/*
* Copyright (c) 2016, Gabor Adam Fodor <fogggab@yahoo.com>
* All rights reserved.
*
* License:
*
* This program is provided for scientific and educational purposed only.
* Feel free to use and/or modify it for such purposes, but you are kindly
* asked not to redistribute this or derivative works in source or executable
* form. A license must be obtained from the author of the code for any other use.
*
*/
#include "FFST.h"
#include <iostream>
#include <fstream>

/**
* Fast finite shearlet tranform implementation following the Matlab implementation
* of the image processing group of TU Kaiserslautern. The matlab code and detailed
* publications on their method can be found at:
* http://www.mathematik.uni-kl.de/imagepro/software/ffst/
* A copy of the same matlab files is also available via the Platypus project at
* http://www.project-platypus.net/cradledownload.html
*
* NOTE: This implementation ONLY works for images of size 512x512, skipping over
* generating the filter-banks on the run to speed up computation speed. Instead,
* all frequency domain filters are stored in a big lookup table that is initalized
* the first time the transform is called.
*
* For a more detailed explanation on the functions in this file, check out the Matlab
* implementation.
**/

namespace FFST{

	std::vector<cv::Mat> Psi;

	cv::Mat inverseShearletTransformSpect(std::vector<cv::Mat> &ST){

		//Only supports 512x512 images
		cv::Mat img(512, 512, CV_32F, cv::Scalar(0));
		
		//Check if Psi is initialized
		if (Psi.size() == 0){
			//Read coefficients
			set_coeffs_512x512();
		}

		cv::Mat dftout, scc, src;
		scc = cv::Mat(512, 512, CV_32F, cv::Scalar(0));
		src = cv::Mat(512, 512, CV_32F, cv::Scalar(0));

		for (int k = 0; k < ST.size(); k++){
			cv::Mat cc, rc;
			cv::dft(ST[k], dftout, cv::DFT_COMPLEX_OUTPUT);

			cc = cv::Mat(512, 512, CV_32F, cv::Scalar(0));
			rc = cv::Mat(512, 512, CV_32F, cv::Scalar(0));

			//Separate complex output
			for (int i = 0; i < cc.rows; i++){
				for (int j = 0; j < cc.cols; j++){
					rc.at<float>(i, j) = dftout.at<float>(i, j * 2);
					cc.at<float>(i, j) = dftout.at<float>(i, j * 2 + 1);
				}
			}

			rc = fftshift(rc);
			cc = fftshift(cc);

			for (int i = 0; i < cc.rows; i++){
				for (int j = 0; j < cc.cols; j++){
					rc.at<float>(i, j) *= Psi[k].at<float>(i, j);
					cc.at<float>(i, j) *= Psi[k].at<float>(i, j);
				}
			}

			src += rc;
			scc += cc;
		}
		
		src = ifftshift(src);
		scc = ifftshift(scc);

		//Put it all together for forward transform
		cv::Mat xx(src.rows, src.cols, CV_32FC2, cv::Scalar(0));
		for (int i = 0; i < xx.rows; i++){
			for (int j = 0; j < xx.cols; j++){
				xx.at<cv::Point2f>(i, j).x = src.at<float>(i, j);	//Real part
				xx.at<cv::Point2f>(i, j).y = scc.at<float>(i, j);	//Imaginary part
			}
		}

		cv::dft(xx, img, cv::DFT_SCALE | cv::DFT_INVERSE | cv::DFT_REAL_OUTPUT);

		return img;
	}

	std::vector<cv::Mat> shearletTransformSpect(cv::Mat &img){
		std::vector<cv::Mat> ST(61);

		if (img.rows != 512 || img.cols != 512){
			std::cout << "Image resolution not supported!" << std::endl;
		}

		//Check if Psi is initialized
		if (Psi.size() == 0){
			//Read coefficients
			set_coeffs_512x512();
		}

		//fftshift(fft2(A))
		cv::Mat dftout, cc, rc;
		cv::dft(img, dftout, cv::DFT_COMPLEX_OUTPUT);

		//Separate complex and real channel
		cc = cv::Mat(img.rows, img.cols, CV_32F, cv::Scalar(0));
		rc = cv::Mat(img.rows, img.cols, CV_32F, cv::Scalar(0));

		for (int i = 0; i < img.rows; i++){
			for (int j = 0; j < img.cols; j++){
				rc.at<float>(i, j) = dftout.at<float>(i, j * 2);
				cc.at<float>(i, j) = dftout.at<float>(i, j * 2 + 1);
			}
		}
		rc = fftshift(rc);
		cc = fftshift(cc);

		for (int k = 0; k < 61; k++){
			cv::Mat lrc, lcc;
			lcc = cv::Mat(cc.rows, cc.cols, CV_32F);
			lrc = cv::Mat(rc.rows, rc.cols, CV_32F);

			for (int i = 0; i < rc.rows; i++){
				for (int j = 0; j < rc.cols; j++){
					// Psi(:,:,k) .* fft2(img)
					lrc.at<float>(i, j) = Psi[k].at<float>(i, j) * rc.at<float>(i, j);
					lcc.at<float>(i, j) = Psi[k].at<float>(i, j) * cc.at<float>(i, j);
				}
			}

			lrc = fftshift(lrc);
			lcc = fftshift(lcc);

			//Put it all together for forward transform
			cv::Mat xx(lrc.rows, lrc.cols, CV_32FC2, cv::Scalar(0));
			for (int i = 0; i < xx.rows; i++){
				for (int j = 0; j < xx.cols; j++){
					xx.at<cv::Point2f>(i, j).x = lrc.at<float>(i, j);	//Real part
					xx.at<cv::Point2f>(i, j).y = lcc.at<float>(i, j);	//Imaginary part
				}
			}

			cv::dft(xx, ST[k], cv::DFT_SCALE | cv::DFT_INVERSE | cv::DFT_REAL_OUTPUT);
		}

		return ST;
	}

	void set_coeffs_512x512(){
		Psi = std::vector<cv::Mat>(61);
		for (int i = 0; i < Psi.size(); i++){
			Psi[i] = cv::Mat(512, 512, CV_32F, cv::Scalar(0));
		}

		for (int i = 0; i<616952; i++){
			int a = coeff_pos[i][0];
			int b = coeff_pos[i][1];
			int c = coeff_pos[i][2];
			float v = coeff_val[i];
			Psi[c - 1].at<float>(a - 1, b - 1) = v;
		}
	}

	cv::Mat fftshift(cv::Mat &x){
		//Inverse FFT shift
		// FFTSHIFT(X) swaps the first and third
		// quadrants and the second and fourth quadrants.For N - D
		// arrays, FFTSHIFT(X) swaps "half-spaces" of X along each
		// dimension.

		return circshift(x, std::floor(x.rows / 2), std::floor(x.cols / 2));
	}

	cv::Mat ifftshift(cv::Mat &x){
		//Inverse FFT shift
		// IFFTSHIFT(X) swaps the first and third
		// quadrants and the second and fourth quadrants.For N - D
		// arrays, IFFTSHIFT(X) swaps "half-spaces" of X along each
		// dimension.

		return circshift(x, std::ceil(x.rows * 1.0 / 2), std::ceil(x.cols * 1.0 / 2));
	}

	cv::Mat circshift(cv::Mat &M, int x, int y){
		// Circularly shifts the values in matrix M
		// by x along the first dimension and by y along the second one

		if ((x == 0) && (y == 0))
			return M;

		int X = M.rows;
		int Y = M.cols;

		cv::Mat res(X, Y, CV_32F, cv::Scalar(0));
		for (int i = 0; i < M.rows; i++){
			for (int j = 0; j < M.cols; j++){
				res.at<float>((i + x + X) % X, (j + y + Y) % Y) = M.at<float>(i, j);
			}
		}
		return res;
	}

	float coeff_val[616952] = {
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.011143,
		0.011143,
		0.036394,
		0.093895,
		0.093895,
		0.036394,
		0.065143,
		0.294361,
		0.181483,
		0.181483,
		0.294361,
		0.065143,
		0.089384,
		0.5,
		0.535021,
		0.265625,
		0.265625,
		0.535021,
		0.5,
		0.089384,
		0.104444,
		0.642924,
		0.837568,
		0.707107,
		0.416289,
		0.416289,
		0.707107,
		0.837568,
		0.642924,
		0.104444,
		0.110041,
		0.700845,
		0.977154,
		0.964076,
		0.909232,
		0.707107,
		0.707107,
		0.909232,
		0.964076,
		0.977154,
		0.700845,
		0.110041,
		0.110603,
		0.707107,
		0.993865,
		1,
		1,
		1,
		1,
		1,
		1,
		0.993865,
		0.707107,
		0.110603,
		0.110041,
		0.700845,
		0.977154,
		0.964076,
		0.909232,
		0.707107,
		0.707107,
		0.909232,
		0.964076,
		0.977154,
		0.700845,
		0.110041,
		0.104444,
		0.642924,
		0.837568,
		0.707107,
		0.416289,
		0.416289,
		0.707107,
		0.837568,
		0.642924,
		0.104444,
		0.089384,
		0.5,
		0.535021,
		0.265625,
		0.265625,
		0.535021,
		0.5,
		0.089384,
		0.065143,
		0.294361,
		0.181483,
		0.181483,
		0.294361,
		0.065143,
		0.036394,
		0.093895,
		0.093895,
		0.036394,
		0.011143,
		0.011143,
		0.011143,
		0.036394,
		0.065143,
		0.089384,
		0.104444,
		0.110041,
		0.110603,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.110041,
		0.181483,
		0.535021,
		0.837568,
		0.977154,
		0.993865,
		0.700845,
		0.104444,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.977154,
		0.642924,
		0.089384,
		0.416289,
		0.909232,
		1,
		0.964076,
		0.837568,
		0.5,
		0.065143,
		0.707107,
		1,
		0.909232,
		0.707107,
		0.535021,
		0.294361,
		0.036394,
		0.707107,
		0.416289,
		0.265625,
		0.181483,
		0.093895,
		0.011143,
		0.011143,
		0.093895,
		0.181483,
		0.265625,
		0.416289,
		0.707107,
		0.036394,
		0.294361,
		0.535021,
		0.707107,
		0.909232,
		1,
		0.707107,
		0.065143,
		0.5,
		0.837568,
		0.964076,
		1,
		0.909232,
		0.416289,
		0.089384,
		0.642924,
		0.977154,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.104444,
		0.700845,
		0.993865,
		0.977154,
		0.837568,
		0.535021,
		0.181483,
		0.110041,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.110603,
		0.110041,
		0.104444,
		0.089384,
		0.065143,
		0.036394,
		0.011143,
		0.011143,
		0.036394,
		0.065143,
		0.089384,
		0.104444,
		0.110041,
		0.110603,
		0.110041,
		0.104444,
		0.089384,
		0.065143,
		0.036394,
		0.011143,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.181483,
		0.535021,
		0.837568,
		0.977154,
		0.993865,
		0.977154,
		0.837568,
		0.535021,
		0.181483,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.416289,
		0.909232,
		1,
		0.909232,
		0.416289,
		0.707107,
		1,
		0.707107,
		0.707107,
		1,
		0.707107,
		0.416289,
		0.909232,
		1,
		0.909232,
		0.416289,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.181483,
		0.535021,
		0.837568,
		0.977154,
		0.993865,
		0.977154,
		0.837568,
		0.535021,
		0.181483,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.011143,
		0.036394,
		0.065143,
		0.089384,
		0.104444,
		0.110041,
		0.110603,
		0.110041,
		0.104444,
		0.089384,
		0.065143,
		0.036394,
		0.011143,
		0.110603,
		0.110041,
		0.104444,
		0.089384,
		0.065143,
		0.036394,
		0.011143,
		0.110041,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.104444,
		0.700845,
		0.993865,
		0.977154,
		0.837568,
		0.535021,
		0.181483,
		0.089384,
		0.642924,
		0.977154,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.065143,
		0.5,
		0.837568,
		0.964076,
		1,
		0.909232,
		0.416289,
		0.036394,
		0.294361,
		0.535021,
		0.707107,
		0.909232,
		1,
		0.707107,
		0.011143,
		0.093895,
		0.181483,
		0.265625,
		0.416289,
		0.707107,
		0.707107,
		0.416289,
		0.265625,
		0.181483,
		0.093895,
		0.011143,
		0.707107,
		1,
		0.909232,
		0.707107,
		0.535021,
		0.294361,
		0.036394,
		0.416289,
		0.909232,
		1,
		0.964076,
		0.837568,
		0.5,
		0.065143,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.977154,
		0.642924,
		0.089384,
		0.181483,
		0.535021,
		0.837568,
		0.977154,
		0.993865,
		0.700845,
		0.104444,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.110041,
		0.011143,
		0.036394,
		0.065143,
		0.089384,
		0.104444,
		0.110041,
		0.110603,
		3.5e-005,
		0.000237,
		0.000283,
		0.000283,
		0.000237,
		3.5e-005,
		9e-005,
		0.000871,
		0.002331,
		0.003057,
		0.001726,
		0.001726,
		0.003057,
		0.002331,
		0.000871,
		9e-005,
		0.000161,
		0.00179,
		0.005905,
		0.011143,
		0.014139,
		0.011859,
		0.004895,
		0.004895,
		0.011859,
		0.014139,
		0.011143,
		0.005905,
		0.00179,
		0.000161,
		0.000243,
		0.00289,
		0.010502,
		0.022723,
		0.035553,
		0.04289,
		0.039786,
		0.026203,
		0.008906,
		0.008906,
		0.026203,
		0.039786,
		0.04289,
		0.035553,
		0.022723,
		0.010502,
		0.00289,
		0.000243,
		0.000328,
		0.00408,
		0.015662,
		0.036394,
		0.062693,
		0.086698,
		0.099147,
		0.093895,
		0.071706,
		0.040531,
		0.01234,
		0.01234,
		0.040531,
		0.071706,
		0.093895,
		0.099147,
		0.086698,
		0.062693,
		0.036394,
		0.015662,
		0.00408,
		0.000328,
		0.000412,
		0.005276,
		0.020976,
		0.0509,
		0.092608,
		0.137498,
		0.173066,
		0.187825,
		0.176778,
		0.143636,
		0.098223,
		0.05191,
		0.015352,
		0.015352,
		0.05191,
		0.098223,
		0.143636,
		0.176778,
		0.187825,
		0.173066,
		0.137498,
		0.092608,
		0.0509,
		0.020976,
		0.005276,
		0.000412,
		0.000491,
		0.006407,
		0.02609,
		0.065143,
		0.122709,
		0.190179,
		0.252728,
		0.294361,
		0.304801,
		0.283695,
		0.238991,
		0.181483,
		0.120501,
		0.063554,
		0.019046,
		0.019046,
		0.063554,
		0.120501,
		0.181483,
		0.238991,
		0.283695,
		0.304801,
		0.294361,
		0.252728,
		0.190179,
		0.122709,
		0.065143,
		0.02609,
		0.006407,
		0.000491,
		0.00056,
		0.007416,
		0.030711,
		0.078208,
		0.150806,
		0.240381,
		0.330552,
		0.401677,
		0.438887,
		0.438158,
		0.405858,
		0.352837,
		0.288436,
		0.218452,
		0.146884,
		0.078987,
		0.024223,
		0.024223,
		0.078987,
		0.146884,
		0.218452,
		0.288436,
		0.352837,
		0.405858,
		0.438158,
		0.438887,
		0.401677,
		0.330552,
		0.240381,
		0.150806,
		0.078208,
		0.030711,
		0.007416,
		0.00056,
		0.000617,
		0.008259,
		0.034615,
		0.089384,
		0.175172,
		0.284601,
		0.400354,
		0.5,
		0.564909,
		0.587947,
		0.574223,
		0.535021,
		0.480503,
		0.416269,
		0.344293,
		0.265625,
		0.182603,
		0.100747,
		0.031834,
		0.031834,
		0.100747,
		0.182603,
		0.265625,
		0.344293,
		0.416269,
		0.480503,
		0.535021,
		0.574223,
		0.587947,
		0.564909,
		0.5,
		0.400354,
		0.284601,
		0.175172,
		0.089384,
		0.034615,
		0.008259,
		0.000617,
		0.000661,
		0.00891,
		0.037661,
		0.098195,
		0.194608,
		0.320332,
		0.457578,
		0.581942,
		0.67194,
		0.71801,
		0.724327,
		0.702768,
		0.664693,
		0.616381,
		0.559361,
		0.492972,
		0.416289,
		0.32905,
		0.232725,
		0.132788,
		0.043674,
		0.043674,
		0.132788,
		0.232725,
		0.32905,
		0.416289,
		0.492972,
		0.559361,
		0.616381,
		0.664693,
		0.702768,
		0.724327,
		0.71801,
		0.67194,
		0.581942,
		0.457578,
		0.320332,
		0.194608,
		0.098195,
		0.037661,
		0.00891,
		0.000661,
		0.000691,
		0.009364,
		0.039801,
		0.104444,
		0.208535,
		0.346229,
		0.499578,
		0.642924,
		0.752845,
		0.818069,
		0.842154,
		0.837568,
		0.816874,
		0.787387,
		0.751076,
		0.707107,
		0.653763,
		0.588982,
		0.510503,
		0.416289,
		0.3057,
		0.182603,
		0.063557,
		0.063557,
		0.182603,
		0.3057,
		0.416289,
		0.510503,
		0.588982,
		0.653763,
		0.707107,
		0.751076,
		0.787387,
		0.816874,
		0.837568,
		0.842154,
		0.818069,
		0.752845,
		0.642924,
		0.499578,
		0.346229,
		0.208535,
		0.104444,
		0.039801,
		0.009364,
		0.000691,
		0.000709,
		0.009635,
		0.041091,
		0.108244,
		0.217088,
		0.362304,
		0.52595,
		0.681705,
		0.805015,
		0.883593,
		0.92065,
		0.929125,
		0.92254,
		0.909189,
		0.891783,
		0.870045,
		0.842738,
		0.808146,
		0.763962,
		0.707107,
		0.63355,
		0.538323,
		0.416289,
		0.265625,
		0.100129,
		0.100129,
		0.265625,
		0.416289,
		0.538323,
		0.63355,
		0.707107,
		0.763962,
		0.808146,
		0.842738,
		0.870045,
		0.891783,
		0.909189,
		0.92254,
		0.929125,
		0.92065,
		0.883593,
		0.805015,
		0.681705,
		0.52595,
		0.362304,
		0.217088,
		0.108244,
		0.041091,
		0.009635,
		0.000709,
		0.000717,
		0.009761,
		0.041695,
		0.110041,
		0.22117,
		0.370055,
		0.53881,
		0.700845,
		0.831105,
		0.916836,
		0.961111,
		0.977154,
		0.979067,
		0.975798,
		0.970681,
		0.964076,
		0.955526,
		0.944313,
		0.929393,
		0.909232,
		0.881522,
		0.842738,
		0.787425,
		0.707107,
		0.585368,
		0.294361,
		0.020197,
		0.020197,
		0.294361,
		0.585368,
		0.707107,
		0.787425,
		0.842738,
		0.881522,
		0.909232,
		0.929393,
		0.944313,
		0.955526,
		0.964076,
		0.970681,
		0.975798,
		0.979067,
		0.977154,
		0.961111,
		0.916836,
		0.831105,
		0.700845,
		0.53881,
		0.370055,
		0.22117,
		0.110041,
		0.041695,
		0.009761,
		0.000717,
		0.000719,
		0.009797,
		0.041868,
		0.110561,
		0.222364,
		0.372344,
		0.542647,
		0.706621,
		0.839077,
		0.927134,
		0.973833,
		0.992508,
		0.997471,
		0.99793,
		0.997501,
		0.996876,
		0.996042,
		0.994912,
		0.99335,
		0.991144,
		0.987949,
		0.983187,
		0.975845,
		0.964076,
		0.938519,
		0.642924,
		0.09321,
		0.09321,
		0.642924,
		0.938519,
		0.964076,
		0.975845,
		0.983187,
		0.987949,
		0.991144,
		0.99335,
		0.994912,
		0.996042,
		0.996876,
		0.997501,
		0.99793,
		0.997471,
		0.992508,
		0.973833,
		0.927134,
		0.839077,
		0.706621,
		0.542647,
		0.372344,
		0.222364,
		0.110561,
		0.041868,
		0.009797,
		0.000719,
		0.000719,
		0.0098,
		0.041882,
		0.110603,
		0.222461,
		0.372533,
		0.542966,
		0.707107,
		0.839755,
		0.928019,
		0.974941,
		0.993865,
		0.999123,
		0.999952,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.993865,
		0.707107,
		0.110603,
		0.110603,
		0.707107,
		0.993865,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999952,
		0.999123,
		0.993865,
		0.974941,
		0.928019,
		0.839755,
		0.707107,
		0.542966,
		0.372533,
		0.222461,
		0.110603,
		0.041882,
		0.0098,
		0.000719,
		0.000719,
		0.009797,
		0.041868,
		0.110561,
		0.222364,
		0.372344,
		0.542647,
		0.706621,
		0.839077,
		0.927134,
		0.973833,
		0.992508,
		0.997471,
		0.99793,
		0.997501,
		0.996876,
		0.996042,
		0.994912,
		0.99335,
		0.991144,
		0.987949,
		0.983187,
		0.975845,
		0.964076,
		0.938519,
		0.642924,
		0.09321,
		0.09321,
		0.642924,
		0.938519,
		0.964076,
		0.975845,
		0.983187,
		0.987949,
		0.991144,
		0.99335,
		0.994912,
		0.996042,
		0.996876,
		0.997501,
		0.99793,
		0.997471,
		0.992508,
		0.973833,
		0.927134,
		0.839077,
		0.706621,
		0.542647,
		0.372344,
		0.222364,
		0.110561,
		0.041868,
		0.009797,
		0.000719,
		0.000717,
		0.009761,
		0.041695,
		0.110041,
		0.22117,
		0.370055,
		0.53881,
		0.700845,
		0.831105,
		0.916836,
		0.961111,
		0.977154,
		0.979067,
		0.975798,
		0.970681,
		0.964076,
		0.955526,
		0.944313,
		0.929393,
		0.909232,
		0.881522,
		0.842738,
		0.787425,
		0.707107,
		0.585368,
		0.294361,
		0.020197,
		0.020197,
		0.294361,
		0.585368,
		0.707107,
		0.787425,
		0.842738,
		0.881522,
		0.909232,
		0.929393,
		0.944313,
		0.955526,
		0.964076,
		0.970681,
		0.975798,
		0.979067,
		0.977154,
		0.961111,
		0.916836,
		0.831105,
		0.700845,
		0.53881,
		0.370055,
		0.22117,
		0.110041,
		0.041695,
		0.009761,
		0.000717,
		0.000709,
		0.009635,
		0.041091,
		0.108244,
		0.217088,
		0.362304,
		0.52595,
		0.681705,
		0.805015,
		0.883593,
		0.92065,
		0.929125,
		0.92254,
		0.909189,
		0.891783,
		0.870045,
		0.842738,
		0.808146,
		0.763962,
		0.707107,
		0.63355,
		0.538323,
		0.416289,
		0.265625,
		0.100129,
		0.100129,
		0.265625,
		0.416289,
		0.538323,
		0.63355,
		0.707107,
		0.763962,
		0.808146,
		0.842738,
		0.870045,
		0.891783,
		0.909189,
		0.92254,
		0.929125,
		0.92065,
		0.883593,
		0.805015,
		0.681705,
		0.52595,
		0.362304,
		0.217088,
		0.108244,
		0.041091,
		0.009635,
		0.000709,
		0.000691,
		0.009364,
		0.039801,
		0.104444,
		0.208535,
		0.346229,
		0.499578,
		0.642924,
		0.752845,
		0.818069,
		0.842154,
		0.837568,
		0.816874,
		0.787387,
		0.751076,
		0.707107,
		0.653763,
		0.588982,
		0.510503,
		0.416289,
		0.3057,
		0.182603,
		0.063557,
		0.063557,
		0.182603,
		0.3057,
		0.416289,
		0.510503,
		0.588982,
		0.653763,
		0.707107,
		0.751076,
		0.787387,
		0.816874,
		0.837568,
		0.842154,
		0.818069,
		0.752845,
		0.642924,
		0.499578,
		0.346229,
		0.208535,
		0.104444,
		0.039801,
		0.009364,
		0.000691,
		0.000661,
		0.00891,
		0.037661,
		0.098195,
		0.194608,
		0.320332,
		0.457578,
		0.581942,
		0.67194,
		0.71801,
		0.724327,
		0.702768,
		0.664693,
		0.616381,
		0.559361,
		0.492972,
		0.416289,
		0.32905,
		0.232725,
		0.132788,
		0.043674,
		0.043674,
		0.132788,
		0.232725,
		0.32905,
		0.416289,
		0.492972,
		0.559361,
		0.616381,
		0.664693,
		0.702768,
		0.724327,
		0.71801,
		0.67194,
		0.581942,
		0.457578,
		0.320332,
		0.194608,
		0.098195,
		0.037661,
		0.00891,
		0.000661,
		0.000617,
		0.008259,
		0.034615,
		0.089384,
		0.175172,
		0.284601,
		0.400354,
		0.5,
		0.564909,
		0.587947,
		0.574223,
		0.535021,
		0.480503,
		0.416269,
		0.344293,
		0.265625,
		0.182603,
		0.100747,
		0.031834,
		0.031834,
		0.100747,
		0.182603,
		0.265625,
		0.344293,
		0.416269,
		0.480503,
		0.535021,
		0.574223,
		0.587947,
		0.564909,
		0.5,
		0.400354,
		0.284601,
		0.175172,
		0.089384,
		0.034615,
		0.008259,
		0.000617,
		0.00056,
		0.007416,
		0.030711,
		0.078208,
		0.150806,
		0.240381,
		0.330552,
		0.401677,
		0.438887,
		0.438158,
		0.405858,
		0.352837,
		0.288436,
		0.218452,
		0.146884,
		0.078987,
		0.024223,
		0.024223,
		0.078987,
		0.146884,
		0.218452,
		0.288436,
		0.352837,
		0.405858,
		0.438158,
		0.438887,
		0.401677,
		0.330552,
		0.240381,
		0.150806,
		0.078208,
		0.030711,
		0.007416,
		0.00056,
		0.000491,
		0.006407,
		0.02609,
		0.065143,
		0.122709,
		0.190179,
		0.252728,
		0.294361,
		0.304801,
		0.283695,
		0.238991,
		0.181483,
		0.120501,
		0.063554,
		0.019046,
		0.019046,
		0.063554,
		0.120501,
		0.181483,
		0.238991,
		0.283695,
		0.304801,
		0.294361,
		0.252728,
		0.190179,
		0.122709,
		0.065143,
		0.02609,
		0.006407,
		0.000491,
		0.000412,
		0.005276,
		0.020976,
		0.0509,
		0.092608,
		0.137498,
		0.173066,
		0.187825,
		0.176778,
		0.143636,
		0.098223,
		0.05191,
		0.015352,
		0.015352,
		0.05191,
		0.098223,
		0.143636,
		0.176778,
		0.187825,
		0.173066,
		0.137498,
		0.092608,
		0.0509,
		0.020976,
		0.005276,
		0.000412,
		0.000328,
		0.00408,
		0.015662,
		0.036394,
		0.062693,
		0.086698,
		0.099147,
		0.093895,
		0.071706,
		0.040531,
		0.01234,
		0.01234,
		0.040531,
		0.071706,
		0.093895,
		0.099147,
		0.086698,
		0.062693,
		0.036394,
		0.015662,
		0.00408,
		0.000328,
		0.000243,
		0.00289,
		0.010502,
		0.022723,
		0.035553,
		0.04289,
		0.039786,
		0.026203,
		0.008906,
		0.008906,
		0.026203,
		0.039786,
		0.04289,
		0.035553,
		0.022723,
		0.010502,
		0.00289,
		0.000243,
		0.000161,
		0.00179,
		0.005905,
		0.011143,
		0.014139,
		0.011859,
		0.004895,
		0.004895,
		0.011859,
		0.014139,
		0.011143,
		0.005905,
		0.00179,
		0.000161,
		9e-005,
		0.000871,
		0.002331,
		0.003057,
		0.001726,
		0.001726,
		0.003057,
		0.002331,
		0.000871,
		9e-005,
		3.5e-005,
		0.000237,
		0.000283,
		0.000283,
		0.000237,
		3.5e-005,
		1.6e-005,
		0.000237,
		6e-005,
		0.001083,
		0.000871,
		0.000124,
		0.003057,
		0.00396,
		0.00179,
		0.000201,
		0.006591,
		0.011143,
		0.008105,
		0.00289,
		0.000285,
		0.011859,
		0.023933,
		0.022723,
		0.013038,
		0.00408,
		0.000371,
		0.018622,
		0.04289,
		0.0486,
		0.036394,
		0.018324,
		0.005276,
		0.000452,
		0.026203,
		0.067052,
		0.086698,
		0.077474,
		0.0509,
		0.023578,
		0.006407,
		0.000526,
		0.033727,
		0.093895,
		0.134865,
		0.137498,
		0.107782,
		0.065143,
		0.028478,
		0.007416,
		0.00059,
		0.040531,
		0.120219,
		0.187825,
		0.212684,
		0.190179,
		0.137128,
		0.078208,
		0.032763,
		0.008259,
		0.00064,
		0.046472,
		0.143636,
		0.239036,
		0.294361,
		0.292291,
		0.240381,
		0.163544,
		0.089384,
		0.036251,
		0.00891,
		0.000677,
		0.05191,
		0.163637,
		0.283695,
		0.372038,
		0.401677,
		0.366781,
		0.284601,
		0.185558,
		0.098195,
		0.038844,
		0.009364,
		0.000701,
		0.057429,
		0.181483,
		0.320804,
		0.438158,
		0.50367,
		0.5,
		0.430753,
		0.320332,
		0.202269,
		0.104444,
		0.040544,
		0.009635,
		0.000714,
		0.063554,
		0.199181,
		0.352837,
		0.491211,
		0.587947,
		0.621342,
		0.581942,
		0.480563,
		0.346229,
		0.213445,
		0.108244,
		0.041464,
		0.009761,
		0.000719,
		0.070648,
		0.218452,
		0.383622,
		0.535021,
		0.652581,
		0.71801,
		0.715937,
		0.642924,
		0.514646,
		0.362304,
		0.219602,
		0.110041,
		0.041817,
		0.009797,
		0.000719,
		0.078987,
		0.24037,
		0.416269,
		0.575293,
		0.702768,
		0.787895,
		0.818069,
		0.782486,
		0.681705,
		0.533837,
		0.370055,
		0.222012,
		0.110561,
		0.041881,
		0.0098,
		0.000719,
		0.088884,
		0.265625,
		0.452462,
		0.616381,
		0.745942,
		0.837568,
		0.886448,
		0.883593,
		0.820937,
		0.700845,
		0.541506,
		0.372344,
		0.222455,
		0.110603,
		0.041881,
		0.009797,
		0.000719,
		0.100747,
		0.294906,
		0.492972,
		0.660215,
		0.787387,
		0.875992,
		0.929125,
		0.945195,
		0.916836,
		0.836688,
		0.706621,
		0.542944,
		0.372533,
		0.222455,
		0.110561,
		0.041817,
		0.009761,
		0.000714,
		0.11513,
		0.32905,
		0.538323,
		0.707107,
		0.828924,
		0.909189,
		0.956583,
		0.977154,
		0.969981,
		0.927134,
		0.839707,
		0.707107,
		0.542944,
		0.372344,
		0.222012,
		0.110041,
		0.041464,
		0.009635,
		0.000701,
		0.132788,
		0.369091,
		0.588982,
		0.756699,
		0.870045,
		0.938862,
		0.975798,
		0.991829,
		0.992508,
		0.974863,
		0.928019,
		0.839707,
		0.706621,
		0.541506,
		0.370055,
		0.219602,
		0.108244,
		0.040544,
		0.009364,
		0.000677,
		0.154777,
		0.416289,
		0.645261,
		0.808146,
		0.909232,
		0.964076,
		0.989153,
		0.99793,
		0.999005,
		0.993865,
		0.974863,
		0.927134,
		0.836688,
		0.700845,
		0.533837,
		0.362304,
		0.213445,
		0.104444,
		0.038844,
		0.00891,
		0.00064,
		0.182603,
		0.472143,
		0.707107,
		0.859876,
		0.944313,
		0.983187,
		0.996876,
		0.999818,
		0.999952,
		0.999005,
		0.992508,
		0.969981,
		0.916836,
		0.820937,
		0.681705,
		0.514646,
		0.346229,
		0.202269,
		0.098195,
		0.036251,
		0.008259,
		0.00059,
		0.218463,
		0.538323,
		0.773702,
		0.909232,
		0.972543,
		0.994912,
		0.999707,
		1,
		0.999818,
		0.99793,
		0.991829,
		0.977154,
		0.945195,
		0.883593,
		0.782486,
		0.642924,
		0.480563,
		0.320332,
		0.185558,
		0.089384,
		0.032763,
		0.007416,
		0.000526,
		0.265625,
		0.61641,
		0.842738,
		0.952128,
		0.991144,
		0.999493,
		1,
		0.999707,
		0.996876,
		0.989153,
		0.975798,
		0.956583,
		0.929125,
		0.886448,
		0.818069,
		0.715937,
		0.581942,
		0.430753,
		0.284601,
		0.163544,
		0.078208,
		0.028478,
		0.006407,
		0.000452,
		0.327031,
		0.707107,
		0.909232,
		0.983187,
		0.999046,
		1,
		0.999493,
		0.994912,
		0.983187,
		0.964076,
		0.938862,
		0.909189,
		0.875992,
		0.837568,
		0.787895,
		0.71801,
		0.621342,
		0.5,
		0.366781,
		0.240381,
		0.137128,
		0.065143,
		0.023578,
		0.005276,
		0.000371,
		0.294361,
		0.803188,
		0.964076,
		0.997978,
		1,
		0.999046,
		0.991144,
		0.972543,
		0.944313,
		0.909232,
		0.870045,
		0.828924,
		0.787387,
		0.745942,
		0.702768,
		0.652581,
		0.587947,
		0.50367,
		0.401677,
		0.292291,
		0.190179,
		0.107782,
		0.0509,
		0.018324,
		0.00408,
		0.000285,
		0.05954,
		0.642924,
		0.988808,
		1,
		0.997978,
		0.983187,
		0.952128,
		0.909232,
		0.859876,
		0.808146,
		0.756699,
		0.707107,
		0.660215,
		0.616381,
		0.575293,
		0.535021,
		0.491211,
		0.438158,
		0.372038,
		0.294361,
		0.212684,
		0.137498,
		0.077474,
		0.036394,
		0.013038,
		0.00289,
		0.000201,
		0.108744,
		0.707107,
		0.988808,
		0.964076,
		0.909232,
		0.842738,
		0.773702,
		0.707107,
		0.645261,
		0.588982,
		0.538323,
		0.492972,
		0.452462,
		0.416269,
		0.383622,
		0.352837,
		0.320804,
		0.283695,
		0.239036,
		0.187825,
		0.134865,
		0.086698,
		0.0486,
		0.022723,
		0.008105,
		0.00179,
		0.000124,
		0.108744,
		0.642924,
		0.803188,
		0.707107,
		0.61641,
		0.538323,
		0.472143,
		0.416289,
		0.369091,
		0.32905,
		0.294906,
		0.265625,
		0.24037,
		0.218452,
		0.199181,
		0.181483,
		0.163637,
		0.143636,
		0.120219,
		0.093895,
		0.067052,
		0.04289,
		0.023933,
		0.011143,
		0.00396,
		0.000871,
		6e-005,
		0.05954,
		0.294361,
		0.327031,
		0.265625,
		0.218463,
		0.182603,
		0.154777,
		0.132788,
		0.11513,
		0.100747,
		0.088884,
		0.078987,
		0.070648,
		0.063554,
		0.057429,
		0.05191,
		0.046472,
		0.040531,
		0.033727,
		0.026203,
		0.018622,
		0.011859,
		0.006591,
		0.003057,
		0.001083,
		0.000237,
		1.6e-005,
		1.6e-005,
		0.000237,
		0.001083,
		0.003057,
		0.006591,
		0.011859,
		0.018622,
		0.026203,
		0.033727,
		0.040531,
		0.046472,
		0.05191,
		0.057429,
		0.063554,
		0.070648,
		0.078987,
		0.088884,
		0.100747,
		0.11513,
		0.132788,
		0.154777,
		0.182603,
		0.218463,
		0.265625,
		0.327031,
		0.294361,
		0.05954,
		6e-005,
		0.000871,
		0.00396,
		0.011143,
		0.023933,
		0.04289,
		0.067052,
		0.093895,
		0.120219,
		0.143636,
		0.163637,
		0.181483,
		0.199181,
		0.218452,
		0.24037,
		0.265625,
		0.294906,
		0.32905,
		0.369091,
		0.416289,
		0.472143,
		0.538323,
		0.61641,
		0.707107,
		0.803188,
		0.642924,
		0.108744,
		0.000124,
		0.00179,
		0.008105,
		0.022723,
		0.0486,
		0.086698,
		0.134865,
		0.187825,
		0.239036,
		0.283695,
		0.320804,
		0.352837,
		0.383622,
		0.416269,
		0.452462,
		0.492972,
		0.538323,
		0.588982,
		0.645261,
		0.707107,
		0.773702,
		0.842738,
		0.909232,
		0.964076,
		0.988808,
		0.707107,
		0.108744,
		0.000201,
		0.00289,
		0.013038,
		0.036394,
		0.077474,
		0.137498,
		0.212684,
		0.294361,
		0.372038,
		0.438158,
		0.491211,
		0.535021,
		0.575293,
		0.616381,
		0.660215,
		0.707107,
		0.756699,
		0.808146,
		0.859876,
		0.909232,
		0.952128,
		0.983187,
		0.997978,
		1,
		0.988808,
		0.642924,
		0.05954,
		0.000285,
		0.00408,
		0.018324,
		0.0509,
		0.107782,
		0.190179,
		0.292291,
		0.401677,
		0.50367,
		0.587947,
		0.652581,
		0.702768,
		0.745942,
		0.787387,
		0.828924,
		0.870045,
		0.909232,
		0.944313,
		0.972543,
		0.991144,
		0.999046,
		1,
		0.997978,
		0.964076,
		0.803188,
		0.294361,
		0.000371,
		0.005276,
		0.023578,
		0.065143,
		0.137128,
		0.240381,
		0.366781,
		0.5,
		0.621342,
		0.71801,
		0.787895,
		0.837568,
		0.875992,
		0.909189,
		0.938862,
		0.964076,
		0.983187,
		0.994912,
		0.999493,
		1,
		0.999046,
		0.983187,
		0.909232,
		0.707107,
		0.327031,
		0.000452,
		0.006407,
		0.028478,
		0.078208,
		0.163544,
		0.284601,
		0.430753,
		0.581942,
		0.715937,
		0.818069,
		0.886448,
		0.929125,
		0.956583,
		0.975798,
		0.989153,
		0.996876,
		0.999707,
		1,
		0.999493,
		0.991144,
		0.952128,
		0.842738,
		0.61641,
		0.265625,
		0.000526,
		0.007416,
		0.032763,
		0.089384,
		0.185558,
		0.320332,
		0.480563,
		0.642924,
		0.782486,
		0.883593,
		0.945195,
		0.977154,
		0.991829,
		0.99793,
		0.999818,
		1,
		0.999707,
		0.994912,
		0.972543,
		0.909232,
		0.773702,
		0.538323,
		0.218463,
		0.00059,
		0.008259,
		0.036251,
		0.098195,
		0.202269,
		0.346229,
		0.514646,
		0.681705,
		0.820937,
		0.916836,
		0.969981,
		0.992508,
		0.999005,
		0.999952,
		0.999818,
		0.996876,
		0.983187,
		0.944313,
		0.859876,
		0.707107,
		0.472143,
		0.182603,
		0.00064,
		0.00891,
		0.038844,
		0.104444,
		0.213445,
		0.362304,
		0.533837,
		0.700845,
		0.836688,
		0.927134,
		0.974863,
		0.993865,
		0.999005,
		0.99793,
		0.989153,
		0.964076,
		0.909232,
		0.808146,
		0.645261,
		0.416289,
		0.154777,
		0.000677,
		0.009364,
		0.040544,
		0.108244,
		0.219602,
		0.370055,
		0.541506,
		0.706621,
		0.839707,
		0.928019,
		0.974863,
		0.992508,
		0.991829,
		0.975798,
		0.938862,
		0.870045,
		0.756699,
		0.588982,
		0.369091,
		0.132788,
		0.000701,
		0.009635,
		0.041464,
		0.110041,
		0.222012,
		0.372344,
		0.542944,
		0.707107,
		0.839707,
		0.927134,
		0.969981,
		0.977154,
		0.956583,
		0.909189,
		0.828924,
		0.707107,
		0.538323,
		0.32905,
		0.11513,
		0.000714,
		0.009761,
		0.041817,
		0.110561,
		0.222455,
		0.372533,
		0.542944,
		0.706621,
		0.836688,
		0.916836,
		0.945195,
		0.929125,
		0.875992,
		0.787387,
		0.660215,
		0.492972,
		0.294906,
		0.100747,
		0.000719,
		0.009797,
		0.041881,
		0.110603,
		0.222455,
		0.372344,
		0.541506,
		0.700845,
		0.820937,
		0.883593,
		0.886448,
		0.837568,
		0.745942,
		0.616381,
		0.452462,
		0.265625,
		0.088884,
		0.000719,
		0.0098,
		0.041881,
		0.110561,
		0.222012,
		0.370055,
		0.533837,
		0.681705,
		0.782486,
		0.818069,
		0.787895,
		0.702768,
		0.575293,
		0.416269,
		0.24037,
		0.078987,
		0.000719,
		0.009797,
		0.041817,
		0.110041,
		0.219602,
		0.362304,
		0.514646,
		0.642924,
		0.715937,
		0.71801,
		0.652581,
		0.535021,
		0.383622,
		0.218452,
		0.070648,
		0.000719,
		0.009761,
		0.041464,
		0.108244,
		0.213445,
		0.346229,
		0.480563,
		0.581942,
		0.621342,
		0.587947,
		0.491211,
		0.352837,
		0.199181,
		0.063554,
		0.000714,
		0.009635,
		0.040544,
		0.104444,
		0.202269,
		0.320332,
		0.430753,
		0.5,
		0.50367,
		0.438158,
		0.320804,
		0.181483,
		0.057429,
		0.000701,
		0.009364,
		0.038844,
		0.098195,
		0.185558,
		0.284601,
		0.366781,
		0.401677,
		0.372038,
		0.283695,
		0.163637,
		0.05191,
		0.000677,
		0.00891,
		0.036251,
		0.089384,
		0.163544,
		0.240381,
		0.292291,
		0.294361,
		0.239036,
		0.143636,
		0.046472,
		0.00064,
		0.008259,
		0.032763,
		0.078208,
		0.137128,
		0.190179,
		0.212684,
		0.187825,
		0.120219,
		0.040531,
		0.00059,
		0.007416,
		0.028478,
		0.065143,
		0.107782,
		0.137498,
		0.134865,
		0.093895,
		0.033727,
		0.000526,
		0.006407,
		0.023578,
		0.0509,
		0.077474,
		0.086698,
		0.067052,
		0.026203,
		0.000452,
		0.005276,
		0.018324,
		0.036394,
		0.0486,
		0.04289,
		0.018622,
		0.000371,
		0.00408,
		0.013038,
		0.022723,
		0.023933,
		0.011859,
		0.000285,
		0.00289,
		0.008105,
		0.011143,
		0.006591,
		0.000201,
		0.00179,
		0.00396,
		0.003057,
		0.000124,
		0.000871,
		0.001083,
		6e-005,
		0.000237,
		1.6e-005,
		3.5e-005,
		9e-005,
		0.000161,
		0.000243,
		0.000328,
		0.000412,
		0.000491,
		0.00056,
		0.000617,
		0.000661,
		0.000691,
		0.000709,
		0.000717,
		0.000719,
		0.000719,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.000719,
		0.000283,
		0.002331,
		0.005905,
		0.010502,
		0.015662,
		0.020976,
		0.02609,
		0.030711,
		0.034615,
		0.037661,
		0.039801,
		0.041091,
		0.041695,
		0.041868,
		0.041882,
		0.009797,
		0.000717,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.041868,
		0.009761,
		0.000709,
		0.001726,
		0.014139,
		0.035553,
		0.062693,
		0.092608,
		0.122709,
		0.150806,
		0.175172,
		0.194608,
		0.208535,
		0.217088,
		0.22117,
		0.222364,
		0.222461,
		0.110561,
		0.041695,
		0.009635,
		0.000691,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.222364,
		0.110041,
		0.041091,
		0.009364,
		0.000661,
		0.004895,
		0.039786,
		0.099147,
		0.173066,
		0.252728,
		0.330552,
		0.400354,
		0.457578,
		0.499578,
		0.52595,
		0.53881,
		0.542647,
		0.542966,
		0.372344,
		0.22117,
		0.108244,
		0.039801,
		0.00891,
		0.000617,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.542647,
		0.370055,
		0.217088,
		0.104444,
		0.037661,
		0.008259,
		0.00056,
		0.008906,
		0.071706,
		0.176778,
		0.304801,
		0.438887,
		0.564909,
		0.67194,
		0.752845,
		0.805015,
		0.831105,
		0.839077,
		0.839755,
		0.706621,
		0.53881,
		0.362304,
		0.208535,
		0.098195,
		0.034615,
		0.007416,
		0.000491,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.839077,
		0.700845,
		0.52595,
		0.346229,
		0.194608,
		0.089384,
		0.030711,
		0.006407,
		0.000412,
		0.01234,
		0.098223,
		0.238991,
		0.405858,
		0.574223,
		0.724327,
		0.842154,
		0.92065,
		0.961111,
		0.973833,
		0.974941,
		0.927134,
		0.831105,
		0.681705,
		0.499578,
		0.320332,
		0.175172,
		0.078208,
		0.02609,
		0.005276,
		0.000328,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.973833,
		0.916836,
		0.805015,
		0.642924,
		0.457578,
		0.284601,
		0.150806,
		0.065143,
		0.020976,
		0.00408,
		0.000243,
		0.015352,
		0.120501,
		0.288436,
		0.480503,
		0.664693,
		0.816874,
		0.92254,
		0.979067,
		0.997471,
		0.999123,
		0.992508,
		0.961111,
		0.883593,
		0.752845,
		0.581942,
		0.400354,
		0.240381,
		0.122709,
		0.0509,
		0.015662,
		0.00289,
		0.000161,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.997471,
		0.977154,
		0.92065,
		0.818069,
		0.67194,
		0.5,
		0.330552,
		0.190179,
		0.092608,
		0.036394,
		0.010502,
		0.00179,
		9e-005,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751076,
		0.891783,
		0.970681,
		0.997501,
		1,
		0.99793,
		0.979067,
		0.929125,
		0.842154,
		0.71801,
		0.564909,
		0.401677,
		0.252728,
		0.137498,
		0.062693,
		0.022723,
		0.005905,
		0.000871,
		3.5e-005,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.997501,
		0.975798,
		0.92254,
		0.837568,
		0.724327,
		0.587947,
		0.438887,
		0.294361,
		0.173066,
		0.086698,
		0.035553,
		0.011143,
		0.002331,
		0.000237,
		0.024223,
		0.182603,
		0.416289,
		0.653763,
		0.842738,
		0.955526,
		0.996042,
		1,
		0.996876,
		0.970681,
		0.909189,
		0.816874,
		0.702768,
		0.574223,
		0.438158,
		0.304801,
		0.187825,
		0.099147,
		0.04289,
		0.014139,
		0.003057,
		0.000283,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.996042,
		0.964076,
		0.891783,
		0.787387,
		0.664693,
		0.535021,
		0.405858,
		0.283695,
		0.176778,
		0.093895,
		0.039786,
		0.011859,
		0.001726,
		0.031834,
		0.232725,
		0.510503,
		0.763962,
		0.929393,
		0.99335,
		1,
		0.994912,
		0.955526,
		0.870045,
		0.751076,
		0.616381,
		0.480503,
		0.352837,
		0.238991,
		0.143636,
		0.071706,
		0.026203,
		0.004895,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.99335,
		0.944313,
		0.842738,
		0.707107,
		0.559361,
		0.416269,
		0.288436,
		0.181483,
		0.098223,
		0.040531,
		0.008906,
		0.043674,
		0.3057,
		0.63355,
		0.881522,
		0.987949,
		1,
		0.991144,
		0.929393,
		0.808146,
		0.653763,
		0.492972,
		0.344293,
		0.218452,
		0.120501,
		0.05191,
		0.01234,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.987949,
		0.909232,
		0.763962,
		0.588982,
		0.416289,
		0.265625,
		0.146884,
		0.063554,
		0.015352,
		0.063557,
		0.416289,
		0.787425,
		0.975845,
		1,
		0.983187,
		0.881522,
		0.707107,
		0.510503,
		0.32905,
		0.182603,
		0.078987,
		0.019046,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.975845,
		0.842738,
		0.63355,
		0.416289,
		0.232725,
		0.100747,
		0.024223,
		0.100129,
		0.585368,
		0.938519,
		0.993865,
		0.964076,
		0.787425,
		0.538323,
		0.3057,
		0.132788,
		0.031834,
		0.294361,
		0.642924,
		0.707107,
		0.938519,
		0.707107,
		0.416289,
		0.182603,
		0.043674,
		0.020197,
		0.09321,
		0.110603,
		0.642924,
		0.585368,
		0.265625,
		0.063557,
		0.09321,
		0.294361,
		0.100129,
		0.020197,
		0.020197,
		0.100129,
		0.294361,
		0.09321,
		0.063557,
		0.265625,
		0.585368,
		0.642924,
		0.110603,
		0.09321,
		0.020197,
		0.043674,
		0.182603,
		0.416289,
		0.707107,
		0.938519,
		0.707107,
		0.642924,
		0.294361,
		0.031834,
		0.132788,
		0.3057,
		0.538323,
		0.787425,
		0.964076,
		0.993865,
		0.938519,
		0.585368,
		0.100129,
		0.024223,
		0.100747,
		0.232725,
		0.416289,
		0.63355,
		0.842738,
		0.975845,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.019046,
		0.078987,
		0.182603,
		0.32905,
		0.510503,
		0.707107,
		0.881522,
		0.983187,
		1,
		0.975845,
		0.787425,
		0.416289,
		0.063557,
		0.015352,
		0.063554,
		0.146884,
		0.265625,
		0.416289,
		0.588982,
		0.763962,
		0.909232,
		0.987949,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.01234,
		0.05191,
		0.120501,
		0.218452,
		0.344293,
		0.492972,
		0.653763,
		0.808146,
		0.929393,
		0.991144,
		1,
		0.987949,
		0.881522,
		0.63355,
		0.3057,
		0.043674,
		0.008906,
		0.040531,
		0.098223,
		0.181483,
		0.288436,
		0.416269,
		0.559361,
		0.707107,
		0.842738,
		0.944313,
		0.99335,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.004895,
		0.026203,
		0.071706,
		0.143636,
		0.238991,
		0.352837,
		0.480503,
		0.616381,
		0.751076,
		0.870045,
		0.955526,
		0.994912,
		1,
		0.99335,
		0.929393,
		0.763962,
		0.510503,
		0.232725,
		0.031834,
		0.001726,
		0.011859,
		0.039786,
		0.093895,
		0.176778,
		0.283695,
		0.405858,
		0.535021,
		0.664693,
		0.787387,
		0.891783,
		0.964076,
		0.996042,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.000283,
		0.003057,
		0.014139,
		0.04289,
		0.099147,
		0.187825,
		0.304801,
		0.438158,
		0.574223,
		0.702768,
		0.816874,
		0.909189,
		0.970681,
		0.996876,
		1,
		0.996042,
		0.955526,
		0.842738,
		0.653763,
		0.416289,
		0.182603,
		0.024223,
		0.000237,
		0.002331,
		0.011143,
		0.035553,
		0.086698,
		0.173066,
		0.294361,
		0.438887,
		0.587947,
		0.724327,
		0.837568,
		0.92254,
		0.975798,
		0.997501,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		3.5e-005,
		0.000871,
		0.005905,
		0.022723,
		0.062693,
		0.137498,
		0.252728,
		0.401677,
		0.564909,
		0.71801,
		0.842154,
		0.929125,
		0.979067,
		0.99793,
		1,
		0.997501,
		0.970681,
		0.891783,
		0.751076,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		9e-005,
		0.00179,
		0.010502,
		0.036394,
		0.092608,
		0.190179,
		0.330552,
		0.5,
		0.67194,
		0.818069,
		0.92065,
		0.977154,
		0.997471,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		0.000161,
		0.00289,
		0.015662,
		0.0509,
		0.122709,
		0.240381,
		0.400354,
		0.581942,
		0.752845,
		0.883593,
		0.961111,
		0.992508,
		0.999123,
		0.997471,
		0.979067,
		0.92254,
		0.816874,
		0.664693,
		0.480503,
		0.288436,
		0.120501,
		0.015352,
		0.000243,
		0.00408,
		0.020976,
		0.065143,
		0.150806,
		0.284601,
		0.457578,
		0.642924,
		0.805015,
		0.916836,
		0.973833,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.000328,
		0.005276,
		0.02609,
		0.078208,
		0.175172,
		0.320332,
		0.499578,
		0.681705,
		0.831105,
		0.927134,
		0.974941,
		0.973833,
		0.961111,
		0.92065,
		0.842154,
		0.724327,
		0.574223,
		0.405858,
		0.238991,
		0.098223,
		0.01234,
		0.000412,
		0.006407,
		0.030711,
		0.089384,
		0.194608,
		0.346229,
		0.52595,
		0.700845,
		0.839077,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.000491,
		0.007416,
		0.034615,
		0.098195,
		0.208535,
		0.362304,
		0.53881,
		0.706621,
		0.839755,
		0.839077,
		0.831105,
		0.805015,
		0.752845,
		0.67194,
		0.564909,
		0.438887,
		0.304801,
		0.176778,
		0.071706,
		0.008906,
		0.00056,
		0.008259,
		0.037661,
		0.104444,
		0.217088,
		0.370055,
		0.542647,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.000617,
		0.00891,
		0.039801,
		0.108244,
		0.22117,
		0.372344,
		0.542966,
		0.542647,
		0.53881,
		0.52595,
		0.499578,
		0.457578,
		0.400354,
		0.330552,
		0.252728,
		0.173066,
		0.099147,
		0.039786,
		0.004895,
		0.000661,
		0.009364,
		0.041091,
		0.110041,
		0.222364,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.000691,
		0.009635,
		0.041695,
		0.110561,
		0.222461,
		0.222364,
		0.22117,
		0.217088,
		0.208535,
		0.194608,
		0.175172,
		0.150806,
		0.122709,
		0.092608,
		0.062693,
		0.035553,
		0.014139,
		0.001726,
		0.000709,
		0.009761,
		0.041868,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.000717,
		0.009797,
		0.041882,
		0.041868,
		0.041695,
		0.041091,
		0.039801,
		0.037661,
		0.034615,
		0.030711,
		0.02609,
		0.020976,
		0.015662,
		0.010502,
		0.005905,
		0.002331,
		0.000283,
		0.000719,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		0.000719,
		0.000719,
		0.000717,
		0.000709,
		0.000691,
		0.000661,
		0.000617,
		0.00056,
		0.000491,
		0.000412,
		0.000328,
		0.000243,
		0.000161,
		9e-005,
		3.5e-005,
		1.6e-005,
		6e-005,
		0.000124,
		0.000201,
		0.000285,
		0.000371,
		0.000452,
		0.000526,
		0.00059,
		0.00064,
		0.000677,
		0.000701,
		0.000714,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000714,
		0.000701,
		0.000677,
		0.00064,
		0.00059,
		0.000526,
		0.000452,
		0.000371,
		0.000285,
		0.000201,
		0.000124,
		6e-005,
		1.6e-005,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		0.001083,
		0.00396,
		0.008105,
		0.013038,
		0.018324,
		0.023578,
		0.028478,
		0.032763,
		0.036251,
		0.038844,
		0.040544,
		0.041464,
		0.041817,
		0.041881,
		0.041881,
		0.041817,
		0.041464,
		0.040544,
		0.038844,
		0.036251,
		0.032763,
		0.028478,
		0.023578,
		0.018324,
		0.013038,
		0.008105,
		0.00396,
		0.001083,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.006591,
		0.023933,
		0.0486,
		0.077474,
		0.107782,
		0.137128,
		0.163544,
		0.185558,
		0.202269,
		0.213445,
		0.219602,
		0.222012,
		0.222455,
		0.222455,
		0.222012,
		0.219602,
		0.213445,
		0.202269,
		0.185558,
		0.163544,
		0.137128,
		0.107782,
		0.077474,
		0.0486,
		0.023933,
		0.006591,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.018622,
		0.067052,
		0.134865,
		0.212684,
		0.292291,
		0.366781,
		0.430753,
		0.480563,
		0.514646,
		0.533837,
		0.541506,
		0.542944,
		0.542944,
		0.541506,
		0.533837,
		0.514646,
		0.480563,
		0.430753,
		0.366781,
		0.292291,
		0.212684,
		0.134865,
		0.067052,
		0.018622,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.033727,
		0.120219,
		0.239036,
		0.372038,
		0.50367,
		0.621342,
		0.715937,
		0.782486,
		0.820937,
		0.836688,
		0.839707,
		0.839707,
		0.836688,
		0.820937,
		0.782486,
		0.715937,
		0.621342,
		0.50367,
		0.372038,
		0.239036,
		0.120219,
		0.033727,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.046472,
		0.163637,
		0.320804,
		0.491211,
		0.652581,
		0.787895,
		0.886448,
		0.945195,
		0.969981,
		0.974863,
		0.974863,
		0.969981,
		0.945195,
		0.886448,
		0.787895,
		0.652581,
		0.491211,
		0.320804,
		0.163637,
		0.046472,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.057429,
		0.199181,
		0.383622,
		0.575293,
		0.745942,
		0.875992,
		0.956583,
		0.991829,
		0.999005,
		0.999005,
		0.991829,
		0.956583,
		0.875992,
		0.745942,
		0.575293,
		0.383622,
		0.199181,
		0.057429,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989153,
		0.999818,
		0.999818,
		0.989153,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.088884,
		0.294906,
		0.538323,
		0.756699,
		0.909232,
		0.983187,
		0.999707,
		0.999707,
		0.983187,
		0.909232,
		0.756699,
		0.538323,
		0.294906,
		0.088884,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.11513,
		0.369091,
		0.645261,
		0.859876,
		0.972543,
		0.999493,
		0.999493,
		0.972543,
		0.859876,
		0.645261,
		0.369091,
		0.11513,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.154777,
		0.472143,
		0.773702,
		0.952128,
		0.999046,
		0.999046,
		0.952128,
		0.773702,
		0.472143,
		0.154777,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.218463,
		0.61641,
		0.909232,
		0.997978,
		0.997978,
		0.909232,
		0.61641,
		0.218463,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.327031,
		0.803188,
		0.988808,
		0.988808,
		0.803188,
		0.327031,
		0.294361,
		0.642924,
		0.707107,
		0.642924,
		0.294361,
		0.05954,
		0.108744,
		0.108744,
		0.05954,
		0.05954,
		0.108744,
		0.108744,
		0.05954,
		0.294361,
		0.642924,
		0.707107,
		0.642924,
		0.294361,
		0.327031,
		0.803188,
		0.988808,
		0.988808,
		0.803188,
		0.327031,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.218463,
		0.61641,
		0.909232,
		0.997978,
		0.997978,
		0.909232,
		0.61641,
		0.218463,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.154777,
		0.472143,
		0.773702,
		0.952128,
		0.999046,
		0.999046,
		0.952128,
		0.773702,
		0.472143,
		0.154777,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.11513,
		0.369091,
		0.645261,
		0.859876,
		0.972543,
		0.999493,
		0.999493,
		0.972543,
		0.859876,
		0.645261,
		0.369091,
		0.11513,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.088884,
		0.294906,
		0.538323,
		0.756699,
		0.909232,
		0.983187,
		0.999707,
		0.999707,
		0.983187,
		0.909232,
		0.756699,
		0.538323,
		0.294906,
		0.088884,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989153,
		0.999818,
		0.999818,
		0.989153,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		0.057429,
		0.199181,
		0.383622,
		0.575293,
		0.745942,
		0.875992,
		0.956583,
		0.991829,
		0.999005,
		0.999005,
		0.991829,
		0.956583,
		0.875992,
		0.745942,
		0.575293,
		0.383622,
		0.199181,
		0.057429,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.046472,
		0.163637,
		0.320804,
		0.491211,
		0.652581,
		0.787895,
		0.886448,
		0.945195,
		0.969981,
		0.974863,
		0.974863,
		0.969981,
		0.945195,
		0.886448,
		0.787895,
		0.652581,
		0.491211,
		0.320804,
		0.163637,
		0.046472,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.033727,
		0.120219,
		0.239036,
		0.372038,
		0.50367,
		0.621342,
		0.715937,
		0.782486,
		0.820937,
		0.836688,
		0.839707,
		0.839707,
		0.836688,
		0.820937,
		0.782486,
		0.715937,
		0.621342,
		0.50367,
		0.372038,
		0.239036,
		0.120219,
		0.033727,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.018622,
		0.067052,
		0.134865,
		0.212684,
		0.292291,
		0.366781,
		0.430753,
		0.480563,
		0.514646,
		0.533837,
		0.541506,
		0.542944,
		0.542944,
		0.541506,
		0.533837,
		0.514646,
		0.480563,
		0.430753,
		0.366781,
		0.292291,
		0.212684,
		0.134865,
		0.067052,
		0.018622,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.006591,
		0.023933,
		0.0486,
		0.077474,
		0.107782,
		0.137128,
		0.163544,
		0.185558,
		0.202269,
		0.213445,
		0.219602,
		0.222012,
		0.222455,
		0.222455,
		0.222012,
		0.219602,
		0.213445,
		0.202269,
		0.185558,
		0.163544,
		0.137128,
		0.107782,
		0.077474,
		0.0486,
		0.023933,
		0.006591,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.001083,
		0.00396,
		0.008105,
		0.013038,
		0.018324,
		0.023578,
		0.028478,
		0.032763,
		0.036251,
		0.038844,
		0.040544,
		0.041464,
		0.041817,
		0.041881,
		0.041881,
		0.041817,
		0.041464,
		0.040544,
		0.038844,
		0.036251,
		0.032763,
		0.028478,
		0.023578,
		0.018324,
		0.013038,
		0.008105,
		0.00396,
		0.001083,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		1.6e-005,
		6e-005,
		0.000124,
		0.000201,
		0.000285,
		0.000371,
		0.000452,
		0.000526,
		0.00059,
		0.00064,
		0.000677,
		0.000701,
		0.000714,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000714,
		0.000701,
		0.000677,
		0.00064,
		0.00059,
		0.000526,
		0.000452,
		0.000371,
		0.000285,
		0.000201,
		0.000124,
		6e-005,
		1.6e-005,
		3.5e-005,
		9e-005,
		0.000161,
		0.000243,
		0.000328,
		0.000412,
		0.000491,
		0.00056,
		0.000617,
		0.000661,
		0.000691,
		0.000709,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000709,
		0.000691,
		0.000661,
		0.000617,
		0.00056,
		0.000491,
		0.000412,
		0.000328,
		0.000243,
		0.000161,
		9e-005,
		3.5e-005,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		0.000283,
		0.002331,
		0.005905,
		0.010502,
		0.015662,
		0.020976,
		0.02609,
		0.030711,
		0.034615,
		0.037661,
		0.039801,
		0.041091,
		0.041695,
		0.041868,
		0.041882,
		0.041868,
		0.041695,
		0.041091,
		0.039801,
		0.037661,
		0.034615,
		0.030711,
		0.02609,
		0.020976,
		0.015662,
		0.010502,
		0.005905,
		0.002331,
		0.000283,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.001726,
		0.014139,
		0.035553,
		0.062693,
		0.092608,
		0.122709,
		0.150806,
		0.175172,
		0.194608,
		0.208535,
		0.217088,
		0.22117,
		0.222364,
		0.222461,
		0.222364,
		0.22117,
		0.217088,
		0.208535,
		0.194608,
		0.175172,
		0.150806,
		0.122709,
		0.092608,
		0.062693,
		0.035553,
		0.014139,
		0.001726,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.004895,
		0.039786,
		0.099147,
		0.173066,
		0.252728,
		0.330552,
		0.400354,
		0.457578,
		0.499578,
		0.52595,
		0.53881,
		0.542647,
		0.542966,
		0.542647,
		0.53881,
		0.52595,
		0.499578,
		0.457578,
		0.400354,
		0.330552,
		0.252728,
		0.173066,
		0.099147,
		0.039786,
		0.004895,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.008906,
		0.071706,
		0.176778,
		0.304801,
		0.438887,
		0.564909,
		0.67194,
		0.752845,
		0.805015,
		0.831105,
		0.839077,
		0.839755,
		0.839077,
		0.831105,
		0.805015,
		0.752845,
		0.67194,
		0.564909,
		0.438887,
		0.304801,
		0.176778,
		0.071706,
		0.008906,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.01234,
		0.098223,
		0.238991,
		0.405858,
		0.574223,
		0.724327,
		0.842154,
		0.92065,
		0.961111,
		0.973833,
		0.974941,
		0.973833,
		0.961111,
		0.92065,
		0.842154,
		0.724327,
		0.574223,
		0.405858,
		0.238991,
		0.098223,
		0.01234,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.015352,
		0.120501,
		0.288436,
		0.480503,
		0.664693,
		0.816874,
		0.92254,
		0.979067,
		0.997471,
		0.999123,
		0.997471,
		0.979067,
		0.92254,
		0.816874,
		0.664693,
		0.480503,
		0.288436,
		0.120501,
		0.015352,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751076,
		0.891783,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891783,
		0.751076,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.024223,
		0.182603,
		0.416289,
		0.653763,
		0.842738,
		0.955526,
		0.996042,
		1,
		0.996042,
		0.955526,
		0.842738,
		0.653763,
		0.416289,
		0.182603,
		0.024223,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.031834,
		0.232725,
		0.510503,
		0.763962,
		0.929393,
		0.99335,
		1,
		0.99335,
		0.929393,
		0.763962,
		0.510503,
		0.232725,
		0.031834,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.043674,
		0.3057,
		0.63355,
		0.881522,
		0.987949,
		1,
		0.987949,
		0.881522,
		0.63355,
		0.3057,
		0.043674,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.063557,
		0.416289,
		0.787425,
		0.975845,
		1,
		0.975845,
		0.787425,
		0.416289,
		0.063557,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.100129,
		0.585368,
		0.938519,
		0.993865,
		0.938519,
		0.585368,
		0.100129,
		0.294361,
		0.642924,
		0.707107,
		0.642924,
		0.294361,
		0.020197,
		0.09321,
		0.110603,
		0.09321,
		0.020197,
		0.020197,
		0.09321,
		0.110603,
		0.09321,
		0.020197,
		0.294361,
		0.642924,
		0.707107,
		0.642924,
		0.294361,
		0.100129,
		0.585368,
		0.938519,
		0.993865,
		0.938519,
		0.585368,
		0.100129,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.063557,
		0.416289,
		0.787425,
		0.975845,
		1,
		0.975845,
		0.787425,
		0.416289,
		0.063557,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.043674,
		0.3057,
		0.63355,
		0.881522,
		0.987949,
		1,
		0.987949,
		0.881522,
		0.63355,
		0.3057,
		0.043674,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.031834,
		0.232725,
		0.510503,
		0.763962,
		0.929393,
		0.99335,
		1,
		0.99335,
		0.929393,
		0.763962,
		0.510503,
		0.232725,
		0.031834,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.024223,
		0.182603,
		0.416289,
		0.653763,
		0.842738,
		0.955526,
		0.996042,
		1,
		0.996042,
		0.955526,
		0.842738,
		0.653763,
		0.416289,
		0.182603,
		0.024223,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751076,
		0.891783,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891783,
		0.751076,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		0.015352,
		0.120501,
		0.288436,
		0.480503,
		0.664693,
		0.816874,
		0.92254,
		0.979067,
		0.997471,
		0.999123,
		0.997471,
		0.979067,
		0.92254,
		0.816874,
		0.664693,
		0.480503,
		0.288436,
		0.120501,
		0.015352,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.01234,
		0.098223,
		0.238991,
		0.405858,
		0.574223,
		0.724327,
		0.842154,
		0.92065,
		0.961111,
		0.973833,
		0.974941,
		0.973833,
		0.961111,
		0.92065,
		0.842154,
		0.724327,
		0.574223,
		0.405858,
		0.238991,
		0.098223,
		0.01234,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.008906,
		0.071706,
		0.176778,
		0.304801,
		0.438887,
		0.564909,
		0.67194,
		0.752845,
		0.805015,
		0.831105,
		0.839077,
		0.839755,
		0.839077,
		0.831105,
		0.805015,
		0.752845,
		0.67194,
		0.564909,
		0.438887,
		0.304801,
		0.176778,
		0.071706,
		0.008906,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.004895,
		0.039786,
		0.099147,
		0.173066,
		0.252728,
		0.330552,
		0.400354,
		0.457578,
		0.499578,
		0.52595,
		0.53881,
		0.542647,
		0.542966,
		0.542647,
		0.53881,
		0.52595,
		0.499578,
		0.457578,
		0.400354,
		0.330552,
		0.252728,
		0.173066,
		0.099147,
		0.039786,
		0.004895,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.001726,
		0.014139,
		0.035553,
		0.062693,
		0.092608,
		0.122709,
		0.150806,
		0.175172,
		0.194608,
		0.208535,
		0.217088,
		0.22117,
		0.222364,
		0.222461,
		0.222364,
		0.22117,
		0.217088,
		0.208535,
		0.194608,
		0.175172,
		0.150806,
		0.122709,
		0.092608,
		0.062693,
		0.035553,
		0.014139,
		0.001726,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.000283,
		0.002331,
		0.005905,
		0.010502,
		0.015662,
		0.020976,
		0.02609,
		0.030711,
		0.034615,
		0.037661,
		0.039801,
		0.041091,
		0.041695,
		0.041868,
		0.041882,
		0.041868,
		0.041695,
		0.041091,
		0.039801,
		0.037661,
		0.034615,
		0.030711,
		0.02609,
		0.020976,
		0.015662,
		0.010502,
		0.005905,
		0.002331,
		0.000283,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		3.5e-005,
		9e-005,
		0.000161,
		0.000243,
		0.000328,
		0.000412,
		0.000491,
		0.00056,
		0.000617,
		0.000661,
		0.000691,
		0.000709,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000709,
		0.000691,
		0.000661,
		0.000617,
		0.00056,
		0.000491,
		0.000412,
		0.000328,
		0.000243,
		0.000161,
		9e-005,
		3.5e-005,
		1.6e-005,
		6e-005,
		0.000124,
		0.000201,
		0.000285,
		0.000371,
		0.000452,
		0.000526,
		0.00059,
		0.00064,
		0.000677,
		0.000701,
		0.000714,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000714,
		0.000701,
		0.000677,
		0.00064,
		0.00059,
		0.000526,
		0.000452,
		0.000371,
		0.000285,
		0.000201,
		0.000124,
		6e-005,
		1.6e-005,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		0.001083,
		0.00396,
		0.008105,
		0.013038,
		0.018324,
		0.023578,
		0.028478,
		0.032763,
		0.036251,
		0.038844,
		0.040544,
		0.041464,
		0.041817,
		0.041881,
		0.041881,
		0.041817,
		0.041464,
		0.040544,
		0.038844,
		0.036251,
		0.032763,
		0.028478,
		0.023578,
		0.018324,
		0.013038,
		0.008105,
		0.00396,
		0.001083,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.006591,
		0.023933,
		0.0486,
		0.077474,
		0.107782,
		0.137128,
		0.163544,
		0.185558,
		0.202269,
		0.213445,
		0.219602,
		0.222012,
		0.222455,
		0.222455,
		0.222012,
		0.219602,
		0.213445,
		0.202269,
		0.185558,
		0.163544,
		0.137128,
		0.107782,
		0.077474,
		0.0486,
		0.023933,
		0.006591,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.018622,
		0.067052,
		0.134865,
		0.212684,
		0.292291,
		0.366781,
		0.430753,
		0.480563,
		0.514646,
		0.533837,
		0.541506,
		0.542944,
		0.542944,
		0.541506,
		0.533837,
		0.514646,
		0.480563,
		0.430753,
		0.366781,
		0.292291,
		0.212684,
		0.134865,
		0.067052,
		0.018622,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.033727,
		0.120219,
		0.239036,
		0.372038,
		0.50367,
		0.621342,
		0.715937,
		0.782486,
		0.820937,
		0.836688,
		0.839707,
		0.839707,
		0.836688,
		0.820937,
		0.782486,
		0.715937,
		0.621342,
		0.50367,
		0.372038,
		0.239036,
		0.120219,
		0.033727,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.046472,
		0.163637,
		0.320804,
		0.491211,
		0.652581,
		0.787895,
		0.886448,
		0.945195,
		0.969981,
		0.974863,
		0.974863,
		0.969981,
		0.945195,
		0.886448,
		0.787895,
		0.652581,
		0.491211,
		0.320804,
		0.163637,
		0.046472,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.057429,
		0.199181,
		0.383622,
		0.575293,
		0.745942,
		0.875992,
		0.956583,
		0.991829,
		0.999005,
		0.999005,
		0.991829,
		0.956583,
		0.875992,
		0.745942,
		0.575293,
		0.383622,
		0.199181,
		0.057429,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989153,
		0.999818,
		0.999818,
		0.989153,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.088884,
		0.294906,
		0.538323,
		0.756699,
		0.909232,
		0.983187,
		0.999707,
		0.999707,
		0.983187,
		0.909232,
		0.756699,
		0.538323,
		0.294906,
		0.088884,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.11513,
		0.369091,
		0.645261,
		0.859876,
		0.972543,
		0.999493,
		0.999493,
		0.972543,
		0.859876,
		0.645261,
		0.369091,
		0.11513,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.154777,
		0.472143,
		0.773702,
		0.952128,
		0.999046,
		0.999046,
		0.952128,
		0.773702,
		0.472143,
		0.154777,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.218463,
		0.61641,
		0.909232,
		0.997978,
		0.997978,
		0.909232,
		0.61641,
		0.218463,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.327031,
		0.803188,
		0.988808,
		0.988808,
		0.803188,
		0.327031,
		0.294361,
		0.642924,
		0.707107,
		0.642924,
		0.294361,
		0.05954,
		0.108744,
		0.108744,
		0.05954,
		0.05954,
		0.108744,
		0.108744,
		0.05954,
		0.294361,
		0.642924,
		0.707107,
		0.642924,
		0.294361,
		0.327031,
		0.803188,
		0.988808,
		0.988808,
		0.803188,
		0.327031,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.218463,
		0.61641,
		0.909232,
		0.997978,
		0.997978,
		0.909232,
		0.61641,
		0.218463,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.154777,
		0.472143,
		0.773702,
		0.952128,
		0.999046,
		0.999046,
		0.952128,
		0.773702,
		0.472143,
		0.154777,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.11513,
		0.369091,
		0.645261,
		0.859876,
		0.972543,
		0.999493,
		0.999493,
		0.972543,
		0.859876,
		0.645261,
		0.369091,
		0.11513,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.088884,
		0.294906,
		0.538323,
		0.756699,
		0.909232,
		0.983187,
		0.999707,
		0.999707,
		0.983187,
		0.909232,
		0.756699,
		0.538323,
		0.294906,
		0.088884,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989153,
		0.999818,
		0.999818,
		0.989153,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		0.057429,
		0.199181,
		0.383622,
		0.575293,
		0.745942,
		0.875992,
		0.956583,
		0.991829,
		0.999005,
		0.999005,
		0.991829,
		0.956583,
		0.875992,
		0.745942,
		0.575293,
		0.383622,
		0.199181,
		0.057429,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.046472,
		0.163637,
		0.320804,
		0.491211,
		0.652581,
		0.787895,
		0.886448,
		0.945195,
		0.969981,
		0.974863,
		0.974863,
		0.969981,
		0.945195,
		0.886448,
		0.787895,
		0.652581,
		0.491211,
		0.320804,
		0.163637,
		0.046472,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.033727,
		0.120219,
		0.239036,
		0.372038,
		0.50367,
		0.621342,
		0.715937,
		0.782486,
		0.820937,
		0.836688,
		0.839707,
		0.839707,
		0.836688,
		0.820937,
		0.782486,
		0.715937,
		0.621342,
		0.50367,
		0.372038,
		0.239036,
		0.120219,
		0.033727,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.018622,
		0.067052,
		0.134865,
		0.212684,
		0.292291,
		0.366781,
		0.430753,
		0.480563,
		0.514646,
		0.533837,
		0.541506,
		0.542944,
		0.542944,
		0.541506,
		0.533837,
		0.514646,
		0.480563,
		0.430753,
		0.366781,
		0.292291,
		0.212684,
		0.134865,
		0.067052,
		0.018622,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.006591,
		0.023933,
		0.0486,
		0.077474,
		0.107782,
		0.137128,
		0.163544,
		0.185558,
		0.202269,
		0.213445,
		0.219602,
		0.222012,
		0.222455,
		0.222455,
		0.222012,
		0.219602,
		0.213445,
		0.202269,
		0.185558,
		0.163544,
		0.137128,
		0.107782,
		0.077474,
		0.0486,
		0.023933,
		0.006591,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.001083,
		0.00396,
		0.008105,
		0.013038,
		0.018324,
		0.023578,
		0.028478,
		0.032763,
		0.036251,
		0.038844,
		0.040544,
		0.041464,
		0.041817,
		0.041881,
		0.041881,
		0.041817,
		0.041464,
		0.040544,
		0.038844,
		0.036251,
		0.032763,
		0.028478,
		0.023578,
		0.018324,
		0.013038,
		0.008105,
		0.00396,
		0.001083,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		1.6e-005,
		6e-005,
		0.000124,
		0.000201,
		0.000285,
		0.000371,
		0.000452,
		0.000526,
		0.00059,
		0.00064,
		0.000677,
		0.000701,
		0.000714,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000714,
		0.000701,
		0.000677,
		0.00064,
		0.00059,
		0.000526,
		0.000452,
		0.000371,
		0.000285,
		0.000201,
		0.000124,
		6e-005,
		1.6e-005,
		0.000719,
		0.000719,
		0.000717,
		0.000709,
		0.000691,
		0.000661,
		0.000617,
		0.00056,
		0.000491,
		0.000412,
		0.000328,
		0.000243,
		0.000161,
		9e-005,
		3.5e-005,
		0.000719,
		0.0098,
		0.009797,
		0.009761,
		0.009635,
		0.009364,
		0.00891,
		0.008259,
		0.007416,
		0.006407,
		0.005276,
		0.00408,
		0.00289,
		0.00179,
		0.000871,
		0.000237,
		0.000717,
		0.009797,
		0.041882,
		0.041868,
		0.041695,
		0.041091,
		0.039801,
		0.037661,
		0.034615,
		0.030711,
		0.02609,
		0.020976,
		0.015662,
		0.010502,
		0.005905,
		0.002331,
		0.000283,
		0.000709,
		0.009761,
		0.041868,
		0.110603,
		0.110561,
		0.110041,
		0.108244,
		0.104444,
		0.098195,
		0.089384,
		0.078208,
		0.065143,
		0.0509,
		0.036394,
		0.022723,
		0.011143,
		0.003057,
		0.000691,
		0.009635,
		0.041695,
		0.110561,
		0.222461,
		0.222364,
		0.22117,
		0.217088,
		0.208535,
		0.194608,
		0.175172,
		0.150806,
		0.122709,
		0.092608,
		0.062693,
		0.035553,
		0.014139,
		0.001726,
		0.000661,
		0.009364,
		0.041091,
		0.110041,
		0.222364,
		0.372533,
		0.372344,
		0.370055,
		0.362304,
		0.346229,
		0.320332,
		0.284601,
		0.240381,
		0.190179,
		0.137498,
		0.086698,
		0.04289,
		0.011859,
		0.000617,
		0.00891,
		0.039801,
		0.108244,
		0.22117,
		0.372344,
		0.542966,
		0.542647,
		0.53881,
		0.52595,
		0.499578,
		0.457578,
		0.400354,
		0.330552,
		0.252728,
		0.173066,
		0.099147,
		0.039786,
		0.004895,
		0.00056,
		0.008259,
		0.037661,
		0.104444,
		0.217088,
		0.370055,
		0.542647,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.000491,
		0.007416,
		0.034615,
		0.098195,
		0.208535,
		0.362304,
		0.53881,
		0.706621,
		0.839755,
		0.839077,
		0.831105,
		0.805015,
		0.752845,
		0.67194,
		0.564909,
		0.438887,
		0.304801,
		0.176778,
		0.071706,
		0.008906,
		0.000412,
		0.006407,
		0.030711,
		0.089384,
		0.194608,
		0.346229,
		0.52595,
		0.700845,
		0.839077,
		0.928019,
		0.927134,
		0.916836,
		0.883593,
		0.818069,
		0.71801,
		0.587947,
		0.438158,
		0.283695,
		0.143636,
		0.040531,
		0.000328,
		0.005276,
		0.02609,
		0.078208,
		0.175172,
		0.320332,
		0.499578,
		0.681705,
		0.831105,
		0.927134,
		0.974941,
		0.973833,
		0.961111,
		0.92065,
		0.842154,
		0.724327,
		0.574223,
		0.405858,
		0.238991,
		0.098223,
		0.01234,
		0.000243,
		0.00408,
		0.020976,
		0.065143,
		0.150806,
		0.284601,
		0.457578,
		0.642924,
		0.805015,
		0.916836,
		0.973833,
		0.993865,
		0.992508,
		0.977154,
		0.929125,
		0.837568,
		0.702768,
		0.535021,
		0.352837,
		0.181483,
		0.05191,
		0.000161,
		0.00289,
		0.015662,
		0.0509,
		0.122709,
		0.240381,
		0.400354,
		0.581942,
		0.752845,
		0.883593,
		0.961111,
		0.992508,
		0.999123,
		0.997471,
		0.979067,
		0.92254,
		0.816874,
		0.664693,
		0.480503,
		0.288436,
		0.120501,
		0.015352,
		9e-005,
		0.00179,
		0.010502,
		0.036394,
		0.092608,
		0.190179,
		0.330552,
		0.5,
		0.67194,
		0.818069,
		0.92065,
		0.977154,
		0.997471,
		0.999952,
		0.99793,
		0.975798,
		0.909189,
		0.787387,
		0.616381,
		0.416269,
		0.218452,
		0.063554,
		3.5e-005,
		0.000871,
		0.005905,
		0.022723,
		0.062693,
		0.137498,
		0.252728,
		0.401677,
		0.564909,
		0.71801,
		0.842154,
		0.929125,
		0.979067,
		0.99793,
		1,
		0.997501,
		0.970681,
		0.891783,
		0.751076,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.000237,
		0.002331,
		0.011143,
		0.035553,
		0.086698,
		0.173066,
		0.294361,
		0.438887,
		0.587947,
		0.724327,
		0.837568,
		0.92254,
		0.975798,
		0.997501,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.000283,
		0.003057,
		0.014139,
		0.04289,
		0.099147,
		0.187825,
		0.304801,
		0.438158,
		0.574223,
		0.702768,
		0.816874,
		0.909189,
		0.970681,
		0.996876,
		1,
		0.996042,
		0.955526,
		0.842738,
		0.653763,
		0.416289,
		0.182603,
		0.024223,
		0.001726,
		0.011859,
		0.039786,
		0.093895,
		0.176778,
		0.283695,
		0.405858,
		0.535021,
		0.664693,
		0.787387,
		0.891783,
		0.964076,
		0.996042,
		1,
		0.994912,
		0.944313,
		0.808146,
		0.588982,
		0.32905,
		0.100747,
		0.004895,
		0.026203,
		0.071706,
		0.143636,
		0.238991,
		0.352837,
		0.480503,
		0.616381,
		0.751076,
		0.870045,
		0.955526,
		0.994912,
		1,
		0.99335,
		0.929393,
		0.763962,
		0.510503,
		0.232725,
		0.031834,
		0.008906,
		0.040531,
		0.098223,
		0.181483,
		0.288436,
		0.416269,
		0.559361,
		0.707107,
		0.842738,
		0.944313,
		0.99335,
		1,
		0.991144,
		0.909232,
		0.707107,
		0.416289,
		0.132788,
		0.01234,
		0.05191,
		0.120501,
		0.218452,
		0.344293,
		0.492972,
		0.653763,
		0.808146,
		0.929393,
		0.991144,
		1,
		0.987949,
		0.881522,
		0.63355,
		0.3057,
		0.043674,
		0.015352,
		0.063554,
		0.146884,
		0.265625,
		0.416289,
		0.588982,
		0.763962,
		0.909232,
		0.987949,
		1,
		0.983187,
		0.842738,
		0.538323,
		0.182603,
		0.019046,
		0.078987,
		0.182603,
		0.32905,
		0.510503,
		0.707107,
		0.881522,
		0.983187,
		1,
		0.975845,
		0.787425,
		0.416289,
		0.063557,
		0.024223,
		0.100747,
		0.232725,
		0.416289,
		0.63355,
		0.842738,
		0.975845,
		1,
		0.964076,
		0.707107,
		0.265625,
		0.031834,
		0.132788,
		0.3057,
		0.538323,
		0.787425,
		0.964076,
		0.993865,
		0.938519,
		0.585368,
		0.100129,
		0.043674,
		0.182603,
		0.416289,
		0.707107,
		0.938519,
		0.707107,
		0.642924,
		0.294361,
		0.063557,
		0.265625,
		0.585368,
		0.642924,
		0.110603,
		0.09321,
		0.020197,
		0.100129,
		0.294361,
		0.09321,
		0.020197,
		0.020197,
		0.09321,
		0.294361,
		0.100129,
		0.020197,
		0.09321,
		0.110603,
		0.642924,
		0.585368,
		0.265625,
		0.063557,
		0.294361,
		0.642924,
		0.707107,
		0.938519,
		0.707107,
		0.416289,
		0.182603,
		0.043674,
		0.100129,
		0.585368,
		0.938519,
		0.993865,
		0.964076,
		0.787425,
		0.538323,
		0.3057,
		0.132788,
		0.031834,
		0.265625,
		0.707107,
		0.964076,
		1,
		0.975845,
		0.842738,
		0.63355,
		0.416289,
		0.232725,
		0.100747,
		0.024223,
		0.063557,
		0.416289,
		0.787425,
		0.975845,
		1,
		0.983187,
		0.881522,
		0.707107,
		0.510503,
		0.32905,
		0.182603,
		0.078987,
		0.019046,
		0.182603,
		0.538323,
		0.842738,
		0.983187,
		1,
		0.987949,
		0.909232,
		0.763962,
		0.588982,
		0.416289,
		0.265625,
		0.146884,
		0.063554,
		0.015352,
		0.043674,
		0.3057,
		0.63355,
		0.881522,
		0.987949,
		1,
		0.991144,
		0.929393,
		0.808146,
		0.653763,
		0.492972,
		0.344293,
		0.218452,
		0.120501,
		0.05191,
		0.01234,
		0.132788,
		0.416289,
		0.707107,
		0.909232,
		0.991144,
		1,
		0.99335,
		0.944313,
		0.842738,
		0.707107,
		0.559361,
		0.416269,
		0.288436,
		0.181483,
		0.098223,
		0.040531,
		0.008906,
		0.031834,
		0.232725,
		0.510503,
		0.763962,
		0.929393,
		0.99335,
		1,
		0.994912,
		0.955526,
		0.870045,
		0.751076,
		0.616381,
		0.480503,
		0.352837,
		0.238991,
		0.143636,
		0.071706,
		0.026203,
		0.004895,
		0.100747,
		0.32905,
		0.588982,
		0.808146,
		0.944313,
		0.994912,
		1,
		0.996042,
		0.964076,
		0.891783,
		0.787387,
		0.664693,
		0.535021,
		0.405858,
		0.283695,
		0.176778,
		0.093895,
		0.039786,
		0.011859,
		0.001726,
		0.024223,
		0.182603,
		0.416289,
		0.653763,
		0.842738,
		0.955526,
		0.996042,
		1,
		0.996876,
		0.970681,
		0.909189,
		0.816874,
		0.702768,
		0.574223,
		0.438158,
		0.304801,
		0.187825,
		0.099147,
		0.04289,
		0.014139,
		0.003057,
		0.000283,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.997501,
		0.975798,
		0.92254,
		0.837568,
		0.724327,
		0.587947,
		0.438887,
		0.294361,
		0.173066,
		0.086698,
		0.035553,
		0.011143,
		0.002331,
		0.000237,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751076,
		0.891783,
		0.970681,
		0.997501,
		1,
		0.99793,
		0.979067,
		0.929125,
		0.842154,
		0.71801,
		0.564909,
		0.401677,
		0.252728,
		0.137498,
		0.062693,
		0.022723,
		0.005905,
		0.000871,
		3.5e-005,
		0.063554,
		0.218452,
		0.416269,
		0.616381,
		0.787387,
		0.909189,
		0.975798,
		0.99793,
		0.999952,
		0.997471,
		0.977154,
		0.92065,
		0.818069,
		0.67194,
		0.5,
		0.330552,
		0.190179,
		0.092608,
		0.036394,
		0.010502,
		0.00179,
		9e-005,
		0.015352,
		0.120501,
		0.288436,
		0.480503,
		0.664693,
		0.816874,
		0.92254,
		0.979067,
		0.997471,
		0.999123,
		0.992508,
		0.961111,
		0.883593,
		0.752845,
		0.581942,
		0.400354,
		0.240381,
		0.122709,
		0.0509,
		0.015662,
		0.00289,
		0.000161,
		0.05191,
		0.181483,
		0.352837,
		0.535021,
		0.702768,
		0.837568,
		0.929125,
		0.977154,
		0.992508,
		0.993865,
		0.973833,
		0.916836,
		0.805015,
		0.642924,
		0.457578,
		0.284601,
		0.150806,
		0.065143,
		0.020976,
		0.00408,
		0.000243,
		0.01234,
		0.098223,
		0.238991,
		0.405858,
		0.574223,
		0.724327,
		0.842154,
		0.92065,
		0.961111,
		0.973833,
		0.974941,
		0.927134,
		0.831105,
		0.681705,
		0.499578,
		0.320332,
		0.175172,
		0.078208,
		0.02609,
		0.005276,
		0.000328,
		0.040531,
		0.143636,
		0.283695,
		0.438158,
		0.587947,
		0.71801,
		0.818069,
		0.883593,
		0.916836,
		0.927134,
		0.928019,
		0.839077,
		0.700845,
		0.52595,
		0.346229,
		0.194608,
		0.089384,
		0.030711,
		0.006407,
		0.000412,
		0.008906,
		0.071706,
		0.176778,
		0.304801,
		0.438887,
		0.564909,
		0.67194,
		0.752845,
		0.805015,
		0.831105,
		0.839077,
		0.839755,
		0.706621,
		0.53881,
		0.362304,
		0.208535,
		0.098195,
		0.034615,
		0.007416,
		0.000491,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.542647,
		0.370055,
		0.217088,
		0.104444,
		0.037661,
		0.008259,
		0.00056,
		0.004895,
		0.039786,
		0.099147,
		0.173066,
		0.252728,
		0.330552,
		0.400354,
		0.457578,
		0.499578,
		0.52595,
		0.53881,
		0.542647,
		0.542966,
		0.372344,
		0.22117,
		0.108244,
		0.039801,
		0.00891,
		0.000617,
		0.011859,
		0.04289,
		0.086698,
		0.137498,
		0.190179,
		0.240381,
		0.284601,
		0.320332,
		0.346229,
		0.362304,
		0.370055,
		0.372344,
		0.372533,
		0.222364,
		0.110041,
		0.041091,
		0.009364,
		0.000661,
		0.001726,
		0.014139,
		0.035553,
		0.062693,
		0.092608,
		0.122709,
		0.150806,
		0.175172,
		0.194608,
		0.208535,
		0.217088,
		0.22117,
		0.222364,
		0.222461,
		0.110561,
		0.041695,
		0.009635,
		0.000691,
		0.003057,
		0.011143,
		0.022723,
		0.036394,
		0.0509,
		0.065143,
		0.078208,
		0.089384,
		0.098195,
		0.104444,
		0.108244,
		0.110041,
		0.110561,
		0.110603,
		0.041868,
		0.009761,
		0.000709,
		0.000283,
		0.002331,
		0.005905,
		0.010502,
		0.015662,
		0.020976,
		0.02609,
		0.030711,
		0.034615,
		0.037661,
		0.039801,
		0.041091,
		0.041695,
		0.041868,
		0.041882,
		0.009797,
		0.000717,
		0.000237,
		0.000871,
		0.00179,
		0.00289,
		0.00408,
		0.005276,
		0.006407,
		0.007416,
		0.008259,
		0.00891,
		0.009364,
		0.009635,
		0.009761,
		0.009797,
		0.0098,
		0.000719,
		3.5e-005,
		9e-005,
		0.000161,
		0.000243,
		0.000328,
		0.000412,
		0.000491,
		0.00056,
		0.000617,
		0.000661,
		0.000691,
		0.000709,
		0.000717,
		0.000719,
		0.000719,
		1.6e-005,
		6e-005,
		0.000237,
		0.000124,
		0.000871,
		0.001083,
		0.000201,
		0.00179,
		0.00396,
		0.003057,
		0.000285,
		0.00289,
		0.008105,
		0.011143,
		0.006591,
		0.000371,
		0.00408,
		0.013038,
		0.022723,
		0.023933,
		0.011859,
		0.000452,
		0.005276,
		0.018324,
		0.036394,
		0.0486,
		0.04289,
		0.018622,
		0.000526,
		0.006407,
		0.023578,
		0.0509,
		0.077474,
		0.086698,
		0.067052,
		0.026203,
		0.00059,
		0.007416,
		0.028478,
		0.065143,
		0.107782,
		0.137498,
		0.134865,
		0.093895,
		0.033727,
		0.00064,
		0.008259,
		0.032763,
		0.078208,
		0.137128,
		0.190179,
		0.212684,
		0.187825,
		0.120219,
		0.040531,
		0.000677,
		0.00891,
		0.036251,
		0.089384,
		0.163544,
		0.240381,
		0.292291,
		0.294361,
		0.239036,
		0.143636,
		0.046472,
		0.000701,
		0.009364,
		0.038844,
		0.098195,
		0.185558,
		0.284601,
		0.366781,
		0.401677,
		0.372038,
		0.283695,
		0.163637,
		0.05191,
		0.000714,
		0.009635,
		0.040544,
		0.104444,
		0.202269,
		0.320332,
		0.430753,
		0.5,
		0.50367,
		0.438158,
		0.320804,
		0.181483,
		0.057429,
		0.000719,
		0.009761,
		0.041464,
		0.108244,
		0.213445,
		0.346229,
		0.480563,
		0.581942,
		0.621342,
		0.587947,
		0.491211,
		0.352837,
		0.199181,
		0.063554,
		0.000719,
		0.009797,
		0.041817,
		0.110041,
		0.219602,
		0.362304,
		0.514646,
		0.642924,
		0.715937,
		0.71801,
		0.652581,
		0.535021,
		0.383622,
		0.218452,
		0.070648,
		0.000719,
		0.0098,
		0.041881,
		0.110561,
		0.222012,
		0.370055,
		0.533837,
		0.681705,
		0.782486,
		0.818069,
		0.787895,
		0.702768,
		0.575293,
		0.416269,
		0.24037,
		0.078987,
		0.000719,
		0.009797,
		0.041881,
		0.110603,
		0.222455,
		0.372344,
		0.541506,
		0.700845,
		0.820937,
		0.883593,
		0.886448,
		0.837568,
		0.745942,
		0.616381,
		0.452462,
		0.265625,
		0.088884,
		0.000714,
		0.009761,
		0.041817,
		0.110561,
		0.222455,
		0.372533,
		0.542944,
		0.706621,
		0.836688,
		0.916836,
		0.945195,
		0.929125,
		0.875992,
		0.787387,
		0.660215,
		0.492972,
		0.294906,
		0.100747,
		0.000701,
		0.009635,
		0.041464,
		0.110041,
		0.222012,
		0.372344,
		0.542944,
		0.707107,
		0.839707,
		0.927134,
		0.969981,
		0.977154,
		0.956583,
		0.909189,
		0.828924,
		0.707107,
		0.538323,
		0.32905,
		0.11513,
		0.000677,
		0.009364,
		0.040544,
		0.108244,
		0.219602,
		0.370055,
		0.541506,
		0.706621,
		0.839707,
		0.928019,
		0.974863,
		0.992508,
		0.991829,
		0.975798,
		0.938862,
		0.870045,
		0.756699,
		0.588982,
		0.369091,
		0.132788,
		0.00064,
		0.00891,
		0.038844,
		0.104444,
		0.213445,
		0.362304,
		0.533837,
		0.700845,
		0.836688,
		0.927134,
		0.974863,
		0.993865,
		0.999005,
		0.99793,
		0.989153,
		0.964076,
		0.909232,
		0.808146,
		0.645261,
		0.416289,
		0.154777,
		0.00059,
		0.008259,
		0.036251,
		0.098195,
		0.202269,
		0.346229,
		0.514646,
		0.681705,
		0.820937,
		0.916836,
		0.969981,
		0.992508,
		0.999005,
		0.999952,
		0.999818,
		0.996876,
		0.983187,
		0.944313,
		0.859876,
		0.707107,
		0.472143,
		0.182603,
		0.000526,
		0.007416,
		0.032763,
		0.089384,
		0.185558,
		0.320332,
		0.480563,
		0.642924,
		0.782486,
		0.883593,
		0.945195,
		0.977154,
		0.991829,
		0.99793,
		0.999818,
		1,
		0.999707,
		0.994912,
		0.972543,
		0.909232,
		0.773702,
		0.538323,
		0.218463,
		0.000452,
		0.006407,
		0.028478,
		0.078208,
		0.163544,
		0.284601,
		0.430753,
		0.581942,
		0.715937,
		0.818069,
		0.886448,
		0.929125,
		0.956583,
		0.975798,
		0.989153,
		0.996876,
		0.999707,
		1,
		0.999493,
		0.991144,
		0.952128,
		0.842738,
		0.61641,
		0.265625,
		0.000371,
		0.005276,
		0.023578,
		0.065143,
		0.137128,
		0.240381,
		0.366781,
		0.5,
		0.621342,
		0.71801,
		0.787895,
		0.837568,
		0.875992,
		0.909189,
		0.938862,
		0.964076,
		0.983187,
		0.994912,
		0.999493,
		1,
		0.999046,
		0.983187,
		0.909232,
		0.707107,
		0.327031,
		0.000285,
		0.00408,
		0.018324,
		0.0509,
		0.107782,
		0.190179,
		0.292291,
		0.401677,
		0.50367,
		0.587947,
		0.652581,
		0.702768,
		0.745942,
		0.787387,
		0.828924,
		0.870045,
		0.909232,
		0.944313,
		0.972543,
		0.991144,
		0.999046,
		1,
		0.997978,
		0.964076,
		0.803188,
		0.294361,
		0.000201,
		0.00289,
		0.013038,
		0.036394,
		0.077474,
		0.137498,
		0.212684,
		0.294361,
		0.372038,
		0.438158,
		0.491211,
		0.535021,
		0.575293,
		0.616381,
		0.660215,
		0.707107,
		0.756699,
		0.808146,
		0.859876,
		0.909232,
		0.952128,
		0.983187,
		0.997978,
		1,
		0.988808,
		0.642924,
		0.05954,
		0.000124,
		0.00179,
		0.008105,
		0.022723,
		0.0486,
		0.086698,
		0.134865,
		0.187825,
		0.239036,
		0.283695,
		0.320804,
		0.352837,
		0.383622,
		0.416269,
		0.452462,
		0.492972,
		0.538323,
		0.588982,
		0.645261,
		0.707107,
		0.773702,
		0.842738,
		0.909232,
		0.964076,
		0.988808,
		0.707107,
		0.108744,
		6e-005,
		0.000871,
		0.00396,
		0.011143,
		0.023933,
		0.04289,
		0.067052,
		0.093895,
		0.120219,
		0.143636,
		0.163637,
		0.181483,
		0.199181,
		0.218452,
		0.24037,
		0.265625,
		0.294906,
		0.32905,
		0.369091,
		0.416289,
		0.472143,
		0.538323,
		0.61641,
		0.707107,
		0.803188,
		0.642924,
		0.108744,
		1.6e-005,
		0.000237,
		0.001083,
		0.003057,
		0.006591,
		0.011859,
		0.018622,
		0.026203,
		0.033727,
		0.040531,
		0.046472,
		0.05191,
		0.057429,
		0.063554,
		0.070648,
		0.078987,
		0.088884,
		0.100747,
		0.11513,
		0.132788,
		0.154777,
		0.182603,
		0.218463,
		0.265625,
		0.327031,
		0.294361,
		0.05954,
		0.05954,
		0.294361,
		0.327031,
		0.265625,
		0.218463,
		0.182603,
		0.154777,
		0.132788,
		0.11513,
		0.100747,
		0.088884,
		0.078987,
		0.070648,
		0.063554,
		0.057429,
		0.05191,
		0.046472,
		0.040531,
		0.033727,
		0.026203,
		0.018622,
		0.011859,
		0.006591,
		0.003057,
		0.001083,
		0.000237,
		1.6e-005,
		0.108744,
		0.642924,
		0.803188,
		0.707107,
		0.61641,
		0.538323,
		0.472143,
		0.416289,
		0.369091,
		0.32905,
		0.294906,
		0.265625,
		0.24037,
		0.218452,
		0.199181,
		0.181483,
		0.163637,
		0.143636,
		0.120219,
		0.093895,
		0.067052,
		0.04289,
		0.023933,
		0.011143,
		0.00396,
		0.000871,
		6e-005,
		0.108744,
		0.707107,
		0.988808,
		0.964076,
		0.909232,
		0.842738,
		0.773702,
		0.707107,
		0.645261,
		0.588982,
		0.538323,
		0.492972,
		0.452462,
		0.416269,
		0.383622,
		0.352837,
		0.320804,
		0.283695,
		0.239036,
		0.187825,
		0.134865,
		0.086698,
		0.0486,
		0.022723,
		0.008105,
		0.00179,
		0.000124,
		0.05954,
		0.642924,
		0.988808,
		1,
		0.997978,
		0.983187,
		0.952128,
		0.909232,
		0.859876,
		0.808146,
		0.756699,
		0.707107,
		0.660215,
		0.616381,
		0.575293,
		0.535021,
		0.491211,
		0.438158,
		0.372038,
		0.294361,
		0.212684,
		0.137498,
		0.077474,
		0.036394,
		0.013038,
		0.00289,
		0.000201,
		0.294361,
		0.803188,
		0.964076,
		0.997978,
		1,
		0.999046,
		0.991144,
		0.972543,
		0.944313,
		0.909232,
		0.870045,
		0.828924,
		0.787387,
		0.745942,
		0.702768,
		0.652581,
		0.587947,
		0.50367,
		0.401677,
		0.292291,
		0.190179,
		0.107782,
		0.0509,
		0.018324,
		0.00408,
		0.000285,
		0.327031,
		0.707107,
		0.909232,
		0.983187,
		0.999046,
		1,
		0.999493,
		0.994912,
		0.983187,
		0.964076,
		0.938862,
		0.909189,
		0.875992,
		0.837568,
		0.787895,
		0.71801,
		0.621342,
		0.5,
		0.366781,
		0.240381,
		0.137128,
		0.065143,
		0.023578,
		0.005276,
		0.000371,
		0.265625,
		0.61641,
		0.842738,
		0.952128,
		0.991144,
		0.999493,
		1,
		0.999707,
		0.996876,
		0.989153,
		0.975798,
		0.956583,
		0.929125,
		0.886448,
		0.818069,
		0.715937,
		0.581942,
		0.430753,
		0.284601,
		0.163544,
		0.078208,
		0.028478,
		0.006407,
		0.000452,
		0.218463,
		0.538323,
		0.773702,
		0.909232,
		0.972543,
		0.994912,
		0.999707,
		1,
		0.999818,
		0.99793,
		0.991829,
		0.977154,
		0.945195,
		0.883593,
		0.782486,
		0.642924,
		0.480563,
		0.320332,
		0.185558,
		0.089384,
		0.032763,
		0.007416,
		0.000526,
		0.182603,
		0.472143,
		0.707107,
		0.859876,
		0.944313,
		0.983187,
		0.996876,
		0.999818,
		0.999952,
		0.999005,
		0.992508,
		0.969981,
		0.916836,
		0.820937,
		0.681705,
		0.514646,
		0.346229,
		0.202269,
		0.098195,
		0.036251,
		0.008259,
		0.00059,
		0.154777,
		0.416289,
		0.645261,
		0.808146,
		0.909232,
		0.964076,
		0.989153,
		0.99793,
		0.999005,
		0.993865,
		0.974863,
		0.927134,
		0.836688,
		0.700845,
		0.533837,
		0.362304,
		0.213445,
		0.104444,
		0.038844,
		0.00891,
		0.00064,
		0.132788,
		0.369091,
		0.588982,
		0.756699,
		0.870045,
		0.938862,
		0.975798,
		0.991829,
		0.992508,
		0.974863,
		0.928019,
		0.839707,
		0.706621,
		0.541506,
		0.370055,
		0.219602,
		0.108244,
		0.040544,
		0.009364,
		0.000677,
		0.11513,
		0.32905,
		0.538323,
		0.707107,
		0.828924,
		0.909189,
		0.956583,
		0.977154,
		0.969981,
		0.927134,
		0.839707,
		0.707107,
		0.542944,
		0.372344,
		0.222012,
		0.110041,
		0.041464,
		0.009635,
		0.000701,
		0.100747,
		0.294906,
		0.492972,
		0.660215,
		0.787387,
		0.875992,
		0.929125,
		0.945195,
		0.916836,
		0.836688,
		0.706621,
		0.542944,
		0.372533,
		0.222455,
		0.110561,
		0.041817,
		0.009761,
		0.000714,
		0.088884,
		0.265625,
		0.452462,
		0.616381,
		0.745942,
		0.837568,
		0.886448,
		0.883593,
		0.820937,
		0.700845,
		0.541506,
		0.372344,
		0.222455,
		0.110603,
		0.041881,
		0.009797,
		0.000719,
		0.078987,
		0.24037,
		0.416269,
		0.575293,
		0.702768,
		0.787895,
		0.818069,
		0.782486,
		0.681705,
		0.533837,
		0.370055,
		0.222012,
		0.110561,
		0.041881,
		0.0098,
		0.000719,
		0.070648,
		0.218452,
		0.383622,
		0.535021,
		0.652581,
		0.71801,
		0.715937,
		0.642924,
		0.514646,
		0.362304,
		0.219602,
		0.110041,
		0.041817,
		0.009797,
		0.000719,
		0.063554,
		0.199181,
		0.352837,
		0.491211,
		0.587947,
		0.621342,
		0.581942,
		0.480563,
		0.346229,
		0.213445,
		0.108244,
		0.041464,
		0.009761,
		0.000719,
		0.057429,
		0.181483,
		0.320804,
		0.438158,
		0.50367,
		0.5,
		0.430753,
		0.320332,
		0.202269,
		0.104444,
		0.040544,
		0.009635,
		0.000714,
		0.05191,
		0.163637,
		0.283695,
		0.372038,
		0.401677,
		0.366781,
		0.284601,
		0.185558,
		0.098195,
		0.038844,
		0.009364,
		0.000701,
		0.046472,
		0.143636,
		0.239036,
		0.294361,
		0.292291,
		0.240381,
		0.163544,
		0.089384,
		0.036251,
		0.00891,
		0.000677,
		0.040531,
		0.120219,
		0.187825,
		0.212684,
		0.190179,
		0.137128,
		0.078208,
		0.032763,
		0.008259,
		0.00064,
		0.033727,
		0.093895,
		0.134865,
		0.137498,
		0.107782,
		0.065143,
		0.028478,
		0.007416,
		0.00059,
		0.026203,
		0.067052,
		0.086698,
		0.077474,
		0.0509,
		0.023578,
		0.006407,
		0.000526,
		0.018622,
		0.04289,
		0.0486,
		0.036394,
		0.018324,
		0.005276,
		0.000452,
		0.011859,
		0.023933,
		0.022723,
		0.013038,
		0.00408,
		0.000371,
		0.006591,
		0.011143,
		0.008105,
		0.00289,
		0.000285,
		0.003057,
		0.00396,
		0.00179,
		0.000201,
		0.001083,
		0.000871,
		0.000124,
		0.000237,
		6e-005,
		1.6e-005,
		1.6e-005,
		3e-005,
		4.5e-005,
		5.7e-005,
		6.2e-005,
		5.5e-005,
		3.7e-005,
		1.3e-005,
		1.3e-005,
		3.7e-005,
		5.5e-005,
		6.2e-005,
		5.7e-005,
		4.5e-005,
		3e-005,
		1.6e-005,
		1.3e-005,
		3.5e-005,
		7.2e-005,
		0.000121,
		0.000179,
		0.000237,
		0.000287,
		0.000316,
		0.000317,
		0.000283,
		0.000215,
		0.000127,
		4.1e-005,
		4.1e-005,
		0.000127,
		0.000215,
		0.000283,
		0.000317,
		0.000316,
		0.000287,
		0.000237,
		0.000179,
		0.000121,
		7.2e-005,
		3.5e-005,
		1.3e-005,
		2.2e-005,
		6e-005,
		0.000127,
		0.000227,
		0.000357,
		0.000512,
		0.000679,
		0.000842,
		0.000983,
		0.001083,
		0.001123,
		0.001091,
		0.000982,
		0.000799,
		0.000562,
		0.000308,
		9.4e-005,
		9.4e-005,
		0.000308,
		0.000562,
		0.000799,
		0.000982,
		0.001091,
		0.001123,
		0.001083,
		0.000983,
		0.000842,
		0.000679,
		0.000512,
		0.000357,
		0.000227,
		0.000127,
		6e-005,
		2.2e-005,
		3.2e-005,
		9e-005,
		0.000196,
		0.000359,
		0.000585,
		0.000871,
		0.001208,
		0.001579,
		0.001962,
		0.002331,
		0.002653,
		0.0029,
		0.003043,
		0.003057,
		0.002929,
		0.002653,
		0.002242,
		0.001726,
		0.001155,
		0.000604,
		0.000176,
		0.000176,
		0.000604,
		0.001155,
		0.001726,
		0.002242,
		0.002653,
		0.002929,
		0.003057,
		0.003043,
		0.0029,
		0.002653,
		0.002331,
		0.001962,
		0.001579,
		0.001208,
		0.000871,
		0.000585,
		0.000359,
		0.000196,
		9e-005,
		3.2e-005,
		4.3e-005,
		0.000124,
		0.000274,
		0.000513,
		0.000854,
		0.001301,
		0.001853,
		0.002495,
		0.003208,
		0.00396,
		0.004715,
		0.005431,
		0.006065,
		0.00657,
		0.006907,
		0.007038,
		0.006937,
		0.006591,
		0.006001,
		0.005191,
		0.004205,
		0.003111,
		0.002007,
		0.001015,
		0.000286,
		0.000286,
		0.001015,
		0.002007,
		0.003111,
		0.004205,
		0.005191,
		0.006001,
		0.006591,
		0.006937,
		0.007038,
		0.006907,
		0.00657,
		0.006065,
		0.005431,
		0.004715,
		0.00396,
		0.003208,
		0.002495,
		0.001853,
		0.001301,
		0.000854,
		0.000513,
		0.000274,
		0.000124,
		4.3e-005,
		1.2e-005,
		5.6e-005,
		0.000161,
		0.000361,
		0.000685,
		0.001156,
		0.00179,
		0.002592,
		0.003559,
		0.004672,
		0.005905,
		0.007219,
		0.008567,
		0.009895,
		0.011143,
		0.01225,
		0.013156,
		0.013802,
		0.014139,
		0.014128,
		0.013746,
		0.012984,
		0.011859,
		0.010409,
		0.008699,
		0.006822,
		0.004895,
		0.003067,
		0.001509,
		0.000415,
		0.000415,
		0.001509,
		0.003067,
		0.004895,
		0.006822,
		0.008699,
		0.010409,
		0.011859,
		0.012984,
		0.013746,
		0.014128,
		0.014139,
		0.013802,
		0.013156,
		0.01225,
		0.011143,
		0.009895,
		0.008567,
		0.007219,
		0.005905,
		0.004672,
		0.003559,
		0.002592,
		0.00179,
		0.001156,
		0.000685,
		0.000361,
		0.000161,
		5.6e-005,
		1.2e-005,
		1.5e-005,
		6.9e-005,
		0.000201,
		0.000454,
		0.00087,
		0.001483,
		0.002323,
		0.003407,
		0.004739,
		0.006313,
		0.008105,
		0.010082,
		0.012195,
		0.014386,
		0.016587,
		0.018722,
		0.020712,
		0.022475,
		0.023933,
		0.025011,
		0.025644,
		0.025783,
		0.025391,
		0.024457,
		0.022989,
		0.021023,
		0.018622,
		0.015878,
		0.012909,
		0.009861,
		0.006903,
		0.004224,
		0.002032,
		0.000547,
		0.000547,
		0.002032,
		0.004224,
		0.006903,
		0.009861,
		0.012909,
		0.015878,
		0.018622,
		0.021023,
		0.022989,
		0.024457,
		0.025391,
		0.025783,
		0.025644,
		0.025011,
		0.023933,
		0.022475,
		0.020712,
		0.018722,
		0.016587,
		0.014386,
		0.012195,
		0.010082,
		0.008105,
		0.006313,
		0.004739,
		0.003407,
		0.002323,
		0.001483,
		0.00087,
		0.000454,
		0.000201,
		6.9e-005,
		1.5e-005,
		1.7e-005,
		8.2e-005,
		0.000243,
		0.000552,
		0.001064,
		0.00183,
		0.00289,
		0.004277,
		0.006008,
		0.008087,
		0.010502,
		0.013224,
		0.016209,
		0.019399,
		0.022723,
		0.026096,
		0.029425,
		0.032612,
		0.035553,
		0.038146,
		0.040291,
		0.041898,
		0.04289,
		0.043203,
		0.042798,
		0.041656,
		0.039786,
		0.037225,
		0.03404,
		0.030325,
		0.026203,
		0.021821,
		0.01735,
		0.012978,
		0.008906,
		0.00535,
		0.00253,
		0.000671,
		0.000671,
		0.00253,
		0.00535,
		0.008906,
		0.012978,
		0.01735,
		0.021821,
		0.026203,
		0.030325,
		0.03404,
		0.037225,
		0.039786,
		0.041656,
		0.042798,
		0.043203,
		0.04289,
		0.041898,
		0.040291,
		0.038146,
		0.035553,
		0.032612,
		0.029425,
		0.026096,
		0.022723,
		0.019399,
		0.016209,
		0.013224,
		0.010502,
		0.008087,
		0.006008,
		0.004277,
		0.00289,
		0.00183,
		0.001064,
		0.000552,
		0.000243,
		8.2e-005,
		1.7e-005,
		2e-005,
		9.6e-005,
		0.000285,
		0.000652,
		0.001265,
		0.002188,
		0.003479,
		0.005185,
		0.007338,
		0.009956,
		0.013038,
		0.016567,
		0.020506,
		0.0248,
		0.029379,
		0.034155,
		0.039027,
		0.043882,
		0.0486,
		0.053053,
		0.057115,
		0.060661,
		0.063573,
		0.065746,
		0.067091,
		0.067541,
		0.067052,
		0.065609,
		0.063227,
		0.059949,
		0.055852,
		0.051039,
		0.045641,
		0.039813,
		0.033727,
		0.027572,
		0.021549,
		0.015865,
		0.010731,
		0.006363,
		0.002974,
		0.00078,
		0.00078,
		0.002974,
		0.006363,
		0.010731,
		0.015865,
		0.021549,
		0.027572,
		0.033727,
		0.039813,
		0.045641,
		0.051039,
		0.055852,
		0.059949,
		0.063227,
		0.065609,
		0.067052,
		0.067541,
		0.067091,
		0.065746,
		0.063573,
		0.060661,
		0.057115,
		0.053053,
		0.0486,
		0.043882,
		0.039027,
		0.034155,
		0.029379,
		0.0248,
		0.020506,
		0.016567,
		0.013038,
		0.009956,
		0.007338,
		0.005185,
		0.003479,
		0.002188,
		0.001265,
		0.000652,
		0.000285,
		9.6e-005,
		2e-005,
		2.3e-005,
		0.00011,
		0.000328,
		0.000754,
		0.001468,
		0.002552,
		0.00408,
		0.006113,
		0.008703,
		0.011881,
		0.015662,
		0.02004,
		0.02499,
		0.030463,
		0.036394,
		0.042695,
		0.049261,
		0.055972,
		0.062693,
		0.069278,
		0.075577,
		0.081434,
		0.086698,
		0.091224,
		0.09488,
		0.097552,
		0.099147,
		0.0996,
		0.098873,
		0.096962,
		0.093895,
		0.089732,
		0.084562,
		0.078506,
		0.071706,
		0.064324,
		0.05654,
		0.048543,
		0.040531,
		0.032703,
		0.025261,
		0.018407,
		0.01234,
		0.007261,
		0.003372,
		0.00088,
		0.00088,
		0.003372,
		0.007261,
		0.01234,
		0.018407,
		0.025261,
		0.032703,
		0.040531,
		0.048543,
		0.05654,
		0.064324,
		0.071706,
		0.078506,
		0.084562,
		0.089732,
		0.093895,
		0.096962,
		0.098873,
		0.0996,
		0.099147,
		0.097552,
		0.09488,
		0.091224,
		0.086698,
		0.081434,
		0.075577,
		0.069278,
		0.062693,
		0.055972,
		0.049261,
		0.042695,
		0.036394,
		0.030463,
		0.02499,
		0.02004,
		0.015662,
		0.011881,
		0.008703,
		0.006113,
		0.00408,
		0.002552,
		0.001468,
		0.000754,
		0.000328,
		0.00011,
		2.3e-005,
		2.6e-005,
		0.000124,
		0.000371,
		0.000854,
		0.001671,
		0.002916,
		0.004682,
		0.007047,
		0.010079,
		0.013828,
		0.018324,
		0.023575,
		0.029569,
		0.036269,
		0.043615,
		0.051523,
		0.05989,
		0.068588,
		0.077474,
		0.08639,
		0.095162,
		0.103613,
		0.11156,
		0.118823,
		0.12523,
		0.130624,
		0.134865,
		0.13784,
		0.139462,
		0.13968,
		0.138472,
		0.135857,
		0.131883,
		0.126633,
		0.120219,
		0.112777,
		0.104466,
		0.095456,
		0.085931,
		0.076079,
		0.066092,
		0.056161,
		0.046472,
		0.037212,
		0.028563,
		0.020707,
		0.013828,
		0.008113,
		0.003761,
		0.000981,
		0.000981,
		0.003761,
		0.008113,
		0.013828,
		0.020707,
		0.028563,
		0.037212,
		0.046472,
		0.056161,
		0.066092,
		0.076079,
		0.085931,
		0.095456,
		0.104466,
		0.112777,
		0.120219,
		0.126633,
		0.131883,
		0.135857,
		0.138472,
		0.13968,
		0.139462,
		0.13784,
		0.134865,
		0.130624,
		0.12523,
		0.118823,
		0.11156,
		0.103613,
		0.095162,
		0.08639,
		0.077474,
		0.068588,
		0.05989,
		0.051523,
		0.043615,
		0.036269,
		0.029569,
		0.023575,
		0.018324,
		0.013828,
		0.010079,
		0.007047,
		0.004682,
		0.002916,
		0.001671,
		0.000854,
		0.000371,
		0.000124,
		2.6e-005,
		2.9e-005,
		0.000138,
		0.000412,
		0.000953,
		0.00187,
		0.003275,
		0.005276,
		0.00797,
		0.011443,
		0.015763,
		0.020976,
		0.027109,
		0.034161,
		0.042108,
		0.0509,
		0.06046,
		0.070687,
		0.081453,
		0.092608,
		0.103984,
		0.115392,
		0.126633,
		0.137498,
		0.147776,
		0.15726,
		0.16575,
		0.173066,
		0.179047,
		0.183562,
		0.186509,
		0.187825,
		0.187482,
		0.18549,
		0.181895,
		0.176778,
		0.170248,
		0.162442,
		0.153515,
		0.143636,
		0.132984,
		0.121743,
		0.110096,
		0.098223,
		0.0863,
		0.074497,
		0.06298,
		0.05191,
		0.041447,
		0.031753,
		0.022996,
		0.015352,
		0.009011,
		0.004181,
		0.001092,
		0.001092,
		0.004181,
		0.009011,
		0.015352,
		0.022996,
		0.031753,
		0.041447,
		0.05191,
		0.06298,
		0.074497,
		0.0863,
		0.098223,
		0.110096,
		0.121743,
		0.132984,
		0.143636,
		0.153515,
		0.162442,
		0.170248,
		0.176778,
		0.181895,
		0.18549,
		0.187482,
		0.187825,
		0.186509,
		0.183562,
		0.179047,
		0.173066,
		0.16575,
		0.15726,
		0.147776,
		0.137498,
		0.126633,
		0.115392,
		0.103984,
		0.092608,
		0.081453,
		0.070687,
		0.06046,
		0.0509,
		0.042108,
		0.034161,
		0.027109,
		0.020976,
		0.015763,
		0.011443,
		0.00797,
		0.005276,
		0.003275,
		0.00187,
		0.000953,
		0.000412,
		0.000138,
		2.9e-005,
		3.1e-005,
		0.000151,
		0.000452,
		0.001049,
		0.002063,
		0.003623,
		0.005853,
		0.00887,
		0.012776,
		0.017657,
		0.023578,
		0.030582,
		0.038685,
		0.047876,
		0.058116,
		0.069338,
		0.081444,
		0.094309,
		0.107782,
		0.121685,
		0.135819,
		0.149966,
		0.163895,
		0.177368,
		0.190145,
		0.20199,
		0.212684,
		0.222021,
		0.229828,
		0.235957,
		0.240301,
		0.242791,
		0.243396,
		0.242128,
		0.239036,
		0.234205,
		0.227748,
		0.219804,
		0.210532,
		0.200101,
		0.188689,
		0.176476,
		0.163637,
		0.150344,
		0.136762,
		0.123044,
		0.109339,
		0.095789,
		0.082529,
		0.069697,
		0.057429,
		0.04587,
		0.035173,
		0.025507,
		0.017057,
		0.010032,
		0.004665,
		0.001221,
		0.001221,
		0.004665,
		0.010032,
		0.017057,
		0.025507,
		0.035173,
		0.04587,
		0.057429,
		0.069697,
		0.082529,
		0.095789,
		0.109339,
		0.123044,
		0.136762,
		0.150344,
		0.163637,
		0.176476,
		0.188689,
		0.200101,
		0.210532,
		0.219804,
		0.227748,
		0.234205,
		0.239036,
		0.242128,
		0.243396,
		0.242791,
		0.240301,
		0.235957,
		0.229828,
		0.222021,
		0.212684,
		0.20199,
		0.190145,
		0.177368,
		0.163895,
		0.149966,
		0.135819,
		0.121685,
		0.107782,
		0.094309,
		0.081444,
		0.069338,
		0.058116,
		0.047876,
		0.038685,
		0.030582,
		0.023578,
		0.017657,
		0.012776,
		0.00887,
		0.005853,
		0.003623,
		0.002063,
		0.001049,
		0.000452,
		0.000151,
		3.1e-005,
		3.4e-005,
		0.000163,
		0.000491,
		0.00114,
		0.002247,
		0.003956,
		0.006407,
		0.009733,
		0.014057,
		0.019481,
		0.02609,
		0.033943,
		0.043071,
		0.05348,
		0.065143,
		0.078003,
		0.09197,
		0.106923,
		0.122709,
		0.139147,
		0.156029,
		0.173123,
		0.190179,
		0.206935,
		0.223122,
		0.238473,
		0.252728,
		0.265647,
		0.27701,
		0.286632,
		0.294361,
		0.300088,
		0.303745,
		0.30531,
		0.304801,
		0.302277,
		0.297832,
		0.291589,
		0.283695,
		0.274314,
		0.263617,
		0.251784,
		0.238991,
		0.225407,
		0.211195,
		0.196507,
		0.181483,
		0.166253,
		0.150937,
		0.13565,
		0.120501,
		0.105601,
		0.091062,
		0.077003,
		0.063554,
		0.050861,
		0.039085,
		0.02841,
		0.019046,
		0.011231,
		0.005236,
		0.001374,
		0.001374,
		0.005236,
		0.011231,
		0.019046,
		0.02841,
		0.039085,
		0.050861,
		0.063554,
		0.077003,
		0.091062,
		0.105601,
		0.120501,
		0.13565,
		0.150937,
		0.166253,
		0.181483,
		0.196507,
		0.211195,
		0.225407,
		0.238991,
		0.251784,
		0.263617,
		0.274314,
		0.283695,
		0.291589,
		0.297832,
		0.302277,
		0.304801,
		0.30531,
		0.303745,
		0.300088,
		0.294361,
		0.286632,
		0.27701,
		0.265647,
		0.252728,
		0.238473,
		0.223122,
		0.206935,
		0.190179,
		0.173123,
		0.156029,
		0.139147,
		0.122709,
		0.106923,
		0.09197,
		0.078003,
		0.065143,
		0.05348,
		0.043071,
		0.033943,
		0.02609,
		0.019481,
		0.014057,
		0.009733,
		0.006407,
		0.003956,
		0.002247,
		0.00114,
		0.000491,
		0.000163,
		3.4e-005,
		3.6e-005,
		0.000175,
		0.000526,
		0.001225,
		0.002421,
		0.00427,
		0.00693,
		0.01055,
		0.015271,
		0.021213,
		0.028478,
		0.037142,
		0.047255,
		0.058836,
		0.071873,
		0.086318,
		0.102092,
		0.119079,
		0.137128,
		0.156055,
		0.175647,
		0.195661,
		0.21583,
		0.235873,
		0.255494,
		0.274398,
		0.292291,
		0.308897,
		0.323959,
		0.337251,
		0.348584,
		0.357808,
		0.364821,
		0.369568,
		0.372038,
		0.372267,
		0.370327,
		0.366327,
		0.360406,
		0.352718,
		0.343437,
		0.33274,
		0.320804,
		0.307804,
		0.293903,
		0.279254,
		0.263995,
		0.248253,
		0.232137,
		0.215749,
		0.199181,
		0.182516,
		0.165839,
		0.149231,
		0.132781,
		0.116585,
		0.100747,
		0.085389,
		0.070648,
		0.056682,
		0.043674,
		0.031834,
		0.021402,
		0.012657,
		0.00592,
		0.001559,
		0.001559,
		0.00592,
		0.012657,
		0.021402,
		0.031834,
		0.043674,
		0.056682,
		0.070648,
		0.085389,
		0.100747,
		0.116585,
		0.132781,
		0.149231,
		0.165839,
		0.182516,
		0.199181,
		0.215749,
		0.232137,
		0.248253,
		0.263995,
		0.279254,
		0.293903,
		0.307804,
		0.320804,
		0.33274,
		0.343437,
		0.352718,
		0.360406,
		0.366327,
		0.370327,
		0.372267,
		0.372038,
		0.369568,
		0.364821,
		0.357808,
		0.348584,
		0.337251,
		0.323959,
		0.308897,
		0.292291,
		0.274398,
		0.255494,
		0.235873,
		0.21583,
		0.195661,
		0.175647,
		0.156055,
		0.137128,
		0.119079,
		0.102092,
		0.086318,
		0.071873,
		0.058836,
		0.047255,
		0.037142,
		0.028478,
		0.021213,
		0.015271,
		0.01055,
		0.00693,
		0.00427,
		0.002421,
		0.001225,
		0.000526,
		0.000175,
		3.6e-005,
		3.8e-005,
		0.000185,
		0.00056,
		0.001305,
		0.002582,
		0.004562,
		0.007416,
		0.011311,
		0.016402,
		0.02283,
		0.030711,
		0.040139,
		0.051181,
		0.06387,
		0.078208,
		0.094161,
		0.111657,
		0.130588,
		0.150806,
		0.172129,
		0.194338,
		0.217181,
		0.240381,
		0.263636,
		0.286631,
		0.309044,
		0.330552,
		0.350845,
		0.369632,
		0.386651,
		0.401677,
		0.414524,
		0.425058,
		0.433192,
		0.438887,
		0.442157,
		0.443057,
		0.441681,
		0.438158,
		0.43264,
		0.425297,
		0.416309,
		0.405858,
		0.394122,
		0.381271,
		0.367463,
		0.352837,
		0.337521,
		0.321622,
		0.305234,
		0.288436,
		0.2713,
		0.253886,
		0.236251,
		0.218452,
		0.200548,
		0.182602,
		0.164687,
		0.146884,
		0.129291,
		0.112019,
		0.0952,
		0.078987,
		0.063557,
		0.049119,
		0.035914,
		0.024223,
		0.014374,
		0.006746,
		0.001783,
		0.001783,
		0.006746,
		0.014374,
		0.024223,
		0.035914,
		0.049119,
		0.063557,
		0.078987,
		0.0952,
		0.112019,
		0.129291,
		0.146884,
		0.164687,
		0.182602,
		0.200548,
		0.218452,
		0.236251,
		0.253886,
		0.2713,
		0.288436,
		0.305234,
		0.321622,
		0.337521,
		0.352837,
		0.367463,
		0.381271,
		0.394122,
		0.405858,
		0.416309,
		0.425297,
		0.43264,
		0.438158,
		0.441681,
		0.443057,
		0.442157,
		0.438887,
		0.433192,
		0.425058,
		0.414524,
		0.401677,
		0.386651,
		0.369632,
		0.350845,
		0.330552,
		0.309044,
		0.286631,
		0.263636,
		0.240381,
		0.217181,
		0.194338,
		0.172129,
		0.150806,
		0.130588,
		0.111657,
		0.094161,
		0.078208,
		0.06387,
		0.051181,
		0.040139,
		0.030711,
		0.02283,
		0.016402,
		0.011311,
		0.007416,
		0.004562,
		0.002582,
		0.001305,
		0.00056,
		0.000185,
		3.8e-005,
		4e-005,
		0.000195,
		0.00059,
		0.001377,
		0.002729,
		0.004828,
		0.00786,
		0.012007,
		0.01744,
		0.024314,
		0.032763,
		0.042898,
		0.0548,
		0.068519,
		0.084068,
		0.101426,
		0.120532,
		0.141286,
		0.163544,
		0.187125,
		0.211808,
		0.237336,
		0.26342,
		0.289746,
		0.315979,
		0.341773,
		0.366781,
		0.390664,
		0.4131,
		0.433793,
		0.452485,
		0.468959,
		0.48305,
		0.494641,
		0.50367,
		0.510129,
		0.514056,
		0.515532,
		0.514679,
		0.511645,
		0.5066,
		0.499726,
		0.491211,
		0.48124,
		0.469991,
		0.457628,
		0.444298,
		0.430131,
		0.415238,
		0.39971,
		0.383622,
		0.367036,
		0.349999,
		0.332553,
		0.314732,
		0.296568,
		0.278094,
		0.259348,
		0.24037,
		0.221211,
		0.201932,
		0.182603,
		0.163312,
		0.144161,
		0.125272,
		0.106789,
		0.088884,
		0.071756,
		0.055645,
		0.04083,
		0.027641,
		0.016464,
		0.007758,
		0.002059,
		0.002059,
		0.007758,
		0.016464,
		0.027641,
		0.04083,
		0.055645,
		0.071756,
		0.088884,
		0.106789,
		0.125272,
		0.144161,
		0.163312,
		0.182603,
		0.201932,
		0.221211,
		0.24037,
		0.259348,
		0.278094,
		0.296568,
		0.314732,
		0.332553,
		0.349999,
		0.367036,
		0.383622,
		0.39971,
		0.415238,
		0.430131,
		0.444298,
		0.457628,
		0.469991,
		0.48124,
		0.491211,
		0.499726,
		0.5066,
		0.511645,
		0.514679,
		0.515532,
		0.514056,
		0.510129,
		0.50367,
		0.494641,
		0.48305,
		0.468959,
		0.452485,
		0.433793,
		0.4131,
		0.390664,
		0.366781,
		0.341773,
		0.315979,
		0.289746,
		0.26342,
		0.237336,
		0.211808,
		0.187125,
		0.163544,
		0.141286,
		0.120532,
		0.101426,
		0.084068,
		0.068519,
		0.0548,
		0.042898,
		0.032763,
		0.024314,
		0.01744,
		0.012007,
		0.00786,
		0.004828,
		0.002729,
		0.001377,
		0.00059,
		0.000195,
		4e-005,
		4.2e-005,
		0.000204,
		0.000617,
		0.001442,
		0.00286,
		0.005066,
		0.008259,
		0.012633,
		0.018373,
		0.025651,
		0.034615,
		0.045391,
		0.058074,
		0.072729,
		0.089384,
		0.108027,
		0.128609,
		0.151036,
		0.175172,
		0.200837,
		0.22781,
		0.25583,
		0.284601,
		0.313795,
		0.343062,
		0.372039,
		0.400354,
		0.427644,
		0.453558,
		0.477773,
		0.5,
		0.519995,
		0.537563,
		0.552562,
		0.564909,
		0.574574,
		0.581581,
		0.586001,
		0.587947,
		0.587564,
		0.585024,
		0.580512,
		0.574223,
		0.56635,
		0.557079,
		0.546584,
		0.535021,
		0.522525,
		0.509214,
		0.495182,
		0.480503,
		0.465237,
		0.449423,
		0.433094,
		0.416269,
		0.398965,
		0.381192,
		0.362964,
		0.344293,
		0.325197,
		0.3057,
		0.28583,
		0.265625,
		0.245133,
		0.224414,
		0.203541,
		0.182603,
		0.16171,
		0.140991,
		0.120607,
		0.100747,
		0.08164,
		0.063557,
		0.046826,
		0.031834,
		0.019046,
		0.009016,
		0.002404,
		0.002404,
		0.009016,
		0.019046,
		0.031834,
		0.046826,
		0.063557,
		0.08164,
		0.100747,
		0.120607,
		0.140991,
		0.16171,
		0.182603,
		0.203541,
		0.224414,
		0.245133,
		0.265625,
		0.28583,
		0.3057,
		0.325197,
		0.344293,
		0.362964,
		0.381192,
		0.398965,
		0.416269,
		0.433094,
		0.449423,
		0.465237,
		0.480503,
		0.495182,
		0.509214,
		0.522525,
		0.535021,
		0.546584,
		0.557079,
		0.56635,
		0.574223,
		0.580512,
		0.585024,
		0.587564,
		0.587947,
		0.586001,
		0.581581,
		0.574574,
		0.564909,
		0.552562,
		0.537563,
		0.519995,
		0.5,
		0.477773,
		0.453558,
		0.427644,
		0.400354,
		0.372039,
		0.343062,
		0.313795,
		0.284601,
		0.25583,
		0.22781,
		0.200837,
		0.175172,
		0.151036,
		0.128609,
		0.108027,
		0.089384,
		0.072729,
		0.058074,
		0.045391,
		0.034615,
		0.025651,
		0.018373,
		0.012633,
		0.008259,
		0.005066,
		0.00286,
		0.001442,
		0.000617,
		0.000204,
		4.2e-005,
		4.4e-005,
		0.000211,
		0.00064,
		0.001498,
		0.002975,
		0.005276,
		0.008609,
		0.013183,
		0.019196,
		0.02683,
		0.036251,
		0.047595,
		0.060974,
		0.076464,
		0.094104,
		0.113897,
		0.135801,
		0.159732,
		0.185558,
		0.213104,
		0.242148,
		0.272428,
		0.303641,
		0.33545,
		0.367494,
		0.399391,
		0.430753,
		0.461192,
		0.490336,
		0.517835,
		0.543376,
		0.566689,
		0.587553,
		0.605806,
		0.621342,
		0.634115,
		0.644135,
		0.651464,
		0.656209,
		0.658512,
		0.658546,
		0.656502,
		0.652581,
		0.646984,
		0.639906,
		0.631533,
		0.622029,
		0.611541,
		0.600193,
		0.588084,
		0.575293,
		0.56188,
		0.547885,
		0.533332,
		0.518237,
		0.502605,
		0.486434,
		0.469721,
		0.452462,
		0.434652,
		0.416289,
		0.397375,
		0.377913,
		0.357915,
		0.337397,
		0.316383,
		0.294906,
		0.273008,
		0.250747,
		0.228195,
		0.205443,
		0.182603,
		0.159817,
		0.137255,
		0.11513,
		0.093699,
		0.073275,
		0.05424,
		0.037057,
		0.022285,
		0.010606,
		0.002844,
		0.002844,
		0.010606,
		0.022285,
		0.037057,
		0.05424,
		0.073275,
		0.093699,
		0.11513,
		0.137255,
		0.159817,
		0.182603,
		0.205443,
		0.228195,
		0.250747,
		0.273008,
		0.294906,
		0.316383,
		0.337397,
		0.357915,
		0.377913,
		0.397375,
		0.416289,
		0.434652,
		0.452462,
		0.469721,
		0.486434,
		0.502605,
		0.518237,
		0.533332,
		0.547885,
		0.56188,
		0.575293,
		0.588084,
		0.600193,
		0.611541,
		0.622029,
		0.631533,
		0.639906,
		0.646984,
		0.652581,
		0.656502,
		0.658546,
		0.658512,
		0.656209,
		0.651464,
		0.644135,
		0.634115,
		0.621342,
		0.605806,
		0.587553,
		0.566689,
		0.543376,
		0.517835,
		0.490336,
		0.461192,
		0.430753,
		0.399391,
		0.367494,
		0.33545,
		0.303641,
		0.272428,
		0.242148,
		0.213104,
		0.185558,
		0.159732,
		0.135801,
		0.113897,
		0.094104,
		0.076464,
		0.060974,
		0.047595,
		0.036251,
		0.02683,
		0.019196,
		0.013183,
		0.008609,
		0.005276,
		0.002975,
		0.001498,
		0.00064,
		0.000211,
		4.4e-005,
		4.5e-005,
		0.000218,
		0.000661,
		0.001547,
		0.003074,
		0.005456,
		0.00891,
		0.013657,
		0.019903,
		0.027847,
		0.037661,
		0.049499,
		0.063481,
		0.079696,
		0.098195,
		0.118991,
		0.142051,
		0.167298,
		0.194608,
		0.223808,
		0.254678,
		0.286955,
		0.320332,
		0.354464,
		0.388981,
		0.423487,
		0.457578,
		0.49085,
		0.522909,
		0.553385,
		0.581942,
		0.608287,
		0.632181,
		0.653439,
		0.67194,
		0.687623,
		0.700485,
		0.71058,
		0.71801,
		0.722916,
		0.725472,
		0.725873,
		0.724327,
		0.721044,
		0.71623,
		0.710079,
		0.702768,
		0.694454,
		0.685268,
		0.67532,
		0.664693,
		0.653452,
		0.641638,
		0.629277,
		0.616381,
		0.60295,
		0.588979,
		0.574454,
		0.559361,
		0.543682,
		0.527402,
		0.510503,
		0.492972,
		0.474793,
		0.455958,
		0.436458,
		0.416289,
		0.395453,
		0.373956,
		0.351814,
		0.32905,
		0.3057,
		0.281814,
		0.25746,
		0.232725,
		0.207724,
		0.182603,
		0.157547,
		0.132788,
		0.108615,
		0.085389,
		0.063557,
		0.043674,
		0.026425,
		0.012657,
		0.003417,
		0.003417,
		0.012657,
		0.026425,
		0.043674,
		0.063557,
		0.085389,
		0.108615,
		0.132788,
		0.157547,
		0.182603,
		0.207724,
		0.232725,
		0.25746,
		0.281814,
		0.3057,
		0.32905,
		0.351814,
		0.373956,
		0.395453,
		0.416289,
		0.436458,
		0.455958,
		0.474793,
		0.492972,
		0.510503,
		0.527402,
		0.543682,
		0.559361,
		0.574454,
		0.588979,
		0.60295,
		0.616381,
		0.629277,
		0.641638,
		0.653452,
		0.664693,
		0.67532,
		0.685268,
		0.694454,
		0.702768,
		0.710079,
		0.71623,
		0.721044,
		0.724327,
		0.725873,
		0.725472,
		0.722916,
		0.71801,
		0.71058,
		0.700485,
		0.687623,
		0.67194,
		0.653439,
		0.632181,
		0.608287,
		0.581942,
		0.553385,
		0.522909,
		0.49085,
		0.457578,
		0.423487,
		0.388981,
		0.354464,
		0.320332,
		0.286955,
		0.254678,
		0.223808,
		0.194608,
		0.167298,
		0.142051,
		0.118991,
		0.098195,
		0.079696,
		0.063481,
		0.049499,
		0.037661,
		0.027847,
		0.019903,
		0.013657,
		0.00891,
		0.005456,
		0.003074,
		0.001547,
		0.000661,
		0.000218,
		4.5e-005,
		4.6e-005,
		0.000223,
		0.000677,
		0.001587,
		0.003156,
		0.005605,
		0.009162,
		0.014052,
		0.020495,
		0.028697,
		0.038844,
		0.051096,
		0.065587,
		0.082414,
		0.101641,
		0.123286,
		0.147328,
		0.173695,
		0.202269,
		0.232882,
		0.265316,
		0.299306,
		0.334543,
		0.370678,
		0.407332,
		0.444099,
		0.480563,
		0.516304,
		0.550912,
		0.584001,
		0.615215,
		0.644244,
		0.67083,
		0.694774,
		0.715937,
		0.734247,
		0.749691,
		0.762317,
		0.772221,
		0.779544,
		0.784462,
		0.787175,
		0.787895,
		0.786843,
		0.784233,
		0.780271,
		0.775143,
		0.769016,
		0.762033,
		0.75431,
		0.745942,
		0.736995,
		0.727518,
		0.71754,
		0.707073,
		0.696118,
		0.684668,
		0.672707,
		0.660215,
		0.647171,
		0.63355,
		0.619329,
		0.604481,
		0.588982,
		0.572806,
		0.555928,
		0.538323,
		0.519969,
		0.500843,
		0.480925,
		0.460201,
		0.438658,
		0.416289,
		0.393097,
		0.369091,
		0.344293,
		0.318742,
		0.292492,
		0.265625,
		0.238249,
		0.210511,
		0.182603,
		0.154777,
		0.127354,
		0.100747,
		0.075484,
		0.05223,
		0.031834,
		0.015366,
		0.004182,
		0.004182,
		0.015366,
		0.031834,
		0.05223,
		0.075484,
		0.100747,
		0.127354,
		0.154777,
		0.182603,
		0.210511,
		0.238249,
		0.265625,
		0.292492,
		0.318742,
		0.344293,
		0.369091,
		0.393097,
		0.416289,
		0.438658,
		0.460201,
		0.480925,
		0.500843,
		0.519969,
		0.538323,
		0.555928,
		0.572806,
		0.588982,
		0.604481,
		0.619329,
		0.63355,
		0.647171,
		0.660215,
		0.672707,
		0.684668,
		0.696118,
		0.707073,
		0.71754,
		0.727518,
		0.736995,
		0.745942,
		0.75431,
		0.762033,
		0.769016,
		0.775143,
		0.780271,
		0.784233,
		0.786843,
		0.787895,
		0.787175,
		0.784462,
		0.779544,
		0.772221,
		0.762317,
		0.749691,
		0.734247,
		0.715937,
		0.694774,
		0.67083,
		0.644244,
		0.615215,
		0.584001,
		0.550912,
		0.516304,
		0.480563,
		0.444099,
		0.407332,
		0.370678,
		0.334543,
		0.299306,
		0.265316,
		0.232882,
		0.202269,
		0.173695,
		0.147328,
		0.123286,
		0.101641,
		0.082414,
		0.065587,
		0.051096,
		0.038844,
		0.028697,
		0.020495,
		0.014052,
		0.009162,
		0.005605,
		0.003156,
		0.001587,
		0.000677,
		0.000223,
		4.6e-005,
		4.7e-005,
		0.000228,
		0.000691,
		0.001619,
		0.003222,
		0.005726,
		0.009364,
		0.014371,
		0.020973,
		0.029385,
		0.039801,
		0.052391,
		0.067296,
		0.084624,
		0.104444,
		0.126786,
		0.151632,
		0.17892,
		0.208535,
		0.240313,
		0.274039,
		0.309448,
		0.346229,
		0.384031,
		0.422467,
		0.461124,
		0.499578,
		0.537395,
		0.574154,
		0.609454,
		0.642924,
		0.674241,
		0.703131,
		0.729382,
		0.752845,
		0.773435,
		0.791133,
		0.80598,
		0.818069,
		0.827542,
		0.834575,
		0.839373,
		0.842154,
		0.843145,
		0.842572,
		0.840647,
		0.837568,
		0.833512,
		0.828631,
		0.823051,
		0.816874,
		0.810175,
		0.803008,
		0.795406,
		0.787387,
		0.778955,
		0.770102,
		0.760816,
		0.751076,
		0.740862,
		0.730146,
		0.718904,
		0.707107,
		0.694725,
		0.681729,
		0.668086,
		0.653763,
		0.638727,
		0.622942,
		0.606372,
		0.588982,
		0.570735,
		0.551596,
		0.531529,
		0.510503,
		0.488488,
		0.465459,
		0.441396,
		0.416289,
		0.39014,
		0.362964,
		0.334797,
		0.3057,
		0.275766,
		0.245133,
		0.213992,
		0.182603,
		0.151319,
		0.120607,
		0.091084,
		0.063557,
		0.039085,
		0.019046,
		0.005236,
		0.005236,
		0.019046,
		0.039085,
		0.063557,
		0.091084,
		0.120607,
		0.151319,
		0.182603,
		0.213992,
		0.245133,
		0.275766,
		0.3057,
		0.334797,
		0.362964,
		0.39014,
		0.416289,
		0.441396,
		0.465459,
		0.488488,
		0.510503,
		0.531529,
		0.551596,
		0.570735,
		0.588982,
		0.606372,
		0.622942,
		0.638727,
		0.653763,
		0.668086,
		0.681729,
		0.694725,
		0.707107,
		0.718904,
		0.730146,
		0.740862,
		0.751076,
		0.760816,
		0.770102,
		0.778955,
		0.787387,
		0.795406,
		0.803008,
		0.810175,
		0.816874,
		0.823051,
		0.828631,
		0.833512,
		0.837568,
		0.840647,
		0.842572,
		0.843145,
		0.842154,
		0.839373,
		0.834575,
		0.827542,
		0.818069,
		0.80598,
		0.791133,
		0.773435,
		0.752845,
		0.729382,
		0.703131,
		0.674241,
		0.642924,
		0.609454,
		0.574154,
		0.537395,
		0.499578,
		0.461124,
		0.422467,
		0.384031,
		0.346229,
		0.309448,
		0.274039,
		0.240313,
		0.208535,
		0.17892,
		0.151632,
		0.126786,
		0.104444,
		0.084624,
		0.067296,
		0.052391,
		0.039801,
		0.029385,
		0.020973,
		0.014371,
		0.009364,
		0.005726,
		0.003222,
		0.001619,
		0.000691,
		0.000228,
		4.7e-005,
		4.7e-005,
		0.000231,
		0.000701,
		0.001644,
		0.003273,
		0.005819,
		0.009521,
		0.014618,
		0.021344,
		0.029919,
		0.040544,
		0.053397,
		0.068626,
		0.086344,
		0.10663,
		0.129518,
		0.154997,
		0.183009,
		0.213445,
		0.246144,
		0.280893,
		0.317427,
		0.355436,
		0.394565,
		0.434424,
		0.474596,
		0.514646,
		0.554134,
		0.59263,
		0.629719,
		0.665023,
		0.698204,
		0.72898,
		0.757127,
		0.782486,
		0.804966,
		0.824541,
		0.841247,
		0.855177,
		0.86647,
		0.875305,
		0.881889,
		0.886448,
		0.889215,
		0.890421,
		0.890289,
		0.889025,
		0.886816,
		0.883821,
		0.880177,
		0.875992,
		0.87135,
		0.86631,
		0.860913,
		0.855181,
		0.849122,
		0.842734,
		0.836006,
		0.828924,
		0.821467,
		0.813614,
		0.80534,
		0.796619,
		0.787425,
		0.777728,
		0.767497,
		0.756699,
		0.745299,
		0.73326,
		0.720543,
		0.707107,
		0.692907,
		0.677898,
		0.662033,
		0.645261,
		0.62753,
		0.608789,
		0.588982,
		0.568056,
		0.545959,
		0.522638,
		0.498046,
		0.472143,
		0.444897,
		0.416289,
		0.38632,
		0.355015,
		0.322436,
		0.28869,
		0.253947,
		0.218463,
		0.182603,
		0.146884,
		0.112019,
		0.078987,
		0.049119,
		0.024222,
		0.00674,
		0.00674,
		0.024222,
		0.049119,
		0.078987,
		0.112019,
		0.146884,
		0.182603,
		0.218463,
		0.253947,
		0.28869,
		0.322436,
		0.355015,
		0.38632,
		0.416289,
		0.444897,
		0.472143,
		0.498046,
		0.522638,
		0.545959,
		0.568056,
		0.588982,
		0.608789,
		0.62753,
		0.645261,
		0.662033,
		0.677898,
		0.692907,
		0.707107,
		0.720543,
		0.73326,
		0.745299,
		0.756699,
		0.767497,
		0.777728,
		0.787425,
		0.796619,
		0.80534,
		0.813614,
		0.821467,
		0.828924,
		0.836006,
		0.842734,
		0.849122,
		0.855181,
		0.860913,
		0.86631,
		0.87135,
		0.875992,
		0.880177,
		0.883821,
		0.886816,
		0.889025,
		0.890289,
		0.890421,
		0.889215,
		0.886448,
		0.881889,
		0.875305,
		0.86647,
		0.855177,
		0.841247,
		0.824541,
		0.804966,
		0.782486,
		0.757127,
		0.72898,
		0.698204,
		0.665023,
		0.629719,
		0.59263,
		0.554134,
		0.514646,
		0.474596,
		0.434424,
		0.394565,
		0.355436,
		0.317427,
		0.280893,
		0.246144,
		0.213445,
		0.183009,
		0.154997,
		0.129518,
		0.10663,
		0.086344,
		0.068626,
		0.053397,
		0.040544,
		0.029919,
		0.021344,
		0.014618,
		0.009521,
		0.005819,
		0.003273,
		0.001644,
		0.000701,
		0.000231,
		4.7e-005,
		4.8e-005,
		0.000233,
		0.000709,
		0.001662,
		0.003311,
		0.005887,
		0.009635,
		0.014799,
		0.021616,
		0.030311,
		0.041091,
		0.054138,
		0.069605,
		0.087614,
		0.108244,
		0.131537,
		0.157487,
		0.186039,
		0.217088,
		0.250476,
		0.285991,
		0.323371,
		0.362304,
		0.402434,
		0.443369,
		0.484687,
		0.52595,
		0.566711,
		0.606532,
		0.644993,
		0.681705,
		0.716324,
		0.748558,
		0.778177,
		0.805015,
		0.828975,
		0.850026,
		0.868202,
		0.883593,
		0.896339,
		0.90662,
		0.914647,
		0.92065,
		0.924865,
		0.927533,
		0.928881,
		0.929125,
		0.928457,
		0.927046,
		0.925035,
		0.92254,
		0.919651,
		0.916436,
		0.912939,
		0.909189,
		0.905199,
		0.900973,
		0.896504,
		0.891783,
		0.886795,
		0.881522,
		0.875946,
		0.870045,
		0.863799,
		0.857183,
		0.850172,
		0.842738,
		0.834852,
		0.826481,
		0.817592,
		0.808146,
		0.798105,
		0.787425,
		0.776061,
		0.763962,
		0.751077,
		0.737347,
		0.722712,
		0.707107,
		0.690463,
		0.672708,
		0.653763,
		0.63355,
		0.611985,
		0.588982,
		0.564456,
		0.538323,
		0.510503,
		0.480925,
		0.449532,
		0.416289,
		0.381194,
		0.344293,
		0.3057,
		0.265625,
		0.224414,
		0.182595,
		0.140868,
		0.100129,
		0.061965,
		0.029542,
		0.007571,
		0.007571,
		0.029542,
		0.061965,
		0.100129,
		0.140868,
		0.182595,
		0.224414,
		0.265625,
		0.3057,
		0.344293,
		0.381194,
		0.416289,
		0.449532,
		0.480925,
		0.510503,
		0.538323,
		0.564456,
		0.588982,
		0.611985,
		0.63355,
		0.653763,
		0.672708,
		0.690463,
		0.707107,
		0.722712,
		0.737347,
		0.751077,
		0.763962,
		0.776061,
		0.787425,
		0.798105,
		0.808146,
		0.817592,
		0.826481,
		0.834852,
		0.842738,
		0.850172,
		0.857183,
		0.863799,
		0.870045,
		0.875946,
		0.881522,
		0.886795,
		0.891783,
		0.896504,
		0.900973,
		0.905199,
		0.909189,
		0.912939,
		0.916436,
		0.919651,
		0.92254,
		0.925035,
		0.927046,
		0.928457,
		0.929125,
		0.928881,
		0.927533,
		0.924865,
		0.92065,
		0.914647,
		0.90662,
		0.896339,
		0.883593,
		0.868202,
		0.850026,
		0.828975,
		0.805015,
		0.778177,
		0.748558,
		0.716324,
		0.681705,
		0.644993,
		0.606532,
		0.566711,
		0.52595,
		0.484687,
		0.443369,
		0.402434,
		0.362304,
		0.323371,
		0.285991,
		0.250476,
		0.217088,
		0.186039,
		0.157487,
		0.131537,
		0.108244,
		0.087614,
		0.069605,
		0.054138,
		0.041091,
		0.030311,
		0.021616,
		0.014799,
		0.009635,
		0.005887,
		0.003311,
		0.001662,
		0.000709,
		0.000233,
		4.8e-005,
		4.8e-005,
		0.000235,
		0.000714,
		0.001675,
		0.003336,
		0.005933,
		0.009713,
		0.014922,
		0.021801,
		0.030578,
		0.041464,
		0.054644,
		0.070277,
		0.088484,
		0.109353,
		0.132926,
		0.159201,
		0.188127,
		0.219602,
		0.253469,
		0.289518,
		0.327489,
		0.367068,
		0.4079,
		0.449592,
		0.491718,
		0.533837,
		0.575499,
		0.616261,
		0.655698,
		0.693416,
		0.729065,
		0.762347,
		0.793028,
		0.820937,
		0.845973,
		0.868102,
		0.887355,
		0.903823,
		0.917645,
		0.929002,
		0.938108,
		0.945195,
		0.950506,
		0.954283,
		0.956762,
		0.958161,
		0.958682,
		0.958497,
		0.957757,
		0.956583,
		0.955071,
		0.953294,
		0.951302,
		0.94913,
		0.946796,
		0.944307,
		0.941665,
		0.938862,
		0.93589,
		0.932739,
		0.929393,
		0.925841,
		0.922066,
		0.918052,
		0.913781,
		0.909232,
		0.904386,
		0.899218,
		0.893703,
		0.887815,
		0.881522,
		0.874792,
		0.86759,
		0.859876,
		0.851607,
		0.842738,
		0.833218,
		0.82299,
		0.811993,
		0.800163,
		0.787425,
		0.773702,
		0.758906,
		0.742944,
		0.725714,
		0.707107,
		0.687003,
		0.665277,
		0.641793,
		0.61641,
		0.588982,
		0.559361,
		0.527402,
		0.492972,
		0.455958,
		0.416269,
		0.373628,
		0.327031,
		0.274752,
		0.215973,
		0.153342,
		0.093895,
		0.046363,
		0.01627,
		0.002816,
		0.002816,
		0.01627,
		0.046363,
		0.093895,
		0.153342,
		0.215973,
		0.274752,
		0.327031,
		0.373628,
		0.416269,
		0.455958,
		0.492972,
		0.527402,
		0.559361,
		0.588982,
		0.61641,
		0.641793,
		0.665277,
		0.687003,
		0.707107,
		0.725714,
		0.742944,
		0.758906,
		0.773702,
		0.787425,
		0.800163,
		0.811993,
		0.82299,
		0.833218,
		0.842738,
		0.851607,
		0.859876,
		0.86759,
		0.874792,
		0.881522,
		0.887815,
		0.893703,
		0.899218,
		0.904386,
		0.909232,
		0.913781,
		0.918052,
		0.922066,
		0.925841,
		0.929393,
		0.932739,
		0.93589,
		0.938862,
		0.941665,
		0.944307,
		0.946796,
		0.94913,
		0.951302,
		0.953294,
		0.955071,
		0.956583,
		0.957757,
		0.958497,
		0.958682,
		0.958161,
		0.956762,
		0.954283,
		0.950506,
		0.945195,
		0.938108,
		0.929002,
		0.917645,
		0.903823,
		0.887355,
		0.868102,
		0.845973,
		0.820937,
		0.793028,
		0.762347,
		0.729065,
		0.693416,
		0.655698,
		0.616261,
		0.575499,
		0.533837,
		0.491718,
		0.449592,
		0.4079,
		0.367068,
		0.327489,
		0.289518,
		0.253469,
		0.219602,
		0.188127,
		0.159201,
		0.132926,
		0.109353,
		0.088484,
		0.070277,
		0.054644,
		0.041464,
		0.030578,
		0.021801,
		0.014922,
		0.009713,
		0.005933,
		0.003336,
		0.001675,
		0.000714,
		0.000235,
		4.8e-005,
		4.8e-005,
		0.000236,
		0.000717,
		0.001682,
		0.003351,
		0.005962,
		0.009761,
		0.014998,
		0.021915,
		0.030743,
		0.041695,
		0.054958,
		0.070692,
		0.089024,
		0.110041,
		0.133789,
		0.160268,
		0.189429,
		0.22117,
		0.255338,
		0.291724,
		0.330067,
		0.370055,
		0.411333,
		0.453504,
		0.496144,
		0.53881,
		0.581049,
		0.622414,
		0.662478,
		0.700845,
		0.73716,
		0.771123,
		0.802496,
		0.831105,
		0.856847,
		0.879686,
		0.899653,
		0.916836,
		0.931376,
		0.943456,
		0.953289,
		0.961111,
		0.967166,
		0.971702,
		0.974957,
		0.977154,
		0.978498,
		0.979167,
		0.979315,
		0.979067,
		0.978526,
		0.977766,
		0.976845,
		0.975798,
		0.97465,
		0.973413,
		0.97209,
		0.970681,
		0.969182,
		0.967586,
		0.965887,
		0.964076,
		0.962145,
		0.960083,
		0.957881,
		0.955526,
		0.953007,
		0.950308,
		0.947416,
		0.944313,
		0.94098,
		0.937398,
		0.933544,
		0.929393,
		0.924919,
		0.92009,
		0.914874,
		0.909232,
		0.903125,
		0.896506,
		0.889324,
		0.881522,
		0.873037,
		0.863799,
		0.853729,
		0.842738,
		0.830729,
		0.817592,
		0.803203,
		0.787425,
		0.770107,
		0.751077,
		0.730146,
		0.707107,
		0.681729,
		0.653732,
		0.622395,
		0.585368,
		0.537774,
		0.473757,
		0.390871,
		0.294361,
		0.197077,
		0.113883,
		0.054533,
		0.020197,
		0.005051,
		0.000623,
		1.4e-005,
		1.4e-005,
		0.000623,
		0.005051,
		0.020197,
		0.054533,
		0.113883,
		0.197077,
		0.294361,
		0.390871,
		0.473757,
		0.537774,
		0.585368,
		0.622395,
		0.653732,
		0.681729,
		0.707107,
		0.730146,
		0.751077,
		0.770107,
		0.787425,
		0.803203,
		0.817592,
		0.830729,
		0.842738,
		0.853729,
		0.863799,
		0.873037,
		0.881522,
		0.889324,
		0.896506,
		0.903125,
		0.909232,
		0.914874,
		0.92009,
		0.924919,
		0.929393,
		0.933544,
		0.937398,
		0.94098,
		0.944313,
		0.947416,
		0.950308,
		0.953007,
		0.955526,
		0.957881,
		0.960083,
		0.962145,
		0.964076,
		0.965887,
		0.967586,
		0.969182,
		0.970681,
		0.97209,
		0.973413,
		0.97465,
		0.975798,
		0.976845,
		0.977766,
		0.978526,
		0.979067,
		0.979315,
		0.979167,
		0.978498,
		0.977154,
		0.974957,
		0.971702,
		0.967166,
		0.961111,
		0.953289,
		0.943456,
		0.931376,
		0.916836,
		0.899653,
		0.879686,
		0.856847,
		0.831105,
		0.802496,
		0.771123,
		0.73716,
		0.700845,
		0.662478,
		0.622414,
		0.581049,
		0.53881,
		0.496144,
		0.453504,
		0.411333,
		0.370055,
		0.330067,
		0.291724,
		0.255338,
		0.22117,
		0.189429,
		0.160268,
		0.133789,
		0.110041,
		0.089024,
		0.070692,
		0.054958,
		0.041695,
		0.030743,
		0.021915,
		0.014998,
		0.009761,
		0.005962,
		0.003351,
		0.001682,
		0.000717,
		0.000236,
		4.8e-005,
		4.9e-005,
		0.000236,
		0.000719,
		0.001686,
		0.00336,
		0.005977,
		0.009787,
		0.015038,
		0.021976,
		0.030831,
		0.041817,
		0.055125,
		0.070914,
		0.089312,
		0.110408,
		0.13425,
		0.160839,
		0.190126,
		0.222012,
		0.256343,
		0.292911,
		0.331456,
		0.371666,
		0.413186,
		0.455619,
		0.498541,
		0.541506,
		0.584062,
		0.62576,
		0.666171,
		0.704898,
		0.741583,
		0.775926,
		0.807686,
		0.836688,
		0.862827,
		0.886069,
		0.906441,
		0.924032,
		0.938984,
		0.951479,
		0.961733,
		0.969981,
		0.976471,
		0.981452,
		0.985163,
		0.987832,
		0.989664,
		0.990841,
		0.99152,
		0.991829,
		0.991872,
		0.991729,
		0.991457,
		0.991097,
		0.990675,
		0.990206,
		0.989698,
		0.989153,
		0.988571,
		0.987949,
		0.987285,
		0.986574,
		0.985814,
		0.984999,
		0.984125,
		0.983187,
		0.982179,
		0.981094,
		0.979927,
		0.978669,
		0.977312,
		0.975845,
		0.97426,
		0.972543,
		0.970681,
		0.968661,
		0.966466,
		0.964076,
		0.961473,
		0.958631,
		0.955526,
		0.952128,
		0.948403,
		0.944313,
		0.939815,
		0.93486,
		0.929393,
		0.92335,
		0.916658,
		0.909232,
		0.900978,
		0.891784,
		0.881522,
		0.870045,
		0.857183,
		0.842698,
		0.825756,
		0.803188,
		0.767694,
		0.708972,
		0.61919,
		0.5,
		0.365257,
		0.236018,
		0.131026,
		0.05954,
		0.020142,
		0.00408,
		0.000248,
		0.000248,
		0.00408,
		0.020142,
		0.05954,
		0.131026,
		0.236018,
		0.365257,
		0.5,
		0.61919,
		0.708972,
		0.767694,
		0.803188,
		0.825756,
		0.842698,
		0.857183,
		0.870045,
		0.881522,
		0.891784,
		0.900978,
		0.909232,
		0.916658,
		0.92335,
		0.929393,
		0.93486,
		0.939815,
		0.944313,
		0.948403,
		0.952128,
		0.955526,
		0.958631,
		0.961473,
		0.964076,
		0.966466,
		0.968661,
		0.970681,
		0.972543,
		0.97426,
		0.975845,
		0.977312,
		0.978669,
		0.979927,
		0.981094,
		0.982179,
		0.983187,
		0.984125,
		0.984999,
		0.985814,
		0.986574,
		0.987285,
		0.987949,
		0.988571,
		0.989153,
		0.989698,
		0.990206,
		0.990675,
		0.991097,
		0.991457,
		0.991729,
		0.991872,
		0.991829,
		0.99152,
		0.990841,
		0.989664,
		0.987832,
		0.985163,
		0.981452,
		0.976471,
		0.969981,
		0.961733,
		0.951479,
		0.938984,
		0.924032,
		0.906441,
		0.886069,
		0.862827,
		0.836688,
		0.807686,
		0.775926,
		0.741583,
		0.704898,
		0.666171,
		0.62576,
		0.584062,
		0.541506,
		0.498541,
		0.455619,
		0.413186,
		0.371666,
		0.331456,
		0.292911,
		0.256343,
		0.222012,
		0.190126,
		0.160839,
		0.13425,
		0.110408,
		0.089312,
		0.070914,
		0.055125,
		0.041817,
		0.030831,
		0.021976,
		0.015038,
		0.009787,
		0.005977,
		0.00336,
		0.001686,
		0.000719,
		0.000236,
		4.9e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003363,
		0.005983,
		0.009797,
		0.015055,
		0.022001,
		0.030867,
		0.041868,
		0.055194,
		0.071006,
		0.089431,
		0.110561,
		0.134442,
		0.161077,
		0.190417,
		0.222364,
		0.256763,
		0.293409,
		0.332039,
		0.372344,
		0.413966,
		0.456511,
		0.499553,
		0.542647,
		0.585338,
		0.627179,
		0.66774,
		0.706621,
		0.743467,
		0.777975,
		0.809903,
		0.839077,
		0.865392,
		0.88881,
		0.909361,
		0.927134,
		0.942269,
		0.95495,
		0.965392,
		0.973833,
		0.98052,
		0.985703,
		0.989623,
		0.992508,
		0.994565,
		0.995977,
		0.996902,
		0.997471,
		0.997787,
		0.997932,
		0.997966,
		0.99793,
		0.997853,
		0.99775,
		0.997631,
		0.997501,
		0.997361,
		0.99721,
		0.997049,
		0.996876,
		0.99669,
		0.99649,
		0.996274,
		0.996042,
		0.995792,
		0.995521,
		0.995229,
		0.994912,
		0.994569,
		0.994196,
		0.993791,
		0.99335,
		0.99287,
		0.992346,
		0.991772,
		0.991144,
		0.990456,
		0.9897,
		0.988867,
		0.987949,
		0.986936,
		0.985814,
		0.984569,
		0.983187,
		0.981647,
		0.979927,
		0.978003,
		0.975845,
		0.973418,
		0.970681,
		0.967586,
		0.964076,
		0.960083,
		0.955481,
		0.949475,
		0.938519,
		0.913908,
		0.862495,
		0.77265,
		0.642924,
		0.486772,
		0.328396,
		0.192162,
		0.09321,
		0.034243,
		0.007717,
		0.00054,
		0.00054,
		0.007717,
		0.034243,
		0.09321,
		0.192162,
		0.328396,
		0.486772,
		0.642924,
		0.77265,
		0.862495,
		0.913908,
		0.938519,
		0.949475,
		0.955481,
		0.960083,
		0.964076,
		0.967586,
		0.970681,
		0.973418,
		0.975845,
		0.978003,
		0.979927,
		0.981647,
		0.983187,
		0.984569,
		0.985814,
		0.986936,
		0.987949,
		0.988867,
		0.9897,
		0.990456,
		0.991144,
		0.991772,
		0.992346,
		0.99287,
		0.99335,
		0.993791,
		0.994196,
		0.994569,
		0.994912,
		0.995229,
		0.995521,
		0.995792,
		0.996042,
		0.996274,
		0.99649,
		0.99669,
		0.996876,
		0.997049,
		0.99721,
		0.997361,
		0.997501,
		0.997631,
		0.99775,
		0.997853,
		0.99793,
		0.997966,
		0.997932,
		0.997787,
		0.997471,
		0.996902,
		0.995977,
		0.994565,
		0.992508,
		0.989623,
		0.985703,
		0.98052,
		0.973833,
		0.965392,
		0.95495,
		0.942269,
		0.927134,
		0.909361,
		0.88881,
		0.865392,
		0.839077,
		0.809903,
		0.777975,
		0.743467,
		0.706621,
		0.66774,
		0.627179,
		0.585338,
		0.542647,
		0.499553,
		0.456511,
		0.413966,
		0.372344,
		0.332039,
		0.293409,
		0.256763,
		0.222364,
		0.190417,
		0.161077,
		0.134442,
		0.110561,
		0.089431,
		0.071006,
		0.055194,
		0.041868,
		0.030867,
		0.022001,
		0.015055,
		0.009797,
		0.005983,
		0.003363,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022007,
		0.030877,
		0.041881,
		0.055212,
		0.071029,
		0.089462,
		0.1106,
		0.134492,
		0.161138,
		0.190492,
		0.222455,
		0.256872,
		0.293537,
		0.33219,
		0.37252,
		0.414169,
		0.456743,
		0.499817,
		0.542944,
		0.585671,
		0.62755,
		0.66815,
		0.707073,
		0.743962,
		0.778514,
		0.810487,
		0.839707,
		0.866069,
		0.889535,
		0.910135,
		0.927957,
		0.943143,
		0.955875,
		0.966369,
		0.974863,
		0.981605,
		0.986844,
		0.990823,
		0.993768,
		0.995889,
		0.997368,
		0.998363,
		0.999005,
		0.999399,
		0.999627,
		0.999749,
		0.999806,
		0.999829,
		0.999832,
		0.999827,
		0.999818,
		0.999808,
		0.999797,
		0.999784,
		0.999771,
		0.999757,
		0.999741,
		0.999725,
		0.999707,
		0.999687,
		0.999666,
		0.999643,
		0.999618,
		0.999591,
		0.999561,
		0.999529,
		0.999493,
		0.999454,
		0.999412,
		0.999365,
		0.999313,
		0.999256,
		0.999193,
		0.999123,
		0.999046,
		0.99896,
		0.998863,
		0.998756,
		0.998635,
		0.998499,
		0.998347,
		0.998174,
		0.997978,
		0.997755,
		0.997501,
		0.99721,
		0.996876,
		0.996489,
		0.995994,
		0.994648,
		0.988808,
		0.969283,
		0.921848,
		0.833327,
		0.700845,
		0.537373,
		0.368043,
		0.219306,
		0.108744,
		0.041042,
		0.009563,
		0.000698,
		0.000698,
		0.009563,
		0.041042,
		0.108744,
		0.219306,
		0.368043,
		0.537373,
		0.700845,
		0.833327,
		0.921848,
		0.969283,
		0.988808,
		0.994648,
		0.995994,
		0.996489,
		0.996876,
		0.99721,
		0.997501,
		0.997755,
		0.997978,
		0.998174,
		0.998347,
		0.998499,
		0.998635,
		0.998756,
		0.998863,
		0.99896,
		0.999046,
		0.999123,
		0.999193,
		0.999256,
		0.999313,
		0.999365,
		0.999412,
		0.999454,
		0.999493,
		0.999529,
		0.999561,
		0.999591,
		0.999618,
		0.999643,
		0.999666,
		0.999687,
		0.999707,
		0.999725,
		0.999741,
		0.999757,
		0.999771,
		0.999784,
		0.999797,
		0.999808,
		0.999818,
		0.999827,
		0.999832,
		0.999829,
		0.999806,
		0.999749,
		0.999627,
		0.999399,
		0.999005,
		0.998363,
		0.997368,
		0.995889,
		0.993768,
		0.990823,
		0.986844,
		0.981605,
		0.974863,
		0.966369,
		0.955875,
		0.943143,
		0.927957,
		0.910135,
		0.889535,
		0.866069,
		0.839707,
		0.810487,
		0.778514,
		0.743962,
		0.707073,
		0.66815,
		0.62755,
		0.585671,
		0.542944,
		0.499817,
		0.456743,
		0.414169,
		0.37252,
		0.33219,
		0.293537,
		0.256872,
		0.222455,
		0.190492,
		0.161138,
		0.134492,
		0.1106,
		0.089462,
		0.071029,
		0.055212,
		0.041881,
		0.030877,
		0.022007,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041882,
		0.055213,
		0.071031,
		0.089465,
		0.110603,
		0.134496,
		0.161143,
		0.190498,
		0.222461,
		0.25688,
		0.293547,
		0.332201,
		0.372533,
		0.414184,
		0.456761,
		0.499836,
		0.542966,
		0.585696,
		0.627578,
		0.668181,
		0.707107,
		0.743999,
		0.778554,
		0.810531,
		0.839755,
		0.86612,
		0.88959,
		0.910193,
		0.928019,
		0.943208,
		0.955945,
		0.966443,
		0.974941,
		0.981688,
		0.986931,
		0.990914,
		0.993865,
		0.99599,
		0.997474,
		0.998475,
		0.999123,
		0.999523,
		0.999758,
		0.999887,
		0.999952,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999952,
		0.999123,
		0.993865,
		0.974941,
		0.928019,
		0.839755,
		0.707107,
		0.542966,
		0.372533,
		0.222461,
		0.110603,
		0.041882,
		0.0098,
		0.000719,
		0.000719,
		0.0098,
		0.041882,
		0.110603,
		0.222461,
		0.372533,
		0.542966,
		0.707107,
		0.839755,
		0.928019,
		0.974941,
		0.993865,
		0.999123,
		0.999952,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999952,
		0.999887,
		0.999758,
		0.999523,
		0.999123,
		0.998475,
		0.997474,
		0.99599,
		0.993865,
		0.990914,
		0.986931,
		0.981688,
		0.974941,
		0.966443,
		0.955945,
		0.943208,
		0.928019,
		0.910193,
		0.88959,
		0.86612,
		0.839755,
		0.810531,
		0.778554,
		0.743999,
		0.707107,
		0.668181,
		0.627578,
		0.585696,
		0.542966,
		0.499836,
		0.456761,
		0.414184,
		0.372533,
		0.332201,
		0.293547,
		0.25688,
		0.222461,
		0.190498,
		0.161143,
		0.134496,
		0.110603,
		0.089465,
		0.071031,
		0.055213,
		0.041882,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022007,
		0.030877,
		0.041881,
		0.055212,
		0.071029,
		0.089462,
		0.1106,
		0.134492,
		0.161138,
		0.190492,
		0.222455,
		0.256872,
		0.293537,
		0.33219,
		0.37252,
		0.414169,
		0.456743,
		0.499817,
		0.542944,
		0.585671,
		0.62755,
		0.66815,
		0.707073,
		0.743962,
		0.778514,
		0.810487,
		0.839707,
		0.866069,
		0.889535,
		0.910135,
		0.927957,
		0.943143,
		0.955875,
		0.966369,
		0.974863,
		0.981605,
		0.986844,
		0.990823,
		0.993768,
		0.995889,
		0.997368,
		0.998363,
		0.999005,
		0.999399,
		0.999627,
		0.999749,
		0.999806,
		0.999829,
		0.999832,
		0.999827,
		0.999818,
		0.999808,
		0.999797,
		0.999784,
		0.999771,
		0.999757,
		0.999741,
		0.999725,
		0.999707,
		0.999687,
		0.999666,
		0.999643,
		0.999618,
		0.999591,
		0.999561,
		0.999529,
		0.999493,
		0.999454,
		0.999412,
		0.999365,
		0.999313,
		0.999256,
		0.999193,
		0.999123,
		0.999046,
		0.99896,
		0.998863,
		0.998756,
		0.998635,
		0.998499,
		0.998347,
		0.998174,
		0.997978,
		0.997755,
		0.997501,
		0.99721,
		0.996876,
		0.996489,
		0.995994,
		0.994648,
		0.988808,
		0.969283,
		0.921848,
		0.833327,
		0.700845,
		0.537373,
		0.368043,
		0.219306,
		0.108744,
		0.041042,
		0.009563,
		0.000698,
		0.000698,
		0.009563,
		0.041042,
		0.108744,
		0.219306,
		0.368043,
		0.537373,
		0.700845,
		0.833327,
		0.921848,
		0.969283,
		0.988808,
		0.994648,
		0.995994,
		0.996489,
		0.996876,
		0.99721,
		0.997501,
		0.997755,
		0.997978,
		0.998174,
		0.998347,
		0.998499,
		0.998635,
		0.998756,
		0.998863,
		0.99896,
		0.999046,
		0.999123,
		0.999193,
		0.999256,
		0.999313,
		0.999365,
		0.999412,
		0.999454,
		0.999493,
		0.999529,
		0.999561,
		0.999591,
		0.999618,
		0.999643,
		0.999666,
		0.999687,
		0.999707,
		0.999725,
		0.999741,
		0.999757,
		0.999771,
		0.999784,
		0.999797,
		0.999808,
		0.999818,
		0.999827,
		0.999832,
		0.999829,
		0.999806,
		0.999749,
		0.999627,
		0.999399,
		0.999005,
		0.998363,
		0.997368,
		0.995889,
		0.993768,
		0.990823,
		0.986844,
		0.981605,
		0.974863,
		0.966369,
		0.955875,
		0.943143,
		0.927957,
		0.910135,
		0.889535,
		0.866069,
		0.839707,
		0.810487,
		0.778514,
		0.743962,
		0.707073,
		0.66815,
		0.62755,
		0.585671,
		0.542944,
		0.499817,
		0.456743,
		0.414169,
		0.37252,
		0.33219,
		0.293537,
		0.256872,
		0.222455,
		0.190492,
		0.161138,
		0.134492,
		0.1106,
		0.089462,
		0.071029,
		0.055212,
		0.041881,
		0.030877,
		0.022007,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003363,
		0.005983,
		0.009797,
		0.015055,
		0.022001,
		0.030867,
		0.041868,
		0.055194,
		0.071006,
		0.089431,
		0.110561,
		0.134442,
		0.161077,
		0.190417,
		0.222364,
		0.256763,
		0.293409,
		0.332039,
		0.372344,
		0.413966,
		0.456511,
		0.499553,
		0.542647,
		0.585338,
		0.627179,
		0.66774,
		0.706621,
		0.743467,
		0.777975,
		0.809903,
		0.839077,
		0.865392,
		0.88881,
		0.909361,
		0.927134,
		0.942269,
		0.95495,
		0.965392,
		0.973833,
		0.98052,
		0.985703,
		0.989623,
		0.992508,
		0.994565,
		0.995977,
		0.996902,
		0.997471,
		0.997787,
		0.997932,
		0.997966,
		0.99793,
		0.997853,
		0.99775,
		0.997631,
		0.997501,
		0.997361,
		0.99721,
		0.997049,
		0.996876,
		0.99669,
		0.99649,
		0.996274,
		0.996042,
		0.995792,
		0.995521,
		0.995229,
		0.994912,
		0.994569,
		0.994196,
		0.993791,
		0.99335,
		0.99287,
		0.992346,
		0.991772,
		0.991144,
		0.990456,
		0.9897,
		0.988867,
		0.987949,
		0.986936,
		0.985814,
		0.984569,
		0.983187,
		0.981647,
		0.979927,
		0.978003,
		0.975845,
		0.973418,
		0.970681,
		0.967586,
		0.964076,
		0.960083,
		0.955481,
		0.949475,
		0.938519,
		0.913908,
		0.862495,
		0.77265,
		0.642924,
		0.486772,
		0.328396,
		0.192162,
		0.09321,
		0.034243,
		0.007717,
		0.00054,
		0.00054,
		0.007717,
		0.034243,
		0.09321,
		0.192162,
		0.328396,
		0.486772,
		0.642924,
		0.77265,
		0.862495,
		0.913908,
		0.938519,
		0.949475,
		0.955481,
		0.960083,
		0.964076,
		0.967586,
		0.970681,
		0.973418,
		0.975845,
		0.978003,
		0.979927,
		0.981647,
		0.983187,
		0.984569,
		0.985814,
		0.986936,
		0.987949,
		0.988867,
		0.9897,
		0.990456,
		0.991144,
		0.991772,
		0.992346,
		0.99287,
		0.99335,
		0.993791,
		0.994196,
		0.994569,
		0.994912,
		0.995229,
		0.995521,
		0.995792,
		0.996042,
		0.996274,
		0.99649,
		0.99669,
		0.996876,
		0.997049,
		0.99721,
		0.997361,
		0.997501,
		0.997631,
		0.99775,
		0.997853,
		0.99793,
		0.997966,
		0.997932,
		0.997787,
		0.997471,
		0.996902,
		0.995977,
		0.994565,
		0.992508,
		0.989623,
		0.985703,
		0.98052,
		0.973833,
		0.965392,
		0.95495,
		0.942269,
		0.927134,
		0.909361,
		0.88881,
		0.865392,
		0.839077,
		0.809903,
		0.777975,
		0.743467,
		0.706621,
		0.66774,
		0.627179,
		0.585338,
		0.542647,
		0.499553,
		0.456511,
		0.413966,
		0.372344,
		0.332039,
		0.293409,
		0.256763,
		0.222364,
		0.190417,
		0.161077,
		0.134442,
		0.110561,
		0.089431,
		0.071006,
		0.055194,
		0.041868,
		0.030867,
		0.022001,
		0.015055,
		0.009797,
		0.005983,
		0.003363,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		4.9e-005,
		0.000236,
		0.000719,
		0.001686,
		0.00336,
		0.005977,
		0.009787,
		0.015038,
		0.021976,
		0.030831,
		0.041817,
		0.055125,
		0.070914,
		0.089312,
		0.110408,
		0.13425,
		0.160839,
		0.190126,
		0.222012,
		0.256343,
		0.292911,
		0.331456,
		0.371666,
		0.413186,
		0.455619,
		0.498541,
		0.541506,
		0.584062,
		0.62576,
		0.666171,
		0.704898,
		0.741583,
		0.775926,
		0.807686,
		0.836688,
		0.862827,
		0.886069,
		0.906441,
		0.924032,
		0.938984,
		0.951479,
		0.961733,
		0.969981,
		0.976471,
		0.981452,
		0.985163,
		0.987832,
		0.989664,
		0.990841,
		0.99152,
		0.991829,
		0.991872,
		0.991729,
		0.991457,
		0.991097,
		0.990675,
		0.990206,
		0.989698,
		0.989153,
		0.988571,
		0.987949,
		0.987285,
		0.986574,
		0.985814,
		0.984999,
		0.984125,
		0.983187,
		0.982179,
		0.981094,
		0.979927,
		0.978669,
		0.977312,
		0.975845,
		0.97426,
		0.972543,
		0.970681,
		0.968661,
		0.966466,
		0.964076,
		0.961473,
		0.958631,
		0.955526,
		0.952128,
		0.948403,
		0.944313,
		0.939815,
		0.93486,
		0.929393,
		0.92335,
		0.916658,
		0.909232,
		0.900978,
		0.891784,
		0.881522,
		0.870045,
		0.857183,
		0.842698,
		0.825756,
		0.803188,
		0.767694,
		0.708972,
		0.61919,
		0.5,
		0.365257,
		0.236018,
		0.131026,
		0.05954,
		0.020142,
		0.00408,
		0.000248,
		0.000248,
		0.00408,
		0.020142,
		0.05954,
		0.131026,
		0.236018,
		0.365257,
		0.5,
		0.61919,
		0.708972,
		0.767694,
		0.803188,
		0.825756,
		0.842698,
		0.857183,
		0.870045,
		0.881522,
		0.891784,
		0.900978,
		0.909232,
		0.916658,
		0.92335,
		0.929393,
		0.93486,
		0.939815,
		0.944313,
		0.948403,
		0.952128,
		0.955526,
		0.958631,
		0.961473,
		0.964076,
		0.966466,
		0.968661,
		0.970681,
		0.972543,
		0.97426,
		0.975845,
		0.977312,
		0.978669,
		0.979927,
		0.981094,
		0.982179,
		0.983187,
		0.984125,
		0.984999,
		0.985814,
		0.986574,
		0.987285,
		0.987949,
		0.988571,
		0.989153,
		0.989698,
		0.990206,
		0.990675,
		0.991097,
		0.991457,
		0.991729,
		0.991872,
		0.991829,
		0.99152,
		0.990841,
		0.989664,
		0.987832,
		0.985163,
		0.981452,
		0.976471,
		0.969981,
		0.961733,
		0.951479,
		0.938984,
		0.924032,
		0.906441,
		0.886069,
		0.862827,
		0.836688,
		0.807686,
		0.775926,
		0.741583,
		0.704898,
		0.666171,
		0.62576,
		0.584062,
		0.541506,
		0.498541,
		0.455619,
		0.413186,
		0.371666,
		0.331456,
		0.292911,
		0.256343,
		0.222012,
		0.190126,
		0.160839,
		0.13425,
		0.110408,
		0.089312,
		0.070914,
		0.055125,
		0.041817,
		0.030831,
		0.021976,
		0.015038,
		0.009787,
		0.005977,
		0.00336,
		0.001686,
		0.000719,
		0.000236,
		4.9e-005,
		4.8e-005,
		0.000236,
		0.000717,
		0.001682,
		0.003351,
		0.005962,
		0.009761,
		0.014998,
		0.021915,
		0.030743,
		0.041695,
		0.054958,
		0.070692,
		0.089024,
		0.110041,
		0.133789,
		0.160268,
		0.189429,
		0.22117,
		0.255338,
		0.291724,
		0.330067,
		0.370055,
		0.411333,
		0.453504,
		0.496144,
		0.53881,
		0.581049,
		0.622414,
		0.662478,
		0.700845,
		0.73716,
		0.771123,
		0.802496,
		0.831105,
		0.856847,
		0.879686,
		0.899653,
		0.916836,
		0.931376,
		0.943456,
		0.953289,
		0.961111,
		0.967166,
		0.971702,
		0.974957,
		0.977154,
		0.978498,
		0.979167,
		0.979315,
		0.979067,
		0.978526,
		0.977766,
		0.976845,
		0.975798,
		0.97465,
		0.973413,
		0.97209,
		0.970681,
		0.969182,
		0.967586,
		0.965887,
		0.964076,
		0.962145,
		0.960083,
		0.957881,
		0.955526,
		0.953007,
		0.950308,
		0.947416,
		0.944313,
		0.94098,
		0.937398,
		0.933544,
		0.929393,
		0.924919,
		0.92009,
		0.914874,
		0.909232,
		0.903125,
		0.896506,
		0.889324,
		0.881522,
		0.873037,
		0.863799,
		0.853729,
		0.842738,
		0.830729,
		0.817592,
		0.803203,
		0.787425,
		0.770107,
		0.751077,
		0.730146,
		0.707107,
		0.681729,
		0.653732,
		0.622395,
		0.585368,
		0.537774,
		0.473757,
		0.390871,
		0.294361,
		0.197077,
		0.113883,
		0.054533,
		0.020197,
		0.005051,
		0.000623,
		1.4e-005,
		1.4e-005,
		0.000623,
		0.005051,
		0.020197,
		0.054533,
		0.113883,
		0.197077,
		0.294361,
		0.390871,
		0.473757,
		0.537774,
		0.585368,
		0.622395,
		0.653732,
		0.681729,
		0.707107,
		0.730146,
		0.751077,
		0.770107,
		0.787425,
		0.803203,
		0.817592,
		0.830729,
		0.842738,
		0.853729,
		0.863799,
		0.873037,
		0.881522,
		0.889324,
		0.896506,
		0.903125,
		0.909232,
		0.914874,
		0.92009,
		0.924919,
		0.929393,
		0.933544,
		0.937398,
		0.94098,
		0.944313,
		0.947416,
		0.950308,
		0.953007,
		0.955526,
		0.957881,
		0.960083,
		0.962145,
		0.964076,
		0.965887,
		0.967586,
		0.969182,
		0.970681,
		0.97209,
		0.973413,
		0.97465,
		0.975798,
		0.976845,
		0.977766,
		0.978526,
		0.979067,
		0.979315,
		0.979167,
		0.978498,
		0.977154,
		0.974957,
		0.971702,
		0.967166,
		0.961111,
		0.953289,
		0.943456,
		0.931376,
		0.916836,
		0.899653,
		0.879686,
		0.856847,
		0.831105,
		0.802496,
		0.771123,
		0.73716,
		0.700845,
		0.662478,
		0.622414,
		0.581049,
		0.53881,
		0.496144,
		0.453504,
		0.411333,
		0.370055,
		0.330067,
		0.291724,
		0.255338,
		0.22117,
		0.189429,
		0.160268,
		0.133789,
		0.110041,
		0.089024,
		0.070692,
		0.054958,
		0.041695,
		0.030743,
		0.021915,
		0.014998,
		0.009761,
		0.005962,
		0.003351,
		0.001682,
		0.000717,
		0.000236,
		4.8e-005,
		4.8e-005,
		0.000235,
		0.000714,
		0.001675,
		0.003336,
		0.005933,
		0.009713,
		0.014922,
		0.021801,
		0.030578,
		0.041464,
		0.054644,
		0.070277,
		0.088484,
		0.109353,
		0.132926,
		0.159201,
		0.188127,
		0.219602,
		0.253469,
		0.289518,
		0.327489,
		0.367068,
		0.4079,
		0.449592,
		0.491718,
		0.533837,
		0.575499,
		0.616261,
		0.655698,
		0.693416,
		0.729065,
		0.762347,
		0.793028,
		0.820937,
		0.845973,
		0.868102,
		0.887355,
		0.903823,
		0.917645,
		0.929002,
		0.938108,
		0.945195,
		0.950506,
		0.954283,
		0.956762,
		0.958161,
		0.958682,
		0.958497,
		0.957757,
		0.956583,
		0.955071,
		0.953294,
		0.951302,
		0.94913,
		0.946796,
		0.944307,
		0.941665,
		0.938862,
		0.93589,
		0.932739,
		0.929393,
		0.925841,
		0.922066,
		0.918052,
		0.913781,
		0.909232,
		0.904386,
		0.899218,
		0.893703,
		0.887815,
		0.881522,
		0.874792,
		0.86759,
		0.859876,
		0.851607,
		0.842738,
		0.833218,
		0.82299,
		0.811993,
		0.800163,
		0.787425,
		0.773702,
		0.758906,
		0.742944,
		0.725714,
		0.707107,
		0.687003,
		0.665277,
		0.641793,
		0.61641,
		0.588982,
		0.559361,
		0.527402,
		0.492972,
		0.455958,
		0.416269,
		0.373628,
		0.327031,
		0.274752,
		0.215973,
		0.153342,
		0.093895,
		0.046363,
		0.01627,
		0.002816,
		0.002816,
		0.01627,
		0.046363,
		0.093895,
		0.153342,
		0.215973,
		0.274752,
		0.327031,
		0.373628,
		0.416269,
		0.455958,
		0.492972,
		0.527402,
		0.559361,
		0.588982,
		0.61641,
		0.641793,
		0.665277,
		0.687003,
		0.707107,
		0.725714,
		0.742944,
		0.758906,
		0.773702,
		0.787425,
		0.800163,
		0.811993,
		0.82299,
		0.833218,
		0.842738,
		0.851607,
		0.859876,
		0.86759,
		0.874792,
		0.881522,
		0.887815,
		0.893703,
		0.899218,
		0.904386,
		0.909232,
		0.913781,
		0.918052,
		0.922066,
		0.925841,
		0.929393,
		0.932739,
		0.93589,
		0.938862,
		0.941665,
		0.944307,
		0.946796,
		0.94913,
		0.951302,
		0.953294,
		0.955071,
		0.956583,
		0.957757,
		0.958497,
		0.958682,
		0.958161,
		0.956762,
		0.954283,
		0.950506,
		0.945195,
		0.938108,
		0.929002,
		0.917645,
		0.903823,
		0.887355,
		0.868102,
		0.845973,
		0.820937,
		0.793028,
		0.762347,
		0.729065,
		0.693416,
		0.655698,
		0.616261,
		0.575499,
		0.533837,
		0.491718,
		0.449592,
		0.4079,
		0.367068,
		0.327489,
		0.289518,
		0.253469,
		0.219602,
		0.188127,
		0.159201,
		0.132926,
		0.109353,
		0.088484,
		0.070277,
		0.054644,
		0.041464,
		0.030578,
		0.021801,
		0.014922,
		0.009713,
		0.005933,
		0.003336,
		0.001675,
		0.000714,
		0.000235,
		4.8e-005,
		4.8e-005,
		0.000233,
		0.000709,
		0.001662,
		0.003311,
		0.005887,
		0.009635,
		0.014799,
		0.021616,
		0.030311,
		0.041091,
		0.054138,
		0.069605,
		0.087614,
		0.108244,
		0.131537,
		0.157487,
		0.186039,
		0.217088,
		0.250476,
		0.285991,
		0.323371,
		0.362304,
		0.402434,
		0.443369,
		0.484687,
		0.52595,
		0.566711,
		0.606532,
		0.644993,
		0.681705,
		0.716324,
		0.748558,
		0.778177,
		0.805015,
		0.828975,
		0.850026,
		0.868202,
		0.883593,
		0.896339,
		0.90662,
		0.914647,
		0.92065,
		0.924865,
		0.927533,
		0.928881,
		0.929125,
		0.928457,
		0.927046,
		0.925035,
		0.92254,
		0.919651,
		0.916436,
		0.912939,
		0.909189,
		0.905199,
		0.900973,
		0.896504,
		0.891783,
		0.886795,
		0.881522,
		0.875946,
		0.870045,
		0.863799,
		0.857183,
		0.850172,
		0.842738,
		0.834852,
		0.826481,
		0.817592,
		0.808146,
		0.798105,
		0.787425,
		0.776061,
		0.763962,
		0.751077,
		0.737347,
		0.722712,
		0.707107,
		0.690463,
		0.672708,
		0.653763,
		0.63355,
		0.611985,
		0.588982,
		0.564456,
		0.538323,
		0.510503,
		0.480925,
		0.449532,
		0.416289,
		0.381194,
		0.344293,
		0.3057,
		0.265625,
		0.224414,
		0.182595,
		0.140868,
		0.100129,
		0.061965,
		0.029542,
		0.007571,
		0.007571,
		0.029542,
		0.061965,
		0.100129,
		0.140868,
		0.182595,
		0.224414,
		0.265625,
		0.3057,
		0.344293,
		0.381194,
		0.416289,
		0.449532,
		0.480925,
		0.510503,
		0.538323,
		0.564456,
		0.588982,
		0.611985,
		0.63355,
		0.653763,
		0.672708,
		0.690463,
		0.707107,
		0.722712,
		0.737347,
		0.751077,
		0.763962,
		0.776061,
		0.787425,
		0.798105,
		0.808146,
		0.817592,
		0.826481,
		0.834852,
		0.842738,
		0.850172,
		0.857183,
		0.863799,
		0.870045,
		0.875946,
		0.881522,
		0.886795,
		0.891783,
		0.896504,
		0.900973,
		0.905199,
		0.909189,
		0.912939,
		0.916436,
		0.919651,
		0.92254,
		0.925035,
		0.927046,
		0.928457,
		0.929125,
		0.928881,
		0.927533,
		0.924865,
		0.92065,
		0.914647,
		0.90662,
		0.896339,
		0.883593,
		0.868202,
		0.850026,
		0.828975,
		0.805015,
		0.778177,
		0.748558,
		0.716324,
		0.681705,
		0.644993,
		0.606532,
		0.566711,
		0.52595,
		0.484687,
		0.443369,
		0.402434,
		0.362304,
		0.323371,
		0.285991,
		0.250476,
		0.217088,
		0.186039,
		0.157487,
		0.131537,
		0.108244,
		0.087614,
		0.069605,
		0.054138,
		0.041091,
		0.030311,
		0.021616,
		0.014799,
		0.009635,
		0.005887,
		0.003311,
		0.001662,
		0.000709,
		0.000233,
		4.8e-005,
		4.7e-005,
		0.000231,
		0.000701,
		0.001644,
		0.003273,
		0.005819,
		0.009521,
		0.014618,
		0.021344,
		0.029919,
		0.040544,
		0.053397,
		0.068626,
		0.086344,
		0.10663,
		0.129518,
		0.154997,
		0.183009,
		0.213445,
		0.246144,
		0.280893,
		0.317427,
		0.355436,
		0.394565,
		0.434424,
		0.474596,
		0.514646,
		0.554134,
		0.59263,
		0.629719,
		0.665023,
		0.698204,
		0.72898,
		0.757127,
		0.782486,
		0.804966,
		0.824541,
		0.841247,
		0.855177,
		0.86647,
		0.875305,
		0.881889,
		0.886448,
		0.889215,
		0.890421,
		0.890289,
		0.889025,
		0.886816,
		0.883821,
		0.880177,
		0.875992,
		0.87135,
		0.86631,
		0.860913,
		0.855181,
		0.849122,
		0.842734,
		0.836006,
		0.828924,
		0.821467,
		0.813614,
		0.80534,
		0.796619,
		0.787425,
		0.777728,
		0.767497,
		0.756699,
		0.745299,
		0.73326,
		0.720543,
		0.707107,
		0.692907,
		0.677898,
		0.662033,
		0.645261,
		0.62753,
		0.608789,
		0.588982,
		0.568056,
		0.545959,
		0.522638,
		0.498046,
		0.472143,
		0.444897,
		0.416289,
		0.38632,
		0.355015,
		0.322436,
		0.28869,
		0.253947,
		0.218463,
		0.182603,
		0.146884,
		0.112019,
		0.078987,
		0.049119,
		0.024222,
		0.00674,
		0.00674,
		0.024222,
		0.049119,
		0.078987,
		0.112019,
		0.146884,
		0.182603,
		0.218463,
		0.253947,
		0.28869,
		0.322436,
		0.355015,
		0.38632,
		0.416289,
		0.444897,
		0.472143,
		0.498046,
		0.522638,
		0.545959,
		0.568056,
		0.588982,
		0.608789,
		0.62753,
		0.645261,
		0.662033,
		0.677898,
		0.692907,
		0.707107,
		0.720543,
		0.73326,
		0.745299,
		0.756699,
		0.767497,
		0.777728,
		0.787425,
		0.796619,
		0.80534,
		0.813614,
		0.821467,
		0.828924,
		0.836006,
		0.842734,
		0.849122,
		0.855181,
		0.860913,
		0.86631,
		0.87135,
		0.875992,
		0.880177,
		0.883821,
		0.886816,
		0.889025,
		0.890289,
		0.890421,
		0.889215,
		0.886448,
		0.881889,
		0.875305,
		0.86647,
		0.855177,
		0.841247,
		0.824541,
		0.804966,
		0.782486,
		0.757127,
		0.72898,
		0.698204,
		0.665023,
		0.629719,
		0.59263,
		0.554134,
		0.514646,
		0.474596,
		0.434424,
		0.394565,
		0.355436,
		0.317427,
		0.280893,
		0.246144,
		0.213445,
		0.183009,
		0.154997,
		0.129518,
		0.10663,
		0.086344,
		0.068626,
		0.053397,
		0.040544,
		0.029919,
		0.021344,
		0.014618,
		0.009521,
		0.005819,
		0.003273,
		0.001644,
		0.000701,
		0.000231,
		4.7e-005,
		4.7e-005,
		0.000228,
		0.000691,
		0.001619,
		0.003222,
		0.005726,
		0.009364,
		0.014371,
		0.020973,
		0.029385,
		0.039801,
		0.052391,
		0.067296,
		0.084624,
		0.104444,
		0.126786,
		0.151632,
		0.17892,
		0.208535,
		0.240313,
		0.274039,
		0.309448,
		0.346229,
		0.384031,
		0.422467,
		0.461124,
		0.499578,
		0.537395,
		0.574154,
		0.609454,
		0.642924,
		0.674241,
		0.703131,
		0.729382,
		0.752845,
		0.773435,
		0.791133,
		0.80598,
		0.818069,
		0.827542,
		0.834575,
		0.839373,
		0.842154,
		0.843145,
		0.842572,
		0.840647,
		0.837568,
		0.833512,
		0.828631,
		0.823051,
		0.816874,
		0.810175,
		0.803008,
		0.795406,
		0.787387,
		0.778955,
		0.770102,
		0.760816,
		0.751076,
		0.740862,
		0.730146,
		0.718904,
		0.707107,
		0.694725,
		0.681729,
		0.668086,
		0.653763,
		0.638727,
		0.622942,
		0.606372,
		0.588982,
		0.570735,
		0.551596,
		0.531529,
		0.510503,
		0.488488,
		0.465459,
		0.441396,
		0.416289,
		0.39014,
		0.362964,
		0.334797,
		0.3057,
		0.275766,
		0.245133,
		0.213992,
		0.182603,
		0.151319,
		0.120607,
		0.091084,
		0.063557,
		0.039085,
		0.019046,
		0.005236,
		0.005236,
		0.019046,
		0.039085,
		0.063557,
		0.091084,
		0.120607,
		0.151319,
		0.182603,
		0.213992,
		0.245133,
		0.275766,
		0.3057,
		0.334797,
		0.362964,
		0.39014,
		0.416289,
		0.441396,
		0.465459,
		0.488488,
		0.510503,
		0.531529,
		0.551596,
		0.570735,
		0.588982,
		0.606372,
		0.622942,
		0.638727,
		0.653763,
		0.668086,
		0.681729,
		0.694725,
		0.707107,
		0.718904,
		0.730146,
		0.740862,
		0.751076,
		0.760816,
		0.770102,
		0.778955,
		0.787387,
		0.795406,
		0.803008,
		0.810175,
		0.816874,
		0.823051,
		0.828631,
		0.833512,
		0.837568,
		0.840647,
		0.842572,
		0.843145,
		0.842154,
		0.839373,
		0.834575,
		0.827542,
		0.818069,
		0.80598,
		0.791133,
		0.773435,
		0.752845,
		0.729382,
		0.703131,
		0.674241,
		0.642924,
		0.609454,
		0.574154,
		0.537395,
		0.499578,
		0.461124,
		0.422467,
		0.384031,
		0.346229,
		0.309448,
		0.274039,
		0.240313,
		0.208535,
		0.17892,
		0.151632,
		0.126786,
		0.104444,
		0.084624,
		0.067296,
		0.052391,
		0.039801,
		0.029385,
		0.020973,
		0.014371,
		0.009364,
		0.005726,
		0.003222,
		0.001619,
		0.000691,
		0.000228,
		4.7e-005,
		4.6e-005,
		0.000223,
		0.000677,
		0.001587,
		0.003156,
		0.005605,
		0.009162,
		0.014052,
		0.020495,
		0.028697,
		0.038844,
		0.051096,
		0.065587,
		0.082414,
		0.101641,
		0.123286,
		0.147328,
		0.173695,
		0.202269,
		0.232882,
		0.265316,
		0.299306,
		0.334543,
		0.370678,
		0.407332,
		0.444099,
		0.480563,
		0.516304,
		0.550912,
		0.584001,
		0.615215,
		0.644244,
		0.67083,
		0.694774,
		0.715937,
		0.734247,
		0.749691,
		0.762317,
		0.772221,
		0.779544,
		0.784462,
		0.787175,
		0.787895,
		0.786843,
		0.784233,
		0.780271,
		0.775143,
		0.769016,
		0.762033,
		0.75431,
		0.745942,
		0.736995,
		0.727518,
		0.71754,
		0.707073,
		0.696118,
		0.684668,
		0.672707,
		0.660215,
		0.647171,
		0.63355,
		0.619329,
		0.604481,
		0.588982,
		0.572806,
		0.555928,
		0.538323,
		0.519969,
		0.500843,
		0.480925,
		0.460201,
		0.438658,
		0.416289,
		0.393097,
		0.369091,
		0.344293,
		0.318742,
		0.292492,
		0.265625,
		0.238249,
		0.210511,
		0.182603,
		0.154777,
		0.127354,
		0.100747,
		0.075484,
		0.05223,
		0.031834,
		0.015366,
		0.004182,
		0.004182,
		0.015366,
		0.031834,
		0.05223,
		0.075484,
		0.100747,
		0.127354,
		0.154777,
		0.182603,
		0.210511,
		0.238249,
		0.265625,
		0.292492,
		0.318742,
		0.344293,
		0.369091,
		0.393097,
		0.416289,
		0.438658,
		0.460201,
		0.480925,
		0.500843,
		0.519969,
		0.538323,
		0.555928,
		0.572806,
		0.588982,
		0.604481,
		0.619329,
		0.63355,
		0.647171,
		0.660215,
		0.672707,
		0.684668,
		0.696118,
		0.707073,
		0.71754,
		0.727518,
		0.736995,
		0.745942,
		0.75431,
		0.762033,
		0.769016,
		0.775143,
		0.780271,
		0.784233,
		0.786843,
		0.787895,
		0.787175,
		0.784462,
		0.779544,
		0.772221,
		0.762317,
		0.749691,
		0.734247,
		0.715937,
		0.694774,
		0.67083,
		0.644244,
		0.615215,
		0.584001,
		0.550912,
		0.516304,
		0.480563,
		0.444099,
		0.407332,
		0.370678,
		0.334543,
		0.299306,
		0.265316,
		0.232882,
		0.202269,
		0.173695,
		0.147328,
		0.123286,
		0.101641,
		0.082414,
		0.065587,
		0.051096,
		0.038844,
		0.028697,
		0.020495,
		0.014052,
		0.009162,
		0.005605,
		0.003156,
		0.001587,
		0.000677,
		0.000223,
		4.6e-005,
		4.5e-005,
		0.000218,
		0.000661,
		0.001547,
		0.003074,
		0.005456,
		0.00891,
		0.013657,
		0.019903,
		0.027847,
		0.037661,
		0.049499,
		0.063481,
		0.079696,
		0.098195,
		0.118991,
		0.142051,
		0.167298,
		0.194608,
		0.223808,
		0.254678,
		0.286955,
		0.320332,
		0.354464,
		0.388981,
		0.423487,
		0.457578,
		0.49085,
		0.522909,
		0.553385,
		0.581942,
		0.608287,
		0.632181,
		0.653439,
		0.67194,
		0.687623,
		0.700485,
		0.71058,
		0.71801,
		0.722916,
		0.725472,
		0.725873,
		0.724327,
		0.721044,
		0.71623,
		0.710079,
		0.702768,
		0.694454,
		0.685268,
		0.67532,
		0.664693,
		0.653452,
		0.641638,
		0.629277,
		0.616381,
		0.60295,
		0.588979,
		0.574454,
		0.559361,
		0.543682,
		0.527402,
		0.510503,
		0.492972,
		0.474793,
		0.455958,
		0.436458,
		0.416289,
		0.395453,
		0.373956,
		0.351814,
		0.32905,
		0.3057,
		0.281814,
		0.25746,
		0.232725,
		0.207724,
		0.182603,
		0.157547,
		0.132788,
		0.108615,
		0.085389,
		0.063557,
		0.043674,
		0.026425,
		0.012657,
		0.003417,
		0.003417,
		0.012657,
		0.026425,
		0.043674,
		0.063557,
		0.085389,
		0.108615,
		0.132788,
		0.157547,
		0.182603,
		0.207724,
		0.232725,
		0.25746,
		0.281814,
		0.3057,
		0.32905,
		0.351814,
		0.373956,
		0.395453,
		0.416289,
		0.436458,
		0.455958,
		0.474793,
		0.492972,
		0.510503,
		0.527402,
		0.543682,
		0.559361,
		0.574454,
		0.588979,
		0.60295,
		0.616381,
		0.629277,
		0.641638,
		0.653452,
		0.664693,
		0.67532,
		0.685268,
		0.694454,
		0.702768,
		0.710079,
		0.71623,
		0.721044,
		0.724327,
		0.725873,
		0.725472,
		0.722916,
		0.71801,
		0.71058,
		0.700485,
		0.687623,
		0.67194,
		0.653439,
		0.632181,
		0.608287,
		0.581942,
		0.553385,
		0.522909,
		0.49085,
		0.457578,
		0.423487,
		0.388981,
		0.354464,
		0.320332,
		0.286955,
		0.254678,
		0.223808,
		0.194608,
		0.167298,
		0.142051,
		0.118991,
		0.098195,
		0.079696,
		0.063481,
		0.049499,
		0.037661,
		0.027847,
		0.019903,
		0.013657,
		0.00891,
		0.005456,
		0.003074,
		0.001547,
		0.000661,
		0.000218,
		4.5e-005,
		4.4e-005,
		0.000211,
		0.00064,
		0.001498,
		0.002975,
		0.005276,
		0.008609,
		0.013183,
		0.019196,
		0.02683,
		0.036251,
		0.047595,
		0.060974,
		0.076464,
		0.094104,
		0.113897,
		0.135801,
		0.159732,
		0.185558,
		0.213104,
		0.242148,
		0.272428,
		0.303641,
		0.33545,
		0.367494,
		0.399391,
		0.430753,
		0.461192,
		0.490336,
		0.517835,
		0.543376,
		0.566689,
		0.587553,
		0.605806,
		0.621342,
		0.634115,
		0.644135,
		0.651464,
		0.656209,
		0.658512,
		0.658546,
		0.656502,
		0.652581,
		0.646984,
		0.639906,
		0.631533,
		0.622029,
		0.611541,
		0.600193,
		0.588084,
		0.575293,
		0.56188,
		0.547885,
		0.533332,
		0.518237,
		0.502605,
		0.486434,
		0.469721,
		0.452462,
		0.434652,
		0.416289,
		0.397375,
		0.377913,
		0.357915,
		0.337397,
		0.316383,
		0.294906,
		0.273008,
		0.250747,
		0.228195,
		0.205443,
		0.182603,
		0.159817,
		0.137255,
		0.11513,
		0.093699,
		0.073275,
		0.05424,
		0.037057,
		0.022285,
		0.010606,
		0.002844,
		0.002844,
		0.010606,
		0.022285,
		0.037057,
		0.05424,
		0.073275,
		0.093699,
		0.11513,
		0.137255,
		0.159817,
		0.182603,
		0.205443,
		0.228195,
		0.250747,
		0.273008,
		0.294906,
		0.316383,
		0.337397,
		0.357915,
		0.377913,
		0.397375,
		0.416289,
		0.434652,
		0.452462,
		0.469721,
		0.486434,
		0.502605,
		0.518237,
		0.533332,
		0.547885,
		0.56188,
		0.575293,
		0.588084,
		0.600193,
		0.611541,
		0.622029,
		0.631533,
		0.639906,
		0.646984,
		0.652581,
		0.656502,
		0.658546,
		0.658512,
		0.656209,
		0.651464,
		0.644135,
		0.634115,
		0.621342,
		0.605806,
		0.587553,
		0.566689,
		0.543376,
		0.517835,
		0.490336,
		0.461192,
		0.430753,
		0.399391,
		0.367494,
		0.33545,
		0.303641,
		0.272428,
		0.242148,
		0.213104,
		0.185558,
		0.159732,
		0.135801,
		0.113897,
		0.094104,
		0.076464,
		0.060974,
		0.047595,
		0.036251,
		0.02683,
		0.019196,
		0.013183,
		0.008609,
		0.005276,
		0.002975,
		0.001498,
		0.00064,
		0.000211,
		4.4e-005,
		4.2e-005,
		0.000204,
		0.000617,
		0.001442,
		0.00286,
		0.005066,
		0.008259,
		0.012633,
		0.018373,
		0.025651,
		0.034615,
		0.045391,
		0.058074,
		0.072729,
		0.089384,
		0.108027,
		0.128609,
		0.151036,
		0.175172,
		0.200837,
		0.22781,
		0.25583,
		0.284601,
		0.313795,
		0.343062,
		0.372039,
		0.400354,
		0.427644,
		0.453558,
		0.477773,
		0.5,
		0.519995,
		0.537563,
		0.552562,
		0.564909,
		0.574574,
		0.581581,
		0.586001,
		0.587947,
		0.587564,
		0.585024,
		0.580512,
		0.574223,
		0.56635,
		0.557079,
		0.546584,
		0.535021,
		0.522525,
		0.509214,
		0.495182,
		0.480503,
		0.465237,
		0.449423,
		0.433094,
		0.416269,
		0.398965,
		0.381192,
		0.362964,
		0.344293,
		0.325197,
		0.3057,
		0.28583,
		0.265625,
		0.245133,
		0.224414,
		0.203541,
		0.182603,
		0.16171,
		0.140991,
		0.120607,
		0.100747,
		0.08164,
		0.063557,
		0.046826,
		0.031834,
		0.019046,
		0.009016,
		0.002404,
		0.002404,
		0.009016,
		0.019046,
		0.031834,
		0.046826,
		0.063557,
		0.08164,
		0.100747,
		0.120607,
		0.140991,
		0.16171,
		0.182603,
		0.203541,
		0.224414,
		0.245133,
		0.265625,
		0.28583,
		0.3057,
		0.325197,
		0.344293,
		0.362964,
		0.381192,
		0.398965,
		0.416269,
		0.433094,
		0.449423,
		0.465237,
		0.480503,
		0.495182,
		0.509214,
		0.522525,
		0.535021,
		0.546584,
		0.557079,
		0.56635,
		0.574223,
		0.580512,
		0.585024,
		0.587564,
		0.587947,
		0.586001,
		0.581581,
		0.574574,
		0.564909,
		0.552562,
		0.537563,
		0.519995,
		0.5,
		0.477773,
		0.453558,
		0.427644,
		0.400354,
		0.372039,
		0.343062,
		0.313795,
		0.284601,
		0.25583,
		0.22781,
		0.200837,
		0.175172,
		0.151036,
		0.128609,
		0.108027,
		0.089384,
		0.072729,
		0.058074,
		0.045391,
		0.034615,
		0.025651,
		0.018373,
		0.012633,
		0.008259,
		0.005066,
		0.00286,
		0.001442,
		0.000617,
		0.000204,
		4.2e-005,
		4e-005,
		0.000195,
		0.00059,
		0.001377,
		0.002729,
		0.004828,
		0.00786,
		0.012007,
		0.01744,
		0.024314,
		0.032763,
		0.042898,
		0.0548,
		0.068519,
		0.084068,
		0.101426,
		0.120532,
		0.141286,
		0.163544,
		0.187125,
		0.211808,
		0.237336,
		0.26342,
		0.289746,
		0.315979,
		0.341773,
		0.366781,
		0.390664,
		0.4131,
		0.433793,
		0.452485,
		0.468959,
		0.48305,
		0.494641,
		0.50367,
		0.510129,
		0.514056,
		0.515532,
		0.514679,
		0.511645,
		0.5066,
		0.499726,
		0.491211,
		0.48124,
		0.469991,
		0.457628,
		0.444298,
		0.430131,
		0.415238,
		0.39971,
		0.383622,
		0.367036,
		0.349999,
		0.332553,
		0.314732,
		0.296568,
		0.278094,
		0.259348,
		0.24037,
		0.221211,
		0.201932,
		0.182603,
		0.163312,
		0.144161,
		0.125272,
		0.106789,
		0.088884,
		0.071756,
		0.055645,
		0.04083,
		0.027641,
		0.016464,
		0.007758,
		0.002059,
		0.002059,
		0.007758,
		0.016464,
		0.027641,
		0.04083,
		0.055645,
		0.071756,
		0.088884,
		0.106789,
		0.125272,
		0.144161,
		0.163312,
		0.182603,
		0.201932,
		0.221211,
		0.24037,
		0.259348,
		0.278094,
		0.296568,
		0.314732,
		0.332553,
		0.349999,
		0.367036,
		0.383622,
		0.39971,
		0.415238,
		0.430131,
		0.444298,
		0.457628,
		0.469991,
		0.48124,
		0.491211,
		0.499726,
		0.5066,
		0.511645,
		0.514679,
		0.515532,
		0.514056,
		0.510129,
		0.50367,
		0.494641,
		0.48305,
		0.468959,
		0.452485,
		0.433793,
		0.4131,
		0.390664,
		0.366781,
		0.341773,
		0.315979,
		0.289746,
		0.26342,
		0.237336,
		0.211808,
		0.187125,
		0.163544,
		0.141286,
		0.120532,
		0.101426,
		0.084068,
		0.068519,
		0.0548,
		0.042898,
		0.032763,
		0.024314,
		0.01744,
		0.012007,
		0.00786,
		0.004828,
		0.002729,
		0.001377,
		0.00059,
		0.000195,
		4e-005,
		3.8e-005,
		0.000185,
		0.00056,
		0.001305,
		0.002582,
		0.004562,
		0.007416,
		0.011311,
		0.016402,
		0.02283,
		0.030711,
		0.040139,
		0.051181,
		0.06387,
		0.078208,
		0.094161,
		0.111657,
		0.130588,
		0.150806,
		0.172129,
		0.194338,
		0.217181,
		0.240381,
		0.263636,
		0.286631,
		0.309044,
		0.330552,
		0.350845,
		0.369632,
		0.386651,
		0.401677,
		0.414524,
		0.425058,
		0.433192,
		0.438887,
		0.442157,
		0.443057,
		0.441681,
		0.438158,
		0.43264,
		0.425297,
		0.416309,
		0.405858,
		0.394122,
		0.381271,
		0.367463,
		0.352837,
		0.337521,
		0.321622,
		0.305234,
		0.288436,
		0.2713,
		0.253886,
		0.236251,
		0.218452,
		0.200548,
		0.182602,
		0.164687,
		0.146884,
		0.129291,
		0.112019,
		0.0952,
		0.078987,
		0.063557,
		0.049119,
		0.035914,
		0.024223,
		0.014374,
		0.006746,
		0.001783,
		0.001783,
		0.006746,
		0.014374,
		0.024223,
		0.035914,
		0.049119,
		0.063557,
		0.078987,
		0.0952,
		0.112019,
		0.129291,
		0.146884,
		0.164687,
		0.182602,
		0.200548,
		0.218452,
		0.236251,
		0.253886,
		0.2713,
		0.288436,
		0.305234,
		0.321622,
		0.337521,
		0.352837,
		0.367463,
		0.381271,
		0.394122,
		0.405858,
		0.416309,
		0.425297,
		0.43264,
		0.438158,
		0.441681,
		0.443057,
		0.442157,
		0.438887,
		0.433192,
		0.425058,
		0.414524,
		0.401677,
		0.386651,
		0.369632,
		0.350845,
		0.330552,
		0.309044,
		0.286631,
		0.263636,
		0.240381,
		0.217181,
		0.194338,
		0.172129,
		0.150806,
		0.130588,
		0.111657,
		0.094161,
		0.078208,
		0.06387,
		0.051181,
		0.040139,
		0.030711,
		0.02283,
		0.016402,
		0.011311,
		0.007416,
		0.004562,
		0.002582,
		0.001305,
		0.00056,
		0.000185,
		3.8e-005,
		3.6e-005,
		0.000175,
		0.000526,
		0.001225,
		0.002421,
		0.00427,
		0.00693,
		0.01055,
		0.015271,
		0.021213,
		0.028478,
		0.037142,
		0.047255,
		0.058836,
		0.071873,
		0.086318,
		0.102092,
		0.119079,
		0.137128,
		0.156055,
		0.175647,
		0.195661,
		0.21583,
		0.235873,
		0.255494,
		0.274398,
		0.292291,
		0.308897,
		0.323959,
		0.337251,
		0.348584,
		0.357808,
		0.364821,
		0.369568,
		0.372038,
		0.372267,
		0.370327,
		0.366327,
		0.360406,
		0.352718,
		0.343437,
		0.33274,
		0.320804,
		0.307804,
		0.293903,
		0.279254,
		0.263995,
		0.248253,
		0.232137,
		0.215749,
		0.199181,
		0.182516,
		0.165839,
		0.149231,
		0.132781,
		0.116585,
		0.100747,
		0.085389,
		0.070648,
		0.056682,
		0.043674,
		0.031834,
		0.021402,
		0.012657,
		0.00592,
		0.001559,
		0.001559,
		0.00592,
		0.012657,
		0.021402,
		0.031834,
		0.043674,
		0.056682,
		0.070648,
		0.085389,
		0.100747,
		0.116585,
		0.132781,
		0.149231,
		0.165839,
		0.182516,
		0.199181,
		0.215749,
		0.232137,
		0.248253,
		0.263995,
		0.279254,
		0.293903,
		0.307804,
		0.320804,
		0.33274,
		0.343437,
		0.352718,
		0.360406,
		0.366327,
		0.370327,
		0.372267,
		0.372038,
		0.369568,
		0.364821,
		0.357808,
		0.348584,
		0.337251,
		0.323959,
		0.308897,
		0.292291,
		0.274398,
		0.255494,
		0.235873,
		0.21583,
		0.195661,
		0.175647,
		0.156055,
		0.137128,
		0.119079,
		0.102092,
		0.086318,
		0.071873,
		0.058836,
		0.047255,
		0.037142,
		0.028478,
		0.021213,
		0.015271,
		0.01055,
		0.00693,
		0.00427,
		0.002421,
		0.001225,
		0.000526,
		0.000175,
		3.6e-005,
		3.4e-005,
		0.000163,
		0.000491,
		0.00114,
		0.002247,
		0.003956,
		0.006407,
		0.009733,
		0.014057,
		0.019481,
		0.02609,
		0.033943,
		0.043071,
		0.05348,
		0.065143,
		0.078003,
		0.09197,
		0.106923,
		0.122709,
		0.139147,
		0.156029,
		0.173123,
		0.190179,
		0.206935,
		0.223122,
		0.238473,
		0.252728,
		0.265647,
		0.27701,
		0.286632,
		0.294361,
		0.300088,
		0.303745,
		0.30531,
		0.304801,
		0.302277,
		0.297832,
		0.291589,
		0.283695,
		0.274314,
		0.263617,
		0.251784,
		0.238991,
		0.225407,
		0.211195,
		0.196507,
		0.181483,
		0.166253,
		0.150937,
		0.13565,
		0.120501,
		0.105601,
		0.091062,
		0.077003,
		0.063554,
		0.050861,
		0.039085,
		0.02841,
		0.019046,
		0.011231,
		0.005236,
		0.001374,
		0.001374,
		0.005236,
		0.011231,
		0.019046,
		0.02841,
		0.039085,
		0.050861,
		0.063554,
		0.077003,
		0.091062,
		0.105601,
		0.120501,
		0.13565,
		0.150937,
		0.166253,
		0.181483,
		0.196507,
		0.211195,
		0.225407,
		0.238991,
		0.251784,
		0.263617,
		0.274314,
		0.283695,
		0.291589,
		0.297832,
		0.302277,
		0.304801,
		0.30531,
		0.303745,
		0.300088,
		0.294361,
		0.286632,
		0.27701,
		0.265647,
		0.252728,
		0.238473,
		0.223122,
		0.206935,
		0.190179,
		0.173123,
		0.156029,
		0.139147,
		0.122709,
		0.106923,
		0.09197,
		0.078003,
		0.065143,
		0.05348,
		0.043071,
		0.033943,
		0.02609,
		0.019481,
		0.014057,
		0.009733,
		0.006407,
		0.003956,
		0.002247,
		0.00114,
		0.000491,
		0.000163,
		3.4e-005,
		3.1e-005,
		0.000151,
		0.000452,
		0.001049,
		0.002063,
		0.003623,
		0.005853,
		0.00887,
		0.012776,
		0.017657,
		0.023578,
		0.030582,
		0.038685,
		0.047876,
		0.058116,
		0.069338,
		0.081444,
		0.094309,
		0.107782,
		0.121685,
		0.135819,
		0.149966,
		0.163895,
		0.177368,
		0.190145,
		0.20199,
		0.212684,
		0.222021,
		0.229828,
		0.235957,
		0.240301,
		0.242791,
		0.243396,
		0.242128,
		0.239036,
		0.234205,
		0.227748,
		0.219804,
		0.210532,
		0.200101,
		0.188689,
		0.176476,
		0.163637,
		0.150344,
		0.136762,
		0.123044,
		0.109339,
		0.095789,
		0.082529,
		0.069697,
		0.057429,
		0.04587,
		0.035173,
		0.025507,
		0.017057,
		0.010032,
		0.004665,
		0.001221,
		0.001221,
		0.004665,
		0.010032,
		0.017057,
		0.025507,
		0.035173,
		0.04587,
		0.057429,
		0.069697,
		0.082529,
		0.095789,
		0.109339,
		0.123044,
		0.136762,
		0.150344,
		0.163637,
		0.176476,
		0.188689,
		0.200101,
		0.210532,
		0.219804,
		0.227748,
		0.234205,
		0.239036,
		0.242128,
		0.243396,
		0.242791,
		0.240301,
		0.235957,
		0.229828,
		0.222021,
		0.212684,
		0.20199,
		0.190145,
		0.177368,
		0.163895,
		0.149966,
		0.135819,
		0.121685,
		0.107782,
		0.094309,
		0.081444,
		0.069338,
		0.058116,
		0.047876,
		0.038685,
		0.030582,
		0.023578,
		0.017657,
		0.012776,
		0.00887,
		0.005853,
		0.003623,
		0.002063,
		0.001049,
		0.000452,
		0.000151,
		3.1e-005,
		2.9e-005,
		0.000138,
		0.000412,
		0.000953,
		0.00187,
		0.003275,
		0.005276,
		0.00797,
		0.011443,
		0.015763,
		0.020976,
		0.027109,
		0.034161,
		0.042108,
		0.0509,
		0.06046,
		0.070687,
		0.081453,
		0.092608,
		0.103984,
		0.115392,
		0.126633,
		0.137498,
		0.147776,
		0.15726,
		0.16575,
		0.173066,
		0.179047,
		0.183562,
		0.186509,
		0.187825,
		0.187482,
		0.18549,
		0.181895,
		0.176778,
		0.170248,
		0.162442,
		0.153515,
		0.143636,
		0.132984,
		0.121743,
		0.110096,
		0.098223,
		0.0863,
		0.074497,
		0.06298,
		0.05191,
		0.041447,
		0.031753,
		0.022996,
		0.015352,
		0.009011,
		0.004181,
		0.001092,
		0.001092,
		0.004181,
		0.009011,
		0.015352,
		0.022996,
		0.031753,
		0.041447,
		0.05191,
		0.06298,
		0.074497,
		0.0863,
		0.098223,
		0.110096,
		0.121743,
		0.132984,
		0.143636,
		0.153515,
		0.162442,
		0.170248,
		0.176778,
		0.181895,
		0.18549,
		0.187482,
		0.187825,
		0.186509,
		0.183562,
		0.179047,
		0.173066,
		0.16575,
		0.15726,
		0.147776,
		0.137498,
		0.126633,
		0.115392,
		0.103984,
		0.092608,
		0.081453,
		0.070687,
		0.06046,
		0.0509,
		0.042108,
		0.034161,
		0.027109,
		0.020976,
		0.015763,
		0.011443,
		0.00797,
		0.005276,
		0.003275,
		0.00187,
		0.000953,
		0.000412,
		0.000138,
		2.9e-005,
		2.6e-005,
		0.000124,
		0.000371,
		0.000854,
		0.001671,
		0.002916,
		0.004682,
		0.007047,
		0.010079,
		0.013828,
		0.018324,
		0.023575,
		0.029569,
		0.036269,
		0.043615,
		0.051523,
		0.05989,
		0.068588,
		0.077474,
		0.08639,
		0.095162,
		0.103613,
		0.11156,
		0.118823,
		0.12523,
		0.130624,
		0.134865,
		0.13784,
		0.139462,
		0.13968,
		0.138472,
		0.135857,
		0.131883,
		0.126633,
		0.120219,
		0.112777,
		0.104466,
		0.095456,
		0.085931,
		0.076079,
		0.066092,
		0.056161,
		0.046472,
		0.037212,
		0.028563,
		0.020707,
		0.013828,
		0.008113,
		0.003761,
		0.000981,
		0.000981,
		0.003761,
		0.008113,
		0.013828,
		0.020707,
		0.028563,
		0.037212,
		0.046472,
		0.056161,
		0.066092,
		0.076079,
		0.085931,
		0.095456,
		0.104466,
		0.112777,
		0.120219,
		0.126633,
		0.131883,
		0.135857,
		0.138472,
		0.13968,
		0.139462,
		0.13784,
		0.134865,
		0.130624,
		0.12523,
		0.118823,
		0.11156,
		0.103613,
		0.095162,
		0.08639,
		0.077474,
		0.068588,
		0.05989,
		0.051523,
		0.043615,
		0.036269,
		0.029569,
		0.023575,
		0.018324,
		0.013828,
		0.010079,
		0.007047,
		0.004682,
		0.002916,
		0.001671,
		0.000854,
		0.000371,
		0.000124,
		2.6e-005,
		2.3e-005,
		0.00011,
		0.000328,
		0.000754,
		0.001468,
		0.002552,
		0.00408,
		0.006113,
		0.008703,
		0.011881,
		0.015662,
		0.02004,
		0.02499,
		0.030463,
		0.036394,
		0.042695,
		0.049261,
		0.055972,
		0.062693,
		0.069278,
		0.075577,
		0.081434,
		0.086698,
		0.091224,
		0.09488,
		0.097552,
		0.099147,
		0.0996,
		0.098873,
		0.096962,
		0.093895,
		0.089732,
		0.084562,
		0.078506,
		0.071706,
		0.064324,
		0.05654,
		0.048543,
		0.040531,
		0.032703,
		0.025261,
		0.018407,
		0.01234,
		0.007261,
		0.003372,
		0.00088,
		0.00088,
		0.003372,
		0.007261,
		0.01234,
		0.018407,
		0.025261,
		0.032703,
		0.040531,
		0.048543,
		0.05654,
		0.064324,
		0.071706,
		0.078506,
		0.084562,
		0.089732,
		0.093895,
		0.096962,
		0.098873,
		0.0996,
		0.099147,
		0.097552,
		0.09488,
		0.091224,
		0.086698,
		0.081434,
		0.075577,
		0.069278,
		0.062693,
		0.055972,
		0.049261,
		0.042695,
		0.036394,
		0.030463,
		0.02499,
		0.02004,
		0.015662,
		0.011881,
		0.008703,
		0.006113,
		0.00408,
		0.002552,
		0.001468,
		0.000754,
		0.000328,
		0.00011,
		2.3e-005,
		2e-005,
		9.6e-005,
		0.000285,
		0.000652,
		0.001265,
		0.002188,
		0.003479,
		0.005185,
		0.007338,
		0.009956,
		0.013038,
		0.016567,
		0.020506,
		0.0248,
		0.029379,
		0.034155,
		0.039027,
		0.043882,
		0.0486,
		0.053053,
		0.057115,
		0.060661,
		0.063573,
		0.065746,
		0.067091,
		0.067541,
		0.067052,
		0.065609,
		0.063227,
		0.059949,
		0.055852,
		0.051039,
		0.045641,
		0.039813,
		0.033727,
		0.027572,
		0.021549,
		0.015865,
		0.010731,
		0.006363,
		0.002974,
		0.00078,
		0.00078,
		0.002974,
		0.006363,
		0.010731,
		0.015865,
		0.021549,
		0.027572,
		0.033727,
		0.039813,
		0.045641,
		0.051039,
		0.055852,
		0.059949,
		0.063227,
		0.065609,
		0.067052,
		0.067541,
		0.067091,
		0.065746,
		0.063573,
		0.060661,
		0.057115,
		0.053053,
		0.0486,
		0.043882,
		0.039027,
		0.034155,
		0.029379,
		0.0248,
		0.020506,
		0.016567,
		0.013038,
		0.009956,
		0.007338,
		0.005185,
		0.003479,
		0.002188,
		0.001265,
		0.000652,
		0.000285,
		9.6e-005,
		2e-005,
		1.7e-005,
		8.2e-005,
		0.000243,
		0.000552,
		0.001064,
		0.00183,
		0.00289,
		0.004277,
		0.006008,
		0.008087,
		0.010502,
		0.013224,
		0.016209,
		0.019399,
		0.022723,
		0.026096,
		0.029425,
		0.032612,
		0.035553,
		0.038146,
		0.040291,
		0.041898,
		0.04289,
		0.043203,
		0.042798,
		0.041656,
		0.039786,
		0.037225,
		0.03404,
		0.030325,
		0.026203,
		0.021821,
		0.01735,
		0.012978,
		0.008906,
		0.00535,
		0.00253,
		0.000671,
		0.000671,
		0.00253,
		0.00535,
		0.008906,
		0.012978,
		0.01735,
		0.021821,
		0.026203,
		0.030325,
		0.03404,
		0.037225,
		0.039786,
		0.041656,
		0.042798,
		0.043203,
		0.04289,
		0.041898,
		0.040291,
		0.038146,
		0.035553,
		0.032612,
		0.029425,
		0.026096,
		0.022723,
		0.019399,
		0.016209,
		0.013224,
		0.010502,
		0.008087,
		0.006008,
		0.004277,
		0.00289,
		0.00183,
		0.001064,
		0.000552,
		0.000243,
		8.2e-005,
		1.7e-005,
		1.5e-005,
		6.9e-005,
		0.000201,
		0.000454,
		0.00087,
		0.001483,
		0.002323,
		0.003407,
		0.004739,
		0.006313,
		0.008105,
		0.010082,
		0.012195,
		0.014386,
		0.016587,
		0.018722,
		0.020712,
		0.022475,
		0.023933,
		0.025011,
		0.025644,
		0.025783,
		0.025391,
		0.024457,
		0.022989,
		0.021023,
		0.018622,
		0.015878,
		0.012909,
		0.009861,
		0.006903,
		0.004224,
		0.002032,
		0.000547,
		0.000547,
		0.002032,
		0.004224,
		0.006903,
		0.009861,
		0.012909,
		0.015878,
		0.018622,
		0.021023,
		0.022989,
		0.024457,
		0.025391,
		0.025783,
		0.025644,
		0.025011,
		0.023933,
		0.022475,
		0.020712,
		0.018722,
		0.016587,
		0.014386,
		0.012195,
		0.010082,
		0.008105,
		0.006313,
		0.004739,
		0.003407,
		0.002323,
		0.001483,
		0.00087,
		0.000454,
		0.000201,
		6.9e-005,
		1.5e-005,
		1.2e-005,
		5.6e-005,
		0.000161,
		0.000361,
		0.000685,
		0.001156,
		0.00179,
		0.002592,
		0.003559,
		0.004672,
		0.005905,
		0.007219,
		0.008567,
		0.009895,
		0.011143,
		0.01225,
		0.013156,
		0.013802,
		0.014139,
		0.014128,
		0.013746,
		0.012984,
		0.011859,
		0.010409,
		0.008699,
		0.006822,
		0.004895,
		0.003067,
		0.001509,
		0.000415,
		0.000415,
		0.001509,
		0.003067,
		0.004895,
		0.006822,
		0.008699,
		0.010409,
		0.011859,
		0.012984,
		0.013746,
		0.014128,
		0.014139,
		0.013802,
		0.013156,
		0.01225,
		0.011143,
		0.009895,
		0.008567,
		0.007219,
		0.005905,
		0.004672,
		0.003559,
		0.002592,
		0.00179,
		0.001156,
		0.000685,
		0.000361,
		0.000161,
		5.6e-005,
		1.2e-005,
		4.3e-005,
		0.000124,
		0.000274,
		0.000513,
		0.000854,
		0.001301,
		0.001853,
		0.002495,
		0.003208,
		0.00396,
		0.004715,
		0.005431,
		0.006065,
		0.00657,
		0.006907,
		0.007038,
		0.006937,
		0.006591,
		0.006001,
		0.005191,
		0.004205,
		0.003111,
		0.002007,
		0.001015,
		0.000286,
		0.000286,
		0.001015,
		0.002007,
		0.003111,
		0.004205,
		0.005191,
		0.006001,
		0.006591,
		0.006937,
		0.007038,
		0.006907,
		0.00657,
		0.006065,
		0.005431,
		0.004715,
		0.00396,
		0.003208,
		0.002495,
		0.001853,
		0.001301,
		0.000854,
		0.000513,
		0.000274,
		0.000124,
		4.3e-005,
		3.2e-005,
		9e-005,
		0.000196,
		0.000359,
		0.000585,
		0.000871,
		0.001208,
		0.001579,
		0.001962,
		0.002331,
		0.002653,
		0.0029,
		0.003043,
		0.003057,
		0.002929,
		0.002653,
		0.002242,
		0.001726,
		0.001155,
		0.000604,
		0.000176,
		0.000176,
		0.000604,
		0.001155,
		0.001726,
		0.002242,
		0.002653,
		0.002929,
		0.003057,
		0.003043,
		0.0029,
		0.002653,
		0.002331,
		0.001962,
		0.001579,
		0.001208,
		0.000871,
		0.000585,
		0.000359,
		0.000196,
		9e-005,
		3.2e-005,
		2.2e-005,
		6e-005,
		0.000127,
		0.000227,
		0.000357,
		0.000512,
		0.000679,
		0.000842,
		0.000983,
		0.001083,
		0.001123,
		0.001091,
		0.000982,
		0.000799,
		0.000562,
		0.000308,
		9.4e-005,
		9.4e-005,
		0.000308,
		0.000562,
		0.000799,
		0.000982,
		0.001091,
		0.001123,
		0.001083,
		0.000983,
		0.000842,
		0.000679,
		0.000512,
		0.000357,
		0.000227,
		0.000127,
		6e-005,
		2.2e-005,
		1.3e-005,
		3.5e-005,
		7.2e-005,
		0.000121,
		0.000179,
		0.000237,
		0.000287,
		0.000316,
		0.000317,
		0.000283,
		0.000215,
		0.000127,
		4.1e-005,
		4.1e-005,
		0.000127,
		0.000215,
		0.000283,
		0.000317,
		0.000316,
		0.000287,
		0.000237,
		0.000179,
		0.000121,
		7.2e-005,
		3.5e-005,
		1.3e-005,
		1.6e-005,
		3e-005,
		4.5e-005,
		5.7e-005,
		6.2e-005,
		5.5e-005,
		3.7e-005,
		1.3e-005,
		1.3e-005,
		3.7e-005,
		5.5e-005,
		6.2e-005,
		5.7e-005,
		4.5e-005,
		3e-005,
		1.6e-005,
		2.1e-005,
		2.2e-005,
		1.6e-005,
		2.5e-005,
		6.2e-005,
		8.2e-005,
		7.9e-005,
		6e-005,
		3.5e-005,
		1.5e-005,
		5.2e-005,
		0.000144,
		0.000213,
		0.000237,
		0.000218,
		0.00017,
		0.000112,
		6e-005,
		2.4e-005,
		9.7e-005,
		0.000283,
		0.000451,
		0.000551,
		0.000567,
		0.000512,
		0.00041,
		0.00029,
		0.000178,
		9e-005,
		3.5e-005,
		0.000162,
		0.000496,
		0.000834,
		0.001083,
		0.0012,
		0.001186,
		0.001064,
		0.000871,
		0.000649,
		0.000434,
		0.000254,
		0.000124,
		4.6e-005,
		1.1e-005,
		0.000253,
		0.000799,
		0.001398,
		0.001898,
		0.002217,
		0.002331,
		0.002251,
		0.002017,
		0.001682,
		0.001301,
		0.000926,
		0.000597,
		0.000339,
		0.000161,
		5.9e-005,
		1.3e-005,
		0.000371,
		0.001206,
		0.002176,
		0.003057,
		0.003714,
		0.004083,
		0.004154,
		0.00396,
		0.003555,
		0.003011,
		0.0024,
		0.00179,
		0.001235,
		0.000776,
		0.000431,
		0.000201,
		7.2e-005,
		1.6e-005,
		0.000519,
		0.001726,
		0.003192,
		0.00461,
		0.005775,
		0.00657,
		0.006952,
		0.00693,
		0.006556,
		0.005905,
		0.005068,
		0.004136,
		0.003197,
		0.002323,
		0.001568,
		0.000966,
		0.000527,
		0.000243,
		8.6e-005,
		1.9e-005,
		0.000697,
		0.002362,
		0.004461,
		0.006591,
		0.008464,
		0.009899,
		0.0108,
		0.011143,
		0.010961,
		0.010326,
		0.009337,
		0.008105,
		0.006746,
		0.005365,
		0.004055,
		0.00289,
		0.001918,
		0.001164,
		0.000627,
		0.000285,
		0.0001,
		2.2e-005,
		0.000902,
		0.003111,
		0.005983,
		0.009012,
		0.011818,
		0.014139,
		0.015814,
		0.016772,
		0.017011,
		0.016587,
		0.015596,
		0.01416,
		0.012416,
		0.010502,
		0.008547,
		0.006667,
		0.004955,
		0.003479,
		0.002279,
		0.001366,
		0.000728,
		0.000328,
		0.000114,
		2.4e-005,
		0.001132,
		0.003962,
		0.007742,
		0.011859,
		0.015835,
		0.019315,
		0.02206,
		0.023933,
		0.024886,
		0.024941,
		0.02418,
		0.022723,
		0.020718,
		0.018328,
		0.015716,
		0.013038,
		0.010433,
		0.008018,
		0.00588,
		0.00408,
		0.002644,
		0.00157,
		0.000829,
		0.000371,
		0.000127,
		2.7e-005,
		0.00138,
		0.004895,
		0.009702,
		0.015087,
		0.020468,
		0.025391,
		0.029527,
		0.03266,
		0.034677,
		0.035553,
		0.035336,
		0.034131,
		0.032087,
		0.029379,
		0.026196,
		0.02273,
		0.019164,
		0.015662,
		0.012367,
		0.009391,
		0.006814,
		0.004682,
		0.003007,
		0.001771,
		0.000929,
		0.000412,
		0.000141,
		3e-005,
		0.001639,
		0.005885,
		0.011815,
		0.018622,
		0.025625,
		0.032269,
		0.038126,
		0.04289,
		0.046366,
		0.048465,
		0.049186,
		0.0486,
		0.046839,
		0.04408,
		0.040526,
		0.036394,
		0.031906,
		0.027273,
		0.022689,
		0.018324,
		0.014314,
		0.010764,
		0.007741,
		0.005276,
		0.003363,
		0.001968,
		0.001025,
		0.000452,
		0.000154,
		3.3e-005,
		0.001902,
		0.006903,
		0.014018,
		0.022365,
		0.031171,
		0.039786,
		0.04768,
		0.054447,
		0.059802,
		0.063573,
		0.06569,
		0.066174,
		0.065122,
		0.062693,
		0.059088,
		0.054541,
		0.049299,
		0.043615,
		0.03773,
		0.031869,
		0.026229,
		0.020976,
		0.016241,
		0.012115,
		0.008648,
		0.005853,
		0.003708,
		0.002157,
		0.001117,
		0.000491,
		0.000166,
		3.5e-005,
		0.002161,
		0.007919,
		0.016247,
		0.026203,
		0.036943,
		0.047728,
		0.057933,
		0.067052,
		0.074699,
		0.080608,
		0.084624,
		0.086698,
		0.086869,
		0.08526,
		0.082051,
		0.077474,
		0.071791,
		0.06528,
		0.058225,
		0.0509,
		0.04356,
		0.036436,
		0.029723,
		0.023578,
		0.01812,
		0.013424,
		0.009521,
		0.006407,
		0.004037,
		0.002336,
		0.001205,
		0.000526,
		0.000177,
		3.7e-005,
		0.00241,
		0.008906,
		0.018437,
		0.030023,
		0.042764,
		0.055852,
		0.068577,
		0.080339,
		0.090652,
		0.099147,
		0.105574,
		0.109792,
		0.111767,
		0.11156,
		0.109311,
		0.105226,
		0.099561,
		0.092608,
		0.084677,
		0.076083,
		0.067132,
		0.058116,
		0.049296,
		0.0409,
		0.033116,
		0.02609,
		0.019924,
		0.014673,
		0.010351,
		0.00693,
		0.004345,
		0.002503,
		0.001285,
		0.00056,
		0.000188,
		3.9e-005,
		0.002646,
		0.009847,
		0.020541,
		0.033727,
		0.048471,
		0.063915,
		0.079284,
		0.093895,
		0.107171,
		0.118641,
		0.127953,
		0.134865,
		0.139251,
		0.141088,
		0.140451,
		0.137498,
		0.132457,
		0.125608,
		0.117269,
		0.107782,
		0.097494,
		0.086747,
		0.075864,
		0.065143,
		0.054846,
		0.045192,
		0.03636,
		0.028478,
		0.021629,
		0.015848,
		0.011126,
		0.007416,
		0.004631,
		0.002657,
		0.00136,
		0.00059,
		0.000197,
		4.1e-005,
		0.002868,
		0.010731,
		0.022525,
		0.037241,
		0.053931,
		0.071706,
		0.089745,
		0.107308,
		0.123736,
		0.138472,
		0.151061,
		0.161159,
		0.168534,
		0.173066,
		0.174741,
		0.173644,
		0.169947,
		0.163895,
		0.155789,
		0.145975,
		0.134822,
		0.122709,
		0.110012,
		0.097092,
		0.084279,
		0.071873,
		0.060128,
		0.049254,
		0.039411,
		0.030711,
		0.023214,
		0.016933,
		0.011839,
		0.00786,
		0.00489,
		0.002796,
		0.001426,
		0.000617,
		0.000206,
		4.3e-005,
		0.003077,
		0.011559,
		0.02438,
		0.040531,
		0.05906,
		0.07907,
		0.099716,
		0.120219,
		0.139866,
		0.158028,
		0.174161,
		0.187825,
		0.198682,
		0.206504,
		0.211175,
		0.212684,
		0.211121,
		0.206667,
		0.19958,
		0.190179,
		0.178832,
		0.165934,
		0.151896,
		0.137128,
		0.122023,
		0.106953,
		0.092251,
		0.078208,
		0.065071,
		0.053032,
		0.042233,
		0.032763,
		0.024662,
		0.01792,
		0.012483,
		0.008259,
		0.005121,
		0.00292,
		0.001485,
		0.00064,
		0.000213,
		4.4e-005,
		0.003275,
		0.01234,
		0.026116,
		0.043596,
		0.063833,
		0.085931,
		0.109042,
		0.132372,
		0.155176,
		0.176778,
		0.196569,
		0.214025,
		0.228713,
		0.240301,
		0.248563,
		0.253379,
		0.254737,
		0.252728,
		0.247536,
		0.239425,
		0.228728,
		0.21583,
		0.201151,
		0.185128,
		0.168203,
		0.150806,
		0.133344,
		0.116188,
		0.099669,
		0.084068,
		0.069614,
		0.056483,
		0.044794,
		0.034615,
		0.025961,
		0.018799,
		0.013053,
		0.008609,
		0.005324,
		0.003027,
		0.001535,
		0.000661,
		0.000219,
		4.5e-005,
		0.003469,
		0.01309,
		0.027762,
		0.046472,
		0.068277,
		0.09229,
		0.117674,
		0.143636,
		0.169425,
		0.194342,
		0.21774,
		0.239036,
		0.257723,
		0.273376,
		0.285665,
		0.294361,
		0.299339,
		0.300582,
		0.298172,
		0.292291,
		0.283205,
		0.271253,
		0.256833,
		0.240381,
		0.22236,
		0.203242,
		0.183489,
		0.163544,
		0.143817,
		0.124676,
		0.106443,
		0.089384,
		0.073709,
		0.059573,
		0.047072,
		0.036251,
		0.0271,
		0.019564,
		0.013546,
		0.00891,
		0.005496,
		0.003117,
		0.001578,
		0.000677,
		0.000224,
		4.6e-005,
		0.003662,
		0.013828,
		0.029357,
		0.049219,
		0.072466,
		0.098223,
		0.125665,
		0.154017,
		0.182539,
		0.210532,
		0.237338,
		0.262345,
		0.284996,
		0.304801,
		0.321343,
		0.334294,
		0.343419,
		0.348584,
		0.349758,
		0.347016,
		0.340527,
		0.330552,
		0.317429,
		0.30156,
		0.283397,
		0.26342,
		0.242125,
		0.220006,
		0.197539,
		0.175172,
		0.153312,
		0.13232,
		0.112501,
		0.094104,
		0.077319,
		0.062277,
		0.049052,
		0.037661,
		0.028075,
		0.020214,
		0.01396,
		0.009162,
		0.005638,
		0.003191,
		0.001612,
		0.000691,
		0.000228,
		4.7e-005,
		0.003862,
		0.014576,
		0.030947,
		0.05191,
		0.076504,
		0.103854,
		0.133151,
		0.163637,
		0.194598,
		0.225354,
		0.255258,
		0.283695,
		0.310092,
		0.333919,
		0.354703,
		0.372038,
		0.385597,
		0.395137,
		0.400512,
		0.401677,
		0.398686,
		0.391695,
		0.38095,
		0.366781,
		0.34959,
		0.32983,
		0.307995,
		0.284601,
		0.260166,
		0.235198,
		0.21018,
		0.185558,
		0.161731,
		0.139047,
		0.117792,
		0.098195,
		0.080424,
		0.064584,
		0.050725,
		0.038844,
		0.028884,
		0.020748,
		0.014298,
		0.009364,
		0.005752,
		0.00325,
		0.001639,
		0.000701,
		0.000232,
		4.8e-005,
		0.004071,
		0.015352,
		0.032575,
		0.054623,
		0.080506,
		0.109339,
		0.140319,
		0.172707,
		0.205814,
		0.238991,
		0.271617,
		0.3031,
		0.332875,
		0.360406,
		0.385195,
		0.40679,
		0.424797,
		0.438887,
		0.448814,
		0.454416,
		0.455628,
		0.452485,
		0.445116,
		0.43375,
		0.4187,
		0.400354,
		0.379165,
		0.35563,
		0.330276,
		0.303641,
		0.276259,
		0.248646,
		0.221283,
		0.194608,
		0.169008,
		0.144812,
		0.122288,
		0.101641,
		0.083014,
		0.06649,
		0.052095,
		0.039801,
		0.029533,
		0.021172,
		0.014563,
		0.009521,
		0.005838,
		0.003294,
		0.001658,
		0.000709,
		0.000234,
		4.8e-005,
		0.004295,
		0.016175,
		0.034281,
		0.057429,
		0.084585,
		0.114839,
		0.147381,
		0.181483,
		0.216478,
		0.251747,
		0.286708,
		0.320804,
		0.3535,
		0.384281,
		0.412655,
		0.438158,
		0.460362,
		0.478887,
		0.493408,
		0.50367,
		0.5095,
		0.510811,
		0.507609,
		0.5,
		0.488182,
		0.472443,
		0.453154,
		0.430753,
		0.405731,
		0.37862,
		0.349969,
		0.320332,
		0.29025,
		0.260237,
		0.230768,
		0.202269,
		0.17511,
		0.1496,
		0.125984,
		0.104444,
		0.085099,
		0.068007,
		0.053172,
		0.040544,
		0.03003,
		0.021491,
		0.014759,
		0.009635,
		0.005901,
		0.003325,
		0.001672,
		0.000714,
		0.000235,
		4.8e-005,
		0.004537,
		0.017057,
		0.036097,
		0.06039,
		0.088841,
		0.120501,
		0.154541,
		0.190228,
		0.22691,
		0.263995,
		0.300936,
		0.337219,
		0.372351,
		0.405858,
		0.437279,
		0.46617,
		0.492103,
		0.514679,
		0.533535,
		0.548352,
		0.558873,
		0.564909,
		0.566352,
		0.563181,
		0.555467,
		0.543376,
		0.527165,
		0.507174,
		0.48382,
		0.457578,
		0.428973,
		0.398556,
		0.366892,
		0.334543,
		0.302048,
		0.269916,
		0.238608,
		0.208535,
		0.180046,
		0.153429,
		0.128904,
		0.10663,
		0.086702,
		0.069157,
		0.053976,
		0.041091,
		0.030388,
		0.021718,
		0.014896,
		0.009713,
		0.005942,
		0.003345,
		0.001681,
		0.000717,
		0.000236,
		4.9e-005,
		0.004799,
		0.018011,
		0.03805,
		0.063554,
		0.093355,
		0.12645,
		0.161974,
		0.199181,
		0.237419,
		0.276112,
		0.314743,
		0.352837,
		0.38995,
		0.425657,
		0.459545,
		0.491211,
		0.520259,
		0.546305,
		0.568982,
		0.587947,
		0.602893,
		0.613563,
		0.619756,
		0.621342,
		0.618271,
		0.610581,
		0.598399,
		0.581942,
		0.561514,
		0.537498,
		0.510346,
		0.480563,
		0.448693,
		0.415302,
		0.38096,
		0.346229,
		0.311642,
		0.277694,
		0.244832,
		0.213445,
		0.183862,
		0.156346,
		0.131095,
		0.108244,
		0.087866,
		0.069976,
		0.054538,
		0.041464,
		0.030628,
		0.021866,
		0.014983,
		0.009761,
		0.005967,
		0.003356,
		0.001686,
		0.000719,
		0.000237,
		4.9e-005,
		0.005084,
		0.019046,
		0.040162,
		0.066962,
		0.098192,
		0.132781,
		0.169822,
		0.208539,
		0.248267,
		0.288436,
		0.328548,
		0.368158,
		0.406867,
		0.444298,
		0.480096,
		0.513912,
		0.545402,
		0.574223,
		0.600037,
		0.622512,
		0.641333,
		0.656209,
		0.666884,
		0.673152,
		0.674863,
		0.67194,
		0.664385,
		0.652282,
		0.635805,
		0.615215,
		0.590853,
		0.563133,
		0.532534,
		0.499578,
		0.464821,
		0.428835,
		0.392189,
		0.355436,
		0.319097,
		0.283651,
		0.249525,
		0.217088,
		0.186644,
		0.158434,
		0.132632,
		0.109353,
		0.088647,
		0.070513,
		0.054895,
		0.041695,
		0.030771,
		0.021951,
		0.015031,
		0.009787,
		0.005979,
		0.003362,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.005396,
		0.020173,
		0.042454,
		0.070648,
		0.103402,
		0.139572,
		0.178194,
		0.218452,
		0.259662,
		0.301242,
		0.342703,
		0.383622,
		0.423631,
		0.462401,
		0.499627,
		0.535021,
		0.568298,
		0.599175,
		0.627366,
		0.652581,
		0.674529,
		0.692926,
		0.707502,
		0.71801,
		0.724239,
		0.726028,
		0.723272,
		0.715937,
		0.704067,
		0.687787,
		0.667309,
		0.642924,
		0.615002,
		0.58398,
		0.55035,
		0.514646,
		0.477426,
		0.439258,
		0.400704,
		0.362304,
		0.324562,
		0.287937,
		0.252835,
		0.219602,
		0.18852,
		0.159806,
		0.133616,
		0.110041,
		0.089116,
		0.070823,
		0.055094,
		0.041817,
		0.030844,
		0.021992,
		0.015052,
		0.009797,
		0.005984,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.005737,
		0.021402,
		0.044948,
		0.074644,
		0.109032,
		0.146884,
		0.187172,
		0.229034,
		0.271752,
		0.314732,
		0.357477,
		0.399576,
		0.440684,
		0.480503,
		0.518777,
		0.555271,
		0.589761,
		0.622029,
		0.651854,
		0.679005,
		0.703245,
		0.724327,
		0.742003,
		0.756027,
		0.766169,
		0.772221,
		0.774009,
		0.771408,
		0.764353,
		0.752845,
		0.736961,
		0.716858,
		0.692773,
		0.665023,
		0.633993,
		0.600132,
		0.56394,
		0.52595,
		0.486716,
		0.446798,
		0.406741,
		0.367068,
		0.328265,
		0.290769,
		0.254963,
		0.22117,
		0.189652,
		0.160605,
		0.134165,
		0.110408,
		0.089354,
		0.070972,
		0.055183,
		0.041868,
		0.030871,
		0.022005,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.006111,
		0.022747,
		0.047667,
		0.078987,
		0.11513,
		0.154777,
		0.196828,
		0.24037,
		0.28465,
		0.329048,
		0.373058,
		0.416269,
		0.458348,
		0.499021,
		0.538067,
		0.575293,
		0.610534,
		0.643631,
		0.674429,
		0.702768,
		0.728476,
		0.751367,
		0.771243,
		0.787895,
		0.801108,
		0.810671,
		0.816383,
		0.818069,
		0.815588,
		0.808844,
		0.7978,
		0.782486,
		0.763003,
		0.73953,
		0.712321,
		0.681705,
		0.648076,
		0.611886,
		0.57363,
		0.533837,
		0.493048,
		0.451806,
		0.410642,
		0.370055,
		0.330511,
		0.292425,
		0.256157,
		0.222012,
		0.190229,
		0.160989,
		0.134412,
		0.110561,
		0.089445,
		0.071023,
		0.05521,
		0.041881,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.006524,
		0.024223,
		0.050639,
		0.083718,
		0.121749,
		0.163312,
		0.207232,
		0.25254,
		0.298442,
		0.344293,
		0.389572,
		0.433863,
		0.476837,
		0.518237,
		0.557864,
		0.595562,
		0.631206,
		0.664693,
		0.695929,
		0.724819,
		0.751262,
		0.775143,
		0.796331,
		0.814675,
		0.830009,
		0.842154,
		0.850921,
		0.856127,
		0.857596,
		0.855177,
		0.848753,
		0.838252,
		0.823657,
		0.805015,
		0.782445,
		0.756135,
		0.726348,
		0.693416,
		0.657731,
		0.61974,
		0.579929,
		0.53881,
		0.496909,
		0.454751,
		0.412843,
		0.371666,
		0.331662,
		0.293225,
		0.256697,
		0.222364,
		0.19045,
		0.161121,
		0.134487,
		0.1106,
		0.089464,
		0.071031,
		0.055213,
		0.041882,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.006979,
		0.025848,
		0.053897,
		0.088884,
		0.128949,
		0.172561,
		0.218463,
		0.265625,
		0.313213,
		0.360554,
		0.407111,
		0.452462,
		0.496281,
		0.53832,
		0.578397,
		0.616381,
		0.652181,
		0.685735,
		0.717001,
		0.745942,
		0.772521,
		0.796695,
		0.818404,
		0.837568,
		0.854087,
		0.867837,
		0.878677,
		0.886448,
		0.890985,
		0.892122,
		0.889702,
		0.883593,
		0.873692,
		0.859942,
		0.842338,
		0.820937,
		0.795861,
		0.767303,
		0.735522,
		0.700845,
		0.663655,
		0.624385,
		0.583504,
		0.541506,
		0.498898,
		0.456182,
		0.413844,
		0.372344,
		0.332104,
		0.293501,
		0.256861,
		0.222455,
		0.190496,
		0.161143,
		0.134495,
		0.110603,
		0.089465,
		0.071031,
		0.055213,
		0.041881,
		0.030876,
		0.022005,
		0.015057,
		0.009797,
		0.005982,
		0.003362,
		0.001687,
		0.000719,
		0.000236,
		4.9e-005,
		0.007484,
		0.027641,
		0.057479,
		0.094539,
		0.136799,
		0.182603,
		0.230606,
		0.279714,
		0.32905,
		0.377913,
		0.425753,
		0.472143,
		0.516759,
		0.559361,
		0.599782,
		0.637912,
		0.673685,
		0.707073,
		0.738071,
		0.766692,
		0.792954,
		0.816874,
		0.83846,
		0.857704,
		0.874577,
		0.889025,
		0.900971,
		0.91031,
		0.916918,
		0.92065,
		0.921354,
		0.918878,
		0.913075,
		0.903823,
		0.891026,
		0.874633,
		0.85464,
		0.831105,
		0.804146,
		0.773948,
		0.740761,
		0.704898,
		0.666724,
		0.626654,
		0.585137,
		0.542647,
		0.499666,
		0.456678,
		0.414148,
		0.37252,
		0.332198,
		0.293546,
		0.25688,
		0.222461,
		0.190498,
		0.161143,
		0.134494,
		0.1106,
		0.089459,
		0.071023,
		0.055202,
		0.041868,
		0.030862,
		0.021992,
		0.015044,
		0.009787,
		0.005974,
		0.003356,
		0.001684,
		0.000717,
		0.000236,
		4.8e-005,
		0.008046,
		0.029626,
		0.061429,
		0.100747,
		0.145377,
		0.193529,
		0.243759,
		0.294906,
		0.346048,
		0.39646,
		0.445578,
		0.492972,
		0.538323,
		0.581406,
		0.622067,
		0.660215,
		0.695809,
		0.728847,
		0.759357,
		0.787387,
		0.813001,
		0.836264,
		0.857243,
		0.875992,
		0.892551,
		0.906936,
		0.919139,
		0.929125,
		0.936827,
		0.942154,
		0.944988,
		0.945195,
		0.942628,
		0.937137,
		0.928581,
		0.916836,
		0.901806,
		0.883435,
		0.861713,
		0.836688,
		0.808465,
		0.777215,
		0.74317,
		0.706621,
		0.667916,
		0.627445,
		0.585637,
		0.542944,
		0.49983,
		0.45676,
		0.414184,
		0.372533,
		0.332201,
		0.293546,
		0.256878,
		0.222455,
		0.190485,
		0.161121,
		0.134464,
		0.110561,
		0.089413,
		0.070972,
		0.055149,
		0.041817,
		0.030815,
		0.021951,
		0.015011,
		0.009761,
		0.005956,
		0.003345,
		0.001677,
		0.000714,
		0.000235,
		4.8e-005,
		0.008674,
		0.031834,
		0.065798,
		0.107582,
		0.154777,
		0.205443,
		0.258031,
		0.31131,
		0.364313,
		0.416289,
		0.466665,
		0.515011,
		0.561021,
		0.604481,
		0.645261,
		0.683291,
		0.718555,
		0.751076,
		0.780912,
		0.808142,
		0.832862,
		0.855181,
		0.875207,
		0.893048,
		0.908799,
		0.92254,
		0.93433,
		0.9442,
		0.952154,
		0.958161,
		0.962163,
		0.964069,
		0.963764,
		0.961111,
		0.955961,
		0.948161,
		0.937562,
		0.924032,
		0.907465,
		0.887791,
		0.864986,
		0.839077,
		0.810153,
		0.778361,
		0.743911,
		0.707073,
		0.668171,
		0.627576,
		0.585696,
		0.542966,
		0.499836,
		0.45676,
		0.414179,
		0.37252,
		0.332175,
		0.293501,
		0.25681,
		0.222364,
		0.190372,
		0.160989,
		0.134318,
		0.110408,
		0.089259,
		0.070823,
		0.055011,
		0.041695,
		0.03071,
		0.021866,
		0.014945,
		0.009713,
		0.005924,
		0.003325,
		0.001666,
		0.000709,
		0.000233,
		4.8e-005,
		0.009378,
		0.034297,
		0.070648,
		0.11513,
		0.165103,
		0.218463,
		0.273547,
		0.32905,
		0.383959,
		0.437501,
		0.489096,
		0.538323,
		0.584888,
		0.628601,
		0.669354,
		0.707107,
		0.741872,
		0.773702,
		0.802682,
		0.828924,
		0.852554,
		0.873711,
		0.892541,
		0.909189,
		0.923796,
		0.936491,
		0.94739,
		0.956583,
		0.964136,
		0.970086,
		0.974436,
		0.977154,
		0.978178,
		0.977409,
		0.974726,
		0.969981,
		0.963014,
		0.953657,
		0.941747,
		0.927134,
		0.909691,
		0.889329,
		0.865999,
		0.839707,
		0.810517,
		0.778551,
		0.743999,
		0.707107,
		0.668181,
		0.627576,
		0.585688,
		0.542944,
		0.49979,
		0.456678,
		0.414053,
		0.372344,
		0.331948,
		0.293225,
		0.256491,
		0.222012,
		0.189997,
		0.160605,
		0.133936,
		0.110041,
		0.088916,
		0.070513,
		0.054739,
		0.041464,
		0.030522,
		0.021718,
		0.014835,
		0.009635,
		0.005872,
		0.003294,
		0.001649,
		0.000701,
		0.00023,
		4.7e-005,
		0.010171,
		0.037057,
		0.076051,
		0.123492,
		0.176478,
		0.232725,
		0.290446,
		0.34826,
		0.405109,
		0.460201,
		0.512956,
		0.562964,
		0.609954,
		0.653763,
		0.694316,
		0.731605,
		0.765677,
		0.796619,
		0.824552,
		0.849616,
		0.87197,
		0.891783,
		0.909231,
		0.92449,
		0.937734,
		0.94913,
		0.958834,
		0.966984,
		0.973698,
		0.979067,
		0.983155,
		0.985988,
		0.987562,
		0.987832,
		0.986721,
		0.984115,
		0.979875,
		0.973833,
		0.965809,
		0.955611,
		0.943052,
		0.927957,
		0.910174,
		0.889586,
		0.86612,
		0.839755,
		0.810531,
		0.778551,
		0.743987,
		0.707073,
		0.668108,
		0.627445,
		0.585481,
		0.542647,
		0.499394,
		0.456182,
		0.413461,
		0.371666,
		0.331199,
		0.292425,
		0.25566,
		0.22117,
		0.189167,
		0.159806,
		0.133186,
		0.109353,
		0.0883,
		0.069976,
		0.054285,
		0.041091,
		0.030225,
		0.021491,
		0.014669,
		0.009521,
		0.005798,
		0.00325,
		0.001626,
		0.000691,
		0.000227,
		4.6e-005,
		0.011069,
		0.040162,
		0.082095,
		0.132788,
		0.189047,
		0.248386,
		0.308888,
		0.369091,
		0.427896,
		0.484498,
		0.538323,
		0.588982,
		0.636233,
		0.679948,
		0.720092,
		0.756699,
		0.789856,
		0.819691,
		0.846361,
		0.870045,
		0.890936,
		0.909232,
		0.92514,
		0.938862,
		0.9506,
		0.960548,
		0.96889,
		0.975798,
		0.981426,
		0.985907,
		0.989349,
		0.991829,
		0.993394,
		0.994054,
		0.993781,
		0.992508,
		0.990133,
		0.986518,
		0.981492,
		0.974863,
		0.966419,
		0.95594,
		0.943208,
		0.928019,
		0.910193,
		0.889586,
		0.866103,
		0.839707,
		0.810427,
		0.778361,
		0.743679,
		0.706621,
		0.667493,
		0.626654,
		0.584512,
		0.541506,
		0.498098,
		0.454751,
		0.411925,
		0.370055,
		0.329549,
		0.290769,
		0.254032,
		0.219602,
		0.187686,
		0.158434,
		0.13194,
		0.108244,
		0.087336,
		0.069157,
		0.053606,
		0.040544,
		0.029799,
		0.021172,
		0.014439,
		0.009364,
		0.005698,
		0.003191,
		0.001596,
		0.000677,
		0.000222,
		4.5e-005,
		0.012092,
		0.043674,
		0.088884,
		0.14316,
		0.202976,
		0.265625,
		0.32905,
		0.391707,
		0.452462,
		0.510503,
		0.565272,
		0.61641,
		0.663715,
		0.707107,
		0.746597,
		0.782268,
		0.814257,
		0.842738,
		0.867913,
		0.890001,
		0.909232,
		0.925841,
		0.940063,
		0.952128,
		0.962262,
		0.970681,
		0.97759,
		0.983181,
		0.987631,
		0.991097,
		0.993717,
		0.995602,
		0.996836,
		0.997471,
		0.997522,
		0.996969,
		0.995751,
		0.993768,
		0.990884,
		0.986925,
		0.981687,
		0.974941,
		0.966443,
		0.95594,
		0.943187,
		0.927957,
		0.910055,
		0.889329,
		0.865681,
		0.839077,
		0.809553,
		0.777215,
		0.742246,
		0.704898,
		0.665486,
		0.624385,
		0.58201,
		0.53881,
		0.495252,
		0.451806,
		0.408933,
		0.367068,
		0.326616,
		0.287937,
		0.251342,
		0.217088,
		0.185375,
		0.156346,
		0.130086,
		0.10663,
		0.085958,
		0.068007,
		0.052669,
		0.039801,
		0.029228,
		0.020748,
		0.014139,
		0.009162,
		0.005571,
		0.003117,
		0.001557,
		0.000661,
		0.000216,
		4.4e-005,
		0.013263,
		0.047667,
		0.096544,
		0.154777,
		0.218463,
		0.28465,
		0.351133,
		0.416289,
		0.478955,
		0.538323,
		0.593863,
		0.645261,
		0.692366,
		0.735156,
		0.773702,
		0.808146,
		0.838682,
		0.865538,
		0.888967,
		0.909232,
		0.926607,
		0.941362,
		0.953766,
		0.964076,
		0.972543,
		0.9794,
		0.984868,
		0.989153,
		0.992443,
		0.994906,
		0.996693,
		0.99793,
		0.998721,
		0.999139,
		0.99923,
		0.999005,
		0.998438,
		0.997467,
		0.99599,
		0.993865,
		0.990914,
		0.986925,
		0.981661,
		0.974863,
		0.966268,
		0.955611,
		0.942641,
		0.927134,
		0.908898,
		0.887791,
		0.863728,
		0.836688,
		0.80672,
		0.773948,
		0.738568,
		0.700845,
		0.661108,
		0.61974,
		0.577165,
		0.533837,
		0.490224,
		0.446798,
		0.404013,
		0.362304,
		0.322066,
		0.283651,
		0.247362,
		0.213445,
		0.18209,
		0.153429,
		0.127539,
		0.104444,
		0.084118,
		0.06649,
		0.051448,
		0.038844,
		0.0285,
		0.020214,
		0.013763,
		0.00891,
		0.005413,
		0.003027,
		0.001511,
		0.00064,
		0.00021,
		4.3e-005,
		0.014613,
		0.05223,
		0.105229,
		0.167843,
		0.23574,
		0.3057,
		0.375363,
		0.443032,
		0.507526,
		0.568056,
		0.624138,
		0.675515,
		0.722107,
		0.763962,
		0.801227,
		0.834112,
		0.862877,
		0.887815,
		0.909232,
		0.927446,
		0.942774,
		0.955526,
		0.966004,
		0.974494,
		0.981267,
		0.986574,
		0.99065,
		0.993706,
		0.995933,
		0.997501,
		0.998558,
		0.99923,
		0.99962,
		0.999806,
		0.999841,
		0.999749,
		0.999523,
		0.999123,
		0.998474,
		0.997467,
		0.995957,
		0.993768,
		0.990698,
		0.986518,
		0.980982,
		0.973833,
		0.96481,
		0.953657,
		0.940135,
		0.924032,
		0.905174,
		0.883435,
		0.858746,
		0.831105,
		0.800578,
		0.767303,
		0.73149,
		0.693416,
		0.653418,
		0.611886,
		0.569245,
		0.52595,
		0.482466,
		0.439258,
		0.396774,
		0.355436,
		0.31563,
		0.277694,
		0.241918,
		0.208535,
		0.177722,
		0.1496,
		0.124235,
		0.101641,
		0.081783,
		0.064584,
		0.049927,
		0.037661,
		0.027608,
		0.019564,
		0.013309,
		0.008609,
		0.005226,
		0.00292,
		0.001456,
		0.000617,
		0.000202,
		4.1e-005,
		0.016179,
		0.057479,
		0.11513,
		0.182603,
		0.255081,
		0.32905,
		0.40199,
		0.472143,
		0.538323,
		0.599783,
		0.656102,
		0.707107,
		0.752804,
		0.793332,
		0.828924,
		0.859876,
		0.886525,
		0.909232,
		0.928371,
		0.944313,
		0.957423,
		0.968053,
		0.976536,
		0.983187,
		0.988294,
		0.992122,
		0.994912,
		0.996876,
		0.9982,
		0.999046,
		0.999548,
		0.999818,
		0.999942,
		0.999983,
		0.999981,
		0.999952,
		0.999886,
		0.999749,
		0.999483,
		0.999005,
		0.99821,
		0.996969,
		0.995127,
		0.992508,
		0.988911,
		0.984115,
		0.977887,
		0.969981,
		0.960152,
		0.948161,
		0.933787,
		0.916836,
		0.897154,
		0.874633,
		0.849223,
		0.820937,
		0.789857,
		0.756135,
		0.719989,
		0.681705,
		0.641623,
		0.600132,
		0.557657,
		0.514646,
		0.471554,
		0.428835,
		0.386925,
		0.346229,
		0.307118,
		0.269916,
		0.234893,
		0.202269,
		0.172206,
		0.144812,
		0.12014,
		0.098195,
		0.078936,
		0.062277,
		0.0481,
		0.036251,
		0.026551,
		0.018799,
		0.012777,
		0.008259,
		0.005009,
		0.002796,
		0.001394,
		0.00059,
		0.000193,
		3.9e-005,
		0.018011,
		0.063557,
		0.12648,
		0.199356,
		0.276811,
		0.355015,
		0.431293,
		0.503836,
		0.571482,
		0.63355,
		0.689715,
		0.739908,
		0.78425,
		0.82299,
		0.856465,
		0.88507,
		0.909232,
		0.929393,
		0.945995,
		0.959469,
		0.97023,
		0.978669,
		0.985151,
		0.990011,
		0.993551,
		0.996042,
		0.997721,
		0.998793,
		0.999429,
		0.999771,
		0.999929,
		0.999986,
		0.999999,
		1,
		0.999998,
		0.999983,
		0.999932,
		0.999806,
		0.999561,
		0.999139,
		0.99847,
		0.997471,
		0.996039,
		0.994054,
		0.991373,
		0.987832,
		0.983245,
		0.977409,
		0.970107,
		0.961111,
		0.950194,
		0.937137,
		0.921738,
		0.903823,
		0.883254,
		0.859942,
		0.833854,
		0.805015,
		0.773521,
		0.73953,
		0.703269,
		0.665023,
		0.625132,
		0.58398,
		0.541984,
		0.499578,
		0.457205,
		0.415302,
		0.374285,
		0.334543,
		0.296425,
		0.260237,
		0.22623,
		0.194608,
		0.165515,
		0.139047,
		0.115245,
		0.094104,
		0.075576,
		0.059573,
		0.04597,
		0.034615,
		0.025331,
		0.01792,
		0.01217,
		0.00786,
		0.004764,
		0.002657,
		0.001323,
		0.00056,
		0.000183,
		3.7e-005,
		0.020173,
		0.070648,
		0.139573,
		0.218463,
		0.301315,
		0.383959,
		0.463572,
		0.538323,
		0.607109,
		0.669354,
		0.72486,
		0.773702,
		0.816138,
		0.852555,
		0.883417,
		0.909232,
		0.930531,
		0.947842,
		0.961681,
		0.972543,
		0.980889,
		0.987147,
		0.991705,
		0.994912,
		0.997073,
		0.998451,
		0.999268,
		0.999707,
		0.999909,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999986,
		0.999937,
		0.999818,
		0.999595,
		0.99923,
		0.998688,
		0.99793,
		0.996917,
		0.995602,
		0.993929,
		0.991829,
		0.989217,
		0.985988,
		0.982016,
		0.977154,
		0.971234,
		0.964069,
		0.955459,
		0.945195,
		0.933069,
		0.918878,
		0.902438,
		0.883593,
		0.862222,
		0.838252,
		0.81166,
		0.782486,
		0.75083,
		0.716858,
		0.680795,
		0.642924,
		0.603579,
		0.563133,
		0.521988,
		0.480563,
		0.43928,
		0.398556,
		0.358785,
		0.320332,
		0.283525,
		0.248646,
		0.21593,
		0.185558,
		0.157661,
		0.13232,
		0.109565,
		0.089384,
		0.07172,
		0.056483,
		0.043547,
		0.032763,
		0.023956,
		0.016933,
		0.011491,
		0.007416,
		0.004491,
		0.002503,
		0.001246,
		0.000526,
		0.000172,
		3.5e-005,
		0.022747,
		0.078987,
		0.154777,
		0.24037,
		0.32905,
		0.416289,
		0.499142,
		0.575799,
		0.645261,
		0.707107,
		0.761317,
		0.808146,
		0.848031,
		0.881522,
		0.909232,
		0.931803,
		0.949876,
		0.964076,
		0.974996,
		0.983187,
		0.989154,
		0.99335,
		0.996177,
		0.997978,
		0.999046,
		0.999618,
		0.999882,
		0.999977,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.99992,
		0.999771,
		0.999493,
		0.999046,
		0.998392,
		0.997501,
		0.996346,
		0.994906,
		0.993163,
		0.991097,
		0.988688,
		0.985907,
		0.982718,
		0.979067,
		0.974887,
		0.970086,
		0.964555,
		0.958161,
		0.950751,
		0.942154,
		0.932184,
		0.92065,
		0.907357,
		0.892122,
		0.874775,
		0.855177,
		0.83322,
		0.808844,
		0.782037,
		0.752845,
		0.721373,
		0.687787,
		0.652309,
		0.615215,
		0.576825,
		0.537498,
		0.497617,
		0.457578,
		0.417783,
		0.37862,
		0.340459,
		0.303641,
		0.268467,
		0.235198,
		0.204045,
		0.175172,
		0.148694,
		0.124676,
		0.103142,
		0.084068,
		0.067396,
		0.053032,
		0.040853,
		0.030711,
		0.022437,
		0.015848,
		0.010746,
		0.00693,
		0.004194,
		0.002336,
		0.001162,
		0.000491,
		0.00016,
		3.3e-005,
		0.025825,
		0.088879,
		0.172561,
		0.265625,
		0.360554,
		0.452462,
		0.538323,
		0.61641,
		0.685901,
		0.746597,
		0.798713,
		0.842738,
		0.879329,
		0.909232,
		0.933236,
		0.952128,
		0.966673,
		0.977591,
		0.985548,
		0.991144,
		0.994912,
		0.997312,
		0.998733,
		0.999493,
		0.999844,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999977,
		0.999897,
		0.999707,
		0.999355,
		0.998793,
		0.997978,
		0.996876,
		0.995458,
		0.993706,
		0.991607,
		0.989153,
		0.986344,
		0.983181,
		0.979666,
		0.975798,
		0.971575,
		0.966984,
		0.961999,
		0.956583,
		0.950676,
		0.9442,
		0.937057,
		0.929125,
		0.920262,
		0.91031,
		0.899099,
		0.886448,
		0.872181,
		0.856127,
		0.838132,
		0.818069,
		0.795845,
		0.771408,
		0.744755,
		0.715937,
		0.685059,
		0.652282,
		0.617822,
		0.581942,
		0.544946,
		0.507174,
		0.468986,
		0.430753,
		0.392847,
		0.35563,
		0.319444,
		0.284601,
		0.251376,
		0.220006,
		0.19068,
		0.163544,
		0.138695,
		0.116188,
		0.096035,
		0.078208,
		0.062646,
		0.049254,
		0.037912,
		0.028478,
		0.02079,
		0.014673,
		0.009942,
		0.006407,
		0.003874,
		0.002157,
		0.001072,
		0.000452,
		0.000147,
		3e-005,
		0.028884,
		0.100129,
		0.19336,
		0.294891,
		0.39646,
		0.492972,
		0.581406,
		0.660215,
		0.728851,
		0.787425,
		0.836467,
		0.876762,
		0.909232,
		0.93486,
		0.95463,
		0.969489,
		0.980326,
		0.987949,
		0.993081,
		0.996348,
		0.99828,
		0.999313,
		0.999788,
		0.999959,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999864,
		0.999618,
		0.999166,
		0.998451,
		0.997422,
		0.996042,
		0.994282,
		0.992122,
		0.989554,
		0.986574,
		0.983187,
		0.9794,
		0.975226,
		0.970681,
		0.965783,
		0.960548,
		0.954993,
		0.94913,
		0.942964,
		0.936491,
		0.929694,
		0.92254,
		0.914977,
		0.906936,
		0.898322,
		0.889025,
		0.878913,
		0.867837,
		0.855639,
		0.842154,
		0.827216,
		0.810671,
		0.792377,
		0.772221,
		0.750119,
		0.726028,
		0.699951,
		0.67194,
		0.642099,
		0.610581,
		0.577591,
		0.543376,
		0.508222,
		0.472443,
		0.436373,
		0.400354,
		0.36473,
		0.32983,
		0.295966,
		0.26342,
		0.232442,
		0.203242,
		0.175988,
		0.150806,
		0.12778,
		0.106953,
		0.088328,
		0.071873,
		0.057525,
		0.045192,
		0.034759,
		0.02609,
		0.019033,
		0.013424,
		0.009089,
		0.005853,
		0.003537,
		0.001968,
		0.000977,
		0.000412,
		0.000134,
		2.7e-005,
		0.028801,
		0.106843,
		0.212988,
		0.327031,
		0.437117,
		0.538297,
		0.628601,
		0.707107,
		0.773702,
		0.828924,
		0.873716,
		0.909232,
		0.936718,
		0.957423,
		0.972543,
		0.983187,
		0.990352,
		0.994912,
		0.997607,
		0.999046,
		0.999707,
		0.999944,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.999819,
		0.999493,
		0.998903,
		0.997978,
		0.996662,
		0.994912,
		0.9927,
		0.990011,
		0.986838,
		0.983187,
		0.979066,
		0.974494,
		0.969489,
		0.964076,
		0.95828,
		0.952128,
		0.945646,
		0.938862,
		0.931802,
		0.92449,
		0.916947,
		0.909189,
		0.901223,
		0.893048,
		0.884648,
		0.875992,
		0.867033,
		0.857704,
		0.847917,
		0.837568,
		0.826533,
		0.814675,
		0.801847,
		0.787895,
		0.77267,
		0.756027,
		0.737842,
		0.71801,
		0.696458,
		0.673152,
		0.648096,
		0.621342,
		0.592988,
		0.563181,
		0.532108,
		0.5,
		0.467118,
		0.43375,
		0.400201,
		0.366781,
		0.333802,
		0.30156,
		0.270336,
		0.240381,
		0.211916,
		0.185128,
		0.160162,
		0.137128,
		0.116093,
		0.097092,
		0.08012,
		0.065143,
		0.052099,
		0.0409,
		0.031435,
		0.023578,
		0.017189,
		0.012115,
		0.008198,
		0.005276,
		0.003186,
		0.001771,
		0.000879,
		0.000371,
		0.000121,
		2.4e-005,
		0.021807,
		0.093895,
		0.208583,
		0.342523,
		0.472358,
		0.585368,
		0.679351,
		0.756663,
		0.819691,
		0.870045,
		0.909232,
		0.938862,
		0.960553,
		0.975845,
		0.986146,
		0.9927,
		0.996571,
		0.998635,
		0.999581,
		0.99992,
		0.999995,
		1,
		0.999996,
		0.999944,
		0.999752,
		0.999313,
		0.998528,
		0.997312,
		0.995601,
		0.99335,
		0.990536,
		0.987147,
		0.983187,
		0.978669,
		0.973615,
		0.968053,
		0.962012,
		0.955526,
		0.948631,
		0.941362,
		0.933754,
		0.925841,
		0.917657,
		0.909232,
		0.900598,
		0.891783,
		0.882813,
		0.873711,
		0.864496,
		0.855181,
		0.845772,
		0.836264,
		0.826642,
		0.816874,
		0.806913,
		0.796695,
		0.786138,
		0.775143,
		0.763595,
		0.751367,
		0.738324,
		0.724327,
		0.709237,
		0.692926,
		0.675281,
		0.656209,
		0.635646,
		0.613563,
		0.589968,
		0.564909,
		0.538479,
		0.510811,
		0.482077,
		0.452485,
		0.422271,
		0.391695,
		0.361029,
		0.330552,
		0.300538,
		0.271253,
		0.242943,
		0.21583,
		0.190107,
		0.165934,
		0.143438,
		0.122709,
		0.103804,
		0.086747,
		0.07153,
		0.058116,
		0.046446,
		0.036436,
		0.027985,
		0.020976,
		0.015282,
		0.010764,
		0.007279,
		0.004682,
		0.002826,
		0.00157,
		0.000779,
		0.000328,
		0.000107,
		2.2e-005,
		0.010604,
		0.057659,
		0.154555,
		0.294361,
		0.45206,
		0.598814,
		0.716734,
		0.803188,
		0.864779,
		0.909189,
		0.941362,
		0.964076,
		0.9794,
		0.989154,
		0.994912,
		0.997978,
		0.999381,
		0.999882,
		0.999993,
		1,
		0.999994,
		0.99992,
		0.999651,
		0.999046,
		0.997978,
		0.996348,
		0.994084,
		0.991144,
		0.987511,
		0.983187,
		0.978187,
		0.972543,
		0.966289,
		0.959469,
		0.952128,
		0.944313,
		0.93607,
		0.927446,
		0.918486,
		0.909232,
		0.899725,
		0.890001,
		0.880097,
		0.870045,
		0.859876,
		0.849616,
		0.839291,
		0.828924,
		0.818535,
		0.808142,
		0.797756,
		0.787387,
		0.777035,
		0.766692,
		0.756338,
		0.745942,
		0.735456,
		0.724819,
		0.713954,
		0.702768,
		0.691157,
		0.679005,
		0.666188,
		0.652581,
		0.63806,
		0.622512,
		0.605834,
		0.587947,
		0.568794,
		0.548352,
		0.526629,
		0.50367,
		0.47956,
		0.454416,
		0.428393,
		0.401677,
		0.374475,
		0.347016,
		0.31954,
		0.292291,
		0.265509,
		0.239425,
		0.214251,
		0.190179,
		0.167375,
		0.145975,
		0.126086,
		0.107782,
		0.091109,
		0.076083,
		0.062691,
		0.0509,
		0.040651,
		0.031869,
		0.024461,
		0.018324,
		0.013341,
		0.009391,
		0.006347,
		0.00408,
		0.002461,
		0.001366,
		0.000678,
		0.000285,
		9.3e-005,
		1.9e-005,
		0.002407,
		0.020197,
		0.073201,
		0.175889,
		0.325662,
		0.5,
		0.666205,
		0.797981,
		0.886448,
		0.938519,
		0.967203,
		0.983139,
		0.992122,
		0.996876,
		0.999046,
		0.999819,
		0.999989,
		1,
		0.999991,
		0.999882,
		0.999493,
		0.998635,
		0.997147,
		0.994912,
		0.991857,
		0.987949,
		0.983187,
		0.977591,
		0.971204,
		0.964076,
		0.956268,
		0.947842,
		0.938862,
		0.929393,
		0.919497,
		0.909232,
		0.898654,
		0.887815,
		0.876762,
		0.865538,
		0.854185,
		0.842738,
		0.83123,
		0.819691,
		0.808146,
		0.796619,
		0.785132,
		0.773702,
		0.762345,
		0.751076,
		0.739907,
		0.728847,
		0.717902,
		0.707073,
		0.696355,
		0.685735,
		0.675192,
		0.664693,
		0.654192,
		0.643631,
		0.632938,
		0.622029,
		0.61081,
		0.599175,
		0.587017,
		0.574223,
		0.560686,
		0.546305,
		0.530993,
		0.514679,
		0.497317,
		0.478887,
		0.459397,
		0.438887,
		0.417432,
		0.395137,
		0.372134,
		0.348584,
		0.324667,
		0.300582,
		0.276532,
		0.252728,
		0.229374,
		0.206667,
		0.184787,
		0.163895,
		0.144131,
		0.125608,
		0.108413,
		0.092608,
		0.078227,
		0.06528,
		0.053754,
		0.043615,
		0.034811,
		0.027273,
		0.020921,
		0.015662,
		0.011397,
		0.008018,
		0.005416,
		0.003479,
		0.002098,
		0.001164,
		0.000577,
		0.000243,
		7.9e-005,
		1.6e-005,
		5.1e-005,
		0.002141,
		0.016081,
		0.05954,
		0.148905,
		0.288229,
		0.462908,
		0.642924,
		0.795954,
		0.902538,
		0.96241,
		0.988808,
		0.997575,
		0.999659,
		0.999982,
		1,
		0.999986,
		0.999819,
		0.999236,
		0.997978,
		0.995843,
		0.9927,
		0.988485,
		0.983187,
		0.976835,
		0.969489,
		0.961225,
		0.952128,
		0.94229,
		0.931803,
		0.920755,
		0.909232,
		0.897313,
		0.88507,
		0.872571,
		0.859876,
		0.84704,
		0.834112,
		0.821134,
		0.808146,
		0.795181,
		0.782268,
		0.769433,
		0.756699,
		0.744084,
		0.731605,
		0.719275,
		0.707107,
		0.695109,
		0.683291,
		0.671658,
		0.660215,
		0.648966,
		0.637912,
		0.627051,
		0.616381,
		0.605889,
		0.595562,
		0.585374,
		0.575293,
		0.565277,
		0.555271,
		0.54521,
		0.535021,
		0.524618,
		0.513912,
		0.502809,
		0.491211,
		0.479026,
		0.46617,
		0.452566,
		0.438158,
		0.422905,
		0.40679,
		0.389823,
		0.372038,
		0.353499,
		0.334294,
		0.314536,
		0.294361,
		0.27392,
		0.253379,
		0.232908,
		0.212684,
		0.192875,
		0.173644,
		0.155141,
		0.137498,
		0.120829,
		0.105226,
		0.090758,
		0.077474,
		0.0654,
		0.054541,
		0.044882,
		0.036394,
		0.02903,
		0.02273,
		0.017426,
		0.013038,
		0.009482,
		0.006667,
		0.004501,
		0.00289,
		0.001742,
		0.000966,
		0.000479,
		0.000201,
		6.5e-005,
		1.3e-005,
		0.000326,
		0.006041,
		0.031269,
		0.09321,
		0.202269,
		0.354699,
		0.530799,
		0.700845,
		0.837497,
		0.927549,
		0.974912,
		0.993865,
		0.9991,
		0.999659,
		0.998793,
		0.996876,
		0.993706,
		0.989154,
		0.983187,
		0.975845,
		0.967218,
		0.957423,
		0.946591,
		0.93486,
		0.922365,
		0.909232,
		0.895582,
		0.881522,
		0.867151,
		0.852555,
		0.837812,
		0.82299,
		0.808146,
		0.793332,
		0.778592,
		0.763962,
		0.749475,
		0.735156,
		0.721027,
		0.707107,
		0.69341,
		0.679948,
		0.66673,
		0.653763,
		0.641052,
		0.628601,
		0.61641,
		0.604481,
		0.592813,
		0.581406,
		0.570256,
		0.559361,
		0.548717,
		0.53832,
		0.528163,
		0.518237,
		0.508529,
		0.499021,
		0.48969,
		0.480503,
		0.471423,
		0.462401,
		0.453381,
		0.444298,
		0.435083,
		0.425657,
		0.415942,
		0.405858,
		0.395328,
		0.384281,
		0.372656,
		0.360406,
		0.347498,
		0.333919,
		0.319677,
		0.304801,
		0.289343,
		0.273376,
		0.256992,
		0.240301,
		0.223428,
		0.206504,
		0.18967,
		0.173066,
		0.156828,
		0.141088,
		0.125963,
		0.11156,
		0.097968,
		0.08526,
		0.073489,
		0.062693,
		0.052889,
		0.04408,
		0.036252,
		0.029379,
		0.023421,
		0.018328,
		0.014044,
		0.010502,
		0.007633,
		0.005365,
		0.00362,
		0.002323,
		0.001399,
		0.000776,
		0.000384,
		0.000161,
		5.2e-005,
		1.1e-005,
		0.000596,
		0.00891,
		0.040099,
		0.108744,
		0.22133,
		0.372177,
		0.542935,
		0.707107,
		0.839721,
		0.927549,
		0.97297,
		0.988808,
		0.989142,
		0.983139,
		0.974494,
		0.964076,
		0.952128,
		0.938862,
		0.924495,
		0.909232,
		0.893264,
		0.876762,
		0.859876,
		0.842738,
		0.825463,
		0.808146,
		0.790869,
		0.773702,
		0.756699,
		0.739908,
		0.723368,
		0.707107,
		0.69115,
		0.675515,
		0.660215,
		0.645261,
		0.630658,
		0.61641,
		0.602518,
		0.588982,
		0.575799,
		0.562964,
		0.550474,
		0.538323,
		0.526505,
		0.515011,
		0.503836,
		0.492972,
		0.48241,
		0.472143,
		0.462163,
		0.452462,
		0.443032,
		0.433863,
		0.424946,
		0.416269,
		0.407819,
		0.399576,
		0.39152,
		0.383622,
		0.375848,
		0.368158,
		0.360506,
		0.352837,
		0.345096,
		0.337219,
		0.329143,
		0.320804,
		0.312142,
		0.3031,
		0.29363,
		0.283695,
		0.27327,
		0.262345,
		0.250926,
		0.239036,
		0.226717,
		0.214025,
		0.201032,
		0.187825,
		0.1745,
		0.161159,
		0.147912,
		0.134865,
		0.122125,
		0.109792,
		0.097956,
		0.086698,
		0.076086,
		0.066174,
		0.057003,
		0.0486,
		0.040976,
		0.034131,
		0.028054,
		0.022723,
		0.018105,
		0.01416,
		0.010844,
		0.008105,
		0.005889,
		0.004136,
		0.00279,
		0.00179,
		0.001077,
		0.000597,
		0.000295,
		0.000124,
		4e-005,
		0.000712,
		0.00978,
		0.041877,
		0.110603,
		0.222444,
		0.372177,
		0.540983,
		0.700845,
		0.825636,
		0.902538,
		0.935426,
		0.938519,
		0.926633,
		0.909189,
		0.890001,
		0.870045,
		0.849616,
		0.828924,
		0.808146,
		0.787425,
		0.766878,
		0.746597,
		0.726654,
		0.707107,
		0.687996,
		0.669354,
		0.651201,
		0.63355,
		0.61641,
		0.599783,
		0.583667,
		0.568056,
		0.552945,
		0.538323,
		0.52418,
		0.510503,
		0.497281,
		0.484498,
		0.472143,
		0.460201,
		0.448658,
		0.437501,
		0.426716,
		0.416289,
		0.406208,
		0.39646,
		0.387032,
		0.377913,
		0.369091,
		0.360554,
		0.352291,
		0.344293,
		0.336549,
		0.329048,
		0.321779,
		0.314732,
		0.307892,
		0.301242,
		0.294765,
		0.288436,
		0.282229,
		0.276112,
		0.270048,
		0.263995,
		0.257911,
		0.251747,
		0.245457,
		0.238991,
		0.232304,
		0.225354,
		0.218106,
		0.210532,
		0.202613,
		0.194342,
		0.185725,
		0.176778,
		0.167531,
		0.158028,
		0.14832,
		0.138472,
		0.128554,
		0.118641,
		0.108813,
		0.099147,
		0.089722,
		0.080608,
		0.071873,
		0.063573,
		0.055758,
		0.048465,
		0.041725,
		0.035553,
		0.029959,
		0.024941,
		0.02049,
		0.016587,
		0.013209,
		0.010326,
		0.007904,
		0.005905,
		0.004288,
		0.003011,
		0.00203,
		0.001301,
		0.000783,
		0.000434,
		0.000215,
		9e-005,
		2.9e-005,
		0.000719,
		0.00978,
		0.041577,
		0.108744,
		0.215674,
		0.354699,
		0.505937,
		0.642924,
		0.743241,
		0.797981,
		0.81321,
		0.803188,
		0.781582,
		0.756663,
		0.731605,
		0.707107,
		0.683291,
		0.660215,
		0.637915,
		0.61641,
		0.595706,
		0.575799,
		0.556677,
		0.538323,
		0.520718,
		0.503836,
		0.487653,
		0.472143,
		0.457278,
		0.443032,
		0.429378,
		0.416289,
		0.403741,
		0.391707,
		0.380165,
		0.369091,
		0.358463,
		0.34826,
		0.338462,
		0.32905,
		0.320005,
		0.31131,
		0.302949,
		0.294906,
		0.287165,
		0.279714,
		0.272538,
		0.265625,
		0.258963,
		0.25254,
		0.246346,
		0.24037,
		0.234602,
		0.229034,
		0.223654,
		0.218452,
		0.213418,
		0.208539,
		0.203799,
		0.199181,
		0.194665,
		0.190228,
		0.185844,
		0.181483,
		0.177115,
		0.172707,
		0.168226,
		0.163637,
		0.15891,
		0.154017,
		0.148931,
		0.143636,
		0.138117,
		0.132372,
		0.126401,
		0.120219,
		0.113845,
		0.107308,
		0.100644,
		0.093895,
		0.08711,
		0.080339,
		0.073636,
		0.067052,
		0.06064,
		0.054447,
		0.048517,
		0.04289,
		0.037596,
		0.03266,
		0.028102,
		0.023933,
		0.020157,
		0.016772,
		0.013772,
		0.011143,
		0.008869,
		0.00693,
		0.005302,
		0.00396,
		0.002874,
		0.002017,
		0.001359,
		0.000871,
		0.000524,
		0.00029,
		0.000143,
		6e-005,
		2e-005,
		0.000676,
		0.00891,
		0.036721,
		0.09321,
		0.179781,
		0.288229,
		0.401745,
		0.5,
		0.567267,
		0.598814,
		0.600964,
		0.585368,
		0.56247,
		0.538297,
		0.515011,
		0.492972,
		0.472143,
		0.452462,
		0.433865,
		0.416289,
		0.399673,
		0.383959,
		0.369091,
		0.355015,
		0.341684,
		0.32905,
		0.317068,
		0.3057,
		0.294906,
		0.28465,
		0.2749,
		0.265625,
		0.256796,
		0.248386,
		0.24037,
		0.232725,
		0.225429,
		0.218463,
		0.211807,
		0.205443,
		0.199356,
		0.193529,
		0.18795,
		0.182603,
		0.177478,
		0.172561,
		0.167843,
		0.163312,
		0.15896,
		0.154777,
		0.150754,
		0.146884,
		0.143159,
		0.139572,
		0.136115,
		0.132781,
		0.129562,
		0.12645,
		0.123433,
		0.120501,
		0.117642,
		0.114839,
		0.112078,
		0.109339,
		0.106605,
		0.103854,
		0.101067,
		0.098223,
		0.095303,
		0.09229,
		0.08917,
		0.085931,
		0.082565,
		0.07907,
		0.075448,
		0.071706,
		0.067856,
		0.063915,
		0.059906,
		0.055852,
		0.051783,
		0.047728,
		0.043719,
		0.039786,
		0.03596,
		0.032269,
		0.028739,
		0.025391,
		0.022245,
		0.019315,
		0.016611,
		0.014139,
		0.011902,
		0.009899,
		0.008124,
		0.00657,
		0.005227,
		0.004083,
		0.003122,
		0.002331,
		0.001691,
		0.001186,
		0.000799,
		0.000512,
		0.000308,
		0.00017,
		8.4e-005,
		3.5e-005,
		1.1e-005,
		0.000475,
		0.006041,
		0.024116,
		0.05954,
		0.112084,
		0.175889,
		0.240551,
		0.294361,
		0.328938,
		0.342523,
		0.339533,
		0.327031,
		0.311037,
		0.294891,
		0.279714,
		0.265625,
		0.25254,
		0.24037,
		0.229035,
		0.218463,
		0.208589,
		0.199356,
		0.19071,
		0.182603,
		0.174994,
		0.167843,
		0.161115,
		0.154777,
		0.148801,
		0.14316,
		0.137829,
		0.132788,
		0.128015,
		0.123492,
		0.119202,
		0.11513,
		0.111261,
		0.107582,
		0.104082,
		0.100747,
		0.09757,
		0.094539,
		0.091646,
		0.088884,
		0.086243,
		0.083718,
		0.081301,
		0.078987,
		0.07677,
		0.074644,
		0.072605,
		0.070648,
		0.068768,
		0.066962,
		0.065226,
		0.063554,
		0.061944,
		0.06039,
		0.058887,
		0.057429,
		0.05601,
		0.054623,
		0.053259,
		0.05191,
		0.050566,
		0.049219,
		0.047857,
		0.046472,
		0.045055,
		0.043596,
		0.04209,
		0.040531,
		0.038915,
		0.037241,
		0.035511,
		0.033727,
		0.031895,
		0.030023,
		0.028122,
		0.026203,
		0.024279,
		0.022365,
		0.020474,
		0.018622,
		0.016822,
		0.015087,
		0.01343,
		0.011859,
		0.010384,
		0.009012,
		0.007747,
		0.006591,
		0.005545,
		0.00461,
		0.003782,
		0.003057,
		0.002431,
		0.001898,
		0.001451,
		0.001083,
		0.000785,
		0.000551,
		0.000371,
		0.000237,
		0.000143,
		7.9e-005,
		3.9e-005,
		1.6e-005,
		0.000173,
		0.002141,
		0.008349,
		0.020197,
		0.037339,
		0.057659,
		0.077731,
		0.093895,
		0.103703,
		0.106843,
		0.104887,
		0.100129,
		0.094456,
		0.088879,
		0.083718,
		0.078987,
		0.074644,
		0.070648,
		0.066963,
		0.063557,
		0.060404,
		0.057479,
		0.054761,
		0.05223,
		0.049871,
		0.047667,
		0.045605,
		0.043674,
		0.041863,
		0.040162,
		0.038563,
		0.037057,
		0.035637,
		0.034297,
		0.033031,
		0.031834,
		0.0307,
		0.029626,
		0.028608,
		0.027641,
		0.026722,
		0.025848,
		0.025016,
		0.024223,
		0.023468,
		0.022747,
		0.022059,
		0.021402,
		0.020774,
		0.020173,
		0.019597,
		0.019046,
		0.018518,
		0.018011,
		0.017525,
		0.017057,
		0.016608,
		0.016175,
		0.015757,
		0.015352,
		0.014959,
		0.014576,
		0.014199,
		0.013828,
		0.013459,
		0.01309,
		0.012718,
		0.01234,
		0.011955,
		0.011559,
		0.011152,
		0.010731,
		0.010297,
		0.009847,
		0.009384,
		0.008906,
		0.008417,
		0.007919,
		0.007413,
		0.006903,
		0.006393,
		0.005885,
		0.005385,
		0.004895,
		0.00442,
		0.003962,
		0.003525,
		0.003111,
		0.002723,
		0.002362,
		0.002029,
		0.001726,
		0.001451,
		0.001206,
		0.000989,
		0.000799,
		0.000635,
		0.000496,
		0.000379,
		0.000283,
		0.000205,
		0.000144,
		9.7e-005,
		6.2e-005,
		3.7e-005,
		2.1e-005,
		1e-005,
		1e-005,
		2.1e-005,
		3.7e-005,
		6.2e-005,
		9.7e-005,
		0.000144,
		0.000205,
		0.000283,
		0.000379,
		0.000496,
		0.000635,
		0.000799,
		0.000989,
		0.001206,
		0.001451,
		0.001726,
		0.002029,
		0.002362,
		0.002723,
		0.003111,
		0.003525,
		0.003962,
		0.00442,
		0.004895,
		0.005385,
		0.005885,
		0.006393,
		0.006903,
		0.007413,
		0.007919,
		0.008417,
		0.008906,
		0.009384,
		0.009847,
		0.010297,
		0.010731,
		0.011152,
		0.011559,
		0.011955,
		0.01234,
		0.012718,
		0.01309,
		0.013459,
		0.013828,
		0.014199,
		0.014576,
		0.014959,
		0.015352,
		0.015757,
		0.016175,
		0.016608,
		0.017057,
		0.017525,
		0.018011,
		0.018518,
		0.019046,
		0.019597,
		0.020173,
		0.020774,
		0.021402,
		0.022059,
		0.022747,
		0.023468,
		0.024223,
		0.025016,
		0.025848,
		0.026722,
		0.027641,
		0.028608,
		0.029626,
		0.0307,
		0.031834,
		0.033031,
		0.034297,
		0.035637,
		0.037057,
		0.038563,
		0.040162,
		0.041863,
		0.043674,
		0.045605,
		0.047667,
		0.049871,
		0.05223,
		0.054761,
		0.057479,
		0.060404,
		0.063557,
		0.066963,
		0.070648,
		0.074644,
		0.078987,
		0.083718,
		0.088879,
		0.094456,
		0.100129,
		0.104887,
		0.106843,
		0.103703,
		0.093895,
		0.077731,
		0.057659,
		0.037339,
		0.020197,
		0.008349,
		0.002141,
		0.000173,
		1.6e-005,
		3.9e-005,
		7.9e-005,
		0.000143,
		0.000237,
		0.000371,
		0.000551,
		0.000785,
		0.001083,
		0.001451,
		0.001898,
		0.002431,
		0.003057,
		0.003782,
		0.00461,
		0.005545,
		0.006591,
		0.007747,
		0.009012,
		0.010384,
		0.011859,
		0.01343,
		0.015087,
		0.016822,
		0.018622,
		0.020474,
		0.022365,
		0.024279,
		0.026203,
		0.028122,
		0.030023,
		0.031895,
		0.033727,
		0.035511,
		0.037241,
		0.038915,
		0.040531,
		0.04209,
		0.043596,
		0.045055,
		0.046472,
		0.047857,
		0.049219,
		0.050566,
		0.05191,
		0.053259,
		0.054623,
		0.05601,
		0.057429,
		0.058887,
		0.06039,
		0.061944,
		0.063554,
		0.065226,
		0.066962,
		0.068768,
		0.070648,
		0.072605,
		0.074644,
		0.07677,
		0.078987,
		0.081301,
		0.083718,
		0.086243,
		0.088884,
		0.091646,
		0.094539,
		0.09757,
		0.100747,
		0.104082,
		0.107582,
		0.111261,
		0.11513,
		0.119202,
		0.123492,
		0.128015,
		0.132788,
		0.137829,
		0.14316,
		0.148801,
		0.154777,
		0.161115,
		0.167843,
		0.174994,
		0.182603,
		0.19071,
		0.199356,
		0.208589,
		0.218463,
		0.229035,
		0.24037,
		0.25254,
		0.265625,
		0.279714,
		0.294891,
		0.311037,
		0.327031,
		0.339533,
		0.342523,
		0.328938,
		0.294361,
		0.240551,
		0.175889,
		0.112084,
		0.05954,
		0.024116,
		0.006041,
		0.000475,
		1.1e-005,
		3.5e-005,
		8.4e-005,
		0.00017,
		0.000308,
		0.000512,
		0.000799,
		0.001186,
		0.001691,
		0.002331,
		0.003122,
		0.004083,
		0.005227,
		0.00657,
		0.008124,
		0.009899,
		0.011902,
		0.014139,
		0.016611,
		0.019315,
		0.022245,
		0.025391,
		0.028739,
		0.032269,
		0.03596,
		0.039786,
		0.043719,
		0.047728,
		0.051783,
		0.055852,
		0.059906,
		0.063915,
		0.067856,
		0.071706,
		0.075448,
		0.07907,
		0.082565,
		0.085931,
		0.08917,
		0.09229,
		0.095303,
		0.098223,
		0.101067,
		0.103854,
		0.106605,
		0.109339,
		0.112078,
		0.114839,
		0.117642,
		0.120501,
		0.123433,
		0.12645,
		0.129562,
		0.132781,
		0.136115,
		0.139572,
		0.143159,
		0.146884,
		0.150754,
		0.154777,
		0.15896,
		0.163312,
		0.167843,
		0.172561,
		0.177478,
		0.182603,
		0.18795,
		0.193529,
		0.199356,
		0.205443,
		0.211807,
		0.218463,
		0.225429,
		0.232725,
		0.24037,
		0.248386,
		0.256796,
		0.265625,
		0.2749,
		0.28465,
		0.294906,
		0.3057,
		0.317068,
		0.32905,
		0.341684,
		0.355015,
		0.369091,
		0.383959,
		0.399673,
		0.416289,
		0.433865,
		0.452462,
		0.472143,
		0.492972,
		0.515011,
		0.538297,
		0.56247,
		0.585368,
		0.600964,
		0.598814,
		0.567267,
		0.5,
		0.401745,
		0.288229,
		0.179781,
		0.09321,
		0.036721,
		0.00891,
		0.000676,
		2e-005,
		6e-005,
		0.000143,
		0.00029,
		0.000524,
		0.000871,
		0.001359,
		0.002017,
		0.002874,
		0.00396,
		0.005302,
		0.00693,
		0.008869,
		0.011143,
		0.013772,
		0.016772,
		0.020157,
		0.023933,
		0.028102,
		0.03266,
		0.037596,
		0.04289,
		0.048517,
		0.054447,
		0.06064,
		0.067052,
		0.073636,
		0.080339,
		0.08711,
		0.093895,
		0.100644,
		0.107308,
		0.113845,
		0.120219,
		0.126401,
		0.132372,
		0.138117,
		0.143636,
		0.148931,
		0.154017,
		0.15891,
		0.163637,
		0.168226,
		0.172707,
		0.177115,
		0.181483,
		0.185844,
		0.190228,
		0.194665,
		0.199181,
		0.203799,
		0.208539,
		0.213418,
		0.218452,
		0.223654,
		0.229034,
		0.234602,
		0.24037,
		0.246346,
		0.25254,
		0.258963,
		0.265625,
		0.272538,
		0.279714,
		0.287165,
		0.294906,
		0.302949,
		0.31131,
		0.320005,
		0.32905,
		0.338462,
		0.34826,
		0.358463,
		0.369091,
		0.380165,
		0.391707,
		0.403741,
		0.416289,
		0.429378,
		0.443032,
		0.457278,
		0.472143,
		0.487653,
		0.503836,
		0.520718,
		0.538323,
		0.556677,
		0.575799,
		0.595706,
		0.61641,
		0.637915,
		0.660215,
		0.683291,
		0.707107,
		0.731605,
		0.756663,
		0.781582,
		0.803188,
		0.81321,
		0.797981,
		0.743241,
		0.642924,
		0.505937,
		0.354699,
		0.215674,
		0.108744,
		0.041577,
		0.00978,
		0.000719,
		2.9e-005,
		9e-005,
		0.000215,
		0.000434,
		0.000783,
		0.001301,
		0.00203,
		0.003011,
		0.004288,
		0.005905,
		0.007904,
		0.010326,
		0.013209,
		0.016587,
		0.02049,
		0.024941,
		0.029959,
		0.035553,
		0.041725,
		0.048465,
		0.055758,
		0.063573,
		0.071873,
		0.080608,
		0.089722,
		0.099147,
		0.108813,
		0.118641,
		0.128554,
		0.138472,
		0.14832,
		0.158028,
		0.167531,
		0.176778,
		0.185725,
		0.194342,
		0.202613,
		0.210532,
		0.218106,
		0.225354,
		0.232304,
		0.238991,
		0.245457,
		0.251747,
		0.257911,
		0.263995,
		0.270048,
		0.276112,
		0.282229,
		0.288436,
		0.294765,
		0.301242,
		0.307892,
		0.314732,
		0.321779,
		0.329048,
		0.336549,
		0.344293,
		0.352291,
		0.360554,
		0.369091,
		0.377913,
		0.387032,
		0.39646,
		0.406208,
		0.416289,
		0.426716,
		0.437501,
		0.448658,
		0.460201,
		0.472143,
		0.484498,
		0.497281,
		0.510503,
		0.52418,
		0.538323,
		0.552945,
		0.568056,
		0.583667,
		0.599783,
		0.61641,
		0.63355,
		0.651201,
		0.669354,
		0.687996,
		0.707107,
		0.726654,
		0.746597,
		0.766878,
		0.787425,
		0.808146,
		0.828924,
		0.849616,
		0.870045,
		0.890001,
		0.909189,
		0.926633,
		0.938519,
		0.935426,
		0.902538,
		0.825636,
		0.700845,
		0.540983,
		0.372177,
		0.222444,
		0.110603,
		0.041877,
		0.00978,
		0.000712,
		4e-005,
		0.000124,
		0.000295,
		0.000597,
		0.001077,
		0.00179,
		0.00279,
		0.004136,
		0.005889,
		0.008105,
		0.010844,
		0.01416,
		0.018105,
		0.022723,
		0.028054,
		0.034131,
		0.040976,
		0.0486,
		0.057003,
		0.066174,
		0.076086,
		0.086698,
		0.097956,
		0.109792,
		0.122125,
		0.134865,
		0.147912,
		0.161159,
		0.1745,
		0.187825,
		0.201032,
		0.214025,
		0.226717,
		0.239036,
		0.250926,
		0.262345,
		0.27327,
		0.283695,
		0.29363,
		0.3031,
		0.312142,
		0.320804,
		0.329143,
		0.337219,
		0.345096,
		0.352837,
		0.360506,
		0.368158,
		0.375848,
		0.383622,
		0.39152,
		0.399576,
		0.407819,
		0.416269,
		0.424946,
		0.433863,
		0.443032,
		0.452462,
		0.462163,
		0.472143,
		0.48241,
		0.492972,
		0.503836,
		0.515011,
		0.526505,
		0.538323,
		0.550474,
		0.562964,
		0.575799,
		0.588982,
		0.602518,
		0.61641,
		0.630658,
		0.645261,
		0.660215,
		0.675515,
		0.69115,
		0.707107,
		0.723368,
		0.739908,
		0.756699,
		0.773702,
		0.790869,
		0.808146,
		0.825463,
		0.842738,
		0.859876,
		0.876762,
		0.893264,
		0.909232,
		0.924495,
		0.938862,
		0.952128,
		0.964076,
		0.974494,
		0.983139,
		0.989142,
		0.988808,
		0.97297,
		0.927549,
		0.839721,
		0.707107,
		0.542935,
		0.372177,
		0.22133,
		0.108744,
		0.040099,
		0.00891,
		0.000596,
		1.1e-005,
		5.2e-005,
		0.000161,
		0.000384,
		0.000776,
		0.001399,
		0.002323,
		0.00362,
		0.005365,
		0.007633,
		0.010502,
		0.014044,
		0.018328,
		0.023421,
		0.029379,
		0.036252,
		0.04408,
		0.052889,
		0.062693,
		0.073489,
		0.08526,
		0.097968,
		0.11156,
		0.125963,
		0.141088,
		0.156828,
		0.173066,
		0.18967,
		0.206504,
		0.223428,
		0.240301,
		0.256992,
		0.273376,
		0.289343,
		0.304801,
		0.319677,
		0.333919,
		0.347498,
		0.360406,
		0.372656,
		0.384281,
		0.395328,
		0.405858,
		0.415942,
		0.425657,
		0.435083,
		0.444298,
		0.453381,
		0.462401,
		0.471423,
		0.480503,
		0.48969,
		0.499021,
		0.508529,
		0.518237,
		0.528163,
		0.53832,
		0.548717,
		0.559361,
		0.570256,
		0.581406,
		0.592813,
		0.604481,
		0.61641,
		0.628601,
		0.641052,
		0.653763,
		0.66673,
		0.679948,
		0.69341,
		0.707107,
		0.721027,
		0.735156,
		0.749475,
		0.763962,
		0.778592,
		0.793332,
		0.808146,
		0.82299,
		0.837812,
		0.852555,
		0.867151,
		0.881522,
		0.895582,
		0.909232,
		0.922365,
		0.93486,
		0.946591,
		0.957423,
		0.967218,
		0.975845,
		0.983187,
		0.989154,
		0.993706,
		0.996876,
		0.998793,
		0.999659,
		0.9991,
		0.993865,
		0.974912,
		0.927549,
		0.837497,
		0.700845,
		0.530799,
		0.354699,
		0.202269,
		0.09321,
		0.031269,
		0.006041,
		0.000326,
		1.3e-005,
		6.5e-005,
		0.000201,
		0.000479,
		0.000966,
		0.001742,
		0.00289,
		0.004501,
		0.006667,
		0.009482,
		0.013038,
		0.017426,
		0.02273,
		0.02903,
		0.036394,
		0.044882,
		0.054541,
		0.0654,
		0.077474,
		0.090758,
		0.105226,
		0.120829,
		0.137498,
		0.155141,
		0.173644,
		0.192875,
		0.212684,
		0.232908,
		0.253379,
		0.27392,
		0.294361,
		0.314536,
		0.334294,
		0.353499,
		0.372038,
		0.389823,
		0.40679,
		0.422905,
		0.438158,
		0.452566,
		0.46617,
		0.479026,
		0.491211,
		0.502809,
		0.513912,
		0.524618,
		0.535021,
		0.54521,
		0.555271,
		0.565277,
		0.575293,
		0.585374,
		0.595562,
		0.605889,
		0.616381,
		0.627051,
		0.637912,
		0.648966,
		0.660215,
		0.671658,
		0.683291,
		0.695109,
		0.707107,
		0.719275,
		0.731605,
		0.744084,
		0.756699,
		0.769433,
		0.782268,
		0.795181,
		0.808146,
		0.821134,
		0.834112,
		0.84704,
		0.859876,
		0.872571,
		0.88507,
		0.897313,
		0.909232,
		0.920755,
		0.931803,
		0.94229,
		0.952128,
		0.961225,
		0.969489,
		0.976835,
		0.983187,
		0.988485,
		0.9927,
		0.995843,
		0.997978,
		0.999236,
		0.999819,
		0.999986,
		1,
		0.999982,
		0.999659,
		0.997575,
		0.988808,
		0.96241,
		0.902538,
		0.795954,
		0.642924,
		0.462908,
		0.288229,
		0.148905,
		0.05954,
		0.016081,
		0.002141,
		5.1e-005,
		1.6e-005,
		7.9e-005,
		0.000243,
		0.000577,
		0.001164,
		0.002098,
		0.003479,
		0.005416,
		0.008018,
		0.011397,
		0.015662,
		0.020921,
		0.027273,
		0.034811,
		0.043615,
		0.053754,
		0.06528,
		0.078227,
		0.092608,
		0.108413,
		0.125608,
		0.144131,
		0.163895,
		0.184787,
		0.206667,
		0.229374,
		0.252728,
		0.276532,
		0.300582,
		0.324667,
		0.348584,
		0.372134,
		0.395137,
		0.417432,
		0.438887,
		0.459397,
		0.478887,
		0.497317,
		0.514679,
		0.530993,
		0.546305,
		0.560686,
		0.574223,
		0.587017,
		0.599175,
		0.61081,
		0.622029,
		0.632938,
		0.643631,
		0.654192,
		0.664693,
		0.675192,
		0.685735,
		0.696355,
		0.707073,
		0.717902,
		0.728847,
		0.739907,
		0.751076,
		0.762345,
		0.773702,
		0.785132,
		0.796619,
		0.808146,
		0.819691,
		0.83123,
		0.842738,
		0.854185,
		0.865538,
		0.876762,
		0.887815,
		0.898654,
		0.909232,
		0.919497,
		0.929393,
		0.938862,
		0.947842,
		0.956268,
		0.964076,
		0.971204,
		0.977591,
		0.983187,
		0.987949,
		0.991857,
		0.994912,
		0.997147,
		0.998635,
		0.999493,
		0.999882,
		0.999991,
		1,
		0.999989,
		0.999819,
		0.999046,
		0.996876,
		0.992122,
		0.983139,
		0.967203,
		0.938519,
		0.886448,
		0.797981,
		0.666205,
		0.5,
		0.325662,
		0.175889,
		0.073201,
		0.020197,
		0.002407,
		1.9e-005,
		9.3e-005,
		0.000285,
		0.000678,
		0.001366,
		0.002461,
		0.00408,
		0.006347,
		0.009391,
		0.013341,
		0.018324,
		0.024461,
		0.031869,
		0.040651,
		0.0509,
		0.062691,
		0.076083,
		0.091109,
		0.107782,
		0.126086,
		0.145975,
		0.167375,
		0.190179,
		0.214251,
		0.239425,
		0.265509,
		0.292291,
		0.31954,
		0.347016,
		0.374475,
		0.401677,
		0.428393,
		0.454416,
		0.47956,
		0.50367,
		0.526629,
		0.548352,
		0.568794,
		0.587947,
		0.605834,
		0.622512,
		0.63806,
		0.652581,
		0.666188,
		0.679005,
		0.691157,
		0.702768,
		0.713954,
		0.724819,
		0.735456,
		0.745942,
		0.756338,
		0.766692,
		0.777035,
		0.787387,
		0.797756,
		0.808142,
		0.818535,
		0.828924,
		0.839291,
		0.849616,
		0.859876,
		0.870045,
		0.880097,
		0.890001,
		0.899725,
		0.909232,
		0.918486,
		0.927446,
		0.93607,
		0.944313,
		0.952128,
		0.959469,
		0.966289,
		0.972543,
		0.978187,
		0.983187,
		0.987511,
		0.991144,
		0.994084,
		0.996348,
		0.997978,
		0.999046,
		0.999651,
		0.99992,
		0.999994,
		1,
		0.999993,
		0.999882,
		0.999381,
		0.997978,
		0.994912,
		0.989154,
		0.9794,
		0.964076,
		0.941362,
		0.909189,
		0.864779,
		0.803188,
		0.716734,
		0.598814,
		0.45206,
		0.294361,
		0.154555,
		0.057659,
		0.010604,
		2.2e-005,
		0.000107,
		0.000328,
		0.000779,
		0.00157,
		0.002826,
		0.004682,
		0.007279,
		0.010764,
		0.015282,
		0.020976,
		0.027985,
		0.036436,
		0.046446,
		0.058116,
		0.07153,
		0.086747,
		0.103804,
		0.122709,
		0.143438,
		0.165934,
		0.190107,
		0.21583,
		0.242943,
		0.271253,
		0.300538,
		0.330552,
		0.361029,
		0.391695,
		0.422271,
		0.452485,
		0.482077,
		0.510811,
		0.538479,
		0.564909,
		0.589968,
		0.613563,
		0.635646,
		0.656209,
		0.675281,
		0.692926,
		0.709237,
		0.724327,
		0.738324,
		0.751367,
		0.763595,
		0.775143,
		0.786138,
		0.796695,
		0.806913,
		0.816874,
		0.826642,
		0.836264,
		0.845772,
		0.855181,
		0.864496,
		0.873711,
		0.882813,
		0.891783,
		0.900598,
		0.909232,
		0.917657,
		0.925841,
		0.933754,
		0.941362,
		0.948631,
		0.955526,
		0.962012,
		0.968053,
		0.973615,
		0.978669,
		0.983187,
		0.987147,
		0.990536,
		0.99335,
		0.995601,
		0.997312,
		0.998528,
		0.999313,
		0.999752,
		0.999944,
		0.999996,
		1,
		0.999995,
		0.99992,
		0.999581,
		0.998635,
		0.996571,
		0.9927,
		0.986146,
		0.975845,
		0.960553,
		0.938862,
		0.909232,
		0.870045,
		0.819691,
		0.756663,
		0.679351,
		0.585368,
		0.472358,
		0.342523,
		0.208583,
		0.093895,
		0.021807,
		2.4e-005,
		0.000121,
		0.000371,
		0.000879,
		0.001771,
		0.003186,
		0.005276,
		0.008198,
		0.012115,
		0.017189,
		0.023578,
		0.031435,
		0.0409,
		0.052099,
		0.065143,
		0.08012,
		0.097092,
		0.116093,
		0.137128,
		0.160162,
		0.185128,
		0.211916,
		0.240381,
		0.270336,
		0.30156,
		0.333802,
		0.366781,
		0.400201,
		0.43375,
		0.467118,
		0.5,
		0.532108,
		0.563181,
		0.592988,
		0.621342,
		0.648096,
		0.673152,
		0.696458,
		0.71801,
		0.737842,
		0.756027,
		0.77267,
		0.787895,
		0.801847,
		0.814675,
		0.826533,
		0.837568,
		0.847917,
		0.857704,
		0.867033,
		0.875992,
		0.884648,
		0.893048,
		0.901223,
		0.909189,
		0.916947,
		0.92449,
		0.931802,
		0.938862,
		0.945646,
		0.952128,
		0.95828,
		0.964076,
		0.969489,
		0.974494,
		0.979066,
		0.983187,
		0.986838,
		0.990011,
		0.9927,
		0.994912,
		0.996662,
		0.997978,
		0.998903,
		0.999493,
		0.999819,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999944,
		0.999707,
		0.999046,
		0.997607,
		0.994912,
		0.990352,
		0.983187,
		0.972543,
		0.957423,
		0.936718,
		0.909232,
		0.873716,
		0.828924,
		0.773702,
		0.707107,
		0.628601,
		0.538297,
		0.437117,
		0.327031,
		0.212988,
		0.106843,
		0.028801,
		2.7e-005,
		0.000134,
		0.000412,
		0.000977,
		0.001968,
		0.003537,
		0.005853,
		0.009089,
		0.013424,
		0.019033,
		0.02609,
		0.034759,
		0.045192,
		0.057525,
		0.071873,
		0.088328,
		0.106953,
		0.12778,
		0.150806,
		0.175988,
		0.203242,
		0.232442,
		0.26342,
		0.295966,
		0.32983,
		0.36473,
		0.400354,
		0.436373,
		0.472443,
		0.508222,
		0.543376,
		0.577591,
		0.610581,
		0.642099,
		0.67194,
		0.699951,
		0.726028,
		0.750119,
		0.772221,
		0.792377,
		0.810671,
		0.827216,
		0.842154,
		0.855639,
		0.867837,
		0.878913,
		0.889025,
		0.898322,
		0.906936,
		0.914977,
		0.92254,
		0.929694,
		0.936491,
		0.942964,
		0.94913,
		0.954993,
		0.960548,
		0.965783,
		0.970681,
		0.975226,
		0.9794,
		0.983187,
		0.986574,
		0.989554,
		0.992122,
		0.994282,
		0.996042,
		0.997422,
		0.998451,
		0.999166,
		0.999618,
		0.999864,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999959,
		0.999788,
		0.999313,
		0.99828,
		0.996348,
		0.993081,
		0.987949,
		0.980326,
		0.969489,
		0.95463,
		0.93486,
		0.909232,
		0.876762,
		0.836467,
		0.787425,
		0.728851,
		0.660215,
		0.581406,
		0.492972,
		0.39646,
		0.294891,
		0.19336,
		0.100129,
		0.028884,
		3e-005,
		0.000147,
		0.000452,
		0.001072,
		0.002157,
		0.003874,
		0.006407,
		0.009942,
		0.014673,
		0.02079,
		0.028478,
		0.037912,
		0.049254,
		0.062646,
		0.078208,
		0.096035,
		0.116188,
		0.138695,
		0.163544,
		0.19068,
		0.220006,
		0.251376,
		0.284601,
		0.319444,
		0.35563,
		0.392847,
		0.430753,
		0.468986,
		0.507174,
		0.544946,
		0.581942,
		0.617822,
		0.652282,
		0.685059,
		0.715937,
		0.744755,
		0.771408,
		0.795845,
		0.818069,
		0.838132,
		0.856127,
		0.872181,
		0.886448,
		0.899099,
		0.91031,
		0.920262,
		0.929125,
		0.937057,
		0.9442,
		0.950676,
		0.956583,
		0.961999,
		0.966984,
		0.971575,
		0.975798,
		0.979666,
		0.983181,
		0.986344,
		0.989153,
		0.991607,
		0.993706,
		0.995458,
		0.996876,
		0.997978,
		0.998793,
		0.999355,
		0.999707,
		0.999897,
		0.999977,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999844,
		0.999493,
		0.998733,
		0.997312,
		0.994912,
		0.991144,
		0.985548,
		0.977591,
		0.966673,
		0.952128,
		0.933236,
		0.909232,
		0.879329,
		0.842738,
		0.798713,
		0.746597,
		0.685901,
		0.61641,
		0.538323,
		0.452462,
		0.360554,
		0.265625,
		0.172561,
		0.088879,
		0.025825,
		3.3e-005,
		0.00016,
		0.000491,
		0.001162,
		0.002336,
		0.004194,
		0.00693,
		0.010746,
		0.015848,
		0.022437,
		0.030711,
		0.040853,
		0.053032,
		0.067396,
		0.084068,
		0.103142,
		0.124676,
		0.148694,
		0.175172,
		0.204045,
		0.235198,
		0.268467,
		0.303641,
		0.340459,
		0.37862,
		0.417783,
		0.457578,
		0.497617,
		0.537498,
		0.576825,
		0.615215,
		0.652309,
		0.687787,
		0.721373,
		0.752845,
		0.782037,
		0.808844,
		0.83322,
		0.855177,
		0.874775,
		0.892122,
		0.907357,
		0.92065,
		0.932184,
		0.942154,
		0.950751,
		0.958161,
		0.964555,
		0.970086,
		0.974887,
		0.979067,
		0.982718,
		0.985907,
		0.988688,
		0.991097,
		0.993163,
		0.994906,
		0.996346,
		0.997501,
		0.998392,
		0.999046,
		0.999493,
		0.999771,
		0.99992,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999977,
		0.999882,
		0.999618,
		0.999046,
		0.997978,
		0.996177,
		0.99335,
		0.989154,
		0.983187,
		0.974996,
		0.964076,
		0.949876,
		0.931803,
		0.909232,
		0.881522,
		0.848031,
		0.808146,
		0.761317,
		0.707107,
		0.645261,
		0.575799,
		0.499142,
		0.416289,
		0.32905,
		0.24037,
		0.154777,
		0.078987,
		0.022747,
		3.5e-005,
		0.000172,
		0.000526,
		0.001246,
		0.002503,
		0.004491,
		0.007416,
		0.011491,
		0.016933,
		0.023956,
		0.032763,
		0.043547,
		0.056483,
		0.07172,
		0.089384,
		0.109565,
		0.13232,
		0.157661,
		0.185558,
		0.21593,
		0.248646,
		0.283525,
		0.320332,
		0.358785,
		0.398556,
		0.43928,
		0.480563,
		0.521988,
		0.563133,
		0.603579,
		0.642924,
		0.680795,
		0.716858,
		0.75083,
		0.782486,
		0.81166,
		0.838252,
		0.862222,
		0.883593,
		0.902438,
		0.918878,
		0.933069,
		0.945195,
		0.955459,
		0.964069,
		0.971234,
		0.977154,
		0.982016,
		0.985988,
		0.989217,
		0.991829,
		0.993929,
		0.995602,
		0.996917,
		0.99793,
		0.998688,
		0.99923,
		0.999595,
		0.999818,
		0.999937,
		0.999986,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999909,
		0.999707,
		0.999268,
		0.998451,
		0.997073,
		0.994912,
		0.991705,
		0.987147,
		0.980889,
		0.972543,
		0.961681,
		0.947842,
		0.930531,
		0.909232,
		0.883417,
		0.852555,
		0.816138,
		0.773702,
		0.72486,
		0.669354,
		0.607109,
		0.538323,
		0.463572,
		0.383959,
		0.301315,
		0.218463,
		0.139573,
		0.070648,
		0.020173,
		3.7e-005,
		0.000183,
		0.00056,
		0.001323,
		0.002657,
		0.004764,
		0.00786,
		0.01217,
		0.01792,
		0.025331,
		0.034615,
		0.04597,
		0.059573,
		0.075576,
		0.094104,
		0.115245,
		0.139047,
		0.165515,
		0.194608,
		0.22623,
		0.260237,
		0.296425,
		0.334543,
		0.374285,
		0.415302,
		0.457205,
		0.499578,
		0.541984,
		0.58398,
		0.625132,
		0.665023,
		0.703269,
		0.73953,
		0.773521,
		0.805015,
		0.833854,
		0.859942,
		0.883254,
		0.903823,
		0.921738,
		0.937137,
		0.950194,
		0.961111,
		0.970107,
		0.977409,
		0.983245,
		0.987832,
		0.991373,
		0.994054,
		0.996039,
		0.997471,
		0.99847,
		0.999139,
		0.999561,
		0.999806,
		0.999932,
		0.999983,
		0.999998,
		1,
		0.999999,
		0.999986,
		0.999929,
		0.999771,
		0.999429,
		0.998793,
		0.997721,
		0.996042,
		0.993551,
		0.990011,
		0.985151,
		0.978669,
		0.97023,
		0.959469,
		0.945995,
		0.929393,
		0.909232,
		0.88507,
		0.856465,
		0.82299,
		0.78425,
		0.739908,
		0.689715,
		0.63355,
		0.571482,
		0.503836,
		0.431293,
		0.355015,
		0.276811,
		0.199356,
		0.12648,
		0.063557,
		0.018011,
		3.9e-005,
		0.000193,
		0.00059,
		0.001394,
		0.002796,
		0.005009,
		0.008259,
		0.012777,
		0.018799,
		0.026551,
		0.036251,
		0.0481,
		0.062277,
		0.078936,
		0.098195,
		0.12014,
		0.144812,
		0.172206,
		0.202269,
		0.234893,
		0.269916,
		0.307118,
		0.346229,
		0.386925,
		0.428835,
		0.471554,
		0.514646,
		0.557657,
		0.600132,
		0.641623,
		0.681705,
		0.719989,
		0.756135,
		0.789857,
		0.820937,
		0.849223,
		0.874633,
		0.897154,
		0.916836,
		0.933787,
		0.948161,
		0.960152,
		0.969981,
		0.977887,
		0.984115,
		0.988911,
		0.992508,
		0.995127,
		0.996969,
		0.99821,
		0.999005,
		0.999483,
		0.999749,
		0.999886,
		0.999952,
		0.999981,
		0.999983,
		0.999942,
		0.999818,
		0.999548,
		0.999046,
		0.9982,
		0.996876,
		0.994912,
		0.992122,
		0.988294,
		0.983187,
		0.976536,
		0.968053,
		0.957423,
		0.944313,
		0.928371,
		0.909232,
		0.886525,
		0.859876,
		0.828924,
		0.793332,
		0.752804,
		0.707107,
		0.656102,
		0.599783,
		0.538323,
		0.472143,
		0.40199,
		0.32905,
		0.255081,
		0.182603,
		0.11513,
		0.057479,
		0.016179,
		4.1e-005,
		0.000202,
		0.000617,
		0.001456,
		0.00292,
		0.005226,
		0.008609,
		0.013309,
		0.019564,
		0.027608,
		0.037661,
		0.049927,
		0.064584,
		0.081783,
		0.101641,
		0.124235,
		0.1496,
		0.177722,
		0.208535,
		0.241918,
		0.277694,
		0.31563,
		0.355436,
		0.396774,
		0.439258,
		0.482466,
		0.52595,
		0.569245,
		0.611886,
		0.653418,
		0.693416,
		0.73149,
		0.767303,
		0.800578,
		0.831105,
		0.858746,
		0.883435,
		0.905174,
		0.924032,
		0.940135,
		0.953657,
		0.96481,
		0.973833,
		0.980982,
		0.986518,
		0.990698,
		0.993768,
		0.995957,
		0.997467,
		0.998474,
		0.999123,
		0.999523,
		0.999749,
		0.999841,
		0.999806,
		0.99962,
		0.99923,
		0.998558,
		0.997501,
		0.995933,
		0.993706,
		0.99065,
		0.986574,
		0.981267,
		0.974494,
		0.966004,
		0.955526,
		0.942774,
		0.927446,
		0.909232,
		0.887815,
		0.862877,
		0.834112,
		0.801227,
		0.763962,
		0.722107,
		0.675515,
		0.624138,
		0.568056,
		0.507526,
		0.443032,
		0.375363,
		0.3057,
		0.23574,
		0.167843,
		0.105229,
		0.05223,
		0.014613,
		4.3e-005,
		0.00021,
		0.00064,
		0.001511,
		0.003027,
		0.005413,
		0.00891,
		0.013763,
		0.020214,
		0.0285,
		0.038844,
		0.051448,
		0.06649,
		0.084118,
		0.104444,
		0.127539,
		0.153429,
		0.18209,
		0.213445,
		0.247362,
		0.283651,
		0.322066,
		0.362304,
		0.404013,
		0.446798,
		0.490224,
		0.533837,
		0.577165,
		0.61974,
		0.661108,
		0.700845,
		0.738568,
		0.773948,
		0.80672,
		0.836688,
		0.863728,
		0.887791,
		0.908898,
		0.927134,
		0.942641,
		0.955611,
		0.966268,
		0.974863,
		0.981661,
		0.986925,
		0.990914,
		0.993865,
		0.99599,
		0.997467,
		0.998438,
		0.999005,
		0.99923,
		0.999139,
		0.998721,
		0.99793,
		0.996693,
		0.994906,
		0.992443,
		0.989153,
		0.984868,
		0.9794,
		0.972543,
		0.964076,
		0.953766,
		0.941362,
		0.926607,
		0.909232,
		0.888967,
		0.865538,
		0.838682,
		0.808146,
		0.773702,
		0.735156,
		0.692366,
		0.645261,
		0.593863,
		0.538323,
		0.478955,
		0.416289,
		0.351133,
		0.28465,
		0.218463,
		0.154777,
		0.096544,
		0.047667,
		0.013263,
		4.4e-005,
		0.000216,
		0.000661,
		0.001557,
		0.003117,
		0.005571,
		0.009162,
		0.014139,
		0.020748,
		0.029228,
		0.039801,
		0.052669,
		0.068007,
		0.085958,
		0.10663,
		0.130086,
		0.156346,
		0.185375,
		0.217088,
		0.251342,
		0.287937,
		0.326616,
		0.367068,
		0.408933,
		0.451806,
		0.495252,
		0.53881,
		0.58201,
		0.624385,
		0.665486,
		0.704898,
		0.742246,
		0.777215,
		0.809553,
		0.839077,
		0.865681,
		0.889329,
		0.910055,
		0.927957,
		0.943187,
		0.95594,
		0.966443,
		0.974941,
		0.981687,
		0.986925,
		0.990884,
		0.993768,
		0.995751,
		0.996969,
		0.997522,
		0.997471,
		0.996836,
		0.995602,
		0.993717,
		0.991097,
		0.987631,
		0.983181,
		0.97759,
		0.970681,
		0.962262,
		0.952128,
		0.940063,
		0.925841,
		0.909232,
		0.890001,
		0.867913,
		0.842738,
		0.814257,
		0.782268,
		0.746597,
		0.707107,
		0.663715,
		0.61641,
		0.565272,
		0.510503,
		0.452462,
		0.391707,
		0.32905,
		0.265625,
		0.202976,
		0.14316,
		0.088884,
		0.043674,
		0.012092,
		4.5e-005,
		0.000222,
		0.000677,
		0.001596,
		0.003191,
		0.005698,
		0.009364,
		0.014439,
		0.021172,
		0.029799,
		0.040544,
		0.053606,
		0.069157,
		0.087336,
		0.108244,
		0.13194,
		0.158434,
		0.187686,
		0.219602,
		0.254032,
		0.290769,
		0.329549,
		0.370055,
		0.411925,
		0.454751,
		0.498098,
		0.541506,
		0.584512,
		0.626654,
		0.667493,
		0.706621,
		0.743679,
		0.778361,
		0.810427,
		0.839707,
		0.866103,
		0.889586,
		0.910193,
		0.928019,
		0.943208,
		0.95594,
		0.966419,
		0.974863,
		0.981492,
		0.986518,
		0.990133,
		0.992508,
		0.993781,
		0.994054,
		0.993394,
		0.991829,
		0.989349,
		0.985907,
		0.981426,
		0.975798,
		0.96889,
		0.960548,
		0.9506,
		0.938862,
		0.92514,
		0.909232,
		0.890936,
		0.870045,
		0.846361,
		0.819691,
		0.789856,
		0.756699,
		0.720092,
		0.679948,
		0.636233,
		0.588982,
		0.538323,
		0.484498,
		0.427896,
		0.369091,
		0.308888,
		0.248386,
		0.189047,
		0.132788,
		0.082095,
		0.040162,
		0.011069,
		4.6e-005,
		0.000227,
		0.000691,
		0.001626,
		0.00325,
		0.005798,
		0.009521,
		0.014669,
		0.021491,
		0.030225,
		0.041091,
		0.054285,
		0.069976,
		0.0883,
		0.109353,
		0.133186,
		0.159806,
		0.189167,
		0.22117,
		0.25566,
		0.292425,
		0.331199,
		0.371666,
		0.413461,
		0.456182,
		0.499394,
		0.542647,
		0.585481,
		0.627445,
		0.668108,
		0.707073,
		0.743987,
		0.778551,
		0.810531,
		0.839755,
		0.86612,
		0.889586,
		0.910174,
		0.927957,
		0.943052,
		0.955611,
		0.965809,
		0.973833,
		0.979875,
		0.984115,
		0.986721,
		0.987832,
		0.987562,
		0.985988,
		0.983155,
		0.979067,
		0.973698,
		0.966984,
		0.958834,
		0.94913,
		0.937734,
		0.92449,
		0.909231,
		0.891783,
		0.87197,
		0.849616,
		0.824552,
		0.796619,
		0.765677,
		0.731605,
		0.694316,
		0.653763,
		0.609954,
		0.562964,
		0.512956,
		0.460201,
		0.405109,
		0.34826,
		0.290446,
		0.232725,
		0.176478,
		0.123492,
		0.076051,
		0.037057,
		0.010171,
		4.7e-005,
		0.00023,
		0.000701,
		0.001649,
		0.003294,
		0.005872,
		0.009635,
		0.014835,
		0.021718,
		0.030522,
		0.041464,
		0.054739,
		0.070513,
		0.088916,
		0.110041,
		0.133936,
		0.160605,
		0.189997,
		0.222012,
		0.256491,
		0.293225,
		0.331948,
		0.372344,
		0.414053,
		0.456678,
		0.49979,
		0.542944,
		0.585688,
		0.627576,
		0.668181,
		0.707107,
		0.743999,
		0.778551,
		0.810517,
		0.839707,
		0.865999,
		0.889329,
		0.909691,
		0.927134,
		0.941747,
		0.953657,
		0.963014,
		0.969981,
		0.974726,
		0.977409,
		0.978178,
		0.977154,
		0.974436,
		0.970086,
		0.964136,
		0.956583,
		0.94739,
		0.936491,
		0.923796,
		0.909189,
		0.892541,
		0.873711,
		0.852554,
		0.828924,
		0.802682,
		0.773702,
		0.741872,
		0.707107,
		0.669354,
		0.628601,
		0.584888,
		0.538323,
		0.489096,
		0.437501,
		0.383959,
		0.32905,
		0.273547,
		0.218463,
		0.165103,
		0.11513,
		0.070648,
		0.034297,
		0.009378,
		4.8e-005,
		0.000233,
		0.000709,
		0.001666,
		0.003325,
		0.005924,
		0.009713,
		0.014945,
		0.021866,
		0.03071,
		0.041695,
		0.055011,
		0.070823,
		0.089259,
		0.110408,
		0.134318,
		0.160989,
		0.190372,
		0.222364,
		0.25681,
		0.293501,
		0.332175,
		0.37252,
		0.414179,
		0.45676,
		0.499836,
		0.542966,
		0.585696,
		0.627576,
		0.668171,
		0.707073,
		0.743911,
		0.778361,
		0.810153,
		0.839077,
		0.864986,
		0.887791,
		0.907465,
		0.924032,
		0.937562,
		0.948161,
		0.955961,
		0.961111,
		0.963764,
		0.964069,
		0.962163,
		0.958161,
		0.952154,
		0.9442,
		0.93433,
		0.92254,
		0.908799,
		0.893048,
		0.875207,
		0.855181,
		0.832862,
		0.808142,
		0.780912,
		0.751076,
		0.718555,
		0.683291,
		0.645261,
		0.604481,
		0.561021,
		0.515011,
		0.466665,
		0.416289,
		0.364313,
		0.31131,
		0.258031,
		0.205443,
		0.154777,
		0.107582,
		0.065798,
		0.031834,
		0.008674,
		4.8e-005,
		0.000235,
		0.000714,
		0.001677,
		0.003345,
		0.005956,
		0.009761,
		0.015011,
		0.021951,
		0.030815,
		0.041817,
		0.055149,
		0.070972,
		0.089413,
		0.110561,
		0.134464,
		0.161121,
		0.190485,
		0.222455,
		0.256878,
		0.293546,
		0.332201,
		0.372533,
		0.414184,
		0.45676,
		0.49983,
		0.542944,
		0.585637,
		0.627445,
		0.667916,
		0.706621,
		0.74317,
		0.777215,
		0.808465,
		0.836688,
		0.861713,
		0.883435,
		0.901806,
		0.916836,
		0.928581,
		0.937137,
		0.942628,
		0.945195,
		0.944988,
		0.942154,
		0.936827,
		0.929125,
		0.919139,
		0.906936,
		0.892551,
		0.875992,
		0.857243,
		0.836264,
		0.813001,
		0.787387,
		0.759357,
		0.728847,
		0.695809,
		0.660215,
		0.622067,
		0.581406,
		0.538323,
		0.492972,
		0.445578,
		0.39646,
		0.346048,
		0.294906,
		0.243759,
		0.193529,
		0.145377,
		0.100747,
		0.061429,
		0.029626,
		0.008046,
		4.8e-005,
		0.000236,
		0.000717,
		0.001684,
		0.003356,
		0.005974,
		0.009787,
		0.015044,
		0.021992,
		0.030862,
		0.041868,
		0.055202,
		0.071023,
		0.089459,
		0.1106,
		0.134494,
		0.161143,
		0.190498,
		0.222461,
		0.25688,
		0.293546,
		0.332198,
		0.37252,
		0.414148,
		0.456678,
		0.499666,
		0.542647,
		0.585137,
		0.626654,
		0.666724,
		0.704898,
		0.740761,
		0.773948,
		0.804146,
		0.831105,
		0.85464,
		0.874633,
		0.891026,
		0.903823,
		0.913075,
		0.918878,
		0.921354,
		0.92065,
		0.916918,
		0.91031,
		0.900971,
		0.889025,
		0.874577,
		0.857704,
		0.83846,
		0.816874,
		0.792954,
		0.766692,
		0.738071,
		0.707073,
		0.673685,
		0.637912,
		0.599782,
		0.559361,
		0.516759,
		0.472143,
		0.425753,
		0.377913,
		0.32905,
		0.279714,
		0.230606,
		0.182603,
		0.136799,
		0.094539,
		0.057479,
		0.027641,
		0.007484,
		4.9e-005,
		0.000236,
		0.000719,
		0.001687,
		0.003362,
		0.005982,
		0.009797,
		0.015057,
		0.022005,
		0.030876,
		0.041881,
		0.055213,
		0.071031,
		0.089465,
		0.110603,
		0.134495,
		0.161143,
		0.190496,
		0.222455,
		0.256861,
		0.293501,
		0.332104,
		0.372344,
		0.413844,
		0.456182,
		0.498898,
		0.541506,
		0.583504,
		0.624385,
		0.663655,
		0.700845,
		0.735522,
		0.767303,
		0.795861,
		0.820937,
		0.842338,
		0.859942,
		0.873692,
		0.883593,
		0.889702,
		0.892122,
		0.890985,
		0.886448,
		0.878677,
		0.867837,
		0.854087,
		0.837568,
		0.818404,
		0.796695,
		0.772521,
		0.745942,
		0.717001,
		0.685735,
		0.652181,
		0.616381,
		0.578397,
		0.53832,
		0.496281,
		0.452462,
		0.407111,
		0.360554,
		0.313213,
		0.265625,
		0.218463,
		0.172561,
		0.128949,
		0.088884,
		0.053897,
		0.025848,
		0.006979,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041882,
		0.055213,
		0.071031,
		0.089464,
		0.1106,
		0.134487,
		0.161121,
		0.19045,
		0.222364,
		0.256697,
		0.293225,
		0.331662,
		0.371666,
		0.412843,
		0.454751,
		0.496909,
		0.53881,
		0.579929,
		0.61974,
		0.657731,
		0.693416,
		0.726348,
		0.756135,
		0.782445,
		0.805015,
		0.823657,
		0.838252,
		0.848753,
		0.855177,
		0.857596,
		0.856127,
		0.850921,
		0.842154,
		0.830009,
		0.814675,
		0.796331,
		0.775143,
		0.751262,
		0.724819,
		0.695929,
		0.664693,
		0.631206,
		0.595562,
		0.557864,
		0.518237,
		0.476837,
		0.433863,
		0.389572,
		0.344293,
		0.298442,
		0.25254,
		0.207232,
		0.163312,
		0.121749,
		0.083718,
		0.050639,
		0.024223,
		0.006524,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041881,
		0.05521,
		0.071023,
		0.089445,
		0.110561,
		0.134412,
		0.160989,
		0.190229,
		0.222012,
		0.256157,
		0.292425,
		0.330511,
		0.370055,
		0.410642,
		0.451806,
		0.493048,
		0.533837,
		0.57363,
		0.611886,
		0.648076,
		0.681705,
		0.712321,
		0.73953,
		0.763003,
		0.782486,
		0.7978,
		0.808844,
		0.815588,
		0.818069,
		0.816383,
		0.810671,
		0.801108,
		0.787895,
		0.771243,
		0.751367,
		0.728476,
		0.702768,
		0.674429,
		0.643631,
		0.610534,
		0.575293,
		0.538067,
		0.499021,
		0.458348,
		0.416269,
		0.373058,
		0.329048,
		0.28465,
		0.24037,
		0.196828,
		0.154777,
		0.11513,
		0.078987,
		0.047667,
		0.022747,
		0.006111,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022005,
		0.030871,
		0.041868,
		0.055183,
		0.070972,
		0.089354,
		0.110408,
		0.134165,
		0.160605,
		0.189652,
		0.22117,
		0.254963,
		0.290769,
		0.328265,
		0.367068,
		0.406741,
		0.446798,
		0.486716,
		0.52595,
		0.56394,
		0.600132,
		0.633993,
		0.665023,
		0.692773,
		0.716858,
		0.736961,
		0.752845,
		0.764353,
		0.771408,
		0.774009,
		0.772221,
		0.766169,
		0.756027,
		0.742003,
		0.724327,
		0.703245,
		0.679005,
		0.651854,
		0.622029,
		0.589761,
		0.555271,
		0.518777,
		0.480503,
		0.440684,
		0.399576,
		0.357477,
		0.314732,
		0.271752,
		0.229034,
		0.187172,
		0.146884,
		0.109032,
		0.074644,
		0.044948,
		0.021402,
		0.005737,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005984,
		0.009797,
		0.015052,
		0.021992,
		0.030844,
		0.041817,
		0.055094,
		0.070823,
		0.089116,
		0.110041,
		0.133616,
		0.159806,
		0.18852,
		0.219602,
		0.252835,
		0.287937,
		0.324562,
		0.362304,
		0.400704,
		0.439258,
		0.477426,
		0.514646,
		0.55035,
		0.58398,
		0.615002,
		0.642924,
		0.667309,
		0.687787,
		0.704067,
		0.715937,
		0.723272,
		0.726028,
		0.724239,
		0.71801,
		0.707502,
		0.692926,
		0.674529,
		0.652581,
		0.627366,
		0.599175,
		0.568298,
		0.535021,
		0.499627,
		0.462401,
		0.423631,
		0.383622,
		0.342703,
		0.301242,
		0.259662,
		0.218452,
		0.178194,
		0.139572,
		0.103402,
		0.070648,
		0.042454,
		0.020173,
		0.005396,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003362,
		0.005979,
		0.009787,
		0.015031,
		0.021951,
		0.030771,
		0.041695,
		0.054895,
		0.070513,
		0.088647,
		0.109353,
		0.132632,
		0.158434,
		0.186644,
		0.217088,
		0.249525,
		0.283651,
		0.319097,
		0.355436,
		0.392189,
		0.428835,
		0.464821,
		0.499578,
		0.532534,
		0.563133,
		0.590853,
		0.615215,
		0.635805,
		0.652282,
		0.664385,
		0.67194,
		0.674863,
		0.673152,
		0.666884,
		0.656209,
		0.641333,
		0.622512,
		0.600037,
		0.574223,
		0.545402,
		0.513912,
		0.480096,
		0.444298,
		0.406867,
		0.368158,
		0.328548,
		0.288436,
		0.248267,
		0.208539,
		0.169822,
		0.132781,
		0.098192,
		0.066962,
		0.040162,
		0.019046,
		0.005084,
		4.9e-005,
		0.000237,
		0.000719,
		0.001686,
		0.003356,
		0.005967,
		0.009761,
		0.014983,
		0.021866,
		0.030628,
		0.041464,
		0.054538,
		0.069976,
		0.087866,
		0.108244,
		0.131095,
		0.156346,
		0.183862,
		0.213445,
		0.244832,
		0.277694,
		0.311642,
		0.346229,
		0.38096,
		0.415302,
		0.448693,
		0.480563,
		0.510346,
		0.537498,
		0.561514,
		0.581942,
		0.598399,
		0.610581,
		0.618271,
		0.621342,
		0.619756,
		0.613563,
		0.602893,
		0.587947,
		0.568982,
		0.546305,
		0.520259,
		0.491211,
		0.459545,
		0.425657,
		0.38995,
		0.352837,
		0.314743,
		0.276112,
		0.237419,
		0.199181,
		0.161974,
		0.12645,
		0.093355,
		0.063554,
		0.03805,
		0.018011,
		0.004799,
		4.9e-005,
		0.000236,
		0.000717,
		0.001681,
		0.003345,
		0.005942,
		0.009713,
		0.014896,
		0.021718,
		0.030388,
		0.041091,
		0.053976,
		0.069157,
		0.086702,
		0.10663,
		0.128904,
		0.153429,
		0.180046,
		0.208535,
		0.238608,
		0.269916,
		0.302048,
		0.334543,
		0.366892,
		0.398556,
		0.428973,
		0.457578,
		0.48382,
		0.507174,
		0.527165,
		0.543376,
		0.555467,
		0.563181,
		0.566352,
		0.564909,
		0.558873,
		0.548352,
		0.533535,
		0.514679,
		0.492103,
		0.46617,
		0.437279,
		0.405858,
		0.372351,
		0.337219,
		0.300936,
		0.263995,
		0.22691,
		0.190228,
		0.154541,
		0.120501,
		0.088841,
		0.06039,
		0.036097,
		0.017057,
		0.004537,
		4.8e-005,
		0.000235,
		0.000714,
		0.001672,
		0.003325,
		0.005901,
		0.009635,
		0.014759,
		0.021491,
		0.03003,
		0.040544,
		0.053172,
		0.068007,
		0.085099,
		0.104444,
		0.125984,
		0.1496,
		0.17511,
		0.202269,
		0.230768,
		0.260237,
		0.29025,
		0.320332,
		0.349969,
		0.37862,
		0.405731,
		0.430753,
		0.453154,
		0.472443,
		0.488182,
		0.5,
		0.507609,
		0.510811,
		0.5095,
		0.50367,
		0.493408,
		0.478887,
		0.460362,
		0.438158,
		0.412655,
		0.384281,
		0.3535,
		0.320804,
		0.286708,
		0.251747,
		0.216478,
		0.181483,
		0.147381,
		0.114839,
		0.084585,
		0.057429,
		0.034281,
		0.016175,
		0.004295,
		4.8e-005,
		0.000234,
		0.000709,
		0.001658,
		0.003294,
		0.005838,
		0.009521,
		0.014563,
		0.021172,
		0.029533,
		0.039801,
		0.052095,
		0.06649,
		0.083014,
		0.101641,
		0.122288,
		0.144812,
		0.169008,
		0.194608,
		0.221283,
		0.248646,
		0.276259,
		0.303641,
		0.330276,
		0.35563,
		0.379165,
		0.400354,
		0.4187,
		0.43375,
		0.445116,
		0.452485,
		0.455628,
		0.454416,
		0.448814,
		0.438887,
		0.424797,
		0.40679,
		0.385195,
		0.360406,
		0.332875,
		0.3031,
		0.271617,
		0.238991,
		0.205814,
		0.172707,
		0.140319,
		0.109339,
		0.080506,
		0.054623,
		0.032575,
		0.015352,
		0.004071,
		4.8e-005,
		0.000232,
		0.000701,
		0.001639,
		0.00325,
		0.005752,
		0.009364,
		0.014298,
		0.020748,
		0.028884,
		0.038844,
		0.050725,
		0.064584,
		0.080424,
		0.098195,
		0.117792,
		0.139047,
		0.161731,
		0.185558,
		0.21018,
		0.235198,
		0.260166,
		0.284601,
		0.307995,
		0.32983,
		0.34959,
		0.366781,
		0.38095,
		0.391695,
		0.398686,
		0.401677,
		0.400512,
		0.395137,
		0.385597,
		0.372038,
		0.354703,
		0.333919,
		0.310092,
		0.283695,
		0.255258,
		0.225354,
		0.194598,
		0.163637,
		0.133151,
		0.103854,
		0.076504,
		0.05191,
		0.030947,
		0.014576,
		0.003862,
		4.7e-005,
		0.000228,
		0.000691,
		0.001612,
		0.003191,
		0.005638,
		0.009162,
		0.01396,
		0.020214,
		0.028075,
		0.037661,
		0.049052,
		0.062277,
		0.077319,
		0.094104,
		0.112501,
		0.13232,
		0.153312,
		0.175172,
		0.197539,
		0.220006,
		0.242125,
		0.26342,
		0.283397,
		0.30156,
		0.317429,
		0.330552,
		0.340527,
		0.347016,
		0.349758,
		0.348584,
		0.343419,
		0.334294,
		0.321343,
		0.304801,
		0.284996,
		0.262345,
		0.237338,
		0.210532,
		0.182539,
		0.154017,
		0.125665,
		0.098223,
		0.072466,
		0.049219,
		0.029357,
		0.013828,
		0.003662,
		4.6e-005,
		0.000224,
		0.000677,
		0.001578,
		0.003117,
		0.005496,
		0.00891,
		0.013546,
		0.019564,
		0.0271,
		0.036251,
		0.047072,
		0.059573,
		0.073709,
		0.089384,
		0.106443,
		0.124676,
		0.143817,
		0.163544,
		0.183489,
		0.203242,
		0.22236,
		0.240381,
		0.256833,
		0.271253,
		0.283205,
		0.292291,
		0.298172,
		0.300582,
		0.299339,
		0.294361,
		0.285665,
		0.273376,
		0.257723,
		0.239036,
		0.21774,
		0.194342,
		0.169425,
		0.143636,
		0.117674,
		0.09229,
		0.068277,
		0.046472,
		0.027762,
		0.01309,
		0.003469,
		4.5e-005,
		0.000219,
		0.000661,
		0.001535,
		0.003027,
		0.005324,
		0.008609,
		0.013053,
		0.018799,
		0.025961,
		0.034615,
		0.044794,
		0.056483,
		0.069614,
		0.084068,
		0.099669,
		0.116188,
		0.133344,
		0.150806,
		0.168203,
		0.185128,
		0.201151,
		0.21583,
		0.228728,
		0.239425,
		0.247536,
		0.252728,
		0.254737,
		0.253379,
		0.248563,
		0.240301,
		0.228713,
		0.214025,
		0.196569,
		0.176778,
		0.155176,
		0.132372,
		0.109042,
		0.085931,
		0.063833,
		0.043596,
		0.026116,
		0.01234,
		0.003275,
		4.4e-005,
		0.000213,
		0.00064,
		0.001485,
		0.00292,
		0.005121,
		0.008259,
		0.012483,
		0.01792,
		0.024662,
		0.032763,
		0.042233,
		0.053032,
		0.065071,
		0.078208,
		0.092251,
		0.106953,
		0.122023,
		0.137128,
		0.151896,
		0.165934,
		0.178832,
		0.190179,
		0.19958,
		0.206667,
		0.211121,
		0.212684,
		0.211175,
		0.206504,
		0.198682,
		0.187825,
		0.174161,
		0.158028,
		0.139866,
		0.120219,
		0.099716,
		0.07907,
		0.05906,
		0.040531,
		0.02438,
		0.011559,
		0.003077,
		4.3e-005,
		0.000206,
		0.000617,
		0.001426,
		0.002796,
		0.00489,
		0.00786,
		0.011839,
		0.016933,
		0.023214,
		0.030711,
		0.039411,
		0.049254,
		0.060128,
		0.071873,
		0.084279,
		0.097092,
		0.110012,
		0.122709,
		0.134822,
		0.145975,
		0.155789,
		0.163895,
		0.169947,
		0.173644,
		0.174741,
		0.173066,
		0.168534,
		0.161159,
		0.151061,
		0.138472,
		0.123736,
		0.107308,
		0.089745,
		0.071706,
		0.053931,
		0.037241,
		0.022525,
		0.010731,
		0.002868,
		4.1e-005,
		0.000197,
		0.00059,
		0.00136,
		0.002657,
		0.004631,
		0.007416,
		0.011126,
		0.015848,
		0.021629,
		0.028478,
		0.03636,
		0.045192,
		0.054846,
		0.065143,
		0.075864,
		0.086747,
		0.097494,
		0.107782,
		0.117269,
		0.125608,
		0.132457,
		0.137498,
		0.140451,
		0.141088,
		0.139251,
		0.134865,
		0.127953,
		0.118641,
		0.107171,
		0.093895,
		0.079284,
		0.063915,
		0.048471,
		0.033727,
		0.020541,
		0.009847,
		0.002646,
		3.9e-005,
		0.000188,
		0.00056,
		0.001285,
		0.002503,
		0.004345,
		0.00693,
		0.010351,
		0.014673,
		0.019924,
		0.02609,
		0.033116,
		0.0409,
		0.049296,
		0.058116,
		0.067132,
		0.076083,
		0.084677,
		0.092608,
		0.099561,
		0.105226,
		0.109311,
		0.11156,
		0.111767,
		0.109792,
		0.105574,
		0.099147,
		0.090652,
		0.080339,
		0.068577,
		0.055852,
		0.042764,
		0.030023,
		0.018437,
		0.008906,
		0.00241,
		3.7e-005,
		0.000177,
		0.000526,
		0.001205,
		0.002336,
		0.004037,
		0.006407,
		0.009521,
		0.013424,
		0.01812,
		0.023578,
		0.029723,
		0.036436,
		0.04356,
		0.0509,
		0.058225,
		0.06528,
		0.071791,
		0.077474,
		0.082051,
		0.08526,
		0.086869,
		0.086698,
		0.084624,
		0.080608,
		0.074699,
		0.067052,
		0.057933,
		0.047728,
		0.036943,
		0.026203,
		0.016247,
		0.007919,
		0.002161,
		3.5e-005,
		0.000166,
		0.000491,
		0.001117,
		0.002157,
		0.003708,
		0.005853,
		0.008648,
		0.012115,
		0.016241,
		0.020976,
		0.026229,
		0.031869,
		0.03773,
		0.043615,
		0.049299,
		0.054541,
		0.059088,
		0.062693,
		0.065122,
		0.066174,
		0.06569,
		0.063573,
		0.059802,
		0.054447,
		0.04768,
		0.039786,
		0.031171,
		0.022365,
		0.014018,
		0.006903,
		0.001902,
		3.3e-005,
		0.000154,
		0.000452,
		0.001025,
		0.001968,
		0.003363,
		0.005276,
		0.007741,
		0.010764,
		0.014314,
		0.018324,
		0.022689,
		0.027273,
		0.031906,
		0.036394,
		0.040526,
		0.04408,
		0.046839,
		0.0486,
		0.049186,
		0.048465,
		0.046366,
		0.04289,
		0.038126,
		0.032269,
		0.025625,
		0.018622,
		0.011815,
		0.005885,
		0.001639,
		3e-005,
		0.000141,
		0.000412,
		0.000929,
		0.001771,
		0.003007,
		0.004682,
		0.006814,
		0.009391,
		0.012367,
		0.015662,
		0.019164,
		0.02273,
		0.026196,
		0.029379,
		0.032087,
		0.034131,
		0.035336,
		0.035553,
		0.034677,
		0.03266,
		0.029527,
		0.025391,
		0.020468,
		0.015087,
		0.009702,
		0.004895,
		0.00138,
		2.7e-005,
		0.000127,
		0.000371,
		0.000829,
		0.00157,
		0.002644,
		0.00408,
		0.00588,
		0.008018,
		0.010433,
		0.013038,
		0.015716,
		0.018328,
		0.020718,
		0.022723,
		0.02418,
		0.024941,
		0.024886,
		0.023933,
		0.02206,
		0.019315,
		0.015835,
		0.011859,
		0.007742,
		0.003962,
		0.001132,
		2.4e-005,
		0.000114,
		0.000328,
		0.000728,
		0.001366,
		0.002279,
		0.003479,
		0.004955,
		0.006667,
		0.008547,
		0.010502,
		0.012416,
		0.01416,
		0.015596,
		0.016587,
		0.017011,
		0.016772,
		0.015814,
		0.014139,
		0.011818,
		0.009012,
		0.005983,
		0.003111,
		0.000902,
		2.2e-005,
		0.0001,
		0.000285,
		0.000627,
		0.001164,
		0.001918,
		0.00289,
		0.004055,
		0.005365,
		0.006746,
		0.008105,
		0.009337,
		0.010326,
		0.010961,
		0.011143,
		0.0108,
		0.009899,
		0.008464,
		0.006591,
		0.004461,
		0.002362,
		0.000697,
		1.9e-005,
		8.6e-005,
		0.000243,
		0.000527,
		0.000966,
		0.001568,
		0.002323,
		0.003197,
		0.004136,
		0.005068,
		0.005905,
		0.006556,
		0.00693,
		0.006952,
		0.00657,
		0.005775,
		0.00461,
		0.003192,
		0.001726,
		0.000519,
		1.6e-005,
		7.2e-005,
		0.000201,
		0.000431,
		0.000776,
		0.001235,
		0.00179,
		0.0024,
		0.003011,
		0.003555,
		0.00396,
		0.004154,
		0.004083,
		0.003714,
		0.003057,
		0.002176,
		0.001206,
		0.000371,
		1.3e-005,
		5.9e-005,
		0.000161,
		0.000339,
		0.000597,
		0.000926,
		0.001301,
		0.001682,
		0.002017,
		0.002251,
		0.002331,
		0.002217,
		0.001898,
		0.001398,
		0.000799,
		0.000253,
		1.1e-005,
		4.6e-005,
		0.000124,
		0.000254,
		0.000434,
		0.000649,
		0.000871,
		0.001064,
		0.001186,
		0.0012,
		0.001083,
		0.000834,
		0.000496,
		0.000162,
		3.5e-005,
		9e-005,
		0.000178,
		0.00029,
		0.00041,
		0.000512,
		0.000567,
		0.000551,
		0.000451,
		0.000283,
		9.7e-005,
		2.4e-005,
		6e-005,
		0.000112,
		0.00017,
		0.000218,
		0.000237,
		0.000213,
		0.000144,
		5.2e-005,
		1.5e-005,
		3.5e-005,
		6e-005,
		7.9e-005,
		8.2e-005,
		6.2e-005,
		2.5e-005,
		1.6e-005,
		2.2e-005,
		2.1e-005,
		1.6e-005,
		1.6e-005,
		2.1e-005,
		4.5e-005,
		4.9e-005,
		3.5e-005,
		1.7e-005,
		6.2e-005,
		0.00011,
		0.000121,
		9.8e-005,
		6e-005,
		2.7e-005,
		3.7e-005,
		0.000149,
		0.000237,
		0.000261,
		0.000227,
		0.00016,
		9e-005,
		3.7e-005,
		0.000116,
		0.000316,
		0.000464,
		0.000512,
		0.000466,
		0.000359,
		0.000234,
		0.000124,
		4.9e-005,
		1.2e-005,
		2.4e-005,
		0.000283,
		0.000608,
		0.000842,
		0.000927,
		0.000871,
		0.000715,
		0.000513,
		0.000317,
		0.000161,
		6.2e-005,
		1.5e-005,
		0.000127,
		0.000586,
		0.001083,
		0.001437,
		0.001579,
		0.001517,
		0.001301,
		0.001001,
		0.000685,
		0.000407,
		0.000201,
		7.5e-005,
		1.7e-005,
		0.000361,
		0.001091,
		0.001816,
		0.002331,
		0.002555,
		0.002495,
		0.002212,
		0.00179,
		0.001317,
		0.00087,
		0.000503,
		0.000243,
		8.9e-005,
		2e-005,
		6.4e-005,
		0.000799,
		0.001882,
		0.0029,
		0.003623,
		0.00396,
		0.003915,
		0.003559,
		0.002991,
		0.002323,
		0.001654,
		0.001064,
		0.000602,
		0.000285,
		0.000103,
		2.3e-005,
		0.000308,
		0.001528,
		0.003057,
		0.004443,
		0.005431,
		0.005917,
		0.005905,
		0.005473,
		0.004739,
		0.003836,
		0.00289,
		0.002008,
		0.001265,
		0.000703,
		0.000328,
		0.000117,
		2.6e-005,
		0.000826,
		0.002653,
		0.004732,
		0.00657,
		0.007888,
		0.008567,
		0.008611,
		0.008105,
		0.007186,
		0.006008,
		0.004727,
		0.003479,
		0.00237,
		0.001468,
		0.000804,
		0.000371,
		0.000131,
		2.9e-005,
		0.000131,
		0.001726,
		0.004295,
		0.007038,
		0.009423,
		0.011143,
		0.012067,
		0.012195,
		0.011623,
		0.010502,
		0.009012,
		0.007338,
		0.005648,
		0.00408,
		0.002735,
		0.001671,
		0.000904,
		0.000412,
		0.000144,
		3.1e-005,
		0.000604,
		0.003134,
		0.006591,
		0.010121,
		0.013156,
		0.01536,
		0.016587,
		0.016836,
		0.016209,
		0.014875,
		0.013038,
		0.010914,
		0.008703,
		0.00658,
		0.004682,
		0.003097,
		0.00187,
		0.001001,
		0.000452,
		0.000157,
		3.4e-005,
		0.00155,
		0.005191,
		0.009688,
		0.014139,
		0.017935,
		0.020712,
		0.02231,
		0.022723,
		0.022059,
		0.020506,
		0.018292,
		0.015662,
		0.012854,
		0.010079,
		0.007511,
		0.005276,
		0.003451,
		0.002063,
		0.001095,
		0.000491,
		0.000169,
		3.6e-005,
		0.000228,
		0.003111,
		0.008047,
		0.013746,
		0.019257,
		0.023933,
		0.027382,
		0.029425,
		0.030053,
		0.029379,
		0.027607,
		0.02499,
		0.021804,
		0.018324,
		0.014799,
		0.011443,
		0.008424,
		0.005853,
		0.003792,
		0.002247,
		0.001183,
		0.000526,
		0.00018,
		3.8e-005,
		0.001015,
		0.005442,
		0.011859,
		0.018926,
		0.025644,
		0.031327,
		0.035553,
		0.038125,
		0.039027,
		0.038377,
		0.036394,
		0.033356,
		0.029569,
		0.025346,
		0.020976,
		0.016717,
		0.012776,
		0.009307,
		0.006407,
		0.004116,
		0.002421,
		0.001266,
		0.00056,
		0.00019,
		4e-005,
		0.002517,
		0.008699,
		0.016787,
		0.025391,
		0.033468,
		0.040291,
		0.045407,
		0.0486,
		0.049843,
		0.049261,
		0.047085,
		0.043615,
		0.039191,
		0.034161,
		0.028856,
		0.023578,
		0.018579,
		0.014057,
		0.010148,
		0.00693,
		0.004419,
		0.002582,
		0.001342,
		0.00059,
		0.000199,
		4.2e-005,
		0.000349,
		0.004895,
		0.013042,
		0.022989,
		0.0333,
		0.04289,
		0.050989,
		0.057115,
		0.061029,
		0.062693,
		0.062232,
		0.05989,
		0.055989,
		0.0509,
		0.045007,
		0.038685,
		0.03228,
		0.02609,
		0.02036,
		0.015271,
		0.010938,
		0.007416,
		0.004698,
		0.002729,
		0.00141,
		0.000617,
		0.000208,
		4.4e-005,
		0.001509,
		0.008309,
		0.018622,
		0.03061,
		0.042798,
		0.054055,
		0.063573,
		0.070835,
		0.075577,
		0.077749,
		0.077474,
		0.075006,
		0.070687,
		0.064916,
		0.058116,
		0.050704,
		0.043071,
		0.035565,
		0.028478,
		0.022037,
		0.016402,
		0.011667,
		0.00786,
		0.004951,
		0.00286,
		0.001471,
		0.00064,
		0.000215,
		4.5e-005,
		0.003642,
		0.012909,
		0.02558,
		0.039786,
		0.054013,
		0.067091,
		0.07817,
		0.086698,
		0.092384,
		0.095162,
		0.095151,
		0.092608,
		0.087897,
		0.081444,
		0.073706,
		0.065143,
		0.056194,
		0.047255,
		0.038668,
		0.030711,
		0.023589,
		0.01744,
		0.012329,
		0.008259,
		0.005175,
		0.002975,
		0.001523,
		0.000661,
		0.000221,
		4.6e-005,
		0.000482,
		0.006903,
		0.018834,
		0.03404,
		0.050629,
		0.067052,
		0.082099,
		0.09488,
		0.104807,
		0.11156,
		0.115052,
		0.115392,
		0.112844,
		0.107782,
		0.10066,
		0.09197,
		0.082212,
		0.071873,
		0.061398,
		0.051181,
		0.041551,
		0.032763,
		0.025001,
		0.018373,
		0.012918,
		0.008609,
		0.005369,
		0.003074,
		0.001568,
		0.000677,
		0.000225,
		4.7e-005,
		0.002032,
		0.011432,
		0.026203,
		0.044105,
		0.063227,
		0.08199,
		0.099147,
		0.11377,
		0.12523,
		0.133176,
		0.137498,
		0.138292,
		0.135819,
		0.130466,
		0.122709,
		0.11307,
		0.102092,
		0.090306,
		0.078208,
		0.066246,
		0.0548,
		0.044179,
		0.034615,
		0.026261,
		0.019196,
		0.01343,
		0.00891,
		0.005534,
		0.003156,
		0.001604,
		0.000691,
		0.000229,
		4.7e-005,
		0.004798,
		0.01735,
		0.035114,
		0.055852,
		0.07764,
		0.098873,
		0.118271,
		0.134865,
		0.14799,
		0.15726,
		0.162535,
		0.163895,
		0.161595,
		0.156029,
		0.147689,
		0.137128,
		0.124925,
		0.111657,
		0.097872,
		0.084068,
		0.070681,
		0.058074,
		0.04653,
		0.036251,
		0.027359,
		0.019903,
		0.013864,
		0.009162,
		0.005669,
		0.003222,
		0.001633,
		0.000701,
		0.000232,
		4.8e-005,
		0.000611,
		0.008906,
		0.024756,
		0.045641,
		0.069331,
		0.093895,
		0.11771,
		0.139462,
		0.158149,
		0.173066,
		0.183787,
		0.190145,
		0.192194,
		0.190179,
		0.184495,
		0.175647,
		0.164213,
		0.150806,
		0.136047,
		0.120532,
		0.104814,
		0.089384,
		0.074658,
		0.060974,
		0.048585,
		0.037661,
		0.028292,
		0.020495,
		0.014221,
		0.009364,
		0.005776,
		0.003273,
		0.001654,
		0.000709,
		0.000234,
		4.8e-005,
		0.00253,
		0.014463,
		0.033727,
		0.057832,
		0.084562,
		0.111989,
		0.138472,
		0.162676,
		0.183562,
		0.200385,
		0.212684,
		0.220253,
		0.223122,
		0.221518,
		0.21583,
		0.206571,
		0.194338,
		0.179776,
		0.163544,
		0.146287,
		0.128609,
		0.111057,
		0.094104,
		0.078143,
		0.063481,
		0.050336,
		0.038844,
		0.029061,
		0.020973,
		0.014503,
		0.009521,
		0.005856,
		0.003311,
		0.001669,
		0.000714,
		0.000235,
		4.8e-005,
		0.005873,
		0.021549,
		0.044312,
		0.071706,
		0.101537,
		0.131883,
		0.161098,
		0.187825,
		0.210995,
		0.229828,
		0.243819,
		0.252728,
		0.256552,
		0.255494,
		0.249933,
		0.240381,
		0.227449,
		0.211808,
		0.194153,
		0.175172,
		0.155521,
		0.135801,
		0.116543,
		0.098195,
		0.081119,
		0.065587,
		0.051781,
		0.039801,
		0.029671,
		0.021344,
		0.014716,
		0.009635,
		0.005913,
		0.003336,
		0.001679,
		0.000717,
		0.000236,
		4.9e-005,
		0.000727,
		0.010731,
		0.030225,
		0.05654,
		0.087259,
		0.120219,
		0.15351,
		0.18549,
		0.214785,
		0.240301,
		0.261224,
		0.27701,
		0.287379,
		0.292291,
		0.291922,
		0.286631,
		0.276925,
		0.26342,
		0.246807,
		0.22781,
		0.207159,
		0.185558,
		0.163659,
		0.142051,
		0.121239,
		0.101641,
		0.083582,
		0.067296,
		0.052929,
		0.040544,
		0.030132,
		0.021616,
		0.014867,
		0.009713,
		0.00595,
		0.003351,
		0.001685,
		0.000719,
		0.000237,
		4.9e-005,
		0.002974,
		0.017177,
		0.040531,
		0.070417,
		0.104466,
		0.140548,
		0.176778,
		0.21152,
		0.243396,
		0.271292,
		0.294361,
		0.312024,
		0.323959,
		0.330088,
		0.330552,
		0.325684,
		0.315979,
		0.302051,
		0.284601,
		0.264379,
		0.242148,
		0.218656,
		0.194608,
		0.170644,
		0.147328,
		0.125134,
		0.104444,
		0.085543,
		0.068626,
		0.053799,
		0.041091,
		0.030459,
		0.021801,
		0.014965,
		0.009761,
		0.005971,
		0.00336,
		0.001687,
		0.000719,
		0.000237,
		4.9e-005,
		0.006823,
		0.025261,
		0.052485,
		0.085931,
		0.123278,
		0.162442,
		0.201568,
		0.239036,
		0.273471,
		0.303745,
		0.328987,
		0.348584,
		0.362172,
		0.369632,
		0.371068,
		0.366781,
		0.357244,
		0.343062,
		0.324939,
		0.303641,
		0.279959,
		0.254678,
		0.228551,
		0.202269,
		0.176453,
		0.151632,
		0.128245,
		0.10663,
		0.087033,
		0.069605,
		0.054418,
		0.041464,
		0.030672,
		0.021915,
		0.015022,
		0.009787,
		0.005981,
		0.003363,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.000831,
		0.01234,
		0.035019,
		0.066092,
		0.103053,
		0.143636,
		0.185806,
		0.227748,
		0.267868,
		0.304801,
		0.337415,
		0.364821,
		0.386375,
		0.401677,
		0.410564,
		0.4131,
		0.409549,
		0.400354,
		0.386103,
		0.367494,
		0.3453,
		0.320332,
		0.293406,
		0.265316,
		0.236802,
		0.208535,
		0.181102,
		0.154997,
		0.130612,
		0.108244,
		0.088095,
		0.070277,
		0.054822,
		0.041695,
		0.030795,
		0.021976,
		0.015049,
		0.009797,
		0.005984,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.003372,
		0.019574,
		0.046472,
		0.081344,
		0.121743,
		0.165466,
		0.210532,
		0.255173,
		0.297832,
		0.337163,
		0.372038,
		0.40156,
		0.425058,
		0.442101,
		0.452485,
		0.456229,
		0.453558,
		0.444878,
		0.430753,
		0.411863,
		0.388981,
		0.362928,
		0.334543,
		0.304652,
		0.274039,
		0.243424,
		0.213445,
		0.18465,
		0.157487,
		0.132304,
		0.109353,
		0.088791,
		0.070692,
		0.055057,
		0.041817,
		0.030854,
		0.022001,
		0.015058,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.007687,
		0.028563,
		0.059631,
		0.098223,
		0.141955,
		0.188689,
		0.236507,
		0.283695,
		0.32874,
		0.370327,
		0.407343,
		0.438887,
		0.464277,
		0.48305,
		0.494965,
		0.5,
		0.498335,
		0.490336,
		0.47653,
		0.457578,
		0.434238,
		0.407332,
		0.377712,
		0.346229,
		0.313702,
		0.280893,
		0.248488,
		0.217088,
		0.187192,
		0.159201,
		0.133415,
		0.110041,
		0.089194,
		0.070914,
		0.055169,
		0.041868,
		0.030874,
		0.022007,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.00093,
		0.013828,
		0.039335,
		0.074497,
		0.116705,
		0.163637,
		0.213225,
		0.263617,
		0.313165,
		0.360406,
		0.404065,
		0.443057,
		0.476489,
		0.50367,
		0.524119,
		0.537563,
		0.543936,
		0.543376,
		0.536205,
		0.522909,
		0.504116,
		0.480563,
		0.453061,
		0.422467,
		0.389643,
		0.355436,
		0.320642,
		0.285991,
		0.252127,
		0.219602,
		0.188866,
		0.160268,
		0.134061,
		0.110408,
		0.089387,
		0.071006,
		0.055207,
		0.041881,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005984,
		0.003363,
		0.001687,
		0.000719,
		0.000236,
		4.8e-005,
		0.003761,
		0.021838,
		0.05191,
		0.091069,
		0.136762,
		0.186735,
		0.238991,
		0.291752,
		0.343437,
		0.392648,
		0.438158,
		0.478918,
		0.514056,
		0.542884,
		0.564909,
		0.579832,
		0.587553,
		0.588164,
		0.581942,
		0.569328,
		0.550912,
		0.527398,
		0.499578,
		0.468297,
		0.434424,
		0.398819,
		0.362304,
		0.325642,
		0.289518,
		0.254529,
		0.22117,
		0.18984,
		0.160839,
		0.134371,
		0.110561,
		0.089454,
		0.071029,
		0.055213,
		0.041882,
		0.030877,
		0.022007,
		0.015058,
		0.009797,
		0.005981,
		0.00336,
		0.001685,
		0.000717,
		0.000235,
		4.8e-005,
		0.008551,
		0.031753,
		0.066304,
		0.109339,
		0.158365,
		0.211195,
		0.265908,
		0.320804,
		0.374377,
		0.425297,
		0.472399,
		0.514679,
		0.551297,
		0.581581,
		0.605034,
		0.621342,
		0.630373,
		0.632181,
		0.626996,
		0.615215,
		0.597379,
		0.574154,
		0.546301,
		0.514646,
		0.480048,
		0.443369,
		0.405447,
		0.367068,
		0.32895,
		0.291724,
		0.255931,
		0.222012,
		0.19031,
		0.161077,
		0.134478,
		0.1106,
		0.089464,
		0.071031,
		0.055213,
		0.041881,
		0.030874,
		0.022001,
		0.015049,
		0.009787,
		0.005971,
		0.003351,
		0.001679,
		0.000714,
		0.000234,
		4.8e-005,
		0.001034,
		0.015352,
		0.043608,
		0.082529,
		0.129299,
		0.181483,
		0.236966,
		0.293903,
		0.350675,
		0.405858,
		0.458198,
		0.5066,
		0.550115,
		0.587947,
		0.619449,
		0.644135,
		0.661683,
		0.67194,
		0.674927,
		0.67083,
		0.659998,
		0.642924,
		0.620228,
		0.59263,
		0.560924,
		0.52595,
		0.48856,
		0.449592,
		0.409844,
		0.370055,
		0.330886,
		0.292911,
		0.256608,
		0.222364,
		0.190471,
		0.161138,
		0.134495,
		0.110603,
		0.089464,
		0.071029,
		0.055207,
		0.041868,
		0.030854,
		0.021976,
		0.015022,
		0.009761,
		0.00595,
		0.003336,
		0.001669,
		0.000709,
		0.000232,
		4.7e-005,
		0.004181,
		0.024212,
		0.057429,
		0.100594,
		0.150937,
		0.206085,
		0.263995,
		0.322902,
		0.381271,
		0.437764,
		0.491211,
		0.540591,
		0.585024,
		0.623765,
		0.656209,
		0.681888,
		0.700485,
		0.711837,
		0.715937,
		0.712934,
		0.703131,
		0.686971,
		0.665023,
		0.637959,
		0.606532,
		0.571549,
		0.533837,
		0.494223,
		0.453504,
		0.412426,
		0.371666,
		0.331824,
		0.293409,
		0.256841,
		0.222455,
		0.190498,
		0.161143,
		0.134495,
		0.1106,
		0.089454,
		0.071006,
		0.055169,
		0.041817,
		0.030795,
		0.021915,
		0.014965,
		0.009713,
		0.005913,
		0.003311,
		0.001654,
		0.000701,
		0.000229,
		4.7e-005,
		0.009502,
		0.035173,
		0.07324,
		0.120501,
		0.174237,
		0.232137,
		0.292232,
		0.352837,
		0.412507,
		0.469991,
		0.524209,
		0.574223,
		0.619229,
		0.658546,
		0.691617,
		0.71801,
		0.737423,
		0.749691,
		0.754792,
		0.752845,
		0.744109,
		0.72898,
		0.707972,
		0.681705,
		0.650881,
		0.616261,
		0.578638,
		0.53881,
		0.497557,
		0.455619,
		0.413678,
		0.372344,
		0.332147,
		0.293537,
		0.25688,
		0.222461,
		0.190498,
		0.161138,
		0.134478,
		0.110561,
		0.089387,
		0.070914,
		0.055057,
		0.041695,
		0.030672,
		0.021801,
		0.014867,
		0.009635,
		0.005856,
		0.003273,
		0.001633,
		0.000691,
		0.000225,
		4.6e-005,
		0.001154,
		0.017057,
		0.048276,
		0.091062,
		0.142252,
		0.199181,
		0.259602,
		0.321622,
		0.38364,
		0.444298,
		0.502441,
		0.557079,
		0.607367,
		0.652581,
		0.692115,
		0.725472,
		0.752265,
		0.772221,
		0.785187,
		0.791133,
		0.790158,
		0.782486,
		0.768466,
		0.748558,
		0.723327,
		0.693416,
		0.65953,
		0.622414,
		0.582824,
		0.541506,
		0.499179,
		0.456511,
		0.414112,
		0.37252,
		0.332201,
		0.293547,
		0.25688,
		0.222455,
		0.190471,
		0.161077,
		0.134371,
		0.110408,
		0.089194,
		0.070692,
		0.054822,
		0.041464,
		0.030459,
		0.021616,
		0.014716,
		0.009521,
		0.005776,
		0.003222,
		0.001604,
		0.000677,
		0.000221,
		4.5e-005,
		0.004665,
		0.026901,
		0.063554,
		0.110907,
		0.165839,
		0.225742,
		0.288436,
		0.352107,
		0.415238,
		0.476563,
		0.535021,
		0.589719,
		0.639906,
		0.684951,
		0.724327,
		0.757607,
		0.784462,
		0.80466,
		0.818069,
		0.824668,
		0.824541,
		0.817888,
		0.805015,
		0.786334,
		0.762347,
		0.733637,
		0.700845,
		0.664653,
		0.62576,
		0.584866,
		0.542647,
		0.499742,
		0.456743,
		0.414183,
		0.372533,
		0.332201,
		0.293537,
		0.256841,
		0.222364,
		0.19031,
		0.160839,
		0.134061,
		0.110041,
		0.088791,
		0.070277,
		0.054418,
		0.041091,
		0.030132,
		0.021344,
		0.014503,
		0.009364,
		0.005669,
		0.003156,
		0.001568,
		0.000661,
		0.000215,
		4.4e-005,
		0.010606,
		0.039085,
		0.081038,
		0.132781,
		0.191247,
		0.253886,
		0.31859,
		0.383622,
		0.447554,
		0.509214,
		0.567636,
		0.622029,
		0.671739,
		0.71623,
		0.755064,
		0.787895,
		0.814459,
		0.834575,
		0.84815,
		0.855177,
		0.855742,
		0.850026,
		0.838304,
		0.820937,
		0.798372,
		0.771123,
		0.739762,
		0.704898,
		0.667159,
		0.627179,
		0.585577,
		0.542944,
		0.499835,
		0.456761,
		0.414183,
		0.37252,
		0.332147,
		0.293409,
		0.256608,
		0.222012,
		0.18984,
		0.160268,
		0.133415,
		0.109353,
		0.088095,
		0.069605,
		0.053799,
		0.040544,
		0.029671,
		0.020973,
		0.014221,
		0.009162,
		0.005534,
		0.003074,
		0.001523,
		0.00064,
		0.000208,
		4.2e-005,
		0.001294,
		0.019046,
		0.053654,
		0.100747,
		0.156684,
		0.218452,
		0.283568,
		0.349999,
		0.416091,
		0.480503,
		0.542159,
		0.600193,
		0.653913,
		0.702768,
		0.746323,
		0.784233,
		0.81624,
		0.842154,
		0.861858,
		0.875305,
		0.882517,
		0.883593,
		0.878705,
		0.868102,
		0.852106,
		0.831105,
		0.805547,
		0.775926,
		0.742769,
		0.706621,
		0.668034,
		0.62755,
		0.585694,
		0.542966,
		0.499835,
		0.456743,
		0.414112,
		0.372344,
		0.331824,
		0.292911,
		0.255931,
		0.22117,
		0.188866,
		0.159201,
		0.132304,
		0.108244,
		0.087033,
		0.068626,
		0.052929,
		0.039801,
		0.029061,
		0.020495,
		0.013864,
		0.00891,
		0.005369,
		0.002975,
		0.001471,
		0.000617,
		0.000199,
		4e-005,
		0.005236,
		0.030049,
		0.070648,
		0.122695,
		0.182602,
		0.247418,
		0.314732,
		0.38259,
		0.449423,
		0.513985,
		0.575293,
		0.632584,
		0.685268,
		0.732899,
		0.775143,
		0.811757,
		0.842572,
		0.867484,
		0.886448,
		0.899472,
		0.90662,
		0.908012,
		0.903823,
		0.894285,
		0.879686,
		0.86036,
		0.836688,
		0.80908,
		0.777975,
		0.743821,
		0.707073,
		0.668179,
		0.627578,
		0.585694,
		0.542944,
		0.499742,
		0.456511,
		0.413678,
		0.371666,
		0.330886,
		0.291724,
		0.254529,
		0.219602,
		0.187192,
		0.157487,
		0.130612,
		0.10663,
		0.085543,
		0.067296,
		0.051781,
		0.038844,
		0.028292,
		0.019903,
		0.01343,
		0.008609,
		0.005175,
		0.00286,
		0.00141,
		0.00059,
		0.00019,
		3.8e-005,
		0.011912,
		0.043674,
		0.090096,
		0.146884,
		0.210511,
		0.278094,
		0.347303,
		0.416269,
		0.483514,
		0.547885,
		0.608498,
		0.664693,
		0.715985,
		0.762033,
		0.802607,
		0.837568,
		0.866844,
		0.890421,
		0.908331,
		0.92065,
		0.92749,
		0.929002,
		0.925377,
		0.916836,
		0.903638,
		0.886069,
		0.864439,
		0.839077,
		0.810321,
		0.778514,
		0.743996,
		0.707107,
		0.668179,
		0.62755,
		0.585577,
		0.542647,
		0.499179,
		0.455619,
		0.412426,
		0.370055,
		0.32895,
		0.289518,
		0.252127,
		0.217088,
		0.18465,
		0.154997,
		0.128245,
		0.104444,
		0.083582,
		0.065587,
		0.050336,
		0.037661,
		0.027359,
		0.019196,
		0.012918,
		0.008259,
		0.004951,
		0.002729,
		0.001342,
		0.00056,
		0.00018,
		3.6e-005,
		0.001462,
		0.021402,
		0.059973,
		0.112019,
		0.173305,
		0.24037,
		0.310414,
		0.381192,
		0.450927,
		0.518237,
		0.582068,
		0.641638,
		0.696388,
		0.745942,
		0.790063,
		0.828631,
		0.861608,
		0.889025,
		0.910961,
		0.927533,
		0.938887,
		0.945195,
		0.946649,
		0.943456,
		0.935839,
		0.924032,
		0.908275,
		0.88881,
		0.865876,
		0.839707,
		0.810528,
		0.778554,
		0.743996,
		0.707073,
		0.668034,
		0.627179,
		0.584866,
		0.541506,
		0.497557,
		0.453504,
		0.409844,
		0.367068,
		0.325642,
		0.285991,
		0.248488,
		0.213445,
		0.181102,
		0.151632,
		0.125134,
		0.101641,
		0.081119,
		0.063481,
		0.048585,
		0.036251,
		0.026261,
		0.018373,
		0.012329,
		0.00786,
		0.004698,
		0.002582,
		0.001266,
		0.000526,
		0.000169,
		3.4e-005,
		0.00592,
		0.033782,
		0.078987,
		0.136427,
		0.201932,
		0.272124,
		0.344293,
		0.416289,
		0.486434,
		0.553448,
		0.616381,
		0.674555,
		0.727518,
		0.775,
		0.816874,
		0.853125,
		0.883821,
		0.909096,
		0.929125,
		0.944112,
		0.954283,
		0.95987,
		0.961111,
		0.958238,
		0.951479,
		0.941046,
		0.927134,
		0.909914,
		0.889535,
		0.866116,
		0.839755,
		0.810528,
		0.778514,
		0.743821,
		0.706621,
		0.667159,
		0.62576,
		0.582824,
		0.53881,
		0.494223,
		0.449592,
		0.405447,
		0.362304,
		0.320642,
		0.280893,
		0.243424,
		0.208535,
		0.176453,
		0.147328,
		0.121239,
		0.098195,
		0.078143,
		0.060974,
		0.04653,
		0.034615,
		0.025001,
		0.01744,
		0.011667,
		0.007416,
		0.004419,
		0.002421,
		0.001183,
		0.000491,
		0.000157,
		3.1e-005,
		0.013475,
		0.049119,
		0.100747,
		0.163312,
		0.232725,
		0.3057,
		0.379627,
		0.452462,
		0.522637,
		0.588979,
		0.650647,
		0.707073,
		0.757914,
		0.803008,
		0.842337,
		0.875992,
		0.904151,
		0.927046,
		0.944951,
		0.958161,
		0.966979,
		0.971702,
		0.972616,
		0.969981,
		0.964028,
		0.95495,
		0.942893,
		0.927957,
		0.910189,
		0.88959,
		0.866116,
		0.839707,
		0.810321,
		0.777975,
		0.742769,
		0.704898,
		0.664653,
		0.622414,
		0.578638,
		0.533837,
		0.48856,
		0.443369,
		0.398819,
		0.355436,
		0.313702,
		0.274039,
		0.236802,
		0.202269,
		0.170644,
		0.142051,
		0.116543,
		0.094104,
		0.074658,
		0.058074,
		0.044179,
		0.032763,
		0.023589,
		0.016402,
		0.010938,
		0.00693,
		0.004116,
		0.002247,
		0.001095,
		0.000452,
		0.000144,
		2.9e-005,
		0.001665,
		0.024223,
		0.067471,
		0.125272,
		0.192655,
		0.265625,
		0.341002,
		0.416289,
		0.489559,
		0.559361,
		0.624639,
		0.684668,
		0.738994,
		0.787387,
		0.829796,
		0.86631,
		0.897131,
		0.92254,
		0.942872,
		0.958497,
		0.969798,
		0.977154,
		0.980929,
		0.981452,
		0.979009,
		0.973833,
		0.96609,
		0.955875,
		0.943204,
		0.928019,
		0.910189,
		0.889535,
		0.865876,
		0.839077,
		0.80908,
		0.775926,
		0.739762,
		0.700845,
		0.65953,
		0.616261,
		0.571549,
		0.52595,
		0.480048,
		0.434424,
		0.389643,
		0.346229,
		0.304652,
		0.265316,
		0.228551,
		0.194608,
		0.163659,
		0.135801,
		0.111057,
		0.089384,
		0.070681,
		0.0548,
		0.041551,
		0.030711,
		0.022037,
		0.015271,
		0.010148,
		0.006407,
		0.003792,
		0.002063,
		0.001001,
		0.000412,
		0.000131,
		2.6e-005,
		0.006746,
		0.038254,
		0.088884,
		0.152563,
		0.224414,
		0.300551,
		0.377913,
		0.45413,
		0.527402,
		0.596407,
		0.660215,
		0.718225,
		0.770102,
		0.815734,
		0.855181,
		0.888645,
		0.916436,
		0.938936,
		0.956583,
		0.969837,
		0.979167,
		0.985025,
		0.987832,
		0.987957,
		0.985703,
		0.981295,
		0.974863,
		0.966438,
		0.955945,
		0.943204,
		0.927957,
		0.909914,
		0.88881,
		0.864439,
		0.836688,
		0.805547,
		0.771123,
		0.733637,
		0.693416,
		0.650881,
		0.606532,
		0.560924,
		0.514646,
		0.468297,
		0.422467,
		0.377712,
		0.334543,
		0.293406,
		0.254678,
		0.218656,
		0.185558,
		0.155521,
		0.128609,
		0.104814,
		0.084068,
		0.066246,
		0.051181,
		0.038668,
		0.028478,
		0.02036,
		0.014057,
		0.009307,
		0.005853,
		0.003451,
		0.00187,
		0.000904,
		0.000371,
		0.000117,
		2.3e-005,
		0.015366,
		0.055645,
		0.113386,
		0.182603,
		0.258527,
		0.337397,
		0.416289,
		0.492972,
		0.565786,
		0.63355,
		0.69547,
		0.751076,
		0.800162,
		0.842734,
		0.878971,
		0.909189,
		0.9338,
		0.953294,
		0.968199,
		0.979067,
		0.986442,
		0.990841,
		0.992731,
		0.992508,
		0.990479,
		0.986844,
		0.981682,
		0.974941,
		0.966438,
		0.955875,
		0.942893,
		0.927134,
		0.908275,
		0.886069,
		0.86036,
		0.831105,
		0.798372,
		0.762347,
		0.723327,
		0.681705,
		0.637959,
		0.59263,
		0.546301,
		0.499578,
		0.453061,
		0.407332,
		0.362928,
		0.320332,
		0.279959,
		0.242148,
		0.207159,
		0.175172,
		0.146287,
		0.120532,
		0.097872,
		0.078208,
		0.061398,
		0.047255,
		0.035565,
		0.02609,
		0.018579,
		0.012776,
		0.008424,
		0.005276,
		0.003097,
		0.001671,
		0.000804,
		0.000328,
		0.000103,
		2e-005,
		0.001913,
		0.027641,
		0.076461,
		0.140991,
		0.215353,
		0.294906,
		0.376031,
		0.455958,
		0.532609,
		0.604481,
		0.670542,
		0.730146,
		0.782964,
		0.828924,
		0.86816,
		0.900973,
		0.927789,
		0.94913,
		0.965582,
		0.977766,
		0.986312,
		0.991829,
		0.994885,
		0.995977,
		0.995509,
		0.993768,
		0.990908,
		0.986931,
		0.981682,
		0.974863,
		0.96609,
		0.95495,
		0.941046,
		0.924032,
		0.903638,
		0.879686,
		0.852106,
		0.820937,
		0.786334,
		0.748558,
		0.707972,
		0.665023,
		0.620228,
		0.574154,
		0.527398,
		0.480563,
		0.434238,
		0.388981,
		0.3453,
		0.303641,
		0.264379,
		0.22781,
		0.194153,
		0.163544,
		0.136047,
		0.111657,
		0.090306,
		0.071873,
		0.056194,
		0.043071,
		0.03228,
		0.023578,
		0.016717,
		0.011443,
		0.007511,
		0.004682,
		0.002735,
		0.001468,
		0.000703,
		0.000285,
		8.9e-005,
		1.7e-005,
		0.007758,
		0.043674,
		0.100747,
		0.171689,
		0.250747,
		0.333432,
		0.416289,
		0.496716,
		0.572806,
		0.643226,
		0.707107,
		0.763962,
		0.813614,
		0.856131,
		0.891783,
		0.920994,
		0.944307,
		0.962349,
		0.975798,
		0.98536,
		0.991729,
		0.995567,
		0.997471,
		0.997943,
		0.997368,
		0.995983,
		0.993865,
		0.990908,
		0.986844,
		0.981295,
		0.973833,
		0.964028,
		0.951479,
		0.935839,
		0.916836,
		0.894285,
		0.868102,
		0.838304,
		0.805015,
		0.768466,
		0.72898,
		0.686971,
		0.642924,
		0.597379,
		0.550912,
		0.504116,
		0.457578,
		0.411863,
		0.367494,
		0.324939,
		0.284601,
		0.246807,
		0.211808,
		0.179776,
		0.150806,
		0.124925,
		0.102092,
		0.082212,
		0.065143,
		0.050704,
		0.038685,
		0.028856,
		0.020976,
		0.014799,
		0.010079,
		0.00658,
		0.00408,
		0.00237,
		0.001265,
		0.000602,
		0.000243,
		7.5e-005,
		1.5e-005,
		0.017685,
		0.063557,
		0.128533,
		0.205443,
		0.28869,
		0.373956,
		0.457972,
		0.538323,
		0.613292,
		0.681729,
		0.742944,
		0.796619,
		0.842738,
		0.881522,
		0.913379,
		0.938862,
		0.95863,
		0.973413,
		0.983977,
		0.991097,
		0.995519,
		0.997932,
		0.998935,
		0.999005,
		0.998467,
		0.997474,
		0.995983,
		0.993768,
		0.990479,
		0.985703,
		0.979009,
		0.969981,
		0.958238,
		0.943456,
		0.925377,
		0.903823,
		0.878705,
		0.850026,
		0.817888,
		0.782486,
		0.744109,
		0.703131,
		0.659998,
		0.615215,
		0.569328,
		0.522909,
		0.47653,
		0.430753,
		0.386103,
		0.343062,
		0.302051,
		0.26342,
		0.227449,
		0.194338,
		0.164213,
		0.137128,
		0.11307,
		0.09197,
		0.073706,
		0.058116,
		0.045007,
		0.034161,
		0.025346,
		0.018324,
		0.012854,
		0.008703,
		0.005648,
		0.003479,
		0.002008,
		0.001064,
		0.000503,
		0.000201,
		6.2e-005,
		1.2e-005,
		0.002221,
		0.031834,
		0.08736,
		0.159817,
		0.242186,
		0.32905,
		0.416289,
		0.500843,
		0.580504,
		0.653763,
		0.71967,
		0.777728,
		0.827802,
		0.870045,
		0.904839,
		0.932739,
		0.954428,
		0.970681,
		0.982326,
		0.990206,
		0.99515,
		0.99793,
		0.999235,
		0.999627,
		0.999515,
		0.999123,
		0.998467,
		0.997368,
		0.995509,
		0.992508,
		0.987957,
		0.981452,
		0.972616,
		0.961111,
		0.946649,
		0.929002,
		0.908012,
		0.883593,
		0.855742,
		0.824541,
		0.790158,
		0.752845,
		0.712934,
		0.67083,
		0.626996,
		0.581942,
		0.536205,
		0.490336,
		0.444878,
		0.400354,
		0.357244,
		0.315979,
		0.276925,
		0.240381,
		0.206571,
		0.175647,
		0.147689,
		0.122709,
		0.10066,
		0.081444,
		0.064916,
		0.0509,
		0.039191,
		0.029569,
		0.021804,
		0.015662,
		0.010914,
		0.007338,
		0.004727,
		0.00289,
		0.001654,
		0.00087,
		0.000407,
		0.000161,
		4.9e-005,
		0.009016,
		0.05033,
		0.11513,
		0.19457,
		0.281814,
		0.371655,
		0.460201,
		0.544621,
		0.622942,
		0.693878,
		0.756699,
		0.811114,
		0.857183,
		0.895243,
		0.925841,
		0.949683,
		0.967586,
		0.980438,
		0.989153,
		0.994638,
		0.99775,
		0.999258,
		0.999806,
		0.999877,
		0.999758,
		0.999515,
		0.999005,
		0.997943,
		0.995977,
		0.992731,
		0.987832,
		0.980929,
		0.971702,
		0.95987,
		0.945195,
		0.92749,
		0.90662,
		0.882517,
		0.855177,
		0.824668,
		0.791133,
		0.754792,
		0.715937,
		0.674927,
		0.632181,
		0.588164,
		0.543376,
		0.498335,
		0.453558,
		0.409549,
		0.366781,
		0.325684,
		0.286631,
		0.249933,
		0.21583,
		0.184495,
		0.156029,
		0.130466,
		0.107782,
		0.087897,
		0.070687,
		0.055989,
		0.043615,
		0.033356,
		0.02499,
		0.018292,
		0.013038,
		0.009012,
		0.006008,
		0.003836,
		0.002323,
		0.001317,
		0.000685,
		0.000317,
		0.000124,
		3.7e-005,
		0.02057,
		0.073275,
		0.146884,
		0.232725,
		0.324181,
		0.416289,
		0.505413,
		0.588982,
		0.665277,
		0.73326,
		0.792437,
		0.842738,
		0.884434,
		0.918052,
		0.944313,
		0.964076,
		0.978292,
		0.987949,
		0.994038,
		0.997501,
		0.999192,
		0.999832,
		0.999972,
		0.999952,
		0.999877,
		0.999627,
		0.998935,
		0.997471,
		0.994885,
		0.990841,
		0.985025,
		0.977154,
		0.966979,
		0.954283,
		0.938887,
		0.92065,
		0.899472,
		0.875305,
		0.84815,
		0.818069,
		0.785187,
		0.749691,
		0.711837,
		0.67194,
		0.630373,
		0.587553,
		0.543936,
		0.5,
		0.456229,
		0.4131,
		0.371068,
		0.330552,
		0.291922,
		0.255494,
		0.221518,
		0.190179,
		0.161595,
		0.135819,
		0.112844,
		0.092608,
		0.075006,
		0.05989,
		0.047085,
		0.036394,
		0.027607,
		0.020506,
		0.014875,
		0.010502,
		0.007186,
		0.004739,
		0.002991,
		0.00179,
		0.001001,
		0.000513,
		0.000234,
		9e-005,
		2.7e-005,
		0.00261,
		0.037057,
		0.100747,
		0.182603,
		0.27417,
		0.369091,
		0.462682,
		0.551596,
		0.63355,
		0.707107,
		0.771497,
		0.826481,
		0.87223,
		0.909232,
		0.938217,
		0.960083,
		0.975845,
		0.986574,
		0.99335,
		0.99721,
		0.999098,
		0.999818,
		0.999987,
		0.999994,
		0.999972,
		0.999806,
		0.999235,
		0.997932,
		0.995567,
		0.991829,
		0.986442,
		0.979167,
		0.969798,
		0.958161,
		0.944112,
		0.927533,
		0.908331,
		0.886448,
		0.861858,
		0.834575,
		0.80466,
		0.772221,
		0.737423,
		0.700485,
		0.661683,
		0.621342,
		0.579832,
		0.537563,
		0.494965,
		0.452485,
		0.410564,
		0.369632,
		0.330088,
		0.292291,
		0.256552,
		0.223122,
		0.192194,
		0.163895,
		0.138292,
		0.115392,
		0.095151,
		0.077474,
		0.062232,
		0.049261,
		0.038377,
		0.029379,
		0.022059,
		0.016209,
		0.011623,
		0.008105,
		0.005473,
		0.003559,
		0.002212,
		0.001301,
		0.000715,
		0.000359,
		0.00016,
		6e-005,
		1.7e-005,
		0.010606,
		0.058623,
		0.132788,
		0.222223,
		0.318742,
		0.416289,
		0.510503,
		0.598361,
		0.677898,
		0.747985,
		0.808146,
		0.858415,
		0.899218,
		0.931276,
		0.955526,
		0.973047,
		0.984999,
		0.992561,
		0.996876,
		0.998989,
		0.999797,
		0.999987,
		1,
		0.999987,
		0.999832,
		0.999258,
		0.99793,
		0.995519,
		0.991729,
		0.986312,
		0.979067,
		0.969837,
		0.958497,
		0.944951,
		0.929125,
		0.910961,
		0.890421,
		0.867484,
		0.842154,
		0.814459,
		0.784462,
		0.752265,
		0.71801,
		0.681888,
		0.644135,
		0.605034,
		0.564909,
		0.524119,
		0.48305,
		0.442101,
		0.401677,
		0.362172,
		0.323959,
		0.287379,
		0.252728,
		0.220253,
		0.190145,
		0.162535,
		0.137498,
		0.115052,
		0.095162,
		0.077749,
		0.062693,
		0.049843,
		0.039027,
		0.030053,
		0.022723,
		0.016836,
		0.012195,
		0.008611,
		0.005905,
		0.003915,
		0.002495,
		0.001517,
		0.000871,
		0.000466,
		0.000227,
		9.8e-005,
		3.5e-005,
		0.024223,
		0.085389,
		0.169394,
		0.265625,
		0.366214,
		0.465459,
		0.559361,
		0.645261,
		0.721547,
		0.787425,
		0.842738,
		0.887815,
		0.92335,
		0.950308,
		0.969836,
		0.983187,
		0.991651,
		0.99649,
		0.998863,
		0.999771,
		0.999985,
		1,
		0.999987,
		0.999818,
		0.999192,
		0.99775,
		0.99515,
		0.991097,
		0.98536,
		0.977766,
		0.968199,
		0.956583,
		0.942872,
		0.927046,
		0.909096,
		0.889025,
		0.866844,
		0.842572,
		0.81624,
		0.787895,
		0.757607,
		0.725472,
		0.691617,
		0.656209,
		0.619449,
		0.581581,
		0.542884,
		0.50367,
		0.464277,
		0.425058,
		0.386375,
		0.348584,
		0.312024,
		0.27701,
		0.243819,
		0.212684,
		0.183787,
		0.15726,
		0.133176,
		0.11156,
		0.092384,
		0.075577,
		0.061029,
		0.0486,
		0.038125,
		0.029425,
		0.02231,
		0.016587,
		0.012067,
		0.008567,
		0.005917,
		0.00396,
		0.002555,
		0.001579,
		0.000927,
		0.000512,
		0.000261,
		0.000121,
		4.9e-005,
		1.6e-005,
		0.003111,
		0.043674,
		0.117431,
		0.210511,
		0.312626,
		0.416289,
		0.516206,
		0.608789,
		0.691782,
		0.763962,
		0.824905,
		0.874792,
		0.914268,
		0.944313,
		0.966137,
		0.981094,
		0.990599,
		0.996042,
		0.998717,
		0.999741,
		0.999984,
		1,
		0.999985,
		0.999797,
		0.999098,
		0.997501,
		0.994638,
		0.990206,
		0.983977,
		0.975798,
		0.965582,
		0.953294,
		0.938936,
		0.92254,
		0.904151,
		0.883821,
		0.861608,
		0.837568,
		0.811757,
		0.784233,
		0.755064,
		0.724327,
		0.692115,
		0.658546,
		0.623765,
		0.587947,
		0.551297,
		0.514056,
		0.476489,
		0.438887,
		0.40156,
		0.364821,
		0.328987,
		0.294361,
		0.261224,
		0.229828,
		0.200385,
		0.173066,
		0.14799,
		0.12523,
		0.104807,
		0.086698,
		0.070835,
		0.057115,
		0.045407,
		0.035553,
		0.027382,
		0.020712,
		0.01536,
		0.011143,
		0.007888,
		0.005431,
		0.003623,
		0.002331,
		0.001437,
		0.000842,
		0.000464,
		0.000237,
		0.00011,
		4.5e-005,
		1.6e-005,
		0.012657,
		0.069138,
		0.154777,
		0.256014,
		0.362964,
		0.468589,
		0.568056,
		0.658245,
		0.737347,
		0.804566,
		0.859876,
		0.903828,
		0.937398,
		0.961859,
		0.978669,
		0.989376,
		0.995521,
		0.998546,
		0.999707,
		0.999981,
		1,
		0.999984,
		0.999771,
		0.998989,
		0.99721,
		0.994038,
		0.989153,
		0.982326,
		0.973413,
		0.962349,
		0.94913,
		0.9338,
		0.916436,
		0.897131,
		0.875992,
		0.853125,
		0.828631,
		0.802607,
		0.775143,
		0.746323,
		0.71623,
		0.684951,
		0.652581,
		0.619229,
		0.585024,
		0.550115,
		0.514679,
		0.478918,
		0.443057,
		0.407343,
		0.372038,
		0.337415,
		0.303745,
		0.271292,
		0.240301,
		0.210995,
		0.183562,
		0.158149,
		0.134865,
		0.11377,
		0.09488,
		0.07817,
		0.063573,
		0.050989,
		0.040291,
		0.031327,
		0.023933,
		0.017935,
		0.013156,
		0.009423,
		0.00657,
		0.004443,
		0.0029,
		0.001816,
		0.001083,
		0.000608,
		0.000316,
		0.000149,
		6.2e-005,
		2.1e-005,
		0.028941,
		0.100747,
		0.197385,
		0.3057,
		0.416289,
		0.522638,
		0.620445,
		0.707107,
		0.781314,
		0.842738,
		0.891784,
		0.929393,
		0.956891,
		0.975845,
		0.987949,
		0.994912,
		0.998347,
		0.999666,
		0.999979,
		1,
		0.999981,
		0.999741,
		0.998863,
		0.996876,
		0.99335,
		0.987949,
		0.980438,
		0.970681,
		0.95863,
		0.944307,
		0.927789,
		0.909189,
		0.888645,
		0.86631,
		0.842337,
		0.816874,
		0.790063,
		0.762033,
		0.732899,
		0.702768,
		0.671739,
		0.639906,
		0.607367,
		0.574223,
		0.540591,
		0.5066,
		0.472399,
		0.438158,
		0.404065,
		0.370327,
		0.337163,
		0.304801,
		0.273471,
		0.243396,
		0.214785,
		0.187825,
		0.162676,
		0.139462,
		0.118271,
		0.099147,
		0.082099,
		0.067091,
		0.054055,
		0.04289,
		0.033468,
		0.025644,
		0.019257,
		0.014139,
		0.010121,
		0.007038,
		0.004732,
		0.003057,
		0.001882,
		0.001091,
		0.000586,
		0.000283,
		0.000116,
		3.7e-005,
		0.003771,
		0.05223,
		0.138573,
		0.245133,
		0.359265,
		0.472143,
		0.577857,
		0.672708,
		0.754672,
		0.82299,
		0.877839,
		0.92009,
		0.951098,
		0.972543,
		0.986276,
		0.994196,
		0.998111,
		0.999618,
		0.999976,
		1,
		0.999979,
		0.999707,
		0.998717,
		0.99649,
		0.992561,
		0.986574,
		0.978292,
		0.967586,
		0.954428,
		0.938862,
		0.920994,
		0.900973,
		0.878971,
		0.855181,
		0.829796,
		0.803008,
		0.775,
		0.745942,
		0.715985,
		0.685268,
		0.653913,
		0.622029,
		0.589719,
		0.557079,
		0.524209,
		0.491211,
		0.458198,
		0.425297,
		0.392648,
		0.360406,
		0.32874,
		0.297832,
		0.267868,
		0.239036,
		0.21152,
		0.18549,
		0.161098,
		0.138472,
		0.11771,
		0.098873,
		0.08199,
		0.067052,
		0.054013,
		0.042798,
		0.0333,
		0.025391,
		0.018926,
		0.013746,
		0.009688,
		0.006591,
		0.004295,
		0.002653,
		0.001528,
		0.000799,
		0.000361,
		0.000127,
		2.4e-005,
		0.015366,
		0.082738,
		0.182603,
		0.297795,
		0.416289,
		0.529947,
		0.63355,
		0.724081,
		0.800163,
		0.861636,
		0.909232,
		0.944313,
		0.968661,
		0.984305,
		0.99335,
		0.997833,
		0.999561,
		0.999972,
		1,
		0.999976,
		0.999666,
		0.998546,
		0.996042,
		0.991651,
		0.984999,
		0.975845,
		0.964076,
		0.949683,
		0.932739,
		0.913379,
		0.891783,
		0.86816,
		0.842734,
		0.815734,
		0.787387,
		0.757914,
		0.727518,
		0.696388,
		0.664693,
		0.632584,
		0.600193,
		0.567636,
		0.535021,
		0.502441,
		0.469991,
		0.437764,
		0.405858,
		0.374377,
		0.343437,
		0.313165,
		0.283695,
		0.255173,
		0.227748,
		0.201568,
		0.176778,
		0.15351,
		0.131883,
		0.111989,
		0.093895,
		0.07764,
		0.063227,
		0.050629,
		0.039786,
		0.03061,
		0.022989,
		0.016787,
		0.011859,
		0.008047,
		0.005191,
		0.003134,
		0.001726,
		0.000826,
		0.000308,
		6.4e-005,
		0.035182,
		0.120607,
		0.232725,
		0.355015,
		0.476215,
		0.588982,
		0.688896,
		0.773702,
		0.842738,
		0.896506,
		0.936326,
		0.964076,
		0.981968,
		0.992346,
		0.997501,
		0.999493,
		0.999968,
		1,
		0.999972,
		0.999618,
		0.998347,
		0.995521,
		0.990599,
		0.983187,
		0.973047,
		0.960083,
		0.944313,
		0.925841,
		0.904839,
		0.881522,
		0.856131,
		0.828924,
		0.800162,
		0.770102,
		0.738994,
		0.707073,
		0.674555,
		0.641638,
		0.608498,
		0.575293,
		0.542159,
		0.509214,
		0.476563,
		0.444298,
		0.412507,
		0.381271,
		0.350675,
		0.320804,
		0.291752,
		0.263617,
		0.236507,
		0.210532,
		0.185806,
		0.162442,
		0.140548,
		0.120219,
		0.101537,
		0.084562,
		0.069331,
		0.055852,
		0.044105,
		0.03404,
		0.02558,
		0.018622,
		0.013042,
		0.008699,
		0.005442,
		0.003111,
		0.00155,
		0.000604,
		0.000131,
		0.004665,
		0.063557,
		0.165879,
		0.28869,
		0.416289,
		0.538323,
		0.64838,
		0.742944,
		0.820615,
		0.881522,
		0.926878,
		0.958631,
		0.979184,
		0.991144,
		0.997104,
		0.999412,
		0.999962,
		1,
		0.999968,
		0.999561,
		0.998111,
		0.994912,
		0.989376,
		0.981094,
		0.969836,
		0.955526,
		0.938217,
		0.918052,
		0.895243,
		0.870045,
		0.842738,
		0.813614,
		0.782964,
		0.751076,
		0.718225,
		0.684668,
		0.650647,
		0.616381,
		0.582068,
		0.547885,
		0.513985,
		0.480503,
		0.447554,
		0.415238,
		0.38364,
		0.352837,
		0.322902,
		0.293903,
		0.265908,
		0.238991,
		0.213225,
		0.188689,
		0.165466,
		0.143636,
		0.123278,
		0.104466,
		0.087259,
		0.071706,
		0.057832,
		0.045641,
		0.035114,
		0.026203,
		0.018834,
		0.012909,
		0.008309,
		0.004895,
		0.002517,
		0.001015,
		0.000228,
		0.019046,
		0.100747,
		0.218463,
		0.350084,
		0.480925,
		0.601712,
		0.707107,
		0.794618,
		0.863799,
		0.915644,
		0.952128,
		0.975845,
		0.9897,
		0.996625,
		0.999313,
		0.999956,
		1,
		0.999962,
		0.999493,
		0.997833,
		0.994196,
		0.987949,
		0.978669,
		0.966137,
		0.950308,
		0.931276,
		0.909232,
		0.884434,
		0.857183,
		0.827802,
		0.796619,
		0.763962,
		0.730146,
		0.69547,
		0.660215,
		0.624639,
		0.588979,
		0.553448,
		0.518237,
		0.483514,
		0.449423,
		0.416091,
		0.383622,
		0.352107,
		0.321622,
		0.292232,
		0.263995,
		0.236966,
		0.211195,
		0.186735,
		0.163637,
		0.141955,
		0.121743,
		0.103053,
		0.085931,
		0.070417,
		0.05654,
		0.044312,
		0.033727,
		0.024756,
		0.01735,
		0.011432,
		0.006903,
		0.003642,
		0.001509,
		0.000349,
		0.043674,
		0.146884,
		0.278096,
		0.416289,
		0.548017,
		0.665277,
		0.763962,
		0.842738,
		0.902212,
		0.944313,
		0.971816,
		0.987949,
		0.996042,
		0.999193,
		0.999948,
		1,
		0.999956,
		0.999412,
		0.997501,
		0.99335,
		0.986276,
		0.975845,
		0.961859,
		0.944313,
		0.92335,
		0.899218,
		0.87223,
		0.842738,
		0.811114,
		0.777728,
		0.742944,
		0.707107,
		0.670542,
		0.63355,
		0.596407,
		0.559361,
		0.522637,
		0.486434,
		0.450927,
		0.416269,
		0.38259,
		0.349999,
		0.31859,
		0.288436,
		0.259602,
		0.232137,
		0.206085,
		0.181483,
		0.158365,
		0.136762,
		0.116705,
		0.098223,
		0.081344,
		0.066092,
		0.052485,
		0.040531,
		0.030225,
		0.021549,
		0.014463,
		0.008906,
		0.004798,
		0.002032,
		0.000482,
		0.00592,
		0.078987,
		0.201932,
		0.344293,
		0.486437,
		0.61641,
		0.727694,
		0.817592,
		0.886059,
		0.93486,
		0.96692,
		0.985814,
		0.995329,
		0.999046,
		0.999939,
		1,
		0.999948,
		0.999313,
		0.997104,
		0.992346,
		0.984305,
		0.972543,
		0.956891,
		0.937398,
		0.914268,
		0.887815,
		0.858415,
		0.826481,
		0.792437,
		0.756699,
		0.71967,
		0.681729,
		0.643226,
		0.604481,
		0.565786,
		0.527402,
		0.489559,
		0.452462,
		0.416289,
		0.381192,
		0.347303,
		0.314732,
		0.283568,
		0.253886,
		0.225742,
		0.199181,
		0.174237,
		0.150937,
		0.129299,
		0.109339,
		0.091069,
		0.074497,
		0.059631,
		0.046472,
		0.035019,
		0.025261,
		0.017177,
		0.010731,
		0.005873,
		0.00253,
		0.000611,
		0.024222,
		0.125272,
		0.265625,
		0.416289,
		0.559361,
		0.684672,
		0.787425,
		0.86652,
		0.92335,
		0.960924,
		0.983187,
		0.994448,
		0.998863,
		0.999927,
		1,
		0.999939,
		0.999193,
		0.996625,
		0.991144,
		0.981968,
		0.968661,
		0.951098,
		0.929393,
		0.903828,
		0.874792,
		0.842738,
		0.808146,
		0.771497,
		0.73326,
		0.693878,
		0.653763,
		0.613292,
		0.572806,
		0.532609,
		0.492972,
		0.45413,
		0.416289,
		0.379627,
		0.344293,
		0.310414,
		0.278094,
		0.247418,
		0.218452,
		0.191247,
		0.165839,
		0.142252,
		0.120501,
		0.100594,
		0.082529,
		0.066304,
		0.05191,
		0.039335,
		0.028563,
		0.019574,
		0.01234,
		0.006823,
		0.002974,
		0.000727,
		0.055596,
		0.182595,
		0.337397,
		0.492972,
		0.63355,
		0.751077,
		0.842738,
		0.909232,
		0.953525,
		0.979927,
		0.99335,
		0.998635,
		0.999912,
		1,
		0.999927,
		0.999046,
		0.996042,
		0.9897,
		0.979184,
		0.964076,
		0.944313,
		0.92009,
		0.891784,
		0.859876,
		0.824905,
		0.787425,
		0.747985,
		0.707107,
		0.665277,
		0.622942,
		0.580504,
		0.538323,
		0.496716,
		0.455958,
		0.416289,
		0.377913,
		0.341002,
		0.3057,
		0.272124,
		0.24037,
		0.210511,
		0.182602,
		0.156684,
		0.132781,
		0.110907,
		0.091062,
		0.07324,
		0.057429,
		0.043608,
		0.031753,
		0.021838,
		0.013828,
		0.007687,
		0.003372,
		0.000831,
		0.007563,
		0.100129,
		0.250527,
		0.416269,
		0.572806,
		0.707107,
		0.813614,
		0.891784,
		0.944313,
		0.975845,
		0.991969,
		0.998347,
		0.999893,
		1,
		0.999912,
		0.998863,
		0.995329,
		0.987949,
		0.975845,
		0.958631,
		0.936326,
		0.909232,
		0.877839,
		0.842738,
		0.804566,
		0.763962,
		0.721547,
		0.677898,
		0.63355,
		0.588982,
		0.544621,
		0.500843,
		0.457972,
		0.416289,
		0.376031,
		0.337397,
		0.300551,
		0.265625,
		0.232725,
		0.201932,
		0.173305,
		0.146884,
		0.122695,
		0.100747,
		0.081038,
		0.063554,
		0.048276,
		0.035173,
		0.024212,
		0.015352,
		0.008551,
		0.003761,
		0.00093,
		0.029542,
		0.155812,
		0.327031,
		0.500403,
		0.653732,
		0.777728,
		0.870045,
		0.932739,
		0.970681,
		0.990212,
		0.997978,
		0.999869,
		1,
		0.999893,
		0.998635,
		0.994448,
		0.985814,
		0.971816,
		0.952128,
		0.926878,
		0.896506,
		0.861636,
		0.82299,
		0.781314,
		0.737347,
		0.691782,
		0.645261,
		0.598361,
		0.551596,
		0.505413,
		0.460201,
		0.416289,
		0.373956,
		0.333432,
		0.294906,
		0.258527,
		0.224414,
		0.192655,
		0.163312,
		0.136427,
		0.112019,
		0.090096,
		0.070648,
		0.053654,
		0.039085,
		0.026901,
		0.017057,
		0.009502,
		0.004181,
		0.001034,
		0.061533,
		0.215973,
		0.405858,
		0.585368,
		0.732617,
		0.842698,
		0.918052,
		0.964076,
		0.987949,
		0.997501,
		0.999838,
		1,
		0.999869,
		0.998347,
		0.99335,
		0.983187,
		0.96692,
		0.944313,
		0.915644,
		0.881522,
		0.842738,
		0.800163,
		0.754672,
		0.707107,
		0.658245,
		0.608789,
		0.559361,
		0.510503,
		0.462682,
		0.416289,
		0.371655,
		0.32905,
		0.28869,
		0.250747,
		0.215353,
		0.182603,
		0.152563,
		0.125272,
		0.100747,
		0.078987,
		0.059973,
		0.043674,
		0.030049,
		0.019046,
		0.010606,
		0.004665,
		0.001154,
		0.005759,
		0.093895,
		0.267665,
		0.473757,
		0.660911,
		0.803188,
		0.898429,
		0.955481,
		0.984999,
		0.996876,
		0.999797,
		1,
		0.999838,
		0.997978,
		0.991969,
		0.979927,
		0.960924,
		0.93486,
		0.902212,
		0.863799,
		0.820615,
		0.773702,
		0.724081,
		0.672708,
		0.620445,
		0.568056,
		0.516206,
		0.465459,
		0.416289,
		0.369091,
		0.324181,
		0.281814,
		0.242186,
		0.205443,
		0.171689,
		0.140991,
		0.113386,
		0.088884,
		0.067471,
		0.049119,
		0.033782,
		0.021402,
		0.011912,
		0.005236,
		0.001294,
		0.01627,
		0.1143,
		0.294361,
		0.511233,
		0.708972,
		0.852871,
		0.938519,
		0.980234,
		0.995994,
		0.999741,
		1,
		0.999797,
		0.997501,
		0.990212,
		0.975845,
		0.953525,
		0.92335,
		0.886059,
		0.842738,
		0.794618,
		0.742944,
		0.688896,
		0.63355,
		0.577857,
		0.522638,
		0.468589,
		0.416289,
		0.366214,
		0.318742,
		0.27417,
		0.232725,
		0.19457,
		0.159817,
		0.128533,
		0.100747,
		0.076461,
		0.055645,
		0.038254,
		0.024223,
		0.013475,
		0.00592,
		0.001462,
		0.022412,
		0.113883,
		0.283774,
		0.5,
		0.707694,
		0.862495,
		0.951392,
		0.988808,
		0.998789,
		0.999952,
		0.999741,
		0.996876,
		0.987949,
		0.970681,
		0.944313,
		0.909232,
		0.86652,
		0.817592,
		0.763962,
		0.707107,
		0.64838,
		0.588982,
		0.529947,
		0.472143,
		0.416289,
		0.362964,
		0.312626,
		0.265625,
		0.222223,
		0.182603,
		0.146884,
		0.11513,
		0.08736,
		0.063557,
		0.043674,
		0.027641,
		0.015366,
		0.006746,
		0.001665,
		0.000644,
		0.020197,
		0.092608,
		0.236018,
		0.434461,
		0.642924,
		0.813438,
		0.921848,
		0.974514,
		0.993865,
		0.998789,
		0.995994,
		0.984999,
		0.964076,
		0.932739,
		0.891784,
		0.842738,
		0.787425,
		0.727694,
		0.665277,
		0.601712,
		0.538323,
		0.476215,
		0.416289,
		0.359265,
		0.3057,
		0.256014,
		0.210511,
		0.169394,
		0.132788,
		0.100747,
		0.073275,
		0.05033,
		0.031834,
		0.017685,
		0.007758,
		0.001913,
		0.000623,
		0.012091,
		0.05954,
		0.165276,
		0.328396,
		0.520504,
		0.700845,
		0.839261,
		0.928019,
		0.974514,
		0.988808,
		0.980234,
		0.955481,
		0.918052,
		0.870045,
		0.813614,
		0.751077,
		0.684672,
		0.61641,
		0.548017,
		0.480925,
		0.416289,
		0.355015,
		0.297795,
		0.245133,
		0.197385,
		0.154777,
		0.117431,
		0.085389,
		0.058623,
		0.037057,
		0.02057,
		0.009016,
		0.002221,
		0.000106,
		0.00408,
		0.027863,
		0.09321,
		0.210073,
		0.368043,
		0.542528,
		0.707107,
		0.839261,
		0.921848,
		0.951392,
		0.938519,
		0.898429,
		0.842698,
		0.777728,
		0.707107,
		0.63355,
		0.559361,
		0.486437,
		0.416289,
		0.350084,
		0.28869,
		0.232725,
		0.182603,
		0.138573,
		0.100747,
		0.069138,
		0.043674,
		0.024223,
		0.010606,
		0.00261,
		0.000402,
		0.007717,
		0.038672,
		0.108744,
		0.222209,
		0.372533,
		0.542528,
		0.700845,
		0.813438,
		0.862495,
		0.852871,
		0.803188,
		0.732617,
		0.653732,
		0.572806,
		0.492972,
		0.416289,
		0.344293,
		0.278096,
		0.218463,
		0.165879,
		0.120607,
		0.082738,
		0.05223,
		0.028941,
		0.012657,
		0.003111,
		0.000642,
		0.009563,
		0.041813,
		0.110603,
		0.222209,
		0.368043,
		0.520504,
		0.642924,
		0.707694,
		0.708972,
		0.660911,
		0.585368,
		0.500403,
		0.416269,
		0.337397,
		0.265625,
		0.201932,
		0.146884,
		0.100747,
		0.063557,
		0.035182,
		0.015366,
		0.003771,
		0.000718,
		0.0098,
		0.041813,
		0.108744,
		0.210073,
		0.328396,
		0.434461,
		0.5,
		0.511233,
		0.473757,
		0.405858,
		0.327031,
		0.250527,
		0.182595,
		0.125272,
		0.078987,
		0.043674,
		0.019046,
		0.004665,
		0.000718,
		0.009563,
		0.038672,
		0.09321,
		0.165276,
		0.236018,
		0.283774,
		0.294361,
		0.267665,
		0.215973,
		0.155812,
		0.100129,
		0.055596,
		0.024222,
		0.00592,
		0.000642,
		0.007717,
		0.027863,
		0.05954,
		0.092608,
		0.113883,
		0.1143,
		0.093895,
		0.061533,
		0.029542,
		0.007563,
		0.000402,
		0.00408,
		0.012091,
		0.020197,
		0.022412,
		0.01627,
		0.005759,
		0.000106,
		0.000623,
		0.000644,
		0.000644,
		0.000623,
		0.000106,
		0.005759,
		0.01627,
		0.022412,
		0.020197,
		0.012091,
		0.00408,
		0.000402,
		0.007563,
		0.029542,
		0.061533,
		0.093895,
		0.1143,
		0.113883,
		0.092608,
		0.05954,
		0.027863,
		0.007717,
		0.000642,
		0.00592,
		0.024222,
		0.055596,
		0.100129,
		0.155812,
		0.215973,
		0.267665,
		0.294361,
		0.283774,
		0.236018,
		0.165276,
		0.09321,
		0.038672,
		0.009563,
		0.000718,
		0.004665,
		0.019046,
		0.043674,
		0.078987,
		0.125272,
		0.182595,
		0.250527,
		0.327031,
		0.405858,
		0.473757,
		0.511233,
		0.5,
		0.434461,
		0.328396,
		0.210073,
		0.108744,
		0.041813,
		0.0098,
		0.000718,
		0.003771,
		0.015366,
		0.035182,
		0.063557,
		0.100747,
		0.146884,
		0.201932,
		0.265625,
		0.337397,
		0.416269,
		0.500403,
		0.585368,
		0.660911,
		0.708972,
		0.707694,
		0.642924,
		0.520504,
		0.368043,
		0.222209,
		0.110603,
		0.041813,
		0.009563,
		0.000642,
		0.003111,
		0.012657,
		0.028941,
		0.05223,
		0.082738,
		0.120607,
		0.165879,
		0.218463,
		0.278096,
		0.344293,
		0.416289,
		0.492972,
		0.572806,
		0.653732,
		0.732617,
		0.803188,
		0.852871,
		0.862495,
		0.813438,
		0.700845,
		0.542528,
		0.372533,
		0.222209,
		0.108744,
		0.038672,
		0.007717,
		0.000402,
		0.00261,
		0.010606,
		0.024223,
		0.043674,
		0.069138,
		0.100747,
		0.138573,
		0.182603,
		0.232725,
		0.28869,
		0.350084,
		0.416289,
		0.486437,
		0.559361,
		0.63355,
		0.707107,
		0.777728,
		0.842698,
		0.898429,
		0.938519,
		0.951392,
		0.921848,
		0.839261,
		0.707107,
		0.542528,
		0.368043,
		0.210073,
		0.09321,
		0.027863,
		0.00408,
		0.000106,
		0.002221,
		0.009016,
		0.02057,
		0.037057,
		0.058623,
		0.085389,
		0.117431,
		0.154777,
		0.197385,
		0.245133,
		0.297795,
		0.355015,
		0.416289,
		0.480925,
		0.548017,
		0.61641,
		0.684672,
		0.751077,
		0.813614,
		0.870045,
		0.918052,
		0.955481,
		0.980234,
		0.988808,
		0.974514,
		0.928019,
		0.839261,
		0.700845,
		0.520504,
		0.328396,
		0.165276,
		0.05954,
		0.012091,
		0.000623,
		0.001913,
		0.007758,
		0.017685,
		0.031834,
		0.05033,
		0.073275,
		0.100747,
		0.132788,
		0.169394,
		0.210511,
		0.256014,
		0.3057,
		0.359265,
		0.416289,
		0.476215,
		0.538323,
		0.601712,
		0.665277,
		0.727694,
		0.787425,
		0.842738,
		0.891784,
		0.932739,
		0.964076,
		0.984999,
		0.995994,
		0.998789,
		0.993865,
		0.974514,
		0.921848,
		0.813438,
		0.642924,
		0.434461,
		0.236018,
		0.092608,
		0.020197,
		0.000644,
		0.001665,
		0.006746,
		0.015366,
		0.027641,
		0.043674,
		0.063557,
		0.08736,
		0.11513,
		0.146884,
		0.182603,
		0.222223,
		0.265625,
		0.312626,
		0.362964,
		0.416289,
		0.472143,
		0.529947,
		0.588982,
		0.64838,
		0.707107,
		0.763962,
		0.817592,
		0.86652,
		0.909232,
		0.944313,
		0.970681,
		0.987949,
		0.996876,
		0.999741,
		0.999952,
		0.998789,
		0.988808,
		0.951392,
		0.862495,
		0.707694,
		0.5,
		0.283774,
		0.113883,
		0.022412,
		0.001462,
		0.00592,
		0.013475,
		0.024223,
		0.038254,
		0.055645,
		0.076461,
		0.100747,
		0.128533,
		0.159817,
		0.19457,
		0.232725,
		0.27417,
		0.318742,
		0.366214,
		0.416289,
		0.468589,
		0.522638,
		0.577857,
		0.63355,
		0.688896,
		0.742944,
		0.794618,
		0.842738,
		0.886059,
		0.92335,
		0.953525,
		0.975845,
		0.990212,
		0.997501,
		0.999797,
		1,
		0.999741,
		0.995994,
		0.980234,
		0.938519,
		0.852871,
		0.708972,
		0.511233,
		0.294361,
		0.1143,
		0.01627,
		0.001294,
		0.005236,
		0.011912,
		0.021402,
		0.033782,
		0.049119,
		0.067471,
		0.088884,
		0.113386,
		0.140991,
		0.171689,
		0.205443,
		0.242186,
		0.281814,
		0.324181,
		0.369091,
		0.416289,
		0.465459,
		0.516206,
		0.568056,
		0.620445,
		0.672708,
		0.724081,
		0.773702,
		0.820615,
		0.863799,
		0.902212,
		0.93486,
		0.960924,
		0.979927,
		0.991969,
		0.997978,
		0.999838,
		1,
		0.999797,
		0.996876,
		0.984999,
		0.955481,
		0.898429,
		0.803188,
		0.660911,
		0.473757,
		0.267665,
		0.093895,
		0.005759,
		0.001154,
		0.004665,
		0.010606,
		0.019046,
		0.030049,
		0.043674,
		0.059973,
		0.078987,
		0.100747,
		0.125272,
		0.152563,
		0.182603,
		0.215353,
		0.250747,
		0.28869,
		0.32905,
		0.371655,
		0.416289,
		0.462682,
		0.510503,
		0.559361,
		0.608789,
		0.658245,
		0.707107,
		0.754672,
		0.800163,
		0.842738,
		0.881522,
		0.915644,
		0.944313,
		0.96692,
		0.983187,
		0.99335,
		0.998347,
		0.999869,
		1,
		0.999838,
		0.997501,
		0.987949,
		0.964076,
		0.918052,
		0.842698,
		0.732617,
		0.585368,
		0.405858,
		0.215973,
		0.061533,
		0.001034,
		0.004181,
		0.009502,
		0.017057,
		0.026901,
		0.039085,
		0.053654,
		0.070648,
		0.090096,
		0.112019,
		0.136427,
		0.163312,
		0.192655,
		0.224414,
		0.258527,
		0.294906,
		0.333432,
		0.373956,
		0.416289,
		0.460201,
		0.505413,
		0.551596,
		0.598361,
		0.645261,
		0.691782,
		0.737347,
		0.781314,
		0.82299,
		0.861636,
		0.896506,
		0.926878,
		0.952128,
		0.971816,
		0.985814,
		0.994448,
		0.998635,
		0.999893,
		1,
		0.999869,
		0.997978,
		0.990212,
		0.970681,
		0.932739,
		0.870045,
		0.777728,
		0.653732,
		0.500403,
		0.327031,
		0.155812,
		0.029542,
		0.00093,
		0.003761,
		0.008551,
		0.015352,
		0.024212,
		0.035173,
		0.048276,
		0.063554,
		0.081038,
		0.100747,
		0.122695,
		0.146884,
		0.173305,
		0.201932,
		0.232725,
		0.265625,
		0.300551,
		0.337397,
		0.376031,
		0.416289,
		0.457972,
		0.500843,
		0.544621,
		0.588982,
		0.63355,
		0.677898,
		0.721547,
		0.763962,
		0.804566,
		0.842738,
		0.877839,
		0.909232,
		0.936326,
		0.958631,
		0.975845,
		0.987949,
		0.995329,
		0.998863,
		0.999912,
		1,
		0.999893,
		0.998347,
		0.991969,
		0.975845,
		0.944313,
		0.891784,
		0.813614,
		0.707107,
		0.572806,
		0.416269,
		0.250527,
		0.100129,
		0.007563,
		0.000831,
		0.003372,
		0.007687,
		0.013828,
		0.021838,
		0.031753,
		0.043608,
		0.057429,
		0.07324,
		0.091062,
		0.110907,
		0.132781,
		0.156684,
		0.182602,
		0.210511,
		0.24037,
		0.272124,
		0.3057,
		0.341002,
		0.377913,
		0.416289,
		0.455958,
		0.496716,
		0.538323,
		0.580504,
		0.622942,
		0.665277,
		0.707107,
		0.747985,
		0.787425,
		0.824905,
		0.859876,
		0.891784,
		0.92009,
		0.944313,
		0.964076,
		0.979184,
		0.9897,
		0.996042,
		0.999046,
		0.999927,
		1,
		0.999912,
		0.998635,
		0.99335,
		0.979927,
		0.953525,
		0.909232,
		0.842738,
		0.751077,
		0.63355,
		0.492972,
		0.337397,
		0.182595,
		0.055596,
		0.000727,
		0.002974,
		0.006823,
		0.01234,
		0.019574,
		0.028563,
		0.039335,
		0.05191,
		0.066304,
		0.082529,
		0.100594,
		0.120501,
		0.142252,
		0.165839,
		0.191247,
		0.218452,
		0.247418,
		0.278094,
		0.310414,
		0.344293,
		0.379627,
		0.416289,
		0.45413,
		0.492972,
		0.532609,
		0.572806,
		0.613292,
		0.653763,
		0.693878,
		0.73326,
		0.771497,
		0.808146,
		0.842738,
		0.874792,
		0.903828,
		0.929393,
		0.951098,
		0.968661,
		0.981968,
		0.991144,
		0.996625,
		0.999193,
		0.999939,
		1,
		0.999927,
		0.998863,
		0.994448,
		0.983187,
		0.960924,
		0.92335,
		0.86652,
		0.787425,
		0.684672,
		0.559361,
		0.416289,
		0.265625,
		0.125272,
		0.024222,
		0.000611,
		0.00253,
		0.005873,
		0.010731,
		0.017177,
		0.025261,
		0.035019,
		0.046472,
		0.059631,
		0.074497,
		0.091069,
		0.109339,
		0.129299,
		0.150937,
		0.174237,
		0.199181,
		0.225742,
		0.253886,
		0.283568,
		0.314732,
		0.347303,
		0.381192,
		0.416289,
		0.452462,
		0.489559,
		0.527402,
		0.565786,
		0.604481,
		0.643226,
		0.681729,
		0.71967,
		0.756699,
		0.792437,
		0.826481,
		0.858415,
		0.887815,
		0.914268,
		0.937398,
		0.956891,
		0.972543,
		0.984305,
		0.992346,
		0.997104,
		0.999313,
		0.999948,
		1,
		0.999939,
		0.999046,
		0.995329,
		0.985814,
		0.96692,
		0.93486,
		0.886059,
		0.817592,
		0.727694,
		0.61641,
		0.486437,
		0.344293,
		0.201932,
		0.078987,
		0.00592,
		0.000482,
		0.002032,
		0.004798,
		0.008906,
		0.014463,
		0.021549,
		0.030225,
		0.040531,
		0.052485,
		0.066092,
		0.081344,
		0.098223,
		0.116705,
		0.136762,
		0.158365,
		0.181483,
		0.206085,
		0.232137,
		0.259602,
		0.288436,
		0.31859,
		0.349999,
		0.38259,
		0.416269,
		0.450927,
		0.486434,
		0.522637,
		0.559361,
		0.596407,
		0.63355,
		0.670542,
		0.707107,
		0.742944,
		0.777728,
		0.811114,
		0.842738,
		0.87223,
		0.899218,
		0.92335,
		0.944313,
		0.961859,
		0.975845,
		0.986276,
		0.99335,
		0.997501,
		0.999412,
		0.999956,
		1,
		0.999948,
		0.999193,
		0.996042,
		0.987949,
		0.971816,
		0.944313,
		0.902212,
		0.842738,
		0.763962,
		0.665277,
		0.548017,
		0.416289,
		0.278096,
		0.146884,
		0.043674,
		0.000349,
		0.001509,
		0.003642,
		0.006903,
		0.011432,
		0.01735,
		0.024756,
		0.033727,
		0.044312,
		0.05654,
		0.070417,
		0.085931,
		0.103053,
		0.121743,
		0.141955,
		0.163637,
		0.186735,
		0.211195,
		0.236966,
		0.263995,
		0.292232,
		0.321622,
		0.352107,
		0.383622,
		0.416091,
		0.449423,
		0.483514,
		0.518237,
		0.553448,
		0.588979,
		0.624639,
		0.660215,
		0.69547,
		0.730146,
		0.763962,
		0.796619,
		0.827802,
		0.857183,
		0.884434,
		0.909232,
		0.931276,
		0.950308,
		0.966137,
		0.978669,
		0.987949,
		0.994196,
		0.997833,
		0.999493,
		0.999962,
		1,
		0.999956,
		0.999313,
		0.996625,
		0.9897,
		0.975845,
		0.952128,
		0.915644,
		0.863799,
		0.794618,
		0.707107,
		0.601712,
		0.480925,
		0.350084,
		0.218463,
		0.100747,
		0.019046,
		0.000228,
		0.001015,
		0.002517,
		0.004895,
		0.008309,
		0.012909,
		0.018834,
		0.026203,
		0.035114,
		0.045641,
		0.057832,
		0.071706,
		0.087259,
		0.104466,
		0.123278,
		0.143636,
		0.165466,
		0.188689,
		0.213225,
		0.238991,
		0.265908,
		0.293903,
		0.322902,
		0.352837,
		0.38364,
		0.415238,
		0.447554,
		0.480503,
		0.513985,
		0.547885,
		0.582068,
		0.616381,
		0.650647,
		0.684668,
		0.718225,
		0.751076,
		0.782964,
		0.813614,
		0.842738,
		0.870045,
		0.895243,
		0.918052,
		0.938217,
		0.955526,
		0.969836,
		0.981094,
		0.989376,
		0.994912,
		0.998111,
		0.999561,
		0.999968,
		1,
		0.999962,
		0.999412,
		0.997104,
		0.991144,
		0.979184,
		0.958631,
		0.926878,
		0.881522,
		0.820615,
		0.742944,
		0.64838,
		0.538323,
		0.416289,
		0.28869,
		0.165879,
		0.063557,
		0.004665,
		0.000131,
		0.000604,
		0.00155,
		0.003111,
		0.005442,
		0.008699,
		0.013042,
		0.018622,
		0.02558,
		0.03404,
		0.044105,
		0.055852,
		0.069331,
		0.084562,
		0.101537,
		0.120219,
		0.140548,
		0.162442,
		0.185806,
		0.210532,
		0.236507,
		0.263617,
		0.291752,
		0.320804,
		0.350675,
		0.381271,
		0.412507,
		0.444298,
		0.476563,
		0.509214,
		0.542159,
		0.575293,
		0.608498,
		0.641638,
		0.674555,
		0.707073,
		0.738994,
		0.770102,
		0.800162,
		0.828924,
		0.856131,
		0.881522,
		0.904839,
		0.925841,
		0.944313,
		0.960083,
		0.973047,
		0.983187,
		0.990599,
		0.995521,
		0.998347,
		0.999618,
		0.999972,
		1,
		0.999968,
		0.999493,
		0.997501,
		0.992346,
		0.981968,
		0.964076,
		0.936326,
		0.896506,
		0.842738,
		0.773702,
		0.688896,
		0.588982,
		0.476215,
		0.355015,
		0.232725,
		0.120607,
		0.035182,
		6.4e-005,
		0.000308,
		0.000826,
		0.001726,
		0.003134,
		0.005191,
		0.008047,
		0.011859,
		0.016787,
		0.022989,
		0.03061,
		0.039786,
		0.050629,
		0.063227,
		0.07764,
		0.093895,
		0.111989,
		0.131883,
		0.15351,
		0.176778,
		0.201568,
		0.227748,
		0.255173,
		0.283695,
		0.313165,
		0.343437,
		0.374377,
		0.405858,
		0.437764,
		0.469991,
		0.502441,
		0.535021,
		0.567636,
		0.600193,
		0.632584,
		0.664693,
		0.696388,
		0.727518,
		0.757914,
		0.787387,
		0.815734,
		0.842734,
		0.86816,
		0.891783,
		0.913379,
		0.932739,
		0.949683,
		0.964076,
		0.975845,
		0.984999,
		0.991651,
		0.996042,
		0.998546,
		0.999666,
		0.999976,
		1,
		0.999972,
		0.999561,
		0.997833,
		0.99335,
		0.984305,
		0.968661,
		0.944313,
		0.909232,
		0.861636,
		0.800163,
		0.724081,
		0.63355,
		0.529947,
		0.416289,
		0.297795,
		0.182603,
		0.082738,
		0.015366,
		2.4e-005,
		0.000127,
		0.000361,
		0.000799,
		0.001528,
		0.002653,
		0.004295,
		0.006591,
		0.009688,
		0.013746,
		0.018926,
		0.025391,
		0.0333,
		0.042798,
		0.054013,
		0.067052,
		0.08199,
		0.098873,
		0.11771,
		0.138472,
		0.161098,
		0.18549,
		0.21152,
		0.239036,
		0.267868,
		0.297832,
		0.32874,
		0.360406,
		0.392648,
		0.425297,
		0.458198,
		0.491211,
		0.524209,
		0.557079,
		0.589719,
		0.622029,
		0.653913,
		0.685268,
		0.715985,
		0.745942,
		0.775,
		0.803008,
		0.829796,
		0.855181,
		0.878971,
		0.900973,
		0.920994,
		0.938862,
		0.954428,
		0.967586,
		0.978292,
		0.986574,
		0.992561,
		0.99649,
		0.998717,
		0.999707,
		0.999979,
		1,
		0.999976,
		0.999618,
		0.998111,
		0.994196,
		0.986276,
		0.972543,
		0.951098,
		0.92009,
		0.877839,
		0.82299,
		0.754672,
		0.672708,
		0.577857,
		0.472143,
		0.359265,
		0.245133,
		0.138573,
		0.05223,
		0.003771,
		3.7e-005,
		0.000116,
		0.000283,
		0.000586,
		0.001091,
		0.001882,
		0.003057,
		0.004732,
		0.007038,
		0.010121,
		0.014139,
		0.019257,
		0.025644,
		0.033468,
		0.04289,
		0.054055,
		0.067091,
		0.082099,
		0.099147,
		0.118271,
		0.139462,
		0.162676,
		0.187825,
		0.214785,
		0.243396,
		0.273471,
		0.304801,
		0.337163,
		0.370327,
		0.404065,
		0.438158,
		0.472399,
		0.5066,
		0.540591,
		0.574223,
		0.607367,
		0.639906,
		0.671739,
		0.702768,
		0.732899,
		0.762033,
		0.790063,
		0.816874,
		0.842337,
		0.86631,
		0.888645,
		0.909189,
		0.927789,
		0.944307,
		0.95863,
		0.970681,
		0.980438,
		0.987949,
		0.99335,
		0.996876,
		0.998863,
		0.999741,
		0.999981,
		1,
		0.999979,
		0.999666,
		0.998347,
		0.994912,
		0.987949,
		0.975845,
		0.956891,
		0.929393,
		0.891784,
		0.842738,
		0.781314,
		0.707107,
		0.620445,
		0.522638,
		0.416289,
		0.3057,
		0.197385,
		0.100747,
		0.028941,
		2.1e-005,
		6.2e-005,
		0.000149,
		0.000316,
		0.000608,
		0.001083,
		0.001816,
		0.0029,
		0.004443,
		0.00657,
		0.009423,
		0.013156,
		0.017935,
		0.023933,
		0.031327,
		0.040291,
		0.050989,
		0.063573,
		0.07817,
		0.09488,
		0.11377,
		0.134865,
		0.158149,
		0.183562,
		0.210995,
		0.240301,
		0.271292,
		0.303745,
		0.337415,
		0.372038,
		0.407343,
		0.443057,
		0.478918,
		0.514679,
		0.550115,
		0.585024,
		0.619229,
		0.652581,
		0.684951,
		0.71623,
		0.746323,
		0.775143,
		0.802607,
		0.828631,
		0.853125,
		0.875992,
		0.897131,
		0.916436,
		0.9338,
		0.94913,
		0.962349,
		0.973413,
		0.982326,
		0.989153,
		0.994038,
		0.99721,
		0.998989,
		0.999771,
		0.999984,
		1,
		0.999981,
		0.999707,
		0.998546,
		0.995521,
		0.989376,
		0.978669,
		0.961859,
		0.937398,
		0.903828,
		0.859876,
		0.804566,
		0.737347,
		0.658245,
		0.568056,
		0.468589,
		0.362964,
		0.256014,
		0.154777,
		0.069138,
		0.012657,
		1.6e-005,
		4.5e-005,
		0.00011,
		0.000237,
		0.000464,
		0.000842,
		0.001437,
		0.002331,
		0.003623,
		0.005431,
		0.007888,
		0.011143,
		0.01536,
		0.020712,
		0.027382,
		0.035553,
		0.045407,
		0.057115,
		0.070835,
		0.086698,
		0.104807,
		0.12523,
		0.14799,
		0.173066,
		0.200385,
		0.229828,
		0.261224,
		0.294361,
		0.328987,
		0.364821,
		0.40156,
		0.438887,
		0.476489,
		0.514056,
		0.551297,
		0.587947,
		0.623765,
		0.658546,
		0.692115,
		0.724327,
		0.755064,
		0.784233,
		0.811757,
		0.837568,
		0.861608,
		0.883821,
		0.904151,
		0.92254,
		0.938936,
		0.953294,
		0.965582,
		0.975798,
		0.983977,
		0.990206,
		0.994638,
		0.997501,
		0.999098,
		0.999797,
		0.999985,
		1,
		0.999984,
		0.999741,
		0.998717,
		0.996042,
		0.990599,
		0.981094,
		0.966137,
		0.944313,
		0.914268,
		0.874792,
		0.824905,
		0.763962,
		0.691782,
		0.608789,
		0.516206,
		0.416289,
		0.312626,
		0.210511,
		0.117431,
		0.043674,
		0.003111,
		1.6e-005,
		4.9e-005,
		0.000121,
		0.000261,
		0.000512,
		0.000927,
		0.001579,
		0.002555,
		0.00396,
		0.005917,
		0.008567,
		0.012067,
		0.016587,
		0.02231,
		0.029425,
		0.038125,
		0.0486,
		0.061029,
		0.075577,
		0.092384,
		0.11156,
		0.133176,
		0.15726,
		0.183787,
		0.212684,
		0.243819,
		0.27701,
		0.312024,
		0.348584,
		0.386375,
		0.425058,
		0.464277,
		0.50367,
		0.542884,
		0.581581,
		0.619449,
		0.656209,
		0.691617,
		0.725472,
		0.757607,
		0.787895,
		0.81624,
		0.842572,
		0.866844,
		0.889025,
		0.909096,
		0.927046,
		0.942872,
		0.956583,
		0.968199,
		0.977766,
		0.98536,
		0.991097,
		0.99515,
		0.99775,
		0.999192,
		0.999818,
		0.999987,
		1,
		0.999985,
		0.999771,
		0.998863,
		0.99649,
		0.991651,
		0.983187,
		0.969836,
		0.950308,
		0.92335,
		0.887815,
		0.842738,
		0.787425,
		0.721547,
		0.645261,
		0.559361,
		0.465459,
		0.366214,
		0.265625,
		0.169394,
		0.085389,
		0.024223,
		3.5e-005,
		9.8e-005,
		0.000227,
		0.000466,
		0.000871,
		0.001517,
		0.002495,
		0.003915,
		0.005905,
		0.008611,
		0.012195,
		0.016836,
		0.022723,
		0.030053,
		0.039027,
		0.049843,
		0.062693,
		0.077749,
		0.095162,
		0.115052,
		0.137498,
		0.162535,
		0.190145,
		0.220253,
		0.252728,
		0.287379,
		0.323959,
		0.362172,
		0.401677,
		0.442101,
		0.48305,
		0.524119,
		0.564909,
		0.605034,
		0.644135,
		0.681888,
		0.71801,
		0.752265,
		0.784462,
		0.814459,
		0.842154,
		0.867484,
		0.890421,
		0.910961,
		0.929125,
		0.944951,
		0.958497,
		0.969837,
		0.979067,
		0.986312,
		0.991729,
		0.995519,
		0.99793,
		0.999258,
		0.999832,
		0.999987,
		1,
		0.999987,
		0.999797,
		0.998989,
		0.996876,
		0.992561,
		0.984999,
		0.973047,
		0.955526,
		0.931276,
		0.899218,
		0.858415,
		0.808146,
		0.747985,
		0.677898,
		0.598361,
		0.510503,
		0.416289,
		0.318742,
		0.222223,
		0.132788,
		0.058623,
		0.010606,
		1.7e-005,
		6e-005,
		0.00016,
		0.000359,
		0.000715,
		0.001301,
		0.002212,
		0.003559,
		0.005473,
		0.008105,
		0.011623,
		0.016209,
		0.022059,
		0.029379,
		0.038377,
		0.049261,
		0.062232,
		0.077474,
		0.095151,
		0.115392,
		0.138292,
		0.163895,
		0.192194,
		0.223122,
		0.256552,
		0.292291,
		0.330088,
		0.369632,
		0.410564,
		0.452485,
		0.494965,
		0.537563,
		0.579832,
		0.621342,
		0.661683,
		0.700485,
		0.737423,
		0.772221,
		0.80466,
		0.834575,
		0.861858,
		0.886448,
		0.908331,
		0.927533,
		0.944112,
		0.958161,
		0.969798,
		0.979167,
		0.986442,
		0.991829,
		0.995567,
		0.997932,
		0.999235,
		0.999806,
		0.999972,
		0.999994,
		0.999987,
		0.999818,
		0.999098,
		0.99721,
		0.99335,
		0.986574,
		0.975845,
		0.960083,
		0.938217,
		0.909232,
		0.87223,
		0.826481,
		0.771497,
		0.707107,
		0.63355,
		0.551596,
		0.462682,
		0.369091,
		0.27417,
		0.182603,
		0.100747,
		0.037057,
		0.00261,
		2.7e-005,
		9e-005,
		0.000234,
		0.000513,
		0.001001,
		0.00179,
		0.002991,
		0.004739,
		0.007186,
		0.010502,
		0.014875,
		0.020506,
		0.027607,
		0.036394,
		0.047085,
		0.05989,
		0.075006,
		0.092608,
		0.112844,
		0.135819,
		0.161595,
		0.190179,
		0.221518,
		0.255494,
		0.291922,
		0.330552,
		0.371068,
		0.4131,
		0.456229,
		0.5,
		0.543936,
		0.587553,
		0.630373,
		0.67194,
		0.711837,
		0.749691,
		0.785187,
		0.818069,
		0.84815,
		0.875305,
		0.899472,
		0.92065,
		0.938887,
		0.954283,
		0.966979,
		0.977154,
		0.985025,
		0.990841,
		0.994885,
		0.997471,
		0.998935,
		0.999627,
		0.999877,
		0.999952,
		0.999972,
		0.999832,
		0.999192,
		0.997501,
		0.994038,
		0.987949,
		0.978292,
		0.964076,
		0.944313,
		0.918052,
		0.884434,
		0.842738,
		0.792437,
		0.73326,
		0.665277,
		0.588982,
		0.505413,
		0.416289,
		0.324181,
		0.232725,
		0.146884,
		0.073275,
		0.02057,
		3.7e-005,
		0.000124,
		0.000317,
		0.000685,
		0.001317,
		0.002323,
		0.003836,
		0.006008,
		0.009012,
		0.013038,
		0.018292,
		0.02499,
		0.033356,
		0.043615,
		0.055989,
		0.070687,
		0.087897,
		0.107782,
		0.130466,
		0.156029,
		0.184495,
		0.21583,
		0.249933,
		0.286631,
		0.325684,
		0.366781,
		0.409549,
		0.453558,
		0.498335,
		0.543376,
		0.588164,
		0.632181,
		0.674927,
		0.715937,
		0.754792,
		0.791133,
		0.824668,
		0.855177,
		0.882517,
		0.90662,
		0.92749,
		0.945195,
		0.95987,
		0.971702,
		0.980929,
		0.987832,
		0.992731,
		0.995977,
		0.997943,
		0.999005,
		0.999515,
		0.999758,
		0.999877,
		0.999806,
		0.999258,
		0.99775,
		0.994638,
		0.989153,
		0.980438,
		0.967586,
		0.949683,
		0.925841,
		0.895243,
		0.857183,
		0.811114,
		0.756699,
		0.693878,
		0.622942,
		0.544621,
		0.460201,
		0.371655,
		0.281814,
		0.19457,
		0.11513,
		0.05033,
		0.009016,
		4.9e-005,
		0.000161,
		0.000407,
		0.00087,
		0.001654,
		0.00289,
		0.004727,
		0.007338,
		0.010914,
		0.015662,
		0.021804,
		0.029569,
		0.039191,
		0.0509,
		0.064916,
		0.081444,
		0.10066,
		0.122709,
		0.147689,
		0.175647,
		0.206571,
		0.240381,
		0.276925,
		0.315979,
		0.357244,
		0.400354,
		0.444878,
		0.490336,
		0.536205,
		0.581942,
		0.626996,
		0.67083,
		0.712934,
		0.752845,
		0.790158,
		0.824541,
		0.855742,
		0.883593,
		0.908012,
		0.929002,
		0.946649,
		0.961111,
		0.972616,
		0.981452,
		0.987957,
		0.992508,
		0.995509,
		0.997368,
		0.998467,
		0.999123,
		0.999515,
		0.999627,
		0.999235,
		0.99793,
		0.99515,
		0.990206,
		0.982326,
		0.970681,
		0.954428,
		0.932739,
		0.904839,
		0.870045,
		0.827802,
		0.777728,
		0.71967,
		0.653763,
		0.580504,
		0.500843,
		0.416289,
		0.32905,
		0.242186,
		0.159817,
		0.08736,
		0.031834,
		0.002221,
		1.2e-005,
		6.2e-005,
		0.000201,
		0.000503,
		0.001064,
		0.002008,
		0.003479,
		0.005648,
		0.008703,
		0.012854,
		0.018324,
		0.025346,
		0.034161,
		0.045007,
		0.058116,
		0.073706,
		0.09197,
		0.11307,
		0.137128,
		0.164213,
		0.194338,
		0.227449,
		0.26342,
		0.302051,
		0.343062,
		0.386103,
		0.430753,
		0.47653,
		0.522909,
		0.569328,
		0.615215,
		0.659998,
		0.703131,
		0.744109,
		0.782486,
		0.817888,
		0.850026,
		0.878705,
		0.903823,
		0.925377,
		0.943456,
		0.958238,
		0.969981,
		0.979009,
		0.985703,
		0.990479,
		0.993768,
		0.995983,
		0.997474,
		0.998467,
		0.999005,
		0.998935,
		0.997932,
		0.995519,
		0.991097,
		0.983977,
		0.973413,
		0.95863,
		0.938862,
		0.913379,
		0.881522,
		0.842738,
		0.796619,
		0.742944,
		0.681729,
		0.613292,
		0.538323,
		0.457972,
		0.373956,
		0.28869,
		0.205443,
		0.128533,
		0.063557,
		0.017685,
		1.5e-005,
		7.5e-005,
		0.000243,
		0.000602,
		0.001265,
		0.00237,
		0.00408,
		0.00658,
		0.010079,
		0.014799,
		0.020976,
		0.028856,
		0.038685,
		0.050704,
		0.065143,
		0.082212,
		0.102092,
		0.124925,
		0.150806,
		0.179776,
		0.211808,
		0.246807,
		0.284601,
		0.324939,
		0.367494,
		0.411863,
		0.457578,
		0.504116,
		0.550912,
		0.597379,
		0.642924,
		0.686971,
		0.72898,
		0.768466,
		0.805015,
		0.838304,
		0.868102,
		0.894285,
		0.916836,
		0.935839,
		0.951479,
		0.964028,
		0.973833,
		0.981295,
		0.986844,
		0.990908,
		0.993865,
		0.995983,
		0.997368,
		0.997943,
		0.997471,
		0.995567,
		0.991729,
		0.98536,
		0.975798,
		0.962349,
		0.944307,
		0.920994,
		0.891783,
		0.856131,
		0.813614,
		0.763962,
		0.707107,
		0.643226,
		0.572806,
		0.496716,
		0.416289,
		0.333432,
		0.250747,
		0.171689,
		0.100747,
		0.043674,
		0.007758,
		1.7e-005,
		8.9e-005,
		0.000285,
		0.000703,
		0.001468,
		0.002735,
		0.004682,
		0.007511,
		0.011443,
		0.016717,
		0.023578,
		0.03228,
		0.043071,
		0.056194,
		0.071873,
		0.090306,
		0.111657,
		0.136047,
		0.163544,
		0.194153,
		0.22781,
		0.264379,
		0.303641,
		0.3453,
		0.388981,
		0.434238,
		0.480563,
		0.527398,
		0.574154,
		0.620228,
		0.665023,
		0.707972,
		0.748558,
		0.786334,
		0.820937,
		0.852106,
		0.879686,
		0.903638,
		0.924032,
		0.941046,
		0.95495,
		0.96609,
		0.974863,
		0.981682,
		0.986931,
		0.990908,
		0.993768,
		0.995509,
		0.995977,
		0.994885,
		0.991829,
		0.986312,
		0.977766,
		0.965582,
		0.94913,
		0.927789,
		0.900973,
		0.86816,
		0.828924,
		0.782964,
		0.730146,
		0.670542,
		0.604481,
		0.532609,
		0.455958,
		0.376031,
		0.294906,
		0.215353,
		0.140991,
		0.076461,
		0.027641,
		0.001913,
		2e-005,
		0.000103,
		0.000328,
		0.000804,
		0.001671,
		0.003097,
		0.005276,
		0.008424,
		0.012776,
		0.018579,
		0.02609,
		0.035565,
		0.047255,
		0.061398,
		0.078208,
		0.097872,
		0.120532,
		0.146287,
		0.175172,
		0.207159,
		0.242148,
		0.279959,
		0.320332,
		0.362928,
		0.407332,
		0.453061,
		0.499578,
		0.546301,
		0.59263,
		0.637959,
		0.681705,
		0.723327,
		0.762347,
		0.798372,
		0.831105,
		0.86036,
		0.886069,
		0.908275,
		0.927134,
		0.942893,
		0.955875,
		0.966438,
		0.974941,
		0.981682,
		0.986844,
		0.990479,
		0.992508,
		0.992731,
		0.990841,
		0.986442,
		0.979067,
		0.968199,
		0.953294,
		0.9338,
		0.909189,
		0.878971,
		0.842734,
		0.800162,
		0.751076,
		0.69547,
		0.63355,
		0.565786,
		0.492972,
		0.416289,
		0.337397,
		0.258527,
		0.182603,
		0.113386,
		0.055645,
		0.015366,
		2.3e-005,
		0.000117,
		0.000371,
		0.000904,
		0.00187,
		0.003451,
		0.005853,
		0.009307,
		0.014057,
		0.02036,
		0.028478,
		0.038668,
		0.051181,
		0.066246,
		0.084068,
		0.104814,
		0.128609,
		0.155521,
		0.185558,
		0.218656,
		0.254678,
		0.293406,
		0.334543,
		0.377712,
		0.422467,
		0.468297,
		0.514646,
		0.560924,
		0.606532,
		0.650881,
		0.693416,
		0.733637,
		0.771123,
		0.805547,
		0.836688,
		0.864439,
		0.88881,
		0.909914,
		0.927957,
		0.943204,
		0.955945,
		0.966438,
		0.974863,
		0.981295,
		0.985703,
		0.987957,
		0.987832,
		0.985025,
		0.979167,
		0.969837,
		0.956583,
		0.938936,
		0.916436,
		0.888645,
		0.855181,
		0.815734,
		0.770102,
		0.718225,
		0.660215,
		0.596407,
		0.527402,
		0.45413,
		0.377913,
		0.300551,
		0.224414,
		0.152563,
		0.088884,
		0.038254,
		0.006746,
		2.6e-005,
		0.000131,
		0.000412,
		0.001001,
		0.002063,
		0.003792,
		0.006407,
		0.010148,
		0.015271,
		0.022037,
		0.030711,
		0.041551,
		0.0548,
		0.070681,
		0.089384,
		0.111057,
		0.135801,
		0.163659,
		0.194608,
		0.228551,
		0.265316,
		0.304652,
		0.346229,
		0.389643,
		0.434424,
		0.480048,
		0.52595,
		0.571549,
		0.616261,
		0.65953,
		0.700845,
		0.739762,
		0.775926,
		0.80908,
		0.839077,
		0.865876,
		0.889535,
		0.910189,
		0.928019,
		0.943204,
		0.955875,
		0.96609,
		0.973833,
		0.979009,
		0.981452,
		0.980929,
		0.977154,
		0.969798,
		0.958497,
		0.942872,
		0.92254,
		0.897131,
		0.86631,
		0.829796,
		0.787387,
		0.738994,
		0.684668,
		0.624639,
		0.559361,
		0.489559,
		0.416289,
		0.341002,
		0.265625,
		0.192655,
		0.125272,
		0.067471,
		0.024223,
		0.001665,
		2.9e-005,
		0.000144,
		0.000452,
		0.001095,
		0.002247,
		0.004116,
		0.00693,
		0.010938,
		0.016402,
		0.023589,
		0.032763,
		0.044179,
		0.058074,
		0.074658,
		0.094104,
		0.116543,
		0.142051,
		0.170644,
		0.202269,
		0.236802,
		0.274039,
		0.313702,
		0.355436,
		0.398819,
		0.443369,
		0.48856,
		0.533837,
		0.578638,
		0.622414,
		0.664653,
		0.704898,
		0.742769,
		0.777975,
		0.810321,
		0.839707,
		0.866116,
		0.88959,
		0.910189,
		0.927957,
		0.942893,
		0.95495,
		0.964028,
		0.969981,
		0.972616,
		0.971702,
		0.966979,
		0.958161,
		0.944951,
		0.927046,
		0.904151,
		0.875992,
		0.842337,
		0.803008,
		0.757914,
		0.707073,
		0.650647,
		0.588979,
		0.522637,
		0.452462,
		0.379627,
		0.3057,
		0.232725,
		0.163312,
		0.100747,
		0.049119,
		0.013475,
		3.1e-005,
		0.000157,
		0.000491,
		0.001183,
		0.002421,
		0.004419,
		0.007416,
		0.011667,
		0.01744,
		0.025001,
		0.034615,
		0.04653,
		0.060974,
		0.078143,
		0.098195,
		0.121239,
		0.147328,
		0.176453,
		0.208535,
		0.243424,
		0.280893,
		0.320642,
		0.362304,
		0.405447,
		0.449592,
		0.494223,
		0.53881,
		0.582824,
		0.62576,
		0.667159,
		0.706621,
		0.743821,
		0.778514,
		0.810528,
		0.839755,
		0.866116,
		0.889535,
		0.909914,
		0.927134,
		0.941046,
		0.951479,
		0.958238,
		0.961111,
		0.95987,
		0.954283,
		0.944112,
		0.929125,
		0.909096,
		0.883821,
		0.853125,
		0.816874,
		0.775,
		0.727518,
		0.674555,
		0.616381,
		0.553448,
		0.486434,
		0.416289,
		0.344293,
		0.272124,
		0.201932,
		0.136427,
		0.078987,
		0.033782,
		0.00592,
		3.4e-005,
		0.000169,
		0.000526,
		0.001266,
		0.002582,
		0.004698,
		0.00786,
		0.012329,
		0.018373,
		0.026261,
		0.036251,
		0.048585,
		0.063481,
		0.081119,
		0.101641,
		0.125134,
		0.151632,
		0.181102,
		0.213445,
		0.248488,
		0.285991,
		0.325642,
		0.367068,
		0.409844,
		0.453504,
		0.497557,
		0.541506,
		0.584866,
		0.627179,
		0.668034,
		0.707073,
		0.743996,
		0.778554,
		0.810528,
		0.839707,
		0.865876,
		0.88881,
		0.908275,
		0.924032,
		0.935839,
		0.943456,
		0.946649,
		0.945195,
		0.938887,
		0.927533,
		0.910961,
		0.889025,
		0.861608,
		0.828631,
		0.790063,
		0.745942,
		0.696388,
		0.641638,
		0.582068,
		0.518237,
		0.450927,
		0.381192,
		0.310414,
		0.24037,
		0.173305,
		0.112019,
		0.059973,
		0.021402,
		0.001462,
		3.6e-005,
		0.00018,
		0.00056,
		0.001342,
		0.002729,
		0.004951,
		0.008259,
		0.012918,
		0.019196,
		0.027359,
		0.037661,
		0.050336,
		0.065587,
		0.083582,
		0.104444,
		0.128245,
		0.154997,
		0.18465,
		0.217088,
		0.252127,
		0.289518,
		0.32895,
		0.370055,
		0.412426,
		0.455619,
		0.499179,
		0.542647,
		0.585577,
		0.62755,
		0.668179,
		0.707107,
		0.743996,
		0.778514,
		0.810321,
		0.839077,
		0.864439,
		0.886069,
		0.903638,
		0.916836,
		0.925377,
		0.929002,
		0.92749,
		0.92065,
		0.908331,
		0.890421,
		0.866844,
		0.837568,
		0.802607,
		0.762033,
		0.715985,
		0.664693,
		0.608498,
		0.547885,
		0.483514,
		0.416269,
		0.347303,
		0.278094,
		0.210511,
		0.146884,
		0.090096,
		0.043674,
		0.011912,
		3.8e-005,
		0.00019,
		0.00059,
		0.00141,
		0.00286,
		0.005175,
		0.008609,
		0.01343,
		0.019903,
		0.028292,
		0.038844,
		0.051781,
		0.067296,
		0.085543,
		0.10663,
		0.130612,
		0.157487,
		0.187192,
		0.219602,
		0.254529,
		0.291724,
		0.330886,
		0.371666,
		0.413678,
		0.456511,
		0.499742,
		0.542944,
		0.585694,
		0.627578,
		0.668179,
		0.707073,
		0.743821,
		0.777975,
		0.80908,
		0.836688,
		0.86036,
		0.879686,
		0.894285,
		0.903823,
		0.908012,
		0.90662,
		0.899472,
		0.886448,
		0.867484,
		0.842572,
		0.811757,
		0.775143,
		0.732899,
		0.685268,
		0.632584,
		0.575293,
		0.513985,
		0.449423,
		0.38259,
		0.314732,
		0.247418,
		0.182602,
		0.122695,
		0.070648,
		0.030049,
		0.005236,
		4e-005,
		0.000199,
		0.000617,
		0.001471,
		0.002975,
		0.005369,
		0.00891,
		0.013864,
		0.020495,
		0.029061,
		0.039801,
		0.052929,
		0.068626,
		0.087033,
		0.108244,
		0.132304,
		0.159201,
		0.188866,
		0.22117,
		0.255931,
		0.292911,
		0.331824,
		0.372344,
		0.414112,
		0.456743,
		0.499835,
		0.542966,
		0.585694,
		0.62755,
		0.668034,
		0.706621,
		0.742769,
		0.775926,
		0.805547,
		0.831105,
		0.852106,
		0.868102,
		0.878705,
		0.883593,
		0.882517,
		0.875305,
		0.861858,
		0.842154,
		0.81624,
		0.784233,
		0.746323,
		0.702768,
		0.653913,
		0.600193,
		0.542159,
		0.480503,
		0.416091,
		0.349999,
		0.283568,
		0.218452,
		0.156684,
		0.100747,
		0.053654,
		0.019046,
		0.001294,
		4.2e-005,
		0.000208,
		0.00064,
		0.001523,
		0.003074,
		0.005534,
		0.009162,
		0.014221,
		0.020973,
		0.029671,
		0.040544,
		0.053799,
		0.069605,
		0.088095,
		0.109353,
		0.133415,
		0.160268,
		0.18984,
		0.222012,
		0.256608,
		0.293409,
		0.332147,
		0.37252,
		0.414183,
		0.456761,
		0.499835,
		0.542944,
		0.585577,
		0.627179,
		0.667159,
		0.704898,
		0.739762,
		0.771123,
		0.798372,
		0.820937,
		0.838304,
		0.850026,
		0.855742,
		0.855177,
		0.84815,
		0.834575,
		0.814459,
		0.787895,
		0.755064,
		0.71623,
		0.671739,
		0.622029,
		0.567636,
		0.509214,
		0.447554,
		0.383622,
		0.31859,
		0.253886,
		0.191247,
		0.132781,
		0.081038,
		0.039085,
		0.010606,
		4.4e-005,
		0.000215,
		0.000661,
		0.001568,
		0.003156,
		0.005669,
		0.009364,
		0.014503,
		0.021344,
		0.030132,
		0.041091,
		0.054418,
		0.070277,
		0.088791,
		0.110041,
		0.134061,
		0.160839,
		0.19031,
		0.222364,
		0.256841,
		0.293537,
		0.332201,
		0.372533,
		0.414183,
		0.456743,
		0.499742,
		0.542647,
		0.584866,
		0.62576,
		0.664653,
		0.700845,
		0.733637,
		0.762347,
		0.786334,
		0.805015,
		0.817888,
		0.824541,
		0.824668,
		0.818069,
		0.80466,
		0.784462,
		0.757607,
		0.724327,
		0.684951,
		0.639906,
		0.589719,
		0.535021,
		0.476563,
		0.415238,
		0.352107,
		0.288436,
		0.225742,
		0.165839,
		0.110907,
		0.063554,
		0.026901,
		0.004665,
		4.5e-005,
		0.000221,
		0.000677,
		0.001604,
		0.003222,
		0.005776,
		0.009521,
		0.014716,
		0.021616,
		0.030459,
		0.041464,
		0.054822,
		0.070692,
		0.089194,
		0.110408,
		0.134371,
		0.161077,
		0.190471,
		0.222455,
		0.25688,
		0.293547,
		0.332201,
		0.37252,
		0.414112,
		0.456511,
		0.499179,
		0.541506,
		0.582824,
		0.622414,
		0.65953,
		0.693416,
		0.723327,
		0.748558,
		0.768466,
		0.782486,
		0.790158,
		0.791133,
		0.785187,
		0.772221,
		0.752265,
		0.725472,
		0.692115,
		0.652581,
		0.607367,
		0.557079,
		0.502441,
		0.444298,
		0.38364,
		0.321622,
		0.259602,
		0.199181,
		0.142252,
		0.091062,
		0.048276,
		0.017057,
		0.001154,
		4.6e-005,
		0.000225,
		0.000691,
		0.001633,
		0.003273,
		0.005856,
		0.009635,
		0.014867,
		0.021801,
		0.030672,
		0.041695,
		0.055057,
		0.070914,
		0.089387,
		0.110561,
		0.134478,
		0.161138,
		0.190498,
		0.222461,
		0.25688,
		0.293537,
		0.332147,
		0.372344,
		0.413678,
		0.455619,
		0.497557,
		0.53881,
		0.578638,
		0.616261,
		0.650881,
		0.681705,
		0.707972,
		0.72898,
		0.744109,
		0.752845,
		0.754792,
		0.749691,
		0.737423,
		0.71801,
		0.691617,
		0.658546,
		0.619229,
		0.574223,
		0.524209,
		0.469991,
		0.412507,
		0.352837,
		0.292232,
		0.232137,
		0.174237,
		0.120501,
		0.07324,
		0.035173,
		0.009502,
		4.7e-005,
		0.000229,
		0.000701,
		0.001654,
		0.003311,
		0.005913,
		0.009713,
		0.014965,
		0.021915,
		0.030795,
		0.041817,
		0.055169,
		0.071006,
		0.089454,
		0.1106,
		0.134495,
		0.161143,
		0.190498,
		0.222455,
		0.256841,
		0.293409,
		0.331824,
		0.371666,
		0.412426,
		0.453504,
		0.494223,
		0.533837,
		0.571549,
		0.606532,
		0.637959,
		0.665023,
		0.686971,
		0.703131,
		0.712934,
		0.715937,
		0.711837,
		0.700485,
		0.681888,
		0.656209,
		0.623765,
		0.585024,
		0.540591,
		0.491211,
		0.437764,
		0.381271,
		0.322902,
		0.263995,
		0.206085,
		0.150937,
		0.100594,
		0.057429,
		0.024212,
		0.004181,
		4.7e-005,
		0.000232,
		0.000709,
		0.001669,
		0.003336,
		0.00595,
		0.009761,
		0.015022,
		0.021976,
		0.030854,
		0.041868,
		0.055207,
		0.071029,
		0.089464,
		0.110603,
		0.134495,
		0.161138,
		0.190471,
		0.222364,
		0.256608,
		0.292911,
		0.330886,
		0.370055,
		0.409844,
		0.449592,
		0.48856,
		0.52595,
		0.560924,
		0.59263,
		0.620228,
		0.642924,
		0.659998,
		0.67083,
		0.674927,
		0.67194,
		0.661683,
		0.644135,
		0.619449,
		0.587947,
		0.550115,
		0.5066,
		0.458198,
		0.405858,
		0.350675,
		0.293903,
		0.236966,
		0.181483,
		0.129299,
		0.082529,
		0.043608,
		0.015352,
		0.001034,
		4.8e-005,
		0.000234,
		0.000714,
		0.001679,
		0.003351,
		0.005971,
		0.009787,
		0.015049,
		0.022001,
		0.030874,
		0.041881,
		0.055213,
		0.071031,
		0.089464,
		0.1106,
		0.134478,
		0.161077,
		0.19031,
		0.222012,
		0.255931,
		0.291724,
		0.32895,
		0.367068,
		0.405447,
		0.443369,
		0.480048,
		0.514646,
		0.546301,
		0.574154,
		0.597379,
		0.615215,
		0.626996,
		0.632181,
		0.630373,
		0.621342,
		0.605034,
		0.581581,
		0.551297,
		0.514679,
		0.472399,
		0.425297,
		0.374377,
		0.320804,
		0.265908,
		0.211195,
		0.158365,
		0.109339,
		0.066304,
		0.031753,
		0.008551,
		4.8e-005,
		0.000235,
		0.000717,
		0.001685,
		0.00336,
		0.005981,
		0.009797,
		0.015058,
		0.022007,
		0.030877,
		0.041882,
		0.055213,
		0.071029,
		0.089454,
		0.110561,
		0.134371,
		0.160839,
		0.18984,
		0.22117,
		0.254529,
		0.289518,
		0.325642,
		0.362304,
		0.398819,
		0.434424,
		0.468297,
		0.499578,
		0.527398,
		0.550912,
		0.569328,
		0.581942,
		0.588164,
		0.587553,
		0.579832,
		0.564909,
		0.542884,
		0.514056,
		0.478918,
		0.438158,
		0.392648,
		0.343437,
		0.291752,
		0.238991,
		0.186735,
		0.136762,
		0.091069,
		0.05191,
		0.021838,
		0.003761,
		4.8e-005,
		0.000236,
		0.000719,
		0.001687,
		0.003363,
		0.005984,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041881,
		0.055207,
		0.071006,
		0.089387,
		0.110408,
		0.134061,
		0.160268,
		0.188866,
		0.219602,
		0.252127,
		0.285991,
		0.320642,
		0.355436,
		0.389643,
		0.422467,
		0.453061,
		0.480563,
		0.504116,
		0.522909,
		0.536205,
		0.543376,
		0.543936,
		0.537563,
		0.524119,
		0.50367,
		0.476489,
		0.443057,
		0.404065,
		0.360406,
		0.313165,
		0.263617,
		0.213225,
		0.163637,
		0.116705,
		0.074497,
		0.039335,
		0.013828,
		0.00093,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022007,
		0.030874,
		0.041868,
		0.055169,
		0.070914,
		0.089194,
		0.110041,
		0.133415,
		0.159201,
		0.187192,
		0.217088,
		0.248488,
		0.280893,
		0.313702,
		0.346229,
		0.377712,
		0.407332,
		0.434238,
		0.457578,
		0.47653,
		0.490336,
		0.498335,
		0.5,
		0.494965,
		0.48305,
		0.464277,
		0.438887,
		0.407343,
		0.370327,
		0.32874,
		0.283695,
		0.236507,
		0.188689,
		0.141955,
		0.098223,
		0.059631,
		0.028563,
		0.007687,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015058,
		0.022001,
		0.030854,
		0.041817,
		0.055057,
		0.070692,
		0.088791,
		0.109353,
		0.132304,
		0.157487,
		0.18465,
		0.213445,
		0.243424,
		0.274039,
		0.304652,
		0.334543,
		0.362928,
		0.388981,
		0.411863,
		0.430753,
		0.444878,
		0.453558,
		0.456229,
		0.452485,
		0.442101,
		0.425058,
		0.40156,
		0.372038,
		0.337163,
		0.297832,
		0.255173,
		0.210532,
		0.165466,
		0.121743,
		0.081344,
		0.046472,
		0.019574,
		0.003372,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005984,
		0.009797,
		0.015049,
		0.021976,
		0.030795,
		0.041695,
		0.054822,
		0.070277,
		0.088095,
		0.108244,
		0.130612,
		0.154997,
		0.181102,
		0.208535,
		0.236802,
		0.265316,
		0.293406,
		0.320332,
		0.3453,
		0.367494,
		0.386103,
		0.400354,
		0.409549,
		0.4131,
		0.410564,
		0.401677,
		0.386375,
		0.364821,
		0.337415,
		0.304801,
		0.267868,
		0.227748,
		0.185806,
		0.143636,
		0.103053,
		0.066092,
		0.035019,
		0.01234,
		0.000831,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003363,
		0.005981,
		0.009787,
		0.015022,
		0.021915,
		0.030672,
		0.041464,
		0.054418,
		0.069605,
		0.087033,
		0.10663,
		0.128245,
		0.151632,
		0.176453,
		0.202269,
		0.228551,
		0.254678,
		0.279959,
		0.303641,
		0.324939,
		0.343062,
		0.357244,
		0.366781,
		0.371068,
		0.369632,
		0.362172,
		0.348584,
		0.328987,
		0.303745,
		0.273471,
		0.239036,
		0.201568,
		0.162442,
		0.123278,
		0.085931,
		0.052485,
		0.025261,
		0.006823,
		4.9e-005,
		0.000237,
		0.000719,
		0.001687,
		0.00336,
		0.005971,
		0.009761,
		0.014965,
		0.021801,
		0.030459,
		0.041091,
		0.053799,
		0.068626,
		0.085543,
		0.104444,
		0.125134,
		0.147328,
		0.170644,
		0.194608,
		0.218656,
		0.242148,
		0.264379,
		0.284601,
		0.302051,
		0.315979,
		0.325684,
		0.330552,
		0.330088,
		0.323959,
		0.312024,
		0.294361,
		0.271292,
		0.243396,
		0.21152,
		0.176778,
		0.140548,
		0.104466,
		0.070417,
		0.040531,
		0.017177,
		0.002974,
		4.9e-005,
		0.000237,
		0.000719,
		0.001685,
		0.003351,
		0.00595,
		0.009713,
		0.014867,
		0.021616,
		0.030132,
		0.040544,
		0.052929,
		0.067296,
		0.083582,
		0.101641,
		0.121239,
		0.142051,
		0.163659,
		0.185558,
		0.207159,
		0.22781,
		0.246807,
		0.26342,
		0.276925,
		0.286631,
		0.291922,
		0.292291,
		0.287379,
		0.27701,
		0.261224,
		0.240301,
		0.214785,
		0.18549,
		0.15351,
		0.120219,
		0.087259,
		0.05654,
		0.030225,
		0.010731,
		0.000727,
		4.9e-005,
		0.000236,
		0.000717,
		0.001679,
		0.003336,
		0.005913,
		0.009635,
		0.014716,
		0.021344,
		0.029671,
		0.039801,
		0.051781,
		0.065587,
		0.081119,
		0.098195,
		0.116543,
		0.135801,
		0.155521,
		0.175172,
		0.194153,
		0.211808,
		0.227449,
		0.240381,
		0.249933,
		0.255494,
		0.256552,
		0.252728,
		0.243819,
		0.229828,
		0.210995,
		0.187825,
		0.161098,
		0.131883,
		0.101537,
		0.071706,
		0.044312,
		0.021549,
		0.005873,
		4.8e-005,
		0.000235,
		0.000714,
		0.001669,
		0.003311,
		0.005856,
		0.009521,
		0.014503,
		0.020973,
		0.029061,
		0.038844,
		0.050336,
		0.063481,
		0.078143,
		0.094104,
		0.111057,
		0.128609,
		0.146287,
		0.163544,
		0.179776,
		0.194338,
		0.206571,
		0.21583,
		0.221518,
		0.223122,
		0.220253,
		0.212684,
		0.200385,
		0.183562,
		0.162676,
		0.138472,
		0.111989,
		0.084562,
		0.057832,
		0.033727,
		0.014463,
		0.00253,
		4.8e-005,
		0.000234,
		0.000709,
		0.001654,
		0.003273,
		0.005776,
		0.009364,
		0.014221,
		0.020495,
		0.028292,
		0.037661,
		0.048585,
		0.060974,
		0.074658,
		0.089384,
		0.104814,
		0.120532,
		0.136047,
		0.150806,
		0.164213,
		0.175647,
		0.184495,
		0.190179,
		0.192194,
		0.190145,
		0.183787,
		0.173066,
		0.158149,
		0.139462,
		0.11771,
		0.093895,
		0.069331,
		0.045641,
		0.024756,
		0.008906,
		0.000611,
		4.8e-005,
		0.000232,
		0.000701,
		0.001633,
		0.003222,
		0.005669,
		0.009162,
		0.013864,
		0.019903,
		0.027359,
		0.036251,
		0.04653,
		0.058074,
		0.070681,
		0.084068,
		0.097872,
		0.111657,
		0.124925,
		0.137128,
		0.147689,
		0.156029,
		0.161595,
		0.163895,
		0.162535,
		0.15726,
		0.14799,
		0.134865,
		0.118271,
		0.098873,
		0.07764,
		0.055852,
		0.035114,
		0.01735,
		0.004798,
		4.7e-005,
		0.000229,
		0.000691,
		0.001604,
		0.003156,
		0.005534,
		0.00891,
		0.01343,
		0.019196,
		0.026261,
		0.034615,
		0.044179,
		0.0548,
		0.066246,
		0.078208,
		0.090306,
		0.102092,
		0.11307,
		0.122709,
		0.130466,
		0.135819,
		0.138292,
		0.137498,
		0.133176,
		0.12523,
		0.11377,
		0.099147,
		0.08199,
		0.063227,
		0.044105,
		0.026203,
		0.011432,
		0.002032,
		4.7e-005,
		0.000225,
		0.000677,
		0.001568,
		0.003074,
		0.005369,
		0.008609,
		0.012918,
		0.018373,
		0.025001,
		0.032763,
		0.041551,
		0.051181,
		0.061398,
		0.071873,
		0.082212,
		0.09197,
		0.10066,
		0.107782,
		0.112844,
		0.115392,
		0.115052,
		0.11156,
		0.104807,
		0.09488,
		0.082099,
		0.067052,
		0.050629,
		0.03404,
		0.018834,
		0.006903,
		0.000482,
		4.6e-005,
		0.000221,
		0.000661,
		0.001523,
		0.002975,
		0.005175,
		0.008259,
		0.012329,
		0.01744,
		0.023589,
		0.030711,
		0.038668,
		0.047255,
		0.056194,
		0.065143,
		0.073706,
		0.081444,
		0.087897,
		0.092608,
		0.095151,
		0.095162,
		0.092384,
		0.086698,
		0.07817,
		0.067091,
		0.054013,
		0.039786,
		0.02558,
		0.012909,
		0.003642,
		4.5e-005,
		0.000215,
		0.00064,
		0.001471,
		0.00286,
		0.004951,
		0.00786,
		0.011667,
		0.016402,
		0.022037,
		0.028478,
		0.035565,
		0.043071,
		0.050704,
		0.058116,
		0.064916,
		0.070687,
		0.075006,
		0.077474,
		0.077749,
		0.075577,
		0.070835,
		0.063573,
		0.054055,
		0.042798,
		0.03061,
		0.018622,
		0.008309,
		0.001509,
		4.4e-005,
		0.000208,
		0.000617,
		0.00141,
		0.002729,
		0.004698,
		0.007416,
		0.010938,
		0.015271,
		0.02036,
		0.02609,
		0.03228,
		0.038685,
		0.045007,
		0.0509,
		0.055989,
		0.05989,
		0.062232,
		0.062693,
		0.061029,
		0.057115,
		0.050989,
		0.04289,
		0.0333,
		0.022989,
		0.013042,
		0.004895,
		0.000349,
		4.2e-005,
		0.000199,
		0.00059,
		0.001342,
		0.002582,
		0.004419,
		0.00693,
		0.010148,
		0.014057,
		0.018579,
		0.023578,
		0.028856,
		0.034161,
		0.039191,
		0.043615,
		0.047085,
		0.049261,
		0.049843,
		0.0486,
		0.045407,
		0.040291,
		0.033468,
		0.025391,
		0.016787,
		0.008699,
		0.002517,
		4e-005,
		0.00019,
		0.00056,
		0.001266,
		0.002421,
		0.004116,
		0.006407,
		0.009307,
		0.012776,
		0.016717,
		0.020976,
		0.025346,
		0.029569,
		0.033356,
		0.036394,
		0.038377,
		0.039027,
		0.038125,
		0.035553,
		0.031327,
		0.025644,
		0.018926,
		0.011859,
		0.005442,
		0.001015,
		3.8e-005,
		0.00018,
		0.000526,
		0.001183,
		0.002247,
		0.003792,
		0.005853,
		0.008424,
		0.011443,
		0.014799,
		0.018324,
		0.021804,
		0.02499,
		0.027607,
		0.029379,
		0.030053,
		0.029425,
		0.027382,
		0.023933,
		0.019257,
		0.013746,
		0.008047,
		0.003111,
		0.000228,
		3.6e-005,
		0.000169,
		0.000491,
		0.001095,
		0.002063,
		0.003451,
		0.005276,
		0.007511,
		0.010079,
		0.012854,
		0.015662,
		0.018292,
		0.020506,
		0.022059,
		0.022723,
		0.02231,
		0.020712,
		0.017935,
		0.014139,
		0.009688,
		0.005191,
		0.00155,
		3.4e-005,
		0.000157,
		0.000452,
		0.001001,
		0.00187,
		0.003097,
		0.004682,
		0.00658,
		0.008703,
		0.010914,
		0.013038,
		0.014875,
		0.016209,
		0.016836,
		0.016587,
		0.01536,
		0.013156,
		0.010121,
		0.006591,
		0.003134,
		0.000604,
		3.1e-005,
		0.000144,
		0.000412,
		0.000904,
		0.001671,
		0.002735,
		0.00408,
		0.005648,
		0.007338,
		0.009012,
		0.010502,
		0.011623,
		0.012195,
		0.012067,
		0.011143,
		0.009423,
		0.007038,
		0.004295,
		0.001726,
		0.000131,
		2.9e-005,
		0.000131,
		0.000371,
		0.000804,
		0.001468,
		0.00237,
		0.003479,
		0.004727,
		0.006008,
		0.007186,
		0.008105,
		0.008611,
		0.008567,
		0.007888,
		0.00657,
		0.004732,
		0.002653,
		0.000826,
		2.6e-005,
		0.000117,
		0.000328,
		0.000703,
		0.001265,
		0.002008,
		0.00289,
		0.003836,
		0.004739,
		0.005473,
		0.005905,
		0.005917,
		0.005431,
		0.004443,
		0.003057,
		0.001528,
		0.000308,
		2.3e-005,
		0.000103,
		0.000285,
		0.000602,
		0.001064,
		0.001654,
		0.002323,
		0.002991,
		0.003559,
		0.003915,
		0.00396,
		0.003623,
		0.0029,
		0.001882,
		0.000799,
		6.4e-005,
		2e-005,
		8.9e-005,
		0.000243,
		0.000503,
		0.00087,
		0.001317,
		0.00179,
		0.002212,
		0.002495,
		0.002555,
		0.002331,
		0.001816,
		0.001091,
		0.000361,
		1.7e-005,
		7.5e-005,
		0.000201,
		0.000407,
		0.000685,
		0.001001,
		0.001301,
		0.001517,
		0.001579,
		0.001437,
		0.001083,
		0.000586,
		0.000127,
		1.5e-005,
		6.2e-005,
		0.000161,
		0.000317,
		0.000513,
		0.000715,
		0.000871,
		0.000927,
		0.000842,
		0.000608,
		0.000283,
		2.4e-005,
		1.2e-005,
		4.9e-005,
		0.000124,
		0.000234,
		0.000359,
		0.000466,
		0.000512,
		0.000464,
		0.000316,
		0.000116,
		3.7e-005,
		9e-005,
		0.00016,
		0.000227,
		0.000261,
		0.000237,
		0.000149,
		3.7e-005,
		2.7e-005,
		6e-005,
		9.8e-005,
		0.000121,
		0.00011,
		6.2e-005,
		1.7e-005,
		3.5e-005,
		4.9e-005,
		4.5e-005,
		2.1e-005,
		1.6e-005,
		1.6e-005,
		1e-005,
		1.6e-005,
		1.1e-005,
		2.1e-005,
		3.9e-005,
		3.5e-005,
		2e-005,
		3.7e-005,
		7.9e-005,
		8.4e-005,
		6e-005,
		2.9e-005,
		6.2e-005,
		0.000143,
		0.00017,
		0.000143,
		9e-005,
		4e-005,
		1.1e-005,
		9.7e-005,
		0.000237,
		0.000308,
		0.00029,
		0.000215,
		0.000124,
		5.2e-005,
		1.3e-005,
		0.000144,
		0.000371,
		0.000512,
		0.000524,
		0.000434,
		0.000295,
		0.000161,
		6.5e-005,
		1.6e-005,
		0.000205,
		0.000551,
		0.000799,
		0.000871,
		0.000783,
		0.000597,
		0.000384,
		0.000201,
		7.9e-005,
		1.9e-005,
		0.000283,
		0.000785,
		0.001186,
		0.001359,
		0.001301,
		0.001077,
		0.000776,
		0.000479,
		0.000243,
		9.3e-005,
		2.2e-005,
		0.000379,
		0.001083,
		0.001691,
		0.002017,
		0.00203,
		0.00179,
		0.001399,
		0.000966,
		0.000577,
		0.000285,
		0.000107,
		2.4e-005,
		0.000496,
		0.001451,
		0.002331,
		0.002874,
		0.003011,
		0.00279,
		0.002323,
		0.001742,
		0.001164,
		0.000678,
		0.000328,
		0.000121,
		2.7e-005,
		0.000635,
		0.001898,
		0.003122,
		0.00396,
		0.004288,
		0.004136,
		0.00362,
		0.00289,
		0.002098,
		0.001366,
		0.000779,
		0.000371,
		0.000134,
		3e-005,
		0.000799,
		0.002431,
		0.004083,
		0.005302,
		0.005905,
		0.005889,
		0.005365,
		0.004501,
		0.003479,
		0.002461,
		0.00157,
		0.000879,
		0.000412,
		0.000147,
		3.3e-005,
		0.000989,
		0.003057,
		0.005227,
		0.00693,
		0.007904,
		0.008105,
		0.007633,
		0.006667,
		0.005416,
		0.00408,
		0.002826,
		0.001771,
		0.000977,
		0.000452,
		0.00016,
		3.5e-005,
		0.001206,
		0.003782,
		0.00657,
		0.008869,
		0.010326,
		0.010844,
		0.010502,
		0.009482,
		0.008018,
		0.006347,
		0.004682,
		0.003186,
		0.001968,
		0.001072,
		0.000491,
		0.000172,
		3.7e-005,
		0.001451,
		0.00461,
		0.008124,
		0.011143,
		0.013209,
		0.01416,
		0.014044,
		0.013038,
		0.011397,
		0.009391,
		0.007279,
		0.005276,
		0.003537,
		0.002157,
		0.001162,
		0.000526,
		0.000183,
		3.9e-005,
		0.001726,
		0.005545,
		0.009899,
		0.013772,
		0.016587,
		0.018105,
		0.018328,
		0.017426,
		0.015662,
		0.013341,
		0.010764,
		0.008198,
		0.005853,
		0.003874,
		0.002336,
		0.001246,
		0.00056,
		0.000193,
		4.1e-005,
		0.002029,
		0.006591,
		0.011902,
		0.016772,
		0.02049,
		0.022723,
		0.023421,
		0.02273,
		0.020921,
		0.018324,
		0.015282,
		0.012115,
		0.009089,
		0.006407,
		0.004194,
		0.002503,
		0.001323,
		0.00059,
		0.000202,
		4.3e-005,
		0.002362,
		0.007747,
		0.014139,
		0.020157,
		0.024941,
		0.028054,
		0.029379,
		0.02903,
		0.027273,
		0.024461,
		0.020976,
		0.017189,
		0.013424,
		0.009942,
		0.00693,
		0.004491,
		0.002657,
		0.001394,
		0.000617,
		0.00021,
		4.4e-005,
		0.002723,
		0.009012,
		0.016611,
		0.023933,
		0.029959,
		0.034131,
		0.036252,
		0.036394,
		0.034811,
		0.031869,
		0.027985,
		0.023578,
		0.019033,
		0.014673,
		0.010746,
		0.007416,
		0.004764,
		0.002796,
		0.001456,
		0.00064,
		0.000216,
		4.5e-005,
		0.003111,
		0.010384,
		0.019315,
		0.028102,
		0.035553,
		0.040976,
		0.04408,
		0.044882,
		0.043615,
		0.040651,
		0.036436,
		0.031435,
		0.02609,
		0.02079,
		0.015848,
		0.011491,
		0.00786,
		0.005009,
		0.00292,
		0.001511,
		0.000661,
		0.000222,
		4.6e-005,
		0.003525,
		0.011859,
		0.022245,
		0.03266,
		0.041725,
		0.0486,
		0.052889,
		0.054541,
		0.053754,
		0.0509,
		0.046446,
		0.0409,
		0.034759,
		0.028478,
		0.022437,
		0.016933,
		0.01217,
		0.008259,
		0.005226,
		0.003027,
		0.001557,
		0.000677,
		0.000227,
		4.7e-005,
		0.003962,
		0.01343,
		0.025391,
		0.037596,
		0.048465,
		0.057003,
		0.062693,
		0.0654,
		0.06528,
		0.062691,
		0.058116,
		0.052099,
		0.045192,
		0.037912,
		0.030711,
		0.023956,
		0.01792,
		0.012777,
		0.008609,
		0.005413,
		0.003117,
		0.001596,
		0.000691,
		0.00023,
		4.8e-005,
		0.00442,
		0.015087,
		0.028739,
		0.04289,
		0.055758,
		0.066174,
		0.073489,
		0.077474,
		0.078227,
		0.076083,
		0.07153,
		0.065143,
		0.057525,
		0.049254,
		0.040853,
		0.032763,
		0.025331,
		0.018799,
		0.013309,
		0.00891,
		0.005571,
		0.003191,
		0.001626,
		0.000701,
		0.000233,
		4.8e-005,
		0.004895,
		0.016822,
		0.032269,
		0.048517,
		0.063573,
		0.076086,
		0.08526,
		0.090758,
		0.092608,
		0.091109,
		0.086747,
		0.08012,
		0.071873,
		0.062646,
		0.053032,
		0.043547,
		0.034615,
		0.026551,
		0.019564,
		0.013763,
		0.009162,
		0.005698,
		0.00325,
		0.001649,
		0.000709,
		0.000235,
		4.8e-005,
		0.005385,
		0.018622,
		0.03596,
		0.054447,
		0.071873,
		0.086698,
		0.097968,
		0.105226,
		0.108413,
		0.107782,
		0.103804,
		0.097092,
		0.088328,
		0.078208,
		0.067396,
		0.056483,
		0.04597,
		0.036251,
		0.027608,
		0.020214,
		0.014139,
		0.009364,
		0.005798,
		0.003294,
		0.001666,
		0.000714,
		0.000236,
		4.9e-005,
		0.005885,
		0.020474,
		0.039786,
		0.06064,
		0.080608,
		0.097956,
		0.11156,
		0.120829,
		0.125608,
		0.126086,
		0.122709,
		0.116093,
		0.106953,
		0.096035,
		0.084068,
		0.07172,
		0.059573,
		0.0481,
		0.037661,
		0.0285,
		0.020748,
		0.014439,
		0.009521,
		0.005872,
		0.003325,
		0.001677,
		0.000717,
		0.000236,
		4.9e-005,
		0.006393,
		0.022365,
		0.043719,
		0.067052,
		0.089722,
		0.109792,
		0.125963,
		0.137498,
		0.144131,
		0.145975,
		0.143438,
		0.137128,
		0.12778,
		0.116188,
		0.103142,
		0.089384,
		0.075576,
		0.062277,
		0.049927,
		0.038844,
		0.029228,
		0.021172,
		0.014669,
		0.009635,
		0.005924,
		0.003345,
		0.001684,
		0.000719,
		0.000237,
		4.9e-005,
		0.006903,
		0.024279,
		0.047728,
		0.073636,
		0.099147,
		0.122125,
		0.141088,
		0.155141,
		0.163895,
		0.167375,
		0.165934,
		0.160162,
		0.150806,
		0.138695,
		0.124676,
		0.109565,
		0.094104,
		0.078936,
		0.064584,
		0.051448,
		0.039801,
		0.029799,
		0.021491,
		0.014835,
		0.009713,
		0.005956,
		0.003356,
		0.001687,
		0.000719,
		0.000237,
		4.9e-005,
		0.007413,
		0.026203,
		0.051783,
		0.080339,
		0.108813,
		0.134865,
		0.156828,
		0.173644,
		0.184787,
		0.190179,
		0.190107,
		0.185128,
		0.175988,
		0.163544,
		0.148694,
		0.13232,
		0.115245,
		0.098195,
		0.081783,
		0.06649,
		0.052669,
		0.040544,
		0.030225,
		0.021718,
		0.014945,
		0.009761,
		0.005974,
		0.003362,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.007919,
		0.028122,
		0.055852,
		0.08711,
		0.118641,
		0.147912,
		0.173066,
		0.192875,
		0.206667,
		0.214251,
		0.21583,
		0.211916,
		0.203242,
		0.19068,
		0.175172,
		0.157661,
		0.139047,
		0.12014,
		0.101641,
		0.084118,
		0.068007,
		0.053606,
		0.041091,
		0.030522,
		0.021866,
		0.015011,
		0.009787,
		0.005982,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.008417,
		0.030023,
		0.059906,
		0.093895,
		0.128554,
		0.161159,
		0.18967,
		0.212684,
		0.229374,
		0.239425,
		0.242943,
		0.240381,
		0.232442,
		0.220006,
		0.204045,
		0.185558,
		0.165515,
		0.144812,
		0.124235,
		0.104444,
		0.085958,
		0.069157,
		0.054285,
		0.041464,
		0.03071,
		0.021951,
		0.015044,
		0.009797,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.008906,
		0.031895,
		0.063915,
		0.100644,
		0.138472,
		0.1745,
		0.206504,
		0.232908,
		0.252728,
		0.265509,
		0.271253,
		0.270336,
		0.26342,
		0.251376,
		0.235198,
		0.21593,
		0.194608,
		0.172206,
		0.1496,
		0.127539,
		0.10663,
		0.087336,
		0.069976,
		0.054739,
		0.041695,
		0.030815,
		0.021992,
		0.015057,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.009384,
		0.033727,
		0.067856,
		0.107308,
		0.14832,
		0.187825,
		0.223428,
		0.253379,
		0.276532,
		0.292291,
		0.300538,
		0.30156,
		0.295966,
		0.284601,
		0.268467,
		0.248646,
		0.22623,
		0.202269,
		0.177722,
		0.153429,
		0.130086,
		0.108244,
		0.0883,
		0.070513,
		0.055011,
		0.041817,
		0.030862,
		0.022005,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000236,
		4.8e-005,
		0.009847,
		0.035511,
		0.071706,
		0.113845,
		0.158028,
		0.201032,
		0.240301,
		0.27392,
		0.300582,
		0.31954,
		0.330552,
		0.333802,
		0.32983,
		0.319444,
		0.303641,
		0.283525,
		0.260237,
		0.234893,
		0.208535,
		0.18209,
		0.156346,
		0.13194,
		0.109353,
		0.088916,
		0.070823,
		0.055149,
		0.041868,
		0.030876,
		0.022008,
		0.015059,
		0.0098,
		0.005984,
		0.003362,
		0.001686,
		0.000717,
		0.000235,
		4.8e-005,
		0.010297,
		0.037241,
		0.075448,
		0.120219,
		0.167531,
		0.214025,
		0.256992,
		0.294361,
		0.324667,
		0.347016,
		0.361029,
		0.366781,
		0.36473,
		0.35563,
		0.340459,
		0.320332,
		0.296425,
		0.269916,
		0.241918,
		0.213445,
		0.185375,
		0.158434,
		0.133186,
		0.110041,
		0.089259,
		0.070972,
		0.055202,
		0.041881,
		0.030877,
		0.022008,
		0.015059,
		0.009797,
		0.005979,
		0.003356,
		0.001681,
		0.000714,
		0.000234,
		4.8e-005,
		0.010731,
		0.038915,
		0.07907,
		0.126401,
		0.176778,
		0.226717,
		0.273376,
		0.314536,
		0.348584,
		0.374475,
		0.391695,
		0.400201,
		0.400354,
		0.392847,
		0.37862,
		0.358785,
		0.334543,
		0.307118,
		0.277694,
		0.247362,
		0.217088,
		0.187686,
		0.159806,
		0.133936,
		0.110408,
		0.089413,
		0.071023,
		0.055213,
		0.041882,
		0.030877,
		0.022005,
		0.015052,
		0.009787,
		0.005967,
		0.003345,
		0.001672,
		0.000709,
		0.000232,
		4.7e-005,
		0.011152,
		0.040531,
		0.082565,
		0.132372,
		0.185725,
		0.239036,
		0.289343,
		0.334294,
		0.372134,
		0.401677,
		0.422271,
		0.43375,
		0.436373,
		0.430753,
		0.417783,
		0.398556,
		0.374285,
		0.346229,
		0.31563,
		0.283651,
		0.251342,
		0.219602,
		0.189167,
		0.160605,
		0.134318,
		0.110561,
		0.089459,
		0.071031,
		0.055213,
		0.041881,
		0.030871,
		0.021992,
		0.015031,
		0.009761,
		0.005942,
		0.003325,
		0.001658,
		0.000701,
		0.000228,
		4.6e-005,
		0.011559,
		0.04209,
		0.085931,
		0.138117,
		0.194342,
		0.250926,
		0.304801,
		0.353499,
		0.395137,
		0.428393,
		0.452485,
		0.467118,
		0.472443,
		0.468986,
		0.457578,
		0.43928,
		0.415302,
		0.386925,
		0.355436,
		0.322066,
		0.287937,
		0.254032,
		0.22117,
		0.189997,
		0.160989,
		0.134464,
		0.1106,
		0.089465,
		0.071031,
		0.05521,
		0.041868,
		0.030844,
		0.021951,
		0.014983,
		0.009713,
		0.005901,
		0.003294,
		0.001639,
		0.000691,
		0.000224,
		4.5e-005,
		0.011955,
		0.043596,
		0.08917,
		0.143636,
		0.202613,
		0.262345,
		0.319677,
		0.372038,
		0.417432,
		0.454416,
		0.482077,
		0.5,
		0.508222,
		0.507174,
		0.497617,
		0.480563,
		0.457205,
		0.428835,
		0.396774,
		0.362304,
		0.326616,
		0.290769,
		0.25566,
		0.222012,
		0.190372,
		0.161121,
		0.134494,
		0.110603,
		0.089464,
		0.071023,
		0.055183,
		0.041817,
		0.030771,
		0.021866,
		0.014896,
		0.009635,
		0.005838,
		0.00325,
		0.001612,
		0.000677,
		0.000219,
		4.4e-005,
		0.01234,
		0.045055,
		0.09229,
		0.148931,
		0.210532,
		0.27327,
		0.333919,
		0.389823,
		0.438887,
		0.47956,
		0.510811,
		0.532108,
		0.543376,
		0.544946,
		0.537498,
		0.521988,
		0.499578,
		0.471554,
		0.439258,
		0.404013,
		0.367068,
		0.329549,
		0.292425,
		0.256491,
		0.222364,
		0.190485,
		0.161143,
		0.134495,
		0.1106,
		0.089445,
		0.070972,
		0.055094,
		0.041695,
		0.030628,
		0.021718,
		0.014759,
		0.009521,
		0.005752,
		0.003191,
		0.001578,
		0.000661,
		0.000213,
		4.3e-005,
		0.012718,
		0.046472,
		0.095303,
		0.154017,
		0.218106,
		0.283695,
		0.347498,
		0.40679,
		0.459397,
		0.50367,
		0.538479,
		0.563181,
		0.577591,
		0.581942,
		0.576825,
		0.563133,
		0.541984,
		0.514646,
		0.482466,
		0.446798,
		0.408933,
		0.370055,
		0.331199,
		0.293225,
		0.25681,
		0.222455,
		0.190498,
		0.161143,
		0.134487,
		0.110561,
		0.089354,
		0.070823,
		0.054895,
		0.041464,
		0.030388,
		0.021491,
		0.014563,
		0.009364,
		0.005638,
		0.003117,
		0.001535,
		0.00064,
		0.000206,
		4.1e-005,
		0.01309,
		0.047857,
		0.098223,
		0.15891,
		0.225354,
		0.29363,
		0.360406,
		0.422905,
		0.478887,
		0.526629,
		0.564909,
		0.592988,
		0.610581,
		0.617822,
		0.615215,
		0.603579,
		0.58398,
		0.557657,
		0.52595,
		0.490224,
		0.451806,
		0.411925,
		0.371666,
		0.331948,
		0.293501,
		0.256878,
		0.222461,
		0.190496,
		0.161121,
		0.134412,
		0.110408,
		0.089116,
		0.070513,
		0.054538,
		0.041091,
		0.03003,
		0.021172,
		0.014298,
		0.009162,
		0.005496,
		0.003027,
		0.001485,
		0.000617,
		0.000197,
		3.9e-005,
		0.013459,
		0.049219,
		0.101067,
		0.163637,
		0.232304,
		0.3031,
		0.372656,
		0.438158,
		0.497317,
		0.548352,
		0.589968,
		0.621342,
		0.642099,
		0.652282,
		0.652309,
		0.642924,
		0.625132,
		0.600132,
		0.569245,
		0.533837,
		0.495252,
		0.454751,
		0.413461,
		0.372344,
		0.332175,
		0.293546,
		0.25688,
		0.222455,
		0.19045,
		0.160989,
		0.134165,
		0.110041,
		0.088647,
		0.069976,
		0.053976,
		0.040544,
		0.029533,
		0.020748,
		0.01396,
		0.00891,
		0.005324,
		0.00292,
		0.001426,
		0.00059,
		0.000188,
		3.7e-005,
		0.013828,
		0.050566,
		0.103854,
		0.168226,
		0.238991,
		0.312142,
		0.384281,
		0.452566,
		0.514679,
		0.568794,
		0.613563,
		0.648096,
		0.67194,
		0.685059,
		0.687787,
		0.680795,
		0.665023,
		0.641623,
		0.611886,
		0.577165,
		0.53881,
		0.498098,
		0.456182,
		0.414053,
		0.37252,
		0.332201,
		0.293546,
		0.256861,
		0.222364,
		0.190229,
		0.160605,
		0.133616,
		0.109353,
		0.087866,
		0.069157,
		0.053172,
		0.039801,
		0.028884,
		0.020214,
		0.013546,
		0.008609,
		0.005121,
		0.002796,
		0.00136,
		0.00056,
		0.000177,
		3.5e-005,
		0.014199,
		0.05191,
		0.106605,
		0.172707,
		0.245457,
		0.320804,
		0.395328,
		0.46617,
		0.530993,
		0.587947,
		0.635646,
		0.673152,
		0.699951,
		0.715937,
		0.721373,
		0.716858,
		0.703269,
		0.681705,
		0.653418,
		0.61974,
		0.58201,
		0.541506,
		0.499394,
		0.456678,
		0.414179,
		0.372533,
		0.332198,
		0.293501,
		0.256697,
		0.222012,
		0.189652,
		0.159806,
		0.132632,
		0.108244,
		0.086702,
		0.068007,
		0.052095,
		0.038844,
		0.028075,
		0.019564,
		0.013053,
		0.008259,
		0.00489,
		0.002657,
		0.001285,
		0.000526,
		0.000166,
		3.3e-005,
		0.014576,
		0.053259,
		0.109339,
		0.177115,
		0.251747,
		0.329143,
		0.405858,
		0.479026,
		0.546305,
		0.605834,
		0.656209,
		0.696458,
		0.726028,
		0.744755,
		0.752845,
		0.75083,
		0.73953,
		0.719989,
		0.693416,
		0.661108,
		0.624385,
		0.584512,
		0.542647,
		0.49979,
		0.45676,
		0.414184,
		0.37252,
		0.332104,
		0.293225,
		0.256157,
		0.22117,
		0.18852,
		0.158434,
		0.131095,
		0.10663,
		0.085099,
		0.06649,
		0.050725,
		0.037661,
		0.0271,
		0.018799,
		0.012483,
		0.00786,
		0.004631,
		0.002503,
		0.001205,
		0.000491,
		0.000154,
		3e-005,
		0.014959,
		0.054623,
		0.112078,
		0.181483,
		0.257911,
		0.337219,
		0.415942,
		0.491211,
		0.560686,
		0.622512,
		0.675281,
		0.71801,
		0.750119,
		0.771408,
		0.782037,
		0.782486,
		0.773521,
		0.756135,
		0.73149,
		0.700845,
		0.665486,
		0.626654,
		0.585481,
		0.542944,
		0.499836,
		0.45676,
		0.414148,
		0.372344,
		0.331662,
		0.292425,
		0.254963,
		0.219602,
		0.186644,
		0.156346,
		0.128904,
		0.104444,
		0.083014,
		0.064584,
		0.049052,
		0.036251,
		0.025961,
		0.01792,
		0.011839,
		0.007416,
		0.004345,
		0.002336,
		0.001117,
		0.000452,
		0.000141,
		2.7e-005,
		0.015352,
		0.05601,
		0.114839,
		0.185844,
		0.263995,
		0.345096,
		0.425657,
		0.502809,
		0.574223,
		0.63806,
		0.692926,
		0.737842,
		0.772221,
		0.795845,
		0.808844,
		0.81166,
		0.805015,
		0.789857,
		0.767303,
		0.738568,
		0.704898,
		0.667493,
		0.627445,
		0.585688,
		0.542966,
		0.49983,
		0.456678,
		0.413844,
		0.371666,
		0.330511,
		0.290769,
		0.252835,
		0.217088,
		0.183862,
		0.153429,
		0.125984,
		0.101641,
		0.080424,
		0.062277,
		0.047072,
		0.034615,
		0.024662,
		0.016933,
		0.011126,
		0.00693,
		0.004037,
		0.002157,
		0.001025,
		0.000412,
		0.000127,
		2.4e-005,
		0.015757,
		0.057429,
		0.117642,
		0.190228,
		0.270048,
		0.352837,
		0.435083,
		0.513912,
		0.587017,
		0.652581,
		0.709237,
		0.756027,
		0.792377,
		0.818069,
		0.83322,
		0.838252,
		0.833854,
		0.820937,
		0.800578,
		0.773948,
		0.742246,
		0.706621,
		0.668108,
		0.627576,
		0.585696,
		0.542944,
		0.499666,
		0.456182,
		0.412843,
		0.370055,
		0.328265,
		0.287937,
		0.249525,
		0.213445,
		0.180046,
		0.1496,
		0.122288,
		0.098195,
		0.077319,
		0.059573,
		0.044794,
		0.032763,
		0.023214,
		0.015848,
		0.010351,
		0.006407,
		0.003708,
		0.001968,
		0.000929,
		0.000371,
		0.000114,
		2.2e-005,
		0.016175,
		0.058887,
		0.120501,
		0.194665,
		0.276112,
		0.360506,
		0.444298,
		0.524618,
		0.599175,
		0.666188,
		0.724327,
		0.77267,
		0.810671,
		0.838132,
		0.855177,
		0.862222,
		0.859942,
		0.849223,
		0.831105,
		0.80672,
		0.777215,
		0.743679,
		0.707073,
		0.668181,
		0.627576,
		0.585637,
		0.542647,
		0.498898,
		0.454751,
		0.410642,
		0.367068,
		0.324562,
		0.283651,
		0.244832,
		0.208535,
		0.17511,
		0.144812,
		0.117792,
		0.094104,
		0.073709,
		0.056483,
		0.042233,
		0.030711,
		0.021629,
		0.014673,
		0.009521,
		0.005853,
		0.003363,
		0.001771,
		0.000829,
		0.000328,
		0.0001,
		1.9e-005,
		0.016608,
		0.06039,
		0.123433,
		0.199181,
		0.282229,
		0.368158,
		0.453381,
		0.535021,
		0.61081,
		0.679005,
		0.738324,
		0.787895,
		0.827216,
		0.856127,
		0.874775,
		0.883593,
		0.883254,
		0.874633,
		0.858746,
		0.836688,
		0.809553,
		0.778361,
		0.743987,
		0.707107,
		0.668171,
		0.627445,
		0.585137,
		0.541506,
		0.496909,
		0.451806,
		0.406741,
		0.362304,
		0.319097,
		0.277694,
		0.238608,
		0.202269,
		0.169008,
		0.139047,
		0.112501,
		0.089384,
		0.069614,
		0.053032,
		0.039411,
		0.028478,
		0.019924,
		0.013424,
		0.008648,
		0.005276,
		0.003007,
		0.00157,
		0.000728,
		0.000285,
		8.6e-005,
		1.6e-005,
		0.017057,
		0.061944,
		0.12645,
		0.203799,
		0.288436,
		0.375848,
		0.462401,
		0.54521,
		0.622029,
		0.691157,
		0.751367,
		0.801847,
		0.842154,
		0.872181,
		0.892122,
		0.902438,
		0.903823,
		0.897154,
		0.883435,
		0.863728,
		0.839077,
		0.810427,
		0.778551,
		0.743999,
		0.707073,
		0.667916,
		0.626654,
		0.583504,
		0.53881,
		0.493048,
		0.446798,
		0.400704,
		0.355436,
		0.311642,
		0.269916,
		0.230768,
		0.194608,
		0.161731,
		0.13232,
		0.106443,
		0.084068,
		0.065071,
		0.049254,
		0.03636,
		0.02609,
		0.01812,
		0.012115,
		0.007741,
		0.004682,
		0.002644,
		0.001366,
		0.000627,
		0.000243,
		7.2e-005,
		1.3e-005,
		0.017525,
		0.063554,
		0.129562,
		0.208539,
		0.294765,
		0.383622,
		0.471423,
		0.555271,
		0.632938,
		0.702768,
		0.763595,
		0.814675,
		0.855639,
		0.886448,
		0.907357,
		0.918878,
		0.921738,
		0.916836,
		0.905174,
		0.887791,
		0.865681,
		0.839707,
		0.810531,
		0.778551,
		0.743911,
		0.706621,
		0.666724,
		0.624385,
		0.579929,
		0.533837,
		0.486716,
		0.439258,
		0.392189,
		0.346229,
		0.302048,
		0.260237,
		0.221283,
		0.185558,
		0.153312,
		0.124676,
		0.099669,
		0.078208,
		0.060128,
		0.045192,
		0.033116,
		0.023578,
		0.016241,
		0.010764,
		0.006814,
		0.00408,
		0.002279,
		0.001164,
		0.000527,
		0.000201,
		5.9e-005,
		1.1e-005,
		0.018011,
		0.065226,
		0.132781,
		0.213418,
		0.301242,
		0.39152,
		0.480503,
		0.565277,
		0.643631,
		0.713954,
		0.775143,
		0.826533,
		0.867837,
		0.899099,
		0.92065,
		0.933069,
		0.937137,
		0.933787,
		0.924032,
		0.908898,
		0.889329,
		0.866103,
		0.839755,
		0.810517,
		0.778361,
		0.74317,
		0.704898,
		0.663655,
		0.61974,
		0.57363,
		0.52595,
		0.477426,
		0.428835,
		0.38096,
		0.334543,
		0.29025,
		0.248646,
		0.21018,
		0.175172,
		0.143817,
		0.116188,
		0.092251,
		0.071873,
		0.054846,
		0.0409,
		0.029723,
		0.020976,
		0.014314,
		0.009391,
		0.00588,
		0.003479,
		0.001918,
		0.000966,
		0.000431,
		0.000161,
		4.6e-005,
		0.018518,
		0.066962,
		0.136115,
		0.218452,
		0.307892,
		0.399576,
		0.48969,
		0.575293,
		0.654192,
		0.724819,
		0.786138,
		0.837568,
		0.878913,
		0.91031,
		0.932184,
		0.945195,
		0.950194,
		0.948161,
		0.940135,
		0.927134,
		0.910055,
		0.889586,
		0.86612,
		0.839707,
		0.810153,
		0.777215,
		0.740761,
		0.700845,
		0.657731,
		0.611886,
		0.56394,
		0.514646,
		0.464821,
		0.415302,
		0.366892,
		0.320332,
		0.276259,
		0.235198,
		0.197539,
		0.163544,
		0.133344,
		0.106953,
		0.084279,
		0.065143,
		0.049296,
		0.036436,
		0.026229,
		0.018324,
		0.012367,
		0.008018,
		0.004955,
		0.00289,
		0.001568,
		0.000776,
		0.000339,
		0.000124,
		3.5e-005,
		0.019046,
		0.068768,
		0.139572,
		0.223654,
		0.314732,
		0.407819,
		0.499021,
		0.585374,
		0.664693,
		0.735456,
		0.796695,
		0.847917,
		0.889025,
		0.920262,
		0.942154,
		0.955459,
		0.961111,
		0.960152,
		0.953657,
		0.942641,
		0.927957,
		0.910193,
		0.889586,
		0.865999,
		0.839077,
		0.808465,
		0.773948,
		0.735522,
		0.693416,
		0.648076,
		0.600132,
		0.55035,
		0.499578,
		0.448693,
		0.398556,
		0.349969,
		0.303641,
		0.260166,
		0.220006,
		0.183489,
		0.150806,
		0.122023,
		0.097092,
		0.075864,
		0.058116,
		0.04356,
		0.031869,
		0.022689,
		0.015662,
		0.010433,
		0.006667,
		0.004055,
		0.002323,
		0.001235,
		0.000597,
		0.000254,
		9e-005,
		2.4e-005,
		0.019597,
		0.070648,
		0.143159,
		0.229034,
		0.321779,
		0.416269,
		0.508529,
		0.595562,
		0.675192,
		0.745942,
		0.806913,
		0.857704,
		0.898322,
		0.929125,
		0.950751,
		0.964069,
		0.970107,
		0.969981,
		0.96481,
		0.955611,
		0.943187,
		0.928019,
		0.910174,
		0.889329,
		0.864986,
		0.836688,
		0.804146,
		0.767303,
		0.726348,
		0.681705,
		0.633993,
		0.58398,
		0.532534,
		0.480563,
		0.428973,
		0.37862,
		0.330276,
		0.284601,
		0.242125,
		0.203242,
		0.168203,
		0.137128,
		0.110012,
		0.086747,
		0.067132,
		0.0509,
		0.03773,
		0.027273,
		0.019164,
		0.013038,
		0.008547,
		0.005365,
		0.003197,
		0.00179,
		0.000926,
		0.000434,
		0.000178,
		6e-005,
		1.5e-005,
		0.020173,
		0.072605,
		0.146884,
		0.234602,
		0.329048,
		0.424946,
		0.518237,
		0.605889,
		0.685735,
		0.756338,
		0.816874,
		0.867033,
		0.906936,
		0.937057,
		0.958161,
		0.971234,
		0.977409,
		0.977887,
		0.973833,
		0.966268,
		0.95594,
		0.943208,
		0.927957,
		0.909691,
		0.887791,
		0.861713,
		0.831105,
		0.795861,
		0.756135,
		0.712321,
		0.665023,
		0.615002,
		0.563133,
		0.510346,
		0.457578,
		0.405731,
		0.35563,
		0.307995,
		0.26342,
		0.22236,
		0.185128,
		0.151896,
		0.122709,
		0.097494,
		0.076083,
		0.058225,
		0.043615,
		0.031906,
		0.02273,
		0.015716,
		0.010502,
		0.006746,
		0.004136,
		0.0024,
		0.001301,
		0.000649,
		0.00029,
		0.000112,
		3.5e-005,
		0.020774,
		0.074644,
		0.150754,
		0.24037,
		0.336549,
		0.433863,
		0.528163,
		0.616381,
		0.696355,
		0.766692,
		0.826642,
		0.875992,
		0.914977,
		0.9442,
		0.964555,
		0.977154,
		0.983245,
		0.984115,
		0.980982,
		0.974863,
		0.966443,
		0.95594,
		0.943052,
		0.927134,
		0.907465,
		0.883435,
		0.85464,
		0.820937,
		0.782445,
		0.73953,
		0.692773,
		0.642924,
		0.590853,
		0.537498,
		0.48382,
		0.430753,
		0.379165,
		0.32983,
		0.283397,
		0.240381,
		0.201151,
		0.165934,
		0.134822,
		0.107782,
		0.084677,
		0.06528,
		0.049299,
		0.036394,
		0.026196,
		0.018328,
		0.012416,
		0.008105,
		0.005068,
		0.003011,
		0.001682,
		0.000871,
		0.00041,
		0.00017,
		6e-005,
		1.6e-005,
		0.021402,
		0.07677,
		0.154777,
		0.246346,
		0.344293,
		0.443032,
		0.53832,
		0.627051,
		0.707073,
		0.777035,
		0.836264,
		0.884648,
		0.92254,
		0.950676,
		0.970086,
		0.982016,
		0.987832,
		0.988911,
		0.986518,
		0.981661,
		0.974941,
		0.966419,
		0.955611,
		0.941747,
		0.924032,
		0.901806,
		0.874633,
		0.842338,
		0.805015,
		0.763003,
		0.716858,
		0.667309,
		0.615215,
		0.561514,
		0.507174,
		0.453154,
		0.400354,
		0.34959,
		0.30156,
		0.256833,
		0.21583,
		0.178832,
		0.145975,
		0.117269,
		0.092608,
		0.071791,
		0.054541,
		0.040526,
		0.029379,
		0.020718,
		0.01416,
		0.009337,
		0.005905,
		0.003555,
		0.002017,
		0.001064,
		0.000512,
		0.000218,
		7.9e-005,
		2.2e-005,
		0.022059,
		0.078987,
		0.15896,
		0.25254,
		0.352291,
		0.452462,
		0.548717,
		0.637912,
		0.717902,
		0.787387,
		0.845772,
		0.893048,
		0.929694,
		0.956583,
		0.974887,
		0.985988,
		0.991373,
		0.992508,
		0.990698,
		0.986925,
		0.981687,
		0.974863,
		0.965809,
		0.953657,
		0.937562,
		0.916836,
		0.891026,
		0.859942,
		0.823657,
		0.782486,
		0.736961,
		0.687787,
		0.635805,
		0.581942,
		0.527165,
		0.472443,
		0.4187,
		0.366781,
		0.317429,
		0.271253,
		0.228728,
		0.190179,
		0.155789,
		0.125608,
		0.099561,
		0.077474,
		0.059088,
		0.04408,
		0.032087,
		0.022723,
		0.015596,
		0.010326,
		0.006556,
		0.00396,
		0.002251,
		0.001186,
		0.000567,
		0.000237,
		8.2e-005,
		2.1e-005,
		0.022747,
		0.081301,
		0.163312,
		0.258963,
		0.360554,
		0.462163,
		0.559361,
		0.648966,
		0.728847,
		0.797756,
		0.855181,
		0.901223,
		0.936491,
		0.961999,
		0.979067,
		0.989217,
		0.994054,
		0.995127,
		0.993768,
		0.990914,
		0.986925,
		0.981492,
		0.973833,
		0.963014,
		0.948161,
		0.928581,
		0.903823,
		0.873692,
		0.838252,
		0.7978,
		0.752845,
		0.704067,
		0.652282,
		0.598399,
		0.543376,
		0.488182,
		0.43375,
		0.38095,
		0.330552,
		0.283205,
		0.239425,
		0.19958,
		0.163895,
		0.132457,
		0.105226,
		0.082051,
		0.062693,
		0.046839,
		0.034131,
		0.02418,
		0.016587,
		0.010961,
		0.00693,
		0.004154,
		0.002331,
		0.0012,
		0.000551,
		0.000213,
		6.2e-005,
		0.023468,
		0.083718,
		0.167843,
		0.265625,
		0.369091,
		0.472143,
		0.570256,
		0.660215,
		0.739907,
		0.808142,
		0.864496,
		0.909189,
		0.942964,
		0.966984,
		0.982718,
		0.991829,
		0.996039,
		0.996969,
		0.995957,
		0.993865,
		0.990884,
		0.986518,
		0.979875,
		0.969981,
		0.955961,
		0.937137,
		0.913075,
		0.883593,
		0.848753,
		0.808844,
		0.764353,
		0.715937,
		0.664385,
		0.610581,
		0.555467,
		0.5,
		0.445116,
		0.391695,
		0.340527,
		0.292291,
		0.247536,
		0.206667,
		0.169947,
		0.137498,
		0.109311,
		0.08526,
		0.065122,
		0.0486,
		0.035336,
		0.024941,
		0.017011,
		0.011143,
		0.006952,
		0.004083,
		0.002217,
		0.001083,
		0.000451,
		0.000144,
		2.5e-005,
		0.024223,
		0.086243,
		0.172561,
		0.272538,
		0.377913,
		0.48241,
		0.581406,
		0.671658,
		0.751076,
		0.818535,
		0.873711,
		0.916947,
		0.94913,
		0.971575,
		0.985907,
		0.993929,
		0.997471,
		0.99821,
		0.997467,
		0.99599,
		0.993768,
		0.990133,
		0.984115,
		0.974726,
		0.961111,
		0.942628,
		0.918878,
		0.889702,
		0.855177,
		0.815588,
		0.771408,
		0.723272,
		0.67194,
		0.618271,
		0.563181,
		0.507609,
		0.452485,
		0.398686,
		0.347016,
		0.298172,
		0.252728,
		0.211121,
		0.173644,
		0.140451,
		0.11156,
		0.086869,
		0.066174,
		0.049186,
		0.035553,
		0.024886,
		0.016772,
		0.0108,
		0.00657,
		0.003714,
		0.001898,
		0.000834,
		0.000283,
		5.2e-005,
		0.025016,
		0.088884,
		0.177478,
		0.279714,
		0.387032,
		0.492972,
		0.592813,
		0.683291,
		0.762345,
		0.828924,
		0.882813,
		0.92449,
		0.954993,
		0.975798,
		0.988688,
		0.995602,
		0.99847,
		0.999005,
		0.998474,
		0.997467,
		0.995751,
		0.992508,
		0.986721,
		0.977409,
		0.963764,
		0.945195,
		0.921354,
		0.892122,
		0.857596,
		0.818069,
		0.774009,
		0.726028,
		0.674863,
		0.621342,
		0.566352,
		0.510811,
		0.455628,
		0.401677,
		0.349758,
		0.300582,
		0.254737,
		0.212684,
		0.174741,
		0.141088,
		0.111767,
		0.086698,
		0.06569,
		0.048465,
		0.034677,
		0.023933,
		0.015814,
		0.009899,
		0.005775,
		0.003057,
		0.001398,
		0.000496,
		9.7e-005,
		0.025848,
		0.091646,
		0.182603,
		0.287165,
		0.39646,
		0.503836,
		0.604481,
		0.695109,
		0.773702,
		0.839291,
		0.891783,
		0.931802,
		0.960548,
		0.979666,
		0.991097,
		0.996917,
		0.999139,
		0.999483,
		0.999123,
		0.998438,
		0.996969,
		0.993781,
		0.987832,
		0.978178,
		0.964069,
		0.944988,
		0.92065,
		0.890985,
		0.856127,
		0.816383,
		0.772221,
		0.724239,
		0.673152,
		0.619756,
		0.564909,
		0.5095,
		0.454416,
		0.400512,
		0.348584,
		0.299339,
		0.253379,
		0.211175,
		0.173066,
		0.139251,
		0.109792,
		0.084624,
		0.063573,
		0.046366,
		0.03266,
		0.02206,
		0.014139,
		0.008464,
		0.00461,
		0.002176,
		0.000799,
		0.000162,
		0.026722,
		0.094539,
		0.18795,
		0.294906,
		0.406208,
		0.515011,
		0.61641,
		0.707107,
		0.785132,
		0.849616,
		0.900598,
		0.938862,
		0.965783,
		0.983181,
		0.993163,
		0.99793,
		0.999561,
		0.999749,
		0.999523,
		0.999005,
		0.997522,
		0.994054,
		0.987562,
		0.977154,
		0.962163,
		0.942154,
		0.916918,
		0.886448,
		0.850921,
		0.810671,
		0.766169,
		0.71801,
		0.666884,
		0.613563,
		0.558873,
		0.50367,
		0.448814,
		0.395137,
		0.343419,
		0.294361,
		0.248563,
		0.206504,
		0.168534,
		0.134865,
		0.105574,
		0.080608,
		0.059802,
		0.04289,
		0.029527,
		0.019315,
		0.011818,
		0.006591,
		0.003192,
		0.001206,
		0.000253,
		0.027641,
		0.09757,
		0.193529,
		0.302949,
		0.416289,
		0.526505,
		0.628601,
		0.719275,
		0.796619,
		0.859876,
		0.909232,
		0.945646,
		0.970681,
		0.986344,
		0.994906,
		0.998688,
		0.999806,
		0.999886,
		0.999749,
		0.99923,
		0.997471,
		0.993394,
		0.985988,
		0.974436,
		0.958161,
		0.936827,
		0.91031,
		0.878677,
		0.842154,
		0.801108,
		0.756027,
		0.707502,
		0.656209,
		0.602893,
		0.548352,
		0.493408,
		0.438887,
		0.385597,
		0.334294,
		0.285665,
		0.240301,
		0.198682,
		0.161159,
		0.127953,
		0.099147,
		0.074699,
		0.054447,
		0.038126,
		0.025391,
		0.015835,
		0.009012,
		0.004461,
		0.001726,
		0.000371,
		0.028608,
		0.100747,
		0.199356,
		0.31131,
		0.426716,
		0.538323,
		0.641052,
		0.731605,
		0.808146,
		0.870045,
		0.917657,
		0.952128,
		0.975226,
		0.989153,
		0.996346,
		0.99923,
		0.999932,
		0.999952,
		0.999841,
		0.999139,
		0.996836,
		0.991829,
		0.983155,
		0.970086,
		0.952154,
		0.929125,
		0.900971,
		0.867837,
		0.830009,
		0.787895,
		0.742003,
		0.692926,
		0.641333,
		0.587947,
		0.533535,
		0.478887,
		0.424797,
		0.372038,
		0.321343,
		0.273376,
		0.228713,
		0.187825,
		0.151061,
		0.118641,
		0.090652,
		0.067052,
		0.04768,
		0.032269,
		0.020468,
		0.011859,
		0.005983,
		0.002362,
		0.000519,
		0.029626,
		0.104082,
		0.205443,
		0.320005,
		0.437501,
		0.550474,
		0.653763,
		0.744084,
		0.819691,
		0.880097,
		0.925841,
		0.95828,
		0.9794,
		0.991607,
		0.997501,
		0.999595,
		0.999983,
		0.999981,
		0.999806,
		0.998721,
		0.995602,
		0.989349,
		0.979067,
		0.964136,
		0.9442,
		0.919139,
		0.889025,
		0.854087,
		0.814675,
		0.771243,
		0.724327,
		0.674529,
		0.622512,
		0.568982,
		0.514679,
		0.460362,
		0.40679,
		0.354703,
		0.304801,
		0.257723,
		0.214025,
		0.174161,
		0.138472,
		0.107171,
		0.080339,
		0.057933,
		0.039786,
		0.025625,
		0.015087,
		0.007742,
		0.003111,
		0.000697,
		0.0307,
		0.107582,
		0.211807,
		0.32905,
		0.448658,
		0.562964,
		0.66673,
		0.756699,
		0.83123,
		0.890001,
		0.933754,
		0.964076,
		0.983187,
		0.993706,
		0.998392,
		0.999818,
		0.999998,
		0.999983,
		0.99962,
		0.99793,
		0.993717,
		0.985907,
		0.973698,
		0.956583,
		0.93433,
		0.906936,
		0.874577,
		0.837568,
		0.796331,
		0.751367,
		0.703245,
		0.652581,
		0.600037,
		0.546305,
		0.492103,
		0.438158,
		0.385195,
		0.333919,
		0.284996,
		0.239036,
		0.196569,
		0.158028,
		0.123736,
		0.093895,
		0.068577,
		0.047728,
		0.031171,
		0.018622,
		0.009702,
		0.003962,
		0.000902,
		0.031834,
		0.111261,
		0.218463,
		0.338462,
		0.460201,
		0.575799,
		0.679948,
		0.769433,
		0.842738,
		0.899725,
		0.941362,
		0.969489,
		0.986574,
		0.995458,
		0.999046,
		0.999937,
		1,
		0.999942,
		0.99923,
		0.996693,
		0.991097,
		0.981426,
		0.966984,
		0.94739,
		0.92254,
		0.892551,
		0.857704,
		0.818404,
		0.775143,
		0.728476,
		0.679005,
		0.627366,
		0.574223,
		0.520259,
		0.46617,
		0.412655,
		0.360406,
		0.310092,
		0.262345,
		0.21774,
		0.176778,
		0.139866,
		0.107308,
		0.079284,
		0.055852,
		0.036943,
		0.022365,
		0.011815,
		0.004895,
		0.001132,
		0.033031,
		0.11513,
		0.225429,
		0.34826,
		0.472143,
		0.588982,
		0.69341,
		0.782268,
		0.854185,
		0.909232,
		0.948631,
		0.974494,
		0.989554,
		0.996876,
		0.999493,
		0.999986,
		0.999999,
		0.999818,
		0.998558,
		0.994906,
		0.987631,
		0.975798,
		0.958834,
		0.936491,
		0.908799,
		0.875992,
		0.83846,
		0.796695,
		0.751262,
		0.702768,
		0.651854,
		0.599175,
		0.545402,
		0.491211,
		0.437279,
		0.384281,
		0.332875,
		0.283695,
		0.237338,
		0.194342,
		0.155176,
		0.120219,
		0.089745,
		0.063915,
		0.042764,
		0.026203,
		0.014018,
		0.005885,
		0.00138,
		0.034297,
		0.119202,
		0.232725,
		0.358463,
		0.484498,
		0.602518,
		0.707107,
		0.795181,
		0.865538,
		0.918486,
		0.955526,
		0.979066,
		0.992122,
		0.997978,
		0.999771,
		0.999999,
		0.999986,
		0.999548,
		0.997501,
		0.992443,
		0.983181,
		0.96889,
		0.94913,
		0.923796,
		0.893048,
		0.857243,
		0.816874,
		0.772521,
		0.724819,
		0.674429,
		0.622029,
		0.568298,
		0.513912,
		0.459545,
		0.405858,
		0.3535,
		0.3031,
		0.255258,
		0.210532,
		0.169425,
		0.132372,
		0.099716,
		0.071706,
		0.048471,
		0.030023,
		0.016247,
		0.006903,
		0.001639,
		0.035637,
		0.123492,
		0.24037,
		0.369091,
		0.497281,
		0.61641,
		0.721027,
		0.808146,
		0.876762,
		0.927446,
		0.962012,
		0.983187,
		0.994282,
		0.998793,
		0.99992,
		1,
		0.999929,
		0.999046,
		0.995933,
		0.989153,
		0.97759,
		0.960548,
		0.937734,
		0.909189,
		0.875207,
		0.836264,
		0.792954,
		0.745942,
		0.695929,
		0.643631,
		0.589761,
		0.535021,
		0.480096,
		0.425657,
		0.372351,
		0.320804,
		0.271617,
		0.225354,
		0.182539,
		0.143636,
		0.109042,
		0.07907,
		0.053931,
		0.033727,
		0.018437,
		0.007919,
		0.001902,
		0.037057,
		0.128015,
		0.248386,
		0.380165,
		0.510503,
		0.630658,
		0.735156,
		0.821134,
		0.887815,
		0.93607,
		0.968053,
		0.986838,
		0.996042,
		0.999355,
		0.999982,
		0.999999,
		0.999771,
		0.9982,
		0.993706,
		0.984868,
		0.970681,
		0.9506,
		0.92449,
		0.892541,
		0.855181,
		0.813001,
		0.766692,
		0.717001,
		0.664693,
		0.610534,
		0.555271,
		0.499627,
		0.444298,
		0.38995,
		0.337219,
		0.286708,
		0.238991,
		0.194598,
		0.154017,
		0.117674,
		0.085931,
		0.05906,
		0.037241,
		0.020541,
		0.008906,
		0.002161,
		0.038563,
		0.132788,
		0.256796,
		0.391707,
		0.52418,
		0.645261,
		0.749475,
		0.834112,
		0.898654,
		0.944313,
		0.973615,
		0.990011,
		0.997422,
		0.999707,
		0.999999,
		0.999982,
		0.999429,
		0.996876,
		0.99065,
		0.9794,
		0.962262,
		0.938862,
		0.909231,
		0.873711,
		0.832862,
		0.787387,
		0.738071,
		0.685735,
		0.631206,
		0.575293,
		0.518777,
		0.462401,
		0.406867,
		0.352837,
		0.300936,
		0.251747,
		0.205814,
		0.163637,
		0.125665,
		0.09229,
		0.063833,
		0.040531,
		0.022525,
		0.009847,
		0.00241,
		0.040162,
		0.137829,
		0.265625,
		0.403741,
		0.538323,
		0.660215,
		0.763962,
		0.84704,
		0.909232,
		0.952128,
		0.978669,
		0.9927,
		0.998451,
		0.999897,
		1,
		0.999909,
		0.998793,
		0.994912,
		0.986574,
		0.972543,
		0.952128,
		0.92514,
		0.891783,
		0.852554,
		0.808142,
		0.759357,
		0.707073,
		0.652181,
		0.595562,
		0.538067,
		0.480503,
		0.423631,
		0.368158,
		0.314743,
		0.263995,
		0.216478,
		0.172707,
		0.133151,
		0.098223,
		0.068277,
		0.043596,
		0.02438,
		0.010731,
		0.002646,
		0.041863,
		0.14316,
		0.2749,
		0.416289,
		0.552945,
		0.675515,
		0.778592,
		0.859876,
		0.919497,
		0.959469,
		0.983187,
		0.994912,
		0.999166,
		0.999977,
		0.999999,
		0.999707,
		0.997721,
		0.992122,
		0.981267,
		0.964076,
		0.940063,
		0.909232,
		0.87197,
		0.828924,
		0.780912,
		0.728847,
		0.673685,
		0.616381,
		0.557864,
		0.499021,
		0.440684,
		0.383622,
		0.328548,
		0.276112,
		0.22691,
		0.181483,
		0.140319,
		0.103854,
		0.072466,
		0.046472,
		0.026116,
		0.011559,
		0.002868,
		0.043674,
		0.148801,
		0.28465,
		0.429378,
		0.568056,
		0.69115,
		0.793332,
		0.872571,
		0.929393,
		0.966289,
		0.987147,
		0.996662,
		0.999618,
		0.999998,
		0.999977,
		0.999268,
		0.996042,
		0.988294,
		0.974494,
		0.953766,
		0.925841,
		0.890936,
		0.849616,
		0.802682,
		0.751076,
		0.695809,
		0.637912,
		0.578397,
		0.518237,
		0.458348,
		0.399576,
		0.342703,
		0.288436,
		0.237419,
		0.190228,
		0.147381,
		0.109339,
		0.076504,
		0.049219,
		0.027762,
		0.01234,
		0.003077,
		0.045605,
		0.154777,
		0.294906,
		0.443032,
		0.583667,
		0.707107,
		0.808146,
		0.88507,
		0.938862,
		0.972543,
		0.990536,
		0.997978,
		0.999864,
		1,
		0.999882,
		0.998451,
		0.993551,
		0.983187,
		0.966004,
		0.941362,
		0.909232,
		0.870045,
		0.824552,
		0.773702,
		0.718555,
		0.660215,
		0.599782,
		0.53832,
		0.476837,
		0.416269,
		0.357477,
		0.301242,
		0.248267,
		0.199181,
		0.154541,
		0.114839,
		0.080506,
		0.05191,
		0.029357,
		0.01309,
		0.003275,
		0.047667,
		0.161115,
		0.3057,
		0.457278,
		0.599783,
		0.723368,
		0.82299,
		0.897313,
		0.947842,
		0.978187,
		0.99335,
		0.998903,
		0.99997,
		0.999998,
		0.999618,
		0.997073,
		0.990011,
		0.976536,
		0.955526,
		0.926607,
		0.890001,
		0.846361,
		0.796619,
		0.741872,
		0.683291,
		0.622067,
		0.559361,
		0.496281,
		0.433863,
		0.373058,
		0.314732,
		0.259662,
		0.208539,
		0.161974,
		0.120501,
		0.084585,
		0.054623,
		0.030947,
		0.013828,
		0.003469,
		0.049871,
		0.167843,
		0.317068,
		0.472143,
		0.61641,
		0.739908,
		0.837812,
		0.909232,
		0.956268,
		0.983187,
		0.995601,
		0.999493,
		0.999998,
		0.99997,
		0.999046,
		0.994912,
		0.985151,
		0.968053,
		0.942774,
		0.909232,
		0.867913,
		0.819691,
		0.765677,
		0.707107,
		0.645261,
		0.581406,
		0.516759,
		0.452462,
		0.389572,
		0.329048,
		0.271752,
		0.218452,
		0.169822,
		0.12645,
		0.088841,
		0.057429,
		0.032575,
		0.014576,
		0.003662,
		0.05223,
		0.174994,
		0.32905,
		0.487653,
		0.63355,
		0.756699,
		0.852555,
		0.920755,
		0.964076,
		0.987511,
		0.997312,
		0.999819,
		1,
		0.999844,
		0.997978,
		0.991705,
		0.978669,
		0.957423,
		0.927446,
		0.888967,
		0.842738,
		0.789856,
		0.731605,
		0.669354,
		0.604481,
		0.538323,
		0.472143,
		0.407111,
		0.344293,
		0.28465,
		0.229034,
		0.178194,
		0.132781,
		0.093355,
		0.06039,
		0.034281,
		0.015352,
		0.003862,
		0.054761,
		0.182603,
		0.341684,
		0.503836,
		0.651201,
		0.773702,
		0.867151,
		0.931803,
		0.971204,
		0.991144,
		0.998528,
		0.999959,
		0.999998,
		0.999493,
		0.996177,
		0.987147,
		0.97023,
		0.944313,
		0.909232,
		0.865538,
		0.814257,
		0.756699,
		0.694316,
		0.628601,
		0.561021,
		0.492972,
		0.425753,
		0.360554,
		0.298442,
		0.24037,
		0.187172,
		0.139572,
		0.098192,
		0.063554,
		0.036097,
		0.016175,
		0.004071,
		0.057479,
		0.19071,
		0.355015,
		0.520718,
		0.669354,
		0.790869,
		0.881522,
		0.94229,
		0.977591,
		0.994084,
		0.999313,
		0.999997,
		0.999959,
		0.998733,
		0.99335,
		0.980889,
		0.959469,
		0.928371,
		0.887815,
		0.838682,
		0.782268,
		0.720092,
		0.653763,
		0.584888,
		0.515011,
		0.445578,
		0.377913,
		0.313213,
		0.25254,
		0.196828,
		0.146884,
		0.103402,
		0.066962,
		0.03805,
		0.017057,
		0.004295,
		0.060404,
		0.199356,
		0.369091,
		0.538323,
		0.687996,
		0.808146,
		0.895582,
		0.952128,
		0.983187,
		0.996348,
		0.999752,
		1,
		0.999788,
		0.997312,
		0.989154,
		0.972543,
		0.945995,
		0.909232,
		0.862877,
		0.808146,
		0.746597,
		0.679948,
		0.609954,
		0.538323,
		0.466665,
		0.39646,
		0.32905,
		0.265625,
		0.207232,
		0.154777,
		0.109032,
		0.070648,
		0.040162,
		0.018011,
		0.004537,
		0.063557,
		0.208589,
		0.383959,
		0.556677,
		0.707107,
		0.825463,
		0.909232,
		0.961225,
		0.987949,
		0.997978,
		0.999944,
		0.999997,
		0.999313,
		0.994912,
		0.983187,
		0.961681,
		0.929393,
		0.886525,
		0.834112,
		0.773702,
		0.707107,
		0.636233,
		0.562964,
		0.489096,
		0.416289,
		0.346048,
		0.279714,
		0.218463,
		0.163312,
		0.11513,
		0.074644,
		0.042454,
		0.019046,
		0.004799,
		0.066963,
		0.218463,
		0.399673,
		0.575799,
		0.726654,
		0.842738,
		0.922365,
		0.969489,
		0.991857,
		0.999046,
		0.999996,
		0.999944,
		0.99828,
		0.991144,
		0.974996,
		0.947842,
		0.909232,
		0.859876,
		0.801227,
		0.735156,
		0.663715,
		0.588982,
		0.512956,
		0.437501,
		0.364313,
		0.294906,
		0.230606,
		0.172561,
		0.121749,
		0.078987,
		0.044948,
		0.020173,
		0.005084,
		0.070648,
		0.229035,
		0.416289,
		0.595706,
		0.746597,
		0.859876,
		0.93486,
		0.976835,
		0.994912,
		0.999651,
		1,
		0.999707,
		0.996348,
		0.985548,
		0.964076,
		0.930531,
		0.88507,
		0.828924,
		0.763962,
		0.692366,
		0.61641,
		0.538323,
		0.460201,
		0.383959,
		0.31131,
		0.243759,
		0.182603,
		0.128949,
		0.083718,
		0.047667,
		0.021402,
		0.005396,
		0.074644,
		0.24037,
		0.433865,
		0.61641,
		0.766878,
		0.876762,
		0.946591,
		0.983187,
		0.997147,
		0.99992,
		0.999995,
		0.999046,
		0.993081,
		0.977591,
		0.949876,
		0.909232,
		0.856465,
		0.793332,
		0.722107,
		0.645261,
		0.565272,
		0.484498,
		0.405109,
		0.32905,
		0.258031,
		0.193529,
		0.136799,
		0.088884,
		0.050639,
		0.022747,
		0.005737,
		0.078987,
		0.25254,
		0.452462,
		0.637915,
		0.787425,
		0.893264,
		0.957423,
		0.988485,
		0.998635,
		0.999994,
		0.99992,
		0.997607,
		0.987949,
		0.966673,
		0.931803,
		0.883417,
		0.82299,
		0.752804,
		0.675515,
		0.593863,
		0.510503,
		0.427896,
		0.34826,
		0.273547,
		0.205443,
		0.145377,
		0.094539,
		0.053897,
		0.024223,
		0.006111,
		0.083718,
		0.265625,
		0.472143,
		0.660215,
		0.808146,
		0.909232,
		0.967218,
		0.9927,
		0.999493,
		1,
		0.999581,
		0.994912,
		0.980326,
		0.952128,
		0.909232,
		0.852555,
		0.78425,
		0.707107,
		0.624138,
		0.538323,
		0.452462,
		0.369091,
		0.290446,
		0.218463,
		0.154777,
		0.100747,
		0.057479,
		0.025848,
		0.006524,
		0.088879,
		0.279714,
		0.492972,
		0.683291,
		0.828924,
		0.924495,
		0.975845,
		0.995843,
		0.999882,
		0.999993,
		0.998635,
		0.990352,
		0.969489,
		0.933236,
		0.881522,
		0.816138,
		0.739908,
		0.656102,
		0.568056,
		0.478955,
		0.391707,
		0.308888,
		0.232725,
		0.165103,
		0.107582,
		0.061429,
		0.027641,
		0.006979,
		0.094456,
		0.294891,
		0.515011,
		0.707107,
		0.849616,
		0.938862,
		0.983187,
		0.997978,
		0.999991,
		0.999882,
		0.996571,
		0.983187,
		0.95463,
		0.909232,
		0.848031,
		0.773702,
		0.689715,
		0.599783,
		0.507526,
		0.416289,
		0.32905,
		0.248386,
		0.176478,
		0.11513,
		0.065798,
		0.029626,
		0.007484,
		0.100129,
		0.311037,
		0.538297,
		0.731605,
		0.870045,
		0.952128,
		0.989154,
		0.999236,
		1,
		0.999381,
		0.9927,
		0.972543,
		0.93486,
		0.879329,
		0.808146,
		0.72486,
		0.63355,
		0.538323,
		0.443032,
		0.351133,
		0.265625,
		0.189047,
		0.123492,
		0.070648,
		0.031834,
		0.008046,
		0.104887,
		0.327031,
		0.56247,
		0.756663,
		0.890001,
		0.964076,
		0.993706,
		0.999819,
		0.999989,
		0.997978,
		0.986146,
		0.957423,
		0.909232,
		0.842738,
		0.761317,
		0.669354,
		0.571482,
		0.472143,
		0.375363,
		0.28465,
		0.202976,
		0.132788,
		0.076051,
		0.034297,
		0.008674,
		0.106843,
		0.339533,
		0.585368,
		0.781582,
		0.909189,
		0.974494,
		0.996876,
		0.999986,
		0.999819,
		0.994912,
		0.975845,
		0.936718,
		0.876762,
		0.798713,
		0.707107,
		0.607109,
		0.503836,
		0.40199,
		0.3057,
		0.218463,
		0.14316,
		0.082095,
		0.037057,
		0.009378,
		0.103703,
		0.342523,
		0.600964,
		0.803188,
		0.926633,
		0.983139,
		0.998793,
		1,
		0.999046,
		0.989154,
		0.960553,
		0.909232,
		0.836467,
		0.746597,
		0.645261,
		0.538323,
		0.431293,
		0.32905,
		0.23574,
		0.154777,
		0.088884,
		0.040162,
		0.010171,
		0.093895,
		0.328938,
		0.598814,
		0.81321,
		0.938519,
		0.989142,
		0.999659,
		0.999982,
		0.996876,
		0.9794,
		0.938862,
		0.873716,
		0.787425,
		0.685901,
		0.575799,
		0.463572,
		0.355015,
		0.255081,
		0.167843,
		0.096544,
		0.043674,
		0.011069,
		0.077731,
		0.294361,
		0.567267,
		0.797981,
		0.935426,
		0.988808,
		0.9991,
		0.999659,
		0.992122,
		0.964076,
		0.909232,
		0.828924,
		0.728851,
		0.61641,
		0.499142,
		0.383959,
		0.276811,
		0.182603,
		0.105229,
		0.047667,
		0.012092,
		0.057659,
		0.240551,
		0.5,
		0.743241,
		0.902538,
		0.97297,
		0.993865,
		0.997575,
		0.983139,
		0.941362,
		0.870045,
		0.773702,
		0.660215,
		0.538323,
		0.416289,
		0.301315,
		0.199356,
		0.11513,
		0.05223,
		0.013263,
		0.037339,
		0.175889,
		0.401745,
		0.642924,
		0.825636,
		0.927549,
		0.974912,
		0.988808,
		0.967203,
		0.909189,
		0.819691,
		0.707107,
		0.581406,
		0.452462,
		0.32905,
		0.218463,
		0.12648,
		0.057479,
		0.014613,
		0.020197,
		0.112084,
		0.288229,
		0.505937,
		0.700845,
		0.839721,
		0.927549,
		0.96241,
		0.938519,
		0.864779,
		0.756663,
		0.628601,
		0.492972,
		0.360554,
		0.24037,
		0.139573,
		0.063557,
		0.016179,
		0.008349,
		0.05954,
		0.179781,
		0.354699,
		0.540983,
		0.707107,
		0.837497,
		0.902538,
		0.886448,
		0.803188,
		0.679351,
		0.538297,
		0.39646,
		0.265625,
		0.154777,
		0.070648,
		0.018011,
		0.002141,
		0.024116,
		0.09321,
		0.215674,
		0.372177,
		0.542935,
		0.700845,
		0.795954,
		0.797981,
		0.716734,
		0.585368,
		0.437117,
		0.294891,
		0.172561,
		0.078987,
		0.020173,
		0.000173,
		0.006041,
		0.036721,
		0.108744,
		0.222444,
		0.372177,
		0.530799,
		0.642924,
		0.666205,
		0.598814,
		0.472358,
		0.327031,
		0.19336,
		0.088879,
		0.022747,
		0.000475,
		0.00891,
		0.041577,
		0.110603,
		0.22133,
		0.354699,
		0.462908,
		0.5,
		0.45206,
		0.342523,
		0.212988,
		0.100129,
		0.025825,
		0.000676,
		0.00978,
		0.041877,
		0.108744,
		0.202269,
		0.288229,
		0.325662,
		0.294361,
		0.208583,
		0.106843,
		0.028884,
		0.000719,
		0.00978,
		0.040099,
		0.09321,
		0.148905,
		0.175889,
		0.154555,
		0.093895,
		0.028801,
		0.000712,
		0.00891,
		0.031269,
		0.05954,
		0.073201,
		0.057659,
		0.021807,
		0.000596,
		0.006041,
		0.016081,
		0.020197,
		0.010604,
		0.000326,
		0.002141,
		0.002407,
		5.1e-005,
		5.1e-005,
		0.002407,
		0.002141,
		0.000326,
		0.010604,
		0.020197,
		0.016081,
		0.006041,
		0.000596,
		0.021807,
		0.057659,
		0.073201,
		0.05954,
		0.031269,
		0.00891,
		0.000712,
		0.028801,
		0.093895,
		0.154555,
		0.175889,
		0.148905,
		0.09321,
		0.040099,
		0.00978,
		0.000719,
		0.028884,
		0.106843,
		0.208583,
		0.294361,
		0.325662,
		0.288229,
		0.202269,
		0.108744,
		0.041877,
		0.00978,
		0.000676,
		0.025825,
		0.100129,
		0.212988,
		0.342523,
		0.45206,
		0.5,
		0.462908,
		0.354699,
		0.22133,
		0.110603,
		0.041577,
		0.00891,
		0.000475,
		0.022747,
		0.088879,
		0.19336,
		0.327031,
		0.472358,
		0.598814,
		0.666205,
		0.642924,
		0.530799,
		0.372177,
		0.222444,
		0.108744,
		0.036721,
		0.006041,
		0.000173,
		0.020173,
		0.078987,
		0.172561,
		0.294891,
		0.437117,
		0.585368,
		0.716734,
		0.797981,
		0.795954,
		0.700845,
		0.542935,
		0.372177,
		0.215674,
		0.09321,
		0.024116,
		0.002141,
		0.018011,
		0.070648,
		0.154777,
		0.265625,
		0.39646,
		0.538297,
		0.679351,
		0.803188,
		0.886448,
		0.902538,
		0.837497,
		0.707107,
		0.540983,
		0.354699,
		0.179781,
		0.05954,
		0.008349,
		0.016179,
		0.063557,
		0.139573,
		0.24037,
		0.360554,
		0.492972,
		0.628601,
		0.756663,
		0.864779,
		0.938519,
		0.96241,
		0.927549,
		0.839721,
		0.700845,
		0.505937,
		0.288229,
		0.112084,
		0.020197,
		0.014613,
		0.057479,
		0.12648,
		0.218463,
		0.32905,
		0.452462,
		0.581406,
		0.707107,
		0.819691,
		0.909189,
		0.967203,
		0.988808,
		0.974912,
		0.927549,
		0.825636,
		0.642924,
		0.401745,
		0.175889,
		0.037339,
		0.013263,
		0.05223,
		0.11513,
		0.199356,
		0.301315,
		0.416289,
		0.538323,
		0.660215,
		0.773702,
		0.870045,
		0.941362,
		0.983139,
		0.997575,
		0.993865,
		0.97297,
		0.902538,
		0.743241,
		0.5,
		0.240551,
		0.057659,
		0.012092,
		0.047667,
		0.105229,
		0.182603,
		0.276811,
		0.383959,
		0.499142,
		0.61641,
		0.728851,
		0.828924,
		0.909232,
		0.964076,
		0.992122,
		0.999659,
		0.9991,
		0.988808,
		0.935426,
		0.797981,
		0.567267,
		0.294361,
		0.077731,
		0.011069,
		0.043674,
		0.096544,
		0.167843,
		0.255081,
		0.355015,
		0.463572,
		0.575799,
		0.685901,
		0.787425,
		0.873716,
		0.938862,
		0.9794,
		0.996876,
		0.999982,
		0.999659,
		0.989142,
		0.938519,
		0.81321,
		0.598814,
		0.328938,
		0.093895,
		0.010171,
		0.040162,
		0.088884,
		0.154777,
		0.23574,
		0.32905,
		0.431293,
		0.538323,
		0.645261,
		0.746597,
		0.836467,
		0.909232,
		0.960553,
		0.989154,
		0.999046,
		1,
		0.998793,
		0.983139,
		0.926633,
		0.803188,
		0.600964,
		0.342523,
		0.103703,
		0.009378,
		0.037057,
		0.082095,
		0.14316,
		0.218463,
		0.3057,
		0.40199,
		0.503836,
		0.607109,
		0.707107,
		0.798713,
		0.876762,
		0.936718,
		0.975845,
		0.994912,
		0.999819,
		0.999986,
		0.996876,
		0.974494,
		0.909189,
		0.781582,
		0.585368,
		0.339533,
		0.106843,
		0.008674,
		0.034297,
		0.076051,
		0.132788,
		0.202976,
		0.28465,
		0.375363,
		0.472143,
		0.571482,
		0.669354,
		0.761317,
		0.842738,
		0.909232,
		0.957423,
		0.986146,
		0.997978,
		0.999989,
		0.999819,
		0.993706,
		0.964076,
		0.890001,
		0.756663,
		0.56247,
		0.327031,
		0.104887,
		0.008046,
		0.031834,
		0.070648,
		0.123492,
		0.189047,
		0.265625,
		0.351133,
		0.443032,
		0.538323,
		0.63355,
		0.72486,
		0.808146,
		0.879329,
		0.93486,
		0.972543,
		0.9927,
		0.999381,
		1,
		0.999236,
		0.989154,
		0.952128,
		0.870045,
		0.731605,
		0.538297,
		0.311037,
		0.100129,
		0.007484,
		0.029626,
		0.065798,
		0.11513,
		0.176478,
		0.248386,
		0.32905,
		0.416289,
		0.507526,
		0.599783,
		0.689715,
		0.773702,
		0.848031,
		0.909232,
		0.95463,
		0.983187,
		0.996571,
		0.999882,
		0.999991,
		0.997978,
		0.983187,
		0.938862,
		0.849616,
		0.707107,
		0.515011,
		0.294891,
		0.094456,
		0.006979,
		0.027641,
		0.061429,
		0.107582,
		0.165103,
		0.232725,
		0.308888,
		0.391707,
		0.478955,
		0.568056,
		0.656102,
		0.739908,
		0.816138,
		0.881522,
		0.933236,
		0.969489,
		0.990352,
		0.998635,
		0.999993,
		0.999882,
		0.995843,
		0.975845,
		0.924495,
		0.828924,
		0.683291,
		0.492972,
		0.279714,
		0.088879,
		0.006524,
		0.025848,
		0.057479,
		0.100747,
		0.154777,
		0.218463,
		0.290446,
		0.369091,
		0.452462,
		0.538323,
		0.624138,
		0.707107,
		0.78425,
		0.852555,
		0.909232,
		0.952128,
		0.980326,
		0.994912,
		0.999581,
		1,
		0.999493,
		0.9927,
		0.967218,
		0.909232,
		0.808146,
		0.660215,
		0.472143,
		0.265625,
		0.083718,
		0.006111,
		0.024223,
		0.053897,
		0.094539,
		0.145377,
		0.205443,
		0.273547,
		0.34826,
		0.427896,
		0.510503,
		0.593863,
		0.675515,
		0.752804,
		0.82299,
		0.883417,
		0.931803,
		0.966673,
		0.987949,
		0.997607,
		0.99992,
		0.999994,
		0.998635,
		0.988485,
		0.957423,
		0.893264,
		0.787425,
		0.637915,
		0.452462,
		0.25254,
		0.078987,
		0.005737,
		0.022747,
		0.050639,
		0.088884,
		0.136799,
		0.193529,
		0.258031,
		0.32905,
		0.405109,
		0.484498,
		0.565272,
		0.645261,
		0.722107,
		0.793332,
		0.856465,
		0.909232,
		0.949876,
		0.977591,
		0.993081,
		0.999046,
		0.999995,
		0.99992,
		0.997147,
		0.983187,
		0.946591,
		0.876762,
		0.766878,
		0.61641,
		0.433865,
		0.24037,
		0.074644,
		0.005396,
		0.021402,
		0.047667,
		0.083718,
		0.128949,
		0.182603,
		0.243759,
		0.31131,
		0.383959,
		0.460201,
		0.538323,
		0.61641,
		0.692366,
		0.763962,
		0.828924,
		0.88507,
		0.930531,
		0.964076,
		0.985548,
		0.996348,
		0.999707,
		1,
		0.999651,
		0.994912,
		0.976835,
		0.93486,
		0.859876,
		0.746597,
		0.595706,
		0.416289,
		0.229035,
		0.070648,
		0.005084,
		0.020173,
		0.044948,
		0.078987,
		0.121749,
		0.172561,
		0.230606,
		0.294906,
		0.364313,
		0.437501,
		0.512956,
		0.588982,
		0.663715,
		0.735156,
		0.801227,
		0.859876,
		0.909232,
		0.947842,
		0.974996,
		0.991144,
		0.99828,
		0.999944,
		0.999996,
		0.999046,
		0.991857,
		0.969489,
		0.922365,
		0.842738,
		0.726654,
		0.575799,
		0.399673,
		0.218463,
		0.066963,
		0.004799,
		0.019046,
		0.042454,
		0.074644,
		0.11513,
		0.163312,
		0.218463,
		0.279714,
		0.346048,
		0.416289,
		0.489096,
		0.562964,
		0.636233,
		0.707107,
		0.773702,
		0.834112,
		0.886525,
		0.929393,
		0.961681,
		0.983187,
		0.994912,
		0.999313,
		0.999997,
		0.999944,
		0.997978,
		0.987949,
		0.961225,
		0.909232,
		0.825463,
		0.707107,
		0.556677,
		0.383959,
		0.208589,
		0.063557,
		0.004537,
		0.018011,
		0.040162,
		0.070648,
		0.109032,
		0.154777,
		0.207232,
		0.265625,
		0.32905,
		0.39646,
		0.466665,
		0.538323,
		0.609954,
		0.679948,
		0.746597,
		0.808146,
		0.862877,
		0.909232,
		0.945995,
		0.972543,
		0.989154,
		0.997312,
		0.999788,
		1,
		0.999752,
		0.996348,
		0.983187,
		0.952128,
		0.895582,
		0.808146,
		0.687996,
		0.538323,
		0.369091,
		0.199356,
		0.060404,
		0.004295,
		0.017057,
		0.03805,
		0.066962,
		0.103402,
		0.146884,
		0.196828,
		0.25254,
		0.313213,
		0.377913,
		0.445578,
		0.515011,
		0.584888,
		0.653763,
		0.720092,
		0.782268,
		0.838682,
		0.887815,
		0.928371,
		0.959469,
		0.980889,
		0.99335,
		0.998733,
		0.999959,
		0.999997,
		0.999313,
		0.994084,
		0.977591,
		0.94229,
		0.881522,
		0.790869,
		0.669354,
		0.520718,
		0.355015,
		0.19071,
		0.057479,
		0.004071,
		0.016175,
		0.036097,
		0.063554,
		0.098192,
		0.139572,
		0.187172,
		0.24037,
		0.298442,
		0.360554,
		0.425753,
		0.492972,
		0.561021,
		0.628601,
		0.694316,
		0.756699,
		0.814257,
		0.865538,
		0.909232,
		0.944313,
		0.97023,
		0.987147,
		0.996177,
		0.999493,
		0.999998,
		0.999959,
		0.998528,
		0.991144,
		0.971204,
		0.931803,
		0.867151,
		0.773702,
		0.651201,
		0.503836,
		0.341684,
		0.182603,
		0.054761,
		0.003862,
		0.015352,
		0.034281,
		0.06039,
		0.093355,
		0.132781,
		0.178194,
		0.229034,
		0.28465,
		0.344293,
		0.407111,
		0.472143,
		0.538323,
		0.604481,
		0.669354,
		0.731605,
		0.789856,
		0.842738,
		0.888967,
		0.927446,
		0.957423,
		0.978669,
		0.991705,
		0.997978,
		0.999844,
		1,
		0.999819,
		0.997312,
		0.987511,
		0.964076,
		0.920755,
		0.852555,
		0.756699,
		0.63355,
		0.487653,
		0.32905,
		0.174994,
		0.05223,
		0.003662,
		0.014576,
		0.032575,
		0.057429,
		0.088841,
		0.12645,
		0.169822,
		0.218452,
		0.271752,
		0.329048,
		0.389572,
		0.452462,
		0.516759,
		0.581406,
		0.645261,
		0.707107,
		0.765677,
		0.819691,
		0.867913,
		0.909232,
		0.942774,
		0.968053,
		0.985151,
		0.994912,
		0.999046,
		0.99997,
		0.999998,
		0.999493,
		0.995601,
		0.983187,
		0.956268,
		0.909232,
		0.837812,
		0.739908,
		0.61641,
		0.472143,
		0.317068,
		0.167843,
		0.049871,
		0.003469,
		0.013828,
		0.030947,
		0.054623,
		0.084585,
		0.120501,
		0.161974,
		0.208539,
		0.259662,
		0.314732,
		0.373058,
		0.433863,
		0.496281,
		0.559361,
		0.622067,
		0.683291,
		0.741872,
		0.796619,
		0.846361,
		0.890001,
		0.926607,
		0.955526,
		0.976536,
		0.990011,
		0.997073,
		0.999618,
		0.999998,
		0.99997,
		0.998903,
		0.99335,
		0.978187,
		0.947842,
		0.897313,
		0.82299,
		0.723368,
		0.599783,
		0.457278,
		0.3057,
		0.161115,
		0.047667,
		0.003275,
		0.01309,
		0.029357,
		0.05191,
		0.080506,
		0.114839,
		0.154541,
		0.199181,
		0.248267,
		0.301242,
		0.357477,
		0.416269,
		0.476837,
		0.53832,
		0.599782,
		0.660215,
		0.718555,
		0.773702,
		0.824552,
		0.870045,
		0.909232,
		0.941362,
		0.966004,
		0.983187,
		0.993551,
		0.998451,
		0.999882,
		1,
		0.999864,
		0.997978,
		0.990536,
		0.972543,
		0.938862,
		0.88507,
		0.808146,
		0.707107,
		0.583667,
		0.443032,
		0.294906,
		0.154777,
		0.045605,
		0.003077,
		0.01234,
		0.027762,
		0.049219,
		0.076504,
		0.109339,
		0.147381,
		0.190228,
		0.237419,
		0.288436,
		0.342703,
		0.399576,
		0.458348,
		0.518237,
		0.578397,
		0.637912,
		0.695809,
		0.751076,
		0.802682,
		0.849616,
		0.890936,
		0.925841,
		0.953766,
		0.974494,
		0.988294,
		0.996042,
		0.999268,
		0.999977,
		0.999998,
		0.999618,
		0.996662,
		0.987147,
		0.966289,
		0.929393,
		0.872571,
		0.793332,
		0.69115,
		0.568056,
		0.429378,
		0.28465,
		0.148801,
		0.043674,
		0.002868,
		0.011559,
		0.026116,
		0.046472,
		0.072466,
		0.103854,
		0.140319,
		0.181483,
		0.22691,
		0.276112,
		0.328548,
		0.383622,
		0.440684,
		0.499021,
		0.557864,
		0.616381,
		0.673685,
		0.728847,
		0.780912,
		0.828924,
		0.87197,
		0.909232,
		0.940063,
		0.964076,
		0.981267,
		0.992122,
		0.997721,
		0.999707,
		0.999999,
		0.999977,
		0.999166,
		0.994912,
		0.983187,
		0.959469,
		0.919497,
		0.859876,
		0.778592,
		0.675515,
		0.552945,
		0.416289,
		0.2749,
		0.14316,
		0.041863,
		0.002646,
		0.010731,
		0.02438,
		0.043596,
		0.068277,
		0.098223,
		0.133151,
		0.172707,
		0.216478,
		0.263995,
		0.314743,
		0.368158,
		0.423631,
		0.480503,
		0.538067,
		0.595562,
		0.652181,
		0.707073,
		0.759357,
		0.808142,
		0.852554,
		0.891783,
		0.92514,
		0.952128,
		0.972543,
		0.986574,
		0.994912,
		0.998793,
		0.999909,
		1,
		0.999897,
		0.998451,
		0.9927,
		0.978669,
		0.952128,
		0.909232,
		0.84704,
		0.763962,
		0.660215,
		0.538323,
		0.403741,
		0.265625,
		0.137829,
		0.040162,
		0.00241,
		0.009847,
		0.022525,
		0.040531,
		0.063833,
		0.09229,
		0.125665,
		0.163637,
		0.205814,
		0.251747,
		0.300936,
		0.352837,
		0.406867,
		0.462401,
		0.518777,
		0.575293,
		0.631206,
		0.685735,
		0.738071,
		0.787387,
		0.832862,
		0.873711,
		0.909231,
		0.938862,
		0.962262,
		0.9794,
		0.99065,
		0.996876,
		0.999429,
		0.999982,
		0.999999,
		0.999707,
		0.997422,
		0.990011,
		0.973615,
		0.944313,
		0.898654,
		0.834112,
		0.749475,
		0.645261,
		0.52418,
		0.391707,
		0.256796,
		0.132788,
		0.038563,
		0.002161,
		0.008906,
		0.020541,
		0.037241,
		0.05906,
		0.085931,
		0.117674,
		0.154017,
		0.194598,
		0.238991,
		0.286708,
		0.337219,
		0.38995,
		0.444298,
		0.499627,
		0.555271,
		0.610534,
		0.664693,
		0.717001,
		0.766692,
		0.813001,
		0.855181,
		0.892541,
		0.92449,
		0.9506,
		0.970681,
		0.984868,
		0.993706,
		0.9982,
		0.999771,
		0.999999,
		0.999982,
		0.999355,
		0.996042,
		0.986838,
		0.968053,
		0.93607,
		0.887815,
		0.821134,
		0.735156,
		0.630658,
		0.510503,
		0.380165,
		0.248386,
		0.128015,
		0.037057,
		0.001902,
		0.007919,
		0.018437,
		0.033727,
		0.053931,
		0.07907,
		0.109042,
		0.143636,
		0.182539,
		0.225354,
		0.271617,
		0.320804,
		0.372351,
		0.425657,
		0.480096,
		0.535021,
		0.589761,
		0.643631,
		0.695929,
		0.745942,
		0.792954,
		0.836264,
		0.875207,
		0.909189,
		0.937734,
		0.960548,
		0.97759,
		0.989153,
		0.995933,
		0.999046,
		0.999929,
		1,
		0.99992,
		0.998793,
		0.994282,
		0.983187,
		0.962012,
		0.927446,
		0.876762,
		0.808146,
		0.721027,
		0.61641,
		0.497281,
		0.369091,
		0.24037,
		0.123492,
		0.035637,
		0.001639,
		0.006903,
		0.016247,
		0.030023,
		0.048471,
		0.071706,
		0.099716,
		0.132372,
		0.169425,
		0.210532,
		0.255258,
		0.3031,
		0.3535,
		0.405858,
		0.459545,
		0.513912,
		0.568298,
		0.622029,
		0.674429,
		0.724819,
		0.772521,
		0.816874,
		0.857243,
		0.893048,
		0.923796,
		0.94913,
		0.96889,
		0.983181,
		0.992443,
		0.997501,
		0.999548,
		0.999986,
		0.999999,
		0.999771,
		0.997978,
		0.992122,
		0.979066,
		0.955526,
		0.918486,
		0.865538,
		0.795181,
		0.707107,
		0.602518,
		0.484498,
		0.358463,
		0.232725,
		0.119202,
		0.034297,
		0.00138,
		0.005885,
		0.014018,
		0.026203,
		0.042764,
		0.063915,
		0.089745,
		0.120219,
		0.155176,
		0.194342,
		0.237338,
		0.283695,
		0.332875,
		0.384281,
		0.437279,
		0.491211,
		0.545402,
		0.599175,
		0.651854,
		0.702768,
		0.751262,
		0.796695,
		0.83846,
		0.875992,
		0.908799,
		0.936491,
		0.958834,
		0.975798,
		0.987631,
		0.994906,
		0.998558,
		0.999818,
		0.999999,
		0.999986,
		0.999493,
		0.996876,
		0.989554,
		0.974494,
		0.948631,
		0.909232,
		0.854185,
		0.782268,
		0.69341,
		0.588982,
		0.472143,
		0.34826,
		0.225429,
		0.11513,
		0.033031,
		0.001132,
		0.004895,
		0.011815,
		0.022365,
		0.036943,
		0.055852,
		0.079284,
		0.107308,
		0.139866,
		0.176778,
		0.21774,
		0.262345,
		0.310092,
		0.360406,
		0.412655,
		0.46617,
		0.520259,
		0.574223,
		0.627366,
		0.679005,
		0.728476,
		0.775143,
		0.818404,
		0.857704,
		0.892551,
		0.92254,
		0.94739,
		0.966984,
		0.981426,
		0.991097,
		0.996693,
		0.99923,
		0.999942,
		1,
		0.999937,
		0.999046,
		0.995458,
		0.986574,
		0.969489,
		0.941362,
		0.899725,
		0.842738,
		0.769433,
		0.679948,
		0.575799,
		0.460201,
		0.338462,
		0.218463,
		0.111261,
		0.031834,
		0.000902,
		0.003962,
		0.009702,
		0.018622,
		0.031171,
		0.047728,
		0.068577,
		0.093895,
		0.123736,
		0.158028,
		0.196569,
		0.239036,
		0.284996,
		0.333919,
		0.385195,
		0.438158,
		0.492103,
		0.546305,
		0.600037,
		0.652581,
		0.703245,
		0.751367,
		0.796331,
		0.837568,
		0.874577,
		0.906936,
		0.93433,
		0.956583,
		0.973698,
		0.985907,
		0.993717,
		0.99793,
		0.99962,
		0.999983,
		0.999998,
		0.999818,
		0.998392,
		0.993706,
		0.983187,
		0.964076,
		0.933754,
		0.890001,
		0.83123,
		0.756699,
		0.66673,
		0.562964,
		0.448658,
		0.32905,
		0.211807,
		0.107582,
		0.0307,
		0.000697,
		0.003111,
		0.007742,
		0.015087,
		0.025625,
		0.039786,
		0.057933,
		0.080339,
		0.107171,
		0.138472,
		0.174161,
		0.214025,
		0.257723,
		0.304801,
		0.354703,
		0.40679,
		0.460362,
		0.514679,
		0.568982,
		0.622512,
		0.674529,
		0.724327,
		0.771243,
		0.814675,
		0.854087,
		0.889025,
		0.919139,
		0.9442,
		0.964136,
		0.979067,
		0.989349,
		0.995602,
		0.998721,
		0.999806,
		0.999981,
		0.999983,
		0.999595,
		0.997501,
		0.991607,
		0.9794,
		0.95828,
		0.925841,
		0.880097,
		0.819691,
		0.744084,
		0.653763,
		0.550474,
		0.437501,
		0.320005,
		0.205443,
		0.104082,
		0.029626,
		0.000519,
		0.002362,
		0.005983,
		0.011859,
		0.020468,
		0.032269,
		0.04768,
		0.067052,
		0.090652,
		0.118641,
		0.151061,
		0.187825,
		0.228713,
		0.273376,
		0.321343,
		0.372038,
		0.424797,
		0.478887,
		0.533535,
		0.587947,
		0.641333,
		0.692926,
		0.742003,
		0.787895,
		0.830009,
		0.867837,
		0.900971,
		0.929125,
		0.952154,
		0.970086,
		0.983155,
		0.991829,
		0.996836,
		0.999139,
		0.999841,
		0.999952,
		0.999932,
		0.99923,
		0.996346,
		0.989153,
		0.975226,
		0.952128,
		0.917657,
		0.870045,
		0.808146,
		0.731605,
		0.641052,
		0.538323,
		0.426716,
		0.31131,
		0.199356,
		0.100747,
		0.028608,
		0.000371,
		0.001726,
		0.004461,
		0.009012,
		0.015835,
		0.025391,
		0.038126,
		0.054447,
		0.074699,
		0.099147,
		0.127953,
		0.161159,
		0.198682,
		0.240301,
		0.285665,
		0.334294,
		0.385597,
		0.438887,
		0.493408,
		0.548352,
		0.602893,
		0.656209,
		0.707502,
		0.756027,
		0.801108,
		0.842154,
		0.878677,
		0.91031,
		0.936827,
		0.958161,
		0.974436,
		0.985988,
		0.993394,
		0.997471,
		0.99923,
		0.999749,
		0.999886,
		0.999806,
		0.998688,
		0.994906,
		0.986344,
		0.970681,
		0.945646,
		0.909232,
		0.859876,
		0.796619,
		0.719275,
		0.628601,
		0.526505,
		0.416289,
		0.302949,
		0.193529,
		0.09757,
		0.027641,
		0.000253,
		0.001206,
		0.003192,
		0.006591,
		0.011818,
		0.019315,
		0.029527,
		0.04289,
		0.059802,
		0.080608,
		0.105574,
		0.134865,
		0.168534,
		0.206504,
		0.248563,
		0.294361,
		0.343419,
		0.395137,
		0.448814,
		0.50367,
		0.558873,
		0.613563,
		0.666884,
		0.71801,
		0.766169,
		0.810671,
		0.850921,
		0.886448,
		0.916918,
		0.942154,
		0.962163,
		0.977154,
		0.987562,
		0.994054,
		0.997522,
		0.999005,
		0.999523,
		0.999749,
		0.999561,
		0.99793,
		0.993163,
		0.983181,
		0.965783,
		0.938862,
		0.900598,
		0.849616,
		0.785132,
		0.707107,
		0.61641,
		0.515011,
		0.406208,
		0.294906,
		0.18795,
		0.094539,
		0.026722,
		0.000162,
		0.000799,
		0.002176,
		0.00461,
		0.008464,
		0.014139,
		0.02206,
		0.03266,
		0.046366,
		0.063573,
		0.084624,
		0.109792,
		0.139251,
		0.173066,
		0.211175,
		0.253379,
		0.299339,
		0.348584,
		0.400512,
		0.454416,
		0.5095,
		0.564909,
		0.619756,
		0.673152,
		0.724239,
		0.772221,
		0.816383,
		0.856127,
		0.890985,
		0.92065,
		0.944988,
		0.964069,
		0.978178,
		0.987832,
		0.993781,
		0.996969,
		0.998438,
		0.999123,
		0.999483,
		0.999139,
		0.996917,
		0.991097,
		0.979666,
		0.960548,
		0.931802,
		0.891783,
		0.839291,
		0.773702,
		0.695109,
		0.604481,
		0.503836,
		0.39646,
		0.287165,
		0.182603,
		0.091646,
		0.025848,
		9.7e-005,
		0.000496,
		0.001398,
		0.003057,
		0.005775,
		0.009899,
		0.015814,
		0.023933,
		0.034677,
		0.048465,
		0.06569,
		0.086698,
		0.111767,
		0.141088,
		0.174741,
		0.212684,
		0.254737,
		0.300582,
		0.349758,
		0.401677,
		0.455628,
		0.510811,
		0.566352,
		0.621342,
		0.674863,
		0.726028,
		0.774009,
		0.818069,
		0.857596,
		0.892122,
		0.921354,
		0.945195,
		0.963764,
		0.977409,
		0.986721,
		0.992508,
		0.995751,
		0.997467,
		0.998474,
		0.999005,
		0.99847,
		0.995602,
		0.988688,
		0.975798,
		0.954993,
		0.92449,
		0.882813,
		0.828924,
		0.762345,
		0.683291,
		0.592813,
		0.492972,
		0.387032,
		0.279714,
		0.177478,
		0.088884,
		0.025016,
		5.2e-005,
		0.000283,
		0.000834,
		0.001898,
		0.003714,
		0.00657,
		0.0108,
		0.016772,
		0.024886,
		0.035553,
		0.049186,
		0.066174,
		0.086869,
		0.11156,
		0.140451,
		0.173644,
		0.211121,
		0.252728,
		0.298172,
		0.347016,
		0.398686,
		0.452485,
		0.507609,
		0.563181,
		0.618271,
		0.67194,
		0.723272,
		0.771408,
		0.815588,
		0.855177,
		0.889702,
		0.918878,
		0.942628,
		0.961111,
		0.974726,
		0.984115,
		0.990133,
		0.993768,
		0.99599,
		0.997467,
		0.99821,
		0.997471,
		0.993929,
		0.985907,
		0.971575,
		0.94913,
		0.916947,
		0.873711,
		0.818535,
		0.751076,
		0.671658,
		0.581406,
		0.48241,
		0.377913,
		0.272538,
		0.172561,
		0.086243,
		0.024223,
		2.5e-005,
		0.000144,
		0.000451,
		0.001083,
		0.002217,
		0.004083,
		0.006952,
		0.011143,
		0.017011,
		0.024941,
		0.035336,
		0.0486,
		0.065122,
		0.08526,
		0.109311,
		0.137498,
		0.169947,
		0.206667,
		0.247536,
		0.292291,
		0.340527,
		0.391695,
		0.445116,
		0.5,
		0.555467,
		0.610581,
		0.664385,
		0.715937,
		0.764353,
		0.808844,
		0.848753,
		0.883593,
		0.913075,
		0.937137,
		0.955961,
		0.969981,
		0.979875,
		0.986518,
		0.990884,
		0.993865,
		0.995957,
		0.996969,
		0.996039,
		0.991829,
		0.982718,
		0.966984,
		0.942964,
		0.909189,
		0.864496,
		0.808142,
		0.739907,
		0.660215,
		0.570256,
		0.472143,
		0.369091,
		0.265625,
		0.167843,
		0.083718,
		0.023468,
		6.2e-005,
		0.000213,
		0.000551,
		0.0012,
		0.002331,
		0.004154,
		0.00693,
		0.010961,
		0.016587,
		0.02418,
		0.034131,
		0.046839,
		0.062693,
		0.082051,
		0.105226,
		0.132457,
		0.163895,
		0.19958,
		0.239425,
		0.283205,
		0.330552,
		0.38095,
		0.43375,
		0.488182,
		0.543376,
		0.598399,
		0.652282,
		0.704067,
		0.752845,
		0.7978,
		0.838252,
		0.873692,
		0.903823,
		0.928581,
		0.948161,
		0.963014,
		0.973833,
		0.981492,
		0.986925,
		0.990914,
		0.993768,
		0.995127,
		0.994054,
		0.989217,
		0.979067,
		0.961999,
		0.936491,
		0.901223,
		0.855181,
		0.797756,
		0.728847,
		0.648966,
		0.559361,
		0.462163,
		0.360554,
		0.258963,
		0.163312,
		0.081301,
		0.022747,
		2.1e-005,
		8.2e-005,
		0.000237,
		0.000567,
		0.001186,
		0.002251,
		0.00396,
		0.006556,
		0.010326,
		0.015596,
		0.022723,
		0.032087,
		0.04408,
		0.059088,
		0.077474,
		0.099561,
		0.125608,
		0.155789,
		0.190179,
		0.228728,
		0.271253,
		0.317429,
		0.366781,
		0.4187,
		0.472443,
		0.527165,
		0.581942,
		0.635805,
		0.687787,
		0.736961,
		0.782486,
		0.823657,
		0.859942,
		0.891026,
		0.916836,
		0.937562,
		0.953657,
		0.965809,
		0.974863,
		0.981687,
		0.986925,
		0.990698,
		0.992508,
		0.991373,
		0.985988,
		0.974887,
		0.956583,
		0.929694,
		0.893048,
		0.845772,
		0.787387,
		0.717902,
		0.637912,
		0.548717,
		0.452462,
		0.352291,
		0.25254,
		0.15896,
		0.078987,
		0.022059,
		2.2e-005,
		7.9e-005,
		0.000218,
		0.000512,
		0.001064,
		0.002017,
		0.003555,
		0.005905,
		0.009337,
		0.01416,
		0.020718,
		0.029379,
		0.040526,
		0.054541,
		0.071791,
		0.092608,
		0.117269,
		0.145975,
		0.178832,
		0.21583,
		0.256833,
		0.30156,
		0.34959,
		0.400354,
		0.453154,
		0.507174,
		0.561514,
		0.615215,
		0.667309,
		0.716858,
		0.763003,
		0.805015,
		0.842338,
		0.874633,
		0.901806,
		0.924032,
		0.941747,
		0.955611,
		0.966419,
		0.974941,
		0.981661,
		0.986518,
		0.988911,
		0.987832,
		0.982016,
		0.970086,
		0.950676,
		0.92254,
		0.884648,
		0.836264,
		0.777035,
		0.707073,
		0.627051,
		0.53832,
		0.443032,
		0.344293,
		0.246346,
		0.154777,
		0.07677,
		0.021402,
		1.6e-005,
		6e-005,
		0.00017,
		0.00041,
		0.000871,
		0.001682,
		0.003011,
		0.005068,
		0.008105,
		0.012416,
		0.018328,
		0.026196,
		0.036394,
		0.049299,
		0.06528,
		0.084677,
		0.107782,
		0.134822,
		0.165934,
		0.201151,
		0.240381,
		0.283397,
		0.32983,
		0.379165,
		0.430753,
		0.48382,
		0.537498,
		0.590853,
		0.642924,
		0.692773,
		0.73953,
		0.782445,
		0.820937,
		0.85464,
		0.883435,
		0.907465,
		0.927134,
		0.943052,
		0.95594,
		0.966443,
		0.974863,
		0.980982,
		0.984115,
		0.983245,
		0.977154,
		0.964555,
		0.9442,
		0.914977,
		0.875992,
		0.826642,
		0.766692,
		0.696355,
		0.616381,
		0.528163,
		0.433863,
		0.336549,
		0.24037,
		0.150754,
		0.074644,
		0.020774,
		3.5e-005,
		0.000112,
		0.00029,
		0.000649,
		0.001301,
		0.0024,
		0.004136,
		0.006746,
		0.010502,
		0.015716,
		0.02273,
		0.031906,
		0.043615,
		0.058225,
		0.076083,
		0.097494,
		0.122709,
		0.151896,
		0.185128,
		0.22236,
		0.26342,
		0.307995,
		0.35563,
		0.405731,
		0.457578,
		0.510346,
		0.563133,
		0.615002,
		0.665023,
		0.712321,
		0.756135,
		0.795861,
		0.831105,
		0.861713,
		0.887791,
		0.909691,
		0.927957,
		0.943208,
		0.95594,
		0.966268,
		0.973833,
		0.977887,
		0.977409,
		0.971234,
		0.958161,
		0.937057,
		0.906936,
		0.867033,
		0.816874,
		0.756338,
		0.685735,
		0.605889,
		0.518237,
		0.424946,
		0.329048,
		0.234602,
		0.146884,
		0.072605,
		0.020173,
		1.5e-005,
		6e-005,
		0.000178,
		0.000434,
		0.000926,
		0.00179,
		0.003197,
		0.005365,
		0.008547,
		0.013038,
		0.019164,
		0.027273,
		0.03773,
		0.0509,
		0.067132,
		0.086747,
		0.110012,
		0.137128,
		0.168203,
		0.203242,
		0.242125,
		0.284601,
		0.330276,
		0.37862,
		0.428973,
		0.480563,
		0.532534,
		0.58398,
		0.633993,
		0.681705,
		0.726348,
		0.767303,
		0.804146,
		0.836688,
		0.864986,
		0.889329,
		0.910174,
		0.928019,
		0.943187,
		0.955611,
		0.96481,
		0.969981,
		0.970107,
		0.964069,
		0.950751,
		0.929125,
		0.898322,
		0.857704,
		0.806913,
		0.745942,
		0.675192,
		0.595562,
		0.508529,
		0.416269,
		0.321779,
		0.229034,
		0.143159,
		0.070648,
		0.019597,
		2.4e-005,
		9e-005,
		0.000254,
		0.000597,
		0.001235,
		0.002323,
		0.004055,
		0.006667,
		0.010433,
		0.015662,
		0.022689,
		0.031869,
		0.04356,
		0.058116,
		0.075864,
		0.097092,
		0.122023,
		0.150806,
		0.183489,
		0.220006,
		0.260166,
		0.303641,
		0.349969,
		0.398556,
		0.448693,
		0.499578,
		0.55035,
		0.600132,
		0.648076,
		0.693416,
		0.735522,
		0.773948,
		0.808465,
		0.839077,
		0.865999,
		0.889586,
		0.910193,
		0.927957,
		0.942641,
		0.953657,
		0.960152,
		0.961111,
		0.955459,
		0.942154,
		0.920262,
		0.889025,
		0.847917,
		0.796695,
		0.735456,
		0.664693,
		0.585374,
		0.499021,
		0.407819,
		0.314732,
		0.223654,
		0.139572,
		0.068768,
		0.019046,
		3.5e-005,
		0.000124,
		0.000339,
		0.000776,
		0.001568,
		0.00289,
		0.004955,
		0.008018,
		0.012367,
		0.018324,
		0.026229,
		0.036436,
		0.049296,
		0.065143,
		0.084279,
		0.106953,
		0.133344,
		0.163544,
		0.197539,
		0.235198,
		0.276259,
		0.320332,
		0.366892,
		0.415302,
		0.464821,
		0.514646,
		0.56394,
		0.611886,
		0.657731,
		0.700845,
		0.740761,
		0.777215,
		0.810153,
		0.839707,
		0.86612,
		0.889586,
		0.910055,
		0.927134,
		0.940135,
		0.948161,
		0.950194,
		0.945195,
		0.932184,
		0.91031,
		0.878913,
		0.837568,
		0.786138,
		0.724819,
		0.654192,
		0.575293,
		0.48969,
		0.399576,
		0.307892,
		0.218452,
		0.136115,
		0.066962,
		0.018518,
		4.6e-005,
		0.000161,
		0.000431,
		0.000966,
		0.001918,
		0.003479,
		0.00588,
		0.009391,
		0.014314,
		0.020976,
		0.029723,
		0.0409,
		0.054846,
		0.071873,
		0.092251,
		0.116188,
		0.143817,
		0.175172,
		0.21018,
		0.248646,
		0.29025,
		0.334543,
		0.38096,
		0.428835,
		0.477426,
		0.52595,
		0.57363,
		0.61974,
		0.663655,
		0.704898,
		0.74317,
		0.778361,
		0.810517,
		0.839755,
		0.866103,
		0.889329,
		0.908898,
		0.924032,
		0.933787,
		0.937137,
		0.933069,
		0.92065,
		0.899099,
		0.867837,
		0.826533,
		0.775143,
		0.713954,
		0.643631,
		0.565277,
		0.480503,
		0.39152,
		0.301242,
		0.213418,
		0.132781,
		0.065226,
		0.018011,
		1.1e-005,
		5.9e-005,
		0.000201,
		0.000527,
		0.001164,
		0.002279,
		0.00408,
		0.006814,
		0.010764,
		0.016241,
		0.023578,
		0.033116,
		0.045192,
		0.060128,
		0.078208,
		0.099669,
		0.124676,
		0.153312,
		0.185558,
		0.221283,
		0.260237,
		0.302048,
		0.346229,
		0.392189,
		0.439258,
		0.486716,
		0.533837,
		0.579929,
		0.624385,
		0.666724,
		0.706621,
		0.743911,
		0.778551,
		0.810531,
		0.839707,
		0.865681,
		0.887791,
		0.905174,
		0.916836,
		0.921738,
		0.918878,
		0.907357,
		0.886448,
		0.855639,
		0.814675,
		0.763595,
		0.702768,
		0.632938,
		0.555271,
		0.471423,
		0.383622,
		0.294765,
		0.208539,
		0.129562,
		0.063554,
		0.017525,
		1.3e-005,
		7.2e-005,
		0.000243,
		0.000627,
		0.001366,
		0.002644,
		0.004682,
		0.007741,
		0.012115,
		0.01812,
		0.02609,
		0.03636,
		0.049254,
		0.065071,
		0.084068,
		0.106443,
		0.13232,
		0.161731,
		0.194608,
		0.230768,
		0.269916,
		0.311642,
		0.355436,
		0.400704,
		0.446798,
		0.493048,
		0.53881,
		0.583504,
		0.626654,
		0.667916,
		0.707073,
		0.743999,
		0.778551,
		0.810427,
		0.839077,
		0.863728,
		0.883435,
		0.897154,
		0.903823,
		0.902438,
		0.892122,
		0.872181,
		0.842154,
		0.801847,
		0.751367,
		0.691157,
		0.622029,
		0.54521,
		0.462401,
		0.375848,
		0.288436,
		0.203799,
		0.12645,
		0.061944,
		0.017057,
		1.6e-005,
		8.6e-005,
		0.000285,
		0.000728,
		0.00157,
		0.003007,
		0.005276,
		0.008648,
		0.013424,
		0.019924,
		0.028478,
		0.039411,
		0.053032,
		0.069614,
		0.089384,
		0.112501,
		0.139047,
		0.169008,
		0.202269,
		0.238608,
		0.277694,
		0.319097,
		0.362304,
		0.406741,
		0.451806,
		0.496909,
		0.541506,
		0.585137,
		0.627445,
		0.668171,
		0.707107,
		0.743987,
		0.778361,
		0.809553,
		0.836688,
		0.858746,
		0.874633,
		0.883254,
		0.883593,
		0.874775,
		0.856127,
		0.827216,
		0.787895,
		0.738324,
		0.679005,
		0.61081,
		0.535021,
		0.453381,
		0.368158,
		0.282229,
		0.199181,
		0.123433,
		0.06039,
		0.016608,
		1.9e-005,
		0.0001,
		0.000328,
		0.000829,
		0.001771,
		0.003363,
		0.005853,
		0.009521,
		0.014673,
		0.021629,
		0.030711,
		0.042233,
		0.056483,
		0.073709,
		0.094104,
		0.117792,
		0.144812,
		0.17511,
		0.208535,
		0.244832,
		0.283651,
		0.324562,
		0.367068,
		0.410642,
		0.454751,
		0.498898,
		0.542647,
		0.585637,
		0.627576,
		0.668181,
		0.707073,
		0.743679,
		0.777215,
		0.80672,
		0.831105,
		0.849223,
		0.859942,
		0.862222,
		0.855177,
		0.838132,
		0.810671,
		0.77267,
		0.724327,
		0.666188,
		0.599175,
		0.524618,
		0.444298,
		0.360506,
		0.276112,
		0.194665,
		0.120501,
		0.058887,
		0.016175,
		2.2e-005,
		0.000114,
		0.000371,
		0.000929,
		0.001968,
		0.003708,
		0.006407,
		0.010351,
		0.015848,
		0.023214,
		0.032763,
		0.044794,
		0.059573,
		0.077319,
		0.098195,
		0.122288,
		0.1496,
		0.180046,
		0.213445,
		0.249525,
		0.287937,
		0.328265,
		0.370055,
		0.412843,
		0.456182,
		0.499666,
		0.542944,
		0.585696,
		0.627576,
		0.668108,
		0.706621,
		0.742246,
		0.773948,
		0.800578,
		0.820937,
		0.833854,
		0.838252,
		0.83322,
		0.818069,
		0.792377,
		0.756027,
		0.709237,
		0.652581,
		0.587017,
		0.513912,
		0.435083,
		0.352837,
		0.270048,
		0.190228,
		0.117642,
		0.057429,
		0.015757,
		2.4e-005,
		0.000127,
		0.000412,
		0.001025,
		0.002157,
		0.004037,
		0.00693,
		0.011126,
		0.016933,
		0.024662,
		0.034615,
		0.047072,
		0.062277,
		0.080424,
		0.101641,
		0.125984,
		0.153429,
		0.183862,
		0.217088,
		0.252835,
		0.290769,
		0.330511,
		0.371666,
		0.413844,
		0.456678,
		0.49983,
		0.542966,
		0.585688,
		0.627445,
		0.667493,
		0.704898,
		0.738568,
		0.767303,
		0.789857,
		0.805015,
		0.81166,
		0.808844,
		0.795845,
		0.772221,
		0.737842,
		0.692926,
		0.63806,
		0.574223,
		0.502809,
		0.425657,
		0.345096,
		0.263995,
		0.185844,
		0.114839,
		0.05601,
		0.015352,
		2.7e-005,
		0.000141,
		0.000452,
		0.001117,
		0.002336,
		0.004345,
		0.007416,
		0.011839,
		0.01792,
		0.025961,
		0.036251,
		0.049052,
		0.064584,
		0.083014,
		0.104444,
		0.128904,
		0.156346,
		0.186644,
		0.219602,
		0.254963,
		0.292425,
		0.331662,
		0.372344,
		0.414148,
		0.45676,
		0.499836,
		0.542944,
		0.585481,
		0.626654,
		0.665486,
		0.700845,
		0.73149,
		0.756135,
		0.773521,
		0.782486,
		0.782037,
		0.771408,
		0.750119,
		0.71801,
		0.675281,
		0.622512,
		0.560686,
		0.491211,
		0.415942,
		0.337219,
		0.257911,
		0.181483,
		0.112078,
		0.054623,
		0.014959,
		3e-005,
		0.000154,
		0.000491,
		0.001205,
		0.002503,
		0.004631,
		0.00786,
		0.012483,
		0.018799,
		0.0271,
		0.037661,
		0.050725,
		0.06649,
		0.085099,
		0.10663,
		0.131095,
		0.158434,
		0.18852,
		0.22117,
		0.256157,
		0.293225,
		0.332104,
		0.37252,
		0.414184,
		0.45676,
		0.49979,
		0.542647,
		0.584512,
		0.624385,
		0.661108,
		0.693416,
		0.719989,
		0.73953,
		0.75083,
		0.752845,
		0.744755,
		0.726028,
		0.696458,
		0.656209,
		0.605834,
		0.546305,
		0.479026,
		0.405858,
		0.329143,
		0.251747,
		0.177115,
		0.109339,
		0.053259,
		0.014576,
		3.3e-005,
		0.000166,
		0.000526,
		0.001285,
		0.002657,
		0.00489,
		0.008259,
		0.013053,
		0.019564,
		0.028075,
		0.038844,
		0.052095,
		0.068007,
		0.086702,
		0.108244,
		0.132632,
		0.159806,
		0.189652,
		0.222012,
		0.256697,
		0.293501,
		0.332198,
		0.372533,
		0.414179,
		0.456678,
		0.499394,
		0.541506,
		0.58201,
		0.61974,
		0.653418,
		0.681705,
		0.703269,
		0.716858,
		0.721373,
		0.715937,
		0.699951,
		0.673152,
		0.635646,
		0.587947,
		0.530993,
		0.46617,
		0.395328,
		0.320804,
		0.245457,
		0.172707,
		0.106605,
		0.05191,
		0.014199,
		3.5e-005,
		0.000177,
		0.00056,
		0.00136,
		0.002796,
		0.005121,
		0.008609,
		0.013546,
		0.020214,
		0.028884,
		0.039801,
		0.053172,
		0.069157,
		0.087866,
		0.109353,
		0.133616,
		0.160605,
		0.190229,
		0.222364,
		0.256861,
		0.293546,
		0.332201,
		0.37252,
		0.414053,
		0.456182,
		0.498098,
		0.53881,
		0.577165,
		0.611886,
		0.641623,
		0.665023,
		0.680795,
		0.687787,
		0.685059,
		0.67194,
		0.648096,
		0.613563,
		0.568794,
		0.514679,
		0.452566,
		0.384281,
		0.312142,
		0.238991,
		0.168226,
		0.103854,
		0.050566,
		0.013828,
		3.7e-005,
		0.000188,
		0.00059,
		0.001426,
		0.00292,
		0.005324,
		0.00891,
		0.01396,
		0.020748,
		0.029533,
		0.040544,
		0.053976,
		0.069976,
		0.088647,
		0.110041,
		0.134165,
		0.160989,
		0.19045,
		0.222455,
		0.25688,
		0.293546,
		0.332175,
		0.372344,
		0.413461,
		0.454751,
		0.495252,
		0.533837,
		0.569245,
		0.600132,
		0.625132,
		0.642924,
		0.652309,
		0.652282,
		0.642099,
		0.621342,
		0.589968,
		0.548352,
		0.497317,
		0.438158,
		0.372656,
		0.3031,
		0.232304,
		0.163637,
		0.101067,
		0.049219,
		0.013459,
		3.9e-005,
		0.000197,
		0.000617,
		0.001485,
		0.003027,
		0.005496,
		0.009162,
		0.014298,
		0.021172,
		0.03003,
		0.041091,
		0.054538,
		0.070513,
		0.089116,
		0.110408,
		0.134412,
		0.161121,
		0.190496,
		0.222461,
		0.256878,
		0.293501,
		0.331948,
		0.371666,
		0.411925,
		0.451806,
		0.490224,
		0.52595,
		0.557657,
		0.58398,
		0.603579,
		0.615215,
		0.617822,
		0.610581,
		0.592988,
		0.564909,
		0.526629,
		0.478887,
		0.422905,
		0.360406,
		0.29363,
		0.225354,
		0.15891,
		0.098223,
		0.047857,
		0.01309,
		4.1e-005,
		0.000206,
		0.00064,
		0.001535,
		0.003117,
		0.005638,
		0.009364,
		0.014563,
		0.021491,
		0.030388,
		0.041464,
		0.054895,
		0.070823,
		0.089354,
		0.110561,
		0.134487,
		0.161143,
		0.190498,
		0.222455,
		0.25681,
		0.293225,
		0.331199,
		0.370055,
		0.408933,
		0.446798,
		0.482466,
		0.514646,
		0.541984,
		0.563133,
		0.576825,
		0.581942,
		0.577591,
		0.563181,
		0.538479,
		0.50367,
		0.459397,
		0.40679,
		0.347498,
		0.283695,
		0.218106,
		0.154017,
		0.095303,
		0.046472,
		0.012718,
		4.3e-005,
		0.000213,
		0.000661,
		0.001578,
		0.003191,
		0.005752,
		0.009521,
		0.014759,
		0.021718,
		0.030628,
		0.041695,
		0.055094,
		0.070972,
		0.089445,
		0.1106,
		0.134495,
		0.161143,
		0.190485,
		0.222364,
		0.256491,
		0.292425,
		0.329549,
		0.367068,
		0.404013,
		0.439258,
		0.471554,
		0.499578,
		0.521988,
		0.537498,
		0.544946,
		0.543376,
		0.532108,
		0.510811,
		0.47956,
		0.438887,
		0.389823,
		0.333919,
		0.27327,
		0.210532,
		0.148931,
		0.09229,
		0.045055,
		0.01234,
		4.4e-005,
		0.000219,
		0.000677,
		0.001612,
		0.00325,
		0.005838,
		0.009635,
		0.014896,
		0.021866,
		0.030771,
		0.041817,
		0.055183,
		0.071023,
		0.089464,
		0.110603,
		0.134494,
		0.161121,
		0.190372,
		0.222012,
		0.25566,
		0.290769,
		0.326616,
		0.362304,
		0.396774,
		0.428835,
		0.457205,
		0.480563,
		0.497617,
		0.507174,
		0.508222,
		0.5,
		0.482077,
		0.454416,
		0.417432,
		0.372038,
		0.319677,
		0.262345,
		0.202613,
		0.143636,
		0.08917,
		0.043596,
		0.011955,
		4.5e-005,
		0.000224,
		0.000691,
		0.001639,
		0.003294,
		0.005901,
		0.009713,
		0.014983,
		0.021951,
		0.030844,
		0.041868,
		0.05521,
		0.071031,
		0.089465,
		0.1106,
		0.134464,
		0.160989,
		0.189997,
		0.22117,
		0.254032,
		0.287937,
		0.322066,
		0.355436,
		0.386925,
		0.415302,
		0.43928,
		0.457578,
		0.468986,
		0.472443,
		0.467118,
		0.452485,
		0.428393,
		0.395137,
		0.353499,
		0.304801,
		0.250926,
		0.194342,
		0.138117,
		0.085931,
		0.04209,
		0.011559,
		4.6e-005,
		0.000228,
		0.000701,
		0.001658,
		0.003325,
		0.005942,
		0.009761,
		0.015031,
		0.021992,
		0.030871,
		0.041881,
		0.055213,
		0.071031,
		0.089459,
		0.110561,
		0.134318,
		0.160605,
		0.189167,
		0.219602,
		0.251342,
		0.283651,
		0.31563,
		0.346229,
		0.374285,
		0.398556,
		0.417783,
		0.430753,
		0.436373,
		0.43375,
		0.422271,
		0.401677,
		0.372134,
		0.334294,
		0.289343,
		0.239036,
		0.185725,
		0.132372,
		0.082565,
		0.040531,
		0.011152,
		4.7e-005,
		0.000232,
		0.000709,
		0.001672,
		0.003345,
		0.005967,
		0.009787,
		0.015052,
		0.022005,
		0.030877,
		0.041882,
		0.055213,
		0.071023,
		0.089413,
		0.110408,
		0.133936,
		0.159806,
		0.187686,
		0.217088,
		0.247362,
		0.277694,
		0.307118,
		0.334543,
		0.358785,
		0.37862,
		0.392847,
		0.400354,
		0.400201,
		0.391695,
		0.374475,
		0.348584,
		0.314536,
		0.273376,
		0.226717,
		0.176778,
		0.126401,
		0.07907,
		0.038915,
		0.010731,
		4.8e-005,
		0.000234,
		0.000714,
		0.001681,
		0.003356,
		0.005979,
		0.009797,
		0.015059,
		0.022008,
		0.030877,
		0.041881,
		0.055202,
		0.070972,
		0.089259,
		0.110041,
		0.133186,
		0.158434,
		0.185375,
		0.213445,
		0.241918,
		0.269916,
		0.296425,
		0.320332,
		0.340459,
		0.35563,
		0.36473,
		0.366781,
		0.361029,
		0.347016,
		0.324667,
		0.294361,
		0.256992,
		0.214025,
		0.167531,
		0.120219,
		0.075448,
		0.037241,
		0.010297,
		4.8e-005,
		0.000235,
		0.000717,
		0.001686,
		0.003362,
		0.005984,
		0.0098,
		0.015059,
		0.022008,
		0.030876,
		0.041868,
		0.055149,
		0.070823,
		0.088916,
		0.109353,
		0.13194,
		0.156346,
		0.18209,
		0.208535,
		0.234893,
		0.260237,
		0.283525,
		0.303641,
		0.319444,
		0.32983,
		0.333802,
		0.330552,
		0.31954,
		0.300582,
		0.27392,
		0.240301,
		0.201032,
		0.158028,
		0.113845,
		0.071706,
		0.035511,
		0.009847,
		4.8e-005,
		0.000236,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022005,
		0.030862,
		0.041817,
		0.055011,
		0.070513,
		0.0883,
		0.108244,
		0.130086,
		0.153429,
		0.177722,
		0.202269,
		0.22623,
		0.248646,
		0.268467,
		0.284601,
		0.295966,
		0.30156,
		0.300538,
		0.292291,
		0.276532,
		0.253379,
		0.223428,
		0.187825,
		0.14832,
		0.107308,
		0.067856,
		0.033727,
		0.009384,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015057,
		0.021992,
		0.030815,
		0.041695,
		0.054739,
		0.069976,
		0.087336,
		0.10663,
		0.127539,
		0.1496,
		0.172206,
		0.194608,
		0.21593,
		0.235198,
		0.251376,
		0.26342,
		0.270336,
		0.271253,
		0.265509,
		0.252728,
		0.232908,
		0.206504,
		0.1745,
		0.138472,
		0.100644,
		0.063915,
		0.031895,
		0.008906,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.009797,
		0.015044,
		0.021951,
		0.03071,
		0.041464,
		0.054285,
		0.069157,
		0.085958,
		0.104444,
		0.124235,
		0.144812,
		0.165515,
		0.185558,
		0.204045,
		0.220006,
		0.232442,
		0.240381,
		0.242943,
		0.239425,
		0.229374,
		0.212684,
		0.18967,
		0.161159,
		0.128554,
		0.093895,
		0.059906,
		0.030023,
		0.008417,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005982,
		0.009787,
		0.015011,
		0.021866,
		0.030522,
		0.041091,
		0.053606,
		0.068007,
		0.084118,
		0.101641,
		0.12014,
		0.139047,
		0.157661,
		0.175172,
		0.19068,
		0.203242,
		0.211916,
		0.21583,
		0.214251,
		0.206667,
		0.192875,
		0.173066,
		0.147912,
		0.118641,
		0.08711,
		0.055852,
		0.028122,
		0.007919,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003362,
		0.005974,
		0.009761,
		0.014945,
		0.021718,
		0.030225,
		0.040544,
		0.052669,
		0.06649,
		0.081783,
		0.098195,
		0.115245,
		0.13232,
		0.148694,
		0.163544,
		0.175988,
		0.185128,
		0.190107,
		0.190179,
		0.184787,
		0.173644,
		0.156828,
		0.134865,
		0.108813,
		0.080339,
		0.051783,
		0.026203,
		0.007413,
		4.9e-005,
		0.000237,
		0.000719,
		0.001687,
		0.003356,
		0.005956,
		0.009713,
		0.014835,
		0.021491,
		0.029799,
		0.039801,
		0.051448,
		0.064584,
		0.078936,
		0.094104,
		0.109565,
		0.124676,
		0.138695,
		0.150806,
		0.160162,
		0.165934,
		0.167375,
		0.163895,
		0.155141,
		0.141088,
		0.122125,
		0.099147,
		0.073636,
		0.047728,
		0.024279,
		0.006903,
		4.9e-005,
		0.000237,
		0.000719,
		0.001684,
		0.003345,
		0.005924,
		0.009635,
		0.014669,
		0.021172,
		0.029228,
		0.038844,
		0.049927,
		0.062277,
		0.075576,
		0.089384,
		0.103142,
		0.116188,
		0.12778,
		0.137128,
		0.143438,
		0.145975,
		0.144131,
		0.137498,
		0.125963,
		0.109792,
		0.089722,
		0.067052,
		0.043719,
		0.022365,
		0.006393,
		4.9e-005,
		0.000236,
		0.000717,
		0.001677,
		0.003325,
		0.005872,
		0.009521,
		0.014439,
		0.020748,
		0.0285,
		0.037661,
		0.0481,
		0.059573,
		0.07172,
		0.084068,
		0.096035,
		0.106953,
		0.116093,
		0.122709,
		0.126086,
		0.125608,
		0.120829,
		0.11156,
		0.097956,
		0.080608,
		0.06064,
		0.039786,
		0.020474,
		0.005885,
		4.9e-005,
		0.000236,
		0.000714,
		0.001666,
		0.003294,
		0.005798,
		0.009364,
		0.014139,
		0.020214,
		0.027608,
		0.036251,
		0.04597,
		0.056483,
		0.067396,
		0.078208,
		0.088328,
		0.097092,
		0.103804,
		0.107782,
		0.108413,
		0.105226,
		0.097968,
		0.086698,
		0.071873,
		0.054447,
		0.03596,
		0.018622,
		0.005385,
		4.8e-005,
		0.000235,
		0.000709,
		0.001649,
		0.00325,
		0.005698,
		0.009162,
		0.013763,
		0.019564,
		0.026551,
		0.034615,
		0.043547,
		0.053032,
		0.062646,
		0.071873,
		0.08012,
		0.086747,
		0.091109,
		0.092608,
		0.090758,
		0.08526,
		0.076086,
		0.063573,
		0.048517,
		0.032269,
		0.016822,
		0.004895,
		4.8e-005,
		0.000233,
		0.000701,
		0.001626,
		0.003191,
		0.005571,
		0.00891,
		0.013309,
		0.018799,
		0.025331,
		0.032763,
		0.040853,
		0.049254,
		0.057525,
		0.065143,
		0.07153,
		0.076083,
		0.078227,
		0.077474,
		0.073489,
		0.066174,
		0.055758,
		0.04289,
		0.028739,
		0.015087,
		0.00442,
		4.8e-005,
		0.00023,
		0.000691,
		0.001596,
		0.003117,
		0.005413,
		0.008609,
		0.012777,
		0.01792,
		0.023956,
		0.030711,
		0.037912,
		0.045192,
		0.052099,
		0.058116,
		0.062691,
		0.06528,
		0.0654,
		0.062693,
		0.057003,
		0.048465,
		0.037596,
		0.025391,
		0.01343,
		0.003962,
		4.7e-005,
		0.000227,
		0.000677,
		0.001557,
		0.003027,
		0.005226,
		0.008259,
		0.01217,
		0.016933,
		0.022437,
		0.028478,
		0.034759,
		0.0409,
		0.046446,
		0.0509,
		0.053754,
		0.054541,
		0.052889,
		0.0486,
		0.041725,
		0.03266,
		0.022245,
		0.011859,
		0.003525,
		4.6e-005,
		0.000222,
		0.000661,
		0.001511,
		0.00292,
		0.005009,
		0.00786,
		0.011491,
		0.015848,
		0.02079,
		0.02609,
		0.031435,
		0.036436,
		0.040651,
		0.043615,
		0.044882,
		0.04408,
		0.040976,
		0.035553,
		0.028102,
		0.019315,
		0.010384,
		0.003111,
		4.5e-005,
		0.000216,
		0.00064,
		0.001456,
		0.002796,
		0.004764,
		0.007416,
		0.010746,
		0.014673,
		0.019033,
		0.023578,
		0.027985,
		0.031869,
		0.034811,
		0.036394,
		0.036252,
		0.034131,
		0.029959,
		0.023933,
		0.016611,
		0.009012,
		0.002723,
		4.4e-005,
		0.00021,
		0.000617,
		0.001394,
		0.002657,
		0.004491,
		0.00693,
		0.009942,
		0.013424,
		0.017189,
		0.020976,
		0.024461,
		0.027273,
		0.02903,
		0.029379,
		0.028054,
		0.024941,
		0.020157,
		0.014139,
		0.007747,
		0.002362,
		4.3e-005,
		0.000202,
		0.00059,
		0.001323,
		0.002503,
		0.004194,
		0.006407,
		0.009089,
		0.012115,
		0.015282,
		0.018324,
		0.020921,
		0.02273,
		0.023421,
		0.022723,
		0.02049,
		0.016772,
		0.011902,
		0.006591,
		0.002029,
		4.1e-005,
		0.000193,
		0.00056,
		0.001246,
		0.002336,
		0.003874,
		0.005853,
		0.008198,
		0.010764,
		0.013341,
		0.015662,
		0.017426,
		0.018328,
		0.018105,
		0.016587,
		0.013772,
		0.009899,
		0.005545,
		0.001726,
		3.9e-005,
		0.000183,
		0.000526,
		0.001162,
		0.002157,
		0.003537,
		0.005276,
		0.007279,
		0.009391,
		0.011397,
		0.013038,
		0.014044,
		0.01416,
		0.013209,
		0.011143,
		0.008124,
		0.00461,
		0.001451,
		3.7e-005,
		0.000172,
		0.000491,
		0.001072,
		0.001968,
		0.003186,
		0.004682,
		0.006347,
		0.008018,
		0.009482,
		0.010502,
		0.010844,
		0.010326,
		0.008869,
		0.00657,
		0.003782,
		0.001206,
		3.5e-005,
		0.00016,
		0.000452,
		0.000977,
		0.001771,
		0.002826,
		0.00408,
		0.005416,
		0.006667,
		0.007633,
		0.008105,
		0.007904,
		0.00693,
		0.005227,
		0.003057,
		0.000989,
		3.3e-005,
		0.000147,
		0.000412,
		0.000879,
		0.00157,
		0.002461,
		0.003479,
		0.004501,
		0.005365,
		0.005889,
		0.005905,
		0.005302,
		0.004083,
		0.002431,
		0.000799,
		3e-005,
		0.000134,
		0.000371,
		0.000779,
		0.001366,
		0.002098,
		0.00289,
		0.00362,
		0.004136,
		0.004288,
		0.00396,
		0.003122,
		0.001898,
		0.000635,
		2.7e-005,
		0.000121,
		0.000328,
		0.000678,
		0.001164,
		0.001742,
		0.002323,
		0.00279,
		0.003011,
		0.002874,
		0.002331,
		0.001451,
		0.000496,
		2.4e-005,
		0.000107,
		0.000285,
		0.000577,
		0.000966,
		0.001399,
		0.00179,
		0.00203,
		0.002017,
		0.001691,
		0.001083,
		0.000379,
		2.2e-005,
		9.3e-005,
		0.000243,
		0.000479,
		0.000776,
		0.001077,
		0.001301,
		0.001359,
		0.001186,
		0.000785,
		0.000283,
		1.9e-005,
		7.9e-005,
		0.000201,
		0.000384,
		0.000597,
		0.000783,
		0.000871,
		0.000799,
		0.000551,
		0.000205,
		1.6e-005,
		6.5e-005,
		0.000161,
		0.000295,
		0.000434,
		0.000524,
		0.000512,
		0.000371,
		0.000144,
		1.3e-005,
		5.2e-005,
		0.000124,
		0.000215,
		0.00029,
		0.000308,
		0.000237,
		9.7e-005,
		1.1e-005,
		4e-005,
		9e-005,
		0.000143,
		0.00017,
		0.000143,
		6.2e-005,
		2.9e-005,
		6e-005,
		8.4e-005,
		7.9e-005,
		3.7e-005,
		2e-005,
		3.5e-005,
		3.9e-005,
		2.1e-005,
		1.1e-005,
		1.6e-005,
		1e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		1.3e-005,
		2.2e-005,
		3.2e-005,
		4.3e-005,
		5.6e-005,
		6.9e-005,
		8.2e-005,
		9.6e-005,
		0.00011,
		0.000124,
		0.000138,
		0.000151,
		0.000163,
		0.000175,
		0.000185,
		0.000195,
		0.000204,
		0.000211,
		0.000218,
		0.000223,
		0.000228,
		0.000231,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		4.9e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000237,
		4.9e-005,
		3e-005,
		7.2e-005,
		0.000127,
		0.000196,
		0.000274,
		0.000361,
		0.000454,
		0.000552,
		0.000652,
		0.000754,
		0.000854,
		0.000953,
		0.001049,
		0.00114,
		0.001225,
		0.001305,
		0.001377,
		0.001442,
		0.001498,
		0.001547,
		0.001587,
		0.001619,
		0.001644,
		0.001662,
		0.001675,
		0.001682,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.001688,
		0.000719,
		0.000236,
		4.8e-005,
		5.7e-005,
		0.000179,
		0.000357,
		0.000585,
		0.000854,
		0.001156,
		0.001483,
		0.00183,
		0.002188,
		0.002552,
		0.002916,
		0.003275,
		0.003623,
		0.003956,
		0.00427,
		0.004562,
		0.004828,
		0.005066,
		0.005276,
		0.005456,
		0.005605,
		0.005726,
		0.005819,
		0.005887,
		0.005933,
		0.005962,
		0.005977,
		0.005983,
		0.005985,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000236,
		4.8e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.005985,
		0.003363,
		0.001686,
		0.000717,
		0.000235,
		4.8e-005,
		5.5e-005,
		0.000287,
		0.000679,
		0.001208,
		0.001853,
		0.002592,
		0.003407,
		0.004277,
		0.005185,
		0.006113,
		0.007047,
		0.00797,
		0.00887,
		0.009733,
		0.01055,
		0.011311,
		0.012007,
		0.012633,
		0.013183,
		0.013657,
		0.014052,
		0.014371,
		0.014618,
		0.014799,
		0.014922,
		0.014998,
		0.015038,
		0.015055,
		0.015059,
		0.015059,
		0.0098,
		0.005983,
		0.00336,
		0.001682,
		0.000714,
		0.000233,
		4.7e-005,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.015059,
		0.009797,
		0.005977,
		0.003351,
		0.001675,
		0.000709,
		0.000231,
		4.7e-005,
		1.3e-005,
		0.000317,
		0.000983,
		0.001962,
		0.003208,
		0.004672,
		0.006313,
		0.008087,
		0.009956,
		0.011881,
		0.013828,
		0.015763,
		0.017657,
		0.019481,
		0.021213,
		0.02283,
		0.024314,
		0.025651,
		0.02683,
		0.027847,
		0.028697,
		0.029385,
		0.029919,
		0.030311,
		0.030578,
		0.030743,
		0.030831,
		0.030867,
		0.030877,
		0.030877,
		0.022007,
		0.015055,
		0.009787,
		0.005962,
		0.003336,
		0.001662,
		0.000701,
		0.000228,
		4.6e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.030877,
		0.022001,
		0.015038,
		0.009761,
		0.005933,
		0.003311,
		0.001644,
		0.000691,
		0.000223,
		4.5e-005,
		0.000215,
		0.001123,
		0.002653,
		0.004715,
		0.007219,
		0.010082,
		0.013224,
		0.016567,
		0.02004,
		0.023575,
		0.027109,
		0.030582,
		0.033943,
		0.037142,
		0.040139,
		0.042898,
		0.045391,
		0.047595,
		0.049499,
		0.051096,
		0.052391,
		0.053397,
		0.054138,
		0.054644,
		0.054958,
		0.055125,
		0.055194,
		0.055212,
		0.055213,
		0.041881,
		0.030867,
		0.021976,
		0.014998,
		0.009713,
		0.005887,
		0.003273,
		0.001619,
		0.000677,
		0.000218,
		4.4e-005,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.055212,
		0.041868,
		0.030831,
		0.021915,
		0.014922,
		0.009635,
		0.005819,
		0.003222,
		0.001587,
		0.000661,
		0.000211,
		4.2e-005,
		4.1e-005,
		0.000982,
		0.003043,
		0.006065,
		0.009895,
		0.014386,
		0.019399,
		0.0248,
		0.030463,
		0.036269,
		0.042108,
		0.047876,
		0.05348,
		0.058836,
		0.06387,
		0.068519,
		0.072729,
		0.076464,
		0.079696,
		0.082414,
		0.084624,
		0.086344,
		0.087614,
		0.088484,
		0.089024,
		0.089312,
		0.089431,
		0.089462,
		0.089465,
		0.071029,
		0.055194,
		0.041817,
		0.030743,
		0.021801,
		0.014799,
		0.009521,
		0.005726,
		0.003156,
		0.001547,
		0.00064,
		0.000204,
		4e-005,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.089462,
		0.071006,
		0.055125,
		0.041695,
		0.030578,
		0.021616,
		0.014618,
		0.009364,
		0.005605,
		0.003074,
		0.001498,
		0.000617,
		0.000195,
		3.8e-005,
		0.000562,
		0.002929,
		0.006907,
		0.01225,
		0.018722,
		0.026096,
		0.034155,
		0.042695,
		0.051523,
		0.06046,
		0.069338,
		0.078003,
		0.086318,
		0.094161,
		0.101426,
		0.108027,
		0.113897,
		0.118991,
		0.123286,
		0.126786,
		0.129518,
		0.131537,
		0.132926,
		0.133789,
		0.13425,
		0.134442,
		0.134492,
		0.134496,
		0.1106,
		0.089431,
		0.070914,
		0.054958,
		0.041464,
		0.030311,
		0.021344,
		0.014371,
		0.009162,
		0.005456,
		0.002975,
		0.001442,
		0.00059,
		0.000185,
		3.6e-005,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.134492,
		0.110561,
		0.089312,
		0.070692,
		0.054644,
		0.041091,
		0.029919,
		0.020973,
		0.014052,
		0.00891,
		0.005276,
		0.00286,
		0.001377,
		0.00056,
		0.000175,
		3.4e-005,
		9.4e-005,
		0.002242,
		0.006937,
		0.013802,
		0.022475,
		0.032612,
		0.043882,
		0.055972,
		0.068588,
		0.081453,
		0.094309,
		0.106923,
		0.119079,
		0.130588,
		0.141286,
		0.151036,
		0.159732,
		0.167298,
		0.173695,
		0.17892,
		0.183009,
		0.186039,
		0.188127,
		0.189429,
		0.190126,
		0.190417,
		0.190492,
		0.190498,
		0.161138,
		0.134442,
		0.110408,
		0.089024,
		0.070277,
		0.054138,
		0.040544,
		0.029385,
		0.020495,
		0.013657,
		0.008609,
		0.005066,
		0.002729,
		0.001305,
		0.000526,
		0.000163,
		3.1e-005,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.190492,
		0.161077,
		0.13425,
		0.110041,
		0.088484,
		0.069605,
		0.053397,
		0.039801,
		0.028697,
		0.019903,
		0.013183,
		0.008259,
		0.004828,
		0.002582,
		0.001225,
		0.000491,
		0.000151,
		2.9e-005,
		0.001155,
		0.006001,
		0.014128,
		0.025011,
		0.038146,
		0.053053,
		0.069278,
		0.08639,
		0.103984,
		0.121685,
		0.139147,
		0.156055,
		0.172129,
		0.187125,
		0.200837,
		0.213104,
		0.223808,
		0.232882,
		0.240313,
		0.246144,
		0.250476,
		0.253469,
		0.255338,
		0.256343,
		0.256763,
		0.256872,
		0.25688,
		0.222455,
		0.190417,
		0.160839,
		0.133789,
		0.109353,
		0.087614,
		0.068626,
		0.052391,
		0.038844,
		0.027847,
		0.019196,
		0.012633,
		0.00786,
		0.004562,
		0.002421,
		0.00114,
		0.000452,
		0.000138,
		2.6e-005,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.256872,
		0.222364,
		0.190126,
		0.160268,
		0.132926,
		0.108244,
		0.086344,
		0.067296,
		0.051096,
		0.037661,
		0.02683,
		0.018373,
		0.012007,
		0.007416,
		0.00427,
		0.002247,
		0.001049,
		0.000412,
		0.000124,
		2.3e-005,
		0.000176,
		0.004205,
		0.012984,
		0.025783,
		0.041898,
		0.060661,
		0.081434,
		0.103613,
		0.126633,
		0.149966,
		0.173123,
		0.195661,
		0.217181,
		0.237336,
		0.25583,
		0.272428,
		0.286955,
		0.299306,
		0.309448,
		0.317427,
		0.323371,
		0.327489,
		0.330067,
		0.331456,
		0.332039,
		0.33219,
		0.332201,
		0.293537,
		0.256763,
		0.222012,
		0.189429,
		0.159201,
		0.131537,
		0.10663,
		0.084624,
		0.065587,
		0.049499,
		0.036251,
		0.025651,
		0.01744,
		0.011311,
		0.00693,
		0.003956,
		0.002063,
		0.000953,
		0.000371,
		0.00011,
		2e-005,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.33219,
		0.293409,
		0.256343,
		0.22117,
		0.188127,
		0.157487,
		0.129518,
		0.104444,
		0.082414,
		0.063481,
		0.047595,
		0.034615,
		0.024314,
		0.016402,
		0.01055,
		0.006407,
		0.003623,
		0.00187,
		0.000854,
		0.000328,
		9.6e-005,
		1.7e-005,
		0.002007,
		0.010409,
		0.024457,
		0.043203,
		0.065746,
		0.091224,
		0.118823,
		0.147776,
		0.177368,
		0.206935,
		0.235873,
		0.263636,
		0.289746,
		0.313795,
		0.33545,
		0.354464,
		0.370678,
		0.384031,
		0.394565,
		0.402434,
		0.4079,
		0.411333,
		0.413186,
		0.413966,
		0.414169,
		0.414184,
		0.37252,
		0.332039,
		0.292911,
		0.255338,
		0.219602,
		0.186039,
		0.154997,
		0.126786,
		0.101641,
		0.079696,
		0.060974,
		0.045391,
		0.032763,
		0.02283,
		0.015271,
		0.009733,
		0.005853,
		0.003275,
		0.001671,
		0.000754,
		0.000285,
		8.2e-005,
		1.5e-005,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.414169,
		0.372344,
		0.331456,
		0.291724,
		0.253469,
		0.217088,
		0.183009,
		0.151632,
		0.123286,
		0.098195,
		0.076464,
		0.058074,
		0.042898,
		0.030711,
		0.021213,
		0.014057,
		0.00887,
		0.005276,
		0.002916,
		0.001468,
		0.000652,
		0.000243,
		6.9e-005,
		1.2e-005,
		0.000286,
		0.006822,
		0.021023,
		0.041656,
		0.067541,
		0.097552,
		0.130624,
		0.16575,
		0.20199,
		0.238473,
		0.274398,
		0.309044,
		0.341773,
		0.372039,
		0.399391,
		0.423487,
		0.444099,
		0.461124,
		0.474596,
		0.484687,
		0.491718,
		0.496144,
		0.498541,
		0.499553,
		0.499817,
		0.499836,
		0.456743,
		0.413966,
		0.371666,
		0.330067,
		0.289518,
		0.250476,
		0.213445,
		0.17892,
		0.147328,
		0.118991,
		0.094104,
		0.072729,
		0.0548,
		0.040139,
		0.028478,
		0.019481,
		0.012776,
		0.00797,
		0.004682,
		0.002552,
		0.001265,
		0.000552,
		0.000201,
		5.6e-005,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.499817,
		0.456511,
		0.413186,
		0.370055,
		0.327489,
		0.285991,
		0.246144,
		0.208535,
		0.173695,
		0.142051,
		0.113897,
		0.089384,
		0.068519,
		0.051181,
		0.037142,
		0.02609,
		0.017657,
		0.011443,
		0.007047,
		0.00408,
		0.002188,
		0.001064,
		0.000454,
		0.000161,
		4.3e-005,
		0.003067,
		0.015878,
		0.037225,
		0.065609,
		0.0996,
		0.13784,
		0.179047,
		0.222021,
		0.265647,
		0.308897,
		0.350845,
		0.390664,
		0.427644,
		0.461192,
		0.49085,
		0.516304,
		0.537395,
		0.554134,
		0.566711,
		0.575499,
		0.581049,
		0.584062,
		0.585338,
		0.585671,
		0.585696,
		0.542944,
		0.499553,
		0.455619,
		0.411333,
		0.367068,
		0.323371,
		0.280893,
		0.240313,
		0.202269,
		0.167298,
		0.135801,
		0.108027,
		0.084068,
		0.06387,
		0.047255,
		0.033943,
		0.023578,
		0.015763,
		0.010079,
		0.006113,
		0.003479,
		0.00183,
		0.00087,
		0.000361,
		0.000124,
		3.2e-005,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.585671,
		0.542647,
		0.498541,
		0.453504,
		0.4079,
		0.362304,
		0.317427,
		0.274039,
		0.232882,
		0.194608,
		0.159732,
		0.128609,
		0.101426,
		0.078208,
		0.058836,
		0.043071,
		0.030582,
		0.020976,
		0.013828,
		0.008703,
		0.005185,
		0.00289,
		0.001483,
		0.000685,
		0.000274,
		9e-005,
		2.2e-005,
		0.000415,
		0.009861,
		0.030325,
		0.059949,
		0.096962,
		0.13968,
		0.186509,
		0.235957,
		0.286632,
		0.337251,
		0.386651,
		0.433793,
		0.477773,
		0.517835,
		0.553385,
		0.584001,
		0.609454,
		0.629719,
		0.644993,
		0.655698,
		0.662478,
		0.666171,
		0.66774,
		0.66815,
		0.668181,
		0.62755,
		0.585338,
		0.541506,
		0.496144,
		0.449592,
		0.402434,
		0.355436,
		0.309448,
		0.265316,
		0.223808,
		0.185558,
		0.151036,
		0.120532,
		0.094161,
		0.071873,
		0.05348,
		0.038685,
		0.027109,
		0.018324,
		0.011881,
		0.007338,
		0.004277,
		0.002323,
		0.001156,
		0.000513,
		0.000196,
		6e-005,
		1.3e-005,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.66815,
		0.627179,
		0.584062,
		0.53881,
		0.491718,
		0.443369,
		0.394565,
		0.346229,
		0.299306,
		0.254678,
		0.213104,
		0.175172,
		0.141286,
		0.111657,
		0.086318,
		0.065143,
		0.047876,
		0.034161,
		0.023575,
		0.015662,
		0.009956,
		0.006008,
		0.003407,
		0.00179,
		0.000854,
		0.000359,
		0.000127,
		3.5e-005,
		0.004224,
		0.021821,
		0.051039,
		0.089732,
		0.135857,
		0.187482,
		0.242791,
		0.300088,
		0.357808,
		0.414524,
		0.468959,
		0.519995,
		0.566689,
		0.608287,
		0.644244,
		0.674241,
		0.698204,
		0.716324,
		0.729065,
		0.73716,
		0.741583,
		0.743467,
		0.743962,
		0.743999,
		0.707073,
		0.66774,
		0.62576,
		0.581049,
		0.533837,
		0.484687,
		0.434424,
		0.384031,
		0.334543,
		0.286955,
		0.242148,
		0.200837,
		0.163544,
		0.130588,
		0.102092,
		0.078003,
		0.058116,
		0.042108,
		0.029569,
		0.02004,
		0.013038,
		0.008087,
		0.004739,
		0.002592,
		0.001301,
		0.000585,
		0.000227,
		7.2e-005,
		1.6e-005,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.743962,
		0.706621,
		0.666171,
		0.622414,
		0.575499,
		0.52595,
		0.474596,
		0.422467,
		0.370678,
		0.320332,
		0.272428,
		0.22781,
		0.187125,
		0.150806,
		0.119079,
		0.09197,
		0.069338,
		0.0509,
		0.036269,
		0.02499,
		0.016567,
		0.010502,
		0.006313,
		0.003559,
		0.001853,
		0.000871,
		0.000357,
		0.000121,
		3e-005,
		0.000547,
		0.012978,
		0.039813,
		0.078506,
		0.126633,
		0.181895,
		0.242128,
		0.30531,
		0.369568,
		0.433192,
		0.494641,
		0.552562,
		0.605806,
		0.653439,
		0.694774,
		0.729382,
		0.757127,
		0.778177,
		0.793028,
		0.802496,
		0.807686,
		0.809903,
		0.810487,
		0.810531,
		0.778514,
		0.743467,
		0.704898,
		0.662478,
		0.616261,
		0.566711,
		0.514646,
		0.461124,
		0.407332,
		0.354464,
		0.303641,
		0.25583,
		0.211808,
		0.172129,
		0.137128,
		0.106923,
		0.081444,
		0.06046,
		0.043615,
		0.030463,
		0.020506,
		0.013224,
		0.008105,
		0.004672,
		0.002495,
		0.001208,
		0.000512,
		0.000179,
		4.5e-005,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.810487,
		0.777975,
		0.741583,
		0.700845,
		0.655698,
		0.606532,
		0.554134,
		0.499578,
		0.444099,
		0.388981,
		0.33545,
		0.284601,
		0.237336,
		0.194338,
		0.156055,
		0.122709,
		0.094309,
		0.070687,
		0.051523,
		0.036394,
		0.0248,
		0.016209,
		0.010082,
		0.005905,
		0.003208,
		0.001579,
		0.000679,
		0.000237,
		5.7e-005,
		0.00535,
		0.027572,
		0.064324,
		0.112777,
		0.170248,
		0.234205,
		0.302277,
		0.372267,
		0.442157,
		0.510129,
		0.574574,
		0.634115,
		0.687623,
		0.734247,
		0.773435,
		0.804966,
		0.828975,
		0.845973,
		0.856847,
		0.862827,
		0.865392,
		0.866069,
		0.86612,
		0.839707,
		0.809903,
		0.775926,
		0.73716,
		0.693416,
		0.644993,
		0.59263,
		0.537395,
		0.480563,
		0.423487,
		0.367494,
		0.313795,
		0.26342,
		0.217181,
		0.175647,
		0.139147,
		0.107782,
		0.081453,
		0.05989,
		0.042695,
		0.029379,
		0.019399,
		0.012195,
		0.007219,
		0.00396,
		0.001962,
		0.000842,
		0.000287,
		6.2e-005,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.866069,
		0.839077,
		0.807686,
		0.771123,
		0.729065,
		0.681705,
		0.629719,
		0.574154,
		0.516304,
		0.457578,
		0.399391,
		0.343062,
		0.289746,
		0.240381,
		0.195661,
		0.156029,
		0.121685,
		0.092608,
		0.068588,
		0.049261,
		0.034155,
		0.022723,
		0.014386,
		0.008567,
		0.004715,
		0.002331,
		0.000983,
		0.000316,
		5.5e-005,
		0.000671,
		0.015865,
		0.048543,
		0.095456,
		0.153515,
		0.219804,
		0.291589,
		0.366327,
		0.441681,
		0.515532,
		0.586001,
		0.651464,
		0.71058,
		0.762317,
		0.80598,
		0.841247,
		0.868202,
		0.887355,
		0.899653,
		0.906441,
		0.909361,
		0.910135,
		0.910193,
		0.889535,
		0.865392,
		0.836688,
		0.802496,
		0.762347,
		0.716324,
		0.665023,
		0.609454,
		0.550912,
		0.49085,
		0.430753,
		0.372039,
		0.315979,
		0.263636,
		0.21583,
		0.173123,
		0.135819,
		0.103984,
		0.077474,
		0.055972,
		0.039027,
		0.026096,
		0.016587,
		0.009895,
		0.005431,
		0.002653,
		0.001083,
		0.000317,
		3.7e-005,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.910135,
		0.88881,
		0.862827,
		0.831105,
		0.793028,
		0.748558,
		0.698204,
		0.642924,
		0.584001,
		0.522909,
		0.461192,
		0.400354,
		0.341773,
		0.286631,
		0.235873,
		0.190179,
		0.149966,
		0.115392,
		0.08639,
		0.062693,
		0.043882,
		0.029425,
		0.018722,
		0.011143,
		0.006065,
		0.0029,
		0.001123,
		0.000283,
		1.3e-005,
		0.006363,
		0.032703,
		0.076079,
		0.132984,
		0.200101,
		0.274314,
		0.352718,
		0.43264,
		0.511645,
		0.587564,
		0.658512,
		0.722916,
		0.779544,
		0.827542,
		0.86647,
		0.896339,
		0.917645,
		0.931376,
		0.938984,
		0.942269,
		0.943143,
		0.943208,
		0.927957,
		0.909361,
		0.886069,
		0.856847,
		0.820937,
		0.778177,
		0.72898,
		0.674241,
		0.615215,
		0.553385,
		0.490336,
		0.427644,
		0.366781,
		0.309044,
		0.255494,
		0.206935,
		0.163895,
		0.126633,
		0.095162,
		0.069278,
		0.0486,
		0.032612,
		0.020712,
		0.01225,
		0.00657,
		0.003043,
		0.001091,
		0.000215,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.943143,
		0.927134,
		0.906441,
		0.879686,
		0.845973,
		0.805015,
		0.757127,
		0.703131,
		0.644244,
		0.581942,
		0.517835,
		0.453558,
		0.390664,
		0.330552,
		0.274398,
		0.223122,
		0.177368,
		0.137498,
		0.103613,
		0.075577,
		0.053053,
		0.035553,
		0.022475,
		0.013156,
		0.006907,
		0.003057,
		0.000982,
		0.000127,
		0.00078,
		0.018407,
		0.056161,
		0.110096,
		0.176476,
		0.251784,
		0.33274,
		0.416309,
		0.499726,
		0.580512,
		0.656502,
		0.725873,
		0.787175,
		0.839373,
		0.881889,
		0.914647,
		0.938108,
		0.953289,
		0.961733,
		0.965392,
		0.966369,
		0.966443,
		0.955875,
		0.942269,
		0.924032,
		0.899653,
		0.868102,
		0.828975,
		0.782486,
		0.729382,
		0.67083,
		0.608287,
		0.543376,
		0.477773,
		0.4131,
		0.350845,
		0.292291,
		0.238473,
		0.190145,
		0.147776,
		0.11156,
		0.081434,
		0.057115,
		0.038146,
		0.023933,
		0.013802,
		0.007038,
		0.002929,
		0.000799,
		4.1e-005,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.966369,
		0.95495,
		0.938984,
		0.916836,
		0.887355,
		0.850026,
		0.804966,
		0.752845,
		0.694774,
		0.632181,
		0.566689,
		0.5,
		0.433793,
		0.369632,
		0.308897,
		0.252728,
		0.20199,
		0.15726,
		0.118823,
		0.086698,
		0.060661,
		0.040291,
		0.025011,
		0.014139,
		0.006937,
		0.002653,
		0.000562,
		0.007261,
		0.037212,
		0.0863,
		0.150344,
		0.225407,
		0.307804,
		0.394122,
		0.48124,
		0.56635,
		0.646984,
		0.721044,
		0.786843,
		0.843145,
		0.889215,
		0.924865,
		0.950506,
		0.967166,
		0.976471,
		0.98052,
		0.981605,
		0.981688,
		0.974863,
		0.965392,
		0.951479,
		0.931376,
		0.903823,
		0.868202,
		0.824541,
		0.773435,
		0.715937,
		0.653439,
		0.587553,
		0.519995,
		0.452485,
		0.386651,
		0.323959,
		0.265647,
		0.212684,
		0.16575,
		0.12523,
		0.091224,
		0.063573,
		0.041898,
		0.025644,
		0.014128,
		0.006591,
		0.002242,
		0.000308,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.981605,
		0.973833,
		0.961733,
		0.943456,
		0.917645,
		0.883593,
		0.841247,
		0.791133,
		0.734247,
		0.67194,
		0.605806,
		0.537563,
		0.468959,
		0.401677,
		0.337251,
		0.27701,
		0.222021,
		0.173066,
		0.130624,
		0.09488,
		0.065746,
		0.04289,
		0.025783,
		0.013746,
		0.006001,
		0.001726,
		9.4e-005,
		0.00088,
		0.020707,
		0.06298,
		0.123044,
		0.196507,
		0.279254,
		0.367463,
		0.457628,
		0.546584,
		0.631533,
		0.710079,
		0.780271,
		0.840647,
		0.890289,
		0.928881,
		0.956762,
		0.974957,
		0.985163,
		0.989623,
		0.990823,
		0.990914,
		0.986844,
		0.98052,
		0.969981,
		0.953289,
		0.929002,
		0.896339,
		0.855177,
		0.80598,
		0.749691,
		0.687623,
		0.621342,
		0.552562,
		0.48305,
		0.414524,
		0.348584,
		0.286632,
		0.229828,
		0.179047,
		0.134865,
		0.097552,
		0.067091,
		0.043203,
		0.025391,
		0.012984,
		0.005191,
		0.001155,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.990823,
		0.985703,
		0.976471,
		0.961111,
		0.938108,
		0.90662,
		0.86647,
		0.818069,
		0.762317,
		0.700485,
		0.634115,
		0.564909,
		0.494641,
		0.425058,
		0.357808,
		0.294361,
		0.235957,
		0.183562,
		0.13784,
		0.099147,
		0.067541,
		0.042798,
		0.024457,
		0.011859,
		0.004205,
		0.000604,
		0.008113,
		0.041447,
		0.095789,
		0.166253,
		0.248253,
		0.337521,
		0.430131,
		0.522525,
		0.611541,
		0.694454,
		0.769016,
		0.833512,
		0.886816,
		0.928457,
		0.958682,
		0.978498,
		0.989664,
		0.994565,
		0.995889,
		0.99599,
		0.993768,
		0.989623,
		0.981452,
		0.967166,
		0.945195,
		0.914647,
		0.875305,
		0.827542,
		0.772221,
		0.71058,
		0.644135,
		0.574574,
		0.50367,
		0.433192,
		0.364821,
		0.300088,
		0.240301,
		0.186509,
		0.139462,
		0.0996,
		0.067052,
		0.041656,
		0.022989,
		0.010409,
		0.003111,
		0.000176,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.995889,
		0.992508,
		0.985163,
		0.971702,
		0.950506,
		0.92065,
		0.881889,
		0.834575,
		0.779544,
		0.71801,
		0.651464,
		0.581581,
		0.510129,
		0.438887,
		0.369568,
		0.303745,
		0.242791,
		0.187825,
		0.13968,
		0.098873,
		0.065609,
		0.039786,
		0.021023,
		0.008699,
		0.002007,
		0.000981,
		0.022996,
		0.069697,
		0.13565,
		0.215749,
		0.305234,
		0.39971,
		0.495182,
		0.588084,
		0.67532,
		0.75431,
		0.823051,
		0.880177,
		0.925035,
		0.957757,
		0.979315,
		0.99152,
		0.996902,
		0.998363,
		0.998475,
		0.997368,
		0.994565,
		0.987832,
		0.974957,
		0.954283,
		0.924865,
		0.886448,
		0.839373,
		0.784462,
		0.722916,
		0.656209,
		0.586001,
		0.514056,
		0.442157,
		0.372038,
		0.30531,
		0.243396,
		0.187482,
		0.138472,
		0.096962,
		0.063227,
		0.037225,
		0.018622,
		0.006822,
		0.001015,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.998363,
		0.995977,
		0.989664,
		0.977154,
		0.956762,
		0.927533,
		0.889215,
		0.842154,
		0.787175,
		0.725472,
		0.658512,
		0.587947,
		0.515532,
		0.443057,
		0.372267,
		0.304801,
		0.242128,
		0.18549,
		0.135857,
		0.093895,
		0.059949,
		0.03404,
		0.015878,
		0.004895,
		0.000286,
		0.009011,
		0.04587,
		0.105601,
		0.182516,
		0.2713,
		0.367036,
		0.465237,
		0.56188,
		0.653452,
		0.736995,
		0.810175,
		0.87135,
		0.919651,
		0.955071,
		0.978526,
		0.991872,
		0.997787,
		0.999399,
		0.999523,
		0.999005,
		0.996902,
		0.990841,
		0.978498,
		0.958161,
		0.928881,
		0.890421,
		0.843145,
		0.787895,
		0.725873,
		0.658546,
		0.587564,
		0.514679,
		0.441681,
		0.370327,
		0.302277,
		0.239036,
		0.181895,
		0.131883,
		0.089732,
		0.055852,
		0.030325,
		0.012909,
		0.003067,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999399,
		0.997471,
		0.99152,
		0.979167,
		0.958682,
		0.929125,
		0.890289,
		0.842572,
		0.786843,
		0.724327,
		0.656502,
		0.585024,
		0.511645,
		0.438158,
		0.366327,
		0.297832,
		0.234205,
		0.176778,
		0.126633,
		0.084562,
		0.051039,
		0.026203,
		0.009861,
		0.001509,
		0.001092,
		0.025507,
		0.077003,
		0.149231,
		0.236251,
		0.332553,
		0.433094,
		0.533332,
		0.629277,
		0.71754,
		0.795406,
		0.860913,
		0.912939,
		0.951302,
		0.976845,
		0.991457,
		0.997966,
		0.999749,
		0.999887,
		0.999627,
		0.997787,
		0.991829,
		0.979315,
		0.958497,
		0.928457,
		0.889025,
		0.840647,
		0.784233,
		0.721044,
		0.652581,
		0.580512,
		0.5066,
		0.43264,
		0.360406,
		0.291589,
		0.227748,
		0.170248,
		0.120219,
		0.078506,
		0.045641,
		0.021821,
		0.006903,
		0.000415,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999749,
		0.997932,
		0.991872,
		0.979067,
		0.957757,
		0.927046,
		0.886816,
		0.837568,
		0.780271,
		0.71623,
		0.646984,
		0.574223,
		0.499726,
		0.425297,
		0.352718,
		0.283695,
		0.219804,
		0.162442,
		0.112777,
		0.071706,
		0.039813,
		0.01735,
		0.004224,
		0.010032,
		0.050861,
		0.116585,
		0.200548,
		0.296568,
		0.398965,
		0.502605,
		0.60295,
		0.696118,
		0.778955,
		0.849122,
		0.905199,
		0.946796,
		0.97465,
		0.990675,
		0.997853,
		0.999829,
		0.999982,
		0.999806,
		0.997966,
		0.991729,
		0.978526,
		0.956583,
		0.925035,
		0.883821,
		0.833512,
		0.775143,
		0.710079,
		0.639906,
		0.56635,
		0.491211,
		0.416309,
		0.343437,
		0.274314,
		0.210532,
		0.153515,
		0.104466,
		0.064324,
		0.033727,
		0.012978,
		0.002032,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999829,
		0.99793,
		0.991457,
		0.977766,
		0.955071,
		0.92254,
		0.880177,
		0.828631,
		0.769016,
		0.702768,
		0.631533,
		0.557079,
		0.48124,
		0.405858,
		0.33274,
		0.263617,
		0.200101,
		0.143636,
		0.095456,
		0.05654,
		0.027572,
		0.008906,
		0.000547,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469721,
		0.574454,
		0.672707,
		0.760816,
		0.836006,
		0.896504,
		0.941665,
		0.97209,
		0.989698,
		0.997631,
		0.999827,
		0.999999,
		0.999832,
		0.997853,
		0.991097,
		0.976845,
		0.953294,
		0.919651,
		0.875992,
		0.823051,
		0.762033,
		0.694454,
		0.622029,
		0.546584,
		0.469991,
		0.394122,
		0.320804,
		0.251784,
		0.188689,
		0.132984,
		0.085931,
		0.048543,
		0.021549,
		0.00535,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999827,
		0.99775,
		0.990675,
		0.975798,
		0.951302,
		0.916436,
		0.87135,
		0.816874,
		0.75431,
		0.685268,
		0.611541,
		0.535021,
		0.457628,
		0.381271,
		0.307804,
		0.238991,
		0.176476,
		0.121743,
		0.076079,
		0.040531,
		0.015865,
		0.00253,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325197,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999818,
		0.997631,
		0.990206,
		0.97465,
		0.94913,
		0.912939,
		0.86631,
		0.810175,
		0.745942,
		0.67532,
		0.600193,
		0.522525,
		0.444298,
		0.367463,
		0.293903,
		0.225407,
		0.163637,
		0.110096,
		0.066092,
		0.032703,
		0.010731,
		0.000671,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999808,
		0.997501,
		0.989698,
		0.973413,
		0.946796,
		0.909189,
		0.860913,
		0.803008,
		0.736995,
		0.664693,
		0.588084,
		0.509214,
		0.430131,
		0.352837,
		0.279254,
		0.211195,
		0.150344,
		0.098223,
		0.056161,
		0.025261,
		0.006363,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999797,
		0.997361,
		0.989153,
		0.97209,
		0.944307,
		0.905199,
		0.855181,
		0.795406,
		0.727518,
		0.653452,
		0.575293,
		0.495182,
		0.415238,
		0.337521,
		0.263995,
		0.196507,
		0.136762,
		0.0863,
		0.046472,
		0.018407,
		0.002974,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999784,
		0.99721,
		0.988571,
		0.970681,
		0.941665,
		0.900973,
		0.849122,
		0.787387,
		0.71754,
		0.641638,
		0.56188,
		0.480503,
		0.39971,
		0.321622,
		0.248253,
		0.181483,
		0.123044,
		0.074497,
		0.037212,
		0.01234,
		0.00078,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999771,
		0.997049,
		0.987949,
		0.969182,
		0.938862,
		0.896504,
		0.842734,
		0.778955,
		0.707073,
		0.629277,
		0.547885,
		0.465237,
		0.383622,
		0.305234,
		0.232137,
		0.166253,
		0.109339,
		0.06298,
		0.028563,
		0.007261,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999757,
		0.996876,
		0.987285,
		0.967586,
		0.93589,
		0.891783,
		0.836006,
		0.770102,
		0.696118,
		0.616381,
		0.533332,
		0.449423,
		0.367036,
		0.288436,
		0.215749,
		0.150937,
		0.095789,
		0.05191,
		0.020707,
		0.003372,
		0.001559,
		0.035914,
		0.106789,
		0.203541,
		0.316383,
		0.436458,
		0.555928,
		0.668086,
		0.767497,
		0.850172,
		0.913781,
		0.957881,
		0.984125,
		0.996274,
		0.999725,
		1,
		0.999741,
		0.99669,
		0.986574,
		0.965887,
		0.932739,
		0.886795,
		0.828924,
		0.760816,
		0.684668,
		0.60295,
		0.518237,
		0.433094,
		0.349999,
		0.2713,
		0.199181,
		0.13565,
		0.082529,
		0.041447,
		0.013828,
		0.00088,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999725,
		0.99649,
		0.985814,
		0.964076,
		0.929393,
		0.881522,
		0.821467,
		0.751076,
		0.672707,
		0.588979,
		0.502605,
		0.416269,
		0.332553,
		0.253886,
		0.182516,
		0.120501,
		0.069697,
		0.031753,
		0.008113,
		0.014374,
		0.071756,
		0.16171,
		0.273008,
		0.395453,
		0.519969,
		0.638727,
		0.745299,
		0.834852,
		0.904386,
		0.953007,
		0.982179,
		0.995792,
		0.999687,
		1,
		0.999707,
		0.996274,
		0.984999,
		0.962145,
		0.925841,
		0.875946,
		0.813614,
		0.740862,
		0.660215,
		0.574454,
		0.486434,
		0.398965,
		0.314732,
		0.236251,
		0.165839,
		0.105601,
		0.057429,
		0.022996,
		0.003761,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999687,
		0.996042,
		0.984125,
		0.960083,
		0.922066,
		0.870045,
		0.80534,
		0.730146,
		0.647171,
		0.559361,
		0.469721,
		0.381192,
		0.296568,
		0.218452,
		0.149231,
		0.091062,
		0.04587,
		0.015352,
		0.000981,
		0.001783,
		0.04083,
		0.120607,
		0.228195,
		0.351814,
		0.480925,
		0.606372,
		0.720543,
		0.817592,
		0.893703,
		0.947416,
		0.979927,
		0.995229,
		0.999643,
		1,
		0.999666,
		0.995792,
		0.983187,
		0.957881,
		0.918052,
		0.863799,
		0.796619,
		0.718904,
		0.63355,
		0.543682,
		0.452462,
		0.362964,
		0.278094,
		0.200548,
		0.132781,
		0.077003,
		0.035173,
		0.009011,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999643,
		0.995521,
		0.982179,
		0.955526,
		0.913781,
		0.857183,
		0.787425,
		0.707107,
		0.619329,
		0.527402,
		0.434652,
		0.344293,
		0.259348,
		0.182602,
		0.116585,
		0.063554,
		0.025507,
		0.004181,
		0.016464,
		0.08164,
		0.182603,
		0.3057,
		0.438658,
		0.570735,
		0.692907,
		0.798105,
		0.881522,
		0.94098,
		0.977312,
		0.994569,
		0.999591,
		1,
		0.999618,
		0.995229,
		0.981094,
		0.953007,
		0.909232,
		0.850172,
		0.777728,
		0.694725,
		0.604481,
		0.510503,
		0.416289,
		0.325197,
		0.24037,
		0.164687,
		0.100747,
		0.050861,
		0.017057,
		0.001092,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999591,
		0.994912,
		0.979927,
		0.950308,
		0.904386,
		0.842738,
		0.767497,
		0.681729,
		0.588982,
		0.492972,
		0.397375,
		0.3057,
		0.221211,
		0.146884,
		0.085389,
		0.039085,
		0.010032,
		0.002059,
		0.046826,
		0.137255,
		0.25746,
		0.393097,
		0.531529,
		0.662033,
		0.776061,
		0.86759,
		0.933544,
		0.97426,
		0.993791,
		0.999529,
		1,
		0.999561,
		0.994569,
		0.978669,
		0.947416,
		0.899218,
		0.834852,
		0.756699,
		0.668086,
		0.572806,
		0.474793,
		0.377913,
		0.28583,
		0.201932,
		0.129291,
		0.070648,
		0.02841,
		0.004665,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999529,
		0.994196,
		0.977312,
		0.944313,
		0.893703,
		0.826481,
		0.745299,
		0.653763,
		0.555928,
		0.455958,
		0.357915,
		0.265625,
		0.182603,
		0.112019,
		0.056682,
		0.019046,
		0.001221,
		0.019046,
		0.093699,
		0.207724,
		0.344293,
		0.488488,
		0.62753,
		0.751077,
		0.851607,
		0.924919,
		0.970681,
		0.99287,
		0.999454,
		1,
		0.999493,
		0.993791,
		0.975845,
		0.94098,
		0.887815,
		0.817592,
		0.73326,
		0.638727,
		0.538323,
		0.436458,
		0.337397,
		0.245133,
		0.163312,
		0.0952,
		0.043674,
		0.011231,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999454,
		0.99335,
		0.97426,
		0.937398,
		0.881522,
		0.808146,
		0.720543,
		0.622942,
		0.519969,
		0.416289,
		0.316383,
		0.224414,
		0.144161,
		0.078987,
		0.031834,
		0.005236,
		0.002404,
		0.05424,
		0.157547,
		0.292492,
		0.441396,
		0.588982,
		0.722712,
		0.833218,
		0.914874,
		0.966466,
		0.991772,
		0.999365,
		1,
		0.999412,
		0.99287,
		0.972543,
		0.933544,
		0.874792,
		0.798105,
		0.707107,
		0.606372,
		0.500843,
		0.395453,
		0.294906,
		0.203541,
		0.125272,
		0.063557,
		0.021402,
		0.001374,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999365,
		0.992346,
		0.970681,
		0.929393,
		0.86759,
		0.787425,
		0.692907,
		0.588982,
		0.480925,
		0.373956,
		0.273008,
		0.182603,
		0.106789,
		0.049119,
		0.012657,
		0.022285,
		0.108615,
		0.238249,
		0.39014,
		0.545959,
		0.690463,
		0.811993,
		0.903125,
		0.961473,
		0.990456,
		0.999256,
		1,
		0.999313,
		0.991772,
		0.968661,
		0.924919,
		0.859876,
		0.776061,
		0.677898,
		0.570735,
		0.460201,
		0.351814,
		0.250747,
		0.16171,
		0.088884,
		0.035914,
		0.00592,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999256,
		0.991144,
		0.966466,
		0.92009,
		0.851607,
		0.763962,
		0.662033,
		0.551596,
		0.438658,
		0.32905,
		0.228195,
		0.140991,
		0.071756,
		0.024223,
		0.001559,
		0.002844,
		0.063557,
		0.182603,
		0.334797,
		0.498046,
		0.653763,
		0.787425,
		0.889324,
		0.955526,
		0.988867,
		0.999123,
		1,
		0.999193,
		0.990456,
		0.964076,
		0.914874,
		0.842738,
		0.751077,
		0.645261,
		0.531529,
		0.416289,
		0.3057,
		0.205443,
		0.120607,
		0.055645,
		0.014374,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999123,
		0.9897,
		0.961473,
		0.909232,
		0.833218,
		0.737347,
		0.62753,
		0.510503,
		0.393097,
		0.281814,
		0.182603,
		0.100747,
		0.04083,
		0.006746,
		0.026425,
		0.127354,
		0.275766,
		0.444897,
		0.611985,
		0.758906,
		0.873037,
		0.948403,
		0.986936,
		0.99896,
		1,
		0.999046,
		0.988867,
		0.958631,
		0.903125,
		0.82299,
		0.722712,
		0.608789,
		0.488488,
		0.369091,
		0.25746,
		0.159817,
		0.08164,
		0.027641,
		0.001783,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.99896,
		0.987949,
		0.955526,
		0.896506,
		0.811993,
		0.707107,
		0.588982,
		0.465459,
		0.344293,
		0.232725,
		0.137255,
		0.063557,
		0.016464,
		0.003417,
		0.075484,
		0.213992,
		0.38632,
		0.564456,
		0.725714,
		0.853729,
		0.939815,
		0.984569,
		0.998756,
		1,
		0.998863,
		0.986936,
		0.952128,
		0.889324,
		0.800163,
		0.690463,
		0.568056,
		0.441396,
		0.318742,
		0.207724,
		0.11513,
		0.046826,
		0.007758,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998756,
		0.985814,
		0.948403,
		0.881522,
		0.787425,
		0.672708,
		0.545959,
		0.416289,
		0.292492,
		0.182603,
		0.093699,
		0.031834,
		0.002059,
		0.031834,
		0.151319,
		0.322436,
		0.510503,
		0.687003,
		0.830729,
		0.929393,
		0.981647,
		0.998499,
		1,
		0.998635,
		0.984569,
		0.944313,
		0.873037,
		0.773702,
		0.653763,
		0.522638,
		0.39014,
		0.265625,
		0.157547,
		0.073275,
		0.019046,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998499,
		0.983187,
		0.939815,
		0.863799,
		0.758906,
		0.63355,
		0.498046,
		0.362964,
		0.238249,
		0.132788,
		0.05424,
		0.009016,
		0.004182,
		0.091084,
		0.253947,
		0.449532,
		0.641793,
		0.803203,
		0.916658,
		0.978003,
		0.998174,
		1,
		0.998347,
		0.981647,
		0.93486,
		0.853729,
		0.742944,
		0.611985,
		0.472143,
		0.334797,
		0.210511,
		0.108615,
		0.037057,
		0.002404,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.998174,
		0.979927,
		0.929393,
		0.842738,
		0.725714,
		0.588982,
		0.444897,
		0.3057,
		0.182603,
		0.085389,
		0.022285,
		0.039085,
		0.182603,
		0.381194,
		0.588982,
		0.770107,
		0.900978,
		0.973418,
		0.997755,
		1,
		0.997978,
		0.978003,
		0.92335,
		0.830729,
		0.707107,
		0.564456,
		0.416289,
		0.275766,
		0.154777,
		0.063557,
		0.010606,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997755,
		0.975845,
		0.916658,
		0.817592,
		0.687003,
		0.538323,
		0.38632,
		0.245133,
		0.127354,
		0.043674,
		0.002844,
		0.005236,
		0.112019,
		0.3057,
		0.527402,
		0.730146,
		0.881522,
		0.967586,
		0.99721,
		1,
		0.997501,
		0.973418,
		0.909232,
		0.803203,
		0.665277,
		0.510503,
		0.355015,
		0.213992,
		0.100747,
		0.026425,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.99721,
		0.970681,
		0.900978,
		0.787425,
		0.641793,
		0.480925,
		0.322436,
		0.182603,
		0.075484,
		0.012657,
		0.049119,
		0.224414,
		0.455958,
		0.681729,
		0.857183,
		0.960083,
		0.996489,
		1,
		0.996876,
		0.967586,
		0.891784,
		0.770107,
		0.61641,
		0.449532,
		0.28869,
		0.151319,
		0.05223,
		0.003417,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.996489,
		0.964076,
		0.881522,
		0.751077,
		0.588982,
		0.416289,
		0.253947,
		0.120607,
		0.031834,
		0.00674,
		0.140868,
		0.373628,
		0.622395,
		0.825756,
		0.949475,
		0.994648,
		0.999123,
		0.995994,
		0.960083,
		0.870045,
		0.730146,
		0.559361,
		0.381194,
		0.218463,
		0.091084,
		0.015366,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.994648,
		0.955481,
		0.857183,
		0.707107,
		0.527402,
		0.344293,
		0.182603,
		0.063557,
		0.004182,
		0.061965,
		0.274752,
		0.537774,
		0.767694,
		0.913908,
		0.969283,
		0.974941,
		0.988808,
		0.949475,
		0.842698,
		0.681729,
		0.492972,
		0.3057,
		0.146884,
		0.039085,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.969283,
		0.938519,
		0.825756,
		0.653732,
		0.455958,
		0.265625,
		0.112019,
		0.019046,
		0.007571,
		0.153342,
		0.390871,
		0.61919,
		0.77265,
		0.833327,
		0.839755,
		0.921848,
		0.913908,
		0.803188,
		0.622395,
		0.416269,
		0.224414,
		0.078987,
		0.005236,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.833327,
		0.862495,
		0.767694,
		0.585368,
		0.373628,
		0.182595,
		0.049119,
		0.046363,
		0.197077,
		0.365257,
		0.486772,
		0.537373,
		0.542966,
		0.700845,
		0.77265,
		0.708972,
		0.537774,
		0.327031,
		0.140868,
		0.024222,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.537373,
		0.642924,
		0.61919,
		0.473757,
		0.274752,
		0.100129,
		0.00674,
		0.002816,
		0.054533,
		0.131026,
		0.192162,
		0.219306,
		0.222461,
		0.368043,
		0.486772,
		0.5,
		0.390871,
		0.215973,
		0.061965,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.219306,
		0.328396,
		0.365257,
		0.294361,
		0.153342,
		0.029542,
		0.005051,
		0.020142,
		0.034243,
		0.041042,
		0.041882,
		0.108744,
		0.192162,
		0.236018,
		0.197077,
		0.093895,
		0.007571,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.041042,
		0.09321,
		0.131026,
		0.113883,
		0.046363,
		1.4e-005,
		0.000248,
		0.00054,
		0.000698,
		0.000719,
		0.009563,
		0.034243,
		0.05954,
		0.054533,
		0.01627,
		0.000698,
		0.007717,
		0.020142,
		0.020197,
		0.002816,
		0.00054,
		0.00408,
		0.005051,
		0.000248,
		0.000623,
		1.4e-005,
		1.4e-005,
		0.000623,
		0.000248,
		0.005051,
		0.00408,
		0.00054,
		0.002816,
		0.020197,
		0.020142,
		0.007717,
		0.000698,
		0.01627,
		0.054533,
		0.05954,
		0.034243,
		0.009563,
		0.000719,
		0.000698,
		0.00054,
		0.000248,
		1.4e-005,
		0.046363,
		0.113883,
		0.131026,
		0.09321,
		0.041042,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		0.007571,
		0.093895,
		0.197077,
		0.236018,
		0.192162,
		0.108744,
		0.041882,
		0.041042,
		0.034243,
		0.020142,
		0.005051,
		0.029542,
		0.153342,
		0.294361,
		0.365257,
		0.328396,
		0.219306,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.061965,
		0.215973,
		0.390871,
		0.5,
		0.486772,
		0.368043,
		0.222461,
		0.219306,
		0.192162,
		0.131026,
		0.054533,
		0.002816,
		0.00674,
		0.100129,
		0.274752,
		0.473757,
		0.61919,
		0.642924,
		0.537373,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.024222,
		0.140868,
		0.327031,
		0.537774,
		0.708972,
		0.77265,
		0.700845,
		0.542966,
		0.537373,
		0.486772,
		0.365257,
		0.197077,
		0.046363,
		0.049119,
		0.182595,
		0.373628,
		0.585368,
		0.767694,
		0.862495,
		0.833327,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.005236,
		0.078987,
		0.224414,
		0.416269,
		0.622395,
		0.803188,
		0.913908,
		0.921848,
		0.839755,
		0.833327,
		0.77265,
		0.61919,
		0.390871,
		0.153342,
		0.007571,
		0.019046,
		0.112019,
		0.265625,
		0.455958,
		0.653732,
		0.825756,
		0.938519,
		0.969283,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.039085,
		0.146884,
		0.3057,
		0.492972,
		0.681729,
		0.842698,
		0.949475,
		0.988808,
		0.974941,
		0.969283,
		0.913908,
		0.767694,
		0.537774,
		0.274752,
		0.061965,
		0.004182,
		0.063557,
		0.182603,
		0.344293,
		0.527402,
		0.707107,
		0.857183,
		0.955481,
		0.994648,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.015366,
		0.091084,
		0.218463,
		0.381194,
		0.559361,
		0.730146,
		0.870045,
		0.960083,
		0.995994,
		0.999123,
		0.994648,
		0.949475,
		0.825756,
		0.622395,
		0.373628,
		0.140868,
		0.00674,
		0.031834,
		0.120607,
		0.253947,
		0.416289,
		0.588982,
		0.751077,
		0.881522,
		0.964076,
		0.996489,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.003417,
		0.05223,
		0.151319,
		0.28869,
		0.449532,
		0.61641,
		0.770107,
		0.891784,
		0.967586,
		0.996876,
		1,
		0.996489,
		0.960083,
		0.857183,
		0.681729,
		0.455958,
		0.224414,
		0.049119,
		0.012657,
		0.075484,
		0.182603,
		0.322436,
		0.480925,
		0.641793,
		0.787425,
		0.900978,
		0.970681,
		0.99721,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.026425,
		0.100747,
		0.213992,
		0.355015,
		0.510503,
		0.665277,
		0.803203,
		0.909232,
		0.973418,
		0.997501,
		1,
		0.99721,
		0.967586,
		0.881522,
		0.730146,
		0.527402,
		0.3057,
		0.112019,
		0.005236,
		0.002844,
		0.043674,
		0.127354,
		0.245133,
		0.38632,
		0.538323,
		0.687003,
		0.817592,
		0.916658,
		0.975845,
		0.997755,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.010606,
		0.063557,
		0.154777,
		0.275766,
		0.416289,
		0.564456,
		0.707107,
		0.830729,
		0.92335,
		0.978003,
		0.997978,
		1,
		0.997755,
		0.973418,
		0.900978,
		0.770107,
		0.588982,
		0.381194,
		0.182603,
		0.039085,
		0.022285,
		0.085389,
		0.182603,
		0.3057,
		0.444897,
		0.588982,
		0.725714,
		0.842738,
		0.929393,
		0.979927,
		0.998174,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.002404,
		0.037057,
		0.108615,
		0.210511,
		0.334797,
		0.472143,
		0.611985,
		0.742944,
		0.853729,
		0.93486,
		0.981647,
		0.998347,
		1,
		0.998174,
		0.978003,
		0.916658,
		0.803203,
		0.641793,
		0.449532,
		0.253947,
		0.091084,
		0.004182,
		0.009016,
		0.05424,
		0.132788,
		0.238249,
		0.362964,
		0.498046,
		0.63355,
		0.758906,
		0.863799,
		0.939815,
		0.983187,
		0.998499,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.019046,
		0.073275,
		0.157547,
		0.265625,
		0.39014,
		0.522638,
		0.653763,
		0.773702,
		0.873037,
		0.944313,
		0.984569,
		0.998635,
		1,
		0.998499,
		0.981647,
		0.929393,
		0.830729,
		0.687003,
		0.510503,
		0.322436,
		0.151319,
		0.031834,
		0.002059,
		0.031834,
		0.093699,
		0.182603,
		0.292492,
		0.416289,
		0.545959,
		0.672708,
		0.787425,
		0.881522,
		0.948403,
		0.985814,
		0.998756,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.007758,
		0.046826,
		0.11513,
		0.207724,
		0.318742,
		0.441396,
		0.568056,
		0.690463,
		0.800163,
		0.889324,
		0.952128,
		0.986936,
		0.998863,
		1,
		0.998756,
		0.984569,
		0.939815,
		0.853729,
		0.725714,
		0.564456,
		0.38632,
		0.213992,
		0.075484,
		0.003417,
		0.016464,
		0.063557,
		0.137255,
		0.232725,
		0.344293,
		0.465459,
		0.588982,
		0.707107,
		0.811993,
		0.896506,
		0.955526,
		0.987949,
		0.99896,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.001783,
		0.027641,
		0.08164,
		0.159817,
		0.25746,
		0.369091,
		0.488488,
		0.608789,
		0.722712,
		0.82299,
		0.903125,
		0.958631,
		0.988867,
		0.999046,
		1,
		0.99896,
		0.986936,
		0.948403,
		0.873037,
		0.758906,
		0.611985,
		0.444897,
		0.275766,
		0.127354,
		0.026425,
		0.006746,
		0.04083,
		0.100747,
		0.182603,
		0.281814,
		0.393097,
		0.510503,
		0.62753,
		0.737347,
		0.833218,
		0.909232,
		0.961473,
		0.9897,
		0.999123,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.014374,
		0.055645,
		0.120607,
		0.205443,
		0.3057,
		0.416289,
		0.531529,
		0.645261,
		0.751077,
		0.842738,
		0.914874,
		0.964076,
		0.990456,
		0.999193,
		1,
		0.999123,
		0.988867,
		0.955526,
		0.889324,
		0.787425,
		0.653763,
		0.498046,
		0.334797,
		0.182603,
		0.063557,
		0.002844,
		0.001559,
		0.024223,
		0.071756,
		0.140991,
		0.228195,
		0.32905,
		0.438658,
		0.551596,
		0.662033,
		0.763962,
		0.851607,
		0.92009,
		0.966466,
		0.991144,
		0.999256,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.00592,
		0.035914,
		0.088884,
		0.16171,
		0.250747,
		0.351814,
		0.460201,
		0.570735,
		0.677898,
		0.776061,
		0.859876,
		0.924919,
		0.968661,
		0.991772,
		0.999313,
		1,
		0.999256,
		0.990456,
		0.961473,
		0.903125,
		0.811993,
		0.690463,
		0.545959,
		0.39014,
		0.238249,
		0.108615,
		0.022285,
		0.012657,
		0.049119,
		0.106789,
		0.182603,
		0.273008,
		0.373956,
		0.480925,
		0.588982,
		0.692907,
		0.787425,
		0.86759,
		0.929393,
		0.970681,
		0.992346,
		0.999365,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.001374,
		0.021402,
		0.063557,
		0.125272,
		0.203541,
		0.294906,
		0.395453,
		0.500843,
		0.606372,
		0.707107,
		0.798105,
		0.874792,
		0.933544,
		0.972543,
		0.99287,
		0.999412,
		1,
		0.999365,
		0.991772,
		0.966466,
		0.914874,
		0.833218,
		0.722712,
		0.588982,
		0.441396,
		0.292492,
		0.157547,
		0.05424,
		0.002404,
		0.005236,
		0.031834,
		0.078987,
		0.144161,
		0.224414,
		0.316383,
		0.416289,
		0.519969,
		0.622942,
		0.720543,
		0.808146,
		0.881522,
		0.937398,
		0.97426,
		0.99335,
		0.999454,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.011231,
		0.043674,
		0.0952,
		0.163312,
		0.245133,
		0.337397,
		0.436458,
		0.538323,
		0.638727,
		0.73326,
		0.817592,
		0.887815,
		0.94098,
		0.975845,
		0.993791,
		0.999493,
		1,
		0.999454,
		0.99287,
		0.970681,
		0.924919,
		0.851607,
		0.751077,
		0.62753,
		0.488488,
		0.344293,
		0.207724,
		0.093699,
		0.019046,
		0.001221,
		0.019046,
		0.056682,
		0.112019,
		0.182603,
		0.265625,
		0.357915,
		0.455958,
		0.555928,
		0.653763,
		0.745299,
		0.826481,
		0.893703,
		0.944313,
		0.977312,
		0.994196,
		0.999529,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.004665,
		0.02841,
		0.070648,
		0.129291,
		0.201932,
		0.28583,
		0.377913,
		0.474793,
		0.572806,
		0.668086,
		0.756699,
		0.834852,
		0.899218,
		0.947416,
		0.978669,
		0.994569,
		0.999561,
		1,
		0.999529,
		0.993791,
		0.97426,
		0.933544,
		0.86759,
		0.776061,
		0.662033,
		0.531529,
		0.393097,
		0.25746,
		0.137255,
		0.046826,
		0.002059,
		0.010032,
		0.039085,
		0.085389,
		0.146884,
		0.221211,
		0.3057,
		0.397375,
		0.492972,
		0.588982,
		0.681729,
		0.767497,
		0.842738,
		0.904386,
		0.950308,
		0.979927,
		0.994912,
		0.999591,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.001092,
		0.017057,
		0.050861,
		0.100747,
		0.164687,
		0.24037,
		0.325197,
		0.416289,
		0.510503,
		0.604481,
		0.694725,
		0.777728,
		0.850172,
		0.909232,
		0.953007,
		0.981094,
		0.995229,
		0.999618,
		1,
		0.999591,
		0.994569,
		0.977312,
		0.94098,
		0.881522,
		0.798105,
		0.692907,
		0.570735,
		0.438658,
		0.3057,
		0.182603,
		0.08164,
		0.016464,
		0.004181,
		0.025507,
		0.063554,
		0.116585,
		0.182602,
		0.259348,
		0.344293,
		0.434652,
		0.527402,
		0.619329,
		0.707107,
		0.787425,
		0.857183,
		0.913781,
		0.955526,
		0.982179,
		0.995521,
		0.999643,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.009011,
		0.035173,
		0.077003,
		0.132781,
		0.200548,
		0.278094,
		0.362964,
		0.452462,
		0.543682,
		0.63355,
		0.718904,
		0.796619,
		0.863799,
		0.918052,
		0.957881,
		0.983187,
		0.995792,
		0.999666,
		1,
		0.999643,
		0.995229,
		0.979927,
		0.947416,
		0.893703,
		0.817592,
		0.720543,
		0.606372,
		0.480925,
		0.351814,
		0.228195,
		0.120607,
		0.04083,
		0.001783,
		0.000981,
		0.015352,
		0.04587,
		0.091062,
		0.149231,
		0.218452,
		0.296568,
		0.381192,
		0.469721,
		0.559361,
		0.647171,
		0.730146,
		0.80534,
		0.870045,
		0.922066,
		0.960083,
		0.984125,
		0.996042,
		0.999687,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.003761,
		0.022996,
		0.057429,
		0.105601,
		0.165839,
		0.236251,
		0.314732,
		0.398965,
		0.486434,
		0.574454,
		0.660215,
		0.740862,
		0.813614,
		0.875946,
		0.925841,
		0.962145,
		0.984999,
		0.996274,
		0.999707,
		1,
		0.999687,
		0.995792,
		0.982179,
		0.953007,
		0.904386,
		0.834852,
		0.745299,
		0.638727,
		0.519969,
		0.395453,
		0.273008,
		0.16171,
		0.071756,
		0.014374,
		0.008113,
		0.031753,
		0.069697,
		0.120501,
		0.182516,
		0.253886,
		0.332553,
		0.416269,
		0.502605,
		0.588979,
		0.672707,
		0.751076,
		0.821467,
		0.881522,
		0.929393,
		0.964076,
		0.985814,
		0.99649,
		0.999725,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.00088,
		0.013828,
		0.041447,
		0.082529,
		0.13565,
		0.199181,
		0.2713,
		0.349999,
		0.433094,
		0.518237,
		0.60295,
		0.684668,
		0.760816,
		0.828924,
		0.886795,
		0.932739,
		0.965887,
		0.986574,
		0.99669,
		0.999741,
		1,
		0.999725,
		0.996274,
		0.984125,
		0.957881,
		0.913781,
		0.850172,
		0.767497,
		0.668086,
		0.555928,
		0.436458,
		0.316383,
		0.203541,
		0.106789,
		0.035914,
		0.001559,
		0.003372,
		0.020707,
		0.05191,
		0.095789,
		0.150937,
		0.215749,
		0.288436,
		0.367036,
		0.449423,
		0.533332,
		0.616381,
		0.696118,
		0.770102,
		0.836006,
		0.891783,
		0.93589,
		0.967586,
		0.987285,
		0.996876,
		0.999757,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.007261,
		0.028563,
		0.06298,
		0.109339,
		0.166253,
		0.232137,
		0.305234,
		0.383622,
		0.465237,
		0.547885,
		0.629277,
		0.707073,
		0.778955,
		0.842734,
		0.896504,
		0.938862,
		0.969182,
		0.987949,
		0.997049,
		0.999771,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.00078,
		0.01234,
		0.037212,
		0.074497,
		0.123044,
		0.181483,
		0.248253,
		0.321622,
		0.39971,
		0.480503,
		0.56188,
		0.641638,
		0.71754,
		0.787387,
		0.849122,
		0.900973,
		0.941665,
		0.970681,
		0.988571,
		0.99721,
		0.999784,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.002974,
		0.018407,
		0.046472,
		0.0863,
		0.136762,
		0.196507,
		0.263995,
		0.337521,
		0.415238,
		0.495182,
		0.575293,
		0.653452,
		0.727518,
		0.795406,
		0.855181,
		0.905199,
		0.944307,
		0.97209,
		0.989153,
		0.997361,
		0.999797,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.006363,
		0.025261,
		0.056161,
		0.098223,
		0.150344,
		0.211195,
		0.279254,
		0.352837,
		0.430131,
		0.509214,
		0.588084,
		0.664693,
		0.736995,
		0.803008,
		0.860913,
		0.909189,
		0.946796,
		0.973413,
		0.989698,
		0.997501,
		0.999808,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.000671,
		0.010731,
		0.032703,
		0.066092,
		0.110096,
		0.163637,
		0.225407,
		0.293903,
		0.367463,
		0.444298,
		0.522525,
		0.600193,
		0.67532,
		0.745942,
		0.810175,
		0.86631,
		0.912939,
		0.94913,
		0.97465,
		0.990206,
		0.997631,
		0.999818,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325197,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.00253,
		0.015865,
		0.040531,
		0.076079,
		0.121743,
		0.176476,
		0.238991,
		0.307804,
		0.381271,
		0.457628,
		0.535021,
		0.611541,
		0.685268,
		0.75431,
		0.816874,
		0.87135,
		0.916436,
		0.951302,
		0.975798,
		0.990675,
		0.99775,
		0.999827,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.00535,
		0.021549,
		0.048543,
		0.085931,
		0.132984,
		0.188689,
		0.251784,
		0.320804,
		0.394122,
		0.469991,
		0.546584,
		0.622029,
		0.694454,
		0.762033,
		0.823051,
		0.875992,
		0.919651,
		0.953294,
		0.976845,
		0.991097,
		0.997853,
		0.999832,
		0.999999,
		0.999827,
		0.997631,
		0.989698,
		0.97209,
		0.941665,
		0.896504,
		0.836006,
		0.760816,
		0.672707,
		0.574454,
		0.469721,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.000547,
		0.008906,
		0.027572,
		0.05654,
		0.095456,
		0.143636,
		0.200101,
		0.263617,
		0.33274,
		0.405858,
		0.48124,
		0.557079,
		0.631533,
		0.702768,
		0.769016,
		0.828631,
		0.880177,
		0.92254,
		0.955071,
		0.977766,
		0.991457,
		0.99793,
		0.999829,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.002032,
		0.012978,
		0.033727,
		0.064324,
		0.104466,
		0.153515,
		0.210532,
		0.274314,
		0.343437,
		0.416309,
		0.491211,
		0.56635,
		0.639906,
		0.710079,
		0.775143,
		0.833512,
		0.883821,
		0.925035,
		0.956583,
		0.978526,
		0.991729,
		0.997966,
		0.999806,
		0.999982,
		0.999829,
		0.997853,
		0.990675,
		0.97465,
		0.946796,
		0.905199,
		0.849122,
		0.778955,
		0.696118,
		0.60295,
		0.502605,
		0.398965,
		0.296568,
		0.200548,
		0.116585,
		0.050861,
		0.010032,
		0.004224,
		0.01735,
		0.039813,
		0.071706,
		0.112777,
		0.162442,
		0.219804,
		0.283695,
		0.352718,
		0.425297,
		0.499726,
		0.574223,
		0.646984,
		0.71623,
		0.780271,
		0.837568,
		0.886816,
		0.927046,
		0.957757,
		0.979067,
		0.991872,
		0.997932,
		0.999749,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.000415,
		0.006903,
		0.021821,
		0.045641,
		0.078506,
		0.120219,
		0.170248,
		0.227748,
		0.291589,
		0.360406,
		0.43264,
		0.5066,
		0.580512,
		0.652581,
		0.721044,
		0.784233,
		0.840647,
		0.889025,
		0.928457,
		0.958497,
		0.979315,
		0.991829,
		0.997787,
		0.999627,
		0.999887,
		0.999749,
		0.997966,
		0.991457,
		0.976845,
		0.951302,
		0.912939,
		0.860913,
		0.795406,
		0.71754,
		0.629277,
		0.533332,
		0.433094,
		0.332553,
		0.236251,
		0.149231,
		0.077003,
		0.025507,
		0.001092,
		0.001509,
		0.009861,
		0.026203,
		0.051039,
		0.084562,
		0.126633,
		0.176778,
		0.234205,
		0.297832,
		0.366327,
		0.438158,
		0.511645,
		0.585024,
		0.656502,
		0.724327,
		0.786843,
		0.842572,
		0.890289,
		0.929125,
		0.958682,
		0.979167,
		0.99152,
		0.997471,
		0.999399,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.003067,
		0.012909,
		0.030325,
		0.055852,
		0.089732,
		0.131883,
		0.181895,
		0.239036,
		0.302277,
		0.370327,
		0.441681,
		0.514679,
		0.587564,
		0.658546,
		0.725873,
		0.787895,
		0.843145,
		0.890421,
		0.928881,
		0.958161,
		0.978498,
		0.990841,
		0.996902,
		0.999005,
		0.999523,
		0.999399,
		0.997787,
		0.991872,
		0.978526,
		0.955071,
		0.919651,
		0.87135,
		0.810175,
		0.736995,
		0.653452,
		0.56188,
		0.465237,
		0.367036,
		0.2713,
		0.182516,
		0.105601,
		0.04587,
		0.009011,
		0.000286,
		0.004895,
		0.015878,
		0.03404,
		0.059949,
		0.093895,
		0.135857,
		0.18549,
		0.242128,
		0.304801,
		0.372267,
		0.443057,
		0.515532,
		0.587947,
		0.658512,
		0.725472,
		0.787175,
		0.842154,
		0.889215,
		0.927533,
		0.956762,
		0.977154,
		0.989664,
		0.995977,
		0.998363,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.001015,
		0.006822,
		0.018622,
		0.037225,
		0.063227,
		0.096962,
		0.138472,
		0.187482,
		0.243396,
		0.30531,
		0.372038,
		0.442157,
		0.514056,
		0.586001,
		0.656209,
		0.722916,
		0.784462,
		0.839373,
		0.886448,
		0.924865,
		0.954283,
		0.974957,
		0.987832,
		0.994565,
		0.997368,
		0.998475,
		0.998363,
		0.996902,
		0.99152,
		0.979315,
		0.957757,
		0.925035,
		0.880177,
		0.823051,
		0.75431,
		0.67532,
		0.588084,
		0.495182,
		0.39971,
		0.305234,
		0.215749,
		0.13565,
		0.069697,
		0.022996,
		0.000981,
		0.002007,
		0.008699,
		0.021023,
		0.039786,
		0.065609,
		0.098873,
		0.13968,
		0.187825,
		0.242791,
		0.303745,
		0.369568,
		0.438887,
		0.510129,
		0.581581,
		0.651464,
		0.71801,
		0.779544,
		0.834575,
		0.881889,
		0.92065,
		0.950506,
		0.971702,
		0.985163,
		0.992508,
		0.995889,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.000176,
		0.003111,
		0.010409,
		0.022989,
		0.041656,
		0.067052,
		0.0996,
		0.139462,
		0.186509,
		0.240301,
		0.300088,
		0.364821,
		0.433192,
		0.50367,
		0.574574,
		0.644135,
		0.71058,
		0.772221,
		0.827542,
		0.875305,
		0.914647,
		0.945195,
		0.967166,
		0.981452,
		0.989623,
		0.993768,
		0.99599,
		0.995889,
		0.994565,
		0.989664,
		0.978498,
		0.958682,
		0.928457,
		0.886816,
		0.833512,
		0.769016,
		0.694454,
		0.611541,
		0.522525,
		0.430131,
		0.337521,
		0.248253,
		0.166253,
		0.095789,
		0.041447,
		0.008113,
		0.000604,
		0.004205,
		0.011859,
		0.024457,
		0.042798,
		0.067541,
		0.099147,
		0.13784,
		0.183562,
		0.235957,
		0.294361,
		0.357808,
		0.425058,
		0.494641,
		0.564909,
		0.634115,
		0.700485,
		0.762317,
		0.818069,
		0.86647,
		0.90662,
		0.938108,
		0.961111,
		0.976471,
		0.985703,
		0.990823,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.001155,
		0.005191,
		0.012984,
		0.025391,
		0.043203,
		0.067091,
		0.097552,
		0.134865,
		0.179047,
		0.229828,
		0.286632,
		0.348584,
		0.414524,
		0.48305,
		0.552562,
		0.621342,
		0.687623,
		0.749691,
		0.80598,
		0.855177,
		0.896339,
		0.929002,
		0.953289,
		0.969981,
		0.98052,
		0.986844,
		0.990914,
		0.990823,
		0.989623,
		0.985163,
		0.974957,
		0.956762,
		0.928881,
		0.890289,
		0.840647,
		0.780271,
		0.710079,
		0.631533,
		0.546584,
		0.457628,
		0.367463,
		0.279254,
		0.196507,
		0.123044,
		0.06298,
		0.020707,
		0.00088,
		9.4e-005,
		0.001726,
		0.006001,
		0.013746,
		0.025783,
		0.04289,
		0.065746,
		0.09488,
		0.130624,
		0.173066,
		0.222021,
		0.27701,
		0.337251,
		0.401677,
		0.468959,
		0.537563,
		0.605806,
		0.67194,
		0.734247,
		0.791133,
		0.841247,
		0.883593,
		0.917645,
		0.943456,
		0.961733,
		0.973833,
		0.981605,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.000308,
		0.002242,
		0.006591,
		0.014128,
		0.025644,
		0.041898,
		0.063573,
		0.091224,
		0.12523,
		0.16575,
		0.212684,
		0.265647,
		0.323959,
		0.386651,
		0.452485,
		0.519995,
		0.587553,
		0.653439,
		0.715937,
		0.773435,
		0.824541,
		0.868202,
		0.903823,
		0.931376,
		0.951479,
		0.965392,
		0.974863,
		0.981688,
		0.981605,
		0.98052,
		0.976471,
		0.967166,
		0.950506,
		0.924865,
		0.889215,
		0.843145,
		0.786843,
		0.721044,
		0.646984,
		0.56635,
		0.48124,
		0.394122,
		0.307804,
		0.225407,
		0.150344,
		0.0863,
		0.037212,
		0.007261,
		0.000562,
		0.002653,
		0.006937,
		0.014139,
		0.025011,
		0.040291,
		0.060661,
		0.086698,
		0.118823,
		0.15726,
		0.20199,
		0.252728,
		0.308897,
		0.369632,
		0.433793,
		0.5,
		0.566689,
		0.632181,
		0.694774,
		0.752845,
		0.804966,
		0.850026,
		0.887355,
		0.916836,
		0.938984,
		0.95495,
		0.966369,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		4.1e-005,
		0.000799,
		0.002929,
		0.007038,
		0.013802,
		0.023933,
		0.038146,
		0.057115,
		0.081434,
		0.11156,
		0.147776,
		0.190145,
		0.238473,
		0.292291,
		0.350845,
		0.4131,
		0.477773,
		0.543376,
		0.608287,
		0.67083,
		0.729382,
		0.782486,
		0.828975,
		0.868102,
		0.899653,
		0.924032,
		0.942269,
		0.955875,
		0.966443,
		0.966369,
		0.965392,
		0.961733,
		0.953289,
		0.938108,
		0.914647,
		0.881889,
		0.839373,
		0.787175,
		0.725873,
		0.656502,
		0.580512,
		0.499726,
		0.416309,
		0.33274,
		0.251784,
		0.176476,
		0.110096,
		0.056161,
		0.018407,
		0.00078,
		0.000127,
		0.000982,
		0.003057,
		0.006907,
		0.013156,
		0.022475,
		0.035553,
		0.053053,
		0.075577,
		0.103613,
		0.137498,
		0.177368,
		0.223122,
		0.274398,
		0.330552,
		0.390664,
		0.453558,
		0.517835,
		0.581942,
		0.644244,
		0.703131,
		0.757127,
		0.805015,
		0.845973,
		0.879686,
		0.906441,
		0.927134,
		0.943143,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		0.000215,
		0.001091,
		0.003043,
		0.00657,
		0.01225,
		0.020712,
		0.032612,
		0.0486,
		0.069278,
		0.095162,
		0.126633,
		0.163895,
		0.206935,
		0.255494,
		0.309044,
		0.366781,
		0.427644,
		0.490336,
		0.553385,
		0.615215,
		0.674241,
		0.72898,
		0.778177,
		0.820937,
		0.856847,
		0.886069,
		0.909361,
		0.927957,
		0.943208,
		0.943143,
		0.942269,
		0.938984,
		0.931376,
		0.917645,
		0.896339,
		0.86647,
		0.827542,
		0.779544,
		0.722916,
		0.658512,
		0.587564,
		0.511645,
		0.43264,
		0.352718,
		0.274314,
		0.200101,
		0.132984,
		0.076079,
		0.032703,
		0.006363,
		1.3e-005,
		0.000283,
		0.001123,
		0.0029,
		0.006065,
		0.011143,
		0.018722,
		0.029425,
		0.043882,
		0.062693,
		0.08639,
		0.115392,
		0.149966,
		0.190179,
		0.235873,
		0.286631,
		0.341773,
		0.400354,
		0.461192,
		0.522909,
		0.584001,
		0.642924,
		0.698204,
		0.748558,
		0.793028,
		0.831105,
		0.862827,
		0.88881,
		0.910135,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		3.7e-005,
		0.000317,
		0.001083,
		0.002653,
		0.005431,
		0.009895,
		0.016587,
		0.026096,
		0.039027,
		0.055972,
		0.077474,
		0.103984,
		0.135819,
		0.173123,
		0.21583,
		0.263636,
		0.315979,
		0.372039,
		0.430753,
		0.49085,
		0.550912,
		0.609454,
		0.665023,
		0.716324,
		0.762347,
		0.802496,
		0.836688,
		0.865392,
		0.889535,
		0.910193,
		0.910135,
		0.909361,
		0.906441,
		0.899653,
		0.887355,
		0.868202,
		0.841247,
		0.80598,
		0.762317,
		0.71058,
		0.651464,
		0.586001,
		0.515532,
		0.441681,
		0.366327,
		0.291589,
		0.219804,
		0.153515,
		0.095456,
		0.048543,
		0.015865,
		0.000671,
		5.5e-005,
		0.000316,
		0.000983,
		0.002331,
		0.004715,
		0.008567,
		0.014386,
		0.022723,
		0.034155,
		0.049261,
		0.068588,
		0.092608,
		0.121685,
		0.156029,
		0.195661,
		0.240381,
		0.289746,
		0.343062,
		0.399391,
		0.457578,
		0.516304,
		0.574154,
		0.629719,
		0.681705,
		0.729065,
		0.771123,
		0.807686,
		0.839077,
		0.866069,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		6.2e-005,
		0.000287,
		0.000842,
		0.001962,
		0.00396,
		0.007219,
		0.012195,
		0.019399,
		0.029379,
		0.042695,
		0.05989,
		0.081453,
		0.107782,
		0.139147,
		0.175647,
		0.217181,
		0.26342,
		0.313795,
		0.367494,
		0.423487,
		0.480563,
		0.537395,
		0.59263,
		0.644993,
		0.693416,
		0.73716,
		0.775926,
		0.809903,
		0.839707,
		0.86612,
		0.866069,
		0.865392,
		0.862827,
		0.856847,
		0.845973,
		0.828975,
		0.804966,
		0.773435,
		0.734247,
		0.687623,
		0.634115,
		0.574574,
		0.510129,
		0.442157,
		0.372267,
		0.302277,
		0.234205,
		0.170248,
		0.112777,
		0.064324,
		0.027572,
		0.00535,
		5.7e-005,
		0.000237,
		0.000679,
		0.001579,
		0.003208,
		0.005905,
		0.010082,
		0.016209,
		0.0248,
		0.036394,
		0.051523,
		0.070687,
		0.094309,
		0.122709,
		0.156055,
		0.194338,
		0.237336,
		0.284601,
		0.33545,
		0.388981,
		0.444099,
		0.499578,
		0.554134,
		0.606532,
		0.655698,
		0.700845,
		0.741583,
		0.777975,
		0.810487,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		4.5e-005,
		0.000179,
		0.000512,
		0.001208,
		0.002495,
		0.004672,
		0.008105,
		0.013224,
		0.020506,
		0.030463,
		0.043615,
		0.06046,
		0.081444,
		0.106923,
		0.137128,
		0.172129,
		0.211808,
		0.25583,
		0.303641,
		0.354464,
		0.407332,
		0.461124,
		0.514646,
		0.566711,
		0.616261,
		0.662478,
		0.704898,
		0.743467,
		0.778514,
		0.810531,
		0.810487,
		0.809903,
		0.807686,
		0.802496,
		0.793028,
		0.778177,
		0.757127,
		0.729382,
		0.694774,
		0.653439,
		0.605806,
		0.552562,
		0.494641,
		0.433192,
		0.369568,
		0.30531,
		0.242128,
		0.181895,
		0.126633,
		0.078506,
		0.039813,
		0.012978,
		0.000547,
		3e-005,
		0.000121,
		0.000357,
		0.000871,
		0.001853,
		0.003559,
		0.006313,
		0.010502,
		0.016567,
		0.02499,
		0.036269,
		0.0509,
		0.069338,
		0.09197,
		0.119079,
		0.150806,
		0.187125,
		0.22781,
		0.272428,
		0.320332,
		0.370678,
		0.422467,
		0.474596,
		0.52595,
		0.575499,
		0.622414,
		0.666171,
		0.706621,
		0.743962,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		1.6e-005,
		7.2e-005,
		0.000227,
		0.000585,
		0.001301,
		0.002592,
		0.004739,
		0.008087,
		0.013038,
		0.02004,
		0.029569,
		0.042108,
		0.058116,
		0.078003,
		0.102092,
		0.130588,
		0.163544,
		0.200837,
		0.242148,
		0.286955,
		0.334543,
		0.384031,
		0.434424,
		0.484687,
		0.533837,
		0.581049,
		0.62576,
		0.66774,
		0.707073,
		0.743999,
		0.743962,
		0.743467,
		0.741583,
		0.73716,
		0.729065,
		0.716324,
		0.698204,
		0.674241,
		0.644244,
		0.608287,
		0.566689,
		0.519995,
		0.468959,
		0.414524,
		0.357808,
		0.300088,
		0.242791,
		0.187482,
		0.135857,
		0.089732,
		0.051039,
		0.021821,
		0.004224,
		3.5e-005,
		0.000127,
		0.000359,
		0.000854,
		0.00179,
		0.003407,
		0.006008,
		0.009956,
		0.015662,
		0.023575,
		0.034161,
		0.047876,
		0.065143,
		0.086318,
		0.111657,
		0.141286,
		0.175172,
		0.213104,
		0.254678,
		0.299306,
		0.346229,
		0.394565,
		0.443369,
		0.491718,
		0.53881,
		0.584062,
		0.627179,
		0.66815,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		1.3e-005,
		6e-005,
		0.000196,
		0.000513,
		0.001156,
		0.002323,
		0.004277,
		0.007338,
		0.011881,
		0.018324,
		0.027109,
		0.038685,
		0.05348,
		0.071873,
		0.094161,
		0.120532,
		0.151036,
		0.185558,
		0.223808,
		0.265316,
		0.309448,
		0.355436,
		0.402434,
		0.449592,
		0.496144,
		0.541506,
		0.585338,
		0.62755,
		0.668181,
		0.66815,
		0.66774,
		0.666171,
		0.662478,
		0.655698,
		0.644993,
		0.629719,
		0.609454,
		0.584001,
		0.553385,
		0.517835,
		0.477773,
		0.433793,
		0.386651,
		0.337251,
		0.286632,
		0.235957,
		0.186509,
		0.13968,
		0.096962,
		0.059949,
		0.030325,
		0.009861,
		0.000415,
		2.2e-005,
		9e-005,
		0.000274,
		0.000685,
		0.001483,
		0.00289,
		0.005185,
		0.008703,
		0.013828,
		0.020976,
		0.030582,
		0.043071,
		0.058836,
		0.078208,
		0.101426,
		0.128609,
		0.159732,
		0.194608,
		0.232882,
		0.274039,
		0.317427,
		0.362304,
		0.4079,
		0.453504,
		0.498541,
		0.542647,
		0.585671,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		3.2e-005,
		0.000124,
		0.000361,
		0.00087,
		0.00183,
		0.003479,
		0.006113,
		0.010079,
		0.015763,
		0.023578,
		0.033943,
		0.047255,
		0.06387,
		0.084068,
		0.108027,
		0.135801,
		0.167298,
		0.202269,
		0.240313,
		0.280893,
		0.323371,
		0.367068,
		0.411333,
		0.455619,
		0.499553,
		0.542944,
		0.585696,
		0.585671,
		0.585338,
		0.584062,
		0.581049,
		0.575499,
		0.566711,
		0.554134,
		0.537395,
		0.516304,
		0.49085,
		0.461192,
		0.427644,
		0.390664,
		0.350845,
		0.308897,
		0.265647,
		0.222021,
		0.179047,
		0.13784,
		0.0996,
		0.065609,
		0.037225,
		0.015878,
		0.003067,
		4.3e-005,
		0.000161,
		0.000454,
		0.001064,
		0.002188,
		0.00408,
		0.007047,
		0.011443,
		0.017657,
		0.02609,
		0.037142,
		0.051181,
		0.068519,
		0.089384,
		0.113897,
		0.142051,
		0.173695,
		0.208535,
		0.246144,
		0.285991,
		0.327489,
		0.370055,
		0.413186,
		0.456511,
		0.499817,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		5.6e-005,
		0.000201,
		0.000552,
		0.001265,
		0.002552,
		0.004682,
		0.00797,
		0.012776,
		0.019481,
		0.028478,
		0.040139,
		0.0548,
		0.072729,
		0.094104,
		0.118991,
		0.147328,
		0.17892,
		0.213445,
		0.250476,
		0.289518,
		0.330067,
		0.371666,
		0.413966,
		0.456743,
		0.499836,
		0.499817,
		0.499553,
		0.498541,
		0.496144,
		0.491718,
		0.484687,
		0.474596,
		0.461124,
		0.444099,
		0.423487,
		0.399391,
		0.372039,
		0.341773,
		0.309044,
		0.274398,
		0.238473,
		0.20199,
		0.16575,
		0.130624,
		0.097552,
		0.067541,
		0.041656,
		0.021023,
		0.006822,
		0.000286,
		1.2e-005,
		6.9e-005,
		0.000243,
		0.000652,
		0.001468,
		0.002916,
		0.005276,
		0.00887,
		0.014057,
		0.021213,
		0.030711,
		0.042898,
		0.058074,
		0.076464,
		0.098195,
		0.123286,
		0.151632,
		0.183009,
		0.217088,
		0.253469,
		0.291724,
		0.331456,
		0.372344,
		0.414169,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		1.5e-005,
		8.2e-005,
		0.000285,
		0.000754,
		0.001671,
		0.003275,
		0.005853,
		0.009733,
		0.015271,
		0.02283,
		0.032763,
		0.045391,
		0.060974,
		0.079696,
		0.101641,
		0.126786,
		0.154997,
		0.186039,
		0.219602,
		0.255338,
		0.292911,
		0.332039,
		0.37252,
		0.414184,
		0.414169,
		0.413966,
		0.413186,
		0.411333,
		0.4079,
		0.402434,
		0.394565,
		0.384031,
		0.370678,
		0.354464,
		0.33545,
		0.313795,
		0.289746,
		0.263636,
		0.235873,
		0.206935,
		0.177368,
		0.147776,
		0.118823,
		0.091224,
		0.065746,
		0.043203,
		0.024457,
		0.010409,
		0.002007,
		1.7e-005,
		9.6e-005,
		0.000328,
		0.000854,
		0.00187,
		0.003623,
		0.006407,
		0.01055,
		0.016402,
		0.024314,
		0.034615,
		0.047595,
		0.063481,
		0.082414,
		0.104444,
		0.129518,
		0.157487,
		0.188127,
		0.22117,
		0.256343,
		0.293409,
		0.33219,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		2e-005,
		0.00011,
		0.000371,
		0.000953,
		0.002063,
		0.003956,
		0.00693,
		0.011311,
		0.01744,
		0.025651,
		0.036251,
		0.049499,
		0.065587,
		0.084624,
		0.10663,
		0.131537,
		0.159201,
		0.189429,
		0.222012,
		0.256763,
		0.293537,
		0.332201,
		0.33219,
		0.332039,
		0.331456,
		0.330067,
		0.327489,
		0.323371,
		0.317427,
		0.309448,
		0.299306,
		0.286955,
		0.272428,
		0.25583,
		0.237336,
		0.217181,
		0.195661,
		0.173123,
		0.149966,
		0.126633,
		0.103613,
		0.081434,
		0.060661,
		0.041898,
		0.025783,
		0.012984,
		0.004205,
		0.000176,
		2.3e-005,
		0.000124,
		0.000412,
		0.001049,
		0.002247,
		0.00427,
		0.007416,
		0.012007,
		0.018373,
		0.02683,
		0.037661,
		0.051096,
		0.067296,
		0.086344,
		0.108244,
		0.132926,
		0.160268,
		0.190126,
		0.222364,
		0.256872,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		2.6e-005,
		0.000138,
		0.000452,
		0.00114,
		0.002421,
		0.004562,
		0.00786,
		0.012633,
		0.019196,
		0.027847,
		0.038844,
		0.052391,
		0.068626,
		0.087614,
		0.109353,
		0.133789,
		0.160839,
		0.190417,
		0.222455,
		0.25688,
		0.256872,
		0.256763,
		0.256343,
		0.255338,
		0.253469,
		0.250476,
		0.246144,
		0.240313,
		0.232882,
		0.223808,
		0.213104,
		0.200837,
		0.187125,
		0.172129,
		0.156055,
		0.139147,
		0.121685,
		0.103984,
		0.08639,
		0.069278,
		0.053053,
		0.038146,
		0.025011,
		0.014128,
		0.006001,
		0.001155,
		2.9e-005,
		0.000151,
		0.000491,
		0.001225,
		0.002582,
		0.004828,
		0.008259,
		0.013183,
		0.019903,
		0.028697,
		0.039801,
		0.053397,
		0.069605,
		0.088484,
		0.110041,
		0.13425,
		0.161077,
		0.190492,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		3.1e-005,
		0.000163,
		0.000526,
		0.001305,
		0.002729,
		0.005066,
		0.008609,
		0.013657,
		0.020495,
		0.029385,
		0.040544,
		0.054138,
		0.070277,
		0.089024,
		0.110408,
		0.134442,
		0.161138,
		0.190498,
		0.190492,
		0.190417,
		0.190126,
		0.189429,
		0.188127,
		0.186039,
		0.183009,
		0.17892,
		0.173695,
		0.167298,
		0.159732,
		0.151036,
		0.141286,
		0.130588,
		0.119079,
		0.106923,
		0.094309,
		0.081453,
		0.068588,
		0.055972,
		0.043882,
		0.032612,
		0.022475,
		0.013802,
		0.006937,
		0.002242,
		9.4e-005,
		3.4e-005,
		0.000175,
		0.00056,
		0.001377,
		0.00286,
		0.005276,
		0.00891,
		0.014052,
		0.020973,
		0.029919,
		0.041091,
		0.054644,
		0.070692,
		0.089312,
		0.110561,
		0.134492,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		3.6e-005,
		0.000185,
		0.00059,
		0.001442,
		0.002975,
		0.005456,
		0.009162,
		0.014371,
		0.021344,
		0.030311,
		0.041464,
		0.054958,
		0.070914,
		0.089431,
		0.1106,
		0.134496,
		0.134492,
		0.134442,
		0.13425,
		0.133789,
		0.132926,
		0.131537,
		0.129518,
		0.126786,
		0.123286,
		0.118991,
		0.113897,
		0.108027,
		0.101426,
		0.094161,
		0.086318,
		0.078003,
		0.069338,
		0.06046,
		0.051523,
		0.042695,
		0.034155,
		0.026096,
		0.018722,
		0.01225,
		0.006907,
		0.002929,
		0.000562,
		3.8e-005,
		0.000195,
		0.000617,
		0.001498,
		0.003074,
		0.005605,
		0.009364,
		0.014618,
		0.021616,
		0.030578,
		0.041695,
		0.055125,
		0.071006,
		0.089462,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		4e-005,
		0.000204,
		0.00064,
		0.001547,
		0.003156,
		0.005726,
		0.009521,
		0.014799,
		0.021801,
		0.030743,
		0.041817,
		0.055194,
		0.071029,
		0.089465,
		0.089462,
		0.089431,
		0.089312,
		0.089024,
		0.088484,
		0.087614,
		0.086344,
		0.084624,
		0.082414,
		0.079696,
		0.076464,
		0.072729,
		0.068519,
		0.06387,
		0.058836,
		0.05348,
		0.047876,
		0.042108,
		0.036269,
		0.030463,
		0.0248,
		0.019399,
		0.014386,
		0.009895,
		0.006065,
		0.003043,
		0.000982,
		4.1e-005,
		4.2e-005,
		0.000211,
		0.000661,
		0.001587,
		0.003222,
		0.005819,
		0.009635,
		0.014922,
		0.021915,
		0.030831,
		0.041868,
		0.055212,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		4.4e-005,
		0.000218,
		0.000677,
		0.001619,
		0.003273,
		0.005887,
		0.009713,
		0.014998,
		0.021976,
		0.030867,
		0.041881,
		0.055213,
		0.055212,
		0.055194,
		0.055125,
		0.054958,
		0.054644,
		0.054138,
		0.053397,
		0.052391,
		0.051096,
		0.049499,
		0.047595,
		0.045391,
		0.042898,
		0.040139,
		0.037142,
		0.033943,
		0.030582,
		0.027109,
		0.023575,
		0.02004,
		0.016567,
		0.013224,
		0.010082,
		0.007219,
		0.004715,
		0.002653,
		0.001123,
		0.000215,
		4.5e-005,
		0.000223,
		0.000691,
		0.001644,
		0.003311,
		0.005933,
		0.009761,
		0.015038,
		0.022001,
		0.030877,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		4.6e-005,
		0.000228,
		0.000701,
		0.001662,
		0.003336,
		0.005962,
		0.009787,
		0.015055,
		0.022007,
		0.030877,
		0.030877,
		0.030867,
		0.030831,
		0.030743,
		0.030578,
		0.030311,
		0.029919,
		0.029385,
		0.028697,
		0.027847,
		0.02683,
		0.025651,
		0.024314,
		0.02283,
		0.021213,
		0.019481,
		0.017657,
		0.015763,
		0.013828,
		0.011881,
		0.009956,
		0.008087,
		0.006313,
		0.004672,
		0.003208,
		0.001962,
		0.000983,
		0.000317,
		1.3e-005,
		4.7e-005,
		0.000231,
		0.000709,
		0.001675,
		0.003351,
		0.005977,
		0.009797,
		0.015059,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		4.7e-005,
		0.000233,
		0.000714,
		0.001682,
		0.00336,
		0.005983,
		0.0098,
		0.015059,
		0.015059,
		0.015055,
		0.015038,
		0.014998,
		0.014922,
		0.014799,
		0.014618,
		0.014371,
		0.014052,
		0.013657,
		0.013183,
		0.012633,
		0.012007,
		0.011311,
		0.01055,
		0.009733,
		0.00887,
		0.00797,
		0.007047,
		0.006113,
		0.005185,
		0.004277,
		0.003407,
		0.002592,
		0.001853,
		0.001208,
		0.000679,
		0.000287,
		5.5e-005,
		4.8e-005,
		0.000235,
		0.000717,
		0.001686,
		0.003363,
		0.005985,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		4.8e-005,
		0.000236,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.005985,
		0.005983,
		0.005977,
		0.005962,
		0.005933,
		0.005887,
		0.005819,
		0.005726,
		0.005605,
		0.005456,
		0.005276,
		0.005066,
		0.004828,
		0.004562,
		0.00427,
		0.003956,
		0.003623,
		0.003275,
		0.002916,
		0.002552,
		0.002188,
		0.00183,
		0.001483,
		0.001156,
		0.000854,
		0.000585,
		0.000357,
		0.000179,
		5.7e-005,
		4.8e-005,
		0.000236,
		0.000719,
		0.001688,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001682,
		0.001675,
		0.001662,
		0.001644,
		0.001619,
		0.001587,
		0.001547,
		0.001498,
		0.001442,
		0.001377,
		0.001305,
		0.001225,
		0.00114,
		0.001049,
		0.000953,
		0.000854,
		0.000754,
		0.000652,
		0.000552,
		0.000454,
		0.000361,
		0.000274,
		0.000196,
		0.000127,
		7.2e-005,
		3e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		4.9e-005,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.000231,
		0.000228,
		0.000223,
		0.000218,
		0.000211,
		0.000204,
		0.000195,
		0.000185,
		0.000175,
		0.000163,
		0.000151,
		0.000138,
		0.000124,
		0.00011,
		9.6e-005,
		8.2e-005,
		6.9e-005,
		5.6e-005,
		4.3e-005,
		3.2e-005,
		2.2e-005,
		1.3e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		1.5e-005,
		2.4e-005,
		3.5e-005,
		4.6e-005,
		5.9e-005,
		7.2e-005,
		8.6e-005,
		0.0001,
		0.000114,
		0.000127,
		0.000141,
		0.000154,
		0.000166,
		0.000177,
		0.000188,
		0.000197,
		0.000206,
		0.000213,
		0.000219,
		0.000224,
		0.000228,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.00023,
		0.000227,
		0.000222,
		0.000216,
		0.00021,
		0.000202,
		0.000193,
		0.000183,
		0.000172,
		0.00016,
		0.000147,
		0.000134,
		0.000121,
		0.000107,
		9.3e-005,
		7.9e-005,
		6.5e-005,
		5.2e-005,
		4e-005,
		2.9e-005,
		2e-005,
		1.1e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		2.2e-005,
		6e-005,
		0.000112,
		0.000178,
		0.000254,
		0.000339,
		0.000431,
		0.000527,
		0.000627,
		0.000728,
		0.000829,
		0.000929,
		0.001025,
		0.001117,
		0.001205,
		0.001285,
		0.00136,
		0.001426,
		0.001485,
		0.001535,
		0.001578,
		0.001612,
		0.001639,
		0.001658,
		0.001672,
		0.001681,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001684,
		0.001677,
		0.001666,
		0.001649,
		0.001626,
		0.001596,
		0.001557,
		0.001511,
		0.001456,
		0.001394,
		0.001323,
		0.001246,
		0.001162,
		0.001072,
		0.000977,
		0.000879,
		0.000779,
		0.000678,
		0.000577,
		0.000479,
		0.000384,
		0.000295,
		0.000215,
		0.000143,
		8.4e-005,
		3.9e-005,
		1e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		8.2e-005,
		0.000218,
		0.00041,
		0.000649,
		0.000926,
		0.001235,
		0.001568,
		0.001918,
		0.002279,
		0.002644,
		0.003007,
		0.003363,
		0.003708,
		0.004037,
		0.004345,
		0.004631,
		0.00489,
		0.005121,
		0.005324,
		0.005496,
		0.005638,
		0.005752,
		0.005838,
		0.005901,
		0.005942,
		0.005967,
		0.005979,
		0.005984,
		0.005985,
		0.005985,
		0.005985,
		0.005982,
		0.005974,
		0.005956,
		0.005924,
		0.005872,
		0.005798,
		0.005698,
		0.005571,
		0.005413,
		0.005226,
		0.005009,
		0.004764,
		0.004491,
		0.004194,
		0.003874,
		0.003537,
		0.003186,
		0.002826,
		0.002461,
		0.002098,
		0.001742,
		0.001399,
		0.001077,
		0.000783,
		0.000524,
		0.000308,
		0.000143,
		3.7e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		2.5e-005,
		0.000213,
		0.000567,
		0.001064,
		0.001682,
		0.0024,
		0.003197,
		0.004055,
		0.004955,
		0.00588,
		0.006814,
		0.007741,
		0.008648,
		0.009521,
		0.010351,
		0.011126,
		0.011839,
		0.012483,
		0.013053,
		0.013546,
		0.01396,
		0.014298,
		0.014563,
		0.014759,
		0.014896,
		0.014983,
		0.015031,
		0.015052,
		0.015059,
		0.015059,
		0.015059,
		0.015057,
		0.015044,
		0.015011,
		0.014945,
		0.014835,
		0.014669,
		0.014439,
		0.014139,
		0.013763,
		0.013309,
		0.012777,
		0.01217,
		0.011491,
		0.010746,
		0.009942,
		0.009089,
		0.008198,
		0.007279,
		0.006347,
		0.005416,
		0.004501,
		0.00362,
		0.00279,
		0.00203,
		0.001359,
		0.000799,
		0.000371,
		9.7e-005,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		5.2e-005,
		0.000451,
		0.0012,
		0.002251,
		0.003555,
		0.005068,
		0.006746,
		0.008547,
		0.010433,
		0.012367,
		0.014314,
		0.016241,
		0.01812,
		0.019924,
		0.021629,
		0.023214,
		0.024662,
		0.025961,
		0.0271,
		0.028075,
		0.028884,
		0.029533,
		0.03003,
		0.030388,
		0.030628,
		0.030771,
		0.030844,
		0.030871,
		0.030877,
		0.030877,
		0.030876,
		0.030862,
		0.030815,
		0.03071,
		0.030522,
		0.030225,
		0.029799,
		0.029228,
		0.0285,
		0.027608,
		0.026551,
		0.025331,
		0.023956,
		0.022437,
		0.02079,
		0.019033,
		0.017189,
		0.015282,
		0.013341,
		0.011397,
		0.009482,
		0.007633,
		0.005889,
		0.004288,
		0.002874,
		0.001691,
		0.000785,
		0.000205,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		9.7e-005,
		0.000834,
		0.002217,
		0.004154,
		0.006556,
		0.009337,
		0.012416,
		0.015716,
		0.019164,
		0.022689,
		0.026229,
		0.029723,
		0.033116,
		0.03636,
		0.039411,
		0.042233,
		0.044794,
		0.047072,
		0.049052,
		0.050725,
		0.052095,
		0.053172,
		0.053976,
		0.054538,
		0.054895,
		0.055094,
		0.055183,
		0.05521,
		0.055213,
		0.055213,
		0.055202,
		0.055149,
		0.055011,
		0.054739,
		0.054285,
		0.053606,
		0.052669,
		0.051448,
		0.049927,
		0.0481,
		0.04597,
		0.043547,
		0.040853,
		0.037912,
		0.034759,
		0.031435,
		0.027985,
		0.024461,
		0.020921,
		0.017426,
		0.014044,
		0.010844,
		0.007904,
		0.005302,
		0.003122,
		0.001451,
		0.000379,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		0.000162,
		0.001398,
		0.003714,
		0.006952,
		0.010961,
		0.015596,
		0.020718,
		0.026196,
		0.031906,
		0.03773,
		0.04356,
		0.049296,
		0.054846,
		0.060128,
		0.065071,
		0.069614,
		0.073709,
		0.077319,
		0.080424,
		0.083014,
		0.085099,
		0.086702,
		0.087866,
		0.088647,
		0.089116,
		0.089354,
		0.089445,
		0.089464,
		0.089465,
		0.089459,
		0.089413,
		0.089259,
		0.088916,
		0.0883,
		0.087336,
		0.085958,
		0.084118,
		0.081783,
		0.078936,
		0.075576,
		0.07172,
		0.067396,
		0.062646,
		0.057525,
		0.052099,
		0.046446,
		0.040651,
		0.034811,
		0.02903,
		0.023421,
		0.018105,
		0.013209,
		0.008869,
		0.005227,
		0.002431,
		0.000635,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000253,
		0.002176,
		0.005775,
		0.0108,
		0.017011,
		0.02418,
		0.032087,
		0.040526,
		0.049299,
		0.058225,
		0.067132,
		0.075864,
		0.084279,
		0.092251,
		0.099669,
		0.106443,
		0.112501,
		0.117792,
		0.122288,
		0.125984,
		0.128904,
		0.131095,
		0.132632,
		0.133616,
		0.134165,
		0.134412,
		0.134487,
		0.134495,
		0.134494,
		0.134464,
		0.134318,
		0.133936,
		0.133186,
		0.13194,
		0.130086,
		0.127539,
		0.124235,
		0.12014,
		0.115245,
		0.109565,
		0.103142,
		0.096035,
		0.088328,
		0.08012,
		0.07153,
		0.062691,
		0.053754,
		0.044882,
		0.036252,
		0.028054,
		0.02049,
		0.013772,
		0.008124,
		0.003782,
		0.000989,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.000371,
		0.003192,
		0.008464,
		0.015814,
		0.024886,
		0.035336,
		0.046839,
		0.059088,
		0.071791,
		0.084677,
		0.097494,
		0.110012,
		0.122023,
		0.133344,
		0.143817,
		0.153312,
		0.161731,
		0.169008,
		0.17511,
		0.180046,
		0.183862,
		0.186644,
		0.18852,
		0.189652,
		0.190229,
		0.19045,
		0.190496,
		0.190498,
		0.190485,
		0.190372,
		0.189997,
		0.189167,
		0.187686,
		0.185375,
		0.18209,
		0.177722,
		0.172206,
		0.165515,
		0.157661,
		0.148694,
		0.138695,
		0.12778,
		0.116093,
		0.103804,
		0.091109,
		0.078227,
		0.0654,
		0.052889,
		0.040976,
		0.029959,
		0.020157,
		0.011902,
		0.005545,
		0.001451,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000519,
		0.004461,
		0.011818,
		0.02206,
		0.034677,
		0.049186,
		0.065122,
		0.082051,
		0.099561,
		0.117269,
		0.134822,
		0.151896,
		0.168203,
		0.183489,
		0.197539,
		0.21018,
		0.221283,
		0.230768,
		0.238608,
		0.244832,
		0.249525,
		0.252835,
		0.254963,
		0.256157,
		0.256697,
		0.256861,
		0.25688,
		0.256878,
		0.25681,
		0.256491,
		0.25566,
		0.254032,
		0.251342,
		0.247362,
		0.241918,
		0.234893,
		0.22623,
		0.21593,
		0.204045,
		0.19068,
		0.175988,
		0.160162,
		0.143438,
		0.126086,
		0.108413,
		0.090758,
		0.073489,
		0.057003,
		0.041725,
		0.028102,
		0.016611,
		0.007747,
		0.002029,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.000697,
		0.005983,
		0.015835,
		0.029527,
		0.046366,
		0.06569,
		0.086869,
		0.109311,
		0.132457,
		0.155789,
		0.178832,
		0.201151,
		0.22236,
		0.242125,
		0.260166,
		0.276259,
		0.29025,
		0.302048,
		0.311642,
		0.319097,
		0.324562,
		0.328265,
		0.330511,
		0.331662,
		0.332104,
		0.332198,
		0.332201,
		0.332175,
		0.331948,
		0.331199,
		0.329549,
		0.326616,
		0.322066,
		0.31563,
		0.307118,
		0.296425,
		0.283525,
		0.268467,
		0.251376,
		0.232442,
		0.211916,
		0.190107,
		0.167375,
		0.144131,
		0.120829,
		0.097968,
		0.076086,
		0.055758,
		0.037596,
		0.022245,
		0.010384,
		0.002723,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.000902,
		0.007742,
		0.020468,
		0.038126,
		0.059802,
		0.084624,
		0.111767,
		0.140451,
		0.169947,
		0.19958,
		0.228728,
		0.256833,
		0.283397,
		0.307995,
		0.330276,
		0.349969,
		0.366892,
		0.38096,
		0.392189,
		0.400704,
		0.406741,
		0.410642,
		0.412843,
		0.413844,
		0.414148,
		0.414184,
		0.414179,
		0.414053,
		0.413461,
		0.411925,
		0.408933,
		0.404013,
		0.396774,
		0.386925,
		0.374285,
		0.358785,
		0.340459,
		0.319444,
		0.295966,
		0.270336,
		0.242943,
		0.214251,
		0.184787,
		0.155141,
		0.125963,
		0.097956,
		0.071873,
		0.048517,
		0.028739,
		0.01343,
		0.003525,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.001132,
		0.009702,
		0.025625,
		0.04768,
		0.074699,
		0.105574,
		0.139251,
		0.174741,
		0.211121,
		0.247536,
		0.283205,
		0.317429,
		0.34959,
		0.379165,
		0.405731,
		0.428973,
		0.448693,
		0.464821,
		0.477426,
		0.486716,
		0.493048,
		0.496909,
		0.498898,
		0.499666,
		0.49983,
		0.499836,
		0.49979,
		0.499394,
		0.498098,
		0.495252,
		0.490224,
		0.482466,
		0.471554,
		0.457205,
		0.43928,
		0.417783,
		0.392847,
		0.36473,
		0.333802,
		0.300538,
		0.265509,
		0.229374,
		0.192875,
		0.156828,
		0.122125,
		0.089722,
		0.06064,
		0.03596,
		0.016822,
		0.00442,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.00138,
		0.011815,
		0.031171,
		0.057933,
		0.090652,
		0.127953,
		0.168534,
		0.211175,
		0.254737,
		0.298172,
		0.340527,
		0.38095,
		0.4187,
		0.453154,
		0.48382,
		0.510346,
		0.532534,
		0.55035,
		0.56394,
		0.57363,
		0.579929,
		0.583504,
		0.585137,
		0.585637,
		0.585696,
		0.585688,
		0.585481,
		0.584512,
		0.58201,
		0.577165,
		0.569245,
		0.557657,
		0.541984,
		0.521988,
		0.497617,
		0.468986,
		0.436373,
		0.400201,
		0.361029,
		0.31954,
		0.276532,
		0.232908,
		0.18967,
		0.147912,
		0.108813,
		0.073636,
		0.043719,
		0.020474,
		0.005385,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.001639,
		0.014018,
		0.036943,
		0.068577,
		0.107171,
		0.151061,
		0.198682,
		0.248563,
		0.299339,
		0.349758,
		0.398686,
		0.445116,
		0.488182,
		0.527165,
		0.561514,
		0.590853,
		0.615002,
		0.633993,
		0.648076,
		0.657731,
		0.663655,
		0.666724,
		0.667916,
		0.668171,
		0.668181,
		0.668108,
		0.667493,
		0.665486,
		0.661108,
		0.653418,
		0.641623,
		0.625132,
		0.603579,
		0.576825,
		0.544946,
		0.508222,
		0.467118,
		0.422271,
		0.374475,
		0.324667,
		0.27392,
		0.223428,
		0.1745,
		0.128554,
		0.08711,
		0.051783,
		0.024279,
		0.006393,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.001902,
		0.016247,
		0.042764,
		0.079284,
		0.123736,
		0.174161,
		0.228713,
		0.285665,
		0.343419,
		0.400512,
		0.455628,
		0.507609,
		0.555467,
		0.598399,
		0.635805,
		0.667309,
		0.692773,
		0.712321,
		0.726348,
		0.735522,
		0.740761,
		0.74317,
		0.743911,
		0.743999,
		0.743987,
		0.743679,
		0.742246,
		0.738568,
		0.73149,
		0.719989,
		0.703269,
		0.680795,
		0.652309,
		0.617822,
		0.577591,
		0.532108,
		0.482077,
		0.428393,
		0.372134,
		0.314536,
		0.256992,
		0.201032,
		0.14832,
		0.100644,
		0.059906,
		0.028122,
		0.007413,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.002161,
		0.018437,
		0.048471,
		0.089745,
		0.139866,
		0.196569,
		0.257723,
		0.321343,
		0.385597,
		0.448814,
		0.5095,
		0.566352,
		0.618271,
		0.664385,
		0.704067,
		0.736961,
		0.763003,
		0.782445,
		0.795861,
		0.804146,
		0.808465,
		0.810153,
		0.810517,
		0.810531,
		0.810427,
		0.809553,
		0.80672,
		0.800578,
		0.789857,
		0.773521,
		0.75083,
		0.721373,
		0.685059,
		0.642099,
		0.592988,
		0.538479,
		0.47956,
		0.417432,
		0.353499,
		0.289343,
		0.226717,
		0.167531,
		0.113845,
		0.067856,
		0.031895,
		0.008417,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.00241,
		0.020541,
		0.053931,
		0.099716,
		0.155176,
		0.21774,
		0.284996,
		0.354703,
		0.424797,
		0.493408,
		0.558873,
		0.619756,
		0.674863,
		0.723272,
		0.764353,
		0.7978,
		0.823657,
		0.842338,
		0.85464,
		0.861713,
		0.864986,
		0.865999,
		0.86612,
		0.866103,
		0.865681,
		0.863728,
		0.858746,
		0.849223,
		0.833854,
		0.81166,
		0.782037,
		0.744755,
		0.699951,
		0.648096,
		0.589968,
		0.526629,
		0.459397,
		0.389823,
		0.319677,
		0.250926,
		0.185725,
		0.126401,
		0.075448,
		0.035511,
		0.009384,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.002646,
		0.022525,
		0.05906,
		0.109042,
		0.169425,
		0.237338,
		0.310092,
		0.385195,
		0.460362,
		0.533535,
		0.602893,
		0.666884,
		0.724239,
		0.774009,
		0.815588,
		0.848753,
		0.873692,
		0.891026,
		0.901806,
		0.907465,
		0.909691,
		0.910174,
		0.910193,
		0.910055,
		0.908898,
		0.905174,
		0.897154,
		0.883254,
		0.862222,
		0.83322,
		0.795845,
		0.750119,
		0.696458,
		0.635646,
		0.568794,
		0.497317,
		0.422905,
		0.347498,
		0.27327,
		0.202613,
		0.138117,
		0.082565,
		0.038915,
		0.010297,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.002868,
		0.02438,
		0.063833,
		0.117674,
		0.182539,
		0.255258,
		0.332875,
		0.412655,
		0.492103,
		0.568982,
		0.641333,
		0.707502,
		0.766169,
		0.816383,
		0.857596,
		0.889702,
		0.913075,
		0.928581,
		0.937562,
		0.941747,
		0.943052,
		0.943208,
		0.943187,
		0.942641,
		0.940135,
		0.933787,
		0.921738,
		0.902438,
		0.874775,
		0.838132,
		0.792377,
		0.737842,
		0.675281,
		0.605834,
		0.530993,
		0.452566,
		0.372656,
		0.29363,
		0.218106,
		0.148931,
		0.08917,
		0.04209,
		0.011152,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.003077,
		0.026116,
		0.068277,
		0.125665,
		0.194598,
		0.271617,
		0.3535,
		0.437279,
		0.520259,
		0.600037,
		0.674529,
		0.742003,
		0.801108,
		0.850921,
		0.890985,
		0.921354,
		0.942628,
		0.955961,
		0.963014,
		0.965809,
		0.966419,
		0.966443,
		0.966268,
		0.96481,
		0.960152,
		0.950194,
		0.933069,
		0.907357,
		0.872181,
		0.827216,
		0.77267,
		0.709237,
		0.63806,
		0.560686,
		0.479026,
		0.395328,
		0.312142,
		0.232304,
		0.15891,
		0.095303,
		0.045055,
		0.011955,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.003275,
		0.027762,
		0.072466,
		0.133151,
		0.205814,
		0.286708,
		0.372351,
		0.459545,
		0.545402,
		0.627366,
		0.703245,
		0.771243,
		0.830009,
		0.878677,
		0.916918,
		0.944988,
		0.963764,
		0.974726,
		0.979875,
		0.981492,
		0.981687,
		0.981661,
		0.980982,
		0.977887,
		0.970107,
		0.955459,
		0.932184,
		0.899099,
		0.855639,
		0.801847,
		0.738324,
		0.666188,
		0.587017,
		0.502809,
		0.415942,
		0.329143,
		0.245457,
		0.168226,
		0.101067,
		0.047857,
		0.012718,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.003469,
		0.029357,
		0.076504,
		0.140319,
		0.216478,
		0.300936,
		0.38995,
		0.480096,
		0.568298,
		0.651854,
		0.728476,
		0.796331,
		0.854087,
		0.900971,
		0.936827,
		0.962163,
		0.978178,
		0.986721,
		0.990133,
		0.990884,
		0.990914,
		0.990698,
		0.988911,
		0.983245,
		0.971234,
		0.950751,
		0.920262,
		0.878913,
		0.826533,
		0.763595,
		0.691157,
		0.61081,
		0.524618,
		0.435083,
		0.345096,
		0.257911,
		0.177115,
		0.106605,
		0.050566,
		0.013459,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.003662,
		0.030947,
		0.080506,
		0.147381,
		0.22691,
		0.314743,
		0.406867,
		0.499627,
		0.589761,
		0.674429,
		0.751262,
		0.818404,
		0.874577,
		0.919139,
		0.952154,
		0.974436,
		0.987562,
		0.993781,
		0.995751,
		0.99599,
		0.995957,
		0.995127,
		0.991373,
		0.982016,
		0.964555,
		0.937057,
		0.898322,
		0.847917,
		0.786138,
		0.713954,
		0.632938,
		0.54521,
		0.453381,
		0.360506,
		0.270048,
		0.185844,
		0.112078,
		0.053259,
		0.014199,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.003862,
		0.032575,
		0.084585,
		0.154541,
		0.237419,
		0.328548,
		0.423631,
		0.518777,
		0.610534,
		0.695929,
		0.772521,
		0.83846,
		0.892551,
		0.93433,
		0.964136,
		0.983155,
		0.993394,
		0.997522,
		0.998438,
		0.998474,
		0.99821,
		0.996039,
		0.989217,
		0.974887,
		0.950676,
		0.914977,
		0.867033,
		0.806913,
		0.735456,
		0.654192,
		0.565277,
		0.471423,
		0.375848,
		0.282229,
		0.194665,
		0.117642,
		0.05601,
		0.014959,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.004071,
		0.034281,
		0.088841,
		0.161974,
		0.248267,
		0.342703,
		0.440684,
		0.538067,
		0.631206,
		0.717001,
		0.792954,
		0.857243,
		0.908799,
		0.94739,
		0.973698,
		0.989349,
		0.996836,
		0.99923,
		0.999523,
		0.999483,
		0.99847,
		0.993929,
		0.982718,
		0.961999,
		0.929694,
		0.884648,
		0.826642,
		0.756338,
		0.675192,
		0.585374,
		0.48969,
		0.39152,
		0.294765,
		0.203799,
		0.123433,
		0.058887,
		0.015757,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.004295,
		0.036097,
		0.093355,
		0.169822,
		0.259662,
		0.357477,
		0.458348,
		0.557864,
		0.652181,
		0.738071,
		0.813001,
		0.875207,
		0.923796,
		0.958834,
		0.981426,
		0.993717,
		0.998721,
		0.999841,
		0.999886,
		0.999561,
		0.996917,
		0.988688,
		0.971575,
		0.942964,
		0.901223,
		0.845772,
		0.777035,
		0.696355,
		0.605889,
		0.508529,
		0.407819,
		0.307892,
		0.213418,
		0.129562,
		0.061944,
		0.016608,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.004537,
		0.03805,
		0.098192,
		0.178194,
		0.271752,
		0.373058,
		0.476837,
		0.578397,
		0.673685,
		0.759357,
		0.832862,
		0.892541,
		0.937734,
		0.96889,
		0.987631,
		0.996693,
		0.99962,
		0.999981,
		0.999932,
		0.998688,
		0.993163,
		0.979666,
		0.954993,
		0.916947,
		0.864496,
		0.797756,
		0.717902,
		0.627051,
		0.528163,
		0.424946,
		0.321779,
		0.223654,
		0.136115,
		0.065226,
		0.017525,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389572,
		0.496281,
		0.599782,
		0.695809,
		0.780912,
		0.852554,
		0.909231,
		0.9506,
		0.97759,
		0.992443,
		0.998558,
		0.999942,
		0.999998,
		0.999595,
		0.996346,
		0.986344,
		0.965783,
		0.931802,
		0.882813,
		0.818535,
		0.739907,
		0.648966,
		0.548717,
		0.443032,
		0.336549,
		0.234602,
		0.143159,
		0.068768,
		0.018518,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.006111,
		0.050639,
		0.128949,
		0.230606,
		0.346048,
		0.466665,
		0.584888,
		0.694316,
		0.789856,
		0.867913,
		0.926607,
		0.966004,
		0.988294,
		0.997721,
		0.999909,
		0.999999,
		0.999355,
		0.994282,
		0.979066,
		0.948631,
		0.899725,
		0.83123,
		0.744084,
		0.641052,
		0.526505,
		0.406208,
		0.287165,
		0.177478,
		0.086243,
		0.023468,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.006524,
		0.053897,
		0.136799,
		0.243759,
		0.364313,
		0.489096,
		0.609954,
		0.720092,
		0.814257,
		0.888967,
		0.942774,
		0.976536,
		0.993551,
		0.999268,
		0.999999,
		0.999897,
		0.997422,
		0.986838,
		0.962012,
		0.918486,
		0.854185,
		0.769433,
		0.66673,
		0.550474,
		0.426716,
		0.302949,
		0.18795,
		0.091646,
		0.025016,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.006979,
		0.057479,
		0.145377,
		0.258031,
		0.383959,
		0.512956,
		0.636233,
		0.746597,
		0.838682,
		0.909232,
		0.957423,
		0.985151,
		0.997073,
		0.999882,
		0.999998,
		0.999166,
		0.9927,
		0.973615,
		0.93607,
		0.876762,
		0.795181,
		0.69341,
		0.575799,
		0.448658,
		0.320005,
		0.199356,
		0.09757,
		0.026722,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.007484,
		0.061429,
		0.154777,
		0.273547,
		0.405109,
		0.538323,
		0.663715,
		0.773702,
		0.862877,
		0.928371,
		0.97023,
		0.991705,
		0.999046,
		0.999998,
		0.999864,
		0.996662,
		0.983187,
		0.952128,
		0.898654,
		0.821134,
		0.721027,
		0.602518,
		0.472143,
		0.338462,
		0.211807,
		0.104082,
		0.028608,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.008046,
		0.065798,
		0.165103,
		0.290446,
		0.427896,
		0.565272,
		0.692366,
		0.801227,
		0.886525,
		0.945995,
		0.980889,
		0.996177,
		0.999844,
		0.999998,
		0.998903,
		0.990536,
		0.966289,
		0.919497,
		0.84704,
		0.749475,
		0.630658,
		0.497281,
		0.358463,
		0.225429,
		0.111261,
		0.0307,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.008674,
		0.070648,
		0.176478,
		0.308888,
		0.452462,
		0.593863,
		0.722107,
		0.828924,
		0.909232,
		0.961681,
		0.989154,
		0.998733,
		0.999998,
		0.999819,
		0.995601,
		0.978187,
		0.938862,
		0.872571,
		0.778592,
		0.660215,
		0.52418,
		0.380165,
		0.24037,
		0.119202,
		0.033031,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.009378,
		0.076051,
		0.189047,
		0.32905,
		0.478955,
		0.624138,
		0.752804,
		0.856465,
		0.930531,
		0.974996,
		0.994912,
		0.999788,
		0.999997,
		0.998528,
		0.987511,
		0.956268,
		0.897313,
		0.808146,
		0.69115,
		0.552945,
		0.403741,
		0.256796,
		0.128015,
		0.035637,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.010171,
		0.082095,
		0.202976,
		0.351133,
		0.507526,
		0.656102,
		0.78425,
		0.883417,
		0.949876,
		0.985548,
		0.99828,
		0.999997,
		0.999752,
		0.994084,
		0.971204,
		0.920755,
		0.837812,
		0.723368,
		0.583667,
		0.429378,
		0.2749,
		0.137829,
		0.038563,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.011069,
		0.088884,
		0.218463,
		0.375363,
		0.538323,
		0.689715,
		0.816138,
		0.909232,
		0.966673,
		0.993081,
		0.999707,
		0.999996,
		0.997978,
		0.983187,
		0.94229,
		0.867151,
		0.756699,
		0.61641,
		0.457278,
		0.294906,
		0.148801,
		0.041863,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.012092,
		0.096544,
		0.23574,
		0.40199,
		0.571482,
		0.72486,
		0.848031,
		0.933236,
		0.980326,
		0.997607,
		0.999995,
		0.999651,
		0.991857,
		0.961225,
		0.895582,
		0.790869,
		0.651201,
		0.487653,
		0.317068,
		0.161115,
		0.045605,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.013263,
		0.105229,
		0.255081,
		0.431293,
		0.607109,
		0.761317,
		0.879329,
		0.95463,
		0.990352,
		0.999581,
		0.999994,
		0.997147,
		0.976835,
		0.922365,
		0.825463,
		0.687996,
		0.520718,
		0.341684,
		0.174994,
		0.049871,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.014613,
		0.11513,
		0.276811,
		0.463572,
		0.645261,
		0.798713,
		0.909232,
		0.972543,
		0.996571,
		0.999993,
		0.999493,
		0.988485,
		0.946591,
		0.859876,
		0.726654,
		0.556677,
		0.369091,
		0.19071,
		0.054761,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.016179,
		0.12648,
		0.301315,
		0.499142,
		0.685901,
		0.836467,
		0.936718,
		0.986146,
		0.999381,
		0.999991,
		0.995843,
		0.967218,
		0.893264,
		0.766878,
		0.595706,
		0.399673,
		0.208589,
		0.060404,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.018011,
		0.139573,
		0.32905,
		0.538323,
		0.728851,
		0.873716,
		0.960553,
		0.994912,
		0.999989,
		0.999236,
		0.983187,
		0.924495,
		0.808146,
		0.637915,
		0.433865,
		0.229035,
		0.066963,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.020173,
		0.154777,
		0.360554,
		0.581406,
		0.773702,
		0.909232,
		0.9794,
		0.999046,
		0.999986,
		0.993706,
		0.952128,
		0.849616,
		0.683291,
		0.472143,
		0.25254,
		0.074644,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.022747,
		0.172561,
		0.39646,
		0.628601,
		0.819691,
		0.941362,
		0.992122,
		0.999982,
		0.998793,
		0.974494,
		0.890001,
		0.731605,
		0.515011,
		0.279714,
		0.083718,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.025825,
		0.19336,
		0.437117,
		0.679351,
		0.864779,
		0.967203,
		0.997575,
		0.9991,
		0.989142,
		0.926633,
		0.781582,
		0.56247,
		0.311037,
		0.094456,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.028884,
		0.212988,
		0.472358,
		0.716734,
		0.886448,
		0.96241,
		0.974912,
		0.97297,
		0.935426,
		0.81321,
		0.600964,
		0.339533,
		0.104887,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.028801,
		0.208583,
		0.45206,
		0.666205,
		0.795954,
		0.837497,
		0.839721,
		0.825636,
		0.743241,
		0.567267,
		0.328938,
		0.103703,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.021807,
		0.154555,
		0.325662,
		0.462908,
		0.530799,
		0.542935,
		0.540983,
		0.505937,
		0.401745,
		0.240551,
		0.077731,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.010604,
		0.073201,
		0.148905,
		0.202269,
		0.22133,
		0.222444,
		0.215674,
		0.179781,
		0.112084,
		0.037339,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.002407,
		0.016081,
		0.031269,
		0.040099,
		0.041877,
		0.041577,
		0.036721,
		0.024116,
		0.008349,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		5.1e-005,
		0.000326,
		0.000596,
		0.000712,
		0.000719,
		0.000676,
		0.000475,
		0.000173,
		0.000173,
		0.000475,
		0.000676,
		0.000719,
		0.000712,
		0.000596,
		0.000326,
		5.1e-005,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		0.008349,
		0.024116,
		0.036721,
		0.041577,
		0.041877,
		0.040099,
		0.031269,
		0.016081,
		0.002407,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.037339,
		0.112084,
		0.179781,
		0.215674,
		0.222444,
		0.22133,
		0.202269,
		0.148905,
		0.073201,
		0.010604,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.077731,
		0.240551,
		0.401745,
		0.505937,
		0.540983,
		0.542935,
		0.530799,
		0.462908,
		0.325662,
		0.154555,
		0.021807,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.103703,
		0.328938,
		0.567267,
		0.743241,
		0.825636,
		0.839721,
		0.837497,
		0.795954,
		0.666205,
		0.45206,
		0.208583,
		0.028801,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.104887,
		0.339533,
		0.600964,
		0.81321,
		0.935426,
		0.97297,
		0.974912,
		0.96241,
		0.886448,
		0.716734,
		0.472358,
		0.212988,
		0.028884,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.094456,
		0.311037,
		0.56247,
		0.781582,
		0.926633,
		0.989142,
		0.9991,
		0.997575,
		0.967203,
		0.864779,
		0.679351,
		0.437117,
		0.19336,
		0.025825,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.083718,
		0.279714,
		0.515011,
		0.731605,
		0.890001,
		0.974494,
		0.998793,
		0.999982,
		0.992122,
		0.941362,
		0.819691,
		0.628601,
		0.39646,
		0.172561,
		0.022747,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.074644,
		0.25254,
		0.472143,
		0.683291,
		0.849616,
		0.952128,
		0.993706,
		0.999986,
		0.999046,
		0.9794,
		0.909232,
		0.773702,
		0.581406,
		0.360554,
		0.154777,
		0.020173,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.066963,
		0.229035,
		0.433865,
		0.637915,
		0.808146,
		0.924495,
		0.983187,
		0.999236,
		0.999989,
		0.994912,
		0.960553,
		0.873716,
		0.728851,
		0.538323,
		0.32905,
		0.139573,
		0.018011,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.060404,
		0.208589,
		0.399673,
		0.595706,
		0.766878,
		0.893264,
		0.967218,
		0.995843,
		0.999991,
		0.999381,
		0.986146,
		0.936718,
		0.836467,
		0.685901,
		0.499142,
		0.301315,
		0.12648,
		0.016179,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.054761,
		0.19071,
		0.369091,
		0.556677,
		0.726654,
		0.859876,
		0.946591,
		0.988485,
		0.999493,
		0.999993,
		0.996571,
		0.972543,
		0.909232,
		0.798713,
		0.645261,
		0.463572,
		0.276811,
		0.11513,
		0.014613,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.049871,
		0.174994,
		0.341684,
		0.520718,
		0.687996,
		0.825463,
		0.922365,
		0.976835,
		0.997147,
		0.999994,
		0.999581,
		0.990352,
		0.95463,
		0.879329,
		0.761317,
		0.607109,
		0.431293,
		0.255081,
		0.105229,
		0.013263,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.045605,
		0.161115,
		0.317068,
		0.487653,
		0.651201,
		0.790869,
		0.895582,
		0.961225,
		0.991857,
		0.999651,
		0.999995,
		0.997607,
		0.980326,
		0.933236,
		0.848031,
		0.72486,
		0.571482,
		0.40199,
		0.23574,
		0.096544,
		0.012092,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.041863,
		0.148801,
		0.294906,
		0.457278,
		0.61641,
		0.756699,
		0.867151,
		0.94229,
		0.983187,
		0.997978,
		0.999996,
		0.999707,
		0.993081,
		0.966673,
		0.909232,
		0.816138,
		0.689715,
		0.538323,
		0.375363,
		0.218463,
		0.088884,
		0.011069,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.038563,
		0.137829,
		0.2749,
		0.429378,
		0.583667,
		0.723368,
		0.837812,
		0.920755,
		0.971204,
		0.994084,
		0.999752,
		0.999997,
		0.99828,
		0.985548,
		0.949876,
		0.883417,
		0.78425,
		0.656102,
		0.507526,
		0.351133,
		0.202976,
		0.082095,
		0.010171,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.035637,
		0.128015,
		0.256796,
		0.403741,
		0.552945,
		0.69115,
		0.808146,
		0.897313,
		0.956268,
		0.987511,
		0.998528,
		0.999997,
		0.999788,
		0.994912,
		0.974996,
		0.930531,
		0.856465,
		0.752804,
		0.624138,
		0.478955,
		0.32905,
		0.189047,
		0.076051,
		0.009378,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.033031,
		0.119202,
		0.24037,
		0.380165,
		0.52418,
		0.660215,
		0.778592,
		0.872571,
		0.938862,
		0.978187,
		0.995601,
		0.999819,
		0.999998,
		0.998733,
		0.989154,
		0.961681,
		0.909232,
		0.828924,
		0.722107,
		0.593863,
		0.452462,
		0.308888,
		0.176478,
		0.070648,
		0.008674,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.0307,
		0.111261,
		0.225429,
		0.358463,
		0.497281,
		0.630658,
		0.749475,
		0.84704,
		0.919497,
		0.966289,
		0.990536,
		0.998903,
		0.999998,
		0.999844,
		0.996177,
		0.980889,
		0.945995,
		0.886525,
		0.801227,
		0.692366,
		0.565272,
		0.427896,
		0.290446,
		0.165103,
		0.065798,
		0.008046,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.028608,
		0.104082,
		0.211807,
		0.338462,
		0.472143,
		0.602518,
		0.721027,
		0.821134,
		0.898654,
		0.952128,
		0.983187,
		0.996662,
		0.999864,
		0.999998,
		0.999046,
		0.991705,
		0.97023,
		0.928371,
		0.862877,
		0.773702,
		0.663715,
		0.538323,
		0.405109,
		0.273547,
		0.154777,
		0.061429,
		0.007484,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.026722,
		0.09757,
		0.199356,
		0.320005,
		0.448658,
		0.575799,
		0.69341,
		0.795181,
		0.876762,
		0.93607,
		0.973615,
		0.9927,
		0.999166,
		0.999998,
		0.999882,
		0.997073,
		0.985151,
		0.957423,
		0.909232,
		0.838682,
		0.746597,
		0.636233,
		0.512956,
		0.383959,
		0.258031,
		0.145377,
		0.057479,
		0.006979,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.025016,
		0.091646,
		0.18795,
		0.302949,
		0.426716,
		0.550474,
		0.66673,
		0.769433,
		0.854185,
		0.918486,
		0.962012,
		0.986838,
		0.997422,
		0.999897,
		0.999999,
		0.999268,
		0.993551,
		0.976536,
		0.942774,
		0.888967,
		0.814257,
		0.720092,
		0.609954,
		0.489096,
		0.364313,
		0.243759,
		0.136799,
		0.053897,
		0.006524,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.023468,
		0.086243,
		0.177478,
		0.287165,
		0.406208,
		0.526505,
		0.641052,
		0.744084,
		0.83123,
		0.899725,
		0.948631,
		0.979066,
		0.994282,
		0.999355,
		0.999999,
		0.999909,
		0.997721,
		0.988294,
		0.966004,
		0.926607,
		0.867913,
		0.789856,
		0.694316,
		0.584888,
		0.466665,
		0.346048,
		0.230606,
		0.128949,
		0.050639,
		0.006111,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.018518,
		0.068768,
		0.143159,
		0.234602,
		0.336549,
		0.443032,
		0.548717,
		0.648966,
		0.739907,
		0.818535,
		0.882813,
		0.931802,
		0.965783,
		0.986344,
		0.996346,
		0.999595,
		0.999998,
		0.999942,
		0.998558,
		0.992443,
		0.97759,
		0.9506,
		0.909231,
		0.852554,
		0.780912,
		0.695809,
		0.599782,
		0.496281,
		0.389572,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.017525,
		0.065226,
		0.136115,
		0.223654,
		0.321779,
		0.424946,
		0.528163,
		0.627051,
		0.717902,
		0.797756,
		0.864496,
		0.916947,
		0.954993,
		0.979666,
		0.993163,
		0.998688,
		0.999932,
		0.999981,
		0.99962,
		0.996693,
		0.987631,
		0.96889,
		0.937734,
		0.892541,
		0.832862,
		0.759357,
		0.673685,
		0.578397,
		0.476837,
		0.373058,
		0.271752,
		0.178194,
		0.098192,
		0.03805,
		0.004537,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.016608,
		0.061944,
		0.129562,
		0.213418,
		0.307892,
		0.407819,
		0.508529,
		0.605889,
		0.696355,
		0.777035,
		0.845772,
		0.901223,
		0.942964,
		0.971575,
		0.988688,
		0.996917,
		0.999561,
		0.999886,
		0.999841,
		0.998721,
		0.993717,
		0.981426,
		0.958834,
		0.923796,
		0.875207,
		0.813001,
		0.738071,
		0.652181,
		0.557864,
		0.458348,
		0.357477,
		0.259662,
		0.169822,
		0.093355,
		0.036097,
		0.004295,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.015757,
		0.058887,
		0.123433,
		0.203799,
		0.294765,
		0.39152,
		0.48969,
		0.585374,
		0.675192,
		0.756338,
		0.826642,
		0.884648,
		0.929694,
		0.961999,
		0.982718,
		0.993929,
		0.99847,
		0.999483,
		0.999523,
		0.99923,
		0.996836,
		0.989349,
		0.973698,
		0.94739,
		0.908799,
		0.857243,
		0.792954,
		0.717001,
		0.631206,
		0.538067,
		0.440684,
		0.342703,
		0.248267,
		0.161974,
		0.088841,
		0.034281,
		0.004071,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.014959,
		0.05601,
		0.117642,
		0.194665,
		0.282229,
		0.375848,
		0.471423,
		0.565277,
		0.654192,
		0.735456,
		0.806913,
		0.867033,
		0.914977,
		0.950676,
		0.974887,
		0.989217,
		0.996039,
		0.99821,
		0.998474,
		0.998438,
		0.997522,
		0.993394,
		0.983155,
		0.964136,
		0.93433,
		0.892551,
		0.83846,
		0.772521,
		0.695929,
		0.610534,
		0.518777,
		0.423631,
		0.328548,
		0.237419,
		0.154541,
		0.084585,
		0.032575,
		0.003862,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.014199,
		0.053259,
		0.112078,
		0.185844,
		0.270048,
		0.360506,
		0.453381,
		0.54521,
		0.632938,
		0.713954,
		0.786138,
		0.847917,
		0.898322,
		0.937057,
		0.964555,
		0.982016,
		0.991373,
		0.995127,
		0.995957,
		0.99599,
		0.995751,
		0.993781,
		0.987562,
		0.974436,
		0.952154,
		0.919139,
		0.874577,
		0.818404,
		0.751262,
		0.674429,
		0.589761,
		0.499627,
		0.406867,
		0.314743,
		0.22691,
		0.147381,
		0.080506,
		0.030947,
		0.003662,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.013459,
		0.050566,
		0.106605,
		0.177115,
		0.257911,
		0.345096,
		0.435083,
		0.524618,
		0.61081,
		0.691157,
		0.763595,
		0.826533,
		0.878913,
		0.920262,
		0.950751,
		0.971234,
		0.983245,
		0.988911,
		0.990698,
		0.990914,
		0.990884,
		0.990133,
		0.986721,
		0.978178,
		0.962163,
		0.936827,
		0.900971,
		0.854087,
		0.796331,
		0.728476,
		0.651854,
		0.568298,
		0.480096,
		0.38995,
		0.300936,
		0.216478,
		0.140319,
		0.076504,
		0.029357,
		0.003469,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.012718,
		0.047857,
		0.101067,
		0.168226,
		0.245457,
		0.329143,
		0.415942,
		0.502809,
		0.587017,
		0.666188,
		0.738324,
		0.801847,
		0.855639,
		0.899099,
		0.932184,
		0.955459,
		0.970107,
		0.977887,
		0.980982,
		0.981661,
		0.981687,
		0.981492,
		0.979875,
		0.974726,
		0.963764,
		0.944988,
		0.916918,
		0.878677,
		0.830009,
		0.771243,
		0.703245,
		0.627366,
		0.545402,
		0.459545,
		0.372351,
		0.286708,
		0.205814,
		0.133151,
		0.072466,
		0.027762,
		0.003275,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.011955,
		0.045055,
		0.095303,
		0.15891,
		0.232304,
		0.312142,
		0.395328,
		0.479026,
		0.560686,
		0.63806,
		0.709237,
		0.77267,
		0.827216,
		0.872181,
		0.907357,
		0.933069,
		0.950194,
		0.960152,
		0.96481,
		0.966268,
		0.966443,
		0.966419,
		0.965809,
		0.963014,
		0.955961,
		0.942628,
		0.921354,
		0.890985,
		0.850921,
		0.801108,
		0.742003,
		0.674529,
		0.600037,
		0.520259,
		0.437279,
		0.3535,
		0.271617,
		0.194598,
		0.125665,
		0.068277,
		0.026116,
		0.003077,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.011152,
		0.04209,
		0.08917,
		0.148931,
		0.218106,
		0.29363,
		0.372656,
		0.452566,
		0.530993,
		0.605834,
		0.675281,
		0.737842,
		0.792377,
		0.838132,
		0.874775,
		0.902438,
		0.921738,
		0.933787,
		0.940135,
		0.942641,
		0.943187,
		0.943208,
		0.943052,
		0.941747,
		0.937562,
		0.928581,
		0.913075,
		0.889702,
		0.857596,
		0.816383,
		0.766169,
		0.707502,
		0.641333,
		0.568982,
		0.492103,
		0.412655,
		0.332875,
		0.255258,
		0.182539,
		0.117674,
		0.063833,
		0.02438,
		0.002868,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.010297,
		0.038915,
		0.082565,
		0.138117,
		0.202613,
		0.27327,
		0.347498,
		0.422905,
		0.497317,
		0.568794,
		0.635646,
		0.696458,
		0.750119,
		0.795845,
		0.83322,
		0.862222,
		0.883254,
		0.897154,
		0.905174,
		0.908898,
		0.910055,
		0.910193,
		0.910174,
		0.909691,
		0.907465,
		0.901806,
		0.891026,
		0.873692,
		0.848753,
		0.815588,
		0.774009,
		0.724239,
		0.666884,
		0.602893,
		0.533535,
		0.460362,
		0.385195,
		0.310092,
		0.237338,
		0.169425,
		0.109042,
		0.05906,
		0.022525,
		0.002646,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.009384,
		0.035511,
		0.075448,
		0.126401,
		0.185725,
		0.250926,
		0.319677,
		0.389823,
		0.459397,
		0.526629,
		0.589968,
		0.648096,
		0.699951,
		0.744755,
		0.782037,
		0.81166,
		0.833854,
		0.849223,
		0.858746,
		0.863728,
		0.865681,
		0.866103,
		0.86612,
		0.865999,
		0.864986,
		0.861713,
		0.85464,
		0.842338,
		0.823657,
		0.7978,
		0.764353,
		0.723272,
		0.674863,
		0.619756,
		0.558873,
		0.493408,
		0.424797,
		0.354703,
		0.284996,
		0.21774,
		0.155176,
		0.099716,
		0.053931,
		0.020541,
		0.00241,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.008417,
		0.031895,
		0.067856,
		0.113845,
		0.167531,
		0.226717,
		0.289343,
		0.353499,
		0.417432,
		0.47956,
		0.538479,
		0.592988,
		0.642099,
		0.685059,
		0.721373,
		0.75083,
		0.773521,
		0.789857,
		0.800578,
		0.80672,
		0.809553,
		0.810427,
		0.810531,
		0.810517,
		0.810153,
		0.808465,
		0.804146,
		0.795861,
		0.782445,
		0.763003,
		0.736961,
		0.704067,
		0.664385,
		0.618271,
		0.566352,
		0.5095,
		0.448814,
		0.385597,
		0.321343,
		0.257723,
		0.196569,
		0.139866,
		0.089745,
		0.048471,
		0.018437,
		0.002161,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.007413,
		0.028122,
		0.059906,
		0.100644,
		0.14832,
		0.201032,
		0.256992,
		0.314536,
		0.372134,
		0.428393,
		0.482077,
		0.532108,
		0.577591,
		0.617822,
		0.652309,
		0.680795,
		0.703269,
		0.719989,
		0.73149,
		0.738568,
		0.742246,
		0.743679,
		0.743987,
		0.743999,
		0.743911,
		0.74317,
		0.740761,
		0.735522,
		0.726348,
		0.712321,
		0.692773,
		0.667309,
		0.635805,
		0.598399,
		0.555467,
		0.507609,
		0.455628,
		0.400512,
		0.343419,
		0.285665,
		0.228713,
		0.174161,
		0.123736,
		0.079284,
		0.042764,
		0.016247,
		0.001902,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.006393,
		0.024279,
		0.051783,
		0.08711,
		0.128554,
		0.1745,
		0.223428,
		0.27392,
		0.324667,
		0.374475,
		0.422271,
		0.467118,
		0.508222,
		0.544946,
		0.576825,
		0.603579,
		0.625132,
		0.641623,
		0.653418,
		0.661108,
		0.665486,
		0.667493,
		0.668108,
		0.668181,
		0.668171,
		0.667916,
		0.666724,
		0.663655,
		0.657731,
		0.648076,
		0.633993,
		0.615002,
		0.590853,
		0.561514,
		0.527165,
		0.488182,
		0.445116,
		0.398686,
		0.349758,
		0.299339,
		0.248563,
		0.198682,
		0.151061,
		0.107171,
		0.068577,
		0.036943,
		0.014018,
		0.001639,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.005385,
		0.020474,
		0.043719,
		0.073636,
		0.108813,
		0.147912,
		0.18967,
		0.232908,
		0.276532,
		0.31954,
		0.361029,
		0.400201,
		0.436373,
		0.468986,
		0.497617,
		0.521988,
		0.541984,
		0.557657,
		0.569245,
		0.577165,
		0.58201,
		0.584512,
		0.585481,
		0.585688,
		0.585696,
		0.585637,
		0.585137,
		0.583504,
		0.579929,
		0.57363,
		0.56394,
		0.55035,
		0.532534,
		0.510346,
		0.48382,
		0.453154,
		0.4187,
		0.38095,
		0.340527,
		0.298172,
		0.254737,
		0.211175,
		0.168534,
		0.127953,
		0.090652,
		0.057933,
		0.031171,
		0.011815,
		0.00138,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.00442,
		0.016822,
		0.03596,
		0.06064,
		0.089722,
		0.122125,
		0.156828,
		0.192875,
		0.229374,
		0.265509,
		0.300538,
		0.333802,
		0.36473,
		0.392847,
		0.417783,
		0.43928,
		0.457205,
		0.471554,
		0.482466,
		0.490224,
		0.495252,
		0.498098,
		0.499394,
		0.49979,
		0.499836,
		0.49983,
		0.499666,
		0.498898,
		0.496909,
		0.493048,
		0.486716,
		0.477426,
		0.464821,
		0.448693,
		0.428973,
		0.405731,
		0.379165,
		0.34959,
		0.317429,
		0.283205,
		0.247536,
		0.211121,
		0.174741,
		0.139251,
		0.105574,
		0.074699,
		0.04768,
		0.025625,
		0.009702,
		0.001132,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.003525,
		0.01343,
		0.028739,
		0.048517,
		0.071873,
		0.097956,
		0.125963,
		0.155141,
		0.184787,
		0.214251,
		0.242943,
		0.270336,
		0.295966,
		0.319444,
		0.340459,
		0.358785,
		0.374285,
		0.386925,
		0.396774,
		0.404013,
		0.408933,
		0.411925,
		0.413461,
		0.414053,
		0.414179,
		0.414184,
		0.414148,
		0.413844,
		0.412843,
		0.410642,
		0.406741,
		0.400704,
		0.392189,
		0.38096,
		0.366892,
		0.349969,
		0.330276,
		0.307995,
		0.283397,
		0.256833,
		0.228728,
		0.19958,
		0.169947,
		0.140451,
		0.111767,
		0.084624,
		0.059802,
		0.038126,
		0.020468,
		0.007742,
		0.000902,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.002723,
		0.010384,
		0.022245,
		0.037596,
		0.055758,
		0.076086,
		0.097968,
		0.120829,
		0.144131,
		0.167375,
		0.190107,
		0.211916,
		0.232442,
		0.251376,
		0.268467,
		0.283525,
		0.296425,
		0.307118,
		0.31563,
		0.322066,
		0.326616,
		0.329549,
		0.331199,
		0.331948,
		0.332175,
		0.332201,
		0.332198,
		0.332104,
		0.331662,
		0.330511,
		0.328265,
		0.324562,
		0.319097,
		0.311642,
		0.302048,
		0.29025,
		0.276259,
		0.260166,
		0.242125,
		0.22236,
		0.201151,
		0.178832,
		0.155789,
		0.132457,
		0.109311,
		0.086869,
		0.06569,
		0.046366,
		0.029527,
		0.015835,
		0.005983,
		0.000697,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.002029,
		0.007747,
		0.016611,
		0.028102,
		0.041725,
		0.057003,
		0.073489,
		0.090758,
		0.108413,
		0.126086,
		0.143438,
		0.160162,
		0.175988,
		0.19068,
		0.204045,
		0.21593,
		0.22623,
		0.234893,
		0.241918,
		0.247362,
		0.251342,
		0.254032,
		0.25566,
		0.256491,
		0.25681,
		0.256878,
		0.25688,
		0.256861,
		0.256697,
		0.256157,
		0.254963,
		0.252835,
		0.249525,
		0.244832,
		0.238608,
		0.230768,
		0.221283,
		0.21018,
		0.197539,
		0.183489,
		0.168203,
		0.151896,
		0.134822,
		0.117269,
		0.099561,
		0.082051,
		0.065122,
		0.049186,
		0.034677,
		0.02206,
		0.011818,
		0.004461,
		0.000519,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.001451,
		0.005545,
		0.011902,
		0.020157,
		0.029959,
		0.040976,
		0.052889,
		0.0654,
		0.078227,
		0.091109,
		0.103804,
		0.116093,
		0.12778,
		0.138695,
		0.148694,
		0.157661,
		0.165515,
		0.172206,
		0.177722,
		0.18209,
		0.185375,
		0.187686,
		0.189167,
		0.189997,
		0.190372,
		0.190485,
		0.190498,
		0.190496,
		0.19045,
		0.190229,
		0.189652,
		0.18852,
		0.186644,
		0.183862,
		0.180046,
		0.17511,
		0.169008,
		0.161731,
		0.153312,
		0.143817,
		0.133344,
		0.122023,
		0.110012,
		0.097494,
		0.084677,
		0.071791,
		0.059088,
		0.046839,
		0.035336,
		0.024886,
		0.015814,
		0.008464,
		0.003192,
		0.000371,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.000989,
		0.003782,
		0.008124,
		0.013772,
		0.02049,
		0.028054,
		0.036252,
		0.044882,
		0.053754,
		0.062691,
		0.07153,
		0.08012,
		0.088328,
		0.096035,
		0.103142,
		0.109565,
		0.115245,
		0.12014,
		0.124235,
		0.127539,
		0.130086,
		0.13194,
		0.133186,
		0.133936,
		0.134318,
		0.134464,
		0.134494,
		0.134495,
		0.134487,
		0.134412,
		0.134165,
		0.133616,
		0.132632,
		0.131095,
		0.128904,
		0.125984,
		0.122288,
		0.117792,
		0.112501,
		0.106443,
		0.099669,
		0.092251,
		0.084279,
		0.075864,
		0.067132,
		0.058225,
		0.049299,
		0.040526,
		0.032087,
		0.02418,
		0.017011,
		0.0108,
		0.005775,
		0.002176,
		0.000253,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000635,
		0.002431,
		0.005227,
		0.008869,
		0.013209,
		0.018105,
		0.023421,
		0.02903,
		0.034811,
		0.040651,
		0.046446,
		0.052099,
		0.057525,
		0.062646,
		0.067396,
		0.07172,
		0.075576,
		0.078936,
		0.081783,
		0.084118,
		0.085958,
		0.087336,
		0.0883,
		0.088916,
		0.089259,
		0.089413,
		0.089459,
		0.089465,
		0.089464,
		0.089445,
		0.089354,
		0.089116,
		0.088647,
		0.087866,
		0.086702,
		0.085099,
		0.083014,
		0.080424,
		0.077319,
		0.073709,
		0.069614,
		0.065071,
		0.060128,
		0.054846,
		0.049296,
		0.04356,
		0.03773,
		0.031906,
		0.026196,
		0.020718,
		0.015596,
		0.010961,
		0.006952,
		0.003714,
		0.001398,
		0.000162,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		0.000379,
		0.001451,
		0.003122,
		0.005302,
		0.007904,
		0.010844,
		0.014044,
		0.017426,
		0.020921,
		0.024461,
		0.027985,
		0.031435,
		0.034759,
		0.037912,
		0.040853,
		0.043547,
		0.04597,
		0.0481,
		0.049927,
		0.051448,
		0.052669,
		0.053606,
		0.054285,
		0.054739,
		0.055011,
		0.055149,
		0.055202,
		0.055213,
		0.055213,
		0.05521,
		0.055183,
		0.055094,
		0.054895,
		0.054538,
		0.053976,
		0.053172,
		0.052095,
		0.050725,
		0.049052,
		0.047072,
		0.044794,
		0.042233,
		0.039411,
		0.03636,
		0.033116,
		0.029723,
		0.026229,
		0.022689,
		0.019164,
		0.015716,
		0.012416,
		0.009337,
		0.006556,
		0.004154,
		0.002217,
		0.000834,
		9.7e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		0.000205,
		0.000785,
		0.001691,
		0.002874,
		0.004288,
		0.005889,
		0.007633,
		0.009482,
		0.011397,
		0.013341,
		0.015282,
		0.017189,
		0.019033,
		0.02079,
		0.022437,
		0.023956,
		0.025331,
		0.026551,
		0.027608,
		0.0285,
		0.029228,
		0.029799,
		0.030225,
		0.030522,
		0.03071,
		0.030815,
		0.030862,
		0.030876,
		0.030877,
		0.030877,
		0.030871,
		0.030844,
		0.030771,
		0.030628,
		0.030388,
		0.03003,
		0.029533,
		0.028884,
		0.028075,
		0.0271,
		0.025961,
		0.024662,
		0.023214,
		0.021629,
		0.019924,
		0.01812,
		0.016241,
		0.014314,
		0.012367,
		0.010433,
		0.008547,
		0.006746,
		0.005068,
		0.003555,
		0.002251,
		0.0012,
		0.000451,
		5.2e-005,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		9.7e-005,
		0.000371,
		0.000799,
		0.001359,
		0.00203,
		0.00279,
		0.00362,
		0.004501,
		0.005416,
		0.006347,
		0.007279,
		0.008198,
		0.009089,
		0.009942,
		0.010746,
		0.011491,
		0.01217,
		0.012777,
		0.013309,
		0.013763,
		0.014139,
		0.014439,
		0.014669,
		0.014835,
		0.014945,
		0.015011,
		0.015044,
		0.015057,
		0.015059,
		0.015059,
		0.015059,
		0.015052,
		0.015031,
		0.014983,
		0.014896,
		0.014759,
		0.014563,
		0.014298,
		0.01396,
		0.013546,
		0.013053,
		0.012483,
		0.011839,
		0.011126,
		0.010351,
		0.009521,
		0.008648,
		0.007741,
		0.006814,
		0.00588,
		0.004955,
		0.004055,
		0.003197,
		0.0024,
		0.001682,
		0.001064,
		0.000567,
		0.000213,
		2.5e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		3.7e-005,
		0.000143,
		0.000308,
		0.000524,
		0.000783,
		0.001077,
		0.001399,
		0.001742,
		0.002098,
		0.002461,
		0.002826,
		0.003186,
		0.003537,
		0.003874,
		0.004194,
		0.004491,
		0.004764,
		0.005009,
		0.005226,
		0.005413,
		0.005571,
		0.005698,
		0.005798,
		0.005872,
		0.005924,
		0.005956,
		0.005974,
		0.005982,
		0.005985,
		0.005985,
		0.005985,
		0.005984,
		0.005979,
		0.005967,
		0.005942,
		0.005901,
		0.005838,
		0.005752,
		0.005638,
		0.005496,
		0.005324,
		0.005121,
		0.00489,
		0.004631,
		0.004345,
		0.004037,
		0.003708,
		0.003363,
		0.003007,
		0.002644,
		0.002279,
		0.001918,
		0.001568,
		0.001235,
		0.000926,
		0.000649,
		0.00041,
		0.000218,
		8.2e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		1e-005,
		3.9e-005,
		8.4e-005,
		0.000143,
		0.000215,
		0.000295,
		0.000384,
		0.000479,
		0.000577,
		0.000678,
		0.000779,
		0.000879,
		0.000977,
		0.001072,
		0.001162,
		0.001246,
		0.001323,
		0.001394,
		0.001456,
		0.001511,
		0.001557,
		0.001596,
		0.001626,
		0.001649,
		0.001666,
		0.001677,
		0.001684,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001681,
		0.001672,
		0.001658,
		0.001639,
		0.001612,
		0.001578,
		0.001535,
		0.001485,
		0.001426,
		0.00136,
		0.001285,
		0.001205,
		0.001117,
		0.001025,
		0.000929,
		0.000829,
		0.000728,
		0.000627,
		0.000527,
		0.000431,
		0.000339,
		0.000254,
		0.000178,
		0.000112,
		6e-005,
		2.2e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.1e-005,
		2e-005,
		2.9e-005,
		4e-005,
		5.2e-005,
		6.5e-005,
		7.9e-005,
		9.3e-005,
		0.000107,
		0.000121,
		0.000134,
		0.000147,
		0.00016,
		0.000172,
		0.000183,
		0.000193,
		0.000202,
		0.00021,
		0.000216,
		0.000222,
		0.000227,
		0.00023,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000228,
		0.000224,
		0.000219,
		0.000213,
		0.000206,
		0.000197,
		0.000188,
		0.000177,
		0.000166,
		0.000154,
		0.000141,
		0.000127,
		0.000114,
		0.0001,
		8.6e-005,
		7.2e-005,
		5.9e-005,
		4.6e-005,
		3.5e-005,
		2.4e-005,
		1.5e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		1.7e-005,
		2.7e-005,
		3.7e-005,
		4.9e-005,
		6.2e-005,
		7.5e-005,
		8.9e-005,
		0.000103,
		0.000117,
		0.000131,
		0.000144,
		0.000157,
		0.000169,
		0.00018,
		0.00019,
		0.000199,
		0.000208,
		0.000215,
		0.000221,
		0.000225,
		0.000229,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000229,
		0.000225,
		0.000221,
		0.000215,
		0.000208,
		0.000199,
		0.00019,
		0.00018,
		0.000169,
		0.000157,
		0.000144,
		0.000131,
		0.000117,
		0.000103,
		8.9e-005,
		7.5e-005,
		6.2e-005,
		4.9e-005,
		3.7e-005,
		2.7e-005,
		1.7e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.6e-005,
		4.9e-005,
		9.8e-005,
		0.00016,
		0.000234,
		0.000317,
		0.000407,
		0.000503,
		0.000602,
		0.000703,
		0.000804,
		0.000904,
		0.001001,
		0.001095,
		0.001183,
		0.001266,
		0.001342,
		0.00141,
		0.001471,
		0.001523,
		0.001568,
		0.001604,
		0.001633,
		0.001654,
		0.001669,
		0.001679,
		0.001685,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001685,
		0.001679,
		0.001669,
		0.001654,
		0.001633,
		0.001604,
		0.001568,
		0.001523,
		0.001471,
		0.00141,
		0.001342,
		0.001266,
		0.001183,
		0.001095,
		0.001001,
		0.000904,
		0.000804,
		0.000703,
		0.000602,
		0.000503,
		0.000407,
		0.000317,
		0.000234,
		0.00016,
		9.8e-005,
		4.9e-005,
		1.6e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		2.1e-005,
		0.00011,
		0.000261,
		0.000466,
		0.000715,
		0.001001,
		0.001317,
		0.001654,
		0.002008,
		0.00237,
		0.002735,
		0.003097,
		0.003451,
		0.003792,
		0.004116,
		0.004419,
		0.004698,
		0.004951,
		0.005175,
		0.005369,
		0.005534,
		0.005669,
		0.005776,
		0.005856,
		0.005913,
		0.00595,
		0.005971,
		0.005981,
		0.005984,
		0.005985,
		0.005985,
		0.005984,
		0.005981,
		0.005971,
		0.00595,
		0.005913,
		0.005856,
		0.005776,
		0.005669,
		0.005534,
		0.005369,
		0.005175,
		0.004951,
		0.004698,
		0.004419,
		0.004116,
		0.003792,
		0.003451,
		0.003097,
		0.002735,
		0.00237,
		0.002008,
		0.001654,
		0.001317,
		0.001001,
		0.000715,
		0.000466,
		0.000261,
		0.00011,
		2.1e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		0.000149,
		0.000464,
		0.000927,
		0.001517,
		0.002212,
		0.002991,
		0.003836,
		0.004727,
		0.005648,
		0.00658,
		0.007511,
		0.008424,
		0.009307,
		0.010148,
		0.010938,
		0.011667,
		0.012329,
		0.012918,
		0.01343,
		0.013864,
		0.014221,
		0.014503,
		0.014716,
		0.014867,
		0.014965,
		0.015022,
		0.015049,
		0.015058,
		0.015059,
		0.015059,
		0.015058,
		0.015049,
		0.015022,
		0.014965,
		0.014867,
		0.014716,
		0.014503,
		0.014221,
		0.013864,
		0.01343,
		0.012918,
		0.012329,
		0.011667,
		0.010938,
		0.010148,
		0.009307,
		0.008424,
		0.007511,
		0.00658,
		0.005648,
		0.004727,
		0.003836,
		0.002991,
		0.002212,
		0.001517,
		0.000927,
		0.000464,
		0.000149,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		0.000116,
		0.000608,
		0.001437,
		0.002555,
		0.003915,
		0.005473,
		0.007186,
		0.009012,
		0.010914,
		0.012854,
		0.014799,
		0.016717,
		0.018579,
		0.02036,
		0.022037,
		0.023589,
		0.025001,
		0.026261,
		0.027359,
		0.028292,
		0.029061,
		0.029671,
		0.030132,
		0.030459,
		0.030672,
		0.030795,
		0.030854,
		0.030874,
		0.030877,
		0.030877,
		0.030874,
		0.030854,
		0.030795,
		0.030672,
		0.030459,
		0.030132,
		0.029671,
		0.029061,
		0.028292,
		0.027359,
		0.026261,
		0.025001,
		0.023589,
		0.022037,
		0.02036,
		0.018579,
		0.016717,
		0.014799,
		0.012854,
		0.010914,
		0.009012,
		0.007186,
		0.005473,
		0.003915,
		0.002555,
		0.001437,
		0.000608,
		0.000116,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		2.4e-005,
		0.000586,
		0.001816,
		0.003623,
		0.005917,
		0.008611,
		0.011623,
		0.014875,
		0.018292,
		0.021804,
		0.025346,
		0.028856,
		0.03228,
		0.035565,
		0.038668,
		0.041551,
		0.044179,
		0.04653,
		0.048585,
		0.050336,
		0.051781,
		0.052929,
		0.053799,
		0.054418,
		0.054822,
		0.055057,
		0.055169,
		0.055207,
		0.055213,
		0.055213,
		0.055207,
		0.055169,
		0.055057,
		0.054822,
		0.054418,
		0.053799,
		0.052929,
		0.051781,
		0.050336,
		0.048585,
		0.04653,
		0.044179,
		0.041551,
		0.038668,
		0.035565,
		0.03228,
		0.028856,
		0.025346,
		0.021804,
		0.018292,
		0.014875,
		0.011623,
		0.008611,
		0.005917,
		0.003623,
		0.001816,
		0.000586,
		2.4e-005,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		0.000361,
		0.001882,
		0.004443,
		0.007888,
		0.012067,
		0.016836,
		0.022059,
		0.027607,
		0.033356,
		0.039191,
		0.045007,
		0.050704,
		0.056194,
		0.061398,
		0.066246,
		0.070681,
		0.074658,
		0.078143,
		0.081119,
		0.083582,
		0.085543,
		0.087033,
		0.088095,
		0.088791,
		0.089194,
		0.089387,
		0.089454,
		0.089464,
		0.089464,
		0.089454,
		0.089387,
		0.089194,
		0.088791,
		0.088095,
		0.087033,
		0.085543,
		0.083582,
		0.081119,
		0.078143,
		0.074658,
		0.070681,
		0.066246,
		0.061398,
		0.056194,
		0.050704,
		0.045007,
		0.039191,
		0.033356,
		0.027607,
		0.022059,
		0.016836,
		0.012067,
		0.007888,
		0.004443,
		0.001882,
		0.000361,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		6.4e-005,
		0.001528,
		0.004732,
		0.009423,
		0.01536,
		0.02231,
		0.030053,
		0.038377,
		0.047085,
		0.055989,
		0.064916,
		0.073706,
		0.082212,
		0.090306,
		0.097872,
		0.104814,
		0.111057,
		0.116543,
		0.121239,
		0.125134,
		0.128245,
		0.130612,
		0.132304,
		0.133415,
		0.134061,
		0.134371,
		0.134478,
		0.134495,
		0.134495,
		0.134478,
		0.134371,
		0.134061,
		0.133415,
		0.132304,
		0.130612,
		0.128245,
		0.125134,
		0.121239,
		0.116543,
		0.111057,
		0.104814,
		0.097872,
		0.090306,
		0.082212,
		0.073706,
		0.064916,
		0.055989,
		0.047085,
		0.038377,
		0.030053,
		0.02231,
		0.01536,
		0.009423,
		0.004732,
		0.001528,
		6.4e-005,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		0.000826,
		0.004295,
		0.010121,
		0.017935,
		0.027382,
		0.038125,
		0.049843,
		0.062232,
		0.075006,
		0.087897,
		0.10066,
		0.11307,
		0.124925,
		0.136047,
		0.146287,
		0.155521,
		0.163659,
		0.170644,
		0.176453,
		0.181102,
		0.18465,
		0.187192,
		0.188866,
		0.18984,
		0.19031,
		0.190471,
		0.190498,
		0.190498,
		0.190471,
		0.19031,
		0.18984,
		0.188866,
		0.187192,
		0.18465,
		0.181102,
		0.176453,
		0.170644,
		0.163659,
		0.155521,
		0.146287,
		0.136047,
		0.124925,
		0.11307,
		0.10066,
		0.087897,
		0.075006,
		0.062232,
		0.049843,
		0.038125,
		0.027382,
		0.017935,
		0.010121,
		0.004295,
		0.000826,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000131,
		0.003134,
		0.009688,
		0.019257,
		0.031327,
		0.045407,
		0.061029,
		0.077749,
		0.095151,
		0.112844,
		0.130466,
		0.147689,
		0.164213,
		0.179776,
		0.194153,
		0.207159,
		0.218656,
		0.228551,
		0.236802,
		0.243424,
		0.248488,
		0.252127,
		0.254529,
		0.255931,
		0.256608,
		0.256841,
		0.25688,
		0.25688,
		0.256841,
		0.256608,
		0.255931,
		0.254529,
		0.252127,
		0.248488,
		0.243424,
		0.236802,
		0.228551,
		0.218656,
		0.207159,
		0.194153,
		0.179776,
		0.164213,
		0.147689,
		0.130466,
		0.112844,
		0.095151,
		0.077749,
		0.061029,
		0.045407,
		0.031327,
		0.019257,
		0.009688,
		0.003134,
		0.000131,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		0.00155,
		0.008047,
		0.018926,
		0.033468,
		0.050989,
		0.070835,
		0.092384,
		0.115052,
		0.138292,
		0.161595,
		0.184495,
		0.206571,
		0.227449,
		0.246807,
		0.264379,
		0.279959,
		0.293406,
		0.304652,
		0.313702,
		0.320642,
		0.325642,
		0.32895,
		0.330886,
		0.331824,
		0.332147,
		0.332201,
		0.332201,
		0.332147,
		0.331824,
		0.330886,
		0.32895,
		0.325642,
		0.320642,
		0.313702,
		0.304652,
		0.293406,
		0.279959,
		0.264379,
		0.246807,
		0.227449,
		0.206571,
		0.184495,
		0.161595,
		0.138292,
		0.115052,
		0.092384,
		0.070835,
		0.050989,
		0.033468,
		0.018926,
		0.008047,
		0.00155,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.000228,
		0.005442,
		0.016787,
		0.0333,
		0.054055,
		0.07817,
		0.104807,
		0.133176,
		0.162535,
		0.192194,
		0.221518,
		0.249933,
		0.276925,
		0.302051,
		0.324939,
		0.3453,
		0.362928,
		0.377712,
		0.389643,
		0.398819,
		0.405447,
		0.409844,
		0.412426,
		0.413678,
		0.414112,
		0.414183,
		0.414183,
		0.414112,
		0.413678,
		0.412426,
		0.409844,
		0.405447,
		0.398819,
		0.389643,
		0.377712,
		0.362928,
		0.3453,
		0.324939,
		0.302051,
		0.276925,
		0.249933,
		0.221518,
		0.192194,
		0.162535,
		0.133176,
		0.104807,
		0.07817,
		0.054055,
		0.0333,
		0.016787,
		0.005442,
		0.000228,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		0.002517,
		0.013042,
		0.03061,
		0.054013,
		0.082099,
		0.11377,
		0.14799,
		0.183787,
		0.220253,
		0.256552,
		0.291922,
		0.325684,
		0.357244,
		0.386103,
		0.411863,
		0.434238,
		0.453061,
		0.468297,
		0.480048,
		0.48856,
		0.494223,
		0.497557,
		0.499179,
		0.499742,
		0.499835,
		0.499835,
		0.499742,
		0.499179,
		0.497557,
		0.494223,
		0.48856,
		0.480048,
		0.468297,
		0.453061,
		0.434238,
		0.411863,
		0.386103,
		0.357244,
		0.325684,
		0.291922,
		0.256552,
		0.220253,
		0.183787,
		0.14799,
		0.11377,
		0.082099,
		0.054013,
		0.03061,
		0.013042,
		0.002517,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.000349,
		0.008309,
		0.02558,
		0.050629,
		0.08199,
		0.118271,
		0.158149,
		0.200385,
		0.243819,
		0.287379,
		0.330088,
		0.371068,
		0.409549,
		0.444878,
		0.47653,
		0.504116,
		0.527398,
		0.546301,
		0.560924,
		0.571549,
		0.578638,
		0.582824,
		0.584866,
		0.585577,
		0.585694,
		0.585694,
		0.585577,
		0.584866,
		0.582824,
		0.578638,
		0.571549,
		0.560924,
		0.546301,
		0.527398,
		0.504116,
		0.47653,
		0.444878,
		0.409549,
		0.371068,
		0.330088,
		0.287379,
		0.243819,
		0.200385,
		0.158149,
		0.118271,
		0.08199,
		0.050629,
		0.02558,
		0.008309,
		0.000349,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		0.003642,
		0.018834,
		0.044105,
		0.07764,
		0.11771,
		0.162676,
		0.210995,
		0.261224,
		0.312024,
		0.362172,
		0.410564,
		0.456229,
		0.498335,
		0.536205,
		0.569328,
		0.597379,
		0.620228,
		0.637959,
		0.650881,
		0.65953,
		0.664653,
		0.667159,
		0.668034,
		0.668179,
		0.668179,
		0.668034,
		0.667159,
		0.664653,
		0.65953,
		0.650881,
		0.637959,
		0.620228,
		0.597379,
		0.569328,
		0.536205,
		0.498335,
		0.456229,
		0.410564,
		0.362172,
		0.312024,
		0.261224,
		0.210995,
		0.162676,
		0.11771,
		0.07764,
		0.044105,
		0.018834,
		0.003642,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.000482,
		0.011432,
		0.035114,
		0.069331,
		0.111989,
		0.161098,
		0.214785,
		0.271292,
		0.328987,
		0.386375,
		0.442101,
		0.494965,
		0.543936,
		0.588164,
		0.626996,
		0.659998,
		0.686971,
		0.707972,
		0.723327,
		0.733637,
		0.739762,
		0.742769,
		0.743821,
		0.743996,
		0.743996,
		0.743821,
		0.742769,
		0.739762,
		0.733637,
		0.723327,
		0.707972,
		0.686971,
		0.659998,
		0.626996,
		0.588164,
		0.543936,
		0.494965,
		0.442101,
		0.386375,
		0.328987,
		0.271292,
		0.214785,
		0.161098,
		0.111989,
		0.069331,
		0.035114,
		0.011432,
		0.000482,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		0.004798,
		0.024756,
		0.057832,
		0.101537,
		0.15351,
		0.21152,
		0.273471,
		0.337415,
		0.40156,
		0.464277,
		0.524119,
		0.579832,
		0.630373,
		0.674927,
		0.712934,
		0.744109,
		0.768466,
		0.786334,
		0.798372,
		0.805547,
		0.80908,
		0.810321,
		0.810528,
		0.810528,
		0.810321,
		0.80908,
		0.805547,
		0.798372,
		0.786334,
		0.768466,
		0.744109,
		0.712934,
		0.674927,
		0.630373,
		0.579832,
		0.524119,
		0.464277,
		0.40156,
		0.337415,
		0.273471,
		0.21152,
		0.15351,
		0.101537,
		0.057832,
		0.024756,
		0.004798,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.000611,
		0.014463,
		0.044312,
		0.087259,
		0.140548,
		0.201568,
		0.267868,
		0.337163,
		0.407343,
		0.476489,
		0.542884,
		0.605034,
		0.661683,
		0.711837,
		0.754792,
		0.790158,
		0.817888,
		0.838304,
		0.852106,
		0.86036,
		0.864439,
		0.865876,
		0.866116,
		0.866116,
		0.865876,
		0.864439,
		0.86036,
		0.852106,
		0.838304,
		0.817888,
		0.790158,
		0.754792,
		0.711837,
		0.661683,
		0.605034,
		0.542884,
		0.476489,
		0.407343,
		0.337163,
		0.267868,
		0.201568,
		0.140548,
		0.087259,
		0.044312,
		0.014463,
		0.000611,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		0.005873,
		0.030225,
		0.070417,
		0.123278,
		0.185806,
		0.255173,
		0.32874,
		0.404065,
		0.478918,
		0.551297,
		0.619449,
		0.681888,
		0.737423,
		0.785187,
		0.824668,
		0.855742,
		0.878705,
		0.894285,
		0.903638,
		0.908275,
		0.909914,
		0.910189,
		0.910189,
		0.909914,
		0.908275,
		0.903638,
		0.894285,
		0.878705,
		0.855742,
		0.824668,
		0.785187,
		0.737423,
		0.681888,
		0.619449,
		0.551297,
		0.478918,
		0.404065,
		0.32874,
		0.255173,
		0.185806,
		0.123278,
		0.070417,
		0.030225,
		0.005873,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.000727,
		0.017177,
		0.052485,
		0.103053,
		0.165466,
		0.236507,
		0.313165,
		0.392648,
		0.472399,
		0.550115,
		0.623765,
		0.691617,
		0.752265,
		0.80466,
		0.84815,
		0.882517,
		0.908012,
		0.925377,
		0.935839,
		0.941046,
		0.942893,
		0.943204,
		0.943204,
		0.942893,
		0.941046,
		0.935839,
		0.925377,
		0.908012,
		0.882517,
		0.84815,
		0.80466,
		0.752265,
		0.691617,
		0.623765,
		0.550115,
		0.472399,
		0.392648,
		0.313165,
		0.236507,
		0.165466,
		0.103053,
		0.052485,
		0.017177,
		0.000727,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		0.006823,
		0.035019,
		0.081344,
		0.141955,
		0.213225,
		0.291752,
		0.374377,
		0.458198,
		0.540591,
		0.619229,
		0.692115,
		0.757607,
		0.814459,
		0.861858,
		0.899472,
		0.92749,
		0.946649,
		0.958238,
		0.964028,
		0.96609,
		0.966438,
		0.966438,
		0.96609,
		0.964028,
		0.958238,
		0.946649,
		0.92749,
		0.899472,
		0.861858,
		0.814459,
		0.757607,
		0.692115,
		0.619229,
		0.540591,
		0.458198,
		0.374377,
		0.291752,
		0.213225,
		0.141955,
		0.081344,
		0.035019,
		0.006823,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.000831,
		0.019574,
		0.059631,
		0.116705,
		0.186735,
		0.265908,
		0.350675,
		0.437764,
		0.524209,
		0.607367,
		0.684951,
		0.755064,
		0.81624,
		0.867484,
		0.908331,
		0.938887,
		0.95987,
		0.972616,
		0.979009,
		0.981295,
		0.981682,
		0.981682,
		0.981295,
		0.979009,
		0.972616,
		0.95987,
		0.938887,
		0.908331,
		0.867484,
		0.81624,
		0.755064,
		0.684951,
		0.607367,
		0.524209,
		0.437764,
		0.350675,
		0.265908,
		0.186735,
		0.116705,
		0.059631,
		0.019574,
		0.000831,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.007687,
		0.039335,
		0.091069,
		0.158365,
		0.236966,
		0.322902,
		0.412507,
		0.502441,
		0.589719,
		0.671739,
		0.746323,
		0.811757,
		0.866844,
		0.910961,
		0.944112,
		0.966979,
		0.980929,
		0.987957,
		0.990479,
		0.990908,
		0.990908,
		0.990479,
		0.987957,
		0.980929,
		0.966979,
		0.944112,
		0.910961,
		0.866844,
		0.811757,
		0.746323,
		0.671739,
		0.589719,
		0.502441,
		0.412507,
		0.322902,
		0.236966,
		0.158365,
		0.091069,
		0.039335,
		0.007687,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.00093,
		0.021838,
		0.066304,
		0.129299,
		0.206085,
		0.292232,
		0.38364,
		0.476563,
		0.567636,
		0.653913,
		0.732899,
		0.802607,
		0.861608,
		0.909096,
		0.944951,
		0.969798,
		0.985025,
		0.992731,
		0.995509,
		0.995983,
		0.995983,
		0.995509,
		0.992731,
		0.985025,
		0.969798,
		0.944951,
		0.909096,
		0.861608,
		0.802607,
		0.732899,
		0.653913,
		0.567636,
		0.476563,
		0.38364,
		0.292232,
		0.206085,
		0.129299,
		0.066304,
		0.021838,
		0.00093,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.008551,
		0.043608,
		0.100594,
		0.174237,
		0.259602,
		0.352107,
		0.447554,
		0.542159,
		0.632584,
		0.715985,
		0.790063,
		0.853125,
		0.904151,
		0.942872,
		0.969837,
		0.986442,
		0.994885,
		0.997943,
		0.998467,
		0.998467,
		0.997943,
		0.994885,
		0.986442,
		0.969837,
		0.942872,
		0.904151,
		0.853125,
		0.790063,
		0.715985,
		0.632584,
		0.542159,
		0.447554,
		0.352107,
		0.259602,
		0.174237,
		0.100594,
		0.043608,
		0.008551,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.001034,
		0.024212,
		0.07324,
		0.142252,
		0.225742,
		0.31859,
		0.416091,
		0.513985,
		0.608498,
		0.696388,
		0.775,
		0.842337,
		0.897131,
		0.938936,
		0.968199,
		0.986312,
		0.995567,
		0.998935,
		0.999515,
		0.999515,
		0.998935,
		0.995567,
		0.986312,
		0.968199,
		0.938936,
		0.897131,
		0.842337,
		0.775,
		0.696388,
		0.608498,
		0.513985,
		0.416091,
		0.31859,
		0.225742,
		0.142252,
		0.07324,
		0.024212,
		0.001034,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.009502,
		0.048276,
		0.110907,
		0.191247,
		0.283568,
		0.38259,
		0.483514,
		0.582068,
		0.674555,
		0.757914,
		0.829796,
		0.888645,
		0.9338,
		0.965582,
		0.98536,
		0.995519,
		0.999235,
		0.999877,
		0.999877,
		0.999235,
		0.995519,
		0.98536,
		0.965582,
		0.9338,
		0.888645,
		0.829796,
		0.757914,
		0.674555,
		0.582068,
		0.483514,
		0.38259,
		0.283568,
		0.191247,
		0.110907,
		0.048276,
		0.009502,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.001154,
		0.026901,
		0.081038,
		0.156684,
		0.247418,
		0.347303,
		0.450927,
		0.553448,
		0.650647,
		0.738994,
		0.815734,
		0.878971,
		0.927789,
		0.962349,
		0.983977,
		0.99515,
		0.999258,
		0.999972,
		0.999972,
		0.999258,
		0.99515,
		0.983977,
		0.962349,
		0.927789,
		0.878971,
		0.815734,
		0.738994,
		0.650647,
		0.553448,
		0.450927,
		0.347303,
		0.247418,
		0.156684,
		0.081038,
		0.026901,
		0.001154,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310414,
		0.416289,
		0.522637,
		0.624639,
		0.718225,
		0.800162,
		0.86816,
		0.920994,
		0.95863,
		0.982326,
		0.994638,
		0.999192,
		0.999987,
		0.999987,
		0.999192,
		0.994638,
		0.982326,
		0.95863,
		0.920994,
		0.86816,
		0.800162,
		0.718225,
		0.624639,
		0.522637,
		0.416289,
		0.310414,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.013475,
		0.067471,
		0.152563,
		0.258527,
		0.376031,
		0.496716,
		0.613292,
		0.71967,
		0.811114,
		0.884434,
		0.938217,
		0.973047,
		0.991651,
		0.998717,
		0.999981,
		0.999981,
		0.998717,
		0.991651,
		0.973047,
		0.938217,
		0.884434,
		0.811114,
		0.71967,
		0.613292,
		0.496716,
		0.376031,
		0.258527,
		0.152563,
		0.067471,
		0.013475,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.001665,
		0.038254,
		0.113386,
		0.215353,
		0.333432,
		0.457972,
		0.580504,
		0.693878,
		0.792437,
		0.87223,
		0.931276,
		0.969836,
		0.990599,
		0.998546,
		0.999979,
		0.999979,
		0.998546,
		0.990599,
		0.969836,
		0.931276,
		0.87223,
		0.792437,
		0.693878,
		0.580504,
		0.457972,
		0.333432,
		0.215353,
		0.113386,
		0.038254,
		0.001665,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.015366,
		0.076461,
		0.171689,
		0.28869,
		0.416289,
		0.544621,
		0.665277,
		0.771497,
		0.858415,
		0.92335,
		0.966137,
		0.989376,
		0.998347,
		0.999976,
		0.999976,
		0.998347,
		0.989376,
		0.966137,
		0.92335,
		0.858415,
		0.771497,
		0.665277,
		0.544621,
		0.416289,
		0.28869,
		0.171689,
		0.076461,
		0.015366,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.001913,
		0.043674,
		0.128533,
		0.242186,
		0.371655,
		0.505413,
		0.63355,
		0.747985,
		0.842738,
		0.914268,
		0.961859,
		0.987949,
		0.998111,
		0.999972,
		0.999972,
		0.998111,
		0.987949,
		0.961859,
		0.914268,
		0.842738,
		0.747985,
		0.63355,
		0.505413,
		0.371655,
		0.242186,
		0.128533,
		0.043674,
		0.001913,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.017685,
		0.08736,
		0.19457,
		0.324181,
		0.462682,
		0.598361,
		0.721547,
		0.824905,
		0.903828,
		0.956891,
		0.986276,
		0.997833,
		0.999968,
		0.999968,
		0.997833,
		0.986276,
		0.956891,
		0.903828,
		0.824905,
		0.721547,
		0.598361,
		0.462682,
		0.324181,
		0.19457,
		0.08736,
		0.017685,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.002221,
		0.05033,
		0.146884,
		0.27417,
		0.416289,
		0.559361,
		0.691782,
		0.804566,
		0.891784,
		0.951098,
		0.984305,
		0.997501,
		0.999962,
		0.999962,
		0.997501,
		0.984305,
		0.951098,
		0.891784,
		0.804566,
		0.691782,
		0.559361,
		0.416289,
		0.27417,
		0.146884,
		0.05033,
		0.002221,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.02057,
		0.100747,
		0.222223,
		0.366214,
		0.516206,
		0.658245,
		0.781314,
		0.877839,
		0.944313,
		0.981968,
		0.997104,
		0.999956,
		0.999956,
		0.997104,
		0.981968,
		0.944313,
		0.877839,
		0.781314,
		0.658245,
		0.516206,
		0.366214,
		0.222223,
		0.100747,
		0.02057,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.00261,
		0.058623,
		0.169394,
		0.312626,
		0.468589,
		0.620445,
		0.754672,
		0.861636,
		0.936326,
		0.979184,
		0.996625,
		0.999948,
		0.999948,
		0.996625,
		0.979184,
		0.936326,
		0.861636,
		0.754672,
		0.620445,
		0.468589,
		0.312626,
		0.169394,
		0.058623,
		0.00261,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.024223,
		0.117431,
		0.256014,
		0.416289,
		0.577857,
		0.724081,
		0.842738,
		0.926878,
		0.975845,
		0.996042,
		0.999939,
		0.999939,
		0.996042,
		0.975845,
		0.926878,
		0.842738,
		0.724081,
		0.577857,
		0.416289,
		0.256014,
		0.117431,
		0.024223,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.003111,
		0.069138,
		0.197385,
		0.359265,
		0.529947,
		0.688896,
		0.820615,
		0.915644,
		0.971816,
		0.995329,
		0.999927,
		0.999927,
		0.995329,
		0.971816,
		0.915644,
		0.820615,
		0.688896,
		0.529947,
		0.359265,
		0.197385,
		0.069138,
		0.003111,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.028941,
		0.138573,
		0.297795,
		0.476215,
		0.64838,
		0.794618,
		0.902212,
		0.96692,
		0.994448,
		0.999912,
		0.999912,
		0.994448,
		0.96692,
		0.902212,
		0.794618,
		0.64838,
		0.476215,
		0.297795,
		0.138573,
		0.028941,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.003771,
		0.082738,
		0.232725,
		0.416289,
		0.601712,
		0.763962,
		0.886059,
		0.960924,
		0.99335,
		0.999893,
		0.999893,
		0.99335,
		0.960924,
		0.886059,
		0.763962,
		0.601712,
		0.416289,
		0.232725,
		0.082738,
		0.003771,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.035182,
		0.165879,
		0.350084,
		0.548017,
		0.727694,
		0.86652,
		0.953525,
		0.991969,
		0.999869,
		0.999869,
		0.991969,
		0.953525,
		0.86652,
		0.727694,
		0.548017,
		0.350084,
		0.165879,
		0.035182,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.004665,
		0.100747,
		0.278096,
		0.486437,
		0.684672,
		0.842738,
		0.944313,
		0.990212,
		0.999838,
		0.999838,
		0.990212,
		0.944313,
		0.842738,
		0.684672,
		0.486437,
		0.278096,
		0.100747,
		0.004665,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.043674,
		0.201932,
		0.416289,
		0.63355,
		0.813614,
		0.932739,
		0.987949,
		0.999797,
		0.999797,
		0.987949,
		0.932739,
		0.813614,
		0.63355,
		0.416289,
		0.201932,
		0.043674,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.00592,
		0.125272,
		0.337397,
		0.572806,
		0.777728,
		0.918052,
		0.984999,
		0.999741,
		0.999741,
		0.984999,
		0.918052,
		0.777728,
		0.572806,
		0.337397,
		0.125272,
		0.00592,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.055596,
		0.250527,
		0.500403,
		0.732617,
		0.898429,
		0.980234,
		0.998789,
		0.998789,
		0.980234,
		0.898429,
		0.732617,
		0.500403,
		0.250527,
		0.055596,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.007563,
		0.155812,
		0.405858,
		0.660911,
		0.852871,
		0.951392,
		0.974514,
		0.974514,
		0.951392,
		0.852871,
		0.660911,
		0.405858,
		0.155812,
		0.007563,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.061533,
		0.267665,
		0.511233,
		0.707694,
		0.813438,
		0.839261,
		0.839261,
		0.813438,
		0.707694,
		0.511233,
		0.267665,
		0.061533,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.005759,
		0.1143,
		0.283774,
		0.434461,
		0.520504,
		0.542528,
		0.542528,
		0.520504,
		0.434461,
		0.283774,
		0.1143,
		0.005759,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.022412,
		0.092608,
		0.165276,
		0.210073,
		0.222209,
		0.222209,
		0.210073,
		0.165276,
		0.092608,
		0.022412,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.000644,
		0.012091,
		0.027863,
		0.038672,
		0.041813,
		0.041813,
		0.038672,
		0.027863,
		0.012091,
		0.000644,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		0.000106,
		0.000402,
		0.000642,
		0.000718,
		0.000718,
		0.000642,
		0.000402,
		0.000106,
		0.000106,
		0.000402,
		0.000642,
		0.000718,
		0.000718,
		0.000642,
		0.000402,
		0.000106,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		0.000644,
		0.012091,
		0.027863,
		0.038672,
		0.041813,
		0.041813,
		0.038672,
		0.027863,
		0.012091,
		0.000644,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.022412,
		0.092608,
		0.165276,
		0.210073,
		0.222209,
		0.222209,
		0.210073,
		0.165276,
		0.092608,
		0.022412,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.005759,
		0.1143,
		0.283774,
		0.434461,
		0.520504,
		0.542528,
		0.542528,
		0.520504,
		0.434461,
		0.283774,
		0.1143,
		0.005759,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.061533,
		0.267665,
		0.511233,
		0.707694,
		0.813438,
		0.839261,
		0.839261,
		0.813438,
		0.707694,
		0.511233,
		0.267665,
		0.061533,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.007563,
		0.155812,
		0.405858,
		0.660911,
		0.852871,
		0.951392,
		0.974514,
		0.974514,
		0.951392,
		0.852871,
		0.660911,
		0.405858,
		0.155812,
		0.007563,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.055596,
		0.250527,
		0.500403,
		0.732617,
		0.898429,
		0.980234,
		0.998789,
		0.998789,
		0.980234,
		0.898429,
		0.732617,
		0.500403,
		0.250527,
		0.055596,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.00592,
		0.125272,
		0.337397,
		0.572806,
		0.777728,
		0.918052,
		0.984999,
		0.999741,
		0.999741,
		0.984999,
		0.918052,
		0.777728,
		0.572806,
		0.337397,
		0.125272,
		0.00592,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.043674,
		0.201932,
		0.416289,
		0.63355,
		0.813614,
		0.932739,
		0.987949,
		0.999797,
		0.999797,
		0.987949,
		0.932739,
		0.813614,
		0.63355,
		0.416289,
		0.201932,
		0.043674,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.004665,
		0.100747,
		0.278096,
		0.486437,
		0.684672,
		0.842738,
		0.944313,
		0.990212,
		0.999838,
		0.999838,
		0.990212,
		0.944313,
		0.842738,
		0.684672,
		0.486437,
		0.278096,
		0.100747,
		0.004665,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.035182,
		0.165879,
		0.350084,
		0.548017,
		0.727694,
		0.86652,
		0.953525,
		0.991969,
		0.999869,
		0.999869,
		0.991969,
		0.953525,
		0.86652,
		0.727694,
		0.548017,
		0.350084,
		0.165879,
		0.035182,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.003771,
		0.082738,
		0.232725,
		0.416289,
		0.601712,
		0.763962,
		0.886059,
		0.960924,
		0.99335,
		0.999893,
		0.999893,
		0.99335,
		0.960924,
		0.886059,
		0.763962,
		0.601712,
		0.416289,
		0.232725,
		0.082738,
		0.003771,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.028941,
		0.138573,
		0.297795,
		0.476215,
		0.64838,
		0.794618,
		0.902212,
		0.96692,
		0.994448,
		0.999912,
		0.999912,
		0.994448,
		0.96692,
		0.902212,
		0.794618,
		0.64838,
		0.476215,
		0.297795,
		0.138573,
		0.028941,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.003111,
		0.069138,
		0.197385,
		0.359265,
		0.529947,
		0.688896,
		0.820615,
		0.915644,
		0.971816,
		0.995329,
		0.999927,
		0.999927,
		0.995329,
		0.971816,
		0.915644,
		0.820615,
		0.688896,
		0.529947,
		0.359265,
		0.197385,
		0.069138,
		0.003111,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.024223,
		0.117431,
		0.256014,
		0.416289,
		0.577857,
		0.724081,
		0.842738,
		0.926878,
		0.975845,
		0.996042,
		0.999939,
		0.999939,
		0.996042,
		0.975845,
		0.926878,
		0.842738,
		0.724081,
		0.577857,
		0.416289,
		0.256014,
		0.117431,
		0.024223,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.00261,
		0.058623,
		0.169394,
		0.312626,
		0.468589,
		0.620445,
		0.754672,
		0.861636,
		0.936326,
		0.979184,
		0.996625,
		0.999948,
		0.999948,
		0.996625,
		0.979184,
		0.936326,
		0.861636,
		0.754672,
		0.620445,
		0.468589,
		0.312626,
		0.169394,
		0.058623,
		0.00261,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.02057,
		0.100747,
		0.222223,
		0.366214,
		0.516206,
		0.658245,
		0.781314,
		0.877839,
		0.944313,
		0.981968,
		0.997104,
		0.999956,
		0.999956,
		0.997104,
		0.981968,
		0.944313,
		0.877839,
		0.781314,
		0.658245,
		0.516206,
		0.366214,
		0.222223,
		0.100747,
		0.02057,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.002221,
		0.05033,
		0.146884,
		0.27417,
		0.416289,
		0.559361,
		0.691782,
		0.804566,
		0.891784,
		0.951098,
		0.984305,
		0.997501,
		0.999962,
		0.999962,
		0.997501,
		0.984305,
		0.951098,
		0.891784,
		0.804566,
		0.691782,
		0.559361,
		0.416289,
		0.27417,
		0.146884,
		0.05033,
		0.002221,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.017685,
		0.08736,
		0.19457,
		0.324181,
		0.462682,
		0.598361,
		0.721547,
		0.824905,
		0.903828,
		0.956891,
		0.986276,
		0.997833,
		0.999968,
		0.999968,
		0.997833,
		0.986276,
		0.956891,
		0.903828,
		0.824905,
		0.721547,
		0.598361,
		0.462682,
		0.324181,
		0.19457,
		0.08736,
		0.017685,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.001913,
		0.043674,
		0.128533,
		0.242186,
		0.371655,
		0.505413,
		0.63355,
		0.747985,
		0.842738,
		0.914268,
		0.961859,
		0.987949,
		0.998111,
		0.999972,
		0.999972,
		0.998111,
		0.987949,
		0.961859,
		0.914268,
		0.842738,
		0.747985,
		0.63355,
		0.505413,
		0.371655,
		0.242186,
		0.128533,
		0.043674,
		0.001913,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.015366,
		0.076461,
		0.171689,
		0.28869,
		0.416289,
		0.544621,
		0.665277,
		0.771497,
		0.858415,
		0.92335,
		0.966137,
		0.989376,
		0.998347,
		0.999976,
		0.999976,
		0.998347,
		0.989376,
		0.966137,
		0.92335,
		0.858415,
		0.771497,
		0.665277,
		0.544621,
		0.416289,
		0.28869,
		0.171689,
		0.076461,
		0.015366,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.001665,
		0.038254,
		0.113386,
		0.215353,
		0.333432,
		0.457972,
		0.580504,
		0.693878,
		0.792437,
		0.87223,
		0.931276,
		0.969836,
		0.990599,
		0.998546,
		0.999979,
		0.999979,
		0.998546,
		0.990599,
		0.969836,
		0.931276,
		0.87223,
		0.792437,
		0.693878,
		0.580504,
		0.457972,
		0.333432,
		0.215353,
		0.113386,
		0.038254,
		0.001665,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.013475,
		0.067471,
		0.152563,
		0.258527,
		0.376031,
		0.496716,
		0.613292,
		0.71967,
		0.811114,
		0.884434,
		0.938217,
		0.973047,
		0.991651,
		0.998717,
		0.999981,
		0.999981,
		0.998717,
		0.991651,
		0.973047,
		0.938217,
		0.884434,
		0.811114,
		0.71967,
		0.613292,
		0.496716,
		0.376031,
		0.258527,
		0.152563,
		0.067471,
		0.013475,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310414,
		0.416289,
		0.522637,
		0.624639,
		0.718225,
		0.800162,
		0.86816,
		0.920994,
		0.95863,
		0.982326,
		0.994638,
		0.999192,
		0.999987,
		0.999987,
		0.999192,
		0.994638,
		0.982326,
		0.95863,
		0.920994,
		0.86816,
		0.800162,
		0.718225,
		0.624639,
		0.522637,
		0.416289,
		0.310414,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.001154,
		0.026901,
		0.081038,
		0.156684,
		0.247418,
		0.347303,
		0.450927,
		0.553448,
		0.650647,
		0.738994,
		0.815734,
		0.878971,
		0.927789,
		0.962349,
		0.983977,
		0.99515,
		0.999258,
		0.999972,
		0.999972,
		0.999258,
		0.99515,
		0.983977,
		0.962349,
		0.927789,
		0.878971,
		0.815734,
		0.738994,
		0.650647,
		0.553448,
		0.450927,
		0.347303,
		0.247418,
		0.156684,
		0.081038,
		0.026901,
		0.001154,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.009502,
		0.048276,
		0.110907,
		0.191247,
		0.283568,
		0.38259,
		0.483514,
		0.582068,
		0.674555,
		0.757914,
		0.829796,
		0.888645,
		0.9338,
		0.965582,
		0.98536,
		0.995519,
		0.999235,
		0.999877,
		0.999877,
		0.999235,
		0.995519,
		0.98536,
		0.965582,
		0.9338,
		0.888645,
		0.829796,
		0.757914,
		0.674555,
		0.582068,
		0.483514,
		0.38259,
		0.283568,
		0.191247,
		0.110907,
		0.048276,
		0.009502,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.001034,
		0.024212,
		0.07324,
		0.142252,
		0.225742,
		0.31859,
		0.416091,
		0.513985,
		0.608498,
		0.696388,
		0.775,
		0.842337,
		0.897131,
		0.938936,
		0.968199,
		0.986312,
		0.995567,
		0.998935,
		0.999515,
		0.999515,
		0.998935,
		0.995567,
		0.986312,
		0.968199,
		0.938936,
		0.897131,
		0.842337,
		0.775,
		0.696388,
		0.608498,
		0.513985,
		0.416091,
		0.31859,
		0.225742,
		0.142252,
		0.07324,
		0.024212,
		0.001034,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.008551,
		0.043608,
		0.100594,
		0.174237,
		0.259602,
		0.352107,
		0.447554,
		0.542159,
		0.632584,
		0.715985,
		0.790063,
		0.853125,
		0.904151,
		0.942872,
		0.969837,
		0.986442,
		0.994885,
		0.997943,
		0.998467,
		0.998467,
		0.997943,
		0.994885,
		0.986442,
		0.969837,
		0.942872,
		0.904151,
		0.853125,
		0.790063,
		0.715985,
		0.632584,
		0.542159,
		0.447554,
		0.352107,
		0.259602,
		0.174237,
		0.100594,
		0.043608,
		0.008551,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.00093,
		0.021838,
		0.066304,
		0.129299,
		0.206085,
		0.292232,
		0.38364,
		0.476563,
		0.567636,
		0.653913,
		0.732899,
		0.802607,
		0.861608,
		0.909096,
		0.944951,
		0.969798,
		0.985025,
		0.992731,
		0.995509,
		0.995983,
		0.995983,
		0.995509,
		0.992731,
		0.985025,
		0.969798,
		0.944951,
		0.909096,
		0.861608,
		0.802607,
		0.732899,
		0.653913,
		0.567636,
		0.476563,
		0.38364,
		0.292232,
		0.206085,
		0.129299,
		0.066304,
		0.021838,
		0.00093,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.007687,
		0.039335,
		0.091069,
		0.158365,
		0.236966,
		0.322902,
		0.412507,
		0.502441,
		0.589719,
		0.671739,
		0.746323,
		0.811757,
		0.866844,
		0.910961,
		0.944112,
		0.966979,
		0.980929,
		0.987957,
		0.990479,
		0.990908,
		0.990908,
		0.990479,
		0.987957,
		0.980929,
		0.966979,
		0.944112,
		0.910961,
		0.866844,
		0.811757,
		0.746323,
		0.671739,
		0.589719,
		0.502441,
		0.412507,
		0.322902,
		0.236966,
		0.158365,
		0.091069,
		0.039335,
		0.007687,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.000831,
		0.019574,
		0.059631,
		0.116705,
		0.186735,
		0.265908,
		0.350675,
		0.437764,
		0.524209,
		0.607367,
		0.684951,
		0.755064,
		0.81624,
		0.867484,
		0.908331,
		0.938887,
		0.95987,
		0.972616,
		0.979009,
		0.981295,
		0.981682,
		0.981682,
		0.981295,
		0.979009,
		0.972616,
		0.95987,
		0.938887,
		0.908331,
		0.867484,
		0.81624,
		0.755064,
		0.684951,
		0.607367,
		0.524209,
		0.437764,
		0.350675,
		0.265908,
		0.186735,
		0.116705,
		0.059631,
		0.019574,
		0.000831,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.006823,
		0.035019,
		0.081344,
		0.141955,
		0.213225,
		0.291752,
		0.374377,
		0.458198,
		0.540591,
		0.619229,
		0.692115,
		0.757607,
		0.814459,
		0.861858,
		0.899472,
		0.92749,
		0.946649,
		0.958238,
		0.964028,
		0.96609,
		0.966438,
		0.966438,
		0.96609,
		0.964028,
		0.958238,
		0.946649,
		0.92749,
		0.899472,
		0.861858,
		0.814459,
		0.757607,
		0.692115,
		0.619229,
		0.540591,
		0.458198,
		0.374377,
		0.291752,
		0.213225,
		0.141955,
		0.081344,
		0.035019,
		0.006823,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		0.000727,
		0.017177,
		0.052485,
		0.103053,
		0.165466,
		0.236507,
		0.313165,
		0.392648,
		0.472399,
		0.550115,
		0.623765,
		0.691617,
		0.752265,
		0.80466,
		0.84815,
		0.882517,
		0.908012,
		0.925377,
		0.935839,
		0.941046,
		0.942893,
		0.943204,
		0.943204,
		0.942893,
		0.941046,
		0.935839,
		0.925377,
		0.908012,
		0.882517,
		0.84815,
		0.80466,
		0.752265,
		0.691617,
		0.623765,
		0.550115,
		0.472399,
		0.392648,
		0.313165,
		0.236507,
		0.165466,
		0.103053,
		0.052485,
		0.017177,
		0.000727,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.005873,
		0.030225,
		0.070417,
		0.123278,
		0.185806,
		0.255173,
		0.32874,
		0.404065,
		0.478918,
		0.551297,
		0.619449,
		0.681888,
		0.737423,
		0.785187,
		0.824668,
		0.855742,
		0.878705,
		0.894285,
		0.903638,
		0.908275,
		0.909914,
		0.910189,
		0.910189,
		0.909914,
		0.908275,
		0.903638,
		0.894285,
		0.878705,
		0.855742,
		0.824668,
		0.785187,
		0.737423,
		0.681888,
		0.619449,
		0.551297,
		0.478918,
		0.404065,
		0.32874,
		0.255173,
		0.185806,
		0.123278,
		0.070417,
		0.030225,
		0.005873,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		0.000611,
		0.014463,
		0.044312,
		0.087259,
		0.140548,
		0.201568,
		0.267868,
		0.337163,
		0.407343,
		0.476489,
		0.542884,
		0.605034,
		0.661683,
		0.711837,
		0.754792,
		0.790158,
		0.817888,
		0.838304,
		0.852106,
		0.86036,
		0.864439,
		0.865876,
		0.866116,
		0.866116,
		0.865876,
		0.864439,
		0.86036,
		0.852106,
		0.838304,
		0.817888,
		0.790158,
		0.754792,
		0.711837,
		0.661683,
		0.605034,
		0.542884,
		0.476489,
		0.407343,
		0.337163,
		0.267868,
		0.201568,
		0.140548,
		0.087259,
		0.044312,
		0.014463,
		0.000611,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.004798,
		0.024756,
		0.057832,
		0.101537,
		0.15351,
		0.21152,
		0.273471,
		0.337415,
		0.40156,
		0.464277,
		0.524119,
		0.579832,
		0.630373,
		0.674927,
		0.712934,
		0.744109,
		0.768466,
		0.786334,
		0.798372,
		0.805547,
		0.80908,
		0.810321,
		0.810528,
		0.810528,
		0.810321,
		0.80908,
		0.805547,
		0.798372,
		0.786334,
		0.768466,
		0.744109,
		0.712934,
		0.674927,
		0.630373,
		0.579832,
		0.524119,
		0.464277,
		0.40156,
		0.337415,
		0.273471,
		0.21152,
		0.15351,
		0.101537,
		0.057832,
		0.024756,
		0.004798,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		0.000482,
		0.011432,
		0.035114,
		0.069331,
		0.111989,
		0.161098,
		0.214785,
		0.271292,
		0.328987,
		0.386375,
		0.442101,
		0.494965,
		0.543936,
		0.588164,
		0.626996,
		0.659998,
		0.686971,
		0.707972,
		0.723327,
		0.733637,
		0.739762,
		0.742769,
		0.743821,
		0.743996,
		0.743996,
		0.743821,
		0.742769,
		0.739762,
		0.733637,
		0.723327,
		0.707972,
		0.686971,
		0.659998,
		0.626996,
		0.588164,
		0.543936,
		0.494965,
		0.442101,
		0.386375,
		0.328987,
		0.271292,
		0.214785,
		0.161098,
		0.111989,
		0.069331,
		0.035114,
		0.011432,
		0.000482,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.003642,
		0.018834,
		0.044105,
		0.07764,
		0.11771,
		0.162676,
		0.210995,
		0.261224,
		0.312024,
		0.362172,
		0.410564,
		0.456229,
		0.498335,
		0.536205,
		0.569328,
		0.597379,
		0.620228,
		0.637959,
		0.650881,
		0.65953,
		0.664653,
		0.667159,
		0.668034,
		0.668179,
		0.668179,
		0.668034,
		0.667159,
		0.664653,
		0.65953,
		0.650881,
		0.637959,
		0.620228,
		0.597379,
		0.569328,
		0.536205,
		0.498335,
		0.456229,
		0.410564,
		0.362172,
		0.312024,
		0.261224,
		0.210995,
		0.162676,
		0.11771,
		0.07764,
		0.044105,
		0.018834,
		0.003642,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		0.000349,
		0.008309,
		0.02558,
		0.050629,
		0.08199,
		0.118271,
		0.158149,
		0.200385,
		0.243819,
		0.287379,
		0.330088,
		0.371068,
		0.409549,
		0.444878,
		0.47653,
		0.504116,
		0.527398,
		0.546301,
		0.560924,
		0.571549,
		0.578638,
		0.582824,
		0.584866,
		0.585577,
		0.585694,
		0.585694,
		0.585577,
		0.584866,
		0.582824,
		0.578638,
		0.571549,
		0.560924,
		0.546301,
		0.527398,
		0.504116,
		0.47653,
		0.444878,
		0.409549,
		0.371068,
		0.330088,
		0.287379,
		0.243819,
		0.200385,
		0.158149,
		0.118271,
		0.08199,
		0.050629,
		0.02558,
		0.008309,
		0.000349,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.002517,
		0.013042,
		0.03061,
		0.054013,
		0.082099,
		0.11377,
		0.14799,
		0.183787,
		0.220253,
		0.256552,
		0.291922,
		0.325684,
		0.357244,
		0.386103,
		0.411863,
		0.434238,
		0.453061,
		0.468297,
		0.480048,
		0.48856,
		0.494223,
		0.497557,
		0.499179,
		0.499742,
		0.499835,
		0.499835,
		0.499742,
		0.499179,
		0.497557,
		0.494223,
		0.48856,
		0.480048,
		0.468297,
		0.453061,
		0.434238,
		0.411863,
		0.386103,
		0.357244,
		0.325684,
		0.291922,
		0.256552,
		0.220253,
		0.183787,
		0.14799,
		0.11377,
		0.082099,
		0.054013,
		0.03061,
		0.013042,
		0.002517,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		0.000228,
		0.005442,
		0.016787,
		0.0333,
		0.054055,
		0.07817,
		0.104807,
		0.133176,
		0.162535,
		0.192194,
		0.221518,
		0.249933,
		0.276925,
		0.302051,
		0.324939,
		0.3453,
		0.362928,
		0.377712,
		0.389643,
		0.398819,
		0.405447,
		0.409844,
		0.412426,
		0.413678,
		0.414112,
		0.414183,
		0.414183,
		0.414112,
		0.413678,
		0.412426,
		0.409844,
		0.405447,
		0.398819,
		0.389643,
		0.377712,
		0.362928,
		0.3453,
		0.324939,
		0.302051,
		0.276925,
		0.249933,
		0.221518,
		0.192194,
		0.162535,
		0.133176,
		0.104807,
		0.07817,
		0.054055,
		0.0333,
		0.016787,
		0.005442,
		0.000228,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.00155,
		0.008047,
		0.018926,
		0.033468,
		0.050989,
		0.070835,
		0.092384,
		0.115052,
		0.138292,
		0.161595,
		0.184495,
		0.206571,
		0.227449,
		0.246807,
		0.264379,
		0.279959,
		0.293406,
		0.304652,
		0.313702,
		0.320642,
		0.325642,
		0.32895,
		0.330886,
		0.331824,
		0.332147,
		0.332201,
		0.332201,
		0.332147,
		0.331824,
		0.330886,
		0.32895,
		0.325642,
		0.320642,
		0.313702,
		0.304652,
		0.293406,
		0.279959,
		0.264379,
		0.246807,
		0.227449,
		0.206571,
		0.184495,
		0.161595,
		0.138292,
		0.115052,
		0.092384,
		0.070835,
		0.050989,
		0.033468,
		0.018926,
		0.008047,
		0.00155,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		0.000131,
		0.003134,
		0.009688,
		0.019257,
		0.031327,
		0.045407,
		0.061029,
		0.077749,
		0.095151,
		0.112844,
		0.130466,
		0.147689,
		0.164213,
		0.179776,
		0.194153,
		0.207159,
		0.218656,
		0.228551,
		0.236802,
		0.243424,
		0.248488,
		0.252127,
		0.254529,
		0.255931,
		0.256608,
		0.256841,
		0.25688,
		0.25688,
		0.256841,
		0.256608,
		0.255931,
		0.254529,
		0.252127,
		0.248488,
		0.243424,
		0.236802,
		0.228551,
		0.218656,
		0.207159,
		0.194153,
		0.179776,
		0.164213,
		0.147689,
		0.130466,
		0.112844,
		0.095151,
		0.077749,
		0.061029,
		0.045407,
		0.031327,
		0.019257,
		0.009688,
		0.003134,
		0.000131,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000826,
		0.004295,
		0.010121,
		0.017935,
		0.027382,
		0.038125,
		0.049843,
		0.062232,
		0.075006,
		0.087897,
		0.10066,
		0.11307,
		0.124925,
		0.136047,
		0.146287,
		0.155521,
		0.163659,
		0.170644,
		0.176453,
		0.181102,
		0.18465,
		0.187192,
		0.188866,
		0.18984,
		0.19031,
		0.190471,
		0.190498,
		0.190498,
		0.190471,
		0.19031,
		0.18984,
		0.188866,
		0.187192,
		0.18465,
		0.181102,
		0.176453,
		0.170644,
		0.163659,
		0.155521,
		0.146287,
		0.136047,
		0.124925,
		0.11307,
		0.10066,
		0.087897,
		0.075006,
		0.062232,
		0.049843,
		0.038125,
		0.027382,
		0.017935,
		0.010121,
		0.004295,
		0.000826,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		6.4e-005,
		0.001528,
		0.004732,
		0.009423,
		0.01536,
		0.02231,
		0.030053,
		0.038377,
		0.047085,
		0.055989,
		0.064916,
		0.073706,
		0.082212,
		0.090306,
		0.097872,
		0.104814,
		0.111057,
		0.116543,
		0.121239,
		0.125134,
		0.128245,
		0.130612,
		0.132304,
		0.133415,
		0.134061,
		0.134371,
		0.134478,
		0.134495,
		0.134495,
		0.134478,
		0.134371,
		0.134061,
		0.133415,
		0.132304,
		0.130612,
		0.128245,
		0.125134,
		0.121239,
		0.116543,
		0.111057,
		0.104814,
		0.097872,
		0.090306,
		0.082212,
		0.073706,
		0.064916,
		0.055989,
		0.047085,
		0.038377,
		0.030053,
		0.02231,
		0.01536,
		0.009423,
		0.004732,
		0.001528,
		6.4e-005,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000361,
		0.001882,
		0.004443,
		0.007888,
		0.012067,
		0.016836,
		0.022059,
		0.027607,
		0.033356,
		0.039191,
		0.045007,
		0.050704,
		0.056194,
		0.061398,
		0.066246,
		0.070681,
		0.074658,
		0.078143,
		0.081119,
		0.083582,
		0.085543,
		0.087033,
		0.088095,
		0.088791,
		0.089194,
		0.089387,
		0.089454,
		0.089464,
		0.089464,
		0.089454,
		0.089387,
		0.089194,
		0.088791,
		0.088095,
		0.087033,
		0.085543,
		0.083582,
		0.081119,
		0.078143,
		0.074658,
		0.070681,
		0.066246,
		0.061398,
		0.056194,
		0.050704,
		0.045007,
		0.039191,
		0.033356,
		0.027607,
		0.022059,
		0.016836,
		0.012067,
		0.007888,
		0.004443,
		0.001882,
		0.000361,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		2.4e-005,
		0.000586,
		0.001816,
		0.003623,
		0.005917,
		0.008611,
		0.011623,
		0.014875,
		0.018292,
		0.021804,
		0.025346,
		0.028856,
		0.03228,
		0.035565,
		0.038668,
		0.041551,
		0.044179,
		0.04653,
		0.048585,
		0.050336,
		0.051781,
		0.052929,
		0.053799,
		0.054418,
		0.054822,
		0.055057,
		0.055169,
		0.055207,
		0.055213,
		0.055213,
		0.055207,
		0.055169,
		0.055057,
		0.054822,
		0.054418,
		0.053799,
		0.052929,
		0.051781,
		0.050336,
		0.048585,
		0.04653,
		0.044179,
		0.041551,
		0.038668,
		0.035565,
		0.03228,
		0.028856,
		0.025346,
		0.021804,
		0.018292,
		0.014875,
		0.011623,
		0.008611,
		0.005917,
		0.003623,
		0.001816,
		0.000586,
		2.4e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		0.000116,
		0.000608,
		0.001437,
		0.002555,
		0.003915,
		0.005473,
		0.007186,
		0.009012,
		0.010914,
		0.012854,
		0.014799,
		0.016717,
		0.018579,
		0.02036,
		0.022037,
		0.023589,
		0.025001,
		0.026261,
		0.027359,
		0.028292,
		0.029061,
		0.029671,
		0.030132,
		0.030459,
		0.030672,
		0.030795,
		0.030854,
		0.030874,
		0.030877,
		0.030877,
		0.030874,
		0.030854,
		0.030795,
		0.030672,
		0.030459,
		0.030132,
		0.029671,
		0.029061,
		0.028292,
		0.027359,
		0.026261,
		0.025001,
		0.023589,
		0.022037,
		0.02036,
		0.018579,
		0.016717,
		0.014799,
		0.012854,
		0.010914,
		0.009012,
		0.007186,
		0.005473,
		0.003915,
		0.002555,
		0.001437,
		0.000608,
		0.000116,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		0.000149,
		0.000464,
		0.000927,
		0.001517,
		0.002212,
		0.002991,
		0.003836,
		0.004727,
		0.005648,
		0.00658,
		0.007511,
		0.008424,
		0.009307,
		0.010148,
		0.010938,
		0.011667,
		0.012329,
		0.012918,
		0.01343,
		0.013864,
		0.014221,
		0.014503,
		0.014716,
		0.014867,
		0.014965,
		0.015022,
		0.015049,
		0.015058,
		0.015059,
		0.015059,
		0.015058,
		0.015049,
		0.015022,
		0.014965,
		0.014867,
		0.014716,
		0.014503,
		0.014221,
		0.013864,
		0.01343,
		0.012918,
		0.012329,
		0.011667,
		0.010938,
		0.010148,
		0.009307,
		0.008424,
		0.007511,
		0.00658,
		0.005648,
		0.004727,
		0.003836,
		0.002991,
		0.002212,
		0.001517,
		0.000927,
		0.000464,
		0.000149,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		2.1e-005,
		0.00011,
		0.000261,
		0.000466,
		0.000715,
		0.001001,
		0.001317,
		0.001654,
		0.002008,
		0.00237,
		0.002735,
		0.003097,
		0.003451,
		0.003792,
		0.004116,
		0.004419,
		0.004698,
		0.004951,
		0.005175,
		0.005369,
		0.005534,
		0.005669,
		0.005776,
		0.005856,
		0.005913,
		0.00595,
		0.005971,
		0.005981,
		0.005984,
		0.005985,
		0.005985,
		0.005984,
		0.005981,
		0.005971,
		0.00595,
		0.005913,
		0.005856,
		0.005776,
		0.005669,
		0.005534,
		0.005369,
		0.005175,
		0.004951,
		0.004698,
		0.004419,
		0.004116,
		0.003792,
		0.003451,
		0.003097,
		0.002735,
		0.00237,
		0.002008,
		0.001654,
		0.001317,
		0.001001,
		0.000715,
		0.000466,
		0.000261,
		0.00011,
		2.1e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		1.6e-005,
		4.9e-005,
		9.8e-005,
		0.00016,
		0.000234,
		0.000317,
		0.000407,
		0.000503,
		0.000602,
		0.000703,
		0.000804,
		0.000904,
		0.001001,
		0.001095,
		0.001183,
		0.001266,
		0.001342,
		0.00141,
		0.001471,
		0.001523,
		0.001568,
		0.001604,
		0.001633,
		0.001654,
		0.001669,
		0.001679,
		0.001685,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001685,
		0.001679,
		0.001669,
		0.001654,
		0.001633,
		0.001604,
		0.001568,
		0.001523,
		0.001471,
		0.00141,
		0.001342,
		0.001266,
		0.001183,
		0.001095,
		0.001001,
		0.000904,
		0.000804,
		0.000703,
		0.000602,
		0.000503,
		0.000407,
		0.000317,
		0.000234,
		0.00016,
		9.8e-005,
		4.9e-005,
		1.6e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.7e-005,
		2.7e-005,
		3.7e-005,
		4.9e-005,
		6.2e-005,
		7.5e-005,
		8.9e-005,
		0.000103,
		0.000117,
		0.000131,
		0.000144,
		0.000157,
		0.000169,
		0.00018,
		0.00019,
		0.000199,
		0.000208,
		0.000215,
		0.000221,
		0.000225,
		0.000229,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000229,
		0.000225,
		0.000221,
		0.000215,
		0.000208,
		0.000199,
		0.00019,
		0.00018,
		0.000169,
		0.000157,
		0.000144,
		0.000131,
		0.000117,
		0.000103,
		8.9e-005,
		7.5e-005,
		6.2e-005,
		4.9e-005,
		3.7e-005,
		2.7e-005,
		1.7e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		1.1e-005,
		2e-005,
		2.9e-005,
		4e-005,
		5.2e-005,
		6.5e-005,
		7.9e-005,
		9.3e-005,
		0.000107,
		0.000121,
		0.000134,
		0.000147,
		0.00016,
		0.000172,
		0.000183,
		0.000193,
		0.000202,
		0.00021,
		0.000216,
		0.000222,
		0.000227,
		0.00023,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000228,
		0.000224,
		0.000219,
		0.000213,
		0.000206,
		0.000197,
		0.000188,
		0.000177,
		0.000166,
		0.000154,
		0.000141,
		0.000127,
		0.000114,
		0.0001,
		8.6e-005,
		7.2e-005,
		5.9e-005,
		4.6e-005,
		3.5e-005,
		2.4e-005,
		1.5e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1e-005,
		3.9e-005,
		8.4e-005,
		0.000143,
		0.000215,
		0.000295,
		0.000384,
		0.000479,
		0.000577,
		0.000678,
		0.000779,
		0.000879,
		0.000977,
		0.001072,
		0.001162,
		0.001246,
		0.001323,
		0.001394,
		0.001456,
		0.001511,
		0.001557,
		0.001596,
		0.001626,
		0.001649,
		0.001666,
		0.001677,
		0.001684,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001681,
		0.001672,
		0.001658,
		0.001639,
		0.001612,
		0.001578,
		0.001535,
		0.001485,
		0.001426,
		0.00136,
		0.001285,
		0.001205,
		0.001117,
		0.001025,
		0.000929,
		0.000829,
		0.000728,
		0.000627,
		0.000527,
		0.000431,
		0.000339,
		0.000254,
		0.000178,
		0.000112,
		6e-005,
		2.2e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		3.7e-005,
		0.000143,
		0.000308,
		0.000524,
		0.000783,
		0.001077,
		0.001399,
		0.001742,
		0.002098,
		0.002461,
		0.002826,
		0.003186,
		0.003537,
		0.003874,
		0.004194,
		0.004491,
		0.004764,
		0.005009,
		0.005226,
		0.005413,
		0.005571,
		0.005698,
		0.005798,
		0.005872,
		0.005924,
		0.005956,
		0.005974,
		0.005982,
		0.005985,
		0.005985,
		0.005985,
		0.005984,
		0.005979,
		0.005967,
		0.005942,
		0.005901,
		0.005838,
		0.005752,
		0.005638,
		0.005496,
		0.005324,
		0.005121,
		0.00489,
		0.004631,
		0.004345,
		0.004037,
		0.003708,
		0.003363,
		0.003007,
		0.002644,
		0.002279,
		0.001918,
		0.001568,
		0.001235,
		0.000926,
		0.000649,
		0.00041,
		0.000218,
		8.2e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		9.7e-005,
		0.000371,
		0.000799,
		0.001359,
		0.00203,
		0.00279,
		0.00362,
		0.004501,
		0.005416,
		0.006347,
		0.007279,
		0.008198,
		0.009089,
		0.009942,
		0.010746,
		0.011491,
		0.01217,
		0.012777,
		0.013309,
		0.013763,
		0.014139,
		0.014439,
		0.014669,
		0.014835,
		0.014945,
		0.015011,
		0.015044,
		0.015057,
		0.015059,
		0.015059,
		0.015059,
		0.015052,
		0.015031,
		0.014983,
		0.014896,
		0.014759,
		0.014563,
		0.014298,
		0.01396,
		0.013546,
		0.013053,
		0.012483,
		0.011839,
		0.011126,
		0.010351,
		0.009521,
		0.008648,
		0.007741,
		0.006814,
		0.00588,
		0.004955,
		0.004055,
		0.003197,
		0.0024,
		0.001682,
		0.001064,
		0.000567,
		0.000213,
		2.5e-005,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		0.000205,
		0.000785,
		0.001691,
		0.002874,
		0.004288,
		0.005889,
		0.007633,
		0.009482,
		0.011397,
		0.013341,
		0.015282,
		0.017189,
		0.019033,
		0.02079,
		0.022437,
		0.023956,
		0.025331,
		0.026551,
		0.027608,
		0.0285,
		0.029228,
		0.029799,
		0.030225,
		0.030522,
		0.03071,
		0.030815,
		0.030862,
		0.030876,
		0.030877,
		0.030877,
		0.030871,
		0.030844,
		0.030771,
		0.030628,
		0.030388,
		0.03003,
		0.029533,
		0.028884,
		0.028075,
		0.0271,
		0.025961,
		0.024662,
		0.023214,
		0.021629,
		0.019924,
		0.01812,
		0.016241,
		0.014314,
		0.012367,
		0.010433,
		0.008547,
		0.006746,
		0.005068,
		0.003555,
		0.002251,
		0.0012,
		0.000451,
		5.2e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		0.000379,
		0.001451,
		0.003122,
		0.005302,
		0.007904,
		0.010844,
		0.014044,
		0.017426,
		0.020921,
		0.024461,
		0.027985,
		0.031435,
		0.034759,
		0.037912,
		0.040853,
		0.043547,
		0.04597,
		0.0481,
		0.049927,
		0.051448,
		0.052669,
		0.053606,
		0.054285,
		0.054739,
		0.055011,
		0.055149,
		0.055202,
		0.055213,
		0.055213,
		0.05521,
		0.055183,
		0.055094,
		0.054895,
		0.054538,
		0.053976,
		0.053172,
		0.052095,
		0.050725,
		0.049052,
		0.047072,
		0.044794,
		0.042233,
		0.039411,
		0.03636,
		0.033116,
		0.029723,
		0.026229,
		0.022689,
		0.019164,
		0.015716,
		0.012416,
		0.009337,
		0.006556,
		0.004154,
		0.002217,
		0.000834,
		9.7e-005,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		0.000635,
		0.002431,
		0.005227,
		0.008869,
		0.013209,
		0.018105,
		0.023421,
		0.02903,
		0.034811,
		0.040651,
		0.046446,
		0.052099,
		0.057525,
		0.062646,
		0.067396,
		0.07172,
		0.075576,
		0.078936,
		0.081783,
		0.084118,
		0.085958,
		0.087336,
		0.0883,
		0.088916,
		0.089259,
		0.089413,
		0.089459,
		0.089465,
		0.089464,
		0.089445,
		0.089354,
		0.089116,
		0.088647,
		0.087866,
		0.086702,
		0.085099,
		0.083014,
		0.080424,
		0.077319,
		0.073709,
		0.069614,
		0.065071,
		0.060128,
		0.054846,
		0.049296,
		0.04356,
		0.03773,
		0.031906,
		0.026196,
		0.020718,
		0.015596,
		0.010961,
		0.006952,
		0.003714,
		0.001398,
		0.000162,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000989,
		0.003782,
		0.008124,
		0.013772,
		0.02049,
		0.028054,
		0.036252,
		0.044882,
		0.053754,
		0.062691,
		0.07153,
		0.08012,
		0.088328,
		0.096035,
		0.103142,
		0.109565,
		0.115245,
		0.12014,
		0.124235,
		0.127539,
		0.130086,
		0.13194,
		0.133186,
		0.133936,
		0.134318,
		0.134464,
		0.134494,
		0.134495,
		0.134487,
		0.134412,
		0.134165,
		0.133616,
		0.132632,
		0.131095,
		0.128904,
		0.125984,
		0.122288,
		0.117792,
		0.112501,
		0.106443,
		0.099669,
		0.092251,
		0.084279,
		0.075864,
		0.067132,
		0.058225,
		0.049299,
		0.040526,
		0.032087,
		0.02418,
		0.017011,
		0.0108,
		0.005775,
		0.002176,
		0.000253,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.001451,
		0.005545,
		0.011902,
		0.020157,
		0.029959,
		0.040976,
		0.052889,
		0.0654,
		0.078227,
		0.091109,
		0.103804,
		0.116093,
		0.12778,
		0.138695,
		0.148694,
		0.157661,
		0.165515,
		0.172206,
		0.177722,
		0.18209,
		0.185375,
		0.187686,
		0.189167,
		0.189997,
		0.190372,
		0.190485,
		0.190498,
		0.190496,
		0.19045,
		0.190229,
		0.189652,
		0.18852,
		0.186644,
		0.183862,
		0.180046,
		0.17511,
		0.169008,
		0.161731,
		0.153312,
		0.143817,
		0.133344,
		0.122023,
		0.110012,
		0.097494,
		0.084677,
		0.071791,
		0.059088,
		0.046839,
		0.035336,
		0.024886,
		0.015814,
		0.008464,
		0.003192,
		0.000371,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.002029,
		0.007747,
		0.016611,
		0.028102,
		0.041725,
		0.057003,
		0.073489,
		0.090758,
		0.108413,
		0.126086,
		0.143438,
		0.160162,
		0.175988,
		0.19068,
		0.204045,
		0.21593,
		0.22623,
		0.234893,
		0.241918,
		0.247362,
		0.251342,
		0.254032,
		0.25566,
		0.256491,
		0.25681,
		0.256878,
		0.25688,
		0.256861,
		0.256697,
		0.256157,
		0.254963,
		0.252835,
		0.249525,
		0.244832,
		0.238608,
		0.230768,
		0.221283,
		0.21018,
		0.197539,
		0.183489,
		0.168203,
		0.151896,
		0.134822,
		0.117269,
		0.099561,
		0.082051,
		0.065122,
		0.049186,
		0.034677,
		0.02206,
		0.011818,
		0.004461,
		0.000519,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.002723,
		0.010384,
		0.022245,
		0.037596,
		0.055758,
		0.076086,
		0.097968,
		0.120829,
		0.144131,
		0.167375,
		0.190107,
		0.211916,
		0.232442,
		0.251376,
		0.268467,
		0.283525,
		0.296425,
		0.307118,
		0.31563,
		0.322066,
		0.326616,
		0.329549,
		0.331199,
		0.331948,
		0.332175,
		0.332201,
		0.332198,
		0.332104,
		0.331662,
		0.330511,
		0.328265,
		0.324562,
		0.319097,
		0.311642,
		0.302048,
		0.29025,
		0.276259,
		0.260166,
		0.242125,
		0.22236,
		0.201151,
		0.178832,
		0.155789,
		0.132457,
		0.109311,
		0.086869,
		0.06569,
		0.046366,
		0.029527,
		0.015835,
		0.005983,
		0.000697,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.003525,
		0.01343,
		0.028739,
		0.048517,
		0.071873,
		0.097956,
		0.125963,
		0.155141,
		0.184787,
		0.214251,
		0.242943,
		0.270336,
		0.295966,
		0.319444,
		0.340459,
		0.358785,
		0.374285,
		0.386925,
		0.396774,
		0.404013,
		0.408933,
		0.411925,
		0.413461,
		0.414053,
		0.414179,
		0.414184,
		0.414148,
		0.413844,
		0.412843,
		0.410642,
		0.406741,
		0.400704,
		0.392189,
		0.38096,
		0.366892,
		0.349969,
		0.330276,
		0.307995,
		0.283397,
		0.256833,
		0.228728,
		0.19958,
		0.169947,
		0.140451,
		0.111767,
		0.084624,
		0.059802,
		0.038126,
		0.020468,
		0.007742,
		0.000902,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.00442,
		0.016822,
		0.03596,
		0.06064,
		0.089722,
		0.122125,
		0.156828,
		0.192875,
		0.229374,
		0.265509,
		0.300538,
		0.333802,
		0.36473,
		0.392847,
		0.417783,
		0.43928,
		0.457205,
		0.471554,
		0.482466,
		0.490224,
		0.495252,
		0.498098,
		0.499394,
		0.49979,
		0.499836,
		0.49983,
		0.499666,
		0.498898,
		0.496909,
		0.493048,
		0.486716,
		0.477426,
		0.464821,
		0.448693,
		0.428973,
		0.405731,
		0.379165,
		0.34959,
		0.317429,
		0.283205,
		0.247536,
		0.211121,
		0.174741,
		0.139251,
		0.105574,
		0.074699,
		0.04768,
		0.025625,
		0.009702,
		0.001132,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.005385,
		0.020474,
		0.043719,
		0.073636,
		0.108813,
		0.147912,
		0.18967,
		0.232908,
		0.276532,
		0.31954,
		0.361029,
		0.400201,
		0.436373,
		0.468986,
		0.497617,
		0.521988,
		0.541984,
		0.557657,
		0.569245,
		0.577165,
		0.58201,
		0.584512,
		0.585481,
		0.585688,
		0.585696,
		0.585637,
		0.585137,
		0.583504,
		0.579929,
		0.57363,
		0.56394,
		0.55035,
		0.532534,
		0.510346,
		0.48382,
		0.453154,
		0.4187,
		0.38095,
		0.340527,
		0.298172,
		0.254737,
		0.211175,
		0.168534,
		0.127953,
		0.090652,
		0.057933,
		0.031171,
		0.011815,
		0.00138,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.006393,
		0.024279,
		0.051783,
		0.08711,
		0.128554,
		0.1745,
		0.223428,
		0.27392,
		0.324667,
		0.374475,
		0.422271,
		0.467118,
		0.508222,
		0.544946,
		0.576825,
		0.603579,
		0.625132,
		0.641623,
		0.653418,
		0.661108,
		0.665486,
		0.667493,
		0.668108,
		0.668181,
		0.668171,
		0.667916,
		0.666724,
		0.663655,
		0.657731,
		0.648076,
		0.633993,
		0.615002,
		0.590853,
		0.561514,
		0.527165,
		0.488182,
		0.445116,
		0.398686,
		0.349758,
		0.299339,
		0.248563,
		0.198682,
		0.151061,
		0.107171,
		0.068577,
		0.036943,
		0.014018,
		0.001639,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.007413,
		0.028122,
		0.059906,
		0.100644,
		0.14832,
		0.201032,
		0.256992,
		0.314536,
		0.372134,
		0.428393,
		0.482077,
		0.532108,
		0.577591,
		0.617822,
		0.652309,
		0.680795,
		0.703269,
		0.719989,
		0.73149,
		0.738568,
		0.742246,
		0.743679,
		0.743987,
		0.743999,
		0.743911,
		0.74317,
		0.740761,
		0.735522,
		0.726348,
		0.712321,
		0.692773,
		0.667309,
		0.635805,
		0.598399,
		0.555467,
		0.507609,
		0.455628,
		0.400512,
		0.343419,
		0.285665,
		0.228713,
		0.174161,
		0.123736,
		0.079284,
		0.042764,
		0.016247,
		0.001902,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.008417,
		0.031895,
		0.067856,
		0.113845,
		0.167531,
		0.226717,
		0.289343,
		0.353499,
		0.417432,
		0.47956,
		0.538479,
		0.592988,
		0.642099,
		0.685059,
		0.721373,
		0.75083,
		0.773521,
		0.789857,
		0.800578,
		0.80672,
		0.809553,
		0.810427,
		0.810531,
		0.810517,
		0.810153,
		0.808465,
		0.804146,
		0.795861,
		0.782445,
		0.763003,
		0.736961,
		0.704067,
		0.664385,
		0.618271,
		0.566352,
		0.5095,
		0.448814,
		0.385597,
		0.321343,
		0.257723,
		0.196569,
		0.139866,
		0.089745,
		0.048471,
		0.018437,
		0.002161,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.009384,
		0.035511,
		0.075448,
		0.126401,
		0.185725,
		0.250926,
		0.319677,
		0.389823,
		0.459397,
		0.526629,
		0.589968,
		0.648096,
		0.699951,
		0.744755,
		0.782037,
		0.81166,
		0.833854,
		0.849223,
		0.858746,
		0.863728,
		0.865681,
		0.866103,
		0.86612,
		0.865999,
		0.864986,
		0.861713,
		0.85464,
		0.842338,
		0.823657,
		0.7978,
		0.764353,
		0.723272,
		0.674863,
		0.619756,
		0.558873,
		0.493408,
		0.424797,
		0.354703,
		0.284996,
		0.21774,
		0.155176,
		0.099716,
		0.053931,
		0.020541,
		0.00241,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.010297,
		0.038915,
		0.082565,
		0.138117,
		0.202613,
		0.27327,
		0.347498,
		0.422905,
		0.497317,
		0.568794,
		0.635646,
		0.696458,
		0.750119,
		0.795845,
		0.83322,
		0.862222,
		0.883254,
		0.897154,
		0.905174,
		0.908898,
		0.910055,
		0.910193,
		0.910174,
		0.909691,
		0.907465,
		0.901806,
		0.891026,
		0.873692,
		0.848753,
		0.815588,
		0.774009,
		0.724239,
		0.666884,
		0.602893,
		0.533535,
		0.460362,
		0.385195,
		0.310092,
		0.237338,
		0.169425,
		0.109042,
		0.05906,
		0.022525,
		0.002646,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.011152,
		0.04209,
		0.08917,
		0.148931,
		0.218106,
		0.29363,
		0.372656,
		0.452566,
		0.530993,
		0.605834,
		0.675281,
		0.737842,
		0.792377,
		0.838132,
		0.874775,
		0.902438,
		0.921738,
		0.933787,
		0.940135,
		0.942641,
		0.943187,
		0.943208,
		0.943052,
		0.941747,
		0.937562,
		0.928581,
		0.913075,
		0.889702,
		0.857596,
		0.816383,
		0.766169,
		0.707502,
		0.641333,
		0.568982,
		0.492103,
		0.412655,
		0.332875,
		0.255258,
		0.182539,
		0.117674,
		0.063833,
		0.02438,
		0.002868,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.011955,
		0.045055,
		0.095303,
		0.15891,
		0.232304,
		0.312142,
		0.395328,
		0.479026,
		0.560686,
		0.63806,
		0.709237,
		0.77267,
		0.827216,
		0.872181,
		0.907357,
		0.933069,
		0.950194,
		0.960152,
		0.96481,
		0.966268,
		0.966443,
		0.966419,
		0.965809,
		0.963014,
		0.955961,
		0.942628,
		0.921354,
		0.890985,
		0.850921,
		0.801108,
		0.742003,
		0.674529,
		0.600037,
		0.520259,
		0.437279,
		0.3535,
		0.271617,
		0.194598,
		0.125665,
		0.068277,
		0.026116,
		0.003077,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.012718,
		0.047857,
		0.101067,
		0.168226,
		0.245457,
		0.329143,
		0.415942,
		0.502809,
		0.587017,
		0.666188,
		0.738324,
		0.801847,
		0.855639,
		0.899099,
		0.932184,
		0.955459,
		0.970107,
		0.977887,
		0.980982,
		0.981661,
		0.981687,
		0.981492,
		0.979875,
		0.974726,
		0.963764,
		0.944988,
		0.916918,
		0.878677,
		0.830009,
		0.771243,
		0.703245,
		0.627366,
		0.545402,
		0.459545,
		0.372351,
		0.286708,
		0.205814,
		0.133151,
		0.072466,
		0.027762,
		0.003275,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.013459,
		0.050566,
		0.106605,
		0.177115,
		0.257911,
		0.345096,
		0.435083,
		0.524618,
		0.61081,
		0.691157,
		0.763595,
		0.826533,
		0.878913,
		0.920262,
		0.950751,
		0.971234,
		0.983245,
		0.988911,
		0.990698,
		0.990914,
		0.990884,
		0.990133,
		0.986721,
		0.978178,
		0.962163,
		0.936827,
		0.900971,
		0.854087,
		0.796331,
		0.728476,
		0.651854,
		0.568298,
		0.480096,
		0.38995,
		0.300936,
		0.216478,
		0.140319,
		0.076504,
		0.029357,
		0.003469,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.014199,
		0.053259,
		0.112078,
		0.185844,
		0.270048,
		0.360506,
		0.453381,
		0.54521,
		0.632938,
		0.713954,
		0.786138,
		0.847917,
		0.898322,
		0.937057,
		0.964555,
		0.982016,
		0.991373,
		0.995127,
		0.995957,
		0.99599,
		0.995751,
		0.993781,
		0.987562,
		0.974436,
		0.952154,
		0.919139,
		0.874577,
		0.818404,
		0.751262,
		0.674429,
		0.589761,
		0.499627,
		0.406867,
		0.314743,
		0.22691,
		0.147381,
		0.080506,
		0.030947,
		0.003662,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.014959,
		0.05601,
		0.117642,
		0.194665,
		0.282229,
		0.375848,
		0.471423,
		0.565277,
		0.654192,
		0.735456,
		0.806913,
		0.867033,
		0.914977,
		0.950676,
		0.974887,
		0.989217,
		0.996039,
		0.99821,
		0.998474,
		0.998438,
		0.997522,
		0.993394,
		0.983155,
		0.964136,
		0.93433,
		0.892551,
		0.83846,
		0.772521,
		0.695929,
		0.610534,
		0.518777,
		0.423631,
		0.328548,
		0.237419,
		0.154541,
		0.084585,
		0.032575,
		0.003862,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.015757,
		0.058887,
		0.123433,
		0.203799,
		0.294765,
		0.39152,
		0.48969,
		0.585374,
		0.675192,
		0.756338,
		0.826642,
		0.884648,
		0.929694,
		0.961999,
		0.982718,
		0.993929,
		0.99847,
		0.999483,
		0.999523,
		0.99923,
		0.996836,
		0.989349,
		0.973698,
		0.94739,
		0.908799,
		0.857243,
		0.792954,
		0.717001,
		0.631206,
		0.538067,
		0.440684,
		0.342703,
		0.248267,
		0.161974,
		0.088841,
		0.034281,
		0.004071,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.016608,
		0.061944,
		0.129562,
		0.213418,
		0.307892,
		0.407819,
		0.508529,
		0.605889,
		0.696355,
		0.777035,
		0.845772,
		0.901223,
		0.942964,
		0.971575,
		0.988688,
		0.996917,
		0.999561,
		0.999886,
		0.999841,
		0.998721,
		0.993717,
		0.981426,
		0.958834,
		0.923796,
		0.875207,
		0.813001,
		0.738071,
		0.652181,
		0.557864,
		0.458348,
		0.357477,
		0.259662,
		0.169822,
		0.093355,
		0.036097,
		0.004295,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.017525,
		0.065226,
		0.136115,
		0.223654,
		0.321779,
		0.424946,
		0.528163,
		0.627051,
		0.717902,
		0.797756,
		0.864496,
		0.916947,
		0.954993,
		0.979666,
		0.993163,
		0.998688,
		0.999932,
		0.999981,
		0.99962,
		0.996693,
		0.987631,
		0.96889,
		0.937734,
		0.892541,
		0.832862,
		0.759357,
		0.673685,
		0.578397,
		0.476837,
		0.373058,
		0.271752,
		0.178194,
		0.098192,
		0.03805,
		0.004537,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.018518,
		0.068768,
		0.143159,
		0.234602,
		0.336549,
		0.443032,
		0.548717,
		0.648966,
		0.739907,
		0.818535,
		0.882813,
		0.931802,
		0.965783,
		0.986344,
		0.996346,
		0.999595,
		0.999998,
		0.999942,
		0.998558,
		0.992443,
		0.97759,
		0.9506,
		0.909231,
		0.852554,
		0.780912,
		0.695809,
		0.599782,
		0.496281,
		0.389572,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.023468,
		0.086243,
		0.177478,
		0.287165,
		0.406208,
		0.526505,
		0.641052,
		0.744084,
		0.83123,
		0.899725,
		0.948631,
		0.979066,
		0.994282,
		0.999355,
		0.999999,
		0.999909,
		0.997721,
		0.988294,
		0.966004,
		0.926607,
		0.867913,
		0.789856,
		0.694316,
		0.584888,
		0.466665,
		0.346048,
		0.230606,
		0.128949,
		0.050639,
		0.006111,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.025016,
		0.091646,
		0.18795,
		0.302949,
		0.426716,
		0.550474,
		0.66673,
		0.769433,
		0.854185,
		0.918486,
		0.962012,
		0.986838,
		0.997422,
		0.999897,
		0.999999,
		0.999268,
		0.993551,
		0.976536,
		0.942774,
		0.888967,
		0.814257,
		0.720092,
		0.609954,
		0.489096,
		0.364313,
		0.243759,
		0.136799,
		0.053897,
		0.006524,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.026722,
		0.09757,
		0.199356,
		0.320005,
		0.448658,
		0.575799,
		0.69341,
		0.795181,
		0.876762,
		0.93607,
		0.973615,
		0.9927,
		0.999166,
		0.999998,
		0.999882,
		0.997073,
		0.985151,
		0.957423,
		0.909232,
		0.838682,
		0.746597,
		0.636233,
		0.512956,
		0.383959,
		0.258031,
		0.145377,
		0.057479,
		0.006979,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.028608,
		0.104082,
		0.211807,
		0.338462,
		0.472143,
		0.602518,
		0.721027,
		0.821134,
		0.898654,
		0.952128,
		0.983187,
		0.996662,
		0.999864,
		0.999998,
		0.999046,
		0.991705,
		0.97023,
		0.928371,
		0.862877,
		0.773702,
		0.663715,
		0.538323,
		0.405109,
		0.273547,
		0.154777,
		0.061429,
		0.007484,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.0307,
		0.111261,
		0.225429,
		0.358463,
		0.497281,
		0.630658,
		0.749475,
		0.84704,
		0.919497,
		0.966289,
		0.990536,
		0.998903,
		0.999998,
		0.999844,
		0.996177,
		0.980889,
		0.945995,
		0.886525,
		0.801227,
		0.692366,
		0.565272,
		0.427896,
		0.290446,
		0.165103,
		0.065798,
		0.008046,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.033031,
		0.119202,
		0.24037,
		0.380165,
		0.52418,
		0.660215,
		0.778592,
		0.872571,
		0.938862,
		0.978187,
		0.995601,
		0.999819,
		0.999998,
		0.998733,
		0.989154,
		0.961681,
		0.909232,
		0.828924,
		0.722107,
		0.593863,
		0.452462,
		0.308888,
		0.176478,
		0.070648,
		0.008674,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.035637,
		0.128015,
		0.256796,
		0.403741,
		0.552945,
		0.69115,
		0.808146,
		0.897313,
		0.956268,
		0.987511,
		0.998528,
		0.999997,
		0.999788,
		0.994912,
		0.974996,
		0.930531,
		0.856465,
		0.752804,
		0.624138,
		0.478955,
		0.32905,
		0.189047,
		0.076051,
		0.009378,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.038563,
		0.137829,
		0.2749,
		0.429378,
		0.583667,
		0.723368,
		0.837812,
		0.920755,
		0.971204,
		0.994084,
		0.999752,
		0.999997,
		0.99828,
		0.985548,
		0.949876,
		0.883417,
		0.78425,
		0.656102,
		0.507526,
		0.351133,
		0.202976,
		0.082095,
		0.010171,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.041863,
		0.148801,
		0.294906,
		0.457278,
		0.61641,
		0.756699,
		0.867151,
		0.94229,
		0.983187,
		0.997978,
		0.999996,
		0.999707,
		0.993081,
		0.966673,
		0.909232,
		0.816138,
		0.689715,
		0.538323,
		0.375363,
		0.218463,
		0.088884,
		0.011069,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.045605,
		0.161115,
		0.317068,
		0.487653,
		0.651201,
		0.790869,
		0.895582,
		0.961225,
		0.991857,
		0.999651,
		0.999995,
		0.997607,
		0.980326,
		0.933236,
		0.848031,
		0.72486,
		0.571482,
		0.40199,
		0.23574,
		0.096544,
		0.012092,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.049871,
		0.174994,
		0.341684,
		0.520718,
		0.687996,
		0.825463,
		0.922365,
		0.976835,
		0.997147,
		0.999994,
		0.999581,
		0.990352,
		0.95463,
		0.879329,
		0.761317,
		0.607109,
		0.431293,
		0.255081,
		0.105229,
		0.013263,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.054761,
		0.19071,
		0.369091,
		0.556677,
		0.726654,
		0.859876,
		0.946591,
		0.988485,
		0.999493,
		0.999993,
		0.996571,
		0.972543,
		0.909232,
		0.798713,
		0.645261,
		0.463572,
		0.276811,
		0.11513,
		0.014613,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.060404,
		0.208589,
		0.399673,
		0.595706,
		0.766878,
		0.893264,
		0.967218,
		0.995843,
		0.999991,
		0.999381,
		0.986146,
		0.936718,
		0.836467,
		0.685901,
		0.499142,
		0.301315,
		0.12648,
		0.016179,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.066963,
		0.229035,
		0.433865,
		0.637915,
		0.808146,
		0.924495,
		0.983187,
		0.999236,
		0.999989,
		0.994912,
		0.960553,
		0.873716,
		0.728851,
		0.538323,
		0.32905,
		0.139573,
		0.018011,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.074644,
		0.25254,
		0.472143,
		0.683291,
		0.849616,
		0.952128,
		0.993706,
		0.999986,
		0.999046,
		0.9794,
		0.909232,
		0.773702,
		0.581406,
		0.360554,
		0.154777,
		0.020173,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.083718,
		0.279714,
		0.515011,
		0.731605,
		0.890001,
		0.974494,
		0.998793,
		0.999982,
		0.992122,
		0.941362,
		0.819691,
		0.628601,
		0.39646,
		0.172561,
		0.022747,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.094456,
		0.311037,
		0.56247,
		0.781582,
		0.926633,
		0.989142,
		0.9991,
		0.997575,
		0.967203,
		0.864779,
		0.679351,
		0.437117,
		0.19336,
		0.025825,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.104887,
		0.339533,
		0.600964,
		0.81321,
		0.935426,
		0.97297,
		0.974912,
		0.96241,
		0.886448,
		0.716734,
		0.472358,
		0.212988,
		0.028884,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.103703,
		0.328938,
		0.567267,
		0.743241,
		0.825636,
		0.839721,
		0.837497,
		0.795954,
		0.666205,
		0.45206,
		0.208583,
		0.028801,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.077731,
		0.240551,
		0.401745,
		0.505937,
		0.540983,
		0.542935,
		0.530799,
		0.462908,
		0.325662,
		0.154555,
		0.021807,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.037339,
		0.112084,
		0.179781,
		0.215674,
		0.222444,
		0.22133,
		0.202269,
		0.148905,
		0.073201,
		0.010604,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.008349,
		0.024116,
		0.036721,
		0.041577,
		0.041877,
		0.040099,
		0.031269,
		0.016081,
		0.002407,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		0.000173,
		0.000475,
		0.000676,
		0.000719,
		0.000712,
		0.000596,
		0.000326,
		5.1e-005,
		5.1e-005,
		0.000326,
		0.000596,
		0.000712,
		0.000719,
		0.000676,
		0.000475,
		0.000173,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		0.002407,
		0.016081,
		0.031269,
		0.040099,
		0.041877,
		0.041577,
		0.036721,
		0.024116,
		0.008349,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.010604,
		0.073201,
		0.148905,
		0.202269,
		0.22133,
		0.222444,
		0.215674,
		0.179781,
		0.112084,
		0.037339,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.021807,
		0.154555,
		0.325662,
		0.462908,
		0.530799,
		0.542935,
		0.540983,
		0.505937,
		0.401745,
		0.240551,
		0.077731,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.028801,
		0.208583,
		0.45206,
		0.666205,
		0.795954,
		0.837497,
		0.839721,
		0.825636,
		0.743241,
		0.567267,
		0.328938,
		0.103703,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.028884,
		0.212988,
		0.472358,
		0.716734,
		0.886448,
		0.96241,
		0.974912,
		0.97297,
		0.935426,
		0.81321,
		0.600964,
		0.339533,
		0.104887,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.025825,
		0.19336,
		0.437117,
		0.679351,
		0.864779,
		0.967203,
		0.997575,
		0.9991,
		0.989142,
		0.926633,
		0.781582,
		0.56247,
		0.311037,
		0.094456,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.022747,
		0.172561,
		0.39646,
		0.628601,
		0.819691,
		0.941362,
		0.992122,
		0.999982,
		0.998793,
		0.974494,
		0.890001,
		0.731605,
		0.515011,
		0.279714,
		0.083718,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.020173,
		0.154777,
		0.360554,
		0.581406,
		0.773702,
		0.909232,
		0.9794,
		0.999046,
		0.999986,
		0.993706,
		0.952128,
		0.849616,
		0.683291,
		0.472143,
		0.25254,
		0.074644,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.018011,
		0.139573,
		0.32905,
		0.538323,
		0.728851,
		0.873716,
		0.960553,
		0.994912,
		0.999989,
		0.999236,
		0.983187,
		0.924495,
		0.808146,
		0.637915,
		0.433865,
		0.229035,
		0.066963,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.016179,
		0.12648,
		0.301315,
		0.499142,
		0.685901,
		0.836467,
		0.936718,
		0.986146,
		0.999381,
		0.999991,
		0.995843,
		0.967218,
		0.893264,
		0.766878,
		0.595706,
		0.399673,
		0.208589,
		0.060404,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.014613,
		0.11513,
		0.276811,
		0.463572,
		0.645261,
		0.798713,
		0.909232,
		0.972543,
		0.996571,
		0.999993,
		0.999493,
		0.988485,
		0.946591,
		0.859876,
		0.726654,
		0.556677,
		0.369091,
		0.19071,
		0.054761,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.013263,
		0.105229,
		0.255081,
		0.431293,
		0.607109,
		0.761317,
		0.879329,
		0.95463,
		0.990352,
		0.999581,
		0.999994,
		0.997147,
		0.976835,
		0.922365,
		0.825463,
		0.687996,
		0.520718,
		0.341684,
		0.174994,
		0.049871,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.012092,
		0.096544,
		0.23574,
		0.40199,
		0.571482,
		0.72486,
		0.848031,
		0.933236,
		0.980326,
		0.997607,
		0.999995,
		0.999651,
		0.991857,
		0.961225,
		0.895582,
		0.790869,
		0.651201,
		0.487653,
		0.317068,
		0.161115,
		0.045605,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.011069,
		0.088884,
		0.218463,
		0.375363,
		0.538323,
		0.689715,
		0.816138,
		0.909232,
		0.966673,
		0.993081,
		0.999707,
		0.999996,
		0.997978,
		0.983187,
		0.94229,
		0.867151,
		0.756699,
		0.61641,
		0.457278,
		0.294906,
		0.148801,
		0.041863,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.010171,
		0.082095,
		0.202976,
		0.351133,
		0.507526,
		0.656102,
		0.78425,
		0.883417,
		0.949876,
		0.985548,
		0.99828,
		0.999997,
		0.999752,
		0.994084,
		0.971204,
		0.920755,
		0.837812,
		0.723368,
		0.583667,
		0.429378,
		0.2749,
		0.137829,
		0.038563,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.009378,
		0.076051,
		0.189047,
		0.32905,
		0.478955,
		0.624138,
		0.752804,
		0.856465,
		0.930531,
		0.974996,
		0.994912,
		0.999788,
		0.999997,
		0.998528,
		0.987511,
		0.956268,
		0.897313,
		0.808146,
		0.69115,
		0.552945,
		0.403741,
		0.256796,
		0.128015,
		0.035637,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.008674,
		0.070648,
		0.176478,
		0.308888,
		0.452462,
		0.593863,
		0.722107,
		0.828924,
		0.909232,
		0.961681,
		0.989154,
		0.998733,
		0.999998,
		0.999819,
		0.995601,
		0.978187,
		0.938862,
		0.872571,
		0.778592,
		0.660215,
		0.52418,
		0.380165,
		0.24037,
		0.119202,
		0.033031,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.008046,
		0.065798,
		0.165103,
		0.290446,
		0.427896,
		0.565272,
		0.692366,
		0.801227,
		0.886525,
		0.945995,
		0.980889,
		0.996177,
		0.999844,
		0.999998,
		0.998903,
		0.990536,
		0.966289,
		0.919497,
		0.84704,
		0.749475,
		0.630658,
		0.497281,
		0.358463,
		0.225429,
		0.111261,
		0.0307,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.007484,
		0.061429,
		0.154777,
		0.273547,
		0.405109,
		0.538323,
		0.663715,
		0.773702,
		0.862877,
		0.928371,
		0.97023,
		0.991705,
		0.999046,
		0.999998,
		0.999864,
		0.996662,
		0.983187,
		0.952128,
		0.898654,
		0.821134,
		0.721027,
		0.602518,
		0.472143,
		0.338462,
		0.211807,
		0.104082,
		0.028608,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.006979,
		0.057479,
		0.145377,
		0.258031,
		0.383959,
		0.512956,
		0.636233,
		0.746597,
		0.838682,
		0.909232,
		0.957423,
		0.985151,
		0.997073,
		0.999882,
		0.999998,
		0.999166,
		0.9927,
		0.973615,
		0.93607,
		0.876762,
		0.795181,
		0.69341,
		0.575799,
		0.448658,
		0.320005,
		0.199356,
		0.09757,
		0.026722,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.006524,
		0.053897,
		0.136799,
		0.243759,
		0.364313,
		0.489096,
		0.609954,
		0.720092,
		0.814257,
		0.888967,
		0.942774,
		0.976536,
		0.993551,
		0.999268,
		0.999999,
		0.999897,
		0.997422,
		0.986838,
		0.962012,
		0.918486,
		0.854185,
		0.769433,
		0.66673,
		0.550474,
		0.426716,
		0.302949,
		0.18795,
		0.091646,
		0.025016,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.006111,
		0.050639,
		0.128949,
		0.230606,
		0.346048,
		0.466665,
		0.584888,
		0.694316,
		0.789856,
		0.867913,
		0.926607,
		0.966004,
		0.988294,
		0.997721,
		0.999909,
		0.999999,
		0.999355,
		0.994282,
		0.979066,
		0.948631,
		0.899725,
		0.83123,
		0.744084,
		0.641052,
		0.526505,
		0.406208,
		0.287165,
		0.177478,
		0.086243,
		0.023468,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389572,
		0.496281,
		0.599782,
		0.695809,
		0.780912,
		0.852554,
		0.909231,
		0.9506,
		0.97759,
		0.992443,
		0.998558,
		0.999942,
		0.999998,
		0.999595,
		0.996346,
		0.986344,
		0.965783,
		0.931802,
		0.882813,
		0.818535,
		0.739907,
		0.648966,
		0.548717,
		0.443032,
		0.336549,
		0.234602,
		0.143159,
		0.068768,
		0.018518,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.004537,
		0.03805,
		0.098192,
		0.178194,
		0.271752,
		0.373058,
		0.476837,
		0.578397,
		0.673685,
		0.759357,
		0.832862,
		0.892541,
		0.937734,
		0.96889,
		0.987631,
		0.996693,
		0.99962,
		0.999981,
		0.999932,
		0.998688,
		0.993163,
		0.979666,
		0.954993,
		0.916947,
		0.864496,
		0.797756,
		0.717902,
		0.627051,
		0.528163,
		0.424946,
		0.321779,
		0.223654,
		0.136115,
		0.065226,
		0.017525,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.004295,
		0.036097,
		0.093355,
		0.169822,
		0.259662,
		0.357477,
		0.458348,
		0.557864,
		0.652181,
		0.738071,
		0.813001,
		0.875207,
		0.923796,
		0.958834,
		0.981426,
		0.993717,
		0.998721,
		0.999841,
		0.999886,
		0.999561,
		0.996917,
		0.988688,
		0.971575,
		0.942964,
		0.901223,
		0.845772,
		0.777035,
		0.696355,
		0.605889,
		0.508529,
		0.407819,
		0.307892,
		0.213418,
		0.129562,
		0.061944,
		0.016608,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.004071,
		0.034281,
		0.088841,
		0.161974,
		0.248267,
		0.342703,
		0.440684,
		0.538067,
		0.631206,
		0.717001,
		0.792954,
		0.857243,
		0.908799,
		0.94739,
		0.973698,
		0.989349,
		0.996836,
		0.99923,
		0.999523,
		0.999483,
		0.99847,
		0.993929,
		0.982718,
		0.961999,
		0.929694,
		0.884648,
		0.826642,
		0.756338,
		0.675192,
		0.585374,
		0.48969,
		0.39152,
		0.294765,
		0.203799,
		0.123433,
		0.058887,
		0.015757,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.003862,
		0.032575,
		0.084585,
		0.154541,
		0.237419,
		0.328548,
		0.423631,
		0.518777,
		0.610534,
		0.695929,
		0.772521,
		0.83846,
		0.892551,
		0.93433,
		0.964136,
		0.983155,
		0.993394,
		0.997522,
		0.998438,
		0.998474,
		0.99821,
		0.996039,
		0.989217,
		0.974887,
		0.950676,
		0.914977,
		0.867033,
		0.806913,
		0.735456,
		0.654192,
		0.565277,
		0.471423,
		0.375848,
		0.282229,
		0.194665,
		0.117642,
		0.05601,
		0.014959,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.003662,
		0.030947,
		0.080506,
		0.147381,
		0.22691,
		0.314743,
		0.406867,
		0.499627,
		0.589761,
		0.674429,
		0.751262,
		0.818404,
		0.874577,
		0.919139,
		0.952154,
		0.974436,
		0.987562,
		0.993781,
		0.995751,
		0.99599,
		0.995957,
		0.995127,
		0.991373,
		0.982016,
		0.964555,
		0.937057,
		0.898322,
		0.847917,
		0.786138,
		0.713954,
		0.632938,
		0.54521,
		0.453381,
		0.360506,
		0.270048,
		0.185844,
		0.112078,
		0.053259,
		0.014199,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.003469,
		0.029357,
		0.076504,
		0.140319,
		0.216478,
		0.300936,
		0.38995,
		0.480096,
		0.568298,
		0.651854,
		0.728476,
		0.796331,
		0.854087,
		0.900971,
		0.936827,
		0.962163,
		0.978178,
		0.986721,
		0.990133,
		0.990884,
		0.990914,
		0.990698,
		0.988911,
		0.983245,
		0.971234,
		0.950751,
		0.920262,
		0.878913,
		0.826533,
		0.763595,
		0.691157,
		0.61081,
		0.524618,
		0.435083,
		0.345096,
		0.257911,
		0.177115,
		0.106605,
		0.050566,
		0.013459,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.003275,
		0.027762,
		0.072466,
		0.133151,
		0.205814,
		0.286708,
		0.372351,
		0.459545,
		0.545402,
		0.627366,
		0.703245,
		0.771243,
		0.830009,
		0.878677,
		0.916918,
		0.944988,
		0.963764,
		0.974726,
		0.979875,
		0.981492,
		0.981687,
		0.981661,
		0.980982,
		0.977887,
		0.970107,
		0.955459,
		0.932184,
		0.899099,
		0.855639,
		0.801847,
		0.738324,
		0.666188,
		0.587017,
		0.502809,
		0.415942,
		0.329143,
		0.245457,
		0.168226,
		0.101067,
		0.047857,
		0.012718,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.003077,
		0.026116,
		0.068277,
		0.125665,
		0.194598,
		0.271617,
		0.3535,
		0.437279,
		0.520259,
		0.600037,
		0.674529,
		0.742003,
		0.801108,
		0.850921,
		0.890985,
		0.921354,
		0.942628,
		0.955961,
		0.963014,
		0.965809,
		0.966419,
		0.966443,
		0.966268,
		0.96481,
		0.960152,
		0.950194,
		0.933069,
		0.907357,
		0.872181,
		0.827216,
		0.77267,
		0.709237,
		0.63806,
		0.560686,
		0.479026,
		0.395328,
		0.312142,
		0.232304,
		0.15891,
		0.095303,
		0.045055,
		0.011955,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.002868,
		0.02438,
		0.063833,
		0.117674,
		0.182539,
		0.255258,
		0.332875,
		0.412655,
		0.492103,
		0.568982,
		0.641333,
		0.707502,
		0.766169,
		0.816383,
		0.857596,
		0.889702,
		0.913075,
		0.928581,
		0.937562,
		0.941747,
		0.943052,
		0.943208,
		0.943187,
		0.942641,
		0.940135,
		0.933787,
		0.921738,
		0.902438,
		0.874775,
		0.838132,
		0.792377,
		0.737842,
		0.675281,
		0.605834,
		0.530993,
		0.452566,
		0.372656,
		0.29363,
		0.218106,
		0.148931,
		0.08917,
		0.04209,
		0.011152,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.002646,
		0.022525,
		0.05906,
		0.109042,
		0.169425,
		0.237338,
		0.310092,
		0.385195,
		0.460362,
		0.533535,
		0.602893,
		0.666884,
		0.724239,
		0.774009,
		0.815588,
		0.848753,
		0.873692,
		0.891026,
		0.901806,
		0.907465,
		0.909691,
		0.910174,
		0.910193,
		0.910055,
		0.908898,
		0.905174,
		0.897154,
		0.883254,
		0.862222,
		0.83322,
		0.795845,
		0.750119,
		0.696458,
		0.635646,
		0.568794,
		0.497317,
		0.422905,
		0.347498,
		0.27327,
		0.202613,
		0.138117,
		0.082565,
		0.038915,
		0.010297,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.00241,
		0.020541,
		0.053931,
		0.099716,
		0.155176,
		0.21774,
		0.284996,
		0.354703,
		0.424797,
		0.493408,
		0.558873,
		0.619756,
		0.674863,
		0.723272,
		0.764353,
		0.7978,
		0.823657,
		0.842338,
		0.85464,
		0.861713,
		0.864986,
		0.865999,
		0.86612,
		0.866103,
		0.865681,
		0.863728,
		0.858746,
		0.849223,
		0.833854,
		0.81166,
		0.782037,
		0.744755,
		0.699951,
		0.648096,
		0.589968,
		0.526629,
		0.459397,
		0.389823,
		0.319677,
		0.250926,
		0.185725,
		0.126401,
		0.075448,
		0.035511,
		0.009384,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.002161,
		0.018437,
		0.048471,
		0.089745,
		0.139866,
		0.196569,
		0.257723,
		0.321343,
		0.385597,
		0.448814,
		0.5095,
		0.566352,
		0.618271,
		0.664385,
		0.704067,
		0.736961,
		0.763003,
		0.782445,
		0.795861,
		0.804146,
		0.808465,
		0.810153,
		0.810517,
		0.810531,
		0.810427,
		0.809553,
		0.80672,
		0.800578,
		0.789857,
		0.773521,
		0.75083,
		0.721373,
		0.685059,
		0.642099,
		0.592988,
		0.538479,
		0.47956,
		0.417432,
		0.353499,
		0.289343,
		0.226717,
		0.167531,
		0.113845,
		0.067856,
		0.031895,
		0.008417,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.001902,
		0.016247,
		0.042764,
		0.079284,
		0.123736,
		0.174161,
		0.228713,
		0.285665,
		0.343419,
		0.400512,
		0.455628,
		0.507609,
		0.555467,
		0.598399,
		0.635805,
		0.667309,
		0.692773,
		0.712321,
		0.726348,
		0.735522,
		0.740761,
		0.74317,
		0.743911,
		0.743999,
		0.743987,
		0.743679,
		0.742246,
		0.738568,
		0.73149,
		0.719989,
		0.703269,
		0.680795,
		0.652309,
		0.617822,
		0.577591,
		0.532108,
		0.482077,
		0.428393,
		0.372134,
		0.314536,
		0.256992,
		0.201032,
		0.14832,
		0.100644,
		0.059906,
		0.028122,
		0.007413,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.001639,
		0.014018,
		0.036943,
		0.068577,
		0.107171,
		0.151061,
		0.198682,
		0.248563,
		0.299339,
		0.349758,
		0.398686,
		0.445116,
		0.488182,
		0.527165,
		0.561514,
		0.590853,
		0.615002,
		0.633993,
		0.648076,
		0.657731,
		0.663655,
		0.666724,
		0.667916,
		0.668171,
		0.668181,
		0.668108,
		0.667493,
		0.665486,
		0.661108,
		0.653418,
		0.641623,
		0.625132,
		0.603579,
		0.576825,
		0.544946,
		0.508222,
		0.467118,
		0.422271,
		0.374475,
		0.324667,
		0.27392,
		0.223428,
		0.1745,
		0.128554,
		0.08711,
		0.051783,
		0.024279,
		0.006393,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.00138,
		0.011815,
		0.031171,
		0.057933,
		0.090652,
		0.127953,
		0.168534,
		0.211175,
		0.254737,
		0.298172,
		0.340527,
		0.38095,
		0.4187,
		0.453154,
		0.48382,
		0.510346,
		0.532534,
		0.55035,
		0.56394,
		0.57363,
		0.579929,
		0.583504,
		0.585137,
		0.585637,
		0.585696,
		0.585688,
		0.585481,
		0.584512,
		0.58201,
		0.577165,
		0.569245,
		0.557657,
		0.541984,
		0.521988,
		0.497617,
		0.468986,
		0.436373,
		0.400201,
		0.361029,
		0.31954,
		0.276532,
		0.232908,
		0.18967,
		0.147912,
		0.108813,
		0.073636,
		0.043719,
		0.020474,
		0.005385,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.001132,
		0.009702,
		0.025625,
		0.04768,
		0.074699,
		0.105574,
		0.139251,
		0.174741,
		0.211121,
		0.247536,
		0.283205,
		0.317429,
		0.34959,
		0.379165,
		0.405731,
		0.428973,
		0.448693,
		0.464821,
		0.477426,
		0.486716,
		0.493048,
		0.496909,
		0.498898,
		0.499666,
		0.49983,
		0.499836,
		0.49979,
		0.499394,
		0.498098,
		0.495252,
		0.490224,
		0.482466,
		0.471554,
		0.457205,
		0.43928,
		0.417783,
		0.392847,
		0.36473,
		0.333802,
		0.300538,
		0.265509,
		0.229374,
		0.192875,
		0.156828,
		0.122125,
		0.089722,
		0.06064,
		0.03596,
		0.016822,
		0.00442,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.000902,
		0.007742,
		0.020468,
		0.038126,
		0.059802,
		0.084624,
		0.111767,
		0.140451,
		0.169947,
		0.19958,
		0.228728,
		0.256833,
		0.283397,
		0.307995,
		0.330276,
		0.349969,
		0.366892,
		0.38096,
		0.392189,
		0.400704,
		0.406741,
		0.410642,
		0.412843,
		0.413844,
		0.414148,
		0.414184,
		0.414179,
		0.414053,
		0.413461,
		0.411925,
		0.408933,
		0.404013,
		0.396774,
		0.386925,
		0.374285,
		0.358785,
		0.340459,
		0.319444,
		0.295966,
		0.270336,
		0.242943,
		0.214251,
		0.184787,
		0.155141,
		0.125963,
		0.097956,
		0.071873,
		0.048517,
		0.028739,
		0.01343,
		0.003525,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.000697,
		0.005983,
		0.015835,
		0.029527,
		0.046366,
		0.06569,
		0.086869,
		0.109311,
		0.132457,
		0.155789,
		0.178832,
		0.201151,
		0.22236,
		0.242125,
		0.260166,
		0.276259,
		0.29025,
		0.302048,
		0.311642,
		0.319097,
		0.324562,
		0.328265,
		0.330511,
		0.331662,
		0.332104,
		0.332198,
		0.332201,
		0.332175,
		0.331948,
		0.331199,
		0.329549,
		0.326616,
		0.322066,
		0.31563,
		0.307118,
		0.296425,
		0.283525,
		0.268467,
		0.251376,
		0.232442,
		0.211916,
		0.190107,
		0.167375,
		0.144131,
		0.120829,
		0.097968,
		0.076086,
		0.055758,
		0.037596,
		0.022245,
		0.010384,
		0.002723,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.000519,
		0.004461,
		0.011818,
		0.02206,
		0.034677,
		0.049186,
		0.065122,
		0.082051,
		0.099561,
		0.117269,
		0.134822,
		0.151896,
		0.168203,
		0.183489,
		0.197539,
		0.21018,
		0.221283,
		0.230768,
		0.238608,
		0.244832,
		0.249525,
		0.252835,
		0.254963,
		0.256157,
		0.256697,
		0.256861,
		0.25688,
		0.256878,
		0.25681,
		0.256491,
		0.25566,
		0.254032,
		0.251342,
		0.247362,
		0.241918,
		0.234893,
		0.22623,
		0.21593,
		0.204045,
		0.19068,
		0.175988,
		0.160162,
		0.143438,
		0.126086,
		0.108413,
		0.090758,
		0.073489,
		0.057003,
		0.041725,
		0.028102,
		0.016611,
		0.007747,
		0.002029,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000371,
		0.003192,
		0.008464,
		0.015814,
		0.024886,
		0.035336,
		0.046839,
		0.059088,
		0.071791,
		0.084677,
		0.097494,
		0.110012,
		0.122023,
		0.133344,
		0.143817,
		0.153312,
		0.161731,
		0.169008,
		0.17511,
		0.180046,
		0.183862,
		0.186644,
		0.18852,
		0.189652,
		0.190229,
		0.19045,
		0.190496,
		0.190498,
		0.190485,
		0.190372,
		0.189997,
		0.189167,
		0.187686,
		0.185375,
		0.18209,
		0.177722,
		0.172206,
		0.165515,
		0.157661,
		0.148694,
		0.138695,
		0.12778,
		0.116093,
		0.103804,
		0.091109,
		0.078227,
		0.0654,
		0.052889,
		0.040976,
		0.029959,
		0.020157,
		0.011902,
		0.005545,
		0.001451,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.000253,
		0.002176,
		0.005775,
		0.0108,
		0.017011,
		0.02418,
		0.032087,
		0.040526,
		0.049299,
		0.058225,
		0.067132,
		0.075864,
		0.084279,
		0.092251,
		0.099669,
		0.106443,
		0.112501,
		0.117792,
		0.122288,
		0.125984,
		0.128904,
		0.131095,
		0.132632,
		0.133616,
		0.134165,
		0.134412,
		0.134487,
		0.134495,
		0.134494,
		0.134464,
		0.134318,
		0.133936,
		0.133186,
		0.13194,
		0.130086,
		0.127539,
		0.124235,
		0.12014,
		0.115245,
		0.109565,
		0.103142,
		0.096035,
		0.088328,
		0.08012,
		0.07153,
		0.062691,
		0.053754,
		0.044882,
		0.036252,
		0.028054,
		0.02049,
		0.013772,
		0.008124,
		0.003782,
		0.000989,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000162,
		0.001398,
		0.003714,
		0.006952,
		0.010961,
		0.015596,
		0.020718,
		0.026196,
		0.031906,
		0.03773,
		0.04356,
		0.049296,
		0.054846,
		0.060128,
		0.065071,
		0.069614,
		0.073709,
		0.077319,
		0.080424,
		0.083014,
		0.085099,
		0.086702,
		0.087866,
		0.088647,
		0.089116,
		0.089354,
		0.089445,
		0.089464,
		0.089465,
		0.089459,
		0.089413,
		0.089259,
		0.088916,
		0.0883,
		0.087336,
		0.085958,
		0.084118,
		0.081783,
		0.078936,
		0.075576,
		0.07172,
		0.067396,
		0.062646,
		0.057525,
		0.052099,
		0.046446,
		0.040651,
		0.034811,
		0.02903,
		0.023421,
		0.018105,
		0.013209,
		0.008869,
		0.005227,
		0.002431,
		0.000635,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		9.7e-005,
		0.000834,
		0.002217,
		0.004154,
		0.006556,
		0.009337,
		0.012416,
		0.015716,
		0.019164,
		0.022689,
		0.026229,
		0.029723,
		0.033116,
		0.03636,
		0.039411,
		0.042233,
		0.044794,
		0.047072,
		0.049052,
		0.050725,
		0.052095,
		0.053172,
		0.053976,
		0.054538,
		0.054895,
		0.055094,
		0.055183,
		0.05521,
		0.055213,
		0.055213,
		0.055202,
		0.055149,
		0.055011,
		0.054739,
		0.054285,
		0.053606,
		0.052669,
		0.051448,
		0.049927,
		0.0481,
		0.04597,
		0.043547,
		0.040853,
		0.037912,
		0.034759,
		0.031435,
		0.027985,
		0.024461,
		0.020921,
		0.017426,
		0.014044,
		0.010844,
		0.007904,
		0.005302,
		0.003122,
		0.001451,
		0.000379,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		5.2e-005,
		0.000451,
		0.0012,
		0.002251,
		0.003555,
		0.005068,
		0.006746,
		0.008547,
		0.010433,
		0.012367,
		0.014314,
		0.016241,
		0.01812,
		0.019924,
		0.021629,
		0.023214,
		0.024662,
		0.025961,
		0.0271,
		0.028075,
		0.028884,
		0.029533,
		0.03003,
		0.030388,
		0.030628,
		0.030771,
		0.030844,
		0.030871,
		0.030877,
		0.030877,
		0.030876,
		0.030862,
		0.030815,
		0.03071,
		0.030522,
		0.030225,
		0.029799,
		0.029228,
		0.0285,
		0.027608,
		0.026551,
		0.025331,
		0.023956,
		0.022437,
		0.02079,
		0.019033,
		0.017189,
		0.015282,
		0.013341,
		0.011397,
		0.009482,
		0.007633,
		0.005889,
		0.004288,
		0.002874,
		0.001691,
		0.000785,
		0.000205,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		2.5e-005,
		0.000213,
		0.000567,
		0.001064,
		0.001682,
		0.0024,
		0.003197,
		0.004055,
		0.004955,
		0.00588,
		0.006814,
		0.007741,
		0.008648,
		0.009521,
		0.010351,
		0.011126,
		0.011839,
		0.012483,
		0.013053,
		0.013546,
		0.01396,
		0.014298,
		0.014563,
		0.014759,
		0.014896,
		0.014983,
		0.015031,
		0.015052,
		0.015059,
		0.015059,
		0.015059,
		0.015057,
		0.015044,
		0.015011,
		0.014945,
		0.014835,
		0.014669,
		0.014439,
		0.014139,
		0.013763,
		0.013309,
		0.012777,
		0.01217,
		0.011491,
		0.010746,
		0.009942,
		0.009089,
		0.008198,
		0.007279,
		0.006347,
		0.005416,
		0.004501,
		0.00362,
		0.00279,
		0.00203,
		0.001359,
		0.000799,
		0.000371,
		9.7e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		8.2e-005,
		0.000218,
		0.00041,
		0.000649,
		0.000926,
		0.001235,
		0.001568,
		0.001918,
		0.002279,
		0.002644,
		0.003007,
		0.003363,
		0.003708,
		0.004037,
		0.004345,
		0.004631,
		0.00489,
		0.005121,
		0.005324,
		0.005496,
		0.005638,
		0.005752,
		0.005838,
		0.005901,
		0.005942,
		0.005967,
		0.005979,
		0.005984,
		0.005985,
		0.005985,
		0.005985,
		0.005982,
		0.005974,
		0.005956,
		0.005924,
		0.005872,
		0.005798,
		0.005698,
		0.005571,
		0.005413,
		0.005226,
		0.005009,
		0.004764,
		0.004491,
		0.004194,
		0.003874,
		0.003537,
		0.003186,
		0.002826,
		0.002461,
		0.002098,
		0.001742,
		0.001399,
		0.001077,
		0.000783,
		0.000524,
		0.000308,
		0.000143,
		3.7e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		2.2e-005,
		6e-005,
		0.000112,
		0.000178,
		0.000254,
		0.000339,
		0.000431,
		0.000527,
		0.000627,
		0.000728,
		0.000829,
		0.000929,
		0.001025,
		0.001117,
		0.001205,
		0.001285,
		0.00136,
		0.001426,
		0.001485,
		0.001535,
		0.001578,
		0.001612,
		0.001639,
		0.001658,
		0.001672,
		0.001681,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001684,
		0.001677,
		0.001666,
		0.001649,
		0.001626,
		0.001596,
		0.001557,
		0.001511,
		0.001456,
		0.001394,
		0.001323,
		0.001246,
		0.001162,
		0.001072,
		0.000977,
		0.000879,
		0.000779,
		0.000678,
		0.000577,
		0.000479,
		0.000384,
		0.000295,
		0.000215,
		0.000143,
		8.4e-005,
		3.9e-005,
		1e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.5e-005,
		2.4e-005,
		3.5e-005,
		4.6e-005,
		5.9e-005,
		7.2e-005,
		8.6e-005,
		0.0001,
		0.000114,
		0.000127,
		0.000141,
		0.000154,
		0.000166,
		0.000177,
		0.000188,
		0.000197,
		0.000206,
		0.000213,
		0.000219,
		0.000224,
		0.000228,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.00023,
		0.000227,
		0.000222,
		0.000216,
		0.00021,
		0.000202,
		0.000193,
		0.000183,
		0.000172,
		0.00016,
		0.000147,
		0.000134,
		0.000121,
		0.000107,
		9.3e-005,
		7.9e-005,
		6.5e-005,
		5.2e-005,
		4e-005,
		2.9e-005,
		2e-005,
		1.1e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		1.3e-005,
		2.2e-005,
		3.2e-005,
		4.3e-005,
		5.6e-005,
		6.9e-005,
		8.2e-005,
		9.6e-005,
		0.00011,
		0.000124,
		0.000138,
		0.000151,
		0.000163,
		0.000175,
		0.000185,
		0.000195,
		0.000204,
		0.000211,
		0.000218,
		0.000223,
		0.000228,
		0.000231,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.000231,
		0.000228,
		0.000223,
		0.000218,
		0.000211,
		0.000204,
		0.000195,
		0.000185,
		0.000175,
		0.000163,
		0.000151,
		0.000138,
		0.000124,
		0.00011,
		9.6e-005,
		8.2e-005,
		6.9e-005,
		5.6e-005,
		4.3e-005,
		3.2e-005,
		2.2e-005,
		1.3e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		3e-005,
		7.2e-005,
		0.000127,
		0.000196,
		0.000274,
		0.000361,
		0.000454,
		0.000552,
		0.000652,
		0.000754,
		0.000854,
		0.000953,
		0.001049,
		0.00114,
		0.001225,
		0.001305,
		0.001377,
		0.001442,
		0.001498,
		0.001547,
		0.001587,
		0.001619,
		0.001644,
		0.001662,
		0.001675,
		0.001682,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001682,
		0.001675,
		0.001662,
		0.001644,
		0.001619,
		0.001587,
		0.001547,
		0.001498,
		0.001442,
		0.001377,
		0.001305,
		0.001225,
		0.00114,
		0.001049,
		0.000953,
		0.000854,
		0.000754,
		0.000652,
		0.000552,
		0.000454,
		0.000361,
		0.000274,
		0.000196,
		0.000127,
		7.2e-005,
		3e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		5.7e-005,
		0.000179,
		0.000357,
		0.000585,
		0.000854,
		0.001156,
		0.001483,
		0.00183,
		0.002188,
		0.002552,
		0.002916,
		0.003275,
		0.003623,
		0.003956,
		0.00427,
		0.004562,
		0.004828,
		0.005066,
		0.005276,
		0.005456,
		0.005605,
		0.005726,
		0.005819,
		0.005887,
		0.005933,
		0.005962,
		0.005977,
		0.005983,
		0.005985,
		0.005985,
		0.005985,
		0.005983,
		0.005977,
		0.005962,
		0.005933,
		0.005887,
		0.005819,
		0.005726,
		0.005605,
		0.005456,
		0.005276,
		0.005066,
		0.004828,
		0.004562,
		0.00427,
		0.003956,
		0.003623,
		0.003275,
		0.002916,
		0.002552,
		0.002188,
		0.00183,
		0.001483,
		0.001156,
		0.000854,
		0.000585,
		0.000357,
		0.000179,
		5.7e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		5.5e-005,
		0.000287,
		0.000679,
		0.001208,
		0.001853,
		0.002592,
		0.003407,
		0.004277,
		0.005185,
		0.006113,
		0.007047,
		0.00797,
		0.00887,
		0.009733,
		0.01055,
		0.011311,
		0.012007,
		0.012633,
		0.013183,
		0.013657,
		0.014052,
		0.014371,
		0.014618,
		0.014799,
		0.014922,
		0.014998,
		0.015038,
		0.015055,
		0.015059,
		0.015059,
		0.015059,
		0.015055,
		0.015038,
		0.014998,
		0.014922,
		0.014799,
		0.014618,
		0.014371,
		0.014052,
		0.013657,
		0.013183,
		0.012633,
		0.012007,
		0.011311,
		0.01055,
		0.009733,
		0.00887,
		0.00797,
		0.007047,
		0.006113,
		0.005185,
		0.004277,
		0.003407,
		0.002592,
		0.001853,
		0.001208,
		0.000679,
		0.000287,
		5.5e-005,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		1.3e-005,
		0.000317,
		0.000983,
		0.001962,
		0.003208,
		0.004672,
		0.006313,
		0.008087,
		0.009956,
		0.011881,
		0.013828,
		0.015763,
		0.017657,
		0.019481,
		0.021213,
		0.02283,
		0.024314,
		0.025651,
		0.02683,
		0.027847,
		0.028697,
		0.029385,
		0.029919,
		0.030311,
		0.030578,
		0.030743,
		0.030831,
		0.030867,
		0.030877,
		0.030877,
		0.030877,
		0.030867,
		0.030831,
		0.030743,
		0.030578,
		0.030311,
		0.029919,
		0.029385,
		0.028697,
		0.027847,
		0.02683,
		0.025651,
		0.024314,
		0.02283,
		0.021213,
		0.019481,
		0.017657,
		0.015763,
		0.013828,
		0.011881,
		0.009956,
		0.008087,
		0.006313,
		0.004672,
		0.003208,
		0.001962,
		0.000983,
		0.000317,
		1.3e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		0.000215,
		0.001123,
		0.002653,
		0.004715,
		0.007219,
		0.010082,
		0.013224,
		0.016567,
		0.02004,
		0.023575,
		0.027109,
		0.030582,
		0.033943,
		0.037142,
		0.040139,
		0.042898,
		0.045391,
		0.047595,
		0.049499,
		0.051096,
		0.052391,
		0.053397,
		0.054138,
		0.054644,
		0.054958,
		0.055125,
		0.055194,
		0.055212,
		0.055213,
		0.055212,
		0.055194,
		0.055125,
		0.054958,
		0.054644,
		0.054138,
		0.053397,
		0.052391,
		0.051096,
		0.049499,
		0.047595,
		0.045391,
		0.042898,
		0.040139,
		0.037142,
		0.033943,
		0.030582,
		0.027109,
		0.023575,
		0.02004,
		0.016567,
		0.013224,
		0.010082,
		0.007219,
		0.004715,
		0.002653,
		0.001123,
		0.000215,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		4.1e-005,
		0.000982,
		0.003043,
		0.006065,
		0.009895,
		0.014386,
		0.019399,
		0.0248,
		0.030463,
		0.036269,
		0.042108,
		0.047876,
		0.05348,
		0.058836,
		0.06387,
		0.068519,
		0.072729,
		0.076464,
		0.079696,
		0.082414,
		0.084624,
		0.086344,
		0.087614,
		0.088484,
		0.089024,
		0.089312,
		0.089431,
		0.089462,
		0.089465,
		0.089462,
		0.089431,
		0.089312,
		0.089024,
		0.088484,
		0.087614,
		0.086344,
		0.084624,
		0.082414,
		0.079696,
		0.076464,
		0.072729,
		0.068519,
		0.06387,
		0.058836,
		0.05348,
		0.047876,
		0.042108,
		0.036269,
		0.030463,
		0.0248,
		0.019399,
		0.014386,
		0.009895,
		0.006065,
		0.003043,
		0.000982,
		4.1e-005,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000562,
		0.002929,
		0.006907,
		0.01225,
		0.018722,
		0.026096,
		0.034155,
		0.042695,
		0.051523,
		0.06046,
		0.069338,
		0.078003,
		0.086318,
		0.094161,
		0.101426,
		0.108027,
		0.113897,
		0.118991,
		0.123286,
		0.126786,
		0.129518,
		0.131537,
		0.132926,
		0.133789,
		0.13425,
		0.134442,
		0.134492,
		0.134496,
		0.134492,
		0.134442,
		0.13425,
		0.133789,
		0.132926,
		0.131537,
		0.129518,
		0.126786,
		0.123286,
		0.118991,
		0.113897,
		0.108027,
		0.101426,
		0.094161,
		0.086318,
		0.078003,
		0.069338,
		0.06046,
		0.051523,
		0.042695,
		0.034155,
		0.026096,
		0.018722,
		0.01225,
		0.006907,
		0.002929,
		0.000562,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		9.4e-005,
		0.002242,
		0.006937,
		0.013802,
		0.022475,
		0.032612,
		0.043882,
		0.055972,
		0.068588,
		0.081453,
		0.094309,
		0.106923,
		0.119079,
		0.130588,
		0.141286,
		0.151036,
		0.159732,
		0.167298,
		0.173695,
		0.17892,
		0.183009,
		0.186039,
		0.188127,
		0.189429,
		0.190126,
		0.190417,
		0.190492,
		0.190498,
		0.190492,
		0.190417,
		0.190126,
		0.189429,
		0.188127,
		0.186039,
		0.183009,
		0.17892,
		0.173695,
		0.167298,
		0.159732,
		0.151036,
		0.141286,
		0.130588,
		0.119079,
		0.106923,
		0.094309,
		0.081453,
		0.068588,
		0.055972,
		0.043882,
		0.032612,
		0.022475,
		0.013802,
		0.006937,
		0.002242,
		9.4e-005,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.001155,
		0.006001,
		0.014128,
		0.025011,
		0.038146,
		0.053053,
		0.069278,
		0.08639,
		0.103984,
		0.121685,
		0.139147,
		0.156055,
		0.172129,
		0.187125,
		0.200837,
		0.213104,
		0.223808,
		0.232882,
		0.240313,
		0.246144,
		0.250476,
		0.253469,
		0.255338,
		0.256343,
		0.256763,
		0.256872,
		0.25688,
		0.256872,
		0.256763,
		0.256343,
		0.255338,
		0.253469,
		0.250476,
		0.246144,
		0.240313,
		0.232882,
		0.223808,
		0.213104,
		0.200837,
		0.187125,
		0.172129,
		0.156055,
		0.139147,
		0.121685,
		0.103984,
		0.08639,
		0.069278,
		0.053053,
		0.038146,
		0.025011,
		0.014128,
		0.006001,
		0.001155,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		0.000176,
		0.004205,
		0.012984,
		0.025783,
		0.041898,
		0.060661,
		0.081434,
		0.103613,
		0.126633,
		0.149966,
		0.173123,
		0.195661,
		0.217181,
		0.237336,
		0.25583,
		0.272428,
		0.286955,
		0.299306,
		0.309448,
		0.317427,
		0.323371,
		0.327489,
		0.330067,
		0.331456,
		0.332039,
		0.33219,
		0.332201,
		0.33219,
		0.332039,
		0.331456,
		0.330067,
		0.327489,
		0.323371,
		0.317427,
		0.309448,
		0.299306,
		0.286955,
		0.272428,
		0.25583,
		0.237336,
		0.217181,
		0.195661,
		0.173123,
		0.149966,
		0.126633,
		0.103613,
		0.081434,
		0.060661,
		0.041898,
		0.025783,
		0.012984,
		0.004205,
		0.000176,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.002007,
		0.010409,
		0.024457,
		0.043203,
		0.065746,
		0.091224,
		0.118823,
		0.147776,
		0.177368,
		0.206935,
		0.235873,
		0.263636,
		0.289746,
		0.313795,
		0.33545,
		0.354464,
		0.370678,
		0.384031,
		0.394565,
		0.402434,
		0.4079,
		0.411333,
		0.413186,
		0.413966,
		0.414169,
		0.414184,
		0.414169,
		0.413966,
		0.413186,
		0.411333,
		0.4079,
		0.402434,
		0.394565,
		0.384031,
		0.370678,
		0.354464,
		0.33545,
		0.313795,
		0.289746,
		0.263636,
		0.235873,
		0.206935,
		0.177368,
		0.147776,
		0.118823,
		0.091224,
		0.065746,
		0.043203,
		0.024457,
		0.010409,
		0.002007,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		0.000286,
		0.006822,
		0.021023,
		0.041656,
		0.067541,
		0.097552,
		0.130624,
		0.16575,
		0.20199,
		0.238473,
		0.274398,
		0.309044,
		0.341773,
		0.372039,
		0.399391,
		0.423487,
		0.444099,
		0.461124,
		0.474596,
		0.484687,
		0.491718,
		0.496144,
		0.498541,
		0.499553,
		0.499817,
		0.499836,
		0.499817,
		0.499553,
		0.498541,
		0.496144,
		0.491718,
		0.484687,
		0.474596,
		0.461124,
		0.444099,
		0.423487,
		0.399391,
		0.372039,
		0.341773,
		0.309044,
		0.274398,
		0.238473,
		0.20199,
		0.16575,
		0.130624,
		0.097552,
		0.067541,
		0.041656,
		0.021023,
		0.006822,
		0.000286,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.003067,
		0.015878,
		0.037225,
		0.065609,
		0.0996,
		0.13784,
		0.179047,
		0.222021,
		0.265647,
		0.308897,
		0.350845,
		0.390664,
		0.427644,
		0.461192,
		0.49085,
		0.516304,
		0.537395,
		0.554134,
		0.566711,
		0.575499,
		0.581049,
		0.584062,
		0.585338,
		0.585671,
		0.585696,
		0.585671,
		0.585338,
		0.584062,
		0.581049,
		0.575499,
		0.566711,
		0.554134,
		0.537395,
		0.516304,
		0.49085,
		0.461192,
		0.427644,
		0.390664,
		0.350845,
		0.308897,
		0.265647,
		0.222021,
		0.179047,
		0.13784,
		0.0996,
		0.065609,
		0.037225,
		0.015878,
		0.003067,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		0.000415,
		0.009861,
		0.030325,
		0.059949,
		0.096962,
		0.13968,
		0.186509,
		0.235957,
		0.286632,
		0.337251,
		0.386651,
		0.433793,
		0.477773,
		0.517835,
		0.553385,
		0.584001,
		0.609454,
		0.629719,
		0.644993,
		0.655698,
		0.662478,
		0.666171,
		0.66774,
		0.66815,
		0.668181,
		0.66815,
		0.66774,
		0.666171,
		0.662478,
		0.655698,
		0.644993,
		0.629719,
		0.609454,
		0.584001,
		0.553385,
		0.517835,
		0.477773,
		0.433793,
		0.386651,
		0.337251,
		0.286632,
		0.235957,
		0.186509,
		0.13968,
		0.096962,
		0.059949,
		0.030325,
		0.009861,
		0.000415,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.004224,
		0.021821,
		0.051039,
		0.089732,
		0.135857,
		0.187482,
		0.242791,
		0.300088,
		0.357808,
		0.414524,
		0.468959,
		0.519995,
		0.566689,
		0.608287,
		0.644244,
		0.674241,
		0.698204,
		0.716324,
		0.729065,
		0.73716,
		0.741583,
		0.743467,
		0.743962,
		0.743999,
		0.743962,
		0.743467,
		0.741583,
		0.73716,
		0.729065,
		0.716324,
		0.698204,
		0.674241,
		0.644244,
		0.608287,
		0.566689,
		0.519995,
		0.468959,
		0.414524,
		0.357808,
		0.300088,
		0.242791,
		0.187482,
		0.135857,
		0.089732,
		0.051039,
		0.021821,
		0.004224,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		0.000547,
		0.012978,
		0.039813,
		0.078506,
		0.126633,
		0.181895,
		0.242128,
		0.30531,
		0.369568,
		0.433192,
		0.494641,
		0.552562,
		0.605806,
		0.653439,
		0.694774,
		0.729382,
		0.757127,
		0.778177,
		0.793028,
		0.802496,
		0.807686,
		0.809903,
		0.810487,
		0.810531,
		0.810487,
		0.809903,
		0.807686,
		0.802496,
		0.793028,
		0.778177,
		0.757127,
		0.729382,
		0.694774,
		0.653439,
		0.605806,
		0.552562,
		0.494641,
		0.433192,
		0.369568,
		0.30531,
		0.242128,
		0.181895,
		0.126633,
		0.078506,
		0.039813,
		0.012978,
		0.000547,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.00535,
		0.027572,
		0.064324,
		0.112777,
		0.170248,
		0.234205,
		0.302277,
		0.372267,
		0.442157,
		0.510129,
		0.574574,
		0.634115,
		0.687623,
		0.734247,
		0.773435,
		0.804966,
		0.828975,
		0.845973,
		0.856847,
		0.862827,
		0.865392,
		0.866069,
		0.86612,
		0.866069,
		0.865392,
		0.862827,
		0.856847,
		0.845973,
		0.828975,
		0.804966,
		0.773435,
		0.734247,
		0.687623,
		0.634115,
		0.574574,
		0.510129,
		0.442157,
		0.372267,
		0.302277,
		0.234205,
		0.170248,
		0.112777,
		0.064324,
		0.027572,
		0.00535,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		0.000671,
		0.015865,
		0.048543,
		0.095456,
		0.153515,
		0.219804,
		0.291589,
		0.366327,
		0.441681,
		0.515532,
		0.586001,
		0.651464,
		0.71058,
		0.762317,
		0.80598,
		0.841247,
		0.868202,
		0.887355,
		0.899653,
		0.906441,
		0.909361,
		0.910135,
		0.910193,
		0.910135,
		0.909361,
		0.906441,
		0.899653,
		0.887355,
		0.868202,
		0.841247,
		0.80598,
		0.762317,
		0.71058,
		0.651464,
		0.586001,
		0.515532,
		0.441681,
		0.366327,
		0.291589,
		0.219804,
		0.153515,
		0.095456,
		0.048543,
		0.015865,
		0.000671,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.006363,
		0.032703,
		0.076079,
		0.132984,
		0.200101,
		0.274314,
		0.352718,
		0.43264,
		0.511645,
		0.587564,
		0.658512,
		0.722916,
		0.779544,
		0.827542,
		0.86647,
		0.896339,
		0.917645,
		0.931376,
		0.938984,
		0.942269,
		0.943143,
		0.943208,
		0.943143,
		0.942269,
		0.938984,
		0.931376,
		0.917645,
		0.896339,
		0.86647,
		0.827542,
		0.779544,
		0.722916,
		0.658512,
		0.587564,
		0.511645,
		0.43264,
		0.352718,
		0.274314,
		0.200101,
		0.132984,
		0.076079,
		0.032703,
		0.006363,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		0.00078,
		0.018407,
		0.056161,
		0.110096,
		0.176476,
		0.251784,
		0.33274,
		0.416309,
		0.499726,
		0.580512,
		0.656502,
		0.725873,
		0.787175,
		0.839373,
		0.881889,
		0.914647,
		0.938108,
		0.953289,
		0.961733,
		0.965392,
		0.966369,
		0.966443,
		0.966369,
		0.965392,
		0.961733,
		0.953289,
		0.938108,
		0.914647,
		0.881889,
		0.839373,
		0.787175,
		0.725873,
		0.656502,
		0.580512,
		0.499726,
		0.416309,
		0.33274,
		0.251784,
		0.176476,
		0.110096,
		0.056161,
		0.018407,
		0.00078,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.007261,
		0.037212,
		0.0863,
		0.150344,
		0.225407,
		0.307804,
		0.394122,
		0.48124,
		0.56635,
		0.646984,
		0.721044,
		0.786843,
		0.843145,
		0.889215,
		0.924865,
		0.950506,
		0.967166,
		0.976471,
		0.98052,
		0.981605,
		0.981688,
		0.981605,
		0.98052,
		0.976471,
		0.967166,
		0.950506,
		0.924865,
		0.889215,
		0.843145,
		0.786843,
		0.721044,
		0.646984,
		0.56635,
		0.48124,
		0.394122,
		0.307804,
		0.225407,
		0.150344,
		0.0863,
		0.037212,
		0.007261,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.00088,
		0.020707,
		0.06298,
		0.123044,
		0.196507,
		0.279254,
		0.367463,
		0.457628,
		0.546584,
		0.631533,
		0.710079,
		0.780271,
		0.840647,
		0.890289,
		0.928881,
		0.956762,
		0.974957,
		0.985163,
		0.989623,
		0.990823,
		0.990914,
		0.990823,
		0.989623,
		0.985163,
		0.974957,
		0.956762,
		0.928881,
		0.890289,
		0.840647,
		0.780271,
		0.710079,
		0.631533,
		0.546584,
		0.457628,
		0.367463,
		0.279254,
		0.196507,
		0.123044,
		0.06298,
		0.020707,
		0.00088,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.008113,
		0.041447,
		0.095789,
		0.166253,
		0.248253,
		0.337521,
		0.430131,
		0.522525,
		0.611541,
		0.694454,
		0.769016,
		0.833512,
		0.886816,
		0.928457,
		0.958682,
		0.978498,
		0.989664,
		0.994565,
		0.995889,
		0.99599,
		0.995889,
		0.994565,
		0.989664,
		0.978498,
		0.958682,
		0.928457,
		0.886816,
		0.833512,
		0.769016,
		0.694454,
		0.611541,
		0.522525,
		0.430131,
		0.337521,
		0.248253,
		0.166253,
		0.095789,
		0.041447,
		0.008113,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.000981,
		0.022996,
		0.069697,
		0.13565,
		0.215749,
		0.305234,
		0.39971,
		0.495182,
		0.588084,
		0.67532,
		0.75431,
		0.823051,
		0.880177,
		0.925035,
		0.957757,
		0.979315,
		0.99152,
		0.996902,
		0.998363,
		0.998475,
		0.998363,
		0.996902,
		0.99152,
		0.979315,
		0.957757,
		0.925035,
		0.880177,
		0.823051,
		0.75431,
		0.67532,
		0.588084,
		0.495182,
		0.39971,
		0.305234,
		0.215749,
		0.13565,
		0.069697,
		0.022996,
		0.000981,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.009011,
		0.04587,
		0.105601,
		0.182516,
		0.2713,
		0.367036,
		0.465237,
		0.56188,
		0.653452,
		0.736995,
		0.810175,
		0.87135,
		0.919651,
		0.955071,
		0.978526,
		0.991872,
		0.997787,
		0.999399,
		0.999523,
		0.999399,
		0.997787,
		0.991872,
		0.978526,
		0.955071,
		0.919651,
		0.87135,
		0.810175,
		0.736995,
		0.653452,
		0.56188,
		0.465237,
		0.367036,
		0.2713,
		0.182516,
		0.105601,
		0.04587,
		0.009011,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.001092,
		0.025507,
		0.077003,
		0.149231,
		0.236251,
		0.332553,
		0.433094,
		0.533332,
		0.629277,
		0.71754,
		0.795406,
		0.860913,
		0.912939,
		0.951302,
		0.976845,
		0.991457,
		0.997966,
		0.999749,
		0.999887,
		0.999749,
		0.997966,
		0.991457,
		0.976845,
		0.951302,
		0.912939,
		0.860913,
		0.795406,
		0.71754,
		0.629277,
		0.533332,
		0.433094,
		0.332553,
		0.236251,
		0.149231,
		0.077003,
		0.025507,
		0.001092,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.010032,
		0.050861,
		0.116585,
		0.200548,
		0.296568,
		0.398965,
		0.502605,
		0.60295,
		0.696118,
		0.778955,
		0.849122,
		0.905199,
		0.946796,
		0.97465,
		0.990675,
		0.997853,
		0.999829,
		0.999982,
		0.999829,
		0.997853,
		0.990675,
		0.97465,
		0.946796,
		0.905199,
		0.849122,
		0.778955,
		0.696118,
		0.60295,
		0.502605,
		0.398965,
		0.296568,
		0.200548,
		0.116585,
		0.050861,
		0.010032,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469721,
		0.574454,
		0.672707,
		0.760816,
		0.836006,
		0.896504,
		0.941665,
		0.97209,
		0.989698,
		0.997631,
		0.999827,
		0.999999,
		0.999827,
		0.997631,
		0.989698,
		0.97209,
		0.941665,
		0.896504,
		0.836006,
		0.760816,
		0.672707,
		0.574454,
		0.469721,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325197,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325197,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.001559,
		0.035914,
		0.106789,
		0.203541,
		0.316383,
		0.436458,
		0.555928,
		0.668086,
		0.767497,
		0.850172,
		0.913781,
		0.957881,
		0.984125,
		0.996274,
		0.999725,
		1,
		0.999725,
		0.996274,
		0.984125,
		0.957881,
		0.913781,
		0.850172,
		0.767497,
		0.668086,
		0.555928,
		0.436458,
		0.316383,
		0.203541,
		0.106789,
		0.035914,
		0.001559,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.014374,
		0.071756,
		0.16171,
		0.273008,
		0.395453,
		0.519969,
		0.638727,
		0.745299,
		0.834852,
		0.904386,
		0.953007,
		0.982179,
		0.995792,
		0.999687,
		1,
		0.999687,
		0.995792,
		0.982179,
		0.953007,
		0.904386,
		0.834852,
		0.745299,
		0.638727,
		0.519969,
		0.395453,
		0.273008,
		0.16171,
		0.071756,
		0.014374,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.001783,
		0.04083,
		0.120607,
		0.228195,
		0.351814,
		0.480925,
		0.606372,
		0.720543,
		0.817592,
		0.893703,
		0.947416,
		0.979927,
		0.995229,
		0.999643,
		1,
		0.999643,
		0.995229,
		0.979927,
		0.947416,
		0.893703,
		0.817592,
		0.720543,
		0.606372,
		0.480925,
		0.351814,
		0.228195,
		0.120607,
		0.04083,
		0.001783,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.016464,
		0.08164,
		0.182603,
		0.3057,
		0.438658,
		0.570735,
		0.692907,
		0.798105,
		0.881522,
		0.94098,
		0.977312,
		0.994569,
		0.999591,
		1,
		0.999591,
		0.994569,
		0.977312,
		0.94098,
		0.881522,
		0.798105,
		0.692907,
		0.570735,
		0.438658,
		0.3057,
		0.182603,
		0.08164,
		0.016464,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.002059,
		0.046826,
		0.137255,
		0.25746,
		0.393097,
		0.531529,
		0.662033,
		0.776061,
		0.86759,
		0.933544,
		0.97426,
		0.993791,
		0.999529,
		1,
		0.999529,
		0.993791,
		0.97426,
		0.933544,
		0.86759,
		0.776061,
		0.662033,
		0.531529,
		0.393097,
		0.25746,
		0.137255,
		0.046826,
		0.002059,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.019046,
		0.093699,
		0.207724,
		0.344293,
		0.488488,
		0.62753,
		0.751077,
		0.851607,
		0.924919,
		0.970681,
		0.99287,
		0.999454,
		1,
		0.999454,
		0.99287,
		0.970681,
		0.924919,
		0.851607,
		0.751077,
		0.62753,
		0.488488,
		0.344293,
		0.207724,
		0.093699,
		0.019046,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.002404,
		0.05424,
		0.157547,
		0.292492,
		0.441396,
		0.588982,
		0.722712,
		0.833218,
		0.914874,
		0.966466,
		0.991772,
		0.999365,
		1,
		0.999365,
		0.991772,
		0.966466,
		0.914874,
		0.833218,
		0.722712,
		0.588982,
		0.441396,
		0.292492,
		0.157547,
		0.05424,
		0.002404,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.022285,
		0.108615,
		0.238249,
		0.39014,
		0.545959,
		0.690463,
		0.811993,
		0.903125,
		0.961473,
		0.990456,
		0.999256,
		1,
		0.999256,
		0.990456,
		0.961473,
		0.903125,
		0.811993,
		0.690463,
		0.545959,
		0.39014,
		0.238249,
		0.108615,
		0.022285,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.002844,
		0.063557,
		0.182603,
		0.334797,
		0.498046,
		0.653763,
		0.787425,
		0.889324,
		0.955526,
		0.988867,
		0.999123,
		1,
		0.999123,
		0.988867,
		0.955526,
		0.889324,
		0.787425,
		0.653763,
		0.498046,
		0.334797,
		0.182603,
		0.063557,
		0.002844,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.026425,
		0.127354,
		0.275766,
		0.444897,
		0.611985,
		0.758906,
		0.873037,
		0.948403,
		0.986936,
		0.99896,
		1,
		0.99896,
		0.986936,
		0.948403,
		0.873037,
		0.758906,
		0.611985,
		0.444897,
		0.275766,
		0.127354,
		0.026425,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.003417,
		0.075484,
		0.213992,
		0.38632,
		0.564456,
		0.725714,
		0.853729,
		0.939815,
		0.984569,
		0.998756,
		1,
		0.998756,
		0.984569,
		0.939815,
		0.853729,
		0.725714,
		0.564456,
		0.38632,
		0.213992,
		0.075484,
		0.003417,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.031834,
		0.151319,
		0.322436,
		0.510503,
		0.687003,
		0.830729,
		0.929393,
		0.981647,
		0.998499,
		1,
		0.998499,
		0.981647,
		0.929393,
		0.830729,
		0.687003,
		0.510503,
		0.322436,
		0.151319,
		0.031834,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.004182,
		0.091084,
		0.253947,
		0.449532,
		0.641793,
		0.803203,
		0.916658,
		0.978003,
		0.998174,
		1,
		0.998174,
		0.978003,
		0.916658,
		0.803203,
		0.641793,
		0.449532,
		0.253947,
		0.091084,
		0.004182,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.039085,
		0.182603,
		0.381194,
		0.588982,
		0.770107,
		0.900978,
		0.973418,
		0.997755,
		1,
		0.997755,
		0.973418,
		0.900978,
		0.770107,
		0.588982,
		0.381194,
		0.182603,
		0.039085,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.005236,
		0.112019,
		0.3057,
		0.527402,
		0.730146,
		0.881522,
		0.967586,
		0.99721,
		1,
		0.99721,
		0.967586,
		0.881522,
		0.730146,
		0.527402,
		0.3057,
		0.112019,
		0.005236,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.049119,
		0.224414,
		0.455958,
		0.681729,
		0.857183,
		0.960083,
		0.996489,
		1,
		0.996489,
		0.960083,
		0.857183,
		0.681729,
		0.455958,
		0.224414,
		0.049119,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.00674,
		0.140868,
		0.373628,
		0.622395,
		0.825756,
		0.949475,
		0.994648,
		0.999123,
		0.994648,
		0.949475,
		0.825756,
		0.622395,
		0.373628,
		0.140868,
		0.00674,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.061965,
		0.274752,
		0.537774,
		0.767694,
		0.913908,
		0.969283,
		0.974941,
		0.969283,
		0.913908,
		0.767694,
		0.537774,
		0.274752,
		0.061965,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.007571,
		0.153342,
		0.390871,
		0.61919,
		0.77265,
		0.833327,
		0.839755,
		0.833327,
		0.77265,
		0.61919,
		0.390871,
		0.153342,
		0.007571,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.046363,
		0.197077,
		0.365257,
		0.486772,
		0.537373,
		0.542966,
		0.537373,
		0.486772,
		0.365257,
		0.197077,
		0.046363,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.002816,
		0.054533,
		0.131026,
		0.192162,
		0.219306,
		0.222461,
		0.219306,
		0.192162,
		0.131026,
		0.054533,
		0.002816,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.005051,
		0.020142,
		0.034243,
		0.041042,
		0.041882,
		0.041042,
		0.034243,
		0.020142,
		0.005051,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		1.4e-005,
		0.000248,
		0.00054,
		0.000698,
		0.000719,
		0.000698,
		0.00054,
		0.000248,
		1.4e-005,
		1.4e-005,
		0.000248,
		0.00054,
		0.000698,
		0.000719,
		0.000698,
		0.00054,
		0.000248,
		1.4e-005,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		0.005051,
		0.020142,
		0.034243,
		0.041042,
		0.041882,
		0.041042,
		0.034243,
		0.020142,
		0.005051,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.002816,
		0.054533,
		0.131026,
		0.192162,
		0.219306,
		0.222461,
		0.219306,
		0.192162,
		0.131026,
		0.054533,
		0.002816,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.046363,
		0.197077,
		0.365257,
		0.486772,
		0.537373,
		0.542966,
		0.537373,
		0.486772,
		0.365257,
		0.197077,
		0.046363,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.007571,
		0.153342,
		0.390871,
		0.61919,
		0.77265,
		0.833327,
		0.839755,
		0.833327,
		0.77265,
		0.61919,
		0.390871,
		0.153342,
		0.007571,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.061965,
		0.274752,
		0.537774,
		0.767694,
		0.913908,
		0.969283,
		0.974941,
		0.969283,
		0.913908,
		0.767694,
		0.537774,
		0.274752,
		0.061965,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.00674,
		0.140868,
		0.373628,
		0.622395,
		0.825756,
		0.949475,
		0.994648,
		0.999123,
		0.994648,
		0.949475,
		0.825756,
		0.622395,
		0.373628,
		0.140868,
		0.00674,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.049119,
		0.224414,
		0.455958,
		0.681729,
		0.857183,
		0.960083,
		0.996489,
		1,
		0.996489,
		0.960083,
		0.857183,
		0.681729,
		0.455958,
		0.224414,
		0.049119,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.005236,
		0.112019,
		0.3057,
		0.527402,
		0.730146,
		0.881522,
		0.967586,
		0.99721,
		1,
		0.99721,
		0.967586,
		0.881522,
		0.730146,
		0.527402,
		0.3057,
		0.112019,
		0.005236,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.039085,
		0.182603,
		0.381194,
		0.588982,
		0.770107,
		0.900978,
		0.973418,
		0.997755,
		1,
		0.997755,
		0.973418,
		0.900978,
		0.770107,
		0.588982,
		0.381194,
		0.182603,
		0.039085,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.004182,
		0.091084,
		0.253947,
		0.449532,
		0.641793,
		0.803203,
		0.916658,
		0.978003,
		0.998174,
		1,
		0.998174,
		0.978003,
		0.916658,
		0.803203,
		0.641793,
		0.449532,
		0.253947,
		0.091084,
		0.004182,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.031834,
		0.151319,
		0.322436,
		0.510503,
		0.687003,
		0.830729,
		0.929393,
		0.981647,
		0.998499,
		1,
		0.998499,
		0.981647,
		0.929393,
		0.830729,
		0.687003,
		0.510503,
		0.322436,
		0.151319,
		0.031834,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.003417,
		0.075484,
		0.213992,
		0.38632,
		0.564456,
		0.725714,
		0.853729,
		0.939815,
		0.984569,
		0.998756,
		1,
		0.998756,
		0.984569,
		0.939815,
		0.853729,
		0.725714,
		0.564456,
		0.38632,
		0.213992,
		0.075484,
		0.003417,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.026425,
		0.127354,
		0.275766,
		0.444897,
		0.611985,
		0.758906,
		0.873037,
		0.948403,
		0.986936,
		0.99896,
		1,
		0.99896,
		0.986936,
		0.948403,
		0.873037,
		0.758906,
		0.611985,
		0.444897,
		0.275766,
		0.127354,
		0.026425,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.002844,
		0.063557,
		0.182603,
		0.334797,
		0.498046,
		0.653763,
		0.787425,
		0.889324,
		0.955526,
		0.988867,
		0.999123,
		1,
		0.999123,
		0.988867,
		0.955526,
		0.889324,
		0.787425,
		0.653763,
		0.498046,
		0.334797,
		0.182603,
		0.063557,
		0.002844,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.022285,
		0.108615,
		0.238249,
		0.39014,
		0.545959,
		0.690463,
		0.811993,
		0.903125,
		0.961473,
		0.990456,
		0.999256,
		1,
		0.999256,
		0.990456,
		0.961473,
		0.903125,
		0.811993,
		0.690463,
		0.545959,
		0.39014,
		0.238249,
		0.108615,
		0.022285,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.002404,
		0.05424,
		0.157547,
		0.292492,
		0.441396,
		0.588982,
		0.722712,
		0.833218,
		0.914874,
		0.966466,
		0.991772,
		0.999365,
		1,
		0.999365,
		0.991772,
		0.966466,
		0.914874,
		0.833218,
		0.722712,
		0.588982,
		0.441396,
		0.292492,
		0.157547,
		0.05424,
		0.002404,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.019046,
		0.093699,
		0.207724,
		0.344293,
		0.488488,
		0.62753,
		0.751077,
		0.851607,
		0.924919,
		0.970681,
		0.99287,
		0.999454,
		1,
		0.999454,
		0.99287,
		0.970681,
		0.924919,
		0.851607,
		0.751077,
		0.62753,
		0.488488,
		0.344293,
		0.207724,
		0.093699,
		0.019046,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.002059,
		0.046826,
		0.137255,
		0.25746,
		0.393097,
		0.531529,
		0.662033,
		0.776061,
		0.86759,
		0.933544,
		0.97426,
		0.993791,
		0.999529,
		1,
		0.999529,
		0.993791,
		0.97426,
		0.933544,
		0.86759,
		0.776061,
		0.662033,
		0.531529,
		0.393097,
		0.25746,
		0.137255,
		0.046826,
		0.002059,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.016464,
		0.08164,
		0.182603,
		0.3057,
		0.438658,
		0.570735,
		0.692907,
		0.798105,
		0.881522,
		0.94098,
		0.977312,
		0.994569,
		0.999591,
		1,
		0.999591,
		0.994569,
		0.977312,
		0.94098,
		0.881522,
		0.798105,
		0.692907,
		0.570735,
		0.438658,
		0.3057,
		0.182603,
		0.08164,
		0.016464,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.001783,
		0.04083,
		0.120607,
		0.228195,
		0.351814,
		0.480925,
		0.606372,
		0.720543,
		0.817592,
		0.893703,
		0.947416,
		0.979927,
		0.995229,
		0.999643,
		1,
		0.999643,
		0.995229,
		0.979927,
		0.947416,
		0.893703,
		0.817592,
		0.720543,
		0.606372,
		0.480925,
		0.351814,
		0.228195,
		0.120607,
		0.04083,
		0.001783,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.014374,
		0.071756,
		0.16171,
		0.273008,
		0.395453,
		0.519969,
		0.638727,
		0.745299,
		0.834852,
		0.904386,
		0.953007,
		0.982179,
		0.995792,
		0.999687,
		1,
		0.999687,
		0.995792,
		0.982179,
		0.953007,
		0.904386,
		0.834852,
		0.745299,
		0.638727,
		0.519969,
		0.395453,
		0.273008,
		0.16171,
		0.071756,
		0.014374,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.001559,
		0.035914,
		0.106789,
		0.203541,
		0.316383,
		0.436458,
		0.555928,
		0.668086,
		0.767497,
		0.850172,
		0.913781,
		0.957881,
		0.984125,
		0.996274,
		0.999725,
		1,
		0.999725,
		0.996274,
		0.984125,
		0.957881,
		0.913781,
		0.850172,
		0.767497,
		0.668086,
		0.555928,
		0.436458,
		0.316383,
		0.203541,
		0.106789,
		0.035914,
		0.001559,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325197,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325197,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469721,
		0.574454,
		0.672707,
		0.760816,
		0.836006,
		0.896504,
		0.941665,
		0.97209,
		0.989698,
		0.997631,
		0.999827,
		0.999999,
		0.999827,
		0.997631,
		0.989698,
		0.97209,
		0.941665,
		0.896504,
		0.836006,
		0.760816,
		0.672707,
		0.574454,
		0.469721,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.010032,
		0.050861,
		0.116585,
		0.200548,
		0.296568,
		0.398965,
		0.502605,
		0.60295,
		0.696118,
		0.778955,
		0.849122,
		0.905199,
		0.946796,
		0.97465,
		0.990675,
		0.997853,
		0.999829,
		0.999982,
		0.999829,
		0.997853,
		0.990675,
		0.97465,
		0.946796,
		0.905199,
		0.849122,
		0.778955,
		0.696118,
		0.60295,
		0.502605,
		0.398965,
		0.296568,
		0.200548,
		0.116585,
		0.050861,
		0.010032,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.001092,
		0.025507,
		0.077003,
		0.149231,
		0.236251,
		0.332553,
		0.433094,
		0.533332,
		0.629277,
		0.71754,
		0.795406,
		0.860913,
		0.912939,
		0.951302,
		0.976845,
		0.991457,
		0.997966,
		0.999749,
		0.999887,
		0.999749,
		0.997966,
		0.991457,
		0.976845,
		0.951302,
		0.912939,
		0.860913,
		0.795406,
		0.71754,
		0.629277,
		0.533332,
		0.433094,
		0.332553,
		0.236251,
		0.149231,
		0.077003,
		0.025507,
		0.001092,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.009011,
		0.04587,
		0.105601,
		0.182516,
		0.2713,
		0.367036,
		0.465237,
		0.56188,
		0.653452,
		0.736995,
		0.810175,
		0.87135,
		0.919651,
		0.955071,
		0.978526,
		0.991872,
		0.997787,
		0.999399,
		0.999523,
		0.999399,
		0.997787,
		0.991872,
		0.978526,
		0.955071,
		0.919651,
		0.87135,
		0.810175,
		0.736995,
		0.653452,
		0.56188,
		0.465237,
		0.367036,
		0.2713,
		0.182516,
		0.105601,
		0.04587,
		0.009011,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.000981,
		0.022996,
		0.069697,
		0.13565,
		0.215749,
		0.305234,
		0.39971,
		0.495182,
		0.588084,
		0.67532,
		0.75431,
		0.823051,
		0.880177,
		0.925035,
		0.957757,
		0.979315,
		0.99152,
		0.996902,
		0.998363,
		0.998475,
		0.998363,
		0.996902,
		0.99152,
		0.979315,
		0.957757,
		0.925035,
		0.880177,
		0.823051,
		0.75431,
		0.67532,
		0.588084,
		0.495182,
		0.39971,
		0.305234,
		0.215749,
		0.13565,
		0.069697,
		0.022996,
		0.000981,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.008113,
		0.041447,
		0.095789,
		0.166253,
		0.248253,
		0.337521,
		0.430131,
		0.522525,
		0.611541,
		0.694454,
		0.769016,
		0.833512,
		0.886816,
		0.928457,
		0.958682,
		0.978498,
		0.989664,
		0.994565,
		0.995889,
		0.99599,
		0.995889,
		0.994565,
		0.989664,
		0.978498,
		0.958682,
		0.928457,
		0.886816,
		0.833512,
		0.769016,
		0.694454,
		0.611541,
		0.522525,
		0.430131,
		0.337521,
		0.248253,
		0.166253,
		0.095789,
		0.041447,
		0.008113,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.00088,
		0.020707,
		0.06298,
		0.123044,
		0.196507,
		0.279254,
		0.367463,
		0.457628,
		0.546584,
		0.631533,
		0.710079,
		0.780271,
		0.840647,
		0.890289,
		0.928881,
		0.956762,
		0.974957,
		0.985163,
		0.989623,
		0.990823,
		0.990914,
		0.990823,
		0.989623,
		0.985163,
		0.974957,
		0.956762,
		0.928881,
		0.890289,
		0.840647,
		0.780271,
		0.710079,
		0.631533,
		0.546584,
		0.457628,
		0.367463,
		0.279254,
		0.196507,
		0.123044,
		0.06298,
		0.020707,
		0.00088,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.007261,
		0.037212,
		0.0863,
		0.150344,
		0.225407,
		0.307804,
		0.394122,
		0.48124,
		0.56635,
		0.646984,
		0.721044,
		0.786843,
		0.843145,
		0.889215,
		0.924865,
		0.950506,
		0.967166,
		0.976471,
		0.98052,
		0.981605,
		0.981688,
		0.981605,
		0.98052,
		0.976471,
		0.967166,
		0.950506,
		0.924865,
		0.889215,
		0.843145,
		0.786843,
		0.721044,
		0.646984,
		0.56635,
		0.48124,
		0.394122,
		0.307804,
		0.225407,
		0.150344,
		0.0863,
		0.037212,
		0.007261,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.00078,
		0.018407,
		0.056161,
		0.110096,
		0.176476,
		0.251784,
		0.33274,
		0.416309,
		0.499726,
		0.580512,
		0.656502,
		0.725873,
		0.787175,
		0.839373,
		0.881889,
		0.914647,
		0.938108,
		0.953289,
		0.961733,
		0.965392,
		0.966369,
		0.966443,
		0.966369,
		0.965392,
		0.961733,
		0.953289,
		0.938108,
		0.914647,
		0.881889,
		0.839373,
		0.787175,
		0.725873,
		0.656502,
		0.580512,
		0.499726,
		0.416309,
		0.33274,
		0.251784,
		0.176476,
		0.110096,
		0.056161,
		0.018407,
		0.00078,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		0.006363,
		0.032703,
		0.076079,
		0.132984,
		0.200101,
		0.274314,
		0.352718,
		0.43264,
		0.511645,
		0.587564,
		0.658512,
		0.722916,
		0.779544,
		0.827542,
		0.86647,
		0.896339,
		0.917645,
		0.931376,
		0.938984,
		0.942269,
		0.943143,
		0.943208,
		0.943143,
		0.942269,
		0.938984,
		0.931376,
		0.917645,
		0.896339,
		0.86647,
		0.827542,
		0.779544,
		0.722916,
		0.658512,
		0.587564,
		0.511645,
		0.43264,
		0.352718,
		0.274314,
		0.200101,
		0.132984,
		0.076079,
		0.032703,
		0.006363,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.000671,
		0.015865,
		0.048543,
		0.095456,
		0.153515,
		0.219804,
		0.291589,
		0.366327,
		0.441681,
		0.515532,
		0.586001,
		0.651464,
		0.71058,
		0.762317,
		0.80598,
		0.841247,
		0.868202,
		0.887355,
		0.899653,
		0.906441,
		0.909361,
		0.910135,
		0.910193,
		0.910135,
		0.909361,
		0.906441,
		0.899653,
		0.887355,
		0.868202,
		0.841247,
		0.80598,
		0.762317,
		0.71058,
		0.651464,
		0.586001,
		0.515532,
		0.441681,
		0.366327,
		0.291589,
		0.219804,
		0.153515,
		0.095456,
		0.048543,
		0.015865,
		0.000671,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		0.00535,
		0.027572,
		0.064324,
		0.112777,
		0.170248,
		0.234205,
		0.302277,
		0.372267,
		0.442157,
		0.510129,
		0.574574,
		0.634115,
		0.687623,
		0.734247,
		0.773435,
		0.804966,
		0.828975,
		0.845973,
		0.856847,
		0.862827,
		0.865392,
		0.866069,
		0.86612,
		0.866069,
		0.865392,
		0.862827,
		0.856847,
		0.845973,
		0.828975,
		0.804966,
		0.773435,
		0.734247,
		0.687623,
		0.634115,
		0.574574,
		0.510129,
		0.442157,
		0.372267,
		0.302277,
		0.234205,
		0.170248,
		0.112777,
		0.064324,
		0.027572,
		0.00535,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.000547,
		0.012978,
		0.039813,
		0.078506,
		0.126633,
		0.181895,
		0.242128,
		0.30531,
		0.369568,
		0.433192,
		0.494641,
		0.552562,
		0.605806,
		0.653439,
		0.694774,
		0.729382,
		0.757127,
		0.778177,
		0.793028,
		0.802496,
		0.807686,
		0.809903,
		0.810487,
		0.810531,
		0.810487,
		0.809903,
		0.807686,
		0.802496,
		0.793028,
		0.778177,
		0.757127,
		0.729382,
		0.694774,
		0.653439,
		0.605806,
		0.552562,
		0.494641,
		0.433192,
		0.369568,
		0.30531,
		0.242128,
		0.181895,
		0.126633,
		0.078506,
		0.039813,
		0.012978,
		0.000547,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		0.004224,
		0.021821,
		0.051039,
		0.089732,
		0.135857,
		0.187482,
		0.242791,
		0.300088,
		0.357808,
		0.414524,
		0.468959,
		0.519995,
		0.566689,
		0.608287,
		0.644244,
		0.674241,
		0.698204,
		0.716324,
		0.729065,
		0.73716,
		0.741583,
		0.743467,
		0.743962,
		0.743999,
		0.743962,
		0.743467,
		0.741583,
		0.73716,
		0.729065,
		0.716324,
		0.698204,
		0.674241,
		0.644244,
		0.608287,
		0.566689,
		0.519995,
		0.468959,
		0.414524,
		0.357808,
		0.300088,
		0.242791,
		0.187482,
		0.135857,
		0.089732,
		0.051039,
		0.021821,
		0.004224,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.000415,
		0.009861,
		0.030325,
		0.059949,
		0.096962,
		0.13968,
		0.186509,
		0.235957,
		0.286632,
		0.337251,
		0.386651,
		0.433793,
		0.477773,
		0.517835,
		0.553385,
		0.584001,
		0.609454,
		0.629719,
		0.644993,
		0.655698,
		0.662478,
		0.666171,
		0.66774,
		0.66815,
		0.668181,
		0.66815,
		0.66774,
		0.666171,
		0.662478,
		0.655698,
		0.644993,
		0.629719,
		0.609454,
		0.584001,
		0.553385,
		0.517835,
		0.477773,
		0.433793,
		0.386651,
		0.337251,
		0.286632,
		0.235957,
		0.186509,
		0.13968,
		0.096962,
		0.059949,
		0.030325,
		0.009861,
		0.000415,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		0.003067,
		0.015878,
		0.037225,
		0.065609,
		0.0996,
		0.13784,
		0.179047,
		0.222021,
		0.265647,
		0.308897,
		0.350845,
		0.390664,
		0.427644,
		0.461192,
		0.49085,
		0.516304,
		0.537395,
		0.554134,
		0.566711,
		0.575499,
		0.581049,
		0.584062,
		0.585338,
		0.585671,
		0.585696,
		0.585671,
		0.585338,
		0.584062,
		0.581049,
		0.575499,
		0.566711,
		0.554134,
		0.537395,
		0.516304,
		0.49085,
		0.461192,
		0.427644,
		0.390664,
		0.350845,
		0.308897,
		0.265647,
		0.222021,
		0.179047,
		0.13784,
		0.0996,
		0.065609,
		0.037225,
		0.015878,
		0.003067,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.000286,
		0.006822,
		0.021023,
		0.041656,
		0.067541,
		0.097552,
		0.130624,
		0.16575,
		0.20199,
		0.238473,
		0.274398,
		0.309044,
		0.341773,
		0.372039,
		0.399391,
		0.423487,
		0.444099,
		0.461124,
		0.474596,
		0.484687,
		0.491718,
		0.496144,
		0.498541,
		0.499553,
		0.499817,
		0.499836,
		0.499817,
		0.499553,
		0.498541,
		0.496144,
		0.491718,
		0.484687,
		0.474596,
		0.461124,
		0.444099,
		0.423487,
		0.399391,
		0.372039,
		0.341773,
		0.309044,
		0.274398,
		0.238473,
		0.20199,
		0.16575,
		0.130624,
		0.097552,
		0.067541,
		0.041656,
		0.021023,
		0.006822,
		0.000286,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		0.002007,
		0.010409,
		0.024457,
		0.043203,
		0.065746,
		0.091224,
		0.118823,
		0.147776,
		0.177368,
		0.206935,
		0.235873,
		0.263636,
		0.289746,
		0.313795,
		0.33545,
		0.354464,
		0.370678,
		0.384031,
		0.394565,
		0.402434,
		0.4079,
		0.411333,
		0.413186,
		0.413966,
		0.414169,
		0.414184,
		0.414169,
		0.413966,
		0.413186,
		0.411333,
		0.4079,
		0.402434,
		0.394565,
		0.384031,
		0.370678,
		0.354464,
		0.33545,
		0.313795,
		0.289746,
		0.263636,
		0.235873,
		0.206935,
		0.177368,
		0.147776,
		0.118823,
		0.091224,
		0.065746,
		0.043203,
		0.024457,
		0.010409,
		0.002007,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.000176,
		0.004205,
		0.012984,
		0.025783,
		0.041898,
		0.060661,
		0.081434,
		0.103613,
		0.126633,
		0.149966,
		0.173123,
		0.195661,
		0.217181,
		0.237336,
		0.25583,
		0.272428,
		0.286955,
		0.299306,
		0.309448,
		0.317427,
		0.323371,
		0.327489,
		0.330067,
		0.331456,
		0.332039,
		0.33219,
		0.332201,
		0.33219,
		0.332039,
		0.331456,
		0.330067,
		0.327489,
		0.323371,
		0.317427,
		0.309448,
		0.299306,
		0.286955,
		0.272428,
		0.25583,
		0.237336,
		0.217181,
		0.195661,
		0.173123,
		0.149966,
		0.126633,
		0.103613,
		0.081434,
		0.060661,
		0.041898,
		0.025783,
		0.012984,
		0.004205,
		0.000176,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		0.001155,
		0.006001,
		0.014128,
		0.025011,
		0.038146,
		0.053053,
		0.069278,
		0.08639,
		0.103984,
		0.121685,
		0.139147,
		0.156055,
		0.172129,
		0.187125,
		0.200837,
		0.213104,
		0.223808,
		0.232882,
		0.240313,
		0.246144,
		0.250476,
		0.253469,
		0.255338,
		0.256343,
		0.256763,
		0.256872,
		0.25688,
		0.256872,
		0.256763,
		0.256343,
		0.255338,
		0.253469,
		0.250476,
		0.246144,
		0.240313,
		0.232882,
		0.223808,
		0.213104,
		0.200837,
		0.187125,
		0.172129,
		0.156055,
		0.139147,
		0.121685,
		0.103984,
		0.08639,
		0.069278,
		0.053053,
		0.038146,
		0.025011,
		0.014128,
		0.006001,
		0.001155,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		9.4e-005,
		0.002242,
		0.006937,
		0.013802,
		0.022475,
		0.032612,
		0.043882,
		0.055972,
		0.068588,
		0.081453,
		0.094309,
		0.106923,
		0.119079,
		0.130588,
		0.141286,
		0.151036,
		0.159732,
		0.167298,
		0.173695,
		0.17892,
		0.183009,
		0.186039,
		0.188127,
		0.189429,
		0.190126,
		0.190417,
		0.190492,
		0.190498,
		0.190492,
		0.190417,
		0.190126,
		0.189429,
		0.188127,
		0.186039,
		0.183009,
		0.17892,
		0.173695,
		0.167298,
		0.159732,
		0.151036,
		0.141286,
		0.130588,
		0.119079,
		0.106923,
		0.094309,
		0.081453,
		0.068588,
		0.055972,
		0.043882,
		0.032612,
		0.022475,
		0.013802,
		0.006937,
		0.002242,
		9.4e-005,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		0.000562,
		0.002929,
		0.006907,
		0.01225,
		0.018722,
		0.026096,
		0.034155,
		0.042695,
		0.051523,
		0.06046,
		0.069338,
		0.078003,
		0.086318,
		0.094161,
		0.101426,
		0.108027,
		0.113897,
		0.118991,
		0.123286,
		0.126786,
		0.129518,
		0.131537,
		0.132926,
		0.133789,
		0.13425,
		0.134442,
		0.134492,
		0.134496,
		0.134492,
		0.134442,
		0.13425,
		0.133789,
		0.132926,
		0.131537,
		0.129518,
		0.126786,
		0.123286,
		0.118991,
		0.113897,
		0.108027,
		0.101426,
		0.094161,
		0.086318,
		0.078003,
		0.069338,
		0.06046,
		0.051523,
		0.042695,
		0.034155,
		0.026096,
		0.018722,
		0.01225,
		0.006907,
		0.002929,
		0.000562,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		4.1e-005,
		0.000982,
		0.003043,
		0.006065,
		0.009895,
		0.014386,
		0.019399,
		0.0248,
		0.030463,
		0.036269,
		0.042108,
		0.047876,
		0.05348,
		0.058836,
		0.06387,
		0.068519,
		0.072729,
		0.076464,
		0.079696,
		0.082414,
		0.084624,
		0.086344,
		0.087614,
		0.088484,
		0.089024,
		0.089312,
		0.089431,
		0.089462,
		0.089465,
		0.089462,
		0.089431,
		0.089312,
		0.089024,
		0.088484,
		0.087614,
		0.086344,
		0.084624,
		0.082414,
		0.079696,
		0.076464,
		0.072729,
		0.068519,
		0.06387,
		0.058836,
		0.05348,
		0.047876,
		0.042108,
		0.036269,
		0.030463,
		0.0248,
		0.019399,
		0.014386,
		0.009895,
		0.006065,
		0.003043,
		0.000982,
		4.1e-005,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		0.000215,
		0.001123,
		0.002653,
		0.004715,
		0.007219,
		0.010082,
		0.013224,
		0.016567,
		0.02004,
		0.023575,
		0.027109,
		0.030582,
		0.033943,
		0.037142,
		0.040139,
		0.042898,
		0.045391,
		0.047595,
		0.049499,
		0.051096,
		0.052391,
		0.053397,
		0.054138,
		0.054644,
		0.054958,
		0.055125,
		0.055194,
		0.055212,
		0.055213,
		0.055212,
		0.055194,
		0.055125,
		0.054958,
		0.054644,
		0.054138,
		0.053397,
		0.052391,
		0.051096,
		0.049499,
		0.047595,
		0.045391,
		0.042898,
		0.040139,
		0.037142,
		0.033943,
		0.030582,
		0.027109,
		0.023575,
		0.02004,
		0.016567,
		0.013224,
		0.010082,
		0.007219,
		0.004715,
		0.002653,
		0.001123,
		0.000215,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		1.3e-005,
		0.000317,
		0.000983,
		0.001962,
		0.003208,
		0.004672,
		0.006313,
		0.008087,
		0.009956,
		0.011881,
		0.013828,
		0.015763,
		0.017657,
		0.019481,
		0.021213,
		0.02283,
		0.024314,
		0.025651,
		0.02683,
		0.027847,
		0.028697,
		0.029385,
		0.029919,
		0.030311,
		0.030578,
		0.030743,
		0.030831,
		0.030867,
		0.030877,
		0.030877,
		0.030877,
		0.030867,
		0.030831,
		0.030743,
		0.030578,
		0.030311,
		0.029919,
		0.029385,
		0.028697,
		0.027847,
		0.02683,
		0.025651,
		0.024314,
		0.02283,
		0.021213,
		0.019481,
		0.017657,
		0.015763,
		0.013828,
		0.011881,
		0.009956,
		0.008087,
		0.006313,
		0.004672,
		0.003208,
		0.001962,
		0.000983,
		0.000317,
		1.3e-005,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		5.5e-005,
		0.000287,
		0.000679,
		0.001208,
		0.001853,
		0.002592,
		0.003407,
		0.004277,
		0.005185,
		0.006113,
		0.007047,
		0.00797,
		0.00887,
		0.009733,
		0.01055,
		0.011311,
		0.012007,
		0.012633,
		0.013183,
		0.013657,
		0.014052,
		0.014371,
		0.014618,
		0.014799,
		0.014922,
		0.014998,
		0.015038,
		0.015055,
		0.015059,
		0.015059,
		0.015059,
		0.015055,
		0.015038,
		0.014998,
		0.014922,
		0.014799,
		0.014618,
		0.014371,
		0.014052,
		0.013657,
		0.013183,
		0.012633,
		0.012007,
		0.011311,
		0.01055,
		0.009733,
		0.00887,
		0.00797,
		0.007047,
		0.006113,
		0.005185,
		0.004277,
		0.003407,
		0.002592,
		0.001853,
		0.001208,
		0.000679,
		0.000287,
		5.5e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		5.7e-005,
		0.000179,
		0.000357,
		0.000585,
		0.000854,
		0.001156,
		0.001483,
		0.00183,
		0.002188,
		0.002552,
		0.002916,
		0.003275,
		0.003623,
		0.003956,
		0.00427,
		0.004562,
		0.004828,
		0.005066,
		0.005276,
		0.005456,
		0.005605,
		0.005726,
		0.005819,
		0.005887,
		0.005933,
		0.005962,
		0.005977,
		0.005983,
		0.005985,
		0.005985,
		0.005985,
		0.005983,
		0.005977,
		0.005962,
		0.005933,
		0.005887,
		0.005819,
		0.005726,
		0.005605,
		0.005456,
		0.005276,
		0.005066,
		0.004828,
		0.004562,
		0.00427,
		0.003956,
		0.003623,
		0.003275,
		0.002916,
		0.002552,
		0.002188,
		0.00183,
		0.001483,
		0.001156,
		0.000854,
		0.000585,
		0.000357,
		0.000179,
		5.7e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		3e-005,
		7.2e-005,
		0.000127,
		0.000196,
		0.000274,
		0.000361,
		0.000454,
		0.000552,
		0.000652,
		0.000754,
		0.000854,
		0.000953,
		0.001049,
		0.00114,
		0.001225,
		0.001305,
		0.001377,
		0.001442,
		0.001498,
		0.001547,
		0.001587,
		0.001619,
		0.001644,
		0.001662,
		0.001675,
		0.001682,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001682,
		0.001675,
		0.001662,
		0.001644,
		0.001619,
		0.001587,
		0.001547,
		0.001498,
		0.001442,
		0.001377,
		0.001305,
		0.001225,
		0.00114,
		0.001049,
		0.000953,
		0.000854,
		0.000754,
		0.000652,
		0.000552,
		0.000454,
		0.000361,
		0.000274,
		0.000196,
		0.000127,
		7.2e-005,
		3e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.3e-005,
		2.2e-005,
		3.2e-005,
		4.3e-005,
		5.6e-005,
		6.9e-005,
		8.2e-005,
		9.6e-005,
		0.00011,
		0.000124,
		0.000138,
		0.000151,
		0.000163,
		0.000175,
		0.000185,
		0.000195,
		0.000204,
		0.000211,
		0.000218,
		0.000223,
		0.000228,
		0.000231,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.000231,
		0.000228,
		0.000223,
		0.000218,
		0.000211,
		0.000204,
		0.000195,
		0.000185,
		0.000175,
		0.000163,
		0.000151,
		0.000138,
		0.000124,
		0.00011,
		9.6e-005,
		8.2e-005,
		6.9e-005,
		5.6e-005,
		4.3e-005,
		3.2e-005,
		2.2e-005,
		1.3e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		1.5e-005,
		2.4e-005,
		3.5e-005,
		4.6e-005,
		5.9e-005,
		7.2e-005,
		8.6e-005,
		0.0001,
		0.000114,
		0.000127,
		0.000141,
		0.000154,
		0.000166,
		0.000177,
		0.000188,
		0.000197,
		0.000206,
		0.000213,
		0.000219,
		0.000224,
		0.000228,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.00023,
		0.000227,
		0.000222,
		0.000216,
		0.00021,
		0.000202,
		0.000193,
		0.000183,
		0.000172,
		0.00016,
		0.000147,
		0.000134,
		0.000121,
		0.000107,
		9.3e-005,
		7.9e-005,
		6.5e-005,
		5.2e-005,
		4e-005,
		2.9e-005,
		2e-005,
		1.1e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		2.2e-005,
		6e-005,
		0.000112,
		0.000178,
		0.000254,
		0.000339,
		0.000431,
		0.000527,
		0.000627,
		0.000728,
		0.000829,
		0.000929,
		0.001025,
		0.001117,
		0.001205,
		0.001285,
		0.00136,
		0.001426,
		0.001485,
		0.001535,
		0.001578,
		0.001612,
		0.001639,
		0.001658,
		0.001672,
		0.001681,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001684,
		0.001677,
		0.001666,
		0.001649,
		0.001626,
		0.001596,
		0.001557,
		0.001511,
		0.001456,
		0.001394,
		0.001323,
		0.001246,
		0.001162,
		0.001072,
		0.000977,
		0.000879,
		0.000779,
		0.000678,
		0.000577,
		0.000479,
		0.000384,
		0.000295,
		0.000215,
		0.000143,
		8.4e-005,
		3.9e-005,
		1e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		8.2e-005,
		0.000218,
		0.00041,
		0.000649,
		0.000926,
		0.001235,
		0.001568,
		0.001918,
		0.002279,
		0.002644,
		0.003007,
		0.003363,
		0.003708,
		0.004037,
		0.004345,
		0.004631,
		0.00489,
		0.005121,
		0.005324,
		0.005496,
		0.005638,
		0.005752,
		0.005838,
		0.005901,
		0.005942,
		0.005967,
		0.005979,
		0.005984,
		0.005985,
		0.005985,
		0.005985,
		0.005982,
		0.005974,
		0.005956,
		0.005924,
		0.005872,
		0.005798,
		0.005698,
		0.005571,
		0.005413,
		0.005226,
		0.005009,
		0.004764,
		0.004491,
		0.004194,
		0.003874,
		0.003537,
		0.003186,
		0.002826,
		0.002461,
		0.002098,
		0.001742,
		0.001399,
		0.001077,
		0.000783,
		0.000524,
		0.000308,
		0.000143,
		3.7e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		2.5e-005,
		0.000213,
		0.000567,
		0.001064,
		0.001682,
		0.0024,
		0.003197,
		0.004055,
		0.004955,
		0.00588,
		0.006814,
		0.007741,
		0.008648,
		0.009521,
		0.010351,
		0.011126,
		0.011839,
		0.012483,
		0.013053,
		0.013546,
		0.01396,
		0.014298,
		0.014563,
		0.014759,
		0.014896,
		0.014983,
		0.015031,
		0.015052,
		0.015059,
		0.015059,
		0.015059,
		0.015057,
		0.015044,
		0.015011,
		0.014945,
		0.014835,
		0.014669,
		0.014439,
		0.014139,
		0.013763,
		0.013309,
		0.012777,
		0.01217,
		0.011491,
		0.010746,
		0.009942,
		0.009089,
		0.008198,
		0.007279,
		0.006347,
		0.005416,
		0.004501,
		0.00362,
		0.00279,
		0.00203,
		0.001359,
		0.000799,
		0.000371,
		9.7e-005,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		5.2e-005,
		0.000451,
		0.0012,
		0.002251,
		0.003555,
		0.005068,
		0.006746,
		0.008547,
		0.010433,
		0.012367,
		0.014314,
		0.016241,
		0.01812,
		0.019924,
		0.021629,
		0.023214,
		0.024662,
		0.025961,
		0.0271,
		0.028075,
		0.028884,
		0.029533,
		0.03003,
		0.030388,
		0.030628,
		0.030771,
		0.030844,
		0.030871,
		0.030877,
		0.030877,
		0.030876,
		0.030862,
		0.030815,
		0.03071,
		0.030522,
		0.030225,
		0.029799,
		0.029228,
		0.0285,
		0.027608,
		0.026551,
		0.025331,
		0.023956,
		0.022437,
		0.02079,
		0.019033,
		0.017189,
		0.015282,
		0.013341,
		0.011397,
		0.009482,
		0.007633,
		0.005889,
		0.004288,
		0.002874,
		0.001691,
		0.000785,
		0.000205,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		9.7e-005,
		0.000834,
		0.002217,
		0.004154,
		0.006556,
		0.009337,
		0.012416,
		0.015716,
		0.019164,
		0.022689,
		0.026229,
		0.029723,
		0.033116,
		0.03636,
		0.039411,
		0.042233,
		0.044794,
		0.047072,
		0.049052,
		0.050725,
		0.052095,
		0.053172,
		0.053976,
		0.054538,
		0.054895,
		0.055094,
		0.055183,
		0.05521,
		0.055213,
		0.055213,
		0.055202,
		0.055149,
		0.055011,
		0.054739,
		0.054285,
		0.053606,
		0.052669,
		0.051448,
		0.049927,
		0.0481,
		0.04597,
		0.043547,
		0.040853,
		0.037912,
		0.034759,
		0.031435,
		0.027985,
		0.024461,
		0.020921,
		0.017426,
		0.014044,
		0.010844,
		0.007904,
		0.005302,
		0.003122,
		0.001451,
		0.000379,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		0.000162,
		0.001398,
		0.003714,
		0.006952,
		0.010961,
		0.015596,
		0.020718,
		0.026196,
		0.031906,
		0.03773,
		0.04356,
		0.049296,
		0.054846,
		0.060128,
		0.065071,
		0.069614,
		0.073709,
		0.077319,
		0.080424,
		0.083014,
		0.085099,
		0.086702,
		0.087866,
		0.088647,
		0.089116,
		0.089354,
		0.089445,
		0.089464,
		0.089465,
		0.089459,
		0.089413,
		0.089259,
		0.088916,
		0.0883,
		0.087336,
		0.085958,
		0.084118,
		0.081783,
		0.078936,
		0.075576,
		0.07172,
		0.067396,
		0.062646,
		0.057525,
		0.052099,
		0.046446,
		0.040651,
		0.034811,
		0.02903,
		0.023421,
		0.018105,
		0.013209,
		0.008869,
		0.005227,
		0.002431,
		0.000635,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000253,
		0.002176,
		0.005775,
		0.0108,
		0.017011,
		0.02418,
		0.032087,
		0.040526,
		0.049299,
		0.058225,
		0.067132,
		0.075864,
		0.084279,
		0.092251,
		0.099669,
		0.106443,
		0.112501,
		0.117792,
		0.122288,
		0.125984,
		0.128904,
		0.131095,
		0.132632,
		0.133616,
		0.134165,
		0.134412,
		0.134487,
		0.134495,
		0.134494,
		0.134464,
		0.134318,
		0.133936,
		0.133186,
		0.13194,
		0.130086,
		0.127539,
		0.124235,
		0.12014,
		0.115245,
		0.109565,
		0.103142,
		0.096035,
		0.088328,
		0.08012,
		0.07153,
		0.062691,
		0.053754,
		0.044882,
		0.036252,
		0.028054,
		0.02049,
		0.013772,
		0.008124,
		0.003782,
		0.000989,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.000371,
		0.003192,
		0.008464,
		0.015814,
		0.024886,
		0.035336,
		0.046839,
		0.059088,
		0.071791,
		0.084677,
		0.097494,
		0.110012,
		0.122023,
		0.133344,
		0.143817,
		0.153312,
		0.161731,
		0.169008,
		0.17511,
		0.180046,
		0.183862,
		0.186644,
		0.18852,
		0.189652,
		0.190229,
		0.19045,
		0.190496,
		0.190498,
		0.190485,
		0.190372,
		0.189997,
		0.189167,
		0.187686,
		0.185375,
		0.18209,
		0.177722,
		0.172206,
		0.165515,
		0.157661,
		0.148694,
		0.138695,
		0.12778,
		0.116093,
		0.103804,
		0.091109,
		0.078227,
		0.0654,
		0.052889,
		0.040976,
		0.029959,
		0.020157,
		0.011902,
		0.005545,
		0.001451,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000519,
		0.004461,
		0.011818,
		0.02206,
		0.034677,
		0.049186,
		0.065122,
		0.082051,
		0.099561,
		0.117269,
		0.134822,
		0.151896,
		0.168203,
		0.183489,
		0.197539,
		0.21018,
		0.221283,
		0.230768,
		0.238608,
		0.244832,
		0.249525,
		0.252835,
		0.254963,
		0.256157,
		0.256697,
		0.256861,
		0.25688,
		0.256878,
		0.25681,
		0.256491,
		0.25566,
		0.254032,
		0.251342,
		0.247362,
		0.241918,
		0.234893,
		0.22623,
		0.21593,
		0.204045,
		0.19068,
		0.175988,
		0.160162,
		0.143438,
		0.126086,
		0.108413,
		0.090758,
		0.073489,
		0.057003,
		0.041725,
		0.028102,
		0.016611,
		0.007747,
		0.002029,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.000697,
		0.005983,
		0.015835,
		0.029527,
		0.046366,
		0.06569,
		0.086869,
		0.109311,
		0.132457,
		0.155789,
		0.178832,
		0.201151,
		0.22236,
		0.242125,
		0.260166,
		0.276259,
		0.29025,
		0.302048,
		0.311642,
		0.319097,
		0.324562,
		0.328265,
		0.330511,
		0.331662,
		0.332104,
		0.332198,
		0.332201,
		0.332175,
		0.331948,
		0.331199,
		0.329549,
		0.326616,
		0.322066,
		0.31563,
		0.307118,
		0.296425,
		0.283525,
		0.268467,
		0.251376,
		0.232442,
		0.211916,
		0.190107,
		0.167375,
		0.144131,
		0.120829,
		0.097968,
		0.076086,
		0.055758,
		0.037596,
		0.022245,
		0.010384,
		0.002723,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.000902,
		0.007742,
		0.020468,
		0.038126,
		0.059802,
		0.084624,
		0.111767,
		0.140451,
		0.169947,
		0.19958,
		0.228728,
		0.256833,
		0.283397,
		0.307995,
		0.330276,
		0.349969,
		0.366892,
		0.38096,
		0.392189,
		0.400704,
		0.406741,
		0.410642,
		0.412843,
		0.413844,
		0.414148,
		0.414184,
		0.414179,
		0.414053,
		0.413461,
		0.411925,
		0.408933,
		0.404013,
		0.396774,
		0.386925,
		0.374285,
		0.358785,
		0.340459,
		0.319444,
		0.295966,
		0.270336,
		0.242943,
		0.214251,
		0.184787,
		0.155141,
		0.125963,
		0.097956,
		0.071873,
		0.048517,
		0.028739,
		0.01343,
		0.003525,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.001132,
		0.009702,
		0.025625,
		0.04768,
		0.074699,
		0.105574,
		0.139251,
		0.174741,
		0.211121,
		0.247536,
		0.283205,
		0.317429,
		0.34959,
		0.379165,
		0.405731,
		0.428973,
		0.448693,
		0.464821,
		0.477426,
		0.486716,
		0.493048,
		0.496909,
		0.498898,
		0.499666,
		0.49983,
		0.499836,
		0.49979,
		0.499394,
		0.498098,
		0.495252,
		0.490224,
		0.482466,
		0.471554,
		0.457205,
		0.43928,
		0.417783,
		0.392847,
		0.36473,
		0.333802,
		0.300538,
		0.265509,
		0.229374,
		0.192875,
		0.156828,
		0.122125,
		0.089722,
		0.06064,
		0.03596,
		0.016822,
		0.00442,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.00138,
		0.011815,
		0.031171,
		0.057933,
		0.090652,
		0.127953,
		0.168534,
		0.211175,
		0.254737,
		0.298172,
		0.340527,
		0.38095,
		0.4187,
		0.453154,
		0.48382,
		0.510346,
		0.532534,
		0.55035,
		0.56394,
		0.57363,
		0.579929,
		0.583504,
		0.585137,
		0.585637,
		0.585696,
		0.585688,
		0.585481,
		0.584512,
		0.58201,
		0.577165,
		0.569245,
		0.557657,
		0.541984,
		0.521988,
		0.497617,
		0.468986,
		0.436373,
		0.400201,
		0.361029,
		0.31954,
		0.276532,
		0.232908,
		0.18967,
		0.147912,
		0.108813,
		0.073636,
		0.043719,
		0.020474,
		0.005385,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.001639,
		0.014018,
		0.036943,
		0.068577,
		0.107171,
		0.151061,
		0.198682,
		0.248563,
		0.299339,
		0.349758,
		0.398686,
		0.445116,
		0.488182,
		0.527165,
		0.561514,
		0.590853,
		0.615002,
		0.633993,
		0.648076,
		0.657731,
		0.663655,
		0.666724,
		0.667916,
		0.668171,
		0.668181,
		0.668108,
		0.667493,
		0.665486,
		0.661108,
		0.653418,
		0.641623,
		0.625132,
		0.603579,
		0.576825,
		0.544946,
		0.508222,
		0.467118,
		0.422271,
		0.374475,
		0.324667,
		0.27392,
		0.223428,
		0.1745,
		0.128554,
		0.08711,
		0.051783,
		0.024279,
		0.006393,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.001902,
		0.016247,
		0.042764,
		0.079284,
		0.123736,
		0.174161,
		0.228713,
		0.285665,
		0.343419,
		0.400512,
		0.455628,
		0.507609,
		0.555467,
		0.598399,
		0.635805,
		0.667309,
		0.692773,
		0.712321,
		0.726348,
		0.735522,
		0.740761,
		0.74317,
		0.743911,
		0.743999,
		0.743987,
		0.743679,
		0.742246,
		0.738568,
		0.73149,
		0.719989,
		0.703269,
		0.680795,
		0.652309,
		0.617822,
		0.577591,
		0.532108,
		0.482077,
		0.428393,
		0.372134,
		0.314536,
		0.256992,
		0.201032,
		0.14832,
		0.100644,
		0.059906,
		0.028122,
		0.007413,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.002161,
		0.018437,
		0.048471,
		0.089745,
		0.139866,
		0.196569,
		0.257723,
		0.321343,
		0.385597,
		0.448814,
		0.5095,
		0.566352,
		0.618271,
		0.664385,
		0.704067,
		0.736961,
		0.763003,
		0.782445,
		0.795861,
		0.804146,
		0.808465,
		0.810153,
		0.810517,
		0.810531,
		0.810427,
		0.809553,
		0.80672,
		0.800578,
		0.789857,
		0.773521,
		0.75083,
		0.721373,
		0.685059,
		0.642099,
		0.592988,
		0.538479,
		0.47956,
		0.417432,
		0.353499,
		0.289343,
		0.226717,
		0.167531,
		0.113845,
		0.067856,
		0.031895,
		0.008417,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.00241,
		0.020541,
		0.053931,
		0.099716,
		0.155176,
		0.21774,
		0.284996,
		0.354703,
		0.424797,
		0.493408,
		0.558873,
		0.619756,
		0.674863,
		0.723272,
		0.764353,
		0.7978,
		0.823657,
		0.842338,
		0.85464,
		0.861713,
		0.864986,
		0.865999,
		0.86612,
		0.866103,
		0.865681,
		0.863728,
		0.858746,
		0.849223,
		0.833854,
		0.81166,
		0.782037,
		0.744755,
		0.699951,
		0.648096,
		0.589968,
		0.526629,
		0.459397,
		0.389823,
		0.319677,
		0.250926,
		0.185725,
		0.126401,
		0.075448,
		0.035511,
		0.009384,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.002646,
		0.022525,
		0.05906,
		0.109042,
		0.169425,
		0.237338,
		0.310092,
		0.385195,
		0.460362,
		0.533535,
		0.602893,
		0.666884,
		0.724239,
		0.774009,
		0.815588,
		0.848753,
		0.873692,
		0.891026,
		0.901806,
		0.907465,
		0.909691,
		0.910174,
		0.910193,
		0.910055,
		0.908898,
		0.905174,
		0.897154,
		0.883254,
		0.862222,
		0.83322,
		0.795845,
		0.750119,
		0.696458,
		0.635646,
		0.568794,
		0.497317,
		0.422905,
		0.347498,
		0.27327,
		0.202613,
		0.138117,
		0.082565,
		0.038915,
		0.010297,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.002868,
		0.02438,
		0.063833,
		0.117674,
		0.182539,
		0.255258,
		0.332875,
		0.412655,
		0.492103,
		0.568982,
		0.641333,
		0.707502,
		0.766169,
		0.816383,
		0.857596,
		0.889702,
		0.913075,
		0.928581,
		0.937562,
		0.941747,
		0.943052,
		0.943208,
		0.943187,
		0.942641,
		0.940135,
		0.933787,
		0.921738,
		0.902438,
		0.874775,
		0.838132,
		0.792377,
		0.737842,
		0.675281,
		0.605834,
		0.530993,
		0.452566,
		0.372656,
		0.29363,
		0.218106,
		0.148931,
		0.08917,
		0.04209,
		0.011152,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.003077,
		0.026116,
		0.068277,
		0.125665,
		0.194598,
		0.271617,
		0.3535,
		0.437279,
		0.520259,
		0.600037,
		0.674529,
		0.742003,
		0.801108,
		0.850921,
		0.890985,
		0.921354,
		0.942628,
		0.955961,
		0.963014,
		0.965809,
		0.966419,
		0.966443,
		0.966268,
		0.96481,
		0.960152,
		0.950194,
		0.933069,
		0.907357,
		0.872181,
		0.827216,
		0.77267,
		0.709237,
		0.63806,
		0.560686,
		0.479026,
		0.395328,
		0.312142,
		0.232304,
		0.15891,
		0.095303,
		0.045055,
		0.011955,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.003275,
		0.027762,
		0.072466,
		0.133151,
		0.205814,
		0.286708,
		0.372351,
		0.459545,
		0.545402,
		0.627366,
		0.703245,
		0.771243,
		0.830009,
		0.878677,
		0.916918,
		0.944988,
		0.963764,
		0.974726,
		0.979875,
		0.981492,
		0.981687,
		0.981661,
		0.980982,
		0.977887,
		0.970107,
		0.955459,
		0.932184,
		0.899099,
		0.855639,
		0.801847,
		0.738324,
		0.666188,
		0.587017,
		0.502809,
		0.415942,
		0.329143,
		0.245457,
		0.168226,
		0.101067,
		0.047857,
		0.012718,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.003469,
		0.029357,
		0.076504,
		0.140319,
		0.216478,
		0.300936,
		0.38995,
		0.480096,
		0.568298,
		0.651854,
		0.728476,
		0.796331,
		0.854087,
		0.900971,
		0.936827,
		0.962163,
		0.978178,
		0.986721,
		0.990133,
		0.990884,
		0.990914,
		0.990698,
		0.988911,
		0.983245,
		0.971234,
		0.950751,
		0.920262,
		0.878913,
		0.826533,
		0.763595,
		0.691157,
		0.61081,
		0.524618,
		0.435083,
		0.345096,
		0.257911,
		0.177115,
		0.106605,
		0.050566,
		0.013459,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.003662,
		0.030947,
		0.080506,
		0.147381,
		0.22691,
		0.314743,
		0.406867,
		0.499627,
		0.589761,
		0.674429,
		0.751262,
		0.818404,
		0.874577,
		0.919139,
		0.952154,
		0.974436,
		0.987562,
		0.993781,
		0.995751,
		0.99599,
		0.995957,
		0.995127,
		0.991373,
		0.982016,
		0.964555,
		0.937057,
		0.898322,
		0.847917,
		0.786138,
		0.713954,
		0.632938,
		0.54521,
		0.453381,
		0.360506,
		0.270048,
		0.185844,
		0.112078,
		0.053259,
		0.014199,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.003862,
		0.032575,
		0.084585,
		0.154541,
		0.237419,
		0.328548,
		0.423631,
		0.518777,
		0.610534,
		0.695929,
		0.772521,
		0.83846,
		0.892551,
		0.93433,
		0.964136,
		0.983155,
		0.993394,
		0.997522,
		0.998438,
		0.998474,
		0.99821,
		0.996039,
		0.989217,
		0.974887,
		0.950676,
		0.914977,
		0.867033,
		0.806913,
		0.735456,
		0.654192,
		0.565277,
		0.471423,
		0.375848,
		0.282229,
		0.194665,
		0.117642,
		0.05601,
		0.014959,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.004071,
		0.034281,
		0.088841,
		0.161974,
		0.248267,
		0.342703,
		0.440684,
		0.538067,
		0.631206,
		0.717001,
		0.792954,
		0.857243,
		0.908799,
		0.94739,
		0.973698,
		0.989349,
		0.996836,
		0.99923,
		0.999523,
		0.999483,
		0.99847,
		0.993929,
		0.982718,
		0.961999,
		0.929694,
		0.884648,
		0.826642,
		0.756338,
		0.675192,
		0.585374,
		0.48969,
		0.39152,
		0.294765,
		0.203799,
		0.123433,
		0.058887,
		0.015757,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.004295,
		0.036097,
		0.093355,
		0.169822,
		0.259662,
		0.357477,
		0.458348,
		0.557864,
		0.652181,
		0.738071,
		0.813001,
		0.875207,
		0.923796,
		0.958834,
		0.981426,
		0.993717,
		0.998721,
		0.999841,
		0.999886,
		0.999561,
		0.996917,
		0.988688,
		0.971575,
		0.942964,
		0.901223,
		0.845772,
		0.777035,
		0.696355,
		0.605889,
		0.508529,
		0.407819,
		0.307892,
		0.213418,
		0.129562,
		0.061944,
		0.016608,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.004537,
		0.03805,
		0.098192,
		0.178194,
		0.271752,
		0.373058,
		0.476837,
		0.578397,
		0.673685,
		0.759357,
		0.832862,
		0.892541,
		0.937734,
		0.96889,
		0.987631,
		0.996693,
		0.99962,
		0.999981,
		0.999932,
		0.998688,
		0.993163,
		0.979666,
		0.954993,
		0.916947,
		0.864496,
		0.797756,
		0.717902,
		0.627051,
		0.528163,
		0.424946,
		0.321779,
		0.223654,
		0.136115,
		0.065226,
		0.017525,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389572,
		0.496281,
		0.599782,
		0.695809,
		0.780912,
		0.852554,
		0.909231,
		0.9506,
		0.97759,
		0.992443,
		0.998558,
		0.999942,
		0.999998,
		0.999595,
		0.996346,
		0.986344,
		0.965783,
		0.931802,
		0.882813,
		0.818535,
		0.739907,
		0.648966,
		0.548717,
		0.443032,
		0.336549,
		0.234602,
		0.143159,
		0.068768,
		0.018518,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.006111,
		0.050639,
		0.128949,
		0.230606,
		0.346048,
		0.466665,
		0.584888,
		0.694316,
		0.789856,
		0.867913,
		0.926607,
		0.966004,
		0.988294,
		0.997721,
		0.999909,
		0.999999,
		0.999355,
		0.994282,
		0.979066,
		0.948631,
		0.899725,
		0.83123,
		0.744084,
		0.641052,
		0.526505,
		0.406208,
		0.287165,
		0.177478,
		0.086243,
		0.023468,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.006524,
		0.053897,
		0.136799,
		0.243759,
		0.364313,
		0.489096,
		0.609954,
		0.720092,
		0.814257,
		0.888967,
		0.942774,
		0.976536,
		0.993551,
		0.999268,
		0.999999,
		0.999897,
		0.997422,
		0.986838,
		0.962012,
		0.918486,
		0.854185,
		0.769433,
		0.66673,
		0.550474,
		0.426716,
		0.302949,
		0.18795,
		0.091646,
		0.025016,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.006979,
		0.057479,
		0.145377,
		0.258031,
		0.383959,
		0.512956,
		0.636233,
		0.746597,
		0.838682,
		0.909232,
		0.957423,
		0.985151,
		0.997073,
		0.999882,
		0.999998,
		0.999166,
		0.9927,
		0.973615,
		0.93607,
		0.876762,
		0.795181,
		0.69341,
		0.575799,
		0.448658,
		0.320005,
		0.199356,
		0.09757,
		0.026722,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.007484,
		0.061429,
		0.154777,
		0.273547,
		0.405109,
		0.538323,
		0.663715,
		0.773702,
		0.862877,
		0.928371,
		0.97023,
		0.991705,
		0.999046,
		0.999998,
		0.999864,
		0.996662,
		0.983187,
		0.952128,
		0.898654,
		0.821134,
		0.721027,
		0.602518,
		0.472143,
		0.338462,
		0.211807,
		0.104082,
		0.028608,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.008046,
		0.065798,
		0.165103,
		0.290446,
		0.427896,
		0.565272,
		0.692366,
		0.801227,
		0.886525,
		0.945995,
		0.980889,
		0.996177,
		0.999844,
		0.999998,
		0.998903,
		0.990536,
		0.966289,
		0.919497,
		0.84704,
		0.749475,
		0.630658,
		0.497281,
		0.358463,
		0.225429,
		0.111261,
		0.0307,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.008674,
		0.070648,
		0.176478,
		0.308888,
		0.452462,
		0.593863,
		0.722107,
		0.828924,
		0.909232,
		0.961681,
		0.989154,
		0.998733,
		0.999998,
		0.999819,
		0.995601,
		0.978187,
		0.938862,
		0.872571,
		0.778592,
		0.660215,
		0.52418,
		0.380165,
		0.24037,
		0.119202,
		0.033031,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.009378,
		0.076051,
		0.189047,
		0.32905,
		0.478955,
		0.624138,
		0.752804,
		0.856465,
		0.930531,
		0.974996,
		0.994912,
		0.999788,
		0.999997,
		0.998528,
		0.987511,
		0.956268,
		0.897313,
		0.808146,
		0.69115,
		0.552945,
		0.403741,
		0.256796,
		0.128015,
		0.035637,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.010171,
		0.082095,
		0.202976,
		0.351133,
		0.507526,
		0.656102,
		0.78425,
		0.883417,
		0.949876,
		0.985548,
		0.99828,
		0.999997,
		0.999752,
		0.994084,
		0.971204,
		0.920755,
		0.837812,
		0.723368,
		0.583667,
		0.429378,
		0.2749,
		0.137829,
		0.038563,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.011069,
		0.088884,
		0.218463,
		0.375363,
		0.538323,
		0.689715,
		0.816138,
		0.909232,
		0.966673,
		0.993081,
		0.999707,
		0.999996,
		0.997978,
		0.983187,
		0.94229,
		0.867151,
		0.756699,
		0.61641,
		0.457278,
		0.294906,
		0.148801,
		0.041863,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.012092,
		0.096544,
		0.23574,
		0.40199,
		0.571482,
		0.72486,
		0.848031,
		0.933236,
		0.980326,
		0.997607,
		0.999995,
		0.999651,
		0.991857,
		0.961225,
		0.895582,
		0.790869,
		0.651201,
		0.487653,
		0.317068,
		0.161115,
		0.045605,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.013263,
		0.105229,
		0.255081,
		0.431293,
		0.607109,
		0.761317,
		0.879329,
		0.95463,
		0.990352,
		0.999581,
		0.999994,
		0.997147,
		0.976835,
		0.922365,
		0.825463,
		0.687996,
		0.520718,
		0.341684,
		0.174994,
		0.049871,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.014613,
		0.11513,
		0.276811,
		0.463572,
		0.645261,
		0.798713,
		0.909232,
		0.972543,
		0.996571,
		0.999993,
		0.999493,
		0.988485,
		0.946591,
		0.859876,
		0.726654,
		0.556677,
		0.369091,
		0.19071,
		0.054761,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.016179,
		0.12648,
		0.301315,
		0.499142,
		0.685901,
		0.836467,
		0.936718,
		0.986146,
		0.999381,
		0.999991,
		0.995843,
		0.967218,
		0.893264,
		0.766878,
		0.595706,
		0.399673,
		0.208589,
		0.060404,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.018011,
		0.139573,
		0.32905,
		0.538323,
		0.728851,
		0.873716,
		0.960553,
		0.994912,
		0.999989,
		0.999236,
		0.983187,
		0.924495,
		0.808146,
		0.637915,
		0.433865,
		0.229035,
		0.066963,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.020173,
		0.154777,
		0.360554,
		0.581406,
		0.773702,
		0.909232,
		0.9794,
		0.999046,
		0.999986,
		0.993706,
		0.952128,
		0.849616,
		0.683291,
		0.472143,
		0.25254,
		0.074644,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.022747,
		0.172561,
		0.39646,
		0.628601,
		0.819691,
		0.941362,
		0.992122,
		0.999982,
		0.998793,
		0.974494,
		0.890001,
		0.731605,
		0.515011,
		0.279714,
		0.083718,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.025825,
		0.19336,
		0.437117,
		0.679351,
		0.864779,
		0.967203,
		0.997575,
		0.9991,
		0.989142,
		0.926633,
		0.781582,
		0.56247,
		0.311037,
		0.094456,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.028884,
		0.212988,
		0.472358,
		0.716734,
		0.886448,
		0.96241,
		0.974912,
		0.97297,
		0.935426,
		0.81321,
		0.600964,
		0.339533,
		0.104887,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.028801,
		0.208583,
		0.45206,
		0.666205,
		0.795954,
		0.837497,
		0.839721,
		0.825636,
		0.743241,
		0.567267,
		0.328938,
		0.103703,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.021807,
		0.154555,
		0.325662,
		0.462908,
		0.530799,
		0.542935,
		0.540983,
		0.505937,
		0.401745,
		0.240551,
		0.077731,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.010604,
		0.073201,
		0.148905,
		0.202269,
		0.22133,
		0.222444,
		0.215674,
		0.179781,
		0.112084,
		0.037339,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.002407,
		0.016081,
		0.031269,
		0.040099,
		0.041877,
		0.041577,
		0.036721,
		0.024116,
		0.008349,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		5.1e-005,
		0.000326,
		0.000596,
		0.000712,
		0.000719,
		0.000676,
		0.000475,
		0.000173,
		0.000173,
		0.000475,
		0.000676,
		0.000719,
		0.000712,
		0.000596,
		0.000326,
		5.1e-005,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		0.008349,
		0.024116,
		0.036721,
		0.041577,
		0.041877,
		0.040099,
		0.031269,
		0.016081,
		0.002407,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.037339,
		0.112084,
		0.179781,
		0.215674,
		0.222444,
		0.22133,
		0.202269,
		0.148905,
		0.073201,
		0.010604,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.077731,
		0.240551,
		0.401745,
		0.505937,
		0.540983,
		0.542935,
		0.530799,
		0.462908,
		0.325662,
		0.154555,
		0.021807,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.103703,
		0.328938,
		0.567267,
		0.743241,
		0.825636,
		0.839721,
		0.837497,
		0.795954,
		0.666205,
		0.45206,
		0.208583,
		0.028801,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.104887,
		0.339533,
		0.600964,
		0.81321,
		0.935426,
		0.97297,
		0.974912,
		0.96241,
		0.886448,
		0.716734,
		0.472358,
		0.212988,
		0.028884,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.094456,
		0.311037,
		0.56247,
		0.781582,
		0.926633,
		0.989142,
		0.9991,
		0.997575,
		0.967203,
		0.864779,
		0.679351,
		0.437117,
		0.19336,
		0.025825,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.083718,
		0.279714,
		0.515011,
		0.731605,
		0.890001,
		0.974494,
		0.998793,
		0.999982,
		0.992122,
		0.941362,
		0.819691,
		0.628601,
		0.39646,
		0.172561,
		0.022747,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.074644,
		0.25254,
		0.472143,
		0.683291,
		0.849616,
		0.952128,
		0.993706,
		0.999986,
		0.999046,
		0.9794,
		0.909232,
		0.773702,
		0.581406,
		0.360554,
		0.154777,
		0.020173,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.066963,
		0.229035,
		0.433865,
		0.637915,
		0.808146,
		0.924495,
		0.983187,
		0.999236,
		0.999989,
		0.994912,
		0.960553,
		0.873716,
		0.728851,
		0.538323,
		0.32905,
		0.139573,
		0.018011,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.060404,
		0.208589,
		0.399673,
		0.595706,
		0.766878,
		0.893264,
		0.967218,
		0.995843,
		0.999991,
		0.999381,
		0.986146,
		0.936718,
		0.836467,
		0.685901,
		0.499142,
		0.301315,
		0.12648,
		0.016179,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.054761,
		0.19071,
		0.369091,
		0.556677,
		0.726654,
		0.859876,
		0.946591,
		0.988485,
		0.999493,
		0.999993,
		0.996571,
		0.972543,
		0.909232,
		0.798713,
		0.645261,
		0.463572,
		0.276811,
		0.11513,
		0.014613,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.049871,
		0.174994,
		0.341684,
		0.520718,
		0.687996,
		0.825463,
		0.922365,
		0.976835,
		0.997147,
		0.999994,
		0.999581,
		0.990352,
		0.95463,
		0.879329,
		0.761317,
		0.607109,
		0.431293,
		0.255081,
		0.105229,
		0.013263,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.045605,
		0.161115,
		0.317068,
		0.487653,
		0.651201,
		0.790869,
		0.895582,
		0.961225,
		0.991857,
		0.999651,
		0.999995,
		0.997607,
		0.980326,
		0.933236,
		0.848031,
		0.72486,
		0.571482,
		0.40199,
		0.23574,
		0.096544,
		0.012092,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.041863,
		0.148801,
		0.294906,
		0.457278,
		0.61641,
		0.756699,
		0.867151,
		0.94229,
		0.983187,
		0.997978,
		0.999996,
		0.999707,
		0.993081,
		0.966673,
		0.909232,
		0.816138,
		0.689715,
		0.538323,
		0.375363,
		0.218463,
		0.088884,
		0.011069,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.038563,
		0.137829,
		0.2749,
		0.429378,
		0.583667,
		0.723368,
		0.837812,
		0.920755,
		0.971204,
		0.994084,
		0.999752,
		0.999997,
		0.99828,
		0.985548,
		0.949876,
		0.883417,
		0.78425,
		0.656102,
		0.507526,
		0.351133,
		0.202976,
		0.082095,
		0.010171,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.035637,
		0.128015,
		0.256796,
		0.403741,
		0.552945,
		0.69115,
		0.808146,
		0.897313,
		0.956268,
		0.987511,
		0.998528,
		0.999997,
		0.999788,
		0.994912,
		0.974996,
		0.930531,
		0.856465,
		0.752804,
		0.624138,
		0.478955,
		0.32905,
		0.189047,
		0.076051,
		0.009378,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.033031,
		0.119202,
		0.24037,
		0.380165,
		0.52418,
		0.660215,
		0.778592,
		0.872571,
		0.938862,
		0.978187,
		0.995601,
		0.999819,
		0.999998,
		0.998733,
		0.989154,
		0.961681,
		0.909232,
		0.828924,
		0.722107,
		0.593863,
		0.452462,
		0.308888,
		0.176478,
		0.070648,
		0.008674,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.0307,
		0.111261,
		0.225429,
		0.358463,
		0.497281,
		0.630658,
		0.749475,
		0.84704,
		0.919497,
		0.966289,
		0.990536,
		0.998903,
		0.999998,
		0.999844,
		0.996177,
		0.980889,
		0.945995,
		0.886525,
		0.801227,
		0.692366,
		0.565272,
		0.427896,
		0.290446,
		0.165103,
		0.065798,
		0.008046,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.028608,
		0.104082,
		0.211807,
		0.338462,
		0.472143,
		0.602518,
		0.721027,
		0.821134,
		0.898654,
		0.952128,
		0.983187,
		0.996662,
		0.999864,
		0.999998,
		0.999046,
		0.991705,
		0.97023,
		0.928371,
		0.862877,
		0.773702,
		0.663715,
		0.538323,
		0.405109,
		0.273547,
		0.154777,
		0.061429,
		0.007484,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.026722,
		0.09757,
		0.199356,
		0.320005,
		0.448658,
		0.575799,
		0.69341,
		0.795181,
		0.876762,
		0.93607,
		0.973615,
		0.9927,
		0.999166,
		0.999998,
		0.999882,
		0.997073,
		0.985151,
		0.957423,
		0.909232,
		0.838682,
		0.746597,
		0.636233,
		0.512956,
		0.383959,
		0.258031,
		0.145377,
		0.057479,
		0.006979,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.025016,
		0.091646,
		0.18795,
		0.302949,
		0.426716,
		0.550474,
		0.66673,
		0.769433,
		0.854185,
		0.918486,
		0.962012,
		0.986838,
		0.997422,
		0.999897,
		0.999999,
		0.999268,
		0.993551,
		0.976536,
		0.942774,
		0.888967,
		0.814257,
		0.720092,
		0.609954,
		0.489096,
		0.364313,
		0.243759,
		0.136799,
		0.053897,
		0.006524,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.023468,
		0.086243,
		0.177478,
		0.287165,
		0.406208,
		0.526505,
		0.641052,
		0.744084,
		0.83123,
		0.899725,
		0.948631,
		0.979066,
		0.994282,
		0.999355,
		0.999999,
		0.999909,
		0.997721,
		0.988294,
		0.966004,
		0.926607,
		0.867913,
		0.789856,
		0.694316,
		0.584888,
		0.466665,
		0.346048,
		0.230606,
		0.128949,
		0.050639,
		0.006111,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.018518,
		0.068768,
		0.143159,
		0.234602,
		0.336549,
		0.443032,
		0.548717,
		0.648966,
		0.739907,
		0.818535,
		0.882813,
		0.931802,
		0.965783,
		0.986344,
		0.996346,
		0.999595,
		0.999998,
		0.999942,
		0.998558,
		0.992443,
		0.97759,
		0.9506,
		0.909231,
		0.852554,
		0.780912,
		0.695809,
		0.599782,
		0.496281,
		0.389572,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.017525,
		0.065226,
		0.136115,
		0.223654,
		0.321779,
		0.424946,
		0.528163,
		0.627051,
		0.717902,
		0.797756,
		0.864496,
		0.916947,
		0.954993,
		0.979666,
		0.993163,
		0.998688,
		0.999932,
		0.999981,
		0.99962,
		0.996693,
		0.987631,
		0.96889,
		0.937734,
		0.892541,
		0.832862,
		0.759357,
		0.673685,
		0.578397,
		0.476837,
		0.373058,
		0.271752,
		0.178194,
		0.098192,
		0.03805,
		0.004537,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.016608,
		0.061944,
		0.129562,
		0.213418,
		0.307892,
		0.407819,
		0.508529,
		0.605889,
		0.696355,
		0.777035,
		0.845772,
		0.901223,
		0.942964,
		0.971575,
		0.988688,
		0.996917,
		0.999561,
		0.999886,
		0.999841,
		0.998721,
		0.993717,
		0.981426,
		0.958834,
		0.923796,
		0.875207,
		0.813001,
		0.738071,
		0.652181,
		0.557864,
		0.458348,
		0.357477,
		0.259662,
		0.169822,
		0.093355,
		0.036097,
		0.004295,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.015757,
		0.058887,
		0.123433,
		0.203799,
		0.294765,
		0.39152,
		0.48969,
		0.585374,
		0.675192,
		0.756338,
		0.826642,
		0.884648,
		0.929694,
		0.961999,
		0.982718,
		0.993929,
		0.99847,
		0.999483,
		0.999523,
		0.99923,
		0.996836,
		0.989349,
		0.973698,
		0.94739,
		0.908799,
		0.857243,
		0.792954,
		0.717001,
		0.631206,
		0.538067,
		0.440684,
		0.342703,
		0.248267,
		0.161974,
		0.088841,
		0.034281,
		0.004071,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.014959,
		0.05601,
		0.117642,
		0.194665,
		0.282229,
		0.375848,
		0.471423,
		0.565277,
		0.654192,
		0.735456,
		0.806913,
		0.867033,
		0.914977,
		0.950676,
		0.974887,
		0.989217,
		0.996039,
		0.99821,
		0.998474,
		0.998438,
		0.997522,
		0.993394,
		0.983155,
		0.964136,
		0.93433,
		0.892551,
		0.83846,
		0.772521,
		0.695929,
		0.610534,
		0.518777,
		0.423631,
		0.328548,
		0.237419,
		0.154541,
		0.084585,
		0.032575,
		0.003862,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.014199,
		0.053259,
		0.112078,
		0.185844,
		0.270048,
		0.360506,
		0.453381,
		0.54521,
		0.632938,
		0.713954,
		0.786138,
		0.847917,
		0.898322,
		0.937057,
		0.964555,
		0.982016,
		0.991373,
		0.995127,
		0.995957,
		0.99599,
		0.995751,
		0.993781,
		0.987562,
		0.974436,
		0.952154,
		0.919139,
		0.874577,
		0.818404,
		0.751262,
		0.674429,
		0.589761,
		0.499627,
		0.406867,
		0.314743,
		0.22691,
		0.147381,
		0.080506,
		0.030947,
		0.003662,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.013459,
		0.050566,
		0.106605,
		0.177115,
		0.257911,
		0.345096,
		0.435083,
		0.524618,
		0.61081,
		0.691157,
		0.763595,
		0.826533,
		0.878913,
		0.920262,
		0.950751,
		0.971234,
		0.983245,
		0.988911,
		0.990698,
		0.990914,
		0.990884,
		0.990133,
		0.986721,
		0.978178,
		0.962163,
		0.936827,
		0.900971,
		0.854087,
		0.796331,
		0.728476,
		0.651854,
		0.568298,
		0.480096,
		0.38995,
		0.300936,
		0.216478,
		0.140319,
		0.076504,
		0.029357,
		0.003469,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.012718,
		0.047857,
		0.101067,
		0.168226,
		0.245457,
		0.329143,
		0.415942,
		0.502809,
		0.587017,
		0.666188,
		0.738324,
		0.801847,
		0.855639,
		0.899099,
		0.932184,
		0.955459,
		0.970107,
		0.977887,
		0.980982,
		0.981661,
		0.981687,
		0.981492,
		0.979875,
		0.974726,
		0.963764,
		0.944988,
		0.916918,
		0.878677,
		0.830009,
		0.771243,
		0.703245,
		0.627366,
		0.545402,
		0.459545,
		0.372351,
		0.286708,
		0.205814,
		0.133151,
		0.072466,
		0.027762,
		0.003275,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.011955,
		0.045055,
		0.095303,
		0.15891,
		0.232304,
		0.312142,
		0.395328,
		0.479026,
		0.560686,
		0.63806,
		0.709237,
		0.77267,
		0.827216,
		0.872181,
		0.907357,
		0.933069,
		0.950194,
		0.960152,
		0.96481,
		0.966268,
		0.966443,
		0.966419,
		0.965809,
		0.963014,
		0.955961,
		0.942628,
		0.921354,
		0.890985,
		0.850921,
		0.801108,
		0.742003,
		0.674529,
		0.600037,
		0.520259,
		0.437279,
		0.3535,
		0.271617,
		0.194598,
		0.125665,
		0.068277,
		0.026116,
		0.003077,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.011152,
		0.04209,
		0.08917,
		0.148931,
		0.218106,
		0.29363,
		0.372656,
		0.452566,
		0.530993,
		0.605834,
		0.675281,
		0.737842,
		0.792377,
		0.838132,
		0.874775,
		0.902438,
		0.921738,
		0.933787,
		0.940135,
		0.942641,
		0.943187,
		0.943208,
		0.943052,
		0.941747,
		0.937562,
		0.928581,
		0.913075,
		0.889702,
		0.857596,
		0.816383,
		0.766169,
		0.707502,
		0.641333,
		0.568982,
		0.492103,
		0.412655,
		0.332875,
		0.255258,
		0.182539,
		0.117674,
		0.063833,
		0.02438,
		0.002868,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.010297,
		0.038915,
		0.082565,
		0.138117,
		0.202613,
		0.27327,
		0.347498,
		0.422905,
		0.497317,
		0.568794,
		0.635646,
		0.696458,
		0.750119,
		0.795845,
		0.83322,
		0.862222,
		0.883254,
		0.897154,
		0.905174,
		0.908898,
		0.910055,
		0.910193,
		0.910174,
		0.909691,
		0.907465,
		0.901806,
		0.891026,
		0.873692,
		0.848753,
		0.815588,
		0.774009,
		0.724239,
		0.666884,
		0.602893,
		0.533535,
		0.460362,
		0.385195,
		0.310092,
		0.237338,
		0.169425,
		0.109042,
		0.05906,
		0.022525,
		0.002646,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.009384,
		0.035511,
		0.075448,
		0.126401,
		0.185725,
		0.250926,
		0.319677,
		0.389823,
		0.459397,
		0.526629,
		0.589968,
		0.648096,
		0.699951,
		0.744755,
		0.782037,
		0.81166,
		0.833854,
		0.849223,
		0.858746,
		0.863728,
		0.865681,
		0.866103,
		0.86612,
		0.865999,
		0.864986,
		0.861713,
		0.85464,
		0.842338,
		0.823657,
		0.7978,
		0.764353,
		0.723272,
		0.674863,
		0.619756,
		0.558873,
		0.493408,
		0.424797,
		0.354703,
		0.284996,
		0.21774,
		0.155176,
		0.099716,
		0.053931,
		0.020541,
		0.00241,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.008417,
		0.031895,
		0.067856,
		0.113845,
		0.167531,
		0.226717,
		0.289343,
		0.353499,
		0.417432,
		0.47956,
		0.538479,
		0.592988,
		0.642099,
		0.685059,
		0.721373,
		0.75083,
		0.773521,
		0.789857,
		0.800578,
		0.80672,
		0.809553,
		0.810427,
		0.810531,
		0.810517,
		0.810153,
		0.808465,
		0.804146,
		0.795861,
		0.782445,
		0.763003,
		0.736961,
		0.704067,
		0.664385,
		0.618271,
		0.566352,
		0.5095,
		0.448814,
		0.385597,
		0.321343,
		0.257723,
		0.196569,
		0.139866,
		0.089745,
		0.048471,
		0.018437,
		0.002161,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.007413,
		0.028122,
		0.059906,
		0.100644,
		0.14832,
		0.201032,
		0.256992,
		0.314536,
		0.372134,
		0.428393,
		0.482077,
		0.532108,
		0.577591,
		0.617822,
		0.652309,
		0.680795,
		0.703269,
		0.719989,
		0.73149,
		0.738568,
		0.742246,
		0.743679,
		0.743987,
		0.743999,
		0.743911,
		0.74317,
		0.740761,
		0.735522,
		0.726348,
		0.712321,
		0.692773,
		0.667309,
		0.635805,
		0.598399,
		0.555467,
		0.507609,
		0.455628,
		0.400512,
		0.343419,
		0.285665,
		0.228713,
		0.174161,
		0.123736,
		0.079284,
		0.042764,
		0.016247,
		0.001902,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.006393,
		0.024279,
		0.051783,
		0.08711,
		0.128554,
		0.1745,
		0.223428,
		0.27392,
		0.324667,
		0.374475,
		0.422271,
		0.467118,
		0.508222,
		0.544946,
		0.576825,
		0.603579,
		0.625132,
		0.641623,
		0.653418,
		0.661108,
		0.665486,
		0.667493,
		0.668108,
		0.668181,
		0.668171,
		0.667916,
		0.666724,
		0.663655,
		0.657731,
		0.648076,
		0.633993,
		0.615002,
		0.590853,
		0.561514,
		0.527165,
		0.488182,
		0.445116,
		0.398686,
		0.349758,
		0.299339,
		0.248563,
		0.198682,
		0.151061,
		0.107171,
		0.068577,
		0.036943,
		0.014018,
		0.001639,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.005385,
		0.020474,
		0.043719,
		0.073636,
		0.108813,
		0.147912,
		0.18967,
		0.232908,
		0.276532,
		0.31954,
		0.361029,
		0.400201,
		0.436373,
		0.468986,
		0.497617,
		0.521988,
		0.541984,
		0.557657,
		0.569245,
		0.577165,
		0.58201,
		0.584512,
		0.585481,
		0.585688,
		0.585696,
		0.585637,
		0.585137,
		0.583504,
		0.579929,
		0.57363,
		0.56394,
		0.55035,
		0.532534,
		0.510346,
		0.48382,
		0.453154,
		0.4187,
		0.38095,
		0.340527,
		0.298172,
		0.254737,
		0.211175,
		0.168534,
		0.127953,
		0.090652,
		0.057933,
		0.031171,
		0.011815,
		0.00138,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.00442,
		0.016822,
		0.03596,
		0.06064,
		0.089722,
		0.122125,
		0.156828,
		0.192875,
		0.229374,
		0.265509,
		0.300538,
		0.333802,
		0.36473,
		0.392847,
		0.417783,
		0.43928,
		0.457205,
		0.471554,
		0.482466,
		0.490224,
		0.495252,
		0.498098,
		0.499394,
		0.49979,
		0.499836,
		0.49983,
		0.499666,
		0.498898,
		0.496909,
		0.493048,
		0.486716,
		0.477426,
		0.464821,
		0.448693,
		0.428973,
		0.405731,
		0.379165,
		0.34959,
		0.317429,
		0.283205,
		0.247536,
		0.211121,
		0.174741,
		0.139251,
		0.105574,
		0.074699,
		0.04768,
		0.025625,
		0.009702,
		0.001132,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.003525,
		0.01343,
		0.028739,
		0.048517,
		0.071873,
		0.097956,
		0.125963,
		0.155141,
		0.184787,
		0.214251,
		0.242943,
		0.270336,
		0.295966,
		0.319444,
		0.340459,
		0.358785,
		0.374285,
		0.386925,
		0.396774,
		0.404013,
		0.408933,
		0.411925,
		0.413461,
		0.414053,
		0.414179,
		0.414184,
		0.414148,
		0.413844,
		0.412843,
		0.410642,
		0.406741,
		0.400704,
		0.392189,
		0.38096,
		0.366892,
		0.349969,
		0.330276,
		0.307995,
		0.283397,
		0.256833,
		0.228728,
		0.19958,
		0.169947,
		0.140451,
		0.111767,
		0.084624,
		0.059802,
		0.038126,
		0.020468,
		0.007742,
		0.000902,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.002723,
		0.010384,
		0.022245,
		0.037596,
		0.055758,
		0.076086,
		0.097968,
		0.120829,
		0.144131,
		0.167375,
		0.190107,
		0.211916,
		0.232442,
		0.251376,
		0.268467,
		0.283525,
		0.296425,
		0.307118,
		0.31563,
		0.322066,
		0.326616,
		0.329549,
		0.331199,
		0.331948,
		0.332175,
		0.332201,
		0.332198,
		0.332104,
		0.331662,
		0.330511,
		0.328265,
		0.324562,
		0.319097,
		0.311642,
		0.302048,
		0.29025,
		0.276259,
		0.260166,
		0.242125,
		0.22236,
		0.201151,
		0.178832,
		0.155789,
		0.132457,
		0.109311,
		0.086869,
		0.06569,
		0.046366,
		0.029527,
		0.015835,
		0.005983,
		0.000697,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.002029,
		0.007747,
		0.016611,
		0.028102,
		0.041725,
		0.057003,
		0.073489,
		0.090758,
		0.108413,
		0.126086,
		0.143438,
		0.160162,
		0.175988,
		0.19068,
		0.204045,
		0.21593,
		0.22623,
		0.234893,
		0.241918,
		0.247362,
		0.251342,
		0.254032,
		0.25566,
		0.256491,
		0.25681,
		0.256878,
		0.25688,
		0.256861,
		0.256697,
		0.256157,
		0.254963,
		0.252835,
		0.249525,
		0.244832,
		0.238608,
		0.230768,
		0.221283,
		0.21018,
		0.197539,
		0.183489,
		0.168203,
		0.151896,
		0.134822,
		0.117269,
		0.099561,
		0.082051,
		0.065122,
		0.049186,
		0.034677,
		0.02206,
		0.011818,
		0.004461,
		0.000519,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.001451,
		0.005545,
		0.011902,
		0.020157,
		0.029959,
		0.040976,
		0.052889,
		0.0654,
		0.078227,
		0.091109,
		0.103804,
		0.116093,
		0.12778,
		0.138695,
		0.148694,
		0.157661,
		0.165515,
		0.172206,
		0.177722,
		0.18209,
		0.185375,
		0.187686,
		0.189167,
		0.189997,
		0.190372,
		0.190485,
		0.190498,
		0.190496,
		0.19045,
		0.190229,
		0.189652,
		0.18852,
		0.186644,
		0.183862,
		0.180046,
		0.17511,
		0.169008,
		0.161731,
		0.153312,
		0.143817,
		0.133344,
		0.122023,
		0.110012,
		0.097494,
		0.084677,
		0.071791,
		0.059088,
		0.046839,
		0.035336,
		0.024886,
		0.015814,
		0.008464,
		0.003192,
		0.000371,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.000989,
		0.003782,
		0.008124,
		0.013772,
		0.02049,
		0.028054,
		0.036252,
		0.044882,
		0.053754,
		0.062691,
		0.07153,
		0.08012,
		0.088328,
		0.096035,
		0.103142,
		0.109565,
		0.115245,
		0.12014,
		0.124235,
		0.127539,
		0.130086,
		0.13194,
		0.133186,
		0.133936,
		0.134318,
		0.134464,
		0.134494,
		0.134495,
		0.134487,
		0.134412,
		0.134165,
		0.133616,
		0.132632,
		0.131095,
		0.128904,
		0.125984,
		0.122288,
		0.117792,
		0.112501,
		0.106443,
		0.099669,
		0.092251,
		0.084279,
		0.075864,
		0.067132,
		0.058225,
		0.049299,
		0.040526,
		0.032087,
		0.02418,
		0.017011,
		0.0108,
		0.005775,
		0.002176,
		0.000253,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000635,
		0.002431,
		0.005227,
		0.008869,
		0.013209,
		0.018105,
		0.023421,
		0.02903,
		0.034811,
		0.040651,
		0.046446,
		0.052099,
		0.057525,
		0.062646,
		0.067396,
		0.07172,
		0.075576,
		0.078936,
		0.081783,
		0.084118,
		0.085958,
		0.087336,
		0.0883,
		0.088916,
		0.089259,
		0.089413,
		0.089459,
		0.089465,
		0.089464,
		0.089445,
		0.089354,
		0.089116,
		0.088647,
		0.087866,
		0.086702,
		0.085099,
		0.083014,
		0.080424,
		0.077319,
		0.073709,
		0.069614,
		0.065071,
		0.060128,
		0.054846,
		0.049296,
		0.04356,
		0.03773,
		0.031906,
		0.026196,
		0.020718,
		0.015596,
		0.010961,
		0.006952,
		0.003714,
		0.001398,
		0.000162,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		0.000379,
		0.001451,
		0.003122,
		0.005302,
		0.007904,
		0.010844,
		0.014044,
		0.017426,
		0.020921,
		0.024461,
		0.027985,
		0.031435,
		0.034759,
		0.037912,
		0.040853,
		0.043547,
		0.04597,
		0.0481,
		0.049927,
		0.051448,
		0.052669,
		0.053606,
		0.054285,
		0.054739,
		0.055011,
		0.055149,
		0.055202,
		0.055213,
		0.055213,
		0.05521,
		0.055183,
		0.055094,
		0.054895,
		0.054538,
		0.053976,
		0.053172,
		0.052095,
		0.050725,
		0.049052,
		0.047072,
		0.044794,
		0.042233,
		0.039411,
		0.03636,
		0.033116,
		0.029723,
		0.026229,
		0.022689,
		0.019164,
		0.015716,
		0.012416,
		0.009337,
		0.006556,
		0.004154,
		0.002217,
		0.000834,
		9.7e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		0.000205,
		0.000785,
		0.001691,
		0.002874,
		0.004288,
		0.005889,
		0.007633,
		0.009482,
		0.011397,
		0.013341,
		0.015282,
		0.017189,
		0.019033,
		0.02079,
		0.022437,
		0.023956,
		0.025331,
		0.026551,
		0.027608,
		0.0285,
		0.029228,
		0.029799,
		0.030225,
		0.030522,
		0.03071,
		0.030815,
		0.030862,
		0.030876,
		0.030877,
		0.030877,
		0.030871,
		0.030844,
		0.030771,
		0.030628,
		0.030388,
		0.03003,
		0.029533,
		0.028884,
		0.028075,
		0.0271,
		0.025961,
		0.024662,
		0.023214,
		0.021629,
		0.019924,
		0.01812,
		0.016241,
		0.014314,
		0.012367,
		0.010433,
		0.008547,
		0.006746,
		0.005068,
		0.003555,
		0.002251,
		0.0012,
		0.000451,
		5.2e-005,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		9.7e-005,
		0.000371,
		0.000799,
		0.001359,
		0.00203,
		0.00279,
		0.00362,
		0.004501,
		0.005416,
		0.006347,
		0.007279,
		0.008198,
		0.009089,
		0.009942,
		0.010746,
		0.011491,
		0.01217,
		0.012777,
		0.013309,
		0.013763,
		0.014139,
		0.014439,
		0.014669,
		0.014835,
		0.014945,
		0.015011,
		0.015044,
		0.015057,
		0.015059,
		0.015059,
		0.015059,
		0.015052,
		0.015031,
		0.014983,
		0.014896,
		0.014759,
		0.014563,
		0.014298,
		0.01396,
		0.013546,
		0.013053,
		0.012483,
		0.011839,
		0.011126,
		0.010351,
		0.009521,
		0.008648,
		0.007741,
		0.006814,
		0.00588,
		0.004955,
		0.004055,
		0.003197,
		0.0024,
		0.001682,
		0.001064,
		0.000567,
		0.000213,
		2.5e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		3.7e-005,
		0.000143,
		0.000308,
		0.000524,
		0.000783,
		0.001077,
		0.001399,
		0.001742,
		0.002098,
		0.002461,
		0.002826,
		0.003186,
		0.003537,
		0.003874,
		0.004194,
		0.004491,
		0.004764,
		0.005009,
		0.005226,
		0.005413,
		0.005571,
		0.005698,
		0.005798,
		0.005872,
		0.005924,
		0.005956,
		0.005974,
		0.005982,
		0.005985,
		0.005985,
		0.005985,
		0.005984,
		0.005979,
		0.005967,
		0.005942,
		0.005901,
		0.005838,
		0.005752,
		0.005638,
		0.005496,
		0.005324,
		0.005121,
		0.00489,
		0.004631,
		0.004345,
		0.004037,
		0.003708,
		0.003363,
		0.003007,
		0.002644,
		0.002279,
		0.001918,
		0.001568,
		0.001235,
		0.000926,
		0.000649,
		0.00041,
		0.000218,
		8.2e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		1e-005,
		3.9e-005,
		8.4e-005,
		0.000143,
		0.000215,
		0.000295,
		0.000384,
		0.000479,
		0.000577,
		0.000678,
		0.000779,
		0.000879,
		0.000977,
		0.001072,
		0.001162,
		0.001246,
		0.001323,
		0.001394,
		0.001456,
		0.001511,
		0.001557,
		0.001596,
		0.001626,
		0.001649,
		0.001666,
		0.001677,
		0.001684,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001681,
		0.001672,
		0.001658,
		0.001639,
		0.001612,
		0.001578,
		0.001535,
		0.001485,
		0.001426,
		0.00136,
		0.001285,
		0.001205,
		0.001117,
		0.001025,
		0.000929,
		0.000829,
		0.000728,
		0.000627,
		0.000527,
		0.000431,
		0.000339,
		0.000254,
		0.000178,
		0.000112,
		6e-005,
		2.2e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.1e-005,
		2e-005,
		2.9e-005,
		4e-005,
		5.2e-005,
		6.5e-005,
		7.9e-005,
		9.3e-005,
		0.000107,
		0.000121,
		0.000134,
		0.000147,
		0.00016,
		0.000172,
		0.000183,
		0.000193,
		0.000202,
		0.00021,
		0.000216,
		0.000222,
		0.000227,
		0.00023,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000228,
		0.000224,
		0.000219,
		0.000213,
		0.000206,
		0.000197,
		0.000188,
		0.000177,
		0.000166,
		0.000154,
		0.000141,
		0.000127,
		0.000114,
		0.0001,
		8.6e-005,
		7.2e-005,
		5.9e-005,
		4.6e-005,
		3.5e-005,
		2.4e-005,
		1.5e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		1.7e-005,
		2.7e-005,
		3.7e-005,
		4.9e-005,
		6.2e-005,
		7.5e-005,
		8.9e-005,
		0.000103,
		0.000117,
		0.000131,
		0.000144,
		0.000157,
		0.000169,
		0.00018,
		0.00019,
		0.000199,
		0.000208,
		0.000215,
		0.000221,
		0.000225,
		0.000229,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000229,
		0.000225,
		0.000221,
		0.000215,
		0.000208,
		0.000199,
		0.00019,
		0.00018,
		0.000169,
		0.000157,
		0.000144,
		0.000131,
		0.000117,
		0.000103,
		8.9e-005,
		7.5e-005,
		6.2e-005,
		4.9e-005,
		3.7e-005,
		2.7e-005,
		1.7e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.6e-005,
		4.9e-005,
		9.8e-005,
		0.00016,
		0.000234,
		0.000317,
		0.000407,
		0.000503,
		0.000602,
		0.000703,
		0.000804,
		0.000904,
		0.001001,
		0.001095,
		0.001183,
		0.001266,
		0.001342,
		0.00141,
		0.001471,
		0.001523,
		0.001568,
		0.001604,
		0.001633,
		0.001654,
		0.001669,
		0.001679,
		0.001685,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001685,
		0.001679,
		0.001669,
		0.001654,
		0.001633,
		0.001604,
		0.001568,
		0.001523,
		0.001471,
		0.00141,
		0.001342,
		0.001266,
		0.001183,
		0.001095,
		0.001001,
		0.000904,
		0.000804,
		0.000703,
		0.000602,
		0.000503,
		0.000407,
		0.000317,
		0.000234,
		0.00016,
		9.8e-005,
		4.9e-005,
		1.6e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		2.1e-005,
		0.00011,
		0.000261,
		0.000466,
		0.000715,
		0.001001,
		0.001317,
		0.001654,
		0.002008,
		0.00237,
		0.002735,
		0.003097,
		0.003451,
		0.003792,
		0.004116,
		0.004419,
		0.004698,
		0.004951,
		0.005175,
		0.005369,
		0.005534,
		0.005669,
		0.005776,
		0.005856,
		0.005913,
		0.00595,
		0.005971,
		0.005981,
		0.005984,
		0.005985,
		0.005985,
		0.005984,
		0.005981,
		0.005971,
		0.00595,
		0.005913,
		0.005856,
		0.005776,
		0.005669,
		0.005534,
		0.005369,
		0.005175,
		0.004951,
		0.004698,
		0.004419,
		0.004116,
		0.003792,
		0.003451,
		0.003097,
		0.002735,
		0.00237,
		0.002008,
		0.001654,
		0.001317,
		0.001001,
		0.000715,
		0.000466,
		0.000261,
		0.00011,
		2.1e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		0.000149,
		0.000464,
		0.000927,
		0.001517,
		0.002212,
		0.002991,
		0.003836,
		0.004727,
		0.005648,
		0.00658,
		0.007511,
		0.008424,
		0.009307,
		0.010148,
		0.010938,
		0.011667,
		0.012329,
		0.012918,
		0.01343,
		0.013864,
		0.014221,
		0.014503,
		0.014716,
		0.014867,
		0.014965,
		0.015022,
		0.015049,
		0.015058,
		0.015059,
		0.015059,
		0.015058,
		0.015049,
		0.015022,
		0.014965,
		0.014867,
		0.014716,
		0.014503,
		0.014221,
		0.013864,
		0.01343,
		0.012918,
		0.012329,
		0.011667,
		0.010938,
		0.010148,
		0.009307,
		0.008424,
		0.007511,
		0.00658,
		0.005648,
		0.004727,
		0.003836,
		0.002991,
		0.002212,
		0.001517,
		0.000927,
		0.000464,
		0.000149,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		0.000116,
		0.000608,
		0.001437,
		0.002555,
		0.003915,
		0.005473,
		0.007186,
		0.009012,
		0.010914,
		0.012854,
		0.014799,
		0.016717,
		0.018579,
		0.02036,
		0.022037,
		0.023589,
		0.025001,
		0.026261,
		0.027359,
		0.028292,
		0.029061,
		0.029671,
		0.030132,
		0.030459,
		0.030672,
		0.030795,
		0.030854,
		0.030874,
		0.030877,
		0.030877,
		0.030874,
		0.030854,
		0.030795,
		0.030672,
		0.030459,
		0.030132,
		0.029671,
		0.029061,
		0.028292,
		0.027359,
		0.026261,
		0.025001,
		0.023589,
		0.022037,
		0.02036,
		0.018579,
		0.016717,
		0.014799,
		0.012854,
		0.010914,
		0.009012,
		0.007186,
		0.005473,
		0.003915,
		0.002555,
		0.001437,
		0.000608,
		0.000116,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		2.4e-005,
		0.000586,
		0.001816,
		0.003623,
		0.005917,
		0.008611,
		0.011623,
		0.014875,
		0.018292,
		0.021804,
		0.025346,
		0.028856,
		0.03228,
		0.035565,
		0.038668,
		0.041551,
		0.044179,
		0.04653,
		0.048585,
		0.050336,
		0.051781,
		0.052929,
		0.053799,
		0.054418,
		0.054822,
		0.055057,
		0.055169,
		0.055207,
		0.055213,
		0.055213,
		0.055207,
		0.055169,
		0.055057,
		0.054822,
		0.054418,
		0.053799,
		0.052929,
		0.051781,
		0.050336,
		0.048585,
		0.04653,
		0.044179,
		0.041551,
		0.038668,
		0.035565,
		0.03228,
		0.028856,
		0.025346,
		0.021804,
		0.018292,
		0.014875,
		0.011623,
		0.008611,
		0.005917,
		0.003623,
		0.001816,
		0.000586,
		2.4e-005,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		0.000361,
		0.001882,
		0.004443,
		0.007888,
		0.012067,
		0.016836,
		0.022059,
		0.027607,
		0.033356,
		0.039191,
		0.045007,
		0.050704,
		0.056194,
		0.061398,
		0.066246,
		0.070681,
		0.074658,
		0.078143,
		0.081119,
		0.083582,
		0.085543,
		0.087033,
		0.088095,
		0.088791,
		0.089194,
		0.089387,
		0.089454,
		0.089464,
		0.089464,
		0.089454,
		0.089387,
		0.089194,
		0.088791,
		0.088095,
		0.087033,
		0.085543,
		0.083582,
		0.081119,
		0.078143,
		0.074658,
		0.070681,
		0.066246,
		0.061398,
		0.056194,
		0.050704,
		0.045007,
		0.039191,
		0.033356,
		0.027607,
		0.022059,
		0.016836,
		0.012067,
		0.007888,
		0.004443,
		0.001882,
		0.000361,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		6.4e-005,
		0.001528,
		0.004732,
		0.009423,
		0.01536,
		0.02231,
		0.030053,
		0.038377,
		0.047085,
		0.055989,
		0.064916,
		0.073706,
		0.082212,
		0.090306,
		0.097872,
		0.104814,
		0.111057,
		0.116543,
		0.121239,
		0.125134,
		0.128245,
		0.130612,
		0.132304,
		0.133415,
		0.134061,
		0.134371,
		0.134478,
		0.134495,
		0.134495,
		0.134478,
		0.134371,
		0.134061,
		0.133415,
		0.132304,
		0.130612,
		0.128245,
		0.125134,
		0.121239,
		0.116543,
		0.111057,
		0.104814,
		0.097872,
		0.090306,
		0.082212,
		0.073706,
		0.064916,
		0.055989,
		0.047085,
		0.038377,
		0.030053,
		0.02231,
		0.01536,
		0.009423,
		0.004732,
		0.001528,
		6.4e-005,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		0.000826,
		0.004295,
		0.010121,
		0.017935,
		0.027382,
		0.038125,
		0.049843,
		0.062232,
		0.075006,
		0.087897,
		0.10066,
		0.11307,
		0.124925,
		0.136047,
		0.146287,
		0.155521,
		0.163659,
		0.170644,
		0.176453,
		0.181102,
		0.18465,
		0.187192,
		0.188866,
		0.18984,
		0.19031,
		0.190471,
		0.190498,
		0.190498,
		0.190471,
		0.19031,
		0.18984,
		0.188866,
		0.187192,
		0.18465,
		0.181102,
		0.176453,
		0.170644,
		0.163659,
		0.155521,
		0.146287,
		0.136047,
		0.124925,
		0.11307,
		0.10066,
		0.087897,
		0.075006,
		0.062232,
		0.049843,
		0.038125,
		0.027382,
		0.017935,
		0.010121,
		0.004295,
		0.000826,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000131,
		0.003134,
		0.009688,
		0.019257,
		0.031327,
		0.045407,
		0.061029,
		0.077749,
		0.095151,
		0.112844,
		0.130466,
		0.147689,
		0.164213,
		0.179776,
		0.194153,
		0.207159,
		0.218656,
		0.228551,
		0.236802,
		0.243424,
		0.248488,
		0.252127,
		0.254529,
		0.255931,
		0.256608,
		0.256841,
		0.25688,
		0.25688,
		0.256841,
		0.256608,
		0.255931,
		0.254529,
		0.252127,
		0.248488,
		0.243424,
		0.236802,
		0.228551,
		0.218656,
		0.207159,
		0.194153,
		0.179776,
		0.164213,
		0.147689,
		0.130466,
		0.112844,
		0.095151,
		0.077749,
		0.061029,
		0.045407,
		0.031327,
		0.019257,
		0.009688,
		0.003134,
		0.000131,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		0.00155,
		0.008047,
		0.018926,
		0.033468,
		0.050989,
		0.070835,
		0.092384,
		0.115052,
		0.138292,
		0.161595,
		0.184495,
		0.206571,
		0.227449,
		0.246807,
		0.264379,
		0.279959,
		0.293406,
		0.304652,
		0.313702,
		0.320642,
		0.325642,
		0.32895,
		0.330886,
		0.331824,
		0.332147,
		0.332201,
		0.332201,
		0.332147,
		0.331824,
		0.330886,
		0.32895,
		0.325642,
		0.320642,
		0.313702,
		0.304652,
		0.293406,
		0.279959,
		0.264379,
		0.246807,
		0.227449,
		0.206571,
		0.184495,
		0.161595,
		0.138292,
		0.115052,
		0.092384,
		0.070835,
		0.050989,
		0.033468,
		0.018926,
		0.008047,
		0.00155,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.000228,
		0.005442,
		0.016787,
		0.0333,
		0.054055,
		0.07817,
		0.104807,
		0.133176,
		0.162535,
		0.192194,
		0.221518,
		0.249933,
		0.276925,
		0.302051,
		0.324939,
		0.3453,
		0.362928,
		0.377712,
		0.389643,
		0.398819,
		0.405447,
		0.409844,
		0.412426,
		0.413678,
		0.414112,
		0.414183,
		0.414183,
		0.414112,
		0.413678,
		0.412426,
		0.409844,
		0.405447,
		0.398819,
		0.389643,
		0.377712,
		0.362928,
		0.3453,
		0.324939,
		0.302051,
		0.276925,
		0.249933,
		0.221518,
		0.192194,
		0.162535,
		0.133176,
		0.104807,
		0.07817,
		0.054055,
		0.0333,
		0.016787,
		0.005442,
		0.000228,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		0.002517,
		0.013042,
		0.03061,
		0.054013,
		0.082099,
		0.11377,
		0.14799,
		0.183787,
		0.220253,
		0.256552,
		0.291922,
		0.325684,
		0.357244,
		0.386103,
		0.411863,
		0.434238,
		0.453061,
		0.468297,
		0.480048,
		0.48856,
		0.494223,
		0.497557,
		0.499179,
		0.499742,
		0.499835,
		0.499835,
		0.499742,
		0.499179,
		0.497557,
		0.494223,
		0.48856,
		0.480048,
		0.468297,
		0.453061,
		0.434238,
		0.411863,
		0.386103,
		0.357244,
		0.325684,
		0.291922,
		0.256552,
		0.220253,
		0.183787,
		0.14799,
		0.11377,
		0.082099,
		0.054013,
		0.03061,
		0.013042,
		0.002517,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.000349,
		0.008309,
		0.02558,
		0.050629,
		0.08199,
		0.118271,
		0.158149,
		0.200385,
		0.243819,
		0.287379,
		0.330088,
		0.371068,
		0.409549,
		0.444878,
		0.47653,
		0.504116,
		0.527398,
		0.546301,
		0.560924,
		0.571549,
		0.578638,
		0.582824,
		0.584866,
		0.585577,
		0.585694,
		0.585694,
		0.585577,
		0.584866,
		0.582824,
		0.578638,
		0.571549,
		0.560924,
		0.546301,
		0.527398,
		0.504116,
		0.47653,
		0.444878,
		0.409549,
		0.371068,
		0.330088,
		0.287379,
		0.243819,
		0.200385,
		0.158149,
		0.118271,
		0.08199,
		0.050629,
		0.02558,
		0.008309,
		0.000349,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		0.003642,
		0.018834,
		0.044105,
		0.07764,
		0.11771,
		0.162676,
		0.210995,
		0.261224,
		0.312024,
		0.362172,
		0.410564,
		0.456229,
		0.498335,
		0.536205,
		0.569328,
		0.597379,
		0.620228,
		0.637959,
		0.650881,
		0.65953,
		0.664653,
		0.667159,
		0.668034,
		0.668179,
		0.668179,
		0.668034,
		0.667159,
		0.664653,
		0.65953,
		0.650881,
		0.637959,
		0.620228,
		0.597379,
		0.569328,
		0.536205,
		0.498335,
		0.456229,
		0.410564,
		0.362172,
		0.312024,
		0.261224,
		0.210995,
		0.162676,
		0.11771,
		0.07764,
		0.044105,
		0.018834,
		0.003642,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.000482,
		0.011432,
		0.035114,
		0.069331,
		0.111989,
		0.161098,
		0.214785,
		0.271292,
		0.328987,
		0.386375,
		0.442101,
		0.494965,
		0.543936,
		0.588164,
		0.626996,
		0.659998,
		0.686971,
		0.707972,
		0.723327,
		0.733637,
		0.739762,
		0.742769,
		0.743821,
		0.743996,
		0.743996,
		0.743821,
		0.742769,
		0.739762,
		0.733637,
		0.723327,
		0.707972,
		0.686971,
		0.659998,
		0.626996,
		0.588164,
		0.543936,
		0.494965,
		0.442101,
		0.386375,
		0.328987,
		0.271292,
		0.214785,
		0.161098,
		0.111989,
		0.069331,
		0.035114,
		0.011432,
		0.000482,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		0.004798,
		0.024756,
		0.057832,
		0.101537,
		0.15351,
		0.21152,
		0.273471,
		0.337415,
		0.40156,
		0.464277,
		0.524119,
		0.579832,
		0.630373,
		0.674927,
		0.712934,
		0.744109,
		0.768466,
		0.786334,
		0.798372,
		0.805547,
		0.80908,
		0.810321,
		0.810528,
		0.810528,
		0.810321,
		0.80908,
		0.805547,
		0.798372,
		0.786334,
		0.768466,
		0.744109,
		0.712934,
		0.674927,
		0.630373,
		0.579832,
		0.524119,
		0.464277,
		0.40156,
		0.337415,
		0.273471,
		0.21152,
		0.15351,
		0.101537,
		0.057832,
		0.024756,
		0.004798,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.000611,
		0.014463,
		0.044312,
		0.087259,
		0.140548,
		0.201568,
		0.267868,
		0.337163,
		0.407343,
		0.476489,
		0.542884,
		0.605034,
		0.661683,
		0.711837,
		0.754792,
		0.790158,
		0.817888,
		0.838304,
		0.852106,
		0.86036,
		0.864439,
		0.865876,
		0.866116,
		0.866116,
		0.865876,
		0.864439,
		0.86036,
		0.852106,
		0.838304,
		0.817888,
		0.790158,
		0.754792,
		0.711837,
		0.661683,
		0.605034,
		0.542884,
		0.476489,
		0.407343,
		0.337163,
		0.267868,
		0.201568,
		0.140548,
		0.087259,
		0.044312,
		0.014463,
		0.000611,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		0.005873,
		0.030225,
		0.070417,
		0.123278,
		0.185806,
		0.255173,
		0.32874,
		0.404065,
		0.478918,
		0.551297,
		0.619449,
		0.681888,
		0.737423,
		0.785187,
		0.824668,
		0.855742,
		0.878705,
		0.894285,
		0.903638,
		0.908275,
		0.909914,
		0.910189,
		0.910189,
		0.909914,
		0.908275,
		0.903638,
		0.894285,
		0.878705,
		0.855742,
		0.824668,
		0.785187,
		0.737423,
		0.681888,
		0.619449,
		0.551297,
		0.478918,
		0.404065,
		0.32874,
		0.255173,
		0.185806,
		0.123278,
		0.070417,
		0.030225,
		0.005873,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.000727,
		0.017177,
		0.052485,
		0.103053,
		0.165466,
		0.236507,
		0.313165,
		0.392648,
		0.472399,
		0.550115,
		0.623765,
		0.691617,
		0.752265,
		0.80466,
		0.84815,
		0.882517,
		0.908012,
		0.925377,
		0.935839,
		0.941046,
		0.942893,
		0.943204,
		0.943204,
		0.942893,
		0.941046,
		0.935839,
		0.925377,
		0.908012,
		0.882517,
		0.84815,
		0.80466,
		0.752265,
		0.691617,
		0.623765,
		0.550115,
		0.472399,
		0.392648,
		0.313165,
		0.236507,
		0.165466,
		0.103053,
		0.052485,
		0.017177,
		0.000727,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		0.006823,
		0.035019,
		0.081344,
		0.141955,
		0.213225,
		0.291752,
		0.374377,
		0.458198,
		0.540591,
		0.619229,
		0.692115,
		0.757607,
		0.814459,
		0.861858,
		0.899472,
		0.92749,
		0.946649,
		0.958238,
		0.964028,
		0.96609,
		0.966438,
		0.966438,
		0.96609,
		0.964028,
		0.958238,
		0.946649,
		0.92749,
		0.899472,
		0.861858,
		0.814459,
		0.757607,
		0.692115,
		0.619229,
		0.540591,
		0.458198,
		0.374377,
		0.291752,
		0.213225,
		0.141955,
		0.081344,
		0.035019,
		0.006823,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.000831,
		0.019574,
		0.059631,
		0.116705,
		0.186735,
		0.265908,
		0.350675,
		0.437764,
		0.524209,
		0.607367,
		0.684951,
		0.755064,
		0.81624,
		0.867484,
		0.908331,
		0.938887,
		0.95987,
		0.972616,
		0.979009,
		0.981295,
		0.981682,
		0.981682,
		0.981295,
		0.979009,
		0.972616,
		0.95987,
		0.938887,
		0.908331,
		0.867484,
		0.81624,
		0.755064,
		0.684951,
		0.607367,
		0.524209,
		0.437764,
		0.350675,
		0.265908,
		0.186735,
		0.116705,
		0.059631,
		0.019574,
		0.000831,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.007687,
		0.039335,
		0.091069,
		0.158365,
		0.236966,
		0.322902,
		0.412507,
		0.502441,
		0.589719,
		0.671739,
		0.746323,
		0.811757,
		0.866844,
		0.910961,
		0.944112,
		0.966979,
		0.980929,
		0.987957,
		0.990479,
		0.990908,
		0.990908,
		0.990479,
		0.987957,
		0.980929,
		0.966979,
		0.944112,
		0.910961,
		0.866844,
		0.811757,
		0.746323,
		0.671739,
		0.589719,
		0.502441,
		0.412507,
		0.322902,
		0.236966,
		0.158365,
		0.091069,
		0.039335,
		0.007687,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.00093,
		0.021838,
		0.066304,
		0.129299,
		0.206085,
		0.292232,
		0.38364,
		0.476563,
		0.567636,
		0.653913,
		0.732899,
		0.802607,
		0.861608,
		0.909096,
		0.944951,
		0.969798,
		0.985025,
		0.992731,
		0.995509,
		0.995983,
		0.995983,
		0.995509,
		0.992731,
		0.985025,
		0.969798,
		0.944951,
		0.909096,
		0.861608,
		0.802607,
		0.732899,
		0.653913,
		0.567636,
		0.476563,
		0.38364,
		0.292232,
		0.206085,
		0.129299,
		0.066304,
		0.021838,
		0.00093,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.008551,
		0.043608,
		0.100594,
		0.174237,
		0.259602,
		0.352107,
		0.447554,
		0.542159,
		0.632584,
		0.715985,
		0.790063,
		0.853125,
		0.904151,
		0.942872,
		0.969837,
		0.986442,
		0.994885,
		0.997943,
		0.998467,
		0.998467,
		0.997943,
		0.994885,
		0.986442,
		0.969837,
		0.942872,
		0.904151,
		0.853125,
		0.790063,
		0.715985,
		0.632584,
		0.542159,
		0.447554,
		0.352107,
		0.259602,
		0.174237,
		0.100594,
		0.043608,
		0.008551,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.001034,
		0.024212,
		0.07324,
		0.142252,
		0.225742,
		0.31859,
		0.416091,
		0.513985,
		0.608498,
		0.696388,
		0.775,
		0.842337,
		0.897131,
		0.938936,
		0.968199,
		0.986312,
		0.995567,
		0.998935,
		0.999515,
		0.999515,
		0.998935,
		0.995567,
		0.986312,
		0.968199,
		0.938936,
		0.897131,
		0.842337,
		0.775,
		0.696388,
		0.608498,
		0.513985,
		0.416091,
		0.31859,
		0.225742,
		0.142252,
		0.07324,
		0.024212,
		0.001034,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.009502,
		0.048276,
		0.110907,
		0.191247,
		0.283568,
		0.38259,
		0.483514,
		0.582068,
		0.674555,
		0.757914,
		0.829796,
		0.888645,
		0.9338,
		0.965582,
		0.98536,
		0.995519,
		0.999235,
		0.999877,
		0.999877,
		0.999235,
		0.995519,
		0.98536,
		0.965582,
		0.9338,
		0.888645,
		0.829796,
		0.757914,
		0.674555,
		0.582068,
		0.483514,
		0.38259,
		0.283568,
		0.191247,
		0.110907,
		0.048276,
		0.009502,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.001154,
		0.026901,
		0.081038,
		0.156684,
		0.247418,
		0.347303,
		0.450927,
		0.553448,
		0.650647,
		0.738994,
		0.815734,
		0.878971,
		0.927789,
		0.962349,
		0.983977,
		0.99515,
		0.999258,
		0.999972,
		0.999972,
		0.999258,
		0.99515,
		0.983977,
		0.962349,
		0.927789,
		0.878971,
		0.815734,
		0.738994,
		0.650647,
		0.553448,
		0.450927,
		0.347303,
		0.247418,
		0.156684,
		0.081038,
		0.026901,
		0.001154,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310414,
		0.416289,
		0.522637,
		0.624639,
		0.718225,
		0.800162,
		0.86816,
		0.920994,
		0.95863,
		0.982326,
		0.994638,
		0.999192,
		0.999987,
		0.999987,
		0.999192,
		0.994638,
		0.982326,
		0.95863,
		0.920994,
		0.86816,
		0.800162,
		0.718225,
		0.624639,
		0.522637,
		0.416289,
		0.310414,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.013475,
		0.067471,
		0.152563,
		0.258527,
		0.376031,
		0.496716,
		0.613292,
		0.71967,
		0.811114,
		0.884434,
		0.938217,
		0.973047,
		0.991651,
		0.998717,
		0.999981,
		0.999981,
		0.998717,
		0.991651,
		0.973047,
		0.938217,
		0.884434,
		0.811114,
		0.71967,
		0.613292,
		0.496716,
		0.376031,
		0.258527,
		0.152563,
		0.067471,
		0.013475,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.001665,
		0.038254,
		0.113386,
		0.215353,
		0.333432,
		0.457972,
		0.580504,
		0.693878,
		0.792437,
		0.87223,
		0.931276,
		0.969836,
		0.990599,
		0.998546,
		0.999979,
		0.999979,
		0.998546,
		0.990599,
		0.969836,
		0.931276,
		0.87223,
		0.792437,
		0.693878,
		0.580504,
		0.457972,
		0.333432,
		0.215353,
		0.113386,
		0.038254,
		0.001665,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.015366,
		0.076461,
		0.171689,
		0.28869,
		0.416289,
		0.544621,
		0.665277,
		0.771497,
		0.858415,
		0.92335,
		0.966137,
		0.989376,
		0.998347,
		0.999976,
		0.999976,
		0.998347,
		0.989376,
		0.966137,
		0.92335,
		0.858415,
		0.771497,
		0.665277,
		0.544621,
		0.416289,
		0.28869,
		0.171689,
		0.076461,
		0.015366,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.001913,
		0.043674,
		0.128533,
		0.242186,
		0.371655,
		0.505413,
		0.63355,
		0.747985,
		0.842738,
		0.914268,
		0.961859,
		0.987949,
		0.998111,
		0.999972,
		0.999972,
		0.998111,
		0.987949,
		0.961859,
		0.914268,
		0.842738,
		0.747985,
		0.63355,
		0.505413,
		0.371655,
		0.242186,
		0.128533,
		0.043674,
		0.001913,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.017685,
		0.08736,
		0.19457,
		0.324181,
		0.462682,
		0.598361,
		0.721547,
		0.824905,
		0.903828,
		0.956891,
		0.986276,
		0.997833,
		0.999968,
		0.999968,
		0.997833,
		0.986276,
		0.956891,
		0.903828,
		0.824905,
		0.721547,
		0.598361,
		0.462682,
		0.324181,
		0.19457,
		0.08736,
		0.017685,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.002221,
		0.05033,
		0.146884,
		0.27417,
		0.416289,
		0.559361,
		0.691782,
		0.804566,
		0.891784,
		0.951098,
		0.984305,
		0.997501,
		0.999962,
		0.999962,
		0.997501,
		0.984305,
		0.951098,
		0.891784,
		0.804566,
		0.691782,
		0.559361,
		0.416289,
		0.27417,
		0.146884,
		0.05033,
		0.002221,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.02057,
		0.100747,
		0.222223,
		0.366214,
		0.516206,
		0.658245,
		0.781314,
		0.877839,
		0.944313,
		0.981968,
		0.997104,
		0.999956,
		0.999956,
		0.997104,
		0.981968,
		0.944313,
		0.877839,
		0.781314,
		0.658245,
		0.516206,
		0.366214,
		0.222223,
		0.100747,
		0.02057,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.00261,
		0.058623,
		0.169394,
		0.312626,
		0.468589,
		0.620445,
		0.754672,
		0.861636,
		0.936326,
		0.979184,
		0.996625,
		0.999948,
		0.999948,
		0.996625,
		0.979184,
		0.936326,
		0.861636,
		0.754672,
		0.620445,
		0.468589,
		0.312626,
		0.169394,
		0.058623,
		0.00261,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.024223,
		0.117431,
		0.256014,
		0.416289,
		0.577857,
		0.724081,
		0.842738,
		0.926878,
		0.975845,
		0.996042,
		0.999939,
		0.999939,
		0.996042,
		0.975845,
		0.926878,
		0.842738,
		0.724081,
		0.577857,
		0.416289,
		0.256014,
		0.117431,
		0.024223,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.003111,
		0.069138,
		0.197385,
		0.359265,
		0.529947,
		0.688896,
		0.820615,
		0.915644,
		0.971816,
		0.995329,
		0.999927,
		0.999927,
		0.995329,
		0.971816,
		0.915644,
		0.820615,
		0.688896,
		0.529947,
		0.359265,
		0.197385,
		0.069138,
		0.003111,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.028941,
		0.138573,
		0.297795,
		0.476215,
		0.64838,
		0.794618,
		0.902212,
		0.96692,
		0.994448,
		0.999912,
		0.999912,
		0.994448,
		0.96692,
		0.902212,
		0.794618,
		0.64838,
		0.476215,
		0.297795,
		0.138573,
		0.028941,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.003771,
		0.082738,
		0.232725,
		0.416289,
		0.601712,
		0.763962,
		0.886059,
		0.960924,
		0.99335,
		0.999893,
		0.999893,
		0.99335,
		0.960924,
		0.886059,
		0.763962,
		0.601712,
		0.416289,
		0.232725,
		0.082738,
		0.003771,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.035182,
		0.165879,
		0.350084,
		0.548017,
		0.727694,
		0.86652,
		0.953525,
		0.991969,
		0.999869,
		0.999869,
		0.991969,
		0.953525,
		0.86652,
		0.727694,
		0.548017,
		0.350084,
		0.165879,
		0.035182,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.004665,
		0.100747,
		0.278096,
		0.486437,
		0.684672,
		0.842738,
		0.944313,
		0.990212,
		0.999838,
		0.999838,
		0.990212,
		0.944313,
		0.842738,
		0.684672,
		0.486437,
		0.278096,
		0.100747,
		0.004665,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.043674,
		0.201932,
		0.416289,
		0.63355,
		0.813614,
		0.932739,
		0.987949,
		0.999797,
		0.999797,
		0.987949,
		0.932739,
		0.813614,
		0.63355,
		0.416289,
		0.201932,
		0.043674,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.00592,
		0.125272,
		0.337397,
		0.572806,
		0.777728,
		0.918052,
		0.984999,
		0.999741,
		0.999741,
		0.984999,
		0.918052,
		0.777728,
		0.572806,
		0.337397,
		0.125272,
		0.00592,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.055596,
		0.250527,
		0.500403,
		0.732617,
		0.898429,
		0.980234,
		0.998789,
		0.998789,
		0.980234,
		0.898429,
		0.732617,
		0.500403,
		0.250527,
		0.055596,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.007563,
		0.155812,
		0.405858,
		0.660911,
		0.852871,
		0.951392,
		0.974514,
		0.974514,
		0.951392,
		0.852871,
		0.660911,
		0.405858,
		0.155812,
		0.007563,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.061533,
		0.267665,
		0.511233,
		0.707694,
		0.813438,
		0.839261,
		0.839261,
		0.813438,
		0.707694,
		0.511233,
		0.267665,
		0.061533,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.005759,
		0.1143,
		0.283774,
		0.434461,
		0.520504,
		0.542528,
		0.542528,
		0.520504,
		0.434461,
		0.283774,
		0.1143,
		0.005759,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.022412,
		0.092608,
		0.165276,
		0.210073,
		0.222209,
		0.222209,
		0.210073,
		0.165276,
		0.092608,
		0.022412,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.000644,
		0.012091,
		0.027863,
		0.038672,
		0.041813,
		0.041813,
		0.038672,
		0.027863,
		0.012091,
		0.000644,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		0.000106,
		0.000402,
		0.000642,
		0.000718,
		0.000718,
		0.000642,
		0.000402,
		0.000106,
		0.000106,
		0.000402,
		0.000642,
		0.000718,
		0.000718,
		0.000642,
		0.000402,
		0.000106,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		0.000644,
		0.012091,
		0.027863,
		0.038672,
		0.041813,
		0.041813,
		0.038672,
		0.027863,
		0.012091,
		0.000644,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.022412,
		0.092608,
		0.165276,
		0.210073,
		0.222209,
		0.222209,
		0.210073,
		0.165276,
		0.092608,
		0.022412,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.005759,
		0.1143,
		0.283774,
		0.434461,
		0.520504,
		0.542528,
		0.542528,
		0.520504,
		0.434461,
		0.283774,
		0.1143,
		0.005759,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.061533,
		0.267665,
		0.511233,
		0.707694,
		0.813438,
		0.839261,
		0.839261,
		0.813438,
		0.707694,
		0.511233,
		0.267665,
		0.061533,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.007563,
		0.155812,
		0.405858,
		0.660911,
		0.852871,
		0.951392,
		0.974514,
		0.974514,
		0.951392,
		0.852871,
		0.660911,
		0.405858,
		0.155812,
		0.007563,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.055596,
		0.250527,
		0.500403,
		0.732617,
		0.898429,
		0.980234,
		0.998789,
		0.998789,
		0.980234,
		0.898429,
		0.732617,
		0.500403,
		0.250527,
		0.055596,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.00592,
		0.125272,
		0.337397,
		0.572806,
		0.777728,
		0.918052,
		0.984999,
		0.999741,
		0.999741,
		0.984999,
		0.918052,
		0.777728,
		0.572806,
		0.337397,
		0.125272,
		0.00592,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.043674,
		0.201932,
		0.416289,
		0.63355,
		0.813614,
		0.932739,
		0.987949,
		0.999797,
		0.999797,
		0.987949,
		0.932739,
		0.813614,
		0.63355,
		0.416289,
		0.201932,
		0.043674,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.004665,
		0.100747,
		0.278096,
		0.486437,
		0.684672,
		0.842738,
		0.944313,
		0.990212,
		0.999838,
		0.999838,
		0.990212,
		0.944313,
		0.842738,
		0.684672,
		0.486437,
		0.278096,
		0.100747,
		0.004665,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.035182,
		0.165879,
		0.350084,
		0.548017,
		0.727694,
		0.86652,
		0.953525,
		0.991969,
		0.999869,
		0.999869,
		0.991969,
		0.953525,
		0.86652,
		0.727694,
		0.548017,
		0.350084,
		0.165879,
		0.035182,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.003771,
		0.082738,
		0.232725,
		0.416289,
		0.601712,
		0.763962,
		0.886059,
		0.960924,
		0.99335,
		0.999893,
		0.999893,
		0.99335,
		0.960924,
		0.886059,
		0.763962,
		0.601712,
		0.416289,
		0.232725,
		0.082738,
		0.003771,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.028941,
		0.138573,
		0.297795,
		0.476215,
		0.64838,
		0.794618,
		0.902212,
		0.96692,
		0.994448,
		0.999912,
		0.999912,
		0.994448,
		0.96692,
		0.902212,
		0.794618,
		0.64838,
		0.476215,
		0.297795,
		0.138573,
		0.028941,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.003111,
		0.069138,
		0.197385,
		0.359265,
		0.529947,
		0.688896,
		0.820615,
		0.915644,
		0.971816,
		0.995329,
		0.999927,
		0.999927,
		0.995329,
		0.971816,
		0.915644,
		0.820615,
		0.688896,
		0.529947,
		0.359265,
		0.197385,
		0.069138,
		0.003111,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.024223,
		0.117431,
		0.256014,
		0.416289,
		0.577857,
		0.724081,
		0.842738,
		0.926878,
		0.975845,
		0.996042,
		0.999939,
		0.999939,
		0.996042,
		0.975845,
		0.926878,
		0.842738,
		0.724081,
		0.577857,
		0.416289,
		0.256014,
		0.117431,
		0.024223,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.00261,
		0.058623,
		0.169394,
		0.312626,
		0.468589,
		0.620445,
		0.754672,
		0.861636,
		0.936326,
		0.979184,
		0.996625,
		0.999948,
		0.999948,
		0.996625,
		0.979184,
		0.936326,
		0.861636,
		0.754672,
		0.620445,
		0.468589,
		0.312626,
		0.169394,
		0.058623,
		0.00261,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.02057,
		0.100747,
		0.222223,
		0.366214,
		0.516206,
		0.658245,
		0.781314,
		0.877839,
		0.944313,
		0.981968,
		0.997104,
		0.999956,
		0.999956,
		0.997104,
		0.981968,
		0.944313,
		0.877839,
		0.781314,
		0.658245,
		0.516206,
		0.366214,
		0.222223,
		0.100747,
		0.02057,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.002221,
		0.05033,
		0.146884,
		0.27417,
		0.416289,
		0.559361,
		0.691782,
		0.804566,
		0.891784,
		0.951098,
		0.984305,
		0.997501,
		0.999962,
		0.999962,
		0.997501,
		0.984305,
		0.951098,
		0.891784,
		0.804566,
		0.691782,
		0.559361,
		0.416289,
		0.27417,
		0.146884,
		0.05033,
		0.002221,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.017685,
		0.08736,
		0.19457,
		0.324181,
		0.462682,
		0.598361,
		0.721547,
		0.824905,
		0.903828,
		0.956891,
		0.986276,
		0.997833,
		0.999968,
		0.999968,
		0.997833,
		0.986276,
		0.956891,
		0.903828,
		0.824905,
		0.721547,
		0.598361,
		0.462682,
		0.324181,
		0.19457,
		0.08736,
		0.017685,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.001913,
		0.043674,
		0.128533,
		0.242186,
		0.371655,
		0.505413,
		0.63355,
		0.747985,
		0.842738,
		0.914268,
		0.961859,
		0.987949,
		0.998111,
		0.999972,
		0.999972,
		0.998111,
		0.987949,
		0.961859,
		0.914268,
		0.842738,
		0.747985,
		0.63355,
		0.505413,
		0.371655,
		0.242186,
		0.128533,
		0.043674,
		0.001913,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.015366,
		0.076461,
		0.171689,
		0.28869,
		0.416289,
		0.544621,
		0.665277,
		0.771497,
		0.858415,
		0.92335,
		0.966137,
		0.989376,
		0.998347,
		0.999976,
		0.999976,
		0.998347,
		0.989376,
		0.966137,
		0.92335,
		0.858415,
		0.771497,
		0.665277,
		0.544621,
		0.416289,
		0.28869,
		0.171689,
		0.076461,
		0.015366,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.001665,
		0.038254,
		0.113386,
		0.215353,
		0.333432,
		0.457972,
		0.580504,
		0.693878,
		0.792437,
		0.87223,
		0.931276,
		0.969836,
		0.990599,
		0.998546,
		0.999979,
		0.999979,
		0.998546,
		0.990599,
		0.969836,
		0.931276,
		0.87223,
		0.792437,
		0.693878,
		0.580504,
		0.457972,
		0.333432,
		0.215353,
		0.113386,
		0.038254,
		0.001665,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.013475,
		0.067471,
		0.152563,
		0.258527,
		0.376031,
		0.496716,
		0.613292,
		0.71967,
		0.811114,
		0.884434,
		0.938217,
		0.973047,
		0.991651,
		0.998717,
		0.999981,
		0.999981,
		0.998717,
		0.991651,
		0.973047,
		0.938217,
		0.884434,
		0.811114,
		0.71967,
		0.613292,
		0.496716,
		0.376031,
		0.258527,
		0.152563,
		0.067471,
		0.013475,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310414,
		0.416289,
		0.522637,
		0.624639,
		0.718225,
		0.800162,
		0.86816,
		0.920994,
		0.95863,
		0.982326,
		0.994638,
		0.999192,
		0.999987,
		0.999987,
		0.999192,
		0.994638,
		0.982326,
		0.95863,
		0.920994,
		0.86816,
		0.800162,
		0.718225,
		0.624639,
		0.522637,
		0.416289,
		0.310414,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.001154,
		0.026901,
		0.081038,
		0.156684,
		0.247418,
		0.347303,
		0.450927,
		0.553448,
		0.650647,
		0.738994,
		0.815734,
		0.878971,
		0.927789,
		0.962349,
		0.983977,
		0.99515,
		0.999258,
		0.999972,
		0.999972,
		0.999258,
		0.99515,
		0.983977,
		0.962349,
		0.927789,
		0.878971,
		0.815734,
		0.738994,
		0.650647,
		0.553448,
		0.450927,
		0.347303,
		0.247418,
		0.156684,
		0.081038,
		0.026901,
		0.001154,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.009502,
		0.048276,
		0.110907,
		0.191247,
		0.283568,
		0.38259,
		0.483514,
		0.582068,
		0.674555,
		0.757914,
		0.829796,
		0.888645,
		0.9338,
		0.965582,
		0.98536,
		0.995519,
		0.999235,
		0.999877,
		0.999877,
		0.999235,
		0.995519,
		0.98536,
		0.965582,
		0.9338,
		0.888645,
		0.829796,
		0.757914,
		0.674555,
		0.582068,
		0.483514,
		0.38259,
		0.283568,
		0.191247,
		0.110907,
		0.048276,
		0.009502,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.001034,
		0.024212,
		0.07324,
		0.142252,
		0.225742,
		0.31859,
		0.416091,
		0.513985,
		0.608498,
		0.696388,
		0.775,
		0.842337,
		0.897131,
		0.938936,
		0.968199,
		0.986312,
		0.995567,
		0.998935,
		0.999515,
		0.999515,
		0.998935,
		0.995567,
		0.986312,
		0.968199,
		0.938936,
		0.897131,
		0.842337,
		0.775,
		0.696388,
		0.608498,
		0.513985,
		0.416091,
		0.31859,
		0.225742,
		0.142252,
		0.07324,
		0.024212,
		0.001034,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.008551,
		0.043608,
		0.100594,
		0.174237,
		0.259602,
		0.352107,
		0.447554,
		0.542159,
		0.632584,
		0.715985,
		0.790063,
		0.853125,
		0.904151,
		0.942872,
		0.969837,
		0.986442,
		0.994885,
		0.997943,
		0.998467,
		0.998467,
		0.997943,
		0.994885,
		0.986442,
		0.969837,
		0.942872,
		0.904151,
		0.853125,
		0.790063,
		0.715985,
		0.632584,
		0.542159,
		0.447554,
		0.352107,
		0.259602,
		0.174237,
		0.100594,
		0.043608,
		0.008551,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.00093,
		0.021838,
		0.066304,
		0.129299,
		0.206085,
		0.292232,
		0.38364,
		0.476563,
		0.567636,
		0.653913,
		0.732899,
		0.802607,
		0.861608,
		0.909096,
		0.944951,
		0.969798,
		0.985025,
		0.992731,
		0.995509,
		0.995983,
		0.995983,
		0.995509,
		0.992731,
		0.985025,
		0.969798,
		0.944951,
		0.909096,
		0.861608,
		0.802607,
		0.732899,
		0.653913,
		0.567636,
		0.476563,
		0.38364,
		0.292232,
		0.206085,
		0.129299,
		0.066304,
		0.021838,
		0.00093,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.007687,
		0.039335,
		0.091069,
		0.158365,
		0.236966,
		0.322902,
		0.412507,
		0.502441,
		0.589719,
		0.671739,
		0.746323,
		0.811757,
		0.866844,
		0.910961,
		0.944112,
		0.966979,
		0.980929,
		0.987957,
		0.990479,
		0.990908,
		0.990908,
		0.990479,
		0.987957,
		0.980929,
		0.966979,
		0.944112,
		0.910961,
		0.866844,
		0.811757,
		0.746323,
		0.671739,
		0.589719,
		0.502441,
		0.412507,
		0.322902,
		0.236966,
		0.158365,
		0.091069,
		0.039335,
		0.007687,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.000831,
		0.019574,
		0.059631,
		0.116705,
		0.186735,
		0.265908,
		0.350675,
		0.437764,
		0.524209,
		0.607367,
		0.684951,
		0.755064,
		0.81624,
		0.867484,
		0.908331,
		0.938887,
		0.95987,
		0.972616,
		0.979009,
		0.981295,
		0.981682,
		0.981682,
		0.981295,
		0.979009,
		0.972616,
		0.95987,
		0.938887,
		0.908331,
		0.867484,
		0.81624,
		0.755064,
		0.684951,
		0.607367,
		0.524209,
		0.437764,
		0.350675,
		0.265908,
		0.186735,
		0.116705,
		0.059631,
		0.019574,
		0.000831,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.006823,
		0.035019,
		0.081344,
		0.141955,
		0.213225,
		0.291752,
		0.374377,
		0.458198,
		0.540591,
		0.619229,
		0.692115,
		0.757607,
		0.814459,
		0.861858,
		0.899472,
		0.92749,
		0.946649,
		0.958238,
		0.964028,
		0.96609,
		0.966438,
		0.966438,
		0.96609,
		0.964028,
		0.958238,
		0.946649,
		0.92749,
		0.899472,
		0.861858,
		0.814459,
		0.757607,
		0.692115,
		0.619229,
		0.540591,
		0.458198,
		0.374377,
		0.291752,
		0.213225,
		0.141955,
		0.081344,
		0.035019,
		0.006823,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		0.000727,
		0.017177,
		0.052485,
		0.103053,
		0.165466,
		0.236507,
		0.313165,
		0.392648,
		0.472399,
		0.550115,
		0.623765,
		0.691617,
		0.752265,
		0.80466,
		0.84815,
		0.882517,
		0.908012,
		0.925377,
		0.935839,
		0.941046,
		0.942893,
		0.943204,
		0.943204,
		0.942893,
		0.941046,
		0.935839,
		0.925377,
		0.908012,
		0.882517,
		0.84815,
		0.80466,
		0.752265,
		0.691617,
		0.623765,
		0.550115,
		0.472399,
		0.392648,
		0.313165,
		0.236507,
		0.165466,
		0.103053,
		0.052485,
		0.017177,
		0.000727,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.005873,
		0.030225,
		0.070417,
		0.123278,
		0.185806,
		0.255173,
		0.32874,
		0.404065,
		0.478918,
		0.551297,
		0.619449,
		0.681888,
		0.737423,
		0.785187,
		0.824668,
		0.855742,
		0.878705,
		0.894285,
		0.903638,
		0.908275,
		0.909914,
		0.910189,
		0.910189,
		0.909914,
		0.908275,
		0.903638,
		0.894285,
		0.878705,
		0.855742,
		0.824668,
		0.785187,
		0.737423,
		0.681888,
		0.619449,
		0.551297,
		0.478918,
		0.404065,
		0.32874,
		0.255173,
		0.185806,
		0.123278,
		0.070417,
		0.030225,
		0.005873,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		0.000611,
		0.014463,
		0.044312,
		0.087259,
		0.140548,
		0.201568,
		0.267868,
		0.337163,
		0.407343,
		0.476489,
		0.542884,
		0.605034,
		0.661683,
		0.711837,
		0.754792,
		0.790158,
		0.817888,
		0.838304,
		0.852106,
		0.86036,
		0.864439,
		0.865876,
		0.866116,
		0.866116,
		0.865876,
		0.864439,
		0.86036,
		0.852106,
		0.838304,
		0.817888,
		0.790158,
		0.754792,
		0.711837,
		0.661683,
		0.605034,
		0.542884,
		0.476489,
		0.407343,
		0.337163,
		0.267868,
		0.201568,
		0.140548,
		0.087259,
		0.044312,
		0.014463,
		0.000611,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.004798,
		0.024756,
		0.057832,
		0.101537,
		0.15351,
		0.21152,
		0.273471,
		0.337415,
		0.40156,
		0.464277,
		0.524119,
		0.579832,
		0.630373,
		0.674927,
		0.712934,
		0.744109,
		0.768466,
		0.786334,
		0.798372,
		0.805547,
		0.80908,
		0.810321,
		0.810528,
		0.810528,
		0.810321,
		0.80908,
		0.805547,
		0.798372,
		0.786334,
		0.768466,
		0.744109,
		0.712934,
		0.674927,
		0.630373,
		0.579832,
		0.524119,
		0.464277,
		0.40156,
		0.337415,
		0.273471,
		0.21152,
		0.15351,
		0.101537,
		0.057832,
		0.024756,
		0.004798,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		0.000482,
		0.011432,
		0.035114,
		0.069331,
		0.111989,
		0.161098,
		0.214785,
		0.271292,
		0.328987,
		0.386375,
		0.442101,
		0.494965,
		0.543936,
		0.588164,
		0.626996,
		0.659998,
		0.686971,
		0.707972,
		0.723327,
		0.733637,
		0.739762,
		0.742769,
		0.743821,
		0.743996,
		0.743996,
		0.743821,
		0.742769,
		0.739762,
		0.733637,
		0.723327,
		0.707972,
		0.686971,
		0.659998,
		0.626996,
		0.588164,
		0.543936,
		0.494965,
		0.442101,
		0.386375,
		0.328987,
		0.271292,
		0.214785,
		0.161098,
		0.111989,
		0.069331,
		0.035114,
		0.011432,
		0.000482,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.003642,
		0.018834,
		0.044105,
		0.07764,
		0.11771,
		0.162676,
		0.210995,
		0.261224,
		0.312024,
		0.362172,
		0.410564,
		0.456229,
		0.498335,
		0.536205,
		0.569328,
		0.597379,
		0.620228,
		0.637959,
		0.650881,
		0.65953,
		0.664653,
		0.667159,
		0.668034,
		0.668179,
		0.668179,
		0.668034,
		0.667159,
		0.664653,
		0.65953,
		0.650881,
		0.637959,
		0.620228,
		0.597379,
		0.569328,
		0.536205,
		0.498335,
		0.456229,
		0.410564,
		0.362172,
		0.312024,
		0.261224,
		0.210995,
		0.162676,
		0.11771,
		0.07764,
		0.044105,
		0.018834,
		0.003642,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		0.000349,
		0.008309,
		0.02558,
		0.050629,
		0.08199,
		0.118271,
		0.158149,
		0.200385,
		0.243819,
		0.287379,
		0.330088,
		0.371068,
		0.409549,
		0.444878,
		0.47653,
		0.504116,
		0.527398,
		0.546301,
		0.560924,
		0.571549,
		0.578638,
		0.582824,
		0.584866,
		0.585577,
		0.585694,
		0.585694,
		0.585577,
		0.584866,
		0.582824,
		0.578638,
		0.571549,
		0.560924,
		0.546301,
		0.527398,
		0.504116,
		0.47653,
		0.444878,
		0.409549,
		0.371068,
		0.330088,
		0.287379,
		0.243819,
		0.200385,
		0.158149,
		0.118271,
		0.08199,
		0.050629,
		0.02558,
		0.008309,
		0.000349,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.002517,
		0.013042,
		0.03061,
		0.054013,
		0.082099,
		0.11377,
		0.14799,
		0.183787,
		0.220253,
		0.256552,
		0.291922,
		0.325684,
		0.357244,
		0.386103,
		0.411863,
		0.434238,
		0.453061,
		0.468297,
		0.480048,
		0.48856,
		0.494223,
		0.497557,
		0.499179,
		0.499742,
		0.499835,
		0.499835,
		0.499742,
		0.499179,
		0.497557,
		0.494223,
		0.48856,
		0.480048,
		0.468297,
		0.453061,
		0.434238,
		0.411863,
		0.386103,
		0.357244,
		0.325684,
		0.291922,
		0.256552,
		0.220253,
		0.183787,
		0.14799,
		0.11377,
		0.082099,
		0.054013,
		0.03061,
		0.013042,
		0.002517,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		0.000228,
		0.005442,
		0.016787,
		0.0333,
		0.054055,
		0.07817,
		0.104807,
		0.133176,
		0.162535,
		0.192194,
		0.221518,
		0.249933,
		0.276925,
		0.302051,
		0.324939,
		0.3453,
		0.362928,
		0.377712,
		0.389643,
		0.398819,
		0.405447,
		0.409844,
		0.412426,
		0.413678,
		0.414112,
		0.414183,
		0.414183,
		0.414112,
		0.413678,
		0.412426,
		0.409844,
		0.405447,
		0.398819,
		0.389643,
		0.377712,
		0.362928,
		0.3453,
		0.324939,
		0.302051,
		0.276925,
		0.249933,
		0.221518,
		0.192194,
		0.162535,
		0.133176,
		0.104807,
		0.07817,
		0.054055,
		0.0333,
		0.016787,
		0.005442,
		0.000228,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.00155,
		0.008047,
		0.018926,
		0.033468,
		0.050989,
		0.070835,
		0.092384,
		0.115052,
		0.138292,
		0.161595,
		0.184495,
		0.206571,
		0.227449,
		0.246807,
		0.264379,
		0.279959,
		0.293406,
		0.304652,
		0.313702,
		0.320642,
		0.325642,
		0.32895,
		0.330886,
		0.331824,
		0.332147,
		0.332201,
		0.332201,
		0.332147,
		0.331824,
		0.330886,
		0.32895,
		0.325642,
		0.320642,
		0.313702,
		0.304652,
		0.293406,
		0.279959,
		0.264379,
		0.246807,
		0.227449,
		0.206571,
		0.184495,
		0.161595,
		0.138292,
		0.115052,
		0.092384,
		0.070835,
		0.050989,
		0.033468,
		0.018926,
		0.008047,
		0.00155,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		0.000131,
		0.003134,
		0.009688,
		0.019257,
		0.031327,
		0.045407,
		0.061029,
		0.077749,
		0.095151,
		0.112844,
		0.130466,
		0.147689,
		0.164213,
		0.179776,
		0.194153,
		0.207159,
		0.218656,
		0.228551,
		0.236802,
		0.243424,
		0.248488,
		0.252127,
		0.254529,
		0.255931,
		0.256608,
		0.256841,
		0.25688,
		0.25688,
		0.256841,
		0.256608,
		0.255931,
		0.254529,
		0.252127,
		0.248488,
		0.243424,
		0.236802,
		0.228551,
		0.218656,
		0.207159,
		0.194153,
		0.179776,
		0.164213,
		0.147689,
		0.130466,
		0.112844,
		0.095151,
		0.077749,
		0.061029,
		0.045407,
		0.031327,
		0.019257,
		0.009688,
		0.003134,
		0.000131,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000826,
		0.004295,
		0.010121,
		0.017935,
		0.027382,
		0.038125,
		0.049843,
		0.062232,
		0.075006,
		0.087897,
		0.10066,
		0.11307,
		0.124925,
		0.136047,
		0.146287,
		0.155521,
		0.163659,
		0.170644,
		0.176453,
		0.181102,
		0.18465,
		0.187192,
		0.188866,
		0.18984,
		0.19031,
		0.190471,
		0.190498,
		0.190498,
		0.190471,
		0.19031,
		0.18984,
		0.188866,
		0.187192,
		0.18465,
		0.181102,
		0.176453,
		0.170644,
		0.163659,
		0.155521,
		0.146287,
		0.136047,
		0.124925,
		0.11307,
		0.10066,
		0.087897,
		0.075006,
		0.062232,
		0.049843,
		0.038125,
		0.027382,
		0.017935,
		0.010121,
		0.004295,
		0.000826,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		6.4e-005,
		0.001528,
		0.004732,
		0.009423,
		0.01536,
		0.02231,
		0.030053,
		0.038377,
		0.047085,
		0.055989,
		0.064916,
		0.073706,
		0.082212,
		0.090306,
		0.097872,
		0.104814,
		0.111057,
		0.116543,
		0.121239,
		0.125134,
		0.128245,
		0.130612,
		0.132304,
		0.133415,
		0.134061,
		0.134371,
		0.134478,
		0.134495,
		0.134495,
		0.134478,
		0.134371,
		0.134061,
		0.133415,
		0.132304,
		0.130612,
		0.128245,
		0.125134,
		0.121239,
		0.116543,
		0.111057,
		0.104814,
		0.097872,
		0.090306,
		0.082212,
		0.073706,
		0.064916,
		0.055989,
		0.047085,
		0.038377,
		0.030053,
		0.02231,
		0.01536,
		0.009423,
		0.004732,
		0.001528,
		6.4e-005,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000361,
		0.001882,
		0.004443,
		0.007888,
		0.012067,
		0.016836,
		0.022059,
		0.027607,
		0.033356,
		0.039191,
		0.045007,
		0.050704,
		0.056194,
		0.061398,
		0.066246,
		0.070681,
		0.074658,
		0.078143,
		0.081119,
		0.083582,
		0.085543,
		0.087033,
		0.088095,
		0.088791,
		0.089194,
		0.089387,
		0.089454,
		0.089464,
		0.089464,
		0.089454,
		0.089387,
		0.089194,
		0.088791,
		0.088095,
		0.087033,
		0.085543,
		0.083582,
		0.081119,
		0.078143,
		0.074658,
		0.070681,
		0.066246,
		0.061398,
		0.056194,
		0.050704,
		0.045007,
		0.039191,
		0.033356,
		0.027607,
		0.022059,
		0.016836,
		0.012067,
		0.007888,
		0.004443,
		0.001882,
		0.000361,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		2.4e-005,
		0.000586,
		0.001816,
		0.003623,
		0.005917,
		0.008611,
		0.011623,
		0.014875,
		0.018292,
		0.021804,
		0.025346,
		0.028856,
		0.03228,
		0.035565,
		0.038668,
		0.041551,
		0.044179,
		0.04653,
		0.048585,
		0.050336,
		0.051781,
		0.052929,
		0.053799,
		0.054418,
		0.054822,
		0.055057,
		0.055169,
		0.055207,
		0.055213,
		0.055213,
		0.055207,
		0.055169,
		0.055057,
		0.054822,
		0.054418,
		0.053799,
		0.052929,
		0.051781,
		0.050336,
		0.048585,
		0.04653,
		0.044179,
		0.041551,
		0.038668,
		0.035565,
		0.03228,
		0.028856,
		0.025346,
		0.021804,
		0.018292,
		0.014875,
		0.011623,
		0.008611,
		0.005917,
		0.003623,
		0.001816,
		0.000586,
		2.4e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		0.000116,
		0.000608,
		0.001437,
		0.002555,
		0.003915,
		0.005473,
		0.007186,
		0.009012,
		0.010914,
		0.012854,
		0.014799,
		0.016717,
		0.018579,
		0.02036,
		0.022037,
		0.023589,
		0.025001,
		0.026261,
		0.027359,
		0.028292,
		0.029061,
		0.029671,
		0.030132,
		0.030459,
		0.030672,
		0.030795,
		0.030854,
		0.030874,
		0.030877,
		0.030877,
		0.030874,
		0.030854,
		0.030795,
		0.030672,
		0.030459,
		0.030132,
		0.029671,
		0.029061,
		0.028292,
		0.027359,
		0.026261,
		0.025001,
		0.023589,
		0.022037,
		0.02036,
		0.018579,
		0.016717,
		0.014799,
		0.012854,
		0.010914,
		0.009012,
		0.007186,
		0.005473,
		0.003915,
		0.002555,
		0.001437,
		0.000608,
		0.000116,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		0.000149,
		0.000464,
		0.000927,
		0.001517,
		0.002212,
		0.002991,
		0.003836,
		0.004727,
		0.005648,
		0.00658,
		0.007511,
		0.008424,
		0.009307,
		0.010148,
		0.010938,
		0.011667,
		0.012329,
		0.012918,
		0.01343,
		0.013864,
		0.014221,
		0.014503,
		0.014716,
		0.014867,
		0.014965,
		0.015022,
		0.015049,
		0.015058,
		0.015059,
		0.015059,
		0.015058,
		0.015049,
		0.015022,
		0.014965,
		0.014867,
		0.014716,
		0.014503,
		0.014221,
		0.013864,
		0.01343,
		0.012918,
		0.012329,
		0.011667,
		0.010938,
		0.010148,
		0.009307,
		0.008424,
		0.007511,
		0.00658,
		0.005648,
		0.004727,
		0.003836,
		0.002991,
		0.002212,
		0.001517,
		0.000927,
		0.000464,
		0.000149,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		2.1e-005,
		0.00011,
		0.000261,
		0.000466,
		0.000715,
		0.001001,
		0.001317,
		0.001654,
		0.002008,
		0.00237,
		0.002735,
		0.003097,
		0.003451,
		0.003792,
		0.004116,
		0.004419,
		0.004698,
		0.004951,
		0.005175,
		0.005369,
		0.005534,
		0.005669,
		0.005776,
		0.005856,
		0.005913,
		0.00595,
		0.005971,
		0.005981,
		0.005984,
		0.005985,
		0.005985,
		0.005984,
		0.005981,
		0.005971,
		0.00595,
		0.005913,
		0.005856,
		0.005776,
		0.005669,
		0.005534,
		0.005369,
		0.005175,
		0.004951,
		0.004698,
		0.004419,
		0.004116,
		0.003792,
		0.003451,
		0.003097,
		0.002735,
		0.00237,
		0.002008,
		0.001654,
		0.001317,
		0.001001,
		0.000715,
		0.000466,
		0.000261,
		0.00011,
		2.1e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		1.6e-005,
		4.9e-005,
		9.8e-005,
		0.00016,
		0.000234,
		0.000317,
		0.000407,
		0.000503,
		0.000602,
		0.000703,
		0.000804,
		0.000904,
		0.001001,
		0.001095,
		0.001183,
		0.001266,
		0.001342,
		0.00141,
		0.001471,
		0.001523,
		0.001568,
		0.001604,
		0.001633,
		0.001654,
		0.001669,
		0.001679,
		0.001685,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001685,
		0.001679,
		0.001669,
		0.001654,
		0.001633,
		0.001604,
		0.001568,
		0.001523,
		0.001471,
		0.00141,
		0.001342,
		0.001266,
		0.001183,
		0.001095,
		0.001001,
		0.000904,
		0.000804,
		0.000703,
		0.000602,
		0.000503,
		0.000407,
		0.000317,
		0.000234,
		0.00016,
		9.8e-005,
		4.9e-005,
		1.6e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.7e-005,
		2.7e-005,
		3.7e-005,
		4.9e-005,
		6.2e-005,
		7.5e-005,
		8.9e-005,
		0.000103,
		0.000117,
		0.000131,
		0.000144,
		0.000157,
		0.000169,
		0.00018,
		0.00019,
		0.000199,
		0.000208,
		0.000215,
		0.000221,
		0.000225,
		0.000229,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000229,
		0.000225,
		0.000221,
		0.000215,
		0.000208,
		0.000199,
		0.00019,
		0.00018,
		0.000169,
		0.000157,
		0.000144,
		0.000131,
		0.000117,
		0.000103,
		8.9e-005,
		7.5e-005,
		6.2e-005,
		4.9e-005,
		3.7e-005,
		2.7e-005,
		1.7e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		1.1e-005,
		2e-005,
		2.9e-005,
		4e-005,
		5.2e-005,
		6.5e-005,
		7.9e-005,
		9.3e-005,
		0.000107,
		0.000121,
		0.000134,
		0.000147,
		0.00016,
		0.000172,
		0.000183,
		0.000193,
		0.000202,
		0.00021,
		0.000216,
		0.000222,
		0.000227,
		0.00023,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000235,
		0.000234,
		0.000232,
		0.000228,
		0.000224,
		0.000219,
		0.000213,
		0.000206,
		0.000197,
		0.000188,
		0.000177,
		0.000166,
		0.000154,
		0.000141,
		0.000127,
		0.000114,
		0.0001,
		8.6e-005,
		7.2e-005,
		5.9e-005,
		4.6e-005,
		3.5e-005,
		2.4e-005,
		1.5e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1e-005,
		3.9e-005,
		8.4e-005,
		0.000143,
		0.000215,
		0.000295,
		0.000384,
		0.000479,
		0.000577,
		0.000678,
		0.000779,
		0.000879,
		0.000977,
		0.001072,
		0.001162,
		0.001246,
		0.001323,
		0.001394,
		0.001456,
		0.001511,
		0.001557,
		0.001596,
		0.001626,
		0.001649,
		0.001666,
		0.001677,
		0.001684,
		0.001687,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001681,
		0.001672,
		0.001658,
		0.001639,
		0.001612,
		0.001578,
		0.001535,
		0.001485,
		0.001426,
		0.00136,
		0.001285,
		0.001205,
		0.001117,
		0.001025,
		0.000929,
		0.000829,
		0.000728,
		0.000627,
		0.000527,
		0.000431,
		0.000339,
		0.000254,
		0.000178,
		0.000112,
		6e-005,
		2.2e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		3.7e-005,
		0.000143,
		0.000308,
		0.000524,
		0.000783,
		0.001077,
		0.001399,
		0.001742,
		0.002098,
		0.002461,
		0.002826,
		0.003186,
		0.003537,
		0.003874,
		0.004194,
		0.004491,
		0.004764,
		0.005009,
		0.005226,
		0.005413,
		0.005571,
		0.005698,
		0.005798,
		0.005872,
		0.005924,
		0.005956,
		0.005974,
		0.005982,
		0.005985,
		0.005985,
		0.005985,
		0.005984,
		0.005979,
		0.005967,
		0.005942,
		0.005901,
		0.005838,
		0.005752,
		0.005638,
		0.005496,
		0.005324,
		0.005121,
		0.00489,
		0.004631,
		0.004345,
		0.004037,
		0.003708,
		0.003363,
		0.003007,
		0.002644,
		0.002279,
		0.001918,
		0.001568,
		0.001235,
		0.000926,
		0.000649,
		0.00041,
		0.000218,
		8.2e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		9.7e-005,
		0.000371,
		0.000799,
		0.001359,
		0.00203,
		0.00279,
		0.00362,
		0.004501,
		0.005416,
		0.006347,
		0.007279,
		0.008198,
		0.009089,
		0.009942,
		0.010746,
		0.011491,
		0.01217,
		0.012777,
		0.013309,
		0.013763,
		0.014139,
		0.014439,
		0.014669,
		0.014835,
		0.014945,
		0.015011,
		0.015044,
		0.015057,
		0.015059,
		0.015059,
		0.015059,
		0.015052,
		0.015031,
		0.014983,
		0.014896,
		0.014759,
		0.014563,
		0.014298,
		0.01396,
		0.013546,
		0.013053,
		0.012483,
		0.011839,
		0.011126,
		0.010351,
		0.009521,
		0.008648,
		0.007741,
		0.006814,
		0.00588,
		0.004955,
		0.004055,
		0.003197,
		0.0024,
		0.001682,
		0.001064,
		0.000567,
		0.000213,
		2.5e-005,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		0.000205,
		0.000785,
		0.001691,
		0.002874,
		0.004288,
		0.005889,
		0.007633,
		0.009482,
		0.011397,
		0.013341,
		0.015282,
		0.017189,
		0.019033,
		0.02079,
		0.022437,
		0.023956,
		0.025331,
		0.026551,
		0.027608,
		0.0285,
		0.029228,
		0.029799,
		0.030225,
		0.030522,
		0.03071,
		0.030815,
		0.030862,
		0.030876,
		0.030877,
		0.030877,
		0.030871,
		0.030844,
		0.030771,
		0.030628,
		0.030388,
		0.03003,
		0.029533,
		0.028884,
		0.028075,
		0.0271,
		0.025961,
		0.024662,
		0.023214,
		0.021629,
		0.019924,
		0.01812,
		0.016241,
		0.014314,
		0.012367,
		0.010433,
		0.008547,
		0.006746,
		0.005068,
		0.003555,
		0.002251,
		0.0012,
		0.000451,
		5.2e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		0.000379,
		0.001451,
		0.003122,
		0.005302,
		0.007904,
		0.010844,
		0.014044,
		0.017426,
		0.020921,
		0.024461,
		0.027985,
		0.031435,
		0.034759,
		0.037912,
		0.040853,
		0.043547,
		0.04597,
		0.0481,
		0.049927,
		0.051448,
		0.052669,
		0.053606,
		0.054285,
		0.054739,
		0.055011,
		0.055149,
		0.055202,
		0.055213,
		0.055213,
		0.05521,
		0.055183,
		0.055094,
		0.054895,
		0.054538,
		0.053976,
		0.053172,
		0.052095,
		0.050725,
		0.049052,
		0.047072,
		0.044794,
		0.042233,
		0.039411,
		0.03636,
		0.033116,
		0.029723,
		0.026229,
		0.022689,
		0.019164,
		0.015716,
		0.012416,
		0.009337,
		0.006556,
		0.004154,
		0.002217,
		0.000834,
		9.7e-005,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		0.000635,
		0.002431,
		0.005227,
		0.008869,
		0.013209,
		0.018105,
		0.023421,
		0.02903,
		0.034811,
		0.040651,
		0.046446,
		0.052099,
		0.057525,
		0.062646,
		0.067396,
		0.07172,
		0.075576,
		0.078936,
		0.081783,
		0.084118,
		0.085958,
		0.087336,
		0.0883,
		0.088916,
		0.089259,
		0.089413,
		0.089459,
		0.089465,
		0.089464,
		0.089445,
		0.089354,
		0.089116,
		0.088647,
		0.087866,
		0.086702,
		0.085099,
		0.083014,
		0.080424,
		0.077319,
		0.073709,
		0.069614,
		0.065071,
		0.060128,
		0.054846,
		0.049296,
		0.04356,
		0.03773,
		0.031906,
		0.026196,
		0.020718,
		0.015596,
		0.010961,
		0.006952,
		0.003714,
		0.001398,
		0.000162,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000989,
		0.003782,
		0.008124,
		0.013772,
		0.02049,
		0.028054,
		0.036252,
		0.044882,
		0.053754,
		0.062691,
		0.07153,
		0.08012,
		0.088328,
		0.096035,
		0.103142,
		0.109565,
		0.115245,
		0.12014,
		0.124235,
		0.127539,
		0.130086,
		0.13194,
		0.133186,
		0.133936,
		0.134318,
		0.134464,
		0.134494,
		0.134495,
		0.134487,
		0.134412,
		0.134165,
		0.133616,
		0.132632,
		0.131095,
		0.128904,
		0.125984,
		0.122288,
		0.117792,
		0.112501,
		0.106443,
		0.099669,
		0.092251,
		0.084279,
		0.075864,
		0.067132,
		0.058225,
		0.049299,
		0.040526,
		0.032087,
		0.02418,
		0.017011,
		0.0108,
		0.005775,
		0.002176,
		0.000253,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.001451,
		0.005545,
		0.011902,
		0.020157,
		0.029959,
		0.040976,
		0.052889,
		0.0654,
		0.078227,
		0.091109,
		0.103804,
		0.116093,
		0.12778,
		0.138695,
		0.148694,
		0.157661,
		0.165515,
		0.172206,
		0.177722,
		0.18209,
		0.185375,
		0.187686,
		0.189167,
		0.189997,
		0.190372,
		0.190485,
		0.190498,
		0.190496,
		0.19045,
		0.190229,
		0.189652,
		0.18852,
		0.186644,
		0.183862,
		0.180046,
		0.17511,
		0.169008,
		0.161731,
		0.153312,
		0.143817,
		0.133344,
		0.122023,
		0.110012,
		0.097494,
		0.084677,
		0.071791,
		0.059088,
		0.046839,
		0.035336,
		0.024886,
		0.015814,
		0.008464,
		0.003192,
		0.000371,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.002029,
		0.007747,
		0.016611,
		0.028102,
		0.041725,
		0.057003,
		0.073489,
		0.090758,
		0.108413,
		0.126086,
		0.143438,
		0.160162,
		0.175988,
		0.19068,
		0.204045,
		0.21593,
		0.22623,
		0.234893,
		0.241918,
		0.247362,
		0.251342,
		0.254032,
		0.25566,
		0.256491,
		0.25681,
		0.256878,
		0.25688,
		0.256861,
		0.256697,
		0.256157,
		0.254963,
		0.252835,
		0.249525,
		0.244832,
		0.238608,
		0.230768,
		0.221283,
		0.21018,
		0.197539,
		0.183489,
		0.168203,
		0.151896,
		0.134822,
		0.117269,
		0.099561,
		0.082051,
		0.065122,
		0.049186,
		0.034677,
		0.02206,
		0.011818,
		0.004461,
		0.000519,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.002723,
		0.010384,
		0.022245,
		0.037596,
		0.055758,
		0.076086,
		0.097968,
		0.120829,
		0.144131,
		0.167375,
		0.190107,
		0.211916,
		0.232442,
		0.251376,
		0.268467,
		0.283525,
		0.296425,
		0.307118,
		0.31563,
		0.322066,
		0.326616,
		0.329549,
		0.331199,
		0.331948,
		0.332175,
		0.332201,
		0.332198,
		0.332104,
		0.331662,
		0.330511,
		0.328265,
		0.324562,
		0.319097,
		0.311642,
		0.302048,
		0.29025,
		0.276259,
		0.260166,
		0.242125,
		0.22236,
		0.201151,
		0.178832,
		0.155789,
		0.132457,
		0.109311,
		0.086869,
		0.06569,
		0.046366,
		0.029527,
		0.015835,
		0.005983,
		0.000697,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.003525,
		0.01343,
		0.028739,
		0.048517,
		0.071873,
		0.097956,
		0.125963,
		0.155141,
		0.184787,
		0.214251,
		0.242943,
		0.270336,
		0.295966,
		0.319444,
		0.340459,
		0.358785,
		0.374285,
		0.386925,
		0.396774,
		0.404013,
		0.408933,
		0.411925,
		0.413461,
		0.414053,
		0.414179,
		0.414184,
		0.414148,
		0.413844,
		0.412843,
		0.410642,
		0.406741,
		0.400704,
		0.392189,
		0.38096,
		0.366892,
		0.349969,
		0.330276,
		0.307995,
		0.283397,
		0.256833,
		0.228728,
		0.19958,
		0.169947,
		0.140451,
		0.111767,
		0.084624,
		0.059802,
		0.038126,
		0.020468,
		0.007742,
		0.000902,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.00442,
		0.016822,
		0.03596,
		0.06064,
		0.089722,
		0.122125,
		0.156828,
		0.192875,
		0.229374,
		0.265509,
		0.300538,
		0.333802,
		0.36473,
		0.392847,
		0.417783,
		0.43928,
		0.457205,
		0.471554,
		0.482466,
		0.490224,
		0.495252,
		0.498098,
		0.499394,
		0.49979,
		0.499836,
		0.49983,
		0.499666,
		0.498898,
		0.496909,
		0.493048,
		0.486716,
		0.477426,
		0.464821,
		0.448693,
		0.428973,
		0.405731,
		0.379165,
		0.34959,
		0.317429,
		0.283205,
		0.247536,
		0.211121,
		0.174741,
		0.139251,
		0.105574,
		0.074699,
		0.04768,
		0.025625,
		0.009702,
		0.001132,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.005385,
		0.020474,
		0.043719,
		0.073636,
		0.108813,
		0.147912,
		0.18967,
		0.232908,
		0.276532,
		0.31954,
		0.361029,
		0.400201,
		0.436373,
		0.468986,
		0.497617,
		0.521988,
		0.541984,
		0.557657,
		0.569245,
		0.577165,
		0.58201,
		0.584512,
		0.585481,
		0.585688,
		0.585696,
		0.585637,
		0.585137,
		0.583504,
		0.579929,
		0.57363,
		0.56394,
		0.55035,
		0.532534,
		0.510346,
		0.48382,
		0.453154,
		0.4187,
		0.38095,
		0.340527,
		0.298172,
		0.254737,
		0.211175,
		0.168534,
		0.127953,
		0.090652,
		0.057933,
		0.031171,
		0.011815,
		0.00138,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.006393,
		0.024279,
		0.051783,
		0.08711,
		0.128554,
		0.1745,
		0.223428,
		0.27392,
		0.324667,
		0.374475,
		0.422271,
		0.467118,
		0.508222,
		0.544946,
		0.576825,
		0.603579,
		0.625132,
		0.641623,
		0.653418,
		0.661108,
		0.665486,
		0.667493,
		0.668108,
		0.668181,
		0.668171,
		0.667916,
		0.666724,
		0.663655,
		0.657731,
		0.648076,
		0.633993,
		0.615002,
		0.590853,
		0.561514,
		0.527165,
		0.488182,
		0.445116,
		0.398686,
		0.349758,
		0.299339,
		0.248563,
		0.198682,
		0.151061,
		0.107171,
		0.068577,
		0.036943,
		0.014018,
		0.001639,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.007413,
		0.028122,
		0.059906,
		0.100644,
		0.14832,
		0.201032,
		0.256992,
		0.314536,
		0.372134,
		0.428393,
		0.482077,
		0.532108,
		0.577591,
		0.617822,
		0.652309,
		0.680795,
		0.703269,
		0.719989,
		0.73149,
		0.738568,
		0.742246,
		0.743679,
		0.743987,
		0.743999,
		0.743911,
		0.74317,
		0.740761,
		0.735522,
		0.726348,
		0.712321,
		0.692773,
		0.667309,
		0.635805,
		0.598399,
		0.555467,
		0.507609,
		0.455628,
		0.400512,
		0.343419,
		0.285665,
		0.228713,
		0.174161,
		0.123736,
		0.079284,
		0.042764,
		0.016247,
		0.001902,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.008417,
		0.031895,
		0.067856,
		0.113845,
		0.167531,
		0.226717,
		0.289343,
		0.353499,
		0.417432,
		0.47956,
		0.538479,
		0.592988,
		0.642099,
		0.685059,
		0.721373,
		0.75083,
		0.773521,
		0.789857,
		0.800578,
		0.80672,
		0.809553,
		0.810427,
		0.810531,
		0.810517,
		0.810153,
		0.808465,
		0.804146,
		0.795861,
		0.782445,
		0.763003,
		0.736961,
		0.704067,
		0.664385,
		0.618271,
		0.566352,
		0.5095,
		0.448814,
		0.385597,
		0.321343,
		0.257723,
		0.196569,
		0.139866,
		0.089745,
		0.048471,
		0.018437,
		0.002161,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.009384,
		0.035511,
		0.075448,
		0.126401,
		0.185725,
		0.250926,
		0.319677,
		0.389823,
		0.459397,
		0.526629,
		0.589968,
		0.648096,
		0.699951,
		0.744755,
		0.782037,
		0.81166,
		0.833854,
		0.849223,
		0.858746,
		0.863728,
		0.865681,
		0.866103,
		0.86612,
		0.865999,
		0.864986,
		0.861713,
		0.85464,
		0.842338,
		0.823657,
		0.7978,
		0.764353,
		0.723272,
		0.674863,
		0.619756,
		0.558873,
		0.493408,
		0.424797,
		0.354703,
		0.284996,
		0.21774,
		0.155176,
		0.099716,
		0.053931,
		0.020541,
		0.00241,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.010297,
		0.038915,
		0.082565,
		0.138117,
		0.202613,
		0.27327,
		0.347498,
		0.422905,
		0.497317,
		0.568794,
		0.635646,
		0.696458,
		0.750119,
		0.795845,
		0.83322,
		0.862222,
		0.883254,
		0.897154,
		0.905174,
		0.908898,
		0.910055,
		0.910193,
		0.910174,
		0.909691,
		0.907465,
		0.901806,
		0.891026,
		0.873692,
		0.848753,
		0.815588,
		0.774009,
		0.724239,
		0.666884,
		0.602893,
		0.533535,
		0.460362,
		0.385195,
		0.310092,
		0.237338,
		0.169425,
		0.109042,
		0.05906,
		0.022525,
		0.002646,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.011152,
		0.04209,
		0.08917,
		0.148931,
		0.218106,
		0.29363,
		0.372656,
		0.452566,
		0.530993,
		0.605834,
		0.675281,
		0.737842,
		0.792377,
		0.838132,
		0.874775,
		0.902438,
		0.921738,
		0.933787,
		0.940135,
		0.942641,
		0.943187,
		0.943208,
		0.943052,
		0.941747,
		0.937562,
		0.928581,
		0.913075,
		0.889702,
		0.857596,
		0.816383,
		0.766169,
		0.707502,
		0.641333,
		0.568982,
		0.492103,
		0.412655,
		0.332875,
		0.255258,
		0.182539,
		0.117674,
		0.063833,
		0.02438,
		0.002868,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.011955,
		0.045055,
		0.095303,
		0.15891,
		0.232304,
		0.312142,
		0.395328,
		0.479026,
		0.560686,
		0.63806,
		0.709237,
		0.77267,
		0.827216,
		0.872181,
		0.907357,
		0.933069,
		0.950194,
		0.960152,
		0.96481,
		0.966268,
		0.966443,
		0.966419,
		0.965809,
		0.963014,
		0.955961,
		0.942628,
		0.921354,
		0.890985,
		0.850921,
		0.801108,
		0.742003,
		0.674529,
		0.600037,
		0.520259,
		0.437279,
		0.3535,
		0.271617,
		0.194598,
		0.125665,
		0.068277,
		0.026116,
		0.003077,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.012718,
		0.047857,
		0.101067,
		0.168226,
		0.245457,
		0.329143,
		0.415942,
		0.502809,
		0.587017,
		0.666188,
		0.738324,
		0.801847,
		0.855639,
		0.899099,
		0.932184,
		0.955459,
		0.970107,
		0.977887,
		0.980982,
		0.981661,
		0.981687,
		0.981492,
		0.979875,
		0.974726,
		0.963764,
		0.944988,
		0.916918,
		0.878677,
		0.830009,
		0.771243,
		0.703245,
		0.627366,
		0.545402,
		0.459545,
		0.372351,
		0.286708,
		0.205814,
		0.133151,
		0.072466,
		0.027762,
		0.003275,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.013459,
		0.050566,
		0.106605,
		0.177115,
		0.257911,
		0.345096,
		0.435083,
		0.524618,
		0.61081,
		0.691157,
		0.763595,
		0.826533,
		0.878913,
		0.920262,
		0.950751,
		0.971234,
		0.983245,
		0.988911,
		0.990698,
		0.990914,
		0.990884,
		0.990133,
		0.986721,
		0.978178,
		0.962163,
		0.936827,
		0.900971,
		0.854087,
		0.796331,
		0.728476,
		0.651854,
		0.568298,
		0.480096,
		0.38995,
		0.300936,
		0.216478,
		0.140319,
		0.076504,
		0.029357,
		0.003469,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.014199,
		0.053259,
		0.112078,
		0.185844,
		0.270048,
		0.360506,
		0.453381,
		0.54521,
		0.632938,
		0.713954,
		0.786138,
		0.847917,
		0.898322,
		0.937057,
		0.964555,
		0.982016,
		0.991373,
		0.995127,
		0.995957,
		0.99599,
		0.995751,
		0.993781,
		0.987562,
		0.974436,
		0.952154,
		0.919139,
		0.874577,
		0.818404,
		0.751262,
		0.674429,
		0.589761,
		0.499627,
		0.406867,
		0.314743,
		0.22691,
		0.147381,
		0.080506,
		0.030947,
		0.003662,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.014959,
		0.05601,
		0.117642,
		0.194665,
		0.282229,
		0.375848,
		0.471423,
		0.565277,
		0.654192,
		0.735456,
		0.806913,
		0.867033,
		0.914977,
		0.950676,
		0.974887,
		0.989217,
		0.996039,
		0.99821,
		0.998474,
		0.998438,
		0.997522,
		0.993394,
		0.983155,
		0.964136,
		0.93433,
		0.892551,
		0.83846,
		0.772521,
		0.695929,
		0.610534,
		0.518777,
		0.423631,
		0.328548,
		0.237419,
		0.154541,
		0.084585,
		0.032575,
		0.003862,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.015757,
		0.058887,
		0.123433,
		0.203799,
		0.294765,
		0.39152,
		0.48969,
		0.585374,
		0.675192,
		0.756338,
		0.826642,
		0.884648,
		0.929694,
		0.961999,
		0.982718,
		0.993929,
		0.99847,
		0.999483,
		0.999523,
		0.99923,
		0.996836,
		0.989349,
		0.973698,
		0.94739,
		0.908799,
		0.857243,
		0.792954,
		0.717001,
		0.631206,
		0.538067,
		0.440684,
		0.342703,
		0.248267,
		0.161974,
		0.088841,
		0.034281,
		0.004071,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.016608,
		0.061944,
		0.129562,
		0.213418,
		0.307892,
		0.407819,
		0.508529,
		0.605889,
		0.696355,
		0.777035,
		0.845772,
		0.901223,
		0.942964,
		0.971575,
		0.988688,
		0.996917,
		0.999561,
		0.999886,
		0.999841,
		0.998721,
		0.993717,
		0.981426,
		0.958834,
		0.923796,
		0.875207,
		0.813001,
		0.738071,
		0.652181,
		0.557864,
		0.458348,
		0.357477,
		0.259662,
		0.169822,
		0.093355,
		0.036097,
		0.004295,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.017525,
		0.065226,
		0.136115,
		0.223654,
		0.321779,
		0.424946,
		0.528163,
		0.627051,
		0.717902,
		0.797756,
		0.864496,
		0.916947,
		0.954993,
		0.979666,
		0.993163,
		0.998688,
		0.999932,
		0.999981,
		0.99962,
		0.996693,
		0.987631,
		0.96889,
		0.937734,
		0.892541,
		0.832862,
		0.759357,
		0.673685,
		0.578397,
		0.476837,
		0.373058,
		0.271752,
		0.178194,
		0.098192,
		0.03805,
		0.004537,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.018518,
		0.068768,
		0.143159,
		0.234602,
		0.336549,
		0.443032,
		0.548717,
		0.648966,
		0.739907,
		0.818535,
		0.882813,
		0.931802,
		0.965783,
		0.986344,
		0.996346,
		0.999595,
		0.999998,
		0.999942,
		0.998558,
		0.992443,
		0.97759,
		0.9506,
		0.909231,
		0.852554,
		0.780912,
		0.695809,
		0.599782,
		0.496281,
		0.389572,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.023468,
		0.086243,
		0.177478,
		0.287165,
		0.406208,
		0.526505,
		0.641052,
		0.744084,
		0.83123,
		0.899725,
		0.948631,
		0.979066,
		0.994282,
		0.999355,
		0.999999,
		0.999909,
		0.997721,
		0.988294,
		0.966004,
		0.926607,
		0.867913,
		0.789856,
		0.694316,
		0.584888,
		0.466665,
		0.346048,
		0.230606,
		0.128949,
		0.050639,
		0.006111,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.025016,
		0.091646,
		0.18795,
		0.302949,
		0.426716,
		0.550474,
		0.66673,
		0.769433,
		0.854185,
		0.918486,
		0.962012,
		0.986838,
		0.997422,
		0.999897,
		0.999999,
		0.999268,
		0.993551,
		0.976536,
		0.942774,
		0.888967,
		0.814257,
		0.720092,
		0.609954,
		0.489096,
		0.364313,
		0.243759,
		0.136799,
		0.053897,
		0.006524,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.026722,
		0.09757,
		0.199356,
		0.320005,
		0.448658,
		0.575799,
		0.69341,
		0.795181,
		0.876762,
		0.93607,
		0.973615,
		0.9927,
		0.999166,
		0.999998,
		0.999882,
		0.997073,
		0.985151,
		0.957423,
		0.909232,
		0.838682,
		0.746597,
		0.636233,
		0.512956,
		0.383959,
		0.258031,
		0.145377,
		0.057479,
		0.006979,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.028608,
		0.104082,
		0.211807,
		0.338462,
		0.472143,
		0.602518,
		0.721027,
		0.821134,
		0.898654,
		0.952128,
		0.983187,
		0.996662,
		0.999864,
		0.999998,
		0.999046,
		0.991705,
		0.97023,
		0.928371,
		0.862877,
		0.773702,
		0.663715,
		0.538323,
		0.405109,
		0.273547,
		0.154777,
		0.061429,
		0.007484,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.0307,
		0.111261,
		0.225429,
		0.358463,
		0.497281,
		0.630658,
		0.749475,
		0.84704,
		0.919497,
		0.966289,
		0.990536,
		0.998903,
		0.999998,
		0.999844,
		0.996177,
		0.980889,
		0.945995,
		0.886525,
		0.801227,
		0.692366,
		0.565272,
		0.427896,
		0.290446,
		0.165103,
		0.065798,
		0.008046,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.033031,
		0.119202,
		0.24037,
		0.380165,
		0.52418,
		0.660215,
		0.778592,
		0.872571,
		0.938862,
		0.978187,
		0.995601,
		0.999819,
		0.999998,
		0.998733,
		0.989154,
		0.961681,
		0.909232,
		0.828924,
		0.722107,
		0.593863,
		0.452462,
		0.308888,
		0.176478,
		0.070648,
		0.008674,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.035637,
		0.128015,
		0.256796,
		0.403741,
		0.552945,
		0.69115,
		0.808146,
		0.897313,
		0.956268,
		0.987511,
		0.998528,
		0.999997,
		0.999788,
		0.994912,
		0.974996,
		0.930531,
		0.856465,
		0.752804,
		0.624138,
		0.478955,
		0.32905,
		0.189047,
		0.076051,
		0.009378,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.038563,
		0.137829,
		0.2749,
		0.429378,
		0.583667,
		0.723368,
		0.837812,
		0.920755,
		0.971204,
		0.994084,
		0.999752,
		0.999997,
		0.99828,
		0.985548,
		0.949876,
		0.883417,
		0.78425,
		0.656102,
		0.507526,
		0.351133,
		0.202976,
		0.082095,
		0.010171,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.041863,
		0.148801,
		0.294906,
		0.457278,
		0.61641,
		0.756699,
		0.867151,
		0.94229,
		0.983187,
		0.997978,
		0.999996,
		0.999707,
		0.993081,
		0.966673,
		0.909232,
		0.816138,
		0.689715,
		0.538323,
		0.375363,
		0.218463,
		0.088884,
		0.011069,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.045605,
		0.161115,
		0.317068,
		0.487653,
		0.651201,
		0.790869,
		0.895582,
		0.961225,
		0.991857,
		0.999651,
		0.999995,
		0.997607,
		0.980326,
		0.933236,
		0.848031,
		0.72486,
		0.571482,
		0.40199,
		0.23574,
		0.096544,
		0.012092,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.049871,
		0.174994,
		0.341684,
		0.520718,
		0.687996,
		0.825463,
		0.922365,
		0.976835,
		0.997147,
		0.999994,
		0.999581,
		0.990352,
		0.95463,
		0.879329,
		0.761317,
		0.607109,
		0.431293,
		0.255081,
		0.105229,
		0.013263,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.054761,
		0.19071,
		0.369091,
		0.556677,
		0.726654,
		0.859876,
		0.946591,
		0.988485,
		0.999493,
		0.999993,
		0.996571,
		0.972543,
		0.909232,
		0.798713,
		0.645261,
		0.463572,
		0.276811,
		0.11513,
		0.014613,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.060404,
		0.208589,
		0.399673,
		0.595706,
		0.766878,
		0.893264,
		0.967218,
		0.995843,
		0.999991,
		0.999381,
		0.986146,
		0.936718,
		0.836467,
		0.685901,
		0.499142,
		0.301315,
		0.12648,
		0.016179,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.066963,
		0.229035,
		0.433865,
		0.637915,
		0.808146,
		0.924495,
		0.983187,
		0.999236,
		0.999989,
		0.994912,
		0.960553,
		0.873716,
		0.728851,
		0.538323,
		0.32905,
		0.139573,
		0.018011,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.074644,
		0.25254,
		0.472143,
		0.683291,
		0.849616,
		0.952128,
		0.993706,
		0.999986,
		0.999046,
		0.9794,
		0.909232,
		0.773702,
		0.581406,
		0.360554,
		0.154777,
		0.020173,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.083718,
		0.279714,
		0.515011,
		0.731605,
		0.890001,
		0.974494,
		0.998793,
		0.999982,
		0.992122,
		0.941362,
		0.819691,
		0.628601,
		0.39646,
		0.172561,
		0.022747,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.094456,
		0.311037,
		0.56247,
		0.781582,
		0.926633,
		0.989142,
		0.9991,
		0.997575,
		0.967203,
		0.864779,
		0.679351,
		0.437117,
		0.19336,
		0.025825,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.104887,
		0.339533,
		0.600964,
		0.81321,
		0.935426,
		0.97297,
		0.974912,
		0.96241,
		0.886448,
		0.716734,
		0.472358,
		0.212988,
		0.028884,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.103703,
		0.328938,
		0.567267,
		0.743241,
		0.825636,
		0.839721,
		0.837497,
		0.795954,
		0.666205,
		0.45206,
		0.208583,
		0.028801,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.077731,
		0.240551,
		0.401745,
		0.505937,
		0.540983,
		0.542935,
		0.530799,
		0.462908,
		0.325662,
		0.154555,
		0.021807,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.037339,
		0.112084,
		0.179781,
		0.215674,
		0.222444,
		0.22133,
		0.202269,
		0.148905,
		0.073201,
		0.010604,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.008349,
		0.024116,
		0.036721,
		0.041577,
		0.041877,
		0.040099,
		0.031269,
		0.016081,
		0.002407,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		0.000173,
		0.000475,
		0.000676,
		0.000719,
		0.000712,
		0.000596,
		0.000326,
		5.1e-005,
		5.1e-005,
		0.000326,
		0.000596,
		0.000712,
		0.000719,
		0.000676,
		0.000475,
		0.000173,
		0.002141,
		0.006041,
		0.00891,
		0.00978,
		0.00978,
		0.00891,
		0.006041,
		0.002141,
		0.002407,
		0.016081,
		0.031269,
		0.040099,
		0.041877,
		0.041577,
		0.036721,
		0.024116,
		0.008349,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.010604,
		0.073201,
		0.148905,
		0.202269,
		0.22133,
		0.222444,
		0.215674,
		0.179781,
		0.112084,
		0.037339,
		0.057659,
		0.175889,
		0.288229,
		0.354699,
		0.372177,
		0.372177,
		0.354699,
		0.288229,
		0.175889,
		0.057659,
		0.021807,
		0.154555,
		0.325662,
		0.462908,
		0.530799,
		0.542935,
		0.540983,
		0.505937,
		0.401745,
		0.240551,
		0.077731,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.028801,
		0.208583,
		0.45206,
		0.666205,
		0.795954,
		0.837497,
		0.839721,
		0.825636,
		0.743241,
		0.567267,
		0.328938,
		0.103703,
		0.106843,
		0.342523,
		0.598814,
		0.797981,
		0.902538,
		0.927549,
		0.927549,
		0.902538,
		0.797981,
		0.598814,
		0.342523,
		0.106843,
		0.028884,
		0.212988,
		0.472358,
		0.716734,
		0.886448,
		0.96241,
		0.974912,
		0.97297,
		0.935426,
		0.81321,
		0.600964,
		0.339533,
		0.104887,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.025825,
		0.19336,
		0.437117,
		0.679351,
		0.864779,
		0.967203,
		0.997575,
		0.9991,
		0.989142,
		0.926633,
		0.781582,
		0.56247,
		0.311037,
		0.094456,
		0.088879,
		0.294891,
		0.538297,
		0.756663,
		0.909189,
		0.983139,
		0.999659,
		0.999659,
		0.983139,
		0.909189,
		0.756663,
		0.538297,
		0.294891,
		0.088879,
		0.022747,
		0.172561,
		0.39646,
		0.628601,
		0.819691,
		0.941362,
		0.992122,
		0.999982,
		0.998793,
		0.974494,
		0.890001,
		0.731605,
		0.515011,
		0.279714,
		0.083718,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.020173,
		0.154777,
		0.360554,
		0.581406,
		0.773702,
		0.909232,
		0.9794,
		0.999046,
		0.999986,
		0.993706,
		0.952128,
		0.849616,
		0.683291,
		0.472143,
		0.25254,
		0.074644,
		0.070648,
		0.24037,
		0.452462,
		0.660215,
		0.828924,
		0.938862,
		0.989154,
		0.999819,
		0.999819,
		0.989154,
		0.938862,
		0.828924,
		0.660215,
		0.452462,
		0.24037,
		0.070648,
		0.018011,
		0.139573,
		0.32905,
		0.538323,
		0.728851,
		0.873716,
		0.960553,
		0.994912,
		0.999989,
		0.999236,
		0.983187,
		0.924495,
		0.808146,
		0.637915,
		0.433865,
		0.229035,
		0.066963,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.016179,
		0.12648,
		0.301315,
		0.499142,
		0.685901,
		0.836467,
		0.936718,
		0.986146,
		0.999381,
		0.999991,
		0.995843,
		0.967218,
		0.893264,
		0.766878,
		0.595706,
		0.399673,
		0.208589,
		0.060404,
		0.057479,
		0.199356,
		0.383959,
		0.575799,
		0.746597,
		0.876762,
		0.957423,
		0.9927,
		0.999882,
		0.999882,
		0.9927,
		0.957423,
		0.876762,
		0.746597,
		0.575799,
		0.383959,
		0.199356,
		0.057479,
		0.014613,
		0.11513,
		0.276811,
		0.463572,
		0.645261,
		0.798713,
		0.909232,
		0.972543,
		0.996571,
		0.999993,
		0.999493,
		0.988485,
		0.946591,
		0.859876,
		0.726654,
		0.556677,
		0.369091,
		0.19071,
		0.054761,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.013263,
		0.105229,
		0.255081,
		0.431293,
		0.607109,
		0.761317,
		0.879329,
		0.95463,
		0.990352,
		0.999581,
		0.999994,
		0.997147,
		0.976835,
		0.922365,
		0.825463,
		0.687996,
		0.520718,
		0.341684,
		0.174994,
		0.049871,
		0.047667,
		0.167843,
		0.32905,
		0.503836,
		0.669354,
		0.808146,
		0.909232,
		0.969489,
		0.994912,
		0.99992,
		0.99992,
		0.994912,
		0.969489,
		0.909232,
		0.808146,
		0.669354,
		0.503836,
		0.32905,
		0.167843,
		0.047667,
		0.012092,
		0.096544,
		0.23574,
		0.40199,
		0.571482,
		0.72486,
		0.848031,
		0.933236,
		0.980326,
		0.997607,
		0.999995,
		0.999651,
		0.991857,
		0.961225,
		0.895582,
		0.790869,
		0.651201,
		0.487653,
		0.317068,
		0.161115,
		0.045605,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.011069,
		0.088884,
		0.218463,
		0.375363,
		0.538323,
		0.689715,
		0.816138,
		0.909232,
		0.966673,
		0.993081,
		0.999707,
		0.999996,
		0.997978,
		0.983187,
		0.94229,
		0.867151,
		0.756699,
		0.61641,
		0.457278,
		0.294906,
		0.148801,
		0.041863,
		0.040162,
		0.14316,
		0.28465,
		0.443032,
		0.599783,
		0.739908,
		0.852555,
		0.931803,
		0.977591,
		0.996348,
		0.999944,
		0.999944,
		0.996348,
		0.977591,
		0.931803,
		0.852555,
		0.739908,
		0.599783,
		0.443032,
		0.28465,
		0.14316,
		0.040162,
		0.010171,
		0.082095,
		0.202976,
		0.351133,
		0.507526,
		0.656102,
		0.78425,
		0.883417,
		0.949876,
		0.985548,
		0.99828,
		0.999997,
		0.999752,
		0.994084,
		0.971204,
		0.920755,
		0.837812,
		0.723368,
		0.583667,
		0.429378,
		0.2749,
		0.137829,
		0.038563,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.009378,
		0.076051,
		0.189047,
		0.32905,
		0.478955,
		0.624138,
		0.752804,
		0.856465,
		0.930531,
		0.974996,
		0.994912,
		0.999788,
		0.999997,
		0.998528,
		0.987511,
		0.956268,
		0.897313,
		0.808146,
		0.69115,
		0.552945,
		0.403741,
		0.256796,
		0.128015,
		0.035637,
		0.034297,
		0.123492,
		0.248386,
		0.391707,
		0.538323,
		0.675515,
		0.793332,
		0.88507,
		0.947842,
		0.983187,
		0.997312,
		0.999959,
		0.999959,
		0.997312,
		0.983187,
		0.947842,
		0.88507,
		0.793332,
		0.675515,
		0.538323,
		0.391707,
		0.248386,
		0.123492,
		0.034297,
		0.008674,
		0.070648,
		0.176478,
		0.308888,
		0.452462,
		0.593863,
		0.722107,
		0.828924,
		0.909232,
		0.961681,
		0.989154,
		0.998733,
		0.999998,
		0.999819,
		0.995601,
		0.978187,
		0.938862,
		0.872571,
		0.778592,
		0.660215,
		0.52418,
		0.380165,
		0.24037,
		0.119202,
		0.033031,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.008046,
		0.065798,
		0.165103,
		0.290446,
		0.427896,
		0.565272,
		0.692366,
		0.801227,
		0.886525,
		0.945995,
		0.980889,
		0.996177,
		0.999844,
		0.999998,
		0.998903,
		0.990536,
		0.966289,
		0.919497,
		0.84704,
		0.749475,
		0.630658,
		0.497281,
		0.358463,
		0.225429,
		0.111261,
		0.0307,
		0.029626,
		0.107582,
		0.218463,
		0.34826,
		0.484498,
		0.61641,
		0.735156,
		0.834112,
		0.909232,
		0.959469,
		0.987147,
		0.997978,
		0.99997,
		0.99997,
		0.997978,
		0.987147,
		0.959469,
		0.909232,
		0.834112,
		0.735156,
		0.61641,
		0.484498,
		0.34826,
		0.218463,
		0.107582,
		0.029626,
		0.007484,
		0.061429,
		0.154777,
		0.273547,
		0.405109,
		0.538323,
		0.663715,
		0.773702,
		0.862877,
		0.928371,
		0.97023,
		0.991705,
		0.999046,
		0.999998,
		0.999864,
		0.996662,
		0.983187,
		0.952128,
		0.898654,
		0.821134,
		0.721027,
		0.602518,
		0.472143,
		0.338462,
		0.211807,
		0.104082,
		0.028608,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.006979,
		0.057479,
		0.145377,
		0.258031,
		0.383959,
		0.512956,
		0.636233,
		0.746597,
		0.838682,
		0.909232,
		0.957423,
		0.985151,
		0.997073,
		0.999882,
		0.999998,
		0.999166,
		0.9927,
		0.973615,
		0.93607,
		0.876762,
		0.795181,
		0.69341,
		0.575799,
		0.448658,
		0.320005,
		0.199356,
		0.09757,
		0.026722,
		0.025848,
		0.094539,
		0.193529,
		0.31131,
		0.437501,
		0.562964,
		0.679948,
		0.782268,
		0.865538,
		0.927446,
		0.968053,
		0.990011,
		0.998451,
		0.999977,
		0.999977,
		0.998451,
		0.990011,
		0.968053,
		0.927446,
		0.865538,
		0.782268,
		0.679948,
		0.562964,
		0.437501,
		0.31131,
		0.193529,
		0.094539,
		0.025848,
		0.006524,
		0.053897,
		0.136799,
		0.243759,
		0.364313,
		0.489096,
		0.609954,
		0.720092,
		0.814257,
		0.888967,
		0.942774,
		0.976536,
		0.993551,
		0.999268,
		0.999999,
		0.999897,
		0.997422,
		0.986838,
		0.962012,
		0.918486,
		0.854185,
		0.769433,
		0.66673,
		0.550474,
		0.426716,
		0.302949,
		0.18795,
		0.091646,
		0.025016,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.006111,
		0.050639,
		0.128949,
		0.230606,
		0.346048,
		0.466665,
		0.584888,
		0.694316,
		0.789856,
		0.867913,
		0.926607,
		0.966004,
		0.988294,
		0.997721,
		0.999909,
		0.999999,
		0.999355,
		0.994282,
		0.979066,
		0.948631,
		0.899725,
		0.83123,
		0.744084,
		0.641052,
		0.526505,
		0.406208,
		0.287165,
		0.177478,
		0.086243,
		0.023468,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389572,
		0.496281,
		0.599782,
		0.695809,
		0.780912,
		0.852554,
		0.909231,
		0.9506,
		0.97759,
		0.992443,
		0.998558,
		0.999942,
		0.999998,
		0.999595,
		0.996346,
		0.986344,
		0.965783,
		0.931802,
		0.882813,
		0.818535,
		0.739907,
		0.648966,
		0.548717,
		0.443032,
		0.336549,
		0.234602,
		0.143159,
		0.068768,
		0.018518,
		0.018011,
		0.066962,
		0.139572,
		0.229034,
		0.329048,
		0.433863,
		0.53832,
		0.637912,
		0.728847,
		0.808142,
		0.873711,
		0.92449,
		0.960548,
		0.983181,
		0.994906,
		0.99923,
		0.999983,
		0.999983,
		0.99923,
		0.994906,
		0.983181,
		0.960548,
		0.92449,
		0.873711,
		0.808142,
		0.728847,
		0.637912,
		0.53832,
		0.433863,
		0.329048,
		0.229034,
		0.139572,
		0.066962,
		0.018011,
		0.004537,
		0.03805,
		0.098192,
		0.178194,
		0.271752,
		0.373058,
		0.476837,
		0.578397,
		0.673685,
		0.759357,
		0.832862,
		0.892541,
		0.937734,
		0.96889,
		0.987631,
		0.996693,
		0.99962,
		0.999981,
		0.999932,
		0.998688,
		0.993163,
		0.979666,
		0.954993,
		0.916947,
		0.864496,
		0.797756,
		0.717902,
		0.627051,
		0.528163,
		0.424946,
		0.321779,
		0.223654,
		0.136115,
		0.065226,
		0.017525,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.004295,
		0.036097,
		0.093355,
		0.169822,
		0.259662,
		0.357477,
		0.458348,
		0.557864,
		0.652181,
		0.738071,
		0.813001,
		0.875207,
		0.923796,
		0.958834,
		0.981426,
		0.993717,
		0.998721,
		0.999841,
		0.999886,
		0.999561,
		0.996917,
		0.988688,
		0.971575,
		0.942964,
		0.901223,
		0.845772,
		0.777035,
		0.696355,
		0.605889,
		0.508529,
		0.407819,
		0.307892,
		0.213418,
		0.129562,
		0.061944,
		0.016608,
		0.016175,
		0.06039,
		0.12645,
		0.208539,
		0.301242,
		0.399576,
		0.499021,
		0.595562,
		0.685735,
		0.766692,
		0.836264,
		0.893048,
		0.936491,
		0.966984,
		0.985907,
		0.995602,
		0.999139,
		0.999749,
		0.999749,
		0.999139,
		0.995602,
		0.985907,
		0.966984,
		0.936491,
		0.893048,
		0.836264,
		0.766692,
		0.685735,
		0.595562,
		0.499021,
		0.399576,
		0.301242,
		0.208539,
		0.12645,
		0.06039,
		0.016175,
		0.004071,
		0.034281,
		0.088841,
		0.161974,
		0.248267,
		0.342703,
		0.440684,
		0.538067,
		0.631206,
		0.717001,
		0.792954,
		0.857243,
		0.908799,
		0.94739,
		0.973698,
		0.989349,
		0.996836,
		0.99923,
		0.999523,
		0.999483,
		0.99847,
		0.993929,
		0.982718,
		0.961999,
		0.929694,
		0.884648,
		0.826642,
		0.756338,
		0.675192,
		0.585374,
		0.48969,
		0.39152,
		0.294765,
		0.203799,
		0.123433,
		0.058887,
		0.015757,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.003862,
		0.032575,
		0.084585,
		0.154541,
		0.237419,
		0.328548,
		0.423631,
		0.518777,
		0.610534,
		0.695929,
		0.772521,
		0.83846,
		0.892551,
		0.93433,
		0.964136,
		0.983155,
		0.993394,
		0.997522,
		0.998438,
		0.998474,
		0.99821,
		0.996039,
		0.989217,
		0.974887,
		0.950676,
		0.914977,
		0.867033,
		0.806913,
		0.735456,
		0.654192,
		0.565277,
		0.471423,
		0.375848,
		0.282229,
		0.194665,
		0.117642,
		0.05601,
		0.014959,
		0.014576,
		0.054623,
		0.114839,
		0.190228,
		0.276112,
		0.368158,
		0.462401,
		0.555271,
		0.643631,
		0.724819,
		0.796695,
		0.857704,
		0.906936,
		0.9442,
		0.970086,
		0.985988,
		0.994054,
		0.996969,
		0.997467,
		0.997467,
		0.996969,
		0.994054,
		0.985988,
		0.970086,
		0.9442,
		0.906936,
		0.857704,
		0.796695,
		0.724819,
		0.643631,
		0.555271,
		0.462401,
		0.368158,
		0.276112,
		0.190228,
		0.114839,
		0.054623,
		0.014576,
		0.003662,
		0.030947,
		0.080506,
		0.147381,
		0.22691,
		0.314743,
		0.406867,
		0.499627,
		0.589761,
		0.674429,
		0.751262,
		0.818404,
		0.874577,
		0.919139,
		0.952154,
		0.974436,
		0.987562,
		0.993781,
		0.995751,
		0.99599,
		0.995957,
		0.995127,
		0.991373,
		0.982016,
		0.964555,
		0.937057,
		0.898322,
		0.847917,
		0.786138,
		0.713954,
		0.632938,
		0.54521,
		0.453381,
		0.360506,
		0.270048,
		0.185844,
		0.112078,
		0.053259,
		0.014199,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.003469,
		0.029357,
		0.076504,
		0.140319,
		0.216478,
		0.300936,
		0.38995,
		0.480096,
		0.568298,
		0.651854,
		0.728476,
		0.796331,
		0.854087,
		0.900971,
		0.936827,
		0.962163,
		0.978178,
		0.986721,
		0.990133,
		0.990884,
		0.990914,
		0.990698,
		0.988911,
		0.983245,
		0.971234,
		0.950751,
		0.920262,
		0.878913,
		0.826533,
		0.763595,
		0.691157,
		0.61081,
		0.524618,
		0.435083,
		0.345096,
		0.257911,
		0.177115,
		0.106605,
		0.050566,
		0.013459,
		0.01309,
		0.049219,
		0.103854,
		0.172707,
		0.251747,
		0.337219,
		0.425657,
		0.513912,
		0.599175,
		0.679005,
		0.751367,
		0.814675,
		0.867837,
		0.91031,
		0.942154,
		0.964069,
		0.977409,
		0.984115,
		0.986518,
		0.986925,
		0.986925,
		0.986518,
		0.984115,
		0.977409,
		0.964069,
		0.942154,
		0.91031,
		0.867837,
		0.814675,
		0.751367,
		0.679005,
		0.599175,
		0.513912,
		0.425657,
		0.337219,
		0.251747,
		0.172707,
		0.103854,
		0.049219,
		0.01309,
		0.003275,
		0.027762,
		0.072466,
		0.133151,
		0.205814,
		0.286708,
		0.372351,
		0.459545,
		0.545402,
		0.627366,
		0.703245,
		0.771243,
		0.830009,
		0.878677,
		0.916918,
		0.944988,
		0.963764,
		0.974726,
		0.979875,
		0.981492,
		0.981687,
		0.981661,
		0.980982,
		0.977887,
		0.970107,
		0.955459,
		0.932184,
		0.899099,
		0.855639,
		0.801847,
		0.738324,
		0.666188,
		0.587017,
		0.502809,
		0.415942,
		0.329143,
		0.245457,
		0.168226,
		0.101067,
		0.047857,
		0.012718,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.003077,
		0.026116,
		0.068277,
		0.125665,
		0.194598,
		0.271617,
		0.3535,
		0.437279,
		0.520259,
		0.600037,
		0.674529,
		0.742003,
		0.801108,
		0.850921,
		0.890985,
		0.921354,
		0.942628,
		0.955961,
		0.963014,
		0.965809,
		0.966419,
		0.966443,
		0.966268,
		0.96481,
		0.960152,
		0.950194,
		0.933069,
		0.907357,
		0.872181,
		0.827216,
		0.77267,
		0.709237,
		0.63806,
		0.560686,
		0.479026,
		0.395328,
		0.312142,
		0.232304,
		0.15891,
		0.095303,
		0.045055,
		0.011955,
		0.011559,
		0.043596,
		0.09229,
		0.154017,
		0.225354,
		0.3031,
		0.384281,
		0.46617,
		0.546305,
		0.622512,
		0.692926,
		0.756027,
		0.810671,
		0.856127,
		0.892122,
		0.918878,
		0.937137,
		0.948161,
		0.953657,
		0.955611,
		0.95594,
		0.95594,
		0.955611,
		0.953657,
		0.948161,
		0.937137,
		0.918878,
		0.892122,
		0.856127,
		0.810671,
		0.756027,
		0.692926,
		0.622512,
		0.546305,
		0.46617,
		0.384281,
		0.3031,
		0.225354,
		0.154017,
		0.09229,
		0.043596,
		0.011559,
		0.002868,
		0.02438,
		0.063833,
		0.117674,
		0.182539,
		0.255258,
		0.332875,
		0.412655,
		0.492103,
		0.568982,
		0.641333,
		0.707502,
		0.766169,
		0.816383,
		0.857596,
		0.889702,
		0.913075,
		0.928581,
		0.937562,
		0.941747,
		0.943052,
		0.943208,
		0.943187,
		0.942641,
		0.940135,
		0.933787,
		0.921738,
		0.902438,
		0.874775,
		0.838132,
		0.792377,
		0.737842,
		0.675281,
		0.605834,
		0.530993,
		0.452566,
		0.372656,
		0.29363,
		0.218106,
		0.148931,
		0.08917,
		0.04209,
		0.011152,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.002646,
		0.022525,
		0.05906,
		0.109042,
		0.169425,
		0.237338,
		0.310092,
		0.385195,
		0.460362,
		0.533535,
		0.602893,
		0.666884,
		0.724239,
		0.774009,
		0.815588,
		0.848753,
		0.873692,
		0.891026,
		0.901806,
		0.907465,
		0.909691,
		0.910174,
		0.910193,
		0.910055,
		0.908898,
		0.905174,
		0.897154,
		0.883254,
		0.862222,
		0.83322,
		0.795845,
		0.750119,
		0.696458,
		0.635646,
		0.568794,
		0.497317,
		0.422905,
		0.347498,
		0.27327,
		0.202613,
		0.138117,
		0.082565,
		0.038915,
		0.010297,
		0.009847,
		0.037241,
		0.07907,
		0.132372,
		0.194342,
		0.262345,
		0.333919,
		0.40679,
		0.478887,
		0.548352,
		0.613563,
		0.673152,
		0.726028,
		0.771408,
		0.808844,
		0.838252,
		0.859942,
		0.874633,
		0.883435,
		0.887791,
		0.889329,
		0.889586,
		0.889586,
		0.889329,
		0.887791,
		0.883435,
		0.874633,
		0.859942,
		0.838252,
		0.808844,
		0.771408,
		0.726028,
		0.673152,
		0.613563,
		0.548352,
		0.478887,
		0.40679,
		0.333919,
		0.262345,
		0.194342,
		0.132372,
		0.07907,
		0.037241,
		0.009847,
		0.00241,
		0.020541,
		0.053931,
		0.099716,
		0.155176,
		0.21774,
		0.284996,
		0.354703,
		0.424797,
		0.493408,
		0.558873,
		0.619756,
		0.674863,
		0.723272,
		0.764353,
		0.7978,
		0.823657,
		0.842338,
		0.85464,
		0.861713,
		0.864986,
		0.865999,
		0.86612,
		0.866103,
		0.865681,
		0.863728,
		0.858746,
		0.849223,
		0.833854,
		0.81166,
		0.782037,
		0.744755,
		0.699951,
		0.648096,
		0.589968,
		0.526629,
		0.459397,
		0.389823,
		0.319677,
		0.250926,
		0.185725,
		0.126401,
		0.075448,
		0.035511,
		0.009384,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		0.002161,
		0.018437,
		0.048471,
		0.089745,
		0.139866,
		0.196569,
		0.257723,
		0.321343,
		0.385597,
		0.448814,
		0.5095,
		0.566352,
		0.618271,
		0.664385,
		0.704067,
		0.736961,
		0.763003,
		0.782445,
		0.795861,
		0.804146,
		0.808465,
		0.810153,
		0.810517,
		0.810531,
		0.810427,
		0.809553,
		0.80672,
		0.800578,
		0.789857,
		0.773521,
		0.75083,
		0.721373,
		0.685059,
		0.642099,
		0.592988,
		0.538479,
		0.47956,
		0.417432,
		0.353499,
		0.289343,
		0.226717,
		0.167531,
		0.113845,
		0.067856,
		0.031895,
		0.008417,
		0.007919,
		0.030023,
		0.063915,
		0.107308,
		0.158028,
		0.214025,
		0.273376,
		0.334294,
		0.395137,
		0.454416,
		0.510811,
		0.563181,
		0.610581,
		0.652282,
		0.687787,
		0.716858,
		0.73953,
		0.756135,
		0.767303,
		0.773948,
		0.777215,
		0.778361,
		0.778551,
		0.778551,
		0.778361,
		0.777215,
		0.773948,
		0.767303,
		0.756135,
		0.73953,
		0.716858,
		0.687787,
		0.652282,
		0.610581,
		0.563181,
		0.510811,
		0.454416,
		0.395137,
		0.334294,
		0.273376,
		0.214025,
		0.158028,
		0.107308,
		0.063915,
		0.030023,
		0.007919,
		0.001902,
		0.016247,
		0.042764,
		0.079284,
		0.123736,
		0.174161,
		0.228713,
		0.285665,
		0.343419,
		0.400512,
		0.455628,
		0.507609,
		0.555467,
		0.598399,
		0.635805,
		0.667309,
		0.692773,
		0.712321,
		0.726348,
		0.735522,
		0.740761,
		0.74317,
		0.743911,
		0.743999,
		0.743987,
		0.743679,
		0.742246,
		0.738568,
		0.73149,
		0.719989,
		0.703269,
		0.680795,
		0.652309,
		0.617822,
		0.577591,
		0.532108,
		0.482077,
		0.428393,
		0.372134,
		0.314536,
		0.256992,
		0.201032,
		0.14832,
		0.100644,
		0.059906,
		0.028122,
		0.007413,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		0.001639,
		0.014018,
		0.036943,
		0.068577,
		0.107171,
		0.151061,
		0.198682,
		0.248563,
		0.299339,
		0.349758,
		0.398686,
		0.445116,
		0.488182,
		0.527165,
		0.561514,
		0.590853,
		0.615002,
		0.633993,
		0.648076,
		0.657731,
		0.663655,
		0.666724,
		0.667916,
		0.668171,
		0.668181,
		0.668108,
		0.667493,
		0.665486,
		0.661108,
		0.653418,
		0.641623,
		0.625132,
		0.603579,
		0.576825,
		0.544946,
		0.508222,
		0.467118,
		0.422271,
		0.374475,
		0.324667,
		0.27392,
		0.223428,
		0.1745,
		0.128554,
		0.08711,
		0.051783,
		0.024279,
		0.006393,
		0.005885,
		0.022365,
		0.047728,
		0.080339,
		0.118641,
		0.161159,
		0.206504,
		0.253379,
		0.300582,
		0.347016,
		0.391695,
		0.43375,
		0.472443,
		0.507174,
		0.537498,
		0.563133,
		0.58398,
		0.600132,
		0.611886,
		0.61974,
		0.624385,
		0.626654,
		0.627445,
		0.627576,
		0.627576,
		0.627445,
		0.626654,
		0.624385,
		0.61974,
		0.611886,
		0.600132,
		0.58398,
		0.563133,
		0.537498,
		0.507174,
		0.472443,
		0.43375,
		0.391695,
		0.347016,
		0.300582,
		0.253379,
		0.206504,
		0.161159,
		0.118641,
		0.080339,
		0.047728,
		0.022365,
		0.005885,
		0.00138,
		0.011815,
		0.031171,
		0.057933,
		0.090652,
		0.127953,
		0.168534,
		0.211175,
		0.254737,
		0.298172,
		0.340527,
		0.38095,
		0.4187,
		0.453154,
		0.48382,
		0.510346,
		0.532534,
		0.55035,
		0.56394,
		0.57363,
		0.579929,
		0.583504,
		0.585137,
		0.585637,
		0.585696,
		0.585688,
		0.585481,
		0.584512,
		0.58201,
		0.577165,
		0.569245,
		0.557657,
		0.541984,
		0.521988,
		0.497617,
		0.468986,
		0.436373,
		0.400201,
		0.361029,
		0.31954,
		0.276532,
		0.232908,
		0.18967,
		0.147912,
		0.108813,
		0.073636,
		0.043719,
		0.020474,
		0.005385,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		0.001132,
		0.009702,
		0.025625,
		0.04768,
		0.074699,
		0.105574,
		0.139251,
		0.174741,
		0.211121,
		0.247536,
		0.283205,
		0.317429,
		0.34959,
		0.379165,
		0.405731,
		0.428973,
		0.448693,
		0.464821,
		0.477426,
		0.486716,
		0.493048,
		0.496909,
		0.498898,
		0.499666,
		0.49983,
		0.499836,
		0.49979,
		0.499394,
		0.498098,
		0.495252,
		0.490224,
		0.482466,
		0.471554,
		0.457205,
		0.43928,
		0.417783,
		0.392847,
		0.36473,
		0.333802,
		0.300538,
		0.265509,
		0.229374,
		0.192875,
		0.156828,
		0.122125,
		0.089722,
		0.06064,
		0.03596,
		0.016822,
		0.00442,
		0.003962,
		0.015087,
		0.032269,
		0.054447,
		0.080608,
		0.109792,
		0.141088,
		0.173644,
		0.206667,
		0.239425,
		0.271253,
		0.30156,
		0.32983,
		0.35563,
		0.37862,
		0.398556,
		0.415302,
		0.428835,
		0.439258,
		0.446798,
		0.451806,
		0.454751,
		0.456182,
		0.456678,
		0.45676,
		0.45676,
		0.456678,
		0.456182,
		0.454751,
		0.451806,
		0.446798,
		0.439258,
		0.428835,
		0.415302,
		0.398556,
		0.37862,
		0.35563,
		0.32983,
		0.30156,
		0.271253,
		0.239425,
		0.206667,
		0.173644,
		0.141088,
		0.109792,
		0.080608,
		0.054447,
		0.032269,
		0.015087,
		0.003962,
		0.000902,
		0.007742,
		0.020468,
		0.038126,
		0.059802,
		0.084624,
		0.111767,
		0.140451,
		0.169947,
		0.19958,
		0.228728,
		0.256833,
		0.283397,
		0.307995,
		0.330276,
		0.349969,
		0.366892,
		0.38096,
		0.392189,
		0.400704,
		0.406741,
		0.410642,
		0.412843,
		0.413844,
		0.414148,
		0.414184,
		0.414179,
		0.414053,
		0.413461,
		0.411925,
		0.408933,
		0.404013,
		0.396774,
		0.386925,
		0.374285,
		0.358785,
		0.340459,
		0.319444,
		0.295966,
		0.270336,
		0.242943,
		0.214251,
		0.184787,
		0.155141,
		0.125963,
		0.097956,
		0.071873,
		0.048517,
		0.028739,
		0.01343,
		0.003525,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		0.000697,
		0.005983,
		0.015835,
		0.029527,
		0.046366,
		0.06569,
		0.086869,
		0.109311,
		0.132457,
		0.155789,
		0.178832,
		0.201151,
		0.22236,
		0.242125,
		0.260166,
		0.276259,
		0.29025,
		0.302048,
		0.311642,
		0.319097,
		0.324562,
		0.328265,
		0.330511,
		0.331662,
		0.332104,
		0.332198,
		0.332201,
		0.332175,
		0.331948,
		0.331199,
		0.329549,
		0.326616,
		0.322066,
		0.31563,
		0.307118,
		0.296425,
		0.283525,
		0.268467,
		0.251376,
		0.232442,
		0.211916,
		0.190107,
		0.167375,
		0.144131,
		0.120829,
		0.097968,
		0.076086,
		0.055758,
		0.037596,
		0.022245,
		0.010384,
		0.002723,
		0.002362,
		0.009012,
		0.019315,
		0.03266,
		0.048465,
		0.066174,
		0.08526,
		0.105226,
		0.125608,
		0.145975,
		0.165934,
		0.185128,
		0.203242,
		0.220006,
		0.235198,
		0.248646,
		0.260237,
		0.269916,
		0.277694,
		0.283651,
		0.287937,
		0.290769,
		0.292425,
		0.293225,
		0.293501,
		0.293546,
		0.293546,
		0.293501,
		0.293225,
		0.292425,
		0.290769,
		0.287937,
		0.283651,
		0.277694,
		0.269916,
		0.260237,
		0.248646,
		0.235198,
		0.220006,
		0.203242,
		0.185128,
		0.165934,
		0.145975,
		0.125608,
		0.105226,
		0.08526,
		0.066174,
		0.048465,
		0.03266,
		0.019315,
		0.009012,
		0.002362,
		0.000519,
		0.004461,
		0.011818,
		0.02206,
		0.034677,
		0.049186,
		0.065122,
		0.082051,
		0.099561,
		0.117269,
		0.134822,
		0.151896,
		0.168203,
		0.183489,
		0.197539,
		0.21018,
		0.221283,
		0.230768,
		0.238608,
		0.244832,
		0.249525,
		0.252835,
		0.254963,
		0.256157,
		0.256697,
		0.256861,
		0.25688,
		0.256878,
		0.25681,
		0.256491,
		0.25566,
		0.254032,
		0.251342,
		0.247362,
		0.241918,
		0.234893,
		0.22623,
		0.21593,
		0.204045,
		0.19068,
		0.175988,
		0.160162,
		0.143438,
		0.126086,
		0.108413,
		0.090758,
		0.073489,
		0.057003,
		0.041725,
		0.028102,
		0.016611,
		0.007747,
		0.002029,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		0.000371,
		0.003192,
		0.008464,
		0.015814,
		0.024886,
		0.035336,
		0.046839,
		0.059088,
		0.071791,
		0.084677,
		0.097494,
		0.110012,
		0.122023,
		0.133344,
		0.143817,
		0.153312,
		0.161731,
		0.169008,
		0.17511,
		0.180046,
		0.183862,
		0.186644,
		0.18852,
		0.189652,
		0.190229,
		0.19045,
		0.190496,
		0.190498,
		0.190485,
		0.190372,
		0.189997,
		0.189167,
		0.187686,
		0.185375,
		0.18209,
		0.177722,
		0.172206,
		0.165515,
		0.157661,
		0.148694,
		0.138695,
		0.12778,
		0.116093,
		0.103804,
		0.091109,
		0.078227,
		0.0654,
		0.052889,
		0.040976,
		0.029959,
		0.020157,
		0.011902,
		0.005545,
		0.001451,
		0.001206,
		0.00461,
		0.009899,
		0.016772,
		0.024941,
		0.034131,
		0.04408,
		0.054541,
		0.06528,
		0.076083,
		0.086747,
		0.097092,
		0.106953,
		0.116188,
		0.124676,
		0.13232,
		0.139047,
		0.144812,
		0.1496,
		0.153429,
		0.156346,
		0.158434,
		0.159806,
		0.160605,
		0.160989,
		0.161121,
		0.161143,
		0.161143,
		0.161121,
		0.160989,
		0.160605,
		0.159806,
		0.158434,
		0.156346,
		0.153429,
		0.1496,
		0.144812,
		0.139047,
		0.13232,
		0.124676,
		0.116188,
		0.106953,
		0.097092,
		0.086747,
		0.076083,
		0.06528,
		0.054541,
		0.04408,
		0.034131,
		0.024941,
		0.016772,
		0.009899,
		0.00461,
		0.001206,
		0.000253,
		0.002176,
		0.005775,
		0.0108,
		0.017011,
		0.02418,
		0.032087,
		0.040526,
		0.049299,
		0.058225,
		0.067132,
		0.075864,
		0.084279,
		0.092251,
		0.099669,
		0.106443,
		0.112501,
		0.117792,
		0.122288,
		0.125984,
		0.128904,
		0.131095,
		0.132632,
		0.133616,
		0.134165,
		0.134412,
		0.134487,
		0.134495,
		0.134494,
		0.134464,
		0.134318,
		0.133936,
		0.133186,
		0.13194,
		0.130086,
		0.127539,
		0.124235,
		0.12014,
		0.115245,
		0.109565,
		0.103142,
		0.096035,
		0.088328,
		0.08012,
		0.07153,
		0.062691,
		0.053754,
		0.044882,
		0.036252,
		0.028054,
		0.02049,
		0.013772,
		0.008124,
		0.003782,
		0.000989,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		0.000162,
		0.001398,
		0.003714,
		0.006952,
		0.010961,
		0.015596,
		0.020718,
		0.026196,
		0.031906,
		0.03773,
		0.04356,
		0.049296,
		0.054846,
		0.060128,
		0.065071,
		0.069614,
		0.073709,
		0.077319,
		0.080424,
		0.083014,
		0.085099,
		0.086702,
		0.087866,
		0.088647,
		0.089116,
		0.089354,
		0.089445,
		0.089464,
		0.089465,
		0.089459,
		0.089413,
		0.089259,
		0.088916,
		0.0883,
		0.087336,
		0.085958,
		0.084118,
		0.081783,
		0.078936,
		0.075576,
		0.07172,
		0.067396,
		0.062646,
		0.057525,
		0.052099,
		0.046446,
		0.040651,
		0.034811,
		0.02903,
		0.023421,
		0.018105,
		0.013209,
		0.008869,
		0.005227,
		0.002431,
		0.000635,
		0.000496,
		0.001898,
		0.004083,
		0.00693,
		0.010326,
		0.01416,
		0.018328,
		0.02273,
		0.027273,
		0.031869,
		0.036436,
		0.0409,
		0.045192,
		0.049254,
		0.053032,
		0.056483,
		0.059573,
		0.062277,
		0.064584,
		0.06649,
		0.068007,
		0.069157,
		0.069976,
		0.070513,
		0.070823,
		0.070972,
		0.071023,
		0.071031,
		0.071031,
		0.071023,
		0.070972,
		0.070823,
		0.070513,
		0.069976,
		0.069157,
		0.068007,
		0.06649,
		0.064584,
		0.062277,
		0.059573,
		0.056483,
		0.053032,
		0.049254,
		0.045192,
		0.0409,
		0.036436,
		0.031869,
		0.027273,
		0.02273,
		0.018328,
		0.01416,
		0.010326,
		0.00693,
		0.004083,
		0.001898,
		0.000496,
		9.7e-005,
		0.000834,
		0.002217,
		0.004154,
		0.006556,
		0.009337,
		0.012416,
		0.015716,
		0.019164,
		0.022689,
		0.026229,
		0.029723,
		0.033116,
		0.03636,
		0.039411,
		0.042233,
		0.044794,
		0.047072,
		0.049052,
		0.050725,
		0.052095,
		0.053172,
		0.053976,
		0.054538,
		0.054895,
		0.055094,
		0.055183,
		0.05521,
		0.055213,
		0.055213,
		0.055202,
		0.055149,
		0.055011,
		0.054739,
		0.054285,
		0.053606,
		0.052669,
		0.051448,
		0.049927,
		0.0481,
		0.04597,
		0.043547,
		0.040853,
		0.037912,
		0.034759,
		0.031435,
		0.027985,
		0.024461,
		0.020921,
		0.017426,
		0.014044,
		0.010844,
		0.007904,
		0.005302,
		0.003122,
		0.001451,
		0.000379,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		5.2e-005,
		0.000451,
		0.0012,
		0.002251,
		0.003555,
		0.005068,
		0.006746,
		0.008547,
		0.010433,
		0.012367,
		0.014314,
		0.016241,
		0.01812,
		0.019924,
		0.021629,
		0.023214,
		0.024662,
		0.025961,
		0.0271,
		0.028075,
		0.028884,
		0.029533,
		0.03003,
		0.030388,
		0.030628,
		0.030771,
		0.030844,
		0.030871,
		0.030877,
		0.030877,
		0.030876,
		0.030862,
		0.030815,
		0.03071,
		0.030522,
		0.030225,
		0.029799,
		0.029228,
		0.0285,
		0.027608,
		0.026551,
		0.025331,
		0.023956,
		0.022437,
		0.02079,
		0.019033,
		0.017189,
		0.015282,
		0.013341,
		0.011397,
		0.009482,
		0.007633,
		0.005889,
		0.004288,
		0.002874,
		0.001691,
		0.000785,
		0.000205,
		0.000144,
		0.000551,
		0.001186,
		0.002017,
		0.003011,
		0.004136,
		0.005365,
		0.006667,
		0.008018,
		0.009391,
		0.010764,
		0.012115,
		0.013424,
		0.014673,
		0.015848,
		0.016933,
		0.01792,
		0.018799,
		0.019564,
		0.020214,
		0.020748,
		0.021172,
		0.021491,
		0.021718,
		0.021866,
		0.021951,
		0.021992,
		0.022005,
		0.022008,
		0.022008,
		0.022005,
		0.021992,
		0.021951,
		0.021866,
		0.021718,
		0.021491,
		0.021172,
		0.020748,
		0.020214,
		0.019564,
		0.018799,
		0.01792,
		0.016933,
		0.015848,
		0.014673,
		0.013424,
		0.012115,
		0.010764,
		0.009391,
		0.008018,
		0.006667,
		0.005365,
		0.004136,
		0.003011,
		0.002017,
		0.001186,
		0.000551,
		0.000144,
		2.5e-005,
		0.000213,
		0.000567,
		0.001064,
		0.001682,
		0.0024,
		0.003197,
		0.004055,
		0.004955,
		0.00588,
		0.006814,
		0.007741,
		0.008648,
		0.009521,
		0.010351,
		0.011126,
		0.011839,
		0.012483,
		0.013053,
		0.013546,
		0.01396,
		0.014298,
		0.014563,
		0.014759,
		0.014896,
		0.014983,
		0.015031,
		0.015052,
		0.015059,
		0.015059,
		0.015059,
		0.015057,
		0.015044,
		0.015011,
		0.014945,
		0.014835,
		0.014669,
		0.014439,
		0.014139,
		0.013763,
		0.013309,
		0.012777,
		0.01217,
		0.011491,
		0.010746,
		0.009942,
		0.009089,
		0.008198,
		0.007279,
		0.006347,
		0.005416,
		0.004501,
		0.00362,
		0.00279,
		0.00203,
		0.001359,
		0.000799,
		0.000371,
		9.7e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		8.2e-005,
		0.000218,
		0.00041,
		0.000649,
		0.000926,
		0.001235,
		0.001568,
		0.001918,
		0.002279,
		0.002644,
		0.003007,
		0.003363,
		0.003708,
		0.004037,
		0.004345,
		0.004631,
		0.00489,
		0.005121,
		0.005324,
		0.005496,
		0.005638,
		0.005752,
		0.005838,
		0.005901,
		0.005942,
		0.005967,
		0.005979,
		0.005984,
		0.005985,
		0.005985,
		0.005985,
		0.005982,
		0.005974,
		0.005956,
		0.005924,
		0.005872,
		0.005798,
		0.005698,
		0.005571,
		0.005413,
		0.005226,
		0.005009,
		0.004764,
		0.004491,
		0.004194,
		0.003874,
		0.003537,
		0.003186,
		0.002826,
		0.002461,
		0.002098,
		0.001742,
		0.001399,
		0.001077,
		0.000783,
		0.000524,
		0.000308,
		0.000143,
		3.7e-005,
		2.1e-005,
		7.9e-005,
		0.00017,
		0.00029,
		0.000434,
		0.000597,
		0.000776,
		0.000966,
		0.001164,
		0.001366,
		0.00157,
		0.001771,
		0.001968,
		0.002157,
		0.002336,
		0.002503,
		0.002657,
		0.002796,
		0.00292,
		0.003027,
		0.003117,
		0.003191,
		0.00325,
		0.003294,
		0.003325,
		0.003345,
		0.003356,
		0.003362,
		0.003364,
		0.003364,
		0.003364,
		0.003364,
		0.003362,
		0.003356,
		0.003345,
		0.003325,
		0.003294,
		0.00325,
		0.003191,
		0.003117,
		0.003027,
		0.00292,
		0.002796,
		0.002657,
		0.002503,
		0.002336,
		0.002157,
		0.001968,
		0.001771,
		0.00157,
		0.001366,
		0.001164,
		0.000966,
		0.000776,
		0.000597,
		0.000434,
		0.00029,
		0.00017,
		7.9e-005,
		2.1e-005,
		2.2e-005,
		6e-005,
		0.000112,
		0.000178,
		0.000254,
		0.000339,
		0.000431,
		0.000527,
		0.000627,
		0.000728,
		0.000829,
		0.000929,
		0.001025,
		0.001117,
		0.001205,
		0.001285,
		0.00136,
		0.001426,
		0.001485,
		0.001535,
		0.001578,
		0.001612,
		0.001639,
		0.001658,
		0.001672,
		0.001681,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.001689,
		0.001688,
		0.001687,
		0.001684,
		0.001677,
		0.001666,
		0.001649,
		0.001626,
		0.001596,
		0.001557,
		0.001511,
		0.001456,
		0.001394,
		0.001323,
		0.001246,
		0.001162,
		0.001072,
		0.000977,
		0.000879,
		0.000779,
		0.000678,
		0.000577,
		0.000479,
		0.000384,
		0.000295,
		0.000215,
		0.000143,
		8.4e-005,
		3.9e-005,
		1e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		1.5e-005,
		2.4e-005,
		3.5e-005,
		4.6e-005,
		5.9e-005,
		7.2e-005,
		8.6e-005,
		0.0001,
		0.000114,
		0.000127,
		0.000141,
		0.000154,
		0.000166,
		0.000177,
		0.000188,
		0.000197,
		0.000206,
		0.000213,
		0.000219,
		0.000224,
		0.000228,
		0.000232,
		0.000234,
		0.000235,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.00023,
		0.000227,
		0.000222,
		0.000216,
		0.00021,
		0.000202,
		0.000193,
		0.000183,
		0.000172,
		0.00016,
		0.000147,
		0.000134,
		0.000121,
		0.000107,
		9.3e-005,
		7.9e-005,
		6.5e-005,
		5.2e-005,
		4e-005,
		2.9e-005,
		2e-005,
		1.1e-005,
		1.1e-005,
		1.3e-005,
		1.6e-005,
		1.9e-005,
		2.2e-005,
		2.4e-005,
		2.7e-005,
		3e-005,
		3.3e-005,
		3.5e-005,
		3.7e-005,
		3.9e-005,
		4.1e-005,
		4.3e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.3e-005,
		4.1e-005,
		3.9e-005,
		3.7e-005,
		3.5e-005,
		3.3e-005,
		3e-005,
		2.7e-005,
		2.4e-005,
		2.2e-005,
		1.9e-005,
		1.6e-005,
		1.3e-005,
		1.1e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.7e-005,
		4.7e-005,
		4.6e-005,
		4.5e-005,
		4.4e-005,
		4.2e-005,
		4e-005,
		3.8e-005,
		3.6e-005,
		3.4e-005,
		3.1e-005,
		2.9e-005,
		2.6e-005,
		2.3e-005,
		2e-005,
		1.7e-005,
		1.5e-005,
		1.2e-005,
		4.9e-005,
		0.000237,
		0.000237,
		0.000237,
		0.000236,
		0.000236,
		0.000235,
		0.000233,
		0.000231,
		0.000228,
		0.000223,
		0.000218,
		0.000211,
		0.000204,
		0.000195,
		0.000185,
		0.000175,
		0.000163,
		0.000151,
		0.000138,
		0.000124,
		0.00011,
		9.6e-005,
		8.2e-005,
		6.9e-005,
		5.6e-005,
		4.3e-005,
		3.2e-005,
		2.2e-005,
		1.3e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000717,
		0.000714,
		0.000709,
		0.000701,
		0.000691,
		0.000677,
		0.000661,
		0.00064,
		0.000617,
		0.00059,
		0.00056,
		0.000526,
		0.000491,
		0.000452,
		0.000412,
		0.000371,
		0.000328,
		0.000285,
		0.000243,
		0.000201,
		0.000161,
		0.000124,
		9e-005,
		6e-005,
		3.5e-005,
		1.6e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.001688,
		0.001688,
		0.001686,
		0.001682,
		0.001675,
		0.001662,
		0.001644,
		0.001619,
		0.001587,
		0.001547,
		0.001498,
		0.001442,
		0.001377,
		0.001305,
		0.001225,
		0.00114,
		0.001049,
		0.000953,
		0.000854,
		0.000754,
		0.000652,
		0.000552,
		0.000454,
		0.000361,
		0.000274,
		0.000196,
		0.000127,
		7.2e-005,
		3e-005,
		4.8e-005,
		0.000236,
		0.000719,
		0.001688,
		0.003364,
		0.003364,
		0.003363,
		0.00336,
		0.003351,
		0.003336,
		0.003311,
		0.003273,
		0.003222,
		0.003156,
		0.003074,
		0.002975,
		0.00286,
		0.002729,
		0.002582,
		0.002421,
		0.002247,
		0.002063,
		0.00187,
		0.001671,
		0.001468,
		0.001265,
		0.001064,
		0.00087,
		0.000685,
		0.000513,
		0.000359,
		0.000227,
		0.000121,
		4.5e-005,
		4.8e-005,
		0.000236,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.005985,
		0.005983,
		0.005977,
		0.005962,
		0.005933,
		0.005887,
		0.005819,
		0.005726,
		0.005605,
		0.005456,
		0.005276,
		0.005066,
		0.004828,
		0.004562,
		0.00427,
		0.003956,
		0.003623,
		0.003275,
		0.002916,
		0.002552,
		0.002188,
		0.00183,
		0.001483,
		0.001156,
		0.000854,
		0.000585,
		0.000357,
		0.000179,
		5.7e-005,
		4.8e-005,
		0.000235,
		0.000717,
		0.001686,
		0.003363,
		0.005985,
		0.0098,
		0.0098,
		0.009797,
		0.009787,
		0.009761,
		0.009713,
		0.009635,
		0.009521,
		0.009364,
		0.009162,
		0.00891,
		0.008609,
		0.008259,
		0.00786,
		0.007416,
		0.00693,
		0.006407,
		0.005853,
		0.005276,
		0.004682,
		0.00408,
		0.003479,
		0.00289,
		0.002323,
		0.00179,
		0.001301,
		0.000871,
		0.000512,
		0.000237,
		6.2e-005,
		4.7e-005,
		0.000233,
		0.000714,
		0.001682,
		0.00336,
		0.005983,
		0.0098,
		0.015059,
		0.015059,
		0.015055,
		0.015038,
		0.014998,
		0.014922,
		0.014799,
		0.014618,
		0.014371,
		0.014052,
		0.013657,
		0.013183,
		0.012633,
		0.012007,
		0.011311,
		0.01055,
		0.009733,
		0.00887,
		0.00797,
		0.007047,
		0.006113,
		0.005185,
		0.004277,
		0.003407,
		0.002592,
		0.001853,
		0.001208,
		0.000679,
		0.000287,
		5.5e-005,
		4.7e-005,
		0.000231,
		0.000709,
		0.001675,
		0.003351,
		0.005977,
		0.009797,
		0.015059,
		0.022008,
		0.022007,
		0.022001,
		0.021976,
		0.021915,
		0.021801,
		0.021616,
		0.021344,
		0.020973,
		0.020495,
		0.019903,
		0.019196,
		0.018373,
		0.01744,
		0.016402,
		0.015271,
		0.014057,
		0.012776,
		0.011443,
		0.010079,
		0.008703,
		0.007338,
		0.006008,
		0.004739,
		0.003559,
		0.002495,
		0.001579,
		0.000842,
		0.000316,
		3.7e-005,
		4.6e-005,
		0.000228,
		0.000701,
		0.001662,
		0.003336,
		0.005962,
		0.009787,
		0.015055,
		0.022007,
		0.030877,
		0.030877,
		0.030867,
		0.030831,
		0.030743,
		0.030578,
		0.030311,
		0.029919,
		0.029385,
		0.028697,
		0.027847,
		0.02683,
		0.025651,
		0.024314,
		0.02283,
		0.021213,
		0.019481,
		0.017657,
		0.015763,
		0.013828,
		0.011881,
		0.009956,
		0.008087,
		0.006313,
		0.004672,
		0.003208,
		0.001962,
		0.000983,
		0.000317,
		1.3e-005,
		4.5e-005,
		0.000223,
		0.000691,
		0.001644,
		0.003311,
		0.005933,
		0.009761,
		0.015038,
		0.022001,
		0.030877,
		0.041882,
		0.041881,
		0.041868,
		0.041817,
		0.041695,
		0.041464,
		0.041091,
		0.040544,
		0.039801,
		0.038844,
		0.037661,
		0.036251,
		0.034615,
		0.032763,
		0.030711,
		0.028478,
		0.02609,
		0.023578,
		0.020976,
		0.018324,
		0.015662,
		0.013038,
		0.010502,
		0.008105,
		0.005905,
		0.00396,
		0.002331,
		0.001083,
		0.000283,
		4.4e-005,
		0.000218,
		0.000677,
		0.001619,
		0.003273,
		0.005887,
		0.009713,
		0.014998,
		0.021976,
		0.030867,
		0.041881,
		0.055213,
		0.055212,
		0.055194,
		0.055125,
		0.054958,
		0.054644,
		0.054138,
		0.053397,
		0.052391,
		0.051096,
		0.049499,
		0.047595,
		0.045391,
		0.042898,
		0.040139,
		0.037142,
		0.033943,
		0.030582,
		0.027109,
		0.023575,
		0.02004,
		0.016567,
		0.013224,
		0.010082,
		0.007219,
		0.004715,
		0.002653,
		0.001123,
		0.000215,
		4.2e-005,
		0.000211,
		0.000661,
		0.001587,
		0.003222,
		0.005819,
		0.009635,
		0.014922,
		0.021915,
		0.030831,
		0.041868,
		0.055212,
		0.071031,
		0.071029,
		0.071006,
		0.070914,
		0.070692,
		0.070277,
		0.069605,
		0.068626,
		0.067296,
		0.065587,
		0.063481,
		0.060974,
		0.058074,
		0.0548,
		0.051181,
		0.047255,
		0.043071,
		0.038685,
		0.034161,
		0.029569,
		0.02499,
		0.020506,
		0.016209,
		0.012195,
		0.008567,
		0.005431,
		0.0029,
		0.001091,
		0.000127,
		4e-005,
		0.000204,
		0.00064,
		0.001547,
		0.003156,
		0.005726,
		0.009521,
		0.014799,
		0.021801,
		0.030743,
		0.041817,
		0.055194,
		0.071029,
		0.089465,
		0.089462,
		0.089431,
		0.089312,
		0.089024,
		0.088484,
		0.087614,
		0.086344,
		0.084624,
		0.082414,
		0.079696,
		0.076464,
		0.072729,
		0.068519,
		0.06387,
		0.058836,
		0.05348,
		0.047876,
		0.042108,
		0.036269,
		0.030463,
		0.0248,
		0.019399,
		0.014386,
		0.009895,
		0.006065,
		0.003043,
		0.000982,
		4.1e-005,
		3.8e-005,
		0.000195,
		0.000617,
		0.001498,
		0.003074,
		0.005605,
		0.009364,
		0.014618,
		0.021616,
		0.030578,
		0.041695,
		0.055125,
		0.071006,
		0.089462,
		0.110603,
		0.1106,
		0.110561,
		0.110408,
		0.110041,
		0.109353,
		0.108244,
		0.10663,
		0.104444,
		0.101641,
		0.098195,
		0.094104,
		0.089384,
		0.084068,
		0.078208,
		0.071873,
		0.065143,
		0.058116,
		0.0509,
		0.043615,
		0.036394,
		0.029379,
		0.022723,
		0.016587,
		0.011143,
		0.00657,
		0.003057,
		0.000799,
		3.6e-005,
		0.000185,
		0.00059,
		0.001442,
		0.002975,
		0.005456,
		0.009162,
		0.014371,
		0.021344,
		0.030311,
		0.041464,
		0.054958,
		0.070914,
		0.089431,
		0.1106,
		0.134496,
		0.134492,
		0.134442,
		0.13425,
		0.133789,
		0.132926,
		0.131537,
		0.129518,
		0.126786,
		0.123286,
		0.118991,
		0.113897,
		0.108027,
		0.101426,
		0.094161,
		0.086318,
		0.078003,
		0.069338,
		0.06046,
		0.051523,
		0.042695,
		0.034155,
		0.026096,
		0.018722,
		0.01225,
		0.006907,
		0.002929,
		0.000562,
		3.4e-005,
		0.000175,
		0.00056,
		0.001377,
		0.00286,
		0.005276,
		0.00891,
		0.014052,
		0.020973,
		0.029919,
		0.041091,
		0.054644,
		0.070692,
		0.089312,
		0.110561,
		0.134492,
		0.161143,
		0.161138,
		0.161077,
		0.160839,
		0.160268,
		0.159201,
		0.157487,
		0.154997,
		0.151632,
		0.147328,
		0.142051,
		0.135801,
		0.128609,
		0.120532,
		0.111657,
		0.102092,
		0.09197,
		0.081444,
		0.070687,
		0.05989,
		0.049261,
		0.039027,
		0.029425,
		0.020712,
		0.013156,
		0.007038,
		0.002653,
		0.000308,
		3.1e-005,
		0.000163,
		0.000526,
		0.001305,
		0.002729,
		0.005066,
		0.008609,
		0.013657,
		0.020495,
		0.029385,
		0.040544,
		0.054138,
		0.070277,
		0.089024,
		0.110408,
		0.134442,
		0.161138,
		0.190498,
		0.190492,
		0.190417,
		0.190126,
		0.189429,
		0.188127,
		0.186039,
		0.183009,
		0.17892,
		0.173695,
		0.167298,
		0.159732,
		0.151036,
		0.141286,
		0.130588,
		0.119079,
		0.106923,
		0.094309,
		0.081453,
		0.068588,
		0.055972,
		0.043882,
		0.032612,
		0.022475,
		0.013802,
		0.006937,
		0.002242,
		9.4e-005,
		2.9e-005,
		0.000151,
		0.000491,
		0.001225,
		0.002582,
		0.004828,
		0.008259,
		0.013183,
		0.019903,
		0.028697,
		0.039801,
		0.053397,
		0.069605,
		0.088484,
		0.110041,
		0.13425,
		0.161077,
		0.190492,
		0.222461,
		0.222455,
		0.222364,
		0.222012,
		0.22117,
		0.219602,
		0.217088,
		0.213445,
		0.208535,
		0.202269,
		0.194608,
		0.185558,
		0.175172,
		0.163544,
		0.150806,
		0.137128,
		0.122709,
		0.107782,
		0.092608,
		0.077474,
		0.062693,
		0.0486,
		0.035553,
		0.023933,
		0.014139,
		0.006591,
		0.001726,
		2.6e-005,
		0.000138,
		0.000452,
		0.00114,
		0.002421,
		0.004562,
		0.00786,
		0.012633,
		0.019196,
		0.027847,
		0.038844,
		0.052391,
		0.068626,
		0.087614,
		0.109353,
		0.133789,
		0.160839,
		0.190417,
		0.222455,
		0.25688,
		0.256872,
		0.256763,
		0.256343,
		0.255338,
		0.253469,
		0.250476,
		0.246144,
		0.240313,
		0.232882,
		0.223808,
		0.213104,
		0.200837,
		0.187125,
		0.172129,
		0.156055,
		0.139147,
		0.121685,
		0.103984,
		0.08639,
		0.069278,
		0.053053,
		0.038146,
		0.025011,
		0.014128,
		0.006001,
		0.001155,
		2.3e-005,
		0.000124,
		0.000412,
		0.001049,
		0.002247,
		0.00427,
		0.007416,
		0.012007,
		0.018373,
		0.02683,
		0.037661,
		0.051096,
		0.067296,
		0.086344,
		0.108244,
		0.132926,
		0.160268,
		0.190126,
		0.222364,
		0.256872,
		0.293547,
		0.293537,
		0.293409,
		0.292911,
		0.291724,
		0.289518,
		0.285991,
		0.280893,
		0.274039,
		0.265316,
		0.254678,
		0.242148,
		0.22781,
		0.211808,
		0.194338,
		0.175647,
		0.156029,
		0.135819,
		0.115392,
		0.095162,
		0.075577,
		0.057115,
		0.040291,
		0.025644,
		0.013746,
		0.005191,
		0.000604,
		2e-005,
		0.00011,
		0.000371,
		0.000953,
		0.002063,
		0.003956,
		0.00693,
		0.011311,
		0.01744,
		0.025651,
		0.036251,
		0.049499,
		0.065587,
		0.084624,
		0.10663,
		0.131537,
		0.159201,
		0.189429,
		0.222012,
		0.256763,
		0.293537,
		0.332201,
		0.33219,
		0.332039,
		0.331456,
		0.330067,
		0.327489,
		0.323371,
		0.317427,
		0.309448,
		0.299306,
		0.286955,
		0.272428,
		0.25583,
		0.237336,
		0.217181,
		0.195661,
		0.173123,
		0.149966,
		0.126633,
		0.103613,
		0.081434,
		0.060661,
		0.041898,
		0.025783,
		0.012984,
		0.004205,
		0.000176,
		1.7e-005,
		9.6e-005,
		0.000328,
		0.000854,
		0.00187,
		0.003623,
		0.006407,
		0.01055,
		0.016402,
		0.024314,
		0.034615,
		0.047595,
		0.063481,
		0.082414,
		0.104444,
		0.129518,
		0.157487,
		0.188127,
		0.22117,
		0.256343,
		0.293409,
		0.33219,
		0.372533,
		0.37252,
		0.372344,
		0.371666,
		0.370055,
		0.367068,
		0.362304,
		0.355436,
		0.346229,
		0.334543,
		0.320332,
		0.303641,
		0.284601,
		0.26342,
		0.240381,
		0.21583,
		0.190179,
		0.163895,
		0.137498,
		0.11156,
		0.086698,
		0.063573,
		0.04289,
		0.025391,
		0.011859,
		0.003111,
		1.5e-005,
		8.2e-005,
		0.000285,
		0.000754,
		0.001671,
		0.003275,
		0.005853,
		0.009733,
		0.015271,
		0.02283,
		0.032763,
		0.045391,
		0.060974,
		0.079696,
		0.101641,
		0.126786,
		0.154997,
		0.186039,
		0.219602,
		0.255338,
		0.292911,
		0.332039,
		0.37252,
		0.414184,
		0.414169,
		0.413966,
		0.413186,
		0.411333,
		0.4079,
		0.402434,
		0.394565,
		0.384031,
		0.370678,
		0.354464,
		0.33545,
		0.313795,
		0.289746,
		0.263636,
		0.235873,
		0.206935,
		0.177368,
		0.147776,
		0.118823,
		0.091224,
		0.065746,
		0.043203,
		0.024457,
		0.010409,
		0.002007,
		1.2e-005,
		6.9e-005,
		0.000243,
		0.000652,
		0.001468,
		0.002916,
		0.005276,
		0.00887,
		0.014057,
		0.021213,
		0.030711,
		0.042898,
		0.058074,
		0.076464,
		0.098195,
		0.123286,
		0.151632,
		0.183009,
		0.217088,
		0.253469,
		0.291724,
		0.331456,
		0.372344,
		0.414169,
		0.456761,
		0.456743,
		0.456511,
		0.455619,
		0.453504,
		0.449592,
		0.443369,
		0.434424,
		0.422467,
		0.407332,
		0.388981,
		0.367494,
		0.343062,
		0.315979,
		0.286631,
		0.255494,
		0.223122,
		0.190145,
		0.15726,
		0.12523,
		0.09488,
		0.067091,
		0.042798,
		0.022989,
		0.008699,
		0.001015,
		5.6e-005,
		0.000201,
		0.000552,
		0.001265,
		0.002552,
		0.004682,
		0.00797,
		0.012776,
		0.019481,
		0.028478,
		0.040139,
		0.0548,
		0.072729,
		0.094104,
		0.118991,
		0.147328,
		0.17892,
		0.213445,
		0.250476,
		0.289518,
		0.330067,
		0.371666,
		0.413966,
		0.456743,
		0.499836,
		0.499817,
		0.499553,
		0.498541,
		0.496144,
		0.491718,
		0.484687,
		0.474596,
		0.461124,
		0.444099,
		0.423487,
		0.399391,
		0.372039,
		0.341773,
		0.309044,
		0.274398,
		0.238473,
		0.20199,
		0.16575,
		0.130624,
		0.097552,
		0.067541,
		0.041656,
		0.021023,
		0.006822,
		0.000286,
		4.3e-005,
		0.000161,
		0.000454,
		0.001064,
		0.002188,
		0.00408,
		0.007047,
		0.011443,
		0.017657,
		0.02609,
		0.037142,
		0.051181,
		0.068519,
		0.089384,
		0.113897,
		0.142051,
		0.173695,
		0.208535,
		0.246144,
		0.285991,
		0.327489,
		0.370055,
		0.413186,
		0.456511,
		0.499817,
		0.542966,
		0.542944,
		0.542647,
		0.541506,
		0.53881,
		0.533837,
		0.52595,
		0.514646,
		0.499578,
		0.480563,
		0.457578,
		0.430753,
		0.400354,
		0.366781,
		0.330552,
		0.292291,
		0.252728,
		0.212684,
		0.173066,
		0.134865,
		0.099147,
		0.067052,
		0.039786,
		0.018622,
		0.004895,
		3.2e-005,
		0.000124,
		0.000361,
		0.00087,
		0.00183,
		0.003479,
		0.006113,
		0.010079,
		0.015763,
		0.023578,
		0.033943,
		0.047255,
		0.06387,
		0.084068,
		0.108027,
		0.135801,
		0.167298,
		0.202269,
		0.240313,
		0.280893,
		0.323371,
		0.367068,
		0.411333,
		0.455619,
		0.499553,
		0.542944,
		0.585696,
		0.585671,
		0.585338,
		0.584062,
		0.581049,
		0.575499,
		0.566711,
		0.554134,
		0.537395,
		0.516304,
		0.49085,
		0.461192,
		0.427644,
		0.390664,
		0.350845,
		0.308897,
		0.265647,
		0.222021,
		0.179047,
		0.13784,
		0.0996,
		0.065609,
		0.037225,
		0.015878,
		0.003067,
		2.2e-005,
		9e-005,
		0.000274,
		0.000685,
		0.001483,
		0.00289,
		0.005185,
		0.008703,
		0.013828,
		0.020976,
		0.030582,
		0.043071,
		0.058836,
		0.078208,
		0.101426,
		0.128609,
		0.159732,
		0.194608,
		0.232882,
		0.274039,
		0.317427,
		0.362304,
		0.4079,
		0.453504,
		0.498541,
		0.542647,
		0.585671,
		0.627578,
		0.62755,
		0.627179,
		0.62576,
		0.622414,
		0.616261,
		0.606532,
		0.59263,
		0.574154,
		0.550912,
		0.522909,
		0.490336,
		0.453558,
		0.4131,
		0.369632,
		0.323959,
		0.27701,
		0.229828,
		0.183562,
		0.139462,
		0.098873,
		0.063227,
		0.03404,
		0.012909,
		0.001509,
		1.3e-005,
		6e-005,
		0.000196,
		0.000513,
		0.001156,
		0.002323,
		0.004277,
		0.007338,
		0.011881,
		0.018324,
		0.027109,
		0.038685,
		0.05348,
		0.071873,
		0.094161,
		0.120532,
		0.151036,
		0.185558,
		0.223808,
		0.265316,
		0.309448,
		0.355436,
		0.402434,
		0.449592,
		0.496144,
		0.541506,
		0.585338,
		0.62755,
		0.668181,
		0.66815,
		0.66774,
		0.666171,
		0.662478,
		0.655698,
		0.644993,
		0.629719,
		0.609454,
		0.584001,
		0.553385,
		0.517835,
		0.477773,
		0.433793,
		0.386651,
		0.337251,
		0.286632,
		0.235957,
		0.186509,
		0.13968,
		0.096962,
		0.059949,
		0.030325,
		0.009861,
		0.000415,
		3.5e-005,
		0.000127,
		0.000359,
		0.000854,
		0.00179,
		0.003407,
		0.006008,
		0.009956,
		0.015662,
		0.023575,
		0.034161,
		0.047876,
		0.065143,
		0.086318,
		0.111657,
		0.141286,
		0.175172,
		0.213104,
		0.254678,
		0.299306,
		0.346229,
		0.394565,
		0.443369,
		0.491718,
		0.53881,
		0.584062,
		0.627179,
		0.66815,
		0.707107,
		0.707073,
		0.706621,
		0.704898,
		0.700845,
		0.693416,
		0.681705,
		0.665023,
		0.642924,
		0.615215,
		0.581942,
		0.543376,
		0.5,
		0.452485,
		0.401677,
		0.348584,
		0.294361,
		0.240301,
		0.187825,
		0.138472,
		0.093895,
		0.055852,
		0.026203,
		0.006903,
		1.6e-005,
		7.2e-005,
		0.000227,
		0.000585,
		0.001301,
		0.002592,
		0.004739,
		0.008087,
		0.013038,
		0.02004,
		0.029569,
		0.042108,
		0.058116,
		0.078003,
		0.102092,
		0.130588,
		0.163544,
		0.200837,
		0.242148,
		0.286955,
		0.334543,
		0.384031,
		0.434424,
		0.484687,
		0.533837,
		0.581049,
		0.62576,
		0.66774,
		0.707073,
		0.743999,
		0.743962,
		0.743467,
		0.741583,
		0.73716,
		0.729065,
		0.716324,
		0.698204,
		0.674241,
		0.644244,
		0.608287,
		0.566689,
		0.519995,
		0.468959,
		0.414524,
		0.357808,
		0.300088,
		0.242791,
		0.187482,
		0.135857,
		0.089732,
		0.051039,
		0.021821,
		0.004224,
		3e-005,
		0.000121,
		0.000357,
		0.000871,
		0.001853,
		0.003559,
		0.006313,
		0.010502,
		0.016567,
		0.02499,
		0.036269,
		0.0509,
		0.069338,
		0.09197,
		0.119079,
		0.150806,
		0.187125,
		0.22781,
		0.272428,
		0.320332,
		0.370678,
		0.422467,
		0.474596,
		0.52595,
		0.575499,
		0.622414,
		0.666171,
		0.706621,
		0.743962,
		0.778554,
		0.778514,
		0.777975,
		0.775926,
		0.771123,
		0.762347,
		0.748558,
		0.72898,
		0.703131,
		0.67083,
		0.632181,
		0.587553,
		0.537563,
		0.48305,
		0.425058,
		0.364821,
		0.303745,
		0.243396,
		0.18549,
		0.131883,
		0.084562,
		0.045641,
		0.01735,
		0.002032,
		4.5e-005,
		0.000179,
		0.000512,
		0.001208,
		0.002495,
		0.004672,
		0.008105,
		0.013224,
		0.020506,
		0.030463,
		0.043615,
		0.06046,
		0.081444,
		0.106923,
		0.137128,
		0.172129,
		0.211808,
		0.25583,
		0.303641,
		0.354464,
		0.407332,
		0.461124,
		0.514646,
		0.566711,
		0.616261,
		0.662478,
		0.704898,
		0.743467,
		0.778514,
		0.810531,
		0.810487,
		0.809903,
		0.807686,
		0.802496,
		0.793028,
		0.778177,
		0.757127,
		0.729382,
		0.694774,
		0.653439,
		0.605806,
		0.552562,
		0.494641,
		0.433192,
		0.369568,
		0.30531,
		0.242128,
		0.181895,
		0.126633,
		0.078506,
		0.039813,
		0.012978,
		0.000547,
		5.7e-005,
		0.000237,
		0.000679,
		0.001579,
		0.003208,
		0.005905,
		0.010082,
		0.016209,
		0.0248,
		0.036394,
		0.051523,
		0.070687,
		0.094309,
		0.122709,
		0.156055,
		0.194338,
		0.237336,
		0.284601,
		0.33545,
		0.388981,
		0.444099,
		0.499578,
		0.554134,
		0.606532,
		0.655698,
		0.700845,
		0.741583,
		0.777975,
		0.810487,
		0.839755,
		0.839707,
		0.839077,
		0.836688,
		0.831105,
		0.820937,
		0.805015,
		0.782486,
		0.752845,
		0.715937,
		0.67194,
		0.621342,
		0.564909,
		0.50367,
		0.438887,
		0.372038,
		0.304801,
		0.239036,
		0.176778,
		0.120219,
		0.071706,
		0.033727,
		0.008906,
		6.2e-005,
		0.000287,
		0.000842,
		0.001962,
		0.00396,
		0.007219,
		0.012195,
		0.019399,
		0.029379,
		0.042695,
		0.05989,
		0.081453,
		0.107782,
		0.139147,
		0.175647,
		0.217181,
		0.26342,
		0.313795,
		0.367494,
		0.423487,
		0.480563,
		0.537395,
		0.59263,
		0.644993,
		0.693416,
		0.73716,
		0.775926,
		0.809903,
		0.839707,
		0.86612,
		0.866069,
		0.865392,
		0.862827,
		0.856847,
		0.845973,
		0.828975,
		0.804966,
		0.773435,
		0.734247,
		0.687623,
		0.634115,
		0.574574,
		0.510129,
		0.442157,
		0.372267,
		0.302277,
		0.234205,
		0.170248,
		0.112777,
		0.064324,
		0.027572,
		0.00535,
		5.5e-005,
		0.000316,
		0.000983,
		0.002331,
		0.004715,
		0.008567,
		0.014386,
		0.022723,
		0.034155,
		0.049261,
		0.068588,
		0.092608,
		0.121685,
		0.156029,
		0.195661,
		0.240381,
		0.289746,
		0.343062,
		0.399391,
		0.457578,
		0.516304,
		0.574154,
		0.629719,
		0.681705,
		0.729065,
		0.771123,
		0.807686,
		0.839077,
		0.866069,
		0.88959,
		0.889535,
		0.88881,
		0.886069,
		0.879686,
		0.868102,
		0.850026,
		0.824541,
		0.791133,
		0.749691,
		0.700485,
		0.644135,
		0.581581,
		0.514056,
		0.443057,
		0.370327,
		0.297832,
		0.227748,
		0.162442,
		0.104466,
		0.05654,
		0.021549,
		0.00253,
		3.7e-005,
		0.000317,
		0.001083,
		0.002653,
		0.005431,
		0.009895,
		0.016587,
		0.026096,
		0.039027,
		0.055972,
		0.077474,
		0.103984,
		0.135819,
		0.173123,
		0.21583,
		0.263636,
		0.315979,
		0.372039,
		0.430753,
		0.49085,
		0.550912,
		0.609454,
		0.665023,
		0.716324,
		0.762347,
		0.802496,
		0.836688,
		0.865392,
		0.889535,
		0.910193,
		0.910135,
		0.909361,
		0.906441,
		0.899653,
		0.887355,
		0.868202,
		0.841247,
		0.80598,
		0.762317,
		0.71058,
		0.651464,
		0.586001,
		0.515532,
		0.441681,
		0.366327,
		0.291589,
		0.219804,
		0.153515,
		0.095456,
		0.048543,
		0.015865,
		0.000671,
		1.3e-005,
		0.000283,
		0.001123,
		0.0029,
		0.006065,
		0.011143,
		0.018722,
		0.029425,
		0.043882,
		0.062693,
		0.08639,
		0.115392,
		0.149966,
		0.190179,
		0.235873,
		0.286631,
		0.341773,
		0.400354,
		0.461192,
		0.522909,
		0.584001,
		0.642924,
		0.698204,
		0.748558,
		0.793028,
		0.831105,
		0.862827,
		0.88881,
		0.910135,
		0.928019,
		0.927957,
		0.927134,
		0.924032,
		0.916836,
		0.903823,
		0.883593,
		0.855177,
		0.818069,
		0.772221,
		0.71801,
		0.656209,
		0.587947,
		0.514679,
		0.438158,
		0.360406,
		0.283695,
		0.210532,
		0.143636,
		0.085931,
		0.040531,
		0.010731,
		0.000215,
		0.001091,
		0.003043,
		0.00657,
		0.01225,
		0.020712,
		0.032612,
		0.0486,
		0.069278,
		0.095162,
		0.126633,
		0.163895,
		0.206935,
		0.255494,
		0.309044,
		0.366781,
		0.427644,
		0.490336,
		0.553385,
		0.615215,
		0.674241,
		0.72898,
		0.778177,
		0.820937,
		0.856847,
		0.886069,
		0.909361,
		0.927957,
		0.943208,
		0.943143,
		0.942269,
		0.938984,
		0.931376,
		0.917645,
		0.896339,
		0.86647,
		0.827542,
		0.779544,
		0.722916,
		0.658512,
		0.587564,
		0.511645,
		0.43264,
		0.352718,
		0.274314,
		0.200101,
		0.132984,
		0.076079,
		0.032703,
		0.006363,
		0.000127,
		0.000982,
		0.003057,
		0.006907,
		0.013156,
		0.022475,
		0.035553,
		0.053053,
		0.075577,
		0.103613,
		0.137498,
		0.177368,
		0.223122,
		0.274398,
		0.330552,
		0.390664,
		0.453558,
		0.517835,
		0.581942,
		0.644244,
		0.703131,
		0.757127,
		0.805015,
		0.845973,
		0.879686,
		0.906441,
		0.927134,
		0.943143,
		0.955945,
		0.955875,
		0.95495,
		0.951479,
		0.943456,
		0.929002,
		0.90662,
		0.875305,
		0.834575,
		0.784462,
		0.725472,
		0.658546,
		0.585024,
		0.5066,
		0.425297,
		0.343437,
		0.263617,
		0.188689,
		0.121743,
		0.066092,
		0.025261,
		0.002974,
		4.1e-005,
		0.000799,
		0.002929,
		0.007038,
		0.013802,
		0.023933,
		0.038146,
		0.057115,
		0.081434,
		0.11156,
		0.147776,
		0.190145,
		0.238473,
		0.292291,
		0.350845,
		0.4131,
		0.477773,
		0.543376,
		0.608287,
		0.67083,
		0.729382,
		0.782486,
		0.828975,
		0.868102,
		0.899653,
		0.924032,
		0.942269,
		0.955875,
		0.966443,
		0.966369,
		0.965392,
		0.961733,
		0.953289,
		0.938108,
		0.914647,
		0.881889,
		0.839373,
		0.787175,
		0.725873,
		0.656502,
		0.580512,
		0.499726,
		0.416309,
		0.33274,
		0.251784,
		0.176476,
		0.110096,
		0.056161,
		0.018407,
		0.00078,
		0.000562,
		0.002653,
		0.006937,
		0.014139,
		0.025011,
		0.040291,
		0.060661,
		0.086698,
		0.118823,
		0.15726,
		0.20199,
		0.252728,
		0.308897,
		0.369632,
		0.433793,
		0.5,
		0.566689,
		0.632181,
		0.694774,
		0.752845,
		0.804966,
		0.850026,
		0.887355,
		0.916836,
		0.938984,
		0.95495,
		0.966369,
		0.974941,
		0.974863,
		0.973833,
		0.969981,
		0.961111,
		0.945195,
		0.92065,
		0.886448,
		0.842154,
		0.787895,
		0.724327,
		0.652581,
		0.574223,
		0.491211,
		0.405858,
		0.320804,
		0.238991,
		0.163637,
		0.098223,
		0.046472,
		0.01234,
		0.000308,
		0.002242,
		0.006591,
		0.014128,
		0.025644,
		0.041898,
		0.063573,
		0.091224,
		0.12523,
		0.16575,
		0.212684,
		0.265647,
		0.323959,
		0.386651,
		0.452485,
		0.519995,
		0.587553,
		0.653439,
		0.715937,
		0.773435,
		0.824541,
		0.868202,
		0.903823,
		0.931376,
		0.951479,
		0.965392,
		0.974863,
		0.981688,
		0.981605,
		0.98052,
		0.976471,
		0.967166,
		0.950506,
		0.924865,
		0.889215,
		0.843145,
		0.786843,
		0.721044,
		0.646984,
		0.56635,
		0.48124,
		0.394122,
		0.307804,
		0.225407,
		0.150344,
		0.0863,
		0.037212,
		0.007261,
		9.4e-005,
		0.001726,
		0.006001,
		0.013746,
		0.025783,
		0.04289,
		0.065746,
		0.09488,
		0.130624,
		0.173066,
		0.222021,
		0.27701,
		0.337251,
		0.401677,
		0.468959,
		0.537563,
		0.605806,
		0.67194,
		0.734247,
		0.791133,
		0.841247,
		0.883593,
		0.917645,
		0.943456,
		0.961733,
		0.973833,
		0.981605,
		0.986931,
		0.986844,
		0.985703,
		0.981452,
		0.971702,
		0.954283,
		0.927533,
		0.890421,
		0.842572,
		0.784233,
		0.71623,
		0.639906,
		0.557079,
		0.469991,
		0.381271,
		0.293903,
		0.211195,
		0.136762,
		0.074497,
		0.028563,
		0.003372,
		0.001155,
		0.005191,
		0.012984,
		0.025391,
		0.043203,
		0.067091,
		0.097552,
		0.134865,
		0.179047,
		0.229828,
		0.286632,
		0.348584,
		0.414524,
		0.48305,
		0.552562,
		0.621342,
		0.687623,
		0.749691,
		0.80598,
		0.855177,
		0.896339,
		0.929002,
		0.953289,
		0.969981,
		0.98052,
		0.986844,
		0.990914,
		0.990823,
		0.989623,
		0.985163,
		0.974957,
		0.956762,
		0.928881,
		0.890289,
		0.840647,
		0.780271,
		0.710079,
		0.631533,
		0.546584,
		0.457628,
		0.367463,
		0.279254,
		0.196507,
		0.123044,
		0.06298,
		0.020707,
		0.00088,
		0.000604,
		0.004205,
		0.011859,
		0.024457,
		0.042798,
		0.067541,
		0.099147,
		0.13784,
		0.183562,
		0.235957,
		0.294361,
		0.357808,
		0.425058,
		0.494641,
		0.564909,
		0.634115,
		0.700485,
		0.762317,
		0.818069,
		0.86647,
		0.90662,
		0.938108,
		0.961111,
		0.976471,
		0.985703,
		0.990823,
		0.993865,
		0.993768,
		0.992508,
		0.987832,
		0.977154,
		0.958161,
		0.929125,
		0.889025,
		0.837568,
		0.775143,
		0.702768,
		0.622029,
		0.535021,
		0.444298,
		0.352837,
		0.263995,
		0.181483,
		0.109339,
		0.05191,
		0.013828,
		0.000176,
		0.003111,
		0.010409,
		0.022989,
		0.041656,
		0.067052,
		0.0996,
		0.139462,
		0.186509,
		0.240301,
		0.300088,
		0.364821,
		0.433192,
		0.50367,
		0.574574,
		0.644135,
		0.71058,
		0.772221,
		0.827542,
		0.875305,
		0.914647,
		0.945195,
		0.967166,
		0.981452,
		0.989623,
		0.993768,
		0.99599,
		0.995889,
		0.994565,
		0.989664,
		0.978498,
		0.958682,
		0.928457,
		0.886816,
		0.833512,
		0.769016,
		0.694454,
		0.611541,
		0.522525,
		0.430131,
		0.337521,
		0.248253,
		0.166253,
		0.095789,
		0.041447,
		0.008113,
		0.002007,
		0.008699,
		0.021023,
		0.039786,
		0.065609,
		0.098873,
		0.13968,
		0.187825,
		0.242791,
		0.303745,
		0.369568,
		0.438887,
		0.510129,
		0.581581,
		0.651464,
		0.71801,
		0.779544,
		0.834575,
		0.881889,
		0.92065,
		0.950506,
		0.971702,
		0.985163,
		0.992508,
		0.995889,
		0.997474,
		0.997368,
		0.995977,
		0.990841,
		0.979167,
		0.958497,
		0.927046,
		0.883821,
		0.828631,
		0.762033,
		0.685268,
		0.600193,
		0.509214,
		0.415238,
		0.321622,
		0.232137,
		0.150937,
		0.082529,
		0.031753,
		0.003761,
		0.001015,
		0.006822,
		0.018622,
		0.037225,
		0.063227,
		0.096962,
		0.138472,
		0.187482,
		0.243396,
		0.30531,
		0.372038,
		0.442157,
		0.514056,
		0.586001,
		0.656209,
		0.722916,
		0.784462,
		0.839373,
		0.886448,
		0.924865,
		0.954283,
		0.974957,
		0.987832,
		0.994565,
		0.997368,
		0.998475,
		0.998363,
		0.996902,
		0.99152,
		0.979315,
		0.957757,
		0.925035,
		0.880177,
		0.823051,
		0.75431,
		0.67532,
		0.588084,
		0.495182,
		0.39971,
		0.305234,
		0.215749,
		0.13565,
		0.069697,
		0.022996,
		0.000981,
		0.000286,
		0.004895,
		0.015878,
		0.03404,
		0.059949,
		0.093895,
		0.135857,
		0.18549,
		0.242128,
		0.304801,
		0.372267,
		0.443057,
		0.515532,
		0.587947,
		0.658512,
		0.725472,
		0.787175,
		0.842154,
		0.889215,
		0.927533,
		0.956762,
		0.977154,
		0.989664,
		0.995977,
		0.998363,
		0.999123,
		0.999005,
		0.997471,
		0.991829,
		0.979067,
		0.956583,
		0.92254,
		0.875992,
		0.816874,
		0.745942,
		0.664693,
		0.575293,
		0.480503,
		0.383622,
		0.288436,
		0.199181,
		0.120501,
		0.057429,
		0.015352,
		0.003067,
		0.012909,
		0.030325,
		0.055852,
		0.089732,
		0.131883,
		0.181895,
		0.239036,
		0.302277,
		0.370327,
		0.441681,
		0.514679,
		0.587564,
		0.658546,
		0.725873,
		0.787895,
		0.843145,
		0.890421,
		0.928881,
		0.958161,
		0.978498,
		0.990841,
		0.996902,
		0.999005,
		0.999523,
		0.999399,
		0.997787,
		0.991872,
		0.978526,
		0.955071,
		0.919651,
		0.87135,
		0.810175,
		0.736995,
		0.653452,
		0.56188,
		0.465237,
		0.367036,
		0.2713,
		0.182516,
		0.105601,
		0.04587,
		0.009011,
		0.001509,
		0.009861,
		0.026203,
		0.051039,
		0.084562,
		0.126633,
		0.176778,
		0.234205,
		0.297832,
		0.366327,
		0.438158,
		0.511645,
		0.585024,
		0.656502,
		0.724327,
		0.786843,
		0.842572,
		0.890289,
		0.929125,
		0.958682,
		0.979167,
		0.99152,
		0.997471,
		0.999399,
		0.999758,
		0.999627,
		0.997932,
		0.991729,
		0.977766,
		0.953294,
		0.916436,
		0.86631,
		0.803008,
		0.727518,
		0.641638,
		0.547885,
		0.449423,
		0.349999,
		0.253886,
		0.165839,
		0.091062,
		0.035173,
		0.004181,
		0.000415,
		0.006903,
		0.021821,
		0.045641,
		0.078506,
		0.120219,
		0.170248,
		0.227748,
		0.291589,
		0.360406,
		0.43264,
		0.5066,
		0.580512,
		0.652581,
		0.721044,
		0.784233,
		0.840647,
		0.889025,
		0.928457,
		0.958497,
		0.979315,
		0.991829,
		0.997787,
		0.999627,
		0.999887,
		0.999749,
		0.997966,
		0.991457,
		0.976845,
		0.951302,
		0.912939,
		0.860913,
		0.795406,
		0.71754,
		0.629277,
		0.533332,
		0.433094,
		0.332553,
		0.236251,
		0.149231,
		0.077003,
		0.025507,
		0.001092,
		0.004224,
		0.01735,
		0.039813,
		0.071706,
		0.112777,
		0.162442,
		0.219804,
		0.283695,
		0.352718,
		0.425297,
		0.499726,
		0.574223,
		0.646984,
		0.71623,
		0.780271,
		0.837568,
		0.886816,
		0.927046,
		0.957757,
		0.979067,
		0.991872,
		0.997932,
		0.999749,
		0.999952,
		0.999806,
		0.99793,
		0.991097,
		0.975798,
		0.94913,
		0.909189,
		0.855181,
		0.787387,
		0.707073,
		0.616381,
		0.518237,
		0.416269,
		0.314732,
		0.218452,
		0.132781,
		0.063554,
		0.017057,
		0.002032,
		0.012978,
		0.033727,
		0.064324,
		0.104466,
		0.153515,
		0.210532,
		0.274314,
		0.343437,
		0.416309,
		0.491211,
		0.56635,
		0.639906,
		0.710079,
		0.775143,
		0.833512,
		0.883821,
		0.925035,
		0.956583,
		0.978526,
		0.991729,
		0.997966,
		0.999806,
		0.999982,
		0.999829,
		0.997853,
		0.990675,
		0.97465,
		0.946796,
		0.905199,
		0.849122,
		0.778955,
		0.696118,
		0.60295,
		0.502605,
		0.398965,
		0.296568,
		0.200548,
		0.116585,
		0.050861,
		0.010032,
		0.000547,
		0.008906,
		0.027572,
		0.05654,
		0.095456,
		0.143636,
		0.200101,
		0.263617,
		0.33274,
		0.405858,
		0.48124,
		0.557079,
		0.631533,
		0.702768,
		0.769016,
		0.828631,
		0.880177,
		0.92254,
		0.955071,
		0.977766,
		0.991457,
		0.99793,
		0.999829,
		0.999994,
		0.999832,
		0.99775,
		0.990206,
		0.973413,
		0.944307,
		0.900973,
		0.842734,
		0.770102,
		0.684668,
		0.588979,
		0.486434,
		0.381192,
		0.278094,
		0.182602,
		0.100747,
		0.039085,
		0.004665,
		0.00535,
		0.021549,
		0.048543,
		0.085931,
		0.132984,
		0.188689,
		0.251784,
		0.320804,
		0.394122,
		0.469991,
		0.546584,
		0.622029,
		0.694454,
		0.762033,
		0.823051,
		0.875992,
		0.919651,
		0.953294,
		0.976845,
		0.991097,
		0.997853,
		0.999832,
		0.999999,
		0.999827,
		0.997631,
		0.989698,
		0.97209,
		0.941665,
		0.896504,
		0.836006,
		0.760816,
		0.672707,
		0.574454,
		0.469721,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.00253,
		0.015865,
		0.040531,
		0.076079,
		0.121743,
		0.176476,
		0.238991,
		0.307804,
		0.381271,
		0.457628,
		0.535021,
		0.611541,
		0.685268,
		0.75431,
		0.816874,
		0.87135,
		0.916436,
		0.951302,
		0.975798,
		0.990675,
		0.99775,
		0.999827,
		1,
		0.999818,
		0.997501,
		0.989153,
		0.970681,
		0.938862,
		0.891783,
		0.828924,
		0.751076,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.000671,
		0.010731,
		0.032703,
		0.066092,
		0.110096,
		0.163637,
		0.225407,
		0.293903,
		0.367463,
		0.444298,
		0.522525,
		0.600193,
		0.67532,
		0.745942,
		0.810175,
		0.86631,
		0.912939,
		0.94913,
		0.97465,
		0.990206,
		0.997631,
		0.999818,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325197,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.006363,
		0.025261,
		0.056161,
		0.098223,
		0.150344,
		0.211195,
		0.279254,
		0.352837,
		0.430131,
		0.509214,
		0.588084,
		0.664693,
		0.736995,
		0.803008,
		0.860913,
		0.909189,
		0.946796,
		0.973413,
		0.989698,
		0.997501,
		0.999808,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.002974,
		0.018407,
		0.046472,
		0.0863,
		0.136762,
		0.196507,
		0.263995,
		0.337521,
		0.415238,
		0.495182,
		0.575293,
		0.653452,
		0.727518,
		0.795406,
		0.855181,
		0.905199,
		0.944307,
		0.97209,
		0.989153,
		0.997361,
		0.999797,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.00078,
		0.01234,
		0.037212,
		0.074497,
		0.123044,
		0.181483,
		0.248253,
		0.321622,
		0.39971,
		0.480503,
		0.56188,
		0.641638,
		0.71754,
		0.787387,
		0.849122,
		0.900973,
		0.941665,
		0.970681,
		0.988571,
		0.99721,
		0.999784,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.007261,
		0.028563,
		0.06298,
		0.109339,
		0.166253,
		0.232137,
		0.305234,
		0.383622,
		0.465237,
		0.547885,
		0.629277,
		0.707073,
		0.778955,
		0.842734,
		0.896504,
		0.938862,
		0.969182,
		0.987949,
		0.997049,
		0.999771,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.003372,
		0.020707,
		0.05191,
		0.095789,
		0.150937,
		0.215749,
		0.288436,
		0.367036,
		0.449423,
		0.533332,
		0.616381,
		0.696118,
		0.770102,
		0.836006,
		0.891783,
		0.93589,
		0.967586,
		0.987285,
		0.996876,
		0.999757,
		1,
		0.999741,
		0.99649,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.00088,
		0.013828,
		0.041447,
		0.082529,
		0.13565,
		0.199181,
		0.2713,
		0.349999,
		0.433094,
		0.518237,
		0.60295,
		0.684668,
		0.760816,
		0.828924,
		0.886795,
		0.932739,
		0.965887,
		0.986574,
		0.99669,
		0.999741,
		1,
		0.999725,
		0.996274,
		0.984125,
		0.957881,
		0.913781,
		0.850172,
		0.767497,
		0.668086,
		0.555928,
		0.436458,
		0.316383,
		0.203541,
		0.106789,
		0.035914,
		0.001559,
		0.008113,
		0.031753,
		0.069697,
		0.120501,
		0.182516,
		0.253886,
		0.332553,
		0.416269,
		0.502605,
		0.588979,
		0.672707,
		0.751076,
		0.821467,
		0.881522,
		0.929393,
		0.964076,
		0.985814,
		0.99649,
		0.999725,
		1,
		0.999707,
		0.996042,
		0.983187,
		0.955526,
		0.909232,
		0.842738,
		0.756699,
		0.653763,
		0.538323,
		0.416289,
		0.294906,
		0.182603,
		0.088884,
		0.024223,
		0.003761,
		0.022996,
		0.057429,
		0.105601,
		0.165839,
		0.236251,
		0.314732,
		0.398965,
		0.486434,
		0.574454,
		0.660215,
		0.740862,
		0.813614,
		0.875946,
		0.925841,
		0.962145,
		0.984999,
		0.996274,
		0.999707,
		1,
		0.999687,
		0.995792,
		0.982179,
		0.953007,
		0.904386,
		0.834852,
		0.745299,
		0.638727,
		0.519969,
		0.395453,
		0.273008,
		0.16171,
		0.071756,
		0.014374,
		0.000981,
		0.015352,
		0.04587,
		0.091062,
		0.149231,
		0.218452,
		0.296568,
		0.381192,
		0.469721,
		0.559361,
		0.647171,
		0.730146,
		0.80534,
		0.870045,
		0.922066,
		0.960083,
		0.984125,
		0.996042,
		0.999687,
		1,
		0.999666,
		0.995521,
		0.981094,
		0.950308,
		0.899218,
		0.826481,
		0.73326,
		0.622942,
		0.500843,
		0.373956,
		0.250747,
		0.140991,
		0.055645,
		0.006746,
		0.009011,
		0.035173,
		0.077003,
		0.132781,
		0.200548,
		0.278094,
		0.362964,
		0.452462,
		0.543682,
		0.63355,
		0.718904,
		0.796619,
		0.863799,
		0.918052,
		0.957881,
		0.983187,
		0.995792,
		0.999666,
		1,
		0.999643,
		0.995229,
		0.979927,
		0.947416,
		0.893703,
		0.817592,
		0.720543,
		0.606372,
		0.480925,
		0.351814,
		0.228195,
		0.120607,
		0.04083,
		0.001783,
		0.004181,
		0.025507,
		0.063554,
		0.116585,
		0.182602,
		0.259348,
		0.344293,
		0.434652,
		0.527402,
		0.619329,
		0.707107,
		0.787425,
		0.857183,
		0.913781,
		0.955526,
		0.982179,
		0.995521,
		0.999643,
		1,
		0.999618,
		0.994912,
		0.978669,
		0.944313,
		0.887815,
		0.808146,
		0.707107,
		0.588982,
		0.460201,
		0.32905,
		0.205443,
		0.100747,
		0.027641,
		0.001092,
		0.017057,
		0.050861,
		0.100747,
		0.164687,
		0.24037,
		0.325197,
		0.416289,
		0.510503,
		0.604481,
		0.694725,
		0.777728,
		0.850172,
		0.909232,
		0.953007,
		0.981094,
		0.995229,
		0.999618,
		1,
		0.999591,
		0.994569,
		0.977312,
		0.94098,
		0.881522,
		0.798105,
		0.692907,
		0.570735,
		0.438658,
		0.3057,
		0.182603,
		0.08164,
		0.016464,
		0.010032,
		0.039085,
		0.085389,
		0.146884,
		0.221211,
		0.3057,
		0.397375,
		0.492972,
		0.588982,
		0.681729,
		0.767497,
		0.842738,
		0.904386,
		0.950308,
		0.979927,
		0.994912,
		0.999591,
		1,
		0.999561,
		0.994196,
		0.975845,
		0.937398,
		0.874792,
		0.787425,
		0.677898,
		0.551596,
		0.416289,
		0.281814,
		0.159817,
		0.063557,
		0.007758,
		0.004665,
		0.02841,
		0.070648,
		0.129291,
		0.201932,
		0.28583,
		0.377913,
		0.474793,
		0.572806,
		0.668086,
		0.756699,
		0.834852,
		0.899218,
		0.947416,
		0.978669,
		0.994569,
		0.999561,
		1,
		0.999529,
		0.993791,
		0.97426,
		0.933544,
		0.86759,
		0.776061,
		0.662033,
		0.531529,
		0.393097,
		0.25746,
		0.137255,
		0.046826,
		0.002059,
		0.001221,
		0.019046,
		0.056682,
		0.112019,
		0.182603,
		0.265625,
		0.357915,
		0.455958,
		0.555928,
		0.653763,
		0.745299,
		0.826481,
		0.893703,
		0.944313,
		0.977312,
		0.994196,
		0.999529,
		1,
		0.999493,
		0.99335,
		0.972543,
		0.929393,
		0.859876,
		0.763962,
		0.645261,
		0.510503,
		0.369091,
		0.232725,
		0.11513,
		0.031834,
		0.011231,
		0.043674,
		0.0952,
		0.163312,
		0.245133,
		0.337397,
		0.436458,
		0.538323,
		0.638727,
		0.73326,
		0.817592,
		0.887815,
		0.94098,
		0.975845,
		0.993791,
		0.999493,
		1,
		0.999454,
		0.99287,
		0.970681,
		0.924919,
		0.851607,
		0.751077,
		0.62753,
		0.488488,
		0.344293,
		0.207724,
		0.093699,
		0.019046,
		0.005236,
		0.031834,
		0.078987,
		0.144161,
		0.224414,
		0.316383,
		0.416289,
		0.519969,
		0.622942,
		0.720543,
		0.808146,
		0.881522,
		0.937398,
		0.97426,
		0.99335,
		0.999454,
		1,
		0.999412,
		0.992346,
		0.968661,
		0.92009,
		0.842738,
		0.737347,
		0.608789,
		0.465459,
		0.318742,
		0.182603,
		0.073275,
		0.009016,
		0.001374,
		0.021402,
		0.063557,
		0.125272,
		0.203541,
		0.294906,
		0.395453,
		0.500843,
		0.606372,
		0.707107,
		0.798105,
		0.874792,
		0.933544,
		0.972543,
		0.99287,
		0.999412,
		1,
		0.999365,
		0.991772,
		0.966466,
		0.914874,
		0.833218,
		0.722712,
		0.588982,
		0.441396,
		0.292492,
		0.157547,
		0.05424,
		0.002404,
		0.012657,
		0.049119,
		0.106789,
		0.182603,
		0.273008,
		0.373956,
		0.480925,
		0.588982,
		0.692907,
		0.787425,
		0.86759,
		0.929393,
		0.970681,
		0.992346,
		0.999365,
		1,
		0.999313,
		0.991144,
		0.964076,
		0.909232,
		0.82299,
		0.707107,
		0.568056,
		0.416289,
		0.265625,
		0.132788,
		0.037057,
		0.00592,
		0.035914,
		0.088884,
		0.16171,
		0.250747,
		0.351814,
		0.460201,
		0.570735,
		0.677898,
		0.776061,
		0.859876,
		0.924919,
		0.968661,
		0.991772,
		0.999313,
		1,
		0.999256,
		0.990456,
		0.961473,
		0.903125,
		0.811993,
		0.690463,
		0.545959,
		0.39014,
		0.238249,
		0.108615,
		0.022285,
		0.001559,
		0.024223,
		0.071756,
		0.140991,
		0.228195,
		0.32905,
		0.438658,
		0.551596,
		0.662033,
		0.763962,
		0.851607,
		0.92009,
		0.966466,
		0.991144,
		0.999256,
		1,
		0.999193,
		0.9897,
		0.958631,
		0.896506,
		0.800163,
		0.672708,
		0.522638,
		0.362964,
		0.210511,
		0.085389,
		0.010606,
		0.014374,
		0.055645,
		0.120607,
		0.205443,
		0.3057,
		0.416289,
		0.531529,
		0.645261,
		0.751077,
		0.842738,
		0.914874,
		0.964076,
		0.990456,
		0.999193,
		1,
		0.999123,
		0.988867,
		0.955526,
		0.889324,
		0.787425,
		0.653763,
		0.498046,
		0.334797,
		0.182603,
		0.063557,
		0.002844,
		0.006746,
		0.04083,
		0.100747,
		0.182603,
		0.281814,
		0.393097,
		0.510503,
		0.62753,
		0.737347,
		0.833218,
		0.909232,
		0.961473,
		0.9897,
		0.999123,
		1,
		0.999046,
		0.987949,
		0.952128,
		0.881522,
		0.773702,
		0.63355,
		0.472143,
		0.3057,
		0.154777,
		0.043674,
		0.001783,
		0.027641,
		0.08164,
		0.159817,
		0.25746,
		0.369091,
		0.488488,
		0.608789,
		0.722712,
		0.82299,
		0.903125,
		0.958631,
		0.988867,
		0.999046,
		1,
		0.99896,
		0.986936,
		0.948403,
		0.873037,
		0.758906,
		0.611985,
		0.444897,
		0.275766,
		0.127354,
		0.026425,
		0.016464,
		0.063557,
		0.137255,
		0.232725,
		0.344293,
		0.465459,
		0.588982,
		0.707107,
		0.811993,
		0.896506,
		0.955526,
		0.987949,
		0.99896,
		1,
		0.998863,
		0.985814,
		0.944313,
		0.863799,
		0.742944,
		0.588982,
		0.416289,
		0.245133,
		0.100747,
		0.012657,
		0.007758,
		0.046826,
		0.11513,
		0.207724,
		0.318742,
		0.441396,
		0.568056,
		0.690463,
		0.800163,
		0.889324,
		0.952128,
		0.986936,
		0.998863,
		1,
		0.998756,
		0.984569,
		0.939815,
		0.853729,
		0.725714,
		0.564456,
		0.38632,
		0.213992,
		0.075484,
		0.003417,
		0.002059,
		0.031834,
		0.093699,
		0.182603,
		0.292492,
		0.416289,
		0.545959,
		0.672708,
		0.787425,
		0.881522,
		0.948403,
		0.985814,
		0.998756,
		1,
		0.998635,
		0.983187,
		0.93486,
		0.842738,
		0.707107,
		0.538323,
		0.355015,
		0.182603,
		0.05223,
		0.019046,
		0.073275,
		0.157547,
		0.265625,
		0.39014,
		0.522638,
		0.653763,
		0.773702,
		0.873037,
		0.944313,
		0.984569,
		0.998635,
		1,
		0.998499,
		0.981647,
		0.929393,
		0.830729,
		0.687003,
		0.510503,
		0.322436,
		0.151319,
		0.031834,
		0.009016,
		0.05424,
		0.132788,
		0.238249,
		0.362964,
		0.498046,
		0.63355,
		0.758906,
		0.863799,
		0.939815,
		0.983187,
		0.998499,
		1,
		0.998347,
		0.979927,
		0.92335,
		0.817592,
		0.665277,
		0.480925,
		0.28869,
		0.120607,
		0.015366,
		0.002404,
		0.037057,
		0.108615,
		0.210511,
		0.334797,
		0.472143,
		0.611985,
		0.742944,
		0.853729,
		0.93486,
		0.981647,
		0.998347,
		1,
		0.998174,
		0.978003,
		0.916658,
		0.803203,
		0.641793,
		0.449532,
		0.253947,
		0.091084,
		0.004182,
		0.022285,
		0.085389,
		0.182603,
		0.3057,
		0.444897,
		0.588982,
		0.725714,
		0.842738,
		0.929393,
		0.979927,
		0.998174,
		1,
		0.997978,
		0.975845,
		0.909232,
		0.787425,
		0.61641,
		0.416289,
		0.218463,
		0.063557,
		0.010606,
		0.063557,
		0.154777,
		0.275766,
		0.416289,
		0.564456,
		0.707107,
		0.830729,
		0.92335,
		0.978003,
		0.997978,
		1,
		0.997755,
		0.973418,
		0.900978,
		0.770107,
		0.588982,
		0.381194,
		0.182603,
		0.039085,
		0.002844,
		0.043674,
		0.127354,
		0.245133,
		0.38632,
		0.538323,
		0.687003,
		0.817592,
		0.916658,
		0.975845,
		0.997755,
		1,
		0.997501,
		0.970681,
		0.891784,
		0.751077,
		0.559361,
		0.344293,
		0.146884,
		0.019046,
		0.026425,
		0.100747,
		0.213992,
		0.355015,
		0.510503,
		0.665277,
		0.803203,
		0.909232,
		0.973418,
		0.997501,
		1,
		0.99721,
		0.967586,
		0.881522,
		0.730146,
		0.527402,
		0.3057,
		0.112019,
		0.005236,
		0.012657,
		0.075484,
		0.182603,
		0.322436,
		0.480925,
		0.641793,
		0.787425,
		0.900978,
		0.970681,
		0.99721,
		1,
		0.996876,
		0.964076,
		0.870045,
		0.707107,
		0.492972,
		0.265625,
		0.078987,
		0.003417,
		0.05223,
		0.151319,
		0.28869,
		0.449532,
		0.61641,
		0.770107,
		0.891784,
		0.967586,
		0.996876,
		1,
		0.996489,
		0.960083,
		0.857183,
		0.681729,
		0.455958,
		0.224414,
		0.049119,
		0.031834,
		0.120607,
		0.253947,
		0.416289,
		0.588982,
		0.751077,
		0.881522,
		0.964076,
		0.996489,
		0.999952,
		0.995994,
		0.955481,
		0.842698,
		0.653732,
		0.416269,
		0.182595,
		0.024222,
		0.015366,
		0.091084,
		0.218463,
		0.381194,
		0.559361,
		0.730146,
		0.870045,
		0.960083,
		0.995994,
		0.999123,
		0.994648,
		0.949475,
		0.825756,
		0.622395,
		0.373628,
		0.140868,
		0.00674,
		0.004182,
		0.063557,
		0.182603,
		0.344293,
		0.527402,
		0.707107,
		0.857183,
		0.955481,
		0.994648,
		0.993865,
		0.988808,
		0.938519,
		0.803188,
		0.585368,
		0.327031,
		0.100129,
		0.039085,
		0.146884,
		0.3057,
		0.492972,
		0.681729,
		0.842698,
		0.949475,
		0.988808,
		0.974941,
		0.969283,
		0.913908,
		0.767694,
		0.537774,
		0.274752,
		0.061965,
		0.019046,
		0.112019,
		0.265625,
		0.455958,
		0.653732,
		0.825756,
		0.938519,
		0.969283,
		0.928019,
		0.921848,
		0.862495,
		0.708972,
		0.473757,
		0.215973,
		0.029542,
		0.005236,
		0.078987,
		0.224414,
		0.416269,
		0.622395,
		0.803188,
		0.913908,
		0.921848,
		0.839755,
		0.833327,
		0.77265,
		0.61919,
		0.390871,
		0.153342,
		0.007571,
		0.049119,
		0.182595,
		0.373628,
		0.585368,
		0.767694,
		0.862495,
		0.833327,
		0.707107,
		0.700845,
		0.642924,
		0.5,
		0.294361,
		0.093895,
		0.024222,
		0.140868,
		0.327031,
		0.537774,
		0.708972,
		0.77265,
		0.700845,
		0.542966,
		0.537373,
		0.486772,
		0.365257,
		0.197077,
		0.046363,
		0.00674,
		0.100129,
		0.274752,
		0.473757,
		0.61919,
		0.642924,
		0.537373,
		0.372533,
		0.368043,
		0.328396,
		0.236018,
		0.113883,
		0.01627,
		0.061965,
		0.215973,
		0.390871,
		0.5,
		0.486772,
		0.368043,
		0.222461,
		0.219306,
		0.192162,
		0.131026,
		0.054533,
		0.002816,
		0.029542,
		0.153342,
		0.294361,
		0.365257,
		0.328396,
		0.219306,
		0.110603,
		0.108744,
		0.09321,
		0.05954,
		0.020197,
		0.007571,
		0.093895,
		0.197077,
		0.236018,
		0.192162,
		0.108744,
		0.041882,
		0.041042,
		0.034243,
		0.020142,
		0.005051,
		0.046363,
		0.113883,
		0.131026,
		0.09321,
		0.041042,
		0.0098,
		0.009563,
		0.007717,
		0.00408,
		0.000623,
		0.01627,
		0.054533,
		0.05954,
		0.034243,
		0.009563,
		0.000719,
		0.000698,
		0.00054,
		0.000248,
		1.4e-005,
		0.002816,
		0.020197,
		0.020142,
		0.007717,
		0.000698,
		0.005051,
		0.00408,
		0.00054,
		0.000623,
		0.000248,
		1.4e-005,
		1.4e-005,
		0.000248,
		0.000623,
		0.00054,
		0.00408,
		0.005051,
		0.000698,
		0.007717,
		0.020142,
		0.020197,
		0.002816,
		1.4e-005,
		0.000248,
		0.00054,
		0.000698,
		0.000719,
		0.009563,
		0.034243,
		0.05954,
		0.054533,
		0.01627,
		0.000623,
		0.00408,
		0.007717,
		0.009563,
		0.0098,
		0.041042,
		0.09321,
		0.131026,
		0.113883,
		0.046363,
		0.005051,
		0.020142,
		0.034243,
		0.041042,
		0.041882,
		0.108744,
		0.192162,
		0.236018,
		0.197077,
		0.093895,
		0.007571,
		0.020197,
		0.05954,
		0.09321,
		0.108744,
		0.110603,
		0.219306,
		0.328396,
		0.365257,
		0.294361,
		0.153342,
		0.029542,
		0.002816,
		0.054533,
		0.131026,
		0.192162,
		0.219306,
		0.222461,
		0.368043,
		0.486772,
		0.5,
		0.390871,
		0.215973,
		0.061965,
		0.01627,
		0.113883,
		0.236018,
		0.328396,
		0.368043,
		0.372533,
		0.537373,
		0.642924,
		0.61919,
		0.473757,
		0.274752,
		0.100129,
		0.00674,
		0.046363,
		0.197077,
		0.365257,
		0.486772,
		0.537373,
		0.542966,
		0.700845,
		0.77265,
		0.708972,
		0.537774,
		0.327031,
		0.140868,
		0.024222,
		0.093895,
		0.294361,
		0.5,
		0.642924,
		0.700845,
		0.707107,
		0.833327,
		0.862495,
		0.767694,
		0.585368,
		0.373628,
		0.182595,
		0.049119,
		0.007571,
		0.153342,
		0.390871,
		0.61919,
		0.77265,
		0.833327,
		0.839755,
		0.921848,
		0.913908,
		0.803188,
		0.622395,
		0.416269,
		0.224414,
		0.078987,
		0.005236,
		0.029542,
		0.215973,
		0.473757,
		0.708972,
		0.862495,
		0.921848,
		0.928019,
		0.969283,
		0.938519,
		0.825756,
		0.653732,
		0.455958,
		0.265625,
		0.112019,
		0.019046,
		0.061965,
		0.274752,
		0.537774,
		0.767694,
		0.913908,
		0.969283,
		0.974941,
		0.988808,
		0.949475,
		0.842698,
		0.681729,
		0.492972,
		0.3057,
		0.146884,
		0.039085,
		0.100129,
		0.327031,
		0.585368,
		0.803188,
		0.938519,
		0.988808,
		0.993865,
		0.994648,
		0.955481,
		0.857183,
		0.707107,
		0.527402,
		0.344293,
		0.182603,
		0.063557,
		0.004182,
		0.00674,
		0.140868,
		0.373628,
		0.622395,
		0.825756,
		0.949475,
		0.994648,
		0.999123,
		0.995994,
		0.960083,
		0.870045,
		0.730146,
		0.559361,
		0.381194,
		0.218463,
		0.091084,
		0.015366,
		0.024222,
		0.182595,
		0.416269,
		0.653732,
		0.842698,
		0.955481,
		0.995994,
		0.999952,
		0.996489,
		0.964076,
		0.881522,
		0.751077,
		0.588982,
		0.416289,
		0.253947,
		0.120607,
		0.031834,
		0.049119,
		0.224414,
		0.455958,
		0.681729,
		0.857183,
		0.960083,
		0.996489,
		1,
		0.996876,
		0.967586,
		0.891784,
		0.770107,
		0.61641,
		0.449532,
		0.28869,
		0.151319,
		0.05223,
		0.003417,
		0.078987,
		0.265625,
		0.492972,
		0.707107,
		0.870045,
		0.964076,
		0.996876,
		1,
		0.99721,
		0.970681,
		0.900978,
		0.787425,
		0.641793,
		0.480925,
		0.322436,
		0.182603,
		0.075484,
		0.012657,
		0.005236,
		0.112019,
		0.3057,
		0.527402,
		0.730146,
		0.881522,
		0.967586,
		0.99721,
		1,
		0.997501,
		0.973418,
		0.909232,
		0.803203,
		0.665277,
		0.510503,
		0.355015,
		0.213992,
		0.100747,
		0.026425,
		0.019046,
		0.146884,
		0.344293,
		0.559361,
		0.751077,
		0.891784,
		0.970681,
		0.997501,
		1,
		0.997755,
		0.975845,
		0.916658,
		0.817592,
		0.687003,
		0.538323,
		0.38632,
		0.245133,
		0.127354,
		0.043674,
		0.002844,
		0.039085,
		0.182603,
		0.381194,
		0.588982,
		0.770107,
		0.900978,
		0.973418,
		0.997755,
		1,
		0.997978,
		0.978003,
		0.92335,
		0.830729,
		0.707107,
		0.564456,
		0.416289,
		0.275766,
		0.154777,
		0.063557,
		0.010606,
		0.063557,
		0.218463,
		0.416289,
		0.61641,
		0.787425,
		0.909232,
		0.975845,
		0.997978,
		1,
		0.998174,
		0.979927,
		0.929393,
		0.842738,
		0.725714,
		0.588982,
		0.444897,
		0.3057,
		0.182603,
		0.085389,
		0.022285,
		0.004182,
		0.091084,
		0.253947,
		0.449532,
		0.641793,
		0.803203,
		0.916658,
		0.978003,
		0.998174,
		1,
		0.998347,
		0.981647,
		0.93486,
		0.853729,
		0.742944,
		0.611985,
		0.472143,
		0.334797,
		0.210511,
		0.108615,
		0.037057,
		0.002404,
		0.015366,
		0.120607,
		0.28869,
		0.480925,
		0.665277,
		0.817592,
		0.92335,
		0.979927,
		0.998347,
		1,
		0.998499,
		0.983187,
		0.939815,
		0.863799,
		0.758906,
		0.63355,
		0.498046,
		0.362964,
		0.238249,
		0.132788,
		0.05424,
		0.009016,
		0.031834,
		0.151319,
		0.322436,
		0.510503,
		0.687003,
		0.830729,
		0.929393,
		0.981647,
		0.998499,
		1,
		0.998635,
		0.984569,
		0.944313,
		0.873037,
		0.773702,
		0.653763,
		0.522638,
		0.39014,
		0.265625,
		0.157547,
		0.073275,
		0.019046,
		0.05223,
		0.182603,
		0.355015,
		0.538323,
		0.707107,
		0.842738,
		0.93486,
		0.983187,
		0.998635,
		1,
		0.998756,
		0.985814,
		0.948403,
		0.881522,
		0.787425,
		0.672708,
		0.545959,
		0.416289,
		0.292492,
		0.182603,
		0.093699,
		0.031834,
		0.002059,
		0.003417,
		0.075484,
		0.213992,
		0.38632,
		0.564456,
		0.725714,
		0.853729,
		0.939815,
		0.984569,
		0.998756,
		1,
		0.998863,
		0.986936,
		0.952128,
		0.889324,
		0.800163,
		0.690463,
		0.568056,
		0.441396,
		0.318742,
		0.207724,
		0.11513,
		0.046826,
		0.007758,
		0.012657,
		0.100747,
		0.245133,
		0.416289,
		0.588982,
		0.742944,
		0.863799,
		0.944313,
		0.985814,
		0.998863,
		1,
		0.99896,
		0.987949,
		0.955526,
		0.896506,
		0.811993,
		0.707107,
		0.588982,
		0.465459,
		0.344293,
		0.232725,
		0.137255,
		0.063557,
		0.016464,
		0.026425,
		0.127354,
		0.275766,
		0.444897,
		0.611985,
		0.758906,
		0.873037,
		0.948403,
		0.986936,
		0.99896,
		1,
		0.999046,
		0.988867,
		0.958631,
		0.903125,
		0.82299,
		0.722712,
		0.608789,
		0.488488,
		0.369091,
		0.25746,
		0.159817,
		0.08164,
		0.027641,
		0.001783,
		0.043674,
		0.154777,
		0.3057,
		0.472143,
		0.63355,
		0.773702,
		0.881522,
		0.952128,
		0.987949,
		0.999046,
		1,
		0.999123,
		0.9897,
		0.961473,
		0.909232,
		0.833218,
		0.737347,
		0.62753,
		0.510503,
		0.393097,
		0.281814,
		0.182603,
		0.100747,
		0.04083,
		0.006746,
		0.002844,
		0.063557,
		0.182603,
		0.334797,
		0.498046,
		0.653763,
		0.787425,
		0.889324,
		0.955526,
		0.988867,
		0.999123,
		1,
		0.999193,
		0.990456,
		0.964076,
		0.914874,
		0.842738,
		0.751077,
		0.645261,
		0.531529,
		0.416289,
		0.3057,
		0.205443,
		0.120607,
		0.055645,
		0.014374,
		0.010606,
		0.085389,
		0.210511,
		0.362964,
		0.522638,
		0.672708,
		0.800163,
		0.896506,
		0.958631,
		0.9897,
		0.999193,
		1,
		0.999256,
		0.991144,
		0.966466,
		0.92009,
		0.851607,
		0.763962,
		0.662033,
		0.551596,
		0.438658,
		0.32905,
		0.228195,
		0.140991,
		0.071756,
		0.024223,
		0.001559,
		0.022285,
		0.108615,
		0.238249,
		0.39014,
		0.545959,
		0.690463,
		0.811993,
		0.903125,
		0.961473,
		0.990456,
		0.999256,
		1,
		0.999313,
		0.991772,
		0.968661,
		0.924919,
		0.859876,
		0.776061,
		0.677898,
		0.570735,
		0.460201,
		0.351814,
		0.250747,
		0.16171,
		0.088884,
		0.035914,
		0.00592,
		0.037057,
		0.132788,
		0.265625,
		0.416289,
		0.568056,
		0.707107,
		0.82299,
		0.909232,
		0.964076,
		0.991144,
		0.999313,
		1,
		0.999365,
		0.992346,
		0.970681,
		0.929393,
		0.86759,
		0.787425,
		0.692907,
		0.588982,
		0.480925,
		0.373956,
		0.273008,
		0.182603,
		0.106789,
		0.049119,
		0.012657,
		0.002404,
		0.05424,
		0.157547,
		0.292492,
		0.441396,
		0.588982,
		0.722712,
		0.833218,
		0.914874,
		0.966466,
		0.991772,
		0.999365,
		1,
		0.999412,
		0.99287,
		0.972543,
		0.933544,
		0.874792,
		0.798105,
		0.707107,
		0.606372,
		0.500843,
		0.395453,
		0.294906,
		0.203541,
		0.125272,
		0.063557,
		0.021402,
		0.001374,
		0.009016,
		0.073275,
		0.182603,
		0.318742,
		0.465459,
		0.608789,
		0.737347,
		0.842738,
		0.92009,
		0.968661,
		0.992346,
		0.999412,
		1,
		0.999454,
		0.99335,
		0.97426,
		0.937398,
		0.881522,
		0.808146,
		0.720543,
		0.622942,
		0.519969,
		0.416289,
		0.316383,
		0.224414,
		0.144161,
		0.078987,
		0.031834,
		0.005236,
		0.019046,
		0.093699,
		0.207724,
		0.344293,
		0.488488,
		0.62753,
		0.751077,
		0.851607,
		0.924919,
		0.970681,
		0.99287,
		0.999454,
		1,
		0.999493,
		0.993791,
		0.975845,
		0.94098,
		0.887815,
		0.817592,
		0.73326,
		0.638727,
		0.538323,
		0.436458,
		0.337397,
		0.245133,
		0.163312,
		0.0952,
		0.043674,
		0.011231,
		0.031834,
		0.11513,
		0.232725,
		0.369091,
		0.510503,
		0.645261,
		0.763962,
		0.859876,
		0.929393,
		0.972543,
		0.99335,
		0.999493,
		1,
		0.999529,
		0.994196,
		0.977312,
		0.944313,
		0.893703,
		0.826481,
		0.745299,
		0.653763,
		0.555928,
		0.455958,
		0.357915,
		0.265625,
		0.182603,
		0.112019,
		0.056682,
		0.019046,
		0.001221,
		0.002059,
		0.046826,
		0.137255,
		0.25746,
		0.393097,
		0.531529,
		0.662033,
		0.776061,
		0.86759,
		0.933544,
		0.97426,
		0.993791,
		0.999529,
		1,
		0.999561,
		0.994569,
		0.978669,
		0.947416,
		0.899218,
		0.834852,
		0.756699,
		0.668086,
		0.572806,
		0.474793,
		0.377913,
		0.28583,
		0.201932,
		0.129291,
		0.070648,
		0.02841,
		0.004665,
		0.007758,
		0.063557,
		0.159817,
		0.281814,
		0.416289,
		0.551596,
		0.677898,
		0.787425,
		0.874792,
		0.937398,
		0.975845,
		0.994196,
		0.999561,
		1,
		0.999591,
		0.994912,
		0.979927,
		0.950308,
		0.904386,
		0.842738,
		0.767497,
		0.681729,
		0.588982,
		0.492972,
		0.397375,
		0.3057,
		0.221211,
		0.146884,
		0.085389,
		0.039085,
		0.010032,
		0.016464,
		0.08164,
		0.182603,
		0.3057,
		0.438658,
		0.570735,
		0.692907,
		0.798105,
		0.881522,
		0.94098,
		0.977312,
		0.994569,
		0.999591,
		1,
		0.999618,
		0.995229,
		0.981094,
		0.953007,
		0.909232,
		0.850172,
		0.777728,
		0.694725,
		0.604481,
		0.510503,
		0.416289,
		0.325197,
		0.24037,
		0.164687,
		0.100747,
		0.050861,
		0.017057,
		0.001092,
		0.027641,
		0.100747,
		0.205443,
		0.32905,
		0.460201,
		0.588982,
		0.707107,
		0.808146,
		0.887815,
		0.944313,
		0.978669,
		0.994912,
		0.999618,
		1,
		0.999643,
		0.995521,
		0.982179,
		0.955526,
		0.913781,
		0.857183,
		0.787425,
		0.707107,
		0.619329,
		0.527402,
		0.434652,
		0.344293,
		0.259348,
		0.182602,
		0.116585,
		0.063554,
		0.025507,
		0.004181,
		0.001783,
		0.04083,
		0.120607,
		0.228195,
		0.351814,
		0.480925,
		0.606372,
		0.720543,
		0.817592,
		0.893703,
		0.947416,
		0.979927,
		0.995229,
		0.999643,
		1,
		0.999666,
		0.995792,
		0.983187,
		0.957881,
		0.918052,
		0.863799,
		0.796619,
		0.718904,
		0.63355,
		0.543682,
		0.452462,
		0.362964,
		0.278094,
		0.200548,
		0.132781,
		0.077003,
		0.035173,
		0.009011,
		0.006746,
		0.055645,
		0.140991,
		0.250747,
		0.373956,
		0.500843,
		0.622942,
		0.73326,
		0.826481,
		0.899218,
		0.950308,
		0.981094,
		0.995521,
		0.999666,
		1,
		0.999687,
		0.996042,
		0.984125,
		0.960083,
		0.922066,
		0.870045,
		0.80534,
		0.730146,
		0.647171,
		0.559361,
		0.469721,
		0.381192,
		0.296568,
		0.218452,
		0.149231,
		0.091062,
		0.04587,
		0.015352,
		0.000981,
		0.014374,
		0.071756,
		0.16171,
		0.273008,
		0.395453,
		0.519969,
		0.638727,
		0.745299,
		0.834852,
		0.904386,
		0.953007,
		0.982179,
		0.995792,
		0.999687,
		1,
		0.999707,
		0.996274,
		0.984999,
		0.962145,
		0.925841,
		0.875946,
		0.813614,
		0.740862,
		0.660215,
		0.574454,
		0.486434,
		0.398965,
		0.314732,
		0.236251,
		0.165839,
		0.105601,
		0.057429,
		0.022996,
		0.003761,
		0.024223,
		0.088884,
		0.182603,
		0.294906,
		0.416289,
		0.538323,
		0.653763,
		0.756699,
		0.842738,
		0.909232,
		0.955526,
		0.983187,
		0.996042,
		0.999707,
		1,
		0.999725,
		0.99649,
		0.985814,
		0.964076,
		0.929393,
		0.881522,
		0.821467,
		0.751076,
		0.672707,
		0.588979,
		0.502605,
		0.416269,
		0.332553,
		0.253886,
		0.182516,
		0.120501,
		0.069697,
		0.031753,
		0.008113,
		0.001559,
		0.035914,
		0.106789,
		0.203541,
		0.316383,
		0.436458,
		0.555928,
		0.668086,
		0.767497,
		0.850172,
		0.913781,
		0.957881,
		0.984125,
		0.996274,
		0.999725,
		1,
		0.999741,
		0.99669,
		0.986574,
		0.965887,
		0.932739,
		0.886795,
		0.828924,
		0.760816,
		0.684668,
		0.60295,
		0.518237,
		0.433094,
		0.349999,
		0.2713,
		0.199181,
		0.13565,
		0.082529,
		0.041447,
		0.013828,
		0.00088,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.99649,
		0.999741,
		1,
		0.999757,
		0.996876,
		0.987285,
		0.967586,
		0.93589,
		0.891783,
		0.836006,
		0.770102,
		0.696118,
		0.616381,
		0.533332,
		0.449423,
		0.367036,
		0.288436,
		0.215749,
		0.150937,
		0.095789,
		0.05191,
		0.020707,
		0.003372,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999771,
		0.997049,
		0.987949,
		0.969182,
		0.938862,
		0.896504,
		0.842734,
		0.778955,
		0.707073,
		0.629277,
		0.547885,
		0.465237,
		0.383622,
		0.305234,
		0.232137,
		0.166253,
		0.109339,
		0.06298,
		0.028563,
		0.007261,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999784,
		0.99721,
		0.988571,
		0.970681,
		0.941665,
		0.900973,
		0.849122,
		0.787387,
		0.71754,
		0.641638,
		0.56188,
		0.480503,
		0.39971,
		0.321622,
		0.248253,
		0.181483,
		0.123044,
		0.074497,
		0.037212,
		0.01234,
		0.00078,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999797,
		0.997361,
		0.989153,
		0.97209,
		0.944307,
		0.905199,
		0.855181,
		0.795406,
		0.727518,
		0.653452,
		0.575293,
		0.495182,
		0.415238,
		0.337521,
		0.263995,
		0.196507,
		0.136762,
		0.0863,
		0.046472,
		0.018407,
		0.002974,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999808,
		0.997501,
		0.989698,
		0.973413,
		0.946796,
		0.909189,
		0.860913,
		0.803008,
		0.736995,
		0.664693,
		0.588084,
		0.509214,
		0.430131,
		0.352837,
		0.279254,
		0.211195,
		0.150344,
		0.098223,
		0.056161,
		0.025261,
		0.006363,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325197,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999818,
		0.997631,
		0.990206,
		0.97465,
		0.94913,
		0.912939,
		0.86631,
		0.810175,
		0.745942,
		0.67532,
		0.600193,
		0.522525,
		0.444298,
		0.367463,
		0.293903,
		0.225407,
		0.163637,
		0.110096,
		0.066092,
		0.032703,
		0.010731,
		0.000671,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751076,
		0.828924,
		0.891783,
		0.938862,
		0.970681,
		0.989153,
		0.997501,
		0.999818,
		1,
		0.999827,
		0.99775,
		0.990675,
		0.975798,
		0.951302,
		0.916436,
		0.87135,
		0.816874,
		0.75431,
		0.685268,
		0.611541,
		0.535021,
		0.457628,
		0.381271,
		0.307804,
		0.238991,
		0.176476,
		0.121743,
		0.076079,
		0.040531,
		0.015865,
		0.00253,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469721,
		0.574454,
		0.672707,
		0.760816,
		0.836006,
		0.896504,
		0.941665,
		0.97209,
		0.989698,
		0.997631,
		0.999827,
		0.999999,
		0.999832,
		0.997853,
		0.991097,
		0.976845,
		0.953294,
		0.919651,
		0.875992,
		0.823051,
		0.762033,
		0.694454,
		0.622029,
		0.546584,
		0.469991,
		0.394122,
		0.320804,
		0.251784,
		0.188689,
		0.132984,
		0.085931,
		0.048543,
		0.021549,
		0.00535,
		0.004665,
		0.039085,
		0.100747,
		0.182602,
		0.278094,
		0.381192,
		0.486434,
		0.588979,
		0.684668,
		0.770102,
		0.842734,
		0.900973,
		0.944307,
		0.973413,
		0.990206,
		0.99775,
		0.999832,
		0.999994,
		0.999829,
		0.99793,
		0.991457,
		0.977766,
		0.955071,
		0.92254,
		0.880177,
		0.828631,
		0.769016,
		0.702768,
		0.631533,
		0.557079,
		0.48124,
		0.405858,
		0.33274,
		0.263617,
		0.200101,
		0.143636,
		0.095456,
		0.05654,
		0.027572,
		0.008906,
		0.000547,
		0.010032,
		0.050861,
		0.116585,
		0.200548,
		0.296568,
		0.398965,
		0.502605,
		0.60295,
		0.696118,
		0.778955,
		0.849122,
		0.905199,
		0.946796,
		0.97465,
		0.990675,
		0.997853,
		0.999829,
		0.999982,
		0.999806,
		0.997966,
		0.991729,
		0.978526,
		0.956583,
		0.925035,
		0.883821,
		0.833512,
		0.775143,
		0.710079,
		0.639906,
		0.56635,
		0.491211,
		0.416309,
		0.343437,
		0.274314,
		0.210532,
		0.153515,
		0.104466,
		0.064324,
		0.033727,
		0.012978,
		0.002032,
		0.017057,
		0.063554,
		0.132781,
		0.218452,
		0.314732,
		0.416269,
		0.518237,
		0.616381,
		0.707073,
		0.787387,
		0.855181,
		0.909189,
		0.94913,
		0.975798,
		0.991097,
		0.99793,
		0.999806,
		0.999952,
		0.999749,
		0.997932,
		0.991872,
		0.979067,
		0.957757,
		0.927046,
		0.886816,
		0.837568,
		0.780271,
		0.71623,
		0.646984,
		0.574223,
		0.499726,
		0.425297,
		0.352718,
		0.283695,
		0.219804,
		0.162442,
		0.112777,
		0.071706,
		0.039813,
		0.01735,
		0.004224,
		0.001092,
		0.025507,
		0.077003,
		0.149231,
		0.236251,
		0.332553,
		0.433094,
		0.533332,
		0.629277,
		0.71754,
		0.795406,
		0.860913,
		0.912939,
		0.951302,
		0.976845,
		0.991457,
		0.997966,
		0.999749,
		0.999887,
		0.999627,
		0.997787,
		0.991829,
		0.979315,
		0.958497,
		0.928457,
		0.889025,
		0.840647,
		0.784233,
		0.721044,
		0.652581,
		0.580512,
		0.5066,
		0.43264,
		0.360406,
		0.291589,
		0.227748,
		0.170248,
		0.120219,
		0.078506,
		0.045641,
		0.021821,
		0.006903,
		0.000415,
		0.004181,
		0.035173,
		0.091062,
		0.165839,
		0.253886,
		0.349999,
		0.449423,
		0.547885,
		0.641638,
		0.727518,
		0.803008,
		0.86631,
		0.916436,
		0.953294,
		0.977766,
		0.991729,
		0.997932,
		0.999627,
		0.999758,
		0.999399,
		0.997471,
		0.99152,
		0.979167,
		0.958682,
		0.929125,
		0.890289,
		0.842572,
		0.786843,
		0.724327,
		0.656502,
		0.585024,
		0.511645,
		0.438158,
		0.366327,
		0.297832,
		0.234205,
		0.176778,
		0.126633,
		0.084562,
		0.051039,
		0.026203,
		0.009861,
		0.001509,
		0.009011,
		0.04587,
		0.105601,
		0.182516,
		0.2713,
		0.367036,
		0.465237,
		0.56188,
		0.653452,
		0.736995,
		0.810175,
		0.87135,
		0.919651,
		0.955071,
		0.978526,
		0.991872,
		0.997787,
		0.999399,
		0.999523,
		0.999005,
		0.996902,
		0.990841,
		0.978498,
		0.958161,
		0.928881,
		0.890421,
		0.843145,
		0.787895,
		0.725873,
		0.658546,
		0.587564,
		0.514679,
		0.441681,
		0.370327,
		0.302277,
		0.239036,
		0.181895,
		0.131883,
		0.089732,
		0.055852,
		0.030325,
		0.012909,
		0.003067,
		0.015352,
		0.057429,
		0.120501,
		0.199181,
		0.288436,
		0.383622,
		0.480503,
		0.575293,
		0.664693,
		0.745942,
		0.816874,
		0.875992,
		0.92254,
		0.956583,
		0.979067,
		0.991829,
		0.997471,
		0.999005,
		0.999123,
		0.998363,
		0.995977,
		0.989664,
		0.977154,
		0.956762,
		0.927533,
		0.889215,
		0.842154,
		0.787175,
		0.725472,
		0.658512,
		0.587947,
		0.515532,
		0.443057,
		0.372267,
		0.304801,
		0.242128,
		0.18549,
		0.135857,
		0.093895,
		0.059949,
		0.03404,
		0.015878,
		0.004895,
		0.000286,
		0.000981,
		0.022996,
		0.069697,
		0.13565,
		0.215749,
		0.305234,
		0.39971,
		0.495182,
		0.588084,
		0.67532,
		0.75431,
		0.823051,
		0.880177,
		0.925035,
		0.957757,
		0.979315,
		0.99152,
		0.996902,
		0.998363,
		0.998475,
		0.997368,
		0.994565,
		0.987832,
		0.974957,
		0.954283,
		0.924865,
		0.886448,
		0.839373,
		0.784462,
		0.722916,
		0.656209,
		0.586001,
		0.514056,
		0.442157,
		0.372038,
		0.30531,
		0.243396,
		0.187482,
		0.138472,
		0.096962,
		0.063227,
		0.037225,
		0.018622,
		0.006822,
		0.001015,
		0.003761,
		0.031753,
		0.082529,
		0.150937,
		0.232137,
		0.321622,
		0.415238,
		0.509214,
		0.600193,
		0.685268,
		0.762033,
		0.828631,
		0.883821,
		0.927046,
		0.958497,
		0.979167,
		0.990841,
		0.995977,
		0.997368,
		0.997474,
		0.995889,
		0.992508,
		0.985163,
		0.971702,
		0.950506,
		0.92065,
		0.881889,
		0.834575,
		0.779544,
		0.71801,
		0.651464,
		0.581581,
		0.510129,
		0.438887,
		0.369568,
		0.303745,
		0.242791,
		0.187825,
		0.13968,
		0.098873,
		0.065609,
		0.039786,
		0.021023,
		0.008699,
		0.002007,
		0.008113,
		0.041447,
		0.095789,
		0.166253,
		0.248253,
		0.337521,
		0.430131,
		0.522525,
		0.611541,
		0.694454,
		0.769016,
		0.833512,
		0.886816,
		0.928457,
		0.958682,
		0.978498,
		0.989664,
		0.994565,
		0.995889,
		0.99599,
		0.993768,
		0.989623,
		0.981452,
		0.967166,
		0.945195,
		0.914647,
		0.875305,
		0.827542,
		0.772221,
		0.71058,
		0.644135,
		0.574574,
		0.50367,
		0.433192,
		0.364821,
		0.300088,
		0.240301,
		0.186509,
		0.139462,
		0.0996,
		0.067052,
		0.041656,
		0.022989,
		0.010409,
		0.003111,
		0.000176,
		0.013828,
		0.05191,
		0.109339,
		0.181483,
		0.263995,
		0.352837,
		0.444298,
		0.535021,
		0.622029,
		0.702768,
		0.775143,
		0.837568,
		0.889025,
		0.929125,
		0.958161,
		0.977154,
		0.987832,
		0.992508,
		0.993768,
		0.993865,
		0.990823,
		0.985703,
		0.976471,
		0.961111,
		0.938108,
		0.90662,
		0.86647,
		0.818069,
		0.762317,
		0.700485,
		0.634115,
		0.564909,
		0.494641,
		0.425058,
		0.357808,
		0.294361,
		0.235957,
		0.183562,
		0.13784,
		0.099147,
		0.067541,
		0.042798,
		0.024457,
		0.011859,
		0.004205,
		0.000604,
		0.00088,
		0.020707,
		0.06298,
		0.123044,
		0.196507,
		0.279254,
		0.367463,
		0.457628,
		0.546584,
		0.631533,
		0.710079,
		0.780271,
		0.840647,
		0.890289,
		0.928881,
		0.956762,
		0.974957,
		0.985163,
		0.989623,
		0.990823,
		0.990914,
		0.986844,
		0.98052,
		0.969981,
		0.953289,
		0.929002,
		0.896339,
		0.855177,
		0.80598,
		0.749691,
		0.687623,
		0.621342,
		0.552562,
		0.48305,
		0.414524,
		0.348584,
		0.286632,
		0.229828,
		0.179047,
		0.134865,
		0.097552,
		0.067091,
		0.043203,
		0.025391,
		0.012984,
		0.005191,
		0.001155,
		0.003372,
		0.028563,
		0.074497,
		0.136762,
		0.211195,
		0.293903,
		0.381271,
		0.469991,
		0.557079,
		0.639906,
		0.71623,
		0.784233,
		0.842572,
		0.890421,
		0.927533,
		0.954283,
		0.971702,
		0.981452,
		0.985703,
		0.986844,
		0.986931,
		0.981605,
		0.973833,
		0.961733,
		0.943456,
		0.917645,
		0.883593,
		0.841247,
		0.791133,
		0.734247,
		0.67194,
		0.605806,
		0.537563,
		0.468959,
		0.401677,
		0.337251,
		0.27701,
		0.222021,
		0.173066,
		0.130624,
		0.09488,
		0.065746,
		0.04289,
		0.025783,
		0.013746,
		0.006001,
		0.001726,
		9.4e-005,
		0.007261,
		0.037212,
		0.0863,
		0.150344,
		0.225407,
		0.307804,
		0.394122,
		0.48124,
		0.56635,
		0.646984,
		0.721044,
		0.786843,
		0.843145,
		0.889215,
		0.924865,
		0.950506,
		0.967166,
		0.976471,
		0.98052,
		0.981605,
		0.981688,
		0.974863,
		0.965392,
		0.951479,
		0.931376,
		0.903823,
		0.868202,
		0.824541,
		0.773435,
		0.715937,
		0.653439,
		0.587553,
		0.519995,
		0.452485,
		0.386651,
		0.323959,
		0.265647,
		0.212684,
		0.16575,
		0.12523,
		0.091224,
		0.063573,
		0.041898,
		0.025644,
		0.014128,
		0.006591,
		0.002242,
		0.000308,
		0.01234,
		0.046472,
		0.098223,
		0.163637,
		0.238991,
		0.320804,
		0.405858,
		0.491211,
		0.574223,
		0.652581,
		0.724327,
		0.787895,
		0.842154,
		0.886448,
		0.92065,
		0.945195,
		0.961111,
		0.969981,
		0.973833,
		0.974863,
		0.974941,
		0.966369,
		0.95495,
		0.938984,
		0.916836,
		0.887355,
		0.850026,
		0.804966,
		0.752845,
		0.694774,
		0.632181,
		0.566689,
		0.5,
		0.433793,
		0.369632,
		0.308897,
		0.252728,
		0.20199,
		0.15726,
		0.118823,
		0.086698,
		0.060661,
		0.040291,
		0.025011,
		0.014139,
		0.006937,
		0.002653,
		0.000562,
		0.00078,
		0.018407,
		0.056161,
		0.110096,
		0.176476,
		0.251784,
		0.33274,
		0.416309,
		0.499726,
		0.580512,
		0.656502,
		0.725873,
		0.787175,
		0.839373,
		0.881889,
		0.914647,
		0.938108,
		0.953289,
		0.961733,
		0.965392,
		0.966369,
		0.966443,
		0.955875,
		0.942269,
		0.924032,
		0.899653,
		0.868102,
		0.828975,
		0.782486,
		0.729382,
		0.67083,
		0.608287,
		0.543376,
		0.477773,
		0.4131,
		0.350845,
		0.292291,
		0.238473,
		0.190145,
		0.147776,
		0.11156,
		0.081434,
		0.057115,
		0.038146,
		0.023933,
		0.013802,
		0.007038,
		0.002929,
		0.000799,
		4.1e-005,
		0.002974,
		0.025261,
		0.066092,
		0.121743,
		0.188689,
		0.263617,
		0.343437,
		0.425297,
		0.5066,
		0.585024,
		0.658546,
		0.725472,
		0.784462,
		0.834575,
		0.875305,
		0.90662,
		0.929002,
		0.943456,
		0.951479,
		0.95495,
		0.955875,
		0.955945,
		0.943143,
		0.927134,
		0.906441,
		0.879686,
		0.845973,
		0.805015,
		0.757127,
		0.703131,
		0.644244,
		0.581942,
		0.517835,
		0.453558,
		0.390664,
		0.330552,
		0.274398,
		0.223122,
		0.177368,
		0.137498,
		0.103613,
		0.075577,
		0.053053,
		0.035553,
		0.022475,
		0.013156,
		0.006907,
		0.003057,
		0.000982,
		0.000127,
		0.006363,
		0.032703,
		0.076079,
		0.132984,
		0.200101,
		0.274314,
		0.352718,
		0.43264,
		0.511645,
		0.587564,
		0.658512,
		0.722916,
		0.779544,
		0.827542,
		0.86647,
		0.896339,
		0.917645,
		0.931376,
		0.938984,
		0.942269,
		0.943143,
		0.943208,
		0.927957,
		0.909361,
		0.886069,
		0.856847,
		0.820937,
		0.778177,
		0.72898,
		0.674241,
		0.615215,
		0.553385,
		0.490336,
		0.427644,
		0.366781,
		0.309044,
		0.255494,
		0.206935,
		0.163895,
		0.126633,
		0.095162,
		0.069278,
		0.0486,
		0.032612,
		0.020712,
		0.01225,
		0.00657,
		0.003043,
		0.001091,
		0.000215,
		0.010731,
		0.040531,
		0.085931,
		0.143636,
		0.210532,
		0.283695,
		0.360406,
		0.438158,
		0.514679,
		0.587947,
		0.656209,
		0.71801,
		0.772221,
		0.818069,
		0.855177,
		0.883593,
		0.903823,
		0.916836,
		0.924032,
		0.927134,
		0.927957,
		0.928019,
		0.910135,
		0.88881,
		0.862827,
		0.831105,
		0.793028,
		0.748558,
		0.698204,
		0.642924,
		0.584001,
		0.522909,
		0.461192,
		0.400354,
		0.341773,
		0.286631,
		0.235873,
		0.190179,
		0.149966,
		0.115392,
		0.08639,
		0.062693,
		0.043882,
		0.029425,
		0.018722,
		0.011143,
		0.006065,
		0.0029,
		0.001123,
		0.000283,
		1.3e-005,
		0.000671,
		0.015865,
		0.048543,
		0.095456,
		0.153515,
		0.219804,
		0.291589,
		0.366327,
		0.441681,
		0.515532,
		0.586001,
		0.651464,
		0.71058,
		0.762317,
		0.80598,
		0.841247,
		0.868202,
		0.887355,
		0.899653,
		0.906441,
		0.909361,
		0.910135,
		0.910193,
		0.889535,
		0.865392,
		0.836688,
		0.802496,
		0.762347,
		0.716324,
		0.665023,
		0.609454,
		0.550912,
		0.49085,
		0.430753,
		0.372039,
		0.315979,
		0.263636,
		0.21583,
		0.173123,
		0.135819,
		0.103984,
		0.077474,
		0.055972,
		0.039027,
		0.026096,
		0.016587,
		0.009895,
		0.005431,
		0.002653,
		0.001083,
		0.000317,
		3.7e-005,
		0.00253,
		0.021549,
		0.05654,
		0.104466,
		0.162442,
		0.227748,
		0.297832,
		0.370327,
		0.443057,
		0.514056,
		0.581581,
		0.644135,
		0.700485,
		0.749691,
		0.791133,
		0.824541,
		0.850026,
		0.868102,
		0.879686,
		0.886069,
		0.88881,
		0.889535,
		0.88959,
		0.866069,
		0.839077,
		0.807686,
		0.771123,
		0.729065,
		0.681705,
		0.629719,
		0.574154,
		0.516304,
		0.457578,
		0.399391,
		0.343062,
		0.289746,
		0.240381,
		0.195661,
		0.156029,
		0.121685,
		0.092608,
		0.068588,
		0.049261,
		0.034155,
		0.022723,
		0.014386,
		0.008567,
		0.004715,
		0.002331,
		0.000983,
		0.000316,
		5.5e-005,
		0.00535,
		0.027572,
		0.064324,
		0.112777,
		0.170248,
		0.234205,
		0.302277,
		0.372267,
		0.442157,
		0.510129,
		0.574574,
		0.634115,
		0.687623,
		0.734247,
		0.773435,
		0.804966,
		0.828975,
		0.845973,
		0.856847,
		0.862827,
		0.865392,
		0.866069,
		0.86612,
		0.839707,
		0.809903,
		0.775926,
		0.73716,
		0.693416,
		0.644993,
		0.59263,
		0.537395,
		0.480563,
		0.423487,
		0.367494,
		0.313795,
		0.26342,
		0.217181,
		0.175647,
		0.139147,
		0.107782,
		0.081453,
		0.05989,
		0.042695,
		0.029379,
		0.019399,
		0.012195,
		0.007219,
		0.00396,
		0.001962,
		0.000842,
		0.000287,
		6.2e-005,
		0.008906,
		0.033727,
		0.071706,
		0.120219,
		0.176778,
		0.239036,
		0.304801,
		0.372038,
		0.438887,
		0.50367,
		0.564909,
		0.621342,
		0.67194,
		0.715937,
		0.752845,
		0.782486,
		0.805015,
		0.820937,
		0.831105,
		0.836688,
		0.839077,
		0.839707,
		0.839755,
		0.810487,
		0.777975,
		0.741583,
		0.700845,
		0.655698,
		0.606532,
		0.554134,
		0.499578,
		0.444099,
		0.388981,
		0.33545,
		0.284601,
		0.237336,
		0.194338,
		0.156055,
		0.122709,
		0.094309,
		0.070687,
		0.051523,
		0.036394,
		0.0248,
		0.016209,
		0.010082,
		0.005905,
		0.003208,
		0.001579,
		0.000679,
		0.000237,
		5.7e-005,
		0.000547,
		0.012978,
		0.039813,
		0.078506,
		0.126633,
		0.181895,
		0.242128,
		0.30531,
		0.369568,
		0.433192,
		0.494641,
		0.552562,
		0.605806,
		0.653439,
		0.694774,
		0.729382,
		0.757127,
		0.778177,
		0.793028,
		0.802496,
		0.807686,
		0.809903,
		0.810487,
		0.810531,
		0.778514,
		0.743467,
		0.704898,
		0.662478,
		0.616261,
		0.566711,
		0.514646,
		0.461124,
		0.407332,
		0.354464,
		0.303641,
		0.25583,
		0.211808,
		0.172129,
		0.137128,
		0.106923,
		0.081444,
		0.06046,
		0.043615,
		0.030463,
		0.020506,
		0.013224,
		0.008105,
		0.004672,
		0.002495,
		0.001208,
		0.000512,
		0.000179,
		4.5e-005,
		0.002032,
		0.01735,
		0.045641,
		0.084562,
		0.131883,
		0.18549,
		0.243396,
		0.303745,
		0.364821,
		0.425058,
		0.48305,
		0.537563,
		0.587553,
		0.632181,
		0.67083,
		0.703131,
		0.72898,
		0.748558,
		0.762347,
		0.771123,
		0.775926,
		0.777975,
		0.778514,
		0.778554,
		0.743962,
		0.706621,
		0.666171,
		0.622414,
		0.575499,
		0.52595,
		0.474596,
		0.422467,
		0.370678,
		0.320332,
		0.272428,
		0.22781,
		0.187125,
		0.150806,
		0.119079,
		0.09197,
		0.069338,
		0.0509,
		0.036269,
		0.02499,
		0.016567,
		0.010502,
		0.006313,
		0.003559,
		0.001853,
		0.000871,
		0.000357,
		0.000121,
		3e-005,
		0.004224,
		0.021821,
		0.051039,
		0.089732,
		0.135857,
		0.187482,
		0.242791,
		0.300088,
		0.357808,
		0.414524,
		0.468959,
		0.519995,
		0.566689,
		0.608287,
		0.644244,
		0.674241,
		0.698204,
		0.716324,
		0.729065,
		0.73716,
		0.741583,
		0.743467,
		0.743962,
		0.743999,
		0.707073,
		0.66774,
		0.62576,
		0.581049,
		0.533837,
		0.484687,
		0.434424,
		0.384031,
		0.334543,
		0.286955,
		0.242148,
		0.200837,
		0.163544,
		0.130588,
		0.102092,
		0.078003,
		0.058116,
		0.042108,
		0.029569,
		0.02004,
		0.013038,
		0.008087,
		0.004739,
		0.002592,
		0.001301,
		0.000585,
		0.000227,
		7.2e-005,
		1.6e-005,
		0.006903,
		0.026203,
		0.055852,
		0.093895,
		0.138472,
		0.187825,
		0.240301,
		0.294361,
		0.348584,
		0.401677,
		0.452485,
		0.5,
		0.543376,
		0.581942,
		0.615215,
		0.642924,
		0.665023,
		0.681705,
		0.693416,
		0.700845,
		0.704898,
		0.706621,
		0.707073,
		0.707107,
		0.66815,
		0.627179,
		0.584062,
		0.53881,
		0.491718,
		0.443369,
		0.394565,
		0.346229,
		0.299306,
		0.254678,
		0.213104,
		0.175172,
		0.141286,
		0.111657,
		0.086318,
		0.065143,
		0.047876,
		0.034161,
		0.023575,
		0.015662,
		0.009956,
		0.006008,
		0.003407,
		0.00179,
		0.000854,
		0.000359,
		0.000127,
		3.5e-005,
		0.000415,
		0.009861,
		0.030325,
		0.059949,
		0.096962,
		0.13968,
		0.186509,
		0.235957,
		0.286632,
		0.337251,
		0.386651,
		0.433793,
		0.477773,
		0.517835,
		0.553385,
		0.584001,
		0.609454,
		0.629719,
		0.644993,
		0.655698,
		0.662478,
		0.666171,
		0.66774,
		0.66815,
		0.668181,
		0.62755,
		0.585338,
		0.541506,
		0.496144,
		0.449592,
		0.402434,
		0.355436,
		0.309448,
		0.265316,
		0.223808,
		0.185558,
		0.151036,
		0.120532,
		0.094161,
		0.071873,
		0.05348,
		0.038685,
		0.027109,
		0.018324,
		0.011881,
		0.007338,
		0.004277,
		0.002323,
		0.001156,
		0.000513,
		0.000196,
		6e-005,
		1.3e-005,
		0.001509,
		0.012909,
		0.03404,
		0.063227,
		0.098873,
		0.139462,
		0.183562,
		0.229828,
		0.27701,
		0.323959,
		0.369632,
		0.4131,
		0.453558,
		0.490336,
		0.522909,
		0.550912,
		0.574154,
		0.59263,
		0.606532,
		0.616261,
		0.622414,
		0.62576,
		0.627179,
		0.62755,
		0.627578,
		0.585671,
		0.542647,
		0.498541,
		0.453504,
		0.4079,
		0.362304,
		0.317427,
		0.274039,
		0.232882,
		0.194608,
		0.159732,
		0.128609,
		0.101426,
		0.078208,
		0.058836,
		0.043071,
		0.030582,
		0.020976,
		0.013828,
		0.008703,
		0.005185,
		0.00289,
		0.001483,
		0.000685,
		0.000274,
		9e-005,
		2.2e-005,
		0.003067,
		0.015878,
		0.037225,
		0.065609,
		0.0996,
		0.13784,
		0.179047,
		0.222021,
		0.265647,
		0.308897,
		0.350845,
		0.390664,
		0.427644,
		0.461192,
		0.49085,
		0.516304,
		0.537395,
		0.554134,
		0.566711,
		0.575499,
		0.581049,
		0.584062,
		0.585338,
		0.585671,
		0.585696,
		0.542944,
		0.499553,
		0.455619,
		0.411333,
		0.367068,
		0.323371,
		0.280893,
		0.240313,
		0.202269,
		0.167298,
		0.135801,
		0.108027,
		0.084068,
		0.06387,
		0.047255,
		0.033943,
		0.023578,
		0.015763,
		0.010079,
		0.006113,
		0.003479,
		0.00183,
		0.00087,
		0.000361,
		0.000124,
		3.2e-005,
		0.004895,
		0.018622,
		0.039786,
		0.067052,
		0.099147,
		0.134865,
		0.173066,
		0.212684,
		0.252728,
		0.292291,
		0.330552,
		0.366781,
		0.400354,
		0.430753,
		0.457578,
		0.480563,
		0.499578,
		0.514646,
		0.52595,
		0.533837,
		0.53881,
		0.541506,
		0.542647,
		0.542944,
		0.542966,
		0.499817,
		0.456511,
		0.413186,
		0.370055,
		0.327489,
		0.285991,
		0.246144,
		0.208535,
		0.173695,
		0.142051,
		0.113897,
		0.089384,
		0.068519,
		0.051181,
		0.037142,
		0.02609,
		0.017657,
		0.011443,
		0.007047,
		0.00408,
		0.002188,
		0.001064,
		0.000454,
		0.000161,
		4.3e-005,
		0.000286,
		0.006822,
		0.021023,
		0.041656,
		0.067541,
		0.097552,
		0.130624,
		0.16575,
		0.20199,
		0.238473,
		0.274398,
		0.309044,
		0.341773,
		0.372039,
		0.399391,
		0.423487,
		0.444099,
		0.461124,
		0.474596,
		0.484687,
		0.491718,
		0.496144,
		0.498541,
		0.499553,
		0.499817,
		0.499836,
		0.456743,
		0.413966,
		0.371666,
		0.330067,
		0.289518,
		0.250476,
		0.213445,
		0.17892,
		0.147328,
		0.118991,
		0.094104,
		0.072729,
		0.0548,
		0.040139,
		0.028478,
		0.019481,
		0.012776,
		0.00797,
		0.004682,
		0.002552,
		0.001265,
		0.000552,
		0.000201,
		5.6e-005,
		0.001015,
		0.008699,
		0.022989,
		0.042798,
		0.067091,
		0.09488,
		0.12523,
		0.15726,
		0.190145,
		0.223122,
		0.255494,
		0.286631,
		0.315979,
		0.343062,
		0.367494,
		0.388981,
		0.407332,
		0.422467,
		0.434424,
		0.443369,
		0.449592,
		0.453504,
		0.455619,
		0.456511,
		0.456743,
		0.456761,
		0.414169,
		0.372344,
		0.331456,
		0.291724,
		0.253469,
		0.217088,
		0.183009,
		0.151632,
		0.123286,
		0.098195,
		0.076464,
		0.058074,
		0.042898,
		0.030711,
		0.021213,
		0.014057,
		0.00887,
		0.005276,
		0.002916,
		0.001468,
		0.000652,
		0.000243,
		6.9e-005,
		1.2e-005,
		0.002007,
		0.010409,
		0.024457,
		0.043203,
		0.065746,
		0.091224,
		0.118823,
		0.147776,
		0.177368,
		0.206935,
		0.235873,
		0.263636,
		0.289746,
		0.313795,
		0.33545,
		0.354464,
		0.370678,
		0.384031,
		0.394565,
		0.402434,
		0.4079,
		0.411333,
		0.413186,
		0.413966,
		0.414169,
		0.414184,
		0.37252,
		0.332039,
		0.292911,
		0.255338,
		0.219602,
		0.186039,
		0.154997,
		0.126786,
		0.101641,
		0.079696,
		0.060974,
		0.045391,
		0.032763,
		0.02283,
		0.015271,
		0.009733,
		0.005853,
		0.003275,
		0.001671,
		0.000754,
		0.000285,
		8.2e-005,
		1.5e-005,
		0.003111,
		0.011859,
		0.025391,
		0.04289,
		0.063573,
		0.086698,
		0.11156,
		0.137498,
		0.163895,
		0.190179,
		0.21583,
		0.240381,
		0.26342,
		0.284601,
		0.303641,
		0.320332,
		0.334543,
		0.346229,
		0.355436,
		0.362304,
		0.367068,
		0.370055,
		0.371666,
		0.372344,
		0.37252,
		0.372533,
		0.33219,
		0.293409,
		0.256343,
		0.22117,
		0.188127,
		0.157487,
		0.129518,
		0.104444,
		0.082414,
		0.063481,
		0.047595,
		0.034615,
		0.024314,
		0.016402,
		0.01055,
		0.006407,
		0.003623,
		0.00187,
		0.000854,
		0.000328,
		9.6e-005,
		1.7e-005,
		0.000176,
		0.004205,
		0.012984,
		0.025783,
		0.041898,
		0.060661,
		0.081434,
		0.103613,
		0.126633,
		0.149966,
		0.173123,
		0.195661,
		0.217181,
		0.237336,
		0.25583,
		0.272428,
		0.286955,
		0.299306,
		0.309448,
		0.317427,
		0.323371,
		0.327489,
		0.330067,
		0.331456,
		0.332039,
		0.33219,
		0.332201,
		0.293537,
		0.256763,
		0.222012,
		0.189429,
		0.159201,
		0.131537,
		0.10663,
		0.084624,
		0.065587,
		0.049499,
		0.036251,
		0.025651,
		0.01744,
		0.011311,
		0.00693,
		0.003956,
		0.002063,
		0.000953,
		0.000371,
		0.00011,
		2e-005,
		0.000604,
		0.005191,
		0.013746,
		0.025644,
		0.040291,
		0.057115,
		0.075577,
		0.095162,
		0.115392,
		0.135819,
		0.156029,
		0.175647,
		0.194338,
		0.211808,
		0.22781,
		0.242148,
		0.254678,
		0.265316,
		0.274039,
		0.280893,
		0.285991,
		0.289518,
		0.291724,
		0.292911,
		0.293409,
		0.293537,
		0.293547,
		0.256872,
		0.222364,
		0.190126,
		0.160268,
		0.132926,
		0.108244,
		0.086344,
		0.067296,
		0.051096,
		0.037661,
		0.02683,
		0.018373,
		0.012007,
		0.007416,
		0.00427,
		0.002247,
		0.001049,
		0.000412,
		0.000124,
		2.3e-005,
		0.001155,
		0.006001,
		0.014128,
		0.025011,
		0.038146,
		0.053053,
		0.069278,
		0.08639,
		0.103984,
		0.121685,
		0.139147,
		0.156055,
		0.172129,
		0.187125,
		0.200837,
		0.213104,
		0.223808,
		0.232882,
		0.240313,
		0.246144,
		0.250476,
		0.253469,
		0.255338,
		0.256343,
		0.256763,
		0.256872,
		0.25688,
		0.222455,
		0.190417,
		0.160839,
		0.133789,
		0.109353,
		0.087614,
		0.068626,
		0.052391,
		0.038844,
		0.027847,
		0.019196,
		0.012633,
		0.00786,
		0.004562,
		0.002421,
		0.00114,
		0.000452,
		0.000138,
		2.6e-005,
		0.001726,
		0.006591,
		0.014139,
		0.023933,
		0.035553,
		0.0486,
		0.062693,
		0.077474,
		0.092608,
		0.107782,
		0.122709,
		0.137128,
		0.150806,
		0.163544,
		0.175172,
		0.185558,
		0.194608,
		0.202269,
		0.208535,
		0.213445,
		0.217088,
		0.219602,
		0.22117,
		0.222012,
		0.222364,
		0.222455,
		0.222461,
		0.190492,
		0.161077,
		0.13425,
		0.110041,
		0.088484,
		0.069605,
		0.053397,
		0.039801,
		0.028697,
		0.019903,
		0.013183,
		0.008259,
		0.004828,
		0.002582,
		0.001225,
		0.000491,
		0.000151,
		2.9e-005,
		9.4e-005,
		0.002242,
		0.006937,
		0.013802,
		0.022475,
		0.032612,
		0.043882,
		0.055972,
		0.068588,
		0.081453,
		0.094309,
		0.106923,
		0.119079,
		0.130588,
		0.141286,
		0.151036,
		0.159732,
		0.167298,
		0.173695,
		0.17892,
		0.183009,
		0.186039,
		0.188127,
		0.189429,
		0.190126,
		0.190417,
		0.190492,
		0.190498,
		0.161138,
		0.134442,
		0.110408,
		0.089024,
		0.070277,
		0.054138,
		0.040544,
		0.029385,
		0.020495,
		0.013657,
		0.008609,
		0.005066,
		0.002729,
		0.001305,
		0.000526,
		0.000163,
		3.1e-005,
		0.000308,
		0.002653,
		0.007038,
		0.013156,
		0.020712,
		0.029425,
		0.039027,
		0.049261,
		0.05989,
		0.070687,
		0.081444,
		0.09197,
		0.102092,
		0.111657,
		0.120532,
		0.128609,
		0.135801,
		0.142051,
		0.147328,
		0.151632,
		0.154997,
		0.157487,
		0.159201,
		0.160268,
		0.160839,
		0.161077,
		0.161138,
		0.161143,
		0.134492,
		0.110561,
		0.089312,
		0.070692,
		0.054644,
		0.041091,
		0.029919,
		0.020973,
		0.014052,
		0.00891,
		0.005276,
		0.00286,
		0.001377,
		0.00056,
		0.000175,
		3.4e-005,
		0.000562,
		0.002929,
		0.006907,
		0.01225,
		0.018722,
		0.026096,
		0.034155,
		0.042695,
		0.051523,
		0.06046,
		0.069338,
		0.078003,
		0.086318,
		0.094161,
		0.101426,
		0.108027,
		0.113897,
		0.118991,
		0.123286,
		0.126786,
		0.129518,
		0.131537,
		0.132926,
		0.133789,
		0.13425,
		0.134442,
		0.134492,
		0.134496,
		0.1106,
		0.089431,
		0.070914,
		0.054958,
		0.041464,
		0.030311,
		0.021344,
		0.014371,
		0.009162,
		0.005456,
		0.002975,
		0.001442,
		0.00059,
		0.000185,
		3.6e-005,
		0.000799,
		0.003057,
		0.00657,
		0.011143,
		0.016587,
		0.022723,
		0.029379,
		0.036394,
		0.043615,
		0.0509,
		0.058116,
		0.065143,
		0.071873,
		0.078208,
		0.084068,
		0.089384,
		0.094104,
		0.098195,
		0.101641,
		0.104444,
		0.10663,
		0.108244,
		0.109353,
		0.110041,
		0.110408,
		0.110561,
		0.1106,
		0.110603,
		0.089462,
		0.071006,
		0.055125,
		0.041695,
		0.030578,
		0.021616,
		0.014618,
		0.009364,
		0.005605,
		0.003074,
		0.001498,
		0.000617,
		0.000195,
		3.8e-005,
		4.1e-005,
		0.000982,
		0.003043,
		0.006065,
		0.009895,
		0.014386,
		0.019399,
		0.0248,
		0.030463,
		0.036269,
		0.042108,
		0.047876,
		0.05348,
		0.058836,
		0.06387,
		0.068519,
		0.072729,
		0.076464,
		0.079696,
		0.082414,
		0.084624,
		0.086344,
		0.087614,
		0.088484,
		0.089024,
		0.089312,
		0.089431,
		0.089462,
		0.089465,
		0.071029,
		0.055194,
		0.041817,
		0.030743,
		0.021801,
		0.014799,
		0.009521,
		0.005726,
		0.003156,
		0.001547,
		0.00064,
		0.000204,
		4e-005,
		0.000127,
		0.001091,
		0.0029,
		0.005431,
		0.008567,
		0.012195,
		0.016209,
		0.020506,
		0.02499,
		0.029569,
		0.034161,
		0.038685,
		0.043071,
		0.047255,
		0.051181,
		0.0548,
		0.058074,
		0.060974,
		0.063481,
		0.065587,
		0.067296,
		0.068626,
		0.069605,
		0.070277,
		0.070692,
		0.070914,
		0.071006,
		0.071029,
		0.071031,
		0.055212,
		0.041868,
		0.030831,
		0.021915,
		0.014922,
		0.009635,
		0.005819,
		0.003222,
		0.001587,
		0.000661,
		0.000211,
		4.2e-005,
		0.000215,
		0.001123,
		0.002653,
		0.004715,
		0.007219,
		0.010082,
		0.013224,
		0.016567,
		0.02004,
		0.023575,
		0.027109,
		0.030582,
		0.033943,
		0.037142,
		0.040139,
		0.042898,
		0.045391,
		0.047595,
		0.049499,
		0.051096,
		0.052391,
		0.053397,
		0.054138,
		0.054644,
		0.054958,
		0.055125,
		0.055194,
		0.055212,
		0.055213,
		0.041881,
		0.030867,
		0.021976,
		0.014998,
		0.009713,
		0.005887,
		0.003273,
		0.001619,
		0.000677,
		0.000218,
		4.4e-005,
		0.000283,
		0.001083,
		0.002331,
		0.00396,
		0.005905,
		0.008105,
		0.010502,
		0.013038,
		0.015662,
		0.018324,
		0.020976,
		0.023578,
		0.02609,
		0.028478,
		0.030711,
		0.032763,
		0.034615,
		0.036251,
		0.037661,
		0.038844,
		0.039801,
		0.040544,
		0.041091,
		0.041464,
		0.041695,
		0.041817,
		0.041868,
		0.041881,
		0.041882,
		0.030877,
		0.022001,
		0.015038,
		0.009761,
		0.005933,
		0.003311,
		0.001644,
		0.000691,
		0.000223,
		4.5e-005,
		1.3e-005,
		0.000317,
		0.000983,
		0.001962,
		0.003208,
		0.004672,
		0.006313,
		0.008087,
		0.009956,
		0.011881,
		0.013828,
		0.015763,
		0.017657,
		0.019481,
		0.021213,
		0.02283,
		0.024314,
		0.025651,
		0.02683,
		0.027847,
		0.028697,
		0.029385,
		0.029919,
		0.030311,
		0.030578,
		0.030743,
		0.030831,
		0.030867,
		0.030877,
		0.030877,
		0.022007,
		0.015055,
		0.009787,
		0.005962,
		0.003336,
		0.001662,
		0.000701,
		0.000228,
		4.6e-005,
		3.7e-005,
		0.000316,
		0.000842,
		0.001579,
		0.002495,
		0.003559,
		0.004739,
		0.006008,
		0.007338,
		0.008703,
		0.010079,
		0.011443,
		0.012776,
		0.014057,
		0.015271,
		0.016402,
		0.01744,
		0.018373,
		0.019196,
		0.019903,
		0.020495,
		0.020973,
		0.021344,
		0.021616,
		0.021801,
		0.021915,
		0.021976,
		0.022001,
		0.022007,
		0.022008,
		0.015059,
		0.009797,
		0.005977,
		0.003351,
		0.001675,
		0.000709,
		0.000231,
		4.7e-005,
		5.5e-005,
		0.000287,
		0.000679,
		0.001208,
		0.001853,
		0.002592,
		0.003407,
		0.004277,
		0.005185,
		0.006113,
		0.007047,
		0.00797,
		0.00887,
		0.009733,
		0.01055,
		0.011311,
		0.012007,
		0.012633,
		0.013183,
		0.013657,
		0.014052,
		0.014371,
		0.014618,
		0.014799,
		0.014922,
		0.014998,
		0.015038,
		0.015055,
		0.015059,
		0.015059,
		0.0098,
		0.005983,
		0.00336,
		0.001682,
		0.000714,
		0.000233,
		4.7e-005,
		6.2e-005,
		0.000237,
		0.000512,
		0.000871,
		0.001301,
		0.00179,
		0.002323,
		0.00289,
		0.003479,
		0.00408,
		0.004682,
		0.005276,
		0.005853,
		0.006407,
		0.00693,
		0.007416,
		0.00786,
		0.008259,
		0.008609,
		0.00891,
		0.009162,
		0.009364,
		0.009521,
		0.009635,
		0.009713,
		0.009761,
		0.009787,
		0.009797,
		0.0098,
		0.0098,
		0.005985,
		0.003363,
		0.001686,
		0.000717,
		0.000235,
		4.8e-005,
		5.7e-005,
		0.000179,
		0.000357,
		0.000585,
		0.000854,
		0.001156,
		0.001483,
		0.00183,
		0.002188,
		0.002552,
		0.002916,
		0.003275,
		0.003623,
		0.003956,
		0.00427,
		0.004562,
		0.004828,
		0.005066,
		0.005276,
		0.005456,
		0.005605,
		0.005726,
		0.005819,
		0.005887,
		0.005933,
		0.005962,
		0.005977,
		0.005983,
		0.005985,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000236,
		4.8e-005,
		4.5e-005,
		0.000121,
		0.000227,
		0.000359,
		0.000513,
		0.000685,
		0.00087,
		0.001064,
		0.001265,
		0.001468,
		0.001671,
		0.00187,
		0.002063,
		0.002247,
		0.002421,
		0.002582,
		0.002729,
		0.00286,
		0.002975,
		0.003074,
		0.003156,
		0.003222,
		0.003273,
		0.003311,
		0.003336,
		0.003351,
		0.00336,
		0.003363,
		0.003364,
		0.003364,
		0.001688,
		0.000719,
		0.000236,
		4.8e-005,
		3e-005,
		7.2e-005,
		0.000127,
		0.000196,
		0.000274,
		0.000361,
		0.000454,
		0.000552,
		0.000652,
		0.000754,
		0.000854,
		0.000953,
		0.001049,
		0.00114,
		0.001225,
		0.001305,
		0.001377,
		0.001442,
		0.001498,
		0.001547,
		0.001587,
		0.001619,
		0.001644,
		0.001662,
		0.001675,
		0.001682,
		0.001686,
		0.001688,
		0.001688,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		1.6e-005,
		3.5e-005,
		6e-005,
		9e-005,
		0.000124,
		0.000161,
		0.000201,
		0.000243,
		0.000285,
		0.000328,
		0.000371,
		0.000412,
		0.000452,
		0.000491,
		0.000526,
		0.00056,
		0.00059,
		0.000617,
		0.00064,
		0.000661,
		0.000677,
		0.000691,
		0.000701,
		0.000709,
		0.000714,
		0.000717,
		0.000719,
		0.000719,
		0.000719,
		0.000719,
		0.000237,
		4.9e-005,
		1.3e-005,
		2.2e-005,
		3.2e-005,
		4.3e-005,
		5.6e-005,
		6.9e-005,
		8.2e-005,
		9.6e-005,
		0.00011,
		0.000124,
		0.000138,
		0.000151,
		0.000163,
		0.000175,
		0.000185,
		0.000195,
		0.000204,
		0.000211,
		0.000218,
		0.000223,
		0.000228,
		0.000231,
		0.000233,
		0.000235,
		0.000236,
		0.000236,
		0.000237,
		0.000237,
		0.000237,
		4.9e-005,
		1.2e-005,
		1.5e-005,
		1.7e-005,
		2e-005,
		2.3e-005,
		2.6e-005,
		2.9e-005,
		3.1e-005,
		3.4e-005,
		3.6e-005,
		3.8e-005,
		4e-005,
		4.2e-005,
		4.4e-005,
		4.5e-005,
		4.6e-005,
		4.7e-005,
		4.7e-005,
		4.8e-005,
		4.8e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		4.9e-005,
		1.1e-005,
		1.6e-005,
		1e-005,
		2e-005,
		3.5e-005,
		3.9e-005,
		2.1e-005,
		2.9e-005,
		6e-005,
		8.4e-005,
		7.9e-005,
		3.7e-005,
		1.1e-005,
		4e-005,
		9e-005,
		0.000143,
		0.00017,
		0.000143,
		6.2e-005,
		1.3e-005,
		5.2e-005,
		0.000124,
		0.000215,
		0.00029,
		0.000308,
		0.000237,
		9.7e-005,
		1.6e-005,
		6.5e-005,
		0.000161,
		0.000295,
		0.000434,
		0.000524,
		0.000512,
		0.000371,
		0.000144,
		1.9e-005,
		7.9e-005,
		0.000201,
		0.000384,
		0.000597,
		0.000783,
		0.000871,
		0.000799,
		0.000551,
		0.000205,
		2.2e-005,
		9.3e-005,
		0.000243,
		0.000479,
		0.000776,
		0.001077,
		0.001301,
		0.001359,
		0.001186,
		0.000785,
		0.000283,
		2.4e-005,
		0.000107,
		0.000285,
		0.000577,
		0.000966,
		0.001399,
		0.00179,
		0.00203,
		0.002017,
		0.001691,
		0.001083,
		0.000379,
		2.7e-005,
		0.000121,
		0.000328,
		0.000678,
		0.001164,
		0.001742,
		0.002323,
		0.00279,
		0.003011,
		0.002874,
		0.002331,
		0.001451,
		0.000496,
		3e-005,
		0.000134,
		0.000371,
		0.000779,
		0.001366,
		0.002098,
		0.00289,
		0.00362,
		0.004136,
		0.004288,
		0.00396,
		0.003122,
		0.001898,
		0.000635,
		3.3e-005,
		0.000147,
		0.000412,
		0.000879,
		0.00157,
		0.002461,
		0.003479,
		0.004501,
		0.005365,
		0.005889,
		0.005905,
		0.005302,
		0.004083,
		0.002431,
		0.000799,
		3.5e-005,
		0.00016,
		0.000452,
		0.000977,
		0.001771,
		0.002826,
		0.00408,
		0.005416,
		0.006667,
		0.007633,
		0.008105,
		0.007904,
		0.00693,
		0.005227,
		0.003057,
		0.000989,
		3.7e-005,
		0.000172,
		0.000491,
		0.001072,
		0.001968,
		0.003186,
		0.004682,
		0.006347,
		0.008018,
		0.009482,
		0.010502,
		0.010844,
		0.010326,
		0.008869,
		0.00657,
		0.003782,
		0.001206,
		3.9e-005,
		0.000183,
		0.000526,
		0.001162,
		0.002157,
		0.003537,
		0.005276,
		0.007279,
		0.009391,
		0.011397,
		0.013038,
		0.014044,
		0.01416,
		0.013209,
		0.011143,
		0.008124,
		0.00461,
		0.001451,
		4.1e-005,
		0.000193,
		0.00056,
		0.001246,
		0.002336,
		0.003874,
		0.005853,
		0.008198,
		0.010764,
		0.013341,
		0.015662,
		0.017426,
		0.018328,
		0.018105,
		0.016587,
		0.013772,
		0.009899,
		0.005545,
		0.001726,
		4.3e-005,
		0.000202,
		0.00059,
		0.001323,
		0.002503,
		0.004194,
		0.006407,
		0.009089,
		0.012115,
		0.015282,
		0.018324,
		0.020921,
		0.02273,
		0.023421,
		0.022723,
		0.02049,
		0.016772,
		0.011902,
		0.006591,
		0.002029,
		4.4e-005,
		0.00021,
		0.000617,
		0.001394,
		0.002657,
		0.004491,
		0.00693,
		0.009942,
		0.013424,
		0.017189,
		0.020976,
		0.024461,
		0.027273,
		0.02903,
		0.029379,
		0.028054,
		0.024941,
		0.020157,
		0.014139,
		0.007747,
		0.002362,
		4.5e-005,
		0.000216,
		0.00064,
		0.001456,
		0.002796,
		0.004764,
		0.007416,
		0.010746,
		0.014673,
		0.019033,
		0.023578,
		0.027985,
		0.031869,
		0.034811,
		0.036394,
		0.036252,
		0.034131,
		0.029959,
		0.023933,
		0.016611,
		0.009012,
		0.002723,
		4.6e-005,
		0.000222,
		0.000661,
		0.001511,
		0.00292,
		0.005009,
		0.00786,
		0.011491,
		0.015848,
		0.02079,
		0.02609,
		0.031435,
		0.036436,
		0.040651,
		0.043615,
		0.044882,
		0.04408,
		0.040976,
		0.035553,
		0.028102,
		0.019315,
		0.010384,
		0.003111,
		4.7e-005,
		0.000227,
		0.000677,
		0.001557,
		0.003027,
		0.005226,
		0.008259,
		0.01217,
		0.016933,
		0.022437,
		0.028478,
		0.034759,
		0.0409,
		0.046446,
		0.0509,
		0.053754,
		0.054541,
		0.052889,
		0.0486,
		0.041725,
		0.03266,
		0.022245,
		0.011859,
		0.003525,
		4.8e-005,
		0.00023,
		0.000691,
		0.001596,
		0.003117,
		0.005413,
		0.008609,
		0.012777,
		0.01792,
		0.023956,
		0.030711,
		0.037912,
		0.045192,
		0.052099,
		0.058116,
		0.062691,
		0.06528,
		0.0654,
		0.062693,
		0.057003,
		0.048465,
		0.037596,
		0.025391,
		0.01343,
		0.003962,
		4.8e-005,
		0.000233,
		0.000701,
		0.001626,
		0.003191,
		0.005571,
		0.00891,
		0.013309,
		0.018799,
		0.025331,
		0.032763,
		0.040853,
		0.049254,
		0.057525,
		0.065143,
		0.07153,
		0.076083,
		0.078227,
		0.077474,
		0.073489,
		0.066174,
		0.055758,
		0.04289,
		0.028739,
		0.015087,
		0.00442,
		4.8e-005,
		0.000235,
		0.000709,
		0.001649,
		0.00325,
		0.005698,
		0.009162,
		0.013763,
		0.019564,
		0.026551,
		0.034615,
		0.043547,
		0.053032,
		0.062646,
		0.071873,
		0.08012,
		0.086747,
		0.091109,
		0.092608,
		0.090758,
		0.08526,
		0.076086,
		0.063573,
		0.048517,
		0.032269,
		0.016822,
		0.004895,
		4.9e-005,
		0.000236,
		0.000714,
		0.001666,
		0.003294,
		0.005798,
		0.009364,
		0.014139,
		0.020214,
		0.027608,
		0.036251,
		0.04597,
		0.056483,
		0.067396,
		0.078208,
		0.088328,
		0.097092,
		0.103804,
		0.107782,
		0.108413,
		0.105226,
		0.097968,
		0.086698,
		0.071873,
		0.054447,
		0.03596,
		0.018622,
		0.005385,
		4.9e-005,
		0.000236,
		0.000717,
		0.001677,
		0.003325,
		0.005872,
		0.009521,
		0.014439,
		0.020748,
		0.0285,
		0.037661,
		0.0481,
		0.059573,
		0.07172,
		0.084068,
		0.096035,
		0.106953,
		0.116093,
		0.122709,
		0.126086,
		0.125608,
		0.120829,
		0.11156,
		0.097956,
		0.080608,
		0.06064,
		0.039786,
		0.020474,
		0.005885,
		4.9e-005,
		0.000237,
		0.000719,
		0.001684,
		0.003345,
		0.005924,
		0.009635,
		0.014669,
		0.021172,
		0.029228,
		0.038844,
		0.049927,
		0.062277,
		0.075576,
		0.089384,
		0.103142,
		0.116188,
		0.12778,
		0.137128,
		0.143438,
		0.145975,
		0.144131,
		0.137498,
		0.125963,
		0.109792,
		0.089722,
		0.067052,
		0.043719,
		0.022365,
		0.006393,
		4.9e-005,
		0.000237,
		0.000719,
		0.001687,
		0.003356,
		0.005956,
		0.009713,
		0.014835,
		0.021491,
		0.029799,
		0.039801,
		0.051448,
		0.064584,
		0.078936,
		0.094104,
		0.109565,
		0.124676,
		0.138695,
		0.150806,
		0.160162,
		0.165934,
		0.167375,
		0.163895,
		0.155141,
		0.141088,
		0.122125,
		0.099147,
		0.073636,
		0.047728,
		0.024279,
		0.006903,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003362,
		0.005974,
		0.009761,
		0.014945,
		0.021718,
		0.030225,
		0.040544,
		0.052669,
		0.06649,
		0.081783,
		0.098195,
		0.115245,
		0.13232,
		0.148694,
		0.163544,
		0.175988,
		0.185128,
		0.190107,
		0.190179,
		0.184787,
		0.173644,
		0.156828,
		0.134865,
		0.108813,
		0.080339,
		0.051783,
		0.026203,
		0.007413,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005982,
		0.009787,
		0.015011,
		0.021866,
		0.030522,
		0.041091,
		0.053606,
		0.068007,
		0.084118,
		0.101641,
		0.12014,
		0.139047,
		0.157661,
		0.175172,
		0.19068,
		0.203242,
		0.211916,
		0.21583,
		0.214251,
		0.206667,
		0.192875,
		0.173066,
		0.147912,
		0.118641,
		0.08711,
		0.055852,
		0.028122,
		0.007919,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.009797,
		0.015044,
		0.021951,
		0.03071,
		0.041464,
		0.054285,
		0.069157,
		0.085958,
		0.104444,
		0.124235,
		0.144812,
		0.165515,
		0.185558,
		0.204045,
		0.220006,
		0.232442,
		0.240381,
		0.242943,
		0.239425,
		0.229374,
		0.212684,
		0.18967,
		0.161159,
		0.128554,
		0.093895,
		0.059906,
		0.030023,
		0.008417,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015057,
		0.021992,
		0.030815,
		0.041695,
		0.054739,
		0.069976,
		0.087336,
		0.10663,
		0.127539,
		0.1496,
		0.172206,
		0.194608,
		0.21593,
		0.235198,
		0.251376,
		0.26342,
		0.270336,
		0.271253,
		0.265509,
		0.252728,
		0.232908,
		0.206504,
		0.1745,
		0.138472,
		0.100644,
		0.063915,
		0.031895,
		0.008906,
		4.8e-005,
		0.000236,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022005,
		0.030862,
		0.041817,
		0.055011,
		0.070513,
		0.0883,
		0.108244,
		0.130086,
		0.153429,
		0.177722,
		0.202269,
		0.22623,
		0.248646,
		0.268467,
		0.284601,
		0.295966,
		0.30156,
		0.300538,
		0.292291,
		0.276532,
		0.253379,
		0.223428,
		0.187825,
		0.14832,
		0.107308,
		0.067856,
		0.033727,
		0.009384,
		4.8e-005,
		0.000235,
		0.000717,
		0.001686,
		0.003362,
		0.005984,
		0.0098,
		0.015059,
		0.022008,
		0.030876,
		0.041868,
		0.055149,
		0.070823,
		0.088916,
		0.109353,
		0.13194,
		0.156346,
		0.18209,
		0.208535,
		0.234893,
		0.260237,
		0.283525,
		0.303641,
		0.319444,
		0.32983,
		0.333802,
		0.330552,
		0.31954,
		0.300582,
		0.27392,
		0.240301,
		0.201032,
		0.158028,
		0.113845,
		0.071706,
		0.035511,
		0.009847,
		4.8e-005,
		0.000234,
		0.000714,
		0.001681,
		0.003356,
		0.005979,
		0.009797,
		0.015059,
		0.022008,
		0.030877,
		0.041881,
		0.055202,
		0.070972,
		0.089259,
		0.110041,
		0.133186,
		0.158434,
		0.185375,
		0.213445,
		0.241918,
		0.269916,
		0.296425,
		0.320332,
		0.340459,
		0.35563,
		0.36473,
		0.366781,
		0.361029,
		0.347016,
		0.324667,
		0.294361,
		0.256992,
		0.214025,
		0.167531,
		0.120219,
		0.075448,
		0.037241,
		0.010297,
		4.7e-005,
		0.000232,
		0.000709,
		0.001672,
		0.003345,
		0.005967,
		0.009787,
		0.015052,
		0.022005,
		0.030877,
		0.041882,
		0.055213,
		0.071023,
		0.089413,
		0.110408,
		0.133936,
		0.159806,
		0.187686,
		0.217088,
		0.247362,
		0.277694,
		0.307118,
		0.334543,
		0.358785,
		0.37862,
		0.392847,
		0.400354,
		0.400201,
		0.391695,
		0.374475,
		0.348584,
		0.314536,
		0.273376,
		0.226717,
		0.176778,
		0.126401,
		0.07907,
		0.038915,
		0.010731,
		4.6e-005,
		0.000228,
		0.000701,
		0.001658,
		0.003325,
		0.005942,
		0.009761,
		0.015031,
		0.021992,
		0.030871,
		0.041881,
		0.055213,
		0.071031,
		0.089459,
		0.110561,
		0.134318,
		0.160605,
		0.189167,
		0.219602,
		0.251342,
		0.283651,
		0.31563,
		0.346229,
		0.374285,
		0.398556,
		0.417783,
		0.430753,
		0.436373,
		0.43375,
		0.422271,
		0.401677,
		0.372134,
		0.334294,
		0.289343,
		0.239036,
		0.185725,
		0.132372,
		0.082565,
		0.040531,
		0.011152,
		4.5e-005,
		0.000224,
		0.000691,
		0.001639,
		0.003294,
		0.005901,
		0.009713,
		0.014983,
		0.021951,
		0.030844,
		0.041868,
		0.05521,
		0.071031,
		0.089465,
		0.1106,
		0.134464,
		0.160989,
		0.189997,
		0.22117,
		0.254032,
		0.287937,
		0.322066,
		0.355436,
		0.386925,
		0.415302,
		0.43928,
		0.457578,
		0.468986,
		0.472443,
		0.467118,
		0.452485,
		0.428393,
		0.395137,
		0.353499,
		0.304801,
		0.250926,
		0.194342,
		0.138117,
		0.085931,
		0.04209,
		0.011559,
		4.4e-005,
		0.000219,
		0.000677,
		0.001612,
		0.00325,
		0.005838,
		0.009635,
		0.014896,
		0.021866,
		0.030771,
		0.041817,
		0.055183,
		0.071023,
		0.089464,
		0.110603,
		0.134494,
		0.161121,
		0.190372,
		0.222012,
		0.25566,
		0.290769,
		0.326616,
		0.362304,
		0.396774,
		0.428835,
		0.457205,
		0.480563,
		0.497617,
		0.507174,
		0.508222,
		0.5,
		0.482077,
		0.454416,
		0.417432,
		0.372038,
		0.319677,
		0.262345,
		0.202613,
		0.143636,
		0.08917,
		0.043596,
		0.011955,
		4.3e-005,
		0.000213,
		0.000661,
		0.001578,
		0.003191,
		0.005752,
		0.009521,
		0.014759,
		0.021718,
		0.030628,
		0.041695,
		0.055094,
		0.070972,
		0.089445,
		0.1106,
		0.134495,
		0.161143,
		0.190485,
		0.222364,
		0.256491,
		0.292425,
		0.329549,
		0.367068,
		0.404013,
		0.439258,
		0.471554,
		0.499578,
		0.521988,
		0.537498,
		0.544946,
		0.543376,
		0.532108,
		0.510811,
		0.47956,
		0.438887,
		0.389823,
		0.333919,
		0.27327,
		0.210532,
		0.148931,
		0.09229,
		0.045055,
		0.01234,
		4.1e-005,
		0.000206,
		0.00064,
		0.001535,
		0.003117,
		0.005638,
		0.009364,
		0.014563,
		0.021491,
		0.030388,
		0.041464,
		0.054895,
		0.070823,
		0.089354,
		0.110561,
		0.134487,
		0.161143,
		0.190498,
		0.222455,
		0.25681,
		0.293225,
		0.331199,
		0.370055,
		0.408933,
		0.446798,
		0.482466,
		0.514646,
		0.541984,
		0.563133,
		0.576825,
		0.581942,
		0.577591,
		0.563181,
		0.538479,
		0.50367,
		0.459397,
		0.40679,
		0.347498,
		0.283695,
		0.218106,
		0.154017,
		0.095303,
		0.046472,
		0.012718,
		3.9e-005,
		0.000197,
		0.000617,
		0.001485,
		0.003027,
		0.005496,
		0.009162,
		0.014298,
		0.021172,
		0.03003,
		0.041091,
		0.054538,
		0.070513,
		0.089116,
		0.110408,
		0.134412,
		0.161121,
		0.190496,
		0.222461,
		0.256878,
		0.293501,
		0.331948,
		0.371666,
		0.411925,
		0.451806,
		0.490224,
		0.52595,
		0.557657,
		0.58398,
		0.603579,
		0.615215,
		0.617822,
		0.610581,
		0.592988,
		0.564909,
		0.526629,
		0.478887,
		0.422905,
		0.360406,
		0.29363,
		0.225354,
		0.15891,
		0.098223,
		0.047857,
		0.01309,
		3.7e-005,
		0.000188,
		0.00059,
		0.001426,
		0.00292,
		0.005324,
		0.00891,
		0.01396,
		0.020748,
		0.029533,
		0.040544,
		0.053976,
		0.069976,
		0.088647,
		0.110041,
		0.134165,
		0.160989,
		0.19045,
		0.222455,
		0.25688,
		0.293546,
		0.332175,
		0.372344,
		0.413461,
		0.454751,
		0.495252,
		0.533837,
		0.569245,
		0.600132,
		0.625132,
		0.642924,
		0.652309,
		0.652282,
		0.642099,
		0.621342,
		0.589968,
		0.548352,
		0.497317,
		0.438158,
		0.372656,
		0.3031,
		0.232304,
		0.163637,
		0.101067,
		0.049219,
		0.013459,
		3.5e-005,
		0.000177,
		0.00056,
		0.00136,
		0.002796,
		0.005121,
		0.008609,
		0.013546,
		0.020214,
		0.028884,
		0.039801,
		0.053172,
		0.069157,
		0.087866,
		0.109353,
		0.133616,
		0.160605,
		0.190229,
		0.222364,
		0.256861,
		0.293546,
		0.332201,
		0.37252,
		0.414053,
		0.456182,
		0.498098,
		0.53881,
		0.577165,
		0.611886,
		0.641623,
		0.665023,
		0.680795,
		0.687787,
		0.685059,
		0.67194,
		0.648096,
		0.613563,
		0.568794,
		0.514679,
		0.452566,
		0.384281,
		0.312142,
		0.238991,
		0.168226,
		0.103854,
		0.050566,
		0.013828,
		3.3e-005,
		0.000166,
		0.000526,
		0.001285,
		0.002657,
		0.00489,
		0.008259,
		0.013053,
		0.019564,
		0.028075,
		0.038844,
		0.052095,
		0.068007,
		0.086702,
		0.108244,
		0.132632,
		0.159806,
		0.189652,
		0.222012,
		0.256697,
		0.293501,
		0.332198,
		0.372533,
		0.414179,
		0.456678,
		0.499394,
		0.541506,
		0.58201,
		0.61974,
		0.653418,
		0.681705,
		0.703269,
		0.716858,
		0.721373,
		0.715937,
		0.699951,
		0.673152,
		0.635646,
		0.587947,
		0.530993,
		0.46617,
		0.395328,
		0.320804,
		0.245457,
		0.172707,
		0.106605,
		0.05191,
		0.014199,
		3e-005,
		0.000154,
		0.000491,
		0.001205,
		0.002503,
		0.004631,
		0.00786,
		0.012483,
		0.018799,
		0.0271,
		0.037661,
		0.050725,
		0.06649,
		0.085099,
		0.10663,
		0.131095,
		0.158434,
		0.18852,
		0.22117,
		0.256157,
		0.293225,
		0.332104,
		0.37252,
		0.414184,
		0.45676,
		0.49979,
		0.542647,
		0.584512,
		0.624385,
		0.661108,
		0.693416,
		0.719989,
		0.73953,
		0.75083,
		0.752845,
		0.744755,
		0.726028,
		0.696458,
		0.656209,
		0.605834,
		0.546305,
		0.479026,
		0.405858,
		0.329143,
		0.251747,
		0.177115,
		0.109339,
		0.053259,
		0.014576,
		2.7e-005,
		0.000141,
		0.000452,
		0.001117,
		0.002336,
		0.004345,
		0.007416,
		0.011839,
		0.01792,
		0.025961,
		0.036251,
		0.049052,
		0.064584,
		0.083014,
		0.104444,
		0.128904,
		0.156346,
		0.186644,
		0.219602,
		0.254963,
		0.292425,
		0.331662,
		0.372344,
		0.414148,
		0.45676,
		0.499836,
		0.542944,
		0.585481,
		0.626654,
		0.665486,
		0.700845,
		0.73149,
		0.756135,
		0.773521,
		0.782486,
		0.782037,
		0.771408,
		0.750119,
		0.71801,
		0.675281,
		0.622512,
		0.560686,
		0.491211,
		0.415942,
		0.337219,
		0.257911,
		0.181483,
		0.112078,
		0.054623,
		0.014959,
		2.4e-005,
		0.000127,
		0.000412,
		0.001025,
		0.002157,
		0.004037,
		0.00693,
		0.011126,
		0.016933,
		0.024662,
		0.034615,
		0.047072,
		0.062277,
		0.080424,
		0.101641,
		0.125984,
		0.153429,
		0.183862,
		0.217088,
		0.252835,
		0.290769,
		0.330511,
		0.371666,
		0.413844,
		0.456678,
		0.49983,
		0.542966,
		0.585688,
		0.627445,
		0.667493,
		0.704898,
		0.738568,
		0.767303,
		0.789857,
		0.805015,
		0.81166,
		0.808844,
		0.795845,
		0.772221,
		0.737842,
		0.692926,
		0.63806,
		0.574223,
		0.502809,
		0.425657,
		0.345096,
		0.263995,
		0.185844,
		0.114839,
		0.05601,
		0.015352,
		2.2e-005,
		0.000114,
		0.000371,
		0.000929,
		0.001968,
		0.003708,
		0.006407,
		0.010351,
		0.015848,
		0.023214,
		0.032763,
		0.044794,
		0.059573,
		0.077319,
		0.098195,
		0.122288,
		0.1496,
		0.180046,
		0.213445,
		0.249525,
		0.287937,
		0.328265,
		0.370055,
		0.412843,
		0.456182,
		0.499666,
		0.542944,
		0.585696,
		0.627576,
		0.668108,
		0.706621,
		0.742246,
		0.773948,
		0.800578,
		0.820937,
		0.833854,
		0.838252,
		0.83322,
		0.818069,
		0.792377,
		0.756027,
		0.709237,
		0.652581,
		0.587017,
		0.513912,
		0.435083,
		0.352837,
		0.270048,
		0.190228,
		0.117642,
		0.057429,
		0.015757,
		1.9e-005,
		0.0001,
		0.000328,
		0.000829,
		0.001771,
		0.003363,
		0.005853,
		0.009521,
		0.014673,
		0.021629,
		0.030711,
		0.042233,
		0.056483,
		0.073709,
		0.094104,
		0.117792,
		0.144812,
		0.17511,
		0.208535,
		0.244832,
		0.283651,
		0.324562,
		0.367068,
		0.410642,
		0.454751,
		0.498898,
		0.542647,
		0.585637,
		0.627576,
		0.668181,
		0.707073,
		0.743679,
		0.777215,
		0.80672,
		0.831105,
		0.849223,
		0.859942,
		0.862222,
		0.855177,
		0.838132,
		0.810671,
		0.77267,
		0.724327,
		0.666188,
		0.599175,
		0.524618,
		0.444298,
		0.360506,
		0.276112,
		0.194665,
		0.120501,
		0.058887,
		0.016175,
		1.6e-005,
		8.6e-005,
		0.000285,
		0.000728,
		0.00157,
		0.003007,
		0.005276,
		0.008648,
		0.013424,
		0.019924,
		0.028478,
		0.039411,
		0.053032,
		0.069614,
		0.089384,
		0.112501,
		0.139047,
		0.169008,
		0.202269,
		0.238608,
		0.277694,
		0.319097,
		0.362304,
		0.406741,
		0.451806,
		0.496909,
		0.541506,
		0.585137,
		0.627445,
		0.668171,
		0.707107,
		0.743987,
		0.778361,
		0.809553,
		0.836688,
		0.858746,
		0.874633,
		0.883254,
		0.883593,
		0.874775,
		0.856127,
		0.827216,
		0.787895,
		0.738324,
		0.679005,
		0.61081,
		0.535021,
		0.453381,
		0.368158,
		0.282229,
		0.199181,
		0.123433,
		0.06039,
		0.016608,
		1.3e-005,
		7.2e-005,
		0.000243,
		0.000627,
		0.001366,
		0.002644,
		0.004682,
		0.007741,
		0.012115,
		0.01812,
		0.02609,
		0.03636,
		0.049254,
		0.065071,
		0.084068,
		0.106443,
		0.13232,
		0.161731,
		0.194608,
		0.230768,
		0.269916,
		0.311642,
		0.355436,
		0.400704,
		0.446798,
		0.493048,
		0.53881,
		0.583504,
		0.626654,
		0.667916,
		0.707073,
		0.743999,
		0.778551,
		0.810427,
		0.839077,
		0.863728,
		0.883435,
		0.897154,
		0.903823,
		0.902438,
		0.892122,
		0.872181,
		0.842154,
		0.801847,
		0.751367,
		0.691157,
		0.622029,
		0.54521,
		0.462401,
		0.375848,
		0.288436,
		0.203799,
		0.12645,
		0.061944,
		0.017057,
		1.1e-005,
		5.9e-005,
		0.000201,
		0.000527,
		0.001164,
		0.002279,
		0.00408,
		0.006814,
		0.010764,
		0.016241,
		0.023578,
		0.033116,
		0.045192,
		0.060128,
		0.078208,
		0.099669,
		0.124676,
		0.153312,
		0.185558,
		0.221283,
		0.260237,
		0.302048,
		0.346229,
		0.392189,
		0.439258,
		0.486716,
		0.533837,
		0.579929,
		0.624385,
		0.666724,
		0.706621,
		0.743911,
		0.778551,
		0.810531,
		0.839707,
		0.865681,
		0.887791,
		0.905174,
		0.916836,
		0.921738,
		0.918878,
		0.907357,
		0.886448,
		0.855639,
		0.814675,
		0.763595,
		0.702768,
		0.632938,
		0.555271,
		0.471423,
		0.383622,
		0.294765,
		0.208539,
		0.129562,
		0.063554,
		0.017525,
		4.6e-005,
		0.000161,
		0.000431,
		0.000966,
		0.001918,
		0.003479,
		0.00588,
		0.009391,
		0.014314,
		0.020976,
		0.029723,
		0.0409,
		0.054846,
		0.071873,
		0.092251,
		0.116188,
		0.143817,
		0.175172,
		0.21018,
		0.248646,
		0.29025,
		0.334543,
		0.38096,
		0.428835,
		0.477426,
		0.52595,
		0.57363,
		0.61974,
		0.663655,
		0.704898,
		0.74317,
		0.778361,
		0.810517,
		0.839755,
		0.866103,
		0.889329,
		0.908898,
		0.924032,
		0.933787,
		0.937137,
		0.933069,
		0.92065,
		0.899099,
		0.867837,
		0.826533,
		0.775143,
		0.713954,
		0.643631,
		0.565277,
		0.480503,
		0.39152,
		0.301242,
		0.213418,
		0.132781,
		0.065226,
		0.018011,
		3.5e-005,
		0.000124,
		0.000339,
		0.000776,
		0.001568,
		0.00289,
		0.004955,
		0.008018,
		0.012367,
		0.018324,
		0.026229,
		0.036436,
		0.049296,
		0.065143,
		0.084279,
		0.106953,
		0.133344,
		0.163544,
		0.197539,
		0.235198,
		0.276259,
		0.320332,
		0.366892,
		0.415302,
		0.464821,
		0.514646,
		0.56394,
		0.611886,
		0.657731,
		0.700845,
		0.740761,
		0.777215,
		0.810153,
		0.839707,
		0.86612,
		0.889586,
		0.910055,
		0.927134,
		0.940135,
		0.948161,
		0.950194,
		0.945195,
		0.932184,
		0.91031,
		0.878913,
		0.837568,
		0.786138,
		0.724819,
		0.654192,
		0.575293,
		0.48969,
		0.399576,
		0.307892,
		0.218452,
		0.136115,
		0.066962,
		0.018518,
		2.4e-005,
		9e-005,
		0.000254,
		0.000597,
		0.001235,
		0.002323,
		0.004055,
		0.006667,
		0.010433,
		0.015662,
		0.022689,
		0.031869,
		0.04356,
		0.058116,
		0.075864,
		0.097092,
		0.122023,
		0.150806,
		0.183489,
		0.220006,
		0.260166,
		0.303641,
		0.349969,
		0.398556,
		0.448693,
		0.499578,
		0.55035,
		0.600132,
		0.648076,
		0.693416,
		0.735522,
		0.773948,
		0.808465,
		0.839077,
		0.865999,
		0.889586,
		0.910193,
		0.927957,
		0.942641,
		0.953657,
		0.960152,
		0.961111,
		0.955459,
		0.942154,
		0.920262,
		0.889025,
		0.847917,
		0.796695,
		0.735456,
		0.664693,
		0.585374,
		0.499021,
		0.407819,
		0.314732,
		0.223654,
		0.139572,
		0.068768,
		0.019046,
		1.5e-005,
		6e-005,
		0.000178,
		0.000434,
		0.000926,
		0.00179,
		0.003197,
		0.005365,
		0.008547,
		0.013038,
		0.019164,
		0.027273,
		0.03773,
		0.0509,
		0.067132,
		0.086747,
		0.110012,
		0.137128,
		0.168203,
		0.203242,
		0.242125,
		0.284601,
		0.330276,
		0.37862,
		0.428973,
		0.480563,
		0.532534,
		0.58398,
		0.633993,
		0.681705,
		0.726348,
		0.767303,
		0.804146,
		0.836688,
		0.864986,
		0.889329,
		0.910174,
		0.928019,
		0.943187,
		0.955611,
		0.96481,
		0.969981,
		0.970107,
		0.964069,
		0.950751,
		0.929125,
		0.898322,
		0.857704,
		0.806913,
		0.745942,
		0.675192,
		0.595562,
		0.508529,
		0.416269,
		0.321779,
		0.229034,
		0.143159,
		0.070648,
		0.019597,
		3.5e-005,
		0.000112,
		0.00029,
		0.000649,
		0.001301,
		0.0024,
		0.004136,
		0.006746,
		0.010502,
		0.015716,
		0.02273,
		0.031906,
		0.043615,
		0.058225,
		0.076083,
		0.097494,
		0.122709,
		0.151896,
		0.185128,
		0.22236,
		0.26342,
		0.307995,
		0.35563,
		0.405731,
		0.457578,
		0.510346,
		0.563133,
		0.615002,
		0.665023,
		0.712321,
		0.756135,
		0.795861,
		0.831105,
		0.861713,
		0.887791,
		0.909691,
		0.927957,
		0.943208,
		0.95594,
		0.966268,
		0.973833,
		0.977887,
		0.977409,
		0.971234,
		0.958161,
		0.937057,
		0.906936,
		0.867033,
		0.816874,
		0.756338,
		0.685735,
		0.605889,
		0.518237,
		0.424946,
		0.329048,
		0.234602,
		0.146884,
		0.072605,
		0.020173,
		1.6e-005,
		6e-005,
		0.00017,
		0.00041,
		0.000871,
		0.001682,
		0.003011,
		0.005068,
		0.008105,
		0.012416,
		0.018328,
		0.026196,
		0.036394,
		0.049299,
		0.06528,
		0.084677,
		0.107782,
		0.134822,
		0.165934,
		0.201151,
		0.240381,
		0.283397,
		0.32983,
		0.379165,
		0.430753,
		0.48382,
		0.537498,
		0.590853,
		0.642924,
		0.692773,
		0.73953,
		0.782445,
		0.820937,
		0.85464,
		0.883435,
		0.907465,
		0.927134,
		0.943052,
		0.95594,
		0.966443,
		0.974863,
		0.980982,
		0.984115,
		0.983245,
		0.977154,
		0.964555,
		0.9442,
		0.914977,
		0.875992,
		0.826642,
		0.766692,
		0.696355,
		0.616381,
		0.528163,
		0.433863,
		0.336549,
		0.24037,
		0.150754,
		0.074644,
		0.020774,
		2.2e-005,
		7.9e-005,
		0.000218,
		0.000512,
		0.001064,
		0.002017,
		0.003555,
		0.005905,
		0.009337,
		0.01416,
		0.020718,
		0.029379,
		0.040526,
		0.054541,
		0.071791,
		0.092608,
		0.117269,
		0.145975,
		0.178832,
		0.21583,
		0.256833,
		0.30156,
		0.34959,
		0.400354,
		0.453154,
		0.507174,
		0.561514,
		0.615215,
		0.667309,
		0.716858,
		0.763003,
		0.805015,
		0.842338,
		0.874633,
		0.901806,
		0.924032,
		0.941747,
		0.955611,
		0.966419,
		0.974941,
		0.981661,
		0.986518,
		0.988911,
		0.987832,
		0.982016,
		0.970086,
		0.950676,
		0.92254,
		0.884648,
		0.836264,
		0.777035,
		0.707073,
		0.627051,
		0.53832,
		0.443032,
		0.344293,
		0.246346,
		0.154777,
		0.07677,
		0.021402,
		2.1e-005,
		8.2e-005,
		0.000237,
		0.000567,
		0.001186,
		0.002251,
		0.00396,
		0.006556,
		0.010326,
		0.015596,
		0.022723,
		0.032087,
		0.04408,
		0.059088,
		0.077474,
		0.099561,
		0.125608,
		0.155789,
		0.190179,
		0.228728,
		0.271253,
		0.317429,
		0.366781,
		0.4187,
		0.472443,
		0.527165,
		0.581942,
		0.635805,
		0.687787,
		0.736961,
		0.782486,
		0.823657,
		0.859942,
		0.891026,
		0.916836,
		0.937562,
		0.953657,
		0.965809,
		0.974863,
		0.981687,
		0.986925,
		0.990698,
		0.992508,
		0.991373,
		0.985988,
		0.974887,
		0.956583,
		0.929694,
		0.893048,
		0.845772,
		0.787387,
		0.717902,
		0.637912,
		0.548717,
		0.452462,
		0.352291,
		0.25254,
		0.15896,
		0.078987,
		0.022059,
		6.2e-005,
		0.000213,
		0.000551,
		0.0012,
		0.002331,
		0.004154,
		0.00693,
		0.010961,
		0.016587,
		0.02418,
		0.034131,
		0.046839,
		0.062693,
		0.082051,
		0.105226,
		0.132457,
		0.163895,
		0.19958,
		0.239425,
		0.283205,
		0.330552,
		0.38095,
		0.43375,
		0.488182,
		0.543376,
		0.598399,
		0.652282,
		0.704067,
		0.752845,
		0.7978,
		0.838252,
		0.873692,
		0.903823,
		0.928581,
		0.948161,
		0.963014,
		0.973833,
		0.981492,
		0.986925,
		0.990914,
		0.993768,
		0.995127,
		0.994054,
		0.989217,
		0.979067,
		0.961999,
		0.936491,
		0.901223,
		0.855181,
		0.797756,
		0.728847,
		0.648966,
		0.559361,
		0.462163,
		0.360554,
		0.258963,
		0.163312,
		0.081301,
		0.022747,
		2.5e-005,
		0.000144,
		0.000451,
		0.001083,
		0.002217,
		0.004083,
		0.006952,
		0.011143,
		0.017011,
		0.024941,
		0.035336,
		0.0486,
		0.065122,
		0.08526,
		0.109311,
		0.137498,
		0.169947,
		0.206667,
		0.247536,
		0.292291,
		0.340527,
		0.391695,
		0.445116,
		0.5,
		0.555467,
		0.610581,
		0.664385,
		0.715937,
		0.764353,
		0.808844,
		0.848753,
		0.883593,
		0.913075,
		0.937137,
		0.955961,
		0.969981,
		0.979875,
		0.986518,
		0.990884,
		0.993865,
		0.995957,
		0.996969,
		0.996039,
		0.991829,
		0.982718,
		0.966984,
		0.942964,
		0.909189,
		0.864496,
		0.808142,
		0.739907,
		0.660215,
		0.570256,
		0.472143,
		0.369091,
		0.265625,
		0.167843,
		0.083718,
		0.023468,
		5.2e-005,
		0.000283,
		0.000834,
		0.001898,
		0.003714,
		0.00657,
		0.0108,
		0.016772,
		0.024886,
		0.035553,
		0.049186,
		0.066174,
		0.086869,
		0.11156,
		0.140451,
		0.173644,
		0.211121,
		0.252728,
		0.298172,
		0.347016,
		0.398686,
		0.452485,
		0.507609,
		0.563181,
		0.618271,
		0.67194,
		0.723272,
		0.771408,
		0.815588,
		0.855177,
		0.889702,
		0.918878,
		0.942628,
		0.961111,
		0.974726,
		0.984115,
		0.990133,
		0.993768,
		0.99599,
		0.997467,
		0.99821,
		0.997471,
		0.993929,
		0.985907,
		0.971575,
		0.94913,
		0.916947,
		0.873711,
		0.818535,
		0.751076,
		0.671658,
		0.581406,
		0.48241,
		0.377913,
		0.272538,
		0.172561,
		0.086243,
		0.024223,
		9.7e-005,
		0.000496,
		0.001398,
		0.003057,
		0.005775,
		0.009899,
		0.015814,
		0.023933,
		0.034677,
		0.048465,
		0.06569,
		0.086698,
		0.111767,
		0.141088,
		0.174741,
		0.212684,
		0.254737,
		0.300582,
		0.349758,
		0.401677,
		0.455628,
		0.510811,
		0.566352,
		0.621342,
		0.674863,
		0.726028,
		0.774009,
		0.818069,
		0.857596,
		0.892122,
		0.921354,
		0.945195,
		0.963764,
		0.977409,
		0.986721,
		0.992508,
		0.995751,
		0.997467,
		0.998474,
		0.999005,
		0.99847,
		0.995602,
		0.988688,
		0.975798,
		0.954993,
		0.92449,
		0.882813,
		0.828924,
		0.762345,
		0.683291,
		0.592813,
		0.492972,
		0.387032,
		0.279714,
		0.177478,
		0.088884,
		0.025016,
		0.000162,
		0.000799,
		0.002176,
		0.00461,
		0.008464,
		0.014139,
		0.02206,
		0.03266,
		0.046366,
		0.063573,
		0.084624,
		0.109792,
		0.139251,
		0.173066,
		0.211175,
		0.253379,
		0.299339,
		0.348584,
		0.400512,
		0.454416,
		0.5095,
		0.564909,
		0.619756,
		0.673152,
		0.724239,
		0.772221,
		0.816383,
		0.856127,
		0.890985,
		0.92065,
		0.944988,
		0.964069,
		0.978178,
		0.987832,
		0.993781,
		0.996969,
		0.998438,
		0.999123,
		0.999483,
		0.999139,
		0.996917,
		0.991097,
		0.979666,
		0.960548,
		0.931802,
		0.891783,
		0.839291,
		0.773702,
		0.695109,
		0.604481,
		0.503836,
		0.39646,
		0.287165,
		0.182603,
		0.091646,
		0.025848,
		0.000253,
		0.001206,
		0.003192,
		0.006591,
		0.011818,
		0.019315,
		0.029527,
		0.04289,
		0.059802,
		0.080608,
		0.105574,
		0.134865,
		0.168534,
		0.206504,
		0.248563,
		0.294361,
		0.343419,
		0.395137,
		0.448814,
		0.50367,
		0.558873,
		0.613563,
		0.666884,
		0.71801,
		0.766169,
		0.810671,
		0.850921,
		0.886448,
		0.916918,
		0.942154,
		0.962163,
		0.977154,
		0.987562,
		0.994054,
		0.997522,
		0.999005,
		0.999523,
		0.999749,
		0.999561,
		0.99793,
		0.993163,
		0.983181,
		0.965783,
		0.938862,
		0.900598,
		0.849616,
		0.785132,
		0.707107,
		0.61641,
		0.515011,
		0.406208,
		0.294906,
		0.18795,
		0.094539,
		0.026722,
		0.000371,
		0.001726,
		0.004461,
		0.009012,
		0.015835,
		0.025391,
		0.038126,
		0.054447,
		0.074699,
		0.099147,
		0.127953,
		0.161159,
		0.198682,
		0.240301,
		0.285665,
		0.334294,
		0.385597,
		0.438887,
		0.493408,
		0.548352,
		0.602893,
		0.656209,
		0.707502,
		0.756027,
		0.801108,
		0.842154,
		0.878677,
		0.91031,
		0.936827,
		0.958161,
		0.974436,
		0.985988,
		0.993394,
		0.997471,
		0.99923,
		0.999749,
		0.999886,
		0.999806,
		0.998688,
		0.994906,
		0.986344,
		0.970681,
		0.945646,
		0.909232,
		0.859876,
		0.796619,
		0.719275,
		0.628601,
		0.526505,
		0.416289,
		0.302949,
		0.193529,
		0.09757,
		0.027641,
		0.000519,
		0.002362,
		0.005983,
		0.011859,
		0.020468,
		0.032269,
		0.04768,
		0.067052,
		0.090652,
		0.118641,
		0.151061,
		0.187825,
		0.228713,
		0.273376,
		0.321343,
		0.372038,
		0.424797,
		0.478887,
		0.533535,
		0.587947,
		0.641333,
		0.692926,
		0.742003,
		0.787895,
		0.830009,
		0.867837,
		0.900971,
		0.929125,
		0.952154,
		0.970086,
		0.983155,
		0.991829,
		0.996836,
		0.999139,
		0.999841,
		0.999952,
		0.999932,
		0.99923,
		0.996346,
		0.989153,
		0.975226,
		0.952128,
		0.917657,
		0.870045,
		0.808146,
		0.731605,
		0.641052,
		0.538323,
		0.426716,
		0.31131,
		0.199356,
		0.100747,
		0.028608,
		0.000697,
		0.003111,
		0.007742,
		0.015087,
		0.025625,
		0.039786,
		0.057933,
		0.080339,
		0.107171,
		0.138472,
		0.174161,
		0.214025,
		0.257723,
		0.304801,
		0.354703,
		0.40679,
		0.460362,
		0.514679,
		0.568982,
		0.622512,
		0.674529,
		0.724327,
		0.771243,
		0.814675,
		0.854087,
		0.889025,
		0.919139,
		0.9442,
		0.964136,
		0.979067,
		0.989349,
		0.995602,
		0.998721,
		0.999806,
		0.999981,
		0.999983,
		0.999595,
		0.997501,
		0.991607,
		0.9794,
		0.95828,
		0.925841,
		0.880097,
		0.819691,
		0.744084,
		0.653763,
		0.550474,
		0.437501,
		0.320005,
		0.205443,
		0.104082,
		0.029626,
		0.000902,
		0.003962,
		0.009702,
		0.018622,
		0.031171,
		0.047728,
		0.068577,
		0.093895,
		0.123736,
		0.158028,
		0.196569,
		0.239036,
		0.284996,
		0.333919,
		0.385195,
		0.438158,
		0.492103,
		0.546305,
		0.600037,
		0.652581,
		0.703245,
		0.751367,
		0.796331,
		0.837568,
		0.874577,
		0.906936,
		0.93433,
		0.956583,
		0.973698,
		0.985907,
		0.993717,
		0.99793,
		0.99962,
		0.999983,
		0.999998,
		0.999818,
		0.998392,
		0.993706,
		0.983187,
		0.964076,
		0.933754,
		0.890001,
		0.83123,
		0.756699,
		0.66673,
		0.562964,
		0.448658,
		0.32905,
		0.211807,
		0.107582,
		0.0307,
		0.001132,
		0.004895,
		0.011815,
		0.022365,
		0.036943,
		0.055852,
		0.079284,
		0.107308,
		0.139866,
		0.176778,
		0.21774,
		0.262345,
		0.310092,
		0.360406,
		0.412655,
		0.46617,
		0.520259,
		0.574223,
		0.627366,
		0.679005,
		0.728476,
		0.775143,
		0.818404,
		0.857704,
		0.892551,
		0.92254,
		0.94739,
		0.966984,
		0.981426,
		0.991097,
		0.996693,
		0.99923,
		0.999942,
		1,
		0.999937,
		0.999046,
		0.995458,
		0.986574,
		0.969489,
		0.941362,
		0.899725,
		0.842738,
		0.769433,
		0.679948,
		0.575799,
		0.460201,
		0.338462,
		0.218463,
		0.111261,
		0.031834,
		0.00138,
		0.005885,
		0.014018,
		0.026203,
		0.042764,
		0.063915,
		0.089745,
		0.120219,
		0.155176,
		0.194342,
		0.237338,
		0.283695,
		0.332875,
		0.384281,
		0.437279,
		0.491211,
		0.545402,
		0.599175,
		0.651854,
		0.702768,
		0.751262,
		0.796695,
		0.83846,
		0.875992,
		0.908799,
		0.936491,
		0.958834,
		0.975798,
		0.987631,
		0.994906,
		0.998558,
		0.999818,
		0.999999,
		0.999986,
		0.999493,
		0.996876,
		0.989554,
		0.974494,
		0.948631,
		0.909232,
		0.854185,
		0.782268,
		0.69341,
		0.588982,
		0.472143,
		0.34826,
		0.225429,
		0.11513,
		0.033031,
		0.001639,
		0.006903,
		0.016247,
		0.030023,
		0.048471,
		0.071706,
		0.099716,
		0.132372,
		0.169425,
		0.210532,
		0.255258,
		0.3031,
		0.3535,
		0.405858,
		0.459545,
		0.513912,
		0.568298,
		0.622029,
		0.674429,
		0.724819,
		0.772521,
		0.816874,
		0.857243,
		0.893048,
		0.923796,
		0.94913,
		0.96889,
		0.983181,
		0.992443,
		0.997501,
		0.999548,
		0.999986,
		0.999999,
		0.999771,
		0.997978,
		0.992122,
		0.979066,
		0.955526,
		0.918486,
		0.865538,
		0.795181,
		0.707107,
		0.602518,
		0.484498,
		0.358463,
		0.232725,
		0.119202,
		0.034297,
		0.001902,
		0.007919,
		0.018437,
		0.033727,
		0.053931,
		0.07907,
		0.109042,
		0.143636,
		0.182539,
		0.225354,
		0.271617,
		0.320804,
		0.372351,
		0.425657,
		0.480096,
		0.535021,
		0.589761,
		0.643631,
		0.695929,
		0.745942,
		0.792954,
		0.836264,
		0.875207,
		0.909189,
		0.937734,
		0.960548,
		0.97759,
		0.989153,
		0.995933,
		0.999046,
		0.999929,
		1,
		0.99992,
		0.998793,
		0.994282,
		0.983187,
		0.962012,
		0.927446,
		0.876762,
		0.808146,
		0.721027,
		0.61641,
		0.497281,
		0.369091,
		0.24037,
		0.123492,
		0.035637,
		0.002161,
		0.008906,
		0.020541,
		0.037241,
		0.05906,
		0.085931,
		0.117674,
		0.154017,
		0.194598,
		0.238991,
		0.286708,
		0.337219,
		0.38995,
		0.444298,
		0.499627,
		0.555271,
		0.610534,
		0.664693,
		0.717001,
		0.766692,
		0.813001,
		0.855181,
		0.892541,
		0.92449,
		0.9506,
		0.970681,
		0.984868,
		0.993706,
		0.9982,
		0.999771,
		0.999999,
		0.999982,
		0.999355,
		0.996042,
		0.986838,
		0.968053,
		0.93607,
		0.887815,
		0.821134,
		0.735156,
		0.630658,
		0.510503,
		0.380165,
		0.248386,
		0.128015,
		0.037057,
		0.00241,
		0.009847,
		0.022525,
		0.040531,
		0.063833,
		0.09229,
		0.125665,
		0.163637,
		0.205814,
		0.251747,
		0.300936,
		0.352837,
		0.406867,
		0.462401,
		0.518777,
		0.575293,
		0.631206,
		0.685735,
		0.738071,
		0.787387,
		0.832862,
		0.873711,
		0.909231,
		0.938862,
		0.962262,
		0.9794,
		0.99065,
		0.996876,
		0.999429,
		0.999982,
		0.999999,
		0.999707,
		0.997422,
		0.990011,
		0.973615,
		0.944313,
		0.898654,
		0.834112,
		0.749475,
		0.645261,
		0.52418,
		0.391707,
		0.256796,
		0.132788,
		0.038563,
		0.002646,
		0.010731,
		0.02438,
		0.043596,
		0.068277,
		0.098223,
		0.133151,
		0.172707,
		0.216478,
		0.263995,
		0.314743,
		0.368158,
		0.423631,
		0.480503,
		0.538067,
		0.595562,
		0.652181,
		0.707073,
		0.759357,
		0.808142,
		0.852554,
		0.891783,
		0.92514,
		0.952128,
		0.972543,
		0.986574,
		0.994912,
		0.998793,
		0.999909,
		1,
		0.999897,
		0.998451,
		0.9927,
		0.978669,
		0.952128,
		0.909232,
		0.84704,
		0.763962,
		0.660215,
		0.538323,
		0.403741,
		0.265625,
		0.137829,
		0.040162,
		0.002868,
		0.011559,
		0.026116,
		0.046472,
		0.072466,
		0.103854,
		0.140319,
		0.181483,
		0.22691,
		0.276112,
		0.328548,
		0.383622,
		0.440684,
		0.499021,
		0.557864,
		0.616381,
		0.673685,
		0.728847,
		0.780912,
		0.828924,
		0.87197,
		0.909232,
		0.940063,
		0.964076,
		0.981267,
		0.992122,
		0.997721,
		0.999707,
		0.999999,
		0.999977,
		0.999166,
		0.994912,
		0.983187,
		0.959469,
		0.919497,
		0.859876,
		0.778592,
		0.675515,
		0.552945,
		0.416289,
		0.2749,
		0.14316,
		0.041863,
		0.003077,
		0.01234,
		0.027762,
		0.049219,
		0.076504,
		0.109339,
		0.147381,
		0.190228,
		0.237419,
		0.288436,
		0.342703,
		0.399576,
		0.458348,
		0.518237,
		0.578397,
		0.637912,
		0.695809,
		0.751076,
		0.802682,
		0.849616,
		0.890936,
		0.925841,
		0.953766,
		0.974494,
		0.988294,
		0.996042,
		0.999268,
		0.999977,
		0.999998,
		0.999618,
		0.996662,
		0.987147,
		0.966289,
		0.929393,
		0.872571,
		0.793332,
		0.69115,
		0.568056,
		0.429378,
		0.28465,
		0.148801,
		0.043674,
		0.003275,
		0.01309,
		0.029357,
		0.05191,
		0.080506,
		0.114839,
		0.154541,
		0.199181,
		0.248267,
		0.301242,
		0.357477,
		0.416269,
		0.476837,
		0.53832,
		0.599782,
		0.660215,
		0.718555,
		0.773702,
		0.824552,
		0.870045,
		0.909232,
		0.941362,
		0.966004,
		0.983187,
		0.993551,
		0.998451,
		0.999882,
		1,
		0.999864,
		0.997978,
		0.990536,
		0.972543,
		0.938862,
		0.88507,
		0.808146,
		0.707107,
		0.583667,
		0.443032,
		0.294906,
		0.154777,
		0.045605,
		0.003469,
		0.013828,
		0.030947,
		0.054623,
		0.084585,
		0.120501,
		0.161974,
		0.208539,
		0.259662,
		0.314732,
		0.373058,
		0.433863,
		0.496281,
		0.559361,
		0.622067,
		0.683291,
		0.741872,
		0.796619,
		0.846361,
		0.890001,
		0.926607,
		0.955526,
		0.976536,
		0.990011,
		0.997073,
		0.999618,
		0.999998,
		0.99997,
		0.998903,
		0.99335,
		0.978187,
		0.947842,
		0.897313,
		0.82299,
		0.723368,
		0.599783,
		0.457278,
		0.3057,
		0.161115,
		0.047667,
		0.003662,
		0.014576,
		0.032575,
		0.057429,
		0.088841,
		0.12645,
		0.169822,
		0.218452,
		0.271752,
		0.329048,
		0.389572,
		0.452462,
		0.516759,
		0.581406,
		0.645261,
		0.707107,
		0.765677,
		0.819691,
		0.867913,
		0.909232,
		0.942774,
		0.968053,
		0.985151,
		0.994912,
		0.999046,
		0.99997,
		0.999998,
		0.999493,
		0.995601,
		0.983187,
		0.956268,
		0.909232,
		0.837812,
		0.739908,
		0.61641,
		0.472143,
		0.317068,
		0.167843,
		0.049871,
		0.003862,
		0.015352,
		0.034281,
		0.06039,
		0.093355,
		0.132781,
		0.178194,
		0.229034,
		0.28465,
		0.344293,
		0.407111,
		0.472143,
		0.538323,
		0.604481,
		0.669354,
		0.731605,
		0.789856,
		0.842738,
		0.888967,
		0.927446,
		0.957423,
		0.978669,
		0.991705,
		0.997978,
		0.999844,
		1,
		0.999819,
		0.997312,
		0.987511,
		0.964076,
		0.920755,
		0.852555,
		0.756699,
		0.63355,
		0.487653,
		0.32905,
		0.174994,
		0.05223,
		0.004071,
		0.016175,
		0.036097,
		0.063554,
		0.098192,
		0.139572,
		0.187172,
		0.24037,
		0.298442,
		0.360554,
		0.425753,
		0.492972,
		0.561021,
		0.628601,
		0.694316,
		0.756699,
		0.814257,
		0.865538,
		0.909232,
		0.944313,
		0.97023,
		0.987147,
		0.996177,
		0.999493,
		0.999998,
		0.999959,
		0.998528,
		0.991144,
		0.971204,
		0.931803,
		0.867151,
		0.773702,
		0.651201,
		0.503836,
		0.341684,
		0.182603,
		0.054761,
		0.004295,
		0.017057,
		0.03805,
		0.066962,
		0.103402,
		0.146884,
		0.196828,
		0.25254,
		0.313213,
		0.377913,
		0.445578,
		0.515011,
		0.584888,
		0.653763,
		0.720092,
		0.782268,
		0.838682,
		0.887815,
		0.928371,
		0.959469,
		0.980889,
		0.99335,
		0.998733,
		0.999959,
		0.999997,
		0.999313,
		0.994084,
		0.977591,
		0.94229,
		0.881522,
		0.790869,
		0.669354,
		0.520718,
		0.355015,
		0.19071,
		0.057479,
		0.004537,
		0.018011,
		0.040162,
		0.070648,
		0.109032,
		0.154777,
		0.207232,
		0.265625,
		0.32905,
		0.39646,
		0.466665,
		0.538323,
		0.609954,
		0.679948,
		0.746597,
		0.808146,
		0.862877,
		0.909232,
		0.945995,
		0.972543,
		0.989154,
		0.997312,
		0.999788,
		1,
		0.999752,
		0.996348,
		0.983187,
		0.952128,
		0.895582,
		0.808146,
		0.687996,
		0.538323,
		0.369091,
		0.199356,
		0.060404,
		0.004799,
		0.019046,
		0.042454,
		0.074644,
		0.11513,
		0.163312,
		0.218463,
		0.279714,
		0.346048,
		0.416289,
		0.489096,
		0.562964,
		0.636233,
		0.707107,
		0.773702,
		0.834112,
		0.886525,
		0.929393,
		0.961681,
		0.983187,
		0.994912,
		0.999313,
		0.999997,
		0.999944,
		0.997978,
		0.987949,
		0.961225,
		0.909232,
		0.825463,
		0.707107,
		0.556677,
		0.383959,
		0.208589,
		0.063557,
		0.005084,
		0.020173,
		0.044948,
		0.078987,
		0.121749,
		0.172561,
		0.230606,
		0.294906,
		0.364313,
		0.437501,
		0.512956,
		0.588982,
		0.663715,
		0.735156,
		0.801227,
		0.859876,
		0.909232,
		0.947842,
		0.974996,
		0.991144,
		0.99828,
		0.999944,
		0.999996,
		0.999046,
		0.991857,
		0.969489,
		0.922365,
		0.842738,
		0.726654,
		0.575799,
		0.399673,
		0.218463,
		0.066963,
		0.005396,
		0.021402,
		0.047667,
		0.083718,
		0.128949,
		0.182603,
		0.243759,
		0.31131,
		0.383959,
		0.460201,
		0.538323,
		0.61641,
		0.692366,
		0.763962,
		0.828924,
		0.88507,
		0.930531,
		0.964076,
		0.985548,
		0.996348,
		0.999707,
		1,
		0.999651,
		0.994912,
		0.976835,
		0.93486,
		0.859876,
		0.746597,
		0.595706,
		0.416289,
		0.229035,
		0.070648,
		0.005737,
		0.022747,
		0.050639,
		0.088884,
		0.136799,
		0.193529,
		0.258031,
		0.32905,
		0.405109,
		0.484498,
		0.565272,
		0.645261,
		0.722107,
		0.793332,
		0.856465,
		0.909232,
		0.949876,
		0.977591,
		0.993081,
		0.999046,
		0.999995,
		0.99992,
		0.997147,
		0.983187,
		0.946591,
		0.876762,
		0.766878,
		0.61641,
		0.433865,
		0.24037,
		0.074644,
		0.006111,
		0.024223,
		0.053897,
		0.094539,
		0.145377,
		0.205443,
		0.273547,
		0.34826,
		0.427896,
		0.510503,
		0.593863,
		0.675515,
		0.752804,
		0.82299,
		0.883417,
		0.931803,
		0.966673,
		0.987949,
		0.997607,
		0.99992,
		0.999994,
		0.998635,
		0.988485,
		0.957423,
		0.893264,
		0.787425,
		0.637915,
		0.452462,
		0.25254,
		0.078987,
		0.006524,
		0.025848,
		0.057479,
		0.100747,
		0.154777,
		0.218463,
		0.290446,
		0.369091,
		0.452462,
		0.538323,
		0.624138,
		0.707107,
		0.78425,
		0.852555,
		0.909232,
		0.952128,
		0.980326,
		0.994912,
		0.999581,
		1,
		0.999493,
		0.9927,
		0.967218,
		0.909232,
		0.808146,
		0.660215,
		0.472143,
		0.265625,
		0.083718,
		0.006979,
		0.027641,
		0.061429,
		0.107582,
		0.165103,
		0.232725,
		0.308888,
		0.391707,
		0.478955,
		0.568056,
		0.656102,
		0.739908,
		0.816138,
		0.881522,
		0.933236,
		0.969489,
		0.990352,
		0.998635,
		0.999993,
		0.999882,
		0.995843,
		0.975845,
		0.924495,
		0.828924,
		0.683291,
		0.492972,
		0.279714,
		0.088879,
		0.007484,
		0.029626,
		0.065798,
		0.11513,
		0.176478,
		0.248386,
		0.32905,
		0.416289,
		0.507526,
		0.599783,
		0.689715,
		0.773702,
		0.848031,
		0.909232,
		0.95463,
		0.983187,
		0.996571,
		0.999882,
		0.999991,
		0.997978,
		0.983187,
		0.938862,
		0.849616,
		0.707107,
		0.515011,
		0.294891,
		0.094456,
		0.008046,
		0.031834,
		0.070648,
		0.123492,
		0.189047,
		0.265625,
		0.351133,
		0.443032,
		0.538323,
		0.63355,
		0.72486,
		0.808146,
		0.879329,
		0.93486,
		0.972543,
		0.9927,
		0.999381,
		1,
		0.999236,
		0.989154,
		0.952128,
		0.870045,
		0.731605,
		0.538297,
		0.311037,
		0.100129,
		0.008674,
		0.034297,
		0.076051,
		0.132788,
		0.202976,
		0.28465,
		0.375363,
		0.472143,
		0.571482,
		0.669354,
		0.761317,
		0.842738,
		0.909232,
		0.957423,
		0.986146,
		0.997978,
		0.999989,
		0.999819,
		0.993706,
		0.964076,
		0.890001,
		0.756663,
		0.56247,
		0.327031,
		0.104887,
		0.009378,
		0.037057,
		0.082095,
		0.14316,
		0.218463,
		0.3057,
		0.40199,
		0.503836,
		0.607109,
		0.707107,
		0.798713,
		0.876762,
		0.936718,
		0.975845,
		0.994912,
		0.999819,
		0.999986,
		0.996876,
		0.974494,
		0.909189,
		0.781582,
		0.585368,
		0.339533,
		0.106843,
		0.010171,
		0.040162,
		0.088884,
		0.154777,
		0.23574,
		0.32905,
		0.431293,
		0.538323,
		0.645261,
		0.746597,
		0.836467,
		0.909232,
		0.960553,
		0.989154,
		0.999046,
		1,
		0.998793,
		0.983139,
		0.926633,
		0.803188,
		0.600964,
		0.342523,
		0.103703,
		0.011069,
		0.043674,
		0.096544,
		0.167843,
		0.255081,
		0.355015,
		0.463572,
		0.575799,
		0.685901,
		0.787425,
		0.873716,
		0.938862,
		0.9794,
		0.996876,
		0.999982,
		0.999659,
		0.989142,
		0.938519,
		0.81321,
		0.598814,
		0.328938,
		0.093895,
		0.012092,
		0.047667,
		0.105229,
		0.182603,
		0.276811,
		0.383959,
		0.499142,
		0.61641,
		0.728851,
		0.828924,
		0.909232,
		0.964076,
		0.992122,
		0.999659,
		0.9991,
		0.988808,
		0.935426,
		0.797981,
		0.567267,
		0.294361,
		0.077731,
		0.013263,
		0.05223,
		0.11513,
		0.199356,
		0.301315,
		0.416289,
		0.538323,
		0.660215,
		0.773702,
		0.870045,
		0.941362,
		0.983139,
		0.997575,
		0.993865,
		0.97297,
		0.902538,
		0.743241,
		0.5,
		0.240551,
		0.057659,
		0.014613,
		0.057479,
		0.12648,
		0.218463,
		0.32905,
		0.452462,
		0.581406,
		0.707107,
		0.819691,
		0.909189,
		0.967203,
		0.988808,
		0.974912,
		0.927549,
		0.825636,
		0.642924,
		0.401745,
		0.175889,
		0.037339,
		0.016179,
		0.063557,
		0.139573,
		0.24037,
		0.360554,
		0.492972,
		0.628601,
		0.756663,
		0.864779,
		0.938519,
		0.96241,
		0.927549,
		0.839721,
		0.700845,
		0.505937,
		0.288229,
		0.112084,
		0.020197,
		0.018011,
		0.070648,
		0.154777,
		0.265625,
		0.39646,
		0.538297,
		0.679351,
		0.803188,
		0.886448,
		0.902538,
		0.837497,
		0.707107,
		0.540983,
		0.354699,
		0.179781,
		0.05954,
		0.008349,
		0.020173,
		0.078987,
		0.172561,
		0.294891,
		0.437117,
		0.585368,
		0.716734,
		0.797981,
		0.795954,
		0.700845,
		0.542935,
		0.372177,
		0.215674,
		0.09321,
		0.024116,
		0.002141,
		0.022747,
		0.088879,
		0.19336,
		0.327031,
		0.472358,
		0.598814,
		0.666205,
		0.642924,
		0.530799,
		0.372177,
		0.222444,
		0.108744,
		0.036721,
		0.006041,
		0.000173,
		0.025825,
		0.100129,
		0.212988,
		0.342523,
		0.45206,
		0.5,
		0.462908,
		0.354699,
		0.22133,
		0.110603,
		0.041577,
		0.00891,
		0.000475,
		0.028884,
		0.106843,
		0.208583,
		0.294361,
		0.325662,
		0.288229,
		0.202269,
		0.108744,
		0.041877,
		0.00978,
		0.000676,
		0.028801,
		0.093895,
		0.154555,
		0.175889,
		0.148905,
		0.09321,
		0.040099,
		0.00978,
		0.000719,
		0.021807,
		0.057659,
		0.073201,
		0.05954,
		0.031269,
		0.00891,
		0.000712,
		0.010604,
		0.020197,
		0.016081,
		0.006041,
		0.000596,
		0.002407,
		0.002141,
		0.000326,
		5.1e-005,
		5.1e-005,
		0.000326,
		0.002141,
		0.002407,
		0.000596,
		0.006041,
		0.016081,
		0.020197,
		0.010604,
		0.000712,
		0.00891,
		0.031269,
		0.05954,
		0.073201,
		0.057659,
		0.021807,
		0.000719,
		0.00978,
		0.040099,
		0.09321,
		0.148905,
		0.175889,
		0.154555,
		0.093895,
		0.028801,
		0.000676,
		0.00978,
		0.041877,
		0.108744,
		0.202269,
		0.288229,
		0.325662,
		0.294361,
		0.208583,
		0.106843,
		0.028884,
		0.000475,
		0.00891,
		0.041577,
		0.110603,
		0.22133,
		0.354699,
		0.462908,
		0.5,
		0.45206,
		0.342523,
		0.212988,
		0.100129,
		0.025825,
		0.000173,
		0.006041,
		0.036721,
		0.108744,
		0.222444,
		0.372177,
		0.530799,
		0.642924,
		0.666205,
		0.598814,
		0.472358,
		0.327031,
		0.19336,
		0.088879,
		0.022747,
		0.002141,
		0.024116,
		0.09321,
		0.215674,
		0.372177,
		0.542935,
		0.700845,
		0.795954,
		0.797981,
		0.716734,
		0.585368,
		0.437117,
		0.294891,
		0.172561,
		0.078987,
		0.020173,
		0.008349,
		0.05954,
		0.179781,
		0.354699,
		0.540983,
		0.707107,
		0.837497,
		0.902538,
		0.886448,
		0.803188,
		0.679351,
		0.538297,
		0.39646,
		0.265625,
		0.154777,
		0.070648,
		0.018011,
		0.020197,
		0.112084,
		0.288229,
		0.505937,
		0.700845,
		0.839721,
		0.927549,
		0.96241,
		0.938519,
		0.864779,
		0.756663,
		0.628601,
		0.492972,
		0.360554,
		0.24037,
		0.139573,
		0.063557,
		0.016179,
		0.037339,
		0.175889,
		0.401745,
		0.642924,
		0.825636,
		0.927549,
		0.974912,
		0.988808,
		0.967203,
		0.909189,
		0.819691,
		0.707107,
		0.581406,
		0.452462,
		0.32905,
		0.218463,
		0.12648,
		0.057479,
		0.014613,
		0.057659,
		0.240551,
		0.5,
		0.743241,
		0.902538,
		0.97297,
		0.993865,
		0.997575,
		0.983139,
		0.941362,
		0.870045,
		0.773702,
		0.660215,
		0.538323,
		0.416289,
		0.301315,
		0.199356,
		0.11513,
		0.05223,
		0.013263,
		0.077731,
		0.294361,
		0.567267,
		0.797981,
		0.935426,
		0.988808,
		0.9991,
		0.999659,
		0.992122,
		0.964076,
		0.909232,
		0.828924,
		0.728851,
		0.61641,
		0.499142,
		0.383959,
		0.276811,
		0.182603,
		0.105229,
		0.047667,
		0.012092,
		0.093895,
		0.328938,
		0.598814,
		0.81321,
		0.938519,
		0.989142,
		0.999659,
		0.999982,
		0.996876,
		0.9794,
		0.938862,
		0.873716,
		0.787425,
		0.685901,
		0.575799,
		0.463572,
		0.355015,
		0.255081,
		0.167843,
		0.096544,
		0.043674,
		0.011069,
		0.103703,
		0.342523,
		0.600964,
		0.803188,
		0.926633,
		0.983139,
		0.998793,
		1,
		0.999046,
		0.989154,
		0.960553,
		0.909232,
		0.836467,
		0.746597,
		0.645261,
		0.538323,
		0.431293,
		0.32905,
		0.23574,
		0.154777,
		0.088884,
		0.040162,
		0.010171,
		0.106843,
		0.339533,
		0.585368,
		0.781582,
		0.909189,
		0.974494,
		0.996876,
		0.999986,
		0.999819,
		0.994912,
		0.975845,
		0.936718,
		0.876762,
		0.798713,
		0.707107,
		0.607109,
		0.503836,
		0.40199,
		0.3057,
		0.218463,
		0.14316,
		0.082095,
		0.037057,
		0.009378,
		0.104887,
		0.327031,
		0.56247,
		0.756663,
		0.890001,
		0.964076,
		0.993706,
		0.999819,
		0.999989,
		0.997978,
		0.986146,
		0.957423,
		0.909232,
		0.842738,
		0.761317,
		0.669354,
		0.571482,
		0.472143,
		0.375363,
		0.28465,
		0.202976,
		0.132788,
		0.076051,
		0.034297,
		0.008674,
		0.100129,
		0.311037,
		0.538297,
		0.731605,
		0.870045,
		0.952128,
		0.989154,
		0.999236,
		1,
		0.999381,
		0.9927,
		0.972543,
		0.93486,
		0.879329,
		0.808146,
		0.72486,
		0.63355,
		0.538323,
		0.443032,
		0.351133,
		0.265625,
		0.189047,
		0.123492,
		0.070648,
		0.031834,
		0.008046,
		0.094456,
		0.294891,
		0.515011,
		0.707107,
		0.849616,
		0.938862,
		0.983187,
		0.997978,
		0.999991,
		0.999882,
		0.996571,
		0.983187,
		0.95463,
		0.909232,
		0.848031,
		0.773702,
		0.689715,
		0.599783,
		0.507526,
		0.416289,
		0.32905,
		0.248386,
		0.176478,
		0.11513,
		0.065798,
		0.029626,
		0.007484,
		0.088879,
		0.279714,
		0.492972,
		0.683291,
		0.828924,
		0.924495,
		0.975845,
		0.995843,
		0.999882,
		0.999993,
		0.998635,
		0.990352,
		0.969489,
		0.933236,
		0.881522,
		0.816138,
		0.739908,
		0.656102,
		0.568056,
		0.478955,
		0.391707,
		0.308888,
		0.232725,
		0.165103,
		0.107582,
		0.061429,
		0.027641,
		0.006979,
		0.083718,
		0.265625,
		0.472143,
		0.660215,
		0.808146,
		0.909232,
		0.967218,
		0.9927,
		0.999493,
		1,
		0.999581,
		0.994912,
		0.980326,
		0.952128,
		0.909232,
		0.852555,
		0.78425,
		0.707107,
		0.624138,
		0.538323,
		0.452462,
		0.369091,
		0.290446,
		0.218463,
		0.154777,
		0.100747,
		0.057479,
		0.025848,
		0.006524,
		0.078987,
		0.25254,
		0.452462,
		0.637915,
		0.787425,
		0.893264,
		0.957423,
		0.988485,
		0.998635,
		0.999994,
		0.99992,
		0.997607,
		0.987949,
		0.966673,
		0.931803,
		0.883417,
		0.82299,
		0.752804,
		0.675515,
		0.593863,
		0.510503,
		0.427896,
		0.34826,
		0.273547,
		0.205443,
		0.145377,
		0.094539,
		0.053897,
		0.024223,
		0.006111,
		0.074644,
		0.24037,
		0.433865,
		0.61641,
		0.766878,
		0.876762,
		0.946591,
		0.983187,
		0.997147,
		0.99992,
		0.999995,
		0.999046,
		0.993081,
		0.977591,
		0.949876,
		0.909232,
		0.856465,
		0.793332,
		0.722107,
		0.645261,
		0.565272,
		0.484498,
		0.405109,
		0.32905,
		0.258031,
		0.193529,
		0.136799,
		0.088884,
		0.050639,
		0.022747,
		0.005737,
		0.070648,
		0.229035,
		0.416289,
		0.595706,
		0.746597,
		0.859876,
		0.93486,
		0.976835,
		0.994912,
		0.999651,
		1,
		0.999707,
		0.996348,
		0.985548,
		0.964076,
		0.930531,
		0.88507,
		0.828924,
		0.763962,
		0.692366,
		0.61641,
		0.538323,
		0.460201,
		0.383959,
		0.31131,
		0.243759,
		0.182603,
		0.128949,
		0.083718,
		0.047667,
		0.021402,
		0.005396,
		0.066963,
		0.218463,
		0.399673,
		0.575799,
		0.726654,
		0.842738,
		0.922365,
		0.969489,
		0.991857,
		0.999046,
		0.999996,
		0.999944,
		0.99828,
		0.991144,
		0.974996,
		0.947842,
		0.909232,
		0.859876,
		0.801227,
		0.735156,
		0.663715,
		0.588982,
		0.512956,
		0.437501,
		0.364313,
		0.294906,
		0.230606,
		0.172561,
		0.121749,
		0.078987,
		0.044948,
		0.020173,
		0.005084,
		0.063557,
		0.208589,
		0.383959,
		0.556677,
		0.707107,
		0.825463,
		0.909232,
		0.961225,
		0.987949,
		0.997978,
		0.999944,
		0.999997,
		0.999313,
		0.994912,
		0.983187,
		0.961681,
		0.929393,
		0.886525,
		0.834112,
		0.773702,
		0.707107,
		0.636233,
		0.562964,
		0.489096,
		0.416289,
		0.346048,
		0.279714,
		0.218463,
		0.163312,
		0.11513,
		0.074644,
		0.042454,
		0.019046,
		0.004799,
		0.060404,
		0.199356,
		0.369091,
		0.538323,
		0.687996,
		0.808146,
		0.895582,
		0.952128,
		0.983187,
		0.996348,
		0.999752,
		1,
		0.999788,
		0.997312,
		0.989154,
		0.972543,
		0.945995,
		0.909232,
		0.862877,
		0.808146,
		0.746597,
		0.679948,
		0.609954,
		0.538323,
		0.466665,
		0.39646,
		0.32905,
		0.265625,
		0.207232,
		0.154777,
		0.109032,
		0.070648,
		0.040162,
		0.018011,
		0.004537,
		0.057479,
		0.19071,
		0.355015,
		0.520718,
		0.669354,
		0.790869,
		0.881522,
		0.94229,
		0.977591,
		0.994084,
		0.999313,
		0.999997,
		0.999959,
		0.998733,
		0.99335,
		0.980889,
		0.959469,
		0.928371,
		0.887815,
		0.838682,
		0.782268,
		0.720092,
		0.653763,
		0.584888,
		0.515011,
		0.445578,
		0.377913,
		0.313213,
		0.25254,
		0.196828,
		0.146884,
		0.103402,
		0.066962,
		0.03805,
		0.017057,
		0.004295,
		0.054761,
		0.182603,
		0.341684,
		0.503836,
		0.651201,
		0.773702,
		0.867151,
		0.931803,
		0.971204,
		0.991144,
		0.998528,
		0.999959,
		0.999998,
		0.999493,
		0.996177,
		0.987147,
		0.97023,
		0.944313,
		0.909232,
		0.865538,
		0.814257,
		0.756699,
		0.694316,
		0.628601,
		0.561021,
		0.492972,
		0.425753,
		0.360554,
		0.298442,
		0.24037,
		0.187172,
		0.139572,
		0.098192,
		0.063554,
		0.036097,
		0.016175,
		0.004071,
		0.05223,
		0.174994,
		0.32905,
		0.487653,
		0.63355,
		0.756699,
		0.852555,
		0.920755,
		0.964076,
		0.987511,
		0.997312,
		0.999819,
		1,
		0.999844,
		0.997978,
		0.991705,
		0.978669,
		0.957423,
		0.927446,
		0.888967,
		0.842738,
		0.789856,
		0.731605,
		0.669354,
		0.604481,
		0.538323,
		0.472143,
		0.407111,
		0.344293,
		0.28465,
		0.229034,
		0.178194,
		0.132781,
		0.093355,
		0.06039,
		0.034281,
		0.015352,
		0.003862,
		0.049871,
		0.167843,
		0.317068,
		0.472143,
		0.61641,
		0.739908,
		0.837812,
		0.909232,
		0.956268,
		0.983187,
		0.995601,
		0.999493,
		0.999998,
		0.99997,
		0.999046,
		0.994912,
		0.985151,
		0.968053,
		0.942774,
		0.909232,
		0.867913,
		0.819691,
		0.765677,
		0.707107,
		0.645261,
		0.581406,
		0.516759,
		0.452462,
		0.389572,
		0.329048,
		0.271752,
		0.218452,
		0.169822,
		0.12645,
		0.088841,
		0.057429,
		0.032575,
		0.014576,
		0.003662,
		0.047667,
		0.161115,
		0.3057,
		0.457278,
		0.599783,
		0.723368,
		0.82299,
		0.897313,
		0.947842,
		0.978187,
		0.99335,
		0.998903,
		0.99997,
		0.999998,
		0.999618,
		0.997073,
		0.990011,
		0.976536,
		0.955526,
		0.926607,
		0.890001,
		0.846361,
		0.796619,
		0.741872,
		0.683291,
		0.622067,
		0.559361,
		0.496281,
		0.433863,
		0.373058,
		0.314732,
		0.259662,
		0.208539,
		0.161974,
		0.120501,
		0.084585,
		0.054623,
		0.030947,
		0.013828,
		0.003469,
		0.045605,
		0.154777,
		0.294906,
		0.443032,
		0.583667,
		0.707107,
		0.808146,
		0.88507,
		0.938862,
		0.972543,
		0.990536,
		0.997978,
		0.999864,
		1,
		0.999882,
		0.998451,
		0.993551,
		0.983187,
		0.966004,
		0.941362,
		0.909232,
		0.870045,
		0.824552,
		0.773702,
		0.718555,
		0.660215,
		0.599782,
		0.53832,
		0.476837,
		0.416269,
		0.357477,
		0.301242,
		0.248267,
		0.199181,
		0.154541,
		0.114839,
		0.080506,
		0.05191,
		0.029357,
		0.01309,
		0.003275,
		0.043674,
		0.148801,
		0.28465,
		0.429378,
		0.568056,
		0.69115,
		0.793332,
		0.872571,
		0.929393,
		0.966289,
		0.987147,
		0.996662,
		0.999618,
		0.999998,
		0.999977,
		0.999268,
		0.996042,
		0.988294,
		0.974494,
		0.953766,
		0.925841,
		0.890936,
		0.849616,
		0.802682,
		0.751076,
		0.695809,
		0.637912,
		0.578397,
		0.518237,
		0.458348,
		0.399576,
		0.342703,
		0.288436,
		0.237419,
		0.190228,
		0.147381,
		0.109339,
		0.076504,
		0.049219,
		0.027762,
		0.01234,
		0.003077,
		0.041863,
		0.14316,
		0.2749,
		0.416289,
		0.552945,
		0.675515,
		0.778592,
		0.859876,
		0.919497,
		0.959469,
		0.983187,
		0.994912,
		0.999166,
		0.999977,
		0.999999,
		0.999707,
		0.997721,
		0.992122,
		0.981267,
		0.964076,
		0.940063,
		0.909232,
		0.87197,
		0.828924,
		0.780912,
		0.728847,
		0.673685,
		0.616381,
		0.557864,
		0.499021,
		0.440684,
		0.383622,
		0.328548,
		0.276112,
		0.22691,
		0.181483,
		0.140319,
		0.103854,
		0.072466,
		0.046472,
		0.026116,
		0.011559,
		0.002868,
		0.040162,
		0.137829,
		0.265625,
		0.403741,
		0.538323,
		0.660215,
		0.763962,
		0.84704,
		0.909232,
		0.952128,
		0.978669,
		0.9927,
		0.998451,
		0.999897,
		1,
		0.999909,
		0.998793,
		0.994912,
		0.986574,
		0.972543,
		0.952128,
		0.92514,
		0.891783,
		0.852554,
		0.808142,
		0.759357,
		0.707073,
		0.652181,
		0.595562,
		0.538067,
		0.480503,
		0.423631,
		0.368158,
		0.314743,
		0.263995,
		0.216478,
		0.172707,
		0.133151,
		0.098223,
		0.068277,
		0.043596,
		0.02438,
		0.010731,
		0.002646,
		0.038563,
		0.132788,
		0.256796,
		0.391707,
		0.52418,
		0.645261,
		0.749475,
		0.834112,
		0.898654,
		0.944313,
		0.973615,
		0.990011,
		0.997422,
		0.999707,
		0.999999,
		0.999982,
		0.999429,
		0.996876,
		0.99065,
		0.9794,
		0.962262,
		0.938862,
		0.909231,
		0.873711,
		0.832862,
		0.787387,
		0.738071,
		0.685735,
		0.631206,
		0.575293,
		0.518777,
		0.462401,
		0.406867,
		0.352837,
		0.300936,
		0.251747,
		0.205814,
		0.163637,
		0.125665,
		0.09229,
		0.063833,
		0.040531,
		0.022525,
		0.009847,
		0.00241,
		0.037057,
		0.128015,
		0.248386,
		0.380165,
		0.510503,
		0.630658,
		0.735156,
		0.821134,
		0.887815,
		0.93607,
		0.968053,
		0.986838,
		0.996042,
		0.999355,
		0.999982,
		0.999999,
		0.999771,
		0.9982,
		0.993706,
		0.984868,
		0.970681,
		0.9506,
		0.92449,
		0.892541,
		0.855181,
		0.813001,
		0.766692,
		0.717001,
		0.664693,
		0.610534,
		0.555271,
		0.499627,
		0.444298,
		0.38995,
		0.337219,
		0.286708,
		0.238991,
		0.194598,
		0.154017,
		0.117674,
		0.085931,
		0.05906,
		0.037241,
		0.020541,
		0.008906,
		0.002161,
		0.035637,
		0.123492,
		0.24037,
		0.369091,
		0.497281,
		0.61641,
		0.721027,
		0.808146,
		0.876762,
		0.927446,
		0.962012,
		0.983187,
		0.994282,
		0.998793,
		0.99992,
		1,
		0.999929,
		0.999046,
		0.995933,
		0.989153,
		0.97759,
		0.960548,
		0.937734,
		0.909189,
		0.875207,
		0.836264,
		0.792954,
		0.745942,
		0.695929,
		0.643631,
		0.589761,
		0.535021,
		0.480096,
		0.425657,
		0.372351,
		0.320804,
		0.271617,
		0.225354,
		0.182539,
		0.143636,
		0.109042,
		0.07907,
		0.053931,
		0.033727,
		0.018437,
		0.007919,
		0.001902,
		0.034297,
		0.119202,
		0.232725,
		0.358463,
		0.484498,
		0.602518,
		0.707107,
		0.795181,
		0.865538,
		0.918486,
		0.955526,
		0.979066,
		0.992122,
		0.997978,
		0.999771,
		0.999999,
		0.999986,
		0.999548,
		0.997501,
		0.992443,
		0.983181,
		0.96889,
		0.94913,
		0.923796,
		0.893048,
		0.857243,
		0.816874,
		0.772521,
		0.724819,
		0.674429,
		0.622029,
		0.568298,
		0.513912,
		0.459545,
		0.405858,
		0.3535,
		0.3031,
		0.255258,
		0.210532,
		0.169425,
		0.132372,
		0.099716,
		0.071706,
		0.048471,
		0.030023,
		0.016247,
		0.006903,
		0.001639,
		0.033031,
		0.11513,
		0.225429,
		0.34826,
		0.472143,
		0.588982,
		0.69341,
		0.782268,
		0.854185,
		0.909232,
		0.948631,
		0.974494,
		0.989554,
		0.996876,
		0.999493,
		0.999986,
		0.999999,
		0.999818,
		0.998558,
		0.994906,
		0.987631,
		0.975798,
		0.958834,
		0.936491,
		0.908799,
		0.875992,
		0.83846,
		0.796695,
		0.751262,
		0.702768,
		0.651854,
		0.599175,
		0.545402,
		0.491211,
		0.437279,
		0.384281,
		0.332875,
		0.283695,
		0.237338,
		0.194342,
		0.155176,
		0.120219,
		0.089745,
		0.063915,
		0.042764,
		0.026203,
		0.014018,
		0.005885,
		0.00138,
		0.031834,
		0.111261,
		0.218463,
		0.338462,
		0.460201,
		0.575799,
		0.679948,
		0.769433,
		0.842738,
		0.899725,
		0.941362,
		0.969489,
		0.986574,
		0.995458,
		0.999046,
		0.999937,
		1,
		0.999942,
		0.99923,
		0.996693,
		0.991097,
		0.981426,
		0.966984,
		0.94739,
		0.92254,
		0.892551,
		0.857704,
		0.818404,
		0.775143,
		0.728476,
		0.679005,
		0.627366,
		0.574223,
		0.520259,
		0.46617,
		0.412655,
		0.360406,
		0.310092,
		0.262345,
		0.21774,
		0.176778,
		0.139866,
		0.107308,
		0.079284,
		0.055852,
		0.036943,
		0.022365,
		0.011815,
		0.004895,
		0.001132,
		0.0307,
		0.107582,
		0.211807,
		0.32905,
		0.448658,
		0.562964,
		0.66673,
		0.756699,
		0.83123,
		0.890001,
		0.933754,
		0.964076,
		0.983187,
		0.993706,
		0.998392,
		0.999818,
		0.999998,
		0.999983,
		0.99962,
		0.99793,
		0.993717,
		0.985907,
		0.973698,
		0.956583,
		0.93433,
		0.906936,
		0.874577,
		0.837568,
		0.796331,
		0.751367,
		0.703245,
		0.652581,
		0.600037,
		0.546305,
		0.492103,
		0.438158,
		0.385195,
		0.333919,
		0.284996,
		0.239036,
		0.196569,
		0.158028,
		0.123736,
		0.093895,
		0.068577,
		0.047728,
		0.031171,
		0.018622,
		0.009702,
		0.003962,
		0.000902,
		0.029626,
		0.104082,
		0.205443,
		0.320005,
		0.437501,
		0.550474,
		0.653763,
		0.744084,
		0.819691,
		0.880097,
		0.925841,
		0.95828,
		0.9794,
		0.991607,
		0.997501,
		0.999595,
		0.999983,
		0.999981,
		0.999806,
		0.998721,
		0.995602,
		0.989349,
		0.979067,
		0.964136,
		0.9442,
		0.919139,
		0.889025,
		0.854087,
		0.814675,
		0.771243,
		0.724327,
		0.674529,
		0.622512,
		0.568982,
		0.514679,
		0.460362,
		0.40679,
		0.354703,
		0.304801,
		0.257723,
		0.214025,
		0.174161,
		0.138472,
		0.107171,
		0.080339,
		0.057933,
		0.039786,
		0.025625,
		0.015087,
		0.007742,
		0.003111,
		0.000697,
		0.028608,
		0.100747,
		0.199356,
		0.31131,
		0.426716,
		0.538323,
		0.641052,
		0.731605,
		0.808146,
		0.870045,
		0.917657,
		0.952128,
		0.975226,
		0.989153,
		0.996346,
		0.99923,
		0.999932,
		0.999952,
		0.999841,
		0.999139,
		0.996836,
		0.991829,
		0.983155,
		0.970086,
		0.952154,
		0.929125,
		0.900971,
		0.867837,
		0.830009,
		0.787895,
		0.742003,
		0.692926,
		0.641333,
		0.587947,
		0.533535,
		0.478887,
		0.424797,
		0.372038,
		0.321343,
		0.273376,
		0.228713,
		0.187825,
		0.151061,
		0.118641,
		0.090652,
		0.067052,
		0.04768,
		0.032269,
		0.020468,
		0.011859,
		0.005983,
		0.002362,
		0.000519,
		0.027641,
		0.09757,
		0.193529,
		0.302949,
		0.416289,
		0.526505,
		0.628601,
		0.719275,
		0.796619,
		0.859876,
		0.909232,
		0.945646,
		0.970681,
		0.986344,
		0.994906,
		0.998688,
		0.999806,
		0.999886,
		0.999749,
		0.99923,
		0.997471,
		0.993394,
		0.985988,
		0.974436,
		0.958161,
		0.936827,
		0.91031,
		0.878677,
		0.842154,
		0.801108,
		0.756027,
		0.707502,
		0.656209,
		0.602893,
		0.548352,
		0.493408,
		0.438887,
		0.385597,
		0.334294,
		0.285665,
		0.240301,
		0.198682,
		0.161159,
		0.127953,
		0.099147,
		0.074699,
		0.054447,
		0.038126,
		0.025391,
		0.015835,
		0.009012,
		0.004461,
		0.001726,
		0.000371,
		0.026722,
		0.094539,
		0.18795,
		0.294906,
		0.406208,
		0.515011,
		0.61641,
		0.707107,
		0.785132,
		0.849616,
		0.900598,
		0.938862,
		0.965783,
		0.983181,
		0.993163,
		0.99793,
		0.999561,
		0.999749,
		0.999523,
		0.999005,
		0.997522,
		0.994054,
		0.987562,
		0.977154,
		0.962163,
		0.942154,
		0.916918,
		0.886448,
		0.850921,
		0.810671,
		0.766169,
		0.71801,
		0.666884,
		0.613563,
		0.558873,
		0.50367,
		0.448814,
		0.395137,
		0.343419,
		0.294361,
		0.248563,
		0.206504,
		0.168534,
		0.134865,
		0.105574,
		0.080608,
		0.059802,
		0.04289,
		0.029527,
		0.019315,
		0.011818,
		0.006591,
		0.003192,
		0.001206,
		0.000253,
		0.025848,
		0.091646,
		0.182603,
		0.287165,
		0.39646,
		0.503836,
		0.604481,
		0.695109,
		0.773702,
		0.839291,
		0.891783,
		0.931802,
		0.960548,
		0.979666,
		0.991097,
		0.996917,
		0.999139,
		0.999483,
		0.999123,
		0.998438,
		0.996969,
		0.993781,
		0.987832,
		0.978178,
		0.964069,
		0.944988,
		0.92065,
		0.890985,
		0.856127,
		0.816383,
		0.772221,
		0.724239,
		0.673152,
		0.619756,
		0.564909,
		0.5095,
		0.454416,
		0.400512,
		0.348584,
		0.299339,
		0.253379,
		0.211175,
		0.173066,
		0.139251,
		0.109792,
		0.084624,
		0.063573,
		0.046366,
		0.03266,
		0.02206,
		0.014139,
		0.008464,
		0.00461,
		0.002176,
		0.000799,
		0.000162,
		0.025016,
		0.088884,
		0.177478,
		0.279714,
		0.387032,
		0.492972,
		0.592813,
		0.683291,
		0.762345,
		0.828924,
		0.882813,
		0.92449,
		0.954993,
		0.975798,
		0.988688,
		0.995602,
		0.99847,
		0.999005,
		0.998474,
		0.997467,
		0.995751,
		0.992508,
		0.986721,
		0.977409,
		0.963764,
		0.945195,
		0.921354,
		0.892122,
		0.857596,
		0.818069,
		0.774009,
		0.726028,
		0.674863,
		0.621342,
		0.566352,
		0.510811,
		0.455628,
		0.401677,
		0.349758,
		0.300582,
		0.254737,
		0.212684,
		0.174741,
		0.141088,
		0.111767,
		0.086698,
		0.06569,
		0.048465,
		0.034677,
		0.023933,
		0.015814,
		0.009899,
		0.005775,
		0.003057,
		0.001398,
		0.000496,
		9.7e-005,
		0.024223,
		0.086243,
		0.172561,
		0.272538,
		0.377913,
		0.48241,
		0.581406,
		0.671658,
		0.751076,
		0.818535,
		0.873711,
		0.916947,
		0.94913,
		0.971575,
		0.985907,
		0.993929,
		0.997471,
		0.99821,
		0.997467,
		0.99599,
		0.993768,
		0.990133,
		0.984115,
		0.974726,
		0.961111,
		0.942628,
		0.918878,
		0.889702,
		0.855177,
		0.815588,
		0.771408,
		0.723272,
		0.67194,
		0.618271,
		0.563181,
		0.507609,
		0.452485,
		0.398686,
		0.347016,
		0.298172,
		0.252728,
		0.211121,
		0.173644,
		0.140451,
		0.11156,
		0.086869,
		0.066174,
		0.049186,
		0.035553,
		0.024886,
		0.016772,
		0.0108,
		0.00657,
		0.003714,
		0.001898,
		0.000834,
		0.000283,
		5.2e-005,
		0.023468,
		0.083718,
		0.167843,
		0.265625,
		0.369091,
		0.472143,
		0.570256,
		0.660215,
		0.739907,
		0.808142,
		0.864496,
		0.909189,
		0.942964,
		0.966984,
		0.982718,
		0.991829,
		0.996039,
		0.996969,
		0.995957,
		0.993865,
		0.990884,
		0.986518,
		0.979875,
		0.969981,
		0.955961,
		0.937137,
		0.913075,
		0.883593,
		0.848753,
		0.808844,
		0.764353,
		0.715937,
		0.664385,
		0.610581,
		0.555467,
		0.5,
		0.445116,
		0.391695,
		0.340527,
		0.292291,
		0.247536,
		0.206667,
		0.169947,
		0.137498,
		0.109311,
		0.08526,
		0.065122,
		0.0486,
		0.035336,
		0.024941,
		0.017011,
		0.011143,
		0.006952,
		0.004083,
		0.002217,
		0.001083,
		0.000451,
		0.000144,
		2.5e-005,
		0.022747,
		0.081301,
		0.163312,
		0.258963,
		0.360554,
		0.462163,
		0.559361,
		0.648966,
		0.728847,
		0.797756,
		0.855181,
		0.901223,
		0.936491,
		0.961999,
		0.979067,
		0.989217,
		0.994054,
		0.995127,
		0.993768,
		0.990914,
		0.986925,
		0.981492,
		0.973833,
		0.963014,
		0.948161,
		0.928581,
		0.903823,
		0.873692,
		0.838252,
		0.7978,
		0.752845,
		0.704067,
		0.652282,
		0.598399,
		0.543376,
		0.488182,
		0.43375,
		0.38095,
		0.330552,
		0.283205,
		0.239425,
		0.19958,
		0.163895,
		0.132457,
		0.105226,
		0.082051,
		0.062693,
		0.046839,
		0.034131,
		0.02418,
		0.016587,
		0.010961,
		0.00693,
		0.004154,
		0.002331,
		0.0012,
		0.000551,
		0.000213,
		6.2e-005,
		0.022059,
		0.078987,
		0.15896,
		0.25254,
		0.352291,
		0.452462,
		0.548717,
		0.637912,
		0.717902,
		0.787387,
		0.845772,
		0.893048,
		0.929694,
		0.956583,
		0.974887,
		0.985988,
		0.991373,
		0.992508,
		0.990698,
		0.986925,
		0.981687,
		0.974863,
		0.965809,
		0.953657,
		0.937562,
		0.916836,
		0.891026,
		0.859942,
		0.823657,
		0.782486,
		0.736961,
		0.687787,
		0.635805,
		0.581942,
		0.527165,
		0.472443,
		0.4187,
		0.366781,
		0.317429,
		0.271253,
		0.228728,
		0.190179,
		0.155789,
		0.125608,
		0.099561,
		0.077474,
		0.059088,
		0.04408,
		0.032087,
		0.022723,
		0.015596,
		0.010326,
		0.006556,
		0.00396,
		0.002251,
		0.001186,
		0.000567,
		0.000237,
		8.2e-005,
		2.1e-005,
		0.021402,
		0.07677,
		0.154777,
		0.246346,
		0.344293,
		0.443032,
		0.53832,
		0.627051,
		0.707073,
		0.777035,
		0.836264,
		0.884648,
		0.92254,
		0.950676,
		0.970086,
		0.982016,
		0.987832,
		0.988911,
		0.986518,
		0.981661,
		0.974941,
		0.966419,
		0.955611,
		0.941747,
		0.924032,
		0.901806,
		0.874633,
		0.842338,
		0.805015,
		0.763003,
		0.716858,
		0.667309,
		0.615215,
		0.561514,
		0.507174,
		0.453154,
		0.400354,
		0.34959,
		0.30156,
		0.256833,
		0.21583,
		0.178832,
		0.145975,
		0.117269,
		0.092608,
		0.071791,
		0.054541,
		0.040526,
		0.029379,
		0.020718,
		0.01416,
		0.009337,
		0.005905,
		0.003555,
		0.002017,
		0.001064,
		0.000512,
		0.000218,
		7.9e-005,
		2.2e-005,
		0.020774,
		0.074644,
		0.150754,
		0.24037,
		0.336549,
		0.433863,
		0.528163,
		0.616381,
		0.696355,
		0.766692,
		0.826642,
		0.875992,
		0.914977,
		0.9442,
		0.964555,
		0.977154,
		0.983245,
		0.984115,
		0.980982,
		0.974863,
		0.966443,
		0.95594,
		0.943052,
		0.927134,
		0.907465,
		0.883435,
		0.85464,
		0.820937,
		0.782445,
		0.73953,
		0.692773,
		0.642924,
		0.590853,
		0.537498,
		0.48382,
		0.430753,
		0.379165,
		0.32983,
		0.283397,
		0.240381,
		0.201151,
		0.165934,
		0.134822,
		0.107782,
		0.084677,
		0.06528,
		0.049299,
		0.036394,
		0.026196,
		0.018328,
		0.012416,
		0.008105,
		0.005068,
		0.003011,
		0.001682,
		0.000871,
		0.00041,
		0.00017,
		6e-005,
		1.6e-005,
		0.020173,
		0.072605,
		0.146884,
		0.234602,
		0.329048,
		0.424946,
		0.518237,
		0.605889,
		0.685735,
		0.756338,
		0.816874,
		0.867033,
		0.906936,
		0.937057,
		0.958161,
		0.971234,
		0.977409,
		0.977887,
		0.973833,
		0.966268,
		0.95594,
		0.943208,
		0.927957,
		0.909691,
		0.887791,
		0.861713,
		0.831105,
		0.795861,
		0.756135,
		0.712321,
		0.665023,
		0.615002,
		0.563133,
		0.510346,
		0.457578,
		0.405731,
		0.35563,
		0.307995,
		0.26342,
		0.22236,
		0.185128,
		0.151896,
		0.122709,
		0.097494,
		0.076083,
		0.058225,
		0.043615,
		0.031906,
		0.02273,
		0.015716,
		0.010502,
		0.006746,
		0.004136,
		0.0024,
		0.001301,
		0.000649,
		0.00029,
		0.000112,
		3.5e-005,
		0.019597,
		0.070648,
		0.143159,
		0.229034,
		0.321779,
		0.416269,
		0.508529,
		0.595562,
		0.675192,
		0.745942,
		0.806913,
		0.857704,
		0.898322,
		0.929125,
		0.950751,
		0.964069,
		0.970107,
		0.969981,
		0.96481,
		0.955611,
		0.943187,
		0.928019,
		0.910174,
		0.889329,
		0.864986,
		0.836688,
		0.804146,
		0.767303,
		0.726348,
		0.681705,
		0.633993,
		0.58398,
		0.532534,
		0.480563,
		0.428973,
		0.37862,
		0.330276,
		0.284601,
		0.242125,
		0.203242,
		0.168203,
		0.137128,
		0.110012,
		0.086747,
		0.067132,
		0.0509,
		0.03773,
		0.027273,
		0.019164,
		0.013038,
		0.008547,
		0.005365,
		0.003197,
		0.00179,
		0.000926,
		0.000434,
		0.000178,
		6e-005,
		1.5e-005,
		0.019046,
		0.068768,
		0.139572,
		0.223654,
		0.314732,
		0.407819,
		0.499021,
		0.585374,
		0.664693,
		0.735456,
		0.796695,
		0.847917,
		0.889025,
		0.920262,
		0.942154,
		0.955459,
		0.961111,
		0.960152,
		0.953657,
		0.942641,
		0.927957,
		0.910193,
		0.889586,
		0.865999,
		0.839077,
		0.808465,
		0.773948,
		0.735522,
		0.693416,
		0.648076,
		0.600132,
		0.55035,
		0.499578,
		0.448693,
		0.398556,
		0.349969,
		0.303641,
		0.260166,
		0.220006,
		0.183489,
		0.150806,
		0.122023,
		0.097092,
		0.075864,
		0.058116,
		0.04356,
		0.031869,
		0.022689,
		0.015662,
		0.010433,
		0.006667,
		0.004055,
		0.002323,
		0.001235,
		0.000597,
		0.000254,
		9e-005,
		2.4e-005,
		0.018518,
		0.066962,
		0.136115,
		0.218452,
		0.307892,
		0.399576,
		0.48969,
		0.575293,
		0.654192,
		0.724819,
		0.786138,
		0.837568,
		0.878913,
		0.91031,
		0.932184,
		0.945195,
		0.950194,
		0.948161,
		0.940135,
		0.927134,
		0.910055,
		0.889586,
		0.86612,
		0.839707,
		0.810153,
		0.777215,
		0.740761,
		0.700845,
		0.657731,
		0.611886,
		0.56394,
		0.514646,
		0.464821,
		0.415302,
		0.366892,
		0.320332,
		0.276259,
		0.235198,
		0.197539,
		0.163544,
		0.133344,
		0.106953,
		0.084279,
		0.065143,
		0.049296,
		0.036436,
		0.026229,
		0.018324,
		0.012367,
		0.008018,
		0.004955,
		0.00289,
		0.001568,
		0.000776,
		0.000339,
		0.000124,
		3.5e-005,
		0.018011,
		0.065226,
		0.132781,
		0.213418,
		0.301242,
		0.39152,
		0.480503,
		0.565277,
		0.643631,
		0.713954,
		0.775143,
		0.826533,
		0.867837,
		0.899099,
		0.92065,
		0.933069,
		0.937137,
		0.933787,
		0.924032,
		0.908898,
		0.889329,
		0.866103,
		0.839755,
		0.810517,
		0.778361,
		0.74317,
		0.704898,
		0.663655,
		0.61974,
		0.57363,
		0.52595,
		0.477426,
		0.428835,
		0.38096,
		0.334543,
		0.29025,
		0.248646,
		0.21018,
		0.175172,
		0.143817,
		0.116188,
		0.092251,
		0.071873,
		0.054846,
		0.0409,
		0.029723,
		0.020976,
		0.014314,
		0.009391,
		0.00588,
		0.003479,
		0.001918,
		0.000966,
		0.000431,
		0.000161,
		4.6e-005,
		0.017525,
		0.063554,
		0.129562,
		0.208539,
		0.294765,
		0.383622,
		0.471423,
		0.555271,
		0.632938,
		0.702768,
		0.763595,
		0.814675,
		0.855639,
		0.886448,
		0.907357,
		0.918878,
		0.921738,
		0.916836,
		0.905174,
		0.887791,
		0.865681,
		0.839707,
		0.810531,
		0.778551,
		0.743911,
		0.706621,
		0.666724,
		0.624385,
		0.579929,
		0.533837,
		0.486716,
		0.439258,
		0.392189,
		0.346229,
		0.302048,
		0.260237,
		0.221283,
		0.185558,
		0.153312,
		0.124676,
		0.099669,
		0.078208,
		0.060128,
		0.045192,
		0.033116,
		0.023578,
		0.016241,
		0.010764,
		0.006814,
		0.00408,
		0.002279,
		0.001164,
		0.000527,
		0.000201,
		5.9e-005,
		1.1e-005,
		0.017057,
		0.061944,
		0.12645,
		0.203799,
		0.288436,
		0.375848,
		0.462401,
		0.54521,
		0.622029,
		0.691157,
		0.751367,
		0.801847,
		0.842154,
		0.872181,
		0.892122,
		0.902438,
		0.903823,
		0.897154,
		0.883435,
		0.863728,
		0.839077,
		0.810427,
		0.778551,
		0.743999,
		0.707073,
		0.667916,
		0.626654,
		0.583504,
		0.53881,
		0.493048,
		0.446798,
		0.400704,
		0.355436,
		0.311642,
		0.269916,
		0.230768,
		0.194608,
		0.161731,
		0.13232,
		0.106443,
		0.084068,
		0.065071,
		0.049254,
		0.03636,
		0.02609,
		0.01812,
		0.012115,
		0.007741,
		0.004682,
		0.002644,
		0.001366,
		0.000627,
		0.000243,
		7.2e-005,
		1.3e-005,
		0.016608,
		0.06039,
		0.123433,
		0.199181,
		0.282229,
		0.368158,
		0.453381,
		0.535021,
		0.61081,
		0.679005,
		0.738324,
		0.787895,
		0.827216,
		0.856127,
		0.874775,
		0.883593,
		0.883254,
		0.874633,
		0.858746,
		0.836688,
		0.809553,
		0.778361,
		0.743987,
		0.707107,
		0.668171,
		0.627445,
		0.585137,
		0.541506,
		0.496909,
		0.451806,
		0.406741,
		0.362304,
		0.319097,
		0.277694,
		0.238608,
		0.202269,
		0.169008,
		0.139047,
		0.112501,
		0.089384,
		0.069614,
		0.053032,
		0.039411,
		0.028478,
		0.019924,
		0.013424,
		0.008648,
		0.005276,
		0.003007,
		0.00157,
		0.000728,
		0.000285,
		8.6e-005,
		1.6e-005,
		0.016175,
		0.058887,
		0.120501,
		0.194665,
		0.276112,
		0.360506,
		0.444298,
		0.524618,
		0.599175,
		0.666188,
		0.724327,
		0.77267,
		0.810671,
		0.838132,
		0.855177,
		0.862222,
		0.859942,
		0.849223,
		0.831105,
		0.80672,
		0.777215,
		0.743679,
		0.707073,
		0.668181,
		0.627576,
		0.585637,
		0.542647,
		0.498898,
		0.454751,
		0.410642,
		0.367068,
		0.324562,
		0.283651,
		0.244832,
		0.208535,
		0.17511,
		0.144812,
		0.117792,
		0.094104,
		0.073709,
		0.056483,
		0.042233,
		0.030711,
		0.021629,
		0.014673,
		0.009521,
		0.005853,
		0.003363,
		0.001771,
		0.000829,
		0.000328,
		0.0001,
		1.9e-005,
		0.015757,
		0.057429,
		0.117642,
		0.190228,
		0.270048,
		0.352837,
		0.435083,
		0.513912,
		0.587017,
		0.652581,
		0.709237,
		0.756027,
		0.792377,
		0.818069,
		0.83322,
		0.838252,
		0.833854,
		0.820937,
		0.800578,
		0.773948,
		0.742246,
		0.706621,
		0.668108,
		0.627576,
		0.585696,
		0.542944,
		0.499666,
		0.456182,
		0.412843,
		0.370055,
		0.328265,
		0.287937,
		0.249525,
		0.213445,
		0.180046,
		0.1496,
		0.122288,
		0.098195,
		0.077319,
		0.059573,
		0.044794,
		0.032763,
		0.023214,
		0.015848,
		0.010351,
		0.006407,
		0.003708,
		0.001968,
		0.000929,
		0.000371,
		0.000114,
		2.2e-005,
		0.015352,
		0.05601,
		0.114839,
		0.185844,
		0.263995,
		0.345096,
		0.425657,
		0.502809,
		0.574223,
		0.63806,
		0.692926,
		0.737842,
		0.772221,
		0.795845,
		0.808844,
		0.81166,
		0.805015,
		0.789857,
		0.767303,
		0.738568,
		0.704898,
		0.667493,
		0.627445,
		0.585688,
		0.542966,
		0.49983,
		0.456678,
		0.413844,
		0.371666,
		0.330511,
		0.290769,
		0.252835,
		0.217088,
		0.183862,
		0.153429,
		0.125984,
		0.101641,
		0.080424,
		0.062277,
		0.047072,
		0.034615,
		0.024662,
		0.016933,
		0.011126,
		0.00693,
		0.004037,
		0.002157,
		0.001025,
		0.000412,
		0.000127,
		2.4e-005,
		0.014959,
		0.054623,
		0.112078,
		0.181483,
		0.257911,
		0.337219,
		0.415942,
		0.491211,
		0.560686,
		0.622512,
		0.675281,
		0.71801,
		0.750119,
		0.771408,
		0.782037,
		0.782486,
		0.773521,
		0.756135,
		0.73149,
		0.700845,
		0.665486,
		0.626654,
		0.585481,
		0.542944,
		0.499836,
		0.45676,
		0.414148,
		0.372344,
		0.331662,
		0.292425,
		0.254963,
		0.219602,
		0.186644,
		0.156346,
		0.128904,
		0.104444,
		0.083014,
		0.064584,
		0.049052,
		0.036251,
		0.025961,
		0.01792,
		0.011839,
		0.007416,
		0.004345,
		0.002336,
		0.001117,
		0.000452,
		0.000141,
		2.7e-005,
		0.014576,
		0.053259,
		0.109339,
		0.177115,
		0.251747,
		0.329143,
		0.405858,
		0.479026,
		0.546305,
		0.605834,
		0.656209,
		0.696458,
		0.726028,
		0.744755,
		0.752845,
		0.75083,
		0.73953,
		0.719989,
		0.693416,
		0.661108,
		0.624385,
		0.584512,
		0.542647,
		0.49979,
		0.45676,
		0.414184,
		0.37252,
		0.332104,
		0.293225,
		0.256157,
		0.22117,
		0.18852,
		0.158434,
		0.131095,
		0.10663,
		0.085099,
		0.06649,
		0.050725,
		0.037661,
		0.0271,
		0.018799,
		0.012483,
		0.00786,
		0.004631,
		0.002503,
		0.001205,
		0.000491,
		0.000154,
		3e-005,
		0.014199,
		0.05191,
		0.106605,
		0.172707,
		0.245457,
		0.320804,
		0.395328,
		0.46617,
		0.530993,
		0.587947,
		0.635646,
		0.673152,
		0.699951,
		0.715937,
		0.721373,
		0.716858,
		0.703269,
		0.681705,
		0.653418,
		0.61974,
		0.58201,
		0.541506,
		0.499394,
		0.456678,
		0.414179,
		0.372533,
		0.332198,
		0.293501,
		0.256697,
		0.222012,
		0.189652,
		0.159806,
		0.132632,
		0.108244,
		0.086702,
		0.068007,
		0.052095,
		0.038844,
		0.028075,
		0.019564,
		0.013053,
		0.008259,
		0.00489,
		0.002657,
		0.001285,
		0.000526,
		0.000166,
		3.3e-005,
		0.013828,
		0.050566,
		0.103854,
		0.168226,
		0.238991,
		0.312142,
		0.384281,
		0.452566,
		0.514679,
		0.568794,
		0.613563,
		0.648096,
		0.67194,
		0.685059,
		0.687787,
		0.680795,
		0.665023,
		0.641623,
		0.611886,
		0.577165,
		0.53881,
		0.498098,
		0.456182,
		0.414053,
		0.37252,
		0.332201,
		0.293546,
		0.256861,
		0.222364,
		0.190229,
		0.160605,
		0.133616,
		0.109353,
		0.087866,
		0.069157,
		0.053172,
		0.039801,
		0.028884,
		0.020214,
		0.013546,
		0.008609,
		0.005121,
		0.002796,
		0.00136,
		0.00056,
		0.000177,
		3.5e-005,
		0.013459,
		0.049219,
		0.101067,
		0.163637,
		0.232304,
		0.3031,
		0.372656,
		0.438158,
		0.497317,
		0.548352,
		0.589968,
		0.621342,
		0.642099,
		0.652282,
		0.652309,
		0.642924,
		0.625132,
		0.600132,
		0.569245,
		0.533837,
		0.495252,
		0.454751,
		0.413461,
		0.372344,
		0.332175,
		0.293546,
		0.25688,
		0.222455,
		0.19045,
		0.160989,
		0.134165,
		0.110041,
		0.088647,
		0.069976,
		0.053976,
		0.040544,
		0.029533,
		0.020748,
		0.01396,
		0.00891,
		0.005324,
		0.00292,
		0.001426,
		0.00059,
		0.000188,
		3.7e-005,
		0.01309,
		0.047857,
		0.098223,
		0.15891,
		0.225354,
		0.29363,
		0.360406,
		0.422905,
		0.478887,
		0.526629,
		0.564909,
		0.592988,
		0.610581,
		0.617822,
		0.615215,
		0.603579,
		0.58398,
		0.557657,
		0.52595,
		0.490224,
		0.451806,
		0.411925,
		0.371666,
		0.331948,
		0.293501,
		0.256878,
		0.222461,
		0.190496,
		0.161121,
		0.134412,
		0.110408,
		0.089116,
		0.070513,
		0.054538,
		0.041091,
		0.03003,
		0.021172,
		0.014298,
		0.009162,
		0.005496,
		0.003027,
		0.001485,
		0.000617,
		0.000197,
		3.9e-005,
		0.012718,
		0.046472,
		0.095303,
		0.154017,
		0.218106,
		0.283695,
		0.347498,
		0.40679,
		0.459397,
		0.50367,
		0.538479,
		0.563181,
		0.577591,
		0.581942,
		0.576825,
		0.563133,
		0.541984,
		0.514646,
		0.482466,
		0.446798,
		0.408933,
		0.370055,
		0.331199,
		0.293225,
		0.25681,
		0.222455,
		0.190498,
		0.161143,
		0.134487,
		0.110561,
		0.089354,
		0.070823,
		0.054895,
		0.041464,
		0.030388,
		0.021491,
		0.014563,
		0.009364,
		0.005638,
		0.003117,
		0.001535,
		0.00064,
		0.000206,
		4.1e-005,
		0.01234,
		0.045055,
		0.09229,
		0.148931,
		0.210532,
		0.27327,
		0.333919,
		0.389823,
		0.438887,
		0.47956,
		0.510811,
		0.532108,
		0.543376,
		0.544946,
		0.537498,
		0.521988,
		0.499578,
		0.471554,
		0.439258,
		0.404013,
		0.367068,
		0.329549,
		0.292425,
		0.256491,
		0.222364,
		0.190485,
		0.161143,
		0.134495,
		0.1106,
		0.089445,
		0.070972,
		0.055094,
		0.041695,
		0.030628,
		0.021718,
		0.014759,
		0.009521,
		0.005752,
		0.003191,
		0.001578,
		0.000661,
		0.000213,
		4.3e-005,
		0.011955,
		0.043596,
		0.08917,
		0.143636,
		0.202613,
		0.262345,
		0.319677,
		0.372038,
		0.417432,
		0.454416,
		0.482077,
		0.5,
		0.508222,
		0.507174,
		0.497617,
		0.480563,
		0.457205,
		0.428835,
		0.396774,
		0.362304,
		0.326616,
		0.290769,
		0.25566,
		0.222012,
		0.190372,
		0.161121,
		0.134494,
		0.110603,
		0.089464,
		0.071023,
		0.055183,
		0.041817,
		0.030771,
		0.021866,
		0.014896,
		0.009635,
		0.005838,
		0.00325,
		0.001612,
		0.000677,
		0.000219,
		4.4e-005,
		0.011559,
		0.04209,
		0.085931,
		0.138117,
		0.194342,
		0.250926,
		0.304801,
		0.353499,
		0.395137,
		0.428393,
		0.452485,
		0.467118,
		0.472443,
		0.468986,
		0.457578,
		0.43928,
		0.415302,
		0.386925,
		0.355436,
		0.322066,
		0.287937,
		0.254032,
		0.22117,
		0.189997,
		0.160989,
		0.134464,
		0.1106,
		0.089465,
		0.071031,
		0.05521,
		0.041868,
		0.030844,
		0.021951,
		0.014983,
		0.009713,
		0.005901,
		0.003294,
		0.001639,
		0.000691,
		0.000224,
		4.5e-005,
		0.011152,
		0.040531,
		0.082565,
		0.132372,
		0.185725,
		0.239036,
		0.289343,
		0.334294,
		0.372134,
		0.401677,
		0.422271,
		0.43375,
		0.436373,
		0.430753,
		0.417783,
		0.398556,
		0.374285,
		0.346229,
		0.31563,
		0.283651,
		0.251342,
		0.219602,
		0.189167,
		0.160605,
		0.134318,
		0.110561,
		0.089459,
		0.071031,
		0.055213,
		0.041881,
		0.030871,
		0.021992,
		0.015031,
		0.009761,
		0.005942,
		0.003325,
		0.001658,
		0.000701,
		0.000228,
		4.6e-005,
		0.010731,
		0.038915,
		0.07907,
		0.126401,
		0.176778,
		0.226717,
		0.273376,
		0.314536,
		0.348584,
		0.374475,
		0.391695,
		0.400201,
		0.400354,
		0.392847,
		0.37862,
		0.358785,
		0.334543,
		0.307118,
		0.277694,
		0.247362,
		0.217088,
		0.187686,
		0.159806,
		0.133936,
		0.110408,
		0.089413,
		0.071023,
		0.055213,
		0.041882,
		0.030877,
		0.022005,
		0.015052,
		0.009787,
		0.005967,
		0.003345,
		0.001672,
		0.000709,
		0.000232,
		4.7e-005,
		0.010297,
		0.037241,
		0.075448,
		0.120219,
		0.167531,
		0.214025,
		0.256992,
		0.294361,
		0.324667,
		0.347016,
		0.361029,
		0.366781,
		0.36473,
		0.35563,
		0.340459,
		0.320332,
		0.296425,
		0.269916,
		0.241918,
		0.213445,
		0.185375,
		0.158434,
		0.133186,
		0.110041,
		0.089259,
		0.070972,
		0.055202,
		0.041881,
		0.030877,
		0.022008,
		0.015059,
		0.009797,
		0.005979,
		0.003356,
		0.001681,
		0.000714,
		0.000234,
		4.8e-005,
		0.009847,
		0.035511,
		0.071706,
		0.113845,
		0.158028,
		0.201032,
		0.240301,
		0.27392,
		0.300582,
		0.31954,
		0.330552,
		0.333802,
		0.32983,
		0.319444,
		0.303641,
		0.283525,
		0.260237,
		0.234893,
		0.208535,
		0.18209,
		0.156346,
		0.13194,
		0.109353,
		0.088916,
		0.070823,
		0.055149,
		0.041868,
		0.030876,
		0.022008,
		0.015059,
		0.0098,
		0.005984,
		0.003362,
		0.001686,
		0.000717,
		0.000235,
		4.8e-005,
		0.009384,
		0.033727,
		0.067856,
		0.107308,
		0.14832,
		0.187825,
		0.223428,
		0.253379,
		0.276532,
		0.292291,
		0.300538,
		0.30156,
		0.295966,
		0.284601,
		0.268467,
		0.248646,
		0.22623,
		0.202269,
		0.177722,
		0.153429,
		0.130086,
		0.108244,
		0.0883,
		0.070513,
		0.055011,
		0.041817,
		0.030862,
		0.022005,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000236,
		4.8e-005,
		0.008906,
		0.031895,
		0.063915,
		0.100644,
		0.138472,
		0.1745,
		0.206504,
		0.232908,
		0.252728,
		0.265509,
		0.271253,
		0.270336,
		0.26342,
		0.251376,
		0.235198,
		0.21593,
		0.194608,
		0.172206,
		0.1496,
		0.127539,
		0.10663,
		0.087336,
		0.069976,
		0.054739,
		0.041695,
		0.030815,
		0.021992,
		0.015057,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.008417,
		0.030023,
		0.059906,
		0.093895,
		0.128554,
		0.161159,
		0.18967,
		0.212684,
		0.229374,
		0.239425,
		0.242943,
		0.240381,
		0.232442,
		0.220006,
		0.204045,
		0.185558,
		0.165515,
		0.144812,
		0.124235,
		0.104444,
		0.085958,
		0.069157,
		0.054285,
		0.041464,
		0.03071,
		0.021951,
		0.015044,
		0.009797,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.007919,
		0.028122,
		0.055852,
		0.08711,
		0.118641,
		0.147912,
		0.173066,
		0.192875,
		0.206667,
		0.214251,
		0.21583,
		0.211916,
		0.203242,
		0.19068,
		0.175172,
		0.157661,
		0.139047,
		0.12014,
		0.101641,
		0.084118,
		0.068007,
		0.053606,
		0.041091,
		0.030522,
		0.021866,
		0.015011,
		0.009787,
		0.005982,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.007413,
		0.026203,
		0.051783,
		0.080339,
		0.108813,
		0.134865,
		0.156828,
		0.173644,
		0.184787,
		0.190179,
		0.190107,
		0.185128,
		0.175988,
		0.163544,
		0.148694,
		0.13232,
		0.115245,
		0.098195,
		0.081783,
		0.06649,
		0.052669,
		0.040544,
		0.030225,
		0.021718,
		0.014945,
		0.009761,
		0.005974,
		0.003362,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.006903,
		0.024279,
		0.047728,
		0.073636,
		0.099147,
		0.122125,
		0.141088,
		0.155141,
		0.163895,
		0.167375,
		0.165934,
		0.160162,
		0.150806,
		0.138695,
		0.124676,
		0.109565,
		0.094104,
		0.078936,
		0.064584,
		0.051448,
		0.039801,
		0.029799,
		0.021491,
		0.014835,
		0.009713,
		0.005956,
		0.003356,
		0.001687,
		0.000719,
		0.000237,
		4.9e-005,
		0.006393,
		0.022365,
		0.043719,
		0.067052,
		0.089722,
		0.109792,
		0.125963,
		0.137498,
		0.144131,
		0.145975,
		0.143438,
		0.137128,
		0.12778,
		0.116188,
		0.103142,
		0.089384,
		0.075576,
		0.062277,
		0.049927,
		0.038844,
		0.029228,
		0.021172,
		0.014669,
		0.009635,
		0.005924,
		0.003345,
		0.001684,
		0.000719,
		0.000237,
		4.9e-005,
		0.005885,
		0.020474,
		0.039786,
		0.06064,
		0.080608,
		0.097956,
		0.11156,
		0.120829,
		0.125608,
		0.126086,
		0.122709,
		0.116093,
		0.106953,
		0.096035,
		0.084068,
		0.07172,
		0.059573,
		0.0481,
		0.037661,
		0.0285,
		0.020748,
		0.014439,
		0.009521,
		0.005872,
		0.003325,
		0.001677,
		0.000717,
		0.000236,
		4.9e-005,
		0.005385,
		0.018622,
		0.03596,
		0.054447,
		0.071873,
		0.086698,
		0.097968,
		0.105226,
		0.108413,
		0.107782,
		0.103804,
		0.097092,
		0.088328,
		0.078208,
		0.067396,
		0.056483,
		0.04597,
		0.036251,
		0.027608,
		0.020214,
		0.014139,
		0.009364,
		0.005798,
		0.003294,
		0.001666,
		0.000714,
		0.000236,
		4.9e-005,
		0.004895,
		0.016822,
		0.032269,
		0.048517,
		0.063573,
		0.076086,
		0.08526,
		0.090758,
		0.092608,
		0.091109,
		0.086747,
		0.08012,
		0.071873,
		0.062646,
		0.053032,
		0.043547,
		0.034615,
		0.026551,
		0.019564,
		0.013763,
		0.009162,
		0.005698,
		0.00325,
		0.001649,
		0.000709,
		0.000235,
		4.8e-005,
		0.00442,
		0.015087,
		0.028739,
		0.04289,
		0.055758,
		0.066174,
		0.073489,
		0.077474,
		0.078227,
		0.076083,
		0.07153,
		0.065143,
		0.057525,
		0.049254,
		0.040853,
		0.032763,
		0.025331,
		0.018799,
		0.013309,
		0.00891,
		0.005571,
		0.003191,
		0.001626,
		0.000701,
		0.000233,
		4.8e-005,
		0.003962,
		0.01343,
		0.025391,
		0.037596,
		0.048465,
		0.057003,
		0.062693,
		0.0654,
		0.06528,
		0.062691,
		0.058116,
		0.052099,
		0.045192,
		0.037912,
		0.030711,
		0.023956,
		0.01792,
		0.012777,
		0.008609,
		0.005413,
		0.003117,
		0.001596,
		0.000691,
		0.00023,
		4.8e-005,
		0.003525,
		0.011859,
		0.022245,
		0.03266,
		0.041725,
		0.0486,
		0.052889,
		0.054541,
		0.053754,
		0.0509,
		0.046446,
		0.0409,
		0.034759,
		0.028478,
		0.022437,
		0.016933,
		0.01217,
		0.008259,
		0.005226,
		0.003027,
		0.001557,
		0.000677,
		0.000227,
		4.7e-005,
		0.003111,
		0.010384,
		0.019315,
		0.028102,
		0.035553,
		0.040976,
		0.04408,
		0.044882,
		0.043615,
		0.040651,
		0.036436,
		0.031435,
		0.02609,
		0.02079,
		0.015848,
		0.011491,
		0.00786,
		0.005009,
		0.00292,
		0.001511,
		0.000661,
		0.000222,
		4.6e-005,
		0.002723,
		0.009012,
		0.016611,
		0.023933,
		0.029959,
		0.034131,
		0.036252,
		0.036394,
		0.034811,
		0.031869,
		0.027985,
		0.023578,
		0.019033,
		0.014673,
		0.010746,
		0.007416,
		0.004764,
		0.002796,
		0.001456,
		0.00064,
		0.000216,
		4.5e-005,
		0.002362,
		0.007747,
		0.014139,
		0.020157,
		0.024941,
		0.028054,
		0.029379,
		0.02903,
		0.027273,
		0.024461,
		0.020976,
		0.017189,
		0.013424,
		0.009942,
		0.00693,
		0.004491,
		0.002657,
		0.001394,
		0.000617,
		0.00021,
		4.4e-005,
		0.002029,
		0.006591,
		0.011902,
		0.016772,
		0.02049,
		0.022723,
		0.023421,
		0.02273,
		0.020921,
		0.018324,
		0.015282,
		0.012115,
		0.009089,
		0.006407,
		0.004194,
		0.002503,
		0.001323,
		0.00059,
		0.000202,
		4.3e-005,
		0.001726,
		0.005545,
		0.009899,
		0.013772,
		0.016587,
		0.018105,
		0.018328,
		0.017426,
		0.015662,
		0.013341,
		0.010764,
		0.008198,
		0.005853,
		0.003874,
		0.002336,
		0.001246,
		0.00056,
		0.000193,
		4.1e-005,
		0.001451,
		0.00461,
		0.008124,
		0.011143,
		0.013209,
		0.01416,
		0.014044,
		0.013038,
		0.011397,
		0.009391,
		0.007279,
		0.005276,
		0.003537,
		0.002157,
		0.001162,
		0.000526,
		0.000183,
		3.9e-005,
		0.001206,
		0.003782,
		0.00657,
		0.008869,
		0.010326,
		0.010844,
		0.010502,
		0.009482,
		0.008018,
		0.006347,
		0.004682,
		0.003186,
		0.001968,
		0.001072,
		0.000491,
		0.000172,
		3.7e-005,
		0.000989,
		0.003057,
		0.005227,
		0.00693,
		0.007904,
		0.008105,
		0.007633,
		0.006667,
		0.005416,
		0.00408,
		0.002826,
		0.001771,
		0.000977,
		0.000452,
		0.00016,
		3.5e-005,
		0.000799,
		0.002431,
		0.004083,
		0.005302,
		0.005905,
		0.005889,
		0.005365,
		0.004501,
		0.003479,
		0.002461,
		0.00157,
		0.000879,
		0.000412,
		0.000147,
		3.3e-005,
		0.000635,
		0.001898,
		0.003122,
		0.00396,
		0.004288,
		0.004136,
		0.00362,
		0.00289,
		0.002098,
		0.001366,
		0.000779,
		0.000371,
		0.000134,
		3e-005,
		0.000496,
		0.001451,
		0.002331,
		0.002874,
		0.003011,
		0.00279,
		0.002323,
		0.001742,
		0.001164,
		0.000678,
		0.000328,
		0.000121,
		2.7e-005,
		0.000379,
		0.001083,
		0.001691,
		0.002017,
		0.00203,
		0.00179,
		0.001399,
		0.000966,
		0.000577,
		0.000285,
		0.000107,
		2.4e-005,
		0.000283,
		0.000785,
		0.001186,
		0.001359,
		0.001301,
		0.001077,
		0.000776,
		0.000479,
		0.000243,
		9.3e-005,
		2.2e-005,
		0.000205,
		0.000551,
		0.000799,
		0.000871,
		0.000783,
		0.000597,
		0.000384,
		0.000201,
		7.9e-005,
		1.9e-005,
		0.000144,
		0.000371,
		0.000512,
		0.000524,
		0.000434,
		0.000295,
		0.000161,
		6.5e-005,
		1.6e-005,
		9.7e-005,
		0.000237,
		0.000308,
		0.00029,
		0.000215,
		0.000124,
		5.2e-005,
		1.3e-005,
		6.2e-005,
		0.000143,
		0.00017,
		0.000143,
		9e-005,
		4e-005,
		1.1e-005,
		3.7e-005,
		7.9e-005,
		8.4e-005,
		6e-005,
		2.9e-005,
		2.1e-005,
		3.9e-005,
		3.5e-005,
		2e-005,
		1e-005,
		1.6e-005,
		1.1e-005,
		1.6e-005,
		1.6e-005,
		1.7e-005,
		3.5e-005,
		4.9e-005,
		4.5e-005,
		2.1e-005,
		2.7e-005,
		6e-005,
		9.8e-005,
		0.000121,
		0.00011,
		6.2e-005,
		3.7e-005,
		9e-005,
		0.00016,
		0.000227,
		0.000261,
		0.000237,
		0.000149,
		3.7e-005,
		1.2e-005,
		4.9e-005,
		0.000124,
		0.000234,
		0.000359,
		0.000466,
		0.000512,
		0.000464,
		0.000316,
		0.000116,
		1.5e-005,
		6.2e-005,
		0.000161,
		0.000317,
		0.000513,
		0.000715,
		0.000871,
		0.000927,
		0.000842,
		0.000608,
		0.000283,
		2.4e-005,
		1.7e-005,
		7.5e-005,
		0.000201,
		0.000407,
		0.000685,
		0.001001,
		0.001301,
		0.001517,
		0.001579,
		0.001437,
		0.001083,
		0.000586,
		0.000127,
		2e-005,
		8.9e-005,
		0.000243,
		0.000503,
		0.00087,
		0.001317,
		0.00179,
		0.002212,
		0.002495,
		0.002555,
		0.002331,
		0.001816,
		0.001091,
		0.000361,
		2.3e-005,
		0.000103,
		0.000285,
		0.000602,
		0.001064,
		0.001654,
		0.002323,
		0.002991,
		0.003559,
		0.003915,
		0.00396,
		0.003623,
		0.0029,
		0.001882,
		0.000799,
		6.4e-005,
		2.6e-005,
		0.000117,
		0.000328,
		0.000703,
		0.001265,
		0.002008,
		0.00289,
		0.003836,
		0.004739,
		0.005473,
		0.005905,
		0.005917,
		0.005431,
		0.004443,
		0.003057,
		0.001528,
		0.000308,
		2.9e-005,
		0.000131,
		0.000371,
		0.000804,
		0.001468,
		0.00237,
		0.003479,
		0.004727,
		0.006008,
		0.007186,
		0.008105,
		0.008611,
		0.008567,
		0.007888,
		0.00657,
		0.004732,
		0.002653,
		0.000826,
		3.1e-005,
		0.000144,
		0.000412,
		0.000904,
		0.001671,
		0.002735,
		0.00408,
		0.005648,
		0.007338,
		0.009012,
		0.010502,
		0.011623,
		0.012195,
		0.012067,
		0.011143,
		0.009423,
		0.007038,
		0.004295,
		0.001726,
		0.000131,
		3.4e-005,
		0.000157,
		0.000452,
		0.001001,
		0.00187,
		0.003097,
		0.004682,
		0.00658,
		0.008703,
		0.010914,
		0.013038,
		0.014875,
		0.016209,
		0.016836,
		0.016587,
		0.01536,
		0.013156,
		0.010121,
		0.006591,
		0.003134,
		0.000604,
		3.6e-005,
		0.000169,
		0.000491,
		0.001095,
		0.002063,
		0.003451,
		0.005276,
		0.007511,
		0.010079,
		0.012854,
		0.015662,
		0.018292,
		0.020506,
		0.022059,
		0.022723,
		0.02231,
		0.020712,
		0.017935,
		0.014139,
		0.009688,
		0.005191,
		0.00155,
		3.8e-005,
		0.00018,
		0.000526,
		0.001183,
		0.002247,
		0.003792,
		0.005853,
		0.008424,
		0.011443,
		0.014799,
		0.018324,
		0.021804,
		0.02499,
		0.027607,
		0.029379,
		0.030053,
		0.029425,
		0.027382,
		0.023933,
		0.019257,
		0.013746,
		0.008047,
		0.003111,
		0.000228,
		4e-005,
		0.00019,
		0.00056,
		0.001266,
		0.002421,
		0.004116,
		0.006407,
		0.009307,
		0.012776,
		0.016717,
		0.020976,
		0.025346,
		0.029569,
		0.033356,
		0.036394,
		0.038377,
		0.039027,
		0.038125,
		0.035553,
		0.031327,
		0.025644,
		0.018926,
		0.011859,
		0.005442,
		0.001015,
		4.2e-005,
		0.000199,
		0.00059,
		0.001342,
		0.002582,
		0.004419,
		0.00693,
		0.010148,
		0.014057,
		0.018579,
		0.023578,
		0.028856,
		0.034161,
		0.039191,
		0.043615,
		0.047085,
		0.049261,
		0.049843,
		0.0486,
		0.045407,
		0.040291,
		0.033468,
		0.025391,
		0.016787,
		0.008699,
		0.002517,
		4.4e-005,
		0.000208,
		0.000617,
		0.00141,
		0.002729,
		0.004698,
		0.007416,
		0.010938,
		0.015271,
		0.02036,
		0.02609,
		0.03228,
		0.038685,
		0.045007,
		0.0509,
		0.055989,
		0.05989,
		0.062232,
		0.062693,
		0.061029,
		0.057115,
		0.050989,
		0.04289,
		0.0333,
		0.022989,
		0.013042,
		0.004895,
		0.000349,
		4.5e-005,
		0.000215,
		0.00064,
		0.001471,
		0.00286,
		0.004951,
		0.00786,
		0.011667,
		0.016402,
		0.022037,
		0.028478,
		0.035565,
		0.043071,
		0.050704,
		0.058116,
		0.064916,
		0.070687,
		0.075006,
		0.077474,
		0.077749,
		0.075577,
		0.070835,
		0.063573,
		0.054055,
		0.042798,
		0.03061,
		0.018622,
		0.008309,
		0.001509,
		4.6e-005,
		0.000221,
		0.000661,
		0.001523,
		0.002975,
		0.005175,
		0.008259,
		0.012329,
		0.01744,
		0.023589,
		0.030711,
		0.038668,
		0.047255,
		0.056194,
		0.065143,
		0.073706,
		0.081444,
		0.087897,
		0.092608,
		0.095151,
		0.095162,
		0.092384,
		0.086698,
		0.07817,
		0.067091,
		0.054013,
		0.039786,
		0.02558,
		0.012909,
		0.003642,
		4.7e-005,
		0.000225,
		0.000677,
		0.001568,
		0.003074,
		0.005369,
		0.008609,
		0.012918,
		0.018373,
		0.025001,
		0.032763,
		0.041551,
		0.051181,
		0.061398,
		0.071873,
		0.082212,
		0.09197,
		0.10066,
		0.107782,
		0.112844,
		0.115392,
		0.115052,
		0.11156,
		0.104807,
		0.09488,
		0.082099,
		0.067052,
		0.050629,
		0.03404,
		0.018834,
		0.006903,
		0.000482,
		4.7e-005,
		0.000229,
		0.000691,
		0.001604,
		0.003156,
		0.005534,
		0.00891,
		0.01343,
		0.019196,
		0.026261,
		0.034615,
		0.044179,
		0.0548,
		0.066246,
		0.078208,
		0.090306,
		0.102092,
		0.11307,
		0.122709,
		0.130466,
		0.135819,
		0.138292,
		0.137498,
		0.133176,
		0.12523,
		0.11377,
		0.099147,
		0.08199,
		0.063227,
		0.044105,
		0.026203,
		0.011432,
		0.002032,
		4.8e-005,
		0.000232,
		0.000701,
		0.001633,
		0.003222,
		0.005669,
		0.009162,
		0.013864,
		0.019903,
		0.027359,
		0.036251,
		0.04653,
		0.058074,
		0.070681,
		0.084068,
		0.097872,
		0.111657,
		0.124925,
		0.137128,
		0.147689,
		0.156029,
		0.161595,
		0.163895,
		0.162535,
		0.15726,
		0.14799,
		0.134865,
		0.118271,
		0.098873,
		0.07764,
		0.055852,
		0.035114,
		0.01735,
		0.004798,
		4.8e-005,
		0.000234,
		0.000709,
		0.001654,
		0.003273,
		0.005776,
		0.009364,
		0.014221,
		0.020495,
		0.028292,
		0.037661,
		0.048585,
		0.060974,
		0.074658,
		0.089384,
		0.104814,
		0.120532,
		0.136047,
		0.150806,
		0.164213,
		0.175647,
		0.184495,
		0.190179,
		0.192194,
		0.190145,
		0.183787,
		0.173066,
		0.158149,
		0.139462,
		0.11771,
		0.093895,
		0.069331,
		0.045641,
		0.024756,
		0.008906,
		0.000611,
		4.8e-005,
		0.000235,
		0.000714,
		0.001669,
		0.003311,
		0.005856,
		0.009521,
		0.014503,
		0.020973,
		0.029061,
		0.038844,
		0.050336,
		0.063481,
		0.078143,
		0.094104,
		0.111057,
		0.128609,
		0.146287,
		0.163544,
		0.179776,
		0.194338,
		0.206571,
		0.21583,
		0.221518,
		0.223122,
		0.220253,
		0.212684,
		0.200385,
		0.183562,
		0.162676,
		0.138472,
		0.111989,
		0.084562,
		0.057832,
		0.033727,
		0.014463,
		0.00253,
		4.9e-005,
		0.000236,
		0.000717,
		0.001679,
		0.003336,
		0.005913,
		0.009635,
		0.014716,
		0.021344,
		0.029671,
		0.039801,
		0.051781,
		0.065587,
		0.081119,
		0.098195,
		0.116543,
		0.135801,
		0.155521,
		0.175172,
		0.194153,
		0.211808,
		0.227449,
		0.240381,
		0.249933,
		0.255494,
		0.256552,
		0.252728,
		0.243819,
		0.229828,
		0.210995,
		0.187825,
		0.161098,
		0.131883,
		0.101537,
		0.071706,
		0.044312,
		0.021549,
		0.005873,
		4.9e-005,
		0.000237,
		0.000719,
		0.001685,
		0.003351,
		0.00595,
		0.009713,
		0.014867,
		0.021616,
		0.030132,
		0.040544,
		0.052929,
		0.067296,
		0.083582,
		0.101641,
		0.121239,
		0.142051,
		0.163659,
		0.185558,
		0.207159,
		0.22781,
		0.246807,
		0.26342,
		0.276925,
		0.286631,
		0.291922,
		0.292291,
		0.287379,
		0.27701,
		0.261224,
		0.240301,
		0.214785,
		0.18549,
		0.15351,
		0.120219,
		0.087259,
		0.05654,
		0.030225,
		0.010731,
		0.000727,
		4.9e-005,
		0.000237,
		0.000719,
		0.001687,
		0.00336,
		0.005971,
		0.009761,
		0.014965,
		0.021801,
		0.030459,
		0.041091,
		0.053799,
		0.068626,
		0.085543,
		0.104444,
		0.125134,
		0.147328,
		0.170644,
		0.194608,
		0.218656,
		0.242148,
		0.264379,
		0.284601,
		0.302051,
		0.315979,
		0.325684,
		0.330552,
		0.330088,
		0.323959,
		0.312024,
		0.294361,
		0.271292,
		0.243396,
		0.21152,
		0.176778,
		0.140548,
		0.104466,
		0.070417,
		0.040531,
		0.017177,
		0.002974,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003363,
		0.005981,
		0.009787,
		0.015022,
		0.021915,
		0.030672,
		0.041464,
		0.054418,
		0.069605,
		0.087033,
		0.10663,
		0.128245,
		0.151632,
		0.176453,
		0.202269,
		0.228551,
		0.254678,
		0.279959,
		0.303641,
		0.324939,
		0.343062,
		0.357244,
		0.366781,
		0.371068,
		0.369632,
		0.362172,
		0.348584,
		0.328987,
		0.303745,
		0.273471,
		0.239036,
		0.201568,
		0.162442,
		0.123278,
		0.085931,
		0.052485,
		0.025261,
		0.006823,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005984,
		0.009797,
		0.015049,
		0.021976,
		0.030795,
		0.041695,
		0.054822,
		0.070277,
		0.088095,
		0.108244,
		0.130612,
		0.154997,
		0.181102,
		0.208535,
		0.236802,
		0.265316,
		0.293406,
		0.320332,
		0.3453,
		0.367494,
		0.386103,
		0.400354,
		0.409549,
		0.4131,
		0.410564,
		0.401677,
		0.386375,
		0.364821,
		0.337415,
		0.304801,
		0.267868,
		0.227748,
		0.185806,
		0.143636,
		0.103053,
		0.066092,
		0.035019,
		0.01234,
		0.000831,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015058,
		0.022001,
		0.030854,
		0.041817,
		0.055057,
		0.070692,
		0.088791,
		0.109353,
		0.132304,
		0.157487,
		0.18465,
		0.213445,
		0.243424,
		0.274039,
		0.304652,
		0.334543,
		0.362928,
		0.388981,
		0.411863,
		0.430753,
		0.444878,
		0.453558,
		0.456229,
		0.452485,
		0.442101,
		0.425058,
		0.40156,
		0.372038,
		0.337163,
		0.297832,
		0.255173,
		0.210532,
		0.165466,
		0.121743,
		0.081344,
		0.046472,
		0.019574,
		0.003372,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022007,
		0.030874,
		0.041868,
		0.055169,
		0.070914,
		0.089194,
		0.110041,
		0.133415,
		0.159201,
		0.187192,
		0.217088,
		0.248488,
		0.280893,
		0.313702,
		0.346229,
		0.377712,
		0.407332,
		0.434238,
		0.457578,
		0.47653,
		0.490336,
		0.498335,
		0.5,
		0.494965,
		0.48305,
		0.464277,
		0.438887,
		0.407343,
		0.370327,
		0.32874,
		0.283695,
		0.236507,
		0.188689,
		0.141955,
		0.098223,
		0.059631,
		0.028563,
		0.007687,
		4.8e-005,
		0.000236,
		0.000719,
		0.001687,
		0.003363,
		0.005984,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041881,
		0.055207,
		0.071006,
		0.089387,
		0.110408,
		0.134061,
		0.160268,
		0.188866,
		0.219602,
		0.252127,
		0.285991,
		0.320642,
		0.355436,
		0.389643,
		0.422467,
		0.453061,
		0.480563,
		0.504116,
		0.522909,
		0.536205,
		0.543376,
		0.543936,
		0.537563,
		0.524119,
		0.50367,
		0.476489,
		0.443057,
		0.404065,
		0.360406,
		0.313165,
		0.263617,
		0.213225,
		0.163637,
		0.116705,
		0.074497,
		0.039335,
		0.013828,
		0.00093,
		4.8e-005,
		0.000235,
		0.000717,
		0.001685,
		0.00336,
		0.005981,
		0.009797,
		0.015058,
		0.022007,
		0.030877,
		0.041882,
		0.055213,
		0.071029,
		0.089454,
		0.110561,
		0.134371,
		0.160839,
		0.18984,
		0.22117,
		0.254529,
		0.289518,
		0.325642,
		0.362304,
		0.398819,
		0.434424,
		0.468297,
		0.499578,
		0.527398,
		0.550912,
		0.569328,
		0.581942,
		0.588164,
		0.587553,
		0.579832,
		0.564909,
		0.542884,
		0.514056,
		0.478918,
		0.438158,
		0.392648,
		0.343437,
		0.291752,
		0.238991,
		0.186735,
		0.136762,
		0.091069,
		0.05191,
		0.021838,
		0.003761,
		4.8e-005,
		0.000234,
		0.000714,
		0.001679,
		0.003351,
		0.005971,
		0.009787,
		0.015049,
		0.022001,
		0.030874,
		0.041881,
		0.055213,
		0.071031,
		0.089464,
		0.1106,
		0.134478,
		0.161077,
		0.19031,
		0.222012,
		0.255931,
		0.291724,
		0.32895,
		0.367068,
		0.405447,
		0.443369,
		0.480048,
		0.514646,
		0.546301,
		0.574154,
		0.597379,
		0.615215,
		0.626996,
		0.632181,
		0.630373,
		0.621342,
		0.605034,
		0.581581,
		0.551297,
		0.514679,
		0.472399,
		0.425297,
		0.374377,
		0.320804,
		0.265908,
		0.211195,
		0.158365,
		0.109339,
		0.066304,
		0.031753,
		0.008551,
		4.7e-005,
		0.000232,
		0.000709,
		0.001669,
		0.003336,
		0.00595,
		0.009761,
		0.015022,
		0.021976,
		0.030854,
		0.041868,
		0.055207,
		0.071029,
		0.089464,
		0.110603,
		0.134495,
		0.161138,
		0.190471,
		0.222364,
		0.256608,
		0.292911,
		0.330886,
		0.370055,
		0.409844,
		0.449592,
		0.48856,
		0.52595,
		0.560924,
		0.59263,
		0.620228,
		0.642924,
		0.659998,
		0.67083,
		0.674927,
		0.67194,
		0.661683,
		0.644135,
		0.619449,
		0.587947,
		0.550115,
		0.5066,
		0.458198,
		0.405858,
		0.350675,
		0.293903,
		0.236966,
		0.181483,
		0.129299,
		0.082529,
		0.043608,
		0.015352,
		0.001034,
		4.7e-005,
		0.000229,
		0.000701,
		0.001654,
		0.003311,
		0.005913,
		0.009713,
		0.014965,
		0.021915,
		0.030795,
		0.041817,
		0.055169,
		0.071006,
		0.089454,
		0.1106,
		0.134495,
		0.161143,
		0.190498,
		0.222455,
		0.256841,
		0.293409,
		0.331824,
		0.371666,
		0.412426,
		0.453504,
		0.494223,
		0.533837,
		0.571549,
		0.606532,
		0.637959,
		0.665023,
		0.686971,
		0.703131,
		0.712934,
		0.715937,
		0.711837,
		0.700485,
		0.681888,
		0.656209,
		0.623765,
		0.585024,
		0.540591,
		0.491211,
		0.437764,
		0.381271,
		0.322902,
		0.263995,
		0.206085,
		0.150937,
		0.100594,
		0.057429,
		0.024212,
		0.004181,
		4.6e-005,
		0.000225,
		0.000691,
		0.001633,
		0.003273,
		0.005856,
		0.009635,
		0.014867,
		0.021801,
		0.030672,
		0.041695,
		0.055057,
		0.070914,
		0.089387,
		0.110561,
		0.134478,
		0.161138,
		0.190498,
		0.222461,
		0.25688,
		0.293537,
		0.332147,
		0.372344,
		0.413678,
		0.455619,
		0.497557,
		0.53881,
		0.578638,
		0.616261,
		0.650881,
		0.681705,
		0.707972,
		0.72898,
		0.744109,
		0.752845,
		0.754792,
		0.749691,
		0.737423,
		0.71801,
		0.691617,
		0.658546,
		0.619229,
		0.574223,
		0.524209,
		0.469991,
		0.412507,
		0.352837,
		0.292232,
		0.232137,
		0.174237,
		0.120501,
		0.07324,
		0.035173,
		0.009502,
		4.5e-005,
		0.000221,
		0.000677,
		0.001604,
		0.003222,
		0.005776,
		0.009521,
		0.014716,
		0.021616,
		0.030459,
		0.041464,
		0.054822,
		0.070692,
		0.089194,
		0.110408,
		0.134371,
		0.161077,
		0.190471,
		0.222455,
		0.25688,
		0.293547,
		0.332201,
		0.37252,
		0.414112,
		0.456511,
		0.499179,
		0.541506,
		0.582824,
		0.622414,
		0.65953,
		0.693416,
		0.723327,
		0.748558,
		0.768466,
		0.782486,
		0.790158,
		0.791133,
		0.785187,
		0.772221,
		0.752265,
		0.725472,
		0.692115,
		0.652581,
		0.607367,
		0.557079,
		0.502441,
		0.444298,
		0.38364,
		0.321622,
		0.259602,
		0.199181,
		0.142252,
		0.091062,
		0.048276,
		0.017057,
		0.001154,
		4.4e-005,
		0.000215,
		0.000661,
		0.001568,
		0.003156,
		0.005669,
		0.009364,
		0.014503,
		0.021344,
		0.030132,
		0.041091,
		0.054418,
		0.070277,
		0.088791,
		0.110041,
		0.134061,
		0.160839,
		0.19031,
		0.222364,
		0.256841,
		0.293537,
		0.332201,
		0.372533,
		0.414183,
		0.456743,
		0.499742,
		0.542647,
		0.584866,
		0.62576,
		0.664653,
		0.700845,
		0.733637,
		0.762347,
		0.786334,
		0.805015,
		0.817888,
		0.824541,
		0.824668,
		0.818069,
		0.80466,
		0.784462,
		0.757607,
		0.724327,
		0.684951,
		0.639906,
		0.589719,
		0.535021,
		0.476563,
		0.415238,
		0.352107,
		0.288436,
		0.225742,
		0.165839,
		0.110907,
		0.063554,
		0.026901,
		0.004665,
		4.2e-005,
		0.000208,
		0.00064,
		0.001523,
		0.003074,
		0.005534,
		0.009162,
		0.014221,
		0.020973,
		0.029671,
		0.040544,
		0.053799,
		0.069605,
		0.088095,
		0.109353,
		0.133415,
		0.160268,
		0.18984,
		0.222012,
		0.256608,
		0.293409,
		0.332147,
		0.37252,
		0.414183,
		0.456761,
		0.499835,
		0.542944,
		0.585577,
		0.627179,
		0.667159,
		0.704898,
		0.739762,
		0.771123,
		0.798372,
		0.820937,
		0.838304,
		0.850026,
		0.855742,
		0.855177,
		0.84815,
		0.834575,
		0.814459,
		0.787895,
		0.755064,
		0.71623,
		0.671739,
		0.622029,
		0.567636,
		0.509214,
		0.447554,
		0.383622,
		0.31859,
		0.253886,
		0.191247,
		0.132781,
		0.081038,
		0.039085,
		0.010606,
		4e-005,
		0.000199,
		0.000617,
		0.001471,
		0.002975,
		0.005369,
		0.00891,
		0.013864,
		0.020495,
		0.029061,
		0.039801,
		0.052929,
		0.068626,
		0.087033,
		0.108244,
		0.132304,
		0.159201,
		0.188866,
		0.22117,
		0.255931,
		0.292911,
		0.331824,
		0.372344,
		0.414112,
		0.456743,
		0.499835,
		0.542966,
		0.585694,
		0.62755,
		0.668034,
		0.706621,
		0.742769,
		0.775926,
		0.805547,
		0.831105,
		0.852106,
		0.868102,
		0.878705,
		0.883593,
		0.882517,
		0.875305,
		0.861858,
		0.842154,
		0.81624,
		0.784233,
		0.746323,
		0.702768,
		0.653913,
		0.600193,
		0.542159,
		0.480503,
		0.416091,
		0.349999,
		0.283568,
		0.218452,
		0.156684,
		0.100747,
		0.053654,
		0.019046,
		0.001294,
		3.8e-005,
		0.00019,
		0.00059,
		0.00141,
		0.00286,
		0.005175,
		0.008609,
		0.01343,
		0.019903,
		0.028292,
		0.038844,
		0.051781,
		0.067296,
		0.085543,
		0.10663,
		0.130612,
		0.157487,
		0.187192,
		0.219602,
		0.254529,
		0.291724,
		0.330886,
		0.371666,
		0.413678,
		0.456511,
		0.499742,
		0.542944,
		0.585694,
		0.627578,
		0.668179,
		0.707073,
		0.743821,
		0.777975,
		0.80908,
		0.836688,
		0.86036,
		0.879686,
		0.894285,
		0.903823,
		0.908012,
		0.90662,
		0.899472,
		0.886448,
		0.867484,
		0.842572,
		0.811757,
		0.775143,
		0.732899,
		0.685268,
		0.632584,
		0.575293,
		0.513985,
		0.449423,
		0.38259,
		0.314732,
		0.247418,
		0.182602,
		0.122695,
		0.070648,
		0.030049,
		0.005236,
		3.6e-005,
		0.00018,
		0.00056,
		0.001342,
		0.002729,
		0.004951,
		0.008259,
		0.012918,
		0.019196,
		0.027359,
		0.037661,
		0.050336,
		0.065587,
		0.083582,
		0.104444,
		0.128245,
		0.154997,
		0.18465,
		0.217088,
		0.252127,
		0.289518,
		0.32895,
		0.370055,
		0.412426,
		0.455619,
		0.499179,
		0.542647,
		0.585577,
		0.62755,
		0.668179,
		0.707107,
		0.743996,
		0.778514,
		0.810321,
		0.839077,
		0.864439,
		0.886069,
		0.903638,
		0.916836,
		0.925377,
		0.929002,
		0.92749,
		0.92065,
		0.908331,
		0.890421,
		0.866844,
		0.837568,
		0.802607,
		0.762033,
		0.715985,
		0.664693,
		0.608498,
		0.547885,
		0.483514,
		0.416269,
		0.347303,
		0.278094,
		0.210511,
		0.146884,
		0.090096,
		0.043674,
		0.011912,
		3.4e-005,
		0.000169,
		0.000526,
		0.001266,
		0.002582,
		0.004698,
		0.00786,
		0.012329,
		0.018373,
		0.026261,
		0.036251,
		0.048585,
		0.063481,
		0.081119,
		0.101641,
		0.125134,
		0.151632,
		0.181102,
		0.213445,
		0.248488,
		0.285991,
		0.325642,
		0.367068,
		0.409844,
		0.453504,
		0.497557,
		0.541506,
		0.584866,
		0.627179,
		0.668034,
		0.707073,
		0.743996,
		0.778554,
		0.810528,
		0.839707,
		0.865876,
		0.88881,
		0.908275,
		0.924032,
		0.935839,
		0.943456,
		0.946649,
		0.945195,
		0.938887,
		0.927533,
		0.910961,
		0.889025,
		0.861608,
		0.828631,
		0.790063,
		0.745942,
		0.696388,
		0.641638,
		0.582068,
		0.518237,
		0.450927,
		0.381192,
		0.310414,
		0.24037,
		0.173305,
		0.112019,
		0.059973,
		0.021402,
		0.001462,
		3.1e-005,
		0.000157,
		0.000491,
		0.001183,
		0.002421,
		0.004419,
		0.007416,
		0.011667,
		0.01744,
		0.025001,
		0.034615,
		0.04653,
		0.060974,
		0.078143,
		0.098195,
		0.121239,
		0.147328,
		0.176453,
		0.208535,
		0.243424,
		0.280893,
		0.320642,
		0.362304,
		0.405447,
		0.449592,
		0.494223,
		0.53881,
		0.582824,
		0.62576,
		0.667159,
		0.706621,
		0.743821,
		0.778514,
		0.810528,
		0.839755,
		0.866116,
		0.889535,
		0.909914,
		0.927134,
		0.941046,
		0.951479,
		0.958238,
		0.961111,
		0.95987,
		0.954283,
		0.944112,
		0.929125,
		0.909096,
		0.883821,
		0.853125,
		0.816874,
		0.775,
		0.727518,
		0.674555,
		0.616381,
		0.553448,
		0.486434,
		0.416289,
		0.344293,
		0.272124,
		0.201932,
		0.136427,
		0.078987,
		0.033782,
		0.00592,
		2.9e-005,
		0.000144,
		0.000452,
		0.001095,
		0.002247,
		0.004116,
		0.00693,
		0.010938,
		0.016402,
		0.023589,
		0.032763,
		0.044179,
		0.058074,
		0.074658,
		0.094104,
		0.116543,
		0.142051,
		0.170644,
		0.202269,
		0.236802,
		0.274039,
		0.313702,
		0.355436,
		0.398819,
		0.443369,
		0.48856,
		0.533837,
		0.578638,
		0.622414,
		0.664653,
		0.704898,
		0.742769,
		0.777975,
		0.810321,
		0.839707,
		0.866116,
		0.88959,
		0.910189,
		0.927957,
		0.942893,
		0.95495,
		0.964028,
		0.969981,
		0.972616,
		0.971702,
		0.966979,
		0.958161,
		0.944951,
		0.927046,
		0.904151,
		0.875992,
		0.842337,
		0.803008,
		0.757914,
		0.707073,
		0.650647,
		0.588979,
		0.522637,
		0.452462,
		0.379627,
		0.3057,
		0.232725,
		0.163312,
		0.100747,
		0.049119,
		0.013475,
		2.6e-005,
		0.000131,
		0.000412,
		0.001001,
		0.002063,
		0.003792,
		0.006407,
		0.010148,
		0.015271,
		0.022037,
		0.030711,
		0.041551,
		0.0548,
		0.070681,
		0.089384,
		0.111057,
		0.135801,
		0.163659,
		0.194608,
		0.228551,
		0.265316,
		0.304652,
		0.346229,
		0.389643,
		0.434424,
		0.480048,
		0.52595,
		0.571549,
		0.616261,
		0.65953,
		0.700845,
		0.739762,
		0.775926,
		0.80908,
		0.839077,
		0.865876,
		0.889535,
		0.910189,
		0.928019,
		0.943204,
		0.955875,
		0.96609,
		0.973833,
		0.979009,
		0.981452,
		0.980929,
		0.977154,
		0.969798,
		0.958497,
		0.942872,
		0.92254,
		0.897131,
		0.86631,
		0.829796,
		0.787387,
		0.738994,
		0.684668,
		0.624639,
		0.559361,
		0.489559,
		0.416289,
		0.341002,
		0.265625,
		0.192655,
		0.125272,
		0.067471,
		0.024223,
		0.001665,
		2.3e-005,
		0.000117,
		0.000371,
		0.000904,
		0.00187,
		0.003451,
		0.005853,
		0.009307,
		0.014057,
		0.02036,
		0.028478,
		0.038668,
		0.051181,
		0.066246,
		0.084068,
		0.104814,
		0.128609,
		0.155521,
		0.185558,
		0.218656,
		0.254678,
		0.293406,
		0.334543,
		0.377712,
		0.422467,
		0.468297,
		0.514646,
		0.560924,
		0.606532,
		0.650881,
		0.693416,
		0.733637,
		0.771123,
		0.805547,
		0.836688,
		0.864439,
		0.88881,
		0.909914,
		0.927957,
		0.943204,
		0.955945,
		0.966438,
		0.974863,
		0.981295,
		0.985703,
		0.987957,
		0.987832,
		0.985025,
		0.979167,
		0.969837,
		0.956583,
		0.938936,
		0.916436,
		0.888645,
		0.855181,
		0.815734,
		0.770102,
		0.718225,
		0.660215,
		0.596407,
		0.527402,
		0.45413,
		0.377913,
		0.300551,
		0.224414,
		0.152563,
		0.088884,
		0.038254,
		0.006746,
		2e-005,
		0.000103,
		0.000328,
		0.000804,
		0.001671,
		0.003097,
		0.005276,
		0.008424,
		0.012776,
		0.018579,
		0.02609,
		0.035565,
		0.047255,
		0.061398,
		0.078208,
		0.097872,
		0.120532,
		0.146287,
		0.175172,
		0.207159,
		0.242148,
		0.279959,
		0.320332,
		0.362928,
		0.407332,
		0.453061,
		0.499578,
		0.546301,
		0.59263,
		0.637959,
		0.681705,
		0.723327,
		0.762347,
		0.798372,
		0.831105,
		0.86036,
		0.886069,
		0.908275,
		0.927134,
		0.942893,
		0.955875,
		0.966438,
		0.974941,
		0.981682,
		0.986844,
		0.990479,
		0.992508,
		0.992731,
		0.990841,
		0.986442,
		0.979067,
		0.968199,
		0.953294,
		0.9338,
		0.909189,
		0.878971,
		0.842734,
		0.800162,
		0.751076,
		0.69547,
		0.63355,
		0.565786,
		0.492972,
		0.416289,
		0.337397,
		0.258527,
		0.182603,
		0.113386,
		0.055645,
		0.015366,
		1.7e-005,
		8.9e-005,
		0.000285,
		0.000703,
		0.001468,
		0.002735,
		0.004682,
		0.007511,
		0.011443,
		0.016717,
		0.023578,
		0.03228,
		0.043071,
		0.056194,
		0.071873,
		0.090306,
		0.111657,
		0.136047,
		0.163544,
		0.194153,
		0.22781,
		0.264379,
		0.303641,
		0.3453,
		0.388981,
		0.434238,
		0.480563,
		0.527398,
		0.574154,
		0.620228,
		0.665023,
		0.707972,
		0.748558,
		0.786334,
		0.820937,
		0.852106,
		0.879686,
		0.903638,
		0.924032,
		0.941046,
		0.95495,
		0.96609,
		0.974863,
		0.981682,
		0.986931,
		0.990908,
		0.993768,
		0.995509,
		0.995977,
		0.994885,
		0.991829,
		0.986312,
		0.977766,
		0.965582,
		0.94913,
		0.927789,
		0.900973,
		0.86816,
		0.828924,
		0.782964,
		0.730146,
		0.670542,
		0.604481,
		0.532609,
		0.455958,
		0.376031,
		0.294906,
		0.215353,
		0.140991,
		0.076461,
		0.027641,
		0.001913,
		1.5e-005,
		7.5e-005,
		0.000243,
		0.000602,
		0.001265,
		0.00237,
		0.00408,
		0.00658,
		0.010079,
		0.014799,
		0.020976,
		0.028856,
		0.038685,
		0.050704,
		0.065143,
		0.082212,
		0.102092,
		0.124925,
		0.150806,
		0.179776,
		0.211808,
		0.246807,
		0.284601,
		0.324939,
		0.367494,
		0.411863,
		0.457578,
		0.504116,
		0.550912,
		0.597379,
		0.642924,
		0.686971,
		0.72898,
		0.768466,
		0.805015,
		0.838304,
		0.868102,
		0.894285,
		0.916836,
		0.935839,
		0.951479,
		0.964028,
		0.973833,
		0.981295,
		0.986844,
		0.990908,
		0.993865,
		0.995983,
		0.997368,
		0.997943,
		0.997471,
		0.995567,
		0.991729,
		0.98536,
		0.975798,
		0.962349,
		0.944307,
		0.920994,
		0.891783,
		0.856131,
		0.813614,
		0.763962,
		0.707107,
		0.643226,
		0.572806,
		0.496716,
		0.416289,
		0.333432,
		0.250747,
		0.171689,
		0.100747,
		0.043674,
		0.007758,
		1.2e-005,
		6.2e-005,
		0.000201,
		0.000503,
		0.001064,
		0.002008,
		0.003479,
		0.005648,
		0.008703,
		0.012854,
		0.018324,
		0.025346,
		0.034161,
		0.045007,
		0.058116,
		0.073706,
		0.09197,
		0.11307,
		0.137128,
		0.164213,
		0.194338,
		0.227449,
		0.26342,
		0.302051,
		0.343062,
		0.386103,
		0.430753,
		0.47653,
		0.522909,
		0.569328,
		0.615215,
		0.659998,
		0.703131,
		0.744109,
		0.782486,
		0.817888,
		0.850026,
		0.878705,
		0.903823,
		0.925377,
		0.943456,
		0.958238,
		0.969981,
		0.979009,
		0.985703,
		0.990479,
		0.993768,
		0.995983,
		0.997474,
		0.998467,
		0.999005,
		0.998935,
		0.997932,
		0.995519,
		0.991097,
		0.983977,
		0.973413,
		0.95863,
		0.938862,
		0.913379,
		0.881522,
		0.842738,
		0.796619,
		0.742944,
		0.681729,
		0.613292,
		0.538323,
		0.457972,
		0.373956,
		0.28869,
		0.205443,
		0.128533,
		0.063557,
		0.017685,
		4.9e-005,
		0.000161,
		0.000407,
		0.00087,
		0.001654,
		0.00289,
		0.004727,
		0.007338,
		0.010914,
		0.015662,
		0.021804,
		0.029569,
		0.039191,
		0.0509,
		0.064916,
		0.081444,
		0.10066,
		0.122709,
		0.147689,
		0.175647,
		0.206571,
		0.240381,
		0.276925,
		0.315979,
		0.357244,
		0.400354,
		0.444878,
		0.490336,
		0.536205,
		0.581942,
		0.626996,
		0.67083,
		0.712934,
		0.752845,
		0.790158,
		0.824541,
		0.855742,
		0.883593,
		0.908012,
		0.929002,
		0.946649,
		0.961111,
		0.972616,
		0.981452,
		0.987957,
		0.992508,
		0.995509,
		0.997368,
		0.998467,
		0.999123,
		0.999515,
		0.999627,
		0.999235,
		0.99793,
		0.99515,
		0.990206,
		0.982326,
		0.970681,
		0.954428,
		0.932739,
		0.904839,
		0.870045,
		0.827802,
		0.777728,
		0.71967,
		0.653763,
		0.580504,
		0.500843,
		0.416289,
		0.32905,
		0.242186,
		0.159817,
		0.08736,
		0.031834,
		0.002221,
		3.7e-005,
		0.000124,
		0.000317,
		0.000685,
		0.001317,
		0.002323,
		0.003836,
		0.006008,
		0.009012,
		0.013038,
		0.018292,
		0.02499,
		0.033356,
		0.043615,
		0.055989,
		0.070687,
		0.087897,
		0.107782,
		0.130466,
		0.156029,
		0.184495,
		0.21583,
		0.249933,
		0.286631,
		0.325684,
		0.366781,
		0.409549,
		0.453558,
		0.498335,
		0.543376,
		0.588164,
		0.632181,
		0.674927,
		0.715937,
		0.754792,
		0.791133,
		0.824668,
		0.855177,
		0.882517,
		0.90662,
		0.92749,
		0.945195,
		0.95987,
		0.971702,
		0.980929,
		0.987832,
		0.992731,
		0.995977,
		0.997943,
		0.999005,
		0.999515,
		0.999758,
		0.999877,
		0.999806,
		0.999258,
		0.99775,
		0.994638,
		0.989153,
		0.980438,
		0.967586,
		0.949683,
		0.925841,
		0.895243,
		0.857183,
		0.811114,
		0.756699,
		0.693878,
		0.622942,
		0.544621,
		0.460201,
		0.371655,
		0.281814,
		0.19457,
		0.11513,
		0.05033,
		0.009016,
		2.7e-005,
		9e-005,
		0.000234,
		0.000513,
		0.001001,
		0.00179,
		0.002991,
		0.004739,
		0.007186,
		0.010502,
		0.014875,
		0.020506,
		0.027607,
		0.036394,
		0.047085,
		0.05989,
		0.075006,
		0.092608,
		0.112844,
		0.135819,
		0.161595,
		0.190179,
		0.221518,
		0.255494,
		0.291922,
		0.330552,
		0.371068,
		0.4131,
		0.456229,
		0.5,
		0.543936,
		0.587553,
		0.630373,
		0.67194,
		0.711837,
		0.749691,
		0.785187,
		0.818069,
		0.84815,
		0.875305,
		0.899472,
		0.92065,
		0.938887,
		0.954283,
		0.966979,
		0.977154,
		0.985025,
		0.990841,
		0.994885,
		0.997471,
		0.998935,
		0.999627,
		0.999877,
		0.999952,
		0.999972,
		0.999832,
		0.999192,
		0.997501,
		0.994038,
		0.987949,
		0.978292,
		0.964076,
		0.944313,
		0.918052,
		0.884434,
		0.842738,
		0.792437,
		0.73326,
		0.665277,
		0.588982,
		0.505413,
		0.416289,
		0.324181,
		0.232725,
		0.146884,
		0.073275,
		0.02057,
		1.7e-005,
		6e-005,
		0.00016,
		0.000359,
		0.000715,
		0.001301,
		0.002212,
		0.003559,
		0.005473,
		0.008105,
		0.011623,
		0.016209,
		0.022059,
		0.029379,
		0.038377,
		0.049261,
		0.062232,
		0.077474,
		0.095151,
		0.115392,
		0.138292,
		0.163895,
		0.192194,
		0.223122,
		0.256552,
		0.292291,
		0.330088,
		0.369632,
		0.410564,
		0.452485,
		0.494965,
		0.537563,
		0.579832,
		0.621342,
		0.661683,
		0.700485,
		0.737423,
		0.772221,
		0.80466,
		0.834575,
		0.861858,
		0.886448,
		0.908331,
		0.927533,
		0.944112,
		0.958161,
		0.969798,
		0.979167,
		0.986442,
		0.991829,
		0.995567,
		0.997932,
		0.999235,
		0.999806,
		0.999972,
		0.999994,
		0.999987,
		0.999818,
		0.999098,
		0.99721,
		0.99335,
		0.986574,
		0.975845,
		0.960083,
		0.938217,
		0.909232,
		0.87223,
		0.826481,
		0.771497,
		0.707107,
		0.63355,
		0.551596,
		0.462682,
		0.369091,
		0.27417,
		0.182603,
		0.100747,
		0.037057,
		0.00261,
		3.5e-005,
		9.8e-005,
		0.000227,
		0.000466,
		0.000871,
		0.001517,
		0.002495,
		0.003915,
		0.005905,
		0.008611,
		0.012195,
		0.016836,
		0.022723,
		0.030053,
		0.039027,
		0.049843,
		0.062693,
		0.077749,
		0.095162,
		0.115052,
		0.137498,
		0.162535,
		0.190145,
		0.220253,
		0.252728,
		0.287379,
		0.323959,
		0.362172,
		0.401677,
		0.442101,
		0.48305,
		0.524119,
		0.564909,
		0.605034,
		0.644135,
		0.681888,
		0.71801,
		0.752265,
		0.784462,
		0.814459,
		0.842154,
		0.867484,
		0.890421,
		0.910961,
		0.929125,
		0.944951,
		0.958497,
		0.969837,
		0.979067,
		0.986312,
		0.991729,
		0.995519,
		0.99793,
		0.999258,
		0.999832,
		0.999987,
		1,
		0.999987,
		0.999797,
		0.998989,
		0.996876,
		0.992561,
		0.984999,
		0.973047,
		0.955526,
		0.931276,
		0.899218,
		0.858415,
		0.808146,
		0.747985,
		0.677898,
		0.598361,
		0.510503,
		0.416289,
		0.318742,
		0.222223,
		0.132788,
		0.058623,
		0.010606,
		1.6e-005,
		4.9e-005,
		0.000121,
		0.000261,
		0.000512,
		0.000927,
		0.001579,
		0.002555,
		0.00396,
		0.005917,
		0.008567,
		0.012067,
		0.016587,
		0.02231,
		0.029425,
		0.038125,
		0.0486,
		0.061029,
		0.075577,
		0.092384,
		0.11156,
		0.133176,
		0.15726,
		0.183787,
		0.212684,
		0.243819,
		0.27701,
		0.312024,
		0.348584,
		0.386375,
		0.425058,
		0.464277,
		0.50367,
		0.542884,
		0.581581,
		0.619449,
		0.656209,
		0.691617,
		0.725472,
		0.757607,
		0.787895,
		0.81624,
		0.842572,
		0.866844,
		0.889025,
		0.909096,
		0.927046,
		0.942872,
		0.956583,
		0.968199,
		0.977766,
		0.98536,
		0.991097,
		0.99515,
		0.99775,
		0.999192,
		0.999818,
		0.999987,
		1,
		0.999985,
		0.999771,
		0.998863,
		0.99649,
		0.991651,
		0.983187,
		0.969836,
		0.950308,
		0.92335,
		0.887815,
		0.842738,
		0.787425,
		0.721547,
		0.645261,
		0.559361,
		0.465459,
		0.366214,
		0.265625,
		0.169394,
		0.085389,
		0.024223,
		1.6e-005,
		4.5e-005,
		0.00011,
		0.000237,
		0.000464,
		0.000842,
		0.001437,
		0.002331,
		0.003623,
		0.005431,
		0.007888,
		0.011143,
		0.01536,
		0.020712,
		0.027382,
		0.035553,
		0.045407,
		0.057115,
		0.070835,
		0.086698,
		0.104807,
		0.12523,
		0.14799,
		0.173066,
		0.200385,
		0.229828,
		0.261224,
		0.294361,
		0.328987,
		0.364821,
		0.40156,
		0.438887,
		0.476489,
		0.514056,
		0.551297,
		0.587947,
		0.623765,
		0.658546,
		0.692115,
		0.724327,
		0.755064,
		0.784233,
		0.811757,
		0.837568,
		0.861608,
		0.883821,
		0.904151,
		0.92254,
		0.938936,
		0.953294,
		0.965582,
		0.975798,
		0.983977,
		0.990206,
		0.994638,
		0.997501,
		0.999098,
		0.999797,
		0.999985,
		1,
		0.999984,
		0.999741,
		0.998717,
		0.996042,
		0.990599,
		0.981094,
		0.966137,
		0.944313,
		0.914268,
		0.874792,
		0.824905,
		0.763962,
		0.691782,
		0.608789,
		0.516206,
		0.416289,
		0.312626,
		0.210511,
		0.117431,
		0.043674,
		0.003111,
		2.1e-005,
		6.2e-005,
		0.000149,
		0.000316,
		0.000608,
		0.001083,
		0.001816,
		0.0029,
		0.004443,
		0.00657,
		0.009423,
		0.013156,
		0.017935,
		0.023933,
		0.031327,
		0.040291,
		0.050989,
		0.063573,
		0.07817,
		0.09488,
		0.11377,
		0.134865,
		0.158149,
		0.183562,
		0.210995,
		0.240301,
		0.271292,
		0.303745,
		0.337415,
		0.372038,
		0.407343,
		0.443057,
		0.478918,
		0.514679,
		0.550115,
		0.585024,
		0.619229,
		0.652581,
		0.684951,
		0.71623,
		0.746323,
		0.775143,
		0.802607,
		0.828631,
		0.853125,
		0.875992,
		0.897131,
		0.916436,
		0.9338,
		0.94913,
		0.962349,
		0.973413,
		0.982326,
		0.989153,
		0.994038,
		0.99721,
		0.998989,
		0.999771,
		0.999984,
		1,
		0.999981,
		0.999707,
		0.998546,
		0.995521,
		0.989376,
		0.978669,
		0.961859,
		0.937398,
		0.903828,
		0.859876,
		0.804566,
		0.737347,
		0.658245,
		0.568056,
		0.468589,
		0.362964,
		0.256014,
		0.154777,
		0.069138,
		0.012657,
		3.7e-005,
		0.000116,
		0.000283,
		0.000586,
		0.001091,
		0.001882,
		0.003057,
		0.004732,
		0.007038,
		0.010121,
		0.014139,
		0.019257,
		0.025644,
		0.033468,
		0.04289,
		0.054055,
		0.067091,
		0.082099,
		0.099147,
		0.118271,
		0.139462,
		0.162676,
		0.187825,
		0.214785,
		0.243396,
		0.273471,
		0.304801,
		0.337163,
		0.370327,
		0.404065,
		0.438158,
		0.472399,
		0.5066,
		0.540591,
		0.574223,
		0.607367,
		0.639906,
		0.671739,
		0.702768,
		0.732899,
		0.762033,
		0.790063,
		0.816874,
		0.842337,
		0.86631,
		0.888645,
		0.909189,
		0.927789,
		0.944307,
		0.95863,
		0.970681,
		0.980438,
		0.987949,
		0.99335,
		0.996876,
		0.998863,
		0.999741,
		0.999981,
		1,
		0.999979,
		0.999666,
		0.998347,
		0.994912,
		0.987949,
		0.975845,
		0.956891,
		0.929393,
		0.891784,
		0.842738,
		0.781314,
		0.707107,
		0.620445,
		0.522638,
		0.416289,
		0.3057,
		0.197385,
		0.100747,
		0.028941,
		2.4e-005,
		0.000127,
		0.000361,
		0.000799,
		0.001528,
		0.002653,
		0.004295,
		0.006591,
		0.009688,
		0.013746,
		0.018926,
		0.025391,
		0.0333,
		0.042798,
		0.054013,
		0.067052,
		0.08199,
		0.098873,
		0.11771,
		0.138472,
		0.161098,
		0.18549,
		0.21152,
		0.239036,
		0.267868,
		0.297832,
		0.32874,
		0.360406,
		0.392648,
		0.425297,
		0.458198,
		0.491211,
		0.524209,
		0.557079,
		0.589719,
		0.622029,
		0.653913,
		0.685268,
		0.715985,
		0.745942,
		0.775,
		0.803008,
		0.829796,
		0.855181,
		0.878971,
		0.900973,
		0.920994,
		0.938862,
		0.954428,
		0.967586,
		0.978292,
		0.986574,
		0.992561,
		0.99649,
		0.998717,
		0.999707,
		0.999979,
		1,
		0.999976,
		0.999618,
		0.998111,
		0.994196,
		0.986276,
		0.972543,
		0.951098,
		0.92009,
		0.877839,
		0.82299,
		0.754672,
		0.672708,
		0.577857,
		0.472143,
		0.359265,
		0.245133,
		0.138573,
		0.05223,
		0.003771,
		6.4e-005,
		0.000308,
		0.000826,
		0.001726,
		0.003134,
		0.005191,
		0.008047,
		0.011859,
		0.016787,
		0.022989,
		0.03061,
		0.039786,
		0.050629,
		0.063227,
		0.07764,
		0.093895,
		0.111989,
		0.131883,
		0.15351,
		0.176778,
		0.201568,
		0.227748,
		0.255173,
		0.283695,
		0.313165,
		0.343437,
		0.374377,
		0.405858,
		0.437764,
		0.469991,
		0.502441,
		0.535021,
		0.567636,
		0.600193,
		0.632584,
		0.664693,
		0.696388,
		0.727518,
		0.757914,
		0.787387,
		0.815734,
		0.842734,
		0.86816,
		0.891783,
		0.913379,
		0.932739,
		0.949683,
		0.964076,
		0.975845,
		0.984999,
		0.991651,
		0.996042,
		0.998546,
		0.999666,
		0.999976,
		1,
		0.999972,
		0.999561,
		0.997833,
		0.99335,
		0.984305,
		0.968661,
		0.944313,
		0.909232,
		0.861636,
		0.800163,
		0.724081,
		0.63355,
		0.529947,
		0.416289,
		0.297795,
		0.182603,
		0.082738,
		0.015366,
		0.000131,
		0.000604,
		0.00155,
		0.003111,
		0.005442,
		0.008699,
		0.013042,
		0.018622,
		0.02558,
		0.03404,
		0.044105,
		0.055852,
		0.069331,
		0.084562,
		0.101537,
		0.120219,
		0.140548,
		0.162442,
		0.185806,
		0.210532,
		0.236507,
		0.263617,
		0.291752,
		0.320804,
		0.350675,
		0.381271,
		0.412507,
		0.444298,
		0.476563,
		0.509214,
		0.542159,
		0.575293,
		0.608498,
		0.641638,
		0.674555,
		0.707073,
		0.738994,
		0.770102,
		0.800162,
		0.828924,
		0.856131,
		0.881522,
		0.904839,
		0.925841,
		0.944313,
		0.960083,
		0.973047,
		0.983187,
		0.990599,
		0.995521,
		0.998347,
		0.999618,
		0.999972,
		1,
		0.999968,
		0.999493,
		0.997501,
		0.992346,
		0.981968,
		0.964076,
		0.936326,
		0.896506,
		0.842738,
		0.773702,
		0.688896,
		0.588982,
		0.476215,
		0.355015,
		0.232725,
		0.120607,
		0.035182,
		0.000228,
		0.001015,
		0.002517,
		0.004895,
		0.008309,
		0.012909,
		0.018834,
		0.026203,
		0.035114,
		0.045641,
		0.057832,
		0.071706,
		0.087259,
		0.104466,
		0.123278,
		0.143636,
		0.165466,
		0.188689,
		0.213225,
		0.238991,
		0.265908,
		0.293903,
		0.322902,
		0.352837,
		0.38364,
		0.415238,
		0.447554,
		0.480503,
		0.513985,
		0.547885,
		0.582068,
		0.616381,
		0.650647,
		0.684668,
		0.718225,
		0.751076,
		0.782964,
		0.813614,
		0.842738,
		0.870045,
		0.895243,
		0.918052,
		0.938217,
		0.955526,
		0.969836,
		0.981094,
		0.989376,
		0.994912,
		0.998111,
		0.999561,
		0.999968,
		1,
		0.999962,
		0.999412,
		0.997104,
		0.991144,
		0.979184,
		0.958631,
		0.926878,
		0.881522,
		0.820615,
		0.742944,
		0.64838,
		0.538323,
		0.416289,
		0.28869,
		0.165879,
		0.063557,
		0.004665,
		0.000349,
		0.001509,
		0.003642,
		0.006903,
		0.011432,
		0.01735,
		0.024756,
		0.033727,
		0.044312,
		0.05654,
		0.070417,
		0.085931,
		0.103053,
		0.121743,
		0.141955,
		0.163637,
		0.186735,
		0.211195,
		0.236966,
		0.263995,
		0.292232,
		0.321622,
		0.352107,
		0.383622,
		0.416091,
		0.449423,
		0.483514,
		0.518237,
		0.553448,
		0.588979,
		0.624639,
		0.660215,
		0.69547,
		0.730146,
		0.763962,
		0.796619,
		0.827802,
		0.857183,
		0.884434,
		0.909232,
		0.931276,
		0.950308,
		0.966137,
		0.978669,
		0.987949,
		0.994196,
		0.997833,
		0.999493,
		0.999962,
		1,
		0.999956,
		0.999313,
		0.996625,
		0.9897,
		0.975845,
		0.952128,
		0.915644,
		0.863799,
		0.794618,
		0.707107,
		0.601712,
		0.480925,
		0.350084,
		0.218463,
		0.100747,
		0.019046,
		0.000482,
		0.002032,
		0.004798,
		0.008906,
		0.014463,
		0.021549,
		0.030225,
		0.040531,
		0.052485,
		0.066092,
		0.081344,
		0.098223,
		0.116705,
		0.136762,
		0.158365,
		0.181483,
		0.206085,
		0.232137,
		0.259602,
		0.288436,
		0.31859,
		0.349999,
		0.38259,
		0.416269,
		0.450927,
		0.486434,
		0.522637,
		0.559361,
		0.596407,
		0.63355,
		0.670542,
		0.707107,
		0.742944,
		0.777728,
		0.811114,
		0.842738,
		0.87223,
		0.899218,
		0.92335,
		0.944313,
		0.961859,
		0.975845,
		0.986276,
		0.99335,
		0.997501,
		0.999412,
		0.999956,
		1,
		0.999948,
		0.999193,
		0.996042,
		0.987949,
		0.971816,
		0.944313,
		0.902212,
		0.842738,
		0.763962,
		0.665277,
		0.548017,
		0.416289,
		0.278096,
		0.146884,
		0.043674,
		0.000611,
		0.00253,
		0.005873,
		0.010731,
		0.017177,
		0.025261,
		0.035019,
		0.046472,
		0.059631,
		0.074497,
		0.091069,
		0.109339,
		0.129299,
		0.150937,
		0.174237,
		0.199181,
		0.225742,
		0.253886,
		0.283568,
		0.314732,
		0.347303,
		0.381192,
		0.416289,
		0.452462,
		0.489559,
		0.527402,
		0.565786,
		0.604481,
		0.643226,
		0.681729,
		0.71967,
		0.756699,
		0.792437,
		0.826481,
		0.858415,
		0.887815,
		0.914268,
		0.937398,
		0.956891,
		0.972543,
		0.984305,
		0.992346,
		0.997104,
		0.999313,
		0.999948,
		1,
		0.999939,
		0.999046,
		0.995329,
		0.985814,
		0.96692,
		0.93486,
		0.886059,
		0.817592,
		0.727694,
		0.61641,
		0.486437,
		0.344293,
		0.201932,
		0.078987,
		0.00592,
		0.000727,
		0.002974,
		0.006823,
		0.01234,
		0.019574,
		0.028563,
		0.039335,
		0.05191,
		0.066304,
		0.082529,
		0.100594,
		0.120501,
		0.142252,
		0.165839,
		0.191247,
		0.218452,
		0.247418,
		0.278094,
		0.310414,
		0.344293,
		0.379627,
		0.416289,
		0.45413,
		0.492972,
		0.532609,
		0.572806,
		0.613292,
		0.653763,
		0.693878,
		0.73326,
		0.771497,
		0.808146,
		0.842738,
		0.874792,
		0.903828,
		0.929393,
		0.951098,
		0.968661,
		0.981968,
		0.991144,
		0.996625,
		0.999193,
		0.999939,
		1,
		0.999927,
		0.998863,
		0.994448,
		0.983187,
		0.960924,
		0.92335,
		0.86652,
		0.787425,
		0.684672,
		0.559361,
		0.416289,
		0.265625,
		0.125272,
		0.024222,
		0.000831,
		0.003372,
		0.007687,
		0.013828,
		0.021838,
		0.031753,
		0.043608,
		0.057429,
		0.07324,
		0.091062,
		0.110907,
		0.132781,
		0.156684,
		0.182602,
		0.210511,
		0.24037,
		0.272124,
		0.3057,
		0.341002,
		0.377913,
		0.416289,
		0.455958,
		0.496716,
		0.538323,
		0.580504,
		0.622942,
		0.665277,
		0.707107,
		0.747985,
		0.787425,
		0.824905,
		0.859876,
		0.891784,
		0.92009,
		0.944313,
		0.964076,
		0.979184,
		0.9897,
		0.996042,
		0.999046,
		0.999927,
		1,
		0.999912,
		0.998635,
		0.99335,
		0.979927,
		0.953525,
		0.909232,
		0.842738,
		0.751077,
		0.63355,
		0.492972,
		0.337397,
		0.182595,
		0.055596,
		0.00093,
		0.003761,
		0.008551,
		0.015352,
		0.024212,
		0.035173,
		0.048276,
		0.063554,
		0.081038,
		0.100747,
		0.122695,
		0.146884,
		0.173305,
		0.201932,
		0.232725,
		0.265625,
		0.300551,
		0.337397,
		0.376031,
		0.416289,
		0.457972,
		0.500843,
		0.544621,
		0.588982,
		0.63355,
		0.677898,
		0.721547,
		0.763962,
		0.804566,
		0.842738,
		0.877839,
		0.909232,
		0.936326,
		0.958631,
		0.975845,
		0.987949,
		0.995329,
		0.998863,
		0.999912,
		1,
		0.999893,
		0.998347,
		0.991969,
		0.975845,
		0.944313,
		0.891784,
		0.813614,
		0.707107,
		0.572806,
		0.416269,
		0.250527,
		0.100129,
		0.007563,
		0.001034,
		0.004181,
		0.009502,
		0.017057,
		0.026901,
		0.039085,
		0.053654,
		0.070648,
		0.090096,
		0.112019,
		0.136427,
		0.163312,
		0.192655,
		0.224414,
		0.258527,
		0.294906,
		0.333432,
		0.373956,
		0.416289,
		0.460201,
		0.505413,
		0.551596,
		0.598361,
		0.645261,
		0.691782,
		0.737347,
		0.781314,
		0.82299,
		0.861636,
		0.896506,
		0.926878,
		0.952128,
		0.971816,
		0.985814,
		0.994448,
		0.998635,
		0.999893,
		1,
		0.999869,
		0.997978,
		0.990212,
		0.970681,
		0.932739,
		0.870045,
		0.777728,
		0.653732,
		0.500403,
		0.327031,
		0.155812,
		0.029542,
		0.001154,
		0.004665,
		0.010606,
		0.019046,
		0.030049,
		0.043674,
		0.059973,
		0.078987,
		0.100747,
		0.125272,
		0.152563,
		0.182603,
		0.215353,
		0.250747,
		0.28869,
		0.32905,
		0.371655,
		0.416289,
		0.462682,
		0.510503,
		0.559361,
		0.608789,
		0.658245,
		0.707107,
		0.754672,
		0.800163,
		0.842738,
		0.881522,
		0.915644,
		0.944313,
		0.96692,
		0.983187,
		0.99335,
		0.998347,
		0.999869,
		1,
		0.999838,
		0.997501,
		0.987949,
		0.964076,
		0.918052,
		0.842698,
		0.732617,
		0.585368,
		0.405858,
		0.215973,
		0.061533,
		0.001294,
		0.005236,
		0.011912,
		0.021402,
		0.033782,
		0.049119,
		0.067471,
		0.088884,
		0.113386,
		0.140991,
		0.171689,
		0.205443,
		0.242186,
		0.281814,
		0.324181,
		0.369091,
		0.416289,
		0.465459,
		0.516206,
		0.568056,
		0.620445,
		0.672708,
		0.724081,
		0.773702,
		0.820615,
		0.863799,
		0.902212,
		0.93486,
		0.960924,
		0.979927,
		0.991969,
		0.997978,
		0.999838,
		1,
		0.999797,
		0.996876,
		0.984999,
		0.955481,
		0.898429,
		0.803188,
		0.660911,
		0.473757,
		0.267665,
		0.093895,
		0.005759,
		0.001462,
		0.00592,
		0.013475,
		0.024223,
		0.038254,
		0.055645,
		0.076461,
		0.100747,
		0.128533,
		0.159817,
		0.19457,
		0.232725,
		0.27417,
		0.318742,
		0.366214,
		0.416289,
		0.468589,
		0.522638,
		0.577857,
		0.63355,
		0.688896,
		0.742944,
		0.794618,
		0.842738,
		0.886059,
		0.92335,
		0.953525,
		0.975845,
		0.990212,
		0.997501,
		0.999797,
		1,
		0.999741,
		0.995994,
		0.980234,
		0.938519,
		0.852871,
		0.708972,
		0.511233,
		0.294361,
		0.1143,
		0.01627,
		0.001665,
		0.006746,
		0.015366,
		0.027641,
		0.043674,
		0.063557,
		0.08736,
		0.11513,
		0.146884,
		0.182603,
		0.222223,
		0.265625,
		0.312626,
		0.362964,
		0.416289,
		0.472143,
		0.529947,
		0.588982,
		0.64838,
		0.707107,
		0.763962,
		0.817592,
		0.86652,
		0.909232,
		0.944313,
		0.970681,
		0.987949,
		0.996876,
		0.999741,
		0.999952,
		0.998789,
		0.988808,
		0.951392,
		0.862495,
		0.707694,
		0.5,
		0.283774,
		0.113883,
		0.022412,
		0.001913,
		0.007758,
		0.017685,
		0.031834,
		0.05033,
		0.073275,
		0.100747,
		0.132788,
		0.169394,
		0.210511,
		0.256014,
		0.3057,
		0.359265,
		0.416289,
		0.476215,
		0.538323,
		0.601712,
		0.665277,
		0.727694,
		0.787425,
		0.842738,
		0.891784,
		0.932739,
		0.964076,
		0.984999,
		0.995994,
		0.998789,
		0.993865,
		0.974514,
		0.921848,
		0.813438,
		0.642924,
		0.434461,
		0.236018,
		0.092608,
		0.020197,
		0.000644,
		0.002221,
		0.009016,
		0.02057,
		0.037057,
		0.058623,
		0.085389,
		0.117431,
		0.154777,
		0.197385,
		0.245133,
		0.297795,
		0.355015,
		0.416289,
		0.480925,
		0.548017,
		0.61641,
		0.684672,
		0.751077,
		0.813614,
		0.870045,
		0.918052,
		0.955481,
		0.980234,
		0.988808,
		0.974514,
		0.928019,
		0.839261,
		0.700845,
		0.520504,
		0.328396,
		0.165276,
		0.05954,
		0.012091,
		0.000623,
		0.00261,
		0.010606,
		0.024223,
		0.043674,
		0.069138,
		0.100747,
		0.138573,
		0.182603,
		0.232725,
		0.28869,
		0.350084,
		0.416289,
		0.486437,
		0.559361,
		0.63355,
		0.707107,
		0.777728,
		0.842698,
		0.898429,
		0.938519,
		0.951392,
		0.921848,
		0.839261,
		0.707107,
		0.542528,
		0.368043,
		0.210073,
		0.09321,
		0.027863,
		0.00408,
		0.000106,
		0.003111,
		0.012657,
		0.028941,
		0.05223,
		0.082738,
		0.120607,
		0.165879,
		0.218463,
		0.278096,
		0.344293,
		0.416289,
		0.492972,
		0.572806,
		0.653732,
		0.732617,
		0.803188,
		0.852871,
		0.862495,
		0.813438,
		0.700845,
		0.542528,
		0.372533,
		0.222209,
		0.108744,
		0.038672,
		0.007717,
		0.000402,
		0.003771,
		0.015366,
		0.035182,
		0.063557,
		0.100747,
		0.146884,
		0.201932,
		0.265625,
		0.337397,
		0.416269,
		0.500403,
		0.585368,
		0.660911,
		0.708972,
		0.707694,
		0.642924,
		0.520504,
		0.368043,
		0.222209,
		0.110603,
		0.041813,
		0.009563,
		0.000642,
		0.004665,
		0.019046,
		0.043674,
		0.078987,
		0.125272,
		0.182595,
		0.250527,
		0.327031,
		0.405858,
		0.473757,
		0.511233,
		0.5,
		0.434461,
		0.328396,
		0.210073,
		0.108744,
		0.041813,
		0.0098,
		0.000718,
		0.00592,
		0.024222,
		0.055596,
		0.100129,
		0.155812,
		0.215973,
		0.267665,
		0.294361,
		0.283774,
		0.236018,
		0.165276,
		0.09321,
		0.038672,
		0.009563,
		0.000718,
		0.007563,
		0.029542,
		0.061533,
		0.093895,
		0.1143,
		0.113883,
		0.092608,
		0.05954,
		0.027863,
		0.007717,
		0.000642,
		0.005759,
		0.01627,
		0.022412,
		0.020197,
		0.012091,
		0.00408,
		0.000402,
		0.000644,
		0.000623,
		0.000106,
		0.000106,
		0.000623,
		0.000644,
		0.000402,
		0.00408,
		0.012091,
		0.020197,
		0.022412,
		0.01627,
		0.005759,
		0.000642,
		0.007717,
		0.027863,
		0.05954,
		0.092608,
		0.113883,
		0.1143,
		0.093895,
		0.061533,
		0.029542,
		0.007563,
		0.000718,
		0.009563,
		0.038672,
		0.09321,
		0.165276,
		0.236018,
		0.283774,
		0.294361,
		0.267665,
		0.215973,
		0.155812,
		0.100129,
		0.055596,
		0.024222,
		0.00592,
		0.000718,
		0.0098,
		0.041813,
		0.108744,
		0.210073,
		0.328396,
		0.434461,
		0.5,
		0.511233,
		0.473757,
		0.405858,
		0.327031,
		0.250527,
		0.182595,
		0.125272,
		0.078987,
		0.043674,
		0.019046,
		0.004665,
		0.000642,
		0.009563,
		0.041813,
		0.110603,
		0.222209,
		0.368043,
		0.520504,
		0.642924,
		0.707694,
		0.708972,
		0.660911,
		0.585368,
		0.500403,
		0.416269,
		0.337397,
		0.265625,
		0.201932,
		0.146884,
		0.100747,
		0.063557,
		0.035182,
		0.015366,
		0.003771,
		0.000402,
		0.007717,
		0.038672,
		0.108744,
		0.222209,
		0.372533,
		0.542528,
		0.700845,
		0.813438,
		0.862495,
		0.852871,
		0.803188,
		0.732617,
		0.653732,
		0.572806,
		0.492972,
		0.416289,
		0.344293,
		0.278096,
		0.218463,
		0.165879,
		0.120607,
		0.082738,
		0.05223,
		0.028941,
		0.012657,
		0.003111,
		0.000106,
		0.00408,
		0.027863,
		0.09321,
		0.210073,
		0.368043,
		0.542528,
		0.707107,
		0.839261,
		0.921848,
		0.951392,
		0.938519,
		0.898429,
		0.842698,
		0.777728,
		0.707107,
		0.63355,
		0.559361,
		0.486437,
		0.416289,
		0.350084,
		0.28869,
		0.232725,
		0.182603,
		0.138573,
		0.100747,
		0.069138,
		0.043674,
		0.024223,
		0.010606,
		0.00261,
		0.000623,
		0.012091,
		0.05954,
		0.165276,
		0.328396,
		0.520504,
		0.700845,
		0.839261,
		0.928019,
		0.974514,
		0.988808,
		0.980234,
		0.955481,
		0.918052,
		0.870045,
		0.813614,
		0.751077,
		0.684672,
		0.61641,
		0.548017,
		0.480925,
		0.416289,
		0.355015,
		0.297795,
		0.245133,
		0.197385,
		0.154777,
		0.117431,
		0.085389,
		0.058623,
		0.037057,
		0.02057,
		0.009016,
		0.002221,
		0.000644,
		0.020197,
		0.092608,
		0.236018,
		0.434461,
		0.642924,
		0.813438,
		0.921848,
		0.974514,
		0.993865,
		0.998789,
		0.995994,
		0.984999,
		0.964076,
		0.932739,
		0.891784,
		0.842738,
		0.787425,
		0.727694,
		0.665277,
		0.601712,
		0.538323,
		0.476215,
		0.416289,
		0.359265,
		0.3057,
		0.256014,
		0.210511,
		0.169394,
		0.132788,
		0.100747,
		0.073275,
		0.05033,
		0.031834,
		0.017685,
		0.007758,
		0.001913,
		0.022412,
		0.113883,
		0.283774,
		0.5,
		0.707694,
		0.862495,
		0.951392,
		0.988808,
		0.998789,
		0.999952,
		0.999741,
		0.996876,
		0.987949,
		0.970681,
		0.944313,
		0.909232,
		0.86652,
		0.817592,
		0.763962,
		0.707107,
		0.64838,
		0.588982,
		0.529947,
		0.472143,
		0.416289,
		0.362964,
		0.312626,
		0.265625,
		0.222223,
		0.182603,
		0.146884,
		0.11513,
		0.08736,
		0.063557,
		0.043674,
		0.027641,
		0.015366,
		0.006746,
		0.001665,
		0.01627,
		0.1143,
		0.294361,
		0.511233,
		0.708972,
		0.852871,
		0.938519,
		0.980234,
		0.995994,
		0.999741,
		1,
		0.999797,
		0.997501,
		0.990212,
		0.975845,
		0.953525,
		0.92335,
		0.886059,
		0.842738,
		0.794618,
		0.742944,
		0.688896,
		0.63355,
		0.577857,
		0.522638,
		0.468589,
		0.416289,
		0.366214,
		0.318742,
		0.27417,
		0.232725,
		0.19457,
		0.159817,
		0.128533,
		0.100747,
		0.076461,
		0.055645,
		0.038254,
		0.024223,
		0.013475,
		0.00592,
		0.001462,
		0.005759,
		0.093895,
		0.267665,
		0.473757,
		0.660911,
		0.803188,
		0.898429,
		0.955481,
		0.984999,
		0.996876,
		0.999797,
		1,
		0.999838,
		0.997978,
		0.991969,
		0.979927,
		0.960924,
		0.93486,
		0.902212,
		0.863799,
		0.820615,
		0.773702,
		0.724081,
		0.672708,
		0.620445,
		0.568056,
		0.516206,
		0.465459,
		0.416289,
		0.369091,
		0.324181,
		0.281814,
		0.242186,
		0.205443,
		0.171689,
		0.140991,
		0.113386,
		0.088884,
		0.067471,
		0.049119,
		0.033782,
		0.021402,
		0.011912,
		0.005236,
		0.001294,
		0.061533,
		0.215973,
		0.405858,
		0.585368,
		0.732617,
		0.842698,
		0.918052,
		0.964076,
		0.987949,
		0.997501,
		0.999838,
		1,
		0.999869,
		0.998347,
		0.99335,
		0.983187,
		0.96692,
		0.944313,
		0.915644,
		0.881522,
		0.842738,
		0.800163,
		0.754672,
		0.707107,
		0.658245,
		0.608789,
		0.559361,
		0.510503,
		0.462682,
		0.416289,
		0.371655,
		0.32905,
		0.28869,
		0.250747,
		0.215353,
		0.182603,
		0.152563,
		0.125272,
		0.100747,
		0.078987,
		0.059973,
		0.043674,
		0.030049,
		0.019046,
		0.010606,
		0.004665,
		0.001154,
		0.029542,
		0.155812,
		0.327031,
		0.500403,
		0.653732,
		0.777728,
		0.870045,
		0.932739,
		0.970681,
		0.990212,
		0.997978,
		0.999869,
		1,
		0.999893,
		0.998635,
		0.994448,
		0.985814,
		0.971816,
		0.952128,
		0.926878,
		0.896506,
		0.861636,
		0.82299,
		0.781314,
		0.737347,
		0.691782,
		0.645261,
		0.598361,
		0.551596,
		0.505413,
		0.460201,
		0.416289,
		0.373956,
		0.333432,
		0.294906,
		0.258527,
		0.224414,
		0.192655,
		0.163312,
		0.136427,
		0.112019,
		0.090096,
		0.070648,
		0.053654,
		0.039085,
		0.026901,
		0.017057,
		0.009502,
		0.004181,
		0.001034,
		0.007563,
		0.100129,
		0.250527,
		0.416269,
		0.572806,
		0.707107,
		0.813614,
		0.891784,
		0.944313,
		0.975845,
		0.991969,
		0.998347,
		0.999893,
		1,
		0.999912,
		0.998863,
		0.995329,
		0.987949,
		0.975845,
		0.958631,
		0.936326,
		0.909232,
		0.877839,
		0.842738,
		0.804566,
		0.763962,
		0.721547,
		0.677898,
		0.63355,
		0.588982,
		0.544621,
		0.500843,
		0.457972,
		0.416289,
		0.376031,
		0.337397,
		0.300551,
		0.265625,
		0.232725,
		0.201932,
		0.173305,
		0.146884,
		0.122695,
		0.100747,
		0.081038,
		0.063554,
		0.048276,
		0.035173,
		0.024212,
		0.015352,
		0.008551,
		0.003761,
		0.00093,
		0.055596,
		0.182595,
		0.337397,
		0.492972,
		0.63355,
		0.751077,
		0.842738,
		0.909232,
		0.953525,
		0.979927,
		0.99335,
		0.998635,
		0.999912,
		1,
		0.999927,
		0.999046,
		0.996042,
		0.9897,
		0.979184,
		0.964076,
		0.944313,
		0.92009,
		0.891784,
		0.859876,
		0.824905,
		0.787425,
		0.747985,
		0.707107,
		0.665277,
		0.622942,
		0.580504,
		0.538323,
		0.496716,
		0.455958,
		0.416289,
		0.377913,
		0.341002,
		0.3057,
		0.272124,
		0.24037,
		0.210511,
		0.182602,
		0.156684,
		0.132781,
		0.110907,
		0.091062,
		0.07324,
		0.057429,
		0.043608,
		0.031753,
		0.021838,
		0.013828,
		0.007687,
		0.003372,
		0.000831,
		0.024222,
		0.125272,
		0.265625,
		0.416289,
		0.559361,
		0.684672,
		0.787425,
		0.86652,
		0.92335,
		0.960924,
		0.983187,
		0.994448,
		0.998863,
		0.999927,
		1,
		0.999939,
		0.999193,
		0.996625,
		0.991144,
		0.981968,
		0.968661,
		0.951098,
		0.929393,
		0.903828,
		0.874792,
		0.842738,
		0.808146,
		0.771497,
		0.73326,
		0.693878,
		0.653763,
		0.613292,
		0.572806,
		0.532609,
		0.492972,
		0.45413,
		0.416289,
		0.379627,
		0.344293,
		0.310414,
		0.278094,
		0.247418,
		0.218452,
		0.191247,
		0.165839,
		0.142252,
		0.120501,
		0.100594,
		0.082529,
		0.066304,
		0.05191,
		0.039335,
		0.028563,
		0.019574,
		0.01234,
		0.006823,
		0.002974,
		0.000727,
		0.00592,
		0.078987,
		0.201932,
		0.344293,
		0.486437,
		0.61641,
		0.727694,
		0.817592,
		0.886059,
		0.93486,
		0.96692,
		0.985814,
		0.995329,
		0.999046,
		0.999939,
		1,
		0.999948,
		0.999313,
		0.997104,
		0.992346,
		0.984305,
		0.972543,
		0.956891,
		0.937398,
		0.914268,
		0.887815,
		0.858415,
		0.826481,
		0.792437,
		0.756699,
		0.71967,
		0.681729,
		0.643226,
		0.604481,
		0.565786,
		0.527402,
		0.489559,
		0.452462,
		0.416289,
		0.381192,
		0.347303,
		0.314732,
		0.283568,
		0.253886,
		0.225742,
		0.199181,
		0.174237,
		0.150937,
		0.129299,
		0.109339,
		0.091069,
		0.074497,
		0.059631,
		0.046472,
		0.035019,
		0.025261,
		0.017177,
		0.010731,
		0.005873,
		0.00253,
		0.000611,
		0.043674,
		0.146884,
		0.278096,
		0.416289,
		0.548017,
		0.665277,
		0.763962,
		0.842738,
		0.902212,
		0.944313,
		0.971816,
		0.987949,
		0.996042,
		0.999193,
		0.999948,
		1,
		0.999956,
		0.999412,
		0.997501,
		0.99335,
		0.986276,
		0.975845,
		0.961859,
		0.944313,
		0.92335,
		0.899218,
		0.87223,
		0.842738,
		0.811114,
		0.777728,
		0.742944,
		0.707107,
		0.670542,
		0.63355,
		0.596407,
		0.559361,
		0.522637,
		0.486434,
		0.450927,
		0.416269,
		0.38259,
		0.349999,
		0.31859,
		0.288436,
		0.259602,
		0.232137,
		0.206085,
		0.181483,
		0.158365,
		0.136762,
		0.116705,
		0.098223,
		0.081344,
		0.066092,
		0.052485,
		0.040531,
		0.030225,
		0.021549,
		0.014463,
		0.008906,
		0.004798,
		0.002032,
		0.000482,
		0.019046,
		0.100747,
		0.218463,
		0.350084,
		0.480925,
		0.601712,
		0.707107,
		0.794618,
		0.863799,
		0.915644,
		0.952128,
		0.975845,
		0.9897,
		0.996625,
		0.999313,
		0.999956,
		1,
		0.999962,
		0.999493,
		0.997833,
		0.994196,
		0.987949,
		0.978669,
		0.966137,
		0.950308,
		0.931276,
		0.909232,
		0.884434,
		0.857183,
		0.827802,
		0.796619,
		0.763962,
		0.730146,
		0.69547,
		0.660215,
		0.624639,
		0.588979,
		0.553448,
		0.518237,
		0.483514,
		0.449423,
		0.416091,
		0.383622,
		0.352107,
		0.321622,
		0.292232,
		0.263995,
		0.236966,
		0.211195,
		0.186735,
		0.163637,
		0.141955,
		0.121743,
		0.103053,
		0.085931,
		0.070417,
		0.05654,
		0.044312,
		0.033727,
		0.024756,
		0.01735,
		0.011432,
		0.006903,
		0.003642,
		0.001509,
		0.000349,
		0.004665,
		0.063557,
		0.165879,
		0.28869,
		0.416289,
		0.538323,
		0.64838,
		0.742944,
		0.820615,
		0.881522,
		0.926878,
		0.958631,
		0.979184,
		0.991144,
		0.997104,
		0.999412,
		0.999962,
		1,
		0.999968,
		0.999561,
		0.998111,
		0.994912,
		0.989376,
		0.981094,
		0.969836,
		0.955526,
		0.938217,
		0.918052,
		0.895243,
		0.870045,
		0.842738,
		0.813614,
		0.782964,
		0.751076,
		0.718225,
		0.684668,
		0.650647,
		0.616381,
		0.582068,
		0.547885,
		0.513985,
		0.480503,
		0.447554,
		0.415238,
		0.38364,
		0.352837,
		0.322902,
		0.293903,
		0.265908,
		0.238991,
		0.213225,
		0.188689,
		0.165466,
		0.143636,
		0.123278,
		0.104466,
		0.087259,
		0.071706,
		0.057832,
		0.045641,
		0.035114,
		0.026203,
		0.018834,
		0.012909,
		0.008309,
		0.004895,
		0.002517,
		0.001015,
		0.000228,
		0.035182,
		0.120607,
		0.232725,
		0.355015,
		0.476215,
		0.588982,
		0.688896,
		0.773702,
		0.842738,
		0.896506,
		0.936326,
		0.964076,
		0.981968,
		0.992346,
		0.997501,
		0.999493,
		0.999968,
		1,
		0.999972,
		0.999618,
		0.998347,
		0.995521,
		0.990599,
		0.983187,
		0.973047,
		0.960083,
		0.944313,
		0.925841,
		0.904839,
		0.881522,
		0.856131,
		0.828924,
		0.800162,
		0.770102,
		0.738994,
		0.707073,
		0.674555,
		0.641638,
		0.608498,
		0.575293,
		0.542159,
		0.509214,
		0.476563,
		0.444298,
		0.412507,
		0.381271,
		0.350675,
		0.320804,
		0.291752,
		0.263617,
		0.236507,
		0.210532,
		0.185806,
		0.162442,
		0.140548,
		0.120219,
		0.101537,
		0.084562,
		0.069331,
		0.055852,
		0.044105,
		0.03404,
		0.02558,
		0.018622,
		0.013042,
		0.008699,
		0.005442,
		0.003111,
		0.00155,
		0.000604,
		0.000131,
		0.015366,
		0.082738,
		0.182603,
		0.297795,
		0.416289,
		0.529947,
		0.63355,
		0.724081,
		0.800163,
		0.861636,
		0.909232,
		0.944313,
		0.968661,
		0.984305,
		0.99335,
		0.997833,
		0.999561,
		0.999972,
		1,
		0.999976,
		0.999666,
		0.998546,
		0.996042,
		0.991651,
		0.984999,
		0.975845,
		0.964076,
		0.949683,
		0.932739,
		0.913379,
		0.891783,
		0.86816,
		0.842734,
		0.815734,
		0.787387,
		0.757914,
		0.727518,
		0.696388,
		0.664693,
		0.632584,
		0.600193,
		0.567636,
		0.535021,
		0.502441,
		0.469991,
		0.437764,
		0.405858,
		0.374377,
		0.343437,
		0.313165,
		0.283695,
		0.255173,
		0.227748,
		0.201568,
		0.176778,
		0.15351,
		0.131883,
		0.111989,
		0.093895,
		0.07764,
		0.063227,
		0.050629,
		0.039786,
		0.03061,
		0.022989,
		0.016787,
		0.011859,
		0.008047,
		0.005191,
		0.003134,
		0.001726,
		0.000826,
		0.000308,
		6.4e-005,
		0.003771,
		0.05223,
		0.138573,
		0.245133,
		0.359265,
		0.472143,
		0.577857,
		0.672708,
		0.754672,
		0.82299,
		0.877839,
		0.92009,
		0.951098,
		0.972543,
		0.986276,
		0.994196,
		0.998111,
		0.999618,
		0.999976,
		1,
		0.999979,
		0.999707,
		0.998717,
		0.99649,
		0.992561,
		0.986574,
		0.978292,
		0.967586,
		0.954428,
		0.938862,
		0.920994,
		0.900973,
		0.878971,
		0.855181,
		0.829796,
		0.803008,
		0.775,
		0.745942,
		0.715985,
		0.685268,
		0.653913,
		0.622029,
		0.589719,
		0.557079,
		0.524209,
		0.491211,
		0.458198,
		0.425297,
		0.392648,
		0.360406,
		0.32874,
		0.297832,
		0.267868,
		0.239036,
		0.21152,
		0.18549,
		0.161098,
		0.138472,
		0.11771,
		0.098873,
		0.08199,
		0.067052,
		0.054013,
		0.042798,
		0.0333,
		0.025391,
		0.018926,
		0.013746,
		0.009688,
		0.006591,
		0.004295,
		0.002653,
		0.001528,
		0.000799,
		0.000361,
		0.000127,
		2.4e-005,
		0.028941,
		0.100747,
		0.197385,
		0.3057,
		0.416289,
		0.522638,
		0.620445,
		0.707107,
		0.781314,
		0.842738,
		0.891784,
		0.929393,
		0.956891,
		0.975845,
		0.987949,
		0.994912,
		0.998347,
		0.999666,
		0.999979,
		1,
		0.999981,
		0.999741,
		0.998863,
		0.996876,
		0.99335,
		0.987949,
		0.980438,
		0.970681,
		0.95863,
		0.944307,
		0.927789,
		0.909189,
		0.888645,
		0.86631,
		0.842337,
		0.816874,
		0.790063,
		0.762033,
		0.732899,
		0.702768,
		0.671739,
		0.639906,
		0.607367,
		0.574223,
		0.540591,
		0.5066,
		0.472399,
		0.438158,
		0.404065,
		0.370327,
		0.337163,
		0.304801,
		0.273471,
		0.243396,
		0.214785,
		0.187825,
		0.162676,
		0.139462,
		0.118271,
		0.099147,
		0.082099,
		0.067091,
		0.054055,
		0.04289,
		0.033468,
		0.025644,
		0.019257,
		0.014139,
		0.010121,
		0.007038,
		0.004732,
		0.003057,
		0.001882,
		0.001091,
		0.000586,
		0.000283,
		0.000116,
		3.7e-005,
		0.012657,
		0.069138,
		0.154777,
		0.256014,
		0.362964,
		0.468589,
		0.568056,
		0.658245,
		0.737347,
		0.804566,
		0.859876,
		0.903828,
		0.937398,
		0.961859,
		0.978669,
		0.989376,
		0.995521,
		0.998546,
		0.999707,
		0.999981,
		1,
		0.999984,
		0.999771,
		0.998989,
		0.99721,
		0.994038,
		0.989153,
		0.982326,
		0.973413,
		0.962349,
		0.94913,
		0.9338,
		0.916436,
		0.897131,
		0.875992,
		0.853125,
		0.828631,
		0.802607,
		0.775143,
		0.746323,
		0.71623,
		0.684951,
		0.652581,
		0.619229,
		0.585024,
		0.550115,
		0.514679,
		0.478918,
		0.443057,
		0.407343,
		0.372038,
		0.337415,
		0.303745,
		0.271292,
		0.240301,
		0.210995,
		0.183562,
		0.158149,
		0.134865,
		0.11377,
		0.09488,
		0.07817,
		0.063573,
		0.050989,
		0.040291,
		0.031327,
		0.023933,
		0.017935,
		0.013156,
		0.009423,
		0.00657,
		0.004443,
		0.0029,
		0.001816,
		0.001083,
		0.000608,
		0.000316,
		0.000149,
		6.2e-005,
		2.1e-005,
		0.003111,
		0.043674,
		0.117431,
		0.210511,
		0.312626,
		0.416289,
		0.516206,
		0.608789,
		0.691782,
		0.763962,
		0.824905,
		0.874792,
		0.914268,
		0.944313,
		0.966137,
		0.981094,
		0.990599,
		0.996042,
		0.998717,
		0.999741,
		0.999984,
		1,
		0.999985,
		0.999797,
		0.999098,
		0.997501,
		0.994638,
		0.990206,
		0.983977,
		0.975798,
		0.965582,
		0.953294,
		0.938936,
		0.92254,
		0.904151,
		0.883821,
		0.861608,
		0.837568,
		0.811757,
		0.784233,
		0.755064,
		0.724327,
		0.692115,
		0.658546,
		0.623765,
		0.587947,
		0.551297,
		0.514056,
		0.476489,
		0.438887,
		0.40156,
		0.364821,
		0.328987,
		0.294361,
		0.261224,
		0.229828,
		0.200385,
		0.173066,
		0.14799,
		0.12523,
		0.104807,
		0.086698,
		0.070835,
		0.057115,
		0.045407,
		0.035553,
		0.027382,
		0.020712,
		0.01536,
		0.011143,
		0.007888,
		0.005431,
		0.003623,
		0.002331,
		0.001437,
		0.000842,
		0.000464,
		0.000237,
		0.00011,
		4.5e-005,
		1.6e-005,
		0.024223,
		0.085389,
		0.169394,
		0.265625,
		0.366214,
		0.465459,
		0.559361,
		0.645261,
		0.721547,
		0.787425,
		0.842738,
		0.887815,
		0.92335,
		0.950308,
		0.969836,
		0.983187,
		0.991651,
		0.99649,
		0.998863,
		0.999771,
		0.999985,
		1,
		0.999987,
		0.999818,
		0.999192,
		0.99775,
		0.99515,
		0.991097,
		0.98536,
		0.977766,
		0.968199,
		0.956583,
		0.942872,
		0.927046,
		0.909096,
		0.889025,
		0.866844,
		0.842572,
		0.81624,
		0.787895,
		0.757607,
		0.725472,
		0.691617,
		0.656209,
		0.619449,
		0.581581,
		0.542884,
		0.50367,
		0.464277,
		0.425058,
		0.386375,
		0.348584,
		0.312024,
		0.27701,
		0.243819,
		0.212684,
		0.183787,
		0.15726,
		0.133176,
		0.11156,
		0.092384,
		0.075577,
		0.061029,
		0.0486,
		0.038125,
		0.029425,
		0.02231,
		0.016587,
		0.012067,
		0.008567,
		0.005917,
		0.00396,
		0.002555,
		0.001579,
		0.000927,
		0.000512,
		0.000261,
		0.000121,
		4.9e-005,
		1.6e-005,
		0.010606,
		0.058623,
		0.132788,
		0.222223,
		0.318742,
		0.416289,
		0.510503,
		0.598361,
		0.677898,
		0.747985,
		0.808146,
		0.858415,
		0.899218,
		0.931276,
		0.955526,
		0.973047,
		0.984999,
		0.992561,
		0.996876,
		0.998989,
		0.999797,
		0.999987,
		1,
		0.999987,
		0.999832,
		0.999258,
		0.99793,
		0.995519,
		0.991729,
		0.986312,
		0.979067,
		0.969837,
		0.958497,
		0.944951,
		0.929125,
		0.910961,
		0.890421,
		0.867484,
		0.842154,
		0.814459,
		0.784462,
		0.752265,
		0.71801,
		0.681888,
		0.644135,
		0.605034,
		0.564909,
		0.524119,
		0.48305,
		0.442101,
		0.401677,
		0.362172,
		0.323959,
		0.287379,
		0.252728,
		0.220253,
		0.190145,
		0.162535,
		0.137498,
		0.115052,
		0.095162,
		0.077749,
		0.062693,
		0.049843,
		0.039027,
		0.030053,
		0.022723,
		0.016836,
		0.012195,
		0.008611,
		0.005905,
		0.003915,
		0.002495,
		0.001517,
		0.000871,
		0.000466,
		0.000227,
		9.8e-005,
		3.5e-005,
		0.00261,
		0.037057,
		0.100747,
		0.182603,
		0.27417,
		0.369091,
		0.462682,
		0.551596,
		0.63355,
		0.707107,
		0.771497,
		0.826481,
		0.87223,
		0.909232,
		0.938217,
		0.960083,
		0.975845,
		0.986574,
		0.99335,
		0.99721,
		0.999098,
		0.999818,
		0.999987,
		0.999994,
		0.999972,
		0.999806,
		0.999235,
		0.997932,
		0.995567,
		0.991829,
		0.986442,
		0.979167,
		0.969798,
		0.958161,
		0.944112,
		0.927533,
		0.908331,
		0.886448,
		0.861858,
		0.834575,
		0.80466,
		0.772221,
		0.737423,
		0.700485,
		0.661683,
		0.621342,
		0.579832,
		0.537563,
		0.494965,
		0.452485,
		0.410564,
		0.369632,
		0.330088,
		0.292291,
		0.256552,
		0.223122,
		0.192194,
		0.163895,
		0.138292,
		0.115392,
		0.095151,
		0.077474,
		0.062232,
		0.049261,
		0.038377,
		0.029379,
		0.022059,
		0.016209,
		0.011623,
		0.008105,
		0.005473,
		0.003559,
		0.002212,
		0.001301,
		0.000715,
		0.000359,
		0.00016,
		6e-005,
		1.7e-005,
		0.02057,
		0.073275,
		0.146884,
		0.232725,
		0.324181,
		0.416289,
		0.505413,
		0.588982,
		0.665277,
		0.73326,
		0.792437,
		0.842738,
		0.884434,
		0.918052,
		0.944313,
		0.964076,
		0.978292,
		0.987949,
		0.994038,
		0.997501,
		0.999192,
		0.999832,
		0.999972,
		0.999952,
		0.999877,
		0.999627,
		0.998935,
		0.997471,
		0.994885,
		0.990841,
		0.985025,
		0.977154,
		0.966979,
		0.954283,
		0.938887,
		0.92065,
		0.899472,
		0.875305,
		0.84815,
		0.818069,
		0.785187,
		0.749691,
		0.711837,
		0.67194,
		0.630373,
		0.587553,
		0.543936,
		0.5,
		0.456229,
		0.4131,
		0.371068,
		0.330552,
		0.291922,
		0.255494,
		0.221518,
		0.190179,
		0.161595,
		0.135819,
		0.112844,
		0.092608,
		0.075006,
		0.05989,
		0.047085,
		0.036394,
		0.027607,
		0.020506,
		0.014875,
		0.010502,
		0.007186,
		0.004739,
		0.002991,
		0.00179,
		0.001001,
		0.000513,
		0.000234,
		9e-005,
		2.7e-005,
		0.009016,
		0.05033,
		0.11513,
		0.19457,
		0.281814,
		0.371655,
		0.460201,
		0.544621,
		0.622942,
		0.693878,
		0.756699,
		0.811114,
		0.857183,
		0.895243,
		0.925841,
		0.949683,
		0.967586,
		0.980438,
		0.989153,
		0.994638,
		0.99775,
		0.999258,
		0.999806,
		0.999877,
		0.999758,
		0.999515,
		0.999005,
		0.997943,
		0.995977,
		0.992731,
		0.987832,
		0.980929,
		0.971702,
		0.95987,
		0.945195,
		0.92749,
		0.90662,
		0.882517,
		0.855177,
		0.824668,
		0.791133,
		0.754792,
		0.715937,
		0.674927,
		0.632181,
		0.588164,
		0.543376,
		0.498335,
		0.453558,
		0.409549,
		0.366781,
		0.325684,
		0.286631,
		0.249933,
		0.21583,
		0.184495,
		0.156029,
		0.130466,
		0.107782,
		0.087897,
		0.070687,
		0.055989,
		0.043615,
		0.033356,
		0.02499,
		0.018292,
		0.013038,
		0.009012,
		0.006008,
		0.003836,
		0.002323,
		0.001317,
		0.000685,
		0.000317,
		0.000124,
		3.7e-005,
		0.002221,
		0.031834,
		0.08736,
		0.159817,
		0.242186,
		0.32905,
		0.416289,
		0.500843,
		0.580504,
		0.653763,
		0.71967,
		0.777728,
		0.827802,
		0.870045,
		0.904839,
		0.932739,
		0.954428,
		0.970681,
		0.982326,
		0.990206,
		0.99515,
		0.99793,
		0.999235,
		0.999627,
		0.999515,
		0.999123,
		0.998467,
		0.997368,
		0.995509,
		0.992508,
		0.987957,
		0.981452,
		0.972616,
		0.961111,
		0.946649,
		0.929002,
		0.908012,
		0.883593,
		0.855742,
		0.824541,
		0.790158,
		0.752845,
		0.712934,
		0.67083,
		0.626996,
		0.581942,
		0.536205,
		0.490336,
		0.444878,
		0.400354,
		0.357244,
		0.315979,
		0.276925,
		0.240381,
		0.206571,
		0.175647,
		0.147689,
		0.122709,
		0.10066,
		0.081444,
		0.064916,
		0.0509,
		0.039191,
		0.029569,
		0.021804,
		0.015662,
		0.010914,
		0.007338,
		0.004727,
		0.00289,
		0.001654,
		0.00087,
		0.000407,
		0.000161,
		4.9e-005,
		0.017685,
		0.063557,
		0.128533,
		0.205443,
		0.28869,
		0.373956,
		0.457972,
		0.538323,
		0.613292,
		0.681729,
		0.742944,
		0.796619,
		0.842738,
		0.881522,
		0.913379,
		0.938862,
		0.95863,
		0.973413,
		0.983977,
		0.991097,
		0.995519,
		0.997932,
		0.998935,
		0.999005,
		0.998467,
		0.997474,
		0.995983,
		0.993768,
		0.990479,
		0.985703,
		0.979009,
		0.969981,
		0.958238,
		0.943456,
		0.925377,
		0.903823,
		0.878705,
		0.850026,
		0.817888,
		0.782486,
		0.744109,
		0.703131,
		0.659998,
		0.615215,
		0.569328,
		0.522909,
		0.47653,
		0.430753,
		0.386103,
		0.343062,
		0.302051,
		0.26342,
		0.227449,
		0.194338,
		0.164213,
		0.137128,
		0.11307,
		0.09197,
		0.073706,
		0.058116,
		0.045007,
		0.034161,
		0.025346,
		0.018324,
		0.012854,
		0.008703,
		0.005648,
		0.003479,
		0.002008,
		0.001064,
		0.000503,
		0.000201,
		6.2e-005,
		1.2e-005,
		0.007758,
		0.043674,
		0.100747,
		0.171689,
		0.250747,
		0.333432,
		0.416289,
		0.496716,
		0.572806,
		0.643226,
		0.707107,
		0.763962,
		0.813614,
		0.856131,
		0.891783,
		0.920994,
		0.944307,
		0.962349,
		0.975798,
		0.98536,
		0.991729,
		0.995567,
		0.997471,
		0.997943,
		0.997368,
		0.995983,
		0.993865,
		0.990908,
		0.986844,
		0.981295,
		0.973833,
		0.964028,
		0.951479,
		0.935839,
		0.916836,
		0.894285,
		0.868102,
		0.838304,
		0.805015,
		0.768466,
		0.72898,
		0.686971,
		0.642924,
		0.597379,
		0.550912,
		0.504116,
		0.457578,
		0.411863,
		0.367494,
		0.324939,
		0.284601,
		0.246807,
		0.211808,
		0.179776,
		0.150806,
		0.124925,
		0.102092,
		0.082212,
		0.065143,
		0.050704,
		0.038685,
		0.028856,
		0.020976,
		0.014799,
		0.010079,
		0.00658,
		0.00408,
		0.00237,
		0.001265,
		0.000602,
		0.000243,
		7.5e-005,
		1.5e-005,
		0.001913,
		0.027641,
		0.076461,
		0.140991,
		0.215353,
		0.294906,
		0.376031,
		0.455958,
		0.532609,
		0.604481,
		0.670542,
		0.730146,
		0.782964,
		0.828924,
		0.86816,
		0.900973,
		0.927789,
		0.94913,
		0.965582,
		0.977766,
		0.986312,
		0.991829,
		0.994885,
		0.995977,
		0.995509,
		0.993768,
		0.990908,
		0.986931,
		0.981682,
		0.974863,
		0.96609,
		0.95495,
		0.941046,
		0.924032,
		0.903638,
		0.879686,
		0.852106,
		0.820937,
		0.786334,
		0.748558,
		0.707972,
		0.665023,
		0.620228,
		0.574154,
		0.527398,
		0.480563,
		0.434238,
		0.388981,
		0.3453,
		0.303641,
		0.264379,
		0.22781,
		0.194153,
		0.163544,
		0.136047,
		0.111657,
		0.090306,
		0.071873,
		0.056194,
		0.043071,
		0.03228,
		0.023578,
		0.016717,
		0.011443,
		0.007511,
		0.004682,
		0.002735,
		0.001468,
		0.000703,
		0.000285,
		8.9e-005,
		1.7e-005,
		0.015366,
		0.055645,
		0.113386,
		0.182603,
		0.258527,
		0.337397,
		0.416289,
		0.492972,
		0.565786,
		0.63355,
		0.69547,
		0.751076,
		0.800162,
		0.842734,
		0.878971,
		0.909189,
		0.9338,
		0.953294,
		0.968199,
		0.979067,
		0.986442,
		0.990841,
		0.992731,
		0.992508,
		0.990479,
		0.986844,
		0.981682,
		0.974941,
		0.966438,
		0.955875,
		0.942893,
		0.927134,
		0.908275,
		0.886069,
		0.86036,
		0.831105,
		0.798372,
		0.762347,
		0.723327,
		0.681705,
		0.637959,
		0.59263,
		0.546301,
		0.499578,
		0.453061,
		0.407332,
		0.362928,
		0.320332,
		0.279959,
		0.242148,
		0.207159,
		0.175172,
		0.146287,
		0.120532,
		0.097872,
		0.078208,
		0.061398,
		0.047255,
		0.035565,
		0.02609,
		0.018579,
		0.012776,
		0.008424,
		0.005276,
		0.003097,
		0.001671,
		0.000804,
		0.000328,
		0.000103,
		2e-005,
		0.006746,
		0.038254,
		0.088884,
		0.152563,
		0.224414,
		0.300551,
		0.377913,
		0.45413,
		0.527402,
		0.596407,
		0.660215,
		0.718225,
		0.770102,
		0.815734,
		0.855181,
		0.888645,
		0.916436,
		0.938936,
		0.956583,
		0.969837,
		0.979167,
		0.985025,
		0.987832,
		0.987957,
		0.985703,
		0.981295,
		0.974863,
		0.966438,
		0.955945,
		0.943204,
		0.927957,
		0.909914,
		0.88881,
		0.864439,
		0.836688,
		0.805547,
		0.771123,
		0.733637,
		0.693416,
		0.650881,
		0.606532,
		0.560924,
		0.514646,
		0.468297,
		0.422467,
		0.377712,
		0.334543,
		0.293406,
		0.254678,
		0.218656,
		0.185558,
		0.155521,
		0.128609,
		0.104814,
		0.084068,
		0.066246,
		0.051181,
		0.038668,
		0.028478,
		0.02036,
		0.014057,
		0.009307,
		0.005853,
		0.003451,
		0.00187,
		0.000904,
		0.000371,
		0.000117,
		2.3e-005,
		0.001665,
		0.024223,
		0.067471,
		0.125272,
		0.192655,
		0.265625,
		0.341002,
		0.416289,
		0.489559,
		0.559361,
		0.624639,
		0.684668,
		0.738994,
		0.787387,
		0.829796,
		0.86631,
		0.897131,
		0.92254,
		0.942872,
		0.958497,
		0.969798,
		0.977154,
		0.980929,
		0.981452,
		0.979009,
		0.973833,
		0.96609,
		0.955875,
		0.943204,
		0.928019,
		0.910189,
		0.889535,
		0.865876,
		0.839077,
		0.80908,
		0.775926,
		0.739762,
		0.700845,
		0.65953,
		0.616261,
		0.571549,
		0.52595,
		0.480048,
		0.434424,
		0.389643,
		0.346229,
		0.304652,
		0.265316,
		0.228551,
		0.194608,
		0.163659,
		0.135801,
		0.111057,
		0.089384,
		0.070681,
		0.0548,
		0.041551,
		0.030711,
		0.022037,
		0.015271,
		0.010148,
		0.006407,
		0.003792,
		0.002063,
		0.001001,
		0.000412,
		0.000131,
		2.6e-005,
		0.013475,
		0.049119,
		0.100747,
		0.163312,
		0.232725,
		0.3057,
		0.379627,
		0.452462,
		0.522637,
		0.588979,
		0.650647,
		0.707073,
		0.757914,
		0.803008,
		0.842337,
		0.875992,
		0.904151,
		0.927046,
		0.944951,
		0.958161,
		0.966979,
		0.971702,
		0.972616,
		0.969981,
		0.964028,
		0.95495,
		0.942893,
		0.927957,
		0.910189,
		0.88959,
		0.866116,
		0.839707,
		0.810321,
		0.777975,
		0.742769,
		0.704898,
		0.664653,
		0.622414,
		0.578638,
		0.533837,
		0.48856,
		0.443369,
		0.398819,
		0.355436,
		0.313702,
		0.274039,
		0.236802,
		0.202269,
		0.170644,
		0.142051,
		0.116543,
		0.094104,
		0.074658,
		0.058074,
		0.044179,
		0.032763,
		0.023589,
		0.016402,
		0.010938,
		0.00693,
		0.004116,
		0.002247,
		0.001095,
		0.000452,
		0.000144,
		2.9e-005,
		0.00592,
		0.033782,
		0.078987,
		0.136427,
		0.201932,
		0.272124,
		0.344293,
		0.416289,
		0.486434,
		0.553448,
		0.616381,
		0.674555,
		0.727518,
		0.775,
		0.816874,
		0.853125,
		0.883821,
		0.909096,
		0.929125,
		0.944112,
		0.954283,
		0.95987,
		0.961111,
		0.958238,
		0.951479,
		0.941046,
		0.927134,
		0.909914,
		0.889535,
		0.866116,
		0.839755,
		0.810528,
		0.778514,
		0.743821,
		0.706621,
		0.667159,
		0.62576,
		0.582824,
		0.53881,
		0.494223,
		0.449592,
		0.405447,
		0.362304,
		0.320642,
		0.280893,
		0.243424,
		0.208535,
		0.176453,
		0.147328,
		0.121239,
		0.098195,
		0.078143,
		0.060974,
		0.04653,
		0.034615,
		0.025001,
		0.01744,
		0.011667,
		0.007416,
		0.004419,
		0.002421,
		0.001183,
		0.000491,
		0.000157,
		3.1e-005,
		0.001462,
		0.021402,
		0.059973,
		0.112019,
		0.173305,
		0.24037,
		0.310414,
		0.381192,
		0.450927,
		0.518237,
		0.582068,
		0.641638,
		0.696388,
		0.745942,
		0.790063,
		0.828631,
		0.861608,
		0.889025,
		0.910961,
		0.927533,
		0.938887,
		0.945195,
		0.946649,
		0.943456,
		0.935839,
		0.924032,
		0.908275,
		0.88881,
		0.865876,
		0.839707,
		0.810528,
		0.778554,
		0.743996,
		0.707073,
		0.668034,
		0.627179,
		0.584866,
		0.541506,
		0.497557,
		0.453504,
		0.409844,
		0.367068,
		0.325642,
		0.285991,
		0.248488,
		0.213445,
		0.181102,
		0.151632,
		0.125134,
		0.101641,
		0.081119,
		0.063481,
		0.048585,
		0.036251,
		0.026261,
		0.018373,
		0.012329,
		0.00786,
		0.004698,
		0.002582,
		0.001266,
		0.000526,
		0.000169,
		3.4e-005,
		0.011912,
		0.043674,
		0.090096,
		0.146884,
		0.210511,
		0.278094,
		0.347303,
		0.416269,
		0.483514,
		0.547885,
		0.608498,
		0.664693,
		0.715985,
		0.762033,
		0.802607,
		0.837568,
		0.866844,
		0.890421,
		0.908331,
		0.92065,
		0.92749,
		0.929002,
		0.925377,
		0.916836,
		0.903638,
		0.886069,
		0.864439,
		0.839077,
		0.810321,
		0.778514,
		0.743996,
		0.707107,
		0.668179,
		0.62755,
		0.585577,
		0.542647,
		0.499179,
		0.455619,
		0.412426,
		0.370055,
		0.32895,
		0.289518,
		0.252127,
		0.217088,
		0.18465,
		0.154997,
		0.128245,
		0.104444,
		0.083582,
		0.065587,
		0.050336,
		0.037661,
		0.027359,
		0.019196,
		0.012918,
		0.008259,
		0.004951,
		0.002729,
		0.001342,
		0.00056,
		0.00018,
		3.6e-005,
		0.005236,
		0.030049,
		0.070648,
		0.122695,
		0.182602,
		0.247418,
		0.314732,
		0.38259,
		0.449423,
		0.513985,
		0.575293,
		0.632584,
		0.685268,
		0.732899,
		0.775143,
		0.811757,
		0.842572,
		0.867484,
		0.886448,
		0.899472,
		0.90662,
		0.908012,
		0.903823,
		0.894285,
		0.879686,
		0.86036,
		0.836688,
		0.80908,
		0.777975,
		0.743821,
		0.707073,
		0.668179,
		0.627578,
		0.585694,
		0.542944,
		0.499742,
		0.456511,
		0.413678,
		0.371666,
		0.330886,
		0.291724,
		0.254529,
		0.219602,
		0.187192,
		0.157487,
		0.130612,
		0.10663,
		0.085543,
		0.067296,
		0.051781,
		0.038844,
		0.028292,
		0.019903,
		0.01343,
		0.008609,
		0.005175,
		0.00286,
		0.00141,
		0.00059,
		0.00019,
		3.8e-005,
		0.001294,
		0.019046,
		0.053654,
		0.100747,
		0.156684,
		0.218452,
		0.283568,
		0.349999,
		0.416091,
		0.480503,
		0.542159,
		0.600193,
		0.653913,
		0.702768,
		0.746323,
		0.784233,
		0.81624,
		0.842154,
		0.861858,
		0.875305,
		0.882517,
		0.883593,
		0.878705,
		0.868102,
		0.852106,
		0.831105,
		0.805547,
		0.775926,
		0.742769,
		0.706621,
		0.668034,
		0.62755,
		0.585694,
		0.542966,
		0.499835,
		0.456743,
		0.414112,
		0.372344,
		0.331824,
		0.292911,
		0.255931,
		0.22117,
		0.188866,
		0.159201,
		0.132304,
		0.108244,
		0.087033,
		0.068626,
		0.052929,
		0.039801,
		0.029061,
		0.020495,
		0.013864,
		0.00891,
		0.005369,
		0.002975,
		0.001471,
		0.000617,
		0.000199,
		4e-005,
		0.010606,
		0.039085,
		0.081038,
		0.132781,
		0.191247,
		0.253886,
		0.31859,
		0.383622,
		0.447554,
		0.509214,
		0.567636,
		0.622029,
		0.671739,
		0.71623,
		0.755064,
		0.787895,
		0.814459,
		0.834575,
		0.84815,
		0.855177,
		0.855742,
		0.850026,
		0.838304,
		0.820937,
		0.798372,
		0.771123,
		0.739762,
		0.704898,
		0.667159,
		0.627179,
		0.585577,
		0.542944,
		0.499835,
		0.456761,
		0.414183,
		0.37252,
		0.332147,
		0.293409,
		0.256608,
		0.222012,
		0.18984,
		0.160268,
		0.133415,
		0.109353,
		0.088095,
		0.069605,
		0.053799,
		0.040544,
		0.029671,
		0.020973,
		0.014221,
		0.009162,
		0.005534,
		0.003074,
		0.001523,
		0.00064,
		0.000208,
		4.2e-005,
		0.004665,
		0.026901,
		0.063554,
		0.110907,
		0.165839,
		0.225742,
		0.288436,
		0.352107,
		0.415238,
		0.476563,
		0.535021,
		0.589719,
		0.639906,
		0.684951,
		0.724327,
		0.757607,
		0.784462,
		0.80466,
		0.818069,
		0.824668,
		0.824541,
		0.817888,
		0.805015,
		0.786334,
		0.762347,
		0.733637,
		0.700845,
		0.664653,
		0.62576,
		0.584866,
		0.542647,
		0.499742,
		0.456743,
		0.414183,
		0.372533,
		0.332201,
		0.293537,
		0.256841,
		0.222364,
		0.19031,
		0.160839,
		0.134061,
		0.110041,
		0.088791,
		0.070277,
		0.054418,
		0.041091,
		0.030132,
		0.021344,
		0.014503,
		0.009364,
		0.005669,
		0.003156,
		0.001568,
		0.000661,
		0.000215,
		4.4e-005,
		0.001154,
		0.017057,
		0.048276,
		0.091062,
		0.142252,
		0.199181,
		0.259602,
		0.321622,
		0.38364,
		0.444298,
		0.502441,
		0.557079,
		0.607367,
		0.652581,
		0.692115,
		0.725472,
		0.752265,
		0.772221,
		0.785187,
		0.791133,
		0.790158,
		0.782486,
		0.768466,
		0.748558,
		0.723327,
		0.693416,
		0.65953,
		0.622414,
		0.582824,
		0.541506,
		0.499179,
		0.456511,
		0.414112,
		0.37252,
		0.332201,
		0.293547,
		0.25688,
		0.222455,
		0.190471,
		0.161077,
		0.134371,
		0.110408,
		0.089194,
		0.070692,
		0.054822,
		0.041464,
		0.030459,
		0.021616,
		0.014716,
		0.009521,
		0.005776,
		0.003222,
		0.001604,
		0.000677,
		0.000221,
		4.5e-005,
		0.009502,
		0.035173,
		0.07324,
		0.120501,
		0.174237,
		0.232137,
		0.292232,
		0.352837,
		0.412507,
		0.469991,
		0.524209,
		0.574223,
		0.619229,
		0.658546,
		0.691617,
		0.71801,
		0.737423,
		0.749691,
		0.754792,
		0.752845,
		0.744109,
		0.72898,
		0.707972,
		0.681705,
		0.650881,
		0.616261,
		0.578638,
		0.53881,
		0.497557,
		0.455619,
		0.413678,
		0.372344,
		0.332147,
		0.293537,
		0.25688,
		0.222461,
		0.190498,
		0.161138,
		0.134478,
		0.110561,
		0.089387,
		0.070914,
		0.055057,
		0.041695,
		0.030672,
		0.021801,
		0.014867,
		0.009635,
		0.005856,
		0.003273,
		0.001633,
		0.000691,
		0.000225,
		4.6e-005,
		0.004181,
		0.024212,
		0.057429,
		0.100594,
		0.150937,
		0.206085,
		0.263995,
		0.322902,
		0.381271,
		0.437764,
		0.491211,
		0.540591,
		0.585024,
		0.623765,
		0.656209,
		0.681888,
		0.700485,
		0.711837,
		0.715937,
		0.712934,
		0.703131,
		0.686971,
		0.665023,
		0.637959,
		0.606532,
		0.571549,
		0.533837,
		0.494223,
		0.453504,
		0.412426,
		0.371666,
		0.331824,
		0.293409,
		0.256841,
		0.222455,
		0.190498,
		0.161143,
		0.134495,
		0.1106,
		0.089454,
		0.071006,
		0.055169,
		0.041817,
		0.030795,
		0.021915,
		0.014965,
		0.009713,
		0.005913,
		0.003311,
		0.001654,
		0.000701,
		0.000229,
		4.7e-005,
		0.001034,
		0.015352,
		0.043608,
		0.082529,
		0.129299,
		0.181483,
		0.236966,
		0.293903,
		0.350675,
		0.405858,
		0.458198,
		0.5066,
		0.550115,
		0.587947,
		0.619449,
		0.644135,
		0.661683,
		0.67194,
		0.674927,
		0.67083,
		0.659998,
		0.642924,
		0.620228,
		0.59263,
		0.560924,
		0.52595,
		0.48856,
		0.449592,
		0.409844,
		0.370055,
		0.330886,
		0.292911,
		0.256608,
		0.222364,
		0.190471,
		0.161138,
		0.134495,
		0.110603,
		0.089464,
		0.071029,
		0.055207,
		0.041868,
		0.030854,
		0.021976,
		0.015022,
		0.009761,
		0.00595,
		0.003336,
		0.001669,
		0.000709,
		0.000232,
		4.7e-005,
		0.008551,
		0.031753,
		0.066304,
		0.109339,
		0.158365,
		0.211195,
		0.265908,
		0.320804,
		0.374377,
		0.425297,
		0.472399,
		0.514679,
		0.551297,
		0.581581,
		0.605034,
		0.621342,
		0.630373,
		0.632181,
		0.626996,
		0.615215,
		0.597379,
		0.574154,
		0.546301,
		0.514646,
		0.480048,
		0.443369,
		0.405447,
		0.367068,
		0.32895,
		0.291724,
		0.255931,
		0.222012,
		0.19031,
		0.161077,
		0.134478,
		0.1106,
		0.089464,
		0.071031,
		0.055213,
		0.041881,
		0.030874,
		0.022001,
		0.015049,
		0.009787,
		0.005971,
		0.003351,
		0.001679,
		0.000714,
		0.000234,
		4.8e-005,
		0.003761,
		0.021838,
		0.05191,
		0.091069,
		0.136762,
		0.186735,
		0.238991,
		0.291752,
		0.343437,
		0.392648,
		0.438158,
		0.478918,
		0.514056,
		0.542884,
		0.564909,
		0.579832,
		0.587553,
		0.588164,
		0.581942,
		0.569328,
		0.550912,
		0.527398,
		0.499578,
		0.468297,
		0.434424,
		0.398819,
		0.362304,
		0.325642,
		0.289518,
		0.254529,
		0.22117,
		0.18984,
		0.160839,
		0.134371,
		0.110561,
		0.089454,
		0.071029,
		0.055213,
		0.041882,
		0.030877,
		0.022007,
		0.015058,
		0.009797,
		0.005981,
		0.00336,
		0.001685,
		0.000717,
		0.000235,
		4.8e-005,
		0.00093,
		0.013828,
		0.039335,
		0.074497,
		0.116705,
		0.163637,
		0.213225,
		0.263617,
		0.313165,
		0.360406,
		0.404065,
		0.443057,
		0.476489,
		0.50367,
		0.524119,
		0.537563,
		0.543936,
		0.543376,
		0.536205,
		0.522909,
		0.504116,
		0.480563,
		0.453061,
		0.422467,
		0.389643,
		0.355436,
		0.320642,
		0.285991,
		0.252127,
		0.219602,
		0.188866,
		0.160268,
		0.134061,
		0.110408,
		0.089387,
		0.071006,
		0.055207,
		0.041881,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005984,
		0.003363,
		0.001687,
		0.000719,
		0.000236,
		4.8e-005,
		0.007687,
		0.028563,
		0.059631,
		0.098223,
		0.141955,
		0.188689,
		0.236507,
		0.283695,
		0.32874,
		0.370327,
		0.407343,
		0.438887,
		0.464277,
		0.48305,
		0.494965,
		0.5,
		0.498335,
		0.490336,
		0.47653,
		0.457578,
		0.434238,
		0.407332,
		0.377712,
		0.346229,
		0.313702,
		0.280893,
		0.248488,
		0.217088,
		0.187192,
		0.159201,
		0.133415,
		0.110041,
		0.089194,
		0.070914,
		0.055169,
		0.041868,
		0.030874,
		0.022007,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.003372,
		0.019574,
		0.046472,
		0.081344,
		0.121743,
		0.165466,
		0.210532,
		0.255173,
		0.297832,
		0.337163,
		0.372038,
		0.40156,
		0.425058,
		0.442101,
		0.452485,
		0.456229,
		0.453558,
		0.444878,
		0.430753,
		0.411863,
		0.388981,
		0.362928,
		0.334543,
		0.304652,
		0.274039,
		0.243424,
		0.213445,
		0.18465,
		0.157487,
		0.132304,
		0.109353,
		0.088791,
		0.070692,
		0.055057,
		0.041817,
		0.030854,
		0.022001,
		0.015058,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.000831,
		0.01234,
		0.035019,
		0.066092,
		0.103053,
		0.143636,
		0.185806,
		0.227748,
		0.267868,
		0.304801,
		0.337415,
		0.364821,
		0.386375,
		0.401677,
		0.410564,
		0.4131,
		0.409549,
		0.400354,
		0.386103,
		0.367494,
		0.3453,
		0.320332,
		0.293406,
		0.265316,
		0.236802,
		0.208535,
		0.181102,
		0.154997,
		0.130612,
		0.108244,
		0.088095,
		0.070277,
		0.054822,
		0.041695,
		0.030795,
		0.021976,
		0.015049,
		0.009797,
		0.005984,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.006823,
		0.025261,
		0.052485,
		0.085931,
		0.123278,
		0.162442,
		0.201568,
		0.239036,
		0.273471,
		0.303745,
		0.328987,
		0.348584,
		0.362172,
		0.369632,
		0.371068,
		0.366781,
		0.357244,
		0.343062,
		0.324939,
		0.303641,
		0.279959,
		0.254678,
		0.228551,
		0.202269,
		0.176453,
		0.151632,
		0.128245,
		0.10663,
		0.087033,
		0.069605,
		0.054418,
		0.041464,
		0.030672,
		0.021915,
		0.015022,
		0.009787,
		0.005981,
		0.003363,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.002974,
		0.017177,
		0.040531,
		0.070417,
		0.104466,
		0.140548,
		0.176778,
		0.21152,
		0.243396,
		0.271292,
		0.294361,
		0.312024,
		0.323959,
		0.330088,
		0.330552,
		0.325684,
		0.315979,
		0.302051,
		0.284601,
		0.264379,
		0.242148,
		0.218656,
		0.194608,
		0.170644,
		0.147328,
		0.125134,
		0.104444,
		0.085543,
		0.068626,
		0.053799,
		0.041091,
		0.030459,
		0.021801,
		0.014965,
		0.009761,
		0.005971,
		0.00336,
		0.001687,
		0.000719,
		0.000237,
		4.9e-005,
		0.000727,
		0.010731,
		0.030225,
		0.05654,
		0.087259,
		0.120219,
		0.15351,
		0.18549,
		0.214785,
		0.240301,
		0.261224,
		0.27701,
		0.287379,
		0.292291,
		0.291922,
		0.286631,
		0.276925,
		0.26342,
		0.246807,
		0.22781,
		0.207159,
		0.185558,
		0.163659,
		0.142051,
		0.121239,
		0.101641,
		0.083582,
		0.067296,
		0.052929,
		0.040544,
		0.030132,
		0.021616,
		0.014867,
		0.009713,
		0.00595,
		0.003351,
		0.001685,
		0.000719,
		0.000237,
		4.9e-005,
		0.005873,
		0.021549,
		0.044312,
		0.071706,
		0.101537,
		0.131883,
		0.161098,
		0.187825,
		0.210995,
		0.229828,
		0.243819,
		0.252728,
		0.256552,
		0.255494,
		0.249933,
		0.240381,
		0.227449,
		0.211808,
		0.194153,
		0.175172,
		0.155521,
		0.135801,
		0.116543,
		0.098195,
		0.081119,
		0.065587,
		0.051781,
		0.039801,
		0.029671,
		0.021344,
		0.014716,
		0.009635,
		0.005913,
		0.003336,
		0.001679,
		0.000717,
		0.000236,
		4.9e-005,
		0.00253,
		0.014463,
		0.033727,
		0.057832,
		0.084562,
		0.111989,
		0.138472,
		0.162676,
		0.183562,
		0.200385,
		0.212684,
		0.220253,
		0.223122,
		0.221518,
		0.21583,
		0.206571,
		0.194338,
		0.179776,
		0.163544,
		0.146287,
		0.128609,
		0.111057,
		0.094104,
		0.078143,
		0.063481,
		0.050336,
		0.038844,
		0.029061,
		0.020973,
		0.014503,
		0.009521,
		0.005856,
		0.003311,
		0.001669,
		0.000714,
		0.000235,
		4.8e-005,
		0.000611,
		0.008906,
		0.024756,
		0.045641,
		0.069331,
		0.093895,
		0.11771,
		0.139462,
		0.158149,
		0.173066,
		0.183787,
		0.190145,
		0.192194,
		0.190179,
		0.184495,
		0.175647,
		0.164213,
		0.150806,
		0.136047,
		0.120532,
		0.104814,
		0.089384,
		0.074658,
		0.060974,
		0.048585,
		0.037661,
		0.028292,
		0.020495,
		0.014221,
		0.009364,
		0.005776,
		0.003273,
		0.001654,
		0.000709,
		0.000234,
		4.8e-005,
		0.004798,
		0.01735,
		0.035114,
		0.055852,
		0.07764,
		0.098873,
		0.118271,
		0.134865,
		0.14799,
		0.15726,
		0.162535,
		0.163895,
		0.161595,
		0.156029,
		0.147689,
		0.137128,
		0.124925,
		0.111657,
		0.097872,
		0.084068,
		0.070681,
		0.058074,
		0.04653,
		0.036251,
		0.027359,
		0.019903,
		0.013864,
		0.009162,
		0.005669,
		0.003222,
		0.001633,
		0.000701,
		0.000232,
		4.8e-005,
		0.002032,
		0.011432,
		0.026203,
		0.044105,
		0.063227,
		0.08199,
		0.099147,
		0.11377,
		0.12523,
		0.133176,
		0.137498,
		0.138292,
		0.135819,
		0.130466,
		0.122709,
		0.11307,
		0.102092,
		0.090306,
		0.078208,
		0.066246,
		0.0548,
		0.044179,
		0.034615,
		0.026261,
		0.019196,
		0.01343,
		0.00891,
		0.005534,
		0.003156,
		0.001604,
		0.000691,
		0.000229,
		4.7e-005,
		0.000482,
		0.006903,
		0.018834,
		0.03404,
		0.050629,
		0.067052,
		0.082099,
		0.09488,
		0.104807,
		0.11156,
		0.115052,
		0.115392,
		0.112844,
		0.107782,
		0.10066,
		0.09197,
		0.082212,
		0.071873,
		0.061398,
		0.051181,
		0.041551,
		0.032763,
		0.025001,
		0.018373,
		0.012918,
		0.008609,
		0.005369,
		0.003074,
		0.001568,
		0.000677,
		0.000225,
		4.7e-005,
		0.003642,
		0.012909,
		0.02558,
		0.039786,
		0.054013,
		0.067091,
		0.07817,
		0.086698,
		0.092384,
		0.095162,
		0.095151,
		0.092608,
		0.087897,
		0.081444,
		0.073706,
		0.065143,
		0.056194,
		0.047255,
		0.038668,
		0.030711,
		0.023589,
		0.01744,
		0.012329,
		0.008259,
		0.005175,
		0.002975,
		0.001523,
		0.000661,
		0.000221,
		4.6e-005,
		0.001509,
		0.008309,
		0.018622,
		0.03061,
		0.042798,
		0.054055,
		0.063573,
		0.070835,
		0.075577,
		0.077749,
		0.077474,
		0.075006,
		0.070687,
		0.064916,
		0.058116,
		0.050704,
		0.043071,
		0.035565,
		0.028478,
		0.022037,
		0.016402,
		0.011667,
		0.00786,
		0.004951,
		0.00286,
		0.001471,
		0.00064,
		0.000215,
		4.5e-005,
		0.000349,
		0.004895,
		0.013042,
		0.022989,
		0.0333,
		0.04289,
		0.050989,
		0.057115,
		0.061029,
		0.062693,
		0.062232,
		0.05989,
		0.055989,
		0.0509,
		0.045007,
		0.038685,
		0.03228,
		0.02609,
		0.02036,
		0.015271,
		0.010938,
		0.007416,
		0.004698,
		0.002729,
		0.00141,
		0.000617,
		0.000208,
		4.4e-005,
		0.002517,
		0.008699,
		0.016787,
		0.025391,
		0.033468,
		0.040291,
		0.045407,
		0.0486,
		0.049843,
		0.049261,
		0.047085,
		0.043615,
		0.039191,
		0.034161,
		0.028856,
		0.023578,
		0.018579,
		0.014057,
		0.010148,
		0.00693,
		0.004419,
		0.002582,
		0.001342,
		0.00059,
		0.000199,
		4.2e-005,
		0.001015,
		0.005442,
		0.011859,
		0.018926,
		0.025644,
		0.031327,
		0.035553,
		0.038125,
		0.039027,
		0.038377,
		0.036394,
		0.033356,
		0.029569,
		0.025346,
		0.020976,
		0.016717,
		0.012776,
		0.009307,
		0.006407,
		0.004116,
		0.002421,
		0.001266,
		0.00056,
		0.00019,
		4e-005,
		0.000228,
		0.003111,
		0.008047,
		0.013746,
		0.019257,
		0.023933,
		0.027382,
		0.029425,
		0.030053,
		0.029379,
		0.027607,
		0.02499,
		0.021804,
		0.018324,
		0.014799,
		0.011443,
		0.008424,
		0.005853,
		0.003792,
		0.002247,
		0.001183,
		0.000526,
		0.00018,
		3.8e-005,
		0.00155,
		0.005191,
		0.009688,
		0.014139,
		0.017935,
		0.020712,
		0.02231,
		0.022723,
		0.022059,
		0.020506,
		0.018292,
		0.015662,
		0.012854,
		0.010079,
		0.007511,
		0.005276,
		0.003451,
		0.002063,
		0.001095,
		0.000491,
		0.000169,
		3.6e-005,
		0.000604,
		0.003134,
		0.006591,
		0.010121,
		0.013156,
		0.01536,
		0.016587,
		0.016836,
		0.016209,
		0.014875,
		0.013038,
		0.010914,
		0.008703,
		0.00658,
		0.004682,
		0.003097,
		0.00187,
		0.001001,
		0.000452,
		0.000157,
		3.4e-005,
		0.000131,
		0.001726,
		0.004295,
		0.007038,
		0.009423,
		0.011143,
		0.012067,
		0.012195,
		0.011623,
		0.010502,
		0.009012,
		0.007338,
		0.005648,
		0.00408,
		0.002735,
		0.001671,
		0.000904,
		0.000412,
		0.000144,
		3.1e-005,
		0.000826,
		0.002653,
		0.004732,
		0.00657,
		0.007888,
		0.008567,
		0.008611,
		0.008105,
		0.007186,
		0.006008,
		0.004727,
		0.003479,
		0.00237,
		0.001468,
		0.000804,
		0.000371,
		0.000131,
		2.9e-005,
		0.000308,
		0.001528,
		0.003057,
		0.004443,
		0.005431,
		0.005917,
		0.005905,
		0.005473,
		0.004739,
		0.003836,
		0.00289,
		0.002008,
		0.001265,
		0.000703,
		0.000328,
		0.000117,
		2.6e-005,
		6.4e-005,
		0.000799,
		0.001882,
		0.0029,
		0.003623,
		0.00396,
		0.003915,
		0.003559,
		0.002991,
		0.002323,
		0.001654,
		0.001064,
		0.000602,
		0.000285,
		0.000103,
		2.3e-005,
		0.000361,
		0.001091,
		0.001816,
		0.002331,
		0.002555,
		0.002495,
		0.002212,
		0.00179,
		0.001317,
		0.00087,
		0.000503,
		0.000243,
		8.9e-005,
		2e-005,
		0.000127,
		0.000586,
		0.001083,
		0.001437,
		0.001579,
		0.001517,
		0.001301,
		0.001001,
		0.000685,
		0.000407,
		0.000201,
		7.5e-005,
		1.7e-005,
		2.4e-005,
		0.000283,
		0.000608,
		0.000842,
		0.000927,
		0.000871,
		0.000715,
		0.000513,
		0.000317,
		0.000161,
		6.2e-005,
		1.5e-005,
		0.000116,
		0.000316,
		0.000464,
		0.000512,
		0.000466,
		0.000359,
		0.000234,
		0.000124,
		4.9e-005,
		1.2e-005,
		3.7e-005,
		0.000149,
		0.000237,
		0.000261,
		0.000227,
		0.00016,
		9e-005,
		3.7e-005,
		6.2e-005,
		0.00011,
		0.000121,
		9.8e-005,
		6e-005,
		2.7e-005,
		2.1e-005,
		4.5e-005,
		4.9e-005,
		3.5e-005,
		1.7e-005,
		1.6e-005,
		1.6e-005,
		1.6e-005,
		2.2e-005,
		2.1e-005,
		1.5e-005,
		3.5e-005,
		6e-005,
		7.9e-005,
		8.2e-005,
		6.2e-005,
		2.5e-005,
		2.4e-005,
		6e-005,
		0.000112,
		0.00017,
		0.000218,
		0.000237,
		0.000213,
		0.000144,
		5.2e-005,
		3.5e-005,
		9e-005,
		0.000178,
		0.00029,
		0.00041,
		0.000512,
		0.000567,
		0.000551,
		0.000451,
		0.000283,
		9.7e-005,
		1.1e-005,
		4.6e-005,
		0.000124,
		0.000254,
		0.000434,
		0.000649,
		0.000871,
		0.001064,
		0.001186,
		0.0012,
		0.001083,
		0.000834,
		0.000496,
		0.000162,
		1.3e-005,
		5.9e-005,
		0.000161,
		0.000339,
		0.000597,
		0.000926,
		0.001301,
		0.001682,
		0.002017,
		0.002251,
		0.002331,
		0.002217,
		0.001898,
		0.001398,
		0.000799,
		0.000253,
		1.6e-005,
		7.2e-005,
		0.000201,
		0.000431,
		0.000776,
		0.001235,
		0.00179,
		0.0024,
		0.003011,
		0.003555,
		0.00396,
		0.004154,
		0.004083,
		0.003714,
		0.003057,
		0.002176,
		0.001206,
		0.000371,
		1.9e-005,
		8.6e-005,
		0.000243,
		0.000527,
		0.000966,
		0.001568,
		0.002323,
		0.003197,
		0.004136,
		0.005068,
		0.005905,
		0.006556,
		0.00693,
		0.006952,
		0.00657,
		0.005775,
		0.00461,
		0.003192,
		0.001726,
		0.000519,
		2.2e-005,
		0.0001,
		0.000285,
		0.000627,
		0.001164,
		0.001918,
		0.00289,
		0.004055,
		0.005365,
		0.006746,
		0.008105,
		0.009337,
		0.010326,
		0.010961,
		0.011143,
		0.0108,
		0.009899,
		0.008464,
		0.006591,
		0.004461,
		0.002362,
		0.000697,
		2.4e-005,
		0.000114,
		0.000328,
		0.000728,
		0.001366,
		0.002279,
		0.003479,
		0.004955,
		0.006667,
		0.008547,
		0.010502,
		0.012416,
		0.01416,
		0.015596,
		0.016587,
		0.017011,
		0.016772,
		0.015814,
		0.014139,
		0.011818,
		0.009012,
		0.005983,
		0.003111,
		0.000902,
		2.7e-005,
		0.000127,
		0.000371,
		0.000829,
		0.00157,
		0.002644,
		0.00408,
		0.00588,
		0.008018,
		0.010433,
		0.013038,
		0.015716,
		0.018328,
		0.020718,
		0.022723,
		0.02418,
		0.024941,
		0.024886,
		0.023933,
		0.02206,
		0.019315,
		0.015835,
		0.011859,
		0.007742,
		0.003962,
		0.001132,
		3e-005,
		0.000141,
		0.000412,
		0.000929,
		0.001771,
		0.003007,
		0.004682,
		0.006814,
		0.009391,
		0.012367,
		0.015662,
		0.019164,
		0.02273,
		0.026196,
		0.029379,
		0.032087,
		0.034131,
		0.035336,
		0.035553,
		0.034677,
		0.03266,
		0.029527,
		0.025391,
		0.020468,
		0.015087,
		0.009702,
		0.004895,
		0.00138,
		3.3e-005,
		0.000154,
		0.000452,
		0.001025,
		0.001968,
		0.003363,
		0.005276,
		0.007741,
		0.010764,
		0.014314,
		0.018324,
		0.022689,
		0.027273,
		0.031906,
		0.036394,
		0.040526,
		0.04408,
		0.046839,
		0.0486,
		0.049186,
		0.048465,
		0.046366,
		0.04289,
		0.038126,
		0.032269,
		0.025625,
		0.018622,
		0.011815,
		0.005885,
		0.001639,
		3.5e-005,
		0.000166,
		0.000491,
		0.001117,
		0.002157,
		0.003708,
		0.005853,
		0.008648,
		0.012115,
		0.016241,
		0.020976,
		0.026229,
		0.031869,
		0.03773,
		0.043615,
		0.049299,
		0.054541,
		0.059088,
		0.062693,
		0.065122,
		0.066174,
		0.06569,
		0.063573,
		0.059802,
		0.054447,
		0.04768,
		0.039786,
		0.031171,
		0.022365,
		0.014018,
		0.006903,
		0.001902,
		3.7e-005,
		0.000177,
		0.000526,
		0.001205,
		0.002336,
		0.004037,
		0.006407,
		0.009521,
		0.013424,
		0.01812,
		0.023578,
		0.029723,
		0.036436,
		0.04356,
		0.0509,
		0.058225,
		0.06528,
		0.071791,
		0.077474,
		0.082051,
		0.08526,
		0.086869,
		0.086698,
		0.084624,
		0.080608,
		0.074699,
		0.067052,
		0.057933,
		0.047728,
		0.036943,
		0.026203,
		0.016247,
		0.007919,
		0.002161,
		3.9e-005,
		0.000188,
		0.00056,
		0.001285,
		0.002503,
		0.004345,
		0.00693,
		0.010351,
		0.014673,
		0.019924,
		0.02609,
		0.033116,
		0.0409,
		0.049296,
		0.058116,
		0.067132,
		0.076083,
		0.084677,
		0.092608,
		0.099561,
		0.105226,
		0.109311,
		0.11156,
		0.111767,
		0.109792,
		0.105574,
		0.099147,
		0.090652,
		0.080339,
		0.068577,
		0.055852,
		0.042764,
		0.030023,
		0.018437,
		0.008906,
		0.00241,
		4.1e-005,
		0.000197,
		0.00059,
		0.00136,
		0.002657,
		0.004631,
		0.007416,
		0.011126,
		0.015848,
		0.021629,
		0.028478,
		0.03636,
		0.045192,
		0.054846,
		0.065143,
		0.075864,
		0.086747,
		0.097494,
		0.107782,
		0.117269,
		0.125608,
		0.132457,
		0.137498,
		0.140451,
		0.141088,
		0.139251,
		0.134865,
		0.127953,
		0.118641,
		0.107171,
		0.093895,
		0.079284,
		0.063915,
		0.048471,
		0.033727,
		0.020541,
		0.009847,
		0.002646,
		4.3e-005,
		0.000206,
		0.000617,
		0.001426,
		0.002796,
		0.00489,
		0.00786,
		0.011839,
		0.016933,
		0.023214,
		0.030711,
		0.039411,
		0.049254,
		0.060128,
		0.071873,
		0.084279,
		0.097092,
		0.110012,
		0.122709,
		0.134822,
		0.145975,
		0.155789,
		0.163895,
		0.169947,
		0.173644,
		0.174741,
		0.173066,
		0.168534,
		0.161159,
		0.151061,
		0.138472,
		0.123736,
		0.107308,
		0.089745,
		0.071706,
		0.053931,
		0.037241,
		0.022525,
		0.010731,
		0.002868,
		4.4e-005,
		0.000213,
		0.00064,
		0.001485,
		0.00292,
		0.005121,
		0.008259,
		0.012483,
		0.01792,
		0.024662,
		0.032763,
		0.042233,
		0.053032,
		0.065071,
		0.078208,
		0.092251,
		0.106953,
		0.122023,
		0.137128,
		0.151896,
		0.165934,
		0.178832,
		0.190179,
		0.19958,
		0.206667,
		0.211121,
		0.212684,
		0.211175,
		0.206504,
		0.198682,
		0.187825,
		0.174161,
		0.158028,
		0.139866,
		0.120219,
		0.099716,
		0.07907,
		0.05906,
		0.040531,
		0.02438,
		0.011559,
		0.003077,
		4.5e-005,
		0.000219,
		0.000661,
		0.001535,
		0.003027,
		0.005324,
		0.008609,
		0.013053,
		0.018799,
		0.025961,
		0.034615,
		0.044794,
		0.056483,
		0.069614,
		0.084068,
		0.099669,
		0.116188,
		0.133344,
		0.150806,
		0.168203,
		0.185128,
		0.201151,
		0.21583,
		0.228728,
		0.239425,
		0.247536,
		0.252728,
		0.254737,
		0.253379,
		0.248563,
		0.240301,
		0.228713,
		0.214025,
		0.196569,
		0.176778,
		0.155176,
		0.132372,
		0.109042,
		0.085931,
		0.063833,
		0.043596,
		0.026116,
		0.01234,
		0.003275,
		4.6e-005,
		0.000224,
		0.000677,
		0.001578,
		0.003117,
		0.005496,
		0.00891,
		0.013546,
		0.019564,
		0.0271,
		0.036251,
		0.047072,
		0.059573,
		0.073709,
		0.089384,
		0.106443,
		0.124676,
		0.143817,
		0.163544,
		0.183489,
		0.203242,
		0.22236,
		0.240381,
		0.256833,
		0.271253,
		0.283205,
		0.292291,
		0.298172,
		0.300582,
		0.299339,
		0.294361,
		0.285665,
		0.273376,
		0.257723,
		0.239036,
		0.21774,
		0.194342,
		0.169425,
		0.143636,
		0.117674,
		0.09229,
		0.068277,
		0.046472,
		0.027762,
		0.01309,
		0.003469,
		4.7e-005,
		0.000228,
		0.000691,
		0.001612,
		0.003191,
		0.005638,
		0.009162,
		0.01396,
		0.020214,
		0.028075,
		0.037661,
		0.049052,
		0.062277,
		0.077319,
		0.094104,
		0.112501,
		0.13232,
		0.153312,
		0.175172,
		0.197539,
		0.220006,
		0.242125,
		0.26342,
		0.283397,
		0.30156,
		0.317429,
		0.330552,
		0.340527,
		0.347016,
		0.349758,
		0.348584,
		0.343419,
		0.334294,
		0.321343,
		0.304801,
		0.284996,
		0.262345,
		0.237338,
		0.210532,
		0.182539,
		0.154017,
		0.125665,
		0.098223,
		0.072466,
		0.049219,
		0.029357,
		0.013828,
		0.003662,
		4.8e-005,
		0.000232,
		0.000701,
		0.001639,
		0.00325,
		0.005752,
		0.009364,
		0.014298,
		0.020748,
		0.028884,
		0.038844,
		0.050725,
		0.064584,
		0.080424,
		0.098195,
		0.117792,
		0.139047,
		0.161731,
		0.185558,
		0.21018,
		0.235198,
		0.260166,
		0.284601,
		0.307995,
		0.32983,
		0.34959,
		0.366781,
		0.38095,
		0.391695,
		0.398686,
		0.401677,
		0.400512,
		0.395137,
		0.385597,
		0.372038,
		0.354703,
		0.333919,
		0.310092,
		0.283695,
		0.255258,
		0.225354,
		0.194598,
		0.163637,
		0.133151,
		0.103854,
		0.076504,
		0.05191,
		0.030947,
		0.014576,
		0.003862,
		4.8e-005,
		0.000234,
		0.000709,
		0.001658,
		0.003294,
		0.005838,
		0.009521,
		0.014563,
		0.021172,
		0.029533,
		0.039801,
		0.052095,
		0.06649,
		0.083014,
		0.101641,
		0.122288,
		0.144812,
		0.169008,
		0.194608,
		0.221283,
		0.248646,
		0.276259,
		0.303641,
		0.330276,
		0.35563,
		0.379165,
		0.400354,
		0.4187,
		0.43375,
		0.445116,
		0.452485,
		0.455628,
		0.454416,
		0.448814,
		0.438887,
		0.424797,
		0.40679,
		0.385195,
		0.360406,
		0.332875,
		0.3031,
		0.271617,
		0.238991,
		0.205814,
		0.172707,
		0.140319,
		0.109339,
		0.080506,
		0.054623,
		0.032575,
		0.015352,
		0.004071,
		4.8e-005,
		0.000235,
		0.000714,
		0.001672,
		0.003325,
		0.005901,
		0.009635,
		0.014759,
		0.021491,
		0.03003,
		0.040544,
		0.053172,
		0.068007,
		0.085099,
		0.104444,
		0.125984,
		0.1496,
		0.17511,
		0.202269,
		0.230768,
		0.260237,
		0.29025,
		0.320332,
		0.349969,
		0.37862,
		0.405731,
		0.430753,
		0.453154,
		0.472443,
		0.488182,
		0.5,
		0.507609,
		0.510811,
		0.5095,
		0.50367,
		0.493408,
		0.478887,
		0.460362,
		0.438158,
		0.412655,
		0.384281,
		0.3535,
		0.320804,
		0.286708,
		0.251747,
		0.216478,
		0.181483,
		0.147381,
		0.114839,
		0.084585,
		0.057429,
		0.034281,
		0.016175,
		0.004295,
		4.9e-005,
		0.000236,
		0.000717,
		0.001681,
		0.003345,
		0.005942,
		0.009713,
		0.014896,
		0.021718,
		0.030388,
		0.041091,
		0.053976,
		0.069157,
		0.086702,
		0.10663,
		0.128904,
		0.153429,
		0.180046,
		0.208535,
		0.238608,
		0.269916,
		0.302048,
		0.334543,
		0.366892,
		0.398556,
		0.428973,
		0.457578,
		0.48382,
		0.507174,
		0.527165,
		0.543376,
		0.555467,
		0.563181,
		0.566352,
		0.564909,
		0.558873,
		0.548352,
		0.533535,
		0.514679,
		0.492103,
		0.46617,
		0.437279,
		0.405858,
		0.372351,
		0.337219,
		0.300936,
		0.263995,
		0.22691,
		0.190228,
		0.154541,
		0.120501,
		0.088841,
		0.06039,
		0.036097,
		0.017057,
		0.004537,
		4.9e-005,
		0.000237,
		0.000719,
		0.001686,
		0.003356,
		0.005967,
		0.009761,
		0.014983,
		0.021866,
		0.030628,
		0.041464,
		0.054538,
		0.069976,
		0.087866,
		0.108244,
		0.131095,
		0.156346,
		0.183862,
		0.213445,
		0.244832,
		0.277694,
		0.311642,
		0.346229,
		0.38096,
		0.415302,
		0.448693,
		0.480563,
		0.510346,
		0.537498,
		0.561514,
		0.581942,
		0.598399,
		0.610581,
		0.618271,
		0.621342,
		0.619756,
		0.613563,
		0.602893,
		0.587947,
		0.568982,
		0.546305,
		0.520259,
		0.491211,
		0.459545,
		0.425657,
		0.38995,
		0.352837,
		0.314743,
		0.276112,
		0.237419,
		0.199181,
		0.161974,
		0.12645,
		0.093355,
		0.063554,
		0.03805,
		0.018011,
		0.004799,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003362,
		0.005979,
		0.009787,
		0.015031,
		0.021951,
		0.030771,
		0.041695,
		0.054895,
		0.070513,
		0.088647,
		0.109353,
		0.132632,
		0.158434,
		0.186644,
		0.217088,
		0.249525,
		0.283651,
		0.319097,
		0.355436,
		0.392189,
		0.428835,
		0.464821,
		0.499578,
		0.532534,
		0.563133,
		0.590853,
		0.615215,
		0.635805,
		0.652282,
		0.664385,
		0.67194,
		0.674863,
		0.673152,
		0.666884,
		0.656209,
		0.641333,
		0.622512,
		0.600037,
		0.574223,
		0.545402,
		0.513912,
		0.480096,
		0.444298,
		0.406867,
		0.368158,
		0.328548,
		0.288436,
		0.248267,
		0.208539,
		0.169822,
		0.132781,
		0.098192,
		0.066962,
		0.040162,
		0.019046,
		0.005084,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005984,
		0.009797,
		0.015052,
		0.021992,
		0.030844,
		0.041817,
		0.055094,
		0.070823,
		0.089116,
		0.110041,
		0.133616,
		0.159806,
		0.18852,
		0.219602,
		0.252835,
		0.287937,
		0.324562,
		0.362304,
		0.400704,
		0.439258,
		0.477426,
		0.514646,
		0.55035,
		0.58398,
		0.615002,
		0.642924,
		0.667309,
		0.687787,
		0.704067,
		0.715937,
		0.723272,
		0.726028,
		0.724239,
		0.71801,
		0.707502,
		0.692926,
		0.674529,
		0.652581,
		0.627366,
		0.599175,
		0.568298,
		0.535021,
		0.499627,
		0.462401,
		0.423631,
		0.383622,
		0.342703,
		0.301242,
		0.259662,
		0.218452,
		0.178194,
		0.139572,
		0.103402,
		0.070648,
		0.042454,
		0.020173,
		0.005396,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022005,
		0.030871,
		0.041868,
		0.055183,
		0.070972,
		0.089354,
		0.110408,
		0.134165,
		0.160605,
		0.189652,
		0.22117,
		0.254963,
		0.290769,
		0.328265,
		0.367068,
		0.406741,
		0.446798,
		0.486716,
		0.52595,
		0.56394,
		0.600132,
		0.633993,
		0.665023,
		0.692773,
		0.716858,
		0.736961,
		0.752845,
		0.764353,
		0.771408,
		0.774009,
		0.772221,
		0.766169,
		0.756027,
		0.742003,
		0.724327,
		0.703245,
		0.679005,
		0.651854,
		0.622029,
		0.589761,
		0.555271,
		0.518777,
		0.480503,
		0.440684,
		0.399576,
		0.357477,
		0.314732,
		0.271752,
		0.229034,
		0.187172,
		0.146884,
		0.109032,
		0.074644,
		0.044948,
		0.021402,
		0.005737,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041881,
		0.05521,
		0.071023,
		0.089445,
		0.110561,
		0.134412,
		0.160989,
		0.190229,
		0.222012,
		0.256157,
		0.292425,
		0.330511,
		0.370055,
		0.410642,
		0.451806,
		0.493048,
		0.533837,
		0.57363,
		0.611886,
		0.648076,
		0.681705,
		0.712321,
		0.73953,
		0.763003,
		0.782486,
		0.7978,
		0.808844,
		0.815588,
		0.818069,
		0.816383,
		0.810671,
		0.801108,
		0.787895,
		0.771243,
		0.751367,
		0.728476,
		0.702768,
		0.674429,
		0.643631,
		0.610534,
		0.575293,
		0.538067,
		0.499021,
		0.458348,
		0.416269,
		0.373058,
		0.329048,
		0.28465,
		0.24037,
		0.196828,
		0.154777,
		0.11513,
		0.078987,
		0.047667,
		0.022747,
		0.006111,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041882,
		0.055213,
		0.071031,
		0.089464,
		0.1106,
		0.134487,
		0.161121,
		0.19045,
		0.222364,
		0.256697,
		0.293225,
		0.331662,
		0.371666,
		0.412843,
		0.454751,
		0.496909,
		0.53881,
		0.579929,
		0.61974,
		0.657731,
		0.693416,
		0.726348,
		0.756135,
		0.782445,
		0.805015,
		0.823657,
		0.838252,
		0.848753,
		0.855177,
		0.857596,
		0.856127,
		0.850921,
		0.842154,
		0.830009,
		0.814675,
		0.796331,
		0.775143,
		0.751262,
		0.724819,
		0.695929,
		0.664693,
		0.631206,
		0.595562,
		0.557864,
		0.518237,
		0.476837,
		0.433863,
		0.389572,
		0.344293,
		0.298442,
		0.25254,
		0.207232,
		0.163312,
		0.121749,
		0.083718,
		0.050639,
		0.024223,
		0.006524,
		4.9e-005,
		0.000236,
		0.000719,
		0.001687,
		0.003362,
		0.005982,
		0.009797,
		0.015057,
		0.022005,
		0.030876,
		0.041881,
		0.055213,
		0.071031,
		0.089465,
		0.110603,
		0.134495,
		0.161143,
		0.190496,
		0.222455,
		0.256861,
		0.293501,
		0.332104,
		0.372344,
		0.413844,
		0.456182,
		0.498898,
		0.541506,
		0.583504,
		0.624385,
		0.663655,
		0.700845,
		0.735522,
		0.767303,
		0.795861,
		0.820937,
		0.842338,
		0.859942,
		0.873692,
		0.883593,
		0.889702,
		0.892122,
		0.890985,
		0.886448,
		0.878677,
		0.867837,
		0.854087,
		0.837568,
		0.818404,
		0.796695,
		0.772521,
		0.745942,
		0.717001,
		0.685735,
		0.652181,
		0.616381,
		0.578397,
		0.53832,
		0.496281,
		0.452462,
		0.407111,
		0.360554,
		0.313213,
		0.265625,
		0.218463,
		0.172561,
		0.128949,
		0.088884,
		0.053897,
		0.025848,
		0.006979,
		4.8e-005,
		0.000236,
		0.000717,
		0.001684,
		0.003356,
		0.005974,
		0.009787,
		0.015044,
		0.021992,
		0.030862,
		0.041868,
		0.055202,
		0.071023,
		0.089459,
		0.1106,
		0.134494,
		0.161143,
		0.190498,
		0.222461,
		0.25688,
		0.293546,
		0.332198,
		0.37252,
		0.414148,
		0.456678,
		0.499666,
		0.542647,
		0.585137,
		0.626654,
		0.666724,
		0.704898,
		0.740761,
		0.773948,
		0.804146,
		0.831105,
		0.85464,
		0.874633,
		0.891026,
		0.903823,
		0.913075,
		0.918878,
		0.921354,
		0.92065,
		0.916918,
		0.91031,
		0.900971,
		0.889025,
		0.874577,
		0.857704,
		0.83846,
		0.816874,
		0.792954,
		0.766692,
		0.738071,
		0.707073,
		0.673685,
		0.637912,
		0.599782,
		0.559361,
		0.516759,
		0.472143,
		0.425753,
		0.377913,
		0.32905,
		0.279714,
		0.230606,
		0.182603,
		0.136799,
		0.094539,
		0.057479,
		0.027641,
		0.007484,
		4.8e-005,
		0.000235,
		0.000714,
		0.001677,
		0.003345,
		0.005956,
		0.009761,
		0.015011,
		0.021951,
		0.030815,
		0.041817,
		0.055149,
		0.070972,
		0.089413,
		0.110561,
		0.134464,
		0.161121,
		0.190485,
		0.222455,
		0.256878,
		0.293546,
		0.332201,
		0.372533,
		0.414184,
		0.45676,
		0.49983,
		0.542944,
		0.585637,
		0.627445,
		0.667916,
		0.706621,
		0.74317,
		0.777215,
		0.808465,
		0.836688,
		0.861713,
		0.883435,
		0.901806,
		0.916836,
		0.928581,
		0.937137,
		0.942628,
		0.945195,
		0.944988,
		0.942154,
		0.936827,
		0.929125,
		0.919139,
		0.906936,
		0.892551,
		0.875992,
		0.857243,
		0.836264,
		0.813001,
		0.787387,
		0.759357,
		0.728847,
		0.695809,
		0.660215,
		0.622067,
		0.581406,
		0.538323,
		0.492972,
		0.445578,
		0.39646,
		0.346048,
		0.294906,
		0.243759,
		0.193529,
		0.145377,
		0.100747,
		0.061429,
		0.029626,
		0.008046,
		4.8e-005,
		0.000233,
		0.000709,
		0.001666,
		0.003325,
		0.005924,
		0.009713,
		0.014945,
		0.021866,
		0.03071,
		0.041695,
		0.055011,
		0.070823,
		0.089259,
		0.110408,
		0.134318,
		0.160989,
		0.190372,
		0.222364,
		0.25681,
		0.293501,
		0.332175,
		0.37252,
		0.414179,
		0.45676,
		0.499836,
		0.542966,
		0.585696,
		0.627576,
		0.668171,
		0.707073,
		0.743911,
		0.778361,
		0.810153,
		0.839077,
		0.864986,
		0.887791,
		0.907465,
		0.924032,
		0.937562,
		0.948161,
		0.955961,
		0.961111,
		0.963764,
		0.964069,
		0.962163,
		0.958161,
		0.952154,
		0.9442,
		0.93433,
		0.92254,
		0.908799,
		0.893048,
		0.875207,
		0.855181,
		0.832862,
		0.808142,
		0.780912,
		0.751076,
		0.718555,
		0.683291,
		0.645261,
		0.604481,
		0.561021,
		0.515011,
		0.466665,
		0.416289,
		0.364313,
		0.31131,
		0.258031,
		0.205443,
		0.154777,
		0.107582,
		0.065798,
		0.031834,
		0.008674,
		4.7e-005,
		0.00023,
		0.000701,
		0.001649,
		0.003294,
		0.005872,
		0.009635,
		0.014835,
		0.021718,
		0.030522,
		0.041464,
		0.054739,
		0.070513,
		0.088916,
		0.110041,
		0.133936,
		0.160605,
		0.189997,
		0.222012,
		0.256491,
		0.293225,
		0.331948,
		0.372344,
		0.414053,
		0.456678,
		0.49979,
		0.542944,
		0.585688,
		0.627576,
		0.668181,
		0.707107,
		0.743999,
		0.778551,
		0.810517,
		0.839707,
		0.865999,
		0.889329,
		0.909691,
		0.927134,
		0.941747,
		0.953657,
		0.963014,
		0.969981,
		0.974726,
		0.977409,
		0.978178,
		0.977154,
		0.974436,
		0.970086,
		0.964136,
		0.956583,
		0.94739,
		0.936491,
		0.923796,
		0.909189,
		0.892541,
		0.873711,
		0.852554,
		0.828924,
		0.802682,
		0.773702,
		0.741872,
		0.707107,
		0.669354,
		0.628601,
		0.584888,
		0.538323,
		0.489096,
		0.437501,
		0.383959,
		0.32905,
		0.273547,
		0.218463,
		0.165103,
		0.11513,
		0.070648,
		0.034297,
		0.009378,
		4.6e-005,
		0.000227,
		0.000691,
		0.001626,
		0.00325,
		0.005798,
		0.009521,
		0.014669,
		0.021491,
		0.030225,
		0.041091,
		0.054285,
		0.069976,
		0.0883,
		0.109353,
		0.133186,
		0.159806,
		0.189167,
		0.22117,
		0.25566,
		0.292425,
		0.331199,
		0.371666,
		0.413461,
		0.456182,
		0.499394,
		0.542647,
		0.585481,
		0.627445,
		0.668108,
		0.707073,
		0.743987,
		0.778551,
		0.810531,
		0.839755,
		0.86612,
		0.889586,
		0.910174,
		0.927957,
		0.943052,
		0.955611,
		0.965809,
		0.973833,
		0.979875,
		0.984115,
		0.986721,
		0.987832,
		0.987562,
		0.985988,
		0.983155,
		0.979067,
		0.973698,
		0.966984,
		0.958834,
		0.94913,
		0.937734,
		0.92449,
		0.909231,
		0.891783,
		0.87197,
		0.849616,
		0.824552,
		0.796619,
		0.765677,
		0.731605,
		0.694316,
		0.653763,
		0.609954,
		0.562964,
		0.512956,
		0.460201,
		0.405109,
		0.34826,
		0.290446,
		0.232725,
		0.176478,
		0.123492,
		0.076051,
		0.037057,
		0.010171,
		4.5e-005,
		0.000222,
		0.000677,
		0.001596,
		0.003191,
		0.005698,
		0.009364,
		0.014439,
		0.021172,
		0.029799,
		0.040544,
		0.053606,
		0.069157,
		0.087336,
		0.108244,
		0.13194,
		0.158434,
		0.187686,
		0.219602,
		0.254032,
		0.290769,
		0.329549,
		0.370055,
		0.411925,
		0.454751,
		0.498098,
		0.541506,
		0.584512,
		0.626654,
		0.667493,
		0.706621,
		0.743679,
		0.778361,
		0.810427,
		0.839707,
		0.866103,
		0.889586,
		0.910193,
		0.928019,
		0.943208,
		0.95594,
		0.966419,
		0.974863,
		0.981492,
		0.986518,
		0.990133,
		0.992508,
		0.993781,
		0.994054,
		0.993394,
		0.991829,
		0.989349,
		0.985907,
		0.981426,
		0.975798,
		0.96889,
		0.960548,
		0.9506,
		0.938862,
		0.92514,
		0.909232,
		0.890936,
		0.870045,
		0.846361,
		0.819691,
		0.789856,
		0.756699,
		0.720092,
		0.679948,
		0.636233,
		0.588982,
		0.538323,
		0.484498,
		0.427896,
		0.369091,
		0.308888,
		0.248386,
		0.189047,
		0.132788,
		0.082095,
		0.040162,
		0.011069,
		4.4e-005,
		0.000216,
		0.000661,
		0.001557,
		0.003117,
		0.005571,
		0.009162,
		0.014139,
		0.020748,
		0.029228,
		0.039801,
		0.052669,
		0.068007,
		0.085958,
		0.10663,
		0.130086,
		0.156346,
		0.185375,
		0.217088,
		0.251342,
		0.287937,
		0.326616,
		0.367068,
		0.408933,
		0.451806,
		0.495252,
		0.53881,
		0.58201,
		0.624385,
		0.665486,
		0.704898,
		0.742246,
		0.777215,
		0.809553,
		0.839077,
		0.865681,
		0.889329,
		0.910055,
		0.927957,
		0.943187,
		0.95594,
		0.966443,
		0.974941,
		0.981687,
		0.986925,
		0.990884,
		0.993768,
		0.995751,
		0.996969,
		0.997522,
		0.997471,
		0.996836,
		0.995602,
		0.993717,
		0.991097,
		0.987631,
		0.983181,
		0.97759,
		0.970681,
		0.962262,
		0.952128,
		0.940063,
		0.925841,
		0.909232,
		0.890001,
		0.867913,
		0.842738,
		0.814257,
		0.782268,
		0.746597,
		0.707107,
		0.663715,
		0.61641,
		0.565272,
		0.510503,
		0.452462,
		0.391707,
		0.32905,
		0.265625,
		0.202976,
		0.14316,
		0.088884,
		0.043674,
		0.012092,
		4.3e-005,
		0.00021,
		0.00064,
		0.001511,
		0.003027,
		0.005413,
		0.00891,
		0.013763,
		0.020214,
		0.0285,
		0.038844,
		0.051448,
		0.06649,
		0.084118,
		0.104444,
		0.127539,
		0.153429,
		0.18209,
		0.213445,
		0.247362,
		0.283651,
		0.322066,
		0.362304,
		0.404013,
		0.446798,
		0.490224,
		0.533837,
		0.577165,
		0.61974,
		0.661108,
		0.700845,
		0.738568,
		0.773948,
		0.80672,
		0.836688,
		0.863728,
		0.887791,
		0.908898,
		0.927134,
		0.942641,
		0.955611,
		0.966268,
		0.974863,
		0.981661,
		0.986925,
		0.990914,
		0.993865,
		0.99599,
		0.997467,
		0.998438,
		0.999005,
		0.99923,
		0.999139,
		0.998721,
		0.99793,
		0.996693,
		0.994906,
		0.992443,
		0.989153,
		0.984868,
		0.9794,
		0.972543,
		0.964076,
		0.953766,
		0.941362,
		0.926607,
		0.909232,
		0.888967,
		0.865538,
		0.838682,
		0.808146,
		0.773702,
		0.735156,
		0.692366,
		0.645261,
		0.593863,
		0.538323,
		0.478955,
		0.416289,
		0.351133,
		0.28465,
		0.218463,
		0.154777,
		0.096544,
		0.047667,
		0.013263,
		4.1e-005,
		0.000202,
		0.000617,
		0.001456,
		0.00292,
		0.005226,
		0.008609,
		0.013309,
		0.019564,
		0.027608,
		0.037661,
		0.049927,
		0.064584,
		0.081783,
		0.101641,
		0.124235,
		0.1496,
		0.177722,
		0.208535,
		0.241918,
		0.277694,
		0.31563,
		0.355436,
		0.396774,
		0.439258,
		0.482466,
		0.52595,
		0.569245,
		0.611886,
		0.653418,
		0.693416,
		0.73149,
		0.767303,
		0.800578,
		0.831105,
		0.858746,
		0.883435,
		0.905174,
		0.924032,
		0.940135,
		0.953657,
		0.96481,
		0.973833,
		0.980982,
		0.986518,
		0.990698,
		0.993768,
		0.995957,
		0.997467,
		0.998474,
		0.999123,
		0.999523,
		0.999749,
		0.999841,
		0.999806,
		0.99962,
		0.99923,
		0.998558,
		0.997501,
		0.995933,
		0.993706,
		0.99065,
		0.986574,
		0.981267,
		0.974494,
		0.966004,
		0.955526,
		0.942774,
		0.927446,
		0.909232,
		0.887815,
		0.862877,
		0.834112,
		0.801227,
		0.763962,
		0.722107,
		0.675515,
		0.624138,
		0.568056,
		0.507526,
		0.443032,
		0.375363,
		0.3057,
		0.23574,
		0.167843,
		0.105229,
		0.05223,
		0.014613,
		3.9e-005,
		0.000193,
		0.00059,
		0.001394,
		0.002796,
		0.005009,
		0.008259,
		0.012777,
		0.018799,
		0.026551,
		0.036251,
		0.0481,
		0.062277,
		0.078936,
		0.098195,
		0.12014,
		0.144812,
		0.172206,
		0.202269,
		0.234893,
		0.269916,
		0.307118,
		0.346229,
		0.386925,
		0.428835,
		0.471554,
		0.514646,
		0.557657,
		0.600132,
		0.641623,
		0.681705,
		0.719989,
		0.756135,
		0.789857,
		0.820937,
		0.849223,
		0.874633,
		0.897154,
		0.916836,
		0.933787,
		0.948161,
		0.960152,
		0.969981,
		0.977887,
		0.984115,
		0.988911,
		0.992508,
		0.995127,
		0.996969,
		0.99821,
		0.999005,
		0.999483,
		0.999749,
		0.999886,
		0.999952,
		0.999981,
		0.999983,
		0.999942,
		0.999818,
		0.999548,
		0.999046,
		0.9982,
		0.996876,
		0.994912,
		0.992122,
		0.988294,
		0.983187,
		0.976536,
		0.968053,
		0.957423,
		0.944313,
		0.928371,
		0.909232,
		0.886525,
		0.859876,
		0.828924,
		0.793332,
		0.752804,
		0.707107,
		0.656102,
		0.599783,
		0.538323,
		0.472143,
		0.40199,
		0.32905,
		0.255081,
		0.182603,
		0.11513,
		0.057479,
		0.016179,
		3.7e-005,
		0.000183,
		0.00056,
		0.001323,
		0.002657,
		0.004764,
		0.00786,
		0.01217,
		0.01792,
		0.025331,
		0.034615,
		0.04597,
		0.059573,
		0.075576,
		0.094104,
		0.115245,
		0.139047,
		0.165515,
		0.194608,
		0.22623,
		0.260237,
		0.296425,
		0.334543,
		0.374285,
		0.415302,
		0.457205,
		0.499578,
		0.541984,
		0.58398,
		0.625132,
		0.665023,
		0.703269,
		0.73953,
		0.773521,
		0.805015,
		0.833854,
		0.859942,
		0.883254,
		0.903823,
		0.921738,
		0.937137,
		0.950194,
		0.961111,
		0.970107,
		0.977409,
		0.983245,
		0.987832,
		0.991373,
		0.994054,
		0.996039,
		0.997471,
		0.99847,
		0.999139,
		0.999561,
		0.999806,
		0.999932,
		0.999983,
		0.999998,
		1,
		0.999999,
		0.999986,
		0.999929,
		0.999771,
		0.999429,
		0.998793,
		0.997721,
		0.996042,
		0.993551,
		0.990011,
		0.985151,
		0.978669,
		0.97023,
		0.959469,
		0.945995,
		0.929393,
		0.909232,
		0.88507,
		0.856465,
		0.82299,
		0.78425,
		0.739908,
		0.689715,
		0.63355,
		0.571482,
		0.503836,
		0.431293,
		0.355015,
		0.276811,
		0.199356,
		0.12648,
		0.063557,
		0.018011,
		3.5e-005,
		0.000172,
		0.000526,
		0.001246,
		0.002503,
		0.004491,
		0.007416,
		0.011491,
		0.016933,
		0.023956,
		0.032763,
		0.043547,
		0.056483,
		0.07172,
		0.089384,
		0.109565,
		0.13232,
		0.157661,
		0.185558,
		0.21593,
		0.248646,
		0.283525,
		0.320332,
		0.358785,
		0.398556,
		0.43928,
		0.480563,
		0.521988,
		0.563133,
		0.603579,
		0.642924,
		0.680795,
		0.716858,
		0.75083,
		0.782486,
		0.81166,
		0.838252,
		0.862222,
		0.883593,
		0.902438,
		0.918878,
		0.933069,
		0.945195,
		0.955459,
		0.964069,
		0.971234,
		0.977154,
		0.982016,
		0.985988,
		0.989217,
		0.991829,
		0.993929,
		0.995602,
		0.996917,
		0.99793,
		0.998688,
		0.99923,
		0.999595,
		0.999818,
		0.999937,
		0.999986,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999909,
		0.999707,
		0.999268,
		0.998451,
		0.997073,
		0.994912,
		0.991705,
		0.987147,
		0.980889,
		0.972543,
		0.961681,
		0.947842,
		0.930531,
		0.909232,
		0.883417,
		0.852555,
		0.816138,
		0.773702,
		0.72486,
		0.669354,
		0.607109,
		0.538323,
		0.463572,
		0.383959,
		0.301315,
		0.218463,
		0.139573,
		0.070648,
		0.020173,
		3.3e-005,
		0.00016,
		0.000491,
		0.001162,
		0.002336,
		0.004194,
		0.00693,
		0.010746,
		0.015848,
		0.022437,
		0.030711,
		0.040853,
		0.053032,
		0.067396,
		0.084068,
		0.103142,
		0.124676,
		0.148694,
		0.175172,
		0.204045,
		0.235198,
		0.268467,
		0.303641,
		0.340459,
		0.37862,
		0.417783,
		0.457578,
		0.497617,
		0.537498,
		0.576825,
		0.615215,
		0.652309,
		0.687787,
		0.721373,
		0.752845,
		0.782037,
		0.808844,
		0.83322,
		0.855177,
		0.874775,
		0.892122,
		0.907357,
		0.92065,
		0.932184,
		0.942154,
		0.950751,
		0.958161,
		0.964555,
		0.970086,
		0.974887,
		0.979067,
		0.982718,
		0.985907,
		0.988688,
		0.991097,
		0.993163,
		0.994906,
		0.996346,
		0.997501,
		0.998392,
		0.999046,
		0.999493,
		0.999771,
		0.99992,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999977,
		0.999882,
		0.999618,
		0.999046,
		0.997978,
		0.996177,
		0.99335,
		0.989154,
		0.983187,
		0.974996,
		0.964076,
		0.949876,
		0.931803,
		0.909232,
		0.881522,
		0.848031,
		0.808146,
		0.761317,
		0.707107,
		0.645261,
		0.575799,
		0.499142,
		0.416289,
		0.32905,
		0.24037,
		0.154777,
		0.078987,
		0.022747,
		3e-005,
		0.000147,
		0.000452,
		0.001072,
		0.002157,
		0.003874,
		0.006407,
		0.009942,
		0.014673,
		0.02079,
		0.028478,
		0.037912,
		0.049254,
		0.062646,
		0.078208,
		0.096035,
		0.116188,
		0.138695,
		0.163544,
		0.19068,
		0.220006,
		0.251376,
		0.284601,
		0.319444,
		0.35563,
		0.392847,
		0.430753,
		0.468986,
		0.507174,
		0.544946,
		0.581942,
		0.617822,
		0.652282,
		0.685059,
		0.715937,
		0.744755,
		0.771408,
		0.795845,
		0.818069,
		0.838132,
		0.856127,
		0.872181,
		0.886448,
		0.899099,
		0.91031,
		0.920262,
		0.929125,
		0.937057,
		0.9442,
		0.950676,
		0.956583,
		0.961999,
		0.966984,
		0.971575,
		0.975798,
		0.979666,
		0.983181,
		0.986344,
		0.989153,
		0.991607,
		0.993706,
		0.995458,
		0.996876,
		0.997978,
		0.998793,
		0.999355,
		0.999707,
		0.999897,
		0.999977,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999844,
		0.999493,
		0.998733,
		0.997312,
		0.994912,
		0.991144,
		0.985548,
		0.977591,
		0.966673,
		0.952128,
		0.933236,
		0.909232,
		0.879329,
		0.842738,
		0.798713,
		0.746597,
		0.685901,
		0.61641,
		0.538323,
		0.452462,
		0.360554,
		0.265625,
		0.172561,
		0.088879,
		0.025825,
		2.7e-005,
		0.000134,
		0.000412,
		0.000977,
		0.001968,
		0.003537,
		0.005853,
		0.009089,
		0.013424,
		0.019033,
		0.02609,
		0.034759,
		0.045192,
		0.057525,
		0.071873,
		0.088328,
		0.106953,
		0.12778,
		0.150806,
		0.175988,
		0.203242,
		0.232442,
		0.26342,
		0.295966,
		0.32983,
		0.36473,
		0.400354,
		0.436373,
		0.472443,
		0.508222,
		0.543376,
		0.577591,
		0.610581,
		0.642099,
		0.67194,
		0.699951,
		0.726028,
		0.750119,
		0.772221,
		0.792377,
		0.810671,
		0.827216,
		0.842154,
		0.855639,
		0.867837,
		0.878913,
		0.889025,
		0.898322,
		0.906936,
		0.914977,
		0.92254,
		0.929694,
		0.936491,
		0.942964,
		0.94913,
		0.954993,
		0.960548,
		0.965783,
		0.970681,
		0.975226,
		0.9794,
		0.983187,
		0.986574,
		0.989554,
		0.992122,
		0.994282,
		0.996042,
		0.997422,
		0.998451,
		0.999166,
		0.999618,
		0.999864,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999959,
		0.999788,
		0.999313,
		0.99828,
		0.996348,
		0.993081,
		0.987949,
		0.980326,
		0.969489,
		0.95463,
		0.93486,
		0.909232,
		0.876762,
		0.836467,
		0.787425,
		0.728851,
		0.660215,
		0.581406,
		0.492972,
		0.39646,
		0.294891,
		0.19336,
		0.100129,
		0.028884,
		2.4e-005,
		0.000121,
		0.000371,
		0.000879,
		0.001771,
		0.003186,
		0.005276,
		0.008198,
		0.012115,
		0.017189,
		0.023578,
		0.031435,
		0.0409,
		0.052099,
		0.065143,
		0.08012,
		0.097092,
		0.116093,
		0.137128,
		0.160162,
		0.185128,
		0.211916,
		0.240381,
		0.270336,
		0.30156,
		0.333802,
		0.366781,
		0.400201,
		0.43375,
		0.467118,
		0.5,
		0.532108,
		0.563181,
		0.592988,
		0.621342,
		0.648096,
		0.673152,
		0.696458,
		0.71801,
		0.737842,
		0.756027,
		0.77267,
		0.787895,
		0.801847,
		0.814675,
		0.826533,
		0.837568,
		0.847917,
		0.857704,
		0.867033,
		0.875992,
		0.884648,
		0.893048,
		0.901223,
		0.909189,
		0.916947,
		0.92449,
		0.931802,
		0.938862,
		0.945646,
		0.952128,
		0.95828,
		0.964076,
		0.969489,
		0.974494,
		0.979066,
		0.983187,
		0.986838,
		0.990011,
		0.9927,
		0.994912,
		0.996662,
		0.997978,
		0.998903,
		0.999493,
		0.999819,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999944,
		0.999707,
		0.999046,
		0.997607,
		0.994912,
		0.990352,
		0.983187,
		0.972543,
		0.957423,
		0.936718,
		0.909232,
		0.873716,
		0.828924,
		0.773702,
		0.707107,
		0.628601,
		0.538297,
		0.437117,
		0.327031,
		0.212988,
		0.106843,
		0.028801,
		2.2e-005,
		0.000107,
		0.000328,
		0.000779,
		0.00157,
		0.002826,
		0.004682,
		0.007279,
		0.010764,
		0.015282,
		0.020976,
		0.027985,
		0.036436,
		0.046446,
		0.058116,
		0.07153,
		0.086747,
		0.103804,
		0.122709,
		0.143438,
		0.165934,
		0.190107,
		0.21583,
		0.242943,
		0.271253,
		0.300538,
		0.330552,
		0.361029,
		0.391695,
		0.422271,
		0.452485,
		0.482077,
		0.510811,
		0.538479,
		0.564909,
		0.589968,
		0.613563,
		0.635646,
		0.656209,
		0.675281,
		0.692926,
		0.709237,
		0.724327,
		0.738324,
		0.751367,
		0.763595,
		0.775143,
		0.786138,
		0.796695,
		0.806913,
		0.816874,
		0.826642,
		0.836264,
		0.845772,
		0.855181,
		0.864496,
		0.873711,
		0.882813,
		0.891783,
		0.900598,
		0.909232,
		0.917657,
		0.925841,
		0.933754,
		0.941362,
		0.948631,
		0.955526,
		0.962012,
		0.968053,
		0.973615,
		0.978669,
		0.983187,
		0.987147,
		0.990536,
		0.99335,
		0.995601,
		0.997312,
		0.998528,
		0.999313,
		0.999752,
		0.999944,
		0.999996,
		1,
		0.999995,
		0.99992,
		0.999581,
		0.998635,
		0.996571,
		0.9927,
		0.986146,
		0.975845,
		0.960553,
		0.938862,
		0.909232,
		0.870045,
		0.819691,
		0.756663,
		0.679351,
		0.585368,
		0.472358,
		0.342523,
		0.208583,
		0.093895,
		0.021807,
		1.9e-005,
		9.3e-005,
		0.000285,
		0.000678,
		0.001366,
		0.002461,
		0.00408,
		0.006347,
		0.009391,
		0.013341,
		0.018324,
		0.024461,
		0.031869,
		0.040651,
		0.0509,
		0.062691,
		0.076083,
		0.091109,
		0.107782,
		0.126086,
		0.145975,
		0.167375,
		0.190179,
		0.214251,
		0.239425,
		0.265509,
		0.292291,
		0.31954,
		0.347016,
		0.374475,
		0.401677,
		0.428393,
		0.454416,
		0.47956,
		0.50367,
		0.526629,
		0.548352,
		0.568794,
		0.587947,
		0.605834,
		0.622512,
		0.63806,
		0.652581,
		0.666188,
		0.679005,
		0.691157,
		0.702768,
		0.713954,
		0.724819,
		0.735456,
		0.745942,
		0.756338,
		0.766692,
		0.777035,
		0.787387,
		0.797756,
		0.808142,
		0.818535,
		0.828924,
		0.839291,
		0.849616,
		0.859876,
		0.870045,
		0.880097,
		0.890001,
		0.899725,
		0.909232,
		0.918486,
		0.927446,
		0.93607,
		0.944313,
		0.952128,
		0.959469,
		0.966289,
		0.972543,
		0.978187,
		0.983187,
		0.987511,
		0.991144,
		0.994084,
		0.996348,
		0.997978,
		0.999046,
		0.999651,
		0.99992,
		0.999994,
		1,
		0.999993,
		0.999882,
		0.999381,
		0.997978,
		0.994912,
		0.989154,
		0.9794,
		0.964076,
		0.941362,
		0.909189,
		0.864779,
		0.803188,
		0.716734,
		0.598814,
		0.45206,
		0.294361,
		0.154555,
		0.057659,
		0.010604,
		1.6e-005,
		7.9e-005,
		0.000243,
		0.000577,
		0.001164,
		0.002098,
		0.003479,
		0.005416,
		0.008018,
		0.011397,
		0.015662,
		0.020921,
		0.027273,
		0.034811,
		0.043615,
		0.053754,
		0.06528,
		0.078227,
		0.092608,
		0.108413,
		0.125608,
		0.144131,
		0.163895,
		0.184787,
		0.206667,
		0.229374,
		0.252728,
		0.276532,
		0.300582,
		0.324667,
		0.348584,
		0.372134,
		0.395137,
		0.417432,
		0.438887,
		0.459397,
		0.478887,
		0.497317,
		0.514679,
		0.530993,
		0.546305,
		0.560686,
		0.574223,
		0.587017,
		0.599175,
		0.61081,
		0.622029,
		0.632938,
		0.643631,
		0.654192,
		0.664693,
		0.675192,
		0.685735,
		0.696355,
		0.707073,
		0.717902,
		0.728847,
		0.739907,
		0.751076,
		0.762345,
		0.773702,
		0.785132,
		0.796619,
		0.808146,
		0.819691,
		0.83123,
		0.842738,
		0.854185,
		0.865538,
		0.876762,
		0.887815,
		0.898654,
		0.909232,
		0.919497,
		0.929393,
		0.938862,
		0.947842,
		0.956268,
		0.964076,
		0.971204,
		0.977591,
		0.983187,
		0.987949,
		0.991857,
		0.994912,
		0.997147,
		0.998635,
		0.999493,
		0.999882,
		0.999991,
		1,
		0.999989,
		0.999819,
		0.999046,
		0.996876,
		0.992122,
		0.983139,
		0.967203,
		0.938519,
		0.886448,
		0.797981,
		0.666205,
		0.5,
		0.325662,
		0.175889,
		0.073201,
		0.020197,
		0.002407,
		1.3e-005,
		6.5e-005,
		0.000201,
		0.000479,
		0.000966,
		0.001742,
		0.00289,
		0.004501,
		0.006667,
		0.009482,
		0.013038,
		0.017426,
		0.02273,
		0.02903,
		0.036394,
		0.044882,
		0.054541,
		0.0654,
		0.077474,
		0.090758,
		0.105226,
		0.120829,
		0.137498,
		0.155141,
		0.173644,
		0.192875,
		0.212684,
		0.232908,
		0.253379,
		0.27392,
		0.294361,
		0.314536,
		0.334294,
		0.353499,
		0.372038,
		0.389823,
		0.40679,
		0.422905,
		0.438158,
		0.452566,
		0.46617,
		0.479026,
		0.491211,
		0.502809,
		0.513912,
		0.524618,
		0.535021,
		0.54521,
		0.555271,
		0.565277,
		0.575293,
		0.585374,
		0.595562,
		0.605889,
		0.616381,
		0.627051,
		0.637912,
		0.648966,
		0.660215,
		0.671658,
		0.683291,
		0.695109,
		0.707107,
		0.719275,
		0.731605,
		0.744084,
		0.756699,
		0.769433,
		0.782268,
		0.795181,
		0.808146,
		0.821134,
		0.834112,
		0.84704,
		0.859876,
		0.872571,
		0.88507,
		0.897313,
		0.909232,
		0.920755,
		0.931803,
		0.94229,
		0.952128,
		0.961225,
		0.969489,
		0.976835,
		0.983187,
		0.988485,
		0.9927,
		0.995843,
		0.997978,
		0.999236,
		0.999819,
		0.999986,
		1,
		0.999982,
		0.999659,
		0.997575,
		0.988808,
		0.96241,
		0.902538,
		0.795954,
		0.642924,
		0.462908,
		0.288229,
		0.148905,
		0.05954,
		0.016081,
		0.002141,
		5.1e-005,
		1.1e-005,
		5.2e-005,
		0.000161,
		0.000384,
		0.000776,
		0.001399,
		0.002323,
		0.00362,
		0.005365,
		0.007633,
		0.010502,
		0.014044,
		0.018328,
		0.023421,
		0.029379,
		0.036252,
		0.04408,
		0.052889,
		0.062693,
		0.073489,
		0.08526,
		0.097968,
		0.11156,
		0.125963,
		0.141088,
		0.156828,
		0.173066,
		0.18967,
		0.206504,
		0.223428,
		0.240301,
		0.256992,
		0.273376,
		0.289343,
		0.304801,
		0.319677,
		0.333919,
		0.347498,
		0.360406,
		0.372656,
		0.384281,
		0.395328,
		0.405858,
		0.415942,
		0.425657,
		0.435083,
		0.444298,
		0.453381,
		0.462401,
		0.471423,
		0.480503,
		0.48969,
		0.499021,
		0.508529,
		0.518237,
		0.528163,
		0.53832,
		0.548717,
		0.559361,
		0.570256,
		0.581406,
		0.592813,
		0.604481,
		0.61641,
		0.628601,
		0.641052,
		0.653763,
		0.66673,
		0.679948,
		0.69341,
		0.707107,
		0.721027,
		0.735156,
		0.749475,
		0.763962,
		0.778592,
		0.793332,
		0.808146,
		0.82299,
		0.837812,
		0.852555,
		0.867151,
		0.881522,
		0.895582,
		0.909232,
		0.922365,
		0.93486,
		0.946591,
		0.957423,
		0.967218,
		0.975845,
		0.983187,
		0.989154,
		0.993706,
		0.996876,
		0.998793,
		0.999659,
		0.9991,
		0.993865,
		0.974912,
		0.927549,
		0.837497,
		0.700845,
		0.530799,
		0.354699,
		0.202269,
		0.09321,
		0.031269,
		0.006041,
		0.000326,
		4e-005,
		0.000124,
		0.000295,
		0.000597,
		0.001077,
		0.00179,
		0.00279,
		0.004136,
		0.005889,
		0.008105,
		0.010844,
		0.01416,
		0.018105,
		0.022723,
		0.028054,
		0.034131,
		0.040976,
		0.0486,
		0.057003,
		0.066174,
		0.076086,
		0.086698,
		0.097956,
		0.109792,
		0.122125,
		0.134865,
		0.147912,
		0.161159,
		0.1745,
		0.187825,
		0.201032,
		0.214025,
		0.226717,
		0.239036,
		0.250926,
		0.262345,
		0.27327,
		0.283695,
		0.29363,
		0.3031,
		0.312142,
		0.320804,
		0.329143,
		0.337219,
		0.345096,
		0.352837,
		0.360506,
		0.368158,
		0.375848,
		0.383622,
		0.39152,
		0.399576,
		0.407819,
		0.416269,
		0.424946,
		0.433863,
		0.443032,
		0.452462,
		0.462163,
		0.472143,
		0.48241,
		0.492972,
		0.503836,
		0.515011,
		0.526505,
		0.538323,
		0.550474,
		0.562964,
		0.575799,
		0.588982,
		0.602518,
		0.61641,
		0.630658,
		0.645261,
		0.660215,
		0.675515,
		0.69115,
		0.707107,
		0.723368,
		0.739908,
		0.756699,
		0.773702,
		0.790869,
		0.808146,
		0.825463,
		0.842738,
		0.859876,
		0.876762,
		0.893264,
		0.909232,
		0.924495,
		0.938862,
		0.952128,
		0.964076,
		0.974494,
		0.983139,
		0.989142,
		0.988808,
		0.97297,
		0.927549,
		0.839721,
		0.707107,
		0.542935,
		0.372177,
		0.22133,
		0.108744,
		0.040099,
		0.00891,
		0.000596,
		2.9e-005,
		9e-005,
		0.000215,
		0.000434,
		0.000783,
		0.001301,
		0.00203,
		0.003011,
		0.004288,
		0.005905,
		0.007904,
		0.010326,
		0.013209,
		0.016587,
		0.02049,
		0.024941,
		0.029959,
		0.035553,
		0.041725,
		0.048465,
		0.055758,
		0.063573,
		0.071873,
		0.080608,
		0.089722,
		0.099147,
		0.108813,
		0.118641,
		0.128554,
		0.138472,
		0.14832,
		0.158028,
		0.167531,
		0.176778,
		0.185725,
		0.194342,
		0.202613,
		0.210532,
		0.218106,
		0.225354,
		0.232304,
		0.238991,
		0.245457,
		0.251747,
		0.257911,
		0.263995,
		0.270048,
		0.276112,
		0.282229,
		0.288436,
		0.294765,
		0.301242,
		0.307892,
		0.314732,
		0.321779,
		0.329048,
		0.336549,
		0.344293,
		0.352291,
		0.360554,
		0.369091,
		0.377913,
		0.387032,
		0.39646,
		0.406208,
		0.416289,
		0.426716,
		0.437501,
		0.448658,
		0.460201,
		0.472143,
		0.484498,
		0.497281,
		0.510503,
		0.52418,
		0.538323,
		0.552945,
		0.568056,
		0.583667,
		0.599783,
		0.61641,
		0.63355,
		0.651201,
		0.669354,
		0.687996,
		0.707107,
		0.726654,
		0.746597,
		0.766878,
		0.787425,
		0.808146,
		0.828924,
		0.849616,
		0.870045,
		0.890001,
		0.909189,
		0.926633,
		0.938519,
		0.935426,
		0.902538,
		0.825636,
		0.700845,
		0.540983,
		0.372177,
		0.222444,
		0.110603,
		0.041877,
		0.00978,
		0.000712,
		2e-005,
		6e-005,
		0.000143,
		0.00029,
		0.000524,
		0.000871,
		0.001359,
		0.002017,
		0.002874,
		0.00396,
		0.005302,
		0.00693,
		0.008869,
		0.011143,
		0.013772,
		0.016772,
		0.020157,
		0.023933,
		0.028102,
		0.03266,
		0.037596,
		0.04289,
		0.048517,
		0.054447,
		0.06064,
		0.067052,
		0.073636,
		0.080339,
		0.08711,
		0.093895,
		0.100644,
		0.107308,
		0.113845,
		0.120219,
		0.126401,
		0.132372,
		0.138117,
		0.143636,
		0.148931,
		0.154017,
		0.15891,
		0.163637,
		0.168226,
		0.172707,
		0.177115,
		0.181483,
		0.185844,
		0.190228,
		0.194665,
		0.199181,
		0.203799,
		0.208539,
		0.213418,
		0.218452,
		0.223654,
		0.229034,
		0.234602,
		0.24037,
		0.246346,
		0.25254,
		0.258963,
		0.265625,
		0.272538,
		0.279714,
		0.287165,
		0.294906,
		0.302949,
		0.31131,
		0.320005,
		0.32905,
		0.338462,
		0.34826,
		0.358463,
		0.369091,
		0.380165,
		0.391707,
		0.403741,
		0.416289,
		0.429378,
		0.443032,
		0.457278,
		0.472143,
		0.487653,
		0.503836,
		0.520718,
		0.538323,
		0.556677,
		0.575799,
		0.595706,
		0.61641,
		0.637915,
		0.660215,
		0.683291,
		0.707107,
		0.731605,
		0.756663,
		0.781582,
		0.803188,
		0.81321,
		0.797981,
		0.743241,
		0.642924,
		0.505937,
		0.354699,
		0.215674,
		0.108744,
		0.041577,
		0.00978,
		0.000719,
		1.1e-005,
		3.5e-005,
		8.4e-005,
		0.00017,
		0.000308,
		0.000512,
		0.000799,
		0.001186,
		0.001691,
		0.002331,
		0.003122,
		0.004083,
		0.005227,
		0.00657,
		0.008124,
		0.009899,
		0.011902,
		0.014139,
		0.016611,
		0.019315,
		0.022245,
		0.025391,
		0.028739,
		0.032269,
		0.03596,
		0.039786,
		0.043719,
		0.047728,
		0.051783,
		0.055852,
		0.059906,
		0.063915,
		0.067856,
		0.071706,
		0.075448,
		0.07907,
		0.082565,
		0.085931,
		0.08917,
		0.09229,
		0.095303,
		0.098223,
		0.101067,
		0.103854,
		0.106605,
		0.109339,
		0.112078,
		0.114839,
		0.117642,
		0.120501,
		0.123433,
		0.12645,
		0.129562,
		0.132781,
		0.136115,
		0.139572,
		0.143159,
		0.146884,
		0.150754,
		0.154777,
		0.15896,
		0.163312,
		0.167843,
		0.172561,
		0.177478,
		0.182603,
		0.18795,
		0.193529,
		0.199356,
		0.205443,
		0.211807,
		0.218463,
		0.225429,
		0.232725,
		0.24037,
		0.248386,
		0.256796,
		0.265625,
		0.2749,
		0.28465,
		0.294906,
		0.3057,
		0.317068,
		0.32905,
		0.341684,
		0.355015,
		0.369091,
		0.383959,
		0.399673,
		0.416289,
		0.433865,
		0.452462,
		0.472143,
		0.492972,
		0.515011,
		0.538297,
		0.56247,
		0.585368,
		0.600964,
		0.598814,
		0.567267,
		0.5,
		0.401745,
		0.288229,
		0.179781,
		0.09321,
		0.036721,
		0.00891,
		0.000676,
		1.6e-005,
		3.9e-005,
		7.9e-005,
		0.000143,
		0.000237,
		0.000371,
		0.000551,
		0.000785,
		0.001083,
		0.001451,
		0.001898,
		0.002431,
		0.003057,
		0.003782,
		0.00461,
		0.005545,
		0.006591,
		0.007747,
		0.009012,
		0.010384,
		0.011859,
		0.01343,
		0.015087,
		0.016822,
		0.018622,
		0.020474,
		0.022365,
		0.024279,
		0.026203,
		0.028122,
		0.030023,
		0.031895,
		0.033727,
		0.035511,
		0.037241,
		0.038915,
		0.040531,
		0.04209,
		0.043596,
		0.045055,
		0.046472,
		0.047857,
		0.049219,
		0.050566,
		0.05191,
		0.053259,
		0.054623,
		0.05601,
		0.057429,
		0.058887,
		0.06039,
		0.061944,
		0.063554,
		0.065226,
		0.066962,
		0.068768,
		0.070648,
		0.072605,
		0.074644,
		0.07677,
		0.078987,
		0.081301,
		0.083718,
		0.086243,
		0.088884,
		0.091646,
		0.094539,
		0.09757,
		0.100747,
		0.104082,
		0.107582,
		0.111261,
		0.11513,
		0.119202,
		0.123492,
		0.128015,
		0.132788,
		0.137829,
		0.14316,
		0.148801,
		0.154777,
		0.161115,
		0.167843,
		0.174994,
		0.182603,
		0.19071,
		0.199356,
		0.208589,
		0.218463,
		0.229035,
		0.24037,
		0.25254,
		0.265625,
		0.279714,
		0.294891,
		0.311037,
		0.327031,
		0.339533,
		0.342523,
		0.328938,
		0.294361,
		0.240551,
		0.175889,
		0.112084,
		0.05954,
		0.024116,
		0.006041,
		0.000475,
		1e-005,
		2.1e-005,
		3.7e-005,
		6.2e-005,
		9.7e-005,
		0.000144,
		0.000205,
		0.000283,
		0.000379,
		0.000496,
		0.000635,
		0.000799,
		0.000989,
		0.001206,
		0.001451,
		0.001726,
		0.002029,
		0.002362,
		0.002723,
		0.003111,
		0.003525,
		0.003962,
		0.00442,
		0.004895,
		0.005385,
		0.005885,
		0.006393,
		0.006903,
		0.007413,
		0.007919,
		0.008417,
		0.008906,
		0.009384,
		0.009847,
		0.010297,
		0.010731,
		0.011152,
		0.011559,
		0.011955,
		0.01234,
		0.012718,
		0.01309,
		0.013459,
		0.013828,
		0.014199,
		0.014576,
		0.014959,
		0.015352,
		0.015757,
		0.016175,
		0.016608,
		0.017057,
		0.017525,
		0.018011,
		0.018518,
		0.019046,
		0.019597,
		0.020173,
		0.020774,
		0.021402,
		0.022059,
		0.022747,
		0.023468,
		0.024223,
		0.025016,
		0.025848,
		0.026722,
		0.027641,
		0.028608,
		0.029626,
		0.0307,
		0.031834,
		0.033031,
		0.034297,
		0.035637,
		0.037057,
		0.038563,
		0.040162,
		0.041863,
		0.043674,
		0.045605,
		0.047667,
		0.049871,
		0.05223,
		0.054761,
		0.057479,
		0.060404,
		0.063557,
		0.066963,
		0.070648,
		0.074644,
		0.078987,
		0.083718,
		0.088879,
		0.094456,
		0.100129,
		0.104887,
		0.106843,
		0.103703,
		0.093895,
		0.077731,
		0.057659,
		0.037339,
		0.020197,
		0.008349,
		0.002141,
		0.000173,
		0.000173,
		0.002141,
		0.008349,
		0.020197,
		0.037339,
		0.057659,
		0.077731,
		0.093895,
		0.103703,
		0.106843,
		0.104887,
		0.100129,
		0.094456,
		0.088879,
		0.083718,
		0.078987,
		0.074644,
		0.070648,
		0.066963,
		0.063557,
		0.060404,
		0.057479,
		0.054761,
		0.05223,
		0.049871,
		0.047667,
		0.045605,
		0.043674,
		0.041863,
		0.040162,
		0.038563,
		0.037057,
		0.035637,
		0.034297,
		0.033031,
		0.031834,
		0.0307,
		0.029626,
		0.028608,
		0.027641,
		0.026722,
		0.025848,
		0.025016,
		0.024223,
		0.023468,
		0.022747,
		0.022059,
		0.021402,
		0.020774,
		0.020173,
		0.019597,
		0.019046,
		0.018518,
		0.018011,
		0.017525,
		0.017057,
		0.016608,
		0.016175,
		0.015757,
		0.015352,
		0.014959,
		0.014576,
		0.014199,
		0.013828,
		0.013459,
		0.01309,
		0.012718,
		0.01234,
		0.011955,
		0.011559,
		0.011152,
		0.010731,
		0.010297,
		0.009847,
		0.009384,
		0.008906,
		0.008417,
		0.007919,
		0.007413,
		0.006903,
		0.006393,
		0.005885,
		0.005385,
		0.004895,
		0.00442,
		0.003962,
		0.003525,
		0.003111,
		0.002723,
		0.002362,
		0.002029,
		0.001726,
		0.001451,
		0.001206,
		0.000989,
		0.000799,
		0.000635,
		0.000496,
		0.000379,
		0.000283,
		0.000205,
		0.000144,
		9.7e-005,
		6.2e-005,
		3.7e-005,
		2.1e-005,
		1e-005,
		0.000475,
		0.006041,
		0.024116,
		0.05954,
		0.112084,
		0.175889,
		0.240551,
		0.294361,
		0.328938,
		0.342523,
		0.339533,
		0.327031,
		0.311037,
		0.294891,
		0.279714,
		0.265625,
		0.25254,
		0.24037,
		0.229035,
		0.218463,
		0.208589,
		0.199356,
		0.19071,
		0.182603,
		0.174994,
		0.167843,
		0.161115,
		0.154777,
		0.148801,
		0.14316,
		0.137829,
		0.132788,
		0.128015,
		0.123492,
		0.119202,
		0.11513,
		0.111261,
		0.107582,
		0.104082,
		0.100747,
		0.09757,
		0.094539,
		0.091646,
		0.088884,
		0.086243,
		0.083718,
		0.081301,
		0.078987,
		0.07677,
		0.074644,
		0.072605,
		0.070648,
		0.068768,
		0.066962,
		0.065226,
		0.063554,
		0.061944,
		0.06039,
		0.058887,
		0.057429,
		0.05601,
		0.054623,
		0.053259,
		0.05191,
		0.050566,
		0.049219,
		0.047857,
		0.046472,
		0.045055,
		0.043596,
		0.04209,
		0.040531,
		0.038915,
		0.037241,
		0.035511,
		0.033727,
		0.031895,
		0.030023,
		0.028122,
		0.026203,
		0.024279,
		0.022365,
		0.020474,
		0.018622,
		0.016822,
		0.015087,
		0.01343,
		0.011859,
		0.010384,
		0.009012,
		0.007747,
		0.006591,
		0.005545,
		0.00461,
		0.003782,
		0.003057,
		0.002431,
		0.001898,
		0.001451,
		0.001083,
		0.000785,
		0.000551,
		0.000371,
		0.000237,
		0.000143,
		7.9e-005,
		3.9e-005,
		1.6e-005,
		0.000676,
		0.00891,
		0.036721,
		0.09321,
		0.179781,
		0.288229,
		0.401745,
		0.5,
		0.567267,
		0.598814,
		0.600964,
		0.585368,
		0.56247,
		0.538297,
		0.515011,
		0.492972,
		0.472143,
		0.452462,
		0.433865,
		0.416289,
		0.399673,
		0.383959,
		0.369091,
		0.355015,
		0.341684,
		0.32905,
		0.317068,
		0.3057,
		0.294906,
		0.28465,
		0.2749,
		0.265625,
		0.256796,
		0.248386,
		0.24037,
		0.232725,
		0.225429,
		0.218463,
		0.211807,
		0.205443,
		0.199356,
		0.193529,
		0.18795,
		0.182603,
		0.177478,
		0.172561,
		0.167843,
		0.163312,
		0.15896,
		0.154777,
		0.150754,
		0.146884,
		0.143159,
		0.139572,
		0.136115,
		0.132781,
		0.129562,
		0.12645,
		0.123433,
		0.120501,
		0.117642,
		0.114839,
		0.112078,
		0.109339,
		0.106605,
		0.103854,
		0.101067,
		0.098223,
		0.095303,
		0.09229,
		0.08917,
		0.085931,
		0.082565,
		0.07907,
		0.075448,
		0.071706,
		0.067856,
		0.063915,
		0.059906,
		0.055852,
		0.051783,
		0.047728,
		0.043719,
		0.039786,
		0.03596,
		0.032269,
		0.028739,
		0.025391,
		0.022245,
		0.019315,
		0.016611,
		0.014139,
		0.011902,
		0.009899,
		0.008124,
		0.00657,
		0.005227,
		0.004083,
		0.003122,
		0.002331,
		0.001691,
		0.001186,
		0.000799,
		0.000512,
		0.000308,
		0.00017,
		8.4e-005,
		3.5e-005,
		1.1e-005,
		0.000719,
		0.00978,
		0.041577,
		0.108744,
		0.215674,
		0.354699,
		0.505937,
		0.642924,
		0.743241,
		0.797981,
		0.81321,
		0.803188,
		0.781582,
		0.756663,
		0.731605,
		0.707107,
		0.683291,
		0.660215,
		0.637915,
		0.61641,
		0.595706,
		0.575799,
		0.556677,
		0.538323,
		0.520718,
		0.503836,
		0.487653,
		0.472143,
		0.457278,
		0.443032,
		0.429378,
		0.416289,
		0.403741,
		0.391707,
		0.380165,
		0.369091,
		0.358463,
		0.34826,
		0.338462,
		0.32905,
		0.320005,
		0.31131,
		0.302949,
		0.294906,
		0.287165,
		0.279714,
		0.272538,
		0.265625,
		0.258963,
		0.25254,
		0.246346,
		0.24037,
		0.234602,
		0.229034,
		0.223654,
		0.218452,
		0.213418,
		0.208539,
		0.203799,
		0.199181,
		0.194665,
		0.190228,
		0.185844,
		0.181483,
		0.177115,
		0.172707,
		0.168226,
		0.163637,
		0.15891,
		0.154017,
		0.148931,
		0.143636,
		0.138117,
		0.132372,
		0.126401,
		0.120219,
		0.113845,
		0.107308,
		0.100644,
		0.093895,
		0.08711,
		0.080339,
		0.073636,
		0.067052,
		0.06064,
		0.054447,
		0.048517,
		0.04289,
		0.037596,
		0.03266,
		0.028102,
		0.023933,
		0.020157,
		0.016772,
		0.013772,
		0.011143,
		0.008869,
		0.00693,
		0.005302,
		0.00396,
		0.002874,
		0.002017,
		0.001359,
		0.000871,
		0.000524,
		0.00029,
		0.000143,
		6e-005,
		2e-005,
		0.000712,
		0.00978,
		0.041877,
		0.110603,
		0.222444,
		0.372177,
		0.540983,
		0.700845,
		0.825636,
		0.902538,
		0.935426,
		0.938519,
		0.926633,
		0.909189,
		0.890001,
		0.870045,
		0.849616,
		0.828924,
		0.808146,
		0.787425,
		0.766878,
		0.746597,
		0.726654,
		0.707107,
		0.687996,
		0.669354,
		0.651201,
		0.63355,
		0.61641,
		0.599783,
		0.583667,
		0.568056,
		0.552945,
		0.538323,
		0.52418,
		0.510503,
		0.497281,
		0.484498,
		0.472143,
		0.460201,
		0.448658,
		0.437501,
		0.426716,
		0.416289,
		0.406208,
		0.39646,
		0.387032,
		0.377913,
		0.369091,
		0.360554,
		0.352291,
		0.344293,
		0.336549,
		0.329048,
		0.321779,
		0.314732,
		0.307892,
		0.301242,
		0.294765,
		0.288436,
		0.282229,
		0.276112,
		0.270048,
		0.263995,
		0.257911,
		0.251747,
		0.245457,
		0.238991,
		0.232304,
		0.225354,
		0.218106,
		0.210532,
		0.202613,
		0.194342,
		0.185725,
		0.176778,
		0.167531,
		0.158028,
		0.14832,
		0.138472,
		0.128554,
		0.118641,
		0.108813,
		0.099147,
		0.089722,
		0.080608,
		0.071873,
		0.063573,
		0.055758,
		0.048465,
		0.041725,
		0.035553,
		0.029959,
		0.024941,
		0.02049,
		0.016587,
		0.013209,
		0.010326,
		0.007904,
		0.005905,
		0.004288,
		0.003011,
		0.00203,
		0.001301,
		0.000783,
		0.000434,
		0.000215,
		9e-005,
		2.9e-005,
		0.000596,
		0.00891,
		0.040099,
		0.108744,
		0.22133,
		0.372177,
		0.542935,
		0.707107,
		0.839721,
		0.927549,
		0.97297,
		0.988808,
		0.989142,
		0.983139,
		0.974494,
		0.964076,
		0.952128,
		0.938862,
		0.924495,
		0.909232,
		0.893264,
		0.876762,
		0.859876,
		0.842738,
		0.825463,
		0.808146,
		0.790869,
		0.773702,
		0.756699,
		0.739908,
		0.723368,
		0.707107,
		0.69115,
		0.675515,
		0.660215,
		0.645261,
		0.630658,
		0.61641,
		0.602518,
		0.588982,
		0.575799,
		0.562964,
		0.550474,
		0.538323,
		0.526505,
		0.515011,
		0.503836,
		0.492972,
		0.48241,
		0.472143,
		0.462163,
		0.452462,
		0.443032,
		0.433863,
		0.424946,
		0.416269,
		0.407819,
		0.399576,
		0.39152,
		0.383622,
		0.375848,
		0.368158,
		0.360506,
		0.352837,
		0.345096,
		0.337219,
		0.329143,
		0.320804,
		0.312142,
		0.3031,
		0.29363,
		0.283695,
		0.27327,
		0.262345,
		0.250926,
		0.239036,
		0.226717,
		0.214025,
		0.201032,
		0.187825,
		0.1745,
		0.161159,
		0.147912,
		0.134865,
		0.122125,
		0.109792,
		0.097956,
		0.086698,
		0.076086,
		0.066174,
		0.057003,
		0.0486,
		0.040976,
		0.034131,
		0.028054,
		0.022723,
		0.018105,
		0.01416,
		0.010844,
		0.008105,
		0.005889,
		0.004136,
		0.00279,
		0.00179,
		0.001077,
		0.000597,
		0.000295,
		0.000124,
		4e-005,
		0.000326,
		0.006041,
		0.031269,
		0.09321,
		0.202269,
		0.354699,
		0.530799,
		0.700845,
		0.837497,
		0.927549,
		0.974912,
		0.993865,
		0.9991,
		0.999659,
		0.998793,
		0.996876,
		0.993706,
		0.989154,
		0.983187,
		0.975845,
		0.967218,
		0.957423,
		0.946591,
		0.93486,
		0.922365,
		0.909232,
		0.895582,
		0.881522,
		0.867151,
		0.852555,
		0.837812,
		0.82299,
		0.808146,
		0.793332,
		0.778592,
		0.763962,
		0.749475,
		0.735156,
		0.721027,
		0.707107,
		0.69341,
		0.679948,
		0.66673,
		0.653763,
		0.641052,
		0.628601,
		0.61641,
		0.604481,
		0.592813,
		0.581406,
		0.570256,
		0.559361,
		0.548717,
		0.53832,
		0.528163,
		0.518237,
		0.508529,
		0.499021,
		0.48969,
		0.480503,
		0.471423,
		0.462401,
		0.453381,
		0.444298,
		0.435083,
		0.425657,
		0.415942,
		0.405858,
		0.395328,
		0.384281,
		0.372656,
		0.360406,
		0.347498,
		0.333919,
		0.319677,
		0.304801,
		0.289343,
		0.273376,
		0.256992,
		0.240301,
		0.223428,
		0.206504,
		0.18967,
		0.173066,
		0.156828,
		0.141088,
		0.125963,
		0.11156,
		0.097968,
		0.08526,
		0.073489,
		0.062693,
		0.052889,
		0.04408,
		0.036252,
		0.029379,
		0.023421,
		0.018328,
		0.014044,
		0.010502,
		0.007633,
		0.005365,
		0.00362,
		0.002323,
		0.001399,
		0.000776,
		0.000384,
		0.000161,
		5.2e-005,
		1.1e-005,
		5.1e-005,
		0.002141,
		0.016081,
		0.05954,
		0.148905,
		0.288229,
		0.462908,
		0.642924,
		0.795954,
		0.902538,
		0.96241,
		0.988808,
		0.997575,
		0.999659,
		0.999982,
		1,
		0.999986,
		0.999819,
		0.999236,
		0.997978,
		0.995843,
		0.9927,
		0.988485,
		0.983187,
		0.976835,
		0.969489,
		0.961225,
		0.952128,
		0.94229,
		0.931803,
		0.920755,
		0.909232,
		0.897313,
		0.88507,
		0.872571,
		0.859876,
		0.84704,
		0.834112,
		0.821134,
		0.808146,
		0.795181,
		0.782268,
		0.769433,
		0.756699,
		0.744084,
		0.731605,
		0.719275,
		0.707107,
		0.695109,
		0.683291,
		0.671658,
		0.660215,
		0.648966,
		0.637912,
		0.627051,
		0.616381,
		0.605889,
		0.595562,
		0.585374,
		0.575293,
		0.565277,
		0.555271,
		0.54521,
		0.535021,
		0.524618,
		0.513912,
		0.502809,
		0.491211,
		0.479026,
		0.46617,
		0.452566,
		0.438158,
		0.422905,
		0.40679,
		0.389823,
		0.372038,
		0.353499,
		0.334294,
		0.314536,
		0.294361,
		0.27392,
		0.253379,
		0.232908,
		0.212684,
		0.192875,
		0.173644,
		0.155141,
		0.137498,
		0.120829,
		0.105226,
		0.090758,
		0.077474,
		0.0654,
		0.054541,
		0.044882,
		0.036394,
		0.02903,
		0.02273,
		0.017426,
		0.013038,
		0.009482,
		0.006667,
		0.004501,
		0.00289,
		0.001742,
		0.000966,
		0.000479,
		0.000201,
		6.5e-005,
		1.3e-005,
		0.002407,
		0.020197,
		0.073201,
		0.175889,
		0.325662,
		0.5,
		0.666205,
		0.797981,
		0.886448,
		0.938519,
		0.967203,
		0.983139,
		0.992122,
		0.996876,
		0.999046,
		0.999819,
		0.999989,
		1,
		0.999991,
		0.999882,
		0.999493,
		0.998635,
		0.997147,
		0.994912,
		0.991857,
		0.987949,
		0.983187,
		0.977591,
		0.971204,
		0.964076,
		0.956268,
		0.947842,
		0.938862,
		0.929393,
		0.919497,
		0.909232,
		0.898654,
		0.887815,
		0.876762,
		0.865538,
		0.854185,
		0.842738,
		0.83123,
		0.819691,
		0.808146,
		0.796619,
		0.785132,
		0.773702,
		0.762345,
		0.751076,
		0.739907,
		0.728847,
		0.717902,
		0.707073,
		0.696355,
		0.685735,
		0.675192,
		0.664693,
		0.654192,
		0.643631,
		0.632938,
		0.622029,
		0.61081,
		0.599175,
		0.587017,
		0.574223,
		0.560686,
		0.546305,
		0.530993,
		0.514679,
		0.497317,
		0.478887,
		0.459397,
		0.438887,
		0.417432,
		0.395137,
		0.372134,
		0.348584,
		0.324667,
		0.300582,
		0.276532,
		0.252728,
		0.229374,
		0.206667,
		0.184787,
		0.163895,
		0.144131,
		0.125608,
		0.108413,
		0.092608,
		0.078227,
		0.06528,
		0.053754,
		0.043615,
		0.034811,
		0.027273,
		0.020921,
		0.015662,
		0.011397,
		0.008018,
		0.005416,
		0.003479,
		0.002098,
		0.001164,
		0.000577,
		0.000243,
		7.9e-005,
		1.6e-005,
		0.010604,
		0.057659,
		0.154555,
		0.294361,
		0.45206,
		0.598814,
		0.716734,
		0.803188,
		0.864779,
		0.909189,
		0.941362,
		0.964076,
		0.9794,
		0.989154,
		0.994912,
		0.997978,
		0.999381,
		0.999882,
		0.999993,
		1,
		0.999994,
		0.99992,
		0.999651,
		0.999046,
		0.997978,
		0.996348,
		0.994084,
		0.991144,
		0.987511,
		0.983187,
		0.978187,
		0.972543,
		0.966289,
		0.959469,
		0.952128,
		0.944313,
		0.93607,
		0.927446,
		0.918486,
		0.909232,
		0.899725,
		0.890001,
		0.880097,
		0.870045,
		0.859876,
		0.849616,
		0.839291,
		0.828924,
		0.818535,
		0.808142,
		0.797756,
		0.787387,
		0.777035,
		0.766692,
		0.756338,
		0.745942,
		0.735456,
		0.724819,
		0.713954,
		0.702768,
		0.691157,
		0.679005,
		0.666188,
		0.652581,
		0.63806,
		0.622512,
		0.605834,
		0.587947,
		0.568794,
		0.548352,
		0.526629,
		0.50367,
		0.47956,
		0.454416,
		0.428393,
		0.401677,
		0.374475,
		0.347016,
		0.31954,
		0.292291,
		0.265509,
		0.239425,
		0.214251,
		0.190179,
		0.167375,
		0.145975,
		0.126086,
		0.107782,
		0.091109,
		0.076083,
		0.062691,
		0.0509,
		0.040651,
		0.031869,
		0.024461,
		0.018324,
		0.013341,
		0.009391,
		0.006347,
		0.00408,
		0.002461,
		0.001366,
		0.000678,
		0.000285,
		9.3e-005,
		1.9e-005,
		0.021807,
		0.093895,
		0.208583,
		0.342523,
		0.472358,
		0.585368,
		0.679351,
		0.756663,
		0.819691,
		0.870045,
		0.909232,
		0.938862,
		0.960553,
		0.975845,
		0.986146,
		0.9927,
		0.996571,
		0.998635,
		0.999581,
		0.99992,
		0.999995,
		1,
		0.999996,
		0.999944,
		0.999752,
		0.999313,
		0.998528,
		0.997312,
		0.995601,
		0.99335,
		0.990536,
		0.987147,
		0.983187,
		0.978669,
		0.973615,
		0.968053,
		0.962012,
		0.955526,
		0.948631,
		0.941362,
		0.933754,
		0.925841,
		0.917657,
		0.909232,
		0.900598,
		0.891783,
		0.882813,
		0.873711,
		0.864496,
		0.855181,
		0.845772,
		0.836264,
		0.826642,
		0.816874,
		0.806913,
		0.796695,
		0.786138,
		0.775143,
		0.763595,
		0.751367,
		0.738324,
		0.724327,
		0.709237,
		0.692926,
		0.675281,
		0.656209,
		0.635646,
		0.613563,
		0.589968,
		0.564909,
		0.538479,
		0.510811,
		0.482077,
		0.452485,
		0.422271,
		0.391695,
		0.361029,
		0.330552,
		0.300538,
		0.271253,
		0.242943,
		0.21583,
		0.190107,
		0.165934,
		0.143438,
		0.122709,
		0.103804,
		0.086747,
		0.07153,
		0.058116,
		0.046446,
		0.036436,
		0.027985,
		0.020976,
		0.015282,
		0.010764,
		0.007279,
		0.004682,
		0.002826,
		0.00157,
		0.000779,
		0.000328,
		0.000107,
		2.2e-005,
		0.028801,
		0.106843,
		0.212988,
		0.327031,
		0.437117,
		0.538297,
		0.628601,
		0.707107,
		0.773702,
		0.828924,
		0.873716,
		0.909232,
		0.936718,
		0.957423,
		0.972543,
		0.983187,
		0.990352,
		0.994912,
		0.997607,
		0.999046,
		0.999707,
		0.999944,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.999819,
		0.999493,
		0.998903,
		0.997978,
		0.996662,
		0.994912,
		0.9927,
		0.990011,
		0.986838,
		0.983187,
		0.979066,
		0.974494,
		0.969489,
		0.964076,
		0.95828,
		0.952128,
		0.945646,
		0.938862,
		0.931802,
		0.92449,
		0.916947,
		0.909189,
		0.901223,
		0.893048,
		0.884648,
		0.875992,
		0.867033,
		0.857704,
		0.847917,
		0.837568,
		0.826533,
		0.814675,
		0.801847,
		0.787895,
		0.77267,
		0.756027,
		0.737842,
		0.71801,
		0.696458,
		0.673152,
		0.648096,
		0.621342,
		0.592988,
		0.563181,
		0.532108,
		0.5,
		0.467118,
		0.43375,
		0.400201,
		0.366781,
		0.333802,
		0.30156,
		0.270336,
		0.240381,
		0.211916,
		0.185128,
		0.160162,
		0.137128,
		0.116093,
		0.097092,
		0.08012,
		0.065143,
		0.052099,
		0.0409,
		0.031435,
		0.023578,
		0.017189,
		0.012115,
		0.008198,
		0.005276,
		0.003186,
		0.001771,
		0.000879,
		0.000371,
		0.000121,
		2.4e-005,
		0.028884,
		0.100129,
		0.19336,
		0.294891,
		0.39646,
		0.492972,
		0.581406,
		0.660215,
		0.728851,
		0.787425,
		0.836467,
		0.876762,
		0.909232,
		0.93486,
		0.95463,
		0.969489,
		0.980326,
		0.987949,
		0.993081,
		0.996348,
		0.99828,
		0.999313,
		0.999788,
		0.999959,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999864,
		0.999618,
		0.999166,
		0.998451,
		0.997422,
		0.996042,
		0.994282,
		0.992122,
		0.989554,
		0.986574,
		0.983187,
		0.9794,
		0.975226,
		0.970681,
		0.965783,
		0.960548,
		0.954993,
		0.94913,
		0.942964,
		0.936491,
		0.929694,
		0.92254,
		0.914977,
		0.906936,
		0.898322,
		0.889025,
		0.878913,
		0.867837,
		0.855639,
		0.842154,
		0.827216,
		0.810671,
		0.792377,
		0.772221,
		0.750119,
		0.726028,
		0.699951,
		0.67194,
		0.642099,
		0.610581,
		0.577591,
		0.543376,
		0.508222,
		0.472443,
		0.436373,
		0.400354,
		0.36473,
		0.32983,
		0.295966,
		0.26342,
		0.232442,
		0.203242,
		0.175988,
		0.150806,
		0.12778,
		0.106953,
		0.088328,
		0.071873,
		0.057525,
		0.045192,
		0.034759,
		0.02609,
		0.019033,
		0.013424,
		0.009089,
		0.005853,
		0.003537,
		0.001968,
		0.000977,
		0.000412,
		0.000134,
		2.7e-005,
		0.025825,
		0.088879,
		0.172561,
		0.265625,
		0.360554,
		0.452462,
		0.538323,
		0.61641,
		0.685901,
		0.746597,
		0.798713,
		0.842738,
		0.879329,
		0.909232,
		0.933236,
		0.952128,
		0.966673,
		0.977591,
		0.985548,
		0.991144,
		0.994912,
		0.997312,
		0.998733,
		0.999493,
		0.999844,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999977,
		0.999897,
		0.999707,
		0.999355,
		0.998793,
		0.997978,
		0.996876,
		0.995458,
		0.993706,
		0.991607,
		0.989153,
		0.986344,
		0.983181,
		0.979666,
		0.975798,
		0.971575,
		0.966984,
		0.961999,
		0.956583,
		0.950676,
		0.9442,
		0.937057,
		0.929125,
		0.920262,
		0.91031,
		0.899099,
		0.886448,
		0.872181,
		0.856127,
		0.838132,
		0.818069,
		0.795845,
		0.771408,
		0.744755,
		0.715937,
		0.685059,
		0.652282,
		0.617822,
		0.581942,
		0.544946,
		0.507174,
		0.468986,
		0.430753,
		0.392847,
		0.35563,
		0.319444,
		0.284601,
		0.251376,
		0.220006,
		0.19068,
		0.163544,
		0.138695,
		0.116188,
		0.096035,
		0.078208,
		0.062646,
		0.049254,
		0.037912,
		0.028478,
		0.02079,
		0.014673,
		0.009942,
		0.006407,
		0.003874,
		0.002157,
		0.001072,
		0.000452,
		0.000147,
		3e-005,
		0.022747,
		0.078987,
		0.154777,
		0.24037,
		0.32905,
		0.416289,
		0.499142,
		0.575799,
		0.645261,
		0.707107,
		0.761317,
		0.808146,
		0.848031,
		0.881522,
		0.909232,
		0.931803,
		0.949876,
		0.964076,
		0.974996,
		0.983187,
		0.989154,
		0.99335,
		0.996177,
		0.997978,
		0.999046,
		0.999618,
		0.999882,
		0.999977,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.99992,
		0.999771,
		0.999493,
		0.999046,
		0.998392,
		0.997501,
		0.996346,
		0.994906,
		0.993163,
		0.991097,
		0.988688,
		0.985907,
		0.982718,
		0.979067,
		0.974887,
		0.970086,
		0.964555,
		0.958161,
		0.950751,
		0.942154,
		0.932184,
		0.92065,
		0.907357,
		0.892122,
		0.874775,
		0.855177,
		0.83322,
		0.808844,
		0.782037,
		0.752845,
		0.721373,
		0.687787,
		0.652309,
		0.615215,
		0.576825,
		0.537498,
		0.497617,
		0.457578,
		0.417783,
		0.37862,
		0.340459,
		0.303641,
		0.268467,
		0.235198,
		0.204045,
		0.175172,
		0.148694,
		0.124676,
		0.103142,
		0.084068,
		0.067396,
		0.053032,
		0.040853,
		0.030711,
		0.022437,
		0.015848,
		0.010746,
		0.00693,
		0.004194,
		0.002336,
		0.001162,
		0.000491,
		0.00016,
		3.3e-005,
		0.020173,
		0.070648,
		0.139573,
		0.218463,
		0.301315,
		0.383959,
		0.463572,
		0.538323,
		0.607109,
		0.669354,
		0.72486,
		0.773702,
		0.816138,
		0.852555,
		0.883417,
		0.909232,
		0.930531,
		0.947842,
		0.961681,
		0.972543,
		0.980889,
		0.987147,
		0.991705,
		0.994912,
		0.997073,
		0.998451,
		0.999268,
		0.999707,
		0.999909,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999986,
		0.999937,
		0.999818,
		0.999595,
		0.99923,
		0.998688,
		0.99793,
		0.996917,
		0.995602,
		0.993929,
		0.991829,
		0.989217,
		0.985988,
		0.982016,
		0.977154,
		0.971234,
		0.964069,
		0.955459,
		0.945195,
		0.933069,
		0.918878,
		0.902438,
		0.883593,
		0.862222,
		0.838252,
		0.81166,
		0.782486,
		0.75083,
		0.716858,
		0.680795,
		0.642924,
		0.603579,
		0.563133,
		0.521988,
		0.480563,
		0.43928,
		0.398556,
		0.358785,
		0.320332,
		0.283525,
		0.248646,
		0.21593,
		0.185558,
		0.157661,
		0.13232,
		0.109565,
		0.089384,
		0.07172,
		0.056483,
		0.043547,
		0.032763,
		0.023956,
		0.016933,
		0.011491,
		0.007416,
		0.004491,
		0.002503,
		0.001246,
		0.000526,
		0.000172,
		3.5e-005,
		0.018011,
		0.063557,
		0.12648,
		0.199356,
		0.276811,
		0.355015,
		0.431293,
		0.503836,
		0.571482,
		0.63355,
		0.689715,
		0.739908,
		0.78425,
		0.82299,
		0.856465,
		0.88507,
		0.909232,
		0.929393,
		0.945995,
		0.959469,
		0.97023,
		0.978669,
		0.985151,
		0.990011,
		0.993551,
		0.996042,
		0.997721,
		0.998793,
		0.999429,
		0.999771,
		0.999929,
		0.999986,
		0.999999,
		1,
		0.999998,
		0.999983,
		0.999932,
		0.999806,
		0.999561,
		0.999139,
		0.99847,
		0.997471,
		0.996039,
		0.994054,
		0.991373,
		0.987832,
		0.983245,
		0.977409,
		0.970107,
		0.961111,
		0.950194,
		0.937137,
		0.921738,
		0.903823,
		0.883254,
		0.859942,
		0.833854,
		0.805015,
		0.773521,
		0.73953,
		0.703269,
		0.665023,
		0.625132,
		0.58398,
		0.541984,
		0.499578,
		0.457205,
		0.415302,
		0.374285,
		0.334543,
		0.296425,
		0.260237,
		0.22623,
		0.194608,
		0.165515,
		0.139047,
		0.115245,
		0.094104,
		0.075576,
		0.059573,
		0.04597,
		0.034615,
		0.025331,
		0.01792,
		0.01217,
		0.00786,
		0.004764,
		0.002657,
		0.001323,
		0.00056,
		0.000183,
		3.7e-005,
		0.016179,
		0.057479,
		0.11513,
		0.182603,
		0.255081,
		0.32905,
		0.40199,
		0.472143,
		0.538323,
		0.599783,
		0.656102,
		0.707107,
		0.752804,
		0.793332,
		0.828924,
		0.859876,
		0.886525,
		0.909232,
		0.928371,
		0.944313,
		0.957423,
		0.968053,
		0.976536,
		0.983187,
		0.988294,
		0.992122,
		0.994912,
		0.996876,
		0.9982,
		0.999046,
		0.999548,
		0.999818,
		0.999942,
		0.999983,
		0.999981,
		0.999952,
		0.999886,
		0.999749,
		0.999483,
		0.999005,
		0.99821,
		0.996969,
		0.995127,
		0.992508,
		0.988911,
		0.984115,
		0.977887,
		0.969981,
		0.960152,
		0.948161,
		0.933787,
		0.916836,
		0.897154,
		0.874633,
		0.849223,
		0.820937,
		0.789857,
		0.756135,
		0.719989,
		0.681705,
		0.641623,
		0.600132,
		0.557657,
		0.514646,
		0.471554,
		0.428835,
		0.386925,
		0.346229,
		0.307118,
		0.269916,
		0.234893,
		0.202269,
		0.172206,
		0.144812,
		0.12014,
		0.098195,
		0.078936,
		0.062277,
		0.0481,
		0.036251,
		0.026551,
		0.018799,
		0.012777,
		0.008259,
		0.005009,
		0.002796,
		0.001394,
		0.00059,
		0.000193,
		3.9e-005,
		0.014613,
		0.05223,
		0.105229,
		0.167843,
		0.23574,
		0.3057,
		0.375363,
		0.443032,
		0.507526,
		0.568056,
		0.624138,
		0.675515,
		0.722107,
		0.763962,
		0.801227,
		0.834112,
		0.862877,
		0.887815,
		0.909232,
		0.927446,
		0.942774,
		0.955526,
		0.966004,
		0.974494,
		0.981267,
		0.986574,
		0.99065,
		0.993706,
		0.995933,
		0.997501,
		0.998558,
		0.99923,
		0.99962,
		0.999806,
		0.999841,
		0.999749,
		0.999523,
		0.999123,
		0.998474,
		0.997467,
		0.995957,
		0.993768,
		0.990698,
		0.986518,
		0.980982,
		0.973833,
		0.96481,
		0.953657,
		0.940135,
		0.924032,
		0.905174,
		0.883435,
		0.858746,
		0.831105,
		0.800578,
		0.767303,
		0.73149,
		0.693416,
		0.653418,
		0.611886,
		0.569245,
		0.52595,
		0.482466,
		0.439258,
		0.396774,
		0.355436,
		0.31563,
		0.277694,
		0.241918,
		0.208535,
		0.177722,
		0.1496,
		0.124235,
		0.101641,
		0.081783,
		0.064584,
		0.049927,
		0.037661,
		0.027608,
		0.019564,
		0.013309,
		0.008609,
		0.005226,
		0.00292,
		0.001456,
		0.000617,
		0.000202,
		4.1e-005,
		0.013263,
		0.047667,
		0.096544,
		0.154777,
		0.218463,
		0.28465,
		0.351133,
		0.416289,
		0.478955,
		0.538323,
		0.593863,
		0.645261,
		0.692366,
		0.735156,
		0.773702,
		0.808146,
		0.838682,
		0.865538,
		0.888967,
		0.909232,
		0.926607,
		0.941362,
		0.953766,
		0.964076,
		0.972543,
		0.9794,
		0.984868,
		0.989153,
		0.992443,
		0.994906,
		0.996693,
		0.99793,
		0.998721,
		0.999139,
		0.99923,
		0.999005,
		0.998438,
		0.997467,
		0.99599,
		0.993865,
		0.990914,
		0.986925,
		0.981661,
		0.974863,
		0.966268,
		0.955611,
		0.942641,
		0.927134,
		0.908898,
		0.887791,
		0.863728,
		0.836688,
		0.80672,
		0.773948,
		0.738568,
		0.700845,
		0.661108,
		0.61974,
		0.577165,
		0.533837,
		0.490224,
		0.446798,
		0.404013,
		0.362304,
		0.322066,
		0.283651,
		0.247362,
		0.213445,
		0.18209,
		0.153429,
		0.127539,
		0.104444,
		0.084118,
		0.06649,
		0.051448,
		0.038844,
		0.0285,
		0.020214,
		0.013763,
		0.00891,
		0.005413,
		0.003027,
		0.001511,
		0.00064,
		0.00021,
		4.3e-005,
		0.012092,
		0.043674,
		0.088884,
		0.14316,
		0.202976,
		0.265625,
		0.32905,
		0.391707,
		0.452462,
		0.510503,
		0.565272,
		0.61641,
		0.663715,
		0.707107,
		0.746597,
		0.782268,
		0.814257,
		0.842738,
		0.867913,
		0.890001,
		0.909232,
		0.925841,
		0.940063,
		0.952128,
		0.962262,
		0.970681,
		0.97759,
		0.983181,
		0.987631,
		0.991097,
		0.993717,
		0.995602,
		0.996836,
		0.997471,
		0.997522,
		0.996969,
		0.995751,
		0.993768,
		0.990884,
		0.986925,
		0.981687,
		0.974941,
		0.966443,
		0.95594,
		0.943187,
		0.927957,
		0.910055,
		0.889329,
		0.865681,
		0.839077,
		0.809553,
		0.777215,
		0.742246,
		0.704898,
		0.665486,
		0.624385,
		0.58201,
		0.53881,
		0.495252,
		0.451806,
		0.408933,
		0.367068,
		0.326616,
		0.287937,
		0.251342,
		0.217088,
		0.185375,
		0.156346,
		0.130086,
		0.10663,
		0.085958,
		0.068007,
		0.052669,
		0.039801,
		0.029228,
		0.020748,
		0.014139,
		0.009162,
		0.005571,
		0.003117,
		0.001557,
		0.000661,
		0.000216,
		4.4e-005,
		0.011069,
		0.040162,
		0.082095,
		0.132788,
		0.189047,
		0.248386,
		0.308888,
		0.369091,
		0.427896,
		0.484498,
		0.538323,
		0.588982,
		0.636233,
		0.679948,
		0.720092,
		0.756699,
		0.789856,
		0.819691,
		0.846361,
		0.870045,
		0.890936,
		0.909232,
		0.92514,
		0.938862,
		0.9506,
		0.960548,
		0.96889,
		0.975798,
		0.981426,
		0.985907,
		0.989349,
		0.991829,
		0.993394,
		0.994054,
		0.993781,
		0.992508,
		0.990133,
		0.986518,
		0.981492,
		0.974863,
		0.966419,
		0.95594,
		0.943208,
		0.928019,
		0.910193,
		0.889586,
		0.866103,
		0.839707,
		0.810427,
		0.778361,
		0.743679,
		0.706621,
		0.667493,
		0.626654,
		0.584512,
		0.541506,
		0.498098,
		0.454751,
		0.411925,
		0.370055,
		0.329549,
		0.290769,
		0.254032,
		0.219602,
		0.187686,
		0.158434,
		0.13194,
		0.108244,
		0.087336,
		0.069157,
		0.053606,
		0.040544,
		0.029799,
		0.021172,
		0.014439,
		0.009364,
		0.005698,
		0.003191,
		0.001596,
		0.000677,
		0.000222,
		4.5e-005,
		0.010171,
		0.037057,
		0.076051,
		0.123492,
		0.176478,
		0.232725,
		0.290446,
		0.34826,
		0.405109,
		0.460201,
		0.512956,
		0.562964,
		0.609954,
		0.653763,
		0.694316,
		0.731605,
		0.765677,
		0.796619,
		0.824552,
		0.849616,
		0.87197,
		0.891783,
		0.909231,
		0.92449,
		0.937734,
		0.94913,
		0.958834,
		0.966984,
		0.973698,
		0.979067,
		0.983155,
		0.985988,
		0.987562,
		0.987832,
		0.986721,
		0.984115,
		0.979875,
		0.973833,
		0.965809,
		0.955611,
		0.943052,
		0.927957,
		0.910174,
		0.889586,
		0.86612,
		0.839755,
		0.810531,
		0.778551,
		0.743987,
		0.707073,
		0.668108,
		0.627445,
		0.585481,
		0.542647,
		0.499394,
		0.456182,
		0.413461,
		0.371666,
		0.331199,
		0.292425,
		0.25566,
		0.22117,
		0.189167,
		0.159806,
		0.133186,
		0.109353,
		0.0883,
		0.069976,
		0.054285,
		0.041091,
		0.030225,
		0.021491,
		0.014669,
		0.009521,
		0.005798,
		0.00325,
		0.001626,
		0.000691,
		0.000227,
		4.6e-005,
		0.009378,
		0.034297,
		0.070648,
		0.11513,
		0.165103,
		0.218463,
		0.273547,
		0.32905,
		0.383959,
		0.437501,
		0.489096,
		0.538323,
		0.584888,
		0.628601,
		0.669354,
		0.707107,
		0.741872,
		0.773702,
		0.802682,
		0.828924,
		0.852554,
		0.873711,
		0.892541,
		0.909189,
		0.923796,
		0.936491,
		0.94739,
		0.956583,
		0.964136,
		0.970086,
		0.974436,
		0.977154,
		0.978178,
		0.977409,
		0.974726,
		0.969981,
		0.963014,
		0.953657,
		0.941747,
		0.927134,
		0.909691,
		0.889329,
		0.865999,
		0.839707,
		0.810517,
		0.778551,
		0.743999,
		0.707107,
		0.668181,
		0.627576,
		0.585688,
		0.542944,
		0.49979,
		0.456678,
		0.414053,
		0.372344,
		0.331948,
		0.293225,
		0.256491,
		0.222012,
		0.189997,
		0.160605,
		0.133936,
		0.110041,
		0.088916,
		0.070513,
		0.054739,
		0.041464,
		0.030522,
		0.021718,
		0.014835,
		0.009635,
		0.005872,
		0.003294,
		0.001649,
		0.000701,
		0.00023,
		4.7e-005,
		0.008674,
		0.031834,
		0.065798,
		0.107582,
		0.154777,
		0.205443,
		0.258031,
		0.31131,
		0.364313,
		0.416289,
		0.466665,
		0.515011,
		0.561021,
		0.604481,
		0.645261,
		0.683291,
		0.718555,
		0.751076,
		0.780912,
		0.808142,
		0.832862,
		0.855181,
		0.875207,
		0.893048,
		0.908799,
		0.92254,
		0.93433,
		0.9442,
		0.952154,
		0.958161,
		0.962163,
		0.964069,
		0.963764,
		0.961111,
		0.955961,
		0.948161,
		0.937562,
		0.924032,
		0.907465,
		0.887791,
		0.864986,
		0.839077,
		0.810153,
		0.778361,
		0.743911,
		0.707073,
		0.668171,
		0.627576,
		0.585696,
		0.542966,
		0.499836,
		0.45676,
		0.414179,
		0.37252,
		0.332175,
		0.293501,
		0.25681,
		0.222364,
		0.190372,
		0.160989,
		0.134318,
		0.110408,
		0.089259,
		0.070823,
		0.055011,
		0.041695,
		0.03071,
		0.021866,
		0.014945,
		0.009713,
		0.005924,
		0.003325,
		0.001666,
		0.000709,
		0.000233,
		4.8e-005,
		0.008046,
		0.029626,
		0.061429,
		0.100747,
		0.145377,
		0.193529,
		0.243759,
		0.294906,
		0.346048,
		0.39646,
		0.445578,
		0.492972,
		0.538323,
		0.581406,
		0.622067,
		0.660215,
		0.695809,
		0.728847,
		0.759357,
		0.787387,
		0.813001,
		0.836264,
		0.857243,
		0.875992,
		0.892551,
		0.906936,
		0.919139,
		0.929125,
		0.936827,
		0.942154,
		0.944988,
		0.945195,
		0.942628,
		0.937137,
		0.928581,
		0.916836,
		0.901806,
		0.883435,
		0.861713,
		0.836688,
		0.808465,
		0.777215,
		0.74317,
		0.706621,
		0.667916,
		0.627445,
		0.585637,
		0.542944,
		0.49983,
		0.45676,
		0.414184,
		0.372533,
		0.332201,
		0.293546,
		0.256878,
		0.222455,
		0.190485,
		0.161121,
		0.134464,
		0.110561,
		0.089413,
		0.070972,
		0.055149,
		0.041817,
		0.030815,
		0.021951,
		0.015011,
		0.009761,
		0.005956,
		0.003345,
		0.001677,
		0.000714,
		0.000235,
		4.8e-005,
		0.007484,
		0.027641,
		0.057479,
		0.094539,
		0.136799,
		0.182603,
		0.230606,
		0.279714,
		0.32905,
		0.377913,
		0.425753,
		0.472143,
		0.516759,
		0.559361,
		0.599782,
		0.637912,
		0.673685,
		0.707073,
		0.738071,
		0.766692,
		0.792954,
		0.816874,
		0.83846,
		0.857704,
		0.874577,
		0.889025,
		0.900971,
		0.91031,
		0.916918,
		0.92065,
		0.921354,
		0.918878,
		0.913075,
		0.903823,
		0.891026,
		0.874633,
		0.85464,
		0.831105,
		0.804146,
		0.773948,
		0.740761,
		0.704898,
		0.666724,
		0.626654,
		0.585137,
		0.542647,
		0.499666,
		0.456678,
		0.414148,
		0.37252,
		0.332198,
		0.293546,
		0.25688,
		0.222461,
		0.190498,
		0.161143,
		0.134494,
		0.1106,
		0.089459,
		0.071023,
		0.055202,
		0.041868,
		0.030862,
		0.021992,
		0.015044,
		0.009787,
		0.005974,
		0.003356,
		0.001684,
		0.000717,
		0.000236,
		4.8e-005,
		0.006979,
		0.025848,
		0.053897,
		0.088884,
		0.128949,
		0.172561,
		0.218463,
		0.265625,
		0.313213,
		0.360554,
		0.407111,
		0.452462,
		0.496281,
		0.53832,
		0.578397,
		0.616381,
		0.652181,
		0.685735,
		0.717001,
		0.745942,
		0.772521,
		0.796695,
		0.818404,
		0.837568,
		0.854087,
		0.867837,
		0.878677,
		0.886448,
		0.890985,
		0.892122,
		0.889702,
		0.883593,
		0.873692,
		0.859942,
		0.842338,
		0.820937,
		0.795861,
		0.767303,
		0.735522,
		0.700845,
		0.663655,
		0.624385,
		0.583504,
		0.541506,
		0.498898,
		0.456182,
		0.413844,
		0.372344,
		0.332104,
		0.293501,
		0.256861,
		0.222455,
		0.190496,
		0.161143,
		0.134495,
		0.110603,
		0.089465,
		0.071031,
		0.055213,
		0.041881,
		0.030876,
		0.022005,
		0.015057,
		0.009797,
		0.005982,
		0.003362,
		0.001687,
		0.000719,
		0.000236,
		4.9e-005,
		0.006524,
		0.024223,
		0.050639,
		0.083718,
		0.121749,
		0.163312,
		0.207232,
		0.25254,
		0.298442,
		0.344293,
		0.389572,
		0.433863,
		0.476837,
		0.518237,
		0.557864,
		0.595562,
		0.631206,
		0.664693,
		0.695929,
		0.724819,
		0.751262,
		0.775143,
		0.796331,
		0.814675,
		0.830009,
		0.842154,
		0.850921,
		0.856127,
		0.857596,
		0.855177,
		0.848753,
		0.838252,
		0.823657,
		0.805015,
		0.782445,
		0.756135,
		0.726348,
		0.693416,
		0.657731,
		0.61974,
		0.579929,
		0.53881,
		0.496909,
		0.454751,
		0.412843,
		0.371666,
		0.331662,
		0.293225,
		0.256697,
		0.222364,
		0.19045,
		0.161121,
		0.134487,
		0.1106,
		0.089464,
		0.071031,
		0.055213,
		0.041882,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.006111,
		0.022747,
		0.047667,
		0.078987,
		0.11513,
		0.154777,
		0.196828,
		0.24037,
		0.28465,
		0.329048,
		0.373058,
		0.416269,
		0.458348,
		0.499021,
		0.538067,
		0.575293,
		0.610534,
		0.643631,
		0.674429,
		0.702768,
		0.728476,
		0.751367,
		0.771243,
		0.787895,
		0.801108,
		0.810671,
		0.816383,
		0.818069,
		0.815588,
		0.808844,
		0.7978,
		0.782486,
		0.763003,
		0.73953,
		0.712321,
		0.681705,
		0.648076,
		0.611886,
		0.57363,
		0.533837,
		0.493048,
		0.451806,
		0.410642,
		0.370055,
		0.330511,
		0.292425,
		0.256157,
		0.222012,
		0.190229,
		0.160989,
		0.134412,
		0.110561,
		0.089445,
		0.071023,
		0.05521,
		0.041881,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.005737,
		0.021402,
		0.044948,
		0.074644,
		0.109032,
		0.146884,
		0.187172,
		0.229034,
		0.271752,
		0.314732,
		0.357477,
		0.399576,
		0.440684,
		0.480503,
		0.518777,
		0.555271,
		0.589761,
		0.622029,
		0.651854,
		0.679005,
		0.703245,
		0.724327,
		0.742003,
		0.756027,
		0.766169,
		0.772221,
		0.774009,
		0.771408,
		0.764353,
		0.752845,
		0.736961,
		0.716858,
		0.692773,
		0.665023,
		0.633993,
		0.600132,
		0.56394,
		0.52595,
		0.486716,
		0.446798,
		0.406741,
		0.367068,
		0.328265,
		0.290769,
		0.254963,
		0.22117,
		0.189652,
		0.160605,
		0.134165,
		0.110408,
		0.089354,
		0.070972,
		0.055183,
		0.041868,
		0.030871,
		0.022005,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		0.005396,
		0.020173,
		0.042454,
		0.070648,
		0.103402,
		0.139572,
		0.178194,
		0.218452,
		0.259662,
		0.301242,
		0.342703,
		0.383622,
		0.423631,
		0.462401,
		0.499627,
		0.535021,
		0.568298,
		0.599175,
		0.627366,
		0.652581,
		0.674529,
		0.692926,
		0.707502,
		0.71801,
		0.724239,
		0.726028,
		0.723272,
		0.715937,
		0.704067,
		0.687787,
		0.667309,
		0.642924,
		0.615002,
		0.58398,
		0.55035,
		0.514646,
		0.477426,
		0.439258,
		0.400704,
		0.362304,
		0.324562,
		0.287937,
		0.252835,
		0.219602,
		0.18852,
		0.159806,
		0.133616,
		0.110041,
		0.089116,
		0.070823,
		0.055094,
		0.041817,
		0.030844,
		0.021992,
		0.015052,
		0.009797,
		0.005984,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.005084,
		0.019046,
		0.040162,
		0.066962,
		0.098192,
		0.132781,
		0.169822,
		0.208539,
		0.248267,
		0.288436,
		0.328548,
		0.368158,
		0.406867,
		0.444298,
		0.480096,
		0.513912,
		0.545402,
		0.574223,
		0.600037,
		0.622512,
		0.641333,
		0.656209,
		0.666884,
		0.673152,
		0.674863,
		0.67194,
		0.664385,
		0.652282,
		0.635805,
		0.615215,
		0.590853,
		0.563133,
		0.532534,
		0.499578,
		0.464821,
		0.428835,
		0.392189,
		0.355436,
		0.319097,
		0.283651,
		0.249525,
		0.217088,
		0.186644,
		0.158434,
		0.132632,
		0.109353,
		0.088647,
		0.070513,
		0.054895,
		0.041695,
		0.030771,
		0.021951,
		0.015031,
		0.009787,
		0.005979,
		0.003362,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.004799,
		0.018011,
		0.03805,
		0.063554,
		0.093355,
		0.12645,
		0.161974,
		0.199181,
		0.237419,
		0.276112,
		0.314743,
		0.352837,
		0.38995,
		0.425657,
		0.459545,
		0.491211,
		0.520259,
		0.546305,
		0.568982,
		0.587947,
		0.602893,
		0.613563,
		0.619756,
		0.621342,
		0.618271,
		0.610581,
		0.598399,
		0.581942,
		0.561514,
		0.537498,
		0.510346,
		0.480563,
		0.448693,
		0.415302,
		0.38096,
		0.346229,
		0.311642,
		0.277694,
		0.244832,
		0.213445,
		0.183862,
		0.156346,
		0.131095,
		0.108244,
		0.087866,
		0.069976,
		0.054538,
		0.041464,
		0.030628,
		0.021866,
		0.014983,
		0.009761,
		0.005967,
		0.003356,
		0.001686,
		0.000719,
		0.000237,
		4.9e-005,
		0.004537,
		0.017057,
		0.036097,
		0.06039,
		0.088841,
		0.120501,
		0.154541,
		0.190228,
		0.22691,
		0.263995,
		0.300936,
		0.337219,
		0.372351,
		0.405858,
		0.437279,
		0.46617,
		0.492103,
		0.514679,
		0.533535,
		0.548352,
		0.558873,
		0.564909,
		0.566352,
		0.563181,
		0.555467,
		0.543376,
		0.527165,
		0.507174,
		0.48382,
		0.457578,
		0.428973,
		0.398556,
		0.366892,
		0.334543,
		0.302048,
		0.269916,
		0.238608,
		0.208535,
		0.180046,
		0.153429,
		0.128904,
		0.10663,
		0.086702,
		0.069157,
		0.053976,
		0.041091,
		0.030388,
		0.021718,
		0.014896,
		0.009713,
		0.005942,
		0.003345,
		0.001681,
		0.000717,
		0.000236,
		4.9e-005,
		0.004295,
		0.016175,
		0.034281,
		0.057429,
		0.084585,
		0.114839,
		0.147381,
		0.181483,
		0.216478,
		0.251747,
		0.286708,
		0.320804,
		0.3535,
		0.384281,
		0.412655,
		0.438158,
		0.460362,
		0.478887,
		0.493408,
		0.50367,
		0.5095,
		0.510811,
		0.507609,
		0.5,
		0.488182,
		0.472443,
		0.453154,
		0.430753,
		0.405731,
		0.37862,
		0.349969,
		0.320332,
		0.29025,
		0.260237,
		0.230768,
		0.202269,
		0.17511,
		0.1496,
		0.125984,
		0.104444,
		0.085099,
		0.068007,
		0.053172,
		0.040544,
		0.03003,
		0.021491,
		0.014759,
		0.009635,
		0.005901,
		0.003325,
		0.001672,
		0.000714,
		0.000235,
		4.8e-005,
		0.004071,
		0.015352,
		0.032575,
		0.054623,
		0.080506,
		0.109339,
		0.140319,
		0.172707,
		0.205814,
		0.238991,
		0.271617,
		0.3031,
		0.332875,
		0.360406,
		0.385195,
		0.40679,
		0.424797,
		0.438887,
		0.448814,
		0.454416,
		0.455628,
		0.452485,
		0.445116,
		0.43375,
		0.4187,
		0.400354,
		0.379165,
		0.35563,
		0.330276,
		0.303641,
		0.276259,
		0.248646,
		0.221283,
		0.194608,
		0.169008,
		0.144812,
		0.122288,
		0.101641,
		0.083014,
		0.06649,
		0.052095,
		0.039801,
		0.029533,
		0.021172,
		0.014563,
		0.009521,
		0.005838,
		0.003294,
		0.001658,
		0.000709,
		0.000234,
		4.8e-005,
		0.003862,
		0.014576,
		0.030947,
		0.05191,
		0.076504,
		0.103854,
		0.133151,
		0.163637,
		0.194598,
		0.225354,
		0.255258,
		0.283695,
		0.310092,
		0.333919,
		0.354703,
		0.372038,
		0.385597,
		0.395137,
		0.400512,
		0.401677,
		0.398686,
		0.391695,
		0.38095,
		0.366781,
		0.34959,
		0.32983,
		0.307995,
		0.284601,
		0.260166,
		0.235198,
		0.21018,
		0.185558,
		0.161731,
		0.139047,
		0.117792,
		0.098195,
		0.080424,
		0.064584,
		0.050725,
		0.038844,
		0.028884,
		0.020748,
		0.014298,
		0.009364,
		0.005752,
		0.00325,
		0.001639,
		0.000701,
		0.000232,
		4.8e-005,
		0.003662,
		0.013828,
		0.029357,
		0.049219,
		0.072466,
		0.098223,
		0.125665,
		0.154017,
		0.182539,
		0.210532,
		0.237338,
		0.262345,
		0.284996,
		0.304801,
		0.321343,
		0.334294,
		0.343419,
		0.348584,
		0.349758,
		0.347016,
		0.340527,
		0.330552,
		0.317429,
		0.30156,
		0.283397,
		0.26342,
		0.242125,
		0.220006,
		0.197539,
		0.175172,
		0.153312,
		0.13232,
		0.112501,
		0.094104,
		0.077319,
		0.062277,
		0.049052,
		0.037661,
		0.028075,
		0.020214,
		0.01396,
		0.009162,
		0.005638,
		0.003191,
		0.001612,
		0.000691,
		0.000228,
		4.7e-005,
		0.003469,
		0.01309,
		0.027762,
		0.046472,
		0.068277,
		0.09229,
		0.117674,
		0.143636,
		0.169425,
		0.194342,
		0.21774,
		0.239036,
		0.257723,
		0.273376,
		0.285665,
		0.294361,
		0.299339,
		0.300582,
		0.298172,
		0.292291,
		0.283205,
		0.271253,
		0.256833,
		0.240381,
		0.22236,
		0.203242,
		0.183489,
		0.163544,
		0.143817,
		0.124676,
		0.106443,
		0.089384,
		0.073709,
		0.059573,
		0.047072,
		0.036251,
		0.0271,
		0.019564,
		0.013546,
		0.00891,
		0.005496,
		0.003117,
		0.001578,
		0.000677,
		0.000224,
		4.6e-005,
		0.003275,
		0.01234,
		0.026116,
		0.043596,
		0.063833,
		0.085931,
		0.109042,
		0.132372,
		0.155176,
		0.176778,
		0.196569,
		0.214025,
		0.228713,
		0.240301,
		0.248563,
		0.253379,
		0.254737,
		0.252728,
		0.247536,
		0.239425,
		0.228728,
		0.21583,
		0.201151,
		0.185128,
		0.168203,
		0.150806,
		0.133344,
		0.116188,
		0.099669,
		0.084068,
		0.069614,
		0.056483,
		0.044794,
		0.034615,
		0.025961,
		0.018799,
		0.013053,
		0.008609,
		0.005324,
		0.003027,
		0.001535,
		0.000661,
		0.000219,
		4.5e-005,
		0.003077,
		0.011559,
		0.02438,
		0.040531,
		0.05906,
		0.07907,
		0.099716,
		0.120219,
		0.139866,
		0.158028,
		0.174161,
		0.187825,
		0.198682,
		0.206504,
		0.211175,
		0.212684,
		0.211121,
		0.206667,
		0.19958,
		0.190179,
		0.178832,
		0.165934,
		0.151896,
		0.137128,
		0.122023,
		0.106953,
		0.092251,
		0.078208,
		0.065071,
		0.053032,
		0.042233,
		0.032763,
		0.024662,
		0.01792,
		0.012483,
		0.008259,
		0.005121,
		0.00292,
		0.001485,
		0.00064,
		0.000213,
		4.4e-005,
		0.002868,
		0.010731,
		0.022525,
		0.037241,
		0.053931,
		0.071706,
		0.089745,
		0.107308,
		0.123736,
		0.138472,
		0.151061,
		0.161159,
		0.168534,
		0.173066,
		0.174741,
		0.173644,
		0.169947,
		0.163895,
		0.155789,
		0.145975,
		0.134822,
		0.122709,
		0.110012,
		0.097092,
		0.084279,
		0.071873,
		0.060128,
		0.049254,
		0.039411,
		0.030711,
		0.023214,
		0.016933,
		0.011839,
		0.00786,
		0.00489,
		0.002796,
		0.001426,
		0.000617,
		0.000206,
		4.3e-005,
		0.002646,
		0.009847,
		0.020541,
		0.033727,
		0.048471,
		0.063915,
		0.079284,
		0.093895,
		0.107171,
		0.118641,
		0.127953,
		0.134865,
		0.139251,
		0.141088,
		0.140451,
		0.137498,
		0.132457,
		0.125608,
		0.117269,
		0.107782,
		0.097494,
		0.086747,
		0.075864,
		0.065143,
		0.054846,
		0.045192,
		0.03636,
		0.028478,
		0.021629,
		0.015848,
		0.011126,
		0.007416,
		0.004631,
		0.002657,
		0.00136,
		0.00059,
		0.000197,
		4.1e-005,
		0.00241,
		0.008906,
		0.018437,
		0.030023,
		0.042764,
		0.055852,
		0.068577,
		0.080339,
		0.090652,
		0.099147,
		0.105574,
		0.109792,
		0.111767,
		0.11156,
		0.109311,
		0.105226,
		0.099561,
		0.092608,
		0.084677,
		0.076083,
		0.067132,
		0.058116,
		0.049296,
		0.0409,
		0.033116,
		0.02609,
		0.019924,
		0.014673,
		0.010351,
		0.00693,
		0.004345,
		0.002503,
		0.001285,
		0.00056,
		0.000188,
		3.9e-005,
		0.002161,
		0.007919,
		0.016247,
		0.026203,
		0.036943,
		0.047728,
		0.057933,
		0.067052,
		0.074699,
		0.080608,
		0.084624,
		0.086698,
		0.086869,
		0.08526,
		0.082051,
		0.077474,
		0.071791,
		0.06528,
		0.058225,
		0.0509,
		0.04356,
		0.036436,
		0.029723,
		0.023578,
		0.01812,
		0.013424,
		0.009521,
		0.006407,
		0.004037,
		0.002336,
		0.001205,
		0.000526,
		0.000177,
		3.7e-005,
		0.001902,
		0.006903,
		0.014018,
		0.022365,
		0.031171,
		0.039786,
		0.04768,
		0.054447,
		0.059802,
		0.063573,
		0.06569,
		0.066174,
		0.065122,
		0.062693,
		0.059088,
		0.054541,
		0.049299,
		0.043615,
		0.03773,
		0.031869,
		0.026229,
		0.020976,
		0.016241,
		0.012115,
		0.008648,
		0.005853,
		0.003708,
		0.002157,
		0.001117,
		0.000491,
		0.000166,
		3.5e-005,
		0.001639,
		0.005885,
		0.011815,
		0.018622,
		0.025625,
		0.032269,
		0.038126,
		0.04289,
		0.046366,
		0.048465,
		0.049186,
		0.0486,
		0.046839,
		0.04408,
		0.040526,
		0.036394,
		0.031906,
		0.027273,
		0.022689,
		0.018324,
		0.014314,
		0.010764,
		0.007741,
		0.005276,
		0.003363,
		0.001968,
		0.001025,
		0.000452,
		0.000154,
		3.3e-005,
		0.00138,
		0.004895,
		0.009702,
		0.015087,
		0.020468,
		0.025391,
		0.029527,
		0.03266,
		0.034677,
		0.035553,
		0.035336,
		0.034131,
		0.032087,
		0.029379,
		0.026196,
		0.02273,
		0.019164,
		0.015662,
		0.012367,
		0.009391,
		0.006814,
		0.004682,
		0.003007,
		0.001771,
		0.000929,
		0.000412,
		0.000141,
		3e-005,
		0.001132,
		0.003962,
		0.007742,
		0.011859,
		0.015835,
		0.019315,
		0.02206,
		0.023933,
		0.024886,
		0.024941,
		0.02418,
		0.022723,
		0.020718,
		0.018328,
		0.015716,
		0.013038,
		0.010433,
		0.008018,
		0.00588,
		0.00408,
		0.002644,
		0.00157,
		0.000829,
		0.000371,
		0.000127,
		2.7e-005,
		0.000902,
		0.003111,
		0.005983,
		0.009012,
		0.011818,
		0.014139,
		0.015814,
		0.016772,
		0.017011,
		0.016587,
		0.015596,
		0.01416,
		0.012416,
		0.010502,
		0.008547,
		0.006667,
		0.004955,
		0.003479,
		0.002279,
		0.001366,
		0.000728,
		0.000328,
		0.000114,
		2.4e-005,
		0.000697,
		0.002362,
		0.004461,
		0.006591,
		0.008464,
		0.009899,
		0.0108,
		0.011143,
		0.010961,
		0.010326,
		0.009337,
		0.008105,
		0.006746,
		0.005365,
		0.004055,
		0.00289,
		0.001918,
		0.001164,
		0.000627,
		0.000285,
		0.0001,
		2.2e-005,
		0.000519,
		0.001726,
		0.003192,
		0.00461,
		0.005775,
		0.00657,
		0.006952,
		0.00693,
		0.006556,
		0.005905,
		0.005068,
		0.004136,
		0.003197,
		0.002323,
		0.001568,
		0.000966,
		0.000527,
		0.000243,
		8.6e-005,
		1.9e-005,
		0.000371,
		0.001206,
		0.002176,
		0.003057,
		0.003714,
		0.004083,
		0.004154,
		0.00396,
		0.003555,
		0.003011,
		0.0024,
		0.00179,
		0.001235,
		0.000776,
		0.000431,
		0.000201,
		7.2e-005,
		1.6e-005,
		0.000253,
		0.000799,
		0.001398,
		0.001898,
		0.002217,
		0.002331,
		0.002251,
		0.002017,
		0.001682,
		0.001301,
		0.000926,
		0.000597,
		0.000339,
		0.000161,
		5.9e-005,
		1.3e-005,
		0.000162,
		0.000496,
		0.000834,
		0.001083,
		0.0012,
		0.001186,
		0.001064,
		0.000871,
		0.000649,
		0.000434,
		0.000254,
		0.000124,
		4.6e-005,
		1.1e-005,
		9.7e-005,
		0.000283,
		0.000451,
		0.000551,
		0.000567,
		0.000512,
		0.00041,
		0.00029,
		0.000178,
		9e-005,
		3.5e-005,
		5.2e-005,
		0.000144,
		0.000213,
		0.000237,
		0.000218,
		0.00017,
		0.000112,
		6e-005,
		2.4e-005,
		2.5e-005,
		6.2e-005,
		8.2e-005,
		7.9e-005,
		6e-005,
		3.5e-005,
		1.5e-005,
		2.1e-005,
		2.2e-005,
		1.6e-005,
		0.005562,
		0.004312,
		0.003208,
		0.002256,
		0.001462,
		0.000833,
		0.000375,
		9.5e-005,
		9.5e-005,
		0.000375,
		0.000833,
		0.001462,
		0.002256,
		0.003208,
		0.004312,
		0.021402,
		0.019046,
		0.0168,
		0.014668,
		0.012657,
		0.010771,
		0.009016,
		0.007397,
		0.00592,
		0.004591,
		0.003417,
		0.002404,
		0.001559,
		0.000888,
		0.0004,
		0.000101,
		0.000101,
		0.0004,
		0.000888,
		0.001559,
		0.002404,
		0.003417,
		0.004591,
		0.00592,
		0.007397,
		0.009016,
		0.010771,
		0.012657,
		0.014668,
		0.0168,
		0.019046,
		0.046277,
		0.043031,
		0.039864,
		0.03678,
		0.033782,
		0.030875,
		0.028064,
		0.025353,
		0.022747,
		0.020251,
		0.01787,
		0.015609,
		0.013475,
		0.011471,
		0.009606,
		0.007884,
		0.006313,
		0.004898,
		0.003647,
		0.002567,
		0.001665,
		0.000949,
		0.000428,
		0.000108,
		0.000108,
		0.000428,
		0.000949,
		0.001665,
		0.002567,
		0.003647,
		0.004898,
		0.006313,
		0.007884,
		0.009606,
		0.011471,
		0.013475,
		0.015609,
		0.01787,
		0.020251,
		0.022747,
		0.025353,
		0.028064,
		0.030875,
		0.033782,
		0.03678,
		0.039864,
		0.043031,
		0.078987,
		0.075048,
		0.071162,
		0.06733,
		0.063557,
		0.059846,
		0.0562,
		0.052623,
		0.049119,
		0.045692,
		0.042346,
		0.039085,
		0.035914,
		0.032838,
		0.02986,
		0.026987,
		0.024223,
		0.021574,
		0.019046,
		0.016644,
		0.014374,
		0.012242,
		0.010256,
		0.008421,
		0.006746,
		0.005236,
		0.003901,
		0.002747,
		0.001783,
		0.001017,
		0.000458,
		0.000116,
		0.000116,
		0.000458,
		0.001017,
		0.001783,
		0.002747,
		0.003901,
		0.005236,
		0.006746,
		0.008421,
		0.010256,
		0.012242,
		0.014374,
		0.016644,
		0.019046,
		0.021574,
		0.024223,
		0.026987,
		0.02986,
		0.032838,
		0.035914,
		0.039085,
		0.042346,
		0.045692,
		0.049119,
		0.052623,
		0.0562,
		0.059846,
		0.063557,
		0.06733,
		0.071162,
		0.075048,
		0.118371,
		0.113919,
		0.109496,
		0.105104,
		0.100747,
		0.096427,
		0.092147,
		0.08791,
		0.083718,
		0.079575,
		0.075484,
		0.071448,
		0.067471,
		0.063557,
		0.05971,
		0.055933,
		0.05223,
		0.048607,
		0.045066,
		0.041614,
		0.038254,
		0.034993,
		0.031834,
		0.028784,
		0.025848,
		0.023032,
		0.020342,
		0.017784,
		0.015366,
		0.013093,
		0.010974,
		0.009016,
		0.007225,
		0.005611,
		0.004182,
		0.002946,
		0.001913,
		0.001092,
		0.000492,
		0.000125,
		0.000125,
		0.000492,
		0.001092,
		0.001913,
		0.002946,
		0.004182,
		0.005611,
		0.007225,
		0.009016,
		0.010974,
		0.013093,
		0.015366,
		0.017784,
		0.020342,
		0.023032,
		0.025848,
		0.028784,
		0.031834,
		0.034993,
		0.038254,
		0.041614,
		0.045066,
		0.048607,
		0.05223,
		0.055933,
		0.05971,
		0.063557,
		0.067471,
		0.071448,
		0.075484,
		0.079575,
		0.083718,
		0.08791,
		0.092147,
		0.096427,
		0.100747,
		0.105104,
		0.109496,
		0.113919,
		0.163312,
		0.158507,
		0.153712,
		0.14893,
		0.144161,
		0.139409,
		0.134676,
		0.129963,
		0.125272,
		0.120607,
		0.11597,
		0.111363,
		0.106789,
		0.102252,
		0.097753,
		0.093296,
		0.088884,
		0.08452,
		0.080209,
		0.075953,
		0.071756,
		0.067623,
		0.063557,
		0.059563,
		0.055645,
		0.051808,
		0.048056,
		0.044395,
		0.04083,
		0.037366,
		0.034009,
		0.030765,
		0.027641,
		0.024641,
		0.021774,
		0.019046,
		0.016464,
		0.014037,
		0.011771,
		0.009675,
		0.007758,
		0.006028,
		0.004495,
		0.003169,
		0.002059,
		0.001176,
		0.00053,
		0.000135,
		0.000135,
		0.00053,
		0.001176,
		0.002059,
		0.003169,
		0.004495,
		0.006028,
		0.007758,
		0.009675,
		0.011771,
		0.014037,
		0.016464,
		0.019046,
		0.021774,
		0.024641,
		0.027641,
		0.030765,
		0.034009,
		0.037366,
		0.04083,
		0.044395,
		0.048056,
		0.051808,
		0.055645,
		0.059563,
		0.063557,
		0.067623,
		0.071756,
		0.075953,
		0.080209,
		0.08452,
		0.088884,
		0.093296,
		0.097753,
		0.102252,
		0.106789,
		0.111363,
		0.11597,
		0.120607,
		0.125272,
		0.129963,
		0.134676,
		0.139409,
		0.144161,
		0.14893,
		0.153712,
		0.158507,
		0.212739,
		0.207724,
		0.202704,
		0.197681,
		0.192655,
		0.187629,
		0.182603,
		0.17758,
		0.172561,
		0.167548,
		0.162543,
		0.157547,
		0.152563,
		0.147593,
		0.142639,
		0.137703,
		0.132788,
		0.127896,
		0.12303,
		0.118192,
		0.113386,
		0.108615,
		0.103881,
		0.099188,
		0.094539,
		0.089938,
		0.085389,
		0.080895,
		0.076461,
		0.07209,
		0.067787,
		0.063557,
		0.059405,
		0.055335,
		0.051353,
		0.047464,
		0.043674,
		0.039989,
		0.036415,
		0.032959,
		0.029626,
		0.026425,
		0.023363,
		0.020447,
		0.017685,
		0.015085,
		0.012657,
		0.010409,
		0.008351,
		0.006493,
		0.004845,
		0.003417,
		0.002221,
		0.001269,
		0.000573,
		0.000146,
		0.000146,
		0.000573,
		0.001269,
		0.002221,
		0.003417,
		0.004845,
		0.006493,
		0.008351,
		0.010409,
		0.012657,
		0.015085,
		0.017685,
		0.020447,
		0.023363,
		0.026425,
		0.029626,
		0.032959,
		0.036415,
		0.039989,
		0.043674,
		0.047464,
		0.051353,
		0.055335,
		0.059405,
		0.063557,
		0.067787,
		0.07209,
		0.076461,
		0.080895,
		0.085389,
		0.089938,
		0.094539,
		0.099188,
		0.103881,
		0.108615,
		0.113386,
		0.118192,
		0.12303,
		0.127896,
		0.132788,
		0.137703,
		0.142639,
		0.147593,
		0.152563,
		0.157547,
		0.162543,
		0.167548,
		0.172561,
		0.17758,
		0.182603,
		0.187629,
		0.192655,
		0.197681,
		0.202704,
		0.207724,
		0.265625,
		0.260527,
		0.255412,
		0.25028,
		0.245133,
		0.239972,
		0.234798,
		0.229612,
		0.224414,
		0.219207,
		0.213992,
		0.208769,
		0.203541,
		0.198309,
		0.193074,
		0.187838,
		0.182603,
		0.177371,
		0.172143,
		0.166922,
		0.16171,
		0.156508,
		0.151319,
		0.146146,
		0.140991,
		0.135857,
		0.130747,
		0.125662,
		0.120607,
		0.115585,
		0.110599,
		0.105651,
		0.100747,
		0.09589,
		0.091084,
		0.086332,
		0.08164,
		0.077011,
		0.072452,
		0.067965,
		0.063557,
		0.059234,
		0.055,
		0.050862,
		0.046826,
		0.042898,
		0.039085,
		0.035395,
		0.031834,
		0.02841,
		0.025132,
		0.022008,
		0.019046,
		0.016256,
		0.013648,
		0.011231,
		0.009016,
		0.007014,
		0.005236,
		0.003696,
		0.002404,
		0.001374,
		0.000621,
		0.000158,
		0.000158,
		0.000621,
		0.001374,
		0.002404,
		0.003696,
		0.005236,
		0.007014,
		0.009016,
		0.011231,
		0.013648,
		0.016256,
		0.019046,
		0.022008,
		0.025132,
		0.02841,
		0.031834,
		0.035395,
		0.039085,
		0.042898,
		0.046826,
		0.050862,
		0.055,
		0.059234,
		0.063557,
		0.067965,
		0.072452,
		0.077011,
		0.08164,
		0.086332,
		0.091084,
		0.09589,
		0.100747,
		0.105651,
		0.110599,
		0.115585,
		0.120607,
		0.125662,
		0.130747,
		0.135857,
		0.140991,
		0.146146,
		0.151319,
		0.156508,
		0.16171,
		0.166922,
		0.172143,
		0.177371,
		0.182603,
		0.187838,
		0.193074,
		0.198309,
		0.203541,
		0.208769,
		0.213992,
		0.219207,
		0.224414,
		0.229612,
		0.234798,
		0.239972,
		0.245133,
		0.25028,
		0.255412,
		0.260527,
		0.320992,
		0.315921,
		0.310823,
		0.3057,
		0.300551,
		0.295377,
		0.290179,
		0.284958,
		0.279714,
		0.274448,
		0.269161,
		0.263854,
		0.258527,
		0.253183,
		0.247821,
		0.242442,
		0.237049,
		0.231643,
		0.226223,
		0.220793,
		0.215353,
		0.209905,
		0.204451,
		0.198992,
		0.193529,
		0.188066,
		0.182603,
		0.177144,
		0.171689,
		0.166242,
		0.160804,
		0.155379,
		0.149968,
		0.144575,
		0.139203,
		0.133854,
		0.128533,
		0.123241,
		0.117983,
		0.112762,
		0.107582,
		0.102448,
		0.097363,
		0.092332,
		0.08736,
		0.082451,
		0.077611,
		0.072845,
		0.068159,
		0.063557,
		0.059048,
		0.054636,
		0.05033,
		0.046135,
		0.042059,
		0.03811,
		0.034297,
		0.030627,
		0.02711,
		0.023754,
		0.02057,
		0.017568,
		0.014759,
		0.012153,
		0.009762,
		0.0076,
		0.005678,
		0.00401,
		0.00261,
		0.001493,
		0.000675,
		0.000172,
		0.000172,
		0.000675,
		0.001493,
		0.00261,
		0.00401,
		0.005678,
		0.0076,
		0.009762,
		0.012153,
		0.014759,
		0.017568,
		0.02057,
		0.023754,
		0.02711,
		0.030627,
		0.034297,
		0.03811,
		0.042059,
		0.046135,
		0.05033,
		0.054636,
		0.059048,
		0.063557,
		0.068159,
		0.072845,
		0.077611,
		0.082451,
		0.08736,
		0.092332,
		0.097363,
		0.102448,
		0.107582,
		0.112762,
		0.117983,
		0.123241,
		0.128533,
		0.133854,
		0.139203,
		0.144575,
		0.149968,
		0.155379,
		0.160804,
		0.166242,
		0.171689,
		0.177144,
		0.182603,
		0.188066,
		0.193529,
		0.198992,
		0.204451,
		0.209905,
		0.215353,
		0.220793,
		0.226223,
		0.231643,
		0.237049,
		0.242442,
		0.247821,
		0.253183,
		0.258527,
		0.263854,
		0.269161,
		0.274448,
		0.279714,
		0.284958,
		0.290179,
		0.295377,
		0.300551,
		0.3057,
		0.310823,
		0.315921,
		0.377913,
		0.372963,
		0.36798,
		0.362964,
		0.357915,
		0.352834,
		0.34772,
		0.342574,
		0.337397,
		0.332189,
		0.32695,
		0.321682,
		0.316383,
		0.311056,
		0.3057,
		0.300316,
		0.294906,
		0.289469,
		0.284006,
		0.278519,
		0.273008,
		0.267474,
		0.261919,
		0.256343,
		0.250747,
		0.245133,
		0.239503,
		0.233856,
		0.228195,
		0.222522,
		0.216838,
		0.211144,
		0.205443,
		0.199736,
		0.194026,
		0.188314,
		0.182603,
		0.176896,
		0.171193,
		0.1655,
		0.159817,
		0.154148,
		0.148496,
		0.142864,
		0.137255,
		0.131674,
		0.126123,
		0.120607,
		0.11513,
		0.109696,
		0.10431,
		0.098976,
		0.093699,
		0.088485,
		0.083339,
		0.078267,
		0.073275,
		0.06837,
		0.063557,
		0.058845,
		0.05424,
		0.049751,
		0.045385,
		0.04115,
		0.037057,
		0.033113,
		0.02933,
		0.025717,
		0.022285,
		0.019046,
		0.016011,
		0.013194,
		0.010606,
		0.008262,
		0.006177,
		0.004366,
		0.002844,
		0.001628,
		0.000737,
		0.000188,
		0.000188,
		0.000737,
		0.001628,
		0.002844,
		0.004366,
		0.006177,
		0.008262,
		0.010606,
		0.013194,
		0.016011,
		0.019046,
		0.022285,
		0.025717,
		0.02933,
		0.033113,
		0.037057,
		0.04115,
		0.045385,
		0.049751,
		0.05424,
		0.058845,
		0.063557,
		0.06837,
		0.073275,
		0.078267,
		0.083339,
		0.088485,
		0.093699,
		0.098976,
		0.10431,
		0.109696,
		0.11513,
		0.120607,
		0.126123,
		0.131674,
		0.137255,
		0.142864,
		0.148496,
		0.154148,
		0.159817,
		0.1655,
		0.171193,
		0.176896,
		0.182603,
		0.188314,
		0.194026,
		0.199736,
		0.205443,
		0.211144,
		0.216838,
		0.222522,
		0.228195,
		0.233856,
		0.239503,
		0.245133,
		0.250747,
		0.256343,
		0.261919,
		0.267474,
		0.273008,
		0.278519,
		0.284006,
		0.289469,
		0.294906,
		0.300316,
		0.3057,
		0.311056,
		0.316383,
		0.321682,
		0.32695,
		0.332189,
		0.337397,
		0.342574,
		0.34772,
		0.352834,
		0.357915,
		0.362964,
		0.36798,
		0.372963,
		0.435513,
		0.430764,
		0.425977,
		0.421152,
		0.416289,
		0.411389,
		0.40645,
		0.401474,
		0.39646,
		0.391409,
		0.38632,
		0.381194,
		0.376031,
		0.370832,
		0.365596,
		0.360324,
		0.355015,
		0.349672,
		0.344293,
		0.33888,
		0.333432,
		0.327951,
		0.322436,
		0.316889,
		0.31131,
		0.3057,
		0.300059,
		0.294389,
		0.28869,
		0.282963,
		0.277209,
		0.271429,
		0.265625,
		0.259797,
		0.253947,
		0.248076,
		0.242186,
		0.236278,
		0.230353,
		0.224414,
		0.218463,
		0.2125,
		0.206529,
		0.200552,
		0.19457,
		0.188586,
		0.182603,
		0.176624,
		0.170651,
		0.164687,
		0.158736,
		0.1528,
		0.146884,
		0.140991,
		0.135126,
		0.129291,
		0.123492,
		0.117733,
		0.112019,
		0.106356,
		0.100747,
		0.0952,
		0.08972,
		0.084314,
		0.078987,
		0.073747,
		0.068601,
		0.063557,
		0.058623,
		0.053808,
		0.049119,
		0.044568,
		0.040162,
		0.035914,
		0.031834,
		0.027933,
		0.024223,
		0.020718,
		0.01743,
		0.014374,
		0.011564,
		0.009016,
		0.006746,
		0.004772,
		0.003111,
		0.001783,
		0.000807,
		0.000206,
		0.000206,
		0.000807,
		0.001783,
		0.003111,
		0.004772,
		0.006746,
		0.009016,
		0.011564,
		0.014374,
		0.01743,
		0.020718,
		0.024223,
		0.027933,
		0.031834,
		0.035914,
		0.040162,
		0.044568,
		0.049119,
		0.053808,
		0.058623,
		0.063557,
		0.068601,
		0.073747,
		0.078987,
		0.084314,
		0.08972,
		0.0952,
		0.100747,
		0.106356,
		0.112019,
		0.117733,
		0.123492,
		0.129291,
		0.135126,
		0.140991,
		0.146884,
		0.1528,
		0.158736,
		0.164687,
		0.170651,
		0.176624,
		0.182603,
		0.188586,
		0.19457,
		0.200552,
		0.206529,
		0.2125,
		0.218463,
		0.224414,
		0.230353,
		0.236278,
		0.242186,
		0.248076,
		0.253947,
		0.259797,
		0.265625,
		0.271429,
		0.277209,
		0.282963,
		0.28869,
		0.294389,
		0.300059,
		0.3057,
		0.31131,
		0.316889,
		0.322436,
		0.327951,
		0.333432,
		0.33888,
		0.344293,
		0.349672,
		0.355015,
		0.360324,
		0.365596,
		0.370832,
		0.376031,
		0.381194,
		0.38632,
		0.391409,
		0.39646,
		0.401474,
		0.40645,
		0.411389,
		0.416289,
		0.421152,
		0.425977,
		0.430764,
		0.492972,
		0.488488,
		0.483964,
		0.479399,
		0.474793,
		0.470147,
		0.465459,
		0.460729,
		0.455958,
		0.451146,
		0.446292,
		0.441396,
		0.436458,
		0.431479,
		0.426457,
		0.421394,
		0.416289,
		0.411143,
		0.405954,
		0.400724,
		0.395453,
		0.39014,
		0.384786,
		0.379391,
		0.373956,
		0.36848,
		0.362964,
		0.357409,
		0.351814,
		0.34618,
		0.340507,
		0.334797,
		0.32905,
		0.323265,
		0.317445,
		0.31159,
		0.3057,
		0.299776,
		0.29382,
		0.287832,
		0.281814,
		0.275766,
		0.269691,
		0.263588,
		0.25746,
		0.251308,
		0.245133,
		0.238939,
		0.232725,
		0.226495,
		0.22025,
		0.213992,
		0.207724,
		0.201449,
		0.195168,
		0.188885,
		0.182603,
		0.176325,
		0.170054,
		0.163793,
		0.157547,
		0.151319,
		0.145114,
		0.138935,
		0.132788,
		0.126677,
		0.120607,
		0.114585,
		0.108615,
		0.102704,
		0.096858,
		0.091084,
		0.085389,
		0.079781,
		0.074267,
		0.068856,
		0.063557,
		0.05838,
		0.053333,
		0.048428,
		0.043674,
		0.039085,
		0.034672,
		0.030448,
		0.026425,
		0.02262,
		0.019046,
		0.015719,
		0.012657,
		0.009877,
		0.007397,
		0.005236,
		0.003417,
		0.00196,
		0.000888,
		0.000227,
		0.000227,
		0.000888,
		0.00196,
		0.003417,
		0.005236,
		0.007397,
		0.009877,
		0.012657,
		0.015719,
		0.019046,
		0.02262,
		0.026425,
		0.030448,
		0.034672,
		0.039085,
		0.043674,
		0.048428,
		0.053333,
		0.05838,
		0.063557,
		0.068856,
		0.074267,
		0.079781,
		0.085389,
		0.091084,
		0.096858,
		0.102704,
		0.108615,
		0.114585,
		0.120607,
		0.126677,
		0.132788,
		0.138935,
		0.145114,
		0.151319,
		0.157547,
		0.163793,
		0.170054,
		0.176325,
		0.182603,
		0.188885,
		0.195168,
		0.201449,
		0.207724,
		0.213992,
		0.22025,
		0.226495,
		0.232725,
		0.238939,
		0.245133,
		0.251308,
		0.25746,
		0.263588,
		0.269691,
		0.275766,
		0.281814,
		0.287832,
		0.29382,
		0.299776,
		0.3057,
		0.31159,
		0.317445,
		0.323265,
		0.32905,
		0.334797,
		0.340507,
		0.34618,
		0.351814,
		0.357409,
		0.362964,
		0.36848,
		0.373956,
		0.379391,
		0.384786,
		0.39014,
		0.395453,
		0.400724,
		0.405954,
		0.411143,
		0.416289,
		0.421394,
		0.426457,
		0.431479,
		0.436458,
		0.441396,
		0.446292,
		0.451146,
		0.455958,
		0.460729,
		0.465459,
		0.470147,
		0.474793,
		0.479399,
		0.483964,
		0.488488,
		0.549528,
		0.545361,
		0.541152,
		0.536902,
		0.532609,
		0.528274,
		0.523896,
		0.519475,
		0.515011,
		0.510503,
		0.505952,
		0.501356,
		0.496716,
		0.492031,
		0.487302,
		0.482527,
		0.477707,
		0.472842,
		0.467931,
		0.462975,
		0.457972,
		0.452924,
		0.447829,
		0.442688,
		0.437501,
		0.432268,
		0.426988,
		0.421662,
		0.416289,
		0.410871,
		0.405406,
		0.399895,
		0.394338,
		0.388735,
		0.383087,
		0.377394,
		0.371655,
		0.365872,
		0.360045,
		0.354174,
		0.34826,
		0.342303,
		0.336303,
		0.330263,
		0.324181,
		0.31806,
		0.311899,
		0.3057,
		0.299464,
		0.293191,
		0.286884,
		0.280543,
		0.27417,
		0.267766,
		0.261333,
		0.254872,
		0.248386,
		0.241875,
		0.235343,
		0.228792,
		0.222223,
		0.21564,
		0.209044,
		0.20244,
		0.19583,
		0.189216,
		0.182603,
		0.175995,
		0.169394,
		0.162806,
		0.156234,
		0.149684,
		0.14316,
		0.136666,
		0.13021,
		0.123796,
		0.117431,
		0.111122,
		0.104874,
		0.098696,
		0.092596,
		0.086581,
		0.08066,
		0.074843,
		0.069138,
		0.063557,
		0.058111,
		0.05281,
		0.047667,
		0.042694,
		0.037906,
		0.033317,
		0.028941,
		0.024796,
		0.020897,
		0.017263,
		0.013913,
		0.010867,
		0.008146,
		0.005773,
		0.003771,
		0.002165,
		0.000982,
		0.000251,
		0.000251,
		0.000982,
		0.002165,
		0.003771,
		0.005773,
		0.008146,
		0.010867,
		0.013913,
		0.017263,
		0.020897,
		0.024796,
		0.028941,
		0.033317,
		0.037906,
		0.042694,
		0.047667,
		0.05281,
		0.058111,
		0.063557,
		0.069138,
		0.074843,
		0.08066,
		0.086581,
		0.092596,
		0.098696,
		0.104874,
		0.111122,
		0.117431,
		0.123796,
		0.13021,
		0.136666,
		0.14316,
		0.149684,
		0.156234,
		0.162806,
		0.169394,
		0.175995,
		0.182603,
		0.189216,
		0.19583,
		0.20244,
		0.209044,
		0.21564,
		0.222223,
		0.228792,
		0.235343,
		0.241875,
		0.248386,
		0.254872,
		0.261333,
		0.267766,
		0.27417,
		0.280543,
		0.286884,
		0.293191,
		0.299464,
		0.3057,
		0.311899,
		0.31806,
		0.324181,
		0.330263,
		0.336303,
		0.342303,
		0.34826,
		0.354174,
		0.360045,
		0.365872,
		0.371655,
		0.377394,
		0.383087,
		0.388735,
		0.394338,
		0.399895,
		0.405406,
		0.410871,
		0.416289,
		0.421662,
		0.426988,
		0.432268,
		0.437501,
		0.442688,
		0.447829,
		0.452924,
		0.457972,
		0.462975,
		0.467931,
		0.472842,
		0.477707,
		0.482527,
		0.487302,
		0.492031,
		0.496716,
		0.501356,
		0.505952,
		0.510503,
		0.515011,
		0.519475,
		0.523896,
		0.528274,
		0.532609,
		0.536902,
		0.541152,
		0.545361,
		0.604481,
		0.600669,
		0.596815,
		0.592919,
		0.588982,
		0.585003,
		0.58098,
		0.576915,
		0.572806,
		0.568653,
		0.564456,
		0.560215,
		0.555928,
		0.551596,
		0.547218,
		0.542794,
		0.538323,
		0.533806,
		0.529241,
		0.524629,
		0.519969,
		0.51526,
		0.510503,
		0.505698,
		0.500843,
		0.495938,
		0.490984,
		0.48598,
		0.480925,
		0.475821,
		0.470665,
		0.465459,
		0.460201,
		0.454892,
		0.449532,
		0.444121,
		0.438658,
		0.433143,
		0.427577,
		0.421959,
		0.416289,
		0.410568,
		0.404796,
		0.398972,
		0.393097,
		0.387171,
		0.381194,
		0.375167,
		0.369091,
		0.362964,
		0.356789,
		0.350565,
		0.344293,
		0.337974,
		0.331609,
		0.325198,
		0.318742,
		0.312242,
		0.3057,
		0.299116,
		0.292492,
		0.28583,
		0.27913,
		0.272394,
		0.265625,
		0.258824,
		0.251992,
		0.245133,
		0.238249,
		0.231342,
		0.224414,
		0.21747,
		0.210511,
		0.203541,
		0.196564,
		0.189584,
		0.182603,
		0.175628,
		0.168662,
		0.16171,
		0.154777,
		0.147869,
		0.140991,
		0.134151,
		0.127354,
		0.120607,
		0.113919,
		0.107296,
		0.100747,
		0.094282,
		0.08791,
		0.08164,
		0.075484,
		0.069452,
		0.063557,
		0.057812,
		0.05223,
		0.046826,
		0.041614,
		0.036611,
		0.031834,
		0.027301,
		0.023032,
		0.019046,
		0.015366,
		0.012014,
		0.009016,
		0.006396,
		0.004182,
		0.002404,
		0.001092,
		0.000279,
		0.000279,
		0.001092,
		0.002404,
		0.004182,
		0.006396,
		0.009016,
		0.012014,
		0.015366,
		0.019046,
		0.023032,
		0.027301,
		0.031834,
		0.036611,
		0.041614,
		0.046826,
		0.05223,
		0.057812,
		0.063557,
		0.069452,
		0.075484,
		0.08164,
		0.08791,
		0.094282,
		0.100747,
		0.107296,
		0.113919,
		0.120607,
		0.127354,
		0.134151,
		0.140991,
		0.147869,
		0.154777,
		0.16171,
		0.168662,
		0.175628,
		0.182603,
		0.189584,
		0.196564,
		0.203541,
		0.210511,
		0.21747,
		0.224414,
		0.231342,
		0.238249,
		0.245133,
		0.251992,
		0.258824,
		0.265625,
		0.272394,
		0.27913,
		0.28583,
		0.292492,
		0.299116,
		0.3057,
		0.312242,
		0.318742,
		0.325198,
		0.331609,
		0.337974,
		0.344293,
		0.350565,
		0.356789,
		0.362964,
		0.369091,
		0.375167,
		0.381194,
		0.387171,
		0.393097,
		0.398972,
		0.404796,
		0.410568,
		0.416289,
		0.421959,
		0.427577,
		0.433143,
		0.438658,
		0.444121,
		0.449532,
		0.454892,
		0.460201,
		0.465459,
		0.470665,
		0.475821,
		0.480925,
		0.48598,
		0.490984,
		0.495938,
		0.500843,
		0.505698,
		0.510503,
		0.51526,
		0.519969,
		0.524629,
		0.529241,
		0.533806,
		0.538323,
		0.542794,
		0.547218,
		0.551596,
		0.555928,
		0.560215,
		0.564456,
		0.568653,
		0.572806,
		0.576915,
		0.58098,
		0.585003,
		0.588982,
		0.592919,
		0.596815,
		0.600669,
		0.657196,
		0.653763,
		0.650291,
		0.646778,
		0.643226,
		0.639632,
		0.635997,
		0.63232,
		0.628601,
		0.624839,
		0.621034,
		0.617185,
		0.613292,
		0.609355,
		0.605372,
		0.601344,
		0.59727,
		0.59315,
		0.588982,
		0.584767,
		0.580504,
		0.576193,
		0.571833,
		0.567424,
		0.562964,
		0.558455,
		0.553895,
		0.549284,
		0.544621,
		0.539907,
		0.535139,
		0.53032,
		0.525446,
		0.52052,
		0.515539,
		0.510503,
		0.505413,
		0.500268,
		0.495067,
		0.489811,
		0.484498,
		0.479129,
		0.473704,
		0.468221,
		0.462682,
		0.457085,
		0.45143,
		0.445718,
		0.439948,
		0.43412,
		0.428234,
		0.422291,
		0.416289,
		0.41023,
		0.404113,
		0.397939,
		0.391707,
		0.385418,
		0.379073,
		0.372671,
		0.366214,
		0.359701,
		0.353134,
		0.346512,
		0.339837,
		0.333111,
		0.326332,
		0.319503,
		0.312626,
		0.3057,
		0.298728,
		0.29171,
		0.28465,
		0.277548,
		0.270407,
		0.263228,
		0.256014,
		0.248768,
		0.241492,
		0.234189,
		0.226862,
		0.219514,
		0.212149,
		0.204772,
		0.197385,
		0.189994,
		0.182603,
		0.175218,
		0.167843,
		0.160485,
		0.153149,
		0.145843,
		0.138573,
		0.131346,
		0.124172,
		0.117059,
		0.110014,
		0.10305,
		0.096175,
		0.0894,
		0.082738,
		0.076202,
		0.069803,
		0.063557,
		0.057479,
		0.051585,
		0.045892,
		0.040417,
		0.035182,
		0.030205,
		0.02551,
		0.021119,
		0.017058,
		0.013353,
		0.010032,
		0.007125,
		0.004665,
		0.002685,
		0.001221,
		0.000312,
		0.000312,
		0.001221,
		0.002685,
		0.004665,
		0.007125,
		0.010032,
		0.013353,
		0.017058,
		0.021119,
		0.02551,
		0.030205,
		0.035182,
		0.040417,
		0.045892,
		0.051585,
		0.057479,
		0.063557,
		0.069803,
		0.076202,
		0.082738,
		0.0894,
		0.096175,
		0.10305,
		0.110014,
		0.117059,
		0.124172,
		0.131346,
		0.138573,
		0.145843,
		0.153149,
		0.160485,
		0.167843,
		0.175218,
		0.182603,
		0.189994,
		0.197385,
		0.204772,
		0.212149,
		0.219514,
		0.226862,
		0.234189,
		0.241492,
		0.248768,
		0.256014,
		0.263228,
		0.270407,
		0.277548,
		0.28465,
		0.29171,
		0.298728,
		0.3057,
		0.312626,
		0.319503,
		0.326332,
		0.333111,
		0.339837,
		0.346512,
		0.353134,
		0.359701,
		0.366214,
		0.372671,
		0.379073,
		0.385418,
		0.391707,
		0.397939,
		0.404113,
		0.41023,
		0.416289,
		0.422291,
		0.428234,
		0.43412,
		0.439948,
		0.445718,
		0.45143,
		0.457085,
		0.462682,
		0.468221,
		0.473704,
		0.479129,
		0.484498,
		0.489811,
		0.495067,
		0.500268,
		0.505413,
		0.510503,
		0.515539,
		0.52052,
		0.525446,
		0.53032,
		0.535139,
		0.539907,
		0.544621,
		0.549284,
		0.553895,
		0.558455,
		0.562964,
		0.567424,
		0.571833,
		0.576193,
		0.580504,
		0.584767,
		0.588982,
		0.59315,
		0.59727,
		0.601344,
		0.605372,
		0.609355,
		0.613292,
		0.617185,
		0.621034,
		0.624839,
		0.628601,
		0.63232,
		0.635997,
		0.639632,
		0.643226,
		0.646778,
		0.650291,
		0.653763,
		0.707107,
		0.704067,
		0.700991,
		0.697877,
		0.694725,
		0.691535,
		0.688306,
		0.685037,
		0.681729,
		0.67838,
		0.67499,
		0.671559,
		0.668086,
		0.66457,
		0.661012,
		0.65741,
		0.653763,
		0.650073,
		0.646337,
		0.642555,
		0.638727,
		0.634852,
		0.63093,
		0.62696,
		0.622942,
		0.618874,
		0.614757,
		0.61059,
		0.606372,
		0.602103,
		0.597782,
		0.593409,
		0.588982,
		0.584502,
		0.579968,
		0.575379,
		0.570735,
		0.566035,
		0.561279,
		0.556466,
		0.551596,
		0.546667,
		0.541681,
		0.536635,
		0.531529,
		0.526364,
		0.521138,
		0.515852,
		0.510503,
		0.505093,
		0.499621,
		0.494086,
		0.488488,
		0.482827,
		0.477101,
		0.471312,
		0.465459,
		0.45954,
		0.453557,
		0.447509,
		0.441396,
		0.435217,
		0.428973,
		0.422664,
		0.416289,
		0.409849,
		0.403344,
		0.396775,
		0.39014,
		0.383441,
		0.376679,
		0.369853,
		0.362964,
		0.356013,
		0.349001,
		0.341929,
		0.334797,
		0.327607,
		0.32036,
		0.313057,
		0.3057,
		0.29829,
		0.29083,
		0.283322,
		0.275766,
		0.268167,
		0.260527,
		0.252848,
		0.245133,
		0.237387,
		0.229612,
		0.221812,
		0.213992,
		0.206156,
		0.198309,
		0.190456,
		0.182603,
		0.174757,
		0.166922,
		0.159107,
		0.151319,
		0.143566,
		0.135857,
		0.128201,
		0.120607,
		0.113087,
		0.105651,
		0.098313,
		0.091084,
		0.083978,
		0.077011,
		0.070199,
		0.063557,
		0.057105,
		0.050862,
		0.044848,
		0.039085,
		0.033598,
		0.02841,
		0.02355,
		0.019046,
		0.014928,
		0.011231,
		0.007987,
		0.005236,
		0.003018,
		0.001374,
		0.000352,
		0.000352,
		0.001374,
		0.003018,
		0.005236,
		0.007987,
		0.011231,
		0.014928,
		0.019046,
		0.02355,
		0.02841,
		0.033598,
		0.039085,
		0.044848,
		0.050862,
		0.057105,
		0.063557,
		0.070199,
		0.077011,
		0.083978,
		0.091084,
		0.098313,
		0.105651,
		0.113087,
		0.120607,
		0.128201,
		0.135857,
		0.143566,
		0.151319,
		0.159107,
		0.166922,
		0.174757,
		0.182603,
		0.190456,
		0.198309,
		0.206156,
		0.213992,
		0.221812,
		0.229612,
		0.237387,
		0.245133,
		0.252848,
		0.260527,
		0.268167,
		0.275766,
		0.283322,
		0.29083,
		0.29829,
		0.3057,
		0.313057,
		0.32036,
		0.327607,
		0.334797,
		0.341929,
		0.349001,
		0.356013,
		0.362964,
		0.369853,
		0.376679,
		0.383441,
		0.39014,
		0.396775,
		0.403344,
		0.409849,
		0.416289,
		0.422664,
		0.428973,
		0.435217,
		0.441396,
		0.447509,
		0.453557,
		0.45954,
		0.465459,
		0.471312,
		0.477101,
		0.482827,
		0.488488,
		0.494086,
		0.499621,
		0.505093,
		0.510503,
		0.515852,
		0.521138,
		0.526364,
		0.531529,
		0.536635,
		0.541681,
		0.546667,
		0.551596,
		0.556466,
		0.561279,
		0.566035,
		0.570735,
		0.575379,
		0.579968,
		0.584502,
		0.588982,
		0.593409,
		0.597782,
		0.602103,
		0.606372,
		0.61059,
		0.614757,
		0.618874,
		0.622942,
		0.62696,
		0.63093,
		0.634852,
		0.638727,
		0.642555,
		0.646337,
		0.650073,
		0.653763,
		0.65741,
		0.661012,
		0.66457,
		0.668086,
		0.671559,
		0.67499,
		0.67838,
		0.681729,
		0.685037,
		0.688306,
		0.691535,
		0.694725,
		0.697877,
		0.700991,
		0.704067,
		0.753719,
		0.751077,
		0.7484,
		0.745689,
		0.742944,
		0.740163,
		0.737347,
		0.734494,
		0.731605,
		0.728678,
		0.725714,
		0.722712,
		0.71967,
		0.71659,
		0.713469,
		0.710309,
		0.707107,
		0.703863,
		0.700578,
		0.69725,
		0.693878,
		0.690463,
		0.687003,
		0.683499,
		0.679948,
		0.676351,
		0.672708,
		0.669016,
		0.665277,
		0.661489,
		0.657651,
		0.653763,
		0.649825,
		0.645835,
		0.641793,
		0.637698,
		0.63355,
		0.629348,
		0.625091,
		0.620779,
		0.61641,
		0.611985,
		0.607502,
		0.602961,
		0.598361,
		0.593702,
		0.588982,
		0.584202,
		0.579359,
		0.574455,
		0.569487,
		0.564456,
		0.559361,
		0.554201,
		0.548975,
		0.543682,
		0.538323,
		0.532897,
		0.527402,
		0.521839,
		0.516206,
		0.510503,
		0.504731,
		0.498887,
		0.492972,
		0.486985,
		0.480925,
		0.474793,
		0.468589,
		0.46231,
		0.455958,
		0.449532,
		0.443032,
		0.436458,
		0.429809,
		0.423087,
		0.416289,
		0.409418,
		0.402472,
		0.395453,
		0.38836,
		0.381194,
		0.373956,
		0.366646,
		0.359265,
		0.351814,
		0.344293,
		0.336705,
		0.32905,
		0.321329,
		0.313545,
		0.3057,
		0.297795,
		0.289832,
		0.281814,
		0.273744,
		0.265625,
		0.25746,
		0.249252,
		0.241006,
		0.232725,
		0.224414,
		0.216079,
		0.207724,
		0.199356,
		0.19098,
		0.182603,
		0.174234,
		0.165879,
		0.157547,
		0.149248,
		0.140991,
		0.132788,
		0.124649,
		0.116587,
		0.108615,
		0.100747,
		0.093,
		0.085389,
		0.077932,
		0.070648,
		0.063557,
		0.056682,
		0.050046,
		0.043674,
		0.037594,
		0.031834,
		0.026425,
		0.021402,
		0.0168,
		0.012657,
		0.009016,
		0.00592,
		0.003417,
		0.001559,
		0.0004,
		0.0004,
		0.001559,
		0.003417,
		0.00592,
		0.009016,
		0.012657,
		0.0168,
		0.021402,
		0.026425,
		0.031834,
		0.037594,
		0.043674,
		0.050046,
		0.056682,
		0.063557,
		0.070648,
		0.077932,
		0.085389,
		0.093,
		0.100747,
		0.108615,
		0.116587,
		0.124649,
		0.132788,
		0.140991,
		0.149248,
		0.157547,
		0.165879,
		0.174234,
		0.182603,
		0.19098,
		0.199356,
		0.207724,
		0.216079,
		0.224414,
		0.232725,
		0.241006,
		0.249252,
		0.25746,
		0.265625,
		0.273744,
		0.281814,
		0.289832,
		0.297795,
		0.3057,
		0.313545,
		0.321329,
		0.32905,
		0.336705,
		0.344293,
		0.351814,
		0.359265,
		0.366646,
		0.373956,
		0.381194,
		0.38836,
		0.395453,
		0.402472,
		0.409418,
		0.416289,
		0.423087,
		0.429809,
		0.436458,
		0.443032,
		0.449532,
		0.455958,
		0.46231,
		0.468589,
		0.474793,
		0.480925,
		0.486985,
		0.492972,
		0.498887,
		0.504731,
		0.510503,
		0.516206,
		0.521839,
		0.527402,
		0.532897,
		0.538323,
		0.543682,
		0.548975,
		0.554201,
		0.559361,
		0.564456,
		0.569487,
		0.574455,
		0.579359,
		0.584202,
		0.588982,
		0.593702,
		0.598361,
		0.602961,
		0.607502,
		0.611985,
		0.61641,
		0.620779,
		0.625091,
		0.629348,
		0.63355,
		0.637698,
		0.641793,
		0.645835,
		0.649825,
		0.653763,
		0.657651,
		0.661489,
		0.665277,
		0.669016,
		0.672708,
		0.676351,
		0.679948,
		0.683499,
		0.687003,
		0.690463,
		0.693878,
		0.69725,
		0.700578,
		0.703863,
		0.707107,
		0.710309,
		0.713469,
		0.71659,
		0.71967,
		0.722712,
		0.725714,
		0.728678,
		0.731605,
		0.734494,
		0.737347,
		0.740163,
		0.742944,
		0.745689,
		0.7484,
		0.751077,
		0.796619,
		0.794366,
		0.792083,
		0.78977,
		0.787425,
		0.785049,
		0.782641,
		0.780201,
		0.777728,
		0.775222,
		0.772681,
		0.770107,
		0.767497,
		0.764852,
		0.762171,
		0.759454,
		0.756699,
		0.753907,
		0.751077,
		0.748208,
		0.745299,
		0.742351,
		0.739362,
		0.736332,
		0.73326,
		0.730146,
		0.726989,
		0.723788,
		0.720543,
		0.717253,
		0.713918,
		0.710536,
		0.707107,
		0.70363,
		0.700105,
		0.696531,
		0.692907,
		0.689233,
		0.685507,
		0.681729,
		0.677898,
		0.674014,
		0.670076,
		0.666082,
		0.662033,
		0.657927,
		0.653763,
		0.649542,
		0.645261,
		0.64092,
		0.636519,
		0.632056,
		0.62753,
		0.622942,
		0.618289,
		0.613572,
		0.608789,
		0.603939,
		0.599022,
		0.594037,
		0.588982,
		0.583858,
		0.578663,
		0.573396,
		0.568056,
		0.562644,
		0.557157,
		0.551596,
		0.545959,
		0.540245,
		0.534454,
		0.528585,
		0.522638,
		0.516611,
		0.510503,
		0.504315,
		0.498046,
		0.491695,
		0.485261,
		0.478744,
		0.472143,
		0.465459,
		0.45869,
		0.451836,
		0.444897,
		0.437873,
		0.430764,
		0.423569,
		0.416289,
		0.408924,
		0.401474,
		0.393939,
		0.38632,
		0.378617,
		0.370832,
		0.362964,
		0.355015,
		0.346987,
		0.33888,
		0.330696,
		0.322436,
		0.314104,
		0.3057,
		0.297228,
		0.28869,
		0.280089,
		0.271429,
		0.262714,
		0.253947,
		0.245133,
		0.236278,
		0.227386,
		0.218463,
		0.209516,
		0.200552,
		0.191578,
		0.182603,
		0.173636,
		0.164687,
		0.155766,
		0.146884,
		0.138055,
		0.129291,
		0.120607,
		0.112019,
		0.103544,
		0.0952,
		0.087007,
		0.078987,
		0.071162,
		0.063557,
		0.0562,
		0.049119,
		0.042346,
		0.035914,
		0.02986,
		0.024222,
		0.019044,
		0.01437,
		0.010251,
		0.00674,
		0.003895,
		0.001778,
		0.000457,
		0.000457,
		0.001778,
		0.003895,
		0.00674,
		0.010251,
		0.01437,
		0.019044,
		0.024222,
		0.02986,
		0.035914,
		0.042346,
		0.049119,
		0.0562,
		0.063557,
		0.071162,
		0.078987,
		0.087007,
		0.0952,
		0.103544,
		0.112019,
		0.120607,
		0.129291,
		0.138055,
		0.146884,
		0.155766,
		0.164687,
		0.173636,
		0.182603,
		0.191578,
		0.200552,
		0.209516,
		0.218463,
		0.227386,
		0.236278,
		0.245133,
		0.253947,
		0.262714,
		0.271429,
		0.280089,
		0.28869,
		0.297228,
		0.3057,
		0.314104,
		0.322436,
		0.330696,
		0.33888,
		0.346987,
		0.355015,
		0.362964,
		0.370832,
		0.378617,
		0.38632,
		0.393939,
		0.401474,
		0.408924,
		0.416289,
		0.423569,
		0.430764,
		0.437873,
		0.444897,
		0.451836,
		0.45869,
		0.465459,
		0.472143,
		0.478744,
		0.485261,
		0.491695,
		0.498046,
		0.504315,
		0.510503,
		0.516611,
		0.522638,
		0.528585,
		0.534454,
		0.540245,
		0.545959,
		0.551596,
		0.557157,
		0.562644,
		0.568056,
		0.573396,
		0.578663,
		0.583858,
		0.588982,
		0.594037,
		0.599022,
		0.603939,
		0.608789,
		0.613572,
		0.618289,
		0.622942,
		0.62753,
		0.632056,
		0.636519,
		0.64092,
		0.645261,
		0.649542,
		0.653763,
		0.657927,
		0.662033,
		0.666082,
		0.670076,
		0.674014,
		0.677898,
		0.681729,
		0.685507,
		0.689233,
		0.692907,
		0.696531,
		0.700105,
		0.70363,
		0.707107,
		0.710536,
		0.713918,
		0.717253,
		0.720543,
		0.723788,
		0.726989,
		0.730146,
		0.73326,
		0.736332,
		0.739362,
		0.742351,
		0.745299,
		0.748208,
		0.751077,
		0.753907,
		0.756699,
		0.759454,
		0.762171,
		0.764852,
		0.767497,
		0.770107,
		0.772681,
		0.775222,
		0.777728,
		0.780201,
		0.782641,
		0.785049,
		0.787425,
		0.78977,
		0.792083,
		0.794366,
		0.835475,
		0.833597,
		0.831692,
		0.82976,
		0.827802,
		0.825816,
		0.823803,
		0.821761,
		0.819691,
		0.817592,
		0.815463,
		0.813304,
		0.811114,
		0.808893,
		0.806641,
		0.804357,
		0.80204,
		0.79969,
		0.797307,
		0.794889,
		0.792437,
		0.789949,
		0.787425,
		0.784865,
		0.782268,
		0.779633,
		0.77696,
		0.774249,
		0.771497,
		0.768706,
		0.765874,
		0.763,
		0.760084,
		0.757125,
		0.754123,
		0.751077,
		0.747985,
		0.744848,
		0.741665,
		0.738434,
		0.735156,
		0.731828,
		0.728452,
		0.725025,
		0.721547,
		0.718017,
		0.714434,
		0.710798,
		0.707107,
		0.703361,
		0.699558,
		0.695699,
		0.691782,
		0.687806,
		0.68377,
		0.679673,
		0.675515,
		0.671294,
		0.667009,
		0.66266,
		0.658245,
		0.653763,
		0.649214,
		0.644597,
		0.63991,
		0.635152,
		0.630323,
		0.62542,
		0.620445,
		0.615394,
		0.610268,
		0.605064,
		0.599783,
		0.594422,
		0.588982,
		0.583461,
		0.577857,
		0.57217,
		0.566399,
		0.560543,
		0.5546,
		0.54857,
		0.542452,
		0.536244,
		0.529947,
		0.523558,
		0.517077,
		0.510503,
		0.503836,
		0.497074,
		0.490217,
		0.483264,
		0.476215,
		0.469068,
		0.461824,
		0.454482,
		0.447041,
		0.439502,
		0.431863,
		0.424126,
		0.416289,
		0.408354,
		0.40032,
		0.392188,
		0.383959,
		0.375633,
		0.367211,
		0.358694,
		0.350084,
		0.341382,
		0.332591,
		0.323712,
		0.314747,
		0.3057,
		0.296573,
		0.287371,
		0.278096,
		0.268753,
		0.259348,
		0.249885,
		0.24037,
		0.23081,
		0.221211,
		0.211582,
		0.201932,
		0.192269,
		0.182603,
		0.172947,
		0.163312,
		0.153712,
		0.144161,
		0.134676,
		0.125272,
		0.11597,
		0.106789,
		0.097751,
		0.088879,
		0.0802,
		0.071739,
		0.063527,
		0.055596,
		0.047983,
		0.040727,
		0.033873,
		0.027471,
		0.021576,
		0.016249,
		0.011555,
		0.007563,
		0.004344,
		0.001968,
		0.0005,
		0.0005,
		0.001968,
		0.004344,
		0.007563,
		0.011555,
		0.016249,
		0.021576,
		0.027471,
		0.033873,
		0.040727,
		0.047983,
		0.055596,
		0.063527,
		0.071739,
		0.0802,
		0.088879,
		0.097751,
		0.106789,
		0.11597,
		0.125272,
		0.134676,
		0.144161,
		0.153712,
		0.163312,
		0.172947,
		0.182603,
		0.192269,
		0.201932,
		0.211582,
		0.221211,
		0.23081,
		0.24037,
		0.249885,
		0.259348,
		0.268753,
		0.278096,
		0.287371,
		0.296573,
		0.3057,
		0.314747,
		0.323712,
		0.332591,
		0.341382,
		0.350084,
		0.358694,
		0.367211,
		0.375633,
		0.383959,
		0.392188,
		0.40032,
		0.408354,
		0.416289,
		0.424126,
		0.431863,
		0.439502,
		0.447041,
		0.454482,
		0.461824,
		0.469068,
		0.476215,
		0.483264,
		0.490217,
		0.497074,
		0.503836,
		0.510503,
		0.517077,
		0.523558,
		0.529947,
		0.536244,
		0.542452,
		0.54857,
		0.5546,
		0.560543,
		0.566399,
		0.57217,
		0.577857,
		0.583461,
		0.588982,
		0.594422,
		0.599783,
		0.605064,
		0.610268,
		0.615394,
		0.620445,
		0.62542,
		0.630323,
		0.635152,
		0.63991,
		0.644597,
		0.649214,
		0.653763,
		0.658245,
		0.66266,
		0.667009,
		0.671294,
		0.675515,
		0.679673,
		0.68377,
		0.687806,
		0.691782,
		0.695699,
		0.699558,
		0.703361,
		0.707107,
		0.710798,
		0.714434,
		0.718017,
		0.721547,
		0.725025,
		0.728452,
		0.731828,
		0.735156,
		0.738434,
		0.741665,
		0.744848,
		0.747985,
		0.751077,
		0.754123,
		0.757125,
		0.760084,
		0.763,
		0.765874,
		0.768706,
		0.771497,
		0.774249,
		0.77696,
		0.779633,
		0.782268,
		0.784865,
		0.787425,
		0.789949,
		0.792437,
		0.794889,
		0.797307,
		0.79969,
		0.80204,
		0.804357,
		0.806641,
		0.808893,
		0.811114,
		0.813304,
		0.815463,
		0.817592,
		0.819691,
		0.821761,
		0.823803,
		0.825816,
		0.827802,
		0.82976,
		0.831692,
		0.833597,
		0.870045,
		0.868517,
		0.866967,
		0.865395,
		0.863799,
		0.862181,
		0.860539,
		0.858873,
		0.857183,
		0.855469,
		0.853729,
		0.851964,
		0.850172,
		0.848355,
		0.84651,
		0.844638,
		0.842738,
		0.84081,
		0.838854,
		0.836868,
		0.834852,
		0.832806,
		0.830729,
		0.828621,
		0.826481,
		0.824309,
		0.822103,
		0.819865,
		0.817592,
		0.815284,
		0.812941,
		0.810562,
		0.808146,
		0.805693,
		0.803203,
		0.800673,
		0.798105,
		0.795497,
		0.792848,
		0.790157,
		0.787425,
		0.78465,
		0.781832,
		0.778969,
		0.776061,
		0.773107,
		0.770107,
		0.767059,
		0.763962,
		0.760817,
		0.757621,
		0.754375,
		0.751077,
		0.747726,
		0.744321,
		0.740862,
		0.737347,
		0.733775,
		0.730146,
		0.726459,
		0.722712,
		0.718904,
		0.715035,
		0.711103,
		0.707107,
		0.703046,
		0.698919,
		0.694725,
		0.690463,
		0.686131,
		0.681729,
		0.677255,
		0.672708,
		0.668086,
		0.663389,
		0.658615,
		0.653763,
		0.648832,
		0.643821,
		0.638727,
		0.63355,
		0.628289,
		0.622942,
		0.617508,
		0.611985,
		0.606372,
		0.600669,
		0.594872,
		0.588982,
		0.582997,
		0.576915,
		0.570735,
		0.564456,
		0.558077,
		0.551596,
		0.545012,
		0.538323,
		0.531529,
		0.524629,
		0.517621,
		0.510503,
		0.503276,
		0.495938,
		0.488488,
		0.480925,
		0.473249,
		0.465459,
		0.457553,
		0.449532,
		0.441396,
		0.433143,
		0.424774,
		0.416289,
		0.407688,
		0.398972,
		0.39014,
		0.381194,
		0.372135,
		0.362964,
		0.353683,
		0.344293,
		0.334797,
		0.325198,
		0.315497,
		0.3057,
		0.295809,
		0.28583,
		0.275766,
		0.265625,
		0.255412,
		0.245133,
		0.234798,
		0.224414,
		0.213992,
		0.20354,
		0.193071,
		0.182595,
		0.172124,
		0.16167,
		0.151247,
		0.140868,
		0.130547,
		0.120303,
		0.110155,
		0.100129,
		0.090256,
		0.080573,
		0.071125,
		0.061965,
		0.053154,
		0.044763,
		0.036866,
		0.029542,
		0.022875,
		0.016943,
		0.01182,
		0.007571,
		0.004244,
		0.001872,
		0.000462,
		0.000462,
		0.001872,
		0.004244,
		0.007571,
		0.01182,
		0.016943,
		0.022875,
		0.029542,
		0.036866,
		0.044763,
		0.053154,
		0.061965,
		0.071125,
		0.080573,
		0.090256,
		0.100129,
		0.110155,
		0.120303,
		0.130547,
		0.140868,
		0.151247,
		0.16167,
		0.172124,
		0.182595,
		0.193071,
		0.20354,
		0.213992,
		0.224414,
		0.234798,
		0.245133,
		0.255412,
		0.265625,
		0.275766,
		0.28583,
		0.295809,
		0.3057,
		0.315497,
		0.325198,
		0.334797,
		0.344293,
		0.353683,
		0.362964,
		0.372135,
		0.381194,
		0.39014,
		0.398972,
		0.407688,
		0.416289,
		0.424774,
		0.433143,
		0.441396,
		0.449532,
		0.457553,
		0.465459,
		0.473249,
		0.480925,
		0.488488,
		0.495938,
		0.503276,
		0.510503,
		0.517621,
		0.524629,
		0.531529,
		0.538323,
		0.545012,
		0.551596,
		0.558077,
		0.564456,
		0.570735,
		0.576915,
		0.582997,
		0.588982,
		0.594872,
		0.600669,
		0.606372,
		0.611985,
		0.617508,
		0.622942,
		0.628289,
		0.63355,
		0.638727,
		0.643821,
		0.648832,
		0.653763,
		0.658615,
		0.663389,
		0.668086,
		0.672708,
		0.677255,
		0.681729,
		0.686131,
		0.690463,
		0.694725,
		0.698919,
		0.703046,
		0.707107,
		0.711103,
		0.715035,
		0.718904,
		0.722712,
		0.726459,
		0.730146,
		0.733775,
		0.737347,
		0.740862,
		0.744321,
		0.747726,
		0.751077,
		0.754375,
		0.757621,
		0.760817,
		0.763962,
		0.767059,
		0.770107,
		0.773107,
		0.776061,
		0.778969,
		0.781832,
		0.78465,
		0.787425,
		0.790157,
		0.792848,
		0.795497,
		0.798105,
		0.800673,
		0.803203,
		0.805693,
		0.808146,
		0.810562,
		0.812941,
		0.815284,
		0.817592,
		0.819865,
		0.822103,
		0.824309,
		0.826481,
		0.828621,
		0.830729,
		0.832806,
		0.834852,
		0.836868,
		0.838854,
		0.84081,
		0.842738,
		0.844638,
		0.84651,
		0.848355,
		0.850172,
		0.851964,
		0.853729,
		0.855469,
		0.857183,
		0.858873,
		0.860539,
		0.862181,
		0.863799,
		0.865395,
		0.866967,
		0.868517,
		0.900183,
		0.898975,
		0.897749,
		0.896506,
		0.895243,
		0.893962,
		0.892661,
		0.891341,
		0.890001,
		0.888641,
		0.88726,
		0.885858,
		0.884434,
		0.882989,
		0.881522,
		0.880032,
		0.878519,
		0.876983,
		0.875423,
		0.873839,
		0.87223,
		0.870596,
		0.868936,
		0.867251,
		0.865538,
		0.863799,
		0.862033,
		0.860238,
		0.858415,
		0.856563,
		0.854681,
		0.852769,
		0.850827,
		0.848853,
		0.846847,
		0.844809,
		0.842738,
		0.840634,
		0.838495,
		0.836321,
		0.834112,
		0.831866,
		0.829583,
		0.827263,
		0.824905,
		0.822507,
		0.820069,
		0.817592,
		0.815072,
		0.812511,
		0.809907,
		0.807258,
		0.804566,
		0.801828,
		0.799043,
		0.796212,
		0.793332,
		0.790404,
		0.787425,
		0.784396,
		0.781314,
		0.77818,
		0.774992,
		0.771749,
		0.76845,
		0.765094,
		0.76168,
		0.758206,
		0.754672,
		0.751077,
		0.747418,
		0.743696,
		0.739908,
		0.736055,
		0.732133,
		0.728142,
		0.724081,
		0.719949,
		0.715743,
		0.711463,
		0.707107,
		0.702674,
		0.698162,
		0.69357,
		0.688896,
		0.684139,
		0.679298,
		0.67437,
		0.669354,
		0.664249,
		0.659052,
		0.653763,
		0.64838,
		0.642901,
		0.637323,
		0.631647,
		0.625869,
		0.619989,
		0.614003,
		0.607912,
		0.601712,
		0.595403,
		0.588982,
		0.582448,
		0.575799,
		0.569033,
		0.562148,
		0.555144,
		0.548017,
		0.540768,
		0.533393,
		0.525892,
		0.518262,
		0.510503,
		0.502614,
		0.494592,
		0.486437,
		0.478147,
		0.469722,
		0.461161,
		0.452462,
		0.443626,
		0.434652,
		0.42554,
		0.416289,
		0.406901,
		0.397375,
		0.387712,
		0.377913,
		0.36798,
		0.357915,
		0.34772,
		0.337397,
		0.32695,
		0.316381,
		0.305694,
		0.294891,
		0.283974,
		0.272942,
		0.261794,
		0.250527,
		0.239137,
		0.227619,
		0.215968,
		0.204182,
		0.192263,
		0.180217,
		0.168059,
		0.155812,
		0.143513,
		0.131208,
		0.118961,
		0.106843,
		0.094942,
		0.083354,
		0.072182,
		0.061533,
		0.051515,
		0.042229,
		0.033766,
		0.026203,
		0.019598,
		0.013986,
		0.009378,
		0.005759,
		0.003088,
		0.001299,
		0.000305,
		0.000305,
		0.001299,
		0.003088,
		0.005759,
		0.009378,
		0.013986,
		0.019598,
		0.026203,
		0.033766,
		0.042229,
		0.051515,
		0.061533,
		0.072182,
		0.083354,
		0.094942,
		0.106843,
		0.118961,
		0.131208,
		0.143513,
		0.155812,
		0.168059,
		0.180217,
		0.192263,
		0.204182,
		0.215968,
		0.227619,
		0.239137,
		0.250527,
		0.261794,
		0.272942,
		0.283974,
		0.294891,
		0.305694,
		0.316381,
		0.32695,
		0.337397,
		0.34772,
		0.357915,
		0.36798,
		0.377913,
		0.387712,
		0.397375,
		0.406901,
		0.416289,
		0.42554,
		0.434652,
		0.443626,
		0.452462,
		0.461161,
		0.469722,
		0.478147,
		0.486437,
		0.494592,
		0.502614,
		0.510503,
		0.518262,
		0.525892,
		0.533393,
		0.540768,
		0.548017,
		0.555144,
		0.562148,
		0.569033,
		0.575799,
		0.582448,
		0.588982,
		0.595403,
		0.601712,
		0.607912,
		0.614003,
		0.619989,
		0.625869,
		0.631647,
		0.637323,
		0.642901,
		0.64838,
		0.653763,
		0.659052,
		0.664249,
		0.669354,
		0.67437,
		0.679298,
		0.684139,
		0.688896,
		0.69357,
		0.698162,
		0.702674,
		0.707107,
		0.711463,
		0.715743,
		0.719949,
		0.724081,
		0.728142,
		0.732133,
		0.736055,
		0.739908,
		0.743696,
		0.747418,
		0.751077,
		0.754672,
		0.758206,
		0.76168,
		0.765094,
		0.76845,
		0.771749,
		0.774992,
		0.77818,
		0.781314,
		0.784396,
		0.787425,
		0.790404,
		0.793332,
		0.796212,
		0.799043,
		0.801828,
		0.804566,
		0.807258,
		0.809907,
		0.812511,
		0.815072,
		0.817592,
		0.820069,
		0.822507,
		0.824905,
		0.827263,
		0.829583,
		0.831866,
		0.834112,
		0.836321,
		0.838495,
		0.840634,
		0.842738,
		0.844809,
		0.846847,
		0.848853,
		0.850827,
		0.852769,
		0.854681,
		0.856563,
		0.858415,
		0.860238,
		0.862033,
		0.863799,
		0.865538,
		0.867251,
		0.868936,
		0.870596,
		0.87223,
		0.873839,
		0.875423,
		0.876983,
		0.878519,
		0.880032,
		0.881522,
		0.882989,
		0.884434,
		0.885858,
		0.88726,
		0.888641,
		0.890001,
		0.891341,
		0.892661,
		0.893962,
		0.895243,
		0.896506,
		0.897749,
		0.898975,
		0.925841,
		0.924919,
		0.923982,
		0.923032,
		0.922066,
		0.921086,
		0.92009,
		0.919079,
		0.918052,
		0.917009,
		0.91595,
		0.914874,
		0.913781,
		0.91267,
		0.911542,
		0.910396,
		0.909232,
		0.908049,
		0.906848,
		0.905627,
		0.904386,
		0.903125,
		0.901844,
		0.900541,
		0.899218,
		0.897873,
		0.896506,
		0.895116,
		0.893703,
		0.892267,
		0.890808,
		0.889324,
		0.887815,
		0.886281,
		0.884721,
		0.883135,
		0.881522,
		0.879882,
		0.878214,
		0.876517,
		0.874792,
		0.873037,
		0.871252,
		0.869437,
		0.86759,
		0.865711,
		0.863799,
		0.861854,
		0.859876,
		0.857862,
		0.855814,
		0.853729,
		0.851607,
		0.849448,
		0.847251,
		0.845015,
		0.842738,
		0.840421,
		0.838063,
		0.835662,
		0.833218,
		0.830729,
		0.828196,
		0.825616,
		0.82299,
		0.820315,
		0.817592,
		0.814818,
		0.811993,
		0.809117,
		0.806187,
		0.803203,
		0.800163,
		0.797066,
		0.793912,
		0.790699,
		0.787425,
		0.78409,
		0.780692,
		0.77723,
		0.773702,
		0.770107,
		0.766443,
		0.76271,
		0.758906,
		0.755028,
		0.751077,
		0.747049,
		0.742944,
		0.738759,
		0.734494,
		0.730146,
		0.725714,
		0.721196,
		0.71659,
		0.711894,
		0.707107,
		0.702226,
		0.69725,
		0.692176,
		0.687003,
		0.681729,
		0.676351,
		0.670868,
		0.665277,
		0.659576,
		0.653763,
		0.647836,
		0.641793,
		0.635631,
		0.629348,
		0.622942,
		0.61641,
		0.609751,
		0.602961,
		0.596039,
		0.588982,
		0.581788,
		0.574455,
		0.56698,
		0.559361,
		0.551596,
		0.543682,
		0.535618,
		0.527402,
		0.519031,
		0.510503,
		0.501818,
		0.492972,
		0.483964,
		0.474793,
		0.465459,
		0.455958,
		0.446291,
		0.436456,
		0.42645,
		0.416269,
		0.405908,
		0.395357,
		0.384603,
		0.373628,
		0.362411,
		0.350925,
		0.339142,
		0.327031,
		0.314561,
		0.301705,
		0.28844,
		0.274752,
		0.260641,
		0.246117,
		0.231212,
		0.215973,
		0.20047,
		0.184789,
		0.169039,
		0.153342,
		0.137834,
		0.122659,
		0.107965,
		0.093895,
		0.080587,
		0.068164,
		0.056729,
		0.046363,
		0.037121,
		0.029031,
		0.02209,
		0.01627,
		0.011518,
		0.007757,
		0.004893,
		0.002816,
		0.001409,
		0.000551,
		0.000119,
		0.000119,
		0.000551,
		0.001409,
		0.002816,
		0.004893,
		0.007757,
		0.011518,
		0.01627,
		0.02209,
		0.029031,
		0.037121,
		0.046363,
		0.056729,
		0.068164,
		0.080587,
		0.093895,
		0.107965,
		0.122659,
		0.137834,
		0.153342,
		0.169039,
		0.184789,
		0.20047,
		0.215973,
		0.231212,
		0.246117,
		0.260641,
		0.274752,
		0.28844,
		0.301705,
		0.314561,
		0.327031,
		0.339142,
		0.350925,
		0.362411,
		0.373628,
		0.384603,
		0.395357,
		0.405908,
		0.416269,
		0.42645,
		0.436456,
		0.446291,
		0.455958,
		0.465459,
		0.474793,
		0.483964,
		0.492972,
		0.501818,
		0.510503,
		0.519031,
		0.527402,
		0.535618,
		0.543682,
		0.551596,
		0.559361,
		0.56698,
		0.574455,
		0.581788,
		0.588982,
		0.596039,
		0.602961,
		0.609751,
		0.61641,
		0.622942,
		0.629348,
		0.635631,
		0.641793,
		0.647836,
		0.653763,
		0.659576,
		0.665277,
		0.670868,
		0.676351,
		0.681729,
		0.687003,
		0.692176,
		0.69725,
		0.702226,
		0.707107,
		0.711894,
		0.71659,
		0.721196,
		0.725714,
		0.730146,
		0.734494,
		0.738759,
		0.742944,
		0.747049,
		0.751077,
		0.755028,
		0.758906,
		0.76271,
		0.766443,
		0.770107,
		0.773702,
		0.77723,
		0.780692,
		0.78409,
		0.787425,
		0.790699,
		0.793912,
		0.797066,
		0.800163,
		0.803203,
		0.806187,
		0.809117,
		0.811993,
		0.814818,
		0.817592,
		0.820315,
		0.82299,
		0.825616,
		0.828196,
		0.830729,
		0.833218,
		0.835662,
		0.838063,
		0.840421,
		0.842738,
		0.845015,
		0.847251,
		0.849448,
		0.851607,
		0.853729,
		0.855814,
		0.857862,
		0.859876,
		0.861854,
		0.863799,
		0.865711,
		0.86759,
		0.869437,
		0.871252,
		0.873037,
		0.874792,
		0.876517,
		0.878214,
		0.879882,
		0.881522,
		0.883135,
		0.884721,
		0.886281,
		0.887815,
		0.889324,
		0.890808,
		0.892267,
		0.893703,
		0.895116,
		0.896506,
		0.897873,
		0.899218,
		0.900541,
		0.901844,
		0.903125,
		0.904386,
		0.905627,
		0.906848,
		0.908049,
		0.909232,
		0.910396,
		0.911542,
		0.91267,
		0.913781,
		0.914874,
		0.91595,
		0.917009,
		0.918052,
		0.919079,
		0.92009,
		0.921086,
		0.922066,
		0.923032,
		0.923982,
		0.924919,
		0.947082,
		0.946406,
		0.945719,
		0.945021,
		0.944313,
		0.943593,
		0.942861,
		0.942118,
		0.941362,
		0.940595,
		0.939815,
		0.939022,
		0.938217,
		0.937398,
		0.936566,
		0.93572,
		0.93486,
		0.933986,
		0.933098,
		0.932195,
		0.931276,
		0.930343,
		0.929393,
		0.928428,
		0.927446,
		0.926448,
		0.925433,
		0.9244,
		0.92335,
		0.922282,
		0.921195,
		0.92009,
		0.918966,
		0.917822,
		0.916658,
		0.915473,
		0.914268,
		0.913042,
		0.911794,
		0.910525,
		0.909232,
		0.907917,
		0.906578,
		0.905215,
		0.903828,
		0.902416,
		0.900978,
		0.899514,
		0.898023,
		0.896506,
		0.89496,
		0.893386,
		0.891784,
		0.890151,
		0.888488,
		0.886795,
		0.88507,
		0.883312,
		0.881522,
		0.879698,
		0.877839,
		0.875946,
		0.874016,
		0.872049,
		0.870045,
		0.868003,
		0.865921,
		0.863799,
		0.861636,
		0.859431,
		0.857183,
		0.854892,
		0.852555,
		0.850172,
		0.847743,
		0.845265,
		0.842738,
		0.840161,
		0.837533,
		0.834852,
		0.832117,
		0.829327,
		0.826481,
		0.823577,
		0.820615,
		0.817592,
		0.814507,
		0.811359,
		0.808146,
		0.804867,
		0.801521,
		0.798105,
		0.794618,
		0.791059,
		0.787425,
		0.783716,
		0.779928,
		0.776061,
		0.772112,
		0.76808,
		0.763962,
		0.759757,
		0.755463,
		0.751077,
		0.746597,
		0.742021,
		0.737347,
		0.732572,
		0.727694,
		0.722712,
		0.717621,
		0.71242,
		0.707107,
		0.701678,
		0.696131,
		0.690463,
		0.684672,
		0.678754,
		0.672708,
		0.666529,
		0.660215,
		0.653763,
		0.647171,
		0.640434,
		0.63355,
		0.626516,
		0.619329,
		0.611985,
		0.604481,
		0.596815,
		0.588982,
		0.58098,
		0.572806,
		0.564456,
		0.555925,
		0.547208,
		0.538297,
		0.529181,
		0.519843,
		0.51026,
		0.500403,
		0.490235,
		0.479711,
		0.468778,
		0.457378,
		0.445448,
		0.432925,
		0.419747,
		0.405858,
		0.391212,
		0.375779,
		0.359546,
		0.342523,
		0.324747,
		0.30628,
		0.287213,
		0.267665,
		0.247779,
		0.227721,
		0.207672,
		0.187825,
		0.168377,
		0.14952,
		0.131439,
		0.1143,
		0.09825,
		0.083406,
		0.069857,
		0.057659,
		0.046838,
		0.037384,
		0.029263,
		0.022412,
		0.016747,
		0.012164,
		0.008548,
		0.005777,
		0.003723,
		0.002261,
		0.001272,
		0.000644,
		0.000278,
		9.2e-005,
		1.6e-005,
		1.6e-005,
		9.2e-005,
		0.000278,
		0.000644,
		0.001272,
		0.002261,
		0.003723,
		0.005777,
		0.008548,
		0.012164,
		0.016747,
		0.022412,
		0.029263,
		0.037384,
		0.046838,
		0.057659,
		0.069857,
		0.083406,
		0.09825,
		0.1143,
		0.131439,
		0.14952,
		0.168377,
		0.187825,
		0.207672,
		0.227721,
		0.247779,
		0.267665,
		0.287213,
		0.30628,
		0.324747,
		0.342523,
		0.359546,
		0.375779,
		0.391212,
		0.405858,
		0.419747,
		0.432925,
		0.445448,
		0.457378,
		0.468778,
		0.479711,
		0.490235,
		0.500403,
		0.51026,
		0.519843,
		0.529181,
		0.538297,
		0.547208,
		0.555925,
		0.564456,
		0.572806,
		0.58098,
		0.588982,
		0.596815,
		0.604481,
		0.611985,
		0.619329,
		0.626516,
		0.63355,
		0.640434,
		0.647171,
		0.653763,
		0.660215,
		0.666529,
		0.672708,
		0.678754,
		0.684672,
		0.690463,
		0.696131,
		0.701678,
		0.707107,
		0.71242,
		0.717621,
		0.722712,
		0.727694,
		0.732572,
		0.737347,
		0.742021,
		0.746597,
		0.751077,
		0.755463,
		0.759757,
		0.763962,
		0.76808,
		0.772112,
		0.776061,
		0.779928,
		0.783716,
		0.787425,
		0.791059,
		0.794618,
		0.798105,
		0.801521,
		0.804867,
		0.808146,
		0.811359,
		0.814507,
		0.817592,
		0.820615,
		0.823577,
		0.826481,
		0.829327,
		0.832117,
		0.834852,
		0.837533,
		0.840161,
		0.842738,
		0.845265,
		0.847743,
		0.850172,
		0.852555,
		0.854892,
		0.857183,
		0.859431,
		0.861636,
		0.863799,
		0.865921,
		0.868003,
		0.870045,
		0.872049,
		0.874016,
		0.875946,
		0.877839,
		0.879698,
		0.881522,
		0.883312,
		0.88507,
		0.886795,
		0.888488,
		0.890151,
		0.891784,
		0.893386,
		0.89496,
		0.896506,
		0.898023,
		0.899514,
		0.900978,
		0.902416,
		0.903828,
		0.905215,
		0.906578,
		0.907917,
		0.909232,
		0.910525,
		0.911794,
		0.913042,
		0.914268,
		0.915473,
		0.916658,
		0.917822,
		0.918966,
		0.92009,
		0.921195,
		0.922282,
		0.92335,
		0.9244,
		0.925433,
		0.926448,
		0.927446,
		0.928428,
		0.929393,
		0.930343,
		0.931276,
		0.932195,
		0.933098,
		0.933986,
		0.93486,
		0.93572,
		0.936566,
		0.937398,
		0.938217,
		0.939022,
		0.939815,
		0.940595,
		0.941362,
		0.942118,
		0.942861,
		0.943593,
		0.944313,
		0.945021,
		0.945719,
		0.946406,
		0.964076,
		0.963605,
		0.963126,
		0.96264,
		0.962145,
		0.961642,
		0.961131,
		0.960612,
		0.960083,
		0.959546,
		0.959001,
		0.958445,
		0.957881,
		0.957307,
		0.956724,
		0.95613,
		0.955526,
		0.954912,
		0.954288,
		0.953653,
		0.953007,
		0.95235,
		0.951681,
		0.951001,
		0.950308,
		0.949604,
		0.948887,
		0.948158,
		0.947416,
		0.946661,
		0.945892,
		0.945109,
		0.944313,
		0.943502,
		0.942676,
		0.941836,
		0.94098,
		0.940109,
		0.939222,
		0.938318,
		0.937398,
		0.936461,
		0.935507,
		0.934534,
		0.933544,
		0.932535,
		0.931507,
		0.93046,
		0.929393,
		0.928306,
		0.927198,
		0.926069,
		0.924919,
		0.923746,
		0.922551,
		0.921332,
		0.92009,
		0.918824,
		0.917532,
		0.916216,
		0.914874,
		0.913505,
		0.912108,
		0.910685,
		0.909232,
		0.907751,
		0.90624,
		0.904698,
		0.903125,
		0.90152,
		0.899882,
		0.898211,
		0.896506,
		0.894765,
		0.892988,
		0.891175,
		0.889324,
		0.887434,
		0.885504,
		0.883534,
		0.881522,
		0.879467,
		0.877369,
		0.875226,
		0.873037,
		0.870801,
		0.868517,
		0.866184,
		0.863799,
		0.861363,
		0.858873,
		0.856329,
		0.853729,
		0.851071,
		0.848355,
		0.845578,
		0.842738,
		0.839836,
		0.836868,
		0.833833,
		0.830729,
		0.827555,
		0.824309,
		0.820988,
		0.817592,
		0.814117,
		0.810562,
		0.806924,
		0.803203,
		0.799394,
		0.795497,
		0.791508,
		0.787425,
		0.783246,
		0.778969,
		0.77459,
		0.770107,
		0.765517,
		0.760817,
		0.756005,
		0.751077,
		0.74603,
		0.740862,
		0.735568,
		0.730146,
		0.724593,
		0.718904,
		0.713077,
		0.707107,
		0.700991,
		0.694725,
		0.688306,
		0.681729,
		0.674989,
		0.668082,
		0.661,
		0.653732,
		0.646263,
		0.638572,
		0.630629,
		0.622395,
		0.61382,
		0.604841,
		0.595385,
		0.585368,
		0.574699,
		0.563276,
		0.551001,
		0.537774,
		0.523504,
		0.508113,
		0.491542,
		0.473757,
		0.454752,
		0.434554,
		0.413227,
		0.390871,
		0.367622,
		0.34365,
		0.319156,
		0.294361,
		0.269507,
		0.244843,
		0.220619,
		0.197077,
		0.174444,
		0.152922,
		0.132688,
		0.113883,
		0.096614,
		0.08095,
		0.066924,
		0.054533,
		0.043741,
		0.034483,
		0.026672,
		0.020197,
		0.014934,
		0.010748,
		0.007501,
		0.005051,
		0.003262,
		0.002005,
		0.00116,
		0.000623,
		0.000304,
		0.000131,
		4.8e-005,
		1.4e-005,
		1.4e-005,
		4.8e-005,
		0.000131,
		0.000304,
		0.000623,
		0.00116,
		0.002005,
		0.003262,
		0.005051,
		0.007501,
		0.010748,
		0.014934,
		0.020197,
		0.026672,
		0.034483,
		0.043741,
		0.054533,
		0.066924,
		0.08095,
		0.096614,
		0.113883,
		0.132688,
		0.152922,
		0.174444,
		0.197077,
		0.220619,
		0.244843,
		0.269507,
		0.294361,
		0.319156,
		0.34365,
		0.367622,
		0.390871,
		0.413227,
		0.434554,
		0.454752,
		0.473757,
		0.491542,
		0.508113,
		0.523504,
		0.537774,
		0.551001,
		0.563276,
		0.574699,
		0.585368,
		0.595385,
		0.604841,
		0.61382,
		0.622395,
		0.630629,
		0.638572,
		0.646263,
		0.653732,
		0.661,
		0.668082,
		0.674989,
		0.681729,
		0.688306,
		0.694725,
		0.700991,
		0.707107,
		0.713077,
		0.718904,
		0.724593,
		0.730146,
		0.735568,
		0.740862,
		0.74603,
		0.751077,
		0.756005,
		0.760817,
		0.765517,
		0.770107,
		0.77459,
		0.778969,
		0.783246,
		0.787425,
		0.791508,
		0.795497,
		0.799394,
		0.803203,
		0.806924,
		0.810562,
		0.814117,
		0.817592,
		0.820988,
		0.824309,
		0.827555,
		0.830729,
		0.833833,
		0.836868,
		0.839836,
		0.842738,
		0.845578,
		0.848355,
		0.851071,
		0.853729,
		0.856329,
		0.858873,
		0.861363,
		0.863799,
		0.866184,
		0.868517,
		0.870801,
		0.873037,
		0.875226,
		0.877369,
		0.879467,
		0.881522,
		0.883534,
		0.885504,
		0.887434,
		0.889324,
		0.891175,
		0.892988,
		0.894765,
		0.896506,
		0.898211,
		0.899882,
		0.90152,
		0.903125,
		0.904698,
		0.90624,
		0.907751,
		0.909232,
		0.910685,
		0.912108,
		0.913505,
		0.914874,
		0.916216,
		0.917532,
		0.918824,
		0.92009,
		0.921332,
		0.922551,
		0.923746,
		0.924919,
		0.926069,
		0.927198,
		0.928306,
		0.929393,
		0.93046,
		0.931507,
		0.932535,
		0.933544,
		0.934534,
		0.935507,
		0.936461,
		0.937398,
		0.938318,
		0.939222,
		0.940109,
		0.94098,
		0.941836,
		0.942676,
		0.943502,
		0.944313,
		0.945109,
		0.945892,
		0.946661,
		0.947416,
		0.948158,
		0.948887,
		0.949604,
		0.950308,
		0.951001,
		0.951681,
		0.95235,
		0.953007,
		0.953653,
		0.954288,
		0.954912,
		0.955526,
		0.95613,
		0.956724,
		0.957307,
		0.957881,
		0.958445,
		0.959001,
		0.959546,
		0.960083,
		0.960612,
		0.961131,
		0.961642,
		0.962145,
		0.96264,
		0.963126,
		0.963605,
		0.977109,
		0.976801,
		0.976488,
		0.976169,
		0.975845,
		0.975516,
		0.975181,
		0.97484,
		0.974494,
		0.974141,
		0.973783,
		0.973418,
		0.973047,
		0.97267,
		0.972286,
		0.971895,
		0.971498,
		0.971093,
		0.970681,
		0.970262,
		0.969836,
		0.969402,
		0.96896,
		0.96851,
		0.968053,
		0.967586,
		0.967112,
		0.966629,
		0.966137,
		0.965636,
		0.965125,
		0.964606,
		0.964076,
		0.963537,
		0.962988,
		0.962429,
		0.961859,
		0.961278,
		0.960686,
		0.960083,
		0.959469,
		0.958843,
		0.958205,
		0.957554,
		0.956891,
		0.956215,
		0.955526,
		0.954824,
		0.954108,
		0.953377,
		0.952633,
		0.951873,
		0.951098,
		0.950308,
		0.949502,
		0.94868,
		0.947842,
		0.946986,
		0.946113,
		0.945222,
		0.944313,
		0.943385,
		0.942438,
		0.941471,
		0.940484,
		0.939477,
		0.938448,
		0.937398,
		0.936326,
		0.935231,
		0.934112,
		0.93297,
		0.931803,
		0.930611,
		0.929393,
		0.928149,
		0.926878,
		0.925579,
		0.924251,
		0.922895,
		0.921508,
		0.92009,
		0.918641,
		0.917159,
		0.915644,
		0.914095,
		0.91251,
		0.91089,
		0.909232,
		0.907537,
		0.905802,
		0.904028,
		0.902212,
		0.900354,
		0.898452,
		0.896506,
		0.894513,
		0.892474,
		0.890386,
		0.888248,
		0.886059,
		0.883818,
		0.881522,
		0.87917,
		0.876762,
		0.874294,
		0.871765,
		0.869175,
		0.86652,
		0.863799,
		0.861011,
		0.858152,
		0.855222,
		0.852217,
		0.849137,
		0.845978,
		0.842738,
		0.839416,
		0.836008,
		0.832511,
		0.828924,
		0.825244,
		0.821467,
		0.817592,
		0.813614,
		0.809531,
		0.80534,
		0.801037,
		0.796619,
		0.792083,
		0.787425,
		0.782641,
		0.777728,
		0.77268,
		0.767493,
		0.762157,
		0.756663,
		0.750991,
		0.745119,
		0.739009,
		0.732617,
		0.72588,
		0.718723,
		0.711055,
		0.702768,
		0.693744,
		0.683852,
		0.672954,
		0.660911,
		0.64759,
		0.632867,
		0.616635,
		0.598814,
		0.579355,
		0.558244,
		0.535513,
		0.511233,
		0.485526,
		0.458554,
		0.430524,
		0.401677,
		0.372282,
		0.342631,
		0.313028,
		0.283774,
		0.255168,
		0.227488,
		0.200987,
		0.175889,
		0.152377,
		0.130598,
		0.110655,
		0.092608,
		0.07648,
		0.062253,
		0.049875,
		0.039266,
		0.030318,
		0.022903,
		0.016879,
		0.012091,
		0.008381,
		0.005589,
		0.003558,
		0.002141,
		0.0012,
		0.000614,
		0.000278,
		0.000106,
		3.1e-005,
		3.1e-005,
		0.000106,
		0.000278,
		0.000614,
		0.0012,
		0.002141,
		0.003558,
		0.005589,
		0.008381,
		0.012091,
		0.016879,
		0.022903,
		0.030318,
		0.039266,
		0.049875,
		0.062253,
		0.07648,
		0.092608,
		0.110655,
		0.130598,
		0.152377,
		0.175889,
		0.200987,
		0.227488,
		0.255168,
		0.283774,
		0.313028,
		0.342631,
		0.372282,
		0.401677,
		0.430524,
		0.458554,
		0.485526,
		0.511233,
		0.535513,
		0.558244,
		0.579355,
		0.598814,
		0.616635,
		0.632867,
		0.64759,
		0.660911,
		0.672954,
		0.683852,
		0.693744,
		0.702768,
		0.711055,
		0.718723,
		0.72588,
		0.732617,
		0.739009,
		0.745119,
		0.750991,
		0.756663,
		0.762157,
		0.767493,
		0.77268,
		0.777728,
		0.782641,
		0.787425,
		0.792083,
		0.796619,
		0.801037,
		0.80534,
		0.809531,
		0.813614,
		0.817592,
		0.821467,
		0.825244,
		0.828924,
		0.832511,
		0.836008,
		0.839416,
		0.842738,
		0.845978,
		0.849137,
		0.852217,
		0.855222,
		0.858152,
		0.861011,
		0.863799,
		0.86652,
		0.869175,
		0.871765,
		0.874294,
		0.876762,
		0.87917,
		0.881522,
		0.883818,
		0.886059,
		0.888248,
		0.890386,
		0.892474,
		0.894513,
		0.896506,
		0.898452,
		0.900354,
		0.902212,
		0.904028,
		0.905802,
		0.907537,
		0.909232,
		0.91089,
		0.91251,
		0.914095,
		0.915644,
		0.917159,
		0.918641,
		0.92009,
		0.921508,
		0.922895,
		0.924251,
		0.925579,
		0.926878,
		0.928149,
		0.929393,
		0.930611,
		0.931803,
		0.93297,
		0.934112,
		0.935231,
		0.936326,
		0.937398,
		0.938448,
		0.939477,
		0.940484,
		0.941471,
		0.942438,
		0.943385,
		0.944313,
		0.945222,
		0.946113,
		0.946986,
		0.947842,
		0.94868,
		0.949502,
		0.950308,
		0.951098,
		0.951873,
		0.952633,
		0.953377,
		0.954108,
		0.954824,
		0.955526,
		0.956215,
		0.956891,
		0.957554,
		0.958205,
		0.958843,
		0.959469,
		0.960083,
		0.960686,
		0.961278,
		0.961859,
		0.962429,
		0.962988,
		0.963537,
		0.964076,
		0.964606,
		0.965125,
		0.965636,
		0.966137,
		0.966629,
		0.967112,
		0.967586,
		0.968053,
		0.96851,
		0.96896,
		0.969402,
		0.969836,
		0.970262,
		0.970681,
		0.971093,
		0.971498,
		0.971895,
		0.972286,
		0.97267,
		0.973047,
		0.973418,
		0.973783,
		0.974141,
		0.974494,
		0.97484,
		0.975181,
		0.975516,
		0.975845,
		0.976169,
		0.976488,
		0.976801,
		0.986574,
		0.986389,
		0.986201,
		0.986009,
		0.985814,
		0.985615,
		0.985413,
		0.985208,
		0.984999,
		0.984786,
		0.984569,
		0.984349,
		0.984125,
		0.983896,
		0.983664,
		0.983428,
		0.983187,
		0.982941,
		0.982692,
		0.982438,
		0.982179,
		0.981915,
		0.981647,
		0.981373,
		0.981094,
		0.980811,
		0.980522,
		0.980227,
		0.979927,
		0.979622,
		0.97931,
		0.978993,
		0.978669,
		0.978339,
		0.978003,
		0.977661,
		0.977312,
		0.976956,
		0.976593,
		0.976223,
		0.975845,
		0.97546,
		0.975068,
		0.974668,
		0.97426,
		0.973843,
		0.973418,
		0.972985,
		0.972543,
		0.972091,
		0.971631,
		0.971161,
		0.970681,
		0.970192,
		0.969692,
		0.969182,
		0.968661,
		0.968129,
		0.967586,
		0.967032,
		0.966466,
		0.965887,
		0.965297,
		0.964693,
		0.964076,
		0.963446,
		0.962803,
		0.962145,
		0.961473,
		0.960786,
		0.960083,
		0.959365,
		0.958631,
		0.957881,
		0.957114,
		0.956329,
		0.955526,
		0.954706,
		0.953866,
		0.953007,
		0.952128,
		0.951229,
		0.950308,
		0.949367,
		0.948403,
		0.947416,
		0.946406,
		0.945372,
		0.944313,
		0.943228,
		0.942118,
		0.94098,
		0.939815,
		0.938621,
		0.937398,
		0.936145,
		0.93486,
		0.933544,
		0.932195,
		0.930812,
		0.929393,
		0.927939,
		0.926448,
		0.924919,
		0.92335,
		0.921741,
		0.92009,
		0.918396,
		0.916658,
		0.914874,
		0.913042,
		0.911162,
		0.909232,
		0.90725,
		0.905215,
		0.903125,
		0.900978,
		0.898772,
		0.896506,
		0.894177,
		0.891784,
		0.889324,
		0.886795,
		0.884195,
		0.881522,
		0.878773,
		0.875946,
		0.873037,
		0.870045,
		0.866967,
		0.863799,
		0.860539,
		0.857183,
		0.853728,
		0.850168,
		0.846495,
		0.842698,
		0.838759,
		0.83465,
		0.830333,
		0.825756,
		0.820849,
		0.815526,
		0.809681,
		0.803188,
		0.795905,
		0.787675,
		0.778329,
		0.767694,
		0.755596,
		0.741871,
		0.726371,
		0.708972,
		0.689582,
		0.66815,
		0.644671,
		0.61919,
		0.591806,
		0.56267,
		0.531985,
		0.5,
		0.467005,
		0.433319,
		0.399286,
		0.365257,
		0.331586,
		0.298613,
		0.26666,
		0.236018,
		0.206942,
		0.179646,
		0.1543,
		0.131026,
		0.109901,
		0.090958,
		0.074187,
		0.05954,
		0.046936,
		0.036262,
		0.027382,
		0.020142,
		0.014372,
		0.009893,
		0.006523,
		0.00408,
		0.002388,
		0.001282,
		0.000613,
		0.000248,
		7.7e-005,
		1.5e-005,
		1.5e-005,
		7.7e-005,
		0.000248,
		0.000613,
		0.001282,
		0.002388,
		0.00408,
		0.006523,
		0.009893,
		0.014372,
		0.020142,
		0.027382,
		0.036262,
		0.046936,
		0.05954,
		0.074187,
		0.090958,
		0.109901,
		0.131026,
		0.1543,
		0.179646,
		0.206942,
		0.236018,
		0.26666,
		0.298613,
		0.331586,
		0.365257,
		0.399286,
		0.433319,
		0.467005,
		0.5,
		0.531985,
		0.56267,
		0.591806,
		0.61919,
		0.644671,
		0.66815,
		0.689582,
		0.708972,
		0.726371,
		0.741871,
		0.755596,
		0.767694,
		0.778329,
		0.787675,
		0.795905,
		0.803188,
		0.809681,
		0.815526,
		0.820849,
		0.825756,
		0.830333,
		0.83465,
		0.838759,
		0.842698,
		0.846495,
		0.850168,
		0.853728,
		0.857183,
		0.860539,
		0.863799,
		0.866967,
		0.870045,
		0.873037,
		0.875946,
		0.878773,
		0.881522,
		0.884195,
		0.886795,
		0.889324,
		0.891784,
		0.894177,
		0.896506,
		0.898772,
		0.900978,
		0.903125,
		0.905215,
		0.90725,
		0.909232,
		0.911162,
		0.913042,
		0.914874,
		0.916658,
		0.918396,
		0.92009,
		0.921741,
		0.92335,
		0.924919,
		0.926448,
		0.927939,
		0.929393,
		0.930812,
		0.932195,
		0.933544,
		0.93486,
		0.936145,
		0.937398,
		0.938621,
		0.939815,
		0.94098,
		0.942118,
		0.943228,
		0.944313,
		0.945372,
		0.946406,
		0.947416,
		0.948403,
		0.949367,
		0.950308,
		0.951229,
		0.952128,
		0.953007,
		0.953866,
		0.954706,
		0.955526,
		0.956329,
		0.957114,
		0.957881,
		0.958631,
		0.959365,
		0.960083,
		0.960786,
		0.961473,
		0.962145,
		0.962803,
		0.963446,
		0.964076,
		0.964693,
		0.965297,
		0.965887,
		0.966466,
		0.967032,
		0.967586,
		0.968129,
		0.968661,
		0.969182,
		0.969692,
		0.970192,
		0.970681,
		0.971161,
		0.971631,
		0.972091,
		0.972543,
		0.972985,
		0.973418,
		0.973843,
		0.97426,
		0.974668,
		0.975068,
		0.97546,
		0.975845,
		0.976223,
		0.976593,
		0.976956,
		0.977312,
		0.977661,
		0.978003,
		0.978339,
		0.978669,
		0.978993,
		0.97931,
		0.979622,
		0.979927,
		0.980227,
		0.980522,
		0.980811,
		0.981094,
		0.981373,
		0.981647,
		0.981915,
		0.982179,
		0.982438,
		0.982692,
		0.982941,
		0.983187,
		0.983428,
		0.983664,
		0.983896,
		0.984125,
		0.984349,
		0.984569,
		0.984786,
		0.984999,
		0.985208,
		0.985413,
		0.985615,
		0.985814,
		0.986009,
		0.986201,
		0.986389,
		0.992969,
		0.99287,
		0.992769,
		0.992666,
		0.992561,
		0.992454,
		0.992346,
		0.992235,
		0.992122,
		0.992008,
		0.991891,
		0.991772,
		0.991651,
		0.991528,
		0.991403,
		0.991275,
		0.991144,
		0.991012,
		0.990877,
		0.990739,
		0.990599,
		0.990456,
		0.99031,
		0.990162,
		0.990011,
		0.989857,
		0.9897,
		0.98954,
		0.989376,
		0.98921,
		0.98904,
		0.988867,
		0.988691,
		0.988511,
		0.988327,
		0.98814,
		0.987949,
		0.987755,
		0.987556,
		0.987353,
		0.987147,
		0.986936,
		0.98672,
		0.986501,
		0.986276,
		0.986047,
		0.985814,
		0.985575,
		0.985332,
		0.985083,
		0.984829,
		0.984569,
		0.984305,
		0.984034,
		0.983758,
		0.983475,
		0.983187,
		0.982892,
		0.982591,
		0.982283,
		0.981968,
		0.981647,
		0.981318,
		0.980982,
		0.980638,
		0.980287,
		0.979927,
		0.97956,
		0.979184,
		0.978799,
		0.978406,
		0.978003,
		0.977591,
		0.97717,
		0.976739,
		0.976297,
		0.975845,
		0.975383,
		0.974909,
		0.974424,
		0.973927,
		0.973418,
		0.972897,
		0.972363,
		0.971816,
		0.971256,
		0.970681,
		0.970093,
		0.969489,
		0.968871,
		0.968237,
		0.967586,
		0.96692,
		0.966236,
		0.965534,
		0.964815,
		0.964076,
		0.963319,
		0.962541,
		0.961743,
		0.960924,
		0.960083,
		0.95922,
		0.958333,
		0.957423,
		0.956487,
		0.955526,
		0.954539,
		0.953525,
		0.952482,
		0.95141,
		0.950308,
		0.949176,
		0.948011,
		0.946813,
		0.94558,
		0.944313,
		0.943008,
		0.941666,
		0.940284,
		0.938862,
		0.937398,
		0.93589,
		0.934338,
		0.932739,
		0.931091,
		0.929393,
		0.927644,
		0.925841,
		0.923982,
		0.922066,
		0.92009,
		0.918052,
		0.915948,
		0.913775,
		0.911526,
		0.909189,
		0.906745,
		0.904167,
		0.901414,
		0.898429,
		0.895138,
		0.891446,
		0.887236,
		0.882368,
		0.876682,
		0.870001,
		0.862132,
		0.852871,
		0.842016,
		0.829368,
		0.814743,
		0.797981,
		0.778956,
		0.757581,
		0.733821,
		0.707694,
		0.679276,
		0.648705,
		0.616177,
		0.581942,
		0.546299,
		0.509589,
		0.47218,
		0.434461,
		0.396826,
		0.359665,
		0.32335,
		0.288229,
		0.254614,
		0.222774,
		0.192937,
		0.165276,
		0.13992,
		0.116944,
		0.096378,
		0.078208,
		0.062379,
		0.048799,
		0.037343,
		0.027863,
		0.020186,
		0.014124,
		0.009478,
		0.006041,
		0.003609,
		0.001981,
		0.00097,
		0.000402,
		0.000129,
		2.6e-005,
		2.6e-005,
		0.000129,
		0.000402,
		0.00097,
		0.001981,
		0.003609,
		0.006041,
		0.009478,
		0.014124,
		0.020186,
		0.027863,
		0.037343,
		0.048799,
		0.062379,
		0.078208,
		0.096378,
		0.116944,
		0.13992,
		0.165276,
		0.192937,
		0.222774,
		0.254614,
		0.288229,
		0.32335,
		0.359665,
		0.396826,
		0.434461,
		0.47218,
		0.509589,
		0.546299,
		0.581942,
		0.616177,
		0.648705,
		0.679276,
		0.707694,
		0.733821,
		0.757581,
		0.778956,
		0.797981,
		0.814743,
		0.829368,
		0.842016,
		0.852871,
		0.862132,
		0.870001,
		0.876682,
		0.882368,
		0.887236,
		0.891446,
		0.895138,
		0.898429,
		0.901414,
		0.904167,
		0.906745,
		0.909189,
		0.911526,
		0.913775,
		0.915948,
		0.918052,
		0.92009,
		0.922066,
		0.923982,
		0.925841,
		0.927644,
		0.929393,
		0.931091,
		0.932739,
		0.934338,
		0.93589,
		0.937398,
		0.938862,
		0.940284,
		0.941666,
		0.943008,
		0.944313,
		0.94558,
		0.946813,
		0.948011,
		0.949176,
		0.950308,
		0.95141,
		0.952482,
		0.953525,
		0.954539,
		0.955526,
		0.956487,
		0.957423,
		0.958333,
		0.95922,
		0.960083,
		0.960924,
		0.961743,
		0.962541,
		0.963319,
		0.964076,
		0.964815,
		0.965534,
		0.966236,
		0.96692,
		0.967586,
		0.968237,
		0.968871,
		0.969489,
		0.970093,
		0.970681,
		0.971256,
		0.971816,
		0.972363,
		0.972897,
		0.973418,
		0.973927,
		0.974424,
		0.974909,
		0.975383,
		0.975845,
		0.976297,
		0.976739,
		0.97717,
		0.977591,
		0.978003,
		0.978406,
		0.978799,
		0.979184,
		0.97956,
		0.979927,
		0.980287,
		0.980638,
		0.980982,
		0.981318,
		0.981647,
		0.981968,
		0.982283,
		0.982591,
		0.982892,
		0.983187,
		0.983475,
		0.983758,
		0.984034,
		0.984305,
		0.984569,
		0.984829,
		0.985083,
		0.985332,
		0.985575,
		0.985814,
		0.986047,
		0.986276,
		0.986501,
		0.98672,
		0.986936,
		0.987147,
		0.987353,
		0.987556,
		0.987755,
		0.987949,
		0.98814,
		0.988327,
		0.988511,
		0.988691,
		0.988867,
		0.98904,
		0.98921,
		0.989376,
		0.98954,
		0.9897,
		0.989857,
		0.990011,
		0.990162,
		0.99031,
		0.990456,
		0.990599,
		0.990739,
		0.990877,
		0.991012,
		0.991144,
		0.991275,
		0.991403,
		0.991528,
		0.991651,
		0.991772,
		0.991891,
		0.992008,
		0.992122,
		0.992235,
		0.992346,
		0.992454,
		0.992561,
		0.992666,
		0.992769,
		0.99287,
		0.996876,
		0.99683,
		0.996784,
		0.996737,
		0.99669,
		0.996641,
		0.996591,
		0.996541,
		0.99649,
		0.996437,
		0.996384,
		0.996329,
		0.996274,
		0.996218,
		0.99616,
		0.996102,
		0.996042,
		0.995981,
		0.995919,
		0.995856,
		0.995792,
		0.995726,
		0.995659,
		0.995591,
		0.995521,
		0.99545,
		0.995378,
		0.995304,
		0.995229,
		0.995152,
		0.995074,
		0.994994,
		0.994912,
		0.994829,
		0.994744,
		0.994657,
		0.994569,
		0.994478,
		0.994386,
		0.994292,
		0.994196,
		0.994098,
		0.993998,
		0.993896,
		0.993791,
		0.993685,
		0.993576,
		0.993464,
		0.99335,
		0.993234,
		0.993115,
		0.992994,
		0.99287,
		0.992743,
		0.992614,
		0.992481,
		0.992346,
		0.992207,
		0.992065,
		0.99192,
		0.991772,
		0.991621,
		0.991466,
		0.991307,
		0.991144,
		0.990978,
		0.990808,
		0.990634,
		0.990456,
		0.990274,
		0.990087,
		0.989896,
		0.9897,
		0.989499,
		0.989293,
		0.989083,
		0.988867,
		0.988646,
		0.98842,
		0.988187,
		0.987949,
		0.987705,
		0.987455,
		0.987199,
		0.986936,
		0.986666,
		0.986389,
		0.986105,
		0.985814,
		0.985515,
		0.985208,
		0.984893,
		0.984569,
		0.984237,
		0.983896,
		0.983546,
		0.983187,
		0.982817,
		0.982438,
		0.982047,
		0.981647,
		0.981234,
		0.980811,
		0.980375,
		0.979927,
		0.979466,
		0.978993,
		0.978505,
		0.978003,
		0.977487,
		0.976956,
		0.976409,
		0.975845,
		0.975265,
		0.974668,
		0.974052,
		0.973418,
		0.972765,
		0.972091,
		0.971397,
		0.970681,
		0.969943,
		0.969182,
		0.968397,
		0.967586,
		0.96675,
		0.965887,
		0.964996,
		0.964076,
		0.963126,
		0.962145,
		0.961131,
		0.960083,
		0.958999,
		0.957876,
		0.956706,
		0.955481,
		0.95418,
		0.952776,
		0.951227,
		0.949475,
		0.94744,
		0.945023,
		0.942099,
		0.938519,
		0.934111,
		0.928683,
		0.922022,
		0.913908,
		0.904114,
		0.892416,
		0.878606,
		0.862495,
		0.84393,
		0.822798,
		0.799039,
		0.77265,
		0.743688,
		0.712278,
		0.678608,
		0.642924,
		0.605532,
		0.566781,
		0.527057,
		0.486772,
		0.446348,
		0.406208,
		0.366762,
		0.328396,
		0.291463,
		0.256277,
		0.223105,
		0.192162,
		0.163614,
		0.137572,
		0.1141,
		0.09321,
		0.07487,
		0.059008,
		0.045513,
		0.034243,
		0.025028,
		0.017677,
		0.01198,
		0.007717,
		0.004662,
		0.002591,
		0.001285,
		0.00054,
		0.000175,
		3.5e-005,
		3.5e-005,
		0.000175,
		0.00054,
		0.001285,
		0.002591,
		0.004662,
		0.007717,
		0.01198,
		0.017677,
		0.025028,
		0.034243,
		0.045513,
		0.059008,
		0.07487,
		0.09321,
		0.1141,
		0.137572,
		0.163614,
		0.192162,
		0.223105,
		0.256277,
		0.291463,
		0.328396,
		0.366762,
		0.406208,
		0.446348,
		0.486772,
		0.527057,
		0.566781,
		0.605532,
		0.642924,
		0.678608,
		0.712278,
		0.743688,
		0.77265,
		0.799039,
		0.822798,
		0.84393,
		0.862495,
		0.878606,
		0.892416,
		0.904114,
		0.913908,
		0.922022,
		0.928683,
		0.934111,
		0.938519,
		0.942099,
		0.945023,
		0.94744,
		0.949475,
		0.951227,
		0.952776,
		0.95418,
		0.955481,
		0.956706,
		0.957876,
		0.958999,
		0.960083,
		0.961131,
		0.962145,
		0.963126,
		0.964076,
		0.964996,
		0.965887,
		0.96675,
		0.967586,
		0.968397,
		0.969182,
		0.969943,
		0.970681,
		0.971397,
		0.972091,
		0.972765,
		0.973418,
		0.974052,
		0.974668,
		0.975265,
		0.975845,
		0.976409,
		0.976956,
		0.977487,
		0.978003,
		0.978505,
		0.978993,
		0.979466,
		0.979927,
		0.980375,
		0.980811,
		0.981234,
		0.981647,
		0.982047,
		0.982438,
		0.982817,
		0.983187,
		0.983546,
		0.983896,
		0.984237,
		0.984569,
		0.984893,
		0.985208,
		0.985515,
		0.985814,
		0.986105,
		0.986389,
		0.986666,
		0.986936,
		0.987199,
		0.987455,
		0.987705,
		0.987949,
		0.988187,
		0.98842,
		0.988646,
		0.988867,
		0.989083,
		0.989293,
		0.989499,
		0.9897,
		0.989896,
		0.990087,
		0.990274,
		0.990456,
		0.990634,
		0.990808,
		0.990978,
		0.991144,
		0.991307,
		0.991466,
		0.991621,
		0.991772,
		0.99192,
		0.992065,
		0.992207,
		0.992346,
		0.992481,
		0.992614,
		0.992743,
		0.99287,
		0.992994,
		0.993115,
		0.993234,
		0.99335,
		0.993464,
		0.993576,
		0.993685,
		0.993791,
		0.993896,
		0.993998,
		0.994098,
		0.994196,
		0.994292,
		0.994386,
		0.994478,
		0.994569,
		0.994657,
		0.994744,
		0.994829,
		0.994912,
		0.994994,
		0.995074,
		0.995152,
		0.995229,
		0.995304,
		0.995378,
		0.99545,
		0.995521,
		0.995591,
		0.995659,
		0.995726,
		0.995792,
		0.995856,
		0.995919,
		0.995981,
		0.996042,
		0.996102,
		0.99616,
		0.996218,
		0.996274,
		0.996329,
		0.996384,
		0.996437,
		0.99649,
		0.996541,
		0.996591,
		0.996641,
		0.99669,
		0.996737,
		0.996784,
		0.99683,
		0.998929,
		0.998913,
		0.998897,
		0.99888,
		0.998863,
		0.998846,
		0.998829,
		0.998811,
		0.998793,
		0.998774,
		0.998756,
		0.998737,
		0.998717,
		0.998697,
		0.998677,
		0.998656,
		0.998635,
		0.998614,
		0.998592,
		0.998569,
		0.998546,
		0.998523,
		0.998499,
		0.998475,
		0.998451,
		0.998425,
		0.9984,
		0.998373,
		0.998347,
		0.998319,
		0.998291,
		0.998263,
		0.998234,
		0.998204,
		0.998174,
		0.998143,
		0.998111,
		0.998079,
		0.998046,
		0.998013,
		0.997978,
		0.997943,
		0.997907,
		0.99787,
		0.997833,
		0.997795,
		0.997755,
		0.997715,
		0.997675,
		0.997633,
		0.99759,
		0.997546,
		0.997501,
		0.997456,
		0.997409,
		0.997361,
		0.997312,
		0.997262,
		0.99721,
		0.997158,
		0.997104,
		0.997049,
		0.996992,
		0.996935,
		0.996876,
		0.996815,
		0.996753,
		0.99669,
		0.996625,
		0.996558,
		0.99649,
		0.996419,
		0.996348,
		0.996274,
		0.996199,
		0.996121,
		0.996042,
		0.995961,
		0.995877,
		0.995792,
		0.995704,
		0.995614,
		0.995521,
		0.995426,
		0.995329,
		0.995229,
		0.995126,
		0.99502,
		0.994912,
		0.994801,
		0.994686,
		0.994569,
		0.994448,
		0.994324,
		0.994196,
		0.994065,
		0.99393,
		0.993791,
		0.993648,
		0.993502,
		0.99335,
		0.993195,
		0.993035,
		0.99287,
		0.9927,
		0.992526,
		0.992346,
		0.99216,
		0.991969,
		0.991772,
		0.991569,
		0.99136,
		0.991144,
		0.990922,
		0.990693,
		0.990456,
		0.990212,
		0.98996,
		0.9897,
		0.989431,
		0.989154,
		0.988867,
		0.988571,
		0.988265,
		0.987949,
		0.987623,
		0.987285,
		0.986936,
		0.986574,
		0.986201,
		0.985814,
		0.985413,
		0.984999,
		0.984568,
		0.984119,
		0.983646,
		0.983139,
		0.98258,
		0.981941,
		0.981178,
		0.980234,
		0.979026,
		0.977452,
		0.975383,
		0.972665,
		0.969117,
		0.964539,
		0.958709,
		0.951392,
		0.942348,
		0.931338,
		0.918136,
		0.902538,
		0.884372,
		0.863508,
		0.83987,
		0.813438,
		0.784259,
		0.752446,
		0.718179,
		0.681705,
		0.643327,
		0.603399,
		0.562317,
		0.520504,
		0.4784,
		0.436447,
		0.395076,
		0.354699,
		0.315694,
		0.278401,
		0.243113,
		0.210073,
		0.179472,
		0.151444,
		0.126076,
		0.103399,
		0.083398,
		0.066016,
		0.051152,
		0.038672,
		0.02841,
		0.020174,
		0.01375,
		0.00891,
		0.005418,
		0.003031,
		0.001514,
		0.000642,
		0.00021,
		4.3e-005,
		4.3e-005,
		0.00021,
		0.000642,
		0.001514,
		0.003031,
		0.005418,
		0.00891,
		0.01375,
		0.020174,
		0.02841,
		0.038672,
		0.051152,
		0.066016,
		0.083398,
		0.103399,
		0.126076,
		0.151444,
		0.179472,
		0.210073,
		0.243113,
		0.278401,
		0.315694,
		0.354699,
		0.395076,
		0.436447,
		0.4784,
		0.520504,
		0.562317,
		0.603399,
		0.643327,
		0.681705,
		0.718179,
		0.752446,
		0.784259,
		0.813438,
		0.83987,
		0.863508,
		0.884372,
		0.902538,
		0.918136,
		0.931338,
		0.942348,
		0.951392,
		0.958709,
		0.964539,
		0.969117,
		0.972665,
		0.975383,
		0.977452,
		0.979026,
		0.980234,
		0.981178,
		0.981941,
		0.98258,
		0.983139,
		0.983646,
		0.984119,
		0.984568,
		0.984999,
		0.985413,
		0.985814,
		0.986201,
		0.986574,
		0.986936,
		0.987285,
		0.987623,
		0.987949,
		0.988265,
		0.988571,
		0.988867,
		0.989154,
		0.989431,
		0.9897,
		0.98996,
		0.990212,
		0.990456,
		0.990693,
		0.990922,
		0.991144,
		0.99136,
		0.991569,
		0.991772,
		0.991969,
		0.99216,
		0.992346,
		0.992526,
		0.9927,
		0.99287,
		0.993035,
		0.993195,
		0.99335,
		0.993502,
		0.993648,
		0.993791,
		0.99393,
		0.994065,
		0.994196,
		0.994324,
		0.994448,
		0.994569,
		0.994686,
		0.994801,
		0.994912,
		0.99502,
		0.995126,
		0.995229,
		0.995329,
		0.995426,
		0.995521,
		0.995614,
		0.995704,
		0.995792,
		0.995877,
		0.995961,
		0.996042,
		0.996121,
		0.996199,
		0.996274,
		0.996348,
		0.996419,
		0.99649,
		0.996558,
		0.996625,
		0.99669,
		0.996753,
		0.996815,
		0.996876,
		0.996935,
		0.996992,
		0.997049,
		0.997104,
		0.997158,
		0.99721,
		0.997262,
		0.997312,
		0.997361,
		0.997409,
		0.997456,
		0.997501,
		0.997546,
		0.99759,
		0.997633,
		0.997675,
		0.997715,
		0.997755,
		0.997795,
		0.997833,
		0.99787,
		0.997907,
		0.997943,
		0.997978,
		0.998013,
		0.998046,
		0.998079,
		0.998111,
		0.998143,
		0.998174,
		0.998204,
		0.998234,
		0.998263,
		0.998291,
		0.998319,
		0.998347,
		0.998373,
		0.9984,
		0.998425,
		0.998451,
		0.998475,
		0.998499,
		0.998523,
		0.998546,
		0.998569,
		0.998592,
		0.998614,
		0.998635,
		0.998656,
		0.998677,
		0.998697,
		0.998717,
		0.998737,
		0.998756,
		0.998774,
		0.998793,
		0.998811,
		0.998829,
		0.998846,
		0.998863,
		0.99888,
		0.998897,
		0.998913,
		0.999771,
		0.999767,
		0.999764,
		0.99976,
		0.999757,
		0.999753,
		0.999749,
		0.999745,
		0.999741,
		0.999737,
		0.999733,
		0.999729,
		0.999725,
		0.99972,
		0.999716,
		0.999711,
		0.999707,
		0.999702,
		0.999697,
		0.999692,
		0.999687,
		0.999682,
		0.999677,
		0.999671,
		0.999666,
		0.99966,
		0.999655,
		0.999649,
		0.999643,
		0.999637,
		0.999631,
		0.999624,
		0.999618,
		0.999611,
		0.999605,
		0.999598,
		0.999591,
		0.999584,
		0.999576,
		0.999569,
		0.999561,
		0.999553,
		0.999545,
		0.999537,
		0.999529,
		0.99952,
		0.999511,
		0.999502,
		0.999493,
		0.999484,
		0.999474,
		0.999464,
		0.999454,
		0.999444,
		0.999434,
		0.999423,
		0.999412,
		0.9994,
		0.999389,
		0.999377,
		0.999365,
		0.999352,
		0.99934,
		0.999327,
		0.999313,
		0.999299,
		0.999285,
		0.999271,
		0.999256,
		0.999241,
		0.999225,
		0.99921,
		0.999193,
		0.999176,
		0.999159,
		0.999141,
		0.999123,
		0.999105,
		0.999086,
		0.999066,
		0.999046,
		0.999025,
		0.999004,
		0.998982,
		0.99896,
		0.998936,
		0.998913,
		0.998888,
		0.998863,
		0.998838,
		0.998811,
		0.998784,
		0.998756,
		0.998727,
		0.998697,
		0.998667,
		0.998635,
		0.998603,
		0.998569,
		0.998535,
		0.998499,
		0.998463,
		0.998425,
		0.998387,
		0.998347,
		0.998306,
		0.998263,
		0.998219,
		0.998174,
		0.998127,
		0.998079,
		0.998029,
		0.997978,
		0.997925,
		0.99787,
		0.997814,
		0.997755,
		0.997695,
		0.997633,
		0.997568,
		0.997501,
		0.997432,
		0.997361,
		0.997287,
		0.99721,
		0.997131,
		0.997049,
		0.996964,
		0.996876,
		0.996784,
		0.99669,
		0.996591,
		0.996489,
		0.996382,
		0.996268,
		0.996142,
		0.995994,
		0.995806,
		0.99555,
		0.995184,
		0.994648,
		0.993859,
		0.992715,
		0.991083,
		0.988808,
		0.985706,
		0.981571,
		0.976177,
		0.969283,
		0.960643,
		0.950009,
		0.937149,
		0.921848,
		0.903927,
		0.883247,
		0.859722,
		0.833327,
		0.8041,
		0.772148,
		0.737649,
		0.700845,
		0.662039,
		0.621588,
		0.57989,
		0.537373,
		0.494485,
		0.451676,
		0.409388,
		0.368043,
		0.328034,
		0.289712,
		0.253384,
		0.219306,
		0.18768,
		0.158656,
		0.13233,
		0.108744,
		0.087893,
		0.069727,
		0.054154,
		0.041042,
		0.030229,
		0.021524,
		0.014712,
		0.009563,
		0.005833,
		0.003275,
		0.001641,
		0.000698,
		0.000229,
		4.7e-005,
		4.7e-005,
		0.000229,
		0.000698,
		0.001641,
		0.003275,
		0.005833,
		0.009563,
		0.014712,
		0.021524,
		0.030229,
		0.041042,
		0.054154,
		0.069727,
		0.087893,
		0.108744,
		0.13233,
		0.158656,
		0.18768,
		0.219306,
		0.253384,
		0.289712,
		0.328034,
		0.368043,
		0.409388,
		0.451676,
		0.494485,
		0.537373,
		0.57989,
		0.621588,
		0.662039,
		0.700845,
		0.737649,
		0.772148,
		0.8041,
		0.833327,
		0.859722,
		0.883247,
		0.903927,
		0.921848,
		0.937149,
		0.950009,
		0.960643,
		0.969283,
		0.976177,
		0.981571,
		0.985706,
		0.988808,
		0.991083,
		0.992715,
		0.993859,
		0.994648,
		0.995184,
		0.99555,
		0.995806,
		0.995994,
		0.996142,
		0.996268,
		0.996382,
		0.996489,
		0.996591,
		0.99669,
		0.996784,
		0.996876,
		0.996964,
		0.997049,
		0.997131,
		0.99721,
		0.997287,
		0.997361,
		0.997432,
		0.997501,
		0.997568,
		0.997633,
		0.997695,
		0.997755,
		0.997814,
		0.99787,
		0.997925,
		0.997978,
		0.998029,
		0.998079,
		0.998127,
		0.998174,
		0.998219,
		0.998263,
		0.998306,
		0.998347,
		0.998387,
		0.998425,
		0.998463,
		0.998499,
		0.998535,
		0.998569,
		0.998603,
		0.998635,
		0.998667,
		0.998697,
		0.998727,
		0.998756,
		0.998784,
		0.998811,
		0.998838,
		0.998863,
		0.998888,
		0.998913,
		0.998936,
		0.99896,
		0.998982,
		0.999004,
		0.999025,
		0.999046,
		0.999066,
		0.999086,
		0.999105,
		0.999123,
		0.999141,
		0.999159,
		0.999176,
		0.999193,
		0.99921,
		0.999225,
		0.999241,
		0.999256,
		0.999271,
		0.999285,
		0.999299,
		0.999313,
		0.999327,
		0.99934,
		0.999352,
		0.999365,
		0.999377,
		0.999389,
		0.9994,
		0.999412,
		0.999423,
		0.999434,
		0.999444,
		0.999454,
		0.999464,
		0.999474,
		0.999484,
		0.999493,
		0.999502,
		0.999511,
		0.99952,
		0.999529,
		0.999537,
		0.999545,
		0.999553,
		0.999561,
		0.999569,
		0.999576,
		0.999584,
		0.999591,
		0.999598,
		0.999605,
		0.999611,
		0.999618,
		0.999624,
		0.999631,
		0.999637,
		0.999643,
		0.999649,
		0.999655,
		0.99966,
		0.999666,
		0.999671,
		0.999677,
		0.999682,
		0.999687,
		0.999692,
		0.999697,
		0.999702,
		0.999707,
		0.999711,
		0.999716,
		0.99972,
		0.999725,
		0.999729,
		0.999733,
		0.999737,
		0.999741,
		0.999745,
		0.999749,
		0.999753,
		0.999757,
		0.99976,
		0.999764,
		0.999767,
		0.999985,
		0.999984,
		0.999984,
		0.999984,
		0.999984,
		0.999983,
		0.999983,
		0.999983,
		0.999982,
		0.999982,
		0.999982,
		0.999982,
		0.999981,
		0.999981,
		0.999981,
		0.99998,
		0.99998,
		0.99998,
		0.999979,
		0.999979,
		0.999979,
		0.999978,
		0.999978,
		0.999978,
		0.999977,
		0.999977,
		0.999976,
		0.999976,
		0.999976,
		0.999975,
		0.999975,
		0.999974,
		0.999974,
		0.999973,
		0.999973,
		0.999972,
		0.999972,
		0.999971,
		0.999971,
		0.99997,
		0.99997,
		0.999969,
		0.999969,
		0.999968,
		0.999968,
		0.999967,
		0.999966,
		0.999966,
		0.999965,
		0.999964,
		0.999964,
		0.999963,
		0.999962,
		0.999962,
		0.999961,
		0.99996,
		0.999959,
		0.999959,
		0.999958,
		0.999957,
		0.999956,
		0.999955,
		0.999954,
		0.999953,
		0.999952,
		0.999951,
		0.99995,
		0.999949,
		0.999948,
		0.999947,
		0.999946,
		0.999945,
		0.999944,
		0.999943,
		0.999941,
		0.99994,
		0.999939,
		0.999937,
		0.999936,
		0.999935,
		0.999933,
		0.999932,
		0.99993,
		0.999929,
		0.999927,
		0.999925,
		0.999923,
		0.999922,
		0.99992,
		0.999918,
		0.999916,
		0.999914,
		0.999912,
		0.99991,
		0.999908,
		0.999906,
		0.999903,
		0.999901,
		0.999898,
		0.999896,
		0.999893,
		0.999891,
		0.999888,
		0.999885,
		0.999882,
		0.999879,
		0.999876,
		0.999872,
		0.999869,
		0.999866,
		0.999862,
		0.999858,
		0.999854,
		0.999851,
		0.999846,
		0.999842,
		0.999838,
		0.999833,
		0.999829,
		0.999824,
		0.999819,
		0.999813,
		0.999808,
		0.999802,
		0.999797,
		0.99979,
		0.999784,
		0.999778,
		0.999771,
		0.999764,
		0.999757,
		0.999749,
		0.999741,
		0.999732,
		0.999719,
		0.999698,
		0.999659,
		0.999584,
		0.999445,
		0.9992,
		0.998789,
		0.99813,
		0.997118,
		0.995622,
		0.993485,
		0.990522,
		0.986527,
		0.981272,
		0.974514,
		0.966004,
		0.955494,
		0.942748,
		0.927549,
		0.909715,
		0.889104,
		0.865629,
		0.839261,
		0.810035,
		0.778059,
		0.743507,
		0.706621,
		0.667704,
		0.627111,
		0.585242,
		0.542528,
		0.499416,
		0.45636,
		0.413805,
		0.372177,
		0.33187,
		0.293242,
		0.256601,
		0.222209,
		0.190271,
		0.160942,
		0.13432,
		0.110452,
		0.089337,
		0.070925,
		0.055126,
		0.041813,
		0.030824,
		0.021968,
		0.01503,
		0.00978,
		0.005972,
		0.003356,
		0.001685,
		0.000718,
		0.000236,
		4.8e-005,
		4.8e-005,
		0.000236,
		0.000718,
		0.001685,
		0.003356,
		0.005972,
		0.00978,
		0.01503,
		0.021968,
		0.030824,
		0.041813,
		0.055126,
		0.070925,
		0.089337,
		0.110452,
		0.13432,
		0.160942,
		0.190271,
		0.222209,
		0.256601,
		0.293242,
		0.33187,
		0.372177,
		0.413805,
		0.45636,
		0.499416,
		0.542528,
		0.585242,
		0.627111,
		0.667704,
		0.706621,
		0.743507,
		0.778059,
		0.810035,
		0.839261,
		0.865629,
		0.889104,
		0.909715,
		0.927549,
		0.942748,
		0.955494,
		0.966004,
		0.974514,
		0.981272,
		0.986527,
		0.990522,
		0.993485,
		0.995622,
		0.997118,
		0.99813,
		0.998789,
		0.9992,
		0.999445,
		0.999584,
		0.999659,
		0.999698,
		0.999719,
		0.999732,
		0.999741,
		0.999749,
		0.999757,
		0.999764,
		0.999771,
		0.999778,
		0.999784,
		0.99979,
		0.999797,
		0.999802,
		0.999808,
		0.999813,
		0.999819,
		0.999824,
		0.999829,
		0.999833,
		0.999838,
		0.999842,
		0.999846,
		0.999851,
		0.999854,
		0.999858,
		0.999862,
		0.999866,
		0.999869,
		0.999872,
		0.999876,
		0.999879,
		0.999882,
		0.999885,
		0.999888,
		0.999891,
		0.999893,
		0.999896,
		0.999898,
		0.999901,
		0.999903,
		0.999906,
		0.999908,
		0.99991,
		0.999912,
		0.999914,
		0.999916,
		0.999918,
		0.99992,
		0.999922,
		0.999923,
		0.999925,
		0.999927,
		0.999929,
		0.99993,
		0.999932,
		0.999933,
		0.999935,
		0.999936,
		0.999937,
		0.999939,
		0.99994,
		0.999941,
		0.999943,
		0.999944,
		0.999945,
		0.999946,
		0.999947,
		0.999948,
		0.999949,
		0.99995,
		0.999951,
		0.999952,
		0.999953,
		0.999954,
		0.999955,
		0.999956,
		0.999957,
		0.999958,
		0.999959,
		0.999959,
		0.99996,
		0.999961,
		0.999962,
		0.999962,
		0.999963,
		0.999964,
		0.999964,
		0.999965,
		0.999966,
		0.999966,
		0.999967,
		0.999968,
		0.999968,
		0.999969,
		0.999969,
		0.99997,
		0.99997,
		0.999971,
		0.999971,
		0.999972,
		0.999972,
		0.999973,
		0.999973,
		0.999974,
		0.999974,
		0.999975,
		0.999975,
		0.999976,
		0.999976,
		0.999976,
		0.999977,
		0.999977,
		0.999978,
		0.999978,
		0.999978,
		0.999979,
		0.999979,
		0.999979,
		0.99998,
		0.99998,
		0.99998,
		0.999981,
		0.999981,
		0.999981,
		0.999982,
		0.999982,
		0.999982,
		0.999982,
		0.999983,
		0.999983,
		0.999983,
		0.999984,
		0.999984,
		0.999984,
		0.999984,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999952,
		0.999887,
		0.999758,
		0.999523,
		0.999123,
		0.998475,
		0.997474,
		0.99599,
		0.993865,
		0.990914,
		0.986931,
		0.981688,
		0.974941,
		0.966443,
		0.955945,
		0.943208,
		0.928019,
		0.910193,
		0.88959,
		0.86612,
		0.839755,
		0.810531,
		0.778554,
		0.743999,
		0.707107,
		0.668181,
		0.627578,
		0.585696,
		0.542966,
		0.499836,
		0.456761,
		0.414184,
		0.372533,
		0.332201,
		0.293547,
		0.25688,
		0.222461,
		0.190498,
		0.161143,
		0.134496,
		0.110603,
		0.089465,
		0.071031,
		0.055213,
		0.041882,
		0.030877,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001689,
		0.000719,
		0.000237,
		4.9e-005,
		4.9e-005,
		0.000237,
		0.000719,
		0.001689,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030877,
		0.041882,
		0.055213,
		0.071031,
		0.089465,
		0.110603,
		0.134496,
		0.161143,
		0.190498,
		0.222461,
		0.25688,
		0.293547,
		0.332201,
		0.372533,
		0.414184,
		0.456761,
		0.499836,
		0.542966,
		0.585696,
		0.627578,
		0.668181,
		0.707107,
		0.743999,
		0.778554,
		0.810531,
		0.839755,
		0.86612,
		0.88959,
		0.910193,
		0.928019,
		0.943208,
		0.955945,
		0.966443,
		0.974941,
		0.981688,
		0.986931,
		0.990914,
		0.993865,
		0.99599,
		0.997474,
		0.998475,
		0.999123,
		0.999523,
		0.999758,
		0.999887,
		0.999952,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999985,
		0.999984,
		0.999984,
		0.999984,
		0.999984,
		0.999983,
		0.999983,
		0.999983,
		0.999982,
		0.999982,
		0.999982,
		0.999982,
		0.999981,
		0.999981,
		0.999981,
		0.99998,
		0.99998,
		0.99998,
		0.999979,
		0.999979,
		0.999979,
		0.999978,
		0.999978,
		0.999978,
		0.999977,
		0.999977,
		0.999976,
		0.999976,
		0.999976,
		0.999975,
		0.999975,
		0.999974,
		0.999974,
		0.999973,
		0.999973,
		0.999972,
		0.999972,
		0.999971,
		0.999971,
		0.99997,
		0.99997,
		0.999969,
		0.999969,
		0.999968,
		0.999968,
		0.999967,
		0.999966,
		0.999966,
		0.999965,
		0.999964,
		0.999964,
		0.999963,
		0.999962,
		0.999962,
		0.999961,
		0.99996,
		0.999959,
		0.999959,
		0.999958,
		0.999957,
		0.999956,
		0.999955,
		0.999954,
		0.999953,
		0.999952,
		0.999951,
		0.99995,
		0.999949,
		0.999948,
		0.999947,
		0.999946,
		0.999945,
		0.999944,
		0.999943,
		0.999941,
		0.99994,
		0.999939,
		0.999937,
		0.999936,
		0.999935,
		0.999933,
		0.999932,
		0.99993,
		0.999929,
		0.999927,
		0.999925,
		0.999923,
		0.999922,
		0.99992,
		0.999918,
		0.999916,
		0.999914,
		0.999912,
		0.99991,
		0.999908,
		0.999906,
		0.999903,
		0.999901,
		0.999898,
		0.999896,
		0.999893,
		0.999891,
		0.999888,
		0.999885,
		0.999882,
		0.999879,
		0.999876,
		0.999872,
		0.999869,
		0.999866,
		0.999862,
		0.999858,
		0.999854,
		0.999851,
		0.999846,
		0.999842,
		0.999838,
		0.999833,
		0.999829,
		0.999824,
		0.999819,
		0.999813,
		0.999808,
		0.999802,
		0.999797,
		0.99979,
		0.999784,
		0.999778,
		0.999771,
		0.999764,
		0.999757,
		0.999749,
		0.999741,
		0.999732,
		0.999719,
		0.999698,
		0.999659,
		0.999584,
		0.999445,
		0.9992,
		0.998789,
		0.99813,
		0.997118,
		0.995622,
		0.993485,
		0.990522,
		0.986527,
		0.981272,
		0.974514,
		0.966004,
		0.955494,
		0.942748,
		0.927549,
		0.909715,
		0.889104,
		0.865629,
		0.839261,
		0.810035,
		0.778059,
		0.743507,
		0.706621,
		0.667704,
		0.627111,
		0.585242,
		0.542528,
		0.499416,
		0.45636,
		0.413805,
		0.372177,
		0.33187,
		0.293242,
		0.256601,
		0.222209,
		0.190271,
		0.160942,
		0.13432,
		0.110452,
		0.089337,
		0.070925,
		0.055126,
		0.041813,
		0.030824,
		0.021968,
		0.01503,
		0.00978,
		0.005972,
		0.003356,
		0.001685,
		0.000718,
		0.000236,
		4.8e-005,
		4.8e-005,
		0.000236,
		0.000718,
		0.001685,
		0.003356,
		0.005972,
		0.00978,
		0.01503,
		0.021968,
		0.030824,
		0.041813,
		0.055126,
		0.070925,
		0.089337,
		0.110452,
		0.13432,
		0.160942,
		0.190271,
		0.222209,
		0.256601,
		0.293242,
		0.33187,
		0.372177,
		0.413805,
		0.45636,
		0.499416,
		0.542528,
		0.585242,
		0.627111,
		0.667704,
		0.706621,
		0.743507,
		0.778059,
		0.810035,
		0.839261,
		0.865629,
		0.889104,
		0.909715,
		0.927549,
		0.942748,
		0.955494,
		0.966004,
		0.974514,
		0.981272,
		0.986527,
		0.990522,
		0.993485,
		0.995622,
		0.997118,
		0.99813,
		0.998789,
		0.9992,
		0.999445,
		0.999584,
		0.999659,
		0.999698,
		0.999719,
		0.999732,
		0.999741,
		0.999749,
		0.999757,
		0.999764,
		0.999771,
		0.999778,
		0.999784,
		0.99979,
		0.999797,
		0.999802,
		0.999808,
		0.999813,
		0.999819,
		0.999824,
		0.999829,
		0.999833,
		0.999838,
		0.999842,
		0.999846,
		0.999851,
		0.999854,
		0.999858,
		0.999862,
		0.999866,
		0.999869,
		0.999872,
		0.999876,
		0.999879,
		0.999882,
		0.999885,
		0.999888,
		0.999891,
		0.999893,
		0.999896,
		0.999898,
		0.999901,
		0.999903,
		0.999906,
		0.999908,
		0.99991,
		0.999912,
		0.999914,
		0.999916,
		0.999918,
		0.99992,
		0.999922,
		0.999923,
		0.999925,
		0.999927,
		0.999929,
		0.99993,
		0.999932,
		0.999933,
		0.999935,
		0.999936,
		0.999937,
		0.999939,
		0.99994,
		0.999941,
		0.999943,
		0.999944,
		0.999945,
		0.999946,
		0.999947,
		0.999948,
		0.999949,
		0.99995,
		0.999951,
		0.999952,
		0.999953,
		0.999954,
		0.999955,
		0.999956,
		0.999957,
		0.999958,
		0.999959,
		0.999959,
		0.99996,
		0.999961,
		0.999962,
		0.999962,
		0.999963,
		0.999964,
		0.999964,
		0.999965,
		0.999966,
		0.999966,
		0.999967,
		0.999968,
		0.999968,
		0.999969,
		0.999969,
		0.99997,
		0.99997,
		0.999971,
		0.999971,
		0.999972,
		0.999972,
		0.999973,
		0.999973,
		0.999974,
		0.999974,
		0.999975,
		0.999975,
		0.999976,
		0.999976,
		0.999976,
		0.999977,
		0.999977,
		0.999978,
		0.999978,
		0.999978,
		0.999979,
		0.999979,
		0.999979,
		0.99998,
		0.99998,
		0.99998,
		0.999981,
		0.999981,
		0.999981,
		0.999982,
		0.999982,
		0.999982,
		0.999982,
		0.999983,
		0.999983,
		0.999983,
		0.999984,
		0.999984,
		0.999984,
		0.999984,
		0.999771,
		0.999767,
		0.999764,
		0.99976,
		0.999757,
		0.999753,
		0.999749,
		0.999745,
		0.999741,
		0.999737,
		0.999733,
		0.999729,
		0.999725,
		0.99972,
		0.999716,
		0.999711,
		0.999707,
		0.999702,
		0.999697,
		0.999692,
		0.999687,
		0.999682,
		0.999677,
		0.999671,
		0.999666,
		0.99966,
		0.999655,
		0.999649,
		0.999643,
		0.999637,
		0.999631,
		0.999624,
		0.999618,
		0.999611,
		0.999605,
		0.999598,
		0.999591,
		0.999584,
		0.999576,
		0.999569,
		0.999561,
		0.999553,
		0.999545,
		0.999537,
		0.999529,
		0.99952,
		0.999511,
		0.999502,
		0.999493,
		0.999484,
		0.999474,
		0.999464,
		0.999454,
		0.999444,
		0.999434,
		0.999423,
		0.999412,
		0.9994,
		0.999389,
		0.999377,
		0.999365,
		0.999352,
		0.99934,
		0.999327,
		0.999313,
		0.999299,
		0.999285,
		0.999271,
		0.999256,
		0.999241,
		0.999225,
		0.99921,
		0.999193,
		0.999176,
		0.999159,
		0.999141,
		0.999123,
		0.999105,
		0.999086,
		0.999066,
		0.999046,
		0.999025,
		0.999004,
		0.998982,
		0.99896,
		0.998936,
		0.998913,
		0.998888,
		0.998863,
		0.998838,
		0.998811,
		0.998784,
		0.998756,
		0.998727,
		0.998697,
		0.998667,
		0.998635,
		0.998603,
		0.998569,
		0.998535,
		0.998499,
		0.998463,
		0.998425,
		0.998387,
		0.998347,
		0.998306,
		0.998263,
		0.998219,
		0.998174,
		0.998127,
		0.998079,
		0.998029,
		0.997978,
		0.997925,
		0.99787,
		0.997814,
		0.997755,
		0.997695,
		0.997633,
		0.997568,
		0.997501,
		0.997432,
		0.997361,
		0.997287,
		0.99721,
		0.997131,
		0.997049,
		0.996964,
		0.996876,
		0.996784,
		0.99669,
		0.996591,
		0.996489,
		0.996382,
		0.996268,
		0.996142,
		0.995994,
		0.995806,
		0.99555,
		0.995184,
		0.994648,
		0.993859,
		0.992715,
		0.991083,
		0.988808,
		0.985706,
		0.981571,
		0.976177,
		0.969283,
		0.960643,
		0.950009,
		0.937149,
		0.921848,
		0.903927,
		0.883247,
		0.859722,
		0.833327,
		0.8041,
		0.772148,
		0.737649,
		0.700845,
		0.662039,
		0.621588,
		0.57989,
		0.537373,
		0.494485,
		0.451676,
		0.409388,
		0.368043,
		0.328034,
		0.289712,
		0.253384,
		0.219306,
		0.18768,
		0.158656,
		0.13233,
		0.108744,
		0.087893,
		0.069727,
		0.054154,
		0.041042,
		0.030229,
		0.021524,
		0.014712,
		0.009563,
		0.005833,
		0.003275,
		0.001641,
		0.000698,
		0.000229,
		4.7e-005,
		4.7e-005,
		0.000229,
		0.000698,
		0.001641,
		0.003275,
		0.005833,
		0.009563,
		0.014712,
		0.021524,
		0.030229,
		0.041042,
		0.054154,
		0.069727,
		0.087893,
		0.108744,
		0.13233,
		0.158656,
		0.18768,
		0.219306,
		0.253384,
		0.289712,
		0.328034,
		0.368043,
		0.409388,
		0.451676,
		0.494485,
		0.537373,
		0.57989,
		0.621588,
		0.662039,
		0.700845,
		0.737649,
		0.772148,
		0.8041,
		0.833327,
		0.859722,
		0.883247,
		0.903927,
		0.921848,
		0.937149,
		0.950009,
		0.960643,
		0.969283,
		0.976177,
		0.981571,
		0.985706,
		0.988808,
		0.991083,
		0.992715,
		0.993859,
		0.994648,
		0.995184,
		0.99555,
		0.995806,
		0.995994,
		0.996142,
		0.996268,
		0.996382,
		0.996489,
		0.996591,
		0.99669,
		0.996784,
		0.996876,
		0.996964,
		0.997049,
		0.997131,
		0.99721,
		0.997287,
		0.997361,
		0.997432,
		0.997501,
		0.997568,
		0.997633,
		0.997695,
		0.997755,
		0.997814,
		0.99787,
		0.997925,
		0.997978,
		0.998029,
		0.998079,
		0.998127,
		0.998174,
		0.998219,
		0.998263,
		0.998306,
		0.998347,
		0.998387,
		0.998425,
		0.998463,
		0.998499,
		0.998535,
		0.998569,
		0.998603,
		0.998635,
		0.998667,
		0.998697,
		0.998727,
		0.998756,
		0.998784,
		0.998811,
		0.998838,
		0.998863,
		0.998888,
		0.998913,
		0.998936,
		0.99896,
		0.998982,
		0.999004,
		0.999025,
		0.999046,
		0.999066,
		0.999086,
		0.999105,
		0.999123,
		0.999141,
		0.999159,
		0.999176,
		0.999193,
		0.99921,
		0.999225,
		0.999241,
		0.999256,
		0.999271,
		0.999285,
		0.999299,
		0.999313,
		0.999327,
		0.99934,
		0.999352,
		0.999365,
		0.999377,
		0.999389,
		0.9994,
		0.999412,
		0.999423,
		0.999434,
		0.999444,
		0.999454,
		0.999464,
		0.999474,
		0.999484,
		0.999493,
		0.999502,
		0.999511,
		0.99952,
		0.999529,
		0.999537,
		0.999545,
		0.999553,
		0.999561,
		0.999569,
		0.999576,
		0.999584,
		0.999591,
		0.999598,
		0.999605,
		0.999611,
		0.999618,
		0.999624,
		0.999631,
		0.999637,
		0.999643,
		0.999649,
		0.999655,
		0.99966,
		0.999666,
		0.999671,
		0.999677,
		0.999682,
		0.999687,
		0.999692,
		0.999697,
		0.999702,
		0.999707,
		0.999711,
		0.999716,
		0.99972,
		0.999725,
		0.999729,
		0.999733,
		0.999737,
		0.999741,
		0.999745,
		0.999749,
		0.999753,
		0.999757,
		0.99976,
		0.999764,
		0.999767,
		0.998929,
		0.998913,
		0.998897,
		0.99888,
		0.998863,
		0.998846,
		0.998829,
		0.998811,
		0.998793,
		0.998774,
		0.998756,
		0.998737,
		0.998717,
		0.998697,
		0.998677,
		0.998656,
		0.998635,
		0.998614,
		0.998592,
		0.998569,
		0.998546,
		0.998523,
		0.998499,
		0.998475,
		0.998451,
		0.998425,
		0.9984,
		0.998373,
		0.998347,
		0.998319,
		0.998291,
		0.998263,
		0.998234,
		0.998204,
		0.998174,
		0.998143,
		0.998111,
		0.998079,
		0.998046,
		0.998013,
		0.997978,
		0.997943,
		0.997907,
		0.99787,
		0.997833,
		0.997795,
		0.997755,
		0.997715,
		0.997675,
		0.997633,
		0.99759,
		0.997546,
		0.997501,
		0.997456,
		0.997409,
		0.997361,
		0.997312,
		0.997262,
		0.99721,
		0.997158,
		0.997104,
		0.997049,
		0.996992,
		0.996935,
		0.996876,
		0.996815,
		0.996753,
		0.99669,
		0.996625,
		0.996558,
		0.99649,
		0.996419,
		0.996348,
		0.996274,
		0.996199,
		0.996121,
		0.996042,
		0.995961,
		0.995877,
		0.995792,
		0.995704,
		0.995614,
		0.995521,
		0.995426,
		0.995329,
		0.995229,
		0.995126,
		0.99502,
		0.994912,
		0.994801,
		0.994686,
		0.994569,
		0.994448,
		0.994324,
		0.994196,
		0.994065,
		0.99393,
		0.993791,
		0.993648,
		0.993502,
		0.99335,
		0.993195,
		0.993035,
		0.99287,
		0.9927,
		0.992526,
		0.992346,
		0.99216,
		0.991969,
		0.991772,
		0.991569,
		0.99136,
		0.991144,
		0.990922,
		0.990693,
		0.990456,
		0.990212,
		0.98996,
		0.9897,
		0.989431,
		0.989154,
		0.988867,
		0.988571,
		0.988265,
		0.987949,
		0.987623,
		0.987285,
		0.986936,
		0.986574,
		0.986201,
		0.985814,
		0.985413,
		0.984999,
		0.984568,
		0.984119,
		0.983646,
		0.983139,
		0.98258,
		0.981941,
		0.981178,
		0.980234,
		0.979026,
		0.977452,
		0.975383,
		0.972665,
		0.969117,
		0.964539,
		0.958709,
		0.951392,
		0.942348,
		0.931338,
		0.918136,
		0.902538,
		0.884372,
		0.863508,
		0.83987,
		0.813438,
		0.784259,
		0.752446,
		0.718179,
		0.681705,
		0.643327,
		0.603399,
		0.562317,
		0.520504,
		0.4784,
		0.436447,
		0.395076,
		0.354699,
		0.315694,
		0.278401,
		0.243113,
		0.210073,
		0.179472,
		0.151444,
		0.126076,
		0.103399,
		0.083398,
		0.066016,
		0.051152,
		0.038672,
		0.02841,
		0.020174,
		0.01375,
		0.00891,
		0.005418,
		0.003031,
		0.001514,
		0.000642,
		0.00021,
		4.3e-005,
		4.3e-005,
		0.00021,
		0.000642,
		0.001514,
		0.003031,
		0.005418,
		0.00891,
		0.01375,
		0.020174,
		0.02841,
		0.038672,
		0.051152,
		0.066016,
		0.083398,
		0.103399,
		0.126076,
		0.151444,
		0.179472,
		0.210073,
		0.243113,
		0.278401,
		0.315694,
		0.354699,
		0.395076,
		0.436447,
		0.4784,
		0.520504,
		0.562317,
		0.603399,
		0.643327,
		0.681705,
		0.718179,
		0.752446,
		0.784259,
		0.813438,
		0.83987,
		0.863508,
		0.884372,
		0.902538,
		0.918136,
		0.931338,
		0.942348,
		0.951392,
		0.958709,
		0.964539,
		0.969117,
		0.972665,
		0.975383,
		0.977452,
		0.979026,
		0.980234,
		0.981178,
		0.981941,
		0.98258,
		0.983139,
		0.983646,
		0.984119,
		0.984568,
		0.984999,
		0.985413,
		0.985814,
		0.986201,
		0.986574,
		0.986936,
		0.987285,
		0.987623,
		0.987949,
		0.988265,
		0.988571,
		0.988867,
		0.989154,
		0.989431,
		0.9897,
		0.98996,
		0.990212,
		0.990456,
		0.990693,
		0.990922,
		0.991144,
		0.99136,
		0.991569,
		0.991772,
		0.991969,
		0.99216,
		0.992346,
		0.992526,
		0.9927,
		0.99287,
		0.993035,
		0.993195,
		0.99335,
		0.993502,
		0.993648,
		0.993791,
		0.99393,
		0.994065,
		0.994196,
		0.994324,
		0.994448,
		0.994569,
		0.994686,
		0.994801,
		0.994912,
		0.99502,
		0.995126,
		0.995229,
		0.995329,
		0.995426,
		0.995521,
		0.995614,
		0.995704,
		0.995792,
		0.995877,
		0.995961,
		0.996042,
		0.996121,
		0.996199,
		0.996274,
		0.996348,
		0.996419,
		0.99649,
		0.996558,
		0.996625,
		0.99669,
		0.996753,
		0.996815,
		0.996876,
		0.996935,
		0.996992,
		0.997049,
		0.997104,
		0.997158,
		0.99721,
		0.997262,
		0.997312,
		0.997361,
		0.997409,
		0.997456,
		0.997501,
		0.997546,
		0.99759,
		0.997633,
		0.997675,
		0.997715,
		0.997755,
		0.997795,
		0.997833,
		0.99787,
		0.997907,
		0.997943,
		0.997978,
		0.998013,
		0.998046,
		0.998079,
		0.998111,
		0.998143,
		0.998174,
		0.998204,
		0.998234,
		0.998263,
		0.998291,
		0.998319,
		0.998347,
		0.998373,
		0.9984,
		0.998425,
		0.998451,
		0.998475,
		0.998499,
		0.998523,
		0.998546,
		0.998569,
		0.998592,
		0.998614,
		0.998635,
		0.998656,
		0.998677,
		0.998697,
		0.998717,
		0.998737,
		0.998756,
		0.998774,
		0.998793,
		0.998811,
		0.998829,
		0.998846,
		0.998863,
		0.99888,
		0.998897,
		0.998913,
		0.996876,
		0.99683,
		0.996784,
		0.996737,
		0.99669,
		0.996641,
		0.996591,
		0.996541,
		0.99649,
		0.996437,
		0.996384,
		0.996329,
		0.996274,
		0.996218,
		0.99616,
		0.996102,
		0.996042,
		0.995981,
		0.995919,
		0.995856,
		0.995792,
		0.995726,
		0.995659,
		0.995591,
		0.995521,
		0.99545,
		0.995378,
		0.995304,
		0.995229,
		0.995152,
		0.995074,
		0.994994,
		0.994912,
		0.994829,
		0.994744,
		0.994657,
		0.994569,
		0.994478,
		0.994386,
		0.994292,
		0.994196,
		0.994098,
		0.993998,
		0.993896,
		0.993791,
		0.993685,
		0.993576,
		0.993464,
		0.99335,
		0.993234,
		0.993115,
		0.992994,
		0.99287,
		0.992743,
		0.992614,
		0.992481,
		0.992346,
		0.992207,
		0.992065,
		0.99192,
		0.991772,
		0.991621,
		0.991466,
		0.991307,
		0.991144,
		0.990978,
		0.990808,
		0.990634,
		0.990456,
		0.990274,
		0.990087,
		0.989896,
		0.9897,
		0.989499,
		0.989293,
		0.989083,
		0.988867,
		0.988646,
		0.98842,
		0.988187,
		0.987949,
		0.987705,
		0.987455,
		0.987199,
		0.986936,
		0.986666,
		0.986389,
		0.986105,
		0.985814,
		0.985515,
		0.985208,
		0.984893,
		0.984569,
		0.984237,
		0.983896,
		0.983546,
		0.983187,
		0.982817,
		0.982438,
		0.982047,
		0.981647,
		0.981234,
		0.980811,
		0.980375,
		0.979927,
		0.979466,
		0.978993,
		0.978505,
		0.978003,
		0.977487,
		0.976956,
		0.976409,
		0.975845,
		0.975265,
		0.974668,
		0.974052,
		0.973418,
		0.972765,
		0.972091,
		0.971397,
		0.970681,
		0.969943,
		0.969182,
		0.968397,
		0.967586,
		0.96675,
		0.965887,
		0.964996,
		0.964076,
		0.963126,
		0.962145,
		0.961131,
		0.960083,
		0.958999,
		0.957876,
		0.956706,
		0.955481,
		0.95418,
		0.952776,
		0.951227,
		0.949475,
		0.94744,
		0.945023,
		0.942099,
		0.938519,
		0.934111,
		0.928683,
		0.922022,
		0.913908,
		0.904114,
		0.892416,
		0.878606,
		0.862495,
		0.84393,
		0.822798,
		0.799039,
		0.77265,
		0.743688,
		0.712278,
		0.678608,
		0.642924,
		0.605532,
		0.566781,
		0.527057,
		0.486772,
		0.446348,
		0.406208,
		0.366762,
		0.328396,
		0.291463,
		0.256277,
		0.223105,
		0.192162,
		0.163614,
		0.137572,
		0.1141,
		0.09321,
		0.07487,
		0.059008,
		0.045513,
		0.034243,
		0.025028,
		0.017677,
		0.01198,
		0.007717,
		0.004662,
		0.002591,
		0.001285,
		0.00054,
		0.000175,
		3.5e-005,
		3.5e-005,
		0.000175,
		0.00054,
		0.001285,
		0.002591,
		0.004662,
		0.007717,
		0.01198,
		0.017677,
		0.025028,
		0.034243,
		0.045513,
		0.059008,
		0.07487,
		0.09321,
		0.1141,
		0.137572,
		0.163614,
		0.192162,
		0.223105,
		0.256277,
		0.291463,
		0.328396,
		0.366762,
		0.406208,
		0.446348,
		0.486772,
		0.527057,
		0.566781,
		0.605532,
		0.642924,
		0.678608,
		0.712278,
		0.743688,
		0.77265,
		0.799039,
		0.822798,
		0.84393,
		0.862495,
		0.878606,
		0.892416,
		0.904114,
		0.913908,
		0.922022,
		0.928683,
		0.934111,
		0.938519,
		0.942099,
		0.945023,
		0.94744,
		0.949475,
		0.951227,
		0.952776,
		0.95418,
		0.955481,
		0.956706,
		0.957876,
		0.958999,
		0.960083,
		0.961131,
		0.962145,
		0.963126,
		0.964076,
		0.964996,
		0.965887,
		0.96675,
		0.967586,
		0.968397,
		0.969182,
		0.969943,
		0.970681,
		0.971397,
		0.972091,
		0.972765,
		0.973418,
		0.974052,
		0.974668,
		0.975265,
		0.975845,
		0.976409,
		0.976956,
		0.977487,
		0.978003,
		0.978505,
		0.978993,
		0.979466,
		0.979927,
		0.980375,
		0.980811,
		0.981234,
		0.981647,
		0.982047,
		0.982438,
		0.982817,
		0.983187,
		0.983546,
		0.983896,
		0.984237,
		0.984569,
		0.984893,
		0.985208,
		0.985515,
		0.985814,
		0.986105,
		0.986389,
		0.986666,
		0.986936,
		0.987199,
		0.987455,
		0.987705,
		0.987949,
		0.988187,
		0.98842,
		0.988646,
		0.988867,
		0.989083,
		0.989293,
		0.989499,
		0.9897,
		0.989896,
		0.990087,
		0.990274,
		0.990456,
		0.990634,
		0.990808,
		0.990978,
		0.991144,
		0.991307,
		0.991466,
		0.991621,
		0.991772,
		0.99192,
		0.992065,
		0.992207,
		0.992346,
		0.992481,
		0.992614,
		0.992743,
		0.99287,
		0.992994,
		0.993115,
		0.993234,
		0.99335,
		0.993464,
		0.993576,
		0.993685,
		0.993791,
		0.993896,
		0.993998,
		0.994098,
		0.994196,
		0.994292,
		0.994386,
		0.994478,
		0.994569,
		0.994657,
		0.994744,
		0.994829,
		0.994912,
		0.994994,
		0.995074,
		0.995152,
		0.995229,
		0.995304,
		0.995378,
		0.99545,
		0.995521,
		0.995591,
		0.995659,
		0.995726,
		0.995792,
		0.995856,
		0.995919,
		0.995981,
		0.996042,
		0.996102,
		0.99616,
		0.996218,
		0.996274,
		0.996329,
		0.996384,
		0.996437,
		0.99649,
		0.996541,
		0.996591,
		0.996641,
		0.99669,
		0.996737,
		0.996784,
		0.99683,
		0.992969,
		0.99287,
		0.992769,
		0.992666,
		0.992561,
		0.992454,
		0.992346,
		0.992235,
		0.992122,
		0.992008,
		0.991891,
		0.991772,
		0.991651,
		0.991528,
		0.991403,
		0.991275,
		0.991144,
		0.991012,
		0.990877,
		0.990739,
		0.990599,
		0.990456,
		0.99031,
		0.990162,
		0.990011,
		0.989857,
		0.9897,
		0.98954,
		0.989376,
		0.98921,
		0.98904,
		0.988867,
		0.988691,
		0.988511,
		0.988327,
		0.98814,
		0.987949,
		0.987755,
		0.987556,
		0.987353,
		0.987147,
		0.986936,
		0.98672,
		0.986501,
		0.986276,
		0.986047,
		0.985814,
		0.985575,
		0.985332,
		0.985083,
		0.984829,
		0.984569,
		0.984305,
		0.984034,
		0.983758,
		0.983475,
		0.983187,
		0.982892,
		0.982591,
		0.982283,
		0.981968,
		0.981647,
		0.981318,
		0.980982,
		0.980638,
		0.980287,
		0.979927,
		0.97956,
		0.979184,
		0.978799,
		0.978406,
		0.978003,
		0.977591,
		0.97717,
		0.976739,
		0.976297,
		0.975845,
		0.975383,
		0.974909,
		0.974424,
		0.973927,
		0.973418,
		0.972897,
		0.972363,
		0.971816,
		0.971256,
		0.970681,
		0.970093,
		0.969489,
		0.968871,
		0.968237,
		0.967586,
		0.96692,
		0.966236,
		0.965534,
		0.964815,
		0.964076,
		0.963319,
		0.962541,
		0.961743,
		0.960924,
		0.960083,
		0.95922,
		0.958333,
		0.957423,
		0.956487,
		0.955526,
		0.954539,
		0.953525,
		0.952482,
		0.95141,
		0.950308,
		0.949176,
		0.948011,
		0.946813,
		0.94558,
		0.944313,
		0.943008,
		0.941666,
		0.940284,
		0.938862,
		0.937398,
		0.93589,
		0.934338,
		0.932739,
		0.931091,
		0.929393,
		0.927644,
		0.925841,
		0.923982,
		0.922066,
		0.92009,
		0.918052,
		0.915948,
		0.913775,
		0.911526,
		0.909189,
		0.906745,
		0.904167,
		0.901414,
		0.898429,
		0.895138,
		0.891446,
		0.887236,
		0.882368,
		0.876682,
		0.870001,
		0.862132,
		0.852871,
		0.842016,
		0.829368,
		0.814743,
		0.797981,
		0.778956,
		0.757581,
		0.733821,
		0.707694,
		0.679276,
		0.648705,
		0.616177,
		0.581942,
		0.546299,
		0.509589,
		0.47218,
		0.434461,
		0.396826,
		0.359665,
		0.32335,
		0.288229,
		0.254614,
		0.222774,
		0.192937,
		0.165276,
		0.13992,
		0.116944,
		0.096378,
		0.078208,
		0.062379,
		0.048799,
		0.037343,
		0.027863,
		0.020186,
		0.014124,
		0.009478,
		0.006041,
		0.003609,
		0.001981,
		0.00097,
		0.000402,
		0.000129,
		2.6e-005,
		2.6e-005,
		0.000129,
		0.000402,
		0.00097,
		0.001981,
		0.003609,
		0.006041,
		0.009478,
		0.014124,
		0.020186,
		0.027863,
		0.037343,
		0.048799,
		0.062379,
		0.078208,
		0.096378,
		0.116944,
		0.13992,
		0.165276,
		0.192937,
		0.222774,
		0.254614,
		0.288229,
		0.32335,
		0.359665,
		0.396826,
		0.434461,
		0.47218,
		0.509589,
		0.546299,
		0.581942,
		0.616177,
		0.648705,
		0.679276,
		0.707694,
		0.733821,
		0.757581,
		0.778956,
		0.797981,
		0.814743,
		0.829368,
		0.842016,
		0.852871,
		0.862132,
		0.870001,
		0.876682,
		0.882368,
		0.887236,
		0.891446,
		0.895138,
		0.898429,
		0.901414,
		0.904167,
		0.906745,
		0.909189,
		0.911526,
		0.913775,
		0.915948,
		0.918052,
		0.92009,
		0.922066,
		0.923982,
		0.925841,
		0.927644,
		0.929393,
		0.931091,
		0.932739,
		0.934338,
		0.93589,
		0.937398,
		0.938862,
		0.940284,
		0.941666,
		0.943008,
		0.944313,
		0.94558,
		0.946813,
		0.948011,
		0.949176,
		0.950308,
		0.95141,
		0.952482,
		0.953525,
		0.954539,
		0.955526,
		0.956487,
		0.957423,
		0.958333,
		0.95922,
		0.960083,
		0.960924,
		0.961743,
		0.962541,
		0.963319,
		0.964076,
		0.964815,
		0.965534,
		0.966236,
		0.96692,
		0.967586,
		0.968237,
		0.968871,
		0.969489,
		0.970093,
		0.970681,
		0.971256,
		0.971816,
		0.972363,
		0.972897,
		0.973418,
		0.973927,
		0.974424,
		0.974909,
		0.975383,
		0.975845,
		0.976297,
		0.976739,
		0.97717,
		0.977591,
		0.978003,
		0.978406,
		0.978799,
		0.979184,
		0.97956,
		0.979927,
		0.980287,
		0.980638,
		0.980982,
		0.981318,
		0.981647,
		0.981968,
		0.982283,
		0.982591,
		0.982892,
		0.983187,
		0.983475,
		0.983758,
		0.984034,
		0.984305,
		0.984569,
		0.984829,
		0.985083,
		0.985332,
		0.985575,
		0.985814,
		0.986047,
		0.986276,
		0.986501,
		0.98672,
		0.986936,
		0.987147,
		0.987353,
		0.987556,
		0.987755,
		0.987949,
		0.98814,
		0.988327,
		0.988511,
		0.988691,
		0.988867,
		0.98904,
		0.98921,
		0.989376,
		0.98954,
		0.9897,
		0.989857,
		0.990011,
		0.990162,
		0.99031,
		0.990456,
		0.990599,
		0.990739,
		0.990877,
		0.991012,
		0.991144,
		0.991275,
		0.991403,
		0.991528,
		0.991651,
		0.991772,
		0.991891,
		0.992008,
		0.992122,
		0.992235,
		0.992346,
		0.992454,
		0.992561,
		0.992666,
		0.992769,
		0.99287,
		0.986574,
		0.986389,
		0.986201,
		0.986009,
		0.985814,
		0.985615,
		0.985413,
		0.985208,
		0.984999,
		0.984786,
		0.984569,
		0.984349,
		0.984125,
		0.983896,
		0.983664,
		0.983428,
		0.983187,
		0.982941,
		0.982692,
		0.982438,
		0.982179,
		0.981915,
		0.981647,
		0.981373,
		0.981094,
		0.980811,
		0.980522,
		0.980227,
		0.979927,
		0.979622,
		0.97931,
		0.978993,
		0.978669,
		0.978339,
		0.978003,
		0.977661,
		0.977312,
		0.976956,
		0.976593,
		0.976223,
		0.975845,
		0.97546,
		0.975068,
		0.974668,
		0.97426,
		0.973843,
		0.973418,
		0.972985,
		0.972543,
		0.972091,
		0.971631,
		0.971161,
		0.970681,
		0.970192,
		0.969692,
		0.969182,
		0.968661,
		0.968129,
		0.967586,
		0.967032,
		0.966466,
		0.965887,
		0.965297,
		0.964693,
		0.964076,
		0.963446,
		0.962803,
		0.962145,
		0.961473,
		0.960786,
		0.960083,
		0.959365,
		0.958631,
		0.957881,
		0.957114,
		0.956329,
		0.955526,
		0.954706,
		0.953866,
		0.953007,
		0.952128,
		0.951229,
		0.950308,
		0.949367,
		0.948403,
		0.947416,
		0.946406,
		0.945372,
		0.944313,
		0.943228,
		0.942118,
		0.94098,
		0.939815,
		0.938621,
		0.937398,
		0.936145,
		0.93486,
		0.933544,
		0.932195,
		0.930812,
		0.929393,
		0.927939,
		0.926448,
		0.924919,
		0.92335,
		0.921741,
		0.92009,
		0.918396,
		0.916658,
		0.914874,
		0.913042,
		0.911162,
		0.909232,
		0.90725,
		0.905215,
		0.903125,
		0.900978,
		0.898772,
		0.896506,
		0.894177,
		0.891784,
		0.889324,
		0.886795,
		0.884195,
		0.881522,
		0.878773,
		0.875946,
		0.873037,
		0.870045,
		0.866967,
		0.863799,
		0.860539,
		0.857183,
		0.853728,
		0.850168,
		0.846495,
		0.842698,
		0.838759,
		0.83465,
		0.830333,
		0.825756,
		0.820849,
		0.815526,
		0.809681,
		0.803188,
		0.795905,
		0.787675,
		0.778329,
		0.767694,
		0.755596,
		0.741871,
		0.726371,
		0.708972,
		0.689582,
		0.66815,
		0.644671,
		0.61919,
		0.591806,
		0.56267,
		0.531985,
		0.5,
		0.467005,
		0.433319,
		0.399286,
		0.365257,
		0.331586,
		0.298613,
		0.26666,
		0.236018,
		0.206942,
		0.179646,
		0.1543,
		0.131026,
		0.109901,
		0.090958,
		0.074187,
		0.05954,
		0.046936,
		0.036262,
		0.027382,
		0.020142,
		0.014372,
		0.009893,
		0.006523,
		0.00408,
		0.002388,
		0.001282,
		0.000613,
		0.000248,
		7.7e-005,
		1.5e-005,
		1.5e-005,
		7.7e-005,
		0.000248,
		0.000613,
		0.001282,
		0.002388,
		0.00408,
		0.006523,
		0.009893,
		0.014372,
		0.020142,
		0.027382,
		0.036262,
		0.046936,
		0.05954,
		0.074187,
		0.090958,
		0.109901,
		0.131026,
		0.1543,
		0.179646,
		0.206942,
		0.236018,
		0.26666,
		0.298613,
		0.331586,
		0.365257,
		0.399286,
		0.433319,
		0.467005,
		0.5,
		0.531985,
		0.56267,
		0.591806,
		0.61919,
		0.644671,
		0.66815,
		0.689582,
		0.708972,
		0.726371,
		0.741871,
		0.755596,
		0.767694,
		0.778329,
		0.787675,
		0.795905,
		0.803188,
		0.809681,
		0.815526,
		0.820849,
		0.825756,
		0.830333,
		0.83465,
		0.838759,
		0.842698,
		0.846495,
		0.850168,
		0.853728,
		0.857183,
		0.860539,
		0.863799,
		0.866967,
		0.870045,
		0.873037,
		0.875946,
		0.878773,
		0.881522,
		0.884195,
		0.886795,
		0.889324,
		0.891784,
		0.894177,
		0.896506,
		0.898772,
		0.900978,
		0.903125,
		0.905215,
		0.90725,
		0.909232,
		0.911162,
		0.913042,
		0.914874,
		0.916658,
		0.918396,
		0.92009,
		0.921741,
		0.92335,
		0.924919,
		0.926448,
		0.927939,
		0.929393,
		0.930812,
		0.932195,
		0.933544,
		0.93486,
		0.936145,
		0.937398,
		0.938621,
		0.939815,
		0.94098,
		0.942118,
		0.943228,
		0.944313,
		0.945372,
		0.946406,
		0.947416,
		0.948403,
		0.949367,
		0.950308,
		0.951229,
		0.952128,
		0.953007,
		0.953866,
		0.954706,
		0.955526,
		0.956329,
		0.957114,
		0.957881,
		0.958631,
		0.959365,
		0.960083,
		0.960786,
		0.961473,
		0.962145,
		0.962803,
		0.963446,
		0.964076,
		0.964693,
		0.965297,
		0.965887,
		0.966466,
		0.967032,
		0.967586,
		0.968129,
		0.968661,
		0.969182,
		0.969692,
		0.970192,
		0.970681,
		0.971161,
		0.971631,
		0.972091,
		0.972543,
		0.972985,
		0.973418,
		0.973843,
		0.97426,
		0.974668,
		0.975068,
		0.97546,
		0.975845,
		0.976223,
		0.976593,
		0.976956,
		0.977312,
		0.977661,
		0.978003,
		0.978339,
		0.978669,
		0.978993,
		0.97931,
		0.979622,
		0.979927,
		0.980227,
		0.980522,
		0.980811,
		0.981094,
		0.981373,
		0.981647,
		0.981915,
		0.982179,
		0.982438,
		0.982692,
		0.982941,
		0.983187,
		0.983428,
		0.983664,
		0.983896,
		0.984125,
		0.984349,
		0.984569,
		0.984786,
		0.984999,
		0.985208,
		0.985413,
		0.985615,
		0.985814,
		0.986009,
		0.986201,
		0.986389,
		0.977109,
		0.976801,
		0.976488,
		0.976169,
		0.975845,
		0.975516,
		0.975181,
		0.97484,
		0.974494,
		0.974141,
		0.973783,
		0.973418,
		0.973047,
		0.97267,
		0.972286,
		0.971895,
		0.971498,
		0.971093,
		0.970681,
		0.970262,
		0.969836,
		0.969402,
		0.96896,
		0.96851,
		0.968053,
		0.967586,
		0.967112,
		0.966629,
		0.966137,
		0.965636,
		0.965125,
		0.964606,
		0.964076,
		0.963537,
		0.962988,
		0.962429,
		0.961859,
		0.961278,
		0.960686,
		0.960083,
		0.959469,
		0.958843,
		0.958205,
		0.957554,
		0.956891,
		0.956215,
		0.955526,
		0.954824,
		0.954108,
		0.953377,
		0.952633,
		0.951873,
		0.951098,
		0.950308,
		0.949502,
		0.94868,
		0.947842,
		0.946986,
		0.946113,
		0.945222,
		0.944313,
		0.943385,
		0.942438,
		0.941471,
		0.940484,
		0.939477,
		0.938448,
		0.937398,
		0.936326,
		0.935231,
		0.934112,
		0.93297,
		0.931803,
		0.930611,
		0.929393,
		0.928149,
		0.926878,
		0.925579,
		0.924251,
		0.922895,
		0.921508,
		0.92009,
		0.918641,
		0.917159,
		0.915644,
		0.914095,
		0.91251,
		0.91089,
		0.909232,
		0.907537,
		0.905802,
		0.904028,
		0.902212,
		0.900354,
		0.898452,
		0.896506,
		0.894513,
		0.892474,
		0.890386,
		0.888248,
		0.886059,
		0.883818,
		0.881522,
		0.87917,
		0.876762,
		0.874294,
		0.871765,
		0.869175,
		0.86652,
		0.863799,
		0.861011,
		0.858152,
		0.855222,
		0.852217,
		0.849137,
		0.845978,
		0.842738,
		0.839416,
		0.836008,
		0.832511,
		0.828924,
		0.825244,
		0.821467,
		0.817592,
		0.813614,
		0.809531,
		0.80534,
		0.801037,
		0.796619,
		0.792083,
		0.787425,
		0.782641,
		0.777728,
		0.77268,
		0.767493,
		0.762157,
		0.756663,
		0.750991,
		0.745119,
		0.739009,
		0.732617,
		0.72588,
		0.718723,
		0.711055,
		0.702768,
		0.693744,
		0.683852,
		0.672954,
		0.660911,
		0.64759,
		0.632867,
		0.616635,
		0.598814,
		0.579355,
		0.558244,
		0.535513,
		0.511233,
		0.485526,
		0.458554,
		0.430524,
		0.401677,
		0.372282,
		0.342631,
		0.313028,
		0.283774,
		0.255168,
		0.227488,
		0.200987,
		0.175889,
		0.152377,
		0.130598,
		0.110655,
		0.092608,
		0.07648,
		0.062253,
		0.049875,
		0.039266,
		0.030318,
		0.022903,
		0.016879,
		0.012091,
		0.008381,
		0.005589,
		0.003558,
		0.002141,
		0.0012,
		0.000614,
		0.000278,
		0.000106,
		3.1e-005,
		3.1e-005,
		0.000106,
		0.000278,
		0.000614,
		0.0012,
		0.002141,
		0.003558,
		0.005589,
		0.008381,
		0.012091,
		0.016879,
		0.022903,
		0.030318,
		0.039266,
		0.049875,
		0.062253,
		0.07648,
		0.092608,
		0.110655,
		0.130598,
		0.152377,
		0.175889,
		0.200987,
		0.227488,
		0.255168,
		0.283774,
		0.313028,
		0.342631,
		0.372282,
		0.401677,
		0.430524,
		0.458554,
		0.485526,
		0.511233,
		0.535513,
		0.558244,
		0.579355,
		0.598814,
		0.616635,
		0.632867,
		0.64759,
		0.660911,
		0.672954,
		0.683852,
		0.693744,
		0.702768,
		0.711055,
		0.718723,
		0.72588,
		0.732617,
		0.739009,
		0.745119,
		0.750991,
		0.756663,
		0.762157,
		0.767493,
		0.77268,
		0.777728,
		0.782641,
		0.787425,
		0.792083,
		0.796619,
		0.801037,
		0.80534,
		0.809531,
		0.813614,
		0.817592,
		0.821467,
		0.825244,
		0.828924,
		0.832511,
		0.836008,
		0.839416,
		0.842738,
		0.845978,
		0.849137,
		0.852217,
		0.855222,
		0.858152,
		0.861011,
		0.863799,
		0.86652,
		0.869175,
		0.871765,
		0.874294,
		0.876762,
		0.87917,
		0.881522,
		0.883818,
		0.886059,
		0.888248,
		0.890386,
		0.892474,
		0.894513,
		0.896506,
		0.898452,
		0.900354,
		0.902212,
		0.904028,
		0.905802,
		0.907537,
		0.909232,
		0.91089,
		0.91251,
		0.914095,
		0.915644,
		0.917159,
		0.918641,
		0.92009,
		0.921508,
		0.922895,
		0.924251,
		0.925579,
		0.926878,
		0.928149,
		0.929393,
		0.930611,
		0.931803,
		0.93297,
		0.934112,
		0.935231,
		0.936326,
		0.937398,
		0.938448,
		0.939477,
		0.940484,
		0.941471,
		0.942438,
		0.943385,
		0.944313,
		0.945222,
		0.946113,
		0.946986,
		0.947842,
		0.94868,
		0.949502,
		0.950308,
		0.951098,
		0.951873,
		0.952633,
		0.953377,
		0.954108,
		0.954824,
		0.955526,
		0.956215,
		0.956891,
		0.957554,
		0.958205,
		0.958843,
		0.959469,
		0.960083,
		0.960686,
		0.961278,
		0.961859,
		0.962429,
		0.962988,
		0.963537,
		0.964076,
		0.964606,
		0.965125,
		0.965636,
		0.966137,
		0.966629,
		0.967112,
		0.967586,
		0.968053,
		0.96851,
		0.96896,
		0.969402,
		0.969836,
		0.970262,
		0.970681,
		0.971093,
		0.971498,
		0.971895,
		0.972286,
		0.97267,
		0.973047,
		0.973418,
		0.973783,
		0.974141,
		0.974494,
		0.97484,
		0.975181,
		0.975516,
		0.975845,
		0.976169,
		0.976488,
		0.976801,
		0.964076,
		0.963605,
		0.963126,
		0.96264,
		0.962145,
		0.961642,
		0.961131,
		0.960612,
		0.960083,
		0.959546,
		0.959001,
		0.958445,
		0.957881,
		0.957307,
		0.956724,
		0.95613,
		0.955526,
		0.954912,
		0.954288,
		0.953653,
		0.953007,
		0.95235,
		0.951681,
		0.951001,
		0.950308,
		0.949604,
		0.948887,
		0.948158,
		0.947416,
		0.946661,
		0.945892,
		0.945109,
		0.944313,
		0.943502,
		0.942676,
		0.941836,
		0.94098,
		0.940109,
		0.939222,
		0.938318,
		0.937398,
		0.936461,
		0.935507,
		0.934534,
		0.933544,
		0.932535,
		0.931507,
		0.93046,
		0.929393,
		0.928306,
		0.927198,
		0.926069,
		0.924919,
		0.923746,
		0.922551,
		0.921332,
		0.92009,
		0.918824,
		0.917532,
		0.916216,
		0.914874,
		0.913505,
		0.912108,
		0.910685,
		0.909232,
		0.907751,
		0.90624,
		0.904698,
		0.903125,
		0.90152,
		0.899882,
		0.898211,
		0.896506,
		0.894765,
		0.892988,
		0.891175,
		0.889324,
		0.887434,
		0.885504,
		0.883534,
		0.881522,
		0.879467,
		0.877369,
		0.875226,
		0.873037,
		0.870801,
		0.868517,
		0.866184,
		0.863799,
		0.861363,
		0.858873,
		0.856329,
		0.853729,
		0.851071,
		0.848355,
		0.845578,
		0.842738,
		0.839836,
		0.836868,
		0.833833,
		0.830729,
		0.827555,
		0.824309,
		0.820988,
		0.817592,
		0.814117,
		0.810562,
		0.806924,
		0.803203,
		0.799394,
		0.795497,
		0.791508,
		0.787425,
		0.783246,
		0.778969,
		0.77459,
		0.770107,
		0.765517,
		0.760817,
		0.756005,
		0.751077,
		0.74603,
		0.740862,
		0.735568,
		0.730146,
		0.724593,
		0.718904,
		0.713077,
		0.707107,
		0.700991,
		0.694725,
		0.688306,
		0.681729,
		0.674989,
		0.668082,
		0.661,
		0.653732,
		0.646263,
		0.638572,
		0.630629,
		0.622395,
		0.61382,
		0.604841,
		0.595385,
		0.585368,
		0.574699,
		0.563276,
		0.551001,
		0.537774,
		0.523504,
		0.508113,
		0.491542,
		0.473757,
		0.454752,
		0.434554,
		0.413227,
		0.390871,
		0.367622,
		0.34365,
		0.319156,
		0.294361,
		0.269507,
		0.244843,
		0.220619,
		0.197077,
		0.174444,
		0.152922,
		0.132688,
		0.113883,
		0.096614,
		0.08095,
		0.066924,
		0.054533,
		0.043741,
		0.034483,
		0.026672,
		0.020197,
		0.014934,
		0.010748,
		0.007501,
		0.005051,
		0.003262,
		0.002005,
		0.00116,
		0.000623,
		0.000304,
		0.000131,
		4.8e-005,
		1.4e-005,
		1.4e-005,
		4.8e-005,
		0.000131,
		0.000304,
		0.000623,
		0.00116,
		0.002005,
		0.003262,
		0.005051,
		0.007501,
		0.010748,
		0.014934,
		0.020197,
		0.026672,
		0.034483,
		0.043741,
		0.054533,
		0.066924,
		0.08095,
		0.096614,
		0.113883,
		0.132688,
		0.152922,
		0.174444,
		0.197077,
		0.220619,
		0.244843,
		0.269507,
		0.294361,
		0.319156,
		0.34365,
		0.367622,
		0.390871,
		0.413227,
		0.434554,
		0.454752,
		0.473757,
		0.491542,
		0.508113,
		0.523504,
		0.537774,
		0.551001,
		0.563276,
		0.574699,
		0.585368,
		0.595385,
		0.604841,
		0.61382,
		0.622395,
		0.630629,
		0.638572,
		0.646263,
		0.653732,
		0.661,
		0.668082,
		0.674989,
		0.681729,
		0.688306,
		0.694725,
		0.700991,
		0.707107,
		0.713077,
		0.718904,
		0.724593,
		0.730146,
		0.735568,
		0.740862,
		0.74603,
		0.751077,
		0.756005,
		0.760817,
		0.765517,
		0.770107,
		0.77459,
		0.778969,
		0.783246,
		0.787425,
		0.791508,
		0.795497,
		0.799394,
		0.803203,
		0.806924,
		0.810562,
		0.814117,
		0.817592,
		0.820988,
		0.824309,
		0.827555,
		0.830729,
		0.833833,
		0.836868,
		0.839836,
		0.842738,
		0.845578,
		0.848355,
		0.851071,
		0.853729,
		0.856329,
		0.858873,
		0.861363,
		0.863799,
		0.866184,
		0.868517,
		0.870801,
		0.873037,
		0.875226,
		0.877369,
		0.879467,
		0.881522,
		0.883534,
		0.885504,
		0.887434,
		0.889324,
		0.891175,
		0.892988,
		0.894765,
		0.896506,
		0.898211,
		0.899882,
		0.90152,
		0.903125,
		0.904698,
		0.90624,
		0.907751,
		0.909232,
		0.910685,
		0.912108,
		0.913505,
		0.914874,
		0.916216,
		0.917532,
		0.918824,
		0.92009,
		0.921332,
		0.922551,
		0.923746,
		0.924919,
		0.926069,
		0.927198,
		0.928306,
		0.929393,
		0.93046,
		0.931507,
		0.932535,
		0.933544,
		0.934534,
		0.935507,
		0.936461,
		0.937398,
		0.938318,
		0.939222,
		0.940109,
		0.94098,
		0.941836,
		0.942676,
		0.943502,
		0.944313,
		0.945109,
		0.945892,
		0.946661,
		0.947416,
		0.948158,
		0.948887,
		0.949604,
		0.950308,
		0.951001,
		0.951681,
		0.95235,
		0.953007,
		0.953653,
		0.954288,
		0.954912,
		0.955526,
		0.95613,
		0.956724,
		0.957307,
		0.957881,
		0.958445,
		0.959001,
		0.959546,
		0.960083,
		0.960612,
		0.961131,
		0.961642,
		0.962145,
		0.96264,
		0.963126,
		0.963605,
		0.947082,
		0.946406,
		0.945719,
		0.945021,
		0.944313,
		0.943593,
		0.942861,
		0.942118,
		0.941362,
		0.940595,
		0.939815,
		0.939022,
		0.938217,
		0.937398,
		0.936566,
		0.93572,
		0.93486,
		0.933986,
		0.933098,
		0.932195,
		0.931276,
		0.930343,
		0.929393,
		0.928428,
		0.927446,
		0.926448,
		0.925433,
		0.9244,
		0.92335,
		0.922282,
		0.921195,
		0.92009,
		0.918966,
		0.917822,
		0.916658,
		0.915473,
		0.914268,
		0.913042,
		0.911794,
		0.910525,
		0.909232,
		0.907917,
		0.906578,
		0.905215,
		0.903828,
		0.902416,
		0.900978,
		0.899514,
		0.898023,
		0.896506,
		0.89496,
		0.893386,
		0.891784,
		0.890151,
		0.888488,
		0.886795,
		0.88507,
		0.883312,
		0.881522,
		0.879698,
		0.877839,
		0.875946,
		0.874016,
		0.872049,
		0.870045,
		0.868003,
		0.865921,
		0.863799,
		0.861636,
		0.859431,
		0.857183,
		0.854892,
		0.852555,
		0.850172,
		0.847743,
		0.845265,
		0.842738,
		0.840161,
		0.837533,
		0.834852,
		0.832117,
		0.829327,
		0.826481,
		0.823577,
		0.820615,
		0.817592,
		0.814507,
		0.811359,
		0.808146,
		0.804867,
		0.801521,
		0.798105,
		0.794618,
		0.791059,
		0.787425,
		0.783716,
		0.779928,
		0.776061,
		0.772112,
		0.76808,
		0.763962,
		0.759757,
		0.755463,
		0.751077,
		0.746597,
		0.742021,
		0.737347,
		0.732572,
		0.727694,
		0.722712,
		0.717621,
		0.71242,
		0.707107,
		0.701678,
		0.696131,
		0.690463,
		0.684672,
		0.678754,
		0.672708,
		0.666529,
		0.660215,
		0.653763,
		0.647171,
		0.640434,
		0.63355,
		0.626516,
		0.619329,
		0.611985,
		0.604481,
		0.596815,
		0.588982,
		0.58098,
		0.572806,
		0.564456,
		0.555925,
		0.547208,
		0.538297,
		0.529181,
		0.519843,
		0.51026,
		0.500403,
		0.490235,
		0.479711,
		0.468778,
		0.457378,
		0.445448,
		0.432925,
		0.419747,
		0.405858,
		0.391212,
		0.375779,
		0.359546,
		0.342523,
		0.324747,
		0.30628,
		0.287213,
		0.267665,
		0.247779,
		0.227721,
		0.207672,
		0.187825,
		0.168377,
		0.14952,
		0.131439,
		0.1143,
		0.09825,
		0.083406,
		0.069857,
		0.057659,
		0.046838,
		0.037384,
		0.029263,
		0.022412,
		0.016747,
		0.012164,
		0.008548,
		0.005777,
		0.003723,
		0.002261,
		0.001272,
		0.000644,
		0.000278,
		9.2e-005,
		1.6e-005,
		1.6e-005,
		9.2e-005,
		0.000278,
		0.000644,
		0.001272,
		0.002261,
		0.003723,
		0.005777,
		0.008548,
		0.012164,
		0.016747,
		0.022412,
		0.029263,
		0.037384,
		0.046838,
		0.057659,
		0.069857,
		0.083406,
		0.09825,
		0.1143,
		0.131439,
		0.14952,
		0.168377,
		0.187825,
		0.207672,
		0.227721,
		0.247779,
		0.267665,
		0.287213,
		0.30628,
		0.324747,
		0.342523,
		0.359546,
		0.375779,
		0.391212,
		0.405858,
		0.419747,
		0.432925,
		0.445448,
		0.457378,
		0.468778,
		0.479711,
		0.490235,
		0.500403,
		0.51026,
		0.519843,
		0.529181,
		0.538297,
		0.547208,
		0.555925,
		0.564456,
		0.572806,
		0.58098,
		0.588982,
		0.596815,
		0.604481,
		0.611985,
		0.619329,
		0.626516,
		0.63355,
		0.640434,
		0.647171,
		0.653763,
		0.660215,
		0.666529,
		0.672708,
		0.678754,
		0.684672,
		0.690463,
		0.696131,
		0.701678,
		0.707107,
		0.71242,
		0.717621,
		0.722712,
		0.727694,
		0.732572,
		0.737347,
		0.742021,
		0.746597,
		0.751077,
		0.755463,
		0.759757,
		0.763962,
		0.76808,
		0.772112,
		0.776061,
		0.779928,
		0.783716,
		0.787425,
		0.791059,
		0.794618,
		0.798105,
		0.801521,
		0.804867,
		0.808146,
		0.811359,
		0.814507,
		0.817592,
		0.820615,
		0.823577,
		0.826481,
		0.829327,
		0.832117,
		0.834852,
		0.837533,
		0.840161,
		0.842738,
		0.845265,
		0.847743,
		0.850172,
		0.852555,
		0.854892,
		0.857183,
		0.859431,
		0.861636,
		0.863799,
		0.865921,
		0.868003,
		0.870045,
		0.872049,
		0.874016,
		0.875946,
		0.877839,
		0.879698,
		0.881522,
		0.883312,
		0.88507,
		0.886795,
		0.888488,
		0.890151,
		0.891784,
		0.893386,
		0.89496,
		0.896506,
		0.898023,
		0.899514,
		0.900978,
		0.902416,
		0.903828,
		0.905215,
		0.906578,
		0.907917,
		0.909232,
		0.910525,
		0.911794,
		0.913042,
		0.914268,
		0.915473,
		0.916658,
		0.917822,
		0.918966,
		0.92009,
		0.921195,
		0.922282,
		0.92335,
		0.9244,
		0.925433,
		0.926448,
		0.927446,
		0.928428,
		0.929393,
		0.930343,
		0.931276,
		0.932195,
		0.933098,
		0.933986,
		0.93486,
		0.93572,
		0.936566,
		0.937398,
		0.938217,
		0.939022,
		0.939815,
		0.940595,
		0.941362,
		0.942118,
		0.942861,
		0.943593,
		0.944313,
		0.945021,
		0.945719,
		0.946406,
		0.925841,
		0.924919,
		0.923982,
		0.923032,
		0.922066,
		0.921086,
		0.92009,
		0.919079,
		0.918052,
		0.917009,
		0.91595,
		0.914874,
		0.913781,
		0.91267,
		0.911542,
		0.910396,
		0.909232,
		0.908049,
		0.906848,
		0.905627,
		0.904386,
		0.903125,
		0.901844,
		0.900541,
		0.899218,
		0.897873,
		0.896506,
		0.895116,
		0.893703,
		0.892267,
		0.890808,
		0.889324,
		0.887815,
		0.886281,
		0.884721,
		0.883135,
		0.881522,
		0.879882,
		0.878214,
		0.876517,
		0.874792,
		0.873037,
		0.871252,
		0.869437,
		0.86759,
		0.865711,
		0.863799,
		0.861854,
		0.859876,
		0.857862,
		0.855814,
		0.853729,
		0.851607,
		0.849448,
		0.847251,
		0.845015,
		0.842738,
		0.840421,
		0.838063,
		0.835662,
		0.833218,
		0.830729,
		0.828196,
		0.825616,
		0.82299,
		0.820315,
		0.817592,
		0.814818,
		0.811993,
		0.809117,
		0.806187,
		0.803203,
		0.800163,
		0.797066,
		0.793912,
		0.790699,
		0.787425,
		0.78409,
		0.780692,
		0.77723,
		0.773702,
		0.770107,
		0.766443,
		0.76271,
		0.758906,
		0.755028,
		0.751077,
		0.747049,
		0.742944,
		0.738759,
		0.734494,
		0.730146,
		0.725714,
		0.721196,
		0.71659,
		0.711894,
		0.707107,
		0.702226,
		0.69725,
		0.692176,
		0.687003,
		0.681729,
		0.676351,
		0.670868,
		0.665277,
		0.659576,
		0.653763,
		0.647836,
		0.641793,
		0.635631,
		0.629348,
		0.622942,
		0.61641,
		0.609751,
		0.602961,
		0.596039,
		0.588982,
		0.581788,
		0.574455,
		0.56698,
		0.559361,
		0.551596,
		0.543682,
		0.535618,
		0.527402,
		0.519031,
		0.510503,
		0.501818,
		0.492972,
		0.483964,
		0.474793,
		0.465459,
		0.455958,
		0.446291,
		0.436456,
		0.42645,
		0.416269,
		0.405908,
		0.395357,
		0.384603,
		0.373628,
		0.362411,
		0.350925,
		0.339142,
		0.327031,
		0.314561,
		0.301705,
		0.28844,
		0.274752,
		0.260641,
		0.246117,
		0.231212,
		0.215973,
		0.20047,
		0.184789,
		0.169039,
		0.153342,
		0.137834,
		0.122659,
		0.107965,
		0.093895,
		0.080587,
		0.068164,
		0.056729,
		0.046363,
		0.037121,
		0.029031,
		0.02209,
		0.01627,
		0.011518,
		0.007757,
		0.004893,
		0.002816,
		0.001409,
		0.000551,
		0.000119,
		0.000119,
		0.000551,
		0.001409,
		0.002816,
		0.004893,
		0.007757,
		0.011518,
		0.01627,
		0.02209,
		0.029031,
		0.037121,
		0.046363,
		0.056729,
		0.068164,
		0.080587,
		0.093895,
		0.107965,
		0.122659,
		0.137834,
		0.153342,
		0.169039,
		0.184789,
		0.20047,
		0.215973,
		0.231212,
		0.246117,
		0.260641,
		0.274752,
		0.28844,
		0.301705,
		0.314561,
		0.327031,
		0.339142,
		0.350925,
		0.362411,
		0.373628,
		0.384603,
		0.395357,
		0.405908,
		0.416269,
		0.42645,
		0.436456,
		0.446291,
		0.455958,
		0.465459,
		0.474793,
		0.483964,
		0.492972,
		0.501818,
		0.510503,
		0.519031,
		0.527402,
		0.535618,
		0.543682,
		0.551596,
		0.559361,
		0.56698,
		0.574455,
		0.581788,
		0.588982,
		0.596039,
		0.602961,
		0.609751,
		0.61641,
		0.622942,
		0.629348,
		0.635631,
		0.641793,
		0.647836,
		0.653763,
		0.659576,
		0.665277,
		0.670868,
		0.676351,
		0.681729,
		0.687003,
		0.692176,
		0.69725,
		0.702226,
		0.707107,
		0.711894,
		0.71659,
		0.721196,
		0.725714,
		0.730146,
		0.734494,
		0.738759,
		0.742944,
		0.747049,
		0.751077,
		0.755028,
		0.758906,
		0.76271,
		0.766443,
		0.770107,
		0.773702,
		0.77723,
		0.780692,
		0.78409,
		0.787425,
		0.790699,
		0.793912,
		0.797066,
		0.800163,
		0.803203,
		0.806187,
		0.809117,
		0.811993,
		0.814818,
		0.817592,
		0.820315,
		0.82299,
		0.825616,
		0.828196,
		0.830729,
		0.833218,
		0.835662,
		0.838063,
		0.840421,
		0.842738,
		0.845015,
		0.847251,
		0.849448,
		0.851607,
		0.853729,
		0.855814,
		0.857862,
		0.859876,
		0.861854,
		0.863799,
		0.865711,
		0.86759,
		0.869437,
		0.871252,
		0.873037,
		0.874792,
		0.876517,
		0.878214,
		0.879882,
		0.881522,
		0.883135,
		0.884721,
		0.886281,
		0.887815,
		0.889324,
		0.890808,
		0.892267,
		0.893703,
		0.895116,
		0.896506,
		0.897873,
		0.899218,
		0.900541,
		0.901844,
		0.903125,
		0.904386,
		0.905627,
		0.906848,
		0.908049,
		0.909232,
		0.910396,
		0.911542,
		0.91267,
		0.913781,
		0.914874,
		0.91595,
		0.917009,
		0.918052,
		0.919079,
		0.92009,
		0.921086,
		0.922066,
		0.923032,
		0.923982,
		0.924919,
		0.900183,
		0.898975,
		0.897749,
		0.896506,
		0.895243,
		0.893962,
		0.892661,
		0.891341,
		0.890001,
		0.888641,
		0.88726,
		0.885858,
		0.884434,
		0.882989,
		0.881522,
		0.880032,
		0.878519,
		0.876983,
		0.875423,
		0.873839,
		0.87223,
		0.870596,
		0.868936,
		0.867251,
		0.865538,
		0.863799,
		0.862033,
		0.860238,
		0.858415,
		0.856563,
		0.854681,
		0.852769,
		0.850827,
		0.848853,
		0.846847,
		0.844809,
		0.842738,
		0.840634,
		0.838495,
		0.836321,
		0.834112,
		0.831866,
		0.829583,
		0.827263,
		0.824905,
		0.822507,
		0.820069,
		0.817592,
		0.815072,
		0.812511,
		0.809907,
		0.807258,
		0.804566,
		0.801828,
		0.799043,
		0.796212,
		0.793332,
		0.790404,
		0.787425,
		0.784396,
		0.781314,
		0.77818,
		0.774992,
		0.771749,
		0.76845,
		0.765094,
		0.76168,
		0.758206,
		0.754672,
		0.751077,
		0.747418,
		0.743696,
		0.739908,
		0.736055,
		0.732133,
		0.728142,
		0.724081,
		0.719949,
		0.715743,
		0.711463,
		0.707107,
		0.702674,
		0.698162,
		0.69357,
		0.688896,
		0.684139,
		0.679298,
		0.67437,
		0.669354,
		0.664249,
		0.659052,
		0.653763,
		0.64838,
		0.642901,
		0.637323,
		0.631647,
		0.625869,
		0.619989,
		0.614003,
		0.607912,
		0.601712,
		0.595403,
		0.588982,
		0.582448,
		0.575799,
		0.569033,
		0.562148,
		0.555144,
		0.548017,
		0.540768,
		0.533393,
		0.525892,
		0.518262,
		0.510503,
		0.502614,
		0.494592,
		0.486437,
		0.478147,
		0.469722,
		0.461161,
		0.452462,
		0.443626,
		0.434652,
		0.42554,
		0.416289,
		0.406901,
		0.397375,
		0.387712,
		0.377913,
		0.36798,
		0.357915,
		0.34772,
		0.337397,
		0.32695,
		0.316381,
		0.305694,
		0.294891,
		0.283974,
		0.272942,
		0.261794,
		0.250527,
		0.239137,
		0.227619,
		0.215968,
		0.204182,
		0.192263,
		0.180217,
		0.168059,
		0.155812,
		0.143513,
		0.131208,
		0.118961,
		0.106843,
		0.094942,
		0.083354,
		0.072182,
		0.061533,
		0.051515,
		0.042229,
		0.033766,
		0.026203,
		0.019598,
		0.013986,
		0.009378,
		0.005759,
		0.003088,
		0.001299,
		0.000305,
		0.000305,
		0.001299,
		0.003088,
		0.005759,
		0.009378,
		0.013986,
		0.019598,
		0.026203,
		0.033766,
		0.042229,
		0.051515,
		0.061533,
		0.072182,
		0.083354,
		0.094942,
		0.106843,
		0.118961,
		0.131208,
		0.143513,
		0.155812,
		0.168059,
		0.180217,
		0.192263,
		0.204182,
		0.215968,
		0.227619,
		0.239137,
		0.250527,
		0.261794,
		0.272942,
		0.283974,
		0.294891,
		0.305694,
		0.316381,
		0.32695,
		0.337397,
		0.34772,
		0.357915,
		0.36798,
		0.377913,
		0.387712,
		0.397375,
		0.406901,
		0.416289,
		0.42554,
		0.434652,
		0.443626,
		0.452462,
		0.461161,
		0.469722,
		0.478147,
		0.486437,
		0.494592,
		0.502614,
		0.510503,
		0.518262,
		0.525892,
		0.533393,
		0.540768,
		0.548017,
		0.555144,
		0.562148,
		0.569033,
		0.575799,
		0.582448,
		0.588982,
		0.595403,
		0.601712,
		0.607912,
		0.614003,
		0.619989,
		0.625869,
		0.631647,
		0.637323,
		0.642901,
		0.64838,
		0.653763,
		0.659052,
		0.664249,
		0.669354,
		0.67437,
		0.679298,
		0.684139,
		0.688896,
		0.69357,
		0.698162,
		0.702674,
		0.707107,
		0.711463,
		0.715743,
		0.719949,
		0.724081,
		0.728142,
		0.732133,
		0.736055,
		0.739908,
		0.743696,
		0.747418,
		0.751077,
		0.754672,
		0.758206,
		0.76168,
		0.765094,
		0.76845,
		0.771749,
		0.774992,
		0.77818,
		0.781314,
		0.784396,
		0.787425,
		0.790404,
		0.793332,
		0.796212,
		0.799043,
		0.801828,
		0.804566,
		0.807258,
		0.809907,
		0.812511,
		0.815072,
		0.817592,
		0.820069,
		0.822507,
		0.824905,
		0.827263,
		0.829583,
		0.831866,
		0.834112,
		0.836321,
		0.838495,
		0.840634,
		0.842738,
		0.844809,
		0.846847,
		0.848853,
		0.850827,
		0.852769,
		0.854681,
		0.856563,
		0.858415,
		0.860238,
		0.862033,
		0.863799,
		0.865538,
		0.867251,
		0.868936,
		0.870596,
		0.87223,
		0.873839,
		0.875423,
		0.876983,
		0.878519,
		0.880032,
		0.881522,
		0.882989,
		0.884434,
		0.885858,
		0.88726,
		0.888641,
		0.890001,
		0.891341,
		0.892661,
		0.893962,
		0.895243,
		0.896506,
		0.897749,
		0.898975,
		0.870045,
		0.868517,
		0.866967,
		0.865395,
		0.863799,
		0.862181,
		0.860539,
		0.858873,
		0.857183,
		0.855469,
		0.853729,
		0.851964,
		0.850172,
		0.848355,
		0.84651,
		0.844638,
		0.842738,
		0.84081,
		0.838854,
		0.836868,
		0.834852,
		0.832806,
		0.830729,
		0.828621,
		0.826481,
		0.824309,
		0.822103,
		0.819865,
		0.817592,
		0.815284,
		0.812941,
		0.810562,
		0.808146,
		0.805693,
		0.803203,
		0.800673,
		0.798105,
		0.795497,
		0.792848,
		0.790157,
		0.787425,
		0.78465,
		0.781832,
		0.778969,
		0.776061,
		0.773107,
		0.770107,
		0.767059,
		0.763962,
		0.760817,
		0.757621,
		0.754375,
		0.751077,
		0.747726,
		0.744321,
		0.740862,
		0.737347,
		0.733775,
		0.730146,
		0.726459,
		0.722712,
		0.718904,
		0.715035,
		0.711103,
		0.707107,
		0.703046,
		0.698919,
		0.694725,
		0.690463,
		0.686131,
		0.681729,
		0.677255,
		0.672708,
		0.668086,
		0.663389,
		0.658615,
		0.653763,
		0.648832,
		0.643821,
		0.638727,
		0.63355,
		0.628289,
		0.622942,
		0.617508,
		0.611985,
		0.606372,
		0.600669,
		0.594872,
		0.588982,
		0.582997,
		0.576915,
		0.570735,
		0.564456,
		0.558077,
		0.551596,
		0.545012,
		0.538323,
		0.531529,
		0.524629,
		0.517621,
		0.510503,
		0.503276,
		0.495938,
		0.488488,
		0.480925,
		0.473249,
		0.465459,
		0.457553,
		0.449532,
		0.441396,
		0.433143,
		0.424774,
		0.416289,
		0.407688,
		0.398972,
		0.39014,
		0.381194,
		0.372135,
		0.362964,
		0.353683,
		0.344293,
		0.334797,
		0.325198,
		0.315497,
		0.3057,
		0.295809,
		0.28583,
		0.275766,
		0.265625,
		0.255412,
		0.245133,
		0.234798,
		0.224414,
		0.213992,
		0.20354,
		0.193071,
		0.182595,
		0.172124,
		0.16167,
		0.151247,
		0.140868,
		0.130547,
		0.120303,
		0.110155,
		0.100129,
		0.090256,
		0.080573,
		0.071125,
		0.061965,
		0.053154,
		0.044763,
		0.036866,
		0.029542,
		0.022875,
		0.016943,
		0.01182,
		0.007571,
		0.004244,
		0.001872,
		0.000462,
		0.000462,
		0.001872,
		0.004244,
		0.007571,
		0.01182,
		0.016943,
		0.022875,
		0.029542,
		0.036866,
		0.044763,
		0.053154,
		0.061965,
		0.071125,
		0.080573,
		0.090256,
		0.100129,
		0.110155,
		0.120303,
		0.130547,
		0.140868,
		0.151247,
		0.16167,
		0.172124,
		0.182595,
		0.193071,
		0.20354,
		0.213992,
		0.224414,
		0.234798,
		0.245133,
		0.255412,
		0.265625,
		0.275766,
		0.28583,
		0.295809,
		0.3057,
		0.315497,
		0.325198,
		0.334797,
		0.344293,
		0.353683,
		0.362964,
		0.372135,
		0.381194,
		0.39014,
		0.398972,
		0.407688,
		0.416289,
		0.424774,
		0.433143,
		0.441396,
		0.449532,
		0.457553,
		0.465459,
		0.473249,
		0.480925,
		0.488488,
		0.495938,
		0.503276,
		0.510503,
		0.517621,
		0.524629,
		0.531529,
		0.538323,
		0.545012,
		0.551596,
		0.558077,
		0.564456,
		0.570735,
		0.576915,
		0.582997,
		0.588982,
		0.594872,
		0.600669,
		0.606372,
		0.611985,
		0.617508,
		0.622942,
		0.628289,
		0.63355,
		0.638727,
		0.643821,
		0.648832,
		0.653763,
		0.658615,
		0.663389,
		0.668086,
		0.672708,
		0.677255,
		0.681729,
		0.686131,
		0.690463,
		0.694725,
		0.698919,
		0.703046,
		0.707107,
		0.711103,
		0.715035,
		0.718904,
		0.722712,
		0.726459,
		0.730146,
		0.733775,
		0.737347,
		0.740862,
		0.744321,
		0.747726,
		0.751077,
		0.754375,
		0.757621,
		0.760817,
		0.763962,
		0.767059,
		0.770107,
		0.773107,
		0.776061,
		0.778969,
		0.781832,
		0.78465,
		0.787425,
		0.790157,
		0.792848,
		0.795497,
		0.798105,
		0.800673,
		0.803203,
		0.805693,
		0.808146,
		0.810562,
		0.812941,
		0.815284,
		0.817592,
		0.819865,
		0.822103,
		0.824309,
		0.826481,
		0.828621,
		0.830729,
		0.832806,
		0.834852,
		0.836868,
		0.838854,
		0.84081,
		0.842738,
		0.844638,
		0.84651,
		0.848355,
		0.850172,
		0.851964,
		0.853729,
		0.855469,
		0.857183,
		0.858873,
		0.860539,
		0.862181,
		0.863799,
		0.865395,
		0.866967,
		0.868517,
		0.835475,
		0.833597,
		0.831692,
		0.82976,
		0.827802,
		0.825816,
		0.823803,
		0.821761,
		0.819691,
		0.817592,
		0.815463,
		0.813304,
		0.811114,
		0.808893,
		0.806641,
		0.804357,
		0.80204,
		0.79969,
		0.797307,
		0.794889,
		0.792437,
		0.789949,
		0.787425,
		0.784865,
		0.782268,
		0.779633,
		0.77696,
		0.774249,
		0.771497,
		0.768706,
		0.765874,
		0.763,
		0.760084,
		0.757125,
		0.754123,
		0.751077,
		0.747985,
		0.744848,
		0.741665,
		0.738434,
		0.735156,
		0.731828,
		0.728452,
		0.725025,
		0.721547,
		0.718017,
		0.714434,
		0.710798,
		0.707107,
		0.703361,
		0.699558,
		0.695699,
		0.691782,
		0.687806,
		0.68377,
		0.679673,
		0.675515,
		0.671294,
		0.667009,
		0.66266,
		0.658245,
		0.653763,
		0.649214,
		0.644597,
		0.63991,
		0.635152,
		0.630323,
		0.62542,
		0.620445,
		0.615394,
		0.610268,
		0.605064,
		0.599783,
		0.594422,
		0.588982,
		0.583461,
		0.577857,
		0.57217,
		0.566399,
		0.560543,
		0.5546,
		0.54857,
		0.542452,
		0.536244,
		0.529947,
		0.523558,
		0.517077,
		0.510503,
		0.503836,
		0.497074,
		0.490217,
		0.483264,
		0.476215,
		0.469068,
		0.461824,
		0.454482,
		0.447041,
		0.439502,
		0.431863,
		0.424126,
		0.416289,
		0.408354,
		0.40032,
		0.392188,
		0.383959,
		0.375633,
		0.367211,
		0.358694,
		0.350084,
		0.341382,
		0.332591,
		0.323712,
		0.314747,
		0.3057,
		0.296573,
		0.287371,
		0.278096,
		0.268753,
		0.259348,
		0.249885,
		0.24037,
		0.23081,
		0.221211,
		0.211582,
		0.201932,
		0.192269,
		0.182603,
		0.172947,
		0.163312,
		0.153712,
		0.144161,
		0.134676,
		0.125272,
		0.11597,
		0.106789,
		0.097751,
		0.088879,
		0.0802,
		0.071739,
		0.063527,
		0.055596,
		0.047983,
		0.040727,
		0.033873,
		0.027471,
		0.021576,
		0.016249,
		0.011555,
		0.007563,
		0.004344,
		0.001968,
		0.0005,
		0.0005,
		0.001968,
		0.004344,
		0.007563,
		0.011555,
		0.016249,
		0.021576,
		0.027471,
		0.033873,
		0.040727,
		0.047983,
		0.055596,
		0.063527,
		0.071739,
		0.0802,
		0.088879,
		0.097751,
		0.106789,
		0.11597,
		0.125272,
		0.134676,
		0.144161,
		0.153712,
		0.163312,
		0.172947,
		0.182603,
		0.192269,
		0.201932,
		0.211582,
		0.221211,
		0.23081,
		0.24037,
		0.249885,
		0.259348,
		0.268753,
		0.278096,
		0.287371,
		0.296573,
		0.3057,
		0.314747,
		0.323712,
		0.332591,
		0.341382,
		0.350084,
		0.358694,
		0.367211,
		0.375633,
		0.383959,
		0.392188,
		0.40032,
		0.408354,
		0.416289,
		0.424126,
		0.431863,
		0.439502,
		0.447041,
		0.454482,
		0.461824,
		0.469068,
		0.476215,
		0.483264,
		0.490217,
		0.497074,
		0.503836,
		0.510503,
		0.517077,
		0.523558,
		0.529947,
		0.536244,
		0.542452,
		0.54857,
		0.5546,
		0.560543,
		0.566399,
		0.57217,
		0.577857,
		0.583461,
		0.588982,
		0.594422,
		0.599783,
		0.605064,
		0.610268,
		0.615394,
		0.620445,
		0.62542,
		0.630323,
		0.635152,
		0.63991,
		0.644597,
		0.649214,
		0.653763,
		0.658245,
		0.66266,
		0.667009,
		0.671294,
		0.675515,
		0.679673,
		0.68377,
		0.687806,
		0.691782,
		0.695699,
		0.699558,
		0.703361,
		0.707107,
		0.710798,
		0.714434,
		0.718017,
		0.721547,
		0.725025,
		0.728452,
		0.731828,
		0.735156,
		0.738434,
		0.741665,
		0.744848,
		0.747985,
		0.751077,
		0.754123,
		0.757125,
		0.760084,
		0.763,
		0.765874,
		0.768706,
		0.771497,
		0.774249,
		0.77696,
		0.779633,
		0.782268,
		0.784865,
		0.787425,
		0.789949,
		0.792437,
		0.794889,
		0.797307,
		0.79969,
		0.80204,
		0.804357,
		0.806641,
		0.808893,
		0.811114,
		0.813304,
		0.815463,
		0.817592,
		0.819691,
		0.821761,
		0.823803,
		0.825816,
		0.827802,
		0.82976,
		0.831692,
		0.833597,
		0.796619,
		0.794366,
		0.792083,
		0.78977,
		0.787425,
		0.785049,
		0.782641,
		0.780201,
		0.777728,
		0.775222,
		0.772681,
		0.770107,
		0.767497,
		0.764852,
		0.762171,
		0.759454,
		0.756699,
		0.753907,
		0.751077,
		0.748208,
		0.745299,
		0.742351,
		0.739362,
		0.736332,
		0.73326,
		0.730146,
		0.726989,
		0.723788,
		0.720543,
		0.717253,
		0.713918,
		0.710536,
		0.707107,
		0.70363,
		0.700105,
		0.696531,
		0.692907,
		0.689233,
		0.685507,
		0.681729,
		0.677898,
		0.674014,
		0.670076,
		0.666082,
		0.662033,
		0.657927,
		0.653763,
		0.649542,
		0.645261,
		0.64092,
		0.636519,
		0.632056,
		0.62753,
		0.622942,
		0.618289,
		0.613572,
		0.608789,
		0.603939,
		0.599022,
		0.594037,
		0.588982,
		0.583858,
		0.578663,
		0.573396,
		0.568056,
		0.562644,
		0.557157,
		0.551596,
		0.545959,
		0.540245,
		0.534454,
		0.528585,
		0.522638,
		0.516611,
		0.510503,
		0.504315,
		0.498046,
		0.491695,
		0.485261,
		0.478744,
		0.472143,
		0.465459,
		0.45869,
		0.451836,
		0.444897,
		0.437873,
		0.430764,
		0.423569,
		0.416289,
		0.408924,
		0.401474,
		0.393939,
		0.38632,
		0.378617,
		0.370832,
		0.362964,
		0.355015,
		0.346987,
		0.33888,
		0.330696,
		0.322436,
		0.314104,
		0.3057,
		0.297228,
		0.28869,
		0.280089,
		0.271429,
		0.262714,
		0.253947,
		0.245133,
		0.236278,
		0.227386,
		0.218463,
		0.209516,
		0.200552,
		0.191578,
		0.182603,
		0.173636,
		0.164687,
		0.155766,
		0.146884,
		0.138055,
		0.129291,
		0.120607,
		0.112019,
		0.103544,
		0.0952,
		0.087007,
		0.078987,
		0.071162,
		0.063557,
		0.0562,
		0.049119,
		0.042346,
		0.035914,
		0.02986,
		0.024222,
		0.019044,
		0.01437,
		0.010251,
		0.00674,
		0.003895,
		0.001778,
		0.000457,
		0.000457,
		0.001778,
		0.003895,
		0.00674,
		0.010251,
		0.01437,
		0.019044,
		0.024222,
		0.02986,
		0.035914,
		0.042346,
		0.049119,
		0.0562,
		0.063557,
		0.071162,
		0.078987,
		0.087007,
		0.0952,
		0.103544,
		0.112019,
		0.120607,
		0.129291,
		0.138055,
		0.146884,
		0.155766,
		0.164687,
		0.173636,
		0.182603,
		0.191578,
		0.200552,
		0.209516,
		0.218463,
		0.227386,
		0.236278,
		0.245133,
		0.253947,
		0.262714,
		0.271429,
		0.280089,
		0.28869,
		0.297228,
		0.3057,
		0.314104,
		0.322436,
		0.330696,
		0.33888,
		0.346987,
		0.355015,
		0.362964,
		0.370832,
		0.378617,
		0.38632,
		0.393939,
		0.401474,
		0.408924,
		0.416289,
		0.423569,
		0.430764,
		0.437873,
		0.444897,
		0.451836,
		0.45869,
		0.465459,
		0.472143,
		0.478744,
		0.485261,
		0.491695,
		0.498046,
		0.504315,
		0.510503,
		0.516611,
		0.522638,
		0.528585,
		0.534454,
		0.540245,
		0.545959,
		0.551596,
		0.557157,
		0.562644,
		0.568056,
		0.573396,
		0.578663,
		0.583858,
		0.588982,
		0.594037,
		0.599022,
		0.603939,
		0.608789,
		0.613572,
		0.618289,
		0.622942,
		0.62753,
		0.632056,
		0.636519,
		0.64092,
		0.645261,
		0.649542,
		0.653763,
		0.657927,
		0.662033,
		0.666082,
		0.670076,
		0.674014,
		0.677898,
		0.681729,
		0.685507,
		0.689233,
		0.692907,
		0.696531,
		0.700105,
		0.70363,
		0.707107,
		0.710536,
		0.713918,
		0.717253,
		0.720543,
		0.723788,
		0.726989,
		0.730146,
		0.73326,
		0.736332,
		0.739362,
		0.742351,
		0.745299,
		0.748208,
		0.751077,
		0.753907,
		0.756699,
		0.759454,
		0.762171,
		0.764852,
		0.767497,
		0.770107,
		0.772681,
		0.775222,
		0.777728,
		0.780201,
		0.782641,
		0.785049,
		0.787425,
		0.78977,
		0.792083,
		0.794366,
		0.753719,
		0.751077,
		0.7484,
		0.745689,
		0.742944,
		0.740163,
		0.737347,
		0.734494,
		0.731605,
		0.728678,
		0.725714,
		0.722712,
		0.71967,
		0.71659,
		0.713469,
		0.710309,
		0.707107,
		0.703863,
		0.700578,
		0.69725,
		0.693878,
		0.690463,
		0.687003,
		0.683499,
		0.679948,
		0.676351,
		0.672708,
		0.669016,
		0.665277,
		0.661489,
		0.657651,
		0.653763,
		0.649825,
		0.645835,
		0.641793,
		0.637698,
		0.63355,
		0.629348,
		0.625091,
		0.620779,
		0.61641,
		0.611985,
		0.607502,
		0.602961,
		0.598361,
		0.593702,
		0.588982,
		0.584202,
		0.579359,
		0.574455,
		0.569487,
		0.564456,
		0.559361,
		0.554201,
		0.548975,
		0.543682,
		0.538323,
		0.532897,
		0.527402,
		0.521839,
		0.516206,
		0.510503,
		0.504731,
		0.498887,
		0.492972,
		0.486985,
		0.480925,
		0.474793,
		0.468589,
		0.46231,
		0.455958,
		0.449532,
		0.443032,
		0.436458,
		0.429809,
		0.423087,
		0.416289,
		0.409418,
		0.402472,
		0.395453,
		0.38836,
		0.381194,
		0.373956,
		0.366646,
		0.359265,
		0.351814,
		0.344293,
		0.336705,
		0.32905,
		0.321329,
		0.313545,
		0.3057,
		0.297795,
		0.289832,
		0.281814,
		0.273744,
		0.265625,
		0.25746,
		0.249252,
		0.241006,
		0.232725,
		0.224414,
		0.216079,
		0.207724,
		0.199356,
		0.19098,
		0.182603,
		0.174234,
		0.165879,
		0.157547,
		0.149248,
		0.140991,
		0.132788,
		0.124649,
		0.116587,
		0.108615,
		0.100747,
		0.093,
		0.085389,
		0.077932,
		0.070648,
		0.063557,
		0.056682,
		0.050046,
		0.043674,
		0.037594,
		0.031834,
		0.026425,
		0.021402,
		0.0168,
		0.012657,
		0.009016,
		0.00592,
		0.003417,
		0.001559,
		0.0004,
		0.0004,
		0.001559,
		0.003417,
		0.00592,
		0.009016,
		0.012657,
		0.0168,
		0.021402,
		0.026425,
		0.031834,
		0.037594,
		0.043674,
		0.050046,
		0.056682,
		0.063557,
		0.070648,
		0.077932,
		0.085389,
		0.093,
		0.100747,
		0.108615,
		0.116587,
		0.124649,
		0.132788,
		0.140991,
		0.149248,
		0.157547,
		0.165879,
		0.174234,
		0.182603,
		0.19098,
		0.199356,
		0.207724,
		0.216079,
		0.224414,
		0.232725,
		0.241006,
		0.249252,
		0.25746,
		0.265625,
		0.273744,
		0.281814,
		0.289832,
		0.297795,
		0.3057,
		0.313545,
		0.321329,
		0.32905,
		0.336705,
		0.344293,
		0.351814,
		0.359265,
		0.366646,
		0.373956,
		0.381194,
		0.38836,
		0.395453,
		0.402472,
		0.409418,
		0.416289,
		0.423087,
		0.429809,
		0.436458,
		0.443032,
		0.449532,
		0.455958,
		0.46231,
		0.468589,
		0.474793,
		0.480925,
		0.486985,
		0.492972,
		0.498887,
		0.504731,
		0.510503,
		0.516206,
		0.521839,
		0.527402,
		0.532897,
		0.538323,
		0.543682,
		0.548975,
		0.554201,
		0.559361,
		0.564456,
		0.569487,
		0.574455,
		0.579359,
		0.584202,
		0.588982,
		0.593702,
		0.598361,
		0.602961,
		0.607502,
		0.611985,
		0.61641,
		0.620779,
		0.625091,
		0.629348,
		0.63355,
		0.637698,
		0.641793,
		0.645835,
		0.649825,
		0.653763,
		0.657651,
		0.661489,
		0.665277,
		0.669016,
		0.672708,
		0.676351,
		0.679948,
		0.683499,
		0.687003,
		0.690463,
		0.693878,
		0.69725,
		0.700578,
		0.703863,
		0.707107,
		0.710309,
		0.713469,
		0.71659,
		0.71967,
		0.722712,
		0.725714,
		0.728678,
		0.731605,
		0.734494,
		0.737347,
		0.740163,
		0.742944,
		0.745689,
		0.7484,
		0.751077,
		0.707107,
		0.704067,
		0.700991,
		0.697877,
		0.694725,
		0.691535,
		0.688306,
		0.685037,
		0.681729,
		0.67838,
		0.67499,
		0.671559,
		0.668086,
		0.66457,
		0.661012,
		0.65741,
		0.653763,
		0.650073,
		0.646337,
		0.642555,
		0.638727,
		0.634852,
		0.63093,
		0.62696,
		0.622942,
		0.618874,
		0.614757,
		0.61059,
		0.606372,
		0.602103,
		0.597782,
		0.593409,
		0.588982,
		0.584502,
		0.579968,
		0.575379,
		0.570735,
		0.566035,
		0.561279,
		0.556466,
		0.551596,
		0.546667,
		0.541681,
		0.536635,
		0.531529,
		0.526364,
		0.521138,
		0.515852,
		0.510503,
		0.505093,
		0.499621,
		0.494086,
		0.488488,
		0.482827,
		0.477101,
		0.471312,
		0.465459,
		0.45954,
		0.453557,
		0.447509,
		0.441396,
		0.435217,
		0.428973,
		0.422664,
		0.416289,
		0.409849,
		0.403344,
		0.396775,
		0.39014,
		0.383441,
		0.376679,
		0.369853,
		0.362964,
		0.356013,
		0.349001,
		0.341929,
		0.334797,
		0.327607,
		0.32036,
		0.313057,
		0.3057,
		0.29829,
		0.29083,
		0.283322,
		0.275766,
		0.268167,
		0.260527,
		0.252848,
		0.245133,
		0.237387,
		0.229612,
		0.221812,
		0.213992,
		0.206156,
		0.198309,
		0.190456,
		0.182603,
		0.174757,
		0.166922,
		0.159107,
		0.151319,
		0.143566,
		0.135857,
		0.128201,
		0.120607,
		0.113087,
		0.105651,
		0.098313,
		0.091084,
		0.083978,
		0.077011,
		0.070199,
		0.063557,
		0.057105,
		0.050862,
		0.044848,
		0.039085,
		0.033598,
		0.02841,
		0.02355,
		0.019046,
		0.014928,
		0.011231,
		0.007987,
		0.005236,
		0.003018,
		0.001374,
		0.000352,
		0.000352,
		0.001374,
		0.003018,
		0.005236,
		0.007987,
		0.011231,
		0.014928,
		0.019046,
		0.02355,
		0.02841,
		0.033598,
		0.039085,
		0.044848,
		0.050862,
		0.057105,
		0.063557,
		0.070199,
		0.077011,
		0.083978,
		0.091084,
		0.098313,
		0.105651,
		0.113087,
		0.120607,
		0.128201,
		0.135857,
		0.143566,
		0.151319,
		0.159107,
		0.166922,
		0.174757,
		0.182603,
		0.190456,
		0.198309,
		0.206156,
		0.213992,
		0.221812,
		0.229612,
		0.237387,
		0.245133,
		0.252848,
		0.260527,
		0.268167,
		0.275766,
		0.283322,
		0.29083,
		0.29829,
		0.3057,
		0.313057,
		0.32036,
		0.327607,
		0.334797,
		0.341929,
		0.349001,
		0.356013,
		0.362964,
		0.369853,
		0.376679,
		0.383441,
		0.39014,
		0.396775,
		0.403344,
		0.409849,
		0.416289,
		0.422664,
		0.428973,
		0.435217,
		0.441396,
		0.447509,
		0.453557,
		0.45954,
		0.465459,
		0.471312,
		0.477101,
		0.482827,
		0.488488,
		0.494086,
		0.499621,
		0.505093,
		0.510503,
		0.515852,
		0.521138,
		0.526364,
		0.531529,
		0.536635,
		0.541681,
		0.546667,
		0.551596,
		0.556466,
		0.561279,
		0.566035,
		0.570735,
		0.575379,
		0.579968,
		0.584502,
		0.588982,
		0.593409,
		0.597782,
		0.602103,
		0.606372,
		0.61059,
		0.614757,
		0.618874,
		0.622942,
		0.62696,
		0.63093,
		0.634852,
		0.638727,
		0.642555,
		0.646337,
		0.650073,
		0.653763,
		0.65741,
		0.661012,
		0.66457,
		0.668086,
		0.671559,
		0.67499,
		0.67838,
		0.681729,
		0.685037,
		0.688306,
		0.691535,
		0.694725,
		0.697877,
		0.700991,
		0.704067,
		0.657196,
		0.653763,
		0.650291,
		0.646778,
		0.643226,
		0.639632,
		0.635997,
		0.63232,
		0.628601,
		0.624839,
		0.621034,
		0.617185,
		0.613292,
		0.609355,
		0.605372,
		0.601344,
		0.59727,
		0.59315,
		0.588982,
		0.584767,
		0.580504,
		0.576193,
		0.571833,
		0.567424,
		0.562964,
		0.558455,
		0.553895,
		0.549284,
		0.544621,
		0.539907,
		0.535139,
		0.53032,
		0.525446,
		0.52052,
		0.515539,
		0.510503,
		0.505413,
		0.500268,
		0.495067,
		0.489811,
		0.484498,
		0.479129,
		0.473704,
		0.468221,
		0.462682,
		0.457085,
		0.45143,
		0.445718,
		0.439948,
		0.43412,
		0.428234,
		0.422291,
		0.416289,
		0.41023,
		0.404113,
		0.397939,
		0.391707,
		0.385418,
		0.379073,
		0.372671,
		0.366214,
		0.359701,
		0.353134,
		0.346512,
		0.339837,
		0.333111,
		0.326332,
		0.319503,
		0.312626,
		0.3057,
		0.298728,
		0.29171,
		0.28465,
		0.277548,
		0.270407,
		0.263228,
		0.256014,
		0.248768,
		0.241492,
		0.234189,
		0.226862,
		0.219514,
		0.212149,
		0.204772,
		0.197385,
		0.189994,
		0.182603,
		0.175218,
		0.167843,
		0.160485,
		0.153149,
		0.145843,
		0.138573,
		0.131346,
		0.124172,
		0.117059,
		0.110014,
		0.10305,
		0.096175,
		0.0894,
		0.082738,
		0.076202,
		0.069803,
		0.063557,
		0.057479,
		0.051585,
		0.045892,
		0.040417,
		0.035182,
		0.030205,
		0.02551,
		0.021119,
		0.017058,
		0.013353,
		0.010032,
		0.007125,
		0.004665,
		0.002685,
		0.001221,
		0.000312,
		0.000312,
		0.001221,
		0.002685,
		0.004665,
		0.007125,
		0.010032,
		0.013353,
		0.017058,
		0.021119,
		0.02551,
		0.030205,
		0.035182,
		0.040417,
		0.045892,
		0.051585,
		0.057479,
		0.063557,
		0.069803,
		0.076202,
		0.082738,
		0.0894,
		0.096175,
		0.10305,
		0.110014,
		0.117059,
		0.124172,
		0.131346,
		0.138573,
		0.145843,
		0.153149,
		0.160485,
		0.167843,
		0.175218,
		0.182603,
		0.189994,
		0.197385,
		0.204772,
		0.212149,
		0.219514,
		0.226862,
		0.234189,
		0.241492,
		0.248768,
		0.256014,
		0.263228,
		0.270407,
		0.277548,
		0.28465,
		0.29171,
		0.298728,
		0.3057,
		0.312626,
		0.319503,
		0.326332,
		0.333111,
		0.339837,
		0.346512,
		0.353134,
		0.359701,
		0.366214,
		0.372671,
		0.379073,
		0.385418,
		0.391707,
		0.397939,
		0.404113,
		0.41023,
		0.416289,
		0.422291,
		0.428234,
		0.43412,
		0.439948,
		0.445718,
		0.45143,
		0.457085,
		0.462682,
		0.468221,
		0.473704,
		0.479129,
		0.484498,
		0.489811,
		0.495067,
		0.500268,
		0.505413,
		0.510503,
		0.515539,
		0.52052,
		0.525446,
		0.53032,
		0.535139,
		0.539907,
		0.544621,
		0.549284,
		0.553895,
		0.558455,
		0.562964,
		0.567424,
		0.571833,
		0.576193,
		0.580504,
		0.584767,
		0.588982,
		0.59315,
		0.59727,
		0.601344,
		0.605372,
		0.609355,
		0.613292,
		0.617185,
		0.621034,
		0.624839,
		0.628601,
		0.63232,
		0.635997,
		0.639632,
		0.643226,
		0.646778,
		0.650291,
		0.653763,
		0.604481,
		0.600669,
		0.596815,
		0.592919,
		0.588982,
		0.585003,
		0.58098,
		0.576915,
		0.572806,
		0.568653,
		0.564456,
		0.560215,
		0.555928,
		0.551596,
		0.547218,
		0.542794,
		0.538323,
		0.533806,
		0.529241,
		0.524629,
		0.519969,
		0.51526,
		0.510503,
		0.505698,
		0.500843,
		0.495938,
		0.490984,
		0.48598,
		0.480925,
		0.475821,
		0.470665,
		0.465459,
		0.460201,
		0.454892,
		0.449532,
		0.444121,
		0.438658,
		0.433143,
		0.427577,
		0.421959,
		0.416289,
		0.410568,
		0.404796,
		0.398972,
		0.393097,
		0.387171,
		0.381194,
		0.375167,
		0.369091,
		0.362964,
		0.356789,
		0.350565,
		0.344293,
		0.337974,
		0.331609,
		0.325198,
		0.318742,
		0.312242,
		0.3057,
		0.299116,
		0.292492,
		0.28583,
		0.27913,
		0.272394,
		0.265625,
		0.258824,
		0.251992,
		0.245133,
		0.238249,
		0.231342,
		0.224414,
		0.21747,
		0.210511,
		0.203541,
		0.196564,
		0.189584,
		0.182603,
		0.175628,
		0.168662,
		0.16171,
		0.154777,
		0.147869,
		0.140991,
		0.134151,
		0.127354,
		0.120607,
		0.113919,
		0.107296,
		0.100747,
		0.094282,
		0.08791,
		0.08164,
		0.075484,
		0.069452,
		0.063557,
		0.057812,
		0.05223,
		0.046826,
		0.041614,
		0.036611,
		0.031834,
		0.027301,
		0.023032,
		0.019046,
		0.015366,
		0.012014,
		0.009016,
		0.006396,
		0.004182,
		0.002404,
		0.001092,
		0.000279,
		0.000279,
		0.001092,
		0.002404,
		0.004182,
		0.006396,
		0.009016,
		0.012014,
		0.015366,
		0.019046,
		0.023032,
		0.027301,
		0.031834,
		0.036611,
		0.041614,
		0.046826,
		0.05223,
		0.057812,
		0.063557,
		0.069452,
		0.075484,
		0.08164,
		0.08791,
		0.094282,
		0.100747,
		0.107296,
		0.113919,
		0.120607,
		0.127354,
		0.134151,
		0.140991,
		0.147869,
		0.154777,
		0.16171,
		0.168662,
		0.175628,
		0.182603,
		0.189584,
		0.196564,
		0.203541,
		0.210511,
		0.21747,
		0.224414,
		0.231342,
		0.238249,
		0.245133,
		0.251992,
		0.258824,
		0.265625,
		0.272394,
		0.27913,
		0.28583,
		0.292492,
		0.299116,
		0.3057,
		0.312242,
		0.318742,
		0.325198,
		0.331609,
		0.337974,
		0.344293,
		0.350565,
		0.356789,
		0.362964,
		0.369091,
		0.375167,
		0.381194,
		0.387171,
		0.393097,
		0.398972,
		0.404796,
		0.410568,
		0.416289,
		0.421959,
		0.427577,
		0.433143,
		0.438658,
		0.444121,
		0.449532,
		0.454892,
		0.460201,
		0.465459,
		0.470665,
		0.475821,
		0.480925,
		0.48598,
		0.490984,
		0.495938,
		0.500843,
		0.505698,
		0.510503,
		0.51526,
		0.519969,
		0.524629,
		0.529241,
		0.533806,
		0.538323,
		0.542794,
		0.547218,
		0.551596,
		0.555928,
		0.560215,
		0.564456,
		0.568653,
		0.572806,
		0.576915,
		0.58098,
		0.585003,
		0.588982,
		0.592919,
		0.596815,
		0.600669,
		0.549528,
		0.545361,
		0.541152,
		0.536902,
		0.532609,
		0.528274,
		0.523896,
		0.519475,
		0.515011,
		0.510503,
		0.505952,
		0.501356,
		0.496716,
		0.492031,
		0.487302,
		0.482527,
		0.477707,
		0.472842,
		0.467931,
		0.462975,
		0.457972,
		0.452924,
		0.447829,
		0.442688,
		0.437501,
		0.432268,
		0.426988,
		0.421662,
		0.416289,
		0.410871,
		0.405406,
		0.399895,
		0.394338,
		0.388735,
		0.383087,
		0.377394,
		0.371655,
		0.365872,
		0.360045,
		0.354174,
		0.34826,
		0.342303,
		0.336303,
		0.330263,
		0.324181,
		0.31806,
		0.311899,
		0.3057,
		0.299464,
		0.293191,
		0.286884,
		0.280543,
		0.27417,
		0.267766,
		0.261333,
		0.254872,
		0.248386,
		0.241875,
		0.235343,
		0.228792,
		0.222223,
		0.21564,
		0.209044,
		0.20244,
		0.19583,
		0.189216,
		0.182603,
		0.175995,
		0.169394,
		0.162806,
		0.156234,
		0.149684,
		0.14316,
		0.136666,
		0.13021,
		0.123796,
		0.117431,
		0.111122,
		0.104874,
		0.098696,
		0.092596,
		0.086581,
		0.08066,
		0.074843,
		0.069138,
		0.063557,
		0.058111,
		0.05281,
		0.047667,
		0.042694,
		0.037906,
		0.033317,
		0.028941,
		0.024796,
		0.020897,
		0.017263,
		0.013913,
		0.010867,
		0.008146,
		0.005773,
		0.003771,
		0.002165,
		0.000982,
		0.000251,
		0.000251,
		0.000982,
		0.002165,
		0.003771,
		0.005773,
		0.008146,
		0.010867,
		0.013913,
		0.017263,
		0.020897,
		0.024796,
		0.028941,
		0.033317,
		0.037906,
		0.042694,
		0.047667,
		0.05281,
		0.058111,
		0.063557,
		0.069138,
		0.074843,
		0.08066,
		0.086581,
		0.092596,
		0.098696,
		0.104874,
		0.111122,
		0.117431,
		0.123796,
		0.13021,
		0.136666,
		0.14316,
		0.149684,
		0.156234,
		0.162806,
		0.169394,
		0.175995,
		0.182603,
		0.189216,
		0.19583,
		0.20244,
		0.209044,
		0.21564,
		0.222223,
		0.228792,
		0.235343,
		0.241875,
		0.248386,
		0.254872,
		0.261333,
		0.267766,
		0.27417,
		0.280543,
		0.286884,
		0.293191,
		0.299464,
		0.3057,
		0.311899,
		0.31806,
		0.324181,
		0.330263,
		0.336303,
		0.342303,
		0.34826,
		0.354174,
		0.360045,
		0.365872,
		0.371655,
		0.377394,
		0.383087,
		0.388735,
		0.394338,
		0.399895,
		0.405406,
		0.410871,
		0.416289,
		0.421662,
		0.426988,
		0.432268,
		0.437501,
		0.442688,
		0.447829,
		0.452924,
		0.457972,
		0.462975,
		0.467931,
		0.472842,
		0.477707,
		0.482527,
		0.487302,
		0.492031,
		0.496716,
		0.501356,
		0.505952,
		0.510503,
		0.515011,
		0.519475,
		0.523896,
		0.528274,
		0.532609,
		0.536902,
		0.541152,
		0.545361,
		0.492972,
		0.488488,
		0.483964,
		0.479399,
		0.474793,
		0.470147,
		0.465459,
		0.460729,
		0.455958,
		0.451146,
		0.446292,
		0.441396,
		0.436458,
		0.431479,
		0.426457,
		0.421394,
		0.416289,
		0.411143,
		0.405954,
		0.400724,
		0.395453,
		0.39014,
		0.384786,
		0.379391,
		0.373956,
		0.36848,
		0.362964,
		0.357409,
		0.351814,
		0.34618,
		0.340507,
		0.334797,
		0.32905,
		0.323265,
		0.317445,
		0.31159,
		0.3057,
		0.299776,
		0.29382,
		0.287832,
		0.281814,
		0.275766,
		0.269691,
		0.263588,
		0.25746,
		0.251308,
		0.245133,
		0.238939,
		0.232725,
		0.226495,
		0.22025,
		0.213992,
		0.207724,
		0.201449,
		0.195168,
		0.188885,
		0.182603,
		0.176325,
		0.170054,
		0.163793,
		0.157547,
		0.151319,
		0.145114,
		0.138935,
		0.132788,
		0.126677,
		0.120607,
		0.114585,
		0.108615,
		0.102704,
		0.096858,
		0.091084,
		0.085389,
		0.079781,
		0.074267,
		0.068856,
		0.063557,
		0.05838,
		0.053333,
		0.048428,
		0.043674,
		0.039085,
		0.034672,
		0.030448,
		0.026425,
		0.02262,
		0.019046,
		0.015719,
		0.012657,
		0.009877,
		0.007397,
		0.005236,
		0.003417,
		0.00196,
		0.000888,
		0.000227,
		0.000227,
		0.000888,
		0.00196,
		0.003417,
		0.005236,
		0.007397,
		0.009877,
		0.012657,
		0.015719,
		0.019046,
		0.02262,
		0.026425,
		0.030448,
		0.034672,
		0.039085,
		0.043674,
		0.048428,
		0.053333,
		0.05838,
		0.063557,
		0.068856,
		0.074267,
		0.079781,
		0.085389,
		0.091084,
		0.096858,
		0.102704,
		0.108615,
		0.114585,
		0.120607,
		0.126677,
		0.132788,
		0.138935,
		0.145114,
		0.151319,
		0.157547,
		0.163793,
		0.170054,
		0.176325,
		0.182603,
		0.188885,
		0.195168,
		0.201449,
		0.207724,
		0.213992,
		0.22025,
		0.226495,
		0.232725,
		0.238939,
		0.245133,
		0.251308,
		0.25746,
		0.263588,
		0.269691,
		0.275766,
		0.281814,
		0.287832,
		0.29382,
		0.299776,
		0.3057,
		0.31159,
		0.317445,
		0.323265,
		0.32905,
		0.334797,
		0.340507,
		0.34618,
		0.351814,
		0.357409,
		0.362964,
		0.36848,
		0.373956,
		0.379391,
		0.384786,
		0.39014,
		0.395453,
		0.400724,
		0.405954,
		0.411143,
		0.416289,
		0.421394,
		0.426457,
		0.431479,
		0.436458,
		0.441396,
		0.446292,
		0.451146,
		0.455958,
		0.460729,
		0.465459,
		0.470147,
		0.474793,
		0.479399,
		0.483964,
		0.488488,
		0.435513,
		0.430764,
		0.425977,
		0.421152,
		0.416289,
		0.411389,
		0.40645,
		0.401474,
		0.39646,
		0.391409,
		0.38632,
		0.381194,
		0.376031,
		0.370832,
		0.365596,
		0.360324,
		0.355015,
		0.349672,
		0.344293,
		0.33888,
		0.333432,
		0.327951,
		0.322436,
		0.316889,
		0.31131,
		0.3057,
		0.300059,
		0.294389,
		0.28869,
		0.282963,
		0.277209,
		0.271429,
		0.265625,
		0.259797,
		0.253947,
		0.248076,
		0.242186,
		0.236278,
		0.230353,
		0.224414,
		0.218463,
		0.2125,
		0.206529,
		0.200552,
		0.19457,
		0.188586,
		0.182603,
		0.176624,
		0.170651,
		0.164687,
		0.158736,
		0.1528,
		0.146884,
		0.140991,
		0.135126,
		0.129291,
		0.123492,
		0.117733,
		0.112019,
		0.106356,
		0.100747,
		0.0952,
		0.08972,
		0.084314,
		0.078987,
		0.073747,
		0.068601,
		0.063557,
		0.058623,
		0.053808,
		0.049119,
		0.044568,
		0.040162,
		0.035914,
		0.031834,
		0.027933,
		0.024223,
		0.020718,
		0.01743,
		0.014374,
		0.011564,
		0.009016,
		0.006746,
		0.004772,
		0.003111,
		0.001783,
		0.000807,
		0.000206,
		0.000206,
		0.000807,
		0.001783,
		0.003111,
		0.004772,
		0.006746,
		0.009016,
		0.011564,
		0.014374,
		0.01743,
		0.020718,
		0.024223,
		0.027933,
		0.031834,
		0.035914,
		0.040162,
		0.044568,
		0.049119,
		0.053808,
		0.058623,
		0.063557,
		0.068601,
		0.073747,
		0.078987,
		0.084314,
		0.08972,
		0.0952,
		0.100747,
		0.106356,
		0.112019,
		0.117733,
		0.123492,
		0.129291,
		0.135126,
		0.140991,
		0.146884,
		0.1528,
		0.158736,
		0.164687,
		0.170651,
		0.176624,
		0.182603,
		0.188586,
		0.19457,
		0.200552,
		0.206529,
		0.2125,
		0.218463,
		0.224414,
		0.230353,
		0.236278,
		0.242186,
		0.248076,
		0.253947,
		0.259797,
		0.265625,
		0.271429,
		0.277209,
		0.282963,
		0.28869,
		0.294389,
		0.300059,
		0.3057,
		0.31131,
		0.316889,
		0.322436,
		0.327951,
		0.333432,
		0.33888,
		0.344293,
		0.349672,
		0.355015,
		0.360324,
		0.365596,
		0.370832,
		0.376031,
		0.381194,
		0.38632,
		0.391409,
		0.39646,
		0.401474,
		0.40645,
		0.411389,
		0.416289,
		0.421152,
		0.425977,
		0.430764,
		0.377913,
		0.372963,
		0.36798,
		0.362964,
		0.357915,
		0.352834,
		0.34772,
		0.342574,
		0.337397,
		0.332189,
		0.32695,
		0.321682,
		0.316383,
		0.311056,
		0.3057,
		0.300316,
		0.294906,
		0.289469,
		0.284006,
		0.278519,
		0.273008,
		0.267474,
		0.261919,
		0.256343,
		0.250747,
		0.245133,
		0.239503,
		0.233856,
		0.228195,
		0.222522,
		0.216838,
		0.211144,
		0.205443,
		0.199736,
		0.194026,
		0.188314,
		0.182603,
		0.176896,
		0.171193,
		0.1655,
		0.159817,
		0.154148,
		0.148496,
		0.142864,
		0.137255,
		0.131674,
		0.126123,
		0.120607,
		0.11513,
		0.109696,
		0.10431,
		0.098976,
		0.093699,
		0.088485,
		0.083339,
		0.078267,
		0.073275,
		0.06837,
		0.063557,
		0.058845,
		0.05424,
		0.049751,
		0.045385,
		0.04115,
		0.037057,
		0.033113,
		0.02933,
		0.025717,
		0.022285,
		0.019046,
		0.016011,
		0.013194,
		0.010606,
		0.008262,
		0.006177,
		0.004366,
		0.002844,
		0.001628,
		0.000737,
		0.000188,
		0.000188,
		0.000737,
		0.001628,
		0.002844,
		0.004366,
		0.006177,
		0.008262,
		0.010606,
		0.013194,
		0.016011,
		0.019046,
		0.022285,
		0.025717,
		0.02933,
		0.033113,
		0.037057,
		0.04115,
		0.045385,
		0.049751,
		0.05424,
		0.058845,
		0.063557,
		0.06837,
		0.073275,
		0.078267,
		0.083339,
		0.088485,
		0.093699,
		0.098976,
		0.10431,
		0.109696,
		0.11513,
		0.120607,
		0.126123,
		0.131674,
		0.137255,
		0.142864,
		0.148496,
		0.154148,
		0.159817,
		0.1655,
		0.171193,
		0.176896,
		0.182603,
		0.188314,
		0.194026,
		0.199736,
		0.205443,
		0.211144,
		0.216838,
		0.222522,
		0.228195,
		0.233856,
		0.239503,
		0.245133,
		0.250747,
		0.256343,
		0.261919,
		0.267474,
		0.273008,
		0.278519,
		0.284006,
		0.289469,
		0.294906,
		0.300316,
		0.3057,
		0.311056,
		0.316383,
		0.321682,
		0.32695,
		0.332189,
		0.337397,
		0.342574,
		0.34772,
		0.352834,
		0.357915,
		0.362964,
		0.36798,
		0.372963,
		0.320992,
		0.315921,
		0.310823,
		0.3057,
		0.300551,
		0.295377,
		0.290179,
		0.284958,
		0.279714,
		0.274448,
		0.269161,
		0.263854,
		0.258527,
		0.253183,
		0.247821,
		0.242442,
		0.237049,
		0.231643,
		0.226223,
		0.220793,
		0.215353,
		0.209905,
		0.204451,
		0.198992,
		0.193529,
		0.188066,
		0.182603,
		0.177144,
		0.171689,
		0.166242,
		0.160804,
		0.155379,
		0.149968,
		0.144575,
		0.139203,
		0.133854,
		0.128533,
		0.123241,
		0.117983,
		0.112762,
		0.107582,
		0.102448,
		0.097363,
		0.092332,
		0.08736,
		0.082451,
		0.077611,
		0.072845,
		0.068159,
		0.063557,
		0.059048,
		0.054636,
		0.05033,
		0.046135,
		0.042059,
		0.03811,
		0.034297,
		0.030627,
		0.02711,
		0.023754,
		0.02057,
		0.017568,
		0.014759,
		0.012153,
		0.009762,
		0.0076,
		0.005678,
		0.00401,
		0.00261,
		0.001493,
		0.000675,
		0.000172,
		0.000172,
		0.000675,
		0.001493,
		0.00261,
		0.00401,
		0.005678,
		0.0076,
		0.009762,
		0.012153,
		0.014759,
		0.017568,
		0.02057,
		0.023754,
		0.02711,
		0.030627,
		0.034297,
		0.03811,
		0.042059,
		0.046135,
		0.05033,
		0.054636,
		0.059048,
		0.063557,
		0.068159,
		0.072845,
		0.077611,
		0.082451,
		0.08736,
		0.092332,
		0.097363,
		0.102448,
		0.107582,
		0.112762,
		0.117983,
		0.123241,
		0.128533,
		0.133854,
		0.139203,
		0.144575,
		0.149968,
		0.155379,
		0.160804,
		0.166242,
		0.171689,
		0.177144,
		0.182603,
		0.188066,
		0.193529,
		0.198992,
		0.204451,
		0.209905,
		0.215353,
		0.220793,
		0.226223,
		0.231643,
		0.237049,
		0.242442,
		0.247821,
		0.253183,
		0.258527,
		0.263854,
		0.269161,
		0.274448,
		0.279714,
		0.284958,
		0.290179,
		0.295377,
		0.300551,
		0.3057,
		0.310823,
		0.315921,
		0.265625,
		0.260527,
		0.255412,
		0.25028,
		0.245133,
		0.239972,
		0.234798,
		0.229612,
		0.224414,
		0.219207,
		0.213992,
		0.208769,
		0.203541,
		0.198309,
		0.193074,
		0.187838,
		0.182603,
		0.177371,
		0.172143,
		0.166922,
		0.16171,
		0.156508,
		0.151319,
		0.146146,
		0.140991,
		0.135857,
		0.130747,
		0.125662,
		0.120607,
		0.115585,
		0.110599,
		0.105651,
		0.100747,
		0.09589,
		0.091084,
		0.086332,
		0.08164,
		0.077011,
		0.072452,
		0.067965,
		0.063557,
		0.059234,
		0.055,
		0.050862,
		0.046826,
		0.042898,
		0.039085,
		0.035395,
		0.031834,
		0.02841,
		0.025132,
		0.022008,
		0.019046,
		0.016256,
		0.013648,
		0.011231,
		0.009016,
		0.007014,
		0.005236,
		0.003696,
		0.002404,
		0.001374,
		0.000621,
		0.000158,
		0.000158,
		0.000621,
		0.001374,
		0.002404,
		0.003696,
		0.005236,
		0.007014,
		0.009016,
		0.011231,
		0.013648,
		0.016256,
		0.019046,
		0.022008,
		0.025132,
		0.02841,
		0.031834,
		0.035395,
		0.039085,
		0.042898,
		0.046826,
		0.050862,
		0.055,
		0.059234,
		0.063557,
		0.067965,
		0.072452,
		0.077011,
		0.08164,
		0.086332,
		0.091084,
		0.09589,
		0.100747,
		0.105651,
		0.110599,
		0.115585,
		0.120607,
		0.125662,
		0.130747,
		0.135857,
		0.140991,
		0.146146,
		0.151319,
		0.156508,
		0.16171,
		0.166922,
		0.172143,
		0.177371,
		0.182603,
		0.187838,
		0.193074,
		0.198309,
		0.203541,
		0.208769,
		0.213992,
		0.219207,
		0.224414,
		0.229612,
		0.234798,
		0.239972,
		0.245133,
		0.25028,
		0.255412,
		0.260527,
		0.212739,
		0.207724,
		0.202704,
		0.197681,
		0.192655,
		0.187629,
		0.182603,
		0.17758,
		0.172561,
		0.167548,
		0.162543,
		0.157547,
		0.152563,
		0.147593,
		0.142639,
		0.137703,
		0.132788,
		0.127896,
		0.12303,
		0.118192,
		0.113386,
		0.108615,
		0.103881,
		0.099188,
		0.094539,
		0.089938,
		0.085389,
		0.080895,
		0.076461,
		0.07209,
		0.067787,
		0.063557,
		0.059405,
		0.055335,
		0.051353,
		0.047464,
		0.043674,
		0.039989,
		0.036415,
		0.032959,
		0.029626,
		0.026425,
		0.023363,
		0.020447,
		0.017685,
		0.015085,
		0.012657,
		0.010409,
		0.008351,
		0.006493,
		0.004845,
		0.003417,
		0.002221,
		0.001269,
		0.000573,
		0.000146,
		0.000146,
		0.000573,
		0.001269,
		0.002221,
		0.003417,
		0.004845,
		0.006493,
		0.008351,
		0.010409,
		0.012657,
		0.015085,
		0.017685,
		0.020447,
		0.023363,
		0.026425,
		0.029626,
		0.032959,
		0.036415,
		0.039989,
		0.043674,
		0.047464,
		0.051353,
		0.055335,
		0.059405,
		0.063557,
		0.067787,
		0.07209,
		0.076461,
		0.080895,
		0.085389,
		0.089938,
		0.094539,
		0.099188,
		0.103881,
		0.108615,
		0.113386,
		0.118192,
		0.12303,
		0.127896,
		0.132788,
		0.137703,
		0.142639,
		0.147593,
		0.152563,
		0.157547,
		0.162543,
		0.167548,
		0.172561,
		0.17758,
		0.182603,
		0.187629,
		0.192655,
		0.197681,
		0.202704,
		0.207724,
		0.163312,
		0.158507,
		0.153712,
		0.14893,
		0.144161,
		0.139409,
		0.134676,
		0.129963,
		0.125272,
		0.120607,
		0.11597,
		0.111363,
		0.106789,
		0.102252,
		0.097753,
		0.093296,
		0.088884,
		0.08452,
		0.080209,
		0.075953,
		0.071756,
		0.067623,
		0.063557,
		0.059563,
		0.055645,
		0.051808,
		0.048056,
		0.044395,
		0.04083,
		0.037366,
		0.034009,
		0.030765,
		0.027641,
		0.024641,
		0.021774,
		0.019046,
		0.016464,
		0.014037,
		0.011771,
		0.009675,
		0.007758,
		0.006028,
		0.004495,
		0.003169,
		0.002059,
		0.001176,
		0.00053,
		0.000135,
		0.000135,
		0.00053,
		0.001176,
		0.002059,
		0.003169,
		0.004495,
		0.006028,
		0.007758,
		0.009675,
		0.011771,
		0.014037,
		0.016464,
		0.019046,
		0.021774,
		0.024641,
		0.027641,
		0.030765,
		0.034009,
		0.037366,
		0.04083,
		0.044395,
		0.048056,
		0.051808,
		0.055645,
		0.059563,
		0.063557,
		0.067623,
		0.071756,
		0.075953,
		0.080209,
		0.08452,
		0.088884,
		0.093296,
		0.097753,
		0.102252,
		0.106789,
		0.111363,
		0.11597,
		0.120607,
		0.125272,
		0.129963,
		0.134676,
		0.139409,
		0.144161,
		0.14893,
		0.153712,
		0.158507,
		0.118371,
		0.113919,
		0.109496,
		0.105104,
		0.100747,
		0.096427,
		0.092147,
		0.08791,
		0.083718,
		0.079575,
		0.075484,
		0.071448,
		0.067471,
		0.063557,
		0.05971,
		0.055933,
		0.05223,
		0.048607,
		0.045066,
		0.041614,
		0.038254,
		0.034993,
		0.031834,
		0.028784,
		0.025848,
		0.023032,
		0.020342,
		0.017784,
		0.015366,
		0.013093,
		0.010974,
		0.009016,
		0.007225,
		0.005611,
		0.004182,
		0.002946,
		0.001913,
		0.001092,
		0.000492,
		0.000125,
		0.000125,
		0.000492,
		0.001092,
		0.001913,
		0.002946,
		0.004182,
		0.005611,
		0.007225,
		0.009016,
		0.010974,
		0.013093,
		0.015366,
		0.017784,
		0.020342,
		0.023032,
		0.025848,
		0.028784,
		0.031834,
		0.034993,
		0.038254,
		0.041614,
		0.045066,
		0.048607,
		0.05223,
		0.055933,
		0.05971,
		0.063557,
		0.067471,
		0.071448,
		0.075484,
		0.079575,
		0.083718,
		0.08791,
		0.092147,
		0.096427,
		0.100747,
		0.105104,
		0.109496,
		0.113919,
		0.078987,
		0.075048,
		0.071162,
		0.06733,
		0.063557,
		0.059846,
		0.0562,
		0.052623,
		0.049119,
		0.045692,
		0.042346,
		0.039085,
		0.035914,
		0.032838,
		0.02986,
		0.026987,
		0.024223,
		0.021574,
		0.019046,
		0.016644,
		0.014374,
		0.012242,
		0.010256,
		0.008421,
		0.006746,
		0.005236,
		0.003901,
		0.002747,
		0.001783,
		0.001017,
		0.000458,
		0.000116,
		0.000116,
		0.000458,
		0.001017,
		0.001783,
		0.002747,
		0.003901,
		0.005236,
		0.006746,
		0.008421,
		0.010256,
		0.012242,
		0.014374,
		0.016644,
		0.019046,
		0.021574,
		0.024223,
		0.026987,
		0.02986,
		0.032838,
		0.035914,
		0.039085,
		0.042346,
		0.045692,
		0.049119,
		0.052623,
		0.0562,
		0.059846,
		0.063557,
		0.06733,
		0.071162,
		0.075048,
		0.046277,
		0.043031,
		0.039864,
		0.03678,
		0.033782,
		0.030875,
		0.028064,
		0.025353,
		0.022747,
		0.020251,
		0.01787,
		0.015609,
		0.013475,
		0.011471,
		0.009606,
		0.007884,
		0.006313,
		0.004898,
		0.003647,
		0.002567,
		0.001665,
		0.000949,
		0.000428,
		0.000108,
		0.000108,
		0.000428,
		0.000949,
		0.001665,
		0.002567,
		0.003647,
		0.004898,
		0.006313,
		0.007884,
		0.009606,
		0.011471,
		0.013475,
		0.015609,
		0.01787,
		0.020251,
		0.022747,
		0.025353,
		0.028064,
		0.030875,
		0.033782,
		0.03678,
		0.039864,
		0.043031,
		0.021402,
		0.019046,
		0.0168,
		0.014668,
		0.012657,
		0.010771,
		0.009016,
		0.007397,
		0.00592,
		0.004591,
		0.003417,
		0.002404,
		0.001559,
		0.000888,
		0.0004,
		0.000101,
		0.000101,
		0.0004,
		0.000888,
		0.001559,
		0.002404,
		0.003417,
		0.004591,
		0.00592,
		0.007397,
		0.009016,
		0.010771,
		0.012657,
		0.014668,
		0.0168,
		0.019046,
		0.005562,
		0.004312,
		0.003208,
		0.002256,
		0.001462,
		0.000833,
		0.000375,
		9.5e-005,
		9.5e-005,
		0.000375,
		0.000833,
		0.001462,
		0.002256,
		0.003208,
		0.004312,
		0.000366,
		0.00144,
		0.003183,
		0.000378,
		0.001486,
		0.003284,
		0.005737,
		0.008808,
		0.012464,
		0.016673,
		0.00039,
		0.001534,
		0.00339,
		0.00592,
		0.009086,
		0.012854,
		0.01719,
		0.022059,
		0.027432,
		0.033279,
		0.039569,
		0.000403,
		0.001584,
		0.003501,
		0.006111,
		0.009378,
		0.013263,
		0.017731,
		0.022747,
		0.02828,
		0.034297,
		0.040769,
		0.047667,
		0.054963,
		0.062632,
		0.070648,
		0.000417,
		0.001638,
		0.003617,
		0.006313,
		0.009684,
		0.013691,
		0.018298,
		0.023468,
		0.029167,
		0.035363,
		0.042023,
		0.049119,
		0.056622,
		0.064503,
		0.072738,
		0.081301,
		0.090168,
		0.099316,
		0.108724,
		0.000431,
		0.001693,
		0.003739,
		0.006524,
		0.010005,
		0.014141,
		0.018893,
		0.024223,
		0.030097,
		0.036479,
		0.043337,
		0.050639,
		0.058356,
		0.06646,
		0.074922,
		0.083718,
		0.092821,
		0.102209,
		0.111859,
		0.121749,
		0.131859,
		0.142168,
		0.152659,
		0.000447,
		0.001752,
		0.003868,
		0.006746,
		0.010342,
		0.014613,
		0.019517,
		0.025016,
		0.031071,
		0.037648,
		0.044712,
		0.05223,
		0.060171,
		0.068506,
		0.077206,
		0.086243,
		0.095593,
		0.105229,
		0.11513,
		0.125272,
		0.135635,
		0.146196,
		0.156939,
		0.167843,
		0.178891,
		0.190067,
		0.201355,
		0.000462,
		0.001814,
		0.004003,
		0.006979,
		0.010696,
		0.015108,
		0.020173,
		0.025848,
		0.032094,
		0.038875,
		0.046154,
		0.053897,
		0.062072,
		0.070648,
		0.079594,
		0.088884,
		0.098489,
		0.108384,
		0.118545,
		0.128949,
		0.139573,
		0.150396,
		0.161399,
		0.172561,
		0.183866,
		0.195295,
		0.206833,
		0.218463,
		0.230171,
		0.241943,
		0.253765,
		0.000479,
		0.001879,
		0.004145,
		0.007225,
		0.011069,
		0.01563,
		0.020862,
		0.026722,
		0.033168,
		0.040162,
		0.047667,
		0.055645,
		0.064064,
		0.072891,
		0.082095,
		0.091646,
		0.101517,
		0.111681,
		0.122113,
		0.132788,
		0.143683,
		0.154777,
		0.166048,
		0.177478,
		0.189047,
		0.200737,
		0.212531,
		0.224414,
		0.236371,
		0.248386,
		0.260446,
		0.272538,
		0.28465,
		0.29677,
		0.308888,
		0.000497,
		0.001948,
		0.004296,
		0.007484,
		0.011462,
		0.016179,
		0.021587,
		0.027641,
		0.034297,
		0.041515,
		0.049255,
		0.057479,
		0.066153,
		0.075242,
		0.084714,
		0.094539,
		0.104687,
		0.11513,
		0.125842,
		0.136799,
		0.147975,
		0.159349,
		0.170898,
		0.182603,
		0.194445,
		0.206404,
		0.218463,
		0.230606,
		0.242817,
		0.255081,
		0.267385,
		0.279714,
		0.292057,
		0.304401,
		0.316735,
		0.32905,
		0.341334,
		0.353578,
		0.365774,
		0.000516,
		0.002021,
		0.004454,
		0.007758,
		0.011876,
		0.016757,
		0.02235,
		0.028608,
		0.035484,
		0.042936,
		0.050923,
		0.059405,
		0.068345,
		0.077708,
		0.087461,
		0.09757,
		0.108006,
		0.118739,
		0.129743,
		0.140991,
		0.152459,
		0.164123,
		0.17596,
		0.18795,
		0.200072,
		0.212308,
		0.224639,
		0.237049,
		0.249522,
		0.262041,
		0.274594,
		0.287165,
		0.299743,
		0.312315,
		0.32487,
		0.337397,
		0.349886,
		0.362328,
		0.374712,
		0.387032,
		0.39928,
		0.411446,
		0.423526,
		0.000536,
		0.002098,
		0.004622,
		0.008046,
		0.012313,
		0.017367,
		0.023155,
		0.029626,
		0.036734,
		0.044432,
		0.052677,
		0.061429,
		0.070648,
		0.080297,
		0.090341,
		0.100747,
		0.111483,
		0.122519,
		0.133826,
		0.145377,
		0.157147,
		0.169111,
		0.181245,
		0.193529,
		0.205942,
		0.218463,
		0.231074,
		0.243759,
		0.256499,
		0.26928,
		0.282087,
		0.294906,
		0.307723,
		0.320527,
		0.333306,
		0.346048,
		0.358744,
		0.371384,
		0.383959,
		0.39646,
		0.40888,
		0.42121,
		0.433445,
		0.445578,
		0.457602,
		0.469512,
		0.481304,
		0.000557,
		0.002179,
		0.004799,
		0.008351,
		0.012775,
		0.018011,
		0.024004,
		0.0307,
		0.038051,
		0.046007,
		0.054523,
		0.063557,
		0.073068,
		0.083016,
		0.093366,
		0.104082,
		0.11513,
		0.12648,
		0.138102,
		0.149968,
		0.162051,
		0.174326,
		0.186768,
		0.199356,
		0.212067,
		0.224882,
		0.237781,
		0.250747,
		0.263763,
		0.276811,
		0.289878,
		0.302949,
		0.31601,
		0.32905,
		0.342055,
		0.355015,
		0.36792,
		0.38076,
		0.393526,
		0.406208,
		0.4188,
		0.431293,
		0.443682,
		0.455958,
		0.468117,
		0.480153,
		0.492061,
		0.503836,
		0.515474,
		0.526971,
		0.538323,
		0.000579,
		0.002265,
		0.004986,
		0.008674,
		0.013263,
		0.018691,
		0.0249,
		0.031834,
		0.039439,
		0.047667,
		0.056468,
		0.065798,
		0.075614,
		0.085875,
		0.096544,
		0.107582,
		0.118957,
		0.130634,
		0.142584,
		0.154777,
		0.167184,
		0.179781,
		0.192542,
		0.205443,
		0.218463,
		0.23158,
		0.244776,
		0.258031,
		0.271328,
		0.28465,
		0.297982,
		0.31131,
		0.324619,
		0.337898,
		0.351133,
		0.364313,
		0.377429,
		0.390469,
		0.403426,
		0.416289,
		0.429052,
		0.441707,
		0.454246,
		0.466665,
		0.478955,
		0.491113,
		0.503133,
		0.515011,
		0.526742,
		0.538323,
		0.54975,
		0.561021,
		0.572131,
		0.583079,
		0.593863,
		0.000602,
		0.002356,
		0.005185,
		0.009016,
		0.013779,
		0.019411,
		0.025848,
		0.033031,
		0.040905,
		0.049418,
		0.058518,
		0.068159,
		0.078295,
		0.088884,
		0.099885,
		0.111261,
		0.122976,
		0.134994,
		0.147285,
		0.159817,
		0.172561,
		0.185491,
		0.198581,
		0.211807,
		0.225145,
		0.238574,
		0.252073,
		0.265625,
		0.27921,
		0.292812,
		0.306416,
		0.320005,
		0.333566,
		0.347086,
		0.360554,
		0.373956,
		0.387283,
		0.400525,
		0.413672,
		0.426716,
		0.439648,
		0.452462,
		0.465151,
		0.477707,
		0.490126,
		0.502402,
		0.514529,
		0.526505,
		0.538323,
		0.549982,
		0.561477,
		0.572806,
		0.583966,
		0.594955,
		0.60577,
		0.61641,
		0.626874,
		0.63716,
		0.647268,
		0.000627,
		0.002453,
		0.005396,
		0.009378,
		0.014327,
		0.020173,
		0.02685,
		0.034297,
		0.042454,
		0.051267,
		0.060681,
		0.070648,
		0.081119,
		0.092052,
		0.103402,
		0.11513,
		0.127199,
		0.139573,
		0.152218,
		0.165103,
		0.178197,
		0.191473,
		0.204903,
		0.218463,
		0.232129,
		0.245878,
		0.259691,
		0.273547,
		0.287427,
		0.301315,
		0.315194,
		0.32905,
		0.342867,
		0.356632,
		0.370333,
		0.383959,
		0.397498,
		0.410941,
		0.424278,
		0.437501,
		0.450601,
		0.463572,
		0.476406,
		0.489096,
		0.501639,
		0.514027,
		0.526257,
		0.538323,
		0.550223,
		0.561953,
		0.573509,
		0.584888,
		0.596089,
		0.607109,
		0.617947,
		0.628601,
		0.639069,
		0.649351,
		0.659446,
		0.669354,
		0.679074,
		0.688606,
		0.69795,
		0.000654,
		0.002556,
		0.00562,
		0.009762,
		0.014907,
		0.02098,
		0.027912,
		0.035637,
		0.044093,
		0.053221,
		0.062966,
		0.073275,
		0.084099,
		0.095391,
		0.107106,
		0.119202,
		0.131641,
		0.144385,
		0.1574,
		0.170651,
		0.184108,
		0.197741,
		0.211524,
		0.225429,
		0.239433,
		0.253513,
		0.267647,
		0.281814,
		0.295997,
		0.310176,
		0.324337,
		0.338462,
		0.352538,
		0.366551,
		0.380488,
		0.394338,
		0.40809,
		0.421733,
		0.435259,
		0.448658,
		0.461923,
		0.475047,
		0.488022,
		0.500843,
		0.513503,
		0.525998,
		0.538323,
		0.550474,
		0.562448,
		0.57424,
		0.585848,
		0.59727,
		0.608503,
		0.619546,
		0.630396,
		0.641052,
		0.651515,
		0.661782,
		0.671853,
		0.681729,
		0.691409,
		0.700893,
		0.710181,
		0.719275,
		0.728175,
		0.736881,
		0.745396,
		0.000682,
		0.002666,
		0.005858,
		0.010171,
		0.015523,
		0.021837,
		0.029038,
		0.037057,
		0.045828,
		0.055289,
		0.065382,
		0.076051,
		0.087245,
		0.098913,
		0.11101,
		0.123492,
		0.136317,
		0.149448,
		0.162846,
		0.176478,
		0.190312,
		0.204316,
		0.218463,
		0.232725,
		0.247077,
		0.261496,
		0.27596,
		0.290446,
		0.304937,
		0.319414,
		0.333861,
		0.34826,
		0.362598,
		0.37686,
		0.391034,
		0.405109,
		0.419073,
		0.432915,
		0.446628,
		0.460201,
		0.473628,
		0.4869,
		0.500011,
		0.512956,
		0.525728,
		0.538323,
		0.550737,
		0.562964,
		0.575003,
		0.586849,
		0.5985,
		0.609954,
		0.621209,
		0.632263,
		0.643115,
		0.653763,
		0.664208,
		0.674449,
		0.684484,
		0.694316,
		0.703943,
		0.713366,
		0.722587,
		0.731605,
		0.740422,
		0.749038,
		0.757456,
		0.765677,
		0.773702,
		0.781532,
		0.789171,
		0.000713,
		0.002783,
		0.006111,
		0.010606,
		0.016179,
		0.022747,
		0.030233,
		0.038563,
		0.047667,
		0.057479,
		0.067939,
		0.078987,
		0.090569,
		0.102633,
		0.11513,
		0.128015,
		0.141244,
		0.154777,
		0.168575,
		0.182603,
		0.196828,
		0.211216,
		0.22574,
		0.24037,
		0.255081,
		0.269849,
		0.28465,
		0.299464,
		0.31427,
		0.32905,
		0.343786,
		0.358463,
		0.373065,
		0.387578,
		0.40199,
		0.416289,
		0.430464,
		0.444504,
		0.4584,
		0.472143,
		0.485726,
		0.499142,
		0.512384,
		0.525446,
		0.538323,
		0.55101,
		0.563503,
		0.575799,
		0.587893,
		0.599783,
		0.611467,
		0.622942,
		0.634207,
		0.645261,
		0.656102,
		0.66673,
		0.677145,
		0.687346,
		0.697333,
		0.707107,
		0.716668,
		0.726017,
		0.735156,
		0.744084,
		0.752804,
		0.761317,
		0.769624,
		0.777728,
		0.78563,
		0.793332,
		0.800837,
		0.808146,
		0.815262,
		0.822187,
		0.828924,
		0.000745,
		0.002907,
		0.006382,
		0.011069,
		0.016877,
		0.023716,
		0.031504,
		0.040162,
		0.049618,
		0.059802,
		0.070648,
		0.082095,
		0.094085,
		0.106564,
		0.119481,
		0.132788,
		0.146439,
		0.160392,
		0.174607,
		0.189047,
		0.203676,
		0.218463,
		0.233376,
		0.248386,
		0.263466,
		0.278592,
		0.29374,
		0.308888,
		0.324015,
		0.339103,
		0.354133,
		0.369091,
		0.383959,
		0.398725,
		0.413374,
		0.427896,
		0.442279,
		0.456513,
		0.470589,
		0.484498,
		0.498233,
		0.511786,
		0.525152,
		0.538323,
		0.551297,
		0.564067,
		0.57663,
		0.588982,
		0.601121,
		0.613044,
		0.624748,
		0.636233,
		0.647496,
		0.658536,
		0.669354,
		0.679948,
		0.690319,
		0.700466,
		0.71039,
		0.720092,
		0.729573,
		0.738833,
		0.747875,
		0.756699,
		0.765307,
		0.773702,
		0.781884,
		0.789856,
		0.79762,
		0.805179,
		0.812535,
		0.819691,
		0.826649,
		0.833411,
		0.839981,
		0.846361,
		0.852555,
		0.858565,
		0.864394,
		0.00078,
		0.003041,
		0.006671,
		0.011564,
		0.017621,
		0.024747,
		0.032856,
		0.041863,
		0.051692,
		0.062267,
		0.073521,
		0.085389,
		0.097809,
		0.110724,
		0.124081,
		0.137829,
		0.151921,
		0.166312,
		0.180962,
		0.19583,
		0.21088,
		0.226079,
		0.241394,
		0.256796,
		0.272257,
		0.28775,
		0.303253,
		0.318742,
		0.334196,
		0.349596,
		0.364925,
		0.380165,
		0.395301,
		0.410318,
		0.425205,
		0.439948,
		0.454537,
		0.468961,
		0.483212,
		0.497281,
		0.51116,
		0.524843,
		0.538323,
		0.551596,
		0.564656,
		0.577498,
		0.59012,
		0.602518,
		0.61469,
		0.626633,
		0.638345,
		0.649825,
		0.661072,
		0.672085,
		0.682865,
		0.69341,
		0.703721,
		0.7138,
		0.723646,
		0.73326,
		0.742645,
		0.751801,
		0.76073,
		0.769433,
		0.777914,
		0.786174,
		0.794215,
		0.80204,
		0.809651,
		0.817052,
		0.824244,
		0.83123,
		0.838015,
		0.8446,
		0.850988,
		0.857183,
		0.863189,
		0.869007,
		0.874642,
		0.880097,
		0.885375,
		0.89048,
		0.895415,
		0.000817,
		0.003183,
		0.006979,
		0.012092,
		0.018415,
		0.025848,
		0.034297,
		0.043674,
		0.053897,
		0.064888,
		0.076573,
		0.088884,
		0.101756,
		0.11513,
		0.128949,
		0.14316,
		0.157713,
		0.172561,
		0.187663,
		0.202976,
		0.218463,
		0.234088,
		0.249819,
		0.265625,
		0.281477,
		0.297347,
		0.313213,
		0.32905,
		0.344836,
		0.360554,
		0.376183,
		0.391707,
		0.407111,
		0.42238,
		0.437501,
		0.452462,
		0.467253,
		0.481862,
		0.496282,
		0.510503,
		0.524519,
		0.538323,
		0.551909,
		0.565272,
		0.578407,
		0.591311,
		0.603979,
		0.61641,
		0.628601,
		0.640549,
		0.652255,
		0.663715,
		0.674931,
		0.685901,
		0.696627,
		0.707107,
		0.717343,
		0.727336,
		0.737086,
		0.746597,
		0.755868,
		0.764902,
		0.773702,
		0.782268,
		0.790604,
		0.798713,
		0.806596,
		0.814257,
		0.821699,
		0.828924,
		0.835936,
		0.842738,
		0.849334,
		0.855726,
		0.861918,
		0.867913,
		0.873716,
		0.879329,
		0.884756,
		0.890001,
		0.895068,
		0.899959,
		0.90468,
		0.909232,
		0.913621,
		0.91785,
		0.921922,
		0.000857,
		0.003336,
		0.00731,
		0.012657,
		0.019264,
		0.027023,
		0.035835,
		0.045605,
		0.056247,
		0.067676,
		0.079817,
		0.092596,
		0.105945,
		0.119801,
		0.134105,
		0.148801,
		0.163836,
		0.179163,
		0.194735,
		0.210511,
		0.226451,
		0.242518,
		0.258679,
		0.2749,
		0.291153,
		0.307411,
		0.323646,
		0.339837,
		0.355962,
		0.371999,
		0.387931,
		0.403741,
		0.419412,
		0.43493,
		0.450283,
		0.465459,
		0.480445,
		0.495233,
		0.509814,
		0.52418,
		0.538323,
		0.552238,
		0.565918,
		0.579359,
		0.592557,
		0.605508,
		0.618209,
		0.630658,
		0.642853,
		0.654791,
		0.666474,
		0.677898,
		0.689066,
		0.699976,
		0.710629,
		0.721027,
		0.73117,
		0.741059,
		0.750696,
		0.760084,
		0.769224,
		0.778118,
		0.78677,
		0.795181,
		0.803355,
		0.811294,
		0.819002,
		0.826481,
		0.833736,
		0.840769,
		0.847584,
		0.854185,
		0.860575,
		0.866759,
		0.872739,
		0.878519,
		0.884104,
		0.889497,
		0.894703,
		0.899725,
		0.904566,
		0.909232,
		0.913726,
		0.918052,
		0.922214,
		0.926215,
		0.930061,
		0.933754,
		0.937299,
		0.940699,
		0.943959,
		0.000899,
		0.003501,
		0.007665,
		0.013263,
		0.020173,
		0.02828,
		0.037478,
		0.047667,
		0.058753,
		0.070648,
		0.08327,
		0.096544,
		0.110396,
		0.12476,
		0.139573,
		0.154777,
		0.170317,
		0.186142,
		0.202206,
		0.218463,
		0.234873,
		0.251397,
		0.268,
		0.28465,
		0.301315,
		0.317968,
		0.334582,
		0.351133,
		0.367599,
		0.383959,
		0.400195,
		0.416289,
		0.432226,
		0.447991,
		0.463572,
		0.478955,
		0.494131,
		0.50909,
		0.523824,
		0.538323,
		0.552583,
		0.566596,
		0.580358,
		0.593863,
		0.607109,
		0.620093,
		0.63281,
		0.645261,
		0.657442,
		0.669354,
		0.680995,
		0.692366,
		0.703467,
		0.714298,
		0.72486,
		0.735156,
		0.745185,
		0.754952,
		0.764456,
		0.773702,
		0.78269,
		0.791425,
		0.79991,
		0.808146,
		0.816138,
		0.823889,
		0.831403,
		0.838682,
		0.845732,
		0.852555,
		0.859156,
		0.865538,
		0.871707,
		0.877665,
		0.883417,
		0.888967,
		0.894319,
		0.899478,
		0.904447,
		0.909232,
		0.913836,
		0.918264,
		0.922519,
		0.926607,
		0.930531,
		0.934295,
		0.937904,
		0.941362,
		0.944673,
		0.947842,
		0.950871,
		0.953766,
		0.95653,
		0.959167,
		0.961681,
		0.000945,
		0.003677,
		0.008046,
		0.013913,
		0.021147,
		0.029626,
		0.039236,
		0.049871,
		0.061429,
		0.073818,
		0.086952,
		0.100747,
		0.11513,
		0.130029,
		0.145377,
		0.161115,
		0.177183,
		0.193529,
		0.210104,
		0.226862,
		0.243759,
		0.260756,
		0.277816,
		0.294906,
		0.311993,
		0.32905,
		0.346048,
		0.362964,
		0.379775,
		0.39646,
		0.413,
		0.429378,
		0.445578,
		0.461585,
		0.477387,
		0.492972,
		0.508329,
		0.523449,
		0.538323,
		0.552945,
		0.567308,
		0.581406,
		0.595234,
		0.608789,
		0.622067,
		0.635065,
		0.647781,
		0.660215,
		0.672365,
		0.68423,
		0.69581,
		0.707107,
		0.71812,
		0.728851,
		0.739302,
		0.749475,
		0.759371,
		0.768993,
		0.778343,
		0.787425,
		0.796242,
		0.804797,
		0.813093,
		0.821134,
		0.828924,
		0.836467,
		0.843766,
		0.850827,
		0.857652,
		0.864247,
		0.870615,
		0.876762,
		0.88269,
		0.888407,
		0.893914,
		0.899218,
		0.904322,
		0.909232,
		0.913952,
		0.918486,
		0.92284,
		0.927017,
		0.931022,
		0.93486,
		0.938536,
		0.942053,
		0.945417,
		0.948631,
		0.951701,
		0.95463,
		0.957423,
		0.960083,
		0.962616,
		0.965026,
		0.967315,
		0.969489,
		0.971552,
		0.973507,
		0.975358,
		0.000995,
		0.003868,
		0.008457,
		0.014613,
		0.022194,
		0.031071,
		0.041121,
		0.05223,
		0.064291,
		0.077206,
		0.09088,
		0.105229,
		0.120172,
		0.135635,
		0.151546,
		0.167843,
		0.184465,
		0.201355,
		0.218463,
		0.23574,
		0.253142,
		0.270627,
		0.288158,
		0.3057,
		0.32322,
		0.340688,
		0.358077,
		0.375363,
		0.392521,
		0.409531,
		0.426374,
		0.443032,
		0.45949,
		0.475734,
		0.491749,
		0.507526,
		0.523054,
		0.538323,
		0.553326,
		0.568056,
		0.582507,
		0.596674,
		0.610552,
		0.624138,
		0.637429,
		0.650423,
		0.663118,
		0.675515,
		0.687611,
		0.699408,
		0.710906,
		0.722107,
		0.73301,
		0.743619,
		0.753936,
		0.763962,
		0.773702,
		0.783156,
		0.79233,
		0.801227,
		0.809849,
		0.818201,
		0.826287,
		0.834112,
		0.841678,
		0.848992,
		0.856057,
		0.862877,
		0.869459,
		0.875806,
		0.881923,
		0.887815,
		0.893487,
		0.898944,
		0.904191,
		0.909232,
		0.914074,
		0.91872,
		0.923176,
		0.927446,
		0.931537,
		0.935451,
		0.939196,
		0.942774,
		0.946192,
		0.949453,
		0.952563,
		0.955526,
		0.958348,
		0.961031,
		0.963582,
		0.966004,
		0.968302,
		0.97048,
		0.972543,
		0.974494,
		0.976338,
		0.978079,
		0.97972,
		0.981267,
		0.982721,
		0.984089,
		0.985372,
		0.001049,
		0.004073,
		0.0089,
		0.015366,
		0.023321,
		0.032625,
		0.043145,
		0.054761,
		0.067358,
		0.08083,
		0.09508,
		0.110014,
		0.12555,
		0.141606,
		0.15811,
		0.174994,
		0.192195,
		0.209654,
		0.227317,
		0.245133,
		0.263058,
		0.281048,
		0.299063,
		0.317068,
		0.33503,
		0.352917,
		0.370703,
		0.38836,
		0.405867,
		0.423201,
		0.440344,
		0.457278,
		0.473988,
		0.49046,
		0.50668,
		0.522638,
		0.538323,
		0.553728,
		0.568845,
		0.583667,
		0.598188,
		0.612405,
		0.626313,
		0.63991,
		0.653193,
		0.666161,
		0.678813,
		0.69115,
		0.70317,
		0.714875,
		0.726267,
		0.737347,
		0.748116,
		0.758579,
		0.768736,
		0.778592,
		0.78815,
		0.797413,
		0.806386,
		0.815072,
		0.823476,
		0.831602,
		0.839455,
		0.84704,
		0.854361,
		0.861423,
		0.868232,
		0.874792,
		0.881109,
		0.887188,
		0.893035,
		0.898654,
		0.904051,
		0.909232,
		0.914202,
		0.918966,
		0.923529,
		0.927897,
		0.932076,
		0.93607,
		0.939885,
		0.943527,
		0.946999,
		0.950308,
		0.953459,
		0.956457,
		0.959306,
		0.962012,
		0.964579,
		0.967013,
		0.969317,
		0.971498,
		0.973558,
		0.975504,
		0.977338,
		0.979066,
		0.980692,
		0.98222,
		0.983655,
		0.984999,
		0.986257,
		0.987433,
		0.988531,
		0.989554,
		0.990506,
		0.99139,
		0.99221,
		0.001107,
		0.004296,
		0.009378,
		0.016179,
		0.024536,
		0.034297,
		0.045322,
		0.057479,
		0.070648,
		0.084714,
		0.099574,
		0.11513,
		0.131291,
		0.147975,
		0.165103,
		0.182603,
		0.200411,
		0.218463,
		0.236704,
		0.255081,
		0.273547,
		0.292057,
		0.31057,
		0.32905,
		0.347461,
		0.365774,
		0.383959,
		0.40199,
		0.419845,
		0.437501,
		0.45494,
		0.472143,
		0.489096,
		0.505785,
		0.522198,
		0.538323,
		0.554152,
		0.569676,
		0.584888,
		0.599783,
		0.614355,
		0.628601,
		0.642517,
		0.656102,
		0.669354,
		0.682272,
		0.694856,
		0.707107,
		0.719025,
		0.730613,
		0.741872,
		0.752804,
		0.763413,
		0.773702,
		0.783673,
		0.793332,
		0.802682,
		0.811728,
		0.820474,
		0.828924,
		0.837084,
		0.84496,
		0.852555,
		0.859876,
		0.866927,
		0.873716,
		0.880246,
		0.886525,
		0.892557,
		0.898348,
		0.903905,
		0.909232,
		0.914337,
		0.919224,
		0.9239,
		0.928371,
		0.932642,
		0.936718,
		0.940607,
		0.944313,
		0.947842,
		0.951199,
		0.954391,
		0.957423,
		0.960299,
		0.963026,
		0.965609,
		0.968053,
		0.970362,
		0.972543,
		0.974599,
		0.976536,
		0.978359,
		0.980072,
		0.98168,
		0.983187,
		0.984597,
		0.985916,
		0.987147,
		0.988294,
		0.989361,
		0.990352,
		0.991272,
		0.992122,
		0.992908,
		0.993633,
		0.9943,
		0.994912,
		0.995473,
		0.995985,
		0.996452,
		0.00117,
		0.004537,
		0.009896,
		0.017058,
		0.025848,
		0.036101,
		0.047667,
		0.060404,
		0.074184,
		0.088884,
		0.104393,
		0.120607,
		0.137432,
		0.154777,
		0.172561,
		0.19071,
		0.209152,
		0.227824,
		0.246666,
		0.265625,
		0.28465,
		0.303696,
		0.32272,
		0.341684,
		0.360554,
		0.379296,
		0.397884,
		0.416289,
		0.434489,
		0.452462,
		0.470189,
		0.487653,
		0.504839,
		0.521733,
		0.538323,
		0.5546,
		0.570554,
		0.586177,
		0.601465,
		0.61641,
		0.63101,
		0.645261,
		0.65916,
		0.672708,
		0.685901,
		0.698742,
		0.71123,
		0.723368,
		0.735156,
		0.746597,
		0.757694,
		0.76845,
		0.778869,
		0.788955,
		0.798713,
		0.808146,
		0.81726,
		0.82606,
		0.834551,
		0.842738,
		0.850628,
		0.858226,
		0.865538,
		0.872571,
		0.879329,
		0.88582,
		0.892049,
		0.898023,
		0.903749,
		0.909232,
		0.914479,
		0.919497,
		0.924291,
		0.928869,
		0.933236,
		0.937398,
		0.941362,
		0.945135,
		0.948721,
		0.952128,
		0.955361,
		0.958426,
		0.961329,
		0.964076,
		0.966673,
		0.969125,
		0.971437,
		0.973615,
		0.975665,
		0.977591,
		0.9794,
		0.981094,
		0.982681,
		0.984164,
		0.985548,
		0.986838,
		0.988039,
		0.989154,
		0.990188,
		0.991144,
		0.992029,
		0.992844,
		0.993593,
		0.994282,
		0.994912,
		0.995488,
		0.996013,
		0.99649,
		0.996922,
		0.997312,
		0.997663,
		0.997978,
		0.99826,
		0.998511,
		0.998733,
		0.001239,
		0.004799,
		0.010458,
		0.018011,
		0.027267,
		0.038051,
		0.050198,
		0.063557,
		0.07799,
		0.093366,
		0.109566,
		0.12648,
		0.144007,
		0.162051,
		0.180527,
		0.199356,
		0.218463,
		0.237781,
		0.25725,
		0.276811,
		0.296414,
		0.31601,
		0.335557,
		0.355015,
		0.374349,
		0.393526,
		0.412516,
		0.431293,
		0.449834,
		0.468117,
		0.486123,
		0.503836,
		0.52124,
		0.538323,
		0.555074,
		0.571482,
		0.58754,
		0.603241,
		0.618579,
		0.63355,
		0.648151,
		0.66238,
		0.676234,
		0.689715,
		0.702821,
		0.715554,
		0.727916,
		0.739908,
		0.751535,
		0.762798,
		0.773702,
		0.78425,
		0.794448,
		0.804301,
		0.813813,
		0.82299,
		0.831837,
		0.840361,
		0.848568,
		0.856465,
		0.864056,
		0.87135,
		0.878352,
		0.88507,
		0.89151,
		0.897679,
		0.903584,
		0.909232,
		0.91463,
		0.919785,
		0.924704,
		0.929393,
		0.93386,
		0.938112,
		0.942154,
		0.945995,
		0.94964,
		0.953097,
		0.956371,
		0.959469,
		0.962398,
		0.965164,
		0.967772,
		0.97023,
		0.972543,
		0.974716,
		0.976757,
		0.978669,
		0.980459,
		0.982133,
		0.983695,
		0.985151,
		0.986506,
		0.987764,
		0.988931,
		0.990011,
		0.991008,
		0.991928,
		0.992774,
		0.993551,
		0.994262,
		0.994912,
		0.995504,
		0.996042,
		0.996529,
		0.99697,
		0.997366,
		0.997721,
		0.998039,
		0.998321,
		0.998572,
		0.998793,
		0.998987,
		0.999156,
		0.999303,
		0.999429,
		0.999537,
		0.999629,
		0.999707,
		0.001314,
		0.005084,
		0.011069,
		0.019046,
		0.028807,
		0.040162,
		0.052935,
		0.066963,
		0.082095,
		0.098193,
		0.11513,
		0.132788,
		0.151058,
		0.169842,
		0.189047,
		0.208589,
		0.228392,
		0.248386,
		0.268504,
		0.28869,
		0.308888,
		0.32905,
		0.349131,
		0.369091,
		0.388893,
		0.408505,
		0.427896,
		0.447041,
		0.465916,
		0.484498,
		0.502771,
		0.520718,
		0.538323,
		0.555576,
		0.572465,
		0.588982,
		0.605119,
		0.620871,
		0.636233,
		0.651201,
		0.665773,
		0.679948,
		0.693726,
		0.707107,
		0.720092,
		0.732684,
		0.744885,
		0.756699,
		0.768129,
		0.77918,
		0.789856,
		0.800163,
		0.810106,
		0.819691,
		0.828924,
		0.837812,
		0.846361,
		0.854579,
		0.862471,
		0.870045,
		0.877309,
		0.88427,
		0.890936,
		0.897313,
		0.903409,
		0.909232,
		0.91479,
		0.92009,
		0.92514,
		0.929947,
		0.934518,
		0.938862,
		0.942986,
		0.946896,
		0.950601,
		0.954108,
		0.957423,
		0.960553,
		0.963506,
		0.966289,
		0.968908,
		0.971369,
		0.973679,
		0.975845,
		0.977873,
		0.979768,
		0.981538,
		0.983187,
		0.984721,
		0.986146,
		0.987468,
		0.988691,
		0.989821,
		0.990862,
		0.991821,
		0.9927,
		0.993506,
		0.994242,
		0.994912,
		0.995521,
		0.996073,
		0.996571,
		0.99702,
		0.997422,
		0.997782,
		0.998102,
		0.998385,
		0.998635,
		0.998854,
		0.999046,
		0.999212,
		0.999355,
		0.999477,
		0.999581,
		0.999668,
		0.999741,
		0.999801,
		0.99985,
		0.999889,
		0.99992,
		0.999944,
		0.999962,
		0.999975,
		0.001396,
		0.005396,
		0.011736,
		0.020173,
		0.030481,
		0.042454,
		0.055902,
		0.070648,
		0.086531,
		0.103402,
		0.121124,
		0.139573,
		0.158632,
		0.178197,
		0.19817,
		0.218463,
		0.238994,
		0.259691,
		0.280485,
		0.301315,
		0.322126,
		0.342867,
		0.363491,
		0.383959,
		0.404232,
		0.424278,
		0.444067,
		0.463572,
		0.482769,
		0.501639,
		0.520162,
		0.538323,
		0.556109,
		0.573509,
		0.590511,
		0.607109,
		0.623297,
		0.639069,
		0.654422,
		0.669354,
		0.683863,
		0.69795,
		0.711615,
		0.72486,
		0.737688,
		0.750101,
		0.762104,
		0.773702,
		0.784898,
		0.795699,
		0.80611,
		0.816138,
		0.82579,
		0.835071,
		0.843991,
		0.852555,
		0.860772,
		0.868649,
		0.876195,
		0.883417,
		0.890323,
		0.896922,
		0.903223,
		0.909232,
		0.91496,
		0.920413,
		0.925601,
		0.930531,
		0.935212,
		0.939652,
		0.943859,
		0.947842,
		0.951607,
		0.955164,
		0.958519,
		0.961681,
		0.964657,
		0.967454,
		0.970081,
		0.972543,
		0.974848,
		0.977002,
		0.979014,
		0.980889,
		0.982633,
		0.984253,
		0.985756,
		0.987147,
		0.988431,
		0.989616,
		0.990705,
		0.991705,
		0.992621,
		0.993458,
		0.99422,
		0.994912,
		0.995539,
		0.996106,
		0.996615,
		0.997073,
		0.997481,
		0.997845,
		0.998167,
		0.998451,
		0.9987,
		0.998917,
		0.999106,
		0.999268,
		0.999407,
		0.999525,
		0.999624,
		0.999707,
		0.999775,
		0.99983,
		0.999874,
		0.999909,
		0.999936,
		0.999957,
		0.999972,
		0.999982,
		0.99999,
		0.999995,
		0.999997,
		0.999999,
		1,
		1,
		1,
		0.001486,
		0.005737,
		0.012464,
		0.021402,
		0.032305,
		0.044948,
		0.059124,
		0.074644,
		0.091333,
		0.109032,
		0.127594,
		0.146884,
		0.166781,
		0.187172,
		0.207954,
		0.229035,
		0.250328,
		0.271757,
		0.293251,
		0.314747,
		0.336187,
		0.357518,
		0.378694,
		0.399673,
		0.420418,
		0.440894,
		0.461072,
		0.480925,
		0.500431,
		0.51957,
		0.538323,
		0.556677,
		0.574618,
		0.592135,
		0.609221,
		0.625869,
		0.642073,
		0.657831,
		0.673139,
		0.687996,
		0.702404,
		0.716363,
		0.729875,
		0.742944,
		0.755573,
		0.767766,
		0.77953,
		0.790869,
		0.801791,
		0.812301,
		0.822408,
		0.832117,
		0.841438,
		0.850378,
		0.858946,
		0.867151,
		0.875,
		0.882503,
		0.889668,
		0.896506,
		0.903024,
		0.909232,
		0.91514,
		0.920755,
		0.926089,
		0.931149,
		0.935944,
		0.940484,
		0.944778,
		0.948834,
		0.952661,
		0.956268,
		0.959663,
		0.962855,
		0.965851,
		0.968661,
		0.971292,
		0.973751,
		0.976047,
		0.978187,
		0.980179,
		0.982029,
		0.983744,
		0.985332,
		0.986798,
		0.98815,
		0.989394,
		0.990536,
		0.991581,
		0.992536,
		0.993406,
		0.994196,
		0.994912,
		0.995558,
		0.99614,
		0.996662,
		0.997128,
		0.997543,
		0.99791,
		0.998234,
		0.998518,
		0.998766,
		0.998981,
		0.999166,
		0.999324,
		0.999459,
		0.999572,
		0.999666,
		0.999743,
		0.999806,
		0.999857,
		0.999897,
		0.999927,
		0.999951,
		0.999968,
		0.99998,
		0.999988,
		0.999994,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999995,
		0.999991,
		0.001584,
		0.006111,
		0.013263,
		0.022747,
		0.034297,
		0.047667,
		0.062632,
		0.078987,
		0.096544,
		0.11513,
		0.134589,
		0.154777,
		0.175563,
		0.196828,
		0.218463,
		0.24037,
		0.262459,
		0.28465,
		0.306869,
		0.32905,
		0.351133,
		0.373065,
		0.394798,
		0.416289,
		0.437501,
		0.4584,
		0.478955,
		0.499142,
		0.518938,
		0.538323,
		0.557281,
		0.575799,
		0.593863,
		0.611467,
		0.628601,
		0.645261,
		0.661443,
		0.677145,
		0.692366,
		0.707107,
		0.721369,
		0.735156,
		0.74847,
		0.761317,
		0.773702,
		0.78563,
		0.797109,
		0.808146,
		0.818748,
		0.828924,
		0.838682,
		0.848031,
		0.85698,
		0.865538,
		0.873716,
		0.881522,
		0.888967,
		0.89606,
		0.902812,
		0.909232,
		0.915332,
		0.92112,
		0.926607,
		0.931803,
		0.936718,
		0.941362,
		0.945745,
		0.949876,
		0.953766,
		0.957423,
		0.960857,
		0.964076,
		0.967092,
		0.969911,
		0.972543,
		0.974996,
		0.977279,
		0.9794,
		0.981366,
		0.983187,
		0.984868,
		0.986419,
		0.987845,
		0.989154,
		0.990352,
		0.991447,
		0.992444,
		0.99335,
		0.994171,
		0.994912,
		0.995579,
		0.996177,
		0.996711,
		0.997186,
		0.997607,
		0.997978,
		0.998304,
		0.998588,
		0.998834,
		0.999046,
		0.999227,
		0.999381,
		0.99951,
		0.999618,
		0.999707,
		0.999779,
		0.999836,
		0.999882,
		0.999917,
		0.999944,
		0.999963,
		0.999977,
		0.999987,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999994,
		0.999989,
		0.999982,
		0.999973,
		0.999959,
		0.999942,
		0.99992,
		0.999892,
		0.999859,
		0.999819,
		0.001693,
		0.006522,
		0.014139,
		0.024222,
		0.036478,
		0.050639,
		0.06646,
		0.083718,
		0.102209,
		0.121749,
		0.142168,
		0.163312,
		0.185043,
		0.207232,
		0.229766,
		0.25254,
		0.27546,
		0.298442,
		0.321409,
		0.344293,
		0.367032,
		0.389573,
		0.411865,
		0.433865,
		0.455537,
		0.476846,
		0.497762,
		0.518262,
		0.538323,
		0.557927,
		0.577059,
		0.595706,
		0.613858,
		0.631507,
		0.648648,
		0.665277,
		0.681392,
		0.696992,
		0.712079,
		0.726654,
		0.740722,
		0.754286,
		0.767353,
		0.779928,
		0.792019,
		0.803632,
		0.814777,
		0.825463,
		0.835698,
		0.845492,
		0.854856,
		0.863799,
		0.872333,
		0.880467,
		0.888213,
		0.895582,
		0.902584,
		0.909232,
		0.915536,
		0.921508,
		0.927158,
		0.932498,
		0.937538,
		0.94229,
		0.946765,
		0.950973,
		0.954925,
		0.958631,
		0.962103,
		0.965348,
		0.968379,
		0.971204,
		0.973833,
		0.976276,
		0.978541,
		0.980638,
		0.982575,
		0.984361,
		0.986004,
		0.987511,
		0.988892,
		0.990153,
		0.991302,
		0.992346,
		0.993291,
		0.994144,
		0.994912,
		0.995601,
		0.996216,
		0.996763,
		0.997247,
		0.997675,
		0.998049,
		0.998376,
		0.998659,
		0.998903,
		0.999112,
		0.999289,
		0.999437,
		0.999561,
		0.999663,
		0.999746,
		0.999812,
		0.999864,
		0.999905,
		0.999935,
		0.999958,
		0.999974,
		0.999985,
		0.999992,
		0.999996,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999988,
		0.99998,
		0.999969,
		0.999954,
		0.999934,
		0.999909,
		0.999878,
		0.99984,
		0.999795,
		0.999741,
		0.999679,
		0.999606,
		0.999523,
		0.999429,
		0.999323,
		0.999205,
		0.999074,
		0.001807,
		0.006962,
		0.015085,
		0.025825,
		0.038856,
		0.053884,
		0.07064,
		0.088879,
		0.108382,
		0.128948,
		0.150396,
		0.172561,
		0.195295,
		0.218463,
		0.241943,
		0.265625,
		0.289411,
		0.313213,
		0.336951,
		0.360554,
		0.383959,
		0.407111,
		0.42996,
		0.452462,
		0.474581,
		0.496282,
		0.517538,
		0.538323,
		0.558619,
		0.578407,
		0.597675,
		0.61641,
		0.634605,
		0.652255,
		0.669354,
		0.685901,
		0.701897,
		0.717343,
		0.732241,
		0.746597,
		0.760415,
		0.773702,
		0.786465,
		0.798713,
		0.810454,
		0.821699,
		0.832457,
		0.842738,
		0.852555,
		0.861918,
		0.870838,
		0.879329,
		0.887401,
		0.895068,
		0.902341,
		0.909232,
		0.915755,
		0.921922,
		0.927745,
		0.933236,
		0.938407,
		0.943272,
		0.947842,
		0.952128,
		0.956143,
		0.959898,
		0.963404,
		0.966673,
		0.969716,
		0.972543,
		0.975164,
		0.977591,
		0.979834,
		0.981901,
		0.983803,
		0.985548,
		0.987147,
		0.988607,
		0.989936,
		0.991144,
		0.992239,
		0.993226,
		0.994115,
		0.994912,
		0.995624,
		0.996257,
		0.996818,
		0.997312,
		0.997745,
		0.998123,
		0.998451,
		0.998733,
		0.998974,
		0.999178,
		0.99935,
		0.999493,
		0.999611,
		0.999707,
		0.999783,
		0.999844,
		0.99989,
		0.999926,
		0.999951,
		0.99997,
		0.999982,
		0.999991,
		0.999996,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999996,
		0.999992,
		0.999986,
		0.999977,
		0.999965,
		0.999947,
		0.999925,
		0.999897,
		0.999861,
		0.999819,
		0.999767,
		0.999707,
		0.999636,
		0.999554,
		0.999461,
		0.999355,
		0.999236,
		0.999103,
		0.998956,
		0.998793,
		0.998614,
		0.99842,
		0.998208,
		0.997978,
		0.997731,
		0.997465,
		0.99718,
		0.001912,
		0.007387,
		0.016032,
		0.027471,
		0.041348,
		0.057334,
		0.075127,
		0.094456,
		0.115075,
		0.136765,
		0.159331,
		0.182595,
		0.2064,
		0.230604,
		0.255081,
		0.279714,
		0.304401,
		0.32905,
		0.353578,
		0.377913,
		0.40199,
		0.425753,
		0.449152,
		0.472143,
		0.494689,
		0.516759,
		0.538323,
		0.559361,
		0.579853,
		0.599783,
		0.61914,
		0.637915,
		0.656102,
		0.673697,
		0.690698,
		0.707107,
		0.722924,
		0.738155,
		0.752804,
		0.766878,
		0.780384,
		0.793332,
		0.805731,
		0.817592,
		0.828924,
		0.839741,
		0.850054,
		0.859876,
		0.869219,
		0.878098,
		0.886525,
		0.894513,
		0.902078,
		0.909232,
		0.91599,
		0.922365,
		0.928371,
		0.934022,
		0.939331,
		0.944313,
		0.94898,
		0.953345,
		0.957423,
		0.961225,
		0.964764,
		0.968053,
		0.971103,
		0.973927,
		0.976536,
		0.978942,
		0.981155,
		0.983187,
		0.985047,
		0.986746,
		0.988294,
		0.9897,
		0.990973,
		0.992122,
		0.993157,
		0.994084,
		0.994912,
		0.995649,
		0.996301,
		0.996876,
		0.99738,
		0.997819,
		0.9982,
		0.998528,
		0.998808,
		0.999046,
		0.999246,
		0.999412,
		0.999548,
		0.99966,
		0.999749,
		0.999819,
		0.999873,
		0.999914,
		0.999944,
		0.999965,
		0.99998,
		0.999989,
		0.999995,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999996,
		0.999991,
		0.999984,
		0.999974,
		0.999959,
		0.99994,
		0.999914,
		0.999882,
		0.999842,
		0.999793,
		0.999735,
		0.999666,
		0.999586,
		0.999493,
		0.999388,
		0.999268,
		0.999134,
		0.998984,
		0.998818,
		0.998635,
		0.998435,
		0.998216,
		0.997978,
		0.997721,
		0.997444,
		0.997147,
		0.996829,
		0.99649,
		0.996128,
		0.995745,
		0.99534,
		0.994912,
		0.994461,
		0.993987,
		0.99349,
		0.001978,
		0.007692,
		0.016784,
		0.028884,
		0.043618,
		0.060626,
		0.079568,
		0.100129,
		0.122028,
		0.14501,
		0.168853,
		0.19336,
		0.218359,
		0.243699,
		0.269249,
		0.294891,
		0.320521,
		0.346046,
		0.371384,
		0.39646,
		0.42121,
		0.445578,
		0.469512,
		0.492972,
		0.515919,
		0.538323,
		0.560159,
		0.581406,
		0.602046,
		0.622067,
		0.641458,
		0.660215,
		0.678333,
		0.69581,
		0.712649,
		0.728851,
		0.744423,
		0.759371,
		0.773702,
		0.787425,
		0.800552,
		0.813093,
		0.82506,
		0.836467,
		0.847326,
		0.857652,
		0.867459,
		0.876762,
		0.885575,
		0.893914,
		0.901795,
		0.909232,
		0.916242,
		0.92284,
		0.929041,
		0.93486,
		0.940314,
		0.945417,
		0.950184,
		0.95463,
		0.958769,
		0.962616,
		0.966185,
		0.969489,
		0.972543,
		0.975358,
		0.977948,
		0.980326,
		0.982503,
		0.984492,
		0.986304,
		0.987949,
		0.98944,
		0.990785,
		0.991996,
		0.993081,
		0.99405,
		0.994912,
		0.995675,
		0.996348,
		0.996937,
		0.997451,
		0.997897,
		0.99828,
		0.998608,
		0.998886,
		0.999119,
		0.999313,
		0.999473,
		0.999603,
		0.999707,
		0.999788,
		0.999852,
		0.999899,
		0.999934,
		0.999959,
		0.999976,
		0.999987,
		0.999994,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999995,
		0.99999,
		0.999982,
		0.99997,
		0.999953,
		0.999931,
		0.999901,
		0.999864,
		0.999819,
		0.999763,
		0.999696,
		0.999618,
		0.999527,
		0.999422,
		0.999302,
		0.999166,
		0.999014,
		0.998845,
		0.998657,
		0.998451,
		0.998225,
		0.997978,
		0.997711,
		0.997422,
		0.997111,
		0.996778,
		0.996422,
		0.996042,
		0.995638,
		0.995211,
		0.994759,
		0.994282,
		0.99378,
		0.993253,
		0.9927,
		0.992122,
		0.991519,
		0.99089,
		0.990235,
		0.989554,
		0.988848,
		0.988116,
		0.987358,
		0.001962,
		0.007716,
		0.017013,
		0.029542,
		0.04496,
		0.062899,
		0.082994,
		0.104887,
		0.128242,
		0.152754,
		0.178148,
		0.204182,
		0.230652,
		0.257379,
		0.284216,
		0.311037,
		0.337737,
		0.364225,
		0.390425,
		0.416269,
		0.441699,
		0.466662,
		0.491113,
		0.515011,
		0.538323,
		0.561021,
		0.583079,
		0.604481,
		0.625212,
		0.645261,
		0.664622,
		0.683291,
		0.701268,
		0.718555,
		0.735156,
		0.751077,
		0.766326,
		0.780913,
		0.794849,
		0.808146,
		0.820817,
		0.832877,
		0.84434,
		0.855222,
		0.865538,
		0.875306,
		0.884543,
		0.893264,
		0.901488,
		0.909232,
		0.916514,
		0.92335,
		0.929759,
		0.935757,
		0.941362,
		0.946591,
		0.951461,
		0.955987,
		0.960187,
		0.964076,
		0.967671,
		0.970985,
		0.974035,
		0.976835,
		0.9794,
		0.981742,
		0.983876,
		0.985814,
		0.987569,
		0.989154,
		0.990579,
		0.991857,
		0.992999,
		0.994013,
		0.994912,
		0.995704,
		0.996398,
		0.997003,
		0.997527,
		0.997978,
		0.998364,
		0.99869,
		0.998965,
		0.999193,
		0.999381,
		0.999533,
		0.999656,
		0.999752,
		0.999826,
		0.999882,
		0.999923,
		0.999952,
		0.999972,
		0.999985,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999994,
		0.999988,
		0.999979,
		0.999965,
		0.999946,
		0.99992,
		0.999886,
		0.999844,
		0.999791,
		0.999727,
		0.999651,
		0.999561,
		0.999457,
		0.999337,
		0.9992,
		0.999046,
		0.998873,
		0.998681,
		0.998468,
		0.998234,
		0.997978,
		0.9977,
		0.997398,
		0.997073,
		0.996723,
		0.996348,
		0.995947,
		0.995521,
		0.995069,
		0.99459,
		0.994084,
		0.993551,
		0.992991,
		0.992403,
		0.991788,
		0.991144,
		0.990474,
		0.989775,
		0.989048,
		0.988294,
		0.987511,
		0.986701,
		0.985864,
		0.984999,
		0.984106,
		0.983187,
		0.98224,
		0.981267,
		0.980266,
		0.97924,
		0.978187,
		0.001825,
		0.007301,
		0.01635,
		0.028801,
		0.044403,
		0.062848,
		0.083785,
		0.106843,
		0.131646,
		0.157829,
		0.185047,
		0.212988,
		0.241376,
		0.269972,
		0.298578,
		0.327031,
		0.355202,
		0.382989,
		0.410314,
		0.437117,
		0.463351,
		0.488978,
		0.513969,
		0.538297,
		0.561943,
		0.584885,
		0.607109,
		0.628601,
		0.649351,
		0.669354,
		0.688606,
		0.707107,
		0.72486,
		0.741872,
		0.758149,
		0.773702,
		0.788542,
		0.802682,
		0.816138,
		0.828924,
		0.841057,
		0.852555,
		0.863435,
		0.873716,
		0.883417,
		0.892557,
		0.901155,
		0.909232,
		0.916807,
		0.9239,
		0.930531,
		0.936718,
		0.942482,
		0.947842,
		0.952816,
		0.957423,
		0.961681,
		0.965609,
		0.969224,
		0.972543,
		0.975582,
		0.978359,
		0.980889,
		0.983187,
		0.985268,
		0.987147,
		0.988837,
		0.990352,
		0.991705,
		0.992908,
		0.993974,
		0.994912,
		0.995735,
		0.996452,
		0.997073,
		0.997607,
		0.998064,
		0.998451,
		0.998776,
		0.999046,
		0.999268,
		0.999448,
		0.999593,
		0.999707,
		0.999794,
		0.999861,
		0.999909,
		0.999944,
		0.999967,
		0.999982,
		0.999992,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999986,
		0.999975,
		0.999959,
		0.999937,
		0.999907,
		0.999868,
		0.999819,
		0.999758,
		0.999684,
		0.999596,
		0.999493,
		0.999373,
		0.999236,
		0.999079,
		0.998903,
		0.998706,
		0.998486,
		0.998244,
		0.997978,
		0.997688,
		0.997372,
		0.99703,
		0.996662,
		0.996266,
		0.995843,
		0.995392,
		0.994912,
		0.994403,
		0.993865,
		0.993298,
		0.9927,
		0.992073,
		0.991416,
		0.990728,
		0.990011,
		0.989263,
		0.988485,
		0.987677,
		0.986838,
		0.98597,
		0.985072,
		0.984144,
		0.983187,
		0.9822,
		0.981184,
		0.98014,
		0.979066,
		0.977965,
		0.976835,
		0.975678,
		0.974494,
		0.973282,
		0.972044,
		0.97078,
		0.969489,
		0.968173,
		0.966832,
		0.965467,
		0.001561,
		0.006383,
		0.014591,
		0.026203,
		0.041135,
		0.05921,
		0.080172,
		0.103703,
		0.12944,
		0.156993,
		0.185967,
		0.215973,
		0.246645,
		0.277651,
		0.308696,
		0.339533,
		0.369959,
		0.399815,
		0.428982,
		0.457378,
		0.484947,
		0.51166,
		0.537502,
		0.56247,
		0.586569,
		0.609807,
		0.632191,
		0.653732,
		0.674436,
		0.694312,
		0.713365,
		0.731605,
		0.749038,
		0.765677,
		0.781532,
		0.796619,
		0.810953,
		0.824552,
		0.837433,
		0.849616,
		0.861121,
		0.87197,
		0.882183,
		0.891784,
		0.900792,
		0.909232,
		0.917126,
		0.924495,
		0.931363,
		0.937751,
		0.943681,
		0.949176,
		0.954256,
		0.958943,
		0.963257,
		0.967218,
		0.970847,
		0.974162,
		0.977183,
		0.979927,
		0.982413,
		0.984657,
		0.986675,
		0.988485,
		0.990101,
		0.991538,
		0.992809,
		0.99393,
		0.994912,
		0.995768,
		0.996509,
		0.997147,
		0.997692,
		0.998153,
		0.998541,
		0.998863,
		0.999128,
		0.999343,
		0.999515,
		0.999651,
		0.999755,
		0.999834,
		0.999892,
		0.999933,
		0.999961,
		0.999979,
		0.99999,
		0.999996,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999992,
		0.999984,
		0.999971,
		0.999952,
		0.999926,
		0.999891,
		0.999846,
		0.999788,
		0.999718,
		0.999632,
		0.999531,
		0.999412,
		0.999274,
		0.999115,
		0.998935,
		0.998733,
		0.998506,
		0.998255,
		0.997978,
		0.997675,
		0.997343,
		0.996984,
		0.996595,
		0.996177,
		0.995728,
		0.995249,
		0.994738,
		0.994196,
		0.993622,
		0.993016,
		0.992377,
		0.991705,
		0.991001,
		0.990264,
		0.989494,
		0.988691,
		0.987855,
		0.986986,
		0.986085,
		0.985151,
		0.984185,
		0.983187,
		0.982156,
		0.981094,
		0.980001,
		0.978877,
		0.977722,
		0.976536,
		0.975321,
		0.974075,
		0.972801,
		0.971498,
		0.970166,
		0.968806,
		0.967419,
		0.966004,
		0.964563,
		0.963095,
		0.961602,
		0.960083,
		0.95854,
		0.956972,
		0.955381,
		0.953766,
		0.952128,
		0.950468,
		0.948785,
		0.001204,
		0.005056,
		0.011854,
		0.021807,
		0.035026,
		0.051521,
		0.071204,
		0.093895,
		0.119331,
		0.147183,
		0.177071,
		0.208583,
		0.241294,
		0.274783,
		0.308649,
		0.342523,
		0.37608,
		0.409045,
		0.441194,
		0.472358,
		0.502414,
		0.531288,
		0.558942,
		0.585368,
		0.610585,
		0.634625,
		0.657532,
		0.679351,
		0.700132,
		0.719918,
		0.73875,
		0.756663,
		0.773688,
		0.789852,
		0.805178,
		0.819691,
		0.833411,
		0.846361,
		0.858565,
		0.870045,
		0.880827,
		0.890936,
		0.900395,
		0.909232,
		0.917472,
		0.92514,
		0.932261,
		0.938862,
		0.944967,
		0.950601,
		0.955789,
		0.960553,
		0.964919,
		0.968908,
		0.972543,
		0.975845,
		0.978837,
		0.981538,
		0.983968,
		0.986146,
		0.988091,
		0.989821,
		0.991352,
		0.9927,
		0.993882,
		0.994912,
		0.995804,
		0.996571,
		0.997227,
		0.997782,
		0.998248,
		0.998635,
		0.998953,
		0.999212,
		0.999419,
		0.999581,
		0.999707,
		0.999801,
		0.999871,
		0.99992,
		0.999953,
		0.999975,
		0.999988,
		0.999995,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.99999,
		0.999981,
		0.999966,
		0.999944,
		0.999913,
		0.999872,
		0.999819,
		0.999752,
		0.999669,
		0.99957,
		0.999452,
		0.999313,
		0.999153,
		0.99897,
		0.998762,
		0.998528,
		0.998267,
		0.997978,
		0.99766,
		0.997312,
		0.996932,
		0.996521,
		0.996078,
		0.995601,
		0.99509,
		0.994545,
		0.993965,
		0.99335,
		0.9927,
		0.992015,
		0.991293,
		0.990536,
		0.989742,
		0.988913,
		0.988048,
		0.987147,
		0.98621,
		0.985237,
		0.98423,
		0.983187,
		0.982109,
		0.980996,
		0.97985,
		0.978669,
		0.977455,
		0.976208,
		0.974928,
		0.973615,
		0.972271,
		0.970896,
		0.969489,
		0.968053,
		0.966586,
		0.96509,
		0.963565,
		0.962012,
		0.960431,
		0.958822,
		0.957188,
		0.955526,
		0.95384,
		0.952128,
		0.950392,
		0.948631,
		0.946848,
		0.945041,
		0.943213,
		0.941362,
		0.939491,
		0.937598,
		0.935686,
		0.933754,
		0.931803,
		0.929834,
		0.927846,
		0.000818,
		0.00355,
		0.008587,
		0.01627,
		0.026875,
		0.040599,
		0.057546,
		0.077731,
		0.101068,
		0.127383,
		0.156413,
		0.187825,
		0.221226,
		0.256183,
		0.29224,
		0.328938,
		0.365832,
		0.402506,
		0.438588,
		0.473757,
		0.507751,
		0.540369,
		0.571468,
		0.600964,
		0.628819,
		0.655041,
		0.679669,
		0.702768,
		0.724419,
		0.744711,
		0.763735,
		0.781582,
		0.798332,
		0.81406,
		0.82883,
		0.842698,
		0.85571,
		0.867908,
		0.879328,
		0.890001,
		0.899959,
		0.909232,
		0.91785,
		0.925841,
		0.933236,
		0.940063,
		0.946351,
		0.952128,
		0.957423,
		0.962262,
		0.966673,
		0.970681,
		0.974313,
		0.977591,
		0.980542,
		0.983187,
		0.985548,
		0.987648,
		0.989507,
		0.991144,
		0.992579,
		0.993829,
		0.994912,
		0.995843,
		0.996638,
		0.997312,
		0.997877,
		0.998347,
		0.998733,
		0.999046,
		0.999296,
		0.999493,
		0.999645,
		0.99976,
		0.999844,
		0.999903,
		0.999944,
		0.99997,
		0.999986,
		0.999994,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999995,
		0.999989,
		0.999977,
		0.999959,
		0.999933,
		0.999897,
		0.999848,
		0.999785,
		0.999707,
		0.99961,
		0.999493,
		0.999355,
		0.999193,
		0.999006,
		0.998793,
		0.998551,
		0.99828,
		0.997978,
		0.997644,
		0.997277,
		0.996876,
		0.996439,
		0.995967,
		0.995458,
		0.994912,
		0.994328,
		0.993706,
		0.993045,
		0.992346,
		0.991607,
		0.990828,
		0.990011,
		0.989154,
		0.988257,
		0.987321,
		0.986346,
		0.985332,
		0.984278,
		0.983187,
		0.982057,
		0.980889,
		0.979683,
		0.97844,
		0.977161,
		0.975845,
		0.974494,
		0.973107,
		0.971686,
		0.97023,
		0.968741,
		0.967218,
		0.965663,
		0.964076,
		0.962458,
		0.96081,
		0.959131,
		0.957423,
		0.955686,
		0.953921,
		0.952128,
		0.950308,
		0.948463,
		0.946591,
		0.944695,
		0.942774,
		0.94083,
		0.938862,
		0.936872,
		0.93486,
		0.932827,
		0.930774,
		0.9287,
		0.926607,
		0.924495,
		0.922365,
		0.920217,
		0.918052,
		0.91587,
		0.913673,
		0.91146,
		0.909232,
		0.90699,
		0.904734,
		0.902465,
		0.000471,
		0.002137,
		0.005387,
		0.010604,
		0.018148,
		0.02834,
		0.041445,
		0.057659,
		0.077097,
		0.099785,
		0.125657,
		0.154555,
		0.186232,
		0.220363,
		0.256554,
		0.294361,
		0.333305,
		0.372893,
		0.412633,
		0.45206,
		0.49074,
		0.528295,
		0.564404,
		0.598814,
		0.63134,
		0.661863,
		0.690328,
		0.716734,
		0.741127,
		0.76359,
		0.784235,
		0.803188,
		0.820585,
		0.836564,
		0.851254,
		0.864779,
		0.877246,
		0.888751,
		0.899376,
		0.909189,
		0.918247,
		0.926602,
		0.934294,
		0.941362,
		0.947842,
		0.953766,
		0.959167,
		0.964076,
		0.968525,
		0.972543,
		0.976158,
		0.9794,
		0.982294,
		0.984868,
		0.987147,
		0.989154,
		0.990912,
		0.992444,
		0.993771,
		0.994912,
		0.995886,
		0.996711,
		0.997403,
		0.997978,
		0.998451,
		0.998834,
		0.99914,
		0.999381,
		0.999567,
		0.999707,
		0.999809,
		0.999882,
		0.999931,
		0.999963,
		0.999982,
		0.999993,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999994,
		0.999986,
		0.999973,
		0.999951,
		0.99992,
		0.999876,
		0.999819,
		0.999744,
		0.999651,
		0.999536,
		0.999399,
		0.999236,
		0.999046,
		0.998827,
		0.998577,
		0.998294,
		0.997978,
		0.997627,
		0.997238,
		0.996812,
		0.996348,
		0.995843,
		0.995298,
		0.994712,
		0.994084,
		0.993413,
		0.9927,
		0.991944,
		0.991144,
		0.990301,
		0.989415,
		0.988485,
		0.987511,
		0.986495,
		0.985435,
		0.984332,
		0.983187,
		0.981999,
		0.98077,
		0.979499,
		0.978187,
		0.976835,
		0.975444,
		0.974012,
		0.972543,
		0.971035,
		0.969489,
		0.967907,
		0.966289,
		0.964635,
		0.962947,
		0.961225,
		0.959469,
		0.957681,
		0.955861,
		0.954009,
		0.952128,
		0.950217,
		0.948277,
		0.946308,
		0.944313,
		0.94229,
		0.940242,
		0.938168,
		0.93607,
		0.933948,
		0.931803,
		0.929636,
		0.927446,
		0.925236,
		0.923006,
		0.920755,
		0.918486,
		0.916199,
		0.913893,
		0.911571,
		0.909232,
		0.906878,
		0.904508,
		0.902123,
		0.899725,
		0.897313,
		0.894888,
		0.89245,
		0.890001,
		0.887541,
		0.88507,
		0.882588,
		0.880097,
		0.877597,
		0.875088,
		0.872571,
		0.000214,
		0.001038,
		0.00278,
		0.005777,
		0.010384,
		0.016957,
		0.025838,
		0.037339,
		0.051724,
		0.069201,
		0.089903,
		0.113883,
		0.141108,
		0.171451,
		0.204699,
		0.240551,
		0.278635,
		0.318512,
		0.359697,
		0.401677,
		0.443925,
		0.485925,
		0.527188,
		0.567267,
		0.605771,
		0.642378,
		0.676837,
		0.708972,
		0.73868,
		0.765928,
		0.790745,
		0.81321,
		0.833445,
		0.8516,
		0.867848,
		0.882368,
		0.895339,
		0.906936,
		0.917318,
		0.926633,
		0.935005,
		0.942546,
		0.949345,
		0.955481,
		0.961014,
		0.965998,
		0.970479,
		0.974494,
		0.978079,
		0.981267,
		0.984089,
		0.986574,
		0.988753,
		0.99065,
		0.992293,
		0.993706,
		0.994912,
		0.995933,
		0.99679,
		0.997501,
		0.998086,
		0.99856,
		0.998938,
		0.999236,
		0.999465,
		0.999638,
		0.999765,
		0.999854,
		0.999916,
		0.999955,
		0.999978,
		0.999991,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		0.999998,
		0.999993,
		0.999983,
		0.999967,
		0.999941,
		0.999903,
		0.999851,
		0.999782,
		0.999693,
		0.999581,
		0.999445,
		0.999281,
		0.999088,
		0.998863,
		0.998605,
		0.99831,
		0.997978,
		0.997607,
		0.997195,
		0.996742,
		0.996245,
		0.995704,
		0.995118,
		0.994485,
		0.993807,
		0.993081,
		0.992308,
		0.991486,
		0.990617,
		0.9897,
		0.988734,
		0.98772,
		0.986658,
		0.985548,
		0.984391,
		0.983187,
		0.981935,
		0.980638,
		0.979295,
		0.977906,
		0.976473,
		0.974996,
		0.973475,
		0.971912,
		0.970307,
		0.968661,
		0.966975,
		0.965249,
		0.963484,
		0.961681,
		0.959841,
		0.957965,
		0.956054,
		0.954108,
		0.952128,
		0.950115,
		0.948071,
		0.945995,
		0.943889,
		0.941754,
		0.93959,
		0.937398,
		0.935179,
		0.932935,
		0.930665,
		0.928371,
		0.926053,
		0.923712,
		0.92135,
		0.918966,
		0.916561,
		0.914137,
		0.911694,
		0.909232,
		0.906753,
		0.904257,
		0.901745,
		0.899218,
		0.896676,
		0.894119,
		0.891549,
		0.888967,
		0.886372,
		0.883765,
		0.881147,
		0.878519,
		0.875881,
		0.873234,
		0.870578,
		0.867913,
		0.865241,
		0.862562,
		0.859876,
		0.857183,
		0.854485,
		0.851782,
		0.849074,
		0.846361,
		0.843645,
		0.840925,
		0.838201,
		6.5e-005,
		0.000356,
		0.001059,
		0.002407,
		0.004677,
		0.008178,
		0.013239,
		0.020197,
		0.029382,
		0.041104,
		0.055636,
		0.073201,
		0.09396,
		0.118003,
		0.145338,
		0.175889,
		0.209488,
		0.245885,
		0.284745,
		0.325662,
		0.368169,
		0.411755,
		0.455881,
		0.5,
		0.543575,
		0.586099,
		0.627109,
		0.666205,
		0.703054,
		0.737402,
		0.769076,
		0.797981,
		0.824096,
		0.847468,
		0.868202,
		0.886448,
		0.902391,
		0.916238,
		0.928207,
		0.938519,
		0.947385,
		0.955004,
		0.961557,
		0.967203,
		0.972079,
		0.9763,
		0.979961,
		0.983139,
		0.985898,
		0.988288,
		0.990351,
		0.992122,
		0.993633,
		0.994912,
		0.995985,
		0.996876,
		0.997607,
		0.9982,
		0.998674,
		0.999046,
		0.999333,
		0.999548,
		0.999707,
		0.999819,
		0.999895,
		0.999944,
		0.999973,
		0.999989,
		0.999997,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999991,
		0.99998,
		0.999959,
		0.999928,
		0.999882,
		0.999819,
		0.999735,
		0.999627,
		0.999493,
		0.99933,
		0.999134,
		0.998903,
		0.998635,
		0.998327,
		0.997978,
		0.997585,
		0.997147,
		0.996662,
		0.996128,
		0.995546,
		0.994912,
		0.994227,
		0.99349,
		0.9927,
		0.991857,
		0.990961,
		0.990011,
		0.989007,
		0.987949,
		0.986838,
		0.985674,
		0.984457,
		0.983187,
		0.981865,
		0.980491,
		0.979066,
		0.977591,
		0.976067,
		0.974494,
		0.972873,
		0.971204,
		0.969489,
		0.967729,
		0.965925,
		0.964076,
		0.962186,
		0.960253,
		0.95828,
		0.956268,
		0.954217,
		0.952128,
		0.950003,
		0.947842,
		0.945646,
		0.943417,
		0.941155,
		0.938862,
		0.936538,
		0.934185,
		0.931803,
		0.929393,
		0.926957,
		0.924495,
		0.922008,
		0.919497,
		0.916963,
		0.914407,
		0.91183,
		0.909232,
		0.906615,
		0.903979,
		0.901325,
		0.898654,
		0.895967,
		0.893264,
		0.890546,
		0.887815,
		0.88507,
		0.882312,
		0.879542,
		0.876762,
		0.87397,
		0.871169,
		0.868358,
		0.865538,
		0.862711,
		0.859876,
		0.857034,
		0.854185,
		0.851331,
		0.848471,
		0.845607,
		0.842738,
		0.839866,
		0.83699,
		0.834112,
		0.83123,
		0.828347,
		0.825463,
		0.822577,
		0.819691,
		0.816804,
		0.813918,
		0.811031,
		0.808146,
		0.805262,
		0.802379,
		0.799498,
		6.1e-005,
		0.000228,
		0.000623,
		0.001406,
		0.002784,
		0.005004,
		0.008349,
		0.013129,
		0.019662,
		0.028272,
		0.039266,
		0.052927,
		0.0695,
		0.089175,
		0.112084,
		0.138285,
		0.167757,
		0.200397,
		0.236018,
		0.274347,
		0.315035,
		0.35766,
		0.401745,
		0.446768,
		0.492178,
		0.537419,
		0.581942,
		0.625228,
		0.666804,
		0.706256,
		0.743241,
		0.777497,
		0.808844,
		0.837182,
		0.862495,
		0.884835,
		0.904319,
		0.921114,
		0.935426,
		0.947489,
		0.95755,
		0.96586,
		0.972665,
		0.978195,
		0.982663,
		0.986257,
		0.989142,
		0.991455,
		0.99331,
		0.994799,
		0.995994,
		0.996952,
		0.997716,
		0.99832,
		0.998793,
		0.999156,
		0.999429,
		0.999629,
		0.999771,
		0.999867,
		0.999929,
		0.999966,
		0.999986,
		0.999996,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.999989,
		0.999975,
		0.99995,
		0.999911,
		0.999854,
		0.999777,
		0.999675,
		0.999544,
		0.999381,
		0.999183,
		0.998946,
		0.998669,
		0.998347,
		0.997978,
		0.997561,
		0.997092,
		0.996571,
		0.995996,
		0.995364,
		0.994676,
		0.99393,
		0.993125,
		0.992261,
		0.991337,
		0.990352,
		0.989308,
		0.988203,
		0.987038,
		0.985814,
		0.98453,
		0.983187,
		0.981785,
		0.980326,
		0.97881,
		0.977237,
		0.975609,
		0.973927,
		0.972191,
		0.970403,
		0.968563,
		0.966673,
		0.964734,
		0.962746,
		0.960712,
		0.958631,
		0.956507,
		0.954338,
		0.952128,
		0.949876,
		0.947585,
		0.945255,
		0.942888,
		0.940484,
		0.938045,
		0.935573,
		0.933068,
		0.930531,
		0.927963,
		0.925367,
		0.922742,
		0.92009,
		0.917412,
		0.914709,
		0.911982,
		0.909232,
		0.90646,
		0.903668,
		0.900855,
		0.898023,
		0.895174,
		0.892307,
		0.889423,
		0.886525,
		0.883611,
		0.880684,
		0.877744,
		0.874792,
		0.871829,
		0.868855,
		0.865871,
		0.862877,
		0.859876,
		0.856866,
		0.85385,
		0.850827,
		0.847798,
		0.844764,
		0.841725,
		0.838682,
		0.835636,
		0.832587,
		0.829535,
		0.826481,
		0.823426,
		0.82037,
		0.817314,
		0.814257,
		0.811201,
		0.808146,
		0.805092,
		0.80204,
		0.79899,
		0.795942,
		0.792898,
		0.789856,
		0.786818,
		0.783784,
		0.780754,
		0.777728,
		0.774707,
		0.771692,
		0.768681,
		0.765677,
		0.762678,
		0.759685,
		0.756699,
		1e-005,
		5.1e-005,
		0.000175,
		0.00047,
		0.001066,
		0.002141,
		0.003911,
		0.006631,
		0.010587,
		0.016081,
		0.023426,
		0.032928,
		0.044879,
		0.05954,
		0.077134,
		0.097831,
		0.121741,
		0.148905,
		0.17929,
		0.212781,
		0.249185,
		0.288229,
		0.329566,
		0.37278,
		0.417399,
		0.462908,
		0.508764,
		0.554413,
		0.599306,
		0.642924,
		0.684786,
		0.724468,
		0.761617,
		0.795954,
		0.827286,
		0.855502,
		0.88057,
		0.902538,
		0.921517,
		0.937675,
		0.951225,
		0.96241,
		0.971493,
		0.978745,
		0.984431,
		0.988808,
		0.992111,
		0.994554,
		0.996322,
		0.997575,
		0.998441,
		0.999026,
		0.999412,
		0.999659,
		0.999812,
		0.999904,
		0.999955,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999995,
		0.999986,
		0.999968,
		0.999937,
		0.999889,
		0.999819,
		0.999723,
		0.999596,
		0.999435,
		0.999236,
		0.998994,
		0.998706,
		0.998368,
		0.997978,
		0.997533,
		0.99703,
		0.996468,
		0.995843,
		0.995156,
		0.994403,
		0.993585,
		0.9927,
		0.991748,
		0.990728,
		0.989641,
		0.988485,
		0.987261,
		0.98597,
		0.984612,
		0.983187,
		0.981696,
		0.98014,
		0.978519,
		0.976835,
		0.975089,
		0.973282,
		0.971415,
		0.969489,
		0.967506,
		0.965467,
		0.963372,
		0.961225,
		0.959025,
		0.956775,
		0.954475,
		0.952128,
		0.949734,
		0.947296,
		0.944814,
		0.94229,
		0.939726,
		0.937122,
		0.934481,
		0.931803,
		0.92909,
		0.926344,
		0.923565,
		0.920755,
		0.917916,
		0.915048,
		0.912153,
		0.909232,
		0.906287,
		0.903317,
		0.900326,
		0.897313,
		0.894279,
		0.891227,
		0.888157,
		0.88507,
		0.881966,
		0.878848,
		0.875716,
		0.872571,
		0.869413,
		0.866244,
		0.863065,
		0.859876,
		0.856678,
		0.853472,
		0.850259,
		0.84704,
		0.843815,
		0.840584,
		0.83735,
		0.834112,
		0.83087,
		0.827626,
		0.824381,
		0.821134,
		0.817887,
		0.814639,
		0.811392,
		0.808146,
		0.804901,
		0.801659,
		0.798418,
		0.795181,
		0.791947,
		0.788716,
		0.78549,
		0.782268,
		0.779051,
		0.77584,
		0.772634,
		0.769433,
		0.76624,
		0.763052,
		0.759872,
		0.756699,
		0.753533,
		0.750376,
		0.747226,
		0.744084,
		0.740951,
		0.737827,
		0.734711,
		0.731605,
		0.728508,
		0.72542,
		0.722343,
		0.719275,
		0.716218,
		0.71317,
		0.710133,
		4.7e-005,
		0.000173,
		0.000481,
		0.001107,
		0.002233,
		0.00408,
		0.006903,
		0.010985,
		0.016622,
		0.024116,
		0.033761,
		0.045834,
		0.06058,
		0.078208,
		0.098875,
		0.122681,
		0.149661,
		0.179781,
		0.212934,
		0.248937,
		0.287533,
		0.328396,
		0.371134,
		0.415302,
		0.460409,
		0.505937,
		0.551352,
		0.596121,
		0.639732,
		0.681705,
		0.721612,
		0.759087,
		0.793834,
		0.825636,
		0.854357,
		0.879941,
		0.902408,
		0.921848,
		0.938409,
		0.95229,
		0.963724,
		0.97297,
		0.980301,
		0.985989,
		0.990301,
		0.993485,
		0.99577,
		0.997356,
		0.998418,
		0.9991,
		0.999516,
		0.999756,
		0.999887,
		0.999952,
		0.999982,
		0.999993,
		0.999993,
		0.999982,
		0.999959,
		0.99992,
		0.999859,
		0.999771,
		0.999651,
		0.999493,
		0.999293,
		0.999046,
		0.998747,
		0.998392,
		0.997978,
		0.997501,
		0.996959,
		0.996348,
		0.995666,
		0.994912,
		0.994084,
		0.993181,
		0.992201,
		0.991144,
		0.990011,
		0.9888,
		0.987511,
		0.986146,
		0.984704,
		0.983187,
		0.981594,
		0.979927,
		0.978187,
		0.976376,
		0.974494,
		0.972543,
		0.970524,
		0.968439,
		0.966289,
		0.964076,
		0.961803,
		0.959469,
		0.957078,
		0.95463,
		0.952128,
		0.949573,
		0.946968,
		0.944313,
		0.94161,
		0.938862,
		0.93607,
		0.933236,
		0.930361,
		0.927446,
		0.924495,
		0.921508,
		0.918486,
		0.915432,
		0.912347,
		0.909232,
		0.906089,
		0.90292,
		0.899725,
		0.896506,
		0.893264,
		0.890001,
		0.886718,
		0.883417,
		0.880097,
		0.876762,
		0.873411,
		0.870045,
		0.866667,
		0.863277,
		0.859876,
		0.856465,
		0.853044,
		0.849616,
		0.84618,
		0.842738,
		0.839291,
		0.835839,
		0.832383,
		0.828924,
		0.825463,
		0.822,
		0.818536,
		0.815072,
		0.811609,
		0.808146,
		0.804685,
		0.801227,
		0.797771,
		0.794318,
		0.790869,
		0.787425,
		0.783986,
		0.780552,
		0.777124,
		0.773702,
		0.770286,
		0.766878,
		0.763477,
		0.760084,
		0.756699,
		0.753323,
		0.749955,
		0.746597,
		0.743248,
		0.739908,
		0.736579,
		0.73326,
		0.729952,
		0.726654,
		0.723368,
		0.720092,
		0.716828,
		0.713576,
		0.710335,
		0.707107,
		0.70389,
		0.700686,
		0.697495,
		0.694316,
		0.69115,
		0.687996,
		0.684856,
		0.681729,
		0.678615,
		0.675515,
		0.672427,
		0.669354,
		0.666294,
		0.663248,
		0.660215,
		1.8e-005,
		9.6e-005,
		0.000326,
		0.000838,
		0.001811,
		0.003462,
		0.006041,
		0.009823,
		0.015095,
		0.02215,
		0.031269,
		0.042718,
		0.056733,
		0.073513,
		0.09321,
		0.115924,
		0.141698,
		0.170509,
		0.202269,
		0.236823,
		0.273949,
		0.313356,
		0.354699,
		0.397574,
		0.441538,
		0.486112,
		0.530799,
		0.575095,
		0.618508,
		0.660568,
		0.700845,
		0.738959,
		0.774593,
		0.807497,
		0.837497,
		0.864494,
		0.888462,
		0.909445,
		0.927549,
		0.942932,
		0.955795,
		0.966371,
		0.974912,
		0.981678,
		0.986929,
		0.990914,
		0.993865,
		0.99599,
		0.997473,
		0.998467,
		0.9991,
		0.999471,
		0.999655,
		0.999705,
		0.999659,
		0.999536,
		0.999349,
		0.999102,
		0.998793,
		0.998419,
		0.997978,
		0.997465,
		0.996876,
		0.996208,
		0.995458,
		0.994625,
		0.993706,
		0.9927,
		0.991607,
		0.990425,
		0.989154,
		0.987794,
		0.986346,
		0.98481,
		0.983187,
		0.981477,
		0.979683,
		0.977805,
		0.975845,
		0.973805,
		0.971686,
		0.969489,
		0.967218,
		0.964874,
		0.962458,
		0.959974,
		0.957423,
		0.954807,
		0.952128,
		0.949389,
		0.946591,
		0.943738,
		0.94083,
		0.93787,
		0.93486,
		0.931803,
		0.9287,
		0.925553,
		0.922365,
		0.919137,
		0.91587,
		0.912568,
		0.909232,
		0.905864,
		0.902465,
		0.899037,
		0.895582,
		0.892101,
		0.888597,
		0.88507,
		0.881522,
		0.877955,
		0.874369,
		0.870768,
		0.867151,
		0.863519,
		0.859876,
		0.85622,
		0.852555,
		0.84888,
		0.845197,
		0.841508,
		0.837812,
		0.834112,
		0.830407,
		0.826699,
		0.82299,
		0.819279,
		0.815567,
		0.811856,
		0.808146,
		0.804438,
		0.800733,
		0.797031,
		0.793332,
		0.789639,
		0.785951,
		0.782268,
		0.778592,
		0.774923,
		0.771261,
		0.767608,
		0.763962,
		0.760326,
		0.756699,
		0.753082,
		0.749475,
		0.745878,
		0.742293,
		0.738718,
		0.735156,
		0.731605,
		0.728066,
		0.72454,
		0.721027,
		0.717527,
		0.71404,
		0.710566,
		0.707107,
		0.703661,
		0.70023,
		0.696813,
		0.69341,
		0.690022,
		0.686649,
		0.683291,
		0.679948,
		0.67662,
		0.673308,
		0.670011,
		0.66673,
		0.663465,
		0.660215,
		0.656981,
		0.653763,
		0.650561,
		0.647376,
		0.644206,
		0.641052,
		0.637915,
		0.634794,
		0.631689,
		0.628601,
		0.625529,
		0.622473,
		0.619433,
		0.61641,
		0.613403,
		0.610413,
		0.607439,
		2.8e-005,
		0.000147,
		0.000475,
		0.001175,
		0.002452,
		0.004545,
		0.007717,
		0.012245,
		0.018409,
		0.026482,
		0.036721,
		0.049356,
		0.064584,
		0.082561,
		0.103399,
		0.127154,
		0.153832,
		0.183376,
		0.215674,
		0.25055,
		0.287772,
		0.32705,
		0.368043,
		0.410368,
		0.4536,
		0.497293,
		0.540983,
		0.584203,
		0.626498,
		0.667436,
		0.706621,
		0.743703,
		0.778389,
		0.810449,
		0.839721,
		0.866109,
		0.889587,
		0.910193,
		0.928019,
		0.943208,
		0.955943,
		0.966433,
		0.974912,
		0.98162,
		0.986797,
		0.990679,
		0.993485,
		0.995414,
		0.996642,
		0.997321,
		0.997575,
		0.997502,
		0.997181,
		0.996665,
		0.995994,
		0.995193,
		0.994276,
		0.993251,
		0.992122,
		0.99089,
		0.989554,
		0.988116,
		0.986574,
		0.984931,
		0.983187,
		0.981342,
		0.9794,
		0.97736,
		0.975226,
		0.972999,
		0.970681,
		0.968276,
		0.965784,
		0.963209,
		0.960553,
		0.957819,
		0.95501,
		0.952128,
		0.949176,
		0.946156,
		0.943071,
		0.939924,
		0.936718,
		0.933455,
		0.930138,
		0.926769,
		0.92335,
		0.919885,
		0.916375,
		0.912824,
		0.909232,
		0.905603,
		0.901939,
		0.898242,
		0.894513,
		0.890756,
		0.886971,
		0.883162,
		0.879329,
		0.875474,
		0.8716,
		0.867708,
		0.863799,
		0.859876,
		0.855939,
		0.85199,
		0.848031,
		0.844063,
		0.840087,
		0.836105,
		0.832117,
		0.828126,
		0.824131,
		0.820135,
		0.816138,
		0.812141,
		0.808146,
		0.804153,
		0.800163,
		0.796177,
		0.792195,
		0.78822,
		0.78425,
		0.780288,
		0.776333,
		0.772387,
		0.76845,
		0.764523,
		0.760605,
		0.756699,
		0.752804,
		0.748921,
		0.74505,
		0.741192,
		0.737347,
		0.733515,
		0.729698,
		0.725895,
		0.722107,
		0.718333,
		0.714575,
		0.710833,
		0.707107,
		0.703397,
		0.699703,
		0.696026,
		0.692366,
		0.688723,
		0.685097,
		0.681489,
		0.677898,
		0.674326,
		0.670771,
		0.667234,
		0.663715,
		0.660215,
		0.656733,
		0.65327,
		0.649825,
		0.646399,
		0.642991,
		0.639602,
		0.636233,
		0.632882,
		0.629549,
		0.626236,
		0.622942,
		0.619667,
		0.61641,
		0.613173,
		0.609954,
		0.606755,
		0.603574,
		0.600413,
		0.59727,
		0.594146,
		0.591042,
		0.587956,
		0.584888,
		0.58184,
		0.57881,
		0.575799,
		0.572806,
		0.569832,
		0.566876,
		0.563939,
		0.561021,
		0.55812,
		0.555238,
		0.552374,
		3.8e-005,
		0.00019,
		0.000596,
		0.00144,
		0.002939,
		0.005342,
		0.00891,
		0.013913,
		0.020615,
		0.029267,
		0.040099,
		0.053314,
		0.069081,
		0.087528,
		0.108744,
		0.132767,
		0.159588,
		0.189147,
		0.22133,
		0.255969,
		0.292845,
		0.331687,
		0.372177,
		0.413956,
		0.456627,
		0.499767,
		0.542935,
		0.585686,
		0.627576,
		0.668181,
		0.707107,
		0.743999,
		0.778552,
		0.81052,
		0.839721,
		0.866039,
		0.889428,
		0.909906,
		0.927549,
		0.942488,
		0.954896,
		0.96498,
		0.97297,
		0.979108,
		0.983637,
		0.986795,
		0.988808,
		0.98988,
		0.990193,
		0.989904,
		0.989142,
		0.988014,
		0.986599,
		0.98496,
		0.983139,
		0.981167,
		0.979061,
		0.976834,
		0.974494,
		0.972044,
		0.969489,
		0.966832,
		0.964076,
		0.961225,
		0.95828,
		0.955247,
		0.952128,
		0.948926,
		0.945646,
		0.94229,
		0.938862,
		0.935365,
		0.931803,
		0.928179,
		0.924495,
		0.920755,
		0.916963,
		0.913121,
		0.909232,
		0.905299,
		0.901325,
		0.897313,
		0.893264,
		0.889182,
		0.88507,
		0.880929,
		0.876762,
		0.872571,
		0.868358,
		0.864126,
		0.859876,
		0.85561,
		0.851331,
		0.84704,
		0.842738,
		0.838428,
		0.834112,
		0.829789,
		0.825463,
		0.821134,
		0.816804,
		0.812474,
		0.808146,
		0.80382,
		0.799498,
		0.795181,
		0.790869,
		0.786565,
		0.782268,
		0.77798,
		0.773702,
		0.769433,
		0.765177,
		0.760931,
		0.756699,
		0.75248,
		0.748275,
		0.744084,
		0.739908,
		0.735749,
		0.731605,
		0.727478,
		0.723368,
		0.719275,
		0.715201,
		0.711144,
		0.707107,
		0.703088,
		0.699089,
		0.695109,
		0.69115,
		0.68721,
		0.683291,
		0.679392,
		0.675515,
		0.671658,
		0.667822,
		0.664008,
		0.660215,
		0.656444,
		0.652694,
		0.648967,
		0.645261,
		0.641577,
		0.637915,
		0.634275,
		0.630658,
		0.627063,
		0.62349,
		0.619939,
		0.61641,
		0.612904,
		0.60942,
		0.605958,
		0.602518,
		0.599101,
		0.595706,
		0.592333,
		0.588982,
		0.585653,
		0.582346,
		0.579062,
		0.575799,
		0.572558,
		0.569338,
		0.56614,
		0.562964,
		0.55981,
		0.556677,
		0.553565,
		0.550474,
		0.547405,
		0.544357,
		0.54133,
		0.538323,
		0.535338,
		0.532373,
		0.529428,
		0.526505,
		0.523601,
		0.520718,
		0.517854,
		0.515011,
		0.512188,
		0.509384,
		0.5066,
		0.503836,
		0.501091,
		0.498366,
		0.495659,
		4.4e-005,
		0.000219,
		0.000676,
		0.001605,
		0.003231,
		0.005799,
		0.009563,
		0.014781,
		0.021703,
		0.030563,
		0.041577,
		0.054932,
		0.070788,
		0.089266,
		0.110452,
		0.134389,
		0.161075,
		0.19046,
		0.222444,
		0.256874,
		0.293546,
		0.332201,
		0.372533,
		0.414184,
		0.456759,
		0.499827,
		0.542935,
		0.585621,
		0.627422,
		0.667894,
		0.706621,
		0.743232,
		0.777408,
		0.808892,
		0.837497,
		0.863107,
		0.885676,
		0.905228,
		0.921848,
		0.935677,
		0.946897,
		0.955728,
		0.96241,
		0.967195,
		0.970337,
		0.972083,
		0.972665,
		0.972293,
		0.971156,
		0.969415,
		0.967203,
		0.96463,
		0.961779,
		0.958714,
		0.955481,
		0.952111,
		0.948626,
		0.94504,
		0.941362,
		0.937598,
		0.933754,
		0.929834,
		0.925841,
		0.921781,
		0.917657,
		0.913472,
		0.909232,
		0.90494,
		0.900598,
		0.896212,
		0.891784,
		0.887317,
		0.882814,
		0.87828,
		0.873716,
		0.869125,
		0.864511,
		0.859876,
		0.855222,
		0.850552,
		0.845868,
		0.841172,
		0.836467,
		0.831754,
		0.827036,
		0.822315,
		0.817592,
		0.812868,
		0.808146,
		0.803427,
		0.798713,
		0.794004,
		0.789303,
		0.784611,
		0.779928,
		0.775256,
		0.770596,
		0.76595,
		0.761317,
		0.756699,
		0.752097,
		0.747512,
		0.742944,
		0.738394,
		0.733863,
		0.729352,
		0.72486,
		0.72039,
		0.71594,
		0.711512,
		0.707107,
		0.702724,
		0.698364,
		0.694027,
		0.689715,
		0.685426,
		0.681162,
		0.676922,
		0.672708,
		0.668518,
		0.664354,
		0.660215,
		0.656102,
		0.652015,
		0.647954,
		0.643919,
		0.63991,
		0.635927,
		0.631971,
		0.628041,
		0.624138,
		0.620261,
		0.61641,
		0.612586,
		0.608789,
		0.605018,
		0.601273,
		0.597555,
		0.593863,
		0.590198,
		0.586559,
		0.582946,
		0.579359,
		0.575799,
		0.572264,
		0.568755,
		0.565272,
		0.561815,
		0.558383,
		0.554977,
		0.551596,
		0.54824,
		0.54491,
		0.541604,
		0.538323,
		0.535067,
		0.531836,
		0.528629,
		0.525446,
		0.522288,
		0.519153,
		0.516043,
		0.512956,
		0.509893,
		0.506853,
		0.503836,
		0.500843,
		0.497872,
		0.494924,
		0.491999,
		0.489096,
		0.486216,
		0.483358,
		0.480522,
		0.477707,
		0.474914,
		0.472143,
		0.469393,
		0.466665,
		0.463957,
		0.46127,
		0.458604,
		0.455958,
		0.453333,
		0.450728,
		0.448143,
		0.445578,
		0.443032,
		0.440506,
		0.438,
		4.8e-005,
		0.000233,
		0.000712,
		0.001676,
		0.003347,
		0.005965,
		0.00978,
		0.015042,
		0.021994,
		0.030868,
		0.041877,
		0.055211,
		0.071031,
		0.089464,
		0.110603,
		0.134495,
		0.161142,
		0.190493,
		0.222444,
		0.256834,
		0.293444,
		0.332002,
		0.372177,
		0.413595,
		0.455837,
		0.498456,
		0.540983,
		0.582942,
		0.623865,
		0.663304,
		0.700845,
		0.736123,
		0.768831,
		0.798725,
		0.825636,
		0.849464,
		0.870185,
		0.887842,
		0.902538,
		0.914431,
		0.923719,
		0.930634,
		0.935426,
		0.938357,
		0.939685,
		0.939661,
		0.938519,
		0.936471,
		0.933706,
		0.930382,
		0.926633,
		0.922566,
		0.918264,
		0.91379,
		0.909189,
		0.904491,
		0.89972,
		0.894886,
		0.890001,
		0.88507,
		0.880097,
		0.875088,
		0.870045,
		0.864974,
		0.859876,
		0.854755,
		0.849616,
		0.84446,
		0.839291,
		0.834112,
		0.828924,
		0.823732,
		0.818536,
		0.81334,
		0.808146,
		0.802956,
		0.797771,
		0.792593,
		0.787425,
		0.782268,
		0.777124,
		0.771993,
		0.766878,
		0.761779,
		0.756699,
		0.751638,
		0.746597,
		0.741577,
		0.736579,
		0.731605,
		0.726654,
		0.721729,
		0.716828,
		0.711954,
		0.707107,
		0.702287,
		0.697495,
		0.692731,
		0.687996,
		0.683291,
		0.678615,
		0.673969,
		0.669354,
		0.664769,
		0.660215,
		0.655692,
		0.651201,
		0.64674,
		0.642312,
		0.637915,
		0.63355,
		0.629217,
		0.624916,
		0.620647,
		0.61641,
		0.612205,
		0.608032,
		0.603892,
		0.599783,
		0.595706,
		0.591661,
		0.587648,
		0.583667,
		0.579717,
		0.575799,
		0.571912,
		0.568056,
		0.564232,
		0.560439,
		0.556677,
		0.552945,
		0.549244,
		0.545574,
		0.541933,
		0.538323,
		0.534743,
		0.531193,
		0.527672,
		0.52418,
		0.520718,
		0.517284,
		0.51388,
		0.510503,
		0.507156,
		0.503836,
		0.500544,
		0.497281,
		0.494044,
		0.490835,
		0.487653,
		0.484498,
		0.48137,
		0.478268,
		0.475193,
		0.472143,
		0.469119,
		0.466121,
		0.463149,
		0.460201,
		0.457278,
		0.454381,
		0.451507,
		0.448658,
		0.445833,
		0.443032,
		0.440255,
		0.437501,
		0.43477,
		0.432063,
		0.429378,
		0.426716,
		0.424076,
		0.421459,
		0.418863,
		0.416289,
		0.413737,
		0.411206,
		0.408697,
		0.406208,
		0.403741,
		0.401293,
		0.398867,
		0.39646,
		0.394074,
		0.391707,
		0.38936,
		0.387032,
		0.384724,
		0.382435,
		0.380165,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030876,
		0.041877,
		0.055198,
		0.070995,
		0.089387,
		0.110452,
		0.134224,
		0.160683,
		0.18976,
		0.22133,
		0.255208,
		0.291157,
		0.328883,
		0.368043,
		0.408251,
		0.449081,
		0.490085,
		0.530799,
		0.570757,
		0.609506,
		0.646618,
		0.681705,
		0.714427,
		0.744506,
		0.771729,
		0.795954,
		0.817112,
		0.835202,
		0.850289,
		0.862495,
		0.871991,
		0.878988,
		0.883723,
		0.886448,
		0.887425,
		0.88691,
		0.885148,
		0.882368,
		0.878774,
		0.874547,
		0.86984,
		0.864779,
		0.859466,
		0.853978,
		0.848375,
		0.842698,
		0.836975,
		0.831226,
		0.825462,
		0.819691,
		0.813918,
		0.808146,
		0.802379,
		0.796619,
		0.790869,
		0.785132,
		0.779408,
		0.773702,
		0.768013,
		0.762345,
		0.756699,
		0.751077,
		0.745479,
		0.739908,
		0.734365,
		0.728851,
		0.723368,
		0.717915,
		0.712494,
		0.707107,
		0.701753,
		0.696434,
		0.69115,
		0.685901,
		0.68069,
		0.675515,
		0.670377,
		0.665277,
		0.660215,
		0.655192,
		0.650207,
		0.645261,
		0.640354,
		0.635486,
		0.630658,
		0.625869,
		0.62112,
		0.61641,
		0.61174,
		0.607109,
		0.602518,
		0.597967,
		0.593455,
		0.588982,
		0.584549,
		0.580154,
		0.575799,
		0.571482,
		0.567204,
		0.562964,
		0.558763,
		0.5546,
		0.550474,
		0.546387,
		0.542336,
		0.538323,
		0.534347,
		0.530408,
		0.526505,
		0.522638,
		0.518807,
		0.515011,
		0.511251,
		0.507526,
		0.503836,
		0.50018,
		0.496559,
		0.492972,
		0.489418,
		0.485897,
		0.48241,
		0.478955,
		0.475533,
		0.472143,
		0.468785,
		0.465459,
		0.462163,
		0.458899,
		0.455665,
		0.452462,
		0.449289,
		0.446146,
		0.443032,
		0.439948,
		0.436892,
		0.433865,
		0.430867,
		0.427896,
		0.424954,
		0.422038,
		0.41915,
		0.416289,
		0.413455,
		0.410647,
		0.407865,
		0.405109,
		0.402378,
		0.399673,
		0.396993,
		0.394338,
		0.391707,
		0.3891,
		0.386518,
		0.383959,
		0.381424,
		0.378911,
		0.376422,
		0.373956,
		0.371512,
		0.369091,
		0.366691,
		0.364313,
		0.361957,
		0.359622,
		0.357308,
		0.355015,
		0.352743,
		0.350491,
		0.34826,
		0.346048,
		0.343856,
		0.341684,
		0.339531,
		0.337397,
		0.335282,
		0.333186,
		0.331109,
		0.32905,
		0.327009,
		0.324985,
		0.32298,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003361,
		0.005977,
		0.00978,
		0.015015,
		0.021916,
		0.030705,
		0.041577,
		0.054701,
		0.070213,
		0.088209,
		0.108744,
		0.131824,
		0.15741,
		0.185408,
		0.215674,
		0.248009,
		0.282165,
		0.317842,
		0.354699,
		0.392355,
		0.430401,
		0.468408,
		0.505937,
		0.542556,
		0.577846,
		0.611418,
		0.642924,
		0.672067,
		0.698606,
		0.722367,
		0.743241,
		0.761188,
		0.77623,
		0.788449,
		0.797981,
		0.805002,
		0.809723,
		0.812377,
		0.81321,
		0.812471,
		0.810403,
		0.807238,
		0.803188,
		0.798444,
		0.793172,
		0.787513,
		0.781582,
		0.77547,
		0.769247,
		0.762966,
		0.756663,
		0.750362,
		0.74408,
		0.737825,
		0.731605,
		0.72542,
		0.719275,
		0.71317,
		0.707107,
		0.701086,
		0.695109,
		0.689177,
		0.683291,
		0.677451,
		0.671658,
		0.665912,
		0.660215,
		0.654566,
		0.648967,
		0.643416,
		0.637915,
		0.632464,
		0.627063,
		0.621711,
		0.61641,
		0.611159,
		0.605958,
		0.600807,
		0.595706,
		0.590655,
		0.585653,
		0.580701,
		0.575799,
		0.570945,
		0.56614,
		0.561384,
		0.556677,
		0.552017,
		0.547405,
		0.542841,
		0.538323,
		0.533853,
		0.529428,
		0.52505,
		0.520718,
		0.51643,
		0.512188,
		0.50799,
		0.503836,
		0.499726,
		0.495659,
		0.491635,
		0.487653,
		0.483714,
		0.479816,
		0.475959,
		0.472143,
		0.468368,
		0.464632,
		0.460936,
		0.457278,
		0.45366,
		0.45008,
		0.446537,
		0.443032,
		0.439564,
		0.436133,
		0.432738,
		0.429378,
		0.426054,
		0.422765,
		0.41951,
		0.416289,
		0.413102,
		0.409949,
		0.406828,
		0.403741,
		0.400685,
		0.397661,
		0.394668,
		0.391707,
		0.388776,
		0.385876,
		0.383005,
		0.380165,
		0.377353,
		0.37457,
		0.371816,
		0.369091,
		0.366393,
		0.363722,
		0.361079,
		0.358463,
		0.355873,
		0.353309,
		0.350772,
		0.34826,
		0.345773,
		0.343311,
		0.340874,
		0.338462,
		0.336073,
		0.333709,
		0.331368,
		0.32905,
		0.326755,
		0.324482,
		0.322233,
		0.320005,
		0.317799,
		0.315615,
		0.313452,
		0.31131,
		0.309189,
		0.307089,
		0.305009,
		0.302949,
		0.300909,
		0.298888,
		0.296887,
		0.294906,
		0.292943,
		0.290998,
		0.289073,
		0.287165,
		0.285276,
		0.283404,
		0.28155,
		0.279714,
		0.277895,
		0.276092,
		0.274307,
		0.272538,
		0.270786,
		0.269049,
		0.267329,
		4.8e-005,
		0.000235,
		0.000712,
		0.001665,
		0.003307,
		0.005863,
		0.009563,
		0.014633,
		0.021286,
		0.029718,
		0.040099,
		0.05257,
		0.067237,
		0.084171,
		0.103399,
		0.124906,
		0.148632,
		0.174471,
		0.202269,
		0.231825,
		0.262895,
		0.295193,
		0.328396,
		0.36215,
		0.39608,
		0.429797,
		0.462908,
		0.495029,
		0.525792,
		0.554862,
		0.581942,
		0.606781,
		0.629185,
		0.649019,
		0.666205,
		0.680727,
		0.692627,
		0.701997,
		0.708972,
		0.713725,
		0.716456,
		0.717384,
		0.716734,
		0.714734,
		0.711604,
		0.707552,
		0.702768,
		0.697422,
		0.691658,
		0.685602,
		0.679351,
		0.672987,
		0.666569,
		0.66014,
		0.653732,
		0.647364,
		0.641049,
		0.634793,
		0.628601,
		0.622473,
		0.61641,
		0.610413,
		0.604481,
		0.598615,
		0.592813,
		0.587077,
		0.581406,
		0.575799,
		0.570256,
		0.564777,
		0.559361,
		0.554008,
		0.548718,
		0.54349,
		0.538323,
		0.533218,
		0.528173,
		0.523188,
		0.518262,
		0.513395,
		0.508587,
		0.503836,
		0.499142,
		0.494505,
		0.489923,
		0.485397,
		0.480925,
		0.476508,
		0.472143,
		0.467831,
		0.463572,
		0.459363,
		0.455206,
		0.451099,
		0.447041,
		0.443032,
		0.439072,
		0.435159,
		0.431293,
		0.427474,
		0.423701,
		0.419973,
		0.416289,
		0.41265,
		0.409054,
		0.405501,
		0.40199,
		0.398522,
		0.395094,
		0.391707,
		0.38836,
		0.385053,
		0.381784,
		0.378555,
		0.375363,
		0.372208,
		0.369091,
		0.366009,
		0.362964,
		0.359954,
		0.35698,
		0.354039,
		0.351133,
		0.34826,
		0.34542,
		0.342613,
		0.339837,
		0.337094,
		0.334382,
		0.331701,
		0.32905,
		0.326429,
		0.323837,
		0.321275,
		0.318742,
		0.316237,
		0.31376,
		0.31131,
		0.308888,
		0.306492,
		0.304124,
		0.301781,
		0.299464,
		0.297172,
		0.294906,
		0.292664,
		0.290446,
		0.288253,
		0.286083,
		0.283937,
		0.281814,
		0.279714,
		0.277636,
		0.27558,
		0.273547,
		0.271535,
		0.269544,
		0.267574,
		0.265625,
		0.263696,
		0.261788,
		0.2599,
		0.258031,
		0.256181,
		0.254351,
		0.25254,
		0.250747,
		0.248973,
		0.247217,
		0.245479,
		0.243759,
		0.242056,
		0.24037,
		0.238701,
		0.237049,
		0.235414,
		0.233795,
		0.232193,
		0.230606,
		0.229035,
		0.227479,
		0.225939,
		0.224414,
		0.222905,
		0.221409,
		0.219929,
		0.218463,
		0.217011,
		0.215573,
		0.214149,
		4.6e-005,
		0.000224,
		0.000676,
		0.001573,
		0.00311,
		0.005488,
		0.00891,
		0.013573,
		0.019659,
		0.027329,
		0.036721,
		0.047945,
		0.061078,
		0.076164,
		0.09321,
		0.112184,
		0.133017,
		0.1556,
		0.179781,
		0.205375,
		0.232157,
		0.259871,
		0.288229,
		0.316924,
		0.34563,
		0.374015,
		0.401745,
		0.428498,
		0.453969,
		0.477883,
		0.5,
		0.520121,
		0.538097,
		0.553828,
		0.567267,
		0.578414,
		0.58732,
		0.594078,
		0.598814,
		0.601687,
		0.602874,
		0.602567,
		0.600964,
		0.59826,
		0.594644,
		0.590294,
		0.585368,
		0.580011,
		0.574344,
		0.56847,
		0.56247,
		0.556411,
		0.550341,
		0.544295,
		0.538297,
		0.532363,
		0.526502,
		0.520717,
		0.515011,
		0.509384,
		0.503836,
		0.498366,
		0.492972,
		0.487653,
		0.48241,
		0.47724,
		0.472143,
		0.467118,
		0.462163,
		0.457278,
		0.452462,
		0.447714,
		0.443032,
		0.438416,
		0.433865,
		0.429378,
		0.424954,
		0.420591,
		0.416289,
		0.412048,
		0.407865,
		0.403741,
		0.399673,
		0.395662,
		0.391707,
		0.387806,
		0.383959,
		0.380165,
		0.376422,
		0.372731,
		0.369091,
		0.365499,
		0.361957,
		0.358463,
		0.355015,
		0.351615,
		0.34826,
		0.34495,
		0.341684,
		0.338462,
		0.335282,
		0.332145,
		0.32905,
		0.325995,
		0.32298,
		0.320005,
		0.317068,
		0.31417,
		0.31131,
		0.308487,
		0.3057,
		0.302949,
		0.300233,
		0.297552,
		0.294906,
		0.292293,
		0.289713,
		0.287165,
		0.28465,
		0.282166,
		0.279714,
		0.277292,
		0.2749,
		0.272538,
		0.270205,
		0.267901,
		0.265625,
		0.263377,
		0.261156,
		0.258963,
		0.256796,
		0.254655,
		0.25254,
		0.25045,
		0.248386,
		0.246346,
		0.24433,
		0.242338,
		0.24037,
		0.238425,
		0.236502,
		0.234603,
		0.232725,
		0.230869,
		0.229035,
		0.227222,
		0.225429,
		0.223658,
		0.221906,
		0.220175,
		0.218463,
		0.21677,
		0.215097,
		0.213443,
		0.211807,
		0.210189,
		0.208589,
		0.207007,
		0.205443,
		0.203896,
		0.202366,
		0.200852,
		0.199356,
		0.197875,
		0.196411,
		0.194962,
		0.193529,
		0.192112,
		0.19071,
		0.189322,
		0.18795,
		0.186592,
		0.185248,
		0.183919,
		0.182603,
		0.181302,
		0.180014,
		0.178739,
		0.177478,
		0.17623,
		0.174994,
		0.173771,
		0.172561,
		0.171364,
		0.170178,
		0.169005,
		0.167843,
		0.166693,
		0.165555,
		0.164428,
		4.1e-005,
		0.000199,
		0.000596,
		0.001382,
		0.002719,
		0.004775,
		0.007717,
		0.011703,
		0.016877,
		0.023365,
		0.031269,
		0.040669,
		0.051615,
		0.064131,
		0.078208,
		0.09381,
		0.110866,
		0.129275,
		0.148905,
		0.169596,
		0.191158,
		0.213377,
		0.236018,
		0.25883,
		0.281552,
		0.303917,
		0.325662,
		0.346533,
		0.366296,
		0.384738,
		0.401677,
		0.416966,
		0.430498,
		0.442204,
		0.45206,
		0.460076,
		0.466305,
		0.470829,
		0.473757,
		0.475222,
		0.475373,
		0.474364,
		0.472358,
		0.46951,
		0.465973,
		0.461886,
		0.457378,
		0.452559,
		0.447525,
		0.442357,
		0.437117,
		0.431857,
		0.426612,
		0.421411,
		0.416269,
		0.411199,
		0.406206,
		0.401293,
		0.39646,
		0.391707,
		0.387032,
		0.382435,
		0.377913,
		0.373465,
		0.369091,
		0.364787,
		0.360554,
		0.356389,
		0.352291,
		0.34826,
		0.344293,
		0.34039,
		0.336549,
		0.332769,
		0.32905,
		0.325389,
		0.321785,
		0.318238,
		0.314747,
		0.31131,
		0.307926,
		0.304595,
		0.301315,
		0.298086,
		0.294906,
		0.291774,
		0.28869,
		0.285652,
		0.282661,
		0.279714,
		0.276811,
		0.273952,
		0.271135,
		0.268359,
		0.265625,
		0.262931,
		0.260276,
		0.257659,
		0.255081,
		0.25254,
		0.250035,
		0.247567,
		0.245133,
		0.242735,
		0.24037,
		0.238038,
		0.23574,
		0.233473,
		0.231239,
		0.229035,
		0.226862,
		0.224718,
		0.222604,
		0.220519,
		0.218463,
		0.216434,
		0.214433,
		0.212459,
		0.210511,
		0.208589,
		0.206693,
		0.204822,
		0.202976,
		0.201154,
		0.199356,
		0.197581,
		0.19583,
		0.194101,
		0.192394,
		0.19071,
		0.189047,
		0.187405,
		0.185784,
		0.184184,
		0.182603,
		0.181043,
		0.179502,
		0.177981,
		0.176478,
		0.174994,
		0.173528,
		0.172081,
		0.170651,
		0.169238,
		0.167843,
		0.166465,
		0.165103,
		0.163757,
		0.162428,
		0.161115,
		0.159817,
		0.158534,
		0.157267,
		0.156015,
		0.154777,
		0.153554,
		0.152345,
		0.151149,
		0.149968,
		0.148801,
		0.147646,
		0.146505,
		0.145377,
		0.144262,
		0.14316,
		0.142069,
		0.140991,
		0.139926,
		0.138872,
		0.137829,
		0.136799,
		0.135779,
		0.134771,
		0.133774,
		0.132788,
		0.131812,
		0.130847,
		0.129893,
		0.128949,
		0.128015,
		0.127091,
		0.126177,
		0.125272,
		0.124377,
		0.123492,
		0.122616,
		0.121749,
		0.120891,
		0.120042,
		0.119202,
		3.3e-005,
		0.000159,
		0.000475,
		0.001096,
		0.002146,
		0.003753,
		0.006041,
		0.009125,
		0.01311,
		0.018083,
		0.024116,
		0.031258,
		0.039541,
		0.048973,
		0.05954,
		0.071206,
		0.08391,
		0.097571,
		0.112084,
		0.127325,
		0.143149,
		0.159395,
		0.175889,
		0.192443,
		0.208867,
		0.224966,
		0.240551,
		0.255441,
		0.269468,
		0.282483,
		0.294361,
		0.305002,
		0.314334,
		0.322316,
		0.328938,
		0.334215,
		0.338191,
		0.340931,
		0.342523,
		0.343066,
		0.34267,
		0.341453,
		0.339533,
		0.337025,
		0.334039,
		0.330677,
		0.327031,
		0.323181,
		0.319197,
		0.315133,
		0.311037,
		0.306942,
		0.302876,
		0.298855,
		0.294891,
		0.290993,
		0.287164,
		0.283404,
		0.279714,
		0.276092,
		0.272538,
		0.269049,
		0.265625,
		0.262263,
		0.258963,
		0.255722,
		0.25254,
		0.249415,
		0.246346,
		0.243331,
		0.24037,
		0.237461,
		0.234603,
		0.231794,
		0.229035,
		0.226323,
		0.223658,
		0.221038,
		0.218463,
		0.215931,
		0.213443,
		0.210995,
		0.208589,
		0.206223,
		0.203896,
		0.201607,
		0.199356,
		0.197141,
		0.194962,
		0.192819,
		0.19071,
		0.188634,
		0.186592,
		0.184582,
		0.182603,
		0.180656,
		0.178739,
		0.176852,
		0.174994,
		0.173165,
		0.171364,
		0.16959,
		0.167843,
		0.166123,
		0.164428,
		0.162759,
		0.161115,
		0.159495,
		0.157899,
		0.156326,
		0.154777,
		0.15325,
		0.151745,
		0.150262,
		0.148801,
		0.14736,
		0.14594,
		0.14454,
		0.14316,
		0.141799,
		0.140457,
		0.139134,
		0.137829,
		0.136543,
		0.135274,
		0.134022,
		0.132788,
		0.13157,
		0.130369,
		0.129184,
		0.128015,
		0.126861,
		0.125723,
		0.1246,
		0.123492,
		0.122398,
		0.121319,
		0.120254,
		0.119202,
		0.118164,
		0.11714,
		0.116129,
		0.11513,
		0.114144,
		0.113171,
		0.11221,
		0.111261,
		0.110324,
		0.109399,
		0.108485,
		0.107582,
		0.106691,
		0.10581,
		0.104941,
		0.104082,
		0.103233,
		0.102394,
		0.101566,
		0.100747,
		0.099939,
		0.09914,
		0.09835,
		0.09757,
		0.096799,
		0.096037,
		0.095283,
		0.094539,
		0.093803,
		0.093076,
		0.092357,
		0.091646,
		0.090944,
		0.090249,
		0.089563,
		0.088884,
		0.088212,
		0.087549,
		0.086892,
		0.086243,
		0.085601,
		0.084966,
		0.084339,
		0.083718,
		0.083104,
		0.082496,
		0.081895,
		0.081301,
		0.080713,
		0.080132,
		0.079556,
		2.3e-005,
		0.000109,
		0.000326,
		0.000748,
		0.00146,
		0.002543,
		0.00408,
		0.006142,
		0.008796,
		0.012095,
		0.016081,
		0.020783,
		0.026217,
		0.032382,
		0.039266,
		0.046839,
		0.05506,
		0.063871,
		0.073201,
		0.082967,
		0.093074,
		0.103418,
		0.113883,
		0.124352,
		0.134701,
		0.144809,
		0.154555,
		0.163827,
		0.172521,
		0.180546,
		0.187825,
		0.1943,
		0.199929,
		0.204691,
		0.208583,
		0.211619,
		0.213831,
		0.215263,
		0.215973,
		0.216028,
		0.215498,
		0.21446,
		0.212988,
		0.211158,
		0.209038,
		0.206694,
		0.204182,
		0.201554,
		0.198852,
		0.196111,
		0.19336,
		0.190619,
		0.187904,
		0.185227,
		0.182595,
		0.180011,
		0.177477,
		0.174994,
		0.172561,
		0.170178,
		0.167843,
		0.165555,
		0.163312,
		0.161115,
		0.15896,
		0.156848,
		0.154777,
		0.152746,
		0.150754,
		0.148801,
		0.146884,
		0.145004,
		0.14316,
		0.141349,
		0.139573,
		0.137829,
		0.136118,
		0.134438,
		0.132788,
		0.131168,
		0.129577,
		0.128015,
		0.12648,
		0.124973,
		0.123492,
		0.122037,
		0.120607,
		0.119202,
		0.117821,
		0.116464,
		0.11513,
		0.113819,
		0.112529,
		0.111261,
		0.110014,
		0.108788,
		0.107582,
		0.106396,
		0.105229,
		0.104082,
		0.102952,
		0.101841,
		0.100747,
		0.099671,
		0.098612,
		0.09757,
		0.096544,
		0.095534,
		0.094539,
		0.09356,
		0.092596,
		0.091646,
		0.090711,
		0.089791,
		0.088884,
		0.08799,
		0.08711,
		0.086243,
		0.085389,
		0.084547,
		0.083718,
		0.0829,
		0.082095,
		0.081301,
		0.080519,
		0.079747,
		0.078987,
		0.078237,
		0.077498,
		0.07677,
		0.076051,
		0.075343,
		0.074644,
		0.073955,
		0.073275,
		0.072605,
		0.071944,
		0.071291,
		0.070648,
		0.070013,
		0.069387,
		0.068768,
		0.068159,
		0.067557,
		0.066963,
		0.066377,
		0.065798,
		0.065227,
		0.064663,
		0.064107,
		0.063557,
		0.063015,
		0.06248,
		0.061951,
		0.061429,
		0.060913,
		0.060404,
		0.059902,
		0.059405,
		0.058915,
		0.05843,
		0.057952,
		0.057479,
		0.057012,
		0.056551,
		0.056095,
		0.055645,
		0.0552,
		0.054761,
		0.054327,
		0.053897,
		0.053473,
		0.053054,
		0.05264,
		0.05223,
		0.051826,
		0.051425,
		0.05103,
		0.050639,
		0.050253,
		0.049871,
		0.049493,
		0.049119,
		0.04875,
		0.048385,
		0.048024,
		0.047667,
		0.047314,
		0.046964,
		0.046619,
		1.2e-005,
		5.8e-005,
		0.000173,
		0.000396,
		0.00077,
		0.001339,
		0.002141,
		0.003214,
		0.004591,
		0.006296,
		0.008349,
		0.010764,
		0.013546,
		0.016693,
		0.020197,
		0.02404,
		0.028199,
		0.032644,
		0.037339,
		0.042238,
		0.047295,
		0.052454,
		0.057659,
		0.06285,
		0.067966,
		0.072946,
		0.077731,
		0.082265,
		0.086499,
		0.090387,
		0.093895,
		0.096994,
		0.099666,
		0.101903,
		0.103703,
		0.105077,
		0.106041,
		0.10662,
		0.106843,
		0.106744,
		0.10636,
		0.105728,
		0.104887,
		0.103873,
		0.102721,
		0.101464,
		0.100129,
		0.098742,
		0.097323,
		0.09589,
		0.094456,
		0.093032,
		0.091624,
		0.090239,
		0.088879,
		0.087547,
		0.086243,
		0.084966,
		0.083718,
		0.082496,
		0.081301,
		0.080132,
		0.078987,
		0.077867,
		0.07677,
		0.075696,
		0.074644,
		0.073614,
		0.072605,
		0.071616,
		0.070648,
		0.069699,
		0.068768,
		0.067857,
		0.066963,
		0.066086,
		0.065227,
		0.064384,
		0.063557,
		0.062746,
		0.061951,
		0.06117,
		0.060404,
		0.059653,
		0.058915,
		0.05819,
		0.057479,
		0.056781,
		0.056095,
		0.055422,
		0.054761,
		0.054111,
		0.053473,
		0.052846,
		0.05223,
		0.051625,
		0.05103,
		0.050445,
		0.049871,
		0.049305,
		0.04875,
		0.048204,
		0.047667,
		0.047138,
		0.046619,
		0.046108,
		0.045605,
		0.045111,
		0.044624,
		0.044146,
		0.043674,
		0.043211,
		0.042755,
		0.042305,
		0.041863,
		0.041428,
		0.041,
		0.040578,
		0.040162,
		0.039753,
		0.03935,
		0.038954,
		0.038563,
		0.038178,
		0.037798,
		0.037425,
		0.037057,
		0.036694,
		0.036336,
		0.035984,
		0.035637,
		0.035295,
		0.034957,
		0.034625,
		0.034297,
		0.033974,
		0.033655,
		0.033341,
		0.033031,
		0.032726,
		0.032424,
		0.032127,
		0.031834,
		0.031545,
		0.031259,
		0.030978,
		0.0307,
		0.030427,
		0.030156,
		0.02989,
		0.029626,
		0.029367,
		0.02911,
		0.028857,
		0.028608,
		0.028361,
		0.028118,
		0.027878,
		0.027641,
		0.027406,
		0.027175,
		0.026947,
		0.026722,
		0.026499,
		0.026279,
		0.026062,
		0.025848,
		0.025636,
		0.025427,
		0.02522,
		0.025016,
		0.024814,
		0.024615,
		0.024418,
		0.024223,
		0.024031,
		0.023841,
		0.023653,
		0.023468,
		0.023285,
		0.023103,
		0.022924,
		0.022747,
		0.022572,
		0.022399,
		0.022228,
		0.022059,
		0.021892,
		0.021727,
		0.021564,
		1.7e-005,
		5.1e-005,
		0.000116,
		0.000225,
		0.00039,
		0.000623,
		0.000933,
		0.001329,
		0.001819,
		0.002407,
		0.003097,
		0.00389,
		0.004784,
		0.005777,
		0.006863,
		0.008036,
		0.009287,
		0.010604,
		0.011975,
		0.013387,
		0.014824,
		0.01627,
		0.017708,
		0.019122,
		0.020493,
		0.021807,
		0.023047,
		0.024201,
		0.025256,
		0.026203,
		0.027034,
		0.027745,
		0.028334,
		0.028801,
		0.029149,
		0.029384,
		0.029512,
		0.029542,
		0.029484,
		0.029348,
		0.029144,
		0.028884,
		0.028577,
		0.028234,
		0.027862,
		0.027471,
		0.027066,
		0.026654,
		0.026239,
		0.025825,
		0.025414,
		0.02501,
		0.024612,
		0.024222,
		0.023841,
		0.023468,
		0.023103,
		0.022747,
		0.022399,
		0.022059,
		0.021727,
		0.021402,
		0.021084,
		0.020774,
		0.02047,
		0.020173,
		0.019882,
		0.019597,
		0.019319,
		0.019046,
		0.018779,
		0.018518,
		0.018262,
		0.018011,
		0.017765,
		0.017525,
		0.017289,
		0.017058,
		0.016832,
		0.01661,
		0.016392,
		0.016179,
		0.01597,
		0.015765,
		0.015563,
		0.015366,
		0.015172,
		0.014982,
		0.014796,
		0.014613,
		0.014433,
		0.014256,
		0.014083,
		0.013913,
		0.013746,
		0.013582,
		0.013421,
		0.013263,
		0.013107,
		0.012955,
		0.012805,
		0.012657,
		0.012512,
		0.01237,
		0.01223,
		0.012092,
		0.011957,
		0.011823,
		0.011692,
		0.011564,
		0.011437,
		0.011312,
		0.01119,
		0.011069,
		0.010951,
		0.010834,
		0.010719,
		0.010606,
		0.010495,
		0.010385,
		0.010277,
		0.010171,
		0.010067,
		0.009964,
		0.009862,
		0.009762,
		0.009664,
		0.009567,
		0.009472,
		0.009378,
		0.009285,
		0.009194,
		0.009104,
		0.009016,
		0.008928,
		0.008842,
		0.008758,
		0.008674,
		0.008592,
		0.00851,
		0.00843,
		0.008351,
		0.008273,
		0.008197,
		0.008121,
		0.008046,
		0.007973,
		0.0079,
		0.007828,
		0.007758,
		0.007688,
		0.007619,
		0.007551,
		0.007484,
		0.007418,
		0.007353,
		0.007289,
		0.007225,
		0.007163,
		0.007101,
		0.00704,
		0.006979,
		0.00692,
		0.006861,
		0.006803,
		0.006746,
		0.006689,
		0.006633,
		0.006578,
		0.006524,
		0.00647,
		0.006417,
		0.006364,
		0.006313,
		0.006261,
		0.006211,
		0.006161,
		0.006111,
		0.006063,
		0.006014,
		0.005967,
		0.00592,
		0.005873,
		0.005827,
		0.005782,
		0.005737,
		0.005692,
		0.005649,
		0.005605,
		0.005562,
		0.005605,
		0.005649,
		0.005692,
		0.005737,
		0.005782,
		0.005827,
		0.005873,
		0.00592,
		0.005967,
		0.006014,
		0.006063,
		0.006111,
		0.006161,
		0.006211,
		0.006261,
		0.006313,
		0.006364,
		0.006417,
		0.00647,
		0.006524,
		0.006578,
		0.006633,
		0.006689,
		0.006746,
		0.006803,
		0.006861,
		0.00692,
		0.006979,
		0.00704,
		0.007101,
		0.007163,
		0.007225,
		0.007289,
		0.007353,
		0.007418,
		0.007484,
		0.007551,
		0.007619,
		0.007688,
		0.007758,
		0.007828,
		0.0079,
		0.007973,
		0.008046,
		0.008121,
		0.008197,
		0.008273,
		0.008351,
		0.00843,
		0.00851,
		0.008592,
		0.008674,
		0.008758,
		0.008842,
		0.008928,
		0.009016,
		0.009104,
		0.009194,
		0.009285,
		0.009378,
		0.009472,
		0.009567,
		0.009664,
		0.009762,
		0.009862,
		0.009964,
		0.010067,
		0.010171,
		0.010277,
		0.010385,
		0.010495,
		0.010606,
		0.010719,
		0.010834,
		0.010951,
		0.011069,
		0.01119,
		0.011312,
		0.011437,
		0.011564,
		0.011692,
		0.011823,
		0.011957,
		0.012092,
		0.01223,
		0.01237,
		0.012512,
		0.012657,
		0.012805,
		0.012955,
		0.013107,
		0.013263,
		0.013421,
		0.013582,
		0.013746,
		0.013913,
		0.014083,
		0.014256,
		0.014433,
		0.014613,
		0.014796,
		0.014982,
		0.015172,
		0.015366,
		0.015563,
		0.015765,
		0.01597,
		0.016179,
		0.016392,
		0.01661,
		0.016832,
		0.017058,
		0.017289,
		0.017525,
		0.017765,
		0.018011,
		0.018262,
		0.018518,
		0.018779,
		0.019046,
		0.019319,
		0.019597,
		0.019882,
		0.020173,
		0.02047,
		0.020774,
		0.021084,
		0.021402,
		0.021727,
		0.022059,
		0.022399,
		0.022747,
		0.023103,
		0.023468,
		0.023841,
		0.024222,
		0.024612,
		0.02501,
		0.025414,
		0.025825,
		0.026239,
		0.026654,
		0.027066,
		0.027471,
		0.027862,
		0.028234,
		0.028577,
		0.028884,
		0.029144,
		0.029348,
		0.029484,
		0.029542,
		0.029512,
		0.029384,
		0.029149,
		0.028801,
		0.028334,
		0.027745,
		0.027034,
		0.026203,
		0.025256,
		0.024201,
		0.023047,
		0.021807,
		0.020493,
		0.019122,
		0.017708,
		0.01627,
		0.014824,
		0.013387,
		0.011975,
		0.010604,
		0.009287,
		0.008036,
		0.006863,
		0.005777,
		0.004784,
		0.00389,
		0.003097,
		0.002407,
		0.001819,
		0.001329,
		0.000933,
		0.000623,
		0.00039,
		0.000225,
		0.000116,
		5.1e-005,
		1.7e-005,
		0.021402,
		0.021564,
		0.021727,
		0.021892,
		0.022059,
		0.022228,
		0.022399,
		0.022572,
		0.022747,
		0.022924,
		0.023103,
		0.023285,
		0.023468,
		0.023653,
		0.023841,
		0.024031,
		0.024223,
		0.024418,
		0.024615,
		0.024814,
		0.025016,
		0.02522,
		0.025427,
		0.025636,
		0.025848,
		0.026062,
		0.026279,
		0.026499,
		0.026722,
		0.026947,
		0.027175,
		0.027406,
		0.027641,
		0.027878,
		0.028118,
		0.028361,
		0.028608,
		0.028857,
		0.02911,
		0.029367,
		0.029626,
		0.02989,
		0.030156,
		0.030427,
		0.0307,
		0.030978,
		0.031259,
		0.031545,
		0.031834,
		0.032127,
		0.032424,
		0.032726,
		0.033031,
		0.033341,
		0.033655,
		0.033974,
		0.034297,
		0.034625,
		0.034957,
		0.035295,
		0.035637,
		0.035984,
		0.036336,
		0.036694,
		0.037057,
		0.037425,
		0.037798,
		0.038178,
		0.038563,
		0.038954,
		0.03935,
		0.039753,
		0.040162,
		0.040578,
		0.041,
		0.041428,
		0.041863,
		0.042305,
		0.042755,
		0.043211,
		0.043674,
		0.044146,
		0.044624,
		0.045111,
		0.045605,
		0.046108,
		0.046619,
		0.047138,
		0.047667,
		0.048204,
		0.04875,
		0.049305,
		0.049871,
		0.050445,
		0.05103,
		0.051625,
		0.05223,
		0.052846,
		0.053473,
		0.054111,
		0.054761,
		0.055422,
		0.056095,
		0.056781,
		0.057479,
		0.05819,
		0.058915,
		0.059653,
		0.060404,
		0.06117,
		0.061951,
		0.062746,
		0.063557,
		0.064384,
		0.065227,
		0.066086,
		0.066963,
		0.067857,
		0.068768,
		0.069699,
		0.070648,
		0.071616,
		0.072605,
		0.073614,
		0.074644,
		0.075696,
		0.07677,
		0.077867,
		0.078987,
		0.080132,
		0.081301,
		0.082496,
		0.083718,
		0.084966,
		0.086243,
		0.087547,
		0.088879,
		0.090239,
		0.091624,
		0.093032,
		0.094456,
		0.09589,
		0.097323,
		0.098742,
		0.100129,
		0.101464,
		0.102721,
		0.103873,
		0.104887,
		0.105728,
		0.10636,
		0.106744,
		0.106843,
		0.10662,
		0.106041,
		0.105077,
		0.103703,
		0.101903,
		0.099666,
		0.096994,
		0.093895,
		0.090387,
		0.086499,
		0.082265,
		0.077731,
		0.072946,
		0.067966,
		0.06285,
		0.057659,
		0.052454,
		0.047295,
		0.042238,
		0.037339,
		0.032644,
		0.028199,
		0.02404,
		0.020197,
		0.016693,
		0.013546,
		0.010764,
		0.008349,
		0.006296,
		0.004591,
		0.003214,
		0.002141,
		0.001339,
		0.00077,
		0.000396,
		0.000173,
		5.8e-005,
		1.2e-005,
		0.046277,
		0.046619,
		0.046964,
		0.047314,
		0.047667,
		0.048024,
		0.048385,
		0.04875,
		0.049119,
		0.049493,
		0.049871,
		0.050253,
		0.050639,
		0.05103,
		0.051425,
		0.051826,
		0.05223,
		0.05264,
		0.053054,
		0.053473,
		0.053897,
		0.054327,
		0.054761,
		0.0552,
		0.055645,
		0.056095,
		0.056551,
		0.057012,
		0.057479,
		0.057952,
		0.05843,
		0.058915,
		0.059405,
		0.059902,
		0.060404,
		0.060913,
		0.061429,
		0.061951,
		0.06248,
		0.063015,
		0.063557,
		0.064107,
		0.064663,
		0.065227,
		0.065798,
		0.066377,
		0.066963,
		0.067557,
		0.068159,
		0.068768,
		0.069387,
		0.070013,
		0.070648,
		0.071291,
		0.071944,
		0.072605,
		0.073275,
		0.073955,
		0.074644,
		0.075343,
		0.076051,
		0.07677,
		0.077498,
		0.078237,
		0.078987,
		0.079747,
		0.080519,
		0.081301,
		0.082095,
		0.0829,
		0.083718,
		0.084547,
		0.085389,
		0.086243,
		0.08711,
		0.08799,
		0.088884,
		0.089791,
		0.090711,
		0.091646,
		0.092596,
		0.09356,
		0.094539,
		0.095534,
		0.096544,
		0.09757,
		0.098612,
		0.099671,
		0.100747,
		0.101841,
		0.102952,
		0.104082,
		0.105229,
		0.106396,
		0.107582,
		0.108788,
		0.110014,
		0.111261,
		0.112529,
		0.113819,
		0.11513,
		0.116464,
		0.117821,
		0.119202,
		0.120607,
		0.122037,
		0.123492,
		0.124973,
		0.12648,
		0.128015,
		0.129577,
		0.131168,
		0.132788,
		0.134438,
		0.136118,
		0.137829,
		0.139573,
		0.141349,
		0.14316,
		0.145004,
		0.146884,
		0.148801,
		0.150754,
		0.152746,
		0.154777,
		0.156848,
		0.15896,
		0.161115,
		0.163312,
		0.165555,
		0.167843,
		0.170178,
		0.172561,
		0.174994,
		0.177477,
		0.180011,
		0.182595,
		0.185227,
		0.187904,
		0.190619,
		0.19336,
		0.196111,
		0.198852,
		0.201554,
		0.204182,
		0.206694,
		0.209038,
		0.211158,
		0.212988,
		0.21446,
		0.215498,
		0.216028,
		0.215973,
		0.215263,
		0.213831,
		0.211619,
		0.208583,
		0.204691,
		0.199929,
		0.1943,
		0.187825,
		0.180546,
		0.172521,
		0.163827,
		0.154555,
		0.144809,
		0.134701,
		0.124352,
		0.113883,
		0.103418,
		0.093074,
		0.082967,
		0.073201,
		0.063871,
		0.05506,
		0.046839,
		0.039266,
		0.032382,
		0.026217,
		0.020783,
		0.016081,
		0.012095,
		0.008796,
		0.006142,
		0.00408,
		0.002543,
		0.00146,
		0.000748,
		0.000326,
		0.000109,
		2.3e-005,
		0.078987,
		0.079556,
		0.080132,
		0.080713,
		0.081301,
		0.081895,
		0.082496,
		0.083104,
		0.083718,
		0.084339,
		0.084966,
		0.085601,
		0.086243,
		0.086892,
		0.087549,
		0.088212,
		0.088884,
		0.089563,
		0.090249,
		0.090944,
		0.091646,
		0.092357,
		0.093076,
		0.093803,
		0.094539,
		0.095283,
		0.096037,
		0.096799,
		0.09757,
		0.09835,
		0.09914,
		0.099939,
		0.100747,
		0.101566,
		0.102394,
		0.103233,
		0.104082,
		0.104941,
		0.10581,
		0.106691,
		0.107582,
		0.108485,
		0.109399,
		0.110324,
		0.111261,
		0.11221,
		0.113171,
		0.114144,
		0.11513,
		0.116129,
		0.11714,
		0.118164,
		0.119202,
		0.120254,
		0.121319,
		0.122398,
		0.123492,
		0.1246,
		0.125723,
		0.126861,
		0.128015,
		0.129184,
		0.130369,
		0.13157,
		0.132788,
		0.134022,
		0.135274,
		0.136543,
		0.137829,
		0.139134,
		0.140457,
		0.141799,
		0.14316,
		0.14454,
		0.14594,
		0.14736,
		0.148801,
		0.150262,
		0.151745,
		0.15325,
		0.154777,
		0.156326,
		0.157899,
		0.159495,
		0.161115,
		0.162759,
		0.164428,
		0.166123,
		0.167843,
		0.16959,
		0.171364,
		0.173165,
		0.174994,
		0.176852,
		0.178739,
		0.180656,
		0.182603,
		0.184582,
		0.186592,
		0.188634,
		0.19071,
		0.192819,
		0.194962,
		0.197141,
		0.199356,
		0.201607,
		0.203896,
		0.206223,
		0.208589,
		0.210995,
		0.213443,
		0.215931,
		0.218463,
		0.221038,
		0.223658,
		0.226323,
		0.229035,
		0.231794,
		0.234603,
		0.237461,
		0.24037,
		0.243331,
		0.246346,
		0.249415,
		0.25254,
		0.255722,
		0.258963,
		0.262263,
		0.265625,
		0.269049,
		0.272538,
		0.276092,
		0.279714,
		0.283404,
		0.287164,
		0.290993,
		0.294891,
		0.298855,
		0.302876,
		0.306942,
		0.311037,
		0.315133,
		0.319197,
		0.323181,
		0.327031,
		0.330677,
		0.334039,
		0.337025,
		0.339533,
		0.341453,
		0.34267,
		0.343066,
		0.342523,
		0.340931,
		0.338191,
		0.334215,
		0.328938,
		0.322316,
		0.314334,
		0.305002,
		0.294361,
		0.282483,
		0.269468,
		0.255441,
		0.240551,
		0.224966,
		0.208867,
		0.192443,
		0.175889,
		0.159395,
		0.143149,
		0.127325,
		0.112084,
		0.097571,
		0.08391,
		0.071206,
		0.05954,
		0.048973,
		0.039541,
		0.031258,
		0.024116,
		0.018083,
		0.01311,
		0.009125,
		0.006041,
		0.003753,
		0.002146,
		0.001096,
		0.000475,
		0.000159,
		3.3e-005,
		0.118371,
		0.119202,
		0.120042,
		0.120891,
		0.121749,
		0.122616,
		0.123492,
		0.124377,
		0.125272,
		0.126177,
		0.127091,
		0.128015,
		0.128949,
		0.129893,
		0.130847,
		0.131812,
		0.132788,
		0.133774,
		0.134771,
		0.135779,
		0.136799,
		0.137829,
		0.138872,
		0.139926,
		0.140991,
		0.142069,
		0.14316,
		0.144262,
		0.145377,
		0.146505,
		0.147646,
		0.148801,
		0.149968,
		0.151149,
		0.152345,
		0.153554,
		0.154777,
		0.156015,
		0.157267,
		0.158534,
		0.159817,
		0.161115,
		0.162428,
		0.163757,
		0.165103,
		0.166465,
		0.167843,
		0.169238,
		0.170651,
		0.172081,
		0.173528,
		0.174994,
		0.176478,
		0.177981,
		0.179502,
		0.181043,
		0.182603,
		0.184184,
		0.185784,
		0.187405,
		0.189047,
		0.19071,
		0.192394,
		0.194101,
		0.19583,
		0.197581,
		0.199356,
		0.201154,
		0.202976,
		0.204822,
		0.206693,
		0.208589,
		0.210511,
		0.212459,
		0.214433,
		0.216434,
		0.218463,
		0.220519,
		0.222604,
		0.224718,
		0.226862,
		0.229035,
		0.231239,
		0.233473,
		0.23574,
		0.238038,
		0.24037,
		0.242735,
		0.245133,
		0.247567,
		0.250035,
		0.25254,
		0.255081,
		0.257659,
		0.260276,
		0.262931,
		0.265625,
		0.268359,
		0.271135,
		0.273952,
		0.276811,
		0.279714,
		0.282661,
		0.285652,
		0.28869,
		0.291774,
		0.294906,
		0.298086,
		0.301315,
		0.304595,
		0.307926,
		0.31131,
		0.314747,
		0.318238,
		0.321785,
		0.325389,
		0.32905,
		0.332769,
		0.336549,
		0.34039,
		0.344293,
		0.34826,
		0.352291,
		0.356389,
		0.360554,
		0.364787,
		0.369091,
		0.373465,
		0.377913,
		0.382435,
		0.387032,
		0.391707,
		0.39646,
		0.401293,
		0.406206,
		0.411199,
		0.416269,
		0.421411,
		0.426612,
		0.431857,
		0.437117,
		0.442357,
		0.447525,
		0.452559,
		0.457378,
		0.461886,
		0.465973,
		0.46951,
		0.472358,
		0.474364,
		0.475373,
		0.475222,
		0.473757,
		0.470829,
		0.466305,
		0.460076,
		0.45206,
		0.442204,
		0.430498,
		0.416966,
		0.401677,
		0.384738,
		0.366296,
		0.346533,
		0.325662,
		0.303917,
		0.281552,
		0.25883,
		0.236018,
		0.213377,
		0.191158,
		0.169596,
		0.148905,
		0.129275,
		0.110866,
		0.09381,
		0.078208,
		0.064131,
		0.051615,
		0.040669,
		0.031269,
		0.023365,
		0.016877,
		0.011703,
		0.007717,
		0.004775,
		0.002719,
		0.001382,
		0.000596,
		0.000199,
		4.1e-005,
		0.163312,
		0.164428,
		0.165555,
		0.166693,
		0.167843,
		0.169005,
		0.170178,
		0.171364,
		0.172561,
		0.173771,
		0.174994,
		0.17623,
		0.177478,
		0.178739,
		0.180014,
		0.181302,
		0.182603,
		0.183919,
		0.185248,
		0.186592,
		0.18795,
		0.189322,
		0.19071,
		0.192112,
		0.193529,
		0.194962,
		0.196411,
		0.197875,
		0.199356,
		0.200852,
		0.202366,
		0.203896,
		0.205443,
		0.207007,
		0.208589,
		0.210189,
		0.211807,
		0.213443,
		0.215097,
		0.21677,
		0.218463,
		0.220175,
		0.221906,
		0.223658,
		0.225429,
		0.227222,
		0.229035,
		0.230869,
		0.232725,
		0.234603,
		0.236502,
		0.238425,
		0.24037,
		0.242338,
		0.24433,
		0.246346,
		0.248386,
		0.25045,
		0.25254,
		0.254655,
		0.256796,
		0.258963,
		0.261156,
		0.263377,
		0.265625,
		0.267901,
		0.270205,
		0.272538,
		0.2749,
		0.277292,
		0.279714,
		0.282166,
		0.28465,
		0.287165,
		0.289713,
		0.292293,
		0.294906,
		0.297552,
		0.300233,
		0.302949,
		0.3057,
		0.308487,
		0.31131,
		0.31417,
		0.317068,
		0.320005,
		0.32298,
		0.325995,
		0.32905,
		0.332145,
		0.335282,
		0.338462,
		0.341684,
		0.34495,
		0.34826,
		0.351615,
		0.355015,
		0.358463,
		0.361957,
		0.365499,
		0.369091,
		0.372731,
		0.376422,
		0.380165,
		0.383959,
		0.387806,
		0.391707,
		0.395662,
		0.399673,
		0.403741,
		0.407865,
		0.412048,
		0.416289,
		0.420591,
		0.424954,
		0.429378,
		0.433865,
		0.438416,
		0.443032,
		0.447714,
		0.452462,
		0.457278,
		0.462163,
		0.467118,
		0.472143,
		0.47724,
		0.48241,
		0.487653,
		0.492972,
		0.498366,
		0.503836,
		0.509384,
		0.515011,
		0.520717,
		0.526502,
		0.532363,
		0.538297,
		0.544295,
		0.550341,
		0.556411,
		0.56247,
		0.56847,
		0.574344,
		0.580011,
		0.585368,
		0.590294,
		0.594644,
		0.59826,
		0.600964,
		0.602567,
		0.602874,
		0.601687,
		0.598814,
		0.594078,
		0.58732,
		0.578414,
		0.567267,
		0.553828,
		0.538097,
		0.520121,
		0.5,
		0.477883,
		0.453969,
		0.428498,
		0.401745,
		0.374015,
		0.34563,
		0.316924,
		0.288229,
		0.259871,
		0.232157,
		0.205375,
		0.179781,
		0.1556,
		0.133017,
		0.112184,
		0.09321,
		0.076164,
		0.061078,
		0.047945,
		0.036721,
		0.027329,
		0.019659,
		0.013573,
		0.00891,
		0.005488,
		0.00311,
		0.001573,
		0.000676,
		0.000224,
		4.6e-005,
		0.212739,
		0.214149,
		0.215573,
		0.217011,
		0.218463,
		0.219929,
		0.221409,
		0.222905,
		0.224414,
		0.225939,
		0.227479,
		0.229035,
		0.230606,
		0.232193,
		0.233795,
		0.235414,
		0.237049,
		0.238701,
		0.24037,
		0.242056,
		0.243759,
		0.245479,
		0.247217,
		0.248973,
		0.250747,
		0.25254,
		0.254351,
		0.256181,
		0.258031,
		0.2599,
		0.261788,
		0.263696,
		0.265625,
		0.267574,
		0.269544,
		0.271535,
		0.273547,
		0.27558,
		0.277636,
		0.279714,
		0.281814,
		0.283937,
		0.286083,
		0.288253,
		0.290446,
		0.292664,
		0.294906,
		0.297172,
		0.299464,
		0.301781,
		0.304124,
		0.306492,
		0.308888,
		0.31131,
		0.31376,
		0.316237,
		0.318742,
		0.321275,
		0.323837,
		0.326429,
		0.32905,
		0.331701,
		0.334382,
		0.337094,
		0.339837,
		0.342613,
		0.34542,
		0.34826,
		0.351133,
		0.354039,
		0.35698,
		0.359954,
		0.362964,
		0.366009,
		0.369091,
		0.372208,
		0.375363,
		0.378555,
		0.381784,
		0.385053,
		0.38836,
		0.391707,
		0.395094,
		0.398522,
		0.40199,
		0.405501,
		0.409054,
		0.41265,
		0.416289,
		0.419973,
		0.423701,
		0.427474,
		0.431293,
		0.435159,
		0.439072,
		0.443032,
		0.447041,
		0.451099,
		0.455206,
		0.459363,
		0.463572,
		0.467831,
		0.472143,
		0.476508,
		0.480925,
		0.485397,
		0.489923,
		0.494505,
		0.499142,
		0.503836,
		0.508587,
		0.513395,
		0.518262,
		0.523188,
		0.528173,
		0.533218,
		0.538323,
		0.54349,
		0.548718,
		0.554008,
		0.559361,
		0.564777,
		0.570256,
		0.575799,
		0.581406,
		0.587077,
		0.592813,
		0.598615,
		0.604481,
		0.610413,
		0.61641,
		0.622473,
		0.628601,
		0.634793,
		0.641049,
		0.647364,
		0.653732,
		0.66014,
		0.666569,
		0.672987,
		0.679351,
		0.685602,
		0.691658,
		0.697422,
		0.702768,
		0.707552,
		0.711604,
		0.714734,
		0.716734,
		0.717384,
		0.716456,
		0.713725,
		0.708972,
		0.701997,
		0.692627,
		0.680727,
		0.666205,
		0.649019,
		0.629185,
		0.606781,
		0.581942,
		0.554862,
		0.525792,
		0.495029,
		0.462908,
		0.429797,
		0.39608,
		0.36215,
		0.328396,
		0.295193,
		0.262895,
		0.231825,
		0.202269,
		0.174471,
		0.148632,
		0.124906,
		0.103399,
		0.084171,
		0.067237,
		0.05257,
		0.040099,
		0.029718,
		0.021286,
		0.014633,
		0.009563,
		0.005863,
		0.003307,
		0.001665,
		0.000712,
		0.000235,
		4.8e-005,
		0.265625,
		0.267329,
		0.269049,
		0.270786,
		0.272538,
		0.274307,
		0.276092,
		0.277895,
		0.279714,
		0.28155,
		0.283404,
		0.285276,
		0.287165,
		0.289073,
		0.290998,
		0.292943,
		0.294906,
		0.296887,
		0.298888,
		0.300909,
		0.302949,
		0.305009,
		0.307089,
		0.309189,
		0.31131,
		0.313452,
		0.315615,
		0.317799,
		0.320005,
		0.322233,
		0.324482,
		0.326755,
		0.32905,
		0.331368,
		0.333709,
		0.336073,
		0.338462,
		0.340874,
		0.343311,
		0.345773,
		0.34826,
		0.350772,
		0.353309,
		0.355873,
		0.358463,
		0.361079,
		0.363722,
		0.366393,
		0.369091,
		0.371816,
		0.37457,
		0.377353,
		0.380165,
		0.383005,
		0.385876,
		0.388776,
		0.391707,
		0.394668,
		0.397661,
		0.400685,
		0.403741,
		0.406828,
		0.409949,
		0.413102,
		0.416289,
		0.41951,
		0.422765,
		0.426054,
		0.429378,
		0.432738,
		0.436133,
		0.439564,
		0.443032,
		0.446537,
		0.45008,
		0.45366,
		0.457278,
		0.460936,
		0.464632,
		0.468368,
		0.472143,
		0.475959,
		0.479816,
		0.483714,
		0.487653,
		0.491635,
		0.495659,
		0.499726,
		0.503836,
		0.50799,
		0.512188,
		0.51643,
		0.520718,
		0.52505,
		0.529428,
		0.533853,
		0.538323,
		0.542841,
		0.547405,
		0.552017,
		0.556677,
		0.561384,
		0.56614,
		0.570945,
		0.575799,
		0.580701,
		0.585653,
		0.590655,
		0.595706,
		0.600807,
		0.605958,
		0.611159,
		0.61641,
		0.621711,
		0.627063,
		0.632464,
		0.637915,
		0.643416,
		0.648967,
		0.654566,
		0.660215,
		0.665912,
		0.671658,
		0.677451,
		0.683291,
		0.689177,
		0.695109,
		0.701086,
		0.707107,
		0.71317,
		0.719275,
		0.72542,
		0.731605,
		0.737825,
		0.74408,
		0.750362,
		0.756663,
		0.762966,
		0.769247,
		0.77547,
		0.781582,
		0.787513,
		0.793172,
		0.798444,
		0.803188,
		0.807238,
		0.810403,
		0.812471,
		0.81321,
		0.812377,
		0.809723,
		0.805002,
		0.797981,
		0.788449,
		0.77623,
		0.761188,
		0.743241,
		0.722367,
		0.698606,
		0.672067,
		0.642924,
		0.611418,
		0.577846,
		0.542556,
		0.505937,
		0.468408,
		0.430401,
		0.392355,
		0.354699,
		0.317842,
		0.282165,
		0.248009,
		0.215674,
		0.185408,
		0.15741,
		0.131824,
		0.108744,
		0.088209,
		0.070213,
		0.054701,
		0.041577,
		0.030705,
		0.021916,
		0.015015,
		0.00978,
		0.005977,
		0.003361,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.320992,
		0.32298,
		0.324985,
		0.327009,
		0.32905,
		0.331109,
		0.333186,
		0.335282,
		0.337397,
		0.339531,
		0.341684,
		0.343856,
		0.346048,
		0.34826,
		0.350491,
		0.352743,
		0.355015,
		0.357308,
		0.359622,
		0.361957,
		0.364313,
		0.366691,
		0.369091,
		0.371512,
		0.373956,
		0.376422,
		0.378911,
		0.381424,
		0.383959,
		0.386518,
		0.3891,
		0.391707,
		0.394338,
		0.396993,
		0.399673,
		0.402378,
		0.405109,
		0.407865,
		0.410647,
		0.413455,
		0.416289,
		0.41915,
		0.422038,
		0.424954,
		0.427896,
		0.430867,
		0.433865,
		0.436892,
		0.439948,
		0.443032,
		0.446146,
		0.449289,
		0.452462,
		0.455665,
		0.458899,
		0.462163,
		0.465459,
		0.468785,
		0.472143,
		0.475533,
		0.478955,
		0.48241,
		0.485897,
		0.489418,
		0.492972,
		0.496559,
		0.50018,
		0.503836,
		0.507526,
		0.511251,
		0.515011,
		0.518807,
		0.522638,
		0.526505,
		0.530408,
		0.534347,
		0.538323,
		0.542336,
		0.546387,
		0.550474,
		0.5546,
		0.558763,
		0.562964,
		0.567204,
		0.571482,
		0.575799,
		0.580154,
		0.584549,
		0.588982,
		0.593455,
		0.597967,
		0.602518,
		0.607109,
		0.61174,
		0.61641,
		0.62112,
		0.625869,
		0.630658,
		0.635486,
		0.640354,
		0.645261,
		0.650207,
		0.655192,
		0.660215,
		0.665277,
		0.670377,
		0.675515,
		0.68069,
		0.685901,
		0.69115,
		0.696434,
		0.701753,
		0.707107,
		0.712494,
		0.717915,
		0.723368,
		0.728851,
		0.734365,
		0.739908,
		0.745479,
		0.751077,
		0.756699,
		0.762345,
		0.768013,
		0.773702,
		0.779408,
		0.785132,
		0.790869,
		0.796619,
		0.802379,
		0.808146,
		0.813918,
		0.819691,
		0.825462,
		0.831226,
		0.836975,
		0.842698,
		0.848375,
		0.853978,
		0.859466,
		0.864779,
		0.86984,
		0.874547,
		0.878774,
		0.882368,
		0.885148,
		0.88691,
		0.887425,
		0.886448,
		0.883723,
		0.878988,
		0.871991,
		0.862495,
		0.850289,
		0.835202,
		0.817112,
		0.795954,
		0.771729,
		0.744506,
		0.714427,
		0.681705,
		0.646618,
		0.609506,
		0.570757,
		0.530799,
		0.490085,
		0.449081,
		0.408251,
		0.368043,
		0.328883,
		0.291157,
		0.255208,
		0.22133,
		0.18976,
		0.160683,
		0.134224,
		0.110452,
		0.089387,
		0.070995,
		0.055198,
		0.041877,
		0.030876,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.377913,
		0.380165,
		0.382435,
		0.384724,
		0.387032,
		0.38936,
		0.391707,
		0.394074,
		0.39646,
		0.398867,
		0.401293,
		0.403741,
		0.406208,
		0.408697,
		0.411206,
		0.413737,
		0.416289,
		0.418863,
		0.421459,
		0.424076,
		0.426716,
		0.429378,
		0.432063,
		0.43477,
		0.437501,
		0.440255,
		0.443032,
		0.445833,
		0.448658,
		0.451507,
		0.454381,
		0.457278,
		0.460201,
		0.463149,
		0.466121,
		0.469119,
		0.472143,
		0.475193,
		0.478268,
		0.48137,
		0.484498,
		0.487653,
		0.490835,
		0.494044,
		0.497281,
		0.500544,
		0.503836,
		0.507156,
		0.510503,
		0.51388,
		0.517284,
		0.520718,
		0.52418,
		0.527672,
		0.531193,
		0.534743,
		0.538323,
		0.541933,
		0.545574,
		0.549244,
		0.552945,
		0.556677,
		0.560439,
		0.564232,
		0.568056,
		0.571912,
		0.575799,
		0.579717,
		0.583667,
		0.587648,
		0.591661,
		0.595706,
		0.599783,
		0.603892,
		0.608032,
		0.612205,
		0.61641,
		0.620647,
		0.624916,
		0.629217,
		0.63355,
		0.637915,
		0.642312,
		0.64674,
		0.651201,
		0.655692,
		0.660215,
		0.664769,
		0.669354,
		0.673969,
		0.678615,
		0.683291,
		0.687996,
		0.692731,
		0.697495,
		0.702287,
		0.707107,
		0.711954,
		0.716828,
		0.721729,
		0.726654,
		0.731605,
		0.736579,
		0.741577,
		0.746597,
		0.751638,
		0.756699,
		0.761779,
		0.766878,
		0.771993,
		0.777124,
		0.782268,
		0.787425,
		0.792593,
		0.797771,
		0.802956,
		0.808146,
		0.81334,
		0.818536,
		0.823732,
		0.828924,
		0.834112,
		0.839291,
		0.84446,
		0.849616,
		0.854755,
		0.859876,
		0.864974,
		0.870045,
		0.875088,
		0.880097,
		0.88507,
		0.890001,
		0.894886,
		0.89972,
		0.904491,
		0.909189,
		0.91379,
		0.918264,
		0.922566,
		0.926633,
		0.930382,
		0.933706,
		0.936471,
		0.938519,
		0.939661,
		0.939685,
		0.938357,
		0.935426,
		0.930634,
		0.923719,
		0.914431,
		0.902538,
		0.887842,
		0.870185,
		0.849464,
		0.825636,
		0.798725,
		0.768831,
		0.736123,
		0.700845,
		0.663304,
		0.623865,
		0.582942,
		0.540983,
		0.498456,
		0.455837,
		0.413595,
		0.372177,
		0.332002,
		0.293444,
		0.256834,
		0.222444,
		0.190493,
		0.161142,
		0.134495,
		0.110603,
		0.089464,
		0.071031,
		0.055211,
		0.041877,
		0.030868,
		0.021994,
		0.015042,
		0.00978,
		0.005965,
		0.003347,
		0.001676,
		0.000712,
		0.000233,
		4.8e-005,
		0.435513,
		0.438,
		0.440506,
		0.443032,
		0.445578,
		0.448143,
		0.450728,
		0.453333,
		0.455958,
		0.458604,
		0.46127,
		0.463957,
		0.466665,
		0.469393,
		0.472143,
		0.474914,
		0.477707,
		0.480522,
		0.483358,
		0.486216,
		0.489096,
		0.491999,
		0.494924,
		0.497872,
		0.500843,
		0.503836,
		0.506853,
		0.509893,
		0.512956,
		0.516043,
		0.519153,
		0.522288,
		0.525446,
		0.528629,
		0.531836,
		0.535067,
		0.538323,
		0.541604,
		0.54491,
		0.54824,
		0.551596,
		0.554977,
		0.558383,
		0.561815,
		0.565272,
		0.568755,
		0.572264,
		0.575799,
		0.579359,
		0.582946,
		0.586559,
		0.590198,
		0.593863,
		0.597555,
		0.601273,
		0.605018,
		0.608789,
		0.612586,
		0.61641,
		0.620261,
		0.624138,
		0.628041,
		0.631971,
		0.635927,
		0.63991,
		0.643919,
		0.647954,
		0.652015,
		0.656102,
		0.660215,
		0.664354,
		0.668518,
		0.672708,
		0.676922,
		0.681162,
		0.685426,
		0.689715,
		0.694027,
		0.698364,
		0.702724,
		0.707107,
		0.711512,
		0.71594,
		0.72039,
		0.72486,
		0.729352,
		0.733863,
		0.738394,
		0.742944,
		0.747512,
		0.752097,
		0.756699,
		0.761317,
		0.76595,
		0.770596,
		0.775256,
		0.779928,
		0.784611,
		0.789303,
		0.794004,
		0.798713,
		0.803427,
		0.808146,
		0.812868,
		0.817592,
		0.822315,
		0.827036,
		0.831754,
		0.836467,
		0.841172,
		0.845868,
		0.850552,
		0.855222,
		0.859876,
		0.864511,
		0.869125,
		0.873716,
		0.87828,
		0.882814,
		0.887317,
		0.891784,
		0.896212,
		0.900598,
		0.90494,
		0.909232,
		0.913472,
		0.917657,
		0.921781,
		0.925841,
		0.929834,
		0.933754,
		0.937598,
		0.941362,
		0.94504,
		0.948626,
		0.952111,
		0.955481,
		0.958714,
		0.961779,
		0.96463,
		0.967203,
		0.969415,
		0.971156,
		0.972293,
		0.972665,
		0.972083,
		0.970337,
		0.967195,
		0.96241,
		0.955728,
		0.946897,
		0.935677,
		0.921848,
		0.905228,
		0.885676,
		0.863107,
		0.837497,
		0.808892,
		0.777408,
		0.743232,
		0.706621,
		0.667894,
		0.627422,
		0.585621,
		0.542935,
		0.499827,
		0.456759,
		0.414184,
		0.372533,
		0.332201,
		0.293546,
		0.256874,
		0.222444,
		0.19046,
		0.161075,
		0.134389,
		0.110452,
		0.089266,
		0.070788,
		0.054932,
		0.041577,
		0.030563,
		0.021703,
		0.014781,
		0.009563,
		0.005799,
		0.003231,
		0.001605,
		0.000676,
		0.000219,
		4.4e-005,
		0.492972,
		0.495659,
		0.498366,
		0.501091,
		0.503836,
		0.5066,
		0.509384,
		0.512188,
		0.515011,
		0.517854,
		0.520718,
		0.523601,
		0.526505,
		0.529428,
		0.532373,
		0.535338,
		0.538323,
		0.54133,
		0.544357,
		0.547405,
		0.550474,
		0.553565,
		0.556677,
		0.55981,
		0.562964,
		0.56614,
		0.569338,
		0.572558,
		0.575799,
		0.579062,
		0.582346,
		0.585653,
		0.588982,
		0.592333,
		0.595706,
		0.599101,
		0.602518,
		0.605958,
		0.60942,
		0.612904,
		0.61641,
		0.619939,
		0.62349,
		0.627063,
		0.630658,
		0.634275,
		0.637915,
		0.641577,
		0.645261,
		0.648967,
		0.652694,
		0.656444,
		0.660215,
		0.664008,
		0.667822,
		0.671658,
		0.675515,
		0.679392,
		0.683291,
		0.68721,
		0.69115,
		0.695109,
		0.699089,
		0.703088,
		0.707107,
		0.711144,
		0.715201,
		0.719275,
		0.723368,
		0.727478,
		0.731605,
		0.735749,
		0.739908,
		0.744084,
		0.748275,
		0.75248,
		0.756699,
		0.760931,
		0.765177,
		0.769433,
		0.773702,
		0.77798,
		0.782268,
		0.786565,
		0.790869,
		0.795181,
		0.799498,
		0.80382,
		0.808146,
		0.812474,
		0.816804,
		0.821134,
		0.825463,
		0.829789,
		0.834112,
		0.838428,
		0.842738,
		0.84704,
		0.851331,
		0.85561,
		0.859876,
		0.864126,
		0.868358,
		0.872571,
		0.876762,
		0.880929,
		0.88507,
		0.889182,
		0.893264,
		0.897313,
		0.901325,
		0.905299,
		0.909232,
		0.913121,
		0.916963,
		0.920755,
		0.924495,
		0.928179,
		0.931803,
		0.935365,
		0.938862,
		0.94229,
		0.945646,
		0.948926,
		0.952128,
		0.955247,
		0.95828,
		0.961225,
		0.964076,
		0.966832,
		0.969489,
		0.972044,
		0.974494,
		0.976834,
		0.979061,
		0.981167,
		0.983139,
		0.98496,
		0.986599,
		0.988014,
		0.989142,
		0.989904,
		0.990193,
		0.98988,
		0.988808,
		0.986795,
		0.983637,
		0.979108,
		0.97297,
		0.96498,
		0.954896,
		0.942488,
		0.927549,
		0.909906,
		0.889428,
		0.866039,
		0.839721,
		0.81052,
		0.778552,
		0.743999,
		0.707107,
		0.668181,
		0.627576,
		0.585686,
		0.542935,
		0.499767,
		0.456627,
		0.413956,
		0.372177,
		0.331687,
		0.292845,
		0.255969,
		0.22133,
		0.189147,
		0.159588,
		0.132767,
		0.108744,
		0.087528,
		0.069081,
		0.053314,
		0.040099,
		0.029267,
		0.020615,
		0.013913,
		0.00891,
		0.005342,
		0.002939,
		0.00144,
		0.000596,
		0.00019,
		3.8e-005,
		0.549528,
		0.552374,
		0.555238,
		0.55812,
		0.561021,
		0.563939,
		0.566876,
		0.569832,
		0.572806,
		0.575799,
		0.57881,
		0.58184,
		0.584888,
		0.587956,
		0.591042,
		0.594146,
		0.59727,
		0.600413,
		0.603574,
		0.606755,
		0.609954,
		0.613173,
		0.61641,
		0.619667,
		0.622942,
		0.626236,
		0.629549,
		0.632882,
		0.636233,
		0.639602,
		0.642991,
		0.646399,
		0.649825,
		0.65327,
		0.656733,
		0.660215,
		0.663715,
		0.667234,
		0.670771,
		0.674326,
		0.677898,
		0.681489,
		0.685097,
		0.688723,
		0.692366,
		0.696026,
		0.699703,
		0.703397,
		0.707107,
		0.710833,
		0.714575,
		0.718333,
		0.722107,
		0.725895,
		0.729698,
		0.733515,
		0.737347,
		0.741192,
		0.74505,
		0.748921,
		0.752804,
		0.756699,
		0.760605,
		0.764523,
		0.76845,
		0.772387,
		0.776333,
		0.780288,
		0.78425,
		0.78822,
		0.792195,
		0.796177,
		0.800163,
		0.804153,
		0.808146,
		0.812141,
		0.816138,
		0.820135,
		0.824131,
		0.828126,
		0.832117,
		0.836105,
		0.840087,
		0.844063,
		0.848031,
		0.85199,
		0.855939,
		0.859876,
		0.863799,
		0.867708,
		0.8716,
		0.875474,
		0.879329,
		0.883162,
		0.886971,
		0.890756,
		0.894513,
		0.898242,
		0.901939,
		0.905603,
		0.909232,
		0.912824,
		0.916375,
		0.919885,
		0.92335,
		0.926769,
		0.930138,
		0.933455,
		0.936718,
		0.939924,
		0.943071,
		0.946156,
		0.949176,
		0.952128,
		0.95501,
		0.957819,
		0.960553,
		0.963209,
		0.965784,
		0.968276,
		0.970681,
		0.972999,
		0.975226,
		0.97736,
		0.9794,
		0.981342,
		0.983187,
		0.984931,
		0.986574,
		0.988116,
		0.989554,
		0.99089,
		0.992122,
		0.993251,
		0.994276,
		0.995193,
		0.995994,
		0.996665,
		0.997181,
		0.997502,
		0.997575,
		0.997321,
		0.996642,
		0.995414,
		0.993485,
		0.990679,
		0.986797,
		0.98162,
		0.974912,
		0.966433,
		0.955943,
		0.943208,
		0.928019,
		0.910193,
		0.889587,
		0.866109,
		0.839721,
		0.810449,
		0.778389,
		0.743703,
		0.706621,
		0.667436,
		0.626498,
		0.584203,
		0.540983,
		0.497293,
		0.4536,
		0.410368,
		0.368043,
		0.32705,
		0.287772,
		0.25055,
		0.215674,
		0.183376,
		0.153832,
		0.127154,
		0.103399,
		0.082561,
		0.064584,
		0.049356,
		0.036721,
		0.026482,
		0.018409,
		0.012245,
		0.007717,
		0.004545,
		0.002452,
		0.001175,
		0.000475,
		0.000147,
		2.8e-005,
		0.604481,
		0.607439,
		0.610413,
		0.613403,
		0.61641,
		0.619433,
		0.622473,
		0.625529,
		0.628601,
		0.631689,
		0.634794,
		0.637915,
		0.641052,
		0.644206,
		0.647376,
		0.650561,
		0.653763,
		0.656981,
		0.660215,
		0.663465,
		0.66673,
		0.670011,
		0.673308,
		0.67662,
		0.679948,
		0.683291,
		0.686649,
		0.690022,
		0.69341,
		0.696813,
		0.70023,
		0.703661,
		0.707107,
		0.710566,
		0.71404,
		0.717527,
		0.721027,
		0.72454,
		0.728066,
		0.731605,
		0.735156,
		0.738718,
		0.742293,
		0.745878,
		0.749475,
		0.753082,
		0.756699,
		0.760326,
		0.763962,
		0.767608,
		0.771261,
		0.774923,
		0.778592,
		0.782268,
		0.785951,
		0.789639,
		0.793332,
		0.797031,
		0.800733,
		0.804438,
		0.808146,
		0.811856,
		0.815567,
		0.819279,
		0.82299,
		0.826699,
		0.830407,
		0.834112,
		0.837812,
		0.841508,
		0.845197,
		0.84888,
		0.852555,
		0.85622,
		0.859876,
		0.863519,
		0.867151,
		0.870768,
		0.874369,
		0.877955,
		0.881522,
		0.88507,
		0.888597,
		0.892101,
		0.895582,
		0.899037,
		0.902465,
		0.905864,
		0.909232,
		0.912568,
		0.91587,
		0.919137,
		0.922365,
		0.925553,
		0.9287,
		0.931803,
		0.93486,
		0.93787,
		0.94083,
		0.943738,
		0.946591,
		0.949389,
		0.952128,
		0.954807,
		0.957423,
		0.959974,
		0.962458,
		0.964874,
		0.967218,
		0.969489,
		0.971686,
		0.973805,
		0.975845,
		0.977805,
		0.979683,
		0.981477,
		0.983187,
		0.98481,
		0.986346,
		0.987794,
		0.989154,
		0.990425,
		0.991607,
		0.9927,
		0.993706,
		0.994625,
		0.995458,
		0.996208,
		0.996876,
		0.997465,
		0.997978,
		0.998419,
		0.998793,
		0.999102,
		0.999349,
		0.999536,
		0.999659,
		0.999705,
		0.999655,
		0.999471,
		0.9991,
		0.998467,
		0.997473,
		0.99599,
		0.993865,
		0.990914,
		0.986929,
		0.981678,
		0.974912,
		0.966371,
		0.955795,
		0.942932,
		0.927549,
		0.909445,
		0.888462,
		0.864494,
		0.837497,
		0.807497,
		0.774593,
		0.738959,
		0.700845,
		0.660568,
		0.618508,
		0.575095,
		0.530799,
		0.486112,
		0.441538,
		0.397574,
		0.354699,
		0.313356,
		0.273949,
		0.236823,
		0.202269,
		0.170509,
		0.141698,
		0.115924,
		0.09321,
		0.073513,
		0.056733,
		0.042718,
		0.031269,
		0.02215,
		0.015095,
		0.009823,
		0.006041,
		0.003462,
		0.001811,
		0.000838,
		0.000326,
		9.6e-005,
		1.8e-005,
		0.657196,
		0.660215,
		0.663248,
		0.666294,
		0.669354,
		0.672427,
		0.675515,
		0.678615,
		0.681729,
		0.684856,
		0.687996,
		0.69115,
		0.694316,
		0.697495,
		0.700686,
		0.70389,
		0.707107,
		0.710335,
		0.713576,
		0.716828,
		0.720092,
		0.723368,
		0.726654,
		0.729952,
		0.73326,
		0.736579,
		0.739908,
		0.743248,
		0.746597,
		0.749955,
		0.753323,
		0.756699,
		0.760084,
		0.763477,
		0.766878,
		0.770286,
		0.773702,
		0.777124,
		0.780552,
		0.783986,
		0.787425,
		0.790869,
		0.794318,
		0.797771,
		0.801227,
		0.804685,
		0.808146,
		0.811609,
		0.815072,
		0.818536,
		0.822,
		0.825463,
		0.828924,
		0.832383,
		0.835839,
		0.839291,
		0.842738,
		0.84618,
		0.849616,
		0.853044,
		0.856465,
		0.859876,
		0.863277,
		0.866667,
		0.870045,
		0.873411,
		0.876762,
		0.880097,
		0.883417,
		0.886718,
		0.890001,
		0.893264,
		0.896506,
		0.899725,
		0.90292,
		0.906089,
		0.909232,
		0.912347,
		0.915432,
		0.918486,
		0.921508,
		0.924495,
		0.927446,
		0.930361,
		0.933236,
		0.93607,
		0.938862,
		0.94161,
		0.944313,
		0.946968,
		0.949573,
		0.952128,
		0.95463,
		0.957078,
		0.959469,
		0.961803,
		0.964076,
		0.966289,
		0.968439,
		0.970524,
		0.972543,
		0.974494,
		0.976376,
		0.978187,
		0.979927,
		0.981594,
		0.983187,
		0.984704,
		0.986146,
		0.987511,
		0.9888,
		0.990011,
		0.991144,
		0.992201,
		0.993181,
		0.994084,
		0.994912,
		0.995666,
		0.996348,
		0.996959,
		0.997501,
		0.997978,
		0.998392,
		0.998747,
		0.999046,
		0.999293,
		0.999493,
		0.999651,
		0.999771,
		0.999859,
		0.99992,
		0.999959,
		0.999982,
		0.999993,
		0.999993,
		0.999982,
		0.999952,
		0.999887,
		0.999756,
		0.999516,
		0.9991,
		0.998418,
		0.997356,
		0.99577,
		0.993485,
		0.990301,
		0.985989,
		0.980301,
		0.97297,
		0.963724,
		0.95229,
		0.938409,
		0.921848,
		0.902408,
		0.879941,
		0.854357,
		0.825636,
		0.793834,
		0.759087,
		0.721612,
		0.681705,
		0.639732,
		0.596121,
		0.551352,
		0.505937,
		0.460409,
		0.415302,
		0.371134,
		0.328396,
		0.287533,
		0.248937,
		0.212934,
		0.179781,
		0.149661,
		0.122681,
		0.098875,
		0.078208,
		0.06058,
		0.045834,
		0.033761,
		0.024116,
		0.016622,
		0.010985,
		0.006903,
		0.00408,
		0.002233,
		0.001107,
		0.000481,
		0.000173,
		4.7e-005,
		0.707107,
		0.710133,
		0.71317,
		0.716218,
		0.719275,
		0.722343,
		0.72542,
		0.728508,
		0.731605,
		0.734711,
		0.737827,
		0.740951,
		0.744084,
		0.747226,
		0.750376,
		0.753533,
		0.756699,
		0.759872,
		0.763052,
		0.76624,
		0.769433,
		0.772634,
		0.77584,
		0.779051,
		0.782268,
		0.78549,
		0.788716,
		0.791947,
		0.795181,
		0.798418,
		0.801659,
		0.804901,
		0.808146,
		0.811392,
		0.814639,
		0.817887,
		0.821134,
		0.824381,
		0.827626,
		0.83087,
		0.834112,
		0.83735,
		0.840584,
		0.843815,
		0.84704,
		0.850259,
		0.853472,
		0.856678,
		0.859876,
		0.863065,
		0.866244,
		0.869413,
		0.872571,
		0.875716,
		0.878848,
		0.881966,
		0.88507,
		0.888157,
		0.891227,
		0.894279,
		0.897313,
		0.900326,
		0.903317,
		0.906287,
		0.909232,
		0.912153,
		0.915048,
		0.917916,
		0.920755,
		0.923565,
		0.926344,
		0.92909,
		0.931803,
		0.934481,
		0.937122,
		0.939726,
		0.94229,
		0.944814,
		0.947296,
		0.949734,
		0.952128,
		0.954475,
		0.956775,
		0.959025,
		0.961225,
		0.963372,
		0.965467,
		0.967506,
		0.969489,
		0.971415,
		0.973282,
		0.975089,
		0.976835,
		0.978519,
		0.98014,
		0.981696,
		0.983187,
		0.984612,
		0.98597,
		0.987261,
		0.988485,
		0.989641,
		0.990728,
		0.991748,
		0.9927,
		0.993585,
		0.994403,
		0.995156,
		0.995843,
		0.996468,
		0.99703,
		0.997533,
		0.997978,
		0.998368,
		0.998706,
		0.998994,
		0.999236,
		0.999435,
		0.999596,
		0.999723,
		0.999819,
		0.999889,
		0.999937,
		0.999968,
		0.999986,
		0.999995,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999955,
		0.999904,
		0.999812,
		0.999659,
		0.999412,
		0.999026,
		0.998441,
		0.997575,
		0.996322,
		0.994554,
		0.992111,
		0.988808,
		0.984431,
		0.978745,
		0.971493,
		0.96241,
		0.951225,
		0.937675,
		0.921517,
		0.902538,
		0.88057,
		0.855502,
		0.827286,
		0.795954,
		0.761617,
		0.724468,
		0.684786,
		0.642924,
		0.599306,
		0.554413,
		0.508764,
		0.462908,
		0.417399,
		0.37278,
		0.329566,
		0.288229,
		0.249185,
		0.212781,
		0.17929,
		0.148905,
		0.121741,
		0.097831,
		0.077134,
		0.05954,
		0.044879,
		0.032928,
		0.023426,
		0.016081,
		0.010587,
		0.006631,
		0.003911,
		0.002141,
		0.001066,
		0.00047,
		0.000175,
		5.1e-005,
		1e-005,
		0.753719,
		0.756699,
		0.759685,
		0.762678,
		0.765677,
		0.768681,
		0.771692,
		0.774707,
		0.777728,
		0.780754,
		0.783784,
		0.786818,
		0.789856,
		0.792898,
		0.795942,
		0.79899,
		0.80204,
		0.805092,
		0.808146,
		0.811201,
		0.814257,
		0.817314,
		0.82037,
		0.823426,
		0.826481,
		0.829535,
		0.832587,
		0.835636,
		0.838682,
		0.841725,
		0.844764,
		0.847798,
		0.850827,
		0.85385,
		0.856866,
		0.859876,
		0.862877,
		0.865871,
		0.868855,
		0.871829,
		0.874792,
		0.877744,
		0.880684,
		0.883611,
		0.886525,
		0.889423,
		0.892307,
		0.895174,
		0.898023,
		0.900855,
		0.903668,
		0.90646,
		0.909232,
		0.911982,
		0.914709,
		0.917412,
		0.92009,
		0.922742,
		0.925367,
		0.927963,
		0.930531,
		0.933068,
		0.935573,
		0.938045,
		0.940484,
		0.942888,
		0.945255,
		0.947585,
		0.949876,
		0.952128,
		0.954338,
		0.956507,
		0.958631,
		0.960712,
		0.962746,
		0.964734,
		0.966673,
		0.968563,
		0.970403,
		0.972191,
		0.973927,
		0.975609,
		0.977237,
		0.97881,
		0.980326,
		0.981785,
		0.983187,
		0.98453,
		0.985814,
		0.987038,
		0.988203,
		0.989308,
		0.990352,
		0.991337,
		0.992261,
		0.993125,
		0.99393,
		0.994676,
		0.995364,
		0.995996,
		0.996571,
		0.997092,
		0.997561,
		0.997978,
		0.998347,
		0.998669,
		0.998946,
		0.999183,
		0.999381,
		0.999544,
		0.999675,
		0.999777,
		0.999854,
		0.999911,
		0.99995,
		0.999975,
		0.999989,
		0.999996,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.999986,
		0.999966,
		0.999929,
		0.999867,
		0.999771,
		0.999629,
		0.999429,
		0.999156,
		0.998793,
		0.99832,
		0.997716,
		0.996952,
		0.995994,
		0.994799,
		0.99331,
		0.991455,
		0.989142,
		0.986257,
		0.982663,
		0.978195,
		0.972665,
		0.96586,
		0.95755,
		0.947489,
		0.935426,
		0.921114,
		0.904319,
		0.884835,
		0.862495,
		0.837182,
		0.808844,
		0.777497,
		0.743241,
		0.706256,
		0.666804,
		0.625228,
		0.581942,
		0.537419,
		0.492178,
		0.446768,
		0.401745,
		0.35766,
		0.315035,
		0.274347,
		0.236018,
		0.200397,
		0.167757,
		0.138285,
		0.112084,
		0.089175,
		0.0695,
		0.052927,
		0.039266,
		0.028272,
		0.019662,
		0.013129,
		0.008349,
		0.005004,
		0.002784,
		0.001406,
		0.000623,
		0.000228,
		6.1e-005,
		0.796619,
		0.799498,
		0.802379,
		0.805262,
		0.808146,
		0.811031,
		0.813918,
		0.816804,
		0.819691,
		0.822577,
		0.825463,
		0.828347,
		0.83123,
		0.834112,
		0.83699,
		0.839866,
		0.842738,
		0.845607,
		0.848471,
		0.851331,
		0.854185,
		0.857034,
		0.859876,
		0.862711,
		0.865538,
		0.868358,
		0.871169,
		0.87397,
		0.876762,
		0.879542,
		0.882312,
		0.88507,
		0.887815,
		0.890546,
		0.893264,
		0.895967,
		0.898654,
		0.901325,
		0.903979,
		0.906615,
		0.909232,
		0.91183,
		0.914407,
		0.916963,
		0.919497,
		0.922008,
		0.924495,
		0.926957,
		0.929393,
		0.931803,
		0.934185,
		0.936538,
		0.938862,
		0.941155,
		0.943417,
		0.945646,
		0.947842,
		0.950003,
		0.952128,
		0.954217,
		0.956268,
		0.95828,
		0.960253,
		0.962186,
		0.964076,
		0.965925,
		0.967729,
		0.969489,
		0.971204,
		0.972873,
		0.974494,
		0.976067,
		0.977591,
		0.979066,
		0.980491,
		0.981865,
		0.983187,
		0.984457,
		0.985674,
		0.986838,
		0.987949,
		0.989007,
		0.990011,
		0.990961,
		0.991857,
		0.9927,
		0.99349,
		0.994227,
		0.994912,
		0.995546,
		0.996128,
		0.996662,
		0.997147,
		0.997585,
		0.997978,
		0.998327,
		0.998635,
		0.998903,
		0.999134,
		0.99933,
		0.999493,
		0.999627,
		0.999735,
		0.999819,
		0.999882,
		0.999928,
		0.999959,
		0.99998,
		0.999991,
		0.999997,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999989,
		0.999973,
		0.999944,
		0.999895,
		0.999819,
		0.999707,
		0.999548,
		0.999333,
		0.999046,
		0.998674,
		0.9982,
		0.997607,
		0.996876,
		0.995985,
		0.994912,
		0.993633,
		0.992122,
		0.990351,
		0.988288,
		0.985898,
		0.983139,
		0.979961,
		0.9763,
		0.972079,
		0.967203,
		0.961557,
		0.955004,
		0.947385,
		0.938519,
		0.928207,
		0.916238,
		0.902391,
		0.886448,
		0.868202,
		0.847468,
		0.824096,
		0.797981,
		0.769076,
		0.737402,
		0.703054,
		0.666205,
		0.627109,
		0.586099,
		0.543575,
		0.5,
		0.455881,
		0.411755,
		0.368169,
		0.325662,
		0.284745,
		0.245885,
		0.209488,
		0.175889,
		0.145338,
		0.118003,
		0.09396,
		0.073201,
		0.055636,
		0.041104,
		0.029382,
		0.020197,
		0.013239,
		0.008178,
		0.004677,
		0.002407,
		0.001059,
		0.000356,
		6.5e-005,
		0.835475,
		0.838201,
		0.840925,
		0.843645,
		0.846361,
		0.849074,
		0.851782,
		0.854485,
		0.857183,
		0.859876,
		0.862562,
		0.865241,
		0.867913,
		0.870578,
		0.873234,
		0.875881,
		0.878519,
		0.881147,
		0.883765,
		0.886372,
		0.888967,
		0.891549,
		0.894119,
		0.896676,
		0.899218,
		0.901745,
		0.904257,
		0.906753,
		0.909232,
		0.911694,
		0.914137,
		0.916561,
		0.918966,
		0.92135,
		0.923712,
		0.926053,
		0.928371,
		0.930665,
		0.932935,
		0.935179,
		0.937398,
		0.93959,
		0.941754,
		0.943889,
		0.945995,
		0.948071,
		0.950115,
		0.952128,
		0.954108,
		0.956054,
		0.957965,
		0.959841,
		0.961681,
		0.963484,
		0.965249,
		0.966975,
		0.968661,
		0.970307,
		0.971912,
		0.973475,
		0.974996,
		0.976473,
		0.977906,
		0.979295,
		0.980638,
		0.981935,
		0.983187,
		0.984391,
		0.985548,
		0.986658,
		0.98772,
		0.988734,
		0.9897,
		0.990617,
		0.991486,
		0.992308,
		0.993081,
		0.993807,
		0.994485,
		0.995118,
		0.995704,
		0.996245,
		0.996742,
		0.997195,
		0.997607,
		0.997978,
		0.99831,
		0.998605,
		0.998863,
		0.999088,
		0.999281,
		0.999445,
		0.999581,
		0.999693,
		0.999782,
		0.999851,
		0.999903,
		0.999941,
		0.999967,
		0.999983,
		0.999993,
		0.999998,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999991,
		0.999978,
		0.999955,
		0.999916,
		0.999854,
		0.999765,
		0.999638,
		0.999465,
		0.999236,
		0.998938,
		0.99856,
		0.998086,
		0.997501,
		0.99679,
		0.995933,
		0.994912,
		0.993706,
		0.992293,
		0.99065,
		0.988753,
		0.986574,
		0.984089,
		0.981267,
		0.978079,
		0.974494,
		0.970479,
		0.965998,
		0.961014,
		0.955481,
		0.949345,
		0.942546,
		0.935005,
		0.926633,
		0.917318,
		0.906936,
		0.895339,
		0.882368,
		0.867848,
		0.8516,
		0.833445,
		0.81321,
		0.790745,
		0.765928,
		0.73868,
		0.708972,
		0.676837,
		0.642378,
		0.605771,
		0.567267,
		0.527188,
		0.485925,
		0.443925,
		0.401677,
		0.359697,
		0.318512,
		0.278635,
		0.240551,
		0.204699,
		0.171451,
		0.141108,
		0.113883,
		0.089903,
		0.069201,
		0.051724,
		0.037339,
		0.025838,
		0.016957,
		0.010384,
		0.005777,
		0.00278,
		0.001038,
		0.000214,
		0.870045,
		0.872571,
		0.875088,
		0.877597,
		0.880097,
		0.882588,
		0.88507,
		0.887541,
		0.890001,
		0.89245,
		0.894888,
		0.897313,
		0.899725,
		0.902123,
		0.904508,
		0.906878,
		0.909232,
		0.911571,
		0.913893,
		0.916199,
		0.918486,
		0.920755,
		0.923006,
		0.925236,
		0.927446,
		0.929636,
		0.931803,
		0.933948,
		0.93607,
		0.938168,
		0.940242,
		0.94229,
		0.944313,
		0.946308,
		0.948277,
		0.950217,
		0.952128,
		0.954009,
		0.955861,
		0.957681,
		0.959469,
		0.961225,
		0.962947,
		0.964635,
		0.966289,
		0.967907,
		0.969489,
		0.971035,
		0.972543,
		0.974012,
		0.975444,
		0.976835,
		0.978187,
		0.979499,
		0.98077,
		0.981999,
		0.983187,
		0.984332,
		0.985435,
		0.986495,
		0.987511,
		0.988485,
		0.989415,
		0.990301,
		0.991144,
		0.991944,
		0.9927,
		0.993413,
		0.994084,
		0.994712,
		0.995298,
		0.995843,
		0.996348,
		0.996812,
		0.997238,
		0.997627,
		0.997978,
		0.998294,
		0.998577,
		0.998827,
		0.999046,
		0.999236,
		0.999399,
		0.999536,
		0.999651,
		0.999744,
		0.999819,
		0.999876,
		0.99992,
		0.999951,
		0.999973,
		0.999986,
		0.999994,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999993,
		0.999982,
		0.999963,
		0.999931,
		0.999882,
		0.999809,
		0.999707,
		0.999567,
		0.999381,
		0.99914,
		0.998834,
		0.998451,
		0.997978,
		0.997403,
		0.996711,
		0.995886,
		0.994912,
		0.993771,
		0.992444,
		0.990912,
		0.989154,
		0.987147,
		0.984868,
		0.982294,
		0.9794,
		0.976158,
		0.972543,
		0.968525,
		0.964076,
		0.959167,
		0.953766,
		0.947842,
		0.941362,
		0.934294,
		0.926602,
		0.918247,
		0.909189,
		0.899376,
		0.888751,
		0.877246,
		0.864779,
		0.851254,
		0.836564,
		0.820585,
		0.803188,
		0.784235,
		0.76359,
		0.741127,
		0.716734,
		0.690328,
		0.661863,
		0.63134,
		0.598814,
		0.564404,
		0.528295,
		0.49074,
		0.45206,
		0.412633,
		0.372893,
		0.333305,
		0.294361,
		0.256554,
		0.220363,
		0.186232,
		0.154555,
		0.125657,
		0.099785,
		0.077097,
		0.057659,
		0.041445,
		0.02834,
		0.018148,
		0.010604,
		0.005387,
		0.002137,
		0.000471,
		0.900183,
		0.902465,
		0.904734,
		0.90699,
		0.909232,
		0.91146,
		0.913673,
		0.91587,
		0.918052,
		0.920217,
		0.922365,
		0.924495,
		0.926607,
		0.9287,
		0.930774,
		0.932827,
		0.93486,
		0.936872,
		0.938862,
		0.94083,
		0.942774,
		0.944695,
		0.946591,
		0.948463,
		0.950308,
		0.952128,
		0.953921,
		0.955686,
		0.957423,
		0.959131,
		0.96081,
		0.962458,
		0.964076,
		0.965663,
		0.967218,
		0.968741,
		0.97023,
		0.971686,
		0.973107,
		0.974494,
		0.975845,
		0.977161,
		0.97844,
		0.979683,
		0.980889,
		0.982057,
		0.983187,
		0.984278,
		0.985332,
		0.986346,
		0.987321,
		0.988257,
		0.989154,
		0.990011,
		0.990828,
		0.991607,
		0.992346,
		0.993045,
		0.993706,
		0.994328,
		0.994912,
		0.995458,
		0.995967,
		0.996439,
		0.996876,
		0.997277,
		0.997644,
		0.997978,
		0.99828,
		0.998551,
		0.998793,
		0.999006,
		0.999193,
		0.999355,
		0.999493,
		0.99961,
		0.999707,
		0.999785,
		0.999848,
		0.999897,
		0.999933,
		0.999959,
		0.999977,
		0.999989,
		0.999995,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999994,
		0.999986,
		0.99997,
		0.999944,
		0.999903,
		0.999844,
		0.99976,
		0.999645,
		0.999493,
		0.999296,
		0.999046,
		0.998733,
		0.998347,
		0.997877,
		0.997312,
		0.996638,
		0.995843,
		0.994912,
		0.993829,
		0.992579,
		0.991144,
		0.989507,
		0.987648,
		0.985548,
		0.983187,
		0.980542,
		0.977591,
		0.974313,
		0.970681,
		0.966673,
		0.962262,
		0.957423,
		0.952128,
		0.946351,
		0.940063,
		0.933236,
		0.925841,
		0.91785,
		0.909232,
		0.899959,
		0.890001,
		0.879328,
		0.867908,
		0.85571,
		0.842698,
		0.82883,
		0.81406,
		0.798332,
		0.781582,
		0.763735,
		0.744711,
		0.724419,
		0.702768,
		0.679669,
		0.655041,
		0.628819,
		0.600964,
		0.571468,
		0.540369,
		0.507751,
		0.473757,
		0.438588,
		0.402506,
		0.365832,
		0.328938,
		0.29224,
		0.256183,
		0.221226,
		0.187825,
		0.156413,
		0.127383,
		0.101068,
		0.077731,
		0.057546,
		0.040599,
		0.026875,
		0.01627,
		0.008587,
		0.00355,
		0.000818,
		0.925841,
		0.927846,
		0.929834,
		0.931803,
		0.933754,
		0.935686,
		0.937598,
		0.939491,
		0.941362,
		0.943213,
		0.945041,
		0.946848,
		0.948631,
		0.950392,
		0.952128,
		0.95384,
		0.955526,
		0.957188,
		0.958822,
		0.960431,
		0.962012,
		0.963565,
		0.96509,
		0.966586,
		0.968053,
		0.969489,
		0.970896,
		0.972271,
		0.973615,
		0.974928,
		0.976208,
		0.977455,
		0.978669,
		0.97985,
		0.980996,
		0.982109,
		0.983187,
		0.98423,
		0.985237,
		0.98621,
		0.987147,
		0.988048,
		0.988913,
		0.989742,
		0.990536,
		0.991293,
		0.992015,
		0.9927,
		0.99335,
		0.993965,
		0.994545,
		0.99509,
		0.995601,
		0.996078,
		0.996521,
		0.996932,
		0.997312,
		0.99766,
		0.997978,
		0.998267,
		0.998528,
		0.998762,
		0.99897,
		0.999153,
		0.999313,
		0.999452,
		0.99957,
		0.999669,
		0.999752,
		0.999819,
		0.999872,
		0.999913,
		0.999944,
		0.999966,
		0.999981,
		0.99999,
		0.999996,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999995,
		0.999988,
		0.999975,
		0.999953,
		0.99992,
		0.999871,
		0.999801,
		0.999707,
		0.999581,
		0.999419,
		0.999212,
		0.998953,
		0.998635,
		0.998248,
		0.997782,
		0.997227,
		0.996571,
		0.995804,
		0.994912,
		0.993882,
		0.9927,
		0.991352,
		0.989821,
		0.988091,
		0.986146,
		0.983968,
		0.981538,
		0.978837,
		0.975845,
		0.972543,
		0.968908,
		0.964919,
		0.960553,
		0.955789,
		0.950601,
		0.944967,
		0.938862,
		0.932261,
		0.92514,
		0.917472,
		0.909232,
		0.900395,
		0.890936,
		0.880827,
		0.870045,
		0.858565,
		0.846361,
		0.833411,
		0.819691,
		0.805178,
		0.789852,
		0.773688,
		0.756663,
		0.73875,
		0.719918,
		0.700132,
		0.679351,
		0.657532,
		0.634625,
		0.610585,
		0.585368,
		0.558942,
		0.531288,
		0.502414,
		0.472358,
		0.441194,
		0.409045,
		0.37608,
		0.342523,
		0.308649,
		0.274783,
		0.241294,
		0.208583,
		0.177071,
		0.147183,
		0.119331,
		0.093895,
		0.071204,
		0.051521,
		0.035026,
		0.021807,
		0.011854,
		0.005056,
		0.001204,
		0.947082,
		0.948785,
		0.950468,
		0.952128,
		0.953766,
		0.955381,
		0.956972,
		0.95854,
		0.960083,
		0.961602,
		0.963095,
		0.964563,
		0.966004,
		0.967419,
		0.968806,
		0.970166,
		0.971498,
		0.972801,
		0.974075,
		0.975321,
		0.976536,
		0.977722,
		0.978877,
		0.980001,
		0.981094,
		0.982156,
		0.983187,
		0.984185,
		0.985151,
		0.986085,
		0.986986,
		0.987855,
		0.988691,
		0.989494,
		0.990264,
		0.991001,
		0.991705,
		0.992377,
		0.993016,
		0.993622,
		0.994196,
		0.994738,
		0.995249,
		0.995728,
		0.996177,
		0.996595,
		0.996984,
		0.997343,
		0.997675,
		0.997978,
		0.998255,
		0.998506,
		0.998733,
		0.998935,
		0.999115,
		0.999274,
		0.999412,
		0.999531,
		0.999632,
		0.999718,
		0.999788,
		0.999846,
		0.999891,
		0.999926,
		0.999952,
		0.999971,
		0.999984,
		0.999992,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.99999,
		0.999979,
		0.999961,
		0.999933,
		0.999892,
		0.999834,
		0.999755,
		0.999651,
		0.999515,
		0.999343,
		0.999128,
		0.998863,
		0.998541,
		0.998153,
		0.997692,
		0.997147,
		0.996509,
		0.995768,
		0.994912,
		0.99393,
		0.992809,
		0.991538,
		0.990101,
		0.988485,
		0.986675,
		0.984657,
		0.982413,
		0.979927,
		0.977183,
		0.974162,
		0.970847,
		0.967218,
		0.963257,
		0.958943,
		0.954256,
		0.949176,
		0.943681,
		0.937751,
		0.931363,
		0.924495,
		0.917126,
		0.909232,
		0.900792,
		0.891784,
		0.882183,
		0.87197,
		0.861121,
		0.849616,
		0.837433,
		0.824552,
		0.810953,
		0.796619,
		0.781532,
		0.765677,
		0.749038,
		0.731605,
		0.713365,
		0.694312,
		0.674436,
		0.653732,
		0.632191,
		0.609807,
		0.586569,
		0.56247,
		0.537502,
		0.51166,
		0.484947,
		0.457378,
		0.428982,
		0.399815,
		0.369959,
		0.339533,
		0.308696,
		0.277651,
		0.246645,
		0.215973,
		0.185967,
		0.156993,
		0.12944,
		0.103703,
		0.080172,
		0.05921,
		0.041135,
		0.026203,
		0.014591,
		0.006383,
		0.001561,
		0.964076,
		0.965467,
		0.966832,
		0.968173,
		0.969489,
		0.97078,
		0.972044,
		0.973282,
		0.974494,
		0.975678,
		0.976835,
		0.977965,
		0.979066,
		0.98014,
		0.981184,
		0.9822,
		0.983187,
		0.984144,
		0.985072,
		0.98597,
		0.986838,
		0.987677,
		0.988485,
		0.989263,
		0.990011,
		0.990728,
		0.991416,
		0.992073,
		0.9927,
		0.993298,
		0.993865,
		0.994403,
		0.994912,
		0.995392,
		0.995843,
		0.996266,
		0.996662,
		0.99703,
		0.997372,
		0.997688,
		0.997978,
		0.998244,
		0.998486,
		0.998706,
		0.998903,
		0.999079,
		0.999236,
		0.999373,
		0.999493,
		0.999596,
		0.999684,
		0.999758,
		0.999819,
		0.999868,
		0.999907,
		0.999937,
		0.999959,
		0.999975,
		0.999986,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999992,
		0.999982,
		0.999967,
		0.999944,
		0.999909,
		0.999861,
		0.999794,
		0.999707,
		0.999593,
		0.999448,
		0.999268,
		0.999046,
		0.998776,
		0.998451,
		0.998064,
		0.997607,
		0.997073,
		0.996452,
		0.995735,
		0.994912,
		0.993974,
		0.992908,
		0.991705,
		0.990352,
		0.988837,
		0.987147,
		0.985268,
		0.983187,
		0.980889,
		0.978359,
		0.975582,
		0.972543,
		0.969224,
		0.965609,
		0.961681,
		0.957423,
		0.952816,
		0.947842,
		0.942482,
		0.936718,
		0.930531,
		0.9239,
		0.916807,
		0.909232,
		0.901155,
		0.892557,
		0.883417,
		0.873716,
		0.863435,
		0.852555,
		0.841057,
		0.828924,
		0.816138,
		0.802682,
		0.788542,
		0.773702,
		0.758149,
		0.741872,
		0.72486,
		0.707107,
		0.688606,
		0.669354,
		0.649351,
		0.628601,
		0.607109,
		0.584885,
		0.561943,
		0.538297,
		0.513969,
		0.488978,
		0.463351,
		0.437117,
		0.410314,
		0.382989,
		0.355202,
		0.327031,
		0.298578,
		0.269972,
		0.241376,
		0.212988,
		0.185047,
		0.157829,
		0.131646,
		0.106843,
		0.083785,
		0.062848,
		0.044403,
		0.028801,
		0.01635,
		0.007301,
		0.001825,
		0.977109,
		0.978187,
		0.97924,
		0.980266,
		0.981267,
		0.98224,
		0.983187,
		0.984106,
		0.984999,
		0.985864,
		0.986701,
		0.987511,
		0.988294,
		0.989048,
		0.989775,
		0.990474,
		0.991144,
		0.991788,
		0.992403,
		0.992991,
		0.993551,
		0.994084,
		0.99459,
		0.995069,
		0.995521,
		0.995947,
		0.996348,
		0.996723,
		0.997073,
		0.997398,
		0.9977,
		0.997978,
		0.998234,
		0.998468,
		0.998681,
		0.998873,
		0.999046,
		0.9992,
		0.999337,
		0.999457,
		0.999561,
		0.999651,
		0.999727,
		0.999791,
		0.999844,
		0.999886,
		0.99992,
		0.999946,
		0.999965,
		0.999979,
		0.999988,
		0.999994,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999985,
		0.999972,
		0.999952,
		0.999923,
		0.999882,
		0.999826,
		0.999752,
		0.999656,
		0.999533,
		0.999381,
		0.999193,
		0.998965,
		0.99869,
		0.998364,
		0.997978,
		0.997527,
		0.997003,
		0.996398,
		0.995704,
		0.994912,
		0.994013,
		0.992999,
		0.991857,
		0.990579,
		0.989154,
		0.987569,
		0.985814,
		0.983876,
		0.981742,
		0.9794,
		0.976835,
		0.974035,
		0.970985,
		0.967671,
		0.964076,
		0.960187,
		0.955987,
		0.951461,
		0.946591,
		0.941362,
		0.935757,
		0.929759,
		0.92335,
		0.916514,
		0.909232,
		0.901488,
		0.893264,
		0.884543,
		0.875306,
		0.865538,
		0.855222,
		0.84434,
		0.832877,
		0.820817,
		0.808146,
		0.794849,
		0.780913,
		0.766326,
		0.751077,
		0.735156,
		0.718555,
		0.701268,
		0.683291,
		0.664622,
		0.645261,
		0.625212,
		0.604481,
		0.583079,
		0.561021,
		0.538323,
		0.515011,
		0.491113,
		0.466662,
		0.441699,
		0.416269,
		0.390425,
		0.364225,
		0.337737,
		0.311037,
		0.284216,
		0.257379,
		0.230652,
		0.204182,
		0.178148,
		0.152754,
		0.128242,
		0.104887,
		0.082994,
		0.062899,
		0.04496,
		0.029542,
		0.017013,
		0.007716,
		0.001962,
		0.986574,
		0.987358,
		0.988116,
		0.988848,
		0.989554,
		0.990235,
		0.99089,
		0.991519,
		0.992122,
		0.9927,
		0.993253,
		0.99378,
		0.994282,
		0.994759,
		0.995211,
		0.995638,
		0.996042,
		0.996422,
		0.996778,
		0.997111,
		0.997422,
		0.997711,
		0.997978,
		0.998225,
		0.998451,
		0.998657,
		0.998845,
		0.999014,
		0.999166,
		0.999302,
		0.999422,
		0.999527,
		0.999618,
		0.999696,
		0.999763,
		0.999819,
		0.999864,
		0.999901,
		0.999931,
		0.999953,
		0.99997,
		0.999982,
		0.99999,
		0.999995,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999994,
		0.999987,
		0.999976,
		0.999959,
		0.999934,
		0.999899,
		0.999852,
		0.999788,
		0.999707,
		0.999603,
		0.999473,
		0.999313,
		0.999119,
		0.998886,
		0.998608,
		0.99828,
		0.997897,
		0.997451,
		0.996937,
		0.996348,
		0.995675,
		0.994912,
		0.99405,
		0.993081,
		0.991996,
		0.990785,
		0.98944,
		0.987949,
		0.986304,
		0.984492,
		0.982503,
		0.980326,
		0.977948,
		0.975358,
		0.972543,
		0.969489,
		0.966185,
		0.962616,
		0.958769,
		0.95463,
		0.950184,
		0.945417,
		0.940314,
		0.93486,
		0.929041,
		0.92284,
		0.916242,
		0.909232,
		0.901795,
		0.893914,
		0.885575,
		0.876762,
		0.867459,
		0.857652,
		0.847326,
		0.836467,
		0.82506,
		0.813093,
		0.800552,
		0.787425,
		0.773702,
		0.759371,
		0.744423,
		0.728851,
		0.712649,
		0.69581,
		0.678333,
		0.660215,
		0.641458,
		0.622067,
		0.602046,
		0.581406,
		0.560159,
		0.538323,
		0.515919,
		0.492972,
		0.469512,
		0.445578,
		0.42121,
		0.39646,
		0.371384,
		0.346046,
		0.320521,
		0.294891,
		0.269249,
		0.243699,
		0.218359,
		0.19336,
		0.168853,
		0.14501,
		0.122028,
		0.100129,
		0.079568,
		0.060626,
		0.043618,
		0.028884,
		0.016784,
		0.007692,
		0.001978,
		0.992969,
		0.99349,
		0.993987,
		0.994461,
		0.994912,
		0.99534,
		0.995745,
		0.996128,
		0.99649,
		0.996829,
		0.997147,
		0.997444,
		0.997721,
		0.997978,
		0.998216,
		0.998435,
		0.998635,
		0.998818,
		0.998984,
		0.999134,
		0.999268,
		0.999388,
		0.999493,
		0.999586,
		0.999666,
		0.999735,
		0.999793,
		0.999842,
		0.999882,
		0.999914,
		0.99994,
		0.999959,
		0.999974,
		0.999984,
		0.999991,
		0.999996,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999995,
		0.999989,
		0.99998,
		0.999965,
		0.999944,
		0.999914,
		0.999873,
		0.999819,
		0.999749,
		0.99966,
		0.999548,
		0.999412,
		0.999246,
		0.999046,
		0.998808,
		0.998528,
		0.9982,
		0.997819,
		0.99738,
		0.996876,
		0.996301,
		0.995649,
		0.994912,
		0.994084,
		0.993157,
		0.992122,
		0.990973,
		0.9897,
		0.988294,
		0.986746,
		0.985047,
		0.983187,
		0.981155,
		0.978942,
		0.976536,
		0.973927,
		0.971103,
		0.968053,
		0.964764,
		0.961225,
		0.957423,
		0.953345,
		0.94898,
		0.944313,
		0.939331,
		0.934022,
		0.928371,
		0.922365,
		0.91599,
		0.909232,
		0.902078,
		0.894513,
		0.886525,
		0.878098,
		0.869219,
		0.859876,
		0.850054,
		0.839741,
		0.828924,
		0.817592,
		0.805731,
		0.793332,
		0.780384,
		0.766878,
		0.752804,
		0.738155,
		0.722924,
		0.707107,
		0.690698,
		0.673697,
		0.656102,
		0.637915,
		0.61914,
		0.599783,
		0.579853,
		0.559361,
		0.538323,
		0.516759,
		0.494689,
		0.472143,
		0.449152,
		0.425753,
		0.40199,
		0.377913,
		0.353578,
		0.32905,
		0.304401,
		0.279714,
		0.255081,
		0.230604,
		0.2064,
		0.182595,
		0.159331,
		0.136765,
		0.115075,
		0.094456,
		0.075127,
		0.057334,
		0.041348,
		0.027471,
		0.016032,
		0.007387,
		0.001912,
		0.996876,
		0.99718,
		0.997465,
		0.997731,
		0.997978,
		0.998208,
		0.99842,
		0.998614,
		0.998793,
		0.998956,
		0.999103,
		0.999236,
		0.999355,
		0.999461,
		0.999554,
		0.999636,
		0.999707,
		0.999767,
		0.999819,
		0.999861,
		0.999897,
		0.999925,
		0.999947,
		0.999965,
		0.999977,
		0.999986,
		0.999992,
		0.999996,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999996,
		0.999991,
		0.999982,
		0.99997,
		0.999951,
		0.999926,
		0.99989,
		0.999844,
		0.999783,
		0.999707,
		0.999611,
		0.999493,
		0.99935,
		0.999178,
		0.998974,
		0.998733,
		0.998451,
		0.998123,
		0.997745,
		0.997312,
		0.996818,
		0.996257,
		0.995624,
		0.994912,
		0.994115,
		0.993226,
		0.992239,
		0.991144,
		0.989936,
		0.988607,
		0.987147,
		0.985548,
		0.983803,
		0.981901,
		0.979834,
		0.977591,
		0.975164,
		0.972543,
		0.969716,
		0.966673,
		0.963404,
		0.959898,
		0.956143,
		0.952128,
		0.947842,
		0.943272,
		0.938407,
		0.933236,
		0.927745,
		0.921922,
		0.915755,
		0.909232,
		0.902341,
		0.895068,
		0.887401,
		0.879329,
		0.870838,
		0.861918,
		0.852555,
		0.842738,
		0.832457,
		0.821699,
		0.810454,
		0.798713,
		0.786465,
		0.773702,
		0.760415,
		0.746597,
		0.732241,
		0.717343,
		0.701897,
		0.685901,
		0.669354,
		0.652255,
		0.634605,
		0.61641,
		0.597675,
		0.578407,
		0.558619,
		0.538323,
		0.517538,
		0.496282,
		0.474581,
		0.452462,
		0.42996,
		0.407111,
		0.383959,
		0.360554,
		0.336951,
		0.313213,
		0.289411,
		0.265625,
		0.241943,
		0.218463,
		0.195295,
		0.172561,
		0.150396,
		0.128948,
		0.108382,
		0.088879,
		0.07064,
		0.053884,
		0.038856,
		0.025825,
		0.015085,
		0.006962,
		0.001807,
		0.998929,
		0.999074,
		0.999205,
		0.999323,
		0.999429,
		0.999523,
		0.999606,
		0.999679,
		0.999741,
		0.999795,
		0.99984,
		0.999878,
		0.999909,
		0.999934,
		0.999954,
		0.999969,
		0.99998,
		0.999988,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999996,
		0.999992,
		0.999985,
		0.999974,
		0.999958,
		0.999935,
		0.999905,
		0.999864,
		0.999812,
		0.999746,
		0.999663,
		0.999561,
		0.999437,
		0.999289,
		0.999112,
		0.998903,
		0.998659,
		0.998376,
		0.998049,
		0.997675,
		0.997247,
		0.996763,
		0.996216,
		0.995601,
		0.994912,
		0.994144,
		0.993291,
		0.992346,
		0.991302,
		0.990153,
		0.988892,
		0.987511,
		0.986004,
		0.984361,
		0.982575,
		0.980638,
		0.978541,
		0.976276,
		0.973833,
		0.971204,
		0.968379,
		0.965348,
		0.962103,
		0.958631,
		0.954925,
		0.950973,
		0.946765,
		0.94229,
		0.937538,
		0.932498,
		0.927158,
		0.921508,
		0.915536,
		0.909232,
		0.902584,
		0.895582,
		0.888213,
		0.880467,
		0.872333,
		0.863799,
		0.854856,
		0.845492,
		0.835698,
		0.825463,
		0.814777,
		0.803632,
		0.792019,
		0.779928,
		0.767353,
		0.754286,
		0.740722,
		0.726654,
		0.712079,
		0.696992,
		0.681392,
		0.665277,
		0.648648,
		0.631507,
		0.613858,
		0.595706,
		0.577059,
		0.557927,
		0.538323,
		0.518262,
		0.497762,
		0.476846,
		0.455537,
		0.433865,
		0.411865,
		0.389573,
		0.367032,
		0.344293,
		0.321409,
		0.298442,
		0.27546,
		0.25254,
		0.229766,
		0.207232,
		0.185043,
		0.163312,
		0.142168,
		0.121749,
		0.102209,
		0.083718,
		0.06646,
		0.050639,
		0.036478,
		0.024222,
		0.014139,
		0.006522,
		0.001693,
		0.999771,
		0.999819,
		0.999859,
		0.999892,
		0.99992,
		0.999942,
		0.999959,
		0.999973,
		0.999982,
		0.999989,
		0.999994,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999987,
		0.999977,
		0.999963,
		0.999944,
		0.999917,
		0.999882,
		0.999836,
		0.999779,
		0.999707,
		0.999618,
		0.99951,
		0.999381,
		0.999227,
		0.999046,
		0.998834,
		0.998588,
		0.998304,
		0.997978,
		0.997607,
		0.997186,
		0.996711,
		0.996177,
		0.995579,
		0.994912,
		0.994171,
		0.99335,
		0.992444,
		0.991447,
		0.990352,
		0.989154,
		0.987845,
		0.986419,
		0.984868,
		0.983187,
		0.981366,
		0.9794,
		0.977279,
		0.974996,
		0.972543,
		0.969911,
		0.967092,
		0.964076,
		0.960857,
		0.957423,
		0.953766,
		0.949876,
		0.945745,
		0.941362,
		0.936718,
		0.931803,
		0.926607,
		0.92112,
		0.915332,
		0.909232,
		0.902812,
		0.89606,
		0.888967,
		0.881522,
		0.873716,
		0.865538,
		0.85698,
		0.848031,
		0.838682,
		0.828924,
		0.818748,
		0.808146,
		0.797109,
		0.78563,
		0.773702,
		0.761317,
		0.74847,
		0.735156,
		0.721369,
		0.707107,
		0.692366,
		0.677145,
		0.661443,
		0.645261,
		0.628601,
		0.611467,
		0.593863,
		0.575799,
		0.557281,
		0.538323,
		0.518938,
		0.499142,
		0.478955,
		0.4584,
		0.437501,
		0.416289,
		0.394798,
		0.373065,
		0.351133,
		0.32905,
		0.306869,
		0.28465,
		0.262459,
		0.24037,
		0.218463,
		0.196828,
		0.175563,
		0.154777,
		0.134589,
		0.11513,
		0.096544,
		0.078987,
		0.062632,
		0.047667,
		0.034297,
		0.022747,
		0.013263,
		0.006111,
		0.001584,
		0.999985,
		0.999991,
		0.999995,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999994,
		0.999988,
		0.99998,
		0.999968,
		0.999951,
		0.999927,
		0.999897,
		0.999857,
		0.999806,
		0.999743,
		0.999666,
		0.999572,
		0.999459,
		0.999324,
		0.999166,
		0.998981,
		0.998766,
		0.998518,
		0.998234,
		0.99791,
		0.997543,
		0.997128,
		0.996662,
		0.99614,
		0.995558,
		0.994912,
		0.994196,
		0.993406,
		0.992536,
		0.991581,
		0.990536,
		0.989394,
		0.98815,
		0.986798,
		0.985332,
		0.983744,
		0.982029,
		0.980179,
		0.978187,
		0.976047,
		0.973751,
		0.971292,
		0.968661,
		0.965851,
		0.962855,
		0.959663,
		0.956268,
		0.952661,
		0.948834,
		0.944778,
		0.940484,
		0.935944,
		0.931149,
		0.926089,
		0.920755,
		0.91514,
		0.909232,
		0.903024,
		0.896506,
		0.889668,
		0.882503,
		0.875,
		0.867151,
		0.858946,
		0.850378,
		0.841438,
		0.832117,
		0.822408,
		0.812301,
		0.801791,
		0.790869,
		0.77953,
		0.767766,
		0.755573,
		0.742944,
		0.729875,
		0.716363,
		0.702404,
		0.687996,
		0.673139,
		0.657831,
		0.642073,
		0.625869,
		0.609221,
		0.592135,
		0.574618,
		0.556677,
		0.538323,
		0.51957,
		0.500431,
		0.480925,
		0.461072,
		0.440894,
		0.420418,
		0.399673,
		0.378694,
		0.357518,
		0.336187,
		0.314747,
		0.293251,
		0.271757,
		0.250328,
		0.229035,
		0.207954,
		0.187172,
		0.166781,
		0.146884,
		0.127594,
		0.109032,
		0.091333,
		0.074644,
		0.059124,
		0.044948,
		0.032305,
		0.021402,
		0.012464,
		0.005737,
		0.001486,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999995,
		0.99999,
		0.999982,
		0.999972,
		0.999957,
		0.999936,
		0.999909,
		0.999874,
		0.99983,
		0.999775,
		0.999707,
		0.999624,
		0.999525,
		0.999407,
		0.999268,
		0.999106,
		0.998917,
		0.9987,
		0.998451,
		0.998167,
		0.997845,
		0.997481,
		0.997073,
		0.996615,
		0.996106,
		0.995539,
		0.994912,
		0.99422,
		0.993458,
		0.992621,
		0.991705,
		0.990705,
		0.989616,
		0.988431,
		0.987147,
		0.985756,
		0.984253,
		0.982633,
		0.980889,
		0.979014,
		0.977002,
		0.974848,
		0.972543,
		0.970081,
		0.967454,
		0.964657,
		0.961681,
		0.958519,
		0.955164,
		0.951607,
		0.947842,
		0.943859,
		0.939652,
		0.935212,
		0.930531,
		0.925601,
		0.920413,
		0.91496,
		0.909232,
		0.903223,
		0.896922,
		0.890323,
		0.883417,
		0.876195,
		0.868649,
		0.860772,
		0.852555,
		0.843991,
		0.835071,
		0.82579,
		0.816138,
		0.80611,
		0.795699,
		0.784898,
		0.773702,
		0.762104,
		0.750101,
		0.737688,
		0.72486,
		0.711615,
		0.69795,
		0.683863,
		0.669354,
		0.654422,
		0.639069,
		0.623297,
		0.607109,
		0.590511,
		0.573509,
		0.556109,
		0.538323,
		0.520162,
		0.501639,
		0.482769,
		0.463572,
		0.444067,
		0.424278,
		0.404232,
		0.383959,
		0.363491,
		0.342867,
		0.322126,
		0.301315,
		0.280485,
		0.259691,
		0.238994,
		0.218463,
		0.19817,
		0.178197,
		0.158632,
		0.139573,
		0.121124,
		0.103402,
		0.086531,
		0.070648,
		0.055902,
		0.042454,
		0.030481,
		0.020173,
		0.011736,
		0.005396,
		0.001396,
		0.999985,
		0.999975,
		0.999962,
		0.999944,
		0.99992,
		0.999889,
		0.99985,
		0.999801,
		0.999741,
		0.999668,
		0.999581,
		0.999477,
		0.999355,
		0.999212,
		0.999046,
		0.998854,
		0.998635,
		0.998385,
		0.998102,
		0.997782,
		0.997422,
		0.99702,
		0.996571,
		0.996073,
		0.995521,
		0.994912,
		0.994242,
		0.993506,
		0.9927,
		0.991821,
		0.990862,
		0.989821,
		0.988691,
		0.987468,
		0.986146,
		0.984721,
		0.983187,
		0.981538,
		0.979768,
		0.977873,
		0.975845,
		0.973679,
		0.971369,
		0.968908,
		0.966289,
		0.963506,
		0.960553,
		0.957423,
		0.954108,
		0.950601,
		0.946896,
		0.942986,
		0.938862,
		0.934518,
		0.929947,
		0.92514,
		0.92009,
		0.91479,
		0.909232,
		0.903409,
		0.897313,
		0.890936,
		0.88427,
		0.877309,
		0.870045,
		0.862471,
		0.854579,
		0.846361,
		0.837812,
		0.828924,
		0.819691,
		0.810106,
		0.800163,
		0.789856,
		0.77918,
		0.768129,
		0.756699,
		0.744885,
		0.732684,
		0.720092,
		0.707107,
		0.693726,
		0.679948,
		0.665773,
		0.651201,
		0.636233,
		0.620871,
		0.605119,
		0.588982,
		0.572465,
		0.555576,
		0.538323,
		0.520718,
		0.502771,
		0.484498,
		0.465916,
		0.447041,
		0.427896,
		0.408505,
		0.388893,
		0.369091,
		0.349131,
		0.32905,
		0.308888,
		0.28869,
		0.268504,
		0.248386,
		0.228392,
		0.208589,
		0.189047,
		0.169842,
		0.151058,
		0.132788,
		0.11513,
		0.098193,
		0.082095,
		0.066963,
		0.052935,
		0.040162,
		0.028807,
		0.019046,
		0.011069,
		0.005084,
		0.001314,
		0.999771,
		0.999707,
		0.999629,
		0.999537,
		0.999429,
		0.999303,
		0.999156,
		0.998987,
		0.998793,
		0.998572,
		0.998321,
		0.998039,
		0.997721,
		0.997366,
		0.99697,
		0.996529,
		0.996042,
		0.995504,
		0.994912,
		0.994262,
		0.993551,
		0.992774,
		0.991928,
		0.991008,
		0.990011,
		0.988931,
		0.987764,
		0.986506,
		0.985151,
		0.983695,
		0.982133,
		0.980459,
		0.978669,
		0.976757,
		0.974716,
		0.972543,
		0.97023,
		0.967772,
		0.965164,
		0.962398,
		0.959469,
		0.956371,
		0.953097,
		0.94964,
		0.945995,
		0.942154,
		0.938112,
		0.93386,
		0.929393,
		0.924704,
		0.919785,
		0.91463,
		0.909232,
		0.903584,
		0.897679,
		0.89151,
		0.88507,
		0.878352,
		0.87135,
		0.864056,
		0.856465,
		0.848568,
		0.840361,
		0.831837,
		0.82299,
		0.813813,
		0.804301,
		0.794448,
		0.78425,
		0.773702,
		0.762798,
		0.751535,
		0.739908,
		0.727916,
		0.715554,
		0.702821,
		0.689715,
		0.676234,
		0.66238,
		0.648151,
		0.63355,
		0.618579,
		0.603241,
		0.58754,
		0.571482,
		0.555074,
		0.538323,
		0.52124,
		0.503836,
		0.486123,
		0.468117,
		0.449834,
		0.431293,
		0.412516,
		0.393526,
		0.374349,
		0.355015,
		0.335557,
		0.31601,
		0.296414,
		0.276811,
		0.25725,
		0.237781,
		0.218463,
		0.199356,
		0.180527,
		0.162051,
		0.144007,
		0.12648,
		0.109566,
		0.093366,
		0.07799,
		0.063557,
		0.050198,
		0.038051,
		0.027267,
		0.018011,
		0.010458,
		0.004799,
		0.001239,
		0.998929,
		0.998733,
		0.998511,
		0.99826,
		0.997978,
		0.997663,
		0.997312,
		0.996922,
		0.99649,
		0.996013,
		0.995488,
		0.994912,
		0.994282,
		0.993593,
		0.992844,
		0.992029,
		0.991144,
		0.990188,
		0.989154,
		0.988039,
		0.986838,
		0.985548,
		0.984164,
		0.982681,
		0.981094,
		0.9794,
		0.977591,
		0.975665,
		0.973615,
		0.971437,
		0.969125,
		0.966673,
		0.964076,
		0.961329,
		0.958426,
		0.955361,
		0.952128,
		0.948721,
		0.945135,
		0.941362,
		0.937398,
		0.933236,
		0.928869,
		0.924291,
		0.919497,
		0.914479,
		0.909232,
		0.903749,
		0.898023,
		0.892049,
		0.88582,
		0.879329,
		0.872571,
		0.865538,
		0.858226,
		0.850628,
		0.842738,
		0.834551,
		0.82606,
		0.81726,
		0.808146,
		0.798713,
		0.788955,
		0.778869,
		0.76845,
		0.757694,
		0.746597,
		0.735156,
		0.723368,
		0.71123,
		0.698742,
		0.685901,
		0.672708,
		0.65916,
		0.645261,
		0.63101,
		0.61641,
		0.601465,
		0.586177,
		0.570554,
		0.5546,
		0.538323,
		0.521733,
		0.504839,
		0.487653,
		0.470189,
		0.452462,
		0.434489,
		0.416289,
		0.397884,
		0.379296,
		0.360554,
		0.341684,
		0.32272,
		0.303696,
		0.28465,
		0.265625,
		0.246666,
		0.227824,
		0.209152,
		0.19071,
		0.172561,
		0.154777,
		0.137432,
		0.120607,
		0.104393,
		0.088884,
		0.074184,
		0.060404,
		0.047667,
		0.036101,
		0.025848,
		0.017058,
		0.009896,
		0.004537,
		0.00117,
		0.996876,
		0.996452,
		0.995985,
		0.995473,
		0.994912,
		0.9943,
		0.993633,
		0.992908,
		0.992122,
		0.991272,
		0.990352,
		0.989361,
		0.988294,
		0.987147,
		0.985916,
		0.984597,
		0.983187,
		0.98168,
		0.980072,
		0.978359,
		0.976536,
		0.974599,
		0.972543,
		0.970362,
		0.968053,
		0.965609,
		0.963026,
		0.960299,
		0.957423,
		0.954391,
		0.951199,
		0.947842,
		0.944313,
		0.940607,
		0.936718,
		0.932642,
		0.928371,
		0.9239,
		0.919224,
		0.914337,
		0.909232,
		0.903905,
		0.898348,
		0.892557,
		0.886525,
		0.880246,
		0.873716,
		0.866927,
		0.859876,
		0.852555,
		0.84496,
		0.837084,
		0.828924,
		0.820474,
		0.811728,
		0.802682,
		0.793332,
		0.783673,
		0.773702,
		0.763413,
		0.752804,
		0.741872,
		0.730613,
		0.719025,
		0.707107,
		0.694856,
		0.682272,
		0.669354,
		0.656102,
		0.642517,
		0.628601,
		0.614355,
		0.599783,
		0.584888,
		0.569676,
		0.554152,
		0.538323,
		0.522198,
		0.505785,
		0.489096,
		0.472143,
		0.45494,
		0.437501,
		0.419845,
		0.40199,
		0.383959,
		0.365774,
		0.347461,
		0.32905,
		0.31057,
		0.292057,
		0.273547,
		0.255081,
		0.236704,
		0.218463,
		0.200411,
		0.182603,
		0.165103,
		0.147975,
		0.131291,
		0.11513,
		0.099574,
		0.084714,
		0.070648,
		0.057479,
		0.045322,
		0.034297,
		0.024536,
		0.016179,
		0.009378,
		0.004296,
		0.001107,
		0.992969,
		0.99221,
		0.99139,
		0.990506,
		0.989554,
		0.988531,
		0.987433,
		0.986257,
		0.984999,
		0.983655,
		0.98222,
		0.980692,
		0.979066,
		0.977338,
		0.975504,
		0.973558,
		0.971498,
		0.969317,
		0.967013,
		0.964579,
		0.962012,
		0.959306,
		0.956457,
		0.953459,
		0.950308,
		0.946999,
		0.943527,
		0.939885,
		0.93607,
		0.932076,
		0.927897,
		0.923529,
		0.918966,
		0.914202,
		0.909232,
		0.904051,
		0.898654,
		0.893035,
		0.887188,
		0.881109,
		0.874792,
		0.868232,
		0.861423,
		0.854361,
		0.84704,
		0.839455,
		0.831602,
		0.823476,
		0.815072,
		0.806386,
		0.797413,
		0.78815,
		0.778592,
		0.768736,
		0.758579,
		0.748116,
		0.737347,
		0.726267,
		0.714875,
		0.70317,
		0.69115,
		0.678813,
		0.666161,
		0.653193,
		0.63991,
		0.626313,
		0.612405,
		0.598188,
		0.583667,
		0.568845,
		0.553728,
		0.538323,
		0.522638,
		0.50668,
		0.49046,
		0.473988,
		0.457278,
		0.440344,
		0.423201,
		0.405867,
		0.38836,
		0.370703,
		0.352917,
		0.33503,
		0.317068,
		0.299063,
		0.281048,
		0.263058,
		0.245133,
		0.227317,
		0.209654,
		0.192195,
		0.174994,
		0.15811,
		0.141606,
		0.12555,
		0.110014,
		0.09508,
		0.08083,
		0.067358,
		0.054761,
		0.043145,
		0.032625,
		0.023321,
		0.015366,
		0.0089,
		0.004073,
		0.001049,
		0.986574,
		0.985372,
		0.984089,
		0.982721,
		0.981267,
		0.97972,
		0.978079,
		0.976338,
		0.974494,
		0.972543,
		0.97048,
		0.968302,
		0.966004,
		0.963582,
		0.961031,
		0.958348,
		0.955526,
		0.952563,
		0.949453,
		0.946192,
		0.942774,
		0.939196,
		0.935451,
		0.931537,
		0.927446,
		0.923176,
		0.91872,
		0.914074,
		0.909232,
		0.904191,
		0.898944,
		0.893487,
		0.887815,
		0.881923,
		0.875806,
		0.869459,
		0.862877,
		0.856057,
		0.848992,
		0.841678,
		0.834112,
		0.826287,
		0.818201,
		0.809849,
		0.801227,
		0.79233,
		0.783156,
		0.773702,
		0.763962,
		0.753936,
		0.743619,
		0.73301,
		0.722107,
		0.710906,
		0.699408,
		0.687611,
		0.675515,
		0.663118,
		0.650423,
		0.637429,
		0.624138,
		0.610552,
		0.596674,
		0.582507,
		0.568056,
		0.553326,
		0.538323,
		0.523054,
		0.507526,
		0.491749,
		0.475734,
		0.45949,
		0.443032,
		0.426374,
		0.409531,
		0.392521,
		0.375363,
		0.358077,
		0.340688,
		0.32322,
		0.3057,
		0.288158,
		0.270627,
		0.253142,
		0.23574,
		0.218463,
		0.201355,
		0.184465,
		0.167843,
		0.151546,
		0.135635,
		0.120172,
		0.105229,
		0.09088,
		0.077206,
		0.064291,
		0.05223,
		0.041121,
		0.031071,
		0.022194,
		0.014613,
		0.008457,
		0.003868,
		0.000995,
		0.977109,
		0.975358,
		0.973507,
		0.971552,
		0.969489,
		0.967315,
		0.965026,
		0.962616,
		0.960083,
		0.957423,
		0.95463,
		0.951701,
		0.948631,
		0.945417,
		0.942053,
		0.938536,
		0.93486,
		0.931022,
		0.927017,
		0.92284,
		0.918486,
		0.913952,
		0.909232,
		0.904322,
		0.899218,
		0.893914,
		0.888407,
		0.88269,
		0.876762,
		0.870615,
		0.864247,
		0.857652,
		0.850827,
		0.843766,
		0.836467,
		0.828924,
		0.821134,
		0.813093,
		0.804797,
		0.796242,
		0.787425,
		0.778343,
		0.768993,
		0.759371,
		0.749475,
		0.739302,
		0.728851,
		0.71812,
		0.707107,
		0.69581,
		0.68423,
		0.672365,
		0.660215,
		0.647781,
		0.635065,
		0.622067,
		0.608789,
		0.595234,
		0.581406,
		0.567308,
		0.552945,
		0.538323,
		0.523449,
		0.508329,
		0.492972,
		0.477387,
		0.461585,
		0.445578,
		0.429378,
		0.413,
		0.39646,
		0.379775,
		0.362964,
		0.346048,
		0.32905,
		0.311993,
		0.294906,
		0.277816,
		0.260756,
		0.243759,
		0.226862,
		0.210104,
		0.193529,
		0.177183,
		0.161115,
		0.145377,
		0.130029,
		0.11513,
		0.100747,
		0.086952,
		0.073818,
		0.061429,
		0.049871,
		0.039236,
		0.029626,
		0.021147,
		0.013913,
		0.008046,
		0.003677,
		0.000945,
		0.964076,
		0.961681,
		0.959167,
		0.95653,
		0.953766,
		0.950871,
		0.947842,
		0.944673,
		0.941362,
		0.937904,
		0.934295,
		0.930531,
		0.926607,
		0.922519,
		0.918264,
		0.913836,
		0.909232,
		0.904447,
		0.899478,
		0.894319,
		0.888967,
		0.883417,
		0.877665,
		0.871707,
		0.865538,
		0.859156,
		0.852555,
		0.845732,
		0.838682,
		0.831403,
		0.823889,
		0.816138,
		0.808146,
		0.79991,
		0.791425,
		0.78269,
		0.773702,
		0.764456,
		0.754952,
		0.745185,
		0.735156,
		0.72486,
		0.714298,
		0.703467,
		0.692366,
		0.680995,
		0.669354,
		0.657442,
		0.645261,
		0.63281,
		0.620093,
		0.607109,
		0.593863,
		0.580358,
		0.566596,
		0.552583,
		0.538323,
		0.523824,
		0.50909,
		0.494131,
		0.478955,
		0.463572,
		0.447991,
		0.432226,
		0.416289,
		0.400195,
		0.383959,
		0.367599,
		0.351133,
		0.334582,
		0.317968,
		0.301315,
		0.28465,
		0.268,
		0.251397,
		0.234873,
		0.218463,
		0.202206,
		0.186142,
		0.170317,
		0.154777,
		0.139573,
		0.12476,
		0.110396,
		0.096544,
		0.08327,
		0.070648,
		0.058753,
		0.047667,
		0.037478,
		0.02828,
		0.020173,
		0.013263,
		0.007665,
		0.003501,
		0.000899,
		0.947082,
		0.943959,
		0.940699,
		0.937299,
		0.933754,
		0.930061,
		0.926215,
		0.922214,
		0.918052,
		0.913726,
		0.909232,
		0.904566,
		0.899725,
		0.894703,
		0.889497,
		0.884104,
		0.878519,
		0.872739,
		0.866759,
		0.860575,
		0.854185,
		0.847584,
		0.840769,
		0.833736,
		0.826481,
		0.819002,
		0.811294,
		0.803355,
		0.795181,
		0.78677,
		0.778118,
		0.769224,
		0.760084,
		0.750696,
		0.741059,
		0.73117,
		0.721027,
		0.710629,
		0.699976,
		0.689066,
		0.677898,
		0.666474,
		0.654791,
		0.642853,
		0.630658,
		0.618209,
		0.605508,
		0.592557,
		0.579359,
		0.565918,
		0.552238,
		0.538323,
		0.52418,
		0.509814,
		0.495233,
		0.480445,
		0.465459,
		0.450283,
		0.43493,
		0.419412,
		0.403741,
		0.387931,
		0.371999,
		0.355962,
		0.339837,
		0.323646,
		0.307411,
		0.291153,
		0.2749,
		0.258679,
		0.242518,
		0.226451,
		0.210511,
		0.194735,
		0.179163,
		0.163836,
		0.148801,
		0.134105,
		0.119801,
		0.105945,
		0.092596,
		0.079817,
		0.067676,
		0.056247,
		0.045605,
		0.035835,
		0.027023,
		0.019264,
		0.012657,
		0.00731,
		0.003336,
		0.000857,
		0.925841,
		0.921922,
		0.91785,
		0.913621,
		0.909232,
		0.90468,
		0.899959,
		0.895068,
		0.890001,
		0.884756,
		0.879329,
		0.873716,
		0.867913,
		0.861918,
		0.855726,
		0.849334,
		0.842738,
		0.835936,
		0.828924,
		0.821699,
		0.814257,
		0.806596,
		0.798713,
		0.790604,
		0.782268,
		0.773702,
		0.764902,
		0.755868,
		0.746597,
		0.737086,
		0.727336,
		0.717343,
		0.707107,
		0.696627,
		0.685901,
		0.674931,
		0.663715,
		0.652255,
		0.640549,
		0.628601,
		0.61641,
		0.603979,
		0.591311,
		0.578407,
		0.565272,
		0.551909,
		0.538323,
		0.524519,
		0.510503,
		0.496282,
		0.481862,
		0.467253,
		0.452462,
		0.437501,
		0.42238,
		0.407111,
		0.391707,
		0.376183,
		0.360554,
		0.344836,
		0.32905,
		0.313213,
		0.297347,
		0.281477,
		0.265625,
		0.249819,
		0.234088,
		0.218463,
		0.202976,
		0.187663,
		0.172561,
		0.157713,
		0.14316,
		0.128949,
		0.11513,
		0.101756,
		0.088884,
		0.076573,
		0.064888,
		0.053897,
		0.043674,
		0.034297,
		0.025848,
		0.018415,
		0.012092,
		0.006979,
		0.003183,
		0.000817,
		0.900183,
		0.895415,
		0.89048,
		0.885375,
		0.880097,
		0.874642,
		0.869007,
		0.863189,
		0.857183,
		0.850988,
		0.8446,
		0.838015,
		0.83123,
		0.824244,
		0.817052,
		0.809651,
		0.80204,
		0.794215,
		0.786174,
		0.777914,
		0.769433,
		0.76073,
		0.751801,
		0.742645,
		0.73326,
		0.723646,
		0.7138,
		0.703721,
		0.69341,
		0.682865,
		0.672085,
		0.661072,
		0.649825,
		0.638345,
		0.626633,
		0.61469,
		0.602518,
		0.59012,
		0.577498,
		0.564656,
		0.551596,
		0.538323,
		0.524843,
		0.51116,
		0.497281,
		0.483212,
		0.468961,
		0.454537,
		0.439948,
		0.425205,
		0.410318,
		0.395301,
		0.380165,
		0.364925,
		0.349596,
		0.334196,
		0.318742,
		0.303253,
		0.28775,
		0.272257,
		0.256796,
		0.241394,
		0.226079,
		0.21088,
		0.19583,
		0.180962,
		0.166312,
		0.151921,
		0.137829,
		0.124081,
		0.110724,
		0.097809,
		0.085389,
		0.073521,
		0.062267,
		0.051692,
		0.041863,
		0.032856,
		0.024747,
		0.017621,
		0.011564,
		0.006671,
		0.003041,
		0.00078,
		0.870045,
		0.864394,
		0.858565,
		0.852555,
		0.846361,
		0.839981,
		0.833411,
		0.826649,
		0.819691,
		0.812535,
		0.805179,
		0.79762,
		0.789856,
		0.781884,
		0.773702,
		0.765307,
		0.756699,
		0.747875,
		0.738833,
		0.729573,
		0.720092,
		0.71039,
		0.700466,
		0.690319,
		0.679948,
		0.669354,
		0.658536,
		0.647496,
		0.636233,
		0.624748,
		0.613044,
		0.601121,
		0.588982,
		0.57663,
		0.564067,
		0.551297,
		0.538323,
		0.525152,
		0.511786,
		0.498233,
		0.484498,
		0.470589,
		0.456513,
		0.442279,
		0.427896,
		0.413374,
		0.398725,
		0.383959,
		0.369091,
		0.354133,
		0.339103,
		0.324015,
		0.308888,
		0.29374,
		0.278592,
		0.263466,
		0.248386,
		0.233376,
		0.218463,
		0.203676,
		0.189047,
		0.174607,
		0.160392,
		0.146439,
		0.132788,
		0.119481,
		0.106564,
		0.094085,
		0.082095,
		0.070648,
		0.059802,
		0.049618,
		0.040162,
		0.031504,
		0.023716,
		0.016877,
		0.011069,
		0.006382,
		0.002907,
		0.000745,
		0.835475,
		0.828924,
		0.822187,
		0.815262,
		0.808146,
		0.800837,
		0.793332,
		0.78563,
		0.777728,
		0.769624,
		0.761317,
		0.752804,
		0.744084,
		0.735156,
		0.726017,
		0.716668,
		0.707107,
		0.697333,
		0.687346,
		0.677145,
		0.66673,
		0.656102,
		0.645261,
		0.634207,
		0.622942,
		0.611467,
		0.599783,
		0.587893,
		0.575799,
		0.563503,
		0.55101,
		0.538323,
		0.525446,
		0.512384,
		0.499142,
		0.485726,
		0.472143,
		0.4584,
		0.444504,
		0.430464,
		0.416289,
		0.40199,
		0.387578,
		0.373065,
		0.358463,
		0.343786,
		0.32905,
		0.31427,
		0.299464,
		0.28465,
		0.269849,
		0.255081,
		0.24037,
		0.22574,
		0.211216,
		0.196828,
		0.182603,
		0.168575,
		0.154777,
		0.141244,
		0.128015,
		0.11513,
		0.102633,
		0.090569,
		0.078987,
		0.067939,
		0.057479,
		0.047667,
		0.038563,
		0.030233,
		0.022747,
		0.016179,
		0.010606,
		0.006111,
		0.002783,
		0.000713,
		0.796619,
		0.789171,
		0.781532,
		0.773702,
		0.765677,
		0.757456,
		0.749038,
		0.740422,
		0.731605,
		0.722587,
		0.713366,
		0.703943,
		0.694316,
		0.684484,
		0.674449,
		0.664208,
		0.653763,
		0.643115,
		0.632263,
		0.621209,
		0.609954,
		0.5985,
		0.586849,
		0.575003,
		0.562964,
		0.550737,
		0.538323,
		0.525728,
		0.512956,
		0.500011,
		0.4869,
		0.473628,
		0.460201,
		0.446628,
		0.432915,
		0.419073,
		0.405109,
		0.391034,
		0.37686,
		0.362598,
		0.34826,
		0.333861,
		0.319414,
		0.304937,
		0.290446,
		0.27596,
		0.261496,
		0.247077,
		0.232725,
		0.218463,
		0.204316,
		0.190312,
		0.176478,
		0.162846,
		0.149448,
		0.136317,
		0.123492,
		0.11101,
		0.098913,
		0.087245,
		0.076051,
		0.065382,
		0.055289,
		0.045828,
		0.037057,
		0.029038,
		0.021837,
		0.015523,
		0.010171,
		0.005858,
		0.002666,
		0.000682,
		0.753719,
		0.745396,
		0.736881,
		0.728175,
		0.719275,
		0.710181,
		0.700893,
		0.691409,
		0.681729,
		0.671853,
		0.661782,
		0.651515,
		0.641052,
		0.630396,
		0.619546,
		0.608503,
		0.59727,
		0.585848,
		0.57424,
		0.562448,
		0.550474,
		0.538323,
		0.525998,
		0.513503,
		0.500843,
		0.488022,
		0.475047,
		0.461923,
		0.448658,
		0.435259,
		0.421733,
		0.40809,
		0.394338,
		0.380488,
		0.366551,
		0.352538,
		0.338462,
		0.324337,
		0.310176,
		0.295997,
		0.281814,
		0.267647,
		0.253513,
		0.239433,
		0.225429,
		0.211524,
		0.197741,
		0.184108,
		0.170651,
		0.1574,
		0.144385,
		0.131641,
		0.119202,
		0.107106,
		0.095391,
		0.084099,
		0.073275,
		0.062966,
		0.053221,
		0.044093,
		0.035637,
		0.027912,
		0.02098,
		0.014907,
		0.009762,
		0.00562,
		0.002556,
		0.000654,
		0.707107,
		0.69795,
		0.688606,
		0.679074,
		0.669354,
		0.659446,
		0.649351,
		0.639069,
		0.628601,
		0.617947,
		0.607109,
		0.596089,
		0.584888,
		0.573509,
		0.561953,
		0.550223,
		0.538323,
		0.526257,
		0.514027,
		0.501639,
		0.489096,
		0.476406,
		0.463572,
		0.450601,
		0.437501,
		0.424278,
		0.410941,
		0.397498,
		0.383959,
		0.370333,
		0.356632,
		0.342867,
		0.32905,
		0.315194,
		0.301315,
		0.287427,
		0.273547,
		0.259691,
		0.245878,
		0.232129,
		0.218463,
		0.204903,
		0.191473,
		0.178197,
		0.165103,
		0.152218,
		0.139573,
		0.127199,
		0.11513,
		0.103402,
		0.092052,
		0.081119,
		0.070648,
		0.060681,
		0.051267,
		0.042454,
		0.034297,
		0.02685,
		0.020173,
		0.014327,
		0.009378,
		0.005396,
		0.002453,
		0.000627,
		0.657196,
		0.647268,
		0.63716,
		0.626874,
		0.61641,
		0.60577,
		0.594955,
		0.583966,
		0.572806,
		0.561477,
		0.549982,
		0.538323,
		0.526505,
		0.514529,
		0.502402,
		0.490126,
		0.477707,
		0.465151,
		0.452462,
		0.439648,
		0.426716,
		0.413672,
		0.400525,
		0.387283,
		0.373956,
		0.360554,
		0.347086,
		0.333566,
		0.320005,
		0.306416,
		0.292812,
		0.27921,
		0.265625,
		0.252073,
		0.238574,
		0.225145,
		0.211807,
		0.198581,
		0.185491,
		0.172561,
		0.159817,
		0.147285,
		0.134994,
		0.122976,
		0.111261,
		0.099885,
		0.088884,
		0.078295,
		0.068159,
		0.058518,
		0.049418,
		0.040905,
		0.033031,
		0.025848,
		0.019411,
		0.013779,
		0.009016,
		0.005185,
		0.002356,
		0.000602,
		0.604481,
		0.593863,
		0.583079,
		0.572131,
		0.561021,
		0.54975,
		0.538323,
		0.526742,
		0.515011,
		0.503133,
		0.491113,
		0.478955,
		0.466665,
		0.454246,
		0.441707,
		0.429052,
		0.416289,
		0.403426,
		0.390469,
		0.377429,
		0.364313,
		0.351133,
		0.337898,
		0.324619,
		0.31131,
		0.297982,
		0.28465,
		0.271328,
		0.258031,
		0.244776,
		0.23158,
		0.218463,
		0.205443,
		0.192542,
		0.179781,
		0.167184,
		0.154777,
		0.142584,
		0.130634,
		0.118957,
		0.107582,
		0.096544,
		0.085875,
		0.075614,
		0.065798,
		0.056468,
		0.047667,
		0.039439,
		0.031834,
		0.0249,
		0.018691,
		0.013263,
		0.008674,
		0.004986,
		0.002265,
		0.000579,
		0.549528,
		0.538323,
		0.526971,
		0.515474,
		0.503836,
		0.492061,
		0.480153,
		0.468117,
		0.455958,
		0.443682,
		0.431293,
		0.4188,
		0.406208,
		0.393526,
		0.38076,
		0.36792,
		0.355015,
		0.342055,
		0.32905,
		0.31601,
		0.302949,
		0.289878,
		0.276811,
		0.263763,
		0.250747,
		0.237781,
		0.224882,
		0.212067,
		0.199356,
		0.186768,
		0.174326,
		0.162051,
		0.149968,
		0.138102,
		0.12648,
		0.11513,
		0.104082,
		0.093366,
		0.083016,
		0.073068,
		0.063557,
		0.054523,
		0.046007,
		0.038051,
		0.0307,
		0.024004,
		0.018011,
		0.012775,
		0.008351,
		0.004799,
		0.002179,
		0.000557,
		0.492972,
		0.481304,
		0.469512,
		0.457602,
		0.445578,
		0.433445,
		0.42121,
		0.40888,
		0.39646,
		0.383959,
		0.371384,
		0.358744,
		0.346048,
		0.333306,
		0.320527,
		0.307723,
		0.294906,
		0.282087,
		0.26928,
		0.256499,
		0.243759,
		0.231074,
		0.218463,
		0.205942,
		0.193529,
		0.181245,
		0.169111,
		0.157147,
		0.145377,
		0.133826,
		0.122519,
		0.111483,
		0.100747,
		0.090341,
		0.080297,
		0.070648,
		0.061429,
		0.052677,
		0.044432,
		0.036734,
		0.029626,
		0.023155,
		0.017367,
		0.012313,
		0.008046,
		0.004622,
		0.002098,
		0.000536,
		0.435513,
		0.423526,
		0.411446,
		0.39928,
		0.387032,
		0.374712,
		0.362328,
		0.349886,
		0.337397,
		0.32487,
		0.312315,
		0.299743,
		0.287165,
		0.274594,
		0.262041,
		0.249522,
		0.237049,
		0.224639,
		0.212308,
		0.200072,
		0.18795,
		0.17596,
		0.164123,
		0.152459,
		0.140991,
		0.129743,
		0.118739,
		0.108006,
		0.09757,
		0.087461,
		0.077708,
		0.068345,
		0.059405,
		0.050923,
		0.042936,
		0.035484,
		0.028608,
		0.02235,
		0.016757,
		0.011876,
		0.007758,
		0.004454,
		0.002021,
		0.000516,
		0.377913,
		0.365774,
		0.353578,
		0.341334,
		0.32905,
		0.316735,
		0.304401,
		0.292057,
		0.279714,
		0.267385,
		0.255081,
		0.242817,
		0.230606,
		0.218463,
		0.206404,
		0.194445,
		0.182603,
		0.170898,
		0.159349,
		0.147975,
		0.136799,
		0.125842,
		0.11513,
		0.104687,
		0.094539,
		0.084714,
		0.075242,
		0.066153,
		0.057479,
		0.049255,
		0.041515,
		0.034297,
		0.027641,
		0.021587,
		0.016179,
		0.011462,
		0.007484,
		0.004296,
		0.001948,
		0.000497,
		0.320992,
		0.308888,
		0.29677,
		0.28465,
		0.272538,
		0.260446,
		0.248386,
		0.236371,
		0.224414,
		0.212531,
		0.200737,
		0.189047,
		0.177478,
		0.166048,
		0.154777,
		0.143683,
		0.132788,
		0.122113,
		0.111681,
		0.101517,
		0.091646,
		0.082095,
		0.072891,
		0.064064,
		0.055645,
		0.047667,
		0.040162,
		0.033168,
		0.026722,
		0.020862,
		0.01563,
		0.011069,
		0.007225,
		0.004145,
		0.001879,
		0.000479,
		0.265625,
		0.253765,
		0.241943,
		0.230171,
		0.218463,
		0.206833,
		0.195295,
		0.183866,
		0.172561,
		0.161399,
		0.150396,
		0.139573,
		0.128949,
		0.118545,
		0.108384,
		0.098489,
		0.088884,
		0.079594,
		0.070648,
		0.062072,
		0.053897,
		0.046154,
		0.038875,
		0.032094,
		0.025848,
		0.020173,
		0.015108,
		0.010696,
		0.006979,
		0.004003,
		0.001814,
		0.000462,
		0.212739,
		0.201355,
		0.190067,
		0.178891,
		0.167843,
		0.156939,
		0.146196,
		0.135635,
		0.125272,
		0.11513,
		0.105229,
		0.095593,
		0.086243,
		0.077206,
		0.068506,
		0.060171,
		0.05223,
		0.044712,
		0.037648,
		0.031071,
		0.025016,
		0.019517,
		0.014613,
		0.010342,
		0.006746,
		0.003868,
		0.001752,
		0.000447,
		0.163312,
		0.152659,
		0.142168,
		0.131859,
		0.121749,
		0.111859,
		0.102209,
		0.092821,
		0.083718,
		0.074922,
		0.06646,
		0.058356,
		0.050639,
		0.043337,
		0.036479,
		0.030097,
		0.024223,
		0.018893,
		0.014141,
		0.010005,
		0.006524,
		0.003739,
		0.001693,
		0.000431,
		0.118371,
		0.108724,
		0.099316,
		0.090168,
		0.081301,
		0.072738,
		0.064503,
		0.056622,
		0.049119,
		0.042023,
		0.035363,
		0.029167,
		0.023468,
		0.018298,
		0.013691,
		0.009684,
		0.006313,
		0.003617,
		0.001638,
		0.000417,
		0.078987,
		0.070648,
		0.062632,
		0.054963,
		0.047667,
		0.040769,
		0.034297,
		0.02828,
		0.022747,
		0.017731,
		0.013263,
		0.009378,
		0.006111,
		0.003501,
		0.001584,
		0.000403,
		0.046277,
		0.039569,
		0.033279,
		0.027432,
		0.022059,
		0.01719,
		0.012854,
		0.009086,
		0.00592,
		0.00339,
		0.001534,
		0.00039,
		0.021402,
		0.016673,
		0.012464,
		0.008808,
		0.005737,
		0.003284,
		0.001486,
		0.000378,
		0.005562,
		0.003183,
		0.00144,
		0.000366,
		0.000363,
		0.002221,
		9.3e-005,
		0.001462,
		0.004379,
		0.008741,
		0.014447,
		0.000846,
		0.00331,
		0.007282,
		0.012657,
		0.019337,
		0.027226,
		0.036235,
		0.000387,
		0.002366,
		0.00592,
		0.01094,
		0.017323,
		0.024968,
		0.033782,
		0.043674,
		0.05456,
		0.066356,
		0.0001,
		0.001559,
		0.004665,
		0.009304,
		0.015366,
		0.022747,
		0.031349,
		0.041077,
		0.051842,
		0.063557,
		0.076143,
		0.089522,
		0.103621,
		0.000903,
		0.003529,
		0.007758,
		0.013475,
		0.02057,
		0.028941,
		0.038489,
		0.049119,
		0.060743,
		0.073275,
		0.086636,
		0.100747,
		0.115538,
		0.130939,
		0.146884,
		0.000414,
		0.002525,
		0.006313,
		0.011657,
		0.018444,
		0.026564,
		0.035914,
		0.046396,
		0.057915,
		0.070384,
		0.083718,
		0.097836,
		0.112664,
		0.128128,
		0.144161,
		0.160699,
		0.177679,
		0.195044,
		0.000107,
		0.001665,
		0.004979,
		0.009923,
		0.016376,
		0.024223,
		0.033357,
		0.043674,
		0.055077,
		0.067471,
		0.08077,
		0.094888,
		0.109748,
		0.125272,
		0.141391,
		0.158037,
		0.175145,
		0.192655,
		0.210511,
		0.228658,
		0.247045,
		0.000966,
		0.003771,
		0.008282,
		0.014374,
		0.021925,
		0.030823,
		0.040959,
		0.05223,
		0.064539,
		0.077793,
		0.091904,
		0.106789,
		0.122371,
		0.138573,
		0.155325,
		0.172561,
		0.190218,
		0.208236,
		0.226559,
		0.245133,
		0.26391,
		0.282841,
		0.301882,
		0.000443,
		0.0027,
		0.006746,
		0.012447,
		0.019678,
		0.028318,
		0.038254,
		0.049379,
		0.061589,
		0.074788,
		0.088884,
		0.103789,
		0.119423,
		0.135705,
		0.152563,
		0.169927,
		0.187731,
		0.205913,
		0.224414,
		0.243179,
		0.262156,
		0.281295,
		0.300551,
		0.31988,
		0.339242,
		0.358598,
		0.000114,
		0.001783,
		0.005327,
		0.010606,
		0.017488,
		0.025848,
		0.035564,
		0.046525,
		0.058623,
		0.071756,
		0.085828,
		0.100747,
		0.116428,
		0.132788,
		0.14975,
		0.167242,
		0.185194,
		0.203541,
		0.222223,
		0.241182,
		0.260362,
		0.279714,
		0.299189,
		0.318742,
		0.33833,
		0.357915,
		0.377459,
		0.396928,
		0.416289,
		0.001035,
		0.004038,
		0.008861,
		0.015366,
		0.023419,
		0.032895,
		0.043674,
		0.055645,
		0.0687,
		0.082738,
		0.097664,
		0.113386,
		0.12982,
		0.146884,
		0.164503,
		0.182603,
		0.201118,
		0.219983,
		0.239139,
		0.258527,
		0.278096,
		0.297795,
		0.317576,
		0.337397,
		0.357216,
		0.376995,
		0.396697,
		0.416289,
		0.435741,
		0.455022,
		0.474108,
		0.000475,
		0.002895,
		0.007225,
		0.01332,
		0.02104,
		0.030251,
		0.04083,
		0.052658,
		0.065621,
		0.079615,
		0.094539,
		0.110298,
		0.126801,
		0.143965,
		0.16171,
		0.179959,
		0.198643,
		0.217694,
		0.237049,
		0.25665,
		0.27644,
		0.296367,
		0.316383,
		0.336442,
		0.3565,
		0.376519,
		0.39646,
		0.416289,
		0.435974,
		0.455485,
		0.474793,
		0.493875,
		0.512705,
		0.531263,
		0.000123,
		0.001913,
		0.005711,
		0.011362,
		0.018718,
		0.027641,
		0.037997,
		0.049664,
		0.062522,
		0.076461,
		0.091374,
		0.107161,
		0.12373,
		0.140991,
		0.158861,
		0.17726,
		0.196114,
		0.215353,
		0.234912,
		0.254728,
		0.274744,
		0.294906,
		0.315161,
		0.335463,
		0.355767,
		0.376031,
		0.396218,
		0.416289,
		0.436213,
		0.455958,
		0.475496,
		0.494799,
		0.513845,
		0.532609,
		0.551073,
		0.569217,
		0.587025,
		0.001112,
		0.004334,
		0.009504,
		0.016464,
		0.02507,
		0.035182,
		0.046668,
		0.059405,
		0.073275,
		0.088168,
		0.103978,
		0.120607,
		0.137962,
		0.155955,
		0.174504,
		0.193529,
		0.212959,
		0.232725,
		0.252761,
		0.273008,
		0.293408,
		0.313909,
		0.33446,
		0.355015,
		0.375532,
		0.395969,
		0.416289,
		0.436458,
		0.456443,
		0.476215,
		0.495746,
		0.515011,
		0.533987,
		0.552654,
		0.570991,
		0.588982,
		0.606611,
		0.623865,
		0.64073,
		0.000511,
		0.003111,
		0.007758,
		0.014288,
		0.022547,
		0.032389,
		0.043674,
		0.056273,
		0.070061,
		0.084923,
		0.100747,
		0.117431,
		0.134877,
		0.152992,
		0.171689,
		0.190888,
		0.210511,
		0.230487,
		0.250747,
		0.27123,
		0.291874,
		0.312626,
		0.333432,
		0.354245,
		0.37502,
		0.395714,
		0.416289,
		0.436709,
		0.45694,
		0.476952,
		0.496716,
		0.516206,
		0.535398,
		0.554272,
		0.572806,
		0.590984,
		0.608789,
		0.626207,
		0.643226,
		0.659834,
		0.676022,
		0.691782,
		0.000132,
		0.002059,
		0.00614,
		0.012202,
		0.020082,
		0.029626,
		0.040688,
		0.05313,
		0.066821,
		0.08164,
		0.09747,
		0.114202,
		0.131734,
		0.149968,
		0.168815,
		0.188187,
		0.208006,
		0.228195,
		0.248685,
		0.269407,
		0.290302,
		0.31131,
		0.332378,
		0.353455,
		0.374495,
		0.395453,
		0.416289,
		0.436966,
		0.45745,
		0.477707,
		0.497709,
		0.51743,
		0.536843,
		0.555928,
		0.574663,
		0.593031,
		0.611015,
		0.628601,
		0.645775,
		0.662527,
		0.678846,
		0.694725,
		0.710157,
		0.725136,
		0.739658,
		0.001198,
		0.004665,
		0.010218,
		0.017685,
		0.026902,
		0.037715,
		0.049979,
		0.063557,
		0.07832,
		0.094146,
		0.110919,
		0.128533,
		0.146884,
		0.165879,
		0.185427,
		0.205443,
		0.225849,
		0.246571,
		0.26754,
		0.28869,
		0.309961,
		0.331297,
		0.352645,
		0.373956,
		0.395185,
		0.416289,
		0.43723,
		0.457972,
		0.478481,
		0.498728,
		0.518684,
		0.538323,
		0.557624,
		0.576564,
		0.595126,
		0.613292,
		0.631048,
		0.64838,
		0.665277,
		0.681729,
		0.697728,
		0.713267,
		0.72834,
		0.742944,
		0.757075,
		0.770731,
		0.783913,
		0.000551,
		0.003352,
		0.008351,
		0.015366,
		0.024223,
		0.034761,
		0.046826,
		0.060273,
		0.074965,
		0.090775,
		0.107582,
		0.125272,
		0.143738,
		0.16288,
		0.182603,
		0.20282,
		0.223446,
		0.244406,
		0.265625,
		0.287037,
		0.308577,
		0.330188,
		0.351814,
		0.373403,
		0.39491,
		0.416289,
		0.437501,
		0.458508,
		0.479275,
		0.499772,
		0.519969,
		0.53984,
		0.559361,
		0.578511,
		0.59727,
		0.615622,
		0.63355,
		0.651042,
		0.668086,
		0.684672,
		0.700791,
		0.716437,
		0.731605,
		0.746289,
		0.760489,
		0.774201,
		0.787425,
		0.800163,
		0.812415,
		0.824185,
		0.000143,
		0.002221,
		0.006618,
		0.013138,
		0.021601,
		0.031834,
		0.043674,
		0.05697,
		0.071577,
		0.08736,
		0.104192,
		0.121952,
		0.140529,
		0.159817,
		0.179716,
		0.200135,
		0.220985,
		0.242186,
		0.263662,
		0.285341,
		0.307157,
		0.32905,
		0.35096,
		0.372836,
		0.394628,
		0.416289,
		0.437779,
		0.459058,
		0.48009,
		0.500843,
		0.521286,
		0.541394,
		0.561141,
		0.580504,
		0.599465,
		0.618005,
		0.636109,
		0.653763,
		0.670956,
		0.687676,
		0.703917,
		0.71967,
		0.734932,
		0.749696,
		0.763962,
		0.777728,
		0.790993,
		0.803759,
		0.816028,
		0.827802,
		0.839085,
		0.849883,
		0.860201,
		0.001294,
		0.005035,
		0.011016,
		0.019046,
		0.028941,
		0.040531,
		0.053654,
		0.068159,
		0.083901,
		0.100747,
		0.118572,
		0.137255,
		0.156687,
		0.176763,
		0.197385,
		0.218463,
		0.23991,
		0.261647,
		0.283601,
		0.3057,
		0.327881,
		0.350084,
		0.372254,
		0.394338,
		0.416289,
		0.438064,
		0.459622,
		0.480925,
		0.501941,
		0.522638,
		0.542987,
		0.562964,
		0.582546,
		0.601712,
		0.620445,
		0.638727,
		0.656545,
		0.673888,
		0.690745,
		0.707107,
		0.722968,
		0.738322,
		0.753166,
		0.767497,
		0.781314,
		0.794618,
		0.80741,
		0.819691,
		0.831466,
		0.842738,
		0.853514,
		0.863799,
		0.873601,
		0.882926,
		0.891784,
		0.000596,
		0.003623,
		0.009016,
		0.01657,
		0.026093,
		0.037403,
		0.05033,
		0.064712,
		0.080399,
		0.09725,
		0.11513,
		0.133916,
		0.15349,
		0.173742,
		0.19457,
		0.215878,
		0.237576,
		0.259581,
		0.281814,
		0.304203,
		0.326681,
		0.349184,
		0.371655,
		0.39404,
		0.416289,
		0.438357,
		0.460201,
		0.481783,
		0.503068,
		0.524024,
		0.544621,
		0.564834,
		0.584639,
		0.604014,
		0.622942,
		0.641405,
		0.65939,
		0.676885,
		0.693878,
		0.710363,
		0.726331,
		0.741777,
		0.756699,
		0.771094,
		0.784961,
		0.7983,
		0.811114,
		0.823405,
		0.835177,
		0.846434,
		0.857183,
		0.867431,
		0.877185,
		0.886452,
		0.895243,
		0.903567,
		0.911434,
		0.918855,
		0.000155,
		0.002404,
		0.007154,
		0.014187,
		0.023299,
		0.034297,
		0.047001,
		0.061241,
		0.076857,
		0.093699,
		0.111627,
		0.130509,
		0.150223,
		0.170651,
		0.191686,
		0.213228,
		0.235182,
		0.25746,
		0.27998,
		0.302666,
		0.325448,
		0.34826,
		0.371041,
		0.393734,
		0.416289,
		0.438658,
		0.460796,
		0.482664,
		0.504225,
		0.525446,
		0.546297,
		0.566751,
		0.586783,
		0.606372,
		0.625499,
		0.644146,
		0.6623,
		0.679948,
		0.697079,
		0.713686,
		0.729761,
		0.745299,
		0.760297,
		0.774754,
		0.788668,
		0.80204,
		0.814873,
		0.827169,
		0.838934,
		0.850172,
		0.86089,
		0.871096,
		0.880796,
		0.890001,
		0.89872,
		0.906963,
		0.914741,
		0.922066,
		0.92895,
		0.935405,
		0.941445,
		0.001403,
		0.00545,
		0.011912,
		0.02057,
		0.031222,
		0.043674,
		0.057749,
		0.073275,
		0.090096,
		0.108062,
		0.127035,
		0.146884,
		0.167488,
		0.188732,
		0.210511,
		0.232725,
		0.255282,
		0.278096,
		0.301087,
		0.324181,
		0.34731,
		0.370409,
		0.39342,
		0.416289,
		0.438967,
		0.461407,
		0.483569,
		0.505413,
		0.526907,
		0.548017,
		0.568718,
		0.588982,
		0.608789,
		0.628118,
		0.646952,
		0.665277,
		0.68308,
		0.70035,
		0.717079,
		0.73326,
		0.748889,
		0.763962,
		0.778478,
		0.792437,
		0.805839,
		0.818687,
		0.830985,
		0.842738,
		0.853952,
		0.864634,
		0.874792,
		0.884434,
		0.893571,
		0.902212,
		0.910368,
		0.918052,
		0.925275,
		0.932051,
		0.938392,
		0.944313,
		0.949827,
		0.954949,
		0.959694,
		0.000647,
		0.003928,
		0.009762,
		0.017921,
		0.028187,
		0.040357,
		0.05424,
		0.069658,
		0.086442,
		0.104436,
		0.123492,
		0.143473,
		0.164251,
		0.185705,
		0.207724,
		0.230203,
		0.253045,
		0.27616,
		0.299464,
		0.322879,
		0.346332,
		0.369759,
		0.393097,
		0.416289,
		0.439285,
		0.462036,
		0.484498,
		0.506634,
		0.528406,
		0.549783,
		0.570735,
		0.591237,
		0.611266,
		0.630801,
		0.649825,
		0.668323,
		0.686282,
		0.703692,
		0.720543,
		0.736831,
		0.752549,
		0.767695,
		0.782268,
		0.796268,
		0.809696,
		0.822556,
		0.834852,
		0.846589,
		0.857774,
		0.868415,
		0.878519,
		0.888097,
		0.897159,
		0.905717,
		0.913781,
		0.921364,
		0.928479,
		0.935141,
		0.941362,
		0.947158,
		0.952544,
		0.957534,
		0.962145,
		0.966392,
		0.970291,
		0.973858,
		0.000168,
		0.00261,
		0.007758,
		0.015366,
		0.025204,
		0.037057,
		0.050721,
		0.066007,
		0.082738,
		0.100747,
		0.119879,
		0.139988,
		0.160938,
		0.182603,
		0.204865,
		0.227614,
		0.250747,
		0.27417,
		0.297795,
		0.321539,
		0.345327,
		0.369091,
		0.392764,
		0.416289,
		0.439612,
		0.462682,
		0.485454,
		0.507888,
		0.529947,
		0.551596,
		0.572806,
		0.59355,
		0.613805,
		0.63355,
		0.652767,
		0.67144,
		0.689557,
		0.707107,
		0.724081,
		0.740474,
		0.75628,
		0.771497,
		0.786125,
		0.800163,
		0.813614,
		0.826481,
		0.83877,
		0.850486,
		0.861636,
		0.87223,
		0.882275,
		0.891784,
		0.900765,
		0.909232,
		0.917197,
		0.924674,
		0.931675,
		0.938217,
		0.944313,
		0.949978,
		0.95523,
		0.960083,
		0.964555,
		0.968661,
		0.972419,
		0.975845,
		0.978957,
		0.981771,
		0.984305,
		0.001526,
		0.00592,
		0.012921,
		0.022285,
		0.033782,
		0.047197,
		0.062328,
		0.078987,
		0.096998,
		0.116196,
		0.136427,
		0.157547,
		0.179424,
		0.201932,
		0.224955,
		0.248386,
		0.272124,
		0.296078,
		0.32016,
		0.344293,
		0.368403,
		0.392422,
		0.416289,
		0.439948,
		0.463346,
		0.486437,
		0.509177,
		0.531529,
		0.553458,
		0.574932,
		0.595924,
		0.61641,
		0.636368,
		0.655781,
		0.674631,
		0.692907,
		0.710598,
		0.727694,
		0.744191,
		0.760084,
		0.77537,
		0.790049,
		0.804122,
		0.817592,
		0.830462,
		0.842738,
		0.854428,
		0.865538,
		0.876079,
		0.886059,
		0.895491,
		0.904386,
		0.912756,
		0.920616,
		0.927979,
		0.93486,
		0.941275,
		0.947239,
		0.952769,
		0.957881,
		0.962592,
		0.96692,
		0.970881,
		0.974494,
		0.977776,
		0.980745,
		0.983418,
		0.985814,
		0.987949,
		0.989843,
		0.99151,
		0.000705,
		0.004273,
		0.010606,
		0.019444,
		0.030544,
		0.043674,
		0.058623,
		0.07519,
		0.093187,
		0.112441,
		0.132788,
		0.154076,
		0.176164,
		0.198921,
		0.222223,
		0.245958,
		0.27002,
		0.294311,
		0.318742,
		0.343228,
		0.367695,
		0.39207,
		0.416289,
		0.440294,
		0.46403,
		0.487448,
		0.510503,
		0.533157,
		0.555372,
		0.577116,
		0.598361,
		0.619082,
		0.639258,
		0.658868,
		0.677898,
		0.696335,
		0.714166,
		0.731385,
		0.747985,
		0.763962,
		0.779315,
		0.794042,
		0.808146,
		0.82163,
		0.834498,
		0.846758,
		0.858415,
		0.869479,
		0.87996,
		0.889869,
		0.899218,
		0.908019,
		0.916286,
		0.924033,
		0.931276,
		0.938031,
		0.944313,
		0.950139,
		0.955526,
		0.960493,
		0.965057,
		0.969235,
		0.973047,
		0.976511,
		0.979645,
		0.982468,
		0.984999,
		0.987255,
		0.989256,
		0.991018,
		0.992561,
		0.993901,
		0.995056,
		0.996042,
		0.000183,
		0.002844,
		0.008442,
		0.016698,
		0.027353,
		0.040162,
		0.054899,
		0.07135,
		0.089317,
		0.108615,
		0.12907,
		0.150522,
		0.172822,
		0.19583,
		0.219416,
		0.243461,
		0.267854,
		0.292492,
		0.317281,
		0.342132,
		0.366965,
		0.391707,
		0.416289,
		0.44065,
		0.464734,
		0.488488,
		0.511868,
		0.53483,
		0.557338,
		0.579359,
		0.600863,
		0.621825,
		0.642221,
		0.662033,
		0.681244,
		0.699842,
		0.717815,
		0.735156,
		0.751857,
		0.767917,
		0.783333,
		0.798105,
		0.812236,
		0.825729,
		0.838591,
		0.850827,
		0.862446,
		0.873458,
		0.883873,
		0.893703,
		0.902962,
		0.911662,
		0.919818,
		0.927446,
		0.934562,
		0.941183,
		0.947325,
		0.953007,
		0.958246,
		0.963062,
		0.967473,
		0.971498,
		0.975156,
		0.978467,
		0.98145,
		0.984125,
		0.98651,
		0.988626,
		0.99049,
		0.992122,
		0.993541,
		0.994763,
		0.995807,
		0.99669,
		0.997428,
		0.998037,
		0.998532,
		0.001665,
		0.006452,
		0.014064,
		0.024223,
		0.03667,
		0.051161,
		0.067471,
		0.085389,
		0.104717,
		0.125272,
		0.146884,
		0.169394,
		0.192655,
		0.21653,
		0.240892,
		0.265625,
		0.290619,
		0.315776,
		0.341002,
		0.366214,
		0.391333,
		0.416289,
		0.441017,
		0.465459,
		0.489559,
		0.513271,
		0.536552,
		0.559361,
		0.581665,
		0.603434,
		0.62464,
		0.645261,
		0.665277,
		0.684672,
		0.703432,
		0.721547,
		0.739008,
		0.75581,
		0.771949,
		0.787425,
		0.802239,
		0.816392,
		0.82989,
		0.842738,
		0.854945,
		0.86652,
		0.877473,
		0.887815,
		0.897559,
		0.90672,
		0.915312,
		0.92335,
		0.930852,
		0.937833,
		0.944313,
		0.950308,
		0.955839,
		0.960924,
		0.965583,
		0.969836,
		0.973702,
		0.977203,
		0.980357,
		0.983187,
		0.985711,
		0.987949,
		0.989923,
		0.991651,
		0.993153,
		0.994448,
		0.995554,
		0.99649,
		0.997272,
		0.997917,
		0.998443,
		0.998863,
		0.999193,
		0.999446,
		0.999635,
		0.000771,
		0.004665,
		0.011564,
		0.02117,
		0.033208,
		0.047417,
		0.063557,
		0.081404,
		0.100747,
		0.121393,
		0.14316,
		0.165879,
		0.189395,
		0.213563,
		0.238249,
		0.263329,
		0.28869,
		0.314225,
		0.339837,
		0.365439,
		0.390948,
		0.416289,
		0.441396,
		0.466205,
		0.490663,
		0.514717,
		0.538323,
		0.561441,
		0.584036,
		0.606074,
		0.62753,
		0.64838,
		0.668603,
		0.688183,
		0.707107,
		0.725363,
		0.742944,
		0.759844,
		0.776061,
		0.791594,
		0.806444,
		0.820615,
		0.834112,
		0.846941,
		0.859113,
		0.870636,
		0.881522,
		0.891784,
		0.901435,
		0.91049,
		0.918966,
		0.926878,
		0.934245,
		0.941085,
		0.947416,
		0.953258,
		0.958631,
		0.963556,
		0.968053,
		0.972142,
		0.975845,
		0.979184,
		0.982179,
		0.984851,
		0.987222,
		0.989313,
		0.991144,
		0.992736,
		0.994109,
		0.995282,
		0.996274,
		0.997104,
		0.997789,
		0.998347,
		0.998793,
		0.999143,
		0.999412,
		0.999612,
		0.999757,
		0.999857,
		0.999922,
		0.999962,
		0.000201,
		0.003111,
		0.00922,
		0.018211,
		0.029788,
		0.043674,
		0.059613,
		0.077365,
		0.096707,
		0.117431,
		0.139346,
		0.162273,
		0.186045,
		0.210511,
		0.235528,
		0.260965,
		0.286701,
		0.312626,
		0.338636,
		0.36464,
		0.39055,
		0.416289,
		0.441786,
		0.466975,
		0.4918,
		0.516206,
		0.540147,
		0.563582,
		0.586474,
		0.608789,
		0.630499,
		0.651582,
		0.672015,
		0.691782,
		0.710869,
		0.729267,
		0.746966,
		0.763962,
		0.780253,
		0.795839,
		0.810721,
		0.824905,
		0.838394,
		0.851199,
		0.863328,
		0.874792,
		0.885604,
		0.895777,
		0.905327,
		0.914268,
		0.92262,
		0.930398,
		0.937623,
		0.944313,
		0.950488,
		0.95617,
		0.961379,
		0.966137,
		0.970465,
		0.974386,
		0.977922,
		0.981094,
		0.983926,
		0.98644,
		0.988657,
		0.990599,
		0.992288,
		0.993744,
		0.994989,
		0.996042,
		0.996923,
		0.997651,
		0.998243,
		0.998717,
		0.999089,
		0.999375,
		0.999588,
		0.999741,
		0.999848,
		0.999917,
		0.99996,
		0.999984,
		0.999995,
		0.999999,
		1,
		0.001825,
		0.00706,
		0.015366,
		0.026425,
		0.039943,
		0.055645,
		0.073275,
		0.092596,
		0.113386,
		0.135442,
		0.158573,
		0.182603,
		0.207371,
		0.232725,
		0.258527,
		0.28465,
		0.310976,
		0.337397,
		0.363815,
		0.39014,
		0.416289,
		0.442188,
		0.467769,
		0.492972,
		0.51774,
		0.542026,
		0.565786,
		0.588982,
		0.61158,
		0.63355,
		0.654869,
		0.675515,
		0.69547,
		0.714722,
		0.73326,
		0.751077,
		0.768167,
		0.784529,
		0.800163,
		0.815072,
		0.829262,
		0.842738,
		0.855511,
		0.86759,
		0.878987,
		0.889716,
		0.899793,
		0.909232,
		0.918052,
		0.92627,
		0.933906,
		0.94098,
		0.947512,
		0.953525,
		0.959038,
		0.964076,
		0.968661,
		0.972816,
		0.976563,
		0.979927,
		0.982931,
		0.985597,
		0.987949,
		0.990011,
		0.991804,
		0.99335,
		0.994673,
		0.995792,
		0.996728,
		0.997501,
		0.998131,
		0.998635,
		0.999031,
		0.999334,
		0.999561,
		0.999725,
		0.999838,
		0.999912,
		0.999957,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999995,
		0.000846,
		0.005114,
		0.012657,
		0.023136,
		0.036235,
		0.05166,
		0.069138,
		0.088416,
		0.109257,
		0.131445,
		0.154777,
		0.179066,
		0.204139,
		0.229837,
		0.256014,
		0.282534,
		0.309274,
		0.336118,
		0.362964,
		0.389717,
		0.416289,
		0.442603,
		0.468589,
		0.49418,
		0.519322,
		0.543963,
		0.568056,
		0.591564,
		0.61445,
		0.636686,
		0.658245,
		0.679106,
		0.699252,
		0.718669,
		0.737347,
		0.755278,
		0.772459,
		0.788888,
		0.804566,
		0.819497,
		0.833687,
		0.847143,
		0.859876,
		0.871897,
		0.883219,
		0.893857,
		0.903828,
		0.913148,
		0.921837,
		0.929913,
		0.937398,
		0.944313,
		0.950679,
		0.95652,
		0.961859,
		0.966719,
		0.971124,
		0.9751,
		0.978669,
		0.981857,
		0.984688,
		0.987186,
		0.989376,
		0.991281,
		0.992925,
		0.994331,
		0.995521,
		0.996517,
		0.99734,
		0.99801,
		0.998546,
		0.998968,
		0.999291,
		0.999532,
		0.999707,
		0.999827,
		0.999906,
		0.999955,
		0.999981,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999995,
		0.999984,
		0.999961,
		0.999922,
		0.999861,
		0.000221,
		0.003417,
		0.010111,
		0.019939,
		0.032563,
		0.047667,
		0.064959,
		0.084169,
		0.105044,
		0.127354,
		0.150882,
		0.175429,
		0.200812,
		0.226862,
		0.253422,
		0.280351,
		0.307516,
		0.334797,
		0.362085,
		0.389279,
		0.416289,
		0.443032,
		0.469434,
		0.495428,
		0.520954,
		0.545959,
		0.570395,
		0.594222,
		0.617404,
		0.63991,
		0.661713,
		0.682792,
		0.70313,
		0.722712,
		0.741528,
		0.759572,
		0.776841,
		0.793332,
		0.809049,
		0.823996,
		0.838179,
		0.851607,
		0.864293,
		0.876247,
		0.887485,
		0.898023,
		0.907879,
		0.917071,
		0.925619,
		0.933544,
		0.940869,
		0.947615,
		0.953807,
		0.959469,
		0.964625,
		0.969301,
		0.973521,
		0.977312,
		0.980698,
		0.983707,
		0.986362,
		0.988691,
		0.990717,
		0.992466,
		0.993962,
		0.995229,
		0.996289,
		0.997165,
		0.997879,
		0.998451,
		0.9989,
		0.999244,
		0.999501,
		0.999687,
		0.999816,
		0.9999,
		0.999952,
		0.99998,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999959,
		0.999917,
		0.999852,
		0.999757,
		0.999624,
		0.999446,
		0.999217,
		0.002008,
		0.007758,
		0.016857,
		0.028941,
		0.043674,
		0.060743,
		0.079857,
		0.100747,
		0.123166,
		0.146884,
		0.171689,
		0.197385,
		0.223793,
		0.250747,
		0.278096,
		0.3057,
		0.333432,
		0.361176,
		0.388827,
		0.416289,
		0.443475,
		0.470308,
		0.496716,
		0.522638,
		0.548017,
		0.572806,
		0.596961,
		0.620445,
		0.643226,
		0.665277,
		0.686577,
		0.707107,
		0.726854,
		0.745808,
		0.763962,
		0.781314,
		0.797864,
		0.813614,
		0.828569,
		0.842738,
		0.856131,
		0.86876,
		0.880639,
		0.891784,
		0.902212,
		0.911942,
		0.920996,
		0.929393,
		0.937158,
		0.944313,
		0.950882,
		0.956891,
		0.962366,
		0.967332,
		0.971816,
		0.975845,
		0.979446,
		0.982646,
		0.985471,
		0.987949,
		0.990107,
		0.991969,
		0.993563,
		0.994912,
		0.996042,
		0.996976,
		0.997737,
		0.998347,
		0.998826,
		0.999193,
		0.999468,
		0.999666,
		0.999803,
		0.999893,
		0.999948,
		0.999979,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999981,
		0.999956,
		0.999912,
		0.999843,
		0.999741,
		0.9996,
		0.999412,
		0.999169,
		0.998863,
		0.998488,
		0.998037,
		0.997501,
		0.000934,
		0.005631,
		0.013913,
		0.025389,
		0.039695,
		0.056497,
		0.075484,
		0.096367,
		0.118882,
		0.142782,
		0.167843,
		0.193856,
		0.220628,
		0.247986,
		0.275766,
		0.303823,
		0.332021,
		0.360237,
		0.38836,
		0.416289,
		0.443933,
		0.47121,
		0.498046,
		0.524376,
		0.550142,
		0.575292,
		0.599783,
		0.623575,
		0.646637,
		0.66894,
		0.690463,
		0.711187,
		0.731099,
		0.750188,
		0.76845,
		0.785882,
		0.802484,
		0.81826,
		0.833218,
		0.847365,
		0.860713,
		0.873276,
		0.88507,
		0.896111,
		0.906419,
		0.916014,
		0.924919,
		0.933156,
		0.940749,
		0.947725,
		0.954108,
		0.959925,
		0.965204,
		0.969973,
		0.97426,
		0.978092,
		0.981498,
		0.984507,
		0.987147,
		0.989445,
		0.991431,
		0.99313,
		0.994569,
		0.995774,
		0.996771,
		0.997583,
		0.998234,
		0.998745,
		0.999138,
		0.999431,
		0.999643,
		0.99979,
		0.999886,
		0.999945,
		0.999977,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.99998,
		0.999953,
		0.999906,
		0.999833,
		0.999725,
		0.999575,
		0.999375,
		0.999117,
		0.998793,
		0.998396,
		0.997917,
		0.997351,
		0.99669,
		0.995927,
		0.995056,
		0.994072,
		0.000244,
		0.003771,
		0.011138,
		0.021925,
		0.035743,
		0.05223,
		0.071053,
		0.091904,
		0.114498,
		0.138573,
		0.163887,
		0.190218,
		0.217363,
		0.245133,
		0.273359,
		0.301882,
		0.330561,
		0.359265,
		0.387877,
		0.416289,
		0.444407,
		0.472143,
		0.499421,
		0.526172,
		0.552335,
		0.577857,
		0.602692,
		0.6268,
		0.650148,
		0.672708,
		0.694455,
		0.715374,
		0.735449,
		0.754672,
		0.773038,
		0.790544,
		0.807193,
		0.82299,
		0.837941,
		0.852057,
		0.865352,
		0.877839,
		0.889537,
		0.900464,
		0.910641,
		0.92009,
		0.928835,
		0.9369,
		0.944313,
		0.951098,
		0.957286,
		0.962903,
		0.967979,
		0.972543,
		0.976625,
		0.980254,
		0.983462,
		0.986276,
		0.988728,
		0.990846,
		0.99266,
		0.994196,
		0.995483,
		0.996548,
		0.997416,
		0.998111,
		0.998658,
		0.999078,
		0.999391,
		0.999618,
		0.999775,
		0.999878,
		0.999941,
		0.999976,
		0.999992,
		0.999998,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.999979,
		0.99995,
		0.9999,
		0.999822,
		0.999707,
		0.999547,
		0.999334,
		0.999061,
		0.998717,
		0.998296,
		0.997789,
		0.997189,
		0.99649,
		0.995683,
		0.994763,
		0.993724,
		0.992561,
		0.991269,
		0.989843,
		0.988279,
		0.002221,
		0.008564,
		0.018575,
		0.031834,
		0.047952,
		0.066571,
		0.08736,
		0.110014,
		0.134253,
		0.159817,
		0.186469,
		0.213992,
		0.242186,
		0.270869,
		0.299874,
		0.32905,
		0.358259,
		0.387376,
		0.416289,
		0.444897,
		0.473108,
		0.500843,
		0.528027,
		0.5546,
		0.580504,
		0.605693,
		0.630124,
		0.653763,
		0.676582,
		0.698557,
		0.71967,
		0.739908,
		0.759263,
		0.777728,
		0.795304,
		0.811993,
		0.827802,
		0.842738,
		0.856815,
		0.870045,
		0.882447,
		0.894038,
		0.904839,
		0.914874,
		0.924165,
		0.932739,
		0.940621,
		0.947842,
		0.954428,
		0.960409,
		0.965817,
		0.970681,
		0.975034,
		0.978905,
		0.982327,
		0.985332,
		0.987949,
		0.990212,
		0.992149,
		0.993791,
		0.995167,
		0.996306,
		0.997234,
		0.997978,
		0.998563,
		0.999012,
		0.999348,
		0.999591,
		0.999759,
		0.999869,
		0.999937,
		0.999974,
		0.999992,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999993,
		0.999977,
		0.999947,
		0.999893,
		0.99981,
		0.999687,
		0.999517,
		0.999291,
		0.999,
		0.998635,
		0.998188,
		0.997651,
		0.997015,
		0.996274,
		0.995421,
		0.994448,
		0.99335,
		0.992122,
		0.990759,
		0.989256,
		0.987608,
		0.985814,
		0.983869,
		0.981771,
		0.979518,
		0.001035,
		0.006231,
		0.015366,
		0.027987,
		0.043674,
		0.062044,
		0.082738,
		0.105431,
		0.12982,
		0.155629,
		0.182603,
		0.210511,
		0.239139,
		0.268292,
		0.297795,
		0.327485,
		0.357216,
		0.386858,
		0.416289,
		0.445405,
		0.474108,
		0.502313,
		0.529947,
		0.556941,
		0.583238,
		0.608789,
		0.63355,
		0.657487,
		0.680569,
		0.702773,
		0.724081,
		0.74448,
		0.763962,
		0.782523,
		0.800163,
		0.816885,
		0.832697,
		0.84761,
		0.861636,
		0.874792,
		0.887096,
		0.898568,
		0.909232,
		0.919112,
		0.928233,
		0.936624,
		0.944313,
		0.95133,
		0.957705,
		0.963472,
		0.968661,
		0.973306,
		0.977439,
		0.981094,
		0.984305,
		0.987103,
		0.989522,
		0.991594,
		0.99335,
		0.994823,
		0.996042,
		0.997036,
		0.997833,
		0.99846,
		0.998941,
		0.999301,
		0.999561,
		0.999741,
		0.99986,
		0.999932,
		0.999972,
		0.999991,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999992,
		0.999976,
		0.999943,
		0.999886,
		0.999797,
		0.999666,
		0.999485,
		0.999244,
		0.998934,
		0.998546,
		0.998072,
		0.997501,
		0.996827,
		0.996042,
		0.995138,
		0.994109,
		0.992949,
		0.991651,
		0.990212,
		0.988626,
		0.986889,
		0.984999,
		0.982951,
		0.980745,
		0.978376,
		0.975845,
		0.97315,
		0.970291,
		0.967266,
		0.000272,
		0.004182,
		0.012329,
		0.024223,
		0.039412,
		0.057479,
		0.078042,
		0.100747,
		0.125272,
		0.151319,
		0.178617,
		0.206915,
		0.235987,
		0.265625,
		0.295641,
		0.325863,
		0.356136,
		0.38632,
		0.416289,
		0.44593,
		0.475143,
		0.503836,
		0.531932,
		0.559361,
		0.586062,
		0.611985,
		0.637084,
		0.661323,
		0.684672,
		0.707107,
		0.72861,
		0.749168,
		0.768774,
		0.787425,
		0.805122,
		0.821869,
		0.837676,
		0.852555,
		0.86652,
		0.879589,
		0.891784,
		0.903125,
		0.913638,
		0.92335,
		0.932289,
		0.940484,
		0.947967,
		0.954769,
		0.960924,
		0.966466,
		0.971428,
		0.975845,
		0.979753,
		0.983187,
		0.986181,
		0.98877,
		0.990988,
		0.99287,
		0.994448,
		0.995754,
		0.99682,
		0.997675,
		0.998347,
		0.998863,
		0.999249,
		0.999529,
		0.999722,
		0.999849,
		0.999927,
		0.99997,
		0.99999,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999991,
		0.999974,
		0.999939,
		0.999878,
		0.999782,
		0.999643,
		0.99945,
		0.999193,
		0.998863,
		0.998451,
		0.997946,
		0.99734,
		0.996625,
		0.995792,
		0.994834,
		0.993744,
		0.992516,
		0.991144,
		0.989624,
		0.987949,
		0.986118,
		0.984125,
		0.981968,
		0.979645,
		0.977154,
		0.974494,
		0.971663,
		0.968661,
		0.965488,
		0.962145,
		0.958631,
		0.954949,
		0.951098,
		0.00247,
		0.009504,
		0.02057,
		0.035182,
		0.052888,
		0.073275,
		0.095965,
		0.120607,
		0.146884,
		0.174504,
		0.203198,
		0.232725,
		0.262862,
		0.293408,
		0.324181,
		0.355015,
		0.385763,
		0.416289,
		0.446476,
		0.476215,
		0.505413,
		0.533987,
		0.561865,
		0.588982,
		0.615286,
		0.64073,
		0.665277,
		0.688896,
		0.711563,
		0.73326,
		0.753975,
		0.773702,
		0.792437,
		0.810183,
		0.826947,
		0.842738,
		0.857572,
		0.871464,
		0.884434,
		0.896506,
		0.907703,
		0.918052,
		0.927583,
		0.936326,
		0.944313,
		0.951577,
		0.958153,
		0.964076,
		0.969383,
		0.974109,
		0.978292,
		0.981968,
		0.985175,
		0.987949,
		0.990328,
		0.992346,
		0.994038,
		0.99544,
		0.996584,
		0.997501,
		0.998223,
		0.998778,
		0.999193,
		0.999493,
		0.999701,
		0.999838,
		0.999921,
		0.999968,
		0.99999,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999991,
		0.999972,
		0.999934,
		0.999869,
		0.999767,
		0.999618,
		0.999412,
		0.999138,
		0.998786,
		0.998347,
		0.997809,
		0.997165,
		0.996405,
		0.995521,
		0.994505,
		0.99335,
		0.99205,
		0.990599,
		0.988991,
		0.987222,
		0.985289,
		0.983187,
		0.980913,
		0.978467,
		0.975845,
		0.973047,
		0.970072,
		0.96692,
		0.96359,
		0.960083,
		0.956401,
		0.952544,
		0.948514,
		0.944313,
		0.939942,
		0.935405,
		0.930704,
		0.001154,
		0.006932,
		0.017058,
		0.031004,
		0.048281,
		0.068445,
		0.091084,
		0.115823,
		0.14232,
		0.170259,
		0.199356,
		0.229348,
		0.259999,
		0.291093,
		0.322436,
		0.353853,
		0.385184,
		0.416289,
		0.447041,
		0.477327,
		0.507048,
		0.536117,
		0.564456,
		0.592002,
		0.618696,
		0.644493,
		0.669354,
		0.693246,
		0.716147,
		0.738037,
		0.758906,
		0.778747,
		0.797559,
		0.815347,
		0.832117,
		0.847883,
		0.862659,
		0.876466,
		0.889324,
		0.901258,
		0.912296,
		0.922467,
		0.931803,
		0.940336,
		0.948101,
		0.955135,
		0.961473,
		0.967154,
		0.972215,
		0.976697,
		0.980638,
		0.984077,
		0.987053,
		0.989606,
		0.991772,
		0.99359,
		0.995096,
		0.996325,
		0.997312,
		0.998088,
		0.998685,
		0.999131,
		0.999454,
		0.999678,
		0.999825,
		0.999915,
		0.999965,
		0.999989,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.99999,
		0.99997,
		0.999929,
		0.99986,
		0.99975,
		0.999591,
		0.99937,
		0.999078,
		0.998703,
		0.998234,
		0.997662,
		0.996976,
		0.996168,
		0.995229,
		0.99415,
		0.992925,
		0.991548,
		0.990011,
		0.98831,
		0.98644,
		0.984397,
		0.982179,
		0.979781,
		0.977203,
		0.974442,
		0.971498,
		0.968369,
		0.965057,
		0.96156,
		0.957881,
		0.95402,
		0.949978,
		0.945759,
		0.941362,
		0.936792,
		0.932051,
		0.927141,
		0.922066,
		0.916829,
		0.911434,
		0.905884,
		0.000304,
		0.004665,
		0.013723,
		0.026902,
		0.043674,
		0.063557,
		0.086107,
		0.110919,
		0.137622,
		0.165879,
		0.195381,
		0.225849,
		0.257029,
		0.28869,
		0.320624,
		0.352645,
		0.384583,
		0.416289,
		0.447628,
		0.478481,
		0.508744,
		0.538323,
		0.56714,
		0.595126,
		0.622222,
		0.64838,
		0.673559,
		0.697728,
		0.720862,
		0.742944,
		0.763962,
		0.783913,
		0.802795,
		0.820615,
		0.837381,
		0.853109,
		0.867815,
		0.881522,
		0.894253,
		0.906036,
		0.9169,
		0.926878,
		0.936004,
		0.944313,
		0.951842,
		0.958631,
		0.96472,
		0.970147,
		0.974954,
		0.979184,
		0.982876,
		0.986073,
		0.988816,
		0.991144,
		0.993099,
		0.994719,
		0.996042,
		0.997104,
		0.99794,
		0.998583,
		0.999064,
		0.999412,
		0.999653,
		0.999812,
		0.999909,
		0.999962,
		0.999988,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999989,
		0.999968,
		0.999924,
		0.999849,
		0.999732,
		0.999561,
		0.999325,
		0.999012,
		0.998612,
		0.998111,
		0.997501,
		0.996771,
		0.995911,
		0.994912,
		0.993766,
		0.992466,
		0.991005,
		0.989376,
		0.987575,
		0.985597,
		0.983438,
		0.981094,
		0.978564,
		0.975845,
		0.972936,
		0.969836,
		0.966545,
		0.963062,
		0.959389,
		0.955526,
		0.951476,
		0.947239,
		0.942819,
		0.938217,
		0.933436,
		0.928479,
		0.92335,
		0.918052,
		0.912588,
		0.906963,
		0.90118,
		0.895243,
		0.889157,
		0.882926,
		0.876554,
		0.002763,
		0.010606,
		0.022905,
		0.039085,
		0.058623,
		0.081039,
		0.105894,
		0.132788,
		0.161357,
		0.191269,
		0.222223,
		0.253947,
		0.286194,
		0.318742,
		0.35139,
		0.383959,
		0.416289,
		0.448238,
		0.47968,
		0.510503,
		0.540612,
		0.569922,
		0.598361,
		0.625869,
		0.652395,
		0.677898,
		0.702346,
		0.725714,
		0.747985,
		0.769149,
		0.789202,
		0.808146,
		0.825987,
		0.842738,
		0.858415,
		0.873037,
		0.886629,
		0.899218,
		0.910833,
		0.921508,
		0.931276,
		0.940176,
		0.948246,
		0.955526,
		0.962058,
		0.967884,
		0.973047,
		0.977591,
		0.981561,
		0.984999,
		0.987949,
		0.990456,
		0.992561,
		0.994306,
		0.995731,
		0.996876,
		0.997777,
		0.99847,
		0.998989,
		0.999365,
		0.999625,
		0.999797,
		0.999901,
		0.999959,
		0.999987,
		0.999997,
		1,
		1,
		1,
		0.999998,
		0.999988,
		0.999965,
		0.999918,
		0.999838,
		0.999712,
		0.999529,
		0.999276,
		0.998941,
		0.998512,
		0.997978,
		0.997327,
		0.996548,
		0.995632,
		0.994569,
		0.99335,
		0.991969,
		0.990418,
		0.988691,
		0.986782,
		0.984688,
		0.982404,
		0.979927,
		0.977255,
		0.974386,
		0.971319,
		0.968053,
		0.964588,
		0.960924,
		0.957064,
		0.953007,
		0.948756,
		0.944313,
		0.93968,
		0.93486,
		0.929857,
		0.924674,
		0.919314,
		0.913781,
		0.908079,
		0.902212,
		0.896184,
		0.890001,
		0.883666,
		0.877185,
		0.870561,
		0.863799,
		0.856905,
		0.849883,
		0.842738,
		0.001294,
		0.007758,
		0.019046,
		0.034535,
		0.053654,
		0.075884,
		0.100747,
		0.127813,
		0.156687,
		0.187012,
		0.218463,
		0.250747,
		0.283601,
		0.316784,
		0.350084,
		0.383309,
		0.416289,
		0.448872,
		0.480925,
		0.512331,
		0.542987,
		0.572806,
		0.601712,
		0.629643,
		0.656545,
		0.682378,
		0.707107,
		0.730708,
		0.753166,
		0.77447,
		0.794618,
		0.813614,
		0.831466,
		0.848188,
		0.863799,
		0.878322,
		0.891784,
		0.904213,
		0.915644,
		0.926112,
		0.935655,
		0.944313,
		0.952128,
		0.959144,
		0.965405,
		0.970957,
		0.975845,
		0.980117,
		0.983819,
		0.986998,
		0.9897,
		0.991969,
		0.993851,
		0.995389,
		0.996625,
		0.997598,
		0.998347,
		0.998907,
		0.999313,
		0.999595,
		0.99978,
		0.999893,
		0.999956,
		0.999986,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999988,
		0.999962,
		0.999912,
		0.999825,
		0.99969,
		0.999493,
		0.999222,
		0.998863,
		0.998404,
		0.997833,
		0.997137,
		0.996306,
		0.995329,
		0.994196,
		0.992899,
		0.991431,
		0.989783,
		0.987949,
		0.985926,
		0.983707,
		0.981289,
		0.978669,
		0.975845,
		0.972816,
		0.96958,
		0.966137,
		0.962487,
		0.958631,
		0.954571,
		0.950308,
		0.945845,
		0.941183,
		0.936326,
		0.931276,
		0.926039,
		0.920616,
		0.915012,
		0.909232,
		0.90328,
		0.897159,
		0.890876,
		0.884434,
		0.877839,
		0.871096,
		0.864209,
		0.857183,
		0.850025,
		0.842738,
		0.835329,
		0.827802,
		0.820162,
		0.812415,
		0.804566,
		0.000342,
		0.005236,
		0.015366,
		0.030049,
		0.048665,
		0.070648,
		0.095481,
		0.122695,
		0.151865,
		0.182603,
		0.214561,
		0.247423,
		0.280903,
		0.314747,
		0.348725,
		0.382633,
		0.416289,
		0.449532,
		0.482221,
		0.51423,
		0.545454,
		0.575799,
		0.605186,
		0.63355,
		0.660837,
		0.687003,
		0.712016,
		0.73585,
		0.75849,
		0.779928,
		0.800163,
		0.8192,
		0.837049,
		0.853729,
		0.869259,
		0.883666,
		0.89698,
		0.909232,
		0.920461,
		0.930704,
		0.940003,
		0.948403,
		0.955947,
		0.962684,
		0.968661,
		0.973927,
		0.978531,
		0.982523,
		0.985952,
		0.988867,
		0.991318,
		0.99335,
		0.995012,
		0.996348,
		0.9974,
		0.99821,
		0.998817,
		0.999256,
		0.999561,
		0.999762,
		0.999884,
		0.999952,
		0.999985,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999987,
		0.999959,
		0.999905,
		0.999812,
		0.999666,
		0.999454,
		0.999163,
		0.998778,
		0.998286,
		0.997675,
		0.99693,
		0.996042,
		0.994999,
		0.993791,
		0.99241,
		0.990846,
		0.989094,
		0.987147,
		0.984999,
		0.982646,
		0.980084,
		0.977312,
		0.974325,
		0.971124,
		0.967708,
		0.964076,
		0.96023,
		0.95617,
		0.951898,
		0.947416,
		0.942727,
		0.937833,
		0.932739,
		0.927446,
		0.921961,
		0.916286,
		0.910426,
		0.904386,
		0.89817,
		0.891784,
		0.885232,
		0.878519,
		0.871652,
		0.864634,
		0.857473,
		0.850172,
		0.842738,
		0.835177,
		0.827492,
		0.819691,
		0.811778,
		0.803759,
		0.79564,
		0.787425,
		0.779121,
		0.770731,
		0.762263,
		0.003111,
		0.011912,
		0.025659,
		0.043674,
		0.065341,
		0.090096,
		0.117431,
		0.146884,
		0.178037,
		0.210511,
		0.243966,
		0.278096,
		0.312626,
		0.34731,
		0.381929,
		0.416289,
		0.450219,
		0.483569,
		0.516206,
		0.548017,
		0.578906,
		0.608789,
		0.637598,
		0.665277,
		0.691782,
		0.717079,
		0.741144,
		0.763962,
		0.785527,
		0.805839,
		0.824905,
		0.842738,
		0.859359,
		0.874792,
		0.889065,
		0.902212,
		0.914268,
		0.925275,
		0.935275,
		0.944313,
		0.952436,
		0.959694,
		0.966137,
		0.971816,
		0.976785,
		0.981094,
		0.984799,
		0.987949,
		0.990599,
		0.992798,
		0.994596,
		0.996042,
		0.997182,
		0.99806,
		0.998717,
		0.999193,
		0.999524,
		0.999741,
		0.999874,
		0.999948,
		0.999984,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999985,
		0.999956,
		0.999897,
		0.999797,
		0.999639,
		0.999412,
		0.999098,
		0.998685,
		0.998157,
		0.997501,
		0.996704,
		0.995754,
		0.99464,
		0.99335,
		0.991877,
		0.990212,
		0.988347,
		0.986276,
		0.983995,
		0.981498,
		0.978782,
		0.975845,
		0.972685,
		0.969301,
		0.965692,
		0.961859,
		0.957802,
		0.953525,
		0.949027,
		0.944313,
		0.939384,
		0.934245,
		0.928899,
		0.92335,
		0.917603,
		0.911662,
		0.905532,
		0.899218,
		0.892725,
		0.886059,
		0.879226,
		0.87223,
		0.865077,
		0.857774,
		0.850326,
		0.842738,
		0.835018,
		0.827169,
		0.8192,
		0.811114,
		0.802918,
		0.794618,
		0.78622,
		0.777728,
		0.769149,
		0.760489,
		0.751752,
		0.742944,
		0.73407,
		0.725136,
		0.716147,
		0.001462,
		0.008741,
		0.021402,
		0.038704,
		0.059973,
		0.084596,
		0.112019,
		0.14174,
		0.173305,
		0.206304,
		0.24037,
		0.275172,
		0.310415,
		0.345833,
		0.381194,
		0.416289,
		0.450936,
		0.484973,
		0.518262,
		0.550683,
		0.582134,
		0.612528,
		0.641793,
		0.669872,
		0.69672,
		0.722303,
		0.746597,
		0.769588,
		0.79127,
		0.811648,
		0.830729,
		0.848532,
		0.865077,
		0.880394,
		0.894513,
		0.907473,
		0.919314,
		0.930079,
		0.939815,
		0.948572,
		0.956401,
		0.963355,
		0.969489,
		0.974858,
		0.979518,
		0.983525,
		0.986936,
		0.989805,
		0.992187,
		0.994136,
		0.995704,
		0.99694,
		0.997893,
		0.998606,
		0.999123,
		0.999482,
		0.999719,
		0.999863,
		0.999944,
		0.999982,
		0.999996,
		1,
		1,
		1,
		0.999997,
		0.999984,
		0.999952,
		0.999889,
		0.99978,
		0.99961,
		0.999365,
		0.999027,
		0.998583,
		0.998016,
		0.997312,
		0.996457,
		0.99544,
		0.994248,
		0.99287,
		0.991297,
		0.989522,
		0.987535,
		0.985332,
		0.982906,
		0.980254,
		0.977373,
		0.97426,
		0.970913,
		0.967332,
		0.963517,
		0.959469,
		0.955189,
		0.950679,
		0.945942,
		0.94098,
		0.935798,
		0.930398,
		0.924786,
		0.918966,
		0.912942,
		0.90672,
		0.900305,
		0.893703,
		0.88692,
		0.87996,
		0.872831,
		0.865538,
		0.858088,
		0.850486,
		0.842738,
		0.834852,
		0.826833,
		0.818687,
		0.810421,
		0.80204,
		0.793551,
		0.784961,
		0.776274,
		0.767497,
		0.758636,
		0.749696,
		0.740684,
		0.731605,
		0.722464,
		0.713267,
		0.704019,
		0.694725,
		0.685391,
		0.676022,
		0.666622,
		0.000387,
		0.00592,
		0.017323,
		0.033782,
		0.05456,
		0.078987,
		0.106458,
		0.136427,
		0.168399,
		0.201932,
		0.236626,
		0.272124,
		0.308108,
		0.344293,
		0.380428,
		0.416289,
		0.451682,
		0.486437,
		0.520404,
		0.553458,
		0.58549,
		0.61641,
		0.646144,
		0.674631,
		0.701826,
		0.727694,
		0.752213,
		0.77537,
		0.797161,
		0.817592,
		0.836674,
		0.854428,
		0.870879,
		0.886059,
		0.900005,
		0.912756,
		0.924359,
		0.93486,
		0.944313,
		0.952769,
		0.960286,
		0.96692,
		0.97273,
		0.977776,
		0.982117,
		0.985814,
		0.988925,
		0.99151,
		0.993626,
		0.995329,
		0.996672,
		0.997707,
		0.998483,
		0.999046,
		0.999437,
		0.999694,
		0.999851,
		0.999939,
		0.999981,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999983,
		0.999948,
		0.99988,
		0.999762,
		0.999578,
		0.999313,
		0.998949,
		0.99847,
		0.99786,
		0.997104,
		0.996187,
		0.995096,
		0.993819,
		0.992346,
		0.990665,
		0.98877,
		0.986651,
		0.984305,
		0.981724,
		0.978905,
		0.975845,
		0.972543,
		0.968996,
		0.965204,
		0.961169,
		0.956891,
		0.952372,
		0.947615,
		0.942622,
		0.937398,
		0.931946,
		0.92627,
		0.920376,
		0.914268,
		0.907953,
		0.901435,
		0.89472,
		0.887815,
		0.880725,
		0.873458,
		0.866019,
		0.858415,
		0.850653,
		0.842738,
		0.834679,
		0.826481,
		0.818151,
		0.809696,
		0.801123,
		0.792437,
		0.783645,
		0.774754,
		0.76577,
		0.756699,
		0.747548,
		0.738322,
		0.729027,
		0.71967,
		0.710256,
		0.700791,
		0.69128,
		0.681729,
		0.672143,
		0.662527,
		0.652886,
		0.643226,
		0.63355,
		0.623865,
		0.614173,
		0.003529,
		0.013474,
		0.028941,
		0.049119,
		0.073275,
		0.100747,
		0.130939,
		0.163312,
		0.197385,
		0.232725,
		0.268945,
		0.3057,
		0.342684,
		0.379627,
		0.416289,
		0.452462,
		0.487964,
		0.522638,
		0.556348,
		0.588982,
		0.620445,
		0.650658,
		0.679561,
		0.707107,
		0.73326,
		0.758,
		0.781314,
		0.803203,
		0.823672,
		0.842738,
		0.860425,
		0.876762,
		0.891784,
		0.905532,
		0.918052,
		0.929393,
		0.939609,
		0.948756,
		0.956891,
		0.964076,
		0.970373,
		0.975845,
		0.980556,
		0.984569,
		0.987949,
		0.990759,
		0.99306,
		0.994912,
		0.996374,
		0.997501,
		0.998347,
		0.99896,
		0.999385,
		0.999666,
		0.999838,
		0.999933,
		0.999979,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999981,
		0.999944,
		0.999869,
		0.999741,
		0.999543,
		0.999256,
		0.998863,
		0.998347,
		0.997689,
		0.996876,
		0.99589,
		0.994719,
		0.99335,
		0.991772,
		0.989975,
		0.987949,
		0.985689,
		0.983187,
		0.980438,
		0.977439,
		0.974188,
		0.970681,
		0.96692,
		0.962903,
		0.958631,
		0.954108,
		0.949334,
		0.944313,
		0.939048,
		0.933544,
		0.927805,
		0.921837,
		0.915644,
		0.909232,
		0.902608,
		0.895777,
		0.888746,
		0.881522,
		0.874111,
		0.86652,
		0.858756,
		0.850827,
		0.842738,
		0.834498,
		0.826114,
		0.817592,
		0.808939,
		0.800163,
		0.79127,
		0.782268,
		0.773163,
		0.763962,
		0.754672,
		0.745299,
		0.73585,
		0.726331,
		0.716747,
		0.707107,
		0.697414,
		0.687676,
		0.677898,
		0.668086,
		0.658245,
		0.64838,
		0.638497,
		0.628601,
		0.618696,
		0.608789,
		0.598882,
		0.588982,
		0.579092,
		0.569217,
		0.559361,
		0.001665,
		0.009922,
		0.024222,
		0.043674,
		0.067471,
		0.094888,
		0.125272,
		0.158037,
		0.192655,
		0.228658,
		0.265625,
		0.303183,
		0.341002,
		0.37879,
		0.416289,
		0.453277,
		0.489559,
		0.524968,
		0.559361,
		0.592618,
		0.62464,
		0.655346,
		0.684672,
		0.71257,
		0.739008,
		0.763962,
		0.787425,
		0.809397,
		0.82989,
		0.848922,
		0.86652,
		0.882719,
		0.897559,
		0.911086,
		0.92335,
		0.934406,
		0.944313,
		0.953131,
		0.960924,
		0.967759,
		0.973702,
		0.978822,
		0.983187,
		0.986864,
		0.989923,
		0.992429,
		0.994448,
		0.996042,
		0.997272,
		0.998194,
		0.998863,
		0.999328,
		0.999635,
		0.999823,
		0.999927,
		0.999977,
		0.999995,
		1,
		1,
		1,
		0.999996,
		0.99998,
		0.999939,
		0.999858,
		0.999719,
		0.999504,
		0.999193,
		0.998768,
		0.99821,
		0.997501,
		0.996625,
		0.995564,
		0.994306,
		0.992836,
		0.991144,
		0.98922,
		0.987053,
		0.984638,
		0.981968,
		0.979038,
		0.975845,
		0.972387,
		0.968661,
		0.964669,
		0.960409,
		0.955885,
		0.951098,
		0.946052,
		0.940749,
		0.935195,
		0.929393,
		0.92335,
		0.917071,
		0.910561,
		0.903828,
		0.896877,
		0.889716,
		0.882352,
		0.874792,
		0.867043,
		0.859113,
		0.851009,
		0.842738,
		0.834309,
		0.825729,
		0.817006,
		0.808146,
		0.799158,
		0.790049,
		0.780826,
		0.771497,
		0.762069,
		0.752549,
		0.742944,
		0.73326,
		0.723506,
		0.713686,
		0.703808,
		0.693878,
		0.683903,
		0.673888,
		0.66384,
		0.653763,
		0.643665,
		0.63355,
		0.623424,
		0.613292,
		0.60316,
		0.593031,
		0.582912,
		0.572806,
		0.562719,
		0.552654,
		0.542616,
		0.532609,
		0.522638,
		0.512705,
		0.502815,
		0.000442,
		0.00674,
		0.019668,
		0.038245,
		0.061582,
		0.088879,
		0.11942,
		0.152562,
		0.187731,
		0.224414,
		0.262156,
		0.300551,
		0.339242,
		0.377913,
		0.416289,
		0.45413,
		0.491227,
		0.527402,
		0.562504,
		0.596407,
		0.629005,
		0.660215,
		0.689972,
		0.718226,
		0.744944,
		0.770107,
		0.793707,
		0.815748,
		0.836245,
		0.855222,
		0.872709,
		0.888746,
		0.903379,
		0.916658,
		0.928639,
		0.939384,
		0.948956,
		0.957423,
		0.964853,
		0.971319,
		0.976892,
		0.981647,
		0.985655,
		0.988991,
		0.991726,
		0.99393,
		0.995671,
		0.997015,
		0.998024,
		0.998756,
		0.999265,
		0.9996,
		0.999806,
		0.99992,
		0.999974,
		0.999995,
		1,
		1,
		1,
		0.999995,
		0.999978,
		0.999933,
		0.999845,
		0.999694,
		0.99946,
		0.999123,
		0.998663,
		0.99806,
		0.997294,
		0.996348,
		0.995205,
		0.993851,
		0.992272,
		0.990456,
		0.988393,
		0.986073,
		0.98349,
		0.980638,
		0.977512,
		0.974109,
		0.970427,
		0.966466,
		0.962225,
		0.957705,
		0.95291,
		0.947842,
		0.942504,
		0.9369,
		0.931037,
		0.924919,
		0.918552,
		0.911942,
		0.905097,
		0.898023,
		0.890728,
		0.883219,
		0.875503,
		0.86759,
		0.859486,
		0.851199,
		0.842738,
		0.834112,
		0.825327,
		0.816392,
		0.807315,
		0.798105,
		0.788769,
		0.779315,
		0.769751,
		0.760084,
		0.750323,
		0.740474,
		0.730545,
		0.720543,
		0.710476,
		0.70035,
		0.690172,
		0.679948,
		0.669685,
		0.65939,
		0.649069,
		0.638727,
		0.628371,
		0.618005,
		0.607637,
		0.59727,
		0.586911,
		0.576564,
		0.566235,
		0.555928,
		0.545648,
		0.535398,
		0.525185,
		0.515011,
		0.504881,
		0.494799,
		0.484769,
		0.474793,
		0.464877,
		0.455022,
		0.445233,
		0.004022,
		0.015327,
		0.032845,
		0.055596,
		0.082699,
		0.113359,
		0.146868,
		0.182595,
		0.21998,
		0.258526,
		0.297794,
		0.337397,
		0.376995,
		0.416289,
		0.455022,
		0.492972,
		0.529947,
		0.565786,
		0.600358,
		0.63355,
		0.665277,
		0.69547,
		0.724081,
		0.751077,
		0.776439,
		0.800163,
		0.822257,
		0.842738,
		0.861636,
		0.878987,
		0.894835,
		0.909232,
		0.922235,
		0.933906,
		0.944313,
		0.953525,
		0.961616,
		0.968661,
		0.974739,
		0.979927,
		0.984305,
		0.987949,
		0.990939,
		0.99335,
		0.995256,
		0.996728,
		0.997833,
		0.998635,
		0.999193,
		0.999561,
		0.999787,
		0.999912,
		0.999972,
		0.999994,
		1,
		1,
		1,
		0.999995,
		0.999976,
		0.999927,
		0.99983,
		0.999666,
		0.999412,
		0.999046,
		0.998546,
		0.997893,
		0.997064,
		0.996042,
		0.994809,
		0.99335,
		0.991651,
		0.9897,
		0.987485,
		0.984999,
		0.982234,
		0.979184,
		0.975845,
		0.972215,
		0.968292,
		0.964076,
		0.959568,
		0.954769,
		0.949683,
		0.944313,
		0.938663,
		0.932739,
		0.926546,
		0.92009,
		0.913379,
		0.906419,
		0.899218,
		0.891784,
		0.884124,
		0.876247,
		0.868161,
		0.859876,
		0.851398,
		0.842738,
		0.833904,
		0.824905,
		0.815748,
		0.806444,
		0.797,
		0.787425,
		0.777728,
		0.767917,
		0.758,
		0.747985,
		0.737881,
		0.727694,
		0.717434,
		0.707107,
		0.69672,
		0.686282,
		0.675798,
		0.665277,
		0.654724,
		0.644146,
		0.63355,
		0.622942,
		0.612327,
		0.601712,
		0.591103,
		0.580504,
		0.569922,
		0.559361,
		0.548826,
		0.538323,
		0.527856,
		0.51743,
		0.507048,
		0.496716,
		0.486437,
		0.476215,
		0.466054,
		0.455958,
		0.44593,
		0.435974,
		0.426093,
		0.416289,
		0.406567,
		0.396928,
		0.387376,
		0.001888,
		0.011259,
		0.027471,
		0.049465,
		0.076267,
		0.106998,
		0.140868,
		0.177175,
		0.215301,
		0.254699,
		0.294891,
		0.335457,
		0.376029,
		0.416289,
		0.455958,
		0.494799,
		0.532609,
		0.569217,
		0.604481,
		0.638287,
		0.670542,
		0.701178,
		0.730146,
		0.757413,
		0.782964,
		0.806797,
		0.828924,
		0.849368,
		0.868161,
		0.885348,
		0.900978,
		0.915109,
		0.927805,
		0.939136,
		0.949176,
		0.958001,
		0.965692,
		0.972331,
		0.978003,
		0.982792,
		0.986782,
		0.990058,
		0.9927,
		0.99479,
		0.996405,
		0.997618,
		0.998499,
		0.999113,
		0.999517,
		0.999765,
		0.999903,
		0.999969,
		0.999994,
		1,
		1,
		1,
		0.999994,
		0.999973,
		0.99992,
		0.999814,
		0.999635,
		0.999358,
		0.99896,
		0.998417,
		0.997707,
		0.99681,
		0.995704,
		0.994372,
		0.992798,
		0.990967,
		0.988867,
		0.986488,
		0.983819,
		0.980856,
		0.977591,
		0.974023,
		0.970147,
		0.965963,
		0.961473,
		0.956676,
		0.951577,
		0.946178,
		0.940484,
		0.934501,
		0.928233,
		0.921688,
		0.914874,
		0.907796,
		0.900464,
		0.892886,
		0.88507,
		0.877025,
		0.86876,
		0.860284,
		0.851607,
		0.842738,
		0.833687,
		0.824461,
		0.815072,
		0.805528,
		0.795839,
		0.786014,
		0.776061,
		0.76599,
		0.75581,
		0.745529,
		0.735156,
		0.724699,
		0.714166,
		0.703566,
		0.692907,
		0.682196,
		0.67144,
		0.660647,
		0.649825,
		0.638979,
		0.628118,
		0.617247,
		0.606372,
		0.595501,
		0.584639,
		0.573791,
		0.562964,
		0.552163,
		0.541394,
		0.530661,
		0.519969,
		0.509323,
		0.498728,
		0.488188,
		0.477707,
		0.46729,
		0.45694,
		0.446662,
		0.436458,
		0.426333,
		0.416289,
		0.406331,
		0.39646,
		0.386681,
		0.376995,
		0.367405,
		0.357915,
		0.348527,
		0.339242,
		0.330063,
		0.000494,
		0.007563,
		0.022135,
		0.043104,
		0.069425,
		0.100129,
		0.134336,
		0.171255,
		0.21019,
		0.250527,
		0.291735,
		0.333351,
		0.374977,
		0.416269,
		0.456932,
		0.496713,
		0.535398,
		0.572806,
		0.608789,
		0.643226,
		0.676022,
		0.707107,
		0.73643,
		0.763962,
		0.789689,
		0.813614,
		0.835752,
		0.856131,
		0.874792,
		0.891784,
		0.907163,
		0.920996,
		0.933353,
		0.944313,
		0.953955,
		0.962366,
		0.969633,
		0.975845,
		0.981094,
		0.985471,
		0.989067,
		0.991969,
		0.994266,
		0.996042,
		0.997377,
		0.998347,
		0.999022,
		0.999468,
		0.999741,
		0.999893,
		0.999966,
		0.999993,
		1,
		1,
		1,
		0.999994,
		0.999971,
		0.999912,
		0.999797,
		0.9996,
		0.999297,
		0.998863,
		0.998273,
		0.997501,
		0.996527,
		0.995329,
		0.993888,
		0.992187,
		0.990212,
		0.987949,
		0.985389,
		0.982523,
		0.979343,
		0.975845,
		0.972026,
		0.967884,
		0.963419,
		0.958631,
		0.953525,
		0.948101,
		0.942367,
		0.936326,
		0.929985,
		0.92335,
		0.91643,
		0.909232,
		0.901765,
		0.894038,
		0.886059,
		0.877839,
		0.869387,
		0.860713,
		0.851827,
		0.842738,
		0.833458,
		0.823996,
		0.814362,
		0.804566,
		0.794618,
		0.784529,
		0.774307,
		0.763962,
		0.753505,
		0.742944,
		0.732288,
		0.721547,
		0.710728,
		0.699842,
		0.688896,
		0.677898,
		0.666857,
		0.655781,
		0.644676,
		0.63355,
		0.622411,
		0.611266,
		0.60012,
		0.588982,
		0.577857,
		0.566751,
		0.555671,
		0.544621,
		0.533608,
		0.522638,
		0.511714,
		0.500843,
		0.490028,
		0.479275,
		0.468589,
		0.457972,
		0.44743,
		0.436966,
		0.426585,
		0.416289,
		0.406083,
		0.395969,
		0.385951,
		0.376031,
		0.366214,
		0.3565,
		0.346894,
		0.337397,
		0.328012,
		0.318742,
		0.309587,
		0.300551,
		0.291635,
		0.282841,
		0.27417,
		0.0044,
		0.016906,
		0.036449,
		0.061965,
		0.092422,
		0.126853,
		0.164372,
		0.204182,
		0.245582,
		0.287961,
		0.330792,
		0.373628,
		0.416091,
		0.457861,
		0.498671,
		0.538297,
		0.576554,
		0.613289,
		0.648379,
		0.681729,
		0.713267,
		0.742944,
		0.770731,
		0.796619,
		0.820615,
		0.842738,
		0.863025,
		0.881522,
		0.898284,
		0.913379,
		0.926878,
		0.938862,
		0.949417,
		0.958631,
		0.9666,
		0.973418,
		0.979184,
		0.983995,
		0.987949,
		0.991144,
		0.993675,
		0.995632,
		0.997104,
		0.998174,
		0.99892,
		0.999412,
		0.999714,
		0.999882,
		0.999962,
		0.999993,
		1,
		1,
		1,
		0.999993,
		0.999968,
		0.999903,
		0.999776,
		0.999561,
		0.99923,
		0.998756,
		0.998111,
		0.997272,
		0.996213,
		0.994912,
		0.99335,
		0.99151,
		0.989376,
		0.986936,
		0.984178,
		0.981094,
		0.977679,
		0.973927,
		0.969836,
		0.965405,
		0.960634,
		0.955526,
		0.950084,
		0.944313,
		0.938217,
		0.931803,
		0.925079,
		0.918052,
		0.910731,
		0.903125,
		0.895243,
		0.887096,
		0.878693,
		0.870045,
		0.861163,
		0.852057,
		0.842738,
		0.833218,
		0.823506,
		0.813614,
		0.803552,
		0.793332,
		0.782964,
		0.772459,
		0.761826,
		0.751077,
		0.74022,
		0.729267,
		0.718226,
		0.707107,
		0.695919,
		0.684672,
		0.673374,
		0.662033,
		0.650658,
		0.639258,
		0.627839,
		0.61641,
		0.604978,
		0.59355,
		0.582134,
		0.570735,
		0.559361,
		0.548017,
		0.536711,
		0.525446,
		0.51423,
		0.503068,
		0.491965,
		0.480925,
		0.469955,
		0.459058,
		0.448238,
		0.437501,
		0.42685,
		0.416289,
		0.405822,
		0.395453,
		0.385184,
		0.37502,
		0.364963,
		0.355015,
		0.345181,
		0.335463,
		0.325863,
		0.316383,
		0.307026,
		0.297795,
		0.28869,
		0.279714,
		0.270869,
		0.262156,
		0.253577,
		0.245133,
		0.236827,
		0.228658,
		0.220628,
		0.001976,
		0.011958,
		0.029542,
		0.053735,
		0.083512,
		0.11786,
		0.155812,
		0.19647,
		0.239021,
		0.282748,
		0.327031,
		0.371341,
		0.415238,
		0.458357,
		0.500403,
		0.541136,
		0.580364,
		0.617935,
		0.653732,
		0.687664,
		0.719666,
		0.749695,
		0.777728,
		0.803759,
		0.827802,
		0.849883,
		0.870045,
		0.888342,
		0.904839,
		0.91961,
		0.932739,
		0.944313,
		0.954428,
		0.963183,
		0.970681,
		0.977027,
		0.982327,
		0.986687,
		0.990212,
		0.993005,
		0.995167,
		0.996795,
		0.997978,
		0.998803,
		0.999348,
		0.999683,
		0.999869,
		0.999958,
		0.999992,
		0.999999,
		1,
		1,
		0.999993,
		0.999964,
		0.999893,
		0.999754,
		0.999517,
		0.999154,
		0.998635,
		0.997931,
		0.997015,
		0.995862,
		0.994448,
		0.992753,
		0.990759,
		0.98845,
		0.985814,
		0.982839,
		0.979518,
		0.975845,
		0.971816,
		0.967429,
		0.962684,
		0.957583,
		0.952128,
		0.946324,
		0.940176,
		0.933692,
		0.926878,
		0.919743,
		0.912296,
		0.904547,
		0.896506,
		0.888183,
		0.879589,
		0.870737,
		0.861636,
		0.8523,
		0.842738,
		0.832964,
		0.82299,
		0.812825,
		0.802484,
		0.791976,
		0.781314,
		0.770509,
		0.759572,
		0.748515,
		0.737347,
		0.726079,
		0.714722,
		0.703287,
		0.691782,
		0.680218,
		0.668603,
		0.656948,
		0.645261,
		0.63355,
		0.621825,
		0.610092,
		0.598361,
		0.586639,
		0.574932,
		0.563249,
		0.551596,
		0.539979,
		0.528406,
		0.516882,
		0.505413,
		0.494005,
		0.482664,
		0.471394,
		0.460201,
		0.449089,
		0.438064,
		0.427129,
		0.416289,
		0.405548,
		0.39491,
		0.384378,
		0.373956,
		0.363647,
		0.353455,
		0.343382,
		0.333432,
		0.323607,
		0.313909,
		0.304341,
		0.294906,
		0.285604,
		0.27644,
		0.267414,
		0.258527,
		0.249783,
		0.241182,
		0.232725,
		0.224414,
		0.216251,
		0.208236,
		0.200371,
		0.192655,
		0.185091,
		0.177679,
		0.170419,
		0.000483,
		0.007571,
		0.0226,
		0.044773,
		0.073179,
		0.106843,
		0.144773,
		0.185998,
		0.229604,
		0.274752,
		0.320699,
		0.366798,
		0.412507,
		0.457378,
		0.501051,
		0.543246,
		0.583747,
		0.622395,
		0.659076,
		0.69371,
		0.726248,
		0.756663,
		0.784947,
		0.811109,
		0.835175,
		0.857183,
		0.877185,
		0.895243,
		0.911434,
		0.925841,
		0.938558,
		0.949683,
		0.959323,
		0.967586,
		0.974587,
		0.980438,
		0.985256,
		0.989154,
		0.992245,
		0.99464,
		0.996443,
		0.997755,
		0.998671,
		0.999276,
		0.999648,
		0.999854,
		0.999954,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.999992,
		0.99996,
		0.999882,
		0.999728,
		0.999468,
		0.999069,
		0.998499,
		0.997729,
		0.996728,
		0.99547,
		0.99393,
		0.992087,
		0.989923,
		0.987421,
		0.984569,
		0.981357,
		0.977776,
		0.973821,
		0.969489,
		0.96478,
		0.959694,
		0.954233,
		0.948403,
		0.942207,
		0.935655,
		0.928752,
		0.921508,
		0.913932,
		0.906036,
		0.897829,
		0.889324,
		0.880531,
		0.871464,
		0.862134,
		0.852555,
		0.842738,
		0.832697,
		0.822445,
		0.811993,
		0.801356,
		0.790544,
		0.779572,
		0.76845,
		0.757191,
		0.745808,
		0.734311,
		0.722712,
		0.711022,
		0.699252,
		0.687413,
		0.675515,
		0.663568,
		0.651582,
		0.639566,
		0.62753,
		0.615483,
		0.603434,
		0.591389,
		0.579359,
		0.567351,
		0.555372,
		0.543429,
		0.531529,
		0.51968,
		0.507888,
		0.496159,
		0.484498,
		0.472913,
		0.461407,
		0.449987,
		0.438658,
		0.427424,
		0.416289,
		0.405259,
		0.394338,
		0.383529,
		0.372836,
		0.362263,
		0.351814,
		0.34149,
		0.331297,
		0.321236,
		0.31131,
		0.301522,
		0.291874,
		0.282369,
		0.273008,
		0.263794,
		0.254728,
		0.245813,
		0.237049,
		0.228439,
		0.219983,
		0.211684,
		0.203541,
		0.195557,
		0.187731,
		0.180066,
		0.172561,
		0.165218,
		0.158037,
		0.151018,
		0.144161,
		0.137468,
		0.130939,
		0.124573,
		0.004055,
		0.016015,
		0.0354,
		0.061533,
		0.093595,
		0.130667,
		0.171783,
		0.215973,
		0.262302,
		0.309899,
		0.357979,
		0.405858,
		0.452958,
		0.498808,
		0.543038,
		0.585368,
		0.625599,
		0.663597,
		0.699281,
		0.732617,
		0.763598,
		0.792245,
		0.818594,
		0.842698,
		0.864619,
		0.884429,
		0.90221,
		0.918052,
		0.932051,
		0.944313,
		0.954949,
		0.964076,
		0.971816,
		0.978292,
		0.983628,
		0.987949,
		0.991379,
		0.994038,
		0.996042,
		0.997501,
		0.99852,
		0.999193,
		0.999607,
		0.999838,
		0.999948,
		0.99999,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999956,
		0.999869,
		0.999699,
		0.999412,
		0.998972,
		0.998347,
		0.997501,
		0.996405,
		0.99503,
		0.99335,
		0.991344,
		0.988991,
		0.986276,
		0.983187,
		0.979712,
		0.975845,
		0.971582,
		0.96692,
		0.961859,
		0.956401,
		0.950551,
		0.944313,
		0.937694,
		0.930704,
		0.92335,
		0.915644,
		0.907596,
		0.899218,
		0.890522,
		0.881522,
		0.87223,
		0.862659,
		0.852824,
		0.842738,
		0.832415,
		0.821869,
		0.811114,
		0.800163,
		0.78903,
		0.777728,
		0.766271,
		0.754672,
		0.742944,
		0.731099,
		0.719149,
		0.707107,
		0.694984,
		0.682792,
		0.670542,
		0.658245,
		0.645911,
		0.63355,
		0.621173,
		0.608789,
		0.596407,
		0.584036,
		0.571684,
		0.559361,
		0.547074,
		0.53483,
		0.522638,
		0.510503,
		0.498434,
		0.486437,
		0.474517,
		0.462682,
		0.450936,
		0.439285,
		0.427734,
		0.416289,
		0.404954,
		0.393734,
		0.382633,
		0.371655,
		0.360804,
		0.350084,
		0.339498,
		0.32905,
		0.318742,
		0.308577,
		0.298559,
		0.28869,
		0.278972,
		0.269407,
		0.259999,
		0.250747,
		0.241656,
		0.232725,
		0.223957,
		0.215353,
		0.206915,
		0.198643,
		0.190539,
		0.182603,
		0.174837,
		0.167242,
		0.159817,
		0.152563,
		0.145482,
		0.138573,
		0.131836,
		0.125272,
		0.118882,
		0.112664,
		0.106619,
		0.100747,
		0.095049,
		0.089522,
		0.084169,
		0.001638,
		0.010267,
		0.026203,
		0.049109,
		0.078437,
		0.113465,
		0.153342,
		0.197141,
		0.243899,
		0.292662,
		0.342523,
		0.392648,
		0.442298,
		0.490845,
		0.537774,
		0.582681,
		0.62527,
		0.66534,
		0.702768,
		0.737504,
		0.769549,
		0.798942,
		0.825756,
		0.85008,
		0.872018,
		0.891682,
		0.909189,
		0.924657,
		0.938211,
		0.949977,
		0.960083,
		0.968661,
		0.975845,
		0.981771,
		0.986574,
		0.99039,
		0.99335,
		0.995583,
		0.99721,
		0.998347,
		0.999098,
		0.999561,
		0.999819,
		0.999942,
		0.999988,
		0.999999,
		1,
		0.999999,
		0.99999,
		0.999951,
		0.999854,
		0.999666,
		0.999348,
		0.998863,
		0.998174,
		0.997245,
		0.996042,
		0.994536,
		0.9927,
		0.990511,
		0.987949,
		0.984999,
		0.981647,
		0.977883,
		0.973702,
		0.9691,
		0.964076,
		0.958631,
		0.952769,
		0.946495,
		0.939815,
		0.932739,
		0.925275,
		0.917436,
		0.909232,
		0.900677,
		0.891784,
		0.882566,
		0.873037,
		0.863213,
		0.853109,
		0.842738,
		0.832117,
		0.82126,
		0.810183,
		0.798899,
		0.787425,
		0.775775,
		0.763962,
		0.752002,
		0.739908,
		0.727694,
		0.715374,
		0.702959,
		0.690463,
		0.677898,
		0.665277,
		0.65261,
		0.63991,
		0.627186,
		0.61445,
		0.601712,
		0.588982,
		0.576269,
		0.563582,
		0.550931,
		0.538323,
		0.525768,
		0.513271,
		0.500843,
		0.488488,
		0.476215,
		0.46403,
		0.451939,
		0.439948,
		0.428063,
		0.416289,
		0.404632,
		0.393097,
		0.381687,
		0.370409,
		0.359265,
		0.34826,
		0.337397,
		0.326681,
		0.316114,
		0.3057,
		0.295441,
		0.285341,
		0.275401,
		0.265625,
		0.256014,
		0.246571,
		0.237298,
		0.228195,
		0.219266,
		0.210511,
		0.201932,
		0.193529,
		0.185305,
		0.17726,
		0.169394,
		0.16171,
		0.154206,
		0.146884,
		0.139745,
		0.132788,
		0.126014,
		0.119423,
		0.113014,
		0.106789,
		0.100747,
		0.094888,
		0.089212,
		0.083718,
		0.078406,
		0.073275,
		0.068326,
		0.063557,
		0.058969,
		0.05456,
		0.05033,
		0.000352,
		0.005759,
		0.017889,
		0.036791,
		0.062266,
		0.093895,
		0.131066,
		0.173013,
		0.218859,
		0.267665,
		0.318468,
		0.370327,
		0.422357,
		0.473757,
		0.523831,
		0.572,
		0.617806,
		0.660911,
		0.701088,
		0.73821,
		0.772234,
		0.803188,
		0.831152,
		0.856247,
		0.878618,
		0.898429,
		0.915849,
		0.931051,
		0.944206,
		0.955481,
		0.965039,
		0.973042,
		0.979644,
		0.984999,
		0.989256,
		0.992561,
		0.995056,
		0.996876,
		0.998148,
		0.998989,
		0.999508,
		0.999797,
		0.999935,
		0.999987,
		0.999999,
		1,
		0.999999,
		0.999988,
		0.999945,
		0.999838,
		0.999628,
		0.999276,
		0.998739,
		0.997978,
		0.996954,
		0.995632,
		0.99398,
		0.991969,
		0.989577,
		0.986782,
		0.98357,
		0.979927,
		0.975845,
		0.971319,
		0.966345,
		0.960924,
		0.95506,
		0.948756,
		0.94202,
		0.93486,
		0.927288,
		0.919314,
		0.91095,
		0.902212,
		0.893112,
		0.883666,
		0.87389,
		0.863799,
		0.85341,
		0.842738,
		0.831801,
		0.820615,
		0.809195,
		0.797559,
		0.785723,
		0.773702,
		0.761512,
		0.749168,
		0.736686,
		0.724081,
		0.711367,
		0.698557,
		0.685666,
		0.672708,
		0.659694,
		0.646637,
		0.63355,
		0.620445,
		0.607332,
		0.594222,
		0.581127,
		0.568056,
		0.55502,
		0.542026,
		0.529086,
		0.516206,
		0.503396,
		0.490663,
		0.478015,
		0.465459,
		0.453002,
		0.44065,
		0.428411,
		0.416289,
		0.404291,
		0.392422,
		0.380687,
		0.369091,
		0.357638,
		0.346332,
		0.335179,
		0.324181,
		0.313342,
		0.302666,
		0.292156,
		0.281814,
		0.271644,
		0.261647,
		0.251827,
		0.242186,
		0.232725,
		0.223446,
		0.214352,
		0.205443,
		0.196721,
		0.188187,
		0.179843,
		0.171689,
		0.163726,
		0.155955,
		0.148377,
		0.140991,
		0.133799,
		0.126801,
		0.119997,
		0.113386,
		0.10697,
		0.100747,
		0.094719,
		0.088884,
		0.083242,
		0.077793,
		0.072536,
		0.067471,
		0.062598,
		0.057915,
		0.053423,
		0.049119,
		0.045004,
		0.041077,
		0.037337,
		0.033782,
		0.030412,
		0.027226,
		0.024223,
		0.002672,
		0.011064,
		0.025572,
		0.046363,
		0.073371,
		0.106308,
		0.144681,
		0.187825,
		0.234937,
		0.285117,
		0.337415,
		0.390871,
		0.444555,
		0.497602,
		0.549241,
		0.598814,
		0.645788,
		0.689759,
		0.730447,
		0.767694,
		0.801442,
		0.831725,
		0.858647,
		0.882368,
		0.903084,
		0.921018,
		0.936403,
		0.949475,
		0.960466,
		0.969601,
		0.977092,
		0.983139,
		0.987932,
		0.991646,
		0.994447,
		0.996489,
		0.997917,
		0.998863,
		0.999446,
		0.999771,
		0.999927,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999987,
		0.999939,
		0.999819,
		0.999585,
		0.999193,
		0.998598,
		0.997755,
		0.996625,
		0.995167,
		0.99335,
		0.991144,
		0.988525,
		0.985471,
		0.981968,
		0.978003,
		0.973569,
		0.968661,
		0.963278,
		0.957423,
		0.951098,
		0.944313,
		0.937074,
		0.929393,
		0.921283,
		0.912756,
		0.903828,
		0.894513,
		0.884829,
		0.874792,
		0.864419,
		0.853729,
		0.842738,
		0.831466,
		0.819929,
		0.808146,
		0.796135,
		0.783913,
		0.771497,
		0.758906,
		0.746155,
		0.73326,
		0.720239,
		0.707107,
		0.693878,
		0.680569,
		0.667192,
		0.653763,
		0.640295,
		0.6268,
		0.613292,
		0.599783,
		0.586284,
		0.572806,
		0.559361,
		0.545959,
		0.532609,
		0.519322,
		0.506107,
		0.492972,
		0.479925,
		0.466975,
		0.45413,
		0.441396,
		0.42878,
		0.416289,
		0.40393,
		0.391707,
		0.379627,
		0.367695,
		0.355915,
		0.344293,
		0.332833,
		0.321539,
		0.310415,
		0.299464,
		0.28869,
		0.278096,
		0.267685,
		0.25746,
		0.247423,
		0.237576,
		0.227922,
		0.218463,
		0.2092,
		0.200135,
		0.191269,
		0.182603,
		0.17414,
		0.165879,
		0.157821,
		0.149968,
		0.14232,
		0.134877,
		0.127639,
		0.120607,
		0.113781,
		0.107161,
		0.100747,
		0.094539,
		0.088536,
		0.082738,
		0.077145,
		0.071756,
		0.066571,
		0.061589,
		0.056809,
		0.05223,
		0.047852,
		0.043674,
		0.039695,
		0.035914,
		0.03233,
		0.028941,
		0.025748,
		0.022747,
		0.019939,
		0.017323,
		0.014896,
		0.012657,
		0.010606,
		0.008741,
		0.00706,
		0.000913,
		0.00605,
		0.01627,
		0.032043,
		0.053638,
		0.08111,
		0.1143,
		0.152846,
		0.196197,
		0.243645,
		0.294361,
		0.347429,
		0.401894,
		0.456801,
		0.511233,
		0.564348,
		0.615402,
		0.663773,
		0.708972,
		0.750642,
		0.788563,
		0.822636,
		0.852871,
		0.879373,
		0.90232,
		0.921945,
		0.938519,
		0.952336,
		0.963696,
		0.9729,
		0.980234,
		0.98597,
		0.990359,
		0.993631,
		0.995994,
		0.997633,
		0.998711,
		0.999373,
		0.999741,
		0.999917,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999985,
		0.999931,
		0.999797,
		0.999535,
		0.999098,
		0.998436,
		0.997501,
		0.996249,
		0.99464,
		0.992637,
		0.990212,
		0.987338,
		0.983995,
		0.980168,
		0.975845,
		0.971021,
		0.965692,
		0.959858,
		0.953525,
		0.946697,
		0.939384,
		0.931598,
		0.92335,
		0.914656,
		0.905532,
		0.895994,
		0.886059,
		0.875748,
		0.865077,
		0.854068,
		0.842738,
		0.831109,
		0.8192,
		0.807029,
		0.794618,
		0.781985,
		0.769149,
		0.756129,
		0.742944,
		0.72961,
		0.716147,
		0.70257,
		0.688896,
		0.675142,
		0.661323,
		0.647454,
		0.63355,
		0.619625,
		0.605693,
		0.591766,
		0.577857,
		0.563978,
		0.550142,
		0.536358,
		0.522638,
		0.508991,
		0.495428,
		0.481957,
		0.468589,
		0.455329,
		0.442188,
		0.429172,
		0.416289,
		0.403546,
		0.390948,
		0.378502,
		0.366214,
		0.354089,
		0.342132,
		0.330348,
		0.318742,
		0.307317,
		0.296078,
		0.285028,
		0.27417,
		0.263508,
		0.253045,
		0.242783,
		0.232725,
		0.222873,
		0.213228,
		0.203793,
		0.19457,
		0.185559,
		0.176763,
		0.168182,
		0.159817,
		0.151669,
		0.143738,
		0.136026,
		0.128533,
		0.121258,
		0.114202,
		0.107365,
		0.100747,
		0.094348,
		0.088168,
		0.082205,
		0.076461,
		0.070933,
		0.065621,
		0.060526,
		0.055645,
		0.050979,
		0.046525,
		0.042284,
		0.038254,
		0.034434,
		0.030823,
		0.02742,
		0.024223,
		0.021232,
		0.018444,
		0.015859,
		0.013475,
		0.01129,
		0.009304,
		0.007514,
		0.00592,
		0.004519,
		0.00331,
		0.002292,
		0.001462,
		0.00082,
		0.000363,
		9.1e-005,
		0.000161,
		0.002816,
		0.009308,
		0.020295,
		0.036295,
		0.057659,
		0.084551,
		0.116937,
		0.154586,
		0.197077,
		0.243819,
		0.294076,
		0.347001,
		0.401677,
		0.45715,
		0.512479,
		0.566765,
		0.61919,
		0.669042,
		0.715733,
		0.758816,
		0.797981,
		0.83306,
		0.864009,
		0.890903,
		0.913908,
		0.93327,
		0.949288,
		0.962301,
		0.972665,
		0.980739,
		0.986877,
		0.991411,
		0.994648,
		0.996864,
		0.998305,
		0.999178,
		0.999659,
		0.999888,
		0.999976,
		0.999997,
		1,
		0.999999,
		0.999984,
		0.999922,
		0.999771,
		0.999478,
		0.998989,
		0.998251,
		0.99721,
		0.99582,
		0.994038,
		0.991827,
		0.989154,
		0.985994,
		0.982327,
		0.978139,
		0.973418,
		0.968161,
		0.962366,
		0.956035,
		0.949176,
		0.941795,
		0.933906,
		0.925522,
		0.916658,
		0.907331,
		0.897559,
		0.887363,
		0.876762,
		0.865776,
		0.854428,
		0.842738,
		0.830729,
		0.818422,
		0.805839,
		0.793,
		0.779928,
		0.766643,
		0.753166,
		0.739516,
		0.725714,
		0.711778,
		0.697728,
		0.683581,
		0.669354,
		0.655065,
		0.64073,
		0.626365,
		0.611985,
		0.597605,
		0.583238,
		0.568899,
		0.5546,
		0.540354,
		0.526172,
		0.512066,
		0.498046,
		0.484124,
		0.470308,
		0.456608,
		0.443032,
		0.42959,
		0.416289,
		0.403137,
		0.39014,
		0.377306,
		0.36464,
		0.352149,
		0.339837,
		0.327712,
		0.315776,
		0.304035,
		0.292492,
		0.281153,
		0.27002,
		0.259096,
		0.248386,
		0.237891,
		0.227614,
		0.217558,
		0.207724,
		0.198115,
		0.188732,
		0.179577,
		0.170651,
		0.161955,
		0.15349,
		0.145256,
		0.137255,
		0.129487,
		0.121952,
		0.114651,
		0.107582,
		0.100747,
		0.094146,
		0.087777,
		0.08164,
		0.075735,
		0.070061,
		0.064618,
		0.059405,
		0.054421,
		0.049664,
		0.045134,
		0.04083,
		0.036751,
		0.032895,
		0.029261,
		0.025848,
		0.022654,
		0.019678,
		0.016918,
		0.014374,
		0.012042,
		0.009923,
		0.008014,
		0.006313,
		0.004818,
		0.003529,
		0.002443,
		0.001559,
		0.000874,
		0.000387,
		9.6e-005,
		0.001043,
		0.004664,
		0.011571,
		0.022412,
		0.037732,
		0.057942,
		0.083299,
		0.113883,
		0.149594,
		0.190145,
		0.235077,
		0.283774,
		0.33549,
		0.389377,
		0.444525,
		0.5,
		0.55488,
		0.608295,
		0.659459,
		0.707694,
		0.75245,
		0.793321,
		0.830044,
		0.862495,
		0.890684,
		0.914735,
		0.934871,
		0.951392,
		0.964651,
		0.975038,
		0.982956,
		0.988808,
		0.992978,
		0.995825,
		0.997669,
		0.998789,
		0.999416,
		0.999737,
		0.999885,
		0.999952,
		0.999981,
		0.999976,
		0.999911,
		0.999741,
		0.999412,
		0.998863,
		0.998037,
		0.996876,
		0.995329,
		0.99335,
		0.990902,
		0.987949,
		0.984468,
		0.980438,
		0.975845,
		0.970681,
		0.964943,
		0.958631,
		0.951752,
		0.944313,
		0.936326,
		0.927805,
		0.918768,
		0.909232,
		0.899218,
		0.888746,
		0.877839,
		0.86652,
		0.854812,
		0.842738,
		0.830324,
		0.817592,
		0.804566,
		0.79127,
		0.777728,
		0.763962,
		0.749996,
		0.73585,
		0.721547,
		0.707107,
		0.692551,
		0.677898,
		0.663169,
		0.64838,
		0.63355,
		0.618696,
		0.603835,
		0.588982,
		0.574153,
		0.559361,
		0.544621,
		0.529947,
		0.51535,
		0.500843,
		0.486437,
		0.472143,
		0.457972,
		0.443933,
		0.430036,
		0.416289,
		0.402701,
		0.389279,
		0.376031,
		0.362964,
		0.350084,
		0.337397,
		0.324909,
		0.312626,
		0.300551,
		0.28869,
		0.277047,
		0.265625,
		0.254429,
		0.243461,
		0.232725,
		0.222223,
		0.211958,
		0.201932,
		0.192146,
		0.182603,
		0.173305,
		0.164251,
		0.155444,
		0.146884,
		0.138573,
		0.130509,
		0.122695,
		0.11513,
		0.107814,
		0.100747,
		0.09393,
		0.08736,
		0.081039,
		0.074965,
		0.069138,
		0.063557,
		0.058222,
		0.05313,
		0.048281,
		0.043674,
		0.039308,
		0.035182,
		0.031293,
		0.027641,
		0.024223,
		0.02104,
		0.018088,
		0.015366,
		0.012873,
		0.010606,
		0.008564,
		0.006746,
		0.005149,
		0.003771,
		0.00261,
		0.001665,
		0.000934,
		0.000414,
		0.000103,
		0.000268,
		0.001962,
		0.005777,
		0.012361,
		0.02233,
		0.036236,
		0.054533,
		0.077548,
		0.105461,
		0.138292,
		0.175889,
		0.217932,
		0.263942,
		0.313298,
		0.365257,
		0.418991,
		0.473615,
		0.528227,
		0.581942,
		0.633929,
		0.683445,
		0.729854,
		0.77265,
		0.811468,
		0.846087,
		0.876427,
		0.902538,
		0.924584,
		0.942826,
		0.957597,
		0.969283,
		0.978299,
		0.985067,
		0.99,
		0.993485,
		0.995868,
		0.99745,
		0.998473,
		0.999123,
		0.999522,
		0.999737,
		0.999787,
		0.999659,
		0.999317,
		0.998711,
		0.997788,
		0.996489,
		0.994763,
		0.992561,
		0.989843,
		0.986574,
		0.98273,
		0.978292,
		0.973246,
		0.967586,
		0.961312,
		0.954428,
		0.94694,
		0.938862,
		0.930208,
		0.920996,
		0.911245,
		0.900978,
		0.890217,
		0.878987,
		0.867313,
		0.855222,
		0.842738,
		0.82989,
		0.816702,
		0.803203,
		0.789417,
		0.77537,
		0.761088,
		0.746597,
		0.731919,
		0.717079,
		0.7021,
		0.687003,
		0.671812,
		0.656545,
		0.641225,
		0.625869,
		0.610497,
		0.595126,
		0.579774,
		0.564456,
		0.549189,
		0.533987,
		0.518865,
		0.503836,
		0.488913,
		0.474108,
		0.459432,
		0.444897,
		0.430513,
		0.416289,
		0.402235,
		0.38836,
		0.374671,
		0.361176,
		0.347883,
		0.334797,
		0.321925,
		0.309274,
		0.296847,
		0.28465,
		0.272688,
		0.260965,
		0.249484,
		0.238249,
		0.227264,
		0.21653,
		0.206051,
		0.19583,
		0.185867,
		0.176164,
		0.166724,
		0.157547,
		0.148635,
		0.139988,
		0.131607,
		0.123492,
		0.115644,
		0.108062,
		0.100747,
		0.093699,
		0.086916,
		0.080399,
		0.074147,
		0.068159,
		0.062433,
		0.05697,
		0.051768,
		0.046826,
		0.042142,
		0.037715,
		0.033544,
		0.029626,
		0.025962,
		0.022547,
		0.019382,
		0.016464,
		0.013791,
		0.011362,
		0.009174,
		0.007225,
		0.005514,
		0.004038,
		0.002795,
		0.001783,
		0.000999,
		0.000443,
		0.00011,
		3.2e-005,
		0.000644,
		0.002411,
		0.005877,
		0.011614,
		0.020197,
		0.032163,
		0.047987,
		0.068049,
		0.092608,
		0.121789,
		0.155564,
		0.193752,
		0.236018,
		0.281881,
		0.330732,
		0.381856,
		0.434461,
		0.487707,
		0.540744,
		0.592741,
		0.642924,
		0.690604,
		0.735198,
		0.77625,
		0.813438,
		0.84658,
		0.875627,
		0.900656,
		0.921848,
		0.939475,
		0.953873,
		0.96542,
		0.974514,
		0.98155,
		0.986903,
		0.990912,
		0.993865,
		0.995988,
		0.99745,
		0.998361,
		0.998789,
		0.998768,
		0.998305,
		0.997388,
		0.995994,
		0.994091,
		0.991646,
		0.988624,
		0.984999,
		0.980744,
		0.975845,
		0.970291,
		0.964076,
		0.957205,
		0.949683,
		0.941522,
		0.932739,
		0.92335,
		0.913379,
		0.902848,
		0.891784,
		0.880212,
		0.868161,
		0.855661,
		0.842738,
		0.829424,
		0.815748,
		0.801739,
		0.787425,
		0.772836,
		0.758,
		0.742944,
		0.727694,
		0.712278,
		0.69672,
		0.681045,
		0.665277,
		0.649438,
		0.63355,
		0.617635,
		0.601712,
		0.585802,
		0.569922,
		0.55409,
		0.538323,
		0.522638,
		0.507048,
		0.491569,
		0.476215,
		0.460998,
		0.44593,
		0.431024,
		0.416289,
		0.401737,
		0.387376,
		0.373216,
		0.359265,
		0.345531,
		0.332021,
		0.318742,
		0.3057,
		0.292901,
		0.280351,
		0.268054,
		0.256014,
		0.244237,
		0.232725,
		0.221482,
		0.210511,
		0.199814,
		0.189395,
		0.179254,
		0.169394,
		0.159817,
		0.150522,
		0.141513,
		0.132788,
		0.124349,
		0.116196,
		0.108329,
		0.100747,
		0.093452,
		0.086442,
		0.079717,
		0.073275,
		0.067117,
		0.061241,
		0.055645,
		0.05033,
		0.045292,
		0.040531,
		0.036046,
		0.031834,
		0.027894,
		0.024223,
		0.020821,
		0.017685,
		0.014812,
		0.012202,
		0.009851,
		0.007758,
		0.00592,
		0.004334,
		0.003,
		0.001913,
		0.001072,
		0.000475,
		0.000118,
		0.000143,
		0.000774,
		0.002263,
		0.005051,
		0.009635,
		0.016531,
		0.02625,
		0.039266,
		0.055989,
		0.076739,
		0.101724,
		0.131026,
		0.164591,
		0.202223,
		0.24359,
		0.288229,
		0.335565,
		0.38493,
		0.435589,
		0.486772,
		0.537699,
		0.587617,
		0.635826,
		0.681705,
		0.724736,
		0.764515,
		0.800764,
		0.833327,
		0.86217,
		0.887367,
		0.909081,
		0.927549,
		0.943055,
		0.955914,
		0.966441,
		0.974941,
		0.981686,
		0.986903,
		0.990785,
		0.993485,
		0.995131,
		0.995825,
		0.995644,
		0.994648,
		0.992877,
		0.990359,
		0.98711,
		0.983139,
		0.978449,
		0.973042,
		0.966918,
		0.960083,
		0.952544,
		0.944313,
		0.935405,
		0.925841,
		0.915644,
		0.904839,
		0.893455,
		0.881522,
		0.86907,
		0.856131,
		0.842738,
		0.828924,
		0.814721,
		0.800163,
		0.785281,
		0.770107,
		0.754672,
		0.739008,
		0.723143,
		0.707107,
		0.690927,
		0.674631,
		0.658245,
		0.641793,
		0.6253,
		0.608789,
		0.592281,
		0.575799,
		0.559361,
		0.542987,
		0.526696,
		0.510503,
		0.494427,
		0.478481,
		0.462682,
		0.447041,
		0.431573,
		0.416289,
		0.401201,
		0.38632,
		0.371655,
		0.357216,
		0.343012,
		0.32905,
		0.315337,
		0.301882,
		0.28869,
		0.275766,
		0.263117,
		0.250747,
		0.238661,
		0.226862,
		0.215353,
		0.204139,
		0.193222,
		0.182603,
		0.172287,
		0.162273,
		0.152563,
		0.14316,
		0.134062,
		0.125272,
		0.11679,
		0.108615,
		0.100747,
		0.093187,
		0.085934,
		0.078987,
		0.072345,
		0.066007,
		0.059973,
		0.05424,
		0.048808,
		0.043674,
		0.038838,
		0.034297,
		0.030049,
		0.026093,
		0.022426,
		0.019046,
		0.015951,
		0.013138,
		0.010606,
		0.008351,
		0.006372,
		0.004665,
		0.003228,
		0.002059,
		0.001154,
		0.000511,
		0.000127,
		1.6e-005,
		0.000161,
		0.000623,
		0.00167,
		0.003651,
		0.006974,
		0.012091,
		0.019471,
		0.029569,
		0.042807,
		0.05954,
		0.080042,
		0.104482,
		0.132916,
		0.165276,
		0.201372,
		0.240889,
		0.283404,
		0.328396,
		0.375266,
		0.423361,
		0.472002,
		0.520504,
		0.568211,
		0.614514,
		0.658875,
		0.700845,
		0.74007,
		0.776299,
		0.809382,
		0.839261,
		0.865957,
		0.889556,
		0.910191,
		0.928019,
		0.943206,
		0.955914,
		0.966297,
		0.974514,
		0.980718,
		0.985067,
		0.987715,
		0.988808,
		0.988486,
		0.986877,
		0.984094,
		0.980234,
		0.97538,
		0.969601,
		0.962953,
		0.955481,
		0.947222,
		0.938211,
		0.928478,
		0.918052,
		0.906963,
		0.895243,
		0.882926,
		0.870045,
		0.856637,
		0.842738,
		0.828385,
		0.813614,
		0.798462,
		0.782964,
		0.767158,
		0.751077,
		0.734755,
		0.718226,
		0.701521,
		0.684672,
		0.667708,
		0.650658,
		0.63355,
		0.61641,
		0.599263,
		0.582134,
		0.565045,
		0.548017,
		0.531073,
		0.51423,
		0.497509,
		0.480925,
		0.464497,
		0.448238,
		0.432165,
		0.416289,
		0.400625,
		0.385184,
		0.369978,
		0.355015,
		0.340308,
		0.325863,
		0.311689,
		0.297795,
		0.284186,
		0.270869,
		0.25785,
		0.245133,
		0.232725,
		0.220628,
		0.208847,
		0.197385,
		0.186245,
		0.175429,
		0.164939,
		0.154777,
		0.144944,
		0.135442,
		0.126271,
		0.117431,
		0.108924,
		0.100747,
		0.092903,
		0.085389,
		0.078205,
		0.07135,
		0.064824,
		0.058623,
		0.052748,
		0.047197,
		0.041967,
		0.037057,
		0.032464,
		0.028187,
		0.024223,
		0.02057,
		0.017226,
		0.014187,
		0.011451,
		0.009016,
		0.006878,
		0.005035,
		0.003484,
		0.002221,
		0.001245,
		0.000551,
		0.000137,
		1.4e-005,
		9.1e-005,
		0.000339,
		0.000938,
		0.002141,
		0.004271,
		0.007705,
		0.012854,
		0.020142,
		0.02998,
		0.04274,
		0.058738,
		0.078208,
		0.101297,
		0.128047,
		0.158393,
		0.192162,
		0.229082,
		0.268784,
		0.310826,
		0.354699,
		0.399855,
		0.445728,
		0.491749,
		0.537373,
		0.582095,
		0.625459,
		0.667076,
		0.706621,
		0.743835,
		0.77852,
		0.810529,
		0.839755,
		0.866118,
		0.889556,
		0.910034,
		0.927549,
		0.942136,
		0.953873,
		0.962873,
		0.969283,
		0.973275,
		0.975038,
		0.974769,
		0.972665,
		0.968915,
		0.963696,
		0.957169,
		0.949475,
		0.940734,
		0.931051,
		0.920512,
		0.909189,
		0.897143,
		0.884429,
		0.871094,
		0.857183,
		0.842738,
		0.827802,
		0.812415,
		0.796619,
		0.780455,
		0.763962,
		0.74718,
		0.730146,
		0.712898,
		0.69547,
		0.677898,
		0.660215,
		0.642452,
		0.62464,
		0.606807,
		0.588982,
		0.571191,
		0.553458,
		0.535808,
		0.518262,
		0.500843,
		0.483569,
		0.466459,
		0.449532,
		0.432804,
		0.416289,
		0.400003,
		0.383959,
		0.368169,
		0.352645,
		0.337397,
		0.322436,
		0.30777,
		0.293408,
		0.279358,
		0.265625,
		0.252217,
		0.239139,
		0.226396,
		0.213992,
		0.201932,
		0.190218,
		0.178855,
		0.167843,
		0.157186,
		0.146884,
		0.13694,
		0.127354,
		0.118126,
		0.109257,
		0.100747,
		0.092596,
		0.084802,
		0.077365,
		0.070284,
		0.063557,
		0.057184,
		0.051161,
		0.045488,
		0.040162,
		0.035182,
		0.030544,
		0.026246,
		0.022285,
		0.01866,
		0.015366,
		0.012401,
		0.009762,
		0.007447,
		0.00545,
		0.003771,
		0.002404,
		0.001347,
		0.000596,
		0.000148,
		2.1e-005,
		0.000106,
		0.000355,
		0.000936,
		0.002079,
		0.00408,
		0.007282,
		0.012061,
		0.018798,
		0.027863,
		0.039592,
		0.054265,
		0.072094,
		0.09321,
		0.117656,
		0.145385,
		0.176263,
		0.210073,
		0.246526,
		0.285274,
		0.325921,
		0.368043,
		0.411201,
		0.454953,
		0.498867,
		0.542528,
		0.585544,
		0.627545,
		0.668179,
		0.707107,
		0.743996,
		0.77852,
		0.810366,
		0.839261,
		0.864981,
		0.887367,
		0.906329,
		0.921848,
		0.933976,
		0.942826,
		0.948562,
		0.951392,
		0.95155,
		0.949288,
		0.944861,
		0.938519,
		0.930499,
		0.921018,
		0.910271,
		0.898429,
		0.885637,
		0.872018,
		0.857677,
		0.842698,
		0.827155,
		0.811109,
		0.794617,
		0.777728,
		0.760489,
		0.742944,
		0.725136,
		0.707107,
		0.688896,
		0.670542,
		0.652082,
		0.63355,
		0.614981,
		0.596407,
		0.577857,
		0.559361,
		0.540946,
		0.522638,
		0.50446,
		0.486437,
		0.468589,
		0.450936,
		0.433497,
		0.416289,
		0.39933,
		0.382633,
		0.366214,
		0.350084,
		0.334256,
		0.318742,
		0.30355,
		0.28869,
		0.27417,
		0.259999,
		0.246181,
		0.232725,
		0.219635,
		0.206915,
		0.19457,
		0.182603,
		0.171018,
		0.159817,
		0.149001,
		0.138573,
		0.128533,
		0.118882,
		0.10962,
		0.100747,
		0.092264,
		0.084169,
		0.076461,
		0.069138,
		0.0622,
		0.055645,
		0.049471,
		0.043674,
		0.038254,
		0.033208,
		0.028532,
		0.024223,
		0.02028,
		0.016698,
		0.013475,
		0.010606,
		0.008089,
		0.00592,
		0.004095,
		0.00261,
		0.001462,
		0.000647,
		0.000161,
		6.4e-005,
		0.000248,
		0.000703,
		0.001636,
		0.003312,
		0.006041,
		0.01016,
		0.016015,
		0.023941,
		0.034243,
		0.047176,
		0.062938,
		0.081659,
		0.103399,
		0.128143,
		0.155812,
		0.186264,
		0.219306,
		0.254699,
		0.292176,
		0.33144,
		0.372177,
		0.414057,
		0.456733,
		0.499835,
		0.542966,
		0.585694,
		0.627545,
		0.668022,
		0.706621,
		0.742861,
		0.776299,
		0.806556,
		0.833327,
		0.856393,
		0.875627,
		0.890993,
		0.902538,
		0.910389,
		0.914735,
		0.915816,
		0.913908,
		0.909308,
		0.90232,
		0.893244,
		0.882368,
		0.869955,
		0.856247,
		0.841453,
		0.825756,
		0.80931,
		0.792245,
		0.774666,
		0.756663,
		0.738309,
		0.719666,
		0.70079,
		0.681729,
		0.662527,
		0.643226,
		0.623865,
		0.604481,
		0.585111,
		0.565786,
		0.54654,
		0.527402,
		0.5084,
		0.489559,
		0.470906,
		0.452462,
		0.43425,
		0.416289,
		0.398598,
		0.381194,
		0.364093,
		0.34731,
		0.330857,
		0.314747,
		0.298992,
		0.283601,
		0.268583,
		0.253947,
		0.2397,
		0.225849,
		0.212399,
		0.199356,
		0.186723,
		0.174504,
		0.162702,
		0.151319,
		0.140358,
		0.12982,
		0.119705,
		0.110014,
		0.100747,
		0.091904,
		0.083483,
		0.075484,
		0.067904,
		0.060743,
		0.053998,
		0.047667,
		0.041747,
		0.036235,
		0.031129,
		0.026425,
		0.022121,
		0.018211,
		0.014693,
		0.011564,
		0.008818,
		0.006452,
		0.004462,
		0.002844,
		0.001593,
		0.000705,
		0.000176,
		2e-005,
		0.000116,
		0.000402,
		0.001055,
		0.002303,
		0.004423,
		0.007717,
		0.012498,
		0.01907,
		0.027714,
		0.038672,
		0.052138,
		0.068256,
		0.087112,
		0.108744,
		0.13314,
		0.160248,
		0.189978,
		0.222209,
		0.256786,
		0.293526,
		0.3322,
		0.372533,
		0.414182,
		0.456733,
		0.499696,
		0.542528,
		0.584648,
		0.625459,
		0.664376,
		0.700845,
		0.734367,
		0.764515,
		0.790953,
		0.813438,
		0.83183,
		0.846087,
		0.856264,
		0.862495,
		0.864989,
		0.864009,
		0.859861,
		0.852871,
		0.84338,
		0.831725,
		0.818227,
		0.803188,
		0.786881,
		0.769549,
		0.751401,
		0.732617,
		0.713346,
		0.69371,
		0.673812,
		0.653732,
		0.633539,
		0.613289,
		0.59303,
		0.572806,
		0.552654,
		0.532609,
		0.512705,
		0.492972,
		0.473438,
		0.45413,
		0.435073,
		0.416289,
		0.397801,
		0.379627,
		0.361785,
		0.344293,
		0.327165,
		0.310415,
		0.294054,
		0.278096,
		0.262549,
		0.247423,
		0.232725,
		0.218463,
		0.204642,
		0.191269,
		0.178346,
		0.165879,
		0.153869,
		0.14232,
		0.131232,
		0.120607,
		0.110446,
		0.100747,
		0.091512,
		0.082738,
		0.074425,
		0.066571,
		0.059174,
		0.05223,
		0.045739,
		0.039695,
		0.034098,
		0.028941,
		0.024223,
		0.019939,
		0.016085,
		0.012657,
		0.00965,
		0.00706,
		0.004882,
		0.003111,
		0.001742,
		0.000771,
		0.000192,
		3.1e-005,
		0.000165,
		0.00054,
		0.001351,
		0.002835,
		0.005261,
		0.00891,
		0.014064,
		0.020985,
		0.02991,
		0.041042,
		0.054548,
		0.070559,
		0.089172,
		0.110452,
		0.134436,
		0.161129,
		0.190497,
		0.222461,
		0.256879,
		0.293526,
		0.33209,
		0.372177,
		0.413311,
		0.454953,
		0.496513,
		0.537373,
		0.57691,
		0.614514,
		0.649616,
		0.681705,
		0.710347,
		0.735198,
		0.756013,
		0.77265,
		0.785067,
		0.793321,
		0.797554,
		0.797981,
		0.794878,
		0.788563,
		0.779382,
		0.767694,
		0.753855,
		0.73821,
		0.721083,
		0.702768,
		0.68353,
		0.663597,
		0.643164,
		0.622395,
		0.601425,
		0.580364,
		0.559298,
		0.538297,
		0.51742,
		0.496713,
		0.476214,
		0.455958,
		0.435974,
		0.416289,
		0.396928,
		0.377913,
		0.359265,
		0.341002,
		0.323142,
		0.3057,
		0.28869,
		0.272124,
		0.256014,
		0.24037,
		0.2252,
		0.210511,
		0.19631,
		0.182603,
		0.169394,
		0.156687,
		0.144484,
		0.132788,
		0.121599,
		0.110919,
		0.100747,
		0.091084,
		0.081927,
		0.073275,
		0.065127,
		0.057479,
		0.05033,
		0.043674,
		0.037511,
		0.031834,
		0.02664,
		0.021925,
		0.017685,
		0.013913,
		0.010606,
		0.007758,
		0.005363,
		0.003417,
		0.001913,
		0.000846,
		0.000211,
		4e-005,
		0.000203,
		0.000642,
		0.001555,
		0.003177,
		0.00576,
		0.009563,
		0.014841,
		0.021831,
		0.030755,
		0.041813,
		0.055184,
		0.071024,
		0.089464,
		0.110603,
		0.134495,
		0.161129,
		0.190422,
		0.222209,
		0.256238,
		0.292176,
		0.329606,
		0.368043,
		0.406945,
		0.445728,
		0.483784,
		0.520504,
		0.555299,
		0.587617,
		0.616965,
		0.642924,
		0.665163,
		0.683445,
		0.697635,
		0.707694,
		0.713679,
		0.715733,
		0.714073,
		0.708972,
		0.700751,
		0.689759,
		0.676358,
		0.660911,
		0.643772,
		0.62527,
		0.605712,
		0.585368,
		0.564478,
		0.543246,
		0.52184,
		0.500403,
		0.479047,
		0.457861,
		0.436917,
		0.416269,
		0.395962,
		0.376029,
		0.3565,
		0.337397,
		0.318742,
		0.300551,
		0.282841,
		0.265625,
		0.248916,
		0.232725,
		0.217061,
		0.201932,
		0.187344,
		0.173305,
		0.159817,
		0.146884,
		0.13451,
		0.122695,
		0.111441,
		0.100747,
		0.090614,
		0.081039,
		0.072021,
		0.063557,
		0.055645,
		0.048281,
		0.041461,
		0.035182,
		0.029437,
		0.024223,
		0.019535,
		0.015366,
		0.011711,
		0.008564,
		0.00592,
		0.003771,
		0.002111,
		0.000934,
		0.000232,
		4.5e-005,
		0.000226,
		0.000698,
		0.001659,
		0.003331,
		0.005956,
		0.00978,
		0.01505,
		0.022005,
		0.030877,
		0.041882,
		0.055213,
		0.071024,
		0.089421,
		0.110452,
		0.134094,
		0.160248,
		0.188738,
		0.219306,
		0.251619,
		0.285274,
		0.319805,
		0.354699,
		0.389406,
		0.423361,
		0.455999,
		0.486772,
		0.515172,
		0.540744,
		0.563102,
		0.581942,
		0.597047,
		0.608295,
		0.615657,
		0.61919,
		0.619035,
		0.615402,
		0.608558,
		0.598814,
		0.58651,
		0.572,
		0.555639,
		0.537774,
		0.51873,
		0.498808,
		0.478279,
		0.457378,
		0.436307,
		0.415238,
		0.394307,
		0.373628,
		0.353287,
		0.333351,
		0.313873,
		0.294891,
		0.276435,
		0.258526,
		0.241181,
		0.224414,
		0.208236,
		0.192655,
		0.177679,
		0.163312,
		0.149561,
		0.136427,
		0.123912,
		0.112019,
		0.100747,
		0.090096,
		0.080064,
		0.070648,
		0.061846,
		0.053654,
		0.046069,
		0.039085,
		0.032698,
		0.026902,
		0.02169,
		0.017058,
		0.012998,
		0.009504,
		0.006567,
		0.004182,
		0.002341,
		0.001035,
		0.000257,
		4.8e-005,
		0.000235,
		0.000718,
		0.001687,
		0.003363,
		0.005985,
		0.0098,
		0.015059,
		0.022005,
		0.030859,
		0.041813,
		0.055015,
		0.070559,
		0.08848,
		0.108744,
		0.131247,
		0.155812,
		0.182192,
		0.210073,
		0.239079,
		0.268784,
		0.298721,
		0.328396,
		0.3573,
		0.38493,
		0.4108,
		0.434461,
		0.455512,
		0.473615,
		0.488506,
		0.5,
		0.507996,
		0.512479,
		0.513512,
		0.511233,
		0.505845,
		0.497602,
		0.486799,
		0.473757,
		0.458811,
		0.442298,
		0.424544,
		0.405858,
		0.386524,
		0.366798,
		0.346903,
		0.327031,
		0.30734,
		0.287961,
		0.268996,
		0.250527,
		0.232614,
		0.215301,
		0.198621,
		0.182595,
		0.167239,
		0.152562,
		0.138572,
		0.125272,
		0.112664,
		0.100747,
		0.089522,
		0.078987,
		0.069138,
		0.059973,
		0.051487,
		0.043674,
		0.036531,
		0.030049,
		0.024223,
		0.019046,
		0.014509,
		0.010606,
		0.007327,
		0.004665,
		0.00261,
		0.001154,
		0.000287,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003363,
		0.005981,
		0.00978,
		0.014994,
		0.021831,
		0.030469,
		0.041042,
		0.05363,
		0.068256,
		0.08488,
		0.103399,
		0.123644,
		0.145385,
		0.168337,
		0.192162,
		0.216483,
		0.240889,
		0.26495,
		0.288229,
		0.310294,
		0.330732,
		0.349164,
		0.365257,
		0.378733,
		0.389377,
		0.397048,
		0.401677,
		0.403267,
		0.401894,
		0.397697,
		0.390871,
		0.381656,
		0.370327,
		0.35718,
		0.342523,
		0.326664,
		0.309899,
		0.29251,
		0.274752,
		0.256856,
		0.239021,
		0.221416,
		0.204182,
		0.187433,
		0.171255,
		0.155717,
		0.140868,
		0.126741,
		0.113359,
		0.100736,
		0.088879,
		0.077791,
		0.067471,
		0.057915,
		0.049119,
		0.041077,
		0.033782,
		0.027226,
		0.021402,
		0.0163,
		0.011912,
		0.008227,
		0.005236,
		0.002929,
		0.001294,
		0.000322,
		4.9e-005,
		0.000237,
		0.000718,
		0.001679,
		0.003331,
		0.005889,
		0.009563,
		0.014543,
		0.020985,
		0.029006,
		0.038672,
		0.049997,
		0.062938,
		0.077394,
		0.09321,
		0.110179,
		0.128047,
		0.146521,
		0.165276,
		0.183964,
		0.202223,
		0.219692,
		0.236018,
		0.250868,
		0.263942,
		0.274981,
		0.283774,
		0.29017,
		0.294076,
		0.295462,
		0.294361,
		0.290865,
		0.285117,
		0.277308,
		0.267665,
		0.256439,
		0.243899,
		0.23032,
		0.215973,
		0.201119,
		0.185998,
		0.170831,
		0.155812,
		0.141106,
		0.126853,
		0.113164,
		0.100129,
		0.087814,
		0.076267,
		0.065521,
		0.055596,
		0.046503,
		0.038245,
		0.03082,
		0.024222,
		0.018444,
		0.013474,
		0.009304,
		0.00592,
		0.00331,
		0.001462,
		0.000363,
		4.8e-005,
		0.000232,
		0.000698,
		0.001619,
		0.003177,
		0.005553,
		0.00891,
		0.013384,
		0.01907,
		0.026022,
		0.034243,
		0.043688,
		0.054265,
		0.065832,
		0.078208,
		0.091174,
		0.104482,
		0.11786,
		0.131026,
		0.143689,
		0.155564,
		0.166381,
		0.175889,
		0.18387,
		0.190145,
		0.194576,
		0.197077,
		0.197612,
		0.196197,
		0.192897,
		0.187825,
		0.181135,
		0.173013,
		0.163672,
		0.153342,
		0.142261,
		0.130667,
		0.118789,
		0.106843,
		0.095026,
		0.083512,
		0.07245,
		0.061965,
		0.052157,
		0.043104,
		0.034862,
		0.027471,
		0.020955,
		0.015327,
		0.01059,
		0.00674,
		0.003769,
		0.001665,
		0.000414,
		4.5e-005,
		0.000216,
		0.000642,
		0.001466,
		0.002835,
		0.004882,
		0.007717,
		0.011415,
		0.016015,
		0.021513,
		0.027863,
		0.03498,
		0.04274,
		0.050988,
		0.05954,
		0.068196,
		0.076739,
		0.084949,
		0.092608,
		0.099509,
		0.105461,
		0.110298,
		0.113883,
		0.116117,
		0.116937,
		0.116324,
		0.1143,
		0.110928,
		0.106308,
		0.100576,
		0.093895,
		0.086449,
		0.078437,
		0.070064,
		0.061533,
		0.053042,
		0.044773,
		0.036891,
		0.029542,
		0.022848,
		0.016906,
		0.011792,
		0.007563,
		0.004255,
		0.001888,
		0.000471,
		4e-005,
		0.000185,
		0.00054,
		0.001213,
		0.002303,
		0.003894,
		0.006041,
		0.008767,
		0.012061,
		0.015878,
		0.020142,
		0.024749,
		0.029569,
		0.03446,
		0.039266,
		0.043827,
		0.047987,
		0.0516,
		0.054533,
		0.056675,
		0.057942,
		0.058278,
		0.057659,
		0.056098,
		0.053638,
		0.050357,
		0.046363,
		0.04179,
		0.036791,
		0.031536,
		0.026203,
		0.020971,
		0.016015,
		0.011499,
		0.007571,
		0.004361,
		0.001976,
		0.000502,
		3.1e-005,
		0.000141,
		0.000402,
		0.000882,
		0.001636,
		0.002699,
		0.00408,
		0.005762,
		0.007705,
		0.009842,
		0.012091,
		0.014355,
		0.016531,
		0.018512,
		0.020197,
		0.021495,
		0.02233,
		0.022647,
		0.022412,
		0.021621,
		0.020295,
		0.018485,
		0.01627,
		0.013754,
		0.011064,
		0.008346,
		0.005759,
		0.003467,
		0.001638,
		0.000432,
		2e-005,
		9e-005,
		0.000248,
		0.000524,
		0.000936,
		0.001481,
		0.002141,
		0.00288,
		0.003651,
		0.004394,
		0.005051,
		0.005563,
		0.005877,
		0.005956,
		0.005777,
		0.005339,
		0.004664,
		0.003798,
		0.002816,
		0.001814,
		0.000913,
		0.000256,
		4.1e-005,
		0.000106,
		0.000207,
		0.000339,
		0.000485,
		0.000623,
		0.000727,
		0.000774,
		0.000748,
		0.000644,
		0.000473,
		0.000268,
		8.4e-005,
		1.4e-005,
		1.8e-005,
		1.6e-005,
		1.6e-005,
		1.8e-005,
		1.4e-005,
		8.4e-005,
		0.000268,
		0.000473,
		0.000644,
		0.000748,
		0.000774,
		0.000727,
		0.000623,
		0.000485,
		0.000339,
		0.000207,
		0.000106,
		4.1e-005,
		0.000256,
		0.000913,
		0.001814,
		0.002816,
		0.003798,
		0.004664,
		0.005339,
		0.005777,
		0.005956,
		0.005877,
		0.005563,
		0.005051,
		0.004394,
		0.003651,
		0.00288,
		0.002141,
		0.001481,
		0.000936,
		0.000524,
		0.000248,
		9e-005,
		2e-005,
		0.000432,
		0.001638,
		0.003467,
		0.005759,
		0.008346,
		0.011064,
		0.013754,
		0.01627,
		0.018485,
		0.020295,
		0.021621,
		0.022412,
		0.022647,
		0.02233,
		0.021495,
		0.020197,
		0.018512,
		0.016531,
		0.014355,
		0.012091,
		0.009842,
		0.007705,
		0.005762,
		0.00408,
		0.002699,
		0.001636,
		0.000882,
		0.000402,
		0.000141,
		3.1e-005,
		0.000502,
		0.001976,
		0.004361,
		0.007571,
		0.011499,
		0.016015,
		0.020971,
		0.026203,
		0.031536,
		0.036791,
		0.04179,
		0.046363,
		0.050357,
		0.053638,
		0.056098,
		0.057659,
		0.058278,
		0.057942,
		0.056675,
		0.054533,
		0.0516,
		0.047987,
		0.043827,
		0.039266,
		0.03446,
		0.029569,
		0.024749,
		0.020142,
		0.015878,
		0.012061,
		0.008767,
		0.006041,
		0.003894,
		0.002303,
		0.001213,
		0.00054,
		0.000185,
		4e-005,
		0.000471,
		0.001888,
		0.004255,
		0.007563,
		0.011792,
		0.016906,
		0.022848,
		0.029542,
		0.036891,
		0.044773,
		0.053042,
		0.061533,
		0.070064,
		0.078437,
		0.086449,
		0.093895,
		0.100576,
		0.106308,
		0.110928,
		0.1143,
		0.116324,
		0.116937,
		0.116117,
		0.113883,
		0.110298,
		0.105461,
		0.099509,
		0.092608,
		0.084949,
		0.076739,
		0.068196,
		0.05954,
		0.050988,
		0.04274,
		0.03498,
		0.027863,
		0.021513,
		0.016015,
		0.011415,
		0.007717,
		0.004882,
		0.002835,
		0.001466,
		0.000642,
		0.000216,
		4.5e-005,
		0.000414,
		0.001665,
		0.003769,
		0.00674,
		0.01059,
		0.015327,
		0.020955,
		0.027471,
		0.034862,
		0.043104,
		0.052157,
		0.061965,
		0.07245,
		0.083512,
		0.095026,
		0.106843,
		0.118789,
		0.130667,
		0.142261,
		0.153342,
		0.163672,
		0.173013,
		0.181135,
		0.187825,
		0.192897,
		0.196197,
		0.197612,
		0.197077,
		0.194576,
		0.190145,
		0.18387,
		0.175889,
		0.166381,
		0.155564,
		0.143689,
		0.131026,
		0.11786,
		0.104482,
		0.091174,
		0.078208,
		0.065832,
		0.054265,
		0.043688,
		0.034243,
		0.026022,
		0.01907,
		0.013384,
		0.00891,
		0.005553,
		0.003177,
		0.001619,
		0.000698,
		0.000232,
		4.8e-005,
		0.000363,
		0.001462,
		0.00331,
		0.00592,
		0.009304,
		0.013474,
		0.018444,
		0.024222,
		0.03082,
		0.038245,
		0.046503,
		0.055596,
		0.065521,
		0.076267,
		0.087814,
		0.100129,
		0.113164,
		0.126853,
		0.141106,
		0.155812,
		0.170831,
		0.185998,
		0.201119,
		0.215973,
		0.23032,
		0.243899,
		0.256439,
		0.267665,
		0.277308,
		0.285117,
		0.290865,
		0.294361,
		0.295462,
		0.294076,
		0.29017,
		0.283774,
		0.274981,
		0.263942,
		0.250868,
		0.236018,
		0.219692,
		0.202223,
		0.183964,
		0.165276,
		0.146521,
		0.128047,
		0.110179,
		0.09321,
		0.077394,
		0.062938,
		0.049997,
		0.038672,
		0.029006,
		0.020985,
		0.014543,
		0.009563,
		0.005889,
		0.003331,
		0.001679,
		0.000718,
		0.000237,
		4.9e-005,
		0.000322,
		0.001294,
		0.002929,
		0.005236,
		0.008227,
		0.011912,
		0.0163,
		0.021402,
		0.027226,
		0.033782,
		0.041077,
		0.049119,
		0.057915,
		0.067471,
		0.077791,
		0.088879,
		0.100736,
		0.113359,
		0.126741,
		0.140868,
		0.155717,
		0.171255,
		0.187433,
		0.204182,
		0.221416,
		0.239021,
		0.256856,
		0.274752,
		0.29251,
		0.309899,
		0.326664,
		0.342523,
		0.35718,
		0.370327,
		0.381656,
		0.390871,
		0.397697,
		0.401894,
		0.403267,
		0.401677,
		0.397048,
		0.389377,
		0.378733,
		0.365257,
		0.349164,
		0.330732,
		0.310294,
		0.288229,
		0.26495,
		0.240889,
		0.216483,
		0.192162,
		0.168337,
		0.145385,
		0.123644,
		0.103399,
		0.08488,
		0.068256,
		0.05363,
		0.041042,
		0.030469,
		0.021831,
		0.014994,
		0.00978,
		0.005981,
		0.003363,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.000287,
		0.001154,
		0.00261,
		0.004665,
		0.007327,
		0.010606,
		0.014509,
		0.019046,
		0.024223,
		0.030049,
		0.036531,
		0.043674,
		0.051487,
		0.059973,
		0.069138,
		0.078987,
		0.089522,
		0.100747,
		0.112664,
		0.125272,
		0.138572,
		0.152562,
		0.167239,
		0.182595,
		0.198621,
		0.215301,
		0.232614,
		0.250527,
		0.268996,
		0.287961,
		0.30734,
		0.327031,
		0.346903,
		0.366798,
		0.386524,
		0.405858,
		0.424544,
		0.442298,
		0.458811,
		0.473757,
		0.486799,
		0.497602,
		0.505845,
		0.511233,
		0.513512,
		0.512479,
		0.507996,
		0.5,
		0.488506,
		0.473615,
		0.455512,
		0.434461,
		0.4108,
		0.38493,
		0.3573,
		0.328396,
		0.298721,
		0.268784,
		0.239079,
		0.210073,
		0.182192,
		0.155812,
		0.131247,
		0.108744,
		0.08848,
		0.070559,
		0.055015,
		0.041813,
		0.030859,
		0.022005,
		0.015059,
		0.0098,
		0.005985,
		0.003363,
		0.001687,
		0.000718,
		0.000235,
		4.8e-005,
		0.000257,
		0.001035,
		0.002341,
		0.004182,
		0.006567,
		0.009504,
		0.012998,
		0.017058,
		0.02169,
		0.026902,
		0.032698,
		0.039085,
		0.046069,
		0.053654,
		0.061846,
		0.070648,
		0.080064,
		0.090096,
		0.100747,
		0.112019,
		0.123912,
		0.136427,
		0.149561,
		0.163312,
		0.177679,
		0.192655,
		0.208236,
		0.224414,
		0.241181,
		0.258526,
		0.276435,
		0.294891,
		0.313873,
		0.333351,
		0.353287,
		0.373628,
		0.394307,
		0.415238,
		0.436307,
		0.457378,
		0.478279,
		0.498808,
		0.51873,
		0.537774,
		0.555639,
		0.572,
		0.58651,
		0.598814,
		0.608558,
		0.615402,
		0.619035,
		0.61919,
		0.615657,
		0.608295,
		0.597047,
		0.581942,
		0.563102,
		0.540744,
		0.515172,
		0.486772,
		0.455999,
		0.423361,
		0.389406,
		0.354699,
		0.319805,
		0.285274,
		0.251619,
		0.219306,
		0.188738,
		0.160248,
		0.134094,
		0.110452,
		0.089421,
		0.071024,
		0.055213,
		0.041882,
		0.030877,
		0.022005,
		0.01505,
		0.00978,
		0.005956,
		0.003331,
		0.001659,
		0.000698,
		0.000226,
		4.5e-005,
		0.000232,
		0.000934,
		0.002111,
		0.003771,
		0.00592,
		0.008564,
		0.011711,
		0.015366,
		0.019535,
		0.024223,
		0.029437,
		0.035182,
		0.041461,
		0.048281,
		0.055645,
		0.063557,
		0.072021,
		0.081039,
		0.090614,
		0.100747,
		0.111441,
		0.122695,
		0.13451,
		0.146884,
		0.159817,
		0.173305,
		0.187344,
		0.201932,
		0.217061,
		0.232725,
		0.248916,
		0.265625,
		0.282841,
		0.300551,
		0.318742,
		0.337397,
		0.3565,
		0.376029,
		0.395962,
		0.416269,
		0.436917,
		0.457861,
		0.479047,
		0.500403,
		0.52184,
		0.543246,
		0.564478,
		0.585368,
		0.605712,
		0.62527,
		0.643772,
		0.660911,
		0.676358,
		0.689759,
		0.700751,
		0.708972,
		0.714073,
		0.715733,
		0.713679,
		0.707694,
		0.697635,
		0.683445,
		0.665163,
		0.642924,
		0.616965,
		0.587617,
		0.555299,
		0.520504,
		0.483784,
		0.445728,
		0.406945,
		0.368043,
		0.329606,
		0.292176,
		0.256238,
		0.222209,
		0.190422,
		0.161129,
		0.134495,
		0.110603,
		0.089464,
		0.071024,
		0.055184,
		0.041813,
		0.030755,
		0.021831,
		0.014841,
		0.009563,
		0.00576,
		0.003177,
		0.001555,
		0.000642,
		0.000203,
		4e-005,
		0.000211,
		0.000846,
		0.001913,
		0.003417,
		0.005363,
		0.007758,
		0.010606,
		0.013913,
		0.017685,
		0.021925,
		0.02664,
		0.031834,
		0.037511,
		0.043674,
		0.05033,
		0.057479,
		0.065127,
		0.073275,
		0.081927,
		0.091084,
		0.100747,
		0.110919,
		0.121599,
		0.132788,
		0.144484,
		0.156687,
		0.169394,
		0.182603,
		0.19631,
		0.210511,
		0.2252,
		0.24037,
		0.256014,
		0.272124,
		0.28869,
		0.3057,
		0.323142,
		0.341002,
		0.359265,
		0.377913,
		0.396928,
		0.416289,
		0.435974,
		0.455958,
		0.476214,
		0.496713,
		0.51742,
		0.538297,
		0.559298,
		0.580364,
		0.601425,
		0.622395,
		0.643164,
		0.663597,
		0.68353,
		0.702768,
		0.721083,
		0.73821,
		0.753855,
		0.767694,
		0.779382,
		0.788563,
		0.794878,
		0.797981,
		0.797554,
		0.793321,
		0.785067,
		0.77265,
		0.756013,
		0.735198,
		0.710347,
		0.681705,
		0.649616,
		0.614514,
		0.57691,
		0.537373,
		0.496513,
		0.454953,
		0.413311,
		0.372177,
		0.33209,
		0.293526,
		0.256879,
		0.222461,
		0.190497,
		0.161129,
		0.134436,
		0.110452,
		0.089172,
		0.070559,
		0.054548,
		0.041042,
		0.02991,
		0.020985,
		0.014064,
		0.00891,
		0.005261,
		0.002835,
		0.001351,
		0.00054,
		0.000165,
		3.1e-005,
		0.000192,
		0.000771,
		0.001742,
		0.003111,
		0.004882,
		0.00706,
		0.00965,
		0.012657,
		0.016085,
		0.019939,
		0.024223,
		0.028941,
		0.034098,
		0.039695,
		0.045739,
		0.05223,
		0.059174,
		0.066571,
		0.074425,
		0.082738,
		0.091512,
		0.100747,
		0.110446,
		0.120607,
		0.131232,
		0.14232,
		0.153869,
		0.165879,
		0.178346,
		0.191269,
		0.204642,
		0.218463,
		0.232725,
		0.247423,
		0.262549,
		0.278096,
		0.294054,
		0.310415,
		0.327165,
		0.344293,
		0.361785,
		0.379627,
		0.397801,
		0.416289,
		0.435073,
		0.45413,
		0.473438,
		0.492972,
		0.512705,
		0.532609,
		0.552654,
		0.572806,
		0.59303,
		0.613289,
		0.633539,
		0.653732,
		0.673812,
		0.69371,
		0.713346,
		0.732617,
		0.751401,
		0.769549,
		0.786881,
		0.803188,
		0.818227,
		0.831725,
		0.84338,
		0.852871,
		0.859861,
		0.864009,
		0.864989,
		0.862495,
		0.856264,
		0.846087,
		0.83183,
		0.813438,
		0.790953,
		0.764515,
		0.734367,
		0.700845,
		0.664376,
		0.625459,
		0.584648,
		0.542528,
		0.499696,
		0.456733,
		0.414182,
		0.372533,
		0.3322,
		0.293526,
		0.256786,
		0.222209,
		0.189978,
		0.160248,
		0.13314,
		0.108744,
		0.087112,
		0.068256,
		0.052138,
		0.038672,
		0.027714,
		0.01907,
		0.012498,
		0.007717,
		0.004423,
		0.002303,
		0.001055,
		0.000402,
		0.000116,
		2e-005,
		0.000176,
		0.000705,
		0.001593,
		0.002844,
		0.004462,
		0.006452,
		0.008818,
		0.011564,
		0.014693,
		0.018211,
		0.022121,
		0.026425,
		0.031129,
		0.036235,
		0.041747,
		0.047667,
		0.053998,
		0.060743,
		0.067904,
		0.075484,
		0.083483,
		0.091904,
		0.100747,
		0.110014,
		0.119705,
		0.12982,
		0.140358,
		0.151319,
		0.162702,
		0.174504,
		0.186723,
		0.199356,
		0.212399,
		0.225849,
		0.2397,
		0.253947,
		0.268583,
		0.283601,
		0.298992,
		0.314747,
		0.330857,
		0.34731,
		0.364093,
		0.381194,
		0.398598,
		0.416289,
		0.43425,
		0.452462,
		0.470906,
		0.489559,
		0.5084,
		0.527402,
		0.54654,
		0.565786,
		0.585111,
		0.604481,
		0.623865,
		0.643226,
		0.662527,
		0.681729,
		0.70079,
		0.719666,
		0.738309,
		0.756663,
		0.774666,
		0.792245,
		0.80931,
		0.825756,
		0.841453,
		0.856247,
		0.869955,
		0.882368,
		0.893244,
		0.90232,
		0.909308,
		0.913908,
		0.915816,
		0.914735,
		0.910389,
		0.902538,
		0.890993,
		0.875627,
		0.856393,
		0.833327,
		0.806556,
		0.776299,
		0.742861,
		0.706621,
		0.668022,
		0.627545,
		0.585694,
		0.542966,
		0.499835,
		0.456733,
		0.414057,
		0.372177,
		0.33144,
		0.292176,
		0.254699,
		0.219306,
		0.186264,
		0.155812,
		0.128143,
		0.103399,
		0.081659,
		0.062938,
		0.047176,
		0.034243,
		0.023941,
		0.016015,
		0.01016,
		0.006041,
		0.003312,
		0.001636,
		0.000703,
		0.000248,
		6.4e-005,
		0.000161,
		0.000647,
		0.001462,
		0.00261,
		0.004095,
		0.00592,
		0.008089,
		0.010606,
		0.013475,
		0.016698,
		0.02028,
		0.024223,
		0.028532,
		0.033208,
		0.038254,
		0.043674,
		0.049471,
		0.055645,
		0.0622,
		0.069138,
		0.076461,
		0.084169,
		0.092264,
		0.100747,
		0.10962,
		0.118882,
		0.128533,
		0.138573,
		0.149001,
		0.159817,
		0.171018,
		0.182603,
		0.19457,
		0.206915,
		0.219635,
		0.232725,
		0.246181,
		0.259999,
		0.27417,
		0.28869,
		0.30355,
		0.318742,
		0.334256,
		0.350084,
		0.366214,
		0.382633,
		0.39933,
		0.416289,
		0.433497,
		0.450936,
		0.468589,
		0.486437,
		0.50446,
		0.522638,
		0.540946,
		0.559361,
		0.577857,
		0.596407,
		0.614981,
		0.63355,
		0.652082,
		0.670542,
		0.688896,
		0.707107,
		0.725136,
		0.742944,
		0.760489,
		0.777728,
		0.794617,
		0.811109,
		0.827155,
		0.842698,
		0.857677,
		0.872018,
		0.885637,
		0.898429,
		0.910271,
		0.921018,
		0.930499,
		0.938519,
		0.944861,
		0.949288,
		0.95155,
		0.951392,
		0.948562,
		0.942826,
		0.933976,
		0.921848,
		0.906329,
		0.887367,
		0.864981,
		0.839261,
		0.810366,
		0.77852,
		0.743996,
		0.707107,
		0.668179,
		0.627545,
		0.585544,
		0.542528,
		0.498867,
		0.454953,
		0.411201,
		0.368043,
		0.325921,
		0.285274,
		0.246526,
		0.210073,
		0.176263,
		0.145385,
		0.117656,
		0.09321,
		0.072094,
		0.054265,
		0.039592,
		0.027863,
		0.018798,
		0.012061,
		0.007282,
		0.00408,
		0.002079,
		0.000936,
		0.000355,
		0.000106,
		2.1e-005,
		0.000148,
		0.000596,
		0.001347,
		0.002404,
		0.003771,
		0.00545,
		0.007447,
		0.009762,
		0.012401,
		0.015366,
		0.01866,
		0.022285,
		0.026246,
		0.030544,
		0.035182,
		0.040162,
		0.045488,
		0.051161,
		0.057184,
		0.063557,
		0.070284,
		0.077365,
		0.084802,
		0.092596,
		0.100747,
		0.109257,
		0.118126,
		0.127354,
		0.13694,
		0.146884,
		0.157186,
		0.167843,
		0.178855,
		0.190218,
		0.201932,
		0.213992,
		0.226396,
		0.239139,
		0.252217,
		0.265625,
		0.279358,
		0.293408,
		0.30777,
		0.322436,
		0.337397,
		0.352645,
		0.368169,
		0.383959,
		0.400003,
		0.416289,
		0.432804,
		0.449532,
		0.466459,
		0.483569,
		0.500843,
		0.518262,
		0.535808,
		0.553458,
		0.571191,
		0.588982,
		0.606807,
		0.62464,
		0.642452,
		0.660215,
		0.677898,
		0.69547,
		0.712898,
		0.730146,
		0.74718,
		0.763962,
		0.780455,
		0.796619,
		0.812415,
		0.827802,
		0.842738,
		0.857183,
		0.871094,
		0.884429,
		0.897143,
		0.909189,
		0.920512,
		0.931051,
		0.940734,
		0.949475,
		0.957169,
		0.963696,
		0.968915,
		0.972665,
		0.974769,
		0.975038,
		0.973275,
		0.969283,
		0.962873,
		0.953873,
		0.942136,
		0.927549,
		0.910034,
		0.889556,
		0.866118,
		0.839755,
		0.810529,
		0.77852,
		0.743835,
		0.706621,
		0.667076,
		0.625459,
		0.582095,
		0.537373,
		0.491749,
		0.445728,
		0.399855,
		0.354699,
		0.310826,
		0.268784,
		0.229082,
		0.192162,
		0.158393,
		0.128047,
		0.101297,
		0.078208,
		0.058738,
		0.04274,
		0.02998,
		0.020142,
		0.012854,
		0.007705,
		0.004271,
		0.002141,
		0.000938,
		0.000339,
		9.1e-005,
		1.4e-005,
		0.000137,
		0.000551,
		0.001245,
		0.002221,
		0.003484,
		0.005035,
		0.006878,
		0.009016,
		0.011451,
		0.014187,
		0.017226,
		0.02057,
		0.024223,
		0.028187,
		0.032464,
		0.037057,
		0.041967,
		0.047197,
		0.052748,
		0.058623,
		0.064824,
		0.07135,
		0.078205,
		0.085389,
		0.092903,
		0.100747,
		0.108924,
		0.117431,
		0.126271,
		0.135442,
		0.144944,
		0.154777,
		0.164939,
		0.175429,
		0.186245,
		0.197385,
		0.208847,
		0.220628,
		0.232725,
		0.245133,
		0.25785,
		0.270869,
		0.284186,
		0.297795,
		0.311689,
		0.325863,
		0.340308,
		0.355015,
		0.369978,
		0.385184,
		0.400625,
		0.416289,
		0.432165,
		0.448238,
		0.464497,
		0.480925,
		0.497509,
		0.51423,
		0.531073,
		0.548017,
		0.565045,
		0.582134,
		0.599263,
		0.61641,
		0.63355,
		0.650658,
		0.667708,
		0.684672,
		0.701521,
		0.718226,
		0.734755,
		0.751077,
		0.767158,
		0.782964,
		0.798462,
		0.813614,
		0.828385,
		0.842738,
		0.856637,
		0.870045,
		0.882926,
		0.895243,
		0.906963,
		0.918052,
		0.928478,
		0.938211,
		0.947222,
		0.955481,
		0.962953,
		0.969601,
		0.97538,
		0.980234,
		0.984094,
		0.986877,
		0.988486,
		0.988808,
		0.987715,
		0.985067,
		0.980718,
		0.974514,
		0.966297,
		0.955914,
		0.943206,
		0.928019,
		0.910191,
		0.889556,
		0.865957,
		0.839261,
		0.809382,
		0.776299,
		0.74007,
		0.700845,
		0.658875,
		0.614514,
		0.568211,
		0.520504,
		0.472002,
		0.423361,
		0.375266,
		0.328396,
		0.283404,
		0.240889,
		0.201372,
		0.165276,
		0.132916,
		0.104482,
		0.080042,
		0.05954,
		0.042807,
		0.029569,
		0.019471,
		0.012091,
		0.006974,
		0.003651,
		0.00167,
		0.000623,
		0.000161,
		1.6e-005,
		0.000127,
		0.000511,
		0.001154,
		0.002059,
		0.003228,
		0.004665,
		0.006372,
		0.008351,
		0.010606,
		0.013138,
		0.015951,
		0.019046,
		0.022426,
		0.026093,
		0.030049,
		0.034297,
		0.038838,
		0.043674,
		0.048808,
		0.05424,
		0.059973,
		0.066007,
		0.072345,
		0.078987,
		0.085934,
		0.093187,
		0.100747,
		0.108615,
		0.11679,
		0.125272,
		0.134062,
		0.14316,
		0.152563,
		0.162273,
		0.172287,
		0.182603,
		0.193222,
		0.204139,
		0.215353,
		0.226862,
		0.238661,
		0.250747,
		0.263117,
		0.275766,
		0.28869,
		0.301882,
		0.315337,
		0.32905,
		0.343012,
		0.357216,
		0.371655,
		0.38632,
		0.401201,
		0.416289,
		0.431573,
		0.447041,
		0.462682,
		0.478481,
		0.494427,
		0.510503,
		0.526696,
		0.542987,
		0.559361,
		0.575799,
		0.592281,
		0.608789,
		0.6253,
		0.641793,
		0.658245,
		0.674631,
		0.690927,
		0.707107,
		0.723143,
		0.739008,
		0.754672,
		0.770107,
		0.785281,
		0.800163,
		0.814721,
		0.828924,
		0.842738,
		0.856131,
		0.86907,
		0.881522,
		0.893455,
		0.904839,
		0.915644,
		0.925841,
		0.935405,
		0.944313,
		0.952544,
		0.960083,
		0.966918,
		0.973042,
		0.978449,
		0.983139,
		0.98711,
		0.990359,
		0.992877,
		0.994648,
		0.995644,
		0.995825,
		0.995131,
		0.993485,
		0.990785,
		0.986903,
		0.981686,
		0.974941,
		0.966441,
		0.955914,
		0.943055,
		0.927549,
		0.909081,
		0.887367,
		0.86217,
		0.833327,
		0.800764,
		0.764515,
		0.724736,
		0.681705,
		0.635826,
		0.587617,
		0.537699,
		0.486772,
		0.435589,
		0.38493,
		0.335565,
		0.288229,
		0.24359,
		0.202223,
		0.164591,
		0.131026,
		0.101724,
		0.076739,
		0.055989,
		0.039266,
		0.02625,
		0.016531,
		0.009635,
		0.005051,
		0.002263,
		0.000774,
		0.000143,
		0.000118,
		0.000475,
		0.001072,
		0.001913,
		0.003,
		0.004334,
		0.00592,
		0.007758,
		0.009851,
		0.012202,
		0.014812,
		0.017685,
		0.020821,
		0.024223,
		0.027894,
		0.031834,
		0.036046,
		0.040531,
		0.045292,
		0.05033,
		0.055645,
		0.061241,
		0.067117,
		0.073275,
		0.079717,
		0.086442,
		0.093452,
		0.100747,
		0.108329,
		0.116196,
		0.124349,
		0.132788,
		0.141513,
		0.150522,
		0.159817,
		0.169394,
		0.179254,
		0.189395,
		0.199814,
		0.210511,
		0.221482,
		0.232725,
		0.244237,
		0.256014,
		0.268054,
		0.280351,
		0.292901,
		0.3057,
		0.318742,
		0.332021,
		0.345531,
		0.359265,
		0.373216,
		0.387376,
		0.401737,
		0.416289,
		0.431024,
		0.44593,
		0.460998,
		0.476215,
		0.491569,
		0.507048,
		0.522638,
		0.538323,
		0.55409,
		0.569922,
		0.585802,
		0.601712,
		0.617635,
		0.63355,
		0.649438,
		0.665277,
		0.681045,
		0.69672,
		0.712278,
		0.727694,
		0.742944,
		0.758,
		0.772836,
		0.787425,
		0.801739,
		0.815748,
		0.829424,
		0.842738,
		0.855661,
		0.868161,
		0.880212,
		0.891784,
		0.902848,
		0.913379,
		0.92335,
		0.932739,
		0.941522,
		0.949683,
		0.957205,
		0.964076,
		0.970291,
		0.975845,
		0.980744,
		0.984999,
		0.988624,
		0.991646,
		0.994091,
		0.995994,
		0.997388,
		0.998305,
		0.998768,
		0.998789,
		0.998361,
		0.99745,
		0.995988,
		0.993865,
		0.990912,
		0.986903,
		0.98155,
		0.974514,
		0.96542,
		0.953873,
		0.939475,
		0.921848,
		0.900656,
		0.875627,
		0.84658,
		0.813438,
		0.77625,
		0.735198,
		0.690604,
		0.642924,
		0.592741,
		0.540744,
		0.487707,
		0.434461,
		0.381856,
		0.330732,
		0.281881,
		0.236018,
		0.193752,
		0.155564,
		0.121789,
		0.092608,
		0.068049,
		0.047987,
		0.032163,
		0.020197,
		0.011614,
		0.005877,
		0.002411,
		0.000644,
		3.2e-005,
		0.00011,
		0.000443,
		0.000999,
		0.001783,
		0.002795,
		0.004038,
		0.005514,
		0.007225,
		0.009174,
		0.011362,
		0.013791,
		0.016464,
		0.019382,
		0.022547,
		0.025962,
		0.029626,
		0.033544,
		0.037715,
		0.042142,
		0.046826,
		0.051768,
		0.05697,
		0.062433,
		0.068159,
		0.074147,
		0.080399,
		0.086916,
		0.093699,
		0.100747,
		0.108062,
		0.115644,
		0.123492,
		0.131607,
		0.139988,
		0.148635,
		0.157547,
		0.166724,
		0.176164,
		0.185867,
		0.19583,
		0.206051,
		0.21653,
		0.227264,
		0.238249,
		0.249484,
		0.260965,
		0.272688,
		0.28465,
		0.296847,
		0.309274,
		0.321925,
		0.334797,
		0.347883,
		0.361176,
		0.374671,
		0.38836,
		0.402235,
		0.416289,
		0.430513,
		0.444897,
		0.459432,
		0.474108,
		0.488913,
		0.503836,
		0.518865,
		0.533987,
		0.549189,
		0.564456,
		0.579774,
		0.595126,
		0.610497,
		0.625869,
		0.641225,
		0.656545,
		0.671812,
		0.687003,
		0.7021,
		0.717079,
		0.731919,
		0.746597,
		0.761088,
		0.77537,
		0.789417,
		0.803203,
		0.816702,
		0.82989,
		0.842738,
		0.855222,
		0.867313,
		0.878987,
		0.890217,
		0.900978,
		0.911245,
		0.920996,
		0.930208,
		0.938862,
		0.94694,
		0.954428,
		0.961312,
		0.967586,
		0.973246,
		0.978292,
		0.98273,
		0.986574,
		0.989843,
		0.992561,
		0.994763,
		0.996489,
		0.997788,
		0.998711,
		0.999317,
		0.999659,
		0.999787,
		0.999737,
		0.999522,
		0.999123,
		0.998473,
		0.99745,
		0.995868,
		0.993485,
		0.99,
		0.985067,
		0.978299,
		0.969283,
		0.957597,
		0.942826,
		0.924584,
		0.902538,
		0.876427,
		0.846087,
		0.811468,
		0.77265,
		0.729854,
		0.683445,
		0.633929,
		0.581942,
		0.528227,
		0.473615,
		0.418991,
		0.365257,
		0.313298,
		0.263942,
		0.217932,
		0.175889,
		0.138292,
		0.105461,
		0.077548,
		0.054533,
		0.036236,
		0.02233,
		0.012361,
		0.005777,
		0.001962,
		0.000268,
		0.000103,
		0.000414,
		0.000934,
		0.001665,
		0.00261,
		0.003771,
		0.005149,
		0.006746,
		0.008564,
		0.010606,
		0.012873,
		0.015366,
		0.018088,
		0.02104,
		0.024223,
		0.027641,
		0.031293,
		0.035182,
		0.039308,
		0.043674,
		0.048281,
		0.05313,
		0.058222,
		0.063557,
		0.069138,
		0.074965,
		0.081039,
		0.08736,
		0.09393,
		0.100747,
		0.107814,
		0.11513,
		0.122695,
		0.130509,
		0.138573,
		0.146884,
		0.155444,
		0.164251,
		0.173305,
		0.182603,
		0.192146,
		0.201932,
		0.211958,
		0.222223,
		0.232725,
		0.243461,
		0.254429,
		0.265625,
		0.277047,
		0.28869,
		0.300551,
		0.312626,
		0.324909,
		0.337397,
		0.350084,
		0.362964,
		0.376031,
		0.389279,
		0.402701,
		0.416289,
		0.430036,
		0.443933,
		0.457972,
		0.472143,
		0.486437,
		0.500843,
		0.51535,
		0.529947,
		0.544621,
		0.559361,
		0.574153,
		0.588982,
		0.603835,
		0.618696,
		0.63355,
		0.64838,
		0.663169,
		0.677898,
		0.692551,
		0.707107,
		0.721547,
		0.73585,
		0.749996,
		0.763962,
		0.777728,
		0.79127,
		0.804566,
		0.817592,
		0.830324,
		0.842738,
		0.854812,
		0.86652,
		0.877839,
		0.888746,
		0.899218,
		0.909232,
		0.918768,
		0.927805,
		0.936326,
		0.944313,
		0.951752,
		0.958631,
		0.964943,
		0.970681,
		0.975845,
		0.980438,
		0.984468,
		0.987949,
		0.990902,
		0.99335,
		0.995329,
		0.996876,
		0.998037,
		0.998863,
		0.999412,
		0.999741,
		0.999911,
		0.999976,
		0.999981,
		0.999952,
		0.999885,
		0.999737,
		0.999416,
		0.998789,
		0.997669,
		0.995825,
		0.992978,
		0.988808,
		0.982956,
		0.975038,
		0.964651,
		0.951392,
		0.934871,
		0.914735,
		0.890684,
		0.862495,
		0.830044,
		0.793321,
		0.75245,
		0.707694,
		0.659459,
		0.608295,
		0.55488,
		0.5,
		0.444525,
		0.389377,
		0.33549,
		0.283774,
		0.235077,
		0.190145,
		0.149594,
		0.113883,
		0.083299,
		0.057942,
		0.037732,
		0.022412,
		0.011571,
		0.004664,
		0.001043,
		9.6e-005,
		0.000387,
		0.000874,
		0.001559,
		0.002443,
		0.003529,
		0.004818,
		0.006313,
		0.008014,
		0.009923,
		0.012042,
		0.014374,
		0.016918,
		0.019678,
		0.022654,
		0.025848,
		0.029261,
		0.032895,
		0.036751,
		0.04083,
		0.045134,
		0.049664,
		0.054421,
		0.059405,
		0.064618,
		0.070061,
		0.075735,
		0.08164,
		0.087777,
		0.094146,
		0.100747,
		0.107582,
		0.114651,
		0.121952,
		0.129487,
		0.137255,
		0.145256,
		0.15349,
		0.161955,
		0.170651,
		0.179577,
		0.188732,
		0.198115,
		0.207724,
		0.217558,
		0.227614,
		0.237891,
		0.248386,
		0.259096,
		0.27002,
		0.281153,
		0.292492,
		0.304035,
		0.315776,
		0.327712,
		0.339837,
		0.352149,
		0.36464,
		0.377306,
		0.39014,
		0.403137,
		0.416289,
		0.42959,
		0.443032,
		0.456608,
		0.470308,
		0.484124,
		0.498046,
		0.512066,
		0.526172,
		0.540354,
		0.5546,
		0.568899,
		0.583238,
		0.597605,
		0.611985,
		0.626365,
		0.64073,
		0.655065,
		0.669354,
		0.683581,
		0.697728,
		0.711778,
		0.725714,
		0.739516,
		0.753166,
		0.766643,
		0.779928,
		0.793,
		0.805839,
		0.818422,
		0.830729,
		0.842738,
		0.854428,
		0.865776,
		0.876762,
		0.887363,
		0.897559,
		0.907331,
		0.916658,
		0.925522,
		0.933906,
		0.941795,
		0.949176,
		0.956035,
		0.962366,
		0.968161,
		0.973418,
		0.978139,
		0.982327,
		0.985994,
		0.989154,
		0.991827,
		0.994038,
		0.99582,
		0.99721,
		0.998251,
		0.998989,
		0.999478,
		0.999771,
		0.999922,
		0.999984,
		0.999999,
		1,
		0.999997,
		0.999976,
		0.999888,
		0.999659,
		0.999178,
		0.998305,
		0.996864,
		0.994648,
		0.991411,
		0.986877,
		0.980739,
		0.972665,
		0.962301,
		0.949288,
		0.93327,
		0.913908,
		0.890903,
		0.864009,
		0.83306,
		0.797981,
		0.758816,
		0.715733,
		0.669042,
		0.61919,
		0.566765,
		0.512479,
		0.45715,
		0.401677,
		0.347001,
		0.294076,
		0.243819,
		0.197077,
		0.154586,
		0.116937,
		0.084551,
		0.057659,
		0.036295,
		0.020295,
		0.009308,
		0.002816,
		0.000161,
		9.1e-005,
		0.000363,
		0.00082,
		0.001462,
		0.002292,
		0.00331,
		0.004519,
		0.00592,
		0.007514,
		0.009304,
		0.01129,
		0.013475,
		0.015859,
		0.018444,
		0.021232,
		0.024223,
		0.02742,
		0.030823,
		0.034434,
		0.038254,
		0.042284,
		0.046525,
		0.050979,
		0.055645,
		0.060526,
		0.065621,
		0.070933,
		0.076461,
		0.082205,
		0.088168,
		0.094348,
		0.100747,
		0.107365,
		0.114202,
		0.121258,
		0.128533,
		0.136026,
		0.143738,
		0.151669,
		0.159817,
		0.168182,
		0.176763,
		0.185559,
		0.19457,
		0.203793,
		0.213228,
		0.222873,
		0.232725,
		0.242783,
		0.253045,
		0.263508,
		0.27417,
		0.285028,
		0.296078,
		0.307317,
		0.318742,
		0.330348,
		0.342132,
		0.354089,
		0.366214,
		0.378502,
		0.390948,
		0.403546,
		0.416289,
		0.429172,
		0.442188,
		0.455329,
		0.468589,
		0.481957,
		0.495428,
		0.508991,
		0.522638,
		0.536358,
		0.550142,
		0.563978,
		0.577857,
		0.591766,
		0.605693,
		0.619625,
		0.63355,
		0.647454,
		0.661323,
		0.675142,
		0.688896,
		0.70257,
		0.716147,
		0.72961,
		0.742944,
		0.756129,
		0.769149,
		0.781985,
		0.794618,
		0.807029,
		0.8192,
		0.831109,
		0.842738,
		0.854068,
		0.865077,
		0.875748,
		0.886059,
		0.895994,
		0.905532,
		0.914656,
		0.92335,
		0.931598,
		0.939384,
		0.946697,
		0.953525,
		0.959858,
		0.965692,
		0.971021,
		0.975845,
		0.980168,
		0.983995,
		0.987338,
		0.990212,
		0.992637,
		0.99464,
		0.996249,
		0.997501,
		0.998436,
		0.999098,
		0.999535,
		0.999797,
		0.999931,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999917,
		0.999741,
		0.999373,
		0.998711,
		0.997633,
		0.995994,
		0.993631,
		0.990359,
		0.98597,
		0.980234,
		0.9729,
		0.963696,
		0.952336,
		0.938519,
		0.921945,
		0.90232,
		0.879373,
		0.852871,
		0.822636,
		0.788563,
		0.750642,
		0.708972,
		0.663773,
		0.615402,
		0.564348,
		0.511233,
		0.456801,
		0.401894,
		0.347429,
		0.294361,
		0.243645,
		0.196197,
		0.152846,
		0.1143,
		0.08111,
		0.053638,
		0.032043,
		0.01627,
		0.00605,
		0.000913,
		0.005562,
		0.00706,
		0.008741,
		0.010606,
		0.012657,
		0.014896,
		0.017323,
		0.019939,
		0.022747,
		0.025748,
		0.028941,
		0.03233,
		0.035914,
		0.039695,
		0.043674,
		0.047852,
		0.05223,
		0.056809,
		0.061589,
		0.066571,
		0.071756,
		0.077145,
		0.082738,
		0.088536,
		0.094539,
		0.100747,
		0.107161,
		0.113781,
		0.120607,
		0.127639,
		0.134877,
		0.14232,
		0.149968,
		0.157821,
		0.165879,
		0.17414,
		0.182603,
		0.191269,
		0.200135,
		0.2092,
		0.218463,
		0.227922,
		0.237576,
		0.247423,
		0.25746,
		0.267685,
		0.278096,
		0.28869,
		0.299464,
		0.310415,
		0.321539,
		0.332833,
		0.344293,
		0.355915,
		0.367695,
		0.379627,
		0.391707,
		0.40393,
		0.416289,
		0.42878,
		0.441396,
		0.45413,
		0.466975,
		0.479925,
		0.492972,
		0.506107,
		0.519322,
		0.532609,
		0.545959,
		0.559361,
		0.572806,
		0.586284,
		0.599783,
		0.613292,
		0.6268,
		0.640295,
		0.653763,
		0.667192,
		0.680569,
		0.693878,
		0.707107,
		0.720239,
		0.73326,
		0.746155,
		0.758906,
		0.771497,
		0.783913,
		0.796135,
		0.808146,
		0.819929,
		0.831466,
		0.842738,
		0.853729,
		0.864419,
		0.874792,
		0.884829,
		0.894513,
		0.903828,
		0.912756,
		0.921283,
		0.929393,
		0.937074,
		0.944313,
		0.951098,
		0.957423,
		0.963278,
		0.968661,
		0.973569,
		0.978003,
		0.981968,
		0.985471,
		0.988525,
		0.991144,
		0.99335,
		0.995167,
		0.996625,
		0.997755,
		0.998598,
		0.999193,
		0.999585,
		0.999819,
		0.999939,
		0.999987,
		0.999999,
		1,
		0.999999,
		0.999985,
		0.999927,
		0.999771,
		0.999446,
		0.998863,
		0.997917,
		0.996489,
		0.994447,
		0.991646,
		0.987932,
		0.983139,
		0.977092,
		0.969601,
		0.960466,
		0.949475,
		0.936403,
		0.921018,
		0.903084,
		0.882368,
		0.858647,
		0.831725,
		0.801442,
		0.767694,
		0.730447,
		0.689759,
		0.645788,
		0.598814,
		0.549241,
		0.497602,
		0.444555,
		0.390871,
		0.337415,
		0.285117,
		0.234937,
		0.187825,
		0.144681,
		0.106308,
		0.073371,
		0.046363,
		0.025572,
		0.011064,
		0.002672,
		0.021402,
		0.024223,
		0.027226,
		0.030412,
		0.033782,
		0.037337,
		0.041077,
		0.045004,
		0.049119,
		0.053423,
		0.057915,
		0.062598,
		0.067471,
		0.072536,
		0.077793,
		0.083242,
		0.088884,
		0.094719,
		0.100747,
		0.10697,
		0.113386,
		0.119997,
		0.126801,
		0.133799,
		0.140991,
		0.148377,
		0.155955,
		0.163726,
		0.171689,
		0.179843,
		0.188187,
		0.196721,
		0.205443,
		0.214352,
		0.223446,
		0.232725,
		0.242186,
		0.251827,
		0.261647,
		0.271644,
		0.281814,
		0.292156,
		0.302666,
		0.313342,
		0.324181,
		0.335179,
		0.346332,
		0.357638,
		0.369091,
		0.380687,
		0.392422,
		0.404291,
		0.416289,
		0.428411,
		0.44065,
		0.453002,
		0.465459,
		0.478015,
		0.490663,
		0.503396,
		0.516206,
		0.529086,
		0.542026,
		0.55502,
		0.568056,
		0.581127,
		0.594222,
		0.607332,
		0.620445,
		0.63355,
		0.646637,
		0.659694,
		0.672708,
		0.685666,
		0.698557,
		0.711367,
		0.724081,
		0.736686,
		0.749168,
		0.761512,
		0.773702,
		0.785723,
		0.797559,
		0.809195,
		0.820615,
		0.831801,
		0.842738,
		0.85341,
		0.863799,
		0.87389,
		0.883666,
		0.893112,
		0.902212,
		0.91095,
		0.919314,
		0.927288,
		0.93486,
		0.94202,
		0.948756,
		0.95506,
		0.960924,
		0.966345,
		0.971319,
		0.975845,
		0.979927,
		0.98357,
		0.986782,
		0.989577,
		0.991969,
		0.99398,
		0.995632,
		0.996954,
		0.997978,
		0.998739,
		0.999276,
		0.999628,
		0.999838,
		0.999945,
		0.999988,
		0.999999,
		1,
		0.999999,
		0.999987,
		0.999935,
		0.999797,
		0.999508,
		0.998989,
		0.998148,
		0.996876,
		0.995056,
		0.992561,
		0.989256,
		0.984999,
		0.979644,
		0.973042,
		0.965039,
		0.955481,
		0.944206,
		0.931051,
		0.915849,
		0.898429,
		0.878618,
		0.856247,
		0.831152,
		0.803188,
		0.772234,
		0.73821,
		0.701088,
		0.660911,
		0.617806,
		0.572,
		0.523831,
		0.473757,
		0.422357,
		0.370327,
		0.318468,
		0.267665,
		0.218859,
		0.173013,
		0.131066,
		0.093895,
		0.062266,
		0.036791,
		0.017889,
		0.005759,
		0.000352,
		0.046277,
		0.05033,
		0.05456,
		0.058969,
		0.063557,
		0.068326,
		0.073275,
		0.078406,
		0.083718,
		0.089212,
		0.094888,
		0.100747,
		0.106789,
		0.113014,
		0.119423,
		0.126014,
		0.132788,
		0.139745,
		0.146884,
		0.154206,
		0.16171,
		0.169394,
		0.17726,
		0.185305,
		0.193529,
		0.201932,
		0.210511,
		0.219266,
		0.228195,
		0.237298,
		0.246571,
		0.256014,
		0.265625,
		0.275401,
		0.285341,
		0.295441,
		0.3057,
		0.316114,
		0.326681,
		0.337397,
		0.34826,
		0.359265,
		0.370409,
		0.381687,
		0.393097,
		0.404632,
		0.416289,
		0.428063,
		0.439948,
		0.451939,
		0.46403,
		0.476215,
		0.488488,
		0.500843,
		0.513271,
		0.525768,
		0.538323,
		0.550931,
		0.563582,
		0.576269,
		0.588982,
		0.601712,
		0.61445,
		0.627186,
		0.63991,
		0.65261,
		0.665277,
		0.677898,
		0.690463,
		0.702959,
		0.715374,
		0.727694,
		0.739908,
		0.752002,
		0.763962,
		0.775775,
		0.787425,
		0.798899,
		0.810183,
		0.82126,
		0.832117,
		0.842738,
		0.853109,
		0.863213,
		0.873037,
		0.882566,
		0.891784,
		0.900677,
		0.909232,
		0.917436,
		0.925275,
		0.932739,
		0.939815,
		0.946495,
		0.952769,
		0.958631,
		0.964076,
		0.9691,
		0.973702,
		0.977883,
		0.981647,
		0.984999,
		0.987949,
		0.990511,
		0.9927,
		0.994536,
		0.996042,
		0.997245,
		0.998174,
		0.998863,
		0.999348,
		0.999666,
		0.999854,
		0.999951,
		0.99999,
		0.999999,
		1,
		0.999999,
		0.999988,
		0.999942,
		0.999819,
		0.999561,
		0.999098,
		0.998347,
		0.99721,
		0.995583,
		0.99335,
		0.99039,
		0.986574,
		0.981771,
		0.975845,
		0.968661,
		0.960083,
		0.949977,
		0.938211,
		0.924657,
		0.909189,
		0.891682,
		0.872018,
		0.85008,
		0.825756,
		0.798942,
		0.769549,
		0.737504,
		0.702768,
		0.66534,
		0.62527,
		0.582681,
		0.537774,
		0.490845,
		0.442298,
		0.392648,
		0.342523,
		0.292662,
		0.243899,
		0.197141,
		0.153342,
		0.113465,
		0.078437,
		0.049109,
		0.026203,
		0.010267,
		0.001638,
		0.078987,
		0.084169,
		0.089522,
		0.095049,
		0.100747,
		0.106619,
		0.112664,
		0.118882,
		0.125272,
		0.131836,
		0.138573,
		0.145482,
		0.152563,
		0.159817,
		0.167242,
		0.174837,
		0.182603,
		0.190539,
		0.198643,
		0.206915,
		0.215353,
		0.223957,
		0.232725,
		0.241656,
		0.250747,
		0.259999,
		0.269407,
		0.278972,
		0.28869,
		0.298559,
		0.308577,
		0.318742,
		0.32905,
		0.339498,
		0.350084,
		0.360804,
		0.371655,
		0.382633,
		0.393734,
		0.404954,
		0.416289,
		0.427734,
		0.439285,
		0.450936,
		0.462682,
		0.474517,
		0.486437,
		0.498434,
		0.510503,
		0.522638,
		0.53483,
		0.547074,
		0.559361,
		0.571684,
		0.584036,
		0.596407,
		0.608789,
		0.621173,
		0.63355,
		0.645911,
		0.658245,
		0.670542,
		0.682792,
		0.694984,
		0.707107,
		0.719149,
		0.731099,
		0.742944,
		0.754672,
		0.766271,
		0.777728,
		0.78903,
		0.800163,
		0.811114,
		0.821869,
		0.832415,
		0.842738,
		0.852824,
		0.862659,
		0.87223,
		0.881522,
		0.890522,
		0.899218,
		0.907596,
		0.915644,
		0.92335,
		0.930704,
		0.937694,
		0.944313,
		0.950551,
		0.956401,
		0.961859,
		0.96692,
		0.971582,
		0.975845,
		0.979712,
		0.983187,
		0.986276,
		0.988991,
		0.991344,
		0.99335,
		0.99503,
		0.996405,
		0.997501,
		0.998347,
		0.998972,
		0.999412,
		0.999699,
		0.999869,
		0.999956,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.99999,
		0.999948,
		0.999838,
		0.999607,
		0.999193,
		0.99852,
		0.997501,
		0.996042,
		0.994038,
		0.991379,
		0.987949,
		0.983628,
		0.978292,
		0.971816,
		0.964076,
		0.954949,
		0.944313,
		0.932051,
		0.918052,
		0.90221,
		0.884429,
		0.864619,
		0.842698,
		0.818594,
		0.792245,
		0.763598,
		0.732617,
		0.699281,
		0.663597,
		0.625599,
		0.585368,
		0.543038,
		0.498808,
		0.452958,
		0.405858,
		0.357979,
		0.309899,
		0.262302,
		0.215973,
		0.171783,
		0.130667,
		0.093595,
		0.061533,
		0.0354,
		0.016015,
		0.004055,
		0.118371,
		0.124573,
		0.130939,
		0.137468,
		0.144161,
		0.151018,
		0.158037,
		0.165218,
		0.172561,
		0.180066,
		0.187731,
		0.195557,
		0.203541,
		0.211684,
		0.219983,
		0.228439,
		0.237049,
		0.245813,
		0.254728,
		0.263794,
		0.273008,
		0.282369,
		0.291874,
		0.301522,
		0.31131,
		0.321236,
		0.331297,
		0.34149,
		0.351814,
		0.362263,
		0.372836,
		0.383529,
		0.394338,
		0.405259,
		0.416289,
		0.427424,
		0.438658,
		0.449987,
		0.461407,
		0.472913,
		0.484498,
		0.496159,
		0.507888,
		0.51968,
		0.531529,
		0.543429,
		0.555372,
		0.567351,
		0.579359,
		0.591389,
		0.603434,
		0.615483,
		0.62753,
		0.639566,
		0.651582,
		0.663568,
		0.675515,
		0.687413,
		0.699252,
		0.711022,
		0.722712,
		0.734311,
		0.745808,
		0.757191,
		0.76845,
		0.779572,
		0.790544,
		0.801356,
		0.811993,
		0.822445,
		0.832697,
		0.842738,
		0.852555,
		0.862134,
		0.871464,
		0.880531,
		0.889324,
		0.897829,
		0.906036,
		0.913932,
		0.921508,
		0.928752,
		0.935655,
		0.942207,
		0.948403,
		0.954233,
		0.959694,
		0.96478,
		0.969489,
		0.973821,
		0.977776,
		0.981357,
		0.984569,
		0.987421,
		0.989923,
		0.992087,
		0.99393,
		0.99547,
		0.996728,
		0.997729,
		0.998499,
		0.999069,
		0.999468,
		0.999728,
		0.999882,
		0.99996,
		0.999992,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999954,
		0.999854,
		0.999648,
		0.999276,
		0.998671,
		0.997755,
		0.996443,
		0.99464,
		0.992245,
		0.989154,
		0.985256,
		0.980438,
		0.974587,
		0.967586,
		0.959323,
		0.949683,
		0.938558,
		0.925841,
		0.911434,
		0.895243,
		0.877185,
		0.857183,
		0.835175,
		0.811109,
		0.784947,
		0.756663,
		0.726248,
		0.69371,
		0.659076,
		0.622395,
		0.583747,
		0.543246,
		0.501051,
		0.457378,
		0.412507,
		0.366798,
		0.320699,
		0.274752,
		0.229604,
		0.185998,
		0.144773,
		0.106843,
		0.073179,
		0.044773,
		0.0226,
		0.007571,
		0.000483,
		0.163312,
		0.170419,
		0.177679,
		0.185091,
		0.192655,
		0.200371,
		0.208236,
		0.216251,
		0.224414,
		0.232725,
		0.241182,
		0.249783,
		0.258527,
		0.267414,
		0.27644,
		0.285604,
		0.294906,
		0.304341,
		0.313909,
		0.323607,
		0.333432,
		0.343382,
		0.353455,
		0.363647,
		0.373956,
		0.384378,
		0.39491,
		0.405548,
		0.416289,
		0.427129,
		0.438064,
		0.449089,
		0.460201,
		0.471394,
		0.482664,
		0.494005,
		0.505413,
		0.516882,
		0.528406,
		0.539979,
		0.551596,
		0.563249,
		0.574932,
		0.586639,
		0.598361,
		0.610092,
		0.621825,
		0.63355,
		0.645261,
		0.656948,
		0.668603,
		0.680218,
		0.691782,
		0.703287,
		0.714722,
		0.726079,
		0.737347,
		0.748515,
		0.759572,
		0.770509,
		0.781314,
		0.791976,
		0.802484,
		0.812825,
		0.82299,
		0.832964,
		0.842738,
		0.8523,
		0.861636,
		0.870737,
		0.879589,
		0.888183,
		0.896506,
		0.904547,
		0.912296,
		0.919743,
		0.926878,
		0.933692,
		0.940176,
		0.946324,
		0.952128,
		0.957583,
		0.962684,
		0.967429,
		0.971816,
		0.975845,
		0.979518,
		0.982839,
		0.985814,
		0.98845,
		0.990759,
		0.992753,
		0.994448,
		0.995862,
		0.997015,
		0.997931,
		0.998635,
		0.999154,
		0.999517,
		0.999754,
		0.999893,
		0.999964,
		0.999993,
		1,
		1,
		0.999999,
		0.999992,
		0.999958,
		0.999869,
		0.999683,
		0.999348,
		0.998803,
		0.997978,
		0.996795,
		0.995167,
		0.993005,
		0.990212,
		0.986687,
		0.982327,
		0.977027,
		0.970681,
		0.963183,
		0.954428,
		0.944313,
		0.932739,
		0.91961,
		0.904839,
		0.888342,
		0.870045,
		0.849883,
		0.827802,
		0.803759,
		0.777728,
		0.749695,
		0.719666,
		0.687664,
		0.653732,
		0.617935,
		0.580364,
		0.541136,
		0.500403,
		0.458357,
		0.415238,
		0.371341,
		0.327031,
		0.282748,
		0.239021,
		0.19647,
		0.155812,
		0.11786,
		0.083512,
		0.053735,
		0.029542,
		0.011958,
		0.001976,
		0.212739,
		0.220628,
		0.228658,
		0.236827,
		0.245133,
		0.253577,
		0.262156,
		0.270869,
		0.279714,
		0.28869,
		0.297795,
		0.307026,
		0.316383,
		0.325863,
		0.335463,
		0.345181,
		0.355015,
		0.364963,
		0.37502,
		0.385184,
		0.395453,
		0.405822,
		0.416289,
		0.42685,
		0.437501,
		0.448238,
		0.459058,
		0.469955,
		0.480925,
		0.491965,
		0.503068,
		0.51423,
		0.525446,
		0.536711,
		0.548017,
		0.559361,
		0.570735,
		0.582134,
		0.59355,
		0.604978,
		0.61641,
		0.627839,
		0.639258,
		0.650658,
		0.662033,
		0.673374,
		0.684672,
		0.695919,
		0.707107,
		0.718226,
		0.729267,
		0.74022,
		0.751077,
		0.761826,
		0.772459,
		0.782964,
		0.793332,
		0.803552,
		0.813614,
		0.823506,
		0.833218,
		0.842738,
		0.852057,
		0.861163,
		0.870045,
		0.878693,
		0.887096,
		0.895243,
		0.903125,
		0.910731,
		0.918052,
		0.925079,
		0.931803,
		0.938217,
		0.944313,
		0.950084,
		0.955526,
		0.960634,
		0.965405,
		0.969836,
		0.973927,
		0.977679,
		0.981094,
		0.984178,
		0.986936,
		0.989376,
		0.99151,
		0.99335,
		0.994912,
		0.996213,
		0.997272,
		0.998111,
		0.998756,
		0.99923,
		0.999561,
		0.999776,
		0.999903,
		0.999968,
		0.999993,
		1,
		1,
		1,
		0.999993,
		0.999962,
		0.999882,
		0.999714,
		0.999412,
		0.99892,
		0.998174,
		0.997104,
		0.995632,
		0.993675,
		0.991144,
		0.987949,
		0.983995,
		0.979184,
		0.973418,
		0.9666,
		0.958631,
		0.949417,
		0.938862,
		0.926878,
		0.913379,
		0.898284,
		0.881522,
		0.863025,
		0.842738,
		0.820615,
		0.796619,
		0.770731,
		0.742944,
		0.713267,
		0.681729,
		0.648379,
		0.613289,
		0.576554,
		0.538297,
		0.498671,
		0.457861,
		0.416091,
		0.373628,
		0.330792,
		0.287961,
		0.245582,
		0.204182,
		0.164372,
		0.126853,
		0.092422,
		0.061965,
		0.036449,
		0.016906,
		0.0044,
		0.265625,
		0.27417,
		0.282841,
		0.291635,
		0.300551,
		0.309587,
		0.318742,
		0.328012,
		0.337397,
		0.346894,
		0.3565,
		0.366214,
		0.376031,
		0.385951,
		0.395969,
		0.406083,
		0.416289,
		0.426585,
		0.436966,
		0.44743,
		0.457972,
		0.468589,
		0.479275,
		0.490028,
		0.500843,
		0.511714,
		0.522638,
		0.533608,
		0.544621,
		0.555671,
		0.566751,
		0.577857,
		0.588982,
		0.60012,
		0.611266,
		0.622411,
		0.63355,
		0.644676,
		0.655781,
		0.666857,
		0.677898,
		0.688896,
		0.699842,
		0.710728,
		0.721547,
		0.732288,
		0.742944,
		0.753505,
		0.763962,
		0.774307,
		0.784529,
		0.794618,
		0.804566,
		0.814362,
		0.823996,
		0.833458,
		0.842738,
		0.851827,
		0.860713,
		0.869387,
		0.877839,
		0.886059,
		0.894038,
		0.901765,
		0.909232,
		0.91643,
		0.92335,
		0.929985,
		0.936326,
		0.942367,
		0.948101,
		0.953525,
		0.958631,
		0.963419,
		0.967884,
		0.972026,
		0.975845,
		0.979343,
		0.982523,
		0.985389,
		0.987949,
		0.990212,
		0.992187,
		0.993888,
		0.995329,
		0.996527,
		0.997501,
		0.998273,
		0.998863,
		0.999297,
		0.9996,
		0.999797,
		0.999912,
		0.999971,
		0.999994,
		1,
		1,
		1,
		0.999993,
		0.999966,
		0.999893,
		0.999741,
		0.999468,
		0.999022,
		0.998347,
		0.997377,
		0.996042,
		0.994266,
		0.991969,
		0.989067,
		0.985471,
		0.981094,
		0.975845,
		0.969633,
		0.962366,
		0.953955,
		0.944313,
		0.933353,
		0.920996,
		0.907163,
		0.891784,
		0.874792,
		0.856131,
		0.835752,
		0.813614,
		0.789689,
		0.763962,
		0.73643,
		0.707107,
		0.676022,
		0.643226,
		0.608789,
		0.572806,
		0.535398,
		0.496713,
		0.456932,
		0.416269,
		0.374977,
		0.333351,
		0.291735,
		0.250527,
		0.21019,
		0.171255,
		0.134336,
		0.100129,
		0.069425,
		0.043104,
		0.022135,
		0.007563,
		0.000494,
		0.320992,
		0.330063,
		0.339242,
		0.348527,
		0.357915,
		0.367405,
		0.376995,
		0.386681,
		0.39646,
		0.406331,
		0.416289,
		0.426333,
		0.436458,
		0.446662,
		0.45694,
		0.46729,
		0.477707,
		0.488188,
		0.498728,
		0.509323,
		0.519969,
		0.530661,
		0.541394,
		0.552163,
		0.562964,
		0.573791,
		0.584639,
		0.595501,
		0.606372,
		0.617247,
		0.628118,
		0.638979,
		0.649825,
		0.660647,
		0.67144,
		0.682196,
		0.692907,
		0.703566,
		0.714166,
		0.724699,
		0.735156,
		0.745529,
		0.75581,
		0.76599,
		0.776061,
		0.786014,
		0.795839,
		0.805528,
		0.815072,
		0.824461,
		0.833687,
		0.842738,
		0.851607,
		0.860284,
		0.86876,
		0.877025,
		0.88507,
		0.892886,
		0.900464,
		0.907796,
		0.914874,
		0.921688,
		0.928233,
		0.934501,
		0.940484,
		0.946178,
		0.951577,
		0.956676,
		0.961473,
		0.965963,
		0.970147,
		0.974023,
		0.977591,
		0.980856,
		0.983819,
		0.986488,
		0.988867,
		0.990967,
		0.992798,
		0.994372,
		0.995704,
		0.99681,
		0.997707,
		0.998417,
		0.99896,
		0.999358,
		0.999635,
		0.999814,
		0.99992,
		0.999973,
		0.999994,
		1,
		1,
		1,
		0.999994,
		0.999969,
		0.999903,
		0.999765,
		0.999517,
		0.999113,
		0.998499,
		0.997618,
		0.996405,
		0.99479,
		0.9927,
		0.990058,
		0.986782,
		0.982792,
		0.978003,
		0.972331,
		0.965692,
		0.958001,
		0.949176,
		0.939136,
		0.927805,
		0.915109,
		0.900978,
		0.885348,
		0.868161,
		0.849368,
		0.828924,
		0.806797,
		0.782964,
		0.757413,
		0.730146,
		0.701178,
		0.670542,
		0.638287,
		0.604481,
		0.569217,
		0.532609,
		0.494799,
		0.455958,
		0.416289,
		0.376029,
		0.335457,
		0.294891,
		0.254699,
		0.215301,
		0.177175,
		0.140868,
		0.106998,
		0.076267,
		0.049465,
		0.027471,
		0.011259,
		0.001888,
		0.377913,
		0.387376,
		0.396928,
		0.406567,
		0.416289,
		0.426093,
		0.435974,
		0.44593,
		0.455958,
		0.466054,
		0.476215,
		0.486437,
		0.496716,
		0.507048,
		0.51743,
		0.527856,
		0.538323,
		0.548826,
		0.559361,
		0.569922,
		0.580504,
		0.591103,
		0.601712,
		0.612327,
		0.622942,
		0.63355,
		0.644146,
		0.654724,
		0.665277,
		0.675798,
		0.686282,
		0.69672,
		0.707107,
		0.717434,
		0.727694,
		0.737881,
		0.747985,
		0.758,
		0.767917,
		0.777728,
		0.787425,
		0.797,
		0.806444,
		0.815748,
		0.824905,
		0.833904,
		0.842738,
		0.851398,
		0.859876,
		0.868161,
		0.876247,
		0.884124,
		0.891784,
		0.899218,
		0.906419,
		0.913379,
		0.92009,
		0.926546,
		0.932739,
		0.938663,
		0.944313,
		0.949683,
		0.954769,
		0.959568,
		0.964076,
		0.968292,
		0.972215,
		0.975845,
		0.979184,
		0.982234,
		0.984999,
		0.987485,
		0.9897,
		0.991651,
		0.99335,
		0.994809,
		0.996042,
		0.997064,
		0.997893,
		0.998546,
		0.999046,
		0.999412,
		0.999666,
		0.99983,
		0.999927,
		0.999976,
		0.999995,
		1,
		1,
		1,
		0.999994,
		0.999972,
		0.999912,
		0.999787,
		0.999561,
		0.999193,
		0.998635,
		0.997833,
		0.996728,
		0.995256,
		0.99335,
		0.990939,
		0.987949,
		0.984305,
		0.979927,
		0.974739,
		0.968661,
		0.961616,
		0.953525,
		0.944313,
		0.933906,
		0.922235,
		0.909232,
		0.894835,
		0.878987,
		0.861636,
		0.842738,
		0.822257,
		0.800163,
		0.776439,
		0.751077,
		0.724081,
		0.69547,
		0.665277,
		0.63355,
		0.600358,
		0.565786,
		0.529947,
		0.492972,
		0.455022,
		0.416289,
		0.376995,
		0.337397,
		0.297794,
		0.258526,
		0.21998,
		0.182595,
		0.146868,
		0.113359,
		0.082699,
		0.055596,
		0.032845,
		0.015327,
		0.004022,
		0.435513,
		0.445233,
		0.455022,
		0.464877,
		0.474793,
		0.484769,
		0.494799,
		0.504881,
		0.515011,
		0.525185,
		0.535398,
		0.545648,
		0.555928,
		0.566235,
		0.576564,
		0.586911,
		0.59727,
		0.607637,
		0.618005,
		0.628371,
		0.638727,
		0.649069,
		0.65939,
		0.669685,
		0.679948,
		0.690172,
		0.70035,
		0.710476,
		0.720543,
		0.730545,
		0.740474,
		0.750323,
		0.760084,
		0.769751,
		0.779315,
		0.788769,
		0.798105,
		0.807315,
		0.816392,
		0.825327,
		0.834112,
		0.842738,
		0.851199,
		0.859486,
		0.86759,
		0.875503,
		0.883219,
		0.890728,
		0.898023,
		0.905097,
		0.911942,
		0.918552,
		0.924919,
		0.931037,
		0.9369,
		0.942504,
		0.947842,
		0.95291,
		0.957705,
		0.962225,
		0.966466,
		0.970427,
		0.974109,
		0.977512,
		0.980638,
		0.98349,
		0.986073,
		0.988393,
		0.990456,
		0.992272,
		0.993851,
		0.995205,
		0.996348,
		0.997294,
		0.99806,
		0.998663,
		0.999123,
		0.99946,
		0.999694,
		0.999845,
		0.999933,
		0.999978,
		0.999995,
		1,
		1,
		1,
		0.999995,
		0.999974,
		0.99992,
		0.999806,
		0.9996,
		0.999265,
		0.998756,
		0.998024,
		0.997015,
		0.995671,
		0.99393,
		0.991726,
		0.988991,
		0.985655,
		0.981647,
		0.976892,
		0.971319,
		0.964853,
		0.957423,
		0.948956,
		0.939384,
		0.928639,
		0.916658,
		0.903379,
		0.888746,
		0.872709,
		0.855222,
		0.836245,
		0.815748,
		0.793707,
		0.770107,
		0.744944,
		0.718226,
		0.689972,
		0.660215,
		0.629005,
		0.596407,
		0.562504,
		0.527402,
		0.491227,
		0.45413,
		0.416289,
		0.377913,
		0.339242,
		0.300551,
		0.262156,
		0.224414,
		0.187731,
		0.152562,
		0.11942,
		0.088879,
		0.061582,
		0.038245,
		0.019668,
		0.00674,
		0.000442,
		0.492972,
		0.502815,
		0.512705,
		0.522638,
		0.532609,
		0.542616,
		0.552654,
		0.562719,
		0.572806,
		0.582912,
		0.593031,
		0.60316,
		0.613292,
		0.623424,
		0.63355,
		0.643665,
		0.653763,
		0.66384,
		0.673888,
		0.683903,
		0.693878,
		0.703808,
		0.713686,
		0.723506,
		0.73326,
		0.742944,
		0.752549,
		0.762069,
		0.771497,
		0.780826,
		0.790049,
		0.799158,
		0.808146,
		0.817006,
		0.825729,
		0.834309,
		0.842738,
		0.851009,
		0.859113,
		0.867043,
		0.874792,
		0.882352,
		0.889716,
		0.896877,
		0.903828,
		0.910561,
		0.917071,
		0.92335,
		0.929393,
		0.935195,
		0.940749,
		0.946052,
		0.951098,
		0.955885,
		0.960409,
		0.964669,
		0.968661,
		0.972387,
		0.975845,
		0.979038,
		0.981968,
		0.984638,
		0.987053,
		0.98922,
		0.991144,
		0.992836,
		0.994306,
		0.995564,
		0.996625,
		0.997501,
		0.99821,
		0.998768,
		0.999193,
		0.999504,
		0.999719,
		0.999858,
		0.999939,
		0.99998,
		0.999996,
		1,
		1,
		1,
		0.999995,
		0.999977,
		0.999927,
		0.999823,
		0.999635,
		0.999328,
		0.998863,
		0.998194,
		0.997272,
		0.996042,
		0.994448,
		0.992429,
		0.989923,
		0.986864,
		0.983187,
		0.978822,
		0.973702,
		0.967759,
		0.960924,
		0.953131,
		0.944313,
		0.934406,
		0.92335,
		0.911086,
		0.897559,
		0.882719,
		0.86652,
		0.848922,
		0.82989,
		0.809397,
		0.787425,
		0.763962,
		0.739008,
		0.71257,
		0.684672,
		0.655346,
		0.62464,
		0.592618,
		0.559361,
		0.524968,
		0.489559,
		0.453277,
		0.416289,
		0.37879,
		0.341002,
		0.303183,
		0.265625,
		0.228658,
		0.192655,
		0.158037,
		0.125272,
		0.094888,
		0.067471,
		0.043674,
		0.024222,
		0.009922,
		0.001665,
		0.549528,
		0.559361,
		0.569217,
		0.579092,
		0.588982,
		0.598882,
		0.608789,
		0.618696,
		0.628601,
		0.638497,
		0.64838,
		0.658245,
		0.668086,
		0.677898,
		0.687676,
		0.697414,
		0.707107,
		0.716747,
		0.726331,
		0.73585,
		0.745299,
		0.754672,
		0.763962,
		0.773163,
		0.782268,
		0.79127,
		0.800163,
		0.808939,
		0.817592,
		0.826114,
		0.834498,
		0.842738,
		0.850827,
		0.858756,
		0.86652,
		0.874111,
		0.881522,
		0.888746,
		0.895777,
		0.902608,
		0.909232,
		0.915644,
		0.921837,
		0.927805,
		0.933544,
		0.939048,
		0.944313,
		0.949334,
		0.954108,
		0.958631,
		0.962903,
		0.96692,
		0.970681,
		0.974188,
		0.977439,
		0.980438,
		0.983187,
		0.985689,
		0.987949,
		0.989975,
		0.991772,
		0.99335,
		0.994719,
		0.99589,
		0.996876,
		0.997689,
		0.998347,
		0.998863,
		0.999256,
		0.999543,
		0.999741,
		0.999869,
		0.999944,
		0.999981,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999979,
		0.999933,
		0.999838,
		0.999666,
		0.999385,
		0.99896,
		0.998347,
		0.997501,
		0.996374,
		0.994912,
		0.99306,
		0.990759,
		0.987949,
		0.984569,
		0.980556,
		0.975845,
		0.970373,
		0.964076,
		0.956891,
		0.948756,
		0.939609,
		0.929393,
		0.918052,
		0.905532,
		0.891784,
		0.876762,
		0.860425,
		0.842738,
		0.823672,
		0.803203,
		0.781314,
		0.758,
		0.73326,
		0.707107,
		0.679561,
		0.650658,
		0.620445,
		0.588982,
		0.556348,
		0.522638,
		0.487964,
		0.452462,
		0.416289,
		0.379627,
		0.342684,
		0.3057,
		0.268945,
		0.232725,
		0.197385,
		0.163312,
		0.130939,
		0.100747,
		0.073275,
		0.049119,
		0.028941,
		0.013474,
		0.003529,
		0.604481,
		0.614173,
		0.623865,
		0.63355,
		0.643226,
		0.652886,
		0.662527,
		0.672143,
		0.681729,
		0.69128,
		0.700791,
		0.710256,
		0.71967,
		0.729027,
		0.738322,
		0.747548,
		0.756699,
		0.76577,
		0.774754,
		0.783645,
		0.792437,
		0.801123,
		0.809696,
		0.818151,
		0.826481,
		0.834679,
		0.842738,
		0.850653,
		0.858415,
		0.866019,
		0.873458,
		0.880725,
		0.887815,
		0.89472,
		0.901435,
		0.907953,
		0.914268,
		0.920376,
		0.92627,
		0.931946,
		0.937398,
		0.942622,
		0.947615,
		0.952372,
		0.956891,
		0.961169,
		0.965204,
		0.968996,
		0.972543,
		0.975845,
		0.978905,
		0.981724,
		0.984305,
		0.986651,
		0.98877,
		0.990665,
		0.992346,
		0.993819,
		0.995096,
		0.996187,
		0.997104,
		0.99786,
		0.99847,
		0.998949,
		0.999313,
		0.999578,
		0.999762,
		0.99988,
		0.999948,
		0.999983,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999981,
		0.999939,
		0.999851,
		0.999694,
		0.999437,
		0.999046,
		0.998483,
		0.997707,
		0.996672,
		0.995329,
		0.993626,
		0.99151,
		0.988925,
		0.985814,
		0.982117,
		0.977776,
		0.97273,
		0.96692,
		0.960286,
		0.952769,
		0.944313,
		0.93486,
		0.924359,
		0.912756,
		0.900005,
		0.886059,
		0.870879,
		0.854428,
		0.836674,
		0.817592,
		0.797161,
		0.77537,
		0.752213,
		0.727694,
		0.701826,
		0.674631,
		0.646144,
		0.61641,
		0.58549,
		0.553458,
		0.520404,
		0.486437,
		0.451682,
		0.416289,
		0.380428,
		0.344293,
		0.308108,
		0.272124,
		0.236626,
		0.201932,
		0.168399,
		0.136427,
		0.106458,
		0.078987,
		0.05456,
		0.033782,
		0.017323,
		0.00592,
		0.000387,
		0.657196,
		0.666622,
		0.676022,
		0.685391,
		0.694725,
		0.704019,
		0.713267,
		0.722464,
		0.731605,
		0.740684,
		0.749696,
		0.758636,
		0.767497,
		0.776274,
		0.784961,
		0.793551,
		0.80204,
		0.810421,
		0.818687,
		0.826833,
		0.834852,
		0.842738,
		0.850486,
		0.858088,
		0.865538,
		0.872831,
		0.87996,
		0.88692,
		0.893703,
		0.900305,
		0.90672,
		0.912942,
		0.918966,
		0.924786,
		0.930398,
		0.935798,
		0.94098,
		0.945942,
		0.950679,
		0.955189,
		0.959469,
		0.963517,
		0.967332,
		0.970913,
		0.97426,
		0.977373,
		0.980254,
		0.982906,
		0.985332,
		0.987535,
		0.989522,
		0.991297,
		0.99287,
		0.994248,
		0.99544,
		0.996457,
		0.997312,
		0.998016,
		0.998583,
		0.999027,
		0.999365,
		0.99961,
		0.99978,
		0.999889,
		0.999952,
		0.999984,
		0.999997,
		1,
		1,
		1,
		0.999996,
		0.999982,
		0.999944,
		0.999863,
		0.999719,
		0.999482,
		0.999123,
		0.998606,
		0.997893,
		0.99694,
		0.995704,
		0.994136,
		0.992187,
		0.989805,
		0.986936,
		0.983525,
		0.979518,
		0.974858,
		0.969489,
		0.963355,
		0.956401,
		0.948572,
		0.939815,
		0.930079,
		0.919314,
		0.907473,
		0.894513,
		0.880394,
		0.865077,
		0.848532,
		0.830729,
		0.811648,
		0.79127,
		0.769588,
		0.746597,
		0.722303,
		0.69672,
		0.669872,
		0.641793,
		0.612528,
		0.582134,
		0.550683,
		0.518262,
		0.484973,
		0.450936,
		0.416289,
		0.381194,
		0.345833,
		0.310415,
		0.275172,
		0.24037,
		0.206304,
		0.173305,
		0.14174,
		0.112019,
		0.084596,
		0.059973,
		0.038704,
		0.021402,
		0.008741,
		0.001462,
		0.707107,
		0.716147,
		0.725136,
		0.73407,
		0.742944,
		0.751752,
		0.760489,
		0.769149,
		0.777728,
		0.78622,
		0.794618,
		0.802918,
		0.811114,
		0.8192,
		0.827169,
		0.835018,
		0.842738,
		0.850326,
		0.857774,
		0.865077,
		0.87223,
		0.879226,
		0.886059,
		0.892725,
		0.899218,
		0.905532,
		0.911662,
		0.917603,
		0.92335,
		0.928899,
		0.934245,
		0.939384,
		0.944313,
		0.949027,
		0.953525,
		0.957802,
		0.961859,
		0.965692,
		0.969301,
		0.972685,
		0.975845,
		0.978782,
		0.981498,
		0.983995,
		0.986276,
		0.988347,
		0.990212,
		0.991877,
		0.99335,
		0.99464,
		0.995754,
		0.996704,
		0.997501,
		0.998157,
		0.998685,
		0.999098,
		0.999412,
		0.999639,
		0.999797,
		0.999897,
		0.999956,
		0.999985,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999984,
		0.999948,
		0.999874,
		0.999741,
		0.999524,
		0.999193,
		0.998717,
		0.99806,
		0.997182,
		0.996042,
		0.994596,
		0.992798,
		0.990599,
		0.987949,
		0.984799,
		0.981094,
		0.976785,
		0.971816,
		0.966137,
		0.959694,
		0.952436,
		0.944313,
		0.935275,
		0.925275,
		0.914268,
		0.902212,
		0.889065,
		0.874792,
		0.859359,
		0.842738,
		0.824905,
		0.805839,
		0.785527,
		0.763962,
		0.741144,
		0.717079,
		0.691782,
		0.665277,
		0.637598,
		0.608789,
		0.578906,
		0.548017,
		0.516206,
		0.483569,
		0.450219,
		0.416289,
		0.381929,
		0.34731,
		0.312626,
		0.278096,
		0.243966,
		0.210511,
		0.178037,
		0.146884,
		0.117431,
		0.090096,
		0.065341,
		0.043674,
		0.025659,
		0.011912,
		0.003111,
		0.753719,
		0.762263,
		0.770731,
		0.779121,
		0.787425,
		0.79564,
		0.803759,
		0.811778,
		0.819691,
		0.827492,
		0.835177,
		0.842738,
		0.850172,
		0.857473,
		0.864634,
		0.871652,
		0.878519,
		0.885232,
		0.891784,
		0.89817,
		0.904386,
		0.910426,
		0.916286,
		0.921961,
		0.927446,
		0.932739,
		0.937833,
		0.942727,
		0.947416,
		0.951898,
		0.95617,
		0.96023,
		0.964076,
		0.967708,
		0.971124,
		0.974325,
		0.977312,
		0.980084,
		0.982646,
		0.984999,
		0.987147,
		0.989094,
		0.990846,
		0.99241,
		0.993791,
		0.994999,
		0.996042,
		0.99693,
		0.997675,
		0.998286,
		0.998778,
		0.999163,
		0.999454,
		0.999666,
		0.999812,
		0.999905,
		0.999959,
		0.999987,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999985,
		0.999952,
		0.999884,
		0.999762,
		0.999561,
		0.999256,
		0.998817,
		0.99821,
		0.9974,
		0.996348,
		0.995012,
		0.99335,
		0.991318,
		0.988867,
		0.985952,
		0.982523,
		0.978531,
		0.973927,
		0.968661,
		0.962684,
		0.955947,
		0.948403,
		0.940003,
		0.930704,
		0.920461,
		0.909232,
		0.89698,
		0.883666,
		0.869259,
		0.853729,
		0.837049,
		0.8192,
		0.800163,
		0.779928,
		0.75849,
		0.73585,
		0.712016,
		0.687003,
		0.660837,
		0.63355,
		0.605186,
		0.575799,
		0.545454,
		0.51423,
		0.482221,
		0.449532,
		0.416289,
		0.382633,
		0.348725,
		0.314747,
		0.280903,
		0.247423,
		0.214561,
		0.182603,
		0.151865,
		0.122695,
		0.095481,
		0.070648,
		0.048665,
		0.030049,
		0.015366,
		0.005236,
		0.000342,
		0.796619,
		0.804566,
		0.812415,
		0.820162,
		0.827802,
		0.835329,
		0.842738,
		0.850025,
		0.857183,
		0.864209,
		0.871096,
		0.877839,
		0.884434,
		0.890876,
		0.897159,
		0.90328,
		0.909232,
		0.915012,
		0.920616,
		0.926039,
		0.931276,
		0.936326,
		0.941183,
		0.945845,
		0.950308,
		0.954571,
		0.958631,
		0.962487,
		0.966137,
		0.96958,
		0.972816,
		0.975845,
		0.978669,
		0.981289,
		0.983707,
		0.985926,
		0.987949,
		0.989783,
		0.991431,
		0.992899,
		0.994196,
		0.995329,
		0.996306,
		0.997137,
		0.997833,
		0.998404,
		0.998863,
		0.999222,
		0.999493,
		0.99969,
		0.999825,
		0.999912,
		0.999962,
		0.999988,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999986,
		0.999956,
		0.999893,
		0.99978,
		0.999595,
		0.999313,
		0.998907,
		0.998347,
		0.997598,
		0.996625,
		0.995389,
		0.993851,
		0.991969,
		0.9897,
		0.986998,
		0.983819,
		0.980117,
		0.975845,
		0.970957,
		0.965405,
		0.959144,
		0.952128,
		0.944313,
		0.935655,
		0.926112,
		0.915644,
		0.904213,
		0.891784,
		0.878322,
		0.863799,
		0.848188,
		0.831466,
		0.813614,
		0.794618,
		0.77447,
		0.753166,
		0.730708,
		0.707107,
		0.682378,
		0.656545,
		0.629643,
		0.601712,
		0.572806,
		0.542987,
		0.512331,
		0.480925,
		0.448872,
		0.416289,
		0.383309,
		0.350084,
		0.316784,
		0.283601,
		0.250747,
		0.218463,
		0.187012,
		0.156687,
		0.127813,
		0.100747,
		0.075884,
		0.053654,
		0.034535,
		0.019046,
		0.007758,
		0.001294,
		0.835475,
		0.842738,
		0.849883,
		0.856905,
		0.863799,
		0.870561,
		0.877185,
		0.883666,
		0.890001,
		0.896184,
		0.902212,
		0.908079,
		0.913781,
		0.919314,
		0.924674,
		0.929857,
		0.93486,
		0.93968,
		0.944313,
		0.948756,
		0.953007,
		0.957064,
		0.960924,
		0.964588,
		0.968053,
		0.971319,
		0.974386,
		0.977255,
		0.979927,
		0.982404,
		0.984688,
		0.986782,
		0.988691,
		0.990418,
		0.991969,
		0.99335,
		0.994569,
		0.995632,
		0.996548,
		0.997327,
		0.997978,
		0.998512,
		0.998941,
		0.999276,
		0.999529,
		0.999712,
		0.999838,
		0.999918,
		0.999965,
		0.999988,
		0.999998,
		1,
		1,
		1,
		0.999997,
		0.999987,
		0.999959,
		0.999901,
		0.999797,
		0.999625,
		0.999365,
		0.998989,
		0.99847,
		0.997777,
		0.996876,
		0.995731,
		0.994306,
		0.992561,
		0.990456,
		0.987949,
		0.984999,
		0.981561,
		0.977591,
		0.973047,
		0.967884,
		0.962058,
		0.955526,
		0.948246,
		0.940176,
		0.931276,
		0.921508,
		0.910833,
		0.899218,
		0.886629,
		0.873037,
		0.858415,
		0.842738,
		0.825987,
		0.808146,
		0.789202,
		0.769149,
		0.747985,
		0.725714,
		0.702346,
		0.677898,
		0.652395,
		0.625869,
		0.598361,
		0.569922,
		0.540612,
		0.510503,
		0.47968,
		0.448238,
		0.416289,
		0.383959,
		0.35139,
		0.318742,
		0.286194,
		0.253947,
		0.222223,
		0.191269,
		0.161357,
		0.132788,
		0.105894,
		0.081039,
		0.058623,
		0.039085,
		0.022905,
		0.010606,
		0.002763,
		0.870045,
		0.876554,
		0.882926,
		0.889157,
		0.895243,
		0.90118,
		0.906963,
		0.912588,
		0.918052,
		0.92335,
		0.928479,
		0.933436,
		0.938217,
		0.942819,
		0.947239,
		0.951476,
		0.955526,
		0.959389,
		0.963062,
		0.966545,
		0.969836,
		0.972936,
		0.975845,
		0.978564,
		0.981094,
		0.983438,
		0.985597,
		0.987575,
		0.989376,
		0.991005,
		0.992466,
		0.993766,
		0.994912,
		0.995911,
		0.996771,
		0.997501,
		0.998111,
		0.998612,
		0.999012,
		0.999325,
		0.999561,
		0.999732,
		0.999849,
		0.999924,
		0.999968,
		0.999989,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999988,
		0.999962,
		0.999909,
		0.999812,
		0.999653,
		0.999412,
		0.999064,
		0.998583,
		0.99794,
		0.997104,
		0.996042,
		0.994719,
		0.993099,
		0.991144,
		0.988816,
		0.986073,
		0.982876,
		0.979184,
		0.974954,
		0.970147,
		0.96472,
		0.958631,
		0.951842,
		0.944313,
		0.936004,
		0.926878,
		0.9169,
		0.906036,
		0.894253,
		0.881522,
		0.867815,
		0.853109,
		0.837381,
		0.820615,
		0.802795,
		0.783913,
		0.763962,
		0.742944,
		0.720862,
		0.697728,
		0.673559,
		0.64838,
		0.622222,
		0.595126,
		0.56714,
		0.538323,
		0.508744,
		0.478481,
		0.447628,
		0.416289,
		0.384583,
		0.352645,
		0.320624,
		0.28869,
		0.257029,
		0.225849,
		0.195381,
		0.165879,
		0.137622,
		0.110919,
		0.086107,
		0.063557,
		0.043674,
		0.026902,
		0.013723,
		0.004665,
		0.000304,
		0.900183,
		0.905884,
		0.911434,
		0.916829,
		0.922066,
		0.927141,
		0.932051,
		0.936792,
		0.941362,
		0.945759,
		0.949978,
		0.95402,
		0.957881,
		0.96156,
		0.965057,
		0.968369,
		0.971498,
		0.974442,
		0.977203,
		0.979781,
		0.982179,
		0.984397,
		0.98644,
		0.98831,
		0.990011,
		0.991548,
		0.992925,
		0.99415,
		0.995229,
		0.996168,
		0.996976,
		0.997662,
		0.998234,
		0.998703,
		0.999078,
		0.99937,
		0.999591,
		0.99975,
		0.99986,
		0.999929,
		0.99997,
		0.99999,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999989,
		0.999965,
		0.999915,
		0.999825,
		0.999678,
		0.999454,
		0.999131,
		0.998685,
		0.998088,
		0.997312,
		0.996325,
		0.995096,
		0.99359,
		0.991772,
		0.989606,
		0.987053,
		0.984077,
		0.980638,
		0.976697,
		0.972215,
		0.967154,
		0.961473,
		0.955135,
		0.948101,
		0.940336,
		0.931803,
		0.922467,
		0.912296,
		0.901258,
		0.889324,
		0.876466,
		0.862659,
		0.847883,
		0.832117,
		0.815347,
		0.797559,
		0.778747,
		0.758906,
		0.738037,
		0.716147,
		0.693246,
		0.669354,
		0.644493,
		0.618696,
		0.592002,
		0.564456,
		0.536117,
		0.507048,
		0.477327,
		0.447041,
		0.416289,
		0.385184,
		0.353853,
		0.322436,
		0.291093,
		0.259999,
		0.229348,
		0.199356,
		0.170259,
		0.14232,
		0.115823,
		0.091084,
		0.068445,
		0.048281,
		0.031004,
		0.017058,
		0.006932,
		0.001154,
		0.925841,
		0.930704,
		0.935405,
		0.939942,
		0.944313,
		0.948514,
		0.952544,
		0.956401,
		0.960083,
		0.96359,
		0.96692,
		0.970072,
		0.973047,
		0.975845,
		0.978467,
		0.980913,
		0.983187,
		0.985289,
		0.987222,
		0.988991,
		0.990599,
		0.99205,
		0.99335,
		0.994505,
		0.995521,
		0.996405,
		0.997165,
		0.997809,
		0.998347,
		0.998786,
		0.999138,
		0.999412,
		0.999618,
		0.999767,
		0.999869,
		0.999934,
		0.999972,
		0.999991,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.99999,
		0.999968,
		0.999921,
		0.999838,
		0.999701,
		0.999493,
		0.999193,
		0.998778,
		0.998223,
		0.997501,
		0.996584,
		0.99544,
		0.994038,
		0.992346,
		0.990328,
		0.987949,
		0.985175,
		0.981968,
		0.978292,
		0.974109,
		0.969383,
		0.964076,
		0.958153,
		0.951577,
		0.944313,
		0.936326,
		0.927583,
		0.918052,
		0.907703,
		0.896506,
		0.884434,
		0.871464,
		0.857572,
		0.842738,
		0.826947,
		0.810183,
		0.792437,
		0.773702,
		0.753975,
		0.73326,
		0.711563,
		0.688896,
		0.665277,
		0.64073,
		0.615286,
		0.588982,
		0.561865,
		0.533987,
		0.505413,
		0.476215,
		0.446476,
		0.416289,
		0.385763,
		0.355015,
		0.324181,
		0.293408,
		0.262862,
		0.232725,
		0.203198,
		0.174504,
		0.146884,
		0.120607,
		0.095965,
		0.073275,
		0.052888,
		0.035182,
		0.02057,
		0.009504,
		0.00247,
		0.947082,
		0.951098,
		0.954949,
		0.958631,
		0.962145,
		0.965488,
		0.968661,
		0.971663,
		0.974494,
		0.977154,
		0.979645,
		0.981968,
		0.984125,
		0.986118,
		0.987949,
		0.989624,
		0.991144,
		0.992516,
		0.993744,
		0.994834,
		0.995792,
		0.996625,
		0.99734,
		0.997946,
		0.998451,
		0.998863,
		0.999193,
		0.99945,
		0.999643,
		0.999782,
		0.999878,
		0.999939,
		0.999974,
		0.999991,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.99999,
		0.99997,
		0.999927,
		0.999849,
		0.999722,
		0.999529,
		0.999249,
		0.998863,
		0.998347,
		0.997675,
		0.99682,
		0.995754,
		0.994448,
		0.99287,
		0.990988,
		0.98877,
		0.986181,
		0.983187,
		0.979753,
		0.975845,
		0.971428,
		0.966466,
		0.960924,
		0.954769,
		0.947967,
		0.940484,
		0.932289,
		0.92335,
		0.913638,
		0.903125,
		0.891784,
		0.879589,
		0.86652,
		0.852555,
		0.837676,
		0.821869,
		0.805122,
		0.787425,
		0.768774,
		0.749168,
		0.72861,
		0.707107,
		0.684672,
		0.661323,
		0.637084,
		0.611985,
		0.586062,
		0.559361,
		0.531932,
		0.503836,
		0.475143,
		0.44593,
		0.416289,
		0.38632,
		0.356136,
		0.325863,
		0.295641,
		0.265625,
		0.235987,
		0.206915,
		0.178617,
		0.151319,
		0.125272,
		0.100747,
		0.078042,
		0.057479,
		0.039412,
		0.024223,
		0.012329,
		0.004182,
		0.000272,
		0.964076,
		0.967266,
		0.970291,
		0.97315,
		0.975845,
		0.978376,
		0.980745,
		0.982951,
		0.984999,
		0.986889,
		0.988626,
		0.990212,
		0.991651,
		0.992949,
		0.994109,
		0.995138,
		0.996042,
		0.996827,
		0.997501,
		0.998072,
		0.998546,
		0.998934,
		0.999244,
		0.999485,
		0.999666,
		0.999797,
		0.999886,
		0.999943,
		0.999976,
		0.999992,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999991,
		0.999972,
		0.999932,
		0.99986,
		0.999741,
		0.999561,
		0.999301,
		0.998941,
		0.99846,
		0.997833,
		0.997036,
		0.996042,
		0.994823,
		0.99335,
		0.991594,
		0.989522,
		0.987103,
		0.984305,
		0.981094,
		0.977439,
		0.973306,
		0.968661,
		0.963472,
		0.957705,
		0.95133,
		0.944313,
		0.936624,
		0.928233,
		0.919112,
		0.909232,
		0.898568,
		0.887096,
		0.874792,
		0.861636,
		0.84761,
		0.832697,
		0.816885,
		0.800163,
		0.782523,
		0.763962,
		0.74448,
		0.724081,
		0.702773,
		0.680569,
		0.657487,
		0.63355,
		0.608789,
		0.583238,
		0.556941,
		0.529947,
		0.502313,
		0.474108,
		0.445405,
		0.416289,
		0.386858,
		0.357216,
		0.327485,
		0.297795,
		0.268292,
		0.239139,
		0.210511,
		0.182603,
		0.155629,
		0.12982,
		0.105431,
		0.082738,
		0.062044,
		0.043674,
		0.027987,
		0.015366,
		0.006231,
		0.001035,
		0.977109,
		0.979518,
		0.981771,
		0.983869,
		0.985814,
		0.987608,
		0.989256,
		0.990759,
		0.992122,
		0.99335,
		0.994448,
		0.995421,
		0.996274,
		0.997015,
		0.997651,
		0.998188,
		0.998635,
		0.999,
		0.999291,
		0.999517,
		0.999687,
		0.99981,
		0.999893,
		0.999947,
		0.999977,
		0.999993,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999992,
		0.999974,
		0.999937,
		0.999869,
		0.999759,
		0.999591,
		0.999348,
		0.999012,
		0.998563,
		0.997978,
		0.997234,
		0.996306,
		0.995167,
		0.993791,
		0.992149,
		0.990212,
		0.987949,
		0.985332,
		0.982327,
		0.978905,
		0.975034,
		0.970681,
		0.965817,
		0.960409,
		0.954428,
		0.947842,
		0.940621,
		0.932739,
		0.924165,
		0.914874,
		0.904839,
		0.894038,
		0.882447,
		0.870045,
		0.856815,
		0.842738,
		0.827802,
		0.811993,
		0.795304,
		0.777728,
		0.759263,
		0.739908,
		0.71967,
		0.698557,
		0.676582,
		0.653763,
		0.630124,
		0.605693,
		0.580504,
		0.5546,
		0.528027,
		0.500843,
		0.473108,
		0.444897,
		0.416289,
		0.387376,
		0.358259,
		0.32905,
		0.299874,
		0.270869,
		0.242186,
		0.213992,
		0.186469,
		0.159817,
		0.134253,
		0.110014,
		0.08736,
		0.066571,
		0.047952,
		0.031834,
		0.018575,
		0.008564,
		0.002221,
		0.986574,
		0.988279,
		0.989843,
		0.991269,
		0.992561,
		0.993724,
		0.994763,
		0.995683,
		0.99649,
		0.997189,
		0.997789,
		0.998296,
		0.998717,
		0.999061,
		0.999334,
		0.999547,
		0.999707,
		0.999822,
		0.9999,
		0.99995,
		0.999979,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999998,
		0.999992,
		0.999976,
		0.999941,
		0.999878,
		0.999775,
		0.999618,
		0.999391,
		0.999078,
		0.998658,
		0.998111,
		0.997416,
		0.996548,
		0.995483,
		0.994196,
		0.99266,
		0.990846,
		0.988728,
		0.986276,
		0.983462,
		0.980254,
		0.976625,
		0.972543,
		0.967979,
		0.962903,
		0.957286,
		0.951098,
		0.944313,
		0.9369,
		0.928835,
		0.92009,
		0.910641,
		0.900464,
		0.889537,
		0.877839,
		0.865352,
		0.852057,
		0.837941,
		0.82299,
		0.807193,
		0.790544,
		0.773038,
		0.754672,
		0.735449,
		0.715374,
		0.694455,
		0.672708,
		0.650148,
		0.6268,
		0.602692,
		0.577857,
		0.552335,
		0.526172,
		0.499421,
		0.472143,
		0.444407,
		0.416289,
		0.387877,
		0.359265,
		0.330561,
		0.301882,
		0.273359,
		0.245133,
		0.217363,
		0.190218,
		0.163887,
		0.138573,
		0.114498,
		0.091904,
		0.071053,
		0.05223,
		0.035743,
		0.021925,
		0.011138,
		0.003771,
		0.000244,
		0.992969,
		0.994072,
		0.995056,
		0.995927,
		0.99669,
		0.997351,
		0.997917,
		0.998396,
		0.998793,
		0.999117,
		0.999375,
		0.999575,
		0.999725,
		0.999833,
		0.999906,
		0.999953,
		0.99998,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.999977,
		0.999945,
		0.999886,
		0.99979,
		0.999643,
		0.999431,
		0.999138,
		0.998745,
		0.998234,
		0.997583,
		0.996771,
		0.995774,
		0.994569,
		0.99313,
		0.991431,
		0.989445,
		0.987147,
		0.984507,
		0.981498,
		0.978092,
		0.97426,
		0.969973,
		0.965204,
		0.959925,
		0.954108,
		0.947725,
		0.940749,
		0.933156,
		0.924919,
		0.916014,
		0.906419,
		0.896111,
		0.88507,
		0.873276,
		0.860713,
		0.847365,
		0.833218,
		0.81826,
		0.802484,
		0.785882,
		0.76845,
		0.750188,
		0.731099,
		0.711187,
		0.690463,
		0.66894,
		0.646637,
		0.623575,
		0.599783,
		0.575292,
		0.550142,
		0.524376,
		0.498046,
		0.47121,
		0.443933,
		0.416289,
		0.38836,
		0.360237,
		0.332021,
		0.303823,
		0.275766,
		0.247986,
		0.220628,
		0.193856,
		0.167843,
		0.142782,
		0.118882,
		0.096367,
		0.075484,
		0.056497,
		0.039695,
		0.025389,
		0.013913,
		0.005631,
		0.000934,
		0.996876,
		0.997501,
		0.998037,
		0.998488,
		0.998863,
		0.999169,
		0.999412,
		0.9996,
		0.999741,
		0.999843,
		0.999912,
		0.999956,
		0.999981,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.999979,
		0.999948,
		0.999893,
		0.999803,
		0.999666,
		0.999468,
		0.999193,
		0.998826,
		0.998347,
		0.997737,
		0.996976,
		0.996042,
		0.994912,
		0.993563,
		0.991969,
		0.990107,
		0.987949,
		0.985471,
		0.982646,
		0.979446,
		0.975845,
		0.971816,
		0.967332,
		0.962366,
		0.956891,
		0.950882,
		0.944313,
		0.937158,
		0.929393,
		0.920996,
		0.911942,
		0.902212,
		0.891784,
		0.880639,
		0.86876,
		0.856131,
		0.842738,
		0.828569,
		0.813614,
		0.797864,
		0.781314,
		0.763962,
		0.745808,
		0.726854,
		0.707107,
		0.686577,
		0.665277,
		0.643226,
		0.620445,
		0.596961,
		0.572806,
		0.548017,
		0.522638,
		0.496716,
		0.470308,
		0.443475,
		0.416289,
		0.388827,
		0.361176,
		0.333432,
		0.3057,
		0.278096,
		0.250747,
		0.223793,
		0.197385,
		0.171689,
		0.146884,
		0.123166,
		0.100747,
		0.079857,
		0.060743,
		0.043674,
		0.028941,
		0.016857,
		0.007758,
		0.002008,
		0.998929,
		0.999217,
		0.999446,
		0.999624,
		0.999757,
		0.999852,
		0.999917,
		0.999959,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.99998,
		0.999952,
		0.9999,
		0.999816,
		0.999687,
		0.999501,
		0.999244,
		0.9989,
		0.998451,
		0.997879,
		0.997165,
		0.996289,
		0.995229,
		0.993962,
		0.992466,
		0.990717,
		0.988691,
		0.986362,
		0.983707,
		0.980698,
		0.977312,
		0.973521,
		0.969301,
		0.964625,
		0.959469,
		0.953807,
		0.947615,
		0.940869,
		0.933544,
		0.925619,
		0.917071,
		0.907879,
		0.898023,
		0.887485,
		0.876247,
		0.864293,
		0.851607,
		0.838179,
		0.823996,
		0.809049,
		0.793332,
		0.776841,
		0.759572,
		0.741528,
		0.722712,
		0.70313,
		0.682792,
		0.661713,
		0.63991,
		0.617404,
		0.594222,
		0.570395,
		0.545959,
		0.520954,
		0.495428,
		0.469434,
		0.443032,
		0.416289,
		0.389279,
		0.362085,
		0.334797,
		0.307516,
		0.280351,
		0.253422,
		0.226862,
		0.200812,
		0.175429,
		0.150882,
		0.127354,
		0.105044,
		0.084169,
		0.064959,
		0.047667,
		0.032563,
		0.019939,
		0.010111,
		0.003417,
		0.000221,
		0.999771,
		0.999861,
		0.999922,
		0.999961,
		0.999984,
		0.999995,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999981,
		0.999955,
		0.999906,
		0.999827,
		0.999707,
		0.999532,
		0.999291,
		0.998968,
		0.998546,
		0.99801,
		0.99734,
		0.996517,
		0.995521,
		0.994331,
		0.992925,
		0.991281,
		0.989376,
		0.987186,
		0.984688,
		0.981857,
		0.978669,
		0.9751,
		0.971124,
		0.966719,
		0.961859,
		0.95652,
		0.950679,
		0.944313,
		0.937398,
		0.929913,
		0.921837,
		0.913148,
		0.903828,
		0.893857,
		0.883219,
		0.871897,
		0.859876,
		0.847143,
		0.833687,
		0.819497,
		0.804566,
		0.788888,
		0.772459,
		0.755278,
		0.737347,
		0.718669,
		0.699252,
		0.679106,
		0.658245,
		0.636686,
		0.61445,
		0.591564,
		0.568056,
		0.543963,
		0.519322,
		0.49418,
		0.468589,
		0.442603,
		0.416289,
		0.389717,
		0.362964,
		0.336118,
		0.309274,
		0.282534,
		0.256014,
		0.229837,
		0.204139,
		0.179066,
		0.154777,
		0.131445,
		0.109257,
		0.088416,
		0.069138,
		0.05166,
		0.036235,
		0.023136,
		0.012657,
		0.005114,
		0.000846,
		0.999985,
		0.999995,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999957,
		0.999912,
		0.999838,
		0.999725,
		0.999561,
		0.999334,
		0.999031,
		0.998635,
		0.998131,
		0.997501,
		0.996728,
		0.995792,
		0.994673,
		0.99335,
		0.991804,
		0.990011,
		0.987949,
		0.985597,
		0.982931,
		0.979927,
		0.976563,
		0.972816,
		0.968661,
		0.964076,
		0.959038,
		0.953525,
		0.947512,
		0.94098,
		0.933906,
		0.92627,
		0.918052,
		0.909232,
		0.899793,
		0.889716,
		0.878987,
		0.86759,
		0.855511,
		0.842738,
		0.829262,
		0.815072,
		0.800163,
		0.784529,
		0.768167,
		0.751077,
		0.73326,
		0.714722,
		0.69547,
		0.675515,
		0.654869,
		0.63355,
		0.61158,
		0.588982,
		0.565786,
		0.542026,
		0.51774,
		0.492972,
		0.467769,
		0.442188,
		0.416289,
		0.39014,
		0.363815,
		0.337397,
		0.310976,
		0.28465,
		0.258527,
		0.232725,
		0.207371,
		0.182603,
		0.158573,
		0.135442,
		0.113386,
		0.092596,
		0.073275,
		0.055645,
		0.039943,
		0.026425,
		0.015366,
		0.00706,
		0.001825,
		1,
		1,
		0.999999,
		0.999995,
		0.999984,
		0.99996,
		0.999917,
		0.999848,
		0.999741,
		0.999588,
		0.999375,
		0.999089,
		0.998717,
		0.998243,
		0.997651,
		0.996923,
		0.996042,
		0.994989,
		0.993744,
		0.992288,
		0.990599,
		0.988657,
		0.98644,
		0.983926,
		0.981094,
		0.977922,
		0.974386,
		0.970465,
		0.966137,
		0.961379,
		0.95617,
		0.950488,
		0.944313,
		0.937623,
		0.930398,
		0.92262,
		0.914268,
		0.905327,
		0.895777,
		0.885604,
		0.874792,
		0.863328,
		0.851199,
		0.838394,
		0.824905,
		0.810721,
		0.795839,
		0.780253,
		0.763962,
		0.746966,
		0.729267,
		0.710869,
		0.691782,
		0.672015,
		0.651582,
		0.630499,
		0.608789,
		0.586474,
		0.563582,
		0.540147,
		0.516206,
		0.4918,
		0.466975,
		0.441786,
		0.416289,
		0.39055,
		0.36464,
		0.338636,
		0.312626,
		0.286701,
		0.260965,
		0.235528,
		0.210511,
		0.186045,
		0.162273,
		0.139346,
		0.117431,
		0.096707,
		0.077365,
		0.059613,
		0.043674,
		0.029788,
		0.018211,
		0.00922,
		0.003111,
		0.000201,
		0.999985,
		0.999962,
		0.999922,
		0.999857,
		0.999757,
		0.999612,
		0.999412,
		0.999143,
		0.998793,
		0.998347,
		0.997789,
		0.997104,
		0.996274,
		0.995282,
		0.994109,
		0.992736,
		0.991144,
		0.989313,
		0.987222,
		0.984851,
		0.982179,
		0.979184,
		0.975845,
		0.972142,
		0.968053,
		0.963556,
		0.958631,
		0.953258,
		0.947416,
		0.941085,
		0.934245,
		0.926878,
		0.918966,
		0.91049,
		0.901435,
		0.891784,
		0.881522,
		0.870636,
		0.859113,
		0.846941,
		0.834112,
		0.820615,
		0.806444,
		0.791594,
		0.776061,
		0.759844,
		0.742944,
		0.725363,
		0.707107,
		0.688183,
		0.668603,
		0.64838,
		0.62753,
		0.606074,
		0.584036,
		0.561441,
		0.538323,
		0.514717,
		0.490663,
		0.466205,
		0.441396,
		0.416289,
		0.390948,
		0.365439,
		0.339837,
		0.314225,
		0.28869,
		0.263329,
		0.238249,
		0.213563,
		0.189395,
		0.165879,
		0.14316,
		0.121393,
		0.100747,
		0.081404,
		0.063557,
		0.047417,
		0.033208,
		0.02117,
		0.011564,
		0.004665,
		0.000771,
		0.999771,
		0.999635,
		0.999446,
		0.999193,
		0.998863,
		0.998443,
		0.997917,
		0.997272,
		0.99649,
		0.995554,
		0.994448,
		0.993153,
		0.991651,
		0.989923,
		0.987949,
		0.985711,
		0.983187,
		0.980357,
		0.977203,
		0.973702,
		0.969836,
		0.965583,
		0.960924,
		0.955839,
		0.950308,
		0.944313,
		0.937833,
		0.930852,
		0.92335,
		0.915312,
		0.90672,
		0.897559,
		0.887815,
		0.877473,
		0.86652,
		0.854945,
		0.842738,
		0.82989,
		0.816392,
		0.802239,
		0.787425,
		0.771949,
		0.75581,
		0.739008,
		0.721547,
		0.703432,
		0.684672,
		0.665277,
		0.645261,
		0.62464,
		0.603434,
		0.581665,
		0.559361,
		0.536552,
		0.513271,
		0.489559,
		0.465459,
		0.441017,
		0.416289,
		0.391333,
		0.366214,
		0.341002,
		0.315776,
		0.290619,
		0.265625,
		0.240892,
		0.21653,
		0.192655,
		0.169394,
		0.146884,
		0.125272,
		0.104717,
		0.085389,
		0.067471,
		0.051161,
		0.03667,
		0.024223,
		0.014064,
		0.006452,
		0.001665,
		0.998929,
		0.998532,
		0.998037,
		0.997428,
		0.99669,
		0.995807,
		0.994763,
		0.993541,
		0.992122,
		0.99049,
		0.988626,
		0.98651,
		0.984125,
		0.98145,
		0.978467,
		0.975156,
		0.971498,
		0.967473,
		0.963062,
		0.958246,
		0.953007,
		0.947325,
		0.941183,
		0.934562,
		0.927446,
		0.919818,
		0.911662,
		0.902962,
		0.893703,
		0.883873,
		0.873458,
		0.862446,
		0.850827,
		0.838591,
		0.825729,
		0.812236,
		0.798105,
		0.783333,
		0.767917,
		0.751857,
		0.735156,
		0.717815,
		0.699842,
		0.681244,
		0.662033,
		0.642221,
		0.621825,
		0.600863,
		0.579359,
		0.557338,
		0.53483,
		0.511868,
		0.488488,
		0.464734,
		0.44065,
		0.416289,
		0.391707,
		0.366965,
		0.342132,
		0.317281,
		0.292492,
		0.267854,
		0.243461,
		0.219416,
		0.19583,
		0.172822,
		0.150522,
		0.12907,
		0.108615,
		0.089317,
		0.07135,
		0.054899,
		0.040162,
		0.027353,
		0.016698,
		0.008442,
		0.002844,
		0.000183,
		0.996876,
		0.996042,
		0.995056,
		0.993901,
		0.992561,
		0.991018,
		0.989256,
		0.987255,
		0.984999,
		0.982468,
		0.979645,
		0.976511,
		0.973047,
		0.969235,
		0.965057,
		0.960493,
		0.955526,
		0.950139,
		0.944313,
		0.938031,
		0.931276,
		0.924033,
		0.916286,
		0.908019,
		0.899218,
		0.889869,
		0.87996,
		0.869479,
		0.858415,
		0.846758,
		0.834498,
		0.82163,
		0.808146,
		0.794042,
		0.779315,
		0.763962,
		0.747985,
		0.731385,
		0.714166,
		0.696335,
		0.677898,
		0.658868,
		0.639258,
		0.619082,
		0.598361,
		0.577116,
		0.555372,
		0.533157,
		0.510503,
		0.487448,
		0.46403,
		0.440294,
		0.416289,
		0.39207,
		0.367695,
		0.343228,
		0.318742,
		0.294311,
		0.27002,
		0.245958,
		0.222223,
		0.198921,
		0.176164,
		0.154076,
		0.132788,
		0.112441,
		0.093187,
		0.07519,
		0.058623,
		0.043674,
		0.030544,
		0.019444,
		0.010606,
		0.004273,
		0.000705,
		0.992969,
		0.99151,
		0.989843,
		0.987949,
		0.985814,
		0.983418,
		0.980745,
		0.977776,
		0.974494,
		0.970881,
		0.96692,
		0.962592,
		0.957881,
		0.952769,
		0.947239,
		0.941275,
		0.93486,
		0.927979,
		0.920616,
		0.912756,
		0.904386,
		0.895491,
		0.886059,
		0.876079,
		0.865538,
		0.854428,
		0.842738,
		0.830462,
		0.817592,
		0.804122,
		0.790049,
		0.77537,
		0.760084,
		0.744191,
		0.727694,
		0.710598,
		0.692907,
		0.674631,
		0.655781,
		0.636368,
		0.61641,
		0.595924,
		0.574932,
		0.553458,
		0.531529,
		0.509177,
		0.486437,
		0.463346,
		0.439948,
		0.416289,
		0.392422,
		0.368403,
		0.344293,
		0.32016,
		0.296078,
		0.272124,
		0.248386,
		0.224955,
		0.201932,
		0.179424,
		0.157547,
		0.136427,
		0.116196,
		0.096998,
		0.078987,
		0.062328,
		0.047197,
		0.033782,
		0.022285,
		0.012921,
		0.00592,
		0.001526,
		0.986574,
		0.984305,
		0.981771,
		0.978957,
		0.975845,
		0.972419,
		0.968661,
		0.964555,
		0.960083,
		0.95523,
		0.949978,
		0.944313,
		0.938217,
		0.931675,
		0.924674,
		0.917197,
		0.909232,
		0.900765,
		0.891784,
		0.882275,
		0.87223,
		0.861636,
		0.850486,
		0.83877,
		0.826481,
		0.813614,
		0.800163,
		0.786125,
		0.771497,
		0.75628,
		0.740474,
		0.724081,
		0.707107,
		0.689557,
		0.67144,
		0.652767,
		0.63355,
		0.613805,
		0.59355,
		0.572806,
		0.551596,
		0.529947,
		0.507888,
		0.485454,
		0.462682,
		0.439612,
		0.416289,
		0.392764,
		0.369091,
		0.345327,
		0.321539,
		0.297795,
		0.27417,
		0.250747,
		0.227614,
		0.204865,
		0.182603,
		0.160938,
		0.139988,
		0.119879,
		0.100747,
		0.082738,
		0.066007,
		0.050721,
		0.037057,
		0.025204,
		0.015366,
		0.007758,
		0.00261,
		0.000168,
		0.977109,
		0.973858,
		0.970291,
		0.966392,
		0.962145,
		0.957534,
		0.952544,
		0.947158,
		0.941362,
		0.935141,
		0.928479,
		0.921364,
		0.913781,
		0.905717,
		0.897159,
		0.888097,
		0.878519,
		0.868415,
		0.857774,
		0.846589,
		0.834852,
		0.822556,
		0.809696,
		0.796268,
		0.782268,
		0.767695,
		0.752549,
		0.736831,
		0.720543,
		0.703692,
		0.686282,
		0.668323,
		0.649825,
		0.630801,
		0.611266,
		0.591237,
		0.570735,
		0.549783,
		0.528406,
		0.506634,
		0.484498,
		0.462036,
		0.439285,
		0.416289,
		0.393097,
		0.369759,
		0.346332,
		0.322879,
		0.299464,
		0.27616,
		0.253045,
		0.230203,
		0.207724,
		0.185705,
		0.164251,
		0.143473,
		0.123492,
		0.104436,
		0.086442,
		0.069658,
		0.05424,
		0.040357,
		0.028187,
		0.017921,
		0.009762,
		0.003928,
		0.000647,
		0.964076,
		0.959694,
		0.954949,
		0.949827,
		0.944313,
		0.938392,
		0.932051,
		0.925275,
		0.918052,
		0.910368,
		0.902212,
		0.893571,
		0.884434,
		0.874792,
		0.864634,
		0.853952,
		0.842738,
		0.830985,
		0.818687,
		0.805839,
		0.792437,
		0.778478,
		0.763962,
		0.748889,
		0.73326,
		0.717079,
		0.70035,
		0.68308,
		0.665277,
		0.646952,
		0.628118,
		0.608789,
		0.588982,
		0.568718,
		0.548017,
		0.526907,
		0.505413,
		0.483569,
		0.461407,
		0.438967,
		0.416289,
		0.39342,
		0.370409,
		0.34731,
		0.324181,
		0.301087,
		0.278096,
		0.255282,
		0.232725,
		0.210511,
		0.188732,
		0.167488,
		0.146884,
		0.127035,
		0.108062,
		0.090096,
		0.073275,
		0.057749,
		0.043674,
		0.031222,
		0.02057,
		0.011912,
		0.00545,
		0.001403,
		0.947082,
		0.941445,
		0.935405,
		0.92895,
		0.922066,
		0.914741,
		0.906963,
		0.89872,
		0.890001,
		0.880796,
		0.871096,
		0.86089,
		0.850172,
		0.838934,
		0.827169,
		0.814873,
		0.80204,
		0.788668,
		0.774754,
		0.760297,
		0.745299,
		0.729761,
		0.713686,
		0.697079,
		0.679948,
		0.6623,
		0.644146,
		0.625499,
		0.606372,
		0.586783,
		0.566751,
		0.546297,
		0.525446,
		0.504225,
		0.482664,
		0.460796,
		0.438658,
		0.416289,
		0.393734,
		0.371041,
		0.34826,
		0.325448,
		0.302666,
		0.27998,
		0.25746,
		0.235182,
		0.213228,
		0.191686,
		0.170651,
		0.150223,
		0.130509,
		0.111627,
		0.093699,
		0.076857,
		0.061241,
		0.047001,
		0.034297,
		0.023299,
		0.014187,
		0.007154,
		0.002404,
		0.000155,
		0.925841,
		0.918855,
		0.911434,
		0.903567,
		0.895243,
		0.886452,
		0.877185,
		0.867431,
		0.857183,
		0.846434,
		0.835177,
		0.823405,
		0.811114,
		0.7983,
		0.784961,
		0.771094,
		0.756699,
		0.741777,
		0.726331,
		0.710363,
		0.693878,
		0.676885,
		0.65939,
		0.641405,
		0.622942,
		0.604014,
		0.584639,
		0.564834,
		0.544621,
		0.524024,
		0.503068,
		0.481783,
		0.460201,
		0.438357,
		0.416289,
		0.39404,
		0.371655,
		0.349184,
		0.326681,
		0.304203,
		0.281814,
		0.259581,
		0.237576,
		0.215878,
		0.19457,
		0.173742,
		0.15349,
		0.133916,
		0.11513,
		0.09725,
		0.080399,
		0.064712,
		0.05033,
		0.037403,
		0.026093,
		0.01657,
		0.009016,
		0.003623,
		0.000596,
		0.900183,
		0.891784,
		0.882926,
		0.873601,
		0.863799,
		0.853514,
		0.842738,
		0.831466,
		0.819691,
		0.80741,
		0.794618,
		0.781314,
		0.767497,
		0.753166,
		0.738322,
		0.722968,
		0.707107,
		0.690745,
		0.673888,
		0.656545,
		0.638727,
		0.620445,
		0.601712,
		0.582546,
		0.562964,
		0.542987,
		0.522638,
		0.501941,
		0.480925,
		0.459622,
		0.438064,
		0.416289,
		0.394338,
		0.372254,
		0.350084,
		0.327881,
		0.3057,
		0.283601,
		0.261647,
		0.23991,
		0.218463,
		0.197385,
		0.176763,
		0.156687,
		0.137255,
		0.118572,
		0.100747,
		0.083901,
		0.068159,
		0.053654,
		0.040531,
		0.028941,
		0.019046,
		0.011016,
		0.005035,
		0.001294,
		0.870045,
		0.860201,
		0.849883,
		0.839085,
		0.827802,
		0.816028,
		0.803759,
		0.790993,
		0.777728,
		0.763962,
		0.749696,
		0.734932,
		0.71967,
		0.703917,
		0.687676,
		0.670956,
		0.653763,
		0.636109,
		0.618005,
		0.599465,
		0.580504,
		0.561141,
		0.541394,
		0.521286,
		0.500843,
		0.48009,
		0.459058,
		0.437779,
		0.416289,
		0.394628,
		0.372836,
		0.35096,
		0.32905,
		0.307157,
		0.285341,
		0.263662,
		0.242186,
		0.220985,
		0.200135,
		0.179716,
		0.159817,
		0.140529,
		0.121952,
		0.104192,
		0.08736,
		0.071577,
		0.05697,
		0.043674,
		0.031834,
		0.021601,
		0.013138,
		0.006618,
		0.002221,
		0.000143,
		0.835475,
		0.824185,
		0.812415,
		0.800163,
		0.787425,
		0.774201,
		0.760489,
		0.746289,
		0.731605,
		0.716437,
		0.700791,
		0.684672,
		0.668086,
		0.651042,
		0.63355,
		0.615622,
		0.59727,
		0.578511,
		0.559361,
		0.53984,
		0.519969,
		0.499772,
		0.479275,
		0.458508,
		0.437501,
		0.416289,
		0.39491,
		0.373403,
		0.351814,
		0.330188,
		0.308577,
		0.287037,
		0.265625,
		0.244406,
		0.223446,
		0.20282,
		0.182603,
		0.16288,
		0.143738,
		0.125272,
		0.107582,
		0.090775,
		0.074965,
		0.060273,
		0.046826,
		0.034761,
		0.024223,
		0.015366,
		0.008351,
		0.003352,
		0.000551,
		0.796619,
		0.783913,
		0.770731,
		0.757075,
		0.742944,
		0.72834,
		0.713267,
		0.697728,
		0.681729,
		0.665277,
		0.64838,
		0.631048,
		0.613292,
		0.595126,
		0.576564,
		0.557624,
		0.538323,
		0.518684,
		0.498728,
		0.478481,
		0.457972,
		0.43723,
		0.416289,
		0.395185,
		0.373956,
		0.352645,
		0.331297,
		0.309961,
		0.28869,
		0.26754,
		0.246571,
		0.225849,
		0.205443,
		0.185427,
		0.165879,
		0.146884,
		0.128533,
		0.110919,
		0.094146,
		0.07832,
		0.063557,
		0.049979,
		0.037715,
		0.026902,
		0.017685,
		0.010218,
		0.004665,
		0.001198,
		0.753719,
		0.739658,
		0.725136,
		0.710157,
		0.694725,
		0.678846,
		0.662527,
		0.645775,
		0.628601,
		0.611015,
		0.593031,
		0.574663,
		0.555928,
		0.536843,
		0.51743,
		0.497709,
		0.477707,
		0.45745,
		0.436966,
		0.416289,
		0.395453,
		0.374495,
		0.353455,
		0.332378,
		0.31131,
		0.290302,
		0.269407,
		0.248685,
		0.228195,
		0.208006,
		0.188187,
		0.168815,
		0.149968,
		0.131734,
		0.114202,
		0.09747,
		0.08164,
		0.066821,
		0.05313,
		0.040688,
		0.029626,
		0.020082,
		0.012202,
		0.00614,
		0.002059,
		0.000132,
		0.707107,
		0.691782,
		0.676022,
		0.659834,
		0.643226,
		0.626207,
		0.608789,
		0.590984,
		0.572806,
		0.554272,
		0.535398,
		0.516206,
		0.496716,
		0.476952,
		0.45694,
		0.436709,
		0.416289,
		0.395714,
		0.37502,
		0.354245,
		0.333432,
		0.312626,
		0.291874,
		0.27123,
		0.250747,
		0.230487,
		0.210511,
		0.190888,
		0.171689,
		0.152992,
		0.134877,
		0.117431,
		0.100747,
		0.084923,
		0.070061,
		0.056273,
		0.043674,
		0.032389,
		0.022547,
		0.014288,
		0.007758,
		0.003111,
		0.000511,
		0.657196,
		0.64073,
		0.623865,
		0.606611,
		0.588982,
		0.570991,
		0.552654,
		0.533987,
		0.515011,
		0.495746,
		0.476215,
		0.456443,
		0.436458,
		0.416289,
		0.395969,
		0.375532,
		0.355015,
		0.33446,
		0.313909,
		0.293408,
		0.273008,
		0.252761,
		0.232725,
		0.212959,
		0.193529,
		0.174504,
		0.155955,
		0.137962,
		0.120607,
		0.103978,
		0.088168,
		0.073275,
		0.059405,
		0.046668,
		0.035182,
		0.02507,
		0.016464,
		0.009504,
		0.004334,
		0.001112,
		0.604481,
		0.587025,
		0.569217,
		0.551073,
		0.532609,
		0.513845,
		0.494799,
		0.475496,
		0.455958,
		0.436213,
		0.416289,
		0.396218,
		0.376031,
		0.355767,
		0.335463,
		0.315161,
		0.294906,
		0.274744,
		0.254728,
		0.234912,
		0.215353,
		0.196114,
		0.17726,
		0.158861,
		0.140991,
		0.12373,
		0.107161,
		0.091374,
		0.076461,
		0.062522,
		0.049664,
		0.037997,
		0.027641,
		0.018718,
		0.011362,
		0.005711,
		0.001913,
		0.000123,
		0.549528,
		0.531263,
		0.512705,
		0.493875,
		0.474793,
		0.455485,
		0.435974,
		0.416289,
		0.39646,
		0.376519,
		0.3565,
		0.336442,
		0.316383,
		0.296367,
		0.27644,
		0.25665,
		0.237049,
		0.217694,
		0.198643,
		0.179959,
		0.16171,
		0.143965,
		0.126801,
		0.110298,
		0.094539,
		0.079615,
		0.065621,
		0.052658,
		0.04083,
		0.030251,
		0.02104,
		0.01332,
		0.007225,
		0.002895,
		0.000475,
		0.492972,
		0.474108,
		0.455022,
		0.435741,
		0.416289,
		0.396697,
		0.376995,
		0.357216,
		0.337397,
		0.317576,
		0.297795,
		0.278096,
		0.258527,
		0.239139,
		0.219983,
		0.201118,
		0.182603,
		0.164503,
		0.146884,
		0.12982,
		0.113386,
		0.097664,
		0.082738,
		0.0687,
		0.055645,
		0.043674,
		0.032895,
		0.023419,
		0.015366,
		0.008861,
		0.004038,
		0.001035,
		0.435513,
		0.416289,
		0.396928,
		0.377459,
		0.357915,
		0.33833,
		0.318742,
		0.299189,
		0.279714,
		0.260362,
		0.241182,
		0.222223,
		0.203541,
		0.185194,
		0.167242,
		0.14975,
		0.132788,
		0.116428,
		0.100747,
		0.085828,
		0.071756,
		0.058623,
		0.046525,
		0.035564,
		0.025848,
		0.017488,
		0.010606,
		0.005327,
		0.001783,
		0.000114,
		0.377913,
		0.358598,
		0.339242,
		0.31988,
		0.300551,
		0.281295,
		0.262156,
		0.243179,
		0.224414,
		0.205913,
		0.187731,
		0.169927,
		0.152563,
		0.135705,
		0.119423,
		0.103789,
		0.088884,
		0.074788,
		0.061589,
		0.049379,
		0.038254,
		0.028318,
		0.019678,
		0.012447,
		0.006746,
		0.0027,
		0.000443,
		0.320992,
		0.301882,
		0.282841,
		0.26391,
		0.245133,
		0.226559,
		0.208236,
		0.190218,
		0.172561,
		0.155325,
		0.138573,
		0.122371,
		0.106789,
		0.091904,
		0.077793,
		0.064539,
		0.05223,
		0.040959,
		0.030823,
		0.021925,
		0.014374,
		0.008282,
		0.003771,
		0.000966,
		0.265625,
		0.247045,
		0.228658,
		0.210511,
		0.192655,
		0.175145,
		0.158037,
		0.141391,
		0.125272,
		0.109748,
		0.094888,
		0.08077,
		0.067471,
		0.055077,
		0.043674,
		0.033357,
		0.024223,
		0.016376,
		0.009923,
		0.004979,
		0.001665,
		0.000107,
		0.212739,
		0.195044,
		0.177679,
		0.160699,
		0.144161,
		0.128128,
		0.112664,
		0.097836,
		0.083718,
		0.070384,
		0.057915,
		0.046396,
		0.035914,
		0.026564,
		0.018444,
		0.011657,
		0.006313,
		0.002525,
		0.000414,
		0.163312,
		0.146884,
		0.130939,
		0.115538,
		0.100747,
		0.086636,
		0.073275,
		0.060743,
		0.049119,
		0.038489,
		0.028941,
		0.02057,
		0.013475,
		0.007758,
		0.003529,
		0.000903,
		0.118371,
		0.103621,
		0.089522,
		0.076143,
		0.063557,
		0.051842,
		0.041077,
		0.031349,
		0.022747,
		0.015366,
		0.009304,
		0.004665,
		0.001559,
		0.0001,
		0.078987,
		0.066356,
		0.05456,
		0.043674,
		0.033782,
		0.024968,
		0.017323,
		0.01094,
		0.00592,
		0.002366,
		0.000387,
		0.046277,
		0.036235,
		0.027226,
		0.019337,
		0.012657,
		0.007282,
		0.00331,
		0.000846,
		0.021402,
		0.014447,
		0.008741,
		0.004379,
		0.001462,
		9.3e-005,
		0.005562,
		0.002221,
		0.000363,
		0.001428,
		0.001451,
		0.005649,
		0.01237,
		0.001474,
		0.005737,
		0.01256,
		0.021727,
		0.033031,
		0.001497,
		0.005827,
		0.012755,
		0.022059,
		0.033529,
		0.046964,
		0.062177,
		0.001521,
		0.00592,
		0.012955,
		0.022399,
		0.034038,
		0.047667,
		0.063092,
		0.080132,
		0.098612,
		0.001546,
		0.006014,
		0.013159,
		0.022747,
		0.034559,
		0.048385,
		0.064028,
		0.081301,
		0.100028,
		0.120042,
		0.141187,
		0.001571,
		0.006111,
		0.013368,
		0.023103,
		0.035092,
		0.049119,
		0.064984,
		0.082496,
		0.101474,
		0.121749,
		0.14316,
		0.165555,
		0.188793,
		0.001597,
		0.006211,
		0.013582,
		0.023468,
		0.035637,
		0.049871,
		0.065963,
		0.083718,
		0.102952,
		0.123492,
		0.145174,
		0.167843,
		0.191355,
		0.215573,
		0.24037,
		0.001624,
		0.006313,
		0.013802,
		0.023841,
		0.036195,
		0.050639,
		0.066963,
		0.084966,
		0.104462,
		0.125272,
		0.14723,
		0.170178,
		0.193969,
		0.218463,
		0.24353,
		0.269049,
		0.294906,
		0.001651,
		0.006417,
		0.014026,
		0.024223,
		0.036766,
		0.051425,
		0.067986,
		0.086243,
		0.106005,
		0.127091,
		0.14933,
		0.172561,
		0.196635,
		0.221409,
		0.246752,
		0.272538,
		0.298652,
		0.324985,
		0.351437,
		0.001679,
		0.006524,
		0.014256,
		0.024615,
		0.037351,
		0.05223,
		0.069032,
		0.087549,
		0.107582,
		0.128949,
		0.151474,
		0.174994,
		0.199356,
		0.224414,
		0.250035,
		0.276092,
		0.302467,
		0.32905,
		0.355737,
		0.382435,
		0.409054,
		0.001708,
		0.006633,
		0.014492,
		0.025016,
		0.03795,
		0.053054,
		0.070103,
		0.088884,
		0.109195,
		0.130847,
		0.153664,
		0.177478,
		0.202132,
		0.227479,
		0.253383,
		0.279714,
		0.306352,
		0.333186,
		0.360112,
		0.387032,
		0.413858,
		0.440506,
		0.466901,
		0.001737,
		0.006746,
		0.014734,
		0.025427,
		0.038563,
		0.053897,
		0.071199,
		0.090249,
		0.110843,
		0.132788,
		0.155901,
		0.180014,
		0.204965,
		0.230606,
		0.256796,
		0.283404,
		0.310309,
		0.337397,
		0.364563,
		0.391707,
		0.41874,
		0.445578,
		0.472143,
		0.498366,
		0.52418,
		0.001767,
		0.006861,
		0.014982,
		0.025848,
		0.039191,
		0.054761,
		0.07232,
		0.091646,
		0.112529,
		0.134771,
		0.158187,
		0.182603,
		0.207857,
		0.233795,
		0.260276,
		0.287165,
		0.31434,
		0.341684,
		0.369091,
		0.39646,
		0.423701,
		0.450728,
		0.477463,
		0.503836,
		0.529781,
		0.555238,
		0.580154,
		0.001798,
		0.006979,
		0.015236,
		0.026279,
		0.039835,
		0.055645,
		0.073468,
		0.093076,
		0.114253,
		0.136799,
		0.160523,
		0.185248,
		0.210809,
		0.237049,
		0.263824,
		0.290998,
		0.318445,
		0.346048,
		0.373698,
		0.401293,
		0.428742,
		0.455958,
		0.482863,
		0.509384,
		0.535457,
		0.561021,
		0.586022,
		0.610413,
		0.63415,
		0.00183,
		0.007101,
		0.015497,
		0.026722,
		0.040494,
		0.056551,
		0.074644,
		0.094539,
		0.116017,
		0.138872,
		0.16291,
		0.18795,
		0.213823,
		0.24037,
		0.267443,
		0.294906,
		0.322628,
		0.350491,
		0.378386,
		0.406208,
		0.433865,
		0.46127,
		0.488343,
		0.515011,
		0.541209,
		0.566876,
		0.59196,
		0.61641,
		0.640185,
		0.663248,
		0.685564,
		0.001863,
		0.007225,
		0.015765,
		0.027175,
		0.04117,
		0.057479,
		0.075848,
		0.096037,
		0.117821,
		0.140991,
		0.165349,
		0.19071,
		0.2169,
		0.243759,
		0.271135,
		0.298888,
		0.326889,
		0.355015,
		0.383156,
		0.411206,
		0.439072,
		0.466665,
		0.493904,
		0.520718,
		0.547038,
		0.572806,
		0.597967,
		0.622473,
		0.646281,
		0.669354,
		0.691659,
		0.71317,
		0.733863,
		0.001896,
		0.007353,
		0.016039,
		0.027641,
		0.041863,
		0.05843,
		0.077081,
		0.09757,
		0.119668,
		0.14316,
		0.167843,
		0.193529,
		0.220042,
		0.247217,
		0.2749,
		0.302949,
		0.331231,
		0.359622,
		0.38801,
		0.416289,
		0.444363,
		0.472143,
		0.499548,
		0.526505,
		0.552945,
		0.57881,
		0.604044,
		0.628601,
		0.652437,
		0.675515,
		0.697803,
		0.719275,
		0.739908,
		0.759685,
		0.778592,
		0.001931,
		0.007484,
		0.016321,
		0.028118,
		0.042574,
		0.059405,
		0.078344,
		0.09914,
		0.121558,
		0.145377,
		0.170393,
		0.196411,
		0.223252,
		0.250747,
		0.278741,
		0.307089,
		0.335654,
		0.364313,
		0.39295,
		0.421459,
		0.449741,
		0.477707,
		0.505276,
		0.532373,
		0.55893,
		0.584888,
		0.610192,
		0.634794,
		0.658652,
		0.681729,
		0.703994,
		0.72542,
		0.745987,
		0.765677,
		0.784477,
		0.802379,
		0.819379,
		0.001966,
		0.007619,
		0.01661,
		0.028608,
		0.043303,
		0.060404,
		0.079638,
		0.100747,
		0.123492,
		0.147646,
		0.173,
		0.199356,
		0.22653,
		0.254351,
		0.282661,
		0.31131,
		0.340162,
		0.369091,
		0.397978,
		0.426716,
		0.455206,
		0.483358,
		0.511089,
		0.538323,
		0.564995,
		0.591042,
		0.61641,
		0.641052,
		0.664927,
		0.687996,
		0.710231,
		0.731605,
		0.752097,
		0.771692,
		0.790377,
		0.808146,
		0.824995,
		0.840925,
		0.855939,
		0.002002,
		0.007758,
		0.016907,
		0.02911,
		0.044051,
		0.061429,
		0.080964,
		0.102394,
		0.125472,
		0.149968,
		0.175666,
		0.202366,
		0.229879,
		0.258031,
		0.28666,
		0.315615,
		0.344756,
		0.373956,
		0.403095,
		0.432063,
		0.460761,
		0.489096,
		0.516987,
		0.544357,
		0.571138,
		0.59727,
		0.622699,
		0.647376,
		0.67126,
		0.694316,
		0.716513,
		0.737827,
		0.758237,
		0.777728,
		0.796291,
		0.813918,
		0.830607,
		0.846361,
		0.861185,
		0.875088,
		0.888082,
		0.00204,
		0.0079,
		0.017212,
		0.029626,
		0.044818,
		0.06248,
		0.082324,
		0.104082,
		0.1275,
		0.152345,
		0.178394,
		0.205443,
		0.2333,
		0.261788,
		0.290741,
		0.320005,
		0.349439,
		0.378911,
		0.408303,
		0.437501,
		0.466406,
		0.494924,
		0.522972,
		0.550474,
		0.577362,
		0.603574,
		0.629057,
		0.653763,
		0.677651,
		0.700686,
		0.722839,
		0.744084,
		0.764404,
		0.783784,
		0.802214,
		0.819691,
		0.836212,
		0.851782,
		0.866407,
		0.880097,
		0.892867,
		0.904734,
		0.915718,
		0.002078,
		0.008046,
		0.017525,
		0.030156,
		0.045605,
		0.063557,
		0.083718,
		0.10581,
		0.129577,
		0.154777,
		0.181184,
		0.208589,
		0.236797,
		0.265625,
		0.294906,
		0.324482,
		0.354211,
		0.383959,
		0.413603,
		0.443032,
		0.472143,
		0.500843,
		0.529046,
		0.556677,
		0.583667,
		0.609954,
		0.635486,
		0.660215,
		0.6841,
		0.707107,
		0.729206,
		0.750376,
		0.770596,
		0.789856,
		0.808146,
		0.825463,
		0.841807,
		0.857183,
		0.8716,
		0.88507,
		0.897607,
		0.909232,
		0.919966,
		0.929834,
		0.938862,
		0.002117,
		0.008197,
		0.017847,
		0.0307,
		0.046414,
		0.064663,
		0.085147,
		0.107582,
		0.131705,
		0.157267,
		0.184039,
		0.211807,
		0.24037,
		0.269544,
		0.299157,
		0.32905,
		0.359076,
		0.3891,
		0.418999,
		0.448658,
		0.477974,
		0.506853,
		0.535208,
		0.562964,
		0.590052,
		0.61641,
		0.641985,
		0.66673,
		0.690605,
		0.713576,
		0.735615,
		0.756699,
		0.776812,
		0.795942,
		0.814083,
		0.83123,
		0.847388,
		0.862562,
		0.876762,
		0.890001,
		0.902298,
		0.913673,
		0.924149,
		0.933754,
		0.942516,
		0.950468,
		0.957642,
		0.002158,
		0.008351,
		0.018178,
		0.031259,
		0.047243,
		0.065798,
		0.086613,
		0.109399,
		0.133884,
		0.159817,
		0.186961,
		0.215097,
		0.244022,
		0.273547,
		0.303496,
		0.333709,
		0.364035,
		0.394338,
		0.424491,
		0.454381,
		0.483901,
		0.512956,
		0.541461,
		0.569338,
		0.596519,
		0.622942,
		0.648554,
		0.673308,
		0.697165,
		0.720092,
		0.742062,
		0.763052,
		0.783049,
		0.80204,
		0.820021,
		0.83699,
		0.852952,
		0.867913,
		0.881887,
		0.894888,
		0.906935,
		0.918052,
		0.928264,
		0.937598,
		0.946088,
		0.953766,
		0.960668,
		0.966832,
		0.972299,
		0.0022,
		0.00851,
		0.018518,
		0.031834,
		0.048096,
		0.066963,
		0.088117,
		0.111261,
		0.136118,
		0.162428,
		0.189951,
		0.218463,
		0.247755,
		0.277636,
		0.307926,
		0.338462,
		0.369091,
		0.399673,
		0.430082,
		0.460201,
		0.489923,
		0.519153,
		0.547804,
		0.575799,
		0.603067,
		0.629549,
		0.655192,
		0.679948,
		0.70378,
		0.726654,
		0.748546,
		0.769433,
		0.789303,
		0.808146,
		0.825958,
		0.842738,
		0.858494,
		0.873234,
		0.886971,
		0.899725,
		0.911514,
		0.922365,
		0.932304,
		0.941362,
		0.949573,
		0.956972,
		0.963598,
		0.969489,
		0.974689,
		0.97924,
		0.983187,
		0.002243,
		0.008674,
		0.018867,
		0.032424,
		0.048971,
		0.068159,
		0.08966,
		0.113171,
		0.138407,
		0.165103,
		0.193012,
		0.221906,
		0.251572,
		0.281814,
		0.31245,
		0.343311,
		0.374245,
		0.405109,
		0.435774,
		0.466121,
		0.496045,
		0.525446,
		0.55424,
		0.582346,
		0.609698,
		0.636233,
		0.661898,
		0.686649,
		0.710447,
		0.73326,
		0.755064,
		0.77584,
		0.795573,
		0.814257,
		0.831889,
		0.848471,
		0.864011,
		0.878519,
		0.892012,
		0.904508,
		0.916031,
		0.926607,
		0.936266,
		0.945041,
		0.952968,
		0.960083,
		0.966428,
		0.972044,
		0.976975,
		0.981267,
		0.984964,
		0.988116,
		0.990768,
		0.002287,
		0.008842,
		0.019227,
		0.033031,
		0.049871,
		0.069387,
		0.091244,
		0.11513,
		0.140754,
		0.167843,
		0.196146,
		0.225429,
		0.255475,
		0.286083,
		0.317068,
		0.34826,
		0.379501,
		0.410647,
		0.441568,
		0.472143,
		0.502265,
		0.531836,
		0.560768,
		0.588982,
		0.61641,
		0.642991,
		0.668673,
		0.69341,
		0.717165,
		0.739908,
		0.761615,
		0.782268,
		0.801855,
		0.82037,
		0.837812,
		0.854185,
		0.869498,
		0.883765,
		0.897002,
		0.909232,
		0.92048,
		0.930774,
		0.940146,
		0.948631,
		0.956268,
		0.963095,
		0.969156,
		0.974494,
		0.979155,
		0.983187,
		0.986637,
		0.989554,
		0.991988,
		0.993987,
		0.995601,
		0.002333,
		0.009016,
		0.019597,
		0.033655,
		0.050795,
		0.070648,
		0.09287,
		0.11714,
		0.14316,
		0.170651,
		0.199356,
		0.229035,
		0.259467,
		0.290446,
		0.321785,
		0.353309,
		0.384859,
		0.416289,
		0.447466,
		0.478268,
		0.508587,
		0.538323,
		0.567389,
		0.595706,
		0.623205,
		0.649825,
		0.675515,
		0.70023,
		0.723934,
		0.746597,
		0.768196,
		0.788716,
		0.808146,
		0.826481,
		0.843722,
		0.859876,
		0.874952,
		0.888967,
		0.901939,
		0.913893,
		0.924857,
		0.93486,
		0.943938,
		0.952128,
		0.959469,
		0.966004,
		0.971777,
		0.976835,
		0.981226,
		0.984999,
		0.988203,
		0.99089,
		0.993109,
		0.994912,
		0.996348,
		0.997465,
		0.99831,
		0.00238,
		0.009194,
		0.019978,
		0.034297,
		0.051745,
		0.071944,
		0.094539,
		0.119202,
		0.145627,
		0.173528,
		0.202643,
		0.232725,
		0.263549,
		0.294906,
		0.326602,
		0.358463,
		0.390324,
		0.422038,
		0.453471,
		0.484498,
		0.515011,
		0.54491,
		0.574105,
		0.602518,
		0.630081,
		0.656733,
		0.682423,
		0.707107,
		0.73075,
		0.753323,
		0.774805,
		0.795181,
		0.814442,
		0.832587,
		0.849616,
		0.865538,
		0.880367,
		0.894119,
		0.906817,
		0.918486,
		0.929157,
		0.938862,
		0.947639,
		0.955526,
		0.962567,
		0.968806,
		0.974289,
		0.979066,
		0.983187,
		0.986701,
		0.989663,
		0.992122,
		0.994133,
		0.995745,
		0.997011,
		0.997978,
		0.998695,
		0.999205,
		0.999551,
		0.002428,
		0.009378,
		0.02037,
		0.034957,
		0.052722,
		0.073275,
		0.096253,
		0.121319,
		0.148158,
		0.176478,
		0.20601,
		0.236502,
		0.267725,
		0.299464,
		0.331523,
		0.363722,
		0.395897,
		0.427896,
		0.459584,
		0.490835,
		0.521539,
		0.551596,
		0.580916,
		0.60942,
		0.63704,
		0.663715,
		0.689396,
		0.71404,
		0.737611,
		0.760084,
		0.781437,
		0.801659,
		0.820741,
		0.838682,
		0.855488,
		0.871169,
		0.885739,
		0.899218,
		0.911631,
		0.923006,
		0.933375,
		0.942774,
		0.951243,
		0.958822,
		0.965558,
		0.971498,
		0.976689,
		0.981184,
		0.985035,
		0.988294,
		0.991015,
		0.993253,
		0.99506,
		0.99649,
		0.997593,
		0.99842,
		0.999017,
		0.999429,
		0.999698,
		0.999859,
		0.999945,
		0.002478,
		0.009567,
		0.020774,
		0.035637,
		0.053727,
		0.074644,
		0.098015,
		0.123492,
		0.150754,
		0.179502,
		0.20946,
		0.24037,
		0.271997,
		0.304124,
		0.336549,
		0.369091,
		0.40158,
		0.433865,
		0.465807,
		0.497281,
		0.528173,
		0.558383,
		0.587822,
		0.61641,
		0.64408,
		0.670771,
		0.696434,
		0.721027,
		0.744517,
		0.766878,
		0.788092,
		0.808146,
		0.827036,
		0.844764,
		0.861335,
		0.876762,
		0.891062,
		0.904257,
		0.916375,
		0.927446,
		0.937506,
		0.946591,
		0.954745,
		0.962012,
		0.968439,
		0.974075,
		0.978974,
		0.983187,
		0.986769,
		0.989775,
		0.992261,
		0.994282,
		0.995893,
		0.997147,
		0.998097,
		0.998793,
		0.999281,
		0.999606,
		0.999807,
		0.99992,
		0.999974,
		0.999995,
		1,
		0.00253,
		0.009762,
		0.021189,
		0.036336,
		0.054761,
		0.076051,
		0.099824,
		0.125723,
		0.153419,
		0.182603,
		0.212995,
		0.24433,
		0.276369,
		0.308888,
		0.341684,
		0.37457,
		0.407377,
		0.439948,
		0.472143,
		0.503836,
		0.534913,
		0.565272,
		0.594824,
		0.62349,
		0.651201,
		0.677898,
		0.703534,
		0.728066,
		0.751464,
		0.773702,
		0.794763,
		0.814639,
		0.833326,
		0.850827,
		0.867151,
		0.882312,
		0.896331,
		0.909232,
		0.921045,
		0.931803,
		0.941544,
		0.950308,
		0.958141,
		0.96509,
		0.971204,
		0.976536,
		0.98114,
		0.985072,
		0.988388,
		0.991144,
		0.9934,
		0.995211,
		0.996633,
		0.997721,
		0.998528,
		0.999103,
		0.999493,
		0.999741,
		0.999886,
		0.999959,
		0.99999,
		0.999999,
		1,
		1,
		0.999998,
		0.002583,
		0.009964,
		0.021618,
		0.037057,
		0.055825,
		0.077498,
		0.101684,
		0.128015,
		0.156153,
		0.185784,
		0.216617,
		0.248386,
		0.280842,
		0.31376,
		0.34693,
		0.380165,
		0.413289,
		0.446146,
		0.478594,
		0.510503,
		0.541761,
		0.572264,
		0.601923,
		0.630658,
		0.658402,
		0.685097,
		0.710695,
		0.735156,
		0.758449,
		0.780552,
		0.80145,
		0.821134,
		0.839605,
		0.856866,
		0.872931,
		0.887815,
		0.901541,
		0.914137,
		0.925634,
		0.93607,
		0.945484,
		0.953921,
		0.961427,
		0.968053,
		0.973851,
		0.978877,
		0.983187,
		0.986838,
		0.989891,
		0.992403,
		0.994434,
		0.996042,
		0.997284,
		0.998216,
		0.998889,
		0.999355,
		0.999658,
		0.99984,
		0.999938,
		0.999982,
		0.999997,
		1,
		1,
		0.999999,
		0.99999,
		0.999962,
		0.999896,
		0.002638,
		0.010171,
		0.022059,
		0.037798,
		0.05692,
		0.078987,
		0.103595,
		0.130369,
		0.15896,
		0.189047,
		0.220331,
		0.25254,
		0.285421,
		0.318742,
		0.352291,
		0.385876,
		0.419319,
		0.452462,
		0.48516,
		0.517284,
		0.548718,
		0.579359,
		0.609118,
		0.637915,
		0.665684,
		0.692366,
		0.717915,
		0.742293,
		0.76547,
		0.787425,
		0.808146,
		0.827626,
		0.845868,
		0.862877,
		0.87867,
		0.893264,
		0.906686,
		0.918966,
		0.930138,
		0.940242,
		0.949321,
		0.957423,
		0.964597,
		0.970896,
		0.976376,
		0.981094,
		0.985111,
		0.988485,
		0.991278,
		0.993551,
		0.995364,
		0.996778,
		0.99785,
		0.998635,
		0.999187,
		0.999554,
		0.999782,
		0.999909,
		0.999971,
		0.999994,
		1,
		1,
		1,
		0.999995,
		0.999973,
		0.99992,
		0.999813,
		0.999629,
		0.999343,
		0.002694,
		0.010385,
		0.022514,
		0.038563,
		0.058047,
		0.080519,
		0.105561,
		0.132788,
		0.161842,
		0.192394,
		0.224139,
		0.256796,
		0.290107,
		0.323837,
		0.357769,
		0.391707,
		0.425471,
		0.458899,
		0.491846,
		0.52418,
		0.555785,
		0.586559,
		0.61641,
		0.645261,
		0.673044,
		0.699703,
		0.725192,
		0.749475,
		0.772523,
		0.794318,
		0.814849,
		0.834112,
		0.85211,
		0.868855,
		0.884362,
		0.898654,
		0.91176,
		0.923712,
		0.934549,
		0.944313,
		0.95305,
		0.96081,
		0.967646,
		0.973615,
		0.978775,
		0.983187,
		0.986911,
		0.990011,
		0.992549,
		0.99459,
		0.996194,
		0.997422,
		0.998334,
		0.998984,
		0.999426,
		0.999707,
		0.99987,
		0.999954,
		0.999989,
		0.999999,
		1,
		1,
		0.999997,
		0.999982,
		0.99994,
		0.99985,
		0.999688,
		0.999429,
		0.999046,
		0.998511,
		0.997796,
		0.002753,
		0.010606,
		0.022984,
		0.03935,
		0.059208,
		0.082095,
		0.107582,
		0.135274,
		0.164802,
		0.19583,
		0.228043,
		0.261156,
		0.294906,
		0.32905,
		0.363368,
		0.397661,
		0.431746,
		0.465459,
		0.498652,
		0.531193,
		0.562964,
		0.593863,
		0.623799,
		0.652694,
		0.680482,
		0.707107,
		0.732524,
		0.756699,
		0.779606,
		0.801227,
		0.821553,
		0.840584,
		0.858326,
		0.874792,
		0.890001,
		0.903979,
		0.916757,
		0.928371,
		0.938862,
		0.948277,
		0.956664,
		0.964076,
		0.970571,
		0.976208,
		0.981047,
		0.985151,
		0.988586,
		0.991416,
		0.993706,
		0.995521,
		0.996925,
		0.997978,
		0.998741,
		0.999268,
		0.999612,
		0.999819,
		0.99993,
		0.99998,
		0.999997,
		1,
		1,
		0.999999,
		0.999989,
		0.999957,
		0.999882,
		0.999741,
		0.999508,
		0.999156,
		0.998655,
		0.997978,
		0.997097,
		0.995985,
		0.994617,
		0.002813,
		0.010834,
		0.023468,
		0.040162,
		0.060404,
		0.083718,
		0.109662,
		0.137829,
		0.167843,
		0.199356,
		0.232048,
		0.265625,
		0.299819,
		0.334382,
		0.369091,
		0.403741,
		0.438147,
		0.472143,
		0.50558,
		0.538323,
		0.570256,
		0.601273,
		0.631286,
		0.660215,
		0.687996,
		0.714575,
		0.739908,
		0.763962,
		0.786713,
		0.808146,
		0.828254,
		0.84704,
		0.864511,
		0.880684,
		0.895582,
		0.909232,
		0.92167,
		0.932935,
		0.943071,
		0.952128,
		0.960158,
		0.967218,
		0.973368,
		0.978669,
		0.983187,
		0.986986,
		0.990135,
		0.9927,
		0.994749,
		0.996348,
		0.997561,
		0.998451,
		0.999076,
		0.999493,
		0.999752,
		0.999897,
		0.999967,
		0.999993,
		1,
		1,
		1,
		0.999994,
		0.99997,
		0.999909,
		0.999788,
		0.999581,
		0.999259,
		0.998793,
		0.998153,
		0.997312,
		0.99624,
		0.994912,
		0.993303,
		0.99139,
		0.989154,
		0.002875,
		0.011069,
		0.023968,
		0.041,
		0.061637,
		0.085389,
		0.111802,
		0.140457,
		0.170967,
		0.202976,
		0.236155,
		0.270205,
		0.30485,
		0.339837,
		0.37494,
		0.409949,
		0.444677,
		0.478955,
		0.512632,
		0.545574,
		0.577661,
		0.608789,
		0.638868,
		0.667822,
		0.695586,
		0.722107,
		0.747342,
		0.771261,
		0.793842,
		0.815072,
		0.834948,
		0.853472,
		0.870658,
		0.886525,
		0.901097,
		0.914407,
		0.926493,
		0.937398,
		0.94717,
		0.955861,
		0.963527,
		0.97023,
		0.976031,
		0.980996,
		0.985193,
		0.988691,
		0.991558,
		0.993865,
		0.995681,
		0.997073,
		0.998106,
		0.998845,
		0.999346,
		0.999666,
		0.999852,
		0.999947,
		0.999987,
		0.999999,
		1,
		1,
		0.999997,
		0.99998,
		0.999932,
		0.99983,
		0.999647,
		0.999355,
		0.998923,
		0.998321,
		0.99752,
		0.99649,
		0.995203,
		0.993633,
		0.991758,
		0.989554,
		0.987004,
		0.984089,
		0.980795,
		0.00294,
		0.011312,
		0.024483,
		0.041863,
		0.062907,
		0.08711,
		0.114005,
		0.14316,
		0.174178,
		0.206693,
		0.24037,
		0.2749,
		0.310002,
		0.34542,
		0.380919,
		0.416289,
		0.451339,
		0.485897,
		0.519811,
		0.552945,
		0.58518,
		0.61641,
		0.646547,
		0.675515,
		0.703249,
		0.729698,
		0.754822,
		0.778592,
		0.800988,
		0.822,
		0.841627,
		0.859876,
		0.876762,
		0.892307,
		0.90654,
		0.919497,
		0.931219,
		0.941754,
		0.951151,
		0.959469,
		0.966766,
		0.973107,
		0.978558,
		0.983187,
		0.987065,
		0.990264,
		0.992856,
		0.994912,
		0.996504,
		0.9977,
		0.998566,
		0.999166,
		0.999557,
		0.999793,
		0.99992,
		0.999977,
		0.999996,
		1,
		1,
		0.999999,
		0.999988,
		0.999951,
		0.999866,
		0.999707,
		0.999443,
		0.999046,
		0.998482,
		0.997721,
		0.996733,
		0.995488,
		0.993959,
		0.992122,
		0.989954,
		0.987433,
		0.984543,
		0.981267,
		0.977591,
		0.973507,
		0.969004,
		0.003007,
		0.011564,
		0.025016,
		0.042755,
		0.064217,
		0.088884,
		0.116272,
		0.14594,
		0.177478,
		0.210511,
		0.244695,
		0.279714,
		0.315281,
		0.351133,
		0.387032,
		0.422765,
		0.458135,
		0.492972,
		0.527118,
		0.560439,
		0.592813,
		0.624138,
		0.654322,
		0.683291,
		0.710982,
		0.737347,
		0.762345,
		0.785951,
		0.808146,
		0.828924,
		0.848287,
		0.866244,
		0.882814,
		0.898023,
		0.911904,
		0.924495,
		0.935842,
		0.945995,
		0.95501,
		0.962947,
		0.96987,
		0.975845,
		0.980944,
		0.985237,
		0.9888,
		0.991705,
		0.994029,
		0.995843,
		0.997222,
		0.998234,
		0.998946,
		0.999422,
		0.999716,
		0.999882,
		0.999962,
		0.999992,
		1,
		1,
		1,
		0.999993,
		0.999966,
		0.999897,
		0.99976,
		0.999525,
		0.999161,
		0.998635,
		0.997915,
		0.99697,
		0.995768,
		0.994282,
		0.992484,
		0.990352,
		0.987863,
		0.984999,
		0.981742,
		0.978079,
		0.973997,
		0.969489,
		0.964547,
		0.959167,
		0.953345,
		0.003075,
		0.011823,
		0.025566,
		0.043674,
		0.065569,
		0.090711,
		0.118608,
		0.148801,
		0.180871,
		0.214433,
		0.249134,
		0.28465,
		0.320688,
		0.35698,
		0.393283,
		0.429378,
		0.465069,
		0.50018,
		0.534555,
		0.568056,
		0.600563,
		0.631971,
		0.662191,
		0.69115,
		0.718785,
		0.74505,
		0.769907,
		0.793332,
		0.815311,
		0.835839,
		0.854921,
		0.872571,
		0.88881,
		0.903668,
		0.917181,
		0.929393,
		0.940353,
		0.950115,
		0.958739,
		0.966289,
		0.972832,
		0.97844,
		0.983187,
		0.987147,
		0.990397,
		0.993016,
		0.995079,
		0.996662,
		0.997839,
		0.998681,
		0.999253,
		0.999618,
		0.999831,
		0.99994,
		0.999985,
		0.999998,
		1,
		1,
		0.999996,
		0.999977,
		0.999923,
		0.999806,
		0.999599,
		0.999268,
		0.99878,
		0.998102,
		0.997199,
		0.996042,
		0.994599,
		0.992844,
		0.990749,
		0.988294,
		0.985457,
		0.98222,
		0.97857,
		0.974494,
		0.969981,
		0.965026,
		0.959621,
		0.953766,
		0.947458,
		0.940699,
		0.933492,
		0.003147,
		0.012092,
		0.026134,
		0.044624,
		0.066963,
		0.092596,
		0.121013,
		0.151745,
		0.18436,
		0.218463,
		0.25369,
		0.289713,
		0.326228,
		0.362964,
		0.399673,
		0.436133,
		0.472143,
		0.507526,
		0.542124,
		0.575799,
		0.608428,
		0.63991,
		0.670154,
		0.699089,
		0.726654,
		0.752804,
		0.777504,
		0.800733,
		0.822478,
		0.842738,
		0.861523,
		0.878848,
		0.89474,
		0.909232,
		0.922365,
		0.934185,
		0.944746,
		0.954108,
		0.962333,
		0.969489,
		0.97565,
		0.980889,
		0.985283,
		0.988913,
		0.991857,
		0.994196,
		0.996009,
		0.997372,
		0.998361,
		0.999046,
		0.999493,
		0.999763,
		0.999908,
		0.999974,
		0.999996,
		1,
		1,
		0.999998,
		0.999986,
		0.999944,
		0.999847,
		0.999666,
		0.999367,
		0.998917,
		0.99828,
		0.997422,
		0.99631,
		0.994912,
		0.993199,
		0.991144,
		0.988724,
		0.985916,
		0.982702,
		0.979066,
		0.974996,
		0.97048,
		0.965511,
		0.960083,
		0.954194,
		0.947842,
		0.941027,
		0.933754,
		0.926026,
		0.91785,
		0.909232,
		0.003221,
		0.01237,
		0.026722,
		0.045605,
		0.068402,
		0.094539,
		0.123492,
		0.154777,
		0.18795,
		0.222604,
		0.258369,
		0.294906,
		0.331906,
		0.369091,
		0.406208,
		0.443032,
		0.47936,
		0.515011,
		0.549827,
		0.583667,
		0.61641,
		0.647954,
		0.67821,
		0.707107,
		0.734587,
		0.760605,
		0.785132,
		0.808146,
		0.82964,
		0.849616,
		0.868085,
		0.88507,
		0.900598,
		0.914709,
		0.927446,
		0.938862,
		0.949014,
		0.957965,
		0.965784,
		0.972543,
		0.978317,
		0.983187,
		0.987232,
		0.990536,
		0.993181,
		0.995249,
		0.996822,
		0.997978,
		0.998793,
		0.999337,
		0.999675,
		0.999864,
		0.999956,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.999992,
		0.999961,
		0.999882,
		0.999726,
		0.999459,
		0.999046,
		0.998451,
		0.997637,
		0.996571,
		0.995219,
		0.993551,
		0.991538,
		0.989154,
		0.986377,
		0.983187,
		0.979567,
		0.975504,
		0.970985,
		0.966004,
		0.960553,
		0.95463,
		0.948233,
		0.941362,
		0.934022,
		0.926215,
		0.91795,
		0.909232,
		0.900072,
		0.89048,
		0.880467,
		0.003297,
		0.012657,
		0.027329,
		0.046619,
		0.069887,
		0.096544,
		0.126047,
		0.157899,
		0.191643,
		0.226862,
		0.263174,
		0.300233,
		0.337724,
		0.375363,
		0.412891,
		0.45008,
		0.486723,
		0.522638,
		0.557664,
		0.591661,
		0.624508,
		0.656102,
		0.686356,
		0.715201,
		0.742579,
		0.76845,
		0.792785,
		0.815567,
		0.836792,
		0.856465,
		0.874601,
		0.891227,
		0.906376,
		0.92009,
		0.932418,
		0.943417,
		0.953149,
		0.961681,
		0.969087,
		0.975444,
		0.980831,
		0.985332,
		0.989031,
		0.992015,
		0.994368,
		0.996177,
		0.997523,
		0.998486,
		0.999142,
		0.999561,
		0.999805,
		0.999931,
		0.999983,
		0.999998,
		1,
		1,
		0.999996,
		0.999974,
		0.999911,
		0.999779,
		0.999542,
		0.999166,
		0.998613,
		0.997845,
		0.996826,
		0.995521,
		0.993899,
		0.991928,
		0.989583,
		0.986838,
		0.983674,
		0.980072,
		0.976017,
		0.971498,
		0.966504,
		0.961031,
		0.955074,
		0.948631,
		0.941704,
		0.934295,
		0.926409,
		0.918052,
		0.909232,
		0.899959,
		0.890244,
		0.880097,
		0.869533,
		0.858565,
		0.847207,
		0.003376,
		0.012955,
		0.027957,
		0.047667,
		0.071421,
		0.098612,
		0.128681,
		0.161115,
		0.195443,
		0.231239,
		0.268109,
		0.3057,
		0.343689,
		0.381784,
		0.419726,
		0.457278,
		0.494234,
		0.530408,
		0.565638,
		0.599783,
		0.632722,
		0.664354,
		0.694592,
		0.723368,
		0.750628,
		0.776333,
		0.800458,
		0.82299,
		0.843926,
		0.863277,
		0.881063,
		0.897313,
		0.912065,
		0.925367,
		0.937272,
		0.947842,
		0.957143,
		0.965249,
		0.972236,
		0.978187,
		0.983187,
		0.987321,
		0.990679,
		0.99335,
		0.995423,
		0.996984,
		0.998117,
		0.998903,
		0.999417,
		0.999727,
		0.999894,
		0.99997,
		0.999995,
		1,
		1,
		0.999998,
		0.999984,
		0.999935,
		0.999825,
		0.999618,
		0.999278,
		0.998766,
		0.998044,
		0.997073,
		0.995817,
		0.994242,
		0.992316,
		0.990011,
		0.987301,
		0.984164,
		0.980581,
		0.976536,
		0.972017,
		0.967013,
		0.961517,
		0.955526,
		0.949038,
		0.942053,
		0.934575,
		0.926607,
		0.918157,
		0.909232,
		0.899843,
		0.890001,
		0.879718,
		0.869007,
		0.857883,
		0.846361,
		0.834457,
		0.822187,
		0.809569,
		0.003458,
		0.013263,
		0.028608,
		0.04875,
		0.073006,
		0.100747,
		0.131398,
		0.164428,
		0.199356,
		0.23574,
		0.273179,
		0.31131,
		0.349803,
		0.38836,
		0.426716,
		0.464632,
		0.501896,
		0.538323,
		0.573749,
		0.608032,
		0.641052,
		0.672708,
		0.702914,
		0.731605,
		0.758729,
		0.78425,
		0.808146,
		0.830407,
		0.851035,
		0.870045,
		0.887461,
		0.903317,
		0.917657,
		0.930531,
		0.941999,
		0.952128,
		0.960989,
		0.968661,
		0.975226,
		0.98077,
		0.985382,
		0.989154,
		0.992177,
		0.994545,
		0.996348,
		0.997675,
		0.99861,
		0.999236,
		0.999625,
		0.999844,
		0.99995,
		0.99999,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999955,
		0.999864,
		0.999686,
		0.999381,
		0.99891,
		0.998234,
		0.997312,
		0.996106,
		0.99458,
		0.9927,
		0.990437,
		0.987764,
		0.984657,
		0.981094,
		0.977061,
		0.972543,
		0.967529,
		0.962012,
		0.955987,
		0.949453,
		0.94241,
		0.93486,
		0.926809,
		0.918264,
		0.909232,
		0.899725,
		0.889753,
		0.879329,
		0.868468,
		0.857183,
		0.845492,
		0.833411,
		0.820956,
		0.808146,
		0.794999,
		0.781532,
		0.767766,
		0.003544,
		0.013582,
		0.029281,
		0.049871,
		0.074644,
		0.102952,
		0.1342,
		0.167843,
		0.203384,
		0.24037,
		0.278389,
		0.317068,
		0.356071,
		0.395094,
		0.433865,
		0.472143,
		0.509713,
		0.546387,
		0.582,
		0.61641,
		0.649498,
		0.681162,
		0.71132,
		0.739908,
		0.766878,
		0.792195,
		0.815842,
		0.837812,
		0.858112,
		0.876762,
		0.893788,
		0.909232,
		0.923141,
		0.935573,
		0.946591,
		0.956268,
		0.964681,
		0.971912,
		0.978051,
		0.983187,
		0.987414,
		0.990828,
		0.993526,
		0.995601,
		0.997147,
		0.998255,
		0.999011,
		0.999493,
		0.999775,
		0.99992,
		0.99998,
		0.999998,
		1,
		1,
		0.999995,
		0.99997,
		0.999898,
		0.999746,
		0.999475,
		0.999046,
		0.998415,
		0.997543,
		0.996388,
		0.994912,
		0.993081,
		0.990862,
		0.988228,
		0.985151,
		0.981612,
		0.977591,
		0.973075,
		0.968053,
		0.962515,
		0.956457,
		0.949876,
		0.942774,
		0.935153,
		0.927017,
		0.918374,
		0.909232,
		0.899603,
		0.889497,
		0.878929,
		0.867913,
		0.856465,
		0.8446,
		0.832336,
		0.819691,
		0.806683,
		0.793332,
		0.779657,
		0.765677,
		0.751412,
		0.736881,
		0.722107,
		0.003632,
		0.013913,
		0.029978,
		0.05103,
		0.076337,
		0.105229,
		0.137092,
		0.171364,
		0.207533,
		0.245133,
		0.283743,
		0.32298,
		0.362499,
		0.40199,
		0.441178,
		0.479816,
		0.517687,
		0.5546,
		0.59039,
		0.624916,
		0.658057,
		0.689715,
		0.719808,
		0.748275,
		0.77507,
		0.800163,
		0.823539,
		0.845197,
		0.865149,
		0.883417,
		0.900035,
		0.915048,
		0.928511,
		0.940484,
		0.951039,
		0.960253,
		0.968209,
		0.974996,
		0.980706,
		0.985435,
		0.989282,
		0.992346,
		0.994726,
		0.996521,
		0.997826,
		0.998733,
		0.999326,
		0.999684,
		0.999877,
		0.999965,
		0.999995,
		1,
		1,
		0.999998,
		0.999981,
		0.999926,
		0.999798,
		0.999561,
		0.999172,
		0.998588,
		0.997765,
		0.996662,
		0.995238,
		0.993458,
		0.991285,
		0.988691,
		0.985648,
		0.982133,
		0.978128,
		0.973615,
		0.968585,
		0.963026,
		0.956935,
		0.950308,
		0.943146,
		0.935451,
		0.927229,
		0.918486,
		0.909232,
		0.899478,
		0.889236,
		0.878519,
		0.867344,
		0.855726,
		0.843682,
		0.83123,
		0.81839,
		0.805179,
		0.791619,
		0.777728,
		0.763528,
		0.749038,
		0.73428,
		0.719275,
		0.704043,
		0.688606,
		0.672983,
		0.003724,
		0.014256,
		0.0307,
		0.05223,
		0.078089,
		0.107582,
		0.140077,
		0.174994,
		0.211807,
		0.250035,
		0.289247,
		0.32905,
		0.369091,
		0.409054,
		0.448658,
		0.487653,
		0.52582,
		0.562964,
		0.598922,
		0.63355,
		0.66673,
		0.698364,
		0.728373,
		0.756699,
		0.783299,
		0.808146,
		0.83123,
		0.852555,
		0.872136,
		0.890001,
		0.906191,
		0.920755,
		0.933754,
		0.945255,
		0.955335,
		0.964076,
		0.971569,
		0.977906,
		0.983187,
		0.987511,
		0.990983,
		0.993706,
		0.995782,
		0.997312,
		0.998392,
		0.999115,
		0.999565,
		0.999819,
		0.999942,
		0.999988,
		0.999999,
		1,
		0.999999,
		0.999989,
		0.999948,
		0.999844,
		0.999638,
		0.999289,
		0.99875,
		0.997978,
		0.996928,
		0.995558,
		0.993829,
		0.991705,
		0.989154,
		0.986146,
		0.982658,
		0.978669,
		0.974162,
		0.969125,
		0.963547,
		0.957423,
		0.950749,
		0.943527,
		0.935757,
		0.927446,
		0.918602,
		0.909232,
		0.89935,
		0.888967,
		0.878098,
		0.866759,
		0.854966,
		0.842738,
		0.830094,
		0.817052,
		0.803632,
		0.789856,
		0.775744,
		0.761317,
		0.746597,
		0.731605,
		0.716363,
		0.700893,
		0.685216,
		0.669354,
		0.653328,
		0.63716,
		0.620871,
		0.003819,
		0.014613,
		0.031449,
		0.053473,
		0.079901,
		0.110014,
		0.14316,
		0.178739,
		0.21621,
		0.255081,
		0.294906,
		0.335282,
		0.375851,
		0.416289,
		0.45631,
		0.495659,
		0.534114,
		0.571482,
		0.607595,
		0.642312,
		0.675515,
		0.707107,
		0.737013,
		0.765177,
		0.791559,
		0.816138,
		0.838908,
		0.859876,
		0.879064,
		0.896506,
		0.912247,
		0.926344,
		0.938862,
		0.949876,
		0.959469,
		0.967729,
		0.974752,
		0.980638,
		0.98549,
		0.989415,
		0.99252,
		0.994912,
		0.996697,
		0.997978,
		0.998853,
		0.999412,
		0.999739,
		0.999907,
		0.999977,
		0.999998,
		1,
		1,
		0.999994,
		0.999965,
		0.999882,
		0.999707,
		0.999396,
		0.998903,
		0.998182,
		0.997186,
		0.995871,
		0.994196,
		0.992122,
		0.989616,
		0.986646,
		0.983187,
		0.979216,
		0.974716,
		0.969673,
		0.964076,
		0.95792,
		0.951199,
		0.943915,
		0.93607,
		0.927669,
		0.91872,
		0.909232,
		0.899218,
		0.88869,
		0.877665,
		0.866157,
		0.854185,
		0.841768,
		0.828924,
		0.815675,
		0.80204,
		0.788042,
		0.773702,
		0.759042,
		0.744084,
		0.728851,
		0.713366,
		0.697651,
		0.681729,
		0.665622,
		0.649351,
		0.63294,
		0.61641,
		0.599783,
		0.583079,
		0.566321,
		0.003918,
		0.014982,
		0.032226,
		0.054761,
		0.081776,
		0.112529,
		0.146343,
		0.182603,
		0.22075,
		0.260276,
		0.300724,
		0.341684,
		0.382786,
		0.423701,
		0.464137,
		0.503836,
		0.542574,
		0.580154,
		0.61641,
		0.651201,
		0.684409,
		0.71594,
		0.745722,
		0.773702,
		0.799844,
		0.824131,
		0.846562,
		0.867151,
		0.885923,
		0.90292,
		0.918192,
		0.931803,
		0.943825,
		0.954338,
		0.963433,
		0.971204,
		0.977753,
		0.983187,
		0.987613,
		0.991144,
		0.993892,
		0.995967,
		0.997478,
		0.998528,
		0.999216,
		0.999632,
		0.999857,
		0.999959,
		0.999994,
		1,
		1,
		0.999998,
		0.999978,
		0.999914,
		0.999767,
		0.999493,
		0.999046,
		0.998376,
		0.997435,
		0.996177,
		0.994557,
		0.992536,
		0.990077,
		0.987147,
		0.983719,
		0.979768,
		0.975277,
		0.97023,
		0.964615,
		0.958426,
		0.951659,
		0.944313,
		0.93639,
		0.927897,
		0.918841,
		0.909232,
		0.899083,
		0.888407,
		0.877219,
		0.865538,
		0.853382,
		0.840769,
		0.827721,
		0.814257,
		0.800401,
		0.786174,
		0.771599,
		0.756699,
		0.741497,
		0.726017,
		0.710282,
		0.694316,
		0.678141,
		0.661782,
		0.645261,
		0.628601,
		0.611825,
		0.594955,
		0.578013,
		0.561021,
		0.544,
		0.526971,
		0.509955,
		0.00402,
		0.015366,
		0.033031,
		0.056095,
		0.083718,
		0.11513,
		0.149633,
		0.186592,
		0.225429,
		0.265625,
		0.306709,
		0.34826,
		0.3899,
		0.431293,
		0.472143,
		0.512188,
		0.5512,
		0.588982,
		0.625367,
		0.660215,
		0.69341,
		0.72486,
		0.754496,
		0.782268,
		0.808146,
		0.832117,
		0.854185,
		0.874369,
		0.892703,
		0.909232,
		0.924016,
		0.937122,
		0.948631,
		0.958631,
		0.967218,
		0.974494,
		0.980567,
		0.985548,
		0.989554,
		0.9927,
		0.995103,
		0.996876,
		0.99813,
		0.99897,
		0.999493,
		0.999788,
		0.999932,
		0.999986,
		0.999999,
		1,
		0.999999,
		0.999987,
		0.999939,
		0.999819,
		0.999581,
		0.999178,
		0.99856,
		0.997675,
		0.996474,
		0.994912,
		0.992945,
		0.990536,
		0.987648,
		0.984253,
		0.980326,
		0.975845,
		0.970795,
		0.965164,
		0.958943,
		0.952128,
		0.944719,
		0.936718,
		0.928131,
		0.918966,
		0.909232,
		0.898944,
		0.888115,
		0.876762,
		0.864902,
		0.852555,
		0.839741,
		0.826481,
		0.812798,
		0.798713,
		0.78425,
		0.769433,
		0.754286,
		0.738833,
		0.723099,
		0.707107,
		0.690882,
		0.674449,
		0.657831,
		0.641052,
		0.624138,
		0.607109,
		0.589991,
		0.572806,
		0.555576,
		0.538323,
		0.52107,
		0.503836,
		0.486644,
		0.469512,
		0.452462,
		0.004127,
		0.015765,
		0.033867,
		0.057479,
		0.085729,
		0.117821,
		0.153034,
		0.19071,
		0.230255,
		0.271135,
		0.312866,
		0.355015,
		0.397198,
		0.439072,
		0.480333,
		0.520718,
		0.559995,
		0.597967,
		0.634466,
		0.669354,
		0.702516,
		0.733863,
		0.763329,
		0.790869,
		0.816458,
		0.840087,
		0.861767,
		0.881522,
		0.899393,
		0.915432,
		0.929706,
		0.94229,
		0.953272,
		0.962746,
		0.970816,
		0.977591,
		0.983187,
		0.98772,
		0.991312,
		0.994084,
		0.996156,
		0.997644,
		0.998662,
		0.999313,
		0.999695,
		0.999891,
		0.999973,
		0.999997,
		1,
		1,
		0.999993,
		0.999959,
		0.999863,
		0.99966,
		0.9993,
		0.998733,
		0.997904,
		0.996763,
		0.99526,
		0.99335,
		0.990993,
		0.98815,
		0.984791,
		0.980889,
		0.97642,
		0.971369,
		0.965721,
		0.959469,
		0.952607,
		0.945135,
		0.937054,
		0.928371,
		0.919093,
		0.909232,
		0.898801,
		0.887815,
		0.87629,
		0.864247,
		0.851704,
		0.838682,
		0.825205,
		0.811294,
		0.796974,
		0.782268,
		0.767202,
		0.751801,
		0.736089,
		0.720092,
		0.703836,
		0.687346,
		0.670646,
		0.653763,
		0.636722,
		0.619546,
		0.60226,
		0.584888,
		0.567455,
		0.549982,
		0.532493,
		0.515011,
		0.497557,
		0.480153,
		0.46282,
		0.445578,
		0.428447,
		0.411446,
		0.394596,
		0.004238,
		0.016179,
		0.034735,
		0.058915,
		0.087813,
		0.120607,
		0.156549,
		0.194962,
		0.235233,
		0.276811,
		0.3192,
		0.361957,
		0.404687,
		0.447041,
		0.488711,
		0.529428,
		0.568961,
		0.607109,
		0.643707,
		0.678615,
		0.711723,
		0.742944,
		0.772216,
		0.799498,
		0.82477,
		0.848031,
		0.869296,
		0.888597,
		0.90598,
		0.921508,
		0.935251,
		0.947296,
		0.957735,
		0.966673,
		0.974219,
		0.980491,
		0.98561,
		0.9897,
		0.992887,
		0.995298,
		0.997056,
		0.99828,
		0.999083,
		0.99957,
		0.999833,
		0.999952,
		0.999993,
		1,
		1,
		0.999997,
		0.999974,
		0.999899,
		0.999729,
		0.999412,
		0.998895,
		0.998123,
		0.997042,
		0.995601,
		0.99375,
		0.991447,
		0.988652,
		0.985332,
		0.981456,
		0.977002,
		0.971951,
		0.966289,
		0.960006,
		0.953097,
		0.94556,
		0.937398,
		0.928617,
		0.919224,
		0.909232,
		0.898654,
		0.887506,
		0.875806,
		0.863572,
		0.850827,
		0.837591,
		0.823889,
		0.809744,
		0.795181,
		0.780225,
		0.764902,
		0.749239,
		0.73326,
		0.716994,
		0.700466,
		0.683702,
		0.66673,
		0.649575,
		0.632263,
		0.61482,
		0.59727,
		0.57964,
		0.561953,
		0.544233,
		0.526505,
		0.50879,
		0.491113,
		0.473495,
		0.455958,
		0.438523,
		0.42121,
		0.40404,
		0.387032,
		0.370205,
		0.353578,
		0.337168,
		0.004354,
		0.01661,
		0.035637,
		0.060404,
		0.089974,
		0.123492,
		0.160186,
		0.199356,
		0.24037,
		0.282661,
		0.325719,
		0.369091,
		0.412372,
		0.455206,
		0.497281,
		0.538323,
		0.5781,
		0.61641,
		0.653088,
		0.687996,
		0.721027,
		0.752097,
		0.781149,
		0.808146,
		0.833075,
		0.855939,
		0.876762,
		0.895582,
		0.912454,
		0.927446,
		0.94064,
		0.952128,
		0.962012,
		0.970403,
		0.97742,
		0.983187,
		0.987832,
		0.991486,
		0.994282,
		0.996348,
		0.997811,
		0.998793,
		0.999406,
		0.999752,
		0.99992,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999985,
		0.999929,
		0.999788,
		0.999512,
		0.999046,
		0.998331,
		0.997312,
		0.995933,
		0.994144,
		0.991898,
		0.989154,
		0.985874,
		0.982029,
		0.977591,
		0.972543,
		0.966867,
		0.960553,
		0.953597,
		0.945995,
		0.937751,
		0.928869,
		0.919359,
		0.909232,
		0.898503,
		0.887188,
		0.875306,
		0.862877,
		0.849923,
		0.836467,
		0.822533,
		0.808146,
		0.793332,
		0.778118,
		0.762531,
		0.746597,
		0.730344,
		0.7138,
		0.696992,
		0.679948,
		0.662695,
		0.645261,
		0.627672,
		0.609954,
		0.592135,
		0.57424,
		0.556295,
		0.538323,
		0.520351,
		0.502402,
		0.484498,
		0.466665,
		0.448922,
		0.431293,
		0.413799,
		0.39646,
		0.379296,
		0.362328,
		0.345573,
		0.32905,
		0.312776,
		0.29677,
		0.281048,
		0.004475,
		0.017058,
		0.036574,
		0.061951,
		0.092214,
		0.12648,
		0.163949,
		0.203896,
		0.245671,
		0.28869,
		0.332429,
		0.376422,
		0.420258,
		0.463572,
		0.506046,
		0.547405,
		0.587413,
		0.625869,
		0.662608,
		0.697495,
		0.730424,
		0.761317,
		0.79012,
		0.816804,
		0.84136,
		0.863799,
		0.884152,
		0.902465,
		0.9188,
		0.933236,
		0.94586,
		0.956775,
		0.96609,
		0.973927,
		0.980411,
		0.985674,
		0.989852,
		0.993081,
		0.995499,
		0.997238,
		0.998429,
		0.999193,
		0.999641,
		0.999872,
		0.999969,
		0.999997,
		1,
		1,
		0.999992,
		0.999952,
		0.999839,
		0.999603,
		0.999185,
		0.998528,
		0.997571,
		0.996257,
		0.994532,
		0.992346,
		0.989654,
		0.986419,
		0.982605,
		0.978187,
		0.973143,
		0.967454,
		0.961112,
		0.954108,
		0.94644,
		0.938112,
		0.929128,
		0.919497,
		0.909232,
		0.898348,
		0.886861,
		0.874792,
		0.862161,
		0.848992,
		0.835308,
		0.821134,
		0.806498,
		0.791425,
		0.775945,
		0.760084,
		0.743871,
		0.727336,
		0.710506,
		0.69341,
		0.676077,
		0.658536,
		0.640815,
		0.622942,
		0.604944,
		0.586849,
		0.568684,
		0.550474,
		0.532247,
		0.514027,
		0.495839,
		0.477707,
		0.459655,
		0.441707,
		0.423884,
		0.406208,
		0.388702,
		0.371384,
		0.354276,
		0.337397,
		0.320766,
		0.304401,
		0.288319,
		0.272538,
		0.257074,
		0.241943,
		0.227159,
		0.0046,
		0.017525,
		0.037549,
		0.063557,
		0.094539,
		0.129577,
		0.167843,
		0.208589,
		0.251144,
		0.294906,
		0.339336,
		0.383959,
		0.428351,
		0.472143,
		0.515011,
		0.556677,
		0.596902,
		0.635486,
		0.672265,
		0.707107,
		0.739908,
		0.770596,
		0.799123,
		0.825463,
		0.849616,
		0.8716,
		0.891454,
		0.909232,
		0.925006,
		0.938862,
		0.950898,
		0.961225,
		0.969962,
		0.977237,
		0.983187,
		0.987949,
		0.991668,
		0.994485,
		0.996543,
		0.997978,
		0.998921,
		0.999493,
		0.999803,
		0.999944,
		0.999991,
		1,
		1,
		0.999997,
		0.99997,
		0.999882,
		0.999682,
		0.999313,
		0.998713,
		0.997819,
		0.996571,
		0.994912,
		0.992789,
		0.990153,
		0.986964,
		0.983187,
		0.97879,
		0.973751,
		0.968053,
		0.961681,
		0.95463,
		0.946896,
		0.938482,
		0.929393,
		0.919639,
		0.909232,
		0.898188,
		0.886525,
		0.874262,
		0.861423,
		0.848031,
		0.834112,
		0.819691,
		0.804797,
		0.789457,
		0.773702,
		0.757559,
		0.741059,
		0.724231,
		0.707107,
		0.689715,
		0.672085,
		0.654248,
		0.636233,
		0.618068,
		0.599783,
		0.581406,
		0.562964,
		0.544486,
		0.525998,
		0.507526,
		0.489096,
		0.470734,
		0.452462,
		0.434306,
		0.416289,
		0.398433,
		0.38076,
		0.363292,
		0.346048,
		0.32905,
		0.312315,
		0.295864,
		0.279714,
		0.263882,
		0.248386,
		0.233241,
		0.218463,
		0.204067,
		0.190067,
		0.176478,
		0.004731,
		0.018011,
		0.038563,
		0.065227,
		0.096952,
		0.132788,
		0.171875,
		0.213443,
		0.256796,
		0.301315,
		0.346449,
		0.391707,
		0.436658,
		0.480925,
		0.52418,
		0.56614,
		0.606567,
		0.645261,
		0.682058,
		0.716828,
		0.749475,
		0.779928,
		0.808146,
		0.834112,
		0.85783,
		0.879329,
		0.898654,
		0.91587,
		0.931058,
		0.944313,
		0.955742,
		0.965467,
		0.973615,
		0.980326,
		0.985742,
		0.990011,
		0.993282,
		0.995704,
		0.997422,
		0.998577,
		0.999299,
		0.999707,
		0.999905,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999917,
		0.999752,
		0.999429,
		0.998886,
		0.998056,
		0.996876,
		0.995285,
		0.993226,
		0.99065,
		0.987511,
		0.983772,
		0.9794,
		0.974369,
		0.968661,
		0.962262,
		0.955164,
		0.947363,
		0.938862,
		0.929666,
		0.919785,
		0.909232,
		0.898023,
		0.886177,
		0.873716,
		0.860662,
		0.84704,
		0.832877,
		0.818201,
		0.803041,
		0.787425,
		0.771385,
		0.754952,
		0.738155,
		0.721027,
		0.703599,
		0.685901,
		0.667967,
		0.649825,
		0.631507,
		0.613044,
		0.594464,
		0.575799,
		0.557076,
		0.538323,
		0.51957,
		0.500843,
		0.482168,
		0.463572,
		0.44508,
		0.426716,
		0.408505,
		0.390469,
		0.372632,
		0.355015,
		0.33764,
		0.320527,
		0.303696,
		0.287165,
		0.270954,
		0.255081,
		0.239562,
		0.224414,
		0.209654,
		0.195295,
		0.181354,
		0.167843,
		0.154777,
		0.142168,
		0.130029,
		0.004868,
		0.018518,
		0.039618,
		0.066963,
		0.099458,
		0.136118,
		0.176052,
		0.218463,
		0.262634,
		0.307926,
		0.353774,
		0.399673,
		0.445185,
		0.489923,
		0.533556,
		0.575799,
		0.61641,
		0.655192,
		0.691982,
		0.726654,
		0.759116,
		0.789303,
		0.817181,
		0.842738,
		0.86599,
		0.886971,
		0.905738,
		0.922365,
		0.936941,
		0.949573,
		0.960379,
		0.969489,
		0.977043,
		0.983187,
		0.988073,
		0.991857,
		0.994696,
		0.996742,
		0.998145,
		0.999046,
		0.999575,
		0.999848,
		0.999963,
		0.999996,
		1,
		1,
		0.999991,
		0.999944,
		0.999811,
		0.999533,
		0.999046,
		0.99828,
		0.997169,
		0.995649,
		0.993658,
		0.991144,
		0.988059,
		0.984361,
		0.980016,
		0.974996,
		0.969281,
		0.962855,
		0.95571,
		0.947842,
		0.939252,
		0.929947,
		0.919935,
		0.909232,
		0.897854,
		0.88582,
		0.873152,
		0.859876,
		0.846017,
		0.831602,
		0.816662,
		0.801227,
		0.785326,
		0.768993,
		0.752258,
		0.735156,
		0.717717,
		0.699976,
		0.681965,
		0.663715,
		0.645261,
		0.626633,
		0.607863,
		0.588982,
		0.570021,
		0.55101,
		0.531979,
		0.512956,
		0.493969,
		0.475047,
		0.456215,
		0.437501,
		0.418929,
		0.400525,
		0.382312,
		0.364313,
		0.346552,
		0.32905,
		0.311827,
		0.294906,
		0.278304,
		0.262041,
		0.246136,
		0.230606,
		0.215467,
		0.200737,
		0.18643,
		0.172561,
		0.159146,
		0.146196,
		0.133727,
		0.121749,
		0.110275,
		0.099316,
		0.088884,
		0.00501,
		0.019046,
		0.040718,
		0.068768,
		0.102062,
		0.139573,
		0.18038,
		0.223658,
		0.268666,
		0.314747,
		0.361318,
		0.407865,
		0.453937,
		0.499142,
		0.543144,
		0.585653,
		0.62643,
		0.665277,
		0.702034,
		0.736579,
		0.768825,
		0.798713,
		0.826216,
		0.851331,
		0.874082,
		0.894513,
		0.912691,
		0.9287,
		0.942641,
		0.95463,
		0.964797,
		0.973282,
		0.980236,
		0.985814,
		0.990178,
		0.99349,
		0.995914,
		0.997607,
		0.998722,
		0.999399,
		0.999766,
		0.999933,
		0.99999,
		1,
		1,
		0.999996,
		0.999964,
		0.999861,
		0.999626,
		0.999193,
		0.998492,
		0.997451,
		0.996003,
		0.994084,
		0.991635,
		0.988607,
		0.984953,
		0.980638,
		0.975631,
		0.969911,
		0.963459,
		0.956268,
		0.948332,
		0.939652,
		0.930235,
		0.92009,
		0.909232,
		0.897679,
		0.885451,
		0.872571,
		0.859064,
		0.84496,
		0.830285,
		0.815072,
		0.799352,
		0.783156,
		0.76652,
		0.749475,
		0.732056,
		0.714298,
		0.696234,
		0.677898,
		0.659326,
		0.640549,
		0.621603,
		0.602518,
		0.583329,
		0.564067,
		0.544762,
		0.525446,
		0.506149,
		0.4869,
		0.467727,
		0.448658,
		0.429721,
		0.410941,
		0.392345,
		0.373956,
		0.355799,
		0.337898,
		0.320274,
		0.302949,
		0.285944,
		0.26928,
		0.252975,
		0.237049,
		0.22152,
		0.206404,
		0.191718,
		0.177478,
		0.163699,
		0.150396,
		0.137583,
		0.125272,
		0.113477,
		0.102209,
		0.09148,
		0.081301,
		0.071682,
		0.062632,
		0.054161,
		0.005159,
		0.019597,
		0.041863,
		0.070648,
		0.104768,
		0.14316,
		0.184867,
		0.229035,
		0.2749,
		0.321785,
		0.369091,
		0.416289,
		0.462921,
		0.508587,
		0.552945,
		0.595706,
		0.636628,
		0.675515,
		0.71221,
		0.746597,
		0.778592,
		0.808146,
		0.835238,
		0.859876,
		0.882091,
		0.901939,
		0.919497,
		0.93486,
		0.948142,
		0.959469,
		0.968983,
		0.976835,
		0.983187,
		0.988203,
		0.992055,
		0.994912,
		0.996943,
		0.99831,
		0.999166,
		0.999651,
		0.999887,
		0.999977,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999901,
		0.999707,
		0.999327,
		0.99869,
		0.997721,
		0.996348,
		0.994502,
		0.992122,
		0.989154,
		0.985548,
		0.981267,
		0.976276,
		0.970552,
		0.964076,
		0.956839,
		0.948834,
		0.940063,
		0.930531,
		0.920249,
		0.909232,
		0.897499,
		0.88507,
		0.87197,
		0.858226,
		0.843868,
		0.828924,
		0.813428,
		0.797413,
		0.780913,
		0.763962,
		0.746597,
		0.728851,
		0.710763,
		0.692366,
		0.673697,
		0.654791,
		0.635684,
		0.61641,
		0.597004,
		0.577498,
		0.557927,
		0.538323,
		0.518718,
		0.499142,
		0.479627,
		0.460201,
		0.440894,
		0.421733,
		0.402746,
		0.383959,
		0.365398,
		0.347086,
		0.32905,
		0.31131,
		0.29389,
		0.276811,
		0.260094,
		0.243759,
		0.227824,
		0.212308,
		0.197229,
		0.182603,
		0.168448,
		0.154777,
		0.141606,
		0.128949,
		0.116819,
		0.105229,
		0.094192,
		0.083718,
		0.073818,
		0.064503,
		0.055783,
		0.047667,
		0.040162,
		0.033279,
		0.027023,
		0.005315,
		0.020173,
		0.043058,
		0.072605,
		0.107582,
		0.146884,
		0.18952,
		0.234603,
		0.281345,
		0.32905,
		0.377099,
		0.424954,
		0.472143,
		0.518262,
		0.562964,
		0.605958,
		0.647002,
		0.685901,
		0.722505,
		0.756699,
		0.788409,
		0.817592,
		0.844236,
		0.868358,
		0.890001,
		0.909232,
		0.926139,
		0.94083,
		0.953429,
		0.964076,
		0.972926,
		0.98014,
		0.98589,
		0.990352,
		0.993706,
		0.996128,
		0.997793,
		0.998863,
		0.999493,
		0.999819,
		0.999956,
		0.999995,
		1,
		1,
		0.999989,
		0.999933,
		0.999776,
		0.999448,
		0.998875,
		0.997978,
		0.996681,
		0.994912,
		0.992605,
		0.9897,
		0.986146,
		0.981901,
		0.976929,
		0.971204,
		0.964706,
		0.957423,
		0.949349,
		0.940484,
		0.930835,
		0.920413,
		0.909232,
		0.897313,
		0.884676,
		0.87135,
		0.85736,
		0.842738,
		0.827516,
		0.811728,
		0.795408,
		0.778592,
		0.761317,
		0.743619,
		0.725537,
		0.707107,
		0.688367,
		0.669354,
		0.650105,
		0.630658,
		0.611048,
		0.591311,
		0.571482,
		0.551596,
		0.531686,
		0.511786,
		0.491928,
		0.472143,
		0.452462,
		0.432915,
		0.413531,
		0.394338,
		0.375363,
		0.356632,
		0.338171,
		0.320005,
		0.302157,
		0.28465,
		0.267506,
		0.250747,
		0.234393,
		0.218463,
		0.202976,
		0.18795,
		0.173402,
		0.159349,
		0.145806,
		0.132788,
		0.120309,
		0.108384,
		0.097025,
		0.086243,
		0.076051,
		0.06646,
		0.057479,
		0.049119,
		0.041389,
		0.034297,
		0.027851,
		0.022059,
		0.016928,
		0.012464,
		0.008674,
		0.005478,
		0.020774,
		0.044304,
		0.074644,
		0.110511,
		0.150754,
		0.194346,
		0.24037,
		0.288011,
		0.336549,
		0.385352,
		0.433865,
		0.48161,
		0.528173,
		0.573204,
		0.61641,
		0.657551,
		0.696434,
		0.732912,
		0.766878,
		0.798264,
		0.827036,
		0.853193,
		0.876762,
		0.897796,
		0.916375,
		0.9326,
		0.946591,
		0.958486,
		0.968439,
		0.976614,
		0.983187,
		0.98834,
		0.992261,
		0.995135,
		0.997147,
		0.998474,
		0.999281,
		0.99972,
		0.99992,
		0.999988,
		1,
		1,
		0.999995,
		0.999957,
		0.999834,
		0.999556,
		0.999046,
		0.998222,
		0.997003,
		0.995313,
		0.993081,
		0.990244,
		0.986746,
		0.982541,
		0.977591,
		0.971868,
		0.965348,
		0.95802,
		0.949876,
		0.940917,
		0.931149,
		0.920582,
		0.909232,
		0.897121,
		0.88427,
		0.870708,
		0.856465,
		0.84157,
		0.82606,
		0.809968,
		0.793332,
		0.77619,
		0.758579,
		0.740538,
		0.722107,
		0.703324,
		0.68423,
		0.664862,
		0.645261,
		0.625463,
		0.605508,
		0.585432,
		0.565272,
		0.545064,
		0.524843,
		0.504643,
		0.484498,
		0.464441,
		0.444504,
		0.424716,
		0.405109,
		0.385711,
		0.366551,
		0.347655,
		0.32905,
		0.310761,
		0.292812,
		0.275228,
		0.258031,
		0.241242,
		0.224882,
		0.208972,
		0.193529,
		0.178574,
		0.164123,
		0.150192,
		0.136799,
		0.123957,
		0.111681,
		0.099986,
		0.088884,
		0.078386,
		0.068506,
		0.059254,
		0.050639,
		0.042672,
		0.035363,
		0.028718,
		0.022747,
		0.017457,
		0.012854,
		0.008946,
		0.005737,
		0.003233,
		0.00144,
		0.000361,
		0.005649,
		0.021402,
		0.045605,
		0.07677,
		0.113559,
		0.154777,
		0.199356,
		0.246346,
		0.294906,
		0.344293,
		0.393858,
		0.443032,
		0.491327,
		0.538323,
		0.583667,
		0.627063,
		0.668272,
		0.707107,
		0.743424,
		0.777124,
		0.808146,
		0.836467,
		0.862095,
		0.88507,
		0.905457,
		0.92335,
		0.938862,
		0.952128,
		0.963299,
		0.972543,
		0.980037,
		0.98597,
		0.990536,
		0.99393,
		0.996348,
		0.997978,
		0.999001,
		0.999581,
		0.999864,
		0.999973,
		0.999998,
		1,
		0.999998,
		0.999975,
		0.999882,
		0.999651,
		0.999202,
		0.998451,
		0.997312,
		0.995704,
		0.993551,
		0.990785,
		0.987347,
		0.983187,
		0.978262,
		0.972543,
		0.966004,
		0.958631,
		0.950418,
		0.941362,
		0.931471,
		0.920755,
		0.909232,
		0.896922,
		0.883851,
		0.870045,
		0.855538,
		0.840361,
		0.824552,
		0.808146,
		0.791183,
		0.773702,
		0.755743,
		0.737347,
		0.718555,
		0.699408,
		0.679948,
		0.660215,
		0.64025,
		0.620093,
		0.599783,
		0.579359,
		0.55886,
		0.538323,
		0.517785,
		0.497281,
		0.476846,
		0.456513,
		0.436317,
		0.416289,
		0.39646,
		0.37686,
		0.357518,
		0.338462,
		0.319719,
		0.301315,
		0.283276,
		0.265625,
		0.248386,
		0.23158,
		0.21523,
		0.199356,
		0.183976,
		0.169111,
		0.154777,
		0.140991,
		0.127771,
		0.11513,
		0.103084,
		0.091646,
		0.08083,
		0.070648,
		0.061111,
		0.05223,
		0.044016,
		0.036479,
		0.029626,
		0.023468,
		0.018011,
		0.013263,
		0.00923,
		0.00592,
		0.003336,
		0.001486,
		0.000372,
		0.005827,
		0.022059,
		0.046964,
		0.078987,
		0.116734,
		0.15896,
		0.204557,
		0.25254,
		0.30204,
		0.352291,
		0.402626,
		0.452462,
		0.501302,
		0.548718,
		0.594354,
		0.637915,
		0.679164,
		0.717915,
		0.754033,
		0.787425,
		0.818041,
		0.845868,
		0.870925,
		0.893264,
		0.912966,
		0.930138,
		0.944907,
		0.957423,
		0.967852,
		0.976376,
		0.983187,
		0.988485,
		0.992476,
		0.995364,
		0.997353,
		0.998635,
		0.999391,
		0.999782,
		0.999946,
		0.999994,
		1,
		1,
		0.999987,
		0.99992,
		0.999732,
		0.999343,
		0.998665,
		0.997607,
		0.996084,
		0.994013,
		0.991323,
		0.987949,
		0.983837,
		0.978942,
		0.973229,
		0.966673,
		0.959257,
		0.950973,
		0.94182,
		0.931803,
		0.920935,
		0.909232,
		0.896717,
		0.883417,
		0.869359,
		0.854579,
		0.839109,
		0.82299,
		0.806258,
		0.788955,
		0.771123,
		0.752804,
		0.73404,
		0.714875,
		0.695352,
		0.675515,
		0.655405,
		0.635065,
		0.614537,
		0.593863,
		0.573084,
		0.552238,
		0.531365,
		0.510503,
		0.48969,
		0.468961,
		0.448352,
		0.427896,
		0.407628,
		0.387578,
		0.367779,
		0.34826,
		0.32905,
		0.310176,
		0.291667,
		0.273547,
		0.255841,
		0.238574,
		0.221767,
		0.205443,
		0.189623,
		0.174326,
		0.159571,
		0.145377,
		0.131761,
		0.118739,
		0.106327,
		0.094539,
		0.083389,
		0.072891,
		0.063057,
		0.053897,
		0.045425,
		0.037648,
		0.030578,
		0.024223,
		0.018592,
		0.013691,
		0.009529,
		0.006111,
		0.003444,
		0.001534,
		0.000384,
		0.006014,
		0.022747,
		0.048385,
		0.081301,
		0.120042,
		0.163312,
		0.209959,
		0.258963,
		0.309424,
		0.360554,
		0.411665,
		0.462163,
		0.511539,
		0.559361,
		0.605268,
		0.648967,
		0.690221,
		0.728851,
		0.764729,
		0.797771,
		0.827935,
		0.855222,
		0.879663,
		0.901325,
		0.920303,
		0.936718,
		0.950715,
		0.962458,
		0.97213,
		0.979927,
		0.986055,
		0.990728,
		0.994162,
		0.996571,
		0.998163,
		0.999134,
		0.999662,
		0.999903,
		0.999985,
		1,
		1,
		0.999994,
		0.999949,
		0.999801,
		0.99947,
		0.998863,
		0.997888,
		0.996452,
		0.994468,
		0.991857,
		0.988552,
		0.984492,
		0.97963,
		0.973927,
		0.967356,
		0.959898,
		0.951543,
		0.94229,
		0.932145,
		0.92112,
		0.909232,
		0.896506,
		0.882967,
		0.868649,
		0.853585,
		0.837812,
		0.82137,
		0.804301,
		0.786646,
		0.76845,
		0.749757,
		0.730613,
		0.711062,
		0.69115,
		0.670922,
		0.650423,
		0.629697,
		0.608789,
		0.587741,
		0.566596,
		0.545395,
		0.52418,
		0.502989,
		0.481862,
		0.460836,
		0.439948,
		0.419232,
		0.398725,
		0.378457,
		0.358463,
		0.338772,
		0.319414,
		0.300419,
		0.281814,
		0.263625,
		0.245878,
		0.228598,
		0.211807,
		0.195527,
		0.179781,
		0.164588,
		0.149968,
		0.135939,
		0.122519,
		0.109724,
		0.09757,
		0.086071,
		0.075242,
		0.065096,
		0.055645,
		0.046901,
		0.038875,
		0.031577,
		0.025016,
		0.019201,
		0.014141,
		0.009842,
		0.006313,
		0.003558,
		0.001584,
		0.000397,
		0.006211,
		0.023468,
		0.04987,
		0.083718,
		0.123492,
		0.167843,
		0.215573,
		0.265625,
		0.317068,
		0.369091,
		0.420985,
		0.472143,
		0.522046,
		0.570256,
		0.61641,
		0.660215,
		0.701439,
		0.739908,
		0.775502,
		0.808146,
		0.837812,
		0.864511,
		0.888291,
		0.909232,
		0.927446,
		0.943071,
		0.956268,
		0.967218,
		0.97612,
		0.983187,
		0.988638,
		0.9927,
		0.995601,
		0.997561,
		0.998793,
		0.999493,
		0.999836,
		0.999967,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999858,
		0.999581,
		0.999046,
		0.998153,
		0.996806,
		0.994912,
		0.992386,
		0.989154,
		0.985151,
		0.980326,
		0.974637,
		0.968053,
		0.960553,
		0.952128,
		0.942774,
		0.932498,
		0.921311,
		0.909232,
		0.896287,
		0.882503,
		0.867913,
		0.852555,
		0.836467,
		0.819691,
		0.80227,
		0.78425,
		0.765677,
		0.746597,
		0.727057,
		0.707107,
		0.686792,
		0.666161,
		0.645261,
		0.624138,
		0.602837,
		0.581406,
		0.559887,
		0.538323,
		0.516759,
		0.495233,
		0.473788,
		0.452462,
		0.431293,
		0.410318,
		0.389573,
		0.369091,
		0.348906,
		0.32905,
		0.309553,
		0.290446,
		0.271757,
		0.253513,
		0.23574,
		0.218463,
		0.201706,
		0.185491,
		0.169842,
		0.154777,
		0.140317,
		0.12648,
		0.113285,
		0.100747,
		0.088884,
		0.077708,
		0.067236,
		0.057479,
		0.048451,
		0.040162,
		0.032625,
		0.025848,
		0.019841,
		0.014613,
		0.010171,
		0.006524,
		0.003677,
		0.001638,
		0.00041,
		0.006416,
		0.024222,
		0.051425,
		0.086243,
		0.127091,
		0.172561,
		0.221409,
		0.272538,
		0.324985,
		0.377913,
		0.430596,
		0.48241,
		0.532828,
		0.581406,
		0.62778,
		0.671658,
		0.712812,
		0.751077,
		0.786339,
		0.818536,
		0.847654,
		0.873716,
		0.896786,
		0.916963,
		0.934374,
		0.949176,
		0.961546,
		0.971686,
		0.97981,
		0.986146,
		0.990931,
		0.994403,
		0.996799,
		0.998347,
		0.999261,
		0.999735,
		0.999935,
		0.999993,
		1,
		1,
		0.999984,
		0.999903,
		0.999678,
		0.999212,
		0.998403,
		0.997147,
		0.995346,
		0.992908,
		0.989754,
		0.985814,
		0.98103,
		0.975358,
		0.968764,
		0.961225,
		0.952728,
		0.943272,
		0.932861,
		0.921508,
		0.909232,
		0.89606,
		0.882021,
		0.867151,
		0.851487,
		0.835071,
		0.817948,
		0.800163,
		0.781763,
		0.762798,
		0.743316,
		0.723368,
		0.703003,
		0.682272,
		0.661224,
		0.63991,
		0.618377,
		0.596674,
		0.574848,
		0.552945,
		0.531011,
		0.50909,
		0.487226,
		0.465459,
		0.44383,
		0.42238,
		0.401146,
		0.380165,
		0.359472,
		0.339103,
		0.319089,
		0.299464,
		0.280256,
		0.261496,
		0.243212,
		0.225429,
		0.208175,
		0.191473,
		0.175346,
		0.159817,
		0.144906,
		0.130634,
		0.11702,
		0.104082,
		0.091835,
		0.080297,
		0.069482,
		0.059405,
		0.050078,
		0.041515,
		0.033726,
		0.026722,
		0.020513,
		0.015108,
		0.010517,
		0.006746,
		0.003803,
		0.001693,
		0.000424,
		0.00663,
		0.02501,
		0.053048,
		0.088879,
		0.130845,
		0.177477,
		0.227479,
		0.279714,
		0.333186,
		0.387032,
		0.440506,
		0.492972,
		0.54389,
		0.592813,
		0.639377,
		0.683291,
		0.724333,
		0.762345,
		0.797225,
		0.828924,
		0.85744,
		0.882814,
		0.905127,
		0.924495,
		0.941064,
		0.95501,
		0.966531,
		0.975845,
		0.983187,
		0.9888,
		0.992935,
		0.995843,
		0.997769,
		0.998946,
		0.999588,
		0.999882,
		0.999982,
		1,
		1,
		0.999993,
		0.999938,
		0.99976,
		0.999361,
		0.998635,
		0.997473,
		0.995768,
		0.993424,
		0.990352,
		0.986479,
		0.981742,
		0.976091,
		0.969489,
		0.961912,
		0.953345,
		0.943785,
		0.933236,
		0.921711,
		0.909232,
		0.895825,
		0.881522,
		0.866359,
		0.850378,
		0.833622,
		0.816138,
		0.797974,
		0.77918,
		0.759807,
		0.739908,
		0.719536,
		0.698742,
		0.67758,
		0.656102,
		0.63436,
		0.612405,
		0.590287,
		0.568056,
		0.545761,
		0.523449,
		0.501165,
		0.478955,
		0.456863,
		0.43493,
		0.413199,
		0.391707,
		0.370494,
		0.349596,
		0.32905,
		0.308888,
		0.289144,
		0.269849,
		0.251033,
		0.232725,
		0.214952,
		0.197741,
		0.181117,
		0.165103,
		0.149721,
		0.134994,
		0.120941,
		0.107582,
		0.094935,
		0.083016,
		0.071843,
		0.061429,
		0.051789,
		0.042936,
		0.034883,
		0.027641,
		0.02122,
		0.01563,
		0.01088,
		0.006979,
		0.003934,
		0.001752,
		0.000439,
		0.006851,
		0.025825,
		0.054735,
		0.091624,
		0.134756,
		0.182595,
		0.233791,
		0.287164,
		0.341684,
		0.39646,
		0.450728,
		0.503836,
		0.555238,
		0.604481,
		0.651201,
		0.695109,
		0.735993,
		0.773702,
		0.808146,
		0.839291,
		0.867151,
		0.891784,
		0.913289,
		0.931803,
		0.947493,
		0.960553,
		0.971204,
		0.979683,
		0.986243,
		0.991144,
		0.994653,
		0.99703,
		0.998528,
		0.999381,
		0.999799,
		0.999959,
		0.999997,
		1,
		0.999998,
		0.999963,
		0.999827,
		0.999493,
		0.99885,
		0.997782,
		0.996177,
		0.99393,
		0.990947,
		0.987147,
		0.982461,
		0.976835,
		0.97023,
		0.962616,
		0.953979,
		0.944313,
		0.933622,
		0.921922,
		0.909232,
		0.895582,
		0.881004,
		0.865538,
		0.849227,
		0.832117,
		0.814257,
		0.795699,
		0.776495,
		0.756699,
		0.736367,
		0.715554,
		0.694316,
		0.672708,
		0.650784,
		0.628601,
		0.606211,
		0.583667,
		0.561021,
		0.538323,
		0.515624,
		0.492972,
		0.470412,
		0.447991,
		0.425753,
		0.403741,
		0.381994,
		0.360554,
		0.339457,
		0.318742,
		0.298442,
		0.278592,
		0.259225,
		0.24037,
		0.222058,
		0.204316,
		0.187172,
		0.170651,
		0.154777,
		0.139573,
		0.125061,
		0.111261,
		0.098193,
		0.085875,
		0.074325,
		0.063557,
		0.053588,
		0.044432,
		0.036101,
		0.028608,
		0.021964,
		0.016179,
		0.011263,
		0.007225,
		0.004073,
		0.001814,
		0.000454,
		0.007072,
		0.026654,
		0.056465,
		0.094456,
		0.138805,
		0.187904,
		0.240343,
		0.294891,
		0.350485,
		0.406206,
		0.461269,
		0.515011,
		0.566876,
		0.61641,
		0.663248,
		0.707107,
		0.747781,
		0.785132,
		0.819083,
		0.849616,
		0.876762,
		0.900598,
		0.921246,
		0.938862,
		0.953636,
		0.965784,
		0.975548,
		0.983187,
		0.988971,
		0.993181,
		0.996092,
		0.997978,
		0.999094,
		0.999675,
		0.99992,
		0.999991,
		1,
		0.999999,
		0.999981,
		0.999882,
		0.999608,
		0.999046,
		0.998073,
		0.996571,
		0.994427,
		0.991538,
		0.987816,
		0.983187,
		0.977591,
		0.970985,
		0.963338,
		0.95463,
		0.944857,
		0.934022,
		0.922139,
		0.909232,
		0.895329,
		0.880467,
		0.864686,
		0.848031,
		0.830552,
		0.812301,
		0.793332,
		0.773702,
		0.753466,
		0.732684,
		0.711414,
		0.689715,
		0.667644,
		0.645261,
		0.622621,
		0.599783,
		0.5768,
		0.553728,
		0.53062,
		0.507526,
		0.484498,
		0.461585,
		0.438833,
		0.416289,
		0.393997,
		0.371999,
		0.350337,
		0.32905,
		0.308175,
		0.28775,
		0.267809,
		0.248386,
		0.229511,
		0.211216,
		0.193529,
		0.176478,
		0.160089,
		0.144385,
		0.129392,
		0.11513,
		0.101621,
		0.088884,
		0.076937,
		0.065798,
		0.055483,
		0.046007,
		0.037384,
		0.029626,
		0.022747,
		0.016757,
		0.011667,
		0.007484,
		0.004219,
		0.001879,
		0.000471,
		0.007286,
		0.027471,
		0.058196,
		0.097323,
		0.142941,
		0.19336,
		0.247099,
		0.302876,
		0.359581,
		0.416269,
		0.472135,
		0.526502,
		0.578809,
		0.628601,
		0.675515,
		0.719275,
		0.759685,
		0.796619,
		0.830017,
		0.859876,
		0.886248,
		0.909232,
		0.928972,
		0.945646,
		0.959469,
		0.970681,
		0.979547,
		0.986346,
		0.991369,
		0.994912,
		0.997265,
		0.998706,
		0.999493,
		0.999854,
		0.999977,
		0.999999,
		1,
		0.999991,
		0.999924,
		0.999707,
		0.999223,
		0.998347,
		0.99695,
		0.994912,
		0.992122,
		0.988485,
		0.983919,
		0.978359,
		0.971756,
		0.964076,
		0.955299,
		0.945417,
		0.934434,
		0.922365,
		0.909232,
		0.895068,
		0.879909,
		0.863799,
		0.846787,
		0.828924,
		0.810266,
		0.790869,
		0.770794,
		0.750101,
		0.728851,
		0.707107,
		0.684929,
		0.66238,
		0.63952,
		0.61641,
		0.593109,
		0.569676,
		0.546167,
		0.522638,
		0.499142,
		0.475734,
		0.452462,
		0.429378,
		0.406528,
		0.383959,
		0.361715,
		0.339837,
		0.318369,
		0.297347,
		0.276811,
		0.256796,
		0.237335,
		0.218463,
		0.200209,
		0.182603,
		0.165674,
		0.149448,
		0.133949,
		0.119202,
		0.105229,
		0.092052,
		0.079688,
		0.068159,
		0.057479,
		0.047667,
		0.038736,
		0.0307,
		0.023574,
		0.017367,
		0.012092,
		0.007758,
		0.004374,
		0.001948,
		0.000488,
		0.00748,
		0.028234,
		0.059855,
		0.100129,
		0.147054,
		0.198852,
		0.253963,
		0.311037,
		0.368915,
		0.426612,
		0.483303,
		0.538297,
		0.591031,
		0.641049,
		0.687995,
		0.731605,
		0.771692,
		0.808146,
		0.840925,
		0.870045,
		0.895582,
		0.917657,
		0.936437,
		0.952128,
		0.964969,
		0.975226,
		0.983187,
		0.989154,
		0.993437,
		0.996348,
		0.998186,
		0.999236,
		0.999752,
		0.99995,
		0.999997,
		1,
		0.999997,
		0.999955,
		0.999788,
		0.999381,
		0.9986,
		0.997312,
		0.995385,
		0.9927,
		0.989154,
		0.984657,
		0.979138,
		0.972543,
		0.964833,
		0.955987,
		0.945995,
		0.93486,
		0.922598,
		0.909232,
		0.894796,
		0.879329,
		0.862877,
		0.845492,
		0.827229,
		0.808146,
		0.788304,
		0.767766,
		0.746597,
		0.72486,
		0.702622,
		0.679948,
		0.656903,
		0.63355,
		0.609954,
		0.586177,
		0.56228,
		0.538323,
		0.514364,
		0.49046,
		0.466665,
		0.443032,
		0.419614,
		0.39646,
		0.373618,
		0.351133,
		0.32905,
		0.307411,
		0.286256,
		0.265625,
		0.245554,
		0.226079,
		0.207232,
		0.189047,
		0.171552,
		0.154777,
		0.138748,
		0.123492,
		0.109032,
		0.095391,
		0.082589,
		0.070648,
		0.059585,
		0.049418,
		0.040162,
		0.031834,
		0.024446,
		0.018011,
		0.012541,
		0.008046,
		0.004537,
		0.002021,
		0.000506,
		0.007635,
		0.028884,
		0.061335,
		0.102721,
		0.15096,
		0.204182,
		0.260738,
		0.319197,
		0.378334,
		0.437117,
		0.494688,
		0.550341,
		0.603506,
		0.653732,
		0.700674,
		0.74408,
		0.783783,
		0.819691,
		0.851782,
		0.880097,
		0.904734,
		0.925841,
		0.943612,
		0.95828,
		0.970112,
		0.9794,
		0.986456,
		0.991607,
		0.99518,
		0.997501,
		0.998879,
		0.999596,
		0.9999,
		0.999989,
		1,
		0.999999,
		0.999976,
		0.999854,
		0.999519,
		0.998834,
		0.997655,
		0.995843,
		0.99327,
		0.989821,
		0.985399,
		0.979927,
		0.973345,
		0.965609,
		0.956695,
		0.946591,
		0.935301,
		0.92284,
		0.909232,
		0.894513,
		0.878726,
		0.861918,
		0.844144,
		0.825463,
		0.805937,
		0.78563,
		0.76461,
		0.742944,
		0.720701,
		0.69795,
		0.67476,
		0.651201,
		0.627338,
		0.603241,
		0.578973,
		0.5546,
		0.530184,
		0.505785,
		0.481465,
		0.457278,
		0.433283,
		0.409531,
		0.386075,
		0.362964,
		0.340247,
		0.317968,
		0.296172,
		0.2749,
		0.254193,
		0.234088,
		0.214622,
		0.19583,
		0.177742,
		0.160392,
		0.143807,
		0.128015,
		0.113042,
		0.098913,
		0.08565,
		0.073275,
		0.061808,
		0.051267,
		0.041669,
		0.033031,
		0.025367,
		0.018691,
		0.013016,
		0.008351,
		0.004709,
		0.002098,
		0.000525,
		0.007731,
		0.029348,
		0.062493,
		0.104887,
		0.154387,
		0.209038,
		0.267095,
		0.327031,
		0.38754,
		0.447525,
		0.50608,
		0.56247,
		0.616116,
		0.666569,
		0.713495,
		0.756663,
		0.795928,
		0.831226,
		0.862561,
		0.890001,
		0.913673,
		0.933754,
		0.950468,
		0.964076,
		0.974876,
		0.983187,
		0.989348,
		0.993706,
		0.996609,
		0.998392,
		0.999369,
		0.999819,
		0.999972,
		0.999999,
		1,
		0.999989,
		0.999906,
		0.999638,
		0.999046,
		0.997978,
		0.996286,
		0.993829,
		0.990485,
		0.986146,
		0.980727,
		0.974162,
		0.966404,
		0.957423,
		0.947206,
		0.935757,
		0.92309,
		0.909232,
		0.89422,
		0.878098,
		0.860918,
		0.842738,
		0.823621,
		0.803632,
		0.782841,
		0.761317,
		0.739133,
		0.716363,
		0.693079,
		0.669354,
		0.645261,
		0.620871,
		0.596255,
		0.571482,
		0.546619,
		0.521733,
		0.496887,
		0.472143,
		0.447562,
		0.423201,
		0.399117,
		0.375363,
		0.35199,
		0.32905,
		0.306588,
		0.28465,
		0.26328,
		0.242518,
		0.222405,
		0.202976,
		0.184267,
		0.166312,
		0.149143,
		0.132788,
		0.117276,
		0.102633,
		0.088884,
		0.076051,
		0.064157,
		0.053221,
		0.043262,
		0.034297,
		0.026342,
		0.019411,
		0.013517,
		0.008674,
		0.004891,
		0.002179,
		0.000546,
		0.007746,
		0.029542,
		0.06316,
		0.10636,
		0.156977,
		0.212988,
		0.272552,
		0.334039,
		0.396042,
		0.457378,
		0.517072,
		0.574344,
		0.628589,
		0.679351,
		0.726308,
		0.769247,
		0.808054,
		0.842698,
		0.873218,
		0.89972,
		0.922363,
		0.941362,
		0.956972,
		0.969489,
		0.97924,
		0.986574,
		0.991857,
		0.995458,
		0.99774,
		0.999046,
		0.999689,
		0.999937,
		0.999996,
		1,
		0.999996,
		0.999944,
		0.999738,
		0.999236,
		0.99828,
		0.996711,
		0.994377,
		0.991144,
		0.986896,
		0.981538,
		0.974996,
		0.967218,
		0.958172,
		0.947842,
		0.936229,
		0.92335,
		0.909232,
		0.893914,
		0.877444,
		0.859876,
		0.841272,
		0.821699,
		0.801227,
		0.779928,
		0.757879,
		0.735156,
		0.711836,
		0.687996,
		0.663715,
		0.639069,
		0.614133,
		0.588982,
		0.563688,
		0.538323,
		0.512956,
		0.487653,
		0.462481,
		0.437501,
		0.412775,
		0.38836,
		0.364313,
		0.340688,
		0.317536,
		0.294906,
		0.272844,
		0.251397,
		0.230606,
		0.210511,
		0.191151,
		0.172561,
		0.154777,
		0.137829,
		0.121749,
		0.106564,
		0.092302,
		0.078987,
		0.066642,
		0.055289,
		0.044948,
		0.035637,
		0.027373,
		0.020173,
		0.014049,
		0.009016,
		0.005084,
		0.002265,
		0.000567,
		0.007659,
		0.029384,
		0.063155,
		0.106843,
		0.158311,
		0.215498,
		0.276483,
		0.339533,
		0.403127,
		0.465973,
		0.527004,
		0.585368,
		0.640413,
		0.691658,
		0.73878,
		0.781582,
		0.819979,
		0.853978,
		0.883665,
		0.909189,
		0.930757,
		0.948626,
		0.963094,
		0.974494,
		0.983187,
		0.989554,
		0.993987,
		0.996876,
		0.998595,
		0.999493,
		0.999875,
		0.999986,
		1,
		0.999999,
		0.99997,
		0.999819,
		0.999403,
		0.99856,
		0.997117,
		0.994912,
		0.991798,
		0.987648,
		0.982358,
		0.975845,
		0.968053,
		0.958943,
		0.948498,
		0.936718,
		0.92362,
		0.909232,
		0.893596,
		0.876762,
		0.858788,
		0.839741,
		0.819691,
		0.798713,
		0.776885,
		0.754286,
		0.731,
		0.707107,
		0.68269,
		0.657831,
		0.632611,
		0.607109,
		0.581406,
		0.555576,
		0.529695,
		0.503836,
		0.478069,
		0.452462,
		0.427082,
		0.40199,
		0.37725,
		0.352917,
		0.32905,
		0.3057,
		0.282919,
		0.260756,
		0.239256,
		0.218463,
		0.198419,
		0.179163,
		0.160731,
		0.14316,
		0.12648,
		0.110724,
		0.09592,
		0.082095,
		0.069274,
		0.057479,
		0.046734,
		0.037057,
		0.028466,
		0.02098,
		0.014613,
		0.009378,
		0.005289,
		0.002356,
		0.00059,
		0.007452,
		0.028801,
		0.062312,
		0.106041,
		0.157944,
		0.215973,
		0.278157,
		0.34267,
		0.407876,
		0.472358,
		0.534931,
		0.594644,
		0.650766,
		0.702768,
		0.750302,
		0.793172,
		0.831316,
		0.864779,
		0.893693,
		0.918264,
		0.938756,
		0.955481,
		0.968789,
		0.979061,
		0.9867,
		0.992122,
		0.995745,
		0.997978,
		0.999205,
		0.999771,
		0.999964,
		0.999999,
		1,
		0.999986,
		0.999882,
		0.999548,
		0.998815,
		0.997501,
		0.995431,
		0.992444,
		0.988401,
		0.983187,
		0.976711,
		0.968908,
		0.959736,
		0.949176,
		0.937225,
		0.9239,
		0.909232,
		0.893264,
		0.87605,
		0.857652,
		0.838141,
		0.817592,
		0.796084,
		0.773702,
		0.750529,
		0.726654,
		0.702164,
		0.677145,
		0.651685,
		0.625869,
		0.599783,
		0.573509,
		0.547127,
		0.520718,
		0.494356,
		0.468117,
		0.442072,
		0.416289,
		0.390835,
		0.365774,
		0.341166,
		0.317068,
		0.293538,
		0.270627,
		0.248386,
		0.226862,
		0.2061,
		0.186142,
		0.16703,
		0.148801,
		0.13149,
		0.11513,
		0.099754,
		0.085389,
		0.072063,
		0.059802,
		0.048628,
		0.038563,
		0.029626,
		0.021837,
		0.015211,
		0.009762,
		0.005506,
		0.002453,
		0.000615,
		0.007118,
		0.027745,
		0.060501,
		0.103703,
		0.155471,
		0.213831,
		0.276811,
		0.342523,
		0.409225,
		0.475373,
		0.539646,
		0.600964,
		0.658487,
		0.711604,
		0.75991,
		0.803188,
		0.841376,
		0.874547,
		0.902878,
		0.926633,
		0.94614,
		0.961779,
		0.973965,
		0.983139,
		0.989757,
		0.994276,
		0.997146,
		0.998793,
		0.999606,
		0.99992,
		0.999995,
		1,
		0.999995,
		0.999929,
		0.999671,
		0.999046,
		0.997863,
		0.995933,
		0.993081,
		0.989154,
		0.984024,
		0.977591,
		0.969784,
		0.960553,
		0.949876,
		0.937751,
		0.924192,
		0.909232,
		0.892918,
		0.875306,
		0.856465,
		0.836467,
		0.815394,
		0.793332,
		0.770369,
		0.746597,
		0.722107,
		0.696992,
		0.671346,
		0.645261,
		0.618827,
		0.592135,
		0.565272,
		0.538323,
		0.511372,
		0.484498,
		0.457781,
		0.431293,
		0.405109,
		0.379296,
		0.353922,
		0.32905,
		0.304739,
		0.281048,
		0.258031,
		0.23574,
		0.214224,
		0.193529,
		0.1737,
		0.154777,
		0.136799,
		0.119801,
		0.103819,
		0.088884,
		0.075024,
		0.062267,
		0.050639,
		0.040162,
		0.030859,
		0.022747,
		0.015846,
		0.010171,
		0.005737,
		0.002556,
		0.000641,
		0.006658,
		0.026203,
		0.057659,
		0.099666,
		0.150584,
		0.208583,
		0.271753,
		0.338191,
		0.406079,
		0.473757,
		0.539764,
		0.602874,
		0.662108,
		0.716734,
		0.766258,
		0.810403,
		0.849081,
		0.882368,
		0.910471,
		0.933706,
		0.952465,
		0.967203,
		0.97841,
		0.986599,
		0.99229,
		0.995994,
		0.998198,
		0.999349,
		0.999839,
		0.999982,
		1,
		0.999999,
		0.999962,
		0.999771,
		0.999251,
		0.9982,
		0.996415,
		0.993706,
		0.989904,
		0.984868,
		0.978488,
		0.970681,
		0.961395,
		0.950601,
		0.938296,
		0.924495,
		0.909232,
		0.892557,
		0.874529,
		0.855222,
		0.834714,
		0.813093,
		0.790449,
		0.766878,
		0.742476,
		0.717343,
		0.691577,
		0.665277,
		0.638541,
		0.611467,
		0.584147,
		0.556677,
		0.529145,
		0.501639,
		0.474244,
		0.447041,
		0.42011,
		0.393526,
		0.367361,
		0.341684,
		0.316562,
		0.292057,
		0.268228,
		0.245133,
		0.222825,
		0.201355,
		0.18077,
		0.161115,
		0.142431,
		0.12476,
		0.108136,
		0.092596,
		0.07817,
		0.064888,
		0.052777,
		0.041863,
		0.032169,
		0.023716,
		0.016522,
		0.010606,
		0.005983,
		0.002666,
		0.000668,
		0.006083,
		0.024201,
		0.053801,
		0.093895,
		0.143141,
		0.199929,
		0.26248,
		0.328938,
		0.397462,
		0.466305,
		0.533882,
		0.598814,
		0.659966,
		0.716456,
		0.767663,
		0.81321,
		0.852944,
		0.88691,
		0.91532,
		0.938519,
		0.956957,
		0.971156,
		0.981687,
		0.989142,
		0.994116,
		0.997181,
		0.998871,
		0.999659,
		0.999936,
		0.999993,
		0.999998,
		0.999982,
		0.99985,
		0.999429,
		0.998511,
		0.996876,
		0.994317,
		0.99065,
		0.985719,
		0.9794,
		0.971601,
		0.962262,
		0.951351,
		0.938862,
		0.924811,
		0.909232,
		0.892179,
		0.873716,
		0.85392,
		0.832877,
		0.81068,
		0.787425,
		0.763216,
		0.738155,
		0.712348,
		0.685901,
		0.65892,
		0.631507,
		0.603767,
		0.575799,
		0.547701,
		0.51957,
		0.491497,
		0.463572,
		0.43588,
		0.408505,
		0.381525,
		0.355015,
		0.32905,
		0.303696,
		0.279019,
		0.255081,
		0.231941,
		0.209654,
		0.188271,
		0.167843,
		0.148414,
		0.130029,
		0.112726,
		0.096544,
		0.081516,
		0.067676,
		0.055053,
		0.043674,
		0.033565,
		0.024747,
		0.017243,
		0.011069,
		0.006244,
		0.002783,
		0.000697,
		0.005415,
		0.021807,
		0.049033,
		0.086499,
		0.133206,
		0.187825,
		0.24878,
		0.314334,
		0.382687,
		0.45206,
		0.520775,
		0.58732,
		0.650402,
		0.708972,
		0.76225,
		0.809723,
		0.851131,
		0.886448,
		0.915851,
		0.939685,
		0.95843,
		0.972665,
		0.983029,
		0.990193,
		0.994828,
		0.997575,
		0.999017,
		0.999655,
		0.99988,
		0.999952,
		0.999976,
		0.999904,
		0.99958,
		0.998793,
		0.997312,
		0.994912,
		0.99139,
		0.986574,
		0.980326,
		0.972543,
		0.963156,
		0.952128,
		0.939451,
		0.92514,
		0.909232,
		0.891784,
		0.872864,
		0.852555,
		0.830949,
		0.808146,
		0.78425,
		0.759371,
		0.733619,
		0.707107,
		0.679948,
		0.652255,
		0.624138,
		0.595706,
		0.567067,
		0.538323,
		0.509577,
		0.480925,
		0.452462,
		0.424278,
		0.39646,
		0.369091,
		0.342249,
		0.31601,
		0.290446,
		0.265625,
		0.24161,
		0.218463,
		0.19624,
		0.174994,
		0.154777,
		0.135635,
		0.117611,
		0.100747,
		0.085081,
		0.070648,
		0.057479,
		0.045605,
		0.035053,
		0.025848,
		0.018011,
		0.011564,
		0.006524,
		0.002907,
		0.000729,
		0.004685,
		0.019122,
		0.043541,
		0.077731,
		0.121059,
		0.172521,
		0.230803,
		0.294361,
		0.361507,
		0.430498,
		0.499619,
		0.567267,
		0.632006,
		0.692627,
		0.748179,
		0.797981,
		0.841631,
		0.878988,
		0.910151,
		0.935426,
		0.955287,
		0.970337,
		0.981267,
		0.988808,
		0.993699,
		0.996642,
		0.998268,
		0.9991,
		0.999523,
		0.999756,
		0.999837,
		0.999659,
		0.999028,
		0.997716,
		0.995487,
		0.992122,
		0.987433,
		0.981267,
		0.973507,
		0.964076,
		0.952932,
		0.940063,
		0.925483,
		0.909232,
		0.89137,
		0.87197,
		0.851122,
		0.828924,
		0.805484,
		0.780913,
		0.755329,
		0.728851,
		0.701601,
		0.673697,
		0.645261,
		0.61641,
		0.587261,
		0.557927,
		0.528519,
		0.499142,
		0.469901,
		0.440894,
		0.412216,
		0.383959,
		0.356209,
		0.32905,
		0.302559,
		0.276811,
		0.251877,
		0.227824,
		0.204713,
		0.182603,
		0.161551,
		0.141606,
		0.122817,
		0.105229,
		0.088884,
		0.073818,
		0.060068,
		0.047667,
		0.036642,
		0.027023,
		0.018832,
		0.012092,
		0.006822,
		0.003041,
		0.000762,
		0.003928,
		0.01627,
		0.037571,
		0.067966,
		0.107181,
		0.154555,
		0.209081,
		0.269468,
		0.334211,
		0.401677,
		0.470186,
		0.538097,
		0.603886,
		0.666205,
		0.723937,
		0.77623,
		0.822512,
		0.862495,
		0.896157,
		0.923719,
		0.945607,
		0.96241,
		0.974831,
		0.983637,
		0.989607,
		0.993485,
		0.99593,
		0.997473,
		0.998474,
		0.9991,
		0.99933,
		0.999026,
		0.997988,
		0.995994,
		0.992826,
		0.988288,
		0.982219,
		0.974494,
		0.965026,
		0.953766,
		0.940699,
		0.925841,
		0.909232,
		0.890936,
		0.871032,
		0.849616,
		0.826794,
		0.802682,
		0.777401,
		0.751077,
		0.723836,
		0.69581,
		0.667127,
		0.637915,
		0.608301,
		0.578407,
		0.548355,
		0.518262,
		0.488241,
		0.4584,
		0.428844,
		0.399673,
		0.370984,
		0.342867,
		0.315408,
		0.28869,
		0.26279,
		0.237781,
		0.213733,
		0.19071,
		0.168771,
		0.147975,
		0.128373,
		0.110014,
		0.092945,
		0.077206,
		0.062836,
		0.049871,
		0.038342,
		0.02828,
		0.01971,
		0.012657,
		0.007142,
		0.003183,
		0.000798,
		0.003178,
		0.013387,
		0.031406,
		0.057659,
		0.092199,
		0.134701,
		0.184485,
		0.240551,
		0.30164,
		0.366296,
		0.43295,
		0.5,
		0.565889,
		0.629185,
		0.68864,
		0.743241,
		0.792248,
		0.835202,
		0.871933,
		0.902538,
		0.92735,
		0.946897,
		0.961852,
		0.97297,
		0.981036,
		0.986797,
		0.990906,
		0.993865,
		0.995982,
		0.997356,
		0.997934,
		0.997575,
		0.996096,
		0.99331,
		0.989042,
		0.983139,
		0.975486,
		0.965998,
		0.954629,
		0.941362,
		0.926215,
		0.909232,
		0.89048,
		0.870045,
		0.848031,
		0.824552,
		0.799732,
		0.773702,
		0.746597,
		0.718555,
		0.689715,
		0.660215,
		0.630193,
		0.599783,
		0.569117,
		0.538323,
		0.507526,
		0.476846,
		0.446396,
		0.416289,
		0.38663,
		0.357518,
		0.32905,
		0.301315,
		0.2744,
		0.248386,
		0.223347,
		0.199356,
		0.176478,
		0.154777,
		0.13431,
		0.11513,
		0.097288,
		0.08083,
		0.065798,
		0.05223,
		0.040162,
		0.029626,
		0.020651,
		0.013263,
		0.007484,
		0.003336,
		0.000836,
		0.002468,
		0.010604,
		0.025332,
		0.047295,
		0.076818,
		0.113883,
		0.158129,
		0.208867,
		0.265117,
		0.325662,
		0.389109,
		0.453969,
		0.518731,
		0.581942,
		0.642277,
		0.698606,
		0.750038,
		0.795954,
		0.836021,
		0.870185,
		0.898653,
		0.921848,
		0.940363,
		0.954896,
		0.966179,
		0.974912,
		0.981687,
		0.986929,
		0.99085,
		0.993485,
		0.994761,
		0.994554,
		0.992725,
		0.989142,
		0.983695,
		0.9763,
		0.966903,
		0.955481,
		0.942036,
		0.926602,
		0.909231,
		0.890001,
		0.869007,
		0.846361,
		0.822187,
		0.796619,
		0.769799,
		0.741872,
		0.712986,
		0.683291,
		0.652936,
		0.622067,
		0.590828,
		0.559361,
		0.527802,
		0.496282,
		0.464929,
		0.433865,
		0.403207,
		0.373065,
		0.343545,
		0.314747,
		0.286766,
		0.259691,
		0.233606,
		0.208589,
		0.184715,
		0.162051,
		0.140662,
		0.120607,
		0.101941,
		0.084714,
		0.068974,
		0.054761,
		0.042115,
		0.031071,
		0.021661,
		0.013913,
		0.007852,
		0.003501,
		0.000878,
		0.001827,
		0.008036,
		0.019612,
		0.037339,
		0.061746,
		0.093074,
		0.131251,
		0.175889,
		0.226305,
		0.281552,
		0.34047,
		0.401745,
		0.463986,
		0.525792,
		0.585835,
		0.642924,
		0.696065,
		0.744506,
		0.787765,
		0.825636,
		0.858176,
		0.885676,
		0.908605,
		0.927549,
		0.943133,
		0.955943,
		0.966443,
		0.974912,
		0.981438,
		0.985989,
		0.988479,
		0.988808,
		0.986889,
		0.982663,
		0.9761,
		0.967203,
		0.956001,
		0.942546,
		0.926912,
		0.909189,
		0.889481,
		0.867908,
		0.844598,
		0.819691,
		0.793332,
		0.765677,
		0.736881,
		0.707107,
		0.676513,
		0.645261,
		0.613507,
		0.581406,
		0.549108,
		0.516759,
		0.484498,
		0.452462,
		0.420779,
		0.389573,
		0.358959,
		0.32905,
		0.299949,
		0.271757,
		0.244566,
		0.218463,
		0.193529,
		0.169842,
		0.14747,
		0.12648,
		0.106933,
		0.088884,
		0.072383,
		0.057479,
		0.044213,
		0.032625,
		0.022747,
		0.014613,
		0.008248,
		0.003677,
		0.000922,
		0.001275,
		0.005777,
		0.014469,
		0.028199,
		0.047631,
		0.073201,
		0.105078,
		0.143149,
		0.187018,
		0.236018,
		0.289251,
		0.34563,
		0.403941,
		0.462908,
		0.52127,
		0.577846,
		0.631602,
		0.681705,
		0.727561,
		0.768831,
		0.805429,
		0.837497,
		0.865353,
		0.889428,
		0.910182,
		0.928019,
		0.943198,
		0.955795,
		0.965756,
		0.97297,
		0.977327,
		0.978745,
		0.977186,
		0.972665,
		0.965238,
		0.955004,
		0.942087,
		0.926633,
		0.908799,
		0.888751,
		0.86666,
		0.842698,
		0.817037,
		0.789852,
		0.761316,
		0.731605,
		0.700893,
		0.669354,
		0.63716,
		0.604481,
		0.571482,
		0.538323,
		0.505161,
		0.472143,
		0.439414,
		0.407111,
		0.375363,
		0.344293,
		0.314019,
		0.28465,
		0.25629,
		0.229035,
		0.202976,
		0.178197,
		0.154777,
		0.132788,
		0.112297,
		0.093366,
		0.076051,
		0.060404,
		0.046472,
		0.034297,
		0.023917,
		0.015366,
		0.008674,
		0.003868,
		0.00097,
		0.000827,
		0.00389,
		0.010067,
		0.020197,
		0.035006,
		0.05506,
		0.080714,
		0.112084,
		0.14903,
		0.191158,
		0.237836,
		0.288229,
		0.341342,
		0.39608,
		0.451312,
		0.505937,
		0.558952,
		0.609506,
		0.656947,
		0.700845,
		0.740998,
		0.777408,
		0.810233,
		0.839721,
		0.86612,
		0.889587,
		0.910115,
		0.927549,
		0.941677,
		0.95229,
		0.95923,
		0.96241,
		0.961826,
		0.95755,
		0.949718,
		0.938519,
		0.924176,
		0.906936,
		0.887051,
		0.864779,
		0.840368,
		0.81406,
		0.786085,
		0.756663,
		0.726004,
		0.694312,
		0.661781,
		0.628601,
		0.594955,
		0.561021,
		0.526971,
		0.492972,
		0.459182,
		0.425753,
		0.392831,
		0.360554,
		0.32905,
		0.298442,
		0.268846,
		0.24037,
		0.213114,
		0.187172,
		0.162631,
		0.139573,
		0.118071,
		0.098193,
		0.080003,
		0.063557,
		0.048908,
		0.036101,
		0.025179,
		0.016179,
		0.009134,
		0.004073,
		0.001021,
		0.000487,
		0.002407,
		0.006505,
		0.013546,
		0.024257,
		0.039266,
		0.059053,
		0.08391,
		0.113912,
		0.148905,
		0.188515,
		0.232157,
		0.279078,
		0.328396,
		0.379157,
		0.430401,
		0.481216,
		0.530799,
		0.578504,
		0.623865,
		0.666607,
		0.706621,
		0.743918,
		0.778552,
		0.810531,
		0.839721,
		0.865825,
		0.888462,
		0.907247,
		0.921848,
		0.932032,
		0.937675,
		0.938773,
		0.935426,
		0.927827,
		0.916238,
		0.900971,
		0.882368,
		0.860781,
		0.836564,
		0.810056,
		0.781582,
		0.751442,
		0.719918,
		0.687268,
		0.653732,
		0.619534,
		0.584885,
		0.549981,
		0.515011,
		0.480153,
		0.445578,
		0.411446,
		0.377913,
		0.345123,
		0.313213,
		0.282312,
		0.25254,
		0.22401,
		0.196828,
		0.171089,
		0.146884,
		0.124297,
		0.103402,
		0.084269,
		0.066963,
		0.051539,
		0.038051,
		0.026543,
		0.017058,
		0.009632,
		0.004296,
		0.001077,
		0.00025,
		0.001329,
		0.003813,
		0.008349,
		0.015607,
		0.026217,
		0.040725,
		0.05954,
		0.082904,
		0.110866,
		0.143273,
		0.179781,
		0.219874,
		0.262895,
		0.308099,
		0.354699,
		0.401925,
		0.449081,
		0.495581,
		0.540983,
		0.584989,
		0.627422,
		0.66817,
		0.707107,
		0.743988,
		0.778389,
		0.809757,
		0.837497,
		0.861048,
		0.879941,
		0.893836,
		0.902538,
		0.906002,
		0.904319,
		0.897699,
		0.886448,
		0.870942,
		0.8516,
		0.828867,
		0.803188,
		0.774998,
		0.744711,
		0.712711,
		0.679351,
		0.644953,
		0.609807,
		0.574175,
		0.538297,
		0.502393,
		0.466662,
		0.431293,
		0.39646,
		0.362328,
		0.32905,
		0.29677,
		0.265625,
		0.23574,
		0.207232,
		0.180211,
		0.154777,
		0.131022,
		0.109032,
		0.088884,
		0.070648,
		0.054388,
		0.040162,
		0.028021,
		0.018011,
		0.010171,
		0.004537,
		0.001138,
		0.000105,
		0.000623,
		0.001951,
		0.004591,
		0.009106,
		0.016081,
		0.026068,
		0.039541,
		0.056853,
		0.078208,
		0.103642,
		0.133017,
		0.166038,
		0.202269,
		0.241175,
		0.282165,
		0.32464,
		0.368043,
		0.4119,
		0.455837,
		0.499583,
		0.542935,
		0.585696,
		0.627576,
		0.668102,
		0.706621,
		0.742376,
		0.774593,
		0.80255,
		0.825636,
		0.843385,
		0.855502,
		0.861859,
		0.862495,
		0.857596,
		0.847468,
		0.832516,
		0.81321,
		0.790059,
		0.76359,
		0.734326,
		0.702768,
		0.66939,
		0.634625,
		0.598868,
		0.56247,
		0.525747,
		0.488978,
		0.452411,
		0.416269,
		0.380753,
		0.346046,
		0.312315,
		0.279714,
		0.248386,
		0.218463,
		0.190067,
		0.163312,
		0.138302,
		0.11513,
		0.093885,
		0.074644,
		0.057479,
		0.042454,
		0.029626,
		0.019046,
		0.010757,
		0.004799,
		0.001204,
		3.1e-005,
		0.000225,
		0.000815,
		0.002141,
		0.004637,
		0.008796,
		0.015128,
		0.024116,
		0.036175,
		0.051615,
		0.070615,
		0.09321,
		0.119294,
		0.148632,
		0.180892,
		0.215674,
		0.252561,
		0.291157,
		0.331119,
		0.372177,
		0.414121,
		0.456759,
		0.499836,
		0.542935,
		0.585423,
		0.626498,
		0.665273,
		0.700845,
		0.732363,
		0.759087,
		0.780424,
		0.795954,
		0.805445,
		0.808844,
		0.806267,
		0.797981,
		0.784374,
		0.765928,
		0.743188,
		0.716734,
		0.687158,
		0.655041,
		0.620939,
		0.585368,
		0.548801,
		0.51166,
		0.474322,
		0.437117,
		0.400334,
		0.364225,
		0.329012,
		0.294891,
		0.262037,
		0.230604,
		0.200736,
		0.172561,
		0.146196,
		0.121749,
		0.099316,
		0.078987,
		0.06084,
		0.044948,
		0.031373,
		0.020173,
		0.011395,
		0.005084,
		0.001275,
		5.1e-005,
		0.000242,
		0.00077,
		0.001926,
		0.00408,
		0.007659,
		0.01311,
		0.020858,
		0.031269,
		0.04462,
		0.061078,
		0.080692,
		0.103399,
		0.129044,
		0.15741,
		0.188251,
		0.22133,
		0.256446,
		0.293444,
		0.332194,
		0.372533,
		0.414176,
		0.456627,
		0.499182,
		0.540983,
		0.581082,
		0.618508,
		0.652329,
		0.681705,
		0.705931,
		0.724468,
		0.736961,
		0.743241,
		0.743327,
		0.737402,
		0.7258,
		0.708972,
		0.687459,
		0.661863,
		0.632819,
		0.600964,
		0.566923,
		0.531288,
		0.494607,
		0.457378,
		0.420042,
		0.382989,
		0.346557,
		0.311037,
		0.276679,
		0.243699,
		0.212284,
		0.182595,
		0.154774,
		0.128948,
		0.105229,
		0.083718,
		0.064503,
		0.047667,
		0.033279,
		0.021402,
		0.012092,
		0.005396,
		0.001354,
		3.6e-005,
		0.000173,
		0.000568,
		0.00146,
		0.003161,
		0.006041,
		0.010488,
		0.016877,
		0.025537,
		0.036721,
		0.050596,
		0.067237,
		0.086641,
		0.108744,
		0.133455,
		0.160683,
		0.190361,
		0.222444,
		0.25688,
		0.293546,
		0.332146,
		0.372177,
		0.412943,
		0.4536,
		0.493212,
		0.530799,
		0.565401,
		0.596121,
		0.622176,
		0.642924,
		0.657896,
		0.666804,
		0.669547,
		0.666205,
		0.65702,
		0.642378,
		0.622781,
		0.598814,
		0.57112,
		0.540369,
		0.507231,
		0.472358,
		0.436364,
		0.399815,
		0.36322,
		0.327031,
		0.291638,
		0.257379,
		0.224539,
		0.19336,
		0.164044,
		0.136765,
		0.111669,
		0.088879,
		0.068505,
		0.050639,
		0.035363,
		0.022747,
		0.012854,
		0.005737,
		0.00144,
		1.2e-005,
		8.3e-005,
		0.000326,
		0.000927,
		0.002146,
		0.004297,
		0.007717,
		0.012738,
		0.019659,
		0.02872,
		0.040099,
		0.053909,
		0.070213,
		0.08905,
		0.110452,
		0.134466,
		0.161142,
		0.190498,
		0.222444,
		0.256712,
		0.292845,
		0.330213,
		0.368043,
		0.405462,
		0.441538,
		0.475331,
		0.505937,
		0.532534,
		0.554413,
		0.571013,
		0.581942,
		0.586982,
		0.586099,
		0.579429,
		0.567267,
		0.550042,
		0.528295,
		0.502643,
		0.473757,
		0.442328,
		0.409045,
		0.374572,
		0.339533,
		0.304496,
		0.269972,
		0.236406,
		0.204182,
		0.173627,
		0.14501,
		0.118558,
		0.094456,
		0.072856,
		0.053884,
		0.037644,
		0.024222,
		0.013691,
		0.006111,
		0.001534,
		2.3e-005,
		0.000135,
		0.000475,
		0.001249,
		0.002719,
		0.005174,
		0.00891,
		0.014202,
		0.021286,
		0.030358,
		0.041577,
		0.055078,
		0.070995,
		0.089462,
		0.110603,
		0.134491,
		0.161075,
		0.190146,
		0.22133,
		0.254094,
		0.287772,
		0.321588,
		0.354699,
		0.386226,
		0.415302,
		0.441107,
		0.462908,
		0.48009,
		0.492178,
		0.498862,
		0.5,
		0.495623,
		0.485925,
		0.471247,
		0.45206,
		0.428931,
		0.402506,
		0.373469,
		0.342523,
		0.310362,
		0.277651,
		0.245006,
		0.212988,
		0.182095,
		0.152754,
		0.125331,
		0.100129,
		0.077397,
		0.057334,
		0.040101,
		0.025825,
		0.014606,
		0.006522,
		0.001637,
		3.3e-005,
		0.00018,
		0.000596,
		0.001491,
		0.00311,
		0.005716,
		0.009563,
		0.014891,
		0.021916,
		0.030845,
		0.041877,
		0.055213,
		0.071031,
		0.089443,
		0.110452,
		0.133926,
		0.159588,
		0.18702,
		0.215674,
		0.244896,
		0.273949,
		0.302048,
		0.328396,
		0.352213,
		0.37278,
		0.389462,
		0.401745,
		0.40925,
		0.411755,
		0.409197,
		0.401677,
		0.389445,
		0.372893,
		0.352525,
		0.328938,
		0.302792,
		0.274783,
		0.245614,
		0.215973,
		0.186511,
		0.157829,
		0.130464,
		0.104887,
		0.081496,
		0.060626,
		0.042546,
		0.027471,
		0.015567,
		0.006962,
		0.001749,
		4.1e-005,
		0.000213,
		0.000676,
		0.001631,
		0.003307,
		0.005944,
		0.00978,
		0.015054,
		0.022008,
		0.030877,
		0.041877,
		0.05516,
		0.070788,
		0.088707,
		0.108744,
		0.130593,
		0.153832,
		0.177929,
		0.202269,
		0.226175,
		0.248937,
		0.269848,
		0.288229,
		0.303466,
		0.315035,
		0.322525,
		0.325662,
		0.324316,
		0.318512,
		0.308422,
		0.294361,
		0.276767,
		0.256183,
		0.23323,
		0.208583,
		0.182939,
		0.156993,
		0.131419,
		0.106843,
		0.083836,
		0.062899,
		0.044463,
		0.028884,
		0.01645,
		0.007387,
		0.001862,
		4.6e-005,
		0.000231,
		0.000712,
		0.001682,
		0.003361,
		0.005985,
		0.0098,
		0.015059,
		0.021994,
		0.030794,
		0.041577,
		0.054366,
		0.069081,
		0.085527,
		0.103399,
		0.122288,
		0.141698,
		0.161065,
		0.179781,
		0.197224,
		0.212781,
		0.22588,
		0.236018,
		0.242785,
		0.245885,
		0.24515,
		0.240551,
		0.232205,
		0.220363,
		0.205406,
		0.187825,
		0.168202,
		0.147183,
		0.125451,
		0.103703,
		0.08262,
		0.062848,
		0.04498,
		0.029542,
		0.016988,
		0.007692,
		0.001953,
		4.8e-005,
		0.000236,
		0.000719,
		0.001689,
		0.003364,
		0.005983,
		0.00978,
		0.014966,
		0.021703,
		0.03008,
		0.040099,
		0.051666,
		0.064584,
		0.078559,
		0.09321,
		0.108085,
		0.122681,
		0.136466,
		0.148905,
		0.159489,
		0.167757,
		0.17332,
		0.175889,
		0.175283,
		0.171451,
		0.164471,
		0.154555,
		0.142042,
		0.127383,
		0.111127,
		0.093895,
		0.076359,
		0.05921,
		0.043137,
		0.028801,
		0.016812,
		0.007716,
		0.001983,
		4.9e-005,
		0.000237,
		0.000719,
		0.001686,
		0.003347,
		0.005911,
		0.009563,
		0.014441,
		0.020615,
		0.028075,
		0.036721,
		0.046365,
		0.056733,
		0.067482,
		0.078208,
		0.088475,
		0.097831,
		0.105836,
		0.112084,
		0.116229,
		0.118003,
		0.11724,
		0.113883,
		0.108001,
		0.099785,
		0.089552,
		0.077731,
		0.064851,
		0.051521,
		0.038407,
		0.026203,
		0.015609,
		0.007301,
		0.00191,
		4.9e-005,
		0.000236,
		0.000712,
		0.001651,
		0.003231,
		0.005612,
		0.00891,
		0.013182,
		0.018409,
		0.024496,
		0.031269,
		0.038483,
		0.045834,
		0.052975,
		0.05954,
		0.065163,
		0.0695,
		0.072255,
		0.073201,
		0.072196,
		0.069201,
		0.064291,
		0.057659,
		0.04962,
		0.040599,
		0.031123,
		0.021807,
		0.013323,
		0.006383,
		0.001708,
		4.8e-005,
		0.000228,
		0.000676,
		0.001535,
		0.002939,
		0.004985,
		0.007717,
		0.011116,
		0.015095,
		0.019499,
		0.024116,
		0.028687,
		0.032928,
		0.036547,
		0.039266,
		0.040846,
		0.041104,
		0.039939,
		0.037339,
		0.033402,
		0.02834,
		0.022482,
		0.01627,
		0.01025,
		0.005056,
		0.001391,
		4.4e-005,
		0.000206,
		0.000596,
		0.001319,
		0.002452,
		0.004032,
		0.006041,
		0.008402,
		0.010985,
		0.013614,
		0.016081,
		0.018168,
		0.019662,
		0.020382,
		0.020197,
		0.019045,
		0.016957,
		0.014062,
		0.010604,
		0.006942,
		0.00355,
		0.00101,
		3.8e-005,
		0.00017,
		0.000475,
		0.001013,
		0.001811,
		0.002854,
		0.00408,
		0.005384,
		0.006631,
		0.007669,
		0.008349,
		0.008546,
		0.008178,
		0.007231,
		0.005777,
		0.003985,
		0.002137,
		0.000636,
		2.8e-005,
		0.000122,
		0.000326,
		0.000658,
		0.001107,
		0.001626,
		0.002141,
		0.002558,
		0.002784,
		0.002744,
		0.002407,
		0.001801,
		0.001038,
		0.000329,
		1.8e-005,
		7.1e-005,
		0.000173,
		0.000316,
		0.00047,
		0.000588,
		0.000623,
		0.000544,
		0.000356,
		0.000126,
		2.6e-005,
		5.1e-005,
		6.8e-005,
		6.1e-005,
		2.7e-005,
		2.7e-005,
		6.1e-005,
		6.8e-005,
		5.1e-005,
		2.6e-005,
		0.000126,
		0.000356,
		0.000544,
		0.000623,
		0.000588,
		0.00047,
		0.000316,
		0.000173,
		7.1e-005,
		1.8e-005,
		0.000329,
		0.001038,
		0.001801,
		0.002407,
		0.002744,
		0.002784,
		0.002558,
		0.002141,
		0.001626,
		0.001107,
		0.000658,
		0.000326,
		0.000122,
		2.8e-005,
		0.000636,
		0.002137,
		0.003985,
		0.005777,
		0.007231,
		0.008178,
		0.008546,
		0.008349,
		0.007669,
		0.006631,
		0.005384,
		0.00408,
		0.002854,
		0.001811,
		0.001013,
		0.000475,
		0.00017,
		3.8e-005,
		0.00101,
		0.00355,
		0.006942,
		0.010604,
		0.014062,
		0.016957,
		0.019045,
		0.020197,
		0.020382,
		0.019662,
		0.018168,
		0.016081,
		0.013614,
		0.010985,
		0.008402,
		0.006041,
		0.004032,
		0.002452,
		0.001319,
		0.000596,
		0.000206,
		4.4e-005,
		0.001391,
		0.005056,
		0.01025,
		0.01627,
		0.022482,
		0.02834,
		0.033402,
		0.037339,
		0.039939,
		0.041104,
		0.040846,
		0.039266,
		0.036547,
		0.032928,
		0.028687,
		0.024116,
		0.019499,
		0.015095,
		0.011116,
		0.007717,
		0.004985,
		0.002939,
		0.001535,
		0.000676,
		0.000228,
		4.8e-005,
		0.001708,
		0.006383,
		0.013323,
		0.021807,
		0.031123,
		0.040599,
		0.04962,
		0.057659,
		0.064291,
		0.069201,
		0.072196,
		0.073201,
		0.072255,
		0.0695,
		0.065163,
		0.05954,
		0.052975,
		0.045834,
		0.038483,
		0.031269,
		0.024496,
		0.018409,
		0.013182,
		0.00891,
		0.005612,
		0.003231,
		0.001651,
		0.000712,
		0.000236,
		4.9e-005,
		0.00191,
		0.007301,
		0.015609,
		0.026203,
		0.038407,
		0.051521,
		0.064851,
		0.077731,
		0.089552,
		0.099785,
		0.108001,
		0.113883,
		0.11724,
		0.118003,
		0.116229,
		0.112084,
		0.105836,
		0.097831,
		0.088475,
		0.078208,
		0.067482,
		0.056733,
		0.046365,
		0.036721,
		0.028075,
		0.020615,
		0.014441,
		0.009563,
		0.005911,
		0.003347,
		0.001686,
		0.000719,
		0.000237,
		4.9e-005,
		0.001983,
		0.007716,
		0.016812,
		0.028801,
		0.043137,
		0.05921,
		0.076359,
		0.093895,
		0.111127,
		0.127383,
		0.142042,
		0.154555,
		0.164471,
		0.171451,
		0.175283,
		0.175889,
		0.17332,
		0.167757,
		0.159489,
		0.148905,
		0.136466,
		0.122681,
		0.108085,
		0.09321,
		0.078559,
		0.064584,
		0.051666,
		0.040099,
		0.03008,
		0.021703,
		0.014966,
		0.00978,
		0.005983,
		0.003364,
		0.001689,
		0.000719,
		0.000236,
		4.8e-005,
		0.001953,
		0.007692,
		0.016988,
		0.029542,
		0.04498,
		0.062848,
		0.08262,
		0.103703,
		0.125451,
		0.147183,
		0.168202,
		0.187825,
		0.205406,
		0.220363,
		0.232205,
		0.240551,
		0.24515,
		0.245885,
		0.242785,
		0.236018,
		0.22588,
		0.212781,
		0.197224,
		0.179781,
		0.161065,
		0.141698,
		0.122288,
		0.103399,
		0.085527,
		0.069081,
		0.054366,
		0.041577,
		0.030794,
		0.021994,
		0.015059,
		0.0098,
		0.005985,
		0.003361,
		0.001682,
		0.000712,
		0.000231,
		4.6e-005,
		0.001862,
		0.007387,
		0.01645,
		0.028884,
		0.044463,
		0.062899,
		0.083836,
		0.106843,
		0.131419,
		0.156993,
		0.182939,
		0.208583,
		0.23323,
		0.256183,
		0.276767,
		0.294361,
		0.308422,
		0.318512,
		0.324316,
		0.325662,
		0.322525,
		0.315035,
		0.303466,
		0.288229,
		0.269848,
		0.248937,
		0.226175,
		0.202269,
		0.177929,
		0.153832,
		0.130593,
		0.108744,
		0.088707,
		0.070788,
		0.05516,
		0.041877,
		0.030877,
		0.022008,
		0.015054,
		0.00978,
		0.005944,
		0.003307,
		0.001631,
		0.000676,
		0.000213,
		4.1e-005,
		0.001749,
		0.006962,
		0.015567,
		0.027471,
		0.042546,
		0.060626,
		0.081496,
		0.104887,
		0.130464,
		0.157829,
		0.186511,
		0.215973,
		0.245614,
		0.274783,
		0.302792,
		0.328938,
		0.352525,
		0.372893,
		0.389445,
		0.401677,
		0.409197,
		0.411755,
		0.40925,
		0.401745,
		0.389462,
		0.37278,
		0.352213,
		0.328396,
		0.302048,
		0.273949,
		0.244896,
		0.215674,
		0.18702,
		0.159588,
		0.133926,
		0.110452,
		0.089443,
		0.071031,
		0.055213,
		0.041877,
		0.030845,
		0.021916,
		0.014891,
		0.009563,
		0.005716,
		0.00311,
		0.001491,
		0.000596,
		0.00018,
		3.3e-005,
		0.001637,
		0.006522,
		0.014606,
		0.025825,
		0.040101,
		0.057334,
		0.077397,
		0.100129,
		0.125331,
		0.152754,
		0.182095,
		0.212988,
		0.245006,
		0.277651,
		0.310362,
		0.342523,
		0.373469,
		0.402506,
		0.428931,
		0.45206,
		0.471247,
		0.485925,
		0.495623,
		0.5,
		0.498862,
		0.492178,
		0.48009,
		0.462908,
		0.441107,
		0.415302,
		0.386226,
		0.354699,
		0.321588,
		0.287772,
		0.254094,
		0.22133,
		0.190146,
		0.161075,
		0.134491,
		0.110603,
		0.089462,
		0.070995,
		0.055078,
		0.041577,
		0.030358,
		0.021286,
		0.014202,
		0.00891,
		0.005174,
		0.002719,
		0.001249,
		0.000475,
		0.000135,
		2.3e-005,
		0.001534,
		0.006111,
		0.013691,
		0.024222,
		0.037644,
		0.053884,
		0.072856,
		0.094456,
		0.118558,
		0.14501,
		0.173627,
		0.204182,
		0.236406,
		0.269972,
		0.304496,
		0.339533,
		0.374572,
		0.409045,
		0.442328,
		0.473757,
		0.502643,
		0.528295,
		0.550042,
		0.567267,
		0.579429,
		0.586099,
		0.586982,
		0.581942,
		0.571013,
		0.554413,
		0.532534,
		0.505937,
		0.475331,
		0.441538,
		0.405462,
		0.368043,
		0.330213,
		0.292845,
		0.256712,
		0.222444,
		0.190498,
		0.161142,
		0.134466,
		0.110452,
		0.08905,
		0.070213,
		0.053909,
		0.040099,
		0.02872,
		0.019659,
		0.012738,
		0.007717,
		0.004297,
		0.002146,
		0.000927,
		0.000326,
		8.3e-005,
		1.2e-005,
		0.00144,
		0.005737,
		0.012854,
		0.022747,
		0.035363,
		0.050639,
		0.068505,
		0.088879,
		0.111669,
		0.136765,
		0.164044,
		0.19336,
		0.224539,
		0.257379,
		0.291638,
		0.327031,
		0.36322,
		0.399815,
		0.436364,
		0.472358,
		0.507231,
		0.540369,
		0.57112,
		0.598814,
		0.622781,
		0.642378,
		0.65702,
		0.666205,
		0.669547,
		0.666804,
		0.657896,
		0.642924,
		0.622176,
		0.596121,
		0.565401,
		0.530799,
		0.493212,
		0.4536,
		0.412943,
		0.372177,
		0.332146,
		0.293546,
		0.25688,
		0.222444,
		0.190361,
		0.160683,
		0.133455,
		0.108744,
		0.086641,
		0.067237,
		0.050596,
		0.036721,
		0.025537,
		0.016877,
		0.010488,
		0.006041,
		0.003161,
		0.00146,
		0.000568,
		0.000173,
		3.6e-005,
		0.001354,
		0.005396,
		0.012092,
		0.021402,
		0.033279,
		0.047667,
		0.064503,
		0.083718,
		0.105229,
		0.128948,
		0.154774,
		0.182595,
		0.212284,
		0.243699,
		0.276679,
		0.311037,
		0.346557,
		0.382989,
		0.420042,
		0.457378,
		0.494607,
		0.531288,
		0.566923,
		0.600964,
		0.632819,
		0.661863,
		0.687459,
		0.708972,
		0.7258,
		0.737402,
		0.743327,
		0.743241,
		0.736961,
		0.724468,
		0.705931,
		0.681705,
		0.652329,
		0.618508,
		0.581082,
		0.540983,
		0.499182,
		0.456627,
		0.414176,
		0.372533,
		0.332194,
		0.293444,
		0.256446,
		0.22133,
		0.188251,
		0.15741,
		0.129044,
		0.103399,
		0.080692,
		0.061078,
		0.04462,
		0.031269,
		0.020858,
		0.01311,
		0.007659,
		0.00408,
		0.001926,
		0.00077,
		0.000242,
		5.1e-005,
		0.001275,
		0.005084,
		0.011395,
		0.020173,
		0.031373,
		0.044948,
		0.06084,
		0.078987,
		0.099316,
		0.121749,
		0.146196,
		0.172561,
		0.200736,
		0.230604,
		0.262037,
		0.294891,
		0.329012,
		0.364225,
		0.400334,
		0.437117,
		0.474322,
		0.51166,
		0.548801,
		0.585368,
		0.620939,
		0.655041,
		0.687158,
		0.716734,
		0.743188,
		0.765928,
		0.784374,
		0.797981,
		0.806267,
		0.808844,
		0.805445,
		0.795954,
		0.780424,
		0.759087,
		0.732363,
		0.700845,
		0.665273,
		0.626498,
		0.585423,
		0.542935,
		0.499836,
		0.456759,
		0.414121,
		0.372177,
		0.331119,
		0.291157,
		0.252561,
		0.215674,
		0.180892,
		0.148632,
		0.119294,
		0.09321,
		0.070615,
		0.051615,
		0.036175,
		0.024116,
		0.015128,
		0.008796,
		0.004637,
		0.002141,
		0.000815,
		0.000225,
		3.1e-005,
		0.001204,
		0.004799,
		0.010757,
		0.019046,
		0.029626,
		0.042454,
		0.057479,
		0.074644,
		0.093885,
		0.11513,
		0.138302,
		0.163312,
		0.190067,
		0.218463,
		0.248386,
		0.279714,
		0.312315,
		0.346046,
		0.380753,
		0.416269,
		0.452411,
		0.488978,
		0.525747,
		0.56247,
		0.598868,
		0.634625,
		0.66939,
		0.702768,
		0.734326,
		0.76359,
		0.790059,
		0.81321,
		0.832516,
		0.847468,
		0.857596,
		0.862495,
		0.861859,
		0.855502,
		0.843385,
		0.825636,
		0.80255,
		0.774593,
		0.742376,
		0.706621,
		0.668102,
		0.627576,
		0.585696,
		0.542935,
		0.499583,
		0.455837,
		0.4119,
		0.368043,
		0.32464,
		0.282165,
		0.241175,
		0.202269,
		0.166038,
		0.133017,
		0.103642,
		0.078208,
		0.056853,
		0.039541,
		0.026068,
		0.016081,
		0.009106,
		0.004591,
		0.001951,
		0.000623,
		0.000105,
		0.001138,
		0.004537,
		0.010171,
		0.018011,
		0.028021,
		0.040162,
		0.054388,
		0.070648,
		0.088884,
		0.109032,
		0.131022,
		0.154777,
		0.180211,
		0.207232,
		0.23574,
		0.265625,
		0.29677,
		0.32905,
		0.362328,
		0.39646,
		0.431293,
		0.466662,
		0.502393,
		0.538297,
		0.574175,
		0.609807,
		0.644953,
		0.679351,
		0.712711,
		0.744711,
		0.774998,
		0.803188,
		0.828867,
		0.8516,
		0.870942,
		0.886448,
		0.897699,
		0.904319,
		0.906002,
		0.902538,
		0.893836,
		0.879941,
		0.861048,
		0.837497,
		0.809757,
		0.778389,
		0.743988,
		0.707107,
		0.66817,
		0.627422,
		0.584989,
		0.540983,
		0.495581,
		0.449081,
		0.401925,
		0.354699,
		0.308099,
		0.262895,
		0.219874,
		0.179781,
		0.143273,
		0.110866,
		0.082904,
		0.05954,
		0.040725,
		0.026217,
		0.015607,
		0.008349,
		0.003813,
		0.001329,
		0.00025,
		0.001077,
		0.004296,
		0.009632,
		0.017058,
		0.026543,
		0.038051,
		0.051539,
		0.066963,
		0.084269,
		0.103402,
		0.124297,
		0.146884,
		0.171089,
		0.196828,
		0.22401,
		0.25254,
		0.282312,
		0.313213,
		0.345123,
		0.377913,
		0.411446,
		0.445578,
		0.480153,
		0.515011,
		0.549981,
		0.584885,
		0.619534,
		0.653732,
		0.687268,
		0.719918,
		0.751442,
		0.781582,
		0.810056,
		0.836564,
		0.860781,
		0.882368,
		0.900971,
		0.916238,
		0.927827,
		0.935426,
		0.938773,
		0.937675,
		0.932032,
		0.921848,
		0.907247,
		0.888462,
		0.865825,
		0.839721,
		0.810531,
		0.778552,
		0.743918,
		0.706621,
		0.666607,
		0.623865,
		0.578504,
		0.530799,
		0.481216,
		0.430401,
		0.379157,
		0.328396,
		0.279078,
		0.232157,
		0.188515,
		0.148905,
		0.113912,
		0.08391,
		0.059053,
		0.039266,
		0.024257,
		0.013546,
		0.006505,
		0.002407,
		0.000487,
		0.001021,
		0.004073,
		0.009134,
		0.016179,
		0.025179,
		0.036101,
		0.048908,
		0.063557,
		0.080003,
		0.098193,
		0.118071,
		0.139573,
		0.162631,
		0.187172,
		0.213114,
		0.24037,
		0.268846,
		0.298442,
		0.32905,
		0.360554,
		0.392831,
		0.425753,
		0.459182,
		0.492972,
		0.526971,
		0.561021,
		0.594955,
		0.628601,
		0.661781,
		0.694312,
		0.726004,
		0.756663,
		0.786085,
		0.81406,
		0.840368,
		0.864779,
		0.887051,
		0.906936,
		0.924176,
		0.938519,
		0.949718,
		0.95755,
		0.961826,
		0.96241,
		0.95923,
		0.95229,
		0.941677,
		0.927549,
		0.910115,
		0.889587,
		0.86612,
		0.839721,
		0.810233,
		0.777408,
		0.740998,
		0.700845,
		0.656947,
		0.609506,
		0.558952,
		0.505937,
		0.451312,
		0.39608,
		0.341342,
		0.288229,
		0.237836,
		0.191158,
		0.14903,
		0.112084,
		0.080714,
		0.05506,
		0.035006,
		0.020197,
		0.010067,
		0.00389,
		0.000827,
		0.00097,
		0.003868,
		0.008674,
		0.015366,
		0.023917,
		0.034297,
		0.046472,
		0.060404,
		0.076051,
		0.093366,
		0.112297,
		0.132788,
		0.154777,
		0.178197,
		0.202976,
		0.229035,
		0.25629,
		0.28465,
		0.314019,
		0.344293,
		0.375363,
		0.407111,
		0.439414,
		0.472143,
		0.505161,
		0.538323,
		0.571482,
		0.604481,
		0.63716,
		0.669354,
		0.700893,
		0.731605,
		0.761316,
		0.789852,
		0.817037,
		0.842698,
		0.86666,
		0.888751,
		0.908799,
		0.926633,
		0.942087,
		0.955004,
		0.965238,
		0.972665,
		0.977186,
		0.978745,
		0.977327,
		0.97297,
		0.965756,
		0.955795,
		0.943198,
		0.928019,
		0.910182,
		0.889428,
		0.865353,
		0.837497,
		0.805429,
		0.768831,
		0.727561,
		0.681705,
		0.631602,
		0.577846,
		0.52127,
		0.462908,
		0.403941,
		0.34563,
		0.289251,
		0.236018,
		0.187018,
		0.143149,
		0.105078,
		0.073201,
		0.047631,
		0.028199,
		0.014469,
		0.005777,
		0.001275,
		0.000922,
		0.003677,
		0.008248,
		0.014613,
		0.022747,
		0.032625,
		0.044213,
		0.057479,
		0.072383,
		0.088884,
		0.106933,
		0.12648,
		0.14747,
		0.169842,
		0.193529,
		0.218463,
		0.244566,
		0.271757,
		0.299949,
		0.32905,
		0.358959,
		0.389573,
		0.420779,
		0.452462,
		0.484498,
		0.516759,
		0.549108,
		0.581406,
		0.613507,
		0.645261,
		0.676513,
		0.707107,
		0.736881,
		0.765677,
		0.793332,
		0.819691,
		0.844598,
		0.867908,
		0.889481,
		0.909189,
		0.926912,
		0.942546,
		0.956001,
		0.967203,
		0.9761,
		0.982663,
		0.986889,
		0.988808,
		0.988479,
		0.985989,
		0.981438,
		0.974912,
		0.966443,
		0.955943,
		0.943133,
		0.927549,
		0.908605,
		0.885676,
		0.858176,
		0.825636,
		0.787765,
		0.744506,
		0.696065,
		0.642924,
		0.585835,
		0.525792,
		0.463986,
		0.401745,
		0.34047,
		0.281552,
		0.226305,
		0.175889,
		0.131251,
		0.093074,
		0.061746,
		0.037339,
		0.019612,
		0.008036,
		0.001827,
		0.000878,
		0.003501,
		0.007852,
		0.013913,
		0.021661,
		0.031071,
		0.042115,
		0.054761,
		0.068974,
		0.084714,
		0.101941,
		0.120607,
		0.140662,
		0.162051,
		0.184715,
		0.208589,
		0.233606,
		0.259691,
		0.286766,
		0.314747,
		0.343545,
		0.373065,
		0.403207,
		0.433865,
		0.464929,
		0.496282,
		0.527802,
		0.559361,
		0.590828,
		0.622067,
		0.652936,
		0.683291,
		0.712986,
		0.741872,
		0.769799,
		0.796619,
		0.822187,
		0.846361,
		0.869007,
		0.890001,
		0.909231,
		0.926602,
		0.942036,
		0.955481,
		0.966903,
		0.9763,
		0.983695,
		0.989142,
		0.992725,
		0.994554,
		0.994761,
		0.993485,
		0.99085,
		0.986929,
		0.981687,
		0.974912,
		0.966179,
		0.954896,
		0.940363,
		0.921848,
		0.898653,
		0.870185,
		0.836021,
		0.795954,
		0.750038,
		0.698606,
		0.642277,
		0.581942,
		0.518731,
		0.453969,
		0.389109,
		0.325662,
		0.265117,
		0.208867,
		0.158129,
		0.113883,
		0.076818,
		0.047295,
		0.025332,
		0.010604,
		0.002468,
		0.000836,
		0.003336,
		0.007484,
		0.013263,
		0.020651,
		0.029626,
		0.040162,
		0.05223,
		0.065798,
		0.08083,
		0.097288,
		0.11513,
		0.13431,
		0.154777,
		0.176478,
		0.199356,
		0.223347,
		0.248386,
		0.2744,
		0.301315,
		0.32905,
		0.357518,
		0.38663,
		0.416289,
		0.446396,
		0.476846,
		0.507526,
		0.538323,
		0.569117,
		0.599783,
		0.630193,
		0.660215,
		0.689715,
		0.718555,
		0.746597,
		0.773702,
		0.799732,
		0.824552,
		0.848031,
		0.870045,
		0.89048,
		0.909232,
		0.926215,
		0.941362,
		0.954629,
		0.965998,
		0.975486,
		0.983139,
		0.989042,
		0.99331,
		0.996096,
		0.997575,
		0.997934,
		0.997356,
		0.995982,
		0.993865,
		0.990906,
		0.986797,
		0.981036,
		0.97297,
		0.961852,
		0.946897,
		0.92735,
		0.902538,
		0.871933,
		0.835202,
		0.792248,
		0.743241,
		0.68864,
		0.629185,
		0.565889,
		0.5,
		0.43295,
		0.366296,
		0.30164,
		0.240551,
		0.184485,
		0.134701,
		0.092199,
		0.057659,
		0.031406,
		0.013387,
		0.003178,
		0.000798,
		0.003183,
		0.007142,
		0.012657,
		0.01971,
		0.02828,
		0.038342,
		0.049871,
		0.062836,
		0.077206,
		0.092945,
		0.110014,
		0.128373,
		0.147975,
		0.168771,
		0.19071,
		0.213733,
		0.237781,
		0.26279,
		0.28869,
		0.315408,
		0.342867,
		0.370984,
		0.399673,
		0.428844,
		0.4584,
		0.488241,
		0.518262,
		0.548355,
		0.578407,
		0.608301,
		0.637915,
		0.667127,
		0.69581,
		0.723836,
		0.751077,
		0.777401,
		0.802682,
		0.826794,
		0.849616,
		0.871032,
		0.890936,
		0.909232,
		0.925841,
		0.940699,
		0.953766,
		0.965026,
		0.974494,
		0.982219,
		0.988288,
		0.992826,
		0.995994,
		0.997988,
		0.999026,
		0.99933,
		0.9991,
		0.998474,
		0.997473,
		0.99593,
		0.993485,
		0.989607,
		0.983637,
		0.974831,
		0.96241,
		0.945607,
		0.923719,
		0.896157,
		0.862495,
		0.822512,
		0.77623,
		0.723937,
		0.666205,
		0.603886,
		0.538097,
		0.470186,
		0.401677,
		0.334211,
		0.269468,
		0.209081,
		0.154555,
		0.107181,
		0.067966,
		0.037571,
		0.01627,
		0.003928,
		0.000762,
		0.003041,
		0.006822,
		0.012092,
		0.018832,
		0.027023,
		0.036642,
		0.047667,
		0.060068,
		0.073818,
		0.088884,
		0.105229,
		0.122817,
		0.141606,
		0.161551,
		0.182603,
		0.204713,
		0.227824,
		0.251877,
		0.276811,
		0.302559,
		0.32905,
		0.356209,
		0.383959,
		0.412216,
		0.440894,
		0.469901,
		0.499142,
		0.528519,
		0.557927,
		0.587261,
		0.61641,
		0.645261,
		0.673697,
		0.701601,
		0.728851,
		0.755329,
		0.780913,
		0.805484,
		0.828924,
		0.851122,
		0.87197,
		0.89137,
		0.909232,
		0.925483,
		0.940063,
		0.952932,
		0.964076,
		0.973507,
		0.981267,
		0.987433,
		0.992122,
		0.995487,
		0.997716,
		0.999028,
		0.999659,
		0.999837,
		0.999756,
		0.999523,
		0.9991,
		0.998268,
		0.996642,
		0.993699,
		0.988808,
		0.981267,
		0.970337,
		0.955287,
		0.935426,
		0.910151,
		0.878988,
		0.841631,
		0.797981,
		0.748179,
		0.692627,
		0.632006,
		0.567267,
		0.499619,
		0.430498,
		0.361507,
		0.294361,
		0.230803,
		0.172521,
		0.121059,
		0.077731,
		0.043541,
		0.019122,
		0.004685,
		0.000729,
		0.002907,
		0.006524,
		0.011564,
		0.018011,
		0.025848,
		0.035053,
		0.045605,
		0.057479,
		0.070648,
		0.085081,
		0.100747,
		0.117611,
		0.135635,
		0.154777,
		0.174994,
		0.19624,
		0.218463,
		0.24161,
		0.265625,
		0.290446,
		0.31601,
		0.342249,
		0.369091,
		0.39646,
		0.424278,
		0.452462,
		0.480925,
		0.509577,
		0.538323,
		0.567067,
		0.595706,
		0.624138,
		0.652255,
		0.679948,
		0.707107,
		0.733619,
		0.759371,
		0.78425,
		0.808146,
		0.830949,
		0.852555,
		0.872864,
		0.891784,
		0.909232,
		0.92514,
		0.939451,
		0.952128,
		0.963156,
		0.972543,
		0.980326,
		0.986574,
		0.99139,
		0.994912,
		0.997312,
		0.998793,
		0.99958,
		0.999904,
		0.999976,
		0.999952,
		0.99988,
		0.999655,
		0.999017,
		0.997575,
		0.994828,
		0.990193,
		0.983029,
		0.972665,
		0.95843,
		0.939685,
		0.915851,
		0.886448,
		0.851131,
		0.809723,
		0.76225,
		0.708972,
		0.650402,
		0.58732,
		0.520775,
		0.45206,
		0.382687,
		0.314334,
		0.24878,
		0.187825,
		0.133206,
		0.086499,
		0.049033,
		0.021807,
		0.005415,
		0.000697,
		0.002783,
		0.006244,
		0.011069,
		0.017243,
		0.024747,
		0.033565,
		0.043674,
		0.055053,
		0.067676,
		0.081516,
		0.096544,
		0.112726,
		0.130029,
		0.148414,
		0.167843,
		0.188271,
		0.209654,
		0.231941,
		0.255081,
		0.279019,
		0.303696,
		0.32905,
		0.355015,
		0.381525,
		0.408505,
		0.43588,
		0.463572,
		0.491497,
		0.51957,
		0.547701,
		0.575799,
		0.603767,
		0.631507,
		0.65892,
		0.685901,
		0.712348,
		0.738155,
		0.763216,
		0.787425,
		0.81068,
		0.832877,
		0.85392,
		0.873716,
		0.892179,
		0.909232,
		0.924811,
		0.938862,
		0.951351,
		0.962262,
		0.971601,
		0.9794,
		0.985719,
		0.99065,
		0.994317,
		0.996876,
		0.998511,
		0.999429,
		0.99985,
		0.999982,
		0.999998,
		0.999993,
		0.999936,
		0.999659,
		0.998871,
		0.997181,
		0.994116,
		0.989142,
		0.981687,
		0.971156,
		0.956957,
		0.938519,
		0.91532,
		0.88691,
		0.852944,
		0.81321,
		0.767663,
		0.716456,
		0.659966,
		0.598814,
		0.533882,
		0.466305,
		0.397462,
		0.328938,
		0.26248,
		0.199929,
		0.143141,
		0.093895,
		0.053801,
		0.024201,
		0.006083,
		0.000668,
		0.002666,
		0.005983,
		0.010606,
		0.016522,
		0.023716,
		0.032169,
		0.041863,
		0.052777,
		0.064888,
		0.07817,
		0.092596,
		0.108136,
		0.12476,
		0.142431,
		0.161115,
		0.18077,
		0.201355,
		0.222825,
		0.245133,
		0.268228,
		0.292057,
		0.316562,
		0.341684,
		0.367361,
		0.393526,
		0.42011,
		0.447041,
		0.474244,
		0.501639,
		0.529145,
		0.556677,
		0.584147,
		0.611467,
		0.638541,
		0.665277,
		0.691577,
		0.717343,
		0.742476,
		0.766878,
		0.790449,
		0.813093,
		0.834714,
		0.855222,
		0.874529,
		0.892557,
		0.909232,
		0.924495,
		0.938296,
		0.950601,
		0.961395,
		0.970681,
		0.978488,
		0.984868,
		0.989904,
		0.993706,
		0.996415,
		0.9982,
		0.999251,
		0.999771,
		0.999962,
		0.999999,
		1,
		0.999982,
		0.999839,
		0.999349,
		0.998198,
		0.995994,
		0.99229,
		0.986599,
		0.97841,
		0.967203,
		0.952465,
		0.933706,
		0.910471,
		0.882368,
		0.849081,
		0.810403,
		0.766258,
		0.716734,
		0.662108,
		0.602874,
		0.539764,
		0.473757,
		0.406079,
		0.338191,
		0.271753,
		0.208583,
		0.150584,
		0.099666,
		0.057659,
		0.026203,
		0.006658,
		0.000641,
		0.002556,
		0.005737,
		0.010171,
		0.015846,
		0.022747,
		0.030859,
		0.040162,
		0.050639,
		0.062267,
		0.075024,
		0.088884,
		0.103819,
		0.119801,
		0.136799,
		0.154777,
		0.1737,
		0.193529,
		0.214224,
		0.23574,
		0.258031,
		0.281048,
		0.304739,
		0.32905,
		0.353922,
		0.379296,
		0.405109,
		0.431293,
		0.457781,
		0.484498,
		0.511372,
		0.538323,
		0.565272,
		0.592135,
		0.618827,
		0.645261,
		0.671346,
		0.696992,
		0.722107,
		0.746597,
		0.770369,
		0.793332,
		0.815394,
		0.836467,
		0.856465,
		0.875306,
		0.892918,
		0.909232,
		0.924192,
		0.937751,
		0.949876,
		0.960553,
		0.969784,
		0.977591,
		0.984024,
		0.989154,
		0.993081,
		0.995933,
		0.997863,
		0.999046,
		0.999671,
		0.999929,
		0.999995,
		1,
		0.999995,
		0.99992,
		0.999606,
		0.998793,
		0.997146,
		0.994276,
		0.989757,
		0.983139,
		0.973965,
		0.961779,
		0.94614,
		0.926633,
		0.902878,
		0.874547,
		0.841376,
		0.803188,
		0.75991,
		0.711604,
		0.658487,
		0.600964,
		0.539646,
		0.475373,
		0.409225,
		0.342523,
		0.276811,
		0.213831,
		0.155471,
		0.103703,
		0.060501,
		0.027745,
		0.007118,
		0.000615,
		0.002453,
		0.005506,
		0.009762,
		0.015211,
		0.021837,
		0.029626,
		0.038563,
		0.048628,
		0.059802,
		0.072063,
		0.085389,
		0.099754,
		0.11513,
		0.13149,
		0.148801,
		0.16703,
		0.186142,
		0.2061,
		0.226862,
		0.248386,
		0.270627,
		0.293538,
		0.317068,
		0.341166,
		0.365774,
		0.390835,
		0.416289,
		0.442072,
		0.468117,
		0.494356,
		0.520718,
		0.547127,
		0.573509,
		0.599783,
		0.625869,
		0.651685,
		0.677145,
		0.702164,
		0.726654,
		0.750529,
		0.773702,
		0.796084,
		0.817592,
		0.838141,
		0.857652,
		0.87605,
		0.893264,
		0.909232,
		0.9239,
		0.937225,
		0.949176,
		0.959736,
		0.968908,
		0.976711,
		0.983187,
		0.988401,
		0.992444,
		0.995431,
		0.997501,
		0.998815,
		0.999548,
		0.999882,
		0.999986,
		1,
		0.999999,
		0.999964,
		0.999771,
		0.999205,
		0.997978,
		0.995745,
		0.992122,
		0.9867,
		0.979061,
		0.968789,
		0.955481,
		0.938756,
		0.918264,
		0.893693,
		0.864779,
		0.831316,
		0.793172,
		0.750302,
		0.702768,
		0.650766,
		0.594644,
		0.534931,
		0.472358,
		0.407876,
		0.34267,
		0.278157,
		0.215973,
		0.157944,
		0.106041,
		0.062312,
		0.028801,
		0.007452,
		0.00059,
		0.002356,
		0.005289,
		0.009378,
		0.014613,
		0.02098,
		0.028466,
		0.037057,
		0.046734,
		0.057479,
		0.069274,
		0.082095,
		0.09592,
		0.110724,
		0.12648,
		0.14316,
		0.160731,
		0.179163,
		0.198419,
		0.218463,
		0.239256,
		0.260756,
		0.282919,
		0.3057,
		0.32905,
		0.352917,
		0.37725,
		0.40199,
		0.427082,
		0.452462,
		0.478069,
		0.503836,
		0.529695,
		0.555576,
		0.581406,
		0.607109,
		0.632611,
		0.657831,
		0.68269,
		0.707107,
		0.731,
		0.754286,
		0.776885,
		0.798713,
		0.819691,
		0.839741,
		0.858788,
		0.876762,
		0.893596,
		0.909232,
		0.92362,
		0.936718,
		0.948498,
		0.958943,
		0.968053,
		0.975845,
		0.982358,
		0.987648,
		0.991798,
		0.994912,
		0.997117,
		0.99856,
		0.999403,
		0.999819,
		0.99997,
		0.999999,
		1,
		0.999986,
		0.999875,
		0.999493,
		0.998595,
		0.996876,
		0.993987,
		0.989554,
		0.983187,
		0.974494,
		0.963094,
		0.948626,
		0.930757,
		0.909189,
		0.883665,
		0.853978,
		0.819979,
		0.781582,
		0.73878,
		0.691658,
		0.640413,
		0.585368,
		0.527004,
		0.465973,
		0.403127,
		0.339533,
		0.276483,
		0.215498,
		0.158311,
		0.106843,
		0.063155,
		0.029384,
		0.007659,
		0.000567,
		0.002265,
		0.005084,
		0.009016,
		0.014049,
		0.020173,
		0.027373,
		0.035637,
		0.044948,
		0.055289,
		0.066642,
		0.078987,
		0.092302,
		0.106564,
		0.121749,
		0.137829,
		0.154777,
		0.172561,
		0.191151,
		0.210511,
		0.230606,
		0.251397,
		0.272844,
		0.294906,
		0.317536,
		0.340688,
		0.364313,
		0.38836,
		0.412775,
		0.437501,
		0.462481,
		0.487653,
		0.512956,
		0.538323,
		0.563688,
		0.588982,
		0.614133,
		0.639069,
		0.663715,
		0.687996,
		0.711836,
		0.735156,
		0.757879,
		0.779928,
		0.801227,
		0.821699,
		0.841272,
		0.859876,
		0.877444,
		0.893914,
		0.909232,
		0.92335,
		0.936229,
		0.947842,
		0.958172,
		0.967218,
		0.974996,
		0.981538,
		0.986896,
		0.991144,
		0.994377,
		0.996711,
		0.99828,
		0.999236,
		0.999738,
		0.999944,
		0.999996,
		1,
		0.999996,
		0.999937,
		0.999689,
		0.999046,
		0.99774,
		0.995458,
		0.991857,
		0.986574,
		0.97924,
		0.969489,
		0.956972,
		0.941362,
		0.922363,
		0.89972,
		0.873218,
		0.842698,
		0.808054,
		0.769247,
		0.726308,
		0.679351,
		0.628589,
		0.574344,
		0.517072,
		0.457378,
		0.396042,
		0.334039,
		0.272552,
		0.212988,
		0.156977,
		0.10636,
		0.06316,
		0.029542,
		0.007746,
		0.000546,
		0.002179,
		0.004891,
		0.008674,
		0.013517,
		0.019411,
		0.026342,
		0.034297,
		0.043262,
		0.053221,
		0.064157,
		0.076051,
		0.088884,
		0.102633,
		0.117276,
		0.132788,
		0.149143,
		0.166312,
		0.184267,
		0.202976,
		0.222405,
		0.242518,
		0.26328,
		0.28465,
		0.306588,
		0.32905,
		0.35199,
		0.375363,
		0.399117,
		0.423201,
		0.447562,
		0.472143,
		0.496887,
		0.521733,
		0.546619,
		0.571482,
		0.596255,
		0.620871,
		0.645261,
		0.669354,
		0.693079,
		0.716363,
		0.739133,
		0.761317,
		0.782841,
		0.803632,
		0.823621,
		0.842738,
		0.860918,
		0.878098,
		0.89422,
		0.909232,
		0.92309,
		0.935757,
		0.947206,
		0.957423,
		0.966404,
		0.974162,
		0.980727,
		0.986146,
		0.990485,
		0.993829,
		0.996286,
		0.997978,
		0.999046,
		0.999638,
		0.999906,
		0.999989,
		1,
		0.999999,
		0.999972,
		0.999819,
		0.999369,
		0.998392,
		0.996609,
		0.993706,
		0.989348,
		0.983187,
		0.974876,
		0.964076,
		0.950468,
		0.933754,
		0.913673,
		0.890001,
		0.862561,
		0.831226,
		0.795928,
		0.756663,
		0.713495,
		0.666569,
		0.616116,
		0.56247,
		0.50608,
		0.447525,
		0.38754,
		0.327031,
		0.267095,
		0.209038,
		0.154387,
		0.104887,
		0.062493,
		0.029348,
		0.007731,
		0.000525,
		0.002098,
		0.004709,
		0.008351,
		0.013016,
		0.018691,
		0.025367,
		0.033031,
		0.041669,
		0.051267,
		0.061808,
		0.073275,
		0.08565,
		0.098913,
		0.113042,
		0.128015,
		0.143807,
		0.160392,
		0.177742,
		0.19583,
		0.214622,
		0.234088,
		0.254193,
		0.2749,
		0.296172,
		0.317968,
		0.340247,
		0.362964,
		0.386075,
		0.409531,
		0.433283,
		0.457278,
		0.481465,
		0.505785,
		0.530184,
		0.5546,
		0.578973,
		0.603241,
		0.627338,
		0.651201,
		0.67476,
		0.69795,
		0.720701,
		0.742944,
		0.76461,
		0.78563,
		0.805937,
		0.825463,
		0.844144,
		0.861918,
		0.878726,
		0.894513,
		0.909232,
		0.92284,
		0.935301,
		0.946591,
		0.956695,
		0.965609,
		0.973345,
		0.979927,
		0.985399,
		0.989821,
		0.99327,
		0.995843,
		0.997655,
		0.998834,
		0.999519,
		0.999854,
		0.999976,
		0.999999,
		1,
		0.999989,
		0.9999,
		0.999596,
		0.998879,
		0.997501,
		0.99518,
		0.991607,
		0.986456,
		0.9794,
		0.970112,
		0.95828,
		0.943612,
		0.925841,
		0.904734,
		0.880097,
		0.851782,
		0.819691,
		0.783783,
		0.74408,
		0.700674,
		0.653732,
		0.603506,
		0.550341,
		0.494688,
		0.437117,
		0.378334,
		0.319197,
		0.260738,
		0.204182,
		0.15096,
		0.102721,
		0.061335,
		0.028884,
		0.007635,
		0.000506,
		0.002021,
		0.004537,
		0.008046,
		0.012541,
		0.018011,
		0.024446,
		0.031834,
		0.040162,
		0.049418,
		0.059585,
		0.070648,
		0.082589,
		0.095391,
		0.109032,
		0.123492,
		0.138748,
		0.154777,
		0.171552,
		0.189047,
		0.207232,
		0.226079,
		0.245554,
		0.265625,
		0.286256,
		0.307411,
		0.32905,
		0.351133,
		0.373618,
		0.39646,
		0.419614,
		0.443032,
		0.466665,
		0.49046,
		0.514364,
		0.538323,
		0.56228,
		0.586177,
		0.609954,
		0.63355,
		0.656903,
		0.679948,
		0.702622,
		0.72486,
		0.746597,
		0.767766,
		0.788304,
		0.808146,
		0.827229,
		0.845492,
		0.862877,
		0.879329,
		0.894796,
		0.909232,
		0.922598,
		0.93486,
		0.945995,
		0.955987,
		0.964833,
		0.972543,
		0.979138,
		0.984657,
		0.989154,
		0.9927,
		0.995385,
		0.997312,
		0.9986,
		0.999381,
		0.999788,
		0.999955,
		0.999997,
		1,
		0.999997,
		0.99995,
		0.999752,
		0.999236,
		0.998186,
		0.996348,
		0.993437,
		0.989154,
		0.983187,
		0.975226,
		0.964969,
		0.952128,
		0.936437,
		0.917657,
		0.895582,
		0.870045,
		0.840925,
		0.808146,
		0.771692,
		0.731605,
		0.687995,
		0.641049,
		0.591031,
		0.538297,
		0.483303,
		0.426612,
		0.368915,
		0.311037,
		0.253963,
		0.198852,
		0.147054,
		0.100129,
		0.059855,
		0.028234,
		0.00748,
		0.000488,
		0.001948,
		0.004374,
		0.007758,
		0.012092,
		0.017367,
		0.023574,
		0.0307,
		0.038736,
		0.047667,
		0.057479,
		0.068159,
		0.079688,
		0.092052,
		0.105229,
		0.119202,
		0.133949,
		0.149448,
		0.165674,
		0.182603,
		0.200209,
		0.218463,
		0.237335,
		0.256796,
		0.276811,
		0.297347,
		0.318369,
		0.339837,
		0.361715,
		0.383959,
		0.406528,
		0.429378,
		0.452462,
		0.475734,
		0.499142,
		0.522638,
		0.546167,
		0.569676,
		0.593109,
		0.61641,
		0.63952,
		0.66238,
		0.684929,
		0.707107,
		0.728851,
		0.750101,
		0.770794,
		0.790869,
		0.810266,
		0.828924,
		0.846787,
		0.863799,
		0.879909,
		0.895068,
		0.909232,
		0.922365,
		0.934434,
		0.945417,
		0.955299,
		0.964076,
		0.971756,
		0.978359,
		0.983919,
		0.988485,
		0.992122,
		0.994912,
		0.99695,
		0.998347,
		0.999223,
		0.999707,
		0.999924,
		0.999991,
		1,
		0.999999,
		0.999977,
		0.999854,
		0.999493,
		0.998706,
		0.997265,
		0.994912,
		0.991369,
		0.986346,
		0.979547,
		0.970681,
		0.959469,
		0.945646,
		0.928972,
		0.909232,
		0.886248,
		0.859876,
		0.830017,
		0.796619,
		0.759685,
		0.719275,
		0.675515,
		0.628601,
		0.578809,
		0.526502,
		0.472135,
		0.416269,
		0.359581,
		0.302876,
		0.247099,
		0.19336,
		0.142941,
		0.097323,
		0.058196,
		0.027471,
		0.007286,
		0.000471,
		0.001879,
		0.004219,
		0.007484,
		0.011667,
		0.016757,
		0.022747,
		0.029626,
		0.037384,
		0.046007,
		0.055483,
		0.065798,
		0.076937,
		0.088884,
		0.101621,
		0.11513,
		0.129392,
		0.144385,
		0.160089,
		0.176478,
		0.193529,
		0.211216,
		0.229511,
		0.248386,
		0.267809,
		0.28775,
		0.308175,
		0.32905,
		0.350337,
		0.371999,
		0.393997,
		0.416289,
		0.438833,
		0.461585,
		0.484498,
		0.507526,
		0.53062,
		0.553728,
		0.5768,
		0.599783,
		0.622621,
		0.645261,
		0.667644,
		0.689715,
		0.711414,
		0.732684,
		0.753466,
		0.773702,
		0.793332,
		0.812301,
		0.830552,
		0.848031,
		0.864686,
		0.880467,
		0.895329,
		0.909232,
		0.922139,
		0.934022,
		0.944857,
		0.95463,
		0.963338,
		0.970985,
		0.977591,
		0.983187,
		0.987816,
		0.991538,
		0.994427,
		0.996571,
		0.998073,
		0.999046,
		0.999608,
		0.999882,
		0.999981,
		0.999999,
		1,
		0.999991,
		0.99992,
		0.999675,
		0.999094,
		0.997978,
		0.996092,
		0.993181,
		0.988971,
		0.983187,
		0.975548,
		0.965784,
		0.953636,
		0.938862,
		0.921246,
		0.900598,
		0.876762,
		0.849616,
		0.819083,
		0.785132,
		0.747781,
		0.707107,
		0.663248,
		0.61641,
		0.566876,
		0.515011,
		0.461269,
		0.406206,
		0.350485,
		0.294891,
		0.240343,
		0.187904,
		0.138805,
		0.094456,
		0.056465,
		0.026654,
		0.007072,
		0.000454,
		0.001814,
		0.004073,
		0.007225,
		0.011263,
		0.016179,
		0.021964,
		0.028608,
		0.036101,
		0.044432,
		0.053588,
		0.063557,
		0.074325,
		0.085875,
		0.098193,
		0.111261,
		0.125061,
		0.139573,
		0.154777,
		0.170651,
		0.187172,
		0.204316,
		0.222058,
		0.24037,
		0.259225,
		0.278592,
		0.298442,
		0.318742,
		0.339457,
		0.360554,
		0.381994,
		0.403741,
		0.425753,
		0.447991,
		0.470412,
		0.492972,
		0.515624,
		0.538323,
		0.561021,
		0.583667,
		0.606211,
		0.628601,
		0.650784,
		0.672708,
		0.694316,
		0.715554,
		0.736367,
		0.756699,
		0.776495,
		0.795699,
		0.814257,
		0.832117,
		0.849227,
		0.865538,
		0.881004,
		0.895582,
		0.909232,
		0.921922,
		0.933622,
		0.944313,
		0.953979,
		0.962616,
		0.97023,
		0.976835,
		0.982461,
		0.987147,
		0.990947,
		0.99393,
		0.996177,
		0.997782,
		0.99885,
		0.999493,
		0.999827,
		0.999963,
		0.999998,
		1,
		0.999997,
		0.999959,
		0.999799,
		0.999381,
		0.998528,
		0.99703,
		0.994653,
		0.991144,
		0.986243,
		0.979683,
		0.971204,
		0.960553,
		0.947493,
		0.931803,
		0.913289,
		0.891784,
		0.867151,
		0.839291,
		0.808146,
		0.773702,
		0.735993,
		0.695109,
		0.651201,
		0.604481,
		0.555238,
		0.503836,
		0.450728,
		0.39646,
		0.341684,
		0.287164,
		0.233791,
		0.182595,
		0.134756,
		0.091624,
		0.054735,
		0.025825,
		0.006851,
		0.000439,
		0.001752,
		0.003934,
		0.006979,
		0.01088,
		0.01563,
		0.02122,
		0.027641,
		0.034883,
		0.042936,
		0.051789,
		0.061429,
		0.071843,
		0.083016,
		0.094935,
		0.107582,
		0.120941,
		0.134994,
		0.149721,
		0.165103,
		0.181117,
		0.197741,
		0.214952,
		0.232725,
		0.251033,
		0.269849,
		0.289144,
		0.308888,
		0.32905,
		0.349596,
		0.370494,
		0.391707,
		0.413199,
		0.43493,
		0.456863,
		0.478955,
		0.501165,
		0.523449,
		0.545761,
		0.568056,
		0.590287,
		0.612405,
		0.63436,
		0.656102,
		0.67758,
		0.698742,
		0.719536,
		0.739908,
		0.759807,
		0.77918,
		0.797974,
		0.816138,
		0.833622,
		0.850378,
		0.866359,
		0.881522,
		0.895825,
		0.909232,
		0.921711,
		0.933236,
		0.943785,
		0.953345,
		0.961912,
		0.969489,
		0.976091,
		0.981742,
		0.986479,
		0.990352,
		0.993424,
		0.995768,
		0.997473,
		0.998635,
		0.999361,
		0.99976,
		0.999938,
		0.999993,
		1,
		1,
		0.999982,
		0.999882,
		0.999588,
		0.998946,
		0.997769,
		0.995843,
		0.992935,
		0.9888,
		0.983187,
		0.975845,
		0.966531,
		0.95501,
		0.941064,
		0.924495,
		0.905127,
		0.882814,
		0.85744,
		0.828924,
		0.797225,
		0.762345,
		0.724333,
		0.683291,
		0.639377,
		0.592813,
		0.54389,
		0.492972,
		0.440506,
		0.387032,
		0.333186,
		0.279714,
		0.227479,
		0.177477,
		0.130845,
		0.088879,
		0.053048,
		0.02501,
		0.00663,
		0.000424,
		0.001693,
		0.003803,
		0.006746,
		0.010517,
		0.015108,
		0.020513,
		0.026722,
		0.033726,
		0.041515,
		0.050078,
		0.059405,
		0.069482,
		0.080297,
		0.091835,
		0.104082,
		0.11702,
		0.130634,
		0.144906,
		0.159817,
		0.175346,
		0.191473,
		0.208175,
		0.225429,
		0.243212,
		0.261496,
		0.280256,
		0.299464,
		0.319089,
		0.339103,
		0.359472,
		0.380165,
		0.401146,
		0.42238,
		0.44383,
		0.465459,
		0.487226,
		0.50909,
		0.531011,
		0.552945,
		0.574848,
		0.596674,
		0.618377,
		0.63991,
		0.661224,
		0.682272,
		0.703003,
		0.723368,
		0.743316,
		0.762798,
		0.781763,
		0.800163,
		0.817948,
		0.835071,
		0.851487,
		0.867151,
		0.882021,
		0.89606,
		0.909232,
		0.921508,
		0.932861,
		0.943272,
		0.952728,
		0.961225,
		0.968764,
		0.975358,
		0.98103,
		0.985814,
		0.989754,
		0.992908,
		0.995346,
		0.997147,
		0.998403,
		0.999212,
		0.999678,
		0.999903,
		0.999984,
		1,
		1,
		0.999993,
		0.999935,
		0.999735,
		0.999261,
		0.998347,
		0.996799,
		0.994403,
		0.990931,
		0.986146,
		0.97981,
		0.971686,
		0.961546,
		0.949176,
		0.934374,
		0.916963,
		0.896786,
		0.873716,
		0.847654,
		0.818536,
		0.786339,
		0.751077,
		0.712812,
		0.671658,
		0.62778,
		0.581406,
		0.532828,
		0.48241,
		0.430596,
		0.377913,
		0.324985,
		0.272538,
		0.221409,
		0.172561,
		0.127091,
		0.086243,
		0.051425,
		0.024222,
		0.006416,
		0.00041,
		0.001638,
		0.003677,
		0.006524,
		0.010171,
		0.014613,
		0.019841,
		0.025848,
		0.032625,
		0.040162,
		0.048451,
		0.057479,
		0.067236,
		0.077708,
		0.088884,
		0.100747,
		0.113285,
		0.12648,
		0.140317,
		0.154777,
		0.169842,
		0.185491,
		0.201706,
		0.218463,
		0.23574,
		0.253513,
		0.271757,
		0.290446,
		0.309553,
		0.32905,
		0.348906,
		0.369091,
		0.389573,
		0.410318,
		0.431293,
		0.452462,
		0.473788,
		0.495233,
		0.516759,
		0.538323,
		0.559887,
		0.581406,
		0.602837,
		0.624138,
		0.645261,
		0.666161,
		0.686792,
		0.707107,
		0.727057,
		0.746597,
		0.765677,
		0.78425,
		0.80227,
		0.819691,
		0.836467,
		0.852555,
		0.867913,
		0.882503,
		0.896287,
		0.909232,
		0.921311,
		0.932498,
		0.942774,
		0.952128,
		0.960553,
		0.968053,
		0.974637,
		0.980326,
		0.985151,
		0.989154,
		0.992386,
		0.994912,
		0.996806,
		0.998153,
		0.999046,
		0.999581,
		0.999858,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999967,
		0.999836,
		0.999493,
		0.998793,
		0.997561,
		0.995601,
		0.9927,
		0.988638,
		0.983187,
		0.97612,
		0.967218,
		0.956268,
		0.943071,
		0.927446,
		0.909232,
		0.888291,
		0.864511,
		0.837812,
		0.808146,
		0.775502,
		0.739908,
		0.701439,
		0.660215,
		0.61641,
		0.570256,
		0.522046,
		0.472143,
		0.420985,
		0.369091,
		0.317068,
		0.265625,
		0.215573,
		0.167843,
		0.123492,
		0.083718,
		0.04987,
		0.023468,
		0.006211,
		0.000397,
		0.001584,
		0.003558,
		0.006313,
		0.009842,
		0.014141,
		0.019201,
		0.025016,
		0.031577,
		0.038875,
		0.046901,
		0.055645,
		0.065096,
		0.075242,
		0.086071,
		0.09757,
		0.109724,
		0.122519,
		0.135939,
		0.149968,
		0.164588,
		0.179781,
		0.195527,
		0.211807,
		0.228598,
		0.245878,
		0.263625,
		0.281814,
		0.300419,
		0.319414,
		0.338772,
		0.358463,
		0.378457,
		0.398725,
		0.419232,
		0.439948,
		0.460836,
		0.481862,
		0.502989,
		0.52418,
		0.545395,
		0.566596,
		0.587741,
		0.608789,
		0.629697,
		0.650423,
		0.670922,
		0.69115,
		0.711062,
		0.730613,
		0.749757,
		0.76845,
		0.786646,
		0.804301,
		0.82137,
		0.837812,
		0.853585,
		0.868649,
		0.882967,
		0.896506,
		0.909232,
		0.92112,
		0.932145,
		0.94229,
		0.951543,
		0.959898,
		0.967356,
		0.973927,
		0.97963,
		0.984492,
		0.988552,
		0.991857,
		0.994468,
		0.996452,
		0.997888,
		0.998863,
		0.99947,
		0.999801,
		0.999949,
		0.999994,
		1,
		1,
		0.999985,
		0.999903,
		0.999662,
		0.999134,
		0.998163,
		0.996571,
		0.994162,
		0.990728,
		0.986055,
		0.979927,
		0.97213,
		0.962458,
		0.950715,
		0.936718,
		0.920303,
		0.901325,
		0.879663,
		0.855222,
		0.827935,
		0.797771,
		0.764729,
		0.728851,
		0.690221,
		0.648967,
		0.605268,
		0.559361,
		0.511539,
		0.462163,
		0.411665,
		0.360554,
		0.309424,
		0.258963,
		0.209959,
		0.163312,
		0.120042,
		0.081301,
		0.048385,
		0.022747,
		0.006014,
		0.000384,
		0.001534,
		0.003444,
		0.006111,
		0.009529,
		0.013691,
		0.018592,
		0.024223,
		0.030578,
		0.037648,
		0.045425,
		0.053897,
		0.063057,
		0.072891,
		0.083389,
		0.094539,
		0.106327,
		0.118739,
		0.131761,
		0.145377,
		0.159571,
		0.174326,
		0.189623,
		0.205443,
		0.221767,
		0.238574,
		0.255841,
		0.273547,
		0.291667,
		0.310176,
		0.32905,
		0.34826,
		0.367779,
		0.387578,
		0.407628,
		0.427896,
		0.448352,
		0.468961,
		0.48969,
		0.510503,
		0.531365,
		0.552238,
		0.573084,
		0.593863,
		0.614537,
		0.635065,
		0.655405,
		0.675515,
		0.695352,
		0.714875,
		0.73404,
		0.752804,
		0.771123,
		0.788955,
		0.806258,
		0.82299,
		0.839109,
		0.854579,
		0.869359,
		0.883417,
		0.896717,
		0.909232,
		0.920935,
		0.931803,
		0.94182,
		0.950973,
		0.959257,
		0.966673,
		0.973229,
		0.978942,
		0.983837,
		0.987949,
		0.991323,
		0.994013,
		0.996084,
		0.997607,
		0.998665,
		0.999343,
		0.999732,
		0.99992,
		0.999987,
		1,
		1,
		0.999994,
		0.999946,
		0.999782,
		0.999391,
		0.998635,
		0.997353,
		0.995364,
		0.992476,
		0.988485,
		0.983187,
		0.976376,
		0.967852,
		0.957423,
		0.944907,
		0.930138,
		0.912966,
		0.893264,
		0.870925,
		0.845868,
		0.818041,
		0.787425,
		0.754033,
		0.717915,
		0.679164,
		0.637915,
		0.594354,
		0.548718,
		0.501302,
		0.452462,
		0.402626,
		0.352291,
		0.30204,
		0.25254,
		0.204557,
		0.15896,
		0.116734,
		0.078987,
		0.046964,
		0.022059,
		0.005827,
		0.000372,
		0.001486,
		0.003336,
		0.00592,
		0.00923,
		0.013263,
		0.018011,
		0.023468,
		0.029626,
		0.036479,
		0.044016,
		0.05223,
		0.061111,
		0.070648,
		0.08083,
		0.091646,
		0.103084,
		0.11513,
		0.127771,
		0.140991,
		0.154777,
		0.169111,
		0.183976,
		0.199356,
		0.21523,
		0.23158,
		0.248386,
		0.265625,
		0.283276,
		0.301315,
		0.319719,
		0.338462,
		0.357518,
		0.37686,
		0.39646,
		0.416289,
		0.436317,
		0.456513,
		0.476846,
		0.497281,
		0.517785,
		0.538323,
		0.55886,
		0.579359,
		0.599783,
		0.620093,
		0.64025,
		0.660215,
		0.679948,
		0.699408,
		0.718555,
		0.737347,
		0.755743,
		0.773702,
		0.791183,
		0.808146,
		0.824552,
		0.840361,
		0.855538,
		0.870045,
		0.883851,
		0.896922,
		0.909232,
		0.920755,
		0.931471,
		0.941362,
		0.950418,
		0.958631,
		0.966004,
		0.972543,
		0.978262,
		0.983187,
		0.987347,
		0.990785,
		0.993551,
		0.995704,
		0.997312,
		0.998451,
		0.999202,
		0.999651,
		0.999882,
		0.999975,
		0.999998,
		1,
		0.999998,
		0.999973,
		0.999864,
		0.999581,
		0.999001,
		0.997978,
		0.996348,
		0.99393,
		0.990536,
		0.98597,
		0.980037,
		0.972543,
		0.963299,
		0.952128,
		0.938862,
		0.92335,
		0.905457,
		0.88507,
		0.862095,
		0.836467,
		0.808146,
		0.777124,
		0.743424,
		0.707107,
		0.668272,
		0.627063,
		0.583667,
		0.538323,
		0.491327,
		0.443032,
		0.393858,
		0.344293,
		0.294906,
		0.246346,
		0.199356,
		0.154777,
		0.113559,
		0.07677,
		0.045605,
		0.021402,
		0.005649,
		0.000361,
		0.00144,
		0.003233,
		0.005737,
		0.008946,
		0.012854,
		0.017457,
		0.022747,
		0.028718,
		0.035363,
		0.042672,
		0.050639,
		0.059254,
		0.068506,
		0.078386,
		0.088884,
		0.099986,
		0.111681,
		0.123957,
		0.136799,
		0.150192,
		0.164123,
		0.178574,
		0.193529,
		0.208972,
		0.224882,
		0.241242,
		0.258031,
		0.275228,
		0.292812,
		0.310761,
		0.32905,
		0.347655,
		0.366551,
		0.385711,
		0.405109,
		0.424716,
		0.444504,
		0.464441,
		0.484498,
		0.504643,
		0.524843,
		0.545064,
		0.565272,
		0.585432,
		0.605508,
		0.625463,
		0.645261,
		0.664862,
		0.68423,
		0.703324,
		0.722107,
		0.740538,
		0.758579,
		0.77619,
		0.793332,
		0.809968,
		0.82606,
		0.84157,
		0.856465,
		0.870708,
		0.88427,
		0.897121,
		0.909232,
		0.920582,
		0.931149,
		0.940917,
		0.949876,
		0.95802,
		0.965348,
		0.971868,
		0.977591,
		0.982541,
		0.986746,
		0.990244,
		0.993081,
		0.995313,
		0.997003,
		0.998222,
		0.999046,
		0.999556,
		0.999834,
		0.999957,
		0.999995,
		1,
		1,
		0.999988,
		0.99992,
		0.99972,
		0.999281,
		0.998474,
		0.997147,
		0.995135,
		0.992261,
		0.98834,
		0.983187,
		0.976614,
		0.968439,
		0.958486,
		0.946591,
		0.9326,
		0.916375,
		0.897796,
		0.876762,
		0.853193,
		0.827036,
		0.798264,
		0.766878,
		0.732912,
		0.696434,
		0.657551,
		0.61641,
		0.573204,
		0.528173,
		0.48161,
		0.433865,
		0.385352,
		0.336549,
		0.288011,
		0.24037,
		0.194346,
		0.150754,
		0.110511,
		0.074644,
		0.044304,
		0.020774,
		0.005478,
		0.005562,
		0.008674,
		0.012464,
		0.016928,
		0.022059,
		0.027851,
		0.034297,
		0.041389,
		0.049119,
		0.057479,
		0.06646,
		0.076051,
		0.086243,
		0.097025,
		0.108384,
		0.120309,
		0.132788,
		0.145806,
		0.159349,
		0.173402,
		0.18795,
		0.202976,
		0.218463,
		0.234393,
		0.250747,
		0.267506,
		0.28465,
		0.302157,
		0.320005,
		0.338171,
		0.356632,
		0.375363,
		0.394338,
		0.413531,
		0.432915,
		0.452462,
		0.472143,
		0.491928,
		0.511786,
		0.531686,
		0.551596,
		0.571482,
		0.591311,
		0.611048,
		0.630658,
		0.650105,
		0.669354,
		0.688367,
		0.707107,
		0.725537,
		0.743619,
		0.761317,
		0.778592,
		0.795408,
		0.811728,
		0.827516,
		0.842738,
		0.85736,
		0.87135,
		0.884676,
		0.897313,
		0.909232,
		0.920413,
		0.930835,
		0.940484,
		0.949349,
		0.957423,
		0.964706,
		0.971204,
		0.976929,
		0.981901,
		0.986146,
		0.9897,
		0.992605,
		0.994912,
		0.996681,
		0.997978,
		0.998875,
		0.999448,
		0.999776,
		0.999933,
		0.999989,
		1,
		1,
		0.999995,
		0.999956,
		0.999819,
		0.999493,
		0.998863,
		0.997793,
		0.996128,
		0.993706,
		0.990352,
		0.98589,
		0.98014,
		0.972926,
		0.964076,
		0.953429,
		0.94083,
		0.926139,
		0.909232,
		0.890001,
		0.868358,
		0.844236,
		0.817592,
		0.788409,
		0.756699,
		0.722505,
		0.685901,
		0.647002,
		0.605958,
		0.562964,
		0.518262,
		0.472143,
		0.424954,
		0.377099,
		0.32905,
		0.281345,
		0.234603,
		0.18952,
		0.146884,
		0.107582,
		0.072605,
		0.043058,
		0.020173,
		0.005315,
		0.021402,
		0.027023,
		0.033279,
		0.040162,
		0.047667,
		0.055783,
		0.064503,
		0.073818,
		0.083718,
		0.094192,
		0.105229,
		0.116819,
		0.128949,
		0.141606,
		0.154777,
		0.168448,
		0.182603,
		0.197229,
		0.212308,
		0.227824,
		0.243759,
		0.260094,
		0.276811,
		0.29389,
		0.31131,
		0.32905,
		0.347086,
		0.365398,
		0.383959,
		0.402746,
		0.421733,
		0.440894,
		0.460201,
		0.479627,
		0.499142,
		0.518718,
		0.538323,
		0.557927,
		0.577498,
		0.597004,
		0.61641,
		0.635684,
		0.654791,
		0.673697,
		0.692366,
		0.710763,
		0.728851,
		0.746597,
		0.763962,
		0.780913,
		0.797413,
		0.813428,
		0.828924,
		0.843868,
		0.858226,
		0.87197,
		0.88507,
		0.897499,
		0.909232,
		0.920249,
		0.930531,
		0.940063,
		0.948834,
		0.956839,
		0.964076,
		0.970552,
		0.976276,
		0.981267,
		0.985548,
		0.989154,
		0.992122,
		0.994502,
		0.996348,
		0.997721,
		0.99869,
		0.999327,
		0.999707,
		0.999901,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999977,
		0.999887,
		0.999651,
		0.999166,
		0.99831,
		0.996943,
		0.994912,
		0.992055,
		0.988203,
		0.983187,
		0.976835,
		0.968983,
		0.959469,
		0.948142,
		0.93486,
		0.919497,
		0.901939,
		0.882091,
		0.859876,
		0.835238,
		0.808146,
		0.778592,
		0.746597,
		0.71221,
		0.675515,
		0.636628,
		0.595706,
		0.552945,
		0.508587,
		0.462921,
		0.416289,
		0.369091,
		0.321785,
		0.2749,
		0.229035,
		0.184867,
		0.14316,
		0.104768,
		0.070648,
		0.041863,
		0.019597,
		0.005159,
		0.046277,
		0.054161,
		0.062632,
		0.071682,
		0.081301,
		0.09148,
		0.102209,
		0.113477,
		0.125272,
		0.137583,
		0.150396,
		0.163699,
		0.177478,
		0.191718,
		0.206404,
		0.22152,
		0.237049,
		0.252975,
		0.26928,
		0.285944,
		0.302949,
		0.320274,
		0.337898,
		0.355799,
		0.373956,
		0.392345,
		0.410941,
		0.429721,
		0.448658,
		0.467727,
		0.4869,
		0.506149,
		0.525446,
		0.544762,
		0.564067,
		0.583329,
		0.602518,
		0.621603,
		0.640549,
		0.659326,
		0.677898,
		0.696234,
		0.714298,
		0.732056,
		0.749475,
		0.76652,
		0.783156,
		0.799352,
		0.815072,
		0.830285,
		0.84496,
		0.859064,
		0.872571,
		0.885451,
		0.897679,
		0.909232,
		0.92009,
		0.930235,
		0.939652,
		0.948332,
		0.956268,
		0.963459,
		0.969911,
		0.975631,
		0.980638,
		0.984953,
		0.988607,
		0.991635,
		0.994084,
		0.996003,
		0.997451,
		0.998492,
		0.999193,
		0.999626,
		0.999861,
		0.999964,
		0.999996,
		1,
		1,
		0.99999,
		0.999933,
		0.999766,
		0.999399,
		0.998722,
		0.997607,
		0.995914,
		0.99349,
		0.990178,
		0.985814,
		0.980236,
		0.973282,
		0.964797,
		0.95463,
		0.942641,
		0.9287,
		0.912691,
		0.894513,
		0.874082,
		0.851331,
		0.826216,
		0.798713,
		0.768825,
		0.736579,
		0.702034,
		0.665277,
		0.62643,
		0.585653,
		0.543144,
		0.499142,
		0.453937,
		0.407865,
		0.361318,
		0.314747,
		0.268666,
		0.223658,
		0.18038,
		0.139573,
		0.102062,
		0.068768,
		0.040718,
		0.019046,
		0.00501,
		0.078987,
		0.088884,
		0.099316,
		0.110275,
		0.121749,
		0.133727,
		0.146196,
		0.159146,
		0.172561,
		0.18643,
		0.200737,
		0.215467,
		0.230606,
		0.246136,
		0.262041,
		0.278304,
		0.294906,
		0.311827,
		0.32905,
		0.346552,
		0.364313,
		0.382312,
		0.400525,
		0.418929,
		0.437501,
		0.456215,
		0.475047,
		0.493969,
		0.512956,
		0.531979,
		0.55101,
		0.570021,
		0.588982,
		0.607863,
		0.626633,
		0.645261,
		0.663715,
		0.681965,
		0.699976,
		0.717717,
		0.735156,
		0.752258,
		0.768993,
		0.785326,
		0.801227,
		0.816662,
		0.831602,
		0.846017,
		0.859876,
		0.873152,
		0.88582,
		0.897854,
		0.909232,
		0.919935,
		0.929947,
		0.939252,
		0.947842,
		0.95571,
		0.962855,
		0.969281,
		0.974996,
		0.980016,
		0.984361,
		0.988059,
		0.991144,
		0.993658,
		0.995649,
		0.997169,
		0.99828,
		0.999046,
		0.999533,
		0.999811,
		0.999944,
		0.999991,
		1,
		1,
		0.999996,
		0.999963,
		0.999848,
		0.999575,
		0.999046,
		0.998145,
		0.996742,
		0.994696,
		0.991857,
		0.988073,
		0.983187,
		0.977043,
		0.969489,
		0.960379,
		0.949573,
		0.936941,
		0.922365,
		0.905738,
		0.886971,
		0.86599,
		0.842738,
		0.817181,
		0.789303,
		0.759116,
		0.726654,
		0.691982,
		0.655192,
		0.61641,
		0.575799,
		0.533556,
		0.489923,
		0.445185,
		0.399673,
		0.353774,
		0.307926,
		0.262634,
		0.218463,
		0.176052,
		0.136118,
		0.099458,
		0.066963,
		0.039618,
		0.018518,
		0.004868,
		0.118371,
		0.130029,
		0.142168,
		0.154777,
		0.167843,
		0.181354,
		0.195295,
		0.209654,
		0.224414,
		0.239562,
		0.255081,
		0.270954,
		0.287165,
		0.303696,
		0.320527,
		0.33764,
		0.355015,
		0.372632,
		0.390469,
		0.408505,
		0.426716,
		0.44508,
		0.463572,
		0.482168,
		0.500843,
		0.51957,
		0.538323,
		0.557076,
		0.575799,
		0.594464,
		0.613044,
		0.631507,
		0.649825,
		0.667967,
		0.685901,
		0.703599,
		0.721027,
		0.738155,
		0.754952,
		0.771385,
		0.787425,
		0.803041,
		0.818201,
		0.832877,
		0.84704,
		0.860662,
		0.873716,
		0.886177,
		0.898023,
		0.909232,
		0.919785,
		0.929666,
		0.938862,
		0.947363,
		0.955164,
		0.962262,
		0.968661,
		0.974369,
		0.9794,
		0.983772,
		0.987511,
		0.99065,
		0.993226,
		0.995285,
		0.996876,
		0.998056,
		0.998886,
		0.999429,
		0.999752,
		0.999917,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999905,
		0.999707,
		0.999299,
		0.998577,
		0.997422,
		0.995704,
		0.993282,
		0.990011,
		0.985742,
		0.980326,
		0.973615,
		0.965467,
		0.955742,
		0.944313,
		0.931058,
		0.91587,
		0.898654,
		0.879329,
		0.85783,
		0.834112,
		0.808146,
		0.779928,
		0.749475,
		0.716828,
		0.682058,
		0.645261,
		0.606567,
		0.56614,
		0.52418,
		0.480925,
		0.436658,
		0.391707,
		0.346449,
		0.301315,
		0.256796,
		0.213443,
		0.171875,
		0.132788,
		0.096952,
		0.065227,
		0.038563,
		0.018011,
		0.004731,
		0.163312,
		0.176478,
		0.190067,
		0.204067,
		0.218463,
		0.233241,
		0.248386,
		0.263882,
		0.279714,
		0.295864,
		0.312315,
		0.32905,
		0.346048,
		0.363292,
		0.38076,
		0.398433,
		0.416289,
		0.434306,
		0.452462,
		0.470734,
		0.489096,
		0.507526,
		0.525998,
		0.544486,
		0.562964,
		0.581406,
		0.599783,
		0.618068,
		0.636233,
		0.654248,
		0.672085,
		0.689715,
		0.707107,
		0.724231,
		0.741059,
		0.757559,
		0.773702,
		0.789457,
		0.804797,
		0.819691,
		0.834112,
		0.848031,
		0.861423,
		0.874262,
		0.886525,
		0.898188,
		0.909232,
		0.919639,
		0.929393,
		0.938482,
		0.946896,
		0.95463,
		0.961681,
		0.968053,
		0.973751,
		0.97879,
		0.983187,
		0.986964,
		0.990153,
		0.992789,
		0.994912,
		0.996571,
		0.997819,
		0.998713,
		0.999313,
		0.999682,
		0.999882,
		0.99997,
		0.999997,
		1,
		1,
		0.999991,
		0.999944,
		0.999803,
		0.999493,
		0.998921,
		0.997978,
		0.996543,
		0.994485,
		0.991668,
		0.987949,
		0.983187,
		0.977237,
		0.969962,
		0.961225,
		0.950898,
		0.938862,
		0.925006,
		0.909232,
		0.891454,
		0.8716,
		0.849616,
		0.825463,
		0.799123,
		0.770596,
		0.739908,
		0.707107,
		0.672265,
		0.635486,
		0.596902,
		0.556677,
		0.515011,
		0.472143,
		0.428351,
		0.383959,
		0.339336,
		0.294906,
		0.251144,
		0.208589,
		0.167843,
		0.129577,
		0.094539,
		0.063557,
		0.037549,
		0.017525,
		0.0046,
		0.212739,
		0.227159,
		0.241943,
		0.257074,
		0.272538,
		0.288319,
		0.304401,
		0.320766,
		0.337397,
		0.354276,
		0.371384,
		0.388702,
		0.406208,
		0.423884,
		0.441707,
		0.459655,
		0.477707,
		0.495839,
		0.514027,
		0.532247,
		0.550474,
		0.568684,
		0.586849,
		0.604944,
		0.622942,
		0.640815,
		0.658536,
		0.676077,
		0.69341,
		0.710506,
		0.727336,
		0.743871,
		0.760084,
		0.775945,
		0.791425,
		0.806498,
		0.821134,
		0.835308,
		0.848992,
		0.862161,
		0.874792,
		0.886861,
		0.898348,
		0.909232,
		0.919497,
		0.929128,
		0.938112,
		0.94644,
		0.954108,
		0.961112,
		0.967454,
		0.973143,
		0.978187,
		0.982605,
		0.986419,
		0.989654,
		0.992346,
		0.994532,
		0.996257,
		0.997571,
		0.998528,
		0.999185,
		0.999603,
		0.999839,
		0.999952,
		0.999992,
		1,
		1,
		0.999997,
		0.999969,
		0.999872,
		0.999641,
		0.999193,
		0.998429,
		0.997238,
		0.995499,
		0.993081,
		0.989852,
		0.985674,
		0.980411,
		0.973927,
		0.96609,
		0.956775,
		0.94586,
		0.933236,
		0.9188,
		0.902465,
		0.884152,
		0.863799,
		0.84136,
		0.816804,
		0.79012,
		0.761317,
		0.730424,
		0.697495,
		0.662608,
		0.625869,
		0.587413,
		0.547405,
		0.506046,
		0.463572,
		0.420258,
		0.376422,
		0.332429,
		0.28869,
		0.245671,
		0.203896,
		0.163949,
		0.12648,
		0.092214,
		0.061951,
		0.036574,
		0.017058,
		0.004475,
		0.265625,
		0.281048,
		0.29677,
		0.312776,
		0.32905,
		0.345573,
		0.362328,
		0.379296,
		0.39646,
		0.413799,
		0.431293,
		0.448922,
		0.466665,
		0.484498,
		0.502402,
		0.520351,
		0.538323,
		0.556295,
		0.57424,
		0.592135,
		0.609954,
		0.627672,
		0.645261,
		0.662695,
		0.679948,
		0.696992,
		0.7138,
		0.730344,
		0.746597,
		0.762531,
		0.778118,
		0.793332,
		0.808146,
		0.822533,
		0.836467,
		0.849923,
		0.862877,
		0.875306,
		0.887188,
		0.898503,
		0.909232,
		0.919359,
		0.928869,
		0.937751,
		0.945995,
		0.953597,
		0.960553,
		0.966867,
		0.972543,
		0.977591,
		0.982029,
		0.985874,
		0.989154,
		0.991898,
		0.994144,
		0.995933,
		0.997312,
		0.998331,
		0.999046,
		0.999512,
		0.999788,
		0.999929,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.99992,
		0.999752,
		0.999406,
		0.998793,
		0.997811,
		0.996348,
		0.994282,
		0.991486,
		0.987832,
		0.983187,
		0.97742,
		0.970403,
		0.962012,
		0.952128,
		0.94064,
		0.927446,
		0.912454,
		0.895582,
		0.876762,
		0.855939,
		0.833075,
		0.808146,
		0.781149,
		0.752097,
		0.721027,
		0.687996,
		0.653088,
		0.61641,
		0.5781,
		0.538323,
		0.497281,
		0.455206,
		0.412372,
		0.369091,
		0.325719,
		0.282661,
		0.24037,
		0.199356,
		0.160186,
		0.123492,
		0.089974,
		0.060404,
		0.035637,
		0.01661,
		0.004354,
		0.320992,
		0.337168,
		0.353578,
		0.370205,
		0.387032,
		0.40404,
		0.42121,
		0.438523,
		0.455958,
		0.473495,
		0.491113,
		0.50879,
		0.526505,
		0.544233,
		0.561953,
		0.57964,
		0.59727,
		0.61482,
		0.632263,
		0.649575,
		0.66673,
		0.683702,
		0.700466,
		0.716994,
		0.73326,
		0.749239,
		0.764902,
		0.780225,
		0.795181,
		0.809744,
		0.823889,
		0.837591,
		0.850827,
		0.863572,
		0.875806,
		0.887506,
		0.898654,
		0.909232,
		0.919224,
		0.928617,
		0.937398,
		0.94556,
		0.953097,
		0.960006,
		0.966289,
		0.971951,
		0.977002,
		0.981456,
		0.985332,
		0.988652,
		0.991447,
		0.99375,
		0.995601,
		0.997042,
		0.998123,
		0.998895,
		0.999412,
		0.999729,
		0.999899,
		0.999974,
		0.999997,
		1,
		1,
		0.999993,
		0.999952,
		0.999833,
		0.99957,
		0.999083,
		0.99828,
		0.997056,
		0.995298,
		0.992887,
		0.9897,
		0.98561,
		0.980491,
		0.974219,
		0.966673,
		0.957735,
		0.947296,
		0.935251,
		0.921508,
		0.90598,
		0.888597,
		0.869296,
		0.848031,
		0.82477,
		0.799498,
		0.772216,
		0.742944,
		0.711723,
		0.678615,
		0.643707,
		0.607109,
		0.568961,
		0.529428,
		0.488711,
		0.447041,
		0.404687,
		0.361957,
		0.3192,
		0.276811,
		0.235233,
		0.194962,
		0.156549,
		0.120607,
		0.087813,
		0.058915,
		0.034735,
		0.016179,
		0.004238,
		0.377913,
		0.394596,
		0.411446,
		0.428447,
		0.445578,
		0.46282,
		0.480153,
		0.497557,
		0.515011,
		0.532493,
		0.549982,
		0.567455,
		0.584888,
		0.60226,
		0.619546,
		0.636722,
		0.653763,
		0.670646,
		0.687346,
		0.703836,
		0.720092,
		0.736089,
		0.751801,
		0.767202,
		0.782268,
		0.796974,
		0.811294,
		0.825205,
		0.838682,
		0.851704,
		0.864247,
		0.87629,
		0.887815,
		0.898801,
		0.909232,
		0.919093,
		0.928371,
		0.937054,
		0.945135,
		0.952607,
		0.959469,
		0.965721,
		0.971369,
		0.97642,
		0.980889,
		0.984791,
		0.98815,
		0.990993,
		0.99335,
		0.99526,
		0.996763,
		0.997904,
		0.998733,
		0.9993,
		0.99966,
		0.999863,
		0.999959,
		0.999993,
		1,
		1,
		0.999997,
		0.999973,
		0.999891,
		0.999695,
		0.999313,
		0.998662,
		0.997644,
		0.996156,
		0.994084,
		0.991312,
		0.98772,
		0.983187,
		0.977591,
		0.970816,
		0.962746,
		0.953272,
		0.94229,
		0.929706,
		0.915432,
		0.899393,
		0.881522,
		0.861767,
		0.840087,
		0.816458,
		0.790869,
		0.763329,
		0.733863,
		0.702516,
		0.669354,
		0.634466,
		0.597967,
		0.559995,
		0.520718,
		0.480333,
		0.439072,
		0.397198,
		0.355015,
		0.312866,
		0.271135,
		0.230255,
		0.19071,
		0.153034,
		0.117821,
		0.085729,
		0.057479,
		0.033867,
		0.015765,
		0.004127,
		0.435513,
		0.452462,
		0.469512,
		0.486644,
		0.503836,
		0.52107,
		0.538323,
		0.555576,
		0.572806,
		0.589991,
		0.607109,
		0.624138,
		0.641052,
		0.657831,
		0.674449,
		0.690882,
		0.707107,
		0.723099,
		0.738833,
		0.754286,
		0.769433,
		0.78425,
		0.798713,
		0.812798,
		0.826481,
		0.839741,
		0.852555,
		0.864902,
		0.876762,
		0.888115,
		0.898944,
		0.909232,
		0.918966,
		0.928131,
		0.936718,
		0.944719,
		0.952128,
		0.958943,
		0.965164,
		0.970795,
		0.975845,
		0.980326,
		0.984253,
		0.987648,
		0.990536,
		0.992945,
		0.994912,
		0.996474,
		0.997675,
		0.99856,
		0.999178,
		0.999581,
		0.999819,
		0.999939,
		0.999987,
		0.999999,
		1,
		0.999999,
		0.999986,
		0.999932,
		0.999788,
		0.999493,
		0.99897,
		0.99813,
		0.996876,
		0.995103,
		0.9927,
		0.989554,
		0.985548,
		0.980567,
		0.974494,
		0.967218,
		0.958631,
		0.948631,
		0.937122,
		0.924016,
		0.909232,
		0.892703,
		0.874369,
		0.854185,
		0.832117,
		0.808146,
		0.782268,
		0.754496,
		0.72486,
		0.69341,
		0.660215,
		0.625367,
		0.588982,
		0.5512,
		0.512188,
		0.472143,
		0.431293,
		0.3899,
		0.34826,
		0.306709,
		0.265625,
		0.225429,
		0.186592,
		0.149633,
		0.11513,
		0.083718,
		0.056095,
		0.033031,
		0.015366,
		0.00402,
		0.492972,
		0.509955,
		0.526971,
		0.544,
		0.561021,
		0.578013,
		0.594955,
		0.611825,
		0.628601,
		0.645261,
		0.661782,
		0.678141,
		0.694316,
		0.710282,
		0.726017,
		0.741497,
		0.756699,
		0.771599,
		0.786174,
		0.800401,
		0.814257,
		0.827721,
		0.840769,
		0.853382,
		0.865538,
		0.877219,
		0.888407,
		0.899083,
		0.909232,
		0.918841,
		0.927897,
		0.93639,
		0.944313,
		0.951659,
		0.958426,
		0.964615,
		0.97023,
		0.975277,
		0.979768,
		0.983719,
		0.987147,
		0.990077,
		0.992536,
		0.994557,
		0.996177,
		0.997435,
		0.998376,
		0.999046,
		0.999493,
		0.999767,
		0.999914,
		0.999978,
		0.999998,
		1,
		1,
		0.999994,
		0.999959,
		0.999857,
		0.999632,
		0.999216,
		0.998528,
		0.997478,
		0.995967,
		0.993892,
		0.991144,
		0.987613,
		0.983187,
		0.977753,
		0.971204,
		0.963433,
		0.954338,
		0.943825,
		0.931803,
		0.918192,
		0.90292,
		0.885923,
		0.867151,
		0.846562,
		0.824131,
		0.799844,
		0.773702,
		0.745722,
		0.71594,
		0.684409,
		0.651201,
		0.61641,
		0.580154,
		0.542574,
		0.503836,
		0.464137,
		0.423701,
		0.382786,
		0.341684,
		0.300724,
		0.260276,
		0.22075,
		0.182603,
		0.146343,
		0.112529,
		0.081776,
		0.054761,
		0.032226,
		0.014982,
		0.003918,
		0.549528,
		0.566321,
		0.583079,
		0.599783,
		0.61641,
		0.63294,
		0.649351,
		0.665622,
		0.681729,
		0.697651,
		0.713366,
		0.728851,
		0.744084,
		0.759042,
		0.773702,
		0.788042,
		0.80204,
		0.815675,
		0.828924,
		0.841768,
		0.854185,
		0.866157,
		0.877665,
		0.88869,
		0.899218,
		0.909232,
		0.91872,
		0.927669,
		0.93607,
		0.943915,
		0.951199,
		0.95792,
		0.964076,
		0.969673,
		0.974716,
		0.979216,
		0.983187,
		0.986646,
		0.989616,
		0.992122,
		0.994196,
		0.995871,
		0.997186,
		0.998182,
		0.998903,
		0.999396,
		0.999707,
		0.999882,
		0.999965,
		0.999994,
		1,
		1,
		0.999998,
		0.999977,
		0.999907,
		0.999739,
		0.999412,
		0.998853,
		0.997978,
		0.996697,
		0.994912,
		0.99252,
		0.989415,
		0.98549,
		0.980638,
		0.974752,
		0.967729,
		0.959469,
		0.949876,
		0.938862,
		0.926344,
		0.912247,
		0.896506,
		0.879064,
		0.859876,
		0.838908,
		0.816138,
		0.791559,
		0.765177,
		0.737013,
		0.707107,
		0.675515,
		0.642312,
		0.607595,
		0.571482,
		0.534114,
		0.495659,
		0.45631,
		0.416289,
		0.375851,
		0.335282,
		0.294906,
		0.255081,
		0.21621,
		0.178739,
		0.14316,
		0.110014,
		0.079901,
		0.053473,
		0.031449,
		0.014613,
		0.003819,
		0.604481,
		0.620871,
		0.63716,
		0.653328,
		0.669354,
		0.685216,
		0.700893,
		0.716363,
		0.731605,
		0.746597,
		0.761317,
		0.775744,
		0.789856,
		0.803632,
		0.817052,
		0.830094,
		0.842738,
		0.854966,
		0.866759,
		0.878098,
		0.888967,
		0.89935,
		0.909232,
		0.918602,
		0.927446,
		0.935757,
		0.943527,
		0.950749,
		0.957423,
		0.963547,
		0.969125,
		0.974162,
		0.978669,
		0.982658,
		0.986146,
		0.989154,
		0.991705,
		0.993829,
		0.995558,
		0.996928,
		0.997978,
		0.99875,
		0.999289,
		0.999638,
		0.999844,
		0.999948,
		0.999989,
		0.999999,
		1,
		0.999999,
		0.999988,
		0.999942,
		0.999819,
		0.999565,
		0.999115,
		0.998392,
		0.997312,
		0.995782,
		0.993706,
		0.990983,
		0.987511,
		0.983187,
		0.977906,
		0.971569,
		0.964076,
		0.955335,
		0.945255,
		0.933754,
		0.920755,
		0.906191,
		0.890001,
		0.872136,
		0.852555,
		0.83123,
		0.808146,
		0.783299,
		0.756699,
		0.728373,
		0.698364,
		0.66673,
		0.63355,
		0.598922,
		0.562964,
		0.52582,
		0.487653,
		0.448658,
		0.409054,
		0.369091,
		0.32905,
		0.289247,
		0.250035,
		0.211807,
		0.174994,
		0.140077,
		0.107582,
		0.078089,
		0.05223,
		0.0307,
		0.014256,
		0.003724,
		0.657196,
		0.672983,
		0.688606,
		0.704043,
		0.719275,
		0.73428,
		0.749038,
		0.763528,
		0.777728,
		0.791619,
		0.805179,
		0.81839,
		0.83123,
		0.843682,
		0.855726,
		0.867344,
		0.878519,
		0.889236,
		0.899478,
		0.909232,
		0.918486,
		0.927229,
		0.935451,
		0.943146,
		0.950308,
		0.956935,
		0.963026,
		0.968585,
		0.973615,
		0.978128,
		0.982133,
		0.985648,
		0.988691,
		0.991285,
		0.993458,
		0.995238,
		0.996662,
		0.997765,
		0.998588,
		0.999172,
		0.999561,
		0.999798,
		0.999926,
		0.999981,
		0.999998,
		1,
		1,
		0.999995,
		0.999965,
		0.999877,
		0.999684,
		0.999326,
		0.998733,
		0.997826,
		0.996521,
		0.994726,
		0.992346,
		0.989282,
		0.985435,
		0.980706,
		0.974996,
		0.968209,
		0.960253,
		0.951039,
		0.940484,
		0.928511,
		0.915048,
		0.900035,
		0.883417,
		0.865149,
		0.845197,
		0.823539,
		0.800163,
		0.77507,
		0.748275,
		0.719808,
		0.689715,
		0.658057,
		0.624916,
		0.59039,
		0.5546,
		0.517687,
		0.479816,
		0.441178,
		0.40199,
		0.362499,
		0.32298,
		0.283743,
		0.245133,
		0.207533,
		0.171364,
		0.137092,
		0.105229,
		0.076337,
		0.05103,
		0.029978,
		0.013913,
		0.003632,
		0.707107,
		0.722107,
		0.736881,
		0.751412,
		0.765677,
		0.779657,
		0.793332,
		0.806683,
		0.819691,
		0.832336,
		0.8446,
		0.856465,
		0.867913,
		0.878929,
		0.889497,
		0.899603,
		0.909232,
		0.918374,
		0.927017,
		0.935153,
		0.942774,
		0.949876,
		0.956457,
		0.962515,
		0.968053,
		0.973075,
		0.977591,
		0.981612,
		0.985151,
		0.988228,
		0.990862,
		0.993081,
		0.994912,
		0.996388,
		0.997543,
		0.998415,
		0.999046,
		0.999475,
		0.999746,
		0.999898,
		0.99997,
		0.999995,
		1,
		1,
		0.999998,
		0.99998,
		0.99992,
		0.999775,
		0.999493,
		0.999011,
		0.998255,
		0.997147,
		0.995601,
		0.993526,
		0.990828,
		0.987414,
		0.983187,
		0.978051,
		0.971912,
		0.964681,
		0.956268,
		0.946591,
		0.935573,
		0.923141,
		0.909232,
		0.893788,
		0.876762,
		0.858112,
		0.837812,
		0.815842,
		0.792195,
		0.766878,
		0.739908,
		0.71132,
		0.681162,
		0.649498,
		0.61641,
		0.582,
		0.546387,
		0.509713,
		0.472143,
		0.433865,
		0.395094,
		0.356071,
		0.317068,
		0.278389,
		0.24037,
		0.203384,
		0.167843,
		0.1342,
		0.102952,
		0.074644,
		0.049871,
		0.029281,
		0.013582,
		0.003544,
		0.753719,
		0.767766,
		0.781532,
		0.794999,
		0.808146,
		0.820956,
		0.833411,
		0.845492,
		0.857183,
		0.868468,
		0.879329,
		0.889753,
		0.899725,
		0.909232,
		0.918264,
		0.926809,
		0.93486,
		0.94241,
		0.949453,
		0.955987,
		0.962012,
		0.967529,
		0.972543,
		0.977061,
		0.981094,
		0.984657,
		0.987764,
		0.990437,
		0.9927,
		0.99458,
		0.996106,
		0.997312,
		0.998234,
		0.99891,
		0.999381,
		0.999686,
		0.999864,
		0.999955,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.99999,
		0.99995,
		0.999844,
		0.999625,
		0.999236,
		0.99861,
		0.997675,
		0.996348,
		0.994545,
		0.992177,
		0.989154,
		0.985382,
		0.98077,
		0.975226,
		0.968661,
		0.960989,
		0.952128,
		0.941999,
		0.930531,
		0.917657,
		0.903317,
		0.887461,
		0.870045,
		0.851035,
		0.830407,
		0.808146,
		0.78425,
		0.758729,
		0.731605,
		0.702914,
		0.672708,
		0.641052,
		0.608032,
		0.573749,
		0.538323,
		0.501896,
		0.464632,
		0.426716,
		0.38836,
		0.349803,
		0.31131,
		0.273179,
		0.23574,
		0.199356,
		0.164428,
		0.131398,
		0.100747,
		0.073006,
		0.04875,
		0.028608,
		0.013263,
		0.003458,
		0.796619,
		0.809569,
		0.822187,
		0.834457,
		0.846361,
		0.857883,
		0.869007,
		0.879718,
		0.890001,
		0.899843,
		0.909232,
		0.918157,
		0.926607,
		0.934575,
		0.942053,
		0.949038,
		0.955526,
		0.961517,
		0.967013,
		0.972017,
		0.976536,
		0.980581,
		0.984164,
		0.987301,
		0.990011,
		0.992316,
		0.994242,
		0.995817,
		0.997073,
		0.998044,
		0.998766,
		0.999278,
		0.999618,
		0.999825,
		0.999935,
		0.999984,
		0.999998,
		1,
		1,
		0.999995,
		0.99997,
		0.999894,
		0.999727,
		0.999417,
		0.998903,
		0.998117,
		0.996984,
		0.995423,
		0.99335,
		0.990679,
		0.987321,
		0.983187,
		0.978187,
		0.972236,
		0.965249,
		0.957143,
		0.947842,
		0.937272,
		0.925367,
		0.912065,
		0.897313,
		0.881063,
		0.863277,
		0.843926,
		0.82299,
		0.800458,
		0.776333,
		0.750628,
		0.723368,
		0.694592,
		0.664354,
		0.632722,
		0.599783,
		0.565638,
		0.530408,
		0.494234,
		0.457278,
		0.419726,
		0.381784,
		0.343689,
		0.3057,
		0.268109,
		0.231239,
		0.195443,
		0.161115,
		0.128681,
		0.098612,
		0.071421,
		0.047667,
		0.027957,
		0.012955,
		0.003376,
		0.835475,
		0.847207,
		0.858565,
		0.869533,
		0.880097,
		0.890244,
		0.899959,
		0.909232,
		0.918052,
		0.926409,
		0.934295,
		0.941704,
		0.948631,
		0.955074,
		0.961031,
		0.966504,
		0.971498,
		0.976017,
		0.980072,
		0.983674,
		0.986838,
		0.989583,
		0.991928,
		0.993899,
		0.995521,
		0.996826,
		0.997845,
		0.998613,
		0.999166,
		0.999542,
		0.999779,
		0.999911,
		0.999974,
		0.999996,
		1,
		1,
		0.999998,
		0.999983,
		0.999931,
		0.999805,
		0.999561,
		0.999142,
		0.998486,
		0.997523,
		0.996177,
		0.994368,
		0.992015,
		0.989031,
		0.985332,
		0.980831,
		0.975444,
		0.969087,
		0.961681,
		0.953149,
		0.943417,
		0.932418,
		0.92009,
		0.906376,
		0.891227,
		0.874601,
		0.856465,
		0.836792,
		0.815567,
		0.792785,
		0.76845,
		0.742579,
		0.715201,
		0.686356,
		0.656102,
		0.624508,
		0.591661,
		0.557664,
		0.522638,
		0.486723,
		0.45008,
		0.412891,
		0.375363,
		0.337724,
		0.300233,
		0.263174,
		0.226862,
		0.191643,
		0.157899,
		0.126047,
		0.096544,
		0.069887,
		0.046619,
		0.027329,
		0.012657,
		0.003297,
		0.870045,
		0.880467,
		0.89048,
		0.900072,
		0.909232,
		0.91795,
		0.926215,
		0.934022,
		0.941362,
		0.948233,
		0.95463,
		0.960553,
		0.966004,
		0.970985,
		0.975504,
		0.979567,
		0.983187,
		0.986377,
		0.989154,
		0.991538,
		0.993551,
		0.995219,
		0.996571,
		0.997637,
		0.998451,
		0.999046,
		0.999459,
		0.999726,
		0.999882,
		0.999961,
		0.999992,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999956,
		0.999864,
		0.999675,
		0.999337,
		0.998793,
		0.997978,
		0.996822,
		0.995249,
		0.993181,
		0.990536,
		0.987232,
		0.983187,
		0.978317,
		0.972543,
		0.965784,
		0.957965,
		0.949014,
		0.938862,
		0.927446,
		0.914709,
		0.900598,
		0.88507,
		0.868085,
		0.849616,
		0.82964,
		0.808146,
		0.785132,
		0.760605,
		0.734587,
		0.707107,
		0.67821,
		0.647954,
		0.61641,
		0.583667,
		0.549827,
		0.515011,
		0.47936,
		0.443032,
		0.406208,
		0.369091,
		0.331906,
		0.294906,
		0.258369,
		0.222604,
		0.18795,
		0.154777,
		0.123492,
		0.094539,
		0.068402,
		0.045605,
		0.026722,
		0.01237,
		0.003221,
		0.900183,
		0.909232,
		0.91785,
		0.926026,
		0.933754,
		0.941027,
		0.947842,
		0.954194,
		0.960083,
		0.965511,
		0.97048,
		0.974996,
		0.979066,
		0.982702,
		0.985916,
		0.988724,
		0.991144,
		0.993199,
		0.994912,
		0.99631,
		0.997422,
		0.99828,
		0.998917,
		0.999367,
		0.999666,
		0.999847,
		0.999944,
		0.999986,
		0.999998,
		1,
		1,
		0.999996,
		0.999974,
		0.999908,
		0.999763,
		0.999493,
		0.999046,
		0.998361,
		0.997372,
		0.996009,
		0.994196,
		0.991857,
		0.988913,
		0.985283,
		0.980889,
		0.97565,
		0.969489,
		0.962333,
		0.954108,
		0.944746,
		0.934185,
		0.922365,
		0.909232,
		0.89474,
		0.878848,
		0.861523,
		0.842738,
		0.822478,
		0.800733,
		0.777504,
		0.752804,
		0.726654,
		0.699089,
		0.670154,
		0.63991,
		0.608428,
		0.575799,
		0.542124,
		0.507526,
		0.472143,
		0.436133,
		0.399673,
		0.362964,
		0.326228,
		0.289713,
		0.25369,
		0.218463,
		0.18436,
		0.151745,
		0.121013,
		0.092596,
		0.066963,
		0.044624,
		0.026134,
		0.012092,
		0.003147,
		0.925841,
		0.933492,
		0.940699,
		0.947458,
		0.953766,
		0.959621,
		0.965026,
		0.969981,
		0.974494,
		0.97857,
		0.98222,
		0.985457,
		0.988294,
		0.990749,
		0.992844,
		0.994599,
		0.996042,
		0.997199,
		0.998102,
		0.99878,
		0.999268,
		0.999599,
		0.999806,
		0.999923,
		0.999977,
		0.999996,
		1,
		1,
		0.999998,
		0.999985,
		0.99994,
		0.999831,
		0.999618,
		0.999253,
		0.998681,
		0.997839,
		0.996662,
		0.995079,
		0.993016,
		0.990397,
		0.987147,
		0.983187,
		0.97844,
		0.972832,
		0.966289,
		0.958739,
		0.950115,
		0.940353,
		0.929393,
		0.917181,
		0.903668,
		0.88881,
		0.872571,
		0.854921,
		0.835839,
		0.815311,
		0.793332,
		0.769907,
		0.74505,
		0.718785,
		0.69115,
		0.662191,
		0.631971,
		0.600563,
		0.568056,
		0.534555,
		0.50018,
		0.465069,
		0.429378,
		0.393283,
		0.35698,
		0.320688,
		0.28465,
		0.249134,
		0.214433,
		0.180871,
		0.148801,
		0.118608,
		0.090711,
		0.065569,
		0.043674,
		0.025566,
		0.011823,
		0.003075,
		0.947082,
		0.953345,
		0.959167,
		0.964547,
		0.969489,
		0.973997,
		0.978079,
		0.981742,
		0.984999,
		0.987863,
		0.990352,
		0.992484,
		0.994282,
		0.995768,
		0.99697,
		0.997915,
		0.998635,
		0.999161,
		0.999525,
		0.99976,
		0.999897,
		0.999966,
		0.999993,
		1,
		1,
		1,
		0.999992,
		0.999962,
		0.999882,
		0.999716,
		0.999422,
		0.998946,
		0.998234,
		0.997222,
		0.995843,
		0.994029,
		0.991705,
		0.9888,
		0.985237,
		0.980944,
		0.975845,
		0.96987,
		0.962947,
		0.95501,
		0.945995,
		0.935842,
		0.924495,
		0.911904,
		0.898023,
		0.882814,
		0.866244,
		0.848287,
		0.828924,
		0.808146,
		0.785951,
		0.762345,
		0.737347,
		0.710982,
		0.683291,
		0.654322,
		0.624138,
		0.592813,
		0.560439,
		0.527118,
		0.492972,
		0.458135,
		0.422765,
		0.387032,
		0.351133,
		0.315281,
		0.279714,
		0.244695,
		0.210511,
		0.177478,
		0.14594,
		0.116272,
		0.088884,
		0.064217,
		0.042755,
		0.025016,
		0.011564,
		0.003007,
		0.964076,
		0.969004,
		0.973507,
		0.977591,
		0.981267,
		0.984543,
		0.987433,
		0.989954,
		0.992122,
		0.993959,
		0.995488,
		0.996733,
		0.997721,
		0.998482,
		0.999046,
		0.999443,
		0.999707,
		0.999866,
		0.999951,
		0.999988,
		0.999999,
		1,
		1,
		0.999996,
		0.999977,
		0.99992,
		0.999793,
		0.999557,
		0.999166,
		0.998566,
		0.9977,
		0.996504,
		0.994912,
		0.992856,
		0.990264,
		0.987065,
		0.983187,
		0.978558,
		0.973107,
		0.966766,
		0.959469,
		0.951151,
		0.941754,
		0.931219,
		0.919497,
		0.90654,
		0.892307,
		0.876762,
		0.859876,
		0.841627,
		0.822,
		0.800988,
		0.778592,
		0.754822,
		0.729698,
		0.703249,
		0.675515,
		0.646547,
		0.61641,
		0.58518,
		0.552945,
		0.519811,
		0.485897,
		0.451339,
		0.416289,
		0.380919,
		0.34542,
		0.310002,
		0.2749,
		0.24037,
		0.206693,
		0.174178,
		0.14316,
		0.114005,
		0.08711,
		0.062907,
		0.041863,
		0.024483,
		0.011312,
		0.00294,
		0.977109,
		0.980795,
		0.984089,
		0.987004,
		0.989554,
		0.991758,
		0.993633,
		0.995203,
		0.99649,
		0.99752,
		0.998321,
		0.998923,
		0.999355,
		0.999647,
		0.99983,
		0.999932,
		0.99998,
		0.999997,
		1,
		1,
		0.999999,
		0.999987,
		0.999947,
		0.999852,
		0.999666,
		0.999346,
		0.998845,
		0.998106,
		0.997073,
		0.995681,
		0.993865,
		0.991558,
		0.988691,
		0.985193,
		0.980996,
		0.976031,
		0.97023,
		0.963527,
		0.955861,
		0.94717,
		0.937398,
		0.926493,
		0.914407,
		0.901097,
		0.886525,
		0.870658,
		0.853472,
		0.834948,
		0.815072,
		0.793842,
		0.771261,
		0.747342,
		0.722107,
		0.695586,
		0.667822,
		0.638868,
		0.608789,
		0.577661,
		0.545574,
		0.512632,
		0.478955,
		0.444677,
		0.409949,
		0.37494,
		0.339837,
		0.30485,
		0.270205,
		0.236155,
		0.202976,
		0.170967,
		0.140457,
		0.111802,
		0.085389,
		0.061637,
		0.041,
		0.023968,
		0.011069,
		0.002875,
		0.986574,
		0.989154,
		0.99139,
		0.993303,
		0.994912,
		0.99624,
		0.997312,
		0.998153,
		0.998793,
		0.999259,
		0.999581,
		0.999788,
		0.999909,
		0.99997,
		0.999994,
		1,
		1,
		1,
		0.999993,
		0.999967,
		0.999897,
		0.999752,
		0.999493,
		0.999076,
		0.998451,
		0.997561,
		0.996348,
		0.994749,
		0.9927,
		0.990135,
		0.986986,
		0.983187,
		0.978669,
		0.973368,
		0.967218,
		0.960158,
		0.952128,
		0.943071,
		0.932935,
		0.92167,
		0.909232,
		0.895582,
		0.880684,
		0.864511,
		0.84704,
		0.828254,
		0.808146,
		0.786713,
		0.763962,
		0.739908,
		0.714575,
		0.687996,
		0.660215,
		0.631286,
		0.601273,
		0.570256,
		0.538323,
		0.50558,
		0.472143,
		0.438147,
		0.403741,
		0.369091,
		0.334382,
		0.299819,
		0.265625,
		0.232048,
		0.199356,
		0.167843,
		0.137829,
		0.109662,
		0.083718,
		0.060404,
		0.040162,
		0.023468,
		0.010834,
		0.002813,
		0.992969,
		0.994617,
		0.995985,
		0.997097,
		0.997978,
		0.998655,
		0.999156,
		0.999508,
		0.999741,
		0.999882,
		0.999957,
		0.999989,
		0.999999,
		1,
		1,
		0.999997,
		0.99998,
		0.99993,
		0.999819,
		0.999612,
		0.999268,
		0.998741,
		0.997978,
		0.996925,
		0.995521,
		0.993706,
		0.991416,
		0.988586,
		0.985151,
		0.981047,
		0.976208,
		0.970571,
		0.964076,
		0.956664,
		0.948277,
		0.938862,
		0.928371,
		0.916757,
		0.903979,
		0.890001,
		0.874792,
		0.858326,
		0.840584,
		0.821553,
		0.801227,
		0.779606,
		0.756699,
		0.732524,
		0.707107,
		0.680482,
		0.652694,
		0.623799,
		0.593863,
		0.562964,
		0.531193,
		0.498652,
		0.465459,
		0.431746,
		0.397661,
		0.363368,
		0.32905,
		0.294906,
		0.261156,
		0.228043,
		0.19583,
		0.164802,
		0.135274,
		0.107582,
		0.082095,
		0.059208,
		0.03935,
		0.022984,
		0.010606,
		0.002753,
		0.996876,
		0.997796,
		0.998511,
		0.999046,
		0.999429,
		0.999688,
		0.99985,
		0.99994,
		0.999982,
		0.999997,
		1,
		1,
		0.999999,
		0.999989,
		0.999954,
		0.99987,
		0.999707,
		0.999426,
		0.998984,
		0.998334,
		0.997422,
		0.996194,
		0.99459,
		0.992549,
		0.990011,
		0.986911,
		0.983187,
		0.978775,
		0.973615,
		0.967646,
		0.96081,
		0.95305,
		0.944313,
		0.934549,
		0.923712,
		0.91176,
		0.898654,
		0.884362,
		0.868855,
		0.85211,
		0.834112,
		0.814849,
		0.794318,
		0.772523,
		0.749475,
		0.725192,
		0.699703,
		0.673044,
		0.645261,
		0.61641,
		0.586559,
		0.555785,
		0.52418,
		0.491846,
		0.458899,
		0.425471,
		0.391707,
		0.357769,
		0.323837,
		0.290107,
		0.256796,
		0.224139,
		0.192394,
		0.161842,
		0.132788,
		0.105561,
		0.080519,
		0.058047,
		0.038563,
		0.022514,
		0.010385,
		0.002694,
		0.998929,
		0.999343,
		0.999629,
		0.999813,
		0.99992,
		0.999973,
		0.999995,
		1,
		1,
		1,
		0.999994,
		0.999971,
		0.999909,
		0.999782,
		0.999554,
		0.999187,
		0.998635,
		0.99785,
		0.996778,
		0.995364,
		0.993551,
		0.991278,
		0.988485,
		0.985111,
		0.981094,
		0.976376,
		0.970896,
		0.964597,
		0.957423,
		0.949321,
		0.940242,
		0.930138,
		0.918966,
		0.906686,
		0.893264,
		0.87867,
		0.862877,
		0.845868,
		0.827626,
		0.808146,
		0.787425,
		0.76547,
		0.742293,
		0.717915,
		0.692366,
		0.665684,
		0.637915,
		0.609118,
		0.579359,
		0.548718,
		0.517284,
		0.48516,
		0.452462,
		0.419319,
		0.385876,
		0.352291,
		0.318742,
		0.285421,
		0.25254,
		0.220331,
		0.189047,
		0.15896,
		0.130369,
		0.103595,
		0.078987,
		0.05692,
		0.037798,
		0.022059,
		0.010171,
		0.002638,
		0.999771,
		0.999896,
		0.999962,
		0.99999,
		0.999999,
		1,
		1,
		0.999997,
		0.999982,
		0.999938,
		0.99984,
		0.999658,
		0.999355,
		0.998889,
		0.998216,
		0.997284,
		0.996042,
		0.994434,
		0.992403,
		0.989891,
		0.986838,
		0.983187,
		0.978877,
		0.973851,
		0.968053,
		0.961427,
		0.953921,
		0.945484,
		0.93607,
		0.925634,
		0.914137,
		0.901541,
		0.887815,
		0.872931,
		0.856866,
		0.839605,
		0.821134,
		0.80145,
		0.780552,
		0.758449,
		0.735156,
		0.710695,
		0.685097,
		0.658402,
		0.630658,
		0.601923,
		0.572264,
		0.541761,
		0.510503,
		0.478594,
		0.446146,
		0.413289,
		0.380165,
		0.34693,
		0.31376,
		0.280842,
		0.248386,
		0.216617,
		0.185784,
		0.156153,
		0.128015,
		0.101684,
		0.077498,
		0.055825,
		0.037057,
		0.021618,
		0.009964,
		0.002583,
		0.999985,
		0.999998,
		1,
		1,
		0.999999,
		0.99999,
		0.999959,
		0.999886,
		0.999741,
		0.999493,
		0.999103,
		0.998528,
		0.997721,
		0.996633,
		0.995211,
		0.9934,
		0.991144,
		0.988388,
		0.985072,
		0.98114,
		0.976536,
		0.971204,
		0.96509,
		0.958141,
		0.950308,
		0.941544,
		0.931803,
		0.921045,
		0.909232,
		0.896331,
		0.882312,
		0.867151,
		0.850827,
		0.833326,
		0.814639,
		0.794763,
		0.773702,
		0.751464,
		0.728066,
		0.703534,
		0.677898,
		0.651201,
		0.62349,
		0.594824,
		0.565272,
		0.534913,
		0.503836,
		0.472143,
		0.439948,
		0.407377,
		0.37457,
		0.341684,
		0.308888,
		0.276369,
		0.24433,
		0.212995,
		0.182603,
		0.153419,
		0.125723,
		0.099824,
		0.076051,
		0.054761,
		0.036336,
		0.021189,
		0.009762,
		0.00253,
		1,
		1,
		0.999995,
		0.999974,
		0.99992,
		0.999807,
		0.999606,
		0.999281,
		0.998793,
		0.998097,
		0.997147,
		0.995893,
		0.994282,
		0.992261,
		0.989775,
		0.986769,
		0.983187,
		0.978974,
		0.974075,
		0.968439,
		0.962012,
		0.954745,
		0.946591,
		0.937506,
		0.927446,
		0.916375,
		0.904257,
		0.891062,
		0.876762,
		0.861335,
		0.844764,
		0.827036,
		0.808146,
		0.788092,
		0.766878,
		0.744517,
		0.721027,
		0.696434,
		0.670771,
		0.64408,
		0.61641,
		0.587822,
		0.558383,
		0.528173,
		0.497281,
		0.465807,
		0.433865,
		0.40158,
		0.369091,
		0.336549,
		0.304124,
		0.271997,
		0.24037,
		0.20946,
		0.179502,
		0.150754,
		0.123492,
		0.098015,
		0.074644,
		0.053727,
		0.035637,
		0.020774,
		0.009567,
		0.002478,
		0.999985,
		0.999945,
		0.999859,
		0.999698,
		0.999429,
		0.999017,
		0.99842,
		0.997593,
		0.99649,
		0.99506,
		0.993253,
		0.991015,
		0.988294,
		0.985035,
		0.981184,
		0.976689,
		0.971498,
		0.965558,
		0.958822,
		0.951243,
		0.942774,
		0.933375,
		0.923006,
		0.911631,
		0.899218,
		0.885739,
		0.871169,
		0.855488,
		0.838682,
		0.820741,
		0.801659,
		0.781437,
		0.760084,
		0.737611,
		0.71404,
		0.689396,
		0.663715,
		0.63704,
		0.60942,
		0.580916,
		0.551596,
		0.521539,
		0.490835,
		0.459584,
		0.427896,
		0.395897,
		0.363722,
		0.331523,
		0.299464,
		0.267725,
		0.236502,
		0.20601,
		0.176478,
		0.148158,
		0.121319,
		0.096253,
		0.073275,
		0.052722,
		0.034957,
		0.02037,
		0.009378,
		0.002428,
		0.999771,
		0.999551,
		0.999205,
		0.998695,
		0.997978,
		0.997011,
		0.995745,
		0.994133,
		0.992122,
		0.989663,
		0.986701,
		0.983187,
		0.979066,
		0.974289,
		0.968806,
		0.962567,
		0.955526,
		0.947639,
		0.938862,
		0.929157,
		0.918486,
		0.906817,
		0.894119,
		0.880367,
		0.865538,
		0.849616,
		0.832587,
		0.814442,
		0.795181,
		0.774805,
		0.753323,
		0.73075,
		0.707107,
		0.682423,
		0.656733,
		0.630081,
		0.602518,
		0.574105,
		0.54491,
		0.515011,
		0.484498,
		0.453471,
		0.422038,
		0.390324,
		0.358463,
		0.326602,
		0.294906,
		0.263549,
		0.232725,
		0.202643,
		0.173528,
		0.145627,
		0.119202,
		0.094539,
		0.071944,
		0.051745,
		0.034297,
		0.019978,
		0.009194,
		0.00238,
		0.998929,
		0.99831,
		0.997465,
		0.996348,
		0.994912,
		0.993109,
		0.99089,
		0.988203,
		0.984999,
		0.981226,
		0.976835,
		0.971777,
		0.966004,
		0.959469,
		0.952128,
		0.943938,
		0.93486,
		0.924857,
		0.913893,
		0.901939,
		0.888967,
		0.874952,
		0.859876,
		0.843722,
		0.826481,
		0.808146,
		0.788716,
		0.768196,
		0.746597,
		0.723934,
		0.70023,
		0.675515,
		0.649825,
		0.623205,
		0.595706,
		0.567389,
		0.538323,
		0.508587,
		0.478268,
		0.447466,
		0.416289,
		0.384859,
		0.353309,
		0.321785,
		0.290446,
		0.259467,
		0.229035,
		0.199356,
		0.170651,
		0.14316,
		0.11714,
		0.09287,
		0.070648,
		0.050795,
		0.033655,
		0.019597,
		0.009016,
		0.002333,
		0.996876,
		0.995601,
		0.993987,
		0.991988,
		0.989554,
		0.986637,
		0.983187,
		0.979155,
		0.974494,
		0.969156,
		0.963095,
		0.956268,
		0.948631,
		0.940146,
		0.930774,
		0.92048,
		0.909232,
		0.897002,
		0.883765,
		0.869498,
		0.854185,
		0.837812,
		0.82037,
		0.801855,
		0.782268,
		0.761615,
		0.739908,
		0.717165,
		0.69341,
		0.668673,
		0.642991,
		0.61641,
		0.588982,
		0.560768,
		0.531836,
		0.502265,
		0.472143,
		0.441568,
		0.410647,
		0.379501,
		0.34826,
		0.317068,
		0.286083,
		0.255475,
		0.225429,
		0.196146,
		0.167843,
		0.140754,
		0.11513,
		0.091244,
		0.069387,
		0.049871,
		0.033031,
		0.019227,
		0.008842,
		0.002287,
		0.992969,
		0.990768,
		0.988116,
		0.984964,
		0.981267,
		0.976975,
		0.972044,
		0.966428,
		0.960083,
		0.952968,
		0.945041,
		0.936266,
		0.926607,
		0.916031,
		0.904508,
		0.892012,
		0.878519,
		0.864011,
		0.848471,
		0.831889,
		0.814257,
		0.795573,
		0.77584,
		0.755064,
		0.73326,
		0.710447,
		0.686649,
		0.661898,
		0.636233,
		0.609698,
		0.582346,
		0.55424,
		0.525446,
		0.496045,
		0.466121,
		0.435774,
		0.405109,
		0.374245,
		0.343311,
		0.31245,
		0.281814,
		0.251572,
		0.221906,
		0.193012,
		0.165103,
		0.138407,
		0.113171,
		0.08966,
		0.068159,
		0.048971,
		0.032424,
		0.018867,
		0.008674,
		0.002243,
		0.986574,
		0.983187,
		0.97924,
		0.974689,
		0.969489,
		0.963598,
		0.956972,
		0.949573,
		0.941362,
		0.932304,
		0.922365,
		0.911514,
		0.899725,
		0.886971,
		0.873234,
		0.858494,
		0.842738,
		0.825958,
		0.808146,
		0.789303,
		0.769433,
		0.748546,
		0.726654,
		0.70378,
		0.679948,
		0.655192,
		0.629549,
		0.603067,
		0.575799,
		0.547804,
		0.519153,
		0.489923,
		0.460201,
		0.430082,
		0.399673,
		0.369091,
		0.338462,
		0.307926,
		0.277636,
		0.247755,
		0.218463,
		0.189951,
		0.162428,
		0.136118,
		0.111261,
		0.088117,
		0.066963,
		0.048096,
		0.031834,
		0.018518,
		0.00851,
		0.0022,
		0.977109,
		0.972299,
		0.966832,
		0.960668,
		0.953766,
		0.946088,
		0.937598,
		0.928264,
		0.918052,
		0.906935,
		0.894888,
		0.881887,
		0.867913,
		0.852952,
		0.83699,
		0.820021,
		0.80204,
		0.783049,
		0.763052,
		0.742062,
		0.720092,
		0.697165,
		0.673308,
		0.648554,
		0.622942,
		0.596519,
		0.569338,
		0.541461,
		0.512956,
		0.483901,
		0.454381,
		0.424491,
		0.394338,
		0.364035,
		0.333709,
		0.303496,
		0.273547,
		0.244022,
		0.215097,
		0.186961,
		0.159817,
		0.133884,
		0.109399,
		0.086613,
		0.065798,
		0.047243,
		0.031259,
		0.018178,
		0.008351,
		0.002158,
		0.964076,
		0.957642,
		0.950468,
		0.942516,
		0.933754,
		0.924149,
		0.913673,
		0.902298,
		0.890001,
		0.876762,
		0.862562,
		0.847388,
		0.83123,
		0.814083,
		0.795942,
		0.776812,
		0.756699,
		0.735615,
		0.713576,
		0.690605,
		0.66673,
		0.641985,
		0.61641,
		0.590052,
		0.562964,
		0.535208,
		0.506853,
		0.477974,
		0.448658,
		0.418999,
		0.3891,
		0.359076,
		0.32905,
		0.299157,
		0.269544,
		0.24037,
		0.211807,
		0.184039,
		0.157267,
		0.131705,
		0.107582,
		0.085147,
		0.064663,
		0.046414,
		0.0307,
		0.017847,
		0.008197,
		0.002117,
		0.947082,
		0.938862,
		0.929834,
		0.919966,
		0.909232,
		0.897607,
		0.88507,
		0.8716,
		0.857183,
		0.841807,
		0.825463,
		0.808146,
		0.789856,
		0.770596,
		0.750376,
		0.729206,
		0.707107,
		0.6841,
		0.660215,
		0.635486,
		0.609954,
		0.583667,
		0.556677,
		0.529046,
		0.500843,
		0.472143,
		0.443032,
		0.413603,
		0.383959,
		0.354211,
		0.324482,
		0.294906,
		0.265625,
		0.236797,
		0.208589,
		0.181184,
		0.154777,
		0.129577,
		0.10581,
		0.083718,
		0.063557,
		0.045605,
		0.030156,
		0.017525,
		0.008046,
		0.002078,
		0.925841,
		0.915718,
		0.904734,
		0.892867,
		0.880097,
		0.866407,
		0.851782,
		0.836212,
		0.819691,
		0.802214,
		0.783784,
		0.764404,
		0.744084,
		0.722839,
		0.700686,
		0.677651,
		0.653763,
		0.629057,
		0.603574,
		0.577362,
		0.550474,
		0.522972,
		0.494924,
		0.466406,
		0.437501,
		0.408303,
		0.378911,
		0.349439,
		0.320005,
		0.290741,
		0.261788,
		0.2333,
		0.205443,
		0.178394,
		0.152345,
		0.1275,
		0.104082,
		0.082324,
		0.06248,
		0.044818,
		0.029626,
		0.017212,
		0.0079,
		0.00204,
		0.900183,
		0.888082,
		0.875088,
		0.861185,
		0.846361,
		0.830607,
		0.813918,
		0.796291,
		0.777728,
		0.758237,
		0.737827,
		0.716513,
		0.694316,
		0.67126,
		0.647376,
		0.622699,
		0.59727,
		0.571138,
		0.544357,
		0.516987,
		0.489096,
		0.460761,
		0.432063,
		0.403095,
		0.373956,
		0.344756,
		0.315615,
		0.28666,
		0.258031,
		0.229879,
		0.202366,
		0.175666,
		0.149968,
		0.125472,
		0.102394,
		0.080964,
		0.061429,
		0.044051,
		0.02911,
		0.016907,
		0.007758,
		0.002002,
		0.870045,
		0.855939,
		0.840925,
		0.824995,
		0.808146,
		0.790377,
		0.771692,
		0.752097,
		0.731605,
		0.710231,
		0.687996,
		0.664927,
		0.641052,
		0.61641,
		0.591042,
		0.564995,
		0.538323,
		0.511089,
		0.483358,
		0.455206,
		0.426716,
		0.397978,
		0.369091,
		0.340162,
		0.31131,
		0.282661,
		0.254351,
		0.22653,
		0.199356,
		0.173,
		0.147646,
		0.123492,
		0.100747,
		0.079638,
		0.060404,
		0.043303,
		0.028608,
		0.01661,
		0.007619,
		0.001966,
		0.835475,
		0.819379,
		0.802379,
		0.784477,
		0.765677,
		0.745987,
		0.72542,
		0.703994,
		0.681729,
		0.658652,
		0.634794,
		0.610192,
		0.584888,
		0.55893,
		0.532373,
		0.505276,
		0.477707,
		0.449741,
		0.421459,
		0.39295,
		0.364313,
		0.335654,
		0.307089,
		0.278741,
		0.250747,
		0.223252,
		0.196411,
		0.170393,
		0.145377,
		0.121558,
		0.09914,
		0.078344,
		0.059405,
		0.042574,
		0.028118,
		0.016321,
		0.007484,
		0.001931,
		0.796619,
		0.778592,
		0.759685,
		0.739908,
		0.719275,
		0.697803,
		0.675515,
		0.652437,
		0.628601,
		0.604044,
		0.57881,
		0.552945,
		0.526505,
		0.499548,
		0.472143,
		0.444363,
		0.416289,
		0.38801,
		0.359622,
		0.331231,
		0.302949,
		0.2749,
		0.247217,
		0.220042,
		0.193529,
		0.167843,
		0.14316,
		0.119668,
		0.09757,
		0.077081,
		0.05843,
		0.041863,
		0.027641,
		0.016039,
		0.007353,
		0.001896,
		0.753719,
		0.733863,
		0.71317,
		0.691659,
		0.669354,
		0.646281,
		0.622473,
		0.597967,
		0.572806,
		0.547038,
		0.520718,
		0.493904,
		0.466665,
		0.439072,
		0.411206,
		0.383156,
		0.355015,
		0.326889,
		0.298888,
		0.271135,
		0.243759,
		0.2169,
		0.19071,
		0.165349,
		0.140991,
		0.117821,
		0.096037,
		0.075848,
		0.057479,
		0.04117,
		0.027175,
		0.015765,
		0.007225,
		0.001863,
		0.707107,
		0.685564,
		0.663248,
		0.640185,
		0.61641,
		0.59196,
		0.566876,
		0.541209,
		0.515011,
		0.488343,
		0.46127,
		0.433865,
		0.406208,
		0.378386,
		0.350491,
		0.322628,
		0.294906,
		0.267443,
		0.24037,
		0.213823,
		0.18795,
		0.16291,
		0.138872,
		0.116017,
		0.094539,
		0.074644,
		0.056551,
		0.040494,
		0.026722,
		0.015497,
		0.007101,
		0.00183,
		0.657196,
		0.63415,
		0.610413,
		0.586022,
		0.561021,
		0.535457,
		0.509384,
		0.482863,
		0.455958,
		0.428742,
		0.401293,
		0.373698,
		0.346048,
		0.318445,
		0.290998,
		0.263824,
		0.237049,
		0.210809,
		0.185248,
		0.160523,
		0.136799,
		0.114253,
		0.093076,
		0.073468,
		0.055645,
		0.039835,
		0.026279,
		0.015236,
		0.006979,
		0.001798,
		0.604481,
		0.580154,
		0.555238,
		0.529781,
		0.503836,
		0.477463,
		0.450728,
		0.423701,
		0.39646,
		0.369091,
		0.341684,
		0.31434,
		0.287165,
		0.260276,
		0.233795,
		0.207857,
		0.182603,
		0.158187,
		0.134771,
		0.112529,
		0.091646,
		0.07232,
		0.054761,
		0.039191,
		0.025848,
		0.014982,
		0.006861,
		0.001767,
		0.549528,
		0.52418,
		0.498366,
		0.472143,
		0.445578,
		0.41874,
		0.391707,
		0.364563,
		0.337397,
		0.310309,
		0.283404,
		0.256796,
		0.230606,
		0.204965,
		0.180014,
		0.155901,
		0.132788,
		0.110843,
		0.090249,
		0.071199,
		0.053897,
		0.038563,
		0.025427,
		0.014734,
		0.006746,
		0.001737,
		0.492972,
		0.466901,
		0.440506,
		0.413858,
		0.387032,
		0.360112,
		0.333186,
		0.306352,
		0.279714,
		0.253383,
		0.227479,
		0.202132,
		0.177478,
		0.153664,
		0.130847,
		0.109195,
		0.088884,
		0.070103,
		0.053054,
		0.03795,
		0.025016,
		0.014492,
		0.006633,
		0.001708,
		0.435513,
		0.409054,
		0.382435,
		0.355737,
		0.32905,
		0.302467,
		0.276092,
		0.250035,
		0.224414,
		0.199356,
		0.174994,
		0.151474,
		0.128949,
		0.107582,
		0.087549,
		0.069032,
		0.05223,
		0.037351,
		0.024615,
		0.014256,
		0.006524,
		0.001679,
		0.377913,
		0.351437,
		0.324985,
		0.298652,
		0.272538,
		0.246752,
		0.221409,
		0.196635,
		0.172561,
		0.14933,
		0.127091,
		0.106005,
		0.086243,
		0.067986,
		0.051425,
		0.036766,
		0.024223,
		0.014026,
		0.006417,
		0.001651,
		0.320992,
		0.294906,
		0.269049,
		0.24353,
		0.218463,
		0.193969,
		0.170178,
		0.14723,
		0.125272,
		0.104462,
		0.084966,
		0.066963,
		0.050639,
		0.036195,
		0.023841,
		0.013802,
		0.006313,
		0.001624,
		0.265625,
		0.24037,
		0.215573,
		0.191355,
		0.167843,
		0.145174,
		0.123492,
		0.102952,
		0.083718,
		0.065963,
		0.049871,
		0.035637,
		0.023468,
		0.013582,
		0.006211,
		0.001597,
		0.212739,
		0.188793,
		0.165555,
		0.14316,
		0.121749,
		0.101474,
		0.082496,
		0.064984,
		0.049119,
		0.035092,
		0.023103,
		0.013368,
		0.006111,
		0.001571,
		0.163312,
		0.141187,
		0.120042,
		0.100028,
		0.081301,
		0.064028,
		0.048385,
		0.034559,
		0.022747,
		0.013159,
		0.006014,
		0.001546,
		0.118371,
		0.098612,
		0.080132,
		0.063092,
		0.047667,
		0.034038,
		0.022399,
		0.012955,
		0.00592,
		0.001521,
		0.078987,
		0.062177,
		0.046964,
		0.033529,
		0.022059,
		0.012755,
		0.005827,
		0.001497,
		0.046277,
		0.033031,
		0.021727,
		0.01256,
		0.005737,
		0.001474,
		0.021402,
		0.01237,
		0.005649,
		0.001451,
		0.005562,
		0.001428,
		0.000807,
		9.2e-005,
		0.003208,
		0.010445,
		0.001462,
		0.00717,
		0.0168,
		0.029959,
		0.000375,
		0.004448,
		0.012657,
		0.024591,
		0.039864,
		0.058111,
		0.002328,
		0.009016,
		0.019635,
		0.033782,
		0.051077,
		0.071162,
		0.093699,
		0.00086,
		0.00592,
		0.015128,
		0.028064,
		0.044332,
		0.063557,
		0.085389,
		0.109496,
		0.135567,
		9.8e-005,
		0.003417,
		0.011113,
		0.022747,
		0.037906,
		0.0562,
		0.077262,
		0.100747,
		0.126332,
		0.153712,
		0.182603,
		0.001559,
		0.007634,
		0.01787,
		0.031834,
		0.049119,
		0.069344,
		0.092147,
		0.117192,
		0.144161,
		0.172758,
		0.202704,
		0.233741,
		0.0004,
		0.004741,
		0.013475,
		0.026151,
		0.042346,
		0.06166,
		0.083718,
		0.108167,
		0.134676,
		0.162935,
		0.192655,
		0.223565,
		0.255412,
		0.28796,
		0.002483,
		0.009606,
		0.020897,
		0.035914,
		0.05424,
		0.075484,
		0.099276,
		0.125272,
		0.153149,
		0.182603,
		0.213353,
		0.245133,
		0.2777,
		0.310823,
		0.344293,
		0.000918,
		0.006313,
		0.016114,
		0.02986,
		0.047116,
		0.067471,
		0.090543,
		0.11597,
		0.143415,
		0.172561,
		0.203114,
		0.234798,
		0.267356,
		0.300551,
		0.33416,
		0.36798,
		0.401822,
		0.000105,
		0.003647,
		0.011847,
		0.024223,
		0.04032,
		0.05971,
		0.081992,
		0.106789,
		0.13375,
		0.162543,
		0.19286,
		0.224414,
		0.256937,
		0.290179,
		0.32391,
		0.357915,
		0.391998,
		0.425977,
		0.459685,
		0.001665,
		0.008146,
		0.019046,
		0.033889,
		0.05223,
		0.073649,
		0.097753,
		0.124172,
		0.152563,
		0.182603,
		0.213992,
		0.246449,
		0.279714,
		0.313545,
		0.34772,
		0.382031,
		0.416289,
		0.45032,
		0.483964,
		0.517077,
		0.000428,
		0.005063,
		0.014374,
		0.027864,
		0.045066,
		0.065543,
		0.088884,
		0.114702,
		0.142639,
		0.172357,
		0.203541,
		0.2359,
		0.269161,
		0.303071,
		0.337397,
		0.371923,
		0.40645,
		0.440796,
		0.474793,
		0.508291,
		0.541152,
		0.573252,
		0.002655,
		0.010256,
		0.022285,
		0.038254,
		0.057705,
		0.080209,
		0.105362,
		0.132788,
		0.162135,
		0.193074,
		0.2253,
		0.258527,
		0.292492,
		0.32695,
		0.361676,
		0.39646,
		0.431113,
		0.465459,
		0.499339,
		0.532609,
		0.56514,
		0.596815,
		0.62753,
		0.000982,
		0.006746,
		0.0172,
		0.031834,
		0.050169,
		0.071756,
		0.096175,
		0.12303,
		0.151954,
		0.182603,
		0.214658,
		0.247821,
		0.281814,
		0.316383,
		0.351291,
		0.38632,
		0.42127,
		0.455958,
		0.490217,
		0.523896,
		0.556859,
		0.588982,
		0.620158,
		0.650291,
		0.679298,
		0.000112,
		0.003901,
		0.012657,
		0.025848,
		0.042971,
		0.063557,
		0.087167,
		0.113386,
		0.141832,
		0.172143,
		0.203986,
		0.237049,
		0.271043,
		0.3057,
		0.340772,
		0.376031,
		0.411269,
		0.446292,
		0.480925,
		0.515011,
		0.548406,
		0.58098,
		0.612621,
		0.643226,
		0.672708,
		0.700991,
		0.728012,
		0.001783,
		0.008711,
		0.020342,
		0.03615,
		0.055645,
		0.078366,
		0.103881,
		0.131787,
		0.16171,
		0.193297,
		0.226223,
		0.260186,
		0.294906,
		0.330122,
		0.365596,
		0.401108,
		0.436458,
		0.471461,
		0.505952,
		0.539778,
		0.572806,
		0.604914,
		0.635997,
		0.66596,
		0.694725,
		0.722224,
		0.7484,
		0.77321,
		0.000458,
		0.005419,
		0.015366,
		0.02975,
		0.048056,
		0.069803,
		0.094539,
		0.121842,
		0.151319,
		0.182603,
		0.215353,
		0.249252,
		0.284006,
		0.319344,
		0.355015,
		0.39079,
		0.426457,
		0.461824,
		0.496716,
		0.530974,
		0.564456,
		0.597036,
		0.628601,
		0.659052,
		0.688306,
		0.71629,
		0.742944,
		0.76822,
		0.792083,
		0.814507,
		0.002844,
		0.010974,
		0.023816,
		0.04083,
		0.061512,
		0.085389,
		0.112019,
		0.140991,
		0.171921,
		0.204451,
		0.238249,
		0.273008,
		0.308444,
		0.344293,
		0.380315,
		0.416289,
		0.452013,
		0.487302,
		0.52199,
		0.555928,
		0.588982,
		0.621034,
		0.651979,
		0.681729,
		0.710206,
		0.737347,
		0.7631,
		0.787425,
		0.810295,
		0.831692,
		0.851607,
		0.001053,
		0.007225,
		0.018399,
		0.034009,
		0.053528,
		0.076461,
		0.102344,
		0.130747,
		0.161266,
		0.193529,
		0.227188,
		0.261919,
		0.297425,
		0.333432,
		0.369686,
		0.405954,
		0.442026,
		0.477707,
		0.512824,
		0.547218,
		0.580749,
		0.613292,
		0.644738,
		0.67499,
		0.703969,
		0.731605,
		0.757843,
		0.782641,
		0.805968,
		0.827802,
		0.848135,
		0.866967,
		0.88431,
		0.00012,
		0.004182,
		0.013553,
		0.027641,
		0.045892,
		0.067787,
		0.092843,
		0.120607,
		0.150658,
		0.182603,
		0.216079,
		0.250747,
		0.286296,
		0.322436,
		0.358903,
		0.395453,
		0.431863,
		0.467931,
		0.503473,
		0.538323,
		0.572334,
		0.605372,
		0.637323,
		0.668086,
		0.697574,
		0.725714,
		0.752448,
		0.777728,
		0.801521,
		0.823803,
		0.844563,
		0.863799,
		0.881522,
		0.897749,
		0.91251,
		0.001913,
		0.009337,
		0.021774,
		0.038645,
		0.059405,
		0.083548,
		0.110599,
		0.140116,
		0.171689,
		0.204936,
		0.239503,
		0.275061,
		0.31131,
		0.34797,
		0.384786,
		0.421525,
		0.457972,
		0.493936,
		0.529241,
		0.563732,
		0.59727,
		0.629732,
		0.661012,
		0.691017,
		0.71967,
		0.746909,
		0.772681,
		0.796951,
		0.819691,
		0.840888,
		0.860539,
		0.878652,
		0.895243,
		0.910341,
		0.923982,
		0.936212,
		0.000492,
		0.005814,
		0.016464,
		0.031834,
		0.051353,
		0.07449,
		0.100747,
		0.129662,
		0.160804,
		0.193772,
		0.228195,
		0.26373,
		0.300059,
		0.336891,
		0.373956,
		0.41101,
		0.447829,
		0.48421,
		0.519969,
		0.554942,
		0.588982,
		0.62196,
		0.653763,
		0.684294,
		0.713469,
		0.741222,
		0.767497,
		0.792253,
		0.815463,
		0.837108,
		0.857183,
		0.875696,
		0.892661,
		0.908106,
		0.922066,
		0.934586,
		0.945719,
		0.955526,
		0.003055,
		0.011771,
		0.02551,
		0.043674,
		0.065706,
		0.091084,
		0.119322,
		0.149968,
		0.182603,
		0.216838,
		0.252311,
		0.28869,
		0.325668,
		0.362964,
		0.40032,
		0.437501,
		0.474293,
		0.510503,
		0.545959,
		0.580504,
		0.614003,
		0.646337,
		0.6774,
		0.707107,
		0.735383,
		0.762171,
		0.787425,
		0.811114,
		0.833218,
		0.853729,
		0.872652,
		0.890001,
		0.905802,
		0.92009,
		0.932909,
		0.944313,
		0.954362,
		0.963126,
		0.970681,
		0.001133,
		0.007758,
		0.019728,
		0.036415,
		0.057235,
		0.08164,
		0.109121,
		0.139203,
		0.171447,
		0.205443,
		0.240813,
		0.277209,
		0.314308,
		0.351814,
		0.389456,
		0.426988,
		0.464184,
		0.500843,
		0.53678,
		0.571833,
		0.605858,
		0.638727,
		0.670332,
		0.700578,
		0.729388,
		0.756699,
		0.782462,
		0.806641,
		0.829215,
		0.850172,
		0.869516,
		0.88726,
		0.903427,
		0.918052,
		0.931179,
		0.942861,
		0.95316,
		0.962145,
		0.969893,
		0.976488,
		0.982017,
		0.00013,
		0.004495,
		0.014547,
		0.029626,
		0.049119,
		0.072452,
		0.099088,
		0.128533,
		0.160321,
		0.194026,
		0.229249,
		0.265625,
		0.302814,
		0.340507,
		0.378419,
		0.416289,
		0.453882,
		0.490984,
		0.527402,
		0.562964,
		0.597518,
		0.63093,
		0.663083,
		0.693878,
		0.723232,
		0.751077,
		0.777359,
		0.80204,
		0.825095,
		0.84651,
		0.866286,
		0.884434,
		0.900978,
		0.91595,
		0.929393,
		0.941362,
		0.951918,
		0.961131,
		0.969079,
		0.975845,
		0.981521,
		0.986201,
		0.989983,
		0.002059,
		0.010032,
		0.023363,
		0.041405,
		0.063557,
		0.089257,
		0.117983,
		0.149248,
		0.182603,
		0.217631,
		0.253947,
		0.291195,
		0.32905,
		0.367211,
		0.405406,
		0.443386,
		0.480925,
		0.517822,
		0.553895,
		0.588982,
		0.622942,
		0.655651,
		0.687003,
		0.71691,
		0.745299,
		0.772112,
		0.797307,
		0.820854,
		0.842738,
		0.862958,
		0.881522,
		0.898452,
		0.913781,
		0.927551,
		0.939815,
		0.950636,
		0.960083,
		0.968237,
		0.975181,
		0.981008,
		0.985814,
		0.9897,
		0.992769,
		0.995126,
		0.00053,
		0.006254,
		0.017685,
		0.034144,
		0.055,
		0.079662,
		0.107582,
		0.13825,
		0.171193,
		0.205974,
		0.242186,
		0.279458,
		0.317445,
		0.355834,
		0.394338,
		0.432694,
		0.470665,
		0.508038,
		0.544621,
		0.580244,
		0.614757,
		0.648029,
		0.679948,
		0.710418,
		0.739362,
		0.766717,
		0.792437,
		0.816488,
		0.838854,
		0.859528,
		0.878519,
		0.895847,
		0.911542,
		0.925648,
		0.938217,
		0.949311,
		0.959001,
		0.967366,
		0.974494,
		0.980477,
		0.985413,
		0.989406,
		0.992561,
		0.994985,
		0.996784,
		0.998065,
		0.003289,
		0.012657,
		0.027391,
		0.046826,
		0.070342,
		0.097363,
		0.127354,
		0.159817,
		0.194292,
		0.230353,
		0.26761,
		0.3057,
		0.344293,
		0.383087,
		0.421806,
		0.460201,
		0.498046,
		0.535139,
		0.571301,
		0.606372,
		0.640214,
		0.672708,
		0.703751,
		0.73326,
		0.761169,
		0.787425,
		0.811993,
		0.834852,
		0.855993,
		0.875423,
		0.893159,
		0.909232,
		0.923684,
		0.936566,
		0.947941,
		0.957881,
		0.966466,
		0.973783,
		0.979927,
		0.984999,
		0.989102,
		0.992346,
		0.994838,
		0.99669,
		0.998008,
		0.998897,
		0.999454,
		0.001221,
		0.008351,
		0.021206,
		0.039085,
		0.061339,
		0.08736,
		0.116587,
		0.148496,
		0.182603,
		0.218463,
		0.255662,
		0.29382,
		0.332591,
		0.371655,
		0.410723,
		0.449532,
		0.487844,
		0.525446,
		0.562148,
		0.597782,
		0.632201,
		0.665277,
		0.696903,
		0.726989,
		0.755463,
		0.782268,
		0.807365,
		0.830729,
		0.85235,
		0.87223,
		0.890386,
		0.906848,
		0.921655,
		0.93486,
		0.946526,
		0.956724,
		0.965534,
		0.973047,
		0.979358,
		0.984569,
		0.988787,
		0.992122,
		0.994686,
		0.996591,
		0.997948,
		0.998863,
		0.999438,
		0.999764,
		0.999923,
		0.00014,
		0.004845,
		0.015654,
		0.031834,
		0.052699,
		0.077611,
		0.10598,
		0.137255,
		0.170928,
		0.206529,
		0.243624,
		0.281814,
		0.320733,
		0.360045,
		0.399446,
		0.438658,
		0.47743,
		0.515539,
		0.552782,
		0.588982,
		0.623984,
		0.657651,
		0.68987,
		0.720543,
		0.749594,
		0.77696,
		0.802599,
		0.826481,
		0.848593,
		0.868936,
		0.887524,
		0.904386,
		0.91956,
		0.933098,
		0.945063,
		0.955526,
		0.964571,
		0.972286,
		0.978769,
		0.984125,
		0.988461,
		0.991891,
		0.994529,
		0.99649,
		0.997886,
		0.998829,
		0.999421,
		0.999757,
		0.999921,
		0.999984,
		0.999999,
		0.002221,
		0.010808,
		0.025132,
		0.044472,
		0.068159,
		0.095568,
		0.126123,
		0.159289,
		0.19457,
		0.231511,
		0.269691,
		0.308725,
		0.34826,
		0.387975,
		0.427577,
		0.466802,
		0.505413,
		0.543198,
		0.579968,
		0.615558,
		0.649825,
		0.682646,
		0.713918,
		0.743557,
		0.771497,
		0.79769,
		0.822103,
		0.84472,
		0.865538,
		0.884571,
		0.901844,
		0.917395,
		0.931276,
		0.94355,
		0.954288,
		0.963574,
		0.971498,
		0.978159,
		0.983664,
		0.988123,
		0.991651,
		0.994366,
		0.996384,
		0.997822,
		0.998793,
		0.999403,
		0.999749,
		0.999919,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.000573,
		0.006746,
		0.019046,
		0.036715,
		0.059048,
		0.085389,
		0.11513,
		0.147708,
		0.182603,
		0.219335,
		0.25746,
		0.296573,
		0.336303,
		0.376311,
		0.416289,
		0.455958,
		0.495067,
		0.533393,
		0.570735,
		0.606919,
		0.641793,
		0.675225,
		0.707107,
		0.737347,
		0.765874,
		0.792634,
		0.817592,
		0.840726,
		0.862033,
		0.881522,
		0.899218,
		0.915158,
		0.929393,
		0.941985,
		0.953007,
		0.962541,
		0.970681,
		0.977527,
		0.983187,
		0.987773,
		0.991403,
		0.994196,
		0.996274,
		0.997755,
		0.998756,
		0.999384,
		0.999741,
		0.999916,
		0.999983,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999923,
		0.003552,
		0.013648,
		0.029488,
		0.05033,
		0.075484,
		0.10431,
		0.136215,
		0.170651,
		0.207112,
		0.245133,
		0.284287,
		0.324181,
		0.364459,
		0.404796,
		0.444897,
		0.484498,
		0.523363,
		0.561279,
		0.598062,
		0.63355,
		0.667604,
		0.700105,
		0.730958,
		0.760084,
		0.787425,
		0.812941,
		0.836606,
		0.858415,
		0.878374,
		0.896506,
		0.912847,
		0.927446,
		0.940367,
		0.951681,
		0.961473,
		0.969836,
		0.976872,
		0.982692,
		0.987409,
		0.991144,
		0.99402,
		0.99616,
		0.997686,
		0.998717,
		0.999365,
		0.999733,
		0.999913,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999921,
		0.999764,
		0.999454,
		0.00132,
		0.009016,
		0.022856,
		0.042059,
		0.065898,
		0.093699,
		0.124837,
		0.158736,
		0.194862,
		0.232725,
		0.271875,
		0.311899,
		0.35242,
		0.393097,
		0.433618,
		0.473704,
		0.513104,
		0.551596,
		0.588982,
		0.625091,
		0.659775,
		0.692907,
		0.724384,
		0.754123,
		0.782059,
		0.808146,
		0.832357,
		0.854681,
		0.875123,
		0.893703,
		0.910457,
		0.925433,
		0.938692,
		0.950308,
		0.960366,
		0.96896,
		0.976194,
		0.982179,
		0.987032,
		0.990877,
		0.993838,
		0.996042,
		0.997614,
		0.998677,
		0.999345,
		0.999725,
		0.999911,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999983,
		0.999919,
		0.999757,
		0.999438,
		0.998897,
		0.998065,
		0.000152,
		0.005236,
		0.016893,
		0.034297,
		0.056682,
		0.083339,
		0.113609,
		0.146884,
		0.182603,
		0.22025,
		0.259348,
		0.299464,
		0.3402,
		0.381194,
		0.42212,
		0.462682,
		0.502614,
		0.541681,
		0.579674,
		0.61641,
		0.651733,
		0.685507,
		0.717621,
		0.747985,
		0.776529,
		0.803203,
		0.827973,
		0.850827,
		0.871765,
		0.890808,
		0.907987,
		0.92335,
		0.936959,
		0.948887,
		0.95922,
		0.968053,
		0.97549,
		0.981647,
		0.986641,
		0.990599,
		0.993648,
		0.995919,
		0.99754,
		0.998635,
		0.999324,
		0.999716,
		0.999908,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999916,
		0.999749,
		0.999421,
		0.998863,
		0.998008,
		0.996784,
		0.995126,
		0.002404,
		0.011678,
		0.02711,
		0.047892,
		0.073275,
		0.102567,
		0.135126,
		0.17036,
		0.207724,
		0.246719,
		0.286884,
		0.327802,
		0.369091,
		0.410404,
		0.45143,
		0.491888,
		0.531529,
		0.570132,
		0.607502,
		0.643472,
		0.677898,
		0.710662,
		0.741665,
		0.770831,
		0.798105,
		0.82345,
		0.846847,
		0.868297,
		0.887815,
		0.905432,
		0.921195,
		0.935165,
		0.947416,
		0.958032,
		0.967112,
		0.974761,
		0.981094,
		0.986235,
		0.99031,
		0.993452,
		0.995792,
		0.997462,
		0.998592,
		0.999302,
		0.999707,
		0.999905,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999913,
		0.999741,
		0.999403,
		0.998829,
		0.997948,
		0.99669,
		0.994985,
		0.992769,
		0.989983,
		0.000621,
		0.007298,
		0.02057,
		0.039587,
		0.063557,
		0.091751,
		0.123492,
		0.158156,
		0.195168,
		0.234001,
		0.27417,
		0.315234,
		0.356789,
		0.39847,
		0.439948,
		0.480925,
		0.521138,
		0.560352,
		0.598361,
		0.634987,
		0.670076,
		0.7035,
		0.735156,
		0.764959,
		0.792848,
		0.818782,
		0.842738,
		0.864714,
		0.884721,
		0.90279,
		0.918966,
		0.933308,
		0.945892,
		0.956802,
		0.966137,
		0.974004,
		0.980522,
		0.985814,
		0.990011,
		0.993247,
		0.995659,
		0.997381,
		0.998546,
		0.99928,
		0.999697,
		0.999902,
		0.99998,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999911,
		0.999733,
		0.999384,
		0.998793,
		0.997886,
		0.996591,
		0.994838,
		0.992561,
		0.9897,
		0.986201,
		0.982017,
		0.003848,
		0.014759,
		0.031834,
		0.05424,
		0.081207,
		0.112019,
		0.14602,
		0.182603,
		0.221211,
		0.261333,
		0.302502,
		0.344293,
		0.38632,
		0.428234,
		0.469722,
		0.510503,
		0.55033,
		0.588982,
		0.626271,
		0.662033,
		0.696131,
		0.728452,
		0.758906,
		0.787425,
		0.813964,
		0.838495,
		0.861011,
		0.881522,
		0.900056,
		0.916658,
		0.931385,
		0.944313,
		0.955526,
		0.965125,
		0.973219,
		0.979927,
		0.985376,
		0.9897,
		0.993035,
		0.995521,
		0.997298,
		0.998499,
		0.999256,
		0.999687,
		0.999898,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999908,
		0.999725,
		0.999365,
		0.998756,
		0.997822,
		0.99649,
		0.994686,
		0.992346,
		0.989406,
		0.985814,
		0.981521,
		0.976488,
		0.970681,
		0.001432,
		0.009762,
		0.024706,
		0.045385,
		0.070982,
		0.100747,
		0.133985,
		0.170054,
		0.208367,
		0.248386,
		0.289616,
		0.331609,
		0.373956,
		0.416289,
		0.458277,
		0.499621,
		0.540059,
		0.579359,
		0.617319,
		0.653763,
		0.688547,
		0.721547,
		0.752666,
		0.781832,
		0.80899,
		0.834112,
		0.857183,
		0.878214,
		0.897228,
		0.914268,
		0.929393,
		0.942676,
		0.954204,
		0.964076,
		0.972405,
		0.97931,
		0.984922,
		0.989376,
		0.992814,
		0.995378,
		0.99721,
		0.998451,
		0.999232,
		0.999677,
		0.999895,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.99998,
		0.999905,
		0.999716,
		0.999345,
		0.998717,
		0.997755,
		0.996384,
		0.994529,
		0.992122,
		0.989102,
		0.985413,
		0.981008,
		0.975845,
		0.969893,
		0.963126,
		0.955526,
		0.000165,
		0.005678,
		0.018285,
		0.037057,
		0.061133,
		0.08972,
		0.122084,
		0.157547,
		0.19549,
		0.235343,
		0.276585,
		0.318742,
		0.361381,
		0.404113,
		0.446587,
		0.488488,
		0.529537,
		0.569487,
		0.608124,
		0.645261,
		0.680741,
		0.714434,
		0.746234,
		0.776061,
		0.803856,
		0.829583,
		0.853227,
		0.874792,
		0.894301,
		0.911794,
		0.92733,
		0.94098,
		0.952833,
		0.962988,
		0.971559,
		0.978669,
		0.98445,
		0.98904,
		0.992584,
		0.995229,
		0.997119,
		0.9984,
		0.999206,
		0.999666,
		0.999891,
		0.999978,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999902,
		0.999707,
		0.999324,
		0.998677,
		0.997686,
		0.996274,
		0.994366,
		0.991891,
		0.988787,
		0.984999,
		0.980477,
		0.975181,
		0.969079,
		0.962145,
		0.954362,
		0.945719,
		0.936212,
		0.00261,
		0.012657,
		0.02933,
		0.05172,
		0.078987,
		0.110356,
		0.145114,
		0.182603,
		0.222223,
		0.263423,
		0.3057,
		0.348599,
		0.391707,
		0.434652,
		0.477101,
		0.518758,
		0.559361,
		0.59868,
		0.636519,
		0.672708,
		0.707107,
		0.739603,
		0.770107,
		0.798554,
		0.824905,
		0.849137,
		0.871252,
		0.891271,
		0.909232,
		0.925192,
		0.939222,
		0.95141,
		0.961859,
		0.970681,
		0.978003,
		0.983959,
		0.988691,
		0.992346,
		0.995074,
		0.997025,
		0.998347,
		0.99918,
		0.999655,
		0.999888,
		0.999977,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999898,
		0.999697,
		0.999302,
		0.998635,
		0.997614,
		0.99616,
		0.994196,
		0.991651,
		0.988461,
		0.984569,
		0.979927,
		0.974494,
		0.968237,
		0.961131,
		0.95316,
		0.944313,
		0.934586,
		0.923982,
		0.91251,
		0.000675,
		0.007921,
		0.022285,
		0.042808,
		0.068601,
		0.098846,
		0.132788,
		0.169733,
		0.209044,
		0.250141,
		0.292492,
		0.335615,
		0.379073,
		0.422472,
		0.465459,
		0.507719,
		0.548975,
		0.588982,
		0.62753,
		0.664439,
		0.699558,
		0.732765,
		0.763962,
		0.79308,
		0.820069,
		0.844907,
		0.86759,
		0.888134,
		0.906578,
		0.922975,
		0.937398,
		0.949934,
		0.960686,
		0.96977,
		0.977312,
		0.983449,
		0.988327,
		0.992097,
		0.994912,
		0.996926,
		0.998291,
		0.999152,
		0.999643,
		0.999884,
		0.999976,
		0.999998,
		1,
		0.999999,
		0.999978,
		0.999895,
		0.999687,
		0.99928,
		0.998592,
		0.99754,
		0.996042,
		0.99402,
		0.991403,
		0.988123,
		0.984125,
		0.979358,
		0.973783,
		0.967366,
		0.960083,
		0.951918,
		0.942861,
		0.932909,
		0.922066,
		0.910341,
		0.897749,
		0.88431,
		0.004182,
		0.016011,
		0.034471,
		0.058623,
		0.087601,
		0.120607,
		0.156908,
		0.19583,
		0.236757,
		0.27913,
		0.322436,
		0.366214,
		0.410046,
		0.453557,
		0.496415,
		0.538323,
		0.579023,
		0.618289,
		0.655929,
		0.691782,
		0.725714,
		0.757621,
		0.787425,
		0.815072,
		0.840533,
		0.863799,
		0.884886,
		0.903828,
		0.920678,
		0.935507,
		0.948403,
		0.959469,
		0.968823,
		0.976593,
		0.982919,
		0.987949,
		0.991839,
		0.994744,
		0.996824,
		0.998234,
		0.999123,
		0.999631,
		0.99988,
		0.999976,
		0.999998,
		1,
		0.999998,
		0.999977,
		0.999891,
		0.999677,
		0.999256,
		0.998546,
		0.997462,
		0.995919,
		0.993838,
		0.991144,
		0.987773,
		0.983664,
		0.978769,
		0.973047,
		0.966466,
		0.959001,
		0.950636,
		0.941362,
		0.931179,
		0.92009,
		0.908106,
		0.895243,
		0.881522,
		0.866967,
		0.851607,
		0.001559,
		0.010606,
		0.026791,
		0.049119,
		0.076675,
		0.108615,
		0.144161,
		0.182603,
		0.223289,
		0.265625,
		0.30907,
		0.353134,
		0.397375,
		0.441396,
		0.484843,
		0.527402,
		0.568797,
		0.608789,
		0.647171,
		0.68377,
		0.718443,
		0.751077,
		0.781584,
		0.809907,
		0.836008,
		0.859876,
		0.881522,
		0.900978,
		0.918295,
		0.933544,
		0.946813,
		0.958205,
		0.967838,
		0.975845,
		0.982367,
		0.987556,
		0.991569,
		0.994569,
		0.996717,
		0.998174,
		0.999093,
		0.999618,
		0.999876,
		0.999975,
		0.999998,
		1,
		0.999998,
		0.999976,
		0.999888,
		0.999666,
		0.999232,
		0.998499,
		0.997381,
		0.995792,
		0.993648,
		0.990877,
		0.987409,
		0.983187,
		0.978159,
		0.972286,
		0.965534,
		0.957881,
		0.949311,
		0.939815,
		0.929393,
		0.918052,
		0.905802,
		0.892661,
		0.878652,
		0.863799,
		0.848135,
		0.831692,
		0.814507,
		0.000179,
		0.006177,
		0.019856,
		0.040162,
		0.066128,
		0.096858,
		0.131531,
		0.169394,
		0.209758,
		0.251992,
		0.295525,
		0.339837,
		0.38446,
		0.428973,
		0.473,
		0.516206,
		0.558299,
		0.599022,
		0.638156,
		0.675515,
		0.710944,
		0.744321,
		0.775551,
		0.804566,
		0.831326,
		0.855814,
		0.878037,
		0.898023,
		0.915824,
		0.931507,
		0.945162,
		0.956891,
		0.966816,
		0.975068,
		0.981794,
		0.987147,
		0.991289,
		0.994386,
		0.996606,
		0.998111,
		0.999062,
		0.999605,
		0.999871,
		0.999974,
		0.999998,
		1,
		0.999998,
		0.999976,
		0.999884,
		0.999655,
		0.999206,
		0.998451,
		0.997298,
		0.995659,
		0.993452,
		0.990599,
		0.987032,
		0.982692,
		0.977527,
		0.971498,
		0.964571,
		0.956724,
		0.947941,
		0.938217,
		0.927551,
		0.91595,
		0.903427,
		0.890001,
		0.875696,
		0.860539,
		0.844563,
		0.827802,
		0.810295,
		0.792083,
		0.77321,
		0.002844,
		0.013765,
		0.031834,
		0.056024,
		0.085389,
		0.119058,
		0.156234,
		0.196187,
		0.238249,
		0.281814,
		0.326332,
		0.371306,
		0.416289,
		0.460882,
		0.504731,
		0.547521,
		0.588982,
		0.628878,
		0.667009,
		0.70321,
		0.737347,
		0.769316,
		0.799043,
		0.826481,
		0.851607,
		0.874425,
		0.89496,
		0.91326,
		0.929393,
		0.943447,
		0.955526,
		0.965752,
		0.97426,
		0.981196,
		0.98672,
		0.990997,
		0.994196,
		0.99649,
		0.998046,
		0.999029,
		0.999591,
		0.999867,
		0.999973,
		0.999998,
		1,
		0.999998,
		0.999975,
		0.99988,
		0.999643,
		0.99918,
		0.9984,
		0.99721,
		0.995521,
		0.993247,
		0.99031,
		0.986641,
		0.982179,
		0.976872,
		0.970681,
		0.963574,
		0.955526,
		0.946526,
		0.936566,
		0.925648,
		0.913781,
		0.900978,
		0.88726,
		0.872652,
		0.857183,
		0.840888,
		0.823803,
		0.805968,
		0.787425,
		0.76822,
		0.7484,
		0.728012,
		0.000737,
		0.008627,
		0.024223,
		0.046438,
		0.074267,
		0.106789,
		0.14316,
		0.182603,
		0.224414,
		0.26795,
		0.312626,
		0.357915,
		0.403344,
		0.448489,
		0.492972,
		0.53646,
		0.578663,
		0.619329,
		0.658245,
		0.695232,
		0.730146,
		0.762874,
		0.793332,
		0.821467,
		0.847251,
		0.870682,
		0.891784,
		0.9106,
		0.927198,
		0.941666,
		0.954108,
		0.964646,
		0.973418,
		0.980575,
		0.986276,
		0.990693,
		0.993998,
		0.996368,
		0.997978,
		0.998995,
		0.999576,
		0.999862,
		0.999972,
		0.999998,
		1,
		0.999998,
		0.999974,
		0.999876,
		0.999631,
		0.999152,
		0.998347,
		0.997119,
		0.995378,
		0.993035,
		0.990011,
		0.986235,
		0.981647,
		0.976194,
		0.969836,
		0.962541,
		0.954288,
		0.945063,
		0.93486,
		0.923684,
		0.911542,
		0.898452,
		0.884434,
		0.869516,
		0.853729,
		0.837108,
		0.819691,
		0.801521,
		0.782641,
		0.7631,
		0.742944,
		0.722224,
		0.700991,
		0.679298,
		0.004562,
		0.01743,
		0.03745,
		0.063557,
		0.094776,
		0.13021,
		0.169038,
		0.210511,
		0.253947,
		0.298728,
		0.344293,
		0.39014,
		0.435818,
		0.480925,
		0.525108,
		0.568056,
		0.609502,
		0.649214,
		0.687003,
		0.722712,
		0.756216,
		0.787425,
		0.816277,
		0.842738,
		0.866803,
		0.888488,
		0.907839,
		0.924919,
		0.939815,
		0.952633,
		0.963495,
		0.972543,
		0.979927,
		0.985814,
		0.990375,
		0.993791,
		0.996242,
		0.997907,
		0.99896,
		0.999561,
		0.999857,
		0.999971,
		0.999998,
		1,
		0.999998,
		0.999973,
		0.999871,
		0.999618,
		0.999123,
		0.998291,
		0.997025,
		0.995229,
		0.992814,
		0.9897,
		0.985814,
		0.981094,
		0.97549,
		0.96896,
		0.961473,
		0.953007,
		0.94355,
		0.933098,
		0.921655,
		0.909232,
		0.895847,
		0.881522,
		0.866286,
		0.850172,
		0.833218,
		0.815463,
		0.796951,
		0.777728,
		0.757843,
		0.737347,
		0.71629,
		0.694725,
		0.672708,
		0.650291,
		0.62753,
		0.001703,
		0.011564,
		0.02915,
		0.053333,
		0.083077,
		0.117431,
		0.155525,
		0.196564,
		0.239825,
		0.28465,
		0.330446,
		0.376679,
		0.422868,
		0.468589,
		0.513462,
		0.557157,
		0.599388,
		0.63991,
		0.678515,
		0.715035,
		0.749335,
		0.781314,
		0.810904,
		0.838063,
		0.86278,
		0.88507,
		0.904972,
		0.922551,
		0.937891,
		0.951098,
		0.962298,
		0.971631,
		0.979253,
		0.985332,
		0.990045,
		0.993576,
		0.99611,
		0.997833,
		0.998922,
		0.999545,
		0.999852,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999972,
		0.999867,
		0.999605,
		0.999093,
		0.998234,
		0.996926,
		0.995074,
		0.992584,
		0.989376,
		0.985376,
		0.980522,
		0.974761,
		0.968053,
		0.960366,
		0.951681,
		0.941985,
		0.931276,
		0.91956,
		0.906848,
		0.893159,
		0.878519,
		0.862958,
		0.84651,
		0.829215,
		0.811114,
		0.792253,
		0.772681,
		0.752448,
		0.731605,
		0.710206,
		0.688306,
		0.66596,
		0.643226,
		0.620158,
		0.596815,
		0.573252,
		0.000196,
		0.006746,
		0.021638,
		0.043674,
		0.071756,
		0.104874,
		0.142104,
		0.182603,
		0.225603,
		0.270407,
		0.316383,
		0.362964,
		0.409641,
		0.455958,
		0.501515,
		0.545959,
		0.588982,
		0.630323,
		0.669758,
		0.707107,
		0.742222,
		0.774992,
		0.80534,
		0.833218,
		0.858608,
		0.881522,
		0.901995,
		0.92009,
		0.93589,
		0.949502,
		0.961052,
		0.970681,
		0.97855,
		0.984829,
		0.9897,
		0.99335,
		0.995972,
		0.997755,
		0.998883,
		0.999529,
		0.999846,
		0.999969,
		0.999998,
		1,
		0.999998,
		0.999971,
		0.999862,
		0.999591,
		0.999062,
		0.998174,
		0.996824,
		0.994912,
		0.992346,
		0.98904,
		0.984922,
		0.979927,
		0.974004,
		0.967112,
		0.95922,
		0.950308,
		0.940367,
		0.929393,
		0.917395,
		0.904386,
		0.890386,
		0.875423,
		0.859528,
		0.842738,
		0.825095,
		0.806641,
		0.787425,
		0.767497,
		0.746909,
		0.725714,
		0.703969,
		0.681729,
		0.659052,
		0.635997,
		0.612621,
		0.588982,
		0.56514,
		0.541152,
		0.517077,
		0.003111,
		0.015024,
		0.034672,
		0.060886,
		0.092596,
		0.128819,
		0.168662,
		0.211307,
		0.256014,
		0.302114,
		0.349001,
		0.396135,
		0.443032,
		0.489264,
		0.534454,
		0.578275,
		0.620445,
		0.660725,
		0.698919,
		0.734868,
		0.76845,
		0.799577,
		0.828196,
		0.854281,
		0.877839,
		0.898903,
		0.917532,
		0.93381,
		0.947842,
		0.959754,
		0.969692,
		0.977817,
		0.984305,
		0.98934,
		0.993115,
		0.995829,
		0.997675,
		0.998843,
		0.999511,
		0.999841,
		0.999968,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999857,
		0.999576,
		0.999029,
		0.998111,
		0.996717,
		0.994744,
		0.992097,
		0.988691,
		0.98445,
		0.97931,
		0.973219,
		0.966137,
		0.958032,
		0.948887,
		0.938692,
		0.927446,
		0.915158,
		0.901844,
		0.887524,
		0.87223,
		0.855993,
		0.838854,
		0.820854,
		0.80204,
		0.782462,
		0.762171,
		0.741222,
		0.71967,
		0.697574,
		0.67499,
		0.651979,
		0.628601,
		0.604914,
		0.58098,
		0.556859,
		0.532609,
		0.508291,
		0.483964,
		0.459685,
		0.000807,
		0.009431,
		0.026425,
		0.050548,
		0.08066,
		0.11572,
		0.154777,
		0.196963,
		0.241492,
		0.287651,
		0.334797,
		0.382355,
		0.429809,
		0.476704,
		0.522638,
		0.567259,
		0.610268,
		0.651406,
		0.690463,
		0.727265,
		0.76168,
		0.793609,
		0.82299,
		0.849792,
		0.874016,
		0.895691,
		0.914874,
		0.931646,
		0.946113,
		0.958402,
		0.968661,
		0.977053,
		0.983758,
		0.988964,
		0.99287,
		0.995678,
		0.99759,
		0.9988,
		0.999493,
		0.999835,
		0.999966,
		0.999998,
		1,
		0.999998,
		0.999969,
		0.999852,
		0.999561,
		0.998995,
		0.998046,
		0.996606,
		0.994569,
		0.991839,
		0.988327,
		0.983959,
		0.978669,
		0.972405,
		0.965125,
		0.956802,
		0.947416,
		0.936959,
		0.925433,
		0.912847,
		0.899218,
		0.884571,
		0.868936,
		0.85235,
		0.834852,
		0.816488,
		0.797307,
		0.777359,
		0.756699,
		0.735383,
		0.713469,
		0.691017,
		0.668086,
		0.644738,
		0.621034,
		0.597036,
		0.572806,
		0.548406,
		0.523896,
		0.499339,
		0.474793,
		0.45032,
		0.425977,
		0.401822,
		0.004996,
		0.019046,
		0.04083,
		0.069138,
		0.102863,
		0.140991,
		0.182603,
		0.226862,
		0.273008,
		0.32036,
		0.368303,
		0.416289,
		0.463832,
		0.510503,
		0.555928,
		0.599783,
		0.641793,
		0.681729,
		0.719404,
		0.754672,
		0.787425,
		0.817592,
		0.845133,
		0.870045,
		0.892353,
		0.912108,
		0.929393,
		0.944313,
		0.956994,
		0.967586,
		0.976257,
		0.983187,
		0.988571,
		0.992614,
		0.995521,
		0.997501,
		0.998756,
		0.999474,
		0.999829,
		0.999965,
		0.999998,
		1,
		0.999998,
		0.999968,
		0.999846,
		0.999545,
		0.99896,
		0.997978,
		0.99649,
		0.994386,
		0.991569,
		0.987949,
		0.983449,
		0.978003,
		0.971559,
		0.964076,
		0.955526,
		0.945892,
		0.935165,
		0.92335,
		0.910457,
		0.896506,
		0.881522,
		0.865538,
		0.848593,
		0.830729,
		0.811993,
		0.792437,
		0.772112,
		0.751077,
		0.729388,
		0.707107,
		0.684294,
		0.661012,
		0.637323,
		0.613292,
		0.588982,
		0.564456,
		0.539778,
		0.515011,
		0.490217,
		0.465459,
		0.440796,
		0.416289,
		0.391998,
		0.36798,
		0.344293,
		0.001868,
		0.012657,
		0.031834,
		0.058111,
		0.090309,
		0.127354,
		0.168264,
		0.212149,
		0.258204,
		0.3057,
		0.353984,
		0.402472,
		0.450646,
		0.498046,
		0.544274,
		0.588982,
		0.631876,
		0.672708,
		0.711275,
		0.747418,
		0.781018,
		0.811993,
		0.840298,
		0.865921,
		0.888882,
		0.909232,
		0.927049,
		0.942438,
		0.955526,
		0.966466,
		0.975425,
		0.982591,
		0.988161,
		0.992346,
		0.995357,
		0.997409,
		0.998709,
		0.999454,
		0.999822,
		0.999964,
		0.999998,
		1,
		0.999998,
		0.999966,
		0.999841,
		0.999529,
		0.998922,
		0.997907,
		0.996368,
		0.994196,
		0.991289,
		0.987556,
		0.982919,
		0.977312,
		0.970681,
		0.962988,
		0.954204,
		0.944313,
		0.933308,
		0.921195,
		0.907987,
		0.893703,
		0.878374,
		0.862033,
		0.84472,
		0.826481,
		0.807365,
		0.787425,
		0.766717,
		0.745299,
		0.723232,
		0.700578,
		0.6774,
		0.653763,
		0.629732,
		0.605372,
		0.580749,
		0.555928,
		0.530974,
		0.505952,
		0.480925,
		0.455958,
		0.431113,
		0.40645,
		0.382031,
		0.357915,
		0.33416,
		0.310823,
		0.28796,
		0.000216,
		0.007397,
		0.02367,
		0.047667,
		0.078131,
		0.113919,
		0.153986,
		0.197385,
		0.243258,
		0.29083,
		0.339405,
		0.38836,
		0.437142,
		0.485261,
		0.53229,
		0.577857,
		0.621646,
		0.663389,
		0.702868,
		0.739908,
		0.774379,
		0.806187,
		0.835279,
		0.861636,
		0.885274,
		0.90624,
		0.924608,
		0.940484,
		0.953996,
		0.965297,
		0.974557,
		0.981968,
		0.987733,
		0.992065,
		0.995185,
		0.997312,
		0.99866,
		0.999434,
		0.999815,
		0.999962,
		0.999998,
		1,
		0.999998,
		0.999965,
		0.999835,
		0.999511,
		0.998883,
		0.997833,
		0.996242,
		0.993998,
		0.990997,
		0.987147,
		0.982367,
		0.976593,
		0.96977,
		0.961859,
		0.952833,
		0.942676,
		0.931385,
		0.918966,
		0.905432,
		0.890808,
		0.875123,
		0.858415,
		0.840726,
		0.822103,
		0.802599,
		0.782268,
		0.761169,
		0.739362,
		0.71691,
		0.693878,
		0.670332,
		0.646337,
		0.62196,
		0.59727,
		0.572334,
		0.547218,
		0.52199,
		0.496716,
		0.471461,
		0.446292,
		0.42127,
		0.39646,
		0.371923,
		0.34772,
		0.32391,
		0.300551,
		0.2777,
		0.255412,
		0.233741,
		0.003417,
		0.016464,
		0.037906,
		0.066407,
		0.100747,
		0.139816,
		0.182603,
		0.228195,
		0.275766,
		0.324575,
		0.373956,
		0.42332,
		0.472143,
		0.519969,
		0.566399,
		0.611093,
		0.653763,
		0.694173,
		0.732133,
		0.767497,
		0.800163,
		0.830067,
		0.857183,
		0.881522,
		0.903125,
		0.922066,
		0.938448,
		0.952401,
		0.964076,
		0.973651,
		0.981318,
		0.987285,
		0.991772,
		0.995005,
		0.99721,
		0.998609,
		0.999412,
		0.999808,
		0.999961,
		0.999997,
		1,
		0.999998,
		0.999964,
		0.999829,
		0.999493,
		0.998843,
		0.997755,
		0.99611,
		0.993791,
		0.990693,
		0.98672,
		0.981794,
		0.975845,
		0.968823,
		0.960686,
		0.95141,
		0.94098,
		0.929393,
		0.916658,
		0.90279,
		0.887815,
		0.871765,
		0.854681,
		0.836606,
		0.817592,
		0.79769,
		0.77696,
		0.755463,
		0.73326,
		0.710418,
		0.687003,
		0.663083,
		0.638727,
		0.614003,
		0.588982,
		0.563732,
		0.538323,
		0.512824,
		0.487302,
		0.461824,
		0.436458,
		0.411269,
		0.38632,
		0.361676,
		0.337397,
		0.313545,
		0.290179,
		0.267356,
		0.245133,
		0.223565,
		0.202704,
		0.182603,
		0.000888,
		0.010354,
		0.028941,
		0.055226,
		0.08791,
		0.125807,
		0.167843,
		0.213043,
		0.260527,
		0.309504,
		0.359265,
		0.409179,
		0.45869,
		0.507305,
		0.5546,
		0.600208,
		0.643821,
		0.68518,
		0.724081,
		0.760363,
		0.793912,
		0.824654,
		0.852555,
		0.877618,
		0.899882,
		0.919418,
		0.936326,
		0.950736,
		0.962803,
		0.972704,
		0.980638,
		0.986817,
		0.991466,
		0.994817,
		0.997104,
		0.998556,
		0.999389,
		0.9998,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999962,
		0.999822,
		0.999474,
		0.9988,
		0.997675,
		0.995972,
		0.993576,
		0.990375,
		0.986276,
		0.981196,
		0.975068,
		0.967838,
		0.959469,
		0.949934,
		0.939222,
		0.92733,
		0.914268,
		0.900056,
		0.884721,
		0.868297,
		0.850827,
		0.832357,
		0.812941,
		0.792634,
		0.771497,
		0.749594,
		0.726989,
		0.703751,
		0.679948,
		0.655651,
		0.63093,
		0.605858,
		0.580504,
		0.554942,
		0.529241,
		0.503473,
		0.477707,
		0.452013,
		0.426457,
		0.401108,
		0.376031,
		0.351291,
		0.32695,
		0.303071,
		0.279714,
		0.256937,
		0.234798,
		0.213353,
		0.192655,
		0.172758,
		0.153712,
		0.135567,
		0.005495,
		0.020897,
		0.044689,
		0.075484,
		0.112019,
		0.153149,
		0.197833,
		0.245133,
		0.294205,
		0.344293,
		0.394723,
		0.444897,
		0.494292,
		0.542452,
		0.588982,
		0.63355,
		0.675879,
		0.715743,
		0.752968,
		0.787425,
		0.819031,
		0.847743,
		0.873557,
		0.896506,
		0.916658,
		0.934112,
		0.948998,
		0.961473,
		0.971715,
		0.979927,
		0.986327,
		0.991144,
		0.99462,
		0.996992,
		0.998499,
		0.999365,
		0.999793,
		0.999958,
		0.999997,
		1,
		0.999997,
		0.999961,
		0.999815,
		0.999454,
		0.998756,
		0.99759,
		0.995829,
		0.99335,
		0.990045,
		0.985814,
		0.980575,
		0.97426,
		0.966816,
		0.958205,
		0.948403,
		0.937398,
		0.925192,
		0.911794,
		0.897228,
		0.881522,
		0.864714,
		0.846847,
		0.827973,
		0.808146,
		0.787425,
		0.765874,
		0.743557,
		0.720543,
		0.696903,
		0.672708,
		0.648029,
		0.622942,
		0.597518,
		0.571833,
		0.545959,
		0.519969,
		0.493936,
		0.467931,
		0.442026,
		0.416289,
		0.39079,
		0.365596,
		0.340772,
		0.316383,
		0.292492,
		0.269161,
		0.246449,
		0.224414,
		0.203114,
		0.182603,
		0.162935,
		0.144161,
		0.126332,
		0.109496,
		0.093699,
		0.002059,
		0.013913,
		0.034906,
		0.063557,
		0.098521,
		0.138573,
		0.182603,
		0.229612,
		0.278696,
		0.32905,
		0.379951,
		0.430764,
		0.480925,
		0.529947,
		0.577405,
		0.622942,
		0.666257,
		0.707107,
		0.745299,
		0.780692,
		0.813189,
		0.842738,
		0.869329,
		0.892988,
		0.913781,
		0.931803,
		0.947185,
		0.960083,
		0.970681,
		0.979184,
		0.985814,
		0.990808,
		0.994413,
		0.996876,
		0.998441,
		0.99934,
		0.999784,
		0.999956,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.999808,
		0.999434,
		0.998709,
		0.997501,
		0.995678,
		0.993115,
		0.9897,
		0.985332,
		0.979927,
		0.973418,
		0.965752,
		0.956891,
		0.946813,
		0.935507,
		0.922975,
		0.909232,
		0.894301,
		0.878214,
		0.861011,
		0.842738,
		0.82345,
		0.803203,
		0.782059,
		0.760084,
		0.737347,
		0.713918,
		0.68987,
		0.665277,
		0.640214,
		0.614757,
		0.588982,
		0.562964,
		0.53678,
		0.510503,
		0.48421,
		0.457972,
		0.431863,
		0.405954,
		0.380315,
		0.355015,
		0.330122,
		0.3057,
		0.281814,
		0.258527,
		0.2359,
		0.213992,
		0.19286,
		0.172561,
		0.153149,
		0.134676,
		0.117192,
		0.100747,
		0.085389,
		0.071162,
		0.058111,
		0.000238,
		0.008146,
		0.026003,
		0.05223,
		0.085389,
		0.124172,
		0.167396,
		0.213992,
		0.262996,
		0.313545,
		0.364871,
		0.416289,
		0.4672,
		0.517077,
		0.565468,
		0.611985,
		0.656305,
		0.698162,
		0.737347,
		0.773702,
		0.807118,
		0.837533,
		0.864928,
		0.889324,
		0.91078,
		0.929393,
		0.945291,
		0.958631,
		0.9696,
		0.978406,
		0.985277,
		0.990456,
		0.994196,
		0.996753,
		0.998379,
		0.999313,
		0.999775,
		0.999954,
		0.999997,
		1,
		0.999997,
		0.999958,
		0.9998,
		0.999412,
		0.99866,
		0.997409,
		0.995521,
		0.99287,
		0.98934,
		0.984829,
		0.979253,
		0.972543,
		0.964646,
		0.955526,
		0.945162,
		0.933544,
		0.920678,
		0.906578,
		0.891271,
		0.874792,
		0.857183,
		0.838495,
		0.818782,
		0.798105,
		0.776529,
		0.754123,
		0.730958,
		0.707107,
		0.682646,
		0.657651,
		0.632201,
		0.606372,
		0.580244,
		0.553895,
		0.527402,
		0.500843,
		0.474293,
		0.447829,
		0.421525,
		0.395453,
		0.369686,
		0.344293,
		0.319344,
		0.294906,
		0.271043,
		0.247821,
		0.2253,
		0.203541,
		0.182603,
		0.162543,
		0.143415,
		0.125272,
		0.108167,
		0.092147,
		0.077262,
		0.063557,
		0.051077,
		0.039864,
		0.029959,
		0.003771,
		0.018122,
		0.041614,
		0.07271,
		0.110014,
		0.152262,
		0.198309,
		0.247127,
		0.297795,
		0.349487,
		0.401474,
		0.453111,
		0.503836,
		0.553161,
		0.600669,
		0.64601,
		0.688896,
		0.729099,
		0.766443,
		0.800807,
		0.832117,
		0.860344,
		0.885504,
		0.907651,
		0.926878,
		0.943313,
		0.957114,
		0.968469,
		0.977591,
		0.984714,
		0.990087,
		0.993969,
		0.996625,
		0.998314,
		0.999285,
		0.999766,
		0.999952,
		0.999997,
		1,
		0.999997,
		0.999956,
		0.999793,
		0.999389,
		0.998609,
		0.997312,
		0.995357,
		0.992614,
		0.988964,
		0.984305,
		0.97855,
		0.971631,
		0.963495,
		0.954108,
		0.943447,
		0.931507,
		0.918295,
		0.903828,
		0.888134,
		0.871252,
		0.853227,
		0.834112,
		0.813964,
		0.792848,
		0.770831,
		0.747985,
		0.724384,
		0.700105,
		0.675225,
		0.649825,
		0.623984,
		0.597782,
		0.571301,
		0.544621,
		0.517822,
		0.490984,
		0.464184,
		0.437501,
		0.41101,
		0.384786,
		0.358903,
		0.333432,
		0.308444,
		0.284006,
		0.260186,
		0.237049,
		0.214658,
		0.193074,
		0.172357,
		0.152563,
		0.13375,
		0.11597,
		0.099276,
		0.083718,
		0.069344,
		0.0562,
		0.044332,
		0.033782,
		0.024591,
		0.0168,
		0.010445,
		0.000982,
		0.011419,
		0.031834,
		0.060583,
		0.096175,
		0.137255,
		0.182603,
		0.231119,
		0.281814,
		0.333809,
		0.38632,
		0.438658,
		0.490217,
		0.540475,
		0.588982,
		0.63536,
		0.679298,
		0.720543,
		0.758906,
		0.794247,
		0.826481,
		0.85557,
		0.881522,
		0.904386,
		0.924251,
		0.941245,
		0.955526,
		0.967285,
		0.976739,
		0.984125,
		0.9897,
		0.993731,
		0.99649,
		0.998246,
		0.999256,
		0.999757,
		0.99995,
		0.999997,
		1,
		0.999997,
		0.999954,
		0.999784,
		0.999365,
		0.998556,
		0.99721,
		0.995185,
		0.992346,
		0.988571,
		0.983758,
		0.977817,
		0.970681,
		0.962298,
		0.952633,
		0.941666,
		0.929393,
		0.915824,
		0.900978,
		0.884886,
		0.86759,
		0.849137,
		0.829583,
		0.80899,
		0.787425,
		0.764959,
		0.741665,
		0.717621,
		0.692907,
		0.667604,
		0.641793,
		0.615558,
		0.588982,
		0.562148,
		0.535139,
		0.508038,
		0.480925,
		0.453882,
		0.426988,
		0.40032,
		0.373956,
		0.34797,
		0.322436,
		0.297425,
		0.273008,
		0.249252,
		0.226223,
		0.203986,
		0.182603,
		0.162135,
		0.142639,
		0.124172,
		0.106789,
		0.090543,
		0.075484,
		0.06166,
		0.049119,
		0.037906,
		0.028064,
		0.019635,
		0.012657,
		0.00717,
		0.003208,
		0.000807,
		0.006072,
		0.023032,
		0.049119,
		0.082738,
		0.122442,
		0.166922,
		0.215002,
		0.265625,
		0.317848,
		0.370832,
		0.423837,
		0.476215,
		0.527402,
		0.576915,
		0.624345,
		0.669354,
		0.711668,
		0.751077,
		0.787425,
		0.820615,
		0.850596,
		0.877369,
		0.900978,
		0.921508,
		0.939084,
		0.953866,
		0.966046,
		0.975845,
		0.983507,
		0.989293,
		0.99348,
		0.996348,
		0.998174,
		0.999225,
		0.999747,
		0.999948,
		0.999997,
		1,
		0.999997,
		0.999952,
		0.999775,
		0.99934,
		0.998499,
		0.997104,
		0.995005,
		0.992065,
		0.988161,
		0.983187,
		0.977053,
		0.969692,
		0.961052,
		0.951098,
		0.939815,
		0.927198,
		0.91326,
		0.898023,
		0.881522,
		0.863799,
		0.844907,
		0.824905,
		0.803856,
		0.781832,
		0.758906,
		0.735156,
		0.710662,
		0.685507,
		0.659775,
		0.63355,
		0.606919,
		0.579968,
		0.552782,
		0.525446,
		0.498046,
		0.470665,
		0.443386,
		0.416289,
		0.389456,
		0.362964,
		0.336891,
		0.31131,
		0.286296,
		0.261919,
		0.238249,
		0.215353,
		0.193297,
		0.172143,
		0.151954,
		0.132788,
		0.114702,
		0.097753,
		0.081992,
		0.067471,
		0.05424,
		0.042346,
		0.031834,
		0.022747,
		0.015128,
		0.009016,
		0.004448,
		0.001462,
		9.2e-005,
		0.00228,
		0.015366,
		0.038444,
		0.069803,
		0.107895,
		0.151319,
		0.198815,
		0.249252,
		0.301618,
		0.355015,
		0.40865,
		0.461824,
		0.513933,
		0.564456,
		0.612952,
		0.659052,
		0.702461,
		0.742944,
		0.78033,
		0.814507,
		0.845412,
		0.873037,
		0.89742,
		0.918641,
		0.936823,
		0.952128,
		0.964748,
		0.974909,
		0.982859,
		0.988867,
		0.993217,
		0.996199,
		0.998099,
		0.999193,
		0.999736,
		0.999946,
		0.999997,
		1,
		0.999997,
		0.99995,
		0.999766,
		0.999313,
		0.998441,
		0.996992,
		0.994817,
		0.991772,
		0.987733,
		0.982591,
		0.976257,
		0.968661,
		0.959754,
		0.949502,
		0.937891,
		0.924919,
		0.9106,
		0.89496,
		0.878037,
		0.859876,
		0.840533,
		0.820069,
		0.798554,
		0.776061,
		0.752666,
		0.728452,
		0.7035,
		0.677898,
		0.651733,
		0.625091,
		0.598062,
		0.570735,
		0.543198,
		0.515539,
		0.487844,
		0.460201,
		0.432694,
		0.405406,
		0.378419,
		0.351814,
		0.325668,
		0.300059,
		0.275061,
		0.250747,
		0.227188,
		0.204451,
		0.182603,
		0.16171,
		0.141832,
		0.12303,
		0.105362,
		0.088884,
		0.073649,
		0.05971,
		0.047116,
		0.035914,
		0.026151,
		0.01787,
		0.011113,
		0.00592,
		0.002328,
		0.000375,
		0.000264,
		0.009016,
		0.028698,
		0.057479,
		0.093699,
		0.135857,
		0.182603,
		0.232725,
		0.285138,
		0.33888,
		0.393097,
		0.447041,
		0.500061,
		0.551596,
		0.601168,
		0.64838,
		0.692907,
		0.734494,
		0.77295,
		0.808146,
		0.840008,
		0.868517,
		0.893703,
		0.915644,
		0.934459,
		0.950308,
		0.963389,
		0.973927,
		0.982179,
		0.98842,
		0.992941,
		0.996042,
		0.998019,
		0.999159,
		0.999725,
		0.999944,
		0.999996,
		1,
		0.999997,
		0.999948,
		0.999757,
		0.999285,
		0.998379,
		0.996876,
		0.99462,
		0.991466,
		0.987285,
		0.981968,
		0.975425,
		0.967586,
		0.958402,
		0.947842,
		0.93589,
		0.922551,
		0.907839,
		0.891784,
		0.874425,
		0.855814,
		0.836008,
		0.815072,
		0.79308,
		0.770107,
		0.746234,
		0.721547,
		0.696131,
		0.670076,
		0.643472,
		0.61641,
		0.588982,
		0.561279,
		0.533393,
		0.505413,
		0.47743,
		0.449532,
		0.421806,
		0.394338,
		0.367211,
		0.340507,
		0.314308,
		0.28869,
		0.26373,
		0.239503,
		0.216079,
		0.193529,
		0.171921,
		0.151319,
		0.131787,
		0.113386,
		0.096175,
		0.080209,
		0.065543,
		0.05223,
		0.04032,
		0.02986,
		0.020897,
		0.013475,
		0.007634,
		0.003417,
		0.00086,
		0.004182,
		0.020042,
		0.045892,
		0.079949,
		0.120607,
		0.166417,
		0.216079,
		0.26843,
		0.322436,
		0.377182,
		0.431863,
		0.485779,
		0.538323,
		0.588982,
		0.637323,
		0.682994,
		0.725714,
		0.765272,
		0.801521,
		0.834373,
		0.863799,
		0.889821,
		0.91251,
		0.931984,
		0.948403,
		0.961963,
		0.972897,
		0.981465,
		0.987949,
		0.992651,
		0.995877,
		0.997936,
		0.999123,
		0.999713,
		0.999941,
		0.999996,
		1,
		0.999996,
		0.999946,
		0.999747,
		0.999256,
		0.998314,
		0.996753,
		0.994413,
		0.991144,
		0.986817,
		0.981318,
		0.974557,
		0.966466,
		0.956994,
		0.946113,
		0.93381,
		0.92009,
		0.904972,
		0.888488,
		0.870682,
		0.851607,
		0.831326,
		0.809907,
		0.787425,
		0.763962,
		0.739603,
		0.714434,
		0.688547,
		0.662033,
		0.634987,
		0.607502,
		0.579674,
		0.551596,
		0.523363,
		0.495067,
		0.466802,
		0.438658,
		0.410723,
		0.383087,
		0.355834,
		0.32905,
		0.302814,
		0.277209,
		0.252311,
		0.228195,
		0.204936,
		0.182603,
		0.161266,
		0.140991,
		0.121842,
		0.103881,
		0.087167,
		0.071756,
		0.057705,
		0.045066,
		0.033889,
		0.024223,
		0.016114,
		0.009606,
		0.004741,
		0.001559,
		9.8e-005,
		0.001092,
		0.012657,
		0.035182,
		0.066755,
		0.105651,
		0.150317,
		0.199356,
		0.25152,
		0.3057,
		0.360911,
		0.416289,
		0.471079,
		0.524629,
		0.576381,
		0.625869,
		0.672708,
		0.71659,
		0.757282,
		0.794618,
		0.828496,
		0.858873,
		0.885764,
		0.909232,
		0.929393,
		0.946406,
		0.960468,
		0.971816,
		0.980715,
		0.987455,
		0.992346,
		0.995704,
		0.997848,
		0.999086,
		0.9997,
		0.999939,
		0.999996,
		1,
		0.999996,
		0.999944,
		0.999736,
		0.999225,
		0.998246,
		0.996625,
		0.994196,
		0.990808,
		0.986327,
		0.980638,
		0.973651,
		0.965297,
		0.955526,
		0.944313,
		0.931646,
		0.917532,
		0.901995,
		0.88507,
		0.866803,
		0.847251,
		0.826481,
		0.804566,
		0.781584,
		0.757621,
		0.732765,
		0.707107,
		0.680741,
		0.653763,
		0.626271,
		0.598361,
		0.570132,
		0.541681,
		0.513104,
		0.484498,
		0.455958,
		0.427577,
		0.399446,
		0.371655,
		0.344293,
		0.317445,
		0.291195,
		0.265625,
		0.240813,
		0.216838,
		0.193772,
		0.171689,
		0.150658,
		0.130747,
		0.112019,
		0.094539,
		0.078366,
		0.063557,
		0.050169,
		0.038254,
		0.027864,
		0.019046,
		0.011847,
		0.006313,
		0.002483,
		0.0004,
		0.006746,
		0.02551,
		0.05424,
		0.091084,
		0.134371,
		0.182603,
		0.234441,
		0.28869,
		0.344293,
		0.40032,
		0.455958,
		0.510503,
		0.563354,
		0.614003,
		0.662033,
		0.707107,
		0.748966,
		0.787425,
		0.822365,
		0.853729,
		0.881522,
		0.905802,
		0.92668,
		0.944313,
		0.9589,
		0.970681,
		0.979927,
		0.986936,
		0.992024,
		0.995521,
		0.997755,
		0.999046,
		0.999687,
		0.999936,
		0.999996,
		1,
		0.999996,
		0.999941,
		0.999725,
		0.999193,
		0.998174,
		0.99649,
		0.993969,
		0.990456,
		0.985814,
		0.979927,
		0.972704,
		0.964076,
		0.953996,
		0.942438,
		0.929393,
		0.914874,
		0.898903,
		0.881522,
		0.86278,
		0.842738,
		0.821467,
		0.799043,
		0.775551,
		0.751077,
		0.725714,
		0.699558,
		0.672708,
		0.645261,
		0.617319,
		0.588982,
		0.560352,
		0.531529,
		0.502614,
		0.473704,
		0.444897,
		0.416289,
		0.387975,
		0.360045,
		0.332591,
		0.3057,
		0.279458,
		0.253947,
		0.229249,
		0.205443,
		0.182603,
		0.160804,
		0.140116,
		0.120607,
		0.102344,
		0.085389,
		0.069803,
		0.055645,
		0.042971,
		0.031834,
		0.022285,
		0.014374,
		0.008146,
		0.003647,
		0.000918,
		0.002539,
		0.017058,
		0.042546,
		0.077011,
		0.118659,
		0.165879,
		0.21723,
		0.271429,
		0.32734,
		0.383959,
		0.440412,
		0.495938,
		0.549888,
		0.601712,
		0.650955,
		0.69725,
		0.74031,
		0.779928,
		0.815966,
		0.848355,
		0.877086,
		0.902212,
		0.923837,
		0.942118,
		0.957254,
		0.969489,
		0.979099,
		0.986389,
		0.991686,
		0.995329,
		0.997658,
		0.999004,
		0.999673,
		0.999933,
		0.999996,
		1,
		0.999996,
		0.999939,
		0.999713,
		0.999159,
		0.998099,
		0.996348,
		0.993731,
		0.990087,
		0.985277,
		0.979184,
		0.971715,
		0.962803,
		0.952401,
		0.940484,
		0.927049,
		0.912108,
		0.895691,
		0.877839,
		0.858608,
		0.838063,
		0.816277,
		0.793332,
		0.769316,
		0.744321,
		0.718443,
		0.691782,
		0.664439,
		0.636519,
		0.608124,
		0.579359,
		0.55033,
		0.521138,
		0.491888,
		0.462682,
		0.433618,
		0.404796,
		0.376311,
		0.34826,
		0.320733,
		0.29382,
		0.26761,
		0.242186,
		0.217631,
		0.194026,
		0.171447,
		0.149968,
		0.129662,
		0.110599,
		0.092843,
		0.076461,
		0.061512,
		0.048056,
		0.03615,
		0.025848,
		0.0172,
		0.010256,
		0.005063,
		0.001665,
		0.000105,
		0.000295,
		0.010032,
		0.031834,
		0.063557,
		0.103273,
		0.149248,
		0.199933,
		0.253947,
		0.310066,
		0.367211,
		0.424437,
		0.480925,
		0.535972,
		0.588982,
		0.63946,
		0.687003,
		0.731298,
		0.772112,
		0.809287,
		0.842738,
		0.872446,
		0.898452,
		0.920857,
		0.939815,
		0.955526,
		0.968237,
		0.978228,
		0.985814,
		0.99133,
		0.995126,
		0.997555,
		0.99896,
		0.999658,
		0.99993,
		0.999995,
		1,
		0.999996,
		0.999936,
		0.9997,
		0.999123,
		0.998019,
		0.996199,
		0.99348,
		0.9897,
		0.984714,
		0.978406,
		0.970681,
		0.961473,
		0.950736,
		0.938448,
		0.924608,
		0.909232,
		0.892353,
		0.874016,
		0.854281,
		0.833218,
		0.810904,
		0.787425,
		0.762874,
		0.737347,
		0.710944,
		0.68377,
		0.655929,
		0.62753,
		0.59868,
		0.569487,
		0.540059,
		0.510503,
		0.480925,
		0.45143,
		0.42212,
		0.393097,
		0.364459,
		0.336303,
		0.308725,
		0.281814,
		0.255662,
		0.230353,
		0.205974,
		0.182603,
		0.160321,
		0.139203,
		0.119322,
		0.100747,
		0.083548,
		0.067787,
		0.053528,
		0.04083,
		0.02975,
		0.020342,
		0.012657,
		0.006746,
		0.002655,
		0.000428,
		0.004665,
		0.022285,
		0.050862,
		0.088318,
		0.132788,
		0.182603,
		0.236278,
		0.292492,
		0.350084,
		0.408035,
		0.465459,
		0.521595,
		0.575799,
		0.62753,
		0.676351,
		0.721915,
		0.763962,
		0.802314,
		0.836868,
		0.86759,
		0.894513,
		0.917733,
		0.937398,
		0.953711,
		0.96692,
		0.977312,
		0.985208,
		0.990954,
		0.994912,
		0.997446,
		0.998913,
		0.999643,
		0.999927,
		0.999995,
		1,
		0.999995,
		0.999933,
		0.999687,
		0.999086,
		0.997936,
		0.996042,
		0.993217,
		0.989293,
		0.984125,
		0.977591,
		0.9696,
		0.960083,
		0.948998,
		0.936326,
		0.922066,
		0.90624,
		0.888882,
		0.870045,
		0.849792,
		0.828196,
		0.80534,
		0.781314,
		0.756216,
		0.730146,
		0.70321,
		0.675515,
		0.647171,
		0.618289,
		0.588982,
		0.559361,
		0.529537,
		0.499621,
		0.469722,
		0.439948,
		0.410404,
		0.381194,
		0.35242,
		0.324181,
		0.296573,
		0.269691,
		0.243624,
		0.218463,
		0.194292,
		0.171193,
		0.149248,
		0.128533,
		0.109121,
		0.091084,
		0.07449,
		0.059405,
		0.045892,
		0.034009,
		0.023816,
		0.015366,
		0.008711,
		0.003901,
		0.000982,
		0.001221,
		0.014108,
		0.039085,
		0.073915,
		0.116587,
		0.165303,
		0.218463,
		0.274643,
		0.332591,
		0.391206,
		0.449532,
		0.506746,
		0.562148,
		0.615152,
		0.665277,
		0.712144,
		0.755463,
		0.795032,
		0.830729,
		0.862506,
		0.890386,
		0.914455,
		0.93486,
		0.951803,
		0.965534,
		0.976347,
		0.984569,
		0.990558,
		0.994686,
		0.997332,
		0.998863,
		0.999626,
		0.999923,
		0.999995,
		1,
		0.999995,
		0.99993,
		0.999673,
		0.999046,
		0.997848,
		0.995877,
		0.992941,
		0.988867,
		0.983507,
		0.976739,
		0.968469,
		0.958631,
		0.947185,
		0.934112,
		0.919418,
		0.903125,
		0.885274,
		0.865921,
		0.845133,
		0.82299,
		0.799577,
		0.774992,
		0.749335,
		0.722712,
		0.695232,
		0.667009,
		0.638156,
		0.608789,
		0.579023,
		0.548975,
		0.518758,
		0.488488,
		0.458277,
		0.428234,
		0.39847,
		0.369091,
		0.3402,
		0.311899,
		0.284287,
		0.25746,
		0.231511,
		0.206529,
		0.182603,
		0.159817,
		0.13825,
		0.117983,
		0.099088,
		0.08164,
		0.065706,
		0.051353,
		0.038645,
		0.027641,
		0.018399,
		0.010974,
		0.005419,
		0.001783,
		0.000112,
		0.007538,
		0.02841,
		0.060202,
		0.100747,
		0.148107,
		0.200552,
		0.25655,
		0.314747,
		0.373956,
		0.433143,
		0.491417,
		0.548017,
		0.602308,
		0.653763,
		0.701966,
		0.746597,
		0.787425,
		0.824309,
		0.857183,
		0.886059,
		0.911016,
		0.932195,
		0.949797,
		0.964076,
		0.975331,
		0.983896,
		0.990141,
		0.994448,
		0.99721,
		0.998811,
		0.999609,
		0.99992,
		0.999995,
		1,
		0.999995,
		0.999927,
		0.999658,
		0.999004,
		0.997755,
		0.995704,
		0.992651,
		0.98842,
		0.982859,
		0.975845,
		0.967285,
		0.957114,
		0.945291,
		0.931803,
		0.916658,
		0.899882,
		0.881522,
		0.861636,
		0.840298,
		0.817592,
		0.793609,
		0.76845,
		0.742222,
		0.715035,
		0.687003,
		0.658245,
		0.628878,
		0.599022,
		0.568797,
		0.538323,
		0.507719,
		0.477101,
		0.446587,
		0.416289,
		0.38632,
		0.356789,
		0.327802,
		0.299464,
		0.271875,
		0.245133,
		0.219335,
		0.19457,
		0.170928,
		0.148496,
		0.127354,
		0.107582,
		0.089257,
		0.072452,
		0.057235,
		0.043674,
		0.031834,
		0.021774,
		0.013553,
		0.007225,
		0.002844,
		0.000458,
		0.002844,
		0.019046,
		0.047341,
		0.085389,
		0.131098,
		0.182603,
		0.238249,
		0.296573,
		0.356293,
		0.416289,
		0.475597,
		0.533393,
		0.588982,
		0.641793,
		0.691366,
		0.737347,
		0.779477,
		0.817592,
		0.851607,
		0.881522,
		0.907405,
		0.929393,
		0.947687,
		0.962541,
		0.97426,
		0.983187,
		0.9897,
		0.994196,
		0.997082,
		0.998756,
		0.999591,
		0.999916,
		0.999995,
		1,
		0.999995,
		0.999923,
		0.999643,
		0.99896,
		0.997658,
		0.995521,
		0.992346,
		0.987949,
		0.982179,
		0.974909,
		0.966046,
		0.955526,
		0.943313,
		0.929393,
		0.913781,
		0.896506,
		0.877618,
		0.857183,
		0.835279,
		0.811993,
		0.787425,
		0.76168,
		0.734868,
		0.707107,
		0.678515,
		0.649214,
		0.619329,
		0.588982,
		0.558299,
		0.527402,
		0.496415,
		0.465459,
		0.434652,
		0.404113,
		0.373956,
		0.344293,
		0.315234,
		0.286884,
		0.259348,
		0.232725,
		0.207112,
		0.182603,
		0.159289,
		0.137255,
		0.116587,
		0.097363,
		0.079662,
		0.063557,
		0.049119,
		0.036415,
		0.02551,
		0.016464,
		0.009337,
		0.004182,
		0.001053,
		0.000331,
		0.011231,
		0.035512,
		0.070648,
		0.114378,
		0.164687,
		0.219786,
		0.278096,
		0.338228,
		0.398972,
		0.459281,
		0.518262,
		0.575159,
		0.629348,
		0.680324,
		0.727694,
		0.771171,
		0.810562,
		0.845765,
		0.876762,
		0.903612,
		0.926448,
		0.945467,
		0.960924,
		0.97313,
		0.982438,
		0.989234,
		0.99393,
		0.996946,
		0.998697,
		0.999571,
		0.999912,
		0.999994,
		1,
		0.999995,
		0.99992,
		0.999626,
		0.998913,
		0.997555,
		0.995329,
		0.992024,
		0.987455,
		0.981465,
		0.973927,
		0.964748,
		0.953866,
		0.941245,
		0.926878,
		0.91078,
		0.892988,
		0.873557,
		0.852555,
		0.830067,
		0.806187,
		0.781018,
		0.754672,
		0.727265,
		0.698919,
		0.669758,
		0.63991,
		0.609502,
		0.578663,
		0.547521,
		0.516206,
		0.484843,
		0.453557,
		0.422472,
		0.391707,
		0.361381,
		0.331609,
		0.302502,
		0.27417,
		0.246719,
		0.22025,
		0.194862,
		0.170651,
		0.147708,
		0.126123,
		0.10598,
		0.08736,
		0.070342,
		0.055,
		0.041405,
		0.029626,
		0.019728,
		0.011771,
		0.005814,
		0.001913,
		0.00012,
		0.005236,
		0.024926,
		0.056682,
		0.098062,
		0.146884,
		0.201216,
		0.259348,
		0.319778,
		0.381194,
		0.442464,
		0.502614,
		0.560823,
		0.61641,
		0.668821,
		0.717621,
		0.762488,
		0.803203,
		0.83964,
		0.871765,
		0.899627,
		0.92335,
		0.943128,
		0.95922,
		0.971939,
		0.981647,
		0.988742,
		0.993648,
		0.996803,
		0.998635,
		0.999551,
		0.999908,
		0.999994,
		1,
		0.999994,
		0.999916,
		0.999609,
		0.998863,
		0.997446,
		0.995126,
		0.991686,
		0.986936,
		0.980715,
		0.972897,
		0.963389,
		0.952128,
		0.939084,
		0.924251,
		0.907651,
		0.889324,
		0.869329,
		0.847743,
		0.824654,
		0.800163,
		0.774379,
		0.747418,
		0.719404,
		0.690463,
		0.660725,
		0.630323,
		0.599388,
		0.568056,
		0.53646,
		0.504731,
		0.473,
		0.441396,
		0.410046,
		0.379073,
		0.348599,
		0.318742,
		0.289616,
		0.261333,
		0.234001,
		0.207724,
		0.182603,
		0.158736,
		0.136215,
		0.11513,
		0.095568,
		0.077611,
		0.061339,
		0.046826,
		0.034144,
		0.023363,
		0.014547,
		0.007758,
		0.003055,
		0.000492,
		0.001374,
		0.015823,
		0.043674,
		0.082283,
		0.129291,
		0.182603,
		0.24037,
		0.300964,
		0.362964,
		0.425141,
		0.486437,
		0.545959,
		0.602961,
		0.656837,
		0.707107,
		0.75341,
		0.795497,
		0.833218,
		0.86652,
		0.895439,
		0.92009,
		0.940665,
		0.957423,
		0.970681,
		0.980811,
		0.988221,
		0.99335,
		0.996651,
		0.998569,
		0.999529,
		0.999903,
		0.999994,
		1,
		0.999994,
		0.999912,
		0.999591,
		0.998811,
		0.997332,
		0.994912,
		0.99133,
		0.986389,
		0.979927,
		0.971816,
		0.961963,
		0.950308,
		0.936823,
		0.921508,
		0.904386,
		0.885504,
		0.864928,
		0.842738,
		0.819031,
		0.793912,
		0.767497,
		0.739908,
		0.711275,
		0.681729,
		0.651406,
		0.620445,
		0.588982,
		0.557157,
		0.525108,
		0.492972,
		0.460882,
		0.428973,
		0.397375,
		0.366214,
		0.335615,
		0.3057,
		0.276585,
		0.248386,
		0.221211,
		0.195168,
		0.17036,
		0.146884,
		0.124837,
		0.10431,
		0.085389,
		0.068159,
		0.052699,
		0.039085,
		0.027391,
		0.017685,
		0.010032,
		0.004495,
		0.001133,
		0.008478,
		0.031834,
		0.067199,
		0.112019,
		0.164025,
		0.221211,
		0.281814,
		0.344293,
		0.407312,
		0.469722,
		0.53055,
		0.588982,
		0.644352,
		0.696131,
		0.743917,
		0.787425,
		0.826481,
		0.861011,
		0.891034,
		0.916658,
		0.938069,
		0.955526,
		0.969353,
		0.979927,
		0.98767,
		0.993035,
		0.99649,
		0.998499,
		0.999505,
		0.999898,
		0.999993,
		1,
		0.999994,
		0.999908,
		0.999571,
		0.998756,
		0.99721,
		0.994686,
		0.990954,
		0.985814,
		0.979099,
		0.970681,
		0.960468,
		0.948403,
		0.934459,
		0.918641,
		0.900978,
		0.881522,
		0.860344,
		0.837533,
		0.813189,
		0.787425,
		0.760363,
		0.732133,
		0.702868,
		0.672708,
		0.641793,
		0.610268,
		0.578275,
		0.545959,
		0.513462,
		0.480925,
		0.448489,
		0.416289,
		0.38446,
		0.353134,
		0.322436,
		0.292492,
		0.263423,
		0.235343,
		0.208367,
		0.182603,
		0.158156,
		0.135126,
		0.113609,
		0.093699,
		0.075484,
		0.059048,
		0.044472,
		0.031834,
		0.021206,
		0.012657,
		0.006254,
		0.002059,
		0.00013,
		0.003208,
		0.021402,
		0.05299,
		0.0952,
		0.145573,
		0.201932,
		0.262364,
		0.325198,
		0.38898,
		0.452462,
		0.514584,
		0.574455,
		0.631345,
		0.684672,
		0.733987,
		0.778969,
		0.819413,
		0.855222,
		0.8864,
		0.913042,
		0.935331,
		0.953525,
		0.96795,
		0.978993,
		0.987087,
		0.9927,
		0.996318,
		0.998425,
		0.999481,
		0.999893,
		0.999993,
		1,
		0.999993,
		0.999903,
		0.999551,
		0.998697,
		0.997082,
		0.994448,
		0.990558,
		0.985208,
		0.978228,
		0.969489,
		0.9589,
		0.946406,
		0.931984,
		0.915644,
		0.89742,
		0.877369,
		0.85557,
		0.832117,
		0.807118,
		0.780692,
		0.752968,
		0.724081,
		0.694173,
		0.663389,
		0.631876,
		0.599783,
		0.567259,
		0.534454,
		0.501515,
		0.468589,
		0.435818,
		0.403344,
		0.371306,
		0.339837,
		0.30907,
		0.27913,
		0.250141,
		0.222223,
		0.19549,
		0.170054,
		0.14602,
		0.123492,
		0.102567,
		0.083339,
		0.065898,
		0.05033,
		0.036715,
		0.025132,
		0.015654,
		0.008351,
		0.003289,
		0.00053,
		0.000375,
		0.012657,
		0.039864,
		0.078987,
		0.127354,
		0.182603,
		0.242658,
		0.3057,
		0.370151,
		0.434652,
		0.498046,
		0.559361,
		0.617796,
		0.672708,
		0.723599,
		0.770107,
		0.811993,
		0.849137,
		0.881522,
		0.909232,
		0.932443,
		0.95141,
		0.966466,
		0.978003,
		0.986469,
		0.992346,
		0.996137,
		0.998347,
		0.999454,
		0.999888,
		0.999993,
		1,
		0.999993,
		0.999898,
		0.999529,
		0.998635,
		0.996946,
		0.994196,
		0.990141,
		0.984569,
		0.977312,
		0.968237,
		0.957254,
		0.944313,
		0.929393,
		0.91251,
		0.893703,
		0.873037,
		0.850596,
		0.826481,
		0.800807,
		0.773702,
		0.745299,
		0.715743,
		0.68518,
		0.653763,
		0.621646,
		0.588982,
		0.555928,
		0.522638,
		0.489264,
		0.455958,
		0.422868,
		0.39014,
		0.357915,
		0.326332,
		0.295525,
		0.265625,
		0.236757,
		0.209044,
		0.182603,
		0.157547,
		0.133985,
		0.112019,
		0.091751,
		0.073275,
		0.056682,
		0.042059,
		0.029488,
		0.019046,
		0.010808,
		0.004845,
		0.001221,
		0.00592,
		0.028064,
		0.063557,
		0.109496,
		0.163312,
		0.222749,
		0.28583,
		0.350837,
		0.416289,
		0.480925,
		0.543682,
		0.603682,
		0.660215,
		0.712729,
		0.760817,
		0.804203,
		0.842738,
		0.876386,
		0.905215,
		0.929393,
		0.949176,
		0.964896,
		0.976956,
		0.985814,
		0.991969,
		0.995944,
		0.998263,
		0.999426,
		0.999882,
		0.999992,
		1,
		0.999993,
		0.999893,
		0.999505,
		0.998569,
		0.996803,
		0.99393,
		0.9897,
		0.983896,
		0.976347,
		0.96692,
		0.955526,
		0.942118,
		0.92668,
		0.909232,
		0.889821,
		0.868517,
		0.845412,
		0.820615,
		0.794247,
		0.766443,
		0.737347,
		0.707107,
		0.675879,
		0.643821,
		0.611093,
		0.577857,
		0.544274,
		0.510503,
		0.476704,
		0.443032,
		0.409641,
		0.376679,
		0.344293,
		0.312626,
		0.281814,
		0.251992,
		0.223289,
		0.19583,
		0.169733,
		0.145114,
		0.122084,
		0.100747,
		0.081207,
		0.063557,
		0.047892,
		0.034297,
		0.022856,
		0.013648,
		0.006746,
		0.002221,
		0.00014,
		0.001559,
		0.01787,
		0.049119,
		0.092147,
		0.144161,
		0.202704,
		0.265625,
		0.331053,
		0.397375,
		0.463212,
		0.527402,
		0.588982,
		0.647171,
		0.701355,
		0.751077,
		0.796022,
		0.836008,
		0.870975,
		0.900978,
		0.926173,
		0.946813,
		0.963233,
		0.975845,
		0.985119,
		0.991569,
		0.995739,
		0.998174,
		0.999397,
		0.999876,
		0.999992,
		1,
		0.999992,
		0.999888,
		0.999481,
		0.998499,
		0.996651,
		0.993648,
		0.989234,
		0.983187,
		0.975331,
		0.965534,
		0.953711,
		0.939815,
		0.923837,
		0.905802,
		0.885764,
		0.863799,
		0.840008,
		0.814507,
		0.787425,
		0.758906,
		0.729099,
		0.698162,
		0.666257,
		0.63355,
		0.600208,
		0.566399,
		0.53229,
		0.498046,
		0.463832,
		0.429809,
		0.396135,
		0.362964,
		0.330446,
		0.298728,
		0.26795,
		0.238249,
		0.209758,
		0.182603,
		0.156908,
		0.132788,
		0.110356,
		0.08972,
		0.070982,
		0.05424,
		0.039587,
		0.02711,
		0.016893,
		0.009016,
		0.003552,
		0.000573,
		0.009606,
		0.035914,
		0.075483,
		0.125272,
		0.182603,
		0.245133,
		0.310823,
		0.377913,
		0.444897,
		0.510503,
		0.573675,
		0.63355,
		0.68945,
		0.740862,
		0.787425,
		0.828924,
		0.865273,
		0.896506,
		0.92277,
		0.944313,
		0.961473,
		0.974668,
		0.984381,
		0.991144,
		0.995521,
		0.998079,
		0.999365,
		0.999869,
		0.999991,
		1,
		0.999992,
		0.999882,
		0.999454,
		0.998425,
		0.99649,
		0.99335,
		0.988742,
		0.982438,
		0.97426,
		0.964076,
		0.951803,
		0.937398,
		0.920857,
		0.902212,
		0.881522,
		0.858873,
		0.834373,
		0.808146,
		0.78033,
		0.751077,
		0.720543,
		0.688896,
		0.656305,
		0.622942,
		0.588982,
		0.5546,
		0.519969,
		0.485261,
		0.450646,
		0.416289,
		0.382355,
		0.349001,
		0.316383,
		0.28465,
		0.253947,
		0.224414,
		0.196187,
		0.169394,
		0.144161,
		0.120607,
		0.098846,
		0.078987,
		0.061133,
		0.045385,
		0.031834,
		0.02057,
		0.011678,
		0.005236,
		0.00132,
		0.003646,
		0.024222,
		0.059709,
		0.106789,
		0.162543,
		0.224414,
		0.290179,
		0.357915,
		0.425977,
		0.492972,
		0.557738,
		0.619329,
		0.676989,
		0.730146,
		0.778391,
		0.821467,
		0.85926,
		0.891784,
		0.919171,
		0.941666,
		0.959607,
		0.973418,
		0.983597,
		0.990693,
		0.995289,
		0.997978,
		0.999331,
		0.999862,
		0.999991,
		1,
		0.999991,
		0.999876,
		0.999426,
		0.998347,
		0.996318,
		0.993035,
		0.988221,
		0.981647,
		0.97313,
		0.962541,
		0.949797,
		0.93486,
		0.917733,
		0.898452,
		0.877086,
		0.853729,
		0.828496,
		0.801521,
		0.77295,
		0.742944,
		0.711668,
		0.679298,
		0.64601,
		0.611985,
		0.577405,
		0.542452,
		0.507305,
		0.472143,
		0.437142,
		0.402472,
		0.368303,
		0.334797,
		0.302114,
		0.270407,
		0.239825,
		0.210511,
		0.182603,
		0.156234,
		0.131531,
		0.108615,
		0.087601,
		0.068601,
		0.05172,
		0.037057,
		0.024706,
		0.014759,
		0.007298,
		0.002404,
		0.000152,
		0.000428,
		0.01437,
		0.045061,
		0.088879,
		0.142636,
		0.20354,
		0.269161,
		0.337397,
		0.40645,
		0.474793,
		0.541152,
		0.604481,
		0.663946,
		0.718904,
		0.768893,
		0.813614,
		0.852917,
		0.886795,
		0.915365,
		0.938862,
		0.957627,
		0.972091,
		0.982764,
		0.990212,
		0.995042,
		0.99787,
		0.999295,
		0.999854,
		0.999991,
		1,
		0.999991,
		0.999869,
		0.999397,
		0.998263,
		0.996137,
		0.9927,
		0.98767,
		0.980811,
		0.971939,
		0.960924,
		0.947687,
		0.932195,
		0.914455,
		0.894513,
		0.872446,
		0.848355,
		0.822365,
		0.794618,
		0.765272,
		0.734494,
		0.702461,
		0.669354,
		0.63536,
		0.600669,
		0.565468,
		0.529947,
		0.494292,
		0.45869,
		0.42332,
		0.38836,
		0.353984,
		0.32036,
		0.287651,
		0.256014,
		0.225603,
		0.196564,
		0.169038,
		0.14316,
		0.119058,
		0.096858,
		0.076675,
		0.058623,
		0.042808,
		0.02933,
		0.018285,
		0.009762,
		0.003848,
		0.000621,
		0.00674,
		0.031819,
		0.071739,
		0.123016,
		0.182595,
		0.247816,
		0.316381,
		0.38632,
		0.455958,
		0.523896,
		0.588982,
		0.650291,
		0.707107,
		0.758906,
		0.80534,
		0.846224,
		0.881522,
		0.911335,
		0.93589,
		0.955526,
		0.970681,
		0.981877,
		0.9897,
		0.994778,
		0.997755,
		0.999256,
		0.999846,
		0.99999,
		1,
		0.999991,
		0.999862,
		0.999365,
		0.998174,
		0.995944,
		0.992346,
		0.987087,
		0.979927,
		0.970681,
		0.95922,
		0.945467,
		0.929393,
		0.911016,
		0.890386,
		0.86759,
		0.842738,
		0.815966,
		0.787425,
		0.757282,
		0.725714,
		0.692907,
		0.659052,
		0.624345,
		0.588982,
		0.553161,
		0.517077,
		0.480925,
		0.444897,
		0.409179,
		0.373956,
		0.339405,
		0.3057,
		0.273008,
		0.241492,
		0.211307,
		0.182603,
		0.155525,
		0.13021,
		0.106789,
		0.085389,
		0.066128,
		0.049119,
		0.034471,
		0.022285,
		0.012657,
		0.005678,
		0.001432,
		0.001778,
		0.020311,
		0.055596,
		0.103831,
		0.16167,
		0.226198,
		0.294891,
		0.365589,
		0.436456,
		0.505951,
		0.572806,
		0.635997,
		0.694725,
		0.7484,
		0.796619,
		0.839157,
		0.875946,
		0.907068,
		0.932739,
		0.953295,
		0.969182,
		0.980933,
		0.989154,
		0.994497,
		0.997633,
		0.999215,
		0.999838,
		0.999989,
		1,
		0.99999,
		0.999854,
		0.999331,
		0.998079,
		0.995739,
		0.991969,
		0.986469,
		0.978993,
		0.969353,
		0.957423,
		0.943128,
		0.926448,
		0.907405,
		0.886059,
		0.862506,
		0.836868,
		0.809287,
		0.779928,
		0.748966,
		0.71659,
		0.682994,
		0.64838,
		0.612952,
		0.576915,
		0.540475,
		0.503836,
		0.4672,
		0.430764,
		0.394723,
		0.359265,
		0.324575,
		0.29083,
		0.258204,
		0.226862,
		0.196963,
		0.168662,
		0.142104,
		0.117431,
		0.094776,
		0.074267,
		0.056024,
		0.040162,
		0.026791,
		0.016011,
		0.007921,
		0.00261,
		0.000165,
		0.01093,
		0.040727,
		0.085259,
		0.140868,
		0.204353,
		0.272942,
		0.344254,
		0.416269,
		0.487293,
		0.555925,
		0.621033,
		0.681729,
		0.737347,
		0.787425,
		0.831692,
		0.870045,
		0.902545,
		0.929393,
		0.950924,
		0.967586,
		0.979927,
		0.988571,
		0.994196,
		0.997501,
		0.999171,
		0.999829,
		0.999989,
		1,
		0.999989,
		0.999846,
		0.999295,
		0.997978,
		0.995521,
		0.991569,
		0.985814,
		0.978003,
		0.96795,
		0.955526,
		0.940665,
		0.92335,
		0.903612,
		0.881522,
		0.857183,
		0.830729,
		0.802314,
		0.772112,
		0.74031,
		0.707107,
		0.672708,
		0.637323,
		0.601168,
		0.564456,
		0.527402,
		0.490217,
		0.453111,
		0.416289,
		0.379951,
		0.344293,
		0.309504,
		0.275766,
		0.243258,
		0.212149,
		0.182603,
		0.154777,
		0.128819,
		0.104874,
		0.083077,
		0.063557,
		0.046438,
		0.031834,
		0.019856,
		0.010606,
		0.004182,
		0.000675,
		0.004144,
		0.027471,
		0.067515,
		0.120303,
		0.182325,
		0.250527,
		0.322282,
		0.395357,
		0.467878,
		0.538297,
		0.605362,
		0.668082,
		0.725713,
		0.777728,
		0.823803,
		0.863799,
		0.897749,
		0.925841,
		0.948403,
		0.965887,
		0.978855,
		0.987949,
		0.993875,
		0.997361,
		0.999123,
		0.999819,
		0.999988,
		1,
		0.999989,
		0.999838,
		0.999256,
		0.99787,
		0.995289,
		0.991144,
		0.985119,
		0.976956,
		0.966466,
		0.953525,
		0.938069,
		0.92009,
		0.899627,
		0.876762,
		0.851607,
		0.824309,
		0.795032,
		0.763962,
		0.731298,
		0.69725,
		0.662033,
		0.625869,
		0.588982,
		0.551596,
		0.513933,
		0.476215,
		0.438658,
		0.401474,
		0.364871,
		0.32905,
		0.294205,
		0.260527,
		0.228195,
		0.197385,
		0.168264,
		0.140991,
		0.11572,
		0.092596,
		0.071756,
		0.053333,
		0.03745,
		0.024223,
		0.013765,
		0.006177,
		0.001559,
		0.000483,
		0.016249,
		0.050887,
		0.100129,
		0.160159,
		0.227619,
		0.299601,
		0.373628,
		0.447615,
		0.519843,
		0.588915,
		0.653732,
		0.713457,
		0.767493,
		0.815461,
		0.857183,
		0.892661,
		0.922066,
		0.945719,
		0.964076,
		0.97771,
		0.987285,
		0.993531,
		0.99721,
		0.999073,
		0.999808,
		0.999987,
		1,
		0.999988,
		0.999829,
		0.999215,
		0.997755,
		0.995042,
		0.990693,
		0.984381,
		0.975845,
		0.964896,
		0.95141,
		0.935331,
		0.916658,
		0.895439,
		0.871765,
		0.845765,
		0.817592,
		0.787425,
		0.755463,
		0.721915,
		0.687003,
		0.650955,
		0.614003,
		0.576381,
		0.538323,
		0.500061,
		0.461824,
		0.423837,
		0.38632,
		0.349487,
		0.313545,
		0.278696,
		0.245133,
		0.213043,
		0.182603,
		0.153986,
		0.127354,
		0.102863,
		0.08066,
		0.060886,
		0.043674,
		0.02915,
		0.01743,
		0.008627,
		0.002844,
		0.000179,
		0.007563,
		0.035748,
		0.080573,
		0.137938,
		0.204182,
		0.276097,
		0.350925,
		0.426337,
		0.500403,
		0.57156,
		0.638572,
		0.700499,
		0.756663,
		0.806627,
		0.850168,
		0.887259,
		0.918052,
		0.942861,
		0.962145,
		0.976488,
		0.986574,
		0.993163,
		0.997049,
		0.999018,
		0.999797,
		0.999987,
		1,
		0.999987,
		0.999819,
		0.999171,
		0.997633,
		0.994778,
		0.990212,
		0.983597,
		0.974668,
		0.963233,
		0.949176,
		0.932443,
		0.913042,
		0.891034,
		0.86652,
		0.83964,
		0.810562,
		0.779477,
		0.746597,
		0.712144,
		0.676351,
		0.63946,
		0.601712,
		0.563354,
		0.524629,
		0.485779,
		0.447041,
		0.40865,
		0.370832,
		0.333809,
		0.297795,
		0.262996,
		0.229612,
		0.197833,
		0.167843,
		0.139816,
		0.113919,
		0.090309,
		0.069138,
		0.050548,
		0.034672,
		0.021638,
		0.011564,
		0.004562,
		0.000737,
		0.001968,
		0.022579,
		0.061965,
		0.115822,
		0.180217,
		0.25164,
		0.327031,
		0.40378,
		0.479711,
		0.55305,
		0.622395,
		0.686676,
		0.745119,
		0.797216,
		0.842698,
		0.881506,
		0.913775,
		0.939814,
		0.960083,
		0.975181,
		0.985814,
		0.992769,
		0.996876,
		0.99896,
		0.999784,
		0.999986,
		1,
		0.999987,
		0.999808,
		0.999123,
		0.997501,
		0.994497,
		0.9897,
		0.982764,
		0.973418,
		0.961473,
		0.946813,
		0.929393,
		0.909232,
		0.8864,
		0.861011,
		0.833218,
		0.803203,
		0.771171,
		0.737347,
		0.701966,
		0.665277,
		0.62753,
		0.588982,
		0.549888,
		0.510503,
		0.471079,
		0.431863,
		0.393097,
		0.355015,
		0.317848,
		0.281814,
		0.247127,
		0.213992,
		0.182603,
		0.153149,
		0.125807,
		0.100747,
		0.078131,
		0.058111,
		0.04083,
		0.026425,
		0.015024,
		0.006746,
		0.001703,
		0.011938,
		0.044763,
		0.094096,
		0.155812,
		0.226135,
		0.301705,
		0.379606,
		0.457378,
		0.532994,
		0.604841,
		0.671681,
		0.732617,
		0.78705,
		0.83465,
		0.875324,
		0.909189,
		0.936549,
		0.957876,
		0.973782,
		0.984999,
		0.992346,
		0.99669,
		0.998897,
		0.999771,
		0.999985,
		1,
		0.999986,
		0.999797,
		0.999073,
		0.997361,
		0.994196,
		0.989154,
		0.981877,
		0.972091,
		0.959607,
		0.944313,
		0.926173,
		0.905215,
		0.881522,
		0.855222,
		0.826481,
		0.795497,
		0.762488,
		0.727694,
		0.691366,
		0.653763,
		0.615152,
		0.575799,
		0.535972,
		0.495938,
		0.455958,
		0.416289,
		0.377182,
		0.33888,
		0.301618,
		0.265625,
		0.231119,
		0.198309,
		0.167396,
		0.138573,
		0.112019,
		0.08791,
		0.066407,
		0.047667,
		0.031834,
		0.019046,
		0.009431,
		0.003111,
		0.000196,
		0.00441,
		0.029542,
		0.073204,
		0.131208,
		0.199599,
		0.274752,
		0.353452,
		0.432925,
		0.510855,
		0.585368,
		0.655014,
		0.718723,
		0.775775,
		0.825756,
		0.868522,
		0.904167,
		0.932992,
		0.955481,
		0.972268,
		0.984119,
		0.99189,
		0.996489,
		0.998829,
		0.999757,
		0.999984,
		1,
		0.999985,
		0.999784,
		0.999018,
		0.99721,
		0.993875,
		0.988571,
		0.980933,
		0.970681,
		0.957627,
		0.941666,
		0.92277,
		0.900978,
		0.876386,
		0.849137,
		0.819413,
		0.787425,
		0.75341,
		0.717621,
		0.680324,
		0.641793,
		0.602308,
		0.562148,
		0.521595,
		0.480925,
		0.440412,
		0.40032,
		0.360911,
		0.322436,
		0.285138,
		0.249252,
		0.215002,
		0.182603,
		0.152262,
		0.124172,
		0.098521,
		0.075484,
		0.055226,
		0.037906,
		0.02367,
		0.012657,
		0.004996,
		0.000807,
		0.000496,
		0.016943,
		0.053745,
		0.106843,
		0.172233,
		0.246117,
		0.325018,
		0.405858,
		0.486002,
		0.563276,
		0.635962,
		0.702768,
		0.762802,
		0.815526,
		0.860718,
		0.898429,
		0.92895,
		0.952776,
		0.970571,
		0.983139,
		0.991385,
		0.996268,
		0.998754,
		0.999741,
		0.999983,
		1,
		0.999984,
		0.999771,
		0.99896,
		0.997049,
		0.993531,
		0.987949,
		0.979927,
		0.969182,
		0.955526,
		0.938862,
		0.919171,
		0.896506,
		0.870975,
		0.842738,
		0.811993,
		0.778969,
		0.743917,
		0.707107,
		0.668821,
		0.629348,
		0.588982,
		0.548017,
		0.506746,
		0.465459,
		0.424437,
		0.383959,
		0.344293,
		0.3057,
		0.26843,
		0.232725,
		0.198815,
		0.166922,
		0.137255,
		0.110014,
		0.085389,
		0.063557,
		0.044689,
		0.028941,
		0.016464,
		0.007397,
		0.001868,
		0.007571,
		0.036428,
		0.083354,
		0.14448,
		0.215973,
		0.294186,
		0.375779,
		0.457808,
		0.537774,
		0.613644,
		0.683852,
		0.747271,
		0.803188,
		0.851254,
		0.891446,
		0.924021,
		0.949475,
		0.968498,
		0.981941,
		0.990764,
		0.995994,
		0.998659,
		0.999719,
		0.99998,
		1,
		0.999983,
		0.999757,
		0.998897,
		0.996876,
		0.993163,
		0.987285,
		0.978855,
		0.967586,
		0.953295,
		0.93589,
		0.915365,
		0.891784,
		0.865273,
		0.836008,
		0.804203,
		0.770107,
		0.733987,
		0.696131,
		0.656837,
		0.61641,
		0.575159,
		0.533393,
		0.491417,
		0.449532,
		0.408035,
		0.367211,
		0.32734,
		0.28869,
		0.25152,
		0.216079,
		0.182603,
		0.151319,
		0.122442,
		0.096175,
		0.07271,
		0.05223,
		0.034906,
		0.020897,
		0.010354,
		0.003417,
		0.000216,
		0.001872,
		0.021973,
		0.061533,
		0.117035,
		0.184789,
		0.26112,
		0.342523,
		0.425793,
		0.508113,
		0.587116,
		0.660911,
		0.728083,
		0.787675,
		0.839153,
		0.882368,
		0.917501,
		0.945023,
		0.965637,
		0.980234,
		0.989838,
		0.99555,
		0.998478,
		0.999659,
		0.999963,
		0.999994,
		0.99998,
		0.999741,
		0.998829,
		0.99669,
		0.992769,
		0.986574,
		0.97771,
		0.965887,
		0.950924,
		0.932739,
		0.911335,
		0.886795,
		0.85926,
		0.828924,
		0.796022,
		0.760817,
		0.723599,
		0.684672,
		0.644352,
		0.602961,
		0.560823,
		0.518262,
		0.475597,
		0.433143,
		0.391206,
		0.350084,
		0.310066,
		0.271429,
		0.234441,
		0.199356,
		0.166417,
		0.135857,
		0.107895,
		0.082738,
		0.060583,
		0.041614,
		0.026003,
		0.013913,
		0.005495,
		0.000888,
		0.01098,
		0.042229,
		0.090795,
		0.153342,
		0.226346,
		0.30628,
		0.389776,
		0.473757,
		0.555533,
		0.632867,
		0.704007,
		0.767694,
		0.82314,
		0.870001,
		0.908329,
		0.938519,
		0.961255,
		0.977452,
		0.98819,
		0.994648,
		0.998023,
		0.999445,
		0.999866,
		0.999952,
		0.999963,
		0.999719,
		0.998754,
		0.996489,
		0.992346,
		0.985814,
		0.976488,
		0.964076,
		0.948403,
		0.929393,
		0.907068,
		0.881522,
		0.852917,
		0.821467,
		0.787425,
		0.751077,
		0.712729,
		0.672708,
		0.631345,
		0.588982,
		0.545959,
		0.502614,
		0.459281,
		0.416289,
		0.373956,
		0.332591,
		0.292492,
		0.253947,
		0.21723,
		0.182603,
		0.150317,
		0.120607,
		0.093699,
		0.069803,
		0.049119,
		0.031834,
		0.018122,
		0.008146,
		0.002059,
		0.003794,
		0.026203,
		0.066752,
		0.122659,
		0.190753,
		0.267665,
		0.35001,
		0.434554,
		0.518344,
		0.598814,
		0.67386,
		0.741871,
		0.801745,
		0.852871,
		0.895097,
		0.928683,
		0.954239,
		0.972665,
		0.985074,
		0.992715,
		0.996877,
		0.998789,
		0.999501,
		0.999758,
		0.999866,
		0.999659,
		0.998659,
		0.996268,
		0.99189,
		0.984999,
		0.975181,
		0.962145,
		0.945719,
		0.925841,
		0.902545,
		0.875946,
		0.846224,
		0.813614,
		0.778391,
		0.740862,
		0.701355,
		0.660215,
		0.617796,
		0.574455,
		0.53055,
		0.486437,
		0.442464,
		0.398972,
		0.356293,
		0.314747,
		0.274643,
		0.236278,
		0.199933,
		0.165879,
		0.134371,
		0.105651,
		0.079949,
		0.057479,
		0.038444,
		0.023032,
		0.011419,
		0.003771,
		0.000238,
		0.000395,
		0.013986,
		0.045838,
		0.093895,
		0.155529,
		0.227721,
		0.30725,
		0.390871,
		0.475478,
		0.558244,
		0.636733,
		0.708972,
		0.773497,
		0.829368,
		0.876156,
		0.913908,
		0.943094,
		0.964539,
		0.979348,
		0.988808,
		0.994288,
		0.997118,
		0.998451,
		0.999123,
		0.999501,
		0.999445,
		0.998478,
		0.995994,
		0.991385,
		0.984119,
		0.973782,
		0.960083,
		0.942861,
		0.922066,
		0.897749,
		0.870045,
		0.839157,
		0.80534,
		0.768893,
		0.730146,
		0.68945,
		0.647171,
		0.603682,
		0.559361,
		0.514584,
		0.469722,
		0.425141,
		0.381194,
		0.338228,
		0.296573,
		0.25655,
		0.218463,
		0.182603,
		0.149248,
		0.118659,
		0.091084,
		0.066755,
		0.045892,
		0.028698,
		0.015366,
		0.006072,
		0.000982,
		0.005759,
		0.028769,
		0.068164,
		0.122012,
		0.187825,
		0.262731,
		0.34365,
		0.427476,
		0.511233,
		0.592226,
		0.66815,
		0.737171,
		0.797981,
		0.84981,
		0.892416,
		0.92605,
		0.951392,
		0.969472,
		0.981571,
		0.989105,
		0.993485,
		0.995965,
		0.997474,
		0.998451,
		0.998789,
		0.998023,
		0.99555,
		0.990764,
		0.983139,
		0.972268,
		0.957876,
		0.939814,
		0.918052,
		0.892661,
		0.863799,
		0.831692,
		0.796619,
		0.758906,
		0.718904,
		0.676989,
		0.63355,
		0.588982,
		0.543682,
		0.498046,
		0.452462,
		0.407312,
		0.362964,
		0.319778,
		0.278096,
		0.238249,
		0.200552,
		0.165303,
		0.132788,
		0.103273,
		0.077011,
		0.05424,
		0.035182,
		0.020042,
		0.009016,
		0.00228,
		0.001299,
		0.015912,
		0.046363,
		0.091506,
		0.14952,
		0.218049,
		0.294361,
		0.375519,
		0.458554,
		0.540631,
		0.61919,
		0.692067,
		0.757581,
		0.814587,
		0.862495,
		0.901261,
		0.931338,
		0.953609,
		0.969283,
		0.97977,
		0.986527,
		0.990887,
		0.993865,
		0.995965,
		0.997118,
		0.996877,
		0.994648,
		0.989838,
		0.981941,
		0.970571,
		0.955481,
		0.936549,
		0.913775,
		0.887259,
		0.857183,
		0.823803,
		0.787425,
		0.7484,
		0.707107,
		0.663946,
		0.619329,
		0.573675,
		0.527402,
		0.480925,
		0.434652,
		0.38898,
		0.344293,
		0.300964,
		0.259348,
		0.219786,
		0.182603,
		0.148107,
		0.116587,
		0.088318,
		0.063557,
		0.042546,
		0.02551,
		0.012657,
		0.004182,
		0.000264,
		0.007219,
		0.029031,
		0.065084,
		0.1143,
		0.174964,
		0.244843,
		0.321345,
		0.401677,
		0.483015,
		0.56267,
		0.638232,
		0.707694,
		0.769537,
		0.822798,
		0.86708,
		0.902538,
		0.929828,
		0.950009,
		0.964421,
		0.974514,
		0.981659,
		0.986931,
		0.990887,
		0.993485,
		0.994288,
		0.992715,
		0.98819,
		0.980234,
		0.968498,
		0.952776,
		0.932992,
		0.909189,
		0.881506,
		0.850168,
		0.815461,
		0.777728,
		0.737347,
		0.694725,
		0.650291,
		0.604481,
		0.557738,
		0.510503,
		0.463212,
		0.416289,
		0.370151,
		0.325198,
		0.281814,
		0.24037,
		0.201216,
		0.164687,
		0.131098,
		0.100747,
		0.073915,
		0.050862,
		0.031834,
		0.017058,
		0.006746,
		0.001092,
		0.002241,
		0.01627,
		0.043437,
		0.083406,
		0.135159,
		0.197077,
		0.267053,
		0.342631,
		0.42117,
		0.5,
		0.576593,
		0.648705,
		0.714501,
		0.77265,
		0.822383,
		0.863508,
		0.896384,
		0.921848,
		0.941091,
		0.955494,
		0.966413,
		0.974941,
		0.981659,
		0.986527,
		0.989105,
		0.988808,
		0.985074,
		0.977452,
		0.965637,
		0.949475,
		0.92895,
		0.904167,
		0.875324,
		0.842698,
		0.806627,
		0.767493,
		0.725713,
		0.681729,
		0.635997,
		0.588982,
		0.541152,
		0.492972,
		0.444897,
		0.397375,
		0.350837,
		0.3057,
		0.262364,
		0.221211,
		0.182603,
		0.146884,
		0.114378,
		0.085389,
		0.060202,
		0.039085,
		0.022285,
		0.010032,
		0.002539,
		0.000207,
		0.007757,
		0.026795,
		0.057659,
		0.100022,
		0.152922,
		0.214834,
		0.283774,
		0.357431,
		0.433319,
		0.50894,
		0.581942,
		0.650267,
		0.712278,
		0.766854,
		0.813438,
		0.852052,
		0.883247,
		0.907999,
		0.927549,
		0.943177,
		0.955945,
		0.966413,
		0.974514,
		0.97977,
		0.981571,
		0.979348,
		0.972665,
		0.961255,
		0.945023,
		0.924021,
		0.898429,
		0.868522,
		0.83465,
		0.797216,
		0.756663,
		0.713457,
		0.668082,
		0.621033,
		0.572806,
		0.523896,
		0.474793,
		0.425977,
		0.377913,
		0.331053,
		0.28583,
		0.242658,
		0.201932,
		0.164025,
		0.129291,
		0.098062,
		0.070648,
		0.047341,
		0.02841,
		0.014108,
		0.004665,
		0.000295,
		0.002816,
		0.014927,
		0.037384,
		0.070476,
		0.113883,
		0.166698,
		0.227488,
		0.294395,
		0.365257,
		0.43776,
		0.509589,
		0.578592,
		0.642924,
		0.701172,
		0.752446,
		0.79642,
		0.833327,
		0.863875,
		0.889104,
		0.91016,
		0.928019,
		0.943177,
		0.955494,
		0.964421,
		0.969283,
		0.969472,
		0.964539,
		0.954239,
		0.938519,
		0.917501,
		0.891446,
		0.860718,
		0.825756,
		0.78705,
		0.745119,
		0.700499,
		0.653732,
		0.605362,
		0.555925,
		0.505951,
		0.455958,
		0.40645,
		0.357915,
		0.310823,
		0.265625,
		0.222749,
		0.182603,
		0.145573,
		0.112019,
		0.082283,
		0.056682,
		0.035512,
		0.019046,
		0.007538,
		0.001221,
		0.000551,
		0.007221,
		0.022412,
		0.046907,
		0.08095,
		0.124229,
		0.175889,
		0.234593,
		0.298613,
		0.365947,
		0.434461,
		0.502049,
		0.566781,
		0.627053,
		0.681705,
		0.730099,
		0.772148,
		0.80827,
		0.839261,
		0.866086,
		0.88959,
		0.91016,
		0.927549,
		0.941091,
		0.950009,
		0.953609,
		0.951392,
		0.943094,
		0.928683,
		0.908329,
		0.882368,
		0.851254,
		0.815526,
		0.775775,
		0.732617,
		0.686676,
		0.638572,
		0.588915,
		0.538297,
		0.487293,
		0.436456,
		0.38632,
		0.337397,
		0.290179,
		0.245133,
		0.202704,
		0.163312,
		0.127354,
		0.0952,
		0.067199,
		0.043674,
		0.024926,
		0.011231,
		0.002844,
		0.002811,
		0.012164,
		0.029175,
		0.054533,
		0.088442,
		0.130598,
		0.180203,
		0.236018,
		0.296449,
		0.359665,
		0.423733,
		0.486772,
		0.547106,
		0.603399,
		0.654769,
		0.700845,
		0.741761,
		0.778059,
		0.810497,
		0.839755,
		0.866086,
		0.889104,
		0.907999,
		0.921848,
		0.929828,
		0.931338,
		0.92605,
		0.913908,
		0.895097,
		0.870001,
		0.839153,
		0.803188,
		0.762802,
		0.718723,
		0.671681,
		0.622395,
		0.57156,
		0.519843,
		0.467878,
		0.416269,
		0.365589,
		0.316381,
		0.269161,
		0.224414,
		0.182603,
		0.144161,
		0.109496,
		0.078987,
		0.05299,
		0.031834,
		0.015823,
		0.005236,
		0.000331,
		0.000729,
		0.005777,
		0.016701,
		0.034483,
		0.05973,
		0.092608,
		0.132827,
		0.179646,
		0.23193,
		0.288229,
		0.346893,
		0.406208,
		0.464545,
		0.520504,
		0.573048,
		0.621588,
		0.666012,
		0.706621,
		0.743965,
		0.778554,
		0.810497,
		0.839261,
		0.863875,
		0.883247,
		0.896384,
		0.902538,
		0.901261,
		0.892416,
		0.876156,
		0.852871,
		0.82314,
		0.787675,
		0.747271,
		0.702768,
		0.655014,
		0.604841,
		0.55305,
		0.500403,
		0.447615,
		0.395357,
		0.344254,
		0.294891,
		0.247816,
		0.20354,
		0.162543,
		0.125272,
		0.092147,
		0.063557,
		0.039864,
		0.021402,
		0.008478,
		0.001374,
		5.4e-005,
		0.002261,
		0.008604,
		0.020197,
		0.037903,
		0.062253,
		0.093385,
		0.131026,
		0.174498,
		0.222774,
		0.274558,
		0.328396,
		0.38281,
		0.436447,
		0.48821,
		0.537373,
		0.58364,
		0.627111,
		0.668148,
		0.707107,
		0.743965,
		0.778059,
		0.80827,
		0.833327,
		0.852052,
		0.863508,
		0.86708,
		0.862495,
		0.84981,
		0.829368,
		0.801745,
		0.767694,
		0.728083,
		0.683852,
		0.635962,
		0.585368,
		0.532994,
		0.479711,
		0.426337,
		0.373628,
		0.322282,
		0.272942,
		0.226198,
		0.182595,
		0.142636,
		0.106789,
		0.075483,
		0.049119,
		0.028064,
		0.012657,
		0.003208,
		0.000644,
		0.003854,
		0.010748,
		0.022299,
		0.039266,
		0.062113,
		0.090958,
		0.125548,
		0.165276,
		0.209231,
		0.256277,
		0.305174,
		0.354699,
		0.403792,
		0.451676,
		0.497936,
		0.542528,
		0.585664,
		0.627578,
		0.668148,
		0.706621,
		0.741761,
		0.772148,
		0.79642,
		0.813438,
		0.822383,
		0.822798,
		0.814587,
		0.797981,
		0.773497,
		0.741871,
		0.704007,
		0.660911,
		0.613644,
		0.563276,
		0.510855,
		0.457378,
		0.40378,
		0.350925,
		0.299601,
		0.250527,
		0.204353,
		0.16167,
		0.123016,
		0.088879,
		0.059709,
		0.035914,
		0.01787,
		0.00592,
		0.000375,
		9.2e-005,
		0.001417,
		0.005051,
		0.01193,
		0.022903,
		0.038636,
		0.05954,
		0.085717,
		0.116944,
		0.152688,
		0.192162,
		0.234406,
		0.278401,
		0.323198,
		0.368043,
		0.412477,
		0.45636,
		0.499807,
		0.542966,
		0.585664,
		0.627111,
		0.666012,
		0.700845,
		0.730099,
		0.752446,
		0.766854,
		0.77265,
		0.769537,
		0.757581,
		0.737171,
		0.708972,
		0.67386,
		0.632867,
		0.587116,
		0.537774,
		0.486002,
		0.432925,
		0.379606,
		0.327031,
		0.276097,
		0.227619,
		0.182325,
		0.140868,
		0.103831,
		0.071739,
		0.045061,
		0.024222,
		0.009606,
		0.001559,
		0.000384,
		0.002005,
		0.005638,
		0.012091,
		0.022103,
		0.036262,
		0.054932,
		0.078208,
		0.105905,
		0.137572,
		0.172554,
		0.210073,
		0.249349,
		0.289712,
		0.330713,
		0.372177,
		0.414158,
		0.456761,
		0.499807,
		0.542528,
		0.58364,
		0.621588,
		0.654769,
		0.681705,
		0.701172,
		0.712278,
		0.714501,
		0.707694,
		0.692067,
		0.66815,
		0.636733,
		0.598814,
		0.555533,
		0.508113,
		0.457808,
		0.405858,
		0.353452,
		0.301705,
		0.25164,
		0.204182,
		0.160159,
		0.120303,
		0.085259,
		0.055596,
		0.031819,
		0.01437,
		0.003646,
		6e-005,
		0.000623,
		0.002248,
		0.005589,
		0.011339,
		0.020142,
		0.03252,
		0.048799,
		0.069077,
		0.09321,
		0.120838,
		0.151444,
		0.184447,
		0.219306,
		0.255628,
		0.293242,
		0.332178,
		0.372533,
		0.414158,
		0.45636,
		0.497936,
		0.537373,
		0.573048,
		0.603399,
		0.627053,
		0.642924,
		0.650267,
		0.648705,
		0.638232,
		0.61919,
		0.592226,
		0.558244,
		0.518344,
		0.473757,
		0.425793,
		0.375779,
		0.325018,
		0.274752,
		0.226135,
		0.180217,
		0.137938,
		0.100129,
		0.067515,
		0.040727,
		0.020311,
		0.00674,
		0.000428,
		0.000131,
		0.000698,
		0.002141,
		0.005009,
		0.009893,
		0.017358,
		0.027863,
		0.041711,
		0.059008,
		0.079658,
		0.103399,
		0.12986,
		0.158656,
		0.189486,
		0.222209,
		0.256861,
		0.293547,
		0.332178,
		0.372177,
		0.412477,
		0.451676,
		0.48821,
		0.520504,
		0.547106,
		0.566781,
		0.578592,
		0.581942,
		0.576593,
		0.56267,
		0.540631,
		0.511233,
		0.475478,
		0.434554,
		0.389776,
		0.342523,
		0.294186,
		0.246117,
		0.199599,
		0.155812,
		0.115822,
		0.080573,
		0.050887,
		0.027471,
		0.01093,
		0.001778,
		1.4e-005,
		0.000144,
		0.000614,
		0.001775,
		0.00408,
		0.008033,
		0.014124,
		0.022767,
		0.034243,
		0.048672,
		0.066016,
		0.086113,
		0.108744,
		0.133715,
		0.160942,
		0.190482,
		0.222461,
		0.256861,
		0.293242,
		0.330713,
		0.368043,
		0.403792,
		0.436447,
		0.464545,
		0.486772,
		0.502049,
		0.509589,
		0.50894,
		0.5,
		0.483015,
		0.458554,
		0.427476,
		0.390871,
		0.35001,
		0.30628,
		0.26112,
		0.215973,
		0.172233,
		0.131208,
		0.094096,
		0.061965,
		0.035748,
		0.016249,
		0.004144,
		1.3e-005,
		0.000106,
		0.000438,
		0.001282,
		0.003008,
		0.006041,
		0.010807,
		0.017677,
		0.026918,
		0.038672,
		0.052962,
		0.069727,
		0.088896,
		0.110452,
		0.134483,
		0.161143,
		0.190482,
		0.222209,
		0.255628,
		0.289712,
		0.323198,
		0.354699,
		0.38281,
		0.406208,
		0.423733,
		0.434461,
		0.43776,
		0.433319,
		0.42117,
		0.401677,
		0.375519,
		0.34365,
		0.30725,
		0.267665,
		0.226346,
		0.184789,
		0.14448,
		0.106843,
		0.073204,
		0.044763,
		0.022579,
		0.007563,
		0.000483,
		5.2e-005,
		0.000248,
		0.000793,
		0.001981,
		0.004166,
		0.007717,
		0.012966,
		0.020174,
		0.029504,
		0.041042,
		0.054829,
		0.070925,
		0.089455,
		0.110603,
		0.134483,
		0.160942,
		0.189486,
		0.219306,
		0.249349,
		0.278401,
		0.305174,
		0.328396,
		0.346893,
		0.359665,
		0.365947,
		0.365257,
		0.357431,
		0.342631,
		0.321345,
		0.294361,
		0.262731,
		0.227721,
		0.190753,
		0.153342,
		0.117035,
		0.083354,
		0.053745,
		0.029542,
		0.011938,
		0.001968,
		1.5e-005,
		0.000103,
		0.000402,
		0.001136,
		0.002591,
		0.005082,
		0.00891,
		0.014328,
		0.021524,
		0.030641,
		0.041813,
		0.055207,
		0.071031,
		0.089455,
		0.110452,
		0.133715,
		0.158656,
		0.184447,
		0.210073,
		0.234406,
		0.256277,
		0.274558,
		0.288229,
		0.296449,
		0.298613,
		0.294395,
		0.283774,
		0.267053,
		0.244843,
		0.218049,
		0.187825,
		0.155529,
		0.122659,
		0.090795,
		0.061533,
		0.036428,
		0.016943,
		0.00441,
		2.6e-005,
		0.000153,
		0.00054,
		0.001412,
		0.003031,
		0.005667,
		0.009563,
		0.014932,
		0.021968,
		0.030874,
		0.041882,
		0.055207,
		0.070925,
		0.088896,
		0.108744,
		0.12986,
		0.151444,
		0.172554,
		0.192162,
		0.209231,
		0.222774,
		0.23193,
		0.236018,
		0.234593,
		0.227488,
		0.214834,
		0.197077,
		0.174964,
		0.14952,
		0.122012,
		0.093895,
		0.066752,
		0.042229,
		0.021973,
		0.007571,
		0.000496,
		3.5e-005,
		0.000194,
		0.000642,
		0.00159,
		0.003275,
		0.005929,
		0.00978,
		0.015057,
		0.022008,
		0.030874,
		0.041813,
		0.054829,
		0.069727,
		0.086113,
		0.103399,
		0.120838,
		0.137572,
		0.152688,
		0.165276,
		0.174498,
		0.179646,
		0.180203,
		0.175889,
		0.166698,
		0.152922,
		0.135159,
		0.1143,
		0.091506,
		0.068164,
		0.045838,
		0.026203,
		0.01098,
		0.001872,
		4.3e-005,
		0.000222,
		0.000698,
		0.001671,
		0.003356,
		0.005984,
		0.0098,
		0.015057,
		0.021968,
		0.030641,
		0.041042,
		0.052962,
		0.066016,
		0.079658,
		0.09321,
		0.105905,
		0.116944,
		0.125548,
		0.131026,
		0.132827,
		0.130598,
		0.124229,
		0.113883,
		0.100022,
		0.083406,
		0.065084,
		0.046363,
		0.028769,
		0.013986,
		0.003794,
		4.7e-005,
		0.000234,
		0.000718,
		0.001688,
		0.003364,
		0.005984,
		0.00978,
		0.014932,
		0.021524,
		0.029504,
		0.038672,
		0.048672,
		0.059008,
		0.069077,
		0.078208,
		0.085717,
		0.090958,
		0.093385,
		0.092608,
		0.088442,
		0.08095,
		0.070476,
		0.057659,
		0.043437,
		0.029031,
		0.015912,
		0.005759,
		0.000395,
		4.8e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003356,
		0.005929,
		0.009563,
		0.014328,
		0.020174,
		0.026918,
		0.034243,
		0.041711,
		0.048799,
		0.054932,
		0.05954,
		0.062113,
		0.062253,
		0.05973,
		0.054533,
		0.046907,
		0.037384,
		0.026795,
		0.01627,
		0.007219,
		0.001299,
		4.9e-005,
		0.000237,
		0.000718,
		0.001671,
		0.003275,
		0.005667,
		0.00891,
		0.012966,
		0.017677,
		0.022767,
		0.027863,
		0.03252,
		0.036262,
		0.038636,
		0.039266,
		0.037903,
		0.034483,
		0.029175,
		0.022412,
		0.014927,
		0.007757,
		0.002241,
		4.8e-005,
		0.000234,
		0.000698,
		0.00159,
		0.003031,
		0.005082,
		0.007717,
		0.010807,
		0.014124,
		0.017358,
		0.020142,
		0.022103,
		0.022903,
		0.022299,
		0.020197,
		0.016701,
		0.012164,
		0.007221,
		0.002816,
		0.000207,
		4.7e-005,
		0.000222,
		0.000642,
		0.001412,
		0.002591,
		0.004166,
		0.006041,
		0.008033,
		0.009893,
		0.011339,
		0.012091,
		0.01193,
		0.010748,
		0.008604,
		0.005777,
		0.002811,
		0.000551,
		4.3e-005,
		0.000194,
		0.00054,
		0.001136,
		0.001981,
		0.003008,
		0.00408,
		0.005009,
		0.005589,
		0.005638,
		0.005051,
		0.003854,
		0.002261,
		0.000729,
		3.5e-005,
		0.000153,
		0.000402,
		0.000793,
		0.001282,
		0.001775,
		0.002141,
		0.002248,
		0.002005,
		0.001417,
		0.000644,
		5.4e-005,
		2.6e-005,
		0.000103,
		0.000248,
		0.000438,
		0.000614,
		0.000698,
		0.000623,
		0.000384,
		9.2e-005,
		1.5e-005,
		5.2e-005,
		0.000106,
		0.000144,
		0.000131,
		6e-005,
		1.3e-005,
		1.4e-005,
		1.4e-005,
		1.3e-005,
		6e-005,
		0.000131,
		0.000144,
		0.000106,
		5.2e-005,
		1.5e-005,
		9.2e-005,
		0.000384,
		0.000623,
		0.000698,
		0.000614,
		0.000438,
		0.000248,
		0.000103,
		2.6e-005,
		5.4e-005,
		0.000644,
		0.001417,
		0.002005,
		0.002248,
		0.002141,
		0.001775,
		0.001282,
		0.000793,
		0.000402,
		0.000153,
		3.5e-005,
		0.000729,
		0.002261,
		0.003854,
		0.005051,
		0.005638,
		0.005589,
		0.005009,
		0.00408,
		0.003008,
		0.001981,
		0.001136,
		0.00054,
		0.000194,
		4.3e-005,
		0.000551,
		0.002811,
		0.005777,
		0.008604,
		0.010748,
		0.01193,
		0.012091,
		0.011339,
		0.009893,
		0.008033,
		0.006041,
		0.004166,
		0.002591,
		0.001412,
		0.000642,
		0.000222,
		4.7e-005,
		0.000207,
		0.002816,
		0.007221,
		0.012164,
		0.016701,
		0.020197,
		0.022299,
		0.022903,
		0.022103,
		0.020142,
		0.017358,
		0.014124,
		0.010807,
		0.007717,
		0.005082,
		0.003031,
		0.00159,
		0.000698,
		0.000234,
		4.8e-005,
		0.002241,
		0.007757,
		0.014927,
		0.022412,
		0.029175,
		0.034483,
		0.037903,
		0.039266,
		0.038636,
		0.036262,
		0.03252,
		0.027863,
		0.022767,
		0.017677,
		0.012966,
		0.00891,
		0.005667,
		0.003275,
		0.001671,
		0.000718,
		0.000237,
		4.9e-005,
		0.001299,
		0.007219,
		0.01627,
		0.026795,
		0.037384,
		0.046907,
		0.054533,
		0.05973,
		0.062253,
		0.062113,
		0.05954,
		0.054932,
		0.048799,
		0.041711,
		0.034243,
		0.026918,
		0.020174,
		0.014328,
		0.009563,
		0.005929,
		0.003356,
		0.001688,
		0.000719,
		0.000237,
		4.8e-005,
		0.000395,
		0.005759,
		0.015912,
		0.029031,
		0.043437,
		0.057659,
		0.070476,
		0.08095,
		0.088442,
		0.092608,
		0.093385,
		0.090958,
		0.085717,
		0.078208,
		0.069077,
		0.059008,
		0.048672,
		0.038672,
		0.029504,
		0.021524,
		0.014932,
		0.00978,
		0.005984,
		0.003364,
		0.001688,
		0.000718,
		0.000234,
		4.7e-005,
		0.003794,
		0.013986,
		0.028769,
		0.046363,
		0.065084,
		0.083406,
		0.100022,
		0.113883,
		0.124229,
		0.130598,
		0.132827,
		0.131026,
		0.125548,
		0.116944,
		0.105905,
		0.09321,
		0.079658,
		0.066016,
		0.052962,
		0.041042,
		0.030641,
		0.021968,
		0.015057,
		0.0098,
		0.005984,
		0.003356,
		0.001671,
		0.000698,
		0.000222,
		4.3e-005,
		0.001872,
		0.01098,
		0.026203,
		0.045838,
		0.068164,
		0.091506,
		0.1143,
		0.135159,
		0.152922,
		0.166698,
		0.175889,
		0.180203,
		0.179646,
		0.174498,
		0.165276,
		0.152688,
		0.137572,
		0.120838,
		0.103399,
		0.086113,
		0.069727,
		0.054829,
		0.041813,
		0.030874,
		0.022008,
		0.015057,
		0.00978,
		0.005929,
		0.003275,
		0.00159,
		0.000642,
		0.000194,
		3.5e-005,
		0.000496,
		0.007571,
		0.021973,
		0.042229,
		0.066752,
		0.093895,
		0.122012,
		0.14952,
		0.174964,
		0.197077,
		0.214834,
		0.227488,
		0.234593,
		0.236018,
		0.23193,
		0.222774,
		0.209231,
		0.192162,
		0.172554,
		0.151444,
		0.12986,
		0.108744,
		0.088896,
		0.070925,
		0.055207,
		0.041882,
		0.030874,
		0.021968,
		0.014932,
		0.009563,
		0.005667,
		0.003031,
		0.001412,
		0.00054,
		0.000153,
		2.6e-005,
		0.00441,
		0.016943,
		0.036428,
		0.061533,
		0.090795,
		0.122659,
		0.155529,
		0.187825,
		0.218049,
		0.244843,
		0.267053,
		0.283774,
		0.294395,
		0.298613,
		0.296449,
		0.288229,
		0.274558,
		0.256277,
		0.234406,
		0.210073,
		0.184447,
		0.158656,
		0.133715,
		0.110452,
		0.089455,
		0.071031,
		0.055207,
		0.041813,
		0.030641,
		0.021524,
		0.014328,
		0.00891,
		0.005082,
		0.002591,
		0.001136,
		0.000402,
		0.000103,
		1.5e-005,
		0.001968,
		0.011938,
		0.029542,
		0.053745,
		0.083354,
		0.117035,
		0.153342,
		0.190753,
		0.227721,
		0.262731,
		0.294361,
		0.321345,
		0.342631,
		0.357431,
		0.365257,
		0.365947,
		0.359665,
		0.346893,
		0.328396,
		0.305174,
		0.278401,
		0.249349,
		0.219306,
		0.189486,
		0.160942,
		0.134483,
		0.110603,
		0.089455,
		0.070925,
		0.054829,
		0.041042,
		0.029504,
		0.020174,
		0.012966,
		0.007717,
		0.004166,
		0.001981,
		0.000793,
		0.000248,
		5.2e-005,
		0.000483,
		0.007563,
		0.022579,
		0.044763,
		0.073204,
		0.106843,
		0.14448,
		0.184789,
		0.226346,
		0.267665,
		0.30725,
		0.34365,
		0.375519,
		0.401677,
		0.42117,
		0.433319,
		0.43776,
		0.434461,
		0.423733,
		0.406208,
		0.38281,
		0.354699,
		0.323198,
		0.289712,
		0.255628,
		0.222209,
		0.190482,
		0.161143,
		0.134483,
		0.110452,
		0.088896,
		0.069727,
		0.052962,
		0.038672,
		0.026918,
		0.017677,
		0.010807,
		0.006041,
		0.003008,
		0.001282,
		0.000438,
		0.000106,
		1.3e-005,
		0.004144,
		0.016249,
		0.035748,
		0.061965,
		0.094096,
		0.131208,
		0.172233,
		0.215973,
		0.26112,
		0.30628,
		0.35001,
		0.390871,
		0.427476,
		0.458554,
		0.483015,
		0.5,
		0.50894,
		0.509589,
		0.502049,
		0.486772,
		0.464545,
		0.436447,
		0.403792,
		0.368043,
		0.330713,
		0.293242,
		0.256861,
		0.222461,
		0.190482,
		0.160942,
		0.133715,
		0.108744,
		0.086113,
		0.066016,
		0.048672,
		0.034243,
		0.022767,
		0.014124,
		0.008033,
		0.00408,
		0.001775,
		0.000614,
		0.000144,
		1.4e-005,
		0.001778,
		0.01093,
		0.027471,
		0.050887,
		0.080573,
		0.115822,
		0.155812,
		0.199599,
		0.246117,
		0.294186,
		0.342523,
		0.389776,
		0.434554,
		0.475478,
		0.511233,
		0.540631,
		0.56267,
		0.576593,
		0.581942,
		0.578592,
		0.566781,
		0.547106,
		0.520504,
		0.48821,
		0.451676,
		0.412477,
		0.372177,
		0.332178,
		0.293547,
		0.256861,
		0.222209,
		0.189486,
		0.158656,
		0.12986,
		0.103399,
		0.079658,
		0.059008,
		0.041711,
		0.027863,
		0.017358,
		0.009893,
		0.005009,
		0.002141,
		0.000698,
		0.000131,
		0.000428,
		0.00674,
		0.020311,
		0.040727,
		0.067515,
		0.100129,
		0.137938,
		0.180217,
		0.226135,
		0.274752,
		0.325018,
		0.375779,
		0.425793,
		0.473757,
		0.518344,
		0.558244,
		0.592226,
		0.61919,
		0.638232,
		0.648705,
		0.650267,
		0.642924,
		0.627053,
		0.603399,
		0.573048,
		0.537373,
		0.497936,
		0.45636,
		0.414158,
		0.372533,
		0.332178,
		0.293242,
		0.255628,
		0.219306,
		0.184447,
		0.151444,
		0.120838,
		0.09321,
		0.069077,
		0.048799,
		0.03252,
		0.020142,
		0.011339,
		0.005589,
		0.002248,
		0.000623,
		6e-005,
		0.003646,
		0.01437,
		0.031819,
		0.055596,
		0.085259,
		0.120303,
		0.160159,
		0.204182,
		0.25164,
		0.301705,
		0.353452,
		0.405858,
		0.457808,
		0.508113,
		0.555533,
		0.598814,
		0.636733,
		0.66815,
		0.692067,
		0.707694,
		0.714501,
		0.712278,
		0.701172,
		0.681705,
		0.654769,
		0.621588,
		0.58364,
		0.542528,
		0.499807,
		0.456761,
		0.414158,
		0.372177,
		0.330713,
		0.289712,
		0.249349,
		0.210073,
		0.172554,
		0.137572,
		0.105905,
		0.078208,
		0.054932,
		0.036262,
		0.022103,
		0.012091,
		0.005638,
		0.002005,
		0.000384,
		0.001559,
		0.009606,
		0.024222,
		0.045061,
		0.071739,
		0.103831,
		0.140868,
		0.182325,
		0.227619,
		0.276097,
		0.327031,
		0.379606,
		0.432925,
		0.486002,
		0.537774,
		0.587116,
		0.632867,
		0.67386,
		0.708972,
		0.737171,
		0.757581,
		0.769537,
		0.77265,
		0.766854,
		0.752446,
		0.730099,
		0.700845,
		0.666012,
		0.627111,
		0.585664,
		0.542966,
		0.499807,
		0.45636,
		0.412477,
		0.368043,
		0.323198,
		0.278401,
		0.234406,
		0.192162,
		0.152688,
		0.116944,
		0.085717,
		0.05954,
		0.038636,
		0.022903,
		0.01193,
		0.005051,
		0.001417,
		9.2e-005,
		0.000375,
		0.00592,
		0.01787,
		0.035914,
		0.059709,
		0.088879,
		0.123016,
		0.16167,
		0.204353,
		0.250527,
		0.299601,
		0.350925,
		0.40378,
		0.457378,
		0.510855,
		0.563276,
		0.613644,
		0.660911,
		0.704007,
		0.741871,
		0.773497,
		0.797981,
		0.814587,
		0.822798,
		0.822383,
		0.813438,
		0.79642,
		0.772148,
		0.741761,
		0.706621,
		0.668148,
		0.627578,
		0.585664,
		0.542528,
		0.497936,
		0.451676,
		0.403792,
		0.354699,
		0.305174,
		0.256277,
		0.209231,
		0.165276,
		0.125548,
		0.090958,
		0.062113,
		0.039266,
		0.022299,
		0.010748,
		0.003854,
		0.000644,
		0.003208,
		0.012657,
		0.028064,
		0.049119,
		0.075483,
		0.106789,
		0.142636,
		0.182595,
		0.226198,
		0.272942,
		0.322282,
		0.373628,
		0.426337,
		0.479711,
		0.532994,
		0.585368,
		0.635962,
		0.683852,
		0.728083,
		0.767694,
		0.801745,
		0.829368,
		0.84981,
		0.862495,
		0.86708,
		0.863508,
		0.852052,
		0.833327,
		0.80827,
		0.778059,
		0.743965,
		0.707107,
		0.668148,
		0.627111,
		0.58364,
		0.537373,
		0.48821,
		0.436447,
		0.38281,
		0.328396,
		0.274558,
		0.222774,
		0.174498,
		0.131026,
		0.093385,
		0.062253,
		0.037903,
		0.020197,
		0.008604,
		0.002261,
		5.4e-005,
		0.001374,
		0.008478,
		0.021402,
		0.039864,
		0.063557,
		0.092147,
		0.125272,
		0.162543,
		0.20354,
		0.247816,
		0.294891,
		0.344254,
		0.395357,
		0.447615,
		0.500403,
		0.55305,
		0.604841,
		0.655014,
		0.702768,
		0.747271,
		0.787675,
		0.82314,
		0.852871,
		0.876156,
		0.892416,
		0.901261,
		0.902538,
		0.896384,
		0.883247,
		0.863875,
		0.839261,
		0.810497,
		0.778554,
		0.743965,
		0.706621,
		0.666012,
		0.621588,
		0.573048,
		0.520504,
		0.464545,
		0.406208,
		0.346893,
		0.288229,
		0.23193,
		0.179646,
		0.132827,
		0.092608,
		0.05973,
		0.034483,
		0.016701,
		0.005777,
		0.000729,
		0.000331,
		0.005236,
		0.015823,
		0.031834,
		0.05299,
		0.078987,
		0.109496,
		0.144161,
		0.182603,
		0.224414,
		0.269161,
		0.316381,
		0.365589,
		0.416269,
		0.467878,
		0.519843,
		0.57156,
		0.622395,
		0.671681,
		0.718723,
		0.762802,
		0.803188,
		0.839153,
		0.870001,
		0.895097,
		0.913908,
		0.92605,
		0.931338,
		0.929828,
		0.921848,
		0.907999,
		0.889104,
		0.866086,
		0.839755,
		0.810497,
		0.778059,
		0.741761,
		0.700845,
		0.654769,
		0.603399,
		0.547106,
		0.486772,
		0.423733,
		0.359665,
		0.296449,
		0.236018,
		0.180203,
		0.130598,
		0.088442,
		0.054533,
		0.029175,
		0.012164,
		0.002811,
		0.002844,
		0.011231,
		0.024926,
		0.043674,
		0.067199,
		0.0952,
		0.127354,
		0.163312,
		0.202704,
		0.245133,
		0.290179,
		0.337397,
		0.38632,
		0.436456,
		0.487293,
		0.538297,
		0.588915,
		0.638572,
		0.686676,
		0.732617,
		0.775775,
		0.815526,
		0.851254,
		0.882368,
		0.908329,
		0.928683,
		0.943094,
		0.951392,
		0.953609,
		0.950009,
		0.941091,
		0.927549,
		0.91016,
		0.88959,
		0.866086,
		0.839261,
		0.80827,
		0.772148,
		0.730099,
		0.681705,
		0.627053,
		0.566781,
		0.502049,
		0.434461,
		0.365947,
		0.298613,
		0.234593,
		0.175889,
		0.124229,
		0.08095,
		0.046907,
		0.022412,
		0.007221,
		0.000551,
		0.001221,
		0.007538,
		0.019046,
		0.035512,
		0.056682,
		0.082283,
		0.112019,
		0.145573,
		0.182603,
		0.222749,
		0.265625,
		0.310823,
		0.357915,
		0.40645,
		0.455958,
		0.505951,
		0.555925,
		0.605362,
		0.653732,
		0.700499,
		0.745119,
		0.78705,
		0.825756,
		0.860718,
		0.891446,
		0.917501,
		0.938519,
		0.954239,
		0.964539,
		0.969472,
		0.969283,
		0.964421,
		0.955494,
		0.943177,
		0.928019,
		0.91016,
		0.889104,
		0.863875,
		0.833327,
		0.79642,
		0.752446,
		0.701172,
		0.642924,
		0.578592,
		0.509589,
		0.43776,
		0.365257,
		0.294395,
		0.227488,
		0.166698,
		0.113883,
		0.070476,
		0.037384,
		0.014927,
		0.002816,
		0.000295,
		0.004665,
		0.014108,
		0.02841,
		0.047341,
		0.070648,
		0.098062,
		0.129291,
		0.164025,
		0.201932,
		0.242658,
		0.28583,
		0.331053,
		0.377913,
		0.425977,
		0.474793,
		0.523896,
		0.572806,
		0.621033,
		0.668082,
		0.713457,
		0.756663,
		0.797216,
		0.83465,
		0.868522,
		0.898429,
		0.924021,
		0.945023,
		0.961255,
		0.972665,
		0.979348,
		0.981571,
		0.97977,
		0.974514,
		0.966413,
		0.955945,
		0.943177,
		0.927549,
		0.907999,
		0.883247,
		0.852052,
		0.813438,
		0.766854,
		0.712278,
		0.650267,
		0.581942,
		0.50894,
		0.433319,
		0.357431,
		0.283774,
		0.214834,
		0.152922,
		0.100022,
		0.057659,
		0.026795,
		0.007757,
		0.000207,
		0.002539,
		0.010032,
		0.022285,
		0.039085,
		0.060202,
		0.085389,
		0.114378,
		0.146884,
		0.182603,
		0.221211,
		0.262364,
		0.3057,
		0.350837,
		0.397375,
		0.444897,
		0.492972,
		0.541152,
		0.588982,
		0.635997,
		0.681729,
		0.725713,
		0.767493,
		0.806627,
		0.842698,
		0.875324,
		0.904167,
		0.92895,
		0.949475,
		0.965637,
		0.977452,
		0.985074,
		0.988808,
		0.989105,
		0.986527,
		0.981659,
		0.974941,
		0.966413,
		0.955494,
		0.941091,
		0.921848,
		0.896384,
		0.863508,
		0.822383,
		0.77265,
		0.714501,
		0.648705,
		0.576593,
		0.5,
		0.42117,
		0.342631,
		0.267053,
		0.197077,
		0.135159,
		0.083406,
		0.043437,
		0.01627,
		0.002241,
		0.001092,
		0.006746,
		0.017058,
		0.031834,
		0.050862,
		0.073915,
		0.100747,
		0.131098,
		0.164687,
		0.201216,
		0.24037,
		0.281814,
		0.325198,
		0.370151,
		0.416289,
		0.463212,
		0.510503,
		0.557738,
		0.604481,
		0.650291,
		0.694725,
		0.737347,
		0.777728,
		0.815461,
		0.850168,
		0.881506,
		0.909189,
		0.932992,
		0.952776,
		0.968498,
		0.980234,
		0.98819,
		0.992715,
		0.994288,
		0.993485,
		0.990887,
		0.986931,
		0.981659,
		0.974514,
		0.964421,
		0.950009,
		0.929828,
		0.902538,
		0.86708,
		0.822798,
		0.769537,
		0.707694,
		0.638232,
		0.56267,
		0.483015,
		0.401677,
		0.321345,
		0.244843,
		0.174964,
		0.1143,
		0.065084,
		0.029031,
		0.007219,
		0.000264,
		0.004182,
		0.012657,
		0.02551,
		0.042546,
		0.063557,
		0.088318,
		0.116587,
		0.148107,
		0.182603,
		0.219786,
		0.259348,
		0.300964,
		0.344293,
		0.38898,
		0.434652,
		0.480925,
		0.527402,
		0.573675,
		0.619329,
		0.663946,
		0.707107,
		0.7484,
		0.787425,
		0.823803,
		0.857183,
		0.887259,
		0.913775,
		0.936549,
		0.955481,
		0.970571,
		0.981941,
		0.989838,
		0.994648,
		0.996877,
		0.997118,
		0.995965,
		0.993865,
		0.990887,
		0.986527,
		0.97977,
		0.969283,
		0.953609,
		0.931338,
		0.901261,
		0.862495,
		0.814587,
		0.757581,
		0.692067,
		0.61919,
		0.540631,
		0.458554,
		0.375519,
		0.294361,
		0.218049,
		0.14952,
		0.091506,
		0.046363,
		0.015912,
		0.001299,
		0.00228,
		0.009016,
		0.020042,
		0.035182,
		0.05424,
		0.077011,
		0.103273,
		0.132788,
		0.165303,
		0.200552,
		0.238249,
		0.278096,
		0.319778,
		0.362964,
		0.407312,
		0.452462,
		0.498046,
		0.543682,
		0.588982,
		0.63355,
		0.676989,
		0.718904,
		0.758906,
		0.796619,
		0.831692,
		0.863799,
		0.892661,
		0.918052,
		0.939814,
		0.957876,
		0.972268,
		0.983139,
		0.990764,
		0.99555,
		0.998023,
		0.998789,
		0.998451,
		0.997474,
		0.995965,
		0.993485,
		0.989105,
		0.981571,
		0.969472,
		0.951392,
		0.92605,
		0.892416,
		0.84981,
		0.797981,
		0.737171,
		0.66815,
		0.592226,
		0.511233,
		0.427476,
		0.34365,
		0.262731,
		0.187825,
		0.122012,
		0.068164,
		0.028769,
		0.005759,
		0.000982,
		0.006072,
		0.015366,
		0.028698,
		0.045892,
		0.066755,
		0.091084,
		0.118659,
		0.149248,
		0.182603,
		0.218463,
		0.25655,
		0.296573,
		0.338228,
		0.381194,
		0.425141,
		0.469722,
		0.514584,
		0.559361,
		0.603682,
		0.647171,
		0.68945,
		0.730146,
		0.768893,
		0.80534,
		0.839157,
		0.870045,
		0.897749,
		0.922066,
		0.942861,
		0.960083,
		0.973782,
		0.984119,
		0.991385,
		0.995994,
		0.998478,
		0.999445,
		0.999501,
		0.999123,
		0.998451,
		0.997118,
		0.994288,
		0.988808,
		0.979348,
		0.964539,
		0.943094,
		0.913908,
		0.876156,
		0.829368,
		0.773497,
		0.708972,
		0.636733,
		0.558244,
		0.475478,
		0.390871,
		0.30725,
		0.227721,
		0.155529,
		0.093895,
		0.045838,
		0.013986,
		0.000395,
		0.000238,
		0.003771,
		0.011419,
		0.023032,
		0.038444,
		0.057479,
		0.079949,
		0.105651,
		0.134371,
		0.165879,
		0.199933,
		0.236278,
		0.274643,
		0.314747,
		0.356293,
		0.398972,
		0.442464,
		0.486437,
		0.53055,
		0.574455,
		0.617796,
		0.660215,
		0.701355,
		0.740862,
		0.778391,
		0.813614,
		0.846224,
		0.875946,
		0.902545,
		0.925841,
		0.945719,
		0.962145,
		0.975181,
		0.984999,
		0.99189,
		0.996268,
		0.998659,
		0.999659,
		0.999866,
		0.999758,
		0.999501,
		0.998789,
		0.996877,
		0.992715,
		0.985074,
		0.972665,
		0.954239,
		0.928683,
		0.895097,
		0.852871,
		0.801745,
		0.741871,
		0.67386,
		0.598814,
		0.518344,
		0.434554,
		0.35001,
		0.267665,
		0.190753,
		0.122659,
		0.066752,
		0.026203,
		0.003794,
		0.002059,
		0.008146,
		0.018122,
		0.031834,
		0.049119,
		0.069803,
		0.093699,
		0.120607,
		0.150317,
		0.182603,
		0.21723,
		0.253947,
		0.292492,
		0.332591,
		0.373956,
		0.416289,
		0.459281,
		0.502614,
		0.545959,
		0.588982,
		0.631345,
		0.672708,
		0.712729,
		0.751077,
		0.787425,
		0.821467,
		0.852917,
		0.881522,
		0.907068,
		0.929393,
		0.948403,
		0.964076,
		0.976488,
		0.985814,
		0.992346,
		0.996489,
		0.998754,
		0.999719,
		0.999963,
		0.999952,
		0.999866,
		0.999445,
		0.998023,
		0.994648,
		0.98819,
		0.977452,
		0.961255,
		0.938519,
		0.908329,
		0.870001,
		0.82314,
		0.767694,
		0.704007,
		0.632867,
		0.555533,
		0.473757,
		0.389776,
		0.30628,
		0.226346,
		0.153342,
		0.090795,
		0.042229,
		0.01098,
		0.000888,
		0.005495,
		0.013913,
		0.026003,
		0.041614,
		0.060583,
		0.082738,
		0.107895,
		0.135857,
		0.166417,
		0.199356,
		0.234441,
		0.271429,
		0.310066,
		0.350084,
		0.391206,
		0.433143,
		0.475597,
		0.518262,
		0.560823,
		0.602961,
		0.644352,
		0.684672,
		0.723599,
		0.760817,
		0.796022,
		0.828924,
		0.85926,
		0.886795,
		0.911335,
		0.932739,
		0.950924,
		0.965887,
		0.97771,
		0.986574,
		0.992769,
		0.99669,
		0.998829,
		0.999741,
		0.99998,
		0.999994,
		0.999963,
		0.999659,
		0.998478,
		0.99555,
		0.989838,
		0.980234,
		0.965637,
		0.945023,
		0.917501,
		0.882368,
		0.839153,
		0.787675,
		0.728083,
		0.660911,
		0.587116,
		0.508113,
		0.425793,
		0.342523,
		0.26112,
		0.184789,
		0.117035,
		0.061533,
		0.021973,
		0.001872,
		0.000216,
		0.003417,
		0.010354,
		0.020897,
		0.034906,
		0.05223,
		0.07271,
		0.096175,
		0.122442,
		0.151319,
		0.182603,
		0.216079,
		0.25152,
		0.28869,
		0.32734,
		0.367211,
		0.408035,
		0.449532,
		0.491417,
		0.533393,
		0.575159,
		0.61641,
		0.656837,
		0.696131,
		0.733987,
		0.770107,
		0.804203,
		0.836008,
		0.865273,
		0.891784,
		0.915365,
		0.93589,
		0.953295,
		0.967586,
		0.978855,
		0.987285,
		0.993163,
		0.996876,
		0.998897,
		0.999757,
		0.999983,
		1,
		0.99998,
		0.999719,
		0.998659,
		0.995994,
		0.990764,
		0.981941,
		0.968498,
		0.949475,
		0.924021,
		0.891446,
		0.851254,
		0.803188,
		0.747271,
		0.683852,
		0.613644,
		0.537774,
		0.457808,
		0.375779,
		0.294186,
		0.215973,
		0.14448,
		0.083354,
		0.036428,
		0.007571,
		0.001868,
		0.007397,
		0.016464,
		0.028941,
		0.044689,
		0.063557,
		0.085389,
		0.110014,
		0.137255,
		0.166922,
		0.198815,
		0.232725,
		0.26843,
		0.3057,
		0.344293,
		0.383959,
		0.424437,
		0.465459,
		0.506746,
		0.548017,
		0.588982,
		0.629348,
		0.668821,
		0.707107,
		0.743917,
		0.778969,
		0.811993,
		0.842738,
		0.870975,
		0.896506,
		0.919171,
		0.938862,
		0.955526,
		0.969182,
		0.979927,
		0.987949,
		0.993531,
		0.997049,
		0.99896,
		0.999771,
		0.999984,
		1,
		0.999983,
		0.999741,
		0.998754,
		0.996268,
		0.991385,
		0.983139,
		0.970571,
		0.952776,
		0.92895,
		0.898429,
		0.860718,
		0.815526,
		0.762802,
		0.702768,
		0.635962,
		0.563276,
		0.486002,
		0.405858,
		0.325018,
		0.246117,
		0.172233,
		0.106843,
		0.053745,
		0.016943,
		0.000496,
		0.000807,
		0.004996,
		0.012657,
		0.02367,
		0.037906,
		0.055226,
		0.075484,
		0.098521,
		0.124172,
		0.152262,
		0.182603,
		0.215002,
		0.249252,
		0.285138,
		0.322436,
		0.360911,
		0.40032,
		0.440412,
		0.480925,
		0.521595,
		0.562148,
		0.602308,
		0.641793,
		0.680324,
		0.717621,
		0.75341,
		0.787425,
		0.819413,
		0.849137,
		0.876386,
		0.900978,
		0.92277,
		0.941666,
		0.957627,
		0.970681,
		0.980933,
		0.988571,
		0.993875,
		0.99721,
		0.999018,
		0.999784,
		0.999985,
		1,
		0.999984,
		0.999757,
		0.998829,
		0.996489,
		0.99189,
		0.984119,
		0.972268,
		0.955481,
		0.932992,
		0.904167,
		0.868522,
		0.825756,
		0.775775,
		0.718723,
		0.655014,
		0.585368,
		0.510855,
		0.432925,
		0.353452,
		0.274752,
		0.199599,
		0.131208,
		0.073204,
		0.029542,
		0.00441,
		0.000196,
		0.003111,
		0.009431,
		0.019046,
		0.031834,
		0.047667,
		0.066407,
		0.08791,
		0.112019,
		0.138573,
		0.167396,
		0.198309,
		0.231119,
		0.265625,
		0.301618,
		0.33888,
		0.377182,
		0.416289,
		0.455958,
		0.495938,
		0.535972,
		0.575799,
		0.615152,
		0.653763,
		0.691366,
		0.727694,
		0.762488,
		0.795497,
		0.826481,
		0.855222,
		0.881522,
		0.905215,
		0.926173,
		0.944313,
		0.959607,
		0.972091,
		0.981877,
		0.989154,
		0.994196,
		0.997361,
		0.999073,
		0.999797,
		0.999986,
		1,
		0.999985,
		0.999771,
		0.998897,
		0.99669,
		0.992346,
		0.984999,
		0.973782,
		0.957876,
		0.936549,
		0.909189,
		0.875324,
		0.83465,
		0.78705,
		0.732617,
		0.671681,
		0.604841,
		0.532994,
		0.457378,
		0.379606,
		0.301705,
		0.226135,
		0.155812,
		0.094096,
		0.044763,
		0.011938,
		0.001703,
		0.006746,
		0.015024,
		0.026425,
		0.04083,
		0.058111,
		0.078131,
		0.100747,
		0.125807,
		0.153149,
		0.182603,
		0.213992,
		0.247127,
		0.281814,
		0.317848,
		0.355015,
		0.393097,
		0.431863,
		0.471079,
		0.510503,
		0.549888,
		0.588982,
		0.62753,
		0.665277,
		0.701966,
		0.737347,
		0.771171,
		0.803203,
		0.833218,
		0.861011,
		0.8864,
		0.909232,
		0.929393,
		0.946813,
		0.961473,
		0.973418,
		0.982764,
		0.9897,
		0.994497,
		0.997501,
		0.999123,
		0.999808,
		0.999987,
		1,
		0.999986,
		0.999784,
		0.99896,
		0.996876,
		0.992769,
		0.985814,
		0.975181,
		0.960083,
		0.939814,
		0.913775,
		0.881506,
		0.842698,
		0.797216,
		0.745119,
		0.686676,
		0.622395,
		0.55305,
		0.479711,
		0.40378,
		0.327031,
		0.25164,
		0.180217,
		0.115822,
		0.061965,
		0.022579,
		0.001968,
		0.000737,
		0.004562,
		0.011564,
		0.021638,
		0.034672,
		0.050548,
		0.069138,
		0.090309,
		0.113919,
		0.139816,
		0.167843,
		0.197833,
		0.229612,
		0.262996,
		0.297795,
		0.333809,
		0.370832,
		0.40865,
		0.447041,
		0.485779,
		0.524629,
		0.563354,
		0.601712,
		0.63946,
		0.676351,
		0.712144,
		0.746597,
		0.779477,
		0.810562,
		0.83964,
		0.86652,
		0.891034,
		0.913042,
		0.932443,
		0.949176,
		0.963233,
		0.974668,
		0.983597,
		0.990212,
		0.994778,
		0.997633,
		0.999171,
		0.999819,
		0.999987,
		1,
		0.999987,
		0.999797,
		0.999018,
		0.997049,
		0.993163,
		0.986574,
		0.976488,
		0.962145,
		0.942861,
		0.918052,
		0.887259,
		0.850168,
		0.806627,
		0.756663,
		0.700499,
		0.638572,
		0.57156,
		0.500403,
		0.426337,
		0.350925,
		0.276097,
		0.204182,
		0.137938,
		0.080573,
		0.035748,
		0.007563,
		0.000179,
		0.002844,
		0.008627,
		0.01743,
		0.02915,
		0.043674,
		0.060886,
		0.08066,
		0.102863,
		0.127354,
		0.153986,
		0.182603,
		0.213043,
		0.245133,
		0.278696,
		0.313545,
		0.349487,
		0.38632,
		0.423837,
		0.461824,
		0.500061,
		0.538323,
		0.576381,
		0.614003,
		0.650955,
		0.687003,
		0.721915,
		0.755463,
		0.787425,
		0.817592,
		0.845765,
		0.871765,
		0.895439,
		0.916658,
		0.935331,
		0.95141,
		0.964896,
		0.975845,
		0.984381,
		0.990693,
		0.995042,
		0.997755,
		0.999215,
		0.999829,
		0.999988,
		1,
		0.999987,
		0.999808,
		0.999073,
		0.99721,
		0.993531,
		0.987285,
		0.97771,
		0.964076,
		0.945719,
		0.922066,
		0.892661,
		0.857183,
		0.815461,
		0.767493,
		0.713457,
		0.653732,
		0.588915,
		0.519843,
		0.447615,
		0.373628,
		0.299601,
		0.227619,
		0.160159,
		0.100129,
		0.050887,
		0.016249,
		0.000483,
		0.001559,
		0.006177,
		0.013765,
		0.024223,
		0.03745,
		0.053333,
		0.071756,
		0.092596,
		0.11572,
		0.140991,
		0.168264,
		0.197385,
		0.228195,
		0.260527,
		0.294205,
		0.32905,
		0.364871,
		0.401474,
		0.438658,
		0.476215,
		0.513933,
		0.551596,
		0.588982,
		0.625869,
		0.662033,
		0.69725,
		0.731298,
		0.763962,
		0.795032,
		0.824309,
		0.851607,
		0.876762,
		0.899627,
		0.92009,
		0.938069,
		0.953525,
		0.966466,
		0.976956,
		0.985119,
		0.991144,
		0.995289,
		0.99787,
		0.999256,
		0.999838,
		0.999989,
		1,
		0.999988,
		0.999819,
		0.999123,
		0.997361,
		0.993875,
		0.987949,
		0.978855,
		0.965887,
		0.948403,
		0.925841,
		0.897749,
		0.863799,
		0.823803,
		0.777728,
		0.725713,
		0.668082,
		0.605362,
		0.538297,
		0.467878,
		0.395357,
		0.322282,
		0.250527,
		0.182325,
		0.120303,
		0.067515,
		0.027471,
		0.004144,
		0.000675,
		0.004182,
		0.010606,
		0.019856,
		0.031834,
		0.046438,
		0.063557,
		0.083077,
		0.104874,
		0.128819,
		0.154777,
		0.182603,
		0.212149,
		0.243258,
		0.275766,
		0.309504,
		0.344293,
		0.379951,
		0.416289,
		0.453111,
		0.490217,
		0.527402,
		0.564456,
		0.601168,
		0.637323,
		0.672708,
		0.707107,
		0.74031,
		0.772112,
		0.802314,
		0.830729,
		0.857183,
		0.881522,
		0.903612,
		0.92335,
		0.940665,
		0.955526,
		0.96795,
		0.978003,
		0.985814,
		0.991569,
		0.995521,
		0.997978,
		0.999295,
		0.999846,
		0.999989,
		1,
		0.999989,
		0.999829,
		0.999171,
		0.997501,
		0.994196,
		0.988571,
		0.979927,
		0.967586,
		0.950924,
		0.929393,
		0.902545,
		0.870045,
		0.831692,
		0.787425,
		0.737347,
		0.681729,
		0.621033,
		0.555925,
		0.487293,
		0.416269,
		0.344254,
		0.272942,
		0.204353,
		0.140868,
		0.085259,
		0.040727,
		0.01093,
		0.000165,
		0.00261,
		0.007921,
		0.016011,
		0.026791,
		0.040162,
		0.056024,
		0.074267,
		0.094776,
		0.117431,
		0.142104,
		0.168662,
		0.196963,
		0.226862,
		0.258204,
		0.29083,
		0.324575,
		0.359265,
		0.394723,
		0.430764,
		0.4672,
		0.503836,
		0.540475,
		0.576915,
		0.612952,
		0.64838,
		0.682994,
		0.71659,
		0.748966,
		0.779928,
		0.809287,
		0.836868,
		0.862506,
		0.886059,
		0.907405,
		0.926448,
		0.943128,
		0.957423,
		0.969353,
		0.978993,
		0.986469,
		0.991969,
		0.995739,
		0.998079,
		0.999331,
		0.999854,
		0.99999,
		1,
		0.999989,
		0.999838,
		0.999215,
		0.997633,
		0.994497,
		0.989154,
		0.980933,
		0.969182,
		0.953295,
		0.932739,
		0.907068,
		0.875946,
		0.839157,
		0.796619,
		0.7484,
		0.694725,
		0.635997,
		0.572806,
		0.505951,
		0.436456,
		0.365589,
		0.294891,
		0.226198,
		0.16167,
		0.103831,
		0.055596,
		0.020311,
		0.001778,
		0.001432,
		0.005678,
		0.012657,
		0.022285,
		0.034471,
		0.049119,
		0.066128,
		0.085389,
		0.106789,
		0.13021,
		0.155525,
		0.182603,
		0.211307,
		0.241492,
		0.273008,
		0.3057,
		0.339405,
		0.373956,
		0.409179,
		0.444897,
		0.480925,
		0.517077,
		0.553161,
		0.588982,
		0.624345,
		0.659052,
		0.692907,
		0.725714,
		0.757282,
		0.787425,
		0.815966,
		0.842738,
		0.86759,
		0.890386,
		0.911016,
		0.929393,
		0.945467,
		0.95922,
		0.970681,
		0.979927,
		0.987087,
		0.992346,
		0.995944,
		0.998174,
		0.999365,
		0.999862,
		0.999991,
		1,
		0.99999,
		0.999846,
		0.999256,
		0.997755,
		0.994778,
		0.9897,
		0.981877,
		0.970681,
		0.955526,
		0.93589,
		0.911335,
		0.881522,
		0.846224,
		0.80534,
		0.758906,
		0.707107,
		0.650291,
		0.588982,
		0.523896,
		0.455958,
		0.38632,
		0.316381,
		0.247816,
		0.182595,
		0.123016,
		0.071739,
		0.031819,
		0.00674,
		0.000621,
		0.003848,
		0.009762,
		0.018285,
		0.02933,
		0.042808,
		0.058623,
		0.076675,
		0.096858,
		0.119058,
		0.14316,
		0.169038,
		0.196564,
		0.225603,
		0.256014,
		0.287651,
		0.32036,
		0.353984,
		0.38836,
		0.42332,
		0.45869,
		0.494292,
		0.529947,
		0.565468,
		0.600669,
		0.63536,
		0.669354,
		0.702461,
		0.734494,
		0.765272,
		0.794618,
		0.822365,
		0.848355,
		0.872446,
		0.894513,
		0.914455,
		0.932195,
		0.947687,
		0.960924,
		0.971939,
		0.980811,
		0.98767,
		0.9927,
		0.996137,
		0.998263,
		0.999397,
		0.999869,
		0.999991,
		1,
		0.999991,
		0.999854,
		0.999295,
		0.99787,
		0.995042,
		0.990212,
		0.982764,
		0.972091,
		0.957627,
		0.938862,
		0.915365,
		0.886795,
		0.852917,
		0.813614,
		0.768893,
		0.718904,
		0.663946,
		0.604481,
		0.541152,
		0.474793,
		0.40645,
		0.337397,
		0.269161,
		0.20354,
		0.142636,
		0.088879,
		0.045061,
		0.01437,
		0.000428,
		0.000152,
		0.002404,
		0.007298,
		0.014759,
		0.024706,
		0.037057,
		0.05172,
		0.068601,
		0.087601,
		0.108615,
		0.131531,
		0.156234,
		0.182603,
		0.210511,
		0.239825,
		0.270407,
		0.302114,
		0.334797,
		0.368303,
		0.402472,
		0.437142,
		0.472143,
		0.507305,
		0.542452,
		0.577405,
		0.611985,
		0.64601,
		0.679298,
		0.711668,
		0.742944,
		0.77295,
		0.801521,
		0.828496,
		0.853729,
		0.877086,
		0.898452,
		0.917733,
		0.93486,
		0.949797,
		0.962541,
		0.97313,
		0.981647,
		0.988221,
		0.993035,
		0.996318,
		0.998347,
		0.999426,
		0.999876,
		0.999991,
		1,
		0.999991,
		0.999862,
		0.999331,
		0.997978,
		0.995289,
		0.990693,
		0.983597,
		0.973418,
		0.959607,
		0.941666,
		0.919171,
		0.891784,
		0.85926,
		0.821467,
		0.778391,
		0.730146,
		0.676989,
		0.619329,
		0.557738,
		0.492972,
		0.425977,
		0.357915,
		0.290179,
		0.224414,
		0.162543,
		0.106789,
		0.059709,
		0.024222,
		0.003646,
		0.00132,
		0.005236,
		0.011678,
		0.02057,
		0.031834,
		0.045385,
		0.061133,
		0.078987,
		0.098846,
		0.120607,
		0.144161,
		0.169394,
		0.196187,
		0.224414,
		0.253947,
		0.28465,
		0.316383,
		0.349001,
		0.382355,
		0.416289,
		0.450646,
		0.485261,
		0.519969,
		0.5546,
		0.588982,
		0.622942,
		0.656305,
		0.688896,
		0.720543,
		0.751077,
		0.78033,
		0.808146,
		0.834373,
		0.858873,
		0.881522,
		0.902212,
		0.920857,
		0.937398,
		0.951803,
		0.964076,
		0.97426,
		0.982438,
		0.988742,
		0.99335,
		0.99649,
		0.998425,
		0.999454,
		0.999882,
		0.999992,
		1,
		0.999991,
		0.999869,
		0.999365,
		0.998079,
		0.995521,
		0.991144,
		0.984381,
		0.974668,
		0.961473,
		0.944313,
		0.92277,
		0.896506,
		0.865273,
		0.828924,
		0.787425,
		0.740862,
		0.68945,
		0.63355,
		0.573675,
		0.510503,
		0.444897,
		0.377913,
		0.310823,
		0.245133,
		0.182603,
		0.125272,
		0.075483,
		0.035914,
		0.009606,
		0.000573,
		0.003552,
		0.009016,
		0.016893,
		0.02711,
		0.039587,
		0.05424,
		0.070982,
		0.08972,
		0.110356,
		0.132788,
		0.156908,
		0.182603,
		0.209758,
		0.238249,
		0.26795,
		0.298728,
		0.330446,
		0.362964,
		0.396135,
		0.429809,
		0.463832,
		0.498046,
		0.53229,
		0.566399,
		0.600208,
		0.63355,
		0.666257,
		0.698162,
		0.729099,
		0.758906,
		0.787425,
		0.814507,
		0.840008,
		0.863799,
		0.885764,
		0.905802,
		0.923837,
		0.939815,
		0.953711,
		0.965534,
		0.975331,
		0.983187,
		0.989234,
		0.993648,
		0.996651,
		0.998499,
		0.999481,
		0.999888,
		0.999992,
		1,
		0.999992,
		0.999876,
		0.999397,
		0.998174,
		0.995739,
		0.991569,
		0.985119,
		0.975845,
		0.963233,
		0.946813,
		0.926173,
		0.900978,
		0.870975,
		0.836008,
		0.796022,
		0.751077,
		0.701355,
		0.647171,
		0.588982,
		0.527402,
		0.463212,
		0.397375,
		0.331053,
		0.265625,
		0.202704,
		0.144161,
		0.092147,
		0.049119,
		0.01787,
		0.001559,
		0.00014,
		0.002221,
		0.006746,
		0.013648,
		0.022856,
		0.034297,
		0.047892,
		0.063557,
		0.081207,
		0.100747,
		0.122084,
		0.145114,
		0.169733,
		0.19583,
		0.223289,
		0.251992,
		0.281814,
		0.312626,
		0.344293,
		0.376679,
		0.409641,
		0.443032,
		0.476704,
		0.510503,
		0.544274,
		0.577857,
		0.611093,
		0.643821,
		0.675879,
		0.707107,
		0.737347,
		0.766443,
		0.794247,
		0.820615,
		0.845412,
		0.868517,
		0.889821,
		0.909232,
		0.92668,
		0.942118,
		0.955526,
		0.96692,
		0.976347,
		0.983896,
		0.9897,
		0.99393,
		0.996803,
		0.998569,
		0.999505,
		0.999893,
		0.999993,
		1,
		0.999992,
		0.999882,
		0.999426,
		0.998263,
		0.995944,
		0.991969,
		0.985814,
		0.976956,
		0.964896,
		0.949176,
		0.929393,
		0.905215,
		0.876386,
		0.842738,
		0.804203,
		0.760817,
		0.712729,
		0.660215,
		0.603682,
		0.543682,
		0.480925,
		0.416289,
		0.350837,
		0.28583,
		0.222749,
		0.163312,
		0.109496,
		0.063557,
		0.028064,
		0.00592,
		0.001221,
		0.004845,
		0.010808,
		0.019046,
		0.029488,
		0.042059,
		0.056682,
		0.073275,
		0.091751,
		0.112019,
		0.133985,
		0.157547,
		0.182603,
		0.209044,
		0.236757,
		0.265625,
		0.295525,
		0.326332,
		0.357915,
		0.39014,
		0.422868,
		0.455958,
		0.489264,
		0.522638,
		0.555928,
		0.588982,
		0.621646,
		0.653763,
		0.68518,
		0.715743,
		0.745299,
		0.773702,
		0.800807,
		0.826481,
		0.850596,
		0.873037,
		0.893703,
		0.91251,
		0.929393,
		0.944313,
		0.957254,
		0.968237,
		0.977312,
		0.984569,
		0.990141,
		0.994196,
		0.996946,
		0.998635,
		0.999529,
		0.999898,
		0.999993,
		1,
		0.999993,
		0.999888,
		0.999454,
		0.998347,
		0.996137,
		0.992346,
		0.986469,
		0.978003,
		0.966466,
		0.95141,
		0.932443,
		0.909232,
		0.881522,
		0.849137,
		0.811993,
		0.770107,
		0.723599,
		0.672708,
		0.617796,
		0.559361,
		0.498046,
		0.434652,
		0.370151,
		0.3057,
		0.242658,
		0.182603,
		0.127354,
		0.078987,
		0.039864,
		0.012657,
		0.000375,
		0.00053,
		0.003289,
		0.008351,
		0.015654,
		0.025132,
		0.036715,
		0.05033,
		0.065898,
		0.083339,
		0.102567,
		0.123492,
		0.14602,
		0.170054,
		0.19549,
		0.222223,
		0.250141,
		0.27913,
		0.30907,
		0.339837,
		0.371306,
		0.403344,
		0.435818,
		0.468589,
		0.501515,
		0.534454,
		0.567259,
		0.599783,
		0.631876,
		0.663389,
		0.694173,
		0.724081,
		0.752968,
		0.780692,
		0.807118,
		0.832117,
		0.85557,
		0.877369,
		0.89742,
		0.915644,
		0.931984,
		0.946406,
		0.9589,
		0.969489,
		0.978228,
		0.985208,
		0.990558,
		0.994448,
		0.997082,
		0.998697,
		0.999551,
		0.999903,
		0.999993,
		1,
		0.999993,
		0.999893,
		0.999481,
		0.998425,
		0.996318,
		0.9927,
		0.987087,
		0.978993,
		0.96795,
		0.953525,
		0.935331,
		0.913042,
		0.8864,
		0.855222,
		0.819413,
		0.778969,
		0.733987,
		0.684672,
		0.631345,
		0.574455,
		0.514584,
		0.452462,
		0.38898,
		0.325198,
		0.262364,
		0.201932,
		0.145573,
		0.0952,
		0.05299,
		0.021402,
		0.003208,
		0.00013,
		0.002059,
		0.006254,
		0.012657,
		0.021206,
		0.031834,
		0.044472,
		0.059048,
		0.075484,
		0.093699,
		0.113609,
		0.135126,
		0.158156,
		0.182603,
		0.208367,
		0.235343,
		0.263423,
		0.292492,
		0.322436,
		0.353134,
		0.38446,
		0.416289,
		0.448489,
		0.480925,
		0.513462,
		0.545959,
		0.578275,
		0.610268,
		0.641793,
		0.672708,
		0.702868,
		0.732133,
		0.760363,
		0.787425,
		0.813189,
		0.837533,
		0.860344,
		0.881522,
		0.900978,
		0.918641,
		0.934459,
		0.948403,
		0.960468,
		0.970681,
		0.979099,
		0.985814,
		0.990954,
		0.994686,
		0.99721,
		0.998756,
		0.999571,
		0.999908,
		0.999994,
		1,
		0.999993,
		0.999898,
		0.999505,
		0.998499,
		0.99649,
		0.993035,
		0.98767,
		0.979927,
		0.969353,
		0.955526,
		0.938069,
		0.916658,
		0.891034,
		0.861011,
		0.826481,
		0.787425,
		0.743917,
		0.696131,
		0.644352,
		0.588982,
		0.53055,
		0.469722,
		0.407312,
		0.344293,
		0.281814,
		0.221211,
		0.164025,
		0.112019,
		0.067199,
		0.031834,
		0.008478,
		0.001133,
		0.004495,
		0.010032,
		0.017685,
		0.027391,
		0.039085,
		0.052699,
		0.068159,
		0.085389,
		0.10431,
		0.124837,
		0.146884,
		0.17036,
		0.195168,
		0.221211,
		0.248386,
		0.276585,
		0.3057,
		0.335615,
		0.366214,
		0.397375,
		0.428973,
		0.460882,
		0.492972,
		0.525108,
		0.557157,
		0.588982,
		0.620445,
		0.651406,
		0.681729,
		0.711275,
		0.739908,
		0.767497,
		0.793912,
		0.819031,
		0.842738,
		0.864928,
		0.885504,
		0.904386,
		0.921508,
		0.936823,
		0.950308,
		0.961963,
		0.971816,
		0.979927,
		0.986389,
		0.99133,
		0.994912,
		0.997332,
		0.998811,
		0.999591,
		0.999912,
		0.999994,
		1,
		0.999994,
		0.999903,
		0.999529,
		0.998569,
		0.996651,
		0.99335,
		0.988221,
		0.980811,
		0.970681,
		0.957423,
		0.940665,
		0.92009,
		0.895439,
		0.86652,
		0.833218,
		0.795497,
		0.75341,
		0.707107,
		0.656837,
		0.602961,
		0.545959,
		0.486437,
		0.425141,
		0.362964,
		0.300964,
		0.24037,
		0.182603,
		0.129291,
		0.082283,
		0.043674,
		0.015823,
		0.001374,
		0.000492,
		0.003055,
		0.007758,
		0.014547,
		0.023363,
		0.034144,
		0.046826,
		0.061339,
		0.077611,
		0.095568,
		0.11513,
		0.136215,
		0.158736,
		0.182603,
		0.207724,
		0.234001,
		0.261333,
		0.289616,
		0.318742,
		0.348599,
		0.379073,
		0.410046,
		0.441396,
		0.473,
		0.504731,
		0.53646,
		0.568056,
		0.599388,
		0.630323,
		0.660725,
		0.690463,
		0.719404,
		0.747418,
		0.774379,
		0.800163,
		0.824654,
		0.847743,
		0.869329,
		0.889324,
		0.907651,
		0.924251,
		0.939084,
		0.952128,
		0.963389,
		0.972897,
		0.980715,
		0.986936,
		0.991686,
		0.995126,
		0.997446,
		0.998863,
		0.999609,
		0.999916,
		0.999994,
		1,
		0.999994,
		0.999908,
		0.999551,
		0.998635,
		0.996803,
		0.993648,
		0.988742,
		0.981647,
		0.971939,
		0.95922,
		0.943128,
		0.92335,
		0.899627,
		0.871765,
		0.83964,
		0.803203,
		0.762488,
		0.717621,
		0.668821,
		0.61641,
		0.560823,
		0.502614,
		0.442464,
		0.381194,
		0.319778,
		0.259348,
		0.201216,
		0.146884,
		0.098062,
		0.056682,
		0.024926,
		0.005236,
		0.00012,
		0.001913,
		0.005814,
		0.011771,
		0.019728,
		0.029626,
		0.041405,
		0.055,
		0.070342,
		0.08736,
		0.10598,
		0.126123,
		0.147708,
		0.170651,
		0.194862,
		0.22025,
		0.246719,
		0.27417,
		0.302502,
		0.331609,
		0.361381,
		0.391707,
		0.422472,
		0.453557,
		0.484843,
		0.516206,
		0.547521,
		0.578663,
		0.609502,
		0.63991,
		0.669758,
		0.698919,
		0.727265,
		0.754672,
		0.781018,
		0.806187,
		0.830067,
		0.852555,
		0.873557,
		0.892988,
		0.91078,
		0.926878,
		0.941245,
		0.953866,
		0.964748,
		0.973927,
		0.981465,
		0.987455,
		0.992024,
		0.995329,
		0.997555,
		0.998913,
		0.999626,
		0.99992,
		0.999995,
		1,
		0.999994,
		0.999912,
		0.999571,
		0.998697,
		0.996946,
		0.99393,
		0.989234,
		0.982438,
		0.97313,
		0.960924,
		0.945467,
		0.926448,
		0.903612,
		0.876762,
		0.845765,
		0.810562,
		0.771171,
		0.727694,
		0.680324,
		0.629348,
		0.575159,
		0.518262,
		0.459281,
		0.398972,
		0.338228,
		0.278096,
		0.219786,
		0.164687,
		0.114378,
		0.070648,
		0.035512,
		0.011231,
		0.000331,
		0.001053,
		0.004182,
		0.009337,
		0.016464,
		0.02551,
		0.036415,
		0.049119,
		0.063557,
		0.079662,
		0.097363,
		0.116587,
		0.137255,
		0.159289,
		0.182603,
		0.207112,
		0.232725,
		0.259348,
		0.286884,
		0.315234,
		0.344293,
		0.373956,
		0.404113,
		0.434652,
		0.465459,
		0.496415,
		0.527402,
		0.558299,
		0.588982,
		0.619329,
		0.649214,
		0.678515,
		0.707107,
		0.734868,
		0.76168,
		0.787425,
		0.811993,
		0.835279,
		0.857183,
		0.877618,
		0.896506,
		0.913781,
		0.929393,
		0.943313,
		0.955526,
		0.966046,
		0.974909,
		0.982179,
		0.987949,
		0.992346,
		0.995521,
		0.997658,
		0.99896,
		0.999643,
		0.999923,
		0.999995,
		1,
		0.999995,
		0.999916,
		0.999591,
		0.998756,
		0.997082,
		0.994196,
		0.9897,
		0.983187,
		0.97426,
		0.962541,
		0.947687,
		0.929393,
		0.907405,
		0.881522,
		0.851607,
		0.817592,
		0.779477,
		0.737347,
		0.691366,
		0.641793,
		0.588982,
		0.533393,
		0.475597,
		0.416289,
		0.356293,
		0.296573,
		0.238249,
		0.182603,
		0.131098,
		0.085389,
		0.047341,
		0.019046,
		0.002844,
		0.000458,
		0.002844,
		0.007225,
		0.013553,
		0.021774,
		0.031834,
		0.043674,
		0.057235,
		0.072452,
		0.089257,
		0.107582,
		0.127354,
		0.148496,
		0.170928,
		0.19457,
		0.219335,
		0.245133,
		0.271875,
		0.299464,
		0.327802,
		0.356789,
		0.38632,
		0.416289,
		0.446587,
		0.477101,
		0.507719,
		0.538323,
		0.568797,
		0.599022,
		0.628878,
		0.658245,
		0.687003,
		0.715035,
		0.742222,
		0.76845,
		0.793609,
		0.817592,
		0.840298,
		0.861636,
		0.881522,
		0.899882,
		0.916658,
		0.931803,
		0.945291,
		0.957114,
		0.967285,
		0.975845,
		0.982859,
		0.98842,
		0.992651,
		0.995704,
		0.997755,
		0.999004,
		0.999658,
		0.999927,
		0.999995,
		1,
		0.999995,
		0.99992,
		0.999609,
		0.998811,
		0.99721,
		0.994448,
		0.990141,
		0.983896,
		0.975331,
		0.964076,
		0.949797,
		0.932195,
		0.911016,
		0.886059,
		0.857183,
		0.824309,
		0.787425,
		0.746597,
		0.701966,
		0.653763,
		0.602308,
		0.548017,
		0.491417,
		0.433143,
		0.373956,
		0.314747,
		0.25655,
		0.200552,
		0.148107,
		0.100747,
		0.060202,
		0.02841,
		0.007538,
		0.000112,
		0.001783,
		0.005419,
		0.010974,
		0.018399,
		0.027641,
		0.038645,
		0.051353,
		0.065706,
		0.08164,
		0.099088,
		0.117983,
		0.13825,
		0.159817,
		0.182603,
		0.206529,
		0.231511,
		0.25746,
		0.284287,
		0.311899,
		0.3402,
		0.369091,
		0.39847,
		0.428234,
		0.458277,
		0.488488,
		0.518758,
		0.548975,
		0.579023,
		0.608789,
		0.638156,
		0.667009,
		0.695232,
		0.722712,
		0.749335,
		0.774992,
		0.799577,
		0.82299,
		0.845133,
		0.865921,
		0.885274,
		0.903125,
		0.919418,
		0.934112,
		0.947185,
		0.958631,
		0.968469,
		0.976739,
		0.983507,
		0.988867,
		0.992941,
		0.995877,
		0.997848,
		0.999046,
		0.999673,
		0.99993,
		0.999995,
		1,
		0.999995,
		0.999923,
		0.999626,
		0.998863,
		0.997332,
		0.994686,
		0.990558,
		0.984569,
		0.976347,
		0.965534,
		0.951803,
		0.93486,
		0.914455,
		0.890386,
		0.862506,
		0.830729,
		0.795032,
		0.755463,
		0.712144,
		0.665277,
		0.615152,
		0.562148,
		0.506746,
		0.449532,
		0.391206,
		0.332591,
		0.274643,
		0.218463,
		0.165303,
		0.116587,
		0.073915,
		0.039085,
		0.014108,
		0.001221,
		0.000982,
		0.003901,
		0.008711,
		0.015366,
		0.023816,
		0.034009,
		0.045892,
		0.059405,
		0.07449,
		0.091084,
		0.109121,
		0.128533,
		0.149248,
		0.171193,
		0.194292,
		0.218463,
		0.243624,
		0.269691,
		0.296573,
		0.324181,
		0.35242,
		0.381194,
		0.410404,
		0.439948,
		0.469722,
		0.499621,
		0.529537,
		0.559361,
		0.588982,
		0.618289,
		0.647171,
		0.675515,
		0.70321,
		0.730146,
		0.756216,
		0.781314,
		0.80534,
		0.828196,
		0.849792,
		0.870045,
		0.888882,
		0.90624,
		0.922066,
		0.936326,
		0.948998,
		0.960083,
		0.9696,
		0.977591,
		0.984125,
		0.989293,
		0.993217,
		0.996042,
		0.997936,
		0.999086,
		0.999687,
		0.999933,
		0.999995,
		1,
		0.999995,
		0.999927,
		0.999643,
		0.998913,
		0.997446,
		0.994912,
		0.990954,
		0.985208,
		0.977312,
		0.96692,
		0.953711,
		0.937398,
		0.917733,
		0.894513,
		0.86759,
		0.836868,
		0.802314,
		0.763962,
		0.721915,
		0.676351,
		0.62753,
		0.575799,
		0.521595,
		0.465459,
		0.408035,
		0.350084,
		0.292492,
		0.236278,
		0.182603,
		0.132788,
		0.088318,
		0.050862,
		0.022285,
		0.004665,
		0.000428,
		0.002655,
		0.006746,
		0.012657,
		0.020342,
		0.02975,
		0.04083,
		0.053528,
		0.067787,
		0.083548,
		0.100747,
		0.119322,
		0.139203,
		0.160321,
		0.182603,
		0.205974,
		0.230353,
		0.255662,
		0.281814,
		0.308725,
		0.336303,
		0.364459,
		0.393097,
		0.42212,
		0.45143,
		0.480925,
		0.510503,
		0.540059,
		0.569487,
		0.59868,
		0.62753,
		0.655929,
		0.68377,
		0.710944,
		0.737347,
		0.762874,
		0.787425,
		0.810904,
		0.833218,
		0.854281,
		0.874016,
		0.892353,
		0.909232,
		0.924608,
		0.938448,
		0.950736,
		0.961473,
		0.970681,
		0.978406,
		0.984714,
		0.9897,
		0.99348,
		0.996199,
		0.998019,
		0.999123,
		0.9997,
		0.999936,
		0.999996,
		1,
		0.999995,
		0.99993,
		0.999658,
		0.99896,
		0.997555,
		0.995126,
		0.99133,
		0.985814,
		0.978228,
		0.968237,
		0.955526,
		0.939815,
		0.920857,
		0.898452,
		0.872446,
		0.842738,
		0.809287,
		0.772112,
		0.731298,
		0.687003,
		0.63946,
		0.588982,
		0.535972,
		0.480925,
		0.424437,
		0.367211,
		0.310066,
		0.253947,
		0.199933,
		0.149248,
		0.103273,
		0.063557,
		0.031834,
		0.010032,
		0.000295,
		0.000105,
		0.001665,
		0.005063,
		0.010256,
		0.0172,
		0.025848,
		0.03615,
		0.048056,
		0.061512,
		0.076461,
		0.092843,
		0.110599,
		0.129662,
		0.149968,
		0.171447,
		0.194026,
		0.217631,
		0.242186,
		0.26761,
		0.29382,
		0.320733,
		0.34826,
		0.376311,
		0.404796,
		0.433618,
		0.462682,
		0.491888,
		0.521138,
		0.55033,
		0.579359,
		0.608124,
		0.636519,
		0.664439,
		0.691782,
		0.718443,
		0.744321,
		0.769316,
		0.793332,
		0.816277,
		0.838063,
		0.858608,
		0.877839,
		0.895691,
		0.912108,
		0.927049,
		0.940484,
		0.952401,
		0.962803,
		0.971715,
		0.979184,
		0.985277,
		0.990087,
		0.993731,
		0.996348,
		0.998099,
		0.999159,
		0.999713,
		0.999939,
		0.999996,
		1,
		0.999996,
		0.999933,
		0.999673,
		0.999004,
		0.997658,
		0.995329,
		0.991686,
		0.986389,
		0.979099,
		0.969489,
		0.957254,
		0.942118,
		0.923837,
		0.902212,
		0.877086,
		0.848355,
		0.815966,
		0.779928,
		0.74031,
		0.69725,
		0.650955,
		0.601712,
		0.549888,
		0.495938,
		0.440412,
		0.383959,
		0.32734,
		0.271429,
		0.21723,
		0.165879,
		0.118659,
		0.077011,
		0.042546,
		0.017058,
		0.002539,
		0.000918,
		0.003647,
		0.008146,
		0.014374,
		0.022285,
		0.031834,
		0.042971,
		0.055645,
		0.069803,
		0.085389,
		0.102344,
		0.120607,
		0.140116,
		0.160804,
		0.182603,
		0.205443,
		0.229249,
		0.253947,
		0.279458,
		0.3057,
		0.332591,
		0.360045,
		0.387975,
		0.416289,
		0.444897,
		0.473704,
		0.502614,
		0.531529,
		0.560352,
		0.588982,
		0.617319,
		0.645261,
		0.672708,
		0.699558,
		0.725714,
		0.751077,
		0.775551,
		0.799043,
		0.821467,
		0.842738,
		0.86278,
		0.881522,
		0.898903,
		0.914874,
		0.929393,
		0.942438,
		0.953996,
		0.964076,
		0.972704,
		0.979927,
		0.985814,
		0.990456,
		0.993969,
		0.99649,
		0.998174,
		0.999193,
		0.999725,
		0.999941,
		0.999996,
		1,
		0.999996,
		0.999936,
		0.999687,
		0.999046,
		0.997755,
		0.995521,
		0.992024,
		0.986936,
		0.979927,
		0.970681,
		0.9589,
		0.944313,
		0.92668,
		0.905802,
		0.881522,
		0.853729,
		0.822365,
		0.787425,
		0.748966,
		0.707107,
		0.662033,
		0.614003,
		0.563354,
		0.510503,
		0.455958,
		0.40032,
		0.344293,
		0.28869,
		0.234441,
		0.182603,
		0.134371,
		0.091084,
		0.05424,
		0.02551,
		0.006746,
		0.0004,
		0.002483,
		0.006313,
		0.011847,
		0.019046,
		0.027864,
		0.038254,
		0.050169,
		0.063557,
		0.078366,
		0.094539,
		0.112019,
		0.130747,
		0.150658,
		0.171689,
		0.193772,
		0.216838,
		0.240813,
		0.265625,
		0.291195,
		0.317445,
		0.344293,
		0.371655,
		0.399446,
		0.427577,
		0.455958,
		0.484498,
		0.513104,
		0.541681,
		0.570132,
		0.598361,
		0.626271,
		0.653763,
		0.680741,
		0.707107,
		0.732765,
		0.757621,
		0.781584,
		0.804566,
		0.826481,
		0.847251,
		0.866803,
		0.88507,
		0.901995,
		0.917532,
		0.931646,
		0.944313,
		0.955526,
		0.965297,
		0.973651,
		0.980638,
		0.986327,
		0.990808,
		0.994196,
		0.996625,
		0.998246,
		0.999225,
		0.999736,
		0.999944,
		0.999996,
		1,
		0.999996,
		0.999939,
		0.9997,
		0.999086,
		0.997848,
		0.995704,
		0.992346,
		0.987455,
		0.980715,
		0.971816,
		0.960468,
		0.946406,
		0.929393,
		0.909232,
		0.885764,
		0.858873,
		0.828496,
		0.794618,
		0.757282,
		0.71659,
		0.672708,
		0.625869,
		0.576381,
		0.524629,
		0.471079,
		0.416289,
		0.360911,
		0.3057,
		0.25152,
		0.199356,
		0.150317,
		0.105651,
		0.066755,
		0.035182,
		0.012657,
		0.001092,
		9.8e-005,
		0.001559,
		0.004741,
		0.009606,
		0.016114,
		0.024223,
		0.033889,
		0.045066,
		0.057705,
		0.071756,
		0.087167,
		0.103881,
		0.121842,
		0.140991,
		0.161266,
		0.182603,
		0.204936,
		0.228195,
		0.252311,
		0.277209,
		0.302814,
		0.32905,
		0.355834,
		0.383087,
		0.410723,
		0.438658,
		0.466802,
		0.495067,
		0.523363,
		0.551596,
		0.579674,
		0.607502,
		0.634987,
		0.662033,
		0.688547,
		0.714434,
		0.739603,
		0.763962,
		0.787425,
		0.809907,
		0.831326,
		0.851607,
		0.870682,
		0.888488,
		0.904972,
		0.92009,
		0.93381,
		0.946113,
		0.956994,
		0.966466,
		0.974557,
		0.981318,
		0.986817,
		0.991144,
		0.994413,
		0.996753,
		0.998314,
		0.999256,
		0.999747,
		0.999946,
		0.999996,
		1,
		0.999996,
		0.999941,
		0.999713,
		0.999123,
		0.997936,
		0.995877,
		0.992651,
		0.987949,
		0.981465,
		0.972897,
		0.961963,
		0.948403,
		0.931984,
		0.91251,
		0.889821,
		0.863799,
		0.834373,
		0.801521,
		0.765272,
		0.725714,
		0.682994,
		0.637323,
		0.588982,
		0.538323,
		0.485779,
		0.431863,
		0.377182,
		0.322436,
		0.26843,
		0.216079,
		0.166417,
		0.120607,
		0.079949,
		0.045892,
		0.020042,
		0.004182,
		0.00086,
		0.003417,
		0.007634,
		0.013475,
		0.020897,
		0.02986,
		0.04032,
		0.05223,
		0.065543,
		0.080209,
		0.096175,
		0.113386,
		0.131787,
		0.151319,
		0.171921,
		0.193529,
		0.216079,
		0.239503,
		0.26373,
		0.28869,
		0.314308,
		0.340507,
		0.367211,
		0.394338,
		0.421806,
		0.449532,
		0.47743,
		0.505413,
		0.533393,
		0.561279,
		0.588982,
		0.61641,
		0.643472,
		0.670076,
		0.696131,
		0.721547,
		0.746234,
		0.770107,
		0.79308,
		0.815072,
		0.836008,
		0.855814,
		0.874425,
		0.891784,
		0.907839,
		0.922551,
		0.93589,
		0.947842,
		0.958402,
		0.967586,
		0.975425,
		0.981968,
		0.987285,
		0.991466,
		0.99462,
		0.996876,
		0.998379,
		0.999285,
		0.999757,
		0.999948,
		0.999997,
		1,
		0.999996,
		0.999944,
		0.999725,
		0.999159,
		0.998019,
		0.996042,
		0.992941,
		0.98842,
		0.982179,
		0.973927,
		0.963389,
		0.950308,
		0.934459,
		0.915644,
		0.893703,
		0.868517,
		0.840008,
		0.808146,
		0.77295,
		0.734494,
		0.692907,
		0.64838,
		0.601168,
		0.551596,
		0.500061,
		0.447041,
		0.393097,
		0.33888,
		0.285138,
		0.232725,
		0.182603,
		0.135857,
		0.093699,
		0.057479,
		0.028698,
		0.009016,
		0.000264,
		0.000375,
		0.002328,
		0.00592,
		0.011113,
		0.01787,
		0.026151,
		0.035914,
		0.047116,
		0.05971,
		0.073649,
		0.088884,
		0.105362,
		0.12303,
		0.141832,
		0.16171,
		0.182603,
		0.204451,
		0.227188,
		0.250747,
		0.275061,
		0.300059,
		0.325668,
		0.351814,
		0.378419,
		0.405406,
		0.432694,
		0.460201,
		0.487844,
		0.515539,
		0.543198,
		0.570735,
		0.598062,
		0.625091,
		0.651733,
		0.677898,
		0.7035,
		0.728452,
		0.752666,
		0.776061,
		0.798554,
		0.820069,
		0.840533,
		0.859876,
		0.878037,
		0.89496,
		0.9106,
		0.924919,
		0.937891,
		0.949502,
		0.959754,
		0.968661,
		0.976257,
		0.982591,
		0.987733,
		0.991772,
		0.994817,
		0.996992,
		0.998441,
		0.999313,
		0.999766,
		0.99995,
		0.999997,
		1,
		0.999997,
		0.999946,
		0.999736,
		0.999193,
		0.998099,
		0.996199,
		0.993217,
		0.988867,
		0.982859,
		0.974909,
		0.964748,
		0.952128,
		0.936823,
		0.918641,
		0.89742,
		0.873037,
		0.845412,
		0.814507,
		0.78033,
		0.742944,
		0.702461,
		0.659052,
		0.612952,
		0.564456,
		0.513933,
		0.461824,
		0.40865,
		0.355015,
		0.301618,
		0.249252,
		0.198815,
		0.151319,
		0.107895,
		0.069803,
		0.038444,
		0.015366,
		0.00228,
		9.2e-005,
		0.001462,
		0.004448,
		0.009016,
		0.015128,
		0.022747,
		0.031834,
		0.042346,
		0.05424,
		0.067471,
		0.081992,
		0.097753,
		0.114702,
		0.132788,
		0.151954,
		0.172143,
		0.193297,
		0.215353,
		0.238249,
		0.261919,
		0.286296,
		0.31131,
		0.336891,
		0.362964,
		0.389456,
		0.416289,
		0.443386,
		0.470665,
		0.498046,
		0.525446,
		0.552782,
		0.579968,
		0.606919,
		0.63355,
		0.659775,
		0.685507,
		0.710662,
		0.735156,
		0.758906,
		0.781832,
		0.803856,
		0.824905,
		0.844907,
		0.863799,
		0.881522,
		0.898023,
		0.91326,
		0.927198,
		0.939815,
		0.951098,
		0.961052,
		0.969692,
		0.977053,
		0.983187,
		0.988161,
		0.992065,
		0.995005,
		0.997104,
		0.998499,
		0.99934,
		0.999775,
		0.999952,
		0.999997,
		1,
		0.999997,
		0.999948,
		0.999747,
		0.999225,
		0.998174,
		0.996348,
		0.99348,
		0.989293,
		0.983507,
		0.975845,
		0.966046,
		0.953866,
		0.939084,
		0.921508,
		0.900978,
		0.877369,
		0.850596,
		0.820615,
		0.787425,
		0.751077,
		0.711668,
		0.669354,
		0.624345,
		0.576915,
		0.527402,
		0.476215,
		0.423837,
		0.370832,
		0.317848,
		0.265625,
		0.215002,
		0.166922,
		0.122442,
		0.082738,
		0.049119,
		0.023032,
		0.006072,
		0.000807,
		0.003208,
		0.00717,
		0.012657,
		0.019635,
		0.028064,
		0.037906,
		0.049119,
		0.06166,
		0.075484,
		0.090543,
		0.106789,
		0.124172,
		0.142639,
		0.162135,
		0.182603,
		0.203986,
		0.226223,
		0.249252,
		0.273008,
		0.297425,
		0.322436,
		0.34797,
		0.373956,
		0.40032,
		0.426988,
		0.453882,
		0.480925,
		0.508038,
		0.535139,
		0.562148,
		0.588982,
		0.615558,
		0.641793,
		0.667604,
		0.692907,
		0.717621,
		0.741665,
		0.764959,
		0.787425,
		0.80899,
		0.829583,
		0.849137,
		0.86759,
		0.884886,
		0.900978,
		0.915824,
		0.929393,
		0.941666,
		0.952633,
		0.962298,
		0.970681,
		0.977817,
		0.983758,
		0.988571,
		0.992346,
		0.995185,
		0.99721,
		0.998556,
		0.999365,
		0.999784,
		0.999954,
		0.999997,
		1,
		0.999997,
		0.99995,
		0.999757,
		0.999256,
		0.998246,
		0.99649,
		0.993731,
		0.9897,
		0.984125,
		0.976739,
		0.967285,
		0.955526,
		0.941245,
		0.924251,
		0.904386,
		0.881522,
		0.85557,
		0.826481,
		0.794247,
		0.758906,
		0.720543,
		0.679298,
		0.63536,
		0.588982,
		0.540475,
		0.490217,
		0.438658,
		0.38632,
		0.333809,
		0.281814,
		0.231119,
		0.182603,
		0.137255,
		0.096175,
		0.060583,
		0.031834,
		0.011419,
		0.000982,
		0.005562,
		0.010445,
		0.0168,
		0.024591,
		0.033782,
		0.044332,
		0.0562,
		0.069344,
		0.083718,
		0.099276,
		0.11597,
		0.13375,
		0.152563,
		0.172357,
		0.193074,
		0.214658,
		0.237049,
		0.260186,
		0.284006,
		0.308444,
		0.333432,
		0.358903,
		0.384786,
		0.41101,
		0.437501,
		0.464184,
		0.490984,
		0.517822,
		0.544621,
		0.571301,
		0.597782,
		0.623984,
		0.649825,
		0.675225,
		0.700105,
		0.724384,
		0.747985,
		0.770831,
		0.792848,
		0.813964,
		0.834112,
		0.853227,
		0.871252,
		0.888134,
		0.903828,
		0.918295,
		0.931507,
		0.943447,
		0.954108,
		0.963495,
		0.971631,
		0.97855,
		0.984305,
		0.988964,
		0.992614,
		0.995357,
		0.997312,
		0.998609,
		0.999389,
		0.999793,
		0.999956,
		0.999997,
		1,
		0.999997,
		0.999952,
		0.999766,
		0.999285,
		0.998314,
		0.996625,
		0.993969,
		0.990087,
		0.984714,
		0.977591,
		0.968469,
		0.957114,
		0.943313,
		0.926878,
		0.907651,
		0.885504,
		0.860344,
		0.832117,
		0.800807,
		0.766443,
		0.729099,
		0.688896,
		0.64601,
		0.600669,
		0.553161,
		0.503836,
		0.453111,
		0.401474,
		0.349487,
		0.297795,
		0.247127,
		0.198309,
		0.152262,
		0.110014,
		0.07271,
		0.041614,
		0.018122,
		0.003771,
		0.021402,
		0.029959,
		0.039864,
		0.051077,
		0.063557,
		0.077262,
		0.092147,
		0.108167,
		0.125272,
		0.143415,
		0.162543,
		0.182603,
		0.203541,
		0.2253,
		0.247821,
		0.271043,
		0.294906,
		0.319344,
		0.344293,
		0.369686,
		0.395453,
		0.421525,
		0.447829,
		0.474293,
		0.500843,
		0.527402,
		0.553895,
		0.580244,
		0.606372,
		0.632201,
		0.657651,
		0.682646,
		0.707107,
		0.730958,
		0.754123,
		0.776529,
		0.798105,
		0.818782,
		0.838495,
		0.857183,
		0.874792,
		0.891271,
		0.906578,
		0.920678,
		0.933544,
		0.945162,
		0.955526,
		0.964646,
		0.972543,
		0.979253,
		0.984829,
		0.98934,
		0.99287,
		0.995521,
		0.997409,
		0.99866,
		0.999412,
		0.9998,
		0.999958,
		0.999997,
		1,
		0.999997,
		0.999954,
		0.999775,
		0.999313,
		0.998379,
		0.996753,
		0.994196,
		0.990456,
		0.985277,
		0.978406,
		0.9696,
		0.958631,
		0.945291,
		0.929393,
		0.91078,
		0.889324,
		0.864928,
		0.837533,
		0.807118,
		0.773702,
		0.737347,
		0.698162,
		0.656305,
		0.611985,
		0.565468,
		0.517077,
		0.4672,
		0.416289,
		0.364871,
		0.313545,
		0.262996,
		0.213992,
		0.167396,
		0.124172,
		0.085389,
		0.05223,
		0.026003,
		0.008146,
		0.000238,
		0.046277,
		0.058111,
		0.071162,
		0.085389,
		0.100747,
		0.117192,
		0.134676,
		0.153149,
		0.172561,
		0.19286,
		0.213992,
		0.2359,
		0.258527,
		0.281814,
		0.3057,
		0.330122,
		0.355015,
		0.380315,
		0.405954,
		0.431863,
		0.457972,
		0.48421,
		0.510503,
		0.53678,
		0.562964,
		0.588982,
		0.614757,
		0.640214,
		0.665277,
		0.68987,
		0.713918,
		0.737347,
		0.760084,
		0.782059,
		0.803203,
		0.82345,
		0.842738,
		0.861011,
		0.878214,
		0.894301,
		0.909232,
		0.922975,
		0.935507,
		0.946813,
		0.956891,
		0.965752,
		0.973418,
		0.979927,
		0.985332,
		0.9897,
		0.993115,
		0.995678,
		0.997501,
		0.998709,
		0.999434,
		0.999808,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999956,
		0.999784,
		0.99934,
		0.998441,
		0.996876,
		0.994413,
		0.990808,
		0.985814,
		0.979184,
		0.970681,
		0.960083,
		0.947185,
		0.931803,
		0.913781,
		0.892988,
		0.869329,
		0.842738,
		0.813189,
		0.780692,
		0.745299,
		0.707107,
		0.666257,
		0.622942,
		0.577405,
		0.529947,
		0.480925,
		0.430764,
		0.379951,
		0.32905,
		0.278696,
		0.229612,
		0.182603,
		0.138573,
		0.098521,
		0.063557,
		0.034906,
		0.013913,
		0.002059,
		0.078987,
		0.093699,
		0.109496,
		0.126332,
		0.144161,
		0.162935,
		0.182603,
		0.203114,
		0.224414,
		0.246449,
		0.269161,
		0.292492,
		0.316383,
		0.340772,
		0.365596,
		0.39079,
		0.416289,
		0.442026,
		0.467931,
		0.493936,
		0.519969,
		0.545959,
		0.571833,
		0.597518,
		0.622942,
		0.648029,
		0.672708,
		0.696903,
		0.720543,
		0.743557,
		0.765874,
		0.787425,
		0.808146,
		0.827973,
		0.846847,
		0.864714,
		0.881522,
		0.897228,
		0.911794,
		0.925192,
		0.937398,
		0.948403,
		0.958205,
		0.966816,
		0.97426,
		0.980575,
		0.985814,
		0.990045,
		0.99335,
		0.995829,
		0.99759,
		0.998756,
		0.999454,
		0.999815,
		0.999961,
		0.999997,
		1,
		0.999997,
		0.999958,
		0.999793,
		0.999365,
		0.998499,
		0.996992,
		0.99462,
		0.991144,
		0.986327,
		0.979927,
		0.971715,
		0.961473,
		0.948998,
		0.934112,
		0.916658,
		0.896506,
		0.873557,
		0.847743,
		0.819031,
		0.787425,
		0.752968,
		0.715743,
		0.675879,
		0.63355,
		0.588982,
		0.542452,
		0.494292,
		0.444897,
		0.394723,
		0.344293,
		0.294205,
		0.245133,
		0.197833,
		0.153149,
		0.112019,
		0.075484,
		0.044689,
		0.020897,
		0.005495,
		0.118371,
		0.135567,
		0.153712,
		0.172758,
		0.192655,
		0.213353,
		0.234798,
		0.256937,
		0.279714,
		0.303071,
		0.32695,
		0.351291,
		0.376031,
		0.401108,
		0.426457,
		0.452013,
		0.477707,
		0.503473,
		0.529241,
		0.554942,
		0.580504,
		0.605858,
		0.63093,
		0.655651,
		0.679948,
		0.703751,
		0.726989,
		0.749594,
		0.771497,
		0.792634,
		0.812941,
		0.832357,
		0.850827,
		0.868297,
		0.884721,
		0.900056,
		0.914268,
		0.92733,
		0.939222,
		0.949934,
		0.959469,
		0.967838,
		0.975068,
		0.981196,
		0.986276,
		0.990375,
		0.993576,
		0.995972,
		0.997675,
		0.9988,
		0.999474,
		0.999822,
		0.999962,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.9998,
		0.999389,
		0.998556,
		0.997104,
		0.994817,
		0.991466,
		0.986817,
		0.980638,
		0.972704,
		0.962803,
		0.950736,
		0.936326,
		0.919418,
		0.899882,
		0.877618,
		0.852555,
		0.824654,
		0.793912,
		0.760363,
		0.724081,
		0.68518,
		0.643821,
		0.600208,
		0.5546,
		0.507305,
		0.45869,
		0.409179,
		0.359265,
		0.309504,
		0.260527,
		0.213043,
		0.167843,
		0.125807,
		0.08791,
		0.055226,
		0.028941,
		0.010354,
		0.000888,
		0.163312,
		0.182603,
		0.202704,
		0.223565,
		0.245133,
		0.267356,
		0.290179,
		0.313545,
		0.337397,
		0.361676,
		0.38632,
		0.411269,
		0.436458,
		0.461824,
		0.487302,
		0.512824,
		0.538323,
		0.563732,
		0.588982,
		0.614003,
		0.638727,
		0.663083,
		0.687003,
		0.710418,
		0.73326,
		0.755463,
		0.77696,
		0.79769,
		0.817592,
		0.836606,
		0.854681,
		0.871765,
		0.887815,
		0.90279,
		0.916658,
		0.929393,
		0.94098,
		0.95141,
		0.960686,
		0.968823,
		0.975845,
		0.981794,
		0.98672,
		0.990693,
		0.993791,
		0.99611,
		0.997755,
		0.998843,
		0.999493,
		0.999829,
		0.999964,
		0.999998,
		1,
		0.999997,
		0.999961,
		0.999808,
		0.999412,
		0.998609,
		0.99721,
		0.995005,
		0.991772,
		0.987285,
		0.981318,
		0.973651,
		0.964076,
		0.952401,
		0.938448,
		0.922066,
		0.903125,
		0.881522,
		0.857183,
		0.830067,
		0.800163,
		0.767497,
		0.732133,
		0.694173,
		0.653763,
		0.611093,
		0.566399,
		0.519969,
		0.472143,
		0.42332,
		0.373956,
		0.324575,
		0.275766,
		0.228195,
		0.182603,
		0.139816,
		0.100747,
		0.066407,
		0.037906,
		0.016464,
		0.003417,
		0.212739,
		0.233741,
		0.255412,
		0.2777,
		0.300551,
		0.32391,
		0.34772,
		0.371923,
		0.39646,
		0.42127,
		0.446292,
		0.471461,
		0.496716,
		0.52199,
		0.547218,
		0.572334,
		0.59727,
		0.62196,
		0.646337,
		0.670332,
		0.693878,
		0.71691,
		0.739362,
		0.761169,
		0.782268,
		0.802599,
		0.822103,
		0.840726,
		0.858415,
		0.875123,
		0.890808,
		0.905432,
		0.918966,
		0.931385,
		0.942676,
		0.952833,
		0.961859,
		0.96977,
		0.976593,
		0.982367,
		0.987147,
		0.990997,
		0.993998,
		0.996242,
		0.997833,
		0.998883,
		0.999511,
		0.999835,
		0.999965,
		0.999998,
		1,
		0.999998,
		0.999962,
		0.999815,
		0.999434,
		0.99866,
		0.997312,
		0.995185,
		0.992065,
		0.987733,
		0.981968,
		0.974557,
		0.965297,
		0.953996,
		0.940484,
		0.924608,
		0.90624,
		0.885274,
		0.861636,
		0.835279,
		0.806187,
		0.774379,
		0.739908,
		0.702868,
		0.663389,
		0.621646,
		0.577857,
		0.53229,
		0.485261,
		0.437142,
		0.38836,
		0.339405,
		0.29083,
		0.243258,
		0.197385,
		0.153986,
		0.113919,
		0.078131,
		0.047667,
		0.02367,
		0.007397,
		0.000216,
		0.265625,
		0.28796,
		0.310823,
		0.33416,
		0.357915,
		0.382031,
		0.40645,
		0.431113,
		0.455958,
		0.480925,
		0.505952,
		0.530974,
		0.555928,
		0.580749,
		0.605372,
		0.629732,
		0.653763,
		0.6774,
		0.700578,
		0.723232,
		0.745299,
		0.766717,
		0.787425,
		0.807365,
		0.826481,
		0.84472,
		0.862033,
		0.878374,
		0.893703,
		0.907987,
		0.921195,
		0.933308,
		0.944313,
		0.954204,
		0.962988,
		0.970681,
		0.977312,
		0.982919,
		0.987556,
		0.991289,
		0.994196,
		0.996368,
		0.997907,
		0.998922,
		0.999529,
		0.999841,
		0.999966,
		0.999998,
		1,
		0.999998,
		0.999964,
		0.999822,
		0.999454,
		0.998709,
		0.997409,
		0.995357,
		0.992346,
		0.988161,
		0.982591,
		0.975425,
		0.966466,
		0.955526,
		0.942438,
		0.927049,
		0.909232,
		0.888882,
		0.865921,
		0.840298,
		0.811993,
		0.781018,
		0.747418,
		0.711275,
		0.672708,
		0.631876,
		0.588982,
		0.544274,
		0.498046,
		0.450646,
		0.402472,
		0.353984,
		0.3057,
		0.258204,
		0.212149,
		0.168264,
		0.127354,
		0.090309,
		0.058111,
		0.031834,
		0.012657,
		0.001868,
		0.320992,
		0.344293,
		0.36798,
		0.391998,
		0.416289,
		0.440796,
		0.465459,
		0.490217,
		0.515011,
		0.539778,
		0.564456,
		0.588982,
		0.613292,
		0.637323,
		0.661012,
		0.684294,
		0.707107,
		0.729388,
		0.751077,
		0.772112,
		0.792437,
		0.811993,
		0.830729,
		0.848593,
		0.865538,
		0.881522,
		0.896506,
		0.910457,
		0.92335,
		0.935165,
		0.945892,
		0.955526,
		0.964076,
		0.971559,
		0.978003,
		0.983449,
		0.987949,
		0.991569,
		0.994386,
		0.99649,
		0.997978,
		0.99896,
		0.999545,
		0.999846,
		0.999968,
		0.999998,
		1,
		0.999998,
		0.999965,
		0.999829,
		0.999474,
		0.998756,
		0.997501,
		0.995521,
		0.992614,
		0.988571,
		0.983187,
		0.976257,
		0.967586,
		0.956994,
		0.944313,
		0.929393,
		0.912108,
		0.892353,
		0.870045,
		0.845133,
		0.817592,
		0.787425,
		0.754672,
		0.719404,
		0.681729,
		0.641793,
		0.599783,
		0.555928,
		0.510503,
		0.463832,
		0.416289,
		0.368303,
		0.32036,
		0.273008,
		0.226862,
		0.182603,
		0.140991,
		0.102863,
		0.069138,
		0.04083,
		0.019046,
		0.004996,
		0.377913,
		0.401822,
		0.425977,
		0.45032,
		0.474793,
		0.499339,
		0.523896,
		0.548406,
		0.572806,
		0.597036,
		0.621034,
		0.644738,
		0.668086,
		0.691017,
		0.713469,
		0.735383,
		0.756699,
		0.777359,
		0.797307,
		0.816488,
		0.834852,
		0.85235,
		0.868936,
		0.884571,
		0.899218,
		0.912847,
		0.925433,
		0.936959,
		0.947416,
		0.956802,
		0.965125,
		0.972405,
		0.978669,
		0.983959,
		0.988327,
		0.991839,
		0.994569,
		0.996606,
		0.998046,
		0.998995,
		0.999561,
		0.999852,
		0.999969,
		0.999998,
		1,
		0.999998,
		0.999966,
		0.999835,
		0.999493,
		0.9988,
		0.99759,
		0.995678,
		0.99287,
		0.988964,
		0.983758,
		0.977053,
		0.968661,
		0.958402,
		0.946113,
		0.931646,
		0.914874,
		0.895691,
		0.874016,
		0.849792,
		0.82299,
		0.793609,
		0.76168,
		0.727265,
		0.690463,
		0.651406,
		0.610268,
		0.567259,
		0.522638,
		0.476704,
		0.429809,
		0.382355,
		0.334797,
		0.287651,
		0.241492,
		0.196963,
		0.154777,
		0.11572,
		0.08066,
		0.050548,
		0.026425,
		0.009431,
		0.000807,
		0.435513,
		0.459685,
		0.483964,
		0.508291,
		0.532609,
		0.556859,
		0.58098,
		0.604914,
		0.628601,
		0.651979,
		0.67499,
		0.697574,
		0.71967,
		0.741222,
		0.762171,
		0.782462,
		0.80204,
		0.820854,
		0.838854,
		0.855993,
		0.87223,
		0.887524,
		0.901844,
		0.915158,
		0.927446,
		0.938692,
		0.948887,
		0.958032,
		0.966137,
		0.973219,
		0.97931,
		0.98445,
		0.988691,
		0.992097,
		0.994744,
		0.996717,
		0.998111,
		0.999029,
		0.999576,
		0.999857,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999968,
		0.999841,
		0.999511,
		0.998843,
		0.997675,
		0.995829,
		0.993115,
		0.98934,
		0.984305,
		0.977817,
		0.969692,
		0.959754,
		0.947842,
		0.93381,
		0.917532,
		0.898903,
		0.877839,
		0.854281,
		0.828196,
		0.799577,
		0.76845,
		0.734868,
		0.698919,
		0.660725,
		0.620445,
		0.578275,
		0.534454,
		0.489264,
		0.443032,
		0.396135,
		0.349001,
		0.302114,
		0.256014,
		0.211307,
		0.168662,
		0.128819,
		0.092596,
		0.060886,
		0.034672,
		0.015024,
		0.003111,
		0.492972,
		0.517077,
		0.541152,
		0.56514,
		0.588982,
		0.612621,
		0.635997,
		0.659052,
		0.681729,
		0.703969,
		0.725714,
		0.746909,
		0.767497,
		0.787425,
		0.806641,
		0.825095,
		0.842738,
		0.859528,
		0.875423,
		0.890386,
		0.904386,
		0.917395,
		0.929393,
		0.940367,
		0.950308,
		0.95922,
		0.967112,
		0.974004,
		0.979927,
		0.984922,
		0.98904,
		0.992346,
		0.994912,
		0.996824,
		0.998174,
		0.999062,
		0.999591,
		0.999862,
		0.999971,
		0.999998,
		1,
		0.999998,
		0.999969,
		0.999846,
		0.999529,
		0.998883,
		0.997755,
		0.995972,
		0.99335,
		0.9897,
		0.984829,
		0.97855,
		0.970681,
		0.961052,
		0.949502,
		0.93589,
		0.92009,
		0.901995,
		0.881522,
		0.858608,
		0.833218,
		0.80534,
		0.774992,
		0.742222,
		0.707107,
		0.669758,
		0.630323,
		0.588982,
		0.545959,
		0.501515,
		0.455958,
		0.409641,
		0.362964,
		0.316383,
		0.270407,
		0.225603,
		0.182603,
		0.142104,
		0.104874,
		0.071756,
		0.043674,
		0.021638,
		0.006746,
		0.000196,
		0.549528,
		0.573252,
		0.596815,
		0.620158,
		0.643226,
		0.66596,
		0.688306,
		0.710206,
		0.731605,
		0.752448,
		0.772681,
		0.792253,
		0.811114,
		0.829215,
		0.84651,
		0.862958,
		0.878519,
		0.893159,
		0.906848,
		0.91956,
		0.931276,
		0.941985,
		0.951681,
		0.960366,
		0.968053,
		0.974761,
		0.980522,
		0.985376,
		0.989376,
		0.992584,
		0.995074,
		0.996926,
		0.998234,
		0.999093,
		0.999605,
		0.999867,
		0.999972,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999852,
		0.999545,
		0.998922,
		0.997833,
		0.99611,
		0.993576,
		0.990045,
		0.985332,
		0.979253,
		0.971631,
		0.962298,
		0.951098,
		0.937891,
		0.922551,
		0.904972,
		0.88507,
		0.86278,
		0.838063,
		0.810904,
		0.781314,
		0.749335,
		0.715035,
		0.678515,
		0.63991,
		0.599388,
		0.557157,
		0.513462,
		0.468589,
		0.422868,
		0.376679,
		0.330446,
		0.28465,
		0.239825,
		0.196564,
		0.155525,
		0.117431,
		0.083077,
		0.053333,
		0.02915,
		0.011564,
		0.001703,
		0.604481,
		0.62753,
		0.650291,
		0.672708,
		0.694725,
		0.71629,
		0.737347,
		0.757843,
		0.777728,
		0.796951,
		0.815463,
		0.833218,
		0.850172,
		0.866286,
		0.881522,
		0.895847,
		0.909232,
		0.921655,
		0.933098,
		0.94355,
		0.953007,
		0.961473,
		0.96896,
		0.97549,
		0.981094,
		0.985814,
		0.9897,
		0.992814,
		0.995229,
		0.997025,
		0.998291,
		0.999123,
		0.999618,
		0.999871,
		0.999973,
		0.999998,
		1,
		0.999998,
		0.999971,
		0.999857,
		0.999561,
		0.99896,
		0.997907,
		0.996242,
		0.993791,
		0.990375,
		0.985814,
		0.979927,
		0.972543,
		0.963495,
		0.952633,
		0.939815,
		0.924919,
		0.907839,
		0.888488,
		0.866803,
		0.842738,
		0.816277,
		0.787425,
		0.756216,
		0.722712,
		0.687003,
		0.649214,
		0.609502,
		0.568056,
		0.525108,
		0.480925,
		0.435818,
		0.39014,
		0.344293,
		0.298728,
		0.253947,
		0.210511,
		0.169038,
		0.13021,
		0.094776,
		0.063557,
		0.03745,
		0.01743,
		0.004562,
		0.657196,
		0.679298,
		0.700991,
		0.722224,
		0.742944,
		0.7631,
		0.782641,
		0.801521,
		0.819691,
		0.837108,
		0.853729,
		0.869516,
		0.884434,
		0.898452,
		0.911542,
		0.923684,
		0.93486,
		0.945063,
		0.954288,
		0.962541,
		0.969836,
		0.976194,
		0.981647,
		0.986235,
		0.990011,
		0.993035,
		0.995378,
		0.997119,
		0.998347,
		0.999152,
		0.999631,
		0.999876,
		0.999974,
		0.999998,
		1,
		0.999998,
		0.999972,
		0.999862,
		0.999576,
		0.998995,
		0.997978,
		0.996368,
		0.993998,
		0.990693,
		0.986276,
		0.980575,
		0.973418,
		0.964646,
		0.954108,
		0.941666,
		0.927198,
		0.9106,
		0.891784,
		0.870682,
		0.847251,
		0.821467,
		0.793332,
		0.762874,
		0.730146,
		0.695232,
		0.658245,
		0.619329,
		0.578663,
		0.53646,
		0.492972,
		0.448489,
		0.403344,
		0.357915,
		0.312626,
		0.26795,
		0.224414,
		0.182603,
		0.14316,
		0.106789,
		0.074267,
		0.046438,
		0.024223,
		0.008627,
		0.000737,
		0.707107,
		0.728012,
		0.7484,
		0.76822,
		0.787425,
		0.805968,
		0.823803,
		0.840888,
		0.857183,
		0.872652,
		0.88726,
		0.900978,
		0.913781,
		0.925648,
		0.936566,
		0.946526,
		0.955526,
		0.963574,
		0.970681,
		0.976872,
		0.982179,
		0.986641,
		0.99031,
		0.993247,
		0.995521,
		0.99721,
		0.9984,
		0.99918,
		0.999643,
		0.99988,
		0.999975,
		0.999998,
		1,
		0.999998,
		0.999973,
		0.999867,
		0.999591,
		0.999029,
		0.998046,
		0.99649,
		0.994196,
		0.990997,
		0.98672,
		0.981196,
		0.97426,
		0.965752,
		0.955526,
		0.943447,
		0.929393,
		0.91326,
		0.89496,
		0.874425,
		0.851607,
		0.826481,
		0.799043,
		0.769316,
		0.737347,
		0.70321,
		0.667009,
		0.628878,
		0.588982,
		0.547521,
		0.504731,
		0.460882,
		0.416289,
		0.371306,
		0.326332,
		0.281814,
		0.238249,
		0.196187,
		0.156234,
		0.119058,
		0.085389,
		0.056024,
		0.031834,
		0.013765,
		0.002844,
		0.753719,
		0.77321,
		0.792083,
		0.810295,
		0.827802,
		0.844563,
		0.860539,
		0.875696,
		0.890001,
		0.903427,
		0.91595,
		0.927551,
		0.938217,
		0.947941,
		0.956724,
		0.964571,
		0.971498,
		0.977527,
		0.982692,
		0.987032,
		0.990599,
		0.993452,
		0.995659,
		0.997298,
		0.998451,
		0.999206,
		0.999655,
		0.999884,
		0.999976,
		0.999998,
		1,
		0.999998,
		0.999974,
		0.999871,
		0.999605,
		0.999062,
		0.998111,
		0.996606,
		0.994386,
		0.991289,
		0.987147,
		0.981794,
		0.975068,
		0.966816,
		0.956891,
		0.945162,
		0.931507,
		0.915824,
		0.898023,
		0.878037,
		0.855814,
		0.831326,
		0.804566,
		0.775551,
		0.744321,
		0.710944,
		0.675515,
		0.638156,
		0.599022,
		0.558299,
		0.516206,
		0.473,
		0.428973,
		0.38446,
		0.339837,
		0.295525,
		0.251992,
		0.209758,
		0.169394,
		0.131531,
		0.096858,
		0.066128,
		0.040162,
		0.019856,
		0.006177,
		0.000179,
		0.796619,
		0.814507,
		0.831692,
		0.848135,
		0.863799,
		0.878652,
		0.892661,
		0.905802,
		0.918052,
		0.929393,
		0.939815,
		0.949311,
		0.957881,
		0.965534,
		0.972286,
		0.978159,
		0.983187,
		0.987409,
		0.990877,
		0.993648,
		0.995792,
		0.997381,
		0.998499,
		0.999232,
		0.999666,
		0.999888,
		0.999976,
		0.999998,
		1,
		0.999998,
		0.999975,
		0.999876,
		0.999618,
		0.999093,
		0.998174,
		0.996717,
		0.994569,
		0.991569,
		0.987556,
		0.982367,
		0.975845,
		0.967838,
		0.958205,
		0.946813,
		0.933544,
		0.918295,
		0.900978,
		0.881522,
		0.859876,
		0.836008,
		0.809907,
		0.781584,
		0.751077,
		0.718443,
		0.68377,
		0.647171,
		0.608789,
		0.568797,
		0.527402,
		0.484843,
		0.441396,
		0.397375,
		0.353134,
		0.30907,
		0.265625,
		0.223289,
		0.182603,
		0.144161,
		0.108615,
		0.076675,
		0.049119,
		0.026791,
		0.010606,
		0.001559,
		0.835475,
		0.851607,
		0.866967,
		0.881522,
		0.895243,
		0.908106,
		0.92009,
		0.931179,
		0.941362,
		0.950636,
		0.959001,
		0.966466,
		0.973047,
		0.978769,
		0.983664,
		0.987773,
		0.991144,
		0.993838,
		0.995919,
		0.997462,
		0.998546,
		0.999256,
		0.999677,
		0.999891,
		0.999977,
		0.999998,
		1,
		0.999998,
		0.999976,
		0.99988,
		0.999631,
		0.999123,
		0.998234,
		0.996824,
		0.994744,
		0.991839,
		0.987949,
		0.982919,
		0.976593,
		0.968823,
		0.959469,
		0.948403,
		0.935507,
		0.920678,
		0.903828,
		0.884886,
		0.863799,
		0.840533,
		0.815072,
		0.787425,
		0.757621,
		0.725714,
		0.691782,
		0.655929,
		0.618289,
		0.579023,
		0.538323,
		0.496415,
		0.453557,
		0.410046,
		0.366214,
		0.322436,
		0.27913,
		0.236757,
		0.19583,
		0.156908,
		0.120607,
		0.087601,
		0.058623,
		0.034471,
		0.016011,
		0.004182,
		0.870045,
		0.88431,
		0.897749,
		0.910341,
		0.922066,
		0.932909,
		0.942861,
		0.951918,
		0.960083,
		0.967366,
		0.973783,
		0.979358,
		0.984125,
		0.988123,
		0.991403,
		0.99402,
		0.996042,
		0.99754,
		0.998592,
		0.99928,
		0.999687,
		0.999895,
		0.999978,
		0.999999,
		1,
		0.999998,
		0.999976,
		0.999884,
		0.999643,
		0.999152,
		0.998291,
		0.996926,
		0.994912,
		0.992097,
		0.988327,
		0.983449,
		0.977312,
		0.96977,
		0.960686,
		0.949934,
		0.937398,
		0.922975,
		0.906578,
		0.888134,
		0.86759,
		0.844907,
		0.820069,
		0.79308,
		0.763962,
		0.732765,
		0.699558,
		0.664439,
		0.62753,
		0.588982,
		0.548975,
		0.507719,
		0.465459,
		0.422472,
		0.379073,
		0.335615,
		0.292492,
		0.250141,
		0.209044,
		0.169733,
		0.132788,
		0.098846,
		0.068601,
		0.042808,
		0.022285,
		0.007921,
		0.000675,
		0.900183,
		0.91251,
		0.923982,
		0.934586,
		0.944313,
		0.95316,
		0.961131,
		0.968237,
		0.974494,
		0.979927,
		0.984569,
		0.988461,
		0.991651,
		0.994196,
		0.99616,
		0.997614,
		0.998635,
		0.999302,
		0.999697,
		0.999898,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999977,
		0.999888,
		0.999655,
		0.99918,
		0.998347,
		0.997025,
		0.995074,
		0.992346,
		0.988691,
		0.983959,
		0.978003,
		0.970681,
		0.961859,
		0.95141,
		0.939222,
		0.925192,
		0.909232,
		0.891271,
		0.871252,
		0.849137,
		0.824905,
		0.798554,
		0.770107,
		0.739603,
		0.707107,
		0.672708,
		0.636519,
		0.59868,
		0.559361,
		0.518758,
		0.477101,
		0.434652,
		0.391707,
		0.348599,
		0.3057,
		0.263423,
		0.222223,
		0.182603,
		0.145114,
		0.110356,
		0.078987,
		0.05172,
		0.02933,
		0.012657,
		0.00261,
		0.925841,
		0.936212,
		0.945719,
		0.954362,
		0.962145,
		0.969079,
		0.975181,
		0.980477,
		0.984999,
		0.988787,
		0.991891,
		0.994366,
		0.996274,
		0.997686,
		0.998677,
		0.999324,
		0.999707,
		0.999902,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999978,
		0.999891,
		0.999666,
		0.999206,
		0.9984,
		0.997119,
		0.995229,
		0.992584,
		0.98904,
		0.98445,
		0.978669,
		0.971559,
		0.962988,
		0.952833,
		0.94098,
		0.92733,
		0.911794,
		0.894301,
		0.874792,
		0.853227,
		0.829583,
		0.803856,
		0.776061,
		0.746234,
		0.714434,
		0.680741,
		0.645261,
		0.608124,
		0.569487,
		0.529537,
		0.488488,
		0.446587,
		0.404113,
		0.361381,
		0.318742,
		0.276585,
		0.235343,
		0.19549,
		0.157547,
		0.122084,
		0.08972,
		0.061133,
		0.037057,
		0.018285,
		0.005678,
		0.000165,
		0.947082,
		0.955526,
		0.963126,
		0.969893,
		0.975845,
		0.981008,
		0.985413,
		0.989102,
		0.992122,
		0.994529,
		0.996384,
		0.997755,
		0.998717,
		0.999345,
		0.999716,
		0.999905,
		0.99998,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999895,
		0.999677,
		0.999232,
		0.998451,
		0.99721,
		0.995378,
		0.992814,
		0.989376,
		0.984922,
		0.97931,
		0.972405,
		0.964076,
		0.954204,
		0.942676,
		0.929393,
		0.914268,
		0.897228,
		0.878214,
		0.857183,
		0.834112,
		0.80899,
		0.781832,
		0.752666,
		0.721547,
		0.688547,
		0.653763,
		0.617319,
		0.579359,
		0.540059,
		0.499621,
		0.458277,
		0.416289,
		0.373956,
		0.331609,
		0.289616,
		0.248386,
		0.208367,
		0.170054,
		0.133985,
		0.100747,
		0.070982,
		0.045385,
		0.024706,
		0.009762,
		0.001432,
		0.964076,
		0.970681,
		0.976488,
		0.981521,
		0.985814,
		0.989406,
		0.992346,
		0.994686,
		0.99649,
		0.997822,
		0.998756,
		0.999365,
		0.999725,
		0.999908,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999898,
		0.999687,
		0.999256,
		0.998499,
		0.997298,
		0.995521,
		0.993035,
		0.9897,
		0.985376,
		0.979927,
		0.973219,
		0.965125,
		0.955526,
		0.944313,
		0.931385,
		0.916658,
		0.900056,
		0.881522,
		0.861011,
		0.838495,
		0.813964,
		0.787425,
		0.758906,
		0.728452,
		0.696131,
		0.662033,
		0.626271,
		0.588982,
		0.55033,
		0.510503,
		0.469722,
		0.428234,
		0.38632,
		0.344293,
		0.302502,
		0.261333,
		0.221211,
		0.182603,
		0.14602,
		0.112019,
		0.081207,
		0.05424,
		0.031834,
		0.014759,
		0.003848,
		0.977109,
		0.982017,
		0.986201,
		0.9897,
		0.992561,
		0.994838,
		0.996591,
		0.997886,
		0.998793,
		0.999384,
		0.999733,
		0.999911,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.99998,
		0.999902,
		0.999697,
		0.99928,
		0.998546,
		0.997381,
		0.995659,
		0.993247,
		0.990011,
		0.985814,
		0.980522,
		0.974004,
		0.966137,
		0.956802,
		0.945892,
		0.933308,
		0.918966,
		0.90279,
		0.884721,
		0.864714,
		0.842738,
		0.818782,
		0.792848,
		0.764959,
		0.735156,
		0.7035,
		0.670076,
		0.634987,
		0.598361,
		0.560352,
		0.521138,
		0.480925,
		0.439948,
		0.39847,
		0.356789,
		0.315234,
		0.27417,
		0.234001,
		0.195168,
		0.158156,
		0.123492,
		0.091751,
		0.063557,
		0.039587,
		0.02057,
		0.007298,
		0.000621,
		0.986574,
		0.989983,
		0.992769,
		0.994985,
		0.99669,
		0.997948,
		0.998829,
		0.999403,
		0.999741,
		0.999913,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999905,
		0.999707,
		0.999302,
		0.998592,
		0.997462,
		0.995792,
		0.993452,
		0.99031,
		0.986235,
		0.981094,
		0.974761,
		0.967112,
		0.958032,
		0.947416,
		0.935165,
		0.921195,
		0.905432,
		0.887815,
		0.868297,
		0.846847,
		0.82345,
		0.798105,
		0.770831,
		0.741665,
		0.710662,
		0.677898,
		0.643472,
		0.607502,
		0.570132,
		0.531529,
		0.491888,
		0.45143,
		0.410404,
		0.369091,
		0.327802,
		0.286884,
		0.246719,
		0.207724,
		0.17036,
		0.135126,
		0.102567,
		0.073275,
		0.047892,
		0.02711,
		0.011678,
		0.002404,
		0.992969,
		0.995126,
		0.996784,
		0.998008,
		0.998863,
		0.999421,
		0.999749,
		0.999916,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999908,
		0.999716,
		0.999324,
		0.998635,
		0.99754,
		0.995919,
		0.993648,
		0.990599,
		0.986641,
		0.981647,
		0.97549,
		0.968053,
		0.95922,
		0.948887,
		0.936959,
		0.92335,
		0.907987,
		0.890808,
		0.871765,
		0.850827,
		0.827973,
		0.803203,
		0.776529,
		0.747985,
		0.717621,
		0.685507,
		0.651733,
		0.61641,
		0.579674,
		0.541681,
		0.502614,
		0.462682,
		0.42212,
		0.381194,
		0.3402,
		0.299464,
		0.259348,
		0.22025,
		0.182603,
		0.146884,
		0.113609,
		0.083339,
		0.056682,
		0.034297,
		0.016893,
		0.005236,
		0.000152,
		0.996876,
		0.998065,
		0.998897,
		0.999438,
		0.999757,
		0.999919,
		0.999983,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999911,
		0.999725,
		0.999345,
		0.998677,
		0.997614,
		0.996042,
		0.993838,
		0.990877,
		0.987032,
		0.982179,
		0.976194,
		0.96896,
		0.960366,
		0.950308,
		0.938692,
		0.925433,
		0.910457,
		0.893703,
		0.875123,
		0.854681,
		0.832357,
		0.808146,
		0.782059,
		0.754123,
		0.724384,
		0.692907,
		0.659775,
		0.625091,
		0.588982,
		0.551596,
		0.513104,
		0.473704,
		0.433618,
		0.393097,
		0.35242,
		0.311899,
		0.271875,
		0.232725,
		0.194862,
		0.158736,
		0.124837,
		0.093699,
		0.065898,
		0.042059,
		0.022856,
		0.009016,
		0.00132,
		0.998929,
		0.999454,
		0.999764,
		0.999921,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999913,
		0.999733,
		0.999365,
		0.998717,
		0.997686,
		0.99616,
		0.99402,
		0.991144,
		0.987409,
		0.982692,
		0.976872,
		0.969836,
		0.961473,
		0.951681,
		0.940367,
		0.927446,
		0.912847,
		0.896506,
		0.878374,
		0.858415,
		0.836606,
		0.812941,
		0.787425,
		0.760084,
		0.730958,
		0.700105,
		0.667604,
		0.63355,
		0.598062,
		0.561279,
		0.523363,
		0.484498,
		0.444897,
		0.404796,
		0.364459,
		0.324181,
		0.284287,
		0.245133,
		0.207112,
		0.170651,
		0.136215,
		0.10431,
		0.075484,
		0.05033,
		0.029488,
		0.013648,
		0.003552,
		0.999771,
		0.999923,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999983,
		0.999916,
		0.999741,
		0.999384,
		0.998756,
		0.997755,
		0.996274,
		0.994196,
		0.991403,
		0.987773,
		0.983187,
		0.977527,
		0.970681,
		0.962541,
		0.953007,
		0.941985,
		0.929393,
		0.915158,
		0.899218,
		0.881522,
		0.862033,
		0.840726,
		0.817592,
		0.792634,
		0.765874,
		0.737347,
		0.707107,
		0.675225,
		0.641793,
		0.606919,
		0.570735,
		0.533393,
		0.495067,
		0.455958,
		0.416289,
		0.376311,
		0.336303,
		0.296573,
		0.25746,
		0.219335,
		0.182603,
		0.147708,
		0.11513,
		0.085389,
		0.059048,
		0.036715,
		0.019046,
		0.006746,
		0.000573,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999919,
		0.999749,
		0.999403,
		0.998793,
		0.997822,
		0.996384,
		0.994366,
		0.991651,
		0.988123,
		0.983664,
		0.978159,
		0.971498,
		0.963574,
		0.954288,
		0.94355,
		0.931276,
		0.917395,
		0.901844,
		0.884571,
		0.865538,
		0.84472,
		0.822103,
		0.79769,
		0.771497,
		0.743557,
		0.713918,
		0.682646,
		0.649825,
		0.615558,
		0.579968,
		0.543198,
		0.505413,
		0.466802,
		0.427577,
		0.387975,
		0.34826,
		0.308725,
		0.269691,
		0.231511,
		0.19457,
		0.159289,
		0.126123,
		0.095568,
		0.068159,
		0.044472,
		0.025132,
		0.010808,
		0.002221,
		1,
		0.999999,
		0.999984,
		0.999921,
		0.999757,
		0.999421,
		0.998829,
		0.997886,
		0.99649,
		0.994529,
		0.991891,
		0.988461,
		0.984125,
		0.978769,
		0.972286,
		0.964571,
		0.955526,
		0.945063,
		0.933098,
		0.91956,
		0.904386,
		0.887524,
		0.868936,
		0.848593,
		0.826481,
		0.802599,
		0.77696,
		0.749594,
		0.720543,
		0.68987,
		0.657651,
		0.623984,
		0.588982,
		0.552782,
		0.515539,
		0.47743,
		0.438658,
		0.399446,
		0.360045,
		0.320733,
		0.281814,
		0.243624,
		0.206529,
		0.170928,
		0.137255,
		0.10598,
		0.077611,
		0.052699,
		0.031834,
		0.015654,
		0.004845,
		0.00014,
		0.999985,
		0.999923,
		0.999764,
		0.999438,
		0.998863,
		0.997948,
		0.996591,
		0.994686,
		0.992122,
		0.988787,
		0.984569,
		0.979358,
		0.973047,
		0.965534,
		0.956724,
		0.946526,
		0.93486,
		0.921655,
		0.906848,
		0.890386,
		0.87223,
		0.85235,
		0.830729,
		0.807365,
		0.782268,
		0.755463,
		0.726989,
		0.696903,
		0.665277,
		0.632201,
		0.597782,
		0.562148,
		0.525446,
		0.487844,
		0.449532,
		0.410723,
		0.371655,
		0.332591,
		0.29382,
		0.255662,
		0.218463,
		0.182603,
		0.148496,
		0.116587,
		0.08736,
		0.061339,
		0.039085,
		0.021206,
		0.008351,
		0.001221,
		0.999771,
		0.999454,
		0.998897,
		0.998008,
		0.99669,
		0.994838,
		0.992346,
		0.989102,
		0.984999,
		0.979927,
		0.973783,
		0.966466,
		0.957881,
		0.947941,
		0.936566,
		0.923684,
		0.909232,
		0.893159,
		0.875423,
		0.855993,
		0.834852,
		0.811993,
		0.787425,
		0.761169,
		0.73326,
		0.703751,
		0.672708,
		0.640214,
		0.606372,
		0.571301,
		0.535139,
		0.498046,
		0.460201,
		0.421806,
		0.383087,
		0.344293,
		0.3057,
		0.26761,
		0.230353,
		0.194292,
		0.159817,
		0.127354,
		0.097363,
		0.070342,
		0.046826,
		0.027391,
		0.012657,
		0.003289,
		0.998929,
		0.998065,
		0.996784,
		0.994985,
		0.992561,
		0.989406,
		0.985413,
		0.980477,
		0.974494,
		0.967366,
		0.959001,
		0.949311,
		0.938217,
		0.925648,
		0.911542,
		0.895847,
		0.878519,
		0.859528,
		0.838854,
		0.816488,
		0.792437,
		0.766717,
		0.739362,
		0.710418,
		0.679948,
		0.648029,
		0.614757,
		0.580244,
		0.544621,
		0.508038,
		0.470665,
		0.432694,
		0.394338,
		0.355834,
		0.317445,
		0.279458,
		0.242186,
		0.205974,
		0.171193,
		0.13825,
		0.107582,
		0.079662,
		0.055,
		0.034144,
		0.017685,
		0.006254,
		0.00053,
		0.996876,
		0.995126,
		0.992769,
		0.9897,
		0.985814,
		0.981008,
		0.975181,
		0.968237,
		0.960083,
		0.950636,
		0.939815,
		0.927551,
		0.913781,
		0.898452,
		0.881522,
		0.862958,
		0.842738,
		0.820854,
		0.797307,
		0.772112,
		0.745299,
		0.71691,
		0.687003,
		0.655651,
		0.622942,
		0.588982,
		0.553895,
		0.517822,
		0.480925,
		0.443386,
		0.405406,
		0.367211,
		0.32905,
		0.291195,
		0.253947,
		0.217631,
		0.182603,
		0.149248,
		0.117983,
		0.089257,
		0.063557,
		0.041405,
		0.023363,
		0.010032,
		0.002059,
		0.992969,
		0.989983,
		0.986201,
		0.981521,
		0.975845,
		0.969079,
		0.961131,
		0.951918,
		0.941362,
		0.929393,
		0.91595,
		0.900978,
		0.884434,
		0.866286,
		0.84651,
		0.825095,
		0.80204,
		0.777359,
		0.751077,
		0.723232,
		0.693878,
		0.663083,
		0.63093,
		0.597518,
		0.562964,
		0.527402,
		0.490984,
		0.453882,
		0.416289,
		0.378419,
		0.340507,
		0.302814,
		0.265625,
		0.229249,
		0.194026,
		0.160321,
		0.128533,
		0.099088,
		0.072452,
		0.049119,
		0.029626,
		0.014547,
		0.004495,
		0.00013,
		0.986574,
		0.982017,
		0.976488,
		0.969893,
		0.962145,
		0.95316,
		0.942861,
		0.931179,
		0.918052,
		0.903427,
		0.88726,
		0.869516,
		0.850172,
		0.829215,
		0.806641,
		0.782462,
		0.756699,
		0.729388,
		0.700578,
		0.670332,
		0.638727,
		0.605858,
		0.571833,
		0.53678,
		0.500843,
		0.464184,
		0.426988,
		0.389456,
		0.351814,
		0.314308,
		0.277209,
		0.240813,
		0.205443,
		0.171447,
		0.139203,
		0.109121,
		0.08164,
		0.057235,
		0.036415,
		0.019728,
		0.007758,
		0.001133,
		0.977109,
		0.970681,
		0.963126,
		0.954362,
		0.944313,
		0.932909,
		0.92009,
		0.905802,
		0.890001,
		0.872652,
		0.853729,
		0.833218,
		0.811114,
		0.787425,
		0.762171,
		0.735383,
		0.707107,
		0.6774,
		0.646337,
		0.614003,
		0.580504,
		0.545959,
		0.510503,
		0.474293,
		0.437501,
		0.40032,
		0.362964,
		0.325668,
		0.28869,
		0.252311,
		0.216838,
		0.182603,
		0.149968,
		0.119322,
		0.091084,
		0.065706,
		0.043674,
		0.02551,
		0.011771,
		0.003055,
		0.964076,
		0.955526,
		0.945719,
		0.934586,
		0.922066,
		0.908106,
		0.892661,
		0.875696,
		0.857183,
		0.837108,
		0.815463,
		0.792253,
		0.767497,
		0.741222,
		0.713469,
		0.684294,
		0.653763,
		0.62196,
		0.588982,
		0.554942,
		0.519969,
		0.48421,
		0.447829,
		0.41101,
		0.373956,
		0.336891,
		0.300059,
		0.26373,
		0.228195,
		0.193772,
		0.160804,
		0.129662,
		0.100747,
		0.07449,
		0.051353,
		0.031834,
		0.016464,
		0.005814,
		0.000492,
		0.947082,
		0.936212,
		0.923982,
		0.910341,
		0.895243,
		0.878652,
		0.860539,
		0.840888,
		0.819691,
		0.796951,
		0.772681,
		0.746909,
		0.71967,
		0.691017,
		0.661012,
		0.629732,
		0.59727,
		0.563732,
		0.529241,
		0.493936,
		0.457972,
		0.421525,
		0.384786,
		0.34797,
		0.31131,
		0.275061,
		0.239503,
		0.204936,
		0.171689,
		0.140116,
		0.110599,
		0.083548,
		0.059405,
		0.038645,
		0.021774,
		0.009337,
		0.001913,
		0.925841,
		0.91251,
		0.897749,
		0.881522,
		0.863799,
		0.844563,
		0.823803,
		0.801521,
		0.777728,
		0.752448,
		0.725714,
		0.697574,
		0.668086,
		0.637323,
		0.605372,
		0.572334,
		0.538323,
		0.503473,
		0.467931,
		0.431863,
		0.395453,
		0.358903,
		0.322436,
		0.286296,
		0.250747,
		0.216079,
		0.182603,
		0.150658,
		0.120607,
		0.092843,
		0.067787,
		0.045892,
		0.027641,
		0.013553,
		0.004182,
		0.00012,
		0.900183,
		0.88431,
		0.866967,
		0.848135,
		0.827802,
		0.805968,
		0.782641,
		0.757843,
		0.731605,
		0.703969,
		0.67499,
		0.644738,
		0.613292,
		0.580749,
		0.547218,
		0.512824,
		0.477707,
		0.442026,
		0.405954,
		0.369686,
		0.333432,
		0.297425,
		0.261919,
		0.227188,
		0.193529,
		0.161266,
		0.130747,
		0.102344,
		0.076461,
		0.053528,
		0.034009,
		0.018399,
		0.007225,
		0.001053,
		0.870045,
		0.851607,
		0.831692,
		0.810295,
		0.787425,
		0.7631,
		0.737347,
		0.710206,
		0.681729,
		0.651979,
		0.621034,
		0.588982,
		0.555928,
		0.52199,
		0.487302,
		0.452013,
		0.416289,
		0.380315,
		0.344293,
		0.308444,
		0.273008,
		0.238249,
		0.204451,
		0.171921,
		0.140991,
		0.112019,
		0.085389,
		0.061512,
		0.04083,
		0.023816,
		0.010974,
		0.002844,
		0.835475,
		0.814507,
		0.792083,
		0.76822,
		0.742944,
		0.71629,
		0.688306,
		0.659052,
		0.628601,
		0.597036,
		0.564456,
		0.530974,
		0.496716,
		0.461824,
		0.426457,
		0.39079,
		0.355015,
		0.319344,
		0.284006,
		0.249252,
		0.215353,
		0.182603,
		0.151319,
		0.121842,
		0.094539,
		0.069803,
		0.048056,
		0.02975,
		0.015366,
		0.005419,
		0.000458,
		0.796619,
		0.77321,
		0.7484,
		0.722224,
		0.694725,
		0.66596,
		0.635997,
		0.604914,
		0.572806,
		0.539778,
		0.505952,
		0.471461,
		0.436458,
		0.401108,
		0.365596,
		0.330122,
		0.294906,
		0.260186,
		0.226223,
		0.193297,
		0.16171,
		0.131787,
		0.103881,
		0.078366,
		0.055645,
		0.03615,
		0.020342,
		0.008711,
		0.001783,
		0.753719,
		0.728012,
		0.700991,
		0.672708,
		0.643226,
		0.612621,
		0.58098,
		0.548406,
		0.515011,
		0.480925,
		0.446292,
		0.411269,
		0.376031,
		0.340772,
		0.3057,
		0.271043,
		0.237049,
		0.203986,
		0.172143,
		0.141832,
		0.113386,
		0.087167,
		0.063557,
		0.042971,
		0.025848,
		0.012657,
		0.003901,
		0.000112,
		0.707107,
		0.679298,
		0.650291,
		0.620158,
		0.588982,
		0.556859,
		0.523896,
		0.490217,
		0.455958,
		0.42127,
		0.38632,
		0.351291,
		0.316383,
		0.281814,
		0.247821,
		0.214658,
		0.182603,
		0.151954,
		0.12303,
		0.096175,
		0.071756,
		0.050169,
		0.031834,
		0.0172,
		0.006746,
		0.000982,
		0.657196,
		0.62753,
		0.596815,
		0.56514,
		0.532609,
		0.499339,
		0.465459,
		0.431113,
		0.39646,
		0.361676,
		0.32695,
		0.292492,
		0.258527,
		0.2253,
		0.193074,
		0.162135,
		0.132788,
		0.105362,
		0.080209,
		0.057705,
		0.038254,
		0.022285,
		0.010256,
		0.002655,
		0.604481,
		0.573252,
		0.541152,
		0.508291,
		0.474793,
		0.440796,
		0.40645,
		0.371923,
		0.337397,
		0.303071,
		0.269161,
		0.2359,
		0.203541,
		0.172357,
		0.142639,
		0.114702,
		0.088884,
		0.065543,
		0.045066,
		0.027864,
		0.014374,
		0.005063,
		0.000428,
		0.549528,
		0.517077,
		0.483964,
		0.45032,
		0.416289,
		0.382031,
		0.34772,
		0.313545,
		0.279714,
		0.246449,
		0.213992,
		0.182603,
		0.152563,
		0.124172,
		0.097753,
		0.073649,
		0.05223,
		0.033889,
		0.019046,
		0.008146,
		0.001665,
		0.492972,
		0.459685,
		0.425977,
		0.391998,
		0.357915,
		0.32391,
		0.290179,
		0.256937,
		0.224414,
		0.19286,
		0.162543,
		0.13375,
		0.106789,
		0.081992,
		0.05971,
		0.04032,
		0.024223,
		0.011847,
		0.003647,
		0.000105,
		0.435513,
		0.401822,
		0.36798,
		0.33416,
		0.300551,
		0.267356,
		0.234798,
		0.203114,
		0.172561,
		0.143415,
		0.11597,
		0.090543,
		0.067471,
		0.047116,
		0.02986,
		0.016114,
		0.006313,
		0.000918,
		0.377913,
		0.344293,
		0.310823,
		0.2777,
		0.245133,
		0.213353,
		0.182603,
		0.153149,
		0.125272,
		0.099276,
		0.075484,
		0.05424,
		0.035914,
		0.020897,
		0.009606,
		0.002483,
		0.320992,
		0.28796,
		0.255412,
		0.223565,
		0.192655,
		0.162935,
		0.134676,
		0.108167,
		0.083718,
		0.06166,
		0.042346,
		0.026151,
		0.013475,
		0.004741,
		0.0004,
		0.265625,
		0.233741,
		0.202704,
		0.172758,
		0.144161,
		0.117192,
		0.092147,
		0.069344,
		0.049119,
		0.031834,
		0.01787,
		0.007634,
		0.001559,
		0.212739,
		0.182603,
		0.153712,
		0.126332,
		0.100747,
		0.077262,
		0.0562,
		0.037906,
		0.022747,
		0.011113,
		0.003417,
		9.8e-005,
		0.163312,
		0.135567,
		0.109496,
		0.085389,
		0.063557,
		0.044332,
		0.028064,
		0.015128,
		0.00592,
		0.00086,
		0.118371,
		0.093699,
		0.071162,
		0.051077,
		0.033782,
		0.019635,
		0.009016,
		0.002328,
		0.078987,
		0.058111,
		0.039864,
		0.024591,
		0.012657,
		0.004448,
		0.000375,
		0.046277,
		0.029959,
		0.0168,
		0.00717,
		0.001462,
		0.021402,
		0.010445,
		0.003208,
		9.2e-005,
		0.005562,
		0.000807,
		0.000361,
		0.00144,
		0.008674,
		0.003233,
		0.012464,
		0.027023,
		0.000372,
		0.005737,
		0.016928,
		0.033279,
		0.054161,
		0.001486,
		0.008946,
		0.022059,
		0.040162,
		0.062632,
		0.088884,
		0.003336,
		0.012854,
		0.027851,
		0.047667,
		0.071682,
		0.099316,
		0.130029,
		0.000384,
		0.00592,
		0.017457,
		0.034297,
		0.055783,
		0.081301,
		0.110275,
		0.142168,
		0.176478,
		0.001534,
		0.00923,
		0.022747,
		0.041389,
		0.064503,
		0.09148,
		0.121749,
		0.154777,
		0.190067,
		0.227159,
		0.003444,
		0.013263,
		0.028718,
		0.049119,
		0.073818,
		0.102209,
		0.133727,
		0.167843,
		0.204067,
		0.241943,
		0.281048,
		0.000397,
		0.006111,
		0.018011,
		0.035363,
		0.057479,
		0.083718,
		0.113477,
		0.146196,
		0.181354,
		0.218463,
		0.257074,
		0.29677,
		0.337168,
		0.001584,
		0.009529,
		0.023468,
		0.042672,
		0.06646,
		0.094192,
		0.125272,
		0.159146,
		0.195295,
		0.233241,
		0.272538,
		0.312776,
		0.353578,
		0.394596,
		0.003558,
		0.013691,
		0.029626,
		0.050639,
		0.076051,
		0.105229,
		0.137583,
		0.172561,
		0.209654,
		0.248386,
		0.288319,
		0.32905,
		0.370205,
		0.411446,
		0.452462,
		0.00041,
		0.006313,
		0.018592,
		0.036479,
		0.059254,
		0.086243,
		0.116819,
		0.150396,
		0.18643,
		0.224414,
		0.263882,
		0.304401,
		0.345573,
		0.387032,
		0.428447,
		0.469512,
		0.509955,
		0.001638,
		0.009842,
		0.024223,
		0.044016,
		0.068506,
		0.097025,
		0.128949,
		0.163699,
		0.200737,
		0.239562,
		0.279714,
		0.320766,
		0.362328,
		0.40404,
		0.445578,
		0.486644,
		0.526971,
		0.566321,
		0.003677,
		0.014141,
		0.030578,
		0.05223,
		0.078386,
		0.108384,
		0.141606,
		0.177478,
		0.215467,
		0.255081,
		0.295864,
		0.337397,
		0.379296,
		0.42121,
		0.46282,
		0.503836,
		0.544,
		0.583079,
		0.620871,
		0.000424,
		0.006524,
		0.019201,
		0.037648,
		0.061111,
		0.088884,
		0.120309,
		0.154777,
		0.191718,
		0.230606,
		0.270954,
		0.312315,
		0.354276,
		0.39646,
		0.438523,
		0.480153,
		0.52107,
		0.561021,
		0.599783,
		0.63716,
		0.672983,
		0.001693,
		0.010171,
		0.025016,
		0.045425,
		0.070648,
		0.099986,
		0.132788,
		0.168448,
		0.206404,
		0.246136,
		0.287165,
		0.32905,
		0.371384,
		0.413799,
		0.455958,
		0.497557,
		0.538323,
		0.578013,
		0.61641,
		0.653328,
		0.688606,
		0.722107,
		0.003803,
		0.014613,
		0.031577,
		0.053897,
		0.08083,
		0.111681,
		0.145806,
		0.182603,
		0.22152,
		0.262041,
		0.303696,
		0.346048,
		0.388702,
		0.431293,
		0.473495,
		0.515011,
		0.555576,
		0.594955,
		0.63294,
		0.669354,
		0.704043,
		0.736881,
		0.767766,
		0.000439,
		0.006746,
		0.019841,
		0.038875,
		0.063057,
		0.091646,
		0.123957,
		0.159349,
		0.197229,
		0.237049,
		0.278304,
		0.320527,
		0.363292,
		0.406208,
		0.448922,
		0.491113,
		0.532493,
		0.572806,
		0.611825,
		0.649351,
		0.685216,
		0.719275,
		0.751412,
		0.781532,
		0.809569,
		0.001752,
		0.010517,
		0.025848,
		0.046901,
		0.072891,
		0.103084,
		0.136799,
		0.173402,
		0.212308,
		0.252975,
		0.294906,
		0.33764,
		0.38076,
		0.423884,
		0.466665,
		0.50879,
		0.549982,
		0.589991,
		0.628601,
		0.665622,
		0.700893,
		0.73428,
		0.765677,
		0.794999,
		0.822187,
		0.847207,
		0.003934,
		0.015108,
		0.032625,
		0.055645,
		0.083389,
		0.11513,
		0.150192,
		0.18795,
		0.227824,
		0.26928,
		0.311827,
		0.355015,
		0.398433,
		0.441707,
		0.484498,
		0.526505,
		0.567455,
		0.607109,
		0.645261,
		0.681729,
		0.716363,
		0.749038,
		0.779657,
		0.808146,
		0.834457,
		0.858565,
		0.880467,
		0.000454,
		0.006979,
		0.020513,
		0.040162,
		0.065096,
		0.094539,
		0.127771,
		0.164123,
		0.202976,
		0.243759,
		0.285944,
		0.32905,
		0.372632,
		0.416289,
		0.459655,
		0.502402,
		0.544233,
		0.584888,
		0.624138,
		0.661782,
		0.697651,
		0.731605,
		0.763528,
		0.793332,
		0.820956,
		0.846361,
		0.869533,
		0.89048,
		0.909232,
		0.001814,
		0.01088,
		0.026722,
		0.048451,
		0.075242,
		0.106327,
		0.140991,
		0.178574,
		0.218463,
		0.260094,
		0.302949,
		0.346552,
		0.390469,
		0.434306,
		0.477707,
		0.520351,
		0.561953,
		0.60226,
		0.641052,
		0.678141,
		0.713366,
		0.746597,
		0.777728,
		0.806683,
		0.833411,
		0.857883,
		0.880097,
		0.900072,
		0.91785,
		0.933492,
		0.004073,
		0.01563,
		0.033726,
		0.057479,
		0.086071,
		0.118739,
		0.154777,
		0.193529,
		0.234393,
		0.276811,
		0.320274,
		0.364313,
		0.408505,
		0.452462,
		0.495839,
		0.538323,
		0.57964,
		0.619546,
		0.657831,
		0.694316,
		0.728851,
		0.761317,
		0.791619,
		0.819691,
		0.845492,
		0.869007,
		0.890244,
		0.909232,
		0.926026,
		0.940699,
		0.953345,
		0.000471,
		0.007225,
		0.02122,
		0.041515,
		0.067236,
		0.09757,
		0.131761,
		0.169111,
		0.208972,
		0.250747,
		0.29389,
		0.337898,
		0.382312,
		0.426716,
		0.470734,
		0.514027,
		0.556295,
		0.59727,
		0.636722,
		0.674449,
		0.710282,
		0.744084,
		0.775744,
		0.805179,
		0.832336,
		0.857183,
		0.879718,
		0.899959,
		0.91795,
		0.933754,
		0.947458,
		0.959167,
		0.969004,
		0.001879,
		0.011263,
		0.027641,
		0.050078,
		0.077708,
		0.109724,
		0.145377,
		0.183976,
		0.224882,
		0.267506,
		0.31131,
		0.355799,
		0.400525,
		0.44508,
		0.489096,
		0.532247,
		0.57424,
		0.61482,
		0.653763,
		0.690882,
		0.726017,
		0.759042,
		0.789856,
		0.81839,
		0.8446,
		0.868468,
		0.890001,
		0.909232,
		0.926215,
		0.941027,
		0.953766,
		0.964547,
		0.973507,
		0.980795,
		0.004219,
		0.016179,
		0.034883,
		0.059405,
		0.088884,
		0.122519,
		0.159571,
		0.199356,
		0.241242,
		0.28465,
		0.32905,
		0.373956,
		0.418929,
		0.463572,
		0.507526,
		0.550474,
		0.592135,
		0.632263,
		0.670646,
		0.707107,
		0.741497,
		0.773702,
		0.803632,
		0.83123,
		0.856465,
		0.879329,
		0.899843,
		0.918052,
		0.934022,
		0.947842,
		0.959621,
		0.969489,
		0.977591,
		0.984089,
		0.989154,
		0.000488,
		0.007484,
		0.021964,
		0.042936,
		0.069482,
		0.100747,
		0.135939,
		0.174326,
		0.21523,
		0.258031,
		0.302157,
		0.347086,
		0.392345,
		0.437501,
		0.482168,
		0.525998,
		0.568684,
		0.609954,
		0.649575,
		0.687346,
		0.723099,
		0.756699,
		0.788042,
		0.817052,
		0.843682,
		0.867913,
		0.889753,
		0.909232,
		0.926409,
		0.941362,
		0.954194,
		0.965026,
		0.973997,
		0.981267,
		0.987004,
		0.99139,
		0.994617,
		0.001948,
		0.011667,
		0.028608,
		0.051789,
		0.080297,
		0.113285,
		0.149968,
		0.189623,
		0.23158,
		0.275228,
		0.320005,
		0.365398,
		0.410941,
		0.456215,
		0.500843,
		0.544486,
		0.586849,
		0.627672,
		0.66673,
		0.703836,
		0.738833,
		0.771599,
		0.80204,
		0.830094,
		0.855726,
		0.878929,
		0.899725,
		0.918157,
		0.934295,
		0.948233,
		0.960083,
		0.969981,
		0.978079,
		0.984543,
		0.989554,
		0.993303,
		0.995985,
		0.997796,
		0.004374,
		0.016757,
		0.036101,
		0.061429,
		0.091835,
		0.12648,
		0.164588,
		0.205443,
		0.248386,
		0.292812,
		0.338171,
		0.383959,
		0.429721,
		0.475047,
		0.51957,
		0.562964,
		0.604944,
		0.645261,
		0.683702,
		0.720092,
		0.754286,
		0.786174,
		0.815675,
		0.842738,
		0.867344,
		0.889497,
		0.909232,
		0.926607,
		0.941704,
		0.95463,
		0.965511,
		0.974494,
		0.981742,
		0.987433,
		0.991758,
		0.994912,
		0.997097,
		0.998511,
		0.999343,
		0.000506,
		0.007758,
		0.022747,
		0.044432,
		0.071843,
		0.104082,
		0.140317,
		0.179781,
		0.221767,
		0.265625,
		0.310761,
		0.356632,
		0.402746,
		0.448658,
		0.493969,
		0.538323,
		0.581406,
		0.622942,
		0.662695,
		0.700466,
		0.736089,
		0.769433,
		0.800401,
		0.828924,
		0.854966,
		0.878519,
		0.899603,
		0.918264,
		0.934575,
		0.948631,
		0.960553,
		0.97048,
		0.97857,
		0.984999,
		0.989954,
		0.993633,
		0.99624,
		0.997978,
		0.999046,
		0.999629,
		0.999896,
		0.002021,
		0.012092,
		0.029626,
		0.053588,
		0.083016,
		0.11702,
		0.154777,
		0.195527,
		0.238574,
		0.283276,
		0.32905,
		0.375363,
		0.421733,
		0.467727,
		0.512956,
		0.557076,
		0.599783,
		0.640815,
		0.679948,
		0.716994,
		0.751801,
		0.78425,
		0.814257,
		0.841768,
		0.866759,
		0.889236,
		0.909232,
		0.926809,
		0.942053,
		0.955074,
		0.966004,
		0.974996,
		0.98222,
		0.987863,
		0.992122,
		0.995203,
		0.997312,
		0.998655,
		0.999429,
		0.999813,
		0.999962,
		0.999998,
		0.004537,
		0.017367,
		0.037384,
		0.063557,
		0.094935,
		0.130634,
		0.169842,
		0.211807,
		0.255841,
		0.301315,
		0.347655,
		0.394338,
		0.440894,
		0.4869,
		0.531979,
		0.575799,
		0.618068,
		0.658536,
		0.696992,
		0.73326,
		0.767202,
		0.798713,
		0.827721,
		0.854185,
		0.878098,
		0.899478,
		0.918374,
		0.93486,
		0.949038,
		0.961031,
		0.970985,
		0.979066,
		0.985457,
		0.990352,
		0.993959,
		0.99649,
		0.998153,
		0.999156,
		0.999688,
		0.99992,
		0.99999,
		1,
		1,
		0.000525,
		0.008046,
		0.023574,
		0.046007,
		0.074325,
		0.107582,
		0.144906,
		0.185491,
		0.228598,
		0.273547,
		0.319719,
		0.366551,
		0.413531,
		0.460201,
		0.506149,
		0.55101,
		0.594464,
		0.636233,
		0.676077,
		0.7138,
		0.749239,
		0.782268,
		0.812798,
		0.840769,
		0.866157,
		0.888967,
		0.909232,
		0.927017,
		0.94241,
		0.955526,
		0.966504,
		0.975504,
		0.982702,
		0.988294,
		0.992484,
		0.995488,
		0.99752,
		0.998793,
		0.999508,
		0.99985,
		0.999973,
		0.999999,
		1,
		0.999995,
		0.999945,
		0.002098,
		0.012541,
		0.0307,
		0.055483,
		0.085875,
		0.120941,
		0.159817,
		0.201706,
		0.245878,
		0.291667,
		0.338462,
		0.385711,
		0.432915,
		0.479627,
		0.525446,
		0.570021,
		0.613044,
		0.654248,
		0.69341,
		0.730344,
		0.764902,
		0.796974,
		0.826481,
		0.853382,
		0.877665,
		0.89935,
		0.918486,
		0.935153,
		0.949453,
		0.961517,
		0.971498,
		0.979567,
		0.985916,
		0.990749,
		0.994282,
		0.996733,
		0.998321,
		0.999259,
		0.999741,
		0.99994,
		0.999995,
		1,
		0.999999,
		0.999974,
		0.999859,
		0.999551,
		0.004709,
		0.018011,
		0.038736,
		0.065798,
		0.098193,
		0.134994,
		0.175346,
		0.218463,
		0.263625,
		0.310176,
		0.357518,
		0.405109,
		0.452462,
		0.499142,
		0.544762,
		0.588982,
		0.631507,
		0.672085,
		0.710506,
		0.746597,
		0.780225,
		0.811294,
		0.839741,
		0.865538,
		0.88869,
		0.909232,
		0.927229,
		0.942774,
		0.955987,
		0.967013,
		0.976017,
		0.983187,
		0.988724,
		0.992844,
		0.995768,
		0.997721,
		0.998923,
		0.999581,
		0.999882,
		0.999982,
		1,
		1,
		0.99999,
		0.99992,
		0.999698,
		0.999205,
		0.99831,
		0.000546,
		0.008351,
		0.024446,
		0.047667,
		0.076937,
		0.111261,
		0.149721,
		0.191473,
		0.23574,
		0.281814,
		0.32905,
		0.37686,
		0.424716,
		0.472143,
		0.518718,
		0.564067,
		0.607863,
		0.649825,
		0.689715,
		0.727336,
		0.762531,
		0.795181,
		0.825205,
		0.852555,
		0.877219,
		0.899218,
		0.918602,
		0.935451,
		0.949876,
		0.962012,
		0.972017,
		0.980072,
		0.986377,
		0.991144,
		0.994599,
		0.99697,
		0.998482,
		0.999355,
		0.999788,
		0.999957,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.999807,
		0.999429,
		0.998695,
		0.997465,
		0.995601,
		0.002179,
		0.013016,
		0.031834,
		0.057479,
		0.088884,
		0.125061,
		0.165103,
		0.208175,
		0.253513,
		0.300419,
		0.34826,
		0.39646,
		0.444504,
		0.491928,
		0.538323,
		0.583329,
		0.626633,
		0.667967,
		0.707107,
		0.743871,
		0.778118,
		0.809744,
		0.838682,
		0.864902,
		0.888407,
		0.909232,
		0.927446,
		0.943146,
		0.956457,
		0.967529,
		0.976536,
		0.983674,
		0.989154,
		0.993199,
		0.996042,
		0.997915,
		0.999046,
		0.999647,
		0.999909,
		0.999989,
		1,
		1,
		0.999982,
		0.999886,
		0.999606,
		0.999017,
		0.997978,
		0.996348,
		0.993987,
		0.990768,
		0.004891,
		0.018691,
		0.040162,
		0.068159,
		0.101621,
		0.139573,
		0.181117,
		0.225429,
		0.271757,
		0.319414,
		0.367779,
		0.416289,
		0.464441,
		0.511786,
		0.557927,
		0.602518,
		0.645261,
		0.685901,
		0.724231,
		0.760084,
		0.793332,
		0.823889,
		0.851704,
		0.876762,
		0.899083,
		0.91872,
		0.935757,
		0.950308,
		0.962515,
		0.972543,
		0.980581,
		0.986838,
		0.991538,
		0.994912,
		0.997199,
		0.998635,
		0.999443,
		0.99983,
		0.99997,
		0.999999,
		1,
		0.999994,
		0.999938,
		0.999741,
		0.999281,
		0.99842,
		0.997011,
		0.994912,
		0.991988,
		0.988116,
		0.983187,
		0.000567,
		0.008674,
		0.025367,
		0.049418,
		0.079688,
		0.11513,
		0.154777,
		0.197741,
		0.243212,
		0.290446,
		0.338772,
		0.387578,
		0.436317,
		0.484498,
		0.531686,
		0.577498,
		0.621603,
		0.663715,
		0.703599,
		0.741059,
		0.775945,
		0.808146,
		0.837591,
		0.864247,
		0.888115,
		0.909232,
		0.927669,
		0.943527,
		0.956935,
		0.968053,
		0.977061,
		0.984164,
		0.989583,
		0.993551,
		0.99631,
		0.998102,
		0.999161,
		0.999707,
		0.999932,
		0.999994,
		1,
		0.999999,
		0.999971,
		0.99984,
		0.999493,
		0.998793,
		0.997593,
		0.995745,
		0.993109,
		0.989554,
		0.984964,
		0.97924,
		0.972299,
		0.002265,
		0.013517,
		0.033031,
		0.059585,
		0.092052,
		0.129392,
		0.170651,
		0.214952,
		0.261496,
		0.309553,
		0.358463,
		0.407628,
		0.456513,
		0.504643,
		0.551596,
		0.597004,
		0.640549,
		0.681965,
		0.721027,
		0.757559,
		0.791425,
		0.822533,
		0.850827,
		0.87629,
		0.898944,
		0.918841,
		0.93607,
		0.950749,
		0.963026,
		0.973075,
		0.981094,
		0.987301,
		0.991928,
		0.995219,
		0.997422,
		0.99878,
		0.999525,
		0.999866,
		0.99998,
		1,
		1,
		0.999989,
		0.999909,
		0.999658,
		0.999103,
		0.998097,
		0.99649,
		0.994133,
		0.99089,
		0.986637,
		0.981267,
		0.974689,
		0.966832,
		0.957642,
		0.005084,
		0.019411,
		0.041669,
		0.070648,
		0.105229,
		0.144385,
		0.187172,
		0.232725,
		0.280256,
		0.32905,
		0.378457,
		0.427896,
		0.476846,
		0.524843,
		0.571482,
		0.61641,
		0.659326,
		0.699976,
		0.738155,
		0.773702,
		0.806498,
		0.836467,
		0.863572,
		0.887815,
		0.909232,
		0.927897,
		0.943915,
		0.957423,
		0.968585,
		0.977591,
		0.984657,
		0.990011,
		0.993899,
		0.996571,
		0.99828,
		0.999268,
		0.99976,
		0.999951,
		0.999997,
		1,
		0.999997,
		0.999954,
		0.999782,
		0.999355,
		0.998528,
		0.997147,
		0.99506,
		0.992122,
		0.988203,
		0.983187,
		0.976975,
		0.969489,
		0.960668,
		0.950468,
		0.938862,
		0.00059,
		0.009016,
		0.026342,
		0.051267,
		0.082589,
		0.119202,
		0.160089,
		0.204316,
		0.251033,
		0.299464,
		0.348906,
		0.398725,
		0.448352,
		0.497281,
		0.545064,
		0.591311,
		0.635684,
		0.677898,
		0.717717,
		0.754952,
		0.789457,
		0.821134,
		0.849923,
		0.875806,
		0.898801,
		0.918966,
		0.93639,
		0.951199,
		0.963547,
		0.973615,
		0.981612,
		0.987764,
		0.992316,
		0.995521,
		0.997637,
		0.998917,
		0.999599,
		0.999897,
		0.999988,
		1,
		1,
		0.99998,
		0.99987,
		0.999554,
		0.998889,
		0.997721,
		0.995893,
		0.993253,
		0.989663,
		0.984999,
		0.979155,
		0.972044,
		0.963598,
		0.953766,
		0.942516,
		0.929834,
		0.915718,
		0.002356,
		0.014049,
		0.034297,
		0.061808,
		0.095391,
		0.133949,
		0.176478,
		0.222058,
		0.269849,
		0.319089,
		0.369091,
		0.419232,
		0.468961,
		0.517785,
		0.565272,
		0.611048,
		0.654791,
		0.696234,
		0.735156,
		0.771385,
		0.804797,
		0.835308,
		0.862877,
		0.887506,
		0.909232,
		0.928131,
		0.944313,
		0.95792,
		0.969125,
		0.978128,
		0.985151,
		0.990437,
		0.994242,
		0.996826,
		0.998451,
		0.999367,
		0.999806,
		0.999966,
		0.999999,
		1,
		0.999993,
		0.99993,
		0.999707,
		0.999187,
		0.998216,
		0.996633,
		0.994282,
		0.991015,
		0.986701,
		0.981226,
		0.974494,
		0.966428,
		0.956972,
		0.946088,
		0.933754,
		0.919966,
		0.904734,
		0.888082,
		0.005289,
		0.020173,
		0.043262,
		0.073275,
		0.109032,
		0.149448,
		0.193529,
		0.24037,
		0.289144,
		0.339103,
		0.389573,
		0.439948,
		0.48969,
		0.538323,
		0.585432,
		0.630658,
		0.673697,
		0.714298,
		0.752258,
		0.787425,
		0.819691,
		0.848992,
		0.875306,
		0.898654,
		0.919093,
		0.936718,
		0.951659,
		0.964076,
		0.974162,
		0.982133,
		0.988228,
		0.9927,
		0.995817,
		0.997845,
		0.999046,
		0.999666,
		0.999923,
		0.999993,
		1,
		0.999999,
		0.999967,
		0.999819,
		0.999426,
		0.998635,
		0.997284,
		0.995211,
		0.992261,
		0.988294,
		0.983187,
		0.976835,
		0.969156,
		0.960083,
		0.949573,
		0.937598,
		0.924149,
		0.909232,
		0.892867,
		0.875088,
		0.855939,
		0.000615,
		0.009378,
		0.027373,
		0.053221,
		0.08565,
		0.123492,
		0.165674,
		0.211216,
		0.259225,
		0.308888,
		0.359472,
		0.410318,
		0.460836,
		0.510503,
		0.55886,
		0.605508,
		0.650105,
		0.692366,
		0.732056,
		0.768993,
		0.803041,
		0.834112,
		0.862161,
		0.887188,
		0.909232,
		0.928371,
		0.944719,
		0.958426,
		0.969673,
		0.978669,
		0.985648,
		0.990862,
		0.99458,
		0.997073,
		0.998613,
		0.999459,
		0.999847,
		0.999977,
		1,
		1,
		0.999987,
		0.999897,
		0.999612,
		0.998984,
		0.99785,
		0.996042,
		0.9934,
		0.989775,
		0.985035,
		0.979066,
		0.971777,
		0.963095,
		0.952968,
		0.941362,
		0.928264,
		0.913673,
		0.897607,
		0.880097,
		0.861185,
		0.840925,
		0.819379,
		0.002453,
		0.014613,
		0.035637,
		0.064157,
		0.098913,
		0.138748,
		0.182603,
		0.229511,
		0.278592,
		0.32905,
		0.380165,
		0.431293,
		0.481862,
		0.531365,
		0.579359,
		0.625463,
		0.669354,
		0.710763,
		0.749475,
		0.785326,
		0.818201,
		0.848031,
		0.874792,
		0.898503,
		0.919224,
		0.937054,
		0.952128,
		0.964615,
		0.974716,
		0.982658,
		0.988691,
		0.993081,
		0.996106,
		0.998044,
		0.999166,
		0.999726,
		0.999944,
		0.999996,
		1,
		0.999996,
		0.999947,
		0.999752,
		0.999268,
		0.998334,
		0.996778,
		0.994434,
		0.991144,
		0.986769,
		0.981184,
		0.974289,
		0.966004,
		0.956268,
		0.945041,
		0.932304,
		0.918052,
		0.902298,
		0.88507,
		0.866407,
		0.846361,
		0.824995,
		0.802379,
		0.778592,
		0.005506,
		0.02098,
		0.044948,
		0.076051,
		0.113042,
		0.154777,
		0.200209,
		0.248386,
		0.298442,
		0.349596,
		0.401146,
		0.452462,
		0.502989,
		0.552238,
		0.599783,
		0.645261,
		0.688367,
		0.728851,
		0.76652,
		0.801227,
		0.832877,
		0.861423,
		0.886861,
		0.909232,
		0.928617,
		0.945135,
		0.958943,
		0.97023,
		0.979216,
		0.986146,
		0.991285,
		0.994912,
		0.997312,
		0.998766,
		0.999542,
		0.999882,
		0.999986,
		1,
		1,
		0.999977,
		0.999852,
		0.999493,
		0.998741,
		0.997422,
		0.995364,
		0.992403,
		0.988388,
		0.983187,
		0.976689,
		0.968806,
		0.959469,
		0.948631,
		0.936266,
		0.922365,
		0.906935,
		0.890001,
		0.8716,
		0.851782,
		0.830607,
		0.808146,
		0.784477,
		0.759685,
		0.733863,
		0.000641,
		0.009762,
		0.028466,
		0.055289,
		0.088884,
		0.128015,
		0.171552,
		0.218463,
		0.267809,
		0.318742,
		0.370494,
		0.42238,
		0.473788,
		0.52418,
		0.573084,
		0.620093,
		0.664862,
		0.707107,
		0.746597,
		0.783156,
		0.816662,
		0.84704,
		0.874262,
		0.898348,
		0.919359,
		0.937398,
		0.952607,
		0.965164,
		0.975277,
		0.983187,
		0.989154,
		0.993458,
		0.996388,
		0.998234,
		0.999278,
		0.999779,
		0.999961,
		0.999998,
		1,
		0.999992,
		0.99992,
		0.999666,
		0.999076,
		0.997978,
		0.996194,
		0.993551,
		0.989891,
		0.985072,
		0.978974,
		0.971498,
		0.962567,
		0.952128,
		0.940146,
		0.926607,
		0.911514,
		0.894888,
		0.876762,
		0.857183,
		0.836212,
		0.813918,
		0.790377,
		0.765677,
		0.739908,
		0.71317,
		0.685564,
		0.002556,
		0.015211,
		0.037057,
		0.066642,
		0.102633,
		0.143807,
		0.189047,
		0.237335,
		0.28775,
		0.339457,
		0.391707,
		0.44383,
		0.495233,
		0.545395,
		0.593863,
		0.64025,
		0.68423,
		0.725537,
		0.763962,
		0.799352,
		0.831602,
		0.860662,
		0.886525,
		0.909232,
		0.928869,
		0.94556,
		0.959469,
		0.970795,
		0.979768,
		0.986646,
		0.991705,
		0.995238,
		0.997543,
		0.99891,
		0.999618,
		0.999911,
		0.999992,
		1,
		0.999998,
		0.999962,
		0.999793,
		0.999346,
		0.998451,
		0.996925,
		0.99459,
		0.991278,
		0.986838,
		0.98114,
		0.974075,
		0.965558,
		0.955526,
		0.943938,
		0.930774,
		0.916031,
		0.899725,
		0.881887,
		0.862562,
		0.841807,
		0.819691,
		0.796291,
		0.771692,
		0.745987,
		0.719275,
		0.691659,
		0.663248,
		0.63415,
		0.005737,
		0.021837,
		0.046734,
		0.078987,
		0.117276,
		0.160392,
		0.207232,
		0.256796,
		0.308175,
		0.360554,
		0.413199,
		0.465459,
		0.516759,
		0.566596,
		0.614537,
		0.660215,
		0.703324,
		0.743619,
		0.780913,
		0.815072,
		0.846017,
		0.873716,
		0.898188,
		0.919497,
		0.937751,
		0.953097,
		0.965721,
		0.975845,
		0.983719,
		0.989616,
		0.993829,
		0.996662,
		0.998415,
		0.999381,
		0.999825,
		0.999974,
		0.999999,
		1,
		0.999985,
		0.999882,
		0.999557,
		0.998845,
		0.997561,
		0.995521,
		0.992549,
		0.988485,
		0.983187,
		0.976536,
		0.968439,
		0.958822,
		0.947639,
		0.93486,
		0.92048,
		0.904508,
		0.886971,
		0.867913,
		0.847388,
		0.825463,
		0.802214,
		0.777728,
		0.752097,
		0.72542,
		0.697803,
		0.669354,
		0.640185,
		0.610413,
		0.580154,
		0.000668,
		0.010171,
		0.029626,
		0.057479,
		0.092302,
		0.132788,
		0.177742,
		0.226079,
		0.276811,
		0.32905,
		0.381994,
		0.43493,
		0.487226,
		0.538323,
		0.587741,
		0.635065,
		0.679948,
		0.722107,
		0.761317,
		0.797413,
		0.830285,
		0.859876,
		0.886177,
		0.909232,
		0.929128,
		0.945995,
		0.960006,
		0.971369,
		0.980326,
		0.987147,
		0.992122,
		0.995558,
		0.997765,
		0.999046,
		0.999686,
		0.999935,
		0.999996,
		1,
		0.999996,
		0.99994,
		0.999716,
		0.999166,
		0.998106,
		0.996348,
		0.993706,
		0.990011,
		0.985111,
		0.978877,
		0.971204,
		0.962012,
		0.951243,
		0.938862,
		0.924857,
		0.909232,
		0.892012,
		0.873234,
		0.852952,
		0.83123,
		0.808146,
		0.783784,
		0.758237,
		0.731605,
		0.703994,
		0.675515,
		0.646281,
		0.61641,
		0.586022,
		0.555238,
		0.52418,
		0.002666,
		0.015846,
		0.038563,
		0.069274,
		0.106564,
		0.149143,
		0.19583,
		0.245554,
		0.297347,
		0.350337,
		0.403741,
		0.456863,
		0.50909,
		0.559887,
		0.608789,
		0.655405,
		0.699408,
		0.740538,
		0.778592,
		0.813428,
		0.84496,
		0.873152,
		0.898023,
		0.919639,
		0.938112,
		0.953597,
		0.966289,
		0.97642,
		0.984253,
		0.990077,
		0.994196,
		0.996928,
		0.998588,
		0.999475,
		0.999864,
		0.999984,
		1,
		0.999999,
		0.999974,
		0.999831,
		0.999422,
		0.998566,
		0.997073,
		0.994749,
		0.991416,
		0.986911,
		0.981094,
		0.973851,
		0.96509,
		0.954745,
		0.942774,
		0.929157,
		0.913893,
		0.897002,
		0.878519,
		0.858494,
		0.83699,
		0.814083,
		0.789856,
		0.764404,
		0.737827,
		0.710231,
		0.681729,
		0.652437,
		0.622473,
		0.59196,
		0.561021,
		0.529781,
		0.498366,
		0.466901,
		0.005983,
		0.022747,
		0.048628,
		0.082095,
		0.121749,
		0.166312,
		0.214622,
		0.265625,
		0.318369,
		0.371999,
		0.425753,
		0.478955,
		0.531011,
		0.581406,
		0.629697,
		0.675515,
		0.718555,
		0.758579,
		0.795408,
		0.828924,
		0.859064,
		0.88582,
		0.909232,
		0.929393,
		0.94644,
		0.960553,
		0.971951,
		0.980889,
		0.987648,
		0.992536,
		0.995871,
		0.997978,
		0.999172,
		0.999746,
		0.999955,
		0.999998,
		1,
		0.999991,
		0.999908,
		0.999618,
		0.998946,
		0.9977,
		0.995681,
		0.9927,
		0.988586,
		0.983187,
		0.976376,
		0.968053,
		0.958141,
		0.946591,
		0.933375,
		0.918486,
		0.901939,
		0.883765,
		0.864011,
		0.842738,
		0.820021,
		0.795942,
		0.770596,
		0.744084,
		0.716513,
		0.687996,
		0.658652,
		0.628601,
		0.597967,
		0.566876,
		0.535457,
		0.503836,
		0.472143,
		0.440506,
		0.409054,
		0.000697,
		0.010606,
		0.030859,
		0.059802,
		0.09592,
		0.137829,
		0.184267,
		0.234088,
		0.286256,
		0.339837,
		0.393997,
		0.447991,
		0.501165,
		0.552945,
		0.602837,
		0.650423,
		0.695352,
		0.737347,
		0.77619,
		0.811728,
		0.843868,
		0.872571,
		0.897854,
		0.919785,
		0.938482,
		0.954108,
		0.966867,
		0.977002,
		0.984791,
		0.990536,
		0.994557,
		0.997186,
		0.99875,
		0.999561,
		0.999898,
		0.999991,
		1,
		0.999998,
		0.999956,
		0.999763,
		0.999253,
		0.998234,
		0.996504,
		0.993865,
		0.990135,
		0.985151,
		0.978775,
		0.970896,
		0.961427,
		0.950308,
		0.937506,
		0.923006,
		0.906817,
		0.888967,
		0.869498,
		0.848471,
		0.825958,
		0.80204,
		0.776812,
		0.750376,
		0.722839,
		0.694316,
		0.664927,
		0.634794,
		0.604044,
		0.572806,
		0.541209,
		0.509384,
		0.477463,
		0.445578,
		0.413858,
		0.382435,
		0.351437,
		0.002783,
		0.016522,
		0.040162,
		0.072063,
		0.110724,
		0.154777,
		0.202976,
		0.254193,
		0.307411,
		0.361715,
		0.416289,
		0.470412,
		0.523449,
		0.574848,
		0.624138,
		0.670922,
		0.714875,
		0.755743,
		0.793332,
		0.827516,
		0.858226,
		0.885451,
		0.909232,
		0.929666,
		0.946896,
		0.961112,
		0.972543,
		0.981456,
		0.98815,
		0.992945,
		0.996177,
		0.998182,
		0.999289,
		0.999798,
		0.99997,
		0.999999,
		1,
		0.999983,
		0.999864,
		0.999493,
		0.998681,
		0.997222,
		0.994912,
		0.991558,
		0.986986,
		0.981047,
		0.973615,
		0.964597,
		0.953921,
		0.941544,
		0.927446,
		0.911631,
		0.894119,
		0.874952,
		0.854185,
		0.831889,
		0.808146,
		0.783049,
		0.756699,
		0.729206,
		0.700686,
		0.67126,
		0.641052,
		0.610192,
		0.57881,
		0.547038,
		0.515011,
		0.482863,
		0.450728,
		0.41874,
		0.387032,
		0.355737,
		0.324985,
		0.294906,
		0.006244,
		0.023716,
		0.050639,
		0.085389,
		0.12648,
		0.172561,
		0.222405,
		0.2749,
		0.32905,
		0.383959,
		0.438833,
		0.492972,
		0.545761,
		0.596674,
		0.645261,
		0.69115,
		0.73404,
		0.773702,
		0.809968,
		0.842738,
		0.87197,
		0.897679,
		0.919935,
		0.938862,
		0.95463,
		0.967454,
		0.977591,
		0.985332,
		0.990993,
		0.994912,
		0.997435,
		0.998903,
		0.999638,
		0.999926,
		0.999995,
		1,
		0.999995,
		0.999931,
		0.999675,
		0.999046,
		0.997839,
		0.995843,
		0.992856,
		0.988691,
		0.983187,
		0.976208,
		0.967646,
		0.957423,
		0.945484,
		0.931803,
		0.916375,
		0.899218,
		0.880367,
		0.859876,
		0.837812,
		0.814257,
		0.789303,
		0.763052,
		0.735615,
		0.707107,
		0.677651,
		0.647376,
		0.61641,
		0.584888,
		0.552945,
		0.520718,
		0.488343,
		0.455958,
		0.423701,
		0.391707,
		0.360112,
		0.32905,
		0.298652,
		0.269049,
		0.24037,
		0.000729,
		0.011069,
		0.032169,
		0.062267,
		0.099754,
		0.14316,
		0.191151,
		0.242518,
		0.296172,
		0.351133,
		0.406528,
		0.461585,
		0.515624,
		0.568056,
		0.618377,
		0.666161,
		0.711062,
		0.752804,
		0.791183,
		0.82606,
		0.85736,
		0.88507,
		0.909232,
		0.929947,
		0.947363,
		0.961681,
		0.973143,
		0.982029,
		0.988652,
		0.99335,
		0.996474,
		0.998376,
		0.999396,
		0.999844,
		0.999981,
		1,
		0.999999,
		0.99997,
		0.999805,
		0.999337,
		0.998361,
		0.996662,
		0.994029,
		0.990264,
		0.985193,
		0.978669,
		0.970571,
		0.96081,
		0.949321,
		0.93607,
		0.921045,
		0.904257,
		0.885739,
		0.865538,
		0.843722,
		0.82037,
		0.795573,
		0.769433,
		0.742062,
		0.713576,
		0.6841,
		0.653763,
		0.622699,
		0.591042,
		0.55893,
		0.526505,
		0.493904,
		0.46127,
		0.428742,
		0.39646,
		0.364563,
		0.333186,
		0.302467,
		0.272538,
		0.24353,
		0.215573,
		0.188793,
		0.002907,
		0.017243,
		0.041863,
		0.075024,
		0.11513,
		0.160731,
		0.210511,
		0.26328,
		0.317968,
		0.373618,
		0.429378,
		0.484498,
		0.538323,
		0.590287,
		0.63991,
		0.686792,
		0.730613,
		0.771123,
		0.808146,
		0.84157,
		0.87135,
		0.897499,
		0.92009,
		0.939252,
		0.955164,
		0.968053,
		0.978187,
		0.985874,
		0.991447,
		0.99526,
		0.997675,
		0.999046,
		0.999707,
		0.999948,
		0.999998,
		1,
		0.99999,
		0.999894,
		0.999561,
		0.998793,
		0.997372,
		0.995079,
		0.991705,
		0.987065,
		0.980996,
		0.973368,
		0.964076,
		0.95305,
		0.940242,
		0.925634,
		0.909232,
		0.891062,
		0.871169,
		0.849616,
		0.826481,
		0.801855,
		0.77584,
		0.748546,
		0.720092,
		0.690605,
		0.660215,
		0.629057,
		0.59727,
		0.564995,
		0.532373,
		0.499548,
		0.466665,
		0.433865,
		0.401293,
		0.369091,
		0.337397,
		0.306352,
		0.276092,
		0.246752,
		0.218463,
		0.191355,
		0.165555,
		0.141187,
		0.006524,
		0.024747,
		0.052777,
		0.088884,
		0.13149,
		0.179163,
		0.230606,
		0.28465,
		0.340247,
		0.39646,
		0.452462,
		0.507526,
		0.561021,
		0.612405,
		0.661224,
		0.707107,
		0.749757,
		0.788955,
		0.824552,
		0.856465,
		0.884676,
		0.909232,
		0.930235,
		0.947842,
		0.962262,
		0.973751,
		0.982605,
		0.989154,
		0.99375,
		0.996763,
		0.99856,
		0.999493,
		0.999882,
		0.999989,
		1,
		0.999998,
		0.99995,
		0.999727,
		0.999142,
		0.997978,
		0.996009,
		0.993016,
		0.9888,
		0.983187,
		0.976031,
		0.967218,
		0.956664,
		0.944313,
		0.930138,
		0.914137,
		0.896331,
		0.876762,
		0.855488,
		0.832587,
		0.808146,
		0.782268,
		0.755064,
		0.726654,
		0.697165,
		0.66673,
		0.635486,
		0.603574,
		0.571138,
		0.538323,
		0.505276,
		0.472143,
		0.439072,
		0.406208,
		0.373698,
		0.341684,
		0.310309,
		0.279714,
		0.250035,
		0.221409,
		0.193969,
		0.167843,
		0.14316,
		0.120042,
		0.098612,
		0.000762,
		0.011564,
		0.033565,
		0.064888,
		0.103819,
		0.148801,
		0.198419,
		0.251397,
		0.306588,
		0.362964,
		0.419614,
		0.475734,
		0.53062,
		0.583667,
		0.63436,
		0.682272,
		0.727057,
		0.76845,
		0.806258,
		0.840361,
		0.870708,
		0.897313,
		0.920249,
		0.939652,
		0.95571,
		0.968661,
		0.97879,
		0.986419,
		0.991898,
		0.995601,
		0.997904,
		0.999178,
		0.999767,
		0.999965,
		0.999999,
		1,
		0.99998,
		0.999844,
		0.999417,
		0.998486,
		0.996822,
		0.994196,
		0.990397,
		0.985237,
		0.978558,
		0.97023,
		0.960158,
		0.948277,
		0.934549,
		0.918966,
		0.901541,
		0.882312,
		0.861335,
		0.838682,
		0.814442,
		0.788716,
		0.761615,
		0.73326,
		0.70378,
		0.673308,
		0.641985,
		0.609954,
		0.577362,
		0.544357,
		0.511089,
		0.477707,
		0.444363,
		0.411206,
		0.378386,
		0.346048,
		0.31434,
		0.283404,
		0.253383,
		0.224414,
		0.196635,
		0.170178,
		0.145174,
		0.121749,
		0.100028,
		0.080132,
		0.062177,
		0.003041,
		0.018011,
		0.043674,
		0.07817,
		0.119801,
		0.16703,
		0.218463,
		0.272844,
		0.32905,
		0.386075,
		0.443032,
		0.499142,
		0.553728,
		0.606211,
		0.656102,
		0.703003,
		0.746597,
		0.786646,
		0.82299,
		0.855538,
		0.88427,
		0.909232,
		0.930531,
		0.948332,
		0.962855,
		0.974369,
		0.983187,
		0.989654,
		0.994144,
		0.997042,
		0.998733,
		0.999581,
		0.999914,
		0.999994,
		1,
		0.999995,
		0.99992,
		0.999625,
		0.998903,
		0.997523,
		0.995249,
		0.991857,
		0.987147,
		0.980944,
		0.973107,
		0.963527,
		0.952128,
		0.938862,
		0.923712,
		0.906686,
		0.887815,
		0.867151,
		0.844764,
		0.820741,
		0.795181,
		0.768196,
		0.739908,
		0.710447,
		0.679948,
		0.648554,
		0.61641,
		0.583667,
		0.550474,
		0.516987,
		0.483358,
		0.449741,
		0.416289,
		0.383156,
		0.350491,
		0.318445,
		0.287165,
		0.256796,
		0.227479,
		0.199356,
		0.172561,
		0.14723,
		0.123492,
		0.101474,
		0.081301,
		0.063092,
		0.046964,
		0.033031,
		0.006822,
		0.025848,
		0.055053,
		0.092596,
		0.136799,
		0.186142,
		0.239256,
		0.294906,
		0.35199,
		0.409531,
		0.466665,
		0.522638,
		0.5768,
		0.628601,
		0.67758,
		0.723368,
		0.765677,
		0.804301,
		0.839109,
		0.870045,
		0.897121,
		0.920413,
		0.940063,
		0.956268,
		0.969281,
		0.9794,
		0.986964,
		0.992346,
		0.995933,
		0.998123,
		0.9993,
		0.999819,
		0.999978,
		1,
		0.999999,
		0.999965,
		0.999775,
		0.999236,
		0.998117,
		0.996177,
		0.993181,
		0.988913,
		0.983187,
		0.975845,
		0.966766,
		0.955861,
		0.943071,
		0.928371,
		0.91176,
		0.893264,
		0.872931,
		0.850827,
		0.827036,
		0.801659,
		0.774805,
		0.746597,
		0.717165,
		0.686649,
		0.655192,
		0.622942,
		0.590052,
		0.556677,
		0.522972,
		0.489096,
		0.455206,
		0.421459,
		0.38801,
		0.355015,
		0.322628,
		0.290998,
		0.260276,
		0.230606,
		0.202132,
		0.174994,
		0.14933,
		0.125272,
		0.102952,
		0.082496,
		0.064028,
		0.047667,
		0.033529,
		0.021727,
		0.01237,
		0.000798,
		0.012092,
		0.035053,
		0.067676,
		0.108136,
		0.154777,
		0.2061,
		0.260756,
		0.317536,
		0.375363,
		0.433283,
		0.49046,
		0.546167,
		0.599783,
		0.650784,
		0.698742,
		0.743316,
		0.78425,
		0.82137,
		0.854579,
		0.883851,
		0.909232,
		0.930835,
		0.948834,
		0.963459,
		0.974996,
		0.983772,
		0.990153,
		0.994532,
		0.997312,
		0.998895,
		0.99966,
		0.999939,
		0.999998,
		1,
		0.999988,
		0.999877,
		0.999493,
		0.99861,
		0.996984,
		0.994368,
		0.990536,
		0.985283,
		0.97844,
		0.96987,
		0.959469,
		0.94717,
		0.932935,
		0.916757,
		0.898654,
		0.87867,
		0.856866,
		0.833326,
		0.808146,
		0.781437,
		0.753323,
		0.723934,
		0.69341,
		0.661898,
		0.629549,
		0.596519,
		0.562964,
		0.529046,
		0.494924,
		0.460761,
		0.426716,
		0.39295,
		0.359622,
		0.326889,
		0.294906,
		0.263824,
		0.233795,
		0.204965,
		0.177478,
		0.151474,
		0.127091,
		0.104462,
		0.083718,
		0.064984,
		0.048385,
		0.034038,
		0.022059,
		0.01256,
		0.005649,
		0.001428,
		0.003183,
		0.018832,
		0.045605,
		0.081516,
		0.12476,
		0.1737,
		0.226862,
		0.282919,
		0.340688,
		0.399117,
		0.457278,
		0.514364,
		0.569676,
		0.622621,
		0.672708,
		0.719536,
		0.762798,
		0.80227,
		0.837812,
		0.869359,
		0.896922,
		0.920582,
		0.940484,
		0.956839,
		0.969911,
		0.980016,
		0.987511,
		0.992789,
		0.996257,
		0.998331,
		0.999412,
		0.999863,
		0.999987,
		1,
		0.999998,
		0.999942,
		0.999684,
		0.999011,
		0.997675,
		0.995423,
		0.992015,
		0.987232,
		0.980889,
		0.972832,
		0.962947,
		0.951151,
		0.937398,
		0.92167,
		0.903979,
		0.884362,
		0.862877,
		0.839605,
		0.814639,
		0.788092,
		0.760084,
		0.73075,
		0.70023,
		0.668673,
		0.636233,
		0.603067,
		0.569338,
		0.535208,
		0.500843,
		0.466406,
		0.432063,
		0.397978,
		0.364313,
		0.331231,
		0.298888,
		0.267443,
		0.237049,
		0.207857,
		0.180014,
		0.153664,
		0.128949,
		0.106005,
		0.084966,
		0.065963,
		0.049119,
		0.034559,
		0.022399,
		0.012755,
		0.005737,
		0.001451,
		0.007142,
		0.027023,
		0.057479,
		0.096544,
		0.142431,
		0.193529,
		0.248386,
		0.3057,
		0.364313,
		0.423201,
		0.481465,
		0.538323,
		0.593109,
		0.645261,
		0.694316,
		0.739908,
		0.781763,
		0.819691,
		0.853585,
		0.883417,
		0.909232,
		0.931149,
		0.949349,
		0.964076,
		0.975631,
		0.984361,
		0.99065,
		0.994912,
		0.997571,
		0.999046,
		0.999729,
		0.999959,
		0.999999,
		1,
		0.999977,
		0.999819,
		0.999326,
		0.998255,
		0.996348,
		0.99335,
		0.989031,
		0.983187,
		0.97565,
		0.966289,
		0.95501,
		0.941754,
		0.926493,
		0.909232,
		0.890001,
		0.868855,
		0.845868,
		0.821134,
		0.794763,
		0.766878,
		0.737611,
		0.707107,
		0.675515,
		0.642991,
		0.609698,
		0.575799,
		0.541461,
		0.506853,
		0.472143,
		0.437501,
		0.403095,
		0.369091,
		0.335654,
		0.302949,
		0.271135,
		0.24037,
		0.210809,
		0.182603,
		0.155901,
		0.130847,
		0.107582,
		0.086243,
		0.066963,
		0.049871,
		0.035092,
		0.022747,
		0.012955,
		0.005827,
		0.001474,
		0.000836,
		0.012657,
		0.036642,
		0.070648,
		0.112726,
		0.161115,
		0.214224,
		0.270627,
		0.32905,
		0.38836,
		0.447562,
		0.505785,
		0.56228,
		0.61641,
		0.667644,
		0.715554,
		0.759807,
		0.800163,
		0.836467,
		0.868649,
		0.896717,
		0.920755,
		0.940917,
		0.957423,
		0.970552,
		0.980638,
		0.988059,
		0.993226,
		0.996571,
		0.998528,
		0.999512,
		0.999899,
		0.999993,
		1,
		0.999994,
		0.999907,
		0.999565,
		0.998733,
		0.997147,
		0.994545,
		0.990679,
		0.985332,
		0.978317,
		0.969489,
		0.958739,
		0.945995,
		0.931219,
		0.914407,
		0.895582,
		0.874792,
		0.85211,
		0.827626,
		0.80145,
		0.773702,
		0.744517,
		0.71404,
		0.682423,
		0.649825,
		0.61641,
		0.582346,
		0.547804,
		0.512956,
		0.477974,
		0.443032,
		0.408303,
		0.373956,
		0.340162,
		0.307089,
		0.2749,
		0.243759,
		0.213823,
		0.185248,
		0.158187,
		0.132788,
		0.109195,
		0.087549,
		0.067986,
		0.050639,
		0.035637,
		0.023103,
		0.013159,
		0.00592,
		0.001497,
		0.003336,
		0.01971,
		0.047667,
		0.085081,
		0.130029,
		0.18077,
		0.23574,
		0.293538,
		0.352917,
		0.412775,
		0.472143,
		0.530184,
		0.586177,
		0.63952,
		0.689715,
		0.736367,
		0.77918,
		0.817948,
		0.852555,
		0.882967,
		0.909232,
		0.931471,
		0.949876,
		0.964706,
		0.976276,
		0.984953,
		0.991144,
		0.995285,
		0.997819,
		0.999185,
		0.999788,
		0.999974,
		1,
		0.999999,
		0.999959,
		0.999739,
		0.999115,
		0.997826,
		0.995601,
		0.992177,
		0.987321,
		0.980831,
		0.972543,
		0.962333,
		0.950115,
		0.935842,
		0.919497,
		0.901097,
		0.880684,
		0.858326,
		0.834112,
		0.808146,
		0.780552,
		0.751464,
		0.721027,
		0.689396,
		0.656733,
		0.623205,
		0.588982,
		0.55424,
		0.519153,
		0.483901,
		0.448658,
		0.413603,
		0.378911,
		0.344756,
		0.31131,
		0.278741,
		0.247217,
		0.2169,
		0.18795,
		0.160523,
		0.134771,
		0.110843,
		0.088884,
		0.069032,
		0.051425,
		0.036195,
		0.023468,
		0.013368,
		0.006014,
		0.001521,
		0.007484,
		0.02828,
		0.060068,
		0.100747,
		0.148414,
		0.201355,
		0.258031,
		0.317068,
		0.37725,
		0.437501,
		0.496887,
		0.5546,
		0.609954,
		0.66238,
		0.711414,
		0.756699,
		0.797974,
		0.835071,
		0.867913,
		0.896506,
		0.920935,
		0.941362,
		0.95802,
		0.971204,
		0.981267,
		0.988607,
		0.993658,
		0.996876,
		0.998713,
		0.999603,
		0.999929,
		0.999997,
		1,
		0.999986,
		0.999857,
		0.999412,
		0.998392,
		0.996521,
		0.993526,
		0.989154,
		0.983187,
		0.975444,
		0.965784,
		0.954108,
		0.940353,
		0.924495,
		0.90654,
		0.886525,
		0.864511,
		0.840584,
		0.814849,
		0.787425,
		0.758449,
		0.728066,
		0.696434,
		0.663715,
		0.630081,
		0.595706,
		0.560768,
		0.525446,
		0.489923,
		0.454381,
		0.418999,
		0.383959,
		0.349439,
		0.315615,
		0.282661,
		0.250747,
		0.220042,
		0.19071,
		0.16291,
		0.136799,
		0.112529,
		0.090249,
		0.070103,
		0.05223,
		0.036766,
		0.023841,
		0.013582,
		0.006111,
		0.001546,
		0.000878,
		0.013263,
		0.038342,
		0.073818,
		0.117611,
		0.167843,
		0.222825,
		0.281048,
		0.341166,
		0.40199,
		0.462481,
		0.521733,
		0.578973,
		0.63355,
		0.684929,
		0.732684,
		0.776495,
		0.816138,
		0.851487,
		0.882503,
		0.909232,
		0.931803,
		0.950418,
		0.965348,
		0.976929,
		0.985548,
		0.991635,
		0.995649,
		0.998056,
		0.999313,
		0.999839,
		0.999985,
		1,
		0.999997,
		0.999932,
		0.999632,
		0.998853,
		0.997312,
		0.994726,
		0.990828,
		0.985382,
		0.978187,
		0.969087,
		0.957965,
		0.944746,
		0.929393,
		0.911904,
		0.892307,
		0.870658,
		0.84704,
		0.821553,
		0.794318,
		0.76547,
		0.735156,
		0.703534,
		0.670771,
		0.63704,
		0.602518,
		0.567389,
		0.531836,
		0.496045,
		0.460201,
		0.424491,
		0.3891,
		0.354211,
		0.320005,
		0.28666,
		0.254351,
		0.223252,
		0.193529,
		0.165349,
		0.138872,
		0.114253,
		0.091646,
		0.071199,
		0.053054,
		0.037351,
		0.024223,
		0.013802,
		0.006211,
		0.001571,
		0.003501,
		0.020651,
		0.049871,
		0.088884,
		0.135635,
		0.188271,
		0.245133,
		0.304739,
		0.365774,
		0.427082,
		0.487653,
		0.546619,
		0.603241,
		0.656903,
		0.707107,
		0.753466,
		0.795699,
		0.833622,
		0.867151,
		0.896287,
		0.92112,
		0.94182,
		0.958631,
		0.971868,
		0.981901,
		0.989154,
		0.994084,
		0.997169,
		0.998886,
		0.999682,
		0.999952,
		0.999999,
		1,
		0.999973,
		0.999788,
		0.999216,
		0.997978,
		0.995782,
		0.992346,
		0.987414,
		0.98077,
		0.972236,
		0.961681,
		0.949014,
		0.934185,
		0.917181,
		0.898023,
		0.876762,
		0.853472,
		0.828254,
		0.801227,
		0.772523,
		0.742293,
		0.710695,
		0.677898,
		0.64408,
		0.60942,
		0.574105,
		0.538323,
		0.502265,
		0.466121,
		0.430082,
		0.394338,
		0.359076,
		0.324482,
		0.290741,
		0.258031,
		0.22653,
		0.196411,
		0.167843,
		0.140991,
		0.116017,
		0.093076,
		0.07232,
		0.053897,
		0.03795,
		0.024615,
		0.014026,
		0.006313,
		0.001597,
		0.007852,
		0.029626,
		0.062836,
		0.105229,
		0.154777,
		0.209654,
		0.268228,
		0.32905,
		0.390835,
		0.452462,
		0.512956,
		0.571482,
		0.627338,
		0.679948,
		0.728851,
		0.773702,
		0.814257,
		0.850378,
		0.882021,
		0.909232,
		0.932145,
		0.950973,
		0.966004,
		0.977591,
		0.986146,
		0.992122,
		0.996003,
		0.99828,
		0.999429,
		0.999882,
		0.999992,
		1,
		0.999993,
		0.999891,
		0.999493,
		0.998528,
		0.996697,
		0.993706,
		0.989282,
		0.983187,
		0.975226,
		0.965249,
		0.953149,
		0.938862,
		0.922365,
		0.903668,
		0.882814,
		0.859876,
		0.834948,
		0.808146,
		0.779606,
		0.749475,
		0.717915,
		0.685097,
		0.651201,
		0.61641,
		0.580916,
		0.54491,
		0.508587,
		0.472143,
		0.435774,
		0.399673,
		0.364035,
		0.32905,
		0.294906,
		0.261788,
		0.229879,
		0.199356,
		0.170393,
		0.14316,
		0.117821,
		0.094539,
		0.073468,
		0.054761,
		0.038563,
		0.025016,
		0.014256,
		0.006417,
		0.001624,
		0.000922,
		0.013913,
		0.040162,
		0.077206,
		0.122817,
		0.174994,
		0.231941,
		0.292057,
		0.353922,
		0.416289,
		0.478069,
		0.538323,
		0.596255,
		0.651201,
		0.702622,
		0.750101,
		0.793332,
		0.832117,
		0.866359,
		0.89606,
		0.921311,
		0.94229,
		0.959257,
		0.972543,
		0.982541,
		0.9897,
		0.994502,
		0.997451,
		0.999046,
		0.999752,
		0.99997,
		1,
		0.999999,
		0.999952,
		0.999695,
		0.99897,
		0.997478,
		0.994912,
		0.990983,
		0.985435,
		0.978051,
		0.968661,
		0.957143,
		0.943417,
		0.927446,
		0.909232,
		0.88881,
		0.866244,
		0.841627,
		0.815072,
		0.786713,
		0.756699,
		0.725192,
		0.692366,
		0.658402,
		0.62349,
		0.587822,
		0.551596,
		0.515011,
		0.478268,
		0.441568,
		0.405109,
		0.369091,
		0.333709,
		0.299157,
		0.265625,
		0.2333,
		0.202366,
		0.173,
		0.145377,
		0.119668,
		0.096037,
		0.074644,
		0.055645,
		0.039191,
		0.025427,
		0.014492,
		0.006524,
		0.001651,
		0.003677,
		0.021661,
		0.05223,
		0.092945,
		0.141606,
		0.19624,
		0.255081,
		0.316562,
		0.379296,
		0.442072,
		0.503836,
		0.563688,
		0.620871,
		0.67476,
		0.72486,
		0.770794,
		0.812301,
		0.849227,
		0.881522,
		0.909232,
		0.932498,
		0.951543,
		0.966673,
		0.978262,
		0.986746,
		0.992605,
		0.996348,
		0.998492,
		0.999533,
		0.999917,
		0.999997,
		1,
		0.999984,
		0.999833,
		0.999313,
		0.99813,
		0.995967,
		0.99252,
		0.987511,
		0.980706,
		0.971912,
		0.960989,
		0.947842,
		0.932418,
		0.914709,
		0.89474,
		0.872571,
		0.848287,
		0.822,
		0.793842,
		0.763962,
		0.732524,
		0.699703,
		0.665684,
		0.630658,
		0.594824,
		0.558383,
		0.521539,
		0.484498,
		0.447466,
		0.410647,
		0.374245,
		0.338462,
		0.303496,
		0.269544,
		0.236797,
		0.205443,
		0.175666,
		0.147646,
		0.121558,
		0.09757,
		0.075848,
		0.056551,
		0.039835,
		0.025848,
		0.014734,
		0.006633,
		0.001679,
		0.008248,
		0.031071,
		0.065798,
		0.110014,
		0.161551,
		0.218463,
		0.279019,
		0.341684,
		0.405109,
		0.468117,
		0.529695,
		0.588982,
		0.645261,
		0.69795,
		0.746597,
		0.790869,
		0.830552,
		0.865538,
		0.895825,
		0.921508,
		0.942774,
		0.959898,
		0.973229,
		0.983187,
		0.990244,
		0.994912,
		0.997721,
		0.999193,
		0.999811,
		0.999982,
		1,
		0.999997,
		0.99992,
		0.99957,
		0.998662,
		0.996876,
		0.993892,
		0.989415,
		0.983187,
		0.974996,
		0.964681,
		0.952128,
		0.937272,
		0.92009,
		0.900598,
		0.878848,
		0.854921,
		0.828924,
		0.800988,
		0.771261,
		0.739908,
		0.707107,
		0.673044,
		0.637915,
		0.601923,
		0.565272,
		0.528173,
		0.490835,
		0.453471,
		0.416289,
		0.379501,
		0.343311,
		0.307926,
		0.273547,
		0.24037,
		0.208589,
		0.178394,
		0.149968,
		0.123492,
		0.09914,
		0.077081,
		0.057479,
		0.040494,
		0.026279,
		0.014982,
		0.006746,
		0.001708,
		0.00097,
		0.014613,
		0.042115,
		0.08083,
		0.128373,
		0.182603,
		0.24161,
		0.303696,
		0.367361,
		0.431293,
		0.494356,
		0.555576,
		0.614133,
		0.669354,
		0.720701,
		0.767766,
		0.810266,
		0.848031,
		0.881004,
		0.909232,
		0.932861,
		0.952128,
		0.967356,
		0.978942,
		0.987347,
		0.993081,
		0.996681,
		0.99869,
		0.999626,
		0.999944,
		0.999999,
		1,
		0.999969,
		0.999752,
		0.999083,
		0.997644,
		0.995103,
		0.991144,
		0.98549,
		0.977906,
		0.968209,
		0.956268,
		0.941999,
		0.925367,
		0.906376,
		0.88507,
		0.861523,
		0.835839,
		0.808146,
		0.778592,
		0.747342,
		0.714575,
		0.680482,
		0.645261,
		0.609118,
		0.572264,
		0.534913,
		0.497281,
		0.459584,
		0.422038,
		0.384859,
		0.34826,
		0.31245,
		0.277636,
		0.244022,
		0.211807,
		0.181184,
		0.152345,
		0.125472,
		0.100747,
		0.078344,
		0.05843,
		0.04117,
		0.026722,
		0.015236,
		0.006861,
		0.001737,
		0.003868,
		0.022747,
		0.054761,
		0.097288,
		0.147975,
		0.204713,
		0.265625,
		0.32905,
		0.393526,
		0.457781,
		0.520718,
		0.581406,
		0.639069,
		0.693079,
		0.742944,
		0.788304,
		0.828924,
		0.864686,
		0.895582,
		0.921711,
		0.943272,
		0.960553,
		0.973927,
		0.983837,
		0.990785,
		0.995313,
		0.997978,
		0.999327,
		0.999861,
		0.999991,
		1,
		0.999991,
		0.999872,
		0.999406,
		0.99828,
		0.996156,
		0.9927,
		0.987613,
		0.980638,
		0.971569,
		0.960253,
		0.946591,
		0.930531,
		0.912065,
		0.891227,
		0.868085,
		0.842738,
		0.815311,
		0.785951,
		0.754822,
		0.722107,
		0.687996,
		0.652694,
		0.61641,
		0.579359,
		0.541761,
		0.503836,
		0.465807,
		0.427896,
		0.390324,
		0.353309,
		0.317068,
		0.281814,
		0.247755,
		0.215097,
		0.184039,
		0.154777,
		0.1275,
		0.102394,
		0.079638,
		0.059405,
		0.041863,
		0.027175,
		0.015497,
		0.006979,
		0.001767,
		0.008674,
		0.032625,
		0.068974,
		0.11513,
		0.168771,
		0.227824,
		0.290446,
		0.355015,
		0.42011,
		0.484498,
		0.547127,
		0.607109,
		0.663715,
		0.716363,
		0.76461,
		0.808146,
		0.846787,
		0.880467,
		0.909232,
		0.933236,
		0.952728,
		0.968053,
		0.97963,
		0.987949,
		0.993551,
		0.997003,
		0.998875,
		0.999707,
		0.999964,
		1,
		0.999999,
		0.999944,
		0.999641,
		0.998793,
		0.997056,
		0.994084,
		0.989554,
		0.983187,
		0.974752,
		0.964076,
		0.951039,
		0.935573,
		0.917657,
		0.897313,
		0.874601,
		0.849616,
		0.822478,
		0.793332,
		0.762345,
		0.729698,
		0.695586,
		0.660215,
		0.623799,
		0.586559,
		0.548718,
		0.510503,
		0.472143,
		0.433865,
		0.395897,
		0.358463,
		0.321785,
		0.286083,
		0.251572,
		0.218463,
		0.186961,
		0.157267,
		0.129577,
		0.104082,
		0.080964,
		0.060404,
		0.042574,
		0.027641,
		0.015765,
		0.007101,
		0.001798,
		0.001021,
		0.015366,
		0.044213,
		0.084714,
		0.13431,
		0.19071,
		0.251877,
		0.31601,
		0.381525,
		0.447041,
		0.511372,
		0.573509,
		0.632611,
		0.687996,
		0.739133,
		0.78563,
		0.827229,
		0.863799,
		0.895329,
		0.921922,
		0.943785,
		0.961225,
		0.974637,
		0.984492,
		0.991323,
		0.995704,
		0.998222,
		0.999448,
		0.999901,
		0.999996,
		1,
		0.999981,
		0.999803,
		0.999193,
		0.997811,
		0.995298,
		0.991312,
		0.985548,
		0.977753,
		0.967729,
		0.955335,
		0.940484,
		0.923141,
		0.903317,
		0.881063,
		0.856465,
		0.82964,
		0.800733,
		0.769907,
		0.737347,
		0.703249,
		0.667822,
		0.631286,
		0.593863,
		0.555785,
		0.517284,
		0.478594,
		0.439948,
		0.40158,
		0.363722,
		0.326602,
		0.290446,
		0.255475,
		0.221906,
		0.189951,
		0.159817,
		0.131705,
		0.10581,
		0.082324,
		0.061429,
		0.043303,
		0.028118,
		0.016039,
		0.007225,
		0.00183,
		0.004073,
		0.023917,
		0.057479,
		0.101941,
		0.154777,
		0.213733,
		0.276811,
		0.342249,
		0.408505,
		0.474244,
		0.538323,
		0.599783,
		0.657831,
		0.711836,
		0.761317,
		0.805937,
		0.845492,
		0.879909,
		0.909232,
		0.933622,
		0.953345,
		0.968764,
		0.980326,
		0.988552,
		0.994013,
		0.997312,
		0.999046,
		0.999776,
		0.999979,
		1,
		0.999996,
		0.999905,
		0.999493,
		0.998429,
		0.996348,
		0.992887,
		0.98772,
		0.980567,
		0.971204,
		0.959469,
		0.945255,
		0.928511,
		0.909232,
		0.887461,
		0.863277,
		0.836792,
		0.808146,
		0.777504,
		0.74505,
		0.710982,
		0.675515,
		0.638868,
		0.601273,
		0.562964,
		0.52418,
		0.48516,
		0.446146,
		0.407377,
		0.369091,
		0.331523,
		0.294906,
		0.259467,
		0.225429,
		0.193012,
		0.162428,
		0.133884,
		0.107582,
		0.083718,
		0.06248,
		0.044051,
		0.028608,
		0.016321,
		0.007353,
		0.001863,
		0.009134,
		0.034297,
		0.072383,
		0.120607,
		0.176478,
		0.237781,
		0.302559,
		0.369091,
		0.43588,
		0.501639,
		0.565272,
		0.625869,
		0.68269,
		0.735156,
		0.782841,
		0.825463,
		0.862877,
		0.895068,
		0.922139,
		0.944313,
		0.961912,
		0.975358,
		0.985151,
		0.991857,
		0.996084,
		0.998451,
		0.999556,
		0.999933,
		0.999999,
		1,
		0.999963,
		0.999707,
		0.998921,
		0.997238,
		0.994282,
		0.9897,
		0.983187,
		0.974494,
		0.963433,
		0.949876,
		0.933754,
		0.915048,
		0.893788,
		0.870045,
		0.843926,
		0.815567,
		0.785132,
		0.752804,
		0.718785,
		0.683291,
		0.646547,
		0.608789,
		0.570256,
		0.531193,
		0.491846,
		0.452462,
		0.413289,
		0.37457,
		0.336549,
		0.299464,
		0.263549,
		0.229035,
		0.196146,
		0.165103,
		0.136118,
		0.109399,
		0.085147,
		0.063557,
		0.044818,
		0.02911,
		0.01661,
		0.007484,
		0.001896,
		0.001077,
		0.016179,
		0.046472,
		0.088884,
		0.140662,
		0.199356,
		0.26279,
		0.32905,
		0.39646,
		0.463572,
		0.529145,
		0.592135,
		0.651685,
		0.707107,
		0.757879,
		0.803632,
		0.844144,
		0.879329,
		0.909232,
		0.934022,
		0.953979,
		0.969489,
		0.98103,
		0.989154,
		0.994468,
		0.997607,
		0.999202,
		0.999834,
		0.999989,
		1,
		0.99999,
		0.999848,
		0.999299,
		0.997978,
		0.995499,
		0.991486,
		0.98561,
		0.977591,
		0.967218,
		0.954338,
		0.938862,
		0.920755,
		0.900035,
		0.876762,
		0.851035,
		0.82299,
		0.792785,
		0.760605,
		0.726654,
		0.69115,
		0.654322,
		0.61641,
		0.577661,
		0.538323,
		0.498652,
		0.458899,
		0.419319,
		0.380165,
		0.341684,
		0.304124,
		0.267725,
		0.232725,
		0.199356,
		0.167843,
		0.138407,
		0.111261,
		0.086613,
		0.064663,
		0.045605,
		0.029626,
		0.016907,
		0.007619,
		0.001931,
		0.004296,
		0.025179,
		0.060404,
		0.106933,
		0.162051,
		0.223347,
		0.28869,
		0.356209,
		0.424278,
		0.491497,
		0.556677,
		0.618827,
		0.677145,
		0.731,
		0.779928,
		0.823621,
		0.861918,
		0.894796,
		0.922365,
		0.944857,
		0.962616,
		0.976091,
		0.985814,
		0.992386,
		0.996452,
		0.998665,
		0.999651,
		0.999957,
		1,
		0.999999,
		0.999933,
		0.999575,
		0.998577,
		0.996543,
		0.993081,
		0.987832,
		0.980491,
		0.970816,
		0.958631,
		0.943825,
		0.926344,
		0.906191,
		0.883417,
		0.858112,
		0.830407,
		0.800458,
		0.76845,
		0.734587,
		0.699089,
		0.662191,
		0.624138,
		0.58518,
		0.545574,
		0.50558,
		0.465459,
		0.425471,
		0.385876,
		0.34693,
		0.308888,
		0.271997,
		0.236502,
		0.202643,
		0.170651,
		0.140754,
		0.113171,
		0.088117,
		0.065798,
		0.046414,
		0.030156,
		0.017212,
		0.007758,
		0.001966,
		0.009632,
		0.036101,
		0.076051,
		0.12648,
		0.184715,
		0.248386,
		0.315408,
		0.383959,
		0.452462,
		0.51957,
		0.584147,
		0.645261,
		0.702164,
		0.754286,
		0.801227,
		0.842738,
		0.878726,
		0.909232,
		0.934434,
		0.95463,
		0.97023,
		0.981742,
		0.989754,
		0.994912,
		0.997888,
		0.999343,
		0.999882,
		0.999995,
		1,
		0.999977,
		0.999766,
		0.999046,
		0.997422,
		0.994485,
		0.989852,
		0.983187,
		0.974219,
		0.962746,
		0.948631,
		0.931803,
		0.912247,
		0.890001,
		0.865149,
		0.837812,
		0.808146,
		0.776333,
		0.742579,
		0.707107,
		0.670154,
		0.631971,
		0.592813,
		0.552945,
		0.512632,
		0.472143,
		0.431746,
		0.391707,
		0.352291,
		0.31376,
		0.276369,
		0.24037,
		0.20601,
		0.173528,
		0.14316,
		0.11513,
		0.08966,
		0.066963,
		0.047243,
		0.0307,
		0.017525,
		0.0079,
		0.002002,
		0.001138,
		0.017058,
		0.048908,
		0.093366,
		0.14747,
		0.208589,
		0.2744,
		0.342867,
		0.412216,
		0.480925,
		0.547701,
		0.611467,
		0.671346,
		0.726654,
		0.776885,
		0.821699,
		0.860918,
		0.894513,
		0.922598,
		0.945417,
		0.963338,
		0.976835,
		0.986479,
		0.992908,
		0.996806,
		0.998863,
		0.999732,
		0.999975,
		1,
		0.999995,
		0.999887,
		0.999399,
		0.998145,
		0.995704,
		0.991668,
		0.985674,
		0.97742,
		0.966673,
		0.953272,
		0.937122,
		0.918192,
		0.896506,
		0.872136,
		0.845197,
		0.815842,
		0.78425,
		0.750628,
		0.715201,
		0.67821,
		0.63991,
		0.600563,
		0.560439,
		0.519811,
		0.478955,
		0.438147,
		0.397661,
		0.357769,
		0.318742,
		0.280842,
		0.24433,
		0.20946,
		0.176478,
		0.145627,
		0.11714,
		0.091244,
		0.068159,
		0.048096,
		0.031259,
		0.017847,
		0.008046,
		0.00204,
		0.004537,
		0.026543,
		0.063557,
		0.112297,
		0.169842,
		0.233606,
		0.301315,
		0.370984,
		0.440894,
		0.509577,
		0.575799,
		0.638541,
		0.696992,
		0.750529,
		0.798713,
		0.841272,
		0.878098,
		0.909232,
		0.93486,
		0.955299,
		0.970985,
		0.982461,
		0.990352,
		0.995346,
		0.998153,
		0.99947,
		0.99992,
		0.999998,
		1,
		0.999956,
		0.999651,
		0.998722,
		0.996742,
		0.993282,
		0.987949,
		0.980411,
		0.970403,
		0.957735,
		0.94229,
		0.924016,
		0.90292,
		0.879064,
		0.852555,
		0.823539,
		0.792195,
		0.758729,
		0.723368,
		0.686356,
		0.647954,
		0.608428,
		0.568056,
		0.527118,
		0.485897,
		0.444677,
		0.403741,
		0.363368,
		0.323837,
		0.285421,
		0.248386,
		0.212995,
		0.179502,
		0.148158,
		0.119202,
		0.09287,
		0.069387,
		0.048971,
		0.031834,
		0.018178,
		0.008197,
		0.002078,
		0.010171,
		0.038051,
		0.080003,
		0.132788,
		0.193529,
		0.259691,
		0.32905,
		0.399673,
		0.469901,
		0.538323,
		0.603767,
		0.665277,
		0.722107,
		0.773702,
		0.819691,
		0.859876,
		0.89422,
		0.92284,
		0.945995,
		0.964076,
		0.977591,
		0.987147,
		0.993424,
		0.997147,
		0.999046,
		0.999801,
		0.999987,
		1,
		0.999988,
		0.999819,
		0.999166,
		0.997607,
		0.994696,
		0.990011,
		0.983187,
		0.973927,
		0.962012,
		0.947296,
		0.929706,
		0.909232,
		0.885923,
		0.859876,
		0.83123,
		0.800163,
		0.766878,
		0.731605,
		0.694592,
		0.656102,
		0.61641,
		0.575799,
		0.534555,
		0.492972,
		0.451339,
		0.409949,
		0.369091,
		0.32905,
		0.290107,
		0.25254,
		0.216617,
		0.182603,
		0.150754,
		0.121319,
		0.094539,
		0.070648,
		0.049871,
		0.032424,
		0.018518,
		0.008351,
		0.002117,
		0.001204,
		0.018011,
		0.051539,
		0.098193,
		0.154777,
		0.218463,
		0.286766,
		0.357518,
		0.428844,
		0.499142,
		0.567067,
		0.631507,
		0.691577,
		0.746597,
		0.796084,
		0.839741,
		0.877444,
		0.909232,
		0.935301,
		0.955987,
		0.971756,
		0.983187,
		0.990947,
		0.995768,
		0.998403,
		0.999581,
		0.999949,
		1,
		0.999998,
		0.99992,
		0.999493,
		0.99831,
		0.995914,
		0.991857,
		0.985742,
		0.977237,
		0.96609,
		0.952128,
		0.935251,
		0.915432,
		0.892703,
		0.867151,
		0.838908,
		0.808146,
		0.77507,
		0.739908,
		0.702914,
		0.664354,
		0.624508,
		0.583667,
		0.542124,
		0.50018,
		0.458135,
		0.416289,
		0.37494,
		0.334382,
		0.294906,
		0.256796,
		0.220331,
		0.185784,
		0.153419,
		0.123492,
		0.096253,
		0.071944,
		0.050795,
		0.033031,
		0.018867,
		0.00851,
		0.002158,
		0.004799,
		0.028021,
		0.066963,
		0.118071,
		0.178197,
		0.244566,
		0.314747,
		0.38663,
		0.4584,
		0.528519,
		0.595706,
		0.65892,
		0.717343,
		0.770369,
		0.817592,
		0.858788,
		0.893914,
		0.92309,
		0.946591,
		0.964833,
		0.978359,
		0.987816,
		0.99393,
		0.997473,
		0.999212,
		0.999858,
		0.999994,
		1,
		0.999973,
		0.99972,
		0.998863,
		0.996943,
		0.99349,
		0.988073,
		0.980326,
		0.969962,
		0.956775,
		0.94064,
		0.921508,
		0.899393,
		0.874369,
		0.846562,
		0.816138,
		0.783299,
		0.748275,
		0.71132,
		0.672708,
		0.632722,
		0.591661,
		0.549827,
		0.507526,
		0.465069,
		0.422765,
		0.380919,
		0.339837,
		0.299819,
		0.261156,
		0.224139,
		0.189047,
		0.156153,
		0.125723,
		0.098015,
		0.073275,
		0.051745,
		0.033655,
		0.019227,
		0.008674,
		0.0022,
		0.010757,
		0.040162,
		0.084269,
		0.139573,
		0.202976,
		0.271757,
		0.343545,
		0.416289,
		0.488241,
		0.557927,
		0.624138,
		0.685901,
		0.742476,
		0.793332,
		0.838141,
		0.876762,
		0.909232,
		0.935757,
		0.956695,
		0.972543,
		0.983919,
		0.991538,
		0.996177,
		0.998635,
		0.999678,
		0.99997,
		1,
		0.999994,
		0.999864,
		0.999281,
		0.997793,
		0.994912,
		0.990178,
		0.983187,
		0.973615,
		0.961225,
		0.94586,
		0.927446,
		0.90598,
		0.881522,
		0.854185,
		0.824131,
		0.791559,
		0.756699,
		0.719808,
		0.681162,
		0.641052,
		0.599783,
		0.557664,
		0.515011,
		0.472143,
		0.429378,
		0.387032,
		0.34542,
		0.30485,
		0.265625,
		0.228043,
		0.192394,
		0.15896,
		0.128015,
		0.099824,
		0.074644,
		0.052722,
		0.034297,
		0.019597,
		0.008842,
		0.002243,
		0.001275,
		0.019046,
		0.054388,
		0.103402,
		0.162631,
		0.229035,
		0.299949,
		0.373065,
		0.446396,
		0.518262,
		0.587261,
		0.652255,
		0.712348,
		0.766878,
		0.815394,
		0.857652,
		0.893596,
		0.92335,
		0.947206,
		0.965609,
		0.979138,
		0.988485,
		0.994427,
		0.997782,
		0.999361,
		0.999903,
		0.999998,
		1,
		0.999946,
		0.999581,
		0.998474,
		0.996128,
		0.992055,
		0.985814,
		0.977043,
		0.965467,
		0.950898,
		0.933236,
		0.912454,
		0.888597,
		0.861767,
		0.832117,
		0.799844,
		0.765177,
		0.728373,
		0.689715,
		0.649498,
		0.608032,
		0.565638,
		0.522638,
		0.47936,
		0.436133,
		0.393283,
		0.351133,
		0.310002,
		0.270205,
		0.232048,
		0.19583,
		0.161842,
		0.130369,
		0.101684,
		0.076051,
		0.053727,
		0.034957,
		0.019978,
		0.009016,
		0.002287,
		0.005084,
		0.029626,
		0.070648,
		0.124297,
		0.187172,
		0.25629,
		0.32905,
		0.403207,
		0.476846,
		0.548355,
		0.61641,
		0.679948,
		0.738155,
		0.790449,
		0.836467,
		0.87605,
		0.909232,
		0.936229,
		0.957423,
		0.973345,
		0.984657,
		0.992122,
		0.996571,
		0.99885,
		0.99976,
		0.999984,
		1,
		0.999985,
		0.999782,
		0.999001,
		0.997147,
		0.993706,
		0.988203,
		0.980236,
		0.969489,
		0.955742,
		0.938862,
		0.9188,
		0.895582,
		0.869296,
		0.840087,
		0.808146,
		0.773702,
		0.737013,
		0.698364,
		0.658057,
		0.61641,
		0.573749,
		0.530408,
		0.486723,
		0.443032,
		0.399673,
		0.35698,
		0.315281,
		0.2749,
		0.236155,
		0.199356,
		0.164802,
		0.132788,
		0.103595,
		0.077498,
		0.054761,
		0.035637,
		0.02037,
		0.009194,
		0.002333,
		0.011395,
		0.042454,
		0.088884,
		0.146884,
		0.213114,
		0.28465,
		0.358959,
		0.433865,
		0.507526,
		0.578407,
		0.645261,
		0.707107,
		0.763216,
		0.813093,
		0.856465,
		0.893264,
		0.92362,
		0.947842,
		0.966404,
		0.979927,
		0.989154,
		0.994912,
		0.998073,
		0.999493,
		0.999938,
		1,
		0.999998,
		0.999903,
		0.999391,
		0.997978,
		0.995135,
		0.990352,
		0.983187,
		0.973282,
		0.960379,
		0.944313,
		0.925006,
		0.902465,
		0.876762,
		0.848031,
		0.816458,
		0.782268,
		0.745722,
		0.707107,
		0.66673,
		0.624916,
		0.582,
		0.538323,
		0.494234,
		0.45008,
		0.406208,
		0.362964,
		0.320688,
		0.279714,
		0.24037,
		0.202976,
		0.167843,
		0.135274,
		0.105561,
		0.078987,
		0.055825,
		0.036336,
		0.020774,
		0.009378,
		0.00238,
		0.001354,
		0.020173,
		0.057479,
		0.109032,
		0.171089,
		0.24037,
		0.314019,
		0.389573,
		0.464929,
		0.538323,
		0.608301,
		0.673697,
		0.733619,
		0.787425,
		0.834714,
		0.875306,
		0.909232,
		0.936718,
		0.958172,
		0.974162,
		0.985399,
		0.9927,
		0.99695,
		0.999046,
		0.999827,
		0.999993,
		1,
		0.999967,
		0.999662,
		0.998635,
		0.996348,
		0.992261,
		0.98589,
		0.976835,
		0.964797,
		0.949573,
		0.931058,
		0.909232,
		0.884152,
		0.855939,
		0.82477,
		0.790869,
		0.754496,
		0.71594,
		0.675515,
		0.63355,
		0.59039,
		0.546387,
		0.501896,
		0.457278,
		0.412891,
		0.369091,
		0.326228,
		0.28465,
		0.244695,
		0.206693,
		0.170967,
		0.137829,
		0.107582,
		0.080519,
		0.05692,
		0.037057,
		0.021189,
		0.009567,
		0.002428,
		0.005396,
		0.031373,
		0.074644,
		0.131022,
		0.196828,
		0.268846,
		0.344293,
		0.420779,
		0.496282,
		0.569117,
		0.637915,
		0.701601,
		0.759371,
		0.81068,
		0.855222,
		0.892918,
		0.9239,
		0.948498,
		0.967218,
		0.980727,
		0.989821,
		0.995385,
		0.998347,
		0.999608,
		0.999963,
		1,
		0.999993,
		0.999836,
		0.999134,
		0.997353,
		0.99393,
		0.98834,
		0.98014,
		0.968983,
		0.95463,
		0.936941,
		0.91587,
		0.891454,
		0.863799,
		0.833075,
		0.799498,
		0.763329,
		0.72486,
		0.684409,
		0.642312,
		0.598922,
		0.5546,
		0.509713,
		0.464632,
		0.419726,
		0.375363,
		0.331906,
		0.289713,
		0.249134,
		0.210511,
		0.174178,
		0.140457,
		0.109662,
		0.082095,
		0.058047,
		0.037798,
		0.021618,
		0.009762,
		0.002478,
		0.012092,
		0.044948,
		0.093885,
		0.154777,
		0.22401,
		0.298442,
		0.375363,
		0.452462,
		0.527802,
		0.599783,
		0.667127,
		0.728851,
		0.78425,
		0.832877,
		0.874529,
		0.909232,
		0.937225,
		0.958943,
		0.974996,
		0.986146,
		0.99327,
		0.997312,
		0.999223,
		0.999882,
		0.999998,
		1,
		0.999935,
		0.999493,
		0.998163,
		0.995364,
		0.990536,
		0.983187,
		0.972926,
		0.959469,
		0.942641,
		0.922365,
		0.898654,
		0.8716,
		0.84136,
		0.808146,
		0.772216,
		0.733863,
		0.69341,
		0.651201,
		0.607595,
		0.562964,
		0.517687,
		0.472143,
		0.426716,
		0.381784,
		0.337724,
		0.294906,
		0.25369,
		0.214433,
		0.177478,
		0.14316,
		0.111802,
		0.083718,
		0.059208,
		0.038563,
		0.022059,
		0.009964,
		0.00253,
		0.00144,
		0.021402,
		0.06084,
		0.11513,
		0.180211,
		0.25254,
		0.32905,
		0.407111,
		0.484498,
		0.559361,
		0.630193,
		0.69581,
		0.755329,
		0.808146,
		0.85392,
		0.892557,
		0.924192,
		0.949176,
		0.968053,
		0.981538,
		0.990485,
		0.995843,
		0.9986,
		0.999707,
		0.999981,
		1,
		0.999982,
		0.999735,
		0.998793,
		0.996571,
		0.992476,
		0.98597,
		0.976614,
		0.964076,
		0.948142,
		0.9287,
		0.905738,
		0.879329,
		0.849616,
		0.816804,
		0.781149,
		0.742944,
		0.702516,
		0.660215,
		0.61641,
		0.571482,
		0.52582,
		0.479816,
		0.433865,
		0.38836,
		0.343689,
		0.300233,
		0.258369,
		0.218463,
		0.180871,
		0.14594,
		0.114005,
		0.085389,
		0.060404,
		0.03935,
		0.022514,
		0.010171,
		0.002583,
		0.005737,
		0.033279,
		0.078987,
		0.138302,
		0.207232,
		0.282312,
		0.360554,
		0.439414,
		0.516759,
		0.590828,
		0.660215,
		0.723836,
		0.780913,
		0.830949,
		0.873716,
		0.909232,
		0.937751,
		0.959736,
		0.975845,
		0.986896,
		0.993829,
		0.997655,
		0.999381,
		0.999924,
		0.999999,
		0.999997,
		0.999882,
		0.999261,
		0.997561,
		0.994162,
		0.988485,
		0.980037,
		0.968439,
		0.953429,
		0.93486,
		0.912691,
		0.886971,
		0.85783,
		0.825463,
		0.79012,
		0.752097,
		0.711723,
		0.669354,
		0.625367,
		0.580154,
		0.534114,
		0.487653,
		0.441178,
		0.395094,
		0.349803,
		0.3057,
		0.263174,
		0.222604,
		0.18436,
		0.148801,
		0.116272,
		0.08711,
		0.061637,
		0.040162,
		0.022984,
		0.010385,
		0.002638,
		0.012854,
		0.047667,
		0.099316,
		0.163312,
		0.23574,
		0.313213,
		0.392831,
		0.472143,
		0.549108,
		0.622067,
		0.689715,
		0.751077,
		0.805484,
		0.852555,
		0.892179,
		0.924495,
		0.949876,
		0.968908,
		0.982358,
		0.991144,
		0.996286,
		0.998834,
		0.999788,
		0.999991,
		1,
		0.999959,
		0.999588,
		0.998347,
		0.995601,
		0.990728,
		0.983187,
		0.972543,
		0.958486,
		0.94083,
		0.919497,
		0.894513,
		0.86599,
		0.834112,
		0.799123,
		0.761317,
		0.721027,
		0.678615,
		0.634466,
		0.588982,
		0.542574,
		0.495659,
		0.448658,
		0.40199,
		0.356071,
		0.31131,
		0.268109,
		0.226862,
		0.18795,
		0.151745,
		0.118608,
		0.088884,
		0.062907,
		0.041,
		0.023468,
		0.010606,
		0.002694,
		0.001534,
		0.022747,
		0.064503,
		0.121749,
		0.190067,
		0.265625,
		0.345123,
		0.425753,
		0.505161,
		0.581406,
		0.652936,
		0.718555,
		0.777401,
		0.828924,
		0.872864,
		0.909232,
		0.938296,
		0.960553,
		0.976711,
		0.987648,
		0.994377,
		0.997978,
		0.999519,
		0.999955,
		1,
		0.999991,
		0.999799,
		0.998946,
		0.996799,
		0.9927,
		0.986055,
		0.976376,
		0.963299,
		0.946591,
		0.926139,
		0.901939,
		0.874082,
		0.842738,
		0.808146,
		0.770596,
		0.730424,
		0.687996,
		0.643707,
		0.597967,
		0.5512,
		0.503836,
		0.45631,
		0.409054,
		0.362499,
		0.317068,
		0.273179,
		0.231239,
		0.191643,
		0.154777,
		0.121013,
		0.090711,
		0.064217,
		0.041863,
		0.023968,
		0.010834,
		0.002753,
		0.006111,
		0.035363,
		0.083718,
		0.146196,
		0.218463,
		0.29677,
		0.377913,
		0.459182,
		0.538323,
		0.613507,
		0.683291,
		0.746597,
		0.802682,
		0.851122,
		0.891784,
		0.924811,
		0.950601,
		0.969784,
		0.983187,
		0.991798,
		0.996711,
		0.999046,
		0.999854,
		0.999997,
		0.999999,
		0.99992,
		0.999381,
		0.997769,
		0.994403,
		0.988638,
		0.979927,
		0.967852,
		0.952128,
		0.9326,
		0.909232,
		0.882091,
		0.851331,
		0.817181,
		0.779928,
		0.739908,
		0.697495,
		0.653088,
		0.607109,
		0.559995,
		0.512188,
		0.464137,
		0.416289,
		0.369091,
		0.32298,
		0.278389,
		0.23574,
		0.195443,
		0.157899,
		0.123492,
		0.092596,
		0.065569,
		0.042755,
		0.024483,
		0.011069,
		0.002813,
		0.013691,
		0.050639,
		0.105229,
		0.172561,
		0.248386,
		0.32905,
		0.411446,
		0.492972,
		0.571482,
		0.645261,
		0.712986,
		0.773702,
		0.826794,
		0.87197,
		0.909232,
		0.938862,
		0.961395,
		0.977591,
		0.988401,
		0.994912,
		0.99828,
		0.999638,
		0.999976,
		1,
		0.999977,
		0.999675,
		0.998528,
		0.995843,
		0.990931,
		0.983187,
		0.97213,
		0.957423,
		0.938862,
		0.916375,
		0.890001,
		0.859876,
		0.826216,
		0.789303,
		0.749475,
		0.707107,
		0.662608,
		0.61641,
		0.568961,
		0.520718,
		0.472143,
		0.423701,
		0.375851,
		0.32905,
		0.283743,
		0.24037,
		0.199356,
		0.161115,
		0.126047,
		0.094539,
		0.066963,
		0.043674,
		0.025016,
		0.011312,
		0.002875,
		0.001637,
		0.024222,
		0.068505,
		0.128948,
		0.200736,
		0.279714,
		0.362328,
		0.445578,
		0.526971,
		0.604481,
		0.676513,
		0.741872,
		0.799732,
		0.849616,
		0.89137,
		0.92514,
		0.951351,
		0.970681,
		0.984024,
		0.992444,
		0.997117,
		0.999236,
		0.999906,
		0.999999,
		0.999997,
		0.999854,
		0.999094,
		0.99703,
		0.992935,
		0.986146,
		0.97612,
		0.962458,
		0.944907,
		0.92335,
		0.897796,
		0.868358,
		0.835238,
		0.798713,
		0.759116,
		0.716828,
		0.672265,
		0.625869,
		0.5781,
		0.529428,
		0.480333,
		0.431293,
		0.382786,
		0.335282,
		0.289247,
		0.245133,
		0.203384,
		0.164428,
		0.128681,
		0.096544,
		0.068402,
		0.044624,
		0.025566,
		0.011564,
		0.00294,
		0.006522,
		0.037644,
		0.088879,
		0.154774,
		0.230604,
		0.312315,
		0.39646,
		0.480153,
		0.561021,
		0.63716,
		0.707107,
		0.769799,
		0.824552,
		0.871032,
		0.909232,
		0.939451,
		0.962262,
		0.978488,
		0.989154,
		0.995431,
		0.99856,
		0.999738,
		0.999989,
		1,
		0.99995,
		0.999493,
		0.997978,
		0.994653,
		0.9888,
		0.97981,
		0.967218,
		0.950715,
		0.930138,
		0.905457,
		0.876762,
		0.844236,
		0.808146,
		0.768825,
		0.726654,
		0.682058,
		0.635486,
		0.587413,
		0.538323,
		0.488711,
		0.439072,
		0.3899,
		0.341684,
		0.294906,
		0.250035,
		0.207533,
		0.167843,
		0.131398,
		0.098612,
		0.069887,
		0.045605,
		0.026134,
		0.011823,
		0.003007,
		0.014606,
		0.053884,
		0.111669,
		0.182595,
		0.262037,
		0.346046,
		0.431293,
		0.515011,
		0.594955,
		0.669354,
		0.736881,
		0.796619,
		0.848031,
		0.890936,
		0.925483,
		0.952128,
		0.971601,
		0.984868,
		0.993081,
		0.997501,
		0.999403,
		0.999944,
		1,
		0.999989,
		0.999752,
		0.998706,
		0.996092,
		0.991144,
		0.983187,
		0.971686,
		0.956268,
		0.936718,
		0.912966,
		0.88507,
		0.853193,
		0.817592,
		0.778592,
		0.736579,
		0.691982,
		0.645261,
		0.596902,
		0.547405,
		0.497281,
		0.447041,
		0.397198,
		0.34826,
		0.300724,
		0.255081,
		0.211807,
		0.171364,
		0.1342,
		0.100747,
		0.071421,
		0.046619,
		0.026722,
		0.012092,
		0.003075,
		0.001749,
		0.025825,
		0.072856,
		0.136765,
		0.212284,
		0.294891,
		0.380753,
		0.466662,
		0.549981,
		0.628601,
		0.700893,
		0.765677,
		0.822187,
		0.870045,
		0.909232,
		0.940063,
		0.963156,
		0.9794,
		0.989904,
		0.995933,
		0.998815,
		0.999819,
		0.999996,
		0.999999,
		0.9999,
		0.999236,
		0.997265,
		0.993181,
		0.986243,
		0.975845,
		0.961546,
		0.943071,
		0.920303,
		0.893264,
		0.862095,
		0.827036,
		0.788409,
		0.746597,
		0.702034,
		0.655192,
		0.606567,
		0.556677,
		0.506046,
		0.455206,
		0.404687,
		0.355015,
		0.306709,
		0.260276,
		0.21621,
		0.174994,
		0.137092,
		0.102952,
		0.073006,
		0.047667,
		0.027329,
		0.01237,
		0.003147,
		0.006962,
		0.040101,
		0.094456,
		0.164044,
		0.243699,
		0.329012,
		0.416269,
		0.502393,
		0.584885,
		0.661781,
		0.731605,
		0.793332,
		0.846361,
		0.89048,
		0.925841,
		0.952932,
		0.972543,
		0.985719,
		0.993706,
		0.997863,
		0.999548,
		0.99997,
		1,
		0.999972,
		0.999596,
		0.998186,
		0.994912,
		0.988971,
		0.979683,
		0.966531,
		0.949176,
		0.927446,
		0.901325,
		0.870925,
		0.836467,
		0.798264,
		0.756699,
		0.71221,
		0.665277,
		0.61641,
		0.56614,
		0.515011,
		0.463572,
		0.412372,
		0.361957,
		0.312866,
		0.265625,
		0.22075,
		0.178739,
		0.140077,
		0.105229,
		0.074644,
		0.04875,
		0.027957,
		0.012657,
		0.003221,
		0.015567,
		0.057334,
		0.118558,
		0.19336,
		0.276679,
		0.364225,
		0.452411,
		0.538297,
		0.619534,
		0.694312,
		0.761316,
		0.819691,
		0.869007,
		0.909232,
		0.940699,
		0.964076,
		0.980326,
		0.99065,
		0.996415,
		0.999046,
		0.999882,
		0.999999,
		0.999996,
		0.999819,
		0.998879,
		0.996348,
		0.991369,
		0.983187,
		0.971204,
		0.95501,
		0.934374,
		0.909232,
		0.879663,
		0.845868,
		0.808146,
		0.766878,
		0.722505,
		0.675515,
		0.62643,
		0.575799,
		0.52418,
		0.472143,
		0.420258,
		0.369091,
		0.3192,
		0.271135,
		0.225429,
		0.182603,
		0.14316,
		0.107582,
		0.076337,
		0.049871,
		0.028608,
		0.012955,
		0.003297,
		0.001862,
		0.027471,
		0.077397,
		0.14501,
		0.224539,
		0.311037,
		0.400334,
		0.488978,
		0.574175,
		0.653732,
		0.726004,
		0.789852,
		0.844598,
		0.890001,
		0.926215,
		0.953766,
		0.973507,
		0.986574,
		0.994317,
		0.9982,
		0.999671,
		0.999986,
		1,
		0.999937,
		0.999369,
		0.997501,
		0.993437,
		0.986346,
		0.975548,
		0.960553,
		0.941064,
		0.916963,
		0.888291,
		0.855222,
		0.818041,
		0.777124,
		0.732912,
		0.685901,
		0.636628,
		0.585653,
		0.533556,
		0.480925,
		0.428351,
		0.376422,
		0.325719,
		0.276811,
		0.230255,
		0.186592,
		0.146343,
		0.110014,
		0.078089,
		0.05103,
		0.029281,
		0.013263,
		0.003376,
		0.007387,
		0.042546,
		0.100129,
		0.173627,
		0.257379,
		0.346557,
		0.437117,
		0.525747,
		0.609807,
		0.687268,
		0.756663,
		0.817037,
		0.867908,
		0.909231,
		0.941362,
		0.965026,
		0.981267,
		0.99139,
		0.996876,
		0.999251,
		0.999929,
		1,
		0.999986,
		0.999689,
		0.998392,
		0.99518,
		0.989154,
		0.979547,
		0.965784,
		0.947493,
		0.924495,
		0.896786,
		0.864511,
		0.827935,
		0.787425,
		0.743424,
		0.696434,
		0.647002,
		0.595706,
		0.543144,
		0.489923,
		0.436658,
		0.383959,
		0.332429,
		0.282661,
		0.235233,
		0.19071,
		0.149633,
		0.112529,
		0.079901,
		0.05223,
		0.029978,
		0.013582,
		0.003458,
		0.01645,
		0.060626,
		0.125331,
		0.204182,
		0.291638,
		0.382989,
		0.474322,
		0.56247,
		0.644953,
		0.719918,
		0.786085,
		0.842698,
		0.889481,
		0.926602,
		0.954629,
		0.974494,
		0.987433,
		0.994912,
		0.998511,
		0.999771,
		0.999995,
		0.999999,
		0.999875,
		0.999046,
		0.996609,
		0.991607,
		0.983187,
		0.970681,
		0.953636,
		0.931803,
		0.905127,
		0.873716,
		0.837812,
		0.797771,
		0.754033,
		0.707107,
		0.657551,
		0.605958,
		0.552945,
		0.499142,
		0.445185,
		0.391707,
		0.339336,
		0.28869,
		0.24037,
		0.194962,
		0.153034,
		0.11513,
		0.081776,
		0.053473,
		0.0307,
		0.013913,
		0.003544,
		0.001953,
		0.028884,
		0.081496,
		0.152754,
		0.236406,
		0.327031,
		0.420042,
		0.51166,
		0.598868,
		0.679351,
		0.751442,
		0.81406,
		0.86666,
		0.909189,
		0.942036,
		0.965998,
		0.982219,
		0.992122,
		0.997312,
		0.999429,
		0.999962,
		1,
		0.999964,
		0.999493,
		0.99774,
		0.993706,
		0.986456,
		0.975226,
		0.959469,
		0.938862,
		0.913289,
		0.882814,
		0.847654,
		0.808146,
		0.764729,
		0.717915,
		0.668272,
		0.61641,
		0.562964,
		0.508587,
		0.453937,
		0.399673,
		0.346449,
		0.294906,
		0.245671,
		0.199356,
		0.156549,
		0.117821,
		0.083718,
		0.054761,
		0.031449,
		0.014256,
		0.003632,
		0.007692,
		0.044463,
		0.104887,
		0.182095,
		0.269972,
		0.36322,
		0.457378,
		0.548801,
		0.634625,
		0.712711,
		0.781582,
		0.840368,
		0.888751,
		0.926912,
		0.955481,
		0.975486,
		0.988288,
		0.995487,
		0.998793,
		0.99985,
		0.999999,
		0.999995,
		0.999771,
		0.998595,
		0.995458,
		0.989348,
		0.9794,
		0.964969,
		0.945646,
		0.921246,
		0.891784,
		0.85744,
		0.818536,
		0.775502,
		0.728851,
		0.679164,
		0.627063,
		0.573204,
		0.518262,
		0.462921,
		0.407865,
		0.353774,
		0.301315,
		0.251144,
		0.203896,
		0.160186,
		0.120607,
		0.085729,
		0.056095,
		0.032226,
		0.014613,
		0.003724,
		0.016988,
		0.062899,
		0.130464,
		0.212988,
		0.304496,
		0.399815,
		0.494607,
		0.585368,
		0.66939,
		0.744711,
		0.810056,
		0.864779,
		0.908799,
		0.942546,
		0.966903,
		0.983139,
		0.992826,
		0.997716,
		0.99958,
		0.999982,
		1,
		0.99992,
		0.999205,
		0.996876,
		0.991857,
		0.983187,
		0.970112,
		0.952128,
		0.928972,
		0.900598,
		0.867151,
		0.828924,
		0.786339,
		0.739908,
		0.690221,
		0.637915,
		0.583667,
		0.528173,
		0.472143,
		0.416289,
		0.361318,
		0.307926,
		0.256796,
		0.208589,
		0.163949,
		0.123492,
		0.087813,
		0.057479,
		0.033031,
		0.014982,
		0.003819,
		0.001983,
		0.029542,
		0.083836,
		0.157829,
		0.245006,
		0.339533,
		0.436364,
		0.531288,
		0.620939,
		0.702768,
		0.774998,
		0.836564,
		0.887051,
		0.926633,
		0.956001,
		0.9763,
		0.989042,
		0.995994,
		0.999028,
		0.999904,
		0.999998,
		0.999982,
		0.999606,
		0.997978,
		0.993987,
		0.986574,
		0.974876,
		0.95828,
		0.936437,
		0.909232,
		0.876762,
		0.839291,
		0.797225,
		0.751077,
		0.701439,
		0.648967,
		0.594354,
		0.538323,
		0.48161,
		0.424954,
		0.369091,
		0.314747,
		0.262634,
		0.213443,
		0.167843,
		0.12648,
		0.089974,
		0.058915,
		0.033867,
		0.015366,
		0.003918,
		0.007716,
		0.04498,
		0.106843,
		0.186511,
		0.277651,
		0.374572,
		0.472358,
		0.566923,
		0.655041,
		0.734326,
		0.803188,
		0.860781,
		0.906936,
		0.942087,
		0.967203,
		0.983695,
		0.99331,
		0.997988,
		0.999659,
		0.999976,
		0.999993,
		0.999839,
		0.998793,
		0.995745,
		0.989554,
		0.97924,
		0.964076,
		0.943612,
		0.917657,
		0.886248,
		0.849616,
		0.808146,
		0.762345,
		0.712812,
		0.660215,
		0.605268,
		0.548718,
		0.491327,
		0.433865,
		0.377099,
		0.321785,
		0.268666,
		0.218463,
		0.171875,
		0.129577,
		0.092214,
		0.060404,
		0.034735,
		0.015765,
		0.00402,
		0.016812,
		0.062848,
		0.131419,
		0.215973,
		0.310362,
		0.409045,
		0.507231,
		0.600964,
		0.687158,
		0.76359,
		0.828867,
		0.882368,
		0.924176,
		0.955004,
		0.9761,
		0.989142,
		0.996096,
		0.999026,
		0.999837,
		0.999952,
		0.999936,
		0.999349,
		0.997146,
		0.992122,
		0.983187,
		0.969489,
		0.950468,
		0.925841,
		0.895582,
		0.859876,
		0.819083,
		0.773702,
		0.724333,
		0.671658,
		0.61641,
		0.559361,
		0.501302,
		0.443032,
		0.385352,
		0.32905,
		0.2749,
		0.223658,
		0.176052,
		0.132788,
		0.094539,
		0.061951,
		0.035637,
		0.016179,
		0.004127,
		0.00191,
		0.028801,
		0.08262,
		0.156993,
		0.245614,
		0.342523,
		0.442328,
		0.540369,
		0.632819,
		0.716734,
		0.790059,
		0.8516,
		0.900971,
		0.938519,
		0.965238,
		0.982663,
		0.992725,
		0.997575,
		0.99933,
		0.999756,
		0.99988,
		0.999659,
		0.998198,
		0.994276,
		0.9867,
		0.974494,
		0.956972,
		0.933754,
		0.904734,
		0.870045,
		0.830017,
		0.785132,
		0.735993,
		0.683291,
		0.62778,
		0.570256,
		0.511539,
		0.452462,
		0.393858,
		0.336549,
		0.281345,
		0.229035,
		0.18038,
		0.136118,
		0.096952,
		0.063557,
		0.036574,
		0.01661,
		0.004238,
		0.007301,
		0.043137,
		0.103703,
		0.182939,
		0.274783,
		0.373469,
		0.473757,
		0.57112,
		0.661863,
		0.743188,
		0.81321,
		0.870942,
		0.916238,
		0.949718,
		0.972665,
		0.986889,
		0.994554,
		0.997934,
		0.9991,
		0.999523,
		0.999655,
		0.998871,
		0.995994,
		0.989757,
		0.979061,
		0.963094,
		0.941362,
		0.913673,
		0.880097,
		0.840925,
		0.796619,
		0.747781,
		0.695109,
		0.639377,
		0.581406,
		0.522046,
		0.462163,
		0.402626,
		0.344293,
		0.288011,
		0.234603,
		0.184867,
		0.139573,
		0.099458,
		0.065227,
		0.037549,
		0.017058,
		0.004354,
		0.015609,
		0.05921,
		0.125451,
		0.208583,
		0.302792,
		0.402506,
		0.502643,
		0.598814,
		0.687459,
		0.765928,
		0.832516,
		0.886448,
		0.927827,
		0.95755,
		0.977186,
		0.988808,
		0.994761,
		0.997356,
		0.998474,
		0.9991,
		0.999017,
		0.997181,
		0.99229,
		0.983139,
		0.968789,
		0.948626,
		0.922363,
		0.890001,
		0.851782,
		0.808146,
		0.759685,
		0.707107,
		0.651201,
		0.592813,
		0.532828,
		0.472143,
		0.411665,
		0.352291,
		0.294906,
		0.24037,
		0.18952,
		0.14316,
		0.102062,
		0.066963,
		0.038563,
		0.017525,
		0.004475,
		0.001708,
		0.026203,
		0.076359,
		0.147183,
		0.23323,
		0.328938,
		0.428931,
		0.528295,
		0.622781,
		0.708972,
		0.784374,
		0.847468,
		0.897699,
		0.935426,
		0.961826,
		0.978745,
		0.988479,
		0.993485,
		0.995982,
		0.997473,
		0.998268,
		0.997575,
		0.994116,
		0.986599,
		0.973965,
		0.955481,
		0.930757,
		0.89972,
		0.862561,
		0.819691,
		0.771692,
		0.719275,
		0.663248,
		0.604481,
		0.54389,
		0.48241,
		0.420985,
		0.360554,
		0.30204,
		0.246346,
		0.194346,
		0.146884,
		0.104768,
		0.068768,
		0.039618,
		0.018011,
		0.0046,
		0.006383,
		0.038407,
		0.093895,
		0.168202,
		0.256183,
		0.352525,
		0.45206,
		0.550042,
		0.642378,
		0.7258,
		0.797981,
		0.857596,
		0.904319,
		0.938773,
		0.96241,
		0.977327,
		0.985989,
		0.99085,
		0.993865,
		0.99593,
		0.996642,
		0.994828,
		0.989142,
		0.97841,
		0.961779,
		0.938756,
		0.909189,
		0.873218,
		0.831226,
		0.783783,
		0.731605,
		0.675515,
		0.61641,
		0.555238,
		0.492972,
		0.430596,
		0.369091,
		0.309424,
		0.25254,
		0.199356,
		0.150754,
		0.107582,
		0.070648,
		0.040718,
		0.018518,
		0.004731,
		0.013323,
		0.051521,
		0.111127,
		0.187825,
		0.276767,
		0.372893,
		0.471247,
		0.567267,
		0.65702,
		0.737402,
		0.806267,
		0.862495,
		0.906002,
		0.937675,
		0.95923,
		0.97297,
		0.981438,
		0.986929,
		0.990906,
		0.993485,
		0.993699,
		0.990193,
		0.981687,
		0.967203,
		0.94614,
		0.918264,
		0.883665,
		0.842698,
		0.795928,
		0.74408,
		0.687995,
		0.628601,
		0.566876,
		0.503836,
		0.440506,
		0.377913,
		0.317068,
		0.258963,
		0.204557,
		0.154777,
		0.110511,
		0.072605,
		0.041863,
		0.019046,
		0.004868,
		0.001391,
		0.021807,
		0.064851,
		0.127383,
		0.205406,
		0.294361,
		0.389445,
		0.485925,
		0.579429,
		0.666205,
		0.743327,
		0.808844,
		0.861859,
		0.902538,
		0.932032,
		0.95229,
		0.965756,
		0.974912,
		0.981687,
		0.986797,
		0.989607,
		0.988808,
		0.983029,
		0.971156,
		0.952465,
		0.926633,
		0.893693,
		0.853978,
		0.808054,
		0.756663,
		0.700674,
		0.641049,
		0.578809,
		0.515011,
		0.450728,
		0.387032,
		0.324985,
		0.265625,
		0.209959,
		0.15896,
		0.113559,
		0.074644,
		0.043058,
		0.019597,
		0.00501,
		0.005056,
		0.031123,
		0.077731,
		0.142042,
		0.220363,
		0.308422,
		0.401677,
		0.495623,
		0.586099,
		0.669547,
		0.743241,
		0.805445,
		0.855502,
		0.893836,
		0.921848,
		0.941677,
		0.955795,
		0.966443,
		0.974912,
		0.981036,
		0.983637,
		0.981267,
		0.972665,
		0.956957,
		0.933706,
		0.902878,
		0.864779,
		0.819979,
		0.769247,
		0.713495,
		0.653732,
		0.591031,
		0.526502,
		0.461269,
		0.39646,
		0.333186,
		0.272538,
		0.215573,
		0.163312,
		0.116734,
		0.07677,
		0.044304,
		0.020173,
		0.005159,
		0.01025,
		0.040599,
		0.089552,
		0.154555,
		0.232205,
		0.318512,
		0.409197,
		0.5,
		0.586982,
		0.666804,
		0.736961,
		0.795954,
		0.843385,
		0.879941,
		0.907247,
		0.927549,
		0.943198,
		0.955943,
		0.966179,
		0.97297,
		0.974831,
		0.970337,
		0.95843,
		0.938519,
		0.910471,
		0.874547,
		0.831316,
		0.781582,
		0.726308,
		0.666569,
		0.603506,
		0.538297,
		0.472135,
		0.406206,
		0.341684,
		0.279714,
		0.221409,
		0.167843,
		0.120042,
		0.078987,
		0.045605,
		0.020774,
		0.005315,
		0.00101,
		0.01627,
		0.04962,
		0.099785,
		0.164471,
		0.240551,
		0.324316,
		0.411755,
		0.498862,
		0.581942,
		0.657896,
		0.724468,
		0.780424,
		0.825636,
		0.861048,
		0.888462,
		0.910115,
		0.928019,
		0.943133,
		0.954896,
		0.961852,
		0.96241,
		0.955287,
		0.939685,
		0.91532,
		0.882368,
		0.841376,
		0.793172,
		0.73878,
		0.679351,
		0.616116,
		0.550341,
		0.483303,
		0.416269,
		0.350485,
		0.287164,
		0.227479,
		0.172561,
		0.123492,
		0.081301,
		0.046964,
		0.021402,
		0.005478,
		0.00355,
		0.022482,
		0.057659,
		0.108001,
		0.171451,
		0.24515,
		0.325662,
		0.40925,
		0.492178,
		0.571013,
		0.642924,
		0.705931,
		0.759087,
		0.80255,
		0.837497,
		0.865825,
		0.889587,
		0.910182,
		0.927549,
		0.940363,
		0.946897,
		0.945607,
		0.935426,
		0.915851,
		0.88691,
		0.849081,
		0.803188,
		0.750302,
		0.691658,
		0.628589,
		0.56247,
		0.494688,
		0.426612,
		0.359581,
		0.294891,
		0.233791,
		0.177477,
		0.127091,
		0.083718,
		0.048385,
		0.022059,
		0.005649,
		0.006942,
		0.02834,
		0.064291,
		0.113883,
		0.175283,
		0.245885,
		0.322525,
		0.401745,
		0.48009,
		0.554413,
		0.622176,
		0.681705,
		0.732363,
		0.774593,
		0.809757,
		0.839721,
		0.86612,
		0.889428,
		0.908605,
		0.921848,
		0.92735,
		0.923719,
		0.910151,
		0.886448,
		0.852944,
		0.810403,
		0.75991,
		0.702768,
		0.640413,
		0.574344,
		0.50608,
		0.437117,
		0.368915,
		0.302876,
		0.240343,
		0.182595,
		0.130845,
		0.086243,
		0.04987,
		0.022747,
		0.005827,
		0.000636,
		0.010604,
		0.033402,
		0.069201,
		0.11724,
		0.175889,
		0.242785,
		0.315035,
		0.389462,
		0.462908,
		0.532534,
		0.596121,
		0.652329,
		0.700845,
		0.742376,
		0.778389,
		0.810531,
		0.839721,
		0.865353,
		0.885676,
		0.898653,
		0.902538,
		0.896157,
		0.878988,
		0.851131,
		0.81321,
		0.766258,
		0.711604,
		0.650766,
		0.585368,
		0.517072,
		0.447525,
		0.378334,
		0.311037,
		0.247099,
		0.187904,
		0.134756,
		0.088879,
		0.051425,
		0.023468,
		0.006014,
		0.002137,
		0.014062,
		0.037339,
		0.072196,
		0.118003,
		0.17332,
		0.236018,
		0.303466,
		0.37278,
		0.441107,
		0.505937,
		0.565401,
		0.618508,
		0.665273,
		0.706621,
		0.743988,
		0.778552,
		0.810233,
		0.837497,
		0.858176,
		0.870185,
		0.871933,
		0.862495,
		0.841631,
		0.809723,
		0.767663,
		0.716734,
		0.658487,
		0.594644,
		0.527004,
		0.457378,
		0.38754,
		0.319197,
		0.253963,
		0.19336,
		0.138805,
		0.091624,
		0.053048,
		0.024222,
		0.006211,
		0.003985,
		0.016957,
		0.039939,
		0.073201,
		0.116229,
		0.167757,
		0.22588,
		0.288229,
		0.352213,
		0.415302,
		0.475331,
		0.530799,
		0.581082,
		0.626498,
		0.668102,
		0.707107,
		0.743918,
		0.777408,
		0.805429,
		0.825636,
		0.836021,
		0.835202,
		0.822512,
		0.797981,
		0.76225,
		0.716456,
		0.662108,
		0.600964,
		0.534931,
		0.465973,
		0.396042,
		0.327031,
		0.260738,
		0.198852,
		0.142941,
		0.094456,
		0.054735,
		0.02501,
		0.006416,
		0.000329,
		0.005777,
		0.019045,
		0.041104,
		0.072255,
		0.112084,
		0.159489,
		0.212781,
		0.269848,
		0.328396,
		0.386226,
		0.441538,
		0.493212,
		0.540983,
		0.585423,
		0.627576,
		0.66817,
		0.706621,
		0.740998,
		0.768831,
		0.787765,
		0.795954,
		0.792248,
		0.77623,
		0.748179,
		0.708972,
		0.659966,
		0.602874,
		0.539646,
		0.472358,
		0.403127,
		0.334039,
		0.267095,
		0.204182,
		0.147054,
		0.097323,
		0.056465,
		0.025825,
		0.00663,
		0.001038,
		0.007231,
		0.020197,
		0.040846,
		0.0695,
		0.105836,
		0.148905,
		0.197224,
		0.248937,
		0.302048,
		0.354699,
		0.405462,
		0.4536,
		0.499182,
		0.542935,
		0.585696,
		0.627422,
		0.666607,
		0.700845,
		0.727561,
		0.744506,
		0.750038,
		0.743241,
		0.723937,
		0.692627,
		0.650402,
		0.598814,
		0.539764,
		0.475373,
		0.407876,
		0.339533,
		0.272552,
		0.209038,
		0.15096,
		0.100129,
		0.058196,
		0.026654,
		0.006851,
		0.001801,
		0.008178,
		0.020382,
		0.039266,
		0.065163,
		0.097831,
		0.136466,
		0.179781,
		0.226175,
		0.273949,
		0.321588,
		0.368043,
		0.412943,
		0.456627,
		0.499836,
		0.542935,
		0.584989,
		0.623865,
		0.656947,
		0.681705,
		0.696065,
		0.698606,
		0.68864,
		0.666205,
		0.632006,
		0.58732,
		0.533882,
		0.473757,
		0.409225,
		0.34267,
		0.276483,
		0.212988,
		0.154387,
		0.102721,
		0.059855,
		0.027471,
		0.007072,
		0.000126,
		0.002407,
		0.008546,
		0.019662,
		0.036547,
		0.05954,
		0.088475,
		0.122681,
		0.161065,
		0.202269,
		0.244896,
		0.287772,
		0.330213,
		0.372177,
		0.414176,
		0.456759,
		0.499583,
		0.540983,
		0.578504,
		0.609506,
		0.631602,
		0.642924,
		0.642277,
		0.629185,
		0.603886,
		0.567267,
		0.520775,
		0.466305,
		0.406079,
		0.342523,
		0.278157,
		0.215498,
		0.156977,
		0.104887,
		0.061335,
		0.028234,
		0.007286,
		0.000356,
		0.002744,
		0.008349,
		0.018168,
		0.032928,
		0.052975,
		0.078208,
		0.108085,
		0.141698,
		0.177929,
		0.215674,
		0.254094,
		0.292845,
		0.332146,
		0.372533,
		0.414121,
		0.455837,
		0.495581,
		0.530799,
		0.558952,
		0.577846,
		0.585835,
		0.581942,
		0.565889,
		0.538097,
		0.499619,
		0.45206,
		0.397462,
		0.338191,
		0.276811,
		0.215973,
		0.158311,
		0.10636,
		0.062493,
		0.028884,
		0.00748,
		0.000544,
		0.002784,
		0.007669,
		0.016081,
		0.028687,
		0.045834,
		0.067482,
		0.09321,
		0.122288,
		0.153832,
		0.18702,
		0.22133,
		0.256712,
		0.293546,
		0.332194,
		0.372177,
		0.4119,
		0.449081,
		0.481216,
		0.505937,
		0.52127,
		0.525792,
		0.518731,
		0.5,
		0.470186,
		0.430498,
		0.382687,
		0.328938,
		0.271753,
		0.213831,
		0.157944,
		0.106843,
		0.06316,
		0.029348,
		0.007635,
		2.7e-005,
		0.000623,
		0.002558,
		0.006631,
		0.013614,
		0.024116,
		0.038483,
		0.056733,
		0.078559,
		0.103399,
		0.130593,
		0.159588,
		0.190146,
		0.222444,
		0.25688,
		0.293444,
		0.331119,
		0.368043,
		0.401925,
		0.430401,
		0.451312,
		0.462908,
		0.463986,
		0.453969,
		0.43295,
		0.401677,
		0.361507,
		0.314334,
		0.26248,
		0.208583,
		0.155471,
		0.106041,
		0.063155,
		0.029542,
		0.007731,
		6.1e-005,
		0.000588,
		0.002141,
		0.005384,
		0.010985,
		0.019499,
		0.031269,
		0.046365,
		0.064584,
		0.085527,
		0.108744,
		0.133926,
		0.161075,
		0.190498,
		0.222444,
		0.256446,
		0.291157,
		0.32464,
		0.354699,
		0.379157,
		0.39608,
		0.403941,
		0.401745,
		0.389109,
		0.366296,
		0.334211,
		0.294361,
		0.24878,
		0.199929,
		0.150584,
		0.103703,
		0.062312,
		0.029384,
		0.007746,
		6.8e-005,
		0.00047,
		0.001626,
		0.00408,
		0.008402,
		0.015095,
		0.024496,
		0.036721,
		0.051666,
		0.069081,
		0.088707,
		0.110452,
		0.134491,
		0.161142,
		0.190361,
		0.22133,
		0.252561,
		0.282165,
		0.308099,
		0.328396,
		0.341342,
		0.34563,
		0.34047,
		0.325662,
		0.30164,
		0.269468,
		0.230803,
		0.187825,
		0.143141,
		0.099666,
		0.060501,
		0.028801,
		0.007659,
		5.1e-005,
		0.000316,
		0.001107,
		0.002854,
		0.006041,
		0.011116,
		0.018409,
		0.028075,
		0.040099,
		0.054366,
		0.070788,
		0.089443,
		0.110603,
		0.134466,
		0.160683,
		0.188251,
		0.215674,
		0.241175,
		0.262895,
		0.279078,
		0.288229,
		0.289251,
		0.281552,
		0.265117,
		0.240551,
		0.209081,
		0.172521,
		0.133206,
		0.093895,
		0.057659,
		0.027745,
		0.007452,
		2.6e-005,
		0.000173,
		0.000658,
		0.001811,
		0.004032,
		0.007717,
		0.013182,
		0.020615,
		0.03008,
		0.041577,
		0.05516,
		0.071031,
		0.089462,
		0.110452,
		0.133455,
		0.15741,
		0.180892,
		0.202269,
		0.219874,
		0.232157,
		0.237836,
		0.236018,
		0.226305,
		0.208867,
		0.184485,
		0.154555,
		0.121059,
		0.086499,
		0.053801,
		0.026203,
		0.007118,
		7.1e-005,
		0.000326,
		0.001013,
		0.002452,
		0.004985,
		0.00891,
		0.014441,
		0.021703,
		0.030794,
		0.041877,
		0.055213,
		0.070995,
		0.08905,
		0.108744,
		0.129044,
		0.148632,
		0.166038,
		0.179781,
		0.188515,
		0.191158,
		0.187018,
		0.175889,
		0.158129,
		0.134701,
		0.107181,
		0.077731,
		0.049033,
		0.024201,
		0.006658,
		1.8e-005,
		0.000122,
		0.000475,
		0.001319,
		0.002939,
		0.005612,
		0.009563,
		0.014966,
		0.021994,
		0.030877,
		0.041877,
		0.055078,
		0.070213,
		0.086641,
		0.103399,
		0.119294,
		0.133017,
		0.143273,
		0.148905,
		0.14903,
		0.143149,
		0.131251,
		0.113883,
		0.092199,
		0.067966,
		0.043541,
		0.021807,
		0.006083,
		2.8e-005,
		0.00017,
		0.000596,
		0.001535,
		0.003231,
		0.005911,
		0.00978,
		0.015059,
		0.022008,
		0.030845,
		0.041577,
		0.053909,
		0.067237,
		0.080692,
		0.09321,
		0.103642,
		0.110866,
		0.113912,
		0.112084,
		0.105078,
		0.093074,
		0.076818,
		0.057659,
		0.037571,
		0.019122,
		0.005415,
		3.8e-005,
		0.000206,
		0.000676,
		0.001651,
		0.003347,
		0.005983,
		0.0098,
		0.015054,
		0.021916,
		0.030358,
		0.040099,
		0.050596,
		0.061078,
		0.070615,
		0.078208,
		0.082904,
		0.08391,
		0.080714,
		0.073201,
		0.061746,
		0.047295,
		0.031406,
		0.01627,
		0.004685,
		4.4e-005,
		0.000228,
		0.000712,
		0.001686,
		0.003364,
		0.005985,
		0.00978,
		0.014891,
		0.021286,
		0.02872,
		0.036721,
		0.04462,
		0.051615,
		0.056853,
		0.05954,
		0.059053,
		0.05506,
		0.047631,
		0.037339,
		0.025332,
		0.013387,
		0.003928,
		4.8e-005,
		0.000236,
		0.000719,
		0.001689,
		0.003361,
		0.005944,
		0.009563,
		0.014202,
		0.019659,
		0.025537,
		0.031269,
		0.036175,
		0.039541,
		0.040725,
		0.039266,
		0.035006,
		0.028199,
		0.019612,
		0.010604,
		0.003178,
		4.9e-005,
		0.000237,
		0.000719,
		0.001682,
		0.003307,
		0.005716,
		0.00891,
		0.012738,
		0.016877,
		0.020858,
		0.024116,
		0.026068,
		0.026217,
		0.024257,
		0.020197,
		0.014469,
		0.008036,
		0.002468,
		4.9e-005,
		0.000236,
		0.000712,
		0.001631,
		0.00311,
		0.005174,
		0.007717,
		0.010488,
		0.01311,
		0.015128,
		0.016081,
		0.015607,
		0.013546,
		0.010067,
		0.005777,
		0.001827,
		4.8e-005,
		0.000231,
		0.000676,
		0.001491,
		0.002719,
		0.004297,
		0.006041,
		0.007659,
		0.008796,
		0.009106,
		0.008349,
		0.006505,
		0.00389,
		0.001275,
		4.6e-005,
		0.000213,
		0.000596,
		0.001249,
		0.002146,
		0.003161,
		0.00408,
		0.004637,
		0.004591,
		0.003813,
		0.002407,
		0.000827,
		4.1e-005,
		0.00018,
		0.000475,
		0.000927,
		0.00146,
		0.001926,
		0.002141,
		0.001951,
		0.001329,
		0.000487,
		3.3e-005,
		0.000135,
		0.000326,
		0.000568,
		0.00077,
		0.000815,
		0.000623,
		0.00025,
		2.3e-005,
		8.3e-005,
		0.000173,
		0.000242,
		0.000225,
		0.000105,
		1.2e-005,
		3.6e-005,
		5.1e-005,
		3.1e-005,
		3.1e-005,
		5.1e-005,
		3.6e-005,
		1.2e-005,
		0.000105,
		0.000225,
		0.000242,
		0.000173,
		8.3e-005,
		2.3e-005,
		0.00025,
		0.000623,
		0.000815,
		0.00077,
		0.000568,
		0.000326,
		0.000135,
		3.3e-005,
		0.000487,
		0.001329,
		0.001951,
		0.002141,
		0.001926,
		0.00146,
		0.000927,
		0.000475,
		0.00018,
		4.1e-005,
		0.000827,
		0.002407,
		0.003813,
		0.004591,
		0.004637,
		0.00408,
		0.003161,
		0.002146,
		0.001249,
		0.000596,
		0.000213,
		4.6e-005,
		0.001275,
		0.00389,
		0.006505,
		0.008349,
		0.009106,
		0.008796,
		0.007659,
		0.006041,
		0.004297,
		0.002719,
		0.001491,
		0.000676,
		0.000231,
		4.8e-005,
		0.001827,
		0.005777,
		0.010067,
		0.013546,
		0.015607,
		0.016081,
		0.015128,
		0.01311,
		0.010488,
		0.007717,
		0.005174,
		0.00311,
		0.001631,
		0.000712,
		0.000236,
		4.9e-005,
		0.002468,
		0.008036,
		0.014469,
		0.020197,
		0.024257,
		0.026217,
		0.026068,
		0.024116,
		0.020858,
		0.016877,
		0.012738,
		0.00891,
		0.005716,
		0.003307,
		0.001682,
		0.000719,
		0.000237,
		4.9e-005,
		0.003178,
		0.010604,
		0.019612,
		0.028199,
		0.035006,
		0.039266,
		0.040725,
		0.039541,
		0.036175,
		0.031269,
		0.025537,
		0.019659,
		0.014202,
		0.009563,
		0.005944,
		0.003361,
		0.001689,
		0.000719,
		0.000236,
		4.8e-005,
		0.003928,
		0.013387,
		0.025332,
		0.037339,
		0.047631,
		0.05506,
		0.059053,
		0.05954,
		0.056853,
		0.051615,
		0.04462,
		0.036721,
		0.02872,
		0.021286,
		0.014891,
		0.00978,
		0.005985,
		0.003364,
		0.001686,
		0.000712,
		0.000228,
		4.4e-005,
		0.004685,
		0.01627,
		0.031406,
		0.047295,
		0.061746,
		0.073201,
		0.080714,
		0.08391,
		0.082904,
		0.078208,
		0.070615,
		0.061078,
		0.050596,
		0.040099,
		0.030358,
		0.021916,
		0.015054,
		0.0098,
		0.005983,
		0.003347,
		0.001651,
		0.000676,
		0.000206,
		3.8e-005,
		0.005415,
		0.019122,
		0.037571,
		0.057659,
		0.076818,
		0.093074,
		0.105078,
		0.112084,
		0.113912,
		0.110866,
		0.103642,
		0.09321,
		0.080692,
		0.067237,
		0.053909,
		0.041577,
		0.030845,
		0.022008,
		0.015059,
		0.00978,
		0.005911,
		0.003231,
		0.001535,
		0.000596,
		0.00017,
		2.8e-005,
		0.006083,
		0.021807,
		0.043541,
		0.067966,
		0.092199,
		0.113883,
		0.131251,
		0.143149,
		0.14903,
		0.148905,
		0.143273,
		0.133017,
		0.119294,
		0.103399,
		0.086641,
		0.070213,
		0.055078,
		0.041877,
		0.030877,
		0.021994,
		0.014966,
		0.009563,
		0.005612,
		0.002939,
		0.001319,
		0.000475,
		0.000122,
		1.8e-005,
		0.006658,
		0.024201,
		0.049033,
		0.077731,
		0.107181,
		0.134701,
		0.158129,
		0.175889,
		0.187018,
		0.191158,
		0.188515,
		0.179781,
		0.166038,
		0.148632,
		0.129044,
		0.108744,
		0.08905,
		0.070995,
		0.055213,
		0.041877,
		0.030794,
		0.021703,
		0.014441,
		0.00891,
		0.004985,
		0.002452,
		0.001013,
		0.000326,
		7.1e-005,
		0.007118,
		0.026203,
		0.053801,
		0.086499,
		0.121059,
		0.154555,
		0.184485,
		0.208867,
		0.226305,
		0.236018,
		0.237836,
		0.232157,
		0.219874,
		0.202269,
		0.180892,
		0.15741,
		0.133455,
		0.110452,
		0.089462,
		0.071031,
		0.05516,
		0.041577,
		0.03008,
		0.020615,
		0.013182,
		0.007717,
		0.004032,
		0.001811,
		0.000658,
		0.000173,
		2.6e-005,
		0.007452,
		0.027745,
		0.057659,
		0.093895,
		0.133206,
		0.172521,
		0.209081,
		0.240551,
		0.265117,
		0.281552,
		0.289251,
		0.288229,
		0.279078,
		0.262895,
		0.241175,
		0.215674,
		0.188251,
		0.160683,
		0.134466,
		0.110603,
		0.089443,
		0.070788,
		0.054366,
		0.040099,
		0.028075,
		0.018409,
		0.011116,
		0.006041,
		0.002854,
		0.001107,
		0.000316,
		5.1e-005,
		0.007659,
		0.028801,
		0.060501,
		0.099666,
		0.143141,
		0.187825,
		0.230803,
		0.269468,
		0.30164,
		0.325662,
		0.34047,
		0.34563,
		0.341342,
		0.328396,
		0.308099,
		0.282165,
		0.252561,
		0.22133,
		0.190361,
		0.161142,
		0.134491,
		0.110452,
		0.088707,
		0.069081,
		0.051666,
		0.036721,
		0.024496,
		0.015095,
		0.008402,
		0.00408,
		0.001626,
		0.00047,
		6.8e-005,
		0.007746,
		0.029384,
		0.062312,
		0.103703,
		0.150584,
		0.199929,
		0.24878,
		0.294361,
		0.334211,
		0.366296,
		0.389109,
		0.401745,
		0.403941,
		0.39608,
		0.379157,
		0.354699,
		0.32464,
		0.291157,
		0.256446,
		0.222444,
		0.190498,
		0.161075,
		0.133926,
		0.108744,
		0.085527,
		0.064584,
		0.046365,
		0.031269,
		0.019499,
		0.010985,
		0.005384,
		0.002141,
		0.000588,
		6.1e-005,
		0.007731,
		0.029542,
		0.063155,
		0.106041,
		0.155471,
		0.208583,
		0.26248,
		0.314334,
		0.361507,
		0.401677,
		0.43295,
		0.453969,
		0.463986,
		0.462908,
		0.451312,
		0.430401,
		0.401925,
		0.368043,
		0.331119,
		0.293444,
		0.25688,
		0.222444,
		0.190146,
		0.159588,
		0.130593,
		0.103399,
		0.078559,
		0.056733,
		0.038483,
		0.024116,
		0.013614,
		0.006631,
		0.002558,
		0.000623,
		2.7e-005,
		0.007635,
		0.029348,
		0.06316,
		0.106843,
		0.157944,
		0.213831,
		0.271753,
		0.328938,
		0.382687,
		0.430498,
		0.470186,
		0.5,
		0.518731,
		0.525792,
		0.52127,
		0.505937,
		0.481216,
		0.449081,
		0.4119,
		0.372177,
		0.332194,
		0.293546,
		0.256712,
		0.22133,
		0.18702,
		0.153832,
		0.122288,
		0.09321,
		0.067482,
		0.045834,
		0.028687,
		0.016081,
		0.007669,
		0.002784,
		0.000544,
		0.00748,
		0.028884,
		0.062493,
		0.10636,
		0.158311,
		0.215973,
		0.276811,
		0.338191,
		0.397462,
		0.45206,
		0.499619,
		0.538097,
		0.565889,
		0.581942,
		0.585835,
		0.577846,
		0.558952,
		0.530799,
		0.495581,
		0.455837,
		0.414121,
		0.372533,
		0.332146,
		0.292845,
		0.254094,
		0.215674,
		0.177929,
		0.141698,
		0.108085,
		0.078208,
		0.052975,
		0.032928,
		0.018168,
		0.008349,
		0.002744,
		0.000356,
		0.007286,
		0.028234,
		0.061335,
		0.104887,
		0.156977,
		0.215498,
		0.278157,
		0.342523,
		0.406079,
		0.466305,
		0.520775,
		0.567267,
		0.603886,
		0.629185,
		0.642277,
		0.642924,
		0.631602,
		0.609506,
		0.578504,
		0.540983,
		0.499583,
		0.456759,
		0.414176,
		0.372177,
		0.330213,
		0.287772,
		0.244896,
		0.202269,
		0.161065,
		0.122681,
		0.088475,
		0.05954,
		0.036547,
		0.019662,
		0.008546,
		0.002407,
		0.000126,
		0.007072,
		0.027471,
		0.059855,
		0.102721,
		0.154387,
		0.212988,
		0.276483,
		0.34267,
		0.409225,
		0.473757,
		0.533882,
		0.58732,
		0.632006,
		0.666205,
		0.68864,
		0.698606,
		0.696065,
		0.681705,
		0.656947,
		0.623865,
		0.584989,
		0.542935,
		0.499836,
		0.456627,
		0.412943,
		0.368043,
		0.321588,
		0.273949,
		0.226175,
		0.179781,
		0.136466,
		0.097831,
		0.065163,
		0.039266,
		0.020382,
		0.008178,
		0.001801,
		0.006851,
		0.026654,
		0.058196,
		0.100129,
		0.15096,
		0.209038,
		0.272552,
		0.339533,
		0.407876,
		0.475373,
		0.539764,
		0.598814,
		0.650402,
		0.692627,
		0.723937,
		0.743241,
		0.750038,
		0.744506,
		0.727561,
		0.700845,
		0.666607,
		0.627422,
		0.585696,
		0.542935,
		0.499182,
		0.4536,
		0.405462,
		0.354699,
		0.302048,
		0.248937,
		0.197224,
		0.148905,
		0.105836,
		0.0695,
		0.040846,
		0.020197,
		0.007231,
		0.001038,
		0.00663,
		0.025825,
		0.056465,
		0.097323,
		0.147054,
		0.204182,
		0.267095,
		0.334039,
		0.403127,
		0.472358,
		0.539646,
		0.602874,
		0.659966,
		0.708972,
		0.748179,
		0.77623,
		0.792248,
		0.795954,
		0.787765,
		0.768831,
		0.740998,
		0.706621,
		0.66817,
		0.627576,
		0.585423,
		0.540983,
		0.493212,
		0.441538,
		0.386226,
		0.328396,
		0.269848,
		0.212781,
		0.159489,
		0.112084,
		0.072255,
		0.041104,
		0.019045,
		0.005777,
		0.000329,
		0.006416,
		0.02501,
		0.054735,
		0.094456,
		0.142941,
		0.198852,
		0.260738,
		0.327031,
		0.396042,
		0.465973,
		0.534931,
		0.600964,
		0.662108,
		0.716456,
		0.76225,
		0.797981,
		0.822512,
		0.835202,
		0.836021,
		0.825636,
		0.805429,
		0.777408,
		0.743918,
		0.707107,
		0.668102,
		0.626498,
		0.581082,
		0.530799,
		0.475331,
		0.415302,
		0.352213,
		0.288229,
		0.22588,
		0.167757,
		0.116229,
		0.073201,
		0.039939,
		0.016957,
		0.003985,
		0.006211,
		0.024222,
		0.053048,
		0.091624,
		0.138805,
		0.19336,
		0.253963,
		0.319197,
		0.38754,
		0.457378,
		0.527004,
		0.594644,
		0.658487,
		0.716734,
		0.767663,
		0.809723,
		0.841631,
		0.862495,
		0.871933,
		0.870185,
		0.858176,
		0.837497,
		0.810233,
		0.778552,
		0.743988,
		0.706621,
		0.665273,
		0.618508,
		0.565401,
		0.505937,
		0.441107,
		0.37278,
		0.303466,
		0.236018,
		0.17332,
		0.118003,
		0.072196,
		0.037339,
		0.014062,
		0.002137,
		0.006014,
		0.023468,
		0.051425,
		0.088879,
		0.134756,
		0.187904,
		0.247099,
		0.311037,
		0.378334,
		0.447525,
		0.517072,
		0.585368,
		0.650766,
		0.711604,
		0.766258,
		0.81321,
		0.851131,
		0.878988,
		0.896157,
		0.902538,
		0.898653,
		0.885676,
		0.865353,
		0.839721,
		0.810531,
		0.778389,
		0.742376,
		0.700845,
		0.652329,
		0.596121,
		0.532534,
		0.462908,
		0.389462,
		0.315035,
		0.242785,
		0.175889,
		0.11724,
		0.069201,
		0.033402,
		0.010604,
		0.000636,
		0.005827,
		0.022747,
		0.04987,
		0.086243,
		0.130845,
		0.182595,
		0.240343,
		0.302876,
		0.368915,
		0.437117,
		0.50608,
		0.574344,
		0.640413,
		0.702768,
		0.75991,
		0.810403,
		0.852944,
		0.886448,
		0.910151,
		0.923719,
		0.92735,
		0.921848,
		0.908605,
		0.889428,
		0.86612,
		0.839721,
		0.809757,
		0.774593,
		0.732363,
		0.681705,
		0.622176,
		0.554413,
		0.48009,
		0.401745,
		0.322525,
		0.245885,
		0.175283,
		0.113883,
		0.064291,
		0.02834,
		0.006942,
		0.005649,
		0.022059,
		0.048385,
		0.083718,
		0.127091,
		0.177477,
		0.233791,
		0.294891,
		0.359581,
		0.426612,
		0.494688,
		0.56247,
		0.628589,
		0.691658,
		0.750302,
		0.803188,
		0.849081,
		0.88691,
		0.915851,
		0.935426,
		0.945607,
		0.946897,
		0.940363,
		0.927549,
		0.910182,
		0.889587,
		0.865825,
		0.837497,
		0.80255,
		0.759087,
		0.705931,
		0.642924,
		0.571013,
		0.492178,
		0.40925,
		0.325662,
		0.24515,
		0.171451,
		0.108001,
		0.057659,
		0.022482,
		0.00355,
		0.005478,
		0.021402,
		0.046964,
		0.081301,
		0.123492,
		0.172561,
		0.227479,
		0.287164,
		0.350485,
		0.416269,
		0.483303,
		0.550341,
		0.616116,
		0.679351,
		0.73878,
		0.793172,
		0.841376,
		0.882368,
		0.91532,
		0.939685,
		0.955287,
		0.96241,
		0.961852,
		0.954896,
		0.943133,
		0.928019,
		0.910115,
		0.888462,
		0.861048,
		0.825636,
		0.780424,
		0.724468,
		0.657896,
		0.581942,
		0.498862,
		0.411755,
		0.324316,
		0.240551,
		0.164471,
		0.099785,
		0.04962,
		0.01627,
		0.00101,
		0.005315,
		0.020774,
		0.045605,
		0.078987,
		0.120042,
		0.167843,
		0.221409,
		0.279714,
		0.341684,
		0.406206,
		0.472135,
		0.538297,
		0.603506,
		0.666569,
		0.726308,
		0.781582,
		0.831316,
		0.874547,
		0.910471,
		0.938519,
		0.95843,
		0.970337,
		0.974831,
		0.97297,
		0.966179,
		0.955943,
		0.943198,
		0.927549,
		0.907247,
		0.879941,
		0.843385,
		0.795954,
		0.736961,
		0.666804,
		0.586982,
		0.5,
		0.409197,
		0.318512,
		0.232205,
		0.154555,
		0.089552,
		0.040599,
		0.01025,
		0.005159,
		0.020173,
		0.044304,
		0.07677,
		0.116734,
		0.163312,
		0.215573,
		0.272538,
		0.333186,
		0.39646,
		0.461269,
		0.526502,
		0.591031,
		0.653732,
		0.713495,
		0.769247,
		0.819979,
		0.864779,
		0.902878,
		0.933706,
		0.956957,
		0.972665,
		0.981267,
		0.983637,
		0.981036,
		0.974912,
		0.966443,
		0.955795,
		0.941677,
		0.921848,
		0.893836,
		0.855502,
		0.805445,
		0.743241,
		0.669547,
		0.586099,
		0.495623,
		0.401677,
		0.308422,
		0.220363,
		0.142042,
		0.077731,
		0.031123,
		0.005056,
		0.00501,
		0.019597,
		0.043058,
		0.074644,
		0.113559,
		0.15896,
		0.209959,
		0.265625,
		0.324985,
		0.387032,
		0.450728,
		0.515011,
		0.578809,
		0.641049,
		0.700674,
		0.756663,
		0.808054,
		0.853978,
		0.893693,
		0.926633,
		0.952465,
		0.971156,
		0.983029,
		0.988808,
		0.989607,
		0.986797,
		0.981687,
		0.974912,
		0.965756,
		0.95229,
		0.932032,
		0.902538,
		0.861859,
		0.808844,
		0.743327,
		0.666205,
		0.579429,
		0.485925,
		0.389445,
		0.294361,
		0.205406,
		0.127383,
		0.064851,
		0.021807,
		0.001391,
		0.004868,
		0.019046,
		0.041863,
		0.072605,
		0.110511,
		0.154777,
		0.204557,
		0.258963,
		0.317068,
		0.377913,
		0.440506,
		0.503836,
		0.566876,
		0.628601,
		0.687995,
		0.74408,
		0.795928,
		0.842698,
		0.883665,
		0.918264,
		0.94614,
		0.967203,
		0.981687,
		0.990193,
		0.993699,
		0.993485,
		0.990906,
		0.986929,
		0.981438,
		0.97297,
		0.95923,
		0.937675,
		0.906002,
		0.862495,
		0.806267,
		0.737402,
		0.65702,
		0.567267,
		0.471247,
		0.372893,
		0.276767,
		0.187825,
		0.111127,
		0.051521,
		0.013323,
		0.004731,
		0.018518,
		0.040718,
		0.070648,
		0.107582,
		0.150754,
		0.199356,
		0.25254,
		0.309424,
		0.369091,
		0.430596,
		0.492972,
		0.555238,
		0.61641,
		0.675515,
		0.731605,
		0.783783,
		0.831226,
		0.873218,
		0.909189,
		0.938756,
		0.961779,
		0.97841,
		0.989142,
		0.994828,
		0.996642,
		0.99593,
		0.993865,
		0.99085,
		0.985989,
		0.977327,
		0.96241,
		0.938773,
		0.904319,
		0.857596,
		0.797981,
		0.7258,
		0.642378,
		0.550042,
		0.45206,
		0.352525,
		0.256183,
		0.168202,
		0.093895,
		0.038407,
		0.006383,
		0.0046,
		0.018011,
		0.039618,
		0.068768,
		0.104768,
		0.146884,
		0.194346,
		0.246346,
		0.30204,
		0.360554,
		0.420985,
		0.48241,
		0.54389,
		0.604481,
		0.663248,
		0.719275,
		0.771692,
		0.819691,
		0.862561,
		0.89972,
		0.930757,
		0.955481,
		0.973965,
		0.986599,
		0.994116,
		0.997575,
		0.998268,
		0.997473,
		0.995982,
		0.993485,
		0.988479,
		0.978745,
		0.961826,
		0.935426,
		0.897699,
		0.847468,
		0.784374,
		0.708972,
		0.622781,
		0.528295,
		0.428931,
		0.328938,
		0.23323,
		0.147183,
		0.076359,
		0.026203,
		0.001708,
		0.004475,
		0.017525,
		0.038563,
		0.066963,
		0.102062,
		0.14316,
		0.18952,
		0.24037,
		0.294906,
		0.352291,
		0.411665,
		0.472143,
		0.532828,
		0.592813,
		0.651201,
		0.707107,
		0.759685,
		0.808146,
		0.851782,
		0.890001,
		0.922363,
		0.948626,
		0.968789,
		0.983139,
		0.99229,
		0.997181,
		0.999017,
		0.9991,
		0.998474,
		0.997356,
		0.994761,
		0.988808,
		0.977186,
		0.95755,
		0.927827,
		0.886448,
		0.832516,
		0.765928,
		0.687459,
		0.598814,
		0.502643,
		0.402506,
		0.302792,
		0.208583,
		0.125451,
		0.05921,
		0.015609,
		0.004354,
		0.017058,
		0.037549,
		0.065227,
		0.099458,
		0.139573,
		0.184867,
		0.234603,
		0.288011,
		0.344293,
		0.402626,
		0.462163,
		0.522046,
		0.581406,
		0.639377,
		0.695109,
		0.747781,
		0.796619,
		0.840925,
		0.880097,
		0.913673,
		0.941362,
		0.963094,
		0.979061,
		0.989757,
		0.995994,
		0.998871,
		0.999655,
		0.999523,
		0.9991,
		0.997934,
		0.994554,
		0.986889,
		0.972665,
		0.949718,
		0.916238,
		0.870942,
		0.81321,
		0.743188,
		0.661863,
		0.57112,
		0.473757,
		0.373469,
		0.274783,
		0.182939,
		0.103703,
		0.043137,
		0.007301,
		0.004238,
		0.01661,
		0.036574,
		0.063557,
		0.096952,
		0.136118,
		0.18038,
		0.229035,
		0.281345,
		0.336549,
		0.393858,
		0.452462,
		0.511539,
		0.570256,
		0.62778,
		0.683291,
		0.735993,
		0.785132,
		0.830017,
		0.870045,
		0.904734,
		0.933754,
		0.956972,
		0.974494,
		0.9867,
		0.994276,
		0.998198,
		0.999659,
		0.99988,
		0.999756,
		0.99933,
		0.997575,
		0.992725,
		0.982663,
		0.965238,
		0.938519,
		0.900971,
		0.8516,
		0.790059,
		0.716734,
		0.632819,
		0.540369,
		0.442328,
		0.342523,
		0.245614,
		0.156993,
		0.08262,
		0.028801,
		0.00191,
		0.004127,
		0.016179,
		0.035637,
		0.061951,
		0.094539,
		0.132788,
		0.176052,
		0.223658,
		0.2749,
		0.32905,
		0.385352,
		0.443032,
		0.501302,
		0.559361,
		0.61641,
		0.671658,
		0.724333,
		0.773702,
		0.819083,
		0.859876,
		0.895582,
		0.925841,
		0.950468,
		0.969489,
		0.983187,
		0.992122,
		0.997146,
		0.999349,
		0.999936,
		0.999952,
		0.999837,
		0.999026,
		0.996096,
		0.989142,
		0.9761,
		0.955004,
		0.924176,
		0.882368,
		0.828867,
		0.76359,
		0.687158,
		0.600964,
		0.507231,
		0.409045,
		0.310362,
		0.215973,
		0.131419,
		0.062848,
		0.016812,
		0.00402,
		0.015765,
		0.034735,
		0.060404,
		0.092214,
		0.129577,
		0.171875,
		0.218463,
		0.268666,
		0.321785,
		0.377099,
		0.433865,
		0.491327,
		0.548718,
		0.605268,
		0.660215,
		0.712812,
		0.762345,
		0.808146,
		0.849616,
		0.886248,
		0.917657,
		0.943612,
		0.964076,
		0.97924,
		0.989554,
		0.995745,
		0.998793,
		0.999839,
		0.999993,
		0.999976,
		0.999659,
		0.997988,
		0.99331,
		0.983695,
		0.967203,
		0.942087,
		0.906936,
		0.860781,
		0.803188,
		0.734326,
		0.655041,
		0.566923,
		0.472358,
		0.374572,
		0.277651,
		0.186511,
		0.106843,
		0.04498,
		0.007716,
		0.003918,
		0.015366,
		0.033867,
		0.058915,
		0.089974,
		0.12648,
		0.167843,
		0.213443,
		0.262634,
		0.314747,
		0.369091,
		0.424954,
		0.48161,
		0.538323,
		0.594354,
		0.648967,
		0.701439,
		0.751077,
		0.797225,
		0.839291,
		0.876762,
		0.909232,
		0.936437,
		0.95828,
		0.974876,
		0.986574,
		0.993987,
		0.997978,
		0.999606,
		0.999982,
		0.999998,
		0.999904,
		0.999028,
		0.995994,
		0.989042,
		0.9763,
		0.956001,
		0.926633,
		0.887051,
		0.836564,
		0.774998,
		0.702768,
		0.620939,
		0.531288,
		0.436364,
		0.339533,
		0.245006,
		0.157829,
		0.083836,
		0.029542,
		0.001983,
		0.003819,
		0.014982,
		0.033031,
		0.057479,
		0.087813,
		0.123492,
		0.163949,
		0.208589,
		0.256796,
		0.307926,
		0.361318,
		0.416289,
		0.472143,
		0.528173,
		0.583667,
		0.637915,
		0.690221,
		0.739908,
		0.786339,
		0.828924,
		0.867151,
		0.900598,
		0.928972,
		0.952128,
		0.970112,
		0.983187,
		0.991857,
		0.996876,
		0.999205,
		0.99992,
		1,
		0.999982,
		0.99958,
		0.997716,
		0.992826,
		0.983139,
		0.966903,
		0.942546,
		0.908799,
		0.864779,
		0.810056,
		0.744711,
		0.66939,
		0.585368,
		0.494607,
		0.399815,
		0.304496,
		0.212988,
		0.130464,
		0.062899,
		0.016988,
		0.003724,
		0.014613,
		0.032226,
		0.056095,
		0.085729,
		0.120607,
		0.160186,
		0.203896,
		0.251144,
		0.301315,
		0.353774,
		0.407865,
		0.462921,
		0.518262,
		0.573204,
		0.627063,
		0.679164,
		0.728851,
		0.775502,
		0.818536,
		0.85744,
		0.891784,
		0.921246,
		0.945646,
		0.964969,
		0.9794,
		0.989348,
		0.995458,
		0.998595,
		0.999771,
		0.999995,
		0.999999,
		0.99985,
		0.998793,
		0.995487,
		0.988288,
		0.975486,
		0.955481,
		0.926912,
		0.888751,
		0.840368,
		0.781582,
		0.712711,
		0.634625,
		0.548801,
		0.457378,
		0.36322,
		0.269972,
		0.182095,
		0.104887,
		0.044463,
		0.007692,
		0.003632,
		0.014256,
		0.031449,
		0.054761,
		0.083718,
		0.117821,
		0.156549,
		0.199356,
		0.245671,
		0.294906,
		0.346449,
		0.399673,
		0.453937,
		0.508587,
		0.562964,
		0.61641,
		0.668272,
		0.717915,
		0.764729,
		0.808146,
		0.847654,
		0.882814,
		0.913289,
		0.938862,
		0.959469,
		0.975226,
		0.986456,
		0.993706,
		0.99774,
		0.999493,
		0.999964,
		1,
		0.999962,
		0.999429,
		0.997312,
		0.992122,
		0.982219,
		0.965998,
		0.942036,
		0.909189,
		0.86666,
		0.81406,
		0.751442,
		0.679351,
		0.598868,
		0.51166,
		0.420042,
		0.327031,
		0.236406,
		0.152754,
		0.081496,
		0.028884,
		0.001953,
		0.003544,
		0.013913,
		0.0307,
		0.053473,
		0.081776,
		0.11513,
		0.153034,
		0.194962,
		0.24037,
		0.28869,
		0.339336,
		0.391707,
		0.445185,
		0.499142,
		0.552945,
		0.605958,
		0.657551,
		0.707107,
		0.754033,
		0.797771,
		0.837812,
		0.873716,
		0.905127,
		0.931803,
		0.953636,
		0.970681,
		0.983187,
		0.991607,
		0.996609,
		0.999046,
		0.999875,
		0.999999,
		0.999995,
		0.999771,
		0.998511,
		0.994912,
		0.987433,
		0.974494,
		0.954629,
		0.926602,
		0.889481,
		0.842698,
		0.786085,
		0.719918,
		0.644953,
		0.56247,
		0.474322,
		0.382989,
		0.291638,
		0.204182,
		0.125331,
		0.060626,
		0.01645,
		0.003458,
		0.013582,
		0.029978,
		0.05223,
		0.079901,
		0.112529,
		0.149633,
		0.19071,
		0.235233,
		0.282661,
		0.332429,
		0.383959,
		0.436658,
		0.489923,
		0.543144,
		0.595706,
		0.647002,
		0.696434,
		0.743424,
		0.787425,
		0.827935,
		0.864511,
		0.896786,
		0.924495,
		0.947493,
		0.965784,
		0.979547,
		0.989154,
		0.99518,
		0.998392,
		0.999689,
		0.999986,
		1,
		0.999929,
		0.999251,
		0.996876,
		0.99139,
		0.981267,
		0.965026,
		0.941362,
		0.909231,
		0.867908,
		0.817037,
		0.756663,
		0.687268,
		0.609807,
		0.525747,
		0.437117,
		0.346557,
		0.257379,
		0.173627,
		0.100129,
		0.042546,
		0.007387,
		0.003376,
		0.013263,
		0.029281,
		0.05103,
		0.078089,
		0.110014,
		0.146343,
		0.186592,
		0.230255,
		0.276811,
		0.325719,
		0.376422,
		0.428351,
		0.480925,
		0.533556,
		0.585653,
		0.636628,
		0.685901,
		0.732912,
		0.777124,
		0.818041,
		0.855222,
		0.888291,
		0.916963,
		0.941064,
		0.960553,
		0.975548,
		0.986346,
		0.993437,
		0.997501,
		0.999369,
		0.999937,
		1,
		0.999986,
		0.999671,
		0.9982,
		0.994317,
		0.986574,
		0.973507,
		0.953766,
		0.926215,
		0.890001,
		0.844598,
		0.789852,
		0.726004,
		0.653732,
		0.574175,
		0.488978,
		0.400334,
		0.311037,
		0.224539,
		0.14501,
		0.077397,
		0.027471,
		0.001862,
		0.003297,
		0.012955,
		0.028608,
		0.049871,
		0.076337,
		0.107582,
		0.14316,
		0.182603,
		0.225429,
		0.271135,
		0.3192,
		0.369091,
		0.420258,
		0.472143,
		0.52418,
		0.575799,
		0.62643,
		0.675515,
		0.722505,
		0.766878,
		0.808146,
		0.845868,
		0.879663,
		0.909232,
		0.934374,
		0.95501,
		0.971204,
		0.983187,
		0.991369,
		0.996348,
		0.998879,
		0.999819,
		0.999996,
		0.999999,
		0.999882,
		0.999046,
		0.996415,
		0.99065,
		0.980326,
		0.964076,
		0.940699,
		0.909232,
		0.869007,
		0.819691,
		0.761316,
		0.694312,
		0.619534,
		0.538297,
		0.452411,
		0.364225,
		0.276679,
		0.19336,
		0.118558,
		0.057334,
		0.015567,
		0.003221,
		0.012657,
		0.027957,
		0.04875,
		0.074644,
		0.105229,
		0.140077,
		0.178739,
		0.22075,
		0.265625,
		0.312866,
		0.361957,
		0.412372,
		0.463572,
		0.515011,
		0.56614,
		0.61641,
		0.665277,
		0.71221,
		0.756699,
		0.798264,
		0.836467,
		0.870925,
		0.901325,
		0.927446,
		0.949176,
		0.966531,
		0.979683,
		0.988971,
		0.994912,
		0.998186,
		0.999596,
		0.999972,
		1,
		0.99997,
		0.999548,
		0.997863,
		0.993706,
		0.985719,
		0.972543,
		0.952932,
		0.925841,
		0.89048,
		0.846361,
		0.793332,
		0.731605,
		0.661781,
		0.584885,
		0.502393,
		0.416269,
		0.329012,
		0.243699,
		0.164044,
		0.094456,
		0.040101,
		0.006962,
		0.003147,
		0.01237,
		0.027329,
		0.047667,
		0.073006,
		0.102952,
		0.137092,
		0.174994,
		0.21621,
		0.260276,
		0.306709,
		0.355015,
		0.404687,
		0.455206,
		0.506046,
		0.556677,
		0.606567,
		0.655192,
		0.702034,
		0.746597,
		0.788409,
		0.827036,
		0.862095,
		0.893264,
		0.920303,
		0.943071,
		0.961546,
		0.975845,
		0.986243,
		0.993181,
		0.997265,
		0.999236,
		0.9999,
		0.999999,
		0.999996,
		0.999819,
		0.998815,
		0.995933,
		0.989904,
		0.9794,
		0.963156,
		0.940063,
		0.909232,
		0.870045,
		0.822187,
		0.765677,
		0.700893,
		0.628601,
		0.549981,
		0.466662,
		0.380753,
		0.294891,
		0.212284,
		0.136765,
		0.072856,
		0.025825,
		0.001749,
		0.003075,
		0.012092,
		0.026722,
		0.046619,
		0.071421,
		0.100747,
		0.1342,
		0.171364,
		0.211807,
		0.255081,
		0.300724,
		0.34826,
		0.397198,
		0.447041,
		0.497281,
		0.547405,
		0.596902,
		0.645261,
		0.691982,
		0.736579,
		0.778592,
		0.817592,
		0.853193,
		0.88507,
		0.912966,
		0.936718,
		0.956268,
		0.971686,
		0.983187,
		0.991144,
		0.996092,
		0.998706,
		0.999752,
		0.999989,
		1,
		0.999944,
		0.999403,
		0.997501,
		0.993081,
		0.984868,
		0.971601,
		0.952128,
		0.925483,
		0.890936,
		0.848031,
		0.796619,
		0.736881,
		0.669354,
		0.594955,
		0.515011,
		0.431293,
		0.346046,
		0.262037,
		0.182595,
		0.111669,
		0.053884,
		0.014606,
		0.003007,
		0.011823,
		0.026134,
		0.045605,
		0.069887,
		0.098612,
		0.131398,
		0.167843,
		0.207533,
		0.250035,
		0.294906,
		0.341684,
		0.3899,
		0.439072,
		0.488711,
		0.538323,
		0.587413,
		0.635486,
		0.682058,
		0.726654,
		0.768825,
		0.808146,
		0.844236,
		0.876762,
		0.905457,
		0.930138,
		0.950715,
		0.967218,
		0.97981,
		0.9888,
		0.994653,
		0.997978,
		0.999493,
		0.99995,
		1,
		0.999989,
		0.999738,
		0.99856,
		0.995431,
		0.989154,
		0.978488,
		0.962262,
		0.939451,
		0.909232,
		0.871032,
		0.824552,
		0.769799,
		0.707107,
		0.63716,
		0.561021,
		0.480153,
		0.39646,
		0.312315,
		0.230604,
		0.154774,
		0.088879,
		0.037644,
		0.006522,
		0.00294,
		0.011564,
		0.025566,
		0.044624,
		0.068402,
		0.096544,
		0.128681,
		0.164428,
		0.203384,
		0.245133,
		0.289247,
		0.335282,
		0.382786,
		0.431293,
		0.480333,
		0.529428,
		0.5781,
		0.625869,
		0.672265,
		0.716828,
		0.759116,
		0.798713,
		0.835238,
		0.868358,
		0.897796,
		0.92335,
		0.944907,
		0.962458,
		0.97612,
		0.986146,
		0.992935,
		0.99703,
		0.999094,
		0.999854,
		0.999997,
		0.999999,
		0.999906,
		0.999236,
		0.997117,
		0.992444,
		0.984024,
		0.970681,
		0.951351,
		0.92514,
		0.89137,
		0.849616,
		0.799732,
		0.741872,
		0.676513,
		0.604481,
		0.526971,
		0.445578,
		0.362328,
		0.279714,
		0.200736,
		0.128948,
		0.068505,
		0.024222,
		0.001637,
		0.002875,
		0.011312,
		0.025016,
		0.043674,
		0.066963,
		0.094539,
		0.126047,
		0.161115,
		0.199356,
		0.24037,
		0.283743,
		0.32905,
		0.375851,
		0.423701,
		0.472143,
		0.520718,
		0.568961,
		0.61641,
		0.662608,
		0.707107,
		0.749475,
		0.789303,
		0.826216,
		0.859876,
		0.890001,
		0.916375,
		0.938862,
		0.957423,
		0.97213,
		0.983187,
		0.990931,
		0.995843,
		0.998528,
		0.999675,
		0.999977,
		1,
		0.999976,
		0.999638,
		0.99828,
		0.994912,
		0.988401,
		0.977591,
		0.961395,
		0.938862,
		0.909232,
		0.87197,
		0.826794,
		0.773702,
		0.712986,
		0.645261,
		0.571482,
		0.492972,
		0.411446,
		0.32905,
		0.248386,
		0.172561,
		0.105229,
		0.050639,
		0.013691,
		0.002813,
		0.011069,
		0.024483,
		0.042755,
		0.065569,
		0.092596,
		0.123492,
		0.157899,
		0.195443,
		0.23574,
		0.278389,
		0.32298,
		0.369091,
		0.416289,
		0.464137,
		0.512188,
		0.559995,
		0.607109,
		0.653088,
		0.697495,
		0.739908,
		0.779928,
		0.817181,
		0.851331,
		0.882091,
		0.909232,
		0.9326,
		0.952128,
		0.967852,
		0.979927,
		0.988638,
		0.994403,
		0.997769,
		0.999381,
		0.99992,
		0.999999,
		0.999997,
		0.999854,
		0.999046,
		0.996711,
		0.991798,
		0.983187,
		0.969784,
		0.950601,
		0.924811,
		0.891784,
		0.851122,
		0.802682,
		0.746597,
		0.683291,
		0.613507,
		0.538323,
		0.459182,
		0.377913,
		0.29677,
		0.218463,
		0.146196,
		0.083718,
		0.035363,
		0.006111,
		0.002753,
		0.010834,
		0.023968,
		0.041863,
		0.064217,
		0.090711,
		0.121013,
		0.154777,
		0.191643,
		0.231239,
		0.273179,
		0.317068,
		0.362499,
		0.409054,
		0.45631,
		0.503836,
		0.5512,
		0.597967,
		0.643707,
		0.687996,
		0.730424,
		0.770596,
		0.808146,
		0.842738,
		0.874082,
		0.901939,
		0.926139,
		0.946591,
		0.963299,
		0.976376,
		0.986055,
		0.9927,
		0.996799,
		0.998946,
		0.999799,
		0.999991,
		1,
		0.999955,
		0.999519,
		0.997978,
		0.994377,
		0.987648,
		0.976711,
		0.960553,
		0.938296,
		0.909232,
		0.872864,
		0.828924,
		0.777401,
		0.718555,
		0.652936,
		0.581406,
		0.505161,
		0.425753,
		0.345123,
		0.265625,
		0.190067,
		0.121749,
		0.064503,
		0.022747,
		0.001534,
		0.002694,
		0.010606,
		0.023468,
		0.041,
		0.062907,
		0.088884,
		0.118608,
		0.151745,
		0.18795,
		0.226862,
		0.268109,
		0.31131,
		0.356071,
		0.40199,
		0.448658,
		0.495659,
		0.542574,
		0.588982,
		0.634466,
		0.678615,
		0.721027,
		0.761317,
		0.799123,
		0.834112,
		0.86599,
		0.894513,
		0.919497,
		0.94083,
		0.958486,
		0.972543,
		0.983187,
		0.990728,
		0.995601,
		0.998347,
		0.999588,
		0.999959,
		1,
		0.999991,
		0.999788,
		0.998834,
		0.996286,
		0.991144,
		0.982358,
		0.968908,
		0.949876,
		0.924495,
		0.892179,
		0.852555,
		0.805484,
		0.751077,
		0.689715,
		0.622067,
		0.549108,
		0.472143,
		0.392831,
		0.313213,
		0.23574,
		0.163312,
		0.099316,
		0.047667,
		0.012854,
		0.002638,
		0.010385,
		0.022984,
		0.040162,
		0.061637,
		0.08711,
		0.116272,
		0.148801,
		0.18436,
		0.222604,
		0.263174,
		0.3057,
		0.349803,
		0.395094,
		0.441178,
		0.487653,
		0.534114,
		0.580154,
		0.625367,
		0.669354,
		0.711723,
		0.752097,
		0.79012,
		0.825463,
		0.85783,
		0.886971,
		0.912691,
		0.93486,
		0.953429,
		0.968439,
		0.980037,
		0.988485,
		0.994162,
		0.997561,
		0.999261,
		0.999882,
		0.999997,
		0.999999,
		0.999924,
		0.999381,
		0.997655,
		0.993829,
		0.986896,
		0.975845,
		0.959736,
		0.937751,
		0.909232,
		0.873716,
		0.830949,
		0.780913,
		0.723836,
		0.660215,
		0.590828,
		0.516759,
		0.439414,
		0.360554,
		0.282312,
		0.207232,
		0.138302,
		0.078987,
		0.033279,
		0.005737,
		0.002583,
		0.010171,
		0.022514,
		0.03935,
		0.060404,
		0.085389,
		0.114005,
		0.14594,
		0.180871,
		0.218463,
		0.258369,
		0.300233,
		0.343689,
		0.38836,
		0.433865,
		0.479816,
		0.52582,
		0.571482,
		0.61641,
		0.660215,
		0.702516,
		0.742944,
		0.781149,
		0.816804,
		0.849616,
		0.879329,
		0.905738,
		0.9287,
		0.948142,
		0.964076,
		0.976614,
		0.98597,
		0.992476,
		0.996571,
		0.998793,
		0.999735,
		0.999982,
		1,
		0.999981,
		0.999707,
		0.9986,
		0.995843,
		0.990485,
		0.981538,
		0.968053,
		0.949176,
		0.924192,
		0.892557,
		0.85392,
		0.808146,
		0.755329,
		0.69581,
		0.630193,
		0.559361,
		0.484498,
		0.407111,
		0.32905,
		0.25254,
		0.180211,
		0.11513,
		0.06084,
		0.021402,
		0.00144,
		0.00253,
		0.009964,
		0.022059,
		0.038563,
		0.059208,
		0.083718,
		0.111802,
		0.14316,
		0.177478,
		0.214433,
		0.25369,
		0.294906,
		0.337724,
		0.381784,
		0.426716,
		0.472143,
		0.517687,
		0.562964,
		0.607595,
		0.651201,
		0.69341,
		0.733863,
		0.772216,
		0.808146,
		0.84136,
		0.8716,
		0.898654,
		0.922365,
		0.942641,
		0.959469,
		0.972926,
		0.983187,
		0.990536,
		0.995364,
		0.998163,
		0.999493,
		0.999935,
		1,
		0.999998,
		0.999882,
		0.999223,
		0.997312,
		0.99327,
		0.986146,
		0.974996,
		0.958943,
		0.937225,
		0.909232,
		0.874529,
		0.832877,
		0.78425,
		0.728851,
		0.667127,
		0.599783,
		0.527802,
		0.452462,
		0.375363,
		0.298442,
		0.22401,
		0.154777,
		0.093885,
		0.044948,
		0.012092,
		0.002478,
		0.009762,
		0.021618,
		0.037798,
		0.058047,
		0.082095,
		0.109662,
		0.140457,
		0.174178,
		0.210511,
		0.249134,
		0.289713,
		0.331906,
		0.375363,
		0.419726,
		0.464632,
		0.509713,
		0.5546,
		0.598922,
		0.642312,
		0.684409,
		0.72486,
		0.763329,
		0.799498,
		0.833075,
		0.863799,
		0.891454,
		0.91587,
		0.936941,
		0.95463,
		0.968983,
		0.98014,
		0.98834,
		0.99393,
		0.997353,
		0.999134,
		0.999836,
		0.999993,
		1,
		0.999963,
		0.999608,
		0.998347,
		0.995385,
		0.989821,
		0.980727,
		0.967218,
		0.948498,
		0.9239,
		0.892918,
		0.855222,
		0.81068,
		0.759371,
		0.701601,
		0.637915,
		0.569117,
		0.496282,
		0.420779,
		0.344293,
		0.268846,
		0.196828,
		0.131022,
		0.074644,
		0.031373,
		0.005396,
		0.002428,
		0.009567,
		0.021189,
		0.037057,
		0.05692,
		0.080519,
		0.107582,
		0.137829,
		0.170967,
		0.206693,
		0.244695,
		0.28465,
		0.326228,
		0.369091,
		0.412891,
		0.457278,
		0.501896,
		0.546387,
		0.59039,
		0.63355,
		0.675515,
		0.71594,
		0.754496,
		0.790869,
		0.82477,
		0.855939,
		0.884152,
		0.909232,
		0.931058,
		0.949573,
		0.964797,
		0.976835,
		0.98589,
		0.992261,
		0.996348,
		0.998635,
		0.999662,
		0.999967,
		1,
		0.999993,
		0.999827,
		0.999046,
		0.99695,
		0.9927,
		0.985399,
		0.974162,
		0.958172,
		0.936718,
		0.909232,
		0.875306,
		0.834714,
		0.787425,
		0.733619,
		0.673697,
		0.608301,
		0.538323,
		0.464929,
		0.389573,
		0.314019,
		0.24037,
		0.171089,
		0.109032,
		0.057479,
		0.020173,
		0.001354,
		0.00238,
		0.009378,
		0.020774,
		0.036336,
		0.055825,
		0.078987,
		0.105561,
		0.135274,
		0.167843,
		0.202976,
		0.24037,
		0.279714,
		0.320688,
		0.362964,
		0.406208,
		0.45008,
		0.494234,
		0.538323,
		0.582,
		0.624916,
		0.66673,
		0.707107,
		0.745722,
		0.782268,
		0.816458,
		0.848031,
		0.876762,
		0.902465,
		0.925006,
		0.944313,
		0.960379,
		0.973282,
		0.983187,
		0.990352,
		0.995135,
		0.997978,
		0.999391,
		0.999903,
		0.999998,
		1,
		0.999938,
		0.999493,
		0.998073,
		0.994912,
		0.989154,
		0.979927,
		0.966404,
		0.947842,
		0.92362,
		0.893264,
		0.856465,
		0.813093,
		0.763216,
		0.707107,
		0.645261,
		0.578407,
		0.507526,
		0.433865,
		0.358959,
		0.28465,
		0.213114,
		0.146884,
		0.088884,
		0.042454,
		0.011395,
		0.002333,
		0.009194,
		0.02037,
		0.035637,
		0.054761,
		0.077498,
		0.103595,
		0.132788,
		0.164802,
		0.199356,
		0.236155,
		0.2749,
		0.315281,
		0.35698,
		0.399673,
		0.443032,
		0.486723,
		0.530408,
		0.573749,
		0.61641,
		0.658057,
		0.698364,
		0.737013,
		0.773702,
		0.808146,
		0.840087,
		0.869296,
		0.895582,
		0.9188,
		0.938862,
		0.955742,
		0.969489,
		0.980236,
		0.988203,
		0.993706,
		0.997147,
		0.999001,
		0.999782,
		0.999985,
		1,
		0.999984,
		0.99976,
		0.99885,
		0.996571,
		0.992122,
		0.984657,
		0.973345,
		0.957423,
		0.936229,
		0.909232,
		0.87605,
		0.836467,
		0.790449,
		0.738155,
		0.679948,
		0.61641,
		0.548355,
		0.476846,
		0.403207,
		0.32905,
		0.25629,
		0.187172,
		0.124297,
		0.070648,
		0.029626,
		0.005084,
		0.002287,
		0.009016,
		0.019978,
		0.034957,
		0.053727,
		0.076051,
		0.101684,
		0.130369,
		0.161842,
		0.19583,
		0.232048,
		0.270205,
		0.310002,
		0.351133,
		0.393283,
		0.436133,
		0.47936,
		0.522638,
		0.565638,
		0.608032,
		0.649498,
		0.689715,
		0.728373,
		0.765177,
		0.799844,
		0.832117,
		0.861767,
		0.888597,
		0.912454,
		0.933236,
		0.950898,
		0.965467,
		0.977043,
		0.985814,
		0.992055,
		0.996128,
		0.998474,
		0.999581,
		0.999946,
		1,
		0.999998,
		0.999903,
		0.999361,
		0.997782,
		0.994427,
		0.988485,
		0.979138,
		0.965609,
		0.947206,
		0.92335,
		0.893596,
		0.857652,
		0.815394,
		0.766878,
		0.712348,
		0.652255,
		0.587261,
		0.518262,
		0.446396,
		0.373065,
		0.299949,
		0.229035,
		0.162631,
		0.103402,
		0.054388,
		0.019046,
		0.001275,
		0.002243,
		0.008842,
		0.019597,
		0.034297,
		0.052722,
		0.074644,
		0.099824,
		0.128015,
		0.15896,
		0.192394,
		0.228043,
		0.265625,
		0.30485,
		0.34542,
		0.387032,
		0.429378,
		0.472143,
		0.515011,
		0.557664,
		0.599783,
		0.641052,
		0.681162,
		0.719808,
		0.756699,
		0.791559,
		0.824131,
		0.854185,
		0.881522,
		0.90598,
		0.927446,
		0.94586,
		0.961225,
		0.973615,
		0.983187,
		0.990178,
		0.994912,
		0.997793,
		0.999281,
		0.999864,
		0.999994,
		1,
		0.99997,
		0.999678,
		0.998635,
		0.996177,
		0.991538,
		0.983919,
		0.972543,
		0.956695,
		0.935757,
		0.909232,
		0.876762,
		0.838141,
		0.793332,
		0.742476,
		0.685901,
		0.624138,
		0.557927,
		0.488241,
		0.416289,
		0.343545,
		0.271757,
		0.202976,
		0.139573,
		0.084269,
		0.040162,
		0.010757,
		0.0022,
		0.008674,
		0.019227,
		0.033655,
		0.051745,
		0.073275,
		0.098015,
		0.125723,
		0.156153,
		0.189047,
		0.224139,
		0.261156,
		0.299819,
		0.339837,
		0.380919,
		0.422765,
		0.465069,
		0.507526,
		0.549827,
		0.591661,
		0.632722,
		0.672708,
		0.71132,
		0.748275,
		0.783299,
		0.816138,
		0.846562,
		0.874369,
		0.899393,
		0.921508,
		0.94064,
		0.956775,
		0.969962,
		0.980326,
		0.988073,
		0.99349,
		0.996943,
		0.998863,
		0.99972,
		0.999973,
		1,
		0.999994,
		0.999858,
		0.999212,
		0.997473,
		0.99393,
		0.987816,
		0.978359,
		0.964833,
		0.946591,
		0.92309,
		0.893914,
		0.858788,
		0.817592,
		0.770369,
		0.717343,
		0.65892,
		0.595706,
		0.528519,
		0.4584,
		0.38663,
		0.314747,
		0.244566,
		0.178197,
		0.118071,
		0.066963,
		0.028021,
		0.004799,
		0.002158,
		0.00851,
		0.018867,
		0.033031,
		0.050795,
		0.071944,
		0.096253,
		0.123492,
		0.153419,
		0.185784,
		0.220331,
		0.256796,
		0.294906,
		0.334382,
		0.37494,
		0.416289,
		0.458135,
		0.50018,
		0.542124,
		0.583667,
		0.624508,
		0.664354,
		0.702914,
		0.739908,
		0.77507,
		0.808146,
		0.838908,
		0.867151,
		0.892703,
		0.915432,
		0.935251,
		0.952128,
		0.96609,
		0.977237,
		0.985742,
		0.991857,
		0.995914,
		0.99831,
		0.999493,
		0.99992,
		0.999998,
		1,
		0.999949,
		0.999581,
		0.998403,
		0.995768,
		0.990947,
		0.983187,
		0.971756,
		0.955987,
		0.935301,
		0.909232,
		0.877444,
		0.839741,
		0.796084,
		0.746597,
		0.691577,
		0.631507,
		0.567067,
		0.499142,
		0.428844,
		0.357518,
		0.286766,
		0.218463,
		0.154777,
		0.098193,
		0.051539,
		0.018011,
		0.001204,
		0.002117,
		0.008351,
		0.018518,
		0.032424,
		0.049871,
		0.070648,
		0.094539,
		0.121319,
		0.150754,
		0.182603,
		0.216617,
		0.25254,
		0.290107,
		0.32905,
		0.369091,
		0.409949,
		0.451339,
		0.492972,
		0.534555,
		0.575799,
		0.61641,
		0.656102,
		0.694592,
		0.731605,
		0.766878,
		0.800163,
		0.83123,
		0.859876,
		0.885923,
		0.909232,
		0.929706,
		0.947296,
		0.962012,
		0.973927,
		0.983187,
		0.990011,
		0.994696,
		0.997607,
		0.999166,
		0.999819,
		0.999988,
		1,
		0.999987,
		0.999801,
		0.999046,
		0.997147,
		0.993424,
		0.987147,
		0.977591,
		0.964076,
		0.945995,
		0.92284,
		0.89422,
		0.859876,
		0.819691,
		0.773702,
		0.722107,
		0.665277,
		0.603767,
		0.538323,
		0.469901,
		0.399673,
		0.32905,
		0.259691,
		0.193529,
		0.132788,
		0.080003,
		0.038051,
		0.010171,
		0.002078,
		0.008197,
		0.018178,
		0.031834,
		0.048971,
		0.069387,
		0.09287,
		0.119202,
		0.148158,
		0.179502,
		0.212995,
		0.248386,
		0.285421,
		0.323837,
		0.363368,
		0.403741,
		0.444677,
		0.485897,
		0.527118,
		0.568056,
		0.608428,
		0.647954,
		0.686356,
		0.723368,
		0.758729,
		0.792195,
		0.823539,
		0.852555,
		0.879064,
		0.90292,
		0.924016,
		0.94229,
		0.957735,
		0.970403,
		0.980411,
		0.987949,
		0.993282,
		0.996742,
		0.998722,
		0.999651,
		0.999956,
		1,
		0.999998,
		0.99992,
		0.99947,
		0.998153,
		0.995346,
		0.990352,
		0.982461,
		0.970985,
		0.955299,
		0.93486,
		0.909232,
		0.878098,
		0.841272,
		0.798713,
		0.750529,
		0.696992,
		0.638541,
		0.575799,
		0.509577,
		0.440894,
		0.370984,
		0.301315,
		0.233606,
		0.169842,
		0.112297,
		0.063557,
		0.026543,
		0.004537,
		0.00204,
		0.008046,
		0.017847,
		0.031259,
		0.048096,
		0.068159,
		0.091244,
		0.11714,
		0.145627,
		0.176478,
		0.20946,
		0.24433,
		0.280842,
		0.318742,
		0.357769,
		0.397661,
		0.438147,
		0.478955,
		0.519811,
		0.560439,
		0.600563,
		0.63991,
		0.67821,
		0.715201,
		0.750628,
		0.78425,
		0.815842,
		0.845197,
		0.872136,
		0.896506,
		0.918192,
		0.937122,
		0.953272,
		0.966673,
		0.97742,
		0.985674,
		0.991668,
		0.995704,
		0.998145,
		0.999399,
		0.999887,
		0.999995,
		1,
		0.999975,
		0.999732,
		0.998863,
		0.996806,
		0.992908,
		0.986479,
		0.976835,
		0.963338,
		0.945417,
		0.922598,
		0.894513,
		0.860918,
		0.821699,
		0.776885,
		0.726654,
		0.671346,
		0.611467,
		0.547701,
		0.480925,
		0.412216,
		0.342867,
		0.2744,
		0.208589,
		0.14747,
		0.093366,
		0.048908,
		0.017058,
		0.001138,
		0.002002,
		0.0079,
		0.017525,
		0.0307,
		0.047243,
		0.066963,
		0.08966,
		0.11513,
		0.14316,
		0.173528,
		0.20601,
		0.24037,
		0.276369,
		0.31376,
		0.352291,
		0.391707,
		0.431746,
		0.472143,
		0.512632,
		0.552945,
		0.592813,
		0.631971,
		0.670154,
		0.707107,
		0.742579,
		0.776333,
		0.808146,
		0.837812,
		0.865149,
		0.890001,
		0.912247,
		0.931803,
		0.948631,
		0.962746,
		0.974219,
		0.983187,
		0.989852,
		0.994485,
		0.997422,
		0.999046,
		0.999766,
		0.999977,
		1,
		0.999995,
		0.999882,
		0.999343,
		0.997888,
		0.994912,
		0.989754,
		0.981742,
		0.97023,
		0.95463,
		0.934434,
		0.909232,
		0.878726,
		0.842738,
		0.801227,
		0.754286,
		0.702164,
		0.645261,
		0.584147,
		0.51957,
		0.452462,
		0.383959,
		0.315408,
		0.248386,
		0.184715,
		0.12648,
		0.076051,
		0.036101,
		0.009632,
		0.001966,
		0.007758,
		0.017212,
		0.030156,
		0.046414,
		0.065798,
		0.088117,
		0.113171,
		0.140754,
		0.170651,
		0.202643,
		0.236502,
		0.271997,
		0.308888,
		0.34693,
		0.385876,
		0.425471,
		0.465459,
		0.50558,
		0.545574,
		0.58518,
		0.624138,
		0.662191,
		0.699089,
		0.734587,
		0.76845,
		0.800458,
		0.830407,
		0.858112,
		0.883417,
		0.906191,
		0.926344,
		0.943825,
		0.958631,
		0.970816,
		0.980491,
		0.987832,
		0.993081,
		0.996543,
		0.998577,
		0.999575,
		0.999933,
		0.999999,
		1,
		0.999957,
		0.999651,
		0.998665,
		0.996452,
		0.992386,
		0.985814,
		0.976091,
		0.962616,
		0.944857,
		0.922365,
		0.894796,
		0.861918,
		0.823621,
		0.779928,
		0.731,
		0.677145,
		0.618827,
		0.556677,
		0.491497,
		0.424278,
		0.356209,
		0.28869,
		0.223347,
		0.162051,
		0.106933,
		0.060404,
		0.025179,
		0.004296,
		0.001931,
		0.007619,
		0.016907,
		0.029626,
		0.045605,
		0.064663,
		0.086613,
		0.111261,
		0.138407,
		0.167843,
		0.199356,
		0.232725,
		0.267725,
		0.304124,
		0.341684,
		0.380165,
		0.419319,
		0.458899,
		0.498652,
		0.538323,
		0.577661,
		0.61641,
		0.654322,
		0.69115,
		0.726654,
		0.760605,
		0.792785,
		0.82299,
		0.851035,
		0.876762,
		0.900035,
		0.920755,
		0.938862,
		0.954338,
		0.967218,
		0.977591,
		0.98561,
		0.991486,
		0.995499,
		0.997978,
		0.999299,
		0.999848,
		0.99999,
		1,
		0.999989,
		0.999834,
		0.999202,
		0.997607,
		0.994468,
		0.989154,
		0.98103,
		0.969489,
		0.953979,
		0.934022,
		0.909232,
		0.879329,
		0.844144,
		0.803632,
		0.757879,
		0.707107,
		0.651685,
		0.592135,
		0.529145,
		0.463572,
		0.39646,
		0.32905,
		0.26279,
		0.199356,
		0.140662,
		0.088884,
		0.046472,
		0.016179,
		0.001077,
		0.001896,
		0.007484,
		0.01661,
		0.02911,
		0.044818,
		0.063557,
		0.085147,
		0.109399,
		0.136118,
		0.165103,
		0.196146,
		0.229035,
		0.263549,
		0.299464,
		0.336549,
		0.37457,
		0.413289,
		0.452462,
		0.491846,
		0.531193,
		0.570256,
		0.608789,
		0.646547,
		0.683291,
		0.718785,
		0.752804,
		0.785132,
		0.815567,
		0.843926,
		0.870045,
		0.893788,
		0.915048,
		0.933754,
		0.949876,
		0.963433,
		0.974494,
		0.983187,
		0.9897,
		0.994282,
		0.997238,
		0.998921,
		0.999707,
		0.999963,
		1,
		0.999999,
		0.999933,
		0.999556,
		0.998451,
		0.996084,
		0.991857,
		0.985151,
		0.975358,
		0.961912,
		0.944313,
		0.922139,
		0.895068,
		0.862877,
		0.825463,
		0.782841,
		0.735156,
		0.68269,
		0.625869,
		0.565272,
		0.501639,
		0.43588,
		0.369091,
		0.302559,
		0.237781,
		0.176478,
		0.120607,
		0.072383,
		0.034297,
		0.009134,
		0.001863,
		0.007353,
		0.016321,
		0.028608,
		0.044051,
		0.06248,
		0.083718,
		0.107582,
		0.133884,
		0.162428,
		0.193012,
		0.225429,
		0.259467,
		0.294906,
		0.331523,
		0.369091,
		0.407377,
		0.446146,
		0.48516,
		0.52418,
		0.562964,
		0.601273,
		0.638868,
		0.675515,
		0.710982,
		0.74505,
		0.777504,
		0.808146,
		0.836792,
		0.863277,
		0.887461,
		0.909232,
		0.928511,
		0.945255,
		0.959469,
		0.971204,
		0.980567,
		0.98772,
		0.992887,
		0.996348,
		0.998429,
		0.999493,
		0.999905,
		0.999996,
		1,
		0.999979,
		0.999776,
		0.999046,
		0.997312,
		0.994013,
		0.988552,
		0.980326,
		0.968764,
		0.953345,
		0.933622,
		0.909232,
		0.879909,
		0.845492,
		0.805937,
		0.761317,
		0.711836,
		0.657831,
		0.599783,
		0.538323,
		0.474244,
		0.408505,
		0.342249,
		0.276811,
		0.213733,
		0.154777,
		0.101941,
		0.057479,
		0.023917,
		0.004073,
		0.00183,
		0.007225,
		0.016039,
		0.028118,
		0.043303,
		0.061429,
		0.082324,
		0.10581,
		0.131705,
		0.159817,
		0.189951,
		0.221906,
		0.255475,
		0.290446,
		0.326602,
		0.363722,
		0.40158,
		0.439948,
		0.478594,
		0.517284,
		0.555785,
		0.593863,
		0.631286,
		0.667822,
		0.703249,
		0.737347,
		0.769907,
		0.800733,
		0.82964,
		0.856465,
		0.881063,
		0.903317,
		0.923141,
		0.940484,
		0.955335,
		0.967729,
		0.977753,
		0.985548,
		0.991312,
		0.995298,
		0.997811,
		0.999193,
		0.999803,
		0.999981,
		1,
		0.999996,
		0.999901,
		0.999448,
		0.998222,
		0.995704,
		0.991323,
		0.984492,
		0.974637,
		0.961225,
		0.943785,
		0.921922,
		0.895329,
		0.863799,
		0.827229,
		0.78563,
		0.739133,
		0.687996,
		0.632611,
		0.573509,
		0.511372,
		0.447041,
		0.381525,
		0.31601,
		0.251877,
		0.19071,
		0.13431,
		0.084714,
		0.044213,
		0.015366,
		0.001021,
		0.001798,
		0.007101,
		0.015765,
		0.027641,
		0.042574,
		0.060404,
		0.080964,
		0.104082,
		0.129577,
		0.157267,
		0.186961,
		0.218463,
		0.251572,
		0.286083,
		0.321785,
		0.358463,
		0.395897,
		0.433865,
		0.472143,
		0.510503,
		0.548718,
		0.586559,
		0.623799,
		0.660215,
		0.695586,
		0.729698,
		0.762345,
		0.793332,
		0.822478,
		0.849616,
		0.874601,
		0.897313,
		0.917657,
		0.935573,
		0.951039,
		0.964076,
		0.974752,
		0.983187,
		0.989554,
		0.994084,
		0.997056,
		0.998793,
		0.999641,
		0.999944,
		0.999999,
		1,
		0.999964,
		0.999707,
		0.998875,
		0.997003,
		0.993551,
		0.987949,
		0.97963,
		0.968053,
		0.952728,
		0.933236,
		0.909232,
		0.880467,
		0.846787,
		0.808146,
		0.76461,
		0.716363,
		0.663715,
		0.607109,
		0.547127,
		0.484498,
		0.42011,
		0.355015,
		0.290446,
		0.227824,
		0.168771,
		0.11513,
		0.068974,
		0.032625,
		0.008674,
		0.001767,
		0.006979,
		0.015497,
		0.027175,
		0.041863,
		0.059405,
		0.079638,
		0.102394,
		0.1275,
		0.154777,
		0.184039,
		0.215097,
		0.247755,
		0.281814,
		0.317068,
		0.353309,
		0.390324,
		0.427896,
		0.465807,
		0.503836,
		0.541761,
		0.579359,
		0.61641,
		0.652694,
		0.687996,
		0.722107,
		0.754822,
		0.785951,
		0.815311,
		0.842738,
		0.868085,
		0.891227,
		0.912065,
		0.930531,
		0.946591,
		0.960253,
		0.971569,
		0.980638,
		0.987613,
		0.9927,
		0.996156,
		0.99828,
		0.999406,
		0.999872,
		0.999991,
		1,
		0.999991,
		0.999861,
		0.999327,
		0.997978,
		0.995313,
		0.990785,
		0.983837,
		0.973927,
		0.960553,
		0.943272,
		0.921711,
		0.895582,
		0.864686,
		0.828924,
		0.788304,
		0.742944,
		0.693079,
		0.639069,
		0.581406,
		0.520718,
		0.457781,
		0.393526,
		0.32905,
		0.265625,
		0.204713,
		0.147975,
		0.097288,
		0.054761,
		0.022747,
		0.003868,
		0.001737,
		0.006861,
		0.015236,
		0.026722,
		0.04117,
		0.05843,
		0.078344,
		0.100747,
		0.125472,
		0.152345,
		0.181184,
		0.211807,
		0.244022,
		0.277636,
		0.31245,
		0.34826,
		0.384859,
		0.422038,
		0.459584,
		0.497281,
		0.534913,
		0.572264,
		0.609118,
		0.645261,
		0.680482,
		0.714575,
		0.747342,
		0.778592,
		0.808146,
		0.835839,
		0.861523,
		0.88507,
		0.906376,
		0.925367,
		0.941999,
		0.956268,
		0.968209,
		0.977906,
		0.98549,
		0.991144,
		0.995103,
		0.997644,
		0.999083,
		0.999752,
		0.999969,
		1,
		0.999999,
		0.999944,
		0.999626,
		0.99869,
		0.996681,
		0.993081,
		0.987347,
		0.978942,
		0.967356,
		0.952128,
		0.932861,
		0.909232,
		0.881004,
		0.848031,
		0.810266,
		0.767766,
		0.720701,
		0.669354,
		0.614133,
		0.555576,
		0.494356,
		0.431293,
		0.367361,
		0.303696,
		0.24161,
		0.182603,
		0.128373,
		0.08083,
		0.042115,
		0.014613,
		0.00097,
		0.001708,
		0.006746,
		0.014982,
		0.026279,
		0.040494,
		0.057479,
		0.077081,
		0.09914,
		0.123492,
		0.149968,
		0.178394,
		0.208589,
		0.24037,
		0.273547,
		0.307926,
		0.343311,
		0.379501,
		0.416289,
		0.453471,
		0.490835,
		0.528173,
		0.565272,
		0.601923,
		0.637915,
		0.673044,
		0.707107,
		0.739908,
		0.771261,
		0.800988,
		0.828924,
		0.854921,
		0.878848,
		0.900598,
		0.92009,
		0.937272,
		0.952128,
		0.964681,
		0.974996,
		0.983187,
		0.989415,
		0.993892,
		0.996876,
		0.998662,
		0.99957,
		0.99992,
		0.999997,
		1,
		0.999982,
		0.999811,
		0.999193,
		0.997721,
		0.994912,
		0.990244,
		0.983187,
		0.973229,
		0.959898,
		0.942774,
		0.921508,
		0.895825,
		0.865538,
		0.830552,
		0.790869,
		0.746597,
		0.69795,
		0.645261,
		0.588982,
		0.529695,
		0.468117,
		0.405109,
		0.341684,
		0.279019,
		0.218463,
		0.161551,
		0.110014,
		0.065798,
		0.031071,
		0.008248,
		0.001679,
		0.006633,
		0.014734,
		0.025848,
		0.039835,
		0.056551,
		0.075848,
		0.09757,
		0.121558,
		0.147646,
		0.175666,
		0.205443,
		0.236797,
		0.269544,
		0.303496,
		0.338462,
		0.374245,
		0.410647,
		0.447466,
		0.484498,
		0.521539,
		0.558383,
		0.594824,
		0.630658,
		0.665684,
		0.699703,
		0.732524,
		0.763962,
		0.793842,
		0.822,
		0.848287,
		0.872571,
		0.89474,
		0.914709,
		0.932418,
		0.947842,
		0.960989,
		0.971912,
		0.980706,
		0.987511,
		0.99252,
		0.995967,
		0.99813,
		0.999313,
		0.999833,
		0.999984,
		1,
		0.999997,
		0.999917,
		0.999533,
		0.998492,
		0.996348,
		0.992605,
		0.986746,
		0.978262,
		0.966673,
		0.951543,
		0.932498,
		0.909232,
		0.881522,
		0.849227,
		0.812301,
		0.770794,
		0.72486,
		0.67476,
		0.620871,
		0.563688,
		0.503836,
		0.442072,
		0.379296,
		0.316562,
		0.255081,
		0.19624,
		0.141606,
		0.092945,
		0.05223,
		0.021661,
		0.003677,
		0.001651,
		0.006524,
		0.014492,
		0.025427,
		0.039191,
		0.055645,
		0.074644,
		0.096037,
		0.119668,
		0.145377,
		0.173,
		0.202366,
		0.2333,
		0.265625,
		0.299157,
		0.333709,
		0.369091,
		0.405109,
		0.441568,
		0.478268,
		0.515011,
		0.551596,
		0.587822,
		0.62349,
		0.658402,
		0.692366,
		0.725192,
		0.756699,
		0.786713,
		0.815072,
		0.841627,
		0.866244,
		0.88881,
		0.909232,
		0.927446,
		0.943417,
		0.957143,
		0.968661,
		0.978051,
		0.985435,
		0.990983,
		0.994912,
		0.997478,
		0.99897,
		0.999695,
		0.999952,
		0.999999,
		1,
		0.99997,
		0.999752,
		0.999046,
		0.997451,
		0.994502,
		0.9897,
		0.982541,
		0.972543,
		0.959257,
		0.94229,
		0.921311,
		0.89606,
		0.866359,
		0.832117,
		0.793332,
		0.750101,
		0.702622,
		0.651201,
		0.596255,
		0.538323,
		0.478069,
		0.416289,
		0.353922,
		0.292057,
		0.231941,
		0.174994,
		0.122817,
		0.077206,
		0.040162,
		0.013913,
		0.000922,
		0.001624,
		0.006417,
		0.014256,
		0.025016,
		0.038563,
		0.054761,
		0.073468,
		0.094539,
		0.117821,
		0.14316,
		0.170393,
		0.199356,
		0.229879,
		0.261788,
		0.294906,
		0.32905,
		0.364035,
		0.399673,
		0.435774,
		0.472143,
		0.508587,
		0.54491,
		0.580916,
		0.61641,
		0.651201,
		0.685097,
		0.717915,
		0.749475,
		0.779606,
		0.808146,
		0.834948,
		0.859876,
		0.882814,
		0.903668,
		0.922365,
		0.938862,
		0.953149,
		0.965249,
		0.975226,
		0.983187,
		0.989282,
		0.993706,
		0.996697,
		0.998528,
		0.999493,
		0.999891,
		0.999993,
		1,
		0.999992,
		0.999882,
		0.999429,
		0.99828,
		0.996003,
		0.992122,
		0.986146,
		0.977591,
		0.966004,
		0.950973,
		0.932145,
		0.909232,
		0.882021,
		0.850378,
		0.814257,
		0.773702,
		0.728851,
		0.679948,
		0.627338,
		0.571482,
		0.512956,
		0.452462,
		0.390835,
		0.32905,
		0.268228,
		0.209654,
		0.154777,
		0.105229,
		0.062836,
		0.029626,
		0.007852,
		0.001597,
		0.006313,
		0.014026,
		0.024615,
		0.03795,
		0.053897,
		0.07232,
		0.093076,
		0.116017,
		0.140991,
		0.167843,
		0.196411,
		0.22653,
		0.258031,
		0.290741,
		0.324482,
		0.359076,
		0.394338,
		0.430082,
		0.466121,
		0.502265,
		0.538323,
		0.574105,
		0.60942,
		0.64408,
		0.677898,
		0.710695,
		0.742293,
		0.772523,
		0.801227,
		0.828254,
		0.853472,
		0.876762,
		0.898023,
		0.917181,
		0.934185,
		0.949014,
		0.961681,
		0.972236,
		0.98077,
		0.987414,
		0.992346,
		0.995782,
		0.997978,
		0.999216,
		0.999788,
		0.999973,
		1,
		0.999999,
		0.999952,
		0.999682,
		0.998886,
		0.997169,
		0.994084,
		0.989154,
		0.981901,
		0.971868,
		0.958631,
		0.94182,
		0.92112,
		0.896287,
		0.867151,
		0.833622,
		0.795699,
		0.753466,
		0.707107,
		0.656903,
		0.603241,
		0.546619,
		0.487653,
		0.427082,
		0.365774,
		0.304739,
		0.245133,
		0.188271,
		0.135635,
		0.088884,
		0.049871,
		0.020651,
		0.003501,
		0.001571,
		0.006211,
		0.013802,
		0.024223,
		0.037351,
		0.053054,
		0.071199,
		0.091646,
		0.114253,
		0.138872,
		0.165349,
		0.193529,
		0.223252,
		0.254351,
		0.28666,
		0.320005,
		0.354211,
		0.3891,
		0.424491,
		0.460201,
		0.496045,
		0.531836,
		0.567389,
		0.602518,
		0.63704,
		0.670771,
		0.703534,
		0.735156,
		0.76547,
		0.794318,
		0.821553,
		0.84704,
		0.870658,
		0.892307,
		0.911904,
		0.929393,
		0.944746,
		0.957965,
		0.969087,
		0.978187,
		0.985382,
		0.990828,
		0.994726,
		0.997312,
		0.998853,
		0.999632,
		0.999932,
		0.999997,
		1,
		0.999985,
		0.999839,
		0.999313,
		0.998056,
		0.995649,
		0.991635,
		0.985548,
		0.976929,
		0.965348,
		0.950418,
		0.931803,
		0.909232,
		0.882503,
		0.851487,
		0.816138,
		0.776495,
		0.732684,
		0.684929,
		0.63355,
		0.578973,
		0.521733,
		0.462481,
		0.40199,
		0.341166,
		0.281048,
		0.222825,
		0.167843,
		0.117611,
		0.073818,
		0.038342,
		0.013263,
		0.000878,
		0.001546,
		0.006111,
		0.013582,
		0.023841,
		0.036766,
		0.05223,
		0.070103,
		0.090249,
		0.112529,
		0.136799,
		0.16291,
		0.19071,
		0.220042,
		0.250747,
		0.282661,
		0.315615,
		0.349439,
		0.383959,
		0.418999,
		0.454381,
		0.489923,
		0.525446,
		0.560768,
		0.595706,
		0.630081,
		0.663715,
		0.696434,
		0.728066,
		0.758449,
		0.787425,
		0.814849,
		0.840584,
		0.864511,
		0.886525,
		0.90654,
		0.924495,
		0.940353,
		0.954108,
		0.965784,
		0.975444,
		0.983187,
		0.989154,
		0.993526,
		0.996521,
		0.998392,
		0.999412,
		0.999857,
		0.999986,
		1,
		0.999997,
		0.999929,
		0.999603,
		0.998713,
		0.996876,
		0.993658,
		0.988607,
		0.981267,
		0.971204,
		0.95802,
		0.941362,
		0.920935,
		0.896506,
		0.867913,
		0.835071,
		0.797974,
		0.756699,
		0.711414,
		0.66238,
		0.609954,
		0.5546,
		0.496887,
		0.437501,
		0.37725,
		0.317068,
		0.258031,
		0.201355,
		0.148414,
		0.100747,
		0.060068,
		0.02828,
		0.007484,
		0.001521,
		0.006014,
		0.013368,
		0.023468,
		0.036195,
		0.051425,
		0.069032,
		0.088884,
		0.110843,
		0.134771,
		0.160523,
		0.18795,
		0.2169,
		0.247217,
		0.278741,
		0.31131,
		0.344756,
		0.378911,
		0.413603,
		0.448658,
		0.483901,
		0.519153,
		0.55424,
		0.588982,
		0.623205,
		0.656733,
		0.689396,
		0.721027,
		0.751464,
		0.780552,
		0.808146,
		0.834112,
		0.858326,
		0.880684,
		0.901097,
		0.919497,
		0.935842,
		0.950115,
		0.962333,
		0.972543,
		0.980831,
		0.987321,
		0.992177,
		0.995601,
		0.997826,
		0.999115,
		0.999739,
		0.999959,
		0.999999,
		1,
		0.999974,
		0.999788,
		0.999185,
		0.997819,
		0.995285,
		0.991144,
		0.984953,
		0.976276,
		0.964706,
		0.949876,
		0.931471,
		0.909232,
		0.882967,
		0.852555,
		0.817948,
		0.77918,
		0.736367,
		0.689715,
		0.63952,
		0.586177,
		0.530184,
		0.472143,
		0.412775,
		0.352917,
		0.293538,
		0.23574,
		0.18077,
		0.130029,
		0.085081,
		0.047667,
		0.01971,
		0.003336,
		0.001497,
		0.00592,
		0.013159,
		0.023103,
		0.035637,
		0.050639,
		0.067986,
		0.087549,
		0.109195,
		0.132788,
		0.158187,
		0.185248,
		0.213823,
		0.243759,
		0.2749,
		0.307089,
		0.340162,
		0.373956,
		0.408303,
		0.443032,
		0.477974,
		0.512956,
		0.547804,
		0.582346,
		0.61641,
		0.649825,
		0.682423,
		0.71404,
		0.744517,
		0.773702,
		0.80145,
		0.827626,
		0.85211,
		0.874792,
		0.895582,
		0.914407,
		0.931219,
		0.945995,
		0.958739,
		0.969489,
		0.978317,
		0.985332,
		0.990679,
		0.994545,
		0.997147,
		0.998733,
		0.999565,
		0.999907,
		0.999994,
		1,
		0.999993,
		0.999899,
		0.999512,
		0.998528,
		0.996571,
		0.993226,
		0.988059,
		0.980638,
		0.970552,
		0.957423,
		0.940917,
		0.920755,
		0.896717,
		0.868649,
		0.836467,
		0.800163,
		0.759807,
		0.715554,
		0.667644,
		0.61641,
		0.56228,
		0.505785,
		0.447562,
		0.38836,
		0.32905,
		0.270627,
		0.214224,
		0.161115,
		0.112726,
		0.070648,
		0.036642,
		0.012657,
		0.000836,
		0.001474,
		0.005827,
		0.012955,
		0.022747,
		0.035092,
		0.049871,
		0.066963,
		0.086243,
		0.107582,
		0.130847,
		0.155901,
		0.182603,
		0.210809,
		0.24037,
		0.271135,
		0.302949,
		0.335654,
		0.369091,
		0.403095,
		0.437501,
		0.472143,
		0.506853,
		0.541461,
		0.575799,
		0.609698,
		0.642991,
		0.675515,
		0.707107,
		0.737611,
		0.766878,
		0.794763,
		0.821134,
		0.845868,
		0.868855,
		0.890001,
		0.909232,
		0.926493,
		0.941754,
		0.95501,
		0.966289,
		0.97565,
		0.983187,
		0.989031,
		0.99335,
		0.996348,
		0.998255,
		0.999326,
		0.999819,
		0.999977,
		1,
		0.999999,
		0.999959,
		0.999729,
		0.999046,
		0.997571,
		0.994912,
		0.99065,
		0.984361,
		0.975631,
		0.964076,
		0.949349,
		0.931149,
		0.909232,
		0.883417,
		0.853585,
		0.819691,
		0.781763,
		0.739908,
		0.694316,
		0.645261,
		0.593109,
		0.538323,
		0.481465,
		0.423201,
		0.364313,
		0.3057,
		0.248386,
		0.193529,
		0.142431,
		0.096544,
		0.057479,
		0.027023,
		0.007142,
		0.001451,
		0.005737,
		0.012755,
		0.022399,
		0.034559,
		0.049119,
		0.065963,
		0.084966,
		0.106005,
		0.128949,
		0.153664,
		0.180014,
		0.207857,
		0.237049,
		0.267443,
		0.298888,
		0.331231,
		0.364313,
		0.397978,
		0.432063,
		0.466406,
		0.500843,
		0.535208,
		0.569338,
		0.603067,
		0.636233,
		0.668673,
		0.70023,
		0.73075,
		0.760084,
		0.788092,
		0.814639,
		0.839605,
		0.862877,
		0.884362,
		0.903979,
		0.92167,
		0.937398,
		0.951151,
		0.962947,
		0.972832,
		0.980889,
		0.987232,
		0.992015,
		0.995423,
		0.997675,
		0.999011,
		0.999684,
		0.999942,
		0.999998,
		1,
		0.999987,
		0.999863,
		0.999412,
		0.998331,
		0.996257,
		0.992789,
		0.987511,
		0.980016,
		0.969911,
		0.956839,
		0.940484,
		0.920582,
		0.896922,
		0.869359,
		0.837812,
		0.80227,
		0.762798,
		0.719536,
		0.672708,
		0.622621,
		0.569676,
		0.514364,
		0.457278,
		0.399117,
		0.340688,
		0.282919,
		0.226862,
		0.1737,
		0.12476,
		0.081516,
		0.045605,
		0.018832,
		0.003183,
		0.001428,
		0.005649,
		0.01256,
		0.022059,
		0.034038,
		0.048385,
		0.064984,
		0.083718,
		0.104462,
		0.127091,
		0.151474,
		0.177478,
		0.204965,
		0.233795,
		0.263824,
		0.294906,
		0.326889,
		0.359622,
		0.39295,
		0.426716,
		0.460761,
		0.494924,
		0.529046,
		0.562964,
		0.596519,
		0.629549,
		0.661898,
		0.69341,
		0.723934,
		0.753323,
		0.781437,
		0.808146,
		0.833326,
		0.856866,
		0.87867,
		0.898654,
		0.916757,
		0.932935,
		0.94717,
		0.959469,
		0.96987,
		0.97844,
		0.985283,
		0.990536,
		0.994368,
		0.996984,
		0.99861,
		0.999493,
		0.999877,
		0.999988,
		1,
		0.999998,
		0.999939,
		0.99966,
		0.998895,
		0.997312,
		0.994532,
		0.990153,
		0.983772,
		0.974996,
		0.963459,
		0.948834,
		0.930835,
		0.909232,
		0.883851,
		0.854579,
		0.82137,
		0.78425,
		0.743316,
		0.698742,
		0.650784,
		0.599783,
		0.546167,
		0.49046,
		0.433283,
		0.375363,
		0.317536,
		0.260756,
		0.2061,
		0.154777,
		0.108136,
		0.067676,
		0.035053,
		0.012092,
		0.000798,
		0.005562,
		0.01237,
		0.021727,
		0.033529,
		0.047667,
		0.064028,
		0.082496,
		0.102952,
		0.125272,
		0.14933,
		0.174994,
		0.202132,
		0.230606,
		0.260276,
		0.290998,
		0.322628,
		0.355015,
		0.38801,
		0.421459,
		0.455206,
		0.489096,
		0.522972,
		0.556677,
		0.590052,
		0.622942,
		0.655192,
		0.686649,
		0.717165,
		0.746597,
		0.774805,
		0.801659,
		0.827036,
		0.850827,
		0.872931,
		0.893264,
		0.91176,
		0.928371,
		0.943071,
		0.955861,
		0.966766,
		0.975845,
		0.983187,
		0.988913,
		0.993181,
		0.996177,
		0.998117,
		0.999236,
		0.999775,
		0.999965,
		0.999999,
		1,
		0.999978,
		0.999819,
		0.9993,
		0.998123,
		0.995933,
		0.992346,
		0.986964,
		0.9794,
		0.969281,
		0.956268,
		0.940063,
		0.920413,
		0.897121,
		0.870045,
		0.839109,
		0.804301,
		0.765677,
		0.723368,
		0.67758,
		0.628601,
		0.5768,
		0.522638,
		0.466665,
		0.409531,
		0.35199,
		0.294906,
		0.239256,
		0.186142,
		0.136799,
		0.092596,
		0.055053,
		0.025848,
		0.006822,
		0.021402,
		0.033031,
		0.046964,
		0.063092,
		0.081301,
		0.101474,
		0.123492,
		0.14723,
		0.172561,
		0.199356,
		0.227479,
		0.256796,
		0.287165,
		0.318445,
		0.350491,
		0.383156,
		0.416289,
		0.449741,
		0.483358,
		0.516987,
		0.550474,
		0.583667,
		0.61641,
		0.648554,
		0.679948,
		0.710447,
		0.739908,
		0.768196,
		0.795181,
		0.820741,
		0.844764,
		0.867151,
		0.887815,
		0.906686,
		0.923712,
		0.938862,
		0.952128,
		0.963527,
		0.973107,
		0.980944,
		0.987147,
		0.991857,
		0.995249,
		0.997523,
		0.998903,
		0.999625,
		0.99992,
		0.999995,
		1,
		0.999994,
		0.999914,
		0.999581,
		0.998733,
		0.997042,
		0.994144,
		0.989654,
		0.983187,
		0.974369,
		0.962855,
		0.948332,
		0.930531,
		0.909232,
		0.88427,
		0.855538,
		0.82299,
		0.786646,
		0.746597,
		0.703003,
		0.656102,
		0.606211,
		0.553728,
		0.499142,
		0.443032,
		0.386075,
		0.32905,
		0.272844,
		0.218463,
		0.16703,
		0.119801,
		0.07817,
		0.043674,
		0.018011,
		0.003041,
		0.046277,
		0.062177,
		0.080132,
		0.100028,
		0.121749,
		0.145174,
		0.170178,
		0.196635,
		0.224414,
		0.253383,
		0.283404,
		0.31434,
		0.346048,
		0.378386,
		0.411206,
		0.444363,
		0.477707,
		0.511089,
		0.544357,
		0.577362,
		0.609954,
		0.641985,
		0.673308,
		0.70378,
		0.73326,
		0.761615,
		0.788716,
		0.814442,
		0.838682,
		0.861335,
		0.882312,
		0.901541,
		0.918966,
		0.934549,
		0.948277,
		0.960158,
		0.97023,
		0.978558,
		0.985237,
		0.990397,
		0.994196,
		0.996822,
		0.998486,
		0.999417,
		0.999844,
		0.99998,
		1,
		0.999999,
		0.999965,
		0.999767,
		0.999178,
		0.997904,
		0.995601,
		0.991898,
		0.986419,
		0.97879,
		0.968661,
		0.95571,
		0.939652,
		0.920249,
		0.897313,
		0.870708,
		0.840361,
		0.806258,
		0.76845,
		0.727057,
		0.682272,
		0.63436,
		0.583667,
		0.53062,
		0.475734,
		0.419614,
		0.362964,
		0.306588,
		0.251397,
		0.198419,
		0.148801,
		0.103819,
		0.064888,
		0.033565,
		0.011564,
		0.000762,
		0.078987,
		0.098612,
		0.120042,
		0.14316,
		0.167843,
		0.193969,
		0.221409,
		0.250035,
		0.279714,
		0.310309,
		0.341684,
		0.373698,
		0.406208,
		0.439072,
		0.472143,
		0.505276,
		0.538323,
		0.571138,
		0.603574,
		0.635486,
		0.66673,
		0.697165,
		0.726654,
		0.755064,
		0.782268,
		0.808146,
		0.832587,
		0.855488,
		0.876762,
		0.896331,
		0.914137,
		0.930138,
		0.944313,
		0.956664,
		0.967218,
		0.976031,
		0.983187,
		0.9888,
		0.993016,
		0.996009,
		0.997978,
		0.999142,
		0.999727,
		0.99995,
		0.999998,
		1,
		0.999989,
		0.999882,
		0.999493,
		0.99856,
		0.996763,
		0.99375,
		0.989154,
		0.982605,
		0.973751,
		0.962262,
		0.947842,
		0.930235,
		0.909232,
		0.884676,
		0.856465,
		0.824552,
		0.788955,
		0.749757,
		0.707107,
		0.661224,
		0.612405,
		0.561021,
		0.507526,
		0.452462,
		0.39646,
		0.340247,
		0.28465,
		0.230606,
		0.179163,
		0.13149,
		0.088884,
		0.052777,
		0.024747,
		0.006524,
		0.118371,
		0.141187,
		0.165555,
		0.191355,
		0.218463,
		0.246752,
		0.276092,
		0.306352,
		0.337397,
		0.369091,
		0.401293,
		0.433865,
		0.466665,
		0.499548,
		0.532373,
		0.564995,
		0.59727,
		0.629057,
		0.660215,
		0.690605,
		0.720092,
		0.748546,
		0.77584,
		0.801855,
		0.826481,
		0.849616,
		0.871169,
		0.891062,
		0.909232,
		0.925634,
		0.940242,
		0.95305,
		0.964076,
		0.973368,
		0.980996,
		0.987065,
		0.991705,
		0.995079,
		0.997372,
		0.998793,
		0.999561,
		0.999894,
		0.99999,
		1,
		0.999998,
		0.999948,
		0.999707,
		0.999046,
		0.997675,
		0.99526,
		0.991447,
		0.985874,
		0.978187,
		0.968053,
		0.955164,
		0.939252,
		0.92009,
		0.897499,
		0.87135,
		0.84157,
		0.808146,
		0.771123,
		0.730613,
		0.686792,
		0.63991,
		0.590287,
		0.538323,
		0.484498,
		0.429378,
		0.373618,
		0.317968,
		0.26328,
		0.210511,
		0.160731,
		0.11513,
		0.075024,
		0.041863,
		0.017243,
		0.002907,
		0.163312,
		0.188793,
		0.215573,
		0.24353,
		0.272538,
		0.302467,
		0.333186,
		0.364563,
		0.39646,
		0.428742,
		0.46127,
		0.493904,
		0.526505,
		0.55893,
		0.591042,
		0.622699,
		0.653763,
		0.6841,
		0.713576,
		0.742062,
		0.769433,
		0.795573,
		0.82037,
		0.843722,
		0.865538,
		0.885739,
		0.904257,
		0.921045,
		0.93607,
		0.949321,
		0.96081,
		0.970571,
		0.978669,
		0.985193,
		0.990264,
		0.994029,
		0.996662,
		0.998361,
		0.999337,
		0.999805,
		0.99997,
		0.999999,
		1,
		0.999981,
		0.999844,
		0.999396,
		0.998376,
		0.996474,
		0.99335,
		0.988652,
		0.982029,
		0.973143,
		0.961681,
		0.947363,
		0.929947,
		0.909232,
		0.88507,
		0.85736,
		0.82606,
		0.791183,
		0.752804,
		0.711062,
		0.666161,
		0.618377,
		0.568056,
		0.515624,
		0.461585,
		0.406528,
		0.351133,
		0.296172,
		0.242518,
		0.191151,
		0.14316,
		0.099754,
		0.062267,
		0.032169,
		0.011069,
		0.000729,
		0.212739,
		0.24037,
		0.269049,
		0.298652,
		0.32905,
		0.360112,
		0.391707,
		0.423701,
		0.455958,
		0.488343,
		0.520718,
		0.552945,
		0.584888,
		0.61641,
		0.647376,
		0.677651,
		0.707107,
		0.735615,
		0.763052,
		0.789303,
		0.814257,
		0.837812,
		0.859876,
		0.880367,
		0.899218,
		0.916375,
		0.931803,
		0.945484,
		0.957423,
		0.967646,
		0.976208,
		0.983187,
		0.988691,
		0.992856,
		0.995843,
		0.997839,
		0.999046,
		0.999675,
		0.999931,
		0.999995,
		1,
		0.999995,
		0.999926,
		0.999638,
		0.998903,
		0.997435,
		0.994912,
		0.990993,
		0.985332,
		0.977591,
		0.967454,
		0.95463,
		0.938862,
		0.919935,
		0.897679,
		0.87197,
		0.842738,
		0.809968,
		0.773702,
		0.73404,
		0.69115,
		0.645261,
		0.596674,
		0.545761,
		0.492972,
		0.438833,
		0.383959,
		0.32905,
		0.2749,
		0.222405,
		0.172561,
		0.12648,
		0.085389,
		0.050639,
		0.023716,
		0.006244,
		0.265625,
		0.294906,
		0.324985,
		0.355737,
		0.387032,
		0.41874,
		0.450728,
		0.482863,
		0.515011,
		0.547038,
		0.57881,
		0.610192,
		0.641052,
		0.67126,
		0.700686,
		0.729206,
		0.756699,
		0.783049,
		0.808146,
		0.831889,
		0.854185,
		0.874952,
		0.894119,
		0.911631,
		0.927446,
		0.941544,
		0.953921,
		0.964597,
		0.973615,
		0.981047,
		0.986986,
		0.991558,
		0.994912,
		0.997222,
		0.998681,
		0.999493,
		0.999864,
		0.999983,
		1,
		0.999999,
		0.99997,
		0.999798,
		0.999289,
		0.998182,
		0.996177,
		0.992945,
		0.98815,
		0.981456,
		0.972543,
		0.961112,
		0.946896,
		0.929666,
		0.909232,
		0.885451,
		0.858226,
		0.827516,
		0.793332,
		0.755743,
		0.714875,
		0.670922,
		0.624138,
		0.574848,
		0.523449,
		0.470412,
		0.416289,
		0.361715,
		0.307411,
		0.254193,
		0.202976,
		0.154777,
		0.110724,
		0.072063,
		0.040162,
		0.016522,
		0.002783,
		0.320992,
		0.351437,
		0.382435,
		0.413858,
		0.445578,
		0.477463,
		0.509384,
		0.541209,
		0.572806,
		0.604044,
		0.634794,
		0.664927,
		0.694316,
		0.722839,
		0.750376,
		0.776812,
		0.80204,
		0.825958,
		0.848471,
		0.869498,
		0.888967,
		0.906817,
		0.923006,
		0.937506,
		0.950308,
		0.961427,
		0.970896,
		0.978775,
		0.985151,
		0.990135,
		0.993865,
		0.996504,
		0.998234,
		0.999253,
		0.999763,
		0.999956,
		0.999998,
		1,
		0.999991,
		0.999898,
		0.999561,
		0.99875,
		0.997186,
		0.994557,
		0.990536,
		0.984791,
		0.977002,
		0.966867,
		0.954108,
		0.938482,
		0.919785,
		0.897854,
		0.872571,
		0.843868,
		0.811728,
		0.77619,
		0.737347,
		0.695352,
		0.650423,
		0.602837,
		0.552945,
		0.501165,
		0.447991,
		0.393997,
		0.339837,
		0.286256,
		0.234088,
		0.184267,
		0.137829,
		0.09592,
		0.059802,
		0.030859,
		0.010606,
		0.000697,
		0.377913,
		0.409054,
		0.440506,
		0.472143,
		0.503836,
		0.535457,
		0.566876,
		0.597967,
		0.628601,
		0.658652,
		0.687996,
		0.716513,
		0.744084,
		0.770596,
		0.795942,
		0.820021,
		0.842738,
		0.864011,
		0.883765,
		0.901939,
		0.918486,
		0.933375,
		0.946591,
		0.958141,
		0.968053,
		0.976376,
		0.983187,
		0.988586,
		0.9927,
		0.995681,
		0.9977,
		0.998946,
		0.999618,
		0.999908,
		0.999991,
		1,
		0.999998,
		0.999955,
		0.999746,
		0.999172,
		0.997978,
		0.995871,
		0.992536,
		0.987648,
		0.980889,
		0.971951,
		0.960553,
		0.94644,
		0.929393,
		0.909232,
		0.88582,
		0.859064,
		0.828924,
		0.795408,
		0.758579,
		0.718555,
		0.675515,
		0.629697,
		0.581406,
		0.531011,
		0.478955,
		0.425753,
		0.371999,
		0.318369,
		0.265625,
		0.214622,
		0.166312,
		0.121749,
		0.082095,
		0.048628,
		0.022747,
		0.005983,
		0.435513,
		0.466901,
		0.498366,
		0.529781,
		0.561021,
		0.59196,
		0.622473,
		0.652437,
		0.681729,
		0.710231,
		0.737827,
		0.764404,
		0.789856,
		0.814083,
		0.83699,
		0.858494,
		0.878519,
		0.897002,
		0.913893,
		0.929157,
		0.942774,
		0.954745,
		0.96509,
		0.973851,
		0.981094,
		0.986911,
		0.991416,
		0.994749,
		0.997073,
		0.998566,
		0.999422,
		0.999831,
		0.999974,
		0.999999,
		1,
		0.999984,
		0.999864,
		0.999475,
		0.998588,
		0.996928,
		0.994196,
		0.990077,
		0.984253,
		0.97642,
		0.966289,
		0.953597,
		0.938112,
		0.919639,
		0.898023,
		0.873152,
		0.84496,
		0.813428,
		0.778592,
		0.740538,
		0.699408,
		0.655405,
		0.608789,
		0.559887,
		0.50909,
		0.456863,
		0.403741,
		0.350337,
		0.297347,
		0.245554,
		0.19583,
		0.149143,
		0.106564,
		0.069274,
		0.038563,
		0.015846,
		0.002666,
		0.492972,
		0.52418,
		0.555238,
		0.586022,
		0.61641,
		0.646281,
		0.675515,
		0.703994,
		0.731605,
		0.758237,
		0.783784,
		0.808146,
		0.83123,
		0.852952,
		0.873234,
		0.892012,
		0.909232,
		0.924857,
		0.938862,
		0.951243,
		0.962012,
		0.971204,
		0.978877,
		0.985111,
		0.990011,
		0.993706,
		0.996348,
		0.998106,
		0.999166,
		0.999716,
		0.99994,
		0.999996,
		1,
		0.999996,
		0.999935,
		0.999686,
		0.999046,
		0.997765,
		0.995558,
		0.992122,
		0.987147,
		0.980326,
		0.971369,
		0.960006,
		0.945995,
		0.929128,
		0.909232,
		0.886177,
		0.859876,
		0.830285,
		0.797413,
		0.761317,
		0.722107,
		0.679948,
		0.635065,
		0.587741,
		0.538323,
		0.487226,
		0.43493,
		0.381994,
		0.32905,
		0.276811,
		0.226079,
		0.177742,
		0.132788,
		0.092302,
		0.057479,
		0.029626,
		0.010171,
		0.000668,
		0.549528,
		0.580154,
		0.610413,
		0.640185,
		0.669354,
		0.697803,
		0.72542,
		0.752097,
		0.777728,
		0.802214,
		0.825463,
		0.847388,
		0.867913,
		0.886971,
		0.904508,
		0.92048,
		0.93486,
		0.947639,
		0.958822,
		0.968439,
		0.976536,
		0.983187,
		0.988485,
		0.992549,
		0.995521,
		0.997561,
		0.998845,
		0.999557,
		0.999882,
		0.999985,
		1,
		0.999999,
		0.999974,
		0.999825,
		0.999381,
		0.998415,
		0.996662,
		0.993829,
		0.989616,
		0.983719,
		0.975845,
		0.965721,
		0.953097,
		0.937751,
		0.919497,
		0.898188,
		0.873716,
		0.846017,
		0.815072,
		0.780913,
		0.743619,
		0.703324,
		0.660215,
		0.614537,
		0.566596,
		0.516759,
		0.465459,
		0.413199,
		0.360554,
		0.308175,
		0.256796,
		0.207232,
		0.160392,
		0.117276,
		0.078987,
		0.046734,
		0.021837,
		0.005737,
		0.604481,
		0.63415,
		0.663248,
		0.691659,
		0.719275,
		0.745987,
		0.771692,
		0.796291,
		0.819691,
		0.841807,
		0.862562,
		0.881887,
		0.899725,
		0.916031,
		0.930774,
		0.943938,
		0.955526,
		0.965558,
		0.974075,
		0.98114,
		0.986838,
		0.991278,
		0.99459,
		0.996925,
		0.998451,
		0.999346,
		0.999793,
		0.999962,
		0.999998,
		1,
		0.999992,
		0.999911,
		0.999618,
		0.99891,
		0.997543,
		0.995238,
		0.991705,
		0.986646,
		0.979768,
		0.970795,
		0.959469,
		0.94556,
		0.928869,
		0.909232,
		0.886525,
		0.860662,
		0.831602,
		0.799352,
		0.763962,
		0.725537,
		0.68423,
		0.64025,
		0.593863,
		0.545395,
		0.495233,
		0.44383,
		0.391707,
		0.339457,
		0.28775,
		0.237335,
		0.189047,
		0.143807,
		0.102633,
		0.066642,
		0.037057,
		0.015211,
		0.002556,
		0.657196,
		0.685564,
		0.71317,
		0.739908,
		0.765677,
		0.790377,
		0.813918,
		0.836212,
		0.857183,
		0.876762,
		0.894888,
		0.911514,
		0.926607,
		0.940146,
		0.952128,
		0.962567,
		0.971498,
		0.978974,
		0.985072,
		0.989891,
		0.993551,
		0.996194,
		0.997978,
		0.999076,
		0.999666,
		0.99992,
		0.999992,
		1,
		0.999998,
		0.999961,
		0.999779,
		0.999278,
		0.998234,
		0.996388,
		0.993458,
		0.989154,
		0.983187,
		0.975277,
		0.965164,
		0.952607,
		0.937398,
		0.919359,
		0.898348,
		0.874262,
		0.84704,
		0.816662,
		0.783156,
		0.746597,
		0.707107,
		0.664862,
		0.620093,
		0.573084,
		0.52418,
		0.473788,
		0.42238,
		0.370494,
		0.318742,
		0.267809,
		0.218463,
		0.171552,
		0.128015,
		0.088884,
		0.055289,
		0.028466,
		0.009762,
		0.000641,
		0.707107,
		0.733863,
		0.759685,
		0.784477,
		0.808146,
		0.830607,
		0.851782,
		0.8716,
		0.890001,
		0.906935,
		0.922365,
		0.936266,
		0.948631,
		0.959469,
		0.968806,
		0.976689,
		0.983187,
		0.988388,
		0.992403,
		0.995364,
		0.997422,
		0.998741,
		0.999493,
		0.999852,
		0.999977,
		1,
		1,
		0.999986,
		0.999882,
		0.999542,
		0.998766,
		0.997312,
		0.994912,
		0.991285,
		0.986146,
		0.979216,
		0.97023,
		0.958943,
		0.945135,
		0.928617,
		0.909232,
		0.886861,
		0.861423,
		0.832877,
		0.801227,
		0.76652,
		0.728851,
		0.688367,
		0.645261,
		0.599783,
		0.552238,
		0.502989,
		0.452462,
		0.401146,
		0.349596,
		0.298442,
		0.248386,
		0.200209,
		0.154777,
		0.113042,
		0.076051,
		0.044948,
		0.02098,
		0.005506,
		0.753719,
		0.778592,
		0.802379,
		0.824995,
		0.846361,
		0.866407,
		0.88507,
		0.902298,
		0.918052,
		0.932304,
		0.945041,
		0.956268,
		0.966004,
		0.974289,
		0.981184,
		0.986769,
		0.991144,
		0.994434,
		0.996778,
		0.998334,
		0.999268,
		0.999752,
		0.999947,
		0.999996,
		1,
		0.999996,
		0.999944,
		0.999726,
		0.999166,
		0.998044,
		0.996106,
		0.993081,
		0.988691,
		0.982658,
		0.974716,
		0.964615,
		0.952128,
		0.937054,
		0.919224,
		0.898503,
		0.874792,
		0.848031,
		0.818201,
		0.785326,
		0.749475,
		0.710763,
		0.669354,
		0.625463,
		0.579359,
		0.531365,
		0.481862,
		0.431293,
		0.380165,
		0.32905,
		0.278592,
		0.229511,
		0.182603,
		0.138748,
		0.098913,
		0.064157,
		0.035637,
		0.014613,
		0.002453,
		0.796619,
		0.819379,
		0.840925,
		0.861185,
		0.880097,
		0.897607,
		0.913673,
		0.928264,
		0.941362,
		0.952968,
		0.963095,
		0.971777,
		0.979066,
		0.985035,
		0.989775,
		0.9934,
		0.996042,
		0.99785,
		0.998984,
		0.999612,
		0.999897,
		0.999987,
		1,
		1,
		0.999977,
		0.999847,
		0.999459,
		0.998613,
		0.997073,
		0.99458,
		0.990862,
		0.985648,
		0.978669,
		0.969673,
		0.958426,
		0.944719,
		0.928371,
		0.909232,
		0.887188,
		0.862161,
		0.834112,
		0.803041,
		0.768993,
		0.732056,
		0.692366,
		0.650105,
		0.605508,
		0.55886,
		0.510503,
		0.460836,
		0.410318,
		0.359472,
		0.308888,
		0.259225,
		0.211216,
		0.165674,
		0.123492,
		0.08565,
		0.053221,
		0.027373,
		0.009378,
		0.000615,
		0.835475,
		0.855939,
		0.875088,
		0.892867,
		0.909232,
		0.924149,
		0.937598,
		0.949573,
		0.960083,
		0.969156,
		0.976835,
		0.983187,
		0.988294,
		0.992261,
		0.995211,
		0.997284,
		0.998635,
		0.999426,
		0.999819,
		0.999967,
		0.999999,
		1,
		0.999993,
		0.999923,
		0.999666,
		0.999046,
		0.997845,
		0.995817,
		0.9927,
		0.988228,
		0.982133,
		0.974162,
		0.964076,
		0.951659,
		0.936718,
		0.919093,
		0.898654,
		0.875306,
		0.848992,
		0.819691,
		0.787425,
		0.752258,
		0.714298,
		0.673697,
		0.630658,
		0.585432,
		0.538323,
		0.48969,
		0.439948,
		0.389573,
		0.339103,
		0.289144,
		0.24037,
		0.193529,
		0.149448,
		0.109032,
		0.073275,
		0.043262,
		0.020173,
		0.005289,
		0.870045,
		0.888082,
		0.904734,
		0.919966,
		0.933754,
		0.946088,
		0.956972,
		0.966428,
		0.974494,
		0.981226,
		0.986701,
		0.991015,
		0.994282,
		0.996633,
		0.998216,
		0.999187,
		0.999707,
		0.99993,
		0.999993,
		1,
		0.999999,
		0.999966,
		0.999806,
		0.999367,
		0.998451,
		0.996826,
		0.994242,
		0.990437,
		0.985151,
		0.978128,
		0.969125,
		0.95792,
		0.944313,
		0.928131,
		0.909232,
		0.887506,
		0.862877,
		0.835308,
		0.804797,
		0.771385,
		0.735156,
		0.696234,
		0.654791,
		0.611048,
		0.565272,
		0.517785,
		0.468961,
		0.419232,
		0.369091,
		0.319089,
		0.269849,
		0.222058,
		0.176478,
		0.133949,
		0.095391,
		0.061808,
		0.034297,
		0.014049,
		0.002356,
		0.900183,
		0.915718,
		0.929834,
		0.942516,
		0.953766,
		0.963598,
		0.972044,
		0.979155,
		0.984999,
		0.989663,
		0.993253,
		0.995893,
		0.997721,
		0.998889,
		0.999554,
		0.99987,
		0.99998,
		1,
		1,
		0.999988,
		0.999897,
		0.999599,
		0.998917,
		0.997637,
		0.995521,
		0.992316,
		0.987764,
		0.981612,
		0.973615,
		0.963547,
		0.951199,
		0.93639,
		0.918966,
		0.898801,
		0.875806,
		0.849923,
		0.821134,
		0.789457,
		0.754952,
		0.717717,
		0.677898,
		0.635684,
		0.591311,
		0.545064,
		0.497281,
		0.448352,
		0.398725,
		0.348906,
		0.299464,
		0.251033,
		0.204316,
		0.160089,
		0.119202,
		0.082589,
		0.051267,
		0.026342,
		0.009016,
		0.00059,
		0.925841,
		0.938862,
		0.950468,
		0.960668,
		0.969489,
		0.976975,
		0.983187,
		0.988203,
		0.992122,
		0.99506,
		0.997147,
		0.998528,
		0.999355,
		0.999782,
		0.999954,
		0.999997,
		1,
		0.999997,
		0.999951,
		0.99976,
		0.999268,
		0.99828,
		0.996571,
		0.993899,
		0.990011,
		0.984657,
		0.977591,
		0.968585,
		0.957423,
		0.943915,
		0.927897,
		0.909232,
		0.887815,
		0.863572,
		0.836467,
		0.806498,
		0.773702,
		0.738155,
		0.699976,
		0.659326,
		0.61641,
		0.571482,
		0.524843,
		0.476846,
		0.427896,
		0.378457,
		0.32905,
		0.280256,
		0.232725,
		0.187172,
		0.144385,
		0.105229,
		0.070648,
		0.041669,
		0.019411,
		0.005084,
		0.947082,
		0.957642,
		0.966832,
		0.974689,
		0.981267,
		0.986637,
		0.99089,
		0.994133,
		0.99649,
		0.998097,
		0.999103,
		0.999658,
		0.999909,
		0.999989,
		1,
		1,
		0.99998,
		0.999866,
		0.999525,
		0.99878,
		0.997422,
		0.995219,
		0.991928,
		0.987301,
		0.981094,
		0.973075,
		0.963026,
		0.950749,
		0.93607,
		0.918841,
		0.898944,
		0.87629,
		0.850827,
		0.822533,
		0.791425,
		0.757559,
		0.721027,
		0.681965,
		0.640549,
		0.597004,
		0.551596,
		0.504643,
		0.456513,
		0.407628,
		0.358463,
		0.309553,
		0.261496,
		0.214952,
		0.170651,
		0.129392,
		0.092052,
		0.059585,
		0.033031,
		0.013517,
		0.002265,
		0.964076,
		0.972299,
		0.97924,
		0.984964,
		0.989554,
		0.993109,
		0.995745,
		0.997593,
		0.998793,
		0.999493,
		0.99984,
		0.999971,
		0.999999,
		1,
		0.999994,
		0.999932,
		0.999707,
		0.999161,
		0.998102,
		0.99631,
		0.993551,
		0.989583,
		0.984164,
		0.977061,
		0.968053,
		0.956935,
		0.943527,
		0.927669,
		0.909232,
		0.888115,
		0.864247,
		0.837591,
		0.808146,
		0.775945,
		0.741059,
		0.703599,
		0.663715,
		0.621603,
		0.577498,
		0.531686,
		0.484498,
		0.436317,
		0.387578,
		0.338772,
		0.290446,
		0.243212,
		0.197741,
		0.154777,
		0.11513,
		0.079688,
		0.049418,
		0.025367,
		0.008674,
		0.000567,
		0.977109,
		0.983187,
		0.988116,
		0.991988,
		0.994912,
		0.997011,
		0.99842,
		0.999281,
		0.999741,
		0.999938,
		0.999994,
		1,
		0.999999,
		0.99997,
		0.99983,
		0.999443,
		0.998635,
		0.997199,
		0.994912,
		0.991538,
		0.986838,
		0.980581,
		0.972543,
		0.962515,
		0.950308,
		0.935757,
		0.91872,
		0.899083,
		0.876762,
		0.851704,
		0.823889,
		0.793332,
		0.760084,
		0.724231,
		0.685901,
		0.645261,
		0.602518,
		0.557927,
		0.511786,
		0.464441,
		0.416289,
		0.367779,
		0.319414,
		0.271757,
		0.225429,
		0.181117,
		0.139573,
		0.101621,
		0.068159,
		0.040162,
		0.018691,
		0.004891,
		0.986574,
		0.990768,
		0.993987,
		0.996348,
		0.997978,
		0.999017,
		0.999606,
		0.999886,
		0.999982,
		1,
		1,
		0.999989,
		0.999909,
		0.999647,
		0.999046,
		0.997915,
		0.996042,
		0.993199,
		0.989154,
		0.983674,
		0.976536,
		0.967529,
		0.956457,
		0.943146,
		0.927446,
		0.909232,
		0.888407,
		0.864902,
		0.838682,
		0.809744,
		0.778118,
		0.743871,
		0.707107,
		0.667967,
		0.626633,
		0.583329,
		0.538323,
		0.491928,
		0.444504,
		0.39646,
		0.34826,
		0.300419,
		0.253513,
		0.208175,
		0.165103,
		0.125061,
		0.088884,
		0.057479,
		0.031834,
		0.013016,
		0.002179,
		0.992969,
		0.995601,
		0.997465,
		0.998695,
		0.999429,
		0.999807,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999957,
		0.999788,
		0.999355,
		0.998482,
		0.99697,
		0.994599,
		0.991144,
		0.986377,
		0.980072,
		0.972017,
		0.962012,
		0.949876,
		0.935451,
		0.918602,
		0.899218,
		0.877219,
		0.852555,
		0.825205,
		0.795181,
		0.762531,
		0.727336,
		0.689715,
		0.649825,
		0.607863,
		0.564067,
		0.518718,
		0.472143,
		0.424716,
		0.37686,
		0.32905,
		0.281814,
		0.23574,
		0.191473,
		0.149721,
		0.111261,
		0.076937,
		0.047667,
		0.024446,
		0.008351,
		0.000546,
		0.996876,
		0.99831,
		0.999205,
		0.999698,
		0.99992,
		0.99999,
		1,
		1,
		0.999982,
		0.999882,
		0.999581,
		0.998923,
		0.997721,
		0.995768,
		0.992844,
		0.988724,
		0.983187,
		0.976017,
		0.967013,
		0.955987,
		0.942774,
		0.927229,
		0.909232,
		0.88869,
		0.865538,
		0.839741,
		0.811294,
		0.780225,
		0.746597,
		0.710506,
		0.672085,
		0.631507,
		0.588982,
		0.544762,
		0.499142,
		0.452462,
		0.405109,
		0.357518,
		0.310176,
		0.263625,
		0.218463,
		0.175346,
		0.134994,
		0.098193,
		0.065798,
		0.038736,
		0.018011,
		0.004709,
		0.998929,
		0.999551,
		0.999859,
		0.999974,
		0.999999,
		1,
		0.999995,
		0.99994,
		0.999741,
		0.999259,
		0.998321,
		0.996733,
		0.994282,
		0.990749,
		0.985916,
		0.979567,
		0.971498,
		0.961517,
		0.949453,
		0.935153,
		0.918486,
		0.89935,
		0.877665,
		0.853382,
		0.826481,
		0.796974,
		0.764902,
		0.730344,
		0.69341,
		0.654248,
		0.613044,
		0.570021,
		0.525446,
		0.479627,
		0.432915,
		0.385711,
		0.338462,
		0.291667,
		0.245878,
		0.201706,
		0.159817,
		0.120941,
		0.085875,
		0.055483,
		0.0307,
		0.012541,
		0.002098,
		0.999771,
		0.999945,
		0.999995,
		1,
		0.999999,
		0.999973,
		0.99985,
		0.999508,
		0.998793,
		0.99752,
		0.995488,
		0.992484,
		0.988294,
		0.982702,
		0.975504,
		0.966504,
		0.955526,
		0.94241,
		0.927017,
		0.909232,
		0.888967,
		0.866157,
		0.840769,
		0.812798,
		0.782268,
		0.749239,
		0.7138,
		0.676077,
		0.636233,
		0.594464,
		0.55101,
		0.506149,
		0.460201,
		0.413531,
		0.366551,
		0.319719,
		0.273547,
		0.228598,
		0.185491,
		0.144906,
		0.107582,
		0.074325,
		0.046007,
		0.023574,
		0.008046,
		0.000525,
		0.999985,
		1,
		1,
		0.99999,
		0.99992,
		0.999688,
		0.999156,
		0.998153,
		0.99649,
		0.993959,
		0.990352,
		0.985457,
		0.979066,
		0.970985,
		0.961031,
		0.949038,
		0.93486,
		0.918374,
		0.899478,
		0.878098,
		0.854185,
		0.827721,
		0.798713,
		0.767202,
		0.73326,
		0.696992,
		0.658536,
		0.618068,
		0.575799,
		0.531979,
		0.4869,
		0.440894,
		0.394338,
		0.347655,
		0.301315,
		0.255841,
		0.211807,
		0.169842,
		0.130634,
		0.094935,
		0.063557,
		0.037384,
		0.017367,
		0.004537,
		1,
		0.999998,
		0.999962,
		0.999813,
		0.999429,
		0.998655,
		0.997312,
		0.995203,
		0.992122,
		0.987863,
		0.98222,
		0.974996,
		0.966004,
		0.955074,
		0.942053,
		0.926809,
		0.909232,
		0.889236,
		0.866759,
		0.841768,
		0.814257,
		0.78425,
		0.751801,
		0.716994,
		0.679948,
		0.640815,
		0.599783,
		0.557076,
		0.512956,
		0.467727,
		0.421733,
		0.375363,
		0.32905,
		0.283276,
		0.238574,
		0.195527,
		0.154777,
		0.11702,
		0.083016,
		0.053588,
		0.029626,
		0.012092,
		0.002021,
		0.999985,
		0.999896,
		0.999629,
		0.999046,
		0.997978,
		0.99624,
		0.993633,
		0.989954,
		0.984999,
		0.97857,
		0.97048,
		0.960553,
		0.948631,
		0.934575,
		0.918264,
		0.899603,
		0.878519,
		0.854966,
		0.828924,
		0.800401,
		0.769433,
		0.736089,
		0.700466,
		0.662695,
		0.622942,
		0.581406,
		0.538323,
		0.493969,
		0.448658,
		0.402746,
		0.356632,
		0.310761,
		0.265625,
		0.221767,
		0.179781,
		0.140317,
		0.104082,
		0.071843,
		0.044432,
		0.022747,
		0.007758,
		0.000506,
		0.999771,
		0.999343,
		0.998511,
		0.997097,
		0.994912,
		0.991758,
		0.987433,
		0.981742,
		0.974494,
		0.965511,
		0.95463,
		0.941704,
		0.926607,
		0.909232,
		0.889497,
		0.867344,
		0.842738,
		0.815675,
		0.786174,
		0.754286,
		0.720092,
		0.683702,
		0.645261,
		0.604944,
		0.562964,
		0.51957,
		0.475047,
		0.429721,
		0.383959,
		0.338171,
		0.292812,
		0.248386,
		0.205443,
		0.164588,
		0.12648,
		0.091835,
		0.061429,
		0.036101,
		0.016757,
		0.004374,
		0.998929,
		0.997796,
		0.995985,
		0.993303,
		0.989554,
		0.984543,
		0.978079,
		0.969981,
		0.960083,
		0.948233,
		0.934295,
		0.918157,
		0.899725,
		0.878929,
		0.855726,
		0.830094,
		0.80204,
		0.771599,
		0.738833,
		0.703836,
		0.66673,
		0.627672,
		0.586849,
		0.544486,
		0.500843,
		0.456215,
		0.410941,
		0.365398,
		0.320005,
		0.275228,
		0.23158,
		0.189623,
		0.149968,
		0.113285,
		0.080297,
		0.051789,
		0.028608,
		0.011667,
		0.001948,
		0.996876,
		0.994617,
		0.99139,
		0.987004,
		0.981267,
		0.973997,
		0.965026,
		0.954194,
		0.941362,
		0.926409,
		0.909232,
		0.889753,
		0.867913,
		0.843682,
		0.817052,
		0.788042,
		0.756699,
		0.723099,
		0.687346,
		0.649575,
		0.609954,
		0.568684,
		0.525998,
		0.482168,
		0.437501,
		0.392345,
		0.347086,
		0.302157,
		0.258031,
		0.21523,
		0.174326,
		0.135939,
		0.100747,
		0.069482,
		0.042936,
		0.021964,
		0.007484,
		0.000488,
		0.992969,
		0.989154,
		0.984089,
		0.977591,
		0.969489,
		0.959621,
		0.947842,
		0.934022,
		0.918052,
		0.899843,
		0.879329,
		0.856465,
		0.83123,
		0.803632,
		0.773702,
		0.741497,
		0.707107,
		0.670646,
		0.632263,
		0.592135,
		0.550474,
		0.507526,
		0.463572,
		0.418929,
		0.373956,
		0.32905,
		0.28465,
		0.241242,
		0.199356,
		0.159571,
		0.122519,
		0.088884,
		0.059405,
		0.034883,
		0.016179,
		0.004219,
		0.986574,
		0.980795,
		0.973507,
		0.964547,
		0.953766,
		0.941027,
		0.926215,
		0.909232,
		0.890001,
		0.868468,
		0.8446,
		0.81839,
		0.789856,
		0.759042,
		0.726017,
		0.690882,
		0.653763,
		0.61482,
		0.57424,
		0.532247,
		0.489096,
		0.44508,
		0.400525,
		0.355799,
		0.31131,
		0.267506,
		0.224882,
		0.183976,
		0.145377,
		0.109724,
		0.077708,
		0.050078,
		0.027641,
		0.011263,
		0.001879,
		0.977109,
		0.969004,
		0.959167,
		0.947458,
		0.933754,
		0.91795,
		0.899959,
		0.879718,
		0.857183,
		0.832336,
		0.805179,
		0.775744,
		0.744084,
		0.710282,
		0.674449,
		0.636722,
		0.59727,
		0.556295,
		0.514027,
		0.470734,
		0.426716,
		0.382312,
		0.337898,
		0.29389,
		0.250747,
		0.208972,
		0.169111,
		0.131761,
		0.09757,
		0.067236,
		0.041515,
		0.02122,
		0.007225,
		0.000471,
		0.964076,
		0.953345,
		0.940699,
		0.926026,
		0.909232,
		0.890244,
		0.869007,
		0.845492,
		0.819691,
		0.791619,
		0.761317,
		0.728851,
		0.694316,
		0.657831,
		0.619546,
		0.57964,
		0.538323,
		0.495839,
		0.452462,
		0.408505,
		0.364313,
		0.320274,
		0.276811,
		0.234393,
		0.193529,
		0.154777,
		0.118739,
		0.086071,
		0.057479,
		0.033726,
		0.01563,
		0.004073,
		0.947082,
		0.933492,
		0.91785,
		0.900072,
		0.880097,
		0.857883,
		0.833411,
		0.806683,
		0.777728,
		0.746597,
		0.713366,
		0.678141,
		0.641052,
		0.60226,
		0.561953,
		0.520351,
		0.477707,
		0.434306,
		0.390469,
		0.346552,
		0.302949,
		0.260094,
		0.218463,
		0.178574,
		0.140991,
		0.106327,
		0.075242,
		0.048451,
		0.026722,
		0.01088,
		0.001814,
		0.925841,
		0.909232,
		0.89048,
		0.869533,
		0.846361,
		0.820956,
		0.793332,
		0.763528,
		0.731605,
		0.697651,
		0.661782,
		0.624138,
		0.584888,
		0.544233,
		0.502402,
		0.459655,
		0.416289,
		0.372632,
		0.32905,
		0.285944,
		0.243759,
		0.202976,
		0.164123,
		0.127771,
		0.094539,
		0.065096,
		0.040162,
		0.020513,
		0.006979,
		0.000454,
		0.900183,
		0.880467,
		0.858565,
		0.834457,
		0.808146,
		0.779657,
		0.749038,
		0.716363,
		0.681729,
		0.645261,
		0.607109,
		0.567455,
		0.526505,
		0.484498,
		0.441707,
		0.398433,
		0.355015,
		0.311827,
		0.26928,
		0.227824,
		0.18795,
		0.150192,
		0.11513,
		0.083389,
		0.055645,
		0.032625,
		0.015108,
		0.003934,
		0.870045,
		0.847207,
		0.822187,
		0.794999,
		0.765677,
		0.73428,
		0.700893,
		0.665622,
		0.628601,
		0.589991,
		0.549982,
		0.50879,
		0.466665,
		0.423884,
		0.38076,
		0.33764,
		0.294906,
		0.252975,
		0.212308,
		0.173402,
		0.136799,
		0.103084,
		0.072891,
		0.046901,
		0.025848,
		0.010517,
		0.001752,
		0.835475,
		0.809569,
		0.781532,
		0.751412,
		0.719275,
		0.685216,
		0.649351,
		0.611825,
		0.572806,
		0.532493,
		0.491113,
		0.448922,
		0.406208,
		0.363292,
		0.320527,
		0.278304,
		0.237049,
		0.197229,
		0.159349,
		0.123957,
		0.091646,
		0.063057,
		0.038875,
		0.019841,
		0.006746,
		0.000439,
		0.796619,
		0.767766,
		0.736881,
		0.704043,
		0.669354,
		0.63294,
		0.594955,
		0.555576,
		0.515011,
		0.473495,
		0.431293,
		0.388702,
		0.346048,
		0.303696,
		0.262041,
		0.22152,
		0.182603,
		0.145806,
		0.111681,
		0.08083,
		0.053897,
		0.031577,
		0.014613,
		0.003803,
		0.753719,
		0.722107,
		0.688606,
		0.653328,
		0.61641,
		0.578013,
		0.538323,
		0.497557,
		0.455958,
		0.413799,
		0.371384,
		0.32905,
		0.287165,
		0.246136,
		0.206404,
		0.168448,
		0.132788,
		0.099986,
		0.070648,
		0.045425,
		0.025016,
		0.010171,
		0.001693,
		0.707107,
		0.672983,
		0.63716,
		0.599783,
		0.561021,
		0.52107,
		0.480153,
		0.438523,
		0.39646,
		0.354276,
		0.312315,
		0.270954,
		0.230606,
		0.191718,
		0.154777,
		0.120309,
		0.088884,
		0.061111,
		0.037648,
		0.019201,
		0.006524,
		0.000424,
		0.657196,
		0.620871,
		0.583079,
		0.544,
		0.503836,
		0.46282,
		0.42121,
		0.379296,
		0.337397,
		0.295864,
		0.255081,
		0.215467,
		0.177478,
		0.141606,
		0.108384,
		0.078386,
		0.05223,
		0.030578,
		0.014141,
		0.003677,
		0.604481,
		0.566321,
		0.526971,
		0.486644,
		0.445578,
		0.40404,
		0.362328,
		0.320766,
		0.279714,
		0.239562,
		0.200737,
		0.163699,
		0.128949,
		0.097025,
		0.068506,
		0.044016,
		0.024223,
		0.009842,
		0.001638,
		0.549528,
		0.509955,
		0.469512,
		0.428447,
		0.387032,
		0.345573,
		0.304401,
		0.263882,
		0.224414,
		0.18643,
		0.150396,
		0.116819,
		0.086243,
		0.059254,
		0.036479,
		0.018592,
		0.006313,
		0.00041,
		0.492972,
		0.452462,
		0.411446,
		0.370205,
		0.32905,
		0.288319,
		0.248386,
		0.209654,
		0.172561,
		0.137583,
		0.105229,
		0.076051,
		0.050639,
		0.029626,
		0.013691,
		0.003558,
		0.435513,
		0.394596,
		0.353578,
		0.312776,
		0.272538,
		0.233241,
		0.195295,
		0.159146,
		0.125272,
		0.094192,
		0.06646,
		0.042672,
		0.023468,
		0.009529,
		0.001584,
		0.377913,
		0.337168,
		0.29677,
		0.257074,
		0.218463,
		0.181354,
		0.146196,
		0.113477,
		0.083718,
		0.057479,
		0.035363,
		0.018011,
		0.006111,
		0.000397,
		0.320992,
		0.281048,
		0.241943,
		0.204067,
		0.167843,
		0.133727,
		0.102209,
		0.073818,
		0.049119,
		0.028718,
		0.013263,
		0.003444,
		0.265625,
		0.227159,
		0.190067,
		0.154777,
		0.121749,
		0.09148,
		0.064503,
		0.041389,
		0.022747,
		0.00923,
		0.001534,
		0.212739,
		0.176478,
		0.142168,
		0.110275,
		0.081301,
		0.055783,
		0.034297,
		0.017457,
		0.00592,
		0.000384,
		0.163312,
		0.130029,
		0.099316,
		0.071682,
		0.047667,
		0.027851,
		0.012854,
		0.003336,
		0.118371,
		0.088884,
		0.062632,
		0.040162,
		0.022059,
		0.008946,
		0.001486,
		0.078987,
		0.054161,
		0.033279,
		0.016928,
		0.005737,
		0.000372,
		0.046277,
		0.027023,
		0.012464,
		0.003233,
		0.021402,
		0.008674,
		0.00144,
		0.005562,
		0.000361,
		9.1e-005,
		0.000363,
		0.00706,
		0.00082,
		0.008741,
		0.024223,
		0.001462,
		0.010606,
		0.027226,
		0.05033,
		0.002292,
		0.012657,
		0.030412,
		0.05456,
		0.084169,
		0.00331,
		0.014896,
		0.033782,
		0.058969,
		0.089522,
		0.124573,
		0.004519,
		0.017323,
		0.037337,
		0.063557,
		0.095049,
		0.130939,
		0.170419,
		9.6e-005,
		0.00592,
		0.019939,
		0.041077,
		0.068326,
		0.100747,
		0.137468,
		0.177679,
		0.220628,
		0.000387,
		0.007514,
		0.022747,
		0.045004,
		0.073275,
		0.106619,
		0.144161,
		0.185091,
		0.228658,
		0.27417,
		0.000874,
		0.009304,
		0.025748,
		0.049119,
		0.078406,
		0.112664,
		0.151018,
		0.192655,
		0.236827,
		0.282841,
		0.330063,
		0.001559,
		0.01129,
		0.028941,
		0.053423,
		0.083718,
		0.118882,
		0.158037,
		0.200371,
		0.245133,
		0.291635,
		0.339242,
		0.387376,
		0.002443,
		0.013475,
		0.03233,
		0.057915,
		0.089212,
		0.125272,
		0.165218,
		0.208236,
		0.253577,
		0.300551,
		0.348527,
		0.396928,
		0.445233,
		0.003529,
		0.015859,
		0.035914,
		0.062598,
		0.094888,
		0.131836,
		0.172561,
		0.216251,
		0.262156,
		0.309587,
		0.357915,
		0.406567,
		0.455022,
		0.502815,
		0.004818,
		0.018444,
		0.039695,
		0.067471,
		0.100747,
		0.138573,
		0.180066,
		0.224414,
		0.270869,
		0.318742,
		0.367405,
		0.416289,
		0.464877,
		0.512705,
		0.559361,
		0.000103,
		0.006313,
		0.021232,
		0.043674,
		0.072536,
		0.106789,
		0.145482,
		0.187731,
		0.232725,
		0.279714,
		0.328012,
		0.376995,
		0.426093,
		0.474793,
		0.522638,
		0.569217,
		0.614173,
		0.000414,
		0.008014,
		0.024223,
		0.047852,
		0.077793,
		0.113014,
		0.152563,
		0.195557,
		0.241182,
		0.28869,
		0.337397,
		0.386681,
		0.435974,
		0.484769,
		0.532609,
		0.579092,
		0.623865,
		0.666622,
		0.000934,
		0.009923,
		0.02742,
		0.05223,
		0.083242,
		0.119423,
		0.159817,
		0.203541,
		0.249783,
		0.297795,
		0.346894,
		0.39646,
		0.44593,
		0.494799,
		0.542616,
		0.588982,
		0.63355,
		0.676022,
		0.716147,
		0.001665,
		0.012042,
		0.030823,
		0.056809,
		0.088884,
		0.126014,
		0.167242,
		0.211684,
		0.258527,
		0.307026,
		0.3565,
		0.406331,
		0.455958,
		0.504881,
		0.552654,
		0.598882,
		0.643226,
		0.685391,
		0.725136,
		0.762263,
		0.00261,
		0.014374,
		0.034434,
		0.061589,
		0.094719,
		0.132788,
		0.174837,
		0.219983,
		0.267414,
		0.316383,
		0.366214,
		0.416289,
		0.466054,
		0.515011,
		0.562719,
		0.608789,
		0.652886,
		0.694725,
		0.73407,
		0.770731,
		0.804566,
		0.003771,
		0.016918,
		0.038254,
		0.066571,
		0.100747,
		0.139745,
		0.182603,
		0.228439,
		0.27644,
		0.325863,
		0.376031,
		0.426333,
		0.476215,
		0.525185,
		0.572806,
		0.618696,
		0.662527,
		0.704019,
		0.742944,
		0.779121,
		0.812415,
		0.842738,
		0.005149,
		0.019678,
		0.042284,
		0.071756,
		0.10697,
		0.146884,
		0.190539,
		0.237049,
		0.285604,
		0.335463,
		0.385951,
		0.436458,
		0.486437,
		0.535398,
		0.582912,
		0.628601,
		0.672143,
		0.713267,
		0.751752,
		0.787425,
		0.820162,
		0.849883,
		0.876554,
		0.00011,
		0.006746,
		0.022654,
		0.046525,
		0.077145,
		0.113386,
		0.154206,
		0.198643,
		0.245813,
		0.294906,
		0.345181,
		0.395969,
		0.446662,
		0.496716,
		0.545648,
		0.593031,
		0.638497,
		0.681729,
		0.722464,
		0.760489,
		0.79564,
		0.827802,
		0.856905,
		0.882926,
		0.905884,
		0.000443,
		0.008564,
		0.025848,
		0.050979,
		0.082738,
		0.119997,
		0.16171,
		0.206915,
		0.254728,
		0.304341,
		0.355015,
		0.406083,
		0.45694,
		0.507048,
		0.555928,
		0.60316,
		0.64838,
		0.69128,
		0.731605,
		0.769149,
		0.803759,
		0.835329,
		0.863799,
		0.889157,
		0.911434,
		0.930704,
		0.000999,
		0.010606,
		0.029261,
		0.055645,
		0.088536,
		0.126801,
		0.169394,
		0.215353,
		0.263794,
		0.313909,
		0.364963,
		0.416289,
		0.46729,
		0.51743,
		0.566235,
		0.613292,
		0.658245,
		0.700791,
		0.740684,
		0.777728,
		0.811778,
		0.842738,
		0.870561,
		0.895243,
		0.916829,
		0.935405,
		0.951098,
		0.001783,
		0.012873,
		0.032895,
		0.060526,
		0.094539,
		0.133799,
		0.17726,
		0.223957,
		0.273008,
		0.323607,
		0.37502,
		0.426585,
		0.477707,
		0.527856,
		0.576564,
		0.623424,
		0.668086,
		0.710256,
		0.749696,
		0.78622,
		0.819691,
		0.850025,
		0.877185,
		0.90118,
		0.922066,
		0.939942,
		0.954949,
		0.967266,
		0.002795,
		0.015366,
		0.036751,
		0.065621,
		0.100747,
		0.140991,
		0.185305,
		0.232725,
		0.282369,
		0.333432,
		0.385184,
		0.436966,
		0.488188,
		0.538323,
		0.586911,
		0.63355,
		0.677898,
		0.71967,
		0.758636,
		0.794618,
		0.827492,
		0.857183,
		0.883666,
		0.906963,
		0.927141,
		0.944313,
		0.958631,
		0.970291,
		0.979518,
		0.004038,
		0.018088,
		0.04083,
		0.070933,
		0.107161,
		0.148377,
		0.193529,
		0.241656,
		0.291874,
		0.343382,
		0.395453,
		0.44743,
		0.498728,
		0.548826,
		0.59727,
		0.643665,
		0.687676,
		0.729027,
		0.767497,
		0.802918,
		0.835177,
		0.864209,
		0.890001,
		0.912588,
		0.932051,
		0.948514,
		0.962145,
		0.97315,
		0.981771,
		0.988279,
		0.005514,
		0.02104,
		0.045134,
		0.076461,
		0.113781,
		0.155955,
		0.201932,
		0.250747,
		0.301522,
		0.353455,
		0.405822,
		0.457972,
		0.509323,
		0.559361,
		0.607637,
		0.653763,
		0.697414,
		0.738322,
		0.776274,
		0.811114,
		0.842738,
		0.871096,
		0.896184,
		0.918052,
		0.936792,
		0.952544,
		0.965488,
		0.975845,
		0.983869,
		0.989843,
		0.994072,
		0.000118,
		0.007225,
		0.024223,
		0.049664,
		0.082205,
		0.120607,
		0.163726,
		0.210511,
		0.259999,
		0.31131,
		0.363647,
		0.416289,
		0.468589,
		0.519969,
		0.569922,
		0.618005,
		0.66384,
		0.707107,
		0.747548,
		0.784961,
		0.8192,
		0.850172,
		0.877839,
		0.902212,
		0.92335,
		0.941362,
		0.956401,
		0.968661,
		0.978376,
		0.985814,
		0.991269,
		0.995056,
		0.997501,
		0.000475,
		0.009174,
		0.027641,
		0.054421,
		0.088168,
		0.127639,
		0.171689,
		0.219266,
		0.269407,
		0.321236,
		0.373956,
		0.42685,
		0.479275,
		0.530661,
		0.580504,
		0.628371,
		0.673888,
		0.716747,
		0.756699,
		0.793551,
		0.827169,
		0.857473,
		0.884434,
		0.908079,
		0.928479,
		0.945759,
		0.960083,
		0.971663,
		0.980745,
		0.987608,
		0.992561,
		0.995927,
		0.998037,
		0.999217,
		0.001072,
		0.011362,
		0.031293,
		0.059405,
		0.094348,
		0.134877,
		0.179843,
		0.228195,
		0.278972,
		0.331297,
		0.384378,
		0.437501,
		0.490028,
		0.541394,
		0.591103,
		0.638727,
		0.683903,
		0.726331,
		0.76577,
		0.80204,
		0.835018,
		0.864634,
		0.890876,
		0.913781,
		0.933436,
		0.949978,
		0.96359,
		0.974494,
		0.982951,
		0.989256,
		0.993724,
		0.99669,
		0.998488,
		0.999446,
		0.999861,
		0.001913,
		0.013791,
		0.035182,
		0.064618,
		0.100747,
		0.14232,
		0.188187,
		0.237298,
		0.28869,
		0.34149,
		0.39491,
		0.448238,
		0.500843,
		0.552163,
		0.601712,
		0.649069,
		0.693878,
		0.73585,
		0.774754,
		0.810421,
		0.842738,
		0.871652,
		0.897159,
		0.919314,
		0.938217,
		0.95402,
		0.96692,
		0.977154,
		0.984999,
		0.990759,
		0.994763,
		0.997351,
		0.998863,
		0.999624,
		0.999922,
		0.999995,
		0.003,
		0.016464,
		0.039308,
		0.070061,
		0.107365,
		0.149968,
		0.196721,
		0.246571,
		0.298559,
		0.351814,
		0.405548,
		0.459058,
		0.511714,
		0.562964,
		0.612327,
		0.65939,
		0.703808,
		0.745299,
		0.783645,
		0.818687,
		0.850326,
		0.878519,
		0.90328,
		0.924674,
		0.942819,
		0.957881,
		0.970072,
		0.979645,
		0.986889,
		0.992122,
		0.995683,
		0.997917,
		0.999169,
		0.999757,
		0.999961,
		0.999999,
		1,
		0.004334,
		0.019382,
		0.043674,
		0.075735,
		0.114202,
		0.157821,
		0.205443,
		0.256014,
		0.308577,
		0.362263,
		0.416289,
		0.469955,
		0.522638,
		0.573791,
		0.622942,
		0.669685,
		0.713686,
		0.754672,
		0.792437,
		0.826833,
		0.857774,
		0.885232,
		0.909232,
		0.929857,
		0.947239,
		0.96156,
		0.973047,
		0.981968,
		0.988626,
		0.99335,
		0.99649,
		0.998396,
		0.999412,
		0.999852,
		0.999984,
		1,
		0.999999,
		0.999962,
		0.00592,
		0.022547,
		0.048281,
		0.08164,
		0.121258,
		0.165879,
		0.214352,
		0.265625,
		0.318742,
		0.372836,
		0.427129,
		0.480925,
		0.533608,
		0.584639,
		0.63355,
		0.679948,
		0.723506,
		0.763962,
		0.801123,
		0.834852,
		0.865077,
		0.891784,
		0.915012,
		0.93486,
		0.951476,
		0.965057,
		0.975845,
		0.984125,
		0.990212,
		0.994448,
		0.997189,
		0.998793,
		0.9996,
		0.999917,
		0.999995,
		1,
		0.999995,
		0.999922,
		0.999635,
		0.000127,
		0.007758,
		0.025962,
		0.05313,
		0.087777,
		0.128533,
		0.17414,
		0.223446,
		0.275401,
		0.32905,
		0.383529,
		0.438064,
		0.491965,
		0.544621,
		0.595501,
		0.644146,
		0.690172,
		0.73326,
		0.773163,
		0.809696,
		0.842738,
		0.87223,
		0.89817,
		0.920616,
		0.93968,
		0.955526,
		0.968369,
		0.978467,
		0.986118,
		0.991651,
		0.995421,
		0.997789,
		0.999117,
		0.999741,
		0.999959,
		0.999999,
		1,
		0.999984,
		0.999857,
		0.999446,
		0.998532,
		0.000511,
		0.009851,
		0.029626,
		0.058222,
		0.094146,
		0.136026,
		0.182603,
		0.232725,
		0.285341,
		0.339498,
		0.394338,
		0.449089,
		0.503068,
		0.555671,
		0.606372,
		0.654724,
		0.70035,
		0.742944,
		0.782268,
		0.818151,
		0.850486,
		0.879226,
		0.904386,
		0.926039,
		0.944313,
		0.959389,
		0.971498,
		0.980913,
		0.987949,
		0.992949,
		0.996274,
		0.998296,
		0.999375,
		0.999843,
		0.999982,
		1,
		0.999999,
		0.99996,
		0.999757,
		0.999193,
		0.998037,
		0.996042,
		0.001154,
		0.012202,
		0.033544,
		0.063557,
		0.100747,
		0.143738,
		0.191269,
		0.242186,
		0.295441,
		0.350084,
		0.405259,
		0.460201,
		0.51423,
		0.566751,
		0.617247,
		0.665277,
		0.710476,
		0.752549,
		0.79127,
		0.826481,
		0.858088,
		0.886059,
		0.910426,
		0.931276,
		0.948756,
		0.963062,
		0.974442,
		0.983187,
		0.989624,
		0.994109,
		0.997015,
		0.998717,
		0.999575,
		0.999912,
		0.999994,
		1,
		0.999994,
		0.999917,
		0.999612,
		0.998863,
		0.997428,
		0.995056,
		0.99151,
		0.002059,
		0.014812,
		0.037715,
		0.069138,
		0.107582,
		0.151669,
		0.200135,
		0.251827,
		0.3057,
		0.360804,
		0.416289,
		0.471394,
		0.525446,
		0.577857,
		0.628118,
		0.675798,
		0.720543,
		0.762069,
		0.800163,
		0.834679,
		0.865538,
		0.892725,
		0.916286,
		0.936326,
		0.953007,
		0.966545,
		0.977203,
		0.985289,
		0.991144,
		0.995138,
		0.997651,
		0.999061,
		0.999725,
		0.999956,
		0.999999,
		1,
		0.999982,
		0.999848,
		0.999412,
		0.998443,
		0.99669,
		0.993901,
		0.989843,
		0.984305,
		0.003228,
		0.017685,
		0.042142,
		0.074965,
		0.114651,
		0.159817,
		0.2092,
		0.261647,
		0.316114,
		0.371655,
		0.427424,
		0.482664,
		0.536711,
		0.588982,
		0.638979,
		0.686282,
		0.730545,
		0.771497,
		0.808939,
		0.842738,
		0.872831,
		0.899218,
		0.921961,
		0.941183,
		0.957064,
		0.969836,
		0.979781,
		0.987222,
		0.992516,
		0.996042,
		0.998188,
		0.999334,
		0.999833,
		0.999981,
		1,
		0.999999,
		0.999957,
		0.999741,
		0.999143,
		0.997917,
		0.995807,
		0.992561,
		0.987949,
		0.981771,
		0.973858,
		0.004665,
		0.020821,
		0.046826,
		0.081039,
		0.121952,
		0.168182,
		0.218463,
		0.271644,
		0.326681,
		0.382633,
		0.438658,
		0.494005,
		0.548017,
		0.60012,
		0.649825,
		0.69672,
		0.740474,
		0.780826,
		0.817592,
		0.850653,
		0.87996,
		0.905532,
		0.927446,
		0.945845,
		0.960924,
		0.972936,
		0.982179,
		0.988991,
		0.993744,
		0.996827,
		0.998635,
		0.999547,
		0.999906,
		0.999994,
		1,
		0.999994,
		0.999912,
		0.999588,
		0.998793,
		0.997272,
		0.994763,
		0.991018,
		0.985814,
		0.978957,
		0.970291,
		0.959694,
		0.006372,
		0.024223,
		0.051768,
		0.08736,
		0.129487,
		0.176763,
		0.227922,
		0.281814,
		0.337397,
		0.393734,
		0.449987,
		0.505413,
		0.559361,
		0.611266,
		0.660647,
		0.707107,
		0.750323,
		0.790049,
		0.826114,
		0.858415,
		0.88692,
		0.911662,
		0.932739,
		0.950308,
		0.964588,
		0.975845,
		0.984397,
		0.990599,
		0.994834,
		0.997501,
		0.999,
		0.999707,
		0.999953,
		0.999999,
		1,
		0.999981,
		0.999838,
		0.999375,
		0.998347,
		0.99649,
		0.993541,
		0.989256,
		0.983418,
		0.975845,
		0.966392,
		0.954949,
		0.941445,
		0.000137,
		0.008351,
		0.027894,
		0.05697,
		0.09393,
		0.137255,
		0.185559,
		0.237576,
		0.292156,
		0.34826,
		0.404954,
		0.461407,
		0.516882,
		0.570735,
		0.622411,
		0.67144,
		0.717434,
		0.760084,
		0.799158,
		0.834498,
		0.866019,
		0.893703,
		0.917603,
		0.937833,
		0.954571,
		0.968053,
		0.978564,
		0.98644,
		0.99205,
		0.995792,
		0.998072,
		0.999291,
		0.999822,
		0.99998,
		1,
		0.999999,
		0.999955,
		0.999725,
		0.999089,
		0.997789,
		0.995554,
		0.992122,
		0.987255,
		0.980745,
		0.972419,
		0.962145,
		0.949827,
		0.935405,
		0.918855,
		0.000551,
		0.010606,
		0.031834,
		0.062433,
		0.100747,
		0.145256,
		0.19457,
		0.247423,
		0.302666,
		0.359265,
		0.416289,
		0.472913,
		0.528406,
		0.582134,
		0.63355,
		0.682196,
		0.727694,
		0.769751,
		0.808146,
		0.842738,
		0.873458,
		0.900305,
		0.92335,
		0.942727,
		0.958631,
		0.971319,
		0.981094,
		0.98831,
		0.99335,
		0.996625,
		0.998546,
		0.999517,
		0.9999,
		0.999993,
		1,
		0.999994,
		0.999906,
		0.999561,
		0.998717,
		0.997104,
		0.994448,
		0.99049,
		0.984999,
		0.977776,
		0.968661,
		0.957534,
		0.944313,
		0.92895,
		0.911434,
		0.891784,
		0.001245,
		0.013138,
		0.036046,
		0.068159,
		0.107814,
		0.15349,
		0.203793,
		0.25746,
		0.313342,
		0.370409,
		0.427734,
		0.484498,
		0.539979,
		0.59355,
		0.644676,
		0.692907,
		0.737881,
		0.779315,
		0.817006,
		0.850827,
		0.880725,
		0.90672,
		0.928899,
		0.947416,
		0.962487,
		0.974386,
		0.983438,
		0.990011,
		0.994505,
		0.99734,
		0.998934,
		0.999687,
		0.99995,
		0.999999,
		1,
		0.99998,
		0.999827,
		0.999334,
		0.998243,
		0.996274,
		0.993153,
		0.988626,
		0.982468,
		0.974494,
		0.964555,
		0.952544,
		0.938392,
		0.922066,
		0.903567,
		0.882926,
		0.860201,
		0.002221,
		0.015951,
		0.040531,
		0.074147,
		0.11513,
		0.161955,
		0.213228,
		0.267685,
		0.324181,
		0.381687,
		0.439285,
		0.496159,
		0.551596,
		0.604978,
		0.655781,
		0.703566,
		0.747985,
		0.788769,
		0.825729,
		0.858756,
		0.887815,
		0.912942,
		0.934245,
		0.951898,
		0.966137,
		0.977255,
		0.985597,
		0.991548,
		0.995521,
		0.997946,
		0.999244,
		0.99981,
		0.999979,
		1,
		0.999999,
		0.999952,
		0.999707,
		0.999031,
		0.997651,
		0.995282,
		0.991651,
		0.98651,
		0.979645,
		0.970881,
		0.960083,
		0.947158,
		0.932051,
		0.914741,
		0.895243,
		0.873601,
		0.849883,
		0.824185,
		0.003484,
		0.019046,
		0.045292,
		0.080399,
		0.122695,
		0.170651,
		0.222873,
		0.278096,
		0.335179,
		0.393097,
		0.450936,
		0.507888,
		0.563249,
		0.61641,
		0.666857,
		0.714166,
		0.758,
		0.798105,
		0.834309,
		0.86652,
		0.89472,
		0.918966,
		0.939384,
		0.95617,
		0.96958,
		0.979927,
		0.987575,
		0.992925,
		0.996405,
		0.998451,
		0.999485,
		0.999893,
		0.999993,
		1,
		0.999993,
		0.9999,
		0.999532,
		0.998635,
		0.996923,
		0.994109,
		0.989923,
		0.984125,
		0.976511,
		0.96692,
		0.95523,
		0.941362,
		0.925275,
		0.906963,
		0.886452,
		0.863799,
		0.839085,
		0.812415,
		0.783913,
		0.005035,
		0.022426,
		0.05033,
		0.086916,
		0.130509,
		0.179577,
		0.232725,
		0.28869,
		0.346332,
		0.404632,
		0.462682,
		0.51968,
		0.574932,
		0.627839,
		0.677898,
		0.724699,
		0.767917,
		0.807315,
		0.842738,
		0.874111,
		0.901435,
		0.924786,
		0.944313,
		0.96023,
		0.972816,
		0.982404,
		0.989376,
		0.99415,
		0.997165,
		0.998863,
		0.999666,
		0.999947,
		0.999999,
		1,
		0.999979,
		0.999816,
		0.999291,
		0.998131,
		0.996042,
		0.992736,
		0.987949,
		0.98145,
		0.973047,
		0.962592,
		0.949978,
		0.935141,
		0.918052,
		0.89872,
		0.877185,
		0.853514,
		0.827802,
		0.800163,
		0.770731,
		0.739658,
		0.006878,
		0.026093,
		0.055645,
		0.093699,
		0.138573,
		0.188732,
		0.242783,
		0.299464,
		0.357638,
		0.416289,
		0.474517,
		0.531529,
		0.586639,
		0.639258,
		0.688896,
		0.735156,
		0.777728,
		0.816392,
		0.851009,
		0.881522,
		0.907953,
		0.930398,
		0.949027,
		0.964076,
		0.975845,
		0.984688,
		0.991005,
		0.995229,
		0.997809,
		0.999193,
		0.999797,
		0.999977,
		1,
		0.999999,
		0.999948,
		0.999687,
		0.998968,
		0.997501,
		0.994989,
		0.991144,
		0.985711,
		0.978467,
		0.969235,
		0.957881,
		0.944313,
		0.928479,
		0.910368,
		0.890001,
		0.867431,
		0.842738,
		0.816028,
		0.787425,
		0.757075,
		0.725136,
		0.691782,
		0.000148,
		0.009016,
		0.030049,
		0.061241,
		0.100747,
		0.146884,
		0.198115,
		0.253045,
		0.310415,
		0.369091,
		0.428063,
		0.486437,
		0.543429,
		0.598361,
		0.650658,
		0.699842,
		0.745529,
		0.787425,
		0.825327,
		0.859113,
		0.888746,
		0.914268,
		0.935798,
		0.953525,
		0.967708,
		0.978669,
		0.986782,
		0.992466,
		0.996168,
		0.998347,
		0.99945,
		0.999886,
		0.999993,
		1,
		0.999993,
		0.999893,
		0.999501,
		0.998546,
		0.996728,
		0.993744,
		0.989313,
		0.983187,
		0.975156,
		0.965057,
		0.952769,
		0.938217,
		0.921364,
		0.902212,
		0.880796,
		0.857183,
		0.831466,
		0.803759,
		0.774201,
		0.742944,
		0.710157,
		0.676022,
		0.64073,
		0.000596,
		0.011451,
		0.034297,
		0.067117,
		0.108062,
		0.155444,
		0.207724,
		0.263508,
		0.321539,
		0.380687,
		0.439948,
		0.498434,
		0.555372,
		0.610092,
		0.662033,
		0.710728,
		0.75581,
		0.797,
		0.834112,
		0.867043,
		0.895777,
		0.920376,
		0.94098,
		0.957802,
		0.971124,
		0.981289,
		0.988691,
		0.993766,
		0.996976,
		0.998786,
		0.999643,
		0.999943,
		0.999998,
		1,
		0.999977,
		0.999803,
		0.999244,
		0.99801,
		0.995792,
		0.992288,
		0.987222,
		0.980357,
		0.971498,
		0.960493,
		0.947239,
		0.931675,
		0.913781,
		0.893571,
		0.871096,
		0.846434,
		0.819691,
		0.790993,
		0.760489,
		0.72834,
		0.694725,
		0.659834,
		0.623865,
		0.587025,
		0.001347,
		0.014187,
		0.038838,
		0.073275,
		0.115644,
		0.164251,
		0.217558,
		0.27417,
		0.332833,
		0.392422,
		0.451939,
		0.510503,
		0.567351,
		0.621825,
		0.673374,
		0.721547,
		0.76599,
		0.806444,
		0.842738,
		0.874792,
		0.902608,
		0.92627,
		0.945942,
		0.961859,
		0.974325,
		0.983707,
		0.990418,
		0.994912,
		0.997662,
		0.999138,
		0.999782,
		0.999976,
		1,
		0.999998,
		0.999945,
		0.999666,
		0.9989,
		0.99734,
		0.994673,
		0.990599,
		0.984851,
		0.977203,
		0.967473,
		0.955526,
		0.941275,
		0.924674,
		0.905717,
		0.884434,
		0.86089,
		0.835177,
		0.80741,
		0.777728,
		0.746289,
		0.713267,
		0.678846,
		0.643226,
		0.606611,
		0.569217,
		0.531263,
		0.002404,
		0.017226,
		0.043674,
		0.079717,
		0.123492,
		0.173305,
		0.227614,
		0.285028,
		0.344293,
		0.404291,
		0.46403,
		0.522638,
		0.579359,
		0.63355,
		0.684672,
		0.732288,
		0.776061,
		0.815748,
		0.851199,
		0.882352,
		0.909232,
		0.931946,
		0.950679,
		0.965692,
		0.977312,
		0.985926,
		0.991969,
		0.995911,
		0.998234,
		0.999412,
		0.999878,
		0.999992,
		1,
		0.999992,
		0.999886,
		0.999468,
		0.998451,
		0.996517,
		0.99335,
		0.988657,
		0.982179,
		0.973702,
		0.963062,
		0.950139,
		0.93486,
		0.917197,
		0.897159,
		0.874792,
		0.850172,
		0.823405,
		0.794618,
		0.763962,
		0.731605,
		0.697728,
		0.662527,
		0.626207,
		0.588982,
		0.551073,
		0.512705,
		0.474108,
		0.003771,
		0.02057,
		0.048808,
		0.086442,
		0.131607,
		0.182603,
		0.237891,
		0.296078,
		0.355915,
		0.416289,
		0.476215,
		0.53483,
		0.591389,
		0.645261,
		0.695919,
		0.742944,
		0.786014,
		0.824905,
		0.859486,
		0.889716,
		0.915644,
		0.937398,
		0.955189,
		0.969301,
		0.980084,
		0.987949,
		0.99335,
		0.996771,
		0.998703,
		0.999618,
		0.999939,
		0.999998,
		1,
		0.999976,
		0.99979,
		0.999193,
		0.997879,
		0.995521,
		0.991804,
		0.98644,
		0.979184,
		0.969836,
		0.958246,
		0.944313,
		0.927979,
		0.909232,
		0.888097,
		0.864634,
		0.838934,
		0.811114,
		0.781314,
		0.749696,
		0.716437,
		0.681729,
		0.645775,
		0.608789,
		0.570991,
		0.532609,
		0.493875,
		0.455022,
		0.416289,
		0.00545,
		0.024223,
		0.05424,
		0.093452,
		0.139988,
		0.192146,
		0.248386,
		0.307317,
		0.367695,
		0.428411,
		0.488488,
		0.547074,
		0.603434,
		0.656948,
		0.707107,
		0.753505,
		0.795839,
		0.833904,
		0.86759,
		0.896877,
		0.921837,
		0.942622,
		0.959469,
		0.972685,
		0.982646,
		0.989783,
		0.994569,
		0.997501,
		0.999078,
		0.999767,
		0.999974,
		1,
		0.999998,
		0.999941,
		0.999643,
		0.998826,
		0.997165,
		0.994331,
		0.990011,
		0.983926,
		0.975845,
		0.965583,
		0.953007,
		0.938031,
		0.920616,
		0.900765,
		0.878519,
		0.853952,
		0.827169,
		0.7983,
		0.767497,
		0.734932,
		0.700791,
		0.665277,
		0.628601,
		0.590984,
		0.552654,
		0.513845,
		0.474793,
		0.435741,
		0.396928,
		0.358598,
		0.007447,
		0.028187,
		0.059973,
		0.100747,
		0.148635,
		0.201932,
		0.259096,
		0.318742,
		0.379627,
		0.44065,
		0.500843,
		0.559361,
		0.615483,
		0.668603,
		0.718226,
		0.763962,
		0.805528,
		0.842738,
		0.875503,
		0.903828,
		0.927805,
		0.947615,
		0.963517,
		0.975845,
		0.984999,
		0.991431,
		0.995632,
		0.998111,
		0.99937,
		0.999869,
		0.999991,
		1,
		0.999992,
		0.999878,
		0.999431,
		0.998347,
		0.996289,
		0.992925,
		0.987949,
		0.981094,
		0.972142,
		0.960924,
		0.947325,
		0.931276,
		0.912756,
		0.891784,
		0.868415,
		0.842738,
		0.814873,
		0.784961,
		0.753166,
		0.71967,
		0.684672,
		0.64838,
		0.611015,
		0.572806,
		0.533987,
		0.494799,
		0.455485,
		0.416289,
		0.377459,
		0.339242,
		0.301882,
		0.000161,
		0.009762,
		0.032464,
		0.066007,
		0.108329,
		0.157547,
		0.211958,
		0.27002,
		0.330348,
		0.391707,
		0.453002,
		0.513271,
		0.571684,
		0.62753,
		0.680218,
		0.729267,
		0.774307,
		0.815072,
		0.851398,
		0.883219,
		0.910561,
		0.933544,
		0.952372,
		0.967332,
		0.978782,
		0.987147,
		0.992899,
		0.996548,
		0.998612,
		0.999591,
		0.999934,
		0.999998,
		1,
		0.999974,
		0.999775,
		0.999138,
		0.997737,
		0.995229,
		0.991281,
		0.985597,
		0.977922,
		0.968053,
		0.955839,
		0.941183,
		0.924033,
		0.904386,
		0.882275,
		0.857774,
		0.830985,
		0.80204,
		0.771094,
		0.738322,
		0.703917,
		0.668086,
		0.631048,
		0.593031,
		0.554272,
		0.515011,
		0.475496,
		0.435974,
		0.396697,
		0.357915,
		0.31988,
		0.282841,
		0.247045,
		0.000647,
		0.012401,
		0.037057,
		0.072345,
		0.116196,
		0.166724,
		0.222223,
		0.281153,
		0.342132,
		0.40393,
		0.465459,
		0.525768,
		0.584036,
		0.639566,
		0.691782,
		0.74022,
		0.784529,
		0.824461,
		0.859876,
		0.890728,
		0.917071,
		0.939048,
		0.956891,
		0.970913,
		0.981498,
		0.989094,
		0.994196,
		0.997327,
		0.999012,
		0.99975,
		0.999972,
		1,
		0.999998,
		0.999937,
		0.999618,
		0.998745,
		0.996976,
		0.993962,
		0.989376,
		0.982931,
		0.974386,
		0.963556,
		0.950308,
		0.934562,
		0.916286,
		0.895491,
		0.87223,
		0.846589,
		0.818687,
		0.788668,
		0.756699,
		0.722968,
		0.687676,
		0.651042,
		0.613292,
		0.574663,
		0.535398,
		0.495746,
		0.455958,
		0.416289,
		0.376995,
		0.33833,
		0.300551,
		0.26391,
		0.228658,
		0.195044,
		0.001462,
		0.015366,
		0.041967,
		0.078987,
		0.124349,
		0.176164,
		0.232725,
		0.292492,
		0.354089,
		0.416289,
		0.478015,
		0.538323,
		0.596407,
		0.651582,
		0.703287,
		0.751077,
		0.794618,
		0.833687,
		0.868161,
		0.898023,
		0.92335,
		0.944313,
		0.961169,
		0.97426,
		0.983995,
		0.990846,
		0.995329,
		0.997978,
		0.999325,
		0.99986,
		0.999991,
		1,
		0.999991,
		0.999869,
		0.999391,
		0.998234,
		0.996042,
		0.992466,
		0.987186,
		0.979927,
		0.970465,
		0.958631,
		0.944313,
		0.927446,
		0.908019,
		0.886059,
		0.861636,
		0.834852,
		0.805839,
		0.774754,
		0.741777,
		0.707107,
		0.670956,
		0.63355,
		0.595126,
		0.555928,
		0.516206,
		0.476215,
		0.436213,
		0.39646,
		0.357216,
		0.318742,
		0.281295,
		0.245133,
		0.210511,
		0.177679,
		0.146884,
		0.00261,
		0.01866,
		0.047197,
		0.085934,
		0.132788,
		0.185867,
		0.243461,
		0.304035,
		0.366214,
		0.42878,
		0.490663,
		0.550931,
		0.608789,
		0.663568,
		0.714722,
		0.761826,
		0.804566,
		0.842738,
		0.876247,
		0.905097,
		0.929393,
		0.949334,
		0.965204,
		0.977373,
		0.986276,
		0.99241,
		0.996306,
		0.998512,
		0.999561,
		0.999929,
		0.999998,
		1,
		0.999972,
		0.999759,
		0.999078,
		0.997583,
		0.994912,
		0.990717,
		0.984688,
		0.976563,
		0.966137,
		0.953258,
		0.937833,
		0.919818,
		0.899218,
		0.876079,
		0.850486,
		0.822556,
		0.792437,
		0.760297,
		0.726331,
		0.690745,
		0.653763,
		0.615622,
		0.576564,
		0.536843,
		0.496716,
		0.456443,
		0.416289,
		0.376519,
		0.337397,
		0.299189,
		0.262156,
		0.226559,
		0.192655,
		0.160699,
		0.130939,
		0.103621,
		0.004095,
		0.022285,
		0.052748,
		0.093187,
		0.141513,
		0.19583,
		0.254429,
		0.315776,
		0.378502,
		0.441396,
		0.503396,
		0.563582,
		0.621173,
		0.675515,
		0.726079,
		0.772459,
		0.814362,
		0.851607,
		0.884124,
		0.911942,
		0.935195,
		0.954108,
		0.968996,
		0.980254,
		0.988347,
		0.993791,
		0.997137,
		0.998941,
		0.999732,
		0.99997,
		1,
		0.999998,
		0.999932,
		0.999591,
		0.998658,
		0.996771,
		0.993563,
		0.988691,
		0.981857,
		0.972816,
		0.961379,
		0.947416,
		0.930852,
		0.911662,
		0.889869,
		0.865538,
		0.83877,
		0.809696,
		0.778478,
		0.745299,
		0.710363,
		0.673888,
		0.636109,
		0.59727,
		0.557624,
		0.51743,
		0.476952,
		0.436458,
		0.396218,
		0.3565,
		0.317576,
		0.279714,
		0.243179,
		0.208236,
		0.175145,
		0.144161,
		0.115538,
		0.089522,
		0.066356,
		0.00592,
		0.026246,
		0.058623,
		0.100747,
		0.150522,
		0.206051,
		0.265625,
		0.327712,
		0.390948,
		0.45413,
		0.516206,
		0.576269,
		0.63355,
		0.687413,
		0.737347,
		0.782964,
		0.823996,
		0.860284,
		0.891784,
		0.918552,
		0.940749,
		0.958631,
		0.972543,
		0.982906,
		0.990212,
		0.994999,
		0.997833,
		0.999276,
		0.999849,
		0.99999,
		1,
		0.99999,
		0.99986,
		0.999348,
		0.998111,
		0.995774,
		0.991969,
		0.986362,
		0.978669,
		0.968661,
		0.95617,
		0.941085,
		0.92335,
		0.902962,
		0.87996,
		0.854428,
		0.826481,
		0.796268,
		0.763962,
		0.729761,
		0.693878,
		0.656545,
		0.618005,
		0.578511,
		0.538323,
		0.497709,
		0.45694,
		0.416289,
		0.376031,
		0.336442,
		0.297795,
		0.260362,
		0.224414,
		0.190218,
		0.158037,
		0.128128,
		0.100747,
		0.076143,
		0.05456,
		0.036235,
		0.008089,
		0.030544,
		0.064824,
		0.108615,
		0.159817,
		0.21653,
		0.277047,
		0.339837,
		0.403546,
		0.466975,
		0.529086,
		0.588982,
		0.645911,
		0.699252,
		0.748515,
		0.793332,
		0.833458,
		0.86876,
		0.899218,
		0.924919,
		0.946052,
		0.962903,
		0.975845,
		0.985332,
		0.991877,
		0.996042,
		0.998404,
		0.999529,
		0.999924,
		0.999998,
		1,
		0.99997,
		0.999741,
		0.999012,
		0.997416,
		0.994569,
		0.990107,
		0.983707,
		0.9751,
		0.964076,
		0.950488,
		0.934245,
		0.915312,
		0.893703,
		0.869479,
		0.842738,
		0.813614,
		0.782268,
		0.748889,
		0.713686,
		0.676885,
		0.638727,
		0.599465,
		0.559361,
		0.518684,
		0.477707,
		0.436709,
		0.395969,
		0.355767,
		0.316383,
		0.278096,
		0.241182,
		0.205913,
		0.172561,
		0.141391,
		0.112664,
		0.086636,
		0.063557,
		0.043674,
		0.027226,
		0.014447,
		0.000176,
		0.010606,
		0.035182,
		0.07135,
		0.11679,
		0.169394,
		0.227264,
		0.28869,
		0.352149,
		0.416289,
		0.479925,
		0.542026,
		0.601712,
		0.658245,
		0.711022,
		0.759572,
		0.803552,
		0.842738,
		0.877025,
		0.906419,
		0.931037,
		0.951098,
		0.96692,
		0.978905,
		0.987535,
		0.99335,
		0.99693,
		0.998863,
		0.999712,
		0.999968,
		1,
		0.999998,
		0.999927,
		0.999561,
		0.998563,
		0.996548,
		0.99313,
		0.987949,
		0.980698,
		0.971124,
		0.959038,
		0.944313,
		0.926878,
		0.90672,
		0.883873,
		0.858415,
		0.830462,
		0.800163,
		0.767695,
		0.73326,
		0.697079,
		0.65939,
		0.620445,
		0.580504,
		0.53984,
		0.498728,
		0.45745,
		0.416289,
		0.375532,
		0.335463,
		0.296367,
		0.258527,
		0.222223,
		0.187731,
		0.155325,
		0.125272,
		0.097836,
		0.073275,
		0.051842,
		0.033782,
		0.019337,
		0.008741,
		0.002221,
		0.000705,
		0.013475,
		0.040162,
		0.078205,
		0.125272,
		0.179254,
		0.238249,
		0.300551,
		0.36464,
		0.429172,
		0.492972,
		0.55502,
		0.61445,
		0.670542,
		0.722712,
		0.770509,
		0.813614,
		0.851827,
		0.88507,
		0.913379,
		0.9369,
		0.955885,
		0.970681,
		0.981724,
		0.989522,
		0.99464,
		0.997675,
		0.999222,
		0.999838,
		0.999989,
		1,
		0.99999,
		0.999849,
		0.999301,
		0.997978,
		0.995483,
		0.991431,
		0.985471,
		0.977312,
		0.966719,
		0.953525,
		0.937623,
		0.918966,
		0.897559,
		0.873458,
		0.846758,
		0.817592,
		0.786125,
		0.752549,
		0.717079,
		0.679948,
		0.641405,
		0.601712,
		0.561141,
		0.519969,
		0.478481,
		0.436966,
		0.395714,
		0.355015,
		0.315161,
		0.27644,
		0.239139,
		0.203541,
		0.169927,
		0.138573,
		0.109748,
		0.083718,
		0.060743,
		0.041077,
		0.024968,
		0.012657,
		0.004379,
		0.000363,
		0.001593,
		0.016698,
		0.045488,
		0.085389,
		0.134062,
		0.189395,
		0.249484,
		0.312626,
		0.377306,
		0.442188,
		0.506107,
		0.568056,
		0.627186,
		0.682792,
		0.734311,
		0.781314,
		0.823506,
		0.860713,
		0.892886,
		0.92009,
		0.942504,
		0.960409,
		0.974188,
		0.984305,
		0.991297,
		0.995754,
		0.998286,
		0.999493,
		0.999918,
		0.999998,
		1,
		0.999968,
		0.999722,
		0.998941,
		0.997234,
		0.994196,
		0.989445,
		0.982646,
		0.973521,
		0.961859,
		0.947512,
		0.930398,
		0.91049,
		0.887815,
		0.862446,
		0.834498,
		0.804122,
		0.771497,
		0.736831,
		0.70035,
		0.6623,
		0.622942,
		0.582546,
		0.541394,
		0.499772,
		0.457972,
		0.416289,
		0.37502,
		0.33446,
		0.294906,
		0.25665,
		0.219983,
		0.185194,
		0.152563,
		0.122371,
		0.094888,
		0.070384,
		0.049119,
		0.031349,
		0.017323,
		0.007282,
		0.001462,
		0.002844,
		0.02028,
		0.051161,
		0.092903,
		0.14316,
		0.199814,
		0.260965,
		0.324909,
		0.39014,
		0.455329,
		0.519322,
		0.581127,
		0.63991,
		0.694984,
		0.745808,
		0.791976,
		0.833218,
		0.869387,
		0.900464,
		0.926546,
		0.947842,
		0.964669,
		0.977439,
		0.986651,
		0.99287,
		0.996704,
		0.998778,
		0.99969,
		0.999965,
		1,
		0.999998,
		0.999921,
		0.999529,
		0.99846,
		0.996306,
		0.99266,
		0.987147,
		0.979446,
		0.969301,
		0.95652,
		0.94098,
		0.92262,
		0.901435,
		0.877473,
		0.850827,
		0.82163,
		0.790049,
		0.75628,
		0.720543,
		0.68308,
		0.644146,
		0.604014,
		0.562964,
		0.521286,
		0.479275,
		0.43723,
		0.395453,
		0.354245,
		0.313909,
		0.274744,
		0.237049,
		0.201118,
		0.167242,
		0.135705,
		0.106789,
		0.08077,
		0.057915,
		0.038489,
		0.022747,
		0.01094,
		0.00331,
		9.3e-005,
		0.004462,
		0.024223,
		0.057184,
		0.100747,
		0.152563,
		0.210511,
		0.272688,
		0.337397,
		0.403137,
		0.468589,
		0.532609,
		0.594222,
		0.65261,
		0.707107,
		0.757191,
		0.802484,
		0.842738,
		0.877839,
		0.907796,
		0.932739,
		0.95291,
		0.968661,
		0.980438,
		0.98877,
		0.994248,
		0.997501,
		0.999163,
		0.999825,
		0.999988,
		1,
		0.999989,
		0.999838,
		0.999249,
		0.997833,
		0.995167,
		0.990846,
		0.984507,
		0.975845,
		0.964625,
		0.950679,
		0.933906,
		0.914268,
		0.891784,
		0.86652,
		0.838591,
		0.808146,
		0.77537,
		0.740474,
		0.703692,
		0.665277,
		0.625499,
		0.584639,
		0.542987,
		0.500843,
		0.458508,
		0.416289,
		0.374495,
		0.333432,
		0.293408,
		0.254728,
		0.217694,
		0.182603,
		0.14975,
		0.119423,
		0.091904,
		0.067471,
		0.046396,
		0.028941,
		0.015366,
		0.00592,
		0.000846,
		0.006452,
		0.028532,
		0.063557,
		0.108924,
		0.162273,
		0.221482,
		0.28465,
		0.350084,
		0.416289,
		0.481957,
		0.545959,
		0.607332,
		0.665277,
		0.719149,
		0.76845,
		0.812825,
		0.852057,
		0.886059,
		0.914874,
		0.938663,
		0.957705,
		0.972387,
		0.983187,
		0.990665,
		0.99544,
		0.998157,
		0.999454,
		0.999912,
		0.999998,
		1,
		0.999965,
		0.999701,
		0.998863,
		0.997036,
		0.993791,
		0.988728,
		0.981498,
		0.971816,
		0.959469,
		0.944313,
		0.92627,
		0.905327,
		0.881522,
		0.854945,
		0.825729,
		0.794042,
		0.760084,
		0.724081,
		0.686282,
		0.646952,
		0.606372,
		0.564834,
		0.522638,
		0.48009,
		0.437501,
		0.395185,
		0.353455,
		0.312626,
		0.273008,
		0.234912,
		0.198643,
		0.164503,
		0.132788,
		0.103789,
		0.077793,
		0.055077,
		0.035914,
		0.02057,
		0.009304,
		0.002366,
		0.008818,
		0.033208,
		0.070284,
		0.117431,
		0.172287,
		0.232725,
		0.296847,
		0.362964,
		0.42959,
		0.495428,
		0.559361,
		0.620445,
		0.677898,
		0.731099,
		0.779572,
		0.82299,
		0.861163,
		0.894038,
		0.921688,
		0.944313,
		0.962225,
		0.975845,
		0.985689,
		0.992346,
		0.996457,
		0.998685,
		0.999666,
		0.999962,
		1,
		0.999998,
		0.999915,
		0.999493,
		0.998347,
		0.996042,
		0.992149,
		0.986276,
		0.978092,
		0.967332,
		0.953807,
		0.937398,
		0.918052,
		0.895777,
		0.870636,
		0.842738,
		0.812236,
		0.779315,
		0.744191,
		0.707107,
		0.668323,
		0.628118,
		0.586783,
		0.544621,
		0.501941,
		0.459058,
		0.416289,
		0.373956,
		0.332378,
		0.291874,
		0.252761,
		0.215353,
		0.179959,
		0.146884,
		0.116428,
		0.088884,
		0.064539,
		0.043674,
		0.026564,
		0.013475,
		0.004665,
		0.000387,
		0.000192,
		0.011564,
		0.038254,
		0.077365,
		0.126271,
		0.182603,
		0.244237,
		0.309274,
		0.376031,
		0.443032,
		0.508991,
		0.572806,
		0.63355,
		0.690463,
		0.742944,
		0.790544,
		0.832964,
		0.870045,
		0.901765,
		0.928233,
		0.949683,
		0.966466,
		0.979038,
		0.987949,
		0.993819,
		0.997312,
		0.999098,
		0.999812,
		0.999988,
		1,
		0.999988,
		0.999825,
		0.999193,
		0.997675,
		0.994823,
		0.990212,
		0.983462,
		0.97426,
		0.962366,
		0.947615,
		0.929913,
		0.909232,
		0.885604,
		0.859113,
		0.82989,
		0.798105,
		0.763962,
		0.727694,
		0.689557,
		0.649825,
		0.608789,
		0.566751,
		0.524024,
		0.480925,
		0.437779,
		0.39491,
		0.352645,
		0.31131,
		0.27123,
		0.232725,
		0.196114,
		0.16171,
		0.12982,
		0.100747,
		0.074788,
		0.05223,
		0.033357,
		0.018444,
		0.007758,
		0.001559,
		0.000771,
		0.014693,
		0.043674,
		0.084802,
		0.135442,
		0.193222,
		0.256014,
		0.321925,
		0.389279,
		0.456608,
		0.522638,
		0.586284,
		0.646637,
		0.702959,
		0.754672,
		0.801356,
		0.842738,
		0.878693,
		0.909232,
		0.934501,
		0.954769,
		0.970427,
		0.981968,
		0.989975,
		0.995096,
		0.998016,
		0.999412,
		0.999905,
		0.999997,
		1,
		0.999962,
		0.999678,
		0.998778,
		0.99682,
		0.99335,
		0.987949,
		0.980254,
		0.969973,
		0.956891,
		0.940869,
		0.921837,
		0.899793,
		0.874792,
		0.846941,
		0.816392,
		0.783333,
		0.747985,
		0.710598,
		0.67144,
		0.630801,
		0.588982,
		0.546297,
		0.503068,
		0.459622,
		0.416289,
		0.373403,
		0.331297,
		0.290302,
		0.250747,
		0.212959,
		0.17726,
		0.143965,
		0.113386,
		0.085828,
		0.061589,
		0.040959,
		0.024223,
		0.011657,
		0.003529,
		0.0001,
		0.001742,
		0.018211,
		0.049471,
		0.092596,
		0.144944,
		0.204139,
		0.268054,
		0.334797,
		0.402701,
		0.470308,
		0.536358,
		0.599783,
		0.659694,
		0.715374,
		0.766271,
		0.811993,
		0.8523,
		0.887096,
		0.91643,
		0.940484,
		0.959568,
		0.974109,
		0.984638,
		0.991772,
		0.996187,
		0.998583,
		0.999639,
		0.999959,
		1,
		0.999997,
		0.999909,
		0.999454,
		0.998223,
		0.995754,
		0.991594,
		0.985332,
		0.976625,
		0.965204,
		0.950882,
		0.933544,
		0.913148,
		0.889716,
		0.863328,
		0.834112,
		0.802239,
		0.767917,
		0.731385,
		0.692907,
		0.652767,
		0.611266,
		0.568718,
		0.525446,
		0.481783,
		0.438064,
		0.394628,
		0.351814,
		0.309961,
		0.269407,
		0.230487,
		0.193529,
		0.158861,
		0.126801,
		0.097664,
		0.071756,
		0.049379,
		0.030823,
		0.016376,
		0.006313,
		0.000903,
		0.003111,
		0.022121,
		0.055645,
		0.100747,
		0.154777,
		0.215353,
		0.280351,
		0.347883,
		0.416289,
		0.484124,
		0.550142,
		0.613292,
		0.672708,
		0.727694,
		0.777728,
		0.822445,
		0.861636,
		0.895243,
		0.92335,
		0.946178,
		0.964076,
		0.977512,
		0.987053,
		0.99335,
		0.997104,
		0.999027,
		0.999797,
		0.999987,
		1,
		0.999987,
		0.999812,
		0.999131,
		0.997501,
		0.994448,
		0.989522,
		0.982327,
		0.972543,
		0.959925,
		0.944313,
		0.925619,
		0.903828,
		0.878987,
		0.851199,
		0.820615,
		0.787425,
		0.751857,
		0.714166,
		0.674631,
		0.63355,
		0.591237,
		0.548017,
		0.504225,
		0.460201,
		0.416289,
		0.372836,
		0.330188,
		0.28869,
		0.248685,
		0.210511,
		0.174504,
		0.140991,
		0.110298,
		0.082738,
		0.058623,
		0.038254,
		0.021925,
		0.009923,
		0.002525,
		0.004882,
		0.026425,
		0.0622,
		0.109257,
		0.164939,
		0.226862,
		0.292901,
		0.361176,
		0.430036,
		0.498046,
		0.563978,
		0.6268,
		0.685666,
		0.739908,
		0.78903,
		0.832697,
		0.870737,
		0.903125,
		0.929985,
		0.951577,
		0.968292,
		0.980638,
		0.98922,
		0.994719,
		0.99786,
		0.999365,
		0.999897,
		0.999997,
		1,
		0.999959,
		0.999653,
		0.998685,
		0.996584,
		0.99287,
		0.987103,
		0.978905,
		0.967979,
		0.954108,
		0.937158,
		0.917071,
		0.893857,
		0.86759,
		0.838394,
		0.806444,
		0.771949,
		0.735156,
		0.696335,
		0.655781,
		0.613805,
		0.570735,
		0.526907,
		0.482664,
		0.438357,
		0.394338,
		0.35096,
		0.308577,
		0.26754,
		0.228195,
		0.190888,
		0.155955,
		0.12373,
		0.094539,
		0.0687,
		0.046525,
		0.028318,
		0.014374,
		0.004979,
		0.000414,
		0.00706,
		0.031129,
		0.069138,
		0.118126,
		0.175429,
		0.238661,
		0.3057,
		0.374671,
		0.443933,
		0.512066,
		0.577857,
		0.640295,
		0.698557,
		0.752002,
		0.800163,
		0.842738,
		0.879589,
		0.910731,
		0.936326,
		0.956676,
		0.972215,
		0.98349,
		0.991144,
		0.99589,
		0.99847,
		0.99961,
		0.999956,
		1,
		0.999997,
		0.999901,
		0.999412,
		0.998088,
		0.99544,
		0.990988,
		0.984305,
		0.975034,
		0.962903,
		0.947725,
		0.929393,
		0.907879,
		0.883219,
		0.855511,
		0.824905,
		0.791594,
		0.75581,
		0.717815,
		0.677898,
		0.636368,
		0.59355,
		0.549783,
		0.505413,
		0.460796,
		0.416289,
		0.372254,
		0.32905,
		0.287037,
		0.246571,
		0.208006,
		0.171689,
		0.137962,
		0.107161,
		0.079615,
		0.055645,
		0.035564,
		0.019678,
		0.008282,
		0.001665,
		0.00965,
		0.036235,
		0.076461,
		0.127354,
		0.186245,
		0.250747,
		0.318742,
		0.38836,
		0.457972,
		0.526172,
		0.591766,
		0.653763,
		0.711367,
		0.763962,
		0.811114,
		0.852555,
		0.888183,
		0.918052,
		0.942367,
		0.961473,
		0.975845,
		0.986073,
		0.992836,
		0.996876,
		0.998949,
		0.99978,
		0.999985,
		1,
		0.999986,
		0.999797,
		0.999064,
		0.997312,
		0.994038,
		0.98877,
		0.981094,
		0.970681,
		0.957286,
		0.940749,
		0.920996,
		0.898023,
		0.871897,
		0.842738,
		0.810721,
		0.776061,
		0.739008,
		0.699842,
		0.658868,
		0.61641,
		0.572806,
		0.528406,
		0.483569,
		0.438658,
		0.39404,
		0.350084,
		0.307157,
		0.265625,
		0.225849,
		0.188187,
		0.152992,
		0.120607,
		0.091374,
		0.065621,
		0.043674,
		0.025848,
		0.012447,
		0.003771,
		0.000107,
		0.000211,
		0.012657,
		0.041747,
		0.084169,
		0.13694,
		0.197385,
		0.263117,
		0.332021,
		0.402235,
		0.472143,
		0.540354,
		0.605693,
		0.667192,
		0.724081,
		0.775775,
		0.821869,
		0.862134,
		0.896506,
		0.925079,
		0.948101,
		0.965963,
		0.979184,
		0.988393,
		0.994306,
		0.997689,
		0.999313,
		0.999889,
		0.999997,
		1,
		0.999956,
		0.999625,
		0.998583,
		0.996325,
		0.992346,
		0.986181,
		0.977439,
		0.965817,
		0.951098,
		0.933156,
		0.911942,
		0.887485,
		0.859876,
		0.829262,
		0.795839,
		0.759844,
		0.721547,
		0.681244,
		0.639258,
		0.595924,
		0.551596,
		0.506634,
		0.461407,
		0.416289,
		0.371655,
		0.327881,
		0.285341,
		0.244406,
		0.205443,
		0.168815,
		0.134877,
		0.103978,
		0.076461,
		0.052658,
		0.032895,
		0.017488,
		0.006746,
		0.000966,
		0.000846,
		0.016085,
		0.047667,
		0.092264,
		0.146884,
		0.208847,
		0.275766,
		0.345531,
		0.416289,
		0.486437,
		0.5546,
		0.619625,
		0.680569,
		0.736686,
		0.787425,
		0.832415,
		0.871464,
		0.904547,
		0.931803,
		0.953525,
		0.970147,
		0.982234,
		0.990456,
		0.995564,
		0.998347,
		0.999578,
		0.999952,
		1,
		0.999997,
		0.999893,
		0.999365,
		0.99794,
		0.995096,
		0.990328,
		0.983187,
		0.973306,
		0.960409,
		0.944313,
		0.924919,
		0.902212,
		0.876247,
		0.847143,
		0.815072,
		0.780253,
		0.742944,
		0.703432,
		0.662033,
		0.619082,
		0.574932,
		0.529947,
		0.484498,
		0.438967,
		0.393734,
		0.349184,
		0.3057,
		0.263662,
		0.223446,
		0.185427,
		0.149968,
		0.117431,
		0.088168,
		0.062522,
		0.04083,
		0.023419,
		0.010606,
		0.0027,
		0.001913,
		0.019939,
		0.053998,
		0.100747,
		0.157186,
		0.220628,
		0.28869,
		0.359265,
		0.430513,
		0.500843,
		0.568899,
		0.63355,
		0.693878,
		0.749168,
		0.798899,
		0.842738,
		0.880531,
		0.912296,
		0.938217,
		0.958631,
		0.974023,
		0.984999,
		0.992272,
		0.996625,
		0.998863,
		0.999762,
		0.999984,
		1,
		0.999985,
		0.99978,
		0.998989,
		0.997104,
		0.99359,
		0.987949,
		0.979753,
		0.968661,
		0.954428,
		0.9369,
		0.916014,
		0.891784,
		0.864293,
		0.833687,
		0.800163,
		0.763962,
		0.725363,
		0.684672,
		0.642221,
		0.598361,
		0.553458,
		0.507888,
		0.462036,
		0.416289,
		0.371041,
		0.326681,
		0.283601,
		0.242186,
		0.20282,
		0.165879,
		0.131734,
		0.100747,
		0.073275,
		0.049664,
		0.030251,
		0.015366,
		0.005327,
		0.000443,
		0.003417,
		0.024223,
		0.060743,
		0.10962,
		0.167843,
		0.232725,
		0.301882,
		0.373216,
		0.444897,
		0.51535,
		0.583238,
		0.647454,
		0.707107,
		0.761512,
		0.810183,
		0.852824,
		0.889324,
		0.919743,
		0.944313,
		0.963419,
		0.977591,
		0.987485,
		0.993851,
		0.997501,
		0.999256,
		0.99988,
		0.999997,
		1,
		0.999952,
		0.999595,
		0.99847,
		0.996042,
		0.991772,
		0.985175,
		0.975845,
		0.963472,
		0.947842,
		0.928835,
		0.906419,
		0.880639,
		0.851607,
		0.819497,
		0.784529,
		0.746966,
		0.707107,
		0.665277,
		0.621825,
		0.577116,
		0.531529,
		0.485454,
		0.439285,
		0.39342,
		0.34826,
		0.304203,
		0.261647,
		0.220985,
		0.182603,
		0.146884,
		0.114202,
		0.084923,
		0.059405,
		0.037997,
		0.02104,
		0.008861,
		0.001783,
		0.005363,
		0.028941,
		0.067904,
		0.118882,
		0.178855,
		0.245133,
		0.315337,
		0.387376,
		0.459432,
		0.529947,
		0.597605,
		0.661323,
		0.720239,
		0.773702,
		0.82126,
		0.862659,
		0.897829,
		0.926878,
		0.950084,
		0.967884,
		0.980856,
		0.9897,
		0.995205,
		0.99821,
		0.999543,
		0.999948,
		1,
		0.999997,
		0.999884,
		0.999313,
		0.997777,
		0.994719,
		0.989606,
		0.981968,
		0.971428,
		0.957705,
		0.940621,
		0.92009,
		0.896111,
		0.86876,
		0.838179,
		0.804566,
		0.768167,
		0.729267,
		0.688183,
		0.645261,
		0.600863,
		0.555372,
		0.509177,
		0.462682,
		0.416289,
		0.370409,
		0.325448,
		0.281814,
		0.23991,
		0.200135,
		0.16288,
		0.128533,
		0.09747,
		0.070061,
		0.046668,
		0.027641,
		0.01332,
		0.004038,
		0.000114,
		0.007758,
		0.034098,
		0.075484,
		0.128533,
		0.190218,
		0.25785,
		0.32905,
		0.401737,
		0.474108,
		0.544621,
		0.611985,
		0.675142,
		0.73326,
		0.785723,
		0.832117,
		0.87223,
		0.906036,
		0.933692,
		0.955526,
		0.972026,
		0.983819,
		0.991651,
		0.996348,
		0.998768,
		0.999741,
		0.999983,
		1,
		0.999984,
		0.999762,
		0.998907,
		0.996876,
		0.993099,
		0.987053,
		0.978292,
		0.966466,
		0.95133,
		0.932739,
		0.910641,
		0.88507,
		0.856131,
		0.823996,
		0.788888,
		0.751077,
		0.710869,
		0.668603,
		0.62464,
		0.579359,
		0.533157,
		0.486437,
		0.439612,
		0.393097,
		0.34731,
		0.302666,
		0.259581,
		0.218463,
		0.179716,
		0.143738,
		0.110919,
		0.08164,
		0.056273,
		0.035182,
		0.018718,
		0.007225,
		0.001035,
		0.010606,
		0.039695,
		0.083483,
		0.138573,
		0.201932,
		0.270869,
		0.343012,
		0.416289,
		0.488913,
		0.559361,
		0.626365,
		0.688896,
		0.746155,
		0.797559,
		0.842738,
		0.881522,
		0.913932,
		0.940176,
		0.960634,
		0.975845,
		0.986488,
		0.99335,
		0.997294,
		0.999193,
		0.999869,
		0.999996,
		1,
		0.999948,
		0.999561,
		0.998347,
		0.995731,
		0.991144,
		0.984077,
		0.974109,
		0.960924,
		0.944313,
		0.924165,
		0.900464,
		0.873276,
		0.842738,
		0.809049,
		0.772459,
		0.73326,
		0.691782,
		0.64838,
		0.603434,
		0.557338,
		0.510503,
		0.463346,
		0.416289,
		0.369759,
		0.324181,
		0.27998,
		0.237576,
		0.197385,
		0.159817,
		0.125272,
		0.094146,
		0.066821,
		0.043674,
		0.02507,
		0.011362,
		0.002895,
		0.000232,
		0.013913,
		0.045739,
		0.091904,
		0.149001,
		0.213992,
		0.284186,
		0.357216,
		0.431024,
		0.503836,
		0.574153,
		0.64073,
		0.70257,
		0.758906,
		0.809195,
		0.853109,
		0.890522,
		0.921508,
		0.946324,
		0.965405,
		0.979343,
		0.988867,
		0.994809,
		0.99806,
		0.999504,
		0.999944,
		1,
		0.999996,
		0.999874,
		0.999256,
		0.997598,
		0.994306,
		0.988816,
		0.980638,
		0.969383,
		0.954769,
		0.936624,
		0.914874,
		0.889537,
		0.860713,
		0.828569,
		0.793332,
		0.755278,
		0.714722,
		0.672015,
		0.62753,
		0.581665,
		0.53483,
		0.487448,
		0.439948,
		0.392764,
		0.346332,
		0.301087,
		0.25746,
		0.215878,
		0.176763,
		0.140529,
		0.107582,
		0.07832,
		0.05313,
		0.032389,
		0.016464,
		0.005711,
		0.000475,
		0.000934,
		0.017685,
		0.05223,
		0.100747,
		0.159817,
		0.226396,
		0.297795,
		0.371655,
		0.44593,
		0.518865,
		0.588982,
		0.655065,
		0.716147,
		0.771497,
		0.820615,
		0.863213,
		0.899218,
		0.928752,
		0.952128,
		0.969836,
		0.982523,
		0.990967,
		0.996042,
		0.998663,
		0.999719,
		0.999981,
		1,
		0.999982,
		0.999741,
		0.998817,
		0.996625,
		0.992561,
		0.986073,
		0.976697,
		0.964076,
		0.947967,
		0.928233,
		0.904839,
		0.877839,
		0.847365,
		0.813614,
		0.776841,
		0.737347,
		0.69547,
		0.651582,
		0.606074,
		0.559361,
		0.511868,
		0.46403,
		0.416289,
		0.369091,
		0.322879,
		0.278096,
		0.235182,
		0.19457,
		0.156687,
		0.121952,
		0.090775,
		0.063557,
		0.040688,
		0.022547,
		0.009504,
		0.001913,
		0.002111,
		0.021925,
		0.059174,
		0.110014,
		0.171018,
		0.239139,
		0.311689,
		0.38632,
		0.460998,
		0.533987,
		0.603835,
		0.669354,
		0.72961,
		0.783913,
		0.831801,
		0.873037,
		0.907596,
		0.935655,
		0.957583,
		0.973927,
		0.985389,
		0.992798,
		0.997064,
		0.999123,
		0.999858,
		0.999996,
		1,
		0.999944,
		0.999524,
		0.99821,
		0.995389,
		0.990456,
		0.982876,
		0.972215,
		0.958153,
		0.940484,
		0.919112,
		0.894038,
		0.865352,
		0.833218,
		0.797864,
		0.759572,
		0.718669,
		0.675515,
		0.630499,
		0.584036,
		0.536552,
		0.488488,
		0.440294,
		0.392422,
		0.345327,
		0.299464,
		0.255282,
		0.213228,
		0.173742,
		0.137255,
		0.104192,
		0.074965,
		0.049979,
		0.029626,
		0.014288,
		0.004334,
		0.000123,
		0.003771,
		0.02664,
		0.066571,
		0.119705,
		0.182603,
		0.252217,
		0.325863,
		0.401201,
		0.476215,
		0.549189,
		0.618696,
		0.683581,
		0.742944,
		0.796135,
		0.842738,
		0.882566,
		0.915644,
		0.942207,
		0.962684,
		0.977679,
		0.987949,
		0.994372,
		0.997893,
		0.99946,
		0.999939,
		1,
		0.999996,
		0.999863,
		0.999193,
		0.9974,
		0.993851,
		0.987949,
		0.979184,
		0.967154,
		0.951577,
		0.932289,
		0.909232,
		0.882447,
		0.852057,
		0.81826,
		0.781314,
		0.741528,
		0.699252,
		0.654869,
		0.608789,
		0.561441,
		0.513271,
		0.464734,
		0.416289,
		0.368403,
		0.321539,
		0.27616,
		0.232725,
		0.191686,
		0.15349,
		0.118572,
		0.08736,
		0.060273,
		0.037715,
		0.020082,
		0.007758,
		0.001112,
		0.00592,
		0.031834,
		0.074425,
		0.12982,
		0.19457,
		0.265625,
		0.340308,
		0.416289,
		0.491569,
		0.564456,
		0.63355,
		0.697728,
		0.756129,
		0.808146,
		0.85341,
		0.891784,
		0.92335,
		0.948403,
		0.967429,
		0.981094,
		0.990212,
		0.995704,
		0.998546,
		0.999694,
		0.99998,
		1,
		0.999981,
		0.999719,
		0.998717,
		0.996348,
		0.991969,
		0.984999,
		0.974954,
		0.961473,
		0.944313,
		0.92335,
		0.898568,
		0.870045,
		0.837941,
		0.802484,
		0.763962,
		0.722712,
		0.679106,
		0.63355,
		0.586474,
		0.538323,
		0.489559,
		0.44065,
		0.39207,
		0.344293,
		0.297795,
		0.253045,
		0.210511,
		0.170651,
		0.133916,
		0.100747,
		0.071577,
		0.046826,
		0.026902,
		0.012202,
		0.003111,
		0.008564,
		0.037511,
		0.082738,
		0.140358,
		0.206915,
		0.279358,
		0.355015,
		0.431573,
		0.507048,
		0.579774,
		0.64838,
		0.711778,
		0.769149,
		0.819929,
		0.863799,
		0.900677,
		0.930704,
		0.954233,
		0.971816,
		0.984178,
		0.992187,
		0.99681,
		0.999046,
		0.999845,
		0.999996,
		1,
		0.999939,
		0.999482,
		0.99806,
		0.995012,
		0.9897,
		0.981561,
		0.970147,
		0.955135,
		0.936326,
		0.913638,
		0.887096,
		0.856815,
		0.82299,
		0.785882,
		0.745808,
		0.70313,
		0.658245,
		0.61158,
		0.563582,
		0.514717,
		0.465459,
		0.416289,
		0.367695,
		0.32016,
		0.27417,
		0.230203,
		0.188732,
		0.150223,
		0.11513,
		0.083901,
		0.05697,
		0.034761,
		0.017685,
		0.00614,
		0.000511,
		0.011711,
		0.043674,
		0.091512,
		0.151319,
		0.219635,
		0.293408,
		0.369978,
		0.447041,
		0.522638,
		0.595126,
		0.663169,
		0.725714,
		0.781985,
		0.831466,
		0.87389,
		0.909232,
		0.937694,
		0.959694,
		0.975845,
		0.986936,
		0.993888,
		0.997707,
		0.999412,
		0.999933,
		1,
		0.999996,
		0.999851,
		0.999123,
		0.997182,
		0.99335,
		0.986998,
		0.977591,
		0.96472,
		0.948101,
		0.927583,
		0.903125,
		0.874792,
		0.842738,
		0.807193,
		0.76845,
		0.726854,
		0.682792,
		0.636686,
		0.588982,
		0.540147,
		0.490663,
		0.441017,
		0.391707,
		0.343228,
		0.296078,
		0.250747,
		0.207724,
		0.167488,
		0.130509,
		0.09725,
		0.068159,
		0.043674,
		0.024223,
		0.010218,
		0.002059,
		0.000257,
		0.015366,
		0.05033,
		0.100747,
		0.162702,
		0.232725,
		0.30777,
		0.385184,
		0.462682,
		0.538323,
		0.610497,
		0.677898,
		0.739516,
		0.794618,
		0.842738,
		0.883666,
		0.917436,
		0.944313,
		0.96478,
		0.979518,
		0.989376,
		0.995329,
		0.998417,
		0.999666,
		0.999978,
		1,
		0.999979,
		0.999694,
		0.998606,
		0.996042,
		0.991318,
		0.983819,
		0.973047,
		0.958631,
		0.940336,
		0.918052,
		0.891784,
		0.861636,
		0.827802,
		0.790544,
		0.750188,
		0.707107,
		0.661713,
		0.61445,
		0.565786,
		0.516206,
		0.466205,
		0.416289,
		0.366965,
		0.318742,
		0.272124,
		0.227614,
		0.185705,
		0.146884,
		0.111627,
		0.080399,
		0.053654,
		0.031834,
		0.015366,
		0.004665,
		0.000132,
		0.001035,
		0.019535,
		0.057479,
		0.110446,
		0.174504,
		0.246181,
		0.322436,
		0.400625,
		0.478481,
		0.55409,
		0.625869,
		0.692551,
		0.753166,
		0.807029,
		0.853729,
		0.893112,
		0.925275,
		0.950551,
		0.969489,
		0.982839,
		0.99151,
		0.996527,
		0.99896,
		0.99983,
		0.999995,
		1,
		0.999933,
		0.999437,
		0.997893,
		0.994596,
		0.988867,
		0.980117,
		0.967884,
		0.951842,
		0.931803,
		0.907703,
		0.879589,
		0.84761,
		0.811993,
		0.773038,
		0.731099,
		0.686577,
		0.63991,
		0.591564,
		0.542026,
		0.4918,
		0.441396,
		0.391333,
		0.342132,
		0.294311,
		0.248386,
		0.204865,
		0.164251,
		0.127035,
		0.093699,
		0.064712,
		0.040531,
		0.021601,
		0.008351,
		0.001198,
		0.002341,
		0.024223,
		0.065127,
		0.120607,
		0.186723,
		0.259999,
		0.337397,
		0.416289,
		0.494427,
		0.569922,
		0.641225,
		0.707107,
		0.766643,
		0.8192,
		0.864419,
		0.902212,
		0.932739,
		0.956401,
		0.973821,
		0.985814,
		0.99335,
		0.997501,
		0.999358,
		0.999927,
		1,
		0.999995,
		0.999838,
		0.999046,
		0.99694,
		0.992798,
		0.985952,
		0.975845,
		0.962058,
		0.944313,
		0.922467,
		0.896506,
		0.86652,
		0.832697,
		0.795304,
		0.754672,
		0.711187,
		0.665277,
		0.617404,
		0.568056,
		0.51774,
		0.466975,
		0.416289,
		0.366214,
		0.317281,
		0.27002,
		0.224955,
		0.182603,
		0.143473,
		0.108062,
		0.076857,
		0.05033,
		0.028941,
		0.013138,
		0.003352,
		0.004182,
		0.029437,
		0.073275,
		0.131232,
		0.199356,
		0.27417,
		0.352645,
		0.432165,
		0.510503,
		0.585802,
		0.656545,
		0.721547,
		0.779928,
		0.831109,
		0.874792,
		0.91095,
		0.939815,
		0.961859,
		0.977776,
		0.98845,
		0.994912,
		0.998273,
		0.999635,
		0.999976,
		1,
		0.999977,
		0.999666,
		0.998483,
		0.995704,
		0.990599,
		0.982523,
		0.970957,
		0.955526,
		0.936004,
		0.912296,
		0.884434,
		0.852555,
		0.816885,
		0.777728,
		0.735449,
		0.690463,
		0.643226,
		0.594222,
		0.543963,
		0.492972,
		0.441786,
		0.390948,
		0.341002,
		0.292492,
		0.245958,
		0.201932,
		0.160938,
		0.123492,
		0.090096,
		0.061241,
		0.037403,
		0.019046,
		0.006618,
		0.000551,
		0.006567,
		0.035182,
		0.081927,
		0.14232,
		0.212399,
		0.28869,
		0.368169,
		0.448238,
		0.526696,
		0.601712,
		0.671812,
		0.73585,
		0.793,
		0.842738,
		0.884829,
		0.919314,
		0.946495,
		0.96692,
		0.981357,
		0.990759,
		0.996213,
		0.998863,
		0.999814,
		0.999995,
		1,
		0.999927,
		0.999385,
		0.997707,
		0.994136,
		0.987949,
		0.978531,
		0.965405,
		0.948246,
		0.926878,
		0.901258,
		0.871464,
		0.837676,
		0.800163,
		0.759263,
		0.715374,
		0.66894,
		0.620445,
		0.570395,
		0.519322,
		0.467769,
		0.416289,
		0.365439,
		0.315776,
		0.267854,
		0.222223,
		0.179424,
		0.139988,
		0.104436,
		0.073275,
		0.047001,
		0.026093,
		0.011016,
		0.002221,
		0.009504,
		0.041461,
		0.091084,
		0.153869,
		0.225849,
		0.30355,
		0.383959,
		0.464497,
		0.542987,
		0.617635,
		0.687003,
		0.749996,
		0.805839,
		0.854068,
		0.894513,
		0.927288,
		0.952769,
		0.971582,
		0.984569,
		0.992753,
		0.997272,
		0.999297,
		0.99992,
		1,
		0.999995,
		0.999823,
		0.99896,
		0.996672,
		0.992187,
		0.984799,
		0.973927,
		0.959144,
		0.940176,
		0.9169,
		0.889324,
		0.857572,
		0.821869,
		0.782523,
		0.739908,
		0.694455,
		0.646637,
		0.596961,
		0.545959,
		0.49418,
		0.442188,
		0.39055,
		0.339837,
		0.290619,
		0.243461,
		0.198921,
		0.157547,
		0.119879,
		0.086442,
		0.057749,
		0.034297,
		0.01657,
		0.005035,
		0.000143,
		0.012998,
		0.048281,
		0.100747,
		0.165879,
		0.2397,
		0.318742,
		0.400003,
		0.480925,
		0.559361,
		0.63355,
		0.7021,
		0.763962,
		0.818422,
		0.865077,
		0.903828,
		0.93486,
		0.958631,
		0.975845,
		0.987421,
		0.994448,
		0.998111,
		0.9996,
		0.999973,
		1,
		0.999974,
		0.999635,
		0.998347,
		0.995329,
		0.989805,
		0.981094,
		0.968661,
		0.952128,
		0.931276,
		0.906036,
		0.876466,
		0.842738,
		0.805122,
		0.763962,
		0.71967,
		0.672708,
		0.623575,
		0.572806,
		0.520954,
		0.468589,
		0.416289,
		0.36464,
		0.314225,
		0.265625,
		0.219416,
		0.176164,
		0.136427,
		0.100747,
		0.069658,
		0.043674,
		0.023299,
		0.009016,
		0.001294,
		0.000287,
		0.017058,
		0.055645,
		0.110919,
		0.178346,
		0.253947,
		0.334256,
		0.416289,
		0.497509,
		0.575799,
		0.649438,
		0.717079,
		0.777728,
		0.830729,
		0.875748,
		0.912756,
		0.94202,
		0.964076,
		0.979712,
		0.989923,
		0.995862,
		0.998756,
		0.999797,
		0.999994,
		1,
		0.99992,
		0.999328,
		0.997501,
		0.993626,
		0.986936,
		0.976785,
		0.962684,
		0.944313,
		0.921508,
		0.894253,
		0.862659,
		0.826947,
		0.787425,
		0.74448,
		0.698557,
		0.650148,
		0.599783,
		0.548017,
		0.495428,
		0.442603,
		0.39014,
		0.338636,
		0.28869,
		0.240892,
		0.19583,
		0.154076,
		0.116196,
		0.082738,
		0.05424,
		0.031222,
		0.014187,
		0.003623,
		0.001154,
		0.02169,
		0.063557,
		0.121599,
		0.191269,
		0.268583,
		0.350084,
		0.432804,
		0.51423,
		0.592281,
		0.665277,
		0.731919,
		0.79127,
		0.842738,
		0.886059,
		0.921283,
		0.948756,
		0.9691,
		0.983187,
		0.992087,
		0.997015,
		0.99923,
		0.999912,
		1,
		0.999994,
		0.999806,
		0.998863,
		0.996374,
		0.99151,
		0.983525,
		0.971816,
		0.955947,
		0.935655,
		0.910833,
		0.881522,
		0.847883,
		0.810183,
		0.768774,
		0.724081,
		0.676582,
		0.6268,
		0.575292,
		0.522638,
		0.469434,
		0.416289,
		0.363815,
		0.312626,
		0.263329,
		0.21653,
		0.172822,
		0.132788,
		0.096998,
		0.066007,
		0.040357,
		0.02057,
		0.007154,
		0.000596,
		0.00261,
		0.026902,
		0.072021,
		0.132788,
		0.204642,
		0.283601,
		0.366214,
		0.449532,
		0.531073,
		0.608789,
		0.681045,
		0.746597,
		0.804566,
		0.854428,
		0.895994,
		0.929393,
		0.95506,
		0.973702,
		0.986276,
		0.99393,
		0.997931,
		0.999561,
		0.999971,
		1,
		0.999972,
		0.9996,
		0.998194,
		0.994912,
		0.988925,
		0.979518,
		0.966137,
		0.948403,
		0.926112,
		0.899218,
		0.867815,
		0.832117,
		0.792437,
		0.749168,
		0.702773,
		0.653763,
		0.602692,
		0.550142,
		0.496716,
		0.443032,
		0.389717,
		0.337397,
		0.286701,
		0.238249,
		0.192655,
		0.150522,
		0.112441,
		0.078987,
		0.050721,
		0.028187,
		0.011912,
		0.002404,
		0.004665,
		0.032698,
		0.081039,
		0.144484,
		0.218463,
		0.298992,
		0.382633,
		0.466459,
		0.548017,
		0.6253,
		0.69672,
		0.761088,
		0.817592,
		0.865776,
		0.905532,
		0.937074,
		0.960924,
		0.977883,
		0.988991,
		0.99547,
		0.998635,
		0.999776,
		0.999994,
		1,
		0.999912,
		0.999265,
		0.997272,
		0.99306,
		0.985814,
		0.974858,
		0.959694,
		0.940003,
		0.915644,
		0.886629,
		0.853109,
		0.815347,
		0.773702,
		0.72861,
		0.680569,
		0.630124,
		0.577857,
		0.524376,
		0.470308,
		0.416289,
		0.362964,
		0.310976,
		0.260965,
		0.213563,
		0.169394,
		0.12907,
		0.093187,
		0.062328,
		0.037057,
		0.017921,
		0.00545,
		0.000155,
		0.007327,
		0.039085,
		0.090614,
		0.156687,
		0.232725,
		0.314747,
		0.39933,
		0.483569,
		0.565045,
		0.641793,
		0.712278,
		0.77537,
		0.830324,
		0.876762,
		0.914656,
		0.944313,
		0.966345,
		0.981647,
		0.991344,
		0.996728,
		0.999154,
		0.999903,
		1,
		0.999994,
		0.999787,
		0.998756,
		0.996042,
		0.990759,
		0.982117,
		0.969489,
		0.952436,
		0.930704,
		0.904213,
		0.873037,
		0.837381,
		0.797559,
		0.753975,
		0.707107,
		0.657487,
		0.605693,
		0.552335,
		0.498046,
		0.443475,
		0.389279,
		0.336118,
		0.28465,
		0.235528,
		0.189395,
		0.146884,
		0.108615,
		0.07519,
		0.047197,
		0.025204,
		0.009762,
		0.001403,
		0.010606,
		0.046069,
		0.100747,
		0.169394,
		0.247423,
		0.330857,
		0.416289,
		0.500843,
		0.582134,
		0.658245,
		0.727694,
		0.789417,
		0.842738,
		0.887363,
		0.92335,
		0.951098,
		0.971319,
		0.984999,
		0.99335,
		0.997729,
		0.999517,
		0.999968,
		1,
		0.999969,
		0.999561,
		0.998024,
		0.994448,
		0.987949,
		0.977776,
		0.963355,
		0.944313,
		0.920461,
		0.891784,
		0.858415,
		0.820615,
		0.778747,
		0.73326,
		0.684672,
		0.63355,
		0.580504,
		0.526172,
		0.47121,
		0.416289,
		0.362085,
		0.309274,
		0.258527,
		0.210511,
		0.165879,
		0.125272,
		0.089317,
		0.058623,
		0.033782,
		0.015366,
		0.003928,
		0.014509,
		0.053654,
		0.111441,
		0.182603,
		0.262549,
		0.34731,
		0.433497,
		0.518262,
		0.599263,
		0.674631,
		0.742944,
		0.803203,
		0.854812,
		0.897559,
		0.931598,
		0.957423,
		0.975845,
		0.987949,
		0.99503,
		0.998499,
		0.999754,
		0.999993,
		1,
		0.999903,
		0.999193,
		0.997015,
		0.992429,
		0.984569,
		0.97273,
		0.956401,
		0.935275,
		0.909232,
		0.878322,
		0.842738,
		0.802795,
		0.758906,
		0.711563,
		0.661323,
		0.608789,
		0.5546,
		0.499421,
		0.443933,
		0.388827,
		0.334797,
		0.282534,
		0.232725,
		0.186045,
		0.14316,
		0.104717,
		0.07135,
		0.043674,
		0.022285,
		0.007758,
		0.000647,
		0.000322,
		0.019046,
		0.061846,
		0.122695,
		0.19631,
		0.278096,
		0.364093,
		0.450936,
		0.535808,
		0.61641,
		0.690927,
		0.758,
		0.816702,
		0.86652,
		0.907331,
		0.939384,
		0.963278,
		0.979927,
		0.990511,
		0.996405,
		0.999069,
		0.999893,
		1,
		0.999993,
		0.999765,
		0.998635,
		0.995671,
		0.989923,
		0.980556,
		0.96692,
		0.948572,
		0.925275,
		0.89698,
		0.863799,
		0.825987,
		0.783913,
		0.738037,
		0.688896,
		0.637084,
		0.583238,
		0.528027,
		0.472143,
		0.416289,
		0.361176,
		0.307516,
		0.256014,
		0.207371,
		0.162273,
		0.121393,
		0.085389,
		0.054899,
		0.030544,
		0.012921,
		0.00261,
		0.001294,
		0.024223,
		0.070648,
		0.13451,
		0.210511,
		0.294054,
		0.381194,
		0.468589,
		0.553458,
		0.63355,
		0.707107,
		0.772836,
		0.82989,
		0.877839,
		0.916658,
		0.946697,
		0.968661,
		0.98357,
		0.9927,
		0.997501,
		0.999468,
		0.999964,
		1,
		0.999966,
		0.999517,
		0.997833,
		0.99393,
		0.986864,
		0.975845,
		0.960286,
		0.939815,
		0.914268,
		0.883666,
		0.848188,
		0.808146,
		0.763962,
		0.716147,
		0.665277,
		0.611985,
		0.556941,
		0.500843,
		0.444407,
		0.38836,
		0.333432,
		0.280351,
		0.229837,
		0.182603,
		0.139346,
		0.100747,
		0.067471,
		0.040162,
		0.019444,
		0.00592,
		0.000168,
		0.002929,
		0.030049,
		0.080064,
		0.146884,
		0.2252,
		0.310415,
		0.398598,
		0.486437,
		0.571191,
		0.650658,
		0.723143,
		0.787425,
		0.842738,
		0.888746,
		0.925522,
		0.953525,
		0.973569,
		0.986782,
		0.994536,
		0.998347,
		0.999728,
		0.999993,
		1,
		0.999893,
		0.999113,
		0.996728,
		0.991726,
		0.983187,
		0.970373,
		0.952769,
		0.930079,
		0.902212,
		0.869259,
		0.831466,
		0.789202,
		0.742944,
		0.693246,
		0.64073,
		0.586062,
		0.529947,
		0.473108,
		0.416289,
		0.360237,
		0.3057,
		0.253422,
		0.204139,
		0.158573,
		0.117431,
		0.081404,
		0.051161,
		0.027353,
		0.010606,
		0.001526,
		0.005236,
		0.036531,
		0.090096,
		0.159817,
		0.24037,
		0.327165,
		0.416289,
		0.50446,
		0.588982,
		0.667708,
		0.739008,
		0.801739,
		0.855222,
		0.899218,
		0.933906,
		0.959858,
		0.978003,
		0.989577,
		0.996042,
		0.998972,
		0.999882,
		1,
		0.999993,
		0.999741,
		0.998499,
		0.995256,
		0.988991,
		0.978822,
		0.964076,
		0.944313,
		0.919314,
		0.889065,
		0.853729,
		0.813614,
		0.769149,
		0.720862,
		0.669354,
		0.615286,
		0.559361,
		0.502313,
		0.444897,
		0.387877,
		0.332021,
		0.278096,
		0.226862,
		0.179066,
		0.135442,
		0.096707,
		0.063557,
		0.03667,
		0.016698,
		0.004273,
		0.008227,
		0.043674,
		0.100747,
		0.173305,
		0.256014,
		0.344293,
		0.43425,
		0.522638,
		0.606807,
		0.684672,
		0.754672,
		0.815748,
		0.867313,
		0.909232,
		0.941795,
		0.965692,
		0.981968,
		0.991969,
		0.997245,
		0.999412,
		0.99996,
		1,
		0.999962,
		0.999468,
		0.997618,
		0.99335,
		0.985655,
		0.973702,
		0.956891,
		0.93486,
		0.907473,
		0.874792,
		0.837049,
		0.794618,
		0.747985,
		0.697728,
		0.644493,
		0.588982,
		0.531932,
		0.474108,
		0.416289,
		0.359265,
		0.303823,
		0.250747,
		0.200812,
		0.154777,
		0.113386,
		0.077365,
		0.047417,
		0.024223,
		0.008442,
		0.000705,
		0.011912,
		0.051487,
		0.112019,
		0.187344,
		0.272124,
		0.361785,
		0.452462,
		0.540946,
		0.62464,
		0.701521,
		0.770107,
		0.829424,
		0.878987,
		0.918768,
		0.949176,
		0.971021,
		0.985471,
		0.99398,
		0.998174,
		0.999699,
		0.999992,
		0.999999,
		0.999882,
		0.999022,
		0.996405,
		0.990939,
		0.981647,
		0.967759,
		0.948756,
		0.924359,
		0.894513,
		0.859359,
		0.8192,
		0.77447,
		0.725714,
		0.673559,
		0.618696,
		0.561865,
		0.503836,
		0.445405,
		0.387376,
		0.330561,
		0.275766,
		0.223793,
		0.175429,
		0.131445,
		0.092596,
		0.059613,
		0.033208,
		0.014064,
		0.002844,
		0.0163,
		0.059973,
		0.123912,
		0.201932,
		0.28869,
		0.379627,
		0.470906,
		0.559361,
		0.642452,
		0.718226,
		0.785281,
		0.842738,
		0.890217,
		0.927805,
		0.956035,
		0.975845,
		0.988525,
		0.995632,
		0.998863,
		0.999869,
		0.999999,
		0.999992,
		0.999714,
		0.998347,
		0.99479,
		0.987949,
		0.976892,
		0.960924,
		0.939609,
		0.912756,
		0.880394,
		0.842738,
		0.800163,
		0.753166,
		0.702346,
		0.64838,
		0.592002,
		0.533987,
		0.475143,
		0.416289,
		0.358259,
		0.301882,
		0.247986,
		0.197385,
		0.150882,
		0.109257,
		0.073275,
		0.043674,
		0.02117,
		0.006452,
		0.000183,
		0.000363,
		0.021402,
		0.069138,
		0.136427,
		0.217061,
		0.3057,
		0.397801,
		0.489559,
		0.577857,
		0.660215,
		0.734755,
		0.800163,
		0.855661,
		0.900978,
		0.936326,
		0.962366,
		0.980168,
		0.991144,
		0.996954,
		0.999348,
		0.999956,
		1,
		0.999958,
		0.999412,
		0.997377,
		0.9927,
		0.984305,
		0.971319,
		0.953131,
		0.929393,
		0.900005,
		0.865077,
		0.824905,
		0.779928,
		0.730708,
		0.677898,
		0.622222,
		0.564456,
		0.505413,
		0.44593,
		0.386858,
		0.32905,
		0.273359,
		0.220628,
		0.171689,
		0.127354,
		0.088416,
		0.055645,
		0.029788,
		0.011564,
		0.001665,
		0.001462,
		0.027226,
		0.078987,
		0.149561,
		0.232725,
		0.323142,
		0.416289,
		0.5084,
		0.596407,
		0.677898,
		0.751077,
		0.814721,
		0.868161,
		0.911245,
		0.944313,
		0.968161,
		0.983995,
		0.99335,
		0.997978,
		0.999666,
		0.999991,
		0.999999,
		0.999869,
		0.99892,
		0.996042,
		0.990058,
		0.979927,
		0.964853,
		0.944313,
		0.918052,
		0.886059,
		0.848532,
		0.805839,
		0.75849,
		0.707107,
		0.652395,
		0.595126,
		0.536117,
		0.476215,
		0.416289,
		0.357216,
		0.299874,
		0.245133,
		0.193856,
		0.146884,
		0.105044,
		0.069138,
		0.039943,
		0.018211,
		0.004665,
		0.00331,
		0.033782,
		0.089522,
		0.163312,
		0.248916,
		0.341002,
		0.435073,
		0.527402,
		0.614981,
		0.69547,
		0.767158,
		0.828924,
		0.880212,
		0.920996,
		0.951752,
		0.973418,
		0.987338,
		0.995167,
		0.998739,
		0.999854,
		0.999999,
		0.999991,
		0.999683,
		0.998174,
		0.994266,
		0.986782,
		0.974739,
		0.957423,
		0.934406,
		0.905532,
		0.870879,
		0.830729,
		0.785527,
		0.73585,
		0.682378,
		0.625869,
		0.56714,
		0.507048,
		0.446476,
		0.38632,
		0.327485,
		0.270869,
		0.217363,
		0.167843,
		0.123166,
		0.084169,
		0.05166,
		0.026425,
		0.00922,
		0.000771,
		0.00592,
		0.041077,
		0.100747,
		0.177679,
		0.265625,
		0.359265,
		0.45413,
		0.54654,
		0.63355,
		0.712898,
		0.782964,
		0.842738,
		0.891784,
		0.930208,
		0.958631,
		0.978139,
		0.990212,
		0.996625,
		0.999276,
		0.999951,
		1,
		0.999954,
		0.999348,
		0.997104,
		0.991969,
		0.982792,
		0.968661,
		0.948956,
		0.92335,
		0.891784,
		0.854428,
		0.811648,
		0.763962,
		0.712016,
		0.656545,
		0.598361,
		0.538323,
		0.477327,
		0.416289,
		0.356136,
		0.297795,
		0.242186,
		0.190218,
		0.142782,
		0.100747,
		0.064959,
		0.036235,
		0.015366,
		0.003111,
		0.009304,
		0.049119,
		0.112664,
		0.192655,
		0.282841,
		0.377913,
		0.473438,
		0.565786,
		0.652082,
		0.730146,
		0.798462,
		0.856131,
		0.902848,
		0.938862,
		0.964943,
		0.982327,
		0.992637,
		0.997755,
		0.999628,
		0.99999,
		0.999999,
		0.999854,
		0.998803,
		0.995632,
		0.989067,
		0.978003,
		0.961616,
		0.939384,
		0.911086,
		0.876762,
		0.836674,
		0.79127,
		0.741144,
		0.687003,
		0.629643,
		0.569922,
		0.508744,
		0.447041,
		0.385763,
		0.325863,
		0.268292,
		0.213992,
		0.163887,
		0.118882,
		0.079857,
		0.047667,
		0.023136,
		0.00706,
		0.000201,
		0.013474,
		0.057915,
		0.125272,
		0.208236,
		0.300551,
		0.396928,
		0.492972,
		0.585111,
		0.670542,
		0.74718,
		0.813614,
		0.86907,
		0.913379,
		0.94694,
		0.970681,
		0.985994,
		0.99464,
		0.998598,
		0.999838,
		0.999999,
		0.99999,
		0.999648,
		0.997978,
		0.993675,
		0.985471,
		0.972331,
		0.953525,
		0.928639,
		0.897559,
		0.860425,
		0.817592,
		0.769588,
		0.717079,
		0.660837,
		0.601712,
		0.540612,
		0.478481,
		0.416289,
		0.355015,
		0.295641,
		0.239139,
		0.186469,
		0.138573,
		0.096367,
		0.060743,
		0.032563,
		0.012657,
		0.001825,
		0.018444,
		0.067471,
		0.138572,
		0.224414,
		0.318742,
		0.416289,
		0.512705,
		0.604481,
		0.688896,
		0.763962,
		0.828385,
		0.881522,
		0.92335,
		0.954428,
		0.975845,
		0.989154,
		0.996249,
		0.999193,
		0.999945,
		1,
		0.999948,
		0.999276,
		0.996795,
		0.991144,
		0.981094,
		0.965692,
		0.944313,
		0.916658,
		0.882719,
		0.842738,
		0.797161,
		0.746597,
		0.691782,
		0.63355,
		0.572806,
		0.510503,
		0.447628,
		0.385184,
		0.324181,
		0.265625,
		0.210511,
		0.159817,
		0.114498,
		0.075484,
		0.043674,
		0.019939,
		0.005114,
		0.000414,
		0.024222,
		0.077791,
		0.152562,
		0.241181,
		0.337397,
		0.435974,
		0.532609,
		0.623865,
		0.707107,
		0.780455,
		0.842738,
		0.893455,
		0.932739,
		0.961312,
		0.980438,
		0.991827,
		0.997501,
		0.999585,
		0.999988,
		0.999999,
		0.999838,
		0.998671,
		0.995167,
		0.987949,
		0.975845,
		0.958001,
		0.933906,
		0.903379,
		0.86652,
		0.823672,
		0.77537,
		0.722303,
		0.665277,
		0.605186,
		0.542987,
		0.47968,
		0.416289,
		0.353853,
		0.293408,
		0.235987,
		0.182603,
		0.134253,
		0.091904,
		0.056497,
		0.028941,
		0.010111,
		0.000846,
		0.001665,
		0.03082,
		0.088879,
		0.167239,
		0.258526,
		0.3565,
		0.455958,
		0.552654,
		0.643226,
		0.725136,
		0.796619,
		0.856637,
		0.904839,
		0.941522,
		0.967586,
		0.984468,
		0.994038,
		0.998436,
		0.999819,
		0.999999,
		0.999988,
		0.999607,
		0.997755,
		0.993005,
		0.983995,
		0.969633,
		0.949176,
		0.922235,
		0.888746,
		0.848922,
		0.803203,
		0.752213,
		0.69672,
		0.637598,
		0.575799,
		0.512331,
		0.448238,
		0.384583,
		0.322436,
		0.262862,
		0.206915,
		0.155629,
		0.110014,
		0.071053,
		0.039695,
		0.016857,
		0.003417,
		0.003769,
		0.038245,
		0.100736,
		0.182595,
		0.276435,
		0.376029,
		0.476214,
		0.572806,
		0.662527,
		0.742944,
		0.812415,
		0.870045,
		0.915644,
		0.949683,
		0.973246,
		0.987949,
		0.99582,
		0.999098,
		0.999939,
		1,
		0.999942,
		0.999193,
		0.996443,
		0.990212,
		0.979184,
		0.962366,
		0.939136,
		0.909232,
		0.872709,
		0.82989,
		0.781314,
		0.727694,
		0.669872,
		0.608789,
		0.545454,
		0.480925,
		0.416289,
		0.352645,
		0.291093,
		0.232725,
		0.178617,
		0.12982,
		0.08736,
		0.05223,
		0.025389,
		0.007758,
		0.000221,
		0.00674,
		0.046503,
		0.113359,
		0.198621,
		0.294891,
		0.395962,
		0.496713,
		0.59303,
		0.681729,
		0.760489,
		0.827802,
		0.882926,
		0.925841,
		0.957205,
		0.978292,
		0.990902,
		0.99721,
		0.999535,
		0.999987,
		0.999999,
		0.999819,
		0.99852,
		0.99464,
		0.986687,
		0.973418,
		0.953955,
		0.927805,
		0.894835,
		0.855222,
		0.809397,
		0.758,
		0.701826,
		0.641793,
		0.578906,
		0.51423,
		0.448872,
		0.383959,
		0.320624,
		0.259999,
		0.203198,
		0.151319,
		0.105431,
		0.066571,
		0.035743,
		0.013913,
		0.002008,
		0.01059,
		0.055596,
		0.126741,
		0.215301,
		0.313873,
		0.416269,
		0.51742,
		0.613289,
		0.70079,
		0.777728,
		0.842738,
		0.895243,
		0.935405,
		0.964076,
		0.98273,
		0.99335,
		0.998251,
		0.999797,
		0.999999,
		0.999987,
		0.999561,
		0.997501,
		0.992245,
		0.982327,
		0.9666,
		0.944313,
		0.915109,
		0.878987,
		0.836245,
		0.787425,
		0.73326,
		0.674631,
		0.612528,
		0.548017,
		0.482221,
		0.416289,
		0.35139,
		0.28869,
		0.229348,
		0.174504,
		0.125272,
		0.082738,
		0.047952,
		0.021925,
		0.005631,
		0.015327,
		0.065521,
		0.140868,
		0.232614,
		0.333351,
		0.436917,
		0.538297,
		0.633539,
		0.719666,
		0.794617,
		0.857183,
		0.906963,
		0.944313,
		0.970291,
		0.986574,
		0.995329,
		0.998989,
		0.999931,
		1,
		0.999935,
		0.999098,
		0.996042,
		0.989154,
		0.977027,
		0.958631,
		0.933353,
		0.900978,
		0.861636,
		0.815748,
		0.763962,
		0.707107,
		0.646144,
		0.582134,
		0.516206,
		0.449532,
		0.383309,
		0.318742,
		0.257029,
		0.199356,
		0.146884,
		0.100747,
		0.062044,
		0.031834,
		0.011138,
		0.000934,
		0.020955,
		0.076267,
		0.155717,
		0.250527,
		0.353287,
		0.457861,
		0.559298,
		0.653732,
		0.738309,
		0.811109,
		0.871094,
		0.918052,
		0.952544,
		0.975845,
		0.989843,
		0.996876,
		0.999478,
		0.999985,
		0.999999,
		0.999797,
		0.998347,
		0.994038,
		0.985256,
		0.970681,
		0.949417,
		0.920996,
		0.885348,
		0.842738,
		0.793707,
		0.739008,
		0.679561,
		0.61641,
		0.550683,
		0.483569,
		0.416289,
		0.350084,
		0.286194,
		0.225849,
		0.170259,
		0.120607,
		0.078042,
		0.043674,
		0.018575,
		0.003771,
		0.000471,
		0.027471,
		0.087814,
		0.171255,
		0.268996,
		0.373628,
		0.479047,
		0.580364,
		0.673812,
		0.756663,
		0.827155,
		0.884429,
		0.928478,
		0.960083,
		0.980744,
		0.992561,
		0.998037,
		0.999771,
		0.999999,
		0.999985,
		0.999508,
		0.99721,
		0.991379,
		0.980438,
		0.963183,
		0.938862,
		0.907163,
		0.868161,
		0.822257,
		0.770107,
		0.71257,
		0.650658,
		0.58549,
		0.518262,
		0.450219,
		0.382633,
		0.316784,
		0.253947,
		0.195381,
		0.14232,
		0.095965,
		0.057479,
		0.027987,
		0.008564,
		0.000244,
		0.001888,
		0.034862,
		0.100129,
		0.187433,
		0.287961,
		0.394307,
		0.500403,
		0.601425,
		0.69371,
		0.774666,
		0.842698,
		0.897143,
		0.938211,
		0.966918,
		0.984999,
		0.994763,
		0.998863,
		0.999922,
		1,
		0.999927,
		0.998989,
		0.995583,
		0.987949,
		0.974587,
		0.954428,
		0.926878,
		0.891784,
		0.849368,
		0.800163,
		0.744944,
		0.684672,
		0.620445,
		0.553458,
		0.484973,
		0.416289,
		0.348725,
		0.283601,
		0.222223,
		0.165879,
		0.115823,
		0.073275,
		0.039412,
		0.015366,
		0.002221,
		0.004255,
		0.043104,
		0.113164,
		0.204182,
		0.30734,
		0.415238,
		0.52184,
		0.622395,
		0.713346,
		0.792245,
		0.857677,
		0.909189,
		0.947222,
		0.973042,
		0.988624,
		0.996489,
		0.999412,
		0.999984,
		0.999999,
		0.999771,
		0.998148,
		0.99335,
		0.983628,
		0.967586,
		0.944313,
		0.913379,
		0.874792,
		0.828924,
		0.776439,
		0.718226,
		0.655346,
		0.588982,
		0.520404,
		0.450936,
		0.381929,
		0.314747,
		0.250747,
		0.191269,
		0.137622,
		0.091084,
		0.052888,
		0.024223,
		0.006231,
		0.007563,
		0.052157,
		0.126853,
		0.221416,
		0.327031,
		0.436307,
		0.543246,
		0.643164,
		0.732617,
		0.80931,
		0.872018,
		0.920512,
		0.955481,
		0.978449,
		0.991646,
		0.997788,
		0.999741,
		0.999999,
		0.999984,
		0.999446,
		0.996876,
		0.99039,
		0.978292,
		0.959323,
		0.932739,
		0.898284,
		0.856131,
		0.806797,
		0.751077,
		0.689972,
		0.62464,
		0.556348,
		0.486437,
		0.416289,
		0.34731,
		0.280903,
		0.218463,
		0.161357,
		0.110919,
		0.068445,
		0.035182,
		0.012329,
		0.001035,
		0.011792,
		0.061965,
		0.141106,
		0.239021,
		0.346903,
		0.457378,
		0.564478,
		0.663597,
		0.751401,
		0.825756,
		0.885637,
		0.931051,
		0.962953,
		0.983139,
		0.994091,
		0.998711,
		0.999911,
		1,
		0.999917,
		0.998863,
		0.995056,
		0.986574,
		0.971816,
		0.949683,
		0.91961,
		0.881522,
		0.835752,
		0.782964,
		0.724081,
		0.660215,
		0.592618,
		0.522638,
		0.451682,
		0.381194,
		0.312626,
		0.247423,
		0.187012,
		0.132788,
		0.086107,
		0.048281,
		0.02057,
		0.004182,
		0.016906,
		0.07245,
		0.155812,
		0.256856,
		0.366798,
		0.478279,
		0.585368,
		0.68353,
		0.769549,
		0.841453,
		0.898429,
		0.940734,
		0.969601,
		0.98711,
		0.995994,
		0.999317,
		0.999976,
		0.999997,
		0.999741,
		0.997917,
		0.992561,
		0.981771,
		0.964076,
		0.938558,
		0.904839,
		0.863025,
		0.813614,
		0.757413,
		0.69547,
		0.629005,
		0.559361,
		0.487964,
		0.416289,
		0.345833,
		0.278096,
		0.214561,
		0.156687,
		0.105894,
		0.063557,
		0.031004,
		0.009504,
		0.000272,
		0.022848,
		0.083512,
		0.170831,
		0.274752,
		0.386524,
		0.498808,
		0.605712,
		0.702768,
		0.786881,
		0.856247,
		0.910271,
		0.949475,
		0.97538,
		0.990359,
		0.997388,
		0.999659,
		0.999981,
		0.999976,
		0.999373,
		0.996489,
		0.989256,
		0.975845,
		0.954949,
		0.925841,
		0.888342,
		0.842738,
		0.789689,
		0.730146,
		0.665277,
		0.596407,
		0.524968,
		0.452462,
		0.380428,
		0.310415,
		0.243966,
		0.182603,
		0.127813,
		0.081039,
		0.043674,
		0.017058,
		0.00247,
		0.000502,
		0.029542,
		0.095026,
		0.185998,
		0.29251,
		0.405858,
		0.51873,
		0.62527,
		0.721083,
		0.803188,
		0.869955,
		0.921018,
		0.957169,
		0.980234,
		0.992877,
		0.998305,
		0.999787,
		0.999952,
		0.999888,
		0.998711,
		0.994447,
		0.984999,
		0.968661,
		0.944313,
		0.911434,
		0.870045,
		0.820615,
		0.763962,
		0.701178,
		0.63355,
		0.562504,
		0.489559,
		0.416289,
		0.344293,
		0.275172,
		0.210511,
		0.151865,
		0.100747,
		0.058623,
		0.026902,
		0.006932,
		0.001976,
		0.036891,
		0.106843,
		0.201119,
		0.309899,
		0.424544,
		0.537774,
		0.643772,
		0.73821,
		0.818227,
		0.882368,
		0.930499,
		0.963696,
		0.984094,
		0.994648,
		0.998768,
		0.999737,
		0.999885,
		0.999659,
		0.997633,
		0.991646,
		0.979644,
		0.960083,
		0.932051,
		0.895243,
		0.849883,
		0.796619,
		0.73643,
		0.670542,
		0.600358,
		0.527402,
		0.453277,
		0.379627,
		0.308108,
		0.24037,
		0.178037,
		0.122695,
		0.075884,
		0.039085,
		0.013723,
		0.001154,
		0.004361,
		0.044773,
		0.118789,
		0.215973,
		0.326664,
		0.442298,
		0.555639,
		0.660911,
		0.753855,
		0.831725,
		0.893244,
		0.938519,
		0.968915,
		0.986877,
		0.995644,
		0.998789,
		0.999522,
		0.999737,
		0.999178,
		0.995994,
		0.987932,
		0.973042,
		0.949977,
		0.918052,
		0.877185,
		0.827802,
		0.770731,
		0.707107,
		0.638287,
		0.565786,
		0.491227,
		0.416289,
		0.342684,
		0.272124,
		0.206304,
		0.146884,
		0.095481,
		0.053654,
		0.022905,
		0.004665,
		0.007571,
		0.053042,
		0.130667,
		0.23032,
		0.342523,
		0.458811,
		0.572,
		0.676358,
		0.767694,
		0.84338,
		0.90232,
		0.944861,
		0.972665,
		0.988486,
		0.995825,
		0.998361,
		0.999123,
		0.999416,
		0.998305,
		0.993631,
		0.983139,
		0.965039,
		0.938211,
		0.90221,
		0.857183,
		0.803759,
		0.742944,
		0.676022,
		0.604481,
		0.529947,
		0.45413,
		0.37879,
		0.3057,
		0.236626,
		0.173305,
		0.117431,
		0.070648,
		0.034535,
		0.010606,
		0.000304,
		0.011499,
		0.061533,
		0.142261,
		0.243899,
		0.35718,
		0.473757,
		0.58651,
		0.689759,
		0.779382,
		0.852871,
		0.909308,
		0.949288,
		0.974769,
		0.988808,
		0.995131,
		0.99745,
		0.998473,
		0.998789,
		0.996864,
		0.990359,
		0.977092,
		0.955481,
		0.924657,
		0.884429,
		0.835175,
		0.777728,
		0.713267,
		0.643226,
		0.569217,
		0.492972,
		0.416289,
		0.341002,
		0.268945,
		0.201932,
		0.14174,
		0.090096,
		0.048665,
		0.019046,
		0.002763,
		0.016015,
		0.070064,
		0.153342,
		0.256439,
		0.370327,
		0.486799,
		0.598814,
		0.700751,
		0.788563,
		0.859861,
		0.913908,
		0.95155,
		0.975038,
		0.987715,
		0.993485,
		0.995988,
		0.99745,
		0.997669,
		0.994648,
		0.98597,
		0.969601,
		0.944206,
		0.909189,
		0.864619,
		0.811109,
		0.749695,
		0.681729,
		0.608789,
		0.532609,
		0.455022,
		0.377913,
		0.303183,
		0.232725,
		0.168399,
		0.112019,
		0.065341,
		0.030049,
		0.007758,
		0.020971,
		0.078437,
		0.163672,
		0.267665,
		0.381656,
		0.497602,
		0.608558,
		0.708972,
		0.794878,
		0.864009,
		0.915816,
		0.951392,
		0.973275,
		0.985067,
		0.990785,
		0.993865,
		0.995868,
		0.995825,
		0.991411,
		0.980234,
		0.960466,
		0.931051,
		0.891682,
		0.842698,
		0.784947,
		0.719666,
		0.648379,
		0.572806,
		0.494799,
		0.416289,
		0.339242,
		0.265625,
		0.197385,
		0.136427,
		0.084596,
		0.043674,
		0.015366,
		0.001294,
		0.000432,
		0.026203,
		0.086449,
		0.173013,
		0.277308,
		0.390871,
		0.505845,
		0.615402,
		0.714073,
		0.797981,
		0.864989,
		0.914735,
		0.948562,
		0.969283,
		0.980718,
		0.986903,
		0.990912,
		0.993485,
		0.992978,
		0.986877,
		0.9729,
		0.949475,
		0.915849,
		0.872018,
		0.818594,
		0.756663,
		0.687664,
		0.613289,
		0.535398,
		0.455958,
		0.376995,
		0.300551,
		0.228658,
		0.163312,
		0.106458,
		0.059973,
		0.025659,
		0.005236,
		0.001638,
		0.031536,
		0.093895,
		0.181135,
		0.285117,
		0.397697,
		0.511233,
		0.619035,
		0.715733,
		0.797554,
		0.862495,
		0.910389,
		0.942826,
		0.962873,
		0.974514,
		0.981686,
		0.986903,
		0.99,
		0.988808,
		0.980739,
		0.963696,
		0.936403,
		0.898429,
		0.85008,
		0.792245,
		0.726248,
		0.653732,
		0.576554,
		0.496713,
		0.416289,
		0.337397,
		0.262156,
		0.192655,
		0.130939,
		0.078987,
		0.038704,
		0.011912,
		0.000342,
		0.003467,
		0.036791,
		0.100576,
		0.187825,
		0.290865,
		0.401894,
		0.513512,
		0.61919,
		0.713679,
		0.793321,
		0.856264,
		0.902538,
		0.933976,
		0.953873,
		0.966297,
		0.974941,
		0.98155,
		0.985067,
		0.982956,
		0.972665,
		0.952336,
		0.921018,
		0.878618,
		0.825756,
		0.763598,
		0.69371,
		0.617935,
		0.538297,
		0.456932,
		0.376029,
		0.297794,
		0.224414,
		0.158037,
		0.100747,
		0.05456,
		0.021402,
		0.003111,
		0.005759,
		0.04179,
		0.106308,
		0.192897,
		0.294361,
		0.403267,
		0.512479,
		0.615657,
		0.707694,
		0.785067,
		0.846087,
		0.890993,
		0.921848,
		0.942136,
		0.955914,
		0.966441,
		0.974514,
		0.978299,
		0.975038,
		0.962301,
		0.938519,
		0.903084,
		0.856247,
		0.798942,
		0.732617,
		0.659076,
		0.580364,
		0.498671,
		0.416269,
		0.335457,
		0.258526,
		0.187731,
		0.125272,
		0.073275,
		0.033782,
		0.008741,
		0.008346,
		0.046363,
		0.110928,
		0.196197,
		0.295462,
		0.401677,
		0.507996,
		0.608295,
		0.697635,
		0.77265,
		0.83183,
		0.875627,
		0.906329,
		0.927549,
		0.943206,
		0.955914,
		0.96542,
		0.969283,
		0.964651,
		0.949288,
		0.921945,
		0.882368,
		0.831152,
		0.769549,
		0.699281,
		0.622395,
		0.541136,
		0.457861,
		0.374977,
		0.294891,
		0.21998,
		0.152562,
		0.094888,
		0.049119,
		0.017323,
		0.001462,
		0.011064,
		0.050357,
		0.1143,
		0.197612,
		0.294076,
		0.397048,
		0.5,
		0.597047,
		0.683445,
		0.756013,
		0.813438,
		0.856393,
		0.887367,
		0.910034,
		0.928019,
		0.943055,
		0.953873,
		0.957597,
		0.951392,
		0.93327,
		0.90232,
		0.858647,
		0.803188,
		0.737504,
		0.663597,
		0.583747,
		0.500403,
		0.416091,
		0.333351,
		0.254699,
		0.182595,
		0.11942,
		0.067471,
		0.028941,
		0.00592,
		0.013754,
		0.053638,
		0.116324,
		0.197077,
		0.29017,
		0.389377,
		0.488506,
		0.581942,
		0.665163,
		0.735198,
		0.790953,
		0.833327,
		0.864981,
		0.889556,
		0.910191,
		0.927549,
		0.939475,
		0.942826,
		0.934871,
		0.913908,
		0.879373,
		0.831725,
		0.772234,
		0.702768,
		0.625599,
		0.543246,
		0.458357,
		0.373628,
		0.291735,
		0.215301,
		0.146868,
		0.088879,
		0.043674,
		0.013474,
		0.000387,
		0.000256,
		0.01627,
		0.056098,
		0.116937,
		0.194576,
		0.283774,
		0.378733,
		0.473615,
		0.563102,
		0.642924,
		0.710347,
		0.764515,
		0.806556,
		0.839261,
		0.866118,
		0.889556,
		0.909081,
		0.921848,
		0.924584,
		0.914735,
		0.890903,
		0.852871,
		0.801442,
		0.73821,
		0.66534,
		0.585368,
		0.501051,
		0.415238,
		0.330792,
		0.250527,
		0.177175,
		0.113359,
		0.061582,
		0.024222,
		0.003529,
		0.000913,
		0.018485,
		0.057659,
		0.116117,
		0.190145,
		0.274981,
		0.365257,
		0.455512,
		0.540744,
		0.616965,
		0.681705,
		0.734367,
		0.776299,
		0.810366,
		0.839755,
		0.865957,
		0.887367,
		0.900656,
		0.902538,
		0.890684,
		0.864009,
		0.822636,
		0.767694,
		0.701088,
		0.62527,
		0.543038,
		0.457378,
		0.371341,
		0.287961,
		0.21019,
		0.140868,
		0.082699,
		0.038245,
		0.009922,
		0.001814,
		0.020295,
		0.058278,
		0.113883,
		0.18387,
		0.263942,
		0.349164,
		0.434461,
		0.515172,
		0.587617,
		0.649616,
		0.700845,
		0.742861,
		0.77852,
		0.810529,
		0.839261,
		0.86217,
		0.875627,
		0.876427,
		0.862495,
		0.83306,
		0.788563,
		0.730447,
		0.660911,
		0.582681,
		0.498808,
		0.412507,
		0.327031,
		0.245582,
		0.171255,
		0.106998,
		0.055596,
		0.019668,
		0.001665,
		0.002816,
		0.021621,
		0.057942,
		0.110298,
		0.175889,
		0.250868,
		0.330732,
		0.4108,
		0.486772,
		0.555299,
		0.614514,
		0.664376,
		0.706621,
		0.743996,
		0.77852,
		0.809382,
		0.833327,
		0.84658,
		0.846087,
		0.830044,
		0.797981,
		0.750642,
		0.689759,
		0.617806,
		0.537774,
		0.452958,
		0.366798,
		0.282748,
		0.204182,
		0.134336,
		0.076267,
		0.032845,
		0.00674,
		0.003798,
		0.022412,
		0.056675,
		0.105461,
		0.166381,
		0.236018,
		0.310294,
		0.38493,
		0.455999,
		0.520504,
		0.57691,
		0.625459,
		0.668022,
		0.707107,
		0.743835,
		0.776299,
		0.800764,
		0.813438,
		0.811468,
		0.793321,
		0.758816,
		0.708972,
		0.645788,
		0.572,
		0.490845,
		0.405858,
		0.320699,
		0.239021,
		0.164372,
		0.100129,
		0.049465,
		0.015327,
		0.000442,
		0.004664,
		0.022647,
		0.054533,
		0.099509,
		0.155564,
		0.219692,
		0.288229,
		0.3573,
		0.423361,
		0.483784,
		0.537373,
		0.584648,
		0.627545,
		0.668179,
		0.706621,
		0.74007,
		0.764515,
		0.77625,
		0.77265,
		0.75245,
		0.715733,
		0.663773,
		0.598814,
		0.523831,
		0.442298,
		0.357979,
		0.274752,
		0.19647,
		0.126853,
		0.069425,
		0.027471,
		0.004022,
		0.005339,
		0.02233,
		0.0516,
		0.092608,
		0.143689,
		0.202223,
		0.26495,
		0.328396,
		0.389406,
		0.445728,
		0.496513,
		0.542528,
		0.585694,
		0.627545,
		0.667076,
		0.700845,
		0.724736,
		0.735198,
		0.729854,
		0.707694,
		0.669042,
		0.615402,
		0.549241,
		0.473757,
		0.392648,
		0.309899,
		0.229604,
		0.155812,
		0.092422,
		0.043104,
		0.011259,
		8.4e-005,
		0.005777,
		0.021495,
		0.047987,
		0.084949,
		0.131026,
		0.183964,
		0.240889,
		0.298721,
		0.354699,
		0.406945,
		0.454953,
		0.499696,
		0.542966,
		0.585544,
		0.625459,
		0.658875,
		0.681705,
		0.690604,
		0.683445,
		0.659459,
		0.61919,
		0.564348,
		0.497602,
		0.422357,
		0.342523,
		0.262302,
		0.185998,
		0.11786,
		0.061965,
		0.022135,
		0.001888,
		0.000268,
		0.005956,
		0.020197,
		0.043827,
		0.076739,
		0.11786,
		0.165276,
		0.216483,
		0.268784,
		0.319805,
		0.368043,
		0.413311,
		0.456733,
		0.499835,
		0.542528,
		0.582095,
		0.614514,
		0.635826,
		0.642924,
		0.633929,
		0.608295,
		0.566765,
		0.511233,
		0.444555,
		0.370327,
		0.292662,
		0.215973,
		0.144773,
		0.083512,
		0.036449,
		0.007563,
		0.000473,
		0.005877,
		0.018512,
		0.039266,
		0.068196,
		0.104482,
		0.146521,
		0.192162,
		0.239079,
		0.285274,
		0.329606,
		0.372177,
		0.414182,
		0.456733,
		0.498867,
		0.537373,
		0.568211,
		0.587617,
		0.592741,
		0.581942,
		0.55488,
		0.512479,
		0.456801,
		0.390871,
		0.318468,
		0.243899,
		0.171783,
		0.106843,
		0.053735,
		0.016906,
		0.000494,
		0.000644,
		0.005563,
		0.016531,
		0.03446,
		0.05954,
		0.091174,
		0.128047,
		0.168337,
		0.210073,
		0.251619,
		0.292176,
		0.33209,
		0.372533,
		0.414057,
		0.454953,
		0.491749,
		0.520504,
		0.537699,
		0.540744,
		0.528227,
		0.5,
		0.45715,
		0.401894,
		0.337415,
		0.267665,
		0.197141,
		0.130667,
		0.073179,
		0.029542,
		0.0044,
		0.000748,
		0.005051,
		0.014355,
		0.029569,
		0.050988,
		0.078208,
		0.110179,
		0.145385,
		0.182192,
		0.219306,
		0.256238,
		0.293526,
		0.3322,
		0.372177,
		0.411201,
		0.445728,
		0.472002,
		0.486772,
		0.487707,
		0.473615,
		0.444525,
		0.401677,
		0.347429,
		0.285117,
		0.218859,
		0.153342,
		0.093595,
		0.044773,
		0.011958,
		0.000774,
		0.004394,
		0.012091,
		0.024749,
		0.04274,
		0.065832,
		0.09321,
		0.123644,
		0.155812,
		0.188738,
		0.222209,
		0.256879,
		0.293526,
		0.33144,
		0.368043,
		0.399855,
		0.423361,
		0.435589,
		0.434461,
		0.418991,
		0.389377,
		0.347001,
		0.294361,
		0.234937,
		0.173013,
		0.113465,
		0.061533,
		0.0226,
		0.001976,
		0.000727,
		0.003651,
		0.009842,
		0.020142,
		0.03498,
		0.054265,
		0.077394,
		0.103399,
		0.131247,
		0.160248,
		0.190422,
		0.222461,
		0.256786,
		0.292176,
		0.325921,
		0.354699,
		0.375266,
		0.38493,
		0.381856,
		0.365257,
		0.33549,
		0.294076,
		0.243645,
		0.187825,
		0.131066,
		0.078437,
		0.0354,
		0.007571,
		0.000623,
		0.00288,
		0.007705,
		0.015878,
		0.027863,
		0.043688,
		0.062938,
		0.08488,
		0.108744,
		0.134094,
		0.161129,
		0.190497,
		0.222209,
		0.254699,
		0.285274,
		0.310826,
		0.328396,
		0.335565,
		0.330732,
		0.313298,
		0.283774,
		0.243819,
		0.196197,
		0.144681,
		0.093895,
		0.049109,
		0.016015,
		0.000483,
		1.6e-005,
		0.000485,
		0.002141,
		0.005762,
		0.012061,
		0.021513,
		0.034243,
		0.049997,
		0.068256,
		0.08848,
		0.110452,
		0.134495,
		0.161129,
		0.189978,
		0.219306,
		0.246526,
		0.268784,
		0.283404,
		0.288229,
		0.281881,
		0.263942,
		0.235077,
		0.197077,
		0.152846,
		0.106308,
		0.062266,
		0.026203,
		0.004055,
		1.8e-005,
		0.000339,
		0.001481,
		0.00408,
		0.008767,
		0.016015,
		0.026022,
		0.038672,
		0.05363,
		0.070559,
		0.089421,
		0.110603,
		0.134436,
		0.160248,
		0.186264,
		0.210073,
		0.229082,
		0.240889,
		0.24359,
		0.236018,
		0.217932,
		0.190145,
		0.154586,
		0.1143,
		0.073371,
		0.036791,
		0.010267,
		1.4e-005,
		0.000207,
		0.000936,
		0.002699,
		0.006041,
		0.011415,
		0.01907,
		0.029006,
		0.041042,
		0.055015,
		0.071024,
		0.089464,
		0.110452,
		0.13314,
		0.155812,
		0.176263,
		0.192162,
		0.201372,
		0.202223,
		0.193752,
		0.175889,
		0.149594,
		0.116937,
		0.08111,
		0.046363,
		0.017889,
		0.001638,
		0.000106,
		0.000524,
		0.001636,
		0.003894,
		0.007717,
		0.013384,
		0.020985,
		0.030469,
		0.041813,
		0.055213,
		0.071024,
		0.089172,
		0.108744,
		0.128143,
		0.145385,
		0.158393,
		0.165276,
		0.164591,
		0.155564,
		0.138292,
		0.113883,
		0.084551,
		0.053638,
		0.025572,
		0.005759,
		4.1e-005,
		0.000248,
		0.000882,
		0.002303,
		0.004882,
		0.00891,
		0.014543,
		0.021831,
		0.030859,
		0.041882,
		0.055184,
		0.070559,
		0.087112,
		0.103399,
		0.117656,
		0.128047,
		0.132916,
		0.131026,
		0.121789,
		0.105461,
		0.083299,
		0.057659,
		0.032043,
		0.011064,
		0.000352,
		9e-005,
		0.000402,
		0.001213,
		0.002835,
		0.005553,
		0.009563,
		0.014994,
		0.022005,
		0.030877,
		0.041813,
		0.054548,
		0.068256,
		0.081659,
		0.09321,
		0.101297,
		0.104482,
		0.101724,
		0.092608,
		0.077548,
		0.057942,
		0.036295,
		0.01627,
		0.002672,
		2e-005,
		0.000141,
		0.00054,
		0.001466,
		0.003177,
		0.005889,
		0.00978,
		0.015059,
		0.022005,
		0.030755,
		0.041042,
		0.052138,
		0.062938,
		0.072094,
		0.078208,
		0.080042,
		0.076739,
		0.068049,
		0.054533,
		0.037732,
		0.020295,
		0.00605,
		3.1e-005,
		0.000185,
		0.000642,
		0.001619,
		0.003331,
		0.005981,
		0.0098,
		0.01505,
		0.021831,
		0.02991,
		0.038672,
		0.047176,
		0.054265,
		0.058738,
		0.05954,
		0.055989,
		0.047987,
		0.036236,
		0.022412,
		0.009308,
		0.000913,
		4e-005,
		0.000216,
		0.000698,
		0.001679,
		0.003363,
		0.005985,
		0.00978,
		0.014841,
		0.020985,
		0.027714,
		0.034243,
		0.039592,
		0.04274,
		0.042807,
		0.039266,
		0.032163,
		0.02233,
		0.011571,
		0.002816,
		4.5e-005,
		0.000232,
		0.000718,
		0.001688,
		0.003363,
		0.005956,
		0.009563,
		0.014064,
		0.01907,
		0.023941,
		0.027863,
		0.02998,
		0.029569,
		0.02625,
		0.020197,
		0.012361,
		0.004664,
		0.000161,
		4.8e-005,
		0.000237,
		0.000719,
		0.001687,
		0.003331,
		0.00576,
		0.00891,
		0.012498,
		0.016015,
		0.018798,
		0.020142,
		0.019471,
		0.016531,
		0.011614,
		0.005777,
		0.001043,
		4.9e-005,
		0.000237,
		0.000718,
		0.001659,
		0.003177,
		0.005261,
		0.007717,
		0.01016,
		0.012061,
		0.012854,
		0.012091,
		0.009635,
		0.005877,
		0.001962,
		4.9e-005,
		0.000235,
		0.000698,
		0.001555,
		0.002835,
		0.004423,
		0.006041,
		0.007282,
		0.007705,
		0.006974,
		0.005051,
		0.002411,
		0.000268,
		4.8e-005,
		0.000226,
		0.000642,
		0.001351,
		0.002303,
		0.003312,
		0.00408,
		0.004271,
		0.003651,
		0.002263,
		0.000644,
		4.5e-005,
		0.000203,
		0.00054,
		0.001055,
		0.001636,
		0.002079,
		0.002141,
		0.00167,
		0.000774,
		3.2e-005,
		4e-005,
		0.000165,
		0.000402,
		0.000703,
		0.000936,
		0.000938,
		0.000623,
		0.000143,
		3.1e-005,
		0.000116,
		0.000248,
		0.000355,
		0.000339,
		0.000161,
		2e-005,
		6.4e-005,
		0.000106,
		9.1e-005,
		1.6e-005,
		2.1e-005,
		1.4e-005,
		1.4e-005,
		2.1e-005,
		1.6e-005,
		9.1e-005,
		0.000106,
		6.4e-005,
		2e-005,
		0.000161,
		0.000339,
		0.000355,
		0.000248,
		0.000116,
		3.1e-005,
		0.000143,
		0.000623,
		0.000938,
		0.000936,
		0.000703,
		0.000402,
		0.000165,
		4e-005,
		3.2e-005,
		0.000774,
		0.00167,
		0.002141,
		0.002079,
		0.001636,
		0.001055,
		0.00054,
		0.000203,
		4.5e-005,
		0.000644,
		0.002263,
		0.003651,
		0.004271,
		0.00408,
		0.003312,
		0.002303,
		0.001351,
		0.000642,
		0.000226,
		4.8e-005,
		0.000268,
		0.002411,
		0.005051,
		0.006974,
		0.007705,
		0.007282,
		0.006041,
		0.004423,
		0.002835,
		0.001555,
		0.000698,
		0.000235,
		4.9e-005,
		0.001962,
		0.005877,
		0.009635,
		0.012091,
		0.012854,
		0.012061,
		0.01016,
		0.007717,
		0.005261,
		0.003177,
		0.001659,
		0.000718,
		0.000237,
		4.9e-005,
		0.001043,
		0.005777,
		0.011614,
		0.016531,
		0.019471,
		0.020142,
		0.018798,
		0.016015,
		0.012498,
		0.00891,
		0.00576,
		0.003331,
		0.001687,
		0.000719,
		0.000237,
		4.8e-005,
		0.000161,
		0.004664,
		0.012361,
		0.020197,
		0.02625,
		0.029569,
		0.02998,
		0.027863,
		0.023941,
		0.01907,
		0.014064,
		0.009563,
		0.005956,
		0.003363,
		0.001688,
		0.000718,
		0.000232,
		4.5e-005,
		0.002816,
		0.011571,
		0.02233,
		0.032163,
		0.039266,
		0.042807,
		0.04274,
		0.039592,
		0.034243,
		0.027714,
		0.020985,
		0.014841,
		0.00978,
		0.005985,
		0.003363,
		0.001679,
		0.000698,
		0.000216,
		4e-005,
		0.000913,
		0.009308,
		0.022412,
		0.036236,
		0.047987,
		0.055989,
		0.05954,
		0.058738,
		0.054265,
		0.047176,
		0.038672,
		0.02991,
		0.021831,
		0.01505,
		0.0098,
		0.005981,
		0.003331,
		0.001619,
		0.000642,
		0.000185,
		3.1e-005,
		0.00605,
		0.020295,
		0.037732,
		0.054533,
		0.068049,
		0.076739,
		0.080042,
		0.078208,
		0.072094,
		0.062938,
		0.052138,
		0.041042,
		0.030755,
		0.022005,
		0.015059,
		0.00978,
		0.005889,
		0.003177,
		0.001466,
		0.00054,
		0.000141,
		2e-005,
		0.002672,
		0.01627,
		0.036295,
		0.057942,
		0.077548,
		0.092608,
		0.101724,
		0.104482,
		0.101297,
		0.09321,
		0.081659,
		0.068256,
		0.054548,
		0.041813,
		0.030877,
		0.022005,
		0.014994,
		0.009563,
		0.005553,
		0.002835,
		0.001213,
		0.000402,
		9e-005,
		0.000352,
		0.011064,
		0.032043,
		0.057659,
		0.083299,
		0.105461,
		0.121789,
		0.131026,
		0.132916,
		0.128047,
		0.117656,
		0.103399,
		0.087112,
		0.070559,
		0.055184,
		0.041882,
		0.030859,
		0.021831,
		0.014543,
		0.00891,
		0.004882,
		0.002303,
		0.000882,
		0.000248,
		4.1e-005,
		0.005759,
		0.025572,
		0.053638,
		0.084551,
		0.113883,
		0.138292,
		0.155564,
		0.164591,
		0.165276,
		0.158393,
		0.145385,
		0.128143,
		0.108744,
		0.089172,
		0.071024,
		0.055213,
		0.041813,
		0.030469,
		0.020985,
		0.013384,
		0.007717,
		0.003894,
		0.001636,
		0.000524,
		0.000106,
		0.001638,
		0.017889,
		0.046363,
		0.08111,
		0.116937,
		0.149594,
		0.175889,
		0.193752,
		0.202223,
		0.201372,
		0.192162,
		0.176263,
		0.155812,
		0.13314,
		0.110452,
		0.089464,
		0.071024,
		0.055015,
		0.041042,
		0.029006,
		0.01907,
		0.011415,
		0.006041,
		0.002699,
		0.000936,
		0.000207,
		1.4e-005,
		0.010267,
		0.036791,
		0.073371,
		0.1143,
		0.154586,
		0.190145,
		0.217932,
		0.236018,
		0.24359,
		0.240889,
		0.229082,
		0.210073,
		0.186264,
		0.160248,
		0.134436,
		0.110603,
		0.089421,
		0.070559,
		0.05363,
		0.038672,
		0.026022,
		0.016015,
		0.008767,
		0.00408,
		0.001481,
		0.000339,
		1.8e-005,
		0.004055,
		0.026203,
		0.062266,
		0.106308,
		0.152846,
		0.197077,
		0.235077,
		0.263942,
		0.281881,
		0.288229,
		0.283404,
		0.268784,
		0.246526,
		0.219306,
		0.189978,
		0.161129,
		0.134495,
		0.110452,
		0.08848,
		0.068256,
		0.049997,
		0.034243,
		0.021513,
		0.012061,
		0.005762,
		0.002141,
		0.000485,
		1.6e-005,
		0.000483,
		0.016015,
		0.049109,
		0.093895,
		0.144681,
		0.196197,
		0.243819,
		0.283774,
		0.313298,
		0.330732,
		0.335565,
		0.328396,
		0.310826,
		0.285274,
		0.254699,
		0.222209,
		0.190497,
		0.161129,
		0.134094,
		0.108744,
		0.08488,
		0.062938,
		0.043688,
		0.027863,
		0.015878,
		0.007705,
		0.00288,
		0.000623,
		0.007571,
		0.0354,
		0.078437,
		0.131066,
		0.187825,
		0.243645,
		0.294076,
		0.33549,
		0.365257,
		0.381856,
		0.38493,
		0.375266,
		0.354699,
		0.325921,
		0.292176,
		0.256786,
		0.222461,
		0.190422,
		0.160248,
		0.131247,
		0.103399,
		0.077394,
		0.054265,
		0.03498,
		0.020142,
		0.009842,
		0.003651,
		0.000727,
		0.001976,
		0.0226,
		0.061533,
		0.113465,
		0.173013,
		0.234937,
		0.294361,
		0.347001,
		0.389377,
		0.418991,
		0.434461,
		0.435589,
		0.423361,
		0.399855,
		0.368043,
		0.33144,
		0.293526,
		0.256879,
		0.222209,
		0.188738,
		0.155812,
		0.123644,
		0.09321,
		0.065832,
		0.04274,
		0.024749,
		0.012091,
		0.004394,
		0.000774,
		0.011958,
		0.044773,
		0.093595,
		0.153342,
		0.218859,
		0.285117,
		0.347429,
		0.401677,
		0.444525,
		0.473615,
		0.487707,
		0.486772,
		0.472002,
		0.445728,
		0.411201,
		0.372177,
		0.3322,
		0.293526,
		0.256238,
		0.219306,
		0.182192,
		0.145385,
		0.110179,
		0.078208,
		0.050988,
		0.029569,
		0.014355,
		0.005051,
		0.000748,
		0.0044,
		0.029542,
		0.073179,
		0.130667,
		0.197141,
		0.267665,
		0.337415,
		0.401894,
		0.45715,
		0.5,
		0.528227,
		0.540744,
		0.537699,
		0.520504,
		0.491749,
		0.454953,
		0.414057,
		0.372533,
		0.33209,
		0.292176,
		0.251619,
		0.210073,
		0.168337,
		0.128047,
		0.091174,
		0.05954,
		0.03446,
		0.016531,
		0.005563,
		0.000644,
		0.000494,
		0.016906,
		0.053735,
		0.106843,
		0.171783,
		0.243899,
		0.318468,
		0.390871,
		0.456801,
		0.512479,
		0.55488,
		0.581942,
		0.592741,
		0.587617,
		0.568211,
		0.537373,
		0.498867,
		0.456733,
		0.414182,
		0.372177,
		0.329606,
		0.285274,
		0.239079,
		0.192162,
		0.146521,
		0.104482,
		0.068196,
		0.039266,
		0.018512,
		0.005877,
		0.000473,
		0.007563,
		0.036449,
		0.083512,
		0.144773,
		0.215973,
		0.292662,
		0.370327,
		0.444555,
		0.511233,
		0.566765,
		0.608295,
		0.633929,
		0.642924,
		0.635826,
		0.614514,
		0.582095,
		0.542528,
		0.499835,
		0.456733,
		0.413311,
		0.368043,
		0.319805,
		0.268784,
		0.216483,
		0.165276,
		0.11786,
		0.076739,
		0.043827,
		0.020197,
		0.005956,
		0.000268,
		0.001888,
		0.022135,
		0.061965,
		0.11786,
		0.185998,
		0.262302,
		0.342523,
		0.422357,
		0.497602,
		0.564348,
		0.61919,
		0.659459,
		0.683445,
		0.690604,
		0.681705,
		0.658875,
		0.625459,
		0.585544,
		0.542966,
		0.499696,
		0.454953,
		0.406945,
		0.354699,
		0.298721,
		0.240889,
		0.183964,
		0.131026,
		0.084949,
		0.047987,
		0.021495,
		0.005777,
		8.4e-005,
		0.011259,
		0.043104,
		0.092422,
		0.155812,
		0.229604,
		0.309899,
		0.392648,
		0.473757,
		0.549241,
		0.615402,
		0.669042,
		0.707694,
		0.729854,
		0.735198,
		0.724736,
		0.700845,
		0.667076,
		0.627545,
		0.585694,
		0.542528,
		0.496513,
		0.445728,
		0.389406,
		0.328396,
		0.26495,
		0.202223,
		0.143689,
		0.092608,
		0.0516,
		0.02233,
		0.005339,
		0.004022,
		0.027471,
		0.069425,
		0.126853,
		0.19647,
		0.274752,
		0.357979,
		0.442298,
		0.523831,
		0.598814,
		0.663773,
		0.715733,
		0.75245,
		0.77265,
		0.77625,
		0.764515,
		0.74007,
		0.706621,
		0.668179,
		0.627545,
		0.584648,
		0.537373,
		0.483784,
		0.423361,
		0.3573,
		0.288229,
		0.219692,
		0.155564,
		0.099509,
		0.054533,
		0.022647,
		0.004664,
		0.000442,
		0.015327,
		0.049465,
		0.100129,
		0.164372,
		0.239021,
		0.320699,
		0.405858,
		0.490845,
		0.572,
		0.645788,
		0.708972,
		0.758816,
		0.793321,
		0.811468,
		0.813438,
		0.800764,
		0.776299,
		0.743835,
		0.707107,
		0.668022,
		0.625459,
		0.57691,
		0.520504,
		0.455999,
		0.38493,
		0.310294,
		0.236018,
		0.166381,
		0.105461,
		0.056675,
		0.022412,
		0.003798,
		0.00674,
		0.032845,
		0.076267,
		0.134336,
		0.204182,
		0.282748,
		0.366798,
		0.452958,
		0.537774,
		0.617806,
		0.689759,
		0.750642,
		0.797981,
		0.830044,
		0.846087,
		0.84658,
		0.833327,
		0.809382,
		0.77852,
		0.743996,
		0.706621,
		0.664376,
		0.614514,
		0.555299,
		0.486772,
		0.4108,
		0.330732,
		0.250868,
		0.175889,
		0.110298,
		0.057942,
		0.021621,
		0.002816,
		0.001665,
		0.019668,
		0.055596,
		0.106998,
		0.171255,
		0.245582,
		0.327031,
		0.412507,
		0.498808,
		0.582681,
		0.660911,
		0.730447,
		0.788563,
		0.83306,
		0.862495,
		0.876427,
		0.875627,
		0.86217,
		0.839261,
		0.810529,
		0.77852,
		0.742861,
		0.700845,
		0.649616,
		0.587617,
		0.515172,
		0.434461,
		0.349164,
		0.263942,
		0.18387,
		0.113883,
		0.058278,
		0.020295,
		0.001814,
		0.009922,
		0.038245,
		0.082699,
		0.140868,
		0.21019,
		0.287961,
		0.371341,
		0.457378,
		0.543038,
		0.62527,
		0.701088,
		0.767694,
		0.822636,
		0.864009,
		0.890684,
		0.902538,
		0.900656,
		0.887367,
		0.865957,
		0.839755,
		0.810366,
		0.776299,
		0.734367,
		0.681705,
		0.616965,
		0.540744,
		0.455512,
		0.365257,
		0.274981,
		0.190145,
		0.116117,
		0.057659,
		0.018485,
		0.000913,
		0.003529,
		0.024222,
		0.061582,
		0.113359,
		0.177175,
		0.250527,
		0.330792,
		0.415238,
		0.501051,
		0.585368,
		0.66534,
		0.73821,
		0.801442,
		0.852871,
		0.890903,
		0.914735,
		0.924584,
		0.921848,
		0.909081,
		0.889556,
		0.866118,
		0.839261,
		0.806556,
		0.764515,
		0.710347,
		0.642924,
		0.563102,
		0.473615,
		0.378733,
		0.283774,
		0.194576,
		0.116937,
		0.056098,
		0.01627,
		0.000256,
		0.000387,
		0.013474,
		0.043674,
		0.088879,
		0.146868,
		0.215301,
		0.291735,
		0.373628,
		0.458357,
		0.543246,
		0.625599,
		0.702768,
		0.772234,
		0.831725,
		0.879373,
		0.913908,
		0.934871,
		0.942826,
		0.939475,
		0.927549,
		0.910191,
		0.889556,
		0.864981,
		0.833327,
		0.790953,
		0.735198,
		0.665163,
		0.581942,
		0.488506,
		0.389377,
		0.29017,
		0.197077,
		0.116324,
		0.053638,
		0.013754,
		0.00592,
		0.028941,
		0.067471,
		0.11942,
		0.182595,
		0.254699,
		0.333351,
		0.416091,
		0.500403,
		0.583747,
		0.663597,
		0.737504,
		0.803188,
		0.858647,
		0.90232,
		0.93327,
		0.951392,
		0.957597,
		0.953873,
		0.943055,
		0.928019,
		0.910034,
		0.887367,
		0.856393,
		0.813438,
		0.756013,
		0.683445,
		0.597047,
		0.5,
		0.397048,
		0.294076,
		0.197612,
		0.1143,
		0.050357,
		0.011064,
		0.001462,
		0.017323,
		0.049119,
		0.094888,
		0.152562,
		0.21998,
		0.294891,
		0.374977,
		0.457861,
		0.541136,
		0.622395,
		0.699281,
		0.769549,
		0.831152,
		0.882368,
		0.921945,
		0.949288,
		0.964651,
		0.969283,
		0.96542,
		0.955914,
		0.943206,
		0.927549,
		0.906329,
		0.875627,
		0.83183,
		0.77265,
		0.697635,
		0.608295,
		0.507996,
		0.401677,
		0.295462,
		0.196197,
		0.110928,
		0.046363,
		0.008346,
		0.008741,
		0.033782,
		0.073275,
		0.125272,
		0.187731,
		0.258526,
		0.335457,
		0.416269,
		0.498671,
		0.580364,
		0.659076,
		0.732617,
		0.798942,
		0.856247,
		0.903084,
		0.938519,
		0.962301,
		0.975038,
		0.978299,
		0.974514,
		0.966441,
		0.955914,
		0.942136,
		0.921848,
		0.890993,
		0.846087,
		0.785067,
		0.707694,
		0.615657,
		0.512479,
		0.403267,
		0.294361,
		0.192897,
		0.106308,
		0.04179,
		0.005759,
		0.003111,
		0.021402,
		0.05456,
		0.100747,
		0.158037,
		0.224414,
		0.297794,
		0.376029,
		0.456932,
		0.538297,
		0.617935,
		0.69371,
		0.763598,
		0.825756,
		0.878618,
		0.921018,
		0.952336,
		0.972665,
		0.982956,
		0.985067,
		0.98155,
		0.974941,
		0.966297,
		0.953873,
		0.933976,
		0.902538,
		0.856264,
		0.793321,
		0.713679,
		0.61919,
		0.513512,
		0.401894,
		0.290865,
		0.187825,
		0.100576,
		0.036791,
		0.003467,
		0.000342,
		0.011912,
		0.038704,
		0.078987,
		0.130939,
		0.192655,
		0.262156,
		0.337397,
		0.416289,
		0.496713,
		0.576554,
		0.653732,
		0.726248,
		0.792245,
		0.85008,
		0.898429,
		0.936403,
		0.963696,
		0.980739,
		0.988808,
		0.99,
		0.986903,
		0.981686,
		0.974514,
		0.962873,
		0.942826,
		0.910389,
		0.862495,
		0.797554,
		0.715733,
		0.619035,
		0.511233,
		0.397697,
		0.285117,
		0.181135,
		0.093895,
		0.031536,
		0.001638,
		0.005236,
		0.025659,
		0.059973,
		0.106458,
		0.163312,
		0.228658,
		0.300551,
		0.376995,
		0.455958,
		0.535398,
		0.613289,
		0.687664,
		0.756663,
		0.818594,
		0.872018,
		0.915849,
		0.949475,
		0.9729,
		0.986877,
		0.992978,
		0.993485,
		0.990912,
		0.986903,
		0.980718,
		0.969283,
		0.948562,
		0.914735,
		0.864989,
		0.797981,
		0.714073,
		0.615402,
		0.505845,
		0.390871,
		0.277308,
		0.173013,
		0.086449,
		0.026203,
		0.000432,
		0.001294,
		0.015366,
		0.043674,
		0.084596,
		0.136427,
		0.197385,
		0.265625,
		0.339242,
		0.416289,
		0.494799,
		0.572806,
		0.648379,
		0.719666,
		0.784947,
		0.842698,
		0.891682,
		0.931051,
		0.960466,
		0.980234,
		0.991411,
		0.995825,
		0.995868,
		0.993865,
		0.990785,
		0.985067,
		0.973275,
		0.951392,
		0.915816,
		0.864009,
		0.794878,
		0.708972,
		0.608558,
		0.497602,
		0.381656,
		0.267665,
		0.163672,
		0.078437,
		0.020971,
		0.007758,
		0.030049,
		0.065341,
		0.112019,
		0.168399,
		0.232725,
		0.303183,
		0.377913,
		0.455022,
		0.532609,
		0.608789,
		0.681729,
		0.749695,
		0.811109,
		0.864619,
		0.909189,
		0.944206,
		0.969601,
		0.98597,
		0.994648,
		0.997669,
		0.99745,
		0.995988,
		0.993485,
		0.987715,
		0.975038,
		0.95155,
		0.913908,
		0.859861,
		0.788563,
		0.700751,
		0.598814,
		0.486799,
		0.370327,
		0.256439,
		0.153342,
		0.070064,
		0.016015,
		0.002763,
		0.019046,
		0.048665,
		0.090096,
		0.14174,
		0.201932,
		0.268945,
		0.341002,
		0.416289,
		0.492972,
		0.569217,
		0.643226,
		0.713267,
		0.777728,
		0.835175,
		0.884429,
		0.924657,
		0.955481,
		0.977092,
		0.990359,
		0.996864,
		0.998789,
		0.998473,
		0.99745,
		0.995131,
		0.988808,
		0.974769,
		0.949288,
		0.909308,
		0.852871,
		0.779382,
		0.689759,
		0.58651,
		0.473757,
		0.35718,
		0.243899,
		0.142261,
		0.061533,
		0.011499,
		0.000304,
		0.010606,
		0.034535,
		0.070648,
		0.117431,
		0.173305,
		0.236626,
		0.3057,
		0.37879,
		0.45413,
		0.529947,
		0.604481,
		0.676022,
		0.742944,
		0.803759,
		0.857183,
		0.90221,
		0.938211,
		0.965039,
		0.983139,
		0.993631,
		0.998305,
		0.999416,
		0.999123,
		0.998361,
		0.995825,
		0.988486,
		0.972665,
		0.944861,
		0.90232,
		0.84338,
		0.767694,
		0.676358,
		0.572,
		0.458811,
		0.342523,
		0.23032,
		0.130667,
		0.053042,
		0.007571,
		0.004665,
		0.022905,
		0.053654,
		0.095481,
		0.146884,
		0.206304,
		0.272124,
		0.342684,
		0.416289,
		0.491227,
		0.565786,
		0.638287,
		0.707107,
		0.770731,
		0.827802,
		0.877185,
		0.918052,
		0.949977,
		0.973042,
		0.987932,
		0.995994,
		0.999178,
		0.999737,
		0.999522,
		0.998789,
		0.995644,
		0.986877,
		0.968915,
		0.938519,
		0.893244,
		0.831725,
		0.753855,
		0.660911,
		0.555639,
		0.442298,
		0.326664,
		0.215973,
		0.118789,
		0.044773,
		0.004361,
		0.001154,
		0.013723,
		0.039085,
		0.075884,
		0.122695,
		0.178037,
		0.24037,
		0.308108,
		0.379627,
		0.453277,
		0.527402,
		0.600358,
		0.670542,
		0.73643,
		0.796619,
		0.849883,
		0.895243,
		0.932051,
		0.960083,
		0.979644,
		0.991646,
		0.997633,
		0.999659,
		0.999885,
		0.999737,
		0.998768,
		0.994648,
		0.984094,
		0.963696,
		0.930499,
		0.882368,
		0.818227,
		0.73821,
		0.643772,
		0.537774,
		0.424544,
		0.309899,
		0.201119,
		0.106843,
		0.036891,
		0.001976,
		0.006932,
		0.026902,
		0.058623,
		0.100747,
		0.151865,
		0.210511,
		0.275172,
		0.344293,
		0.416289,
		0.489559,
		0.562504,
		0.63355,
		0.701178,
		0.763962,
		0.820615,
		0.870045,
		0.911434,
		0.944313,
		0.968661,
		0.984999,
		0.994447,
		0.998711,
		0.999888,
		0.999952,
		0.999787,
		0.998305,
		0.992877,
		0.980234,
		0.957169,
		0.921018,
		0.869955,
		0.803188,
		0.721083,
		0.62527,
		0.51873,
		0.405858,
		0.29251,
		0.185998,
		0.095026,
		0.029542,
		0.000502,
		0.00247,
		0.017058,
		0.043674,
		0.081039,
		0.127813,
		0.182603,
		0.243966,
		0.310415,
		0.380428,
		0.452462,
		0.524968,
		0.596407,
		0.665277,
		0.730146,
		0.789689,
		0.842738,
		0.888342,
		0.925841,
		0.954949,
		0.975845,
		0.989256,
		0.996489,
		0.999373,
		0.999976,
		0.999981,
		0.999659,
		0.997388,
		0.990359,
		0.97538,
		0.949475,
		0.910271,
		0.856247,
		0.786881,
		0.702768,
		0.605712,
		0.498808,
		0.386524,
		0.274752,
		0.170831,
		0.083512,
		0.022848,
		0.000272,
		0.009504,
		0.031004,
		0.063557,
		0.105894,
		0.156687,
		0.214561,
		0.278096,
		0.345833,
		0.416289,
		0.487964,
		0.559361,
		0.629005,
		0.69547,
		0.757413,
		0.813614,
		0.863025,
		0.904839,
		0.938558,
		0.964076,
		0.981771,
		0.992561,
		0.997917,
		0.999741,
		0.999997,
		0.999976,
		0.999317,
		0.995994,
		0.98711,
		0.969601,
		0.940734,
		0.898429,
		0.841453,
		0.769549,
		0.68353,
		0.585368,
		0.478279,
		0.366798,
		0.256856,
		0.155812,
		0.07245,
		0.016906,
		0.004182,
		0.02057,
		0.048281,
		0.086107,
		0.132788,
		0.187012,
		0.247423,
		0.312626,
		0.381194,
		0.451682,
		0.522638,
		0.592618,
		0.660215,
		0.724081,
		0.782964,
		0.835752,
		0.881522,
		0.91961,
		0.949683,
		0.971816,
		0.986574,
		0.995056,
		0.998863,
		0.999917,
		1,
		0.999911,
		0.998711,
		0.994091,
		0.983139,
		0.962953,
		0.931051,
		0.885637,
		0.825756,
		0.751401,
		0.663597,
		0.564478,
		0.457378,
		0.346903,
		0.239021,
		0.141106,
		0.061965,
		0.011792,
		0.001035,
		0.012329,
		0.035182,
		0.068445,
		0.110919,
		0.161357,
		0.218463,
		0.280903,
		0.34731,
		0.416289,
		0.486437,
		0.556348,
		0.62464,
		0.689972,
		0.751077,
		0.806797,
		0.856131,
		0.898284,
		0.932739,
		0.959323,
		0.978292,
		0.99039,
		0.996876,
		0.999446,
		0.999984,
		0.999999,
		0.999741,
		0.997788,
		0.991646,
		0.978449,
		0.955481,
		0.920512,
		0.872018,
		0.80931,
		0.732617,
		0.643164,
		0.543246,
		0.436307,
		0.327031,
		0.221416,
		0.126853,
		0.052157,
		0.007563,
		0.006231,
		0.024223,
		0.052888,
		0.091084,
		0.137622,
		0.191269,
		0.250747,
		0.314747,
		0.381929,
		0.450936,
		0.520404,
		0.588982,
		0.655346,
		0.718226,
		0.776439,
		0.828924,
		0.874792,
		0.913379,
		0.944313,
		0.967586,
		0.983628,
		0.99335,
		0.998148,
		0.999771,
		0.999999,
		0.999984,
		0.999412,
		0.996489,
		0.988624,
		0.973042,
		0.947222,
		0.909189,
		0.857677,
		0.792245,
		0.713346,
		0.622395,
		0.52184,
		0.415238,
		0.30734,
		0.204182,
		0.113164,
		0.043104,
		0.004255,
		0.002221,
		0.015366,
		0.039412,
		0.073275,
		0.115823,
		0.165879,
		0.222223,
		0.283601,
		0.348725,
		0.416289,
		0.484973,
		0.553458,
		0.620445,
		0.684672,
		0.744944,
		0.800163,
		0.849368,
		0.891784,
		0.926878,
		0.954428,
		0.974587,
		0.987949,
		0.995583,
		0.998989,
		0.999927,
		1,
		0.999922,
		0.998863,
		0.994763,
		0.984999,
		0.966918,
		0.938211,
		0.897143,
		0.842698,
		0.774666,
		0.69371,
		0.601425,
		0.500403,
		0.394307,
		0.287961,
		0.187433,
		0.100129,
		0.034862,
		0.001888,
		0.000244,
		0.008564,
		0.027987,
		0.057479,
		0.095965,
		0.14232,
		0.195381,
		0.253947,
		0.316784,
		0.382633,
		0.450219,
		0.518262,
		0.58549,
		0.650658,
		0.71257,
		0.770107,
		0.822257,
		0.868161,
		0.907163,
		0.938862,
		0.963183,
		0.980438,
		0.991379,
		0.99721,
		0.999508,
		0.999985,
		0.999999,
		0.999771,
		0.998037,
		0.992561,
		0.980744,
		0.960083,
		0.928478,
		0.884429,
		0.827155,
		0.756663,
		0.673812,
		0.580364,
		0.479047,
		0.373628,
		0.268996,
		0.171255,
		0.087814,
		0.027471,
		0.000471,
		0.003771,
		0.018575,
		0.043674,
		0.078042,
		0.120607,
		0.170259,
		0.225849,
		0.286194,
		0.350084,
		0.416289,
		0.483569,
		0.550683,
		0.61641,
		0.679561,
		0.739008,
		0.793707,
		0.842738,
		0.885348,
		0.920996,
		0.949417,
		0.970681,
		0.985256,
		0.994038,
		0.998347,
		0.999797,
		0.999999,
		0.999985,
		0.999478,
		0.996876,
		0.989843,
		0.975845,
		0.952544,
		0.918052,
		0.871094,
		0.811109,
		0.738309,
		0.653732,
		0.559298,
		0.457861,
		0.353287,
		0.250527,
		0.155717,
		0.076267,
		0.020955,
		0.000934,
		0.011138,
		0.031834,
		0.062044,
		0.100747,
		0.146884,
		0.199356,
		0.257029,
		0.318742,
		0.383309,
		0.449532,
		0.516206,
		0.582134,
		0.646144,
		0.707107,
		0.763962,
		0.815748,
		0.861636,
		0.900978,
		0.933353,
		0.958631,
		0.977027,
		0.989154,
		0.996042,
		0.999098,
		0.999935,
		1,
		0.999931,
		0.998989,
		0.995329,
		0.986574,
		0.970291,
		0.944313,
		0.906963,
		0.857183,
		0.794617,
		0.719666,
		0.633539,
		0.538297,
		0.436917,
		0.333351,
		0.232614,
		0.140868,
		0.065521,
		0.015327,
		0.005631,
		0.021925,
		0.047952,
		0.082738,
		0.125272,
		0.174504,
		0.229348,
		0.28869,
		0.35139,
		0.416289,
		0.482221,
		0.548017,
		0.612528,
		0.674631,
		0.73326,
		0.787425,
		0.836245,
		0.878987,
		0.915109,
		0.944313,
		0.9666,
		0.982327,
		0.992245,
		0.997501,
		0.999561,
		0.999987,
		0.999999,
		0.999797,
		0.998251,
		0.99335,
		0.98273,
		0.964076,
		0.935405,
		0.895243,
		0.842738,
		0.777728,
		0.70079,
		0.613289,
		0.51742,
		0.416269,
		0.313873,
		0.215301,
		0.126741,
		0.055596,
		0.01059,
		0.002008,
		0.013913,
		0.035743,
		0.066571,
		0.105431,
		0.151319,
		0.203198,
		0.259999,
		0.320624,
		0.383959,
		0.448872,
		0.51423,
		0.578906,
		0.641793,
		0.701826,
		0.758,
		0.809397,
		0.855222,
		0.894835,
		0.927805,
		0.953955,
		0.973418,
		0.986687,
		0.99464,
		0.99852,
		0.999819,
		0.999999,
		0.999987,
		0.999535,
		0.99721,
		0.990902,
		0.978292,
		0.957205,
		0.925841,
		0.882926,
		0.827802,
		0.760489,
		0.681729,
		0.59303,
		0.496713,
		0.395962,
		0.294891,
		0.198621,
		0.113359,
		0.046503,
		0.00674,
		0.000221,
		0.007758,
		0.025389,
		0.05223,
		0.08736,
		0.12982,
		0.178617,
		0.232725,
		0.291093,
		0.352645,
		0.416289,
		0.480925,
		0.545454,
		0.608789,
		0.669872,
		0.727694,
		0.781314,
		0.82989,
		0.872709,
		0.909232,
		0.939136,
		0.962366,
		0.979184,
		0.990212,
		0.996443,
		0.999193,
		0.999942,
		1,
		0.999939,
		0.999098,
		0.99582,
		0.987949,
		0.973246,
		0.949683,
		0.915644,
		0.870045,
		0.812415,
		0.742944,
		0.662527,
		0.572806,
		0.476214,
		0.376029,
		0.276435,
		0.182595,
		0.100736,
		0.038245,
		0.003769,
		0.003417,
		0.016857,
		0.039695,
		0.071053,
		0.110014,
		0.155629,
		0.206915,
		0.262862,
		0.322436,
		0.384583,
		0.448238,
		0.512331,
		0.575799,
		0.637598,
		0.69672,
		0.752213,
		0.803203,
		0.848922,
		0.888746,
		0.922235,
		0.949176,
		0.969633,
		0.983995,
		0.993005,
		0.997755,
		0.999607,
		0.999988,
		0.999999,
		0.999819,
		0.998436,
		0.994038,
		0.984468,
		0.967586,
		0.941522,
		0.904839,
		0.856637,
		0.796619,
		0.725136,
		0.643226,
		0.552654,
		0.455958,
		0.3565,
		0.258526,
		0.167239,
		0.088879,
		0.03082,
		0.001665,
		0.000846,
		0.010111,
		0.028941,
		0.056497,
		0.091904,
		0.134253,
		0.182603,
		0.235987,
		0.293408,
		0.353853,
		0.416289,
		0.47968,
		0.542987,
		0.605186,
		0.665277,
		0.722303,
		0.77537,
		0.823672,
		0.86652,
		0.903379,
		0.933906,
		0.958001,
		0.975845,
		0.987949,
		0.995167,
		0.998671,
		0.999838,
		0.999999,
		0.999988,
		0.999585,
		0.997501,
		0.991827,
		0.980438,
		0.961312,
		0.932739,
		0.893455,
		0.842738,
		0.780455,
		0.707107,
		0.623865,
		0.532609,
		0.435974,
		0.337397,
		0.241181,
		0.152562,
		0.077791,
		0.024222,
		0.000414,
		0.005114,
		0.019939,
		0.043674,
		0.075484,
		0.114498,
		0.159817,
		0.210511,
		0.265625,
		0.324181,
		0.385184,
		0.447628,
		0.510503,
		0.572806,
		0.63355,
		0.691782,
		0.746597,
		0.797161,
		0.842738,
		0.882719,
		0.916658,
		0.944313,
		0.965692,
		0.981094,
		0.991144,
		0.996795,
		0.999276,
		0.999948,
		1,
		0.999945,
		0.999193,
		0.996249,
		0.989154,
		0.975845,
		0.954428,
		0.92335,
		0.881522,
		0.828385,
		0.763962,
		0.688896,
		0.604481,
		0.512705,
		0.416289,
		0.318742,
		0.224414,
		0.138572,
		0.067471,
		0.018444,
		0.001825,
		0.012657,
		0.032563,
		0.060743,
		0.096367,
		0.138573,
		0.186469,
		0.239139,
		0.295641,
		0.355015,
		0.416289,
		0.478481,
		0.540612,
		0.601712,
		0.660837,
		0.717079,
		0.769588,
		0.817592,
		0.860425,
		0.897559,
		0.928639,
		0.953525,
		0.972331,
		0.985471,
		0.993675,
		0.997978,
		0.999648,
		0.99999,
		0.999999,
		0.999838,
		0.998598,
		0.99464,
		0.985994,
		0.970681,
		0.94694,
		0.913379,
		0.86907,
		0.813614,
		0.74718,
		0.670542,
		0.585111,
		0.492972,
		0.396928,
		0.300551,
		0.208236,
		0.125272,
		0.057915,
		0.013474,
		0.000201,
		0.00706,
		0.023136,
		0.047667,
		0.079857,
		0.118882,
		0.163887,
		0.213992,
		0.268292,
		0.325863,
		0.385763,
		0.447041,
		0.508744,
		0.569922,
		0.629643,
		0.687003,
		0.741144,
		0.79127,
		0.836674,
		0.876762,
		0.911086,
		0.939384,
		0.961616,
		0.978003,
		0.989067,
		0.995632,
		0.998803,
		0.999854,
		0.999999,
		0.99999,
		0.999628,
		0.997755,
		0.992637,
		0.982327,
		0.964943,
		0.938862,
		0.902848,
		0.856131,
		0.798462,
		0.730146,
		0.652082,
		0.565786,
		0.473438,
		0.377913,
		0.282841,
		0.192655,
		0.112664,
		0.049119,
		0.009304,
		0.003111,
		0.015366,
		0.036235,
		0.064959,
		0.100747,
		0.142782,
		0.190218,
		0.242186,
		0.297795,
		0.356136,
		0.416289,
		0.477327,
		0.538323,
		0.598361,
		0.656545,
		0.712016,
		0.763962,
		0.811648,
		0.854428,
		0.891784,
		0.92335,
		0.948956,
		0.968661,
		0.982792,
		0.991969,
		0.997104,
		0.999348,
		0.999954,
		1,
		0.999951,
		0.999276,
		0.996625,
		0.990212,
		0.978139,
		0.958631,
		0.930208,
		0.891784,
		0.842738,
		0.782964,
		0.712898,
		0.63355,
		0.54654,
		0.45413,
		0.359265,
		0.265625,
		0.177679,
		0.100747,
		0.041077,
		0.00592,
		0.000771,
		0.00922,
		0.026425,
		0.05166,
		0.084169,
		0.123166,
		0.167843,
		0.217363,
		0.270869,
		0.327485,
		0.38632,
		0.446476,
		0.507048,
		0.56714,
		0.625869,
		0.682378,
		0.73585,
		0.785527,
		0.830729,
		0.870879,
		0.905532,
		0.934406,
		0.957423,
		0.974739,
		0.986782,
		0.994266,
		0.998174,
		0.999683,
		0.999991,
		0.999999,
		0.999854,
		0.998739,
		0.995167,
		0.987338,
		0.973418,
		0.951752,
		0.920996,
		0.880212,
		0.828924,
		0.767158,
		0.69547,
		0.614981,
		0.527402,
		0.435073,
		0.341002,
		0.248916,
		0.163312,
		0.089522,
		0.033782,
		0.00331,
		0.004665,
		0.018211,
		0.039943,
		0.069138,
		0.105044,
		0.146884,
		0.193856,
		0.245133,
		0.299874,
		0.357216,
		0.416289,
		0.476215,
		0.536117,
		0.595126,
		0.652395,
		0.707107,
		0.75849,
		0.805839,
		0.848532,
		0.886059,
		0.918052,
		0.944313,
		0.964853,
		0.979927,
		0.990058,
		0.996042,
		0.99892,
		0.999869,
		0.999999,
		0.999991,
		0.999666,
		0.997978,
		0.99335,
		0.983995,
		0.968161,
		0.944313,
		0.911245,
		0.868161,
		0.814721,
		0.751077,
		0.677898,
		0.596407,
		0.5084,
		0.416289,
		0.323142,
		0.232725,
		0.149561,
		0.078987,
		0.027226,
		0.001462,
		0.001665,
		0.011564,
		0.029788,
		0.055645,
		0.088416,
		0.127354,
		0.171689,
		0.220628,
		0.273359,
		0.32905,
		0.386858,
		0.44593,
		0.505413,
		0.564456,
		0.622222,
		0.677898,
		0.730708,
		0.779928,
		0.824905,
		0.865077,
		0.900005,
		0.929393,
		0.953131,
		0.971319,
		0.984305,
		0.9927,
		0.997377,
		0.999412,
		0.999958,
		1,
		0.999956,
		0.999348,
		0.996954,
		0.991144,
		0.980168,
		0.962366,
		0.936326,
		0.900978,
		0.855661,
		0.800163,
		0.734755,
		0.660215,
		0.577857,
		0.489559,
		0.397801,
		0.3057,
		0.217061,
		0.136427,
		0.069138,
		0.021402,
		0.000363,
		0.000183,
		0.006452,
		0.02117,
		0.043674,
		0.073275,
		0.109257,
		0.150882,
		0.197385,
		0.247986,
		0.301882,
		0.358259,
		0.416289,
		0.475143,
		0.533987,
		0.592002,
		0.64838,
		0.702346,
		0.753166,
		0.800163,
		0.842738,
		0.880394,
		0.912756,
		0.939609,
		0.960924,
		0.976892,
		0.987949,
		0.99479,
		0.998347,
		0.999714,
		0.999992,
		0.999999,
		0.999869,
		0.998863,
		0.995632,
		0.988525,
		0.975845,
		0.956035,
		0.927805,
		0.890217,
		0.842738,
		0.785281,
		0.718226,
		0.642452,
		0.559361,
		0.470906,
		0.379627,
		0.28869,
		0.201932,
		0.123912,
		0.059973,
		0.0163,
		0.002844,
		0.014064,
		0.033208,
		0.059613,
		0.092596,
		0.131445,
		0.175429,
		0.223793,
		0.275766,
		0.330561,
		0.387376,
		0.445405,
		0.503836,
		0.561865,
		0.618696,
		0.673559,
		0.725714,
		0.77447,
		0.8192,
		0.859359,
		0.894513,
		0.924359,
		0.948756,
		0.967759,
		0.981647,
		0.990939,
		0.996405,
		0.999022,
		0.999882,
		0.999999,
		0.999992,
		0.999699,
		0.998174,
		0.99398,
		0.985471,
		0.971021,
		0.949176,
		0.918768,
		0.878987,
		0.829424,
		0.770107,
		0.701521,
		0.62464,
		0.540946,
		0.452462,
		0.361785,
		0.272124,
		0.187344,
		0.112019,
		0.051487,
		0.011912,
		0.000705,
		0.008442,
		0.024223,
		0.047417,
		0.077365,
		0.113386,
		0.154777,
		0.200812,
		0.250747,
		0.303823,
		0.359265,
		0.416289,
		0.474108,
		0.531932,
		0.588982,
		0.644493,
		0.697728,
		0.747985,
		0.794618,
		0.837049,
		0.874792,
		0.907473,
		0.93486,
		0.956891,
		0.973702,
		0.985655,
		0.99335,
		0.997618,
		0.999468,
		0.999962,
		1,
		0.99996,
		0.999412,
		0.997245,
		0.991969,
		0.981968,
		0.965692,
		0.941795,
		0.909232,
		0.867313,
		0.815748,
		0.754672,
		0.684672,
		0.606807,
		0.522638,
		0.43425,
		0.344293,
		0.256014,
		0.173305,
		0.100747,
		0.043674,
		0.008227,
		0.004273,
		0.016698,
		0.03667,
		0.063557,
		0.096707,
		0.135442,
		0.179066,
		0.226862,
		0.278096,
		0.332021,
		0.387877,
		0.444897,
		0.502313,
		0.559361,
		0.615286,
		0.669354,
		0.720862,
		0.769149,
		0.813614,
		0.853729,
		0.889065,
		0.919314,
		0.944313,
		0.964076,
		0.978822,
		0.988991,
		0.995256,
		0.998499,
		0.999741,
		0.999993,
		1,
		0.999882,
		0.998972,
		0.996042,
		0.989577,
		0.978003,
		0.959858,
		0.933906,
		0.899218,
		0.855222,
		0.801739,
		0.739008,
		0.667708,
		0.588982,
		0.50446,
		0.416289,
		0.327165,
		0.24037,
		0.159817,
		0.090096,
		0.036531,
		0.005236,
		0.001526,
		0.010606,
		0.027353,
		0.051161,
		0.081404,
		0.117431,
		0.158573,
		0.204139,
		0.253422,
		0.3057,
		0.360237,
		0.416289,
		0.473108,
		0.529947,
		0.586062,
		0.64073,
		0.693246,
		0.742944,
		0.789202,
		0.831466,
		0.869259,
		0.902212,
		0.930079,
		0.952769,
		0.970373,
		0.983187,
		0.991726,
		0.996728,
		0.999113,
		0.999893,
		1,
		0.999993,
		0.999728,
		0.998347,
		0.994536,
		0.986782,
		0.973569,
		0.953525,
		0.925522,
		0.888746,
		0.842738,
		0.787425,
		0.723143,
		0.650658,
		0.571191,
		0.486437,
		0.398598,
		0.310415,
		0.2252,
		0.146884,
		0.080064,
		0.030049,
		0.002929,
		0.000168,
		0.00592,
		0.019444,
		0.040162,
		0.067471,
		0.100747,
		0.139346,
		0.182603,
		0.229837,
		0.280351,
		0.333432,
		0.38836,
		0.444407,
		0.500843,
		0.556941,
		0.611985,
		0.665277,
		0.716147,
		0.763962,
		0.808146,
		0.848188,
		0.883666,
		0.914268,
		0.939815,
		0.960286,
		0.975845,
		0.986864,
		0.99393,
		0.997833,
		0.999517,
		0.999966,
		1,
		0.999964,
		0.999468,
		0.997501,
		0.9927,
		0.98357,
		0.968661,
		0.946697,
		0.916658,
		0.877839,
		0.82989,
		0.772836,
		0.707107,
		0.63355,
		0.553458,
		0.468589,
		0.381194,
		0.294054,
		0.210511,
		0.13451,
		0.070648,
		0.024223,
		0.001294,
		0.00261,
		0.012921,
		0.030544,
		0.054899,
		0.085389,
		0.121393,
		0.162273,
		0.207371,
		0.256014,
		0.307516,
		0.361176,
		0.416289,
		0.472143,
		0.528027,
		0.583238,
		0.637084,
		0.688896,
		0.738037,
		0.783913,
		0.825987,
		0.863799,
		0.89698,
		0.925275,
		0.948572,
		0.96692,
		0.980556,
		0.989923,
		0.995671,
		0.998635,
		0.999765,
		0.999993,
		1,
		0.999893,
		0.999069,
		0.996405,
		0.990511,
		0.979927,
		0.963278,
		0.939384,
		0.907331,
		0.86652,
		0.816702,
		0.758,
		0.690927,
		0.61641,
		0.535808,
		0.450936,
		0.364093,
		0.278096,
		0.19631,
		0.122695,
		0.061846,
		0.019046,
		0.000322,
		0.000647,
		0.007758,
		0.022285,
		0.043674,
		0.07135,
		0.104717,
		0.14316,
		0.186045,
		0.232725,
		0.282534,
		0.334797,
		0.388827,
		0.443933,
		0.499421,
		0.5546,
		0.608789,
		0.661323,
		0.711563,
		0.758906,
		0.802795,
		0.842738,
		0.878322,
		0.909232,
		0.935275,
		0.956401,
		0.97273,
		0.984569,
		0.992429,
		0.997015,
		0.999193,
		0.999903,
		1,
		0.999993,
		0.999754,
		0.998499,
		0.99503,
		0.987949,
		0.975845,
		0.957423,
		0.931598,
		0.897559,
		0.854812,
		0.803203,
		0.742944,
		0.674631,
		0.599263,
		0.518262,
		0.433497,
		0.34731,
		0.262549,
		0.182603,
		0.111441,
		0.053654,
		0.014509,
		0.003928,
		0.015366,
		0.033782,
		0.058623,
		0.089317,
		0.125272,
		0.165879,
		0.210511,
		0.258527,
		0.309274,
		0.362085,
		0.416289,
		0.47121,
		0.526172,
		0.580504,
		0.63355,
		0.684672,
		0.73326,
		0.778747,
		0.820615,
		0.858415,
		0.891784,
		0.920461,
		0.944313,
		0.963355,
		0.977776,
		0.987949,
		0.994448,
		0.998024,
		0.999561,
		0.999969,
		1,
		0.999968,
		0.999517,
		0.997729,
		0.99335,
		0.984999,
		0.971319,
		0.951098,
		0.92335,
		0.887363,
		0.842738,
		0.789417,
		0.727694,
		0.658245,
		0.582134,
		0.500843,
		0.416289,
		0.330857,
		0.247423,
		0.169394,
		0.100747,
		0.046069,
		0.010606,
		0.001403,
		0.009762,
		0.025204,
		0.047197,
		0.07519,
		0.108615,
		0.146884,
		0.189395,
		0.235528,
		0.28465,
		0.336118,
		0.389279,
		0.443475,
		0.498046,
		0.552335,
		0.605693,
		0.657487,
		0.707107,
		0.753975,
		0.797559,
		0.837381,
		0.873037,
		0.904213,
		0.930704,
		0.952436,
		0.969489,
		0.982117,
		0.990759,
		0.996042,
		0.998756,
		0.999787,
		0.999994,
		1,
		0.999903,
		0.999154,
		0.996728,
		0.991344,
		0.981647,
		0.966345,
		0.944313,
		0.914656,
		0.876762,
		0.830324,
		0.77537,
		0.712278,
		0.641793,
		0.565045,
		0.483569,
		0.39933,
		0.314747,
		0.232725,
		0.156687,
		0.090614,
		0.039085,
		0.007327,
		0.000155,
		0.00545,
		0.017921,
		0.037057,
		0.062328,
		0.093187,
		0.12907,
		0.169394,
		0.213563,
		0.260965,
		0.310976,
		0.362964,
		0.416289,
		0.470308,
		0.524376,
		0.577857,
		0.630124,
		0.680569,
		0.72861,
		0.773702,
		0.815347,
		0.853109,
		0.886629,
		0.915644,
		0.940003,
		0.959694,
		0.974858,
		0.985814,
		0.99306,
		0.997272,
		0.999265,
		0.999912,
		1,
		0.999994,
		0.999776,
		0.998635,
		0.99547,
		0.988991,
		0.977883,
		0.960924,
		0.937074,
		0.905532,
		0.865776,
		0.817592,
		0.761088,
		0.69672,
		0.6253,
		0.548017,
		0.466459,
		0.382633,
		0.298992,
		0.218463,
		0.144484,
		0.081039,
		0.032698,
		0.004665,
		0.002404,
		0.011912,
		0.028187,
		0.050721,
		0.078987,
		0.112441,
		0.150522,
		0.192655,
		0.238249,
		0.286701,
		0.337397,
		0.389717,
		0.443032,
		0.496716,
		0.550142,
		0.602692,
		0.653763,
		0.702773,
		0.749168,
		0.792437,
		0.832117,
		0.867815,
		0.899218,
		0.926112,
		0.948403,
		0.966137,
		0.979518,
		0.988925,
		0.994912,
		0.998194,
		0.9996,
		0.999972,
		1,
		0.999971,
		0.999561,
		0.997931,
		0.99393,
		0.986276,
		0.973702,
		0.95506,
		0.929393,
		0.895994,
		0.854428,
		0.804566,
		0.746597,
		0.681045,
		0.608789,
		0.531073,
		0.449532,
		0.366214,
		0.283601,
		0.204642,
		0.132788,
		0.072021,
		0.026902,
		0.00261,
		0.000596,
		0.007154,
		0.02057,
		0.040357,
		0.066007,
		0.096998,
		0.132788,
		0.172822,
		0.21653,
		0.263329,
		0.312626,
		0.363815,
		0.416289,
		0.469434,
		0.522638,
		0.575292,
		0.6268,
		0.676582,
		0.724081,
		0.768774,
		0.810183,
		0.847883,
		0.881522,
		0.910833,
		0.935655,
		0.955947,
		0.971816,
		0.983525,
		0.99151,
		0.996374,
		0.998863,
		0.999806,
		0.999994,
		1,
		0.999912,
		0.99923,
		0.997015,
		0.992087,
		0.983187,
		0.9691,
		0.948756,
		0.921283,
		0.886059,
		0.842738,
		0.79127,
		0.731919,
		0.665277,
		0.592281,
		0.51423,
		0.432804,
		0.350084,
		0.268583,
		0.191269,
		0.121599,
		0.063557,
		0.02169,
		0.001154,
		0.003623,
		0.014187,
		0.031222,
		0.05424,
		0.082738,
		0.116196,
		0.154076,
		0.19583,
		0.240892,
		0.28869,
		0.338636,
		0.39014,
		0.442603,
		0.495428,
		0.548017,
		0.599783,
		0.650148,
		0.698557,
		0.74448,
		0.787425,
		0.826947,
		0.862659,
		0.894253,
		0.921508,
		0.944313,
		0.962684,
		0.976785,
		0.986936,
		0.993626,
		0.997501,
		0.999328,
		0.99992,
		1,
		0.999994,
		0.999797,
		0.998756,
		0.995862,
		0.989923,
		0.979712,
		0.964076,
		0.94202,
		0.912756,
		0.875748,
		0.830729,
		0.777728,
		0.717079,
		0.649438,
		0.575799,
		0.497509,
		0.416289,
		0.334256,
		0.253947,
		0.178346,
		0.110919,
		0.055645,
		0.017058,
		0.000287,
		0.001294,
		0.009016,
		0.023299,
		0.043674,
		0.069658,
		0.100747,
		0.136427,
		0.176164,
		0.219416,
		0.265625,
		0.314225,
		0.36464,
		0.416289,
		0.468589,
		0.520954,
		0.572806,
		0.623575,
		0.672708,
		0.71967,
		0.763962,
		0.805122,
		0.842738,
		0.876466,
		0.906036,
		0.931276,
		0.952128,
		0.968661,
		0.981094,
		0.989805,
		0.995329,
		0.998347,
		0.999635,
		0.999974,
		1,
		0.999973,
		0.9996,
		0.998111,
		0.994448,
		0.987421,
		0.975845,
		0.958631,
		0.93486,
		0.903828,
		0.865077,
		0.818422,
		0.763962,
		0.7021,
		0.63355,
		0.559361,
		0.480925,
		0.400003,
		0.318742,
		0.2397,
		0.165879,
		0.100747,
		0.048281,
		0.012998,
		0.000143,
		0.005035,
		0.01657,
		0.034297,
		0.057749,
		0.086442,
		0.119879,
		0.157547,
		0.198921,
		0.243461,
		0.290619,
		0.339837,
		0.39055,
		0.442188,
		0.49418,
		0.545959,
		0.596961,
		0.646637,
		0.694455,
		0.739908,
		0.782523,
		0.821869,
		0.857572,
		0.889324,
		0.9169,
		0.940176,
		0.959144,
		0.973927,
		0.984799,
		0.992187,
		0.996672,
		0.99896,
		0.999823,
		0.999995,
		1,
		0.99992,
		0.999297,
		0.997272,
		0.992753,
		0.984569,
		0.971582,
		0.952769,
		0.927288,
		0.894513,
		0.854068,
		0.805839,
		0.749996,
		0.687003,
		0.617635,
		0.542987,
		0.464497,
		0.383959,
		0.30355,
		0.225849,
		0.153869,
		0.091084,
		0.041461,
		0.009504,
		0.002221,
		0.011016,
		0.026093,
		0.047001,
		0.073275,
		0.104436,
		0.139988,
		0.179424,
		0.222223,
		0.267854,
		0.315776,
		0.365439,
		0.416289,
		0.467769,
		0.519322,
		0.570395,
		0.620445,
		0.66894,
		0.715374,
		0.759263,
		0.800163,
		0.837676,
		0.871464,
		0.901258,
		0.926878,
		0.948246,
		0.965405,
		0.978531,
		0.987949,
		0.994136,
		0.997707,
		0.999385,
		0.999927,
		1,
		0.999995,
		0.999814,
		0.998863,
		0.996213,
		0.990759,
		0.981357,
		0.96692,
		0.946495,
		0.919314,
		0.884829,
		0.842738,
		0.793,
		0.73585,
		0.671812,
		0.601712,
		0.526696,
		0.448238,
		0.368169,
		0.28869,
		0.212399,
		0.14232,
		0.081927,
		0.035182,
		0.006567,
		0.000551,
		0.006618,
		0.019046,
		0.037403,
		0.061241,
		0.090096,
		0.123492,
		0.160938,
		0.201932,
		0.245958,
		0.292492,
		0.341002,
		0.390948,
		0.441786,
		0.492972,
		0.543963,
		0.594222,
		0.643226,
		0.690463,
		0.735449,
		0.777728,
		0.816885,
		0.852555,
		0.884434,
		0.912296,
		0.936004,
		0.955526,
		0.970957,
		0.982523,
		0.990599,
		0.995704,
		0.998483,
		0.999666,
		0.999977,
		1,
		0.999976,
		0.999635,
		0.998273,
		0.994912,
		0.98845,
		0.977776,
		0.961859,
		0.939815,
		0.91095,
		0.874792,
		0.831109,
		0.779928,
		0.721547,
		0.656545,
		0.585802,
		0.510503,
		0.432165,
		0.352645,
		0.27417,
		0.199356,
		0.131232,
		0.073275,
		0.029437,
		0.004182,
		0.003352,
		0.013138,
		0.028941,
		0.05033,
		0.076857,
		0.108062,
		0.143473,
		0.182603,
		0.224955,
		0.27002,
		0.317281,
		0.366214,
		0.416289,
		0.466975,
		0.51774,
		0.568056,
		0.617404,
		0.665277,
		0.711187,
		0.754672,
		0.795304,
		0.832697,
		0.86652,
		0.896506,
		0.922467,
		0.944313,
		0.962058,
		0.975845,
		0.985952,
		0.992798,
		0.99694,
		0.999046,
		0.999838,
		0.999995,
		1,
		0.999927,
		0.999358,
		0.997501,
		0.99335,
		0.985814,
		0.973821,
		0.956401,
		0.932739,
		0.902212,
		0.864419,
		0.8192,
		0.766643,
		0.707107,
		0.641225,
		0.569922,
		0.494427,
		0.416289,
		0.337397,
		0.259999,
		0.186723,
		0.120607,
		0.065127,
		0.024223,
		0.002341,
		0.001198,
		0.008351,
		0.021601,
		0.040531,
		0.064712,
		0.093699,
		0.127035,
		0.164251,
		0.204865,
		0.248386,
		0.294311,
		0.342132,
		0.391333,
		0.441396,
		0.4918,
		0.542026,
		0.591564,
		0.63991,
		0.686577,
		0.731099,
		0.773038,
		0.811993,
		0.84761,
		0.879589,
		0.907703,
		0.931803,
		0.951842,
		0.967884,
		0.980117,
		0.988867,
		0.994596,
		0.997893,
		0.999437,
		0.999933,
		1,
		0.999995,
		0.99983,
		0.99896,
		0.996527,
		0.99151,
		0.982839,
		0.969489,
		0.950551,
		0.925275,
		0.893112,
		0.853729,
		0.807029,
		0.753166,
		0.692551,
		0.625869,
		0.55409,
		0.478481,
		0.400625,
		0.322436,
		0.246181,
		0.174504,
		0.110446,
		0.057479,
		0.019535,
		0.001035,
		0.000132,
		0.004665,
		0.015366,
		0.031834,
		0.053654,
		0.080399,
		0.111627,
		0.146884,
		0.185705,
		0.227614,
		0.272124,
		0.318742,
		0.366965,
		0.416289,
		0.466205,
		0.516206,
		0.565786,
		0.61445,
		0.661713,
		0.707107,
		0.750188,
		0.790544,
		0.827802,
		0.861636,
		0.891784,
		0.918052,
		0.940336,
		0.958631,
		0.973047,
		0.983819,
		0.991318,
		0.996042,
		0.998606,
		0.999694,
		0.999979,
		1,
		0.999978,
		0.999666,
		0.998417,
		0.995329,
		0.989376,
		0.979518,
		0.96478,
		0.944313,
		0.917436,
		0.883666,
		0.842738,
		0.794618,
		0.739516,
		0.677898,
		0.610497,
		0.538323,
		0.462682,
		0.385184,
		0.30777,
		0.232725,
		0.162702,
		0.100747,
		0.05033,
		0.015366,
		0.000257,
		0.002059,
		0.010218,
		0.024223,
		0.043674,
		0.068159,
		0.09725,
		0.130509,
		0.167488,
		0.207724,
		0.250747,
		0.296078,
		0.343228,
		0.391707,
		0.441017,
		0.490663,
		0.540147,
		0.588982,
		0.636686,
		0.682792,
		0.726854,
		0.76845,
		0.807193,
		0.842738,
		0.874792,
		0.903125,
		0.927583,
		0.948101,
		0.96472,
		0.977591,
		0.986998,
		0.99335,
		0.997182,
		0.999123,
		0.999851,
		0.999996,
		1,
		0.999933,
		0.999412,
		0.997707,
		0.993888,
		0.986936,
		0.975845,
		0.959694,
		0.937694,
		0.909232,
		0.87389,
		0.831466,
		0.781985,
		0.725714,
		0.663169,
		0.595126,
		0.522638,
		0.447041,
		0.369978,
		0.293408,
		0.219635,
		0.151319,
		0.091512,
		0.043674,
		0.011711,
		0.000511,
		0.00614,
		0.017685,
		0.034761,
		0.05697,
		0.083901,
		0.11513,
		0.150223,
		0.188732,
		0.230203,
		0.27417,
		0.32016,
		0.367695,
		0.416289,
		0.465459,
		0.514717,
		0.563582,
		0.61158,
		0.658245,
		0.70313,
		0.745808,
		0.785882,
		0.82299,
		0.856815,
		0.887096,
		0.913638,
		0.936326,
		0.955135,
		0.970147,
		0.981561,
		0.9897,
		0.995012,
		0.99806,
		0.999482,
		0.999939,
		1,
		0.999996,
		0.999845,
		0.999046,
		0.99681,
		0.992187,
		0.984178,
		0.971816,
		0.954233,
		0.930704,
		0.900677,
		0.863799,
		0.819929,
		0.769149,
		0.711778,
		0.64838,
		0.579774,
		0.507048,
		0.431573,
		0.355015,
		0.279358,
		0.206915,
		0.140358,
		0.082738,
		0.037511,
		0.008564,
		0.003111,
		0.012202,
		0.026902,
		0.046826,
		0.071577,
		0.100747,
		0.133916,
		0.170651,
		0.210511,
		0.253045,
		0.297795,
		0.344293,
		0.39207,
		0.44065,
		0.489559,
		0.538323,
		0.586474,
		0.63355,
		0.679106,
		0.722712,
		0.763962,
		0.802484,
		0.837941,
		0.870045,
		0.898568,
		0.92335,
		0.944313,
		0.961473,
		0.974954,
		0.984999,
		0.991969,
		0.996348,
		0.998717,
		0.999719,
		0.999981,
		1,
		0.99998,
		0.999694,
		0.998546,
		0.995704,
		0.990212,
		0.981094,
		0.967429,
		0.948403,
		0.92335,
		0.891784,
		0.85341,
		0.808146,
		0.756129,
		0.697728,
		0.63355,
		0.564456,
		0.491569,
		0.416289,
		0.340308,
		0.265625,
		0.19457,
		0.12982,
		0.074425,
		0.031834,
		0.00592,
		0.001112,
		0.007758,
		0.020082,
		0.037715,
		0.060273,
		0.08736,
		0.118572,
		0.15349,
		0.191686,
		0.232725,
		0.27616,
		0.321539,
		0.368403,
		0.416289,
		0.464734,
		0.513271,
		0.561441,
		0.608789,
		0.654869,
		0.699252,
		0.741528,
		0.781314,
		0.81826,
		0.852057,
		0.882447,
		0.909232,
		0.932289,
		0.951577,
		0.967154,
		0.979184,
		0.987949,
		0.993851,
		0.9974,
		0.999193,
		0.999863,
		0.999996,
		1,
		0.999939,
		0.99946,
		0.997893,
		0.994372,
		0.987949,
		0.977679,
		0.962684,
		0.942207,
		0.915644,
		0.882566,
		0.842738,
		0.796135,
		0.742944,
		0.683581,
		0.618696,
		0.549189,
		0.476215,
		0.401201,
		0.325863,
		0.252217,
		0.182603,
		0.119705,
		0.066571,
		0.02664,
		0.003771,
		0.000123,
		0.004334,
		0.014288,
		0.029626,
		0.049979,
		0.074965,
		0.104192,
		0.137255,
		0.173742,
		0.213228,
		0.255282,
		0.299464,
		0.345327,
		0.392422,
		0.440294,
		0.488488,
		0.536552,
		0.584036,
		0.630499,
		0.675515,
		0.718669,
		0.759572,
		0.797864,
		0.833218,
		0.865352,
		0.894038,
		0.919112,
		0.940484,
		0.958153,
		0.972215,
		0.982876,
		0.990456,
		0.995389,
		0.99821,
		0.999524,
		0.999944,
		1,
		0.999996,
		0.999858,
		0.999123,
		0.997064,
		0.992798,
		0.985389,
		0.973927,
		0.957583,
		0.935655,
		0.907596,
		0.873037,
		0.831801,
		0.783913,
		0.72961,
		0.669354,
		0.603835,
		0.533987,
		0.460998,
		0.38632,
		0.311689,
		0.239139,
		0.171018,
		0.110014,
		0.059174,
		0.021925,
		0.002111,
		0.001913,
		0.009504,
		0.022547,
		0.040688,
		0.063557,
		0.090775,
		0.121952,
		0.156687,
		0.19457,
		0.235182,
		0.278096,
		0.322879,
		0.369091,
		0.416289,
		0.46403,
		0.511868,
		0.559361,
		0.606074,
		0.651582,
		0.69547,
		0.737347,
		0.776841,
		0.813614,
		0.847365,
		0.877839,
		0.904839,
		0.928233,
		0.947967,
		0.964076,
		0.976697,
		0.986073,
		0.992561,
		0.996625,
		0.998817,
		0.999741,
		0.999982,
		1,
		0.999981,
		0.999719,
		0.998663,
		0.996042,
		0.990967,
		0.982523,
		0.969836,
		0.952128,
		0.928752,
		0.899218,
		0.863213,
		0.820615,
		0.771497,
		0.716147,
		0.655065,
		0.588982,
		0.518865,
		0.44593,
		0.371655,
		0.297795,
		0.226396,
		0.159817,
		0.100747,
		0.05223,
		0.017685,
		0.000934,
		0.000475,
		0.005711,
		0.016464,
		0.032389,
		0.05313,
		0.07832,
		0.107582,
		0.140529,
		0.176763,
		0.215878,
		0.25746,
		0.301087,
		0.346332,
		0.392764,
		0.439948,
		0.487448,
		0.53483,
		0.581665,
		0.62753,
		0.672015,
		0.714722,
		0.755278,
		0.793332,
		0.828569,
		0.860713,
		0.889537,
		0.914874,
		0.936624,
		0.954769,
		0.969383,
		0.980638,
		0.988816,
		0.994306,
		0.997598,
		0.999256,
		0.999874,
		0.999996,
		1,
		0.999944,
		0.999504,
		0.99806,
		0.994809,
		0.988867,
		0.979343,
		0.965405,
		0.946324,
		0.921508,
		0.890522,
		0.853109,
		0.809195,
		0.758906,
		0.70257,
		0.64073,
		0.574153,
		0.503836,
		0.431024,
		0.357216,
		0.284186,
		0.213992,
		0.149001,
		0.091904,
		0.045739,
		0.013913,
		0.000232,
		0.002895,
		0.011362,
		0.02507,
		0.043674,
		0.066821,
		0.094146,
		0.125272,
		0.159817,
		0.197385,
		0.237576,
		0.27998,
		0.324181,
		0.369759,
		0.416289,
		0.463346,
		0.510503,
		0.557338,
		0.603434,
		0.64838,
		0.691782,
		0.73326,
		0.772459,
		0.809049,
		0.842738,
		0.873276,
		0.900464,
		0.924165,
		0.944313,
		0.960924,
		0.974109,
		0.984077,
		0.991144,
		0.995731,
		0.998347,
		0.999561,
		0.999948,
		1,
		0.999996,
		0.999869,
		0.999193,
		0.997294,
		0.99335,
		0.986488,
		0.975845,
		0.960634,
		0.940176,
		0.913932,
		0.881522,
		0.842738,
		0.797559,
		0.746155,
		0.688896,
		0.626365,
		0.559361,
		0.488913,
		0.416289,
		0.343012,
		0.270869,
		0.201932,
		0.138573,
		0.083483,
		0.039695,
		0.010606,
		0.001035,
		0.007225,
		0.018718,
		0.035182,
		0.056273,
		0.08164,
		0.110919,
		0.143738,
		0.179716,
		0.218463,
		0.259581,
		0.302666,
		0.34731,
		0.393097,
		0.439612,
		0.486437,
		0.533157,
		0.579359,
		0.62464,
		0.668603,
		0.710869,
		0.751077,
		0.788888,
		0.823996,
		0.856131,
		0.88507,
		0.910641,
		0.932739,
		0.95133,
		0.966466,
		0.978292,
		0.987053,
		0.993099,
		0.996876,
		0.998907,
		0.999762,
		0.999984,
		1,
		0.999983,
		0.999741,
		0.998768,
		0.996348,
		0.991651,
		0.983819,
		0.972026,
		0.955526,
		0.933692,
		0.906036,
		0.87223,
		0.832117,
		0.785723,
		0.73326,
		0.675142,
		0.611985,
		0.544621,
		0.474108,
		0.401737,
		0.32905,
		0.25785,
		0.190218,
		0.128533,
		0.075484,
		0.034098,
		0.007758,
		0.000114,
		0.004038,
		0.01332,
		0.027641,
		0.046668,
		0.070061,
		0.09747,
		0.128533,
		0.16288,
		0.200135,
		0.23991,
		0.281814,
		0.325448,
		0.370409,
		0.416289,
		0.462682,
		0.509177,
		0.555372,
		0.600863,
		0.645261,
		0.688183,
		0.729267,
		0.768167,
		0.804566,
		0.838179,
		0.86876,
		0.896111,
		0.92009,
		0.940621,
		0.957705,
		0.971428,
		0.981968,
		0.989606,
		0.994719,
		0.997777,
		0.999313,
		0.999884,
		0.999997,
		1,
		0.999948,
		0.999543,
		0.99821,
		0.995205,
		0.9897,
		0.980856,
		0.967884,
		0.950084,
		0.926878,
		0.897829,
		0.862659,
		0.82126,
		0.773702,
		0.720239,
		0.661323,
		0.597605,
		0.529947,
		0.459432,
		0.387376,
		0.315337,
		0.245133,
		0.178855,
		0.118882,
		0.067904,
		0.028941,
		0.005363,
		0.001783,
		0.008861,
		0.02104,
		0.037997,
		0.059405,
		0.084923,
		0.114202,
		0.146884,
		0.182603,
		0.220985,
		0.261647,
		0.304203,
		0.34826,
		0.39342,
		0.439285,
		0.485454,
		0.531529,
		0.577116,
		0.621825,
		0.665277,
		0.707107,
		0.746966,
		0.784529,
		0.819497,
		0.851607,
		0.880639,
		0.906419,
		0.928835,
		0.947842,
		0.963472,
		0.975845,
		0.985175,
		0.991772,
		0.996042,
		0.99847,
		0.999595,
		0.999952,
		1,
		0.999997,
		0.99988,
		0.999256,
		0.997501,
		0.993851,
		0.987485,
		0.977591,
		0.963419,
		0.944313,
		0.919743,
		0.889324,
		0.852824,
		0.810183,
		0.761512,
		0.707107,
		0.647454,
		0.583238,
		0.51535,
		0.444897,
		0.373216,
		0.301882,
		0.232725,
		0.167843,
		0.10962,
		0.060743,
		0.024223,
		0.003417,
		0.000443,
		0.005327,
		0.015366,
		0.030251,
		0.049664,
		0.073275,
		0.100747,
		0.131734,
		0.165879,
		0.20282,
		0.242186,
		0.283601,
		0.326681,
		0.371041,
		0.416289,
		0.462036,
		0.507888,
		0.553458,
		0.598361,
		0.642221,
		0.684672,
		0.725363,
		0.763962,
		0.800163,
		0.833687,
		0.864293,
		0.891784,
		0.916014,
		0.9369,
		0.954428,
		0.968661,
		0.979753,
		0.987949,
		0.99359,
		0.997104,
		0.998989,
		0.99978,
		0.999985,
		1,
		0.999984,
		0.999762,
		0.998863,
		0.996625,
		0.992272,
		0.984999,
		0.974023,
		0.958631,
		0.938217,
		0.912296,
		0.880531,
		0.842738,
		0.798899,
		0.749168,
		0.693878,
		0.63355,
		0.568899,
		0.500843,
		0.430513,
		0.359265,
		0.28869,
		0.220628,
		0.157186,
		0.100747,
		0.053998,
		0.019939,
		0.001913,
		0.0027,
		0.010606,
		0.023419,
		0.04083,
		0.062522,
		0.088168,
		0.117431,
		0.149968,
		0.185427,
		0.223446,
		0.263662,
		0.3057,
		0.349184,
		0.393734,
		0.438967,
		0.484498,
		0.529947,
		0.574932,
		0.619082,
		0.662033,
		0.703432,
		0.742944,
		0.780253,
		0.815072,
		0.847143,
		0.876247,
		0.902212,
		0.924919,
		0.944313,
		0.960409,
		0.973306,
		0.983187,
		0.990328,
		0.995096,
		0.99794,
		0.999365,
		0.999893,
		0.999997,
		1,
		0.999952,
		0.999578,
		0.998347,
		0.995564,
		0.990456,
		0.982234,
		0.970147,
		0.953525,
		0.931803,
		0.904547,
		0.871464,
		0.832415,
		0.787425,
		0.736686,
		0.680569,
		0.619625,
		0.5546,
		0.486437,
		0.416289,
		0.345531,
		0.275766,
		0.208847,
		0.146884,
		0.092264,
		0.047667,
		0.016085,
		0.000846,
		0.000966,
		0.006746,
		0.017488,
		0.032895,
		0.052658,
		0.076461,
		0.103978,
		0.134877,
		0.168815,
		0.205443,
		0.244406,
		0.285341,
		0.327881,
		0.371655,
		0.416289,
		0.461407,
		0.506634,
		0.551596,
		0.595924,
		0.639258,
		0.681244,
		0.721547,
		0.759844,
		0.795839,
		0.829262,
		0.859876,
		0.887485,
		0.911942,
		0.933156,
		0.951098,
		0.965817,
		0.977439,
		0.986181,
		0.992346,
		0.996325,
		0.998583,
		0.999625,
		0.999956,
		1,
		0.999997,
		0.999889,
		0.999313,
		0.997689,
		0.994306,
		0.988393,
		0.979184,
		0.965963,
		0.948101,
		0.925079,
		0.896506,
		0.862134,
		0.821869,
		0.775775,
		0.724081,
		0.667192,
		0.605693,
		0.540354,
		0.472143,
		0.402235,
		0.332021,
		0.263117,
		0.197385,
		0.13694,
		0.084169,
		0.041747,
		0.012657,
		0.000211,
		0.000107,
		0.003771,
		0.012447,
		0.025848,
		0.043674,
		0.065621,
		0.091374,
		0.120607,
		0.152992,
		0.188187,
		0.225849,
		0.265625,
		0.307157,
		0.350084,
		0.39404,
		0.438658,
		0.483569,
		0.528406,
		0.572806,
		0.61641,
		0.658868,
		0.699842,
		0.739008,
		0.776061,
		0.810721,
		0.842738,
		0.871897,
		0.898023,
		0.920996,
		0.940749,
		0.957286,
		0.970681,
		0.981094,
		0.98877,
		0.994038,
		0.997312,
		0.999064,
		0.999797,
		0.999986,
		1,
		0.999985,
		0.99978,
		0.998949,
		0.996876,
		0.992836,
		0.986073,
		0.975845,
		0.961473,
		0.942367,
		0.918052,
		0.888183,
		0.852555,
		0.811114,
		0.763962,
		0.711367,
		0.653763,
		0.591766,
		0.526172,
		0.457972,
		0.38836,
		0.318742,
		0.250747,
		0.186245,
		0.127354,
		0.076461,
		0.036235,
		0.00965,
		0.001665,
		0.008282,
		0.019678,
		0.035564,
		0.055645,
		0.079615,
		0.107161,
		0.137962,
		0.171689,
		0.208006,
		0.246571,
		0.287037,
		0.32905,
		0.372254,
		0.416289,
		0.460796,
		0.505413,
		0.549783,
		0.59355,
		0.636368,
		0.677898,
		0.717815,
		0.75581,
		0.791594,
		0.824905,
		0.855511,
		0.883219,
		0.907879,
		0.929393,
		0.947725,
		0.962903,
		0.975034,
		0.984305,
		0.990988,
		0.99544,
		0.998088,
		0.999412,
		0.999901,
		0.999997,
		1,
		0.999956,
		0.99961,
		0.99847,
		0.99589,
		0.991144,
		0.98349,
		0.972215,
		0.956676,
		0.936326,
		0.910731,
		0.879589,
		0.842738,
		0.800163,
		0.752002,
		0.698557,
		0.640295,
		0.577857,
		0.512066,
		0.443933,
		0.374671,
		0.3057,
		0.238661,
		0.175429,
		0.118126,
		0.069138,
		0.031129,
		0.00706,
		0.000414,
		0.004979,
		0.014374,
		0.028318,
		0.046525,
		0.0687,
		0.094539,
		0.12373,
		0.155955,
		0.190888,
		0.228195,
		0.26754,
		0.308577,
		0.35096,
		0.394338,
		0.438357,
		0.482664,
		0.526907,
		0.570735,
		0.613805,
		0.655781,
		0.696335,
		0.735156,
		0.771949,
		0.806444,
		0.838394,
		0.86759,
		0.893857,
		0.917071,
		0.937158,
		0.954108,
		0.967979,
		0.978905,
		0.987103,
		0.99287,
		0.996584,
		0.998685,
		0.999653,
		0.999959,
		1,
		0.999997,
		0.999897,
		0.999365,
		0.99786,
		0.994719,
		0.98922,
		0.980638,
		0.968292,
		0.951577,
		0.929985,
		0.903125,
		0.870737,
		0.832697,
		0.78903,
		0.739908,
		0.685666,
		0.6268,
		0.563978,
		0.498046,
		0.430036,
		0.361176,
		0.292901,
		0.226862,
		0.164939,
		0.109257,
		0.0622,
		0.026425,
		0.004882,
		0.002525,
		0.009923,
		0.021925,
		0.038254,
		0.058623,
		0.082738,
		0.110298,
		0.140991,
		0.174504,
		0.210511,
		0.248685,
		0.28869,
		0.330188,
		0.372836,
		0.416289,
		0.460201,
		0.504225,
		0.548017,
		0.591237,
		0.63355,
		0.674631,
		0.714166,
		0.751857,
		0.787425,
		0.820615,
		0.851199,
		0.878987,
		0.903828,
		0.925619,
		0.944313,
		0.959925,
		0.972543,
		0.982327,
		0.989522,
		0.994448,
		0.997501,
		0.999131,
		0.999812,
		0.999987,
		1,
		0.999987,
		0.999797,
		0.999027,
		0.997104,
		0.99335,
		0.987053,
		0.977512,
		0.964076,
		0.946178,
		0.92335,
		0.895243,
		0.861636,
		0.822445,
		0.777728,
		0.727694,
		0.672708,
		0.613292,
		0.550142,
		0.484124,
		0.416289,
		0.347883,
		0.280351,
		0.215353,
		0.154777,
		0.100747,
		0.055645,
		0.022121,
		0.003111,
		0.000903,
		0.006313,
		0.016376,
		0.030823,
		0.049379,
		0.071756,
		0.097664,
		0.126801,
		0.158861,
		0.193529,
		0.230487,
		0.269407,
		0.309961,
		0.351814,
		0.394628,
		0.438064,
		0.481783,
		0.525446,
		0.568718,
		0.611266,
		0.652767,
		0.692907,
		0.731385,
		0.767917,
		0.802239,
		0.834112,
		0.863328,
		0.889716,
		0.913148,
		0.933544,
		0.950882,
		0.965204,
		0.976625,
		0.985332,
		0.991594,
		0.995754,
		0.998223,
		0.999454,
		0.999909,
		0.999997,
		1,
		0.999959,
		0.999639,
		0.998583,
		0.996187,
		0.991772,
		0.984638,
		0.974109,
		0.959568,
		0.940484,
		0.91643,
		0.887096,
		0.8523,
		0.811993,
		0.766271,
		0.715374,
		0.659694,
		0.599783,
		0.536358,
		0.470308,
		0.402701,
		0.334797,
		0.268054,
		0.204139,
		0.144944,
		0.092596,
		0.049471,
		0.018211,
		0.001742,
		0.0001,
		0.003529,
		0.011657,
		0.024223,
		0.040959,
		0.061589,
		0.085828,
		0.113386,
		0.143965,
		0.17726,
		0.212959,
		0.250747,
		0.290302,
		0.331297,
		0.373403,
		0.416289,
		0.459622,
		0.503068,
		0.546297,
		0.588982,
		0.630801,
		0.67144,
		0.710598,
		0.747985,
		0.783333,
		0.816392,
		0.846941,
		0.874792,
		0.899793,
		0.921837,
		0.940869,
		0.956891,
		0.969973,
		0.980254,
		0.987949,
		0.99335,
		0.99682,
		0.998778,
		0.999678,
		0.999962,
		1,
		0.999997,
		0.999905,
		0.999412,
		0.998016,
		0.995096,
		0.989975,
		0.981968,
		0.970427,
		0.954769,
		0.934501,
		0.909232,
		0.878693,
		0.842738,
		0.801356,
		0.754672,
		0.702959,
		0.646637,
		0.586284,
		0.522638,
		0.456608,
		0.389279,
		0.321925,
		0.256014,
		0.193222,
		0.135442,
		0.084802,
		0.043674,
		0.014693,
		0.000771,
		0.001559,
		0.007758,
		0.018444,
		0.033357,
		0.05223,
		0.074788,
		0.100747,
		0.12982,
		0.16171,
		0.196114,
		0.232725,
		0.27123,
		0.31131,
		0.352645,
		0.39491,
		0.437779,
		0.480925,
		0.524024,
		0.566751,
		0.608789,
		0.649825,
		0.689557,
		0.727694,
		0.763962,
		0.798105,
		0.82989,
		0.859113,
		0.885604,
		0.909232,
		0.929913,
		0.947615,
		0.962366,
		0.97426,
		0.983462,
		0.990212,
		0.994823,
		0.997675,
		0.999193,
		0.999825,
		0.999988,
		1,
		0.999988,
		0.999812,
		0.999098,
		0.997312,
		0.993819,
		0.987949,
		0.979038,
		0.966466,
		0.949683,
		0.928233,
		0.901765,
		0.870045,
		0.832964,
		0.790544,
		0.742944,
		0.690463,
		0.63355,
		0.572806,
		0.508991,
		0.443032,
		0.376031,
		0.309274,
		0.244237,
		0.182603,
		0.126271,
		0.077365,
		0.038254,
		0.011564,
		0.000192,
		0.000387,
		0.004665,
		0.013475,
		0.026564,
		0.043674,
		0.064539,
		0.088884,
		0.116428,
		0.146884,
		0.179959,
		0.215353,
		0.252761,
		0.291874,
		0.332378,
		0.373956,
		0.416289,
		0.459058,
		0.501941,
		0.544621,
		0.586783,
		0.628118,
		0.668323,
		0.707107,
		0.744191,
		0.779315,
		0.812236,
		0.842738,
		0.870636,
		0.895777,
		0.918052,
		0.937398,
		0.953807,
		0.967332,
		0.978092,
		0.986276,
		0.992149,
		0.996042,
		0.998347,
		0.999493,
		0.999915,
		0.999998,
		1,
		0.999962,
		0.999666,
		0.998685,
		0.996457,
		0.992346,
		0.985689,
		0.975845,
		0.962225,
		0.944313,
		0.921688,
		0.894038,
		0.861163,
		0.82299,
		0.779572,
		0.731099,
		0.677898,
		0.620445,
		0.559361,
		0.495428,
		0.42959,
		0.362964,
		0.296847,
		0.232725,
		0.172287,
		0.117431,
		0.070284,
		0.033208,
		0.008818,
		0.002366,
		0.009304,
		0.02057,
		0.035914,
		0.055077,
		0.077793,
		0.103789,
		0.132788,
		0.164503,
		0.198643,
		0.234912,
		0.273008,
		0.312626,
		0.353455,
		0.395185,
		0.437501,
		0.48009,
		0.522638,
		0.564834,
		0.606372,
		0.646952,
		0.686282,
		0.724081,
		0.760084,
		0.794042,
		0.825729,
		0.854945,
		0.881522,
		0.905327,
		0.92627,
		0.944313,
		0.959469,
		0.971816,
		0.981498,
		0.988728,
		0.993791,
		0.997036,
		0.998863,
		0.999701,
		0.999965,
		1,
		0.999998,
		0.999912,
		0.999454,
		0.998157,
		0.99544,
		0.990665,
		0.983187,
		0.972387,
		0.957705,
		0.938663,
		0.914874,
		0.886059,
		0.852057,
		0.812825,
		0.76845,
		0.719149,
		0.665277,
		0.607332,
		0.545959,
		0.481957,
		0.416289,
		0.350084,
		0.28465,
		0.221482,
		0.162273,
		0.108924,
		0.063557,
		0.028532,
		0.006452,
		0.000846,
		0.00592,
		0.015366,
		0.028941,
		0.046396,
		0.067471,
		0.091904,
		0.119423,
		0.14975,
		0.182603,
		0.217694,
		0.254728,
		0.293408,
		0.333432,
		0.374495,
		0.416289,
		0.458508,
		0.500843,
		0.542987,
		0.584639,
		0.625499,
		0.665277,
		0.703692,
		0.740474,
		0.77537,
		0.808146,
		0.838591,
		0.86652,
		0.891784,
		0.914268,
		0.933906,
		0.950679,
		0.964625,
		0.975845,
		0.984507,
		0.990846,
		0.995167,
		0.997833,
		0.999249,
		0.999838,
		0.999989,
		1,
		0.999988,
		0.999825,
		0.999163,
		0.997501,
		0.994248,
		0.98877,
		0.980438,
		0.968661,
		0.95291,
		0.932739,
		0.907796,
		0.877839,
		0.842738,
		0.802484,
		0.757191,
		0.707107,
		0.65261,
		0.594222,
		0.532609,
		0.468589,
		0.403137,
		0.337397,
		0.272688,
		0.210511,
		0.152563,
		0.100747,
		0.057184,
		0.024223,
		0.004462,
		9.3e-005,
		0.00331,
		0.01094,
		0.022747,
		0.038489,
		0.057915,
		0.08077,
		0.106789,
		0.135705,
		0.167242,
		0.201118,
		0.237049,
		0.274744,
		0.313909,
		0.354245,
		0.395453,
		0.43723,
		0.479275,
		0.521286,
		0.562964,
		0.604014,
		0.644146,
		0.68308,
		0.720543,
		0.75628,
		0.790049,
		0.82163,
		0.850827,
		0.877473,
		0.901435,
		0.92262,
		0.94098,
		0.95652,
		0.969301,
		0.979446,
		0.987147,
		0.99266,
		0.996306,
		0.99846,
		0.999529,
		0.999921,
		0.999998,
		1,
		0.999965,
		0.99969,
		0.998778,
		0.996704,
		0.99287,
		0.986651,
		0.977439,
		0.964669,
		0.947842,
		0.926546,
		0.900464,
		0.869387,
		0.833218,
		0.791976,
		0.745808,
		0.694984,
		0.63991,
		0.581127,
		0.519322,
		0.455329,
		0.39014,
		0.324909,
		0.260965,
		0.199814,
		0.14316,
		0.092903,
		0.051161,
		0.02028,
		0.002844,
		0.001462,
		0.007282,
		0.017323,
		0.031349,
		0.049119,
		0.070384,
		0.094888,
		0.122371,
		0.152563,
		0.185194,
		0.219983,
		0.25665,
		0.294906,
		0.33446,
		0.37502,
		0.416289,
		0.457972,
		0.499772,
		0.541394,
		0.582546,
		0.622942,
		0.6623,
		0.70035,
		0.736831,
		0.771497,
		0.804122,
		0.834498,
		0.862446,
		0.887815,
		0.91049,
		0.930398,
		0.947512,
		0.961859,
		0.973521,
		0.982646,
		0.989445,
		0.994196,
		0.997234,
		0.998941,
		0.999722,
		0.999968,
		1,
		0.999998,
		0.999918,
		0.999493,
		0.998286,
		0.995754,
		0.991297,
		0.984305,
		0.974188,
		0.960409,
		0.942504,
		0.92009,
		0.892886,
		0.860713,
		0.823506,
		0.781314,
		0.734311,
		0.682792,
		0.627186,
		0.568056,
		0.506107,
		0.442188,
		0.377306,
		0.312626,
		0.249484,
		0.189395,
		0.134062,
		0.085389,
		0.045488,
		0.016698,
		0.001593,
		0.000363,
		0.004379,
		0.012657,
		0.024968,
		0.041077,
		0.060743,
		0.083718,
		0.109748,
		0.138573,
		0.169927,
		0.203541,
		0.239139,
		0.27644,
		0.315161,
		0.355015,
		0.395714,
		0.436966,
		0.478481,
		0.519969,
		0.561141,
		0.601712,
		0.641405,
		0.679948,
		0.717079,
		0.752549,
		0.786125,
		0.817592,
		0.846758,
		0.873458,
		0.897559,
		0.918966,
		0.937623,
		0.953525,
		0.966719,
		0.977312,
		0.985471,
		0.991431,
		0.995483,
		0.997978,
		0.999301,
		0.999849,
		0.99999,
		1,
		0.999989,
		0.999838,
		0.999222,
		0.997675,
		0.99464,
		0.989522,
		0.981724,
		0.970681,
		0.955885,
		0.9369,
		0.913379,
		0.88507,
		0.851827,
		0.813614,
		0.770509,
		0.722712,
		0.670542,
		0.61445,
		0.55502,
		0.492972,
		0.429172,
		0.36464,
		0.300551,
		0.238249,
		0.179254,
		0.125272,
		0.078205,
		0.040162,
		0.013475,
		0.000705,
		0.002221,
		0.008741,
		0.019337,
		0.033782,
		0.051842,
		0.073275,
		0.097836,
		0.125272,
		0.155325,
		0.187731,
		0.222223,
		0.258527,
		0.296367,
		0.335463,
		0.375532,
		0.416289,
		0.45745,
		0.498728,
		0.53984,
		0.580504,
		0.620445,
		0.65939,
		0.697079,
		0.73326,
		0.767695,
		0.800163,
		0.830462,
		0.858415,
		0.883873,
		0.90672,
		0.926878,
		0.944313,
		0.959038,
		0.971124,
		0.980698,
		0.987949,
		0.99313,
		0.996548,
		0.998563,
		0.999561,
		0.999927,
		0.999998,
		1,
		0.999968,
		0.999712,
		0.998863,
		0.99693,
		0.99335,
		0.987535,
		0.978905,
		0.96692,
		0.951098,
		0.931037,
		0.906419,
		0.877025,
		0.842738,
		0.803552,
		0.759572,
		0.711022,
		0.658245,
		0.601712,
		0.542026,
		0.479925,
		0.416289,
		0.352149,
		0.28869,
		0.227264,
		0.169394,
		0.11679,
		0.07135,
		0.035182,
		0.010606,
		0.000176,
		0.005562,
		0.014447,
		0.027226,
		0.043674,
		0.063557,
		0.086636,
		0.112664,
		0.141391,
		0.172561,
		0.205913,
		0.241182,
		0.278096,
		0.316383,
		0.355767,
		0.395969,
		0.436709,
		0.477707,
		0.518684,
		0.559361,
		0.599465,
		0.638727,
		0.676885,
		0.713686,
		0.748889,
		0.782268,
		0.813614,
		0.842738,
		0.869479,
		0.893703,
		0.915312,
		0.934245,
		0.950488,
		0.964076,
		0.9751,
		0.983707,
		0.990107,
		0.994569,
		0.997416,
		0.999012,
		0.999741,
		0.99997,
		1,
		0.999998,
		0.999924,
		0.999529,
		0.998404,
		0.996042,
		0.991877,
		0.985332,
		0.975845,
		0.962903,
		0.946052,
		0.924919,
		0.899218,
		0.86876,
		0.833458,
		0.793332,
		0.748515,
		0.699252,
		0.645911,
		0.588982,
		0.529086,
		0.466975,
		0.403546,
		0.339837,
		0.277047,
		0.21653,
		0.159817,
		0.108615,
		0.064824,
		0.030544,
		0.008089,
		0.021402,
		0.036235,
		0.05456,
		0.076143,
		0.100747,
		0.128128,
		0.158037,
		0.190218,
		0.224414,
		0.260362,
		0.297795,
		0.336442,
		0.376031,
		0.416289,
		0.45694,
		0.497709,
		0.538323,
		0.578511,
		0.618005,
		0.656545,
		0.693878,
		0.729761,
		0.763962,
		0.796268,
		0.826481,
		0.854428,
		0.87996,
		0.902962,
		0.92335,
		0.941085,
		0.95617,
		0.968661,
		0.978669,
		0.986362,
		0.991969,
		0.995774,
		0.998111,
		0.999348,
		0.99986,
		0.99999,
		1,
		0.99999,
		0.999849,
		0.999276,
		0.997833,
		0.994999,
		0.990212,
		0.982906,
		0.972543,
		0.958631,
		0.940749,
		0.918552,
		0.891784,
		0.860284,
		0.823996,
		0.782964,
		0.737347,
		0.687413,
		0.63355,
		0.576269,
		0.516206,
		0.45413,
		0.390948,
		0.327712,
		0.265625,
		0.206051,
		0.150522,
		0.100747,
		0.058623,
		0.026246,
		0.00592,
		0.046277,
		0.066356,
		0.089522,
		0.115538,
		0.144161,
		0.175145,
		0.208236,
		0.243179,
		0.279714,
		0.317576,
		0.3565,
		0.396218,
		0.436458,
		0.476952,
		0.51743,
		0.557624,
		0.59727,
		0.636109,
		0.673888,
		0.710363,
		0.745299,
		0.778478,
		0.809696,
		0.83877,
		0.865538,
		0.889869,
		0.911662,
		0.930852,
		0.947416,
		0.961379,
		0.972816,
		0.981857,
		0.988691,
		0.993563,
		0.996771,
		0.998658,
		0.999591,
		0.999932,
		0.999998,
		1,
		0.99997,
		0.999732,
		0.998941,
		0.997137,
		0.993791,
		0.988347,
		0.980254,
		0.968996,
		0.954108,
		0.935195,
		0.911942,
		0.884124,
		0.851607,
		0.814362,
		0.772459,
		0.726079,
		0.675515,
		0.621173,
		0.563582,
		0.503396,
		0.441396,
		0.378502,
		0.315776,
		0.254429,
		0.19583,
		0.141513,
		0.093187,
		0.052748,
		0.022285,
		0.004095,
		0.078987,
		0.103621,
		0.130939,
		0.160699,
		0.192655,
		0.226559,
		0.262156,
		0.299189,
		0.337397,
		0.376519,
		0.416289,
		0.456443,
		0.496716,
		0.536843,
		0.576564,
		0.615622,
		0.653763,
		0.690745,
		0.726331,
		0.760297,
		0.792437,
		0.822556,
		0.850486,
		0.876079,
		0.899218,
		0.919818,
		0.937833,
		0.953258,
		0.966137,
		0.976563,
		0.984688,
		0.990717,
		0.994912,
		0.997583,
		0.999078,
		0.999759,
		0.999972,
		1,
		0.999998,
		0.999929,
		0.999561,
		0.998512,
		0.996306,
		0.99241,
		0.986276,
		0.977373,
		0.965204,
		0.949334,
		0.929393,
		0.905097,
		0.876247,
		0.842738,
		0.804566,
		0.761826,
		0.714722,
		0.663568,
		0.608789,
		0.550931,
		0.490663,
		0.42878,
		0.366214,
		0.304035,
		0.243461,
		0.185867,
		0.132788,
		0.085934,
		0.047197,
		0.01866,
		0.00261,
		0.118371,
		0.146884,
		0.177679,
		0.210511,
		0.245133,
		0.281295,
		0.318742,
		0.357216,
		0.39646,
		0.436213,
		0.476215,
		0.516206,
		0.555928,
		0.595126,
		0.63355,
		0.670956,
		0.707107,
		0.741777,
		0.774754,
		0.805839,
		0.834852,
		0.861636,
		0.886059,
		0.908019,
		0.927446,
		0.944313,
		0.958631,
		0.970465,
		0.979927,
		0.987186,
		0.992466,
		0.996042,
		0.998234,
		0.999391,
		0.999869,
		0.999991,
		1,
		0.999991,
		0.99986,
		0.999325,
		0.997978,
		0.995329,
		0.990846,
		0.983995,
		0.97426,
		0.961169,
		0.944313,
		0.92335,
		0.898023,
		0.868161,
		0.833687,
		0.794618,
		0.751077,
		0.703287,
		0.651582,
		0.596407,
		0.538323,
		0.478015,
		0.416289,
		0.354089,
		0.292492,
		0.232725,
		0.176164,
		0.124349,
		0.078987,
		0.041967,
		0.015366,
		0.001462,
		0.163312,
		0.195044,
		0.228658,
		0.26391,
		0.300551,
		0.33833,
		0.376995,
		0.416289,
		0.455958,
		0.495746,
		0.535398,
		0.574663,
		0.613292,
		0.651042,
		0.687676,
		0.722968,
		0.756699,
		0.788668,
		0.818687,
		0.846589,
		0.87223,
		0.895491,
		0.916286,
		0.934562,
		0.950308,
		0.963556,
		0.974386,
		0.982931,
		0.989376,
		0.993962,
		0.996976,
		0.998745,
		0.999618,
		0.999937,
		0.999998,
		1,
		0.999972,
		0.99975,
		0.999012,
		0.997327,
		0.994196,
		0.989094,
		0.981498,
		0.970913,
		0.956891,
		0.939048,
		0.917071,
		0.890728,
		0.859876,
		0.824461,
		0.784529,
		0.74022,
		0.691782,
		0.639566,
		0.584036,
		0.525768,
		0.465459,
		0.40393,
		0.342132,
		0.281153,
		0.222223,
		0.166724,
		0.116196,
		0.072345,
		0.037057,
		0.012401,
		0.000647,
		0.212739,
		0.247045,
		0.282841,
		0.31988,
		0.357915,
		0.396697,
		0.435974,
		0.475496,
		0.515011,
		0.554272,
		0.593031,
		0.631048,
		0.668086,
		0.703917,
		0.738322,
		0.771094,
		0.80204,
		0.830985,
		0.857774,
		0.882275,
		0.904386,
		0.924033,
		0.941183,
		0.955839,
		0.968053,
		0.977922,
		0.985597,
		0.991281,
		0.995229,
		0.997737,
		0.999138,
		0.999775,
		0.999974,
		1,
		0.999998,
		0.999934,
		0.999591,
		0.998612,
		0.996548,
		0.992899,
		0.987147,
		0.978782,
		0.967332,
		0.952372,
		0.933544,
		0.910561,
		0.883219,
		0.851398,
		0.815072,
		0.774307,
		0.729267,
		0.680218,
		0.62753,
		0.571684,
		0.513271,
		0.453002,
		0.391707,
		0.330348,
		0.27002,
		0.211958,
		0.157547,
		0.108329,
		0.066007,
		0.032464,
		0.009762,
		0.000161,
		0.265625,
		0.301882,
		0.339242,
		0.377459,
		0.416289,
		0.455485,
		0.494799,
		0.533987,
		0.572806,
		0.611015,
		0.64838,
		0.684672,
		0.71967,
		0.753166,
		0.784961,
		0.814873,
		0.842738,
		0.868415,
		0.891784,
		0.912756,
		0.931276,
		0.947325,
		0.960924,
		0.972142,
		0.981094,
		0.987949,
		0.992925,
		0.996289,
		0.998347,
		0.999431,
		0.999878,
		0.999992,
		1,
		0.999991,
		0.999869,
		0.99937,
		0.998111,
		0.995632,
		0.991431,
		0.984999,
		0.975845,
		0.963517,
		0.947615,
		0.927805,
		0.903828,
		0.875503,
		0.842738,
		0.805528,
		0.763962,
		0.718226,
		0.668603,
		0.615483,
		0.559361,
		0.500843,
		0.44065,
		0.379627,
		0.318742,
		0.259096,
		0.201932,
		0.148635,
		0.100747,
		0.059973,
		0.028187,
		0.007447,
		0.320992,
		0.358598,
		0.396928,
		0.435741,
		0.474793,
		0.513845,
		0.552654,
		0.590984,
		0.628601,
		0.665277,
		0.700791,
		0.734932,
		0.767497,
		0.7983,
		0.827169,
		0.853952,
		0.878519,
		0.900765,
		0.920616,
		0.938031,
		0.953007,
		0.965583,
		0.975845,
		0.983926,
		0.990011,
		0.994331,
		0.997165,
		0.998826,
		0.999643,
		0.999941,
		0.999998,
		1,
		0.999974,
		0.999767,
		0.999078,
		0.997501,
		0.994569,
		0.989783,
		0.982646,
		0.972685,
		0.959469,
		0.942622,
		0.921837,
		0.896877,
		0.86759,
		0.833904,
		0.795839,
		0.753505,
		0.707107,
		0.656948,
		0.603434,
		0.547074,
		0.488488,
		0.428411,
		0.367695,
		0.307317,
		0.248386,
		0.192146,
		0.139988,
		0.093452,
		0.05424,
		0.024223,
		0.00545,
		0.377913,
		0.416289,
		0.455022,
		0.493875,
		0.532609,
		0.570991,
		0.608789,
		0.645775,
		0.681729,
		0.716437,
		0.749696,
		0.781314,
		0.811114,
		0.838934,
		0.864634,
		0.888097,
		0.909232,
		0.927979,
		0.944313,
		0.958246,
		0.969836,
		0.979184,
		0.98644,
		0.991804,
		0.995521,
		0.997879,
		0.999193,
		0.99979,
		0.999976,
		1,
		0.999998,
		0.999939,
		0.999618,
		0.998703,
		0.996771,
		0.99335,
		0.987949,
		0.980084,
		0.969301,
		0.955189,
		0.937398,
		0.915644,
		0.889716,
		0.859486,
		0.824905,
		0.786014,
		0.742944,
		0.695919,
		0.645261,
		0.591389,
		0.53483,
		0.476215,
		0.416289,
		0.355915,
		0.296078,
		0.237891,
		0.182603,
		0.131607,
		0.086442,
		0.048808,
		0.02057,
		0.003771,
		0.435513,
		0.474108,
		0.512705,
		0.551073,
		0.588982,
		0.626207,
		0.662527,
		0.697728,
		0.731605,
		0.763962,
		0.794618,
		0.823405,
		0.850172,
		0.874792,
		0.897159,
		0.917197,
		0.93486,
		0.950139,
		0.963062,
		0.973702,
		0.982179,
		0.988657,
		0.99335,
		0.996517,
		0.998451,
		0.999468,
		0.999886,
		0.999992,
		1,
		0.999992,
		0.999878,
		0.999412,
		0.998234,
		0.995911,
		0.991969,
		0.985926,
		0.977312,
		0.965692,
		0.950679,
		0.931946,
		0.909232,
		0.882352,
		0.851199,
		0.815748,
		0.776061,
		0.732288,
		0.684672,
		0.63355,
		0.579359,
		0.522638,
		0.46403,
		0.404291,
		0.344293,
		0.285028,
		0.227614,
		0.173305,
		0.123492,
		0.079717,
		0.043674,
		0.017226,
		0.002404,
		0.492972,
		0.531263,
		0.569217,
		0.606611,
		0.643226,
		0.678846,
		0.713267,
		0.746289,
		0.777728,
		0.80741,
		0.835177,
		0.86089,
		0.884434,
		0.905717,
		0.924674,
		0.941275,
		0.955526,
		0.967473,
		0.977203,
		0.984851,
		0.990599,
		0.994673,
		0.99734,
		0.9989,
		0.999666,
		0.999945,
		0.999998,
		1,
		0.999976,
		0.999782,
		0.999138,
		0.997662,
		0.994912,
		0.990418,
		0.983707,
		0.974325,
		0.961859,
		0.945942,
		0.92627,
		0.902608,
		0.874792,
		0.842738,
		0.806444,
		0.76599,
		0.721547,
		0.673374,
		0.621825,
		0.567351,
		0.510503,
		0.451939,
		0.392422,
		0.332833,
		0.27417,
		0.217558,
		0.164251,
		0.115644,
		0.073275,
		0.038838,
		0.014187,
		0.001347,
		0.549528,
		0.587025,
		0.623865,
		0.659834,
		0.694725,
		0.72834,
		0.760489,
		0.790993,
		0.819691,
		0.846434,
		0.871096,
		0.893571,
		0.913781,
		0.931675,
		0.947239,
		0.960493,
		0.971498,
		0.980357,
		0.987222,
		0.992288,
		0.995792,
		0.99801,
		0.999244,
		0.999803,
		0.999977,
		1,
		0.999998,
		0.999943,
		0.999643,
		0.998786,
		0.996976,
		0.993766,
		0.988691,
		0.981289,
		0.971124,
		0.957802,
		0.94098,
		0.920376,
		0.895777,
		0.867043,
		0.834112,
		0.797,
		0.75581,
		0.710728,
		0.662033,
		0.610092,
		0.555372,
		0.498434,
		0.439948,
		0.380687,
		0.321539,
		0.263508,
		0.207724,
		0.155444,
		0.108062,
		0.067117,
		0.034297,
		0.011451,
		0.000596,
		0.604481,
		0.64073,
		0.676022,
		0.710157,
		0.742944,
		0.774201,
		0.803759,
		0.831466,
		0.857183,
		0.880796,
		0.902212,
		0.921364,
		0.938217,
		0.952769,
		0.965057,
		0.975156,
		0.983187,
		0.989313,
		0.993744,
		0.996728,
		0.998546,
		0.999501,
		0.999893,
		0.999993,
		1,
		0.999993,
		0.999886,
		0.99945,
		0.998347,
		0.996168,
		0.992466,
		0.986782,
		0.978669,
		0.967708,
		0.953525,
		0.935798,
		0.914268,
		0.888746,
		0.859113,
		0.825327,
		0.787425,
		0.745529,
		0.699842,
		0.650658,
		0.598361,
		0.543429,
		0.486437,
		0.428063,
		0.369091,
		0.310415,
		0.253045,
		0.198115,
		0.146884,
		0.100747,
		0.061241,
		0.030049,
		0.009016,
		0.000148,
		0.657196,
		0.691782,
		0.725136,
		0.757075,
		0.787425,
		0.816028,
		0.842738,
		0.867431,
		0.890001,
		0.910368,
		0.928479,
		0.944313,
		0.957881,
		0.969235,
		0.978467,
		0.985711,
		0.991144,
		0.994989,
		0.997501,
		0.998968,
		0.999687,
		0.999948,
		0.999999,
		1,
		0.999977,
		0.999797,
		0.999193,
		0.997809,
		0.995229,
		0.991005,
		0.984688,
		0.975845,
		0.964076,
		0.949027,
		0.930398,
		0.907953,
		0.881522,
		0.851009,
		0.816392,
		0.777728,
		0.735156,
		0.688896,
		0.639258,
		0.586639,
		0.531529,
		0.474517,
		0.416289,
		0.357638,
		0.299464,
		0.242783,
		0.188732,
		0.138573,
		0.093699,
		0.055645,
		0.026093,
		0.006878,
		0.707107,
		0.739658,
		0.770731,
		0.800163,
		0.827802,
		0.853514,
		0.877185,
		0.89872,
		0.918052,
		0.935141,
		0.949978,
		0.962592,
		0.973047,
		0.98145,
		0.987949,
		0.992736,
		0.996042,
		0.998131,
		0.999291,
		0.999816,
		0.999979,
		1,
		0.999999,
		0.999947,
		0.999666,
		0.998863,
		0.997165,
		0.99415,
		0.989376,
		0.982404,
		0.972816,
		0.96023,
		0.944313,
		0.924786,
		0.901435,
		0.874111,
		0.842738,
		0.807315,
		0.767917,
		0.724699,
		0.677898,
		0.627839,
		0.574932,
		0.51968,
		0.462682,
		0.404632,
		0.346332,
		0.28869,
		0.232725,
		0.179577,
		0.130509,
		0.086916,
		0.05033,
		0.022426,
		0.005035,
		0.753719,
		0.783913,
		0.812415,
		0.839085,
		0.863799,
		0.886452,
		0.906963,
		0.925275,
		0.941362,
		0.95523,
		0.96692,
		0.976511,
		0.984125,
		0.989923,
		0.994109,
		0.996923,
		0.998635,
		0.999532,
		0.9999,
		0.999993,
		1,
		0.999993,
		0.999893,
		0.999485,
		0.998451,
		0.996405,
		0.992925,
		0.987575,
		0.979927,
		0.96958,
		0.95617,
		0.939384,
		0.918966,
		0.89472,
		0.86652,
		0.834309,
		0.798105,
		0.758,
		0.714166,
		0.666857,
		0.61641,
		0.563249,
		0.507888,
		0.450936,
		0.393097,
		0.335179,
		0.278096,
		0.222873,
		0.170651,
		0.122695,
		0.080399,
		0.045292,
		0.019046,
		0.003484,
		0.796619,
		0.824185,
		0.849883,
		0.873601,
		0.895243,
		0.914741,
		0.932051,
		0.947158,
		0.960083,
		0.970881,
		0.979645,
		0.98651,
		0.991651,
		0.995282,
		0.997651,
		0.999031,
		0.999707,
		0.999952,
		0.999999,
		1,
		0.999979,
		0.99981,
		0.999244,
		0.997946,
		0.995521,
		0.991548,
		0.985597,
		0.977255,
		0.966137,
		0.951898,
		0.934245,
		0.912942,
		0.887815,
		0.858756,
		0.825729,
		0.788769,
		0.747985,
		0.703566,
		0.655781,
		0.604978,
		0.551596,
		0.496159,
		0.439285,
		0.381687,
		0.324181,
		0.267685,
		0.213228,
		0.161955,
		0.11513,
		0.074147,
		0.040531,
		0.015951,
		0.002221,
		0.835475,
		0.860201,
		0.882926,
		0.903567,
		0.922066,
		0.938392,
		0.952544,
		0.964555,
		0.974494,
		0.982468,
		0.988626,
		0.993153,
		0.996274,
		0.998243,
		0.999334,
		0.999827,
		0.99998,
		1,
		0.999999,
		0.99995,
		0.999687,
		0.998934,
		0.99734,
		0.994505,
		0.990011,
		0.983438,
		0.974386,
		0.962487,
		0.947416,
		0.928899,
		0.90672,
		0.880725,
		0.850827,
		0.817006,
		0.779315,
		0.737881,
		0.692907,
		0.644676,
		0.59355,
		0.539979,
		0.484498,
		0.427734,
		0.370409,
		0.313342,
		0.25746,
		0.203793,
		0.15349,
		0.107814,
		0.068159,
		0.036046,
		0.013138,
		0.001245,
		0.870045,
		0.891784,
		0.911434,
		0.92895,
		0.944313,
		0.957534,
		0.968661,
		0.977776,
		0.984999,
		0.99049,
		0.994448,
		0.997104,
		0.998717,
		0.999561,
		0.999906,
		0.999994,
		1,
		0.999993,
		0.9999,
		0.999517,
		0.998546,
		0.996625,
		0.99335,
		0.98831,
		0.981094,
		0.971319,
		0.958631,
		0.942727,
		0.92335,
		0.900305,
		0.873458,
		0.842738,
		0.808146,
		0.769751,
		0.727694,
		0.682196,
		0.63355,
		0.582134,
		0.528406,
		0.472913,
		0.416289,
		0.359265,
		0.302666,
		0.247423,
		0.19457,
		0.145256,
		0.100747,
		0.062433,
		0.031834,
		0.010606,
		0.000551,
		0.900183,
		0.918855,
		0.935405,
		0.949827,
		0.962145,
		0.972419,
		0.980745,
		0.987255,
		0.992122,
		0.995554,
		0.997789,
		0.999089,
		0.999725,
		0.999955,
		0.999999,
		1,
		0.99998,
		0.999822,
		0.999291,
		0.998072,
		0.995792,
		0.99205,
		0.98644,
		0.978564,
		0.968053,
		0.954571,
		0.937833,
		0.917603,
		0.893703,
		0.866019,
		0.834498,
		0.799158,
		0.760084,
		0.717434,
		0.67144,
		0.622411,
		0.570735,
		0.516882,
		0.461407,
		0.404954,
		0.34826,
		0.292156,
		0.237576,
		0.185559,
		0.137255,
		0.09393,
		0.05697,
		0.027894,
		0.008351,
		0.000137,
		0.925841,
		0.941445,
		0.954949,
		0.966392,
		0.975845,
		0.983418,
		0.989256,
		0.993541,
		0.99649,
		0.998347,
		0.999375,
		0.999838,
		0.999981,
		1,
		0.999999,
		0.999953,
		0.999707,
		0.999,
		0.997501,
		0.994834,
		0.990599,
		0.984397,
		0.975845,
		0.964588,
		0.950308,
		0.932739,
		0.911662,
		0.88692,
		0.858415,
		0.826114,
		0.790049,
		0.750323,
		0.707107,
		0.660647,
		0.611266,
		0.559361,
		0.505413,
		0.449987,
		0.393734,
		0.337397,
		0.281814,
		0.227922,
		0.176763,
		0.129487,
		0.08736,
		0.051768,
		0.024223,
		0.006372,
		0.947082,
		0.959694,
		0.970291,
		0.978957,
		0.985814,
		0.991018,
		0.994763,
		0.997272,
		0.998793,
		0.999588,
		0.999912,
		0.999994,
		1,
		0.999994,
		0.999906,
		0.999547,
		0.998635,
		0.996827,
		0.993744,
		0.988991,
		0.982179,
		0.972936,
		0.960924,
		0.945845,
		0.927446,
		0.905532,
		0.87996,
		0.850653,
		0.817592,
		0.780826,
		0.740474,
		0.69672,
		0.649825,
		0.60012,
		0.548017,
		0.494005,
		0.438658,
		0.382633,
		0.326681,
		0.271644,
		0.218463,
		0.168182,
		0.121952,
		0.081039,
		0.046826,
		0.020821,
		0.004665,
		0.964076,
		0.973858,
		0.981771,
		0.987949,
		0.992561,
		0.995807,
		0.997917,
		0.999143,
		0.999741,
		0.999957,
		0.999999,
		1,
		0.999981,
		0.999833,
		0.999334,
		0.998188,
		0.996042,
		0.992516,
		0.987222,
		0.979781,
		0.969836,
		0.957064,
		0.941183,
		0.921961,
		0.899218,
		0.872831,
		0.842738,
		0.808939,
		0.771497,
		0.730545,
		0.686282,
		0.638979,
		0.588982,
		0.536711,
		0.482664,
		0.427424,
		0.371655,
		0.316114,
		0.261647,
		0.2092,
		0.159817,
		0.114651,
		0.074965,
		0.042142,
		0.017685,
		0.003228,
		0.977109,
		0.984305,
		0.989843,
		0.993901,
		0.99669,
		0.998443,
		0.999412,
		0.999848,
		0.999982,
		1,
		0.999999,
		0.999956,
		0.999725,
		0.999061,
		0.997651,
		0.995138,
		0.991144,
		0.985289,
		0.977203,
		0.966545,
		0.953007,
		0.936326,
		0.916286,
		0.892725,
		0.865538,
		0.834679,
		0.800163,
		0.762069,
		0.720543,
		0.675798,
		0.628118,
		0.577857,
		0.525446,
		0.471394,
		0.416289,
		0.360804,
		0.3057,
		0.251827,
		0.200135,
		0.151669,
		0.107582,
		0.069138,
		0.037715,
		0.014812,
		0.002059,
		0.986574,
		0.99151,
		0.995056,
		0.997428,
		0.998863,
		0.999612,
		0.999917,
		0.999994,
		1,
		0.999994,
		0.999912,
		0.999575,
		0.998717,
		0.997015,
		0.994109,
		0.989624,
		0.983187,
		0.974442,
		0.963062,
		0.948756,
		0.931276,
		0.910426,
		0.886059,
		0.858088,
		0.826481,
		0.79127,
		0.752549,
		0.710476,
		0.665277,
		0.617247,
		0.566751,
		0.51423,
		0.460201,
		0.405259,
		0.350084,
		0.295441,
		0.242186,
		0.191269,
		0.143738,
		0.100747,
		0.063557,
		0.033544,
		0.012202,
		0.001154,
		0.992969,
		0.996042,
		0.998037,
		0.999193,
		0.999757,
		0.99996,
		0.999999,
		1,
		0.999982,
		0.999843,
		0.999375,
		0.998296,
		0.996274,
		0.992949,
		0.987949,
		0.980913,
		0.971498,
		0.959389,
		0.944313,
		0.926039,
		0.904386,
		0.879226,
		0.850486,
		0.818151,
		0.782268,
		0.742944,
		0.70035,
		0.654724,
		0.606372,
		0.555671,
		0.503068,
		0.449089,
		0.394338,
		0.339498,
		0.285341,
		0.232725,
		0.182603,
		0.136026,
		0.094146,
		0.058222,
		0.029626,
		0.009851,
		0.000511,
		0.996876,
		0.998532,
		0.999446,
		0.999857,
		0.999984,
		1,
		0.999999,
		0.999959,
		0.999741,
		0.999117,
		0.997789,
		0.995421,
		0.991651,
		0.986118,
		0.978467,
		0.968369,
		0.955526,
		0.93968,
		0.920616,
		0.89817,
		0.87223,
		0.842738,
		0.809696,
		0.773163,
		0.73326,
		0.690172,
		0.644146,
		0.595501,
		0.544621,
		0.491965,
		0.438064,
		0.383529,
		0.32905,
		0.275401,
		0.223446,
		0.17414,
		0.128533,
		0.087777,
		0.05313,
		0.025962,
		0.007758,
		0.000127,
		0.998929,
		0.999635,
		0.999922,
		0.999995,
		1,
		0.999995,
		0.999917,
		0.9996,
		0.998793,
		0.997189,
		0.994448,
		0.990212,
		0.984125,
		0.975845,
		0.965057,
		0.951476,
		0.93486,
		0.915012,
		0.891784,
		0.865077,
		0.834852,
		0.801123,
		0.763962,
		0.723506,
		0.679948,
		0.63355,
		0.584639,
		0.533608,
		0.480925,
		0.427129,
		0.372836,
		0.318742,
		0.265625,
		0.214352,
		0.165879,
		0.121258,
		0.08164,
		0.048281,
		0.022547,
		0.00592,
		0.999771,
		0.999962,
		0.999999,
		1,
		0.999984,
		0.999852,
		0.999412,
		0.998396,
		0.99649,
		0.99335,
		0.988626,
		0.981968,
		0.973047,
		0.96156,
		0.947239,
		0.929857,
		0.909232,
		0.885232,
		0.857774,
		0.826833,
		0.792437,
		0.754672,
		0.713686,
		0.669685,
		0.622942,
		0.573791,
		0.522638,
		0.469955,
		0.416289,
		0.362263,
		0.308577,
		0.256014,
		0.205443,
		0.157821,
		0.114202,
		0.075735,
		0.043674,
		0.019382,
		0.004334,
		0.999985,
		1,
		0.999999,
		0.999961,
		0.999757,
		0.999169,
		0.997917,
		0.995683,
		0.992122,
		0.986889,
		0.979645,
		0.970072,
		0.957881,
		0.942819,
		0.924674,
		0.90328,
		0.878519,
		0.850326,
		0.818687,
		0.783645,
		0.745299,
		0.703808,
		0.65939,
		0.612327,
		0.562964,
		0.511714,
		0.459058,
		0.405548,
		0.351814,
		0.298559,
		0.246571,
		0.196721,
		0.149968,
		0.107365,
		0.070061,
		0.039308,
		0.016464,
		0.003,
		1,
		0.999995,
		0.999922,
		0.999624,
		0.998863,
		0.997351,
		0.994763,
		0.990759,
		0.984999,
		0.977154,
		0.96692,
		0.95402,
		0.938217,
		0.919314,
		0.897159,
		0.871652,
		0.842738,
		0.810421,
		0.774754,
		0.73585,
		0.693878,
		0.649069,
		0.601712,
		0.552163,
		0.500843,
		0.448238,
		0.39491,
		0.34149,
		0.28869,
		0.237298,
		0.188187,
		0.14232,
		0.100747,
		0.064618,
		0.035182,
		0.013791,
		0.001913,
		0.999985,
		0.999861,
		0.999446,
		0.998488,
		0.99669,
		0.993724,
		0.989256,
		0.982951,
		0.974494,
		0.96359,
		0.949978,
		0.933436,
		0.913781,
		0.890876,
		0.864634,
		0.835018,
		0.80204,
		0.76577,
		0.726331,
		0.683903,
		0.638727,
		0.591103,
		0.541394,
		0.490028,
		0.437501,
		0.384378,
		0.331297,
		0.278972,
		0.228195,
		0.179843,
		0.134877,
		0.094348,
		0.059405,
		0.031293,
		0.011362,
		0.001072,
		0.999771,
		0.999217,
		0.998037,
		0.995927,
		0.992561,
		0.987608,
		0.980745,
		0.971663,
		0.960083,
		0.945759,
		0.928479,
		0.908079,
		0.884434,
		0.857473,
		0.827169,
		0.793551,
		0.756699,
		0.716747,
		0.673888,
		0.628371,
		0.580504,
		0.530661,
		0.479275,
		0.42685,
		0.373956,
		0.321236,
		0.269407,
		0.219266,
		0.171689,
		0.127639,
		0.088168,
		0.054421,
		0.027641,
		0.009174,
		0.000475,
		0.998929,
		0.997501,
		0.995056,
		0.991269,
		0.985814,
		0.978376,
		0.968661,
		0.956401,
		0.941362,
		0.92335,
		0.902212,
		0.877839,
		0.850172,
		0.8192,
		0.784961,
		0.747548,
		0.707107,
		0.66384,
		0.618005,
		0.569922,
		0.519969,
		0.468589,
		0.416289,
		0.363647,
		0.31131,
		0.259999,
		0.210511,
		0.163726,
		0.120607,
		0.082205,
		0.049664,
		0.024223,
		0.007225,
		0.000118,
		0.996876,
		0.994072,
		0.989843,
		0.983869,
		0.975845,
		0.965488,
		0.952544,
		0.936792,
		0.918052,
		0.896184,
		0.871096,
		0.842738,
		0.811114,
		0.776274,
		0.738322,
		0.697414,
		0.653763,
		0.607637,
		0.559361,
		0.509323,
		0.457972,
		0.405822,
		0.353455,
		0.301522,
		0.250747,
		0.201932,
		0.155955,
		0.113781,
		0.076461,
		0.045134,
		0.02104,
		0.005514,
		0.992969,
		0.988279,
		0.981771,
		0.97315,
		0.962145,
		0.948514,
		0.932051,
		0.912588,
		0.890001,
		0.864209,
		0.835177,
		0.802918,
		0.767497,
		0.729027,
		0.687676,
		0.643665,
		0.59727,
		0.548826,
		0.498728,
		0.44743,
		0.395453,
		0.343382,
		0.291874,
		0.241656,
		0.193529,
		0.148377,
		0.107161,
		0.070933,
		0.04083,
		0.018088,
		0.004038,
		0.986574,
		0.979518,
		0.970291,
		0.958631,
		0.944313,
		0.927141,
		0.906963,
		0.883666,
		0.857183,
		0.827492,
		0.794618,
		0.758636,
		0.71967,
		0.677898,
		0.63355,
		0.586911,
		0.538323,
		0.488188,
		0.436966,
		0.385184,
		0.333432,
		0.282369,
		0.232725,
		0.185305,
		0.140991,
		0.100747,
		0.065621,
		0.036751,
		0.015366,
		0.002795,
		0.977109,
		0.967266,
		0.954949,
		0.939942,
		0.922066,
		0.90118,
		0.877185,
		0.850025,
		0.819691,
		0.78622,
		0.749696,
		0.710256,
		0.668086,
		0.623424,
		0.576564,
		0.527856,
		0.477707,
		0.426585,
		0.37502,
		0.323607,
		0.273008,
		0.223957,
		0.17726,
		0.133799,
		0.094539,
		0.060526,
		0.032895,
		0.012873,
		0.001783,
		0.964076,
		0.951098,
		0.935405,
		0.916829,
		0.895243,
		0.870561,
		0.842738,
		0.811778,
		0.777728,
		0.740684,
		0.700791,
		0.658245,
		0.613292,
		0.566235,
		0.51743,
		0.46729,
		0.416289,
		0.364963,
		0.313909,
		0.263794,
		0.215353,
		0.169394,
		0.126801,
		0.088536,
		0.055645,
		0.029261,
		0.010606,
		0.000999,
		0.947082,
		0.930704,
		0.911434,
		0.889157,
		0.863799,
		0.835329,
		0.803759,
		0.769149,
		0.731605,
		0.69128,
		0.64838,
		0.60316,
		0.555928,
		0.507048,
		0.45694,
		0.406083,
		0.355015,
		0.304341,
		0.254728,
		0.206915,
		0.16171,
		0.119997,
		0.082738,
		0.050979,
		0.025848,
		0.008564,
		0.000443,
		0.925841,
		0.905884,
		0.882926,
		0.856905,
		0.827802,
		0.79564,
		0.760489,
		0.722464,
		0.681729,
		0.638497,
		0.593031,
		0.545648,
		0.496716,
		0.446662,
		0.395969,
		0.345181,
		0.294906,
		0.245813,
		0.198643,
		0.154206,
		0.113386,
		0.077145,
		0.046525,
		0.022654,
		0.006746,
		0.00011,
		0.900183,
		0.876554,
		0.849883,
		0.820162,
		0.787425,
		0.751752,
		0.713267,
		0.672143,
		0.628601,
		0.582912,
		0.535398,
		0.486437,
		0.436458,
		0.385951,
		0.335463,
		0.285604,
		0.237049,
		0.190539,
		0.146884,
		0.10697,
		0.071756,
		0.042284,
		0.019678,
		0.005149,
		0.870045,
		0.842738,
		0.812415,
		0.779121,
		0.742944,
		0.704019,
		0.662527,
		0.618696,
		0.572806,
		0.525185,
		0.476215,
		0.426333,
		0.376031,
		0.325863,
		0.27644,
		0.228439,
		0.182603,
		0.139745,
		0.100747,
		0.066571,
		0.038254,
		0.016918,
		0.003771,
		0.835475,
		0.804566,
		0.770731,
		0.73407,
		0.694725,
		0.652886,
		0.608789,
		0.562719,
		0.515011,
		0.466054,
		0.416289,
		0.366214,
		0.316383,
		0.267414,
		0.219983,
		0.174837,
		0.132788,
		0.094719,
		0.061589,
		0.034434,
		0.014374,
		0.00261,
		0.796619,
		0.762263,
		0.725136,
		0.685391,
		0.643226,
		0.598882,
		0.552654,
		0.504881,
		0.455958,
		0.406331,
		0.3565,
		0.307026,
		0.258527,
		0.211684,
		0.167242,
		0.126014,
		0.088884,
		0.056809,
		0.030823,
		0.012042,
		0.001665,
		0.753719,
		0.716147,
		0.676022,
		0.63355,
		0.588982,
		0.542616,
		0.494799,
		0.44593,
		0.39646,
		0.346894,
		0.297795,
		0.249783,
		0.203541,
		0.159817,
		0.119423,
		0.083242,
		0.05223,
		0.02742,
		0.009923,
		0.000934,
		0.707107,
		0.666622,
		0.623865,
		0.579092,
		0.532609,
		0.484769,
		0.435974,
		0.386681,
		0.337397,
		0.28869,
		0.241182,
		0.195557,
		0.152563,
		0.113014,
		0.077793,
		0.047852,
		0.024223,
		0.008014,
		0.000414,
		0.657196,
		0.614173,
		0.569217,
		0.522638,
		0.474793,
		0.426093,
		0.376995,
		0.328012,
		0.279714,
		0.232725,
		0.187731,
		0.145482,
		0.106789,
		0.072536,
		0.043674,
		0.021232,
		0.006313,
		0.000103,
		0.604481,
		0.559361,
		0.512705,
		0.464877,
		0.416289,
		0.367405,
		0.318742,
		0.270869,
		0.224414,
		0.180066,
		0.138573,
		0.100747,
		0.067471,
		0.039695,
		0.018444,
		0.004818,
		0.549528,
		0.502815,
		0.455022,
		0.406567,
		0.357915,
		0.309587,
		0.262156,
		0.216251,
		0.172561,
		0.131836,
		0.094888,
		0.062598,
		0.035914,
		0.015859,
		0.003529,
		0.492972,
		0.445233,
		0.396928,
		0.348527,
		0.300551,
		0.253577,
		0.208236,
		0.165218,
		0.125272,
		0.089212,
		0.057915,
		0.03233,
		0.013475,
		0.002443,
		0.435513,
		0.387376,
		0.339242,
		0.291635,
		0.245133,
		0.200371,
		0.158037,
		0.118882,
		0.083718,
		0.053423,
		0.028941,
		0.01129,
		0.001559,
		0.377913,
		0.330063,
		0.282841,
		0.236827,
		0.192655,
		0.151018,
		0.112664,
		0.078406,
		0.049119,
		0.025748,
		0.009304,
		0.000874,
		0.320992,
		0.27417,
		0.228658,
		0.185091,
		0.144161,
		0.106619,
		0.073275,
		0.045004,
		0.022747,
		0.007514,
		0.000387,
		0.265625,
		0.220628,
		0.177679,
		0.137468,
		0.100747,
		0.068326,
		0.041077,
		0.019939,
		0.00592,
		9.6e-005,
		0.212739,
		0.170419,
		0.130939,
		0.095049,
		0.063557,
		0.037337,
		0.017323,
		0.004519,
		0.163312,
		0.124573,
		0.089522,
		0.058969,
		0.033782,
		0.014896,
		0.00331,
		0.118371,
		0.084169,
		0.05456,
		0.030412,
		0.012657,
		0.002292,
		0.078987,
		0.05033,
		0.027226,
		0.010606,
		0.001462,
		0.046277,
		0.024223,
		0.008741,
		0.00082,
		0.021402,
		0.00706,
		0.000363,
		0.005562,
		9.1e-005,
		0.005605,
		0.005649,
		0.021564,
		0.005692,
		0.021727,
		0.046619,
		0.005737,
		0.021892,
		0.046964,
		0.079556,
		0.005782,
		0.022059,
		0.047314,
		0.080132,
		0.119202,
		0.005827,
		0.022228,
		0.047667,
		0.080713,
		0.120042,
		0.164428,
		0.005873,
		0.022399,
		0.048024,
		0.081301,
		0.120891,
		0.165555,
		0.214149,
		0.00592,
		0.022572,
		0.048385,
		0.081895,
		0.121749,
		0.166693,
		0.215573,
		0.267329,
		0.005967,
		0.022747,
		0.04875,
		0.082496,
		0.122616,
		0.167843,
		0.217011,
		0.269049,
		0.32298,
		0.006014,
		0.022924,
		0.049119,
		0.083104,
		0.123492,
		0.169005,
		0.218463,
		0.270786,
		0.324985,
		0.380165,
		0.006063,
		0.023103,
		0.049493,
		0.083718,
		0.124377,
		0.170178,
		0.219929,
		0.272538,
		0.327009,
		0.382435,
		0.438,
		0.006111,
		0.023285,
		0.049871,
		0.084339,
		0.125272,
		0.171364,
		0.221409,
		0.274307,
		0.32905,
		0.384724,
		0.440506,
		0.495659,
		0.006161,
		0.023468,
		0.050253,
		0.084966,
		0.126177,
		0.172561,
		0.222905,
		0.276092,
		0.331109,
		0.387032,
		0.443032,
		0.498366,
		0.552374,
		0.006211,
		0.023653,
		0.050639,
		0.085601,
		0.127091,
		0.173771,
		0.224414,
		0.277895,
		0.333186,
		0.38936,
		0.445578,
		0.501091,
		0.555238,
		0.607439,
		0.006261,
		0.023841,
		0.05103,
		0.086243,
		0.128015,
		0.174994,
		0.225939,
		0.279714,
		0.335282,
		0.391707,
		0.448143,
		0.503836,
		0.55812,
		0.610413,
		0.660215,
		0.006313,
		0.024031,
		0.051425,
		0.086892,
		0.128949,
		0.17623,
		0.227479,
		0.28155,
		0.337397,
		0.394074,
		0.450728,
		0.5066,
		0.561021,
		0.613403,
		0.663248,
		0.710133,
		0.006364,
		0.024223,
		0.051826,
		0.087549,
		0.129893,
		0.177478,
		0.229035,
		0.283404,
		0.339531,
		0.39646,
		0.453333,
		0.509384,
		0.563939,
		0.61641,
		0.666294,
		0.71317,
		0.756699,
		0.006417,
		0.024418,
		0.05223,
		0.088212,
		0.130847,
		0.178739,
		0.230606,
		0.285276,
		0.341684,
		0.398867,
		0.455958,
		0.512188,
		0.566876,
		0.619433,
		0.669354,
		0.716218,
		0.759685,
		0.799498,
		0.00647,
		0.024615,
		0.05264,
		0.088884,
		0.131812,
		0.180014,
		0.232193,
		0.287165,
		0.343856,
		0.401293,
		0.458604,
		0.515011,
		0.569832,
		0.622473,
		0.672427,
		0.719275,
		0.762678,
		0.802379,
		0.838201,
		0.006524,
		0.024814,
		0.053054,
		0.089563,
		0.132788,
		0.181302,
		0.233795,
		0.289073,
		0.346048,
		0.403741,
		0.46127,
		0.517854,
		0.572806,
		0.625529,
		0.675515,
		0.722343,
		0.765677,
		0.805262,
		0.840925,
		0.872571,
		0.006578,
		0.025016,
		0.053473,
		0.090249,
		0.133774,
		0.182603,
		0.235414,
		0.290998,
		0.34826,
		0.406208,
		0.463957,
		0.520718,
		0.575799,
		0.628601,
		0.678615,
		0.72542,
		0.768681,
		0.808146,
		0.843645,
		0.875088,
		0.902465,
		0.006633,
		0.02522,
		0.053897,
		0.090944,
		0.134771,
		0.183919,
		0.237049,
		0.292943,
		0.350491,
		0.408697,
		0.466665,
		0.523601,
		0.57881,
		0.631689,
		0.681729,
		0.728508,
		0.771692,
		0.811031,
		0.846361,
		0.877597,
		0.904734,
		0.927846,
		0.006689,
		0.025427,
		0.054327,
		0.091646,
		0.135779,
		0.185248,
		0.238701,
		0.294906,
		0.352743,
		0.411206,
		0.469393,
		0.526505,
		0.58184,
		0.634794,
		0.684856,
		0.731605,
		0.774707,
		0.813918,
		0.849074,
		0.880097,
		0.90699,
		0.929834,
		0.948785,
		0.006746,
		0.025636,
		0.054761,
		0.092357,
		0.136799,
		0.186592,
		0.24037,
		0.296887,
		0.355015,
		0.413737,
		0.472143,
		0.529428,
		0.584888,
		0.637915,
		0.687996,
		0.734711,
		0.777728,
		0.816804,
		0.851782,
		0.882588,
		0.909232,
		0.931803,
		0.950468,
		0.965467,
		0.006803,
		0.025848,
		0.0552,
		0.093076,
		0.137829,
		0.18795,
		0.242056,
		0.298888,
		0.357308,
		0.416289,
		0.474914,
		0.532373,
		0.587956,
		0.641052,
		0.69115,
		0.737827,
		0.780754,
		0.819691,
		0.854485,
		0.88507,
		0.91146,
		0.933754,
		0.952128,
		0.966832,
		0.978187,
		0.006861,
		0.026062,
		0.055645,
		0.093803,
		0.138872,
		0.189322,
		0.243759,
		0.300909,
		0.359622,
		0.418863,
		0.477707,
		0.535338,
		0.591042,
		0.644206,
		0.694316,
		0.740951,
		0.783784,
		0.822577,
		0.857183,
		0.887541,
		0.913673,
		0.935686,
		0.953766,
		0.968173,
		0.97924,
		0.987358,
		0.00692,
		0.026279,
		0.056095,
		0.094539,
		0.139926,
		0.19071,
		0.245479,
		0.302949,
		0.361957,
		0.421459,
		0.480522,
		0.538323,
		0.594146,
		0.647376,
		0.697495,
		0.744084,
		0.786818,
		0.825463,
		0.859876,
		0.890001,
		0.91587,
		0.937598,
		0.955381,
		0.969489,
		0.980266,
		0.988116,
		0.99349,
		0.006979,
		0.026499,
		0.056551,
		0.095283,
		0.140991,
		0.192112,
		0.247217,
		0.305009,
		0.364313,
		0.424076,
		0.483358,
		0.54133,
		0.59727,
		0.650561,
		0.700686,
		0.747226,
		0.789856,
		0.828347,
		0.862562,
		0.89245,
		0.918052,
		0.939491,
		0.956972,
		0.97078,
		0.981267,
		0.988848,
		0.993987,
		0.99718,
		0.00704,
		0.026722,
		0.057012,
		0.096037,
		0.142069,
		0.193529,
		0.248973,
		0.307089,
		0.366691,
		0.426716,
		0.486216,
		0.544357,
		0.600413,
		0.653763,
		0.70389,
		0.750376,
		0.792898,
		0.83123,
		0.865241,
		0.894888,
		0.920217,
		0.941362,
		0.95854,
		0.972044,
		0.98224,
		0.989554,
		0.994461,
		0.997465,
		0.999074,
		0.007101,
		0.026947,
		0.057479,
		0.096799,
		0.14316,
		0.194962,
		0.250747,
		0.309189,
		0.369091,
		0.429378,
		0.489096,
		0.547405,
		0.603574,
		0.656981,
		0.707107,
		0.753533,
		0.795942,
		0.834112,
		0.867913,
		0.897313,
		0.922365,
		0.943213,
		0.960083,
		0.973282,
		0.983187,
		0.990235,
		0.994912,
		0.997731,
		0.999205,
		0.999819,
		0.007163,
		0.027175,
		0.057952,
		0.09757,
		0.144262,
		0.196411,
		0.25254,
		0.31131,
		0.371512,
		0.432063,
		0.491999,
		0.550474,
		0.606755,
		0.660215,
		0.710335,
		0.756699,
		0.79899,
		0.83699,
		0.870578,
		0.899725,
		0.924495,
		0.945041,
		0.961602,
		0.974494,
		0.984106,
		0.99089,
		0.99534,
		0.997978,
		0.999323,
		0.999859,
		0.999991,
		0.007225,
		0.027406,
		0.05843,
		0.09835,
		0.145377,
		0.197875,
		0.254351,
		0.313452,
		0.373956,
		0.43477,
		0.494924,
		0.553565,
		0.609954,
		0.663465,
		0.713576,
		0.759872,
		0.80204,
		0.839866,
		0.873234,
		0.902123,
		0.926607,
		0.946848,
		0.963095,
		0.975678,
		0.984999,
		0.991519,
		0.995745,
		0.998208,
		0.999429,
		0.999892,
		0.999995,
		1,
		0.007289,
		0.027641,
		0.058915,
		0.09914,
		0.146505,
		0.199356,
		0.256181,
		0.315615,
		0.376422,
		0.437501,
		0.497872,
		0.556677,
		0.613173,
		0.66673,
		0.716828,
		0.763052,
		0.805092,
		0.842738,
		0.875881,
		0.904508,
		0.9287,
		0.948631,
		0.964563,
		0.976835,
		0.985864,
		0.992122,
		0.996128,
		0.99842,
		0.999523,
		0.99992,
		0.999997,
		1,
		0.999975,
		0.007353,
		0.027878,
		0.059405,
		0.099939,
		0.147646,
		0.200852,
		0.258031,
		0.317799,
		0.378911,
		0.440255,
		0.500843,
		0.55981,
		0.61641,
		0.670011,
		0.720092,
		0.76624,
		0.808146,
		0.845607,
		0.878519,
		0.906878,
		0.930774,
		0.950392,
		0.966004,
		0.977965,
		0.986701,
		0.9927,
		0.99649,
		0.998614,
		0.999606,
		0.999942,
		0.999999,
		1,
		0.999962,
		0.999707,
		0.007418,
		0.028118,
		0.059902,
		0.100747,
		0.148801,
		0.202366,
		0.2599,
		0.320005,
		0.381424,
		0.443032,
		0.503836,
		0.562964,
		0.619667,
		0.673308,
		0.723368,
		0.769433,
		0.811201,
		0.848471,
		0.881147,
		0.909232,
		0.932827,
		0.952128,
		0.967419,
		0.979066,
		0.987511,
		0.993253,
		0.996829,
		0.998793,
		0.999679,
		0.999959,
		1,
		0.999999,
		0.999944,
		0.999629,
		0.998733,
		0.007484,
		0.028361,
		0.060404,
		0.101566,
		0.149968,
		0.203896,
		0.261788,
		0.322233,
		0.383959,
		0.445833,
		0.506853,
		0.56614,
		0.622942,
		0.67662,
		0.726654,
		0.772634,
		0.814257,
		0.851331,
		0.883765,
		0.911571,
		0.93486,
		0.95384,
		0.968806,
		0.98014,
		0.988294,
		0.99378,
		0.997147,
		0.998956,
		0.999741,
		0.999973,
		1,
		0.999997,
		0.99992,
		0.999537,
		0.998511,
		0.996452,
		0.007551,
		0.028608,
		0.060913,
		0.102394,
		0.151149,
		0.205443,
		0.263696,
		0.324482,
		0.386518,
		0.448658,
		0.509893,
		0.569338,
		0.626236,
		0.679948,
		0.729952,
		0.77584,
		0.817314,
		0.854185,
		0.886372,
		0.913893,
		0.936872,
		0.955526,
		0.970166,
		0.981184,
		0.989048,
		0.994282,
		0.997444,
		0.999103,
		0.999795,
		0.999982,
		1,
		0.999995,
		0.999889,
		0.999429,
		0.99826,
		0.995985,
		0.99221,
		0.007619,
		0.028857,
		0.061429,
		0.103233,
		0.152345,
		0.207007,
		0.265625,
		0.326755,
		0.3891,
		0.451507,
		0.512956,
		0.572558,
		0.629549,
		0.683291,
		0.73326,
		0.779051,
		0.82037,
		0.857034,
		0.888967,
		0.916199,
		0.938862,
		0.957188,
		0.971498,
		0.9822,
		0.989775,
		0.994759,
		0.997721,
		0.999236,
		0.99984,
		0.999989,
		1,
		0.99999,
		0.99985,
		0.999303,
		0.997978,
		0.995473,
		0.99139,
		0.985372,
		0.007688,
		0.02911,
		0.061951,
		0.104082,
		0.153554,
		0.208589,
		0.267574,
		0.32905,
		0.391707,
		0.454381,
		0.516043,
		0.575799,
		0.632882,
		0.686649,
		0.736579,
		0.782268,
		0.823426,
		0.859876,
		0.891549,
		0.918486,
		0.94083,
		0.958822,
		0.972801,
		0.983187,
		0.990474,
		0.995211,
		0.997978,
		0.999355,
		0.999878,
		0.999994,
		1,
		0.999982,
		0.999801,
		0.999156,
		0.997663,
		0.994912,
		0.990506,
		0.984089,
		0.975358,
		0.007758,
		0.029367,
		0.06248,
		0.104941,
		0.154777,
		0.210189,
		0.269544,
		0.331368,
		0.394338,
		0.457278,
		0.519153,
		0.579062,
		0.636233,
		0.690022,
		0.739908,
		0.78549,
		0.826481,
		0.862711,
		0.894119,
		0.920755,
		0.942774,
		0.960431,
		0.974075,
		0.984144,
		0.991144,
		0.995638,
		0.998216,
		0.999461,
		0.999909,
		0.999997,
		1,
		0.999972,
		0.999741,
		0.998987,
		0.997312,
		0.9943,
		0.989554,
		0.982721,
		0.973507,
		0.961681,
		0.007828,
		0.029626,
		0.063015,
		0.10581,
		0.156015,
		0.211807,
		0.271535,
		0.333709,
		0.396993,
		0.460201,
		0.522288,
		0.582346,
		0.639602,
		0.69341,
		0.743248,
		0.788716,
		0.829535,
		0.865538,
		0.896676,
		0.923006,
		0.944695,
		0.962012,
		0.975321,
		0.985072,
		0.991788,
		0.996042,
		0.998435,
		0.999554,
		0.999934,
		0.999999,
		1,
		0.999957,
		0.999668,
		0.998793,
		0.996922,
		0.993633,
		0.988531,
		0.981267,
		0.971552,
		0.959167,
		0.943959,
		0.0079,
		0.02989,
		0.063557,
		0.106691,
		0.157267,
		0.213443,
		0.273547,
		0.336073,
		0.399673,
		0.463149,
		0.525446,
		0.585653,
		0.642991,
		0.696813,
		0.746597,
		0.791947,
		0.832587,
		0.868358,
		0.899218,
		0.925236,
		0.946591,
		0.963565,
		0.976536,
		0.98597,
		0.992403,
		0.996422,
		0.998635,
		0.999636,
		0.999954,
		1,
		0.999999,
		0.999936,
		0.999581,
		0.998572,
		0.99649,
		0.992908,
		0.987433,
		0.97972,
		0.969489,
		0.95653,
		0.940699,
		0.921922,
		0.007973,
		0.030156,
		0.064107,
		0.107582,
		0.158534,
		0.215097,
		0.27558,
		0.338462,
		0.402378,
		0.466121,
		0.528629,
		0.588982,
		0.646399,
		0.70023,
		0.749955,
		0.795181,
		0.835636,
		0.871169,
		0.901745,
		0.927446,
		0.948463,
		0.96509,
		0.977722,
		0.986838,
		0.992991,
		0.996778,
		0.998818,
		0.999707,
		0.999969,
		1,
		0.999997,
		0.999909,
		0.999477,
		0.998321,
		0.996013,
		0.992122,
		0.986257,
		0.978079,
		0.967315,
		0.953766,
		0.937299,
		0.91785,
		0.895415,
		0.008046,
		0.030427,
		0.064663,
		0.108485,
		0.159817,
		0.21677,
		0.277636,
		0.340874,
		0.405109,
		0.469119,
		0.531836,
		0.592333,
		0.649825,
		0.703661,
		0.753323,
		0.798418,
		0.838682,
		0.87397,
		0.904257,
		0.929636,
		0.950308,
		0.966586,
		0.978877,
		0.987677,
		0.993551,
		0.997111,
		0.998984,
		0.999767,
		0.99998,
		1,
		0.999994,
		0.999874,
		0.999355,
		0.998039,
		0.995488,
		0.991272,
		0.984999,
		0.976338,
		0.965026,
		0.950871,
		0.933754,
		0.913621,
		0.89048,
		0.864394,
		0.008121,
		0.0307,
		0.065227,
		0.109399,
		0.161115,
		0.218463,
		0.279714,
		0.343311,
		0.407865,
		0.472143,
		0.535067,
		0.595706,
		0.65327,
		0.707107,
		0.756699,
		0.801659,
		0.841725,
		0.876762,
		0.906753,
		0.931803,
		0.952128,
		0.968053,
		0.980001,
		0.988485,
		0.994084,
		0.997422,
		0.999134,
		0.999819,
		0.999988,
		1,
		0.999988,
		0.99983,
		0.999212,
		0.997721,
		0.994912,
		0.990352,
		0.983655,
		0.974494,
		0.962616,
		0.947842,
		0.930061,
		0.909232,
		0.885375,
		0.858565,
		0.828924,
		0.008197,
		0.030978,
		0.065798,
		0.110324,
		0.162428,
		0.220175,
		0.281814,
		0.345773,
		0.410647,
		0.475193,
		0.538323,
		0.599101,
		0.656733,
		0.710566,
		0.760084,
		0.804901,
		0.844764,
		0.879542,
		0.909232,
		0.933948,
		0.953921,
		0.969489,
		0.981094,
		0.989263,
		0.99459,
		0.997711,
		0.999268,
		0.999861,
		0.999993,
		1,
		0.99998,
		0.999775,
		0.999046,
		0.997366,
		0.994282,
		0.989361,
		0.98222,
		0.972543,
		0.960083,
		0.944673,
		0.926215,
		0.90468,
		0.880097,
		0.852555,
		0.822187,
		0.789171,
		0.008273,
		0.031259,
		0.066377,
		0.111261,
		0.163757,
		0.221906,
		0.283937,
		0.34826,
		0.413455,
		0.478268,
		0.541604,
		0.602518,
		0.660215,
		0.71404,
		0.763477,
		0.808146,
		0.847798,
		0.882312,
		0.911694,
		0.93607,
		0.955686,
		0.970896,
		0.982156,
		0.990011,
		0.995069,
		0.997978,
		0.999388,
		0.999897,
		0.999997,
		1,
		0.999968,
		0.999707,
		0.998854,
		0.99697,
		0.993593,
		0.988294,
		0.980692,
		0.97048,
		0.957423,
		0.941362,
		0.922214,
		0.899959,
		0.874642,
		0.846361,
		0.815262,
		0.781532,
		0.745396,
		0.008351,
		0.031545,
		0.066963,
		0.11221,
		0.165103,
		0.223658,
		0.286083,
		0.350772,
		0.416289,
		0.48137,
		0.54491,
		0.605958,
		0.663715,
		0.717527,
		0.766878,
		0.811392,
		0.850827,
		0.88507,
		0.914137,
		0.938168,
		0.957423,
		0.972271,
		0.983187,
		0.990728,
		0.995521,
		0.998225,
		0.999493,
		0.999925,
		0.999999,
		1,
		0.999951,
		0.999624,
		0.998635,
		0.996529,
		0.992844,
		0.987147,
		0.979066,
		0.968302,
		0.95463,
		0.937904,
		0.918052,
		0.895068,
		0.869007,
		0.839981,
		0.808146,
		0.773702,
		0.736881,
		0.69795,
		0.00843,
		0.031834,
		0.067557,
		0.113171,
		0.166465,
		0.225429,
		0.288253,
		0.353309,
		0.41915,
		0.484498,
		0.54824,
		0.60942,
		0.667234,
		0.721027,
		0.770286,
		0.814639,
		0.85385,
		0.887815,
		0.916561,
		0.940242,
		0.959131,
		0.973615,
		0.984185,
		0.991416,
		0.995947,
		0.998451,
		0.999586,
		0.999947,
		1,
		0.999999,
		0.999927,
		0.999525,
		0.998385,
		0.996042,
		0.992029,
		0.985916,
		0.977338,
		0.966004,
		0.951701,
		0.934295,
		0.913726,
		0.890001,
		0.863189,
		0.833411,
		0.800837,
		0.765677,
		0.728175,
		0.688606,
		0.647268,
		0.00851,
		0.032127,
		0.068159,
		0.114144,
		0.167843,
		0.227222,
		0.290446,
		0.355873,
		0.422038,
		0.487653,
		0.551596,
		0.612904,
		0.670771,
		0.72454,
		0.773702,
		0.817887,
		0.856866,
		0.890546,
		0.918966,
		0.94229,
		0.96081,
		0.974928,
		0.985151,
		0.992073,
		0.996348,
		0.998657,
		0.999666,
		0.999965,
		1,
		0.999997,
		0.999897,
		0.999407,
		0.998102,
		0.995504,
		0.991144,
		0.984597,
		0.975504,
		0.963582,
		0.948631,
		0.930531,
		0.909232,
		0.884756,
		0.857183,
		0.826649,
		0.793332,
		0.757456,
		0.719275,
		0.679074,
		0.63716,
		0.593863,
		0.008592,
		0.032424,
		0.068768,
		0.11513,
		0.169238,
		0.229035,
		0.292664,
		0.358463,
		0.424954,
		0.490835,
		0.554977,
		0.61641,
		0.674326,
		0.728066,
		0.777124,
		0.821134,
		0.859876,
		0.893264,
		0.92135,
		0.944313,
		0.962458,
		0.976208,
		0.986085,
		0.9927,
		0.996723,
		0.998845,
		0.999735,
		0.999977,
		1,
		0.999993,
		0.999857,
		0.999268,
		0.997782,
		0.994912,
		0.990188,
		0.983187,
		0.973558,
		0.961031,
		0.945417,
		0.926607,
		0.904566,
		0.879329,
		0.850988,
		0.819691,
		0.78563,
		0.749038,
		0.710181,
		0.669354,
		0.626874,
		0.583079,
		0.538323,
		0.008674,
		0.032726,
		0.069387,
		0.116129,
		0.170651,
		0.230869,
		0.294906,
		0.361079,
		0.427896,
		0.494044,
		0.558383,
		0.619939,
		0.677898,
		0.731605,
		0.780552,
		0.824381,
		0.862877,
		0.895967,
		0.923712,
		0.946308,
		0.964076,
		0.977455,
		0.986986,
		0.993298,
		0.997073,
		0.999014,
		0.999793,
		0.999986,
		1,
		0.999987,
		0.999806,
		0.999106,
		0.997422,
		0.994262,
		0.989154,
		0.98168,
		0.971498,
		0.958348,
		0.942053,
		0.922519,
		0.899725,
		0.873716,
		0.8446,
		0.812535,
		0.777728,
		0.740422,
		0.700893,
		0.659446,
		0.61641,
		0.572131,
		0.526971,
		0.481304,
		0.008758,
		0.033031,
		0.070013,
		0.11714,
		0.172081,
		0.232725,
		0.297172,
		0.363722,
		0.430867,
		0.497281,
		0.561815,
		0.62349,
		0.681489,
		0.735156,
		0.783986,
		0.827626,
		0.865871,
		0.898654,
		0.926053,
		0.948277,
		0.965663,
		0.978669,
		0.987855,
		0.993865,
		0.997398,
		0.999166,
		0.999842,
		0.999992,
		1,
		0.999977,
		0.999743,
		0.998917,
		0.99702,
		0.993551,
		0.988039,
		0.980072,
		0.969317,
		0.955526,
		0.938536,
		0.918264,
		0.894703,
		0.867913,
		0.838015,
		0.805179,
		0.769624,
		0.731605,
		0.691409,
		0.649351,
		0.60577,
		0.561021,
		0.515474,
		0.469512,
		0.423526,
		0.008842,
		0.033341,
		0.070648,
		0.118164,
		0.173528,
		0.234603,
		0.299464,
		0.366393,
		0.433865,
		0.500544,
		0.565272,
		0.627063,
		0.685097,
		0.738718,
		0.787425,
		0.83087,
		0.868855,
		0.901325,
		0.928371,
		0.950217,
		0.967218,
		0.97985,
		0.988691,
		0.994403,
		0.9977,
		0.999302,
		0.999882,
		0.999996,
		1,
		0.999963,
		0.999666,
		0.9987,
		0.996571,
		0.992774,
		0.986838,
		0.978359,
		0.967013,
		0.952563,
		0.93486,
		0.913836,
		0.889497,
		0.861918,
		0.83123,
		0.79762,
		0.761317,
		0.722587,
		0.681729,
		0.639069,
		0.594955,
		0.54975,
		0.503836,
		0.457602,
		0.411446,
		0.365774,
		0.008928,
		0.033655,
		0.071291,
		0.119202,
		0.174994,
		0.236502,
		0.301781,
		0.369091,
		0.436892,
		0.503836,
		0.568755,
		0.630658,
		0.688723,
		0.742293,
		0.790869,
		0.834112,
		0.871829,
		0.903979,
		0.930665,
		0.952128,
		0.968741,
		0.980996,
		0.989494,
		0.994912,
		0.997978,
		0.999422,
		0.999914,
		0.999998,
		1,
		0.999944,
		0.999572,
		0.998451,
		0.996073,
		0.991928,
		0.985548,
		0.976536,
		0.964579,
		0.949453,
		0.931022,
		0.909232,
		0.884104,
		0.855726,
		0.824244,
		0.789856,
		0.752804,
		0.713366,
		0.671853,
		0.628601,
		0.583966,
		0.538323,
		0.492061,
		0.445578,
		0.39928,
		0.353578,
		0.308888,
		0.009016,
		0.033974,
		0.071944,
		0.120254,
		0.176478,
		0.238425,
		0.304124,
		0.371816,
		0.439948,
		0.507156,
		0.572264,
		0.634275,
		0.692366,
		0.745878,
		0.794318,
		0.83735,
		0.874792,
		0.906615,
		0.932935,
		0.954009,
		0.97023,
		0.982109,
		0.990264,
		0.995392,
		0.998234,
		0.999527,
		0.99994,
		0.999999,
		0.999998,
		0.999917,
		0.999459,
		0.998167,
		0.995521,
		0.991008,
		0.984164,
		0.974599,
		0.962012,
		0.946192,
		0.927017,
		0.904447,
		0.878519,
		0.849334,
		0.817052,
		0.781884,
		0.744084,
		0.703943,
		0.661782,
		0.617947,
		0.572806,
		0.526742,
		0.480153,
		0.433445,
		0.387032,
		0.341334,
		0.29677,
		0.253765,
		0.009104,
		0.034297,
		0.072605,
		0.121319,
		0.177981,
		0.24037,
		0.306492,
		0.37457,
		0.443032,
		0.510503,
		0.575799,
		0.637915,
		0.696026,
		0.749475,
		0.797771,
		0.840584,
		0.877744,
		0.909232,
		0.935179,
		0.955861,
		0.971686,
		0.983187,
		0.991001,
		0.995843,
		0.998468,
		0.999618,
		0.999959,
		1,
		0.999996,
		0.999882,
		0.999324,
		0.997845,
		0.994912,
		0.990011,
		0.982681,
		0.972543,
		0.959306,
		0.942774,
		0.92284,
		0.899478,
		0.872739,
		0.842738,
		0.809651,
		0.773702,
		0.735156,
		0.694316,
		0.651515,
		0.607109,
		0.561477,
		0.515011,
		0.468117,
		0.42121,
		0.374712,
		0.32905,
		0.28465,
		0.241943,
		0.201355,
		0.009194,
		0.034625,
		0.073275,
		0.122398,
		0.179502,
		0.242338,
		0.308888,
		0.377353,
		0.446146,
		0.51388,
		0.579359,
		0.641577,
		0.699703,
		0.753082,
		0.801227,
		0.843815,
		0.880684,
		0.91183,
		0.937398,
		0.957681,
		0.973107,
		0.98423,
		0.991705,
		0.996266,
		0.998681,
		0.999696,
		0.999974,
		1,
		0.999992,
		0.999836,
		0.999166,
		0.997481,
		0.994242,
		0.988931,
		0.981094,
		0.970362,
		0.956457,
		0.939196,
		0.918486,
		0.894319,
		0.866759,
		0.835936,
		0.80204,
		0.765307,
		0.726017,
		0.684484,
		0.641052,
		0.596089,
		0.549982,
		0.503133,
		0.455958,
		0.40888,
		0.362328,
		0.316735,
		0.272538,
		0.230171,
		0.190067,
		0.152659,
		0.009285,
		0.034957,
		0.073955,
		0.123492,
		0.181043,
		0.24433,
		0.31131,
		0.380165,
		0.449289,
		0.517284,
		0.582946,
		0.645261,
		0.703397,
		0.756699,
		0.804685,
		0.84704,
		0.883611,
		0.914407,
		0.93959,
		0.959469,
		0.974494,
		0.985237,
		0.992377,
		0.996662,
		0.998873,
		0.999763,
		0.999984,
		1,
		0.999985,
		0.999779,
		0.998981,
		0.997073,
		0.993506,
		0.987764,
		0.9794,
		0.968053,
		0.953459,
		0.935451,
		0.913952,
		0.888967,
		0.860575,
		0.828924,
		0.794215,
		0.756699,
		0.716668,
		0.674449,
		0.630396,
		0.584888,
		0.538323,
		0.491113,
		0.443682,
		0.39646,
		0.349886,
		0.304401,
		0.260446,
		0.218463,
		0.178891,
		0.142168,
		0.108724,
		0.009378,
		0.035295,
		0.074644,
		0.1246,
		0.182603,
		0.246346,
		0.31376,
		0.383005,
		0.452462,
		0.520718,
		0.586559,
		0.648967,
		0.707107,
		0.760326,
		0.808146,
		0.850259,
		0.886525,
		0.916963,
		0.941754,
		0.961225,
		0.975845,
		0.98621,
		0.993016,
		0.99703,
		0.999046,
		0.999819,
		0.999991,
		1,
		0.999974,
		0.999707,
		0.998766,
		0.996615,
		0.9927,
		0.986506,
		0.977591,
		0.965609,
		0.950308,
		0.931537,
		0.909232,
		0.883417,
		0.854185,
		0.821699,
		0.786174,
		0.747875,
		0.707107,
		0.664208,
		0.619546,
		0.573509,
		0.526505,
		0.478955,
		0.431293,
		0.383959,
		0.337397,
		0.292057,
		0.248386,
		0.206833,
		0.167843,
		0.131859,
		0.099316,
		0.070648,
		0.009472,
		0.035637,
		0.075343,
		0.125723,
		0.184184,
		0.248386,
		0.316237,
		0.385876,
		0.455665,
		0.52418,
		0.590198,
		0.652694,
		0.710833,
		0.763962,
		0.811609,
		0.853472,
		0.889423,
		0.919497,
		0.943889,
		0.962947,
		0.977161,
		0.987147,
		0.993622,
		0.997372,
		0.9992,
		0.999864,
		0.999996,
		1,
		0.999958,
		0.999618,
		0.998518,
		0.996106,
		0.991821,
		0.985151,
		0.975665,
		0.963026,
		0.946999,
		0.927446,
		0.904322,
		0.877665,
		0.847584,
		0.814257,
		0.777914,
		0.738833,
		0.697333,
		0.653763,
		0.608503,
		0.561953,
		0.514529,
		0.466665,
		0.4188,
		0.371384,
		0.32487,
		0.279714,
		0.236371,
		0.195295,
		0.156939,
		0.121749,
		0.090168,
		0.062632,
		0.039569,
		0.009567,
		0.035984,
		0.076051,
		0.126861,
		0.185784,
		0.25045,
		0.318742,
		0.388776,
		0.458899,
		0.527672,
		0.593863,
		0.656444,
		0.714575,
		0.767608,
		0.815072,
		0.856678,
		0.892307,
		0.922008,
		0.945995,
		0.964635,
		0.97844,
		0.988048,
		0.994196,
		0.997688,
		0.999337,
		0.999901,
		0.999998,
		0.999999,
		0.999935,
		0.99951,
		0.998234,
		0.995539,
		0.990862,
		0.983695,
		0.973615,
		0.960299,
		0.943527,
		0.923176,
		0.899218,
		0.871707,
		0.840769,
		0.806596,
		0.769433,
		0.729573,
		0.687346,
		0.643115,
		0.59727,
		0.550223,
		0.502402,
		0.454246,
		0.406208,
		0.358744,
		0.312315,
		0.267385,
		0.224414,
		0.183866,
		0.146196,
		0.111859,
		0.081301,
		0.054963,
		0.033279,
		0.016673,
		0.009664,
		0.036336,
		0.07677,
		0.128015,
		0.187405,
		0.25254,
		0.321275,
		0.391707,
		0.462163,
		0.531193,
		0.597555,
		0.660215,
		0.718333,
		0.771261,
		0.818536,
		0.859876,
		0.895174,
		0.924495,
		0.948071,
		0.966289,
		0.979683,
		0.988913,
		0.994738,
		0.997978,
		0.999457,
		0.999931,
		0.999999,
		0.999998,
		0.999905,
		0.999381,
		0.99791,
		0.994912,
		0.989821,
		0.982133,
		0.971437,
		0.957423,
		0.939885,
		0.91872,
		0.893914,
		0.865538,
		0.833736,
		0.798713,
		0.76073,
		0.720092,
		0.677145,
		0.632263,
		0.585848,
		0.538323,
		0.490126,
		0.441707,
		0.393526,
		0.346048,
		0.299743,
		0.255081,
		0.212531,
		0.172561,
		0.135635,
		0.102209,
		0.072738,
		0.047667,
		0.027432,
		0.012464,
		0.003183,
		0.009762,
		0.036694,
		0.077498,
		0.129184,
		0.189047,
		0.254655,
		0.323837,
		0.394668,
		0.465459,
		0.534743,
		0.601273,
		0.664008,
		0.722107,
		0.774923,
		0.822,
		0.863065,
		0.898023,
		0.926957,
		0.950115,
		0.967907,
		0.980889,
		0.989742,
		0.995249,
		0.998244,
		0.999561,
		0.999953,
		1,
		0.999996,
		0.999864,
		0.999227,
		0.997543,
		0.99422,
		0.988691,
		0.980459,
		0.969125,
		0.954391,
		0.93607,
		0.914074,
		0.888407,
		0.859156,
		0.826481,
		0.790604,
		0.751801,
		0.71039,
		0.66673,
		0.621209,
		0.57424,
		0.526257,
		0.477707,
		0.429052,
		0.38076,
		0.333306,
		0.287165,
		0.242817,
		0.200737,
		0.161399,
		0.125272,
		0.092821,
		0.064503,
		0.040769,
		0.022059,
		0.008808,
		0.00144,
		0.009862,
		0.037057,
		0.078237,
		0.130369,
		0.19071,
		0.256796,
		0.326429,
		0.397661,
		0.468785,
		0.538323,
		0.605018,
		0.667822,
		0.725895,
		0.778592,
		0.825463,
		0.866244,
		0.900855,
		0.929393,
		0.952128,
		0.969489,
		0.982057,
		0.990536,
		0.995728,
		0.998486,
		0.999651,
		0.99997,
		1,
		0.999991,
		0.999812,
		0.999046,
		0.997128,
		0.993458,
		0.987468,
		0.978669,
		0.966673,
		0.951199,
		0.932076,
		0.909232,
		0.88269,
		0.852555,
		0.819002,
		0.782268,
		0.742645,
		0.700466,
		0.656102,
		0.609954,
		0.562448,
		0.514027,
		0.465151,
		0.416289,
		0.36792,
		0.320527,
		0.274594,
		0.230606,
		0.189047,
		0.150396,
		0.11513,
		0.083718,
		0.056622,
		0.034297,
		0.01719,
		0.005737,
		0.000366,
		0.009964,
		0.037425,
		0.078987,
		0.13157,
		0.192394,
		0.258963,
		0.32905,
		0.400685,
		0.472143,
		0.541933,
		0.608789,
		0.671658,
		0.729698,
		0.782268,
		0.828924,
		0.869413,
		0.903668,
		0.931803,
		0.954108,
		0.971035,
		0.983187,
		0.991293,
		0.996177,
		0.998706,
		0.999727,
		0.999982,
		1,
		0.999982,
		0.999746,
		0.998834,
		0.996662,
		0.992621,
		0.986146,
		0.976757,
		0.964076,
		0.947842,
		0.927897,
		0.904191,
		0.876762,
		0.845732,
		0.811294,
		0.773702,
		0.73326,
		0.690319,
		0.645261,
		0.5985,
		0.550474,
		0.501639,
		0.452462,
		0.403426,
		0.355015,
		0.307723,
		0.262041,
		0.218463,
		0.177478,
		0.139573,
		0.105229,
		0.074922,
		0.049119,
		0.02828,
		0.012854,
		0.003284,
		0.010067,
		0.037798,
		0.079747,
		0.132788,
		0.194101,
		0.261156,
		0.331701,
		0.403741,
		0.475533,
		0.545574,
		0.612586,
		0.675515,
		0.733515,
		0.785951,
		0.832383,
		0.872571,
		0.90646,
		0.934185,
		0.956054,
		0.972543,
		0.984278,
		0.992015,
		0.996595,
		0.998903,
		0.999791,
		0.99999,
		1,
		0.99997,
		0.999663,
		0.998588,
		0.99614,
		0.991705,
		0.984721,
		0.974716,
		0.961329,
		0.944313,
		0.923529,
		0.898944,
		0.870615,
		0.838682,
		0.803355,
		0.764902,
		0.723646,
		0.679948,
		0.634207,
		0.586849,
		0.538323,
		0.489096,
		0.439648,
		0.390469,
		0.342055,
		0.294906,
		0.249522,
		0.206404,
		0.166048,
		0.128949,
		0.095593,
		0.06646,
		0.042023,
		0.022747,
		0.009086,
		0.001486,
		0.010171,
		0.038178,
		0.080519,
		0.134022,
		0.19583,
		0.263377,
		0.334382,
		0.406828,
		0.478955,
		0.549244,
		0.61641,
		0.679392,
		0.737347,
		0.789639,
		0.835839,
		0.875716,
		0.909232,
		0.936538,
		0.957965,
		0.974012,
		0.985332,
		0.9927,
		0.996984,
		0.999079,
		0.999844,
		0.999995,
		1,
		0.999951,
		0.999561,
		0.998304,
		0.995558,
		0.990705,
		0.983187,
		0.972543,
		0.958426,
		0.940607,
		0.918966,
		0.893487,
		0.864247,
		0.831403,
		0.795181,
		0.755868,
		0.7138,
		0.669354,
		0.622942,
		0.575003,
		0.525998,
		0.476406,
		0.426716,
		0.377429,
		0.32905,
		0.282087,
		0.237049,
		0.194445,
		0.154777,
		0.118545,
		0.086243,
		0.058356,
		0.035363,
		0.017731,
		0.00592,
		0.000378,
		0.010277,
		0.038563,
		0.081301,
		0.135274,
		0.197581,
		0.265625,
		0.337094,
		0.409949,
		0.48241,
		0.552945,
		0.620261,
		0.683291,
		0.741192,
		0.793332,
		0.839291,
		0.878848,
		0.911982,
		0.938862,
		0.959841,
		0.975444,
		0.986346,
		0.99335,
		0.997343,
		0.999236,
		0.999886,
		0.999998,
		0.999999,
		0.999926,
		0.999437,
		0.997978,
		0.994912,
		0.989616,
		0.981538,
		0.97023,
		0.955361,
		0.936718,
		0.914202,
		0.887815,
		0.857652,
		0.823889,
		0.78677,
		0.746597,
		0.703721,
		0.658536,
		0.611467,
		0.562964,
		0.513503,
		0.463572,
		0.413672,
		0.364313,
		0.31601,
		0.26928,
		0.224639,
		0.182603,
		0.143683,
		0.108384,
		0.077206,
		0.050639,
		0.029167,
		0.013263,
		0.00339,
		0.010385,
		0.038954,
		0.082095,
		0.136543,
		0.199356,
		0.267901,
		0.339837,
		0.413102,
		0.485897,
		0.556677,
		0.624138,
		0.68721,
		0.74505,
		0.797031,
		0.842738,
		0.881966,
		0.914709,
		0.941155,
		0.961681,
		0.976835,
		0.987321,
		0.993965,
		0.997675,
		0.999373,
		0.99992,
		0.999999,
		0.999998,
		0.99989,
		0.999289,
		0.997607,
		0.994196,
		0.988431,
		0.979768,
		0.967772,
		0.952128,
		0.932642,
		0.909232,
		0.881923,
		0.850827,
		0.816138,
		0.778118,
		0.737086,
		0.69341,
		0.647496,
		0.599783,
		0.550737,
		0.500843,
		0.450601,
		0.400525,
		0.351133,
		0.302949,
		0.256499,
		0.212308,
		0.170898,
		0.132788,
		0.098489,
		0.068506,
		0.043337,
		0.023468,
		0.009378,
		0.001534,
		0.010495,
		0.03935,
		0.0829,
		0.137829,
		0.201154,
		0.270205,
		0.342613,
		0.416289,
		0.489418,
		0.560439,
		0.628041,
		0.69115,
		0.748921,
		0.800733,
		0.84618,
		0.88507,
		0.917412,
		0.943417,
		0.963484,
		0.978187,
		0.988257,
		0.994545,
		0.997978,
		0.999493,
		0.999946,
		1,
		0.999995,
		0.999844,
		0.999112,
		0.997186,
		0.993406,
		0.987147,
		0.977873,
		0.965164,
		0.948721,
		0.928371,
		0.904051,
		0.875806,
		0.843766,
		0.808146,
		0.769224,
		0.727336,
		0.682865,
		0.636233,
		0.587893,
		0.538323,
		0.488022,
		0.437501,
		0.387283,
		0.337898,
		0.289878,
		0.243759,
		0.200072,
		0.159349,
		0.122113,
		0.088884,
		0.060171,
		0.036479,
		0.018298,
		0.006111,
		0.00039,
		0.010606,
		0.039753,
		0.083718,
		0.139134,
		0.202976,
		0.272538,
		0.34542,
		0.41951,
		0.492972,
		0.564232,
		0.631971,
		0.695109,
		0.752804,
		0.804438,
		0.849616,
		0.888157,
		0.92009,
		0.945646,
		0.965249,
		0.979499,
		0.989154,
		0.99509,
		0.998255,
		0.999596,
		0.999965,
		1,
		0.999989,
		0.999783,
		0.998903,
		0.996711,
		0.992536,
		0.985756,
		0.975845,
		0.962398,
		0.945135,
		0.9239,
		0.898654,
		0.869459,
		0.836467,
		0.79991,
		0.760084,
		0.717343,
		0.672085,
		0.624748,
		0.575799,
		0.525728,
		0.475047,
		0.424278,
		0.373956,
		0.324619,
		0.276811,
		0.231074,
		0.18795,
		0.147975,
		0.111681,
		0.079594,
		0.05223,
		0.030097,
		0.013691,
		0.003501,
		0.010719,
		0.040162,
		0.084547,
		0.140457,
		0.204822,
		0.2749,
		0.34826,
		0.422765,
		0.496559,
		0.568056,
		0.635927,
		0.699089,
		0.756699,
		0.808146,
		0.853044,
		0.891227,
		0.922742,
		0.947842,
		0.966975,
		0.98077,
		0.990011,
		0.995601,
		0.998506,
		0.999684,
		0.999979,
		1,
		0.99998,
		0.999707,
		0.998659,
		0.996177,
		0.991581,
		0.984253,
		0.973679,
		0.959469,
		0.941362,
		0.919224,
		0.893035,
		0.862877,
		0.828924,
		0.791425,
		0.750696,
		0.707107,
		0.661072,
		0.613044,
		0.563503,
		0.512956,
		0.461923,
		0.410941,
		0.360554,
		0.31131,
		0.263763,
		0.218463,
		0.17596,
		0.136799,
		0.101517,
		0.070648,
		0.044712,
		0.024223,
		0.009684,
		0.001584,
		0.010834,
		0.040578,
		0.085389,
		0.141799,
		0.206693,
		0.277292,
		0.351133,
		0.426054,
		0.50018,
		0.571912,
		0.63991,
		0.703088,
		0.760605,
		0.811856,
		0.856465,
		0.894279,
		0.925367,
		0.950003,
		0.968661,
		0.981999,
		0.990828,
		0.996078,
		0.998733,
		0.999758,
		0.999988,
		1,
		0.999965,
		0.999611,
		0.998376,
		0.995579,
		0.990536,
		0.982633,
		0.971369,
		0.956371,
		0.937398,
		0.914337,
		0.887188,
		0.856057,
		0.821134,
		0.78269,
		0.741059,
		0.696627,
		0.649825,
		0.601121,
		0.55101,
		0.500011,
		0.448658,
		0.397498,
		0.347086,
		0.297982,
		0.250747,
		0.205942,
		0.164123,
		0.125842,
		0.091646,
		0.062072,
		0.037648,
		0.018893,
		0.006313,
		0.000403,
		0.010951,
		0.041,
		0.086243,
		0.14316,
		0.208589,
		0.279714,
		0.354039,
		0.429378,
		0.503836,
		0.575799,
		0.643919,
		0.707107,
		0.764523,
		0.815567,
		0.859876,
		0.897313,
		0.927963,
		0.952128,
		0.970307,
		0.983187,
		0.991607,
		0.996521,
		0.998935,
		0.999819,
		0.999994,
		1,
		0.999944,
		0.999493,
		0.998049,
		0.994912,
		0.989394,
		0.980889,
		0.968908,
		0.953097,
		0.933236,
		0.909232,
		0.881109,
		0.848992,
		0.813093,
		0.773702,
		0.73117,
		0.685901,
		0.638345,
		0.588982,
		0.538323,
		0.4869,
		0.435259,
		0.383959,
		0.333566,
		0.28465,
		0.237781,
		0.193529,
		0.152459,
		0.11513,
		0.082095,
		0.053897,
		0.031071,
		0.014141,
		0.003617,
		0.011069,
		0.041428,
		0.08711,
		0.14454,
		0.210511,
		0.282166,
		0.35698,
		0.432738,
		0.507526,
		0.579717,
		0.647954,
		0.711144,
		0.76845,
		0.819279,
		0.863277,
		0.900326,
		0.930531,
		0.954217,
		0.971912,
		0.984332,
		0.992346,
		0.996932,
		0.999115,
		0.999868,
		0.999998,
		0.999999,
		0.999914,
		0.99935,
		0.997675,
		0.994171,
		0.98815,
		0.979014,
		0.966289,
		0.94964,
		0.928869,
		0.903905,
		0.874792,
		0.841678,
		0.804797,
		0.764456,
		0.721027,
		0.674931,
		0.626633,
		0.57663,
		0.525446,
		0.473628,
		0.421733,
		0.370333,
		0.320005,
		0.271328,
		0.224882,
		0.181245,
		0.140991,
		0.104687,
		0.072891,
		0.046154,
		0.025016,
		0.010005,
		0.001638,
		0.01119,
		0.041863,
		0.08799,
		0.14594,
		0.212459,
		0.28465,
		0.359954,
		0.436133,
		0.511251,
		0.583667,
		0.652015,
		0.715201,
		0.772387,
		0.82299,
		0.866667,
		0.903317,
		0.933068,
		0.956268,
		0.973475,
		0.985435,
		0.993045,
		0.997312,
		0.999274,
		0.999907,
		0.999999,
		0.999998,
		0.999873,
		0.999178,
		0.997247,
		0.99335,
		0.986798,
		0.977002,
		0.963506,
		0.945995,
		0.924291,
		0.898348,
		0.868232,
		0.834112,
		0.796242,
		0.754952,
		0.710629,
		0.663715,
		0.61469,
		0.564067,
		0.512384,
		0.460201,
		0.40809,
		0.356632,
		0.306416,
		0.258031,
		0.212067,
		0.169111,
		0.129743,
		0.094539,
		0.064064,
		0.038875,
		0.019517,
		0.006524,
		0.000417,
		0.011312,
		0.042305,
		0.088884,
		0.14736,
		0.214433,
		0.287165,
		0.362964,
		0.439564,
		0.515011,
		0.587648,
		0.656102,
		0.719275,
		0.776333,
		0.826699,
		0.870045,
		0.906287,
		0.935573,
		0.95828,
		0.974996,
		0.986495,
		0.993706,
		0.99766,
		0.999412,
		0.999937,
		1,
		0.999994,
		0.999819,
		0.998974,
		0.996763,
		0.992444,
		0.985332,
		0.974848,
		0.960553,
		0.942154,
		0.919497,
		0.892557,
		0.861423,
		0.826287,
		0.787425,
		0.745185,
		0.699976,
		0.652255,
		0.602518,
		0.551297,
		0.499142,
		0.446628,
		0.394338,
		0.342867,
		0.292812,
		0.244776,
		0.199356,
		0.157147,
		0.118739,
		0.084714,
		0.055645,
		0.032094,
		0.014613,
		0.003739,
		0.011437,
		0.042755,
		0.089791,
		0.148801,
		0.216434,
		0.289713,
		0.366009,
		0.443032,
		0.518807,
		0.591661,
		0.660215,
		0.723368,
		0.780288,
		0.830407,
		0.873411,
		0.909232,
		0.938045,
		0.960253,
		0.976473,
		0.987511,
		0.994328,
		0.997978,
		0.999531,
		0.999959,
		1,
		0.999987,
		0.999749,
		0.998733,
		0.996216,
		0.991447,
		0.983744,
		0.972543,
		0.957423,
		0.938112,
		0.914479,
		0.886525,
		0.854361,
		0.818201,
		0.778343,
		0.735156,
		0.689066,
		0.640549,
		0.59012,
		0.538323,
		0.485726,
		0.432915,
		0.380488,
		0.32905,
		0.27921,
		0.23158,
		0.186768,
		0.145377,
		0.108006,
		0.075242,
		0.047667,
		0.025848,
		0.010342,
		0.001693,
		0.011564,
		0.043211,
		0.090711,
		0.150262,
		0.218463,
		0.292293,
		0.369091,
		0.446537,
		0.522638,
		0.595706,
		0.664354,
		0.727478,
		0.78425,
		0.834112,
		0.876762,
		0.912153,
		0.940484,
		0.962186,
		0.977906,
		0.988485,
		0.994912,
		0.998267,
		0.999632,
		0.999975,
		1,
		0.999976,
		0.99966,
		0.998451,
		0.995601,
		0.990352,
		0.982029,
		0.970081,
		0.954108,
		0.93386,
		0.909232,
		0.880246,
		0.84704,
		0.809849,
		0.768993,
		0.72486,
		0.677898,
		0.628601,
		0.577498,
		0.525152,
		0.472143,
		0.419073,
		0.366551,
		0.315194,
		0.265625,
		0.218463,
		0.174326,
		0.133826,
		0.09757,
		0.066153,
		0.040162,
		0.020173,
		0.006746,
		0.000431,
		0.011692,
		0.043674,
		0.091646,
		0.151745,
		0.220519,
		0.294906,
		0.372208,
		0.45008,
		0.526505,
		0.599783,
		0.668518,
		0.731605,
		0.78822,
		0.837812,
		0.880097,
		0.915048,
		0.942888,
		0.964076,
		0.979295,
		0.989415,
		0.995458,
		0.998528,
		0.999718,
		0.999986,
		1,
		0.999959,
		0.999548,
		0.998123,
		0.994912,
		0.989154,
		0.980179,
		0.967454,
		0.950601,
		0.929393,
		0.903749,
		0.873716,
		0.839455,
		0.801227,
		0.759371,
		0.714298,
		0.666474,
		0.61641,
		0.564656,
		0.511786,
		0.4584,
		0.405109,
		0.352538,
		0.301315,
		0.252073,
		0.205443,
		0.162051,
		0.122519,
		0.087461,
		0.057479,
		0.033168,
		0.015108,
		0.003868,
		0.011823,
		0.044146,
		0.092596,
		0.15325,
		0.222604,
		0.297552,
		0.375363,
		0.45366,
		0.530408,
		0.603892,
		0.672708,
		0.735749,
		0.792195,
		0.841508,
		0.883417,
		0.917916,
		0.945255,
		0.965925,
		0.980638,
		0.990301,
		0.995967,
		0.998762,
		0.999788,
		0.999993,
		1,
		0.999934,
		0.999412,
		0.997745,
		0.994144,
		0.987845,
		0.978187,
		0.964657,
		0.946896,
		0.924704,
		0.898023,
		0.866927,
		0.831602,
		0.79233,
		0.749475,
		0.703467,
		0.654791,
		0.603979,
		0.551596,
		0.498233,
		0.444504,
		0.391034,
		0.338462,
		0.287427,
		0.238574,
		0.192542,
		0.149968,
		0.111483,
		0.077708,
		0.049255,
		0.026722,
		0.010696,
		0.001752,
		0.011957,
		0.044624,
		0.09356,
		0.154777,
		0.224718,
		0.300233,
		0.378555,
		0.457278,
		0.534347,
		0.608032,
		0.676922,
		0.739908,
		0.796177,
		0.845197,
		0.886718,
		0.920755,
		0.947585,
		0.967729,
		0.981935,
		0.991144,
		0.996439,
		0.99897,
		0.999846,
		0.999997,
		0.999999,
		0.999899,
		0.999246,
		0.997312,
		0.993291,
		0.986419,
		0.976047,
		0.961681,
		0.942986,
		0.919785,
		0.892049,
		0.859876,
		0.823476,
		0.783156,
		0.739302,
		0.692366,
		0.642853,
		0.591311,
		0.538323,
		0.484498,
		0.430464,
		0.37686,
		0.324337,
		0.273547,
		0.225145,
		0.179781,
		0.138102,
		0.100747,
		0.068345,
		0.041515,
		0.020862,
		0.006979,
		0.000447,
		0.012092,
		0.045111,
		0.094539,
		0.156326,
		0.226862,
		0.302949,
		0.381784,
		0.460936,
		0.538323,
		0.612205,
		0.681162,
		0.744084,
		0.800163,
		0.84888,
		0.890001,
		0.923565,
		0.949876,
		0.969489,
		0.983187,
		0.991944,
		0.996876,
		0.999153,
		0.999891,
		0.999999,
		0.999997,
		0.999852,
		0.999046,
		0.996818,
		0.992346,
		0.984868,
		0.973751,
		0.958519,
		0.938862,
		0.91463,
		0.88582,
		0.852555,
		0.815072,
		0.773702,
		0.728851,
		0.680995,
		0.630658,
		0.578407,
		0.524843,
		0.470589,
		0.416289,
		0.362598,
		0.310176,
		0.259691,
		0.211807,
		0.167184,
		0.12648,
		0.090341,
		0.059405,
		0.034297,
		0.01563,
		0.004003,
		0.01223,
		0.045605,
		0.095534,
		0.157899,
		0.229035,
		0.3057,
		0.385053,
		0.464632,
		0.542336,
		0.61641,
		0.685426,
		0.748275,
		0.804153,
		0.852555,
		0.893264,
		0.926344,
		0.952128,
		0.971204,
		0.984391,
		0.9927,
		0.997277,
		0.999313,
		0.999926,
		1,
		0.999993,
		0.999788,
		0.998808,
		0.996257,
		0.991302,
		0.983187,
		0.971292,
		0.955164,
		0.934518,
		0.909232,
		0.879329,
		0.84496,
		0.806386,
		0.763962,
		0.71812,
		0.669354,
		0.618209,
		0.565272,
		0.51116,
		0.456513,
		0.40199,
		0.34826,
		0.295997,
		0.245878,
		0.198581,
		0.154777,
		0.11513,
		0.080297,
		0.050923,
		0.027641,
		0.011069,
		0.001814,
		0.01237,
		0.046108,
		0.096544,
		0.159495,
		0.231239,
		0.308487,
		0.38836,
		0.468368,
		0.546387,
		0.620647,
		0.689715,
		0.75248,
		0.808146,
		0.85622,
		0.896506,
		0.92909,
		0.954338,
		0.972873,
		0.985548,
		0.993413,
		0.997644,
		0.999452,
		0.999952,
		1,
		0.999985,
		0.999707,
		0.998528,
		0.995624,
		0.990153,
		0.981366,
		0.968661,
		0.951607,
		0.929947,
		0.903584,
		0.872571,
		0.837084,
		0.797413,
		0.753936,
		0.707107,
		0.657442,
		0.605508,
		0.551909,
		0.497281,
		0.442279,
		0.387578,
		0.333861,
		0.281814,
		0.232129,
		0.185491,
		0.142584,
		0.104082,
		0.070648,
		0.042936,
		0.021587,
		0.007225,
		0.000462,
		0.012512,
		0.046619,
		0.09757,
		0.161115,
		0.233473,
		0.31131,
		0.391707,
		0.472143,
		0.550474,
		0.624916,
		0.694027,
		0.756699,
		0.812141,
		0.859876,
		0.899725,
		0.931803,
		0.956507,
		0.974494,
		0.986658,
		0.994084,
		0.997978,
		0.99957,
		0.999971,
		1,
		0.999972,
		0.999603,
		0.9982,
		0.994912,
		0.988892,
		0.9794,
		0.965851,
		0.947842,
		0.92514,
		0.897679,
		0.865538,
		0.828924,
		0.78815,
		0.743619,
		0.69581,
		0.645261,
		0.592557,
		0.538323,
		0.483212,
		0.427896,
		0.373065,
		0.319414,
		0.267647,
		0.218463,
		0.172561,
		0.130634,
		0.093366,
		0.061429,
		0.035484,
		0.016179,
		0.004145,
		0.012657,
		0.047138,
		0.098612,
		0.162759,
		0.23574,
		0.31417,
		0.395094,
		0.475959,
		0.5546,
		0.629217,
		0.698364,
		0.760931,
		0.816138,
		0.863519,
		0.90292,
		0.934481,
		0.958631,
		0.976067,
		0.98772,
		0.994712,
		0.99828,
		0.999669,
		0.999984,
		1,
		0.999952,
		0.999473,
		0.997819,
		0.994115,
		0.987511,
		0.977279,
		0.962855,
		0.943859,
		0.92009,
		0.89151,
		0.858226,
		0.820474,
		0.778592,
		0.73301,
		0.68423,
		0.63281,
		0.579359,
		0.524519,
		0.468961,
		0.413374,
		0.358463,
		0.304937,
		0.253513,
		0.204903,
		0.159817,
		0.118957,
		0.083016,
		0.052677,
		0.028608,
		0.011462,
		0.001879,
		0.012805,
		0.047667,
		0.099671,
		0.164428,
		0.238038,
		0.317068,
		0.398522,
		0.479816,
		0.558763,
		0.63355,
		0.702724,
		0.765177,
		0.820135,
		0.867151,
		0.906089,
		0.937122,
		0.960712,
		0.977591,
		0.988734,
		0.995298,
		0.998551,
		0.999752,
		0.999992,
		1,
		0.999923,
		0.999313,
		0.99738,
		0.993226,
		0.986004,
		0.974996,
		0.959663,
		0.939652,
		0.91479,
		0.88507,
		0.850628,
		0.811728,
		0.768736,
		0.722107,
		0.672365,
		0.620093,
		0.565918,
		0.510503,
		0.454537,
		0.398725,
		0.343786,
		0.290446,
		0.239433,
		0.191473,
		0.147285,
		0.107582,
		0.073068,
		0.044432,
		0.02235,
		0.007484,
		0.000479,
		0.012955,
		0.048204,
		0.100747,
		0.166123,
		0.24037,
		0.320005,
		0.40199,
		0.483714,
		0.562964,
		0.637915,
		0.707107,
		0.769433,
		0.824131,
		0.870768,
		0.909232,
		0.939726,
		0.962746,
		0.979066,
		0.9897,
		0.995843,
		0.998793,
		0.999819,
		0.999997,
		0.999999,
		0.999882,
		0.999119,
		0.996876,
		0.992239,
		0.984361,
		0.972543,
		0.956268,
		0.935212,
		0.909232,
		0.878352,
		0.842738,
		0.802682,
		0.758579,
		0.710906,
		0.660215,
		0.607109,
		0.552238,
		0.496282,
		0.439948,
		0.383959,
		0.32905,
		0.27596,
		0.225429,
		0.178197,
		0.134994,
		0.096544,
		0.063557,
		0.036734,
		0.016757,
		0.004296,
		0.013107,
		0.04875,
		0.101841,
		0.167843,
		0.242735,
		0.32298,
		0.405501,
		0.487653,
		0.567204,
		0.642312,
		0.711512,
		0.773702,
		0.828126,
		0.874369,
		0.912347,
		0.94229,
		0.964734,
		0.980491,
		0.990617,
		0.996348,
		0.999006,
		0.999872,
		0.999999,
		0.999997,
		0.999826,
		0.998886,
		0.996301,
		0.991144,
		0.982575,
		0.969911,
		0.952661,
		0.930531,
		0.903409,
		0.87135,
		0.834551,
		0.793332,
		0.748116,
		0.699408,
		0.647781,
		0.593863,
		0.538323,
		0.481862,
		0.425205,
		0.369091,
		0.31427,
		0.261496,
		0.211524,
		0.165103,
		0.122976,
		0.085875,
		0.054523,
		0.029626,
		0.011876,
		0.001948,
		0.013263,
		0.049305,
		0.102952,
		0.16959,
		0.245133,
		0.325995,
		0.409054,
		0.491635,
		0.571482,
		0.64674,
		0.71594,
		0.77798,
		0.832117,
		0.877955,
		0.915432,
		0.944814,
		0.966673,
		0.981865,
		0.991486,
		0.996812,
		0.999193,
		0.999913,
		1,
		0.999992,
		0.999752,
		0.998608,
		0.995649,
		0.989936,
		0.980638,
		0.967092,
		0.948834,
		0.925601,
		0.897313,
		0.864056,
		0.82606,
		0.783673,
		0.737347,
		0.687611,
		0.635065,
		0.580358,
		0.52418,
		0.467253,
		0.410318,
		0.354133,
		0.299464,
		0.247077,
		0.197741,
		0.152218,
		0.111261,
		0.075614,
		0.046007,
		0.023155,
		0.007758,
		0.000497,
		0.013421,
		0.049871,
		0.104082,
		0.171364,
		0.247567,
		0.32905,
		0.41265,
		0.495659,
		0.575799,
		0.651201,
		0.72039,
		0.782268,
		0.836105,
		0.881522,
		0.918486,
		0.947296,
		0.968563,
		0.983187,
		0.992308,
		0.997238,
		0.999355,
		0.999944,
		1,
		0.999982,
		0.999656,
		0.99828,
		0.994912,
		0.988607,
		0.978541,
		0.964076,
		0.944778,
		0.920413,
		0.890936,
		0.856465,
		0.81726,
		0.773702,
		0.726267,
		0.675515,
		0.622067,
		0.566596,
		0.509814,
		0.452462,
		0.395301,
		0.339103,
		0.28465,
		0.232725,
		0.184108,
		0.139573,
		0.099885,
		0.065798,
		0.038051,
		0.017367,
		0.004454,
		0.013582,
		0.050445,
		0.105229,
		0.173165,
		0.250035,
		0.332145,
		0.416289,
		0.499726,
		0.580154,
		0.655692,
		0.72486,
		0.786565,
		0.840087,
		0.88507,
		0.921508,
		0.949734,
		0.970403,
		0.984457,
		0.993081,
		0.997627,
		0.999493,
		0.999966,
		1,
		0.999967,
		0.999533,
		0.997897,
		0.994084,
		0.987147,
		0.976276,
		0.960857,
		0.940484,
		0.91496,
		0.88427,
		0.848568,
		0.808146,
		0.763413,
		0.714875,
		0.663118,
		0.608789,
		0.552583,
		0.495233,
		0.437501,
		0.380165,
		0.324015,
		0.269849,
		0.218463,
		0.170651,
		0.127199,
		0.088884,
		0.056468,
		0.0307,
		0.012313,
		0.002021,
		0.013746,
		0.05103,
		0.106396,
		0.174994,
		0.25254,
		0.335282,
		0.419973,
		0.503836,
		0.584549,
		0.660215,
		0.729352,
		0.790869,
		0.844063,
		0.888597,
		0.924495,
		0.952128,
		0.972191,
		0.985674,
		0.993807,
		0.997978,
		0.99961,
		0.999981,
		1,
		0.999944,
		0.999381,
		0.997451,
		0.993157,
		0.985548,
		0.973833,
		0.957423,
		0.935944,
		0.909232,
		0.877309,
		0.840361,
		0.798713,
		0.752804,
		0.70317,
		0.650423,
		0.595234,
		0.538323,
		0.480445,
		0.42238,
		0.364925,
		0.308888,
		0.255081,
		0.204316,
		0.1574,
		0.11513,
		0.078295,
		0.047667,
		0.024004,
		0.008046,
		0.000516,
		0.013913,
		0.051625,
		0.107582,
		0.176852,
		0.255081,
		0.338462,
		0.423701,
		0.50799,
		0.588982,
		0.664769,
		0.733863,
		0.795181,
		0.848031,
		0.892101,
		0.927446,
		0.954475,
		0.973927,
		0.986838,
		0.994485,
		0.998294,
		0.999707,
		0.99999,
		1,
		0.999909,
		0.999193,
		0.996937,
		0.992122,
		0.983803,
		0.971204,
		0.953766,
		0.931149,
		0.903223,
		0.870045,
		0.831837,
		0.788955,
		0.741872,
		0.69115,
		0.637429,
		0.581406,
		0.523824,
		0.465459,
		0.407111,
		0.349596,
		0.29374,
		0.24037,
		0.190312,
		0.144385,
		0.103402,
		0.068159,
		0.039439,
		0.018011,
		0.004622,
		0.014083,
		0.05223,
		0.108788,
		0.178739,
		0.257659,
		0.341684,
		0.427474,
		0.512188,
		0.593455,
		0.669354,
		0.738394,
		0.799498,
		0.85199,
		0.895582,
		0.930361,
		0.956775,
		0.975609,
		0.987949,
		0.995118,
		0.998577,
		0.999785,
		0.999996,
		0.999999,
		0.999861,
		0.998965,
		0.996348,
		0.990973,
		0.981901,
		0.968379,
		0.949876,
		0.926089,
		0.896922,
		0.862471,
		0.82299,
		0.778869,
		0.730613,
		0.678813,
		0.624138,
		0.567308,
		0.50909,
		0.450283,
		0.391707,
		0.334196,
		0.278592,
		0.22574,
		0.176478,
		0.131641,
		0.092052,
		0.058518,
		0.031834,
		0.012775,
		0.002098,
		0.014256,
		0.052846,
		0.110014,
		0.180656,
		0.260276,
		0.34495,
		0.431293,
		0.51643,
		0.597967,
		0.673969,
		0.742944,
		0.80382,
		0.855939,
		0.899037,
		0.933236,
		0.959025,
		0.977237,
		0.989007,
		0.995704,
		0.998827,
		0.999848,
		0.999999,
		0.999996,
		0.999794,
		0.99869,
		0.995675,
		0.9897,
		0.979834,
		0.965348,
		0.945745,
		0.920755,
		0.890323,
		0.854579,
		0.813813,
		0.76845,
		0.719025,
		0.666161,
		0.610552,
		0.552945,
		0.494131,
		0.43493,
		0.376183,
		0.318742,
		0.263466,
		0.211216,
		0.162846,
		0.119202,
		0.081119,
		0.049418,
		0.0249,
		0.008351,
		0.000536,
		0.014433,
		0.053473,
		0.111261,
		0.182603,
		0.262931,
		0.34826,
		0.435159,
		0.520718,
		0.602518,
		0.678615,
		0.747512,
		0.808146,
		0.859876,
		0.902465,
		0.93607,
		0.961225,
		0.97881,
		0.990011,
		0.996245,
		0.999046,
		0.999897,
		1,
		0.99999,
		0.999707,
		0.998364,
		0.994912,
		0.988294,
		0.977591,
		0.962103,
		0.941362,
		0.91514,
		0.883417,
		0.846361,
		0.804301,
		0.757694,
		0.707107,
		0.653193,
		0.596674,
		0.538323,
		0.478955,
		0.419412,
		0.360554,
		0.303253,
		0.248386,
		0.196828,
		0.149448,
		0.107106,
		0.070648,
		0.040905,
		0.018691,
		0.004799,
		0.014613,
		0.054111,
		0.112529,
		0.184582,
		0.265625,
		0.351615,
		0.439072,
		0.52505,
		0.607109,
		0.683291,
		0.752097,
		0.812474,
		0.863799,
		0.905864,
		0.938862,
		0.963372,
		0.980326,
		0.990961,
		0.996742,
		0.999236,
		0.999933,
		1,
		0.999979,
		0.999593,
		0.997978,
		0.99405,
		0.986746,
		0.975164,
		0.958631,
		0.936718,
		0.909232,
		0.876195,
		0.837812,
		0.794448,
		0.746597,
		0.694856,
		0.63991,
		0.582507,
		0.523449,
		0.463572,
		0.403741,
		0.344836,
		0.28775,
		0.233376,
		0.182603,
		0.136317,
		0.095391,
		0.060681,
		0.033031,
		0.013263,
		0.002179,
		0.014796,
		0.054761,
		0.113819,
		0.186592,
		0.268359,
		0.355015,
		0.443032,
		0.529428,
		0.61174,
		0.687996,
		0.756699,
		0.816804,
		0.867708,
		0.909232,
		0.94161,
		0.965467,
		0.981785,
		0.991857,
		0.997195,
		0.999399,
		0.999959,
		1,
		0.999961,
		0.999448,
		0.997527,
		0.993081,
		0.985047,
		0.972543,
		0.954925,
		0.931803,
		0.903024,
		0.868649,
		0.828924,
		0.78425,
		0.735156,
		0.682272,
		0.626313,
		0.568056,
		0.508329,
		0.447991,
		0.387931,
		0.32905,
		0.272257,
		0.218463,
		0.168575,
		0.123492,
		0.084099,
		0.051267,
		0.025848,
		0.008674,
		0.000557,
		0.014982,
		0.055422,
		0.11513,
		0.188634,
		0.271135,
		0.358463,
		0.447041,
		0.533853,
		0.61641,
		0.692731,
		0.761317,
		0.821134,
		0.8716,
		0.912568,
		0.944313,
		0.967506,
		0.983187,
		0.9927,
		0.997607,
		0.999536,
		0.999977,
		1,
		0.999933,
		0.999268,
		0.997003,
		0.991996,
		0.983187,
		0.969716,
		0.950973,
		0.926607,
		0.896506,
		0.860772,
		0.819691,
		0.773702,
		0.723368,
		0.669354,
		0.612405,
		0.553326,
		0.492972,
		0.432226,
		0.371999,
		0.313213,
		0.256796,
		0.203676,
		0.154777,
		0.11101,
		0.073275,
		0.042454,
		0.019411,
		0.004986,
		0.015172,
		0.056095,
		0.116464,
		0.19071,
		0.273952,
		0.361957,
		0.451099,
		0.538323,
		0.62112,
		0.697495,
		0.76595,
		0.825463,
		0.875474,
		0.91587,
		0.946968,
		0.969489,
		0.98453,
		0.99349,
		0.997978,
		0.999651,
		0.999989,
		1,
		0.999892,
		0.999046,
		0.996398,
		0.990785,
		0.981155,
		0.966673,
		0.946765,
		0.92112,
		0.889668,
		0.852555,
		0.810106,
		0.762798,
		0.71123,
		0.656102,
		0.598188,
		0.538323,
		0.477387,
		0.416289,
		0.355962,
		0.297347,
		0.241394,
		0.189047,
		0.141244,
		0.098913,
		0.062966,
		0.034297,
		0.013779,
		0.002265,
		0.015366,
		0.056781,
		0.117821,
		0.192819,
		0.276811,
		0.365499,
		0.455206,
		0.542841,
		0.625869,
		0.702287,
		0.770596,
		0.829789,
		0.879329,
		0.919137,
		0.949573,
		0.971415,
		0.985814,
		0.994227,
		0.99831,
		0.999744,
		0.999995,
		0.999998,
		0.999834,
		0.998776,
		0.995704,
		0.98944,
		0.978942,
		0.963404,
		0.94229,
		0.915332,
		0.882503,
		0.843991,
		0.800163,
		0.751535,
		0.698742,
		0.642517,
		0.583667,
		0.523054,
		0.461585,
		0.400195,
		0.339837,
		0.281477,
		0.226079,
		0.174607,
		0.128015,
		0.087245,
		0.053221,
		0.02685,
		0.009016,
		0.000579,
		0.015563,
		0.057479,
		0.119202,
		0.194962,
		0.279714,
		0.369091,
		0.459363,
		0.547405,
		0.630658,
		0.707107,
		0.775256,
		0.834112,
		0.883162,
		0.922365,
		0.952128,
		0.973282,
		0.987038,
		0.994912,
		0.998605,
		0.999819,
		0.999998,
		0.999995,
		0.999755,
		0.998451,
		0.994912,
		0.987949,
		0.976536,
		0.959898,
		0.937538,
		0.909232,
		0.875,
		0.835071,
		0.789856,
		0.739908,
		0.685901,
		0.628601,
		0.568845,
		0.507526,
		0.445578,
		0.383959,
		0.323646,
		0.265625,
		0.21088,
		0.160392,
		0.11513,
		0.076051,
		0.044093,
		0.020173,
		0.005185,
		0.015765,
		0.05819,
		0.120607,
		0.197141,
		0.282661,
		0.372731,
		0.463572,
		0.552017,
		0.635486,
		0.711954,
		0.779928,
		0.838428,
		0.886971,
		0.925553,
		0.95463,
		0.975089,
		0.988203,
		0.995546,
		0.998863,
		0.999876,
		1,
		0.999988,
		0.999651,
		0.998064,
		0.994013,
		0.986304,
		0.973927,
		0.956143,
		0.932498,
		0.902812,
		0.867151,
		0.82579,
		0.77918,
		0.727916,
		0.672708,
		0.614355,
		0.553728,
		0.491749,
		0.429378,
		0.367599,
		0.307411,
		0.249819,
		0.19583,
		0.146439,
		0.102633,
		0.065382,
		0.035637,
		0.014327,
		0.002356,
		0.01597,
		0.058915,
		0.122037,
		0.199356,
		0.285652,
		0.376422,
		0.467831,
		0.556677,
		0.640354,
		0.716828,
		0.784611,
		0.842738,
		0.890756,
		0.9287,
		0.957078,
		0.976835,
		0.989308,
		0.996128,
		0.999088,
		0.99992,
		1,
		0.999975,
		0.999515,
		0.997607,
		0.992999,
		0.984492,
		0.971103,
		0.952128,
		0.927158,
		0.89606,
		0.858946,
		0.816138,
		0.768129,
		0.715554,
		0.65916,
		0.599783,
		0.538323,
		0.475734,
		0.413,
		0.351133,
		0.291153,
		0.234088,
		0.180962,
		0.132788,
		0.090569,
		0.055289,
		0.027912,
		0.009378,
		0.000602,
		0.016179,
		0.059653,
		0.123492,
		0.201607,
		0.28869,
		0.380165,
		0.472143,
		0.561384,
		0.645261,
		0.721729,
		0.789303,
		0.84704,
		0.894513,
		0.931803,
		0.959469,
		0.978519,
		0.990352,
		0.996662,
		0.999281,
		0.999951,
		1,
		0.999953,
		0.999343,
		0.997073,
		0.991857,
		0.982503,
		0.968053,
		0.947842,
		0.921508,
		0.888967,
		0.850378,
		0.80611,
		0.756699,
		0.702821,
		0.645261,
		0.584888,
		0.522638,
		0.45949,
		0.39646,
		0.334582,
		0.2749,
		0.218463,
		0.166312,
		0.119481,
		0.078987,
		0.045828,
		0.02098,
		0.005396,
		0.016392,
		0.060404,
		0.124973,
		0.203896,
		0.291774,
		0.383959,
		0.476508,
		0.56614,
		0.650207,
		0.726654,
		0.794004,
		0.851331,
		0.898242,
		0.93486,
		0.961803,
		0.98014,
		0.991337,
		0.997147,
		0.999445,
		0.999973,
		1,
		0.99992,
		0.999128,
		0.996452,
		0.990579,
		0.980326,
		0.964764,
		0.943272,
		0.915536,
		0.881522,
		0.841438,
		0.795699,
		0.744885,
		0.689715,
		0.63101,
		0.569676,
		0.50668,
		0.443032,
		0.379775,
		0.317968,
		0.258679,
		0.202976,
		0.151921,
		0.106564,
		0.067939,
		0.037057,
		0.014907,
		0.002453,
		0.01661,
		0.06117,
		0.12648,
		0.206223,
		0.294906,
		0.387806,
		0.480925,
		0.570945,
		0.655192,
		0.731605,
		0.798713,
		0.85561,
		0.901939,
		0.93787,
		0.964076,
		0.981696,
		0.992261,
		0.997585,
		0.999581,
		0.999986,
		1,
		0.999871,
		0.998863,
		0.995735,
		0.989154,
		0.977948,
		0.961225,
		0.938407,
		0.909232,
		0.873716,
		0.832117,
		0.784898,
		0.732684,
		0.676234,
		0.61641,
		0.554152,
		0.49046,
		0.426374,
		0.362964,
		0.301315,
		0.242518,
		0.187663,
		0.137829,
		0.094085,
		0.057479,
		0.029038,
		0.009762,
		0.000627,
		0.016832,
		0.061951,
		0.128015,
		0.208589,
		0.298086,
		0.391707,
		0.485397,
		0.575799,
		0.660215,
		0.736579,
		0.803427,
		0.859876,
		0.905603,
		0.94083,
		0.966289,
		0.983187,
		0.993125,
		0.997978,
		0.999693,
		0.999994,
		0.999998,
		0.999801,
		0.998541,
		0.994912,
		0.987569,
		0.975358,
		0.957423,
		0.933236,
		0.902584,
		0.865538,
		0.822408,
		0.773702,
		0.720092,
		0.66238,
		0.601465,
		0.538323,
		0.473988,
		0.409531,
		0.346048,
		0.28465,
		0.226451,
		0.172561,
		0.124081,
		0.082095,
		0.047667,
		0.021837,
		0.00562,
		0.017058,
		0.062746,
		0.129577,
		0.210995,
		0.301315,
		0.395662,
		0.489923,
		0.580701,
		0.665277,
		0.741577,
		0.808146,
		0.864126,
		0.909232,
		0.943738,
		0.968439,
		0.984612,
		0.99393,
		0.998327,
		0.999782,
		0.999998,
		0.999994,
		0.999707,
		0.998153,
		0.993974,
		0.985814,
		0.972543,
		0.953345,
		0.927745,
		0.895582,
		0.85698,
		0.812301,
		0.762104,
		0.707107,
		0.648151,
		0.586177,
		0.522198,
		0.457278,
		0.392521,
		0.32905,
		0.268,
		0.210511,
		0.157713,
		0.110724,
		0.070648,
		0.038563,
		0.015523,
		0.002556,
		0.017289,
		0.063557,
		0.131168,
		0.213443,
		0.304595,
		0.399673,
		0.494505,
		0.585653,
		0.670377,
		0.746597,
		0.812868,
		0.868358,
		0.912824,
		0.946591,
		0.970524,
		0.98597,
		0.994676,
		0.998635,
		0.999851,
		1,
		0.999986,
		0.999581,
		0.997692,
		0.992908,
		0.983876,
		0.969489,
		0.94898,
		0.921922,
		0.888213,
		0.848031,
		0.801791,
		0.750101,
		0.693726,
		0.63355,
		0.570554,
		0.505785,
		0.440344,
		0.375363,
		0.311993,
		0.251397,
		0.194735,
		0.14316,
		0.097809,
		0.059802,
		0.030233,
		0.010171,
		0.000654,
		0.017525,
		0.064384,
		0.132788,
		0.215931,
		0.307926,
		0.403741,
		0.499142,
		0.590655,
		0.675515,
		0.751638,
		0.817592,
		0.872571,
		0.916375,
		0.949389,
		0.972543,
		0.987261,
		0.995364,
		0.998903,
		0.999903,
		1,
		0.99997,
		0.999419,
		0.997147,
		0.991705,
		0.981742,
		0.966185,
		0.944313,
		0.915755,
		0.880467,
		0.838682,
		0.790869,
		0.737688,
		0.679948,
		0.618579,
		0.5546,
		0.489096,
		0.423201,
		0.358077,
		0.294906,
		0.234873,
		0.179163,
		0.128949,
		0.085389,
		0.049618,
		0.022747,
		0.005858,
		0.017765,
		0.065227,
		0.134438,
		0.218463,
		0.31131,
		0.407865,
		0.503836,
		0.595706,
		0.68069,
		0.756699,
		0.822315,
		0.876762,
		0.919885,
		0.952128,
		0.974494,
		0.988485,
		0.995996,
		0.999134,
		0.999941,
		1,
		0.999944,
		0.999212,
		0.996509,
		0.990352,
		0.9794,
		0.962616,
		0.939331,
		0.909232,
		0.872333,
		0.828924,
		0.77953,
		0.72486,
		0.665773,
		0.603241,
		0.538323,
		0.472143,
		0.405867,
		0.340688,
		0.277816,
		0.218463,
		0.163836,
		0.11513,
		0.073521,
		0.040162,
		0.016179,
		0.002666,
		0.018011,
		0.066086,
		0.136118,
		0.221038,
		0.314747,
		0.412048,
		0.508587,
		0.600807,
		0.685901,
		0.761779,
		0.827036,
		0.880929,
		0.92335,
		0.954807,
		0.976376,
		0.989641,
		0.996571,
		0.99933,
		0.999967,
		1,
		0.999903,
		0.998953,
		0.995768,
		0.988837,
		0.976835,
		0.958769,
		0.934022,
		0.902341,
		0.863799,
		0.818748,
		0.767766,
		0.711615,
		0.651201,
		0.58754,
		0.521733,
		0.45494,
		0.38836,
		0.32322,
		0.260756,
		0.202206,
		0.148801,
		0.101756,
		0.062267,
		0.031504,
		0.010606,
		0.000682,
		0.018262,
		0.066963,
		0.137829,
		0.223658,
		0.318238,
		0.416289,
		0.513395,
		0.605958,
		0.69115,
		0.766878,
		0.831754,
		0.88507,
		0.926769,
		0.957423,
		0.978187,
		0.990728,
		0.997092,
		0.999493,
		0.999983,
		1,
		0.999844,
		0.998635,
		0.994912,
		0.987147,
		0.974035,
		0.95463,
		0.928371,
		0.895068,
		0.854856,
		0.808146,
		0.755573,
		0.69795,
		0.636233,
		0.571482,
		0.504839,
		0.437501,
		0.370703,
		0.3057,
		0.243759,
		0.186142,
		0.134105,
		0.088884,
		0.051692,
		0.023716,
		0.006111,
		0.018518,
		0.067857,
		0.139573,
		0.226323,
		0.321785,
		0.420591,
		0.518262,
		0.611159,
		0.696434,
		0.771993,
		0.836467,
		0.889182,
		0.930138,
		0.959974,
		0.979927,
		0.991748,
		0.997561,
		0.999627,
		0.999993,
		0.999998,
		0.99976,
		0.998248,
		0.99393,
		0.985268,
		0.970985,
		0.950184,
		0.922365,
		0.887401,
		0.845492,
		0.797109,
		0.742944,
		0.683863,
		0.620871,
		0.555074,
		0.487653,
		0.419845,
		0.352917,
		0.288158,
		0.226862,
		0.170317,
		0.119801,
		0.076573,
		0.041863,
		0.016877,
		0.002783,
		0.018779,
		0.068768,
		0.141349,
		0.229035,
		0.325389,
		0.424954,
		0.523188,
		0.61641,
		0.701753,
		0.777124,
		0.841172,
		0.893264,
		0.933455,
		0.962458,
		0.981594,
		0.9927,
		0.997978,
		0.999735,
		0.999998,
		0.999993,
		0.999645,
		0.997782,
		0.992809,
		0.983187,
		0.967671,
		0.945417,
		0.91599,
		0.879329,
		0.835698,
		0.78563,
		0.729875,
		0.669354,
		0.605119,
		0.538323,
		0.470189,
		0.40199,
		0.33503,
		0.270627,
		0.210104,
		0.154777,
		0.105945,
		0.064888,
		0.032856,
		0.011069,
		0.000713,
		0.019046,
		0.069699,
		0.14316,
		0.231794,
		0.32905,
		0.429378,
		0.528173,
		0.621711,
		0.707107,
		0.782268,
		0.845868,
		0.897313,
		0.936718,
		0.964874,
		0.983187,
		0.993585,
		0.998347,
		0.999819,
		1,
		0.999982,
		0.999493,
		0.997227,
		0.991538,
		0.980889,
		0.964076,
		0.940314,
		0.909232,
		0.870838,
		0.825463,
		0.773702,
		0.716363,
		0.654422,
		0.588982,
		0.52124,
		0.452462,
		0.383959,
		0.317068,
		0.253142,
		0.193529,
		0.139573,
		0.092596,
		0.053897,
		0.024747,
		0.006382,
		0.019319,
		0.070648,
		0.145004,
		0.234603,
		0.332769,
		0.433865,
		0.533218,
		0.627063,
		0.712494,
		0.787425,
		0.850552,
		0.901325,
		0.939924,
		0.967218,
		0.984704,
		0.994403,
		0.998669,
		0.999882,
		1,
		0.999963,
		0.999296,
		0.996571,
		0.990101,
		0.978359,
		0.960187,
		0.93486,
		0.902078,
		0.861918,
		0.814777,
		0.761317,
		0.702404,
		0.639069,
		0.572465,
		0.503836,
		0.434489,
		0.365774,
		0.299063,
		0.23574,
		0.177183,
		0.12476,
		0.079817,
		0.043674,
		0.017621,
		0.002907,
		0.019597,
		0.071616,
		0.146884,
		0.237461,
		0.336549,
		0.438416,
		0.538323,
		0.632464,
		0.717915,
		0.792593,
		0.855222,
		0.905299,
		0.943071,
		0.969489,
		0.986146,
		0.995156,
		0.998946,
		0.999928,
		1,
		0.999931,
		0.999046,
		0.995804,
		0.988485,
		0.975582,
		0.955987,
		0.929041,
		0.894513,
		0.852555,
		0.803632,
		0.74847,
		0.687996,
		0.623297,
		0.555576,
		0.486123,
		0.416289,
		0.347461,
		0.281048,
		0.218463,
		0.161115,
		0.110396,
		0.067676,
		0.034297,
		0.011564,
		0.000745,
		0.019882,
		0.072605,
		0.148801,
		0.24037,
		0.34039,
		0.443032,
		0.54349,
		0.637915,
		0.723368,
		0.797771,
		0.859876,
		0.909232,
		0.946156,
		0.971686,
		0.987511,
		0.995843,
		0.999183,
		0.999959,
		1,
		0.999882,
		0.998733,
		0.994912,
		0.986675,
		0.972543,
		0.951461,
		0.92284,
		0.886525,
		0.842738,
		0.792019,
		0.735156,
		0.673139,
		0.607109,
		0.538323,
		0.468117,
		0.397884,
		0.32905,
		0.263058,
		0.201355,
		0.145377,
		0.096544,
		0.056247,
		0.025848,
		0.006671,
		0.020173,
		0.073614,
		0.150754,
		0.243331,
		0.344293,
		0.447714,
		0.548718,
		0.643416,
		0.728851,
		0.802956,
		0.864511,
		0.913121,
		0.949176,
		0.973805,
		0.9888,
		0.996468,
		0.999381,
		0.99998,
		0.999999,
		0.999809,
		0.998347,
		0.993882,
		0.984657,
		0.969224,
		0.946591,
		0.916242,
		0.878098,
		0.832457,
		0.779928,
		0.721369,
		0.657831,
		0.590511,
		0.520718,
		0.449834,
		0.379296,
		0.31057,
		0.245133,
		0.184465,
		0.130029,
		0.08327,
		0.045605,
		0.018415,
		0.003041,
		0.02047,
		0.074644,
		0.152746,
		0.246346,
		0.34826,
		0.452462,
		0.554008,
		0.648967,
		0.734365,
		0.808146,
		0.869125,
		0.916963,
		0.952128,
		0.975845,
		0.990011,
		0.99703,
		0.999544,
		0.999991,
		0.999997,
		0.999707,
		0.997877,
		0.9927,
		0.982413,
		0.965609,
		0.941362,
		0.909232,
		0.869219,
		0.821699,
		0.767353,
		0.707107,
		0.642073,
		0.573509,
		0.502771,
		0.431293,
		0.360554,
		0.292057,
		0.227317,
		0.167843,
		0.11513,
		0.070648,
		0.035835,
		0.012092,
		0.00078,
		0.020774,
		0.075696,
		0.154777,
		0.249415,
		0.352291,
		0.457278,
		0.559361,
		0.654566,
		0.739908,
		0.81334,
		0.873716,
		0.920755,
		0.95501,
		0.977805,
		0.991144,
		0.997533,
		0.999675,
		0.999997,
		0.999991,
		0.999567,
		0.997312,
		0.991352,
		0.979927,
		0.961681,
		0.935757,
		0.901795,
		0.859876,
		0.810454,
		0.754286,
		0.692366,
		0.625869,
		0.556109,
		0.484498,
		0.412516,
		0.341684,
		0.273547,
		0.209654,
		0.151546,
		0.100747,
		0.058753,
		0.027023,
		0.006979,
		0.021084,
		0.07677,
		0.156848,
		0.25254,
		0.356389,
		0.462163,
		0.564777,
		0.660215,
		0.745479,
		0.818536,
		0.87828,
		0.924495,
		0.957819,
		0.979683,
		0.992201,
		0.997978,
		0.999777,
		0.999999,
		0.999978,
		0.999381,
		0.996638,
		0.989821,
		0.977183,
		0.957423,
		0.929759,
		0.893914,
		0.850054,
		0.798713,
		0.740722,
		0.677145,
		0.609221,
		0.538323,
		0.465916,
		0.393526,
		0.32272,
		0.255081,
		0.192195,
		0.135635,
		0.086952,
		0.047667,
		0.019264,
		0.003183,
		0.021402,
		0.077867,
		0.15896,
		0.255722,
		0.360554,
		0.467118,
		0.570256,
		0.665912,
		0.751077,
		0.823732,
		0.882814,
		0.928179,
		0.960553,
		0.981477,
		0.993181,
		0.998368,
		0.999854,
		1,
		0.999955,
		0.99914,
		0.995843,
		0.988091,
		0.974162,
		0.952816,
		0.92335,
		0.885575,
		0.839741,
		0.786465,
		0.726654,
		0.661443,
		0.592135,
		0.520162,
		0.447041,
		0.374349,
		0.303696,
		0.236704,
		0.174994,
		0.120172,
		0.073818,
		0.037478,
		0.012657,
		0.000817,
		0.021727,
		0.078987,
		0.161115,
		0.258963,
		0.364787,
		0.472143,
		0.575799,
		0.671658,
		0.756699,
		0.828924,
		0.887317,
		0.931803,
		0.963209,
		0.983187,
		0.994084,
		0.998706,
		0.999911,
		1,
		0.999916,
		0.998834,
		0.994912,
		0.986146,
		0.970847,
		0.947842,
		0.916514,
		0.876762,
		0.828924,
		0.773702,
		0.712079,
		0.645261,
		0.574618,
		0.501639,
		0.427896,
		0.355015,
		0.28465,
		0.218463,
		0.15811,
		0.105229,
		0.061429,
		0.02828,
		0.00731,
		0.022059,
		0.080132,
		0.163312,
		0.262263,
		0.369091,
		0.47724,
		0.581406,
		0.677451,
		0.762345,
		0.834112,
		0.891784,
		0.935365,
		0.965784,
		0.98481,
		0.994912,
		0.998994,
		0.99995,
		1,
		0.999854,
		0.998451,
		0.993829,
		0.983968,
		0.967218,
		0.942482,
		0.909232,
		0.867459,
		0.817592,
		0.760415,
		0.696992,
		0.628601,
		0.556677,
		0.482769,
		0.408505,
		0.335557,
		0.265625,
		0.200411,
		0.141606,
		0.09088,
		0.049871,
		0.020173,
		0.003336,
		0.022399,
		0.081301,
		0.165555,
		0.265625,
		0.373465,
		0.48241,
		0.587077,
		0.683291,
		0.768013,
		0.839291,
		0.896212,
		0.938862,
		0.968276,
		0.986346,
		0.995666,
		0.999236,
		0.999975,
		0.999999,
		0.999765,
		0.997978,
		0.992579,
		0.981538,
		0.963257,
		0.936718,
		0.901488,
		0.857652,
		0.805731,
		0.746597,
		0.681392,
		0.611467,
		0.538323,
		0.463572,
		0.388893,
		0.31601,
		0.246666,
		0.182603,
		0.12555,
		0.077206,
		0.039236,
		0.013263,
		0.000857,
		0.022747,
		0.082496,
		0.167843,
		0.269049,
		0.377913,
		0.487653,
		0.592813,
		0.689177,
		0.773702,
		0.84446,
		0.900598,
		0.94229,
		0.970681,
		0.987794,
		0.996348,
		0.999435,
		0.999989,
		0.999997,
		0.999638,
		0.997403,
		0.991144,
		0.978837,
		0.958943,
		0.930531,
		0.893264,
		0.847326,
		0.793332,
		0.732241,
		0.665277,
		0.593863,
		0.51957,
		0.444067,
		0.369091,
		0.296414,
		0.227824,
		0.165103,
		0.110014,
		0.064291,
		0.029626,
		0.007665,
		0.023103,
		0.083718,
		0.170178,
		0.272538,
		0.382435,
		0.492972,
		0.598615,
		0.695109,
		0.779408,
		0.849616,
		0.90494,
		0.945646,
		0.972999,
		0.989154,
		0.996959,
		0.999596,
		0.999996,
		0.999989,
		0.999465,
		0.996711,
		0.989507,
		0.975845,
		0.954256,
		0.9239,
		0.884543,
		0.836467,
		0.780384,
		0.717343,
		0.648648,
		0.575799,
		0.500431,
		0.424278,
		0.349131,
		0.276811,
		0.209152,
		0.147975,
		0.09508,
		0.05223,
		0.021147,
		0.003501,
		0.023468,
		0.084966,
		0.172561,
		0.276092,
		0.387032,
		0.498366,
		0.604481,
		0.701086,
		0.785132,
		0.854755,
		0.909232,
		0.948926,
		0.975226,
		0.990425,
		0.997501,
		0.999723,
		0.999999,
		0.999973,
		0.999236,
		0.995886,
		0.987648,
		0.972543,
		0.949176,
		0.916807,
		0.875306,
		0.82506,
		0.766878,
		0.701897,
		0.631507,
		0.557281,
		0.480925,
		0.404232,
		0.32905,
		0.25725,
		0.19071,
		0.131291,
		0.08083,
		0.041121,
		0.013913,
		0.000899,
		0.023841,
		0.086243,
		0.174994,
		0.279714,
		0.391707,
		0.503836,
		0.610413,
		0.707107,
		0.790869,
		0.859876,
		0.913472,
		0.952128,
		0.97736,
		0.991607,
		0.997978,
		0.999819,
		1,
		0.999944,
		0.998938,
		0.994912,
		0.985548,
		0.968908,
		0.943681,
		0.909232,
		0.865538,
		0.813093,
		0.752804,
		0.685901,
		0.613858,
		0.538323,
		0.461072,
		0.383959,
		0.308888,
		0.237781,
		0.172561,
		0.11513,
		0.067358,
		0.031071,
		0.008046,
		0.024222,
		0.087547,
		0.177477,
		0.283404,
		0.39646,
		0.509384,
		0.61641,
		0.71317,
		0.796619,
		0.864974,
		0.917657,
		0.955247,
		0.9794,
		0.9927,
		0.998392,
		0.999889,
		1,
		0.999895,
		0.99856,
		0.993771,
		0.983187,
		0.964919,
		0.937751,
		0.901155,
		0.855222,
		0.800552,
		0.738155,
		0.669354,
		0.595706,
		0.518938,
		0.440894,
		0.363491,
		0.28869,
		0.218463,
		0.154777,
		0.099574,
		0.054761,
		0.022194,
		0.003677,
		0.024612,
		0.088879,
		0.180011,
		0.287164,
		0.401293,
		0.515011,
		0.622473,
		0.719275,
		0.802379,
		0.870045,
		0.921781,
		0.95828,
		0.981342,
		0.993706,
		0.998747,
		0.999937,
		1,
		0.999819,
		0.998086,
		0.992444,
		0.980542,
		0.960553,
		0.931363,
		0.892557,
		0.84434,
		0.787425,
		0.722924,
		0.652255,
		0.577059,
		0.499142,
		0.420418,
		0.342867,
		0.268504,
		0.199356,
		0.137432,
		0.084714,
		0.043145,
		0.014613,
		0.000945,
		0.02501,
		0.090239,
		0.182595,
		0.290993,
		0.406206,
		0.520717,
		0.628601,
		0.72542,
		0.808146,
		0.875088,
		0.925841,
		0.961225,
		0.983187,
		0.994625,
		0.999046,
		0.999968,
		0.999999,
		0.999707,
		0.997501,
		0.990912,
		0.977591,
		0.955789,
		0.924495,
		0.883417,
		0.832877,
		0.773702,
		0.707107,
		0.634605,
		0.557927,
		0.478955,
		0.399673,
		0.322126,
		0.248386,
		0.180527,
		0.120607,
		0.070648,
		0.032625,
		0.008457,
		0.025414,
		0.091624,
		0.185227,
		0.294891,
		0.411199,
		0.526502,
		0.634793,
		0.731605,
		0.813918,
		0.880097,
		0.929834,
		0.964076,
		0.984931,
		0.995458,
		0.999293,
		0.999986,
		0.999996,
		0.999548,
		0.99679,
		0.989154,
		0.974313,
		0.950601,
		0.917126,
		0.873716,
		0.820817,
		0.759371,
		0.690698,
		0.61641,
		0.538323,
		0.4584,
		0.378694,
		0.301315,
		0.228392,
		0.162051,
		0.104393,
		0.057479,
		0.023321,
		0.003868,
		0.025825,
		0.093032,
		0.187904,
		0.298855,
		0.416269,
		0.532363,
		0.641049,
		0.737825,
		0.819691,
		0.88507,
		0.933754,
		0.966832,
		0.986574,
		0.996208,
		0.999493,
		0.999995,
		0.999986,
		0.999333,
		0.995933,
		0.987147,
		0.970681,
		0.944967,
		0.909232,
		0.863435,
		0.808146,
		0.744423,
		0.673697,
		0.597675,
		0.518262,
		0.437501,
		0.357518,
		0.280485,
		0.208589,
		0.144007,
		0.088884,
		0.045322,
		0.015366,
		0.000995,
		0.026239,
		0.094456,
		0.190619,
		0.302876,
		0.421411,
		0.538297,
		0.647364,
		0.74408,
		0.825462,
		0.890001,
		0.937598,
		0.969489,
		0.988116,
		0.996876,
		0.999651,
		0.999999,
		0.999966,
		0.999046,
		0.994912,
		0.984868,
		0.966673,
		0.938862,
		0.900792,
		0.852555,
		0.794849,
		0.728851,
		0.656102,
		0.578407,
		0.497762,
		0.416289,
		0.336187,
		0.259691,
		0.189047,
		0.12648,
		0.074184,
		0.034297,
		0.0089,
		0.026654,
		0.09589,
		0.19336,
		0.306942,
		0.426612,
		0.544295,
		0.653732,
		0.750362,
		0.831226,
		0.894886,
		0.941362,
		0.972044,
		0.989554,
		0.997465,
		0.999771,
		1,
		0.999929,
		0.998674,
		0.993706,
		0.982294,
		0.962262,
		0.932261,
		0.891784,
		0.841057,
		0.780913,
		0.712649,
		0.637915,
		0.558619,
		0.476846,
		0.394798,
		0.314747,
		0.238994,
		0.169842,
		0.109566,
		0.060404,
		0.024536,
		0.004073,
		0.027066,
		0.097323,
		0.196111,
		0.311037,
		0.431857,
		0.550341,
		0.66014,
		0.756663,
		0.836975,
		0.89972,
		0.94504,
		0.974494,
		0.99089,
		0.997978,
		0.999859,
		1,
		0.999867,
		0.9982,
		0.992293,
		0.9794,
		0.957423,
		0.92514,
		0.882183,
		0.828924,
		0.766326,
		0.69581,
		0.61914,
		0.538323,
		0.455537,
		0.373065,
		0.293251,
		0.218463,
		0.151058,
		0.093366,
		0.047667,
		0.016179,
		0.001049,
		0.027471,
		0.098742,
		0.198852,
		0.315133,
		0.437117,
		0.556411,
		0.666569,
		0.762966,
		0.842698,
		0.904491,
		0.948626,
		0.976834,
		0.992122,
		0.998419,
		0.99992,
		1,
		0.999771,
		0.997607,
		0.99065,
		0.976158,
		0.952128,
		0.917472,
		0.87197,
		0.816138,
		0.751077,
		0.678333,
		0.599783,
		0.517538,
		0.433865,
		0.351133,
		0.271757,
		0.19817,
		0.132788,
		0.07799,
		0.036101,
		0.009378,
		0.027862,
		0.100129,
		0.201554,
		0.319197,
		0.442357,
		0.56247,
		0.672987,
		0.769247,
		0.848375,
		0.909189,
		0.952111,
		0.979061,
		0.993251,
		0.998793,
		0.999959,
		0.999999,
		0.999629,
		0.996876,
		0.988753,
		0.972543,
		0.946351,
		0.909232,
		0.861121,
		0.802682,
		0.735156,
		0.660215,
		0.579853,
		0.496282,
		0.411865,
		0.32905,
		0.250328,
		0.178197,
		0.11513,
		0.063557,
		0.025848,
		0.004296,
		0.028234,
		0.101464,
		0.204182,
		0.323181,
		0.447525,
		0.56847,
		0.679351,
		0.77547,
		0.853978,
		0.91379,
		0.955481,
		0.981167,
		0.994276,
		0.999102,
		0.999982,
		0.999994,
		0.999429,
		0.995985,
		0.986574,
		0.968525,
		0.940063,
		0.900395,
		0.849616,
		0.788542,
		0.718555,
		0.641458,
		0.559361,
		0.474581,
		0.389573,
		0.306869,
		0.229035,
		0.158632,
		0.098193,
		0.050198,
		0.017058,
		0.001107,
		0.028577,
		0.102721,
		0.206694,
		0.327031,
		0.452559,
		0.574344,
		0.685602,
		0.781582,
		0.859466,
		0.918264,
		0.958714,
		0.983139,
		0.995193,
		0.999349,
		0.999993,
		0.999982,
		0.999156,
		0.994912,
		0.984089,
		0.964076,
		0.933236,
		0.890936,
		0.837433,
		0.773702,
		0.701268,
		0.622067,
		0.538323,
		0.452462,
		0.367032,
		0.28465,
		0.207954,
		0.139573,
		0.082095,
		0.038051,
		0.009896,
		0.028884,
		0.103873,
		0.209038,
		0.330677,
		0.457378,
		0.580011,
		0.691658,
		0.787513,
		0.864779,
		0.922566,
		0.961779,
		0.98496,
		0.995994,
		0.999536,
		0.999993,
		0.999955,
		0.998793,
		0.993633,
		0.981267,
		0.959167,
		0.925841,
		0.880827,
		0.824552,
		0.758149,
		0.683291,
		0.602046,
		0.516759,
		0.42996,
		0.344293,
		0.262459,
		0.187172,
		0.121124,
		0.066963,
		0.027267,
		0.004537,
		0.029144,
		0.104887,
		0.211158,
		0.334039,
		0.461886,
		0.585368,
		0.697422,
		0.793172,
		0.86984,
		0.926633,
		0.96463,
		0.986599,
		0.996665,
		0.999659,
		0.999982,
		0.999904,
		0.99832,
		0.992122,
		0.978079,
		0.953766,
		0.91785,
		0.870045,
		0.810953,
		0.741872,
		0.664622,
		0.581406,
		0.494689,
		0.407111,
		0.321409,
		0.24037,
		0.166781,
		0.103402,
		0.052935,
		0.018011,
		0.00117,
		0.029348,
		0.105728,
		0.212988,
		0.337025,
		0.465973,
		0.590294,
		0.702768,
		0.798444,
		0.874547,
		0.930382,
		0.967203,
		0.988014,
		0.997181,
		0.999705,
		0.999952,
		0.999812,
		0.997716,
		0.990351,
		0.974494,
		0.947842,
		0.909232,
		0.858565,
		0.796619,
		0.72486,
		0.645261,
		0.560159,
		0.472143,
		0.383959,
		0.298442,
		0.218463,
		0.146884,
		0.086531,
		0.040162,
		0.010458,
		0.029484,
		0.10636,
		0.21446,
		0.339533,
		0.46951,
		0.594644,
		0.707552,
		0.803188,
		0.878774,
		0.933706,
		0.969415,
		0.989142,
		0.997502,
		0.999655,
		0.999887,
		0.999659,
		0.996952,
		0.988288,
		0.970479,
		0.941362,
		0.899959,
		0.846361,
		0.781532,
		0.707107,
		0.625212,
		0.538323,
		0.449152,
		0.360554,
		0.27546,
		0.196828,
		0.127594,
		0.070648,
		0.028807,
		0.004799,
		0.029542,
		0.106744,
		0.215498,
		0.341453,
		0.472358,
		0.59826,
		0.711604,
		0.807238,
		0.882368,
		0.936471,
		0.971156,
		0.989904,
		0.997575,
		0.999471,
		0.999756,
		0.999412,
		0.995994,
		0.985898,
		0.965998,
		0.934294,
		0.890001,
		0.833411,
		0.765677,
		0.688606,
		0.604481,
		0.515919,
		0.425753,
		0.336951,
		0.25254,
		0.175563,
		0.109032,
		0.055902,
		0.019046,
		0.001239,
		0.029512,
		0.106843,
		0.216028,
		0.34267,
		0.474364,
		0.600964,
		0.714734,
		0.810403,
		0.885148,
		0.938519,
		0.972293,
		0.990193,
		0.997321,
		0.9991,
		0.999516,
		0.999026,
		0.994799,
		0.983139,
		0.961014,
		0.926602,
		0.879328,
		0.819691,
		0.749038,
		0.669354,
		0.583079,
		0.492972,
		0.40199,
		0.313213,
		0.229766,
		0.154777,
		0.091333,
		0.042454,
		0.011069,
		0.029384,
		0.10662,
		0.215973,
		0.343066,
		0.475373,
		0.602567,
		0.716734,
		0.812471,
		0.88691,
		0.939661,
		0.972665,
		0.98988,
		0.996642,
		0.998467,
		0.9991,
		0.998441,
		0.99331,
		0.979961,
		0.955481,
		0.918247,
		0.867908,
		0.805178,
		0.731605,
		0.649351,
		0.561021,
		0.469512,
		0.377913,
		0.289411,
		0.207232,
		0.134589,
		0.074644,
		0.030481,
		0.005084,
		0.029149,
		0.106041,
		0.215263,
		0.342523,
		0.475222,
		0.602874,
		0.717384,
		0.81321,
		0.887425,
		0.939685,
		0.972083,
		0.988808,
		0.995414,
		0.997473,
		0.998418,
		0.997575,
		0.991455,
		0.9763,
		0.949345,
		0.909189,
		0.85571,
		0.789852,
		0.713365,
		0.628601,
		0.538323,
		0.445578,
		0.353578,
		0.265625,
		0.185043,
		0.11513,
		0.059124,
		0.020173,
		0.001314,
		0.028801,
		0.105077,
		0.213831,
		0.340931,
		0.473757,
		0.601687,
		0.716456,
		0.812377,
		0.886448,
		0.938357,
		0.970337,
		0.986795,
		0.993485,
		0.99599,
		0.997356,
		0.996322,
		0.989142,
		0.972079,
		0.942546,
		0.899376,
		0.842698,
		0.773688,
		0.694312,
		0.607109,
		0.515011,
		0.42121,
		0.32905,
		0.241943,
		0.163312,
		0.096544,
		0.044948,
		0.011736,
		0.028334,
		0.103703,
		0.211619,
		0.338191,
		0.470829,
		0.598814,
		0.713725,
		0.809723,
		0.883723,
		0.935426,
		0.967195,
		0.983637,
		0.990679,
		0.993865,
		0.99577,
		0.994554,
		0.986257,
		0.967203,
		0.935005,
		0.888751,
		0.82883,
		0.756663,
		0.674436,
		0.584885,
		0.491113,
		0.39646,
		0.304401,
		0.218463,
		0.142168,
		0.078987,
		0.032305,
		0.005396,
		0.027745,
		0.101903,
		0.208583,
		0.334215,
		0.466305,
		0.594078,
		0.708972,
		0.805002,
		0.878988,
		0.930634,
		0.96241,
		0.979108,
		0.986797,
		0.990914,
		0.993485,
		0.992111,
		0.982663,
		0.961557,
		0.926633,
		0.877246,
		0.81406,
		0.73875,
		0.653732,
		0.561943,
		0.466662,
		0.371384,
		0.279714,
		0.195295,
		0.121749,
		0.062632,
		0.021402,
		0.001396,
		0.027034,
		0.099666,
		0.204691,
		0.328938,
		0.460076,
		0.58732,
		0.701997,
		0.797981,
		0.871991,
		0.923719,
		0.955728,
		0.97297,
		0.98162,
		0.986929,
		0.990301,
		0.988808,
		0.978195,
		0.955004,
		0.917318,
		0.864779,
		0.798332,
		0.719918,
		0.632191,
		0.538297,
		0.441699,
		0.346046,
		0.255081,
		0.172561,
		0.102209,
		0.047667,
		0.012464,
		0.026203,
		0.096994,
		0.199929,
		0.322316,
		0.45206,
		0.578414,
		0.692627,
		0.788449,
		0.862495,
		0.914431,
		0.946897,
		0.96498,
		0.974912,
		0.981678,
		0.985989,
		0.984431,
		0.972665,
		0.947385,
		0.906936,
		0.851254,
		0.781582,
		0.700132,
		0.609807,
		0.513969,
		0.416269,
		0.320521,
		0.230604,
		0.150396,
		0.083718,
		0.034297,
		0.005737,
		0.025256,
		0.093895,
		0.1943,
		0.314334,
		0.442204,
		0.567267,
		0.680727,
		0.77623,
		0.850289,
		0.902538,
		0.935677,
		0.954896,
		0.966433,
		0.974912,
		0.980301,
		0.978745,
		0.96586,
		0.938519,
		0.895339,
		0.836564,
		0.763735,
		0.679351,
		0.586569,
		0.488978,
		0.390425,
		0.294891,
		0.2064,
		0.128948,
		0.06646,
		0.022747,
		0.001486,
		0.024201,
		0.090387,
		0.187825,
		0.305002,
		0.430498,
		0.553828,
		0.666205,
		0.761188,
		0.835202,
		0.887842,
		0.921848,
		0.942488,
		0.955943,
		0.966371,
		0.97297,
		0.971493,
		0.95755,
		0.928207,
		0.882368,
		0.820585,
		0.744711,
		0.657532,
		0.56247,
		0.463351,
		0.364225,
		0.269249,
		0.182595,
		0.108382,
		0.050639,
		0.013263,
		0.023047,
		0.086499,
		0.180546,
		0.294361,
		0.416966,
		0.538097,
		0.649019,
		0.743241,
		0.817112,
		0.870185,
		0.905228,
		0.927549,
		0.943208,
		0.955795,
		0.963724,
		0.96241,
		0.947489,
		0.916238,
		0.867848,
		0.803188,
		0.724419,
		0.634625,
		0.537502,
		0.437117,
		0.337737,
		0.243699,
		0.159331,
		0.088879,
		0.036478,
		0.006111,
		0.021807,
		0.082265,
		0.172521,
		0.282483,
		0.401677,
		0.520121,
		0.629185,
		0.722367,
		0.795954,
		0.849464,
		0.885676,
		0.909906,
		0.928019,
		0.942932,
		0.95229,
		0.951225,
		0.935426,
		0.902391,
		0.8516,
		0.784235,
		0.702768,
		0.610585,
		0.51166,
		0.410314,
		0.311037,
		0.218359,
		0.136765,
		0.07064,
		0.024222,
		0.001584,
		0.020493,
		0.077731,
		0.163827,
		0.269468,
		0.384738,
		0.5,
		0.606781,
		0.698606,
		0.771729,
		0.825636,
		0.863107,
		0.889428,
		0.910193,
		0.927549,
		0.938409,
		0.937675,
		0.921114,
		0.886448,
		0.833445,
		0.76359,
		0.679669,
		0.585368,
		0.484947,
		0.382989,
		0.284216,
		0.19336,
		0.115075,
		0.053884,
		0.014139,
		0.019122,
		0.072946,
		0.154555,
		0.255441,
		0.366296,
		0.477883,
		0.581942,
		0.672067,
		0.744506,
		0.798725,
		0.837497,
		0.866039,
		0.889587,
		0.909445,
		0.921848,
		0.921517,
		0.904319,
		0.868202,
		0.81321,
		0.741127,
		0.655041,
		0.558942,
		0.457378,
		0.355202,
		0.257379,
		0.168853,
		0.094456,
		0.038856,
		0.006522,
		0.017708,
		0.067966,
		0.144809,
		0.240551,
		0.346533,
		0.453969,
		0.554862,
		0.642924,
		0.714427,
		0.768831,
		0.808892,
		0.839721,
		0.866109,
		0.888462,
		0.902408,
		0.902538,
		0.884835,
		0.847468,
		0.790745,
		0.716734,
		0.628819,
		0.531288,
		0.428982,
		0.327031,
		0.230652,
		0.14501,
		0.075127,
		0.025825,
		0.001693,
		0.01627,
		0.06285,
		0.134701,
		0.224966,
		0.325662,
		0.428498,
		0.525792,
		0.611418,
		0.681705,
		0.736123,
		0.777408,
		0.81052,
		0.839721,
		0.864494,
		0.879941,
		0.88057,
		0.862495,
		0.824096,
		0.765928,
		0.690328,
		0.600964,
		0.502414,
		0.399815,
		0.298578,
		0.204182,
		0.122028,
		0.057334,
		0.015085,
		0.014824,
		0.057659,
		0.124352,
		0.208867,
		0.303917,
		0.401745,
		0.495029,
		0.577846,
		0.646618,
		0.700845,
		0.743232,
		0.778552,
		0.810449,
		0.837497,
		0.854357,
		0.855502,
		0.837182,
		0.797981,
		0.73868,
		0.661863,
		0.571468,
		0.472358,
		0.369959,
		0.269972,
		0.178148,
		0.100129,
		0.041348,
		0.006962,
		0.013387,
		0.052454,
		0.113883,
		0.192443,
		0.281552,
		0.374015,
		0.462908,
		0.542556,
		0.609506,
		0.663304,
		0.706621,
		0.743999,
		0.778389,
		0.807497,
		0.825636,
		0.827286,
		0.808844,
		0.769076,
		0.708972,
		0.63134,
		0.540369,
		0.441194,
		0.339533,
		0.241376,
		0.152754,
		0.079568,
		0.027471,
		0.001807,
		0.011975,
		0.047295,
		0.103418,
		0.175889,
		0.25883,
		0.34563,
		0.429797,
		0.505937,
		0.570757,
		0.623865,
		0.667894,
		0.707107,
		0.743703,
		0.774593,
		0.793834,
		0.795954,
		0.777497,
		0.737402,
		0.676837,
		0.598814,
		0.507751,
		0.409045,
		0.308696,
		0.212988,
		0.128242,
		0.060626,
		0.016032,
		0.010604,
		0.042238,
		0.093074,
		0.159395,
		0.236018,
		0.316924,
		0.39608,
		0.468408,
		0.530799,
		0.582942,
		0.627422,
		0.668181,
		0.706621,
		0.738959,
		0.759087,
		0.761617,
		0.743241,
		0.703054,
		0.642378,
		0.564404,
		0.473757,
		0.37608,
		0.277651,
		0.185047,
		0.104887,
		0.043618,
		0.007387,
		0.009287,
		0.037339,
		0.082967,
		0.143149,
		0.213377,
		0.288229,
		0.36215,
		0.430401,
		0.490085,
		0.540983,
		0.585621,
		0.627576,
		0.667436,
		0.700845,
		0.721612,
		0.724468,
		0.706256,
		0.666205,
		0.605771,
		0.528295,
		0.438588,
		0.342523,
		0.246645,
		0.157829,
		0.082994,
		0.028884,
		0.001912,
		0.008036,
		0.032644,
		0.073201,
		0.127325,
		0.191158,
		0.259871,
		0.328396,
		0.392355,
		0.449081,
		0.498456,
		0.542935,
		0.585686,
		0.626498,
		0.660568,
		0.681705,
		0.684786,
		0.666804,
		0.627109,
		0.567267,
		0.49074,
		0.402506,
		0.308649,
		0.215973,
		0.131646,
		0.062899,
		0.016784,
		0.006863,
		0.028199,
		0.063871,
		0.112084,
		0.169596,
		0.232157,
		0.295193,
		0.354699,
		0.408251,
		0.455837,
		0.499827,
		0.542935,
		0.584203,
		0.618508,
		0.639732,
		0.642924,
		0.625228,
		0.586099,
		0.527188,
		0.45206,
		0.365832,
		0.274783,
		0.185967,
		0.106843,
		0.04496,
		0.007692,
		0.005777,
		0.02404,
		0.05506,
		0.097571,
		0.148905,
		0.205375,
		0.262895,
		0.317842,
		0.368043,
		0.413595,
		0.456759,
		0.499767,
		0.540983,
		0.575095,
		0.596121,
		0.599306,
		0.581942,
		0.543575,
		0.485925,
		0.412633,
		0.328938,
		0.241294,
		0.156993,
		0.083785,
		0.029542,
		0.001978,
		0.004784,
		0.020197,
		0.046839,
		0.08391,
		0.129275,
		0.179781,
		0.231825,
		0.282165,
		0.328883,
		0.372177,
		0.414184,
		0.456627,
		0.497293,
		0.530799,
		0.551352,
		0.554413,
		0.537419,
		0.5,
		0.443925,
		0.372893,
		0.29224,
		0.208583,
		0.12944,
		0.062848,
		0.017013,
		0.00389,
		0.016693,
		0.039266,
		0.071206,
		0.110866,
		0.1556,
		0.202269,
		0.248009,
		0.291157,
		0.332002,
		0.372533,
		0.413956,
		0.4536,
		0.486112,
		0.505937,
		0.508764,
		0.492178,
		0.455881,
		0.401677,
		0.333305,
		0.256183,
		0.177071,
		0.103703,
		0.044403,
		0.007716,
		0.003097,
		0.013546,
		0.032382,
		0.05954,
		0.09381,
		0.133017,
		0.174471,
		0.215674,
		0.255208,
		0.293444,
		0.332201,
		0.372177,
		0.410368,
		0.441538,
		0.460409,
		0.462908,
		0.446768,
		0.411755,
		0.359697,
		0.294361,
		0.221226,
		0.147183,
		0.080172,
		0.028801,
		0.001962,
		0.002407,
		0.010764,
		0.026217,
		0.048973,
		0.078208,
		0.112184,
		0.148632,
		0.185408,
		0.22133,
		0.256834,
		0.293546,
		0.331687,
		0.368043,
		0.397574,
		0.415302,
		0.417399,
		0.401745,
		0.368169,
		0.318512,
		0.256554,
		0.187825,
		0.119331,
		0.05921,
		0.01635,
		0.001819,
		0.008349,
		0.020783,
		0.039541,
		0.064131,
		0.09321,
		0.124906,
		0.15741,
		0.18976,
		0.222444,
		0.256874,
		0.292845,
		0.32705,
		0.354699,
		0.371134,
		0.37278,
		0.35766,
		0.325662,
		0.278635,
		0.220363,
		0.156413,
		0.093895,
		0.041135,
		0.007301,
		0.001329,
		0.006296,
		0.016081,
		0.031258,
		0.051615,
		0.076164,
		0.103399,
		0.131824,
		0.160683,
		0.190493,
		0.222444,
		0.255969,
		0.287772,
		0.313356,
		0.328396,
		0.329566,
		0.315035,
		0.284745,
		0.240551,
		0.186232,
		0.127383,
		0.071204,
		0.026203,
		0.001825,
		0.000933,
		0.004591,
		0.012095,
		0.024116,
		0.040669,
		0.061078,
		0.084171,
		0.108744,
		0.134224,
		0.161142,
		0.19046,
		0.22133,
		0.25055,
		0.273949,
		0.287533,
		0.288229,
		0.274347,
		0.245885,
		0.204699,
		0.154555,
		0.101068,
		0.051521,
		0.014591,
		0.000623,
		0.003214,
		0.008796,
		0.018083,
		0.031269,
		0.047945,
		0.067237,
		0.088209,
		0.110452,
		0.134495,
		0.161075,
		0.189147,
		0.215674,
		0.236823,
		0.248937,
		0.249185,
		0.236018,
		0.209488,
		0.171451,
		0.125657,
		0.077731,
		0.035026,
		0.006383,
		0.00039,
		0.002141,
		0.006142,
		0.01311,
		0.023365,
		0.036721,
		0.05257,
		0.070213,
		0.089387,
		0.110603,
		0.134389,
		0.159588,
		0.183376,
		0.202269,
		0.212934,
		0.212781,
		0.200397,
		0.175889,
		0.141108,
		0.099785,
		0.057546,
		0.021807,
		0.001561,
		0.000225,
		0.001339,
		0.00408,
		0.009125,
		0.016877,
		0.027329,
		0.040099,
		0.054701,
		0.070995,
		0.089464,
		0.110452,
		0.132767,
		0.153832,
		0.170509,
		0.179781,
		0.17929,
		0.167757,
		0.145338,
		0.113883,
		0.077097,
		0.040599,
		0.011854,
		0.000116,
		0.00077,
		0.002543,
		0.006041,
		0.011703,
		0.019659,
		0.029718,
		0.041577,
		0.055198,
		0.071031,
		0.089266,
		0.108744,
		0.127154,
		0.141698,
		0.149661,
		0.148905,
		0.138285,
		0.118003,
		0.089903,
		0.057659,
		0.026875,
		0.005056,
		5.1e-005,
		0.000396,
		0.00146,
		0.003753,
		0.007717,
		0.013573,
		0.021286,
		0.030705,
		0.041877,
		0.055211,
		0.070788,
		0.087528,
		0.103399,
		0.115924,
		0.122681,
		0.121741,
		0.112084,
		0.09396,
		0.069201,
		0.041445,
		0.01627,
		0.001204,
		1.7e-005,
		0.000173,
		0.000748,
		0.002146,
		0.004775,
		0.00891,
		0.014633,
		0.021916,
		0.030876,
		0.041877,
		0.054932,
		0.069081,
		0.082561,
		0.09321,
		0.098875,
		0.097831,
		0.089175,
		0.073201,
		0.051724,
		0.02834,
		0.008587,
		5.8e-005,
		0.000326,
		0.001096,
		0.002719,
		0.005488,
		0.009563,
		0.015015,
		0.022008,
		0.030868,
		0.041577,
		0.053314,
		0.064584,
		0.073513,
		0.078208,
		0.077134,
		0.0695,
		0.055636,
		0.037339,
		0.018148,
		0.00355,
		1.2e-005,
		0.000109,
		0.000475,
		0.001382,
		0.00311,
		0.005863,
		0.00978,
		0.015059,
		0.021994,
		0.030563,
		0.040099,
		0.049356,
		0.056733,
		0.06058,
		0.05954,
		0.052927,
		0.041104,
		0.025838,
		0.010604,
		0.000818,
		2.3e-005,
		0.000159,
		0.000596,
		0.001573,
		0.003307,
		0.005977,
		0.0098,
		0.015042,
		0.021703,
		0.029267,
		0.036721,
		0.042718,
		0.045834,
		0.044879,
		0.039266,
		0.029382,
		0.016957,
		0.005387,
		3.3e-005,
		0.000199,
		0.000676,
		0.001665,
		0.003361,
		0.005985,
		0.00978,
		0.014781,
		0.020615,
		0.026482,
		0.031269,
		0.033761,
		0.032928,
		0.028272,
		0.020197,
		0.010384,
		0.002137,
		4.1e-005,
		0.000224,
		0.000712,
		0.001688,
		0.003364,
		0.005965,
		0.009563,
		0.013913,
		0.018409,
		0.02215,
		0.024116,
		0.023426,
		0.019662,
		0.013239,
		0.005777,
		0.000471,
		4.6e-005,
		0.000235,
		0.000719,
		0.001688,
		0.003347,
		0.005799,
		0.00891,
		0.012245,
		0.015095,
		0.016622,
		0.016081,
		0.013129,
		0.008178,
		0.00278,
		4.8e-005,
		0.000237,
		0.000719,
		0.001676,
		0.003231,
		0.005342,
		0.007717,
		0.009823,
		0.010985,
		0.010587,
		0.008349,
		0.004677,
		0.001038,
		4.9e-005,
		0.000237,
		0.000712,
		0.001605,
		0.002939,
		0.004545,
		0.006041,
		0.006903,
		0.006631,
		0.005004,
		0.002407,
		0.000214,
		4.9e-005,
		0.000233,
		0.000676,
		0.00144,
		0.002452,
		0.003462,
		0.00408,
		0.003911,
		0.002784,
		0.001059,
		4.8e-005,
		0.000219,
		0.000596,
		0.001175,
		0.001811,
		0.002233,
		0.002141,
		0.001406,
		0.000356,
		4.4e-005,
		0.00019,
		0.000475,
		0.000838,
		0.001107,
		0.001066,
		0.000623,
		6.5e-005,
		3.8e-005,
		0.000147,
		0.000326,
		0.000481,
		0.00047,
		0.000228,
		2.8e-005,
		9.6e-005,
		0.000173,
		0.000175,
		6.1e-005,
		1.8e-005,
		4.7e-005,
		5.1e-005,
		1e-005,
		1e-005,
		5.1e-005,
		4.7e-005,
		1.8e-005,
		6.1e-005,
		0.000175,
		0.000173,
		9.6e-005,
		2.8e-005,
		0.000228,
		0.00047,
		0.000481,
		0.000326,
		0.000147,
		3.8e-005,
		6.5e-005,
		0.000623,
		0.001066,
		0.001107,
		0.000838,
		0.000475,
		0.00019,
		4.4e-005,
		0.000356,
		0.001406,
		0.002141,
		0.002233,
		0.001811,
		0.001175,
		0.000596,
		0.000219,
		4.8e-005,
		0.001059,
		0.002784,
		0.003911,
		0.00408,
		0.003462,
		0.002452,
		0.00144,
		0.000676,
		0.000233,
		4.9e-005,
		0.000214,
		0.002407,
		0.005004,
		0.006631,
		0.006903,
		0.006041,
		0.004545,
		0.002939,
		0.001605,
		0.000712,
		0.000237,
		4.9e-005,
		0.001038,
		0.004677,
		0.008349,
		0.010587,
		0.010985,
		0.009823,
		0.007717,
		0.005342,
		0.003231,
		0.001676,
		0.000719,
		0.000237,
		4.8e-005,
		0.00278,
		0.008178,
		0.013129,
		0.016081,
		0.016622,
		0.015095,
		0.012245,
		0.00891,
		0.005799,
		0.003347,
		0.001688,
		0.000719,
		0.000235,
		4.6e-005,
		0.000471,
		0.005777,
		0.013239,
		0.019662,
		0.023426,
		0.024116,
		0.02215,
		0.018409,
		0.013913,
		0.009563,
		0.005965,
		0.003364,
		0.001688,
		0.000712,
		0.000224,
		4.1e-005,
		0.002137,
		0.010384,
		0.020197,
		0.028272,
		0.032928,
		0.033761,
		0.031269,
		0.026482,
		0.020615,
		0.014781,
		0.00978,
		0.005985,
		0.003361,
		0.001665,
		0.000676,
		0.000199,
		3.3e-005,
		0.005387,
		0.016957,
		0.029382,
		0.039266,
		0.044879,
		0.045834,
		0.042718,
		0.036721,
		0.029267,
		0.021703,
		0.015042,
		0.0098,
		0.005977,
		0.003307,
		0.001573,
		0.000596,
		0.000159,
		2.3e-005,
		0.000818,
		0.010604,
		0.025838,
		0.041104,
		0.052927,
		0.05954,
		0.06058,
		0.056733,
		0.049356,
		0.040099,
		0.030563,
		0.021994,
		0.015059,
		0.00978,
		0.005863,
		0.00311,
		0.001382,
		0.000475,
		0.000109,
		1.2e-005,
		0.00355,
		0.018148,
		0.037339,
		0.055636,
		0.0695,
		0.077134,
		0.078208,
		0.073513,
		0.064584,
		0.053314,
		0.041577,
		0.030868,
		0.022008,
		0.015015,
		0.009563,
		0.005488,
		0.002719,
		0.001096,
		0.000326,
		5.8e-005,
		0.008587,
		0.02834,
		0.051724,
		0.073201,
		0.089175,
		0.097831,
		0.098875,
		0.09321,
		0.082561,
		0.069081,
		0.054932,
		0.041877,
		0.030876,
		0.021916,
		0.014633,
		0.00891,
		0.004775,
		0.002146,
		0.000748,
		0.000173,
		1.7e-005,
		0.001204,
		0.01627,
		0.041445,
		0.069201,
		0.09396,
		0.112084,
		0.121741,
		0.122681,
		0.115924,
		0.103399,
		0.087528,
		0.070788,
		0.055211,
		0.041877,
		0.030705,
		0.021286,
		0.013573,
		0.007717,
		0.003753,
		0.00146,
		0.000396,
		5.1e-005,
		0.005056,
		0.026875,
		0.057659,
		0.089903,
		0.118003,
		0.138285,
		0.148905,
		0.149661,
		0.141698,
		0.127154,
		0.108744,
		0.089266,
		0.071031,
		0.055198,
		0.041577,
		0.029718,
		0.019659,
		0.011703,
		0.006041,
		0.002543,
		0.00077,
		0.000116,
		0.011854,
		0.040599,
		0.077097,
		0.113883,
		0.145338,
		0.167757,
		0.17929,
		0.179781,
		0.170509,
		0.153832,
		0.132767,
		0.110452,
		0.089464,
		0.070995,
		0.054701,
		0.040099,
		0.027329,
		0.016877,
		0.009125,
		0.00408,
		0.001339,
		0.000225,
		0.001561,
		0.021807,
		0.057546,
		0.099785,
		0.141108,
		0.175889,
		0.200397,
		0.212781,
		0.212934,
		0.202269,
		0.183376,
		0.159588,
		0.134389,
		0.110603,
		0.089387,
		0.070213,
		0.05257,
		0.036721,
		0.023365,
		0.01311,
		0.006142,
		0.002141,
		0.00039,
		0.006383,
		0.035026,
		0.077731,
		0.125657,
		0.171451,
		0.209488,
		0.236018,
		0.249185,
		0.248937,
		0.236823,
		0.215674,
		0.189147,
		0.161075,
		0.134495,
		0.110452,
		0.088209,
		0.067237,
		0.047945,
		0.031269,
		0.018083,
		0.008796,
		0.003214,
		0.000623,
		0.014591,
		0.051521,
		0.101068,
		0.154555,
		0.204699,
		0.245885,
		0.274347,
		0.288229,
		0.287533,
		0.273949,
		0.25055,
		0.22133,
		0.19046,
		0.161142,
		0.134224,
		0.108744,
		0.084171,
		0.061078,
		0.040669,
		0.024116,
		0.012095,
		0.004591,
		0.000933,
		0.001825,
		0.026203,
		0.071204,
		0.127383,
		0.186232,
		0.240551,
		0.284745,
		0.315035,
		0.329566,
		0.328396,
		0.313356,
		0.287772,
		0.255969,
		0.222444,
		0.190493,
		0.160683,
		0.131824,
		0.103399,
		0.076164,
		0.051615,
		0.031258,
		0.016081,
		0.006296,
		0.001329,
		0.007301,
		0.041135,
		0.093895,
		0.156413,
		0.220363,
		0.278635,
		0.325662,
		0.35766,
		0.37278,
		0.371134,
		0.354699,
		0.32705,
		0.292845,
		0.256874,
		0.222444,
		0.18976,
		0.15741,
		0.124906,
		0.09321,
		0.064131,
		0.039541,
		0.020783,
		0.008349,
		0.001819,
		0.01635,
		0.05921,
		0.119331,
		0.187825,
		0.256554,
		0.318512,
		0.368169,
		0.401745,
		0.417399,
		0.415302,
		0.397574,
		0.368043,
		0.331687,
		0.293546,
		0.256834,
		0.22133,
		0.185408,
		0.148632,
		0.112184,
		0.078208,
		0.048973,
		0.026217,
		0.010764,
		0.002407,
		0.001962,
		0.028801,
		0.080172,
		0.147183,
		0.221226,
		0.294361,
		0.359697,
		0.411755,
		0.446768,
		0.462908,
		0.460409,
		0.441538,
		0.410368,
		0.372177,
		0.332201,
		0.293444,
		0.255208,
		0.215674,
		0.174471,
		0.133017,
		0.09381,
		0.05954,
		0.032382,
		0.013546,
		0.003097,
		0.007716,
		0.044403,
		0.103703,
		0.177071,
		0.256183,
		0.333305,
		0.401677,
		0.455881,
		0.492178,
		0.508764,
		0.505937,
		0.486112,
		0.4536,
		0.413956,
		0.372533,
		0.332002,
		0.291157,
		0.248009,
		0.202269,
		0.1556,
		0.110866,
		0.071206,
		0.039266,
		0.016693,
		0.00389,
		0.017013,
		0.062848,
		0.12944,
		0.208583,
		0.29224,
		0.372893,
		0.443925,
		0.5,
		0.537419,
		0.554413,
		0.551352,
		0.530799,
		0.497293,
		0.456627,
		0.414184,
		0.372177,
		0.328883,
		0.282165,
		0.231825,
		0.179781,
		0.129275,
		0.08391,
		0.046839,
		0.020197,
		0.004784,
		0.001978,
		0.029542,
		0.083785,
		0.156993,
		0.241294,
		0.328938,
		0.412633,
		0.485925,
		0.543575,
		0.581942,
		0.599306,
		0.596121,
		0.575095,
		0.540983,
		0.499767,
		0.456759,
		0.413595,
		0.368043,
		0.317842,
		0.262895,
		0.205375,
		0.148905,
		0.097571,
		0.05506,
		0.02404,
		0.005777,
		0.007692,
		0.04496,
		0.106843,
		0.185967,
		0.274783,
		0.365832,
		0.45206,
		0.527188,
		0.586099,
		0.625228,
		0.642924,
		0.639732,
		0.618508,
		0.584203,
		0.542935,
		0.499827,
		0.455837,
		0.408251,
		0.354699,
		0.295193,
		0.232157,
		0.169596,
		0.112084,
		0.063871,
		0.028199,
		0.006863,
		0.016784,
		0.062899,
		0.131646,
		0.215973,
		0.308649,
		0.402506,
		0.49074,
		0.567267,
		0.627109,
		0.666804,
		0.684786,
		0.681705,
		0.660568,
		0.626498,
		0.585686,
		0.542935,
		0.498456,
		0.449081,
		0.392355,
		0.328396,
		0.259871,
		0.191158,
		0.127325,
		0.073201,
		0.032644,
		0.008036,
		0.001912,
		0.028884,
		0.082994,
		0.157829,
		0.246645,
		0.342523,
		0.438588,
		0.528295,
		0.605771,
		0.666205,
		0.706256,
		0.724468,
		0.721612,
		0.700845,
		0.667436,
		0.627576,
		0.585621,
		0.540983,
		0.490085,
		0.430401,
		0.36215,
		0.288229,
		0.213377,
		0.143149,
		0.082967,
		0.037339,
		0.009287,
		0.007387,
		0.043618,
		0.104887,
		0.185047,
		0.277651,
		0.37608,
		0.473757,
		0.564404,
		0.642378,
		0.703054,
		0.743241,
		0.761617,
		0.759087,
		0.738959,
		0.706621,
		0.668181,
		0.627422,
		0.582942,
		0.530799,
		0.468408,
		0.39608,
		0.316924,
		0.236018,
		0.159395,
		0.093074,
		0.042238,
		0.010604,
		0.016032,
		0.060626,
		0.128242,
		0.212988,
		0.308696,
		0.409045,
		0.507751,
		0.598814,
		0.676837,
		0.737402,
		0.777497,
		0.795954,
		0.793834,
		0.774593,
		0.743703,
		0.707107,
		0.667894,
		0.623865,
		0.570757,
		0.505937,
		0.429797,
		0.34563,
		0.25883,
		0.175889,
		0.103418,
		0.047295,
		0.011975,
		0.001807,
		0.027471,
		0.079568,
		0.152754,
		0.241376,
		0.339533,
		0.441194,
		0.540369,
		0.63134,
		0.708972,
		0.769076,
		0.808844,
		0.827286,
		0.825636,
		0.807497,
		0.778389,
		0.743999,
		0.706621,
		0.663304,
		0.609506,
		0.542556,
		0.462908,
		0.374015,
		0.281552,
		0.192443,
		0.113883,
		0.052454,
		0.013387,
		0.006962,
		0.041348,
		0.100129,
		0.178148,
		0.269972,
		0.369959,
		0.472358,
		0.571468,
		0.661863,
		0.73868,
		0.797981,
		0.837182,
		0.855502,
		0.854357,
		0.837497,
		0.810449,
		0.778552,
		0.743232,
		0.700845,
		0.646618,
		0.577846,
		0.495029,
		0.401745,
		0.303917,
		0.208867,
		0.124352,
		0.057659,
		0.014824,
		0.015085,
		0.057334,
		0.122028,
		0.204182,
		0.298578,
		0.399815,
		0.502414,
		0.600964,
		0.690328,
		0.765928,
		0.824096,
		0.862495,
		0.88057,
		0.879941,
		0.864494,
		0.839721,
		0.81052,
		0.777408,
		0.736123,
		0.681705,
		0.611418,
		0.525792,
		0.428498,
		0.325662,
		0.224966,
		0.134701,
		0.06285,
		0.01627,
		0.001693,
		0.025825,
		0.075127,
		0.14501,
		0.230652,
		0.327031,
		0.428982,
		0.531288,
		0.628819,
		0.716734,
		0.790745,
		0.847468,
		0.884835,
		0.902538,
		0.902408,
		0.888462,
		0.866109,
		0.839721,
		0.808892,
		0.768831,
		0.714427,
		0.642924,
		0.554862,
		0.453969,
		0.346533,
		0.240551,
		0.144809,
		0.067966,
		0.017708,
		0.006522,
		0.038856,
		0.094456,
		0.168853,
		0.257379,
		0.355202,
		0.457378,
		0.558942,
		0.655041,
		0.741127,
		0.81321,
		0.868202,
		0.904319,
		0.921517,
		0.921848,
		0.909445,
		0.889587,
		0.866039,
		0.837497,
		0.798725,
		0.744506,
		0.672067,
		0.581942,
		0.477883,
		0.366296,
		0.255441,
		0.154555,
		0.072946,
		0.019122,
		0.014139,
		0.053884,
		0.115075,
		0.19336,
		0.284216,
		0.382989,
		0.484947,
		0.585368,
		0.679669,
		0.76359,
		0.833445,
		0.886448,
		0.921114,
		0.937675,
		0.938409,
		0.927549,
		0.910193,
		0.889428,
		0.863107,
		0.825636,
		0.771729,
		0.698606,
		0.606781,
		0.5,
		0.384738,
		0.269468,
		0.163827,
		0.077731,
		0.020493,
		0.001584,
		0.024222,
		0.07064,
		0.136765,
		0.218359,
		0.311037,
		0.410314,
		0.51166,
		0.610585,
		0.702768,
		0.784235,
		0.8516,
		0.902391,
		0.935426,
		0.951225,
		0.95229,
		0.942932,
		0.928019,
		0.909906,
		0.885676,
		0.849464,
		0.795954,
		0.722367,
		0.629185,
		0.520121,
		0.401677,
		0.282483,
		0.172521,
		0.082265,
		0.021807,
		0.006111,
		0.036478,
		0.088879,
		0.159331,
		0.243699,
		0.337737,
		0.437117,
		0.537502,
		0.634625,
		0.724419,
		0.803188,
		0.867848,
		0.916238,
		0.947489,
		0.96241,
		0.963724,
		0.955795,
		0.943208,
		0.927549,
		0.905228,
		0.870185,
		0.817112,
		0.743241,
		0.649019,
		0.538097,
		0.416966,
		0.294361,
		0.180546,
		0.086499,
		0.023047,
		0.013263,
		0.050639,
		0.108382,
		0.182595,
		0.269249,
		0.364225,
		0.463351,
		0.56247,
		0.657532,
		0.744711,
		0.820585,
		0.882368,
		0.928207,
		0.95755,
		0.971493,
		0.97297,
		0.966371,
		0.955943,
		0.942488,
		0.921848,
		0.887842,
		0.835202,
		0.761188,
		0.666205,
		0.553828,
		0.430498,
		0.305002,
		0.187825,
		0.090387,
		0.024201,
		0.001486,
		0.022747,
		0.06646,
		0.128948,
		0.2064,
		0.294891,
		0.390425,
		0.488978,
		0.586569,
		0.679351,
		0.763735,
		0.836564,
		0.895339,
		0.938519,
		0.96586,
		0.978745,
		0.980301,
		0.974912,
		0.966433,
		0.954896,
		0.935677,
		0.902538,
		0.850289,
		0.77623,
		0.680727,
		0.567267,
		0.442204,
		0.314334,
		0.1943,
		0.093895,
		0.025256,
		0.005737,
		0.034297,
		0.083718,
		0.150396,
		0.230604,
		0.320521,
		0.416269,
		0.513969,
		0.609807,
		0.700132,
		0.781582,
		0.851254,
		0.906936,
		0.947385,
		0.972665,
		0.984431,
		0.985989,
		0.981678,
		0.974912,
		0.96498,
		0.946897,
		0.914431,
		0.862495,
		0.788449,
		0.692627,
		0.578414,
		0.45206,
		0.322316,
		0.199929,
		0.096994,
		0.026203,
		0.012464,
		0.047667,
		0.102209,
		0.172561,
		0.255081,
		0.346046,
		0.441699,
		0.538297,
		0.632191,
		0.719918,
		0.798332,
		0.864779,
		0.917318,
		0.955004,
		0.978195,
		0.988808,
		0.990301,
		0.986929,
		0.98162,
		0.97297,
		0.955728,
		0.923719,
		0.871991,
		0.797981,
		0.701997,
		0.58732,
		0.460076,
		0.328938,
		0.204691,
		0.099666,
		0.027034,
		0.001396,
		0.021402,
		0.062632,
		0.121749,
		0.195295,
		0.279714,
		0.371384,
		0.466662,
		0.561943,
		0.653732,
		0.73875,
		0.81406,
		0.877246,
		0.926633,
		0.961557,
		0.982663,
		0.992111,
		0.993485,
		0.990914,
		0.986797,
		0.979108,
		0.96241,
		0.930634,
		0.878988,
		0.805002,
		0.708972,
		0.594078,
		0.466305,
		0.334215,
		0.208583,
		0.101903,
		0.027745,
		0.005396,
		0.032305,
		0.078987,
		0.142168,
		0.218463,
		0.304401,
		0.39646,
		0.491113,
		0.584885,
		0.674436,
		0.756663,
		0.82883,
		0.888751,
		0.935005,
		0.967203,
		0.986257,
		0.994554,
		0.99577,
		0.993865,
		0.990679,
		0.983637,
		0.967195,
		0.935426,
		0.883723,
		0.809723,
		0.713725,
		0.598814,
		0.470829,
		0.338191,
		0.211619,
		0.103703,
		0.028334,
		0.011736,
		0.044948,
		0.096544,
		0.163312,
		0.241943,
		0.32905,
		0.42121,
		0.515011,
		0.607109,
		0.694312,
		0.773688,
		0.842698,
		0.899376,
		0.942546,
		0.972079,
		0.989142,
		0.996322,
		0.997356,
		0.99599,
		0.993485,
		0.986795,
		0.970337,
		0.938357,
		0.886448,
		0.812377,
		0.716456,
		0.601687,
		0.473757,
		0.340931,
		0.213831,
		0.105077,
		0.028801,
		0.001314,
		0.020173,
		0.059124,
		0.11513,
		0.185043,
		0.265625,
		0.353578,
		0.445578,
		0.538323,
		0.628601,
		0.713365,
		0.789852,
		0.85571,
		0.909189,
		0.949345,
		0.9763,
		0.991455,
		0.997575,
		0.998418,
		0.997473,
		0.995414,
		0.988808,
		0.972083,
		0.939685,
		0.887425,
		0.81321,
		0.717384,
		0.602874,
		0.475222,
		0.342523,
		0.215263,
		0.106041,
		0.029149,
		0.005084,
		0.030481,
		0.074644,
		0.134589,
		0.207232,
		0.289411,
		0.377913,
		0.469512,
		0.561021,
		0.649351,
		0.731605,
		0.805178,
		0.867908,
		0.918247,
		0.955481,
		0.979961,
		0.99331,
		0.998441,
		0.9991,
		0.998467,
		0.996642,
		0.98988,
		0.972665,
		0.939661,
		0.88691,
		0.812471,
		0.716734,
		0.602567,
		0.475373,
		0.343066,
		0.215973,
		0.10662,
		0.029384,
		0.011069,
		0.042454,
		0.091333,
		0.154777,
		0.229766,
		0.313213,
		0.40199,
		0.492972,
		0.583079,
		0.669354,
		0.749038,
		0.819691,
		0.879328,
		0.926602,
		0.961014,
		0.983139,
		0.994799,
		0.999026,
		0.999516,
		0.9991,
		0.997321,
		0.990193,
		0.972293,
		0.938519,
		0.885148,
		0.810403,
		0.714734,
		0.600964,
		0.474364,
		0.34267,
		0.216028,
		0.106843,
		0.029512,
		0.001239,
		0.019046,
		0.055902,
		0.109032,
		0.175563,
		0.25254,
		0.336951,
		0.425753,
		0.515919,
		0.604481,
		0.688606,
		0.765677,
		0.833411,
		0.890001,
		0.934294,
		0.965998,
		0.985898,
		0.995994,
		0.999412,
		0.999756,
		0.999471,
		0.997575,
		0.989904,
		0.971156,
		0.936471,
		0.882368,
		0.807238,
		0.711604,
		0.59826,
		0.472358,
		0.341453,
		0.215498,
		0.106744,
		0.029542,
		0.004799,
		0.028807,
		0.070648,
		0.127594,
		0.196828,
		0.27546,
		0.360554,
		0.449152,
		0.538323,
		0.625212,
		0.707107,
		0.781532,
		0.846361,
		0.899959,
		0.941362,
		0.970479,
		0.988288,
		0.996952,
		0.999659,
		0.999887,
		0.999655,
		0.997502,
		0.989142,
		0.969415,
		0.933706,
		0.878774,
		0.803188,
		0.707552,
		0.594644,
		0.46951,
		0.339533,
		0.21446,
		0.10636,
		0.029484,
		0.010458,
		0.040162,
		0.086531,
		0.146884,
		0.218463,
		0.298442,
		0.383959,
		0.472143,
		0.560159,
		0.645261,
		0.72486,
		0.796619,
		0.858565,
		0.909232,
		0.947842,
		0.974494,
		0.990351,
		0.997716,
		0.999812,
		0.999952,
		0.999705,
		0.997181,
		0.988014,
		0.967203,
		0.930382,
		0.874547,
		0.798444,
		0.702768,
		0.590294,
		0.465973,
		0.337025,
		0.212988,
		0.105728,
		0.029348,
		0.00117,
		0.018011,
		0.052935,
		0.103402,
		0.166781,
		0.24037,
		0.321409,
		0.407111,
		0.494689,
		0.581406,
		0.664622,
		0.741872,
		0.810953,
		0.870045,
		0.91785,
		0.953766,
		0.978079,
		0.992122,
		0.99832,
		0.999904,
		0.999982,
		0.999659,
		0.996665,
		0.986599,
		0.96463,
		0.926633,
		0.86984,
		0.793172,
		0.697422,
		0.585368,
		0.461886,
		0.334039,
		0.211158,
		0.104887,
		0.029144,
		0.004537,
		0.027267,
		0.066963,
		0.121124,
		0.187172,
		0.262459,
		0.344293,
		0.42996,
		0.516759,
		0.602046,
		0.683291,
		0.758149,
		0.824552,
		0.880827,
		0.925841,
		0.959167,
		0.981267,
		0.993633,
		0.998793,
		0.999955,
		0.999993,
		0.999536,
		0.995994,
		0.98496,
		0.961779,
		0.922566,
		0.864779,
		0.787513,
		0.691658,
		0.580011,
		0.457378,
		0.330677,
		0.209038,
		0.103873,
		0.028884,
		0.009896,
		0.038051,
		0.082095,
		0.139573,
		0.207954,
		0.28465,
		0.367032,
		0.452462,
		0.538323,
		0.622067,
		0.701268,
		0.773702,
		0.837433,
		0.890936,
		0.933236,
		0.964076,
		0.984089,
		0.994912,
		0.999156,
		0.999982,
		0.999993,
		0.999349,
		0.995193,
		0.983139,
		0.958714,
		0.918264,
		0.859466,
		0.781582,
		0.685602,
		0.574344,
		0.452559,
		0.327031,
		0.206694,
		0.102721,
		0.028577,
		0.001107,
		0.017058,
		0.050198,
		0.098193,
		0.158632,
		0.229035,
		0.306869,
		0.389573,
		0.474581,
		0.559361,
		0.641458,
		0.718555,
		0.788542,
		0.849616,
		0.900395,
		0.940063,
		0.968525,
		0.986574,
		0.995985,
		0.999429,
		0.999994,
		0.999982,
		0.999102,
		0.994276,
		0.981167,
		0.955481,
		0.91379,
		0.853978,
		0.77547,
		0.679351,
		0.56847,
		0.447525,
		0.323181,
		0.204182,
		0.101464,
		0.028234,
		0.004296,
		0.025848,
		0.063557,
		0.11513,
		0.178197,
		0.250328,
		0.32905,
		0.411865,
		0.496282,
		0.579853,
		0.660215,
		0.735156,
		0.802682,
		0.861121,
		0.909232,
		0.946351,
		0.972543,
		0.988753,
		0.996876,
		0.999629,
		0.999999,
		0.999959,
		0.998793,
		0.993251,
		0.979061,
		0.952111,
		0.909189,
		0.848375,
		0.769247,
		0.672987,
		0.56247,
		0.442357,
		0.319197,
		0.201554,
		0.100129,
		0.027862,
		0.009378,
		0.036101,
		0.07799,
		0.132788,
		0.19817,
		0.271757,
		0.351133,
		0.433865,
		0.517538,
		0.599783,
		0.678333,
		0.751077,
		0.816138,
		0.87197,
		0.917472,
		0.952128,
		0.976158,
		0.99065,
		0.997607,
		0.999771,
		1,
		0.99992,
		0.998419,
		0.992122,
		0.976834,
		0.948626,
		0.904491,
		0.842698,
		0.762966,
		0.666569,
		0.556411,
		0.437117,
		0.315133,
		0.198852,
		0.098742,
		0.027471,
		0.001049,
		0.016179,
		0.047667,
		0.093366,
		0.151058,
		0.218463,
		0.293251,
		0.373065,
		0.455537,
		0.538323,
		0.61914,
		0.69581,
		0.766326,
		0.828924,
		0.882183,
		0.92514,
		0.957423,
		0.9794,
		0.992293,
		0.9982,
		0.999867,
		1,
		0.999859,
		0.997978,
		0.99089,
		0.974494,
		0.94504,
		0.89972,
		0.836975,
		0.756663,
		0.66014,
		0.550341,
		0.431857,
		0.311037,
		0.196111,
		0.097323,
		0.027066,
		0.004073,
		0.024536,
		0.060404,
		0.109566,
		0.169842,
		0.238994,
		0.314747,
		0.394798,
		0.476846,
		0.558619,
		0.637915,
		0.712649,
		0.780913,
		0.841057,
		0.891784,
		0.932261,
		0.962262,
		0.982294,
		0.993706,
		0.998674,
		0.999929,
		1,
		0.999771,
		0.997465,
		0.989554,
		0.972044,
		0.941362,
		0.894886,
		0.831226,
		0.750362,
		0.653732,
		0.544295,
		0.426612,
		0.306942,
		0.19336,
		0.09589,
		0.026654,
		0.0089,
		0.034297,
		0.074184,
		0.12648,
		0.189047,
		0.259691,
		0.336187,
		0.416289,
		0.497762,
		0.578407,
		0.656102,
		0.728851,
		0.794849,
		0.852555,
		0.900792,
		0.938862,
		0.966673,
		0.984868,
		0.994912,
		0.999046,
		0.999966,
		0.999999,
		0.999651,
		0.996876,
		0.988116,
		0.969489,
		0.937598,
		0.890001,
		0.825462,
		0.74408,
		0.647364,
		0.538297,
		0.421411,
		0.302876,
		0.190619,
		0.094456,
		0.026239,
		0.000995,
		0.015366,
		0.045322,
		0.088884,
		0.144007,
		0.208589,
		0.280485,
		0.357518,
		0.437501,
		0.518262,
		0.597675,
		0.673697,
		0.744423,
		0.808146,
		0.863435,
		0.909232,
		0.944967,
		0.970681,
		0.987147,
		0.995933,
		0.999333,
		0.999986,
		0.999995,
		0.999493,
		0.996208,
		0.986574,
		0.966832,
		0.933754,
		0.88507,
		0.819691,
		0.737825,
		0.641049,
		0.532363,
		0.416269,
		0.298855,
		0.187904,
		0.093032,
		0.025825,
		0.003868,
		0.023321,
		0.057479,
		0.104393,
		0.162051,
		0.228392,
		0.301315,
		0.378694,
		0.4584,
		0.538323,
		0.61641,
		0.690698,
		0.759371,
		0.820817,
		0.873716,
		0.917126,
		0.950601,
		0.974313,
		0.989154,
		0.99679,
		0.999548,
		0.999996,
		0.999986,
		0.999293,
		0.995458,
		0.984931,
		0.964076,
		0.929834,
		0.880097,
		0.813918,
		0.731605,
		0.634793,
		0.526502,
		0.411199,
		0.294891,
		0.185227,
		0.091624,
		0.025414,
		0.008457,
		0.032625,
		0.070648,
		0.120607,
		0.180527,
		0.248386,
		0.322126,
		0.399673,
		0.478955,
		0.557927,
		0.634605,
		0.707107,
		0.773702,
		0.832877,
		0.883417,
		0.924495,
		0.955789,
		0.977591,
		0.990912,
		0.997501,
		0.999707,
		0.999999,
		0.999968,
		0.999046,
		0.994625,
		0.983187,
		0.961225,
		0.925841,
		0.875088,
		0.808146,
		0.72542,
		0.628601,
		0.520717,
		0.406206,
		0.290993,
		0.182595,
		0.090239,
		0.02501,
		0.000945,
		0.014613,
		0.043145,
		0.084714,
		0.137432,
		0.199356,
		0.268504,
		0.342867,
		0.420418,
		0.499142,
		0.577059,
		0.652255,
		0.722924,
		0.787425,
		0.84434,
		0.892557,
		0.931363,
		0.960553,
		0.980542,
		0.992444,
		0.998086,
		0.999819,
		1,
		0.999937,
		0.998747,
		0.993706,
		0.981342,
		0.95828,
		0.921781,
		0.870045,
		0.802379,
		0.719275,
		0.622473,
		0.515011,
		0.401293,
		0.287164,
		0.180011,
		0.088879,
		0.024612,
		0.003677,
		0.022194,
		0.054761,
		0.099574,
		0.154777,
		0.218463,
		0.28869,
		0.363491,
		0.440894,
		0.518938,
		0.595706,
		0.669354,
		0.738155,
		0.800552,
		0.855222,
		0.901155,
		0.937751,
		0.964919,
		0.983187,
		0.993771,
		0.99856,
		0.999895,
		1,
		0.999889,
		0.998392,
		0.9927,
		0.9794,
		0.955247,
		0.917657,
		0.864974,
		0.796619,
		0.71317,
		0.61641,
		0.509384,
		0.39646,
		0.283404,
		0.177477,
		0.087547,
		0.024222,
		0.008046,
		0.031071,
		0.067358,
		0.11513,
		0.172561,
		0.237781,
		0.308888,
		0.383959,
		0.461072,
		0.538323,
		0.613858,
		0.685901,
		0.752804,
		0.813093,
		0.865538,
		0.909232,
		0.943681,
		0.968908,
		0.985548,
		0.994912,
		0.998938,
		0.999944,
		1,
		0.999819,
		0.997978,
		0.991607,
		0.97736,
		0.952128,
		0.913472,
		0.859876,
		0.790869,
		0.707107,
		0.610413,
		0.503836,
		0.391707,
		0.279714,
		0.174994,
		0.086243,
		0.023841,
		0.000899,
		0.013913,
		0.041121,
		0.08083,
		0.131291,
		0.19071,
		0.25725,
		0.32905,
		0.404232,
		0.480925,
		0.557281,
		0.631507,
		0.701897,
		0.766878,
		0.82506,
		0.875306,
		0.916807,
		0.949176,
		0.972543,
		0.987648,
		0.995886,
		0.999236,
		0.999973,
		0.999999,
		0.999723,
		0.997501,
		0.990425,
		0.975226,
		0.948926,
		0.909232,
		0.854755,
		0.785132,
		0.701086,
		0.604481,
		0.498366,
		0.387032,
		0.276092,
		0.172561,
		0.084966,
		0.023468,
		0.003501,
		0.021147,
		0.05223,
		0.09508,
		0.147975,
		0.209152,
		0.276811,
		0.349131,
		0.424278,
		0.500431,
		0.575799,
		0.648648,
		0.717343,
		0.780384,
		0.836467,
		0.884543,
		0.9239,
		0.954256,
		0.975845,
		0.989507,
		0.996711,
		0.999465,
		0.999989,
		0.999996,
		0.999596,
		0.996959,
		0.989154,
		0.972999,
		0.945646,
		0.90494,
		0.849616,
		0.779408,
		0.695109,
		0.598615,
		0.492972,
		0.382435,
		0.272538,
		0.170178,
		0.083718,
		0.023103,
		0.007665,
		0.029626,
		0.064291,
		0.110014,
		0.165103,
		0.227824,
		0.296414,
		0.369091,
		0.444067,
		0.51957,
		0.593863,
		0.665277,
		0.732241,
		0.793332,
		0.847326,
		0.893264,
		0.930531,
		0.958943,
		0.978837,
		0.991144,
		0.997403,
		0.999638,
		0.999997,
		0.999989,
		0.999435,
		0.996348,
		0.987794,
		0.970681,
		0.94229,
		0.900598,
		0.84446,
		0.773702,
		0.689177,
		0.592813,
		0.487653,
		0.377913,
		0.269049,
		0.167843,
		0.082496,
		0.022747,
		0.000857,
		0.013263,
		0.039236,
		0.077206,
		0.12555,
		0.182603,
		0.246666,
		0.31601,
		0.388893,
		0.463572,
		0.538323,
		0.611467,
		0.681392,
		0.746597,
		0.805731,
		0.857652,
		0.901488,
		0.936718,
		0.963257,
		0.981538,
		0.992579,
		0.997978,
		0.999765,
		0.999999,
		0.999975,
		0.999236,
		0.995666,
		0.986346,
		0.968276,
		0.938862,
		0.896212,
		0.839291,
		0.768013,
		0.683291,
		0.587077,
		0.48241,
		0.373465,
		0.265625,
		0.165555,
		0.081301,
		0.022399,
		0.003336,
		0.020173,
		0.049871,
		0.09088,
		0.141606,
		0.200411,
		0.265625,
		0.335557,
		0.408505,
		0.482769,
		0.556677,
		0.628601,
		0.696992,
		0.760415,
		0.817592,
		0.867459,
		0.909232,
		0.942482,
		0.967218,
		0.983968,
		0.993829,
		0.998451,
		0.999854,
		1,
		0.99995,
		0.998994,
		0.994912,
		0.98481,
		0.965784,
		0.935365,
		0.891784,
		0.834112,
		0.762345,
		0.677451,
		0.581406,
		0.47724,
		0.369091,
		0.262263,
		0.163312,
		0.080132,
		0.022059,
		0.00731,
		0.02828,
		0.061429,
		0.105229,
		0.15811,
		0.218463,
		0.28465,
		0.355015,
		0.427896,
		0.501639,
		0.574618,
		0.645261,
		0.712079,
		0.773702,
		0.828924,
		0.876762,
		0.916514,
		0.947842,
		0.970847,
		0.986146,
		0.994912,
		0.998834,
		0.999916,
		1,
		0.999911,
		0.998706,
		0.994084,
		0.983187,
		0.963209,
		0.931803,
		0.887317,
		0.828924,
		0.756699,
		0.671658,
		0.575799,
		0.472143,
		0.364787,
		0.258963,
		0.161115,
		0.078987,
		0.021727,
		0.000817,
		0.012657,
		0.037478,
		0.073818,
		0.120172,
		0.174994,
		0.236704,
		0.303696,
		0.374349,
		0.447041,
		0.520162,
		0.592135,
		0.661443,
		0.726654,
		0.786465,
		0.839741,
		0.885575,
		0.92335,
		0.952816,
		0.974162,
		0.988091,
		0.995843,
		0.99914,
		0.999955,
		1,
		0.999854,
		0.998368,
		0.993181,
		0.981477,
		0.960553,
		0.928179,
		0.882814,
		0.823732,
		0.751077,
		0.665912,
		0.570256,
		0.467118,
		0.360554,
		0.255722,
		0.15896,
		0.077867,
		0.021402,
		0.003183,
		0.019264,
		0.047667,
		0.086952,
		0.135635,
		0.192195,
		0.255081,
		0.32272,
		0.393526,
		0.465916,
		0.538323,
		0.609221,
		0.677145,
		0.740722,
		0.798713,
		0.850054,
		0.893914,
		0.929759,
		0.957423,
		0.977183,
		0.989821,
		0.996638,
		0.999381,
		0.999978,
		0.999999,
		0.999777,
		0.997978,
		0.992201,
		0.979683,
		0.957819,
		0.924495,
		0.87828,
		0.818536,
		0.745479,
		0.660215,
		0.564777,
		0.462163,
		0.356389,
		0.25254,
		0.156848,
		0.07677,
		0.021084,
		0.006979,
		0.027023,
		0.058753,
		0.100747,
		0.151546,
		0.209654,
		0.273547,
		0.341684,
		0.412516,
		0.484498,
		0.556109,
		0.625869,
		0.692366,
		0.754286,
		0.810454,
		0.859876,
		0.901795,
		0.935757,
		0.961681,
		0.979927,
		0.991352,
		0.997312,
		0.999567,
		0.999991,
		0.999997,
		0.999675,
		0.997533,
		0.991144,
		0.977805,
		0.95501,
		0.920755,
		0.873716,
		0.81334,
		0.739908,
		0.654566,
		0.559361,
		0.457278,
		0.352291,
		0.249415,
		0.154777,
		0.075696,
		0.020774,
		0.00078,
		0.012092,
		0.035835,
		0.070648,
		0.11513,
		0.167843,
		0.227317,
		0.292057,
		0.360554,
		0.431293,
		0.502771,
		0.573509,
		0.642073,
		0.707107,
		0.767353,
		0.821699,
		0.869219,
		0.909232,
		0.941362,
		0.965609,
		0.982413,
		0.9927,
		0.997877,
		0.999707,
		0.999997,
		0.999991,
		0.999544,
		0.99703,
		0.990011,
		0.975845,
		0.952128,
		0.916963,
		0.869125,
		0.808146,
		0.734365,
		0.648967,
		0.554008,
		0.452462,
		0.34826,
		0.246346,
		0.152746,
		0.074644,
		0.02047,
		0.003041,
		0.018415,
		0.045605,
		0.08327,
		0.130029,
		0.184465,
		0.245133,
		0.31057,
		0.379296,
		0.449834,
		0.520718,
		0.590511,
		0.657831,
		0.721369,
		0.779928,
		0.832457,
		0.878098,
		0.916242,
		0.946591,
		0.969224,
		0.984657,
		0.993882,
		0.998347,
		0.999809,
		0.999999,
		0.99998,
		0.999381,
		0.996468,
		0.9888,
		0.973805,
		0.949176,
		0.913121,
		0.864511,
		0.802956,
		0.728851,
		0.643416,
		0.548718,
		0.447714,
		0.344293,
		0.243331,
		0.150754,
		0.073614,
		0.020173,
		0.006671,
		0.025848,
		0.056247,
		0.096544,
		0.145377,
		0.201355,
		0.263058,
		0.32905,
		0.397884,
		0.468117,
		0.538323,
		0.607109,
		0.673139,
		0.735156,
		0.792019,
		0.842738,
		0.886525,
		0.92284,
		0.951461,
		0.972543,
		0.986675,
		0.994912,
		0.998733,
		0.999882,
		1,
		0.999959,
		0.999183,
		0.995843,
		0.987511,
		0.971686,
		0.946156,
		0.909232,
		0.859876,
		0.797771,
		0.723368,
		0.637915,
		0.54349,
		0.443032,
		0.34039,
		0.24037,
		0.148801,
		0.072605,
		0.019882,
		0.000745,
		0.011564,
		0.034297,
		0.067676,
		0.110396,
		0.161115,
		0.218463,
		0.281048,
		0.347461,
		0.416289,
		0.486123,
		0.555576,
		0.623297,
		0.687996,
		0.74847,
		0.803632,
		0.852555,
		0.894513,
		0.929041,
		0.955987,
		0.975582,
		0.988485,
		0.995804,
		0.999046,
		0.999931,
		1,
		0.999928,
		0.998946,
		0.995156,
		0.986146,
		0.969489,
		0.943071,
		0.905299,
		0.855222,
		0.792593,
		0.717915,
		0.632464,
		0.538323,
		0.438416,
		0.336549,
		0.237461,
		0.146884,
		0.071616,
		0.019597,
		0.002907,
		0.017621,
		0.043674,
		0.079817,
		0.12476,
		0.177183,
		0.23574,
		0.299063,
		0.365774,
		0.434489,
		0.503836,
		0.572465,
		0.639069,
		0.702404,
		0.761317,
		0.814777,
		0.861918,
		0.902078,
		0.93486,
		0.960187,
		0.978359,
		0.990101,
		0.996571,
		0.999296,
		0.999963,
		1,
		0.999882,
		0.998669,
		0.994403,
		0.984704,
		0.967218,
		0.939924,
		0.901325,
		0.850552,
		0.787425,
		0.712494,
		0.627063,
		0.533218,
		0.433865,
		0.332769,
		0.234603,
		0.145004,
		0.070648,
		0.019319,
		0.006382,
		0.024747,
		0.053897,
		0.092596,
		0.139573,
		0.193529,
		0.253142,
		0.317068,
		0.383959,
		0.452462,
		0.52124,
		0.588982,
		0.654422,
		0.716363,
		0.773702,
		0.825463,
		0.870838,
		0.909232,
		0.940314,
		0.964076,
		0.980889,
		0.991538,
		0.997227,
		0.999493,
		0.999982,
		1,
		0.999819,
		0.998347,
		0.993585,
		0.983187,
		0.964874,
		0.936718,
		0.897313,
		0.845868,
		0.782268,
		0.707107,
		0.621711,
		0.528173,
		0.429378,
		0.32905,
		0.231794,
		0.14316,
		0.069699,
		0.019046,
		0.000713,
		0.011069,
		0.032856,
		0.064888,
		0.105945,
		0.154777,
		0.210104,
		0.270627,
		0.33503,
		0.40199,
		0.470189,
		0.538323,
		0.605119,
		0.669354,
		0.729875,
		0.78563,
		0.835698,
		0.879329,
		0.91599,
		0.945417,
		0.967671,
		0.983187,
		0.992809,
		0.997782,
		0.999645,
		0.999993,
		0.999998,
		0.999735,
		0.997978,
		0.9927,
		0.981594,
		0.962458,
		0.933455,
		0.893264,
		0.841172,
		0.777124,
		0.701753,
		0.61641,
		0.523188,
		0.424954,
		0.325389,
		0.229035,
		0.141349,
		0.068768,
		0.018779,
		0.002783,
		0.016877,
		0.041863,
		0.076573,
		0.119801,
		0.170317,
		0.226862,
		0.288158,
		0.352917,
		0.419845,
		0.487653,
		0.555074,
		0.620871,
		0.683863,
		0.742944,
		0.797109,
		0.845492,
		0.887401,
		0.922365,
		0.950184,
		0.970985,
		0.985268,
		0.99393,
		0.998248,
		0.99976,
		0.999998,
		0.999993,
		0.999627,
		0.997561,
		0.991748,
		0.979927,
		0.959974,
		0.930138,
		0.889182,
		0.836467,
		0.771993,
		0.696434,
		0.611159,
		0.518262,
		0.420591,
		0.321785,
		0.226323,
		0.139573,
		0.067857,
		0.018518,
		0.006111,
		0.023716,
		0.051692,
		0.088884,
		0.134105,
		0.186142,
		0.243759,
		0.3057,
		0.370703,
		0.437501,
		0.504839,
		0.571482,
		0.636233,
		0.69795,
		0.755573,
		0.808146,
		0.854856,
		0.895068,
		0.928371,
		0.95463,
		0.974035,
		0.987147,
		0.994912,
		0.998635,
		0.999844,
		1,
		0.999983,
		0.999493,
		0.997092,
		0.990728,
		0.978187,
		0.957423,
		0.926769,
		0.88507,
		0.831754,
		0.766878,
		0.69115,
		0.605958,
		0.513395,
		0.416289,
		0.318238,
		0.223658,
		0.137829,
		0.066963,
		0.018262,
		0.000682,
		0.010606,
		0.031504,
		0.062267,
		0.101756,
		0.148801,
		0.202206,
		0.260756,
		0.32322,
		0.38836,
		0.45494,
		0.521733,
		0.58754,
		0.651201,
		0.711615,
		0.767766,
		0.818748,
		0.863799,
		0.902341,
		0.934022,
		0.958769,
		0.976835,
		0.988837,
		0.995768,
		0.998953,
		0.999903,
		1,
		0.999967,
		0.99933,
		0.996571,
		0.989641,
		0.976376,
		0.954807,
		0.92335,
		0.880929,
		0.827036,
		0.761779,
		0.685901,
		0.600807,
		0.508587,
		0.412048,
		0.314747,
		0.221038,
		0.136118,
		0.066086,
		0.018011,
		0.002666,
		0.016179,
		0.040162,
		0.073521,
		0.11513,
		0.163836,
		0.218463,
		0.277816,
		0.340688,
		0.405867,
		0.472143,
		0.538323,
		0.603241,
		0.665773,
		0.72486,
		0.77953,
		0.828924,
		0.872333,
		0.909232,
		0.939331,
		0.962616,
		0.9794,
		0.990352,
		0.996509,
		0.999212,
		0.999944,
		1,
		0.999941,
		0.999134,
		0.995996,
		0.988485,
		0.974494,
		0.952128,
		0.919885,
		0.876762,
		0.822315,
		0.756699,
		0.68069,
		0.595706,
		0.503836,
		0.407865,
		0.31131,
		0.218463,
		0.134438,
		0.065227,
		0.017765,
		0.005858,
		0.022747,
		0.049618,
		0.085389,
		0.128949,
		0.179163,
		0.234873,
		0.294906,
		0.358077,
		0.423201,
		0.489096,
		0.5546,
		0.618579,
		0.679948,
		0.737688,
		0.790869,
		0.838682,
		0.880467,
		0.915755,
		0.944313,
		0.966185,
		0.981742,
		0.991705,
		0.997147,
		0.999419,
		0.99997,
		1,
		0.999903,
		0.998903,
		0.995364,
		0.987261,
		0.972543,
		0.949389,
		0.916375,
		0.872571,
		0.817592,
		0.751638,
		0.675515,
		0.590655,
		0.499142,
		0.403741,
		0.307926,
		0.215931,
		0.132788,
		0.064384,
		0.017525,
		0.000654,
		0.010171,
		0.030233,
		0.059802,
		0.097809,
		0.14316,
		0.194735,
		0.251397,
		0.311993,
		0.375363,
		0.440344,
		0.505785,
		0.570554,
		0.63355,
		0.693726,
		0.750101,
		0.801791,
		0.848031,
		0.888213,
		0.921922,
		0.94898,
		0.969489,
		0.983876,
		0.992908,
		0.997692,
		0.999581,
		0.999986,
		1,
		0.999851,
		0.998635,
		0.994676,
		0.98597,
		0.970524,
		0.946591,
		0.912824,
		0.868358,
		0.812868,
		0.746597,
		0.670377,
		0.585653,
		0.494505,
		0.399673,
		0.304595,
		0.213443,
		0.131168,
		0.063557,
		0.017289,
		0.002556,
		0.015523,
		0.038563,
		0.070648,
		0.110724,
		0.157713,
		0.210511,
		0.268,
		0.32905,
		0.392521,
		0.457278,
		0.522198,
		0.586177,
		0.648151,
		0.707107,
		0.762104,
		0.812301,
		0.85698,
		0.895582,
		0.927745,
		0.953345,
		0.972543,
		0.985814,
		0.993974,
		0.998153,
		0.999707,
		0.999994,
		0.999998,
		0.999782,
		0.998327,
		0.99393,
		0.984612,
		0.968439,
		0.943738,
		0.909232,
		0.864126,
		0.808146,
		0.741577,
		0.665277,
		0.580701,
		0.489923,
		0.395662,
		0.301315,
		0.210995,
		0.129577,
		0.062746,
		0.017058,
		0.00562,
		0.021837,
		0.047667,
		0.082095,
		0.124081,
		0.172561,
		0.226451,
		0.28465,
		0.346048,
		0.409531,
		0.473988,
		0.538323,
		0.601465,
		0.66238,
		0.720092,
		0.773702,
		0.822408,
		0.865538,
		0.902584,
		0.933236,
		0.957423,
		0.975358,
		0.987569,
		0.994912,
		0.998541,
		0.999801,
		0.999998,
		0.999994,
		0.999693,
		0.997978,
		0.993125,
		0.983187,
		0.966289,
		0.94083,
		0.905603,
		0.859876,
		0.803427,
		0.736579,
		0.660215,
		0.575799,
		0.485397,
		0.391707,
		0.298086,
		0.208589,
		0.128015,
		0.061951,
		0.016832,
		0.000627,
		0.009762,
		0.029038,
		0.057479,
		0.094085,
		0.137829,
		0.187663,
		0.242518,
		0.301315,
		0.362964,
		0.426374,
		0.49046,
		0.554152,
		0.61641,
		0.676234,
		0.732684,
		0.784898,
		0.832117,
		0.873716,
		0.909232,
		0.938407,
		0.961225,
		0.977948,
		0.989154,
		0.995735,
		0.998863,
		0.999871,
		1,
		0.999986,
		0.999581,
		0.997585,
		0.992261,
		0.981696,
		0.964076,
		0.93787,
		0.901939,
		0.85561,
		0.798713,
		0.731605,
		0.655192,
		0.570945,
		0.480925,
		0.387806,
		0.294906,
		0.206223,
		0.12648,
		0.06117,
		0.01661,
		0.002453,
		0.014907,
		0.037057,
		0.067939,
		0.106564,
		0.151921,
		0.202976,
		0.258679,
		0.317968,
		0.379775,
		0.443032,
		0.50668,
		0.569676,
		0.63101,
		0.689715,
		0.744885,
		0.795699,
		0.841438,
		0.881522,
		0.915536,
		0.943272,
		0.964764,
		0.980326,
		0.990579,
		0.996452,
		0.999128,
		0.99992,
		1,
		0.999973,
		0.999445,
		0.997147,
		0.991337,
		0.98014,
		0.961803,
		0.93486,
		0.898242,
		0.851331,
		0.794004,
		0.726654,
		0.650207,
		0.56614,
		0.476508,
		0.383959,
		0.291774,
		0.203896,
		0.124973,
		0.060404,
		0.016392,
		0.005396,
		0.02098,
		0.045828,
		0.078987,
		0.119481,
		0.166312,
		0.218463,
		0.2749,
		0.334582,
		0.39646,
		0.45949,
		0.522638,
		0.584888,
		0.645261,
		0.702821,
		0.756699,
		0.80611,
		0.850378,
		0.888967,
		0.921508,
		0.947842,
		0.968053,
		0.982503,
		0.991857,
		0.997073,
		0.999343,
		0.999953,
		1,
		0.999951,
		0.999281,
		0.996662,
		0.990352,
		0.978519,
		0.959469,
		0.931803,
		0.894513,
		0.84704,
		0.789303,
		0.721729,
		0.645261,
		0.561384,
		0.472143,
		0.380165,
		0.28869,
		0.201607,
		0.123492,
		0.059653,
		0.016179,
		0.000602,
		0.009378,
		0.027912,
		0.055289,
		0.090569,
		0.132788,
		0.180962,
		0.234088,
		0.291153,
		0.351133,
		0.413,
		0.475734,
		0.538323,
		0.599783,
		0.65916,
		0.715554,
		0.768129,
		0.816138,
		0.858946,
		0.89606,
		0.927158,
		0.952128,
		0.971103,
		0.984492,
		0.992999,
		0.997607,
		0.999515,
		0.999975,
		1,
		0.99992,
		0.999088,
		0.996128,
		0.989308,
		0.976835,
		0.957078,
		0.9287,
		0.890756,
		0.842738,
		0.784611,
		0.716828,
		0.640354,
		0.556677,
		0.467831,
		0.376422,
		0.285652,
		0.199356,
		0.122037,
		0.058915,
		0.01597,
		0.002356,
		0.014327,
		0.035637,
		0.065382,
		0.102633,
		0.146439,
		0.19583,
		0.249819,
		0.307411,
		0.367599,
		0.429378,
		0.491749,
		0.553728,
		0.614355,
		0.672708,
		0.727916,
		0.77918,
		0.82579,
		0.867151,
		0.902812,
		0.932498,
		0.956143,
		0.973927,
		0.986304,
		0.994013,
		0.998064,
		0.999651,
		0.999988,
		1,
		0.999876,
		0.998863,
		0.995546,
		0.988203,
		0.975089,
		0.95463,
		0.925553,
		0.886971,
		0.838428,
		0.779928,
		0.711954,
		0.635486,
		0.552017,
		0.463572,
		0.372731,
		0.282661,
		0.197141,
		0.120607,
		0.05819,
		0.015765,
		0.005185,
		0.020173,
		0.044093,
		0.076051,
		0.11513,
		0.160392,
		0.21088,
		0.265625,
		0.323646,
		0.383959,
		0.445578,
		0.507526,
		0.568845,
		0.628601,
		0.685901,
		0.739908,
		0.789856,
		0.835071,
		0.875,
		0.909232,
		0.937538,
		0.959898,
		0.976536,
		0.987949,
		0.994912,
		0.998451,
		0.999755,
		0.999995,
		0.999998,
		0.999819,
		0.998605,
		0.994912,
		0.987038,
		0.973282,
		0.952128,
		0.922365,
		0.883162,
		0.834112,
		0.775256,
		0.707107,
		0.630658,
		0.547405,
		0.459363,
		0.369091,
		0.279714,
		0.194962,
		0.119202,
		0.057479,
		0.015563,
		0.000579,
		0.009016,
		0.02685,
		0.053221,
		0.087245,
		0.128015,
		0.174607,
		0.226079,
		0.281477,
		0.339837,
		0.400195,
		0.461585,
		0.523054,
		0.583667,
		0.642517,
		0.698742,
		0.751535,
		0.800163,
		0.843991,
		0.882503,
		0.915332,
		0.94229,
		0.963404,
		0.978942,
		0.98944,
		0.995704,
		0.998776,
		0.999834,
		0.999998,
		0.999995,
		0.999744,
		0.99831,
		0.994227,
		0.985814,
		0.971415,
		0.949573,
		0.919137,
		0.879329,
		0.829789,
		0.770596,
		0.702287,
		0.625869,
		0.542841,
		0.455206,
		0.365499,
		0.276811,
		0.192819,
		0.117821,
		0.056781,
		0.015366,
		0.002265,
		0.013779,
		0.034297,
		0.062966,
		0.098913,
		0.141244,
		0.189047,
		0.241394,
		0.297347,
		0.355962,
		0.416289,
		0.477387,
		0.538323,
		0.598188,
		0.656102,
		0.71123,
		0.762798,
		0.810106,
		0.852555,
		0.889668,
		0.92112,
		0.946765,
		0.966673,
		0.981155,
		0.990785,
		0.996398,
		0.999046,
		0.999892,
		1,
		0.999989,
		0.999651,
		0.997978,
		0.99349,
		0.98453,
		0.969489,
		0.946968,
		0.91587,
		0.875474,
		0.825463,
		0.76595,
		0.697495,
		0.62112,
		0.538323,
		0.451099,
		0.361957,
		0.273952,
		0.19071,
		0.116464,
		0.056095,
		0.015172,
		0.004986,
		0.019411,
		0.042454,
		0.073275,
		0.11101,
		0.154777,
		0.203676,
		0.256796,
		0.313213,
		0.371999,
		0.432226,
		0.492972,
		0.553326,
		0.612405,
		0.669354,
		0.723368,
		0.773702,
		0.819691,
		0.860772,
		0.896506,
		0.926607,
		0.950973,
		0.969716,
		0.983187,
		0.991996,
		0.997003,
		0.999268,
		0.999933,
		1,
		0.999977,
		0.999536,
		0.997607,
		0.9927,
		0.983187,
		0.967506,
		0.944313,
		0.912568,
		0.8716,
		0.821134,
		0.761317,
		0.692731,
		0.61641,
		0.533853,
		0.447041,
		0.358463,
		0.271135,
		0.188634,
		0.11513,
		0.055422,
		0.014982,
		0.000557,
		0.008674,
		0.025848,
		0.051267,
		0.084099,
		0.123492,
		0.168575,
		0.218463,
		0.272257,
		0.32905,
		0.387931,
		0.447991,
		0.508329,
		0.568056,
		0.626313,
		0.682272,
		0.735156,
		0.78425,
		0.828924,
		0.868649,
		0.903024,
		0.931803,
		0.954925,
		0.972543,
		0.985047,
		0.993081,
		0.997527,
		0.999448,
		0.999961,
		1,
		0.999959,
		0.999399,
		0.997195,
		0.991857,
		0.981785,
		0.965467,
		0.94161,
		0.909232,
		0.867708,
		0.816804,
		0.756699,
		0.687996,
		0.61174,
		0.529428,
		0.443032,
		0.355015,
		0.268359,
		0.186592,
		0.113819,
		0.054761,
		0.014796,
		0.002179,
		0.013263,
		0.033031,
		0.060681,
		0.095391,
		0.136317,
		0.182603,
		0.233376,
		0.28775,
		0.344836,
		0.403741,
		0.463572,
		0.523449,
		0.582507,
		0.63991,
		0.694856,
		0.746597,
		0.794448,
		0.837812,
		0.876195,
		0.909232,
		0.936718,
		0.958631,
		0.975164,
		0.986746,
		0.99405,
		0.997978,
		0.999593,
		0.999979,
		1,
		0.999933,
		0.999236,
		0.996742,
		0.990961,
		0.980326,
		0.963372,
		0.938862,
		0.905864,
		0.863799,
		0.812474,
		0.752097,
		0.683291,
		0.607109,
		0.52505,
		0.439072,
		0.351615,
		0.265625,
		0.184582,
		0.112529,
		0.054111,
		0.014613,
		0.004799,
		0.018691,
		0.040905,
		0.070648,
		0.107106,
		0.149448,
		0.196828,
		0.248386,
		0.303253,
		0.360554,
		0.419412,
		0.478955,
		0.538323,
		0.596674,
		0.653193,
		0.707107,
		0.757694,
		0.804301,
		0.846361,
		0.883417,
		0.91514,
		0.941362,
		0.962103,
		0.977591,
		0.988294,
		0.994912,
		0.998364,
		0.999707,
		0.99999,
		1,
		0.999897,
		0.999046,
		0.996245,
		0.990011,
		0.97881,
		0.961225,
		0.93607,
		0.902465,
		0.859876,
		0.808146,
		0.747512,
		0.678615,
		0.602518,
		0.520718,
		0.435159,
		0.34826,
		0.262931,
		0.182603,
		0.111261,
		0.053473,
		0.014433,
		0.000536,
		0.008351,
		0.0249,
		0.049418,
		0.081119,
		0.119202,
		0.162846,
		0.211216,
		0.263466,
		0.318742,
		0.376183,
		0.43493,
		0.494131,
		0.552945,
		0.610552,
		0.666161,
		0.719025,
		0.76845,
		0.813813,
		0.854579,
		0.890323,
		0.920755,
		0.945745,
		0.965348,
		0.979834,
		0.9897,
		0.995675,
		0.99869,
		0.999794,
		0.999996,
		0.999999,
		0.999848,
		0.998827,
		0.995704,
		0.989007,
		0.977237,
		0.959025,
		0.933236,
		0.899037,
		0.855939,
		0.80382,
		0.742944,
		0.673969,
		0.597967,
		0.51643,
		0.431293,
		0.34495,
		0.260276,
		0.180656,
		0.110014,
		0.052846,
		0.014256,
		0.002098,
		0.012775,
		0.031834,
		0.058518,
		0.092052,
		0.131641,
		0.176478,
		0.22574,
		0.278592,
		0.334196,
		0.391707,
		0.450283,
		0.50909,
		0.567308,
		0.624138,
		0.678813,
		0.730613,
		0.778869,
		0.82299,
		0.862471,
		0.896922,
		0.926089,
		0.949876,
		0.968379,
		0.981901,
		0.990973,
		0.996348,
		0.998965,
		0.999861,
		0.999999,
		0.999996,
		0.999785,
		0.998577,
		0.995118,
		0.987949,
		0.975609,
		0.956775,
		0.930361,
		0.895582,
		0.85199,
		0.799498,
		0.738394,
		0.669354,
		0.593455,
		0.512188,
		0.427474,
		0.341684,
		0.257659,
		0.178739,
		0.108788,
		0.05223,
		0.014083,
		0.004622,
		0.018011,
		0.039439,
		0.068159,
		0.103402,
		0.144385,
		0.190312,
		0.24037,
		0.29374,
		0.349596,
		0.407111,
		0.465459,
		0.523824,
		0.581406,
		0.637429,
		0.69115,
		0.741872,
		0.788955,
		0.831837,
		0.870045,
		0.903223,
		0.931149,
		0.953766,
		0.971204,
		0.983803,
		0.992122,
		0.996937,
		0.999193,
		0.999909,
		1,
		0.99999,
		0.999707,
		0.998294,
		0.994485,
		0.986838,
		0.973927,
		0.954475,
		0.927446,
		0.892101,
		0.848031,
		0.795181,
		0.733863,
		0.664769,
		0.588982,
		0.50799,
		0.423701,
		0.338462,
		0.255081,
		0.176852,
		0.107582,
		0.051625,
		0.013913,
		0.000516,
		0.008046,
		0.024004,
		0.047667,
		0.078295,
		0.11513,
		0.1574,
		0.204316,
		0.255081,
		0.308888,
		0.364925,
		0.42238,
		0.480445,
		0.538323,
		0.595234,
		0.650423,
		0.70317,
		0.752804,
		0.798713,
		0.840361,
		0.877309,
		0.909232,
		0.935944,
		0.957423,
		0.973833,
		0.985548,
		0.993157,
		0.997451,
		0.999381,
		0.999944,
		1,
		0.999981,
		0.99961,
		0.997978,
		0.993807,
		0.985674,
		0.972191,
		0.952128,
		0.924495,
		0.888597,
		0.844063,
		0.790869,
		0.729352,
		0.660215,
		0.584549,
		0.503836,
		0.419973,
		0.335282,
		0.25254,
		0.174994,
		0.106396,
		0.05103,
		0.013746,
		0.002021,
		0.012313,
		0.0307,
		0.056468,
		0.088884,
		0.127199,
		0.170651,
		0.218463,
		0.269849,
		0.324015,
		0.380165,
		0.437501,
		0.495233,
		0.552583,
		0.608789,
		0.663118,
		0.714875,
		0.763413,
		0.808146,
		0.848568,
		0.88427,
		0.91496,
		0.940484,
		0.960857,
		0.976276,
		0.987147,
		0.994084,
		0.997897,
		0.999533,
		0.999967,
		1,
		0.999966,
		0.999493,
		0.997627,
		0.993081,
		0.984457,
		0.970403,
		0.949734,
		0.921508,
		0.88507,
		0.840087,
		0.786565,
		0.72486,
		0.655692,
		0.580154,
		0.499726,
		0.416289,
		0.332145,
		0.250035,
		0.173165,
		0.105229,
		0.050445,
		0.013582,
		0.004454,
		0.017367,
		0.038051,
		0.065798,
		0.099885,
		0.139573,
		0.184108,
		0.232725,
		0.28465,
		0.339103,
		0.395301,
		0.452462,
		0.509814,
		0.566596,
		0.622067,
		0.675515,
		0.726267,
		0.773702,
		0.81726,
		0.856465,
		0.890936,
		0.920413,
		0.944778,
		0.964076,
		0.978541,
		0.988607,
		0.994912,
		0.99828,
		0.999656,
		0.999982,
		1,
		0.999944,
		0.999355,
		0.997238,
		0.992308,
		0.983187,
		0.968563,
		0.947296,
		0.918486,
		0.881522,
		0.836105,
		0.782268,
		0.72039,
		0.651201,
		0.575799,
		0.495659,
		0.41265,
		0.32905,
		0.247567,
		0.171364,
		0.104082,
		0.049871,
		0.013421,
		0.000497,
		0.007758,
		0.023155,
		0.046007,
		0.075614,
		0.111261,
		0.152218,
		0.197741,
		0.247077,
		0.299464,
		0.354133,
		0.410318,
		0.467253,
		0.52418,
		0.580358,
		0.635065,
		0.687611,
		0.737347,
		0.783673,
		0.82606,
		0.864056,
		0.897313,
		0.925601,
		0.948834,
		0.967092,
		0.980638,
		0.989936,
		0.995649,
		0.998608,
		0.999752,
		0.999992,
		1,
		0.999913,
		0.999193,
		0.996812,
		0.991486,
		0.981865,
		0.966673,
		0.944814,
		0.915432,
		0.877955,
		0.832117,
		0.77798,
		0.71594,
		0.64674,
		0.571482,
		0.491635,
		0.409054,
		0.325995,
		0.245133,
		0.16959,
		0.102952,
		0.049305,
		0.013263,
		0.001948,
		0.011876,
		0.029626,
		0.054523,
		0.085875,
		0.122976,
		0.165103,
		0.211524,
		0.261496,
		0.31427,
		0.369091,
		0.425205,
		0.481862,
		0.538323,
		0.593863,
		0.647781,
		0.699408,
		0.748116,
		0.793332,
		0.834551,
		0.87135,
		0.903409,
		0.930531,
		0.952661,
		0.969911,
		0.982575,
		0.991144,
		0.996301,
		0.998886,
		0.999826,
		0.999997,
		0.999999,
		0.999872,
		0.999006,
		0.996348,
		0.990617,
		0.980491,
		0.964734,
		0.94229,
		0.912347,
		0.874369,
		0.828126,
		0.773702,
		0.711512,
		0.642312,
		0.567204,
		0.487653,
		0.405501,
		0.32298,
		0.242735,
		0.167843,
		0.101841,
		0.04875,
		0.013107,
		0.004296,
		0.016757,
		0.036734,
		0.063557,
		0.096544,
		0.134994,
		0.178197,
		0.225429,
		0.27596,
		0.32905,
		0.383959,
		0.439948,
		0.496282,
		0.552238,
		0.607109,
		0.660215,
		0.710906,
		0.758579,
		0.802682,
		0.842738,
		0.878352,
		0.909232,
		0.935212,
		0.956268,
		0.972543,
		0.984361,
		0.992239,
		0.996876,
		0.999119,
		0.999882,
		0.999999,
		0.999997,
		0.999819,
		0.998793,
		0.995843,
		0.9897,
		0.979066,
		0.962746,
		0.939726,
		0.909232,
		0.870768,
		0.824131,
		0.769433,
		0.707107,
		0.637915,
		0.562964,
		0.483714,
		0.40199,
		0.320005,
		0.24037,
		0.166123,
		0.100747,
		0.048204,
		0.012955,
		0.000479,
		0.007484,
		0.02235,
		0.044432,
		0.073068,
		0.107582,
		0.147285,
		0.191473,
		0.239433,
		0.290446,
		0.343786,
		0.398725,
		0.454537,
		0.510503,
		0.565918,
		0.620093,
		0.672365,
		0.722107,
		0.768736,
		0.811728,
		0.850628,
		0.88507,
		0.91479,
		0.939652,
		0.959663,
		0.974996,
		0.986004,
		0.993226,
		0.99738,
		0.999313,
		0.999923,
		1,
		0.999992,
		0.999752,
		0.998551,
		0.995298,
		0.988734,
		0.977591,
		0.960712,
		0.937122,
		0.906089,
		0.867151,
		0.820135,
		0.765177,
		0.702724,
		0.63355,
		0.558763,
		0.479816,
		0.398522,
		0.317068,
		0.238038,
		0.164428,
		0.099671,
		0.047667,
		0.012805,
		0.001879,
		0.011462,
		0.028608,
		0.052677,
		0.083016,
		0.118957,
		0.159817,
		0.204903,
		0.253513,
		0.304937,
		0.358463,
		0.413374,
		0.468961,
		0.524519,
		0.579359,
		0.63281,
		0.68423,
		0.73301,
		0.778592,
		0.820474,
		0.858226,
		0.89151,
		0.92009,
		0.943859,
		0.962855,
		0.977279,
		0.987511,
		0.994115,
		0.997819,
		0.999473,
		0.999952,
		1,
		0.999984,
		0.999669,
		0.99828,
		0.994712,
		0.98772,
		0.976067,
		0.958631,
		0.934481,
		0.90292,
		0.863519,
		0.816138,
		0.760931,
		0.698364,
		0.629217,
		0.5546,
		0.475959,
		0.395094,
		0.31417,
		0.23574,
		0.162759,
		0.098612,
		0.047138,
		0.012657,
		0.004145,
		0.016179,
		0.035484,
		0.061429,
		0.093366,
		0.130634,
		0.172561,
		0.218463,
		0.267647,
		0.319414,
		0.373065,
		0.427896,
		0.483212,
		0.538323,
		0.592557,
		0.645261,
		0.69581,
		0.743619,
		0.78815,
		0.828924,
		0.865538,
		0.897679,
		0.92514,
		0.947842,
		0.965851,
		0.9794,
		0.988892,
		0.994912,
		0.9982,
		0.999603,
		0.999972,
		1,
		0.999971,
		0.99957,
		0.997978,
		0.994084,
		0.986658,
		0.974494,
		0.956507,
		0.931803,
		0.899725,
		0.859876,
		0.812141,
		0.756699,
		0.694027,
		0.624916,
		0.550474,
		0.472143,
		0.391707,
		0.31131,
		0.233473,
		0.161115,
		0.09757,
		0.046619,
		0.012512,
		0.000462,
		0.007225,
		0.021587,
		0.042936,
		0.070648,
		0.104082,
		0.142584,
		0.185491,
		0.232129,
		0.281814,
		0.333861,
		0.387578,
		0.442279,
		0.497281,
		0.551909,
		0.605508,
		0.657442,
		0.707107,
		0.753936,
		0.797413,
		0.837084,
		0.872571,
		0.903584,
		0.929947,
		0.951607,
		0.968661,
		0.981366,
		0.990153,
		0.995624,
		0.998528,
		0.999707,
		0.999985,
		1,
		0.999952,
		0.999452,
		0.997644,
		0.993413,
		0.985548,
		0.972873,
		0.954338,
		0.92909,
		0.896506,
		0.85622,
		0.808146,
		0.75248,
		0.689715,
		0.620647,
		0.546387,
		0.468368,
		0.38836,
		0.308487,
		0.231239,
		0.159495,
		0.096544,
		0.046108,
		0.01237,
		0.001814,
		0.011069,
		0.027641,
		0.050923,
		0.080297,
		0.11513,
		0.154777,
		0.198581,
		0.245878,
		0.295997,
		0.34826,
		0.40199,
		0.456513,
		0.51116,
		0.565272,
		0.618209,
		0.669354,
		0.71812,
		0.763962,
		0.806386,
		0.84496,
		0.879329,
		0.909232,
		0.934518,
		0.955164,
		0.971292,
		0.983187,
		0.991302,
		0.996257,
		0.998808,
		0.999788,
		0.999993,
		1,
		0.999926,
		0.999313,
		0.997277,
		0.9927,
		0.984391,
		0.971204,
		0.952128,
		0.926344,
		0.893264,
		0.852555,
		0.804153,
		0.748275,
		0.685426,
		0.61641,
		0.542336,
		0.464632,
		0.385053,
		0.3057,
		0.229035,
		0.157899,
		0.095534,
		0.045605,
		0.01223,
		0.004003,
		0.01563,
		0.034297,
		0.059405,
		0.090341,
		0.12648,
		0.167184,
		0.211807,
		0.259691,
		0.310176,
		0.362598,
		0.416289,
		0.470589,
		0.524843,
		0.578407,
		0.630658,
		0.680995,
		0.728851,
		0.773702,
		0.815072,
		0.852555,
		0.88582,
		0.91463,
		0.938862,
		0.958519,
		0.973751,
		0.984868,
		0.992346,
		0.996818,
		0.999046,
		0.999852,
		0.999997,
		0.999999,
		0.999891,
		0.999153,
		0.996876,
		0.991944,
		0.983187,
		0.969489,
		0.949876,
		0.923565,
		0.890001,
		0.84888,
		0.800163,
		0.744084,
		0.681162,
		0.612205,
		0.538323,
		0.460936,
		0.381784,
		0.302949,
		0.226862,
		0.156326,
		0.094539,
		0.045111,
		0.012092,
		0.000447,
		0.006979,
		0.020862,
		0.041515,
		0.068345,
		0.100747,
		0.138102,
		0.179781,
		0.225145,
		0.273547,
		0.324337,
		0.37686,
		0.430464,
		0.484498,
		0.538323,
		0.591311,
		0.642853,
		0.692366,
		0.739302,
		0.783156,
		0.823476,
		0.859876,
		0.892049,
		0.919785,
		0.942986,
		0.961681,
		0.976047,
		0.986419,
		0.993291,
		0.997312,
		0.999246,
		0.999899,
		0.999999,
		0.999997,
		0.999846,
		0.99897,
		0.996439,
		0.991144,
		0.981935,
		0.967729,
		0.947585,
		0.920755,
		0.886718,
		0.845197,
		0.796177,
		0.739908,
		0.676922,
		0.608032,
		0.534347,
		0.457278,
		0.378555,
		0.300233,
		0.224718,
		0.154777,
		0.09356,
		0.044624,
		0.011957,
		0.001752,
		0.010696,
		0.026722,
		0.049255,
		0.077708,
		0.111483,
		0.149968,
		0.192542,
		0.238574,
		0.287427,
		0.338462,
		0.391034,
		0.444504,
		0.498233,
		0.551596,
		0.603979,
		0.654791,
		0.703467,
		0.749475,
		0.79233,
		0.831602,
		0.866927,
		0.898023,
		0.924704,
		0.946896,
		0.964657,
		0.978187,
		0.987845,
		0.994144,
		0.997745,
		0.999412,
		0.999934,
		1,
		0.999993,
		0.999788,
		0.998762,
		0.995967,
		0.990301,
		0.980638,
		0.965925,
		0.945255,
		0.917916,
		0.883417,
		0.841508,
		0.792195,
		0.735749,
		0.672708,
		0.603892,
		0.530408,
		0.45366,
		0.375363,
		0.297552,
		0.222604,
		0.15325,
		0.092596,
		0.044146,
		0.011823,
		0.003868,
		0.015108,
		0.033168,
		0.057479,
		0.087461,
		0.122519,
		0.162051,
		0.205443,
		0.252073,
		0.301315,
		0.352538,
		0.405109,
		0.4584,
		0.511786,
		0.564656,
		0.61641,
		0.666474,
		0.714298,
		0.759371,
		0.801227,
		0.839455,
		0.873716,
		0.903749,
		0.929393,
		0.950601,
		0.967454,
		0.980179,
		0.989154,
		0.994912,
		0.998123,
		0.999548,
		0.999959,
		1,
		0.999986,
		0.999718,
		0.998528,
		0.995458,
		0.989415,
		0.979295,
		0.964076,
		0.942888,
		0.915048,
		0.880097,
		0.837812,
		0.78822,
		0.731605,
		0.668518,
		0.599783,
		0.526505,
		0.45008,
		0.372208,
		0.294906,
		0.220519,
		0.151745,
		0.091646,
		0.043674,
		0.011692,
		0.000431,
		0.006746,
		0.020173,
		0.040162,
		0.066153,
		0.09757,
		0.133826,
		0.174326,
		0.218463,
		0.265625,
		0.315194,
		0.366551,
		0.419073,
		0.472143,
		0.525152,
		0.577498,
		0.628601,
		0.677898,
		0.72486,
		0.768993,
		0.809849,
		0.84704,
		0.880246,
		0.909232,
		0.93386,
		0.954108,
		0.970081,
		0.982029,
		0.990352,
		0.995601,
		0.998451,
		0.99966,
		0.999976,
		1,
		0.999975,
		0.999632,
		0.998267,
		0.994912,
		0.988485,
		0.977906,
		0.962186,
		0.940484,
		0.912153,
		0.876762,
		0.834112,
		0.78425,
		0.727478,
		0.664354,
		0.595706,
		0.522638,
		0.446537,
		0.369091,
		0.292293,
		0.218463,
		0.150262,
		0.090711,
		0.043211,
		0.011564,
		0.001693,
		0.010342,
		0.025848,
		0.047667,
		0.075242,
		0.108006,
		0.145377,
		0.186768,
		0.23158,
		0.27921,
		0.32905,
		0.380488,
		0.432915,
		0.485726,
		0.538323,
		0.59012,
		0.640549,
		0.689066,
		0.735156,
		0.778343,
		0.818201,
		0.854361,
		0.886525,
		0.914479,
		0.938112,
		0.957423,
		0.972543,
		0.983744,
		0.991447,
		0.996216,
		0.998733,
		0.999749,
		0.999987,
		1,
		0.999959,
		0.999531,
		0.997978,
		0.994328,
		0.987511,
		0.976473,
		0.960253,
		0.938045,
		0.909232,
		0.873411,
		0.830407,
		0.780288,
		0.723368,
		0.660215,
		0.591661,
		0.518807,
		0.443032,
		0.366009,
		0.289713,
		0.216434,
		0.148801,
		0.089791,
		0.042755,
		0.011437,
		0.003739,
		0.014613,
		0.032094,
		0.055645,
		0.084714,
		0.118739,
		0.157147,
		0.199356,
		0.244776,
		0.292812,
		0.342867,
		0.394338,
		0.446628,
		0.499142,
		0.551297,
		0.602518,
		0.652255,
		0.699976,
		0.745185,
		0.787425,
		0.826287,
		0.861423,
		0.892557,
		0.919497,
		0.942154,
		0.960553,
		0.974848,
		0.985332,
		0.992444,
		0.996763,
		0.998974,
		0.999819,
		0.999994,
		1,
		0.999937,
		0.999412,
		0.99766,
		0.993706,
		0.986495,
		0.974996,
		0.95828,
		0.935573,
		0.906287,
		0.870045,
		0.826699,
		0.776333,
		0.719275,
		0.656102,
		0.587648,
		0.515011,
		0.439564,
		0.362964,
		0.287165,
		0.214433,
		0.14736,
		0.088884,
		0.042305,
		0.011312,
		0.000417,
		0.006524,
		0.019517,
		0.038875,
		0.064064,
		0.094539,
		0.129743,
		0.169111,
		0.212067,
		0.258031,
		0.306416,
		0.356632,
		0.40809,
		0.460201,
		0.512384,
		0.564067,
		0.61469,
		0.663715,
		0.710629,
		0.754952,
		0.796242,
		0.834112,
		0.868232,
		0.898348,
		0.924291,
		0.945995,
		0.963506,
		0.977002,
		0.986798,
		0.99335,
		0.997247,
		0.999178,
		0.999873,
		0.999998,
		0.999999,
		0.999907,
		0.999274,
		0.997312,
		0.993045,
		0.985435,
		0.973475,
		0.956268,
		0.933068,
		0.903317,
		0.866667,
		0.82299,
		0.772387,
		0.715201,
		0.652015,
		0.583667,
		0.511251,
		0.436133,
		0.359954,
		0.28465,
		0.212459,
		0.14594,
		0.08799,
		0.041863,
		0.01119,
		0.001638,
		0.010005,
		0.025016,
		0.046154,
		0.072891,
		0.104687,
		0.140991,
		0.181245,
		0.224882,
		0.271328,
		0.320005,
		0.370333,
		0.421733,
		0.473628,
		0.525446,
		0.57663,
		0.626633,
		0.674931,
		0.721027,
		0.764456,
		0.804797,
		0.841678,
		0.874792,
		0.903905,
		0.928869,
		0.94964,
		0.966289,
		0.979014,
		0.98815,
		0.994171,
		0.997675,
		0.99935,
		0.999914,
		0.999999,
		0.999998,
		0.999868,
		0.999115,
		0.996932,
		0.992346,
		0.984332,
		0.971912,
		0.954217,
		0.930531,
		0.900326,
		0.863277,
		0.819279,
		0.76845,
		0.711144,
		0.647954,
		0.579717,
		0.507526,
		0.432738,
		0.35698,
		0.282166,
		0.210511,
		0.14454,
		0.08711,
		0.041428,
		0.011069,
		0.003617,
		0.014141,
		0.031071,
		0.053897,
		0.082095,
		0.11513,
		0.152459,
		0.193529,
		0.237781,
		0.28465,
		0.333566,
		0.383959,
		0.435259,
		0.4869,
		0.538323,
		0.588982,
		0.638345,
		0.685901,
		0.73117,
		0.773702,
		0.813093,
		0.848992,
		0.881109,
		0.909232,
		0.933236,
		0.953097,
		0.968908,
		0.980889,
		0.989394,
		0.994912,
		0.998049,
		0.999493,
		0.999944,
		1,
		0.999994,
		0.999819,
		0.998935,
		0.996521,
		0.991607,
		0.983187,
		0.970307,
		0.952128,
		0.927963,
		0.897313,
		0.859876,
		0.815567,
		0.764523,
		0.707107,
		0.643919,
		0.575799,
		0.503836,
		0.429378,
		0.354039,
		0.279714,
		0.208589,
		0.14316,
		0.086243,
		0.041,
		0.010951,
		0.000403,
		0.006313,
		0.018893,
		0.037648,
		0.062072,
		0.091646,
		0.125842,
		0.164123,
		0.205942,
		0.250747,
		0.297982,
		0.347086,
		0.397498,
		0.448658,
		0.500011,
		0.55101,
		0.601121,
		0.649825,
		0.696627,
		0.741059,
		0.78269,
		0.821134,
		0.856057,
		0.887188,
		0.914337,
		0.937398,
		0.956371,
		0.971369,
		0.982633,
		0.990536,
		0.995579,
		0.998376,
		0.999611,
		0.999965,
		1,
		0.999988,
		0.999758,
		0.998733,
		0.996078,
		0.990828,
		0.981999,
		0.968661,
		0.950003,
		0.925367,
		0.894279,
		0.856465,
		0.811856,
		0.760605,
		0.703088,
		0.63991,
		0.571912,
		0.50018,
		0.426054,
		0.351133,
		0.277292,
		0.206693,
		0.141799,
		0.085389,
		0.040578,
		0.010834,
		0.001584,
		0.009684,
		0.024223,
		0.044712,
		0.070648,
		0.101517,
		0.136799,
		0.17596,
		0.218463,
		0.263763,
		0.31131,
		0.360554,
		0.410941,
		0.461923,
		0.512956,
		0.563503,
		0.613044,
		0.661072,
		0.707107,
		0.750696,
		0.791425,
		0.828924,
		0.862877,
		0.893035,
		0.919224,
		0.941362,
		0.959469,
		0.973679,
		0.984253,
		0.991581,
		0.996177,
		0.998659,
		0.999707,
		0.99998,
		1,
		0.999979,
		0.999684,
		0.998506,
		0.995601,
		0.990011,
		0.98077,
		0.966975,
		0.947842,
		0.922742,
		0.891227,
		0.853044,
		0.808146,
		0.756699,
		0.699089,
		0.635927,
		0.568056,
		0.496559,
		0.422765,
		0.34826,
		0.2749,
		0.204822,
		0.140457,
		0.084547,
		0.040162,
		0.010719,
		0.003501,
		0.013691,
		0.030097,
		0.05223,
		0.079594,
		0.111681,
		0.147975,
		0.18795,
		0.231074,
		0.276811,
		0.324619,
		0.373956,
		0.424278,
		0.475047,
		0.525728,
		0.575799,
		0.624748,
		0.672085,
		0.717343,
		0.760084,
		0.79991,
		0.836467,
		0.869459,
		0.898654,
		0.9239,
		0.945135,
		0.962398,
		0.975845,
		0.985756,
		0.992536,
		0.996711,
		0.998903,
		0.999783,
		0.999989,
		1,
		0.999965,
		0.999596,
		0.998255,
		0.99509,
		0.989154,
		0.979499,
		0.965249,
		0.945646,
		0.92009,
		0.888157,
		0.849616,
		0.804438,
		0.752804,
		0.695109,
		0.631971,
		0.564232,
		0.492972,
		0.41951,
		0.34542,
		0.272538,
		0.202976,
		0.139134,
		0.083718,
		0.039753,
		0.010606,
		0.00039,
		0.006111,
		0.018298,
		0.036479,
		0.060171,
		0.088884,
		0.122113,
		0.159349,
		0.200072,
		0.243759,
		0.289878,
		0.337898,
		0.387283,
		0.437501,
		0.488022,
		0.538323,
		0.587893,
		0.636233,
		0.682865,
		0.727336,
		0.769224,
		0.808146,
		0.843766,
		0.875806,
		0.904051,
		0.928371,
		0.948721,
		0.965164,
		0.977873,
		0.987147,
		0.993406,
		0.997186,
		0.999112,
		0.999844,
		0.999995,
		1,
		0.999946,
		0.999493,
		0.997978,
		0.994545,
		0.988257,
		0.978187,
		0.963484,
		0.943417,
		0.917412,
		0.88507,
		0.84618,
		0.800733,
		0.748921,
		0.69115,
		0.628041,
		0.560439,
		0.489418,
		0.416289,
		0.342613,
		0.270205,
		0.201154,
		0.137829,
		0.0829,
		0.03935,
		0.010495,
		0.001534,
		0.009378,
		0.023468,
		0.043337,
		0.068506,
		0.098489,
		0.132788,
		0.170898,
		0.212308,
		0.256499,
		0.302949,
		0.351133,
		0.400525,
		0.450601,
		0.500843,
		0.550737,
		0.599783,
		0.647496,
		0.69341,
		0.737086,
		0.778118,
		0.816138,
		0.850827,
		0.881923,
		0.909232,
		0.932642,
		0.952128,
		0.967772,
		0.979768,
		0.988431,
		0.994196,
		0.997607,
		0.999289,
		0.99989,
		0.999998,
		0.999999,
		0.99992,
		0.999373,
		0.997675,
		0.993965,
		0.987321,
		0.976835,
		0.961681,
		0.941155,
		0.914709,
		0.881966,
		0.842738,
		0.797031,
		0.74505,
		0.68721,
		0.624138,
		0.556677,
		0.485897,
		0.413102,
		0.339837,
		0.267901,
		0.199356,
		0.136543,
		0.082095,
		0.038954,
		0.010385,
		0.00339,
		0.013263,
		0.029167,
		0.050639,
		0.077206,
		0.108384,
		0.143683,
		0.182603,
		0.224639,
		0.26928,
		0.31601,
		0.364313,
		0.413672,
		0.463572,
		0.513503,
		0.562964,
		0.611467,
		0.658536,
		0.703721,
		0.746597,
		0.78677,
		0.823889,
		0.857652,
		0.887815,
		0.914202,
		0.936718,
		0.955361,
		0.97023,
		0.981538,
		0.989616,
		0.994912,
		0.997978,
		0.999437,
		0.999926,
		0.999999,
		0.999998,
		0.999886,
		0.999236,
		0.997343,
		0.99335,
		0.986346,
		0.975444,
		0.959841,
		0.938862,
		0.911982,
		0.878848,
		0.839291,
		0.793332,
		0.741192,
		0.683291,
		0.620261,
		0.552945,
		0.48241,
		0.409949,
		0.337094,
		0.265625,
		0.197581,
		0.135274,
		0.081301,
		0.038563,
		0.010277,
		0.000378,
		0.00592,
		0.017731,
		0.035363,
		0.058356,
		0.086243,
		0.118545,
		0.154777,
		0.194445,
		0.237049,
		0.282087,
		0.32905,
		0.377429,
		0.426716,
		0.476406,
		0.525998,
		0.575003,
		0.622942,
		0.669354,
		0.7138,
		0.755868,
		0.795181,
		0.831403,
		0.864247,
		0.893487,
		0.918966,
		0.940607,
		0.958426,
		0.972543,
		0.983187,
		0.990705,
		0.995558,
		0.998304,
		0.999561,
		0.999951,
		1,
		0.999995,
		0.999844,
		0.999079,
		0.996984,
		0.9927,
		0.985332,
		0.974012,
		0.957965,
		0.936538,
		0.909232,
		0.875716,
		0.835839,
		0.789639,
		0.737347,
		0.679392,
		0.61641,
		0.549244,
		0.478955,
		0.406828,
		0.334382,
		0.263377,
		0.19583,
		0.134022,
		0.080519,
		0.038178,
		0.010171,
		0.001486,
		0.009086,
		0.022747,
		0.042023,
		0.06646,
		0.095593,
		0.128949,
		0.166048,
		0.206404,
		0.249522,
		0.294906,
		0.342055,
		0.390469,
		0.439648,
		0.489096,
		0.538323,
		0.586849,
		0.634207,
		0.679948,
		0.723646,
		0.764902,
		0.803355,
		0.838682,
		0.870615,
		0.898944,
		0.923529,
		0.944313,
		0.961329,
		0.974716,
		0.984721,
		0.991705,
		0.99614,
		0.998588,
		0.999663,
		0.99997,
		1,
		0.99999,
		0.999791,
		0.998903,
		0.996595,
		0.992015,
		0.984278,
		0.972543,
		0.956054,
		0.934185,
		0.90646,
		0.872571,
		0.832383,
		0.785951,
		0.733515,
		0.675515,
		0.612586,
		0.545574,
		0.475533,
		0.403741,
		0.331701,
		0.261156,
		0.194101,
		0.132788,
		0.079747,
		0.037798,
		0.010067,
		0.003284,
		0.012854,
		0.02828,
		0.049119,
		0.074922,
		0.105229,
		0.139573,
		0.177478,
		0.218463,
		0.262041,
		0.307723,
		0.355015,
		0.403426,
		0.452462,
		0.501639,
		0.550474,
		0.5985,
		0.645261,
		0.690319,
		0.73326,
		0.773702,
		0.811294,
		0.845732,
		0.876762,
		0.904191,
		0.927897,
		0.947842,
		0.964076,
		0.976757,
		0.986146,
		0.992621,
		0.996662,
		0.998834,
		0.999746,
		0.999982,
		1,
		0.999982,
		0.999727,
		0.998706,
		0.996177,
		0.991293,
		0.983187,
		0.971035,
		0.954108,
		0.931803,
		0.903668,
		0.869413,
		0.828924,
		0.782268,
		0.729698,
		0.671658,
		0.608789,
		0.541933,
		0.472143,
		0.400685,
		0.32905,
		0.258963,
		0.192394,
		0.13157,
		0.078987,
		0.037425,
		0.009964,
		0.000366,
		0.005737,
		0.01719,
		0.034297,
		0.056622,
		0.083718,
		0.11513,
		0.150396,
		0.189047,
		0.230606,
		0.274594,
		0.320527,
		0.36792,
		0.416289,
		0.465151,
		0.514027,
		0.562448,
		0.609954,
		0.656102,
		0.700466,
		0.742645,
		0.782268,
		0.819002,
		0.852555,
		0.88269,
		0.909232,
		0.932076,
		0.951199,
		0.966673,
		0.978669,
		0.987468,
		0.993458,
		0.997128,
		0.999046,
		0.999812,
		0.999991,
		1,
		0.99997,
		0.999651,
		0.998486,
		0.995728,
		0.990536,
		0.982057,
		0.969489,
		0.952128,
		0.929393,
		0.900855,
		0.866244,
		0.825463,
		0.778592,
		0.725895,
		0.667822,
		0.605018,
		0.538323,
		0.468785,
		0.397661,
		0.326429,
		0.256796,
		0.19071,
		0.130369,
		0.078237,
		0.037057,
		0.009862,
		0.00144,
		0.008808,
		0.022059,
		0.040769,
		0.064503,
		0.092821,
		0.125272,
		0.161399,
		0.200737,
		0.242817,
		0.287165,
		0.333306,
		0.38076,
		0.429052,
		0.477707,
		0.526257,
		0.57424,
		0.621209,
		0.66673,
		0.71039,
		0.751801,
		0.790604,
		0.826481,
		0.859156,
		0.888407,
		0.914074,
		0.93607,
		0.954391,
		0.969125,
		0.980459,
		0.988691,
		0.99422,
		0.997543,
		0.999227,
		0.999864,
		0.999996,
		1,
		0.999953,
		0.999561,
		0.998244,
		0.995249,
		0.989742,
		0.980889,
		0.967907,
		0.950115,
		0.926957,
		0.898023,
		0.863065,
		0.822,
		0.774923,
		0.722107,
		0.664008,
		0.601273,
		0.534743,
		0.465459,
		0.394668,
		0.323837,
		0.254655,
		0.189047,
		0.129184,
		0.077498,
		0.036694,
		0.009762,
		0.003183,
		0.012464,
		0.027432,
		0.047667,
		0.072738,
		0.102209,
		0.135635,
		0.172561,
		0.212531,
		0.255081,
		0.299743,
		0.346048,
		0.393526,
		0.441707,
		0.490126,
		0.538323,
		0.585848,
		0.632263,
		0.677145,
		0.720092,
		0.76073,
		0.798713,
		0.833736,
		0.865538,
		0.893914,
		0.91872,
		0.939885,
		0.957423,
		0.971437,
		0.982133,
		0.989821,
		0.994912,
		0.99791,
		0.999381,
		0.999905,
		0.999998,
		0.999999,
		0.999931,
		0.999457,
		0.997978,
		0.994738,
		0.988913,
		0.979683,
		0.966289,
		0.948071,
		0.924495,
		0.895174,
		0.859876,
		0.818536,
		0.771261,
		0.718333,
		0.660215,
		0.597555,
		0.531193,
		0.462163,
		0.391707,
		0.321275,
		0.25254,
		0.187405,
		0.128015,
		0.07677,
		0.036336,
		0.009664,
		0.005562,
		0.016673,
		0.033279,
		0.054963,
		0.081301,
		0.111859,
		0.146196,
		0.183866,
		0.224414,
		0.267385,
		0.312315,
		0.358744,
		0.406208,
		0.454246,
		0.502402,
		0.550223,
		0.59727,
		0.643115,
		0.687346,
		0.729573,
		0.769433,
		0.806596,
		0.840769,
		0.871707,
		0.899218,
		0.923176,
		0.943527,
		0.960299,
		0.973615,
		0.983695,
		0.990862,
		0.995539,
		0.998234,
		0.99951,
		0.999935,
		0.999999,
		0.999998,
		0.999901,
		0.999337,
		0.997688,
		0.994196,
		0.988048,
		0.97844,
		0.964635,
		0.945995,
		0.922008,
		0.892307,
		0.856678,
		0.815072,
		0.767608,
		0.714575,
		0.656444,
		0.593863,
		0.527672,
		0.458899,
		0.388776,
		0.318742,
		0.25045,
		0.185784,
		0.126861,
		0.076051,
		0.035984,
		0.009567,
		0.021402,
		0.039569,
		0.062632,
		0.090168,
		0.121749,
		0.156939,
		0.195295,
		0.236371,
		0.279714,
		0.32487,
		0.371384,
		0.4188,
		0.466665,
		0.514529,
		0.561953,
		0.608503,
		0.653763,
		0.697333,
		0.738833,
		0.777914,
		0.814257,
		0.847584,
		0.877665,
		0.904322,
		0.927446,
		0.946999,
		0.963026,
		0.975665,
		0.985151,
		0.991821,
		0.996106,
		0.998518,
		0.999618,
		0.999958,
		1,
		0.999996,
		0.999864,
		0.9992,
		0.997372,
		0.993622,
		0.987147,
		0.977161,
		0.962947,
		0.943889,
		0.919497,
		0.889423,
		0.853472,
		0.811609,
		0.763962,
		0.710833,
		0.652694,
		0.590198,
		0.52418,
		0.455665,
		0.385876,
		0.316237,
		0.248386,
		0.184184,
		0.125723,
		0.075343,
		0.035637,
		0.009472,
		0.046277,
		0.070648,
		0.099316,
		0.131859,
		0.167843,
		0.206833,
		0.248386,
		0.292057,
		0.337397,
		0.383959,
		0.431293,
		0.478955,
		0.526505,
		0.573509,
		0.619546,
		0.664208,
		0.707107,
		0.747875,
		0.786174,
		0.821699,
		0.854185,
		0.883417,
		0.909232,
		0.931537,
		0.950308,
		0.965609,
		0.977591,
		0.986506,
		0.9927,
		0.996615,
		0.998766,
		0.999707,
		0.999974,
		1,
		0.999991,
		0.999819,
		0.999046,
		0.99703,
		0.993016,
		0.98621,
		0.975845,
		0.961225,
		0.941754,
		0.916963,
		0.886525,
		0.850259,
		0.808146,
		0.760326,
		0.707107,
		0.648967,
		0.586559,
		0.520718,
		0.452462,
		0.383005,
		0.31376,
		0.246346,
		0.182603,
		0.1246,
		0.074644,
		0.035295,
		0.009378,
		0.078987,
		0.108724,
		0.142168,
		0.178891,
		0.218463,
		0.260446,
		0.304401,
		0.349886,
		0.39646,
		0.443682,
		0.491113,
		0.538323,
		0.584888,
		0.630396,
		0.674449,
		0.716668,
		0.756699,
		0.794215,
		0.828924,
		0.860575,
		0.888967,
		0.913952,
		0.935451,
		0.953459,
		0.968053,
		0.9794,
		0.987764,
		0.993506,
		0.997073,
		0.998981,
		0.999779,
		0.999985,
		1,
		0.999984,
		0.999763,
		0.998873,
		0.996662,
		0.992377,
		0.985237,
		0.974494,
		0.959469,
		0.93959,
		0.914407,
		0.883611,
		0.84704,
		0.804685,
		0.756699,
		0.703397,
		0.645261,
		0.582946,
		0.517284,
		0.449289,
		0.380165,
		0.31131,
		0.24433,
		0.181043,
		0.123492,
		0.073955,
		0.034957,
		0.009285,
		0.118371,
		0.152659,
		0.190067,
		0.230171,
		0.272538,
		0.316735,
		0.362328,
		0.40888,
		0.455958,
		0.503133,
		0.549982,
		0.596089,
		0.641052,
		0.684484,
		0.726017,
		0.765307,
		0.80204,
		0.835936,
		0.866759,
		0.894319,
		0.918486,
		0.939196,
		0.956457,
		0.970362,
		0.981094,
		0.988931,
		0.994242,
		0.997481,
		0.999166,
		0.999836,
		0.999992,
		1,
		0.999974,
		0.999696,
		0.998681,
		0.996266,
		0.991705,
		0.98423,
		0.973107,
		0.957681,
		0.937398,
		0.91183,
		0.880684,
		0.843815,
		0.801227,
		0.753082,
		0.699703,
		0.641577,
		0.579359,
		0.51388,
		0.446146,
		0.377353,
		0.308888,
		0.242338,
		0.179502,
		0.122398,
		0.073275,
		0.034625,
		0.009194,
		0.163312,
		0.201355,
		0.241943,
		0.28465,
		0.32905,
		0.374712,
		0.42121,
		0.468117,
		0.515011,
		0.561477,
		0.607109,
		0.651515,
		0.694316,
		0.735156,
		0.773702,
		0.809651,
		0.842738,
		0.872739,
		0.899478,
		0.92284,
		0.942774,
		0.959306,
		0.972543,
		0.982681,
		0.990011,
		0.994912,
		0.997845,
		0.999324,
		0.999882,
		0.999996,
		1,
		0.999959,
		0.999618,
		0.998468,
		0.995843,
		0.991001,
		0.983187,
		0.971686,
		0.955861,
		0.935179,
		0.909232,
		0.877744,
		0.840584,
		0.797771,
		0.749475,
		0.696026,
		0.637915,
		0.575799,
		0.510503,
		0.443032,
		0.37457,
		0.306492,
		0.24037,
		0.177981,
		0.121319,
		0.072605,
		0.034297,
		0.009104,
		0.212739,
		0.253765,
		0.29677,
		0.341334,
		0.387032,
		0.433445,
		0.480153,
		0.526742,
		0.572806,
		0.617947,
		0.661782,
		0.703943,
		0.744084,
		0.781884,
		0.817052,
		0.849334,
		0.878519,
		0.904447,
		0.927017,
		0.946192,
		0.962012,
		0.974599,
		0.984164,
		0.991008,
		0.995521,
		0.998167,
		0.999459,
		0.999917,
		0.999998,
		0.999999,
		0.99994,
		0.999527,
		0.998234,
		0.995392,
		0.990264,
		0.982109,
		0.97023,
		0.954009,
		0.932935,
		0.906615,
		0.874792,
		0.83735,
		0.794318,
		0.745878,
		0.692366,
		0.634275,
		0.572264,
		0.507156,
		0.439948,
		0.371816,
		0.304124,
		0.238425,
		0.176478,
		0.120254,
		0.071944,
		0.033974,
		0.009016,
		0.265625,
		0.308888,
		0.353578,
		0.39928,
		0.445578,
		0.492061,
		0.538323,
		0.583966,
		0.628601,
		0.671853,
		0.713366,
		0.752804,
		0.789856,
		0.824244,
		0.855726,
		0.884104,
		0.909232,
		0.931022,
		0.949453,
		0.964579,
		0.976536,
		0.985548,
		0.991928,
		0.996073,
		0.998451,
		0.999572,
		0.999944,
		1,
		0.999998,
		0.999914,
		0.999422,
		0.997978,
		0.994912,
		0.989494,
		0.980996,
		0.968741,
		0.952128,
		0.930665,
		0.903979,
		0.871829,
		0.834112,
		0.790869,
		0.742293,
		0.688723,
		0.630658,
		0.568755,
		0.503836,
		0.436892,
		0.369091,
		0.301781,
		0.236502,
		0.174994,
		0.119202,
		0.071291,
		0.033655,
		0.008928,
		0.320992,
		0.365774,
		0.411446,
		0.457602,
		0.503836,
		0.54975,
		0.594955,
		0.639069,
		0.681729,
		0.722587,
		0.761317,
		0.79762,
		0.83123,
		0.861918,
		0.889497,
		0.913836,
		0.93486,
		0.952563,
		0.967013,
		0.978359,
		0.986838,
		0.992774,
		0.996571,
		0.9987,
		0.999666,
		0.999963,
		1,
		0.999996,
		0.999882,
		0.999302,
		0.9977,
		0.994403,
		0.988691,
		0.97985,
		0.967218,
		0.950217,
		0.928371,
		0.901325,
		0.868855,
		0.83087,
		0.787425,
		0.738718,
		0.685097,
		0.627063,
		0.565272,
		0.500544,
		0.433865,
		0.366393,
		0.299464,
		0.234603,
		0.173528,
		0.118164,
		0.070648,
		0.033341,
		0.008842,
		0.377913,
		0.423526,
		0.469512,
		0.515474,
		0.561021,
		0.60577,
		0.649351,
		0.691409,
		0.731605,
		0.769624,
		0.805179,
		0.838015,
		0.867913,
		0.894703,
		0.918264,
		0.938536,
		0.955526,
		0.969317,
		0.980072,
		0.988039,
		0.993551,
		0.99702,
		0.998917,
		0.999743,
		0.999977,
		1,
		0.999992,
		0.999842,
		0.999166,
		0.997398,
		0.993865,
		0.987855,
		0.978669,
		0.965663,
		0.948277,
		0.926053,
		0.898654,
		0.865871,
		0.827626,
		0.783986,
		0.735156,
		0.681489,
		0.62349,
		0.561815,
		0.497281,
		0.430867,
		0.363722,
		0.297172,
		0.232725,
		0.172081,
		0.11714,
		0.070013,
		0.033031,
		0.008758,
		0.435513,
		0.481304,
		0.526971,
		0.572131,
		0.61641,
		0.659446,
		0.700893,
		0.740422,
		0.777728,
		0.812535,
		0.8446,
		0.873716,
		0.899725,
		0.922519,
		0.942053,
		0.958348,
		0.971498,
		0.98168,
		0.989154,
		0.994262,
		0.997422,
		0.999106,
		0.999806,
		0.999987,
		1,
		0.999986,
		0.999793,
		0.999014,
		0.997073,
		0.993298,
		0.986986,
		0.977455,
		0.964076,
		0.946308,
		0.923712,
		0.895967,
		0.862877,
		0.824381,
		0.780552,
		0.731605,
		0.677898,
		0.619939,
		0.558383,
		0.494044,
		0.427896,
		0.361079,
		0.294906,
		0.230869,
		0.170651,
		0.116129,
		0.069387,
		0.032726,
		0.008674,
		0.492972,
		0.538323,
		0.583079,
		0.626874,
		0.669354,
		0.710181,
		0.749038,
		0.78563,
		0.819691,
		0.850988,
		0.879329,
		0.904566,
		0.926607,
		0.945417,
		0.961031,
		0.973558,
		0.983187,
		0.990188,
		0.994912,
		0.997782,
		0.999268,
		0.999857,
		0.999993,
		1,
		0.999977,
		0.999735,
		0.998845,
		0.996723,
		0.9927,
		0.986085,
		0.976208,
		0.962458,
		0.944313,
		0.92135,
		0.893264,
		0.859876,
		0.821134,
		0.777124,
		0.728066,
		0.674326,
		0.61641,
		0.554977,
		0.490835,
		0.424954,
		0.358463,
		0.292664,
		0.229035,
		0.169238,
		0.11513,
		0.068768,
		0.032424,
		0.008592,
		0.549528,
		0.593863,
		0.63716,
		0.679074,
		0.719275,
		0.757456,
		0.793332,
		0.826649,
		0.857183,
		0.884756,
		0.909232,
		0.930531,
		0.948631,
		0.963582,
		0.975504,
		0.984597,
		0.991144,
		0.995504,
		0.998102,
		0.999407,
		0.999897,
		0.999997,
		1,
		0.999965,
		0.999666,
		0.998657,
		0.996348,
		0.992073,
		0.985151,
		0.974928,
		0.96081,
		0.94229,
		0.918966,
		0.890546,
		0.856866,
		0.817887,
		0.773702,
		0.72454,
		0.670771,
		0.612904,
		0.551596,
		0.487653,
		0.422038,
		0.355873,
		0.290446,
		0.227222,
		0.167843,
		0.114144,
		0.068159,
		0.032127,
		0.00851,
		0.604481,
		0.647268,
		0.688606,
		0.728175,
		0.765677,
		0.800837,
		0.833411,
		0.863189,
		0.890001,
		0.913726,
		0.934295,
		0.951701,
		0.966004,
		0.977338,
		0.985916,
		0.992029,
		0.996042,
		0.998385,
		0.999525,
		0.999927,
		0.999999,
		1,
		0.999947,
		0.999586,
		0.998451,
		0.995947,
		0.991416,
		0.984185,
		0.973615,
		0.959131,
		0.940242,
		0.916561,
		0.887815,
		0.85385,
		0.814639,
		0.770286,
		0.721027,
		0.667234,
		0.60942,
		0.54824,
		0.484498,
		0.41915,
		0.353309,
		0.288253,
		0.225429,
		0.166465,
		0.113171,
		0.067557,
		0.031834,
		0.00843,
		0.657196,
		0.69795,
		0.736881,
		0.773702,
		0.808146,
		0.839981,
		0.869007,
		0.895068,
		0.918052,
		0.937904,
		0.95463,
		0.968302,
		0.979066,
		0.987147,
		0.992844,
		0.996529,
		0.998635,
		0.999624,
		0.999951,
		1,
		0.999999,
		0.999925,
		0.999493,
		0.998225,
		0.995521,
		0.990728,
		0.983187,
		0.972271,
		0.957423,
		0.938168,
		0.914137,
		0.88507,
		0.850827,
		0.811392,
		0.766878,
		0.717527,
		0.663715,
		0.605958,
		0.54491,
		0.48137,
		0.416289,
		0.350772,
		0.286083,
		0.223658,
		0.165103,
		0.11221,
		0.066963,
		0.031545,
		0.008351,
		0.707107,
		0.745396,
		0.781532,
		0.815262,
		0.846361,
		0.874642,
		0.899959,
		0.922214,
		0.941362,
		0.957423,
		0.97048,
		0.980692,
		0.988294,
		0.993593,
		0.99697,
		0.998854,
		0.999707,
		0.999968,
		1,
		0.999997,
		0.999897,
		0.999388,
		0.997978,
		0.995069,
		0.990011,
		0.982156,
		0.970896,
		0.955686,
		0.93607,
		0.911694,
		0.882312,
		0.847798,
		0.808146,
		0.763477,
		0.71404,
		0.660215,
		0.602518,
		0.541604,
		0.478268,
		0.413455,
		0.34826,
		0.283937,
		0.221906,
		0.163757,
		0.111261,
		0.066377,
		0.031259,
		0.008273,
		0.753719,
		0.789171,
		0.822187,
		0.852555,
		0.880097,
		0.90468,
		0.926215,
		0.944673,
		0.960083,
		0.972543,
		0.98222,
		0.989361,
		0.994282,
		0.997366,
		0.999046,
		0.999775,
		0.99998,
		1,
		0.999993,
		0.999861,
		0.999268,
		0.997711,
		0.99459,
		0.989263,
		0.981094,
		0.969489,
		0.953921,
		0.933948,
		0.909232,
		0.879542,
		0.844764,
		0.804901,
		0.760084,
		0.710566,
		0.656733,
		0.599101,
		0.538323,
		0.475193,
		0.410647,
		0.345773,
		0.281814,
		0.220175,
		0.162428,
		0.110324,
		0.065798,
		0.030978,
		0.008197,
		0.796619,
		0.828924,
		0.858565,
		0.885375,
		0.909232,
		0.930061,
		0.947842,
		0.962616,
		0.974494,
		0.983655,
		0.990352,
		0.994912,
		0.997721,
		0.999212,
		0.99983,
		0.999988,
		1,
		0.999988,
		0.999819,
		0.999134,
		0.997422,
		0.994084,
		0.988485,
		0.980001,
		0.968053,
		0.952128,
		0.931803,
		0.906753,
		0.876762,
		0.841725,
		0.801659,
		0.756699,
		0.707107,
		0.65327,
		0.595706,
		0.535067,
		0.472143,
		0.407865,
		0.343311,
		0.279714,
		0.218463,
		0.161115,
		0.109399,
		0.065227,
		0.0307,
		0.008121,
		0.835475,
		0.864394,
		0.89048,
		0.913621,
		0.933754,
		0.950871,
		0.965026,
		0.976338,
		0.984999,
		0.991272,
		0.995488,
		0.998039,
		0.999355,
		0.999874,
		0.999994,
		1,
		0.99998,
		0.999767,
		0.998984,
		0.997111,
		0.993551,
		0.987677,
		0.978877,
		0.966586,
		0.950308,
		0.929636,
		0.904257,
		0.87397,
		0.838682,
		0.798418,
		0.753323,
		0.703661,
		0.649825,
		0.592333,
		0.531836,
		0.469119,
		0.405109,
		0.340874,
		0.277636,
		0.21677,
		0.159817,
		0.108485,
		0.064663,
		0.030427,
		0.008046,
		0.870045,
		0.895415,
		0.91785,
		0.937299,
		0.953766,
		0.967315,
		0.978079,
		0.986257,
		0.992122,
		0.996013,
		0.998321,
		0.999477,
		0.999909,
		0.999997,
		1,
		0.999969,
		0.999707,
		0.998818,
		0.996778,
		0.992991,
		0.986838,
		0.977722,
		0.96509,
		0.948463,
		0.927446,
		0.901745,
		0.871169,
		0.835636,
		0.795181,
		0.749955,
		0.70023,
		0.646399,
		0.588982,
		0.528629,
		0.466121,
		0.402378,
		0.338462,
		0.27558,
		0.215097,
		0.158534,
		0.107582,
		0.064107,
		0.030156,
		0.007973,
		0.900183,
		0.921922,
		0.940699,
		0.95653,
		0.969489,
		0.97972,
		0.987433,
		0.992908,
		0.99649,
		0.998572,
		0.999581,
		0.999936,
		0.999999,
		1,
		0.999954,
		0.999636,
		0.998635,
		0.996422,
		0.992403,
		0.98597,
		0.976536,
		0.963565,
		0.946591,
		0.925236,
		0.899218,
		0.868358,
		0.832587,
		0.791947,
		0.746597,
		0.696813,
		0.642991,
		0.585653,
		0.525446,
		0.463149,
		0.399673,
		0.336073,
		0.273547,
		0.213443,
		0.157267,
		0.106691,
		0.063557,
		0.02989,
		0.0079,
		0.925841,
		0.943959,
		0.959167,
		0.971552,
		0.981267,
		0.988531,
		0.993633,
		0.996922,
		0.998793,
		0.999668,
		0.999957,
		1,
		0.999999,
		0.999934,
		0.999554,
		0.998435,
		0.996042,
		0.991788,
		0.985072,
		0.975321,
		0.962012,
		0.944695,
		0.923006,
		0.896676,
		0.865538,
		0.829535,
		0.788716,
		0.743248,
		0.69341,
		0.639602,
		0.582346,
		0.522288,
		0.460201,
		0.396993,
		0.333709,
		0.271535,
		0.211807,
		0.156015,
		0.10581,
		0.063015,
		0.029626,
		0.007828,
		0.947082,
		0.961681,
		0.973507,
		0.982721,
		0.989554,
		0.9943,
		0.997312,
		0.998987,
		0.999741,
		0.999972,
		1,
		0.999997,
		0.999909,
		0.999461,
		0.998216,
		0.995638,
		0.991144,
		0.984144,
		0.974075,
		0.960431,
		0.942774,
		0.920755,
		0.894119,
		0.862711,
		0.826481,
		0.78549,
		0.739908,
		0.690022,
		0.636233,
		0.579062,
		0.519153,
		0.457278,
		0.394338,
		0.331368,
		0.269544,
		0.210189,
		0.154777,
		0.104941,
		0.06248,
		0.029367,
		0.007758,
		0.964076,
		0.975358,
		0.984089,
		0.990506,
		0.994912,
		0.997663,
		0.999156,
		0.999801,
		0.999982,
		1,
		0.999994,
		0.999878,
		0.999355,
		0.997978,
		0.995211,
		0.990474,
		0.983187,
		0.972801,
		0.958822,
		0.94083,
		0.918486,
		0.891549,
		0.859876,
		0.823426,
		0.782268,
		0.736579,
		0.686649,
		0.632882,
		0.575799,
		0.516043,
		0.454381,
		0.391707,
		0.32905,
		0.267574,
		0.208589,
		0.153554,
		0.104082,
		0.061951,
		0.02911,
		0.007688,
		0.977109,
		0.985372,
		0.99139,
		0.995473,
		0.997978,
		0.999303,
		0.99985,
		0.99999,
		1,
		0.999989,
		0.99984,
		0.999236,
		0.997721,
		0.994759,
		0.989775,
		0.9822,
		0.971498,
		0.957188,
		0.938862,
		0.916199,
		0.888967,
		0.857034,
		0.82037,
		0.779051,
		0.73326,
		0.683291,
		0.629549,
		0.572558,
		0.512956,
		0.451507,
		0.3891,
		0.326755,
		0.265625,
		0.207007,
		0.152345,
		0.103233,
		0.061429,
		0.028857,
		0.007619,
		0.986574,
		0.99221,
		0.995985,
		0.99826,
		0.999429,
		0.999889,
		0.999995,
		1,
		0.999982,
		0.999795,
		0.999103,
		0.997444,
		0.994282,
		0.989048,
		0.981184,
		0.970166,
		0.955526,
		0.936872,
		0.913893,
		0.886372,
		0.854185,
		0.817314,
		0.77584,
		0.729952,
		0.679948,
		0.626236,
		0.569338,
		0.509893,
		0.448658,
		0.386518,
		0.324482,
		0.263696,
		0.205443,
		0.151149,
		0.102394,
		0.060913,
		0.028608,
		0.007551,
		0.992969,
		0.996452,
		0.998511,
		0.999537,
		0.99992,
		0.999997,
		1,
		0.999973,
		0.999741,
		0.998956,
		0.997147,
		0.99378,
		0.988294,
		0.98014,
		0.968806,
		0.95384,
		0.93486,
		0.911571,
		0.883765,
		0.851331,
		0.814257,
		0.772634,
		0.726654,
		0.67662,
		0.622942,
		0.56614,
		0.506853,
		0.445833,
		0.383959,
		0.322233,
		0.261788,
		0.203896,
		0.149968,
		0.101566,
		0.060404,
		0.028361,
		0.007484,
		0.996876,
		0.998733,
		0.999629,
		0.999944,
		0.999999,
		1,
		0.999959,
		0.999679,
		0.998793,
		0.996829,
		0.993253,
		0.987511,
		0.979066,
		0.967419,
		0.952128,
		0.932827,
		0.909232,
		0.881147,
		0.848471,
		0.811201,
		0.769433,
		0.723368,
		0.673308,
		0.619667,
		0.562964,
		0.503836,
		0.443032,
		0.381424,
		0.320005,
		0.2599,
		0.202366,
		0.148801,
		0.100747,
		0.059902,
		0.028118,
		0.007418,
		0.998929,
		0.999707,
		0.999962,
		1,
		0.999999,
		0.999942,
		0.999606,
		0.998614,
		0.99649,
		0.9927,
		0.986701,
		0.977965,
		0.966004,
		0.950392,
		0.930774,
		0.906878,
		0.878519,
		0.845607,
		0.808146,
		0.76624,
		0.720092,
		0.670011,
		0.61641,
		0.55981,
		0.500843,
		0.440255,
		0.378911,
		0.317799,
		0.258031,
		0.200852,
		0.147646,
		0.099939,
		0.059405,
		0.027878,
		0.007353,
		0.999771,
		0.999975,
		1,
		0.999997,
		0.99992,
		0.999523,
		0.99842,
		0.996128,
		0.992122,
		0.985864,
		0.976835,
		0.964563,
		0.948631,
		0.9287,
		0.904508,
		0.875881,
		0.842738,
		0.805092,
		0.763052,
		0.716828,
		0.66673,
		0.613173,
		0.556677,
		0.497872,
		0.437501,
		0.376422,
		0.315615,
		0.256181,
		0.199356,
		0.146505,
		0.09914,
		0.058915,
		0.027641,
		0.007289,
		0.999985,
		1,
		0.999995,
		0.999892,
		0.999429,
		0.998208,
		0.995745,
		0.991519,
		0.984999,
		0.975678,
		0.963095,
		0.946848,
		0.926607,
		0.902123,
		0.873234,
		0.839866,
		0.80204,
		0.759872,
		0.713576,
		0.663465,
		0.609954,
		0.553565,
		0.494924,
		0.43477,
		0.373956,
		0.313452,
		0.254351,
		0.197875,
		0.145377,
		0.09835,
		0.05843,
		0.027406,
		0.007225,
		1,
		0.999991,
		0.999859,
		0.999323,
		0.997978,
		0.99534,
		0.99089,
		0.984106,
		0.974494,
		0.961602,
		0.945041,
		0.924495,
		0.899725,
		0.870578,
		0.83699,
		0.79899,
		0.756699,
		0.710335,
		0.660215,
		0.606755,
		0.550474,
		0.491999,
		0.432063,
		0.371512,
		0.31131,
		0.25254,
		0.196411,
		0.144262,
		0.09757,
		0.057952,
		0.027175,
		0.007163,
		0.999985,
		0.999819,
		0.999205,
		0.997731,
		0.994912,
		0.990235,
		0.983187,
		0.973282,
		0.960083,
		0.943213,
		0.922365,
		0.897313,
		0.867913,
		0.834112,
		0.795942,
		0.753533,
		0.707107,
		0.656981,
		0.603574,
		0.547405,
		0.489096,
		0.429378,
		0.369091,
		0.309189,
		0.250747,
		0.194962,
		0.14316,
		0.096799,
		0.057479,
		0.026947,
		0.007101,
		0.999771,
		0.999074,
		0.997465,
		0.994461,
		0.989554,
		0.98224,
		0.972044,
		0.95854,
		0.941362,
		0.920217,
		0.894888,
		0.865241,
		0.83123,
		0.792898,
		0.750376,
		0.70389,
		0.653763,
		0.600413,
		0.544357,
		0.486216,
		0.426716,
		0.366691,
		0.307089,
		0.248973,
		0.193529,
		0.142069,
		0.096037,
		0.057012,
		0.026722,
		0.00704,
		0.998929,
		0.99718,
		0.993987,
		0.988848,
		0.981267,
		0.97078,
		0.956972,
		0.939491,
		0.918052,
		0.89245,
		0.862562,
		0.828347,
		0.789856,
		0.747226,
		0.700686,
		0.650561,
		0.59727,
		0.54133,
		0.483358,
		0.424076,
		0.364313,
		0.305009,
		0.247217,
		0.192112,
		0.140991,
		0.095283,
		0.056551,
		0.026499,
		0.006979,
		0.996876,
		0.99349,
		0.988116,
		0.980266,
		0.969489,
		0.955381,
		0.937598,
		0.91587,
		0.890001,
		0.859876,
		0.825463,
		0.786818,
		0.744084,
		0.697495,
		0.647376,
		0.594146,
		0.538323,
		0.480522,
		0.421459,
		0.361957,
		0.302949,
		0.245479,
		0.19071,
		0.139926,
		0.094539,
		0.056095,
		0.026279,
		0.00692,
		0.992969,
		0.987358,
		0.97924,
		0.968173,
		0.953766,
		0.935686,
		0.913673,
		0.887541,
		0.857183,
		0.822577,
		0.783784,
		0.740951,
		0.694316,
		0.644206,
		0.591042,
		0.535338,
		0.477707,
		0.418863,
		0.359622,
		0.300909,
		0.243759,
		0.189322,
		0.138872,
		0.093803,
		0.055645,
		0.026062,
		0.006861,
		0.986574,
		0.978187,
		0.966832,
		0.952128,
		0.933754,
		0.91146,
		0.88507,
		0.854485,
		0.819691,
		0.780754,
		0.737827,
		0.69115,
		0.641052,
		0.587956,
		0.532373,
		0.474914,
		0.416289,
		0.357308,
		0.298888,
		0.242056,
		0.18795,
		0.137829,
		0.093076,
		0.0552,
		0.025848,
		0.006803,
		0.977109,
		0.965467,
		0.950468,
		0.931803,
		0.909232,
		0.882588,
		0.851782,
		0.816804,
		0.777728,
		0.734711,
		0.687996,
		0.637915,
		0.584888,
		0.529428,
		0.472143,
		0.413737,
		0.355015,
		0.296887,
		0.24037,
		0.186592,
		0.136799,
		0.092357,
		0.054761,
		0.025636,
		0.006746,
		0.964076,
		0.948785,
		0.929834,
		0.90699,
		0.880097,
		0.849074,
		0.813918,
		0.774707,
		0.731605,
		0.684856,
		0.634794,
		0.58184,
		0.526505,
		0.469393,
		0.411206,
		0.352743,
		0.294906,
		0.238701,
		0.185248,
		0.135779,
		0.091646,
		0.054327,
		0.025427,
		0.006689,
		0.947082,
		0.927846,
		0.904734,
		0.877597,
		0.846361,
		0.811031,
		0.771692,
		0.728508,
		0.681729,
		0.631689,
		0.57881,
		0.523601,
		0.466665,
		0.408697,
		0.350491,
		0.292943,
		0.237049,
		0.183919,
		0.134771,
		0.090944,
		0.053897,
		0.02522,
		0.006633,
		0.925841,
		0.902465,
		0.875088,
		0.843645,
		0.808146,
		0.768681,
		0.72542,
		0.678615,
		0.628601,
		0.575799,
		0.520718,
		0.463957,
		0.406208,
		0.34826,
		0.290998,
		0.235414,
		0.182603,
		0.133774,
		0.090249,
		0.053473,
		0.025016,
		0.006578,
		0.900183,
		0.872571,
		0.840925,
		0.805262,
		0.765677,
		0.722343,
		0.675515,
		0.625529,
		0.572806,
		0.517854,
		0.46127,
		0.403741,
		0.346048,
		0.289073,
		0.233795,
		0.181302,
		0.132788,
		0.089563,
		0.053054,
		0.024814,
		0.006524,
		0.870045,
		0.838201,
		0.802379,
		0.762678,
		0.719275,
		0.672427,
		0.622473,
		0.569832,
		0.515011,
		0.458604,
		0.401293,
		0.343856,
		0.287165,
		0.232193,
		0.180014,
		0.131812,
		0.088884,
		0.05264,
		0.024615,
		0.00647,
		0.835475,
		0.799498,
		0.759685,
		0.716218,
		0.669354,
		0.619433,
		0.566876,
		0.512188,
		0.455958,
		0.398867,
		0.341684,
		0.285276,
		0.230606,
		0.178739,
		0.130847,
		0.088212,
		0.05223,
		0.024418,
		0.006417,
		0.796619,
		0.756699,
		0.71317,
		0.666294,
		0.61641,
		0.563939,
		0.509384,
		0.453333,
		0.39646,
		0.339531,
		0.283404,
		0.229035,
		0.177478,
		0.129893,
		0.087549,
		0.051826,
		0.024223,
		0.006364,
		0.753719,
		0.710133,
		0.663248,
		0.613403,
		0.561021,
		0.5066,
		0.450728,
		0.394074,
		0.337397,
		0.28155,
		0.227479,
		0.17623,
		0.128949,
		0.086892,
		0.051425,
		0.024031,
		0.006313,
		0.707107,
		0.660215,
		0.610413,
		0.55812,
		0.503836,
		0.448143,
		0.391707,
		0.335282,
		0.279714,
		0.225939,
		0.174994,
		0.128015,
		0.086243,
		0.05103,
		0.023841,
		0.006261,
		0.657196,
		0.607439,
		0.555238,
		0.501091,
		0.445578,
		0.38936,
		0.333186,
		0.277895,
		0.224414,
		0.173771,
		0.127091,
		0.085601,
		0.050639,
		0.023653,
		0.006211,
		0.604481,
		0.552374,
		0.498366,
		0.443032,
		0.387032,
		0.331109,
		0.276092,
		0.222905,
		0.172561,
		0.126177,
		0.084966,
		0.050253,
		0.023468,
		0.006161,
		0.549528,
		0.495659,
		0.440506,
		0.384724,
		0.32905,
		0.274307,
		0.221409,
		0.171364,
		0.125272,
		0.084339,
		0.049871,
		0.023285,
		0.006111,
		0.492972,
		0.438,
		0.382435,
		0.327009,
		0.272538,
		0.219929,
		0.170178,
		0.124377,
		0.083718,
		0.049493,
		0.023103,
		0.006063,
		0.435513,
		0.380165,
		0.324985,
		0.270786,
		0.218463,
		0.169005,
		0.123492,
		0.083104,
		0.049119,
		0.022924,
		0.006014,
		0.377913,
		0.32298,
		0.269049,
		0.217011,
		0.167843,
		0.122616,
		0.082496,
		0.04875,
		0.022747,
		0.005967,
		0.320992,
		0.267329,
		0.215573,
		0.166693,
		0.121749,
		0.081895,
		0.048385,
		0.022572,
		0.00592,
		0.265625,
		0.214149,
		0.165555,
		0.120891,
		0.081301,
		0.048024,
		0.022399,
		0.005873,
		0.212739,
		0.164428,
		0.120042,
		0.080713,
		0.047667,
		0.022228,
		0.005827,
		0.163312,
		0.119202,
		0.080132,
		0.047314,
		0.022059,
		0.005782,
		0.118371,
		0.079556,
		0.046964,
		0.021892,
		0.005737,
		0.078987,
		0.046619,
		0.021727,
		0.005692,
		0.046277,
		0.021564,
		0.005649,
		0.021402,
		0.005605,
		0.005562,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		0.004312,
		0.019046,
		0.043031,
		0.075048,
		0.113919,
		0.158507,
		0.207724,
		0.260527,
		0.315921,
		0.372963,
		0.430764,
		0.488488,
		0.545361,
		0.600669,
		0.653763,
		0.704067,
		0.751077,
		0.794366,
		0.833597,
		0.868517,
		0.898975,
		0.924919,
		0.946406,
		0.963605,
		0.976801,
		0.986389,
		0.99287,
		0.99683,
		0.998913,
		0.999767,
		0.999984,
		1,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.002256,
		0.014668,
		0.03678,
		0.06733,
		0.105104,
		0.14893,
		0.197681,
		0.25028,
		0.3057,
		0.362964,
		0.421152,
		0.479399,
		0.536902,
		0.592919,
		0.646778,
		0.697877,
		0.745689,
		0.78977,
		0.82976,
		0.865395,
		0.896506,
		0.923032,
		0.945021,
		0.96264,
		0.976169,
		0.986009,
		0.992666,
		0.996737,
		0.99888,
		0.99976,
		0.999984,
		1,
		0.999984,
		0.999767,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998913,
		0.000833,
		0.010771,
		0.030875,
		0.059846,
		0.096427,
		0.139409,
		0.187629,
		0.239972,
		0.295377,
		0.352834,
		0.411389,
		0.470147,
		0.528274,
		0.585003,
		0.639632,
		0.691535,
		0.740163,
		0.785049,
		0.825816,
		0.862181,
		0.893962,
		0.921086,
		0.943593,
		0.961642,
		0.975516,
		0.985615,
		0.992454,
		0.996641,
		0.998846,
		0.999753,
		0.999983,
		1,
		0.999984,
		0.99976,
		0.998897,
		0.99683,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999757,
		0.99888,
		0.996784,
		0.99287,
		9.5e-005,
		0.007397,
		0.025353,
		0.052623,
		0.08791,
		0.129963,
		0.17758,
		0.229612,
		0.284958,
		0.342574,
		0.401474,
		0.460729,
		0.519475,
		0.576915,
		0.63232,
		0.685037,
		0.734494,
		0.780201,
		0.821761,
		0.858873,
		0.891341,
		0.919079,
		0.942118,
		0.960612,
		0.97484,
		0.985208,
		0.992235,
		0.996541,
		0.998811,
		0.999745,
		0.999983,
		1,
		0.999983,
		0.999753,
		0.998863,
		0.996737,
		0.992769,
		0.986389,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999983,
		0.999749,
		0.998846,
		0.99669,
		0.992666,
		0.986201,
		0.976801,
		0.004591,
		0.020251,
		0.045692,
		0.079575,
		0.120607,
		0.167548,
		0.219207,
		0.274448,
		0.332189,
		0.391409,
		0.451146,
		0.510503,
		0.568653,
		0.624839,
		0.67838,
		0.728678,
		0.775222,
		0.817592,
		0.855469,
		0.888641,
		0.917009,
		0.940595,
		0.959546,
		0.974141,
		0.984786,
		0.992008,
		0.996437,
		0.998774,
		0.999737,
		0.999982,
		1,
		0.999982,
		0.999745,
		0.998829,
		0.996641,
		0.992561,
		0.986009,
		0.976488,
		0.963605,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998811,
		0.996591,
		0.992454,
		0.985814,
		0.976169,
		0.963126,
		0.946406,
		0.002404,
		0.015609,
		0.039085,
		0.071448,
		0.111363,
		0.157547,
		0.208769,
		0.263854,
		0.321682,
		0.381194,
		0.441396,
		0.501356,
		0.560215,
		0.617185,
		0.671559,
		0.722712,
		0.770107,
		0.813304,
		0.851964,
		0.885858,
		0.914874,
		0.939022,
		0.958445,
		0.973418,
		0.984349,
		0.991772,
		0.996329,
		0.998737,
		0.999729,
		0.999982,
		1,
		0.999982,
		0.999737,
		0.998793,
		0.996541,
		0.992346,
		0.985615,
		0.975845,
		0.96264,
		0.945719,
		0.924919,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999982,
		0.999733,
		0.998774,
		0.99649,
		0.992235,
		0.985413,
		0.975516,
		0.962145,
		0.945021,
		0.923982,
		0.898975,
		0.000888,
		0.011471,
		0.032838,
		0.063557,
		0.102252,
		0.147593,
		0.198309,
		0.253183,
		0.311056,
		0.370832,
		0.431479,
		0.492031,
		0.551596,
		0.609355,
		0.66457,
		0.71659,
		0.764852,
		0.808893,
		0.848355,
		0.882989,
		0.91267,
		0.937398,
		0.957307,
		0.97267,
		0.983896,
		0.991528,
		0.996218,
		0.998697,
		0.99972,
		0.999981,
		1,
		0.999981,
		0.999729,
		0.998756,
		0.996437,
		0.992122,
		0.985208,
		0.975181,
		0.961642,
		0.944313,
		0.923032,
		0.897749,
		0.868517,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998737,
		0.996384,
		0.992008,
		0.984999,
		0.97484,
		0.961131,
		0.943593,
		0.922066,
		0.896506,
		0.866967,
		0.833597,
		0.000101,
		0.007884,
		0.026987,
		0.055933,
		0.093296,
		0.137703,
		0.187838,
		0.242442,
		0.300316,
		0.360324,
		0.421394,
		0.482527,
		0.542794,
		0.601344,
		0.65741,
		0.710309,
		0.759454,
		0.804357,
		0.844638,
		0.880032,
		0.910396,
		0.93572,
		0.95613,
		0.971895,
		0.983428,
		0.991275,
		0.996102,
		0.998656,
		0.999711,
		0.99998,
		1,
		0.999981,
		0.99972,
		0.998717,
		0.996329,
		0.991891,
		0.984786,
		0.974494,
		0.960612,
		0.942861,
		0.921086,
		0.895243,
		0.865395,
		0.831692,
		0.794366,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999716,
		0.998697,
		0.996274,
		0.991772,
		0.984569,
		0.974141,
		0.960083,
		0.942118,
		0.92009,
		0.893962,
		0.863799,
		0.82976,
		0.792083,
		0.751077,
		0.004898,
		0.021574,
		0.048607,
		0.08452,
		0.127896,
		0.177371,
		0.231643,
		0.289469,
		0.349672,
		0.411143,
		0.472842,
		0.533806,
		0.59315,
		0.650073,
		0.703863,
		0.753907,
		0.79969,
		0.84081,
		0.876983,
		0.908049,
		0.933986,
		0.954912,
		0.971093,
		0.982941,
		0.991012,
		0.995981,
		0.998614,
		0.999702,
		0.99998,
		1,
		0.99998,
		0.999711,
		0.998677,
		0.996218,
		0.991651,
		0.984349,
		0.973783,
		0.959546,
		0.941362,
		0.919079,
		0.892661,
		0.862181,
		0.827802,
		0.78977,
		0.7484,
		0.704067,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.99998,
		0.999707,
		0.998656,
		0.99616,
		0.991528,
		0.984125,
		0.973418,
		0.959001,
		0.940595,
		0.918052,
		0.891341,
		0.860539,
		0.825816,
		0.787425,
		0.745689,
		0.700991,
		0.653763,
		0.002567,
		0.016644,
		0.041614,
		0.075953,
		0.118192,
		0.166922,
		0.220793,
		0.278519,
		0.33888,
		0.400724,
		0.462975,
		0.524629,
		0.584767,
		0.642555,
		0.69725,
		0.748208,
		0.794889,
		0.836868,
		0.873839,
		0.905627,
		0.932195,
		0.953653,
		0.970262,
		0.982438,
		0.990739,
		0.995856,
		0.998569,
		0.999692,
		0.999979,
		1,
		0.999979,
		0.999702,
		0.998635,
		0.996102,
		0.991403,
		0.983896,
		0.973047,
		0.958445,
		0.939815,
		0.917009,
		0.890001,
		0.858873,
		0.823803,
		0.785049,
		0.742944,
		0.697877,
		0.650291,
		0.600669,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998614,
		0.996042,
		0.991275,
		0.983664,
		0.97267,
		0.957881,
		0.939022,
		0.91595,
		0.888641,
		0.857183,
		0.821761,
		0.782641,
		0.740163,
		0.694725,
		0.646778,
		0.596815,
		0.545361,
		0.000949,
		0.012242,
		0.034993,
		0.067623,
		0.108615,
		0.156508,
		0.209905,
		0.267474,
		0.327951,
		0.39014,
		0.452924,
		0.51526,
		0.576193,
		0.634852,
		0.690463,
		0.742351,
		0.789949,
		0.832806,
		0.870596,
		0.903125,
		0.930343,
		0.95235,
		0.969402,
		0.981915,
		0.990456,
		0.995726,
		0.998523,
		0.999682,
		0.999978,
		1,
		0.999979,
		0.999692,
		0.998592,
		0.995981,
		0.991144,
		0.983428,
		0.972286,
		0.957307,
		0.938217,
		0.914874,
		0.88726,
		0.855469,
		0.819691,
		0.780201,
		0.737347,
		0.691535,
		0.643226,
		0.592919,
		0.541152,
		0.488488,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999687,
		0.998569,
		0.995919,
		0.991012,
		0.983187,
		0.971895,
		0.956724,
		0.937398,
		0.913781,
		0.885858,
		0.853729,
		0.817592,
		0.777728,
		0.734494,
		0.688306,
		0.639632,
		0.588982,
		0.536902,
		0.483964,
		0.430764,
		0.000108,
		0.008421,
		0.028784,
		0.059563,
		0.099188,
		0.146146,
		0.198992,
		0.256343,
		0.316889,
		0.379391,
		0.442688,
		0.505698,
		0.567424,
		0.62696,
		0.683499,
		0.736332,
		0.784865,
		0.828621,
		0.867251,
		0.900541,
		0.928428,
		0.951001,
		0.96851,
		0.981373,
		0.990162,
		0.995591,
		0.998475,
		0.999671,
		0.999978,
		1,
		0.999978,
		0.999682,
		0.998546,
		0.995856,
		0.990877,
		0.982941,
		0.971498,
		0.95613,
		0.936566,
		0.91267,
		0.884434,
		0.851964,
		0.815463,
		0.775222,
		0.731605,
		0.685037,
		0.635997,
		0.585003,
		0.532609,
		0.479399,
		0.425977,
		0.372963,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999978,
		0.999677,
		0.998523,
		0.995792,
		0.990739,
		0.982692,
		0.971093,
		0.955526,
		0.93572,
		0.911542,
		0.882989,
		0.850172,
		0.813304,
		0.772681,
		0.728678,
		0.681729,
		0.63232,
		0.58098,
		0.528274,
		0.474793,
		0.421152,
		0.36798,
		0.315921,
		0.005236,
		0.023032,
		0.051808,
		0.089938,
		0.135857,
		0.188066,
		0.245133,
		0.3057,
		0.36848,
		0.432268,
		0.495938,
		0.558455,
		0.618874,
		0.676351,
		0.730146,
		0.779633,
		0.824309,
		0.863799,
		0.897873,
		0.926448,
		0.949604,
		0.967586,
		0.980811,
		0.989857,
		0.99545,
		0.998425,
		0.99966,
		0.999977,
		1,
		0.999977,
		0.999671,
		0.998499,
		0.995726,
		0.990599,
		0.982438,
		0.970681,
		0.954912,
		0.93486,
		0.910396,
		0.881522,
		0.848355,
		0.811114,
		0.770107,
		0.725714,
		0.67838,
		0.628601,
		0.576915,
		0.523896,
		0.470147,
		0.416289,
		0.362964,
		0.310823,
		0.260527,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999977,
		0.999666,
		0.998475,
		0.995659,
		0.990456,
		0.982179,
		0.970262,
		0.954288,
		0.933986,
		0.909232,
		0.880032,
		0.84651,
		0.808893,
		0.767497,
		0.722712,
		0.67499,
		0.624839,
		0.572806,
		0.519475,
		0.465459,
		0.411389,
		0.357915,
		0.3057,
		0.255412,
		0.207724,
		0.002747,
		0.017784,
		0.044395,
		0.080895,
		0.125662,
		0.177144,
		0.233856,
		0.294389,
		0.357409,
		0.421662,
		0.48598,
		0.549284,
		0.61059,
		0.669016,
		0.723788,
		0.774249,
		0.819865,
		0.860238,
		0.895116,
		0.9244,
		0.948158,
		0.966629,
		0.980227,
		0.98954,
		0.995304,
		0.998373,
		0.999649,
		0.999976,
		1,
		0.999976,
		0.99966,
		0.998451,
		0.995591,
		0.99031,
		0.981915,
		0.969836,
		0.953653,
		0.933098,
		0.908049,
		0.878519,
		0.844638,
		0.806641,
		0.764852,
		0.71967,
		0.671559,
		0.621034,
		0.568653,
		0.515011,
		0.460729,
		0.40645,
		0.352834,
		0.300551,
		0.25028,
		0.202704,
		0.158507,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.998425,
		0.995521,
		0.990162,
		0.981647,
		0.969402,
		0.953007,
		0.932195,
		0.906848,
		0.876983,
		0.842738,
		0.804357,
		0.762171,
		0.71659,
		0.668086,
		0.617185,
		0.564456,
		0.510503,
		0.455958,
		0.401474,
		0.34772,
		0.295377,
		0.245133,
		0.197681,
		0.153712,
		0.113919,
		0.001017,
		0.013093,
		0.037366,
		0.07209,
		0.115585,
		0.166242,
		0.222522,
		0.282963,
		0.34618,
		0.410871,
		0.475821,
		0.539907,
		0.602103,
		0.661489,
		0.717253,
		0.768706,
		0.815284,
		0.856563,
		0.892267,
		0.922282,
		0.946661,
		0.965636,
		0.979622,
		0.98921,
		0.995152,
		0.998319,
		0.999637,
		0.999975,
		1,
		0.999976,
		0.999649,
		0.9984,
		0.99545,
		0.990011,
		0.981373,
		0.96896,
		0.95235,
		0.931276,
		0.905627,
		0.875423,
		0.84081,
		0.80204,
		0.759454,
		0.713469,
		0.66457,
		0.613292,
		0.560215,
		0.505952,
		0.451146,
		0.39646,
		0.342574,
		0.290179,
		0.239972,
		0.192655,
		0.14893,
		0.109496,
		0.075048,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999643,
		0.998373,
		0.995378,
		0.989857,
		0.981094,
		0.96851,
		0.951681,
		0.930343,
		0.904386,
		0.873839,
		0.838854,
		0.79969,
		0.756699,
		0.710309,
		0.661012,
		0.609355,
		0.555928,
		0.501356,
		0.446292,
		0.391409,
		0.337397,
		0.284958,
		0.234798,
		0.187629,
		0.144161,
		0.105104,
		0.071162,
		0.043031,
		0.000116,
		0.009016,
		0.030765,
		0.063557,
		0.105651,
		0.155379,
		0.211144,
		0.271429,
		0.334797,
		0.399895,
		0.465459,
		0.53032,
		0.593409,
		0.653763,
		0.710536,
		0.763,
		0.810562,
		0.852769,
		0.889324,
		0.92009,
		0.945109,
		0.964606,
		0.978993,
		0.988867,
		0.994994,
		0.998263,
		0.999624,
		0.999974,
		1,
		0.999975,
		0.999637,
		0.998347,
		0.995304,
		0.9897,
		0.980811,
		0.968053,
		0.951001,
		0.929393,
		0.903125,
		0.87223,
		0.836868,
		0.797307,
		0.753907,
		0.707107,
		0.65741,
		0.605372,
		0.551596,
		0.496716,
		0.441396,
		0.38632,
		0.332189,
		0.279714,
		0.229612,
		0.182603,
		0.139409,
		0.100747,
		0.06733,
		0.039864,
		0.019046,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999631,
		0.998319,
		0.995229,
		0.98954,
		0.980522,
		0.967586,
		0.950308,
		0.928428,
		0.901844,
		0.870596,
		0.834852,
		0.794889,
		0.751077,
		0.703863,
		0.653763,
		0.601344,
		0.547218,
		0.492031,
		0.436458,
		0.381194,
		0.32695,
		0.274448,
		0.224414,
		0.17758,
		0.134676,
		0.096427,
		0.063557,
		0.03678,
		0.0168,
		0.004312,
		0.005611,
		0.024641,
		0.055335,
		0.09589,
		0.144575,
		0.199736,
		0.259797,
		0.323265,
		0.388735,
		0.454892,
		0.52052,
		0.584502,
		0.645835,
		0.70363,
		0.757125,
		0.805693,
		0.848853,
		0.886281,
		0.917822,
		0.943502,
		0.963537,
		0.978339,
		0.988511,
		0.994829,
		0.998204,
		0.999611,
		0.999973,
		1,
		0.999974,
		0.999624,
		0.998291,
		0.995152,
		0.989376,
		0.980227,
		0.967112,
		0.949604,
		0.927446,
		0.900541,
		0.868936,
		0.832806,
		0.792437,
		0.748208,
		0.700578,
		0.650073,
		0.59727,
		0.542794,
		0.487302,
		0.431479,
		0.376031,
		0.321682,
		0.269161,
		0.219207,
		0.172561,
		0.129963,
		0.092147,
		0.059846,
		0.033782,
		0.014668,
		0.003208,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999618,
		0.998263,
		0.995074,
		0.98921,
		0.979927,
		0.966629,
		0.948887,
		0.926448,
		0.899218,
		0.867251,
		0.830729,
		0.789949,
		0.745299,
		0.69725,
		0.646337,
		0.59315,
		0.538323,
		0.482527,
		0.426457,
		0.370832,
		0.316383,
		0.263854,
		0.213992,
		0.167548,
		0.125272,
		0.08791,
		0.0562,
		0.030875,
		0.012657,
		0.002256,
		0.002946,
		0.019046,
		0.047464,
		0.086332,
		0.133854,
		0.188314,
		0.248076,
		0.31159,
		0.377394,
		0.444121,
		0.510503,
		0.575379,
		0.637698,
		0.696531,
		0.751077,
		0.800673,
		0.844809,
		0.883135,
		0.915473,
		0.941836,
		0.962429,
		0.977661,
		0.98814,
		0.994657,
		0.998143,
		0.999598,
		0.999972,
		1,
		0.999973,
		0.999611,
		0.998234,
		0.994994,
		0.98904,
		0.979622,
		0.966137,
		0.948158,
		0.925433,
		0.897873,
		0.865538,
		0.828621,
		0.787425,
		0.742351,
		0.693878,
		0.642555,
		0.588982,
		0.533806,
		0.477707,
		0.421394,
		0.365596,
		0.311056,
		0.258527,
		0.208769,
		0.162543,
		0.120607,
		0.083718,
		0.052623,
		0.028064,
		0.010771,
		0.001462,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999605,
		0.998204,
		0.994912,
		0.988867,
		0.97931,
		0.965636,
		0.947416,
		0.9244,
		0.896506,
		0.863799,
		0.826481,
		0.784865,
		0.739362,
		0.690463,
		0.638727,
		0.584767,
		0.529241,
		0.472842,
		0.416289,
		0.360324,
		0.3057,
		0.253183,
		0.203541,
		0.157547,
		0.11597,
		0.079575,
		0.049119,
		0.025353,
		0.009016,
		0.000833,
		0.001092,
		0.014037,
		0.039989,
		0.077011,
		0.123241,
		0.176896,
		0.236278,
		0.299776,
		0.365872,
		0.433143,
		0.500268,
		0.566035,
		0.629348,
		0.689233,
		0.744848,
		0.795497,
		0.840634,
		0.879882,
		0.913042,
		0.940109,
		0.961278,
		0.976956,
		0.987755,
		0.994478,
		0.998079,
		0.999584,
		0.999971,
		1,
		0.999972,
		0.999598,
		0.998174,
		0.994829,
		0.988691,
		0.978993,
		0.965125,
		0.946661,
		0.92335,
		0.895116,
		0.862033,
		0.824309,
		0.782268,
		0.736332,
		0.687003,
		0.634852,
		0.580504,
		0.524629,
		0.467931,
		0.411143,
		0.355015,
		0.300316,
		0.247821,
		0.198309,
		0.152563,
		0.111363,
		0.075484,
		0.045692,
		0.022747,
		0.007397,
		0.000375,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999591,
		0.998143,
		0.994744,
		0.988511,
		0.978669,
		0.964606,
		0.945892,
		0.922282,
		0.893703,
		0.860238,
		0.822103,
		0.779633,
		0.73326,
		0.683499,
		0.63093,
		0.576193,
		0.519969,
		0.462975,
		0.405954,
		0.349672,
		0.294906,
		0.242442,
		0.193074,
		0.147593,
		0.106789,
		0.071448,
		0.042346,
		0.020251,
		0.00592,
		9.5e-005,
		0.000125,
		0.009675,
		0.032959,
		0.067965,
		0.112762,
		0.1655,
		0.224414,
		0.287832,
		0.354174,
		0.421959,
		0.489811,
		0.556466,
		0.620779,
		0.681729,
		0.738434,
		0.790157,
		0.836321,
		0.876517,
		0.910525,
		0.938318,
		0.960083,
		0.976223,
		0.987353,
		0.994292,
		0.998013,
		0.999569,
		0.99997,
		1,
		0.999971,
		0.999584,
		0.998111,
		0.994657,
		0.988327,
		0.978339,
		0.964076,
		0.945109,
		0.921195,
		0.892267,
		0.858415,
		0.819865,
		0.77696,
		0.730146,
		0.679948,
		0.62696,
		0.571833,
		0.51526,
		0.457972,
		0.400724,
		0.344293,
		0.289469,
		0.237049,
		0.187838,
		0.142639,
		0.102252,
		0.067471,
		0.039085,
		0.01787,
		0.004591,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999576,
		0.998079,
		0.994569,
		0.98814,
		0.978003,
		0.963537,
		0.944313,
		0.92009,
		0.890808,
		0.856563,
		0.817592,
		0.774249,
		0.726989,
		0.676351,
		0.622942,
		0.567424,
		0.510503,
		0.452924,
		0.395453,
		0.33888,
		0.284006,
		0.231643,
		0.182603,
		0.137703,
		0.097753,
		0.063557,
		0.035914,
		0.015609,
		0.003417,
		0.006028,
		0.026425,
		0.059234,
		0.102448,
		0.154148,
		0.2125,
		0.275766,
		0.342303,
		0.410568,
		0.479129,
		0.546667,
		0.611985,
		0.674014,
		0.731828,
		0.78465,
		0.831866,
		0.873037,
		0.907917,
		0.936461,
		0.958843,
		0.97546,
		0.986936,
		0.994098,
		0.997943,
		0.999553,
		0.999969,
		1,
		0.99997,
		0.999569,
		0.998046,
		0.994478,
		0.987949,
		0.977661,
		0.962988,
		0.943502,
		0.918966,
		0.889324,
		0.854681,
		0.815284,
		0.771497,
		0.723788,
		0.672708,
		0.618874,
		0.562964,
		0.505698,
		0.447829,
		0.39014,
		0.333432,
		0.278519,
		0.226223,
		0.177371,
		0.132788,
		0.093296,
		0.05971,
		0.032838,
		0.013475,
		0.002404,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999561,
		0.998013,
		0.994386,
		0.987755,
		0.977312,
		0.962429,
		0.942676,
		0.917822,
		0.887815,
		0.852769,
		0.812941,
		0.768706,
		0.720543,
		0.669016,
		0.614757,
		0.558455,
		0.500843,
		0.442688,
		0.384786,
		0.327951,
		0.273008,
		0.220793,
		0.172143,
		0.127896,
		0.088884,
		0.055933,
		0.02986,
		0.011471,
		0.001559,
		0.003169,
		0.020447,
		0.050862,
		0.092332,
		0.142864,
		0.200552,
		0.263588,
		0.330263,
		0.398972,
		0.468221,
		0.536635,
		0.602961,
		0.666082,
		0.725025,
		0.778969,
		0.827263,
		0.869437,
		0.905215,
		0.934534,
		0.957554,
		0.974668,
		0.986501,
		0.993896,
		0.99787,
		0.999537,
		0.999968,
		1,
		0.999969,
		0.999553,
		0.997978,
		0.994292,
		0.987556,
		0.976956,
		0.961859,
		0.941836,
		0.916658,
		0.886281,
		0.850827,
		0.810562,
		0.765874,
		0.717253,
		0.665277,
		0.61059,
		0.553895,
		0.495938,
		0.437501,
		0.379391,
		0.322436,
		0.267474,
		0.215353,
		0.166922,
		0.12303,
		0.08452,
		0.05223,
		0.026987,
		0.009606,
		0.000888,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999545,
		0.997943,
		0.994196,
		0.987353,
		0.976593,
		0.961278,
		0.94098,
		0.915473,
		0.884721,
		0.848853,
		0.808146,
		0.763,
		0.713918,
		0.661489,
		0.606372,
		0.549284,
		0.490984,
		0.432268,
		0.373956,
		0.316889,
		0.261919,
		0.209905,
		0.16171,
		0.118192,
		0.080209,
		0.048607,
		0.024223,
		0.007884,
		0.0004,
		0.001176,
		0.015085,
		0.042898,
		0.082451,
		0.131674,
		0.188586,
		0.251308,
		0.31806,
		0.387171,
		0.457085,
		0.526364,
		0.593702,
		0.657927,
		0.718017,
		0.773107,
		0.822507,
		0.865711,
		0.902416,
		0.932535,
		0.956215,
		0.973843,
		0.986047,
		0.993685,
		0.997795,
		0.99952,
		0.999967,
		1,
		0.999968,
		0.999537,
		0.997907,
		0.994098,
		0.987147,
		0.976223,
		0.960686,
		0.940109,
		0.914268,
		0.883135,
		0.846847,
		0.805693,
		0.760084,
		0.710536,
		0.657651,
		0.602103,
		0.544621,
		0.48598,
		0.426988,
		0.36848,
		0.31131,
		0.256343,
		0.204451,
		0.156508,
		0.113386,
		0.075953,
		0.045066,
		0.021574,
		0.006313,
		0.000101,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999967,
		0.999529,
		0.99787,
		0.993998,
		0.986936,
		0.975845,
		0.960083,
		0.939222,
		0.913042,
		0.881522,
		0.844809,
		0.803203,
		0.757125,
		0.707107,
		0.653763,
		0.597782,
		0.539907,
		0.480925,
		0.421662,
		0.362964,
		0.3057,
		0.250747,
		0.198992,
		0.151319,
		0.108615,
		0.071756,
		0.041614,
		0.019046,
		0.004898,
		0.000135,
		0.010409,
		0.035395,
		0.072845,
		0.120607,
		0.176624,
		0.238939,
		0.3057,
		0.375167,
		0.445718,
		0.515852,
		0.584202,
		0.649542,
		0.710798,
		0.767059,
		0.817592,
		0.861854,
		0.899514,
		0.93046,
		0.954824,
		0.972985,
		0.985575,
		0.993464,
		0.997715,
		0.999502,
		0.999966,
		1,
		0.999966,
		0.99952,
		0.997833,
		0.993896,
		0.98672,
		0.97546,
		0.959469,
		0.938318,
		0.911794,
		0.879882,
		0.842738,
		0.800673,
		0.754123,
		0.70363,
		0.649825,
		0.593409,
		0.535139,
		0.475821,
		0.416289,
		0.357409,
		0.300059,
		0.245133,
		0.193529,
		0.146146,
		0.103881,
		0.067623,
		0.038254,
		0.016644,
		0.003647,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999966,
		0.999511,
		0.997795,
		0.993791,
		0.986501,
		0.975068,
		0.958843,
		0.937398,
		0.910525,
		0.878214,
		0.840634,
		0.798105,
		0.751077,
		0.700105,
		0.645835,
		0.588982,
		0.53032,
		0.470665,
		0.410871,
		0.351814,
		0.294389,
		0.239503,
		0.188066,
		0.140991,
		0.099188,
		0.063557,
		0.034993,
		0.014374,
		0.002567,
		0.006493,
		0.02841,
		0.063557,
		0.109696,
		0.164687,
		0.226495,
		0.293191,
		0.362964,
		0.43412,
		0.505093,
		0.574455,
		0.64092,
		0.703361,
		0.760817,
		0.812511,
		0.857862,
		0.896506,
		0.928306,
		0.953377,
		0.972091,
		0.985083,
		0.993234,
		0.997633,
		0.999484,
		0.999964,
		1,
		0.999965,
		0.999502,
		0.997755,
		0.993685,
		0.986276,
		0.974668,
		0.958205,
		0.936461,
		0.909232,
		0.876517,
		0.838495,
		0.795497,
		0.747985,
		0.696531,
		0.641793,
		0.584502,
		0.525446,
		0.465459,
		0.405406,
		0.34618,
		0.28869,
		0.233856,
		0.182603,
		0.135857,
		0.094539,
		0.059563,
		0.031834,
		0.012242,
		0.001665,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999493,
		0.997715,
		0.993576,
		0.986047,
		0.97426,
		0.957554,
		0.935507,
		0.907917,
		0.874792,
		0.836321,
		0.792848,
		0.744848,
		0.692907,
		0.637698,
		0.579968,
		0.52052,
		0.460201,
		0.399895,
		0.340507,
		0.282963,
		0.228195,
		0.177144,
		0.130747,
		0.089938,
		0.055645,
		0.028784,
		0.010256,
		0.000949,
		0.003417,
		0.022008,
		0.054636,
		0.098976,
		0.1528,
		0.213992,
		0.280543,
		0.350565,
		0.422291,
		0.494086,
		0.564456,
		0.632056,
		0.695699,
		0.754375,
		0.807258,
		0.853729,
		0.893386,
		0.926069,
		0.951873,
		0.971161,
		0.984569,
		0.992994,
		0.997546,
		0.999464,
		0.999963,
		1,
		0.999964,
		0.999484,
		0.997675,
		0.993464,
		0.985814,
		0.973843,
		0.956891,
		0.934534,
		0.906578,
		0.873037,
		0.834112,
		0.790157,
		0.741665,
		0.689233,
		0.63355,
		0.575379,
		0.515539,
		0.454892,
		0.394338,
		0.334797,
		0.277209,
		0.222522,
		0.171689,
		0.125662,
		0.085389,
		0.051808,
		0.025848,
		0.008421,
		0.000428,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999963,
		0.999474,
		0.997633,
		0.99335,
		0.985575,
		0.973418,
		0.956215,
		0.933544,
		0.905215,
		0.871252,
		0.831866,
		0.787425,
		0.738434,
		0.685507,
		0.629348,
		0.570735,
		0.510503,
		0.449532,
		0.388735,
		0.32905,
		0.271429,
		0.216838,
		0.166242,
		0.120607,
		0.080895,
		0.048056,
		0.023032,
		0.006746,
		0.000108,
		0.001269,
		0.016256,
		0.046135,
		0.088485,
		0.140991,
		0.201449,
		0.267766,
		0.337974,
		0.41023,
		0.482827,
		0.554201,
		0.622942,
		0.687806,
		0.747726,
		0.801828,
		0.849448,
		0.890151,
		0.923746,
		0.950308,
		0.970192,
		0.984034,
		0.992743,
		0.997456,
		0.999444,
		0.999962,
		1,
		0.999962,
		0.999464,
		0.99759,
		0.993234,
		0.985332,
		0.972985,
		0.955526,
		0.932535,
		0.903828,
		0.869437,
		0.829583,
		0.78465,
		0.735156,
		0.681729,
		0.625091,
		0.566035,
		0.505413,
		0.444121,
		0.383087,
		0.323265,
		0.265625,
		0.211144,
		0.160804,
		0.115585,
		0.076461,
		0.044395,
		0.020342,
		0.005236,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999962,
		0.999454,
		0.997546,
		0.993115,
		0.985083,
		0.972543,
		0.954824,
		0.931507,
		0.902416,
		0.86759,
		0.827263,
		0.781832,
		0.731828,
		0.677898,
		0.620779,
		0.561279,
		0.500268,
		0.438658,
		0.377394,
		0.317445,
		0.259797,
		0.205443,
		0.155379,
		0.110599,
		0.07209,
		0.04083,
		0.017784,
		0.003901,
		0.000146,
		0.011231,
		0.03811,
		0.078267,
		0.129291,
		0.188885,
		0.254872,
		0.325198,
		0.397939,
		0.471312,
		0.543682,
		0.613572,
		0.679673,
		0.740862,
		0.796212,
		0.845015,
		0.886795,
		0.921332,
		0.94868,
		0.969182,
		0.983475,
		0.992481,
		0.997361,
		0.999423,
		0.99996,
		1,
		0.999961,
		0.999444,
		0.997501,
		0.992994,
		0.984829,
		0.972091,
		0.954108,
		0.93046,
		0.900978,
		0.865711,
		0.824905,
		0.778969,
		0.728452,
		0.674014,
		0.61641,
		0.556466,
		0.495067,
		0.433143,
		0.371655,
		0.31159,
		0.253947,
		0.199736,
		0.149968,
		0.105651,
		0.067787,
		0.037366,
		0.015366,
		0.002747,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.99996,
		0.999434,
		0.997456,
		0.99287,
		0.984569,
		0.971631,
		0.953377,
		0.929393,
		0.899514,
		0.863799,
		0.822507,
		0.776061,
		0.725025,
		0.670076,
		0.611985,
		0.551596,
		0.489811,
		0.427577,
		0.365872,
		0.3057,
		0.248076,
		0.194026,
		0.144575,
		0.100747,
		0.063557,
		0.034009,
		0.013093,
		0.001783,
		0.007014,
		0.030627,
		0.06837,
		0.117733,
		0.176325,
		0.241875,
		0.312242,
		0.385418,
		0.45954,
		0.532897,
		0.603939,
		0.671294,
		0.733775,
		0.790404,
		0.840421,
		0.883312,
		0.918824,
		0.946986,
		0.968129,
		0.982892,
		0.992207,
		0.997262,
		0.9994,
		0.999959,
		1,
		0.999959,
		0.999423,
		0.997409,
		0.992743,
		0.984305,
		0.971161,
		0.952633,
		0.928306,
		0.898023,
		0.861854,
		0.820069,
		0.773107,
		0.721547,
		0.666082,
		0.607502,
		0.546667,
		0.484498,
		0.421959,
		0.360045,
		0.299776,
		0.242186,
		0.188314,
		0.139203,
		0.09589,
		0.059405,
		0.030765,
		0.010974,
		0.001017,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999959,
		0.999412,
		0.997361,
		0.992614,
		0.984034,
		0.970681,
		0.951873,
		0.927198,
		0.896506,
		0.859876,
		0.817592,
		0.770107,
		0.718017,
		0.662033,
		0.602961,
		0.541681,
		0.479129,
		0.416289,
		0.354174,
		0.29382,
		0.236278,
		0.182603,
		0.133854,
		0.091084,
		0.055335,
		0.027641,
		0.009016,
		0.000458,
		0.003696,
		0.023754,
		0.058845,
		0.106356,
		0.163793,
		0.228792,
		0.299116,
		0.372671,
		0.447509,
		0.521839,
		0.594037,
		0.66266,
		0.726459,
		0.784396,
		0.835662,
		0.879698,
		0.916216,
		0.945222,
		0.967032,
		0.982283,
		0.99192,
		0.997158,
		0.999377,
		0.999957,
		1,
		0.999958,
		0.9994,
		0.997312,
		0.992481,
		0.983758,
		0.970192,
		0.951098,
		0.926069,
		0.89496,
		0.857862,
		0.815072,
		0.767059,
		0.714434,
		0.657927,
		0.598361,
		0.536635,
		0.473704,
		0.410568,
		0.34826,
		0.287832,
		0.230353,
		0.176896,
		0.128533,
		0.086332,
		0.051353,
		0.024641,
		0.007225,
		0.000116,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999957,
		0.999389,
		0.997262,
		0.992346,
		0.983475,
		0.969692,
		0.950308,
		0.924919,
		0.893386,
		0.855814,
		0.812511,
		0.763962,
		0.710798,
		0.653763,
		0.593702,
		0.531529,
		0.468221,
		0.404796,
		0.342303,
		0.281814,
		0.224414,
		0.171193,
		0.123241,
		0.08164,
		0.047464,
		0.021774,
		0.005611,
		0.001374,
		0.017568,
		0.049751,
		0.0952,
		0.151319,
		0.21564,
		0.28583,
		0.359701,
		0.435217,
		0.510503,
		0.583858,
		0.653763,
		0.718904,
		0.77818,
		0.830729,
		0.875946,
		0.913505,
		0.943385,
		0.965887,
		0.981647,
		0.991621,
		0.997049,
		0.999352,
		0.999955,
		1,
		0.999956,
		0.999377,
		0.99721,
		0.992207,
		0.983187,
		0.969182,
		0.949502,
		0.923746,
		0.891784,
		0.853729,
		0.809907,
		0.760817,
		0.707107,
		0.649542,
		0.588982,
		0.526364,
		0.462682,
		0.398972,
		0.336303,
		0.275766,
		0.218463,
		0.1655,
		0.117983,
		0.077011,
		0.043674,
		0.019046,
		0.004182,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999955,
		0.999365,
		0.997158,
		0.992065,
		0.982892,
		0.968661,
		0.94868,
		0.922551,
		0.890151,
		0.851607,
		0.807258,
		0.757621,
		0.703361,
		0.645261,
		0.584202,
		0.521138,
		0.457085,
		0.393097,
		0.330263,
		0.269691,
		0.2125,
		0.159817,
		0.112762,
		0.072452,
		0.039989,
		0.016464,
		0.002946,
		0.000158,
		0.012153,
		0.04115,
		0.084314,
		0.138935,
		0.20244,
		0.272394,
		0.346512,
		0.422664,
		0.498887,
		0.573396,
		0.644597,
		0.711103,
		0.771749,
		0.825616,
		0.872049,
		0.910685,
		0.941471,
		0.964693,
		0.980982,
		0.991307,
		0.996935,
		0.999327,
		0.999953,
		1,
		0.999954,
		0.999352,
		0.997104,
		0.99192,
		0.982591,
		0.968129,
		0.947842,
		0.921332,
		0.888488,
		0.849448,
		0.804566,
		0.754375,
		0.699558,
		0.64092,
		0.579359,
		0.515852,
		0.45143,
		0.387171,
		0.324181,
		0.263588,
		0.206529,
		0.154148,
		0.107582,
		0.067965,
		0.036415,
		0.014037,
		0.001913,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999953,
		0.99934,
		0.997049,
		0.991772,
		0.982283,
		0.967586,
		0.946986,
		0.92009,
		0.886795,
		0.847251,
		0.801828,
		0.751077,
		0.695699,
		0.636519,
		0.574455,
		0.510503,
		0.445718,
		0.381194,
		0.31806,
		0.25746,
		0.200552,
		0.148496,
		0.102448,
		0.063557,
		0.032959,
		0.011771,
		0.001092,
		0.0076,
		0.033113,
		0.073747,
		0.126677,
		0.189216,
		0.258824,
		0.333111,
		0.409849,
		0.486985,
		0.562644,
		0.635152,
		0.703046,
		0.765094,
		0.820315,
		0.868003,
		0.907751,
		0.939477,
		0.963446,
		0.980287,
		0.990978,
		0.996815,
		0.999299,
		0.999951,
		1,
		0.999952,
		0.999327,
		0.996992,
		0.991621,
		0.981968,
		0.967032,
		0.946113,
		0.918824,
		0.88507,
		0.845015,
		0.799043,
		0.747726,
		0.691782,
		0.632056,
		0.569487,
		0.505093,
		0.439948,
		0.375167,
		0.311899,
		0.251308,
		0.19457,
		0.142864,
		0.097363,
		0.059234,
		0.029626,
		0.009675,
		0.000492,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.999951,
		0.999313,
		0.996935,
		0.991466,
		0.981647,
		0.966466,
		0.945222,
		0.917532,
		0.883312,
		0.842738,
		0.796212,
		0.744321,
		0.687806,
		0.62753,
		0.564456,
		0.499621,
		0.43412,
		0.369091,
		0.3057,
		0.245133,
		0.188586,
		0.137255,
		0.092332,
		0.055,
		0.026425,
		0.007758,
		0.000125,
		0.00401,
		0.025717,
		0.063557,
		0.114585,
		0.175995,
		0.245133,
		0.319503,
		0.396775,
		0.474793,
		0.551596,
		0.62542,
		0.694725,
		0.758206,
		0.814818,
		0.863799,
		0.904698,
		0.937398,
		0.962145,
		0.97956,
		0.990634,
		0.99669,
		0.999271,
		0.999949,
		1,
		0.99995,
		0.999299,
		0.996876,
		0.991307,
		0.981318,
		0.965887,
		0.944313,
		0.916216,
		0.881522,
		0.840421,
		0.793332,
		0.740862,
		0.68377,
		0.622942,
		0.559361,
		0.494086,
		0.428234,
		0.362964,
		0.299464,
		0.238939,
		0.182603,
		0.131674,
		0.08736,
		0.050862,
		0.023363,
		0.006028,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999949,
		0.999285,
		0.996815,
		0.991144,
		0.980982,
		0.965297,
		0.943385,
		0.914874,
		0.879698,
		0.838063,
		0.790404,
		0.737347,
		0.679673,
		0.618289,
		0.554201,
		0.488488,
		0.422291,
		0.356789,
		0.293191,
		0.232725,
		0.176624,
		0.126123,
		0.082451,
		0.046826,
		0.020447,
		0.004495,
		0.001493,
		0.019046,
		0.053808,
		0.102704,
		0.162806,
		0.231342,
		0.3057,
		0.383441,
		0.46231,
		0.540245,
		0.615394,
		0.686131,
		0.751077,
		0.809117,
		0.859431,
		0.90152,
		0.935231,
		0.960786,
		0.978799,
		0.990274,
		0.996558,
		0.999241,
		0.999947,
		1,
		0.999948,
		0.999271,
		0.996753,
		0.990978,
		0.980638,
		0.964693,
		0.942438,
		0.913505,
		0.877839,
		0.835662,
		0.787425,
		0.733775,
		0.675515,
		0.613572,
		0.548975,
		0.482827,
		0.416289,
		0.350565,
		0.286884,
		0.226495,
		0.170651,
		0.120607,
		0.077611,
		0.042898,
		0.017685,
		0.003169,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999947,
		0.999256,
		0.99669,
		0.990808,
		0.980287,
		0.964076,
		0.941471,
		0.912108,
		0.875946,
		0.833218,
		0.784396,
		0.730146,
		0.671294,
		0.608789,
		0.543682,
		0.477101,
		0.41023,
		0.344293,
		0.280543,
		0.22025,
		0.164687,
		0.11513,
		0.072845,
		0.039085,
		0.015085,
		0.002059,
		0.000172,
		0.013194,
		0.044568,
		0.091084,
		0.149684,
		0.21747,
		0.29171,
		0.369853,
		0.449532,
		0.528585,
		0.605064,
		0.677255,
		0.743696,
		0.803203,
		0.854892,
		0.898211,
		0.93297,
		0.959365,
		0.978003,
		0.989896,
		0.996419,
		0.99921,
		0.999945,
		1,
		0.999946,
		0.999241,
		0.996625,
		0.990634,
		0.979927,
		0.963446,
		0.940484,
		0.910685,
		0.874016,
		0.830729,
		0.781314,
		0.726459,
		0.667009,
		0.603939,
		0.538323,
		0.471312,
		0.404113,
		0.337974,
		0.27417,
		0.213992,
		0.158736,
		0.109696,
		0.068159,
		0.035395,
		0.012657,
		0.001176,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999945,
		0.999225,
		0.996558,
		0.990456,
		0.97956,
		0.962803,
		0.939477,
		0.909232,
		0.872049,
		0.828196,
		0.77818,
		0.722712,
		0.66266,
		0.599022,
		0.532897,
		0.465459,
		0.397939,
		0.331609,
		0.267766,
		0.207724,
		0.1528,
		0.10431,
		0.063557,
		0.031834,
		0.010409,
		0.00053,
		0.008262,
		0.035914,
		0.079781,
		0.136666,
		0.203541,
		0.277548,
		0.356013,
		0.436458,
		0.516611,
		0.594422,
		0.668086,
		0.736055,
		0.797066,
		0.850172,
		0.894765,
		0.930611,
		0.957881,
		0.97717,
		0.989499,
		0.996274,
		0.999176,
		0.999943,
		1,
		0.999944,
		0.99921,
		0.99649,
		0.990274,
		0.979184,
		0.962145,
		0.938448,
		0.907751,
		0.870045,
		0.825616,
		0.774992,
		0.718904,
		0.658245,
		0.594037,
		0.527402,
		0.45954,
		0.391707,
		0.325198,
		0.261333,
		0.201449,
		0.146884,
		0.098976,
		0.059048,
		0.02841,
		0.008351,
		0.000135,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999943,
		0.999193,
		0.996419,
		0.990087,
		0.978799,
		0.961473,
		0.937398,
		0.90624,
		0.868003,
		0.82299,
		0.771749,
		0.715035,
		0.653763,
		0.588982,
		0.521839,
		0.453557,
		0.385418,
		0.318742,
		0.254872,
		0.195168,
		0.140991,
		0.093699,
		0.054636,
		0.025132,
		0.006493,
		0.004366,
		0.027933,
		0.068856,
		0.123796,
		0.189584,
		0.263228,
		0.341929,
		0.423087,
		0.504315,
		0.583461,
		0.658615,
		0.728142,
		0.790699,
		0.845265,
		0.891175,
		0.928149,
		0.956329,
		0.976297,
		0.989083,
		0.996121,
		0.999141,
		0.99994,
		1,
		0.999941,
		0.999176,
		0.996348,
		0.989896,
		0.978406,
		0.960786,
		0.936326,
		0.904698,
		0.865921,
		0.820315,
		0.76845,
		0.711103,
		0.649214,
		0.583858,
		0.516206,
		0.447509,
		0.379073,
		0.312242,
		0.248386,
		0.188885,
		0.135126,
		0.088485,
		0.05033,
		0.022008,
		0.004845,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.99994,
		0.999159,
		0.996274,
		0.9897,
		0.978003,
		0.960083,
		0.935231,
		0.903125,
		0.863799,
		0.817592,
		0.765094,
		0.707107,
		0.644597,
		0.578663,
		0.510503,
		0.441396,
		0.372671,
		0.3057,
		0.241875,
		0.182603,
		0.129291,
		0.083339,
		0.046135,
		0.019046,
		0.003417,
		0.001628,
		0.020718,
		0.05838,
		0.111122,
		0.175628,
		0.248768,
		0.327607,
		0.409418,
		0.491695,
		0.57217,
		0.648832,
		0.719949,
		0.78409,
		0.840161,
		0.887434,
		0.925579,
		0.954706,
		0.975383,
		0.988646,
		0.995961,
		0.999105,
		0.999937,
		1,
		0.999939,
		0.999141,
		0.996199,
		0.989499,
		0.977591,
		0.959365,
		0.934112,
		0.90152,
		0.861636,
		0.814818,
		0.76168,
		0.703046,
		0.63991,
		0.573396,
		0.504731,
		0.435217,
		0.366214,
		0.299116,
		0.235343,
		0.176325,
		0.123492,
		0.078267,
		0.042059,
		0.016256,
		0.002221,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999937,
		0.999123,
		0.996121,
		0.989293,
		0.97717,
		0.958631,
		0.93297,
		0.899882,
		0.859431,
		0.811993,
		0.758206,
		0.698919,
		0.635152,
		0.568056,
		0.498887,
		0.428973,
		0.359701,
		0.292492,
		0.228792,
		0.170054,
		0.117733,
		0.073275,
		0.03811,
		0.013648,
		0.001269,
		0.000188,
		0.014374,
		0.048428,
		0.098696,
		0.16171,
		0.234189,
		0.313057,
		0.395453,
		0.478744,
		0.560543,
		0.638727,
		0.711463,
		0.77723,
		0.834852,
		0.883534,
		0.922895,
		0.953007,
		0.974424,
		0.988187,
		0.995792,
		0.999066,
		0.999935,
		1,
		0.999936,
		0.999105,
		0.996042,
		0.989083,
		0.976739,
		0.957881,
		0.931803,
		0.898211,
		0.857183,
		0.809117,
		0.754672,
		0.694725,
		0.630323,
		0.562644,
		0.492972,
		0.422664,
		0.353134,
		0.28583,
		0.222223,
		0.163793,
		0.112019,
		0.06837,
		0.034297,
		0.011231,
		0.000573,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999935,
		0.999086,
		0.995961,
		0.988867,
		0.976297,
		0.957114,
		0.930611,
		0.896506,
		0.854892,
		0.806187,
		0.751077,
		0.690463,
		0.62542,
		0.557157,
		0.486985,
		0.416289,
		0.346512,
		0.27913,
		0.21564,
		0.157547,
		0.106356,
		0.063557,
		0.030627,
		0.009016,
		0.000146,
		0.009016,
		0.039085,
		0.086581,
		0.147869,
		0.219514,
		0.29829,
		0.381194,
		0.465459,
		0.54857,
		0.628289,
		0.702674,
		0.770107,
		0.829327,
		0.879467,
		0.92009,
		0.951229,
		0.973418,
		0.987705,
		0.995614,
		0.999025,
		0.999932,
		1,
		0.999933,
		0.999066,
		0.995877,
		0.988646,
		0.975845,
		0.956329,
		0.929393,
		0.894765,
		0.852555,
		0.803203,
		0.747418,
		0.686131,
		0.620445,
		0.551596,
		0.480925,
		0.409849,
		0.339837,
		0.272394,
		0.209044,
		0.151319,
		0.100747,
		0.058845,
		0.02711,
		0.007014,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.999932,
		0.999046,
		0.995792,
		0.98842,
		0.975383,
		0.955526,
		0.928149,
		0.892988,
		0.850172,
		0.800163,
		0.743696,
		0.681729,
		0.615394,
		0.545959,
		0.474793,
		0.403344,
		0.333111,
		0.265625,
		0.20244,
		0.145114,
		0.0952,
		0.05424,
		0.023754,
		0.005236,
		0.004772,
		0.030448,
		0.074843,
		0.134151,
		0.204772,
		0.283322,
		0.366646,
		0.451836,
		0.536244,
		0.617508,
		0.69357,
		0.76271,
		0.823577,
		0.875226,
		0.917159,
		0.949367,
		0.972363,
		0.987199,
		0.995426,
		0.998982,
		0.999929,
		1,
		0.99993,
		0.999025,
		0.995704,
		0.988187,
		0.974909,
		0.954706,
		0.926878,
		0.891175,
		0.847743,
		0.797066,
		0.739908,
		0.677255,
		0.610268,
		0.540245,
		0.468589,
		0.396775,
		0.326332,
		0.258824,
		0.19583,
		0.138935,
		0.08972,
		0.049751,
		0.02057,
		0.003696,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999929,
		0.999004,
		0.995614,
		0.987949,
		0.974424,
		0.953866,
		0.925579,
		0.889324,
		0.845265,
		0.793912,
		0.736055,
		0.672708,
		0.605064,
		0.534454,
		0.46231,
		0.39014,
		0.319503,
		0.251992,
		0.189216,
		0.132788,
		0.084314,
		0.045385,
		0.017568,
		0.002404,
		0.001783,
		0.02262,
		0.063557,
		0.120607,
		0.189994,
		0.268167,
		0.351814,
		0.437873,
		0.523558,
		0.606372,
		0.684139,
		0.755028,
		0.817592,
		0.870801,
		0.914095,
		0.947416,
		0.971256,
		0.986666,
		0.995229,
		0.998936,
		0.999925,
		1,
		0.999927,
		0.998982,
		0.995521,
		0.987705,
		0.973927,
		0.953007,
		0.924251,
		0.887434,
		0.842738,
		0.790699,
		0.732133,
		0.668086,
		0.599783,
		0.528585,
		0.455958,
		0.383441,
		0.312626,
		0.245133,
		0.182603,
		0.126677,
		0.078987,
		0.04115,
		0.014759,
		0.001374,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999925,
		0.99896,
		0.995426,
		0.987455,
		0.973418,
		0.952128,
		0.922895,
		0.885504,
		0.840161,
		0.787425,
		0.728142,
		0.663389,
		0.594422,
		0.522638,
		0.449532,
		0.376679,
		0.3057,
		0.238249,
		0.175995,
		0.120607,
		0.073747,
		0.037057,
		0.012153,
		0.000621,
		0.000206,
		0.015719,
		0.05281,
		0.107296,
		0.175218,
		0.252848,
		0.336705,
		0.423569,
		0.510503,
		0.594872,
		0.67437,
		0.747049,
		0.811359,
		0.866184,
		0.91089,
		0.945372,
		0.970093,
		0.986105,
		0.99502,
		0.998888,
		0.999922,
		1,
		0.999923,
		0.998936,
		0.995329,
		0.987199,
		0.972897,
		0.951229,
		0.921508,
		0.883534,
		0.837533,
		0.78409,
		0.724081,
		0.658615,
		0.588982,
		0.516611,
		0.443032,
		0.369853,
		0.298728,
		0.231342,
		0.169394,
		0.114585,
		0.068601,
		0.033113,
		0.009762,
		0.000158,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.999922,
		0.998913,
		0.995229,
		0.986936,
		0.972363,
		0.950308,
		0.92009,
		0.881522,
		0.834852,
		0.780692,
		0.719949,
		0.653763,
		0.583461,
		0.510503,
		0.436458,
		0.362964,
		0.29171,
		0.224414,
		0.162806,
		0.108615,
		0.063557,
		0.02933,
		0.0076,
		0.009877,
		0.042694,
		0.094282,
		0.160485,
		0.237387,
		0.321329,
		0.408924,
		0.497074,
		0.582997,
		0.664249,
		0.738759,
		0.804867,
		0.861363,
		0.907537,
		0.943228,
		0.968871,
		0.985515,
		0.994801,
		0.998838,
		0.999918,
		1,
		0.99992,
		0.998888,
		0.995126,
		0.986666,
		0.971816,
		0.949367,
		0.918641,
		0.879467,
		0.832117,
		0.77723,
		0.715743,
		0.648832,
		0.577857,
		0.504315,
		0.429809,
		0.356013,
		0.28465,
		0.21747,
		0.156234,
		0.102704,
		0.058623,
		0.025717,
		0.005678,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999918,
		0.998863,
		0.99502,
		0.986389,
		0.971256,
		0.948403,
		0.917159,
		0.877369,
		0.829327,
		0.773702,
		0.711463,
		0.643821,
		0.57217,
		0.498046,
		0.423087,
		0.349001,
		0.277548,
		0.210511,
		0.149684,
		0.096858,
		0.053808,
		0.022285,
		0.00401,
		0.005236,
		0.033317,
		0.08164,
		0.145843,
		0.221812,
		0.3057,
		0.393939,
		0.483264,
		0.570735,
		0.653763,
		0.730146,
		0.798105,
		0.856329,
		0.904028,
		0.94098,
		0.967586,
		0.984893,
		0.994569,
		0.998784,
		0.999914,
		1,
		0.999916,
		0.998838,
		0.994912,
		0.986105,
		0.970681,
		0.947416,
		0.915644,
		0.875226,
		0.826481,
		0.770107,
		0.707107,
		0.638727,
		0.566399,
		0.491695,
		0.416289,
		0.341929,
		0.270407,
		0.203541,
		0.14316,
		0.091084,
		0.049119,
		0.019046,
		0.00261,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999914,
		0.998811,
		0.994801,
		0.985814,
		0.970093,
		0.946406,
		0.914095,
		0.873037,
		0.823577,
		0.766443,
		0.702674,
		0.63355,
		0.560543,
		0.485261,
		0.409418,
		0.334797,
		0.263228,
		0.196564,
		0.136666,
		0.085389,
		0.044568,
		0.016011,
		0.001493,
		0.00196,
		0.024796,
		0.069452,
		0.131346,
		0.206156,
		0.289832,
		0.378617,
		0.469068,
		0.558077,
		0.642901,
		0.721196,
		0.791059,
		0.851071,
		0.900354,
		0.938621,
		0.966236,
		0.984237,
		0.994324,
		0.998727,
		0.99991,
		1,
		0.999912,
		0.998784,
		0.994686,
		0.985515,
		0.969489,
		0.945372,
		0.91251,
		0.870801,
		0.820615,
		0.76271,
		0.698162,
		0.628289,
		0.5546,
		0.478744,
		0.402472,
		0.327607,
		0.256014,
		0.189584,
		0.13021,
		0.079781,
		0.040162,
		0.013194,
		0.000675,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.99991,
		0.998756,
		0.994569,
		0.985208,
		0.968871,
		0.944313,
		0.91089,
		0.868517,
		0.817592,
		0.758906,
		0.69357,
		0.622942,
		0.54857,
		0.472143,
		0.395453,
		0.32036,
		0.248768,
		0.182603,
		0.123796,
		0.074267,
		0.035914,
		0.010606,
		0.000172,
		0.000227,
		0.017263,
		0.057812,
		0.117059,
		0.190456,
		0.273744,
		0.362964,
		0.454482,
		0.545012,
		0.631647,
		0.711894,
		0.783716,
		0.845578,
		0.896506,
		0.936145,
		0.964815,
		0.983546,
		0.994065,
		0.998667,
		0.999906,
		1,
		0.999908,
		0.998727,
		0.994448,
		0.984893,
		0.968237,
		0.943228,
		0.909232,
		0.866184,
		0.814507,
		0.755028,
		0.688896,
		0.617508,
		0.542452,
		0.465459,
		0.38836,
		0.313057,
		0.241492,
		0.175628,
		0.117431,
		0.068856,
		0.031834,
		0.008262,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999906,
		0.998697,
		0.994324,
		0.984569,
		0.967586,
		0.942118,
		0.907537,
		0.863799,
		0.811359,
		0.751077,
		0.684139,
		0.611985,
		0.536244,
		0.45869,
		0.381194,
		0.3057,
		0.234189,
		0.168662,
		0.111122,
		0.063557,
		0.027933,
		0.006177,
		0.010867,
		0.046826,
		0.10305,
		0.174757,
		0.25746,
		0.346987,
		0.439502,
		0.531529,
		0.619989,
		0.702226,
		0.776061,
		0.839836,
		0.892474,
		0.933544,
		0.963319,
		0.982817,
		0.993791,
		0.998603,
		0.999901,
		1,
		0.999903,
		0.998667,
		0.994196,
		0.984237,
		0.96692,
		0.94098,
		0.905802,
		0.861363,
		0.808146,
		0.747049,
		0.679298,
		0.606372,
		0.529947,
		0.451836,
		0.373956,
		0.29829,
		0.226862,
		0.16171,
		0.104874,
		0.05838,
		0.024223,
		0.004366,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999901,
		0.998635,
		0.994065,
		0.983896,
		0.966236,
		0.939815,
		0.904028,
		0.858873,
		0.804867,
		0.742944,
		0.67437,
		0.600669,
		0.523558,
		0.444897,
		0.366646,
		0.29083,
		0.219514,
		0.154777,
		0.098696,
		0.053333,
		0.020718,
		0.002844,
		0.005773,
		0.036611,
		0.0894,
		0.159107,
		0.241006,
		0.330696,
		0.424126,
		0.517621,
		0.607912,
		0.692176,
		0.76808,
		0.833833,
		0.888248,
		0.930812,
		0.961743,
		0.982047,
		0.993502,
		0.998535,
		0.999896,
		1,
		0.999898,
		0.998603,
		0.99393,
		0.983546,
		0.965534,
		0.938621,
		0.902212,
		0.856329,
		0.801521,
		0.738759,
		0.669354,
		0.594872,
		0.517077,
		0.437873,
		0.359265,
		0.283322,
		0.212149,
		0.147869,
		0.092596,
		0.048428,
		0.01743,
		0.001628,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999896,
		0.998569,
		0.993791,
		0.983187,
		0.964815,
		0.937398,
		0.900354,
		0.853729,
		0.798105,
		0.734494,
		0.664249,
		0.588982,
		0.510503,
		0.430764,
		0.351814,
		0.275766,
		0.204772,
		0.140991,
		0.086581,
		0.043674,
		0.014374,
		0.000737,
		0.002165,
		0.027301,
		0.076202,
		0.143566,
		0.224414,
		0.314104,
		0.408354,
		0.503276,
		0.595403,
		0.681729,
		0.759757,
		0.827555,
		0.883818,
		0.927939,
		0.960083,
		0.981234,
		0.993195,
		0.998463,
		0.999891,
		1,
		0.999893,
		0.998535,
		0.993648,
		0.982817,
		0.964076,
		0.936145,
		0.898452,
		0.851071,
		0.794618,
		0.730146,
		0.659052,
		0.582997,
		0.503836,
		0.423569,
		0.344293,
		0.268167,
		0.197385,
		0.134151,
		0.08066,
		0.039085,
		0.011564,
		0.000188,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999891,
		0.998499,
		0.993502,
		0.982438,
		0.963319,
		0.93486,
		0.896506,
		0.848355,
		0.791059,
		0.725714,
		0.653763,
		0.576915,
		0.497074,
		0.416289,
		0.336705,
		0.260527,
		0.189994,
		0.127354,
		0.074843,
		0.034672,
		0.009016,
		0.000251,
		0.019046,
		0.063557,
		0.128201,
		0.207724,
		0.297228,
		0.392188,
		0.488488,
		0.582448,
		0.670868,
		0.751077,
		0.820988,
		0.87917,
		0.924919,
		0.958333,
		0.980375,
		0.99287,
		0.998387,
		0.999885,
		1,
		0.999888,
		0.998463,
		0.99335,
		0.982047,
		0.962541,
		0.933544,
		0.894513,
		0.845578,
		0.787425,
		0.721196,
		0.64838,
		0.570735,
		0.490217,
		0.408924,
		0.32905,
		0.252848,
		0.182603,
		0.120607,
		0.069138,
		0.030448,
		0.006746,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999885,
		0.998425,
		0.993195,
		0.981647,
		0.961743,
		0.932195,
		0.892474,
		0.842738,
		0.783716,
		0.71659,
		0.642901,
		0.564456,
		0.483264,
		0.401474,
		0.321329,
		0.245133,
		0.175218,
		0.113919,
		0.063557,
		0.026425,
		0.004772,
		0.012014,
		0.051585,
		0.113087,
		0.19098,
		0.280089,
		0.375633,
		0.473249,
		0.569033,
		0.659576,
		0.742021,
		0.814117,
		0.874294,
		0.921741,
		0.956487,
		0.979466,
		0.992526,
		0.998306,
		0.999879,
		1,
		0.999882,
		0.998387,
		0.993035,
		0.981234,
		0.960924,
		0.930812,
		0.890386,
		0.839836,
		0.779928,
		0.711894,
		0.637323,
		0.558077,
		0.476215,
		0.393939,
		0.313545,
		0.237387,
		0.167843,
		0.107296,
		0.058111,
		0.02262,
		0.003111,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999879,
		0.998347,
		0.99287,
		0.980811,
		0.960083,
		0.929393,
		0.888248,
		0.836868,
		0.776061,
		0.707107,
		0.631647,
		0.551596,
		0.469068,
		0.38632,
		0.3057,
		0.229612,
		0.160485,
		0.100747,
		0.05281,
		0.019046,
		0.001783,
		0.006396,
		0.040417,
		0.098313,
		0.174234,
		0.262714,
		0.358694,
		0.457553,
		0.555144,
		0.647836,
		0.732572,
		0.806924,
		0.869175,
		0.918396,
		0.954539,
		0.978505,
		0.99216,
		0.998219,
		0.999872,
		1,
		0.999876,
		0.998306,
		0.9927,
		0.980375,
		0.95922,
		0.927939,
		0.886059,
		0.833833,
		0.772112,
		0.702226,
		0.625869,
		0.545012,
		0.461824,
		0.378617,
		0.297795,
		0.221812,
		0.153149,
		0.094282,
		0.047667,
		0.015719,
		0.000807,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999872,
		0.998263,
		0.992526,
		0.979927,
		0.958333,
		0.926448,
		0.883818,
		0.830729,
		0.76808,
		0.69725,
		0.619989,
		0.538323,
		0.454482,
		0.370832,
		0.289832,
		0.213992,
		0.145843,
		0.08791,
		0.042694,
		0.012657,
		0.000206,
		0.002404,
		0.030205,
		0.083978,
		0.157547,
		0.245133,
		0.341382,
		0.441396,
		0.540768,
		0.635631,
		0.722712,
		0.799394,
		0.863799,
		0.914874,
		0.952482,
		0.977487,
		0.991772,
		0.998127,
		0.999866,
		1,
		0.999869,
		0.998219,
		0.992346,
		0.979466,
		0.957423,
		0.924919,
		0.881522,
		0.827555,
		0.763962,
		0.692176,
		0.614003,
		0.531529,
		0.447041,
		0.362964,
		0.281814,
		0.206156,
		0.138573,
		0.08164,
		0.037906,
		0.009877,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999866,
		0.998174,
		0.99216,
		0.978993,
		0.956487,
		0.92335,
		0.87917,
		0.824309,
		0.759757,
		0.687003,
		0.607912,
		0.524629,
		0.439502,
		0.355015,
		0.273744,
		0.198309,
		0.131346,
		0.075484,
		0.033317,
		0.007397,
		0.000279,
		0.021119,
		0.070199,
		0.140991,
		0.227386,
		0.323712,
		0.424774,
		0.525892,
		0.622942,
		0.71242,
		0.791508,
		0.858152,
		0.911162,
		0.950308,
		0.976409,
		0.99136,
		0.998029,
		0.999858,
		1,
		0.999862,
		0.998127,
		0.991969,
		0.978505,
		0.955526,
		0.921741,
		0.876762,
		0.820988,
		0.755463,
		0.681729,
		0.601712,
		0.517621,
		0.431863,
		0.346987,
		0.265625,
		0.190456,
		0.124172,
		0.069452,
		0.028941,
		0.005236,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999858,
		0.998079,
		0.991772,
		0.978003,
		0.954539,
		0.92009,
		0.874294,
		0.817592,
		0.751077,
		0.676351,
		0.595403,
		0.510503,
		0.424126,
		0.33888,
		0.25746,
		0.182603,
		0.117059,
		0.063557,
		0.024796,
		0.003417,
		0.013353,
		0.057105,
		0.124649,
		0.209516,
		0.3057,
		0.407688,
		0.510503,
		0.609751,
		0.701678,
		0.783246,
		0.852217,
		0.90725,
		0.948011,
		0.975265,
		0.990922,
		0.997925,
		0.999851,
		1,
		0.999854,
		0.998029,
		0.991569,
		0.977487,
		0.953525,
		0.918396,
		0.871765,
		0.814117,
		0.746597,
		0.670868,
		0.588982,
		0.503276,
		0.416289,
		0.330696,
		0.249252,
		0.174757,
		0.110014,
		0.057812,
		0.020897,
		0.00196,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999851,
		0.997978,
		0.99136,
		0.976956,
		0.952482,
		0.916658,
		0.869175,
		0.810562,
		0.742021,
		0.665277,
		0.582448,
		0.495938,
		0.408354,
		0.322436,
		0.241006,
		0.166922,
		0.10305,
		0.05223,
		0.017263,
		0.000888,
		0.007125,
		0.044848,
		0.108615,
		0.191578,
		0.287371,
		0.39014,
		0.494592,
		0.596039,
		0.690463,
		0.77459,
		0.845978,
		0.903125,
		0.94558,
		0.974052,
		0.990456,
		0.997814,
		0.999842,
		1,
		0.999846,
		0.997925,
		0.991144,
		0.976409,
		0.95141,
		0.914874,
		0.86652,
		0.806924,
		0.737347,
		0.659576,
		0.575799,
		0.488488,
		0.40032,
		0.314104,
		0.232725,
		0.159107,
		0.096175,
		0.046826,
		0.013913,
		0.000227,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999842,
		0.99787,
		0.990922,
		0.975845,
		0.950308,
		0.913042,
		0.863799,
		0.803203,
		0.732572,
		0.653763,
		0.569033,
		0.480925,
		0.392188,
		0.3057,
		0.224414,
		0.151319,
		0.0894,
		0.041614,
		0.010867,
		0.002685,
		0.033598,
		0.093,
		0.173636,
		0.268753,
		0.372135,
		0.478147,
		0.581788,
		0.678754,
		0.765517,
		0.839416,
		0.898772,
		0.943008,
		0.972765,
		0.98996,
		0.997695,
		0.999833,
		1,
		0.999838,
		0.997814,
		0.990693,
		0.975265,
		0.949176,
		0.911162,
		0.861011,
		0.799394,
		0.727694,
		0.647836,
		0.562148,
		0.473249,
		0.383959,
		0.297228,
		0.216079,
		0.143566,
		0.082738,
		0.036611,
		0.008146,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999833,
		0.997755,
		0.990456,
		0.974668,
		0.948011,
		0.909232,
		0.858152,
		0.795497,
		0.722712,
		0.641793,
		0.555144,
		0.465459,
		0.375633,
		0.28869,
		0.207724,
		0.135857,
		0.076202,
		0.031834,
		0.005773,
		0.000312,
		0.02355,
		0.077932,
		0.155766,
		0.249885,
		0.353683,
		0.461161,
		0.56698,
		0.666529,
		0.756005,
		0.832511,
		0.894177,
		0.940284,
		0.971397,
		0.989431,
		0.997568,
		0.999824,
		1,
		0.999829,
		0.997695,
		0.990212,
		0.974052,
		0.946813,
		0.90725,
		0.855222,
		0.791508,
		0.717621,
		0.635631,
		0.548017,
		0.457553,
		0.367211,
		0.280089,
		0.199356,
		0.128201,
		0.069803,
		0.027301,
		0.003771,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999824,
		0.997633,
		0.98996,
		0.973418,
		0.94558,
		0.905215,
		0.852217,
		0.787425,
		0.71242,
		0.629348,
		0.540768,
		0.449532,
		0.358694,
		0.271429,
		0.19098,
		0.120607,
		0.063557,
		0.023032,
		0.002165,
		0.014928,
		0.063557,
		0.138055,
		0.23081,
		0.334797,
		0.443626,
		0.551596,
		0.653763,
		0.74603,
		0.825244,
		0.889324,
		0.937398,
		0.969943,
		0.988867,
		0.997432,
		0.999813,
		1,
		0.999819,
		0.997568,
		0.9897,
		0.972765,
		0.944313,
		0.903125,
		0.849137,
		0.783246,
		0.707107,
		0.622942,
		0.533393,
		0.441396,
		0.350084,
		0.262714,
		0.182603,
		0.113087,
		0.057479,
		0.019046,
		0.000982,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999813,
		0.997501,
		0.989431,
		0.972091,
		0.943008,
		0.900978,
		0.845978,
		0.778969,
		0.701678,
		0.61641,
		0.525892,
		0.433143,
		0.341382,
		0.253947,
		0.174234,
		0.105651,
		0.051585,
		0.015366,
		0.000251,
		0.007987,
		0.050046,
		0.120607,
		0.211582,
		0.315497,
		0.42554,
		0.535618,
		0.640434,
		0.735568,
		0.817592,
		0.884195,
		0.934338,
		0.968397,
		0.988265,
		0.997287,
		0.999802,
		1,
		0.999808,
		0.997432,
		0.989154,
		0.971397,
		0.941666,
		0.898772,
		0.842738,
		0.77459,
		0.696131,
		0.609751,
		0.518262,
		0.424774,
		0.332591,
		0.245133,
		0.165879,
		0.098313,
		0.045892,
		0.012014,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999802,
		0.997361,
		0.988867,
		0.970681,
		0.940284,
		0.896506,
		0.839416,
		0.770107,
		0.690463,
		0.602961,
		0.510503,
		0.416289,
		0.323712,
		0.236278,
		0.157547,
		0.091084,
		0.040417,
		0.009016,
		0.003018,
		0.037594,
		0.103544,
		0.192269,
		0.295809,
		0.406901,
		0.519031,
		0.626516,
		0.724593,
		0.809531,
		0.878773,
		0.931091,
		0.96675,
		0.987623,
		0.997131,
		0.99979,
		1,
		0.999797,
		0.997287,
		0.988571,
		0.969943,
		0.938862,
		0.894177,
		0.836008,
		0.765517,
		0.684672,
		0.596039,
		0.502614,
		0.407688,
		0.314747,
		0.227386,
		0.149248,
		0.083978,
		0.035182,
		0.006396,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.99979,
		0.99721,
		0.988265,
		0.969182,
		0.937398,
		0.891784,
		0.832511,
		0.760817,
		0.678754,
		0.588982,
		0.494592,
		0.398972,
		0.3057,
		0.218463,
		0.140991,
		0.077011,
		0.030205,
		0.004182,
		0.000352,
		0.026425,
		0.087007,
		0.172947,
		0.275766,
		0.387712,
		0.501818,
		0.611985,
		0.713077,
		0.801037,
		0.873037,
		0.927644,
		0.964996,
		0.986936,
		0.996964,
		0.999778,
		1,
		0.999784,
		0.997131,
		0.987949,
		0.968397,
		0.93589,
		0.889324,
		0.828924,
		0.756005,
		0.672708,
		0.581788,
		0.486437,
		0.39014,
		0.296573,
		0.209516,
		0.132788,
		0.070199,
		0.02551,
		0.002404,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999778,
		0.997049,
		0.987623,
		0.967586,
		0.934338,
		0.886795,
		0.825244,
		0.751077,
		0.666529,
		0.574455,
		0.478147,
		0.381194,
		0.287371,
		0.200552,
		0.124649,
		0.063557,
		0.021119,
		0.001092,
		0.0168,
		0.071162,
		0.153712,
		0.255412,
		0.36798,
		0.483964,
		0.596815,
		0.700991,
		0.792083,
		0.866967,
		0.923982,
		0.963126,
		0.986201,
		0.996784,
		0.999764,
		1,
		0.999771,
		0.996964,
		0.987285,
		0.96675,
		0.932739,
		0.884195,
		0.821467,
		0.74603,
		0.660215,
		0.56698,
		0.469722,
		0.372135,
		0.278096,
		0.191578,
		0.116587,
		0.057105,
		0.017058,
		0.000279,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999764,
		0.996876,
		0.986936,
		0.965887,
		0.931091,
		0.881522,
		0.817592,
		0.740862,
		0.653763,
		0.559361,
		0.461161,
		0.362964,
		0.268753,
		0.182603,
		0.108615,
		0.050862,
		0.013353,
		0.009016,
		0.0562,
		0.134676,
		0.234798,
		0.34772,
		0.465459,
		0.58098,
		0.688306,
		0.782641,
		0.860539,
		0.92009,
		0.961131,
		0.985413,
		0.996591,
		0.999749,
		1,
		0.999757,
		0.996784,
		0.986574,
		0.964996,
		0.929393,
		0.878773,
		0.813614,
		0.735568,
		0.647171,
		0.551596,
		0.452462,
		0.353683,
		0.259348,
		0.173636,
		0.100747,
		0.044848,
		0.010032,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999749,
		0.99669,
		0.986201,
		0.964076,
		0.927644,
		0.875946,
		0.809531,
		0.730146,
		0.640434,
		0.543682,
		0.443626,
		0.344293,
		0.249885,
		0.164687,
		0.093,
		0.039085,
		0.007125,
		0.003417,
		0.042346,
		0.11597,
		0.213992,
		0.32695,
		0.446291,
		0.564456,
		0.674989,
		0.77268,
		0.853728,
		0.915948,
		0.958999,
		0.984568,
		0.996382,
		0.999732,
		0.999999,
		0.999741,
		0.996591,
		0.985814,
		0.963126,
		0.925841,
		0.873037,
		0.80534,
		0.724593,
		0.63355,
		0.535618,
		0.434652,
		0.334797,
		0.24037,
		0.155766,
		0.085389,
		0.033598,
		0.004665,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999732,
		0.996489,
		0.985413,
		0.962145,
		0.923982,
		0.870045,
		0.801037,
		0.718904,
		0.626516,
		0.527402,
		0.42554,
		0.325198,
		0.23081,
		0.146884,
		0.077932,
		0.02841,
		0.002685,
		0.0004,
		0.02986,
		0.097751,
		0.193071,
		0.305694,
		0.42645,
		0.547208,
		0.661,
		0.762157,
		0.846495,
		0.911526,
		0.956706,
		0.983646,
		0.996142,
		0.999698,
		0.999982,
		0.999719,
		0.996382,
		0.984999,
		0.961131,
		0.922066,
		0.866967,
		0.796619,
		0.713077,
		0.619329,
		0.519031,
		0.416289,
		0.315497,
		0.221211,
		0.138055,
		0.070648,
		0.02355,
		0.001221,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999698,
		0.996268,
		0.984568,
		0.960083,
		0.92009,
		0.863799,
		0.792083,
		0.707107,
		0.611985,
		0.510503,
		0.406901,
		0.3057,
		0.211582,
		0.129291,
		0.063557,
		0.019046,
		0.000312,
		0.019044,
		0.0802,
		0.172124,
		0.283974,
		0.405908,
		0.529181,
		0.646263,
		0.750991,
		0.838759,
		0.906745,
		0.95418,
		0.98258,
		0.995806,
		0.999584,
		0.999887,
		0.999659,
		0.996142,
		0.984119,
		0.958999,
		0.918052,
		0.860539,
		0.787425,
		0.700991,
		0.604481,
		0.501818,
		0.397375,
		0.295809,
		0.201932,
		0.120607,
		0.056682,
		0.014928,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999584,
		0.995994,
		0.983646,
		0.957876,
		0.915948,
		0.857183,
		0.782641,
		0.694725,
		0.596815,
		0.492972,
		0.387712,
		0.28583,
		0.192269,
		0.112019,
		0.050046,
		0.011231,
		0.010251,
		0.063527,
		0.151247,
		0.261794,
		0.384603,
		0.51026,
		0.630629,
		0.739009,
		0.830333,
		0.901414,
		0.951227,
		0.981178,
		0.995184,
		0.9992,
		0.999523,
		0.999445,
		0.995806,
		0.983139,
		0.956706,
		0.913775,
		0.853728,
		0.777728,
		0.688306,
		0.588982,
		0.483964,
		0.377913,
		0.275766,
		0.182603,
		0.103544,
		0.043674,
		0.007987,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.9992,
		0.99555,
		0.98258,
		0.955481,
		0.911526,
		0.850168,
		0.77268,
		0.681729,
		0.58098,
		0.474793,
		0.36798,
		0.265625,
		0.172947,
		0.0952,
		0.037594,
		0.005236,
		0.003895,
		0.047983,
		0.130547,
		0.239137,
		0.362411,
		0.490235,
		0.61382,
		0.72588,
		0.820849,
		0.895138,
		0.94744,
		0.979026,
		0.993859,
		0.99813,
		0.998475,
		0.998789,
		0.995184,
		0.981941,
		0.95418,
		0.909189,
		0.846495,
		0.767493,
		0.674989,
		0.572806,
		0.465459,
		0.357915,
		0.255412,
		0.163312,
		0.087007,
		0.031834,
		0.003018,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.99813,
		0.994648,
		0.981178,
		0.952776,
		0.906745,
		0.842698,
		0.762157,
		0.668082,
		0.564456,
		0.455958,
		0.34772,
		0.245133,
		0.153712,
		0.078987,
		0.026425,
		0.001374,
		0.000457,
		0.033873,
		0.110155,
		0.215968,
		0.339142,
		0.468778,
		0.595385,
		0.711055,
		0.809681,
		0.887236,
		0.942099,
		0.975383,
		0.991083,
		0.995622,
		0.99599,
		0.997118,
		0.993859,
		0.980234,
		0.951227,
		0.904167,
		0.838759,
		0.756663,
		0.661,
		0.555925,
		0.446291,
		0.337397,
		0.234798,
		0.144161,
		0.071162,
		0.021402,
		0.000352,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.995622,
		0.992715,
		0.979026,
		0.949475,
		0.901414,
		0.83465,
		0.750991,
		0.653732,
		0.547208,
		0.436456,
		0.32695,
		0.224414,
		0.134676,
		0.063557,
		0.0168,
		0.021576,
		0.090256,
		0.192263,
		0.314561,
		0.445448,
		0.574699,
		0.693744,
		0.795905,
		0.876682,
		0.934111,
		0.969117,
		0.985706,
		0.990522,
		0.990914,
		0.993485,
		0.991083,
		0.977452,
		0.94744,
		0.898429,
		0.830333,
		0.745119,
		0.646263,
		0.538297,
		0.42645,
		0.316381,
		0.213992,
		0.125272,
		0.0562,
		0.012657,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.990522,
		0.988808,
		0.975383,
		0.945023,
		0.895138,
		0.825756,
		0.739009,
		0.638572,
		0.529181,
		0.416269,
		0.305694,
		0.20354,
		0.11597,
		0.049119,
		0.009016,
		0.011555,
		0.071125,
		0.168059,
		0.28844,
		0.419747,
		0.551001,
		0.672954,
		0.778329,
		0.862132,
		0.922022,
		0.958709,
		0.976177,
		0.981272,
		0.981688,
		0.986527,
		0.985706,
		0.972665,
		0.942099,
		0.891446,
		0.820849,
		0.732617,
		0.630629,
		0.519843,
		0.405908,
		0.294891,
		0.193071,
		0.106789,
		0.042346,
		0.00592,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.981272,
		0.981571,
		0.969117,
		0.938519,
		0.887236,
		0.815526,
		0.72588,
		0.622395,
		0.51026,
		0.395357,
		0.283974,
		0.182595,
		0.097751,
		0.035914,
		0.003417,
		0.004344,
		0.053154,
		0.143513,
		0.260641,
		0.391212,
		0.523504,
		0.64759,
		0.755596,
		0.842016,
		0.904114,
		0.942348,
		0.960643,
		0.966004,
		0.966443,
		0.974514,
		0.976177,
		0.964539,
		0.934111,
		0.882368,
		0.809681,
		0.718723,
		0.61382,
		0.500403,
		0.384603,
		0.272942,
		0.172124,
		0.088879,
		0.02986,
		0.001559,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.966004,
		0.969283,
		0.958709,
		0.928683,
		0.876682,
		0.803188,
		0.711055,
		0.604841,
		0.490235,
		0.373628,
		0.261794,
		0.16167,
		0.0802,
		0.024222,
		0.0004,
		0.0005,
		0.036866,
		0.118961,
		0.231212,
		0.359546,
		0.491542,
		0.616635,
		0.726371,
		0.814743,
		0.878606,
		0.918136,
		0.937149,
		0.942748,
		0.943208,
		0.955494,
		0.960643,
		0.951392,
		0.922022,
		0.870001,
		0.795905,
		0.702768,
		0.595385,
		0.479711,
		0.362411,
		0.250527,
		0.151247,
		0.071739,
		0.019044,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.942748,
		0.950009,
		0.942348,
		0.913908,
		0.862132,
		0.787675,
		0.693744,
		0.585368,
		0.468778,
		0.350925,
		0.239137,
		0.140868,
		0.063527,
		0.01437,
		0.022875,
		0.094942,
		0.20047,
		0.324747,
		0.454752,
		0.579355,
		0.689582,
		0.778956,
		0.84393,
		0.884372,
		0.903927,
		0.909715,
		0.910193,
		0.927549,
		0.937149,
		0.931338,
		0.904114,
		0.852871,
		0.778329,
		0.683852,
		0.574699,
		0.457378,
		0.339142,
		0.227619,
		0.130547,
		0.055596,
		0.010251,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.909715,
		0.921848,
		0.918136,
		0.892416,
		0.842016,
		0.767694,
		0.672954,
		0.563276,
		0.445448,
		0.327031,
		0.215968,
		0.120303,
		0.047983,
		0.00674,
		0.01182,
		0.072182,
		0.169039,
		0.287213,
		0.413227,
		0.535513,
		0.644671,
		0.733821,
		0.799039,
		0.83987,
		0.859722,
		0.865629,
		0.86612,
		0.889104,
		0.903927,
		0.902538,
		0.878606,
		0.829368,
		0.755596,
		0.660911,
		0.551001,
		0.432925,
		0.314561,
		0.204182,
		0.110155,
		0.040727,
		0.003895,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.865629,
		0.883247,
		0.884372,
		0.862495,
		0.814743,
		0.741871,
		0.64759,
		0.537774,
		0.419747,
		0.301705,
		0.192263,
		0.100129,
		0.033873,
		0.001778,
		0.004244,
		0.051515,
		0.137834,
		0.247779,
		0.367622,
		0.485526,
		0.591806,
		0.679276,
		0.743688,
		0.784259,
		0.8041,
		0.810035,
		0.810531,
		0.839261,
		0.859722,
		0.863508,
		0.84393,
		0.797981,
		0.726371,
		0.632867,
		0.523504,
		0.405858,
		0.28844,
		0.180217,
		0.090256,
		0.027471,
		0.000457,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.810035,
		0.833327,
		0.83987,
		0.822798,
		0.778956,
		0.708972,
		0.616635,
		0.508113,
		0.391212,
		0.274752,
		0.168059,
		0.080573,
		0.021576,
		0.000462,
		0.033766,
		0.107965,
		0.207672,
		0.319156,
		0.430524,
		0.531985,
		0.616177,
		0.678608,
		0.718179,
		0.737649,
		0.743507,
		0.743999,
		0.778059,
		0.8041,
		0.813438,
		0.799039,
		0.757581,
		0.689582,
		0.598814,
		0.491542,
		0.375779,
		0.260641,
		0.155812,
		0.071125,
		0.016249,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.743507,
		0.772148,
		0.784259,
		0.77265,
		0.733821,
		0.66815,
		0.579355,
		0.473757,
		0.359546,
		0.246117,
		0.143513,
		0.061965,
		0.011555,
		0.019598,
		0.080587,
		0.168377,
		0.269507,
		0.372282,
		0.467005,
		0.546299,
		0.605532,
		0.643327,
		0.662039,
		0.667704,
		0.668181,
		0.706621,
		0.737649,
		0.752446,
		0.743688,
		0.707694,
		0.644671,
		0.558244,
		0.454752,
		0.342523,
		0.231212,
		0.131208,
		0.053154,
		0.007563,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.667704,
		0.700845,
		0.718179,
		0.712278,
		0.679276,
		0.61919,
		0.535513,
		0.434554,
		0.324747,
		0.215973,
		0.118961,
		0.044763,
		0.004344,
		0.009378,
		0.056729,
		0.131439,
		0.220619,
		0.313028,
		0.399286,
		0.47218,
		0.527057,
		0.562317,
		0.57989,
		0.585242,
		0.585696,
		0.627111,
		0.662039,
		0.681705,
		0.678608,
		0.648705,
		0.591806,
		0.511233,
		0.413227,
		0.30628,
		0.20047,
		0.106843,
		0.036866,
		0.001968,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.585242,
		0.621588,
		0.643327,
		0.642924,
		0.616177,
		0.56267,
		0.485526,
		0.390871,
		0.287213,
		0.184789,
		0.094942,
		0.029542,
		0.0005,
		0.003088,
		0.037121,
		0.09825,
		0.174444,
		0.255168,
		0.331586,
		0.396826,
		0.446348,
		0.4784,
		0.494485,
		0.499416,
		0.499836,
		0.542528,
		0.57989,
		0.603399,
		0.605532,
		0.581942,
		0.531985,
		0.458554,
		0.367622,
		0.267665,
		0.169039,
		0.083354,
		0.022875,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.499416,
		0.537373,
		0.562317,
		0.566781,
		0.546299,
		0.5,
		0.430524,
		0.34365,
		0.247779,
		0.153342,
		0.072182,
		0.016943,
		0.000305,
		0.02209,
		0.069857,
		0.132688,
		0.200987,
		0.26666,
		0.32335,
		0.366762,
		0.395076,
		0.409388,
		0.413805,
		0.414184,
		0.45636,
		0.494485,
		0.520504,
		0.527057,
		0.509589,
		0.467005,
		0.401677,
		0.319156,
		0.227721,
		0.137834,
		0.061533,
		0.01182,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.413805,
		0.451676,
		0.4784,
		0.486772,
		0.47218,
		0.433319,
		0.372282,
		0.294361,
		0.207672,
		0.122659,
		0.051515,
		0.007571,
		0.011518,
		0.046838,
		0.096614,
		0.152377,
		0.206942,
		0.254614,
		0.291463,
		0.315694,
		0.328034,
		0.33187,
		0.332201,
		0.372177,
		0.409388,
		0.436447,
		0.446348,
		0.434461,
		0.399286,
		0.342631,
		0.269507,
		0.187825,
		0.107965,
		0.042229,
		0.004244,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.33187,
		0.368043,
		0.395076,
		0.406208,
		0.396826,
		0.365257,
		0.313028,
		0.244843,
		0.168377,
		0.093895,
		0.033766,
		0.001872,
		0.004893,
		0.029263,
		0.066924,
		0.110655,
		0.1543,
		0.192937,
		0.223105,
		0.243113,
		0.253384,
		0.256601,
		0.25688,
		0.293242,
		0.328034,
		0.354699,
		0.366762,
		0.359665,
		0.331586,
		0.283774,
		0.220619,
		0.14952,
		0.080587,
		0.026203,
		0.000462,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.256601,
		0.289712,
		0.315694,
		0.328396,
		0.32335,
		0.298613,
		0.255168,
		0.197077,
		0.131439,
		0.068164,
		0.019598,
		0.001409,
		0.016747,
		0.043741,
		0.07648,
		0.109901,
		0.13992,
		0.163614,
		0.179472,
		0.18768,
		0.190271,
		0.190498,
		0.222209,
		0.253384,
		0.278401,
		0.291463,
		0.288229,
		0.26666,
		0.227488,
		0.174444,
		0.1143,
		0.056729,
		0.013986,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.190271,
		0.219306,
		0.243113,
		0.256277,
		0.254614,
		0.236018,
		0.200987,
		0.152922,
		0.09825,
		0.046363,
		0.009378,
		0.000119,
		0.008548,
		0.026672,
		0.049875,
		0.074187,
		0.096378,
		0.1141,
		0.126076,
		0.13233,
		0.13432,
		0.134496,
		0.160942,
		0.18768,
		0.210073,
		0.223105,
		0.222774,
		0.206942,
		0.175889,
		0.132688,
		0.083406,
		0.037121,
		0.005759,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.13432,
		0.158656,
		0.179472,
		0.192162,
		0.192937,
		0.179646,
		0.152377,
		0.113883,
		0.069857,
		0.029031,
		0.003088,
		0.003723,
		0.014934,
		0.030318,
		0.046936,
		0.062379,
		0.07487,
		0.083398,
		0.087893,
		0.089337,
		0.089465,
		0.110452,
		0.13233,
		0.151444,
		0.163614,
		0.165276,
		0.1543,
		0.130598,
		0.096614,
		0.057659,
		0.02209,
		0.001299,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.089337,
		0.108744,
		0.126076,
		0.137572,
		0.13992,
		0.131026,
		0.110655,
		0.08095,
		0.046838,
		0.01627,
		0.000305,
		0.001272,
		0.007501,
		0.016879,
		0.027382,
		0.037343,
		0.045513,
		0.051152,
		0.054154,
		0.055126,
		0.055213,
		0.070925,
		0.087893,
		0.103399,
		0.1141,
		0.116944,
		0.109901,
		0.092608,
		0.066924,
		0.037384,
		0.011518,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.055126,
		0.069727,
		0.083398,
		0.09321,
		0.096378,
		0.090958,
		0.07648,
		0.054533,
		0.029263,
		0.007757,
		0.000278,
		0.003262,
		0.008381,
		0.014372,
		0.020186,
		0.025028,
		0.02841,
		0.030229,
		0.030824,
		0.030877,
		0.041813,
		0.054154,
		0.066016,
		0.07487,
		0.078208,
		0.074187,
		0.062253,
		0.043741,
		0.022412,
		0.004893,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.030824,
		0.041042,
		0.051152,
		0.059008,
		0.062379,
		0.05954,
		0.049875,
		0.034483,
		0.016747,
		0.002816,
		1.6e-005,
		0.00116,
		0.003558,
		0.006523,
		0.009478,
		0.01198,
		0.01375,
		0.014712,
		0.01503,
		0.015059,
		0.021968,
		0.030229,
		0.038672,
		0.045513,
		0.048799,
		0.046936,
		0.039266,
		0.026672,
		0.012164,
		0.001409,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.01503,
		0.021524,
		0.02841,
		0.034243,
		0.037343,
		0.036262,
		0.030318,
		0.020197,
		0.008548,
		0.000551,
		0.000304,
		0.0012,
		0.002388,
		0.003609,
		0.004662,
		0.005418,
		0.005833,
		0.005972,
		0.005985,
		0.00978,
		0.014712,
		0.020174,
		0.025028,
		0.027863,
		0.027382,
		0.022903,
		0.014934,
		0.005777,
		0.000119,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.005972,
		0.009563,
		0.01375,
		0.017677,
		0.020186,
		0.020142,
		0.016879,
		0.010748,
		0.003723,
		4.8e-005,
		0.000278,
		0.000613,
		0.00097,
		0.001285,
		0.001514,
		0.001641,
		0.001685,
		0.001689,
		0.003356,
		0.005833,
		0.00891,
		0.01198,
		0.014124,
		0.014372,
		0.012091,
		0.007501,
		0.002261,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.001685,
		0.003275,
		0.005418,
		0.007717,
		0.009478,
		0.009893,
		0.008381,
		0.005051,
		0.001272,
		3.1e-005,
		7.7e-005,
		0.000129,
		0.000175,
		0.00021,
		0.000229,
		0.000236,
		0.000237,
		0.000718,
		0.001641,
		0.003031,
		0.004662,
		0.006041,
		0.006523,
		0.005589,
		0.003262,
		0.000644,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		0.000236,
		0.000698,
		0.001514,
		0.002591,
		0.003609,
		0.00408,
		0.003558,
		0.002005,
		0.000278,
		4.8e-005,
		0.000229,
		0.000642,
		0.001285,
		0.001981,
		0.002388,
		0.002141,
		0.00116,
		9.2e-005,
		4.7e-005,
		0.00021,
		0.00054,
		0.00097,
		0.001282,
		0.0012,
		0.000623,
		1.6e-005,
		4.3e-005,
		0.000175,
		0.000402,
		0.000613,
		0.000614,
		0.000304,
		3.5e-005,
		0.000129,
		0.000248,
		0.000278,
		0.000131,
		2.6e-005,
		7.7e-005,
		0.000106,
		4.8e-005,
		1.5e-005,
		3.1e-005,
		1.4e-005,
		1.4e-005,
		3.1e-005,
		1.5e-005,
		4.8e-005,
		0.000106,
		7.7e-005,
		2.6e-005,
		0.000131,
		0.000278,
		0.000248,
		0.000129,
		3.5e-005,
		0.000304,
		0.000614,
		0.000613,
		0.000402,
		0.000175,
		4.3e-005,
		1.6e-005,
		0.000623,
		0.0012,
		0.001282,
		0.00097,
		0.00054,
		0.00021,
		4.7e-005,
		9.2e-005,
		0.00116,
		0.002141,
		0.002388,
		0.001981,
		0.001285,
		0.000642,
		0.000229,
		4.8e-005,
		0.000278,
		0.002005,
		0.003558,
		0.00408,
		0.003609,
		0.002591,
		0.001514,
		0.000698,
		0.000236,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		0.000644,
		0.003262,
		0.005589,
		0.006523,
		0.006041,
		0.004662,
		0.003031,
		0.001641,
		0.000718,
		0.000237,
		0.000236,
		0.000229,
		0.00021,
		0.000175,
		0.000129,
		7.7e-005,
		3.1e-005,
		0.001272,
		0.005051,
		0.008381,
		0.009893,
		0.009478,
		0.007717,
		0.005418,
		0.003275,
		0.001685,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		0.002261,
		0.007501,
		0.012091,
		0.014372,
		0.014124,
		0.01198,
		0.00891,
		0.005833,
		0.003356,
		0.001689,
		0.001685,
		0.001641,
		0.001514,
		0.001285,
		0.00097,
		0.000613,
		0.000278,
		4.8e-005,
		0.003723,
		0.010748,
		0.016879,
		0.020142,
		0.020186,
		0.017677,
		0.01375,
		0.009563,
		0.005972,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		0.000119,
		0.005777,
		0.014934,
		0.022903,
		0.027382,
		0.027863,
		0.025028,
		0.020174,
		0.014712,
		0.00978,
		0.005985,
		0.005972,
		0.005833,
		0.005418,
		0.004662,
		0.003609,
		0.002388,
		0.0012,
		0.000304,
		0.000551,
		0.008548,
		0.020197,
		0.030318,
		0.036262,
		0.037343,
		0.034243,
		0.02841,
		0.021524,
		0.01503,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.001409,
		0.012164,
		0.026672,
		0.039266,
		0.046936,
		0.048799,
		0.045513,
		0.038672,
		0.030229,
		0.021968,
		0.015059,
		0.01503,
		0.014712,
		0.01375,
		0.01198,
		0.009478,
		0.006523,
		0.003558,
		0.00116,
		1.6e-005,
		0.002816,
		0.016747,
		0.034483,
		0.049875,
		0.05954,
		0.062379,
		0.059008,
		0.051152,
		0.041042,
		0.030824,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		0.004893,
		0.022412,
		0.043741,
		0.062253,
		0.074187,
		0.078208,
		0.07487,
		0.066016,
		0.054154,
		0.041813,
		0.030877,
		0.030824,
		0.030229,
		0.02841,
		0.025028,
		0.020186,
		0.014372,
		0.008381,
		0.003262,
		0.000278,
		0.007757,
		0.029263,
		0.054533,
		0.07648,
		0.090958,
		0.096378,
		0.09321,
		0.083398,
		0.069727,
		0.055126,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.011518,
		0.037384,
		0.066924,
		0.092608,
		0.109901,
		0.116944,
		0.1141,
		0.103399,
		0.087893,
		0.070925,
		0.055213,
		0.055126,
		0.054154,
		0.051152,
		0.045513,
		0.037343,
		0.027382,
		0.016879,
		0.007501,
		0.001272,
		0.000305,
		0.01627,
		0.046838,
		0.08095,
		0.110655,
		0.131026,
		0.13992,
		0.137572,
		0.126076,
		0.108744,
		0.089337,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		0.001299,
		0.02209,
		0.057659,
		0.096614,
		0.130598,
		0.1543,
		0.165276,
		0.163614,
		0.151444,
		0.13233,
		0.110452,
		0.089465,
		0.089337,
		0.087893,
		0.083398,
		0.07487,
		0.062379,
		0.046936,
		0.030318,
		0.014934,
		0.003723,
		0.003088,
		0.029031,
		0.069857,
		0.113883,
		0.152377,
		0.179646,
		0.192937,
		0.192162,
		0.179472,
		0.158656,
		0.13432,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.005759,
		0.037121,
		0.083406,
		0.132688,
		0.175889,
		0.206942,
		0.222774,
		0.223105,
		0.210073,
		0.18768,
		0.160942,
		0.134496,
		0.13432,
		0.13233,
		0.126076,
		0.1141,
		0.096378,
		0.074187,
		0.049875,
		0.026672,
		0.008548,
		0.000119,
		0.009378,
		0.046363,
		0.09825,
		0.152922,
		0.200987,
		0.236018,
		0.254614,
		0.256277,
		0.243113,
		0.219306,
		0.190271,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.013986,
		0.056729,
		0.1143,
		0.174444,
		0.227488,
		0.26666,
		0.288229,
		0.291463,
		0.278401,
		0.253384,
		0.222209,
		0.190498,
		0.190271,
		0.18768,
		0.179472,
		0.163614,
		0.13992,
		0.109901,
		0.07648,
		0.043741,
		0.016747,
		0.001409,
		0.019598,
		0.068164,
		0.131439,
		0.197077,
		0.255168,
		0.298613,
		0.32335,
		0.328396,
		0.315694,
		0.289712,
		0.256601,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.000462,
		0.026203,
		0.080587,
		0.14952,
		0.220619,
		0.283774,
		0.331586,
		0.359665,
		0.366762,
		0.354699,
		0.328034,
		0.293242,
		0.25688,
		0.256601,
		0.253384,
		0.243113,
		0.223105,
		0.192937,
		0.1543,
		0.110655,
		0.066924,
		0.029263,
		0.004893,
		0.001872,
		0.033766,
		0.093895,
		0.168377,
		0.244843,
		0.313028,
		0.365257,
		0.396826,
		0.406208,
		0.395076,
		0.368043,
		0.33187,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.004244,
		0.042229,
		0.107965,
		0.187825,
		0.269507,
		0.342631,
		0.399286,
		0.434461,
		0.446348,
		0.436447,
		0.409388,
		0.372177,
		0.332201,
		0.33187,
		0.328034,
		0.315694,
		0.291463,
		0.254614,
		0.206942,
		0.152377,
		0.096614,
		0.046838,
		0.011518,
		0.007571,
		0.051515,
		0.122659,
		0.207672,
		0.294361,
		0.372282,
		0.433319,
		0.47218,
		0.486772,
		0.4784,
		0.451676,
		0.413805,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.01182,
		0.061533,
		0.137834,
		0.227721,
		0.319156,
		0.401677,
		0.467005,
		0.509589,
		0.527057,
		0.520504,
		0.494485,
		0.45636,
		0.414184,
		0.413805,
		0.409388,
		0.395076,
		0.366762,
		0.32335,
		0.26666,
		0.200987,
		0.132688,
		0.069857,
		0.02209,
		0.000305,
		0.016943,
		0.072182,
		0.153342,
		0.247779,
		0.34365,
		0.430524,
		0.5,
		0.546299,
		0.566781,
		0.562317,
		0.537373,
		0.499416,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.022875,
		0.083354,
		0.169039,
		0.267665,
		0.367622,
		0.458554,
		0.531985,
		0.581942,
		0.605532,
		0.603399,
		0.57989,
		0.542528,
		0.499836,
		0.499416,
		0.494485,
		0.4784,
		0.446348,
		0.396826,
		0.331586,
		0.255168,
		0.174444,
		0.09825,
		0.037121,
		0.003088,
		0.0005,
		0.029542,
		0.094942,
		0.184789,
		0.287213,
		0.390871,
		0.485526,
		0.56267,
		0.616177,
		0.642924,
		0.643327,
		0.621588,
		0.585242,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.001968,
		0.036866,
		0.106843,
		0.20047,
		0.30628,
		0.413227,
		0.511233,
		0.591806,
		0.648705,
		0.678608,
		0.681705,
		0.662039,
		0.627111,
		0.585696,
		0.585242,
		0.57989,
		0.562317,
		0.527057,
		0.47218,
		0.399286,
		0.313028,
		0.220619,
		0.131439,
		0.056729,
		0.009378,
		0.004344,
		0.044763,
		0.118961,
		0.215973,
		0.324747,
		0.434554,
		0.535513,
		0.61919,
		0.679276,
		0.712278,
		0.718179,
		0.700845,
		0.667704,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.007563,
		0.053154,
		0.131208,
		0.231212,
		0.342523,
		0.454752,
		0.558244,
		0.644671,
		0.707694,
		0.743688,
		0.752446,
		0.737649,
		0.706621,
		0.668181,
		0.667704,
		0.662039,
		0.643327,
		0.605532,
		0.546299,
		0.467005,
		0.372282,
		0.269507,
		0.168377,
		0.080587,
		0.019598,
		0.011555,
		0.061965,
		0.143513,
		0.246117,
		0.359546,
		0.473757,
		0.579355,
		0.66815,
		0.733821,
		0.77265,
		0.784259,
		0.772148,
		0.743507,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.016249,
		0.071125,
		0.155812,
		0.260641,
		0.375779,
		0.491542,
		0.598814,
		0.689582,
		0.757581,
		0.799039,
		0.813438,
		0.8041,
		0.778059,
		0.743999,
		0.743507,
		0.737649,
		0.718179,
		0.678608,
		0.616177,
		0.531985,
		0.430524,
		0.319156,
		0.207672,
		0.107965,
		0.033766,
		0.000462,
		0.021576,
		0.080573,
		0.168059,
		0.274752,
		0.391212,
		0.508113,
		0.616635,
		0.708972,
		0.778956,
		0.822798,
		0.83987,
		0.833327,
		0.810035,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.000457,
		0.027471,
		0.090256,
		0.180217,
		0.28844,
		0.405858,
		0.523504,
		0.632867,
		0.726371,
		0.797981,
		0.84393,
		0.863508,
		0.859722,
		0.839261,
		0.810531,
		0.810035,
		0.8041,
		0.784259,
		0.743688,
		0.679276,
		0.591806,
		0.485526,
		0.367622,
		0.247779,
		0.137834,
		0.051515,
		0.004244,
		0.001778,
		0.033873,
		0.100129,
		0.192263,
		0.301705,
		0.419747,
		0.537774,
		0.64759,
		0.741871,
		0.814743,
		0.862495,
		0.884372,
		0.883247,
		0.865629,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.003895,
		0.040727,
		0.110155,
		0.204182,
		0.314561,
		0.432925,
		0.551001,
		0.660911,
		0.755596,
		0.829368,
		0.878606,
		0.902538,
		0.903927,
		0.889104,
		0.86612,
		0.865629,
		0.859722,
		0.83987,
		0.799039,
		0.733821,
		0.644671,
		0.535513,
		0.413227,
		0.287213,
		0.169039,
		0.072182,
		0.01182,
		0.00674,
		0.047983,
		0.120303,
		0.215968,
		0.327031,
		0.445448,
		0.563276,
		0.672954,
		0.767694,
		0.842016,
		0.892416,
		0.918136,
		0.921848,
		0.909715,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.010251,
		0.055596,
		0.130547,
		0.227619,
		0.339142,
		0.457378,
		0.574699,
		0.683852,
		0.778329,
		0.852871,
		0.904114,
		0.931338,
		0.937149,
		0.927549,
		0.910193,
		0.909715,
		0.903927,
		0.884372,
		0.84393,
		0.778956,
		0.689582,
		0.579355,
		0.454752,
		0.324747,
		0.20047,
		0.094942,
		0.022875,
		0.01437,
		0.063527,
		0.140868,
		0.239137,
		0.350925,
		0.468778,
		0.585368,
		0.693744,
		0.787675,
		0.862132,
		0.913908,
		0.942348,
		0.950009,
		0.942748,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.019044,
		0.071739,
		0.151247,
		0.250527,
		0.362411,
		0.479711,
		0.595385,
		0.702768,
		0.795905,
		0.870001,
		0.922022,
		0.951392,
		0.960643,
		0.955494,
		0.943208,
		0.942748,
		0.937149,
		0.918136,
		0.878606,
		0.814743,
		0.726371,
		0.616635,
		0.491542,
		0.359546,
		0.231212,
		0.118961,
		0.036866,
		0.0005,
		0.0004,
		0.024222,
		0.0802,
		0.16167,
		0.261794,
		0.373628,
		0.490235,
		0.604841,
		0.711055,
		0.803188,
		0.876682,
		0.928683,
		0.958709,
		0.969283,
		0.966004,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.001559,
		0.02986,
		0.088879,
		0.172124,
		0.272942,
		0.384603,
		0.500403,
		0.61382,
		0.718723,
		0.809681,
		0.882368,
		0.934111,
		0.964539,
		0.976177,
		0.974514,
		0.966443,
		0.966004,
		0.960643,
		0.942348,
		0.904114,
		0.842016,
		0.755596,
		0.64759,
		0.523504,
		0.391212,
		0.260641,
		0.143513,
		0.053154,
		0.004344,
		0.003417,
		0.035914,
		0.097751,
		0.182595,
		0.283974,
		0.395357,
		0.51026,
		0.622395,
		0.72588,
		0.815526,
		0.887236,
		0.938519,
		0.969117,
		0.981571,
		0.981272,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.00592,
		0.042346,
		0.106789,
		0.193071,
		0.294891,
		0.405908,
		0.519843,
		0.630629,
		0.732617,
		0.820849,
		0.891446,
		0.942099,
		0.972665,
		0.985706,
		0.986527,
		0.981688,
		0.981272,
		0.976177,
		0.958709,
		0.922022,
		0.862132,
		0.778329,
		0.672954,
		0.551001,
		0.419747,
		0.28844,
		0.168059,
		0.071125,
		0.011555,
		0.009016,
		0.049119,
		0.11597,
		0.20354,
		0.305694,
		0.416269,
		0.529181,
		0.638572,
		0.739009,
		0.825756,
		0.895138,
		0.945023,
		0.975383,
		0.988808,
		0.990522,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.012657,
		0.0562,
		0.125272,
		0.213992,
		0.316381,
		0.42645,
		0.538297,
		0.646263,
		0.745119,
		0.830333,
		0.898429,
		0.94744,
		0.977452,
		0.991083,
		0.993485,
		0.990914,
		0.990522,
		0.985706,
		0.969117,
		0.934111,
		0.876682,
		0.795905,
		0.693744,
		0.574699,
		0.445448,
		0.314561,
		0.192263,
		0.090256,
		0.021576,
		0.0168,
		0.063557,
		0.134676,
		0.224414,
		0.32695,
		0.436456,
		0.547208,
		0.653732,
		0.750991,
		0.83465,
		0.901414,
		0.949475,
		0.979026,
		0.992715,
		0.995622,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.000352,
		0.021402,
		0.071162,
		0.144161,
		0.234798,
		0.337397,
		0.446291,
		0.555925,
		0.661,
		0.756663,
		0.838759,
		0.904167,
		0.951227,
		0.980234,
		0.993859,
		0.997118,
		0.99599,
		0.995622,
		0.991083,
		0.975383,
		0.942099,
		0.887236,
		0.809681,
		0.711055,
		0.595385,
		0.468778,
		0.339142,
		0.215968,
		0.110155,
		0.033873,
		0.000457,
		0.001374,
		0.026425,
		0.078987,
		0.153712,
		0.245133,
		0.34772,
		0.455958,
		0.564456,
		0.668082,
		0.762157,
		0.842698,
		0.906745,
		0.952776,
		0.981178,
		0.994648,
		0.99813,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.003018,
		0.031834,
		0.087007,
		0.163312,
		0.255412,
		0.357915,
		0.465459,
		0.572806,
		0.674989,
		0.767493,
		0.846495,
		0.909189,
		0.95418,
		0.981941,
		0.995184,
		0.998789,
		0.998475,
		0.99813,
		0.993859,
		0.979026,
		0.94744,
		0.895138,
		0.820849,
		0.72588,
		0.61382,
		0.490235,
		0.362411,
		0.239137,
		0.130547,
		0.047983,
		0.003895,
		0.005236,
		0.037594,
		0.0952,
		0.172947,
		0.265625,
		0.36798,
		0.474793,
		0.58098,
		0.681729,
		0.77268,
		0.850168,
		0.911526,
		0.955481,
		0.98258,
		0.99555,
		0.9992,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.007987,
		0.043674,
		0.103544,
		0.182603,
		0.275766,
		0.377913,
		0.483964,
		0.588982,
		0.688306,
		0.777728,
		0.853728,
		0.913775,
		0.956706,
		0.983139,
		0.995806,
		0.999445,
		0.999523,
		0.9992,
		0.995184,
		0.981178,
		0.951227,
		0.901414,
		0.830333,
		0.739009,
		0.630629,
		0.51026,
		0.384603,
		0.261794,
		0.151247,
		0.063527,
		0.010251,
		0.011231,
		0.050046,
		0.112019,
		0.192269,
		0.28583,
		0.387712,
		0.492972,
		0.596815,
		0.694725,
		0.782641,
		0.857183,
		0.915948,
		0.957876,
		0.983646,
		0.995994,
		0.999584,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.014928,
		0.056682,
		0.120607,
		0.201932,
		0.295809,
		0.397375,
		0.501818,
		0.604481,
		0.700991,
		0.787425,
		0.860539,
		0.918052,
		0.958999,
		0.984119,
		0.996142,
		0.999659,
		0.999887,
		0.999584,
		0.995806,
		0.98258,
		0.95418,
		0.906745,
		0.838759,
		0.750991,
		0.646263,
		0.529181,
		0.405908,
		0.283974,
		0.172124,
		0.0802,
		0.019044,
		0.000312,
		0.019046,
		0.063557,
		0.129291,
		0.211582,
		0.3057,
		0.406901,
		0.510503,
		0.611985,
		0.707107,
		0.792083,
		0.863799,
		0.92009,
		0.960083,
		0.984568,
		0.996268,
		0.999698,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.001221,
		0.02355,
		0.070648,
		0.138055,
		0.221211,
		0.315497,
		0.416289,
		0.519031,
		0.619329,
		0.713077,
		0.796619,
		0.866967,
		0.922066,
		0.961131,
		0.984999,
		0.996382,
		0.999719,
		0.999982,
		0.999698,
		0.996142,
		0.983646,
		0.956706,
		0.911526,
		0.846495,
		0.762157,
		0.661,
		0.547208,
		0.42645,
		0.305694,
		0.193071,
		0.097751,
		0.02986,
		0.0004,
		0.002685,
		0.02841,
		0.077932,
		0.146884,
		0.23081,
		0.325198,
		0.42554,
		0.527402,
		0.626516,
		0.718904,
		0.801037,
		0.870045,
		0.923982,
		0.962145,
		0.985413,
		0.996489,
		0.999732,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.004665,
		0.033598,
		0.085389,
		0.155766,
		0.24037,
		0.334797,
		0.434652,
		0.535618,
		0.63355,
		0.724593,
		0.80534,
		0.873037,
		0.925841,
		0.963126,
		0.985814,
		0.996591,
		0.999741,
		0.999999,
		0.999732,
		0.996382,
		0.984568,
		0.958999,
		0.915948,
		0.853728,
		0.77268,
		0.674989,
		0.564456,
		0.446291,
		0.32695,
		0.213992,
		0.11597,
		0.042346,
		0.003417,
		0.007125,
		0.039085,
		0.093,
		0.164687,
		0.249885,
		0.344293,
		0.443626,
		0.543682,
		0.640434,
		0.730146,
		0.809531,
		0.875946,
		0.927644,
		0.964076,
		0.986201,
		0.99669,
		0.999749,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.010032,
		0.044848,
		0.100747,
		0.173636,
		0.259348,
		0.353683,
		0.452462,
		0.551596,
		0.647171,
		0.735568,
		0.813614,
		0.878773,
		0.929393,
		0.964996,
		0.986574,
		0.996784,
		0.999757,
		1,
		0.999749,
		0.996591,
		0.985413,
		0.961131,
		0.92009,
		0.860539,
		0.782641,
		0.688306,
		0.58098,
		0.465459,
		0.34772,
		0.234798,
		0.134676,
		0.0562,
		0.009016,
		0.013353,
		0.050862,
		0.108615,
		0.182603,
		0.268753,
		0.362964,
		0.461161,
		0.559361,
		0.653763,
		0.740862,
		0.817592,
		0.881522,
		0.931091,
		0.965887,
		0.986936,
		0.996876,
		0.999764,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.000279,
		0.017058,
		0.057105,
		0.116587,
		0.191578,
		0.278096,
		0.372135,
		0.469722,
		0.56698,
		0.660215,
		0.74603,
		0.821467,
		0.884195,
		0.932739,
		0.96675,
		0.987285,
		0.996964,
		0.999771,
		1,
		0.999764,
		0.996784,
		0.986201,
		0.963126,
		0.923982,
		0.866967,
		0.792083,
		0.700991,
		0.596815,
		0.483964,
		0.36798,
		0.255412,
		0.153712,
		0.071162,
		0.0168,
		0.001092,
		0.021119,
		0.063557,
		0.124649,
		0.200552,
		0.287371,
		0.381194,
		0.478147,
		0.574455,
		0.666529,
		0.751077,
		0.825244,
		0.886795,
		0.934338,
		0.967586,
		0.987623,
		0.997049,
		0.999778,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.002404,
		0.02551,
		0.070199,
		0.132788,
		0.209516,
		0.296573,
		0.39014,
		0.486437,
		0.581788,
		0.672708,
		0.756005,
		0.828924,
		0.889324,
		0.93589,
		0.968397,
		0.987949,
		0.997131,
		0.999784,
		1,
		0.999778,
		0.996964,
		0.986936,
		0.964996,
		0.927644,
		0.873037,
		0.801037,
		0.713077,
		0.611985,
		0.501818,
		0.387712,
		0.275766,
		0.172947,
		0.087007,
		0.026425,
		0.000352,
		0.004182,
		0.030205,
		0.077011,
		0.140991,
		0.218463,
		0.3057,
		0.398972,
		0.494592,
		0.588982,
		0.678754,
		0.760817,
		0.832511,
		0.891784,
		0.937398,
		0.969182,
		0.988265,
		0.99721,
		0.99979,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.006396,
		0.035182,
		0.083978,
		0.149248,
		0.227386,
		0.314747,
		0.407688,
		0.502614,
		0.596039,
		0.684672,
		0.765517,
		0.836008,
		0.894177,
		0.938862,
		0.969943,
		0.988571,
		0.997287,
		0.999797,
		1,
		0.99979,
		0.997131,
		0.987623,
		0.96675,
		0.931091,
		0.878773,
		0.809531,
		0.724593,
		0.626516,
		0.519031,
		0.406901,
		0.295809,
		0.192269,
		0.103544,
		0.037594,
		0.003018,
		0.009016,
		0.040417,
		0.091084,
		0.157547,
		0.236278,
		0.323712,
		0.416289,
		0.510503,
		0.602961,
		0.690463,
		0.770107,
		0.839416,
		0.896506,
		0.940284,
		0.970681,
		0.988867,
		0.997361,
		0.999802,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.012014,
		0.045892,
		0.098313,
		0.165879,
		0.245133,
		0.332591,
		0.424774,
		0.518262,
		0.609751,
		0.696131,
		0.77459,
		0.842738,
		0.898772,
		0.941666,
		0.971397,
		0.989154,
		0.997432,
		0.999808,
		1,
		0.999802,
		0.997287,
		0.988265,
		0.968397,
		0.934338,
		0.884195,
		0.817592,
		0.735568,
		0.640434,
		0.535618,
		0.42554,
		0.315497,
		0.211582,
		0.120607,
		0.050046,
		0.007987,
		0.000251,
		0.015366,
		0.051585,
		0.105651,
		0.174234,
		0.253947,
		0.341382,
		0.433143,
		0.525892,
		0.61641,
		0.701678,
		0.778969,
		0.845978,
		0.900978,
		0.943008,
		0.972091,
		0.989431,
		0.997501,
		0.999813,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.000982,
		0.019046,
		0.057479,
		0.113087,
		0.182603,
		0.262714,
		0.350084,
		0.441396,
		0.533393,
		0.622942,
		0.707107,
		0.783246,
		0.849137,
		0.903125,
		0.944313,
		0.972765,
		0.9897,
		0.997568,
		0.999819,
		1,
		0.999813,
		0.997432,
		0.988867,
		0.969943,
		0.937398,
		0.889324,
		0.825244,
		0.74603,
		0.653763,
		0.551596,
		0.443626,
		0.334797,
		0.23081,
		0.138055,
		0.063557,
		0.014928,
		0.002165,
		0.023032,
		0.063557,
		0.120607,
		0.19098,
		0.271429,
		0.358694,
		0.449532,
		0.540768,
		0.629348,
		0.71242,
		0.787425,
		0.852217,
		0.905215,
		0.94558,
		0.973418,
		0.98996,
		0.997633,
		0.999824,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.003771,
		0.027301,
		0.069803,
		0.128201,
		0.199356,
		0.280089,
		0.367211,
		0.457553,
		0.548017,
		0.635631,
		0.717621,
		0.791508,
		0.855222,
		0.90725,
		0.946813,
		0.974052,
		0.990212,
		0.997695,
		0.999829,
		1,
		0.999824,
		0.997568,
		0.989431,
		0.971397,
		0.940284,
		0.894177,
		0.832511,
		0.756005,
		0.666529,
		0.56698,
		0.461161,
		0.353683,
		0.249885,
		0.155766,
		0.077932,
		0.02355,
		0.000312,
		0.005773,
		0.031834,
		0.076202,
		0.135857,
		0.207724,
		0.28869,
		0.375633,
		0.465459,
		0.555144,
		0.641793,
		0.722712,
		0.795497,
		0.858152,
		0.909232,
		0.948011,
		0.974668,
		0.990456,
		0.997755,
		0.999833,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.008146,
		0.036611,
		0.082738,
		0.143566,
		0.216079,
		0.297228,
		0.383959,
		0.473249,
		0.562148,
		0.647836,
		0.727694,
		0.799394,
		0.861011,
		0.911162,
		0.949176,
		0.975265,
		0.990693,
		0.997814,
		0.999838,
		1,
		0.999833,
		0.997695,
		0.98996,
		0.972765,
		0.943008,
		0.898772,
		0.839416,
		0.765517,
		0.678754,
		0.581788,
		0.478147,
		0.372135,
		0.268753,
		0.173636,
		0.093,
		0.033598,
		0.002685,
		0.010867,
		0.041614,
		0.0894,
		0.151319,
		0.224414,
		0.3057,
		0.392188,
		0.480925,
		0.569033,
		0.653763,
		0.732572,
		0.803203,
		0.863799,
		0.913042,
		0.950308,
		0.975845,
		0.990922,
		0.99787,
		0.999842,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.000227,
		0.013913,
		0.046826,
		0.096175,
		0.159107,
		0.232725,
		0.314104,
		0.40032,
		0.488488,
		0.575799,
		0.659576,
		0.737347,
		0.806924,
		0.86652,
		0.914874,
		0.95141,
		0.976409,
		0.991144,
		0.997925,
		0.999846,
		1,
		0.999842,
		0.997814,
		0.990456,
		0.974052,
		0.94558,
		0.903125,
		0.845978,
		0.77459,
		0.690463,
		0.596039,
		0.494592,
		0.39014,
		0.287371,
		0.191578,
		0.108615,
		0.044848,
		0.007125,
		0.000888,
		0.017263,
		0.05223,
		0.10305,
		0.166922,
		0.241006,
		0.322436,
		0.408354,
		0.495938,
		0.582448,
		0.665277,
		0.742021,
		0.810562,
		0.869175,
		0.916658,
		0.952482,
		0.976956,
		0.99136,
		0.997978,
		0.999851,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.00196,
		0.020897,
		0.057812,
		0.110014,
		0.174757,
		0.249252,
		0.330696,
		0.416289,
		0.503276,
		0.588982,
		0.670868,
		0.746597,
		0.814117,
		0.871765,
		0.918396,
		0.953525,
		0.977487,
		0.991569,
		0.998029,
		0.999854,
		1,
		0.999851,
		0.997925,
		0.990922,
		0.975265,
		0.948011,
		0.90725,
		0.852217,
		0.783246,
		0.701678,
		0.609751,
		0.510503,
		0.407688,
		0.3057,
		0.209516,
		0.124649,
		0.057105,
		0.013353,
		0.003417,
		0.024796,
		0.063557,
		0.117059,
		0.182603,
		0.25746,
		0.33888,
		0.424126,
		0.510503,
		0.595403,
		0.676351,
		0.751077,
		0.817592,
		0.874294,
		0.92009,
		0.954539,
		0.978003,
		0.991772,
		0.998079,
		0.999858,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.005236,
		0.028941,
		0.069452,
		0.124172,
		0.190456,
		0.265625,
		0.346987,
		0.431863,
		0.517621,
		0.601712,
		0.681729,
		0.755463,
		0.820988,
		0.876762,
		0.921741,
		0.955526,
		0.978505,
		0.991969,
		0.998127,
		0.999862,
		1,
		0.999858,
		0.998029,
		0.99136,
		0.976409,
		0.950308,
		0.911162,
		0.858152,
		0.791508,
		0.71242,
		0.622942,
		0.525892,
		0.424774,
		0.323712,
		0.227386,
		0.140991,
		0.070199,
		0.021119,
		0.000279,
		0.007397,
		0.033317,
		0.075484,
		0.131346,
		0.198309,
		0.273744,
		0.355015,
		0.439502,
		0.524629,
		0.607912,
		0.687003,
		0.759757,
		0.824309,
		0.87917,
		0.92335,
		0.956487,
		0.978993,
		0.99216,
		0.998174,
		0.999866,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.009877,
		0.037906,
		0.08164,
		0.138573,
		0.206156,
		0.281814,
		0.362964,
		0.447041,
		0.531529,
		0.614003,
		0.692176,
		0.763962,
		0.827555,
		0.881522,
		0.924919,
		0.957423,
		0.979466,
		0.992346,
		0.998219,
		0.999869,
		1,
		0.999866,
		0.998127,
		0.991772,
		0.977487,
		0.952482,
		0.914874,
		0.863799,
		0.799394,
		0.722712,
		0.635631,
		0.540768,
		0.441396,
		0.341382,
		0.245133,
		0.157547,
		0.083978,
		0.030205,
		0.002404,
		0.000206,
		0.012657,
		0.042694,
		0.08791,
		0.145843,
		0.213992,
		0.289832,
		0.370832,
		0.454482,
		0.538323,
		0.619989,
		0.69725,
		0.76808,
		0.830729,
		0.883818,
		0.926448,
		0.958333,
		0.979927,
		0.992526,
		0.998263,
		0.999872,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.000807,
		0.015719,
		0.047667,
		0.094282,
		0.153149,
		0.221812,
		0.297795,
		0.378617,
		0.461824,
		0.545012,
		0.625869,
		0.702226,
		0.772112,
		0.833833,
		0.886059,
		0.927939,
		0.95922,
		0.980375,
		0.9927,
		0.998306,
		0.999876,
		1,
		0.999872,
		0.998219,
		0.99216,
		0.978505,
		0.954539,
		0.918396,
		0.869175,
		0.806924,
		0.732572,
		0.647836,
		0.555144,
		0.457553,
		0.358694,
		0.262714,
		0.174234,
		0.098313,
		0.040417,
		0.006396,
		0.001783,
		0.019046,
		0.05281,
		0.100747,
		0.160485,
		0.229612,
		0.3057,
		0.38632,
		0.469068,
		0.551596,
		0.631647,
		0.707107,
		0.776061,
		0.836868,
		0.888248,
		0.929393,
		0.960083,
		0.980811,
		0.99287,
		0.998347,
		0.999879,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.003111,
		0.02262,
		0.058111,
		0.107296,
		0.167843,
		0.237387,
		0.313545,
		0.393939,
		0.476215,
		0.558077,
		0.637323,
		0.711894,
		0.779928,
		0.839836,
		0.890386,
		0.930812,
		0.960924,
		0.981234,
		0.993035,
		0.998387,
		0.999882,
		1,
		0.999879,
		0.998306,
		0.992526,
		0.979466,
		0.956487,
		0.921741,
		0.874294,
		0.814117,
		0.742021,
		0.659576,
		0.569033,
		0.473249,
		0.375633,
		0.280089,
		0.19098,
		0.113087,
		0.051585,
		0.012014,
		0.004772,
		0.026425,
		0.063557,
		0.113919,
		0.175218,
		0.245133,
		0.321329,
		0.401474,
		0.483264,
		0.564456,
		0.642901,
		0.71659,
		0.783716,
		0.842738,
		0.892474,
		0.932195,
		0.961743,
		0.981647,
		0.993195,
		0.998425,
		0.999885,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.006746,
		0.030448,
		0.069138,
		0.120607,
		0.182603,
		0.252848,
		0.32905,
		0.408924,
		0.490217,
		0.570735,
		0.64838,
		0.721196,
		0.787425,
		0.845578,
		0.894513,
		0.933544,
		0.962541,
		0.982047,
		0.99335,
		0.998463,
		0.999888,
		1,
		0.999885,
		0.998387,
		0.99287,
		0.980375,
		0.958333,
		0.924919,
		0.87917,
		0.820988,
		0.751077,
		0.670868,
		0.582448,
		0.488488,
		0.392188,
		0.297228,
		0.207724,
		0.128201,
		0.063557,
		0.019046,
		0.000251,
		0.009016,
		0.034672,
		0.074843,
		0.127354,
		0.189994,
		0.260527,
		0.336705,
		0.416289,
		0.497074,
		0.576915,
		0.653763,
		0.725714,
		0.791059,
		0.848355,
		0.896506,
		0.93486,
		0.963319,
		0.982438,
		0.993502,
		0.998499,
		0.999891,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.000188,
		0.011564,
		0.039085,
		0.08066,
		0.134151,
		0.197385,
		0.268167,
		0.344293,
		0.423569,
		0.503836,
		0.582997,
		0.659052,
		0.730146,
		0.794618,
		0.851071,
		0.898452,
		0.936145,
		0.964076,
		0.982817,
		0.993648,
		0.998535,
		0.999893,
		1,
		0.999891,
		0.998463,
		0.993195,
		0.981234,
		0.960083,
		0.927939,
		0.883818,
		0.827555,
		0.759757,
		0.681729,
		0.595403,
		0.503276,
		0.408354,
		0.314104,
		0.224414,
		0.143566,
		0.076202,
		0.027301,
		0.002165,
		0.000737,
		0.014374,
		0.043674,
		0.086581,
		0.140991,
		0.204772,
		0.275766,
		0.351814,
		0.430764,
		0.510503,
		0.588982,
		0.664249,
		0.734494,
		0.798105,
		0.853729,
		0.900354,
		0.937398,
		0.964815,
		0.983187,
		0.993791,
		0.998569,
		0.999896,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.001628,
		0.01743,
		0.048428,
		0.092596,
		0.147869,
		0.212149,
		0.283322,
		0.359265,
		0.437873,
		0.517077,
		0.594872,
		0.669354,
		0.738759,
		0.801521,
		0.856329,
		0.902212,
		0.938621,
		0.965534,
		0.983546,
		0.99393,
		0.998603,
		0.999898,
		1,
		0.999896,
		0.998535,
		0.993502,
		0.982047,
		0.961743,
		0.930812,
		0.888248,
		0.833833,
		0.76808,
		0.692176,
		0.607912,
		0.517621,
		0.424126,
		0.330696,
		0.241006,
		0.159107,
		0.0894,
		0.036611,
		0.005773,
		0.002844,
		0.020718,
		0.053333,
		0.098696,
		0.154777,
		0.219514,
		0.29083,
		0.366646,
		0.444897,
		0.523558,
		0.600669,
		0.67437,
		0.742944,
		0.804867,
		0.858873,
		0.904028,
		0.939815,
		0.966236,
		0.983896,
		0.994065,
		0.998635,
		0.999901,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.004366,
		0.024223,
		0.05838,
		0.104874,
		0.16171,
		0.226862,
		0.29829,
		0.373956,
		0.451836,
		0.529947,
		0.606372,
		0.679298,
		0.747049,
		0.808146,
		0.861363,
		0.905802,
		0.94098,
		0.96692,
		0.984237,
		0.994196,
		0.998667,
		0.999903,
		1,
		0.999901,
		0.998603,
		0.993791,
		0.982817,
		0.963319,
		0.933544,
		0.892474,
		0.839836,
		0.776061,
		0.702226,
		0.619989,
		0.531529,
		0.439502,
		0.346987,
		0.25746,
		0.174757,
		0.10305,
		0.046826,
		0.010867,
		0.006177,
		0.027933,
		0.063557,
		0.111122,
		0.168662,
		0.234189,
		0.3057,
		0.381194,
		0.45869,
		0.536244,
		0.611985,
		0.684139,
		0.751077,
		0.811359,
		0.863799,
		0.907537,
		0.942118,
		0.967586,
		0.984569,
		0.994324,
		0.998697,
		0.999906,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.008262,
		0.031834,
		0.068856,
		0.117431,
		0.175628,
		0.241492,
		0.313057,
		0.38836,
		0.465459,
		0.542452,
		0.617508,
		0.688896,
		0.755028,
		0.814507,
		0.866184,
		0.909232,
		0.943228,
		0.968237,
		0.984893,
		0.994448,
		0.998727,
		0.999908,
		1,
		0.999906,
		0.998667,
		0.994065,
		0.983546,
		0.964815,
		0.936145,
		0.896506,
		0.845578,
		0.783716,
		0.711894,
		0.631647,
		0.545012,
		0.454482,
		0.362964,
		0.273744,
		0.190456,
		0.117059,
		0.057812,
		0.017263,
		0.000227,
		0.000172,
		0.010606,
		0.035914,
		0.074267,
		0.123796,
		0.182603,
		0.248768,
		0.32036,
		0.395453,
		0.472143,
		0.54857,
		0.622942,
		0.69357,
		0.758906,
		0.817592,
		0.868517,
		0.91089,
		0.944313,
		0.968871,
		0.985208,
		0.994569,
		0.998756,
		0.99991,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.000675,
		0.013194,
		0.040162,
		0.079781,
		0.13021,
		0.189584,
		0.256014,
		0.327607,
		0.402472,
		0.478744,
		0.5546,
		0.628289,
		0.698162,
		0.76271,
		0.820615,
		0.870801,
		0.91251,
		0.945372,
		0.969489,
		0.985515,
		0.994686,
		0.998784,
		0.999912,
		1,
		0.99991,
		0.998727,
		0.994324,
		0.984237,
		0.966236,
		0.938621,
		0.900354,
		0.851071,
		0.791059,
		0.721196,
		0.642901,
		0.558077,
		0.469068,
		0.378617,
		0.289832,
		0.206156,
		0.131346,
		0.069452,
		0.024796,
		0.00196,
		0.001493,
		0.016011,
		0.044568,
		0.085389,
		0.136666,
		0.196564,
		0.263228,
		0.334797,
		0.409418,
		0.485261,
		0.560543,
		0.63355,
		0.702674,
		0.766443,
		0.823577,
		0.873037,
		0.914095,
		0.946406,
		0.970093,
		0.985814,
		0.994801,
		0.998811,
		0.999914,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.00261,
		0.019046,
		0.049119,
		0.091084,
		0.14316,
		0.203541,
		0.270407,
		0.341929,
		0.416289,
		0.491695,
		0.566399,
		0.638727,
		0.707107,
		0.770107,
		0.826481,
		0.875226,
		0.915644,
		0.947416,
		0.970681,
		0.986105,
		0.994912,
		0.998838,
		0.999916,
		1,
		0.999914,
		0.998784,
		0.994569,
		0.984893,
		0.967586,
		0.94098,
		0.904028,
		0.856329,
		0.798105,
		0.730146,
		0.653763,
		0.570735,
		0.483264,
		0.393939,
		0.3057,
		0.221812,
		0.145843,
		0.08164,
		0.033317,
		0.005236,
		0.00401,
		0.022285,
		0.053808,
		0.096858,
		0.149684,
		0.210511,
		0.277548,
		0.349001,
		0.423087,
		0.498046,
		0.57217,
		0.643821,
		0.711463,
		0.773702,
		0.829327,
		0.877369,
		0.917159,
		0.948403,
		0.971256,
		0.986389,
		0.99502,
		0.998863,
		0.999918,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.005678,
		0.025717,
		0.058623,
		0.102704,
		0.156234,
		0.21747,
		0.28465,
		0.356013,
		0.429809,
		0.504315,
		0.577857,
		0.648832,
		0.715743,
		0.77723,
		0.832117,
		0.879467,
		0.918641,
		0.949367,
		0.971816,
		0.986666,
		0.995126,
		0.998888,
		0.99992,
		1,
		0.999918,
		0.998838,
		0.994801,
		0.985515,
		0.968871,
		0.943228,
		0.907537,
		0.861363,
		0.804867,
		0.738759,
		0.664249,
		0.582997,
		0.497074,
		0.408924,
		0.321329,
		0.237387,
		0.160485,
		0.094282,
		0.042694,
		0.009877,
		0.0076,
		0.02933,
		0.063557,
		0.108615,
		0.162806,
		0.224414,
		0.29171,
		0.362964,
		0.436458,
		0.510503,
		0.583461,
		0.653763,
		0.719949,
		0.780692,
		0.834852,
		0.881522,
		0.92009,
		0.950308,
		0.972363,
		0.986936,
		0.995229,
		0.998913,
		0.999922,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000158,
		0.009762,
		0.033113,
		0.068601,
		0.114585,
		0.169394,
		0.231342,
		0.298728,
		0.369853,
		0.443032,
		0.516611,
		0.588982,
		0.658615,
		0.724081,
		0.78409,
		0.837533,
		0.883534,
		0.921508,
		0.951229,
		0.972897,
		0.987199,
		0.995329,
		0.998936,
		0.999923,
		1,
		0.999922,
		0.998888,
		0.99502,
		0.986105,
		0.970093,
		0.945372,
		0.91089,
		0.866184,
		0.811359,
		0.747049,
		0.67437,
		0.594872,
		0.510503,
		0.423569,
		0.336705,
		0.252848,
		0.175218,
		0.107296,
		0.05281,
		0.015719,
		0.000206,
		0.000621,
		0.012153,
		0.037057,
		0.073747,
		0.120607,
		0.175995,
		0.238249,
		0.3057,
		0.376679,
		0.449532,
		0.522638,
		0.594422,
		0.663389,
		0.728142,
		0.787425,
		0.840161,
		0.885504,
		0.922895,
		0.952128,
		0.973418,
		0.987455,
		0.995426,
		0.99896,
		0.999925,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.001374,
		0.014759,
		0.04115,
		0.078987,
		0.126677,
		0.182603,
		0.245133,
		0.312626,
		0.383441,
		0.455958,
		0.528585,
		0.599783,
		0.668086,
		0.732133,
		0.790699,
		0.842738,
		0.887434,
		0.924251,
		0.953007,
		0.973927,
		0.987705,
		0.995521,
		0.998982,
		0.999927,
		1,
		0.999925,
		0.998936,
		0.995229,
		0.986666,
		0.971256,
		0.947416,
		0.914095,
		0.870801,
		0.817592,
		0.755028,
		0.684139,
		0.606372,
		0.523558,
		0.437873,
		0.351814,
		0.268167,
		0.189994,
		0.120607,
		0.063557,
		0.02262,
		0.001783,
		0.002404,
		0.017568,
		0.045385,
		0.084314,
		0.132788,
		0.189216,
		0.251992,
		0.319503,
		0.39014,
		0.46231,
		0.534454,
		0.605064,
		0.672708,
		0.736055,
		0.793912,
		0.845265,
		0.889324,
		0.925579,
		0.953866,
		0.974424,
		0.987949,
		0.995614,
		0.999004,
		0.999929,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.003696,
		0.02057,
		0.049751,
		0.08972,
		0.138935,
		0.19583,
		0.258824,
		0.326332,
		0.396775,
		0.468589,
		0.540245,
		0.610268,
		0.677255,
		0.739908,
		0.797066,
		0.847743,
		0.891175,
		0.926878,
		0.954706,
		0.974909,
		0.988187,
		0.995704,
		0.999025,
		0.99993,
		1,
		0.999929,
		0.998982,
		0.995426,
		0.987199,
		0.972363,
		0.949367,
		0.917159,
		0.875226,
		0.823577,
		0.76271,
		0.69357,
		0.617508,
		0.536244,
		0.451836,
		0.366646,
		0.283322,
		0.204772,
		0.134151,
		0.074843,
		0.030448,
		0.004772,
		0.005236,
		0.023754,
		0.05424,
		0.0952,
		0.145114,
		0.20244,
		0.265625,
		0.333111,
		0.403344,
		0.474793,
		0.545959,
		0.615394,
		0.681729,
		0.743696,
		0.800163,
		0.850172,
		0.892988,
		0.928149,
		0.955526,
		0.975383,
		0.98842,
		0.995792,
		0.999046,
		0.999932,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.007014,
		0.02711,
		0.058845,
		0.100747,
		0.151319,
		0.209044,
		0.272394,
		0.339837,
		0.409849,
		0.480925,
		0.551596,
		0.620445,
		0.686131,
		0.747418,
		0.803203,
		0.852555,
		0.894765,
		0.929393,
		0.956329,
		0.975845,
		0.988646,
		0.995877,
		0.999066,
		0.999933,
		1,
		0.999932,
		0.999025,
		0.995614,
		0.987705,
		0.973418,
		0.951229,
		0.92009,
		0.879467,
		0.829327,
		0.770107,
		0.702674,
		0.628289,
		0.54857,
		0.465459,
		0.381194,
		0.29829,
		0.219514,
		0.147869,
		0.086581,
		0.039085,
		0.009016,
		0.000146,
		0.009016,
		0.030627,
		0.063557,
		0.106356,
		0.157547,
		0.21564,
		0.27913,
		0.346512,
		0.416289,
		0.486985,
		0.557157,
		0.62542,
		0.690463,
		0.751077,
		0.806187,
		0.854892,
		0.896506,
		0.930611,
		0.957114,
		0.976297,
		0.988867,
		0.995961,
		0.999086,
		0.999935,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000573,
		0.011231,
		0.034297,
		0.06837,
		0.112019,
		0.163793,
		0.222223,
		0.28583,
		0.353134,
		0.422664,
		0.492972,
		0.562644,
		0.630323,
		0.694725,
		0.754672,
		0.809117,
		0.857183,
		0.898211,
		0.931803,
		0.957881,
		0.976739,
		0.989083,
		0.996042,
		0.999105,
		0.999936,
		1,
		0.999935,
		0.999066,
		0.995792,
		0.988187,
		0.974424,
		0.953007,
		0.922895,
		0.883534,
		0.834852,
		0.77723,
		0.711463,
		0.638727,
		0.560543,
		0.478744,
		0.395453,
		0.313057,
		0.234189,
		0.16171,
		0.098696,
		0.048428,
		0.014374,
		0.000188,
		0.001269,
		0.013648,
		0.03811,
		0.073275,
		0.117733,
		0.170054,
		0.228792,
		0.292492,
		0.359701,
		0.428973,
		0.498887,
		0.568056,
		0.635152,
		0.698919,
		0.758206,
		0.811993,
		0.859431,
		0.899882,
		0.93297,
		0.958631,
		0.97717,
		0.989293,
		0.996121,
		0.999123,
		0.999937,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.002221,
		0.016256,
		0.042059,
		0.078267,
		0.123492,
		0.176325,
		0.235343,
		0.299116,
		0.366214,
		0.435217,
		0.504731,
		0.573396,
		0.63991,
		0.703046,
		0.76168,
		0.814818,
		0.861636,
		0.90152,
		0.934112,
		0.959365,
		0.977591,
		0.989499,
		0.996199,
		0.999141,
		0.999939,
		1,
		0.999937,
		0.999105,
		0.995961,
		0.988646,
		0.975383,
		0.954706,
		0.925579,
		0.887434,
		0.840161,
		0.78409,
		0.719949,
		0.648832,
		0.57217,
		0.491695,
		0.409418,
		0.327607,
		0.248768,
		0.175628,
		0.111122,
		0.05838,
		0.020718,
		0.001628,
		0.003417,
		0.019046,
		0.046135,
		0.083339,
		0.129291,
		0.182603,
		0.241875,
		0.3057,
		0.372671,
		0.441396,
		0.510503,
		0.578663,
		0.644597,
		0.707107,
		0.765094,
		0.817592,
		0.863799,
		0.903125,
		0.935231,
		0.960083,
		0.978003,
		0.9897,
		0.996274,
		0.999159,
		0.99994,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.004845,
		0.022008,
		0.05033,
		0.088485,
		0.135126,
		0.188885,
		0.248386,
		0.312242,
		0.379073,
		0.447509,
		0.516206,
		0.583858,
		0.649214,
		0.711103,
		0.76845,
		0.820315,
		0.865921,
		0.904698,
		0.936326,
		0.960786,
		0.978406,
		0.989896,
		0.996348,
		0.999176,
		0.999941,
		1,
		0.99994,
		0.999141,
		0.996121,
		0.989083,
		0.976297,
		0.956329,
		0.928149,
		0.891175,
		0.845265,
		0.790699,
		0.728142,
		0.658615,
		0.583461,
		0.504315,
		0.423087,
		0.341929,
		0.263228,
		0.189584,
		0.123796,
		0.068856,
		0.027933,
		0.004366,
		0.006493,
		0.025132,
		0.054636,
		0.093699,
		0.140991,
		0.195168,
		0.254872,
		0.318742,
		0.385418,
		0.453557,
		0.521839,
		0.588982,
		0.653763,
		0.715035,
		0.771749,
		0.82299,
		0.868003,
		0.90624,
		0.937398,
		0.961473,
		0.978799,
		0.990087,
		0.996419,
		0.999193,
		0.999943,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.000135,
		0.008351,
		0.02841,
		0.059048,
		0.098976,
		0.146884,
		0.201449,
		0.261333,
		0.325198,
		0.391707,
		0.45954,
		0.527402,
		0.594037,
		0.658245,
		0.718904,
		0.774992,
		0.825616,
		0.870045,
		0.907751,
		0.938448,
		0.962145,
		0.979184,
		0.990274,
		0.99649,
		0.99921,
		0.999944,
		1,
		0.999943,
		0.999176,
		0.996274,
		0.989499,
		0.97717,
		0.957881,
		0.930611,
		0.894765,
		0.850172,
		0.797066,
		0.736055,
		0.668086,
		0.594422,
		0.516611,
		0.436458,
		0.356013,
		0.277548,
		0.203541,
		0.136666,
		0.079781,
		0.035914,
		0.008262,
		0.00053,
		0.010409,
		0.031834,
		0.063557,
		0.10431,
		0.1528,
		0.207724,
		0.267766,
		0.331609,
		0.397939,
		0.465459,
		0.532897,
		0.599022,
		0.66266,
		0.722712,
		0.77818,
		0.828196,
		0.872049,
		0.909232,
		0.939477,
		0.962803,
		0.97956,
		0.990456,
		0.996558,
		0.999225,
		0.999945,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.001176,
		0.012657,
		0.035395,
		0.068159,
		0.109696,
		0.158736,
		0.213992,
		0.27417,
		0.337974,
		0.404113,
		0.471312,
		0.538323,
		0.603939,
		0.667009,
		0.726459,
		0.781314,
		0.830729,
		0.874016,
		0.910685,
		0.940484,
		0.963446,
		0.979927,
		0.990634,
		0.996625,
		0.999241,
		0.999946,
		1,
		0.999945,
		0.99921,
		0.996419,
		0.989896,
		0.978003,
		0.959365,
		0.93297,
		0.898211,
		0.854892,
		0.803203,
		0.743696,
		0.677255,
		0.605064,
		0.528585,
		0.449532,
		0.369853,
		0.29171,
		0.21747,
		0.149684,
		0.091084,
		0.044568,
		0.013194,
		0.000172,
		0.002059,
		0.015085,
		0.039085,
		0.072845,
		0.11513,
		0.164687,
		0.22025,
		0.280543,
		0.344293,
		0.41023,
		0.477101,
		0.543682,
		0.608789,
		0.671294,
		0.730146,
		0.784396,
		0.833218,
		0.875946,
		0.912108,
		0.941471,
		0.964076,
		0.980287,
		0.990808,
		0.99669,
		0.999256,
		0.999947,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.003169,
		0.017685,
		0.042898,
		0.077611,
		0.120607,
		0.170651,
		0.226495,
		0.286884,
		0.350565,
		0.416289,
		0.482827,
		0.548975,
		0.613572,
		0.675515,
		0.733775,
		0.787425,
		0.835662,
		0.877839,
		0.913505,
		0.942438,
		0.964693,
		0.980638,
		0.990978,
		0.996753,
		0.999271,
		0.999948,
		1,
		0.999947,
		0.999241,
		0.996558,
		0.990274,
		0.978799,
		0.960786,
		0.935231,
		0.90152,
		0.859431,
		0.809117,
		0.751077,
		0.686131,
		0.615394,
		0.540245,
		0.46231,
		0.383441,
		0.3057,
		0.231342,
		0.162806,
		0.102704,
		0.053808,
		0.019046,
		0.001493,
		0.004495,
		0.020447,
		0.046826,
		0.082451,
		0.126123,
		0.176624,
		0.232725,
		0.293191,
		0.356789,
		0.422291,
		0.488488,
		0.554201,
		0.618289,
		0.679673,
		0.737347,
		0.790404,
		0.838063,
		0.879698,
		0.914874,
		0.943385,
		0.965297,
		0.980982,
		0.991144,
		0.996815,
		0.999285,
		0.999949,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.006028,
		0.023363,
		0.050862,
		0.08736,
		0.131674,
		0.182603,
		0.238939,
		0.299464,
		0.362964,
		0.428234,
		0.494086,
		0.559361,
		0.622942,
		0.68377,
		0.740862,
		0.793332,
		0.840421,
		0.881522,
		0.916216,
		0.944313,
		0.965887,
		0.981318,
		0.991307,
		0.996876,
		0.999299,
		0.99995,
		1,
		0.999949,
		0.999271,
		0.99669,
		0.990634,
		0.97956,
		0.962145,
		0.937398,
		0.904698,
		0.863799,
		0.814818,
		0.758206,
		0.694725,
		0.62542,
		0.551596,
		0.474793,
		0.396775,
		0.319503,
		0.245133,
		0.175995,
		0.114585,
		0.063557,
		0.025717,
		0.00401,
		0.000125,
		0.007758,
		0.026425,
		0.055,
		0.092332,
		0.137255,
		0.188586,
		0.245133,
		0.3057,
		0.369091,
		0.43412,
		0.499621,
		0.564456,
		0.62753,
		0.687806,
		0.744321,
		0.796212,
		0.842738,
		0.883312,
		0.917532,
		0.945222,
		0.966466,
		0.981647,
		0.991466,
		0.996935,
		0.999313,
		0.999951,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.000492,
		0.009675,
		0.029626,
		0.059234,
		0.097363,
		0.142864,
		0.19457,
		0.251308,
		0.311899,
		0.375167,
		0.439948,
		0.505093,
		0.569487,
		0.632056,
		0.691782,
		0.747726,
		0.799043,
		0.845015,
		0.88507,
		0.918824,
		0.946113,
		0.967032,
		0.981968,
		0.991621,
		0.996992,
		0.999327,
		0.999952,
		1,
		0.999951,
		0.999299,
		0.996815,
		0.990978,
		0.980287,
		0.963446,
		0.939477,
		0.907751,
		0.868003,
		0.820315,
		0.765094,
		0.703046,
		0.635152,
		0.562644,
		0.486985,
		0.409849,
		0.333111,
		0.258824,
		0.189216,
		0.126677,
		0.073747,
		0.033113,
		0.0076,
		0.001092,
		0.011771,
		0.032959,
		0.063557,
		0.102448,
		0.148496,
		0.200552,
		0.25746,
		0.31806,
		0.381194,
		0.445718,
		0.510503,
		0.574455,
		0.636519,
		0.695699,
		0.751077,
		0.801828,
		0.847251,
		0.886795,
		0.92009,
		0.946986,
		0.967586,
		0.982283,
		0.991772,
		0.997049,
		0.99934,
		0.999953,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.001913,
		0.014037,
		0.036415,
		0.067965,
		0.107582,
		0.154148,
		0.206529,
		0.263588,
		0.324181,
		0.387171,
		0.45143,
		0.515852,
		0.579359,
		0.64092,
		0.699558,
		0.754375,
		0.804566,
		0.849448,
		0.888488,
		0.921332,
		0.947842,
		0.968129,
		0.982591,
		0.99192,
		0.997104,
		0.999352,
		0.999954,
		1,
		0.999953,
		0.999327,
		0.996935,
		0.991307,
		0.980982,
		0.964693,
		0.941471,
		0.910685,
		0.872049,
		0.825616,
		0.771749,
		0.711103,
		0.644597,
		0.573396,
		0.498887,
		0.422664,
		0.346512,
		0.272394,
		0.20244,
		0.138935,
		0.084314,
		0.04115,
		0.012153,
		0.000158,
		0.002946,
		0.016464,
		0.039989,
		0.072452,
		0.112762,
		0.159817,
		0.2125,
		0.269691,
		0.330263,
		0.393097,
		0.457085,
		0.521138,
		0.584202,
		0.645261,
		0.703361,
		0.757621,
		0.807258,
		0.851607,
		0.890151,
		0.922551,
		0.94868,
		0.968661,
		0.982892,
		0.992065,
		0.997158,
		0.999365,
		0.999955,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.004182,
		0.019046,
		0.043674,
		0.077011,
		0.117983,
		0.1655,
		0.218463,
		0.275766,
		0.336303,
		0.398972,
		0.462682,
		0.526364,
		0.588982,
		0.649542,
		0.707107,
		0.760817,
		0.809907,
		0.853729,
		0.891784,
		0.923746,
		0.949502,
		0.969182,
		0.983187,
		0.992207,
		0.99721,
		0.999377,
		0.999956,
		1,
		0.999955,
		0.999352,
		0.997049,
		0.991621,
		0.981647,
		0.965887,
		0.943385,
		0.913505,
		0.875946,
		0.830729,
		0.77818,
		0.718904,
		0.653763,
		0.583858,
		0.510503,
		0.435217,
		0.359701,
		0.28583,
		0.21564,
		0.151319,
		0.0952,
		0.049751,
		0.017568,
		0.001374,
		0.005611,
		0.021774,
		0.047464,
		0.08164,
		0.123241,
		0.171193,
		0.224414,
		0.281814,
		0.342303,
		0.404796,
		0.468221,
		0.531529,
		0.593702,
		0.653763,
		0.710798,
		0.763962,
		0.812511,
		0.855814,
		0.893386,
		0.924919,
		0.950308,
		0.969692,
		0.983475,
		0.992346,
		0.997262,
		0.999389,
		0.999957,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.000116,
		0.007225,
		0.024641,
		0.051353,
		0.086332,
		0.128533,
		0.176896,
		0.230353,
		0.287832,
		0.34826,
		0.410568,
		0.473704,
		0.536635,
		0.598361,
		0.657927,
		0.714434,
		0.767059,
		0.815072,
		0.857862,
		0.89496,
		0.926069,
		0.951098,
		0.970192,
		0.983758,
		0.992481,
		0.997312,
		0.9994,
		0.999958,
		1,
		0.999957,
		0.999377,
		0.997158,
		0.99192,
		0.982283,
		0.967032,
		0.945222,
		0.916216,
		0.879698,
		0.835662,
		0.784396,
		0.726459,
		0.66266,
		0.594037,
		0.521839,
		0.447509,
		0.372671,
		0.299116,
		0.228792,
		0.163793,
		0.106356,
		0.058845,
		0.023754,
		0.003696,
		0.000458,
		0.009016,
		0.027641,
		0.055335,
		0.091084,
		0.133854,
		0.182603,
		0.236278,
		0.29382,
		0.354174,
		0.416289,
		0.479129,
		0.541681,
		0.602961,
		0.662033,
		0.718017,
		0.770107,
		0.817592,
		0.859876,
		0.896506,
		0.927198,
		0.951873,
		0.970681,
		0.984034,
		0.992614,
		0.997361,
		0.999412,
		0.999959,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.001017,
		0.010974,
		0.030765,
		0.059405,
		0.09589,
		0.139203,
		0.188314,
		0.242186,
		0.299776,
		0.360045,
		0.421959,
		0.484498,
		0.546667,
		0.607502,
		0.666082,
		0.721547,
		0.773107,
		0.820069,
		0.861854,
		0.898023,
		0.928306,
		0.952633,
		0.971161,
		0.984305,
		0.992743,
		0.997409,
		0.999423,
		0.999959,
		1,
		0.999959,
		0.9994,
		0.997262,
		0.992207,
		0.982892,
		0.968129,
		0.946986,
		0.918824,
		0.883312,
		0.840421,
		0.790404,
		0.733775,
		0.671294,
		0.603939,
		0.532897,
		0.45954,
		0.385418,
		0.312242,
		0.241875,
		0.176325,
		0.117733,
		0.06837,
		0.030627,
		0.007014,
		0.001783,
		0.013093,
		0.034009,
		0.063557,
		0.100747,
		0.144575,
		0.194026,
		0.248076,
		0.3057,
		0.365872,
		0.427577,
		0.489811,
		0.551596,
		0.611985,
		0.670076,
		0.725025,
		0.776061,
		0.822507,
		0.863799,
		0.899514,
		0.929393,
		0.953377,
		0.971631,
		0.984569,
		0.99287,
		0.997456,
		0.999434,
		0.99996,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.002747,
		0.015366,
		0.037366,
		0.067787,
		0.105651,
		0.149968,
		0.199736,
		0.253947,
		0.31159,
		0.371655,
		0.433143,
		0.495067,
		0.556466,
		0.61641,
		0.674014,
		0.728452,
		0.778969,
		0.824905,
		0.865711,
		0.900978,
		0.93046,
		0.954108,
		0.972091,
		0.984829,
		0.992994,
		0.997501,
		0.999444,
		0.999961,
		1,
		0.99996,
		0.999423,
		0.997361,
		0.992481,
		0.983475,
		0.969182,
		0.94868,
		0.921332,
		0.886795,
		0.845015,
		0.796212,
		0.740862,
		0.679673,
		0.613572,
		0.543682,
		0.471312,
		0.397939,
		0.325198,
		0.254872,
		0.188885,
		0.129291,
		0.078267,
		0.03811,
		0.011231,
		0.000146,
		0.003901,
		0.017784,
		0.04083,
		0.07209,
		0.110599,
		0.155379,
		0.205443,
		0.259797,
		0.317445,
		0.377394,
		0.438658,
		0.500268,
		0.561279,
		0.620779,
		0.677898,
		0.731828,
		0.781832,
		0.827263,
		0.86759,
		0.902416,
		0.931507,
		0.954824,
		0.972543,
		0.985083,
		0.993115,
		0.997546,
		0.999454,
		0.999962,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.005236,
		0.020342,
		0.044395,
		0.076461,
		0.115585,
		0.160804,
		0.211144,
		0.265625,
		0.323265,
		0.383087,
		0.444121,
		0.505413,
		0.566035,
		0.625091,
		0.681729,
		0.735156,
		0.78465,
		0.829583,
		0.869437,
		0.903828,
		0.932535,
		0.955526,
		0.972985,
		0.985332,
		0.993234,
		0.99759,
		0.999464,
		0.999962,
		1,
		0.999962,
		0.999444,
		0.997456,
		0.992743,
		0.984034,
		0.970192,
		0.950308,
		0.923746,
		0.890151,
		0.849448,
		0.801828,
		0.747726,
		0.687806,
		0.622942,
		0.554201,
		0.482827,
		0.41023,
		0.337974,
		0.267766,
		0.201449,
		0.140991,
		0.088485,
		0.046135,
		0.016256,
		0.001269,
		0.000108,
		0.006746,
		0.023032,
		0.048056,
		0.080895,
		0.120607,
		0.166242,
		0.216838,
		0.271429,
		0.32905,
		0.388735,
		0.449532,
		0.510503,
		0.570735,
		0.629348,
		0.685507,
		0.738434,
		0.787425,
		0.831866,
		0.871252,
		0.905215,
		0.933544,
		0.956215,
		0.973418,
		0.985575,
		0.99335,
		0.997633,
		0.999474,
		0.999963,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.000428,
		0.008421,
		0.025848,
		0.051808,
		0.085389,
		0.125662,
		0.171689,
		0.222522,
		0.277209,
		0.334797,
		0.394338,
		0.454892,
		0.515539,
		0.575379,
		0.63355,
		0.689233,
		0.741665,
		0.790157,
		0.834112,
		0.873037,
		0.906578,
		0.934534,
		0.956891,
		0.973843,
		0.985814,
		0.993464,
		0.997675,
		0.999484,
		0.999964,
		1,
		0.999963,
		0.999464,
		0.997546,
		0.992994,
		0.984569,
		0.971161,
		0.951873,
		0.926069,
		0.893386,
		0.853729,
		0.807258,
		0.754375,
		0.695699,
		0.632056,
		0.564456,
		0.494086,
		0.422291,
		0.350565,
		0.280543,
		0.213992,
		0.1528,
		0.098976,
		0.054636,
		0.022008,
		0.003417,
		0.000949,
		0.010256,
		0.028784,
		0.055645,
		0.089938,
		0.130747,
		0.177144,
		0.228195,
		0.282963,
		0.340507,
		0.399895,
		0.460201,
		0.52052,
		0.579968,
		0.637698,
		0.692907,
		0.744848,
		0.792848,
		0.836321,
		0.874792,
		0.907917,
		0.935507,
		0.957554,
		0.97426,
		0.986047,
		0.993576,
		0.997715,
		0.999493,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.001665,
		0.012242,
		0.031834,
		0.059563,
		0.094539,
		0.135857,
		0.182603,
		0.233856,
		0.28869,
		0.34618,
		0.405406,
		0.465459,
		0.525446,
		0.584502,
		0.641793,
		0.696531,
		0.747985,
		0.795497,
		0.838495,
		0.876517,
		0.909232,
		0.936461,
		0.958205,
		0.974668,
		0.986276,
		0.993685,
		0.997755,
		0.999502,
		0.999965,
		1,
		0.999964,
		0.999484,
		0.997633,
		0.993234,
		0.985083,
		0.972091,
		0.953377,
		0.928306,
		0.896506,
		0.857862,
		0.812511,
		0.760817,
		0.703361,
		0.64092,
		0.574455,
		0.505093,
		0.43412,
		0.362964,
		0.293191,
		0.226495,
		0.164687,
		0.109696,
		0.063557,
		0.02841,
		0.006493,
		0.002567,
		0.014374,
		0.034993,
		0.063557,
		0.099188,
		0.140991,
		0.188066,
		0.239503,
		0.294389,
		0.351814,
		0.410871,
		0.470665,
		0.53032,
		0.588982,
		0.645835,
		0.700105,
		0.751077,
		0.798105,
		0.840634,
		0.878214,
		0.910525,
		0.937398,
		0.958843,
		0.975068,
		0.986501,
		0.993791,
		0.997795,
		0.999511,
		0.999966,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.003647,
		0.016644,
		0.038254,
		0.067623,
		0.103881,
		0.146146,
		0.193529,
		0.245133,
		0.300059,
		0.357409,
		0.416289,
		0.475821,
		0.535139,
		0.593409,
		0.649825,
		0.70363,
		0.754123,
		0.800673,
		0.842738,
		0.879882,
		0.911794,
		0.938318,
		0.959469,
		0.97546,
		0.98672,
		0.993896,
		0.997833,
		0.99952,
		0.999966,
		1,
		0.999966,
		0.999502,
		0.997715,
		0.993464,
		0.985575,
		0.972985,
		0.954824,
		0.93046,
		0.899514,
		0.861854,
		0.817592,
		0.767059,
		0.710798,
		0.649542,
		0.584202,
		0.515852,
		0.445718,
		0.375167,
		0.3057,
		0.238939,
		0.176624,
		0.120607,
		0.072845,
		0.035395,
		0.010409,
		0.000135,
		0.004898,
		0.019046,
		0.041614,
		0.071756,
		0.108615,
		0.151319,
		0.198992,
		0.250747,
		0.3057,
		0.362964,
		0.421662,
		0.480925,
		0.539907,
		0.597782,
		0.653763,
		0.707107,
		0.757125,
		0.803203,
		0.844809,
		0.881522,
		0.913042,
		0.939222,
		0.960083,
		0.975845,
		0.986936,
		0.993998,
		0.99787,
		0.999529,
		0.999967,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.000101,
		0.006313,
		0.021574,
		0.045066,
		0.075953,
		0.113386,
		0.156508,
		0.204451,
		0.256343,
		0.31131,
		0.36848,
		0.426988,
		0.48598,
		0.544621,
		0.602103,
		0.657651,
		0.710536,
		0.760084,
		0.805693,
		0.846847,
		0.883135,
		0.914268,
		0.940109,
		0.960686,
		0.976223,
		0.987147,
		0.994098,
		0.997907,
		0.999537,
		0.999968,
		1,
		0.999967,
		0.99952,
		0.997795,
		0.993685,
		0.986047,
		0.973843,
		0.956215,
		0.932535,
		0.902416,
		0.865711,
		0.822507,
		0.773107,
		0.718017,
		0.657927,
		0.593702,
		0.526364,
		0.457085,
		0.387171,
		0.31806,
		0.251308,
		0.188586,
		0.131674,
		0.082451,
		0.042898,
		0.015085,
		0.001176,
		0.0004,
		0.007884,
		0.024223,
		0.048607,
		0.080209,
		0.118192,
		0.16171,
		0.209905,
		0.261919,
		0.316889,
		0.373956,
		0.432268,
		0.490984,
		0.549284,
		0.606372,
		0.661489,
		0.713918,
		0.763,
		0.808146,
		0.848853,
		0.884721,
		0.915473,
		0.94098,
		0.961278,
		0.976593,
		0.987353,
		0.994196,
		0.997943,
		0.999545,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.000888,
		0.009606,
		0.026987,
		0.05223,
		0.08452,
		0.12303,
		0.166922,
		0.215353,
		0.267474,
		0.322436,
		0.379391,
		0.437501,
		0.495938,
		0.553895,
		0.61059,
		0.665277,
		0.717253,
		0.765874,
		0.810562,
		0.850827,
		0.886281,
		0.916658,
		0.941836,
		0.961859,
		0.976956,
		0.987556,
		0.994292,
		0.997978,
		0.999553,
		0.999969,
		1,
		0.999968,
		0.999537,
		0.99787,
		0.993896,
		0.986501,
		0.974668,
		0.957554,
		0.934534,
		0.905215,
		0.869437,
		0.827263,
		0.778969,
		0.725025,
		0.666082,
		0.602961,
		0.536635,
		0.468221,
		0.398972,
		0.330263,
		0.263588,
		0.200552,
		0.142864,
		0.092332,
		0.050862,
		0.020447,
		0.003169,
		0.001559,
		0.011471,
		0.02986,
		0.055933,
		0.088884,
		0.127896,
		0.172143,
		0.220793,
		0.273008,
		0.327951,
		0.384786,
		0.442688,
		0.500843,
		0.558455,
		0.614757,
		0.669016,
		0.720543,
		0.768706,
		0.812941,
		0.852769,
		0.887815,
		0.917822,
		0.942676,
		0.962429,
		0.977312,
		0.987755,
		0.994386,
		0.998013,
		0.999561,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.002404,
		0.013475,
		0.032838,
		0.05971,
		0.093296,
		0.132788,
		0.177371,
		0.226223,
		0.278519,
		0.333432,
		0.39014,
		0.447829,
		0.505698,
		0.562964,
		0.618874,
		0.672708,
		0.723788,
		0.771497,
		0.815284,
		0.854681,
		0.889324,
		0.918966,
		0.943502,
		0.962988,
		0.977661,
		0.987949,
		0.994478,
		0.998046,
		0.999569,
		0.99997,
		1,
		0.999969,
		0.999553,
		0.997943,
		0.994098,
		0.986936,
		0.97546,
		0.958843,
		0.936461,
		0.907917,
		0.873037,
		0.831866,
		0.78465,
		0.731828,
		0.674014,
		0.611985,
		0.546667,
		0.479129,
		0.410568,
		0.342303,
		0.275766,
		0.2125,
		0.154148,
		0.102448,
		0.059234,
		0.026425,
		0.006028,
		0.003417,
		0.015609,
		0.035914,
		0.063557,
		0.097753,
		0.137703,
		0.182603,
		0.231643,
		0.284006,
		0.33888,
		0.395453,
		0.452924,
		0.510503,
		0.567424,
		0.622942,
		0.676351,
		0.726989,
		0.774249,
		0.817592,
		0.856563,
		0.890808,
		0.92009,
		0.944313,
		0.963537,
		0.978003,
		0.98814,
		0.994569,
		0.998079,
		0.999576,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.004591,
		0.01787,
		0.039085,
		0.067471,
		0.102252,
		0.142639,
		0.187838,
		0.237049,
		0.289469,
		0.344293,
		0.400724,
		0.457972,
		0.51526,
		0.571833,
		0.62696,
		0.679948,
		0.730146,
		0.77696,
		0.819865,
		0.858415,
		0.892267,
		0.921195,
		0.945109,
		0.964076,
		0.978339,
		0.988327,
		0.994657,
		0.998111,
		0.999584,
		0.999971,
		1,
		0.99997,
		0.999569,
		0.998013,
		0.994292,
		0.987353,
		0.976223,
		0.960083,
		0.938318,
		0.910525,
		0.876517,
		0.836321,
		0.790157,
		0.738434,
		0.681729,
		0.620779,
		0.556466,
		0.489811,
		0.421959,
		0.354174,
		0.287832,
		0.224414,
		0.1655,
		0.112762,
		0.067965,
		0.032959,
		0.009675,
		0.000125,
		9.5e-005,
		0.00592,
		0.020251,
		0.042346,
		0.071448,
		0.106789,
		0.147593,
		0.193074,
		0.242442,
		0.294906,
		0.349672,
		0.405954,
		0.462975,
		0.519969,
		0.576193,
		0.63093,
		0.683499,
		0.73326,
		0.779633,
		0.822103,
		0.860238,
		0.893703,
		0.922282,
		0.945892,
		0.964606,
		0.978669,
		0.988511,
		0.994744,
		0.998143,
		0.999591,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.000375,
		0.007397,
		0.022747,
		0.045692,
		0.075484,
		0.111363,
		0.152563,
		0.198309,
		0.247821,
		0.300316,
		0.355015,
		0.411143,
		0.467931,
		0.524629,
		0.580504,
		0.634852,
		0.687003,
		0.736332,
		0.782268,
		0.824309,
		0.862033,
		0.895116,
		0.92335,
		0.946661,
		0.965125,
		0.978993,
		0.988691,
		0.994829,
		0.998174,
		0.999598,
		0.999972,
		1,
		0.999971,
		0.999584,
		0.998079,
		0.994478,
		0.987755,
		0.976956,
		0.961278,
		0.940109,
		0.913042,
		0.879882,
		0.840634,
		0.795497,
		0.744848,
		0.689233,
		0.629348,
		0.566035,
		0.500268,
		0.433143,
		0.365872,
		0.299776,
		0.236278,
		0.176896,
		0.123241,
		0.077011,
		0.039989,
		0.014037,
		0.001092,
		0.000833,
		0.009016,
		0.025353,
		0.049119,
		0.079575,
		0.11597,
		0.157547,
		0.203541,
		0.253183,
		0.3057,
		0.360324,
		0.416289,
		0.472842,
		0.529241,
		0.584767,
		0.638727,
		0.690463,
		0.739362,
		0.784865,
		0.826481,
		0.863799,
		0.896506,
		0.9244,
		0.947416,
		0.965636,
		0.97931,
		0.988867,
		0.994912,
		0.998204,
		0.999605,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.001462,
		0.010771,
		0.028064,
		0.052623,
		0.083718,
		0.120607,
		0.162543,
		0.208769,
		0.258527,
		0.311056,
		0.365596,
		0.421394,
		0.477707,
		0.533806,
		0.588982,
		0.642555,
		0.693878,
		0.742351,
		0.787425,
		0.828621,
		0.865538,
		0.897873,
		0.925433,
		0.948158,
		0.966137,
		0.979622,
		0.98904,
		0.994994,
		0.998234,
		0.999611,
		0.999973,
		1,
		0.999972,
		0.999598,
		0.998143,
		0.994657,
		0.98814,
		0.977661,
		0.962429,
		0.941836,
		0.915473,
		0.883135,
		0.844809,
		0.800673,
		0.751077,
		0.696531,
		0.637698,
		0.575379,
		0.510503,
		0.444121,
		0.377394,
		0.31159,
		0.248076,
		0.188314,
		0.133854,
		0.086332,
		0.047464,
		0.019046,
		0.002946,
		0.002256,
		0.012657,
		0.030875,
		0.0562,
		0.08791,
		0.125272,
		0.167548,
		0.213992,
		0.263854,
		0.316383,
		0.370832,
		0.426457,
		0.482527,
		0.538323,
		0.59315,
		0.646337,
		0.69725,
		0.745299,
		0.789949,
		0.830729,
		0.867251,
		0.899218,
		0.926448,
		0.948887,
		0.966629,
		0.979927,
		0.98921,
		0.995074,
		0.998263,
		0.999618,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.003208,
		0.014668,
		0.033782,
		0.059846,
		0.092147,
		0.129963,
		0.172561,
		0.219207,
		0.269161,
		0.321682,
		0.376031,
		0.431479,
		0.487302,
		0.542794,
		0.59727,
		0.650073,
		0.700578,
		0.748208,
		0.792437,
		0.832806,
		0.868936,
		0.900541,
		0.927446,
		0.949604,
		0.967112,
		0.980227,
		0.989376,
		0.995152,
		0.998291,
		0.999624,
		0.999974,
		1,
		0.999973,
		0.999611,
		0.998204,
		0.994829,
		0.988511,
		0.978339,
		0.963537,
		0.943502,
		0.917822,
		0.886281,
		0.848853,
		0.805693,
		0.757125,
		0.70363,
		0.645835,
		0.584502,
		0.52052,
		0.454892,
		0.388735,
		0.323265,
		0.259797,
		0.199736,
		0.144575,
		0.09589,
		0.055335,
		0.024641,
		0.005611,
		0.004312,
		0.0168,
		0.03678,
		0.063557,
		0.096427,
		0.134676,
		0.17758,
		0.224414,
		0.274448,
		0.32695,
		0.381194,
		0.436458,
		0.492031,
		0.547218,
		0.601344,
		0.653763,
		0.703863,
		0.751077,
		0.794889,
		0.834852,
		0.870596,
		0.901844,
		0.928428,
		0.950308,
		0.967586,
		0.980522,
		0.98954,
		0.995229,
		0.998319,
		0.999631,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.005562,
		0.019046,
		0.039864,
		0.06733,
		0.100747,
		0.139409,
		0.182603,
		0.229612,
		0.279714,
		0.332189,
		0.38632,
		0.441396,
		0.496716,
		0.551596,
		0.605372,
		0.65741,
		0.707107,
		0.753907,
		0.797307,
		0.836868,
		0.87223,
		0.903125,
		0.929393,
		0.951001,
		0.968053,
		0.980811,
		0.9897,
		0.995304,
		0.998347,
		0.999637,
		0.999975,
		1,
		0.999974,
		0.999624,
		0.998263,
		0.994994,
		0.988867,
		0.978993,
		0.964606,
		0.945109,
		0.92009,
		0.889324,
		0.852769,
		0.810562,
		0.763,
		0.710536,
		0.653763,
		0.593409,
		0.53032,
		0.465459,
		0.399895,
		0.334797,
		0.271429,
		0.211144,
		0.155379,
		0.105651,
		0.063557,
		0.030765,
		0.009016,
		0.000116,
		0.021402,
		0.043031,
		0.071162,
		0.105104,
		0.144161,
		0.187629,
		0.234798,
		0.284958,
		0.337397,
		0.391409,
		0.446292,
		0.501356,
		0.555928,
		0.609355,
		0.661012,
		0.710309,
		0.756699,
		0.79969,
		0.838854,
		0.873839,
		0.904386,
		0.930343,
		0.951681,
		0.96851,
		0.981094,
		0.989857,
		0.995378,
		0.998373,
		0.999643,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.046277,
		0.075048,
		0.109496,
		0.14893,
		0.192655,
		0.239972,
		0.290179,
		0.342574,
		0.39646,
		0.451146,
		0.505952,
		0.560215,
		0.613292,
		0.66457,
		0.713469,
		0.759454,
		0.80204,
		0.84081,
		0.875423,
		0.905627,
		0.931276,
		0.95235,
		0.96896,
		0.981373,
		0.990011,
		0.99545,
		0.9984,
		0.999649,
		0.999976,
		1,
		0.999975,
		0.999637,
		0.998319,
		0.995152,
		0.98921,
		0.979622,
		0.965636,
		0.946661,
		0.922282,
		0.892267,
		0.856563,
		0.815284,
		0.768706,
		0.717253,
		0.661489,
		0.602103,
		0.539907,
		0.475821,
		0.410871,
		0.34618,
		0.282963,
		0.222522,
		0.166242,
		0.115585,
		0.07209,
		0.037366,
		0.013093,
		0.001017,
		0.078987,
		0.113919,
		0.153712,
		0.197681,
		0.245133,
		0.295377,
		0.34772,
		0.401474,
		0.455958,
		0.510503,
		0.564456,
		0.617185,
		0.668086,
		0.71659,
		0.762171,
		0.804357,
		0.842738,
		0.876983,
		0.906848,
		0.932195,
		0.953007,
		0.969402,
		0.981647,
		0.990162,
		0.995521,
		0.998425,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.118371,
		0.158507,
		0.202704,
		0.25028,
		0.300551,
		0.352834,
		0.40645,
		0.460729,
		0.515011,
		0.568653,
		0.621034,
		0.671559,
		0.71967,
		0.764852,
		0.806641,
		0.844638,
		0.878519,
		0.908049,
		0.933098,
		0.953653,
		0.969836,
		0.981915,
		0.99031,
		0.995591,
		0.998451,
		0.99966,
		0.999976,
		1,
		0.999976,
		0.999649,
		0.998373,
		0.995304,
		0.98954,
		0.980227,
		0.966629,
		0.948158,
		0.9244,
		0.895116,
		0.860238,
		0.819865,
		0.774249,
		0.723788,
		0.669016,
		0.61059,
		0.549284,
		0.48598,
		0.421662,
		0.357409,
		0.294389,
		0.233856,
		0.177144,
		0.125662,
		0.080895,
		0.044395,
		0.017784,
		0.002747,
		0.163312,
		0.207724,
		0.255412,
		0.3057,
		0.357915,
		0.411389,
		0.465459,
		0.519475,
		0.572806,
		0.624839,
		0.67499,
		0.722712,
		0.767497,
		0.808893,
		0.84651,
		0.880032,
		0.909232,
		0.933986,
		0.954288,
		0.970262,
		0.982179,
		0.990456,
		0.995659,
		0.998475,
		0.999666,
		0.999977,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.212739,
		0.260527,
		0.310823,
		0.362964,
		0.416289,
		0.470147,
		0.523896,
		0.576915,
		0.628601,
		0.67838,
		0.725714,
		0.770107,
		0.811114,
		0.848355,
		0.881522,
		0.910396,
		0.93486,
		0.954912,
		0.970681,
		0.982438,
		0.990599,
		0.995726,
		0.998499,
		0.999671,
		0.999977,
		1,
		0.999977,
		0.99966,
		0.998425,
		0.99545,
		0.989857,
		0.980811,
		0.967586,
		0.949604,
		0.926448,
		0.897873,
		0.863799,
		0.824309,
		0.779633,
		0.730146,
		0.676351,
		0.618874,
		0.558455,
		0.495938,
		0.432268,
		0.36848,
		0.3057,
		0.245133,
		0.188066,
		0.135857,
		0.089938,
		0.051808,
		0.023032,
		0.005236,
		0.265625,
		0.315921,
		0.36798,
		0.421152,
		0.474793,
		0.528274,
		0.58098,
		0.63232,
		0.681729,
		0.728678,
		0.772681,
		0.813304,
		0.850172,
		0.882989,
		0.911542,
		0.93572,
		0.955526,
		0.971093,
		0.982692,
		0.990739,
		0.995792,
		0.998523,
		0.999677,
		0.999978,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.320992,
		0.372963,
		0.425977,
		0.479399,
		0.532609,
		0.585003,
		0.635997,
		0.685037,
		0.731605,
		0.775222,
		0.815463,
		0.851964,
		0.884434,
		0.91267,
		0.936566,
		0.95613,
		0.971498,
		0.982941,
		0.990877,
		0.995856,
		0.998546,
		0.999682,
		0.999978,
		1,
		0.999978,
		0.999671,
		0.998475,
		0.995591,
		0.990162,
		0.981373,
		0.96851,
		0.951001,
		0.928428,
		0.900541,
		0.867251,
		0.828621,
		0.784865,
		0.736332,
		0.683499,
		0.62696,
		0.567424,
		0.505698,
		0.442688,
		0.379391,
		0.316889,
		0.256343,
		0.198992,
		0.146146,
		0.099188,
		0.059563,
		0.028784,
		0.008421,
		0.000108,
		0.377913,
		0.430764,
		0.483964,
		0.536902,
		0.588982,
		0.639632,
		0.688306,
		0.734494,
		0.777728,
		0.817592,
		0.853729,
		0.885858,
		0.913781,
		0.937398,
		0.956724,
		0.971895,
		0.983187,
		0.991012,
		0.995919,
		0.998569,
		0.999687,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.435513,
		0.488488,
		0.541152,
		0.592919,
		0.643226,
		0.691535,
		0.737347,
		0.780201,
		0.819691,
		0.855469,
		0.88726,
		0.914874,
		0.938217,
		0.957307,
		0.972286,
		0.983428,
		0.991144,
		0.995981,
		0.998592,
		0.999692,
		0.999979,
		1,
		0.999978,
		0.999682,
		0.998523,
		0.995726,
		0.990456,
		0.981915,
		0.969402,
		0.95235,
		0.930343,
		0.903125,
		0.870596,
		0.832806,
		0.789949,
		0.742351,
		0.690463,
		0.634852,
		0.576193,
		0.51526,
		0.452924,
		0.39014,
		0.327951,
		0.267474,
		0.209905,
		0.156508,
		0.108615,
		0.067623,
		0.034993,
		0.012242,
		0.000949,
		0.492972,
		0.545361,
		0.596815,
		0.646778,
		0.694725,
		0.740163,
		0.782641,
		0.821761,
		0.857183,
		0.888641,
		0.91595,
		0.939022,
		0.957881,
		0.97267,
		0.983664,
		0.991275,
		0.996042,
		0.998614,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.549528,
		0.600669,
		0.650291,
		0.697877,
		0.742944,
		0.785049,
		0.823803,
		0.858873,
		0.890001,
		0.917009,
		0.939815,
		0.958445,
		0.973047,
		0.983896,
		0.991403,
		0.996102,
		0.998635,
		0.999702,
		0.999979,
		1,
		0.999979,
		0.999692,
		0.998569,
		0.995856,
		0.990739,
		0.982438,
		0.970262,
		0.953653,
		0.932195,
		0.905627,
		0.873839,
		0.836868,
		0.794889,
		0.748208,
		0.69725,
		0.642555,
		0.584767,
		0.524629,
		0.462975,
		0.400724,
		0.33888,
		0.278519,
		0.220793,
		0.166922,
		0.118192,
		0.075953,
		0.041614,
		0.016644,
		0.002567,
		0.604481,
		0.653763,
		0.700991,
		0.745689,
		0.787425,
		0.825816,
		0.860539,
		0.891341,
		0.918052,
		0.940595,
		0.959001,
		0.973418,
		0.984125,
		0.991528,
		0.99616,
		0.998656,
		0.999707,
		0.99998,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.657196,
		0.704067,
		0.7484,
		0.78977,
		0.827802,
		0.862181,
		0.892661,
		0.919079,
		0.941362,
		0.959546,
		0.973783,
		0.984349,
		0.991651,
		0.996218,
		0.998677,
		0.999711,
		0.99998,
		1,
		0.99998,
		0.999702,
		0.998614,
		0.995981,
		0.991012,
		0.982941,
		0.971093,
		0.954912,
		0.933986,
		0.908049,
		0.876983,
		0.84081,
		0.79969,
		0.753907,
		0.703863,
		0.650073,
		0.59315,
		0.533806,
		0.472842,
		0.411143,
		0.349672,
		0.289469,
		0.231643,
		0.177371,
		0.127896,
		0.08452,
		0.048607,
		0.021574,
		0.004898,
		0.707107,
		0.751077,
		0.792083,
		0.82976,
		0.863799,
		0.893962,
		0.92009,
		0.942118,
		0.960083,
		0.974141,
		0.984569,
		0.991772,
		0.996274,
		0.998697,
		0.999716,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.753719,
		0.794366,
		0.831692,
		0.865395,
		0.895243,
		0.921086,
		0.942861,
		0.960612,
		0.974494,
		0.984786,
		0.991891,
		0.996329,
		0.998717,
		0.99972,
		0.999981,
		1,
		0.99998,
		0.999711,
		0.998656,
		0.996102,
		0.991275,
		0.983428,
		0.971895,
		0.95613,
		0.93572,
		0.910396,
		0.880032,
		0.844638,
		0.804357,
		0.759454,
		0.710309,
		0.65741,
		0.601344,
		0.542794,
		0.482527,
		0.421394,
		0.360324,
		0.300316,
		0.242442,
		0.187838,
		0.137703,
		0.093296,
		0.055933,
		0.026987,
		0.007884,
		0.000101,
		0.796619,
		0.833597,
		0.866967,
		0.896506,
		0.922066,
		0.943593,
		0.961131,
		0.97484,
		0.984999,
		0.992008,
		0.996384,
		0.998737,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.835475,
		0.868517,
		0.897749,
		0.923032,
		0.944313,
		0.961642,
		0.975181,
		0.985208,
		0.992122,
		0.996437,
		0.998756,
		0.999729,
		0.999981,
		1,
		0.999981,
		0.99972,
		0.998697,
		0.996218,
		0.991528,
		0.983896,
		0.97267,
		0.957307,
		0.937398,
		0.91267,
		0.882989,
		0.848355,
		0.808893,
		0.764852,
		0.71659,
		0.66457,
		0.609355,
		0.551596,
		0.492031,
		0.431479,
		0.370832,
		0.311056,
		0.253183,
		0.198309,
		0.147593,
		0.102252,
		0.063557,
		0.032838,
		0.011471,
		0.000888,
		0.870045,
		0.898975,
		0.923982,
		0.945021,
		0.962145,
		0.975516,
		0.985413,
		0.992235,
		0.99649,
		0.998774,
		0.999733,
		0.999982,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.900183,
		0.924919,
		0.945719,
		0.96264,
		0.975845,
		0.985615,
		0.992346,
		0.996541,
		0.998793,
		0.999737,
		0.999982,
		1,
		0.999982,
		0.999729,
		0.998737,
		0.996329,
		0.991772,
		0.984349,
		0.973418,
		0.958445,
		0.939022,
		0.914874,
		0.885858,
		0.851964,
		0.813304,
		0.770107,
		0.722712,
		0.671559,
		0.617185,
		0.560215,
		0.501356,
		0.441396,
		0.381194,
		0.321682,
		0.263854,
		0.208769,
		0.157547,
		0.111363,
		0.071448,
		0.039085,
		0.015609,
		0.002404,
		0.925841,
		0.946406,
		0.963126,
		0.976169,
		0.985814,
		0.992454,
		0.996591,
		0.998811,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		0.947082,
		0.963605,
		0.976488,
		0.986009,
		0.992561,
		0.996641,
		0.998829,
		0.999745,
		0.999982,
		1,
		0.999982,
		0.999737,
		0.998774,
		0.996437,
		0.992008,
		0.984786,
		0.974141,
		0.959546,
		0.940595,
		0.917009,
		0.888641,
		0.855469,
		0.817592,
		0.775222,
		0.728678,
		0.67838,
		0.624839,
		0.568653,
		0.510503,
		0.451146,
		0.391409,
		0.332189,
		0.274448,
		0.219207,
		0.167548,
		0.120607,
		0.079575,
		0.045692,
		0.020251,
		0.004591,
		0.964076,
		0.976801,
		0.986201,
		0.992666,
		0.99669,
		0.998846,
		0.999749,
		0.999983,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.977109,
		0.986389,
		0.992769,
		0.996737,
		0.998863,
		0.999753,
		0.999983,
		1,
		0.999983,
		0.999745,
		0.998811,
		0.996541,
		0.992235,
		0.985208,
		0.97484,
		0.960612,
		0.942118,
		0.919079,
		0.891341,
		0.858873,
		0.821761,
		0.780201,
		0.734494,
		0.685037,
		0.63232,
		0.576915,
		0.519475,
		0.460729,
		0.401474,
		0.342574,
		0.284958,
		0.229612,
		0.17758,
		0.129963,
		0.08791,
		0.052623,
		0.025353,
		0.007397,
		9.5e-005,
		0.986574,
		0.99287,
		0.996784,
		0.99888,
		0.999757,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		0.992969,
		0.99683,
		0.998897,
		0.99976,
		0.999984,
		1,
		0.999983,
		0.999753,
		0.998846,
		0.996641,
		0.992454,
		0.985615,
		0.975516,
		0.961642,
		0.943593,
		0.921086,
		0.893962,
		0.862181,
		0.825816,
		0.785049,
		0.740163,
		0.691535,
		0.639632,
		0.585003,
		0.528274,
		0.470147,
		0.411389,
		0.352834,
		0.295377,
		0.239972,
		0.187629,
		0.139409,
		0.096427,
		0.059846,
		0.030875,
		0.010771,
		0.000833,
		0.996876,
		0.998913,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.998929,
		0.999767,
		0.999984,
		1,
		0.999984,
		0.99976,
		0.99888,
		0.996737,
		0.992666,
		0.986009,
		0.976169,
		0.96264,
		0.945021,
		0.923032,
		0.896506,
		0.865395,
		0.82976,
		0.78977,
		0.745689,
		0.697877,
		0.646778,
		0.592919,
		0.536902,
		0.479399,
		0.421152,
		0.362964,
		0.3057,
		0.25028,
		0.197681,
		0.14893,
		0.105104,
		0.06733,
		0.03678,
		0.014668,
		0.002256,
		0.999771,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		0.999985,
		1,
		0.999984,
		0.999767,
		0.998913,
		0.99683,
		0.99287,
		0.986389,
		0.976801,
		0.963605,
		0.946406,
		0.924919,
		0.898975,
		0.868517,
		0.833597,
		0.794366,
		0.751077,
		0.704067,
		0.653763,
		0.600669,
		0.545361,
		0.488488,
		0.430764,
		0.372963,
		0.315921,
		0.260527,
		0.207724,
		0.158507,
		0.113919,
		0.075048,
		0.043031,
		0.019046,
		0.004312,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.003183,
		0.016673,
		0.039569,
		0.070648,
		0.108724,
		0.152659,
		0.201355,
		0.253765,
		0.308888,
		0.365774,
		0.423526,
		0.481304,
		0.538323,
		0.593863,
		0.647268,
		0.69795,
		0.745396,
		0.789171,
		0.828924,
		0.864394,
		0.895415,
		0.921922,
		0.943959,
		0.961681,
		0.975358,
		0.985372,
		0.99221,
		0.996452,
		0.998733,
		0.999707,
		0.999975,
		1,
		0.999991,
		0.999819,
		0.999074,
		0.99718,
		0.99349,
		0.987358,
		0.978187,
		0.965467,
		0.948785,
		0.927846,
		0.902465,
		0.872571,
		0.838201,
		0.799498,
		0.756699,
		0.710133,
		0.660215,
		0.607439,
		0.552374,
		0.495659,
		0.438,
		0.380165,
		0.32298,
		0.267329,
		0.214149,
		0.164428,
		0.119202,
		0.079556,
		0.046619,
		0.021564,
		0.005605,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.000366,
		0.008808,
		0.027432,
		0.054963,
		0.090168,
		0.131859,
		0.178891,
		0.230171,
		0.28465,
		0.341334,
		0.39928,
		0.457602,
		0.515474,
		0.572131,
		0.626874,
		0.679074,
		0.728175,
		0.773702,
		0.815262,
		0.852555,
		0.885375,
		0.913621,
		0.937299,
		0.95653,
		0.971552,
		0.982721,
		0.990506,
		0.995473,
		0.99826,
		0.999537,
		0.999944,
		1,
		0.999997,
		0.999892,
		0.999323,
		0.997731,
		0.994461,
		0.988848,
		0.980266,
		0.968173,
		0.952128,
		0.931803,
		0.90699,
		0.877597,
		0.843645,
		0.805262,
		0.762678,
		0.716218,
		0.666294,
		0.613403,
		0.55812,
		0.501091,
		0.443032,
		0.384724,
		0.327009,
		0.270786,
		0.217011,
		0.166693,
		0.120891,
		0.080713,
		0.047314,
		0.021892,
		0.005692,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.003284,
		0.01719,
		0.040769,
		0.072738,
		0.111859,
		0.156939,
		0.206833,
		0.260446,
		0.316735,
		0.374712,
		0.433445,
		0.492061,
		0.54975,
		0.60577,
		0.659446,
		0.710181,
		0.757456,
		0.800837,
		0.839981,
		0.874642,
		0.90468,
		0.930061,
		0.950871,
		0.967315,
		0.97972,
		0.988531,
		0.9943,
		0.997663,
		0.999303,
		0.999889,
		0.999997,
		1,
		0.999942,
		0.999523,
		0.998208,
		0.99534,
		0.990235,
		0.98224,
		0.97078,
		0.955381,
		0.935686,
		0.91146,
		0.882588,
		0.849074,
		0.811031,
		0.768681,
		0.722343,
		0.672427,
		0.619433,
		0.563939,
		0.5066,
		0.448143,
		0.38936,
		0.331109,
		0.274307,
		0.219929,
		0.169005,
		0.122616,
		0.081895,
		0.048024,
		0.022228,
		0.005782,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.000378,
		0.009086,
		0.02828,
		0.056622,
		0.092821,
		0.135635,
		0.183866,
		0.236371,
		0.292057,
		0.349886,
		0.40888,
		0.468117,
		0.526742,
		0.583966,
		0.639069,
		0.691409,
		0.740422,
		0.78563,
		0.826649,
		0.863189,
		0.895068,
		0.922214,
		0.944673,
		0.962616,
		0.976338,
		0.986257,
		0.992908,
		0.996922,
		0.998987,
		0.999801,
		0.99999,
		1,
		0.999973,
		0.999679,
		0.998614,
		0.996128,
		0.991519,
		0.984106,
		0.973282,
		0.95854,
		0.939491,
		0.91587,
		0.887541,
		0.854485,
		0.816804,
		0.774707,
		0.728508,
		0.678615,
		0.625529,
		0.569832,
		0.512188,
		0.453333,
		0.394074,
		0.335282,
		0.277895,
		0.222905,
		0.171364,
		0.124377,
		0.083104,
		0.04875,
		0.022572,
		0.005873,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.00339,
		0.017731,
		0.042023,
		0.074922,
		0.11513,
		0.161399,
		0.212531,
		0.267385,
		0.32487,
		0.383959,
		0.443682,
		0.503133,
		0.561477,
		0.617947,
		0.671853,
		0.722587,
		0.769624,
		0.812535,
		0.850988,
		0.884756,
		0.913726,
		0.937904,
		0.957423,
		0.972543,
		0.983655,
		0.991272,
		0.996013,
		0.998572,
		0.999668,
		0.999972,
		1,
		0.999989,
		0.999795,
		0.998956,
		0.996829,
		0.9927,
		0.985864,
		0.975678,
		0.961602,
		0.943213,
		0.920217,
		0.89245,
		0.859876,
		0.822577,
		0.780754,
		0.734711,
		0.684856,
		0.631689,
		0.575799,
		0.517854,
		0.458604,
		0.398867,
		0.339531,
		0.28155,
		0.225939,
		0.173771,
		0.126177,
		0.084339,
		0.049493,
		0.022924,
		0.005967,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.00039,
		0.009378,
		0.029167,
		0.058356,
		0.095593,
		0.139573,
		0.189047,
		0.242817,
		0.299743,
		0.358744,
		0.4188,
		0.478955,
		0.538323,
		0.596089,
		0.651515,
		0.703943,
		0.752804,
		0.79762,
		0.838015,
		0.873716,
		0.904566,
		0.930531,
		0.951701,
		0.968302,
		0.980692,
		0.989361,
		0.994912,
		0.998039,
		0.999477,
		0.999936,
		1,
		0.999997,
		0.999878,
		0.999236,
		0.997444,
		0.99378,
		0.987511,
		0.977965,
		0.964563,
		0.946848,
		0.924495,
		0.897313,
		0.865241,
		0.828347,
		0.786818,
		0.740951,
		0.69115,
		0.637915,
		0.58184,
		0.523601,
		0.463957,
		0.403741,
		0.343856,
		0.285276,
		0.229035,
		0.17623,
		0.128015,
		0.085601,
		0.050253,
		0.023285,
		0.006063,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.003501,
		0.018298,
		0.043337,
		0.077206,
		0.118545,
		0.166048,
		0.218463,
		0.274594,
		0.333306,
		0.393526,
		0.454246,
		0.514529,
		0.573509,
		0.630396,
		0.684484,
		0.735156,
		0.781884,
		0.824244,
		0.861918,
		0.894703,
		0.922519,
		0.945417,
		0.963582,
		0.977338,
		0.987147,
		0.993593,
		0.997366,
		0.999212,
		0.999874,
		0.999997,
		1,
		0.999934,
		0.999461,
		0.997978,
		0.994759,
		0.989048,
		0.98014,
		0.967419,
		0.950392,
		0.9287,
		0.902123,
		0.870578,
		0.834112,
		0.792898,
		0.747226,
		0.697495,
		0.644206,
		0.587956,
		0.529428,
		0.469393,
		0.408697,
		0.34826,
		0.289073,
		0.232193,
		0.178739,
		0.129893,
		0.086892,
		0.05103,
		0.023653,
		0.006161,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.000403,
		0.009684,
		0.030097,
		0.060171,
		0.098489,
		0.143683,
		0.194445,
		0.249522,
		0.307723,
		0.36792,
		0.429052,
		0.490126,
		0.550223,
		0.608503,
		0.664208,
		0.716668,
		0.765307,
		0.809651,
		0.849334,
		0.884104,
		0.913836,
		0.938536,
		0.958348,
		0.973558,
		0.984597,
		0.992029,
		0.996529,
		0.998854,
		0.999775,
		0.999988,
		1,
		0.999969,
		0.999636,
		0.998435,
		0.995638,
		0.990474,
		0.9822,
		0.970166,
		0.95384,
		0.932827,
		0.906878,
		0.875881,
		0.839866,
		0.79899,
		0.753533,
		0.70389,
		0.650561,
		0.594146,
		0.535338,
		0.474914,
		0.413737,
		0.352743,
		0.292943,
		0.235414,
		0.181302,
		0.131812,
		0.088212,
		0.051826,
		0.024031,
		0.006261,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.003617,
		0.018893,
		0.044712,
		0.079594,
		0.122113,
		0.170898,
		0.224639,
		0.282087,
		0.342055,
		0.403426,
		0.465151,
		0.526257,
		0.585848,
		0.643115,
		0.697333,
		0.747875,
		0.794215,
		0.835936,
		0.872739,
		0.904447,
		0.931022,
		0.952563,
		0.969317,
		0.98168,
		0.990188,
		0.995504,
		0.998385,
		0.999624,
		0.999968,
		1,
		0.999988,
		0.999767,
		0.998818,
		0.996422,
		0.991788,
		0.984144,
		0.972801,
		0.957188,
		0.936872,
		0.911571,
		0.881147,
		0.845607,
		0.805092,
		0.759872,
		0.710335,
		0.656981,
		0.600413,
		0.54133,
		0.480522,
		0.418863,
		0.357308,
		0.296887,
		0.238701,
		0.183919,
		0.133774,
		0.089563,
		0.05264,
		0.024418,
		0.006364,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.000417,
		0.010005,
		0.031071,
		0.062072,
		0.101517,
		0.147975,
		0.200072,
		0.256499,
		0.31601,
		0.377429,
		0.439648,
		0.501639,
		0.562448,
		0.621209,
		0.677145,
		0.729573,
		0.777914,
		0.821699,
		0.860575,
		0.894319,
		0.92284,
		0.946192,
		0.964579,
		0.978359,
		0.988039,
		0.994262,
		0.997782,
		0.999407,
		0.999927,
		1,
		0.999997,
		0.999861,
		0.999134,
		0.997111,
		0.992991,
		0.98597,
		0.975321,
		0.960431,
		0.94083,
		0.916199,
		0.886372,
		0.851331,
		0.811201,
		0.76624,
		0.716828,
		0.663465,
		0.606755,
		0.547405,
		0.486216,
		0.424076,
		0.361957,
		0.300909,
		0.242056,
		0.186592,
		0.135779,
		0.090944,
		0.053473,
		0.024814,
		0.00647,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.003739,
		0.019517,
		0.046154,
		0.082095,
		0.125842,
		0.17596,
		0.231074,
		0.289878,
		0.351133,
		0.413672,
		0.476406,
		0.538323,
		0.5985,
		0.656102,
		0.71039,
		0.76073,
		0.806596,
		0.847584,
		0.883417,
		0.913952,
		0.939196,
		0.959306,
		0.974599,
		0.985548,
		0.992774,
		0.99702,
		0.999106,
		0.999857,
		0.999997,
		1,
		0.999925,
		0.999388,
		0.997711,
		0.994084,
		0.987677,
		0.977722,
		0.963565,
		0.944695,
		0.920755,
		0.891549,
		0.857034,
		0.817314,
		0.772634,
		0.723368,
		0.670011,
		0.613173,
		0.553565,
		0.491999,
		0.429378,
		0.366691,
		0.305009,
		0.245479,
		0.189322,
		0.137829,
		0.092357,
		0.054327,
		0.02522,
		0.006578,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.000431,
		0.010342,
		0.032094,
		0.064064,
		0.104687,
		0.152459,
		0.205942,
		0.263763,
		0.324619,
		0.387283,
		0.450601,
		0.513503,
		0.575003,
		0.634207,
		0.690319,
		0.742645,
		0.790604,
		0.833736,
		0.871707,
		0.904322,
		0.931537,
		0.953459,
		0.970362,
		0.982681,
		0.991008,
		0.996073,
		0.9987,
		0.999743,
		0.999987,
		1,
		0.999965,
		0.999586,
		0.998225,
		0.995069,
		0.989263,
		0.980001,
		0.966586,
		0.948463,
		0.925236,
		0.896676,
		0.862711,
		0.823426,
		0.779051,
		0.729952,
		0.67662,
		0.619667,
		0.55981,
		0.497872,
		0.43477,
		0.371512,
		0.309189,
		0.248973,
		0.192112,
		0.139926,
		0.093803,
		0.0552,
		0.025636,
		0.006689,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.003868,
		0.020173,
		0.047667,
		0.084714,
		0.129743,
		0.181245,
		0.237781,
		0.297982,
		0.360554,
		0.424278,
		0.488022,
		0.550737,
		0.611467,
		0.669354,
		0.723646,
		0.773702,
		0.819002,
		0.859156,
		0.893914,
		0.923176,
		0.946999,
		0.965609,
		0.9794,
		0.988931,
		0.994912,
		0.998167,
		0.999572,
		0.999963,
		1,
		0.999986,
		0.999735,
		0.998657,
		0.995947,
		0.990728,
		0.982156,
		0.969489,
		0.952128,
		0.929636,
		0.901745,
		0.868358,
		0.829535,
		0.78549,
		0.736579,
		0.683291,
		0.626236,
		0.56614,
		0.503836,
		0.440255,
		0.376422,
		0.313452,
		0.25254,
		0.194962,
		0.142069,
		0.095283,
		0.056095,
		0.026062,
		0.006803,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.000447,
		0.010696,
		0.033168,
		0.066153,
		0.108006,
		0.157147,
		0.212067,
		0.271328,
		0.333566,
		0.397498,
		0.461923,
		0.525728,
		0.587893,
		0.647496,
		0.703721,
		0.755868,
		0.803355,
		0.845732,
		0.88269,
		0.914074,
		0.939885,
		0.960299,
		0.975665,
		0.986506,
		0.993506,
		0.997481,
		0.999324,
		0.999917,
		1,
		0.999996,
		0.999842,
		0.999014,
		0.996723,
		0.992073,
		0.984185,
		0.972271,
		0.955686,
		0.933948,
		0.906753,
		0.87397,
		0.835636,
		0.791947,
		0.743248,
		0.690022,
		0.632882,
		0.572558,
		0.509893,
		0.445833,
		0.381424,
		0.317799,
		0.256181,
		0.197875,
		0.144262,
		0.096799,
		0.057012,
		0.026499,
		0.00692,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.004003,
		0.020862,
		0.049255,
		0.087461,
		0.133826,
		0.186768,
		0.244776,
		0.306416,
		0.370333,
		0.435259,
		0.500011,
		0.563503,
		0.624748,
		0.682865,
		0.737086,
		0.78677,
		0.831403,
		0.870615,
		0.904191,
		0.932076,
		0.954391,
		0.971437,
		0.983695,
		0.991821,
		0.996615,
		0.998981,
		0.999836,
		0.999996,
		0.999999,
		0.999914,
		0.999302,
		0.997398,
		0.993298,
		0.986085,
		0.974928,
		0.959131,
		0.938168,
		0.911694,
		0.879542,
		0.841725,
		0.798418,
		0.749955,
		0.696813,
		0.639602,
		0.579062,
		0.516043,
		0.451507,
		0.386518,
		0.322233,
		0.2599,
		0.200852,
		0.146505,
		0.09835,
		0.057952,
		0.026947,
		0.00704,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.000462,
		0.011069,
		0.034297,
		0.068345,
		0.111483,
		0.162051,
		0.218463,
		0.27921,
		0.342867,
		0.40809,
		0.473628,
		0.538323,
		0.601121,
		0.661072,
		0.717343,
		0.769224,
		0.816138,
		0.857652,
		0.893487,
		0.923529,
		0.947842,
		0.966673,
		0.980459,
		0.989821,
		0.995539,
		0.998518,
		0.999707,
		0.999985,
		1,
		0.999959,
		0.999527,
		0.997978,
		0.994403,
		0.987855,
		0.977455,
		0.962458,
		0.94229,
		0.916561,
		0.88507,
		0.847798,
		0.804901,
		0.756699,
		0.703661,
		0.646399,
		0.585653,
		0.522288,
		0.457278,
		0.391707,
		0.326755,
		0.263696,
		0.203896,
		0.148801,
		0.099939,
		0.058915,
		0.027406,
		0.007163,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.004145,
		0.021587,
		0.050923,
		0.090341,
		0.138102,
		0.192542,
		0.252073,
		0.315194,
		0.380488,
		0.446628,
		0.512384,
		0.57663,
		0.638345,
		0.696627,
		0.750696,
		0.79991,
		0.843766,
		0.881923,
		0.914202,
		0.940607,
		0.961329,
		0.976757,
		0.987468,
		0.99422,
		0.99791,
		0.99951,
		0.999958,
		1,
		0.999984,
		0.999696,
		0.998468,
		0.995392,
		0.989494,
		0.97985,
		0.965663,
		0.946308,
		0.92135,
		0.890546,
		0.85385,
		0.811392,
		0.763477,
		0.710566,
		0.65327,
		0.592333,
		0.528629,
		0.463149,
		0.396993,
		0.331368,
		0.267574,
		0.207007,
		0.151149,
		0.101566,
		0.059902,
		0.027878,
		0.007289,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.000479,
		0.011462,
		0.035484,
		0.070648,
		0.11513,
		0.167184,
		0.225145,
		0.287427,
		0.352538,
		0.419073,
		0.485726,
		0.551297,
		0.61469,
		0.674931,
		0.73117,
		0.78269,
		0.828924,
		0.869459,
		0.904051,
		0.932642,
		0.955361,
		0.972543,
		0.984721,
		0.992621,
		0.997128,
		0.999227,
		0.999905,
		0.999999,
		0.999996,
		0.999819,
		0.998873,
		0.996266,
		0.991001,
		0.982109,
		0.968741,
		0.950217,
		0.926053,
		0.895967,
		0.859876,
		0.817887,
		0.770286,
		0.717527,
		0.660215,
		0.599101,
		0.535067,
		0.469119,
		0.402378,
		0.336073,
		0.271535,
		0.210189,
		0.153554,
		0.103233,
		0.060913,
		0.028361,
		0.007418,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.004296,
		0.02235,
		0.052677,
		0.093366,
		0.142584,
		0.198581,
		0.259691,
		0.324337,
		0.391034,
		0.4584,
		0.525152,
		0.59012,
		0.652255,
		0.710629,
		0.764456,
		0.813093,
		0.856057,
		0.893035,
		0.9239,
		0.948721,
		0.967772,
		0.981538,
		0.990705,
		0.99614,
		0.998834,
		0.999812,
		0.999996,
		0.999999,
		0.999901,
		0.9992,
		0.99703,
		0.992377,
		0.98423,
		0.971686,
		0.954009,
		0.930665,
		0.901325,
		0.865871,
		0.824381,
		0.777124,
		0.72454,
		0.667234,
		0.605958,
		0.541604,
		0.475193,
		0.407865,
		0.340874,
		0.27558,
		0.213443,
		0.156015,
		0.104941,
		0.061951,
		0.028857,
		0.007551,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.000497,
		0.011876,
		0.036734,
		0.073068,
		0.118957,
		0.172561,
		0.232129,
		0.295997,
		0.362598,
		0.430464,
		0.498233,
		0.564656,
		0.628601,
		0.689066,
		0.745185,
		0.796242,
		0.841678,
		0.881109,
		0.914337,
		0.941362,
		0.962398,
		0.977873,
		0.988431,
		0.994912,
		0.998304,
		0.999663,
		0.999982,
		1,
		0.999953,
		0.999457,
		0.997688,
		0.993622,
		0.98621,
		0.974494,
		0.957681,
		0.935179,
		0.906615,
		0.871829,
		0.83087,
		0.783986,
		0.731605,
		0.674326,
		0.612904,
		0.54824,
		0.48137,
		0.413455,
		0.345773,
		0.279714,
		0.21677,
		0.158534,
		0.106691,
		0.063015,
		0.029367,
		0.007688,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.004454,
		0.023155,
		0.054523,
		0.096544,
		0.147285,
		0.204903,
		0.267647,
		0.333861,
		0.40199,
		0.470589,
		0.538323,
		0.603979,
		0.666474,
		0.72486,
		0.778343,
		0.826287,
		0.868232,
		0.903905,
		0.933236,
		0.956371,
		0.973679,
		0.985756,
		0.993406,
		0.997607,
		0.999437,
		0.999951,
		1,
		0.999982,
		0.999651,
		0.998244,
		0.994738,
		0.988048,
		0.977161,
		0.961225,
		0.93959,
		0.91183,
		0.877744,
		0.83735,
		0.790869,
		0.738718,
		0.681489,
		0.619939,
		0.554977,
		0.487653,
		0.41915,
		0.350772,
		0.283937,
		0.220175,
		0.161115,
		0.108485,
		0.064107,
		0.02989,
		0.007828,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.000516,
		0.012313,
		0.038051,
		0.075614,
		0.122976,
		0.178197,
		0.239433,
		0.304937,
		0.373065,
		0.442279,
		0.51116,
		0.578407,
		0.642853,
		0.703467,
		0.759371,
		0.809849,
		0.854361,
		0.892557,
		0.924291,
		0.94964,
		0.968908,
		0.982633,
		0.991581,
		0.996711,
		0.999112,
		0.99989,
		0.999999,
		0.999995,
		0.999791,
		0.998706,
		0.995728,
		0.989742,
		0.979683,
		0.964635,
		0.943889,
		0.916963,
		0.883611,
		0.843815,
		0.797771,
		0.745878,
		0.688723,
		0.627063,
		0.561815,
		0.494044,
		0.424954,
		0.355873,
		0.288253,
		0.223658,
		0.163757,
		0.110324,
		0.065227,
		0.030427,
		0.007973,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.004622,
		0.024004,
		0.056468,
		0.099885,
		0.152218,
		0.211524,
		0.27596,
		0.343786,
		0.413374,
		0.483212,
		0.551909,
		0.618209,
		0.680995,
		0.739302,
		0.79233,
		0.839455,
		0.880246,
		0.914479,
		0.942154,
		0.963506,
		0.979014,
		0.989394,
		0.995579,
		0.998659,
		0.999783,
		0.999995,
		0.999999,
		0.999886,
		0.999079,
		0.996595,
		0.991293,
		0.982057,
		0.967907,
		0.948071,
		0.922008,
		0.889423,
		0.850259,
		0.804685,
		0.753082,
		0.696026,
		0.634275,
		0.568755,
		0.500544,
		0.430867,
		0.361079,
		0.292664,
		0.227222,
		0.166465,
		0.11221,
		0.066377,
		0.030978,
		0.008121,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.000536,
		0.012775,
		0.039439,
		0.078295,
		0.127199,
		0.184108,
		0.247077,
		0.31427,
		0.383959,
		0.454537,
		0.524519,
		0.592557,
		0.657442,
		0.71812,
		0.773702,
		0.823476,
		0.866927,
		0.903749,
		0.93386,
		0.957423,
		0.974848,
		0.986798,
		0.994171,
		0.998049,
		0.999611,
		0.99998,
		1,
		0.999946,
		0.999373,
		0.997343,
		0.9927,
		0.984278,
		0.971035,
		0.952128,
		0.926957,
		0.895174,
		0.856678,
		0.811609,
		0.760326,
		0.703397,
		0.641577,
		0.575799,
		0.507156,
		0.436892,
		0.366393,
		0.297172,
		0.230869,
		0.169238,
		0.114144,
		0.067557,
		0.031545,
		0.008273,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.004799,
		0.0249,
		0.058518,
		0.103402,
		0.1574,
		0.218463,
		0.28465,
		0.354133,
		0.425205,
		0.496282,
		0.565918,
		0.63281,
		0.69581,
		0.753936,
		0.806386,
		0.852555,
		0.892049,
		0.924704,
		0.950601,
		0.970081,
		0.983744,
		0.992444,
		0.997247,
		0.99935,
		0.999944,
		1,
		0.999979,
		0.999596,
		0.997978,
		0.993965,
		0.986346,
		0.974012,
		0.956054,
		0.931803,
		0.900855,
		0.863065,
		0.818536,
		0.767608,
		0.710833,
		0.648967,
		0.582946,
		0.51388,
		0.443032,
		0.371816,
		0.301781,
		0.234603,
		0.172081,
		0.116129,
		0.068768,
		0.032127,
		0.00843,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.000557,
		0.013263,
		0.040905,
		0.081119,
		0.131641,
		0.190312,
		0.255081,
		0.324015,
		0.395301,
		0.467253,
		0.538323,
		0.607109,
		0.672365,
		0.73301,
		0.78815,
		0.837084,
		0.879329,
		0.91463,
		0.942986,
		0.964657,
		0.980179,
		0.990352,
		0.996216,
		0.998974,
		0.999873,
		0.999999,
		0.999994,
		0.999758,
		0.998506,
		0.99509,
		0.988257,
		0.976835,
		0.959841,
		0.936538,
		0.90646,
		0.869413,
		0.825463,
		0.774923,
		0.718333,
		0.656444,
		0.590198,
		0.520718,
		0.449289,
		0.377353,
		0.306492,
		0.238425,
		0.174994,
		0.118164,
		0.070013,
		0.032726,
		0.008592,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.004986,
		0.025848,
		0.060681,
		0.107106,
		0.162846,
		0.22574,
		0.29374,
		0.364925,
		0.437501,
		0.509814,
		0.580358,
		0.647781,
		0.710906,
		0.768736,
		0.820474,
		0.865538,
		0.903584,
		0.934518,
		0.958519,
		0.976047,
		0.987845,
		0.994912,
		0.998451,
		0.999749,
		0.999994,
		0.999999,
		0.999868,
		0.998935,
		0.996078,
		0.990011,
		0.979499,
		0.963484,
		0.941155,
		0.911982,
		0.875716,
		0.832383,
		0.782268,
		0.725895,
		0.664008,
		0.597555,
		0.527672,
		0.455665,
		0.383005,
		0.31131,
		0.242338,
		0.177981,
		0.120254,
		0.071291,
		0.033341,
		0.008758,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.000579,
		0.013779,
		0.042454,
		0.084099,
		0.136317,
		0.196828,
		0.263466,
		0.334196,
		0.407111,
		0.480445,
		0.552583,
		0.622067,
		0.687611,
		0.748116,
		0.802682,
		0.850628,
		0.89151,
		0.92514,
		0.951607,
		0.971292,
		0.984868,
		0.993291,
		0.997745,
		0.999548,
		0.999976,
		1,
		0.999937,
		0.999274,
		0.996932,
		0.991607,
		0.981999,
		0.966975,
		0.945646,
		0.917412,
		0.881966,
		0.839291,
		0.789639,
		0.733515,
		0.671658,
		0.605018,
		0.534743,
		0.462163,
		0.388776,
		0.316237,
		0.246346,
		0.181043,
		0.122398,
		0.072605,
		0.033974,
		0.008928,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.005185,
		0.02685,
		0.062966,
		0.11101,
		0.168575,
		0.233376,
		0.303253,
		0.376183,
		0.450283,
		0.523824,
		0.595234,
		0.663118,
		0.726267,
		0.783673,
		0.834551,
		0.878352,
		0.91479,
		0.943859,
		0.965851,
		0.981366,
		0.991302,
		0.996818,
		0.999246,
		0.999934,
		1,
		0.999975,
		0.999531,
		0.99766,
		0.993045,
		0.984332,
		0.970307,
		0.950003,
		0.922742,
		0.888157,
		0.84618,
		0.797031,
		0.741192,
		0.679392,
		0.612586,
		0.541933,
		0.468785,
		0.394668,
		0.321275,
		0.25045,
		0.184184,
		0.1246,
		0.073955,
		0.034625,
		0.009104,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.000602,
		0.014327,
		0.044093,
		0.087245,
		0.141244,
		0.203676,
		0.272257,
		0.344836,
		0.419412,
		0.494131,
		0.567308,
		0.637429,
		0.70317,
		0.763413,
		0.81726,
		0.864056,
		0.903409,
		0.935212,
		0.959663,
		0.977279,
		0.988892,
		0.995624,
		0.998808,
		0.999852,
		0.999999,
		0.999993,
		0.999718,
		0.998267,
		0.994328,
		0.986495,
		0.973475,
		0.954217,
		0.927963,
		0.894279,
		0.853044,
		0.804438,
		0.748921,
		0.68721,
		0.620261,
		0.549244,
		0.475533,
		0.400685,
		0.326429,
		0.254655,
		0.187405,
		0.126861,
		0.075343,
		0.035295,
		0.009285,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.005396,
		0.027912,
		0.065382,
		0.11513,
		0.174607,
		0.241394,
		0.313213,
		0.387931,
		0.463572,
		0.538323,
		0.610552,
		0.678813,
		0.741872,
		0.798713,
		0.848568,
		0.890936,
		0.925601,
		0.952661,
		0.972543,
		0.986004,
		0.994115,
		0.9982,
		0.999707,
		0.999993,
		0.999999,
		0.999846,
		0.998762,
		0.995458,
		0.988485,
		0.976473,
		0.95828,
		0.933068,
		0.900326,
		0.859876,
		0.811856,
		0.756699,
		0.695109,
		0.628041,
		0.556677,
		0.48241,
		0.406828,
		0.331701,
		0.258963,
		0.19071,
		0.129184,
		0.07677,
		0.035984,
		0.009472,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.000627,
		0.014907,
		0.045828,
		0.090569,
		0.146439,
		0.21088,
		0.281477,
		0.355962,
		0.432226,
		0.508329,
		0.582507,
		0.653193,
		0.719025,
		0.778869,
		0.831837,
		0.877309,
		0.91496,
		0.944778,
		0.967092,
		0.982575,
		0.992239,
		0.99738,
		0.999473,
		0.999972,
		1,
		0.999926,
		0.999153,
		0.996439,
		0.990301,
		0.979295,
		0.962186,
		0.938045,
		0.906287,
		0.866667,
		0.819279,
		0.764523,
		0.703088,
		0.635927,
		0.564232,
		0.489418,
		0.413102,
		0.337094,
		0.263377,
		0.194101,
		0.13157,
		0.078237,
		0.036694,
		0.009664,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.00562,
		0.029038,
		0.067939,
		0.119481,
		0.180962,
		0.249819,
		0.323646,
		0.400195,
		0.477387,
		0.553326,
		0.626313,
		0.694856,
		0.757694,
		0.813813,
		0.862471,
		0.903223,
		0.935944,
		0.960857,
		0.978541,
		0.989936,
		0.996301,
		0.999119,
		0.999923,
		1,
		0.999971,
		0.999452,
		0.997277,
		0.991944,
		0.981935,
		0.965925,
		0.942888,
		0.912153,
		0.873411,
		0.826699,
		0.772387,
		0.711144,
		0.643919,
		0.571912,
		0.496559,
		0.41951,
		0.342613,
		0.267901,
		0.197581,
		0.134022,
		0.079747,
		0.037425,
		0.009862,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.000654,
		0.015523,
		0.047667,
		0.094085,
		0.151921,
		0.218463,
		0.291153,
		0.367599,
		0.445578,
		0.523054,
		0.598188,
		0.669354,
		0.735156,
		0.794448,
		0.846361,
		0.890323,
		0.926089,
		0.953766,
		0.973833,
		0.987147,
		0.994912,
		0.998608,
		0.999826,
		0.999999,
		0.999992,
		0.999669,
		0.997978,
		0.993413,
		0.984391,
		0.969489,
		0.947585,
		0.917916,
		0.880097,
		0.834112,
		0.780288,
		0.719275,
		0.652015,
		0.579717,
		0.503836,
		0.426054,
		0.34826,
		0.272538,
		0.201154,
		0.136543,
		0.081301,
		0.038178,
		0.010067,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.005858,
		0.030233,
		0.070648,
		0.124081,
		0.187663,
		0.258679,
		0.334582,
		0.413,
		0.491749,
		0.568845,
		0.642517,
		0.71123,
		0.773702,
		0.828924,
		0.876195,
		0.91514,
		0.945745,
		0.968379,
		0.983803,
		0.993157,
		0.997897,
		0.999656,
		0.999992,
		0.999999,
		0.999819,
		0.998551,
		0.994712,
		0.986658,
		0.972873,
		0.952128,
		0.923565,
		0.886718,
		0.841508,
		0.78822,
		0.727478,
		0.660215,
		0.587648,
		0.511251,
		0.432738,
		0.354039,
		0.277292,
		0.204822,
		0.139134,
		0.0829,
		0.038954,
		0.010277,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.000682,
		0.016179,
		0.049618,
		0.097809,
		0.157713,
		0.226451,
		0.301315,
		0.379775,
		0.45949,
		0.538323,
		0.614355,
		0.685901,
		0.751535,
		0.810106,
		0.860772,
		0.903024,
		0.936718,
		0.962103,
		0.979834,
		0.990973,
		0.996937,
		0.999381,
		0.999967,
		1,
		0.999913,
		0.999006,
		0.995843,
		0.988734,
		0.976067,
		0.956507,
		0.92909,
		0.893264,
		0.84888,
		0.796177,
		0.735749,
		0.668518,
		0.595706,
		0.518807,
		0.439564,
		0.359954,
		0.282166,
		0.208589,
		0.141799,
		0.084547,
		0.039753,
		0.010495,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.006111,
		0.031504,
		0.073521,
		0.128949,
		0.194735,
		0.268,
		0.346048,
		0.426374,
		0.50668,
		0.584888,
		0.65916,
		0.727916,
		0.789856,
		0.843991,
		0.889668,
		0.926607,
		0.954925,
		0.975164,
		0.988294,
		0.995675,
		0.998965,
		0.999909,
		1,
		0.999966,
		0.999355,
		0.996812,
		0.990617,
		0.979066,
		0.960712,
		0.934481,
		0.899725,
		0.85622,
		0.804153,
		0.744084,
		0.676922,
		0.603892,
		0.526505,
		0.446537,
		0.366009,
		0.287165,
		0.212459,
		0.14454,
		0.086243,
		0.040578,
		0.010719,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.000713,
		0.016877,
		0.051692,
		0.101756,
		0.163836,
		0.234873,
		0.311993,
		0.392521,
		0.473988,
		0.554152,
		0.63101,
		0.702821,
		0.768129,
		0.82579,
		0.875,
		0.915332,
		0.946765,
		0.969716,
		0.985047,
		0.99405,
		0.998364,
		0.999794,
		0.999999,
		0.99999,
		0.99961,
		0.997627,
		0.992308,
		0.981865,
		0.964734,
		0.939726,
		0.906089,
		0.863519,
		0.812141,
		0.75248,
		0.685426,
		0.612205,
		0.534347,
		0.45366,
		0.372208,
		0.292293,
		0.216434,
		0.14736,
		0.08799,
		0.041428,
		0.010951,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.006382,
		0.032856,
		0.076573,
		0.134105,
		0.202206,
		0.277816,
		0.358077,
		0.440344,
		0.522198,
		0.601465,
		0.676234,
		0.744885,
		0.80611,
		0.858946,
		0.902812,
		0.937538,
		0.963404,
		0.981155,
		0.991996,
		0.997527,
		0.999593,
		0.99999,
		0.999999,
		0.999785,
		0.998294,
		0.993807,
		0.984457,
		0.968563,
		0.944814,
		0.912347,
		0.870768,
		0.820135,
		0.760931,
		0.694027,
		0.620647,
		0.542336,
		0.460936,
		0.378555,
		0.297552,
		0.220519,
		0.150262,
		0.089791,
		0.042305,
		0.01119,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.000745,
		0.017621,
		0.053897,
		0.105945,
		0.170317,
		0.243759,
		0.32322,
		0.405867,
		0.489096,
		0.570554,
		0.648151,
		0.720092,
		0.784898,
		0.841438,
		0.888967,
		0.927158,
		0.956143,
		0.976536,
		0.98944,
		0.996398,
		0.999268,
		0.999961,
		1,
		0.999897,
		0.998827,
		0.995118,
		0.986838,
		0.972191,
		0.949734,
		0.918486,
		0.877955,
		0.828126,
		0.769433,
		0.702724,
		0.629217,
		0.550474,
		0.468368,
		0.385053,
		0.302949,
		0.224718,
		0.15325,
		0.091646,
		0.043211,
		0.011437,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.006671,
		0.034297,
		0.079817,
		0.139573,
		0.210104,
		0.288158,
		0.370703,
		0.45494,
		0.538323,
		0.618579,
		0.693726,
		0.762104,
		0.822408,
		0.873716,
		0.915536,
		0.947842,
		0.971103,
		0.986304,
		0.994912,
		0.998776,
		0.999892,
		1,
		0.999959,
		0.999236,
		0.996245,
		0.989007,
		0.975609,
		0.954475,
		0.924495,
		0.88507,
		0.836105,
		0.77798,
		0.711512,
		0.637915,
		0.558763,
		0.475959,
		0.391707,
		0.308487,
		0.229035,
		0.156326,
		0.09356,
		0.044146,
		0.011692,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.00078,
		0.018415,
		0.056247,
		0.110396,
		0.177183,
		0.253142,
		0.33503,
		0.419845,
		0.504839,
		0.58754,
		0.665773,
		0.737688,
		0.801791,
		0.85698,
		0.902584,
		0.938407,
		0.964764,
		0.982503,
		0.992999,
		0.998064,
		0.999755,
		0.999998,
		0.999989,
		0.999536,
		0.997195,
		0.990961,
		0.97881,
		0.959025,
		0.930361,
		0.892101,
		0.844063,
		0.786565,
		0.72039,
		0.64674,
		0.567204,
		0.483714,
		0.398522,
		0.31417,
		0.233473,
		0.159495,
		0.095534,
		0.045111,
		0.011957,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.006979,
		0.035835,
		0.08327,
		0.145377,
		0.218463,
		0.299063,
		0.383959,
		0.470189,
		0.555074,
		0.636233,
		0.711615,
		0.77953,
		0.838682,
		0.888213,
		0.927745,
		0.957423,
		0.977948,
		0.990579,
		0.997073,
		0.999515,
		0.999988,
		0.999998,
		0.999744,
		0.997978,
		0.9927,
		0.981785,
		0.963372,
		0.93607,
		0.899037,
		0.85199,
		0.795181,
		0.729352,
		0.655692,
		0.575799,
		0.491635,
		0.405501,
		0.320005,
		0.238038,
		0.162759,
		0.09757,
		0.046108,
		0.01223,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.000817,
		0.019264,
		0.058753,
		0.11513,
		0.184465,
		0.263058,
		0.347461,
		0.434489,
		0.52124,
		0.605119,
		0.683863,
		0.755573,
		0.818748,
		0.872333,
		0.915755,
		0.94898,
		0.972543,
		0.987569,
		0.995735,
		0.999128,
		0.999953,
		1,
		0.999876,
		0.998605,
		0.994227,
		0.98453,
		0.967506,
		0.94161,
		0.905864,
		0.859876,
		0.80382,
		0.738394,
		0.664769,
		0.584549,
		0.499726,
		0.41265,
		0.325995,
		0.242735,
		0.166123,
		0.099671,
		0.047138,
		0.012512,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.00731,
		0.037478,
		0.086952,
		0.151546,
		0.227317,
		0.31057,
		0.397884,
		0.486123,
		0.572465,
		0.654422,
		0.729875,
		0.797109,
		0.854856,
		0.902341,
		0.939331,
		0.966185,
		0.983876,
		0.993974,
		0.998541,
		0.999871,
		1,
		0.999951,
		0.999088,
		0.995546,
		0.987038,
		0.971415,
		0.946968,
		0.912568,
		0.867708,
		0.812474,
		0.747512,
		0.673969,
		0.593455,
		0.50799,
		0.419973,
		0.332145,
		0.247567,
		0.16959,
		0.101841,
		0.048204,
		0.012805,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.000857,
		0.020173,
		0.061429,
		0.120172,
		0.192195,
		0.273547,
		0.360554,
		0.449834,
		0.538323,
		0.623297,
		0.702404,
		0.773702,
		0.835698,
		0.887401,
		0.928371,
		0.958769,
		0.9794,
		0.991705,
		0.997692,
		0.999707,
		0.999998,
		0.999986,
		0.999445,
		0.996662,
		0.989308,
		0.975089,
		0.952128,
		0.919137,
		0.875474,
		0.821134,
		0.756699,
		0.683291,
		0.602518,
		0.51643,
		0.427474,
		0.338462,
		0.25254,
		0.173165,
		0.104082,
		0.049305,
		0.013107,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.007665,
		0.039236,
		0.09088,
		0.15811,
		0.236704,
		0.32272,
		0.412516,
		0.502771,
		0.590511,
		0.673139,
		0.74847,
		0.814777,
		0.870838,
		0.91599,
		0.950184,
		0.974035,
		0.988837,
		0.996509,
		0.999419,
		0.999986,
		0.999998,
		0.999693,
		0.997585,
		0.991337,
		0.978519,
		0.957078,
		0.925553,
		0.883162,
		0.829789,
		0.76595,
		0.692731,
		0.61174,
		0.52505,
		0.435159,
		0.34495,
		0.257659,
		0.176852,
		0.106396,
		0.050445,
		0.013421,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.000899,
		0.021147,
		0.064291,
		0.12555,
		0.200411,
		0.28465,
		0.374349,
		0.465916,
		0.556109,
		0.642073,
		0.721369,
		0.792019,
		0.852555,
		0.902078,
		0.940314,
		0.967671,
		0.985268,
		0.994912,
		0.998953,
		0.999944,
		1,
		0.999851,
		0.998327,
		0.993125,
		0.981696,
		0.961803,
		0.931803,
		0.890756,
		0.838428,
		0.775256,
		0.702287,
		0.62112,
		0.533853,
		0.443032,
		0.351615,
		0.262931,
		0.180656,
		0.108788,
		0.051625,
		0.013746,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.008046,
		0.041121,
		0.09508,
		0.165103,
		0.246666,
		0.335557,
		0.427896,
		0.520162,
		0.609221,
		0.692366,
		0.767353,
		0.832457,
		0.886525,
		0.929041,
		0.960187,
		0.980889,
		0.992809,
		0.998248,
		0.999844,
		1,
		0.999941,
		0.998903,
		0.994676,
		0.984612,
		0.966289,
		0.93787,
		0.898242,
		0.84704,
		0.784611,
		0.711954,
		0.630658,
		0.542841,
		0.451099,
		0.358463,
		0.268359,
		0.184582,
		0.111261,
		0.052846,
		0.014083,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.000945,
		0.022194,
		0.067358,
		0.131291,
		0.209152,
		0.296414,
		0.388893,
		0.482769,
		0.574618,
		0.661443,
		0.740722,
		0.810454,
		0.869219,
		0.916242,
		0.951461,
		0.975582,
		0.990101,
		0.997227,
		0.999645,
		0.999998,
		0.999983,
		0.99933,
		0.995996,
		0.987261,
		0.970524,
		0.943738,
		0.905603,
		0.85561,
		0.794004,
		0.721729,
		0.640354,
		0.552017,
		0.459363,
		0.365499,
		0.273952,
		0.188634,
		0.113819,
		0.054111,
		0.014433,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.008457,
		0.043145,
		0.099574,
		0.172561,
		0.25725,
		0.349131,
		0.444067,
		0.538323,
		0.628601,
		0.712079,
		0.786465,
		0.850054,
		0.901795,
		0.941362,
		0.969224,
		0.986675,
		0.995804,
		0.999296,
		0.999982,
		0.999998,
		0.999627,
		0.997092,
		0.989641,
		0.974494,
		0.949389,
		0.912824,
		0.864126,
		0.803427,
		0.731605,
		0.650207,
		0.561384,
		0.467831,
		0.372731,
		0.279714,
		0.192819,
		0.116464,
		0.055422,
		0.014796,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.000995,
		0.023321,
		0.070648,
		0.137432,
		0.218463,
		0.308888,
		0.404232,
		0.500431,
		0.593863,
		0.681392,
		0.760415,
		0.828924,
		0.885575,
		0.929759,
		0.961681,
		0.982413,
		0.993882,
		0.998733,
		0.999931,
		1,
		0.999819,
		0.997978,
		0.991748,
		0.978187,
		0.954807,
		0.919885,
		0.872571,
		0.812868,
		0.741577,
		0.660215,
		0.570945,
		0.476508,
		0.380165,
		0.285652,
		0.197141,
		0.119202,
		0.056781,
		0.015172,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.0089,
		0.045322,
		0.104393,
		0.180527,
		0.268504,
		0.363491,
		0.461072,
		0.557281,
		0.648648,
		0.732241,
		0.805731,
		0.867459,
		0.916514,
		0.952816,
		0.977183,
		0.991352,
		0.997877,
		0.999809,
		1,
		0.999928,
		0.998669,
		0.993585,
		0.981594,
		0.959974,
		0.926769,
		0.880929,
		0.822315,
		0.751638,
		0.670377,
		0.580701,
		0.485397,
		0.387806,
		0.291774,
		0.201607,
		0.122037,
		0.05819,
		0.015563,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.001049,
		0.024536,
		0.074184,
		0.144007,
		0.228392,
		0.322126,
		0.420418,
		0.518938,
		0.613858,
		0.701897,
		0.780384,
		0.847326,
		0.901488,
		0.942482,
		0.970847,
		0.988091,
		0.996638,
		0.999567,
		0.999997,
		0.99998,
		0.999183,
		0.995156,
		0.984704,
		0.964874,
		0.933455,
		0.889182,
		0.831754,
		0.761779,
		0.68069,
		0.590655,
		0.494505,
		0.395662,
		0.298086,
		0.206223,
		0.124973,
		0.059653,
		0.01597,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.009378,
		0.047667,
		0.109566,
		0.189047,
		0.280485,
		0.378694,
		0.478955,
		0.577059,
		0.669354,
		0.752804,
		0.82506,
		0.884543,
		0.930531,
		0.963257,
		0.983968,
		0.994912,
		0.99914,
		0.999978,
		0.999997,
		0.999544,
		0.996468,
		0.987511,
		0.969489,
		0.939924,
		0.897313,
		0.841172,
		0.771993,
		0.69115,
		0.600807,
		0.503836,
		0.403741,
		0.304595,
		0.210995,
		0.128015,
		0.06117,
		0.016392,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.001107,
		0.025848,
		0.07799,
		0.151058,
		0.238994,
		0.336187,
		0.437501,
		0.538323,
		0.634605,
		0.722924,
		0.800552,
		0.865538,
		0.916807,
		0.954256,
		0.978837,
		0.992579,
		0.998451,
		0.999916,
		1,
		0.999777,
		0.997533,
		0.990011,
		0.973805,
		0.946156,
		0.905299,
		0.850552,
		0.782268,
		0.701753,
		0.611159,
		0.513395,
		0.412048,
		0.31131,
		0.215931,
		0.131168,
		0.062746,
		0.016832,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.009896,
		0.050198,
		0.11513,
		0.19817,
		0.293251,
		0.394798,
		0.497762,
		0.597675,
		0.690698,
		0.773702,
		0.84434,
		0.901155,
		0.943681,
		0.972543,
		0.989507,
		0.997403,
		0.999765,
		1,
		0.999911,
		0.998368,
		0.992201,
		0.977805,
		0.952128,
		0.913121,
		0.859876,
		0.792593,
		0.712494,
		0.621711,
		0.523188,
		0.420591,
		0.318238,
		0.221038,
		0.134438,
		0.064384,
		0.017289,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.00117,
		0.027267,
		0.082095,
		0.158632,
		0.250328,
		0.351133,
		0.455537,
		0.558619,
		0.656102,
		0.744423,
		0.820817,
		0.883417,
		0.931363,
		0.964919,
		0.985548,
		0.995886,
		0.999465,
		0.999997,
		0.999975,
		0.998994,
		0.994084,
		0.981477,
		0.957819,
		0.920755,
		0.869125,
		0.802956,
		0.723368,
		0.632464,
		0.533218,
		0.429378,
		0.325389,
		0.226323,
		0.137829,
		0.066086,
		0.017765,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.010458,
		0.052935,
		0.121124,
		0.207954,
		0.306869,
		0.411865,
		0.517538,
		0.61914,
		0.712649,
		0.794849,
		0.863435,
		0.917126,
		0.955789,
		0.980542,
		0.993771,
		0.998938,
		0.999973,
		0.999996,
		0.999435,
		0.995666,
		0.98481,
		0.963209,
		0.928179,
		0.87828,
		0.81334,
		0.734365,
		0.643416,
		0.54349,
		0.438416,
		0.332769,
		0.231794,
		0.141349,
		0.067857,
		0.018262,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.001239,
		0.028807,
		0.086531,
		0.166781,
		0.262459,
		0.367032,
		0.474581,
		0.579853,
		0.678333,
		0.766326,
		0.841057,
		0.900792,
		0.944967,
		0.974313,
		0.990912,
		0.998086,
		0.999895,
		1,
		0.999723,
		0.996959,
		0.987794,
		0.968276,
		0.935365,
		0.887317,
		0.823732,
		0.745479,
		0.654566,
		0.554008,
		0.447714,
		0.34039,
		0.237461,
		0.145004,
		0.069699,
		0.018779,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.011069,
		0.055902,
		0.127594,
		0.218463,
		0.321409,
		0.42996,
		0.538323,
		0.641458,
		0.735156,
		0.816138,
		0.882183,
		0.932261,
		0.966673,
		0.987147,
		0.99679,
		0.999707,
		1,
		0.999889,
		0.997978,
		0.990425,
		0.972999,
		0.94229,
		0.896212,
		0.834112,
		0.756699,
		0.665912,
		0.564777,
		0.457278,
		0.34826,
		0.243331,
		0.148801,
		0.071616,
		0.019319,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.001314,
		0.030481,
		0.091333,
		0.175563,
		0.27546,
		0.383959,
		0.494689,
		0.602046,
		0.701268,
		0.788542,
		0.861121,
		0.917472,
		0.957423,
		0.982294,
		0.994912,
		0.999333,
		0.999996,
		0.999968,
		0.998747,
		0.9927,
		0.97736,
		0.948926,
		0.90494,
		0.84446,
		0.768013,
		0.677451,
		0.575799,
		0.467118,
		0.356389,
		0.249415,
		0.152746,
		0.073614,
		0.019882,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.011736,
		0.059124,
		0.134589,
		0.229766,
		0.336951,
		0.449152,
		0.560159,
		0.664622,
		0.758149,
		0.837433,
		0.900395,
		0.946351,
		0.976158,
		0.992293,
		0.998674,
		0.999966,
		0.999995,
		0.999293,
		0.994625,
		0.981342,
		0.955247,
		0.913472,
		0.854755,
		0.779408,
		0.689177,
		0.587077,
		0.47724,
		0.364787,
		0.255722,
		0.156848,
		0.075696,
		0.02047,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.001396,
		0.032305,
		0.096544,
		0.185043,
		0.289411,
		0.40199,
		0.515919,
		0.625212,
		0.72486,
		0.810953,
		0.880827,
		0.933236,
		0.968525,
		0.988753,
		0.997607,
		0.999867,
		1,
		0.999651,
		0.996208,
		0.984931,
		0.961225,
		0.921781,
		0.864974,
		0.790869,
		0.701086,
		0.598615,
		0.487653,
		0.373465,
		0.262263,
		0.161115,
		0.077867,
		0.021084,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.012464,
		0.062632,
		0.142168,
		0.241943,
		0.353578,
		0.469512,
		0.583079,
		0.688606,
		0.781532,
		0.858565,
		0.91785,
		0.959167,
		0.984089,
		0.995985,
		0.999629,
		1,
		0.999859,
		0.997465,
		0.988116,
		0.966832,
		0.929834,
		0.875088,
		0.802379,
		0.71317,
		0.610413,
		0.498366,
		0.382435,
		0.269049,
		0.165555,
		0.080132,
		0.021727,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.001486,
		0.034297,
		0.102209,
		0.195295,
		0.304401,
		0.42121,
		0.538323,
		0.649351,
		0.749038,
		0.833411,
		0.899959,
		0.947842,
		0.978079,
		0.993633,
		0.999156,
		0.999994,
		0.999959,
		0.998419,
		0.99089,
		0.972044,
		0.937598,
		0.88507,
		0.813918,
		0.72542,
		0.622473,
		0.509384,
		0.391707,
		0.276092,
		0.170178,
		0.082496,
		0.022399,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.013263,
		0.06646,
		0.150396,
		0.255081,
		0.371384,
		0.491113,
		0.607109,
		0.713365,
		0.805178,
		0.879328,
		0.934294,
		0.970479,
		0.990351,
		0.99832,
		0.999955,
		0.999993,
		0.999102,
		0.993251,
		0.976834,
		0.94504,
		0.894886,
		0.825462,
		0.737825,
		0.634793,
		0.520717,
		0.401293,
		0.283404,
		0.174994,
		0.084966,
		0.023103,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.001584,
		0.036478,
		0.108382,
		0.2064,
		0.320521,
		0.441699,
		0.561943,
		0.674436,
		0.773688,
		0.85571,
		0.918247,
		0.961014,
		0.985898,
		0.996952,
		0.999812,
		0.999982,
		0.999536,
		0.995193,
		0.981167,
		0.952111,
		0.904491,
		0.836975,
		0.750362,
		0.647364,
		0.532363,
		0.411199,
		0.290993,
		0.180011,
		0.087547,
		0.023841,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.014139,
		0.07064,
		0.159331,
		0.269249,
		0.390425,
		0.513969,
		0.632191,
		0.73875,
		0.82883,
		0.899376,
		0.949345,
		0.979961,
		0.994799,
		0.999412,
		0.999887,
		0.999705,
		0.996665,
		0.98496,
		0.958714,
		0.91379,
		0.848375,
		0.762966,
		0.66014,
		0.544295,
		0.421411,
		0.298855,
		0.185227,
		0.090239,
		0.024612,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.001693,
		0.038856,
		0.115075,
		0.218359,
		0.337737,
		0.463351,
		0.586569,
		0.700132,
		0.798332,
		0.877246,
		0.935005,
		0.972079,
		0.991455,
		0.998441,
		0.999516,
		0.999471,
		0.997502,
		0.988014,
		0.96463,
		0.922566,
		0.859466,
		0.77547,
		0.672987,
		0.556411,
		0.431857,
		0.306942,
		0.190619,
		0.093032,
		0.025414,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.015085,
		0.075127,
		0.168853,
		0.284216,
		0.410314,
		0.537502,
		0.657532,
		0.763735,
		0.851254,
		0.917318,
		0.961557,
		0.986257,
		0.996322,
		0.998418,
		0.998467,
		0.997321,
		0.989904,
		0.969415,
		0.930382,
		0.86984,
		0.787513,
		0.685602,
		0.56847,
		0.442357,
		0.315133,
		0.196111,
		0.09589,
		0.026239,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.001807,
		0.041348,
		0.122028,
		0.230652,
		0.355202,
		0.484947,
		0.610585,
		0.724419,
		0.820585,
		0.895339,
		0.947385,
		0.978195,
		0.992111,
		0.99577,
		0.99599,
		0.995414,
		0.98988,
		0.972293,
		0.936471,
		0.878774,
		0.798444,
		0.697422,
		0.580011,
		0.452559,
		0.323181,
		0.201554,
		0.098742,
		0.027066,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.016032,
		0.079568,
		0.178148,
		0.298578,
		0.428982,
		0.558942,
		0.679669,
		0.784235,
		0.867848,
		0.928207,
		0.96586,
		0.984431,
		0.990301,
		0.990914,
		0.990679,
		0.986795,
		0.972083,
		0.939661,
		0.885148,
		0.807238,
		0.707552,
		0.590294,
		0.461886,
		0.330677,
		0.206694,
		0.101464,
		0.027862,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.001912,
		0.043618,
		0.128242,
		0.241376,
		0.369959,
		0.502414,
		0.628819,
		0.741127,
		0.833445,
		0.902391,
		0.947489,
		0.971493,
		0.980301,
		0.981678,
		0.98162,
		0.979108,
		0.967195,
		0.938357,
		0.887425,
		0.812471,
		0.714734,
		0.59826,
		0.46951,
		0.337025,
		0.211158,
		0.103873,
		0.028577,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.016784,
		0.082994,
		0.185047,
		0.308696,
		0.441194,
		0.571468,
		0.690328,
		0.790745,
		0.868202,
		0.921114,
		0.951225,
		0.963724,
		0.966371,
		0.966433,
		0.96498,
		0.955728,
		0.930634,
		0.883723,
		0.812377,
		0.717384,
		0.602567,
		0.474364,
		0.341453,
		0.21446,
		0.105728,
		0.029144,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.001978,
		0.04496,
		0.131646,
		0.246645,
		0.37608,
		0.507751,
		0.63134,
		0.73868,
		0.824096,
		0.884835,
		0.921517,
		0.938409,
		0.942932,
		0.943208,
		0.942488,
		0.935677,
		0.914431,
		0.871991,
		0.805002,
		0.713725,
		0.601687,
		0.475222,
		0.343066,
		0.216028,
		0.106744,
		0.029484,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.017013,
		0.083785,
		0.185967,
		0.308649,
		0.438588,
		0.564404,
		0.676837,
		0.769076,
		0.837182,
		0.88057,
		0.902408,
		0.909445,
		0.910193,
		0.909906,
		0.905228,
		0.887842,
		0.850289,
		0.788449,
		0.701997,
		0.594078,
		0.470829,
		0.340931,
		0.215263,
		0.10662,
		0.029512,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.001962,
		0.044403,
		0.12944,
		0.241294,
		0.365832,
		0.49074,
		0.605771,
		0.703054,
		0.777497,
		0.827286,
		0.854357,
		0.864494,
		0.866109,
		0.866039,
		0.863107,
		0.849464,
		0.817112,
		0.761188,
		0.680727,
		0.578414,
		0.460076,
		0.334215,
		0.211619,
		0.105077,
		0.029149,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.01635,
		0.080172,
		0.177071,
		0.29224,
		0.412633,
		0.527188,
		0.627109,
		0.706256,
		0.761617,
		0.793834,
		0.807497,
		0.810449,
		0.81052,
		0.808892,
		0.798725,
		0.771729,
		0.722367,
		0.649019,
		0.553828,
		0.442204,
		0.322316,
		0.204691,
		0.101903,
		0.028334,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.001825,
		0.041135,
		0.119331,
		0.221226,
		0.333305,
		0.443925,
		0.543575,
		0.625228,
		0.684786,
		0.721612,
		0.738959,
		0.743703,
		0.743999,
		0.743232,
		0.736123,
		0.714427,
		0.672067,
		0.606781,
		0.520121,
		0.416966,
		0.305002,
		0.1943,
		0.096994,
		0.027034,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.014591,
		0.071204,
		0.156413,
		0.256554,
		0.359697,
		0.455881,
		0.537419,
		0.599306,
		0.639732,
		0.660568,
		0.667436,
		0.668181,
		0.667894,
		0.663304,
		0.646618,
		0.611418,
		0.554862,
		0.477883,
		0.384738,
		0.282483,
		0.180546,
		0.090387,
		0.025256,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.001561,
		0.035026,
		0.101068,
		0.186232,
		0.278635,
		0.368169,
		0.446768,
		0.508764,
		0.551352,
		0.575095,
		0.584203,
		0.585686,
		0.585621,
		0.582942,
		0.570757,
		0.542556,
		0.495029,
		0.428498,
		0.346533,
		0.255441,
		0.163827,
		0.082265,
		0.023047,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.011854,
		0.057546,
		0.125657,
		0.204699,
		0.284745,
		0.35766,
		0.417399,
		0.460409,
		0.486112,
		0.497293,
		0.499767,
		0.499827,
		0.498456,
		0.490085,
		0.468408,
		0.429797,
		0.374015,
		0.303917,
		0.224966,
		0.144809,
		0.072946,
		0.020493,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.001204,
		0.026875,
		0.077097,
		0.141108,
		0.209488,
		0.274347,
		0.329566,
		0.371134,
		0.397574,
		0.410368,
		0.413956,
		0.414184,
		0.413595,
		0.408251,
		0.392355,
		0.36215,
		0.316924,
		0.25883,
		0.192443,
		0.124352,
		0.06285,
		0.017708,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.008587,
		0.041445,
		0.089903,
		0.145338,
		0.200397,
		0.249185,
		0.287533,
		0.313356,
		0.32705,
		0.331687,
		0.332201,
		0.332002,
		0.328883,
		0.317842,
		0.295193,
		0.259871,
		0.213377,
		0.159395,
		0.103418,
		0.052454,
		0.014824,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.000818,
		0.018148,
		0.051724,
		0.09396,
		0.138285,
		0.17929,
		0.212934,
		0.236823,
		0.25055,
		0.255969,
		0.256874,
		0.256834,
		0.255208,
		0.248009,
		0.231825,
		0.205375,
		0.169596,
		0.127325,
		0.082967,
		0.042238,
		0.011975,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.005387,
		0.025838,
		0.055636,
		0.089175,
		0.121741,
		0.149661,
		0.170509,
		0.183376,
		0.189147,
		0.19046,
		0.190493,
		0.18976,
		0.185408,
		0.174471,
		0.1556,
		0.129275,
		0.097571,
		0.063871,
		0.032644,
		0.009287,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.000471,
		0.010384,
		0.029382,
		0.052927,
		0.077134,
		0.098875,
		0.115924,
		0.127154,
		0.132767,
		0.134389,
		0.134495,
		0.134224,
		0.131824,
		0.124906,
		0.112184,
		0.09381,
		0.071206,
		0.046839,
		0.02404,
		0.006863,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.00278,
		0.013239,
		0.028272,
		0.044879,
		0.06058,
		0.073513,
		0.082561,
		0.087528,
		0.089266,
		0.089464,
		0.089387,
		0.088209,
		0.084171,
		0.076164,
		0.064131,
		0.048973,
		0.032382,
		0.016693,
		0.004784,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.000214,
		0.004677,
		0.013129,
		0.023426,
		0.033761,
		0.042718,
		0.049356,
		0.053314,
		0.054932,
		0.055211,
		0.055198,
		0.054701,
		0.05257,
		0.047945,
		0.040669,
		0.031258,
		0.020783,
		0.010764,
		0.003097,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.001059,
		0.005004,
		0.010587,
		0.016622,
		0.02215,
		0.026482,
		0.029267,
		0.030563,
		0.030868,
		0.030876,
		0.030705,
		0.029718,
		0.027329,
		0.023365,
		0.018083,
		0.012095,
		0.006296,
		0.001819,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		6.5e-005,
		0.001406,
		0.003911,
		0.006903,
		0.009823,
		0.012245,
		0.013913,
		0.014781,
		0.015042,
		0.015059,
		0.015015,
		0.014633,
		0.013573,
		0.011703,
		0.009125,
		0.006142,
		0.003214,
		0.000933,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000228,
		0.001066,
		0.002233,
		0.003462,
		0.004545,
		0.005342,
		0.005799,
		0.005965,
		0.005985,
		0.005977,
		0.005863,
		0.005488,
		0.004775,
		0.003753,
		0.002543,
		0.001339,
		0.00039,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		0.000175,
		0.000481,
		0.000838,
		0.001175,
		0.00144,
		0.001605,
		0.001676,
		0.001688,
		0.001688,
		0.001665,
		0.001573,
		0.001382,
		0.001096,
		0.000748,
		0.000396,
		0.000116,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		1e-005,
		4.7e-005,
		9.6e-005,
		0.000147,
		0.00019,
		0.000219,
		0.000233,
		0.000237,
		0.000237,
		0.000235,
		0.000224,
		0.000199,
		0.000159,
		0.000109,
		5.8e-005,
		1.7e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		1.7e-005,
		5.8e-005,
		0.000109,
		0.000159,
		0.000199,
		0.000224,
		0.000235,
		0.000237,
		0.000237,
		0.000233,
		0.000219,
		0.00019,
		0.000147,
		9.6e-005,
		4.7e-005,
		1e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		0.000116,
		0.000396,
		0.000748,
		0.001096,
		0.001382,
		0.001573,
		0.001665,
		0.001688,
		0.001688,
		0.001676,
		0.001605,
		0.00144,
		0.001175,
		0.000838,
		0.000481,
		0.000175,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		0.00039,
		0.001339,
		0.002543,
		0.003753,
		0.004775,
		0.005488,
		0.005863,
		0.005977,
		0.005985,
		0.005965,
		0.005799,
		0.005342,
		0.004545,
		0.003462,
		0.002233,
		0.001066,
		0.000228,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000933,
		0.003214,
		0.006142,
		0.009125,
		0.011703,
		0.013573,
		0.014633,
		0.015015,
		0.015059,
		0.015042,
		0.014781,
		0.013913,
		0.012245,
		0.009823,
		0.006903,
		0.003911,
		0.001406,
		6.5e-005,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.001819,
		0.006296,
		0.012095,
		0.018083,
		0.023365,
		0.027329,
		0.029718,
		0.030705,
		0.030876,
		0.030868,
		0.030563,
		0.029267,
		0.026482,
		0.02215,
		0.016622,
		0.010587,
		0.005004,
		0.001059,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.003097,
		0.010764,
		0.020783,
		0.031258,
		0.040669,
		0.047945,
		0.05257,
		0.054701,
		0.055198,
		0.055211,
		0.054932,
		0.053314,
		0.049356,
		0.042718,
		0.033761,
		0.023426,
		0.013129,
		0.004677,
		0.000214,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.004784,
		0.016693,
		0.032382,
		0.048973,
		0.064131,
		0.076164,
		0.084171,
		0.088209,
		0.089387,
		0.089464,
		0.089266,
		0.087528,
		0.082561,
		0.073513,
		0.06058,
		0.044879,
		0.028272,
		0.013239,
		0.00278,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.006863,
		0.02404,
		0.046839,
		0.071206,
		0.09381,
		0.112184,
		0.124906,
		0.131824,
		0.134224,
		0.134495,
		0.134389,
		0.132767,
		0.127154,
		0.115924,
		0.098875,
		0.077134,
		0.052927,
		0.029382,
		0.010384,
		0.000471,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.009287,
		0.032644,
		0.063871,
		0.097571,
		0.129275,
		0.1556,
		0.174471,
		0.185408,
		0.18976,
		0.190493,
		0.19046,
		0.189147,
		0.183376,
		0.170509,
		0.149661,
		0.121741,
		0.089175,
		0.055636,
		0.025838,
		0.005387,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.011975,
		0.042238,
		0.082967,
		0.127325,
		0.169596,
		0.205375,
		0.231825,
		0.248009,
		0.255208,
		0.256834,
		0.256874,
		0.255969,
		0.25055,
		0.236823,
		0.212934,
		0.17929,
		0.138285,
		0.09396,
		0.051724,
		0.018148,
		0.000818,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.014824,
		0.052454,
		0.103418,
		0.159395,
		0.213377,
		0.259871,
		0.295193,
		0.317842,
		0.328883,
		0.332002,
		0.332201,
		0.331687,
		0.32705,
		0.313356,
		0.287533,
		0.249185,
		0.200397,
		0.145338,
		0.089903,
		0.041445,
		0.008587,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.017708,
		0.06285,
		0.124352,
		0.192443,
		0.25883,
		0.316924,
		0.36215,
		0.392355,
		0.408251,
		0.413595,
		0.414184,
		0.413956,
		0.410368,
		0.397574,
		0.371134,
		0.329566,
		0.274347,
		0.209488,
		0.141108,
		0.077097,
		0.026875,
		0.001204,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.020493,
		0.072946,
		0.144809,
		0.224966,
		0.303917,
		0.374015,
		0.429797,
		0.468408,
		0.490085,
		0.498456,
		0.499827,
		0.499767,
		0.497293,
		0.486112,
		0.460409,
		0.417399,
		0.35766,
		0.284745,
		0.204699,
		0.125657,
		0.057546,
		0.011854,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.023047,
		0.082265,
		0.163827,
		0.255441,
		0.346533,
		0.428498,
		0.495029,
		0.542556,
		0.570757,
		0.582942,
		0.585621,
		0.585686,
		0.584203,
		0.575095,
		0.551352,
		0.508764,
		0.446768,
		0.368169,
		0.278635,
		0.186232,
		0.101068,
		0.035026,
		0.001561,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.025256,
		0.090387,
		0.180546,
		0.282483,
		0.384738,
		0.477883,
		0.554862,
		0.611418,
		0.646618,
		0.663304,
		0.667894,
		0.668181,
		0.667436,
		0.660568,
		0.639732,
		0.599306,
		0.537419,
		0.455881,
		0.359697,
		0.256554,
		0.156413,
		0.071204,
		0.014591,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.027034,
		0.096994,
		0.1943,
		0.305002,
		0.416966,
		0.520121,
		0.606781,
		0.672067,
		0.714427,
		0.736123,
		0.743232,
		0.743999,
		0.743703,
		0.738959,
		0.721612,
		0.684786,
		0.625228,
		0.543575,
		0.443925,
		0.333305,
		0.221226,
		0.119331,
		0.041135,
		0.001825,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.028334,
		0.101903,
		0.204691,
		0.322316,
		0.442204,
		0.553828,
		0.649019,
		0.722367,
		0.771729,
		0.798725,
		0.808892,
		0.81052,
		0.810449,
		0.807497,
		0.793834,
		0.761617,
		0.706256,
		0.627109,
		0.527188,
		0.412633,
		0.29224,
		0.177071,
		0.080172,
		0.01635,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.029149,
		0.105077,
		0.211619,
		0.334215,
		0.460076,
		0.578414,
		0.680727,
		0.761188,
		0.817112,
		0.849464,
		0.863107,
		0.866039,
		0.866109,
		0.864494,
		0.854357,
		0.827286,
		0.777497,
		0.703054,
		0.605771,
		0.49074,
		0.365832,
		0.241294,
		0.12944,
		0.044403,
		0.001962,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.029512,
		0.10662,
		0.215263,
		0.340931,
		0.470829,
		0.594078,
		0.701997,
		0.788449,
		0.850289,
		0.887842,
		0.905228,
		0.909906,
		0.910193,
		0.909445,
		0.902408,
		0.88057,
		0.837182,
		0.769076,
		0.676837,
		0.564404,
		0.438588,
		0.308649,
		0.185967,
		0.083785,
		0.017013,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.029484,
		0.106744,
		0.216028,
		0.343066,
		0.475222,
		0.601687,
		0.713725,
		0.805002,
		0.871991,
		0.914431,
		0.935677,
		0.942488,
		0.943208,
		0.942932,
		0.938409,
		0.921517,
		0.884835,
		0.824096,
		0.73868,
		0.63134,
		0.507751,
		0.37608,
		0.246645,
		0.131646,
		0.04496,
		0.001978,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.029144,
		0.105728,
		0.21446,
		0.341453,
		0.474364,
		0.602567,
		0.717384,
		0.812377,
		0.883723,
		0.930634,
		0.955728,
		0.96498,
		0.966433,
		0.966371,
		0.963724,
		0.951225,
		0.921114,
		0.868202,
		0.790745,
		0.690328,
		0.571468,
		0.441194,
		0.308696,
		0.185047,
		0.082994,
		0.016784,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.028577,
		0.103873,
		0.211158,
		0.337025,
		0.46951,
		0.59826,
		0.714734,
		0.812471,
		0.887425,
		0.938357,
		0.967195,
		0.979108,
		0.98162,
		0.981678,
		0.980301,
		0.971493,
		0.947489,
		0.902391,
		0.833445,
		0.741127,
		0.628819,
		0.502414,
		0.369959,
		0.241376,
		0.128242,
		0.043618,
		0.001912,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.027862,
		0.101464,
		0.206694,
		0.330677,
		0.461886,
		0.590294,
		0.707552,
		0.807238,
		0.885148,
		0.939661,
		0.972083,
		0.986795,
		0.990679,
		0.990914,
		0.990301,
		0.984431,
		0.96586,
		0.928207,
		0.867848,
		0.784235,
		0.679669,
		0.558942,
		0.428982,
		0.298578,
		0.178148,
		0.079568,
		0.016032,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.027066,
		0.098742,
		0.201554,
		0.323181,
		0.452559,
		0.580011,
		0.697422,
		0.798444,
		0.878774,
		0.936471,
		0.972293,
		0.98988,
		0.995414,
		0.99599,
		0.99577,
		0.992111,
		0.978195,
		0.947385,
		0.895339,
		0.820585,
		0.724419,
		0.610585,
		0.484947,
		0.355202,
		0.230652,
		0.122028,
		0.041348,
		0.001807,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.026239,
		0.09589,
		0.196111,
		0.315133,
		0.442357,
		0.56847,
		0.685602,
		0.787513,
		0.86984,
		0.930382,
		0.969415,
		0.989904,
		0.997321,
		0.998467,
		0.998418,
		0.996322,
		0.986257,
		0.961557,
		0.917318,
		0.851254,
		0.763735,
		0.657532,
		0.537502,
		0.410314,
		0.284216,
		0.168853,
		0.075127,
		0.015085,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.025414,
		0.093032,
		0.190619,
		0.306942,
		0.431857,
		0.556411,
		0.672987,
		0.77547,
		0.859466,
		0.922566,
		0.96463,
		0.988014,
		0.997502,
		0.999471,
		0.999516,
		0.998441,
		0.991455,
		0.972079,
		0.935005,
		0.877246,
		0.798332,
		0.700132,
		0.586569,
		0.463351,
		0.337737,
		0.218359,
		0.115075,
		0.038856,
		0.001693,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.024612,
		0.090239,
		0.185227,
		0.298855,
		0.421411,
		0.544295,
		0.66014,
		0.762966,
		0.848375,
		0.91379,
		0.958714,
		0.98496,
		0.996665,
		0.999705,
		0.999887,
		0.999412,
		0.994799,
		0.979961,
		0.949345,
		0.899376,
		0.82883,
		0.73875,
		0.632191,
		0.513969,
		0.390425,
		0.269249,
		0.159331,
		0.07064,
		0.014139,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.023841,
		0.087547,
		0.180011,
		0.290993,
		0.411199,
		0.532363,
		0.647364,
		0.750362,
		0.836975,
		0.904491,
		0.952111,
		0.981167,
		0.995193,
		0.999536,
		0.999982,
		0.999812,
		0.996952,
		0.985898,
		0.961014,
		0.918247,
		0.85571,
		0.773688,
		0.674436,
		0.561943,
		0.441699,
		0.320521,
		0.2064,
		0.108382,
		0.036478,
		0.001584,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.023103,
		0.084966,
		0.174994,
		0.283404,
		0.401293,
		0.520717,
		0.634793,
		0.737825,
		0.825462,
		0.894886,
		0.94504,
		0.976834,
		0.993251,
		0.999102,
		0.999993,
		0.999955,
		0.99832,
		0.990351,
		0.970479,
		0.934294,
		0.879328,
		0.805178,
		0.713365,
		0.607109,
		0.491113,
		0.371384,
		0.255081,
		0.150396,
		0.06646,
		0.013263,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.022399,
		0.082496,
		0.170178,
		0.276092,
		0.391707,
		0.509384,
		0.622473,
		0.72542,
		0.813918,
		0.88507,
		0.937598,
		0.972044,
		0.99089,
		0.998419,
		0.999959,
		0.999994,
		0.999156,
		0.993633,
		0.978079,
		0.947842,
		0.899959,
		0.833411,
		0.749038,
		0.649351,
		0.538323,
		0.42121,
		0.304401,
		0.195295,
		0.102209,
		0.034297,
		0.001486,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.021727,
		0.080132,
		0.165555,
		0.269049,
		0.382435,
		0.498366,
		0.610413,
		0.71317,
		0.802379,
		0.875088,
		0.929834,
		0.966832,
		0.988116,
		0.997465,
		0.999859,
		1,
		0.999629,
		0.995985,
		0.984089,
		0.959167,
		0.91785,
		0.858565,
		0.781532,
		0.688606,
		0.583079,
		0.469512,
		0.353578,
		0.241943,
		0.142168,
		0.062632,
		0.012464,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.021084,
		0.077867,
		0.161115,
		0.262263,
		0.373465,
		0.487653,
		0.598615,
		0.701086,
		0.790869,
		0.864974,
		0.921781,
		0.961225,
		0.984931,
		0.996208,
		0.999651,
		1,
		0.999867,
		0.997607,
		0.988753,
		0.968525,
		0.933236,
		0.880827,
		0.810953,
		0.72486,
		0.625212,
		0.515919,
		0.40199,
		0.289411,
		0.185043,
		0.096544,
		0.032305,
		0.001396,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.02047,
		0.075696,
		0.156848,
		0.255722,
		0.364787,
		0.47724,
		0.587077,
		0.689177,
		0.779408,
		0.854755,
		0.913472,
		0.955247,
		0.981342,
		0.994625,
		0.999293,
		0.999995,
		0.999966,
		0.998674,
		0.992293,
		0.976158,
		0.946351,
		0.900395,
		0.837433,
		0.758149,
		0.664622,
		0.560159,
		0.449152,
		0.336951,
		0.229766,
		0.134589,
		0.059124,
		0.011736,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.019882,
		0.073614,
		0.152746,
		0.249415,
		0.356389,
		0.467118,
		0.575799,
		0.677451,
		0.768013,
		0.84446,
		0.90494,
		0.948926,
		0.97736,
		0.9927,
		0.998747,
		0.999968,
		0.999996,
		0.999333,
		0.994912,
		0.982294,
		0.957423,
		0.917472,
		0.861121,
		0.788542,
		0.701268,
		0.602046,
		0.494689,
		0.383959,
		0.27546,
		0.175563,
		0.091333,
		0.030481,
		0.001314,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.019319,
		0.071616,
		0.148801,
		0.243331,
		0.34826,
		0.457278,
		0.564777,
		0.665912,
		0.756699,
		0.834112,
		0.896212,
		0.94229,
		0.972999,
		0.990425,
		0.997978,
		0.999889,
		1,
		0.999707,
		0.99679,
		0.987147,
		0.966673,
		0.932261,
		0.882183,
		0.816138,
		0.735156,
		0.641458,
		0.538323,
		0.42996,
		0.321409,
		0.218463,
		0.127594,
		0.055902,
		0.011069,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.018779,
		0.069699,
		0.145004,
		0.237461,
		0.34039,
		0.447714,
		0.554008,
		0.654566,
		0.745479,
		0.823732,
		0.887317,
		0.935365,
		0.968276,
		0.987794,
		0.996959,
		0.999723,
		1,
		0.999895,
		0.998086,
		0.990912,
		0.974313,
		0.944967,
		0.900792,
		0.841057,
		0.766326,
		0.678333,
		0.579853,
		0.474581,
		0.367032,
		0.262459,
		0.166781,
		0.086531,
		0.028807,
		0.001239,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.018262,
		0.067857,
		0.141349,
		0.231794,
		0.332769,
		0.438416,
		0.54349,
		0.643416,
		0.734365,
		0.81334,
		0.87828,
		0.928179,
		0.963209,
		0.98481,
		0.995666,
		0.999435,
		0.999996,
		0.999973,
		0.998938,
		0.993771,
		0.980542,
		0.955789,
		0.917126,
		0.863435,
		0.794849,
		0.712649,
		0.61914,
		0.517538,
		0.411865,
		0.306869,
		0.207954,
		0.121124,
		0.052935,
		0.010458,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.017765,
		0.066086,
		0.137829,
		0.226323,
		0.325389,
		0.429378,
		0.533218,
		0.632464,
		0.723368,
		0.802956,
		0.869125,
		0.920755,
		0.957819,
		0.981477,
		0.994084,
		0.998994,
		0.999975,
		0.999997,
		0.999465,
		0.995886,
		0.985548,
		0.964919,
		0.931363,
		0.883417,
		0.820817,
		0.744423,
		0.656102,
		0.558619,
		0.455537,
		0.351133,
		0.250328,
		0.158632,
		0.082095,
		0.027267,
		0.00117,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.017289,
		0.064384,
		0.134438,
		0.221038,
		0.318238,
		0.420591,
		0.523188,
		0.621711,
		0.712494,
		0.792593,
		0.859876,
		0.913121,
		0.952128,
		0.977805,
		0.992201,
		0.998368,
		0.999911,
		1,
		0.999765,
		0.997403,
		0.989507,
		0.972543,
		0.943681,
		0.901155,
		0.84434,
		0.773702,
		0.690698,
		0.597675,
		0.497762,
		0.394798,
		0.293251,
		0.19817,
		0.11513,
		0.050198,
		0.009896,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.016832,
		0.062746,
		0.131168,
		0.215931,
		0.31131,
		0.412048,
		0.513395,
		0.611159,
		0.701753,
		0.782268,
		0.850552,
		0.905299,
		0.946156,
		0.973805,
		0.990011,
		0.997533,
		0.999777,
		1,
		0.999916,
		0.998451,
		0.992579,
		0.978837,
		0.954256,
		0.916807,
		0.865538,
		0.800552,
		0.722924,
		0.634605,
		0.538323,
		0.437501,
		0.336187,
		0.238994,
		0.151058,
		0.07799,
		0.025848,
		0.001107,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.016392,
		0.06117,
		0.128015,
		0.210995,
		0.304595,
		0.403741,
		0.503836,
		0.600807,
		0.69115,
		0.771993,
		0.841172,
		0.897313,
		0.939924,
		0.969489,
		0.987511,
		0.996468,
		0.999544,
		0.999997,
		0.999978,
		0.99914,
		0.994912,
		0.983968,
		0.963257,
		0.930531,
		0.884543,
		0.82506,
		0.752804,
		0.669354,
		0.577059,
		0.478955,
		0.378694,
		0.280485,
		0.189047,
		0.109566,
		0.047667,
		0.009378,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.01597,
		0.059653,
		0.124973,
		0.206223,
		0.298086,
		0.395662,
		0.494505,
		0.590655,
		0.68069,
		0.761779,
		0.831754,
		0.889182,
		0.933455,
		0.964874,
		0.984704,
		0.995156,
		0.999183,
		0.99998,
		0.999997,
		0.999567,
		0.996638,
		0.988091,
		0.970847,
		0.942482,
		0.901488,
		0.847326,
		0.780384,
		0.701897,
		0.613858,
		0.518938,
		0.420418,
		0.322126,
		0.228392,
		0.144007,
		0.074184,
		0.024536,
		0.001049,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.015563,
		0.05819,
		0.122037,
		0.201607,
		0.291774,
		0.387806,
		0.485397,
		0.580701,
		0.670377,
		0.751638,
		0.822315,
		0.880929,
		0.926769,
		0.959974,
		0.981594,
		0.993585,
		0.998669,
		0.999928,
		1,
		0.999809,
		0.997877,
		0.991352,
		0.977183,
		0.952816,
		0.916514,
		0.867459,
		0.805731,
		0.732241,
		0.648648,
		0.557281,
		0.461072,
		0.363491,
		0.268504,
		0.180527,
		0.104393,
		0.045322,
		0.0089,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.015172,
		0.056781,
		0.119202,
		0.197141,
		0.285652,
		0.380165,
		0.476508,
		0.570945,
		0.660215,
		0.741577,
		0.812868,
		0.872571,
		0.919885,
		0.954807,
		0.978187,
		0.991748,
		0.997978,
		0.999819,
		1,
		0.999931,
		0.998733,
		0.993882,
		0.982413,
		0.961681,
		0.929759,
		0.885575,
		0.828924,
		0.760415,
		0.681392,
		0.593863,
		0.500431,
		0.404232,
		0.308888,
		0.218463,
		0.137432,
		0.070648,
		0.023321,
		0.000995,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.014796,
		0.055422,
		0.116464,
		0.192819,
		0.279714,
		0.372731,
		0.467831,
		0.561384,
		0.650207,
		0.731605,
		0.803427,
		0.864126,
		0.912824,
		0.949389,
		0.974494,
		0.989641,
		0.997092,
		0.999627,
		0.999998,
		0.999982,
		0.999296,
		0.995804,
		0.986675,
		0.969224,
		0.941362,
		0.901795,
		0.850054,
		0.786465,
		0.712079,
		0.628601,
		0.538323,
		0.444067,
		0.349131,
		0.25725,
		0.172561,
		0.099574,
		0.043145,
		0.008457,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.014433,
		0.054111,
		0.113819,
		0.188634,
		0.273952,
		0.365499,
		0.459363,
		0.552017,
		0.640354,
		0.721729,
		0.794004,
		0.85561,
		0.905603,
		0.943738,
		0.970524,
		0.987261,
		0.995996,
		0.99933,
		0.999983,
		0.999998,
		0.999645,
		0.997227,
		0.990101,
		0.975582,
		0.951461,
		0.916242,
		0.869219,
		0.810454,
		0.740722,
		0.661443,
		0.574618,
		0.482769,
		0.388893,
		0.296414,
		0.209152,
		0.131291,
		0.067358,
		0.022194,
		0.000945,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.014083,
		0.052846,
		0.111261,
		0.184582,
		0.268359,
		0.358463,
		0.451099,
		0.542841,
		0.630658,
		0.711954,
		0.784611,
		0.84704,
		0.898242,
		0.93787,
		0.966289,
		0.984612,
		0.994676,
		0.998903,
		0.999941,
		1,
		0.999844,
		0.998248,
		0.992809,
		0.980889,
		0.960187,
		0.929041,
		0.886525,
		0.832457,
		0.767353,
		0.692366,
		0.609221,
		0.520162,
		0.427896,
		0.335557,
		0.246666,
		0.165103,
		0.09508,
		0.041121,
		0.008046,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.013746,
		0.051625,
		0.108788,
		0.180656,
		0.262931,
		0.351615,
		0.443032,
		0.533853,
		0.62112,
		0.702287,
		0.775256,
		0.838428,
		0.890756,
		0.931803,
		0.961803,
		0.981696,
		0.993125,
		0.998327,
		0.999851,
		1,
		0.999944,
		0.998953,
		0.994912,
		0.985268,
		0.967671,
		0.940314,
		0.902078,
		0.852555,
		0.792019,
		0.721369,
		0.642073,
		0.556109,
		0.465916,
		0.374349,
		0.28465,
		0.200411,
		0.12555,
		0.064291,
		0.021147,
		0.000899,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.013421,
		0.050445,
		0.106396,
		0.176852,
		0.257659,
		0.34495,
		0.435159,
		0.52505,
		0.61174,
		0.692731,
		0.76595,
		0.829789,
		0.883162,
		0.925553,
		0.957078,
		0.978519,
		0.991337,
		0.997585,
		0.999693,
		0.999998,
		0.999986,
		0.999419,
		0.996509,
		0.988837,
		0.974035,
		0.950184,
		0.91599,
		0.870838,
		0.814777,
		0.74847,
		0.673139,
		0.590511,
		0.502771,
		0.412516,
		0.32272,
		0.236704,
		0.15811,
		0.09088,
		0.039236,
		0.007665,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.013107,
		0.049305,
		0.104082,
		0.173165,
		0.25254,
		0.338462,
		0.427474,
		0.51643,
		0.602518,
		0.683291,
		0.756699,
		0.821134,
		0.875474,
		0.919137,
		0.952128,
		0.975089,
		0.989308,
		0.996662,
		0.999445,
		0.999986,
		0.999998,
		0.999707,
		0.997692,
		0.991705,
		0.9794,
		0.958769,
		0.928371,
		0.887401,
		0.835698,
		0.773702,
		0.702404,
		0.623297,
		0.538323,
		0.449834,
		0.360554,
		0.273547,
		0.192195,
		0.120172,
		0.061429,
		0.020173,
		0.000857,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.012805,
		0.048204,
		0.101841,
		0.16959,
		0.247567,
		0.332145,
		0.419973,
		0.50799,
		0.593455,
		0.673969,
		0.747512,
		0.812474,
		0.867708,
		0.912568,
		0.946968,
		0.971415,
		0.987038,
		0.995546,
		0.999088,
		0.999951,
		1,
		0.999871,
		0.998541,
		0.993974,
		0.983876,
		0.966185,
		0.939331,
		0.902341,
		0.854856,
		0.797109,
		0.729875,
		0.654422,
		0.572465,
		0.486123,
		0.397884,
		0.31057,
		0.227317,
		0.151546,
		0.086952,
		0.037478,
		0.00731,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.012512,
		0.047138,
		0.099671,
		0.166123,
		0.242735,
		0.325995,
		0.41265,
		0.499726,
		0.584549,
		0.664769,
		0.738394,
		0.80382,
		0.859876,
		0.905864,
		0.94161,
		0.967506,
		0.98453,
		0.994227,
		0.998605,
		0.999876,
		1,
		0.999953,
		0.999128,
		0.995735,
		0.987569,
		0.972543,
		0.94898,
		0.915755,
		0.872333,
		0.818748,
		0.755573,
		0.683863,
		0.605119,
		0.52124,
		0.434489,
		0.347461,
		0.263058,
		0.184465,
		0.11513,
		0.058753,
		0.019264,
		0.000817,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.01223,
		0.046108,
		0.09757,
		0.162759,
		0.238038,
		0.320005,
		0.405501,
		0.491635,
		0.575799,
		0.655692,
		0.729352,
		0.795181,
		0.85199,
		0.899037,
		0.93607,
		0.963372,
		0.981785,
		0.9927,
		0.997978,
		0.999744,
		0.999998,
		0.999988,
		0.999515,
		0.997073,
		0.990579,
		0.977948,
		0.957423,
		0.927745,
		0.888213,
		0.838682,
		0.77953,
		0.711615,
		0.636233,
		0.555074,
		0.470189,
		0.383959,
		0.299063,
		0.218463,
		0.145377,
		0.08327,
		0.035835,
		0.006979,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.011957,
		0.045111,
		0.095534,
		0.159495,
		0.233473,
		0.31417,
		0.398522,
		0.483714,
		0.567204,
		0.64674,
		0.72039,
		0.786565,
		0.844063,
		0.892101,
		0.930361,
		0.959025,
		0.97881,
		0.990961,
		0.997195,
		0.999536,
		0.999989,
		0.999998,
		0.999755,
		0.998064,
		0.992999,
		0.982503,
		0.964764,
		0.938407,
		0.902584,
		0.85698,
		0.801791,
		0.737688,
		0.665773,
		0.58754,
		0.504839,
		0.419845,
		0.33503,
		0.253142,
		0.177183,
		0.110396,
		0.056247,
		0.018415,
		0.00078,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.011692,
		0.044146,
		0.09356,
		0.156326,
		0.229035,
		0.308487,
		0.391707,
		0.475959,
		0.558763,
		0.637915,
		0.711512,
		0.77798,
		0.836105,
		0.88507,
		0.924495,
		0.954475,
		0.975609,
		0.989007,
		0.996245,
		0.999236,
		0.999959,
		1,
		0.999892,
		0.998776,
		0.994912,
		0.986304,
		0.971103,
		0.947842,
		0.915536,
		0.873716,
		0.822408,
		0.762104,
		0.693726,
		0.618579,
		0.538323,
		0.45494,
		0.370703,
		0.288158,
		0.210104,
		0.139573,
		0.079817,
		0.034297,
		0.006671,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.011437,
		0.043211,
		0.091646,
		0.15325,
		0.224718,
		0.302949,
		0.385053,
		0.468368,
		0.550474,
		0.629217,
		0.702724,
		0.769433,
		0.828126,
		0.877955,
		0.918486,
		0.949734,
		0.972191,
		0.986838,
		0.995118,
		0.998827,
		0.999897,
		1,
		0.999961,
		0.999268,
		0.996398,
		0.98944,
		0.976536,
		0.956143,
		0.927158,
		0.888967,
		0.841438,
		0.784898,
		0.720092,
		0.648151,
		0.570554,
		0.489096,
		0.405867,
		0.32322,
		0.243759,
		0.170317,
		0.105945,
		0.053897,
		0.017621,
		0.000745,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.01119,
		0.042305,
		0.089791,
		0.150262,
		0.220519,
		0.297552,
		0.378555,
		0.460936,
		0.542336,
		0.620647,
		0.694027,
		0.760931,
		0.820135,
		0.870768,
		0.912347,
		0.944814,
		0.968563,
		0.984457,
		0.993807,
		0.998294,
		0.999785,
		0.999999,
		0.99999,
		0.999593,
		0.997527,
		0.991996,
		0.981155,
		0.963404,
		0.937538,
		0.902812,
		0.858946,
		0.80611,
		0.744885,
		0.676234,
		0.601465,
		0.522198,
		0.440344,
		0.358077,
		0.277816,
		0.202206,
		0.134105,
		0.076573,
		0.032856,
		0.006382,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.010951,
		0.041428,
		0.08799,
		0.14736,
		0.216434,
		0.292293,
		0.372208,
		0.45366,
		0.534347,
		0.612205,
		0.685426,
		0.75248,
		0.812141,
		0.863519,
		0.906089,
		0.939726,
		0.964734,
		0.981865,
		0.992308,
		0.997627,
		0.99961,
		0.99999,
		0.999999,
		0.999794,
		0.998364,
		0.99405,
		0.985047,
		0.969716,
		0.946765,
		0.915332,
		0.875,
		0.82579,
		0.768129,
		0.702821,
		0.63101,
		0.554152,
		0.473988,
		0.392521,
		0.311993,
		0.234873,
		0.163836,
		0.101756,
		0.051692,
		0.016877,
		0.000713,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.010719,
		0.040578,
		0.086243,
		0.14454,
		0.212459,
		0.287165,
		0.366009,
		0.446537,
		0.526505,
		0.603892,
		0.676922,
		0.744084,
		0.804153,
		0.85622,
		0.899725,
		0.934481,
		0.960712,
		0.979066,
		0.990617,
		0.996812,
		0.999355,
		0.999966,
		1,
		0.999909,
		0.998965,
		0.995675,
		0.988294,
		0.975164,
		0.954925,
		0.926607,
		0.889668,
		0.843991,
		0.789856,
		0.727916,
		0.65916,
		0.584888,
		0.50668,
		0.426374,
		0.346048,
		0.268,
		0.194735,
		0.128949,
		0.073521,
		0.031504,
		0.006111,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.010495,
		0.039753,
		0.084547,
		0.141799,
		0.208589,
		0.282166,
		0.359954,
		0.439564,
		0.518807,
		0.595706,
		0.668518,
		0.735749,
		0.796177,
		0.84888,
		0.893264,
		0.92909,
		0.956507,
		0.976067,
		0.988734,
		0.995843,
		0.999006,
		0.999913,
		1,
		0.999967,
		0.999381,
		0.996937,
		0.990973,
		0.979834,
		0.962103,
		0.936718,
		0.903024,
		0.860772,
		0.810106,
		0.751535,
		0.685901,
		0.614355,
		0.538323,
		0.45949,
		0.379775,
		0.301315,
		0.226451,
		0.157713,
		0.097809,
		0.049618,
		0.016179,
		0.000682,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.010277,
		0.038954,
		0.0829,
		0.139134,
		0.204822,
		0.277292,
		0.354039,
		0.432738,
		0.511251,
		0.587648,
		0.660215,
		0.727478,
		0.78822,
		0.841508,
		0.886718,
		0.923565,
		0.952128,
		0.972873,
		0.986658,
		0.994712,
		0.998551,
		0.999819,
		0.999999,
		0.999992,
		0.999656,
		0.997897,
		0.993157,
		0.983803,
		0.968379,
		0.945745,
		0.91514,
		0.876195,
		0.828924,
		0.773702,
		0.71123,
		0.642517,
		0.568845,
		0.491749,
		0.413,
		0.334582,
		0.258679,
		0.187663,
		0.124081,
		0.070648,
		0.030233,
		0.005858,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.010067,
		0.038178,
		0.081301,
		0.136543,
		0.201154,
		0.272538,
		0.34826,
		0.426054,
		0.503836,
		0.579717,
		0.652015,
		0.719275,
		0.780288,
		0.834112,
		0.880097,
		0.917916,
		0.947585,
		0.969489,
		0.984391,
		0.993413,
		0.997978,
		0.999669,
		0.999992,
		0.999999,
		0.999826,
		0.998608,
		0.994912,
		0.987147,
		0.973833,
		0.953766,
		0.926089,
		0.890323,
		0.846361,
		0.794448,
		0.735156,
		0.669354,
		0.598188,
		0.523054,
		0.445578,
		0.367599,
		0.291153,
		0.218463,
		0.151921,
		0.094085,
		0.047667,
		0.015523,
		0.000654,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.009862,
		0.037425,
		0.079747,
		0.134022,
		0.197581,
		0.267901,
		0.342613,
		0.41951,
		0.496559,
		0.571912,
		0.643919,
		0.711144,
		0.772387,
		0.826699,
		0.873411,
		0.912153,
		0.942888,
		0.965925,
		0.981935,
		0.991944,
		0.997277,
		0.999452,
		0.999971,
		1,
		0.999923,
		0.999119,
		0.996301,
		0.989936,
		0.978541,
		0.960857,
		0.935944,
		0.903223,
		0.862471,
		0.813813,
		0.757694,
		0.694856,
		0.626313,
		0.553326,
		0.477387,
		0.400195,
		0.323646,
		0.249819,
		0.180962,
		0.119481,
		0.067939,
		0.029038,
		0.00562,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.009664,
		0.036694,
		0.078237,
		0.13157,
		0.194101,
		0.263377,
		0.337094,
		0.413102,
		0.489418,
		0.564232,
		0.635927,
		0.703088,
		0.764523,
		0.819279,
		0.866667,
		0.906287,
		0.938045,
		0.962186,
		0.979295,
		0.990301,
		0.996439,
		0.999153,
		0.999926,
		1,
		0.999972,
		0.999473,
		0.99738,
		0.992239,
		0.982575,
		0.967092,
		0.944778,
		0.91496,
		0.877309,
		0.831837,
		0.778869,
		0.719025,
		0.653193,
		0.582507,
		0.508329,
		0.432226,
		0.355962,
		0.281477,
		0.21088,
		0.146439,
		0.090569,
		0.045828,
		0.014907,
		0.000627,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.009472,
		0.035984,
		0.07677,
		0.129184,
		0.19071,
		0.258963,
		0.331701,
		0.406828,
		0.48241,
		0.556677,
		0.628041,
		0.695109,
		0.756699,
		0.811856,
		0.859876,
		0.900326,
		0.933068,
		0.95828,
		0.976473,
		0.988485,
		0.995458,
		0.998762,
		0.999846,
		0.999999,
		0.999993,
		0.999707,
		0.9982,
		0.994115,
		0.986004,
		0.972543,
		0.952661,
		0.925601,
		0.890936,
		0.848568,
		0.798713,
		0.741872,
		0.678813,
		0.610552,
		0.538323,
		0.463572,
		0.387931,
		0.313213,
		0.241394,
		0.174607,
		0.11513,
		0.065382,
		0.027912,
		0.005396,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.009285,
		0.035295,
		0.075343,
		0.126861,
		0.187405,
		0.254655,
		0.326429,
		0.400685,
		0.475533,
		0.549244,
		0.620261,
		0.68721,
		0.748921,
		0.804438,
		0.853044,
		0.894279,
		0.927963,
		0.954217,
		0.973475,
		0.986495,
		0.994328,
		0.998267,
		0.999718,
		0.999993,
		0.999999,
		0.999852,
		0.998808,
		0.995624,
		0.988892,
		0.977279,
		0.959663,
		0.935212,
		0.903409,
		0.864056,
		0.81726,
		0.763413,
		0.70317,
		0.637429,
		0.567308,
		0.494131,
		0.419412,
		0.344836,
		0.272257,
		0.203676,
		0.141244,
		0.087245,
		0.044093,
		0.014327,
		0.000602,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.009104,
		0.034625,
		0.073955,
		0.1246,
		0.184184,
		0.25045,
		0.321275,
		0.394668,
		0.468785,
		0.541933,
		0.612586,
		0.679392,
		0.741192,
		0.797031,
		0.84618,
		0.888157,
		0.922742,
		0.950003,
		0.970307,
		0.984332,
		0.993045,
		0.99766,
		0.999531,
		0.999975,
		1,
		0.999934,
		0.999246,
		0.996818,
		0.991302,
		0.981366,
		0.965851,
		0.943859,
		0.91479,
		0.878352,
		0.834551,
		0.783673,
		0.726267,
		0.663118,
		0.595234,
		0.523824,
		0.450283,
		0.376183,
		0.303253,
		0.233376,
		0.168575,
		0.11101,
		0.062966,
		0.02685,
		0.005185,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.008928,
		0.033974,
		0.072605,
		0.122398,
		0.181043,
		0.246346,
		0.316237,
		0.388776,
		0.462163,
		0.534743,
		0.605018,
		0.671658,
		0.733515,
		0.789639,
		0.839291,
		0.881966,
		0.917412,
		0.945646,
		0.966975,
		0.981999,
		0.991607,
		0.996932,
		0.999274,
		0.999937,
		1,
		0.999976,
		0.999548,
		0.997745,
		0.993291,
		0.984868,
		0.971292,
		0.951607,
		0.92514,
		0.89151,
		0.850628,
		0.802682,
		0.748116,
		0.687611,
		0.622067,
		0.552583,
		0.480445,
		0.407111,
		0.334196,
		0.263466,
		0.196828,
		0.136317,
		0.084099,
		0.042454,
		0.013779,
		0.000579,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.008758,
		0.033341,
		0.071291,
		0.120254,
		0.177981,
		0.242338,
		0.31131,
		0.383005,
		0.455665,
		0.527672,
		0.597555,
		0.664008,
		0.725895,
		0.782268,
		0.832383,
		0.875716,
		0.911982,
		0.941155,
		0.963484,
		0.979499,
		0.990011,
		0.996078,
		0.998935,
		0.999868,
		0.999999,
		0.999994,
		0.999749,
		0.998451,
		0.994912,
		0.987845,
		0.976047,
		0.958519,
		0.934518,
		0.903584,
		0.865538,
		0.820474,
		0.768736,
		0.710906,
		0.647781,
		0.580358,
		0.509814,
		0.437501,
		0.364925,
		0.29374,
		0.22574,
		0.162846,
		0.107106,
		0.060681,
		0.025848,
		0.004986,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.008592,
		0.032726,
		0.070013,
		0.118164,
		0.174994,
		0.238425,
		0.306492,
		0.377353,
		0.449289,
		0.520718,
		0.590198,
		0.656444,
		0.718333,
		0.774923,
		0.825463,
		0.869413,
		0.90646,
		0.936538,
		0.959841,
		0.976835,
		0.988257,
		0.99509,
		0.998506,
		0.999758,
		0.999994,
		0.999999,
		0.999873,
		0.998974,
		0.996216,
		0.990352,
		0.980179,
		0.964657,
		0.942986,
		0.91463,
		0.879329,
		0.837084,
		0.78815,
		0.73301,
		0.672365,
		0.607109,
		0.538323,
		0.467253,
		0.395301,
		0.324015,
		0.255081,
		0.190312,
		0.131641,
		0.081119,
		0.040905,
		0.013263,
		0.000557,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.00843,
		0.032127,
		0.068768,
		0.116129,
		0.172081,
		0.234603,
		0.301781,
		0.371816,
		0.443032,
		0.51388,
		0.582946,
		0.648967,
		0.710833,
		0.767608,
		0.818536,
		0.863065,
		0.900855,
		0.931803,
		0.956054,
		0.974012,
		0.986346,
		0.993965,
		0.997978,
		0.999596,
		0.999979,
		1,
		0.999944,
		0.99935,
		0.997247,
		0.992444,
		0.983744,
		0.970081,
		0.950601,
		0.924704,
		0.892049,
		0.852555,
		0.806386,
		0.753936,
		0.69581,
		0.63281,
		0.565918,
		0.496282,
		0.425205,
		0.354133,
		0.28465,
		0.218463,
		0.1574,
		0.103402,
		0.058518,
		0.0249,
		0.004799,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.008273,
		0.031545,
		0.067557,
		0.114144,
		0.169238,
		0.230869,
		0.297172,
		0.366393,
		0.436892,
		0.507156,
		0.575799,
		0.641577,
		0.703397,
		0.760326,
		0.811609,
		0.856678,
		0.895174,
		0.926957,
		0.952128,
		0.971035,
		0.984278,
		0.9927,
		0.997343,
		0.999373,
		0.999946,
		1,
		0.99998,
		0.999611,
		0.998049,
		0.994171,
		0.986798,
		0.974848,
		0.957423,
		0.93386,
		0.903749,
		0.866927,
		0.823476,
		0.773702,
		0.71812,
		0.657442,
		0.592557,
		0.524519,
		0.454537,
		0.383959,
		0.31427,
		0.247077,
		0.184108,
		0.127199,
		0.078295,
		0.039439,
		0.012775,
		0.000536,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.008121,
		0.030978,
		0.066377,
		0.11221,
		0.166465,
		0.227222,
		0.292664,
		0.361079,
		0.430867,
		0.500544,
		0.568755,
		0.634275,
		0.696026,
		0.753082,
		0.804685,
		0.850259,
		0.889423,
		0.922008,
		0.948071,
		0.967907,
		0.982057,
		0.991293,
		0.996595,
		0.999079,
		0.999886,
		0.999999,
		0.999995,
		0.999783,
		0.998659,
		0.995579,
		0.989394,
		0.979014,
		0.963506,
		0.942154,
		0.914479,
		0.880246,
		0.839455,
		0.79233,
		0.739302,
		0.680995,
		0.618209,
		0.551909,
		0.483212,
		0.413374,
		0.343786,
		0.27596,
		0.211524,
		0.152218,
		0.099885,
		0.056468,
		0.024004,
		0.004622,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.007973,
		0.030427,
		0.065227,
		0.110324,
		0.163757,
		0.223658,
		0.288253,
		0.355873,
		0.424954,
		0.494044,
		0.561815,
		0.627063,
		0.688723,
		0.745878,
		0.797771,
		0.843815,
		0.883611,
		0.916963,
		0.943889,
		0.964635,
		0.979683,
		0.989742,
		0.995728,
		0.998706,
		0.999791,
		0.999995,
		0.999999,
		0.99989,
		0.999112,
		0.996711,
		0.991581,
		0.982633,
		0.968908,
		0.94964,
		0.924291,
		0.892557,
		0.854361,
		0.809849,
		0.759371,
		0.703467,
		0.642853,
		0.578407,
		0.51116,
		0.442279,
		0.373065,
		0.304937,
		0.239433,
		0.178197,
		0.122976,
		0.075614,
		0.038051,
		0.012313,
		0.000516,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.007828,
		0.02989,
		0.064107,
		0.108485,
		0.161115,
		0.220175,
		0.283937,
		0.350772,
		0.41915,
		0.487653,
		0.554977,
		0.619939,
		0.681489,
		0.738718,
		0.790869,
		0.83735,
		0.877744,
		0.91183,
		0.93959,
		0.961225,
		0.977161,
		0.988048,
		0.994738,
		0.998244,
		0.999651,
		0.999982,
		1,
		0.999951,
		0.999437,
		0.997607,
		0.993406,
		0.985756,
		0.973679,
		0.956371,
		0.933236,
		0.903905,
		0.868232,
		0.826287,
		0.778343,
		0.72486,
		0.666474,
		0.603979,
		0.538323,
		0.470589,
		0.40199,
		0.333861,
		0.267647,
		0.204903,
		0.147285,
		0.096544,
		0.054523,
		0.023155,
		0.004454,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.007688,
		0.029367,
		0.063015,
		0.106691,
		0.158534,
		0.21677,
		0.279714,
		0.345773,
		0.413455,
		0.48137,
		0.54824,
		0.612904,
		0.674326,
		0.731605,
		0.783986,
		0.83087,
		0.871829,
		0.906615,
		0.935179,
		0.957681,
		0.974494,
		0.98621,
		0.993622,
		0.997688,
		0.999457,
		0.999953,
		1,
		0.999982,
		0.999663,
		0.998304,
		0.994912,
		0.988431,
		0.977873,
		0.962398,
		0.941362,
		0.914337,
		0.881109,
		0.841678,
		0.796242,
		0.745185,
		0.689066,
		0.628601,
		0.564656,
		0.498233,
		0.430464,
		0.362598,
		0.295997,
		0.232129,
		0.172561,
		0.118957,
		0.073068,
		0.036734,
		0.011876,
		0.000497,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.007551,
		0.028857,
		0.061951,
		0.104941,
		0.156015,
		0.213443,
		0.27558,
		0.340874,
		0.407865,
		0.475193,
		0.541604,
		0.605958,
		0.667234,
		0.72454,
		0.777124,
		0.824381,
		0.865871,
		0.901325,
		0.930665,
		0.954009,
		0.971686,
		0.98423,
		0.992377,
		0.99703,
		0.9992,
		0.999901,
		0.999999,
		0.999996,
		0.999812,
		0.998834,
		0.99614,
		0.990705,
		0.981538,
		0.967772,
		0.948721,
		0.9239,
		0.893035,
		0.856057,
		0.813093,
		0.764456,
		0.710629,
		0.652255,
		0.59012,
		0.525152,
		0.4584,
		0.391034,
		0.324337,
		0.259691,
		0.198581,
		0.142584,
		0.093366,
		0.052677,
		0.02235,
		0.004296,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.007418,
		0.028361,
		0.060913,
		0.103233,
		0.153554,
		0.210189,
		0.271535,
		0.336073,
		0.402378,
		0.469119,
		0.535067,
		0.599101,
		0.660215,
		0.717527,
		0.770286,
		0.817887,
		0.859876,
		0.895967,
		0.926053,
		0.950217,
		0.968741,
		0.982109,
		0.991001,
		0.996266,
		0.998873,
		0.999819,
		0.999996,
		0.999999,
		0.999905,
		0.999227,
		0.997128,
		0.992621,
		0.984721,
		0.972543,
		0.955361,
		0.932642,
		0.904051,
		0.869459,
		0.828924,
		0.78269,
		0.73117,
		0.674931,
		0.61469,
		0.551297,
		0.485726,
		0.419073,
		0.352538,
		0.287427,
		0.225145,
		0.167184,
		0.11513,
		0.070648,
		0.035484,
		0.011462,
		0.000479,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.007289,
		0.027878,
		0.059902,
		0.101566,
		0.151149,
		0.207007,
		0.267574,
		0.331368,
		0.396993,
		0.463149,
		0.528629,
		0.592333,
		0.65327,
		0.710566,
		0.763477,
		0.811392,
		0.85385,
		0.890546,
		0.92135,
		0.946308,
		0.965663,
		0.97985,
		0.989494,
		0.995392,
		0.998468,
		0.999696,
		0.999984,
		1,
		0.999958,
		0.99951,
		0.99791,
		0.99422,
		0.987468,
		0.976757,
		0.961329,
		0.940607,
		0.914202,
		0.881923,
		0.843766,
		0.79991,
		0.750696,
		0.696627,
		0.638345,
		0.57663,
		0.512384,
		0.446628,
		0.380488,
		0.315194,
		0.252073,
		0.192542,
		0.138102,
		0.090341,
		0.050923,
		0.021587,
		0.004145,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.007163,
		0.027406,
		0.058915,
		0.099939,
		0.148801,
		0.203896,
		0.263696,
		0.326755,
		0.391707,
		0.457278,
		0.522288,
		0.585653,
		0.646399,
		0.703661,
		0.756699,
		0.804901,
		0.847798,
		0.88507,
		0.916561,
		0.94229,
		0.962458,
		0.977455,
		0.987855,
		0.994403,
		0.997978,
		0.999527,
		0.999959,
		1,
		0.999985,
		0.999707,
		0.998518,
		0.995539,
		0.989821,
		0.980459,
		0.966673,
		0.947842,
		0.923529,
		0.893487,
		0.857652,
		0.816138,
		0.769224,
		0.717343,
		0.661072,
		0.601121,
		0.538323,
		0.473628,
		0.40809,
		0.342867,
		0.27921,
		0.218463,
		0.162051,
		0.111483,
		0.068345,
		0.034297,
		0.011069,
		0.000462,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.00704,
		0.026947,
		0.057952,
		0.09835,
		0.146505,
		0.200852,
		0.2599,
		0.322233,
		0.386518,
		0.451507,
		0.516043,
		0.579062,
		0.639602,
		0.696813,
		0.749955,
		0.798418,
		0.841725,
		0.879542,
		0.911694,
		0.938168,
		0.959131,
		0.974928,
		0.986085,
		0.993298,
		0.997398,
		0.999302,
		0.999914,
		0.999999,
		0.999996,
		0.999836,
		0.998981,
		0.996615,
		0.991821,
		0.983695,
		0.971437,
		0.954391,
		0.932076,
		0.904191,
		0.870615,
		0.831403,
		0.78677,
		0.737086,
		0.682865,
		0.624748,
		0.563503,
		0.500011,
		0.435259,
		0.370333,
		0.306416,
		0.244776,
		0.186768,
		0.133826,
		0.087461,
		0.049255,
		0.020862,
		0.004003,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.00692,
		0.026499,
		0.057012,
		0.096799,
		0.144262,
		0.197875,
		0.256181,
		0.317799,
		0.381424,
		0.445833,
		0.509893,
		0.572558,
		0.632882,
		0.690022,
		0.743248,
		0.791947,
		0.835636,
		0.87397,
		0.906753,
		0.933948,
		0.955686,
		0.972271,
		0.984185,
		0.992073,
		0.996723,
		0.999014,
		0.999842,
		0.999996,
		1,
		0.999917,
		0.999324,
		0.997481,
		0.993506,
		0.986506,
		0.975665,
		0.960299,
		0.939885,
		0.914074,
		0.88269,
		0.845732,
		0.803355,
		0.755868,
		0.703721,
		0.647496,
		0.587893,
		0.525728,
		0.461923,
		0.397498,
		0.333566,
		0.271328,
		0.212067,
		0.157147,
		0.108006,
		0.066153,
		0.033168,
		0.010696,
		0.000447,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.006803,
		0.026062,
		0.056095,
		0.095283,
		0.142069,
		0.194962,
		0.25254,
		0.313452,
		0.376422,
		0.440255,
		0.503836,
		0.56614,
		0.626236,
		0.683291,
		0.736579,
		0.78549,
		0.829535,
		0.868358,
		0.901745,
		0.929636,
		0.952128,
		0.969489,
		0.982156,
		0.990728,
		0.995947,
		0.998657,
		0.999735,
		0.999986,
		1,
		0.999963,
		0.999572,
		0.998167,
		0.994912,
		0.988931,
		0.9794,
		0.965609,
		0.946999,
		0.923176,
		0.893914,
		0.859156,
		0.819002,
		0.773702,
		0.723646,
		0.669354,
		0.611467,
		0.550737,
		0.488022,
		0.424278,
		0.360554,
		0.297982,
		0.237781,
		0.181245,
		0.129743,
		0.084714,
		0.047667,
		0.020173,
		0.003868,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.006689,
		0.025636,
		0.0552,
		0.093803,
		0.139926,
		0.192112,
		0.248973,
		0.309189,
		0.371512,
		0.43477,
		0.497872,
		0.55981,
		0.619667,
		0.67662,
		0.729952,
		0.779051,
		0.823426,
		0.862711,
		0.896676,
		0.925236,
		0.948463,
		0.966586,
		0.980001,
		0.989263,
		0.995069,
		0.998225,
		0.999586,
		0.999965,
		1,
		0.999987,
		0.999743,
		0.9987,
		0.996073,
		0.991008,
		0.982681,
		0.970362,
		0.953459,
		0.931537,
		0.904322,
		0.871707,
		0.833736,
		0.790604,
		0.742645,
		0.690319,
		0.634207,
		0.575003,
		0.513503,
		0.450601,
		0.387283,
		0.324619,
		0.263763,
		0.205942,
		0.152459,
		0.104687,
		0.064064,
		0.032094,
		0.010342,
		0.000431,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.006578,
		0.02522,
		0.054327,
		0.092357,
		0.137829,
		0.189322,
		0.245479,
		0.305009,
		0.366691,
		0.429378,
		0.491999,
		0.553565,
		0.613173,
		0.670011,
		0.723368,
		0.772634,
		0.817314,
		0.857034,
		0.891549,
		0.920755,
		0.944695,
		0.963565,
		0.977722,
		0.987677,
		0.994084,
		0.997711,
		0.999388,
		0.999925,
		1,
		0.999997,
		0.999857,
		0.999106,
		0.99702,
		0.992774,
		0.985548,
		0.974599,
		0.959306,
		0.939196,
		0.913952,
		0.883417,
		0.847584,
		0.806596,
		0.76073,
		0.71039,
		0.656102,
		0.5985,
		0.538323,
		0.476406,
		0.413672,
		0.351133,
		0.289878,
		0.231074,
		0.17596,
		0.125842,
		0.082095,
		0.046154,
		0.019517,
		0.003739,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.00647,
		0.024814,
		0.053473,
		0.090944,
		0.135779,
		0.186592,
		0.242056,
		0.300909,
		0.361957,
		0.424076,
		0.486216,
		0.547405,
		0.606755,
		0.663465,
		0.716828,
		0.76624,
		0.811201,
		0.851331,
		0.886372,
		0.916199,
		0.94083,
		0.960431,
		0.975321,
		0.98597,
		0.992991,
		0.997111,
		0.999134,
		0.999861,
		0.999997,
		1,
		0.999927,
		0.999407,
		0.997782,
		0.994262,
		0.988039,
		0.978359,
		0.964579,
		0.946192,
		0.92284,
		0.894319,
		0.860575,
		0.821699,
		0.777914,
		0.729573,
		0.677145,
		0.621209,
		0.562448,
		0.501639,
		0.439648,
		0.377429,
		0.31601,
		0.256499,
		0.200072,
		0.147975,
		0.101517,
		0.062072,
		0.031071,
		0.010005,
		0.000417,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.006364,
		0.024418,
		0.05264,
		0.089563,
		0.133774,
		0.183919,
		0.238701,
		0.296887,
		0.357308,
		0.418863,
		0.480522,
		0.54133,
		0.600413,
		0.656981,
		0.710335,
		0.759872,
		0.805092,
		0.845607,
		0.881147,
		0.911571,
		0.936872,
		0.957188,
		0.972801,
		0.984144,
		0.991788,
		0.996422,
		0.998818,
		0.999767,
		0.999988,
		1,
		0.999968,
		0.999624,
		0.998385,
		0.995504,
		0.990188,
		0.98168,
		0.969317,
		0.952563,
		0.931022,
		0.904447,
		0.872739,
		0.835936,
		0.794215,
		0.747875,
		0.697333,
		0.643115,
		0.585848,
		0.526257,
		0.465151,
		0.403426,
		0.342055,
		0.282087,
		0.224639,
		0.170898,
		0.122113,
		0.079594,
		0.044712,
		0.018893,
		0.003617,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.006261,
		0.024031,
		0.051826,
		0.088212,
		0.131812,
		0.181302,
		0.235414,
		0.292943,
		0.352743,
		0.413737,
		0.474914,
		0.535338,
		0.594146,
		0.650561,
		0.70389,
		0.753533,
		0.79899,
		0.839866,
		0.875881,
		0.906878,
		0.932827,
		0.95384,
		0.970166,
		0.9822,
		0.990474,
		0.995638,
		0.998435,
		0.999636,
		0.999969,
		1,
		0.999988,
		0.999775,
		0.998854,
		0.996529,
		0.992029,
		0.984597,
		0.973558,
		0.958348,
		0.938536,
		0.913836,
		0.884104,
		0.849334,
		0.809651,
		0.765307,
		0.716668,
		0.664208,
		0.608503,
		0.550223,
		0.490126,
		0.429052,
		0.36792,
		0.307723,
		0.249522,
		0.194445,
		0.143683,
		0.098489,
		0.060171,
		0.030097,
		0.009684,
		0.000403,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.006161,
		0.023653,
		0.05103,
		0.086892,
		0.129893,
		0.178739,
		0.232193,
		0.289073,
		0.34826,
		0.408697,
		0.469393,
		0.529428,
		0.587956,
		0.644206,
		0.697495,
		0.747226,
		0.792898,
		0.834112,
		0.870578,
		0.902123,
		0.9287,
		0.950392,
		0.967419,
		0.98014,
		0.989048,
		0.994759,
		0.997978,
		0.999461,
		0.999934,
		1,
		0.999997,
		0.999874,
		0.999212,
		0.997366,
		0.993593,
		0.987147,
		0.977338,
		0.963582,
		0.945417,
		0.922519,
		0.894703,
		0.861918,
		0.824244,
		0.781884,
		0.735156,
		0.684484,
		0.630396,
		0.573509,
		0.514529,
		0.454246,
		0.393526,
		0.333306,
		0.274594,
		0.218463,
		0.166048,
		0.118545,
		0.077206,
		0.043337,
		0.018298,
		0.003501,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.006063,
		0.023285,
		0.050253,
		0.085601,
		0.128015,
		0.17623,
		0.229035,
		0.285276,
		0.343856,
		0.403741,
		0.463957,
		0.523601,
		0.58184,
		0.637915,
		0.69115,
		0.740951,
		0.786818,
		0.828347,
		0.865241,
		0.897313,
		0.924495,
		0.946848,
		0.964563,
		0.977965,
		0.987511,
		0.99378,
		0.997444,
		0.999236,
		0.999878,
		0.999997,
		1,
		0.999936,
		0.999477,
		0.998039,
		0.994912,
		0.989361,
		0.980692,
		0.968302,
		0.951701,
		0.930531,
		0.904566,
		0.873716,
		0.838015,
		0.79762,
		0.752804,
		0.703943,
		0.651515,
		0.596089,
		0.538323,
		0.478955,
		0.4188,
		0.358744,
		0.299743,
		0.242817,
		0.189047,
		0.139573,
		0.095593,
		0.058356,
		0.029167,
		0.009378,
		0.00039,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.005967,
		0.022924,
		0.049493,
		0.084339,
		0.126177,
		0.173771,
		0.225939,
		0.28155,
		0.339531,
		0.398867,
		0.458604,
		0.517854,
		0.575799,
		0.631689,
		0.684856,
		0.734711,
		0.780754,
		0.822577,
		0.859876,
		0.89245,
		0.920217,
		0.943213,
		0.961602,
		0.975678,
		0.985864,
		0.9927,
		0.996829,
		0.998956,
		0.999795,
		0.999989,
		1,
		0.999972,
		0.999668,
		0.998572,
		0.996013,
		0.991272,
		0.983655,
		0.972543,
		0.957423,
		0.937904,
		0.913726,
		0.884756,
		0.850988,
		0.812535,
		0.769624,
		0.722587,
		0.671853,
		0.617947,
		0.561477,
		0.503133,
		0.443682,
		0.383959,
		0.32487,
		0.267385,
		0.212531,
		0.161399,
		0.11513,
		0.074922,
		0.042023,
		0.017731,
		0.00339,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.005873,
		0.022572,
		0.04875,
		0.083104,
		0.124377,
		0.171364,
		0.222905,
		0.277895,
		0.335282,
		0.394074,
		0.453333,
		0.512188,
		0.569832,
		0.625529,
		0.678615,
		0.728508,
		0.774707,
		0.816804,
		0.854485,
		0.887541,
		0.91587,
		0.939491,
		0.95854,
		0.973282,
		0.984106,
		0.991519,
		0.996128,
		0.998614,
		0.999679,
		0.999973,
		1,
		0.99999,
		0.999801,
		0.998987,
		0.996922,
		0.992908,
		0.986257,
		0.976338,
		0.962616,
		0.944673,
		0.922214,
		0.895068,
		0.863189,
		0.826649,
		0.78563,
		0.740422,
		0.691409,
		0.639069,
		0.583966,
		0.526742,
		0.468117,
		0.40888,
		0.349886,
		0.292057,
		0.236371,
		0.183866,
		0.135635,
		0.092821,
		0.056622,
		0.02828,
		0.009086,
		0.000378,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.005782,
		0.022228,
		0.048024,
		0.081895,
		0.122616,
		0.169005,
		0.219929,
		0.274307,
		0.331109,
		0.38936,
		0.448143,
		0.5066,
		0.563939,
		0.619433,
		0.672427,
		0.722343,
		0.768681,
		0.811031,
		0.849074,
		0.882588,
		0.91146,
		0.935686,
		0.955381,
		0.97078,
		0.98224,
		0.990235,
		0.99534,
		0.998208,
		0.999523,
		0.999942,
		1,
		0.999997,
		0.999889,
		0.999303,
		0.997663,
		0.9943,
		0.988531,
		0.97972,
		0.967315,
		0.950871,
		0.930061,
		0.90468,
		0.874642,
		0.839981,
		0.800837,
		0.757456,
		0.710181,
		0.659446,
		0.60577,
		0.54975,
		0.492061,
		0.433445,
		0.374712,
		0.316735,
		0.260446,
		0.206833,
		0.156939,
		0.111859,
		0.072738,
		0.040769,
		0.01719,
		0.003284,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.005692,
		0.021892,
		0.047314,
		0.080713,
		0.120891,
		0.166693,
		0.217011,
		0.270786,
		0.327009,
		0.384724,
		0.443032,
		0.501091,
		0.55812,
		0.613403,
		0.666294,
		0.716218,
		0.762678,
		0.805262,
		0.843645,
		0.877597,
		0.90699,
		0.931803,
		0.952128,
		0.968173,
		0.980266,
		0.988848,
		0.994461,
		0.997731,
		0.999323,
		0.999892,
		0.999997,
		1,
		0.999944,
		0.999537,
		0.99826,
		0.995473,
		0.990506,
		0.982721,
		0.971552,
		0.95653,
		0.937299,
		0.913621,
		0.885375,
		0.852555,
		0.815262,
		0.773702,
		0.728175,
		0.679074,
		0.626874,
		0.572131,
		0.515474,
		0.457602,
		0.39928,
		0.341334,
		0.28465,
		0.230171,
		0.178891,
		0.131859,
		0.090168,
		0.054963,
		0.027432,
		0.008808,
		0.000366,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.005605,
		0.021564,
		0.046619,
		0.079556,
		0.119202,
		0.164428,
		0.214149,
		0.267329,
		0.32298,
		0.380165,
		0.438,
		0.495659,
		0.552374,
		0.607439,
		0.660215,
		0.710133,
		0.756699,
		0.799498,
		0.838201,
		0.872571,
		0.902465,
		0.927846,
		0.948785,
		0.965467,
		0.978187,
		0.987358,
		0.99349,
		0.99718,
		0.999074,
		0.999819,
		0.999991,
		1,
		0.999975,
		0.999707,
		0.998733,
		0.996452,
		0.99221,
		0.985372,
		0.975358,
		0.961681,
		0.943959,
		0.921922,
		0.895415,
		0.864394,
		0.828924,
		0.789171,
		0.745396,
		0.69795,
		0.647268,
		0.593863,
		0.538323,
		0.481304,
		0.423526,
		0.365774,
		0.308888,
		0.253765,
		0.201355,
		0.152659,
		0.108724,
		0.070648,
		0.039569,
		0.016673,
		0.003183,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.002221,
		0.014447,
		0.036235,
		0.066356,
		0.103621,
		0.146884,
		0.195044,
		0.247045,
		0.301882,
		0.358598,
		0.416289,
		0.474108,
		0.531263,
		0.587025,
		0.64073,
		0.691782,
		0.739658,
		0.783913,
		0.824185,
		0.860201,
		0.891784,
		0.918855,
		0.941445,
		0.959694,
		0.973858,
		0.984305,
		0.99151,
		0.996042,
		0.998532,
		0.999635,
		0.999962,
		1,
		0.999995,
		0.999861,
		0.999217,
		0.997501,
		0.994072,
		0.988279,
		0.979518,
		0.967266,
		0.951098,
		0.930704,
		0.905884,
		0.876554,
		0.842738,
		0.804566,
		0.762263,
		0.716147,
		0.666622,
		0.614173,
		0.559361,
		0.502815,
		0.445233,
		0.387376,
		0.330063,
		0.27417,
		0.220628,
		0.170419,
		0.124573,
		0.084169,
		0.05033,
		0.024223,
		0.00706,
		9.1e-005,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		0.004379,
		0.019337,
		0.043674,
		0.076143,
		0.115538,
		0.160699,
		0.210511,
		0.26391,
		0.31988,
		0.377459,
		0.435741,
		0.493875,
		0.551073,
		0.606611,
		0.659834,
		0.710157,
		0.757075,
		0.800163,
		0.839085,
		0.873601,
		0.903567,
		0.92895,
		0.949827,
		0.966392,
		0.978957,
		0.987949,
		0.993901,
		0.997428,
		0.999193,
		0.999857,
		0.999995,
		1,
		0.999961,
		0.999624,
		0.998488,
		0.995927,
		0.991269,
		0.983869,
		0.97315,
		0.958631,
		0.939942,
		0.916829,
		0.889157,
		0.856905,
		0.820162,
		0.779121,
		0.73407,
		0.685391,
		0.63355,
		0.579092,
		0.522638,
		0.464877,
		0.406567,
		0.348527,
		0.291635,
		0.236827,
		0.185091,
		0.137468,
		0.095049,
		0.058969,
		0.030412,
		0.010606,
		0.00082,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		9.3e-005,
		0.007282,
		0.024968,
		0.051842,
		0.086636,
		0.128128,
		0.175145,
		0.226559,
		0.281295,
		0.33833,
		0.396697,
		0.455485,
		0.513845,
		0.570991,
		0.626207,
		0.678846,
		0.72834,
		0.774201,
		0.816028,
		0.853514,
		0.886452,
		0.914741,
		0.938392,
		0.957534,
		0.972419,
		0.983418,
		0.991018,
		0.995807,
		0.998443,
		0.999612,
		0.99996,
		1,
		0.999995,
		0.999852,
		0.999169,
		0.997351,
		0.993724,
		0.987608,
		0.978376,
		0.965488,
		0.948514,
		0.927141,
		0.90118,
		0.870561,
		0.835329,
		0.79564,
		0.751752,
		0.704019,
		0.652886,
		0.598882,
		0.542616,
		0.484769,
		0.426093,
		0.367405,
		0.309587,
		0.253577,
		0.200371,
		0.151018,
		0.106619,
		0.068326,
		0.037337,
		0.014896,
		0.002292,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		0.000846,
		0.01094,
		0.031349,
		0.060743,
		0.097836,
		0.141391,
		0.190218,
		0.243179,
		0.299189,
		0.357216,
		0.416289,
		0.475496,
		0.533987,
		0.590984,
		0.645775,
		0.697728,
		0.746289,
		0.790993,
		0.831466,
		0.867431,
		0.89872,
		0.925275,
		0.947158,
		0.964555,
		0.977776,
		0.987255,
		0.993541,
		0.997272,
		0.999143,
		0.999848,
		0.999994,
		1,
		0.999959,
		0.9996,
		0.998396,
		0.995683,
		0.990759,
		0.982951,
		0.971663,
		0.956401,
		0.936792,
		0.912588,
		0.883666,
		0.850025,
		0.811778,
		0.769149,
		0.722464,
		0.672143,
		0.618696,
		0.562719,
		0.504881,
		0.44593,
		0.386681,
		0.328012,
		0.270869,
		0.216251,
		0.165218,
		0.118882,
		0.078406,
		0.045004,
		0.019939,
		0.004519,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.002366,
		0.015366,
		0.038489,
		0.070384,
		0.109748,
		0.155325,
		0.205913,
		0.260362,
		0.317576,
		0.376519,
		0.436213,
		0.495746,
		0.554272,
		0.611015,
		0.665277,
		0.716437,
		0.763962,
		0.80741,
		0.846434,
		0.880796,
		0.910368,
		0.935141,
		0.95523,
		0.970881,
		0.982468,
		0.99049,
		0.995554,
		0.998347,
		0.999588,
		0.999957,
		1,
		0.999994,
		0.999843,
		0.999117,
		0.997189,
		0.99335,
		0.986889,
		0.977154,
		0.96359,
		0.945759,
		0.92335,
		0.896184,
		0.864209,
		0.827492,
		0.78622,
		0.740684,
		0.69128,
		0.638497,
		0.582912,
		0.525185,
		0.466054,
		0.406331,
		0.346894,
		0.28869,
		0.232725,
		0.180066,
		0.131836,
		0.089212,
		0.053423,
		0.025748,
		0.007514,
		9.6e-005,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		0.004665,
		0.02057,
		0.046396,
		0.08077,
		0.122371,
		0.169927,
		0.222223,
		0.278096,
		0.336442,
		0.396218,
		0.456443,
		0.516206,
		0.574663,
		0.631048,
		0.684672,
		0.734932,
		0.781314,
		0.823405,
		0.86089,
		0.893571,
		0.921364,
		0.944313,
		0.962592,
		0.976511,
		0.98651,
		0.993153,
		0.997104,
		0.999089,
		0.999838,
		0.999994,
		1,
		0.999956,
		0.999575,
		0.998296,
		0.995421,
		0.990212,
		0.981968,
		0.970072,
		0.95402,
		0.933436,
		0.908079,
		0.877839,
		0.842738,
		0.802918,
		0.758636,
		0.710256,
		0.658245,
		0.60316,
		0.545648,
		0.486437,
		0.426333,
		0.366214,
		0.307026,
		0.249783,
		0.195557,
		0.145482,
		0.100747,
		0.062598,
		0.03233,
		0.01129,
		0.000874,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.0001,
		0.007758,
		0.026564,
		0.055077,
		0.091904,
		0.135705,
		0.185194,
		0.239139,
		0.296367,
		0.355767,
		0.416289,
		0.476952,
		0.536843,
		0.595126,
		0.651042,
		0.703917,
		0.753166,
		0.7983,
		0.838934,
		0.874792,
		0.905717,
		0.931675,
		0.952769,
		0.969235,
		0.98145,
		0.989923,
		0.995282,
		0.998243,
		0.999561,
		0.999955,
		1,
		0.999994,
		0.999833,
		0.999061,
		0.997015,
		0.992949,
		0.986118,
		0.975845,
		0.96156,
		0.942819,
		0.919314,
		0.890876,
		0.857473,
		0.8192,
		0.776274,
		0.729027,
		0.677898,
		0.623424,
		0.566235,
		0.507048,
		0.446662,
		0.385951,
		0.325863,
		0.267414,
		0.211684,
		0.159817,
		0.113014,
		0.072536,
		0.039695,
		0.015859,
		0.002443,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.000903,
		0.011657,
		0.033357,
		0.064539,
		0.103789,
		0.14975,
		0.201118,
		0.25665,
		0.315161,
		0.375532,
		0.436709,
		0.497709,
		0.557624,
		0.615622,
		0.670956,
		0.722968,
		0.771094,
		0.814873,
		0.853952,
		0.888097,
		0.917197,
		0.941275,
		0.960493,
		0.975156,
		0.985711,
		0.992736,
		0.996923,
		0.999031,
		0.999827,
		0.999994,
		1,
		0.999953,
		0.999547,
		0.998188,
		0.995138,
		0.989624,
		0.980913,
		0.968369,
		0.951476,
		0.929857,
		0.90328,
		0.871652,
		0.835018,
		0.793551,
		0.747548,
		0.697414,
		0.643665,
		0.586911,
		0.527856,
		0.46729,
		0.406083,
		0.345181,
		0.285604,
		0.228439,
		0.174837,
		0.126014,
		0.083242,
		0.047852,
		0.021232,
		0.004818,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.002525,
		0.016376,
		0.040959,
		0.074788,
		0.116428,
		0.164503,
		0.217694,
		0.274744,
		0.33446,
		0.395714,
		0.45745,
		0.518684,
		0.578511,
		0.636109,
		0.690745,
		0.741777,
		0.788668,
		0.830985,
		0.868415,
		0.900765,
		0.927979,
		0.950139,
		0.967473,
		0.980357,
		0.989313,
		0.994989,
		0.998131,
		0.999532,
		0.999952,
		1,
		0.999993,
		0.999822,
		0.999,
		0.996827,
		0.992516,
		0.985289,
		0.974442,
		0.959389,
		0.93968,
		0.915012,
		0.885232,
		0.850326,
		0.810421,
		0.76577,
		0.716747,
		0.66384,
		0.607637,
		0.548826,
		0.488188,
		0.426585,
		0.364963,
		0.304341,
		0.245813,
		0.190539,
		0.139745,
		0.094719,
		0.056809,
		0.02742,
		0.008014,
		0.000103,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.004979,
		0.021925,
		0.049379,
		0.085828,
		0.12982,
		0.179959,
		0.234912,
		0.293408,
		0.354245,
		0.416289,
		0.478481,
		0.53984,
		0.599465,
		0.656545,
		0.710363,
		0.760297,
		0.805839,
		0.846589,
		0.882275,
		0.912756,
		0.938031,
		0.958246,
		0.973702,
		0.984851,
		0.992288,
		0.996728,
		0.998968,
		0.999816,
		0.999993,
		1,
		0.99995,
		0.999517,
		0.998072,
		0.994834,
		0.988991,
		0.979781,
		0.966545,
		0.948756,
		0.926039,
		0.89817,
		0.865077,
		0.826833,
		0.783645,
		0.73585,
		0.683903,
		0.628371,
		0.569922,
		0.509323,
		0.44743,
		0.385184,
		0.323607,
		0.263794,
		0.206915,
		0.154206,
		0.10697,
		0.066571,
		0.034434,
		0.012042,
		0.000934,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.000107,
		0.008282,
		0.028318,
		0.058623,
		0.097664,
		0.143965,
		0.196114,
		0.252761,
		0.312626,
		0.374495,
		0.43723,
		0.499772,
		0.561141,
		0.620445,
		0.676885,
		0.729761,
		0.778478,
		0.822556,
		0.861636,
		0.895491,
		0.924033,
		0.947325,
		0.965583,
		0.979184,
		0.988657,
		0.994673,
		0.99801,
		0.999501,
		0.999948,
		1,
		0.999993,
		0.99981,
		0.998934,
		0.996625,
		0.99205,
		0.984397,
		0.972936,
		0.957064,
		0.936326,
		0.910426,
		0.879226,
		0.842738,
		0.801123,
		0.754672,
		0.703808,
		0.649069,
		0.591103,
		0.530661,
		0.468589,
		0.405822,
		0.343382,
		0.282369,
		0.223957,
		0.169394,
		0.119997,
		0.077145,
		0.042284,
		0.016918,
		0.00261,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.000966,
		0.012447,
		0.035564,
		0.0687,
		0.110298,
		0.158861,
		0.212959,
		0.27123,
		0.332378,
		0.395185,
		0.458508,
		0.521286,
		0.582546,
		0.641405,
		0.697079,
		0.748889,
		0.796268,
		0.83877,
		0.876079,
		0.908019,
		0.934562,
		0.955839,
		0.972142,
		0.983926,
		0.991804,
		0.996517,
		0.9989,
		0.999803,
		0.999993,
		1,
		0.999947,
		0.999485,
		0.997946,
		0.994505,
		0.98831,
		0.978564,
		0.964588,
		0.945845,
		0.921961,
		0.892725,
		0.858088,
		0.818151,
		0.773163,
		0.723506,
		0.669685,
		0.612327,
		0.552163,
		0.490028,
		0.42685,
		0.363647,
		0.301522,
		0.241656,
		0.185305,
		0.133799,
		0.088536,
		0.050979,
		0.022654,
		0.005149,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.0027,
		0.017488,
		0.043674,
		0.079615,
		0.12373,
		0.174504,
		0.230487,
		0.290302,
		0.352645,
		0.416289,
		0.48009,
		0.542987,
		0.604014,
		0.6623,
		0.717079,
		0.767695,
		0.813614,
		0.854428,
		0.889869,
		0.919818,
		0.944313,
		0.963556,
		0.977922,
		0.987949,
		0.994331,
		0.997879,
		0.999468,
		0.999945,
		1,
		0.999993,
		0.999797,
		0.998863,
		0.996405,
		0.991548,
		0.983438,
		0.971319,
		0.954571,
		0.932739,
		0.905532,
		0.872831,
		0.834679,
		0.79127,
		0.742944,
		0.690172,
		0.63355,
		0.573791,
		0.511714,
		0.448238,
		0.384378,
		0.321236,
		0.259999,
		0.201932,
		0.148377,
		0.100747,
		0.060526,
		0.029261,
		0.008564,
		0.00011,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.005327,
		0.023419,
		0.052658,
		0.091374,
		0.137962,
		0.190888,
		0.248685,
		0.309961,
		0.373403,
		0.437779,
		0.501941,
		0.564834,
		0.625499,
		0.68308,
		0.736831,
		0.786125,
		0.830462,
		0.869479,
		0.902962,
		0.930852,
		0.953258,
		0.970465,
		0.982931,
		0.991281,
		0.996289,
		0.998826,
		0.99979,
		0.999992,
		1,
		0.999943,
		0.99945,
		0.997809,
		0.99415,
		0.987575,
		0.977255,
		0.962487,
		0.942727,
		0.917603,
		0.88692,
		0.850653,
		0.808939,
		0.762069,
		0.710476,
		0.654724,
		0.595501,
		0.533608,
		0.469955,
		0.405548,
		0.34149,
		0.278972,
		0.219266,
		0.163726,
		0.113781,
		0.070933,
		0.036751,
		0.012873,
		0.000999,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.000114,
		0.008861,
		0.030251,
		0.062522,
		0.103978,
		0.152992,
		0.208006,
		0.26754,
		0.330188,
		0.394628,
		0.459622,
		0.524024,
		0.586783,
		0.646952,
		0.703692,
		0.75628,
		0.804122,
		0.846758,
		0.883873,
		0.915312,
		0.941085,
		0.961379,
		0.976563,
		0.987186,
		0.993962,
		0.997737,
		0.999431,
		0.999941,
		1,
		0.999992,
		0.999782,
		0.998786,
		0.996168,
		0.991005,
		0.982404,
		0.96958,
		0.951898,
		0.928899,
		0.900305,
		0.866019,
		0.826114,
		0.780826,
		0.730545,
		0.675798,
		0.617247,
		0.555671,
		0.491965,
		0.427129,
		0.362263,
		0.298559,
		0.237298,
		0.179843,
		0.127639,
		0.082205,
		0.045134,
		0.018088,
		0.002795,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.001035,
		0.01332,
		0.037997,
		0.073275,
		0.117431,
		0.168815,
		0.225849,
		0.287037,
		0.35096,
		0.416289,
		0.481783,
		0.546297,
		0.608789,
		0.668323,
		0.724081,
		0.77537,
		0.82163,
		0.862446,
		0.897559,
		0.926878,
		0.950488,
		0.968661,
		0.981857,
		0.990717,
		0.996042,
		0.998745,
		0.999775,
		0.999992,
		1,
		0.999939,
		0.999412,
		0.997662,
		0.993766,
		0.986782,
		0.975845,
		0.96023,
		0.939384,
		0.912942,
		0.880725,
		0.842738,
		0.799158,
		0.750323,
		0.69672,
		0.638979,
		0.577857,
		0.51423,
		0.449089,
		0.383529,
		0.318742,
		0.256014,
		0.196721,
		0.14232,
		0.094348,
		0.054421,
		0.024223,
		0.005514,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.002895,
		0.018718,
		0.046668,
		0.084923,
		0.131734,
		0.185427,
		0.244406,
		0.307157,
		0.372254,
		0.438357,
		0.504225,
		0.568718,
		0.630801,
		0.689557,
		0.744191,
		0.794042,
		0.838591,
		0.877473,
		0.91049,
		0.937623,
		0.959038,
		0.9751,
		0.986362,
		0.993563,
		0.997583,
		0.999391,
		0.999937,
		1,
		0.999991,
		0.999767,
		0.998703,
		0.995911,
		0.990418,
		0.981289,
		0.967708,
		0.949027,
		0.924786,
		0.89472,
		0.858756,
		0.817006,
		0.769751,
		0.717434,
		0.660647,
		0.60012,
		0.536711,
		0.471394,
		0.405259,
		0.339498,
		0.275401,
		0.214352,
		0.157821,
		0.107365,
		0.064618,
		0.031293,
		0.009174,
		0.000118,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.005711,
		0.02507,
		0.056273,
		0.09747,
		0.146884,
		0.20282,
		0.263662,
		0.327881,
		0.39404,
		0.460796,
		0.526907,
		0.591237,
		0.652767,
		0.710598,
		0.763962,
		0.812236,
		0.854945,
		0.891784,
		0.92262,
		0.947512,
		0.966719,
		0.980698,
		0.990107,
		0.995774,
		0.998658,
		0.999759,
		0.999991,
		1,
		0.999934,
		0.99937,
		0.997501,
		0.99335,
		0.985926,
		0.974325,
		0.957802,
		0.935798,
		0.907953,
		0.874111,
		0.834309,
		0.788769,
		0.737881,
		0.682196,
		0.622411,
		0.559361,
		0.494005,
		0.427424,
		0.360804,
		0.295441,
		0.232725,
		0.17414,
		0.121258,
		0.075735,
		0.039308,
		0.013791,
		0.001072,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.000123,
		0.009504,
		0.032389,
		0.066821,
		0.110919,
		0.16288,
		0.220985,
		0.283601,
		0.349184,
		0.416289,
		0.483569,
		0.549783,
		0.613805,
		0.674631,
		0.731385,
		0.783333,
		0.82989,
		0.870636,
		0.905327,
		0.933906,
		0.95652,
		0.973521,
		0.985471,
		0.99313,
		0.997416,
		0.999348,
		0.999932,
		1,
		0.999991,
		0.99975,
		0.998612,
		0.995632,
		0.989783,
		0.980084,
		0.965692,
		0.945942,
		0.920376,
		0.888746,
		0.851009,
		0.807315,
		0.758,
		0.703566,
		0.644676,
		0.582134,
		0.516882,
		0.449987,
		0.382633,
		0.316114,
		0.251827,
		0.191269,
		0.136026,
		0.087777,
		0.048281,
		0.019382,
		0.003,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.001112,
		0.014288,
		0.040688,
		0.07832,
		0.125272,
		0.179716,
		0.23991,
		0.304203,
		0.371041,
		0.438967,
		0.506634,
		0.572806,
		0.636368,
		0.696335,
		0.751857,
		0.802239,
		0.846941,
		0.885604,
		0.918052,
		0.944313,
		0.964625,
		0.979446,
		0.989445,
		0.995483,
		0.998563,
		0.999741,
		0.99999,
		1,
		0.999929,
		0.999325,
		0.997327,
		0.992899,
		0.984999,
		0.972685,
		0.955189,
		0.931946,
		0.902608,
		0.867043,
		0.825327,
		0.777728,
		0.724699,
		0.666857,
		0.604978,
		0.539979,
		0.472913,
		0.404954,
		0.337397,
		0.271644,
		0.2092,
		0.151669,
		0.100747,
		0.058222,
		0.025962,
		0.00592,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.003111,
		0.020082,
		0.049979,
		0.090775,
		0.140529,
		0.197385,
		0.259581,
		0.325448,
		0.39342,
		0.462036,
		0.529947,
		0.595924,
		0.658868,
		0.717815,
		0.771949,
		0.820615,
		0.863328,
		0.899793,
		0.929913,
		0.953807,
		0.971816,
		0.984507,
		0.99266,
		0.997234,
		0.999301,
		0.999927,
		1,
		0.99999,
		0.999732,
		0.998512,
		0.995329,
		0.989094,
		0.978782,
		0.963517,
		0.942622,
		0.915644,
		0.882352,
		0.842738,
		0.797,
		0.745529,
		0.688896,
		0.627839,
		0.563249,
		0.496159,
		0.427734,
		0.359265,
		0.292156,
		0.227922,
		0.168182,
		0.114651,
		0.069138,
		0.033544,
		0.009851,
		0.000127,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.00614,
		0.026902,
		0.060273,
		0.104192,
		0.156687,
		0.215878,
		0.27998,
		0.34731,
		0.416289,
		0.485454,
		0.553458,
		0.619082,
		0.681244,
		0.739008,
		0.791594,
		0.838394,
		0.878987,
		0.913148,
		0.940869,
		0.962366,
		0.978092,
		0.988728,
		0.995167,
		0.99846,
		0.999722,
		0.99999,
		1,
		0.999924,
		0.999276,
		0.997137,
		0.99241,
		0.983995,
		0.970913,
		0.952372,
		0.927805,
		0.896877,
		0.859486,
		0.815748,
		0.76599,
		0.710728,
		0.650658,
		0.586639,
		0.51968,
		0.450936,
		0.381687,
		0.313342,
		0.247423,
		0.185559,
		0.129487,
		0.081039,
		0.042142,
		0.014812,
		0.001154,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.000132,
		0.010218,
		0.034761,
		0.071577,
		0.118572,
		0.173742,
		0.235182,
		0.301087,
		0.369759,
		0.439612,
		0.509177,
		0.577116,
		0.642221,
		0.703432,
		0.759844,
		0.810721,
		0.855511,
		0.893857,
		0.925619,
		0.950882,
		0.969973,
		0.983462,
		0.992149,
		0.997036,
		0.999249,
		0.999921,
		1,
		0.999989,
		0.999712,
		0.998404,
		0.994999,
		0.988347,
		0.977373,
		0.961169,
		0.939048,
		0.910561,
		0.875503,
		0.833904,
		0.786014,
		0.732288,
		0.673374,
		0.610092,
		0.543429,
		0.474517,
		0.404632,
		0.335179,
		0.267685,
		0.203793,
		0.145256,
		0.09393,
		0.051768,
		0.020821,
		0.003228,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.001198,
		0.015366,
		0.043674,
		0.083901,
		0.133916,
		0.191686,
		0.255282,
		0.322879,
		0.392764,
		0.463346,
		0.533157,
		0.600863,
		0.665277,
		0.725363,
		0.780253,
		0.829262,
		0.871897,
		0.907879,
		0.937158,
		0.959925,
		0.976625,
		0.987949,
		0.994823,
		0.998347,
		0.999701,
		0.999989,
		1,
		0.999918,
		0.999222,
		0.99693,
		0.991877,
		0.982906,
		0.968996,
		0.949334,
		0.92335,
		0.890728,
		0.851398,
		0.805528,
		0.753505,
		0.695919,
		0.63355,
		0.567351,
		0.498434,
		0.428063,
		0.357638,
		0.28869,
		0.222873,
		0.161955,
		0.107814,
		0.062433,
		0.027894,
		0.006372,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.003352,
		0.021601,
		0.053654,
		0.09725,
		0.150223,
		0.210511,
		0.27616,
		0.345327,
		0.416289,
		0.487448,
		0.557338,
		0.62464,
		0.688183,
		0.746966,
		0.800163,
		0.847143,
		0.887485,
		0.920996,
		0.947725,
		0.967979,
		0.982327,
		0.991594,
		0.99682,
		0.999193,
		0.999915,
		1,
		0.999988,
		0.99969,
		0.998286,
		0.99464,
		0.987535,
		0.975845,
		0.958631,
		0.935195,
		0.905097,
		0.868161,
		0.824461,
		0.774307,
		0.718226,
		0.656948,
		0.591389,
		0.522638,
		0.451939,
		0.380687,
		0.310415,
		0.242783,
		0.179577,
		0.122695,
		0.074147,
		0.036046,
		0.010606,
		0.000137,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.006618,
		0.028941,
		0.064712,
		0.111627,
		0.167488,
		0.230203,
		0.297795,
		0.368403,
		0.440294,
		0.511868,
		0.581665,
		0.64838,
		0.710869,
		0.768167,
		0.819497,
		0.864293,
		0.902212,
		0.933156,
		0.957286,
		0.975034,
		0.987103,
		0.994448,
		0.998223,
		0.999678,
		0.999988,
		1,
		0.999912,
		0.999163,
		0.996704,
		0.991297,
		0.981724,
		0.96692,
		0.946052,
		0.918552,
		0.884124,
		0.842738,
		0.794618,
		0.74022,
		0.680218,
		0.615483,
		0.547074,
		0.476215,
		0.404291,
		0.332833,
		0.263508,
		0.198115,
		0.138573,
		0.086916,
		0.045292,
		0.015951,
		0.001245,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.000143,
		0.011016,
		0.037403,
		0.076857,
		0.127035,
		0.185705,
		0.250747,
		0.32016,
		0.39207,
		0.464734,
		0.536552,
		0.606074,
		0.672015,
		0.73326,
		0.788888,
		0.838179,
		0.880639,
		0.916014,
		0.944313,
		0.965817,
		0.981094,
		0.990988,
		0.996584,
		0.999131,
		0.999909,
		1,
		0.999988,
		0.999666,
		0.998157,
		0.994248,
		0.986651,
		0.974188,
		0.955885,
		0.931037,
		0.899218,
		0.860284,
		0.814362,
		0.761826,
		0.703287,
		0.639566,
		0.571684,
		0.500843,
		0.428411,
		0.355915,
		0.285028,
		0.217558,
		0.155444,
		0.100747,
		0.055645,
		0.022426,
		0.003484,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.001294,
		0.01657,
		0.047001,
		0.090096,
		0.143473,
		0.204865,
		0.272124,
		0.343228,
		0.416289,
		0.489559,
		0.561441,
		0.630499,
		0.69547,
		0.755278,
		0.809049,
		0.856131,
		0.896111,
		0.928835,
		0.954428,
		0.973306,
		0.986181,
		0.994038,
		0.998088,
		0.999653,
		0.999987,
		1,
		0.999905,
		0.999098,
		0.996457,
		0.990665,
		0.980438,
		0.964669,
		0.942504,
		0.913379,
		0.877025,
		0.833458,
		0.782964,
		0.726079,
		0.663568,
		0.596407,
		0.525768,
		0.453002,
		0.379627,
		0.307317,
		0.237891,
		0.173305,
		0.115644,
		0.067117,
		0.030049,
		0.006878,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.003623,
		0.023299,
		0.057749,
		0.104436,
		0.160938,
		0.224955,
		0.294311,
		0.366965,
		0.441017,
		0.514717,
		0.586474,
		0.654869,
		0.718669,
		0.776841,
		0.828569,
		0.873276,
		0.910641,
		0.940621,
		0.963472,
		0.979753,
		0.990328,
		0.996325,
		0.999064,
		0.999901,
		1,
		0.999987,
		0.999639,
		0.998016,
		0.993819,
		0.985689,
		0.972387,
		0.95291,
		0.926546,
		0.892886,
		0.851827,
		0.803552,
		0.748515,
		0.687413,
		0.621173,
		0.550931,
		0.478015,
		0.40393,
		0.330348,
		0.259096,
		0.192146,
		0.131607,
		0.079717,
		0.038838,
		0.011451,
		0.000148,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.007154,
		0.031222,
		0.069658,
		0.119879,
		0.179424,
		0.245958,
		0.317281,
		0.391333,
		0.466205,
		0.540147,
		0.61158,
		0.679106,
		0.741528,
		0.797864,
		0.847365,
		0.889537,
		0.924165,
		0.95133,
		0.971428,
		0.985175,
		0.99359,
		0.99794,
		0.999625,
		0.999986,
		1,
		0.999897,
		0.999027,
		0.996187,
		0.989975,
		0.979038,
		0.962225,
		0.938663,
		0.907796,
		0.869387,
		0.823506,
		0.770509,
		0.711022,
		0.645911,
		0.576269,
		0.503396,
		0.42878,
		0.354089,
		0.281153,
		0.211958,
		0.148635,
		0.093452,
		0.048808,
		0.017226,
		0.001347,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.000155,
		0.011912,
		0.040357,
		0.082738,
		0.136427,
		0.198921,
		0.267854,
		0.341002,
		0.416289,
		0.4918,
		0.565786,
		0.636686,
		0.70313,
		0.763962,
		0.81826,
		0.865352,
		0.904839,
		0.936624,
		0.960924,
		0.978292,
		0.989606,
		0.996042,
		0.998989,
		0.999893,
		1,
		0.999985,
		0.99961,
		0.99786,
		0.99335,
		0.984638,
		0.970427,
		0.949683,
		0.921688,
		0.886059,
		0.842738,
		0.791976,
		0.734311,
		0.670542,
		0.601712,
		0.529086,
		0.45413,
		0.378502,
		0.304035,
		0.232725,
		0.166724,
		0.108329,
		0.059973,
		0.024223,
		0.003771,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.001403,
		0.017921,
		0.050721,
		0.096998,
		0.154076,
		0.219416,
		0.290619,
		0.365439,
		0.441786,
		0.51774,
		0.591564,
		0.661713,
		0.726854,
		0.785882,
		0.837941,
		0.882447,
		0.919112,
		0.947967,
		0.969383,
		0.984077,
		0.993099,
		0.997777,
		0.999595,
		0.999985,
		1,
		0.999889,
		0.998949,
		0.99589,
		0.98922,
		0.977512,
		0.959568,
		0.934501,
		0.901765,
		0.861163,
		0.812825,
		0.757191,
		0.694984,
		0.627186,
		0.55502,
		0.479925,
		0.403546,
		0.327712,
		0.254429,
		0.185867,
		0.124349,
		0.072345,
		0.032464,
		0.007447,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.003928,
		0.025204,
		0.062328,
		0.112441,
		0.172822,
		0.240892,
		0.314225,
		0.39055,
		0.467769,
		0.543963,
		0.617404,
		0.686577,
		0.750188,
		0.807193,
		0.856815,
		0.898568,
		0.932289,
		0.958153,
		0.976697,
		0.988816,
		0.995731,
		0.998907,
		0.999884,
		1,
		0.999984,
		0.999578,
		0.997689,
		0.992836,
		0.98349,
		0.968292,
		0.946178,
		0.91643,
		0.878693,
		0.832964,
		0.779572,
		0.719149,
		0.65261,
		0.581127,
		0.506107,
		0.429172,
		0.352149,
		0.277047,
		0.206051,
		0.141513,
		0.085934,
		0.041967,
		0.012401,
		0.000161,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.007758,
		0.033782,
		0.07519,
		0.12907,
		0.192655,
		0.263329,
		0.338636,
		0.416289,
		0.49418,
		0.570395,
		0.643226,
		0.711187,
		0.773038,
		0.827802,
		0.874792,
		0.913638,
		0.944313,
		0.967154,
		0.982876,
		0.992561,
		0.997598,
		0.999561,
		0.999984,
		1,
		0.99988,
		0.998863,
		0.995564,
		0.988393,
		0.975845,
		0.956676,
		0.929985,
		0.895243,
		0.8523,
		0.801356,
		0.742944,
		0.677898,
		0.607332,
		0.532609,
		0.455329,
		0.377306,
		0.300551,
		0.227264,
		0.159817,
		0.100747,
		0.052748,
		0.01866,
		0.001462,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.000168,
		0.012921,
		0.043674,
		0.089317,
		0.146884,
		0.213563,
		0.286701,
		0.363815,
		0.442603,
		0.520954,
		0.596961,
		0.66894,
		0.735449,
		0.795304,
		0.84761,
		0.891784,
		0.927583,
		0.955135,
		0.974954,
		0.987949,
		0.995389,
		0.998817,
		0.999874,
		1,
		0.999983,
		0.999543,
		0.997501,
		0.992272,
		0.982234,
		0.965963,
		0.942367,
		0.910731,
		0.870737,
		0.822445,
		0.766271,
		0.702959,
		0.63355,
		0.559361,
		0.481957,
		0.403137,
		0.324909,
		0.249484,
		0.179254,
		0.11679,
		0.064824,
		0.026246,
		0.004095,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.001526,
		0.019444,
		0.054899,
		0.104717,
		0.165879,
		0.235528,
		0.310976,
		0.389717,
		0.469434,
		0.548017,
		0.623575,
		0.694455,
		0.759263,
		0.816885,
		0.86652,
		0.907703,
		0.940336,
		0.96472,
		0.981561,
		0.991969,
		0.9974,
		0.999524,
		0.999982,
		1,
		0.999869,
		0.998768,
		0.995205,
		0.987485,
		0.974023,
		0.953525,
		0.925079,
		0.888183,
		0.842738,
		0.78903,
		0.727694,
		0.659694,
		0.586284,
		0.508991,
		0.42959,
		0.350084,
		0.272688,
		0.199814,
		0.134062,
		0.078205,
		0.035182,
		0.008089,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.004273,
		0.027353,
		0.067471,
		0.121393,
		0.186045,
		0.258527,
		0.336118,
		0.416289,
		0.496716,
		0.575292,
		0.650148,
		0.71967,
		0.782523,
		0.837676,
		0.884434,
		0.922467,
		0.951842,
		0.973047,
		0.986998,
		0.995012,
		0.998717,
		0.999863,
		1,
		0.999981,
		0.999504,
		0.997294,
		0.991651,
		0.980856,
		0.963419,
		0.938217,
		0.904547,
		0.862134,
		0.811114,
		0.752002,
		0.685666,
		0.613292,
		0.536358,
		0.456608,
		0.376031,
		0.296847,
		0.221482,
		0.152563,
		0.092903,
		0.045488,
		0.013475,
		0.000176,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.008442,
		0.03667,
		0.081404,
		0.139346,
		0.207371,
		0.282534,
		0.362085,
		0.443475,
		0.524376,
		0.602692,
		0.676582,
		0.74448,
		0.805122,
		0.857572,
		0.901258,
		0.936004,
		0.962058,
		0.980117,
		0.991318,
		0.997182,
		0.999482,
		0.999981,
		1,
		0.999858,
		0.998663,
		0.994809,
		0.986488,
		0.972026,
		0.950084,
		0.919743,
		0.880531,
		0.832415,
		0.775775,
		0.711367,
		0.640295,
		0.563978,
		0.484124,
		0.402701,
		0.321925,
		0.244237,
		0.172287,
		0.108924,
		0.057184,
		0.02028,
		0.001593,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.000183,
		0.014064,
		0.047417,
		0.096707,
		0.158573,
		0.229837,
		0.307516,
		0.388827,
		0.47121,
		0.552335,
		0.630124,
		0.702773,
		0.768774,
		0.826947,
		0.876466,
		0.9169,
		0.948246,
		0.970957,
		0.985952,
		0.994596,
		0.998606,
		0.999851,
		1,
		0.99998,
		0.99946,
		0.997064,
		0.990967,
		0.979343,
		0.960634,
		0.933692,
		0.897829,
		0.852824,
		0.798899,
		0.736686,
		0.667192,
		0.591766,
		0.512066,
		0.430036,
		0.347883,
		0.268054,
		0.193222,
		0.126271,
		0.070284,
		0.028532,
		0.004462,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.001665,
		0.02117,
		0.059613,
		0.113386,
		0.179066,
		0.253422,
		0.333432,
		0.416289,
		0.499421,
		0.580504,
		0.657487,
		0.72861,
		0.792437,
		0.847883,
		0.894253,
		0.931276,
		0.959144,
		0.978531,
		0.990599,
		0.99694,
		0.999437,
		0.999979,
		1,
		0.999845,
		0.998546,
		0.994372,
		0.985389,
		0.969836,
		0.946324,
		0.913932,
		0.87223,
		0.82126,
		0.761512,
		0.693878,
		0.619625,
		0.540354,
		0.457972,
		0.374671,
		0.292901,
		0.215353,
		0.144944,
		0.084802,
		0.038254,
		0.008818,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.004665,
		0.029788,
		0.073275,
		0.131445,
		0.200812,
		0.278096,
		0.360237,
		0.444407,
		0.528027,
		0.608789,
		0.684672,
		0.753975,
		0.815347,
		0.867815,
		0.910833,
		0.944313,
		0.968661,
		0.984799,
		0.994136,
		0.998483,
		0.999838,
		1,
		0.999978,
		0.999412,
		0.99681,
		0.990212,
		0.977679,
		0.957583,
		0.928752,
		0.890522,
		0.842738,
		0.785723,
		0.720239,
		0.647454,
		0.568899,
		0.486437,
		0.402235,
		0.318742,
		0.238661,
		0.164939,
		0.100747,
		0.049471,
		0.014693,
		0.000192,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.00922,
		0.039943,
		0.088416,
		0.150882,
		0.223793,
		0.303823,
		0.387877,
		0.473108,
		0.556941,
		0.637084,
		0.711563,
		0.778747,
		0.837381,
		0.886629,
		0.926112,
		0.955947,
		0.976785,
		0.989805,
		0.996672,
		0.999385,
		0.999977,
		1,
		0.99983,
		0.998417,
		0.993888,
		0.984178,
		0.967429,
		0.942207,
		0.907596,
		0.863213,
		0.809195,
		0.746155,
		0.675142,
		0.597605,
		0.51535,
		0.430513,
		0.345531,
		0.263117,
		0.186245,
		0.118126,
		0.0622,
		0.022121,
		0.001742,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.000201,
		0.015366,
		0.05166,
		0.105044,
		0.171689,
		0.247986,
		0.330561,
		0.416289,
		0.502313,
		0.586062,
		0.665277,
		0.738037,
		0.802795,
		0.858415,
		0.904213,
		0.940003,
		0.966137,
		0.983525,
		0.993626,
		0.998347,
		0.999823,
		1,
		0.999976,
		0.999358,
		0.996527,
		0.989376,
		0.975845,
		0.954233,
		0.92335,
		0.882566,
		0.831801,
		0.771497,
		0.70257,
		0.626365,
		0.544621,
		0.459432,
		0.373216,
		0.28869,
		0.208847,
		0.13694,
		0.076461,
		0.031129,
		0.004882,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.001825,
		0.023136,
		0.064959,
		0.123166,
		0.193856,
		0.273359,
		0.358259,
		0.445405,
		0.531932,
		0.615286,
		0.693246,
		0.763962,
		0.825987,
		0.878322,
		0.920461,
		0.952436,
		0.974858,
		0.988925,
		0.996374,
		0.999328,
		0.999974,
		1,
		0.999814,
		0.998273,
		0.99335,
		0.982839,
		0.96478,
		0.937694,
		0.900677,
		0.85341,
		0.796135,
		0.72961,
		0.655065,
		0.574153,
		0.488913,
		0.401737,
		0.315337,
		0.232725,
		0.157186,
		0.092264,
		0.041747,
		0.00965,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.005114,
		0.032563,
		0.079857,
		0.142782,
		0.217363,
		0.299874,
		0.386858,
		0.475143,
		0.561865,
		0.644493,
		0.720862,
		0.789202,
		0.848188,
		0.89698,
		0.935275,
		0.963355,
		0.982117,
		0.99306,
		0.998194,
		0.999806,
		1,
		0.999973,
		0.999297,
		0.996213,
		0.98845,
		0.973821,
		0.950551,
		0.917436,
		0.87389,
		0.819929,
		0.756129,
		0.683581,
		0.603835,
		0.518865,
		0.431024,
		0.343012,
		0.25785,
		0.178855,
		0.10962,
		0.053998,
		0.016085,
		0.000211,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.010111,
		0.043674,
		0.096367,
		0.163887,
		0.242186,
		0.327485,
		0.416289,
		0.505413,
		0.592002,
		0.673559,
		0.747985,
		0.813614,
		0.869259,
		0.914268,
		0.948572,
		0.97273,
		0.987949,
		0.996042,
		0.999265,
		0.999972,
		1,
		0.999797,
		0.998111,
		0.992753,
		0.981357,
		0.961859,
		0.932739,
		0.893112,
		0.842738,
		0.781985,
		0.711778,
		0.63355,
		0.549189,
		0.460998,
		0.371655,
		0.284186,
		0.201932,
		0.128533,
		0.067904,
		0.024223,
		0.001913,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.000221,
		0.016857,
		0.056497,
		0.114498,
		0.186469,
		0.268292,
		0.356136,
		0.446476,
		0.536117,
		0.622222,
		0.702346,
		0.77447,
		0.837049,
		0.889065,
		0.930079,
		0.960286,
		0.980556,
		0.992429,
		0.998024,
		0.999787,
		1,
		0.999971,
		0.99923,
		0.995862,
		0.987421,
		0.971582,
		0.946495,
		0.91095,
		0.864419,
		0.807029,
		0.739516,
		0.663169,
		0.579774,
		0.491569,
		0.401201,
		0.311689,
		0.226396,
		0.149001,
		0.083483,
		0.034098,
		0.005363,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.002008,
		0.025389,
		0.071053,
		0.134253,
		0.210511,
		0.295641,
		0.385763,
		0.477327,
		0.56714,
		0.652395,
		0.730708,
		0.800163,
		0.859359,
		0.907473,
		0.944313,
		0.970373,
		0.986864,
		0.995671,
		0.999193,
		0.999969,
		1,
		0.999776,
		0.997931,
		0.992087,
		0.979712,
		0.958631,
		0.927288,
		0.884829,
		0.831109,
		0.766643,
		0.692551,
		0.610497,
		0.522638,
		0.431573,
		0.340308,
		0.252217,
		0.171018,
		0.100747,
		0.045739,
		0.010606,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.005631,
		0.035743,
		0.08736,
		0.155629,
		0.235987,
		0.324181,
		0.416289,
		0.508744,
		0.598361,
		0.682378,
		0.75849,
		0.824905,
		0.880394,
		0.924359,
		0.956891,
		0.978822,
		0.991726,
		0.997833,
		0.999765,
		1,
		0.999968,
		0.999154,
		0.99547,
		0.986276,
		0.9691,
		0.94202,
		0.903828,
		0.854068,
		0.793,
		0.721547,
		0.641225,
		0.55409,
		0.462682,
		0.369978,
		0.279358,
		0.19457,
		0.119705,
		0.059174,
		0.017685,
		0.000232,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.011138,
		0.047952,
		0.105431,
		0.178617,
		0.262862,
		0.353853,
		0.447628,
		0.540612,
		0.629643,
		0.712016,
		0.785527,
		0.848532,
		0.900005,
		0.939609,
		0.967759,
		0.985655,
		0.995256,
		0.999113,
		0.999966,
		1,
		0.999754,
		0.997729,
		0.991344,
		0.977883,
		0.95506,
		0.921283,
		0.875748,
		0.818422,
		0.749996,
		0.671812,
		0.585802,
		0.494427,
		0.400625,
		0.30777,
		0.219635,
		0.140358,
		0.074425,
		0.02664,
		0.002111,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.000244,
		0.018575,
		0.062044,
		0.125272,
		0.203198,
		0.291093,
		0.384583,
		0.47968,
		0.572806,
		0.660837,
		0.741144,
		0.811648,
		0.870879,
		0.918052,
		0.953131,
		0.976892,
		0.990939,
		0.997618,
		0.999741,
		1,
		0.999964,
		0.999069,
		0.99503,
		0.984999,
		0.966345,
		0.937074,
		0.895994,
		0.842738,
		0.777728,
		0.7021,
		0.617635,
		0.526696,
		0.432165,
		0.337397,
		0.246181,
		0.162702,
		0.091512,
		0.037511,
		0.00592,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.002221,
		0.027987,
		0.078042,
		0.146884,
		0.229348,
		0.320624,
		0.416289,
		0.512331,
		0.605186,
		0.691782,
		0.769588,
		0.836674,
		0.891784,
		0.934406,
		0.964853,
		0.984305,
		0.99479,
		0.999022,
		0.999962,
		1,
		0.999728,
		0.997501,
		0.990511,
		0.975845,
		0.951098,
		0.914656,
		0.865776,
		0.804566,
		0.731919,
		0.649438,
		0.559361,
		0.464497,
		0.368169,
		0.27417,
		0.186723,
		0.110446,
		0.05033,
		0.011711,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.006231,
		0.039412,
		0.095965,
		0.170259,
		0.257029,
		0.35139,
		0.448872,
		0.545454,
		0.637598,
		0.722303,
		0.797161,
		0.860425,
		0.911086,
		0.948956,
		0.974739,
		0.990058,
		0.997377,
		0.999714,
		0.999999,
		0.99996,
		0.998972,
		0.994536,
		0.98357,
		0.963278,
		0.931598,
		0.887363,
		0.830324,
		0.761088,
		0.681045,
		0.592281,
		0.497509,
		0.400003,
		0.30355,
		0.212399,
		0.131232,
		0.065127,
		0.019535,
		0.000257,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.012329,
		0.052888,
		0.115823,
		0.195381,
		0.286194,
		0.383309,
		0.482221,
		0.578906,
		0.669872,
		0.752213,
		0.823672,
		0.882719,
		0.928639,
		0.961616,
		0.982792,
		0.994266,
		0.99892,
		0.999958,
		0.999999,
		0.999699,
		0.997245,
		0.989577,
		0.973569,
		0.946697,
		0.907331,
		0.854812,
		0.789417,
		0.712278,
		0.6253,
		0.531073,
		0.432804,
		0.334256,
		0.2397,
		0.153869,
		0.081927,
		0.029437,
		0.002341,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.000272,
		0.02057,
		0.068445,
		0.137622,
		0.222223,
		0.316784,
		0.416289,
		0.516206,
		0.612528,
		0.701826,
		0.781314,
		0.848922,
		0.903379,
		0.944313,
		0.972331,
		0.989067,
		0.997104,
		0.999683,
		0.999999,
		0.999956,
		0.998863,
		0.99398,
		0.981968,
		0.959858,
		0.925522,
		0.877839,
		0.816702,
		0.742944,
		0.658245,
		0.565045,
		0.466459,
		0.366214,
		0.268583,
		0.178346,
		0.100747,
		0.041461,
		0.006567,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.00247,
		0.031004,
		0.086107,
		0.161357,
		0.250747,
		0.348725,
		0.450219,
		0.550683,
		0.646144,
		0.73326,
		0.809397,
		0.872709,
		0.922235,
		0.958001,
		0.981094,
		0.993675,
		0.998803,
		0.999954,
		0.999999,
		0.999666,
		0.996954,
		0.988525,
		0.971021,
		0.941795,
		0.899218,
		0.842738,
		0.772836,
		0.690927,
		0.599263,
		0.500843,
		0.39933,
		0.298992,
		0.204642,
		0.121599,
		0.055645,
		0.012998,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.006932,
		0.043674,
		0.105894,
		0.187012,
		0.280903,
		0.381929,
		0.484973,
		0.58549,
		0.679561,
		0.763962,
		0.836245,
		0.894835,
		0.939136,
		0.969633,
		0.987949,
		0.996795,
		0.999648,
		0.999999,
		0.999951,
		0.998739,
		0.99335,
		0.980168,
		0.956035,
		0.918768,
		0.867313,
		0.801739,
		0.723143,
		0.63355,
		0.535808,
		0.433497,
		0.330857,
		0.232725,
		0.144484,
		0.072021,
		0.02169,
		0.000287,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.013723,
		0.058623,
		0.127813,
		0.214561,
		0.312626,
		0.416289,
		0.520404,
		0.620445,
		0.71257,
		0.793707,
		0.861636,
		0.915109,
		0.953955,
		0.979184,
		0.993005,
		0.998671,
		0.999948,
		0.999999,
		0.999628,
		0.996625,
		0.987338,
		0.968161,
		0.936326,
		0.890217,
		0.829424,
		0.754672,
		0.667708,
		0.571191,
		0.468589,
		0.364093,
		0.262549,
		0.169394,
		0.090614,
		0.032698,
		0.00261,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.000304,
		0.022905,
		0.075884,
		0.151865,
		0.243966,
		0.345833,
		0.451682,
		0.556348,
		0.655346,
		0.744944,
		0.822257,
		0.885348,
		0.933353,
		0.9666,
		0.986687,
		0.996443,
		0.999607,
		0.999999,
		0.999945,
		0.998598,
		0.992637,
		0.978139,
		0.951752,
		0.911245,
		0.855661,
		0.785281,
		0.701521,
		0.606807,
		0.50446,
		0.398598,
		0.294054,
		0.19631,
		0.111441,
		0.046069,
		0.007327,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.002763,
		0.034535,
		0.095481,
		0.178037,
		0.275172,
		0.380428,
		0.487964,
		0.592618,
		0.689972,
		0.776439,
		0.849368,
		0.907163,
		0.949417,
		0.977027,
		0.992245,
		0.99852,
		0.999942,
		0.999999,
		0.999585,
		0.996249,
		0.985994,
		0.964943,
		0.930208,
		0.880212,
		0.814721,
		0.734755,
		0.642452,
		0.540946,
		0.43425,
		0.327165,
		0.2252,
		0.13451,
		0.061846,
		0.014509,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.007758,
		0.048665,
		0.117431,
		0.206304,
		0.308108,
		0.416289,
		0.524968,
		0.629005,
		0.724081,
		0.806797,
		0.874792,
		0.926878,
		0.963183,
		0.985256,
		0.996042,
		0.999561,
		0.999999,
		0.999939,
		0.998436,
		0.991827,
		0.975845,
		0.94694,
		0.902848,
		0.842738,
		0.767158,
		0.677898,
		0.577857,
		0.470906,
		0.361785,
		0.256014,
		0.159817,
		0.080064,
		0.024223,
		0.000322,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.015366,
		0.065341,
		0.14174,
		0.236626,
		0.342684,
		0.453277,
		0.562504,
		0.665277,
		0.757413,
		0.835752,
		0.898284,
		0.944313,
		0.974587,
		0.991379,
		0.998347,
		0.999935,
		0.999999,
		0.999535,
		0.99582,
		0.984468,
		0.961312,
		0.92335,
		0.86907,
		0.798462,
		0.712898,
		0.614981,
		0.5084,
		0.397801,
		0.28869,
		0.187344,
		0.100747,
		0.036531,
		0.002929,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.000342,
		0.025659,
		0.084596,
		0.168399,
		0.268945,
		0.37879,
		0.491227,
		0.600358,
		0.701178,
		0.789689,
		0.863025,
		0.91961,
		0.959323,
		0.983628,
		0.995583,
		0.999508,
		0.999999,
		0.999931,
		0.998251,
		0.990902,
		0.973246,
		0.941522,
		0.893455,
		0.828385,
		0.74718,
		0.652082,
		0.54654,
		0.435073,
		0.323142,
		0.217061,
		0.123912,
		0.051487,
		0.008227,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.003111,
		0.038704,
		0.106458,
		0.197385,
		0.303183,
		0.416289,
		0.529947,
		0.638287,
		0.73643,
		0.820615,
		0.888342,
		0.938558,
		0.971816,
		0.99039,
		0.998148,
		0.999927,
		0.999999,
		0.999478,
		0.995329,
		0.98273,
		0.957205,
		0.915644,
		0.856637,
		0.780455,
		0.688896,
		0.585111,
		0.473438,
		0.359265,
		0.248916,
		0.149561,
		0.069138,
		0.0163,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.008741,
		0.05456,
		0.130939,
		0.228658,
		0.339242,
		0.455022,
		0.569217,
		0.676022,
		0.770731,
		0.849883,
		0.911434,
		0.954949,
		0.981771,
		0.995056,
		0.999446,
		0.999999,
		0.999922,
		0.998037,
		0.989843,
		0.970291,
		0.935405,
		0.882926,
		0.812415,
		0.725136,
		0.623865,
		0.512705,
		0.396928,
		0.282841,
		0.177679,
		0.089522,
		0.027226,
		0.000363,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.017323,
		0.073275,
		0.158037,
		0.262156,
		0.376995,
		0.494799,
		0.608789,
		0.713267,
		0.803759,
		0.877185,
		0.932051,
		0.968661,
		0.989256,
		0.997917,
		0.999917,
		0.999999,
		0.999412,
		0.994763,
		0.980744,
		0.952544,
		0.906963,
		0.842738,
		0.760489,
		0.662527,
		0.552654,
		0.435974,
		0.318742,
		0.208236,
		0.112664,
		0.041077,
		0.00331,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.000387,
		0.028941,
		0.094888,
		0.187731,
		0.297794,
		0.416289,
		0.535398,
		0.648379,
		0.749695,
		0.835175,
		0.90221,
		0.949977,
		0.979644,
		0.994447,
		0.999373,
		0.999997,
		0.999911,
		0.997788,
		0.988624,
		0.966918,
		0.928478,
		0.871094,
		0.794617,
		0.70079,
		0.59303,
		0.476214,
		0.3565,
		0.241181,
		0.138572,
		0.057915,
		0.009304,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.003529,
		0.043674,
		0.11942,
		0.21998,
		0.335457,
		0.456932,
		0.576554,
		0.687664,
		0.784947,
		0.864619,
		0.924657,
		0.965039,
		0.987932,
		0.997633,
		0.999888,
		0.999981,
		0.999317,
		0.994091,
		0.978449,
		0.947222,
		0.897143,
		0.827155,
		0.738309,
		0.633539,
		0.51742,
		0.395962,
		0.276435,
		0.167239,
		0.077791,
		0.018444,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.009922,
		0.061582,
		0.146868,
		0.254699,
		0.374977,
		0.498671,
		0.617935,
		0.726248,
		0.818594,
		0.891682,
		0.944206,
		0.977092,
		0.993631,
		0.999178,
		0.999885,
		0.999787,
		0.997388,
		0.98711,
		0.962953,
		0.920512,
		0.857677,
		0.774666,
		0.673812,
		0.559298,
		0.436917,
		0.313873,
		0.198621,
		0.100736,
		0.03082,
		0.000414,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.019668,
		0.082699,
		0.177175,
		0.291735,
		0.416091,
		0.541136,
		0.659076,
		0.763598,
		0.85008,
		0.915849,
		0.960466,
		0.98597,
		0.996864,
		0.999416,
		0.999522,
		0.998768,
		0.992877,
		0.97538,
		0.940734,
		0.885637,
		0.80931,
		0.713346,
		0.601425,
		0.479047,
		0.353287,
		0.232614,
		0.126741,
		0.046503,
		0.003769,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.000442,
		0.032845,
		0.106998,
		0.21019,
		0.330792,
		0.458357,
		0.583747,
		0.699281,
		0.798942,
		0.878618,
		0.936403,
		0.9729,
		0.991411,
		0.997669,
		0.998473,
		0.998361,
		0.995644,
		0.984094,
		0.957169,
		0.910271,
		0.841453,
		0.751401,
		0.643164,
		0.52184,
		0.394307,
		0.268996,
		0.155717,
		0.065521,
		0.01059,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.004022,
		0.049465,
		0.134336,
		0.245582,
		0.371341,
		0.501051,
		0.625599,
		0.737504,
		0.831152,
		0.903084,
		0.952336,
		0.980739,
		0.992978,
		0.995868,
		0.995988,
		0.995131,
		0.988486,
		0.968915,
		0.930499,
		0.869955,
		0.786881,
		0.68353,
		0.564478,
		0.436307,
		0.30734,
		0.187433,
		0.087814,
		0.020955,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.011259,
		0.069425,
		0.164372,
		0.282748,
		0.412507,
		0.543038,
		0.66534,
		0.772234,
		0.858647,
		0.921945,
		0.962301,
		0.982956,
		0.99,
		0.990912,
		0.990785,
		0.987715,
		0.974769,
		0.944861,
		0.893244,
		0.818227,
		0.721083,
		0.605712,
		0.478279,
		0.346903,
		0.221416,
		0.113164,
		0.034862,
		0.000471,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.022135,
		0.092422,
		0.19647,
		0.320699,
		0.452958,
		0.582681,
		0.701088,
		0.801442,
		0.879373,
		0.93327,
		0.964651,
		0.978299,
		0.98155,
		0.981686,
		0.980718,
		0.973275,
		0.95155,
		0.909308,
		0.84338,
		0.753855,
		0.643772,
		0.51873,
		0.386524,
		0.256856,
		0.141106,
		0.052157,
		0.004255,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.000494,
		0.036449,
		0.11786,
		0.229604,
		0.357979,
		0.490845,
		0.617806,
		0.730447,
		0.822636,
		0.890903,
		0.934871,
		0.957597,
		0.96542,
		0.966441,
		0.966297,
		0.962873,
		0.948562,
		0.915816,
		0.859861,
		0.779382,
		0.676358,
		0.555639,
		0.424544,
		0.29251,
		0.170831,
		0.07245,
		0.011792,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.0044,
		0.053735,
		0.144773,
		0.262302,
		0.392648,
		0.523831,
		0.645788,
		0.750642,
		0.83306,
		0.890684,
		0.924584,
		0.939475,
		0.943055,
		0.943206,
		0.942136,
		0.933976,
		0.910389,
		0.864989,
		0.794878,
		0.700751,
		0.58651,
		0.458811,
		0.326664,
		0.201119,
		0.095026,
		0.022848,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.011958,
		0.073179,
		0.171783,
		0.292662,
		0.422357,
		0.549241,
		0.663773,
		0.758816,
		0.830044,
		0.876427,
		0.900656,
		0.909081,
		0.910191,
		0.910034,
		0.906329,
		0.890993,
		0.856264,
		0.797554,
		0.714073,
		0.608558,
		0.486799,
		0.35718,
		0.23032,
		0.118789,
		0.036891,
		0.000502,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.0226,
		0.093595,
		0.197141,
		0.318468,
		0.444555,
		0.564348,
		0.669042,
		0.75245,
		0.811468,
		0.84658,
		0.86217,
		0.865957,
		0.866118,
		0.864981,
		0.856393,
		0.83183,
		0.785067,
		0.713679,
		0.619035,
		0.505845,
		0.381656,
		0.256439,
		0.142261,
		0.053042,
		0.004361,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.000483,
		0.0354,
		0.113465,
		0.218859,
		0.337415,
		0.456801,
		0.566765,
		0.659459,
		0.729854,
		0.77625,
		0.800764,
		0.809382,
		0.810529,
		0.810366,
		0.806556,
		0.790953,
		0.756013,
		0.697635,
		0.615657,
		0.513512,
		0.397697,
		0.277308,
		0.163672,
		0.070064,
		0.011499,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.004055,
		0.049109,
		0.131066,
		0.234937,
		0.347429,
		0.45715,
		0.55488,
		0.633929,
		0.690604,
		0.724736,
		0.74007,
		0.743835,
		0.743996,
		0.742861,
		0.734367,
		0.710347,
		0.665163,
		0.597047,
		0.507996,
		0.403267,
		0.290865,
		0.181135,
		0.086449,
		0.020971,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.010267,
		0.062266,
		0.144681,
		0.243645,
		0.347001,
		0.444525,
		0.528227,
		0.592741,
		0.635826,
		0.658875,
		0.667076,
		0.668179,
		0.668022,
		0.664376,
		0.649616,
		0.616965,
		0.563102,
		0.488506,
		0.397048,
		0.295462,
		0.192897,
		0.100576,
		0.031536,
		0.000432,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.017889,
		0.073371,
		0.152846,
		0.243819,
		0.33549,
		0.418991,
		0.487707,
		0.537699,
		0.568211,
		0.582095,
		0.585544,
		0.585694,
		0.584648,
		0.57691,
		0.555299,
		0.515172,
		0.455512,
		0.378733,
		0.29017,
		0.197612,
		0.110928,
		0.04179,
		0.003467,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.000352,
		0.025572,
		0.08111,
		0.154586,
		0.235077,
		0.313298,
		0.381856,
		0.435589,
		0.472002,
		0.491749,
		0.498867,
		0.499835,
		0.499696,
		0.496513,
		0.483784,
		0.455999,
		0.4108,
		0.349164,
		0.274981,
		0.194576,
		0.116324,
		0.050357,
		0.008346,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.002672,
		0.032043,
		0.084551,
		0.149594,
		0.217932,
		0.281881,
		0.335565,
		0.375266,
		0.399855,
		0.411201,
		0.414057,
		0.414182,
		0.413311,
		0.406945,
		0.389406,
		0.3573,
		0.310294,
		0.250868,
		0.18387,
		0.116117,
		0.056098,
		0.013754,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.00605,
		0.036295,
		0.083299,
		0.138292,
		0.193752,
		0.24359,
		0.283404,
		0.310826,
		0.325921,
		0.33144,
		0.3322,
		0.33209,
		0.329606,
		0.319805,
		0.298721,
		0.26495,
		0.219692,
		0.166381,
		0.110298,
		0.058278,
		0.018485,
		0.000256,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.009308,
		0.037732,
		0.077548,
		0.121789,
		0.164591,
		0.201372,
		0.229082,
		0.246526,
		0.254699,
		0.256786,
		0.256879,
		0.256238,
		0.251619,
		0.239079,
		0.216483,
		0.183964,
		0.143689,
		0.099509,
		0.056675,
		0.021621,
		0.001814,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.000161,
		0.011571,
		0.036236,
		0.068049,
		0.101724,
		0.132916,
		0.158393,
		0.176263,
		0.186264,
		0.189978,
		0.190497,
		0.190422,
		0.188738,
		0.182192,
		0.168337,
		0.146521,
		0.11786,
		0.084949,
		0.0516,
		0.022647,
		0.003798,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.001043,
		0.012361,
		0.032163,
		0.055989,
		0.080042,
		0.101297,
		0.117656,
		0.128143,
		0.13314,
		0.134436,
		0.134495,
		0.134094,
		0.131247,
		0.123644,
		0.110179,
		0.091174,
		0.068196,
		0.043827,
		0.021495,
		0.005339,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001962,
		0.011614,
		0.02625,
		0.042807,
		0.058738,
		0.072094,
		0.081659,
		0.087112,
		0.089172,
		0.089464,
		0.089421,
		0.08848,
		0.08488,
		0.077394,
		0.065832,
		0.050988,
		0.03446,
		0.018512,
		0.005956,
		8.4e-005,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		0.002411,
		0.009635,
		0.019471,
		0.02998,
		0.039592,
		0.047176,
		0.052138,
		0.054548,
		0.055184,
		0.055213,
		0.055015,
		0.05363,
		0.049997,
		0.043688,
		0.03498,
		0.024749,
		0.014355,
		0.005563,
		0.000473,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		3.2e-005,
		0.002263,
		0.006974,
		0.012854,
		0.018798,
		0.023941,
		0.027714,
		0.02991,
		0.030755,
		0.030877,
		0.030859,
		0.030469,
		0.029006,
		0.026022,
		0.021513,
		0.015878,
		0.009842,
		0.004394,
		0.000748,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		0.000143,
		0.00167,
		0.004271,
		0.007282,
		0.01016,
		0.012498,
		0.014064,
		0.014841,
		0.01505,
		0.015059,
		0.014994,
		0.014543,
		0.013384,
		0.011415,
		0.008767,
		0.005762,
		0.00288,
		0.000727,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000161,
		0.000938,
		0.002079,
		0.003312,
		0.004423,
		0.005261,
		0.00576,
		0.005956,
		0.005985,
		0.005981,
		0.005889,
		0.005553,
		0.004882,
		0.003894,
		0.002699,
		0.001481,
		0.000485,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		9.1e-005,
		0.000355,
		0.000703,
		0.001055,
		0.001351,
		0.001555,
		0.001659,
		0.001687,
		0.001688,
		0.001679,
		0.001619,
		0.001466,
		0.001213,
		0.000882,
		0.000524,
		0.000207,
		1.8e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		2.1e-005,
		6.4e-005,
		0.000116,
		0.000165,
		0.000203,
		0.000226,
		0.000235,
		0.000237,
		0.000237,
		0.000232,
		0.000216,
		0.000185,
		0.000141,
		9e-005,
		4.1e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		4.1e-005,
		9e-005,
		0.000141,
		0.000185,
		0.000216,
		0.000232,
		0.000237,
		0.000237,
		0.000235,
		0.000226,
		0.000203,
		0.000165,
		0.000116,
		6.4e-005,
		2.1e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		1.8e-005,
		0.000207,
		0.000524,
		0.000882,
		0.001213,
		0.001466,
		0.001619,
		0.001679,
		0.001688,
		0.001687,
		0.001659,
		0.001555,
		0.001351,
		0.001055,
		0.000703,
		0.000355,
		9.1e-005,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		0.000485,
		0.001481,
		0.002699,
		0.003894,
		0.004882,
		0.005553,
		0.005889,
		0.005981,
		0.005985,
		0.005956,
		0.00576,
		0.005261,
		0.004423,
		0.003312,
		0.002079,
		0.000938,
		0.000161,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000727,
		0.00288,
		0.005762,
		0.008767,
		0.011415,
		0.013384,
		0.014543,
		0.014994,
		0.015059,
		0.01505,
		0.014841,
		0.014064,
		0.012498,
		0.01016,
		0.007282,
		0.004271,
		0.00167,
		0.000143,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		0.000748,
		0.004394,
		0.009842,
		0.015878,
		0.021513,
		0.026022,
		0.029006,
		0.030469,
		0.030859,
		0.030877,
		0.030755,
		0.02991,
		0.027714,
		0.023941,
		0.018798,
		0.012854,
		0.006974,
		0.002263,
		3.2e-005,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.000473,
		0.005563,
		0.014355,
		0.024749,
		0.03498,
		0.043688,
		0.049997,
		0.05363,
		0.055015,
		0.055213,
		0.055184,
		0.054548,
		0.052138,
		0.047176,
		0.039592,
		0.02998,
		0.019471,
		0.009635,
		0.002411,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		8.4e-005,
		0.005956,
		0.018512,
		0.03446,
		0.050988,
		0.065832,
		0.077394,
		0.08488,
		0.08848,
		0.089421,
		0.089464,
		0.089172,
		0.087112,
		0.081659,
		0.072094,
		0.058738,
		0.042807,
		0.02625,
		0.011614,
		0.001962,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.005339,
		0.021495,
		0.043827,
		0.068196,
		0.091174,
		0.110179,
		0.123644,
		0.131247,
		0.134094,
		0.134495,
		0.134436,
		0.13314,
		0.128143,
		0.117656,
		0.101297,
		0.080042,
		0.055989,
		0.032163,
		0.012361,
		0.001043,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.003798,
		0.022647,
		0.0516,
		0.084949,
		0.11786,
		0.146521,
		0.168337,
		0.182192,
		0.188738,
		0.190422,
		0.190497,
		0.189978,
		0.186264,
		0.176263,
		0.158393,
		0.132916,
		0.101724,
		0.068049,
		0.036236,
		0.011571,
		0.000161,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.001814,
		0.021621,
		0.056675,
		0.099509,
		0.143689,
		0.183964,
		0.216483,
		0.239079,
		0.251619,
		0.256238,
		0.256879,
		0.256786,
		0.254699,
		0.246526,
		0.229082,
		0.201372,
		0.164591,
		0.121789,
		0.077548,
		0.037732,
		0.009308,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.000256,
		0.018485,
		0.058278,
		0.110298,
		0.166381,
		0.219692,
		0.26495,
		0.298721,
		0.319805,
		0.329606,
		0.33209,
		0.3322,
		0.33144,
		0.325921,
		0.310826,
		0.283404,
		0.24359,
		0.193752,
		0.138292,
		0.083299,
		0.036295,
		0.00605,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.013754,
		0.056098,
		0.116117,
		0.18387,
		0.250868,
		0.310294,
		0.3573,
		0.389406,
		0.406945,
		0.413311,
		0.414182,
		0.414057,
		0.411201,
		0.399855,
		0.375266,
		0.335565,
		0.281881,
		0.217932,
		0.149594,
		0.084551,
		0.032043,
		0.002672,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.008346,
		0.050357,
		0.116324,
		0.194576,
		0.274981,
		0.349164,
		0.4108,
		0.455999,
		0.483784,
		0.496513,
		0.499696,
		0.499835,
		0.498867,
		0.491749,
		0.472002,
		0.435589,
		0.381856,
		0.313298,
		0.235077,
		0.154586,
		0.08111,
		0.025572,
		0.000352,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.003467,
		0.04179,
		0.110928,
		0.197612,
		0.29017,
		0.378733,
		0.455512,
		0.515172,
		0.555299,
		0.57691,
		0.584648,
		0.585694,
		0.585544,
		0.582095,
		0.568211,
		0.537699,
		0.487707,
		0.418991,
		0.33549,
		0.243819,
		0.152846,
		0.073371,
		0.017889,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.000432,
		0.031536,
		0.100576,
		0.192897,
		0.295462,
		0.397048,
		0.488506,
		0.563102,
		0.616965,
		0.649616,
		0.664376,
		0.668022,
		0.668179,
		0.667076,
		0.658875,
		0.635826,
		0.592741,
		0.528227,
		0.444525,
		0.347001,
		0.243645,
		0.144681,
		0.062266,
		0.010267,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.020971,
		0.086449,
		0.181135,
		0.290865,
		0.403267,
		0.507996,
		0.597047,
		0.665163,
		0.710347,
		0.734367,
		0.742861,
		0.743996,
		0.743835,
		0.74007,
		0.724736,
		0.690604,
		0.633929,
		0.55488,
		0.45715,
		0.347429,
		0.234937,
		0.131066,
		0.049109,
		0.004055,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.011499,
		0.070064,
		0.163672,
		0.277308,
		0.397697,
		0.513512,
		0.615657,
		0.697635,
		0.756013,
		0.790953,
		0.806556,
		0.810366,
		0.810529,
		0.809382,
		0.800764,
		0.77625,
		0.729854,
		0.659459,
		0.566765,
		0.456801,
		0.337415,
		0.218859,
		0.113465,
		0.0354,
		0.000483,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.004361,
		0.053042,
		0.142261,
		0.256439,
		0.381656,
		0.505845,
		0.619035,
		0.713679,
		0.785067,
		0.83183,
		0.856393,
		0.864981,
		0.866118,
		0.865957,
		0.86217,
		0.84658,
		0.811468,
		0.75245,
		0.669042,
		0.564348,
		0.444555,
		0.318468,
		0.197141,
		0.093595,
		0.0226,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.000502,
		0.036891,
		0.118789,
		0.23032,
		0.35718,
		0.486799,
		0.608558,
		0.714073,
		0.797554,
		0.856264,
		0.890993,
		0.906329,
		0.910034,
		0.910191,
		0.909081,
		0.900656,
		0.876427,
		0.830044,
		0.758816,
		0.663773,
		0.549241,
		0.422357,
		0.292662,
		0.171783,
		0.073179,
		0.011958,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.022848,
		0.095026,
		0.201119,
		0.326664,
		0.458811,
		0.58651,
		0.700751,
		0.794878,
		0.864989,
		0.910389,
		0.933976,
		0.942136,
		0.943206,
		0.943055,
		0.939475,
		0.924584,
		0.890684,
		0.83306,
		0.750642,
		0.645788,
		0.523831,
		0.392648,
		0.262302,
		0.144773,
		0.053735,
		0.0044,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.011792,
		0.07245,
		0.170831,
		0.29251,
		0.424544,
		0.555639,
		0.676358,
		0.779382,
		0.859861,
		0.915816,
		0.948562,
		0.962873,
		0.966297,
		0.966441,
		0.96542,
		0.957597,
		0.934871,
		0.890903,
		0.822636,
		0.730447,
		0.617806,
		0.490845,
		0.357979,
		0.229604,
		0.11786,
		0.036449,
		0.000494,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.004255,
		0.052157,
		0.141106,
		0.256856,
		0.386524,
		0.51873,
		0.643772,
		0.753855,
		0.84338,
		0.909308,
		0.95155,
		0.973275,
		0.980718,
		0.981686,
		0.98155,
		0.978299,
		0.964651,
		0.93327,
		0.879373,
		0.801442,
		0.701088,
		0.582681,
		0.452958,
		0.320699,
		0.19647,
		0.092422,
		0.022135,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.000471,
		0.034862,
		0.113164,
		0.221416,
		0.346903,
		0.478279,
		0.605712,
		0.721083,
		0.818227,
		0.893244,
		0.944861,
		0.974769,
		0.987715,
		0.990785,
		0.990912,
		0.99,
		0.982956,
		0.962301,
		0.921945,
		0.858647,
		0.772234,
		0.66534,
		0.543038,
		0.412507,
		0.282748,
		0.164372,
		0.069425,
		0.011259,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.020955,
		0.087814,
		0.187433,
		0.30734,
		0.436307,
		0.564478,
		0.68353,
		0.786881,
		0.869955,
		0.930499,
		0.968915,
		0.988486,
		0.995131,
		0.995988,
		0.995868,
		0.992978,
		0.980739,
		0.952336,
		0.903084,
		0.831152,
		0.737504,
		0.625599,
		0.501051,
		0.371341,
		0.245582,
		0.134336,
		0.049465,
		0.004022,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.01059,
		0.065521,
		0.155717,
		0.268996,
		0.394307,
		0.52184,
		0.643164,
		0.751401,
		0.841453,
		0.910271,
		0.957169,
		0.984094,
		0.995644,
		0.998361,
		0.998473,
		0.997669,
		0.991411,
		0.9729,
		0.936403,
		0.878618,
		0.798942,
		0.699281,
		0.583747,
		0.458357,
		0.330792,
		0.21019,
		0.106998,
		0.032845,
		0.000442,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.003769,
		0.046503,
		0.126741,
		0.232614,
		0.353287,
		0.479047,
		0.601425,
		0.713346,
		0.80931,
		0.885637,
		0.940734,
		0.97538,
		0.992877,
		0.998768,
		0.999522,
		0.999416,
		0.996864,
		0.98597,
		0.960466,
		0.915849,
		0.85008,
		0.763598,
		0.659076,
		0.541136,
		0.416091,
		0.291735,
		0.177175,
		0.082699,
		0.019668,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.000414,
		0.03082,
		0.100736,
		0.198621,
		0.313873,
		0.436917,
		0.559298,
		0.673812,
		0.774666,
		0.857677,
		0.920512,
		0.962953,
		0.98711,
		0.997388,
		0.999787,
		0.999885,
		0.999178,
		0.993631,
		0.977092,
		0.944206,
		0.891682,
		0.818594,
		0.726248,
		0.617935,
		0.498671,
		0.374977,
		0.254699,
		0.146868,
		0.061582,
		0.009922,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.018444,
		0.077791,
		0.167239,
		0.276435,
		0.395962,
		0.51742,
		0.633539,
		0.738309,
		0.827155,
		0.897143,
		0.947222,
		0.978449,
		0.994091,
		0.999317,
		0.999981,
		0.999888,
		0.997633,
		0.987932,
		0.965039,
		0.924657,
		0.864619,
		0.784947,
		0.687664,
		0.576554,
		0.456932,
		0.335457,
		0.21998,
		0.11942,
		0.043674,
		0.003529,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.009304,
		0.057915,
		0.138572,
		0.241181,
		0.3565,
		0.476214,
		0.59303,
		0.70079,
		0.794617,
		0.871094,
		0.928478,
		0.966918,
		0.988624,
		0.997788,
		0.999911,
		0.999997,
		0.999373,
		0.994447,
		0.979644,
		0.949977,
		0.90221,
		0.835175,
		0.749695,
		0.648379,
		0.535398,
		0.416289,
		0.297794,
		0.187731,
		0.094888,
		0.028941,
		0.000387,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.00331,
		0.041077,
		0.112664,
		0.208236,
		0.318742,
		0.435974,
		0.552654,
		0.662527,
		0.760489,
		0.842738,
		0.906963,
		0.952544,
		0.980744,
		0.994763,
		0.999412,
		0.999999,
		0.999917,
		0.997917,
		0.989256,
		0.968661,
		0.932051,
		0.877185,
		0.803759,
		0.713267,
		0.608789,
		0.494799,
		0.376995,
		0.262156,
		0.158037,
		0.073275,
		0.017323,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.000363,
		0.027226,
		0.089522,
		0.177679,
		0.282841,
		0.396928,
		0.512705,
		0.623865,
		0.725136,
		0.812415,
		0.882926,
		0.935405,
		0.970291,
		0.989843,
		0.998037,
		0.999922,
		0.999999,
		0.999446,
		0.995056,
		0.981771,
		0.954949,
		0.911434,
		0.849883,
		0.770731,
		0.676022,
		0.569217,
		0.455022,
		0.339242,
		0.228658,
		0.130939,
		0.05456,
		0.008741,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.0163,
		0.069138,
		0.149561,
		0.248916,
		0.359265,
		0.473438,
		0.585111,
		0.688896,
		0.780455,
		0.856637,
		0.915644,
		0.957205,
		0.98273,
		0.995329,
		0.999478,
		0.999999,
		0.999927,
		0.998148,
		0.99039,
		0.971816,
		0.938558,
		0.888342,
		0.820615,
		0.73643,
		0.638287,
		0.529947,
		0.416289,
		0.303183,
		0.197385,
		0.106458,
		0.038704,
		0.003111,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.008227,
		0.051487,
		0.123912,
		0.217061,
		0.323142,
		0.435073,
		0.54654,
		0.652082,
		0.74718,
		0.828385,
		0.893455,
		0.941522,
		0.973246,
		0.990902,
		0.998251,
		0.999931,
		0.999999,
		0.999508,
		0.995583,
		0.983628,
		0.959323,
		0.91961,
		0.863025,
		0.789689,
		0.701178,
		0.600358,
		0.491227,
		0.37879,
		0.268945,
		0.168399,
		0.084596,
		0.025659,
		0.000342,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.002929,
		0.036531,
		0.100747,
		0.187344,
		0.28869,
		0.397801,
		0.5084,
		0.614981,
		0.712898,
		0.798462,
		0.86907,
		0.92335,
		0.961312,
		0.984468,
		0.99582,
		0.999535,
		0.999999,
		0.999935,
		0.998347,
		0.991379,
		0.974587,
		0.944313,
		0.898284,
		0.835752,
		0.757413,
		0.665277,
		0.562504,
		0.453277,
		0.342684,
		0.236626,
		0.14174,
		0.065341,
		0.015366,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.000322,
		0.024223,
		0.080064,
		0.159817,
		0.256014,
		0.361785,
		0.470906,
		0.577857,
		0.677898,
		0.767158,
		0.842738,
		0.902848,
		0.94694,
		0.975845,
		0.991827,
		0.998436,
		0.999939,
		0.999999,
		0.999561,
		0.996042,
		0.985256,
		0.963183,
		0.926878,
		0.874792,
		0.806797,
		0.724081,
		0.629005,
		0.524968,
		0.416289,
		0.308108,
		0.206304,
		0.117431,
		0.048665,
		0.007758,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.014509,
		0.061846,
		0.13451,
		0.2252,
		0.327165,
		0.43425,
		0.540946,
		0.642452,
		0.734755,
		0.814721,
		0.880212,
		0.930208,
		0.964943,
		0.985994,
		0.996249,
		0.999585,
		0.999999,
		0.999942,
		0.99852,
		0.992245,
		0.977027,
		0.949417,
		0.907163,
		0.849368,
		0.776439,
		0.689972,
		0.592618,
		0.487964,
		0.380428,
		0.275172,
		0.178037,
		0.095481,
		0.034535,
		0.002763,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.007327,
		0.046069,
		0.111441,
		0.19631,
		0.294054,
		0.398598,
		0.50446,
		0.606807,
		0.701521,
		0.785281,
		0.855661,
		0.911245,
		0.951752,
		0.978139,
		0.992637,
		0.998598,
		0.999945,
		0.999999,
		0.999607,
		0.996443,
		0.986687,
		0.9666,
		0.933353,
		0.885348,
		0.822257,
		0.744944,
		0.655346,
		0.556348,
		0.451682,
		0.345833,
		0.243966,
		0.151865,
		0.075884,
		0.022905,
		0.000304,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.00261,
		0.032698,
		0.090614,
		0.169394,
		0.262549,
		0.364093,
		0.468589,
		0.571191,
		0.667708,
		0.754672,
		0.829424,
		0.890217,
		0.936326,
		0.968161,
		0.987338,
		0.996625,
		0.999628,
		0.999999,
		0.999948,
		0.998671,
		0.993005,
		0.979184,
		0.953955,
		0.915109,
		0.861636,
		0.793707,
		0.71257,
		0.620445,
		0.520404,
		0.416289,
		0.312626,
		0.214561,
		0.127813,
		0.058623,
		0.013723,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.000287,
		0.02169,
		0.072021,
		0.144484,
		0.232725,
		0.330857,
		0.433497,
		0.535808,
		0.63355,
		0.723143,
		0.801739,
		0.867313,
		0.918768,
		0.956035,
		0.980168,
		0.99335,
		0.998739,
		0.999951,
		0.999999,
		0.999648,
		0.996795,
		0.987949,
		0.969633,
		0.939136,
		0.894835,
		0.836245,
		0.763962,
		0.679561,
		0.58549,
		0.484973,
		0.381929,
		0.280903,
		0.187012,
		0.105894,
		0.043674,
		0.006932,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.012998,
		0.055645,
		0.121599,
		0.204642,
		0.298992,
		0.39933,
		0.500843,
		0.599263,
		0.690927,
		0.772836,
		0.842738,
		0.899218,
		0.941795,
		0.971021,
		0.988525,
		0.996954,
		0.999666,
		0.999999,
		0.999954,
		0.998803,
		0.993675,
		0.981094,
		0.958001,
		0.922235,
		0.872709,
		0.809397,
		0.73326,
		0.646144,
		0.550683,
		0.450219,
		0.348725,
		0.250747,
		0.161357,
		0.086107,
		0.031004,
		0.00247,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.006567,
		0.041461,
		0.100747,
		0.178346,
		0.268583,
		0.366214,
		0.466459,
		0.565045,
		0.658245,
		0.742944,
		0.816702,
		0.877839,
		0.925522,
		0.959858,
		0.981968,
		0.99398,
		0.998863,
		0.999956,
		0.999999,
		0.999683,
		0.997104,
		0.989067,
		0.972331,
		0.944313,
		0.903379,
		0.848922,
		0.781314,
		0.701826,
		0.612528,
		0.516206,
		0.416289,
		0.316784,
		0.222223,
		0.137622,
		0.068445,
		0.02057,
		0.000272,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.002341,
		0.029437,
		0.081927,
		0.153869,
		0.2397,
		0.334256,
		0.432804,
		0.531073,
		0.6253,
		0.712278,
		0.789417,
		0.854812,
		0.907331,
		0.946697,
		0.973569,
		0.989577,
		0.997245,
		0.999699,
		0.999999,
		0.999958,
		0.99892,
		0.994266,
		0.982792,
		0.961616,
		0.928639,
		0.882719,
		0.823672,
		0.752213,
		0.669872,
		0.578906,
		0.482221,
		0.383309,
		0.286194,
		0.195381,
		0.115823,
		0.052888,
		0.012329,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.000257,
		0.019535,
		0.065127,
		0.131232,
		0.212399,
		0.30355,
		0.400003,
		0.497509,
		0.592281,
		0.681045,
		0.761088,
		0.830324,
		0.887363,
		0.931598,
		0.963278,
		0.98357,
		0.994536,
		0.998972,
		0.99996,
		0.999999,
		0.999714,
		0.997377,
		0.990058,
		0.974739,
		0.948956,
		0.911086,
		0.860425,
		0.797161,
		0.722303,
		0.637598,
		0.545454,
		0.448872,
		0.35139,
		0.257029,
		0.170259,
		0.095965,
		0.039412,
		0.006231,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.011711,
		0.05033,
		0.110446,
		0.186723,
		0.27417,
		0.368169,
		0.464497,
		0.559361,
		0.649438,
		0.731919,
		0.804566,
		0.865776,
		0.914656,
		0.951098,
		0.975845,
		0.990511,
		0.997501,
		0.999728,
		1,
		0.999962,
		0.999022,
		0.99479,
		0.984305,
		0.964853,
		0.934406,
		0.891784,
		0.836674,
		0.769588,
		0.691782,
		0.605186,
		0.512331,
		0.416289,
		0.320624,
		0.229348,
		0.146884,
		0.078042,
		0.027987,
		0.002221,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.00592,
		0.037511,
		0.091512,
		0.162702,
		0.246181,
		0.337397,
		0.432165,
		0.526696,
		0.617635,
		0.7021,
		0.777728,
		0.842738,
		0.895994,
		0.937074,
		0.966345,
		0.984999,
		0.99503,
		0.999069,
		0.999964,
		1,
		0.999741,
		0.997618,
		0.990939,
		0.976892,
		0.953131,
		0.918052,
		0.870879,
		0.811648,
		0.741144,
		0.660837,
		0.572806,
		0.47968,
		0.384583,
		0.291093,
		0.203198,
		0.125272,
		0.062044,
		0.018575,
		0.000244,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.002111,
		0.02664,
		0.074425,
		0.140358,
		0.219635,
		0.30777,
		0.400625,
		0.494427,
		0.585802,
		0.671812,
		0.749996,
		0.818422,
		0.875748,
		0.921283,
		0.95506,
		0.977883,
		0.991344,
		0.997729,
		0.999754,
		1,
		0.999966,
		0.999113,
		0.995256,
		0.985655,
		0.967759,
		0.939609,
		0.900005,
		0.848532,
		0.785527,
		0.712016,
		0.629643,
		0.540612,
		0.447628,
		0.353853,
		0.262862,
		0.178617,
		0.105431,
		0.047952,
		0.011138,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.000232,
		0.017685,
		0.059174,
		0.119705,
		0.19457,
		0.279358,
		0.369978,
		0.462682,
		0.55409,
		0.641225,
		0.721547,
		0.793,
		0.854068,
		0.903828,
		0.94202,
		0.9691,
		0.986276,
		0.99547,
		0.999154,
		0.999968,
		1,
		0.999765,
		0.997833,
		0.991726,
		0.978822,
		0.956891,
		0.924359,
		0.880394,
		0.824905,
		0.75849,
		0.682378,
		0.598361,
		0.508744,
		0.416289,
		0.324181,
		0.235987,
		0.155629,
		0.08736,
		0.035743,
		0.005631,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.010606,
		0.045739,
		0.100747,
		0.171018,
		0.252217,
		0.340308,
		0.431573,
		0.522638,
		0.610497,
		0.692551,
		0.766643,
		0.831109,
		0.884829,
		0.927288,
		0.958631,
		0.979712,
		0.992087,
		0.997931,
		0.999776,
		1,
		0.999969,
		0.999193,
		0.995671,
		0.986864,
		0.970373,
		0.944313,
		0.907473,
		0.859359,
		0.800163,
		0.730708,
		0.652395,
		0.56714,
		0.477327,
		0.385763,
		0.295641,
		0.210511,
		0.134253,
		0.071053,
		0.025389,
		0.002008,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.005363,
		0.034098,
		0.083483,
		0.149001,
		0.226396,
		0.311689,
		0.401201,
		0.491569,
		0.579774,
		0.663169,
		0.739516,
		0.807029,
		0.864419,
		0.91095,
		0.946495,
		0.971582,
		0.987421,
		0.995862,
		0.99923,
		0.999971,
		1,
		0.999787,
		0.998024,
		0.992429,
		0.980556,
		0.960286,
		0.930079,
		0.889065,
		0.837049,
		0.77447,
		0.702346,
		0.622222,
		0.536117,
		0.446476,
		0.356136,
		0.268292,
		0.186469,
		0.114498,
		0.056497,
		0.016857,
		0.000221,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.001913,
		0.024223,
		0.067904,
		0.128533,
		0.201932,
		0.284186,
		0.371655,
		0.460998,
		0.549189,
		0.63355,
		0.711778,
		0.781985,
		0.842738,
		0.893112,
		0.932739,
		0.961859,
		0.981357,
		0.992753,
		0.998111,
		0.999797,
		1,
		0.999972,
		0.999265,
		0.996042,
		0.987949,
		0.97273,
		0.948572,
		0.914268,
		0.869259,
		0.813614,
		0.747985,
		0.673559,
		0.592002,
		0.505413,
		0.416289,
		0.327485,
		0.242186,
		0.163887,
		0.096367,
		0.043674,
		0.010111,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.000211,
		0.016085,
		0.053998,
		0.10962,
		0.178855,
		0.25785,
		0.343012,
		0.431024,
		0.518865,
		0.603835,
		0.683581,
		0.756129,
		0.819929,
		0.87389,
		0.917436,
		0.950551,
		0.973821,
		0.98845,
		0.996213,
		0.999297,
		0.999973,
		1,
		0.999806,
		0.998194,
		0.99306,
		0.982117,
		0.963355,
		0.935275,
		0.89698,
		0.848188,
		0.789202,
		0.720862,
		0.644493,
		0.561865,
		0.475143,
		0.386858,
		0.299874,
		0.217363,
		0.142782,
		0.079857,
		0.032563,
		0.005114,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.00965,
		0.041747,
		0.092264,
		0.157186,
		0.232725,
		0.315337,
		0.401737,
		0.488913,
		0.574153,
		0.655065,
		0.72961,
		0.796135,
		0.85341,
		0.900677,
		0.937694,
		0.96478,
		0.982839,
		0.99335,
		0.998273,
		0.999814,
		1,
		0.999974,
		0.999328,
		0.996374,
		0.988925,
		0.974858,
		0.952436,
		0.920461,
		0.878322,
		0.825987,
		0.763962,
		0.693246,
		0.615286,
		0.531932,
		0.445405,
		0.358259,
		0.273359,
		0.193856,
		0.123166,
		0.064959,
		0.023136,
		0.001825,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.004882,
		0.031129,
		0.076461,
		0.13694,
		0.208847,
		0.28869,
		0.373216,
		0.459432,
		0.544621,
		0.626365,
		0.70257,
		0.771497,
		0.831801,
		0.882566,
		0.92335,
		0.954233,
		0.975845,
		0.989376,
		0.996527,
		0.999358,
		0.999976,
		1,
		0.999823,
		0.998347,
		0.993626,
		0.983525,
		0.966137,
		0.940003,
		0.904213,
		0.858415,
		0.802795,
		0.738037,
		0.665277,
		0.586062,
		0.502313,
		0.416289,
		0.330561,
		0.247986,
		0.171689,
		0.105044,
		0.05166,
		0.015366,
		0.000201,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.001742,
		0.022121,
		0.0622,
		0.118126,
		0.186245,
		0.263117,
		0.345531,
		0.430513,
		0.51535,
		0.597605,
		0.675142,
		0.746155,
		0.809195,
		0.863213,
		0.907596,
		0.942207,
		0.967429,
		0.984178,
		0.993888,
		0.998417,
		0.99983,
		1,
		0.999977,
		0.999385,
		0.996672,
		0.989805,
		0.976785,
		0.955947,
		0.926112,
		0.886629,
		0.837381,
		0.778747,
		0.711563,
		0.637084,
		0.556941,
		0.473108,
		0.387877,
		0.303823,
		0.223793,
		0.150882,
		0.088416,
		0.039943,
		0.00922,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.000192,
		0.014693,
		0.049471,
		0.100747,
		0.164939,
		0.238661,
		0.318742,
		0.402235,
		0.486437,
		0.568899,
		0.647454,
		0.720239,
		0.785723,
		0.842738,
		0.890522,
		0.928752,
		0.957583,
		0.977679,
		0.990212,
		0.99681,
		0.999412,
		0.999978,
		1,
		0.999838,
		0.998483,
		0.994136,
		0.984799,
		0.968661,
		0.944313,
		0.910833,
		0.867815,
		0.815347,
		0.753975,
		0.684672,
		0.608789,
		0.528027,
		0.444407,
		0.360237,
		0.278096,
		0.200812,
		0.131445,
		0.073275,
		0.029788,
		0.004665,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.008818,
		0.038254,
		0.084802,
		0.144944,
		0.215353,
		0.292901,
		0.374671,
		0.457972,
		0.540354,
		0.619625,
		0.693878,
		0.761512,
		0.82126,
		0.87223,
		0.913932,
		0.946324,
		0.969836,
		0.985389,
		0.994372,
		0.998546,
		0.999845,
		1,
		0.999979,
		0.999437,
		0.99694,
		0.990599,
		0.978531,
		0.959144,
		0.931276,
		0.894253,
		0.847883,
		0.792437,
		0.72861,
		0.657487,
		0.580504,
		0.499421,
		0.416289,
		0.333432,
		0.253422,
		0.179066,
		0.113386,
		0.059613,
		0.02117,
		0.001665,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.004462,
		0.028532,
		0.070284,
		0.126271,
		0.193222,
		0.268054,
		0.347883,
		0.430036,
		0.512066,
		0.591766,
		0.667192,
		0.736686,
		0.798899,
		0.852824,
		0.897829,
		0.933692,
		0.960634,
		0.979343,
		0.990967,
		0.997064,
		0.99946,
		0.99998,
		1,
		0.999851,
		0.998606,
		0.994596,
		0.985952,
		0.970957,
		0.948246,
		0.9169,
		0.876466,
		0.826947,
		0.768774,
		0.702773,
		0.630124,
		0.552335,
		0.47121,
		0.388827,
		0.307516,
		0.229837,
		0.158573,
		0.096707,
		0.047417,
		0.014064,
		0.000183,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.001593,
		0.02028,
		0.057184,
		0.108924,
		0.172287,
		0.244237,
		0.321925,
		0.402701,
		0.484124,
		0.563978,
		0.640295,
		0.711367,
		0.775775,
		0.832415,
		0.880531,
		0.919743,
		0.950084,
		0.972026,
		0.986488,
		0.994809,
		0.998663,
		0.999858,
		1,
		0.999981,
		0.999482,
		0.997182,
		0.991318,
		0.980117,
		0.962058,
		0.936004,
		0.901258,
		0.857572,
		0.805122,
		0.74448,
		0.676582,
		0.602692,
		0.524376,
		0.443475,
		0.362085,
		0.282534,
		0.207371,
		0.139346,
		0.081404,
		0.03667,
		0.008442,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.000176,
		0.013475,
		0.045488,
		0.092903,
		0.152563,
		0.221482,
		0.296847,
		0.376031,
		0.456608,
		0.536358,
		0.613292,
		0.685666,
		0.752002,
		0.811114,
		0.862134,
		0.904547,
		0.938217,
		0.963419,
		0.980856,
		0.991651,
		0.997294,
		0.999504,
		0.999981,
		1,
		0.999863,
		0.998717,
		0.995012,
		0.986998,
		0.973047,
		0.951842,
		0.922467,
		0.884434,
		0.837676,
		0.782523,
		0.71967,
		0.650148,
		0.575292,
		0.496716,
		0.416289,
		0.336118,
		0.258527,
		0.186045,
		0.121393,
		0.067471,
		0.027353,
		0.004273,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.008089,
		0.035182,
		0.078205,
		0.134062,
		0.199814,
		0.272688,
		0.350084,
		0.42959,
		0.508991,
		0.586284,
		0.659694,
		0.727694,
		0.78903,
		0.842738,
		0.888183,
		0.925079,
		0.953525,
		0.974023,
		0.987485,
		0.995205,
		0.998768,
		0.999869,
		1,
		0.999982,
		0.999524,
		0.9974,
		0.991969,
		0.981561,
		0.96472,
		0.940336,
		0.907703,
		0.86652,
		0.816885,
		0.759263,
		0.694455,
		0.623575,
		0.548017,
		0.469434,
		0.389717,
		0.310976,
		0.235528,
		0.165879,
		0.104717,
		0.054899,
		0.019444,
		0.001526,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.004095,
		0.026246,
		0.064824,
		0.11679,
		0.179254,
		0.249484,
		0.324909,
		0.403137,
		0.481957,
		0.559361,
		0.63355,
		0.702959,
		0.766271,
		0.822445,
		0.870737,
		0.910731,
		0.942367,
		0.965963,
		0.982234,
		0.992272,
		0.997501,
		0.999543,
		0.999983,
		1,
		0.999874,
		0.998817,
		0.995389,
		0.987949,
		0.974954,
		0.955135,
		0.927583,
		0.891784,
		0.84761,
		0.795304,
		0.735449,
		0.66894,
		0.596961,
		0.520954,
		0.442603,
		0.363815,
		0.286701,
		0.213563,
		0.146884,
		0.089317,
		0.043674,
		0.012921,
		0.000168,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.001462,
		0.01866,
		0.052748,
		0.100747,
		0.159817,
		0.227264,
		0.300551,
		0.377306,
		0.455329,
		0.532609,
		0.607332,
		0.677898,
		0.742944,
		0.801356,
		0.8523,
		0.895243,
		0.929985,
		0.956676,
		0.975845,
		0.988393,
		0.995564,
		0.998863,
		0.99988,
		1,
		0.999984,
		0.999561,
		0.997598,
		0.992561,
		0.982876,
		0.967154,
		0.944313,
		0.913638,
		0.874792,
		0.827802,
		0.773038,
		0.711187,
		0.643226,
		0.570395,
		0.49418,
		0.416289,
		0.338636,
		0.263329,
		0.192655,
		0.12907,
		0.07519,
		0.033782,
		0.007758,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.000161,
		0.012401,
		0.041967,
		0.085934,
		0.141513,
		0.206051,
		0.277047,
		0.352149,
		0.429172,
		0.506107,
		0.581127,
		0.65261,
		0.719149,
		0.779572,
		0.832964,
		0.878693,
		0.91643,
		0.946178,
		0.968292,
		0.98349,
		0.992836,
		0.997689,
		0.999578,
		0.999984,
		1,
		0.999884,
		0.998907,
		0.995731,
		0.988816,
		0.976697,
		0.958153,
		0.932289,
		0.898568,
		0.856815,
		0.807193,
		0.750188,
		0.686577,
		0.617404,
		0.543963,
		0.467769,
		0.39055,
		0.314225,
		0.240892,
		0.172822,
		0.112441,
		0.062328,
		0.025204,
		0.003928,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.007447,
		0.032464,
		0.072345,
		0.124349,
		0.185867,
		0.254429,
		0.327712,
		0.403546,
		0.479925,
		0.55502,
		0.627186,
		0.694984,
		0.757191,
		0.812825,
		0.861163,
		0.901765,
		0.934501,
		0.959568,
		0.977512,
		0.98922,
		0.99589,
		0.998949,
		0.999889,
		1,
		0.999985,
		0.999595,
		0.997777,
		0.993099,
		0.984077,
		0.969383,
		0.947967,
		0.919112,
		0.882447,
		0.837941,
		0.785882,
		0.726854,
		0.661713,
		0.591564,
		0.51774,
		0.441786,
		0.365439,
		0.290619,
		0.219416,
		0.154076,
		0.096998,
		0.050721,
		0.017921,
		0.001403,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.003771,
		0.024223,
		0.059973,
		0.108329,
		0.166724,
		0.232725,
		0.304035,
		0.378502,
		0.45413,
		0.529086,
		0.601712,
		0.670542,
		0.734311,
		0.791976,
		0.842738,
		0.886059,
		0.921688,
		0.949683,
		0.970427,
		0.984638,
		0.99335,
		0.99786,
		0.99961,
		0.999985,
		1,
		0.999893,
		0.998989,
		0.996042,
		0.989606,
		0.978292,
		0.960924,
		0.936624,
		0.904839,
		0.865352,
		0.81826,
		0.763962,
		0.70313,
		0.636686,
		0.565786,
		0.4918,
		0.416289,
		0.341002,
		0.267854,
		0.198921,
		0.136427,
		0.082738,
		0.040357,
		0.011912,
		0.000155,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.001347,
		0.017226,
		0.048808,
		0.093452,
		0.148635,
		0.211958,
		0.281153,
		0.354089,
		0.42878,
		0.503396,
		0.576269,
		0.645911,
		0.711022,
		0.770509,
		0.823506,
		0.869387,
		0.907796,
		0.938663,
		0.962225,
		0.979038,
		0.989975,
		0.996187,
		0.999027,
		0.999897,
		1,
		0.999986,
		0.999625,
		0.99794,
		0.99359,
		0.985175,
		0.971428,
		0.95133,
		0.924165,
		0.889537,
		0.847365,
		0.797864,
		0.741528,
		0.679106,
		0.61158,
		0.540147,
		0.466205,
		0.391333,
		0.317281,
		0.245958,
		0.179424,
		0.119879,
		0.069658,
		0.031222,
		0.007154,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.000148,
		0.011451,
		0.038838,
		0.079717,
		0.131607,
		0.192146,
		0.259096,
		0.330348,
		0.40393,
		0.478015,
		0.550931,
		0.621173,
		0.687413,
		0.748515,
		0.803552,
		0.851827,
		0.892886,
		0.926546,
		0.95291,
		0.972387,
		0.985689,
		0.993819,
		0.998016,
		0.999639,
		0.999987,
		1,
		0.999901,
		0.999064,
		0.996325,
		0.990328,
		0.979753,
		0.963472,
		0.940621,
		0.910641,
		0.873276,
		0.828569,
		0.776841,
		0.718669,
		0.654869,
		0.586474,
		0.514717,
		0.441017,
		0.366965,
		0.294311,
		0.224955,
		0.160938,
		0.104436,
		0.057749,
		0.023299,
		0.003623,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.006878,
		0.030049,
		0.067117,
		0.115644,
		0.173305,
		0.237891,
		0.307317,
		0.379627,
		0.453002,
		0.525768,
		0.596407,
		0.663568,
		0.726079,
		0.782964,
		0.833458,
		0.877025,
		0.913379,
		0.942504,
		0.964669,
		0.980438,
		0.990665,
		0.996457,
		0.999098,
		0.999905,
		1,
		0.999987,
		0.999653,
		0.998088,
		0.994038,
		0.986181,
		0.973306,
		0.954428,
		0.928835,
		0.896111,
		0.856131,
		0.809049,
		0.755278,
		0.69547,
		0.630499,
		0.561441,
		0.489559,
		0.416289,
		0.343228,
		0.272124,
		0.204865,
		0.143473,
		0.090096,
		0.047001,
		0.01657,
		0.001294,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.003484,
		0.022426,
		0.055645,
		0.100747,
		0.155444,
		0.217558,
		0.285028,
		0.355915,
		0.428411,
		0.500843,
		0.571684,
		0.639566,
		0.703287,
		0.761826,
		0.814362,
		0.860284,
		0.899218,
		0.931037,
		0.955885,
		0.974188,
		0.986651,
		0.994248,
		0.998157,
		0.999666,
		0.999988,
		1,
		0.999909,
		0.999131,
		0.996584,
		0.990988,
		0.981094,
		0.965817,
		0.944313,
		0.916014,
		0.880639,
		0.838179,
		0.788888,
		0.73326,
		0.672015,
		0.606074,
		0.536552,
		0.464734,
		0.39207,
		0.32016,
		0.250747,
		0.185705,
		0.127035,
		0.076857,
		0.037403,
		0.011016,
		0.000143,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.001245,
		0.015951,
		0.045292,
		0.086916,
		0.138573,
		0.198115,
		0.263508,
		0.332833,
		0.404291,
		0.476215,
		0.547074,
		0.615483,
		0.680218,
		0.74022,
		0.794618,
		0.842738,
		0.884124,
		0.918552,
		0.946052,
		0.96692,
		0.981724,
		0.991297,
		0.996704,
		0.999163,
		0.999912,
		1,
		0.999988,
		0.999678,
		0.998223,
		0.994448,
		0.987103,
		0.975034,
		0.957286,
		0.933156,
		0.902212,
		0.864293,
		0.819497,
		0.768167,
		0.710869,
		0.64838,
		0.581665,
		0.511868,
		0.440294,
		0.368403,
		0.297795,
		0.230203,
		0.167488,
		0.111627,
		0.064712,
		0.028941,
		0.006618,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.000137,
		0.010606,
		0.036046,
		0.074147,
		0.122695,
		0.179577,
		0.242783,
		0.310415,
		0.380687,
		0.451939,
		0.522638,
		0.591389,
		0.656948,
		0.718226,
		0.774307,
		0.824461,
		0.868161,
		0.905097,
		0.935195,
		0.958631,
		0.975845,
		0.987535,
		0.99464,
		0.998286,
		0.99969,
		0.999988,
		1,
		0.999915,
		0.999193,
		0.99682,
		0.991594,
		0.982327,
		0.967979,
		0.947725,
		0.920996,
		0.887485,
		0.847143,
		0.800163,
		0.746966,
		0.688183,
		0.62464,
		0.557338,
		0.487448,
		0.416289,
		0.345327,
		0.27616,
		0.210511,
		0.150223,
		0.09725,
		0.053654,
		0.021601,
		0.003352,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.006372,
		0.027894,
		0.062433,
		0.107814,
		0.161955,
		0.222873,
		0.28869,
		0.357638,
		0.428063,
		0.498434,
		0.567351,
		0.63355,
		0.695919,
		0.753505,
		0.805528,
		0.851398,
		0.890728,
		0.92335,
		0.949334,
		0.968996,
		0.982906,
		0.991877,
		0.99693,
		0.999222,
		0.999918,
		1,
		0.999989,
		0.999701,
		0.998347,
		0.994823,
		0.987949,
		0.976625,
		0.959925,
		0.937158,
		0.907879,
		0.871897,
		0.829262,
		0.780253,
		0.725363,
		0.665277,
		0.600863,
		0.533157,
		0.463346,
		0.392764,
		0.322879,
		0.255282,
		0.191686,
		0.133916,
		0.083901,
		0.043674,
		0.015366,
		0.001198,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.003228,
		0.020821,
		0.051768,
		0.09393,
		0.145256,
		0.203793,
		0.267685,
		0.335179,
		0.404632,
		0.474517,
		0.543429,
		0.610092,
		0.673374,
		0.732288,
		0.786014,
		0.833904,
		0.875503,
		0.910561,
		0.939048,
		0.961169,
		0.977373,
		0.988347,
		0.994999,
		0.998404,
		0.999712,
		0.999989,
		1,
		0.999921,
		0.999249,
		0.997036,
		0.992149,
		0.983462,
		0.969973,
		0.950882,
		0.925619,
		0.893857,
		0.855511,
		0.810721,
		0.759844,
		0.703432,
		0.642221,
		0.577116,
		0.509177,
		0.439612,
		0.369759,
		0.301087,
		0.235182,
		0.173742,
		0.118572,
		0.071577,
		0.034761,
		0.010218,
		0.000132,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.001154,
		0.014812,
		0.042142,
		0.081039,
		0.129487,
		0.185559,
		0.247423,
		0.313342,
		0.381687,
		0.450936,
		0.51968,
		0.586639,
		0.650658,
		0.710728,
		0.76599,
		0.815748,
		0.859486,
		0.896877,
		0.927805,
		0.952372,
		0.970913,
		0.983995,
		0.99241,
		0.997137,
		0.999276,
		0.999924,
		1,
		0.99999,
		0.999722,
		0.99846,
		0.995167,
		0.988728,
		0.978092,
		0.962366,
		0.940869,
		0.913148,
		0.878987,
		0.838394,
		0.791594,
		0.739008,
		0.681244,
		0.619082,
		0.553458,
		0.485454,
		0.416289,
		0.34731,
		0.27998,
		0.215878,
		0.156687,
		0.104192,
		0.060273,
		0.026902,
		0.00614,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.000127,
		0.009851,
		0.033544,
		0.069138,
		0.114651,
		0.168182,
		0.227922,
		0.292156,
		0.359265,
		0.427734,
		0.496159,
		0.563249,
		0.627839,
		0.688896,
		0.745529,
		0.797,
		0.842738,
		0.882352,
		0.915644,
		0.942622,
		0.963517,
		0.978782,
		0.989094,
		0.995329,
		0.998512,
		0.999732,
		0.99999,
		1,
		0.999927,
		0.999301,
		0.997234,
		0.99266,
		0.984507,
		0.971816,
		0.953807,
		0.929913,
		0.899793,
		0.863328,
		0.820615,
		0.771949,
		0.717815,
		0.658868,
		0.595924,
		0.529947,
		0.462036,
		0.39342,
		0.325448,
		0.259581,
		0.197385,
		0.140529,
		0.090775,
		0.049979,
		0.020082,
		0.003111,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.00592,
		0.025962,
		0.058222,
		0.100747,
		0.151669,
		0.2092,
		0.271644,
		0.337397,
		0.404954,
		0.472913,
		0.539979,
		0.604978,
		0.666857,
		0.724699,
		0.777728,
		0.825327,
		0.867043,
		0.902608,
		0.931946,
		0.955189,
		0.972685,
		0.984999,
		0.992899,
		0.997327,
		0.999325,
		0.999929,
		1,
		0.99999,
		0.999741,
		0.998563,
		0.995483,
		0.989445,
		0.979446,
		0.964625,
		0.944313,
		0.918052,
		0.885604,
		0.846941,
		0.802239,
		0.751857,
		0.696335,
		0.636368,
		0.572806,
		0.506634,
		0.438967,
		0.371041,
		0.304203,
		0.23991,
		0.179716,
		0.125272,
		0.07832,
		0.040688,
		0.014288,
		0.001112,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.003,
		0.019382,
		0.048281,
		0.087777,
		0.136026,
		0.191269,
		0.251827,
		0.316114,
		0.382633,
		0.449987,
		0.516882,
		0.582134,
		0.644676,
		0.703566,
		0.758,
		0.807315,
		0.851009,
		0.888746,
		0.920376,
		0.945942,
		0.965692,
		0.980084,
		0.989783,
		0.995632,
		0.998612,
		0.99975,
		0.999991,
		1,
		0.999932,
		0.999348,
		0.997416,
		0.99313,
		0.985471,
		0.973521,
		0.95652,
		0.933906,
		0.905327,
		0.870636,
		0.82989,
		0.783333,
		0.731385,
		0.674631,
		0.613805,
		0.549783,
		0.483569,
		0.416289,
		0.349184,
		0.283601,
		0.220985,
		0.16288,
		0.110919,
		0.066821,
		0.032389,
		0.009504,
		0.000123,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.001072,
		0.013791,
		0.039308,
		0.075735,
		0.121258,
		0.17414,
		0.232725,
		0.295441,
		0.360804,
		0.427424,
		0.494005,
		0.559361,
		0.622411,
		0.682196,
		0.737881,
		0.788769,
		0.834309,
		0.874111,
		0.907953,
		0.935798,
		0.957802,
		0.974325,
		0.985926,
		0.99335,
		0.997501,
		0.99937,
		0.999934,
		1,
		0.999991,
		0.999759,
		0.998658,
		0.995774,
		0.990107,
		0.980698,
		0.966719,
		0.947512,
		0.92262,
		0.891784,
		0.854945,
		0.812236,
		0.763962,
		0.710598,
		0.652767,
		0.591237,
		0.526907,
		0.460796,
		0.39404,
		0.327881,
		0.263662,
		0.20282,
		0.146884,
		0.09747,
		0.056273,
		0.02507,
		0.005711,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.000118,
		0.009174,
		0.031293,
		0.064618,
		0.107365,
		0.157821,
		0.214352,
		0.275401,
		0.339498,
		0.405259,
		0.471394,
		0.536711,
		0.60012,
		0.660647,
		0.717434,
		0.769751,
		0.817006,
		0.858756,
		0.89472,
		0.924786,
		0.949027,
		0.967708,
		0.981289,
		0.990418,
		0.995911,
		0.998703,
		0.999767,
		0.999991,
		1,
		0.999937,
		0.999391,
		0.997583,
		0.993563,
		0.986362,
		0.9751,
		0.959038,
		0.937623,
		0.91049,
		0.877473,
		0.838591,
		0.794042,
		0.744191,
		0.689557,
		0.630801,
		0.568718,
		0.504225,
		0.438357,
		0.372254,
		0.307157,
		0.244406,
		0.185427,
		0.131734,
		0.084923,
		0.046668,
		0.018718,
		0.002895,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.005514,
		0.024223,
		0.054421,
		0.094348,
		0.14232,
		0.196721,
		0.256014,
		0.318742,
		0.383529,
		0.449089,
		0.51423,
		0.577857,
		0.638979,
		0.69672,
		0.750323,
		0.799158,
		0.842738,
		0.880725,
		0.912942,
		0.939384,
		0.96023,
		0.975845,
		0.986782,
		0.993766,
		0.997662,
		0.999412,
		0.999939,
		1,
		0.999992,
		0.999775,
		0.998745,
		0.996042,
		0.990717,
		0.981857,
		0.968661,
		0.950488,
		0.926878,
		0.897559,
		0.862446,
		0.82163,
		0.77537,
		0.724081,
		0.668323,
		0.608789,
		0.546297,
		0.481783,
		0.416289,
		0.35096,
		0.287037,
		0.225849,
		0.168815,
		0.117431,
		0.073275,
		0.037997,
		0.01332,
		0.001035,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.002795,
		0.018088,
		0.045134,
		0.082205,
		0.127639,
		0.179843,
		0.237298,
		0.298559,
		0.362263,
		0.427129,
		0.491965,
		0.555671,
		0.617247,
		0.675798,
		0.730545,
		0.780826,
		0.826114,
		0.866019,
		0.900305,
		0.928899,
		0.951898,
		0.96958,
		0.982404,
		0.991005,
		0.996168,
		0.998786,
		0.999782,
		0.999992,
		1,
		0.999941,
		0.999431,
		0.997737,
		0.993962,
		0.987186,
		0.976563,
		0.961379,
		0.941085,
		0.915312,
		0.883873,
		0.846758,
		0.804122,
		0.75628,
		0.703692,
		0.646952,
		0.586783,
		0.524024,
		0.459622,
		0.394628,
		0.330188,
		0.26754,
		0.208006,
		0.152992,
		0.103978,
		0.062522,
		0.030251,
		0.008861,
		0.000114,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.000999,
		0.012873,
		0.036751,
		0.070933,
		0.113781,
		0.163726,
		0.219266,
		0.278972,
		0.34149,
		0.405548,
		0.469955,
		0.533608,
		0.595501,
		0.654724,
		0.710476,
		0.762069,
		0.808939,
		0.850653,
		0.88692,
		0.917603,
		0.942727,
		0.962487,
		0.977255,
		0.987575,
		0.99415,
		0.997809,
		0.99945,
		0.999943,
		1,
		0.999992,
		0.99979,
		0.998826,
		0.996289,
		0.991281,
		0.982931,
		0.970465,
		0.953258,
		0.930852,
		0.902962,
		0.869479,
		0.830462,
		0.786125,
		0.736831,
		0.68308,
		0.625499,
		0.564834,
		0.501941,
		0.437779,
		0.373403,
		0.309961,
		0.248685,
		0.190888,
		0.137962,
		0.091374,
		0.052658,
		0.023419,
		0.005327,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.00011,
		0.008564,
		0.029261,
		0.060526,
		0.100747,
		0.148377,
		0.201932,
		0.259999,
		0.321236,
		0.384378,
		0.448238,
		0.511714,
		0.573791,
		0.63355,
		0.690172,
		0.742944,
		0.79127,
		0.834679,
		0.872831,
		0.905532,
		0.932739,
		0.954571,
		0.971319,
		0.983438,
		0.991548,
		0.996405,
		0.998863,
		0.999797,
		0.999993,
		1,
		0.999945,
		0.999468,
		0.997879,
		0.994331,
		0.987949,
		0.977922,
		0.963556,
		0.944313,
		0.919818,
		0.889869,
		0.854428,
		0.813614,
		0.767695,
		0.717079,
		0.6623,
		0.604014,
		0.542987,
		0.48009,
		0.416289,
		0.352645,
		0.290302,
		0.230487,
		0.174504,
		0.12373,
		0.079615,
		0.043674,
		0.017488,
		0.0027,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.005149,
		0.022654,
		0.050979,
		0.088536,
		0.133799,
		0.185305,
		0.241656,
		0.301522,
		0.363647,
		0.42685,
		0.490028,
		0.552163,
		0.612327,
		0.669685,
		0.723506,
		0.773163,
		0.818151,
		0.858088,
		0.892725,
		0.921961,
		0.945845,
		0.964588,
		0.978564,
		0.98831,
		0.994505,
		0.997946,
		0.999485,
		0.999947,
		1,
		0.999993,
		0.999803,
		0.9989,
		0.996517,
		0.991804,
		0.983926,
		0.972142,
		0.955839,
		0.934562,
		0.908019,
		0.876079,
		0.83877,
		0.796268,
		0.748889,
		0.697079,
		0.641405,
		0.582546,
		0.521286,
		0.458508,
		0.395185,
		0.332378,
		0.27123,
		0.212959,
		0.158861,
		0.110298,
		0.0687,
		0.035564,
		0.012447,
		0.000966,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.00261,
		0.016918,
		0.042284,
		0.077145,
		0.119997,
		0.169394,
		0.223957,
		0.282369,
		0.343382,
		0.405822,
		0.468589,
		0.530661,
		0.591103,
		0.649069,
		0.703808,
		0.754672,
		0.801123,
		0.842738,
		0.879226,
		0.910426,
		0.936326,
		0.957064,
		0.972936,
		0.984397,
		0.99205,
		0.996625,
		0.998934,
		0.99981,
		0.999993,
		1,
		0.999948,
		0.999501,
		0.99801,
		0.994673,
		0.988657,
		0.979184,
		0.965583,
		0.947325,
		0.924033,
		0.895491,
		0.861636,
		0.822556,
		0.778478,
		0.729761,
		0.676885,
		0.620445,
		0.561141,
		0.499772,
		0.43723,
		0.374495,
		0.312626,
		0.252761,
		0.196114,
		0.143965,
		0.097664,
		0.058623,
		0.028318,
		0.008282,
		0.000107,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.000934,
		0.012042,
		0.034434,
		0.066571,
		0.10697,
		0.154206,
		0.206915,
		0.263794,
		0.323607,
		0.385184,
		0.44743,
		0.509323,
		0.569922,
		0.628371,
		0.683903,
		0.73585,
		0.783645,
		0.826833,
		0.865077,
		0.89817,
		0.926039,
		0.948756,
		0.966545,
		0.979781,
		0.988991,
		0.994834,
		0.998072,
		0.999517,
		0.99995,
		1,
		0.999993,
		0.999816,
		0.998968,
		0.996728,
		0.992288,
		0.984851,
		0.973702,
		0.958246,
		0.938031,
		0.912756,
		0.882275,
		0.846589,
		0.805839,
		0.760297,
		0.710363,
		0.656545,
		0.599465,
		0.53984,
		0.478481,
		0.416289,
		0.354245,
		0.293408,
		0.234912,
		0.179959,
		0.12982,
		0.085828,
		0.049379,
		0.021925,
		0.004979,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.000103,
		0.008014,
		0.02742,
		0.056809,
		0.094719,
		0.139745,
		0.190539,
		0.245813,
		0.304341,
		0.364963,
		0.426585,
		0.488188,
		0.548826,
		0.607637,
		0.66384,
		0.716747,
		0.76577,
		0.810421,
		0.850326,
		0.885232,
		0.915012,
		0.93968,
		0.959389,
		0.974442,
		0.985289,
		0.992516,
		0.996827,
		0.999,
		0.999822,
		0.999993,
		1,
		0.999952,
		0.999532,
		0.998131,
		0.994989,
		0.989313,
		0.980357,
		0.967473,
		0.950139,
		0.927979,
		0.900765,
		0.868415,
		0.830985,
		0.788668,
		0.741777,
		0.690745,
		0.636109,
		0.578511,
		0.518684,
		0.45745,
		0.395714,
		0.33446,
		0.274744,
		0.217694,
		0.164503,
		0.116428,
		0.074788,
		0.040959,
		0.016376,
		0.002525,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.004818,
		0.021232,
		0.047852,
		0.083242,
		0.126014,
		0.174837,
		0.228439,
		0.285604,
		0.345181,
		0.406083,
		0.46729,
		0.527856,
		0.586911,
		0.643665,
		0.697414,
		0.747548,
		0.793551,
		0.835018,
		0.871652,
		0.90328,
		0.929857,
		0.951476,
		0.968369,
		0.980913,
		0.989624,
		0.995138,
		0.998188,
		0.999547,
		0.999953,
		1,
		0.999994,
		0.999827,
		0.999031,
		0.996923,
		0.992736,
		0.985711,
		0.975156,
		0.960493,
		0.941275,
		0.917197,
		0.888097,
		0.853952,
		0.814873,
		0.771094,
		0.722968,
		0.670956,
		0.615622,
		0.557624,
		0.497709,
		0.436709,
		0.375532,
		0.315161,
		0.25665,
		0.201118,
		0.14975,
		0.103789,
		0.064539,
		0.033357,
		0.011657,
		0.000903,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.002443,
		0.015859,
		0.039695,
		0.072536,
		0.113014,
		0.159817,
		0.211684,
		0.267414,
		0.325863,
		0.385951,
		0.446662,
		0.507048,
		0.566235,
		0.623424,
		0.677898,
		0.729027,
		0.776274,
		0.8192,
		0.857473,
		0.890876,
		0.919314,
		0.942819,
		0.96156,
		0.975845,
		0.986118,
		0.992949,
		0.997015,
		0.999061,
		0.999833,
		0.999994,
		1,
		0.999955,
		0.999561,
		0.998243,
		0.995282,
		0.989923,
		0.98145,
		0.969235,
		0.952769,
		0.931675,
		0.905717,
		0.874792,
		0.838934,
		0.7983,
		0.753166,
		0.703917,
		0.651042,
		0.595126,
		0.536843,
		0.476952,
		0.416289,
		0.355767,
		0.296367,
		0.239139,
		0.185194,
		0.135705,
		0.091904,
		0.055077,
		0.026564,
		0.007758,
		0.0001,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.000874,
		0.01129,
		0.03233,
		0.062598,
		0.100747,
		0.145482,
		0.195557,
		0.249783,
		0.307026,
		0.366214,
		0.426333,
		0.486437,
		0.545648,
		0.60316,
		0.658245,
		0.710256,
		0.758636,
		0.802918,
		0.842738,
		0.877839,
		0.908079,
		0.933436,
		0.95402,
		0.970072,
		0.981968,
		0.990212,
		0.995421,
		0.998296,
		0.999575,
		0.999956,
		1,
		0.999994,
		0.999838,
		0.999089,
		0.997104,
		0.993153,
		0.98651,
		0.976511,
		0.962592,
		0.944313,
		0.921364,
		0.893571,
		0.86089,
		0.823405,
		0.781314,
		0.734932,
		0.684672,
		0.631048,
		0.574663,
		0.516206,
		0.456443,
		0.396218,
		0.336442,
		0.278096,
		0.222223,
		0.169927,
		0.122371,
		0.08077,
		0.046396,
		0.02057,
		0.004665,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		9.6e-005,
		0.007514,
		0.025748,
		0.053423,
		0.089212,
		0.131836,
		0.180066,
		0.232725,
		0.28869,
		0.346894,
		0.406331,
		0.466054,
		0.525185,
		0.582912,
		0.638497,
		0.69128,
		0.740684,
		0.78622,
		0.827492,
		0.864209,
		0.896184,
		0.92335,
		0.945759,
		0.96359,
		0.977154,
		0.986889,
		0.99335,
		0.997189,
		0.999117,
		0.999843,
		0.999994,
		1,
		0.999957,
		0.999588,
		0.998347,
		0.995554,
		0.99049,
		0.982468,
		0.970881,
		0.95523,
		0.935141,
		0.910368,
		0.880796,
		0.846434,
		0.80741,
		0.763962,
		0.716437,
		0.665277,
		0.611015,
		0.554272,
		0.495746,
		0.436213,
		0.376519,
		0.317576,
		0.260362,
		0.205913,
		0.155325,
		0.109748,
		0.070384,
		0.038489,
		0.015366,
		0.002366,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.004519,
		0.019939,
		0.045004,
		0.078406,
		0.118882,
		0.165218,
		0.216251,
		0.270869,
		0.328012,
		0.386681,
		0.44593,
		0.504881,
		0.562719,
		0.618696,
		0.672143,
		0.722464,
		0.769149,
		0.811778,
		0.850025,
		0.883666,
		0.912588,
		0.936792,
		0.956401,
		0.971663,
		0.982951,
		0.990759,
		0.995683,
		0.998396,
		0.9996,
		0.999959,
		1,
		0.999994,
		0.999848,
		0.999143,
		0.997272,
		0.993541,
		0.987255,
		0.977776,
		0.964555,
		0.947158,
		0.925275,
		0.89872,
		0.867431,
		0.831466,
		0.790993,
		0.746289,
		0.697728,
		0.645775,
		0.590984,
		0.533987,
		0.475496,
		0.416289,
		0.357216,
		0.299189,
		0.243179,
		0.190218,
		0.141391,
		0.097836,
		0.060743,
		0.031349,
		0.01094,
		0.000846,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		0.002292,
		0.014896,
		0.037337,
		0.068326,
		0.106619,
		0.151018,
		0.200371,
		0.253577,
		0.309587,
		0.367405,
		0.426093,
		0.484769,
		0.542616,
		0.598882,
		0.652886,
		0.704019,
		0.751752,
		0.79564,
		0.835329,
		0.870561,
		0.90118,
		0.927141,
		0.948514,
		0.965488,
		0.978376,
		0.987608,
		0.993724,
		0.997351,
		0.999169,
		0.999852,
		0.999995,
		1,
		0.99996,
		0.999612,
		0.998443,
		0.995807,
		0.991018,
		0.983418,
		0.972419,
		0.957534,
		0.938392,
		0.914741,
		0.886452,
		0.853514,
		0.816028,
		0.774201,
		0.72834,
		0.678846,
		0.626207,
		0.570991,
		0.513845,
		0.455485,
		0.396697,
		0.33833,
		0.281295,
		0.226559,
		0.175145,
		0.128128,
		0.086636,
		0.051842,
		0.024968,
		0.007282,
		9.3e-005,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.00082,
		0.010606,
		0.030412,
		0.058969,
		0.095049,
		0.137468,
		0.185091,
		0.236827,
		0.291635,
		0.348527,
		0.406567,
		0.464877,
		0.522638,
		0.579092,
		0.63355,
		0.685391,
		0.73407,
		0.779121,
		0.820162,
		0.856905,
		0.889157,
		0.916829,
		0.939942,
		0.958631,
		0.97315,
		0.983869,
		0.991269,
		0.995927,
		0.998488,
		0.999624,
		0.999961,
		1,
		0.999995,
		0.999857,
		0.999193,
		0.997428,
		0.993901,
		0.987949,
		0.978957,
		0.966392,
		0.949827,
		0.92895,
		0.903567,
		0.873601,
		0.839085,
		0.800163,
		0.757075,
		0.710157,
		0.659834,
		0.606611,
		0.551073,
		0.493875,
		0.435741,
		0.377459,
		0.31988,
		0.26391,
		0.210511,
		0.160699,
		0.115538,
		0.076143,
		0.043674,
		0.019337,
		0.004379,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		9.1e-005,
		0.00706,
		0.024223,
		0.05033,
		0.084169,
		0.124573,
		0.170419,
		0.220628,
		0.27417,
		0.330063,
		0.387376,
		0.445233,
		0.502815,
		0.559361,
		0.614173,
		0.666622,
		0.716147,
		0.762263,
		0.804566,
		0.842738,
		0.876554,
		0.905884,
		0.930704,
		0.951098,
		0.967266,
		0.979518,
		0.988279,
		0.994072,
		0.997501,
		0.999217,
		0.999861,
		0.999995,
		1,
		0.999962,
		0.999635,
		0.998532,
		0.996042,
		0.99151,
		0.984305,
		0.973858,
		0.959694,
		0.941445,
		0.918855,
		0.891784,
		0.860201,
		0.824185,
		0.783913,
		0.739658,
		0.691782,
		0.64073,
		0.587025,
		0.531263,
		0.474108,
		0.416289,
		0.358598,
		0.301882,
		0.247045,
		0.195044,
		0.146884,
		0.103621,
		0.066356,
		0.036235,
		0.014447,
		0.002221,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.001428,
		0.01237,
		0.033031,
		0.062177,
		0.098612,
		0.141187,
		0.188793,
		0.24037,
		0.294906,
		0.351437,
		0.409054,
		0.466901,
		0.52418,
		0.580154,
		0.63415,
		0.685564,
		0.733863,
		0.778592,
		0.819379,
		0.855939,
		0.888082,
		0.915718,
		0.938862,
		0.957642,
		0.972299,
		0.983187,
		0.990768,
		0.995601,
		0.99831,
		0.999551,
		0.999945,
		1,
		0.999998,
		0.999896,
		0.999343,
		0.997796,
		0.994617,
		0.989154,
		0.980795,
		0.969004,
		0.953345,
		0.933492,
		0.909232,
		0.880467,
		0.847207,
		0.809569,
		0.767766,
		0.722107,
		0.672983,
		0.620871,
		0.566321,
		0.509955,
		0.452462,
		0.394596,
		0.337168,
		0.281048,
		0.227159,
		0.176478,
		0.130029,
		0.088884,
		0.054161,
		0.027023,
		0.008674,
		0.000361,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.001451,
		0.01256,
		0.033529,
		0.063092,
		0.100028,
		0.14316,
		0.191355,
		0.24353,
		0.298652,
		0.355737,
		0.413858,
		0.472143,
		0.529781,
		0.586022,
		0.640185,
		0.691659,
		0.739908,
		0.784477,
		0.824995,
		0.861185,
		0.892867,
		0.919966,
		0.942516,
		0.960668,
		0.974689,
		0.984964,
		0.991988,
		0.996348,
		0.998695,
		0.999698,
		0.999974,
		1,
		0.99999,
		0.999813,
		0.999046,
		0.997097,
		0.993303,
		0.987004,
		0.977591,
		0.964547,
		0.947458,
		0.926026,
		0.900072,
		0.869533,
		0.834457,
		0.794999,
		0.751412,
		0.704043,
		0.653328,
		0.599783,
		0.544,
		0.486644,
		0.428447,
		0.370205,
		0.312776,
		0.257074,
		0.204067,
		0.154777,
		0.110275,
		0.071682,
		0.040162,
		0.016928,
		0.003233,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.001474,
		0.012755,
		0.034038,
		0.064028,
		0.101474,
		0.145174,
		0.193969,
		0.246752,
		0.302467,
		0.360112,
		0.41874,
		0.477463,
		0.535457,
		0.59196,
		0.646281,
		0.697803,
		0.745987,
		0.790377,
		0.830607,
		0.866407,
		0.897607,
		0.924149,
		0.946088,
		0.963598,
		0.976975,
		0.986637,
		0.993109,
		0.997011,
		0.999017,
		0.999807,
		0.99999,
		1,
		0.999973,
		0.999688,
		0.998655,
		0.99624,
		0.991758,
		0.984543,
		0.973997,
		0.959621,
		0.941027,
		0.91795,
		0.890244,
		0.857883,
		0.820956,
		0.779657,
		0.73428,
		0.685216,
		0.63294,
		0.578013,
		0.52107,
		0.46282,
		0.40404,
		0.345573,
		0.288319,
		0.233241,
		0.181354,
		0.133727,
		0.09148,
		0.055783,
		0.027851,
		0.008946,
		0.000372,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.001497,
		0.012955,
		0.034559,
		0.064984,
		0.102952,
		0.14723,
		0.196635,
		0.250035,
		0.306352,
		0.364563,
		0.423701,
		0.482863,
		0.541209,
		0.597967,
		0.652437,
		0.703994,
		0.752097,
		0.796291,
		0.836212,
		0.8716,
		0.902298,
		0.928264,
		0.949573,
		0.966428,
		0.979155,
		0.988203,
		0.994133,
		0.997593,
		0.999281,
		0.999886,
		0.999997,
		1,
		0.99994,
		0.999508,
		0.998153,
		0.995203,
		0.989954,
		0.981742,
		0.969981,
		0.954194,
		0.934022,
		0.909232,
		0.879718,
		0.845492,
		0.806683,
		0.763528,
		0.716363,
		0.665622,
		0.611825,
		0.555576,
		0.497557,
		0.438523,
		0.379296,
		0.320766,
		0.263882,
		0.209654,
		0.159146,
		0.113477,
		0.073818,
		0.041389,
		0.017457,
		0.003336,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.001521,
		0.013159,
		0.035092,
		0.065963,
		0.104462,
		0.14933,
		0.199356,
		0.253383,
		0.310309,
		0.369091,
		0.428742,
		0.488343,
		0.547038,
		0.604044,
		0.658652,
		0.710231,
		0.758237,
		0.802214,
		0.841807,
		0.876762,
		0.906935,
		0.932304,
		0.952968,
		0.969156,
		0.981226,
		0.989663,
		0.99506,
		0.998097,
		0.999493,
		0.999938,
		1,
		0.999997,
		0.999882,
		0.999259,
		0.99752,
		0.993959,
		0.987863,
		0.97857,
		0.965511,
		0.948233,
		0.926409,
		0.899843,
		0.868468,
		0.832336,
		0.791619,
		0.746597,
		0.697651,
		0.645261,
		0.589991,
		0.532493,
		0.473495,
		0.413799,
		0.354276,
		0.295864,
		0.239562,
		0.18643,
		0.137583,
		0.094192,
		0.057479,
		0.028718,
		0.00923,
		0.000384,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.001546,
		0.013368,
		0.035637,
		0.066963,
		0.106005,
		0.151474,
		0.202132,
		0.256796,
		0.31434,
		0.373698,
		0.433865,
		0.493904,
		0.552945,
		0.610192,
		0.664927,
		0.716513,
		0.764404,
		0.808146,
		0.847388,
		0.881887,
		0.911514,
		0.936266,
		0.956268,
		0.971777,
		0.983187,
		0.991015,
		0.995893,
		0.998528,
		0.999658,
		0.999971,
		1,
		0.999989,
		0.999788,
		0.998923,
		0.996733,
		0.992484,
		0.985457,
		0.974996,
		0.960553,
		0.941704,
		0.918157,
		0.889753,
		0.856465,
		0.81839,
		0.775744,
		0.728851,
		0.678141,
		0.624138,
		0.567455,
		0.50879,
		0.448922,
		0.388702,
		0.32905,
		0.270954,
		0.215467,
		0.163699,
		0.116819,
		0.076051,
		0.042672,
		0.018011,
		0.003444,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.001571,
		0.013582,
		0.036195,
		0.067986,
		0.107582,
		0.153664,
		0.204965,
		0.260276,
		0.318445,
		0.378386,
		0.439072,
		0.499548,
		0.55893,
		0.61641,
		0.67126,
		0.722839,
		0.770596,
		0.814083,
		0.852952,
		0.886971,
		0.916031,
		0.940146,
		0.959469,
		0.974289,
		0.985035,
		0.992261,
		0.996633,
		0.998889,
		0.999782,
		0.999989,
		1,
		0.99997,
		0.999647,
		0.998482,
		0.995768,
		0.990749,
		0.982702,
		0.970985,
		0.955074,
		0.934575,
		0.909232,
		0.878929,
		0.843682,
		0.803632,
		0.759042,
		0.710282,
		0.657831,
		0.60226,
		0.544233,
		0.484498,
		0.423884,
		0.363292,
		0.303696,
		0.246136,
		0.191718,
		0.141606,
		0.097025,
		0.059254,
		0.029626,
		0.009529,
		0.000397,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.001597,
		0.013802,
		0.036766,
		0.069032,
		0.109195,
		0.155901,
		0.207857,
		0.263824,
		0.322628,
		0.383156,
		0.444363,
		0.505276,
		0.564995,
		0.622699,
		0.677651,
		0.729206,
		0.776812,
		0.820021,
		0.858494,
		0.892012,
		0.92048,
		0.943938,
		0.962567,
		0.976689,
		0.986769,
		0.9934,
		0.997284,
		0.999187,
		0.99987,
		0.999997,
		1,
		0.999932,
		0.999443,
		0.997915,
		0.994599,
		0.988724,
		0.979567,
		0.966504,
		0.949038,
		0.926809,
		0.899603,
		0.867344,
		0.830094,
		0.788042,
		0.741497,
		0.690882,
		0.636722,
		0.57964,
		0.520351,
		0.459655,
		0.398433,
		0.33764,
		0.278304,
		0.22152,
		0.168448,
		0.120309,
		0.078386,
		0.044016,
		0.018592,
		0.003558,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.001624,
		0.014026,
		0.037351,
		0.070103,
		0.110843,
		0.158187,
		0.210809,
		0.267443,
		0.326889,
		0.38801,
		0.449741,
		0.511089,
		0.571138,
		0.629057,
		0.6841,
		0.735615,
		0.783049,
		0.825958,
		0.864011,
		0.897002,
		0.924857,
		0.947639,
		0.965558,
		0.978974,
		0.988388,
		0.994434,
		0.99785,
		0.999426,
		0.99993,
		1,
		0.999997,
		0.999866,
		0.999161,
		0.997199,
		0.993199,
		0.986377,
		0.976017,
		0.961517,
		0.94241,
		0.918374,
		0.889236,
		0.854966,
		0.815675,
		0.771599,
		0.723099,
		0.670646,
		0.61482,
		0.556295,
		0.495839,
		0.434306,
		0.372632,
		0.311827,
		0.252975,
		0.197229,
		0.145806,
		0.099986,
		0.061111,
		0.030578,
		0.009842,
		0.00041,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.001651,
		0.014256,
		0.03795,
		0.071199,
		0.112529,
		0.160523,
		0.213823,
		0.271135,
		0.331231,
		0.39295,
		0.455206,
		0.516987,
		0.577362,
		0.635486,
		0.690605,
		0.742062,
		0.789303,
		0.831889,
		0.869498,
		0.901939,
		0.929157,
		0.951243,
		0.968439,
		0.98114,
		0.989891,
		0.995364,
		0.998334,
		0.999612,
		0.999967,
		1,
		0.999988,
		0.99976,
		0.99878,
		0.99631,
		0.991538,
		0.983674,
		0.972017,
		0.955987,
		0.935153,
		0.909232,
		0.878098,
		0.841768,
		0.800401,
		0.754286,
		0.703836,
		0.649575,
		0.592135,
		0.532247,
		0.470734,
		0.408505,
		0.346552,
		0.285944,
		0.227824,
		0.173402,
		0.123957,
		0.08083,
		0.045425,
		0.019201,
		0.003677,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.001679,
		0.014492,
		0.038563,
		0.07232,
		0.114253,
		0.16291,
		0.2169,
		0.2749,
		0.335654,
		0.397978,
		0.460761,
		0.522972,
		0.583667,
		0.641985,
		0.697165,
		0.748546,
		0.795573,
		0.837812,
		0.874952,
		0.906817,
		0.933375,
		0.954745,
		0.971204,
		0.983187,
		0.991278,
		0.996194,
		0.998741,
		0.999752,
		0.999987,
		1,
		0.999966,
		0.999599,
		0.99828,
		0.995219,
		0.989583,
		0.980581,
		0.967529,
		0.949876,
		0.927229,
		0.89935,
		0.866157,
		0.827721,
		0.78425,
		0.736089,
		0.683702,
		0.627672,
		0.568684,
		0.507526,
		0.44508,
		0.382312,
		0.320274,
		0.260094,
		0.202976,
		0.150192,
		0.103084,
		0.063057,
		0.031577,
		0.010171,
		0.000424,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.001708,
		0.014734,
		0.039191,
		0.073468,
		0.116017,
		0.165349,
		0.220042,
		0.278741,
		0.340162,
		0.403095,
		0.466406,
		0.529046,
		0.590052,
		0.648554,
		0.70378,
		0.755064,
		0.801855,
		0.843722,
		0.880367,
		0.911631,
		0.937506,
		0.958141,
		0.973851,
		0.985111,
		0.992549,
		0.996925,
		0.999076,
		0.999852,
		0.999996,
		1,
		0.999923,
		0.999367,
		0.997637,
		0.993899,
		0.987301,
		0.977061,
		0.962515,
		0.943146,
		0.918602,
		0.88869,
		0.853382,
		0.812798,
		0.767202,
		0.716994,
		0.662695,
		0.604944,
		0.544486,
		0.482168,
		0.418929,
		0.355799,
		0.29389,
		0.234393,
		0.178574,
		0.127771,
		0.083389,
		0.046901,
		0.019841,
		0.003803,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.001737,
		0.014982,
		0.039835,
		0.074644,
		0.117821,
		0.167843,
		0.223252,
		0.282661,
		0.344756,
		0.408303,
		0.472143,
		0.535208,
		0.596519,
		0.655192,
		0.710447,
		0.761615,
		0.808146,
		0.849616,
		0.885739,
		0.916375,
		0.941544,
		0.961427,
		0.976376,
		0.986911,
		0.993706,
		0.997561,
		0.999346,
		0.99992,
		1,
		0.999996,
		0.999847,
		0.999046,
		0.996826,
		0.992316,
		0.984657,
		0.973075,
		0.956935,
		0.935757,
		0.909232,
		0.877219,
		0.839741,
		0.796974,
		0.749239,
		0.696992,
		0.640815,
		0.581406,
		0.51957,
		0.456215,
		0.392345,
		0.32905,
		0.267506,
		0.208972,
		0.154777,
		0.106327,
		0.065096,
		0.032625,
		0.010517,
		0.000439,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.001767,
		0.015236,
		0.040494,
		0.075848,
		0.119668,
		0.170393,
		0.22653,
		0.28666,
		0.349439,
		0.413603,
		0.477974,
		0.541461,
		0.603067,
		0.661898,
		0.717165,
		0.768196,
		0.814442,
		0.855488,
		0.891062,
		0.921045,
		0.945484,
		0.964597,
		0.978775,
		0.988586,
		0.994749,
		0.998106,
		0.999557,
		0.999962,
		1,
		0.999986,
		0.999726,
		0.998613,
		0.995817,
		0.990437,
		0.981612,
		0.968585,
		0.950749,
		0.927669,
		0.899083,
		0.864902,
		0.825205,
		0.780225,
		0.730344,
		0.676077,
		0.618068,
		0.557076,
		0.493969,
		0.429721,
		0.365398,
		0.302157,
		0.241242,
		0.183976,
		0.131761,
		0.086071,
		0.048451,
		0.020513,
		0.003934,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.001798,
		0.015497,
		0.04117,
		0.077081,
		0.121558,
		0.173,
		0.229879,
		0.290741,
		0.354211,
		0.418999,
		0.483901,
		0.547804,
		0.609698,
		0.668673,
		0.723934,
		0.774805,
		0.820741,
		0.861335,
		0.896331,
		0.925634,
		0.949321,
		0.967646,
		0.981047,
		0.990135,
		0.995681,
		0.998566,
		0.999716,
		0.999985,
		1,
		0.999961,
		0.999542,
		0.998044,
		0.99458,
		0.988228,
		0.978128,
		0.963547,
		0.943915,
		0.918841,
		0.888115,
		0.851704,
		0.809744,
		0.762531,
		0.710506,
		0.654248,
		0.594464,
		0.531979,
		0.467727,
		0.402746,
		0.338171,
		0.275228,
		0.21523,
		0.159571,
		0.109724,
		0.067236,
		0.033726,
		0.01088,
		0.000454,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.00183,
		0.015765,
		0.041863,
		0.078344,
		0.123492,
		0.175666,
		0.2333,
		0.294906,
		0.359076,
		0.424491,
		0.489923,
		0.55424,
		0.61641,
		0.675515,
		0.73075,
		0.781437,
		0.827036,
		0.867151,
		0.901541,
		0.930138,
		0.95305,
		0.970571,
		0.983187,
		0.991558,
		0.996504,
		0.998946,
		0.999831,
		0.999996,
		0.999999,
		0.999911,
		0.999278,
		0.997312,
		0.993081,
		0.985648,
		0.974162,
		0.95792,
		0.93639,
		0.909232,
		0.87629,
		0.837591,
		0.793332,
		0.743871,
		0.689715,
		0.631507,
		0.570021,
		0.506149,
		0.440894,
		0.375363,
		0.310761,
		0.248386,
		0.189623,
		0.135939,
		0.088884,
		0.050078,
		0.02122,
		0.004073,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.001863,
		0.016039,
		0.042574,
		0.079638,
		0.125472,
		0.178394,
		0.236797,
		0.299157,
		0.364035,
		0.430082,
		0.496045,
		0.560768,
		0.623205,
		0.682423,
		0.737611,
		0.788092,
		0.833326,
		0.872931,
		0.906686,
		0.934549,
		0.956664,
		0.973368,
		0.985193,
		0.992856,
		0.997222,
		0.999253,
		0.999908,
		0.999999,
		0.999996,
		0.999825,
		0.99891,
		0.996388,
		0.991285,
		0.982658,
		0.969673,
		0.951659,
		0.928131,
		0.898801,
		0.863572,
		0.822533,
		0.775945,
		0.724231,
		0.667967,
		0.607863,
		0.544762,
		0.479627,
		0.413531,
		0.347655,
		0.283276,
		0.221767,
		0.164588,
		0.113285,
		0.069482,
		0.034883,
		0.011263,
		0.000471,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.001896,
		0.016321,
		0.043303,
		0.080964,
		0.1275,
		0.181184,
		0.24037,
		0.303496,
		0.369091,
		0.435774,
		0.502265,
		0.567389,
		0.630081,
		0.689396,
		0.744517,
		0.794763,
		0.839605,
		0.87867,
		0.91176,
		0.938862,
		0.960158,
		0.976031,
		0.987065,
		0.994029,
		0.997839,
		0.999493,
		0.999956,
		1,
		0.999984,
		0.999686,
		0.998415,
		0.995238,
		0.989154,
		0.979216,
		0.964615,
		0.944719,
		0.919093,
		0.887506,
		0.849923,
		0.806498,
		0.757559,
		0.703599,
		0.645261,
		0.583329,
		0.518718,
		0.452462,
		0.385711,
		0.319719,
		0.255841,
		0.195527,
		0.140317,
		0.091835,
		0.051789,
		0.021964,
		0.004219,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.001931,
		0.01661,
		0.044051,
		0.082324,
		0.129577,
		0.184039,
		0.244022,
		0.307926,
		0.374245,
		0.441568,
		0.508587,
		0.574105,
		0.63704,
		0.696434,
		0.751464,
		0.80145,
		0.845868,
		0.884362,
		0.916757,
		0.943071,
		0.963527,
		0.978558,
		0.9888,
		0.995079,
		0.998361,
		0.999675,
		0.999983,
		1,
		0.999955,
		0.999475,
		0.997765,
		0.993829,
		0.986646,
		0.975277,
		0.958943,
		0.937054,
		0.909232,
		0.875306,
		0.835308,
		0.789457,
		0.738155,
		0.681965,
		0.621603,
		0.557927,
		0.491928,
		0.424716,
		0.357518,
		0.291667,
		0.228598,
		0.169842,
		0.11702,
		0.071843,
		0.036101,
		0.011667,
		0.000488,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.001966,
		0.016907,
		0.044818,
		0.083718,
		0.131705,
		0.186961,
		0.247755,
		0.31245,
		0.379501,
		0.447466,
		0.515011,
		0.580916,
		0.64408,
		0.703534,
		0.758449,
		0.808146,
		0.85211,
		0.890001,
		0.92167,
		0.94717,
		0.966766,
		0.980944,
		0.990397,
		0.996009,
		0.998793,
		0.999805,
		0.999995,
		0.999999,
		0.999898,
		0.999172,
		0.996928,
		0.992122,
		0.983719,
		0.970795,
		0.952607,
		0.928617,
		0.898503,
		0.862161,
		0.819691,
		0.771385,
		0.717717,
		0.659326,
		0.597004,
		0.531686,
		0.464441,
		0.39646,
		0.32905,
		0.263625,
		0.201706,
		0.144906,
		0.094935,
		0.053588,
		0.022747,
		0.004374,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.002002,
		0.017212,
		0.045605,
		0.085147,
		0.133884,
		0.189951,
		0.251572,
		0.317068,
		0.384859,
		0.453471,
		0.521539,
		0.587822,
		0.651201,
		0.710695,
		0.76547,
		0.814849,
		0.858326,
		0.895582,
		0.926493,
		0.951151,
		0.96987,
		0.983187,
		0.991857,
		0.996822,
		0.999142,
		0.999894,
		0.999999,
		0.999995,
		0.999798,
		0.99875,
		0.995871,
		0.990077,
		0.980326,
		0.965721,
		0.94556,
		0.919359,
		0.886861,
		0.848031,
		0.803041,
		0.752258,
		0.696234,
		0.635684,
		0.571482,
		0.504643,
		0.436317,
		0.367779,
		0.300419,
		0.23574,
		0.175346,
		0.120941,
		0.074325,
		0.037384,
		0.012092,
		0.000506,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.00204,
		0.017525,
		0.046414,
		0.086613,
		0.136118,
		0.193012,
		0.255475,
		0.321785,
		0.390324,
		0.459584,
		0.528173,
		0.594824,
		0.658402,
		0.717915,
		0.772523,
		0.821553,
		0.864511,
		0.901097,
		0.931219,
		0.95501,
		0.972832,
		0.985283,
		0.993181,
		0.997523,
		0.999417,
		0.99995,
		1,
		0.999981,
		0.999638,
		0.998182,
		0.994557,
		0.987648,
		0.97642,
		0.960006,
		0.937751,
		0.909232,
		0.874262,
		0.832877,
		0.785326,
		0.732056,
		0.673697,
		0.611048,
		0.545064,
		0.476846,
		0.407628,
		0.338772,
		0.271757,
		0.208175,
		0.149721,
		0.098193,
		0.055483,
		0.023574,
		0.004537,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.002078,
		0.017847,
		0.047243,
		0.088117,
		0.138407,
		0.196146,
		0.259467,
		0.326602,
		0.395897,
		0.465807,
		0.534913,
		0.601923,
		0.665684,
		0.725192,
		0.779606,
		0.828254,
		0.870658,
		0.90654,
		0.935842,
		0.958739,
		0.97565,
		0.987232,
		0.994368,
		0.998117,
		0.999625,
		0.99998,
		1,
		0.999948,
		0.999396,
		0.997435,
		0.992945,
		0.984791,
		0.971951,
		0.953597,
		0.929128,
		0.898188,
		0.860662,
		0.816662,
		0.76652,
		0.710763,
		0.650105,
		0.585432,
		0.517785,
		0.448352,
		0.378457,
		0.309553,
		0.243212,
		0.181117,
		0.125061,
		0.076937,
		0.038736,
		0.012541,
		0.000525,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.002117,
		0.018178,
		0.048096,
		0.08966,
		0.140754,
		0.199356,
		0.263549,
		0.331523,
		0.40158,
		0.472143,
		0.541761,
		0.609118,
		0.673044,
		0.732524,
		0.786713,
		0.834948,
		0.876762,
		0.911904,
		0.940353,
		0.962333,
		0.978317,
		0.989031,
		0.995423,
		0.99861,
		0.999775,
		0.999995,
		0.999999,
		0.999882,
		0.999046,
		0.996474,
		0.990993,
		0.981456,
		0.966867,
		0.94644,
		0.919639,
		0.886177,
		0.846017,
		0.799352,
		0.746597,
		0.688367,
		0.625463,
		0.55886,
		0.48969,
		0.419232,
		0.348906,
		0.280256,
		0.214952,
		0.154777,
		0.101621,
		0.057479,
		0.024446,
		0.004709,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.002158,
		0.018518,
		0.048971,
		0.091244,
		0.14316,
		0.202643,
		0.267725,
		0.336549,
		0.407377,
		0.478594,
		0.548718,
		0.61641,
		0.680482,
		0.739908,
		0.793842,
		0.841627,
		0.882814,
		0.917181,
		0.944746,
		0.965784,
		0.980831,
		0.990679,
		0.996348,
		0.999011,
		0.999877,
		0.999999,
		0.999994,
		0.999767,
		0.99856,
		0.99526,
		0.988652,
		0.977591,
		0.961112,
		0.938482,
		0.909232,
		0.873152,
		0.830285,
		0.780913,
		0.725537,
		0.664862,
		0.599783,
		0.531365,
		0.460836,
		0.389573,
		0.319089,
		0.251033,
		0.187172,
		0.129392,
		0.079688,
		0.040162,
		0.013016,
		0.000546,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.0022,
		0.018867,
		0.049871,
		0.09287,
		0.145627,
		0.20601,
		0.271997,
		0.341684,
		0.413289,
		0.48516,
		0.555785,
		0.623799,
		0.687996,
		0.747342,
		0.800988,
		0.848287,
		0.88881,
		0.922365,
		0.949014,
		0.969087,
		0.983187,
		0.992177,
		0.997147,
		0.999326,
		0.999942,
		1,
		0.999978,
		0.999581,
		0.997904,
		0.99375,
		0.985874,
		0.973143,
		0.95463,
		0.929666,
		0.897854,
		0.859064,
		0.813428,
		0.761317,
		0.703324,
		0.64025,
		0.573084,
		0.502989,
		0.431293,
		0.359472,
		0.289144,
		0.222058,
		0.160089,
		0.105229,
		0.059585,
		0.025367,
		0.004891,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.002243,
		0.019227,
		0.050795,
		0.094539,
		0.148158,
		0.20946,
		0.276369,
		0.34693,
		0.419319,
		0.491846,
		0.562964,
		0.631286,
		0.695586,
		0.754822,
		0.808146,
		0.854921,
		0.89474,
		0.927446,
		0.953149,
		0.972236,
		0.985382,
		0.993526,
		0.997826,
		0.999565,
		0.999977,
		1,
		0.999939,
		0.9993,
		0.997042,
		0.991898,
		0.982605,
		0.968053,
		0.947363,
		0.919935,
		0.885451,
		0.843868,
		0.795408,
		0.740538,
		0.679948,
		0.614537,
		0.545395,
		0.473788,
		0.401146,
		0.32905,
		0.259225,
		0.193529,
		0.133949,
		0.082589,
		0.041669,
		0.013517,
		0.000567,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.002287,
		0.019597,
		0.051745,
		0.096253,
		0.150754,
		0.212995,
		0.280842,
		0.352291,
		0.425471,
		0.498652,
		0.570256,
		0.638868,
		0.703249,
		0.762345,
		0.815311,
		0.861523,
		0.900598,
		0.932418,
		0.957143,
		0.975226,
		0.987414,
		0.994726,
		0.998392,
		0.999739,
		0.999994,
		0.999999,
		0.999863,
		0.998895,
		0.995933,
		0.989654,
		0.97879,
		0.962262,
		0.939252,
		0.909232,
		0.87197,
		0.827516,
		0.77619,
		0.718555,
		0.655405,
		0.587741,
		0.516759,
		0.44383,
		0.370494,
		0.298442,
		0.229511,
		0.165674,
		0.109032,
		0.061808,
		0.026342,
		0.005084,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.002333,
		0.019978,
		0.052722,
		0.098015,
		0.153419,
		0.216617,
		0.285421,
		0.357769,
		0.431746,
		0.50558,
		0.577661,
		0.646547,
		0.710982,
		0.769907,
		0.822478,
		0.868085,
		0.906376,
		0.937272,
		0.960989,
		0.978051,
		0.989282,
		0.995782,
		0.998853,
		0.999857,
		0.999999,
		0.999993,
		0.999729,
		0.998331,
		0.994532,
		0.986964,
		0.974369,
		0.95571,
		0.930235,
		0.897499,
		0.85736,
		0.809968,
		0.755743,
		0.695352,
		0.629697,
		0.559887,
		0.487226,
		0.413199,
		0.339457,
		0.267809,
		0.200209,
		0.138748,
		0.08565,
		0.043262,
		0.014049,
		0.00059,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.00238,
		0.02037,
		0.053727,
		0.099824,
		0.156153,
		0.220331,
		0.290107,
		0.363368,
		0.438147,
		0.512632,
		0.58518,
		0.654322,
		0.718785,
		0.777504,
		0.82964,
		0.874601,
		0.912065,
		0.941999,
		0.964681,
		0.980706,
		0.990983,
		0.996697,
		0.999216,
		0.999932,
		1,
		0.999974,
		0.999512,
		0.997571,
		0.992789,
		0.983772,
		0.969281,
		0.948332,
		0.920249,
		0.884676,
		0.84157,
		0.791183,
		0.73404,
		0.670922,
		0.602837,
		0.531011,
		0.456863,
		0.381994,
		0.308175,
		0.237335,
		0.171552,
		0.113042,
		0.064157,
		0.027373,
		0.005289,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.002428,
		0.020774,
		0.054761,
		0.101684,
		0.15896,
		0.224139,
		0.294906,
		0.369091,
		0.444677,
		0.519811,
		0.592813,
		0.662191,
		0.726654,
		0.785132,
		0.836792,
		0.881063,
		0.917657,
		0.946591,
		0.968209,
		0.983187,
		0.99252,
		0.997478,
		0.999493,
		0.999973,
		1,
		0.999929,
		0.999185,
		0.996571,
		0.99065,
		0.980016,
		0.963459,
		0.940063,
		0.909232,
		0.870708,
		0.824552,
		0.771123,
		0.711062,
		0.645261,
		0.574848,
		0.501165,
		0.425753,
		0.350337,
		0.276811,
		0.207232,
		0.143807,
		0.088884,
		0.044948,
		0.014613,
		0.000615,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.002478,
		0.021189,
		0.055825,
		0.103595,
		0.161842,
		0.228043,
		0.299819,
		0.37494,
		0.451339,
		0.527118,
		0.600563,
		0.670154,
		0.734587,
		0.792785,
		0.843926,
		0.887461,
		0.923141,
		0.951039,
		0.971569,
		0.98549,
		0.993892,
		0.99813,
		0.999695,
		0.999993,
		0.999999,
		0.999839,
		0.998713,
		0.995285,
		0.988059,
		0.975631,
		0.956839,
		0.930835,
		0.897121,
		0.855538,
		0.806258,
		0.749757,
		0.686792,
		0.618377,
		0.545761,
		0.470412,
		0.393997,
		0.318369,
		0.245554,
		0.177742,
		0.117276,
		0.066642,
		0.028466,
		0.005506,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.00253,
		0.021618,
		0.05692,
		0.105561,
		0.164802,
		0.232048,
		0.30485,
		0.380919,
		0.458135,
		0.534555,
		0.608428,
		0.67821,
		0.742579,
		0.800458,
		0.851035,
		0.893788,
		0.928511,
		0.955335,
		0.974752,
		0.987613,
		0.995103,
		0.998662,
		0.999833,
		0.999999,
		0.999992,
		0.999682,
		0.998056,
		0.993658,
		0.984953,
		0.970552,
		0.949349,
		0.920582,
		0.883851,
		0.839109,
		0.786646,
		0.727057,
		0.661224,
		0.590287,
		0.515624,
		0.438833,
		0.361715,
		0.286256,
		0.214622,
		0.149143,
		0.092302,
		0.046734,
		0.015211,
		0.000641,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.002583,
		0.022059,
		0.058047,
		0.107582,
		0.167843,
		0.236155,
		0.310002,
		0.387032,
		0.465069,
		0.542124,
		0.61641,
		0.686356,
		0.750628,
		0.808146,
		0.858112,
		0.900035,
		0.933754,
		0.959469,
		0.977753,
		0.989554,
		0.996156,
		0.999083,
		0.99992,
		1,
		0.99997,
		0.999429,
		0.997169,
		0.991635,
		0.981267,
		0.964706,
		0.940917,
		0.909232,
		0.869359,
		0.82137,
		0.765677,
		0.703003,
		0.63436,
		0.561021,
		0.484498,
		0.406528,
		0.32905,
		0.254193,
		0.184267,
		0.121749,
		0.069274,
		0.029626,
		0.005737,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.002638,
		0.022514,
		0.059208,
		0.109662,
		0.170967,
		0.24037,
		0.315281,
		0.393283,
		0.472143,
		0.549827,
		0.624508,
		0.694592,
		0.758729,
		0.815842,
		0.865149,
		0.906191,
		0.938862,
		0.963433,
		0.980567,
		0.991312,
		0.997056,
		0.999406,
		0.999969,
		1,
		0.999917,
		0.999046,
		0.996003,
		0.989154,
		0.976929,
		0.95802,
		0.931471,
		0.896717,
		0.853585,
		0.80227,
		0.743316,
		0.67758,
		0.606211,
		0.53062,
		0.452462,
		0.373618,
		0.296172,
		0.222405,
		0.154777,
		0.09592,
		0.048628,
		0.015846,
		0.000668,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.002694,
		0.022984,
		0.060404,
		0.111802,
		0.174178,
		0.244695,
		0.320688,
		0.399673,
		0.47936,
		0.557664,
		0.632722,
		0.702914,
		0.766878,
		0.823539,
		0.872136,
		0.912247,
		0.943825,
		0.967218,
		0.983187,
		0.992887,
		0.997811,
		0.999641,
		0.999991,
		0.999999,
		0.999811,
		0.998492,
		0.994502,
		0.986146,
		0.971868,
		0.950418,
		0.920935,
		0.882967,
		0.836467,
		0.781763,
		0.719536,
		0.650784,
		0.5768,
		0.499142,
		0.419614,
		0.340247,
		0.26328,
		0.191151,
		0.12648,
		0.072063,
		0.030859,
		0.005983,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.002753,
		0.023468,
		0.061637,
		0.114005,
		0.177478,
		0.249134,
		0.326228,
		0.406208,
		0.486723,
		0.565638,
		0.641052,
		0.71132,
		0.77507,
		0.83123,
		0.879064,
		0.918192,
		0.948631,
		0.970816,
		0.98561,
		0.994282,
		0.998429,
		0.999803,
		0.999999,
		0.999991,
		0.999626,
		0.997721,
		0.992605,
		0.982541,
		0.966004,
		0.94182,
		0.909232,
		0.867913,
		0.817948,
		0.759807,
		0.694316,
		0.622621,
		0.546167,
		0.466665,
		0.386075,
		0.306588,
		0.230606,
		0.160731,
		0.099754,
		0.050639,
		0.016522,
		0.000697,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.002813,
		0.023968,
		0.062907,
		0.116272,
		0.180871,
		0.25369,
		0.331906,
		0.412891,
		0.494234,
		0.573749,
		0.649498,
		0.719808,
		0.783299,
		0.838908,
		0.885923,
		0.924016,
		0.953272,
		0.974219,
		0.987832,
		0.995499,
		0.998921,
		0.999905,
		1,
		0.999964,
		0.999327,
		0.996681,
		0.990244,
		0.978262,
		0.959257,
		0.932145,
		0.896287,
		0.851487,
		0.797974,
		0.736367,
		0.667644,
		0.593109,
		0.514364,
		0.433283,
		0.35199,
		0.272844,
		0.198419,
		0.13149,
		0.075024,
		0.032169,
		0.006244,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.002875,
		0.024483,
		0.064217,
		0.118608,
		0.18436,
		0.258369,
		0.337724,
		0.419726,
		0.501896,
		0.582,
		0.658057,
		0.728373,
		0.791559,
		0.846562,
		0.892703,
		0.929706,
		0.957735,
		0.97742,
		0.989852,
		0.996543,
		0.999299,
		0.999963,
		1,
		0.999901,
		0.998875,
		0.995313,
		0.987347,
		0.973229,
		0.951543,
		0.921311,
		0.882021,
		0.833622,
		0.776495,
		0.711414,
		0.63952,
		0.56228,
		0.481465,
		0.399117,
		0.317536,
		0.239256,
		0.16703,
		0.103819,
		0.052777,
		0.017243,
		0.000729,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.00294,
		0.025016,
		0.065569,
		0.121013,
		0.18795,
		0.263174,
		0.343689,
		0.426716,
		0.509713,
		0.59039,
		0.66673,
		0.737013,
		0.799844,
		0.854185,
		0.899393,
		0.935251,
		0.962012,
		0.980411,
		0.991668,
		0.997422,
		0.999575,
		0.99999,
		0.999999,
		0.999776,
		0.998222,
		0.993551,
		0.983837,
		0.967356,
		0.942774,
		0.909232,
		0.866359,
		0.814257,
		0.753466,
		0.684929,
		0.609954,
		0.530184,
		0.447562,
		0.364313,
		0.282919,
		0.2061,
		0.136799,
		0.07817,
		0.033565,
		0.006524,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.003007,
		0.025566,
		0.066963,
		0.123492,
		0.191643,
		0.268109,
		0.349803,
		0.433865,
		0.517687,
		0.598922,
		0.675515,
		0.745722,
		0.808146,
		0.861767,
		0.90598,
		0.94064,
		0.96609,
		0.983187,
		0.993282,
		0.998145,
		0.999766,
		0.999999,
		0.999989,
		0.999556,
		0.997312,
		0.991323,
		0.97963,
		0.960553,
		0.932861,
		0.895825,
		0.849227,
		0.793332,
		0.728851,
		0.656903,
		0.578973,
		0.496887,
		0.412775,
		0.32905,
		0.248386,
		0.1737,
		0.108136,
		0.055053,
		0.018011,
		0.000762,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.003075,
		0.026134,
		0.068402,
		0.126047,
		0.195443,
		0.273179,
		0.356071,
		0.441178,
		0.52582,
		0.607595,
		0.684409,
		0.754496,
		0.816458,
		0.869296,
		0.912454,
		0.94586,
		0.969962,
		0.985742,
		0.994696,
		0.998722,
		0.999887,
		1,
		0.999957,
		0.999202,
		0.996084,
		0.988552,
		0.974637,
		0.952728,
		0.921711,
		0.881004,
		0.830552,
		0.770794,
		0.702622,
		0.627338,
		0.546619,
		0.462481,
		0.37725,
		0.293538,
		0.214224,
		0.142431,
		0.081516,
		0.035053,
		0.006822,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.003147,
		0.026722,
		0.069887,
		0.128681,
		0.199356,
		0.278389,
		0.362499,
		0.448658,
		0.534114,
		0.61641,
		0.69341,
		0.763329,
		0.82477,
		0.876762,
		0.9188,
		0.950898,
		0.973615,
		0.988073,
		0.995914,
		0.999166,
		0.999956,
		1,
		0.999882,
		0.998665,
		0.994468,
		0.985151,
		0.968764,
		0.943785,
		0.909232,
		0.864686,
		0.810266,
		0.746597,
		0.67476,
		0.596255,
		0.512956,
		0.427082,
		0.341166,
		0.258031,
		0.18077,
		0.112726,
		0.057479,
		0.018832,
		0.000798,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.003221,
		0.027329,
		0.071421,
		0.131398,
		0.203384,
		0.283743,
		0.369091,
		0.45631,
		0.542574,
		0.625367,
		0.702516,
		0.772216,
		0.833075,
		0.884152,
		0.925006,
		0.955742,
		0.977043,
		0.990178,
		0.996943,
		0.999493,
		0.999988,
		0.999998,
		0.999732,
		0.997888,
		0.992386,
		0.98103,
		0.961912,
		0.933622,
		0.895329,
		0.846787,
		0.788304,
		0.720701,
		0.645261,
		0.563688,
		0.478069,
		0.390835,
		0.304739,
		0.222825,
		0.148414,
		0.085081,
		0.036642,
		0.007142,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.003297,
		0.027957,
		0.073006,
		0.1342,
		0.207533,
		0.289247,
		0.375851,
		0.464137,
		0.5512,
		0.634466,
		0.711723,
		0.781149,
		0.84136,
		0.891454,
		0.931058,
		0.960379,
		0.980236,
		0.992055,
		0.997793,
		0.99972,
		0.999998,
		0.999987,
		0.99947,
		0.996806,
		0.989754,
		0.976091,
		0.953979,
		0.922139,
		0.879909,
		0.827229,
		0.76461,
		0.693079,
		0.614133,
		0.529695,
		0.442072,
		0.353922,
		0.268228,
		0.188271,
		0.117611,
		0.060068,
		0.01971,
		0.000836,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.003376,
		0.028608,
		0.074644,
		0.137092,
		0.211807,
		0.294906,
		0.382786,
		0.472143,
		0.559995,
		0.643707,
		0.721027,
		0.79012,
		0.849616,
		0.898654,
		0.936941,
		0.964797,
		0.983187,
		0.993706,
		0.998474,
		0.999864,
		1,
		0.999949,
		0.999046,
		0.995346,
		0.986479,
		0.97023,
		0.944857,
		0.909232,
		0.862877,
		0.805937,
		0.739133,
		0.663715,
		0.581406,
		0.494356,
		0.405109,
		0.316562,
		0.231941,
		0.154777,
		0.088884,
		0.038342,
		0.007484,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.003458,
		0.029281,
		0.076337,
		0.140077,
		0.21621,
		0.300724,
		0.3899,
		0.480333,
		0.568961,
		0.653088,
		0.730424,
		0.799123,
		0.85783,
		0.905738,
		0.942641,
		0.968983,
		0.98589,
		0.995135,
		0.999001,
		0.999946,
		1,
		0.999858,
		0.998403,
		0.993424,
		0.982461,
		0.963338,
		0.934434,
		0.894796,
		0.844144,
		0.782841,
		0.711836,
		0.632611,
		0.547127,
		0.457781,
		0.367361,
		0.279019,
		0.19624,
		0.122817,
		0.062836,
		0.020651,
		0.000878,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.003544,
		0.029978,
		0.078089,
		0.14316,
		0.22075,
		0.306709,
		0.397198,
		0.488711,
		0.5781,
		0.662608,
		0.739908,
		0.808146,
		0.86599,
		0.912691,
		0.948142,
		0.972926,
		0.98834,
		0.996348,
		0.999391,
		0.999985,
		0.999998,
		0.999678,
		0.997473,
		0.990947,
		0.977591,
		0.955299,
		0.922598,
		0.878726,
		0.823621,
		0.757879,
		0.68269,
		0.599783,
		0.511372,
		0.42011,
		0.32905,
		0.24161,
		0.161551,
		0.092945,
		0.040162,
		0.007852,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.003632,
		0.0307,
		0.079901,
		0.146343,
		0.225429,
		0.312866,
		0.404687,
		0.497281,
		0.587413,
		0.672265,
		0.749475,
		0.817181,
		0.874082,
		0.919497,
		0.953429,
		0.976614,
		0.990536,
		0.997353,
		0.999662,
		0.999998,
		0.999984,
		0.999361,
		0.996177,
		0.987816,
		0.971756,
		0.945995,
		0.909232,
		0.860918,
		0.801227,
		0.731,
		0.651685,
		0.565272,
		0.474244,
		0.381525,
		0.290446,
		0.204713,
		0.128373,
		0.065798,
		0.021661,
		0.000922,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.003724,
		0.031449,
		0.081776,
		0.149633,
		0.230255,
		0.3192,
		0.412372,
		0.506046,
		0.596902,
		0.682058,
		0.759116,
		0.826216,
		0.882091,
		0.926139,
		0.958486,
		0.980037,
		0.992476,
		0.998163,
		0.999836,
		1,
		0.999938,
		0.99885,
		0.994427,
		0.983919,
		0.964833,
		0.935301,
		0.89422,
		0.841272,
		0.776885,
		0.702164,
		0.618827,
		0.529145,
		0.43588,
		0.342249,
		0.251877,
		0.168771,
		0.097288,
		0.042115,
		0.008248,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.003819,
		0.032226,
		0.083718,
		0.153034,
		0.235233,
		0.325719,
		0.420258,
		0.515011,
		0.606567,
		0.691982,
		0.768825,
		0.835238,
		0.890001,
		0.9326,
		0.963299,
		0.983187,
		0.994162,
		0.998793,
		0.999935,
		1,
		0.999827,
		0.998073,
		0.992122,
		0.979138,
		0.956695,
		0.92309,
		0.877444,
		0.819691,
		0.750529,
		0.671346,
		0.584147,
		0.491497,
		0.39646,
		0.302559,
		0.213733,
		0.13431,
		0.068974,
		0.022747,
		0.00097,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.003918,
		0.033031,
		0.085729,
		0.156549,
		0.24037,
		0.332429,
		0.428351,
		0.52418,
		0.61641,
		0.702034,
		0.778592,
		0.844236,
		0.897796,
		0.938862,
		0.967852,
		0.986055,
		0.995601,
		0.999261,
		0.999982,
		0.999998,
		0.999608,
		0.99695,
		0.989154,
		0.973345,
		0.947206,
		0.909232,
		0.858788,
		0.796084,
		0.722107,
		0.638541,
		0.547701,
		0.452462,
		0.356209,
		0.26279,
		0.176478,
		0.101941,
		0.044213,
		0.008674,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.00402,
		0.033867,
		0.087813,
		0.160186,
		0.245671,
		0.339336,
		0.436658,
		0.533556,
		0.62643,
		0.71221,
		0.788409,
		0.853193,
		0.905457,
		0.944907,
		0.97213,
		0.988638,
		0.996799,
		0.999588,
		0.999997,
		0.999981,
		0.999223,
		0.995385,
		0.985399,
		0.966404,
		0.936229,
		0.893596,
		0.838141,
		0.770369,
		0.691577,
		0.603767,
		0.509577,
		0.412216,
		0.315408,
		0.223347,
		0.140662,
		0.072383,
		0.023917,
		0.001021,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.004127,
		0.034735,
		0.089974,
		0.163949,
		0.251144,
		0.346449,
		0.445185,
		0.543144,
		0.636628,
		0.722505,
		0.798264,
		0.862095,
		0.912966,
		0.950715,
		0.97612,
		0.990931,
		0.997769,
		0.999799,
		1,
		0.999924,
		0.9986,
		0.99327,
		0.980727,
		0.958172,
		0.92362,
		0.87605,
		0.815394,
		0.742476,
		0.65892,
		0.567067,
		0.469901,
		0.370984,
		0.2744,
		0.184715,
		0.106933,
		0.046472,
		0.009134,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.004238,
		0.035637,
		0.092214,
		0.167843,
		0.256796,
		0.353774,
		0.453937,
		0.552945,
		0.647002,
		0.732912,
		0.808146,
		0.870925,
		0.920303,
		0.956268,
		0.97981,
		0.992935,
		0.998528,
		0.99992,
		1,
		0.999788,
		0.997655,
		0.990485,
		0.974996,
		0.948498,
		0.909232,
		0.856465,
		0.790449,
		0.712348,
		0.624138,
		0.528519,
		0.428844,
		0.32905,
		0.233606,
		0.14747,
		0.076051,
		0.025179,
		0.001077,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.004354,
		0.036574,
		0.094539,
		0.171875,
		0.262634,
		0.361318,
		0.462921,
		0.562964,
		0.657551,
		0.743424,
		0.818041,
		0.879663,
		0.927446,
		0.961546,
		0.983187,
		0.994653,
		0.999094,
		0.999977,
		0.999997,
		0.999519,
		0.996286,
		0.986896,
		0.968053,
		0.937225,
		0.892918,
		0.834714,
		0.763216,
		0.679948,
		0.587261,
		0.488241,
		0.38663,
		0.286766,
		0.193529,
		0.112297,
		0.048908,
		0.009632,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.004475,
		0.037549,
		0.096952,
		0.176052,
		0.268666,
		0.369091,
		0.472143,
		0.573204,
		0.668272,
		0.754033,
		0.827935,
		0.888291,
		0.934374,
		0.966531,
		0.986243,
		0.996092,
		0.999493,
		0.999997,
		0.999976,
		0.999046,
		0.994377,
		0.982358,
		0.959736,
		0.924192,
		0.874529,
		0.81068,
		0.733619,
		0.645261,
		0.548355,
		0.446396,
		0.343545,
		0.244566,
		0.154777,
		0.080003,
		0.026543,
		0.001138,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.0046,
		0.038563,
		0.099458,
		0.18038,
		0.2749,
		0.377099,
		0.48161,
		0.583667,
		0.679164,
		0.764729,
		0.837812,
		0.896786,
		0.941064,
		0.971204,
		0.988971,
		0.997265,
		0.999752,
		1,
		0.999906,
		0.99828,
		0.991798,
		0.976711,
		0.949876,
		0.909232,
		0.85392,
		0.78425,
		0.701601,
		0.608301,
		0.507526,
		0.403207,
		0.299949,
		0.202976,
		0.118071,
		0.051539,
		0.010171,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.004731,
		0.039618,
		0.102062,
		0.184867,
		0.281345,
		0.385352,
		0.491327,
		0.594354,
		0.690221,
		0.775502,
		0.847654,
		0.905127,
		0.947493,
		0.975548,
		0.991369,
		0.998186,
		0.9999,
		1,
		0.999738,
		0.997117,
		0.988401,
		0.969784,
		0.938296,
		0.892179,
		0.830949,
		0.755329,
		0.667127,
		0.569117,
		0.464929,
		0.358959,
		0.25629,
		0.162631,
		0.084269,
		0.028021,
		0.001204,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.004868,
		0.040718,
		0.104768,
		0.18952,
		0.288011,
		0.393858,
		0.501302,
		0.605268,
		0.701439,
		0.786339,
		0.85744,
		0.913289,
		0.953636,
		0.979547,
		0.993437,
		0.998879,
		0.999972,
		0.999996,
		0.999403,
		0.995431,
		0.984024,
		0.961395,
		0.924811,
		0.872864,
		0.805484,
		0.723836,
		0.630193,
		0.527802,
		0.420779,
		0.314019,
		0.213114,
		0.124297,
		0.054388,
		0.010757,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.00501,
		0.041863,
		0.107582,
		0.194346,
		0.294906,
		0.402626,
		0.511539,
		0.61641,
		0.712812,
		0.797225,
		0.867151,
		0.921246,
		0.959469,
		0.983187,
		0.99518,
		0.999369,
		0.999996,
		0.99997,
		0.998815,
		0.993081,
		0.978488,
		0.951351,
		0.909232,
		0.851122,
		0.777401,
		0.689715,
		0.590828,
		0.484498,
		0.375363,
		0.268846,
		0.171089,
		0.088884,
		0.029626,
		0.001275,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.005159,
		0.043058,
		0.110511,
		0.199356,
		0.30204,
		0.411665,
		0.522046,
		0.62778,
		0.724333,
		0.808146,
		0.876762,
		0.928972,
		0.964969,
		0.986456,
		0.996609,
		0.999689,
		1,
		0.999882,
		0.997863,
		0.989904,
		0.971601,
		0.939451,
		0.89137,
		0.826794,
		0.746597,
		0.652936,
		0.549108,
		0.439414,
		0.32905,
		0.22401,
		0.131022,
		0.057479,
		0.011395,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005315,
		0.044304,
		0.113559,
		0.204557,
		0.309424,
		0.420985,
		0.532828,
		0.639377,
		0.735993,
		0.819083,
		0.886248,
		0.936437,
		0.970112,
		0.989348,
		0.99774,
		0.999875,
		1,
		0.999671,
		0.996415,
		0.985719,
		0.963156,
		0.925483,
		0.871032,
		0.799732,
		0.712986,
		0.613507,
		0.505161,
		0.392831,
		0.282312,
		0.180211,
		0.093885,
		0.031373,
		0.001354,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.005478,
		0.045605,
		0.116734,
		0.209959,
		0.317068,
		0.430596,
		0.54389,
		0.651201,
		0.747781,
		0.830017,
		0.895582,
		0.943612,
		0.974876,
		0.991857,
		0.998595,
		0.999964,
		0.999995,
		0.999251,
		0.994317,
		0.980326,
		0.952932,
		0.909232,
		0.848031,
		0.769799,
		0.676513,
		0.571482,
		0.459182,
		0.345123,
		0.23574,
		0.138302,
		0.06084,
		0.012092,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005649,
		0.046964,
		0.120042,
		0.215573,
		0.324985,
		0.440506,
		0.555238,
		0.663248,
		0.759685,
		0.840925,
		0.904734,
		0.950468,
		0.97924,
		0.993987,
		0.999205,
		0.999995,
		0.999962,
		0.998511,
		0.99139,
		0.973507,
		0.940699,
		0.89048,
		0.822187,
		0.736881,
		0.63716,
		0.526971,
		0.411446,
		0.29677,
		0.190067,
		0.099316,
		0.033279,
		0.00144,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.005827,
		0.048385,
		0.123492,
		0.221409,
		0.333186,
		0.450728,
		0.566876,
		0.675515,
		0.771692,
		0.851782,
		0.913673,
		0.956972,
		0.983187,
		0.995745,
		0.999606,
		1,
		0.99985,
		0.997312,
		0.987433,
		0.965026,
		0.926215,
		0.869007,
		0.793332,
		0.700893,
		0.594955,
		0.480153,
		0.362328,
		0.248386,
		0.146196,
		0.064503,
		0.012854,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.006014,
		0.04987,
		0.127091,
		0.227479,
		0.341684,
		0.461269,
		0.578809,
		0.687995,
		0.783783,
		0.862561,
		0.922363,
		0.963094,
		0.9867,
		0.997146,
		0.999839,
		0.999998,
		0.99958,
		0.995487,
		0.982219,
		0.954629,
		0.909231,
		0.844598,
		0.761316,
		0.661781,
		0.549981,
		0.431293,
		0.312315,
		0.200736,
		0.105229,
		0.035363,
		0.001534,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.006211,
		0.051425,
		0.130845,
		0.233791,
		0.350485,
		0.472135,
		0.591031,
		0.700674,
		0.795928,
		0.873218,
		0.930757,
		0.968789,
		0.989757,
		0.998198,
		0.999936,
		0.999976,
		0.999028,
		0.992826,
		0.975486,
		0.942036,
		0.889481,
		0.817037,
		0.726004,
		0.619534,
		0.502393,
		0.380753,
		0.262037,
		0.154774,
		0.068505,
		0.013691,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.006416,
		0.053048,
		0.134756,
		0.240343,
		0.359581,
		0.483303,
		0.603506,
		0.713495,
		0.808054,
		0.883665,
		0.938756,
		0.973965,
		0.99229,
		0.998871,
		0.99988,
		0.999837,
		0.997988,
		0.989042,
		0.966903,
		0.926912,
		0.86666,
		0.786085,
		0.687268,
		0.574175,
		0.452411,
		0.329012,
		0.212284,
		0.111669,
		0.037644,
		0.001637,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.00663,
		0.054735,
		0.138805,
		0.247099,
		0.368915,
		0.494688,
		0.616116,
		0.726308,
		0.819979,
		0.893693,
		0.94614,
		0.97841,
		0.994116,
		0.999017,
		0.999523,
		0.99933,
		0.996096,
		0.983695,
		0.956001,
		0.908799,
		0.840368,
		0.751442,
		0.644953,
		0.525747,
		0.400334,
		0.276679,
		0.164044,
		0.072856,
		0.014606,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.006851,
		0.056465,
		0.142941,
		0.253963,
		0.378334,
		0.50608,
		0.628589,
		0.73878,
		0.831316,
		0.902878,
		0.952465,
		0.981687,
		0.994828,
		0.998268,
		0.998474,
		0.997934,
		0.992725,
		0.9761,
		0.942087,
		0.887051,
		0.810056,
		0.712711,
		0.598868,
		0.474322,
		0.346557,
		0.224539,
		0.118558,
		0.040101,
		0.001749,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.007072,
		0.058196,
		0.147054,
		0.260738,
		0.38754,
		0.517072,
		0.640413,
		0.750302,
		0.841376,
		0.910471,
		0.956957,
		0.983029,
		0.993699,
		0.99593,
		0.995982,
		0.994761,
		0.986889,
		0.965238,
		0.924176,
		0.860781,
		0.774998,
		0.66939,
		0.548801,
		0.420042,
		0.291638,
		0.173627,
		0.077397,
		0.015567,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.007286,
		0.059855,
		0.15096,
		0.267095,
		0.396042,
		0.527004,
		0.650766,
		0.75991,
		0.849081,
		0.91532,
		0.95843,
		0.981267,
		0.989607,
		0.990906,
		0.99085,
		0.988479,
		0.977186,
		0.949718,
		0.900971,
		0.828867,
		0.734326,
		0.620939,
		0.494607,
		0.36322,
		0.236406,
		0.125331,
		0.042546,
		0.001862,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.00748,
		0.061335,
		0.154387,
		0.272552,
		0.403127,
		0.534931,
		0.658487,
		0.766258,
		0.852944,
		0.915851,
		0.955287,
		0.974831,
		0.981036,
		0.981687,
		0.981438,
		0.977327,
		0.961826,
		0.927827,
		0.870942,
		0.790059,
		0.687158,
		0.566923,
		0.436364,
		0.304496,
		0.182095,
		0.081496,
		0.01645,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.007635,
		0.062493,
		0.156977,
		0.276483,
		0.407876,
		0.539646,
		0.662108,
		0.767663,
		0.851131,
		0.910151,
		0.945607,
		0.961852,
		0.966179,
		0.966443,
		0.965756,
		0.95923,
		0.938773,
		0.897699,
		0.832516,
		0.743188,
		0.632819,
		0.507231,
		0.374572,
		0.245006,
		0.130464,
		0.044463,
		0.001953,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.007731,
		0.06316,
		0.158311,
		0.278157,
		0.409225,
		0.539764,
		0.659966,
		0.76225,
		0.841631,
		0.896157,
		0.92735,
		0.940363,
		0.943133,
		0.943198,
		0.941677,
		0.932032,
		0.906002,
		0.857596,
		0.784374,
		0.687459,
		0.57112,
		0.442328,
		0.310362,
		0.186511,
		0.083836,
		0.016988,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.007746,
		0.063155,
		0.157944,
		0.276811,
		0.406079,
		0.533882,
		0.650402,
		0.748179,
		0.822512,
		0.871933,
		0.898653,
		0.908605,
		0.910182,
		0.910115,
		0.907247,
		0.893836,
		0.861859,
		0.806267,
		0.7258,
		0.622781,
		0.502643,
		0.373469,
		0.245614,
		0.131419,
		0.04498,
		0.001983,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.007659,
		0.062312,
		0.155471,
		0.271753,
		0.397462,
		0.520775,
		0.632006,
		0.723937,
		0.792248,
		0.836021,
		0.858176,
		0.865353,
		0.86612,
		0.865825,
		0.861048,
		0.843385,
		0.805445,
		0.743327,
		0.65702,
		0.550042,
		0.428931,
		0.302792,
		0.182939,
		0.08262,
		0.016812,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.007452,
		0.060501,
		0.150584,
		0.26248,
		0.382687,
		0.499619,
		0.603886,
		0.68864,
		0.750038,
		0.787765,
		0.805429,
		0.810233,
		0.810531,
		0.809757,
		0.80255,
		0.780424,
		0.736961,
		0.669547,
		0.579429,
		0.471247,
		0.352525,
		0.23323,
		0.125451,
		0.043137,
		0.00191,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.007118,
		0.057659,
		0.143141,
		0.24878,
		0.361507,
		0.470186,
		0.565889,
		0.642277,
		0.696065,
		0.727561,
		0.740998,
		0.743918,
		0.743988,
		0.742376,
		0.732363,
		0.705931,
		0.657896,
		0.586982,
		0.495623,
		0.389445,
		0.276767,
		0.168202,
		0.076359,
		0.015609,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.006658,
		0.053801,
		0.133206,
		0.230803,
		0.334211,
		0.43295,
		0.518731,
		0.585835,
		0.631602,
		0.656947,
		0.666607,
		0.66817,
		0.668102,
		0.665273,
		0.652329,
		0.622176,
		0.571013,
		0.498862,
		0.409197,
		0.308422,
		0.205406,
		0.111127,
		0.038407,
		0.001708,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.006083,
		0.049033,
		0.121059,
		0.209081,
		0.30164,
		0.389109,
		0.463986,
		0.52127,
		0.558952,
		0.578504,
		0.584989,
		0.585696,
		0.585423,
		0.581082,
		0.565401,
		0.532534,
		0.48009,
		0.40925,
		0.324316,
		0.232205,
		0.142042,
		0.064851,
		0.013323,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.005415,
		0.043541,
		0.107181,
		0.184485,
		0.265117,
		0.34047,
		0.403941,
		0.451312,
		0.481216,
		0.495581,
		0.499583,
		0.499836,
		0.499182,
		0.493212,
		0.475331,
		0.441107,
		0.389462,
		0.322525,
		0.24515,
		0.164471,
		0.089552,
		0.031123,
		0.001391,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.004685,
		0.037571,
		0.092199,
		0.158129,
		0.226305,
		0.289251,
		0.341342,
		0.379157,
		0.401925,
		0.4119,
		0.414121,
		0.414176,
		0.412943,
		0.405462,
		0.386226,
		0.352213,
		0.303466,
		0.242785,
		0.175283,
		0.108001,
		0.04962,
		0.01025,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.003928,
		0.031406,
		0.076818,
		0.131251,
		0.187018,
		0.237836,
		0.279078,
		0.308099,
		0.32464,
		0.331119,
		0.332194,
		0.332146,
		0.330213,
		0.321588,
		0.302048,
		0.269848,
		0.22588,
		0.17332,
		0.11724,
		0.064291,
		0.022482,
		0.00101,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.003178,
		0.025332,
		0.061746,
		0.105078,
		0.14903,
		0.188515,
		0.219874,
		0.241175,
		0.252561,
		0.256446,
		0.25688,
		0.256712,
		0.254094,
		0.244896,
		0.226175,
		0.197224,
		0.159489,
		0.116229,
		0.072196,
		0.033402,
		0.006942,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.002468,
		0.019612,
		0.047631,
		0.080714,
		0.113912,
		0.143273,
		0.166038,
		0.180892,
		0.188251,
		0.190361,
		0.190498,
		0.190146,
		0.18702,
		0.177929,
		0.161065,
		0.136466,
		0.105836,
		0.072255,
		0.039939,
		0.014062,
		0.000636,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.001827,
		0.014469,
		0.035006,
		0.059053,
		0.082904,
		0.103642,
		0.119294,
		0.129044,
		0.133455,
		0.134466,
		0.134491,
		0.133926,
		0.130593,
		0.122288,
		0.108085,
		0.088475,
		0.065163,
		0.040846,
		0.019045,
		0.003985,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001275,
		0.010067,
		0.024257,
		0.040725,
		0.056853,
		0.070615,
		0.080692,
		0.086641,
		0.08905,
		0.089462,
		0.089443,
		0.088707,
		0.085527,
		0.078559,
		0.067482,
		0.052975,
		0.036547,
		0.020382,
		0.007231,
		0.000329,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.000827,
		0.006505,
		0.015607,
		0.026068,
		0.036175,
		0.04462,
		0.050596,
		0.053909,
		0.055078,
		0.055213,
		0.05516,
		0.054366,
		0.051666,
		0.046365,
		0.038483,
		0.028687,
		0.018168,
		0.008546,
		0.001801,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000487,
		0.003813,
		0.009106,
		0.015128,
		0.020858,
		0.025537,
		0.02872,
		0.030358,
		0.030845,
		0.030877,
		0.030794,
		0.03008,
		0.028075,
		0.024496,
		0.019499,
		0.013614,
		0.007669,
		0.002744,
		0.000126,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.00025,
		0.001951,
		0.004637,
		0.007659,
		0.010488,
		0.012738,
		0.014202,
		0.014891,
		0.015054,
		0.015059,
		0.014966,
		0.014441,
		0.013182,
		0.011116,
		0.008402,
		0.005384,
		0.002558,
		0.000544,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000105,
		0.000815,
		0.001926,
		0.003161,
		0.004297,
		0.005174,
		0.005716,
		0.005944,
		0.005985,
		0.005983,
		0.005911,
		0.005612,
		0.004985,
		0.004032,
		0.002854,
		0.001626,
		0.000588,
		2.7e-005,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		3.1e-005,
		0.000242,
		0.000568,
		0.000927,
		0.001249,
		0.001491,
		0.001631,
		0.001682,
		0.001689,
		0.001686,
		0.001651,
		0.001535,
		0.001319,
		0.001013,
		0.000658,
		0.000316,
		6.8e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		3.6e-005,
		8.3e-005,
		0.000135,
		0.00018,
		0.000213,
		0.000231,
		0.000236,
		0.000237,
		0.000236,
		0.000228,
		0.000206,
		0.00017,
		0.000122,
		7.1e-005,
		2.6e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		2.6e-005,
		7.1e-005,
		0.000122,
		0.00017,
		0.000206,
		0.000228,
		0.000236,
		0.000237,
		0.000236,
		0.000231,
		0.000213,
		0.00018,
		0.000135,
		8.3e-005,
		3.6e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		6.8e-005,
		0.000316,
		0.000658,
		0.001013,
		0.001319,
		0.001535,
		0.001651,
		0.001686,
		0.001689,
		0.001682,
		0.001631,
		0.001491,
		0.001249,
		0.000927,
		0.000568,
		0.000242,
		3.1e-005,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		2.7e-005,
		0.000588,
		0.001626,
		0.002854,
		0.004032,
		0.004985,
		0.005612,
		0.005911,
		0.005983,
		0.005985,
		0.005944,
		0.005716,
		0.005174,
		0.004297,
		0.003161,
		0.001926,
		0.000815,
		0.000105,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000544,
		0.002558,
		0.005384,
		0.008402,
		0.011116,
		0.013182,
		0.014441,
		0.014966,
		0.015059,
		0.015054,
		0.014891,
		0.014202,
		0.012738,
		0.010488,
		0.007659,
		0.004637,
		0.001951,
		0.00025,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		0.000126,
		0.002744,
		0.007669,
		0.013614,
		0.019499,
		0.024496,
		0.028075,
		0.03008,
		0.030794,
		0.030877,
		0.030845,
		0.030358,
		0.02872,
		0.025537,
		0.020858,
		0.015128,
		0.009106,
		0.003813,
		0.000487,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.001801,
		0.008546,
		0.018168,
		0.028687,
		0.038483,
		0.046365,
		0.051666,
		0.054366,
		0.05516,
		0.055213,
		0.055078,
		0.053909,
		0.050596,
		0.04462,
		0.036175,
		0.026068,
		0.015607,
		0.006505,
		0.000827,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.000329,
		0.007231,
		0.020382,
		0.036547,
		0.052975,
		0.067482,
		0.078559,
		0.085527,
		0.088707,
		0.089443,
		0.089462,
		0.08905,
		0.086641,
		0.080692,
		0.070615,
		0.056853,
		0.040725,
		0.024257,
		0.010067,
		0.001275,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.003985,
		0.019045,
		0.040846,
		0.065163,
		0.088475,
		0.108085,
		0.122288,
		0.130593,
		0.133926,
		0.134491,
		0.134466,
		0.133455,
		0.129044,
		0.119294,
		0.103642,
		0.082904,
		0.059053,
		0.035006,
		0.014469,
		0.001827,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.000636,
		0.014062,
		0.039939,
		0.072255,
		0.105836,
		0.136466,
		0.161065,
		0.177929,
		0.18702,
		0.190146,
		0.190498,
		0.190361,
		0.188251,
		0.180892,
		0.166038,
		0.143273,
		0.113912,
		0.080714,
		0.047631,
		0.019612,
		0.002468,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.006942,
		0.033402,
		0.072196,
		0.116229,
		0.159489,
		0.197224,
		0.226175,
		0.244896,
		0.254094,
		0.256712,
		0.25688,
		0.256446,
		0.252561,
		0.241175,
		0.219874,
		0.188515,
		0.14903,
		0.105078,
		0.061746,
		0.025332,
		0.003178,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.00101,
		0.022482,
		0.064291,
		0.11724,
		0.17332,
		0.22588,
		0.269848,
		0.302048,
		0.321588,
		0.330213,
		0.332146,
		0.332194,
		0.331119,
		0.32464,
		0.308099,
		0.279078,
		0.237836,
		0.187018,
		0.131251,
		0.076818,
		0.031406,
		0.003928,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.01025,
		0.04962,
		0.108001,
		0.175283,
		0.242785,
		0.303466,
		0.352213,
		0.386226,
		0.405462,
		0.412943,
		0.414176,
		0.414121,
		0.4119,
		0.401925,
		0.379157,
		0.341342,
		0.289251,
		0.226305,
		0.158129,
		0.092199,
		0.037571,
		0.004685,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.001391,
		0.031123,
		0.089552,
		0.164471,
		0.24515,
		0.322525,
		0.389462,
		0.441107,
		0.475331,
		0.493212,
		0.499182,
		0.499836,
		0.499583,
		0.495581,
		0.481216,
		0.451312,
		0.403941,
		0.34047,
		0.265117,
		0.184485,
		0.107181,
		0.043541,
		0.005415,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.013323,
		0.064851,
		0.142042,
		0.232205,
		0.324316,
		0.40925,
		0.48009,
		0.532534,
		0.565401,
		0.581082,
		0.585423,
		0.585696,
		0.584989,
		0.578504,
		0.558952,
		0.52127,
		0.463986,
		0.389109,
		0.30164,
		0.209081,
		0.121059,
		0.049033,
		0.006083,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.001708,
		0.038407,
		0.111127,
		0.205406,
		0.308422,
		0.409197,
		0.498862,
		0.571013,
		0.622176,
		0.652329,
		0.665273,
		0.668102,
		0.66817,
		0.666607,
		0.656947,
		0.631602,
		0.585835,
		0.518731,
		0.43295,
		0.334211,
		0.230803,
		0.133206,
		0.053801,
		0.006658,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.015609,
		0.076359,
		0.168202,
		0.276767,
		0.389445,
		0.495623,
		0.586982,
		0.657896,
		0.705931,
		0.732363,
		0.742376,
		0.743988,
		0.743918,
		0.740998,
		0.727561,
		0.696065,
		0.642277,
		0.565889,
		0.470186,
		0.361507,
		0.24878,
		0.143141,
		0.057659,
		0.007118,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.00191,
		0.043137,
		0.125451,
		0.23323,
		0.352525,
		0.471247,
		0.579429,
		0.669547,
		0.736961,
		0.780424,
		0.80255,
		0.809757,
		0.810531,
		0.810233,
		0.805429,
		0.787765,
		0.750038,
		0.68864,
		0.603886,
		0.499619,
		0.382687,
		0.26248,
		0.150584,
		0.060501,
		0.007452,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.016812,
		0.08262,
		0.182939,
		0.302792,
		0.428931,
		0.550042,
		0.65702,
		0.743327,
		0.805445,
		0.843385,
		0.861048,
		0.865825,
		0.86612,
		0.865353,
		0.858176,
		0.836021,
		0.792248,
		0.723937,
		0.632006,
		0.520775,
		0.397462,
		0.271753,
		0.155471,
		0.062312,
		0.007659,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.001983,
		0.04498,
		0.131419,
		0.245614,
		0.373469,
		0.502643,
		0.622781,
		0.7258,
		0.806267,
		0.861859,
		0.893836,
		0.907247,
		0.910115,
		0.910182,
		0.908605,
		0.898653,
		0.871933,
		0.822512,
		0.748179,
		0.650402,
		0.533882,
		0.406079,
		0.276811,
		0.157944,
		0.063155,
		0.007746,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.016988,
		0.083836,
		0.186511,
		0.310362,
		0.442328,
		0.57112,
		0.687459,
		0.784374,
		0.857596,
		0.906002,
		0.932032,
		0.941677,
		0.943198,
		0.943133,
		0.940363,
		0.92735,
		0.896157,
		0.841631,
		0.76225,
		0.659966,
		0.539764,
		0.409225,
		0.278157,
		0.158311,
		0.06316,
		0.007731,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.001953,
		0.044463,
		0.130464,
		0.245006,
		0.374572,
		0.507231,
		0.632819,
		0.743188,
		0.832516,
		0.897699,
		0.938773,
		0.95923,
		0.965756,
		0.966443,
		0.966179,
		0.961852,
		0.945607,
		0.910151,
		0.851131,
		0.767663,
		0.662108,
		0.539646,
		0.407876,
		0.276483,
		0.156977,
		0.062493,
		0.007635,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.01645,
		0.081496,
		0.182095,
		0.304496,
		0.436364,
		0.566923,
		0.687158,
		0.790059,
		0.870942,
		0.927827,
		0.961826,
		0.977327,
		0.981438,
		0.981687,
		0.981036,
		0.974831,
		0.955287,
		0.915851,
		0.852944,
		0.766258,
		0.658487,
		0.534931,
		0.403127,
		0.272552,
		0.154387,
		0.061335,
		0.00748,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.001862,
		0.042546,
		0.125331,
		0.236406,
		0.36322,
		0.494607,
		0.620939,
		0.734326,
		0.828867,
		0.900971,
		0.949718,
		0.977186,
		0.988479,
		0.99085,
		0.990906,
		0.989607,
		0.981267,
		0.95843,
		0.91532,
		0.849081,
		0.75991,
		0.650766,
		0.527004,
		0.396042,
		0.267095,
		0.15096,
		0.059855,
		0.007286,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.015567,
		0.077397,
		0.173627,
		0.291638,
		0.420042,
		0.548801,
		0.66939,
		0.774998,
		0.860781,
		0.924176,
		0.965238,
		0.986889,
		0.994761,
		0.995982,
		0.99593,
		0.993699,
		0.983029,
		0.956957,
		0.910471,
		0.841376,
		0.750302,
		0.640413,
		0.517072,
		0.38754,
		0.260738,
		0.147054,
		0.058196,
		0.007072,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.001749,
		0.040101,
		0.118558,
		0.224539,
		0.346557,
		0.474322,
		0.598868,
		0.712711,
		0.810056,
		0.887051,
		0.942087,
		0.9761,
		0.992725,
		0.997934,
		0.998474,
		0.998268,
		0.994828,
		0.981687,
		0.952465,
		0.902878,
		0.831316,
		0.73878,
		0.628589,
		0.50608,
		0.378334,
		0.253963,
		0.142941,
		0.056465,
		0.006851,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.014606,
		0.072856,
		0.164044,
		0.276679,
		0.400334,
		0.525747,
		0.644953,
		0.751442,
		0.840368,
		0.908799,
		0.956001,
		0.983695,
		0.996096,
		0.99933,
		0.999523,
		0.999017,
		0.994116,
		0.97841,
		0.94614,
		0.893693,
		0.819979,
		0.726308,
		0.616116,
		0.494688,
		0.368915,
		0.247099,
		0.138805,
		0.054735,
		0.00663,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.001637,
		0.037644,
		0.111669,
		0.212284,
		0.329012,
		0.452411,
		0.574175,
		0.687268,
		0.786085,
		0.86666,
		0.926912,
		0.966903,
		0.989042,
		0.997988,
		0.999837,
		0.99988,
		0.998871,
		0.99229,
		0.973965,
		0.938756,
		0.883665,
		0.808054,
		0.713495,
		0.603506,
		0.483303,
		0.359581,
		0.240343,
		0.134756,
		0.053048,
		0.006416,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.013691,
		0.068505,
		0.154774,
		0.262037,
		0.380753,
		0.502393,
		0.619534,
		0.726004,
		0.817037,
		0.889481,
		0.942036,
		0.975486,
		0.992826,
		0.999028,
		0.999976,
		0.999936,
		0.998198,
		0.989757,
		0.968789,
		0.930757,
		0.873218,
		0.795928,
		0.700674,
		0.591031,
		0.472135,
		0.350485,
		0.233791,
		0.130845,
		0.051425,
		0.006211,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.001534,
		0.035363,
		0.105229,
		0.200736,
		0.312315,
		0.431293,
		0.549981,
		0.661781,
		0.761316,
		0.844598,
		0.909231,
		0.954629,
		0.982219,
		0.995487,
		0.99958,
		0.999998,
		0.999839,
		0.997146,
		0.9867,
		0.963094,
		0.922363,
		0.862561,
		0.783783,
		0.687995,
		0.578809,
		0.461269,
		0.341684,
		0.227479,
		0.127091,
		0.04987,
		0.006014,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.012854,
		0.064503,
		0.146196,
		0.248386,
		0.362328,
		0.480153,
		0.594955,
		0.700893,
		0.793332,
		0.869007,
		0.926215,
		0.965026,
		0.987433,
		0.997312,
		0.99985,
		1,
		0.999606,
		0.995745,
		0.983187,
		0.956972,
		0.913673,
		0.851782,
		0.771692,
		0.675515,
		0.566876,
		0.450728,
		0.333186,
		0.221409,
		0.123492,
		0.048385,
		0.005827,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.00144,
		0.033279,
		0.099316,
		0.190067,
		0.29677,
		0.411446,
		0.526971,
		0.63716,
		0.736881,
		0.822187,
		0.89048,
		0.940699,
		0.973507,
		0.99139,
		0.998511,
		0.999962,
		0.999995,
		0.999205,
		0.993987,
		0.97924,
		0.950468,
		0.904734,
		0.840925,
		0.759685,
		0.663248,
		0.555238,
		0.440506,
		0.324985,
		0.215573,
		0.120042,
		0.046964,
		0.005649,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.012092,
		0.06084,
		0.138302,
		0.23574,
		0.345123,
		0.459182,
		0.571482,
		0.676513,
		0.769799,
		0.848031,
		0.909232,
		0.952932,
		0.980326,
		0.994317,
		0.999251,
		0.999995,
		0.999964,
		0.998595,
		0.991857,
		0.974876,
		0.943612,
		0.895582,
		0.830017,
		0.747781,
		0.651201,
		0.54389,
		0.430596,
		0.317068,
		0.209959,
		0.116734,
		0.045605,
		0.005478,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.001354,
		0.031373,
		0.093885,
		0.180211,
		0.282312,
		0.392831,
		0.505161,
		0.613507,
		0.712986,
		0.799732,
		0.871032,
		0.925483,
		0.963156,
		0.985719,
		0.996415,
		0.999671,
		1,
		0.999875,
		0.99774,
		0.989348,
		0.970112,
		0.936437,
		0.886248,
		0.819083,
		0.735993,
		0.639377,
		0.532828,
		0.420985,
		0.309424,
		0.204557,
		0.113559,
		0.044304,
		0.005315,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.011395,
		0.057479,
		0.131022,
		0.22401,
		0.32905,
		0.439414,
		0.549108,
		0.652936,
		0.746597,
		0.826794,
		0.89137,
		0.939451,
		0.971601,
		0.989904,
		0.997863,
		0.999882,
		1,
		0.999689,
		0.996609,
		0.986456,
		0.964969,
		0.928972,
		0.876762,
		0.808146,
		0.724333,
		0.62778,
		0.522046,
		0.411665,
		0.30204,
		0.199356,
		0.110511,
		0.043058,
		0.005159,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.001275,
		0.029626,
		0.088884,
		0.171089,
		0.268846,
		0.375363,
		0.484498,
		0.590828,
		0.689715,
		0.777401,
		0.851122,
		0.909232,
		0.951351,
		0.978488,
		0.993081,
		0.998815,
		0.99997,
		0.999996,
		0.999369,
		0.99518,
		0.983187,
		0.959469,
		0.921246,
		0.867151,
		0.797225,
		0.712812,
		0.61641,
		0.511539,
		0.402626,
		0.294906,
		0.194346,
		0.107582,
		0.041863,
		0.00501,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.010757,
		0.054388,
		0.124297,
		0.213114,
		0.314019,
		0.420779,
		0.527802,
		0.630193,
		0.723836,
		0.805484,
		0.872864,
		0.924811,
		0.961395,
		0.984024,
		0.995431,
		0.999403,
		0.999996,
		0.999972,
		0.998879,
		0.993437,
		0.979547,
		0.953636,
		0.913289,
		0.85744,
		0.786339,
		0.701439,
		0.605268,
		0.501302,
		0.393858,
		0.288011,
		0.18952,
		0.104768,
		0.040718,
		0.004868,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.001204,
		0.028021,
		0.084269,
		0.162631,
		0.25629,
		0.358959,
		0.464929,
		0.569117,
		0.667127,
		0.755329,
		0.830949,
		0.892179,
		0.938296,
		0.969784,
		0.988401,
		0.997117,
		0.999738,
		1,
		0.9999,
		0.998186,
		0.991369,
		0.975548,
		0.947493,
		0.905127,
		0.847654,
		0.775502,
		0.690221,
		0.594354,
		0.491327,
		0.385352,
		0.281345,
		0.184867,
		0.102062,
		0.039618,
		0.004731,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.010171,
		0.051539,
		0.118071,
		0.202976,
		0.299949,
		0.403207,
		0.507526,
		0.608301,
		0.701601,
		0.78425,
		0.85392,
		0.909232,
		0.949876,
		0.976711,
		0.991798,
		0.99828,
		0.999906,
		1,
		0.999752,
		0.997265,
		0.988971,
		0.971204,
		0.941064,
		0.896786,
		0.837812,
		0.764729,
		0.679164,
		0.583667,
		0.48161,
		0.377099,
		0.2749,
		0.18038,
		0.099458,
		0.038563,
		0.0046,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.001138,
		0.026543,
		0.080003,
		0.154777,
		0.244566,
		0.343545,
		0.446396,
		0.548355,
		0.645261,
		0.733619,
		0.81068,
		0.874529,
		0.924192,
		0.959736,
		0.982358,
		0.994377,
		0.999046,
		0.999976,
		0.999997,
		0.999493,
		0.996092,
		0.986243,
		0.966531,
		0.934374,
		0.888291,
		0.827935,
		0.754033,
		0.668272,
		0.573204,
		0.472143,
		0.369091,
		0.268666,
		0.176052,
		0.096952,
		0.037549,
		0.004475,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.009632,
		0.048908,
		0.112297,
		0.193529,
		0.286766,
		0.38663,
		0.488241,
		0.587261,
		0.679948,
		0.763216,
		0.834714,
		0.892918,
		0.937225,
		0.968053,
		0.986896,
		0.996286,
		0.999519,
		0.999997,
		0.999977,
		0.999094,
		0.994653,
		0.983187,
		0.961546,
		0.927446,
		0.879663,
		0.818041,
		0.743424,
		0.657551,
		0.562964,
		0.462921,
		0.361318,
		0.262634,
		0.171875,
		0.094539,
		0.036574,
		0.004354,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.001077,
		0.025179,
		0.076051,
		0.14747,
		0.233606,
		0.32905,
		0.428844,
		0.528519,
		0.624138,
		0.712348,
		0.790449,
		0.856465,
		0.909232,
		0.948498,
		0.974996,
		0.990485,
		0.997655,
		0.999788,
		1,
		0.99992,
		0.998528,
		0.992935,
		0.97981,
		0.956268,
		0.920303,
		0.870925,
		0.808146,
		0.732912,
		0.647002,
		0.552945,
		0.453937,
		0.353774,
		0.256796,
		0.167843,
		0.092214,
		0.035637,
		0.004238,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.009134,
		0.046472,
		0.106933,
		0.184715,
		0.2744,
		0.370984,
		0.469901,
		0.567067,
		0.65892,
		0.742476,
		0.815394,
		0.87605,
		0.92362,
		0.958172,
		0.980727,
		0.99327,
		0.9986,
		0.999924,
		1,
		0.999799,
		0.997769,
		0.990931,
		0.97612,
		0.950715,
		0.912966,
		0.862095,
		0.798264,
		0.722505,
		0.636628,
		0.543144,
		0.445185,
		0.346449,
		0.251144,
		0.163949,
		0.089974,
		0.034735,
		0.004127,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.001021,
		0.023917,
		0.072383,
		0.140662,
		0.223347,
		0.315408,
		0.412216,
		0.509577,
		0.603767,
		0.691577,
		0.770369,
		0.838141,
		0.893596,
		0.936229,
		0.966404,
		0.985399,
		0.995385,
		0.999223,
		0.999981,
		0.999997,
		0.999588,
		0.996799,
		0.988638,
		0.97213,
		0.944907,
		0.905457,
		0.853193,
		0.788409,
		0.71221,
		0.62643,
		0.533556,
		0.436658,
		0.339336,
		0.245671,
		0.160186,
		0.087813,
		0.033867,
		0.00402,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.008674,
		0.044213,
		0.101941,
		0.176478,
		0.26279,
		0.356209,
		0.452462,
		0.547701,
		0.638541,
		0.722107,
		0.796084,
		0.858788,
		0.909232,
		0.947206,
		0.973345,
		0.989154,
		0.99695,
		0.999608,
		0.999998,
		0.999982,
		0.999261,
		0.995601,
		0.986055,
		0.967852,
		0.938862,
		0.897796,
		0.844236,
		0.778592,
		0.702034,
		0.61641,
		0.52418,
		0.428351,
		0.332429,
		0.24037,
		0.156549,
		0.085729,
		0.033031,
		0.003918,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.00097,
		0.022747,
		0.068974,
		0.13431,
		0.213733,
		0.302559,
		0.39646,
		0.491497,
		0.584147,
		0.671346,
		0.750529,
		0.819691,
		0.877444,
		0.92309,
		0.956695,
		0.979138,
		0.992122,
		0.998073,
		0.999827,
		1,
		0.999935,
		0.998793,
		0.994162,
		0.983187,
		0.963299,
		0.9326,
		0.890001,
		0.835238,
		0.768825,
		0.691982,
		0.606567,
		0.515011,
		0.420258,
		0.325719,
		0.235233,
		0.153034,
		0.083718,
		0.032226,
		0.003819,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.008248,
		0.042115,
		0.097288,
		0.168771,
		0.251877,
		0.342249,
		0.43588,
		0.529145,
		0.618827,
		0.702164,
		0.776885,
		0.841272,
		0.89422,
		0.935301,
		0.964833,
		0.983919,
		0.994427,
		0.99885,
		0.999938,
		1,
		0.999836,
		0.998163,
		0.992476,
		0.980037,
		0.958486,
		0.926139,
		0.882091,
		0.826216,
		0.759116,
		0.682058,
		0.596902,
		0.506046,
		0.412372,
		0.3192,
		0.230255,
		0.149633,
		0.081776,
		0.031449,
		0.003724,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.000922,
		0.021661,
		0.065798,
		0.128373,
		0.204713,
		0.290446,
		0.381525,
		0.474244,
		0.565272,
		0.651685,
		0.731,
		0.801227,
		0.860918,
		0.909232,
		0.945995,
		0.971756,
		0.987816,
		0.996177,
		0.999361,
		0.999984,
		0.999998,
		0.999662,
		0.997353,
		0.990536,
		0.976614,
		0.953429,
		0.919497,
		0.874082,
		0.817181,
		0.749475,
		0.672265,
		0.587413,
		0.497281,
		0.404687,
		0.312866,
		0.225429,
		0.146343,
		0.079901,
		0.0307,
		0.003632,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.007852,
		0.040162,
		0.092945,
		0.161551,
		0.24161,
		0.32905,
		0.42011,
		0.511372,
		0.599783,
		0.68269,
		0.757879,
		0.823621,
		0.878726,
		0.922598,
		0.955299,
		0.977591,
		0.990947,
		0.997473,
		0.999678,
		0.999998,
		0.999985,
		0.999391,
		0.996348,
		0.98834,
		0.972926,
		0.948142,
		0.912691,
		0.86599,
		0.808146,
		0.739908,
		0.662608,
		0.5781,
		0.488711,
		0.397198,
		0.306709,
		0.22075,
		0.14316,
		0.078089,
		0.029978,
		0.003544,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.000878,
		0.020651,
		0.062836,
		0.122817,
		0.19624,
		0.279019,
		0.367361,
		0.457781,
		0.547127,
		0.632611,
		0.711836,
		0.782841,
		0.844144,
		0.894796,
		0.934434,
		0.963338,
		0.982461,
		0.993424,
		0.998403,
		0.999858,
		1,
		0.999946,
		0.999001,
		0.995135,
		0.98589,
		0.968983,
		0.942641,
		0.905738,
		0.85783,
		0.799123,
		0.730424,
		0.653088,
		0.568961,
		0.480333,
		0.3899,
		0.300724,
		0.21621,
		0.140077,
		0.076337,
		0.029281,
		0.003458,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.007484,
		0.038342,
		0.088884,
		0.154777,
		0.231941,
		0.316562,
		0.405109,
		0.494356,
		0.581406,
		0.663715,
		0.739133,
		0.805937,
		0.862877,
		0.909232,
		0.944857,
		0.97023,
		0.986479,
		0.995346,
		0.999046,
		0.999949,
		1,
		0.999864,
		0.998474,
		0.993706,
		0.983187,
		0.964797,
		0.936941,
		0.898654,
		0.849616,
		0.79012,
		0.721027,
		0.643707,
		0.559995,
		0.472143,
		0.382786,
		0.294906,
		0.211807,
		0.137092,
		0.074644,
		0.028608,
		0.003376,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.000836,
		0.01971,
		0.060068,
		0.117611,
		0.188271,
		0.268228,
		0.353922,
		0.442072,
		0.529695,
		0.614133,
		0.693079,
		0.76461,
		0.827229,
		0.879909,
		0.922139,
		0.953979,
		0.976091,
		0.989754,
		0.996806,
		0.99947,
		0.999987,
		0.999998,
		0.99972,
		0.997793,
		0.992055,
		0.980236,
		0.960379,
		0.931058,
		0.891454,
		0.84136,
		0.781149,
		0.711723,
		0.634466,
		0.5512,
		0.464137,
		0.375851,
		0.289247,
		0.207533,
		0.1342,
		0.073006,
		0.027957,
		0.003297,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.007142,
		0.036642,
		0.085081,
		0.148414,
		0.222825,
		0.304739,
		0.390835,
		0.478069,
		0.563688,
		0.645261,
		0.720701,
		0.788304,
		0.846787,
		0.895329,
		0.933622,
		0.961912,
		0.98103,
		0.992386,
		0.997888,
		0.999732,
		0.999998,
		0.999988,
		0.999493,
		0.996943,
		0.990178,
		0.977043,
		0.955742,
		0.925006,
		0.884152,
		0.833075,
		0.772216,
		0.702516,
		0.625367,
		0.542574,
		0.45631,
		0.369091,
		0.283743,
		0.203384,
		0.131398,
		0.071421,
		0.027329,
		0.003221,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.000798,
		0.018832,
		0.057479,
		0.112726,
		0.18077,
		0.258031,
		0.341166,
		0.427082,
		0.512956,
		0.596255,
		0.67476,
		0.746597,
		0.810266,
		0.864686,
		0.909232,
		0.943785,
		0.968764,
		0.985151,
		0.994468,
		0.998665,
		0.999882,
		1,
		0.999956,
		0.999166,
		0.995914,
		0.988073,
		0.973615,
		0.950898,
		0.9188,
		0.876762,
		0.82477,
		0.763329,
		0.69341,
		0.61641,
		0.534114,
		0.448658,
		0.362499,
		0.278389,
		0.199356,
		0.128681,
		0.069887,
		0.026722,
		0.003147,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.006822,
		0.035053,
		0.081516,
		0.142431,
		0.214224,
		0.293538,
		0.37725,
		0.462481,
		0.546619,
		0.627338,
		0.702622,
		0.770794,
		0.830552,
		0.881004,
		0.921711,
		0.952728,
		0.974637,
		0.988552,
		0.996084,
		0.999202,
		0.999957,
		1,
		0.999887,
		0.998722,
		0.994696,
		0.985742,
		0.969962,
		0.94586,
		0.912454,
		0.869296,
		0.816458,
		0.754496,
		0.684409,
		0.607595,
		0.52582,
		0.441178,
		0.356071,
		0.273179,
		0.195443,
		0.126047,
		0.068402,
		0.026134,
		0.003075,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.000762,
		0.018011,
		0.055053,
		0.108136,
		0.1737,
		0.248386,
		0.32905,
		0.412775,
		0.496887,
		0.578973,
		0.656903,
		0.728851,
		0.793332,
		0.849227,
		0.895825,
		0.932861,
		0.960553,
		0.97963,
		0.991323,
		0.997312,
		0.999556,
		0.999989,
		0.999999,
		0.999766,
		0.998145,
		0.993282,
		0.983187,
		0.96609,
		0.94064,
		0.90598,
		0.861767,
		0.808146,
		0.745722,
		0.675515,
		0.598922,
		0.517687,
		0.433865,
		0.349803,
		0.268109,
		0.191643,
		0.123492,
		0.066963,
		0.025566,
		0.003007,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.006524,
		0.033565,
		0.07817,
		0.136799,
		0.2061,
		0.282919,
		0.364313,
		0.447562,
		0.530184,
		0.609954,
		0.684929,
		0.753466,
		0.814257,
		0.866359,
		0.909232,
		0.942774,
		0.967356,
		0.983837,
		0.993551,
		0.998222,
		0.999776,
		0.999999,
		0.99999,
		0.999575,
		0.997422,
		0.991668,
		0.980411,
		0.962012,
		0.935251,
		0.899393,
		0.854185,
		0.799844,
		0.737013,
		0.66673,
		0.59039,
		0.509713,
		0.426716,
		0.343689,
		0.263174,
		0.18795,
		0.121013,
		0.065569,
		0.025016,
		0.00294,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.000729,
		0.017243,
		0.052777,
		0.103819,
		0.16703,
		0.239256,
		0.317536,
		0.399117,
		0.481465,
		0.56228,
		0.63952,
		0.711414,
		0.776495,
		0.833622,
		0.882021,
		0.921311,
		0.951543,
		0.973229,
		0.987347,
		0.995313,
		0.998875,
		0.999901,
		1,
		0.999963,
		0.999299,
		0.996543,
		0.989852,
		0.97742,
		0.957735,
		0.929706,
		0.892703,
		0.846562,
		0.791559,
		0.728373,
		0.658057,
		0.582,
		0.501896,
		0.419726,
		0.337724,
		0.258369,
		0.18436,
		0.118608,
		0.064217,
		0.024483,
		0.002875,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.006244,
		0.032169,
		0.075024,
		0.13149,
		0.198419,
		0.272844,
		0.35199,
		0.433283,
		0.514364,
		0.593109,
		0.667644,
		0.736367,
		0.797974,
		0.851487,
		0.896287,
		0.932145,
		0.959257,
		0.978262,
		0.990244,
		0.996681,
		0.999327,
		0.999964,
		1,
		0.999905,
		0.998921,
		0.995499,
		0.987832,
		0.974219,
		0.953272,
		0.924016,
		0.885923,
		0.838908,
		0.783299,
		0.719808,
		0.649498,
		0.573749,
		0.494234,
		0.412891,
		0.331906,
		0.25369,
		0.180871,
		0.116272,
		0.062907,
		0.023968,
		0.002813,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.000697,
		0.016522,
		0.050639,
		0.099754,
		0.160731,
		0.230606,
		0.306588,
		0.386075,
		0.466665,
		0.546167,
		0.622621,
		0.694316,
		0.759807,
		0.817948,
		0.867913,
		0.909232,
		0.94182,
		0.966004,
		0.982541,
		0.992605,
		0.997721,
		0.999626,
		0.999991,
		0.999999,
		0.999803,
		0.998429,
		0.994282,
		0.98561,
		0.970816,
		0.948631,
		0.918192,
		0.879064,
		0.83123,
		0.77507,
		0.71132,
		0.641052,
		0.565638,
		0.486723,
		0.406208,
		0.326228,
		0.249134,
		0.177478,
		0.114005,
		0.061637,
		0.023468,
		0.002753,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.005983,
		0.030859,
		0.072063,
		0.12648,
		0.191151,
		0.26328,
		0.340247,
		0.419614,
		0.499142,
		0.5768,
		0.650784,
		0.719536,
		0.781763,
		0.836467,
		0.882967,
		0.920935,
		0.950418,
		0.971868,
		0.986146,
		0.994502,
		0.998492,
		0.999811,
		0.999999,
		0.999991,
		0.999641,
		0.997811,
		0.992887,
		0.983187,
		0.967218,
		0.943825,
		0.912247,
		0.872136,
		0.823539,
		0.766878,
		0.702914,
		0.632722,
		0.557664,
		0.47936,
		0.399673,
		0.320688,
		0.244695,
		0.174178,
		0.111802,
		0.060404,
		0.022984,
		0.002694,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.000668,
		0.015846,
		0.048628,
		0.09592,
		0.154777,
		0.222405,
		0.296172,
		0.373618,
		0.452462,
		0.53062,
		0.606211,
		0.67758,
		0.743316,
		0.80227,
		0.853585,
		0.896717,
		0.931471,
		0.95802,
		0.976929,
		0.989154,
		0.996003,
		0.999046,
		0.999917,
		1,
		0.999969,
		0.999406,
		0.997056,
		0.991312,
		0.980567,
		0.963433,
		0.938862,
		0.906191,
		0.865149,
		0.815842,
		0.758729,
		0.694592,
		0.624508,
		0.549827,
		0.472143,
		0.393283,
		0.315281,
		0.24037,
		0.170967,
		0.109662,
		0.059208,
		0.022514,
		0.002638,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.005737,
		0.029626,
		0.069274,
		0.121749,
		0.184267,
		0.254193,
		0.32905,
		0.406528,
		0.484498,
		0.561021,
		0.63436,
		0.703003,
		0.765677,
		0.82137,
		0.869359,
		0.909232,
		0.940917,
		0.964706,
		0.981267,
		0.991635,
		0.997169,
		0.999429,
		0.99997,
		1,
		0.99992,
		0.999083,
		0.996156,
		0.989554,
		0.977753,
		0.959469,
		0.933754,
		0.900035,
		0.858112,
		0.808146,
		0.750628,
		0.686356,
		0.61641,
		0.542124,
		0.465069,
		0.387032,
		0.310002,
		0.236155,
		0.167843,
		0.107582,
		0.058047,
		0.022059,
		0.002583,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.000641,
		0.015211,
		0.046734,
		0.092302,
		0.149143,
		0.214622,
		0.286256,
		0.361715,
		0.438833,
		0.515624,
		0.590287,
		0.661224,
		0.727057,
		0.786646,
		0.839109,
		0.883851,
		0.920582,
		0.949349,
		0.970552,
		0.984953,
		0.993658,
		0.998056,
		0.999682,
		0.999992,
		0.999999,
		0.999833,
		0.998662,
		0.995103,
		0.987613,
		0.974752,
		0.955335,
		0.928511,
		0.893788,
		0.851035,
		0.800458,
		0.742579,
		0.67821,
		0.608428,
		0.534555,
		0.458135,
		0.380919,
		0.30485,
		0.232048,
		0.164802,
		0.105561,
		0.05692,
		0.021618,
		0.00253,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.005506,
		0.028466,
		0.066642,
		0.117276,
		0.177742,
		0.245554,
		0.318369,
		0.393997,
		0.470412,
		0.545761,
		0.618377,
		0.686792,
		0.749757,
		0.806258,
		0.855538,
		0.897121,
		0.930835,
		0.956839,
		0.975631,
		0.988059,
		0.995285,
		0.998713,
		0.999839,
		0.999999,
		0.999993,
		0.999695,
		0.99813,
		0.993892,
		0.98549,
		0.971569,
		0.951039,
		0.923141,
		0.887461,
		0.843926,
		0.792785,
		0.734587,
		0.670154,
		0.600563,
		0.527118,
		0.451339,
		0.37494,
		0.299819,
		0.228043,
		0.161842,
		0.103595,
		0.055825,
		0.021189,
		0.002478,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.000615,
		0.014613,
		0.044948,
		0.088884,
		0.143807,
		0.207232,
		0.276811,
		0.350337,
		0.425753,
		0.501165,
		0.574848,
		0.645261,
		0.711062,
		0.771123,
		0.824552,
		0.870708,
		0.909232,
		0.940063,
		0.963459,
		0.980016,
		0.99065,
		0.996571,
		0.999185,
		0.999929,
		1,
		0.999973,
		0.999493,
		0.997478,
		0.99252,
		0.983187,
		0.968209,
		0.946591,
		0.917657,
		0.881063,
		0.836792,
		0.785132,
		0.726654,
		0.662191,
		0.592813,
		0.519811,
		0.444677,
		0.369091,
		0.294906,
		0.224139,
		0.15896,
		0.101684,
		0.054761,
		0.020774,
		0.002428,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.005289,
		0.027373,
		0.064157,
		0.113042,
		0.171552,
		0.237335,
		0.308175,
		0.381994,
		0.456863,
		0.531011,
		0.602837,
		0.670922,
		0.73404,
		0.791183,
		0.84157,
		0.884676,
		0.920249,
		0.948332,
		0.969281,
		0.983772,
		0.992789,
		0.997571,
		0.999512,
		0.999974,
		1,
		0.999932,
		0.999216,
		0.996697,
		0.990983,
		0.980706,
		0.964681,
		0.941999,
		0.912065,
		0.874601,
		0.82964,
		0.777504,
		0.718785,
		0.654322,
		0.58518,
		0.512632,
		0.438147,
		0.363368,
		0.290107,
		0.220331,
		0.156153,
		0.099824,
		0.053727,
		0.02037,
		0.00238,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.00059,
		0.014049,
		0.043262,
		0.08565,
		0.138748,
		0.200209,
		0.267809,
		0.339457,
		0.413199,
		0.487226,
		0.559887,
		0.629697,
		0.695352,
		0.755743,
		0.809968,
		0.85736,
		0.897499,
		0.930235,
		0.95571,
		0.974369,
		0.986964,
		0.994532,
		0.998331,
		0.999729,
		0.999993,
		0.999999,
		0.999857,
		0.998853,
		0.995782,
		0.989282,
		0.978051,
		0.960989,
		0.937272,
		0.906376,
		0.868085,
		0.822478,
		0.769907,
		0.710982,
		0.646547,
		0.577661,
		0.50558,
		0.431746,
		0.357769,
		0.285421,
		0.216617,
		0.153419,
		0.098015,
		0.052722,
		0.019978,
		0.002333,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.005084,
		0.026342,
		0.061808,
		0.109032,
		0.165674,
		0.229511,
		0.298442,
		0.370494,
		0.44383,
		0.516759,
		0.587741,
		0.655405,
		0.718555,
		0.77619,
		0.827516,
		0.87197,
		0.909232,
		0.939252,
		0.962262,
		0.97879,
		0.989654,
		0.995933,
		0.998895,
		0.999863,
		0.999999,
		0.999994,
		0.999739,
		0.998392,
		0.994726,
		0.987414,
		0.975226,
		0.957143,
		0.932418,
		0.900598,
		0.861523,
		0.815311,
		0.762345,
		0.703249,
		0.638868,
		0.570256,
		0.498652,
		0.425471,
		0.352291,
		0.280842,
		0.212995,
		0.150754,
		0.096253,
		0.051745,
		0.019597,
		0.002287,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.000567,
		0.013517,
		0.041669,
		0.082589,
		0.133949,
		0.193529,
		0.259225,
		0.32905,
		0.401146,
		0.473788,
		0.545395,
		0.614537,
		0.679948,
		0.740538,
		0.795408,
		0.843868,
		0.885451,
		0.919935,
		0.947363,
		0.968053,
		0.982605,
		0.991898,
		0.997042,
		0.9993,
		0.999939,
		1,
		0.999977,
		0.999565,
		0.997826,
		0.993526,
		0.985382,
		0.972236,
		0.953149,
		0.927446,
		0.89474,
		0.854921,
		0.808146,
		0.754822,
		0.695586,
		0.631286,
		0.562964,
		0.491846,
		0.419319,
		0.34693,
		0.276369,
		0.20946,
		0.148158,
		0.094539,
		0.050795,
		0.019227,
		0.002243,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.004891,
		0.025367,
		0.059585,
		0.105229,
		0.160089,
		0.222058,
		0.289144,
		0.359472,
		0.431293,
		0.502989,
		0.573084,
		0.64025,
		0.703324,
		0.761317,
		0.813428,
		0.859064,
		0.897854,
		0.929666,
		0.95463,
		0.973143,
		0.985874,
		0.99375,
		0.997904,
		0.999581,
		0.999978,
		1,
		0.999942,
		0.999326,
		0.997147,
		0.992177,
		0.983187,
		0.969087,
		0.949014,
		0.922365,
		0.88881,
		0.848287,
		0.800988,
		0.747342,
		0.687996,
		0.623799,
		0.555785,
		0.48516,
		0.413289,
		0.341684,
		0.271997,
		0.20601,
		0.145627,
		0.09287,
		0.049871,
		0.018867,
		0.0022,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.000546,
		0.013016,
		0.040162,
		0.079688,
		0.129392,
		0.187172,
		0.251033,
		0.319089,
		0.389573,
		0.460836,
		0.531365,
		0.599783,
		0.664862,
		0.725537,
		0.780913,
		0.830285,
		0.873152,
		0.909232,
		0.938482,
		0.961112,
		0.977591,
		0.988652,
		0.99526,
		0.99856,
		0.999767,
		0.999994,
		0.999999,
		0.999877,
		0.999011,
		0.996348,
		0.990679,
		0.980831,
		0.965784,
		0.944746,
		0.917181,
		0.882814,
		0.841627,
		0.793842,
		0.739908,
		0.680482,
		0.61641,
		0.548718,
		0.478594,
		0.407377,
		0.336549,
		0.267725,
		0.202643,
		0.14316,
		0.091244,
		0.048971,
		0.018518,
		0.002158,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.004709,
		0.024446,
		0.057479,
		0.101621,
		0.154777,
		0.214952,
		0.280256,
		0.348906,
		0.419232,
		0.48969,
		0.55886,
		0.625463,
		0.688367,
		0.746597,
		0.799352,
		0.846017,
		0.886177,
		0.919639,
		0.94644,
		0.966867,
		0.981456,
		0.990993,
		0.996474,
		0.999046,
		0.999882,
		0.999999,
		0.999995,
		0.999775,
		0.99861,
		0.995423,
		0.989031,
		0.978317,
		0.962333,
		0.940353,
		0.911904,
		0.876762,
		0.834948,
		0.786713,
		0.732524,
		0.673044,
		0.609118,
		0.541761,
		0.472143,
		0.40158,
		0.331523,
		0.263549,
		0.199356,
		0.140754,
		0.08966,
		0.048096,
		0.018178,
		0.002117,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.000525,
		0.012541,
		0.038736,
		0.076937,
		0.125061,
		0.181117,
		0.243212,
		0.309553,
		0.378457,
		0.448352,
		0.517785,
		0.585432,
		0.650105,
		0.710763,
		0.76652,
		0.816662,
		0.860662,
		0.898188,
		0.929128,
		0.953597,
		0.971951,
		0.984791,
		0.992945,
		0.997435,
		0.999396,
		0.999948,
		1,
		0.99998,
		0.999625,
		0.998117,
		0.994368,
		0.987232,
		0.97565,
		0.958739,
		0.935842,
		0.90654,
		0.870658,
		0.828254,
		0.779606,
		0.725192,
		0.665684,
		0.601923,
		0.534913,
		0.465807,
		0.395897,
		0.326602,
		0.259467,
		0.196146,
		0.138407,
		0.088117,
		0.047243,
		0.017847,
		0.002078,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.004537,
		0.023574,
		0.055483,
		0.098193,
		0.149721,
		0.208175,
		0.271757,
		0.338772,
		0.407628,
		0.476846,
		0.545064,
		0.611048,
		0.673697,
		0.732056,
		0.785326,
		0.832877,
		0.874262,
		0.909232,
		0.937751,
		0.960006,
		0.97642,
		0.987648,
		0.994557,
		0.998182,
		0.999638,
		0.999981,
		1,
		0.99995,
		0.999417,
		0.997523,
		0.993181,
		0.985283,
		0.972832,
		0.95501,
		0.931219,
		0.901097,
		0.864511,
		0.821553,
		0.772523,
		0.717915,
		0.658402,
		0.594824,
		0.528173,
		0.459584,
		0.390324,
		0.321785,
		0.255475,
		0.193012,
		0.136118,
		0.086613,
		0.046414,
		0.017525,
		0.00204,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.000506,
		0.012092,
		0.037384,
		0.074325,
		0.120941,
		0.175346,
		0.23574,
		0.300419,
		0.367779,
		0.436317,
		0.504643,
		0.571482,
		0.635684,
		0.696234,
		0.752258,
		0.803041,
		0.848031,
		0.886861,
		0.919359,
		0.94556,
		0.965721,
		0.980326,
		0.990077,
		0.995871,
		0.99875,
		0.999798,
		0.999995,
		0.999999,
		0.999894,
		0.999142,
		0.996822,
		0.991857,
		0.983187,
		0.96987,
		0.951151,
		0.926493,
		0.895582,
		0.858326,
		0.814849,
		0.76547,
		0.710695,
		0.651201,
		0.587822,
		0.521539,
		0.453471,
		0.384859,
		0.317068,
		0.251572,
		0.189951,
		0.133884,
		0.085147,
		0.045605,
		0.017212,
		0.002002,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.004374,
		0.022747,
		0.053588,
		0.094935,
		0.144906,
		0.201706,
		0.263625,
		0.32905,
		0.39646,
		0.464441,
		0.531686,
		0.597004,
		0.659326,
		0.717717,
		0.771385,
		0.819691,
		0.862161,
		0.898503,
		0.928617,
		0.952607,
		0.970795,
		0.983719,
		0.992122,
		0.996928,
		0.999172,
		0.999898,
		0.999999,
		0.999995,
		0.999805,
		0.998793,
		0.996009,
		0.990397,
		0.980944,
		0.966766,
		0.94717,
		0.92167,
		0.890001,
		0.85211,
		0.808146,
		0.758449,
		0.703534,
		0.64408,
		0.580916,
		0.515011,
		0.447466,
		0.379501,
		0.31245,
		0.247755,
		0.186961,
		0.131705,
		0.083718,
		0.044818,
		0.016907,
		0.001966,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.000488,
		0.011667,
		0.036101,
		0.071843,
		0.11702,
		0.169842,
		0.228598,
		0.291667,
		0.357518,
		0.424716,
		0.491928,
		0.557927,
		0.621603,
		0.681965,
		0.738155,
		0.789457,
		0.835308,
		0.875306,
		0.909232,
		0.937054,
		0.958943,
		0.975277,
		0.986646,
		0.993829,
		0.997765,
		0.999475,
		0.999955,
		1,
		0.999983,
		0.999675,
		0.998361,
		0.995079,
		0.9888,
		0.978558,
		0.963527,
		0.943071,
		0.916757,
		0.884362,
		0.845868,
		0.80145,
		0.751464,
		0.696434,
		0.63704,
		0.574105,
		0.508587,
		0.441568,
		0.374245,
		0.307926,
		0.244022,
		0.184039,
		0.129577,
		0.082324,
		0.044051,
		0.01661,
		0.001931,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.004219,
		0.021964,
		0.051789,
		0.091835,
		0.140317,
		0.195527,
		0.255841,
		0.319719,
		0.385711,
		0.452462,
		0.518718,
		0.583329,
		0.645261,
		0.703599,
		0.757559,
		0.806498,
		0.849923,
		0.887506,
		0.919093,
		0.944719,
		0.964615,
		0.979216,
		0.989154,
		0.995238,
		0.998415,
		0.999686,
		0.999984,
		1,
		0.999956,
		0.999493,
		0.997839,
		0.994029,
		0.987065,
		0.976031,
		0.960158,
		0.938862,
		0.91176,
		0.87867,
		0.839605,
		0.794763,
		0.744517,
		0.689396,
		0.630081,
		0.567389,
		0.502265,
		0.435774,
		0.369091,
		0.303496,
		0.24037,
		0.181184,
		0.1275,
		0.080964,
		0.043303,
		0.016321,
		0.001896,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.000471,
		0.011263,
		0.034883,
		0.069482,
		0.113285,
		0.164588,
		0.221767,
		0.283276,
		0.347655,
		0.413531,
		0.479627,
		0.544762,
		0.607863,
		0.667967,
		0.724231,
		0.775945,
		0.822533,
		0.863572,
		0.898801,
		0.928131,
		0.951659,
		0.969673,
		0.982658,
		0.991285,
		0.996388,
		0.99891,
		0.999825,
		0.999996,
		0.999999,
		0.999908,
		0.999253,
		0.997222,
		0.992856,
		0.985193,
		0.973368,
		0.956664,
		0.934549,
		0.906686,
		0.872931,
		0.833326,
		0.788092,
		0.737611,
		0.682423,
		0.623205,
		0.560768,
		0.496045,
		0.430082,
		0.364035,
		0.299157,
		0.236797,
		0.178394,
		0.125472,
		0.079638,
		0.042574,
		0.016039,
		0.001863,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.004073,
		0.02122,
		0.050078,
		0.088884,
		0.135939,
		0.189623,
		0.248386,
		0.310761,
		0.375363,
		0.440894,
		0.506149,
		0.570021,
		0.631507,
		0.689715,
		0.743871,
		0.793332,
		0.837591,
		0.87629,
		0.909232,
		0.93639,
		0.95792,
		0.974162,
		0.985648,
		0.993081,
		0.997312,
		0.999278,
		0.999911,
		0.999999,
		0.999996,
		0.999831,
		0.998946,
		0.996504,
		0.991558,
		0.983187,
		0.970571,
		0.95305,
		0.930138,
		0.901541,
		0.867151,
		0.827036,
		0.781437,
		0.73075,
		0.675515,
		0.61641,
		0.55424,
		0.489923,
		0.424491,
		0.359076,
		0.294906,
		0.2333,
		0.175666,
		0.123492,
		0.078344,
		0.041863,
		0.015765,
		0.00183,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.000454,
		0.01088,
		0.033726,
		0.067236,
		0.109724,
		0.159571,
		0.21523,
		0.275228,
		0.338171,
		0.402746,
		0.467727,
		0.531979,
		0.594464,
		0.654248,
		0.710506,
		0.762531,
		0.809744,
		0.851704,
		0.888115,
		0.918841,
		0.943915,
		0.963547,
		0.978128,
		0.988228,
		0.99458,
		0.998044,
		0.999542,
		0.999961,
		1,
		0.999985,
		0.999716,
		0.998566,
		0.995681,
		0.990135,
		0.981047,
		0.967646,
		0.949321,
		0.925634,
		0.896331,
		0.861335,
		0.820741,
		0.774805,
		0.723934,
		0.668673,
		0.609698,
		0.547804,
		0.483901,
		0.418999,
		0.354211,
		0.290741,
		0.229879,
		0.173,
		0.121558,
		0.077081,
		0.04117,
		0.015497,
		0.001798,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.003934,
		0.020513,
		0.048451,
		0.086071,
		0.131761,
		0.183976,
		0.241242,
		0.302157,
		0.365398,
		0.429721,
		0.493969,
		0.557076,
		0.618068,
		0.676077,
		0.730344,
		0.780225,
		0.825205,
		0.864902,
		0.899083,
		0.927669,
		0.950749,
		0.968585,
		0.981612,
		0.990437,
		0.995817,
		0.998613,
		0.999726,
		0.999986,
		1,
		0.999962,
		0.999557,
		0.998106,
		0.994749,
		0.988586,
		0.978775,
		0.964597,
		0.945484,
		0.921045,
		0.891062,
		0.855488,
		0.814442,
		0.768196,
		0.717165,
		0.661898,
		0.603067,
		0.541461,
		0.477974,
		0.413603,
		0.349439,
		0.28666,
		0.22653,
		0.170393,
		0.119668,
		0.075848,
		0.040494,
		0.015236,
		0.001767,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.000439,
		0.010517,
		0.032625,
		0.065096,
		0.106327,
		0.154777,
		0.208972,
		0.267506,
		0.32905,
		0.392345,
		0.456215,
		0.51957,
		0.581406,
		0.640815,
		0.696992,
		0.749239,
		0.796974,
		0.839741,
		0.877219,
		0.909232,
		0.935757,
		0.956935,
		0.973075,
		0.984657,
		0.992316,
		0.996826,
		0.999046,
		0.999847,
		0.999996,
		1,
		0.99992,
		0.999346,
		0.997561,
		0.993706,
		0.986911,
		0.976376,
		0.961427,
		0.941544,
		0.916375,
		0.885739,
		0.849616,
		0.808146,
		0.761615,
		0.710447,
		0.655192,
		0.596519,
		0.535208,
		0.472143,
		0.408303,
		0.344756,
		0.282661,
		0.223252,
		0.167843,
		0.117821,
		0.074644,
		0.039835,
		0.014982,
		0.001737,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.003803,
		0.019841,
		0.046901,
		0.083389,
		0.127771,
		0.178574,
		0.234393,
		0.29389,
		0.355799,
		0.418929,
		0.482168,
		0.544486,
		0.604944,
		0.662695,
		0.716994,
		0.767202,
		0.812798,
		0.853382,
		0.88869,
		0.918602,
		0.943146,
		0.962515,
		0.977061,
		0.987301,
		0.993899,
		0.997637,
		0.999367,
		0.999923,
		1,
		0.999996,
		0.999852,
		0.999076,
		0.996925,
		0.992549,
		0.985111,
		0.973851,
		0.958141,
		0.937506,
		0.911631,
		0.880367,
		0.843722,
		0.801855,
		0.755064,
		0.70378,
		0.648554,
		0.590052,
		0.529046,
		0.466406,
		0.403095,
		0.340162,
		0.278741,
		0.220042,
		0.165349,
		0.116017,
		0.073468,
		0.039191,
		0.014734,
		0.001708,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.000424,
		0.010171,
		0.031577,
		0.063057,
		0.103084,
		0.150192,
		0.202976,
		0.260094,
		0.320274,
		0.382312,
		0.44508,
		0.507526,
		0.568684,
		0.627672,
		0.683702,
		0.736089,
		0.78425,
		0.827721,
		0.866157,
		0.89935,
		0.927229,
		0.949876,
		0.967529,
		0.980581,
		0.989583,
		0.995219,
		0.99828,
		0.999599,
		0.999966,
		1,
		0.999987,
		0.999752,
		0.998741,
		0.996194,
		0.991278,
		0.983187,
		0.971204,
		0.954745,
		0.933375,
		0.906817,
		0.874952,
		0.837812,
		0.795573,
		0.748546,
		0.697165,
		0.641985,
		0.583667,
		0.522972,
		0.460761,
		0.397978,
		0.335654,
		0.2749,
		0.2169,
		0.16291,
		0.114253,
		0.07232,
		0.038563,
		0.014492,
		0.001679,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.003677,
		0.019201,
		0.045425,
		0.08083,
		0.123957,
		0.173402,
		0.227824,
		0.285944,
		0.346552,
		0.408505,
		0.470734,
		0.532247,
		0.592135,
		0.649575,
		0.703836,
		0.754286,
		0.800401,
		0.841768,
		0.878098,
		0.909232,
		0.935153,
		0.955987,
		0.972017,
		0.983674,
		0.991538,
		0.99631,
		0.99878,
		0.99976,
		0.999988,
		1,
		0.999967,
		0.999612,
		0.998334,
		0.995364,
		0.989891,
		0.98114,
		0.968439,
		0.951243,
		0.929157,
		0.901939,
		0.869498,
		0.831889,
		0.789303,
		0.742062,
		0.690605,
		0.635486,
		0.577362,
		0.516987,
		0.455206,
		0.39295,
		0.331231,
		0.271135,
		0.213823,
		0.160523,
		0.112529,
		0.071199,
		0.03795,
		0.014256,
		0.001651,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.00041,
		0.009842,
		0.030578,
		0.061111,
		0.099986,
		0.145806,
		0.197229,
		0.252975,
		0.311827,
		0.372632,
		0.434306,
		0.495839,
		0.556295,
		0.61482,
		0.670646,
		0.723099,
		0.771599,
		0.815675,
		0.854966,
		0.889236,
		0.918374,
		0.94241,
		0.961517,
		0.976017,
		0.986377,
		0.993199,
		0.997199,
		0.999161,
		0.999866,
		0.999997,
		1,
		0.99993,
		0.999426,
		0.99785,
		0.994434,
		0.988388,
		0.978974,
		0.965558,
		0.947639,
		0.924857,
		0.897002,
		0.864011,
		0.825958,
		0.783049,
		0.735615,
		0.6841,
		0.629057,
		0.571138,
		0.511089,
		0.449741,
		0.38801,
		0.326889,
		0.267443,
		0.210809,
		0.158187,
		0.110843,
		0.070103,
		0.037351,
		0.014026,
		0.001624,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.003558,
		0.018592,
		0.044016,
		0.078386,
		0.120309,
		0.168448,
		0.22152,
		0.278304,
		0.33764,
		0.398433,
		0.459655,
		0.520351,
		0.57964,
		0.636722,
		0.690882,
		0.741497,
		0.788042,
		0.830094,
		0.867344,
		0.899603,
		0.926809,
		0.949038,
		0.966504,
		0.979567,
		0.988724,
		0.994599,
		0.997915,
		0.999443,
		0.999932,
		1,
		0.999997,
		0.99987,
		0.999187,
		0.997284,
		0.9934,
		0.986769,
		0.976689,
		0.962567,
		0.943938,
		0.92048,
		0.892012,
		0.858494,
		0.820021,
		0.776812,
		0.729206,
		0.677651,
		0.622699,
		0.564995,
		0.505276,
		0.444363,
		0.383156,
		0.322628,
		0.263824,
		0.207857,
		0.155901,
		0.109195,
		0.069032,
		0.036766,
		0.013802,
		0.001597,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.000397,
		0.009529,
		0.029626,
		0.059254,
		0.097025,
		0.141606,
		0.191718,
		0.246136,
		0.303696,
		0.363292,
		0.423884,
		0.484498,
		0.544233,
		0.60226,
		0.657831,
		0.710282,
		0.759042,
		0.803632,
		0.843682,
		0.878929,
		0.909232,
		0.934575,
		0.955074,
		0.970985,
		0.982702,
		0.990749,
		0.995768,
		0.998482,
		0.999647,
		0.99997,
		1,
		0.999989,
		0.999782,
		0.998889,
		0.996633,
		0.992261,
		0.985035,
		0.974289,
		0.959469,
		0.940146,
		0.916031,
		0.886971,
		0.852952,
		0.814083,
		0.770596,
		0.722839,
		0.67126,
		0.61641,
		0.55893,
		0.499548,
		0.439072,
		0.378386,
		0.318445,
		0.260276,
		0.204965,
		0.153664,
		0.107582,
		0.067986,
		0.036195,
		0.013582,
		0.001571,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.003444,
		0.018011,
		0.042672,
		0.076051,
		0.116819,
		0.163699,
		0.215467,
		0.270954,
		0.32905,
		0.388702,
		0.448922,
		0.50879,
		0.567455,
		0.624138,
		0.678141,
		0.728851,
		0.775744,
		0.81839,
		0.856465,
		0.889753,
		0.918157,
		0.941704,
		0.960553,
		0.974996,
		0.985457,
		0.992484,
		0.996733,
		0.998923,
		0.999788,
		0.999989,
		1,
		0.999971,
		0.999658,
		0.998528,
		0.995893,
		0.991015,
		0.983187,
		0.971777,
		0.956268,
		0.936266,
		0.911514,
		0.881887,
		0.847388,
		0.808146,
		0.764404,
		0.716513,
		0.664927,
		0.610192,
		0.552945,
		0.493904,
		0.433865,
		0.373698,
		0.31434,
		0.256796,
		0.202132,
		0.151474,
		0.106005,
		0.066963,
		0.035637,
		0.013368,
		0.001546,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.000384,
		0.00923,
		0.028718,
		0.057479,
		0.094192,
		0.137583,
		0.18643,
		0.239562,
		0.295864,
		0.354276,
		0.413799,
		0.473495,
		0.532493,
		0.589991,
		0.645261,
		0.697651,
		0.746597,
		0.791619,
		0.832336,
		0.868468,
		0.899843,
		0.926409,
		0.948233,
		0.965511,
		0.97857,
		0.987863,
		0.993959,
		0.99752,
		0.999259,
		0.999882,
		0.999997,
		1,
		0.999938,
		0.999493,
		0.998097,
		0.99506,
		0.989663,
		0.981226,
		0.969156,
		0.952968,
		0.932304,
		0.906935,
		0.876762,
		0.841807,
		0.802214,
		0.758237,
		0.710231,
		0.658652,
		0.604044,
		0.547038,
		0.488343,
		0.428742,
		0.369091,
		0.310309,
		0.253383,
		0.199356,
		0.14933,
		0.104462,
		0.065963,
		0.035092,
		0.013159,
		0.001521,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.003336,
		0.017457,
		0.041389,
		0.073818,
		0.113477,
		0.159146,
		0.209654,
		0.263882,
		0.320766,
		0.379296,
		0.438523,
		0.497557,
		0.555576,
		0.611825,
		0.665622,
		0.716363,
		0.763528,
		0.806683,
		0.845492,
		0.879718,
		0.909232,
		0.934022,
		0.954194,
		0.969981,
		0.981742,
		0.989954,
		0.995203,
		0.998153,
		0.999508,
		0.99994,
		1,
		0.999997,
		0.999886,
		0.999281,
		0.997593,
		0.994133,
		0.988203,
		0.979155,
		0.966428,
		0.949573,
		0.928264,
		0.902298,
		0.8716,
		0.836212,
		0.796291,
		0.752097,
		0.703994,
		0.652437,
		0.597967,
		0.541209,
		0.482863,
		0.423701,
		0.364563,
		0.306352,
		0.250035,
		0.196635,
		0.14723,
		0.102952,
		0.064984,
		0.034559,
		0.012955,
		0.001497,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.000372,
		0.008946,
		0.027851,
		0.055783,
		0.09148,
		0.133727,
		0.181354,
		0.233241,
		0.288319,
		0.345573,
		0.40404,
		0.46282,
		0.52107,
		0.578013,
		0.63294,
		0.685216,
		0.73428,
		0.779657,
		0.820956,
		0.857883,
		0.890244,
		0.91795,
		0.941027,
		0.959621,
		0.973997,
		0.984543,
		0.991758,
		0.99624,
		0.998655,
		0.999688,
		0.999973,
		1,
		0.99999,
		0.999807,
		0.999017,
		0.997011,
		0.993109,
		0.986637,
		0.976975,
		0.963598,
		0.946088,
		0.924149,
		0.897607,
		0.866407,
		0.830607,
		0.790377,
		0.745987,
		0.697803,
		0.646281,
		0.59196,
		0.535457,
		0.477463,
		0.41874,
		0.360112,
		0.302467,
		0.246752,
		0.193969,
		0.145174,
		0.101474,
		0.064028,
		0.034038,
		0.012755,
		0.001474,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.003233,
		0.016928,
		0.040162,
		0.071682,
		0.110275,
		0.154777,
		0.204067,
		0.257074,
		0.312776,
		0.370205,
		0.428447,
		0.486644,
		0.544,
		0.599783,
		0.653328,
		0.704043,
		0.751412,
		0.794999,
		0.834457,
		0.869533,
		0.900072,
		0.926026,
		0.947458,
		0.964547,
		0.977591,
		0.987004,
		0.993303,
		0.997097,
		0.999046,
		0.999813,
		0.99999,
		1,
		0.999974,
		0.999698,
		0.998695,
		0.996348,
		0.991988,
		0.984964,
		0.974689,
		0.960668,
		0.942516,
		0.919966,
		0.892867,
		0.861185,
		0.824995,
		0.784477,
		0.739908,
		0.691659,
		0.640185,
		0.586022,
		0.529781,
		0.472143,
		0.413858,
		0.355737,
		0.298652,
		0.24353,
		0.191355,
		0.14316,
		0.100028,
		0.063092,
		0.033529,
		0.01256,
		0.001451,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.000361,
		0.008674,
		0.027023,
		0.054161,
		0.088884,
		0.130029,
		0.176478,
		0.227159,
		0.281048,
		0.337168,
		0.394596,
		0.452462,
		0.509955,
		0.566321,
		0.620871,
		0.672983,
		0.722107,
		0.767766,
		0.809569,
		0.847207,
		0.880467,
		0.909232,
		0.933492,
		0.953345,
		0.969004,
		0.980795,
		0.989154,
		0.994617,
		0.997796,
		0.999343,
		0.999896,
		0.999998,
		1,
		0.999945,
		0.999551,
		0.99831,
		0.995601,
		0.990768,
		0.983187,
		0.972299,
		0.957642,
		0.938862,
		0.915718,
		0.888082,
		0.855939,
		0.819379,
		0.778592,
		0.733863,
		0.685564,
		0.63415,
		0.580154,
		0.52418,
		0.466901,
		0.409054,
		0.351437,
		0.294906,
		0.24037,
		0.188793,
		0.141187,
		0.098612,
		0.062177,
		0.033031,
		0.01237,
		0.001428,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.000807,
		0.010445,
		0.029959,
		0.058111,
		0.093699,
		0.135567,
		0.182603,
		0.233741,
		0.28796,
		0.344293,
		0.401822,
		0.459685,
		0.517077,
		0.573252,
		0.62753,
		0.679298,
		0.728012,
		0.77321,
		0.814507,
		0.851607,
		0.88431,
		0.91251,
		0.936212,
		0.955526,
		0.970681,
		0.982017,
		0.989983,
		0.995126,
		0.998065,
		0.999454,
		0.999923,
		0.999999,
		0.999999,
		0.999923,
		0.999454,
		0.998065,
		0.995126,
		0.989983,
		0.982017,
		0.970681,
		0.955526,
		0.936212,
		0.91251,
		0.88431,
		0.851607,
		0.814507,
		0.77321,
		0.728012,
		0.679298,
		0.62753,
		0.573252,
		0.517077,
		0.459685,
		0.401822,
		0.344293,
		0.28796,
		0.233741,
		0.182603,
		0.135567,
		0.093699,
		0.058111,
		0.029959,
		0.010445,
		0.000807,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		9.2e-005,
		0.00717,
		0.024591,
		0.051077,
		0.085389,
		0.126332,
		0.172758,
		0.223565,
		0.2777,
		0.33416,
		0.391998,
		0.45032,
		0.508291,
		0.56514,
		0.620158,
		0.672708,
		0.722224,
		0.76822,
		0.810295,
		0.848135,
		0.881522,
		0.910341,
		0.934586,
		0.954362,
		0.969893,
		0.981521,
		0.9897,
		0.994985,
		0.998008,
		0.999438,
		0.999921,
		0.999999,
		0.999999,
		0.999921,
		0.999438,
		0.998008,
		0.994985,
		0.9897,
		0.981521,
		0.969893,
		0.954362,
		0.934586,
		0.910341,
		0.881522,
		0.848135,
		0.810295,
		0.76822,
		0.722224,
		0.672708,
		0.620158,
		0.56514,
		0.508291,
		0.45032,
		0.391998,
		0.33416,
		0.2777,
		0.223565,
		0.172758,
		0.126332,
		0.085389,
		0.051077,
		0.024591,
		0.00717,
		9.2e-005,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.004448,
		0.019635,
		0.044332,
		0.077262,
		0.117192,
		0.162935,
		0.213353,
		0.267356,
		0.32391,
		0.382031,
		0.440796,
		0.499339,
		0.556859,
		0.612621,
		0.66596,
		0.71629,
		0.7631,
		0.805968,
		0.844563,
		0.878652,
		0.908106,
		0.932909,
		0.95316,
		0.969079,
		0.981008,
		0.989406,
		0.994838,
		0.997948,
		0.999421,
		0.999919,
		0.999999,
		0.999999,
		0.999919,
		0.999421,
		0.997948,
		0.994838,
		0.989406,
		0.981008,
		0.969079,
		0.95316,
		0.932909,
		0.908106,
		0.878652,
		0.844563,
		0.805968,
		0.7631,
		0.71629,
		0.66596,
		0.612621,
		0.556859,
		0.499339,
		0.440796,
		0.382031,
		0.32391,
		0.267356,
		0.213353,
		0.162935,
		0.117192,
		0.077262,
		0.044332,
		0.019635,
		0.004448,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		0.002328,
		0.015128,
		0.037906,
		0.069344,
		0.108167,
		0.153149,
		0.203114,
		0.256937,
		0.313545,
		0.371923,
		0.431113,
		0.490217,
		0.548406,
		0.604914,
		0.659052,
		0.710206,
		0.757843,
		0.801521,
		0.840888,
		0.875696,
		0.905802,
		0.931179,
		0.951918,
		0.968237,
		0.980477,
		0.989102,
		0.994686,
		0.997886,
		0.999403,
		0.999916,
		0.999999,
		0.999999,
		0.999916,
		0.999403,
		0.997886,
		0.994686,
		0.989102,
		0.980477,
		0.968237,
		0.951918,
		0.931179,
		0.905802,
		0.875696,
		0.840888,
		0.801521,
		0.757843,
		0.710206,
		0.659052,
		0.604914,
		0.548406,
		0.490217,
		0.431113,
		0.371923,
		0.313545,
		0.256937,
		0.203114,
		0.153149,
		0.108167,
		0.069344,
		0.037906,
		0.015128,
		0.002328,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.00086,
		0.011113,
		0.031834,
		0.06166,
		0.099276,
		0.143415,
		0.19286,
		0.246449,
		0.303071,
		0.361676,
		0.42127,
		0.480925,
		0.539778,
		0.597036,
		0.651979,
		0.703969,
		0.752448,
		0.796951,
		0.837108,
		0.872652,
		0.903427,
		0.929393,
		0.950636,
		0.967366,
		0.979927,
		0.988787,
		0.994529,
		0.997822,
		0.999384,
		0.999913,
		0.999999,
		0.999999,
		0.999913,
		0.999384,
		0.997822,
		0.994529,
		0.988787,
		0.979927,
		0.967366,
		0.950636,
		0.929393,
		0.903427,
		0.872652,
		0.837108,
		0.796951,
		0.752448,
		0.703969,
		0.651979,
		0.597036,
		0.539778,
		0.480925,
		0.42127,
		0.361676,
		0.303071,
		0.246449,
		0.19286,
		0.143415,
		0.099276,
		0.06166,
		0.031834,
		0.011113,
		0.00086,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		9.8e-005,
		0.007634,
		0.026151,
		0.05424,
		0.090543,
		0.13375,
		0.182603,
		0.2359,
		0.292492,
		0.351291,
		0.411269,
		0.471461,
		0.530974,
		0.588982,
		0.644738,
		0.697574,
		0.746909,
		0.792253,
		0.833218,
		0.869516,
		0.900978,
		0.927551,
		0.949311,
		0.966466,
		0.979358,
		0.988461,
		0.994366,
		0.997755,
		0.999365,
		0.999911,
		0.999999,
		0.999999,
		0.999911,
		0.999365,
		0.997755,
		0.994366,
		0.988461,
		0.979358,
		0.966466,
		0.949311,
		0.927551,
		0.900978,
		0.869516,
		0.833218,
		0.792253,
		0.746909,
		0.697574,
		0.644738,
		0.588982,
		0.530974,
		0.471461,
		0.411269,
		0.351291,
		0.292492,
		0.2359,
		0.182603,
		0.13375,
		0.090543,
		0.05424,
		0.026151,
		0.007634,
		9.8e-005,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.004741,
		0.020897,
		0.047116,
		0.081992,
		0.124172,
		0.172357,
		0.2253,
		0.281814,
		0.340772,
		0.401108,
		0.461824,
		0.52199,
		0.580749,
		0.637323,
		0.691017,
		0.741222,
		0.787425,
		0.829215,
		0.866286,
		0.898452,
		0.925648,
		0.947941,
		0.965534,
		0.978769,
		0.988123,
		0.994196,
		0.997686,
		0.999345,
		0.999908,
		0.999999,
		0.999999,
		0.999908,
		0.999345,
		0.997686,
		0.994196,
		0.988123,
		0.978769,
		0.965534,
		0.947941,
		0.925648,
		0.898452,
		0.866286,
		0.829215,
		0.787425,
		0.741222,
		0.691017,
		0.637323,
		0.580749,
		0.52199,
		0.461824,
		0.401108,
		0.340772,
		0.281814,
		0.2253,
		0.172357,
		0.124172,
		0.081992,
		0.047116,
		0.020897,
		0.004741,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.002483,
		0.016114,
		0.04032,
		0.073649,
		0.114702,
		0.162135,
		0.214658,
		0.271043,
		0.330122,
		0.39079,
		0.452013,
		0.512824,
		0.572334,
		0.629732,
		0.684294,
		0.735383,
		0.782462,
		0.825095,
		0.862958,
		0.895847,
		0.923684,
		0.946526,
		0.964571,
		0.978159,
		0.987773,
		0.99402,
		0.997614,
		0.999324,
		0.999905,
		0.999999,
		0.999999,
		0.999905,
		0.999324,
		0.997614,
		0.99402,
		0.987773,
		0.978159,
		0.964571,
		0.946526,
		0.923684,
		0.895847,
		0.862958,
		0.825095,
		0.782462,
		0.735383,
		0.684294,
		0.629732,
		0.572334,
		0.512824,
		0.452013,
		0.39079,
		0.330122,
		0.271043,
		0.214658,
		0.162135,
		0.114702,
		0.073649,
		0.04032,
		0.016114,
		0.002483,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000918,
		0.011847,
		0.033889,
		0.065543,
		0.105362,
		0.151954,
		0.203986,
		0.260186,
		0.319344,
		0.380315,
		0.442026,
		0.503473,
		0.563732,
		0.62196,
		0.6774,
		0.729388,
		0.777359,
		0.820854,
		0.859528,
		0.893159,
		0.921655,
		0.945063,
		0.963574,
		0.977527,
		0.987409,
		0.993838,
		0.99754,
		0.999302,
		0.999902,
		0.999999,
		0.999999,
		0.999902,
		0.999302,
		0.99754,
		0.993838,
		0.987409,
		0.977527,
		0.963574,
		0.945063,
		0.921655,
		0.893159,
		0.859528,
		0.820854,
		0.777359,
		0.729388,
		0.6774,
		0.62196,
		0.563732,
		0.503473,
		0.442026,
		0.380315,
		0.319344,
		0.260186,
		0.203986,
		0.151954,
		0.105362,
		0.065543,
		0.033889,
		0.011847,
		0.000918,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.000105,
		0.008146,
		0.027864,
		0.057705,
		0.096175,
		0.141832,
		0.193297,
		0.249252,
		0.308444,
		0.369686,
		0.431863,
		0.493936,
		0.554942,
		0.614003,
		0.670332,
		0.723232,
		0.772112,
		0.816488,
		0.855993,
		0.890386,
		0.91956,
		0.94355,
		0.962541,
		0.976872,
		0.987032,
		0.993648,
		0.997462,
		0.99928,
		0.999898,
		0.999999,
		0.999999,
		0.999898,
		0.99928,
		0.997462,
		0.993648,
		0.987032,
		0.976872,
		0.962541,
		0.94355,
		0.91956,
		0.890386,
		0.855993,
		0.816488,
		0.772112,
		0.723232,
		0.670332,
		0.614003,
		0.554942,
		0.493936,
		0.431863,
		0.369686,
		0.308444,
		0.249252,
		0.193297,
		0.141832,
		0.096175,
		0.057705,
		0.027864,
		0.008146,
		0.000105,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.005063,
		0.022285,
		0.050169,
		0.087167,
		0.131787,
		0.182603,
		0.238249,
		0.297425,
		0.358903,
		0.421525,
		0.48421,
		0.545959,
		0.605858,
		0.663083,
		0.71691,
		0.766717,
		0.811993,
		0.85235,
		0.887524,
		0.917395,
		0.941985,
		0.961473,
		0.976194,
		0.986641,
		0.993452,
		0.997381,
		0.999256,
		0.999895,
		0.999999,
		0.999999,
		0.999895,
		0.999256,
		0.997381,
		0.993452,
		0.986641,
		0.976194,
		0.961473,
		0.941985,
		0.917395,
		0.887524,
		0.85235,
		0.811993,
		0.766717,
		0.71691,
		0.663083,
		0.605858,
		0.545959,
		0.48421,
		0.421525,
		0.358903,
		0.297425,
		0.238249,
		0.182603,
		0.131787,
		0.087167,
		0.050169,
		0.022285,
		0.005063,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.002655,
		0.0172,
		0.042971,
		0.078366,
		0.121842,
		0.171921,
		0.227188,
		0.286296,
		0.34797,
		0.41101,
		0.474293,
		0.53678,
		0.597518,
		0.655651,
		0.710418,
		0.761169,
		0.807365,
		0.848593,
		0.884571,
		0.915158,
		0.940367,
		0.960366,
		0.97549,
		0.986235,
		0.993247,
		0.997298,
		0.999232,
		0.999891,
		0.999999,
		0.999999,
		0.999891,
		0.999232,
		0.997298,
		0.993247,
		0.986235,
		0.97549,
		0.960366,
		0.940367,
		0.915158,
		0.884571,
		0.848593,
		0.807365,
		0.761169,
		0.710418,
		0.655651,
		0.597518,
		0.53678,
		0.474293,
		0.41101,
		0.34797,
		0.286296,
		0.227188,
		0.171921,
		0.121842,
		0.078366,
		0.042971,
		0.0172,
		0.002655,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000982,
		0.012657,
		0.03615,
		0.069803,
		0.112019,
		0.161266,
		0.216079,
		0.275061,
		0.336891,
		0.40032,
		0.464184,
		0.527402,
		0.588982,
		0.648029,
		0.703751,
		0.755463,
		0.802599,
		0.84472,
		0.881522,
		0.912847,
		0.938692,
		0.95922,
		0.974761,
		0.985814,
		0.993035,
		0.99721,
		0.999206,
		0.999888,
		0.999998,
		0.999998,
		0.999888,
		0.999206,
		0.99721,
		0.993035,
		0.985814,
		0.974761,
		0.95922,
		0.938692,
		0.912847,
		0.881522,
		0.84472,
		0.802599,
		0.755463,
		0.703751,
		0.648029,
		0.588982,
		0.527402,
		0.464184,
		0.40032,
		0.336891,
		0.275061,
		0.216079,
		0.161266,
		0.112019,
		0.069803,
		0.03615,
		0.012657,
		0.000982,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.000112,
		0.008711,
		0.02975,
		0.061512,
		0.102344,
		0.150658,
		0.204936,
		0.26373,
		0.325668,
		0.389456,
		0.453882,
		0.517822,
		0.580244,
		0.640214,
		0.696903,
		0.749594,
		0.79769,
		0.840726,
		0.878374,
		0.910457,
		0.936959,
		0.958032,
		0.974004,
		0.985376,
		0.992814,
		0.997119,
		0.99918,
		0.999884,
		0.999998,
		0.999998,
		0.999884,
		0.99918,
		0.997119,
		0.992814,
		0.985376,
		0.974004,
		0.958032,
		0.936959,
		0.910457,
		0.878374,
		0.840726,
		0.79769,
		0.749594,
		0.696903,
		0.640214,
		0.580244,
		0.517822,
		0.453882,
		0.389456,
		0.325668,
		0.26373,
		0.204936,
		0.150658,
		0.102344,
		0.061512,
		0.02975,
		0.008711,
		0.000112,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.005419,
		0.023816,
		0.053528,
		0.092843,
		0.140116,
		0.193772,
		0.252311,
		0.314308,
		0.378419,
		0.443386,
		0.508038,
		0.571301,
		0.632201,
		0.68987,
		0.743557,
		0.792634,
		0.836606,
		0.875123,
		0.907987,
		0.935165,
		0.956802,
		0.973219,
		0.984922,
		0.992584,
		0.997025,
		0.999152,
		0.99988,
		0.999998,
		0.999998,
		0.99988,
		0.999152,
		0.997025,
		0.992584,
		0.984922,
		0.973219,
		0.956802,
		0.935165,
		0.907987,
		0.875123,
		0.836606,
		0.792634,
		0.743557,
		0.68987,
		0.632201,
		0.571301,
		0.508038,
		0.443386,
		0.378419,
		0.314308,
		0.252311,
		0.193772,
		0.140116,
		0.092843,
		0.053528,
		0.023816,
		0.005419,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.002844,
		0.018399,
		0.045892,
		0.083548,
		0.129662,
		0.182603,
		0.240813,
		0.302814,
		0.367211,
		0.432694,
		0.498046,
		0.562148,
		0.623984,
		0.682646,
		0.737347,
		0.787425,
		0.832357,
		0.871765,
		0.905432,
		0.933308,
		0.955526,
		0.972405,
		0.98445,
		0.992346,
		0.996926,
		0.999123,
		0.999876,
		0.999998,
		0.999998,
		0.999876,
		0.999123,
		0.996926,
		0.992346,
		0.98445,
		0.972405,
		0.955526,
		0.933308,
		0.905432,
		0.871765,
		0.832357,
		0.787425,
		0.737347,
		0.682646,
		0.623984,
		0.562148,
		0.498046,
		0.432694,
		0.367211,
		0.302814,
		0.240813,
		0.182603,
		0.129662,
		0.083548,
		0.045892,
		0.018399,
		0.002844,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.001053,
		0.013553,
		0.038645,
		0.07449,
		0.119322,
		0.171447,
		0.229249,
		0.291195,
		0.355834,
		0.421806,
		0.487844,
		0.552782,
		0.615558,
		0.675225,
		0.730958,
		0.782059,
		0.827973,
		0.868297,
		0.90279,
		0.931385,
		0.954204,
		0.971559,
		0.983959,
		0.992097,
		0.996824,
		0.999093,
		0.999871,
		0.999998,
		0.999998,
		0.999871,
		0.999093,
		0.996824,
		0.992097,
		0.983959,
		0.971559,
		0.954204,
		0.931385,
		0.90279,
		0.868297,
		0.827973,
		0.782059,
		0.730958,
		0.675225,
		0.615558,
		0.552782,
		0.487844,
		0.421806,
		0.355834,
		0.291195,
		0.229249,
		0.171447,
		0.119322,
		0.07449,
		0.038645,
		0.013553,
		0.001053,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.00012,
		0.009337,
		0.031834,
		0.065706,
		0.109121,
		0.160321,
		0.217631,
		0.279458,
		0.344293,
		0.410723,
		0.47743,
		0.543198,
		0.606919,
		0.667604,
		0.724384,
		0.776529,
		0.82345,
		0.864714,
		0.900056,
		0.929393,
		0.952833,
		0.970681,
		0.983449,
		0.991839,
		0.996717,
		0.999062,
		0.999867,
		0.999998,
		0.999998,
		0.999867,
		0.999062,
		0.996717,
		0.991839,
		0.983449,
		0.970681,
		0.952833,
		0.929393,
		0.900056,
		0.864714,
		0.82345,
		0.776529,
		0.724384,
		0.667604,
		0.606919,
		0.543198,
		0.47743,
		0.410723,
		0.344293,
		0.279458,
		0.217631,
		0.160321,
		0.109121,
		0.065706,
		0.031834,
		0.009337,
		0.00012,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.005814,
		0.02551,
		0.057235,
		0.099088,
		0.149248,
		0.205974,
		0.26761,
		0.332591,
		0.399446,
		0.466802,
		0.533393,
		0.598062,
		0.659775,
		0.717621,
		0.770831,
		0.818782,
		0.861011,
		0.897228,
		0.92733,
		0.95141,
		0.96977,
		0.982919,
		0.991569,
		0.996606,
		0.999029,
		0.999862,
		0.999998,
		0.999998,
		0.999862,
		0.999029,
		0.996606,
		0.991569,
		0.982919,
		0.96977,
		0.95141,
		0.92733,
		0.897228,
		0.861011,
		0.818782,
		0.770831,
		0.717621,
		0.659775,
		0.598062,
		0.533393,
		0.466802,
		0.399446,
		0.332591,
		0.26761,
		0.205974,
		0.149248,
		0.099088,
		0.057235,
		0.02551,
		0.005814,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.003055,
		0.019728,
		0.049119,
		0.089257,
		0.13825,
		0.194292,
		0.255662,
		0.320733,
		0.387975,
		0.455958,
		0.523363,
		0.588982,
		0.651733,
		0.710662,
		0.764959,
		0.813964,
		0.857183,
		0.894301,
		0.925192,
		0.949934,
		0.968823,
		0.982367,
		0.991289,
		0.99649,
		0.998995,
		0.999857,
		0.999998,
		0.999998,
		0.999857,
		0.998995,
		0.99649,
		0.991289,
		0.982367,
		0.968823,
		0.949934,
		0.925192,
		0.894301,
		0.857183,
		0.813964,
		0.764959,
		0.710662,
		0.651733,
		0.588982,
		0.523363,
		0.455958,
		0.387975,
		0.320733,
		0.255662,
		0.194292,
		0.13825,
		0.089257,
		0.049119,
		0.019728,
		0.003055,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.001133,
		0.014547,
		0.041405,
		0.079662,
		0.127354,
		0.182603,
		0.243624,
		0.308725,
		0.376311,
		0.444897,
		0.513104,
		0.579674,
		0.643472,
		0.7035,
		0.758906,
		0.80899,
		0.853227,
		0.891271,
		0.922975,
		0.948403,
		0.967838,
		0.981794,
		0.990997,
		0.996368,
		0.99896,
		0.999852,
		0.999998,
		0.999998,
		0.999852,
		0.99896,
		0.996368,
		0.990997,
		0.981794,
		0.967838,
		0.948403,
		0.922975,
		0.891271,
		0.853227,
		0.80899,
		0.758906,
		0.7035,
		0.643472,
		0.579674,
		0.513104,
		0.444897,
		0.376311,
		0.308725,
		0.243624,
		0.182603,
		0.127354,
		0.079662,
		0.041405,
		0.014547,
		0.001133,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.00013,
		0.010032,
		0.034144,
		0.070342,
		0.116587,
		0.170928,
		0.231511,
		0.296573,
		0.364459,
		0.433618,
		0.502614,
		0.570132,
		0.634987,
		0.696131,
		0.752666,
		0.803856,
		0.849137,
		0.888134,
		0.920678,
		0.946813,
		0.966816,
		0.981196,
		0.990693,
		0.996242,
		0.998922,
		0.999846,
		0.999998,
		0.999998,
		0.999846,
		0.998922,
		0.996242,
		0.990693,
		0.981196,
		0.966816,
		0.946813,
		0.920678,
		0.888134,
		0.849137,
		0.803856,
		0.752666,
		0.696131,
		0.634987,
		0.570132,
		0.502614,
		0.433618,
		0.364459,
		0.296573,
		0.231511,
		0.170928,
		0.116587,
		0.070342,
		0.034144,
		0.010032,
		0.00013,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.006254,
		0.027391,
		0.061339,
		0.10598,
		0.159289,
		0.219335,
		0.284287,
		0.35242,
		0.42212,
		0.491888,
		0.560352,
		0.626271,
		0.688547,
		0.746234,
		0.798554,
		0.844907,
		0.884886,
		0.918295,
		0.945162,
		0.965752,
		0.980575,
		0.990375,
		0.99611,
		0.998883,
		0.999841,
		0.999998,
		0.999998,
		0.999841,
		0.998883,
		0.99611,
		0.990375,
		0.980575,
		0.965752,
		0.945162,
		0.918295,
		0.884886,
		0.844907,
		0.798554,
		0.746234,
		0.688547,
		0.626271,
		0.560352,
		0.491888,
		0.42212,
		0.35242,
		0.284287,
		0.219335,
		0.159289,
		0.10598,
		0.061339,
		0.027391,
		0.006254,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.003289,
		0.021206,
		0.052699,
		0.095568,
		0.147708,
		0.207112,
		0.271875,
		0.3402,
		0.410404,
		0.480925,
		0.55033,
		0.617319,
		0.680741,
		0.739603,
		0.79308,
		0.840533,
		0.881522,
		0.915824,
		0.943447,
		0.964646,
		0.979927,
		0.990045,
		0.995972,
		0.998843,
		0.999835,
		0.999998,
		0.999998,
		0.999835,
		0.998843,
		0.995972,
		0.990045,
		0.979927,
		0.964646,
		0.943447,
		0.915824,
		0.881522,
		0.840533,
		0.79308,
		0.739603,
		0.680741,
		0.617319,
		0.55033,
		0.480925,
		0.410404,
		0.3402,
		0.271875,
		0.207112,
		0.147708,
		0.095568,
		0.052699,
		0.021206,
		0.003289,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.001221,
		0.015654,
		0.044472,
		0.085389,
		0.136215,
		0.194862,
		0.259348,
		0.327802,
		0.39847,
		0.469722,
		0.540059,
		0.608124,
		0.672708,
		0.732765,
		0.787425,
		0.836008,
		0.878037,
		0.91326,
		0.941666,
		0.963495,
		0.979253,
		0.9897,
		0.995829,
		0.9988,
		0.999829,
		0.999998,
		0.999998,
		0.999829,
		0.9988,
		0.995829,
		0.9897,
		0.979253,
		0.963495,
		0.941666,
		0.91326,
		0.878037,
		0.836008,
		0.787425,
		0.732765,
		0.672708,
		0.608124,
		0.540059,
		0.469722,
		0.39847,
		0.327802,
		0.259348,
		0.194862,
		0.136215,
		0.085389,
		0.044472,
		0.015654,
		0.001221,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.00014,
		0.010808,
		0.036715,
		0.075484,
		0.124837,
		0.182603,
		0.246719,
		0.315234,
		0.38632,
		0.458277,
		0.529537,
		0.59868,
		0.664439,
		0.725714,
		0.781584,
		0.831326,
		0.874425,
		0.9106,
		0.939815,
		0.962298,
		0.97855,
		0.98934,
		0.995678,
		0.998756,
		0.999822,
		0.999998,
		0.999998,
		0.999822,
		0.998756,
		0.995678,
		0.98934,
		0.97855,
		0.962298,
		0.939815,
		0.9106,
		0.874425,
		0.831326,
		0.781584,
		0.725714,
		0.664439,
		0.59868,
		0.529537,
		0.458277,
		0.38632,
		0.315234,
		0.246719,
		0.182603,
		0.124837,
		0.075484,
		0.036715,
		0.010808,
		0.00014,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.006746,
		0.029488,
		0.065898,
		0.113609,
		0.17036,
		0.234001,
		0.302502,
		0.373956,
		0.446587,
		0.518758,
		0.588982,
		0.655929,
		0.718443,
		0.775551,
		0.826481,
		0.870682,
		0.907839,
		0.937891,
		0.961052,
		0.977817,
		0.988964,
		0.995521,
		0.998709,
		0.999815,
		0.999997,
		0.999997,
		0.999815,
		0.998709,
		0.995521,
		0.988964,
		0.977817,
		0.961052,
		0.937891,
		0.907839,
		0.870682,
		0.826481,
		0.775551,
		0.718443,
		0.655929,
		0.588982,
		0.518758,
		0.446587,
		0.373956,
		0.302502,
		0.234001,
		0.17036,
		0.113609,
		0.065898,
		0.029488,
		0.006746,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.003552,
		0.022856,
		0.056682,
		0.102567,
		0.158156,
		0.221211,
		0.289616,
		0.361381,
		0.434652,
		0.507719,
		0.579023,
		0.647171,
		0.710944,
		0.769316,
		0.821467,
		0.866803,
		0.904972,
		0.93589,
		0.959754,
		0.977053,
		0.988571,
		0.995357,
		0.99866,
		0.999808,
		0.999997,
		0.999997,
		0.999808,
		0.99866,
		0.995357,
		0.988571,
		0.977053,
		0.959754,
		0.93589,
		0.904972,
		0.866803,
		0.821467,
		0.769316,
		0.710944,
		0.647171,
		0.579023,
		0.507719,
		0.434652,
		0.361381,
		0.289616,
		0.221211,
		0.158156,
		0.102567,
		0.056682,
		0.022856,
		0.003552,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.00132,
		0.016893,
		0.047892,
		0.091751,
		0.14602,
		0.208367,
		0.276585,
		0.348599,
		0.422472,
		0.496415,
		0.568797,
		0.638156,
		0.70321,
		0.762874,
		0.816277,
		0.86278,
		0.901995,
		0.93381,
		0.958402,
		0.976257,
		0.988161,
		0.995185,
		0.998609,
		0.9998,
		0.999997,
		0.999997,
		0.9998,
		0.998609,
		0.995185,
		0.988161,
		0.976257,
		0.958402,
		0.93381,
		0.901995,
		0.86278,
		0.816277,
		0.762874,
		0.70321,
		0.638156,
		0.568797,
		0.496415,
		0.422472,
		0.348599,
		0.276585,
		0.208367,
		0.14602,
		0.091751,
		0.047892,
		0.016893,
		0.00132,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.000152,
		0.011678,
		0.039587,
		0.081207,
		0.133985,
		0.19549,
		0.263423,
		0.335615,
		0.410046,
		0.484843,
		0.558299,
		0.628878,
		0.695232,
		0.756216,
		0.810904,
		0.858608,
		0.898903,
		0.931646,
		0.956994,
		0.975425,
		0.987733,
		0.995005,
		0.998556,
		0.999793,
		0.999997,
		0.999997,
		0.999793,
		0.998556,
		0.995005,
		0.987733,
		0.975425,
		0.956994,
		0.931646,
		0.898903,
		0.858608,
		0.810904,
		0.756216,
		0.695232,
		0.628878,
		0.558299,
		0.484843,
		0.410046,
		0.335615,
		0.263423,
		0.19549,
		0.133985,
		0.081207,
		0.039587,
		0.011678,
		0.000152,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.007298,
		0.031834,
		0.070982,
		0.122084,
		0.182603,
		0.250141,
		0.322436,
		0.397375,
		0.473,
		0.547521,
		0.619329,
		0.687003,
		0.749335,
		0.80534,
		0.854281,
		0.895691,
		0.929393,
		0.955526,
		0.974557,
		0.987285,
		0.994817,
		0.998499,
		0.999784,
		0.999997,
		0.999997,
		0.999784,
		0.998499,
		0.994817,
		0.987285,
		0.974557,
		0.955526,
		0.929393,
		0.895691,
		0.854281,
		0.80534,
		0.749335,
		0.687003,
		0.619329,
		0.547521,
		0.473,
		0.397375,
		0.322436,
		0.250141,
		0.182603,
		0.122084,
		0.070982,
		0.031834,
		0.007298,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.003848,
		0.024706,
		0.061133,
		0.110356,
		0.169733,
		0.236757,
		0.30907,
		0.38446,
		0.460882,
		0.53646,
		0.609502,
		0.678515,
		0.742222,
		0.799577,
		0.849792,
		0.892353,
		0.927049,
		0.953996,
		0.973651,
		0.986817,
		0.99462,
		0.998441,
		0.999775,
		0.999997,
		0.999997,
		0.999775,
		0.998441,
		0.99462,
		0.986817,
		0.973651,
		0.953996,
		0.927049,
		0.892353,
		0.849792,
		0.799577,
		0.742222,
		0.678515,
		0.609502,
		0.53646,
		0.460882,
		0.38446,
		0.30907,
		0.236757,
		0.169733,
		0.110356,
		0.061133,
		0.024706,
		0.003848,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.001432,
		0.018285,
		0.05172,
		0.098846,
		0.156908,
		0.223289,
		0.295525,
		0.371306,
		0.448489,
		0.525108,
		0.599388,
		0.669758,
		0.734868,
		0.793609,
		0.845133,
		0.888882,
		0.924608,
		0.952401,
		0.972704,
		0.986327,
		0.994413,
		0.998379,
		0.999766,
		0.999997,
		0.999997,
		0.999766,
		0.998379,
		0.994413,
		0.986327,
		0.972704,
		0.952401,
		0.924608,
		0.888882,
		0.845133,
		0.793609,
		0.734868,
		0.669758,
		0.599388,
		0.525108,
		0.448489,
		0.371306,
		0.295525,
		0.223289,
		0.156908,
		0.098846,
		0.05172,
		0.018285,
		0.001432,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.000165,
		0.012657,
		0.042808,
		0.087601,
		0.144161,
		0.209758,
		0.281814,
		0.357915,
		0.435818,
		0.513462,
		0.588982,
		0.660725,
		0.727265,
		0.787425,
		0.840298,
		0.885274,
		0.922066,
		0.950736,
		0.971715,
		0.985814,
		0.994196,
		0.998314,
		0.999757,
		0.999997,
		0.999997,
		0.999757,
		0.998314,
		0.994196,
		0.985814,
		0.971715,
		0.950736,
		0.922066,
		0.885274,
		0.840298,
		0.787425,
		0.727265,
		0.660725,
		0.588982,
		0.513462,
		0.435818,
		0.357915,
		0.281814,
		0.209758,
		0.144161,
		0.087601,
		0.042808,
		0.012657,
		0.000165,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.007921,
		0.034471,
		0.076675,
		0.131531,
		0.196187,
		0.26795,
		0.344293,
		0.422868,
		0.501515,
		0.578275,
		0.651406,
		0.719404,
		0.781018,
		0.835279,
		0.881522,
		0.919418,
		0.948998,
		0.970681,
		0.985277,
		0.993969,
		0.998246,
		0.999747,
		0.999997,
		0.999997,
		0.999747,
		0.998246,
		0.993969,
		0.985277,
		0.970681,
		0.948998,
		0.919418,
		0.881522,
		0.835279,
		0.781018,
		0.719404,
		0.651406,
		0.578275,
		0.501515,
		0.422868,
		0.344293,
		0.26795,
		0.196187,
		0.131531,
		0.076675,
		0.034471,
		0.007921,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.004182,
		0.026791,
		0.066128,
		0.119058,
		0.182603,
		0.253947,
		0.330446,
		0.409641,
		0.489264,
		0.567259,
		0.641793,
		0.711275,
		0.774379,
		0.830067,
		0.877618,
		0.916658,
		0.947185,
		0.9696,
		0.984714,
		0.993731,
		0.998174,
		0.999736,
		0.999996,
		0.999996,
		0.999736,
		0.998174,
		0.993731,
		0.984714,
		0.9696,
		0.947185,
		0.916658,
		0.877618,
		0.830067,
		0.774379,
		0.711275,
		0.641793,
		0.567259,
		0.489264,
		0.409641,
		0.330446,
		0.253947,
		0.182603,
		0.119058,
		0.066128,
		0.026791,
		0.004182,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.001559,
		0.019856,
		0.056024,
		0.106789,
		0.169038,
		0.239825,
		0.316383,
		0.396135,
		0.476704,
		0.555928,
		0.631876,
		0.702868,
		0.767497,
		0.824654,
		0.873557,
		0.913781,
		0.945291,
		0.968469,
		0.984125,
		0.99348,
		0.998099,
		0.999725,
		0.999996,
		0.999996,
		0.999725,
		0.998099,
		0.99348,
		0.984125,
		0.968469,
		0.945291,
		0.913781,
		0.873557,
		0.824654,
		0.767497,
		0.702868,
		0.631876,
		0.555928,
		0.476704,
		0.396135,
		0.316383,
		0.239825,
		0.169038,
		0.106789,
		0.056024,
		0.019856,
		0.001559,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.000179,
		0.013765,
		0.046438,
		0.094776,
		0.155525,
		0.225603,
		0.302114,
		0.382355,
		0.463832,
		0.544274,
		0.621646,
		0.694173,
		0.760363,
		0.819031,
		0.869329,
		0.91078,
		0.943313,
		0.967285,
		0.983507,
		0.993217,
		0.998019,
		0.999713,
		0.999996,
		0.999996,
		0.999713,
		0.998019,
		0.993217,
		0.983507,
		0.967285,
		0.943313,
		0.91078,
		0.869329,
		0.819031,
		0.760363,
		0.694173,
		0.621646,
		0.544274,
		0.463832,
		0.382355,
		0.302114,
		0.225603,
		0.155525,
		0.094776,
		0.046438,
		0.013765,
		0.000179,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.008627,
		0.03745,
		0.083077,
		0.142104,
		0.211307,
		0.287651,
		0.368303,
		0.450646,
		0.53229,
		0.611093,
		0.68518,
		0.752968,
		0.813189,
		0.864928,
		0.907651,
		0.941245,
		0.966046,
		0.982859,
		0.992941,
		0.997936,
		0.9997,
		0.999996,
		0.999996,
		0.9997,
		0.997936,
		0.992941,
		0.982859,
		0.966046,
		0.941245,
		0.907651,
		0.864928,
		0.813189,
		0.752968,
		0.68518,
		0.611093,
		0.53229,
		0.450646,
		0.368303,
		0.287651,
		0.211307,
		0.142104,
		0.083077,
		0.03745,
		0.008627,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.004562,
		0.02915,
		0.071756,
		0.128819,
		0.196963,
		0.273008,
		0.353984,
		0.437142,
		0.519969,
		0.600208,
		0.675879,
		0.745299,
		0.807118,
		0.860344,
		0.904386,
		0.939084,
		0.964748,
		0.982179,
		0.992651,
		0.997848,
		0.999687,
		0.999996,
		0.999996,
		0.999687,
		0.997848,
		0.992651,
		0.982179,
		0.964748,
		0.939084,
		0.904386,
		0.860344,
		0.807118,
		0.745299,
		0.675879,
		0.600208,
		0.519969,
		0.437142,
		0.353984,
		0.273008,
		0.196963,
		0.128819,
		0.071756,
		0.02915,
		0.004562,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.001703,
		0.021638,
		0.060886,
		0.11572,
		0.182603,
		0.258204,
		0.339405,
		0.42332,
		0.507305,
		0.588982,
		0.666257,
		0.737347,
		0.800807,
		0.85557,
		0.900978,
		0.936823,
		0.963389,
		0.981465,
		0.992346,
		0.997755,
		0.999673,
		0.999995,
		0.999995,
		0.999673,
		0.997755,
		0.992346,
		0.981465,
		0.963389,
		0.936823,
		0.900978,
		0.85557,
		0.800807,
		0.737347,
		0.666257,
		0.588982,
		0.507305,
		0.42332,
		0.339405,
		0.258204,
		0.182603,
		0.11572,
		0.060886,
		0.021638,
		0.001703,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.000196,
		0.015024,
		0.050548,
		0.102863,
		0.168264,
		0.243258,
		0.324575,
		0.409179,
		0.494292,
		0.577405,
		0.656305,
		0.729099,
		0.794247,
		0.850596,
		0.89742,
		0.934459,
		0.961963,
		0.980715,
		0.992024,
		0.997658,
		0.999658,
		0.999995,
		0.999995,
		0.999658,
		0.997658,
		0.992024,
		0.980715,
		0.961963,
		0.934459,
		0.89742,
		0.850596,
		0.794247,
		0.729099,
		0.656305,
		0.577405,
		0.494292,
		0.409179,
		0.324575,
		0.243258,
		0.168264,
		0.102863,
		0.050548,
		0.015024,
		0.000196,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.009431,
		0.04083,
		0.090309,
		0.153986,
		0.228195,
		0.309504,
		0.394723,
		0.480925,
		0.565468,
		0.64601,
		0.720543,
		0.787425,
		0.845412,
		0.893703,
		0.931984,
		0.960468,
		0.979927,
		0.991686,
		0.997555,
		0.999643,
		0.999995,
		0.999995,
		0.999643,
		0.997555,
		0.991686,
		0.979927,
		0.960468,
		0.931984,
		0.893703,
		0.845412,
		0.787425,
		0.720543,
		0.64601,
		0.565468,
		0.480925,
		0.394723,
		0.309504,
		0.228195,
		0.153986,
		0.090309,
		0.04083,
		0.009431,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.004996,
		0.031834,
		0.078131,
		0.139816,
		0.213043,
		0.294205,
		0.379951,
		0.4672,
		0.553161,
		0.63536,
		0.711668,
		0.78033,
		0.840008,
		0.889821,
		0.929393,
		0.9589,
		0.979099,
		0.99133,
		0.997446,
		0.999626,
		0.999995,
		0.999995,
		0.999626,
		0.997446,
		0.99133,
		0.979099,
		0.9589,
		0.929393,
		0.889821,
		0.840008,
		0.78033,
		0.711668,
		0.63536,
		0.553161,
		0.4672,
		0.379951,
		0.294205,
		0.213043,
		0.139816,
		0.078131,
		0.031834,
		0.004996,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.001868,
		0.02367,
		0.066407,
		0.125807,
		0.197833,
		0.278696,
		0.364871,
		0.453111,
		0.540475,
		0.624345,
		0.702461,
		0.77295,
		0.834373,
		0.885764,
		0.92668,
		0.957254,
		0.978228,
		0.990954,
		0.997332,
		0.999609,
		0.999995,
		0.999995,
		0.999609,
		0.997332,
		0.990954,
		0.978228,
		0.957254,
		0.92668,
		0.885764,
		0.834373,
		0.77295,
		0.702461,
		0.624345,
		0.540475,
		0.453111,
		0.364871,
		0.278696,
		0.197833,
		0.125807,
		0.066407,
		0.02367,
		0.001868,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.000216,
		0.016464,
		0.055226,
		0.112019,
		0.182603,
		0.262996,
		0.349487,
		0.438658,
		0.527402,
		0.612952,
		0.692907,
		0.765272,
		0.828496,
		0.881522,
		0.923837,
		0.955526,
		0.977312,
		0.990558,
		0.99721,
		0.999591,
		0.999994,
		0.999994,
		0.999591,
		0.99721,
		0.990558,
		0.977312,
		0.955526,
		0.923837,
		0.881522,
		0.828496,
		0.765272,
		0.692907,
		0.612952,
		0.527402,
		0.438658,
		0.349487,
		0.262996,
		0.182603,
		0.112019,
		0.055226,
		0.016464,
		0.000216,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.010354,
		0.044689,
		0.098521,
		0.167396,
		0.247127,
		0.333809,
		0.423837,
		0.513933,
		0.601168,
		0.682994,
		0.757282,
		0.822365,
		0.877086,
		0.920857,
		0.953711,
		0.976347,
		0.990141,
		0.997082,
		0.999571,
		0.999994,
		0.999994,
		0.999571,
		0.997082,
		0.990141,
		0.976347,
		0.953711,
		0.920857,
		0.877086,
		0.822365,
		0.757282,
		0.682994,
		0.601168,
		0.513933,
		0.423837,
		0.333809,
		0.247127,
		0.167396,
		0.098521,
		0.044689,
		0.010354,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.005495,
		0.034906,
		0.085389,
		0.152262,
		0.231119,
		0.317848,
		0.40865,
		0.500061,
		0.588982,
		0.672708,
		0.748966,
		0.815966,
		0.872446,
		0.917733,
		0.951803,
		0.975331,
		0.9897,
		0.996946,
		0.999551,
		0.999994,
		0.999994,
		0.999551,
		0.996946,
		0.9897,
		0.975331,
		0.951803,
		0.917733,
		0.872446,
		0.815966,
		0.748966,
		0.672708,
		0.588982,
		0.500061,
		0.40865,
		0.317848,
		0.231119,
		0.152262,
		0.085389,
		0.034906,
		0.005495,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.002059,
		0.026003,
		0.07271,
		0.137255,
		0.215002,
		0.301618,
		0.393097,
		0.485779,
		0.576381,
		0.662033,
		0.74031,
		0.809287,
		0.86759,
		0.914455,
		0.949797,
		0.97426,
		0.989234,
		0.996803,
		0.999529,
		0.999993,
		0.999993,
		0.999529,
		0.996803,
		0.989234,
		0.97426,
		0.949797,
		0.914455,
		0.86759,
		0.809287,
		0.74031,
		0.662033,
		0.576381,
		0.485779,
		0.393097,
		0.301618,
		0.215002,
		0.137255,
		0.07271,
		0.026003,
		0.002059,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.000238,
		0.018122,
		0.060583,
		0.122442,
		0.198815,
		0.285138,
		0.377182,
		0.471079,
		0.563354,
		0.650955,
		0.731298,
		0.802314,
		0.862506,
		0.911016,
		0.947687,
		0.97313,
		0.988742,
		0.996651,
		0.999505,
		0.999993,
		0.999993,
		0.999505,
		0.996651,
		0.988742,
		0.97313,
		0.947687,
		0.911016,
		0.862506,
		0.802314,
		0.731298,
		0.650955,
		0.563354,
		0.471079,
		0.377182,
		0.285138,
		0.198815,
		0.122442,
		0.060583,
		0.018122,
		0.000238,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.011419,
		0.049119,
		0.107895,
		0.182603,
		0.26843,
		0.360911,
		0.455958,
		0.549888,
		0.63946,
		0.721915,
		0.795032,
		0.857183,
		0.907405,
		0.945467,
		0.971939,
		0.988221,
		0.99649,
		0.999481,
		0.999993,
		0.999993,
		0.999481,
		0.99649,
		0.988221,
		0.971939,
		0.945467,
		0.907405,
		0.857183,
		0.795032,
		0.721915,
		0.63946,
		0.549888,
		0.455958,
		0.360911,
		0.26843,
		0.182603,
		0.107895,
		0.049119,
		0.011419,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.006072,
		0.038444,
		0.093699,
		0.166417,
		0.25152,
		0.344293,
		0.440412,
		0.535972,
		0.62753,
		0.712144,
		0.787425,
		0.851607,
		0.903612,
		0.943128,
		0.970681,
		0.98767,
		0.996318,
		0.999454,
		0.999992,
		0.999992,
		0.999454,
		0.996318,
		0.98767,
		0.970681,
		0.943128,
		0.903612,
		0.851607,
		0.787425,
		0.712144,
		0.62753,
		0.535972,
		0.440412,
		0.344293,
		0.25152,
		0.166417,
		0.093699,
		0.038444,
		0.006072,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.00228,
		0.028698,
		0.079949,
		0.150317,
		0.234441,
		0.32734,
		0.424437,
		0.521595,
		0.615152,
		0.701966,
		0.779477,
		0.845765,
		0.899627,
		0.940665,
		0.969353,
		0.987087,
		0.996137,
		0.999426,
		0.999992,
		0.999992,
		0.999426,
		0.996137,
		0.987087,
		0.969353,
		0.940665,
		0.899627,
		0.845765,
		0.779477,
		0.701966,
		0.615152,
		0.521595,
		0.424437,
		0.32734,
		0.234441,
		0.150317,
		0.079949,
		0.028698,
		0.00228,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.000264,
		0.020042,
		0.066755,
		0.134371,
		0.21723,
		0.310066,
		0.408035,
		0.506746,
		0.602308,
		0.691366,
		0.771171,
		0.83964,
		0.895439,
		0.938069,
		0.96795,
		0.986469,
		0.995944,
		0.999397,
		0.999991,
		0.999991,
		0.999397,
		0.995944,
		0.986469,
		0.96795,
		0.938069,
		0.895439,
		0.83964,
		0.771171,
		0.691366,
		0.602308,
		0.506746,
		0.408035,
		0.310066,
		0.21723,
		0.134371,
		0.066755,
		0.020042,
		0.000264,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.012657,
		0.05424,
		0.118659,
		0.199933,
		0.292492,
		0.391206,
		0.491417,
		0.588982,
		0.680324,
		0.762488,
		0.833218,
		0.891034,
		0.935331,
		0.966466,
		0.985814,
		0.995739,
		0.999365,
		0.999991,
		0.999991,
		0.999365,
		0.995739,
		0.985814,
		0.966466,
		0.935331,
		0.891034,
		0.833218,
		0.762488,
		0.680324,
		0.588982,
		0.491417,
		0.391206,
		0.292492,
		0.199933,
		0.118659,
		0.05424,
		0.012657,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.006746,
		0.042546,
		0.103273,
		0.182603,
		0.274643,
		0.373956,
		0.475597,
		0.575159,
		0.668821,
		0.75341,
		0.826481,
		0.8864,
		0.932443,
		0.964896,
		0.985119,
		0.995521,
		0.999331,
		0.999991,
		0.999991,
		0.999331,
		0.995521,
		0.985119,
		0.964896,
		0.932443,
		0.8864,
		0.826481,
		0.75341,
		0.668821,
		0.575159,
		0.475597,
		0.373956,
		0.274643,
		0.182603,
		0.103273,
		0.042546,
		0.006746,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.002539,
		0.031834,
		0.088318,
		0.165303,
		0.25655,
		0.356293,
		0.459281,
		0.560823,
		0.656837,
		0.743917,
		0.819413,
		0.881522,
		0.929393,
		0.963233,
		0.984381,
		0.995289,
		0.999295,
		0.99999,
		0.99999,
		0.999295,
		0.995289,
		0.984381,
		0.963233,
		0.929393,
		0.881522,
		0.819413,
		0.743917,
		0.656837,
		0.560823,
		0.459281,
		0.356293,
		0.25655,
		0.165303,
		0.088318,
		0.031834,
		0.002539,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.000295,
		0.022285,
		0.073915,
		0.148107,
		0.238249,
		0.338228,
		0.442464,
		0.545959,
		0.644352,
		0.733987,
		0.811993,
		0.876386,
		0.926173,
		0.961473,
		0.983597,
		0.995042,
		0.999256,
		0.999989,
		0.999989,
		0.999256,
		0.995042,
		0.983597,
		0.961473,
		0.926173,
		0.876386,
		0.811993,
		0.733987,
		0.644352,
		0.545959,
		0.442464,
		0.338228,
		0.238249,
		0.148107,
		0.073915,
		0.022285,
		0.000295,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.014108,
		0.060202,
		0.131098,
		0.219786,
		0.319778,
		0.425141,
		0.53055,
		0.631345,
		0.723599,
		0.804203,
		0.870975,
		0.92277,
		0.959607,
		0.982764,
		0.994778,
		0.999215,
		0.999989,
		0.999989,
		0.999215,
		0.994778,
		0.982764,
		0.959607,
		0.92277,
		0.870975,
		0.804203,
		0.723599,
		0.631345,
		0.53055,
		0.425141,
		0.319778,
		0.219786,
		0.131098,
		0.060202,
		0.014108,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.007538,
		0.047341,
		0.114378,
		0.201216,
		0.300964,
		0.407312,
		0.514584,
		0.617796,
		0.712729,
		0.796022,
		0.865273,
		0.919171,
		0.957627,
		0.981877,
		0.994497,
		0.999171,
		0.999988,
		0.999988,
		0.999171,
		0.994497,
		0.981877,
		0.957627,
		0.919171,
		0.865273,
		0.796022,
		0.712729,
		0.617796,
		0.514584,
		0.407312,
		0.300964,
		0.201216,
		0.114378,
		0.047341,
		0.007538,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.002844,
		0.035512,
		0.098062,
		0.182603,
		0.281814,
		0.38898,
		0.498046,
		0.603682,
		0.701355,
		0.787425,
		0.85926,
		0.915365,
		0.955526,
		0.980933,
		0.994196,
		0.999123,
		0.999987,
		0.999987,
		0.999123,
		0.994196,
		0.980933,
		0.955526,
		0.915365,
		0.85926,
		0.787425,
		0.701355,
		0.603682,
		0.498046,
		0.38898,
		0.281814,
		0.182603,
		0.098062,
		0.035512,
		0.002844,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.000331,
		0.024926,
		0.082283,
		0.164025,
		0.262364,
		0.370151,
		0.480925,
		0.588982,
		0.68945,
		0.778391,
		0.852917,
		0.911335,
		0.953295,
		0.979927,
		0.993875,
		0.999073,
		0.999987,
		0.999987,
		0.999073,
		0.993875,
		0.979927,
		0.953295,
		0.911335,
		0.852917,
		0.778391,
		0.68945,
		0.588982,
		0.480925,
		0.370151,
		0.262364,
		0.164025,
		0.082283,
		0.024926,
		0.000331,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.015823,
		0.067199,
		0.145573,
		0.242658,
		0.350837,
		0.463212,
		0.573675,
		0.676989,
		0.768893,
		0.846224,
		0.907068,
		0.950924,
		0.978855,
		0.993531,
		0.999018,
		0.999986,
		0.999986,
		0.999018,
		0.993531,
		0.978855,
		0.950924,
		0.907068,
		0.846224,
		0.768893,
		0.676989,
		0.573675,
		0.463212,
		0.350837,
		0.242658,
		0.145573,
		0.067199,
		0.015823,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.008478,
		0.05299,
		0.127354,
		0.222749,
		0.331053,
		0.444897,
		0.557738,
		0.663946,
		0.758906,
		0.839157,
		0.902545,
		0.948403,
		0.97771,
		0.993163,
		0.99896,
		0.999985,
		0.999985,
		0.99896,
		0.993163,
		0.97771,
		0.948403,
		0.902545,
		0.839157,
		0.758906,
		0.663946,
		0.557738,
		0.444897,
		0.331053,
		0.222749,
		0.127354,
		0.05299,
		0.008478,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.003208,
		0.039864,
		0.109496,
		0.202704,
		0.310823,
		0.425977,
		0.541152,
		0.650291,
		0.7484,
		0.831692,
		0.897749,
		0.945719,
		0.976488,
		0.992769,
		0.998897,
		0.999984,
		0.999984,
		0.998897,
		0.992769,
		0.976488,
		0.945719,
		0.897749,
		0.831692,
		0.7484,
		0.650291,
		0.541152,
		0.425977,
		0.310823,
		0.202704,
		0.109496,
		0.039864,
		0.003208,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.000375,
		0.028064,
		0.092147,
		0.182603,
		0.290179,
		0.40645,
		0.523896,
		0.635997,
		0.737347,
		0.823803,
		0.892661,
		0.942861,
		0.975181,
		0.992346,
		0.998829,
		0.999983,
		0.999983,
		0.998829,
		0.992346,
		0.975181,
		0.942861,
		0.892661,
		0.823803,
		0.737347,
		0.635997,
		0.523896,
		0.40645,
		0.290179,
		0.182603,
		0.092147,
		0.028064,
		0.000375,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.01787,
		0.075483,
		0.162543,
		0.269161,
		0.38632,
		0.505951,
		0.621033,
		0.725713,
		0.815461,
		0.887259,
		0.939814,
		0.973782,
		0.99189,
		0.998754,
		0.99998,
		0.99998,
		0.998754,
		0.99189,
		0.973782,
		0.939814,
		0.887259,
		0.815461,
		0.725713,
		0.621033,
		0.505951,
		0.38632,
		0.269161,
		0.162543,
		0.075483,
		0.01787,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.009606,
		0.059709,
		0.142636,
		0.247816,
		0.365589,
		0.487293,
		0.605362,
		0.713457,
		0.806627,
		0.881506,
		0.936549,
		0.972268,
		0.991385,
		0.998659,
		0.999963,
		0.999963,
		0.998659,
		0.991385,
		0.972268,
		0.936549,
		0.881506,
		0.806627,
		0.713457,
		0.605362,
		0.487293,
		0.365589,
		0.247816,
		0.142636,
		0.059709,
		0.009606,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.003646,
		0.045061,
		0.123016,
		0.226198,
		0.344254,
		0.467878,
		0.588915,
		0.700499,
		0.797216,
		0.875324,
		0.932992,
		0.970571,
		0.990764,
		0.998478,
		0.999866,
		0.999866,
		0.998478,
		0.990764,
		0.970571,
		0.932992,
		0.875324,
		0.797216,
		0.700499,
		0.588915,
		0.467878,
		0.344254,
		0.226198,
		0.123016,
		0.045061,
		0.003646,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.000428,
		0.031819,
		0.103831,
		0.204353,
		0.322282,
		0.447615,
		0.57156,
		0.686676,
		0.78705,
		0.868522,
		0.92895,
		0.968498,
		0.989838,
		0.998023,
		0.999501,
		0.999501,
		0.998023,
		0.989838,
		0.968498,
		0.92895,
		0.868522,
		0.78705,
		0.686676,
		0.57156,
		0.447615,
		0.322282,
		0.204353,
		0.103831,
		0.031819,
		0.000428,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.020311,
		0.085259,
		0.182325,
		0.299601,
		0.426337,
		0.55305,
		0.671681,
		0.775775,
		0.860718,
		0.924021,
		0.965637,
		0.98819,
		0.996877,
		0.998451,
		0.998451,
		0.996877,
		0.98819,
		0.965637,
		0.924021,
		0.860718,
		0.775775,
		0.671681,
		0.55305,
		0.426337,
		0.299601,
		0.182325,
		0.085259,
		0.020311,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.01093,
		0.067515,
		0.160159,
		0.276097,
		0.40378,
		0.532994,
		0.655014,
		0.762802,
		0.851254,
		0.917501,
		0.961255,
		0.985074,
		0.994288,
		0.995965,
		0.995965,
		0.994288,
		0.985074,
		0.961255,
		0.917501,
		0.851254,
		0.762802,
		0.655014,
		0.532994,
		0.40378,
		0.276097,
		0.160159,
		0.067515,
		0.01093,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.004144,
		0.050887,
		0.137938,
		0.25164,
		0.379606,
		0.510855,
		0.635962,
		0.747271,
		0.839153,
		0.908329,
		0.954239,
		0.979348,
		0.989105,
		0.990887,
		0.990887,
		0.989105,
		0.979348,
		0.954239,
		0.908329,
		0.839153,
		0.747271,
		0.635962,
		0.510855,
		0.379606,
		0.25164,
		0.137938,
		0.050887,
		0.004144,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.000483,
		0.035748,
		0.115822,
		0.226135,
		0.353452,
		0.486002,
		0.613644,
		0.728083,
		0.82314,
		0.895097,
		0.943094,
		0.969472,
		0.97977,
		0.981659,
		0.981659,
		0.97977,
		0.969472,
		0.943094,
		0.895097,
		0.82314,
		0.728083,
		0.613644,
		0.486002,
		0.353452,
		0.226135,
		0.115822,
		0.035748,
		0.000483,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.022579,
		0.094096,
		0.199599,
		0.325018,
		0.457808,
		0.587116,
		0.704007,
		0.801745,
		0.876156,
		0.92605,
		0.953609,
		0.964421,
		0.966413,
		0.966413,
		0.964421,
		0.953609,
		0.92605,
		0.876156,
		0.801745,
		0.704007,
		0.587116,
		0.457808,
		0.325018,
		0.199599,
		0.094096,
		0.022579,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.011938,
		0.073204,
		0.172233,
		0.294186,
		0.425793,
		0.555533,
		0.67386,
		0.773497,
		0.84981,
		0.901261,
		0.929828,
		0.941091,
		0.943177,
		0.943177,
		0.941091,
		0.929828,
		0.901261,
		0.84981,
		0.773497,
		0.67386,
		0.555533,
		0.425793,
		0.294186,
		0.172233,
		0.073204,
		0.011938,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.00441,
		0.053745,
		0.14448,
		0.26112,
		0.389776,
		0.518344,
		0.636733,
		0.737171,
		0.814587,
		0.86708,
		0.896384,
		0.907999,
		0.91016,
		0.91016,
		0.907999,
		0.896384,
		0.86708,
		0.814587,
		0.737171,
		0.636733,
		0.518344,
		0.389776,
		0.26112,
		0.14448,
		0.053745,
		0.00441,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.000496,
		0.036428,
		0.117035,
		0.226346,
		0.35001,
		0.475478,
		0.592226,
		0.692067,
		0.769537,
		0.822383,
		0.852052,
		0.863875,
		0.866086,
		0.866086,
		0.863875,
		0.852052,
		0.822383,
		0.769537,
		0.692067,
		0.592226,
		0.475478,
		0.35001,
		0.226346,
		0.117035,
		0.036428,
		0.000496,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.021973,
		0.090795,
		0.190753,
		0.30725,
		0.427476,
		0.540631,
		0.638232,
		0.714501,
		0.766854,
		0.79642,
		0.80827,
		0.810497,
		0.810497,
		0.80827,
		0.79642,
		0.766854,
		0.714501,
		0.638232,
		0.540631,
		0.427476,
		0.30725,
		0.190753,
		0.090795,
		0.021973,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.01098,
		0.066752,
		0.155529,
		0.262731,
		0.375519,
		0.483015,
		0.576593,
		0.650267,
		0.701172,
		0.730099,
		0.741761,
		0.743965,
		0.743965,
		0.741761,
		0.730099,
		0.701172,
		0.650267,
		0.576593,
		0.483015,
		0.375519,
		0.262731,
		0.155529,
		0.066752,
		0.01098,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.003794,
		0.045838,
		0.122012,
		0.218049,
		0.321345,
		0.42117,
		0.50894,
		0.578592,
		0.627053,
		0.654769,
		0.666012,
		0.668148,
		0.668148,
		0.666012,
		0.654769,
		0.627053,
		0.578592,
		0.50894,
		0.42117,
		0.321345,
		0.218049,
		0.122012,
		0.045838,
		0.003794,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.000395,
		0.028769,
		0.091506,
		0.174964,
		0.267053,
		0.357431,
		0.43776,
		0.502049,
		0.547106,
		0.573048,
		0.58364,
		0.585664,
		0.585664,
		0.58364,
		0.573048,
		0.547106,
		0.502049,
		0.43776,
		0.357431,
		0.267053,
		0.174964,
		0.091506,
		0.028769,
		0.000395,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.015912,
		0.065084,
		0.135159,
		0.214834,
		0.294395,
		0.365947,
		0.423733,
		0.464545,
		0.48821,
		0.497936,
		0.499807,
		0.499807,
		0.497936,
		0.48821,
		0.464545,
		0.423733,
		0.365947,
		0.294395,
		0.214834,
		0.135159,
		0.065084,
		0.015912,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.007219,
		0.043437,
		0.100022,
		0.166698,
		0.234593,
		0.296449,
		0.346893,
		0.38281,
		0.403792,
		0.412477,
		0.414158,
		0.414158,
		0.412477,
		0.403792,
		0.38281,
		0.346893,
		0.296449,
		0.234593,
		0.166698,
		0.100022,
		0.043437,
		0.007219,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.002241,
		0.026795,
		0.070476,
		0.124229,
		0.180203,
		0.23193,
		0.274558,
		0.305174,
		0.323198,
		0.330713,
		0.332178,
		0.332178,
		0.330713,
		0.323198,
		0.305174,
		0.274558,
		0.23193,
		0.180203,
		0.124229,
		0.070476,
		0.026795,
		0.002241,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.000207,
		0.014927,
		0.046907,
		0.088442,
		0.132827,
		0.174498,
		0.209231,
		0.234406,
		0.249349,
		0.255628,
		0.256861,
		0.256861,
		0.255628,
		0.249349,
		0.234406,
		0.209231,
		0.174498,
		0.132827,
		0.088442,
		0.046907,
		0.014927,
		0.000207,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.007221,
		0.029175,
		0.05973,
		0.093385,
		0.125548,
		0.152688,
		0.172554,
		0.184447,
		0.189486,
		0.190482,
		0.190482,
		0.189486,
		0.184447,
		0.172554,
		0.152688,
		0.125548,
		0.093385,
		0.05973,
		0.029175,
		0.007221,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.002811,
		0.016701,
		0.037903,
		0.062113,
		0.085717,
		0.105905,
		0.120838,
		0.12986,
		0.133715,
		0.134483,
		0.134483,
		0.133715,
		0.12986,
		0.120838,
		0.105905,
		0.085717,
		0.062113,
		0.037903,
		0.016701,
		0.002811,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.000729,
		0.008604,
		0.022299,
		0.038636,
		0.054932,
		0.069077,
		0.079658,
		0.086113,
		0.088896,
		0.089455,
		0.089455,
		0.088896,
		0.086113,
		0.079658,
		0.069077,
		0.054932,
		0.038636,
		0.022299,
		0.008604,
		0.000729,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		5.4e-005,
		0.003854,
		0.01193,
		0.022103,
		0.03252,
		0.041711,
		0.048672,
		0.052962,
		0.054829,
		0.055207,
		0.055207,
		0.054829,
		0.052962,
		0.048672,
		0.041711,
		0.03252,
		0.022103,
		0.01193,
		0.003854,
		5.4e-005,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.001417,
		0.005638,
		0.011339,
		0.017358,
		0.022767,
		0.026918,
		0.029504,
		0.030641,
		0.030874,
		0.030874,
		0.030641,
		0.029504,
		0.026918,
		0.022767,
		0.017358,
		0.011339,
		0.005638,
		0.001417,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		0.000384,
		0.002248,
		0.005009,
		0.008033,
		0.010807,
		0.012966,
		0.014328,
		0.014932,
		0.015057,
		0.015057,
		0.014932,
		0.014328,
		0.012966,
		0.010807,
		0.008033,
		0.005009,
		0.002248,
		0.000384,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		6e-005,
		0.000698,
		0.001775,
		0.003008,
		0.004166,
		0.005082,
		0.005667,
		0.005929,
		0.005984,
		0.005984,
		0.005929,
		0.005667,
		0.005082,
		0.004166,
		0.003008,
		0.001775,
		0.000698,
		6e-005,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		0.000144,
		0.000438,
		0.000793,
		0.001136,
		0.001412,
		0.00159,
		0.001671,
		0.001688,
		0.001688,
		0.001671,
		0.00159,
		0.001412,
		0.001136,
		0.000793,
		0.000438,
		0.000144,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		1.3e-005,
		5.2e-005,
		0.000103,
		0.000153,
		0.000194,
		0.000222,
		0.000234,
		0.000237,
		0.000237,
		0.000234,
		0.000222,
		0.000194,
		0.000153,
		0.000103,
		5.2e-005,
		1.3e-005,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		1.3e-005,
		5.2e-005,
		0.000103,
		0.000153,
		0.000194,
		0.000222,
		0.000234,
		0.000237,
		0.000237,
		0.000234,
		0.000222,
		0.000194,
		0.000153,
		0.000103,
		5.2e-005,
		1.3e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		0.000144,
		0.000438,
		0.000793,
		0.001136,
		0.001412,
		0.00159,
		0.001671,
		0.001688,
		0.001688,
		0.001671,
		0.00159,
		0.001412,
		0.001136,
		0.000793,
		0.000438,
		0.000144,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		6e-005,
		0.000698,
		0.001775,
		0.003008,
		0.004166,
		0.005082,
		0.005667,
		0.005929,
		0.005984,
		0.005984,
		0.005929,
		0.005667,
		0.005082,
		0.004166,
		0.003008,
		0.001775,
		0.000698,
		6e-005,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000384,
		0.002248,
		0.005009,
		0.008033,
		0.010807,
		0.012966,
		0.014328,
		0.014932,
		0.015057,
		0.015057,
		0.014932,
		0.014328,
		0.012966,
		0.010807,
		0.008033,
		0.005009,
		0.002248,
		0.000384,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		0.001417,
		0.005638,
		0.011339,
		0.017358,
		0.022767,
		0.026918,
		0.029504,
		0.030641,
		0.030874,
		0.030874,
		0.030641,
		0.029504,
		0.026918,
		0.022767,
		0.017358,
		0.011339,
		0.005638,
		0.001417,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		5.4e-005,
		0.003854,
		0.01193,
		0.022103,
		0.03252,
		0.041711,
		0.048672,
		0.052962,
		0.054829,
		0.055207,
		0.055207,
		0.054829,
		0.052962,
		0.048672,
		0.041711,
		0.03252,
		0.022103,
		0.01193,
		0.003854,
		5.4e-005,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		0.000729,
		0.008604,
		0.022299,
		0.038636,
		0.054932,
		0.069077,
		0.079658,
		0.086113,
		0.088896,
		0.089455,
		0.089455,
		0.088896,
		0.086113,
		0.079658,
		0.069077,
		0.054932,
		0.038636,
		0.022299,
		0.008604,
		0.000729,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.002811,
		0.016701,
		0.037903,
		0.062113,
		0.085717,
		0.105905,
		0.120838,
		0.12986,
		0.133715,
		0.134483,
		0.134483,
		0.133715,
		0.12986,
		0.120838,
		0.105905,
		0.085717,
		0.062113,
		0.037903,
		0.016701,
		0.002811,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.007221,
		0.029175,
		0.05973,
		0.093385,
		0.125548,
		0.152688,
		0.172554,
		0.184447,
		0.189486,
		0.190482,
		0.190482,
		0.189486,
		0.184447,
		0.172554,
		0.152688,
		0.125548,
		0.093385,
		0.05973,
		0.029175,
		0.007221,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.000207,
		0.014927,
		0.046907,
		0.088442,
		0.132827,
		0.174498,
		0.209231,
		0.234406,
		0.249349,
		0.255628,
		0.256861,
		0.256861,
		0.255628,
		0.249349,
		0.234406,
		0.209231,
		0.174498,
		0.132827,
		0.088442,
		0.046907,
		0.014927,
		0.000207,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.002241,
		0.026795,
		0.070476,
		0.124229,
		0.180203,
		0.23193,
		0.274558,
		0.305174,
		0.323198,
		0.330713,
		0.332178,
		0.332178,
		0.330713,
		0.323198,
		0.305174,
		0.274558,
		0.23193,
		0.180203,
		0.124229,
		0.070476,
		0.026795,
		0.002241,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.007219,
		0.043437,
		0.100022,
		0.166698,
		0.234593,
		0.296449,
		0.346893,
		0.38281,
		0.403792,
		0.412477,
		0.414158,
		0.414158,
		0.412477,
		0.403792,
		0.38281,
		0.346893,
		0.296449,
		0.234593,
		0.166698,
		0.100022,
		0.043437,
		0.007219,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.015912,
		0.065084,
		0.135159,
		0.214834,
		0.294395,
		0.365947,
		0.423733,
		0.464545,
		0.48821,
		0.497936,
		0.499807,
		0.499807,
		0.497936,
		0.48821,
		0.464545,
		0.423733,
		0.365947,
		0.294395,
		0.214834,
		0.135159,
		0.065084,
		0.015912,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.000395,
		0.028769,
		0.091506,
		0.174964,
		0.267053,
		0.357431,
		0.43776,
		0.502049,
		0.547106,
		0.573048,
		0.58364,
		0.585664,
		0.585664,
		0.58364,
		0.573048,
		0.547106,
		0.502049,
		0.43776,
		0.357431,
		0.267053,
		0.174964,
		0.091506,
		0.028769,
		0.000395,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.003794,
		0.045838,
		0.122012,
		0.218049,
		0.321345,
		0.42117,
		0.50894,
		0.578592,
		0.627053,
		0.654769,
		0.666012,
		0.668148,
		0.668148,
		0.666012,
		0.654769,
		0.627053,
		0.578592,
		0.50894,
		0.42117,
		0.321345,
		0.218049,
		0.122012,
		0.045838,
		0.003794,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.01098,
		0.066752,
		0.155529,
		0.262731,
		0.375519,
		0.483015,
		0.576593,
		0.650267,
		0.701172,
		0.730099,
		0.741761,
		0.743965,
		0.743965,
		0.741761,
		0.730099,
		0.701172,
		0.650267,
		0.576593,
		0.483015,
		0.375519,
		0.262731,
		0.155529,
		0.066752,
		0.01098,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.021973,
		0.090795,
		0.190753,
		0.30725,
		0.427476,
		0.540631,
		0.638232,
		0.714501,
		0.766854,
		0.79642,
		0.80827,
		0.810497,
		0.810497,
		0.80827,
		0.79642,
		0.766854,
		0.714501,
		0.638232,
		0.540631,
		0.427476,
		0.30725,
		0.190753,
		0.090795,
		0.021973,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.000496,
		0.036428,
		0.117035,
		0.226346,
		0.35001,
		0.475478,
		0.592226,
		0.692067,
		0.769537,
		0.822383,
		0.852052,
		0.863875,
		0.866086,
		0.866086,
		0.863875,
		0.852052,
		0.822383,
		0.769537,
		0.692067,
		0.592226,
		0.475478,
		0.35001,
		0.226346,
		0.117035,
		0.036428,
		0.000496,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.00441,
		0.053745,
		0.14448,
		0.26112,
		0.389776,
		0.518344,
		0.636733,
		0.737171,
		0.814587,
		0.86708,
		0.896384,
		0.907999,
		0.91016,
		0.91016,
		0.907999,
		0.896384,
		0.86708,
		0.814587,
		0.737171,
		0.636733,
		0.518344,
		0.389776,
		0.26112,
		0.14448,
		0.053745,
		0.00441,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.011938,
		0.073204,
		0.172233,
		0.294186,
		0.425793,
		0.555533,
		0.67386,
		0.773497,
		0.84981,
		0.901261,
		0.929828,
		0.941091,
		0.943177,
		0.943177,
		0.941091,
		0.929828,
		0.901261,
		0.84981,
		0.773497,
		0.67386,
		0.555533,
		0.425793,
		0.294186,
		0.172233,
		0.073204,
		0.011938,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.022579,
		0.094096,
		0.199599,
		0.325018,
		0.457808,
		0.587116,
		0.704007,
		0.801745,
		0.876156,
		0.92605,
		0.953609,
		0.964421,
		0.966413,
		0.966413,
		0.964421,
		0.953609,
		0.92605,
		0.876156,
		0.801745,
		0.704007,
		0.587116,
		0.457808,
		0.325018,
		0.199599,
		0.094096,
		0.022579,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.000483,
		0.035748,
		0.115822,
		0.226135,
		0.353452,
		0.486002,
		0.613644,
		0.728083,
		0.82314,
		0.895097,
		0.943094,
		0.969472,
		0.97977,
		0.981659,
		0.981659,
		0.97977,
		0.969472,
		0.943094,
		0.895097,
		0.82314,
		0.728083,
		0.613644,
		0.486002,
		0.353452,
		0.226135,
		0.115822,
		0.035748,
		0.000483,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.004144,
		0.050887,
		0.137938,
		0.25164,
		0.379606,
		0.510855,
		0.635962,
		0.747271,
		0.839153,
		0.908329,
		0.954239,
		0.979348,
		0.989105,
		0.990887,
		0.990887,
		0.989105,
		0.979348,
		0.954239,
		0.908329,
		0.839153,
		0.747271,
		0.635962,
		0.510855,
		0.379606,
		0.25164,
		0.137938,
		0.050887,
		0.004144,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.01093,
		0.067515,
		0.160159,
		0.276097,
		0.40378,
		0.532994,
		0.655014,
		0.762802,
		0.851254,
		0.917501,
		0.961255,
		0.985074,
		0.994288,
		0.995965,
		0.995965,
		0.994288,
		0.985074,
		0.961255,
		0.917501,
		0.851254,
		0.762802,
		0.655014,
		0.532994,
		0.40378,
		0.276097,
		0.160159,
		0.067515,
		0.01093,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.020311,
		0.085259,
		0.182325,
		0.299601,
		0.426337,
		0.55305,
		0.671681,
		0.775775,
		0.860718,
		0.924021,
		0.965637,
		0.98819,
		0.996877,
		0.998451,
		0.998451,
		0.996877,
		0.98819,
		0.965637,
		0.924021,
		0.860718,
		0.775775,
		0.671681,
		0.55305,
		0.426337,
		0.299601,
		0.182325,
		0.085259,
		0.020311,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.000428,
		0.031819,
		0.103831,
		0.204353,
		0.322282,
		0.447615,
		0.57156,
		0.686676,
		0.78705,
		0.868522,
		0.92895,
		0.968498,
		0.989838,
		0.998023,
		0.999501,
		0.999501,
		0.998023,
		0.989838,
		0.968498,
		0.92895,
		0.868522,
		0.78705,
		0.686676,
		0.57156,
		0.447615,
		0.322282,
		0.204353,
		0.103831,
		0.031819,
		0.000428,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.003646,
		0.045061,
		0.123016,
		0.226198,
		0.344254,
		0.467878,
		0.588915,
		0.700499,
		0.797216,
		0.875324,
		0.932992,
		0.970571,
		0.990764,
		0.998478,
		0.999866,
		0.999866,
		0.998478,
		0.990764,
		0.970571,
		0.932992,
		0.875324,
		0.797216,
		0.700499,
		0.588915,
		0.467878,
		0.344254,
		0.226198,
		0.123016,
		0.045061,
		0.003646,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.009606,
		0.059709,
		0.142636,
		0.247816,
		0.365589,
		0.487293,
		0.605362,
		0.713457,
		0.806627,
		0.881506,
		0.936549,
		0.972268,
		0.991385,
		0.998659,
		0.999963,
		0.999963,
		0.998659,
		0.991385,
		0.972268,
		0.936549,
		0.881506,
		0.806627,
		0.713457,
		0.605362,
		0.487293,
		0.365589,
		0.247816,
		0.142636,
		0.059709,
		0.009606,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.01787,
		0.075483,
		0.162543,
		0.269161,
		0.38632,
		0.505951,
		0.621033,
		0.725713,
		0.815461,
		0.887259,
		0.939814,
		0.973782,
		0.99189,
		0.998754,
		0.99998,
		0.99998,
		0.998754,
		0.99189,
		0.973782,
		0.939814,
		0.887259,
		0.815461,
		0.725713,
		0.621033,
		0.505951,
		0.38632,
		0.269161,
		0.162543,
		0.075483,
		0.01787,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.000375,
		0.028064,
		0.092147,
		0.182603,
		0.290179,
		0.40645,
		0.523896,
		0.635997,
		0.737347,
		0.823803,
		0.892661,
		0.942861,
		0.975181,
		0.992346,
		0.998829,
		0.999983,
		0.999983,
		0.998829,
		0.992346,
		0.975181,
		0.942861,
		0.892661,
		0.823803,
		0.737347,
		0.635997,
		0.523896,
		0.40645,
		0.290179,
		0.182603,
		0.092147,
		0.028064,
		0.000375,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.003208,
		0.039864,
		0.109496,
		0.202704,
		0.310823,
		0.425977,
		0.541152,
		0.650291,
		0.7484,
		0.831692,
		0.897749,
		0.945719,
		0.976488,
		0.992769,
		0.998897,
		0.999984,
		0.999984,
		0.998897,
		0.992769,
		0.976488,
		0.945719,
		0.897749,
		0.831692,
		0.7484,
		0.650291,
		0.541152,
		0.425977,
		0.310823,
		0.202704,
		0.109496,
		0.039864,
		0.003208,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.008478,
		0.05299,
		0.127354,
		0.222749,
		0.331053,
		0.444897,
		0.557738,
		0.663946,
		0.758906,
		0.839157,
		0.902545,
		0.948403,
		0.97771,
		0.993163,
		0.99896,
		0.999985,
		0.999985,
		0.99896,
		0.993163,
		0.97771,
		0.948403,
		0.902545,
		0.839157,
		0.758906,
		0.663946,
		0.557738,
		0.444897,
		0.331053,
		0.222749,
		0.127354,
		0.05299,
		0.008478,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.015823,
		0.067199,
		0.145573,
		0.242658,
		0.350837,
		0.463212,
		0.573675,
		0.676989,
		0.768893,
		0.846224,
		0.907068,
		0.950924,
		0.978855,
		0.993531,
		0.999018,
		0.999986,
		0.999986,
		0.999018,
		0.993531,
		0.978855,
		0.950924,
		0.907068,
		0.846224,
		0.768893,
		0.676989,
		0.573675,
		0.463212,
		0.350837,
		0.242658,
		0.145573,
		0.067199,
		0.015823,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.000331,
		0.024926,
		0.082283,
		0.164025,
		0.262364,
		0.370151,
		0.480925,
		0.588982,
		0.68945,
		0.778391,
		0.852917,
		0.911335,
		0.953295,
		0.979927,
		0.993875,
		0.999073,
		0.999987,
		0.999987,
		0.999073,
		0.993875,
		0.979927,
		0.953295,
		0.911335,
		0.852917,
		0.778391,
		0.68945,
		0.588982,
		0.480925,
		0.370151,
		0.262364,
		0.164025,
		0.082283,
		0.024926,
		0.000331,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.002844,
		0.035512,
		0.098062,
		0.182603,
		0.281814,
		0.38898,
		0.498046,
		0.603682,
		0.701355,
		0.787425,
		0.85926,
		0.915365,
		0.955526,
		0.980933,
		0.994196,
		0.999123,
		0.999987,
		0.999987,
		0.999123,
		0.994196,
		0.980933,
		0.955526,
		0.915365,
		0.85926,
		0.787425,
		0.701355,
		0.603682,
		0.498046,
		0.38898,
		0.281814,
		0.182603,
		0.098062,
		0.035512,
		0.002844,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.007538,
		0.047341,
		0.114378,
		0.201216,
		0.300964,
		0.407312,
		0.514584,
		0.617796,
		0.712729,
		0.796022,
		0.865273,
		0.919171,
		0.957627,
		0.981877,
		0.994497,
		0.999171,
		0.999988,
		0.999988,
		0.999171,
		0.994497,
		0.981877,
		0.957627,
		0.919171,
		0.865273,
		0.796022,
		0.712729,
		0.617796,
		0.514584,
		0.407312,
		0.300964,
		0.201216,
		0.114378,
		0.047341,
		0.007538,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.014108,
		0.060202,
		0.131098,
		0.219786,
		0.319778,
		0.425141,
		0.53055,
		0.631345,
		0.723599,
		0.804203,
		0.870975,
		0.92277,
		0.959607,
		0.982764,
		0.994778,
		0.999215,
		0.999989,
		0.999989,
		0.999215,
		0.994778,
		0.982764,
		0.959607,
		0.92277,
		0.870975,
		0.804203,
		0.723599,
		0.631345,
		0.53055,
		0.425141,
		0.319778,
		0.219786,
		0.131098,
		0.060202,
		0.014108,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.000295,
		0.022285,
		0.073915,
		0.148107,
		0.238249,
		0.338228,
		0.442464,
		0.545959,
		0.644352,
		0.733987,
		0.811993,
		0.876386,
		0.926173,
		0.961473,
		0.983597,
		0.995042,
		0.999256,
		0.999989,
		0.999989,
		0.999256,
		0.995042,
		0.983597,
		0.961473,
		0.926173,
		0.876386,
		0.811993,
		0.733987,
		0.644352,
		0.545959,
		0.442464,
		0.338228,
		0.238249,
		0.148107,
		0.073915,
		0.022285,
		0.000295,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.002539,
		0.031834,
		0.088318,
		0.165303,
		0.25655,
		0.356293,
		0.459281,
		0.560823,
		0.656837,
		0.743917,
		0.819413,
		0.881522,
		0.929393,
		0.963233,
		0.984381,
		0.995289,
		0.999295,
		0.99999,
		0.99999,
		0.999295,
		0.995289,
		0.984381,
		0.963233,
		0.929393,
		0.881522,
		0.819413,
		0.743917,
		0.656837,
		0.560823,
		0.459281,
		0.356293,
		0.25655,
		0.165303,
		0.088318,
		0.031834,
		0.002539,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.006746,
		0.042546,
		0.103273,
		0.182603,
		0.274643,
		0.373956,
		0.475597,
		0.575159,
		0.668821,
		0.75341,
		0.826481,
		0.8864,
		0.932443,
		0.964896,
		0.985119,
		0.995521,
		0.999331,
		0.999991,
		0.999991,
		0.999331,
		0.995521,
		0.985119,
		0.964896,
		0.932443,
		0.8864,
		0.826481,
		0.75341,
		0.668821,
		0.575159,
		0.475597,
		0.373956,
		0.274643,
		0.182603,
		0.103273,
		0.042546,
		0.006746,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.012657,
		0.05424,
		0.118659,
		0.199933,
		0.292492,
		0.391206,
		0.491417,
		0.588982,
		0.680324,
		0.762488,
		0.833218,
		0.891034,
		0.935331,
		0.966466,
		0.985814,
		0.995739,
		0.999365,
		0.999991,
		0.999991,
		0.999365,
		0.995739,
		0.985814,
		0.966466,
		0.935331,
		0.891034,
		0.833218,
		0.762488,
		0.680324,
		0.588982,
		0.491417,
		0.391206,
		0.292492,
		0.199933,
		0.118659,
		0.05424,
		0.012657,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.000264,
		0.020042,
		0.066755,
		0.134371,
		0.21723,
		0.310066,
		0.408035,
		0.506746,
		0.602308,
		0.691366,
		0.771171,
		0.83964,
		0.895439,
		0.938069,
		0.96795,
		0.986469,
		0.995944,
		0.999397,
		0.999991,
		0.999991,
		0.999397,
		0.995944,
		0.986469,
		0.96795,
		0.938069,
		0.895439,
		0.83964,
		0.771171,
		0.691366,
		0.602308,
		0.506746,
		0.408035,
		0.310066,
		0.21723,
		0.134371,
		0.066755,
		0.020042,
		0.000264,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.00228,
		0.028698,
		0.079949,
		0.150317,
		0.234441,
		0.32734,
		0.424437,
		0.521595,
		0.615152,
		0.701966,
		0.779477,
		0.845765,
		0.899627,
		0.940665,
		0.969353,
		0.987087,
		0.996137,
		0.999426,
		0.999992,
		0.999992,
		0.999426,
		0.996137,
		0.987087,
		0.969353,
		0.940665,
		0.899627,
		0.845765,
		0.779477,
		0.701966,
		0.615152,
		0.521595,
		0.424437,
		0.32734,
		0.234441,
		0.150317,
		0.079949,
		0.028698,
		0.00228,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.006072,
		0.038444,
		0.093699,
		0.166417,
		0.25152,
		0.344293,
		0.440412,
		0.535972,
		0.62753,
		0.712144,
		0.787425,
		0.851607,
		0.903612,
		0.943128,
		0.970681,
		0.98767,
		0.996318,
		0.999454,
		0.999992,
		0.999992,
		0.999454,
		0.996318,
		0.98767,
		0.970681,
		0.943128,
		0.903612,
		0.851607,
		0.787425,
		0.712144,
		0.62753,
		0.535972,
		0.440412,
		0.344293,
		0.25152,
		0.166417,
		0.093699,
		0.038444,
		0.006072,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.011419,
		0.049119,
		0.107895,
		0.182603,
		0.26843,
		0.360911,
		0.455958,
		0.549888,
		0.63946,
		0.721915,
		0.795032,
		0.857183,
		0.907405,
		0.945467,
		0.971939,
		0.988221,
		0.99649,
		0.999481,
		0.999993,
		0.999993,
		0.999481,
		0.99649,
		0.988221,
		0.971939,
		0.945467,
		0.907405,
		0.857183,
		0.795032,
		0.721915,
		0.63946,
		0.549888,
		0.455958,
		0.360911,
		0.26843,
		0.182603,
		0.107895,
		0.049119,
		0.011419,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.000238,
		0.018122,
		0.060583,
		0.122442,
		0.198815,
		0.285138,
		0.377182,
		0.471079,
		0.563354,
		0.650955,
		0.731298,
		0.802314,
		0.862506,
		0.911016,
		0.947687,
		0.97313,
		0.988742,
		0.996651,
		0.999505,
		0.999993,
		0.999993,
		0.999505,
		0.996651,
		0.988742,
		0.97313,
		0.947687,
		0.911016,
		0.862506,
		0.802314,
		0.731298,
		0.650955,
		0.563354,
		0.471079,
		0.377182,
		0.285138,
		0.198815,
		0.122442,
		0.060583,
		0.018122,
		0.000238,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.002059,
		0.026003,
		0.07271,
		0.137255,
		0.215002,
		0.301618,
		0.393097,
		0.485779,
		0.576381,
		0.662033,
		0.74031,
		0.809287,
		0.86759,
		0.914455,
		0.949797,
		0.97426,
		0.989234,
		0.996803,
		0.999529,
		0.999993,
		0.999993,
		0.999529,
		0.996803,
		0.989234,
		0.97426,
		0.949797,
		0.914455,
		0.86759,
		0.809287,
		0.74031,
		0.662033,
		0.576381,
		0.485779,
		0.393097,
		0.301618,
		0.215002,
		0.137255,
		0.07271,
		0.026003,
		0.002059,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.005495,
		0.034906,
		0.085389,
		0.152262,
		0.231119,
		0.317848,
		0.40865,
		0.500061,
		0.588982,
		0.672708,
		0.748966,
		0.815966,
		0.872446,
		0.917733,
		0.951803,
		0.975331,
		0.9897,
		0.996946,
		0.999551,
		0.999994,
		0.999994,
		0.999551,
		0.996946,
		0.9897,
		0.975331,
		0.951803,
		0.917733,
		0.872446,
		0.815966,
		0.748966,
		0.672708,
		0.588982,
		0.500061,
		0.40865,
		0.317848,
		0.231119,
		0.152262,
		0.085389,
		0.034906,
		0.005495,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.010354,
		0.044689,
		0.098521,
		0.167396,
		0.247127,
		0.333809,
		0.423837,
		0.513933,
		0.601168,
		0.682994,
		0.757282,
		0.822365,
		0.877086,
		0.920857,
		0.953711,
		0.976347,
		0.990141,
		0.997082,
		0.999571,
		0.999994,
		0.999994,
		0.999571,
		0.997082,
		0.990141,
		0.976347,
		0.953711,
		0.920857,
		0.877086,
		0.822365,
		0.757282,
		0.682994,
		0.601168,
		0.513933,
		0.423837,
		0.333809,
		0.247127,
		0.167396,
		0.098521,
		0.044689,
		0.010354,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.000216,
		0.016464,
		0.055226,
		0.112019,
		0.182603,
		0.262996,
		0.349487,
		0.438658,
		0.527402,
		0.612952,
		0.692907,
		0.765272,
		0.828496,
		0.881522,
		0.923837,
		0.955526,
		0.977312,
		0.990558,
		0.99721,
		0.999591,
		0.999994,
		0.999994,
		0.999591,
		0.99721,
		0.990558,
		0.977312,
		0.955526,
		0.923837,
		0.881522,
		0.828496,
		0.765272,
		0.692907,
		0.612952,
		0.527402,
		0.438658,
		0.349487,
		0.262996,
		0.182603,
		0.112019,
		0.055226,
		0.016464,
		0.000216,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.001868,
		0.02367,
		0.066407,
		0.125807,
		0.197833,
		0.278696,
		0.364871,
		0.453111,
		0.540475,
		0.624345,
		0.702461,
		0.77295,
		0.834373,
		0.885764,
		0.92668,
		0.957254,
		0.978228,
		0.990954,
		0.997332,
		0.999609,
		0.999995,
		0.999995,
		0.999609,
		0.997332,
		0.990954,
		0.978228,
		0.957254,
		0.92668,
		0.885764,
		0.834373,
		0.77295,
		0.702461,
		0.624345,
		0.540475,
		0.453111,
		0.364871,
		0.278696,
		0.197833,
		0.125807,
		0.066407,
		0.02367,
		0.001868,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.004996,
		0.031834,
		0.078131,
		0.139816,
		0.213043,
		0.294205,
		0.379951,
		0.4672,
		0.553161,
		0.63536,
		0.711668,
		0.78033,
		0.840008,
		0.889821,
		0.929393,
		0.9589,
		0.979099,
		0.99133,
		0.997446,
		0.999626,
		0.999995,
		0.999995,
		0.999626,
		0.997446,
		0.99133,
		0.979099,
		0.9589,
		0.929393,
		0.889821,
		0.840008,
		0.78033,
		0.711668,
		0.63536,
		0.553161,
		0.4672,
		0.379951,
		0.294205,
		0.213043,
		0.139816,
		0.078131,
		0.031834,
		0.004996,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.009431,
		0.04083,
		0.090309,
		0.153986,
		0.228195,
		0.309504,
		0.394723,
		0.480925,
		0.565468,
		0.64601,
		0.720543,
		0.787425,
		0.845412,
		0.893703,
		0.931984,
		0.960468,
		0.979927,
		0.991686,
		0.997555,
		0.999643,
		0.999995,
		0.999995,
		0.999643,
		0.997555,
		0.991686,
		0.979927,
		0.960468,
		0.931984,
		0.893703,
		0.845412,
		0.787425,
		0.720543,
		0.64601,
		0.565468,
		0.480925,
		0.394723,
		0.309504,
		0.228195,
		0.153986,
		0.090309,
		0.04083,
		0.009431,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.000196,
		0.015024,
		0.050548,
		0.102863,
		0.168264,
		0.243258,
		0.324575,
		0.409179,
		0.494292,
		0.577405,
		0.656305,
		0.729099,
		0.794247,
		0.850596,
		0.89742,
		0.934459,
		0.961963,
		0.980715,
		0.992024,
		0.997658,
		0.999658,
		0.999995,
		0.999995,
		0.999658,
		0.997658,
		0.992024,
		0.980715,
		0.961963,
		0.934459,
		0.89742,
		0.850596,
		0.794247,
		0.729099,
		0.656305,
		0.577405,
		0.494292,
		0.409179,
		0.324575,
		0.243258,
		0.168264,
		0.102863,
		0.050548,
		0.015024,
		0.000196,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.001703,
		0.021638,
		0.060886,
		0.11572,
		0.182603,
		0.258204,
		0.339405,
		0.42332,
		0.507305,
		0.588982,
		0.666257,
		0.737347,
		0.800807,
		0.85557,
		0.900978,
		0.936823,
		0.963389,
		0.981465,
		0.992346,
		0.997755,
		0.999673,
		0.999995,
		0.999995,
		0.999673,
		0.997755,
		0.992346,
		0.981465,
		0.963389,
		0.936823,
		0.900978,
		0.85557,
		0.800807,
		0.737347,
		0.666257,
		0.588982,
		0.507305,
		0.42332,
		0.339405,
		0.258204,
		0.182603,
		0.11572,
		0.060886,
		0.021638,
		0.001703,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.004562,
		0.02915,
		0.071756,
		0.128819,
		0.196963,
		0.273008,
		0.353984,
		0.437142,
		0.519969,
		0.600208,
		0.675879,
		0.745299,
		0.807118,
		0.860344,
		0.904386,
		0.939084,
		0.964748,
		0.982179,
		0.992651,
		0.997848,
		0.999687,
		0.999996,
		0.999996,
		0.999687,
		0.997848,
		0.992651,
		0.982179,
		0.964748,
		0.939084,
		0.904386,
		0.860344,
		0.807118,
		0.745299,
		0.675879,
		0.600208,
		0.519969,
		0.437142,
		0.353984,
		0.273008,
		0.196963,
		0.128819,
		0.071756,
		0.02915,
		0.004562,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.008627,
		0.03745,
		0.083077,
		0.142104,
		0.211307,
		0.287651,
		0.368303,
		0.450646,
		0.53229,
		0.611093,
		0.68518,
		0.752968,
		0.813189,
		0.864928,
		0.907651,
		0.941245,
		0.966046,
		0.982859,
		0.992941,
		0.997936,
		0.9997,
		0.999996,
		0.999996,
		0.9997,
		0.997936,
		0.992941,
		0.982859,
		0.966046,
		0.941245,
		0.907651,
		0.864928,
		0.813189,
		0.752968,
		0.68518,
		0.611093,
		0.53229,
		0.450646,
		0.368303,
		0.287651,
		0.211307,
		0.142104,
		0.083077,
		0.03745,
		0.008627,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.000179,
		0.013765,
		0.046438,
		0.094776,
		0.155525,
		0.225603,
		0.302114,
		0.382355,
		0.463832,
		0.544274,
		0.621646,
		0.694173,
		0.760363,
		0.819031,
		0.869329,
		0.91078,
		0.943313,
		0.967285,
		0.983507,
		0.993217,
		0.998019,
		0.999713,
		0.999996,
		0.999996,
		0.999713,
		0.998019,
		0.993217,
		0.983507,
		0.967285,
		0.943313,
		0.91078,
		0.869329,
		0.819031,
		0.760363,
		0.694173,
		0.621646,
		0.544274,
		0.463832,
		0.382355,
		0.302114,
		0.225603,
		0.155525,
		0.094776,
		0.046438,
		0.013765,
		0.000179,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.001559,
		0.019856,
		0.056024,
		0.106789,
		0.169038,
		0.239825,
		0.316383,
		0.396135,
		0.476704,
		0.555928,
		0.631876,
		0.702868,
		0.767497,
		0.824654,
		0.873557,
		0.913781,
		0.945291,
		0.968469,
		0.984125,
		0.99348,
		0.998099,
		0.999725,
		0.999996,
		0.999996,
		0.999725,
		0.998099,
		0.99348,
		0.984125,
		0.968469,
		0.945291,
		0.913781,
		0.873557,
		0.824654,
		0.767497,
		0.702868,
		0.631876,
		0.555928,
		0.476704,
		0.396135,
		0.316383,
		0.239825,
		0.169038,
		0.106789,
		0.056024,
		0.019856,
		0.001559,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.004182,
		0.026791,
		0.066128,
		0.119058,
		0.182603,
		0.253947,
		0.330446,
		0.409641,
		0.489264,
		0.567259,
		0.641793,
		0.711275,
		0.774379,
		0.830067,
		0.877618,
		0.916658,
		0.947185,
		0.9696,
		0.984714,
		0.993731,
		0.998174,
		0.999736,
		0.999996,
		0.999996,
		0.999736,
		0.998174,
		0.993731,
		0.984714,
		0.9696,
		0.947185,
		0.916658,
		0.877618,
		0.830067,
		0.774379,
		0.711275,
		0.641793,
		0.567259,
		0.489264,
		0.409641,
		0.330446,
		0.253947,
		0.182603,
		0.119058,
		0.066128,
		0.026791,
		0.004182,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.007921,
		0.034471,
		0.076675,
		0.131531,
		0.196187,
		0.26795,
		0.344293,
		0.422868,
		0.501515,
		0.578275,
		0.651406,
		0.719404,
		0.781018,
		0.835279,
		0.881522,
		0.919418,
		0.948998,
		0.970681,
		0.985277,
		0.993969,
		0.998246,
		0.999747,
		0.999997,
		0.999997,
		0.999747,
		0.998246,
		0.993969,
		0.985277,
		0.970681,
		0.948998,
		0.919418,
		0.881522,
		0.835279,
		0.781018,
		0.719404,
		0.651406,
		0.578275,
		0.501515,
		0.422868,
		0.344293,
		0.26795,
		0.196187,
		0.131531,
		0.076675,
		0.034471,
		0.007921,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.000165,
		0.012657,
		0.042808,
		0.087601,
		0.144161,
		0.209758,
		0.281814,
		0.357915,
		0.435818,
		0.513462,
		0.588982,
		0.660725,
		0.727265,
		0.787425,
		0.840298,
		0.885274,
		0.922066,
		0.950736,
		0.971715,
		0.985814,
		0.994196,
		0.998314,
		0.999757,
		0.999997,
		0.999997,
		0.999757,
		0.998314,
		0.994196,
		0.985814,
		0.971715,
		0.950736,
		0.922066,
		0.885274,
		0.840298,
		0.787425,
		0.727265,
		0.660725,
		0.588982,
		0.513462,
		0.435818,
		0.357915,
		0.281814,
		0.209758,
		0.144161,
		0.087601,
		0.042808,
		0.012657,
		0.000165,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.001432,
		0.018285,
		0.05172,
		0.098846,
		0.156908,
		0.223289,
		0.295525,
		0.371306,
		0.448489,
		0.525108,
		0.599388,
		0.669758,
		0.734868,
		0.793609,
		0.845133,
		0.888882,
		0.924608,
		0.952401,
		0.972704,
		0.986327,
		0.994413,
		0.998379,
		0.999766,
		0.999997,
		0.999997,
		0.999766,
		0.998379,
		0.994413,
		0.986327,
		0.972704,
		0.952401,
		0.924608,
		0.888882,
		0.845133,
		0.793609,
		0.734868,
		0.669758,
		0.599388,
		0.525108,
		0.448489,
		0.371306,
		0.295525,
		0.223289,
		0.156908,
		0.098846,
		0.05172,
		0.018285,
		0.001432,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.003848,
		0.024706,
		0.061133,
		0.110356,
		0.169733,
		0.236757,
		0.30907,
		0.38446,
		0.460882,
		0.53646,
		0.609502,
		0.678515,
		0.742222,
		0.799577,
		0.849792,
		0.892353,
		0.927049,
		0.953996,
		0.973651,
		0.986817,
		0.99462,
		0.998441,
		0.999775,
		0.999997,
		0.999997,
		0.999775,
		0.998441,
		0.99462,
		0.986817,
		0.973651,
		0.953996,
		0.927049,
		0.892353,
		0.849792,
		0.799577,
		0.742222,
		0.678515,
		0.609502,
		0.53646,
		0.460882,
		0.38446,
		0.30907,
		0.236757,
		0.169733,
		0.110356,
		0.061133,
		0.024706,
		0.003848,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.007298,
		0.031834,
		0.070982,
		0.122084,
		0.182603,
		0.250141,
		0.322436,
		0.397375,
		0.473,
		0.547521,
		0.619329,
		0.687003,
		0.749335,
		0.80534,
		0.854281,
		0.895691,
		0.929393,
		0.955526,
		0.974557,
		0.987285,
		0.994817,
		0.998499,
		0.999784,
		0.999997,
		0.999997,
		0.999784,
		0.998499,
		0.994817,
		0.987285,
		0.974557,
		0.955526,
		0.929393,
		0.895691,
		0.854281,
		0.80534,
		0.749335,
		0.687003,
		0.619329,
		0.547521,
		0.473,
		0.397375,
		0.322436,
		0.250141,
		0.182603,
		0.122084,
		0.070982,
		0.031834,
		0.007298,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.000152,
		0.011678,
		0.039587,
		0.081207,
		0.133985,
		0.19549,
		0.263423,
		0.335615,
		0.410046,
		0.484843,
		0.558299,
		0.628878,
		0.695232,
		0.756216,
		0.810904,
		0.858608,
		0.898903,
		0.931646,
		0.956994,
		0.975425,
		0.987733,
		0.995005,
		0.998556,
		0.999793,
		0.999997,
		0.999997,
		0.999793,
		0.998556,
		0.995005,
		0.987733,
		0.975425,
		0.956994,
		0.931646,
		0.898903,
		0.858608,
		0.810904,
		0.756216,
		0.695232,
		0.628878,
		0.558299,
		0.484843,
		0.410046,
		0.335615,
		0.263423,
		0.19549,
		0.133985,
		0.081207,
		0.039587,
		0.011678,
		0.000152,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.00132,
		0.016893,
		0.047892,
		0.091751,
		0.14602,
		0.208367,
		0.276585,
		0.348599,
		0.422472,
		0.496415,
		0.568797,
		0.638156,
		0.70321,
		0.762874,
		0.816277,
		0.86278,
		0.901995,
		0.93381,
		0.958402,
		0.976257,
		0.988161,
		0.995185,
		0.998609,
		0.9998,
		0.999997,
		0.999997,
		0.9998,
		0.998609,
		0.995185,
		0.988161,
		0.976257,
		0.958402,
		0.93381,
		0.901995,
		0.86278,
		0.816277,
		0.762874,
		0.70321,
		0.638156,
		0.568797,
		0.496415,
		0.422472,
		0.348599,
		0.276585,
		0.208367,
		0.14602,
		0.091751,
		0.047892,
		0.016893,
		0.00132,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.003552,
		0.022856,
		0.056682,
		0.102567,
		0.158156,
		0.221211,
		0.289616,
		0.361381,
		0.434652,
		0.507719,
		0.579023,
		0.647171,
		0.710944,
		0.769316,
		0.821467,
		0.866803,
		0.904972,
		0.93589,
		0.959754,
		0.977053,
		0.988571,
		0.995357,
		0.99866,
		0.999808,
		0.999997,
		0.999997,
		0.999808,
		0.99866,
		0.995357,
		0.988571,
		0.977053,
		0.959754,
		0.93589,
		0.904972,
		0.866803,
		0.821467,
		0.769316,
		0.710944,
		0.647171,
		0.579023,
		0.507719,
		0.434652,
		0.361381,
		0.289616,
		0.221211,
		0.158156,
		0.102567,
		0.056682,
		0.022856,
		0.003552,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.006746,
		0.029488,
		0.065898,
		0.113609,
		0.17036,
		0.234001,
		0.302502,
		0.373956,
		0.446587,
		0.518758,
		0.588982,
		0.655929,
		0.718443,
		0.775551,
		0.826481,
		0.870682,
		0.907839,
		0.937891,
		0.961052,
		0.977817,
		0.988964,
		0.995521,
		0.998709,
		0.999815,
		0.999997,
		0.999997,
		0.999815,
		0.998709,
		0.995521,
		0.988964,
		0.977817,
		0.961052,
		0.937891,
		0.907839,
		0.870682,
		0.826481,
		0.775551,
		0.718443,
		0.655929,
		0.588982,
		0.518758,
		0.446587,
		0.373956,
		0.302502,
		0.234001,
		0.17036,
		0.113609,
		0.065898,
		0.029488,
		0.006746,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.00014,
		0.010808,
		0.036715,
		0.075484,
		0.124837,
		0.182603,
		0.246719,
		0.315234,
		0.38632,
		0.458277,
		0.529537,
		0.59868,
		0.664439,
		0.725714,
		0.781584,
		0.831326,
		0.874425,
		0.9106,
		0.939815,
		0.962298,
		0.97855,
		0.98934,
		0.995678,
		0.998756,
		0.999822,
		0.999998,
		0.999998,
		0.999822,
		0.998756,
		0.995678,
		0.98934,
		0.97855,
		0.962298,
		0.939815,
		0.9106,
		0.874425,
		0.831326,
		0.781584,
		0.725714,
		0.664439,
		0.59868,
		0.529537,
		0.458277,
		0.38632,
		0.315234,
		0.246719,
		0.182603,
		0.124837,
		0.075484,
		0.036715,
		0.010808,
		0.00014,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.001221,
		0.015654,
		0.044472,
		0.085389,
		0.136215,
		0.194862,
		0.259348,
		0.327802,
		0.39847,
		0.469722,
		0.540059,
		0.608124,
		0.672708,
		0.732765,
		0.787425,
		0.836008,
		0.878037,
		0.91326,
		0.941666,
		0.963495,
		0.979253,
		0.9897,
		0.995829,
		0.9988,
		0.999829,
		0.999998,
		0.999998,
		0.999829,
		0.9988,
		0.995829,
		0.9897,
		0.979253,
		0.963495,
		0.941666,
		0.91326,
		0.878037,
		0.836008,
		0.787425,
		0.732765,
		0.672708,
		0.608124,
		0.540059,
		0.469722,
		0.39847,
		0.327802,
		0.259348,
		0.194862,
		0.136215,
		0.085389,
		0.044472,
		0.015654,
		0.001221,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.003289,
		0.021206,
		0.052699,
		0.095568,
		0.147708,
		0.207112,
		0.271875,
		0.3402,
		0.410404,
		0.480925,
		0.55033,
		0.617319,
		0.680741,
		0.739603,
		0.79308,
		0.840533,
		0.881522,
		0.915824,
		0.943447,
		0.964646,
		0.979927,
		0.990045,
		0.995972,
		0.998843,
		0.999835,
		0.999998,
		0.999998,
		0.999835,
		0.998843,
		0.995972,
		0.990045,
		0.979927,
		0.964646,
		0.943447,
		0.915824,
		0.881522,
		0.840533,
		0.79308,
		0.739603,
		0.680741,
		0.617319,
		0.55033,
		0.480925,
		0.410404,
		0.3402,
		0.271875,
		0.207112,
		0.147708,
		0.095568,
		0.052699,
		0.021206,
		0.003289,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.006254,
		0.027391,
		0.061339,
		0.10598,
		0.159289,
		0.219335,
		0.284287,
		0.35242,
		0.42212,
		0.491888,
		0.560352,
		0.626271,
		0.688547,
		0.746234,
		0.798554,
		0.844907,
		0.884886,
		0.918295,
		0.945162,
		0.965752,
		0.980575,
		0.990375,
		0.99611,
		0.998883,
		0.999841,
		0.999998,
		0.999998,
		0.999841,
		0.998883,
		0.99611,
		0.990375,
		0.980575,
		0.965752,
		0.945162,
		0.918295,
		0.884886,
		0.844907,
		0.798554,
		0.746234,
		0.688547,
		0.626271,
		0.560352,
		0.491888,
		0.42212,
		0.35242,
		0.284287,
		0.219335,
		0.159289,
		0.10598,
		0.061339,
		0.027391,
		0.006254,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.00013,
		0.010032,
		0.034144,
		0.070342,
		0.116587,
		0.170928,
		0.231511,
		0.296573,
		0.364459,
		0.433618,
		0.502614,
		0.570132,
		0.634987,
		0.696131,
		0.752666,
		0.803856,
		0.849137,
		0.888134,
		0.920678,
		0.946813,
		0.966816,
		0.981196,
		0.990693,
		0.996242,
		0.998922,
		0.999846,
		0.999998,
		0.999998,
		0.999846,
		0.998922,
		0.996242,
		0.990693,
		0.981196,
		0.966816,
		0.946813,
		0.920678,
		0.888134,
		0.849137,
		0.803856,
		0.752666,
		0.696131,
		0.634987,
		0.570132,
		0.502614,
		0.433618,
		0.364459,
		0.296573,
		0.231511,
		0.170928,
		0.116587,
		0.070342,
		0.034144,
		0.010032,
		0.00013,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.001133,
		0.014547,
		0.041405,
		0.079662,
		0.127354,
		0.182603,
		0.243624,
		0.308725,
		0.376311,
		0.444897,
		0.513104,
		0.579674,
		0.643472,
		0.7035,
		0.758906,
		0.80899,
		0.853227,
		0.891271,
		0.922975,
		0.948403,
		0.967838,
		0.981794,
		0.990997,
		0.996368,
		0.99896,
		0.999852,
		0.999998,
		0.999998,
		0.999852,
		0.99896,
		0.996368,
		0.990997,
		0.981794,
		0.967838,
		0.948403,
		0.922975,
		0.891271,
		0.853227,
		0.80899,
		0.758906,
		0.7035,
		0.643472,
		0.579674,
		0.513104,
		0.444897,
		0.376311,
		0.308725,
		0.243624,
		0.182603,
		0.127354,
		0.079662,
		0.041405,
		0.014547,
		0.001133,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.003055,
		0.019728,
		0.049119,
		0.089257,
		0.13825,
		0.194292,
		0.255662,
		0.320733,
		0.387975,
		0.455958,
		0.523363,
		0.588982,
		0.651733,
		0.710662,
		0.764959,
		0.813964,
		0.857183,
		0.894301,
		0.925192,
		0.949934,
		0.968823,
		0.982367,
		0.991289,
		0.99649,
		0.998995,
		0.999857,
		0.999998,
		0.999998,
		0.999857,
		0.998995,
		0.99649,
		0.991289,
		0.982367,
		0.968823,
		0.949934,
		0.925192,
		0.894301,
		0.857183,
		0.813964,
		0.764959,
		0.710662,
		0.651733,
		0.588982,
		0.523363,
		0.455958,
		0.387975,
		0.320733,
		0.255662,
		0.194292,
		0.13825,
		0.089257,
		0.049119,
		0.019728,
		0.003055,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.005814,
		0.02551,
		0.057235,
		0.099088,
		0.149248,
		0.205974,
		0.26761,
		0.332591,
		0.399446,
		0.466802,
		0.533393,
		0.598062,
		0.659775,
		0.717621,
		0.770831,
		0.818782,
		0.861011,
		0.897228,
		0.92733,
		0.95141,
		0.96977,
		0.982919,
		0.991569,
		0.996606,
		0.999029,
		0.999862,
		0.999998,
		0.999998,
		0.999862,
		0.999029,
		0.996606,
		0.991569,
		0.982919,
		0.96977,
		0.95141,
		0.92733,
		0.897228,
		0.861011,
		0.818782,
		0.770831,
		0.717621,
		0.659775,
		0.598062,
		0.533393,
		0.466802,
		0.399446,
		0.332591,
		0.26761,
		0.205974,
		0.149248,
		0.099088,
		0.057235,
		0.02551,
		0.005814,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.00012,
		0.009337,
		0.031834,
		0.065706,
		0.109121,
		0.160321,
		0.217631,
		0.279458,
		0.344293,
		0.410723,
		0.47743,
		0.543198,
		0.606919,
		0.667604,
		0.724384,
		0.776529,
		0.82345,
		0.864714,
		0.900056,
		0.929393,
		0.952833,
		0.970681,
		0.983449,
		0.991839,
		0.996717,
		0.999062,
		0.999867,
		0.999998,
		0.999998,
		0.999867,
		0.999062,
		0.996717,
		0.991839,
		0.983449,
		0.970681,
		0.952833,
		0.929393,
		0.900056,
		0.864714,
		0.82345,
		0.776529,
		0.724384,
		0.667604,
		0.606919,
		0.543198,
		0.47743,
		0.410723,
		0.344293,
		0.279458,
		0.217631,
		0.160321,
		0.109121,
		0.065706,
		0.031834,
		0.009337,
		0.00012,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.001053,
		0.013553,
		0.038645,
		0.07449,
		0.119322,
		0.171447,
		0.229249,
		0.291195,
		0.355834,
		0.421806,
		0.487844,
		0.552782,
		0.615558,
		0.675225,
		0.730958,
		0.782059,
		0.827973,
		0.868297,
		0.90279,
		0.931385,
		0.954204,
		0.971559,
		0.983959,
		0.992097,
		0.996824,
		0.999093,
		0.999871,
		0.999998,
		0.999998,
		0.999871,
		0.999093,
		0.996824,
		0.992097,
		0.983959,
		0.971559,
		0.954204,
		0.931385,
		0.90279,
		0.868297,
		0.827973,
		0.782059,
		0.730958,
		0.675225,
		0.615558,
		0.552782,
		0.487844,
		0.421806,
		0.355834,
		0.291195,
		0.229249,
		0.171447,
		0.119322,
		0.07449,
		0.038645,
		0.013553,
		0.001053,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.002844,
		0.018399,
		0.045892,
		0.083548,
		0.129662,
		0.182603,
		0.240813,
		0.302814,
		0.367211,
		0.432694,
		0.498046,
		0.562148,
		0.623984,
		0.682646,
		0.737347,
		0.787425,
		0.832357,
		0.871765,
		0.905432,
		0.933308,
		0.955526,
		0.972405,
		0.98445,
		0.992346,
		0.996926,
		0.999123,
		0.999876,
		0.999998,
		0.999998,
		0.999876,
		0.999123,
		0.996926,
		0.992346,
		0.98445,
		0.972405,
		0.955526,
		0.933308,
		0.905432,
		0.871765,
		0.832357,
		0.787425,
		0.737347,
		0.682646,
		0.623984,
		0.562148,
		0.498046,
		0.432694,
		0.367211,
		0.302814,
		0.240813,
		0.182603,
		0.129662,
		0.083548,
		0.045892,
		0.018399,
		0.002844,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.005419,
		0.023816,
		0.053528,
		0.092843,
		0.140116,
		0.193772,
		0.252311,
		0.314308,
		0.378419,
		0.443386,
		0.508038,
		0.571301,
		0.632201,
		0.68987,
		0.743557,
		0.792634,
		0.836606,
		0.875123,
		0.907987,
		0.935165,
		0.956802,
		0.973219,
		0.984922,
		0.992584,
		0.997025,
		0.999152,
		0.99988,
		0.999998,
		0.999998,
		0.99988,
		0.999152,
		0.997025,
		0.992584,
		0.984922,
		0.973219,
		0.956802,
		0.935165,
		0.907987,
		0.875123,
		0.836606,
		0.792634,
		0.743557,
		0.68987,
		0.632201,
		0.571301,
		0.508038,
		0.443386,
		0.378419,
		0.314308,
		0.252311,
		0.193772,
		0.140116,
		0.092843,
		0.053528,
		0.023816,
		0.005419,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.000112,
		0.008711,
		0.02975,
		0.061512,
		0.102344,
		0.150658,
		0.204936,
		0.26373,
		0.325668,
		0.389456,
		0.453882,
		0.517822,
		0.580244,
		0.640214,
		0.696903,
		0.749594,
		0.79769,
		0.840726,
		0.878374,
		0.910457,
		0.936959,
		0.958032,
		0.974004,
		0.985376,
		0.992814,
		0.997119,
		0.99918,
		0.999884,
		0.999998,
		0.999998,
		0.999884,
		0.99918,
		0.997119,
		0.992814,
		0.985376,
		0.974004,
		0.958032,
		0.936959,
		0.910457,
		0.878374,
		0.840726,
		0.79769,
		0.749594,
		0.696903,
		0.640214,
		0.580244,
		0.517822,
		0.453882,
		0.389456,
		0.325668,
		0.26373,
		0.204936,
		0.150658,
		0.102344,
		0.061512,
		0.02975,
		0.008711,
		0.000112,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.000982,
		0.012657,
		0.03615,
		0.069803,
		0.112019,
		0.161266,
		0.216079,
		0.275061,
		0.336891,
		0.40032,
		0.464184,
		0.527402,
		0.588982,
		0.648029,
		0.703751,
		0.755463,
		0.802599,
		0.84472,
		0.881522,
		0.912847,
		0.938692,
		0.95922,
		0.974761,
		0.985814,
		0.993035,
		0.99721,
		0.999206,
		0.999888,
		0.999998,
		0.999998,
		0.999888,
		0.999206,
		0.99721,
		0.993035,
		0.985814,
		0.974761,
		0.95922,
		0.938692,
		0.912847,
		0.881522,
		0.84472,
		0.802599,
		0.755463,
		0.703751,
		0.648029,
		0.588982,
		0.527402,
		0.464184,
		0.40032,
		0.336891,
		0.275061,
		0.216079,
		0.161266,
		0.112019,
		0.069803,
		0.03615,
		0.012657,
		0.000982,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.002655,
		0.0172,
		0.042971,
		0.078366,
		0.121842,
		0.171921,
		0.227188,
		0.286296,
		0.34797,
		0.41101,
		0.474293,
		0.53678,
		0.597518,
		0.655651,
		0.710418,
		0.761169,
		0.807365,
		0.848593,
		0.884571,
		0.915158,
		0.940367,
		0.960366,
		0.97549,
		0.986235,
		0.993247,
		0.997298,
		0.999232,
		0.999891,
		0.999999,
		0.999999,
		0.999891,
		0.999232,
		0.997298,
		0.993247,
		0.986235,
		0.97549,
		0.960366,
		0.940367,
		0.915158,
		0.884571,
		0.848593,
		0.807365,
		0.761169,
		0.710418,
		0.655651,
		0.597518,
		0.53678,
		0.474293,
		0.41101,
		0.34797,
		0.286296,
		0.227188,
		0.171921,
		0.121842,
		0.078366,
		0.042971,
		0.0172,
		0.002655,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.005063,
		0.022285,
		0.050169,
		0.087167,
		0.131787,
		0.182603,
		0.238249,
		0.297425,
		0.358903,
		0.421525,
		0.48421,
		0.545959,
		0.605858,
		0.663083,
		0.71691,
		0.766717,
		0.811993,
		0.85235,
		0.887524,
		0.917395,
		0.941985,
		0.961473,
		0.976194,
		0.986641,
		0.993452,
		0.997381,
		0.999256,
		0.999895,
		0.999999,
		0.999999,
		0.999895,
		0.999256,
		0.997381,
		0.993452,
		0.986641,
		0.976194,
		0.961473,
		0.941985,
		0.917395,
		0.887524,
		0.85235,
		0.811993,
		0.766717,
		0.71691,
		0.663083,
		0.605858,
		0.545959,
		0.48421,
		0.421525,
		0.358903,
		0.297425,
		0.238249,
		0.182603,
		0.131787,
		0.087167,
		0.050169,
		0.022285,
		0.005063,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.000105,
		0.008146,
		0.027864,
		0.057705,
		0.096175,
		0.141832,
		0.193297,
		0.249252,
		0.308444,
		0.369686,
		0.431863,
		0.493936,
		0.554942,
		0.614003,
		0.670332,
		0.723232,
		0.772112,
		0.816488,
		0.855993,
		0.890386,
		0.91956,
		0.94355,
		0.962541,
		0.976872,
		0.987032,
		0.993648,
		0.997462,
		0.99928,
		0.999898,
		0.999999,
		0.999999,
		0.999898,
		0.99928,
		0.997462,
		0.993648,
		0.987032,
		0.976872,
		0.962541,
		0.94355,
		0.91956,
		0.890386,
		0.855993,
		0.816488,
		0.772112,
		0.723232,
		0.670332,
		0.614003,
		0.554942,
		0.493936,
		0.431863,
		0.369686,
		0.308444,
		0.249252,
		0.193297,
		0.141832,
		0.096175,
		0.057705,
		0.027864,
		0.008146,
		0.000105,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.000918,
		0.011847,
		0.033889,
		0.065543,
		0.105362,
		0.151954,
		0.203986,
		0.260186,
		0.319344,
		0.380315,
		0.442026,
		0.503473,
		0.563732,
		0.62196,
		0.6774,
		0.729388,
		0.777359,
		0.820854,
		0.859528,
		0.893159,
		0.921655,
		0.945063,
		0.963574,
		0.977527,
		0.987409,
		0.993838,
		0.99754,
		0.999302,
		0.999902,
		0.999999,
		0.999999,
		0.999902,
		0.999302,
		0.99754,
		0.993838,
		0.987409,
		0.977527,
		0.963574,
		0.945063,
		0.921655,
		0.893159,
		0.859528,
		0.820854,
		0.777359,
		0.729388,
		0.6774,
		0.62196,
		0.563732,
		0.503473,
		0.442026,
		0.380315,
		0.319344,
		0.260186,
		0.203986,
		0.151954,
		0.105362,
		0.065543,
		0.033889,
		0.011847,
		0.000918,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.002483,
		0.016114,
		0.04032,
		0.073649,
		0.114702,
		0.162135,
		0.214658,
		0.271043,
		0.330122,
		0.39079,
		0.452013,
		0.512824,
		0.572334,
		0.629732,
		0.684294,
		0.735383,
		0.782462,
		0.825095,
		0.862958,
		0.895847,
		0.923684,
		0.946526,
		0.964571,
		0.978159,
		0.987773,
		0.99402,
		0.997614,
		0.999324,
		0.999905,
		0.999999,
		0.999999,
		0.999905,
		0.999324,
		0.997614,
		0.99402,
		0.987773,
		0.978159,
		0.964571,
		0.946526,
		0.923684,
		0.895847,
		0.862958,
		0.825095,
		0.782462,
		0.735383,
		0.684294,
		0.629732,
		0.572334,
		0.512824,
		0.452013,
		0.39079,
		0.330122,
		0.271043,
		0.214658,
		0.162135,
		0.114702,
		0.073649,
		0.04032,
		0.016114,
		0.002483,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.004741,
		0.020897,
		0.047116,
		0.081992,
		0.124172,
		0.172357,
		0.2253,
		0.281814,
		0.340772,
		0.401108,
		0.461824,
		0.52199,
		0.580749,
		0.637323,
		0.691017,
		0.741222,
		0.787425,
		0.829215,
		0.866286,
		0.898452,
		0.925648,
		0.947941,
		0.965534,
		0.978769,
		0.988123,
		0.994196,
		0.997686,
		0.999345,
		0.999908,
		0.999999,
		0.999999,
		0.999908,
		0.999345,
		0.997686,
		0.994196,
		0.988123,
		0.978769,
		0.965534,
		0.947941,
		0.925648,
		0.898452,
		0.866286,
		0.829215,
		0.787425,
		0.741222,
		0.691017,
		0.637323,
		0.580749,
		0.52199,
		0.461824,
		0.401108,
		0.340772,
		0.281814,
		0.2253,
		0.172357,
		0.124172,
		0.081992,
		0.047116,
		0.020897,
		0.004741,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		9.8e-005,
		0.007634,
		0.026151,
		0.05424,
		0.090543,
		0.13375,
		0.182603,
		0.2359,
		0.292492,
		0.351291,
		0.411269,
		0.471461,
		0.530974,
		0.588982,
		0.644738,
		0.697574,
		0.746909,
		0.792253,
		0.833218,
		0.869516,
		0.900978,
		0.927551,
		0.949311,
		0.966466,
		0.979358,
		0.988461,
		0.994366,
		0.997755,
		0.999365,
		0.999911,
		0.999999,
		0.999999,
		0.999911,
		0.999365,
		0.997755,
		0.994366,
		0.988461,
		0.979358,
		0.966466,
		0.949311,
		0.927551,
		0.900978,
		0.869516,
		0.833218,
		0.792253,
		0.746909,
		0.697574,
		0.644738,
		0.588982,
		0.530974,
		0.471461,
		0.411269,
		0.351291,
		0.292492,
		0.2359,
		0.182603,
		0.13375,
		0.090543,
		0.05424,
		0.026151,
		0.007634,
		9.8e-005,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		0.00086,
		0.011113,
		0.031834,
		0.06166,
		0.099276,
		0.143415,
		0.19286,
		0.246449,
		0.303071,
		0.361676,
		0.42127,
		0.480925,
		0.539778,
		0.597036,
		0.651979,
		0.703969,
		0.752448,
		0.796951,
		0.837108,
		0.872652,
		0.903427,
		0.929393,
		0.950636,
		0.967366,
		0.979927,
		0.988787,
		0.994529,
		0.997822,
		0.999384,
		0.999913,
		0.999999,
		0.999999,
		0.999913,
		0.999384,
		0.997822,
		0.994529,
		0.988787,
		0.979927,
		0.967366,
		0.950636,
		0.929393,
		0.903427,
		0.872652,
		0.837108,
		0.796951,
		0.752448,
		0.703969,
		0.651979,
		0.597036,
		0.539778,
		0.480925,
		0.42127,
		0.361676,
		0.303071,
		0.246449,
		0.19286,
		0.143415,
		0.099276,
		0.06166,
		0.031834,
		0.011113,
		0.00086,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.002328,
		0.015128,
		0.037906,
		0.069344,
		0.108167,
		0.153149,
		0.203114,
		0.256937,
		0.313545,
		0.371923,
		0.431113,
		0.490217,
		0.548406,
		0.604914,
		0.659052,
		0.710206,
		0.757843,
		0.801521,
		0.840888,
		0.875696,
		0.905802,
		0.931179,
		0.951918,
		0.968237,
		0.980477,
		0.989102,
		0.994686,
		0.997886,
		0.999403,
		0.999916,
		0.999999,
		0.999999,
		0.999916,
		0.999403,
		0.997886,
		0.994686,
		0.989102,
		0.980477,
		0.968237,
		0.951918,
		0.931179,
		0.905802,
		0.875696,
		0.840888,
		0.801521,
		0.757843,
		0.710206,
		0.659052,
		0.604914,
		0.548406,
		0.490217,
		0.431113,
		0.371923,
		0.313545,
		0.256937,
		0.203114,
		0.153149,
		0.108167,
		0.069344,
		0.037906,
		0.015128,
		0.002328,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		0.004448,
		0.019635,
		0.044332,
		0.077262,
		0.117192,
		0.162935,
		0.213353,
		0.267356,
		0.32391,
		0.382031,
		0.440796,
		0.499339,
		0.556859,
		0.612621,
		0.66596,
		0.71629,
		0.7631,
		0.805968,
		0.844563,
		0.878652,
		0.908106,
		0.932909,
		0.95316,
		0.969079,
		0.981008,
		0.989406,
		0.994838,
		0.997948,
		0.999421,
		0.999919,
		0.999999,
		0.999999,
		0.999919,
		0.999421,
		0.997948,
		0.994838,
		0.989406,
		0.981008,
		0.969079,
		0.95316,
		0.932909,
		0.908106,
		0.878652,
		0.844563,
		0.805968,
		0.7631,
		0.71629,
		0.66596,
		0.612621,
		0.556859,
		0.499339,
		0.440796,
		0.382031,
		0.32391,
		0.267356,
		0.213353,
		0.162935,
		0.117192,
		0.077262,
		0.044332,
		0.019635,
		0.004448,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		9.2e-005,
		0.00717,
		0.024591,
		0.051077,
		0.085389,
		0.126332,
		0.172758,
		0.223565,
		0.2777,
		0.33416,
		0.391998,
		0.45032,
		0.508291,
		0.56514,
		0.620158,
		0.672708,
		0.722224,
		0.76822,
		0.810295,
		0.848135,
		0.881522,
		0.910341,
		0.934586,
		0.954362,
		0.969893,
		0.981521,
		0.9897,
		0.994985,
		0.998008,
		0.999438,
		0.999921,
		0.999999,
		0.999999,
		0.999921,
		0.999438,
		0.998008,
		0.994985,
		0.9897,
		0.981521,
		0.969893,
		0.954362,
		0.934586,
		0.910341,
		0.881522,
		0.848135,
		0.810295,
		0.76822,
		0.722224,
		0.672708,
		0.620158,
		0.56514,
		0.508291,
		0.45032,
		0.391998,
		0.33416,
		0.2777,
		0.223565,
		0.172758,
		0.126332,
		0.085389,
		0.051077,
		0.024591,
		0.00717,
		9.2e-005,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		0.000807,
		0.010445,
		0.029959,
		0.058111,
		0.093699,
		0.135567,
		0.182603,
		0.233741,
		0.28796,
		0.344293,
		0.401822,
		0.459685,
		0.517077,
		0.573252,
		0.62753,
		0.679298,
		0.728012,
		0.77321,
		0.814507,
		0.851607,
		0.88431,
		0.91251,
		0.936212,
		0.955526,
		0.970681,
		0.982017,
		0.989983,
		0.995126,
		0.998065,
		0.999454,
		0.999923,
		0.999999,
		0.999999,
		0.999923,
		0.999454,
		0.998065,
		0.995126,
		0.989983,
		0.982017,
		0.970681,
		0.955526,
		0.936212,
		0.91251,
		0.88431,
		0.851607,
		0.814507,
		0.77321,
		0.728012,
		0.679298,
		0.62753,
		0.573252,
		0.517077,
		0.459685,
		0.401822,
		0.344293,
		0.28796,
		0.233741,
		0.182603,
		0.135567,
		0.093699,
		0.058111,
		0.029959,
		0.010445,
		0.000807,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.000361,
		0.008674,
		0.027023,
		0.054161,
		0.088884,
		0.130029,
		0.176478,
		0.227159,
		0.281048,
		0.337168,
		0.394596,
		0.452462,
		0.509955,
		0.566321,
		0.620871,
		0.672983,
		0.722107,
		0.767766,
		0.809569,
		0.847207,
		0.880467,
		0.909232,
		0.933492,
		0.953345,
		0.969004,
		0.980795,
		0.989154,
		0.994617,
		0.997796,
		0.999343,
		0.999896,
		0.999998,
		1,
		0.999945,
		0.999551,
		0.99831,
		0.995601,
		0.990768,
		0.983187,
		0.972299,
		0.957642,
		0.938862,
		0.915718,
		0.888082,
		0.855939,
		0.819379,
		0.778592,
		0.733863,
		0.685564,
		0.63415,
		0.580154,
		0.52418,
		0.466901,
		0.409054,
		0.351437,
		0.294906,
		0.24037,
		0.188793,
		0.141187,
		0.098612,
		0.062177,
		0.033031,
		0.01237,
		0.001428,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.003233,
		0.016928,
		0.040162,
		0.071682,
		0.110275,
		0.154777,
		0.204067,
		0.257074,
		0.312776,
		0.370205,
		0.428447,
		0.486644,
		0.544,
		0.599783,
		0.653328,
		0.704043,
		0.751412,
		0.794999,
		0.834457,
		0.869533,
		0.900072,
		0.926026,
		0.947458,
		0.964547,
		0.977591,
		0.987004,
		0.993303,
		0.997097,
		0.999046,
		0.999813,
		0.99999,
		1,
		0.999974,
		0.999698,
		0.998695,
		0.996348,
		0.991988,
		0.984964,
		0.974689,
		0.960668,
		0.942516,
		0.919966,
		0.892867,
		0.861185,
		0.824995,
		0.784477,
		0.739908,
		0.691659,
		0.640185,
		0.586022,
		0.529781,
		0.472143,
		0.413858,
		0.355737,
		0.298652,
		0.24353,
		0.191355,
		0.14316,
		0.100028,
		0.063092,
		0.033529,
		0.01256,
		0.001451,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.000372,
		0.008946,
		0.027851,
		0.055783,
		0.09148,
		0.133727,
		0.181354,
		0.233241,
		0.288319,
		0.345573,
		0.40404,
		0.46282,
		0.52107,
		0.578013,
		0.63294,
		0.685216,
		0.73428,
		0.779657,
		0.820956,
		0.857883,
		0.890244,
		0.91795,
		0.941027,
		0.959621,
		0.973997,
		0.984543,
		0.991758,
		0.99624,
		0.998655,
		0.999688,
		0.999973,
		1,
		0.99999,
		0.999807,
		0.999017,
		0.997011,
		0.993109,
		0.986637,
		0.976975,
		0.963598,
		0.946088,
		0.924149,
		0.897607,
		0.866407,
		0.830607,
		0.790377,
		0.745987,
		0.697803,
		0.646281,
		0.59196,
		0.535457,
		0.477463,
		0.41874,
		0.360112,
		0.302467,
		0.246752,
		0.193969,
		0.145174,
		0.101474,
		0.064028,
		0.034038,
		0.012755,
		0.001474,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.003336,
		0.017457,
		0.041389,
		0.073818,
		0.113477,
		0.159146,
		0.209654,
		0.263882,
		0.320766,
		0.379296,
		0.438523,
		0.497557,
		0.555576,
		0.611825,
		0.665622,
		0.716363,
		0.763528,
		0.806683,
		0.845492,
		0.879718,
		0.909232,
		0.934022,
		0.954194,
		0.969981,
		0.981742,
		0.989954,
		0.995203,
		0.998153,
		0.999508,
		0.99994,
		1,
		0.999997,
		0.999886,
		0.999281,
		0.997593,
		0.994133,
		0.988203,
		0.979155,
		0.966428,
		0.949573,
		0.928264,
		0.902298,
		0.8716,
		0.836212,
		0.796291,
		0.752097,
		0.703994,
		0.652437,
		0.597967,
		0.541209,
		0.482863,
		0.423701,
		0.364563,
		0.306352,
		0.250035,
		0.196635,
		0.14723,
		0.102952,
		0.064984,
		0.034559,
		0.012955,
		0.001497,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.000384,
		0.00923,
		0.028718,
		0.057479,
		0.094192,
		0.137583,
		0.18643,
		0.239562,
		0.295864,
		0.354276,
		0.413799,
		0.473495,
		0.532493,
		0.589991,
		0.645261,
		0.697651,
		0.746597,
		0.791619,
		0.832336,
		0.868468,
		0.899843,
		0.926409,
		0.948233,
		0.965511,
		0.97857,
		0.987863,
		0.993959,
		0.99752,
		0.999259,
		0.999882,
		0.999997,
		1,
		0.999938,
		0.999493,
		0.998097,
		0.99506,
		0.989663,
		0.981226,
		0.969156,
		0.952968,
		0.932304,
		0.906935,
		0.876762,
		0.841807,
		0.802214,
		0.758237,
		0.710231,
		0.658652,
		0.604044,
		0.547038,
		0.488343,
		0.428742,
		0.369091,
		0.310309,
		0.253383,
		0.199356,
		0.14933,
		0.104462,
		0.065963,
		0.035092,
		0.013159,
		0.001521,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.003444,
		0.018011,
		0.042672,
		0.076051,
		0.116819,
		0.163699,
		0.215467,
		0.270954,
		0.32905,
		0.388702,
		0.448922,
		0.50879,
		0.567455,
		0.624138,
		0.678141,
		0.728851,
		0.775744,
		0.81839,
		0.856465,
		0.889753,
		0.918157,
		0.941704,
		0.960553,
		0.974996,
		0.985457,
		0.992484,
		0.996733,
		0.998923,
		0.999788,
		0.999989,
		1,
		0.999971,
		0.999658,
		0.998528,
		0.995893,
		0.991015,
		0.983187,
		0.971777,
		0.956268,
		0.936266,
		0.911514,
		0.881887,
		0.847388,
		0.808146,
		0.764404,
		0.716513,
		0.664927,
		0.610192,
		0.552945,
		0.493904,
		0.433865,
		0.373698,
		0.31434,
		0.256796,
		0.202132,
		0.151474,
		0.106005,
		0.066963,
		0.035637,
		0.013368,
		0.001546,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.000397,
		0.009529,
		0.029626,
		0.059254,
		0.097025,
		0.141606,
		0.191718,
		0.246136,
		0.303696,
		0.363292,
		0.423884,
		0.484498,
		0.544233,
		0.60226,
		0.657831,
		0.710282,
		0.759042,
		0.803632,
		0.843682,
		0.878929,
		0.909232,
		0.934575,
		0.955074,
		0.970985,
		0.982702,
		0.990749,
		0.995768,
		0.998482,
		0.999647,
		0.99997,
		1,
		0.999989,
		0.999782,
		0.998889,
		0.996633,
		0.992261,
		0.985035,
		0.974289,
		0.959469,
		0.940146,
		0.916031,
		0.886971,
		0.852952,
		0.814083,
		0.770596,
		0.722839,
		0.67126,
		0.61641,
		0.55893,
		0.499548,
		0.439072,
		0.378386,
		0.318445,
		0.260276,
		0.204965,
		0.153664,
		0.107582,
		0.067986,
		0.036195,
		0.013582,
		0.001571,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.003558,
		0.018592,
		0.044016,
		0.078386,
		0.120309,
		0.168448,
		0.22152,
		0.278304,
		0.33764,
		0.398433,
		0.459655,
		0.520351,
		0.57964,
		0.636722,
		0.690882,
		0.741497,
		0.788042,
		0.830094,
		0.867344,
		0.899603,
		0.926809,
		0.949038,
		0.966504,
		0.979567,
		0.988724,
		0.994599,
		0.997915,
		0.999443,
		0.999932,
		1,
		0.999997,
		0.99987,
		0.999187,
		0.997284,
		0.9934,
		0.986769,
		0.976689,
		0.962567,
		0.943938,
		0.92048,
		0.892012,
		0.858494,
		0.820021,
		0.776812,
		0.729206,
		0.677651,
		0.622699,
		0.564995,
		0.505276,
		0.444363,
		0.383156,
		0.322628,
		0.263824,
		0.207857,
		0.155901,
		0.109195,
		0.069032,
		0.036766,
		0.013802,
		0.001597,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.00041,
		0.009842,
		0.030578,
		0.061111,
		0.099986,
		0.145806,
		0.197229,
		0.252975,
		0.311827,
		0.372632,
		0.434306,
		0.495839,
		0.556295,
		0.61482,
		0.670646,
		0.723099,
		0.771599,
		0.815675,
		0.854966,
		0.889236,
		0.918374,
		0.94241,
		0.961517,
		0.976017,
		0.986377,
		0.993199,
		0.997199,
		0.999161,
		0.999866,
		0.999997,
		1,
		0.99993,
		0.999426,
		0.99785,
		0.994434,
		0.988388,
		0.978974,
		0.965558,
		0.947639,
		0.924857,
		0.897002,
		0.864011,
		0.825958,
		0.783049,
		0.735615,
		0.6841,
		0.629057,
		0.571138,
		0.511089,
		0.449741,
		0.38801,
		0.326889,
		0.267443,
		0.210809,
		0.158187,
		0.110843,
		0.070103,
		0.037351,
		0.014026,
		0.001624,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.003677,
		0.019201,
		0.045425,
		0.08083,
		0.123957,
		0.173402,
		0.227824,
		0.285944,
		0.346552,
		0.408505,
		0.470734,
		0.532247,
		0.592135,
		0.649575,
		0.703836,
		0.754286,
		0.800401,
		0.841768,
		0.878098,
		0.909232,
		0.935153,
		0.955987,
		0.972017,
		0.983674,
		0.991538,
		0.99631,
		0.99878,
		0.99976,
		0.999988,
		1,
		0.999967,
		0.999612,
		0.998334,
		0.995364,
		0.989891,
		0.98114,
		0.968439,
		0.951243,
		0.929157,
		0.901939,
		0.869498,
		0.831889,
		0.789303,
		0.742062,
		0.690605,
		0.635486,
		0.577362,
		0.516987,
		0.455206,
		0.39295,
		0.331231,
		0.271135,
		0.213823,
		0.160523,
		0.112529,
		0.071199,
		0.03795,
		0.014256,
		0.001651,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.000424,
		0.010171,
		0.031577,
		0.063057,
		0.103084,
		0.150192,
		0.202976,
		0.260094,
		0.320274,
		0.382312,
		0.44508,
		0.507526,
		0.568684,
		0.627672,
		0.683702,
		0.736089,
		0.78425,
		0.827721,
		0.866157,
		0.89935,
		0.927229,
		0.949876,
		0.967529,
		0.980581,
		0.989583,
		0.995219,
		0.99828,
		0.999599,
		0.999966,
		1,
		0.999987,
		0.999752,
		0.998741,
		0.996194,
		0.991278,
		0.983187,
		0.971204,
		0.954745,
		0.933375,
		0.906817,
		0.874952,
		0.837812,
		0.795573,
		0.748546,
		0.697165,
		0.641985,
		0.583667,
		0.522972,
		0.460761,
		0.397978,
		0.335654,
		0.2749,
		0.2169,
		0.16291,
		0.114253,
		0.07232,
		0.038563,
		0.014492,
		0.001679,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.003803,
		0.019841,
		0.046901,
		0.083389,
		0.127771,
		0.178574,
		0.234393,
		0.29389,
		0.355799,
		0.418929,
		0.482168,
		0.544486,
		0.604944,
		0.662695,
		0.716994,
		0.767202,
		0.812798,
		0.853382,
		0.88869,
		0.918602,
		0.943146,
		0.962515,
		0.977061,
		0.987301,
		0.993899,
		0.997637,
		0.999367,
		0.999923,
		1,
		0.999996,
		0.999852,
		0.999076,
		0.996925,
		0.992549,
		0.985111,
		0.973851,
		0.958141,
		0.937506,
		0.911631,
		0.880367,
		0.843722,
		0.801855,
		0.755064,
		0.70378,
		0.648554,
		0.590052,
		0.529046,
		0.466406,
		0.403095,
		0.340162,
		0.278741,
		0.220042,
		0.165349,
		0.116017,
		0.073468,
		0.039191,
		0.014734,
		0.001708,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000439,
		0.010517,
		0.032625,
		0.065096,
		0.106327,
		0.154777,
		0.208972,
		0.267506,
		0.32905,
		0.392345,
		0.456215,
		0.51957,
		0.581406,
		0.640815,
		0.696992,
		0.749239,
		0.796974,
		0.839741,
		0.877219,
		0.909232,
		0.935757,
		0.956935,
		0.973075,
		0.984657,
		0.992316,
		0.996826,
		0.999046,
		0.999847,
		0.999996,
		1,
		0.99992,
		0.999346,
		0.997561,
		0.993706,
		0.986911,
		0.976376,
		0.961427,
		0.941544,
		0.916375,
		0.885739,
		0.849616,
		0.808146,
		0.761615,
		0.710447,
		0.655192,
		0.596519,
		0.535208,
		0.472143,
		0.408303,
		0.344756,
		0.282661,
		0.223252,
		0.167843,
		0.117821,
		0.074644,
		0.039835,
		0.014982,
		0.001737,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.003934,
		0.020513,
		0.048451,
		0.086071,
		0.131761,
		0.183976,
		0.241242,
		0.302157,
		0.365398,
		0.429721,
		0.493969,
		0.557076,
		0.618068,
		0.676077,
		0.730344,
		0.780225,
		0.825205,
		0.864902,
		0.899083,
		0.927669,
		0.950749,
		0.968585,
		0.981612,
		0.990437,
		0.995817,
		0.998613,
		0.999726,
		0.999986,
		1,
		0.999962,
		0.999557,
		0.998106,
		0.994749,
		0.988586,
		0.978775,
		0.964597,
		0.945484,
		0.921045,
		0.891062,
		0.855488,
		0.814442,
		0.768196,
		0.717165,
		0.661898,
		0.603067,
		0.541461,
		0.477974,
		0.413603,
		0.349439,
		0.28666,
		0.22653,
		0.170393,
		0.119668,
		0.075848,
		0.040494,
		0.015236,
		0.001767,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.000454,
		0.01088,
		0.033726,
		0.067236,
		0.109724,
		0.159571,
		0.21523,
		0.275228,
		0.338171,
		0.402746,
		0.467727,
		0.531979,
		0.594464,
		0.654248,
		0.710506,
		0.762531,
		0.809744,
		0.851704,
		0.888115,
		0.918841,
		0.943915,
		0.963547,
		0.978128,
		0.988228,
		0.99458,
		0.998044,
		0.999542,
		0.999961,
		1,
		0.999985,
		0.999716,
		0.998566,
		0.995681,
		0.990135,
		0.981047,
		0.967646,
		0.949321,
		0.925634,
		0.896331,
		0.861335,
		0.820741,
		0.774805,
		0.723934,
		0.668673,
		0.609698,
		0.547804,
		0.483901,
		0.418999,
		0.354211,
		0.290741,
		0.229879,
		0.173,
		0.121558,
		0.077081,
		0.04117,
		0.015497,
		0.001798,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.004073,
		0.02122,
		0.050078,
		0.088884,
		0.135939,
		0.189623,
		0.248386,
		0.310761,
		0.375363,
		0.440894,
		0.506149,
		0.570021,
		0.631507,
		0.689715,
		0.743871,
		0.793332,
		0.837591,
		0.87629,
		0.909232,
		0.93639,
		0.95792,
		0.974162,
		0.985648,
		0.993081,
		0.997312,
		0.999278,
		0.999911,
		0.999999,
		0.999996,
		0.999831,
		0.998946,
		0.996504,
		0.991558,
		0.983187,
		0.970571,
		0.95305,
		0.930138,
		0.901541,
		0.867151,
		0.827036,
		0.781437,
		0.73075,
		0.675515,
		0.61641,
		0.55424,
		0.489923,
		0.424491,
		0.359076,
		0.294906,
		0.2333,
		0.175666,
		0.123492,
		0.078344,
		0.041863,
		0.015765,
		0.00183,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.000471,
		0.011263,
		0.034883,
		0.069482,
		0.113285,
		0.164588,
		0.221767,
		0.283276,
		0.347655,
		0.413531,
		0.479627,
		0.544762,
		0.607863,
		0.667967,
		0.724231,
		0.775945,
		0.822533,
		0.863572,
		0.898801,
		0.928131,
		0.951659,
		0.969673,
		0.982658,
		0.991285,
		0.996388,
		0.99891,
		0.999825,
		0.999996,
		0.999999,
		0.999908,
		0.999253,
		0.997222,
		0.992856,
		0.985193,
		0.973368,
		0.956664,
		0.934549,
		0.906686,
		0.872931,
		0.833326,
		0.788092,
		0.737611,
		0.682423,
		0.623205,
		0.560768,
		0.496045,
		0.430082,
		0.364035,
		0.299157,
		0.236797,
		0.178394,
		0.125472,
		0.079638,
		0.042574,
		0.016039,
		0.001863,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.004219,
		0.021964,
		0.051789,
		0.091835,
		0.140317,
		0.195527,
		0.255841,
		0.319719,
		0.385711,
		0.452462,
		0.518718,
		0.583329,
		0.645261,
		0.703599,
		0.757559,
		0.806498,
		0.849923,
		0.887506,
		0.919093,
		0.944719,
		0.964615,
		0.979216,
		0.989154,
		0.995238,
		0.998415,
		0.999686,
		0.999984,
		1,
		0.999956,
		0.999493,
		0.997839,
		0.994029,
		0.987065,
		0.976031,
		0.960158,
		0.938862,
		0.91176,
		0.87867,
		0.839605,
		0.794763,
		0.744517,
		0.689396,
		0.630081,
		0.567389,
		0.502265,
		0.435774,
		0.369091,
		0.303496,
		0.24037,
		0.181184,
		0.1275,
		0.080964,
		0.043303,
		0.016321,
		0.001896,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.000488,
		0.011667,
		0.036101,
		0.071843,
		0.11702,
		0.169842,
		0.228598,
		0.291667,
		0.357518,
		0.424716,
		0.491928,
		0.557927,
		0.621603,
		0.681965,
		0.738155,
		0.789457,
		0.835308,
		0.875306,
		0.909232,
		0.937054,
		0.958943,
		0.975277,
		0.986646,
		0.993829,
		0.997765,
		0.999475,
		0.999955,
		1,
		0.999983,
		0.999675,
		0.998361,
		0.995079,
		0.9888,
		0.978558,
		0.963527,
		0.943071,
		0.916757,
		0.884362,
		0.845868,
		0.80145,
		0.751464,
		0.696434,
		0.63704,
		0.574105,
		0.508587,
		0.441568,
		0.374245,
		0.307926,
		0.244022,
		0.184039,
		0.129577,
		0.082324,
		0.044051,
		0.01661,
		0.001931,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.004374,
		0.022747,
		0.053588,
		0.094935,
		0.144906,
		0.201706,
		0.263625,
		0.32905,
		0.39646,
		0.464441,
		0.531686,
		0.597004,
		0.659326,
		0.717717,
		0.771385,
		0.819691,
		0.862161,
		0.898503,
		0.928617,
		0.952607,
		0.970795,
		0.983719,
		0.992122,
		0.996928,
		0.999172,
		0.999898,
		0.999999,
		0.999995,
		0.999805,
		0.998793,
		0.996009,
		0.990397,
		0.980944,
		0.966766,
		0.94717,
		0.92167,
		0.890001,
		0.85211,
		0.808146,
		0.758449,
		0.703534,
		0.64408,
		0.580916,
		0.515011,
		0.447466,
		0.379501,
		0.31245,
		0.247755,
		0.186961,
		0.131705,
		0.083718,
		0.044818,
		0.016907,
		0.001966,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.000506,
		0.012092,
		0.037384,
		0.074325,
		0.120941,
		0.175346,
		0.23574,
		0.300419,
		0.367779,
		0.436317,
		0.504643,
		0.571482,
		0.635684,
		0.696234,
		0.752258,
		0.803041,
		0.848031,
		0.886861,
		0.919359,
		0.94556,
		0.965721,
		0.980326,
		0.990077,
		0.995871,
		0.99875,
		0.999798,
		0.999995,
		0.999999,
		0.999894,
		0.999142,
		0.996822,
		0.991857,
		0.983187,
		0.96987,
		0.951151,
		0.926493,
		0.895582,
		0.858326,
		0.814849,
		0.76547,
		0.710695,
		0.651201,
		0.587822,
		0.521539,
		0.453471,
		0.384859,
		0.317068,
		0.251572,
		0.189951,
		0.133884,
		0.085147,
		0.045605,
		0.017212,
		0.002002,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.004537,
		0.023574,
		0.055483,
		0.098193,
		0.149721,
		0.208175,
		0.271757,
		0.338772,
		0.407628,
		0.476846,
		0.545064,
		0.611048,
		0.673697,
		0.732056,
		0.785326,
		0.832877,
		0.874262,
		0.909232,
		0.937751,
		0.960006,
		0.97642,
		0.987648,
		0.994557,
		0.998182,
		0.999638,
		0.999981,
		1,
		0.99995,
		0.999417,
		0.997523,
		0.993181,
		0.985283,
		0.972832,
		0.95501,
		0.931219,
		0.901097,
		0.864511,
		0.821553,
		0.772523,
		0.717915,
		0.658402,
		0.594824,
		0.528173,
		0.459584,
		0.390324,
		0.321785,
		0.255475,
		0.193012,
		0.136118,
		0.086613,
		0.046414,
		0.017525,
		0.00204,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.000525,
		0.012541,
		0.038736,
		0.076937,
		0.125061,
		0.181117,
		0.243212,
		0.309553,
		0.378457,
		0.448352,
		0.517785,
		0.585432,
		0.650105,
		0.710763,
		0.76652,
		0.816662,
		0.860662,
		0.898188,
		0.929128,
		0.953597,
		0.971951,
		0.984791,
		0.992945,
		0.997435,
		0.999396,
		0.999948,
		1,
		0.99998,
		0.999625,
		0.998117,
		0.994368,
		0.987232,
		0.97565,
		0.958739,
		0.935842,
		0.90654,
		0.870658,
		0.828254,
		0.779606,
		0.725192,
		0.665684,
		0.601923,
		0.534913,
		0.465807,
		0.395897,
		0.326602,
		0.259467,
		0.196146,
		0.138407,
		0.088117,
		0.047243,
		0.017847,
		0.002078,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.004709,
		0.024446,
		0.057479,
		0.101621,
		0.154777,
		0.214952,
		0.280256,
		0.348906,
		0.419232,
		0.48969,
		0.55886,
		0.625463,
		0.688367,
		0.746597,
		0.799352,
		0.846017,
		0.886177,
		0.919639,
		0.94644,
		0.966867,
		0.981456,
		0.990993,
		0.996474,
		0.999046,
		0.999882,
		0.999999,
		0.999995,
		0.999775,
		0.99861,
		0.995423,
		0.989031,
		0.978317,
		0.962333,
		0.940353,
		0.911904,
		0.876762,
		0.834948,
		0.786713,
		0.732524,
		0.673044,
		0.609118,
		0.541761,
		0.472143,
		0.40158,
		0.331523,
		0.263549,
		0.199356,
		0.140754,
		0.08966,
		0.048096,
		0.018178,
		0.002117,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.000546,
		0.013016,
		0.040162,
		0.079688,
		0.129392,
		0.187172,
		0.251033,
		0.319089,
		0.389573,
		0.460836,
		0.531365,
		0.599783,
		0.664862,
		0.725537,
		0.780913,
		0.830285,
		0.873152,
		0.909232,
		0.938482,
		0.961112,
		0.977591,
		0.988652,
		0.99526,
		0.99856,
		0.999767,
		0.999994,
		0.999999,
		0.999877,
		0.999011,
		0.996348,
		0.990679,
		0.980831,
		0.965784,
		0.944746,
		0.917181,
		0.882814,
		0.841627,
		0.793842,
		0.739908,
		0.680482,
		0.61641,
		0.548718,
		0.478594,
		0.407377,
		0.336549,
		0.267725,
		0.202643,
		0.14316,
		0.091244,
		0.048971,
		0.018518,
		0.002158,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.004891,
		0.025367,
		0.059585,
		0.105229,
		0.160089,
		0.222058,
		0.289144,
		0.359472,
		0.431293,
		0.502989,
		0.573084,
		0.64025,
		0.703324,
		0.761317,
		0.813428,
		0.859064,
		0.897854,
		0.929666,
		0.95463,
		0.973143,
		0.985874,
		0.99375,
		0.997904,
		0.999581,
		0.999978,
		1,
		0.999942,
		0.999326,
		0.997147,
		0.992177,
		0.983187,
		0.969087,
		0.949014,
		0.922365,
		0.88881,
		0.848287,
		0.800988,
		0.747342,
		0.687996,
		0.623799,
		0.555785,
		0.48516,
		0.413289,
		0.341684,
		0.271997,
		0.20601,
		0.145627,
		0.09287,
		0.049871,
		0.018867,
		0.0022,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.000567,
		0.013517,
		0.041669,
		0.082589,
		0.133949,
		0.193529,
		0.259225,
		0.32905,
		0.401146,
		0.473788,
		0.545395,
		0.614537,
		0.679948,
		0.740538,
		0.795408,
		0.843868,
		0.885451,
		0.919935,
		0.947363,
		0.968053,
		0.982605,
		0.991898,
		0.997042,
		0.9993,
		0.999939,
		1,
		0.999977,
		0.999565,
		0.997826,
		0.993526,
		0.985382,
		0.972236,
		0.953149,
		0.927446,
		0.89474,
		0.854921,
		0.808146,
		0.754822,
		0.695586,
		0.631286,
		0.562964,
		0.491846,
		0.419319,
		0.34693,
		0.276369,
		0.20946,
		0.148158,
		0.094539,
		0.050795,
		0.019227,
		0.002243,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.005084,
		0.026342,
		0.061808,
		0.109032,
		0.165674,
		0.229511,
		0.298442,
		0.370494,
		0.44383,
		0.516759,
		0.587741,
		0.655405,
		0.718555,
		0.77619,
		0.827516,
		0.87197,
		0.909232,
		0.939252,
		0.962262,
		0.97879,
		0.989654,
		0.995933,
		0.998895,
		0.999863,
		0.999999,
		0.999994,
		0.999739,
		0.998392,
		0.994726,
		0.987414,
		0.975226,
		0.957143,
		0.932418,
		0.900598,
		0.861523,
		0.815311,
		0.762345,
		0.703249,
		0.638868,
		0.570256,
		0.498652,
		0.425471,
		0.352291,
		0.280842,
		0.212995,
		0.150754,
		0.096253,
		0.051745,
		0.019597,
		0.002287,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.00059,
		0.014049,
		0.043262,
		0.08565,
		0.138748,
		0.200209,
		0.267809,
		0.339457,
		0.413199,
		0.487226,
		0.559887,
		0.629697,
		0.695352,
		0.755743,
		0.809968,
		0.85736,
		0.897499,
		0.930235,
		0.95571,
		0.974369,
		0.986964,
		0.994532,
		0.998331,
		0.999729,
		0.999993,
		0.999999,
		0.999857,
		0.998853,
		0.995782,
		0.989282,
		0.978051,
		0.960989,
		0.937272,
		0.906376,
		0.868085,
		0.822478,
		0.769907,
		0.710982,
		0.646547,
		0.577661,
		0.50558,
		0.431746,
		0.357769,
		0.285421,
		0.216617,
		0.153419,
		0.098015,
		0.052722,
		0.019978,
		0.002333,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.005289,
		0.027373,
		0.064157,
		0.113042,
		0.171552,
		0.237335,
		0.308175,
		0.381994,
		0.456863,
		0.531011,
		0.602837,
		0.670922,
		0.73404,
		0.791183,
		0.84157,
		0.884676,
		0.920249,
		0.948332,
		0.969281,
		0.983772,
		0.992789,
		0.997571,
		0.999512,
		0.999974,
		1,
		0.999932,
		0.999216,
		0.996697,
		0.990983,
		0.980706,
		0.964681,
		0.941999,
		0.912065,
		0.874601,
		0.82964,
		0.777504,
		0.718785,
		0.654322,
		0.58518,
		0.512632,
		0.438147,
		0.363368,
		0.290107,
		0.220331,
		0.156153,
		0.099824,
		0.053727,
		0.02037,
		0.00238,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.000615,
		0.014613,
		0.044948,
		0.088884,
		0.143807,
		0.207232,
		0.276811,
		0.350337,
		0.425753,
		0.501165,
		0.574848,
		0.645261,
		0.711062,
		0.771123,
		0.824552,
		0.870708,
		0.909232,
		0.940063,
		0.963459,
		0.980016,
		0.99065,
		0.996571,
		0.999185,
		0.999929,
		1,
		0.999973,
		0.999493,
		0.997478,
		0.99252,
		0.983187,
		0.968209,
		0.946591,
		0.917657,
		0.881063,
		0.836792,
		0.785132,
		0.726654,
		0.662191,
		0.592813,
		0.519811,
		0.444677,
		0.369091,
		0.294906,
		0.224139,
		0.15896,
		0.101684,
		0.054761,
		0.020774,
		0.002428,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.005506,
		0.028466,
		0.066642,
		0.117276,
		0.177742,
		0.245554,
		0.318369,
		0.393997,
		0.470412,
		0.545761,
		0.618377,
		0.686792,
		0.749757,
		0.806258,
		0.855538,
		0.897121,
		0.930835,
		0.956839,
		0.975631,
		0.988059,
		0.995285,
		0.998713,
		0.999839,
		0.999999,
		0.999993,
		0.999695,
		0.99813,
		0.993892,
		0.98549,
		0.971569,
		0.951039,
		0.923141,
		0.887461,
		0.843926,
		0.792785,
		0.734587,
		0.670154,
		0.600563,
		0.527118,
		0.451339,
		0.37494,
		0.299819,
		0.228043,
		0.161842,
		0.103595,
		0.055825,
		0.021189,
		0.002478,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000641,
		0.015211,
		0.046734,
		0.092302,
		0.149143,
		0.214622,
		0.286256,
		0.361715,
		0.438833,
		0.515624,
		0.590287,
		0.661224,
		0.727057,
		0.786646,
		0.839109,
		0.883851,
		0.920582,
		0.949349,
		0.970552,
		0.984953,
		0.993658,
		0.998056,
		0.999682,
		0.999992,
		0.999999,
		0.999833,
		0.998662,
		0.995103,
		0.987613,
		0.974752,
		0.955335,
		0.928511,
		0.893788,
		0.851035,
		0.800458,
		0.742579,
		0.67821,
		0.608428,
		0.534555,
		0.458135,
		0.380919,
		0.30485,
		0.232048,
		0.164802,
		0.105561,
		0.05692,
		0.021618,
		0.00253,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.005737,
		0.029626,
		0.069274,
		0.121749,
		0.184267,
		0.254193,
		0.32905,
		0.406528,
		0.484498,
		0.561021,
		0.63436,
		0.703003,
		0.765677,
		0.82137,
		0.869359,
		0.909232,
		0.940917,
		0.964706,
		0.981267,
		0.991635,
		0.997169,
		0.999429,
		0.99997,
		1,
		0.99992,
		0.999083,
		0.996156,
		0.989554,
		0.977753,
		0.959469,
		0.933754,
		0.900035,
		0.858112,
		0.808146,
		0.750628,
		0.686356,
		0.61641,
		0.542124,
		0.465069,
		0.387032,
		0.310002,
		0.236155,
		0.167843,
		0.107582,
		0.058047,
		0.022059,
		0.002583,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.000668,
		0.015846,
		0.048628,
		0.09592,
		0.154777,
		0.222405,
		0.296172,
		0.373618,
		0.452462,
		0.53062,
		0.606211,
		0.67758,
		0.743316,
		0.80227,
		0.853585,
		0.896717,
		0.931471,
		0.95802,
		0.976929,
		0.989154,
		0.996003,
		0.999046,
		0.999917,
		1,
		0.999969,
		0.999406,
		0.997056,
		0.991312,
		0.980567,
		0.963433,
		0.938862,
		0.906191,
		0.865149,
		0.815842,
		0.758729,
		0.694592,
		0.624508,
		0.549827,
		0.472143,
		0.393283,
		0.315281,
		0.24037,
		0.170967,
		0.109662,
		0.059208,
		0.022514,
		0.002638,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.005983,
		0.030859,
		0.072063,
		0.12648,
		0.191151,
		0.26328,
		0.340247,
		0.419614,
		0.499142,
		0.5768,
		0.650784,
		0.719536,
		0.781763,
		0.836467,
		0.882967,
		0.920935,
		0.950418,
		0.971868,
		0.986146,
		0.994502,
		0.998492,
		0.999811,
		0.999999,
		0.999991,
		0.999641,
		0.997811,
		0.992887,
		0.983187,
		0.967218,
		0.943825,
		0.912247,
		0.872136,
		0.823539,
		0.766878,
		0.702914,
		0.632722,
		0.557664,
		0.47936,
		0.399673,
		0.320688,
		0.244695,
		0.174178,
		0.111802,
		0.060404,
		0.022984,
		0.002694,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000697,
		0.016522,
		0.050639,
		0.099754,
		0.160731,
		0.230606,
		0.306588,
		0.386075,
		0.466665,
		0.546167,
		0.622621,
		0.694316,
		0.759807,
		0.817948,
		0.867913,
		0.909232,
		0.94182,
		0.966004,
		0.982541,
		0.992605,
		0.997721,
		0.999626,
		0.999991,
		0.999999,
		0.999803,
		0.998429,
		0.994282,
		0.98561,
		0.970816,
		0.948631,
		0.918192,
		0.879064,
		0.83123,
		0.77507,
		0.71132,
		0.641052,
		0.565638,
		0.486723,
		0.406208,
		0.326228,
		0.249134,
		0.177478,
		0.114005,
		0.061637,
		0.023468,
		0.002753,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.006244,
		0.032169,
		0.075024,
		0.13149,
		0.198419,
		0.272844,
		0.35199,
		0.433283,
		0.514364,
		0.593109,
		0.667644,
		0.736367,
		0.797974,
		0.851487,
		0.896287,
		0.932145,
		0.959257,
		0.978262,
		0.990244,
		0.996681,
		0.999327,
		0.999964,
		1,
		0.999905,
		0.998921,
		0.995499,
		0.987832,
		0.974219,
		0.953272,
		0.924016,
		0.885923,
		0.838908,
		0.783299,
		0.719808,
		0.649498,
		0.573749,
		0.494234,
		0.412891,
		0.331906,
		0.25369,
		0.180871,
		0.116272,
		0.062907,
		0.023968,
		0.002813,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.000729,
		0.017243,
		0.052777,
		0.103819,
		0.16703,
		0.239256,
		0.317536,
		0.399117,
		0.481465,
		0.56228,
		0.63952,
		0.711414,
		0.776495,
		0.833622,
		0.882021,
		0.921311,
		0.951543,
		0.973229,
		0.987347,
		0.995313,
		0.998875,
		0.999901,
		1,
		0.999963,
		0.999299,
		0.996543,
		0.989852,
		0.97742,
		0.957735,
		0.929706,
		0.892703,
		0.846562,
		0.791559,
		0.728373,
		0.658057,
		0.582,
		0.501896,
		0.419726,
		0.337724,
		0.258369,
		0.18436,
		0.118608,
		0.064217,
		0.024483,
		0.002875,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.006524,
		0.033565,
		0.07817,
		0.136799,
		0.2061,
		0.282919,
		0.364313,
		0.447562,
		0.530184,
		0.609954,
		0.684929,
		0.753466,
		0.814257,
		0.866359,
		0.909232,
		0.942774,
		0.967356,
		0.983837,
		0.993551,
		0.998222,
		0.999776,
		0.999999,
		0.99999,
		0.999575,
		0.997422,
		0.991668,
		0.980411,
		0.962012,
		0.935251,
		0.899393,
		0.854185,
		0.799844,
		0.737013,
		0.66673,
		0.59039,
		0.509713,
		0.426716,
		0.343689,
		0.263174,
		0.18795,
		0.121013,
		0.065569,
		0.025016,
		0.00294,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000762,
		0.018011,
		0.055053,
		0.108136,
		0.1737,
		0.248386,
		0.32905,
		0.412775,
		0.496887,
		0.578973,
		0.656903,
		0.728851,
		0.793332,
		0.849227,
		0.895825,
		0.932861,
		0.960553,
		0.97963,
		0.991323,
		0.997312,
		0.999556,
		0.999989,
		0.999999,
		0.999766,
		0.998145,
		0.993282,
		0.983187,
		0.96609,
		0.94064,
		0.90598,
		0.861767,
		0.808146,
		0.745722,
		0.675515,
		0.598922,
		0.517687,
		0.433865,
		0.349803,
		0.268109,
		0.191643,
		0.123492,
		0.066963,
		0.025566,
		0.003007,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.006822,
		0.035053,
		0.081516,
		0.142431,
		0.214224,
		0.293538,
		0.37725,
		0.462481,
		0.546619,
		0.627338,
		0.702622,
		0.770794,
		0.830552,
		0.881004,
		0.921711,
		0.952728,
		0.974637,
		0.988552,
		0.996084,
		0.999202,
		0.999957,
		1,
		0.999887,
		0.998722,
		0.994696,
		0.985742,
		0.969962,
		0.94586,
		0.912454,
		0.869296,
		0.816458,
		0.754496,
		0.684409,
		0.607595,
		0.52582,
		0.441178,
		0.356071,
		0.273179,
		0.195443,
		0.126047,
		0.068402,
		0.026134,
		0.003075,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.000798,
		0.018832,
		0.057479,
		0.112726,
		0.18077,
		0.258031,
		0.341166,
		0.427082,
		0.512956,
		0.596255,
		0.67476,
		0.746597,
		0.810266,
		0.864686,
		0.909232,
		0.943785,
		0.968764,
		0.985151,
		0.994468,
		0.998665,
		0.999882,
		1,
		0.999956,
		0.999166,
		0.995914,
		0.988073,
		0.973615,
		0.950898,
		0.9188,
		0.876762,
		0.82477,
		0.763329,
		0.69341,
		0.61641,
		0.534114,
		0.448658,
		0.362499,
		0.278389,
		0.199356,
		0.128681,
		0.069887,
		0.026722,
		0.003147,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.007142,
		0.036642,
		0.085081,
		0.148414,
		0.222825,
		0.304739,
		0.390835,
		0.478069,
		0.563688,
		0.645261,
		0.720701,
		0.788304,
		0.846787,
		0.895329,
		0.933622,
		0.961912,
		0.98103,
		0.992386,
		0.997888,
		0.999732,
		0.999998,
		0.999988,
		0.999493,
		0.996943,
		0.990178,
		0.977043,
		0.955742,
		0.925006,
		0.884152,
		0.833075,
		0.772216,
		0.702516,
		0.625367,
		0.542574,
		0.45631,
		0.369091,
		0.283743,
		0.203384,
		0.131398,
		0.071421,
		0.027329,
		0.003221,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000836,
		0.01971,
		0.060068,
		0.117611,
		0.188271,
		0.268228,
		0.353922,
		0.442072,
		0.529695,
		0.614133,
		0.693079,
		0.76461,
		0.827229,
		0.879909,
		0.922139,
		0.953979,
		0.976091,
		0.989754,
		0.996806,
		0.99947,
		0.999987,
		0.999998,
		0.99972,
		0.997793,
		0.992055,
		0.980236,
		0.960379,
		0.931058,
		0.891454,
		0.84136,
		0.781149,
		0.711723,
		0.634466,
		0.5512,
		0.464137,
		0.375851,
		0.289247,
		0.207533,
		0.1342,
		0.073006,
		0.027957,
		0.003297,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.007484,
		0.038342,
		0.088884,
		0.154777,
		0.231941,
		0.316562,
		0.405109,
		0.494356,
		0.581406,
		0.663715,
		0.739133,
		0.805937,
		0.862877,
		0.909232,
		0.944857,
		0.97023,
		0.986479,
		0.995346,
		0.999046,
		0.999949,
		1,
		0.999864,
		0.998474,
		0.993706,
		0.983187,
		0.964797,
		0.936941,
		0.898654,
		0.849616,
		0.79012,
		0.721027,
		0.643707,
		0.559995,
		0.472143,
		0.382786,
		0.294906,
		0.211807,
		0.137092,
		0.074644,
		0.028608,
		0.003376,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.000878,
		0.020651,
		0.062836,
		0.122817,
		0.19624,
		0.279019,
		0.367361,
		0.457781,
		0.547127,
		0.632611,
		0.711836,
		0.782841,
		0.844144,
		0.894796,
		0.934434,
		0.963338,
		0.982461,
		0.993424,
		0.998403,
		0.999858,
		1,
		0.999946,
		0.999001,
		0.995135,
		0.98589,
		0.968983,
		0.942641,
		0.905738,
		0.85783,
		0.799123,
		0.730424,
		0.653088,
		0.568961,
		0.480333,
		0.3899,
		0.300724,
		0.21621,
		0.140077,
		0.076337,
		0.029281,
		0.003458,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.007852,
		0.040162,
		0.092945,
		0.161551,
		0.24161,
		0.32905,
		0.42011,
		0.511372,
		0.599783,
		0.68269,
		0.757879,
		0.823621,
		0.878726,
		0.922598,
		0.955299,
		0.977591,
		0.990947,
		0.997473,
		0.999678,
		0.999998,
		0.999985,
		0.999391,
		0.996348,
		0.98834,
		0.972926,
		0.948142,
		0.912691,
		0.86599,
		0.808146,
		0.739908,
		0.662608,
		0.5781,
		0.488711,
		0.397198,
		0.306709,
		0.22075,
		0.14316,
		0.078089,
		0.029978,
		0.003544,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.000922,
		0.021661,
		0.065798,
		0.128373,
		0.204713,
		0.290446,
		0.381525,
		0.474244,
		0.565272,
		0.651685,
		0.731,
		0.801227,
		0.860918,
		0.909232,
		0.945995,
		0.971756,
		0.987816,
		0.996177,
		0.999361,
		0.999984,
		0.999998,
		0.999662,
		0.997353,
		0.990536,
		0.976614,
		0.953429,
		0.919497,
		0.874082,
		0.817181,
		0.749475,
		0.672265,
		0.587413,
		0.497281,
		0.404687,
		0.312866,
		0.225429,
		0.146343,
		0.079901,
		0.0307,
		0.003632,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.008248,
		0.042115,
		0.097288,
		0.168771,
		0.251877,
		0.342249,
		0.43588,
		0.529145,
		0.618827,
		0.702164,
		0.776885,
		0.841272,
		0.89422,
		0.935301,
		0.964833,
		0.983919,
		0.994427,
		0.99885,
		0.999938,
		1,
		0.999836,
		0.998163,
		0.992476,
		0.980037,
		0.958486,
		0.926139,
		0.882091,
		0.826216,
		0.759116,
		0.682058,
		0.596902,
		0.506046,
		0.412372,
		0.3192,
		0.230255,
		0.149633,
		0.081776,
		0.031449,
		0.003724,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.00097,
		0.022747,
		0.068974,
		0.13431,
		0.213733,
		0.302559,
		0.39646,
		0.491497,
		0.584147,
		0.671346,
		0.750529,
		0.819691,
		0.877444,
		0.92309,
		0.956695,
		0.979138,
		0.992122,
		0.998073,
		0.999827,
		1,
		0.999935,
		0.998793,
		0.994162,
		0.983187,
		0.963299,
		0.9326,
		0.890001,
		0.835238,
		0.768825,
		0.691982,
		0.606567,
		0.515011,
		0.420258,
		0.325719,
		0.235233,
		0.153034,
		0.083718,
		0.032226,
		0.003819,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.008674,
		0.044213,
		0.101941,
		0.176478,
		0.26279,
		0.356209,
		0.452462,
		0.547701,
		0.638541,
		0.722107,
		0.796084,
		0.858788,
		0.909232,
		0.947206,
		0.973345,
		0.989154,
		0.99695,
		0.999608,
		0.999998,
		0.999982,
		0.999261,
		0.995601,
		0.986055,
		0.967852,
		0.938862,
		0.897796,
		0.844236,
		0.778592,
		0.702034,
		0.61641,
		0.52418,
		0.428351,
		0.332429,
		0.24037,
		0.156549,
		0.085729,
		0.033031,
		0.003918,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.001021,
		0.023917,
		0.072383,
		0.140662,
		0.223347,
		0.315408,
		0.412216,
		0.509577,
		0.603767,
		0.691577,
		0.770369,
		0.838141,
		0.893596,
		0.936229,
		0.966404,
		0.985399,
		0.995385,
		0.999223,
		0.999981,
		0.999997,
		0.999588,
		0.996799,
		0.988638,
		0.97213,
		0.944907,
		0.905457,
		0.853193,
		0.788409,
		0.71221,
		0.62643,
		0.533556,
		0.436658,
		0.339336,
		0.245671,
		0.160186,
		0.087813,
		0.033867,
		0.00402,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.009134,
		0.046472,
		0.106933,
		0.184715,
		0.2744,
		0.370984,
		0.469901,
		0.567067,
		0.65892,
		0.742476,
		0.815394,
		0.87605,
		0.92362,
		0.958172,
		0.980727,
		0.99327,
		0.9986,
		0.999924,
		1,
		0.999799,
		0.997769,
		0.990931,
		0.97612,
		0.950715,
		0.912966,
		0.862095,
		0.798264,
		0.722505,
		0.636628,
		0.543144,
		0.445185,
		0.346449,
		0.251144,
		0.163949,
		0.089974,
		0.034735,
		0.004127,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.001077,
		0.025179,
		0.076051,
		0.14747,
		0.233606,
		0.32905,
		0.428844,
		0.528519,
		0.624138,
		0.712348,
		0.790449,
		0.856465,
		0.909232,
		0.948498,
		0.974996,
		0.990485,
		0.997655,
		0.999788,
		1,
		0.99992,
		0.998528,
		0.992935,
		0.97981,
		0.956268,
		0.920303,
		0.870925,
		0.808146,
		0.732912,
		0.647002,
		0.552945,
		0.453937,
		0.353774,
		0.256796,
		0.167843,
		0.092214,
		0.035637,
		0.004238,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.009632,
		0.048908,
		0.112297,
		0.193529,
		0.286766,
		0.38663,
		0.488241,
		0.587261,
		0.679948,
		0.763216,
		0.834714,
		0.892918,
		0.937225,
		0.968053,
		0.986896,
		0.996286,
		0.999519,
		0.999997,
		0.999977,
		0.999094,
		0.994653,
		0.983187,
		0.961546,
		0.927446,
		0.879663,
		0.818041,
		0.743424,
		0.657551,
		0.562964,
		0.462921,
		0.361318,
		0.262634,
		0.171875,
		0.094539,
		0.036574,
		0.004354,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.001138,
		0.026543,
		0.080003,
		0.154777,
		0.244566,
		0.343545,
		0.446396,
		0.548355,
		0.645261,
		0.733619,
		0.81068,
		0.874529,
		0.924192,
		0.959736,
		0.982358,
		0.994377,
		0.999046,
		0.999976,
		0.999997,
		0.999493,
		0.996092,
		0.986243,
		0.966531,
		0.934374,
		0.888291,
		0.827935,
		0.754033,
		0.668272,
		0.573204,
		0.472143,
		0.369091,
		0.268666,
		0.176052,
		0.096952,
		0.037549,
		0.004475,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.010171,
		0.051539,
		0.118071,
		0.202976,
		0.299949,
		0.403207,
		0.507526,
		0.608301,
		0.701601,
		0.78425,
		0.85392,
		0.909232,
		0.949876,
		0.976711,
		0.991798,
		0.99828,
		0.999906,
		1,
		0.999752,
		0.997265,
		0.988971,
		0.971204,
		0.941064,
		0.896786,
		0.837812,
		0.764729,
		0.679164,
		0.583667,
		0.48161,
		0.377099,
		0.2749,
		0.18038,
		0.099458,
		0.038563,
		0.0046,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.001204,
		0.028021,
		0.084269,
		0.162631,
		0.25629,
		0.358959,
		0.464929,
		0.569117,
		0.667127,
		0.755329,
		0.830949,
		0.892179,
		0.938296,
		0.969784,
		0.988401,
		0.997117,
		0.999738,
		1,
		0.9999,
		0.998186,
		0.991369,
		0.975548,
		0.947493,
		0.905127,
		0.847654,
		0.775502,
		0.690221,
		0.594354,
		0.491327,
		0.385352,
		0.281345,
		0.184867,
		0.102062,
		0.039618,
		0.004731,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.010757,
		0.054388,
		0.124297,
		0.213114,
		0.314019,
		0.420779,
		0.527802,
		0.630193,
		0.723836,
		0.805484,
		0.872864,
		0.924811,
		0.961395,
		0.984024,
		0.995431,
		0.999403,
		0.999996,
		0.999972,
		0.998879,
		0.993437,
		0.979547,
		0.953636,
		0.913289,
		0.85744,
		0.786339,
		0.701439,
		0.605268,
		0.501302,
		0.393858,
		0.288011,
		0.18952,
		0.104768,
		0.040718,
		0.004868,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.001275,
		0.029626,
		0.088884,
		0.171089,
		0.268846,
		0.375363,
		0.484498,
		0.590828,
		0.689715,
		0.777401,
		0.851122,
		0.909232,
		0.951351,
		0.978488,
		0.993081,
		0.998815,
		0.99997,
		0.999996,
		0.999369,
		0.99518,
		0.983187,
		0.959469,
		0.921246,
		0.867151,
		0.797225,
		0.712812,
		0.61641,
		0.511539,
		0.402626,
		0.294906,
		0.194346,
		0.107582,
		0.041863,
		0.00501,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.011395,
		0.057479,
		0.131022,
		0.22401,
		0.32905,
		0.439414,
		0.549108,
		0.652936,
		0.746597,
		0.826794,
		0.89137,
		0.939451,
		0.971601,
		0.989904,
		0.997863,
		0.999882,
		1,
		0.999689,
		0.996609,
		0.986456,
		0.964969,
		0.928972,
		0.876762,
		0.808146,
		0.724333,
		0.62778,
		0.522046,
		0.411665,
		0.30204,
		0.199356,
		0.110511,
		0.043058,
		0.005159,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.001354,
		0.031373,
		0.093885,
		0.180211,
		0.282312,
		0.392831,
		0.505161,
		0.613507,
		0.712986,
		0.799732,
		0.871032,
		0.925483,
		0.963156,
		0.985719,
		0.996415,
		0.999671,
		1,
		0.999875,
		0.99774,
		0.989348,
		0.970112,
		0.936437,
		0.886248,
		0.819083,
		0.735993,
		0.639377,
		0.532828,
		0.420985,
		0.309424,
		0.204557,
		0.113559,
		0.044304,
		0.005315,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.012092,
		0.06084,
		0.138302,
		0.23574,
		0.345123,
		0.459182,
		0.571482,
		0.676513,
		0.769799,
		0.848031,
		0.909232,
		0.952932,
		0.980326,
		0.994317,
		0.999251,
		0.999995,
		0.999964,
		0.998595,
		0.991857,
		0.974876,
		0.943612,
		0.895582,
		0.830017,
		0.747781,
		0.651201,
		0.54389,
		0.430596,
		0.317068,
		0.209959,
		0.116734,
		0.045605,
		0.005478,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.00144,
		0.033279,
		0.099316,
		0.190067,
		0.29677,
		0.411446,
		0.526971,
		0.63716,
		0.736881,
		0.822187,
		0.89048,
		0.940699,
		0.973507,
		0.99139,
		0.998511,
		0.999962,
		0.999995,
		0.999205,
		0.993987,
		0.97924,
		0.950468,
		0.904734,
		0.840925,
		0.759685,
		0.663248,
		0.555238,
		0.440506,
		0.324985,
		0.215573,
		0.120042,
		0.046964,
		0.005649,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.012854,
		0.064503,
		0.146196,
		0.248386,
		0.362328,
		0.480153,
		0.594955,
		0.700893,
		0.793332,
		0.869007,
		0.926215,
		0.965026,
		0.987433,
		0.997312,
		0.99985,
		1,
		0.999606,
		0.995745,
		0.983187,
		0.956972,
		0.913673,
		0.851782,
		0.771692,
		0.675515,
		0.566876,
		0.450728,
		0.333186,
		0.221409,
		0.123492,
		0.048385,
		0.005827,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.001534,
		0.035363,
		0.105229,
		0.200736,
		0.312315,
		0.431293,
		0.549981,
		0.661781,
		0.761316,
		0.844598,
		0.909231,
		0.954629,
		0.982219,
		0.995487,
		0.99958,
		0.999998,
		0.999839,
		0.997146,
		0.9867,
		0.963094,
		0.922363,
		0.862561,
		0.783783,
		0.687995,
		0.578809,
		0.461269,
		0.341684,
		0.227479,
		0.127091,
		0.04987,
		0.006014,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.013691,
		0.068505,
		0.154774,
		0.262037,
		0.380753,
		0.502393,
		0.619534,
		0.726004,
		0.817037,
		0.889481,
		0.942036,
		0.975486,
		0.992826,
		0.999028,
		0.999976,
		0.999936,
		0.998198,
		0.989757,
		0.968789,
		0.930757,
		0.873218,
		0.795928,
		0.700674,
		0.591031,
		0.472135,
		0.350485,
		0.233791,
		0.130845,
		0.051425,
		0.006211,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.001637,
		0.037644,
		0.111669,
		0.212284,
		0.329012,
		0.452411,
		0.574175,
		0.687268,
		0.786085,
		0.86666,
		0.926912,
		0.966903,
		0.989042,
		0.997988,
		0.999837,
		0.99988,
		0.998871,
		0.99229,
		0.973965,
		0.938756,
		0.883665,
		0.808054,
		0.713495,
		0.603506,
		0.483303,
		0.359581,
		0.240343,
		0.134756,
		0.053048,
		0.006416,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.014606,
		0.072856,
		0.164044,
		0.276679,
		0.400334,
		0.525747,
		0.644953,
		0.751442,
		0.840368,
		0.908799,
		0.956001,
		0.983695,
		0.996096,
		0.99933,
		0.999523,
		0.999017,
		0.994116,
		0.97841,
		0.94614,
		0.893693,
		0.819979,
		0.726308,
		0.616116,
		0.494688,
		0.368915,
		0.247099,
		0.138805,
		0.054735,
		0.00663,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.001749,
		0.040101,
		0.118558,
		0.224539,
		0.346557,
		0.474322,
		0.598868,
		0.712711,
		0.810056,
		0.887051,
		0.942087,
		0.9761,
		0.992725,
		0.997934,
		0.998474,
		0.998268,
		0.994828,
		0.981687,
		0.952465,
		0.902878,
		0.831316,
		0.73878,
		0.628589,
		0.50608,
		0.378334,
		0.253963,
		0.142941,
		0.056465,
		0.006851,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.015567,
		0.077397,
		0.173627,
		0.291638,
		0.420042,
		0.548801,
		0.66939,
		0.774998,
		0.860781,
		0.924176,
		0.965238,
		0.986889,
		0.994761,
		0.995982,
		0.99593,
		0.993699,
		0.983029,
		0.956957,
		0.910471,
		0.841376,
		0.750302,
		0.640413,
		0.517072,
		0.38754,
		0.260738,
		0.147054,
		0.058196,
		0.007072,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.001862,
		0.042546,
		0.125331,
		0.236406,
		0.36322,
		0.494607,
		0.620939,
		0.734326,
		0.828867,
		0.900971,
		0.949718,
		0.977186,
		0.988479,
		0.99085,
		0.990906,
		0.989607,
		0.981267,
		0.95843,
		0.91532,
		0.849081,
		0.75991,
		0.650766,
		0.527004,
		0.396042,
		0.267095,
		0.15096,
		0.059855,
		0.007286,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.01645,
		0.081496,
		0.182095,
		0.304496,
		0.436364,
		0.566923,
		0.687158,
		0.790059,
		0.870942,
		0.927827,
		0.961826,
		0.977327,
		0.981438,
		0.981687,
		0.981036,
		0.974831,
		0.955287,
		0.915851,
		0.852944,
		0.766258,
		0.658487,
		0.534931,
		0.403127,
		0.272552,
		0.154387,
		0.061335,
		0.00748,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.001953,
		0.044463,
		0.130464,
		0.245006,
		0.374572,
		0.507231,
		0.632819,
		0.743188,
		0.832516,
		0.897699,
		0.938773,
		0.95923,
		0.965756,
		0.966443,
		0.966179,
		0.961852,
		0.945607,
		0.910151,
		0.851131,
		0.767663,
		0.662108,
		0.539646,
		0.407876,
		0.276483,
		0.156977,
		0.062493,
		0.007635,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.016988,
		0.083836,
		0.186511,
		0.310362,
		0.442328,
		0.57112,
		0.687459,
		0.784374,
		0.857596,
		0.906002,
		0.932032,
		0.941677,
		0.943198,
		0.943133,
		0.940363,
		0.92735,
		0.896157,
		0.841631,
		0.76225,
		0.659966,
		0.539764,
		0.409225,
		0.278157,
		0.158311,
		0.06316,
		0.007731,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.001983,
		0.04498,
		0.131419,
		0.245614,
		0.373469,
		0.502643,
		0.622781,
		0.7258,
		0.806267,
		0.861859,
		0.893836,
		0.907247,
		0.910115,
		0.910182,
		0.908605,
		0.898653,
		0.871933,
		0.822512,
		0.748179,
		0.650402,
		0.533882,
		0.406079,
		0.276811,
		0.157944,
		0.063155,
		0.007746,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.016812,
		0.08262,
		0.182939,
		0.302792,
		0.428931,
		0.550042,
		0.65702,
		0.743327,
		0.805445,
		0.843385,
		0.861048,
		0.865825,
		0.86612,
		0.865353,
		0.858176,
		0.836021,
		0.792248,
		0.723937,
		0.632006,
		0.520775,
		0.397462,
		0.271753,
		0.155471,
		0.062312,
		0.007659,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.00191,
		0.043137,
		0.125451,
		0.23323,
		0.352525,
		0.471247,
		0.579429,
		0.669547,
		0.736961,
		0.780424,
		0.80255,
		0.809757,
		0.810531,
		0.810233,
		0.805429,
		0.787765,
		0.750038,
		0.68864,
		0.603886,
		0.499619,
		0.382687,
		0.26248,
		0.150584,
		0.060501,
		0.007452,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.015609,
		0.076359,
		0.168202,
		0.276767,
		0.389445,
		0.495623,
		0.586982,
		0.657896,
		0.705931,
		0.732363,
		0.742376,
		0.743988,
		0.743918,
		0.740998,
		0.727561,
		0.696065,
		0.642277,
		0.565889,
		0.470186,
		0.361507,
		0.24878,
		0.143141,
		0.057659,
		0.007118,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.001708,
		0.038407,
		0.111127,
		0.205406,
		0.308422,
		0.409197,
		0.498862,
		0.571013,
		0.622176,
		0.652329,
		0.665273,
		0.668102,
		0.66817,
		0.666607,
		0.656947,
		0.631602,
		0.585835,
		0.518731,
		0.43295,
		0.334211,
		0.230803,
		0.133206,
		0.053801,
		0.006658,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.013323,
		0.064851,
		0.142042,
		0.232205,
		0.324316,
		0.40925,
		0.48009,
		0.532534,
		0.565401,
		0.581082,
		0.585423,
		0.585696,
		0.584989,
		0.578504,
		0.558952,
		0.52127,
		0.463986,
		0.389109,
		0.30164,
		0.209081,
		0.121059,
		0.049033,
		0.006083,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.001391,
		0.031123,
		0.089552,
		0.164471,
		0.24515,
		0.322525,
		0.389462,
		0.441107,
		0.475331,
		0.493212,
		0.499182,
		0.499836,
		0.499583,
		0.495581,
		0.481216,
		0.451312,
		0.403941,
		0.34047,
		0.265117,
		0.184485,
		0.107181,
		0.043541,
		0.005415,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.01025,
		0.04962,
		0.108001,
		0.175283,
		0.242785,
		0.303466,
		0.352213,
		0.386226,
		0.405462,
		0.412943,
		0.414176,
		0.414121,
		0.4119,
		0.401925,
		0.379157,
		0.341342,
		0.289251,
		0.226305,
		0.158129,
		0.092199,
		0.037571,
		0.004685,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.00101,
		0.022482,
		0.064291,
		0.11724,
		0.17332,
		0.22588,
		0.269848,
		0.302048,
		0.321588,
		0.330213,
		0.332146,
		0.332194,
		0.331119,
		0.32464,
		0.308099,
		0.279078,
		0.237836,
		0.187018,
		0.131251,
		0.076818,
		0.031406,
		0.003928,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.006942,
		0.033402,
		0.072196,
		0.116229,
		0.159489,
		0.197224,
		0.226175,
		0.244896,
		0.254094,
		0.256712,
		0.25688,
		0.256446,
		0.252561,
		0.241175,
		0.219874,
		0.188515,
		0.14903,
		0.105078,
		0.061746,
		0.025332,
		0.003178,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.000636,
		0.014062,
		0.039939,
		0.072255,
		0.105836,
		0.136466,
		0.161065,
		0.177929,
		0.18702,
		0.190146,
		0.190498,
		0.190361,
		0.188251,
		0.180892,
		0.166038,
		0.143273,
		0.113912,
		0.080714,
		0.047631,
		0.019612,
		0.002468,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.003985,
		0.019045,
		0.040846,
		0.065163,
		0.088475,
		0.108085,
		0.122288,
		0.130593,
		0.133926,
		0.134491,
		0.134466,
		0.133455,
		0.129044,
		0.119294,
		0.103642,
		0.082904,
		0.059053,
		0.035006,
		0.014469,
		0.001827,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.000329,
		0.007231,
		0.020382,
		0.036547,
		0.052975,
		0.067482,
		0.078559,
		0.085527,
		0.088707,
		0.089443,
		0.089462,
		0.08905,
		0.086641,
		0.080692,
		0.070615,
		0.056853,
		0.040725,
		0.024257,
		0.010067,
		0.001275,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.001801,
		0.008546,
		0.018168,
		0.028687,
		0.038483,
		0.046365,
		0.051666,
		0.054366,
		0.05516,
		0.055213,
		0.055078,
		0.053909,
		0.050596,
		0.04462,
		0.036175,
		0.026068,
		0.015607,
		0.006505,
		0.000827,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000126,
		0.002744,
		0.007669,
		0.013614,
		0.019499,
		0.024496,
		0.028075,
		0.03008,
		0.030794,
		0.030877,
		0.030845,
		0.030358,
		0.02872,
		0.025537,
		0.020858,
		0.015128,
		0.009106,
		0.003813,
		0.000487,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		0.000544,
		0.002558,
		0.005384,
		0.008402,
		0.011116,
		0.013182,
		0.014441,
		0.014966,
		0.015059,
		0.015054,
		0.014891,
		0.014202,
		0.012738,
		0.010488,
		0.007659,
		0.004637,
		0.001951,
		0.00025,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		2.7e-005,
		0.000588,
		0.001626,
		0.002854,
		0.004032,
		0.004985,
		0.005612,
		0.005911,
		0.005983,
		0.005985,
		0.005944,
		0.005716,
		0.005174,
		0.004297,
		0.003161,
		0.001926,
		0.000815,
		0.000105,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		6.8e-005,
		0.000316,
		0.000658,
		0.001013,
		0.001319,
		0.001535,
		0.001651,
		0.001686,
		0.001689,
		0.001682,
		0.001631,
		0.001491,
		0.001249,
		0.000927,
		0.000568,
		0.000242,
		3.1e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		2.6e-005,
		7.1e-005,
		0.000122,
		0.00017,
		0.000206,
		0.000228,
		0.000236,
		0.000237,
		0.000236,
		0.000231,
		0.000213,
		0.00018,
		0.000135,
		8.3e-005,
		3.6e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		3.6e-005,
		8.3e-005,
		0.000135,
		0.00018,
		0.000213,
		0.000231,
		0.000236,
		0.000237,
		0.000236,
		0.000228,
		0.000206,
		0.00017,
		0.000122,
		7.1e-005,
		2.6e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		3.1e-005,
		0.000242,
		0.000568,
		0.000927,
		0.001249,
		0.001491,
		0.001631,
		0.001682,
		0.001689,
		0.001686,
		0.001651,
		0.001535,
		0.001319,
		0.001013,
		0.000658,
		0.000316,
		6.8e-005,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		0.000105,
		0.000815,
		0.001926,
		0.003161,
		0.004297,
		0.005174,
		0.005716,
		0.005944,
		0.005985,
		0.005983,
		0.005911,
		0.005612,
		0.004985,
		0.004032,
		0.002854,
		0.001626,
		0.000588,
		2.7e-005,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.00025,
		0.001951,
		0.004637,
		0.007659,
		0.010488,
		0.012738,
		0.014202,
		0.014891,
		0.015054,
		0.015059,
		0.014966,
		0.014441,
		0.013182,
		0.011116,
		0.008402,
		0.005384,
		0.002558,
		0.000544,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.000487,
		0.003813,
		0.009106,
		0.015128,
		0.020858,
		0.025537,
		0.02872,
		0.030358,
		0.030845,
		0.030877,
		0.030794,
		0.03008,
		0.028075,
		0.024496,
		0.019499,
		0.013614,
		0.007669,
		0.002744,
		0.000126,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000827,
		0.006505,
		0.015607,
		0.026068,
		0.036175,
		0.04462,
		0.050596,
		0.053909,
		0.055078,
		0.055213,
		0.05516,
		0.054366,
		0.051666,
		0.046365,
		0.038483,
		0.028687,
		0.018168,
		0.008546,
		0.001801,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.001275,
		0.010067,
		0.024257,
		0.040725,
		0.056853,
		0.070615,
		0.080692,
		0.086641,
		0.08905,
		0.089462,
		0.089443,
		0.088707,
		0.085527,
		0.078559,
		0.067482,
		0.052975,
		0.036547,
		0.020382,
		0.007231,
		0.000329,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001827,
		0.014469,
		0.035006,
		0.059053,
		0.082904,
		0.103642,
		0.119294,
		0.129044,
		0.133455,
		0.134466,
		0.134491,
		0.133926,
		0.130593,
		0.122288,
		0.108085,
		0.088475,
		0.065163,
		0.040846,
		0.019045,
		0.003985,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.002468,
		0.019612,
		0.047631,
		0.080714,
		0.113912,
		0.143273,
		0.166038,
		0.180892,
		0.188251,
		0.190361,
		0.190498,
		0.190146,
		0.18702,
		0.177929,
		0.161065,
		0.136466,
		0.105836,
		0.072255,
		0.039939,
		0.014062,
		0.000636,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.003178,
		0.025332,
		0.061746,
		0.105078,
		0.14903,
		0.188515,
		0.219874,
		0.241175,
		0.252561,
		0.256446,
		0.25688,
		0.256712,
		0.254094,
		0.244896,
		0.226175,
		0.197224,
		0.159489,
		0.116229,
		0.072196,
		0.033402,
		0.006942,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.003928,
		0.031406,
		0.076818,
		0.131251,
		0.187018,
		0.237836,
		0.279078,
		0.308099,
		0.32464,
		0.331119,
		0.332194,
		0.332146,
		0.330213,
		0.321588,
		0.302048,
		0.269848,
		0.22588,
		0.17332,
		0.11724,
		0.064291,
		0.022482,
		0.00101,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.004685,
		0.037571,
		0.092199,
		0.158129,
		0.226305,
		0.289251,
		0.341342,
		0.379157,
		0.401925,
		0.4119,
		0.414121,
		0.414176,
		0.412943,
		0.405462,
		0.386226,
		0.352213,
		0.303466,
		0.242785,
		0.175283,
		0.108001,
		0.04962,
		0.01025,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.005415,
		0.043541,
		0.107181,
		0.184485,
		0.265117,
		0.34047,
		0.403941,
		0.451312,
		0.481216,
		0.495581,
		0.499583,
		0.499836,
		0.499182,
		0.493212,
		0.475331,
		0.441107,
		0.389462,
		0.322525,
		0.24515,
		0.164471,
		0.089552,
		0.031123,
		0.001391,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.006083,
		0.049033,
		0.121059,
		0.209081,
		0.30164,
		0.389109,
		0.463986,
		0.52127,
		0.558952,
		0.578504,
		0.584989,
		0.585696,
		0.585423,
		0.581082,
		0.565401,
		0.532534,
		0.48009,
		0.40925,
		0.324316,
		0.232205,
		0.142042,
		0.064851,
		0.013323,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.006658,
		0.053801,
		0.133206,
		0.230803,
		0.334211,
		0.43295,
		0.518731,
		0.585835,
		0.631602,
		0.656947,
		0.666607,
		0.66817,
		0.668102,
		0.665273,
		0.652329,
		0.622176,
		0.571013,
		0.498862,
		0.409197,
		0.308422,
		0.205406,
		0.111127,
		0.038407,
		0.001708,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.007118,
		0.057659,
		0.143141,
		0.24878,
		0.361507,
		0.470186,
		0.565889,
		0.642277,
		0.696065,
		0.727561,
		0.740998,
		0.743918,
		0.743988,
		0.742376,
		0.732363,
		0.705931,
		0.657896,
		0.586982,
		0.495623,
		0.389445,
		0.276767,
		0.168202,
		0.076359,
		0.015609,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.007452,
		0.060501,
		0.150584,
		0.26248,
		0.382687,
		0.499619,
		0.603886,
		0.68864,
		0.750038,
		0.787765,
		0.805429,
		0.810233,
		0.810531,
		0.809757,
		0.80255,
		0.780424,
		0.736961,
		0.669547,
		0.579429,
		0.471247,
		0.352525,
		0.23323,
		0.125451,
		0.043137,
		0.00191,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.007659,
		0.062312,
		0.155471,
		0.271753,
		0.397462,
		0.520775,
		0.632006,
		0.723937,
		0.792248,
		0.836021,
		0.858176,
		0.865353,
		0.86612,
		0.865825,
		0.861048,
		0.843385,
		0.805445,
		0.743327,
		0.65702,
		0.550042,
		0.428931,
		0.302792,
		0.182939,
		0.08262,
		0.016812,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.007746,
		0.063155,
		0.157944,
		0.276811,
		0.406079,
		0.533882,
		0.650402,
		0.748179,
		0.822512,
		0.871933,
		0.898653,
		0.908605,
		0.910182,
		0.910115,
		0.907247,
		0.893836,
		0.861859,
		0.806267,
		0.7258,
		0.622781,
		0.502643,
		0.373469,
		0.245614,
		0.131419,
		0.04498,
		0.001983,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.007731,
		0.06316,
		0.158311,
		0.278157,
		0.409225,
		0.539764,
		0.659966,
		0.76225,
		0.841631,
		0.896157,
		0.92735,
		0.940363,
		0.943133,
		0.943198,
		0.941677,
		0.932032,
		0.906002,
		0.857596,
		0.784374,
		0.687459,
		0.57112,
		0.442328,
		0.310362,
		0.186511,
		0.083836,
		0.016988,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.007635,
		0.062493,
		0.156977,
		0.276483,
		0.407876,
		0.539646,
		0.662108,
		0.767663,
		0.851131,
		0.910151,
		0.945607,
		0.961852,
		0.966179,
		0.966443,
		0.965756,
		0.95923,
		0.938773,
		0.897699,
		0.832516,
		0.743188,
		0.632819,
		0.507231,
		0.374572,
		0.245006,
		0.130464,
		0.044463,
		0.001953,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.00748,
		0.061335,
		0.154387,
		0.272552,
		0.403127,
		0.534931,
		0.658487,
		0.766258,
		0.852944,
		0.915851,
		0.955287,
		0.974831,
		0.981036,
		0.981687,
		0.981438,
		0.977327,
		0.961826,
		0.927827,
		0.870942,
		0.790059,
		0.687158,
		0.566923,
		0.436364,
		0.304496,
		0.182095,
		0.081496,
		0.01645,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.007286,
		0.059855,
		0.15096,
		0.267095,
		0.396042,
		0.527004,
		0.650766,
		0.75991,
		0.849081,
		0.91532,
		0.95843,
		0.981267,
		0.989607,
		0.990906,
		0.99085,
		0.988479,
		0.977186,
		0.949718,
		0.900971,
		0.828867,
		0.734326,
		0.620939,
		0.494607,
		0.36322,
		0.236406,
		0.125331,
		0.042546,
		0.001862,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.007072,
		0.058196,
		0.147054,
		0.260738,
		0.38754,
		0.517072,
		0.640413,
		0.750302,
		0.841376,
		0.910471,
		0.956957,
		0.983029,
		0.993699,
		0.99593,
		0.995982,
		0.994761,
		0.986889,
		0.965238,
		0.924176,
		0.860781,
		0.774998,
		0.66939,
		0.548801,
		0.420042,
		0.291638,
		0.173627,
		0.077397,
		0.015567,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.006851,
		0.056465,
		0.142941,
		0.253963,
		0.378334,
		0.50608,
		0.628589,
		0.73878,
		0.831316,
		0.902878,
		0.952465,
		0.981687,
		0.994828,
		0.998268,
		0.998474,
		0.997934,
		0.992725,
		0.9761,
		0.942087,
		0.887051,
		0.810056,
		0.712711,
		0.598868,
		0.474322,
		0.346557,
		0.224539,
		0.118558,
		0.040101,
		0.001749,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.00663,
		0.054735,
		0.138805,
		0.247099,
		0.368915,
		0.494688,
		0.616116,
		0.726308,
		0.819979,
		0.893693,
		0.94614,
		0.97841,
		0.994116,
		0.999017,
		0.999523,
		0.99933,
		0.996096,
		0.983695,
		0.956001,
		0.908799,
		0.840368,
		0.751442,
		0.644953,
		0.525747,
		0.400334,
		0.276679,
		0.164044,
		0.072856,
		0.014606,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.006416,
		0.053048,
		0.134756,
		0.240343,
		0.359581,
		0.483303,
		0.603506,
		0.713495,
		0.808054,
		0.883665,
		0.938756,
		0.973965,
		0.99229,
		0.998871,
		0.99988,
		0.999837,
		0.997988,
		0.989042,
		0.966903,
		0.926912,
		0.86666,
		0.786085,
		0.687268,
		0.574175,
		0.452411,
		0.329012,
		0.212284,
		0.111669,
		0.037644,
		0.001637,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.006211,
		0.051425,
		0.130845,
		0.233791,
		0.350485,
		0.472135,
		0.591031,
		0.700674,
		0.795928,
		0.873218,
		0.930757,
		0.968789,
		0.989757,
		0.998198,
		0.999936,
		0.999976,
		0.999028,
		0.992826,
		0.975486,
		0.942036,
		0.889481,
		0.817037,
		0.726004,
		0.619534,
		0.502393,
		0.380753,
		0.262037,
		0.154774,
		0.068505,
		0.013691,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.006014,
		0.04987,
		0.127091,
		0.227479,
		0.341684,
		0.461269,
		0.578809,
		0.687995,
		0.783783,
		0.862561,
		0.922363,
		0.963094,
		0.9867,
		0.997146,
		0.999839,
		0.999998,
		0.99958,
		0.995487,
		0.982219,
		0.954629,
		0.909231,
		0.844598,
		0.761316,
		0.661781,
		0.549981,
		0.431293,
		0.312315,
		0.200736,
		0.105229,
		0.035363,
		0.001534,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.005827,
		0.048385,
		0.123492,
		0.221409,
		0.333186,
		0.450728,
		0.566876,
		0.675515,
		0.771692,
		0.851782,
		0.913673,
		0.956972,
		0.983187,
		0.995745,
		0.999606,
		1,
		0.99985,
		0.997312,
		0.987433,
		0.965026,
		0.926215,
		0.869007,
		0.793332,
		0.700893,
		0.594955,
		0.480153,
		0.362328,
		0.248386,
		0.146196,
		0.064503,
		0.012854,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.005649,
		0.046964,
		0.120042,
		0.215573,
		0.324985,
		0.440506,
		0.555238,
		0.663248,
		0.759685,
		0.840925,
		0.904734,
		0.950468,
		0.97924,
		0.993987,
		0.999205,
		0.999995,
		0.999962,
		0.998511,
		0.99139,
		0.973507,
		0.940699,
		0.89048,
		0.822187,
		0.736881,
		0.63716,
		0.526971,
		0.411446,
		0.29677,
		0.190067,
		0.099316,
		0.033279,
		0.00144,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005478,
		0.045605,
		0.116734,
		0.209959,
		0.317068,
		0.430596,
		0.54389,
		0.651201,
		0.747781,
		0.830017,
		0.895582,
		0.943612,
		0.974876,
		0.991857,
		0.998595,
		0.999964,
		0.999995,
		0.999251,
		0.994317,
		0.980326,
		0.952932,
		0.909232,
		0.848031,
		0.769799,
		0.676513,
		0.571482,
		0.459182,
		0.345123,
		0.23574,
		0.138302,
		0.06084,
		0.012092,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.005315,
		0.044304,
		0.113559,
		0.204557,
		0.309424,
		0.420985,
		0.532828,
		0.639377,
		0.735993,
		0.819083,
		0.886248,
		0.936437,
		0.970112,
		0.989348,
		0.99774,
		0.999875,
		1,
		0.999671,
		0.996415,
		0.985719,
		0.963156,
		0.925483,
		0.871032,
		0.799732,
		0.712986,
		0.613507,
		0.505161,
		0.392831,
		0.282312,
		0.180211,
		0.093885,
		0.031373,
		0.001354,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005159,
		0.043058,
		0.110511,
		0.199356,
		0.30204,
		0.411665,
		0.522046,
		0.62778,
		0.724333,
		0.808146,
		0.876762,
		0.928972,
		0.964969,
		0.986456,
		0.996609,
		0.999689,
		1,
		0.999882,
		0.997863,
		0.989904,
		0.971601,
		0.939451,
		0.89137,
		0.826794,
		0.746597,
		0.652936,
		0.549108,
		0.439414,
		0.32905,
		0.22401,
		0.131022,
		0.057479,
		0.011395,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.00501,
		0.041863,
		0.107582,
		0.194346,
		0.294906,
		0.402626,
		0.511539,
		0.61641,
		0.712812,
		0.797225,
		0.867151,
		0.921246,
		0.959469,
		0.983187,
		0.99518,
		0.999369,
		0.999996,
		0.99997,
		0.998815,
		0.993081,
		0.978488,
		0.951351,
		0.909232,
		0.851122,
		0.777401,
		0.689715,
		0.590828,
		0.484498,
		0.375363,
		0.268846,
		0.171089,
		0.088884,
		0.029626,
		0.001275,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.004868,
		0.040718,
		0.104768,
		0.18952,
		0.288011,
		0.393858,
		0.501302,
		0.605268,
		0.701439,
		0.786339,
		0.85744,
		0.913289,
		0.953636,
		0.979547,
		0.993437,
		0.998879,
		0.999972,
		0.999996,
		0.999403,
		0.995431,
		0.984024,
		0.961395,
		0.924811,
		0.872864,
		0.805484,
		0.723836,
		0.630193,
		0.527802,
		0.420779,
		0.314019,
		0.213114,
		0.124297,
		0.054388,
		0.010757,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.004731,
		0.039618,
		0.102062,
		0.184867,
		0.281345,
		0.385352,
		0.491327,
		0.594354,
		0.690221,
		0.775502,
		0.847654,
		0.905127,
		0.947493,
		0.975548,
		0.991369,
		0.998186,
		0.9999,
		1,
		0.999738,
		0.997117,
		0.988401,
		0.969784,
		0.938296,
		0.892179,
		0.830949,
		0.755329,
		0.667127,
		0.569117,
		0.464929,
		0.358959,
		0.25629,
		0.162631,
		0.084269,
		0.028021,
		0.001204,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.0046,
		0.038563,
		0.099458,
		0.18038,
		0.2749,
		0.377099,
		0.48161,
		0.583667,
		0.679164,
		0.764729,
		0.837812,
		0.896786,
		0.941064,
		0.971204,
		0.988971,
		0.997265,
		0.999752,
		1,
		0.999906,
		0.99828,
		0.991798,
		0.976711,
		0.949876,
		0.909232,
		0.85392,
		0.78425,
		0.701601,
		0.608301,
		0.507526,
		0.403207,
		0.299949,
		0.202976,
		0.118071,
		0.051539,
		0.010171,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.004475,
		0.037549,
		0.096952,
		0.176052,
		0.268666,
		0.369091,
		0.472143,
		0.573204,
		0.668272,
		0.754033,
		0.827935,
		0.888291,
		0.934374,
		0.966531,
		0.986243,
		0.996092,
		0.999493,
		0.999997,
		0.999976,
		0.999046,
		0.994377,
		0.982358,
		0.959736,
		0.924192,
		0.874529,
		0.81068,
		0.733619,
		0.645261,
		0.548355,
		0.446396,
		0.343545,
		0.244566,
		0.154777,
		0.080003,
		0.026543,
		0.001138,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.004354,
		0.036574,
		0.094539,
		0.171875,
		0.262634,
		0.361318,
		0.462921,
		0.562964,
		0.657551,
		0.743424,
		0.818041,
		0.879663,
		0.927446,
		0.961546,
		0.983187,
		0.994653,
		0.999094,
		0.999977,
		0.999997,
		0.999519,
		0.996286,
		0.986896,
		0.968053,
		0.937225,
		0.892918,
		0.834714,
		0.763216,
		0.679948,
		0.587261,
		0.488241,
		0.38663,
		0.286766,
		0.193529,
		0.112297,
		0.048908,
		0.009632,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.004238,
		0.035637,
		0.092214,
		0.167843,
		0.256796,
		0.353774,
		0.453937,
		0.552945,
		0.647002,
		0.732912,
		0.808146,
		0.870925,
		0.920303,
		0.956268,
		0.97981,
		0.992935,
		0.998528,
		0.99992,
		1,
		0.999788,
		0.997655,
		0.990485,
		0.974996,
		0.948498,
		0.909232,
		0.856465,
		0.790449,
		0.712348,
		0.624138,
		0.528519,
		0.428844,
		0.32905,
		0.233606,
		0.14747,
		0.076051,
		0.025179,
		0.001077,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.004127,
		0.034735,
		0.089974,
		0.163949,
		0.251144,
		0.346449,
		0.445185,
		0.543144,
		0.636628,
		0.722505,
		0.798264,
		0.862095,
		0.912966,
		0.950715,
		0.97612,
		0.990931,
		0.997769,
		0.999799,
		1,
		0.999924,
		0.9986,
		0.99327,
		0.980727,
		0.958172,
		0.92362,
		0.87605,
		0.815394,
		0.742476,
		0.65892,
		0.567067,
		0.469901,
		0.370984,
		0.2744,
		0.184715,
		0.106933,
		0.046472,
		0.009134,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.00402,
		0.033867,
		0.087813,
		0.160186,
		0.245671,
		0.339336,
		0.436658,
		0.533556,
		0.62643,
		0.71221,
		0.788409,
		0.853193,
		0.905457,
		0.944907,
		0.97213,
		0.988638,
		0.996799,
		0.999588,
		0.999997,
		0.999981,
		0.999223,
		0.995385,
		0.985399,
		0.966404,
		0.936229,
		0.893596,
		0.838141,
		0.770369,
		0.691577,
		0.603767,
		0.509577,
		0.412216,
		0.315408,
		0.223347,
		0.140662,
		0.072383,
		0.023917,
		0.001021,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.003918,
		0.033031,
		0.085729,
		0.156549,
		0.24037,
		0.332429,
		0.428351,
		0.52418,
		0.61641,
		0.702034,
		0.778592,
		0.844236,
		0.897796,
		0.938862,
		0.967852,
		0.986055,
		0.995601,
		0.999261,
		0.999982,
		0.999998,
		0.999608,
		0.99695,
		0.989154,
		0.973345,
		0.947206,
		0.909232,
		0.858788,
		0.796084,
		0.722107,
		0.638541,
		0.547701,
		0.452462,
		0.356209,
		0.26279,
		0.176478,
		0.101941,
		0.044213,
		0.008674,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.003819,
		0.032226,
		0.083718,
		0.153034,
		0.235233,
		0.325719,
		0.420258,
		0.515011,
		0.606567,
		0.691982,
		0.768825,
		0.835238,
		0.890001,
		0.9326,
		0.963299,
		0.983187,
		0.994162,
		0.998793,
		0.999935,
		1,
		0.999827,
		0.998073,
		0.992122,
		0.979138,
		0.956695,
		0.92309,
		0.877444,
		0.819691,
		0.750529,
		0.671346,
		0.584147,
		0.491497,
		0.39646,
		0.302559,
		0.213733,
		0.13431,
		0.068974,
		0.022747,
		0.00097,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.003724,
		0.031449,
		0.081776,
		0.149633,
		0.230255,
		0.3192,
		0.412372,
		0.506046,
		0.596902,
		0.682058,
		0.759116,
		0.826216,
		0.882091,
		0.926139,
		0.958486,
		0.980037,
		0.992476,
		0.998163,
		0.999836,
		1,
		0.999938,
		0.99885,
		0.994427,
		0.983919,
		0.964833,
		0.935301,
		0.89422,
		0.841272,
		0.776885,
		0.702164,
		0.618827,
		0.529145,
		0.43588,
		0.342249,
		0.251877,
		0.168771,
		0.097288,
		0.042115,
		0.008248,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.003632,
		0.0307,
		0.079901,
		0.146343,
		0.225429,
		0.312866,
		0.404687,
		0.497281,
		0.587413,
		0.672265,
		0.749475,
		0.817181,
		0.874082,
		0.919497,
		0.953429,
		0.976614,
		0.990536,
		0.997353,
		0.999662,
		0.999998,
		0.999984,
		0.999361,
		0.996177,
		0.987816,
		0.971756,
		0.945995,
		0.909232,
		0.860918,
		0.801227,
		0.731,
		0.651685,
		0.565272,
		0.474244,
		0.381525,
		0.290446,
		0.204713,
		0.128373,
		0.065798,
		0.021661,
		0.000922,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.003544,
		0.029978,
		0.078089,
		0.14316,
		0.22075,
		0.306709,
		0.397198,
		0.488711,
		0.5781,
		0.662608,
		0.739908,
		0.808146,
		0.86599,
		0.912691,
		0.948142,
		0.972926,
		0.98834,
		0.996348,
		0.999391,
		0.999985,
		0.999998,
		0.999678,
		0.997473,
		0.990947,
		0.977591,
		0.955299,
		0.922598,
		0.878726,
		0.823621,
		0.757879,
		0.68269,
		0.599783,
		0.511372,
		0.42011,
		0.32905,
		0.24161,
		0.161551,
		0.092945,
		0.040162,
		0.007852,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.003458,
		0.029281,
		0.076337,
		0.140077,
		0.21621,
		0.300724,
		0.3899,
		0.480333,
		0.568961,
		0.653088,
		0.730424,
		0.799123,
		0.85783,
		0.905738,
		0.942641,
		0.968983,
		0.98589,
		0.995135,
		0.999001,
		0.999946,
		1,
		0.999858,
		0.998403,
		0.993424,
		0.982461,
		0.963338,
		0.934434,
		0.894796,
		0.844144,
		0.782841,
		0.711836,
		0.632611,
		0.547127,
		0.457781,
		0.367361,
		0.279019,
		0.19624,
		0.122817,
		0.062836,
		0.020651,
		0.000878,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.003376,
		0.028608,
		0.074644,
		0.137092,
		0.211807,
		0.294906,
		0.382786,
		0.472143,
		0.559995,
		0.643707,
		0.721027,
		0.79012,
		0.849616,
		0.898654,
		0.936941,
		0.964797,
		0.983187,
		0.993706,
		0.998474,
		0.999864,
		1,
		0.999949,
		0.999046,
		0.995346,
		0.986479,
		0.97023,
		0.944857,
		0.909232,
		0.862877,
		0.805937,
		0.739133,
		0.663715,
		0.581406,
		0.494356,
		0.405109,
		0.316562,
		0.231941,
		0.154777,
		0.088884,
		0.038342,
		0.007484,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.003297,
		0.027957,
		0.073006,
		0.1342,
		0.207533,
		0.289247,
		0.375851,
		0.464137,
		0.5512,
		0.634466,
		0.711723,
		0.781149,
		0.84136,
		0.891454,
		0.931058,
		0.960379,
		0.980236,
		0.992055,
		0.997793,
		0.99972,
		0.999998,
		0.999987,
		0.99947,
		0.996806,
		0.989754,
		0.976091,
		0.953979,
		0.922139,
		0.879909,
		0.827229,
		0.76461,
		0.693079,
		0.614133,
		0.529695,
		0.442072,
		0.353922,
		0.268228,
		0.188271,
		0.117611,
		0.060068,
		0.01971,
		0.000836,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.003221,
		0.027329,
		0.071421,
		0.131398,
		0.203384,
		0.283743,
		0.369091,
		0.45631,
		0.542574,
		0.625367,
		0.702516,
		0.772216,
		0.833075,
		0.884152,
		0.925006,
		0.955742,
		0.977043,
		0.990178,
		0.996943,
		0.999493,
		0.999988,
		0.999998,
		0.999732,
		0.997888,
		0.992386,
		0.98103,
		0.961912,
		0.933622,
		0.895329,
		0.846787,
		0.788304,
		0.720701,
		0.645261,
		0.563688,
		0.478069,
		0.390835,
		0.304739,
		0.222825,
		0.148414,
		0.085081,
		0.036642,
		0.007142,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.003147,
		0.026722,
		0.069887,
		0.128681,
		0.199356,
		0.278389,
		0.362499,
		0.448658,
		0.534114,
		0.61641,
		0.69341,
		0.763329,
		0.82477,
		0.876762,
		0.9188,
		0.950898,
		0.973615,
		0.988073,
		0.995914,
		0.999166,
		0.999956,
		1,
		0.999882,
		0.998665,
		0.994468,
		0.985151,
		0.968764,
		0.943785,
		0.909232,
		0.864686,
		0.810266,
		0.746597,
		0.67476,
		0.596255,
		0.512956,
		0.427082,
		0.341166,
		0.258031,
		0.18077,
		0.112726,
		0.057479,
		0.018832,
		0.000798,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.003075,
		0.026134,
		0.068402,
		0.126047,
		0.195443,
		0.273179,
		0.356071,
		0.441178,
		0.52582,
		0.607595,
		0.684409,
		0.754496,
		0.816458,
		0.869296,
		0.912454,
		0.94586,
		0.969962,
		0.985742,
		0.994696,
		0.998722,
		0.999887,
		1,
		0.999957,
		0.999202,
		0.996084,
		0.988552,
		0.974637,
		0.952728,
		0.921711,
		0.881004,
		0.830552,
		0.770794,
		0.702622,
		0.627338,
		0.546619,
		0.462481,
		0.37725,
		0.293538,
		0.214224,
		0.142431,
		0.081516,
		0.035053,
		0.006822,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.003007,
		0.025566,
		0.066963,
		0.123492,
		0.191643,
		0.268109,
		0.349803,
		0.433865,
		0.517687,
		0.598922,
		0.675515,
		0.745722,
		0.808146,
		0.861767,
		0.90598,
		0.94064,
		0.96609,
		0.983187,
		0.993282,
		0.998145,
		0.999766,
		0.999999,
		0.999989,
		0.999556,
		0.997312,
		0.991323,
		0.97963,
		0.960553,
		0.932861,
		0.895825,
		0.849227,
		0.793332,
		0.728851,
		0.656903,
		0.578973,
		0.496887,
		0.412775,
		0.32905,
		0.248386,
		0.1737,
		0.108136,
		0.055053,
		0.018011,
		0.000762,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.00294,
		0.025016,
		0.065569,
		0.121013,
		0.18795,
		0.263174,
		0.343689,
		0.426716,
		0.509713,
		0.59039,
		0.66673,
		0.737013,
		0.799844,
		0.854185,
		0.899393,
		0.935251,
		0.962012,
		0.980411,
		0.991668,
		0.997422,
		0.999575,
		0.99999,
		0.999999,
		0.999776,
		0.998222,
		0.993551,
		0.983837,
		0.967356,
		0.942774,
		0.909232,
		0.866359,
		0.814257,
		0.753466,
		0.684929,
		0.609954,
		0.530184,
		0.447562,
		0.364313,
		0.282919,
		0.2061,
		0.136799,
		0.07817,
		0.033565,
		0.006524,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.002875,
		0.024483,
		0.064217,
		0.118608,
		0.18436,
		0.258369,
		0.337724,
		0.419726,
		0.501896,
		0.582,
		0.658057,
		0.728373,
		0.791559,
		0.846562,
		0.892703,
		0.929706,
		0.957735,
		0.97742,
		0.989852,
		0.996543,
		0.999299,
		0.999963,
		1,
		0.999901,
		0.998875,
		0.995313,
		0.987347,
		0.973229,
		0.951543,
		0.921311,
		0.882021,
		0.833622,
		0.776495,
		0.711414,
		0.63952,
		0.56228,
		0.481465,
		0.399117,
		0.317536,
		0.239256,
		0.16703,
		0.103819,
		0.052777,
		0.017243,
		0.000729,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.002813,
		0.023968,
		0.062907,
		0.116272,
		0.180871,
		0.25369,
		0.331906,
		0.412891,
		0.494234,
		0.573749,
		0.649498,
		0.719808,
		0.783299,
		0.838908,
		0.885923,
		0.924016,
		0.953272,
		0.974219,
		0.987832,
		0.995499,
		0.998921,
		0.999905,
		1,
		0.999964,
		0.999327,
		0.996681,
		0.990244,
		0.978262,
		0.959257,
		0.932145,
		0.896287,
		0.851487,
		0.797974,
		0.736367,
		0.667644,
		0.593109,
		0.514364,
		0.433283,
		0.35199,
		0.272844,
		0.198419,
		0.13149,
		0.075024,
		0.032169,
		0.006244,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.002753,
		0.023468,
		0.061637,
		0.114005,
		0.177478,
		0.249134,
		0.326228,
		0.406208,
		0.486723,
		0.565638,
		0.641052,
		0.71132,
		0.77507,
		0.83123,
		0.879064,
		0.918192,
		0.948631,
		0.970816,
		0.98561,
		0.994282,
		0.998429,
		0.999803,
		0.999999,
		0.999991,
		0.999626,
		0.997721,
		0.992605,
		0.982541,
		0.966004,
		0.94182,
		0.909232,
		0.867913,
		0.817948,
		0.759807,
		0.694316,
		0.622621,
		0.546167,
		0.466665,
		0.386075,
		0.306588,
		0.230606,
		0.160731,
		0.099754,
		0.050639,
		0.016522,
		0.000697,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.002694,
		0.022984,
		0.060404,
		0.111802,
		0.174178,
		0.244695,
		0.320688,
		0.399673,
		0.47936,
		0.557664,
		0.632722,
		0.702914,
		0.766878,
		0.823539,
		0.872136,
		0.912247,
		0.943825,
		0.967218,
		0.983187,
		0.992887,
		0.997811,
		0.999641,
		0.999991,
		0.999999,
		0.999811,
		0.998492,
		0.994502,
		0.986146,
		0.971868,
		0.950418,
		0.920935,
		0.882967,
		0.836467,
		0.781763,
		0.719536,
		0.650784,
		0.5768,
		0.499142,
		0.419614,
		0.340247,
		0.26328,
		0.191151,
		0.12648,
		0.072063,
		0.030859,
		0.005983,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.002638,
		0.022514,
		0.059208,
		0.109662,
		0.170967,
		0.24037,
		0.315281,
		0.393283,
		0.472143,
		0.549827,
		0.624508,
		0.694592,
		0.758729,
		0.815842,
		0.865149,
		0.906191,
		0.938862,
		0.963433,
		0.980567,
		0.991312,
		0.997056,
		0.999406,
		0.999969,
		1,
		0.999917,
		0.999046,
		0.996003,
		0.989154,
		0.976929,
		0.95802,
		0.931471,
		0.896717,
		0.853585,
		0.80227,
		0.743316,
		0.67758,
		0.606211,
		0.53062,
		0.452462,
		0.373618,
		0.296172,
		0.222405,
		0.154777,
		0.09592,
		0.048628,
		0.015846,
		0.000668,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.002583,
		0.022059,
		0.058047,
		0.107582,
		0.167843,
		0.236155,
		0.310002,
		0.387032,
		0.465069,
		0.542124,
		0.61641,
		0.686356,
		0.750628,
		0.808146,
		0.858112,
		0.900035,
		0.933754,
		0.959469,
		0.977753,
		0.989554,
		0.996156,
		0.999083,
		0.99992,
		1,
		0.99997,
		0.999429,
		0.997169,
		0.991635,
		0.981267,
		0.964706,
		0.940917,
		0.909232,
		0.869359,
		0.82137,
		0.765677,
		0.703003,
		0.63436,
		0.561021,
		0.484498,
		0.406528,
		0.32905,
		0.254193,
		0.184267,
		0.121749,
		0.069274,
		0.029626,
		0.005737,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.00253,
		0.021618,
		0.05692,
		0.105561,
		0.164802,
		0.232048,
		0.30485,
		0.380919,
		0.458135,
		0.534555,
		0.608428,
		0.67821,
		0.742579,
		0.800458,
		0.851035,
		0.893788,
		0.928511,
		0.955335,
		0.974752,
		0.987613,
		0.995103,
		0.998662,
		0.999833,
		0.999999,
		0.999992,
		0.999682,
		0.998056,
		0.993658,
		0.984953,
		0.970552,
		0.949349,
		0.920582,
		0.883851,
		0.839109,
		0.786646,
		0.727057,
		0.661224,
		0.590287,
		0.515624,
		0.438833,
		0.361715,
		0.286256,
		0.214622,
		0.149143,
		0.092302,
		0.046734,
		0.015211,
		0.000641,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.002478,
		0.021189,
		0.055825,
		0.103595,
		0.161842,
		0.228043,
		0.299819,
		0.37494,
		0.451339,
		0.527118,
		0.600563,
		0.670154,
		0.734587,
		0.792785,
		0.843926,
		0.887461,
		0.923141,
		0.951039,
		0.971569,
		0.98549,
		0.993892,
		0.99813,
		0.999695,
		0.999993,
		0.999999,
		0.999839,
		0.998713,
		0.995285,
		0.988059,
		0.975631,
		0.956839,
		0.930835,
		0.897121,
		0.855538,
		0.806258,
		0.749757,
		0.686792,
		0.618377,
		0.545761,
		0.470412,
		0.393997,
		0.318369,
		0.245554,
		0.177742,
		0.117276,
		0.066642,
		0.028466,
		0.005506,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.002428,
		0.020774,
		0.054761,
		0.101684,
		0.15896,
		0.224139,
		0.294906,
		0.369091,
		0.444677,
		0.519811,
		0.592813,
		0.662191,
		0.726654,
		0.785132,
		0.836792,
		0.881063,
		0.917657,
		0.946591,
		0.968209,
		0.983187,
		0.99252,
		0.997478,
		0.999493,
		0.999973,
		1,
		0.999929,
		0.999185,
		0.996571,
		0.99065,
		0.980016,
		0.963459,
		0.940063,
		0.909232,
		0.870708,
		0.824552,
		0.771123,
		0.711062,
		0.645261,
		0.574848,
		0.501165,
		0.425753,
		0.350337,
		0.276811,
		0.207232,
		0.143807,
		0.088884,
		0.044948,
		0.014613,
		0.000615,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.00238,
		0.02037,
		0.053727,
		0.099824,
		0.156153,
		0.220331,
		0.290107,
		0.363368,
		0.438147,
		0.512632,
		0.58518,
		0.654322,
		0.718785,
		0.777504,
		0.82964,
		0.874601,
		0.912065,
		0.941999,
		0.964681,
		0.980706,
		0.990983,
		0.996697,
		0.999216,
		0.999932,
		1,
		0.999974,
		0.999512,
		0.997571,
		0.992789,
		0.983772,
		0.969281,
		0.948332,
		0.920249,
		0.884676,
		0.84157,
		0.791183,
		0.73404,
		0.670922,
		0.602837,
		0.531011,
		0.456863,
		0.381994,
		0.308175,
		0.237335,
		0.171552,
		0.113042,
		0.064157,
		0.027373,
		0.005289,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.002333,
		0.019978,
		0.052722,
		0.098015,
		0.153419,
		0.216617,
		0.285421,
		0.357769,
		0.431746,
		0.50558,
		0.577661,
		0.646547,
		0.710982,
		0.769907,
		0.822478,
		0.868085,
		0.906376,
		0.937272,
		0.960989,
		0.978051,
		0.989282,
		0.995782,
		0.998853,
		0.999857,
		0.999999,
		0.999993,
		0.999729,
		0.998331,
		0.994532,
		0.986964,
		0.974369,
		0.95571,
		0.930235,
		0.897499,
		0.85736,
		0.809968,
		0.755743,
		0.695352,
		0.629697,
		0.559887,
		0.487226,
		0.413199,
		0.339457,
		0.267809,
		0.200209,
		0.138748,
		0.08565,
		0.043262,
		0.014049,
		0.00059,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.002287,
		0.019597,
		0.051745,
		0.096253,
		0.150754,
		0.212995,
		0.280842,
		0.352291,
		0.425471,
		0.498652,
		0.570256,
		0.638868,
		0.703249,
		0.762345,
		0.815311,
		0.861523,
		0.900598,
		0.932418,
		0.957143,
		0.975226,
		0.987414,
		0.994726,
		0.998392,
		0.999739,
		0.999994,
		0.999999,
		0.999863,
		0.998895,
		0.995933,
		0.989654,
		0.97879,
		0.962262,
		0.939252,
		0.909232,
		0.87197,
		0.827516,
		0.77619,
		0.718555,
		0.655405,
		0.587741,
		0.516759,
		0.44383,
		0.370494,
		0.298442,
		0.229511,
		0.165674,
		0.109032,
		0.061808,
		0.026342,
		0.005084,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.002243,
		0.019227,
		0.050795,
		0.094539,
		0.148158,
		0.20946,
		0.276369,
		0.34693,
		0.419319,
		0.491846,
		0.562964,
		0.631286,
		0.695586,
		0.754822,
		0.808146,
		0.854921,
		0.89474,
		0.927446,
		0.953149,
		0.972236,
		0.985382,
		0.993526,
		0.997826,
		0.999565,
		0.999977,
		1,
		0.999939,
		0.9993,
		0.997042,
		0.991898,
		0.982605,
		0.968053,
		0.947363,
		0.919935,
		0.885451,
		0.843868,
		0.795408,
		0.740538,
		0.679948,
		0.614537,
		0.545395,
		0.473788,
		0.401146,
		0.32905,
		0.259225,
		0.193529,
		0.133949,
		0.082589,
		0.041669,
		0.013517,
		0.000567,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.0022,
		0.018867,
		0.049871,
		0.09287,
		0.145627,
		0.20601,
		0.271997,
		0.341684,
		0.413289,
		0.48516,
		0.555785,
		0.623799,
		0.687996,
		0.747342,
		0.800988,
		0.848287,
		0.88881,
		0.922365,
		0.949014,
		0.969087,
		0.983187,
		0.992177,
		0.997147,
		0.999326,
		0.999942,
		1,
		0.999978,
		0.999581,
		0.997904,
		0.99375,
		0.985874,
		0.973143,
		0.95463,
		0.929666,
		0.897854,
		0.859064,
		0.813428,
		0.761317,
		0.703324,
		0.64025,
		0.573084,
		0.502989,
		0.431293,
		0.359472,
		0.289144,
		0.222058,
		0.160089,
		0.105229,
		0.059585,
		0.025367,
		0.004891,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.002158,
		0.018518,
		0.048971,
		0.091244,
		0.14316,
		0.202643,
		0.267725,
		0.336549,
		0.407377,
		0.478594,
		0.548718,
		0.61641,
		0.680482,
		0.739908,
		0.793842,
		0.841627,
		0.882814,
		0.917181,
		0.944746,
		0.965784,
		0.980831,
		0.990679,
		0.996348,
		0.999011,
		0.999877,
		0.999999,
		0.999994,
		0.999767,
		0.99856,
		0.99526,
		0.988652,
		0.977591,
		0.961112,
		0.938482,
		0.909232,
		0.873152,
		0.830285,
		0.780913,
		0.725537,
		0.664862,
		0.599783,
		0.531365,
		0.460836,
		0.389573,
		0.319089,
		0.251033,
		0.187172,
		0.129392,
		0.079688,
		0.040162,
		0.013016,
		0.000546,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.002117,
		0.018178,
		0.048096,
		0.08966,
		0.140754,
		0.199356,
		0.263549,
		0.331523,
		0.40158,
		0.472143,
		0.541761,
		0.609118,
		0.673044,
		0.732524,
		0.786713,
		0.834948,
		0.876762,
		0.911904,
		0.940353,
		0.962333,
		0.978317,
		0.989031,
		0.995423,
		0.99861,
		0.999775,
		0.999995,
		0.999999,
		0.999882,
		0.999046,
		0.996474,
		0.990993,
		0.981456,
		0.966867,
		0.94644,
		0.919639,
		0.886177,
		0.846017,
		0.799352,
		0.746597,
		0.688367,
		0.625463,
		0.55886,
		0.48969,
		0.419232,
		0.348906,
		0.280256,
		0.214952,
		0.154777,
		0.101621,
		0.057479,
		0.024446,
		0.004709,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.002078,
		0.017847,
		0.047243,
		0.088117,
		0.138407,
		0.196146,
		0.259467,
		0.326602,
		0.395897,
		0.465807,
		0.534913,
		0.601923,
		0.665684,
		0.725192,
		0.779606,
		0.828254,
		0.870658,
		0.90654,
		0.935842,
		0.958739,
		0.97565,
		0.987232,
		0.994368,
		0.998117,
		0.999625,
		0.99998,
		1,
		0.999948,
		0.999396,
		0.997435,
		0.992945,
		0.984791,
		0.971951,
		0.953597,
		0.929128,
		0.898188,
		0.860662,
		0.816662,
		0.76652,
		0.710763,
		0.650105,
		0.585432,
		0.517785,
		0.448352,
		0.378457,
		0.309553,
		0.243212,
		0.181117,
		0.125061,
		0.076937,
		0.038736,
		0.012541,
		0.000525,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.00204,
		0.017525,
		0.046414,
		0.086613,
		0.136118,
		0.193012,
		0.255475,
		0.321785,
		0.390324,
		0.459584,
		0.528173,
		0.594824,
		0.658402,
		0.717915,
		0.772523,
		0.821553,
		0.864511,
		0.901097,
		0.931219,
		0.95501,
		0.972832,
		0.985283,
		0.993181,
		0.997523,
		0.999417,
		0.99995,
		1,
		0.999981,
		0.999638,
		0.998182,
		0.994557,
		0.987648,
		0.97642,
		0.960006,
		0.937751,
		0.909232,
		0.874262,
		0.832877,
		0.785326,
		0.732056,
		0.673697,
		0.611048,
		0.545064,
		0.476846,
		0.407628,
		0.338772,
		0.271757,
		0.208175,
		0.149721,
		0.098193,
		0.055483,
		0.023574,
		0.004537,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.002002,
		0.017212,
		0.045605,
		0.085147,
		0.133884,
		0.189951,
		0.251572,
		0.317068,
		0.384859,
		0.453471,
		0.521539,
		0.587822,
		0.651201,
		0.710695,
		0.76547,
		0.814849,
		0.858326,
		0.895582,
		0.926493,
		0.951151,
		0.96987,
		0.983187,
		0.991857,
		0.996822,
		0.999142,
		0.999894,
		0.999999,
		0.999995,
		0.999798,
		0.99875,
		0.995871,
		0.990077,
		0.980326,
		0.965721,
		0.94556,
		0.919359,
		0.886861,
		0.848031,
		0.803041,
		0.752258,
		0.696234,
		0.635684,
		0.571482,
		0.504643,
		0.436317,
		0.367779,
		0.300419,
		0.23574,
		0.175346,
		0.120941,
		0.074325,
		0.037384,
		0.012092,
		0.000506,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.001966,
		0.016907,
		0.044818,
		0.083718,
		0.131705,
		0.186961,
		0.247755,
		0.31245,
		0.379501,
		0.447466,
		0.515011,
		0.580916,
		0.64408,
		0.703534,
		0.758449,
		0.808146,
		0.85211,
		0.890001,
		0.92167,
		0.94717,
		0.966766,
		0.980944,
		0.990397,
		0.996009,
		0.998793,
		0.999805,
		0.999995,
		0.999999,
		0.999898,
		0.999172,
		0.996928,
		0.992122,
		0.983719,
		0.970795,
		0.952607,
		0.928617,
		0.898503,
		0.862161,
		0.819691,
		0.771385,
		0.717717,
		0.659326,
		0.597004,
		0.531686,
		0.464441,
		0.39646,
		0.32905,
		0.263625,
		0.201706,
		0.144906,
		0.094935,
		0.053588,
		0.022747,
		0.004374,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.001931,
		0.01661,
		0.044051,
		0.082324,
		0.129577,
		0.184039,
		0.244022,
		0.307926,
		0.374245,
		0.441568,
		0.508587,
		0.574105,
		0.63704,
		0.696434,
		0.751464,
		0.80145,
		0.845868,
		0.884362,
		0.916757,
		0.943071,
		0.963527,
		0.978558,
		0.9888,
		0.995079,
		0.998361,
		0.999675,
		0.999983,
		1,
		0.999955,
		0.999475,
		0.997765,
		0.993829,
		0.986646,
		0.975277,
		0.958943,
		0.937054,
		0.909232,
		0.875306,
		0.835308,
		0.789457,
		0.738155,
		0.681965,
		0.621603,
		0.557927,
		0.491928,
		0.424716,
		0.357518,
		0.291667,
		0.228598,
		0.169842,
		0.11702,
		0.071843,
		0.036101,
		0.011667,
		0.000488,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.001896,
		0.016321,
		0.043303,
		0.080964,
		0.1275,
		0.181184,
		0.24037,
		0.303496,
		0.369091,
		0.435774,
		0.502265,
		0.567389,
		0.630081,
		0.689396,
		0.744517,
		0.794763,
		0.839605,
		0.87867,
		0.91176,
		0.938862,
		0.960158,
		0.976031,
		0.987065,
		0.994029,
		0.997839,
		0.999493,
		0.999956,
		1,
		0.999984,
		0.999686,
		0.998415,
		0.995238,
		0.989154,
		0.979216,
		0.964615,
		0.944719,
		0.919093,
		0.887506,
		0.849923,
		0.806498,
		0.757559,
		0.703599,
		0.645261,
		0.583329,
		0.518718,
		0.452462,
		0.385711,
		0.319719,
		0.255841,
		0.195527,
		0.140317,
		0.091835,
		0.051789,
		0.021964,
		0.004219,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.001863,
		0.016039,
		0.042574,
		0.079638,
		0.125472,
		0.178394,
		0.236797,
		0.299157,
		0.364035,
		0.430082,
		0.496045,
		0.560768,
		0.623205,
		0.682423,
		0.737611,
		0.788092,
		0.833326,
		0.872931,
		0.906686,
		0.934549,
		0.956664,
		0.973368,
		0.985193,
		0.992856,
		0.997222,
		0.999253,
		0.999908,
		0.999999,
		0.999996,
		0.999825,
		0.99891,
		0.996388,
		0.991285,
		0.982658,
		0.969673,
		0.951659,
		0.928131,
		0.898801,
		0.863572,
		0.822533,
		0.775945,
		0.724231,
		0.667967,
		0.607863,
		0.544762,
		0.479627,
		0.413531,
		0.347655,
		0.283276,
		0.221767,
		0.164588,
		0.113285,
		0.069482,
		0.034883,
		0.011263,
		0.000471,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.00183,
		0.015765,
		0.041863,
		0.078344,
		0.123492,
		0.175666,
		0.2333,
		0.294906,
		0.359076,
		0.424491,
		0.489923,
		0.55424,
		0.61641,
		0.675515,
		0.73075,
		0.781437,
		0.827036,
		0.867151,
		0.901541,
		0.930138,
		0.95305,
		0.970571,
		0.983187,
		0.991558,
		0.996504,
		0.998946,
		0.999831,
		0.999996,
		0.999999,
		0.999911,
		0.999278,
		0.997312,
		0.993081,
		0.985648,
		0.974162,
		0.95792,
		0.93639,
		0.909232,
		0.87629,
		0.837591,
		0.793332,
		0.743871,
		0.689715,
		0.631507,
		0.570021,
		0.506149,
		0.440894,
		0.375363,
		0.310761,
		0.248386,
		0.189623,
		0.135939,
		0.088884,
		0.050078,
		0.02122,
		0.004073,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.001798,
		0.015497,
		0.04117,
		0.077081,
		0.121558,
		0.173,
		0.229879,
		0.290741,
		0.354211,
		0.418999,
		0.483901,
		0.547804,
		0.609698,
		0.668673,
		0.723934,
		0.774805,
		0.820741,
		0.861335,
		0.896331,
		0.925634,
		0.949321,
		0.967646,
		0.981047,
		0.990135,
		0.995681,
		0.998566,
		0.999716,
		0.999985,
		1,
		0.999961,
		0.999542,
		0.998044,
		0.99458,
		0.988228,
		0.978128,
		0.963547,
		0.943915,
		0.918841,
		0.888115,
		0.851704,
		0.809744,
		0.762531,
		0.710506,
		0.654248,
		0.594464,
		0.531979,
		0.467727,
		0.402746,
		0.338171,
		0.275228,
		0.21523,
		0.159571,
		0.109724,
		0.067236,
		0.033726,
		0.01088,
		0.000454,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.001767,
		0.015236,
		0.040494,
		0.075848,
		0.119668,
		0.170393,
		0.22653,
		0.28666,
		0.349439,
		0.413603,
		0.477974,
		0.541461,
		0.603067,
		0.661898,
		0.717165,
		0.768196,
		0.814442,
		0.855488,
		0.891062,
		0.921045,
		0.945484,
		0.964597,
		0.978775,
		0.988586,
		0.994749,
		0.998106,
		0.999557,
		0.999962,
		1,
		0.999986,
		0.999726,
		0.998613,
		0.995817,
		0.990437,
		0.981612,
		0.968585,
		0.950749,
		0.927669,
		0.899083,
		0.864902,
		0.825205,
		0.780225,
		0.730344,
		0.676077,
		0.618068,
		0.557076,
		0.493969,
		0.429721,
		0.365398,
		0.302157,
		0.241242,
		0.183976,
		0.131761,
		0.086071,
		0.048451,
		0.020513,
		0.003934,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.001737,
		0.014982,
		0.039835,
		0.074644,
		0.117821,
		0.167843,
		0.223252,
		0.282661,
		0.344756,
		0.408303,
		0.472143,
		0.535208,
		0.596519,
		0.655192,
		0.710447,
		0.761615,
		0.808146,
		0.849616,
		0.885739,
		0.916375,
		0.941544,
		0.961427,
		0.976376,
		0.986911,
		0.993706,
		0.997561,
		0.999346,
		0.99992,
		1,
		0.999996,
		0.999847,
		0.999046,
		0.996826,
		0.992316,
		0.984657,
		0.973075,
		0.956935,
		0.935757,
		0.909232,
		0.877219,
		0.839741,
		0.796974,
		0.749239,
		0.696992,
		0.640815,
		0.581406,
		0.51957,
		0.456215,
		0.392345,
		0.32905,
		0.267506,
		0.208972,
		0.154777,
		0.106327,
		0.065096,
		0.032625,
		0.010517,
		0.000439,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.001708,
		0.014734,
		0.039191,
		0.073468,
		0.116017,
		0.165349,
		0.220042,
		0.278741,
		0.340162,
		0.403095,
		0.466406,
		0.529046,
		0.590052,
		0.648554,
		0.70378,
		0.755064,
		0.801855,
		0.843722,
		0.880367,
		0.911631,
		0.937506,
		0.958141,
		0.973851,
		0.985111,
		0.992549,
		0.996925,
		0.999076,
		0.999852,
		0.999996,
		1,
		0.999923,
		0.999367,
		0.997637,
		0.993899,
		0.987301,
		0.977061,
		0.962515,
		0.943146,
		0.918602,
		0.88869,
		0.853382,
		0.812798,
		0.767202,
		0.716994,
		0.662695,
		0.604944,
		0.544486,
		0.482168,
		0.418929,
		0.355799,
		0.29389,
		0.234393,
		0.178574,
		0.127771,
		0.083389,
		0.046901,
		0.019841,
		0.003803,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.001679,
		0.014492,
		0.038563,
		0.07232,
		0.114253,
		0.16291,
		0.2169,
		0.2749,
		0.335654,
		0.397978,
		0.460761,
		0.522972,
		0.583667,
		0.641985,
		0.697165,
		0.748546,
		0.795573,
		0.837812,
		0.874952,
		0.906817,
		0.933375,
		0.954745,
		0.971204,
		0.983187,
		0.991278,
		0.996194,
		0.998741,
		0.999752,
		0.999987,
		1,
		0.999966,
		0.999599,
		0.99828,
		0.995219,
		0.989583,
		0.980581,
		0.967529,
		0.949876,
		0.927229,
		0.89935,
		0.866157,
		0.827721,
		0.78425,
		0.736089,
		0.683702,
		0.627672,
		0.568684,
		0.507526,
		0.44508,
		0.382312,
		0.320274,
		0.260094,
		0.202976,
		0.150192,
		0.103084,
		0.063057,
		0.031577,
		0.010171,
		0.000424,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.001651,
		0.014256,
		0.03795,
		0.071199,
		0.112529,
		0.160523,
		0.213823,
		0.271135,
		0.331231,
		0.39295,
		0.455206,
		0.516987,
		0.577362,
		0.635486,
		0.690605,
		0.742062,
		0.789303,
		0.831889,
		0.869498,
		0.901939,
		0.929157,
		0.951243,
		0.968439,
		0.98114,
		0.989891,
		0.995364,
		0.998334,
		0.999612,
		0.999967,
		1,
		0.999988,
		0.99976,
		0.99878,
		0.99631,
		0.991538,
		0.983674,
		0.972017,
		0.955987,
		0.935153,
		0.909232,
		0.878098,
		0.841768,
		0.800401,
		0.754286,
		0.703836,
		0.649575,
		0.592135,
		0.532247,
		0.470734,
		0.408505,
		0.346552,
		0.285944,
		0.227824,
		0.173402,
		0.123957,
		0.08083,
		0.045425,
		0.019201,
		0.003677,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.001624,
		0.014026,
		0.037351,
		0.070103,
		0.110843,
		0.158187,
		0.210809,
		0.267443,
		0.326889,
		0.38801,
		0.449741,
		0.511089,
		0.571138,
		0.629057,
		0.6841,
		0.735615,
		0.783049,
		0.825958,
		0.864011,
		0.897002,
		0.924857,
		0.947639,
		0.965558,
		0.978974,
		0.988388,
		0.994434,
		0.99785,
		0.999426,
		0.99993,
		1,
		0.999997,
		0.999866,
		0.999161,
		0.997199,
		0.993199,
		0.986377,
		0.976017,
		0.961517,
		0.94241,
		0.918374,
		0.889236,
		0.854966,
		0.815675,
		0.771599,
		0.723099,
		0.670646,
		0.61482,
		0.556295,
		0.495839,
		0.434306,
		0.372632,
		0.311827,
		0.252975,
		0.197229,
		0.145806,
		0.099986,
		0.061111,
		0.030578,
		0.009842,
		0.00041,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.001597,
		0.013802,
		0.036766,
		0.069032,
		0.109195,
		0.155901,
		0.207857,
		0.263824,
		0.322628,
		0.383156,
		0.444363,
		0.505276,
		0.564995,
		0.622699,
		0.677651,
		0.729206,
		0.776812,
		0.820021,
		0.858494,
		0.892012,
		0.92048,
		0.943938,
		0.962567,
		0.976689,
		0.986769,
		0.9934,
		0.997284,
		0.999187,
		0.99987,
		0.999997,
		1,
		0.999932,
		0.999443,
		0.997915,
		0.994599,
		0.988724,
		0.979567,
		0.966504,
		0.949038,
		0.926809,
		0.899603,
		0.867344,
		0.830094,
		0.788042,
		0.741497,
		0.690882,
		0.636722,
		0.57964,
		0.520351,
		0.459655,
		0.398433,
		0.33764,
		0.278304,
		0.22152,
		0.168448,
		0.120309,
		0.078386,
		0.044016,
		0.018592,
		0.003558,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.001571,
		0.013582,
		0.036195,
		0.067986,
		0.107582,
		0.153664,
		0.204965,
		0.260276,
		0.318445,
		0.378386,
		0.439072,
		0.499548,
		0.55893,
		0.61641,
		0.67126,
		0.722839,
		0.770596,
		0.814083,
		0.852952,
		0.886971,
		0.916031,
		0.940146,
		0.959469,
		0.974289,
		0.985035,
		0.992261,
		0.996633,
		0.998889,
		0.999782,
		0.999989,
		1,
		0.99997,
		0.999647,
		0.998482,
		0.995768,
		0.990749,
		0.982702,
		0.970985,
		0.955074,
		0.934575,
		0.909232,
		0.878929,
		0.843682,
		0.803632,
		0.759042,
		0.710282,
		0.657831,
		0.60226,
		0.544233,
		0.484498,
		0.423884,
		0.363292,
		0.303696,
		0.246136,
		0.191718,
		0.141606,
		0.097025,
		0.059254,
		0.029626,
		0.009529,
		0.000397,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.001546,
		0.013368,
		0.035637,
		0.066963,
		0.106005,
		0.151474,
		0.202132,
		0.256796,
		0.31434,
		0.373698,
		0.433865,
		0.493904,
		0.552945,
		0.610192,
		0.664927,
		0.716513,
		0.764404,
		0.808146,
		0.847388,
		0.881887,
		0.911514,
		0.936266,
		0.956268,
		0.971777,
		0.983187,
		0.991015,
		0.995893,
		0.998528,
		0.999658,
		0.999971,
		1,
		0.999989,
		0.999788,
		0.998923,
		0.996733,
		0.992484,
		0.985457,
		0.974996,
		0.960553,
		0.941704,
		0.918157,
		0.889753,
		0.856465,
		0.81839,
		0.775744,
		0.728851,
		0.678141,
		0.624138,
		0.567455,
		0.50879,
		0.448922,
		0.388702,
		0.32905,
		0.270954,
		0.215467,
		0.163699,
		0.116819,
		0.076051,
		0.042672,
		0.018011,
		0.003444,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.001521,
		0.013159,
		0.035092,
		0.065963,
		0.104462,
		0.14933,
		0.199356,
		0.253383,
		0.310309,
		0.369091,
		0.428742,
		0.488343,
		0.547038,
		0.604044,
		0.658652,
		0.710231,
		0.758237,
		0.802214,
		0.841807,
		0.876762,
		0.906935,
		0.932304,
		0.952968,
		0.969156,
		0.981226,
		0.989663,
		0.99506,
		0.998097,
		0.999493,
		0.999938,
		1,
		0.999997,
		0.999882,
		0.999259,
		0.99752,
		0.993959,
		0.987863,
		0.97857,
		0.965511,
		0.948233,
		0.926409,
		0.899843,
		0.868468,
		0.832336,
		0.791619,
		0.746597,
		0.697651,
		0.645261,
		0.589991,
		0.532493,
		0.473495,
		0.413799,
		0.354276,
		0.295864,
		0.239562,
		0.18643,
		0.137583,
		0.094192,
		0.057479,
		0.028718,
		0.00923,
		0.000384,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.001497,
		0.012955,
		0.034559,
		0.064984,
		0.102952,
		0.14723,
		0.196635,
		0.250035,
		0.306352,
		0.364563,
		0.423701,
		0.482863,
		0.541209,
		0.597967,
		0.652437,
		0.703994,
		0.752097,
		0.796291,
		0.836212,
		0.8716,
		0.902298,
		0.928264,
		0.949573,
		0.966428,
		0.979155,
		0.988203,
		0.994133,
		0.997593,
		0.999281,
		0.999886,
		0.999997,
		1,
		0.99994,
		0.999508,
		0.998153,
		0.995203,
		0.989954,
		0.981742,
		0.969981,
		0.954194,
		0.934022,
		0.909232,
		0.879718,
		0.845492,
		0.806683,
		0.763528,
		0.716363,
		0.665622,
		0.611825,
		0.555576,
		0.497557,
		0.438523,
		0.379296,
		0.320766,
		0.263882,
		0.209654,
		0.159146,
		0.113477,
		0.073818,
		0.041389,
		0.017457,
		0.003336,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.001474,
		0.012755,
		0.034038,
		0.064028,
		0.101474,
		0.145174,
		0.193969,
		0.246752,
		0.302467,
		0.360112,
		0.41874,
		0.477463,
		0.535457,
		0.59196,
		0.646281,
		0.697803,
		0.745987,
		0.790377,
		0.830607,
		0.866407,
		0.897607,
		0.924149,
		0.946088,
		0.963598,
		0.976975,
		0.986637,
		0.993109,
		0.997011,
		0.999017,
		0.999807,
		0.99999,
		1,
		0.999973,
		0.999688,
		0.998655,
		0.99624,
		0.991758,
		0.984543,
		0.973997,
		0.959621,
		0.941027,
		0.91795,
		0.890244,
		0.857883,
		0.820956,
		0.779657,
		0.73428,
		0.685216,
		0.63294,
		0.578013,
		0.52107,
		0.46282,
		0.40404,
		0.345573,
		0.288319,
		0.233241,
		0.181354,
		0.133727,
		0.09148,
		0.055783,
		0.027851,
		0.008946,
		0.000372,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.001451,
		0.01256,
		0.033529,
		0.063092,
		0.100028,
		0.14316,
		0.191355,
		0.24353,
		0.298652,
		0.355737,
		0.413858,
		0.472143,
		0.529781,
		0.586022,
		0.640185,
		0.691659,
		0.739908,
		0.784477,
		0.824995,
		0.861185,
		0.892867,
		0.919966,
		0.942516,
		0.960668,
		0.974689,
		0.984964,
		0.991988,
		0.996348,
		0.998695,
		0.999698,
		0.999974,
		1,
		0.99999,
		0.999813,
		0.999046,
		0.997097,
		0.993303,
		0.987004,
		0.977591,
		0.964547,
		0.947458,
		0.926026,
		0.900072,
		0.869533,
		0.834457,
		0.794999,
		0.751412,
		0.704043,
		0.653328,
		0.599783,
		0.544,
		0.486644,
		0.428447,
		0.370205,
		0.312776,
		0.257074,
		0.204067,
		0.154777,
		0.110275,
		0.071682,
		0.040162,
		0.016928,
		0.003233,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.001428,
		0.01237,
		0.033031,
		0.062177,
		0.098612,
		0.141187,
		0.188793,
		0.24037,
		0.294906,
		0.351437,
		0.409054,
		0.466901,
		0.52418,
		0.580154,
		0.63415,
		0.685564,
		0.733863,
		0.778592,
		0.819379,
		0.855939,
		0.888082,
		0.915718,
		0.938862,
		0.957642,
		0.972299,
		0.983187,
		0.990768,
		0.995601,
		0.99831,
		0.999551,
		0.999945,
		1,
		0.999998,
		0.999896,
		0.999343,
		0.997796,
		0.994617,
		0.989154,
		0.980795,
		0.969004,
		0.953345,
		0.933492,
		0.909232,
		0.880467,
		0.847207,
		0.809569,
		0.767766,
		0.722107,
		0.672983,
		0.620871,
		0.566321,
		0.509955,
		0.452462,
		0.394596,
		0.337168,
		0.281048,
		0.227159,
		0.176478,
		0.130029,
		0.088884,
		0.054161,
		0.027023,
		0.008674,
		0.000361,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		9.1e-005,
		0.00706,
		0.024223,
		0.05033,
		0.084169,
		0.124573,
		0.170419,
		0.220628,
		0.27417,
		0.330063,
		0.387376,
		0.445233,
		0.502815,
		0.559361,
		0.614173,
		0.666622,
		0.716147,
		0.762263,
		0.804566,
		0.842738,
		0.876554,
		0.905884,
		0.930704,
		0.951098,
		0.967266,
		0.979518,
		0.988279,
		0.994072,
		0.997501,
		0.999217,
		0.999861,
		0.999995,
		1,
		0.999962,
		0.999635,
		0.998532,
		0.996042,
		0.99151,
		0.984305,
		0.973858,
		0.959694,
		0.941445,
		0.918855,
		0.891784,
		0.860201,
		0.824185,
		0.783913,
		0.739658,
		0.691782,
		0.64073,
		0.587025,
		0.531263,
		0.474108,
		0.416289,
		0.358598,
		0.301882,
		0.247045,
		0.195044,
		0.146884,
		0.103621,
		0.066356,
		0.036235,
		0.014447,
		0.002221,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		0.00082,
		0.010606,
		0.030412,
		0.058969,
		0.095049,
		0.137468,
		0.185091,
		0.236827,
		0.291635,
		0.348527,
		0.406567,
		0.464877,
		0.522638,
		0.579092,
		0.63355,
		0.685391,
		0.73407,
		0.779121,
		0.820162,
		0.856905,
		0.889157,
		0.916829,
		0.939942,
		0.958631,
		0.97315,
		0.983869,
		0.991269,
		0.995927,
		0.998488,
		0.999624,
		0.999961,
		1,
		0.999995,
		0.999857,
		0.999193,
		0.997428,
		0.993901,
		0.987949,
		0.978957,
		0.966392,
		0.949827,
		0.92895,
		0.903567,
		0.873601,
		0.839085,
		0.800163,
		0.757075,
		0.710157,
		0.659834,
		0.606611,
		0.551073,
		0.493875,
		0.435741,
		0.377459,
		0.31988,
		0.26391,
		0.210511,
		0.160699,
		0.115538,
		0.076143,
		0.043674,
		0.019337,
		0.004379,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.002292,
		0.014896,
		0.037337,
		0.068326,
		0.106619,
		0.151018,
		0.200371,
		0.253577,
		0.309587,
		0.367405,
		0.426093,
		0.484769,
		0.542616,
		0.598882,
		0.652886,
		0.704019,
		0.751752,
		0.79564,
		0.835329,
		0.870561,
		0.90118,
		0.927141,
		0.948514,
		0.965488,
		0.978376,
		0.987608,
		0.993724,
		0.997351,
		0.999169,
		0.999852,
		0.999995,
		1,
		0.99996,
		0.999612,
		0.998443,
		0.995807,
		0.991018,
		0.983418,
		0.972419,
		0.957534,
		0.938392,
		0.914741,
		0.886452,
		0.853514,
		0.816028,
		0.774201,
		0.72834,
		0.678846,
		0.626207,
		0.570991,
		0.513845,
		0.455485,
		0.396697,
		0.33833,
		0.281295,
		0.226559,
		0.175145,
		0.128128,
		0.086636,
		0.051842,
		0.024968,
		0.007282,
		9.3e-005,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		0.004519,
		0.019939,
		0.045004,
		0.078406,
		0.118882,
		0.165218,
		0.216251,
		0.270869,
		0.328012,
		0.386681,
		0.44593,
		0.504881,
		0.562719,
		0.618696,
		0.672143,
		0.722464,
		0.769149,
		0.811778,
		0.850025,
		0.883666,
		0.912588,
		0.936792,
		0.956401,
		0.971663,
		0.982951,
		0.990759,
		0.995683,
		0.998396,
		0.9996,
		0.999959,
		1,
		0.999994,
		0.999848,
		0.999143,
		0.997272,
		0.993541,
		0.987255,
		0.977776,
		0.964555,
		0.947158,
		0.925275,
		0.89872,
		0.867431,
		0.831466,
		0.790993,
		0.746289,
		0.697728,
		0.645775,
		0.590984,
		0.533987,
		0.475496,
		0.416289,
		0.357216,
		0.299189,
		0.243179,
		0.190218,
		0.141391,
		0.097836,
		0.060743,
		0.031349,
		0.01094,
		0.000846,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		9.6e-005,
		0.007514,
		0.025748,
		0.053423,
		0.089212,
		0.131836,
		0.180066,
		0.232725,
		0.28869,
		0.346894,
		0.406331,
		0.466054,
		0.525185,
		0.582912,
		0.638497,
		0.69128,
		0.740684,
		0.78622,
		0.827492,
		0.864209,
		0.896184,
		0.92335,
		0.945759,
		0.96359,
		0.977154,
		0.986889,
		0.99335,
		0.997189,
		0.999117,
		0.999843,
		0.999994,
		1,
		0.999957,
		0.999588,
		0.998347,
		0.995554,
		0.99049,
		0.982468,
		0.970881,
		0.95523,
		0.935141,
		0.910368,
		0.880796,
		0.846434,
		0.80741,
		0.763962,
		0.716437,
		0.665277,
		0.611015,
		0.554272,
		0.495746,
		0.436213,
		0.376519,
		0.317576,
		0.260362,
		0.205913,
		0.155325,
		0.109748,
		0.070384,
		0.038489,
		0.015366,
		0.002366,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		0.000874,
		0.01129,
		0.03233,
		0.062598,
		0.100747,
		0.145482,
		0.195557,
		0.249783,
		0.307026,
		0.366214,
		0.426333,
		0.486437,
		0.545648,
		0.60316,
		0.658245,
		0.710256,
		0.758636,
		0.802918,
		0.842738,
		0.877839,
		0.908079,
		0.933436,
		0.95402,
		0.970072,
		0.981968,
		0.990212,
		0.995421,
		0.998296,
		0.999575,
		0.999956,
		1,
		0.999994,
		0.999838,
		0.999089,
		0.997104,
		0.993153,
		0.98651,
		0.976511,
		0.962592,
		0.944313,
		0.921364,
		0.893571,
		0.86089,
		0.823405,
		0.781314,
		0.734932,
		0.684672,
		0.631048,
		0.574663,
		0.516206,
		0.456443,
		0.396218,
		0.336442,
		0.278096,
		0.222223,
		0.169927,
		0.122371,
		0.08077,
		0.046396,
		0.02057,
		0.004665,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.002443,
		0.015859,
		0.039695,
		0.072536,
		0.113014,
		0.159817,
		0.211684,
		0.267414,
		0.325863,
		0.385951,
		0.446662,
		0.507048,
		0.566235,
		0.623424,
		0.677898,
		0.729027,
		0.776274,
		0.8192,
		0.857473,
		0.890876,
		0.919314,
		0.942819,
		0.96156,
		0.975845,
		0.986118,
		0.992949,
		0.997015,
		0.999061,
		0.999833,
		0.999994,
		1,
		0.999955,
		0.999561,
		0.998243,
		0.995282,
		0.989923,
		0.98145,
		0.969235,
		0.952769,
		0.931675,
		0.905717,
		0.874792,
		0.838934,
		0.7983,
		0.753166,
		0.703917,
		0.651042,
		0.595126,
		0.536843,
		0.476952,
		0.416289,
		0.355767,
		0.296367,
		0.239139,
		0.185194,
		0.135705,
		0.091904,
		0.055077,
		0.026564,
		0.007758,
		0.0001,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.004818,
		0.021232,
		0.047852,
		0.083242,
		0.126014,
		0.174837,
		0.228439,
		0.285604,
		0.345181,
		0.406083,
		0.46729,
		0.527856,
		0.586911,
		0.643665,
		0.697414,
		0.747548,
		0.793551,
		0.835018,
		0.871652,
		0.90328,
		0.929857,
		0.951476,
		0.968369,
		0.980913,
		0.989624,
		0.995138,
		0.998188,
		0.999547,
		0.999953,
		1,
		0.999994,
		0.999827,
		0.999031,
		0.996923,
		0.992736,
		0.985711,
		0.975156,
		0.960493,
		0.941275,
		0.917197,
		0.888097,
		0.853952,
		0.814873,
		0.771094,
		0.722968,
		0.670956,
		0.615622,
		0.557624,
		0.497709,
		0.436709,
		0.375532,
		0.315161,
		0.25665,
		0.201118,
		0.14975,
		0.103789,
		0.064539,
		0.033357,
		0.011657,
		0.000903,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000103,
		0.008014,
		0.02742,
		0.056809,
		0.094719,
		0.139745,
		0.190539,
		0.245813,
		0.304341,
		0.364963,
		0.426585,
		0.488188,
		0.548826,
		0.607637,
		0.66384,
		0.716747,
		0.76577,
		0.810421,
		0.850326,
		0.885232,
		0.915012,
		0.93968,
		0.959389,
		0.974442,
		0.985289,
		0.992516,
		0.996827,
		0.999,
		0.999822,
		0.999993,
		1,
		0.999952,
		0.999532,
		0.998131,
		0.994989,
		0.989313,
		0.980357,
		0.967473,
		0.950139,
		0.927979,
		0.900765,
		0.868415,
		0.830985,
		0.788668,
		0.741777,
		0.690745,
		0.636109,
		0.578511,
		0.518684,
		0.45745,
		0.395714,
		0.33446,
		0.274744,
		0.217694,
		0.164503,
		0.116428,
		0.074788,
		0.040959,
		0.016376,
		0.002525,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.000934,
		0.012042,
		0.034434,
		0.066571,
		0.10697,
		0.154206,
		0.206915,
		0.263794,
		0.323607,
		0.385184,
		0.44743,
		0.509323,
		0.569922,
		0.628371,
		0.683903,
		0.73585,
		0.783645,
		0.826833,
		0.865077,
		0.89817,
		0.926039,
		0.948756,
		0.966545,
		0.979781,
		0.988991,
		0.994834,
		0.998072,
		0.999517,
		0.99995,
		1,
		0.999993,
		0.999816,
		0.998968,
		0.996728,
		0.992288,
		0.984851,
		0.973702,
		0.958246,
		0.938031,
		0.912756,
		0.882275,
		0.846589,
		0.805839,
		0.760297,
		0.710363,
		0.656545,
		0.599465,
		0.53984,
		0.478481,
		0.416289,
		0.354245,
		0.293408,
		0.234912,
		0.179959,
		0.12982,
		0.085828,
		0.049379,
		0.021925,
		0.004979,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.00261,
		0.016918,
		0.042284,
		0.077145,
		0.119997,
		0.169394,
		0.223957,
		0.282369,
		0.343382,
		0.405822,
		0.468589,
		0.530661,
		0.591103,
		0.649069,
		0.703808,
		0.754672,
		0.801123,
		0.842738,
		0.879226,
		0.910426,
		0.936326,
		0.957064,
		0.972936,
		0.984397,
		0.99205,
		0.996625,
		0.998934,
		0.99981,
		0.999993,
		1,
		0.999948,
		0.999501,
		0.99801,
		0.994673,
		0.988657,
		0.979184,
		0.965583,
		0.947325,
		0.924033,
		0.895491,
		0.861636,
		0.822556,
		0.778478,
		0.729761,
		0.676885,
		0.620445,
		0.561141,
		0.499772,
		0.43723,
		0.374495,
		0.312626,
		0.252761,
		0.196114,
		0.143965,
		0.097664,
		0.058623,
		0.028318,
		0.008282,
		0.000107,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.005149,
		0.022654,
		0.050979,
		0.088536,
		0.133799,
		0.185305,
		0.241656,
		0.301522,
		0.363647,
		0.42685,
		0.490028,
		0.552163,
		0.612327,
		0.669685,
		0.723506,
		0.773163,
		0.818151,
		0.858088,
		0.892725,
		0.921961,
		0.945845,
		0.964588,
		0.978564,
		0.98831,
		0.994505,
		0.997946,
		0.999485,
		0.999947,
		1,
		0.999993,
		0.999803,
		0.9989,
		0.996517,
		0.991804,
		0.983926,
		0.972142,
		0.955839,
		0.934562,
		0.908019,
		0.876079,
		0.83877,
		0.796268,
		0.748889,
		0.697079,
		0.641405,
		0.582546,
		0.521286,
		0.458508,
		0.395185,
		0.332378,
		0.27123,
		0.212959,
		0.158861,
		0.110298,
		0.0687,
		0.035564,
		0.012447,
		0.000966,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.00011,
		0.008564,
		0.029261,
		0.060526,
		0.100747,
		0.148377,
		0.201932,
		0.259999,
		0.321236,
		0.384378,
		0.448238,
		0.511714,
		0.573791,
		0.63355,
		0.690172,
		0.742944,
		0.79127,
		0.834679,
		0.872831,
		0.905532,
		0.932739,
		0.954571,
		0.971319,
		0.983438,
		0.991548,
		0.996405,
		0.998863,
		0.999797,
		0.999993,
		1,
		0.999945,
		0.999468,
		0.997879,
		0.994331,
		0.987949,
		0.977922,
		0.963556,
		0.944313,
		0.919818,
		0.889869,
		0.854428,
		0.813614,
		0.767695,
		0.717079,
		0.6623,
		0.604014,
		0.542987,
		0.48009,
		0.416289,
		0.352645,
		0.290302,
		0.230487,
		0.174504,
		0.12373,
		0.079615,
		0.043674,
		0.017488,
		0.0027,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.000999,
		0.012873,
		0.036751,
		0.070933,
		0.113781,
		0.163726,
		0.219266,
		0.278972,
		0.34149,
		0.405548,
		0.469955,
		0.533608,
		0.595501,
		0.654724,
		0.710476,
		0.762069,
		0.808939,
		0.850653,
		0.88692,
		0.917603,
		0.942727,
		0.962487,
		0.977255,
		0.987575,
		0.99415,
		0.997809,
		0.99945,
		0.999943,
		1,
		0.999992,
		0.99979,
		0.998826,
		0.996289,
		0.991281,
		0.982931,
		0.970465,
		0.953258,
		0.930852,
		0.902962,
		0.869479,
		0.830462,
		0.786125,
		0.736831,
		0.68308,
		0.625499,
		0.564834,
		0.501941,
		0.437779,
		0.373403,
		0.309961,
		0.248685,
		0.190888,
		0.137962,
		0.091374,
		0.052658,
		0.023419,
		0.005327,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.002795,
		0.018088,
		0.045134,
		0.082205,
		0.127639,
		0.179843,
		0.237298,
		0.298559,
		0.362263,
		0.427129,
		0.491965,
		0.555671,
		0.617247,
		0.675798,
		0.730545,
		0.780826,
		0.826114,
		0.866019,
		0.900305,
		0.928899,
		0.951898,
		0.96958,
		0.982404,
		0.991005,
		0.996168,
		0.998786,
		0.999782,
		0.999992,
		1,
		0.999941,
		0.999431,
		0.997737,
		0.993962,
		0.987186,
		0.976563,
		0.961379,
		0.941085,
		0.915312,
		0.883873,
		0.846758,
		0.804122,
		0.75628,
		0.703692,
		0.646952,
		0.586783,
		0.524024,
		0.459622,
		0.394628,
		0.330188,
		0.26754,
		0.208006,
		0.152992,
		0.103978,
		0.062522,
		0.030251,
		0.008861,
		0.000114,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.005514,
		0.024223,
		0.054421,
		0.094348,
		0.14232,
		0.196721,
		0.256014,
		0.318742,
		0.383529,
		0.449089,
		0.51423,
		0.577857,
		0.638979,
		0.69672,
		0.750323,
		0.799158,
		0.842738,
		0.880725,
		0.912942,
		0.939384,
		0.96023,
		0.975845,
		0.986782,
		0.993766,
		0.997662,
		0.999412,
		0.999939,
		1,
		0.999992,
		0.999775,
		0.998745,
		0.996042,
		0.990717,
		0.981857,
		0.968661,
		0.950488,
		0.926878,
		0.897559,
		0.862446,
		0.82163,
		0.77537,
		0.724081,
		0.668323,
		0.608789,
		0.546297,
		0.481783,
		0.416289,
		0.35096,
		0.287037,
		0.225849,
		0.168815,
		0.117431,
		0.073275,
		0.037997,
		0.01332,
		0.001035,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.000118,
		0.009174,
		0.031293,
		0.064618,
		0.107365,
		0.157821,
		0.214352,
		0.275401,
		0.339498,
		0.405259,
		0.471394,
		0.536711,
		0.60012,
		0.660647,
		0.717434,
		0.769751,
		0.817006,
		0.858756,
		0.89472,
		0.924786,
		0.949027,
		0.967708,
		0.981289,
		0.990418,
		0.995911,
		0.998703,
		0.999767,
		0.999991,
		1,
		0.999937,
		0.999391,
		0.997583,
		0.993563,
		0.986362,
		0.9751,
		0.959038,
		0.937623,
		0.91049,
		0.877473,
		0.838591,
		0.794042,
		0.744191,
		0.689557,
		0.630801,
		0.568718,
		0.504225,
		0.438357,
		0.372254,
		0.307157,
		0.244406,
		0.185427,
		0.131734,
		0.084923,
		0.046668,
		0.018718,
		0.002895,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.001072,
		0.013791,
		0.039308,
		0.075735,
		0.121258,
		0.17414,
		0.232725,
		0.295441,
		0.360804,
		0.427424,
		0.494005,
		0.559361,
		0.622411,
		0.682196,
		0.737881,
		0.788769,
		0.834309,
		0.874111,
		0.907953,
		0.935798,
		0.957802,
		0.974325,
		0.985926,
		0.99335,
		0.997501,
		0.99937,
		0.999934,
		1,
		0.999991,
		0.999759,
		0.998658,
		0.995774,
		0.990107,
		0.980698,
		0.966719,
		0.947512,
		0.92262,
		0.891784,
		0.854945,
		0.812236,
		0.763962,
		0.710598,
		0.652767,
		0.591237,
		0.526907,
		0.460796,
		0.39404,
		0.327881,
		0.263662,
		0.20282,
		0.146884,
		0.09747,
		0.056273,
		0.02507,
		0.005711,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.003,
		0.019382,
		0.048281,
		0.087777,
		0.136026,
		0.191269,
		0.251827,
		0.316114,
		0.382633,
		0.449987,
		0.516882,
		0.582134,
		0.644676,
		0.703566,
		0.758,
		0.807315,
		0.851009,
		0.888746,
		0.920376,
		0.945942,
		0.965692,
		0.980084,
		0.989783,
		0.995632,
		0.998612,
		0.99975,
		0.999991,
		1,
		0.999932,
		0.999348,
		0.997416,
		0.99313,
		0.985471,
		0.973521,
		0.95652,
		0.933906,
		0.905327,
		0.870636,
		0.82989,
		0.783333,
		0.731385,
		0.674631,
		0.613805,
		0.549783,
		0.483569,
		0.416289,
		0.349184,
		0.283601,
		0.220985,
		0.16288,
		0.110919,
		0.066821,
		0.032389,
		0.009504,
		0.000123,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.00592,
		0.025962,
		0.058222,
		0.100747,
		0.151669,
		0.2092,
		0.271644,
		0.337397,
		0.404954,
		0.472913,
		0.539979,
		0.604978,
		0.666857,
		0.724699,
		0.777728,
		0.825327,
		0.867043,
		0.902608,
		0.931946,
		0.955189,
		0.972685,
		0.984999,
		0.992899,
		0.997327,
		0.999325,
		0.999929,
		1,
		0.99999,
		0.999741,
		0.998563,
		0.995483,
		0.989445,
		0.979446,
		0.964625,
		0.944313,
		0.918052,
		0.885604,
		0.846941,
		0.802239,
		0.751857,
		0.696335,
		0.636368,
		0.572806,
		0.506634,
		0.438967,
		0.371041,
		0.304203,
		0.23991,
		0.179716,
		0.125272,
		0.07832,
		0.040688,
		0.014288,
		0.001112,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.000127,
		0.009851,
		0.033544,
		0.069138,
		0.114651,
		0.168182,
		0.227922,
		0.292156,
		0.359265,
		0.427734,
		0.496159,
		0.563249,
		0.627839,
		0.688896,
		0.745529,
		0.797,
		0.842738,
		0.882352,
		0.915644,
		0.942622,
		0.963517,
		0.978782,
		0.989094,
		0.995329,
		0.998512,
		0.999732,
		0.99999,
		1,
		0.999927,
		0.999301,
		0.997234,
		0.99266,
		0.984507,
		0.971816,
		0.953807,
		0.929913,
		0.899793,
		0.863328,
		0.820615,
		0.771949,
		0.717815,
		0.658868,
		0.595924,
		0.529947,
		0.462036,
		0.39342,
		0.325448,
		0.259581,
		0.197385,
		0.140529,
		0.090775,
		0.049979,
		0.020082,
		0.003111,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.001154,
		0.014812,
		0.042142,
		0.081039,
		0.129487,
		0.185559,
		0.247423,
		0.313342,
		0.381687,
		0.450936,
		0.51968,
		0.586639,
		0.650658,
		0.710728,
		0.76599,
		0.815748,
		0.859486,
		0.896877,
		0.927805,
		0.952372,
		0.970913,
		0.983995,
		0.99241,
		0.997137,
		0.999276,
		0.999924,
		1,
		0.99999,
		0.999722,
		0.99846,
		0.995167,
		0.988728,
		0.978092,
		0.962366,
		0.940869,
		0.913148,
		0.878987,
		0.838394,
		0.791594,
		0.739008,
		0.681244,
		0.619082,
		0.553458,
		0.485454,
		0.416289,
		0.34731,
		0.27998,
		0.215878,
		0.156687,
		0.104192,
		0.060273,
		0.026902,
		0.00614,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.003228,
		0.020821,
		0.051768,
		0.09393,
		0.145256,
		0.203793,
		0.267685,
		0.335179,
		0.404632,
		0.474517,
		0.543429,
		0.610092,
		0.673374,
		0.732288,
		0.786014,
		0.833904,
		0.875503,
		0.910561,
		0.939048,
		0.961169,
		0.977373,
		0.988347,
		0.994999,
		0.998404,
		0.999712,
		0.999989,
		1,
		0.999921,
		0.999249,
		0.997036,
		0.992149,
		0.983462,
		0.969973,
		0.950882,
		0.925619,
		0.893857,
		0.855511,
		0.810721,
		0.759844,
		0.703432,
		0.642221,
		0.577116,
		0.509177,
		0.439612,
		0.369759,
		0.301087,
		0.235182,
		0.173742,
		0.118572,
		0.071577,
		0.034761,
		0.010218,
		0.000132,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.006372,
		0.027894,
		0.062433,
		0.107814,
		0.161955,
		0.222873,
		0.28869,
		0.357638,
		0.428063,
		0.498434,
		0.567351,
		0.63355,
		0.695919,
		0.753505,
		0.805528,
		0.851398,
		0.890728,
		0.92335,
		0.949334,
		0.968996,
		0.982906,
		0.991877,
		0.99693,
		0.999222,
		0.999918,
		1,
		0.999989,
		0.999701,
		0.998347,
		0.994823,
		0.987949,
		0.976625,
		0.959925,
		0.937158,
		0.907879,
		0.871897,
		0.829262,
		0.780253,
		0.725363,
		0.665277,
		0.600863,
		0.533157,
		0.463346,
		0.392764,
		0.322879,
		0.255282,
		0.191686,
		0.133916,
		0.083901,
		0.043674,
		0.015366,
		0.001198,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.000137,
		0.010606,
		0.036046,
		0.074147,
		0.122695,
		0.179577,
		0.242783,
		0.310415,
		0.380687,
		0.451939,
		0.522638,
		0.591389,
		0.656948,
		0.718226,
		0.774307,
		0.824461,
		0.868161,
		0.905097,
		0.935195,
		0.958631,
		0.975845,
		0.987535,
		0.99464,
		0.998286,
		0.99969,
		0.999988,
		1,
		0.999915,
		0.999193,
		0.99682,
		0.991594,
		0.982327,
		0.967979,
		0.947725,
		0.920996,
		0.887485,
		0.847143,
		0.800163,
		0.746966,
		0.688183,
		0.62464,
		0.557338,
		0.487448,
		0.416289,
		0.345327,
		0.27616,
		0.210511,
		0.150223,
		0.09725,
		0.053654,
		0.021601,
		0.003352,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.001245,
		0.015951,
		0.045292,
		0.086916,
		0.138573,
		0.198115,
		0.263508,
		0.332833,
		0.404291,
		0.476215,
		0.547074,
		0.615483,
		0.680218,
		0.74022,
		0.794618,
		0.842738,
		0.884124,
		0.918552,
		0.946052,
		0.96692,
		0.981724,
		0.991297,
		0.996704,
		0.999163,
		0.999912,
		1,
		0.999988,
		0.999678,
		0.998223,
		0.994448,
		0.987103,
		0.975034,
		0.957286,
		0.933156,
		0.902212,
		0.864293,
		0.819497,
		0.768167,
		0.710869,
		0.64838,
		0.581665,
		0.511868,
		0.440294,
		0.368403,
		0.297795,
		0.230203,
		0.167488,
		0.111627,
		0.064712,
		0.028941,
		0.006618,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.003484,
		0.022426,
		0.055645,
		0.100747,
		0.155444,
		0.217558,
		0.285028,
		0.355915,
		0.428411,
		0.500843,
		0.571684,
		0.639566,
		0.703287,
		0.761826,
		0.814362,
		0.860284,
		0.899218,
		0.931037,
		0.955885,
		0.974188,
		0.986651,
		0.994248,
		0.998157,
		0.999666,
		0.999988,
		1,
		0.999909,
		0.999131,
		0.996584,
		0.990988,
		0.981094,
		0.965817,
		0.944313,
		0.916014,
		0.880639,
		0.838179,
		0.788888,
		0.73326,
		0.672015,
		0.606074,
		0.536552,
		0.464734,
		0.39207,
		0.32016,
		0.250747,
		0.185705,
		0.127035,
		0.076857,
		0.037403,
		0.011016,
		0.000143,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.006878,
		0.030049,
		0.067117,
		0.115644,
		0.173305,
		0.237891,
		0.307317,
		0.379627,
		0.453002,
		0.525768,
		0.596407,
		0.663568,
		0.726079,
		0.782964,
		0.833458,
		0.877025,
		0.913379,
		0.942504,
		0.964669,
		0.980438,
		0.990665,
		0.996457,
		0.999098,
		0.999905,
		1,
		0.999987,
		0.999653,
		0.998088,
		0.994038,
		0.986181,
		0.973306,
		0.954428,
		0.928835,
		0.896111,
		0.856131,
		0.809049,
		0.755278,
		0.69547,
		0.630499,
		0.561441,
		0.489559,
		0.416289,
		0.343228,
		0.272124,
		0.204865,
		0.143473,
		0.090096,
		0.047001,
		0.01657,
		0.001294,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.000148,
		0.011451,
		0.038838,
		0.079717,
		0.131607,
		0.192146,
		0.259096,
		0.330348,
		0.40393,
		0.478015,
		0.550931,
		0.621173,
		0.687413,
		0.748515,
		0.803552,
		0.851827,
		0.892886,
		0.926546,
		0.95291,
		0.972387,
		0.985689,
		0.993819,
		0.998016,
		0.999639,
		0.999987,
		1,
		0.999901,
		0.999064,
		0.996325,
		0.990328,
		0.979753,
		0.963472,
		0.940621,
		0.910641,
		0.873276,
		0.828569,
		0.776841,
		0.718669,
		0.654869,
		0.586474,
		0.514717,
		0.441017,
		0.366965,
		0.294311,
		0.224955,
		0.160938,
		0.104436,
		0.057749,
		0.023299,
		0.003623,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.001347,
		0.017226,
		0.048808,
		0.093452,
		0.148635,
		0.211958,
		0.281153,
		0.354089,
		0.42878,
		0.503396,
		0.576269,
		0.645911,
		0.711022,
		0.770509,
		0.823506,
		0.869387,
		0.907796,
		0.938663,
		0.962225,
		0.979038,
		0.989975,
		0.996187,
		0.999027,
		0.999897,
		1,
		0.999986,
		0.999625,
		0.99794,
		0.99359,
		0.985175,
		0.971428,
		0.95133,
		0.924165,
		0.889537,
		0.847365,
		0.797864,
		0.741528,
		0.679106,
		0.61158,
		0.540147,
		0.466205,
		0.391333,
		0.317281,
		0.245958,
		0.179424,
		0.119879,
		0.069658,
		0.031222,
		0.007154,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.003771,
		0.024223,
		0.059973,
		0.108329,
		0.166724,
		0.232725,
		0.304035,
		0.378502,
		0.45413,
		0.529086,
		0.601712,
		0.670542,
		0.734311,
		0.791976,
		0.842738,
		0.886059,
		0.921688,
		0.949683,
		0.970427,
		0.984638,
		0.99335,
		0.99786,
		0.99961,
		0.999985,
		1,
		0.999893,
		0.998989,
		0.996042,
		0.989606,
		0.978292,
		0.960924,
		0.936624,
		0.904839,
		0.865352,
		0.81826,
		0.763962,
		0.70313,
		0.636686,
		0.565786,
		0.4918,
		0.416289,
		0.341002,
		0.267854,
		0.198921,
		0.136427,
		0.082738,
		0.040357,
		0.011912,
		0.000155,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.007447,
		0.032464,
		0.072345,
		0.124349,
		0.185867,
		0.254429,
		0.327712,
		0.403546,
		0.479925,
		0.55502,
		0.627186,
		0.694984,
		0.757191,
		0.812825,
		0.861163,
		0.901765,
		0.934501,
		0.959568,
		0.977512,
		0.98922,
		0.99589,
		0.998949,
		0.999889,
		1,
		0.999985,
		0.999595,
		0.997777,
		0.993099,
		0.984077,
		0.969383,
		0.947967,
		0.919112,
		0.882447,
		0.837941,
		0.785882,
		0.726854,
		0.661713,
		0.591564,
		0.51774,
		0.441786,
		0.365439,
		0.290619,
		0.219416,
		0.154076,
		0.096998,
		0.050721,
		0.017921,
		0.001403,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000161,
		0.012401,
		0.041967,
		0.085934,
		0.141513,
		0.206051,
		0.277047,
		0.352149,
		0.429172,
		0.506107,
		0.581127,
		0.65261,
		0.719149,
		0.779572,
		0.832964,
		0.878693,
		0.91643,
		0.946178,
		0.968292,
		0.98349,
		0.992836,
		0.997689,
		0.999578,
		0.999984,
		1,
		0.999884,
		0.998907,
		0.995731,
		0.988816,
		0.976697,
		0.958153,
		0.932289,
		0.898568,
		0.856815,
		0.807193,
		0.750188,
		0.686577,
		0.617404,
		0.543963,
		0.467769,
		0.39055,
		0.314225,
		0.240892,
		0.172822,
		0.112441,
		0.062328,
		0.025204,
		0.003928,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.001462,
		0.01866,
		0.052748,
		0.100747,
		0.159817,
		0.227264,
		0.300551,
		0.377306,
		0.455329,
		0.532609,
		0.607332,
		0.677898,
		0.742944,
		0.801356,
		0.8523,
		0.895243,
		0.929985,
		0.956676,
		0.975845,
		0.988393,
		0.995564,
		0.998863,
		0.99988,
		1,
		0.999984,
		0.999561,
		0.997598,
		0.992561,
		0.982876,
		0.967154,
		0.944313,
		0.913638,
		0.874792,
		0.827802,
		0.773038,
		0.711187,
		0.643226,
		0.570395,
		0.49418,
		0.416289,
		0.338636,
		0.263329,
		0.192655,
		0.12907,
		0.07519,
		0.033782,
		0.007758,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.004095,
		0.026246,
		0.064824,
		0.11679,
		0.179254,
		0.249484,
		0.324909,
		0.403137,
		0.481957,
		0.559361,
		0.63355,
		0.702959,
		0.766271,
		0.822445,
		0.870737,
		0.910731,
		0.942367,
		0.965963,
		0.982234,
		0.992272,
		0.997501,
		0.999543,
		0.999983,
		1,
		0.999874,
		0.998817,
		0.995389,
		0.987949,
		0.974954,
		0.955135,
		0.927583,
		0.891784,
		0.84761,
		0.795304,
		0.735449,
		0.66894,
		0.596961,
		0.520954,
		0.442603,
		0.363815,
		0.286701,
		0.213563,
		0.146884,
		0.089317,
		0.043674,
		0.012921,
		0.000168,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.008089,
		0.035182,
		0.078205,
		0.134062,
		0.199814,
		0.272688,
		0.350084,
		0.42959,
		0.508991,
		0.586284,
		0.659694,
		0.727694,
		0.78903,
		0.842738,
		0.888183,
		0.925079,
		0.953525,
		0.974023,
		0.987485,
		0.995205,
		0.998768,
		0.999869,
		1,
		0.999982,
		0.999524,
		0.9974,
		0.991969,
		0.981561,
		0.96472,
		0.940336,
		0.907703,
		0.86652,
		0.816885,
		0.759263,
		0.694455,
		0.623575,
		0.548017,
		0.469434,
		0.389717,
		0.310976,
		0.235528,
		0.165879,
		0.104717,
		0.054899,
		0.019444,
		0.001526,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000176,
		0.013475,
		0.045488,
		0.092903,
		0.152563,
		0.221482,
		0.296847,
		0.376031,
		0.456608,
		0.536358,
		0.613292,
		0.685666,
		0.752002,
		0.811114,
		0.862134,
		0.904547,
		0.938217,
		0.963419,
		0.980856,
		0.991651,
		0.997294,
		0.999504,
		0.999981,
		1,
		0.999863,
		0.998717,
		0.995012,
		0.986998,
		0.973047,
		0.951842,
		0.922467,
		0.884434,
		0.837676,
		0.782523,
		0.71967,
		0.650148,
		0.575292,
		0.496716,
		0.416289,
		0.336118,
		0.258527,
		0.186045,
		0.121393,
		0.067471,
		0.027353,
		0.004273,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.001593,
		0.02028,
		0.057184,
		0.108924,
		0.172287,
		0.244237,
		0.321925,
		0.402701,
		0.484124,
		0.563978,
		0.640295,
		0.711367,
		0.775775,
		0.832415,
		0.880531,
		0.919743,
		0.950084,
		0.972026,
		0.986488,
		0.994809,
		0.998663,
		0.999858,
		1,
		0.999981,
		0.999482,
		0.997182,
		0.991318,
		0.980117,
		0.962058,
		0.936004,
		0.901258,
		0.857572,
		0.805122,
		0.74448,
		0.676582,
		0.602692,
		0.524376,
		0.443475,
		0.362085,
		0.282534,
		0.207371,
		0.139346,
		0.081404,
		0.03667,
		0.008442,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.004462,
		0.028532,
		0.070284,
		0.126271,
		0.193222,
		0.268054,
		0.347883,
		0.430036,
		0.512066,
		0.591766,
		0.667192,
		0.736686,
		0.798899,
		0.852824,
		0.897829,
		0.933692,
		0.960634,
		0.979343,
		0.990967,
		0.997064,
		0.99946,
		0.99998,
		1,
		0.999851,
		0.998606,
		0.994596,
		0.985952,
		0.970957,
		0.948246,
		0.9169,
		0.876466,
		0.826947,
		0.768774,
		0.702773,
		0.630124,
		0.552335,
		0.47121,
		0.388827,
		0.307516,
		0.229837,
		0.158573,
		0.096707,
		0.047417,
		0.014064,
		0.000183,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.008818,
		0.038254,
		0.084802,
		0.144944,
		0.215353,
		0.292901,
		0.374671,
		0.457972,
		0.540354,
		0.619625,
		0.693878,
		0.761512,
		0.82126,
		0.87223,
		0.913932,
		0.946324,
		0.969836,
		0.985389,
		0.994372,
		0.998546,
		0.999845,
		1,
		0.999979,
		0.999437,
		0.99694,
		0.990599,
		0.978531,
		0.959144,
		0.931276,
		0.894253,
		0.847883,
		0.792437,
		0.72861,
		0.657487,
		0.580504,
		0.499421,
		0.416289,
		0.333432,
		0.253422,
		0.179066,
		0.113386,
		0.059613,
		0.02117,
		0.001665,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000192,
		0.014693,
		0.049471,
		0.100747,
		0.164939,
		0.238661,
		0.318742,
		0.402235,
		0.486437,
		0.568899,
		0.647454,
		0.720239,
		0.785723,
		0.842738,
		0.890522,
		0.928752,
		0.957583,
		0.977679,
		0.990212,
		0.99681,
		0.999412,
		0.999978,
		1,
		0.999838,
		0.998483,
		0.994136,
		0.984799,
		0.968661,
		0.944313,
		0.910833,
		0.867815,
		0.815347,
		0.753975,
		0.684672,
		0.608789,
		0.528027,
		0.444407,
		0.360237,
		0.278096,
		0.200812,
		0.131445,
		0.073275,
		0.029788,
		0.004665,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.001742,
		0.022121,
		0.0622,
		0.118126,
		0.186245,
		0.263117,
		0.345531,
		0.430513,
		0.51535,
		0.597605,
		0.675142,
		0.746155,
		0.809195,
		0.863213,
		0.907596,
		0.942207,
		0.967429,
		0.984178,
		0.993888,
		0.998417,
		0.99983,
		1,
		0.999977,
		0.999385,
		0.996672,
		0.989805,
		0.976785,
		0.955947,
		0.926112,
		0.886629,
		0.837381,
		0.778747,
		0.711563,
		0.637084,
		0.556941,
		0.473108,
		0.387877,
		0.303823,
		0.223793,
		0.150882,
		0.088416,
		0.039943,
		0.00922,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.004882,
		0.031129,
		0.076461,
		0.13694,
		0.208847,
		0.28869,
		0.373216,
		0.459432,
		0.544621,
		0.626365,
		0.70257,
		0.771497,
		0.831801,
		0.882566,
		0.92335,
		0.954233,
		0.975845,
		0.989376,
		0.996527,
		0.999358,
		0.999976,
		1,
		0.999823,
		0.998347,
		0.993626,
		0.983525,
		0.966137,
		0.940003,
		0.904213,
		0.858415,
		0.802795,
		0.738037,
		0.665277,
		0.586062,
		0.502313,
		0.416289,
		0.330561,
		0.247986,
		0.171689,
		0.105044,
		0.05166,
		0.015366,
		0.000201,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.00965,
		0.041747,
		0.092264,
		0.157186,
		0.232725,
		0.315337,
		0.401737,
		0.488913,
		0.574153,
		0.655065,
		0.72961,
		0.796135,
		0.85341,
		0.900677,
		0.937694,
		0.96478,
		0.982839,
		0.99335,
		0.998273,
		0.999814,
		1,
		0.999974,
		0.999328,
		0.996374,
		0.988925,
		0.974858,
		0.952436,
		0.920461,
		0.878322,
		0.825987,
		0.763962,
		0.693246,
		0.615286,
		0.531932,
		0.445405,
		0.358259,
		0.273359,
		0.193856,
		0.123166,
		0.064959,
		0.023136,
		0.001825,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000211,
		0.016085,
		0.053998,
		0.10962,
		0.178855,
		0.25785,
		0.343012,
		0.431024,
		0.518865,
		0.603835,
		0.683581,
		0.756129,
		0.819929,
		0.87389,
		0.917436,
		0.950551,
		0.973821,
		0.98845,
		0.996213,
		0.999297,
		0.999973,
		1,
		0.999806,
		0.998194,
		0.99306,
		0.982117,
		0.963355,
		0.935275,
		0.89698,
		0.848188,
		0.789202,
		0.720862,
		0.644493,
		0.561865,
		0.475143,
		0.386858,
		0.299874,
		0.217363,
		0.142782,
		0.079857,
		0.032563,
		0.005114,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.001913,
		0.024223,
		0.067904,
		0.128533,
		0.201932,
		0.284186,
		0.371655,
		0.460998,
		0.549189,
		0.63355,
		0.711778,
		0.781985,
		0.842738,
		0.893112,
		0.932739,
		0.961859,
		0.981357,
		0.992753,
		0.998111,
		0.999797,
		1,
		0.999972,
		0.999265,
		0.996042,
		0.987949,
		0.97273,
		0.948572,
		0.914268,
		0.869259,
		0.813614,
		0.747985,
		0.673559,
		0.592002,
		0.505413,
		0.416289,
		0.327485,
		0.242186,
		0.163887,
		0.096367,
		0.043674,
		0.010111,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.005363,
		0.034098,
		0.083483,
		0.149001,
		0.226396,
		0.311689,
		0.401201,
		0.491569,
		0.579774,
		0.663169,
		0.739516,
		0.807029,
		0.864419,
		0.91095,
		0.946495,
		0.971582,
		0.987421,
		0.995862,
		0.99923,
		0.999971,
		1,
		0.999787,
		0.998024,
		0.992429,
		0.980556,
		0.960286,
		0.930079,
		0.889065,
		0.837049,
		0.77447,
		0.702346,
		0.622222,
		0.536117,
		0.446476,
		0.356136,
		0.268292,
		0.186469,
		0.114498,
		0.056497,
		0.016857,
		0.000221,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.010606,
		0.045739,
		0.100747,
		0.171018,
		0.252217,
		0.340308,
		0.431573,
		0.522638,
		0.610497,
		0.692551,
		0.766643,
		0.831109,
		0.884829,
		0.927288,
		0.958631,
		0.979712,
		0.992087,
		0.997931,
		0.999776,
		1,
		0.999969,
		0.999193,
		0.995671,
		0.986864,
		0.970373,
		0.944313,
		0.907473,
		0.859359,
		0.800163,
		0.730708,
		0.652395,
		0.56714,
		0.477327,
		0.385763,
		0.295641,
		0.210511,
		0.134253,
		0.071053,
		0.025389,
		0.002008,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.000232,
		0.017685,
		0.059174,
		0.119705,
		0.19457,
		0.279358,
		0.369978,
		0.462682,
		0.55409,
		0.641225,
		0.721547,
		0.793,
		0.854068,
		0.903828,
		0.94202,
		0.9691,
		0.986276,
		0.99547,
		0.999154,
		0.999968,
		1,
		0.999765,
		0.997833,
		0.991726,
		0.978822,
		0.956891,
		0.924359,
		0.880394,
		0.824905,
		0.75849,
		0.682378,
		0.598361,
		0.508744,
		0.416289,
		0.324181,
		0.235987,
		0.155629,
		0.08736,
		0.035743,
		0.005631,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.002111,
		0.02664,
		0.074425,
		0.140358,
		0.219635,
		0.30777,
		0.400625,
		0.494427,
		0.585802,
		0.671812,
		0.749996,
		0.818422,
		0.875748,
		0.921283,
		0.95506,
		0.977883,
		0.991344,
		0.997729,
		0.999754,
		1,
		0.999966,
		0.999113,
		0.995256,
		0.985655,
		0.967759,
		0.939609,
		0.900005,
		0.848532,
		0.785527,
		0.712016,
		0.629643,
		0.540612,
		0.447628,
		0.353853,
		0.262862,
		0.178617,
		0.105431,
		0.047952,
		0.011138,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.00592,
		0.037511,
		0.091512,
		0.162702,
		0.246181,
		0.337397,
		0.432165,
		0.526696,
		0.617635,
		0.7021,
		0.777728,
		0.842738,
		0.895994,
		0.937074,
		0.966345,
		0.984999,
		0.99503,
		0.999069,
		0.999964,
		1,
		0.999741,
		0.997618,
		0.990939,
		0.976892,
		0.953131,
		0.918052,
		0.870879,
		0.811648,
		0.741144,
		0.660837,
		0.572806,
		0.47968,
		0.384583,
		0.291093,
		0.203198,
		0.125272,
		0.062044,
		0.018575,
		0.000244,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.011711,
		0.05033,
		0.110446,
		0.186723,
		0.27417,
		0.368169,
		0.464497,
		0.559361,
		0.649438,
		0.731919,
		0.804566,
		0.865776,
		0.914656,
		0.951098,
		0.975845,
		0.990511,
		0.997501,
		0.999728,
		1,
		0.999962,
		0.999022,
		0.99479,
		0.984305,
		0.964853,
		0.934406,
		0.891784,
		0.836674,
		0.769588,
		0.691782,
		0.605186,
		0.512331,
		0.416289,
		0.320624,
		0.229348,
		0.146884,
		0.078042,
		0.027987,
		0.002221,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.000257,
		0.019535,
		0.065127,
		0.131232,
		0.212399,
		0.30355,
		0.400003,
		0.497509,
		0.592281,
		0.681045,
		0.761088,
		0.830324,
		0.887363,
		0.931598,
		0.963278,
		0.98357,
		0.994536,
		0.998972,
		0.99996,
		0.999999,
		0.999714,
		0.997377,
		0.990058,
		0.974739,
		0.948956,
		0.911086,
		0.860425,
		0.797161,
		0.722303,
		0.637598,
		0.545454,
		0.448872,
		0.35139,
		0.257029,
		0.170259,
		0.095965,
		0.039412,
		0.006231,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.002341,
		0.029437,
		0.081927,
		0.153869,
		0.2397,
		0.334256,
		0.432804,
		0.531073,
		0.6253,
		0.712278,
		0.789417,
		0.854812,
		0.907331,
		0.946697,
		0.973569,
		0.989577,
		0.997245,
		0.999699,
		0.999999,
		0.999958,
		0.99892,
		0.994266,
		0.982792,
		0.961616,
		0.928639,
		0.882719,
		0.823672,
		0.752213,
		0.669872,
		0.578906,
		0.482221,
		0.383309,
		0.286194,
		0.195381,
		0.115823,
		0.052888,
		0.012329,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.006567,
		0.041461,
		0.100747,
		0.178346,
		0.268583,
		0.366214,
		0.466459,
		0.565045,
		0.658245,
		0.742944,
		0.816702,
		0.877839,
		0.925522,
		0.959858,
		0.981968,
		0.99398,
		0.998863,
		0.999956,
		0.999999,
		0.999683,
		0.997104,
		0.989067,
		0.972331,
		0.944313,
		0.903379,
		0.848922,
		0.781314,
		0.701826,
		0.612528,
		0.516206,
		0.416289,
		0.316784,
		0.222223,
		0.137622,
		0.068445,
		0.02057,
		0.000272,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.012998,
		0.055645,
		0.121599,
		0.204642,
		0.298992,
		0.39933,
		0.500843,
		0.599263,
		0.690927,
		0.772836,
		0.842738,
		0.899218,
		0.941795,
		0.971021,
		0.988525,
		0.996954,
		0.999666,
		0.999999,
		0.999954,
		0.998803,
		0.993675,
		0.981094,
		0.958001,
		0.922235,
		0.872709,
		0.809397,
		0.73326,
		0.646144,
		0.550683,
		0.450219,
		0.348725,
		0.250747,
		0.161357,
		0.086107,
		0.031004,
		0.00247,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.000287,
		0.02169,
		0.072021,
		0.144484,
		0.232725,
		0.330857,
		0.433497,
		0.535808,
		0.63355,
		0.723143,
		0.801739,
		0.867313,
		0.918768,
		0.956035,
		0.980168,
		0.99335,
		0.998739,
		0.999951,
		0.999999,
		0.999648,
		0.996795,
		0.987949,
		0.969633,
		0.939136,
		0.894835,
		0.836245,
		0.763962,
		0.679561,
		0.58549,
		0.484973,
		0.381929,
		0.280903,
		0.187012,
		0.105894,
		0.043674,
		0.006932,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.00261,
		0.032698,
		0.090614,
		0.169394,
		0.262549,
		0.364093,
		0.468589,
		0.571191,
		0.667708,
		0.754672,
		0.829424,
		0.890217,
		0.936326,
		0.968161,
		0.987338,
		0.996625,
		0.999628,
		0.999999,
		0.999948,
		0.998671,
		0.993005,
		0.979184,
		0.953955,
		0.915109,
		0.861636,
		0.793707,
		0.71257,
		0.620445,
		0.520404,
		0.416289,
		0.312626,
		0.214561,
		0.127813,
		0.058623,
		0.013723,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.007327,
		0.046069,
		0.111441,
		0.19631,
		0.294054,
		0.398598,
		0.50446,
		0.606807,
		0.701521,
		0.785281,
		0.855661,
		0.911245,
		0.951752,
		0.978139,
		0.992637,
		0.998598,
		0.999945,
		0.999999,
		0.999607,
		0.996443,
		0.986687,
		0.9666,
		0.933353,
		0.885348,
		0.822257,
		0.744944,
		0.655346,
		0.556348,
		0.451682,
		0.345833,
		0.243966,
		0.151865,
		0.075884,
		0.022905,
		0.000304,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.014509,
		0.061846,
		0.13451,
		0.2252,
		0.327165,
		0.43425,
		0.540946,
		0.642452,
		0.734755,
		0.814721,
		0.880212,
		0.930208,
		0.964943,
		0.985994,
		0.996249,
		0.999585,
		0.999999,
		0.999942,
		0.99852,
		0.992245,
		0.977027,
		0.949417,
		0.907163,
		0.849368,
		0.776439,
		0.689972,
		0.592618,
		0.487964,
		0.380428,
		0.275172,
		0.178037,
		0.095481,
		0.034535,
		0.002763,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.000322,
		0.024223,
		0.080064,
		0.159817,
		0.256014,
		0.361785,
		0.470906,
		0.577857,
		0.677898,
		0.767158,
		0.842738,
		0.902848,
		0.94694,
		0.975845,
		0.991827,
		0.998436,
		0.999939,
		0.999999,
		0.999561,
		0.996042,
		0.985256,
		0.963183,
		0.926878,
		0.874792,
		0.806797,
		0.724081,
		0.629005,
		0.524968,
		0.416289,
		0.308108,
		0.206304,
		0.117431,
		0.048665,
		0.007758,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.002929,
		0.036531,
		0.100747,
		0.187344,
		0.28869,
		0.397801,
		0.5084,
		0.614981,
		0.712898,
		0.798462,
		0.86907,
		0.92335,
		0.961312,
		0.984468,
		0.99582,
		0.999535,
		0.999999,
		0.999935,
		0.998347,
		0.991379,
		0.974587,
		0.944313,
		0.898284,
		0.835752,
		0.757413,
		0.665277,
		0.562504,
		0.453277,
		0.342684,
		0.236626,
		0.14174,
		0.065341,
		0.015366,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.008227,
		0.051487,
		0.123912,
		0.217061,
		0.323142,
		0.435073,
		0.54654,
		0.652082,
		0.74718,
		0.828385,
		0.893455,
		0.941522,
		0.973246,
		0.990902,
		0.998251,
		0.999931,
		0.999999,
		0.999508,
		0.995583,
		0.983628,
		0.959323,
		0.91961,
		0.863025,
		0.789689,
		0.701178,
		0.600358,
		0.491227,
		0.37879,
		0.268945,
		0.168399,
		0.084596,
		0.025659,
		0.000342,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.0163,
		0.069138,
		0.149561,
		0.248916,
		0.359265,
		0.473438,
		0.585111,
		0.688896,
		0.780455,
		0.856637,
		0.915644,
		0.957205,
		0.98273,
		0.995329,
		0.999478,
		0.999999,
		0.999927,
		0.998148,
		0.99039,
		0.971816,
		0.938558,
		0.888342,
		0.820615,
		0.73643,
		0.638287,
		0.529947,
		0.416289,
		0.303183,
		0.197385,
		0.106458,
		0.038704,
		0.003111,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.000363,
		0.027226,
		0.089522,
		0.177679,
		0.282841,
		0.396928,
		0.512705,
		0.623865,
		0.725136,
		0.812415,
		0.882926,
		0.935405,
		0.970291,
		0.989843,
		0.998037,
		0.999922,
		0.999999,
		0.999446,
		0.995056,
		0.981771,
		0.954949,
		0.911434,
		0.849883,
		0.770731,
		0.676022,
		0.569217,
		0.455022,
		0.339242,
		0.228658,
		0.130939,
		0.05456,
		0.008741,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.00331,
		0.041077,
		0.112664,
		0.208236,
		0.318742,
		0.435974,
		0.552654,
		0.662527,
		0.760489,
		0.842738,
		0.906963,
		0.952544,
		0.980744,
		0.994763,
		0.999412,
		0.999999,
		0.999917,
		0.997917,
		0.989256,
		0.968661,
		0.932051,
		0.877185,
		0.803759,
		0.713267,
		0.608789,
		0.494799,
		0.376995,
		0.262156,
		0.158037,
		0.073275,
		0.017323,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.009304,
		0.057915,
		0.138572,
		0.241181,
		0.3565,
		0.476214,
		0.59303,
		0.70079,
		0.794617,
		0.871094,
		0.928478,
		0.966918,
		0.988624,
		0.997788,
		0.999911,
		0.999997,
		0.999373,
		0.994447,
		0.979644,
		0.949977,
		0.90221,
		0.835175,
		0.749695,
		0.648379,
		0.535398,
		0.416289,
		0.297794,
		0.187731,
		0.094888,
		0.028941,
		0.000387,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.018444,
		0.077791,
		0.167239,
		0.276435,
		0.395962,
		0.51742,
		0.633539,
		0.738309,
		0.827155,
		0.897143,
		0.947222,
		0.978449,
		0.994091,
		0.999317,
		0.999981,
		0.999888,
		0.997633,
		0.987932,
		0.965039,
		0.924657,
		0.864619,
		0.784947,
		0.687664,
		0.576554,
		0.456932,
		0.335457,
		0.21998,
		0.11942,
		0.043674,
		0.003529,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.000414,
		0.03082,
		0.100736,
		0.198621,
		0.313873,
		0.436917,
		0.559298,
		0.673812,
		0.774666,
		0.857677,
		0.920512,
		0.962953,
		0.98711,
		0.997388,
		0.999787,
		0.999885,
		0.999178,
		0.993631,
		0.977092,
		0.944206,
		0.891682,
		0.818594,
		0.726248,
		0.617935,
		0.498671,
		0.374977,
		0.254699,
		0.146868,
		0.061582,
		0.009922,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.003769,
		0.046503,
		0.126741,
		0.232614,
		0.353287,
		0.479047,
		0.601425,
		0.713346,
		0.80931,
		0.885637,
		0.940734,
		0.97538,
		0.992877,
		0.998768,
		0.999522,
		0.999416,
		0.996864,
		0.98597,
		0.960466,
		0.915849,
		0.85008,
		0.763598,
		0.659076,
		0.541136,
		0.416091,
		0.291735,
		0.177175,
		0.082699,
		0.019668,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.01059,
		0.065521,
		0.155717,
		0.268996,
		0.394307,
		0.52184,
		0.643164,
		0.751401,
		0.841453,
		0.910271,
		0.957169,
		0.984094,
		0.995644,
		0.998361,
		0.998473,
		0.997669,
		0.991411,
		0.9729,
		0.936403,
		0.878618,
		0.798942,
		0.699281,
		0.583747,
		0.458357,
		0.330792,
		0.21019,
		0.106998,
		0.032845,
		0.000442,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.020955,
		0.087814,
		0.187433,
		0.30734,
		0.436307,
		0.564478,
		0.68353,
		0.786881,
		0.869955,
		0.930499,
		0.968915,
		0.988486,
		0.995131,
		0.995988,
		0.995868,
		0.992978,
		0.980739,
		0.952336,
		0.903084,
		0.831152,
		0.737504,
		0.625599,
		0.501051,
		0.371341,
		0.245582,
		0.134336,
		0.049465,
		0.004022,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.000471,
		0.034862,
		0.113164,
		0.221416,
		0.346903,
		0.478279,
		0.605712,
		0.721083,
		0.818227,
		0.893244,
		0.944861,
		0.974769,
		0.987715,
		0.990785,
		0.990912,
		0.99,
		0.982956,
		0.962301,
		0.921945,
		0.858647,
		0.772234,
		0.66534,
		0.543038,
		0.412507,
		0.282748,
		0.164372,
		0.069425,
		0.011259,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.004255,
		0.052157,
		0.141106,
		0.256856,
		0.386524,
		0.51873,
		0.643772,
		0.753855,
		0.84338,
		0.909308,
		0.95155,
		0.973275,
		0.980718,
		0.981686,
		0.98155,
		0.978299,
		0.964651,
		0.93327,
		0.879373,
		0.801442,
		0.701088,
		0.582681,
		0.452958,
		0.320699,
		0.19647,
		0.092422,
		0.022135,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.011792,
		0.07245,
		0.170831,
		0.29251,
		0.424544,
		0.555639,
		0.676358,
		0.779382,
		0.859861,
		0.915816,
		0.948562,
		0.962873,
		0.966297,
		0.966441,
		0.96542,
		0.957597,
		0.934871,
		0.890903,
		0.822636,
		0.730447,
		0.617806,
		0.490845,
		0.357979,
		0.229604,
		0.11786,
		0.036449,
		0.000494,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.022848,
		0.095026,
		0.201119,
		0.326664,
		0.458811,
		0.58651,
		0.700751,
		0.794878,
		0.864989,
		0.910389,
		0.933976,
		0.942136,
		0.943206,
		0.943055,
		0.939475,
		0.924584,
		0.890684,
		0.83306,
		0.750642,
		0.645788,
		0.523831,
		0.392648,
		0.262302,
		0.144773,
		0.053735,
		0.0044,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.000502,
		0.036891,
		0.118789,
		0.23032,
		0.35718,
		0.486799,
		0.608558,
		0.714073,
		0.797554,
		0.856264,
		0.890993,
		0.906329,
		0.910034,
		0.910191,
		0.909081,
		0.900656,
		0.876427,
		0.830044,
		0.758816,
		0.663773,
		0.549241,
		0.422357,
		0.292662,
		0.171783,
		0.073179,
		0.011958,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.004361,
		0.053042,
		0.142261,
		0.256439,
		0.381656,
		0.505845,
		0.619035,
		0.713679,
		0.785067,
		0.83183,
		0.856393,
		0.864981,
		0.866118,
		0.865957,
		0.86217,
		0.84658,
		0.811468,
		0.75245,
		0.669042,
		0.564348,
		0.444555,
		0.318468,
		0.197141,
		0.093595,
		0.0226,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.011499,
		0.070064,
		0.163672,
		0.277308,
		0.397697,
		0.513512,
		0.615657,
		0.697635,
		0.756013,
		0.790953,
		0.806556,
		0.810366,
		0.810529,
		0.809382,
		0.800764,
		0.77625,
		0.729854,
		0.659459,
		0.566765,
		0.456801,
		0.337415,
		0.218859,
		0.113465,
		0.0354,
		0.000483,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.020971,
		0.086449,
		0.181135,
		0.290865,
		0.403267,
		0.507996,
		0.597047,
		0.665163,
		0.710347,
		0.734367,
		0.742861,
		0.743996,
		0.743835,
		0.74007,
		0.724736,
		0.690604,
		0.633929,
		0.55488,
		0.45715,
		0.347429,
		0.234937,
		0.131066,
		0.049109,
		0.004055,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.000432,
		0.031536,
		0.100576,
		0.192897,
		0.295462,
		0.397048,
		0.488506,
		0.563102,
		0.616965,
		0.649616,
		0.664376,
		0.668022,
		0.668179,
		0.667076,
		0.658875,
		0.635826,
		0.592741,
		0.528227,
		0.444525,
		0.347001,
		0.243645,
		0.144681,
		0.062266,
		0.010267,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.003467,
		0.04179,
		0.110928,
		0.197612,
		0.29017,
		0.378733,
		0.455512,
		0.515172,
		0.555299,
		0.57691,
		0.584648,
		0.585694,
		0.585544,
		0.582095,
		0.568211,
		0.537699,
		0.487707,
		0.418991,
		0.33549,
		0.243819,
		0.152846,
		0.073371,
		0.017889,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.008346,
		0.050357,
		0.116324,
		0.194576,
		0.274981,
		0.349164,
		0.4108,
		0.455999,
		0.483784,
		0.496513,
		0.499696,
		0.499835,
		0.498867,
		0.491749,
		0.472002,
		0.435589,
		0.381856,
		0.313298,
		0.235077,
		0.154586,
		0.08111,
		0.025572,
		0.000352,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.013754,
		0.056098,
		0.116117,
		0.18387,
		0.250868,
		0.310294,
		0.3573,
		0.389406,
		0.406945,
		0.413311,
		0.414182,
		0.414057,
		0.411201,
		0.399855,
		0.375266,
		0.335565,
		0.281881,
		0.217932,
		0.149594,
		0.084551,
		0.032043,
		0.002672,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.000256,
		0.018485,
		0.058278,
		0.110298,
		0.166381,
		0.219692,
		0.26495,
		0.298721,
		0.319805,
		0.329606,
		0.33209,
		0.3322,
		0.33144,
		0.325921,
		0.310826,
		0.283404,
		0.24359,
		0.193752,
		0.138292,
		0.083299,
		0.036295,
		0.00605,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.001814,
		0.021621,
		0.056675,
		0.099509,
		0.143689,
		0.183964,
		0.216483,
		0.239079,
		0.251619,
		0.256238,
		0.256879,
		0.256786,
		0.254699,
		0.246526,
		0.229082,
		0.201372,
		0.164591,
		0.121789,
		0.077548,
		0.037732,
		0.009308,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.003798,
		0.022647,
		0.0516,
		0.084949,
		0.11786,
		0.146521,
		0.168337,
		0.182192,
		0.188738,
		0.190422,
		0.190497,
		0.189978,
		0.186264,
		0.176263,
		0.158393,
		0.132916,
		0.101724,
		0.068049,
		0.036236,
		0.011571,
		0.000161,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.005339,
		0.021495,
		0.043827,
		0.068196,
		0.091174,
		0.110179,
		0.123644,
		0.131247,
		0.134094,
		0.134495,
		0.134436,
		0.13314,
		0.128143,
		0.117656,
		0.101297,
		0.080042,
		0.055989,
		0.032163,
		0.012361,
		0.001043,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		8.4e-005,
		0.005956,
		0.018512,
		0.03446,
		0.050988,
		0.065832,
		0.077394,
		0.08488,
		0.08848,
		0.089421,
		0.089464,
		0.089172,
		0.087112,
		0.081659,
		0.072094,
		0.058738,
		0.042807,
		0.02625,
		0.011614,
		0.001962,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		0.000473,
		0.005563,
		0.014355,
		0.024749,
		0.03498,
		0.043688,
		0.049997,
		0.05363,
		0.055015,
		0.055213,
		0.055184,
		0.054548,
		0.052138,
		0.047176,
		0.039592,
		0.02998,
		0.019471,
		0.009635,
		0.002411,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.000748,
		0.004394,
		0.009842,
		0.015878,
		0.021513,
		0.026022,
		0.029006,
		0.030469,
		0.030859,
		0.030877,
		0.030755,
		0.02991,
		0.027714,
		0.023941,
		0.018798,
		0.012854,
		0.006974,
		0.002263,
		3.2e-005,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		0.000727,
		0.00288,
		0.005762,
		0.008767,
		0.011415,
		0.013384,
		0.014543,
		0.014994,
		0.015059,
		0.01505,
		0.014841,
		0.014064,
		0.012498,
		0.01016,
		0.007282,
		0.004271,
		0.00167,
		0.000143,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000485,
		0.001481,
		0.002699,
		0.003894,
		0.004882,
		0.005553,
		0.005889,
		0.005981,
		0.005985,
		0.005956,
		0.00576,
		0.005261,
		0.004423,
		0.003312,
		0.002079,
		0.000938,
		0.000161,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		1.8e-005,
		0.000207,
		0.000524,
		0.000882,
		0.001213,
		0.001466,
		0.001619,
		0.001679,
		0.001688,
		0.001687,
		0.001659,
		0.001555,
		0.001351,
		0.001055,
		0.000703,
		0.000355,
		9.1e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		4.1e-005,
		9e-005,
		0.000141,
		0.000185,
		0.000216,
		0.000232,
		0.000237,
		0.000237,
		0.000235,
		0.000226,
		0.000203,
		0.000165,
		0.000116,
		6.4e-005,
		2.1e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		2.1e-005,
		6.4e-005,
		0.000116,
		0.000165,
		0.000203,
		0.000226,
		0.000235,
		0.000237,
		0.000237,
		0.000232,
		0.000216,
		0.000185,
		0.000141,
		9e-005,
		4.1e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		9.1e-005,
		0.000355,
		0.000703,
		0.001055,
		0.001351,
		0.001555,
		0.001659,
		0.001687,
		0.001688,
		0.001679,
		0.001619,
		0.001466,
		0.001213,
		0.000882,
		0.000524,
		0.000207,
		1.8e-005,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		0.000161,
		0.000938,
		0.002079,
		0.003312,
		0.004423,
		0.005261,
		0.00576,
		0.005956,
		0.005985,
		0.005981,
		0.005889,
		0.005553,
		0.004882,
		0.003894,
		0.002699,
		0.001481,
		0.000485,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000143,
		0.00167,
		0.004271,
		0.007282,
		0.01016,
		0.012498,
		0.014064,
		0.014841,
		0.01505,
		0.015059,
		0.014994,
		0.014543,
		0.013384,
		0.011415,
		0.008767,
		0.005762,
		0.00288,
		0.000727,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		3.2e-005,
		0.002263,
		0.006974,
		0.012854,
		0.018798,
		0.023941,
		0.027714,
		0.02991,
		0.030755,
		0.030877,
		0.030859,
		0.030469,
		0.029006,
		0.026022,
		0.021513,
		0.015878,
		0.009842,
		0.004394,
		0.000748,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.002411,
		0.009635,
		0.019471,
		0.02998,
		0.039592,
		0.047176,
		0.052138,
		0.054548,
		0.055184,
		0.055213,
		0.055015,
		0.05363,
		0.049997,
		0.043688,
		0.03498,
		0.024749,
		0.014355,
		0.005563,
		0.000473,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		0.001962,
		0.011614,
		0.02625,
		0.042807,
		0.058738,
		0.072094,
		0.081659,
		0.087112,
		0.089172,
		0.089464,
		0.089421,
		0.08848,
		0.08488,
		0.077394,
		0.065832,
		0.050988,
		0.03446,
		0.018512,
		0.005956,
		8.4e-005,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001043,
		0.012361,
		0.032163,
		0.055989,
		0.080042,
		0.101297,
		0.117656,
		0.128143,
		0.13314,
		0.134436,
		0.134495,
		0.134094,
		0.131247,
		0.123644,
		0.110179,
		0.091174,
		0.068196,
		0.043827,
		0.021495,
		0.005339,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.000161,
		0.011571,
		0.036236,
		0.068049,
		0.101724,
		0.132916,
		0.158393,
		0.176263,
		0.186264,
		0.189978,
		0.190497,
		0.190422,
		0.188738,
		0.182192,
		0.168337,
		0.146521,
		0.11786,
		0.084949,
		0.0516,
		0.022647,
		0.003798,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.009308,
		0.037732,
		0.077548,
		0.121789,
		0.164591,
		0.201372,
		0.229082,
		0.246526,
		0.254699,
		0.256786,
		0.256879,
		0.256238,
		0.251619,
		0.239079,
		0.216483,
		0.183964,
		0.143689,
		0.099509,
		0.056675,
		0.021621,
		0.001814,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.00605,
		0.036295,
		0.083299,
		0.138292,
		0.193752,
		0.24359,
		0.283404,
		0.310826,
		0.325921,
		0.33144,
		0.3322,
		0.33209,
		0.329606,
		0.319805,
		0.298721,
		0.26495,
		0.219692,
		0.166381,
		0.110298,
		0.058278,
		0.018485,
		0.000256,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.002672,
		0.032043,
		0.084551,
		0.149594,
		0.217932,
		0.281881,
		0.335565,
		0.375266,
		0.399855,
		0.411201,
		0.414057,
		0.414182,
		0.413311,
		0.406945,
		0.389406,
		0.3573,
		0.310294,
		0.250868,
		0.18387,
		0.116117,
		0.056098,
		0.013754,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.000352,
		0.025572,
		0.08111,
		0.154586,
		0.235077,
		0.313298,
		0.381856,
		0.435589,
		0.472002,
		0.491749,
		0.498867,
		0.499835,
		0.499696,
		0.496513,
		0.483784,
		0.455999,
		0.4108,
		0.349164,
		0.274981,
		0.194576,
		0.116324,
		0.050357,
		0.008346,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.017889,
		0.073371,
		0.152846,
		0.243819,
		0.33549,
		0.418991,
		0.487707,
		0.537699,
		0.568211,
		0.582095,
		0.585544,
		0.585694,
		0.584648,
		0.57691,
		0.555299,
		0.515172,
		0.455512,
		0.378733,
		0.29017,
		0.197612,
		0.110928,
		0.04179,
		0.003467,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.010267,
		0.062266,
		0.144681,
		0.243645,
		0.347001,
		0.444525,
		0.528227,
		0.592741,
		0.635826,
		0.658875,
		0.667076,
		0.668179,
		0.668022,
		0.664376,
		0.649616,
		0.616965,
		0.563102,
		0.488506,
		0.397048,
		0.295462,
		0.192897,
		0.100576,
		0.031536,
		0.000432,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.004055,
		0.049109,
		0.131066,
		0.234937,
		0.347429,
		0.45715,
		0.55488,
		0.633929,
		0.690604,
		0.724736,
		0.74007,
		0.743835,
		0.743996,
		0.742861,
		0.734367,
		0.710347,
		0.665163,
		0.597047,
		0.507996,
		0.403267,
		0.290865,
		0.181135,
		0.086449,
		0.020971,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.000483,
		0.0354,
		0.113465,
		0.218859,
		0.337415,
		0.456801,
		0.566765,
		0.659459,
		0.729854,
		0.77625,
		0.800764,
		0.809382,
		0.810529,
		0.810366,
		0.806556,
		0.790953,
		0.756013,
		0.697635,
		0.615657,
		0.513512,
		0.397697,
		0.277308,
		0.163672,
		0.070064,
		0.011499,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.0226,
		0.093595,
		0.197141,
		0.318468,
		0.444555,
		0.564348,
		0.669042,
		0.75245,
		0.811468,
		0.84658,
		0.86217,
		0.865957,
		0.866118,
		0.864981,
		0.856393,
		0.83183,
		0.785067,
		0.713679,
		0.619035,
		0.505845,
		0.381656,
		0.256439,
		0.142261,
		0.053042,
		0.004361,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.011958,
		0.073179,
		0.171783,
		0.292662,
		0.422357,
		0.549241,
		0.663773,
		0.758816,
		0.830044,
		0.876427,
		0.900656,
		0.909081,
		0.910191,
		0.910034,
		0.906329,
		0.890993,
		0.856264,
		0.797554,
		0.714073,
		0.608558,
		0.486799,
		0.35718,
		0.23032,
		0.118789,
		0.036891,
		0.000502,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.0044,
		0.053735,
		0.144773,
		0.262302,
		0.392648,
		0.523831,
		0.645788,
		0.750642,
		0.83306,
		0.890684,
		0.924584,
		0.939475,
		0.943055,
		0.943206,
		0.942136,
		0.933976,
		0.910389,
		0.864989,
		0.794878,
		0.700751,
		0.58651,
		0.458811,
		0.326664,
		0.201119,
		0.095026,
		0.022848,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.000494,
		0.036449,
		0.11786,
		0.229604,
		0.357979,
		0.490845,
		0.617806,
		0.730447,
		0.822636,
		0.890903,
		0.934871,
		0.957597,
		0.96542,
		0.966441,
		0.966297,
		0.962873,
		0.948562,
		0.915816,
		0.859861,
		0.779382,
		0.676358,
		0.555639,
		0.424544,
		0.29251,
		0.170831,
		0.07245,
		0.011792,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.022135,
		0.092422,
		0.19647,
		0.320699,
		0.452958,
		0.582681,
		0.701088,
		0.801442,
		0.879373,
		0.93327,
		0.964651,
		0.978299,
		0.98155,
		0.981686,
		0.980718,
		0.973275,
		0.95155,
		0.909308,
		0.84338,
		0.753855,
		0.643772,
		0.51873,
		0.386524,
		0.256856,
		0.141106,
		0.052157,
		0.004255,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.011259,
		0.069425,
		0.164372,
		0.282748,
		0.412507,
		0.543038,
		0.66534,
		0.772234,
		0.858647,
		0.921945,
		0.962301,
		0.982956,
		0.99,
		0.990912,
		0.990785,
		0.987715,
		0.974769,
		0.944861,
		0.893244,
		0.818227,
		0.721083,
		0.605712,
		0.478279,
		0.346903,
		0.221416,
		0.113164,
		0.034862,
		0.000471,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.004022,
		0.049465,
		0.134336,
		0.245582,
		0.371341,
		0.501051,
		0.625599,
		0.737504,
		0.831152,
		0.903084,
		0.952336,
		0.980739,
		0.992978,
		0.995868,
		0.995988,
		0.995131,
		0.988486,
		0.968915,
		0.930499,
		0.869955,
		0.786881,
		0.68353,
		0.564478,
		0.436307,
		0.30734,
		0.187433,
		0.087814,
		0.020955,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.000442,
		0.032845,
		0.106998,
		0.21019,
		0.330792,
		0.458357,
		0.583747,
		0.699281,
		0.798942,
		0.878618,
		0.936403,
		0.9729,
		0.991411,
		0.997669,
		0.998473,
		0.998361,
		0.995644,
		0.984094,
		0.957169,
		0.910271,
		0.841453,
		0.751401,
		0.643164,
		0.52184,
		0.394307,
		0.268996,
		0.155717,
		0.065521,
		0.01059,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.019668,
		0.082699,
		0.177175,
		0.291735,
		0.416091,
		0.541136,
		0.659076,
		0.763598,
		0.85008,
		0.915849,
		0.960466,
		0.98597,
		0.996864,
		0.999416,
		0.999522,
		0.998768,
		0.992877,
		0.97538,
		0.940734,
		0.885637,
		0.80931,
		0.713346,
		0.601425,
		0.479047,
		0.353287,
		0.232614,
		0.126741,
		0.046503,
		0.003769,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.009922,
		0.061582,
		0.146868,
		0.254699,
		0.374977,
		0.498671,
		0.617935,
		0.726248,
		0.818594,
		0.891682,
		0.944206,
		0.977092,
		0.993631,
		0.999178,
		0.999885,
		0.999787,
		0.997388,
		0.98711,
		0.962953,
		0.920512,
		0.857677,
		0.774666,
		0.673812,
		0.559298,
		0.436917,
		0.313873,
		0.198621,
		0.100736,
		0.03082,
		0.000414,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.003529,
		0.043674,
		0.11942,
		0.21998,
		0.335457,
		0.456932,
		0.576554,
		0.687664,
		0.784947,
		0.864619,
		0.924657,
		0.965039,
		0.987932,
		0.997633,
		0.999888,
		0.999981,
		0.999317,
		0.994091,
		0.978449,
		0.947222,
		0.897143,
		0.827155,
		0.738309,
		0.633539,
		0.51742,
		0.395962,
		0.276435,
		0.167239,
		0.077791,
		0.018444,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.000387,
		0.028941,
		0.094888,
		0.187731,
		0.297794,
		0.416289,
		0.535398,
		0.648379,
		0.749695,
		0.835175,
		0.90221,
		0.949977,
		0.979644,
		0.994447,
		0.999373,
		0.999997,
		0.999911,
		0.997788,
		0.988624,
		0.966918,
		0.928478,
		0.871094,
		0.794617,
		0.70079,
		0.59303,
		0.476214,
		0.3565,
		0.241181,
		0.138572,
		0.057915,
		0.009304,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.017323,
		0.073275,
		0.158037,
		0.262156,
		0.376995,
		0.494799,
		0.608789,
		0.713267,
		0.803759,
		0.877185,
		0.932051,
		0.968661,
		0.989256,
		0.997917,
		0.999917,
		0.999999,
		0.999412,
		0.994763,
		0.980744,
		0.952544,
		0.906963,
		0.842738,
		0.760489,
		0.662527,
		0.552654,
		0.435974,
		0.318742,
		0.208236,
		0.112664,
		0.041077,
		0.00331,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.008741,
		0.05456,
		0.130939,
		0.228658,
		0.339242,
		0.455022,
		0.569217,
		0.676022,
		0.770731,
		0.849883,
		0.911434,
		0.954949,
		0.981771,
		0.995056,
		0.999446,
		0.999999,
		0.999922,
		0.998037,
		0.989843,
		0.970291,
		0.935405,
		0.882926,
		0.812415,
		0.725136,
		0.623865,
		0.512705,
		0.396928,
		0.282841,
		0.177679,
		0.089522,
		0.027226,
		0.000363,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.003111,
		0.038704,
		0.106458,
		0.197385,
		0.303183,
		0.416289,
		0.529947,
		0.638287,
		0.73643,
		0.820615,
		0.888342,
		0.938558,
		0.971816,
		0.99039,
		0.998148,
		0.999927,
		0.999999,
		0.999478,
		0.995329,
		0.98273,
		0.957205,
		0.915644,
		0.856637,
		0.780455,
		0.688896,
		0.585111,
		0.473438,
		0.359265,
		0.248916,
		0.149561,
		0.069138,
		0.0163,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.000342,
		0.025659,
		0.084596,
		0.168399,
		0.268945,
		0.37879,
		0.491227,
		0.600358,
		0.701178,
		0.789689,
		0.863025,
		0.91961,
		0.959323,
		0.983628,
		0.995583,
		0.999508,
		0.999999,
		0.999931,
		0.998251,
		0.990902,
		0.973246,
		0.941522,
		0.893455,
		0.828385,
		0.74718,
		0.652082,
		0.54654,
		0.435073,
		0.323142,
		0.217061,
		0.123912,
		0.051487,
		0.008227,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.015366,
		0.065341,
		0.14174,
		0.236626,
		0.342684,
		0.453277,
		0.562504,
		0.665277,
		0.757413,
		0.835752,
		0.898284,
		0.944313,
		0.974587,
		0.991379,
		0.998347,
		0.999935,
		0.999999,
		0.999535,
		0.99582,
		0.984468,
		0.961312,
		0.92335,
		0.86907,
		0.798462,
		0.712898,
		0.614981,
		0.5084,
		0.397801,
		0.28869,
		0.187344,
		0.100747,
		0.036531,
		0.002929,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.007758,
		0.048665,
		0.117431,
		0.206304,
		0.308108,
		0.416289,
		0.524968,
		0.629005,
		0.724081,
		0.806797,
		0.874792,
		0.926878,
		0.963183,
		0.985256,
		0.996042,
		0.999561,
		0.999999,
		0.999939,
		0.998436,
		0.991827,
		0.975845,
		0.94694,
		0.902848,
		0.842738,
		0.767158,
		0.677898,
		0.577857,
		0.470906,
		0.361785,
		0.256014,
		0.159817,
		0.080064,
		0.024223,
		0.000322,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.002763,
		0.034535,
		0.095481,
		0.178037,
		0.275172,
		0.380428,
		0.487964,
		0.592618,
		0.689972,
		0.776439,
		0.849368,
		0.907163,
		0.949417,
		0.977027,
		0.992245,
		0.99852,
		0.999942,
		0.999999,
		0.999585,
		0.996249,
		0.985994,
		0.964943,
		0.930208,
		0.880212,
		0.814721,
		0.734755,
		0.642452,
		0.540946,
		0.43425,
		0.327165,
		0.2252,
		0.13451,
		0.061846,
		0.014509,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.000304,
		0.022905,
		0.075884,
		0.151865,
		0.243966,
		0.345833,
		0.451682,
		0.556348,
		0.655346,
		0.744944,
		0.822257,
		0.885348,
		0.933353,
		0.9666,
		0.986687,
		0.996443,
		0.999607,
		0.999999,
		0.999945,
		0.998598,
		0.992637,
		0.978139,
		0.951752,
		0.911245,
		0.855661,
		0.785281,
		0.701521,
		0.606807,
		0.50446,
		0.398598,
		0.294054,
		0.19631,
		0.111441,
		0.046069,
		0.007327,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.013723,
		0.058623,
		0.127813,
		0.214561,
		0.312626,
		0.416289,
		0.520404,
		0.620445,
		0.71257,
		0.793707,
		0.861636,
		0.915109,
		0.953955,
		0.979184,
		0.993005,
		0.998671,
		0.999948,
		0.999999,
		0.999628,
		0.996625,
		0.987338,
		0.968161,
		0.936326,
		0.890217,
		0.829424,
		0.754672,
		0.667708,
		0.571191,
		0.468589,
		0.364093,
		0.262549,
		0.169394,
		0.090614,
		0.032698,
		0.00261,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.006932,
		0.043674,
		0.105894,
		0.187012,
		0.280903,
		0.381929,
		0.484973,
		0.58549,
		0.679561,
		0.763962,
		0.836245,
		0.894835,
		0.939136,
		0.969633,
		0.987949,
		0.996795,
		0.999648,
		0.999999,
		0.999951,
		0.998739,
		0.99335,
		0.980168,
		0.956035,
		0.918768,
		0.867313,
		0.801739,
		0.723143,
		0.63355,
		0.535808,
		0.433497,
		0.330857,
		0.232725,
		0.144484,
		0.072021,
		0.02169,
		0.000287,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.00247,
		0.031004,
		0.086107,
		0.161357,
		0.250747,
		0.348725,
		0.450219,
		0.550683,
		0.646144,
		0.73326,
		0.809397,
		0.872709,
		0.922235,
		0.958001,
		0.981094,
		0.993675,
		0.998803,
		0.999954,
		0.999999,
		0.999666,
		0.996954,
		0.988525,
		0.971021,
		0.941795,
		0.899218,
		0.842738,
		0.772836,
		0.690927,
		0.599263,
		0.500843,
		0.39933,
		0.298992,
		0.204642,
		0.121599,
		0.055645,
		0.012998,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.000272,
		0.02057,
		0.068445,
		0.137622,
		0.222223,
		0.316784,
		0.416289,
		0.516206,
		0.612528,
		0.701826,
		0.781314,
		0.848922,
		0.903379,
		0.944313,
		0.972331,
		0.989067,
		0.997104,
		0.999683,
		0.999999,
		0.999956,
		0.998863,
		0.99398,
		0.981968,
		0.959858,
		0.925522,
		0.877839,
		0.816702,
		0.742944,
		0.658245,
		0.565045,
		0.466459,
		0.366214,
		0.268583,
		0.178346,
		0.100747,
		0.041461,
		0.006567,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.012329,
		0.052888,
		0.115823,
		0.195381,
		0.286194,
		0.383309,
		0.482221,
		0.578906,
		0.669872,
		0.752213,
		0.823672,
		0.882719,
		0.928639,
		0.961616,
		0.982792,
		0.994266,
		0.99892,
		0.999958,
		0.999999,
		0.999699,
		0.997245,
		0.989577,
		0.973569,
		0.946697,
		0.907331,
		0.854812,
		0.789417,
		0.712278,
		0.6253,
		0.531073,
		0.432804,
		0.334256,
		0.2397,
		0.153869,
		0.081927,
		0.029437,
		0.002341,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.006231,
		0.039412,
		0.095965,
		0.170259,
		0.257029,
		0.35139,
		0.448872,
		0.545454,
		0.637598,
		0.722303,
		0.797161,
		0.860425,
		0.911086,
		0.948956,
		0.974739,
		0.990058,
		0.997377,
		0.999714,
		0.999999,
		0.99996,
		0.998972,
		0.994536,
		0.98357,
		0.963278,
		0.931598,
		0.887363,
		0.830324,
		0.761088,
		0.681045,
		0.592281,
		0.497509,
		0.400003,
		0.30355,
		0.212399,
		0.131232,
		0.065127,
		0.019535,
		0.000257,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.002221,
		0.027987,
		0.078042,
		0.146884,
		0.229348,
		0.320624,
		0.416289,
		0.512331,
		0.605186,
		0.691782,
		0.769588,
		0.836674,
		0.891784,
		0.934406,
		0.964853,
		0.984305,
		0.99479,
		0.999022,
		0.999962,
		1,
		0.999728,
		0.997501,
		0.990511,
		0.975845,
		0.951098,
		0.914656,
		0.865776,
		0.804566,
		0.731919,
		0.649438,
		0.559361,
		0.464497,
		0.368169,
		0.27417,
		0.186723,
		0.110446,
		0.05033,
		0.011711,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.000244,
		0.018575,
		0.062044,
		0.125272,
		0.203198,
		0.291093,
		0.384583,
		0.47968,
		0.572806,
		0.660837,
		0.741144,
		0.811648,
		0.870879,
		0.918052,
		0.953131,
		0.976892,
		0.990939,
		0.997618,
		0.999741,
		1,
		0.999964,
		0.999069,
		0.99503,
		0.984999,
		0.966345,
		0.937074,
		0.895994,
		0.842738,
		0.777728,
		0.7021,
		0.617635,
		0.526696,
		0.432165,
		0.337397,
		0.246181,
		0.162702,
		0.091512,
		0.037511,
		0.00592,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.011138,
		0.047952,
		0.105431,
		0.178617,
		0.262862,
		0.353853,
		0.447628,
		0.540612,
		0.629643,
		0.712016,
		0.785527,
		0.848532,
		0.900005,
		0.939609,
		0.967759,
		0.985655,
		0.995256,
		0.999113,
		0.999966,
		1,
		0.999754,
		0.997729,
		0.991344,
		0.977883,
		0.95506,
		0.921283,
		0.875748,
		0.818422,
		0.749996,
		0.671812,
		0.585802,
		0.494427,
		0.400625,
		0.30777,
		0.219635,
		0.140358,
		0.074425,
		0.02664,
		0.002111,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.005631,
		0.035743,
		0.08736,
		0.155629,
		0.235987,
		0.324181,
		0.416289,
		0.508744,
		0.598361,
		0.682378,
		0.75849,
		0.824905,
		0.880394,
		0.924359,
		0.956891,
		0.978822,
		0.991726,
		0.997833,
		0.999765,
		1,
		0.999968,
		0.999154,
		0.99547,
		0.986276,
		0.9691,
		0.94202,
		0.903828,
		0.854068,
		0.793,
		0.721547,
		0.641225,
		0.55409,
		0.462682,
		0.369978,
		0.279358,
		0.19457,
		0.119705,
		0.059174,
		0.017685,
		0.000232,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.002008,
		0.025389,
		0.071053,
		0.134253,
		0.210511,
		0.295641,
		0.385763,
		0.477327,
		0.56714,
		0.652395,
		0.730708,
		0.800163,
		0.859359,
		0.907473,
		0.944313,
		0.970373,
		0.986864,
		0.995671,
		0.999193,
		0.999969,
		1,
		0.999776,
		0.997931,
		0.992087,
		0.979712,
		0.958631,
		0.927288,
		0.884829,
		0.831109,
		0.766643,
		0.692551,
		0.610497,
		0.522638,
		0.431573,
		0.340308,
		0.252217,
		0.171018,
		0.100747,
		0.045739,
		0.010606,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.000221,
		0.016857,
		0.056497,
		0.114498,
		0.186469,
		0.268292,
		0.356136,
		0.446476,
		0.536117,
		0.622222,
		0.702346,
		0.77447,
		0.837049,
		0.889065,
		0.930079,
		0.960286,
		0.980556,
		0.992429,
		0.998024,
		0.999787,
		1,
		0.999971,
		0.99923,
		0.995862,
		0.987421,
		0.971582,
		0.946495,
		0.91095,
		0.864419,
		0.807029,
		0.739516,
		0.663169,
		0.579774,
		0.491569,
		0.401201,
		0.311689,
		0.226396,
		0.149001,
		0.083483,
		0.034098,
		0.005363,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.010111,
		0.043674,
		0.096367,
		0.163887,
		0.242186,
		0.327485,
		0.416289,
		0.505413,
		0.592002,
		0.673559,
		0.747985,
		0.813614,
		0.869259,
		0.914268,
		0.948572,
		0.97273,
		0.987949,
		0.996042,
		0.999265,
		0.999972,
		1,
		0.999797,
		0.998111,
		0.992753,
		0.981357,
		0.961859,
		0.932739,
		0.893112,
		0.842738,
		0.781985,
		0.711778,
		0.63355,
		0.549189,
		0.460998,
		0.371655,
		0.284186,
		0.201932,
		0.128533,
		0.067904,
		0.024223,
		0.001913,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.005114,
		0.032563,
		0.079857,
		0.142782,
		0.217363,
		0.299874,
		0.386858,
		0.475143,
		0.561865,
		0.644493,
		0.720862,
		0.789202,
		0.848188,
		0.89698,
		0.935275,
		0.963355,
		0.982117,
		0.99306,
		0.998194,
		0.999806,
		1,
		0.999973,
		0.999297,
		0.996213,
		0.98845,
		0.973821,
		0.950551,
		0.917436,
		0.87389,
		0.819929,
		0.756129,
		0.683581,
		0.603835,
		0.518865,
		0.431024,
		0.343012,
		0.25785,
		0.178855,
		0.10962,
		0.053998,
		0.016085,
		0.000211,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.001825,
		0.023136,
		0.064959,
		0.123166,
		0.193856,
		0.273359,
		0.358259,
		0.445405,
		0.531932,
		0.615286,
		0.693246,
		0.763962,
		0.825987,
		0.878322,
		0.920461,
		0.952436,
		0.974858,
		0.988925,
		0.996374,
		0.999328,
		0.999974,
		1,
		0.999814,
		0.998273,
		0.99335,
		0.982839,
		0.96478,
		0.937694,
		0.900677,
		0.85341,
		0.796135,
		0.72961,
		0.655065,
		0.574153,
		0.488913,
		0.401737,
		0.315337,
		0.232725,
		0.157186,
		0.092264,
		0.041747,
		0.00965,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.000201,
		0.015366,
		0.05166,
		0.105044,
		0.171689,
		0.247986,
		0.330561,
		0.416289,
		0.502313,
		0.586062,
		0.665277,
		0.738037,
		0.802795,
		0.858415,
		0.904213,
		0.940003,
		0.966137,
		0.983525,
		0.993626,
		0.998347,
		0.999823,
		1,
		0.999976,
		0.999358,
		0.996527,
		0.989376,
		0.975845,
		0.954233,
		0.92335,
		0.882566,
		0.831801,
		0.771497,
		0.70257,
		0.626365,
		0.544621,
		0.459432,
		0.373216,
		0.28869,
		0.208847,
		0.13694,
		0.076461,
		0.031129,
		0.004882,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.00922,
		0.039943,
		0.088416,
		0.150882,
		0.223793,
		0.303823,
		0.387877,
		0.473108,
		0.556941,
		0.637084,
		0.711563,
		0.778747,
		0.837381,
		0.886629,
		0.926112,
		0.955947,
		0.976785,
		0.989805,
		0.996672,
		0.999385,
		0.999977,
		1,
		0.99983,
		0.998417,
		0.993888,
		0.984178,
		0.967429,
		0.942207,
		0.907596,
		0.863213,
		0.809195,
		0.746155,
		0.675142,
		0.597605,
		0.51535,
		0.430513,
		0.345531,
		0.263117,
		0.186245,
		0.118126,
		0.0622,
		0.022121,
		0.001742,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.004665,
		0.029788,
		0.073275,
		0.131445,
		0.200812,
		0.278096,
		0.360237,
		0.444407,
		0.528027,
		0.608789,
		0.684672,
		0.753975,
		0.815347,
		0.867815,
		0.910833,
		0.944313,
		0.968661,
		0.984799,
		0.994136,
		0.998483,
		0.999838,
		1,
		0.999978,
		0.999412,
		0.99681,
		0.990212,
		0.977679,
		0.957583,
		0.928752,
		0.890522,
		0.842738,
		0.785723,
		0.720239,
		0.647454,
		0.568899,
		0.486437,
		0.402235,
		0.318742,
		0.238661,
		0.164939,
		0.100747,
		0.049471,
		0.014693,
		0.000192,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.001665,
		0.02117,
		0.059613,
		0.113386,
		0.179066,
		0.253422,
		0.333432,
		0.416289,
		0.499421,
		0.580504,
		0.657487,
		0.72861,
		0.792437,
		0.847883,
		0.894253,
		0.931276,
		0.959144,
		0.978531,
		0.990599,
		0.99694,
		0.999437,
		0.999979,
		1,
		0.999845,
		0.998546,
		0.994372,
		0.985389,
		0.969836,
		0.946324,
		0.913932,
		0.87223,
		0.82126,
		0.761512,
		0.693878,
		0.619625,
		0.540354,
		0.457972,
		0.374671,
		0.292901,
		0.215353,
		0.144944,
		0.084802,
		0.038254,
		0.008818,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.000183,
		0.014064,
		0.047417,
		0.096707,
		0.158573,
		0.229837,
		0.307516,
		0.388827,
		0.47121,
		0.552335,
		0.630124,
		0.702773,
		0.768774,
		0.826947,
		0.876466,
		0.9169,
		0.948246,
		0.970957,
		0.985952,
		0.994596,
		0.998606,
		0.999851,
		1,
		0.99998,
		0.99946,
		0.997064,
		0.990967,
		0.979343,
		0.960634,
		0.933692,
		0.897829,
		0.852824,
		0.798899,
		0.736686,
		0.667192,
		0.591766,
		0.512066,
		0.430036,
		0.347883,
		0.268054,
		0.193222,
		0.126271,
		0.070284,
		0.028532,
		0.004462,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.008442,
		0.03667,
		0.081404,
		0.139346,
		0.207371,
		0.282534,
		0.362085,
		0.443475,
		0.524376,
		0.602692,
		0.676582,
		0.74448,
		0.805122,
		0.857572,
		0.901258,
		0.936004,
		0.962058,
		0.980117,
		0.991318,
		0.997182,
		0.999482,
		0.999981,
		1,
		0.999858,
		0.998663,
		0.994809,
		0.986488,
		0.972026,
		0.950084,
		0.919743,
		0.880531,
		0.832415,
		0.775775,
		0.711367,
		0.640295,
		0.563978,
		0.484124,
		0.402701,
		0.321925,
		0.244237,
		0.172287,
		0.108924,
		0.057184,
		0.02028,
		0.001593,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.004273,
		0.027353,
		0.067471,
		0.121393,
		0.186045,
		0.258527,
		0.336118,
		0.416289,
		0.496716,
		0.575292,
		0.650148,
		0.71967,
		0.782523,
		0.837676,
		0.884434,
		0.922467,
		0.951842,
		0.973047,
		0.986998,
		0.995012,
		0.998717,
		0.999863,
		1,
		0.999981,
		0.999504,
		0.997294,
		0.991651,
		0.980856,
		0.963419,
		0.938217,
		0.904547,
		0.862134,
		0.811114,
		0.752002,
		0.685666,
		0.613292,
		0.536358,
		0.456608,
		0.376031,
		0.296847,
		0.221482,
		0.152563,
		0.092903,
		0.045488,
		0.013475,
		0.000176,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.001526,
		0.019444,
		0.054899,
		0.104717,
		0.165879,
		0.235528,
		0.310976,
		0.389717,
		0.469434,
		0.548017,
		0.623575,
		0.694455,
		0.759263,
		0.816885,
		0.86652,
		0.907703,
		0.940336,
		0.96472,
		0.981561,
		0.991969,
		0.9974,
		0.999524,
		0.999982,
		1,
		0.999869,
		0.998768,
		0.995205,
		0.987485,
		0.974023,
		0.953525,
		0.925079,
		0.888183,
		0.842738,
		0.78903,
		0.727694,
		0.659694,
		0.586284,
		0.508991,
		0.42959,
		0.350084,
		0.272688,
		0.199814,
		0.134062,
		0.078205,
		0.035182,
		0.008089,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.000168,
		0.012921,
		0.043674,
		0.089317,
		0.146884,
		0.213563,
		0.286701,
		0.363815,
		0.442603,
		0.520954,
		0.596961,
		0.66894,
		0.735449,
		0.795304,
		0.84761,
		0.891784,
		0.927583,
		0.955135,
		0.974954,
		0.987949,
		0.995389,
		0.998817,
		0.999874,
		1,
		0.999983,
		0.999543,
		0.997501,
		0.992272,
		0.982234,
		0.965963,
		0.942367,
		0.910731,
		0.870737,
		0.822445,
		0.766271,
		0.702959,
		0.63355,
		0.559361,
		0.481957,
		0.403137,
		0.324909,
		0.249484,
		0.179254,
		0.11679,
		0.064824,
		0.026246,
		0.004095,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.007758,
		0.033782,
		0.07519,
		0.12907,
		0.192655,
		0.263329,
		0.338636,
		0.416289,
		0.49418,
		0.570395,
		0.643226,
		0.711187,
		0.773038,
		0.827802,
		0.874792,
		0.913638,
		0.944313,
		0.967154,
		0.982876,
		0.992561,
		0.997598,
		0.999561,
		0.999984,
		1,
		0.99988,
		0.998863,
		0.995564,
		0.988393,
		0.975845,
		0.956676,
		0.929985,
		0.895243,
		0.8523,
		0.801356,
		0.742944,
		0.677898,
		0.607332,
		0.532609,
		0.455329,
		0.377306,
		0.300551,
		0.227264,
		0.159817,
		0.100747,
		0.052748,
		0.01866,
		0.001462,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.003928,
		0.025204,
		0.062328,
		0.112441,
		0.172822,
		0.240892,
		0.314225,
		0.39055,
		0.467769,
		0.543963,
		0.617404,
		0.686577,
		0.750188,
		0.807193,
		0.856815,
		0.898568,
		0.932289,
		0.958153,
		0.976697,
		0.988816,
		0.995731,
		0.998907,
		0.999884,
		1,
		0.999984,
		0.999578,
		0.997689,
		0.992836,
		0.98349,
		0.968292,
		0.946178,
		0.91643,
		0.878693,
		0.832964,
		0.779572,
		0.719149,
		0.65261,
		0.581127,
		0.506107,
		0.429172,
		0.352149,
		0.277047,
		0.206051,
		0.141513,
		0.085934,
		0.041967,
		0.012401,
		0.000161,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.001403,
		0.017921,
		0.050721,
		0.096998,
		0.154076,
		0.219416,
		0.290619,
		0.365439,
		0.441786,
		0.51774,
		0.591564,
		0.661713,
		0.726854,
		0.785882,
		0.837941,
		0.882447,
		0.919112,
		0.947967,
		0.969383,
		0.984077,
		0.993099,
		0.997777,
		0.999595,
		0.999985,
		1,
		0.999889,
		0.998949,
		0.99589,
		0.98922,
		0.977512,
		0.959568,
		0.934501,
		0.901765,
		0.861163,
		0.812825,
		0.757191,
		0.694984,
		0.627186,
		0.55502,
		0.479925,
		0.403546,
		0.327712,
		0.254429,
		0.185867,
		0.124349,
		0.072345,
		0.032464,
		0.007447,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.000155,
		0.011912,
		0.040357,
		0.082738,
		0.136427,
		0.198921,
		0.267854,
		0.341002,
		0.416289,
		0.4918,
		0.565786,
		0.636686,
		0.70313,
		0.763962,
		0.81826,
		0.865352,
		0.904839,
		0.936624,
		0.960924,
		0.978292,
		0.989606,
		0.996042,
		0.998989,
		0.999893,
		1,
		0.999985,
		0.99961,
		0.99786,
		0.99335,
		0.984638,
		0.970427,
		0.949683,
		0.921688,
		0.886059,
		0.842738,
		0.791976,
		0.734311,
		0.670542,
		0.601712,
		0.529086,
		0.45413,
		0.378502,
		0.304035,
		0.232725,
		0.166724,
		0.108329,
		0.059973,
		0.024223,
		0.003771,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.007154,
		0.031222,
		0.069658,
		0.119879,
		0.179424,
		0.245958,
		0.317281,
		0.391333,
		0.466205,
		0.540147,
		0.61158,
		0.679106,
		0.741528,
		0.797864,
		0.847365,
		0.889537,
		0.924165,
		0.95133,
		0.971428,
		0.985175,
		0.99359,
		0.99794,
		0.999625,
		0.999986,
		1,
		0.999897,
		0.999027,
		0.996187,
		0.989975,
		0.979038,
		0.962225,
		0.938663,
		0.907796,
		0.869387,
		0.823506,
		0.770509,
		0.711022,
		0.645911,
		0.576269,
		0.503396,
		0.42878,
		0.354089,
		0.281153,
		0.211958,
		0.148635,
		0.093452,
		0.048808,
		0.017226,
		0.001347,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.003623,
		0.023299,
		0.057749,
		0.104436,
		0.160938,
		0.224955,
		0.294311,
		0.366965,
		0.441017,
		0.514717,
		0.586474,
		0.654869,
		0.718669,
		0.776841,
		0.828569,
		0.873276,
		0.910641,
		0.940621,
		0.963472,
		0.979753,
		0.990328,
		0.996325,
		0.999064,
		0.999901,
		1,
		0.999987,
		0.999639,
		0.998016,
		0.993819,
		0.985689,
		0.972387,
		0.95291,
		0.926546,
		0.892886,
		0.851827,
		0.803552,
		0.748515,
		0.687413,
		0.621173,
		0.550931,
		0.478015,
		0.40393,
		0.330348,
		0.259096,
		0.192146,
		0.131607,
		0.079717,
		0.038838,
		0.011451,
		0.000148,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.001294,
		0.01657,
		0.047001,
		0.090096,
		0.143473,
		0.204865,
		0.272124,
		0.343228,
		0.416289,
		0.489559,
		0.561441,
		0.630499,
		0.69547,
		0.755278,
		0.809049,
		0.856131,
		0.896111,
		0.928835,
		0.954428,
		0.973306,
		0.986181,
		0.994038,
		0.998088,
		0.999653,
		0.999987,
		1,
		0.999905,
		0.999098,
		0.996457,
		0.990665,
		0.980438,
		0.964669,
		0.942504,
		0.913379,
		0.877025,
		0.833458,
		0.782964,
		0.726079,
		0.663568,
		0.596407,
		0.525768,
		0.453002,
		0.379627,
		0.307317,
		0.237891,
		0.173305,
		0.115644,
		0.067117,
		0.030049,
		0.006878,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.000143,
		0.011016,
		0.037403,
		0.076857,
		0.127035,
		0.185705,
		0.250747,
		0.32016,
		0.39207,
		0.464734,
		0.536552,
		0.606074,
		0.672015,
		0.73326,
		0.788888,
		0.838179,
		0.880639,
		0.916014,
		0.944313,
		0.965817,
		0.981094,
		0.990988,
		0.996584,
		0.999131,
		0.999909,
		1,
		0.999988,
		0.999666,
		0.998157,
		0.994248,
		0.986651,
		0.974188,
		0.955885,
		0.931037,
		0.899218,
		0.860284,
		0.814362,
		0.761826,
		0.703287,
		0.639566,
		0.571684,
		0.500843,
		0.428411,
		0.355915,
		0.285028,
		0.217558,
		0.155444,
		0.100747,
		0.055645,
		0.022426,
		0.003484,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.006618,
		0.028941,
		0.064712,
		0.111627,
		0.167488,
		0.230203,
		0.297795,
		0.368403,
		0.440294,
		0.511868,
		0.581665,
		0.64838,
		0.710869,
		0.768167,
		0.819497,
		0.864293,
		0.902212,
		0.933156,
		0.957286,
		0.975034,
		0.987103,
		0.994448,
		0.998223,
		0.999678,
		0.999988,
		1,
		0.999912,
		0.999163,
		0.996704,
		0.991297,
		0.981724,
		0.96692,
		0.946052,
		0.918552,
		0.884124,
		0.842738,
		0.794618,
		0.74022,
		0.680218,
		0.615483,
		0.547074,
		0.476215,
		0.404291,
		0.332833,
		0.263508,
		0.198115,
		0.138573,
		0.086916,
		0.045292,
		0.015951,
		0.001245,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.003352,
		0.021601,
		0.053654,
		0.09725,
		0.150223,
		0.210511,
		0.27616,
		0.345327,
		0.416289,
		0.487448,
		0.557338,
		0.62464,
		0.688183,
		0.746966,
		0.800163,
		0.847143,
		0.887485,
		0.920996,
		0.947725,
		0.967979,
		0.982327,
		0.991594,
		0.99682,
		0.999193,
		0.999915,
		1,
		0.999988,
		0.99969,
		0.998286,
		0.99464,
		0.987535,
		0.975845,
		0.958631,
		0.935195,
		0.905097,
		0.868161,
		0.824461,
		0.774307,
		0.718226,
		0.656948,
		0.591389,
		0.522638,
		0.451939,
		0.380687,
		0.310415,
		0.242783,
		0.179577,
		0.122695,
		0.074147,
		0.036046,
		0.010606,
		0.000137,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.001198,
		0.015366,
		0.043674,
		0.083901,
		0.133916,
		0.191686,
		0.255282,
		0.322879,
		0.392764,
		0.463346,
		0.533157,
		0.600863,
		0.665277,
		0.725363,
		0.780253,
		0.829262,
		0.871897,
		0.907879,
		0.937158,
		0.959925,
		0.976625,
		0.987949,
		0.994823,
		0.998347,
		0.999701,
		0.999989,
		1,
		0.999918,
		0.999222,
		0.99693,
		0.991877,
		0.982906,
		0.968996,
		0.949334,
		0.92335,
		0.890728,
		0.851398,
		0.805528,
		0.753505,
		0.695919,
		0.63355,
		0.567351,
		0.498434,
		0.428063,
		0.357638,
		0.28869,
		0.222873,
		0.161955,
		0.107814,
		0.062433,
		0.027894,
		0.006372,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.000132,
		0.010218,
		0.034761,
		0.071577,
		0.118572,
		0.173742,
		0.235182,
		0.301087,
		0.369759,
		0.439612,
		0.509177,
		0.577116,
		0.642221,
		0.703432,
		0.759844,
		0.810721,
		0.855511,
		0.893857,
		0.925619,
		0.950882,
		0.969973,
		0.983462,
		0.992149,
		0.997036,
		0.999249,
		0.999921,
		1,
		0.999989,
		0.999712,
		0.998404,
		0.994999,
		0.988347,
		0.977373,
		0.961169,
		0.939048,
		0.910561,
		0.875503,
		0.833904,
		0.786014,
		0.732288,
		0.673374,
		0.610092,
		0.543429,
		0.474517,
		0.404632,
		0.335179,
		0.267685,
		0.203793,
		0.145256,
		0.09393,
		0.051768,
		0.020821,
		0.003228,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.00614,
		0.026902,
		0.060273,
		0.104192,
		0.156687,
		0.215878,
		0.27998,
		0.34731,
		0.416289,
		0.485454,
		0.553458,
		0.619082,
		0.681244,
		0.739008,
		0.791594,
		0.838394,
		0.878987,
		0.913148,
		0.940869,
		0.962366,
		0.978092,
		0.988728,
		0.995167,
		0.99846,
		0.999722,
		0.99999,
		1,
		0.999924,
		0.999276,
		0.997137,
		0.99241,
		0.983995,
		0.970913,
		0.952372,
		0.927805,
		0.896877,
		0.859486,
		0.815748,
		0.76599,
		0.710728,
		0.650658,
		0.586639,
		0.51968,
		0.450936,
		0.381687,
		0.313342,
		0.247423,
		0.185559,
		0.129487,
		0.081039,
		0.042142,
		0.014812,
		0.001154,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.003111,
		0.020082,
		0.049979,
		0.090775,
		0.140529,
		0.197385,
		0.259581,
		0.325448,
		0.39342,
		0.462036,
		0.529947,
		0.595924,
		0.658868,
		0.717815,
		0.771949,
		0.820615,
		0.863328,
		0.899793,
		0.929913,
		0.953807,
		0.971816,
		0.984507,
		0.99266,
		0.997234,
		0.999301,
		0.999927,
		1,
		0.99999,
		0.999732,
		0.998512,
		0.995329,
		0.989094,
		0.978782,
		0.963517,
		0.942622,
		0.915644,
		0.882352,
		0.842738,
		0.797,
		0.745529,
		0.688896,
		0.627839,
		0.563249,
		0.496159,
		0.427734,
		0.359265,
		0.292156,
		0.227922,
		0.168182,
		0.114651,
		0.069138,
		0.033544,
		0.009851,
		0.000127,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.001112,
		0.014288,
		0.040688,
		0.07832,
		0.125272,
		0.179716,
		0.23991,
		0.304203,
		0.371041,
		0.438967,
		0.506634,
		0.572806,
		0.636368,
		0.696335,
		0.751857,
		0.802239,
		0.846941,
		0.885604,
		0.918052,
		0.944313,
		0.964625,
		0.979446,
		0.989445,
		0.995483,
		0.998563,
		0.999741,
		0.99999,
		1,
		0.999929,
		0.999325,
		0.997327,
		0.992899,
		0.984999,
		0.972685,
		0.955189,
		0.931946,
		0.902608,
		0.867043,
		0.825327,
		0.777728,
		0.724699,
		0.666857,
		0.604978,
		0.539979,
		0.472913,
		0.404954,
		0.337397,
		0.271644,
		0.2092,
		0.151669,
		0.100747,
		0.058222,
		0.025962,
		0.00592,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.000123,
		0.009504,
		0.032389,
		0.066821,
		0.110919,
		0.16288,
		0.220985,
		0.283601,
		0.349184,
		0.416289,
		0.483569,
		0.549783,
		0.613805,
		0.674631,
		0.731385,
		0.783333,
		0.82989,
		0.870636,
		0.905327,
		0.933906,
		0.95652,
		0.973521,
		0.985471,
		0.99313,
		0.997416,
		0.999348,
		0.999932,
		1,
		0.999991,
		0.99975,
		0.998612,
		0.995632,
		0.989783,
		0.980084,
		0.965692,
		0.945942,
		0.920376,
		0.888746,
		0.851009,
		0.807315,
		0.758,
		0.703566,
		0.644676,
		0.582134,
		0.516882,
		0.449987,
		0.382633,
		0.316114,
		0.251827,
		0.191269,
		0.136026,
		0.087777,
		0.048281,
		0.019382,
		0.003,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.005711,
		0.02507,
		0.056273,
		0.09747,
		0.146884,
		0.20282,
		0.263662,
		0.327881,
		0.39404,
		0.460796,
		0.526907,
		0.591237,
		0.652767,
		0.710598,
		0.763962,
		0.812236,
		0.854945,
		0.891784,
		0.92262,
		0.947512,
		0.966719,
		0.980698,
		0.990107,
		0.995774,
		0.998658,
		0.999759,
		0.999991,
		1,
		0.999934,
		0.99937,
		0.997501,
		0.99335,
		0.985926,
		0.974325,
		0.957802,
		0.935798,
		0.907953,
		0.874111,
		0.834309,
		0.788769,
		0.737881,
		0.682196,
		0.622411,
		0.559361,
		0.494005,
		0.427424,
		0.360804,
		0.295441,
		0.232725,
		0.17414,
		0.121258,
		0.075735,
		0.039308,
		0.013791,
		0.001072,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.002895,
		0.018718,
		0.046668,
		0.084923,
		0.131734,
		0.185427,
		0.244406,
		0.307157,
		0.372254,
		0.438357,
		0.504225,
		0.568718,
		0.630801,
		0.689557,
		0.744191,
		0.794042,
		0.838591,
		0.877473,
		0.91049,
		0.937623,
		0.959038,
		0.9751,
		0.986362,
		0.993563,
		0.997583,
		0.999391,
		0.999937,
		1,
		0.999991,
		0.999767,
		0.998703,
		0.995911,
		0.990418,
		0.981289,
		0.967708,
		0.949027,
		0.924786,
		0.89472,
		0.858756,
		0.817006,
		0.769751,
		0.717434,
		0.660647,
		0.60012,
		0.536711,
		0.471394,
		0.405259,
		0.339498,
		0.275401,
		0.214352,
		0.157821,
		0.107365,
		0.064618,
		0.031293,
		0.009174,
		0.000118,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.001035,
		0.01332,
		0.037997,
		0.073275,
		0.117431,
		0.168815,
		0.225849,
		0.287037,
		0.35096,
		0.416289,
		0.481783,
		0.546297,
		0.608789,
		0.668323,
		0.724081,
		0.77537,
		0.82163,
		0.862446,
		0.897559,
		0.926878,
		0.950488,
		0.968661,
		0.981857,
		0.990717,
		0.996042,
		0.998745,
		0.999775,
		0.999992,
		1,
		0.999939,
		0.999412,
		0.997662,
		0.993766,
		0.986782,
		0.975845,
		0.96023,
		0.939384,
		0.912942,
		0.880725,
		0.842738,
		0.799158,
		0.750323,
		0.69672,
		0.638979,
		0.577857,
		0.51423,
		0.449089,
		0.383529,
		0.318742,
		0.256014,
		0.196721,
		0.14232,
		0.094348,
		0.054421,
		0.024223,
		0.005514,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.000114,
		0.008861,
		0.030251,
		0.062522,
		0.103978,
		0.152992,
		0.208006,
		0.26754,
		0.330188,
		0.394628,
		0.459622,
		0.524024,
		0.586783,
		0.646952,
		0.703692,
		0.75628,
		0.804122,
		0.846758,
		0.883873,
		0.915312,
		0.941085,
		0.961379,
		0.976563,
		0.987186,
		0.993962,
		0.997737,
		0.999431,
		0.999941,
		1,
		0.999992,
		0.999782,
		0.998786,
		0.996168,
		0.991005,
		0.982404,
		0.96958,
		0.951898,
		0.928899,
		0.900305,
		0.866019,
		0.826114,
		0.780826,
		0.730545,
		0.675798,
		0.617247,
		0.555671,
		0.491965,
		0.427129,
		0.362263,
		0.298559,
		0.237298,
		0.179843,
		0.127639,
		0.082205,
		0.045134,
		0.018088,
		0.002795,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.005327,
		0.023419,
		0.052658,
		0.091374,
		0.137962,
		0.190888,
		0.248685,
		0.309961,
		0.373403,
		0.437779,
		0.501941,
		0.564834,
		0.625499,
		0.68308,
		0.736831,
		0.786125,
		0.830462,
		0.869479,
		0.902962,
		0.930852,
		0.953258,
		0.970465,
		0.982931,
		0.991281,
		0.996289,
		0.998826,
		0.99979,
		0.999992,
		1,
		0.999943,
		0.99945,
		0.997809,
		0.99415,
		0.987575,
		0.977255,
		0.962487,
		0.942727,
		0.917603,
		0.88692,
		0.850653,
		0.808939,
		0.762069,
		0.710476,
		0.654724,
		0.595501,
		0.533608,
		0.469955,
		0.405548,
		0.34149,
		0.278972,
		0.219266,
		0.163726,
		0.113781,
		0.070933,
		0.036751,
		0.012873,
		0.000999,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.0027,
		0.017488,
		0.043674,
		0.079615,
		0.12373,
		0.174504,
		0.230487,
		0.290302,
		0.352645,
		0.416289,
		0.48009,
		0.542987,
		0.604014,
		0.6623,
		0.717079,
		0.767695,
		0.813614,
		0.854428,
		0.889869,
		0.919818,
		0.944313,
		0.963556,
		0.977922,
		0.987949,
		0.994331,
		0.997879,
		0.999468,
		0.999945,
		1,
		0.999993,
		0.999797,
		0.998863,
		0.996405,
		0.991548,
		0.983438,
		0.971319,
		0.954571,
		0.932739,
		0.905532,
		0.872831,
		0.834679,
		0.79127,
		0.742944,
		0.690172,
		0.63355,
		0.573791,
		0.511714,
		0.448238,
		0.384378,
		0.321236,
		0.259999,
		0.201932,
		0.148377,
		0.100747,
		0.060526,
		0.029261,
		0.008564,
		0.00011,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000966,
		0.012447,
		0.035564,
		0.0687,
		0.110298,
		0.158861,
		0.212959,
		0.27123,
		0.332378,
		0.395185,
		0.458508,
		0.521286,
		0.582546,
		0.641405,
		0.697079,
		0.748889,
		0.796268,
		0.83877,
		0.876079,
		0.908019,
		0.934562,
		0.955839,
		0.972142,
		0.983926,
		0.991804,
		0.996517,
		0.9989,
		0.999803,
		0.999993,
		1,
		0.999947,
		0.999485,
		0.997946,
		0.994505,
		0.98831,
		0.978564,
		0.964588,
		0.945845,
		0.921961,
		0.892725,
		0.858088,
		0.818151,
		0.773163,
		0.723506,
		0.669685,
		0.612327,
		0.552163,
		0.490028,
		0.42685,
		0.363647,
		0.301522,
		0.241656,
		0.185305,
		0.133799,
		0.088536,
		0.050979,
		0.022654,
		0.005149,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.000107,
		0.008282,
		0.028318,
		0.058623,
		0.097664,
		0.143965,
		0.196114,
		0.252761,
		0.312626,
		0.374495,
		0.43723,
		0.499772,
		0.561141,
		0.620445,
		0.676885,
		0.729761,
		0.778478,
		0.822556,
		0.861636,
		0.895491,
		0.924033,
		0.947325,
		0.965583,
		0.979184,
		0.988657,
		0.994673,
		0.99801,
		0.999501,
		0.999948,
		1,
		0.999993,
		0.99981,
		0.998934,
		0.996625,
		0.99205,
		0.984397,
		0.972936,
		0.957064,
		0.936326,
		0.910426,
		0.879226,
		0.842738,
		0.801123,
		0.754672,
		0.703808,
		0.649069,
		0.591103,
		0.530661,
		0.468589,
		0.405822,
		0.343382,
		0.282369,
		0.223957,
		0.169394,
		0.119997,
		0.077145,
		0.042284,
		0.016918,
		0.00261,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.004979,
		0.021925,
		0.049379,
		0.085828,
		0.12982,
		0.179959,
		0.234912,
		0.293408,
		0.354245,
		0.416289,
		0.478481,
		0.53984,
		0.599465,
		0.656545,
		0.710363,
		0.760297,
		0.805839,
		0.846589,
		0.882275,
		0.912756,
		0.938031,
		0.958246,
		0.973702,
		0.984851,
		0.992288,
		0.996728,
		0.998968,
		0.999816,
		0.999993,
		1,
		0.99995,
		0.999517,
		0.998072,
		0.994834,
		0.988991,
		0.979781,
		0.966545,
		0.948756,
		0.926039,
		0.89817,
		0.865077,
		0.826833,
		0.783645,
		0.73585,
		0.683903,
		0.628371,
		0.569922,
		0.509323,
		0.44743,
		0.385184,
		0.323607,
		0.263794,
		0.206915,
		0.154206,
		0.10697,
		0.066571,
		0.034434,
		0.012042,
		0.000934,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.002525,
		0.016376,
		0.040959,
		0.074788,
		0.116428,
		0.164503,
		0.217694,
		0.274744,
		0.33446,
		0.395714,
		0.45745,
		0.518684,
		0.578511,
		0.636109,
		0.690745,
		0.741777,
		0.788668,
		0.830985,
		0.868415,
		0.900765,
		0.927979,
		0.950139,
		0.967473,
		0.980357,
		0.989313,
		0.994989,
		0.998131,
		0.999532,
		0.999952,
		1,
		0.999993,
		0.999822,
		0.999,
		0.996827,
		0.992516,
		0.985289,
		0.974442,
		0.959389,
		0.93968,
		0.915012,
		0.885232,
		0.850326,
		0.810421,
		0.76577,
		0.716747,
		0.66384,
		0.607637,
		0.548826,
		0.488188,
		0.426585,
		0.364963,
		0.304341,
		0.245813,
		0.190539,
		0.139745,
		0.094719,
		0.056809,
		0.02742,
		0.008014,
		0.000103,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000903,
		0.011657,
		0.033357,
		0.064539,
		0.103789,
		0.14975,
		0.201118,
		0.25665,
		0.315161,
		0.375532,
		0.436709,
		0.497709,
		0.557624,
		0.615622,
		0.670956,
		0.722968,
		0.771094,
		0.814873,
		0.853952,
		0.888097,
		0.917197,
		0.941275,
		0.960493,
		0.975156,
		0.985711,
		0.992736,
		0.996923,
		0.999031,
		0.999827,
		0.999994,
		1,
		0.999953,
		0.999547,
		0.998188,
		0.995138,
		0.989624,
		0.980913,
		0.968369,
		0.951476,
		0.929857,
		0.90328,
		0.871652,
		0.835018,
		0.793551,
		0.747548,
		0.697414,
		0.643665,
		0.586911,
		0.527856,
		0.46729,
		0.406083,
		0.345181,
		0.285604,
		0.228439,
		0.174837,
		0.126014,
		0.083242,
		0.047852,
		0.021232,
		0.004818,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.0001,
		0.007758,
		0.026564,
		0.055077,
		0.091904,
		0.135705,
		0.185194,
		0.239139,
		0.296367,
		0.355767,
		0.416289,
		0.476952,
		0.536843,
		0.595126,
		0.651042,
		0.703917,
		0.753166,
		0.7983,
		0.838934,
		0.874792,
		0.905717,
		0.931675,
		0.952769,
		0.969235,
		0.98145,
		0.989923,
		0.995282,
		0.998243,
		0.999561,
		0.999955,
		1,
		0.999994,
		0.999833,
		0.999061,
		0.997015,
		0.992949,
		0.986118,
		0.975845,
		0.96156,
		0.942819,
		0.919314,
		0.890876,
		0.857473,
		0.8192,
		0.776274,
		0.729027,
		0.677898,
		0.623424,
		0.566235,
		0.507048,
		0.446662,
		0.385951,
		0.325863,
		0.267414,
		0.211684,
		0.159817,
		0.113014,
		0.072536,
		0.039695,
		0.015859,
		0.002443,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.004665,
		0.02057,
		0.046396,
		0.08077,
		0.122371,
		0.169927,
		0.222223,
		0.278096,
		0.336442,
		0.396218,
		0.456443,
		0.516206,
		0.574663,
		0.631048,
		0.684672,
		0.734932,
		0.781314,
		0.823405,
		0.86089,
		0.893571,
		0.921364,
		0.944313,
		0.962592,
		0.976511,
		0.98651,
		0.993153,
		0.997104,
		0.999089,
		0.999838,
		0.999994,
		1,
		0.999956,
		0.999575,
		0.998296,
		0.995421,
		0.990212,
		0.981968,
		0.970072,
		0.95402,
		0.933436,
		0.908079,
		0.877839,
		0.842738,
		0.802918,
		0.758636,
		0.710256,
		0.658245,
		0.60316,
		0.545648,
		0.486437,
		0.426333,
		0.366214,
		0.307026,
		0.249783,
		0.195557,
		0.145482,
		0.100747,
		0.062598,
		0.03233,
		0.01129,
		0.000874,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		0.002366,
		0.015366,
		0.038489,
		0.070384,
		0.109748,
		0.155325,
		0.205913,
		0.260362,
		0.317576,
		0.376519,
		0.436213,
		0.495746,
		0.554272,
		0.611015,
		0.665277,
		0.716437,
		0.763962,
		0.80741,
		0.846434,
		0.880796,
		0.910368,
		0.935141,
		0.95523,
		0.970881,
		0.982468,
		0.99049,
		0.995554,
		0.998347,
		0.999588,
		0.999957,
		1,
		0.999994,
		0.999843,
		0.999117,
		0.997189,
		0.99335,
		0.986889,
		0.977154,
		0.96359,
		0.945759,
		0.92335,
		0.896184,
		0.864209,
		0.827492,
		0.78622,
		0.740684,
		0.69128,
		0.638497,
		0.582912,
		0.525185,
		0.466054,
		0.406331,
		0.346894,
		0.28869,
		0.232725,
		0.180066,
		0.131836,
		0.089212,
		0.053423,
		0.025748,
		0.007514,
		9.6e-005,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.000846,
		0.01094,
		0.031349,
		0.060743,
		0.097836,
		0.141391,
		0.190218,
		0.243179,
		0.299189,
		0.357216,
		0.416289,
		0.475496,
		0.533987,
		0.590984,
		0.645775,
		0.697728,
		0.746289,
		0.790993,
		0.831466,
		0.867431,
		0.89872,
		0.925275,
		0.947158,
		0.964555,
		0.977776,
		0.987255,
		0.993541,
		0.997272,
		0.999143,
		0.999848,
		0.999994,
		1,
		0.999959,
		0.9996,
		0.998396,
		0.995683,
		0.990759,
		0.982951,
		0.971663,
		0.956401,
		0.936792,
		0.912588,
		0.883666,
		0.850025,
		0.811778,
		0.769149,
		0.722464,
		0.672143,
		0.618696,
		0.562719,
		0.504881,
		0.44593,
		0.386681,
		0.328012,
		0.270869,
		0.216251,
		0.165218,
		0.118882,
		0.078406,
		0.045004,
		0.019939,
		0.004519,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		9.3e-005,
		0.007282,
		0.024968,
		0.051842,
		0.086636,
		0.128128,
		0.175145,
		0.226559,
		0.281295,
		0.33833,
		0.396697,
		0.455485,
		0.513845,
		0.570991,
		0.626207,
		0.678846,
		0.72834,
		0.774201,
		0.816028,
		0.853514,
		0.886452,
		0.914741,
		0.938392,
		0.957534,
		0.972419,
		0.983418,
		0.991018,
		0.995807,
		0.998443,
		0.999612,
		0.99996,
		1,
		0.999995,
		0.999852,
		0.999169,
		0.997351,
		0.993724,
		0.987608,
		0.978376,
		0.965488,
		0.948514,
		0.927141,
		0.90118,
		0.870561,
		0.835329,
		0.79564,
		0.751752,
		0.704019,
		0.652886,
		0.598882,
		0.542616,
		0.484769,
		0.426093,
		0.367405,
		0.309587,
		0.253577,
		0.200371,
		0.151018,
		0.106619,
		0.068326,
		0.037337,
		0.014896,
		0.002292,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.004379,
		0.019337,
		0.043674,
		0.076143,
		0.115538,
		0.160699,
		0.210511,
		0.26391,
		0.31988,
		0.377459,
		0.435741,
		0.493875,
		0.551073,
		0.606611,
		0.659834,
		0.710157,
		0.757075,
		0.800163,
		0.839085,
		0.873601,
		0.903567,
		0.92895,
		0.949827,
		0.966392,
		0.978957,
		0.987949,
		0.993901,
		0.997428,
		0.999193,
		0.999857,
		0.999995,
		1,
		0.999961,
		0.999624,
		0.998488,
		0.995927,
		0.991269,
		0.983869,
		0.97315,
		0.958631,
		0.939942,
		0.916829,
		0.889157,
		0.856905,
		0.820162,
		0.779121,
		0.73407,
		0.685391,
		0.63355,
		0.579092,
		0.522638,
		0.464877,
		0.406567,
		0.348527,
		0.291635,
		0.236827,
		0.185091,
		0.137468,
		0.095049,
		0.058969,
		0.030412,
		0.010606,
		0.00082,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		0.002221,
		0.014447,
		0.036235,
		0.066356,
		0.103621,
		0.146884,
		0.195044,
		0.247045,
		0.301882,
		0.358598,
		0.416289,
		0.474108,
		0.531263,
		0.587025,
		0.64073,
		0.691782,
		0.739658,
		0.783913,
		0.824185,
		0.860201,
		0.891784,
		0.918855,
		0.941445,
		0.959694,
		0.973858,
		0.984305,
		0.99151,
		0.996042,
		0.998532,
		0.999635,
		0.999962,
		1,
		0.999995,
		0.999861,
		0.999217,
		0.997501,
		0.994072,
		0.988279,
		0.979518,
		0.967266,
		0.951098,
		0.930704,
		0.905884,
		0.876554,
		0.842738,
		0.804566,
		0.762263,
		0.716147,
		0.666622,
		0.614173,
		0.559361,
		0.502815,
		0.445233,
		0.387376,
		0.330063,
		0.27417,
		0.220628,
		0.170419,
		0.124573,
		0.084169,
		0.05033,
		0.024223,
		0.00706,
		9.1e-005,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.005605,
		0.021564,
		0.046619,
		0.079556,
		0.119202,
		0.164428,
		0.214149,
		0.267329,
		0.32298,
		0.380165,
		0.438,
		0.495659,
		0.552374,
		0.607439,
		0.660215,
		0.710133,
		0.756699,
		0.799498,
		0.838201,
		0.872571,
		0.902465,
		0.927846,
		0.948785,
		0.965467,
		0.978187,
		0.987358,
		0.99349,
		0.99718,
		0.999074,
		0.999819,
		0.999991,
		1,
		0.999975,
		0.999707,
		0.998733,
		0.996452,
		0.99221,
		0.985372,
		0.975358,
		0.961681,
		0.943959,
		0.921922,
		0.895415,
		0.864394,
		0.828924,
		0.789171,
		0.745396,
		0.69795,
		0.647268,
		0.593863,
		0.538323,
		0.481304,
		0.423526,
		0.365774,
		0.308888,
		0.253765,
		0.201355,
		0.152659,
		0.108724,
		0.070648,
		0.039569,
		0.016673,
		0.003183,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.005692,
		0.021892,
		0.047314,
		0.080713,
		0.120891,
		0.166693,
		0.217011,
		0.270786,
		0.327009,
		0.384724,
		0.443032,
		0.501091,
		0.55812,
		0.613403,
		0.666294,
		0.716218,
		0.762678,
		0.805262,
		0.843645,
		0.877597,
		0.90699,
		0.931803,
		0.952128,
		0.968173,
		0.980266,
		0.988848,
		0.994461,
		0.997731,
		0.999323,
		0.999892,
		0.999997,
		1,
		0.999944,
		0.999537,
		0.99826,
		0.995473,
		0.990506,
		0.982721,
		0.971552,
		0.95653,
		0.937299,
		0.913621,
		0.885375,
		0.852555,
		0.815262,
		0.773702,
		0.728175,
		0.679074,
		0.626874,
		0.572131,
		0.515474,
		0.457602,
		0.39928,
		0.341334,
		0.28465,
		0.230171,
		0.178891,
		0.131859,
		0.090168,
		0.054963,
		0.027432,
		0.008808,
		0.000366,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.005782,
		0.022228,
		0.048024,
		0.081895,
		0.122616,
		0.169005,
		0.219929,
		0.274307,
		0.331109,
		0.38936,
		0.448143,
		0.5066,
		0.563939,
		0.619433,
		0.672427,
		0.722343,
		0.768681,
		0.811031,
		0.849074,
		0.882588,
		0.91146,
		0.935686,
		0.955381,
		0.97078,
		0.98224,
		0.990235,
		0.99534,
		0.998208,
		0.999523,
		0.999942,
		1,
		0.999997,
		0.999889,
		0.999303,
		0.997663,
		0.9943,
		0.988531,
		0.97972,
		0.967315,
		0.950871,
		0.930061,
		0.90468,
		0.874642,
		0.839981,
		0.800837,
		0.757456,
		0.710181,
		0.659446,
		0.60577,
		0.54975,
		0.492061,
		0.433445,
		0.374712,
		0.316735,
		0.260446,
		0.206833,
		0.156939,
		0.111859,
		0.072738,
		0.040769,
		0.01719,
		0.003284,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.005873,
		0.022572,
		0.04875,
		0.083104,
		0.124377,
		0.171364,
		0.222905,
		0.277895,
		0.335282,
		0.394074,
		0.453333,
		0.512188,
		0.569832,
		0.625529,
		0.678615,
		0.728508,
		0.774707,
		0.816804,
		0.854485,
		0.887541,
		0.91587,
		0.939491,
		0.95854,
		0.973282,
		0.984106,
		0.991519,
		0.996128,
		0.998614,
		0.999679,
		0.999973,
		1,
		0.99999,
		0.999801,
		0.998987,
		0.996922,
		0.992908,
		0.986257,
		0.976338,
		0.962616,
		0.944673,
		0.922214,
		0.895068,
		0.863189,
		0.826649,
		0.78563,
		0.740422,
		0.691409,
		0.639069,
		0.583966,
		0.526742,
		0.468117,
		0.40888,
		0.349886,
		0.292057,
		0.236371,
		0.183866,
		0.135635,
		0.092821,
		0.056622,
		0.02828,
		0.009086,
		0.000378,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.005967,
		0.022924,
		0.049493,
		0.084339,
		0.126177,
		0.173771,
		0.225939,
		0.28155,
		0.339531,
		0.398867,
		0.458604,
		0.517854,
		0.575799,
		0.631689,
		0.684856,
		0.734711,
		0.780754,
		0.822577,
		0.859876,
		0.89245,
		0.920217,
		0.943213,
		0.961602,
		0.975678,
		0.985864,
		0.9927,
		0.996829,
		0.998956,
		0.999795,
		0.999989,
		1,
		0.999972,
		0.999668,
		0.998572,
		0.996013,
		0.991272,
		0.983655,
		0.972543,
		0.957423,
		0.937904,
		0.913726,
		0.884756,
		0.850988,
		0.812535,
		0.769624,
		0.722587,
		0.671853,
		0.617947,
		0.561477,
		0.503133,
		0.443682,
		0.383959,
		0.32487,
		0.267385,
		0.212531,
		0.161399,
		0.11513,
		0.074922,
		0.042023,
		0.017731,
		0.00339,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.006063,
		0.023285,
		0.050253,
		0.085601,
		0.128015,
		0.17623,
		0.229035,
		0.285276,
		0.343856,
		0.403741,
		0.463957,
		0.523601,
		0.58184,
		0.637915,
		0.69115,
		0.740951,
		0.786818,
		0.828347,
		0.865241,
		0.897313,
		0.924495,
		0.946848,
		0.964563,
		0.977965,
		0.987511,
		0.99378,
		0.997444,
		0.999236,
		0.999878,
		0.999997,
		1,
		0.999936,
		0.999477,
		0.998039,
		0.994912,
		0.989361,
		0.980692,
		0.968302,
		0.951701,
		0.930531,
		0.904566,
		0.873716,
		0.838015,
		0.79762,
		0.752804,
		0.703943,
		0.651515,
		0.596089,
		0.538323,
		0.478955,
		0.4188,
		0.358744,
		0.299743,
		0.242817,
		0.189047,
		0.139573,
		0.095593,
		0.058356,
		0.029167,
		0.009378,
		0.00039,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.006161,
		0.023653,
		0.05103,
		0.086892,
		0.129893,
		0.178739,
		0.232193,
		0.289073,
		0.34826,
		0.408697,
		0.469393,
		0.529428,
		0.587956,
		0.644206,
		0.697495,
		0.747226,
		0.792898,
		0.834112,
		0.870578,
		0.902123,
		0.9287,
		0.950392,
		0.967419,
		0.98014,
		0.989048,
		0.994759,
		0.997978,
		0.999461,
		0.999934,
		1,
		0.999997,
		0.999874,
		0.999212,
		0.997366,
		0.993593,
		0.987147,
		0.977338,
		0.963582,
		0.945417,
		0.922519,
		0.894703,
		0.861918,
		0.824244,
		0.781884,
		0.735156,
		0.684484,
		0.630396,
		0.573509,
		0.514529,
		0.454246,
		0.393526,
		0.333306,
		0.274594,
		0.218463,
		0.166048,
		0.118545,
		0.077206,
		0.043337,
		0.018298,
		0.003501,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.006261,
		0.024031,
		0.051826,
		0.088212,
		0.131812,
		0.181302,
		0.235414,
		0.292943,
		0.352743,
		0.413737,
		0.474914,
		0.535338,
		0.594146,
		0.650561,
		0.70389,
		0.753533,
		0.79899,
		0.839866,
		0.875881,
		0.906878,
		0.932827,
		0.95384,
		0.970166,
		0.9822,
		0.990474,
		0.995638,
		0.998435,
		0.999636,
		0.999969,
		1,
		0.999988,
		0.999775,
		0.998854,
		0.996529,
		0.992029,
		0.984597,
		0.973558,
		0.958348,
		0.938536,
		0.913836,
		0.884104,
		0.849334,
		0.809651,
		0.765307,
		0.716668,
		0.664208,
		0.608503,
		0.550223,
		0.490126,
		0.429052,
		0.36792,
		0.307723,
		0.249522,
		0.194445,
		0.143683,
		0.098489,
		0.060171,
		0.030097,
		0.009684,
		0.000403,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.006364,
		0.024418,
		0.05264,
		0.089563,
		0.133774,
		0.183919,
		0.238701,
		0.296887,
		0.357308,
		0.418863,
		0.480522,
		0.54133,
		0.600413,
		0.656981,
		0.710335,
		0.759872,
		0.805092,
		0.845607,
		0.881147,
		0.911571,
		0.936872,
		0.957188,
		0.972801,
		0.984144,
		0.991788,
		0.996422,
		0.998818,
		0.999767,
		0.999988,
		1,
		0.999968,
		0.999624,
		0.998385,
		0.995504,
		0.990188,
		0.98168,
		0.969317,
		0.952563,
		0.931022,
		0.904447,
		0.872739,
		0.835936,
		0.794215,
		0.747875,
		0.697333,
		0.643115,
		0.585848,
		0.526257,
		0.465151,
		0.403426,
		0.342055,
		0.282087,
		0.224639,
		0.170898,
		0.122113,
		0.079594,
		0.044712,
		0.018893,
		0.003617,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.00647,
		0.024814,
		0.053473,
		0.090944,
		0.135779,
		0.186592,
		0.242056,
		0.300909,
		0.361957,
		0.424076,
		0.486216,
		0.547405,
		0.606755,
		0.663465,
		0.716828,
		0.76624,
		0.811201,
		0.851331,
		0.886372,
		0.916199,
		0.94083,
		0.960431,
		0.975321,
		0.98597,
		0.992991,
		0.997111,
		0.999134,
		0.999861,
		0.999997,
		1,
		0.999927,
		0.999407,
		0.997782,
		0.994262,
		0.988039,
		0.978359,
		0.964579,
		0.946192,
		0.92284,
		0.894319,
		0.860575,
		0.821699,
		0.777914,
		0.729573,
		0.677145,
		0.621209,
		0.562448,
		0.501639,
		0.439648,
		0.377429,
		0.31601,
		0.256499,
		0.200072,
		0.147975,
		0.101517,
		0.062072,
		0.031071,
		0.010005,
		0.000417,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.006578,
		0.02522,
		0.054327,
		0.092357,
		0.137829,
		0.189322,
		0.245479,
		0.305009,
		0.366691,
		0.429378,
		0.491999,
		0.553565,
		0.613173,
		0.670011,
		0.723368,
		0.772634,
		0.817314,
		0.857034,
		0.891549,
		0.920755,
		0.944695,
		0.963565,
		0.977722,
		0.987677,
		0.994084,
		0.997711,
		0.999388,
		0.999925,
		1,
		0.999997,
		0.999857,
		0.999106,
		0.99702,
		0.992774,
		0.985548,
		0.974599,
		0.959306,
		0.939196,
		0.913952,
		0.883417,
		0.847584,
		0.806596,
		0.76073,
		0.71039,
		0.656102,
		0.5985,
		0.538323,
		0.476406,
		0.413672,
		0.351133,
		0.289878,
		0.231074,
		0.17596,
		0.125842,
		0.082095,
		0.046154,
		0.019517,
		0.003739,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.006689,
		0.025636,
		0.0552,
		0.093803,
		0.139926,
		0.192112,
		0.248973,
		0.309189,
		0.371512,
		0.43477,
		0.497872,
		0.55981,
		0.619667,
		0.67662,
		0.729952,
		0.779051,
		0.823426,
		0.862711,
		0.896676,
		0.925236,
		0.948463,
		0.966586,
		0.980001,
		0.989263,
		0.995069,
		0.998225,
		0.999586,
		0.999965,
		1,
		0.999987,
		0.999743,
		0.9987,
		0.996073,
		0.991008,
		0.982681,
		0.970362,
		0.953459,
		0.931537,
		0.904322,
		0.871707,
		0.833736,
		0.790604,
		0.742645,
		0.690319,
		0.634207,
		0.575003,
		0.513503,
		0.450601,
		0.387283,
		0.324619,
		0.263763,
		0.205942,
		0.152459,
		0.104687,
		0.064064,
		0.032094,
		0.010342,
		0.000431,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.006803,
		0.026062,
		0.056095,
		0.095283,
		0.142069,
		0.194962,
		0.25254,
		0.313452,
		0.376422,
		0.440255,
		0.503836,
		0.56614,
		0.626236,
		0.683291,
		0.736579,
		0.78549,
		0.829535,
		0.868358,
		0.901745,
		0.929636,
		0.952128,
		0.969489,
		0.982156,
		0.990728,
		0.995947,
		0.998657,
		0.999735,
		0.999986,
		1,
		0.999963,
		0.999572,
		0.998167,
		0.994912,
		0.988931,
		0.9794,
		0.965609,
		0.946999,
		0.923176,
		0.893914,
		0.859156,
		0.819002,
		0.773702,
		0.723646,
		0.669354,
		0.611467,
		0.550737,
		0.488022,
		0.424278,
		0.360554,
		0.297982,
		0.237781,
		0.181245,
		0.129743,
		0.084714,
		0.047667,
		0.020173,
		0.003868,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.00692,
		0.026499,
		0.057012,
		0.096799,
		0.144262,
		0.197875,
		0.256181,
		0.317799,
		0.381424,
		0.445833,
		0.509893,
		0.572558,
		0.632882,
		0.690022,
		0.743248,
		0.791947,
		0.835636,
		0.87397,
		0.906753,
		0.933948,
		0.955686,
		0.972271,
		0.984185,
		0.992073,
		0.996723,
		0.999014,
		0.999842,
		0.999996,
		1,
		0.999917,
		0.999324,
		0.997481,
		0.993506,
		0.986506,
		0.975665,
		0.960299,
		0.939885,
		0.914074,
		0.88269,
		0.845732,
		0.803355,
		0.755868,
		0.703721,
		0.647496,
		0.587893,
		0.525728,
		0.461923,
		0.397498,
		0.333566,
		0.271328,
		0.212067,
		0.157147,
		0.108006,
		0.066153,
		0.033168,
		0.010696,
		0.000447,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.00704,
		0.026947,
		0.057952,
		0.09835,
		0.146505,
		0.200852,
		0.2599,
		0.322233,
		0.386518,
		0.451507,
		0.516043,
		0.579062,
		0.639602,
		0.696813,
		0.749955,
		0.798418,
		0.841725,
		0.879542,
		0.911694,
		0.938168,
		0.959131,
		0.974928,
		0.986085,
		0.993298,
		0.997398,
		0.999302,
		0.999914,
		0.999999,
		0.999996,
		0.999836,
		0.998981,
		0.996615,
		0.991821,
		0.983695,
		0.971437,
		0.954391,
		0.932076,
		0.904191,
		0.870615,
		0.831403,
		0.78677,
		0.737086,
		0.682865,
		0.624748,
		0.563503,
		0.500011,
		0.435259,
		0.370333,
		0.306416,
		0.244776,
		0.186768,
		0.133826,
		0.087461,
		0.049255,
		0.020862,
		0.004003,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.007163,
		0.027406,
		0.058915,
		0.099939,
		0.148801,
		0.203896,
		0.263696,
		0.326755,
		0.391707,
		0.457278,
		0.522288,
		0.585653,
		0.646399,
		0.703661,
		0.756699,
		0.804901,
		0.847798,
		0.88507,
		0.916561,
		0.94229,
		0.962458,
		0.977455,
		0.987855,
		0.994403,
		0.997978,
		0.999527,
		0.999959,
		1,
		0.999985,
		0.999707,
		0.998518,
		0.995539,
		0.989821,
		0.980459,
		0.966673,
		0.947842,
		0.923529,
		0.893487,
		0.857652,
		0.816138,
		0.769224,
		0.717343,
		0.661072,
		0.601121,
		0.538323,
		0.473628,
		0.40809,
		0.342867,
		0.27921,
		0.218463,
		0.162051,
		0.111483,
		0.068345,
		0.034297,
		0.011069,
		0.000462,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.007289,
		0.027878,
		0.059902,
		0.101566,
		0.151149,
		0.207007,
		0.267574,
		0.331368,
		0.396993,
		0.463149,
		0.528629,
		0.592333,
		0.65327,
		0.710566,
		0.763477,
		0.811392,
		0.85385,
		0.890546,
		0.92135,
		0.946308,
		0.965663,
		0.97985,
		0.989494,
		0.995392,
		0.998468,
		0.999696,
		0.999984,
		1,
		0.999958,
		0.99951,
		0.99791,
		0.99422,
		0.987468,
		0.976757,
		0.961329,
		0.940607,
		0.914202,
		0.881923,
		0.843766,
		0.79991,
		0.750696,
		0.696627,
		0.638345,
		0.57663,
		0.512384,
		0.446628,
		0.380488,
		0.315194,
		0.252073,
		0.192542,
		0.138102,
		0.090341,
		0.050923,
		0.021587,
		0.004145,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.007418,
		0.028361,
		0.060913,
		0.103233,
		0.153554,
		0.210189,
		0.271535,
		0.336073,
		0.402378,
		0.469119,
		0.535067,
		0.599101,
		0.660215,
		0.717527,
		0.770286,
		0.817887,
		0.859876,
		0.895967,
		0.926053,
		0.950217,
		0.968741,
		0.982109,
		0.991001,
		0.996266,
		0.998873,
		0.999819,
		0.999996,
		0.999999,
		0.999905,
		0.999227,
		0.997128,
		0.992621,
		0.984721,
		0.972543,
		0.955361,
		0.932642,
		0.904051,
		0.869459,
		0.828924,
		0.78269,
		0.73117,
		0.674931,
		0.61469,
		0.551297,
		0.485726,
		0.419073,
		0.352538,
		0.287427,
		0.225145,
		0.167184,
		0.11513,
		0.070648,
		0.035484,
		0.011462,
		0.000479,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.007551,
		0.028857,
		0.061951,
		0.104941,
		0.156015,
		0.213443,
		0.27558,
		0.340874,
		0.407865,
		0.475193,
		0.541604,
		0.605958,
		0.667234,
		0.72454,
		0.777124,
		0.824381,
		0.865871,
		0.901325,
		0.930665,
		0.954009,
		0.971686,
		0.98423,
		0.992377,
		0.99703,
		0.9992,
		0.999901,
		0.999999,
		0.999996,
		0.999812,
		0.998834,
		0.99614,
		0.990705,
		0.981538,
		0.967772,
		0.948721,
		0.9239,
		0.893035,
		0.856057,
		0.813093,
		0.764456,
		0.710629,
		0.652255,
		0.59012,
		0.525152,
		0.4584,
		0.391034,
		0.324337,
		0.259691,
		0.198581,
		0.142584,
		0.093366,
		0.052677,
		0.02235,
		0.004296,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.007688,
		0.029367,
		0.063015,
		0.106691,
		0.158534,
		0.21677,
		0.279714,
		0.345773,
		0.413455,
		0.48137,
		0.54824,
		0.612904,
		0.674326,
		0.731605,
		0.783986,
		0.83087,
		0.871829,
		0.906615,
		0.935179,
		0.957681,
		0.974494,
		0.98621,
		0.993622,
		0.997688,
		0.999457,
		0.999953,
		1,
		0.999982,
		0.999663,
		0.998304,
		0.994912,
		0.988431,
		0.977873,
		0.962398,
		0.941362,
		0.914337,
		0.881109,
		0.841678,
		0.796242,
		0.745185,
		0.689066,
		0.628601,
		0.564656,
		0.498233,
		0.430464,
		0.362598,
		0.295997,
		0.232129,
		0.172561,
		0.118957,
		0.073068,
		0.036734,
		0.011876,
		0.000497,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.007828,
		0.02989,
		0.064107,
		0.108485,
		0.161115,
		0.220175,
		0.283937,
		0.350772,
		0.41915,
		0.487653,
		0.554977,
		0.619939,
		0.681489,
		0.738718,
		0.790869,
		0.83735,
		0.877744,
		0.91183,
		0.93959,
		0.961225,
		0.977161,
		0.988048,
		0.994738,
		0.998244,
		0.999651,
		0.999982,
		1,
		0.999951,
		0.999437,
		0.997607,
		0.993406,
		0.985756,
		0.973679,
		0.956371,
		0.933236,
		0.903905,
		0.868232,
		0.826287,
		0.778343,
		0.72486,
		0.666474,
		0.603979,
		0.538323,
		0.470589,
		0.40199,
		0.333861,
		0.267647,
		0.204903,
		0.147285,
		0.096544,
		0.054523,
		0.023155,
		0.004454,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.007973,
		0.030427,
		0.065227,
		0.110324,
		0.163757,
		0.223658,
		0.288253,
		0.355873,
		0.424954,
		0.494044,
		0.561815,
		0.627063,
		0.688723,
		0.745878,
		0.797771,
		0.843815,
		0.883611,
		0.916963,
		0.943889,
		0.964635,
		0.979683,
		0.989742,
		0.995728,
		0.998706,
		0.999791,
		0.999995,
		0.999999,
		0.99989,
		0.999112,
		0.996711,
		0.991581,
		0.982633,
		0.968908,
		0.94964,
		0.924291,
		0.892557,
		0.854361,
		0.809849,
		0.759371,
		0.703467,
		0.642853,
		0.578407,
		0.51116,
		0.442279,
		0.373065,
		0.304937,
		0.239433,
		0.178197,
		0.122976,
		0.075614,
		0.038051,
		0.012313,
		0.000516,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.008121,
		0.030978,
		0.066377,
		0.11221,
		0.166465,
		0.227222,
		0.292664,
		0.361079,
		0.430867,
		0.500544,
		0.568755,
		0.634275,
		0.696026,
		0.753082,
		0.804685,
		0.850259,
		0.889423,
		0.922008,
		0.948071,
		0.967907,
		0.982057,
		0.991293,
		0.996595,
		0.999079,
		0.999886,
		0.999999,
		0.999995,
		0.999783,
		0.998659,
		0.995579,
		0.989394,
		0.979014,
		0.963506,
		0.942154,
		0.914479,
		0.880246,
		0.839455,
		0.79233,
		0.739302,
		0.680995,
		0.618209,
		0.551909,
		0.483212,
		0.413374,
		0.343786,
		0.27596,
		0.211524,
		0.152218,
		0.099885,
		0.056468,
		0.024004,
		0.004622,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.008273,
		0.031545,
		0.067557,
		0.114144,
		0.169238,
		0.230869,
		0.297172,
		0.366393,
		0.436892,
		0.507156,
		0.575799,
		0.641577,
		0.703397,
		0.760326,
		0.811609,
		0.856678,
		0.895174,
		0.926957,
		0.952128,
		0.971035,
		0.984278,
		0.9927,
		0.997343,
		0.999373,
		0.999946,
		1,
		0.99998,
		0.999611,
		0.998049,
		0.994171,
		0.986798,
		0.974848,
		0.957423,
		0.93386,
		0.903749,
		0.866927,
		0.823476,
		0.773702,
		0.71812,
		0.657442,
		0.592557,
		0.524519,
		0.454537,
		0.383959,
		0.31427,
		0.247077,
		0.184108,
		0.127199,
		0.078295,
		0.039439,
		0.012775,
		0.000536,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.00843,
		0.032127,
		0.068768,
		0.116129,
		0.172081,
		0.234603,
		0.301781,
		0.371816,
		0.443032,
		0.51388,
		0.582946,
		0.648967,
		0.710833,
		0.767608,
		0.818536,
		0.863065,
		0.900855,
		0.931803,
		0.956054,
		0.974012,
		0.986346,
		0.993965,
		0.997978,
		0.999596,
		0.999979,
		1,
		0.999944,
		0.99935,
		0.997247,
		0.992444,
		0.983744,
		0.970081,
		0.950601,
		0.924704,
		0.892049,
		0.852555,
		0.806386,
		0.753936,
		0.69581,
		0.63281,
		0.565918,
		0.496282,
		0.425205,
		0.354133,
		0.28465,
		0.218463,
		0.1574,
		0.103402,
		0.058518,
		0.0249,
		0.004799,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.008592,
		0.032726,
		0.070013,
		0.118164,
		0.174994,
		0.238425,
		0.306492,
		0.377353,
		0.449289,
		0.520718,
		0.590198,
		0.656444,
		0.718333,
		0.774923,
		0.825463,
		0.869413,
		0.90646,
		0.936538,
		0.959841,
		0.976835,
		0.988257,
		0.99509,
		0.998506,
		0.999758,
		0.999994,
		0.999999,
		0.999873,
		0.998974,
		0.996216,
		0.990352,
		0.980179,
		0.964657,
		0.942986,
		0.91463,
		0.879329,
		0.837084,
		0.78815,
		0.73301,
		0.672365,
		0.607109,
		0.538323,
		0.467253,
		0.395301,
		0.324015,
		0.255081,
		0.190312,
		0.131641,
		0.081119,
		0.040905,
		0.013263,
		0.000557,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.008758,
		0.033341,
		0.071291,
		0.120254,
		0.177981,
		0.242338,
		0.31131,
		0.383005,
		0.455665,
		0.527672,
		0.597555,
		0.664008,
		0.725895,
		0.782268,
		0.832383,
		0.875716,
		0.911982,
		0.941155,
		0.963484,
		0.979499,
		0.990011,
		0.996078,
		0.998935,
		0.999868,
		0.999999,
		0.999994,
		0.999749,
		0.998451,
		0.994912,
		0.987845,
		0.976047,
		0.958519,
		0.934518,
		0.903584,
		0.865538,
		0.820474,
		0.768736,
		0.710906,
		0.647781,
		0.580358,
		0.509814,
		0.437501,
		0.364925,
		0.29374,
		0.22574,
		0.162846,
		0.107106,
		0.060681,
		0.025848,
		0.004986,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.008928,
		0.033974,
		0.072605,
		0.122398,
		0.181043,
		0.246346,
		0.316237,
		0.388776,
		0.462163,
		0.534743,
		0.605018,
		0.671658,
		0.733515,
		0.789639,
		0.839291,
		0.881966,
		0.917412,
		0.945646,
		0.966975,
		0.981999,
		0.991607,
		0.996932,
		0.999274,
		0.999937,
		1,
		0.999976,
		0.999548,
		0.997745,
		0.993291,
		0.984868,
		0.971292,
		0.951607,
		0.92514,
		0.89151,
		0.850628,
		0.802682,
		0.748116,
		0.687611,
		0.622067,
		0.552583,
		0.480445,
		0.407111,
		0.334196,
		0.263466,
		0.196828,
		0.136317,
		0.084099,
		0.042454,
		0.013779,
		0.000579,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.009104,
		0.034625,
		0.073955,
		0.1246,
		0.184184,
		0.25045,
		0.321275,
		0.394668,
		0.468785,
		0.541933,
		0.612586,
		0.679392,
		0.741192,
		0.797031,
		0.84618,
		0.888157,
		0.922742,
		0.950003,
		0.970307,
		0.984332,
		0.993045,
		0.99766,
		0.999531,
		0.999975,
		1,
		0.999934,
		0.999246,
		0.996818,
		0.991302,
		0.981366,
		0.965851,
		0.943859,
		0.91479,
		0.878352,
		0.834551,
		0.783673,
		0.726267,
		0.663118,
		0.595234,
		0.523824,
		0.450283,
		0.376183,
		0.303253,
		0.233376,
		0.168575,
		0.11101,
		0.062966,
		0.02685,
		0.005185,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.009285,
		0.035295,
		0.075343,
		0.126861,
		0.187405,
		0.254655,
		0.326429,
		0.400685,
		0.475533,
		0.549244,
		0.620261,
		0.68721,
		0.748921,
		0.804438,
		0.853044,
		0.894279,
		0.927963,
		0.954217,
		0.973475,
		0.986495,
		0.994328,
		0.998267,
		0.999718,
		0.999993,
		0.999999,
		0.999852,
		0.998808,
		0.995624,
		0.988892,
		0.977279,
		0.959663,
		0.935212,
		0.903409,
		0.864056,
		0.81726,
		0.763413,
		0.70317,
		0.637429,
		0.567308,
		0.494131,
		0.419412,
		0.344836,
		0.272257,
		0.203676,
		0.141244,
		0.087245,
		0.044093,
		0.014327,
		0.000602,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.009472,
		0.035984,
		0.07677,
		0.129184,
		0.19071,
		0.258963,
		0.331701,
		0.406828,
		0.48241,
		0.556677,
		0.628041,
		0.695109,
		0.756699,
		0.811856,
		0.859876,
		0.900326,
		0.933068,
		0.95828,
		0.976473,
		0.988485,
		0.995458,
		0.998762,
		0.999846,
		0.999999,
		0.999993,
		0.999707,
		0.9982,
		0.994115,
		0.986004,
		0.972543,
		0.952661,
		0.925601,
		0.890936,
		0.848568,
		0.798713,
		0.741872,
		0.678813,
		0.610552,
		0.538323,
		0.463572,
		0.387931,
		0.313213,
		0.241394,
		0.174607,
		0.11513,
		0.065382,
		0.027912,
		0.005396,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.009664,
		0.036694,
		0.078237,
		0.13157,
		0.194101,
		0.263377,
		0.337094,
		0.413102,
		0.489418,
		0.564232,
		0.635927,
		0.703088,
		0.764523,
		0.819279,
		0.866667,
		0.906287,
		0.938045,
		0.962186,
		0.979295,
		0.990301,
		0.996439,
		0.999153,
		0.999926,
		1,
		0.999972,
		0.999473,
		0.99738,
		0.992239,
		0.982575,
		0.967092,
		0.944778,
		0.91496,
		0.877309,
		0.831837,
		0.778869,
		0.719025,
		0.653193,
		0.582507,
		0.508329,
		0.432226,
		0.355962,
		0.281477,
		0.21088,
		0.146439,
		0.090569,
		0.045828,
		0.014907,
		0.000627,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.009862,
		0.037425,
		0.079747,
		0.134022,
		0.197581,
		0.267901,
		0.342613,
		0.41951,
		0.496559,
		0.571912,
		0.643919,
		0.711144,
		0.772387,
		0.826699,
		0.873411,
		0.912153,
		0.942888,
		0.965925,
		0.981935,
		0.991944,
		0.997277,
		0.999452,
		0.999971,
		1,
		0.999923,
		0.999119,
		0.996301,
		0.989936,
		0.978541,
		0.960857,
		0.935944,
		0.903223,
		0.862471,
		0.813813,
		0.757694,
		0.694856,
		0.626313,
		0.553326,
		0.477387,
		0.400195,
		0.323646,
		0.249819,
		0.180962,
		0.119481,
		0.067939,
		0.029038,
		0.00562,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.010067,
		0.038178,
		0.081301,
		0.136543,
		0.201154,
		0.272538,
		0.34826,
		0.426054,
		0.503836,
		0.579717,
		0.652015,
		0.719275,
		0.780288,
		0.834112,
		0.880097,
		0.917916,
		0.947585,
		0.969489,
		0.984391,
		0.993413,
		0.997978,
		0.999669,
		0.999992,
		0.999999,
		0.999826,
		0.998608,
		0.994912,
		0.987147,
		0.973833,
		0.953766,
		0.926089,
		0.890323,
		0.846361,
		0.794448,
		0.735156,
		0.669354,
		0.598188,
		0.523054,
		0.445578,
		0.367599,
		0.291153,
		0.218463,
		0.151921,
		0.094085,
		0.047667,
		0.015523,
		0.000654,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.010277,
		0.038954,
		0.0829,
		0.139134,
		0.204822,
		0.277292,
		0.354039,
		0.432738,
		0.511251,
		0.587648,
		0.660215,
		0.727478,
		0.78822,
		0.841508,
		0.886718,
		0.923565,
		0.952128,
		0.972873,
		0.986658,
		0.994712,
		0.998551,
		0.999819,
		0.999999,
		0.999992,
		0.999656,
		0.997897,
		0.993157,
		0.983803,
		0.968379,
		0.945745,
		0.91514,
		0.876195,
		0.828924,
		0.773702,
		0.71123,
		0.642517,
		0.568845,
		0.491749,
		0.413,
		0.334582,
		0.258679,
		0.187663,
		0.124081,
		0.070648,
		0.030233,
		0.005858,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.010495,
		0.039753,
		0.084547,
		0.141799,
		0.208589,
		0.282166,
		0.359954,
		0.439564,
		0.518807,
		0.595706,
		0.668518,
		0.735749,
		0.796177,
		0.84888,
		0.893264,
		0.92909,
		0.956507,
		0.976067,
		0.988734,
		0.995843,
		0.999006,
		0.999913,
		1,
		0.999967,
		0.999381,
		0.996937,
		0.990973,
		0.979834,
		0.962103,
		0.936718,
		0.903024,
		0.860772,
		0.810106,
		0.751535,
		0.685901,
		0.614355,
		0.538323,
		0.45949,
		0.379775,
		0.301315,
		0.226451,
		0.157713,
		0.097809,
		0.049618,
		0.016179,
		0.000682,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.010719,
		0.040578,
		0.086243,
		0.14454,
		0.212459,
		0.287165,
		0.366009,
		0.446537,
		0.526505,
		0.603892,
		0.676922,
		0.744084,
		0.804153,
		0.85622,
		0.899725,
		0.934481,
		0.960712,
		0.979066,
		0.990617,
		0.996812,
		0.999355,
		0.999966,
		1,
		0.999909,
		0.998965,
		0.995675,
		0.988294,
		0.975164,
		0.954925,
		0.926607,
		0.889668,
		0.843991,
		0.789856,
		0.727916,
		0.65916,
		0.584888,
		0.50668,
		0.426374,
		0.346048,
		0.268,
		0.194735,
		0.128949,
		0.073521,
		0.031504,
		0.006111,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.010951,
		0.041428,
		0.08799,
		0.14736,
		0.216434,
		0.292293,
		0.372208,
		0.45366,
		0.534347,
		0.612205,
		0.685426,
		0.75248,
		0.812141,
		0.863519,
		0.906089,
		0.939726,
		0.964734,
		0.981865,
		0.992308,
		0.997627,
		0.99961,
		0.99999,
		0.999999,
		0.999794,
		0.998364,
		0.99405,
		0.985047,
		0.969716,
		0.946765,
		0.915332,
		0.875,
		0.82579,
		0.768129,
		0.702821,
		0.63101,
		0.554152,
		0.473988,
		0.392521,
		0.311993,
		0.234873,
		0.163836,
		0.101756,
		0.051692,
		0.016877,
		0.000713,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.01119,
		0.042305,
		0.089791,
		0.150262,
		0.220519,
		0.297552,
		0.378555,
		0.460936,
		0.542336,
		0.620647,
		0.694027,
		0.760931,
		0.820135,
		0.870768,
		0.912347,
		0.944814,
		0.968563,
		0.984457,
		0.993807,
		0.998294,
		0.999785,
		0.999999,
		0.99999,
		0.999593,
		0.997527,
		0.991996,
		0.981155,
		0.963404,
		0.937538,
		0.902812,
		0.858946,
		0.80611,
		0.744885,
		0.676234,
		0.601465,
		0.522198,
		0.440344,
		0.358077,
		0.277816,
		0.202206,
		0.134105,
		0.076573,
		0.032856,
		0.006382,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.011437,
		0.043211,
		0.091646,
		0.15325,
		0.224718,
		0.302949,
		0.385053,
		0.468368,
		0.550474,
		0.629217,
		0.702724,
		0.769433,
		0.828126,
		0.877955,
		0.918486,
		0.949734,
		0.972191,
		0.986838,
		0.995118,
		0.998827,
		0.999897,
		1,
		0.999961,
		0.999268,
		0.996398,
		0.98944,
		0.976536,
		0.956143,
		0.927158,
		0.888967,
		0.841438,
		0.784898,
		0.720092,
		0.648151,
		0.570554,
		0.489096,
		0.405867,
		0.32322,
		0.243759,
		0.170317,
		0.105945,
		0.053897,
		0.017621,
		0.000745,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.011692,
		0.044146,
		0.09356,
		0.156326,
		0.229035,
		0.308487,
		0.391707,
		0.475959,
		0.558763,
		0.637915,
		0.711512,
		0.77798,
		0.836105,
		0.88507,
		0.924495,
		0.954475,
		0.975609,
		0.989007,
		0.996245,
		0.999236,
		0.999959,
		1,
		0.999892,
		0.998776,
		0.994912,
		0.986304,
		0.971103,
		0.947842,
		0.915536,
		0.873716,
		0.822408,
		0.762104,
		0.693726,
		0.618579,
		0.538323,
		0.45494,
		0.370703,
		0.288158,
		0.210104,
		0.139573,
		0.079817,
		0.034297,
		0.006671,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.011957,
		0.045111,
		0.095534,
		0.159495,
		0.233473,
		0.31417,
		0.398522,
		0.483714,
		0.567204,
		0.64674,
		0.72039,
		0.786565,
		0.844063,
		0.892101,
		0.930361,
		0.959025,
		0.97881,
		0.990961,
		0.997195,
		0.999536,
		0.999989,
		0.999998,
		0.999755,
		0.998064,
		0.992999,
		0.982503,
		0.964764,
		0.938407,
		0.902584,
		0.85698,
		0.801791,
		0.737688,
		0.665773,
		0.58754,
		0.504839,
		0.419845,
		0.33503,
		0.253142,
		0.177183,
		0.110396,
		0.056247,
		0.018415,
		0.00078,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.01223,
		0.046108,
		0.09757,
		0.162759,
		0.238038,
		0.320005,
		0.405501,
		0.491635,
		0.575799,
		0.655692,
		0.729352,
		0.795181,
		0.85199,
		0.899037,
		0.93607,
		0.963372,
		0.981785,
		0.9927,
		0.997978,
		0.999744,
		0.999998,
		0.999988,
		0.999515,
		0.997073,
		0.990579,
		0.977948,
		0.957423,
		0.927745,
		0.888213,
		0.838682,
		0.77953,
		0.711615,
		0.636233,
		0.555074,
		0.470189,
		0.383959,
		0.299063,
		0.218463,
		0.145377,
		0.08327,
		0.035835,
		0.006979,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.012512,
		0.047138,
		0.099671,
		0.166123,
		0.242735,
		0.325995,
		0.41265,
		0.499726,
		0.584549,
		0.664769,
		0.738394,
		0.80382,
		0.859876,
		0.905864,
		0.94161,
		0.967506,
		0.98453,
		0.994227,
		0.998605,
		0.999876,
		1,
		0.999953,
		0.999128,
		0.995735,
		0.987569,
		0.972543,
		0.94898,
		0.915755,
		0.872333,
		0.818748,
		0.755573,
		0.683863,
		0.605119,
		0.52124,
		0.434489,
		0.347461,
		0.263058,
		0.184465,
		0.11513,
		0.058753,
		0.019264,
		0.000817,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.012805,
		0.048204,
		0.101841,
		0.16959,
		0.247567,
		0.332145,
		0.419973,
		0.50799,
		0.593455,
		0.673969,
		0.747512,
		0.812474,
		0.867708,
		0.912568,
		0.946968,
		0.971415,
		0.987038,
		0.995546,
		0.999088,
		0.999951,
		1,
		0.999871,
		0.998541,
		0.993974,
		0.983876,
		0.966185,
		0.939331,
		0.902341,
		0.854856,
		0.797109,
		0.729875,
		0.654422,
		0.572465,
		0.486123,
		0.397884,
		0.31057,
		0.227317,
		0.151546,
		0.086952,
		0.037478,
		0.00731,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.013107,
		0.049305,
		0.104082,
		0.173165,
		0.25254,
		0.338462,
		0.427474,
		0.51643,
		0.602518,
		0.683291,
		0.756699,
		0.821134,
		0.875474,
		0.919137,
		0.952128,
		0.975089,
		0.989308,
		0.996662,
		0.999445,
		0.999986,
		0.999998,
		0.999707,
		0.997692,
		0.991705,
		0.9794,
		0.958769,
		0.928371,
		0.887401,
		0.835698,
		0.773702,
		0.702404,
		0.623297,
		0.538323,
		0.449834,
		0.360554,
		0.273547,
		0.192195,
		0.120172,
		0.061429,
		0.020173,
		0.000857,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.013421,
		0.050445,
		0.106396,
		0.176852,
		0.257659,
		0.34495,
		0.435159,
		0.52505,
		0.61174,
		0.692731,
		0.76595,
		0.829789,
		0.883162,
		0.925553,
		0.957078,
		0.978519,
		0.991337,
		0.997585,
		0.999693,
		0.999998,
		0.999986,
		0.999419,
		0.996509,
		0.988837,
		0.974035,
		0.950184,
		0.91599,
		0.870838,
		0.814777,
		0.74847,
		0.673139,
		0.590511,
		0.502771,
		0.412516,
		0.32272,
		0.236704,
		0.15811,
		0.09088,
		0.039236,
		0.007665,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.013746,
		0.051625,
		0.108788,
		0.180656,
		0.262931,
		0.351615,
		0.443032,
		0.533853,
		0.62112,
		0.702287,
		0.775256,
		0.838428,
		0.890756,
		0.931803,
		0.961803,
		0.981696,
		0.993125,
		0.998327,
		0.999851,
		1,
		0.999944,
		0.998953,
		0.994912,
		0.985268,
		0.967671,
		0.940314,
		0.902078,
		0.852555,
		0.792019,
		0.721369,
		0.642073,
		0.556109,
		0.465916,
		0.374349,
		0.28465,
		0.200411,
		0.12555,
		0.064291,
		0.021147,
		0.000899,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.014083,
		0.052846,
		0.111261,
		0.184582,
		0.268359,
		0.358463,
		0.451099,
		0.542841,
		0.630658,
		0.711954,
		0.784611,
		0.84704,
		0.898242,
		0.93787,
		0.966289,
		0.984612,
		0.994676,
		0.998903,
		0.999941,
		1,
		0.999844,
		0.998248,
		0.992809,
		0.980889,
		0.960187,
		0.929041,
		0.886525,
		0.832457,
		0.767353,
		0.692366,
		0.609221,
		0.520162,
		0.427896,
		0.335557,
		0.246666,
		0.165103,
		0.09508,
		0.041121,
		0.008046,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.014433,
		0.054111,
		0.113819,
		0.188634,
		0.273952,
		0.365499,
		0.459363,
		0.552017,
		0.640354,
		0.721729,
		0.794004,
		0.85561,
		0.905603,
		0.943738,
		0.970524,
		0.987261,
		0.995996,
		0.99933,
		0.999983,
		0.999998,
		0.999645,
		0.997227,
		0.990101,
		0.975582,
		0.951461,
		0.916242,
		0.869219,
		0.810454,
		0.740722,
		0.661443,
		0.574618,
		0.482769,
		0.388893,
		0.296414,
		0.209152,
		0.131291,
		0.067358,
		0.022194,
		0.000945,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.014796,
		0.055422,
		0.116464,
		0.192819,
		0.279714,
		0.372731,
		0.467831,
		0.561384,
		0.650207,
		0.731605,
		0.803427,
		0.864126,
		0.912824,
		0.949389,
		0.974494,
		0.989641,
		0.997092,
		0.999627,
		0.999998,
		0.999982,
		0.999296,
		0.995804,
		0.986675,
		0.969224,
		0.941362,
		0.901795,
		0.850054,
		0.786465,
		0.712079,
		0.628601,
		0.538323,
		0.444067,
		0.349131,
		0.25725,
		0.172561,
		0.099574,
		0.043145,
		0.008457,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.015172,
		0.056781,
		0.119202,
		0.197141,
		0.285652,
		0.380165,
		0.476508,
		0.570945,
		0.660215,
		0.741577,
		0.812868,
		0.872571,
		0.919885,
		0.954807,
		0.978187,
		0.991748,
		0.997978,
		0.999819,
		1,
		0.999931,
		0.998733,
		0.993882,
		0.982413,
		0.961681,
		0.929759,
		0.885575,
		0.828924,
		0.760415,
		0.681392,
		0.593863,
		0.500431,
		0.404232,
		0.308888,
		0.218463,
		0.137432,
		0.070648,
		0.023321,
		0.000995,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.015563,
		0.05819,
		0.122037,
		0.201607,
		0.291774,
		0.387806,
		0.485397,
		0.580701,
		0.670377,
		0.751638,
		0.822315,
		0.880929,
		0.926769,
		0.959974,
		0.981594,
		0.993585,
		0.998669,
		0.999928,
		1,
		0.999809,
		0.997877,
		0.991352,
		0.977183,
		0.952816,
		0.916514,
		0.867459,
		0.805731,
		0.732241,
		0.648648,
		0.557281,
		0.461072,
		0.363491,
		0.268504,
		0.180527,
		0.104393,
		0.045322,
		0.0089,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.01597,
		0.059653,
		0.124973,
		0.206223,
		0.298086,
		0.395662,
		0.494505,
		0.590655,
		0.68069,
		0.761779,
		0.831754,
		0.889182,
		0.933455,
		0.964874,
		0.984704,
		0.995156,
		0.999183,
		0.99998,
		0.999997,
		0.999567,
		0.996638,
		0.988091,
		0.970847,
		0.942482,
		0.901488,
		0.847326,
		0.780384,
		0.701897,
		0.613858,
		0.518938,
		0.420418,
		0.322126,
		0.228392,
		0.144007,
		0.074184,
		0.024536,
		0.001049,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.016392,
		0.06117,
		0.128015,
		0.210995,
		0.304595,
		0.403741,
		0.503836,
		0.600807,
		0.69115,
		0.771993,
		0.841172,
		0.897313,
		0.939924,
		0.969489,
		0.987511,
		0.996468,
		0.999544,
		0.999997,
		0.999978,
		0.99914,
		0.994912,
		0.983968,
		0.963257,
		0.930531,
		0.884543,
		0.82506,
		0.752804,
		0.669354,
		0.577059,
		0.478955,
		0.378694,
		0.280485,
		0.189047,
		0.109566,
		0.047667,
		0.009378,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.016832,
		0.062746,
		0.131168,
		0.215931,
		0.31131,
		0.412048,
		0.513395,
		0.611159,
		0.701753,
		0.782268,
		0.850552,
		0.905299,
		0.946156,
		0.973805,
		0.990011,
		0.997533,
		0.999777,
		1,
		0.999916,
		0.998451,
		0.992579,
		0.978837,
		0.954256,
		0.916807,
		0.865538,
		0.800552,
		0.722924,
		0.634605,
		0.538323,
		0.437501,
		0.336187,
		0.238994,
		0.151058,
		0.07799,
		0.025848,
		0.001107,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.017289,
		0.064384,
		0.134438,
		0.221038,
		0.318238,
		0.420591,
		0.523188,
		0.621711,
		0.712494,
		0.792593,
		0.859876,
		0.913121,
		0.952128,
		0.977805,
		0.992201,
		0.998368,
		0.999911,
		1,
		0.999765,
		0.997403,
		0.989507,
		0.972543,
		0.943681,
		0.901155,
		0.84434,
		0.773702,
		0.690698,
		0.597675,
		0.497762,
		0.394798,
		0.293251,
		0.19817,
		0.11513,
		0.050198,
		0.009896,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.017765,
		0.066086,
		0.137829,
		0.226323,
		0.325389,
		0.429378,
		0.533218,
		0.632464,
		0.723368,
		0.802956,
		0.869125,
		0.920755,
		0.957819,
		0.981477,
		0.994084,
		0.998994,
		0.999975,
		0.999997,
		0.999465,
		0.995886,
		0.985548,
		0.964919,
		0.931363,
		0.883417,
		0.820817,
		0.744423,
		0.656102,
		0.558619,
		0.455537,
		0.351133,
		0.250328,
		0.158632,
		0.082095,
		0.027267,
		0.00117,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.018262,
		0.067857,
		0.141349,
		0.231794,
		0.332769,
		0.438416,
		0.54349,
		0.643416,
		0.734365,
		0.81334,
		0.87828,
		0.928179,
		0.963209,
		0.98481,
		0.995666,
		0.999435,
		0.999996,
		0.999973,
		0.998938,
		0.993771,
		0.980542,
		0.955789,
		0.917126,
		0.863435,
		0.794849,
		0.712649,
		0.61914,
		0.517538,
		0.411865,
		0.306869,
		0.207954,
		0.121124,
		0.052935,
		0.010458,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.018779,
		0.069699,
		0.145004,
		0.237461,
		0.34039,
		0.447714,
		0.554008,
		0.654566,
		0.745479,
		0.823732,
		0.887317,
		0.935365,
		0.968276,
		0.987794,
		0.996959,
		0.999723,
		1,
		0.999895,
		0.998086,
		0.990912,
		0.974313,
		0.944967,
		0.900792,
		0.841057,
		0.766326,
		0.678333,
		0.579853,
		0.474581,
		0.367032,
		0.262459,
		0.166781,
		0.086531,
		0.028807,
		0.001239,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.019319,
		0.071616,
		0.148801,
		0.243331,
		0.34826,
		0.457278,
		0.564777,
		0.665912,
		0.756699,
		0.834112,
		0.896212,
		0.94229,
		0.972999,
		0.990425,
		0.997978,
		0.999889,
		1,
		0.999707,
		0.99679,
		0.987147,
		0.966673,
		0.932261,
		0.882183,
		0.816138,
		0.735156,
		0.641458,
		0.538323,
		0.42996,
		0.321409,
		0.218463,
		0.127594,
		0.055902,
		0.011069,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.019882,
		0.073614,
		0.152746,
		0.249415,
		0.356389,
		0.467118,
		0.575799,
		0.677451,
		0.768013,
		0.84446,
		0.90494,
		0.948926,
		0.97736,
		0.9927,
		0.998747,
		0.999968,
		0.999996,
		0.999333,
		0.994912,
		0.982294,
		0.957423,
		0.917472,
		0.861121,
		0.788542,
		0.701268,
		0.602046,
		0.494689,
		0.383959,
		0.27546,
		0.175563,
		0.091333,
		0.030481,
		0.001314,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.02047,
		0.075696,
		0.156848,
		0.255722,
		0.364787,
		0.47724,
		0.587077,
		0.689177,
		0.779408,
		0.854755,
		0.913472,
		0.955247,
		0.981342,
		0.994625,
		0.999293,
		0.999995,
		0.999966,
		0.998674,
		0.992293,
		0.976158,
		0.946351,
		0.900395,
		0.837433,
		0.758149,
		0.664622,
		0.560159,
		0.449152,
		0.336951,
		0.229766,
		0.134589,
		0.059124,
		0.011736,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.021084,
		0.077867,
		0.161115,
		0.262263,
		0.373465,
		0.487653,
		0.598615,
		0.701086,
		0.790869,
		0.864974,
		0.921781,
		0.961225,
		0.984931,
		0.996208,
		0.999651,
		1,
		0.999867,
		0.997607,
		0.988753,
		0.968525,
		0.933236,
		0.880827,
		0.810953,
		0.72486,
		0.625212,
		0.515919,
		0.40199,
		0.289411,
		0.185043,
		0.096544,
		0.032305,
		0.001396,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.021727,
		0.080132,
		0.165555,
		0.269049,
		0.382435,
		0.498366,
		0.610413,
		0.71317,
		0.802379,
		0.875088,
		0.929834,
		0.966832,
		0.988116,
		0.997465,
		0.999859,
		1,
		0.999629,
		0.995985,
		0.984089,
		0.959167,
		0.91785,
		0.858565,
		0.781532,
		0.688606,
		0.583079,
		0.469512,
		0.353578,
		0.241943,
		0.142168,
		0.062632,
		0.012464,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.022399,
		0.082496,
		0.170178,
		0.276092,
		0.391707,
		0.509384,
		0.622473,
		0.72542,
		0.813918,
		0.88507,
		0.937598,
		0.972044,
		0.99089,
		0.998419,
		0.999959,
		0.999994,
		0.999156,
		0.993633,
		0.978079,
		0.947842,
		0.899959,
		0.833411,
		0.749038,
		0.649351,
		0.538323,
		0.42121,
		0.304401,
		0.195295,
		0.102209,
		0.034297,
		0.001486,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.023103,
		0.084966,
		0.174994,
		0.283404,
		0.401293,
		0.520717,
		0.634793,
		0.737825,
		0.825462,
		0.894886,
		0.94504,
		0.976834,
		0.993251,
		0.999102,
		0.999993,
		0.999955,
		0.99832,
		0.990351,
		0.970479,
		0.934294,
		0.879328,
		0.805178,
		0.713365,
		0.607109,
		0.491113,
		0.371384,
		0.255081,
		0.150396,
		0.06646,
		0.013263,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.023841,
		0.087547,
		0.180011,
		0.290993,
		0.411199,
		0.532363,
		0.647364,
		0.750362,
		0.836975,
		0.904491,
		0.952111,
		0.981167,
		0.995193,
		0.999536,
		0.999982,
		0.999812,
		0.996952,
		0.985898,
		0.961014,
		0.918247,
		0.85571,
		0.773688,
		0.674436,
		0.561943,
		0.441699,
		0.320521,
		0.2064,
		0.108382,
		0.036478,
		0.001584,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.024612,
		0.090239,
		0.185227,
		0.298855,
		0.421411,
		0.544295,
		0.66014,
		0.762966,
		0.848375,
		0.91379,
		0.958714,
		0.98496,
		0.996665,
		0.999705,
		0.999887,
		0.999412,
		0.994799,
		0.979961,
		0.949345,
		0.899376,
		0.82883,
		0.73875,
		0.632191,
		0.513969,
		0.390425,
		0.269249,
		0.159331,
		0.07064,
		0.014139,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.025414,
		0.093032,
		0.190619,
		0.306942,
		0.431857,
		0.556411,
		0.672987,
		0.77547,
		0.859466,
		0.922566,
		0.96463,
		0.988014,
		0.997502,
		0.999471,
		0.999516,
		0.998441,
		0.991455,
		0.972079,
		0.935005,
		0.877246,
		0.798332,
		0.700132,
		0.586569,
		0.463351,
		0.337737,
		0.218359,
		0.115075,
		0.038856,
		0.001693,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.026239,
		0.09589,
		0.196111,
		0.315133,
		0.442357,
		0.56847,
		0.685602,
		0.787513,
		0.86984,
		0.930382,
		0.969415,
		0.989904,
		0.997321,
		0.998467,
		0.998418,
		0.996322,
		0.986257,
		0.961557,
		0.917318,
		0.851254,
		0.763735,
		0.657532,
		0.537502,
		0.410314,
		0.284216,
		0.168853,
		0.075127,
		0.015085,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.027066,
		0.098742,
		0.201554,
		0.323181,
		0.452559,
		0.580011,
		0.697422,
		0.798444,
		0.878774,
		0.936471,
		0.972293,
		0.98988,
		0.995414,
		0.99599,
		0.99577,
		0.992111,
		0.978195,
		0.947385,
		0.895339,
		0.820585,
		0.724419,
		0.610585,
		0.484947,
		0.355202,
		0.230652,
		0.122028,
		0.041348,
		0.001807,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.027862,
		0.101464,
		0.206694,
		0.330677,
		0.461886,
		0.590294,
		0.707552,
		0.807238,
		0.885148,
		0.939661,
		0.972083,
		0.986795,
		0.990679,
		0.990914,
		0.990301,
		0.984431,
		0.96586,
		0.928207,
		0.867848,
		0.784235,
		0.679669,
		0.558942,
		0.428982,
		0.298578,
		0.178148,
		0.079568,
		0.016032,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.028577,
		0.103873,
		0.211158,
		0.337025,
		0.46951,
		0.59826,
		0.714734,
		0.812471,
		0.887425,
		0.938357,
		0.967195,
		0.979108,
		0.98162,
		0.981678,
		0.980301,
		0.971493,
		0.947489,
		0.902391,
		0.833445,
		0.741127,
		0.628819,
		0.502414,
		0.369959,
		0.241376,
		0.128242,
		0.043618,
		0.001912,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.029144,
		0.105728,
		0.21446,
		0.341453,
		0.474364,
		0.602567,
		0.717384,
		0.812377,
		0.883723,
		0.930634,
		0.955728,
		0.96498,
		0.966433,
		0.966371,
		0.963724,
		0.951225,
		0.921114,
		0.868202,
		0.790745,
		0.690328,
		0.571468,
		0.441194,
		0.308696,
		0.185047,
		0.082994,
		0.016784,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.029484,
		0.106744,
		0.216028,
		0.343066,
		0.475222,
		0.601687,
		0.713725,
		0.805002,
		0.871991,
		0.914431,
		0.935677,
		0.942488,
		0.943208,
		0.942932,
		0.938409,
		0.921517,
		0.884835,
		0.824096,
		0.73868,
		0.63134,
		0.507751,
		0.37608,
		0.246645,
		0.131646,
		0.04496,
		0.001978,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.029512,
		0.10662,
		0.215263,
		0.340931,
		0.470829,
		0.594078,
		0.701997,
		0.788449,
		0.850289,
		0.887842,
		0.905228,
		0.909906,
		0.910193,
		0.909445,
		0.902408,
		0.88057,
		0.837182,
		0.769076,
		0.676837,
		0.564404,
		0.438588,
		0.308649,
		0.185967,
		0.083785,
		0.017013,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.029149,
		0.105077,
		0.211619,
		0.334215,
		0.460076,
		0.578414,
		0.680727,
		0.761188,
		0.817112,
		0.849464,
		0.863107,
		0.866039,
		0.866109,
		0.864494,
		0.854357,
		0.827286,
		0.777497,
		0.703054,
		0.605771,
		0.49074,
		0.365832,
		0.241294,
		0.12944,
		0.044403,
		0.001962,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.028334,
		0.101903,
		0.204691,
		0.322316,
		0.442204,
		0.553828,
		0.649019,
		0.722367,
		0.771729,
		0.798725,
		0.808892,
		0.81052,
		0.810449,
		0.807497,
		0.793834,
		0.761617,
		0.706256,
		0.627109,
		0.527188,
		0.412633,
		0.29224,
		0.177071,
		0.080172,
		0.01635,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.027034,
		0.096994,
		0.1943,
		0.305002,
		0.416966,
		0.520121,
		0.606781,
		0.672067,
		0.714427,
		0.736123,
		0.743232,
		0.743999,
		0.743703,
		0.738959,
		0.721612,
		0.684786,
		0.625228,
		0.543575,
		0.443925,
		0.333305,
		0.221226,
		0.119331,
		0.041135,
		0.001825,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.025256,
		0.090387,
		0.180546,
		0.282483,
		0.384738,
		0.477883,
		0.554862,
		0.611418,
		0.646618,
		0.663304,
		0.667894,
		0.668181,
		0.667436,
		0.660568,
		0.639732,
		0.599306,
		0.537419,
		0.455881,
		0.359697,
		0.256554,
		0.156413,
		0.071204,
		0.014591,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.023047,
		0.082265,
		0.163827,
		0.255441,
		0.346533,
		0.428498,
		0.495029,
		0.542556,
		0.570757,
		0.582942,
		0.585621,
		0.585686,
		0.584203,
		0.575095,
		0.551352,
		0.508764,
		0.446768,
		0.368169,
		0.278635,
		0.186232,
		0.101068,
		0.035026,
		0.001561,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.020493,
		0.072946,
		0.144809,
		0.224966,
		0.303917,
		0.374015,
		0.429797,
		0.468408,
		0.490085,
		0.498456,
		0.499827,
		0.499767,
		0.497293,
		0.486112,
		0.460409,
		0.417399,
		0.35766,
		0.284745,
		0.204699,
		0.125657,
		0.057546,
		0.011854,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.017708,
		0.06285,
		0.124352,
		0.192443,
		0.25883,
		0.316924,
		0.36215,
		0.392355,
		0.408251,
		0.413595,
		0.414184,
		0.413956,
		0.410368,
		0.397574,
		0.371134,
		0.329566,
		0.274347,
		0.209488,
		0.141108,
		0.077097,
		0.026875,
		0.001204,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.014824,
		0.052454,
		0.103418,
		0.159395,
		0.213377,
		0.259871,
		0.295193,
		0.317842,
		0.328883,
		0.332002,
		0.332201,
		0.331687,
		0.32705,
		0.313356,
		0.287533,
		0.249185,
		0.200397,
		0.145338,
		0.089903,
		0.041445,
		0.008587,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.011975,
		0.042238,
		0.082967,
		0.127325,
		0.169596,
		0.205375,
		0.231825,
		0.248009,
		0.255208,
		0.256834,
		0.256874,
		0.255969,
		0.25055,
		0.236823,
		0.212934,
		0.17929,
		0.138285,
		0.09396,
		0.051724,
		0.018148,
		0.000818,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.009287,
		0.032644,
		0.063871,
		0.097571,
		0.129275,
		0.1556,
		0.174471,
		0.185408,
		0.18976,
		0.190493,
		0.19046,
		0.189147,
		0.183376,
		0.170509,
		0.149661,
		0.121741,
		0.089175,
		0.055636,
		0.025838,
		0.005387,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.006863,
		0.02404,
		0.046839,
		0.071206,
		0.09381,
		0.112184,
		0.124906,
		0.131824,
		0.134224,
		0.134495,
		0.134389,
		0.132767,
		0.127154,
		0.115924,
		0.098875,
		0.077134,
		0.052927,
		0.029382,
		0.010384,
		0.000471,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.004784,
		0.016693,
		0.032382,
		0.048973,
		0.064131,
		0.076164,
		0.084171,
		0.088209,
		0.089387,
		0.089464,
		0.089266,
		0.087528,
		0.082561,
		0.073513,
		0.06058,
		0.044879,
		0.028272,
		0.013239,
		0.00278,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.003097,
		0.010764,
		0.020783,
		0.031258,
		0.040669,
		0.047945,
		0.05257,
		0.054701,
		0.055198,
		0.055211,
		0.054932,
		0.053314,
		0.049356,
		0.042718,
		0.033761,
		0.023426,
		0.013129,
		0.004677,
		0.000214,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.001819,
		0.006296,
		0.012095,
		0.018083,
		0.023365,
		0.027329,
		0.029718,
		0.030705,
		0.030876,
		0.030868,
		0.030563,
		0.029267,
		0.026482,
		0.02215,
		0.016622,
		0.010587,
		0.005004,
		0.001059,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.000933,
		0.003214,
		0.006142,
		0.009125,
		0.011703,
		0.013573,
		0.014633,
		0.015015,
		0.015059,
		0.015042,
		0.014781,
		0.013913,
		0.012245,
		0.009823,
		0.006903,
		0.003911,
		0.001406,
		6.5e-005,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.00039,
		0.001339,
		0.002543,
		0.003753,
		0.004775,
		0.005488,
		0.005863,
		0.005977,
		0.005985,
		0.005965,
		0.005799,
		0.005342,
		0.004545,
		0.003462,
		0.002233,
		0.001066,
		0.000228,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		0.000116,
		0.000396,
		0.000748,
		0.001096,
		0.001382,
		0.001573,
		0.001665,
		0.001688,
		0.001688,
		0.001676,
		0.001605,
		0.00144,
		0.001175,
		0.000838,
		0.000481,
		0.000175,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		1.7e-005,
		5.8e-005,
		0.000109,
		0.000159,
		0.000199,
		0.000224,
		0.000235,
		0.000237,
		0.000237,
		0.000233,
		0.000219,
		0.00019,
		0.000147,
		9.6e-005,
		4.7e-005,
		1e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		1e-005,
		4.7e-005,
		9.6e-005,
		0.000147,
		0.00019,
		0.000219,
		0.000233,
		0.000237,
		0.000237,
		0.000235,
		0.000224,
		0.000199,
		0.000159,
		0.000109,
		5.8e-005,
		1.7e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		0.000175,
		0.000481,
		0.000838,
		0.001175,
		0.00144,
		0.001605,
		0.001676,
		0.001688,
		0.001688,
		0.001665,
		0.001573,
		0.001382,
		0.001096,
		0.000748,
		0.000396,
		0.000116,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		0.000228,
		0.001066,
		0.002233,
		0.003462,
		0.004545,
		0.005342,
		0.005799,
		0.005965,
		0.005985,
		0.005977,
		0.005863,
		0.005488,
		0.004775,
		0.003753,
		0.002543,
		0.001339,
		0.00039,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		6.5e-005,
		0.001406,
		0.003911,
		0.006903,
		0.009823,
		0.012245,
		0.013913,
		0.014781,
		0.015042,
		0.015059,
		0.015015,
		0.014633,
		0.013573,
		0.011703,
		0.009125,
		0.006142,
		0.003214,
		0.000933,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		0.001059,
		0.005004,
		0.010587,
		0.016622,
		0.02215,
		0.026482,
		0.029267,
		0.030563,
		0.030868,
		0.030876,
		0.030705,
		0.029718,
		0.027329,
		0.023365,
		0.018083,
		0.012095,
		0.006296,
		0.001819,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000214,
		0.004677,
		0.013129,
		0.023426,
		0.033761,
		0.042718,
		0.049356,
		0.053314,
		0.054932,
		0.055211,
		0.055198,
		0.054701,
		0.05257,
		0.047945,
		0.040669,
		0.031258,
		0.020783,
		0.010764,
		0.003097,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.00278,
		0.013239,
		0.028272,
		0.044879,
		0.06058,
		0.073513,
		0.082561,
		0.087528,
		0.089266,
		0.089464,
		0.089387,
		0.088209,
		0.084171,
		0.076164,
		0.064131,
		0.048973,
		0.032382,
		0.016693,
		0.004784,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.000471,
		0.010384,
		0.029382,
		0.052927,
		0.077134,
		0.098875,
		0.115924,
		0.127154,
		0.132767,
		0.134389,
		0.134495,
		0.134224,
		0.131824,
		0.124906,
		0.112184,
		0.09381,
		0.071206,
		0.046839,
		0.02404,
		0.006863,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.005387,
		0.025838,
		0.055636,
		0.089175,
		0.121741,
		0.149661,
		0.170509,
		0.183376,
		0.189147,
		0.19046,
		0.190493,
		0.18976,
		0.185408,
		0.174471,
		0.1556,
		0.129275,
		0.097571,
		0.063871,
		0.032644,
		0.009287,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.000818,
		0.018148,
		0.051724,
		0.09396,
		0.138285,
		0.17929,
		0.212934,
		0.236823,
		0.25055,
		0.255969,
		0.256874,
		0.256834,
		0.255208,
		0.248009,
		0.231825,
		0.205375,
		0.169596,
		0.127325,
		0.082967,
		0.042238,
		0.011975,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.008587,
		0.041445,
		0.089903,
		0.145338,
		0.200397,
		0.249185,
		0.287533,
		0.313356,
		0.32705,
		0.331687,
		0.332201,
		0.332002,
		0.328883,
		0.317842,
		0.295193,
		0.259871,
		0.213377,
		0.159395,
		0.103418,
		0.052454,
		0.014824,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.001204,
		0.026875,
		0.077097,
		0.141108,
		0.209488,
		0.274347,
		0.329566,
		0.371134,
		0.397574,
		0.410368,
		0.413956,
		0.414184,
		0.413595,
		0.408251,
		0.392355,
		0.36215,
		0.316924,
		0.25883,
		0.192443,
		0.124352,
		0.06285,
		0.017708,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.011854,
		0.057546,
		0.125657,
		0.204699,
		0.284745,
		0.35766,
		0.417399,
		0.460409,
		0.486112,
		0.497293,
		0.499767,
		0.499827,
		0.498456,
		0.490085,
		0.468408,
		0.429797,
		0.374015,
		0.303917,
		0.224966,
		0.144809,
		0.072946,
		0.020493,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.001561,
		0.035026,
		0.101068,
		0.186232,
		0.278635,
		0.368169,
		0.446768,
		0.508764,
		0.551352,
		0.575095,
		0.584203,
		0.585686,
		0.585621,
		0.582942,
		0.570757,
		0.542556,
		0.495029,
		0.428498,
		0.346533,
		0.255441,
		0.163827,
		0.082265,
		0.023047,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.014591,
		0.071204,
		0.156413,
		0.256554,
		0.359697,
		0.455881,
		0.537419,
		0.599306,
		0.639732,
		0.660568,
		0.667436,
		0.668181,
		0.667894,
		0.663304,
		0.646618,
		0.611418,
		0.554862,
		0.477883,
		0.384738,
		0.282483,
		0.180546,
		0.090387,
		0.025256,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.001825,
		0.041135,
		0.119331,
		0.221226,
		0.333305,
		0.443925,
		0.543575,
		0.625228,
		0.684786,
		0.721612,
		0.738959,
		0.743703,
		0.743999,
		0.743232,
		0.736123,
		0.714427,
		0.672067,
		0.606781,
		0.520121,
		0.416966,
		0.305002,
		0.1943,
		0.096994,
		0.027034,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.01635,
		0.080172,
		0.177071,
		0.29224,
		0.412633,
		0.527188,
		0.627109,
		0.706256,
		0.761617,
		0.793834,
		0.807497,
		0.810449,
		0.81052,
		0.808892,
		0.798725,
		0.771729,
		0.722367,
		0.649019,
		0.553828,
		0.442204,
		0.322316,
		0.204691,
		0.101903,
		0.028334,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.001962,
		0.044403,
		0.12944,
		0.241294,
		0.365832,
		0.49074,
		0.605771,
		0.703054,
		0.777497,
		0.827286,
		0.854357,
		0.864494,
		0.866109,
		0.866039,
		0.863107,
		0.849464,
		0.817112,
		0.761188,
		0.680727,
		0.578414,
		0.460076,
		0.334215,
		0.211619,
		0.105077,
		0.029149,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.017013,
		0.083785,
		0.185967,
		0.308649,
		0.438588,
		0.564404,
		0.676837,
		0.769076,
		0.837182,
		0.88057,
		0.902408,
		0.909445,
		0.910193,
		0.909906,
		0.905228,
		0.887842,
		0.850289,
		0.788449,
		0.701997,
		0.594078,
		0.470829,
		0.340931,
		0.215263,
		0.10662,
		0.029512,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.001978,
		0.04496,
		0.131646,
		0.246645,
		0.37608,
		0.507751,
		0.63134,
		0.73868,
		0.824096,
		0.884835,
		0.921517,
		0.938409,
		0.942932,
		0.943208,
		0.942488,
		0.935677,
		0.914431,
		0.871991,
		0.805002,
		0.713725,
		0.601687,
		0.475222,
		0.343066,
		0.216028,
		0.106744,
		0.029484,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.016784,
		0.082994,
		0.185047,
		0.308696,
		0.441194,
		0.571468,
		0.690328,
		0.790745,
		0.868202,
		0.921114,
		0.951225,
		0.963724,
		0.966371,
		0.966433,
		0.96498,
		0.955728,
		0.930634,
		0.883723,
		0.812377,
		0.717384,
		0.602567,
		0.474364,
		0.341453,
		0.21446,
		0.105728,
		0.029144,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.001912,
		0.043618,
		0.128242,
		0.241376,
		0.369959,
		0.502414,
		0.628819,
		0.741127,
		0.833445,
		0.902391,
		0.947489,
		0.971493,
		0.980301,
		0.981678,
		0.98162,
		0.979108,
		0.967195,
		0.938357,
		0.887425,
		0.812471,
		0.714734,
		0.59826,
		0.46951,
		0.337025,
		0.211158,
		0.103873,
		0.028577,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.016032,
		0.079568,
		0.178148,
		0.298578,
		0.428982,
		0.558942,
		0.679669,
		0.784235,
		0.867848,
		0.928207,
		0.96586,
		0.984431,
		0.990301,
		0.990914,
		0.990679,
		0.986795,
		0.972083,
		0.939661,
		0.885148,
		0.807238,
		0.707552,
		0.590294,
		0.461886,
		0.330677,
		0.206694,
		0.101464,
		0.027862,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.001807,
		0.041348,
		0.122028,
		0.230652,
		0.355202,
		0.484947,
		0.610585,
		0.724419,
		0.820585,
		0.895339,
		0.947385,
		0.978195,
		0.992111,
		0.99577,
		0.99599,
		0.995414,
		0.98988,
		0.972293,
		0.936471,
		0.878774,
		0.798444,
		0.697422,
		0.580011,
		0.452559,
		0.323181,
		0.201554,
		0.098742,
		0.027066,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.015085,
		0.075127,
		0.168853,
		0.284216,
		0.410314,
		0.537502,
		0.657532,
		0.763735,
		0.851254,
		0.917318,
		0.961557,
		0.986257,
		0.996322,
		0.998418,
		0.998467,
		0.997321,
		0.989904,
		0.969415,
		0.930382,
		0.86984,
		0.787513,
		0.685602,
		0.56847,
		0.442357,
		0.315133,
		0.196111,
		0.09589,
		0.026239,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.001693,
		0.038856,
		0.115075,
		0.218359,
		0.337737,
		0.463351,
		0.586569,
		0.700132,
		0.798332,
		0.877246,
		0.935005,
		0.972079,
		0.991455,
		0.998441,
		0.999516,
		0.999471,
		0.997502,
		0.988014,
		0.96463,
		0.922566,
		0.859466,
		0.77547,
		0.672987,
		0.556411,
		0.431857,
		0.306942,
		0.190619,
		0.093032,
		0.025414,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.014139,
		0.07064,
		0.159331,
		0.269249,
		0.390425,
		0.513969,
		0.632191,
		0.73875,
		0.82883,
		0.899376,
		0.949345,
		0.979961,
		0.994799,
		0.999412,
		0.999887,
		0.999705,
		0.996665,
		0.98496,
		0.958714,
		0.91379,
		0.848375,
		0.762966,
		0.66014,
		0.544295,
		0.421411,
		0.298855,
		0.185227,
		0.090239,
		0.024612,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.001584,
		0.036478,
		0.108382,
		0.2064,
		0.320521,
		0.441699,
		0.561943,
		0.674436,
		0.773688,
		0.85571,
		0.918247,
		0.961014,
		0.985898,
		0.996952,
		0.999812,
		0.999982,
		0.999536,
		0.995193,
		0.981167,
		0.952111,
		0.904491,
		0.836975,
		0.750362,
		0.647364,
		0.532363,
		0.411199,
		0.290993,
		0.180011,
		0.087547,
		0.023841,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.013263,
		0.06646,
		0.150396,
		0.255081,
		0.371384,
		0.491113,
		0.607109,
		0.713365,
		0.805178,
		0.879328,
		0.934294,
		0.970479,
		0.990351,
		0.99832,
		0.999955,
		0.999993,
		0.999102,
		0.993251,
		0.976834,
		0.94504,
		0.894886,
		0.825462,
		0.737825,
		0.634793,
		0.520717,
		0.401293,
		0.283404,
		0.174994,
		0.084966,
		0.023103,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.001486,
		0.034297,
		0.102209,
		0.195295,
		0.304401,
		0.42121,
		0.538323,
		0.649351,
		0.749038,
		0.833411,
		0.899959,
		0.947842,
		0.978079,
		0.993633,
		0.999156,
		0.999994,
		0.999959,
		0.998419,
		0.99089,
		0.972044,
		0.937598,
		0.88507,
		0.813918,
		0.72542,
		0.622473,
		0.509384,
		0.391707,
		0.276092,
		0.170178,
		0.082496,
		0.022399,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.012464,
		0.062632,
		0.142168,
		0.241943,
		0.353578,
		0.469512,
		0.583079,
		0.688606,
		0.781532,
		0.858565,
		0.91785,
		0.959167,
		0.984089,
		0.995985,
		0.999629,
		1,
		0.999859,
		0.997465,
		0.988116,
		0.966832,
		0.929834,
		0.875088,
		0.802379,
		0.71317,
		0.610413,
		0.498366,
		0.382435,
		0.269049,
		0.165555,
		0.080132,
		0.021727,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.001396,
		0.032305,
		0.096544,
		0.185043,
		0.289411,
		0.40199,
		0.515919,
		0.625212,
		0.72486,
		0.810953,
		0.880827,
		0.933236,
		0.968525,
		0.988753,
		0.997607,
		0.999867,
		1,
		0.999651,
		0.996208,
		0.984931,
		0.961225,
		0.921781,
		0.864974,
		0.790869,
		0.701086,
		0.598615,
		0.487653,
		0.373465,
		0.262263,
		0.161115,
		0.077867,
		0.021084,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.011736,
		0.059124,
		0.134589,
		0.229766,
		0.336951,
		0.449152,
		0.560159,
		0.664622,
		0.758149,
		0.837433,
		0.900395,
		0.946351,
		0.976158,
		0.992293,
		0.998674,
		0.999966,
		0.999995,
		0.999293,
		0.994625,
		0.981342,
		0.955247,
		0.913472,
		0.854755,
		0.779408,
		0.689177,
		0.587077,
		0.47724,
		0.364787,
		0.255722,
		0.156848,
		0.075696,
		0.02047,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.001314,
		0.030481,
		0.091333,
		0.175563,
		0.27546,
		0.383959,
		0.494689,
		0.602046,
		0.701268,
		0.788542,
		0.861121,
		0.917472,
		0.957423,
		0.982294,
		0.994912,
		0.999333,
		0.999996,
		0.999968,
		0.998747,
		0.9927,
		0.97736,
		0.948926,
		0.90494,
		0.84446,
		0.768013,
		0.677451,
		0.575799,
		0.467118,
		0.356389,
		0.249415,
		0.152746,
		0.073614,
		0.019882,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.011069,
		0.055902,
		0.127594,
		0.218463,
		0.321409,
		0.42996,
		0.538323,
		0.641458,
		0.735156,
		0.816138,
		0.882183,
		0.932261,
		0.966673,
		0.987147,
		0.99679,
		0.999707,
		1,
		0.999889,
		0.997978,
		0.990425,
		0.972999,
		0.94229,
		0.896212,
		0.834112,
		0.756699,
		0.665912,
		0.564777,
		0.457278,
		0.34826,
		0.243331,
		0.148801,
		0.071616,
		0.019319,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.001239,
		0.028807,
		0.086531,
		0.166781,
		0.262459,
		0.367032,
		0.474581,
		0.579853,
		0.678333,
		0.766326,
		0.841057,
		0.900792,
		0.944967,
		0.974313,
		0.990912,
		0.998086,
		0.999895,
		1,
		0.999723,
		0.996959,
		0.987794,
		0.968276,
		0.935365,
		0.887317,
		0.823732,
		0.745479,
		0.654566,
		0.554008,
		0.447714,
		0.34039,
		0.237461,
		0.145004,
		0.069699,
		0.018779,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.010458,
		0.052935,
		0.121124,
		0.207954,
		0.306869,
		0.411865,
		0.517538,
		0.61914,
		0.712649,
		0.794849,
		0.863435,
		0.917126,
		0.955789,
		0.980542,
		0.993771,
		0.998938,
		0.999973,
		0.999996,
		0.999435,
		0.995666,
		0.98481,
		0.963209,
		0.928179,
		0.87828,
		0.81334,
		0.734365,
		0.643416,
		0.54349,
		0.438416,
		0.332769,
		0.231794,
		0.141349,
		0.067857,
		0.018262,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.00117,
		0.027267,
		0.082095,
		0.158632,
		0.250328,
		0.351133,
		0.455537,
		0.558619,
		0.656102,
		0.744423,
		0.820817,
		0.883417,
		0.931363,
		0.964919,
		0.985548,
		0.995886,
		0.999465,
		0.999997,
		0.999975,
		0.998994,
		0.994084,
		0.981477,
		0.957819,
		0.920755,
		0.869125,
		0.802956,
		0.723368,
		0.632464,
		0.533218,
		0.429378,
		0.325389,
		0.226323,
		0.137829,
		0.066086,
		0.017765,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.009896,
		0.050198,
		0.11513,
		0.19817,
		0.293251,
		0.394798,
		0.497762,
		0.597675,
		0.690698,
		0.773702,
		0.84434,
		0.901155,
		0.943681,
		0.972543,
		0.989507,
		0.997403,
		0.999765,
		1,
		0.999911,
		0.998368,
		0.992201,
		0.977805,
		0.952128,
		0.913121,
		0.859876,
		0.792593,
		0.712494,
		0.621711,
		0.523188,
		0.420591,
		0.318238,
		0.221038,
		0.134438,
		0.064384,
		0.017289,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.001107,
		0.025848,
		0.07799,
		0.151058,
		0.238994,
		0.336187,
		0.437501,
		0.538323,
		0.634605,
		0.722924,
		0.800552,
		0.865538,
		0.916807,
		0.954256,
		0.978837,
		0.992579,
		0.998451,
		0.999916,
		1,
		0.999777,
		0.997533,
		0.990011,
		0.973805,
		0.946156,
		0.905299,
		0.850552,
		0.782268,
		0.701753,
		0.611159,
		0.513395,
		0.412048,
		0.31131,
		0.215931,
		0.131168,
		0.062746,
		0.016832,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.009378,
		0.047667,
		0.109566,
		0.189047,
		0.280485,
		0.378694,
		0.478955,
		0.577059,
		0.669354,
		0.752804,
		0.82506,
		0.884543,
		0.930531,
		0.963257,
		0.983968,
		0.994912,
		0.99914,
		0.999978,
		0.999997,
		0.999544,
		0.996468,
		0.987511,
		0.969489,
		0.939924,
		0.897313,
		0.841172,
		0.771993,
		0.69115,
		0.600807,
		0.503836,
		0.403741,
		0.304595,
		0.210995,
		0.128015,
		0.06117,
		0.016392,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.001049,
		0.024536,
		0.074184,
		0.144007,
		0.228392,
		0.322126,
		0.420418,
		0.518938,
		0.613858,
		0.701897,
		0.780384,
		0.847326,
		0.901488,
		0.942482,
		0.970847,
		0.988091,
		0.996638,
		0.999567,
		0.999997,
		0.99998,
		0.999183,
		0.995156,
		0.984704,
		0.964874,
		0.933455,
		0.889182,
		0.831754,
		0.761779,
		0.68069,
		0.590655,
		0.494505,
		0.395662,
		0.298086,
		0.206223,
		0.124973,
		0.059653,
		0.01597,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.0089,
		0.045322,
		0.104393,
		0.180527,
		0.268504,
		0.363491,
		0.461072,
		0.557281,
		0.648648,
		0.732241,
		0.805731,
		0.867459,
		0.916514,
		0.952816,
		0.977183,
		0.991352,
		0.997877,
		0.999809,
		1,
		0.999928,
		0.998669,
		0.993585,
		0.981594,
		0.959974,
		0.926769,
		0.880929,
		0.822315,
		0.751638,
		0.670377,
		0.580701,
		0.485397,
		0.387806,
		0.291774,
		0.201607,
		0.122037,
		0.05819,
		0.015563,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.000995,
		0.023321,
		0.070648,
		0.137432,
		0.218463,
		0.308888,
		0.404232,
		0.500431,
		0.593863,
		0.681392,
		0.760415,
		0.828924,
		0.885575,
		0.929759,
		0.961681,
		0.982413,
		0.993882,
		0.998733,
		0.999931,
		1,
		0.999819,
		0.997978,
		0.991748,
		0.978187,
		0.954807,
		0.919885,
		0.872571,
		0.812868,
		0.741577,
		0.660215,
		0.570945,
		0.476508,
		0.380165,
		0.285652,
		0.197141,
		0.119202,
		0.056781,
		0.015172,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.008457,
		0.043145,
		0.099574,
		0.172561,
		0.25725,
		0.349131,
		0.444067,
		0.538323,
		0.628601,
		0.712079,
		0.786465,
		0.850054,
		0.901795,
		0.941362,
		0.969224,
		0.986675,
		0.995804,
		0.999296,
		0.999982,
		0.999998,
		0.999627,
		0.997092,
		0.989641,
		0.974494,
		0.949389,
		0.912824,
		0.864126,
		0.803427,
		0.731605,
		0.650207,
		0.561384,
		0.467831,
		0.372731,
		0.279714,
		0.192819,
		0.116464,
		0.055422,
		0.014796,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.000945,
		0.022194,
		0.067358,
		0.131291,
		0.209152,
		0.296414,
		0.388893,
		0.482769,
		0.574618,
		0.661443,
		0.740722,
		0.810454,
		0.869219,
		0.916242,
		0.951461,
		0.975582,
		0.990101,
		0.997227,
		0.999645,
		0.999998,
		0.999983,
		0.99933,
		0.995996,
		0.987261,
		0.970524,
		0.943738,
		0.905603,
		0.85561,
		0.794004,
		0.721729,
		0.640354,
		0.552017,
		0.459363,
		0.365499,
		0.273952,
		0.188634,
		0.113819,
		0.054111,
		0.014433,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.008046,
		0.041121,
		0.09508,
		0.165103,
		0.246666,
		0.335557,
		0.427896,
		0.520162,
		0.609221,
		0.692366,
		0.767353,
		0.832457,
		0.886525,
		0.929041,
		0.960187,
		0.980889,
		0.992809,
		0.998248,
		0.999844,
		1,
		0.999941,
		0.998903,
		0.994676,
		0.984612,
		0.966289,
		0.93787,
		0.898242,
		0.84704,
		0.784611,
		0.711954,
		0.630658,
		0.542841,
		0.451099,
		0.358463,
		0.268359,
		0.184582,
		0.111261,
		0.052846,
		0.014083,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.000899,
		0.021147,
		0.064291,
		0.12555,
		0.200411,
		0.28465,
		0.374349,
		0.465916,
		0.556109,
		0.642073,
		0.721369,
		0.792019,
		0.852555,
		0.902078,
		0.940314,
		0.967671,
		0.985268,
		0.994912,
		0.998953,
		0.999944,
		1,
		0.999851,
		0.998327,
		0.993125,
		0.981696,
		0.961803,
		0.931803,
		0.890756,
		0.838428,
		0.775256,
		0.702287,
		0.62112,
		0.533853,
		0.443032,
		0.351615,
		0.262931,
		0.180656,
		0.108788,
		0.051625,
		0.013746,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.007665,
		0.039236,
		0.09088,
		0.15811,
		0.236704,
		0.32272,
		0.412516,
		0.502771,
		0.590511,
		0.673139,
		0.74847,
		0.814777,
		0.870838,
		0.91599,
		0.950184,
		0.974035,
		0.988837,
		0.996509,
		0.999419,
		0.999986,
		0.999998,
		0.999693,
		0.997585,
		0.991337,
		0.978519,
		0.957078,
		0.925553,
		0.883162,
		0.829789,
		0.76595,
		0.692731,
		0.61174,
		0.52505,
		0.435159,
		0.34495,
		0.257659,
		0.176852,
		0.106396,
		0.050445,
		0.013421,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.000857,
		0.020173,
		0.061429,
		0.120172,
		0.192195,
		0.273547,
		0.360554,
		0.449834,
		0.538323,
		0.623297,
		0.702404,
		0.773702,
		0.835698,
		0.887401,
		0.928371,
		0.958769,
		0.9794,
		0.991705,
		0.997692,
		0.999707,
		0.999998,
		0.999986,
		0.999445,
		0.996662,
		0.989308,
		0.975089,
		0.952128,
		0.919137,
		0.875474,
		0.821134,
		0.756699,
		0.683291,
		0.602518,
		0.51643,
		0.427474,
		0.338462,
		0.25254,
		0.173165,
		0.104082,
		0.049305,
		0.013107,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.00731,
		0.037478,
		0.086952,
		0.151546,
		0.227317,
		0.31057,
		0.397884,
		0.486123,
		0.572465,
		0.654422,
		0.729875,
		0.797109,
		0.854856,
		0.902341,
		0.939331,
		0.966185,
		0.983876,
		0.993974,
		0.998541,
		0.999871,
		1,
		0.999951,
		0.999088,
		0.995546,
		0.987038,
		0.971415,
		0.946968,
		0.912568,
		0.867708,
		0.812474,
		0.747512,
		0.673969,
		0.593455,
		0.50799,
		0.419973,
		0.332145,
		0.247567,
		0.16959,
		0.101841,
		0.048204,
		0.012805,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000817,
		0.019264,
		0.058753,
		0.11513,
		0.184465,
		0.263058,
		0.347461,
		0.434489,
		0.52124,
		0.605119,
		0.683863,
		0.755573,
		0.818748,
		0.872333,
		0.915755,
		0.94898,
		0.972543,
		0.987569,
		0.995735,
		0.999128,
		0.999953,
		1,
		0.999876,
		0.998605,
		0.994227,
		0.98453,
		0.967506,
		0.94161,
		0.905864,
		0.859876,
		0.80382,
		0.738394,
		0.664769,
		0.584549,
		0.499726,
		0.41265,
		0.325995,
		0.242735,
		0.166123,
		0.099671,
		0.047138,
		0.012512,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.006979,
		0.035835,
		0.08327,
		0.145377,
		0.218463,
		0.299063,
		0.383959,
		0.470189,
		0.555074,
		0.636233,
		0.711615,
		0.77953,
		0.838682,
		0.888213,
		0.927745,
		0.957423,
		0.977948,
		0.990579,
		0.997073,
		0.999515,
		0.999988,
		0.999998,
		0.999744,
		0.997978,
		0.9927,
		0.981785,
		0.963372,
		0.93607,
		0.899037,
		0.85199,
		0.795181,
		0.729352,
		0.655692,
		0.575799,
		0.491635,
		0.405501,
		0.320005,
		0.238038,
		0.162759,
		0.09757,
		0.046108,
		0.01223,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.00078,
		0.018415,
		0.056247,
		0.110396,
		0.177183,
		0.253142,
		0.33503,
		0.419845,
		0.504839,
		0.58754,
		0.665773,
		0.737688,
		0.801791,
		0.85698,
		0.902584,
		0.938407,
		0.964764,
		0.982503,
		0.992999,
		0.998064,
		0.999755,
		0.999998,
		0.999989,
		0.999536,
		0.997195,
		0.990961,
		0.97881,
		0.959025,
		0.930361,
		0.892101,
		0.844063,
		0.786565,
		0.72039,
		0.64674,
		0.567204,
		0.483714,
		0.398522,
		0.31417,
		0.233473,
		0.159495,
		0.095534,
		0.045111,
		0.011957,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.006671,
		0.034297,
		0.079817,
		0.139573,
		0.210104,
		0.288158,
		0.370703,
		0.45494,
		0.538323,
		0.618579,
		0.693726,
		0.762104,
		0.822408,
		0.873716,
		0.915536,
		0.947842,
		0.971103,
		0.986304,
		0.994912,
		0.998776,
		0.999892,
		1,
		0.999959,
		0.999236,
		0.996245,
		0.989007,
		0.975609,
		0.954475,
		0.924495,
		0.88507,
		0.836105,
		0.77798,
		0.711512,
		0.637915,
		0.558763,
		0.475959,
		0.391707,
		0.308487,
		0.229035,
		0.156326,
		0.09356,
		0.044146,
		0.011692,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000745,
		0.017621,
		0.053897,
		0.105945,
		0.170317,
		0.243759,
		0.32322,
		0.405867,
		0.489096,
		0.570554,
		0.648151,
		0.720092,
		0.784898,
		0.841438,
		0.888967,
		0.927158,
		0.956143,
		0.976536,
		0.98944,
		0.996398,
		0.999268,
		0.999961,
		1,
		0.999897,
		0.998827,
		0.995118,
		0.986838,
		0.972191,
		0.949734,
		0.918486,
		0.877955,
		0.828126,
		0.769433,
		0.702724,
		0.629217,
		0.550474,
		0.468368,
		0.385053,
		0.302949,
		0.224718,
		0.15325,
		0.091646,
		0.043211,
		0.011437,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.006382,
		0.032856,
		0.076573,
		0.134105,
		0.202206,
		0.277816,
		0.358077,
		0.440344,
		0.522198,
		0.601465,
		0.676234,
		0.744885,
		0.80611,
		0.858946,
		0.902812,
		0.937538,
		0.963404,
		0.981155,
		0.991996,
		0.997527,
		0.999593,
		0.99999,
		0.999999,
		0.999785,
		0.998294,
		0.993807,
		0.984457,
		0.968563,
		0.944814,
		0.912347,
		0.870768,
		0.820135,
		0.760931,
		0.694027,
		0.620647,
		0.542336,
		0.460936,
		0.378555,
		0.297552,
		0.220519,
		0.150262,
		0.089791,
		0.042305,
		0.01119,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.000713,
		0.016877,
		0.051692,
		0.101756,
		0.163836,
		0.234873,
		0.311993,
		0.392521,
		0.473988,
		0.554152,
		0.63101,
		0.702821,
		0.768129,
		0.82579,
		0.875,
		0.915332,
		0.946765,
		0.969716,
		0.985047,
		0.99405,
		0.998364,
		0.999794,
		0.999999,
		0.99999,
		0.99961,
		0.997627,
		0.992308,
		0.981865,
		0.964734,
		0.939726,
		0.906089,
		0.863519,
		0.812141,
		0.75248,
		0.685426,
		0.612205,
		0.534347,
		0.45366,
		0.372208,
		0.292293,
		0.216434,
		0.14736,
		0.08799,
		0.041428,
		0.010951,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.006111,
		0.031504,
		0.073521,
		0.128949,
		0.194735,
		0.268,
		0.346048,
		0.426374,
		0.50668,
		0.584888,
		0.65916,
		0.727916,
		0.789856,
		0.843991,
		0.889668,
		0.926607,
		0.954925,
		0.975164,
		0.988294,
		0.995675,
		0.998965,
		0.999909,
		1,
		0.999966,
		0.999355,
		0.996812,
		0.990617,
		0.979066,
		0.960712,
		0.934481,
		0.899725,
		0.85622,
		0.804153,
		0.744084,
		0.676922,
		0.603892,
		0.526505,
		0.446537,
		0.366009,
		0.287165,
		0.212459,
		0.14454,
		0.086243,
		0.040578,
		0.010719,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000682,
		0.016179,
		0.049618,
		0.097809,
		0.157713,
		0.226451,
		0.301315,
		0.379775,
		0.45949,
		0.538323,
		0.614355,
		0.685901,
		0.751535,
		0.810106,
		0.860772,
		0.903024,
		0.936718,
		0.962103,
		0.979834,
		0.990973,
		0.996937,
		0.999381,
		0.999967,
		1,
		0.999913,
		0.999006,
		0.995843,
		0.988734,
		0.976067,
		0.956507,
		0.92909,
		0.893264,
		0.84888,
		0.796177,
		0.735749,
		0.668518,
		0.595706,
		0.518807,
		0.439564,
		0.359954,
		0.282166,
		0.208589,
		0.141799,
		0.084547,
		0.039753,
		0.010495,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.005858,
		0.030233,
		0.070648,
		0.124081,
		0.187663,
		0.258679,
		0.334582,
		0.413,
		0.491749,
		0.568845,
		0.642517,
		0.71123,
		0.773702,
		0.828924,
		0.876195,
		0.91514,
		0.945745,
		0.968379,
		0.983803,
		0.993157,
		0.997897,
		0.999656,
		0.999992,
		0.999999,
		0.999819,
		0.998551,
		0.994712,
		0.986658,
		0.972873,
		0.952128,
		0.923565,
		0.886718,
		0.841508,
		0.78822,
		0.727478,
		0.660215,
		0.587648,
		0.511251,
		0.432738,
		0.354039,
		0.277292,
		0.204822,
		0.139134,
		0.0829,
		0.038954,
		0.010277,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.000654,
		0.015523,
		0.047667,
		0.094085,
		0.151921,
		0.218463,
		0.291153,
		0.367599,
		0.445578,
		0.523054,
		0.598188,
		0.669354,
		0.735156,
		0.794448,
		0.846361,
		0.890323,
		0.926089,
		0.953766,
		0.973833,
		0.987147,
		0.994912,
		0.998608,
		0.999826,
		0.999999,
		0.999992,
		0.999669,
		0.997978,
		0.993413,
		0.984391,
		0.969489,
		0.947585,
		0.917916,
		0.880097,
		0.834112,
		0.780288,
		0.719275,
		0.652015,
		0.579717,
		0.503836,
		0.426054,
		0.34826,
		0.272538,
		0.201154,
		0.136543,
		0.081301,
		0.038178,
		0.010067,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.00562,
		0.029038,
		0.067939,
		0.119481,
		0.180962,
		0.249819,
		0.323646,
		0.400195,
		0.477387,
		0.553326,
		0.626313,
		0.694856,
		0.757694,
		0.813813,
		0.862471,
		0.903223,
		0.935944,
		0.960857,
		0.978541,
		0.989936,
		0.996301,
		0.999119,
		0.999923,
		1,
		0.999971,
		0.999452,
		0.997277,
		0.991944,
		0.981935,
		0.965925,
		0.942888,
		0.912153,
		0.873411,
		0.826699,
		0.772387,
		0.711144,
		0.643919,
		0.571912,
		0.496559,
		0.41951,
		0.342613,
		0.267901,
		0.197581,
		0.134022,
		0.079747,
		0.037425,
		0.009862,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000627,
		0.014907,
		0.045828,
		0.090569,
		0.146439,
		0.21088,
		0.281477,
		0.355962,
		0.432226,
		0.508329,
		0.582507,
		0.653193,
		0.719025,
		0.778869,
		0.831837,
		0.877309,
		0.91496,
		0.944778,
		0.967092,
		0.982575,
		0.992239,
		0.99738,
		0.999473,
		0.999972,
		1,
		0.999926,
		0.999153,
		0.996439,
		0.990301,
		0.979295,
		0.962186,
		0.938045,
		0.906287,
		0.866667,
		0.819279,
		0.764523,
		0.703088,
		0.635927,
		0.564232,
		0.489418,
		0.413102,
		0.337094,
		0.263377,
		0.194101,
		0.13157,
		0.078237,
		0.036694,
		0.009664,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.005396,
		0.027912,
		0.065382,
		0.11513,
		0.174607,
		0.241394,
		0.313213,
		0.387931,
		0.463572,
		0.538323,
		0.610552,
		0.678813,
		0.741872,
		0.798713,
		0.848568,
		0.890936,
		0.925601,
		0.952661,
		0.972543,
		0.986004,
		0.994115,
		0.9982,
		0.999707,
		0.999993,
		0.999999,
		0.999846,
		0.998762,
		0.995458,
		0.988485,
		0.976473,
		0.95828,
		0.933068,
		0.900326,
		0.859876,
		0.811856,
		0.756699,
		0.695109,
		0.628041,
		0.556677,
		0.48241,
		0.406828,
		0.331701,
		0.258963,
		0.19071,
		0.129184,
		0.07677,
		0.035984,
		0.009472,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.000602,
		0.014327,
		0.044093,
		0.087245,
		0.141244,
		0.203676,
		0.272257,
		0.344836,
		0.419412,
		0.494131,
		0.567308,
		0.637429,
		0.70317,
		0.763413,
		0.81726,
		0.864056,
		0.903409,
		0.935212,
		0.959663,
		0.977279,
		0.988892,
		0.995624,
		0.998808,
		0.999852,
		0.999999,
		0.999993,
		0.999718,
		0.998267,
		0.994328,
		0.986495,
		0.973475,
		0.954217,
		0.927963,
		0.894279,
		0.853044,
		0.804438,
		0.748921,
		0.68721,
		0.620261,
		0.549244,
		0.475533,
		0.400685,
		0.326429,
		0.254655,
		0.187405,
		0.126861,
		0.075343,
		0.035295,
		0.009285,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.005185,
		0.02685,
		0.062966,
		0.11101,
		0.168575,
		0.233376,
		0.303253,
		0.376183,
		0.450283,
		0.523824,
		0.595234,
		0.663118,
		0.726267,
		0.783673,
		0.834551,
		0.878352,
		0.91479,
		0.943859,
		0.965851,
		0.981366,
		0.991302,
		0.996818,
		0.999246,
		0.999934,
		1,
		0.999975,
		0.999531,
		0.99766,
		0.993045,
		0.984332,
		0.970307,
		0.950003,
		0.922742,
		0.888157,
		0.84618,
		0.797031,
		0.741192,
		0.679392,
		0.612586,
		0.541933,
		0.468785,
		0.394668,
		0.321275,
		0.25045,
		0.184184,
		0.1246,
		0.073955,
		0.034625,
		0.009104,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.000579,
		0.013779,
		0.042454,
		0.084099,
		0.136317,
		0.196828,
		0.263466,
		0.334196,
		0.407111,
		0.480445,
		0.552583,
		0.622067,
		0.687611,
		0.748116,
		0.802682,
		0.850628,
		0.89151,
		0.92514,
		0.951607,
		0.971292,
		0.984868,
		0.993291,
		0.997745,
		0.999548,
		0.999976,
		1,
		0.999937,
		0.999274,
		0.996932,
		0.991607,
		0.981999,
		0.966975,
		0.945646,
		0.917412,
		0.881966,
		0.839291,
		0.789639,
		0.733515,
		0.671658,
		0.605018,
		0.534743,
		0.462163,
		0.388776,
		0.316237,
		0.246346,
		0.181043,
		0.122398,
		0.072605,
		0.033974,
		0.008928,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.004986,
		0.025848,
		0.060681,
		0.107106,
		0.162846,
		0.22574,
		0.29374,
		0.364925,
		0.437501,
		0.509814,
		0.580358,
		0.647781,
		0.710906,
		0.768736,
		0.820474,
		0.865538,
		0.903584,
		0.934518,
		0.958519,
		0.976047,
		0.987845,
		0.994912,
		0.998451,
		0.999749,
		0.999994,
		0.999999,
		0.999868,
		0.998935,
		0.996078,
		0.990011,
		0.979499,
		0.963484,
		0.941155,
		0.911982,
		0.875716,
		0.832383,
		0.782268,
		0.725895,
		0.664008,
		0.597555,
		0.527672,
		0.455665,
		0.383005,
		0.31131,
		0.242338,
		0.177981,
		0.120254,
		0.071291,
		0.033341,
		0.008758,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.000557,
		0.013263,
		0.040905,
		0.081119,
		0.131641,
		0.190312,
		0.255081,
		0.324015,
		0.395301,
		0.467253,
		0.538323,
		0.607109,
		0.672365,
		0.73301,
		0.78815,
		0.837084,
		0.879329,
		0.91463,
		0.942986,
		0.964657,
		0.980179,
		0.990352,
		0.996216,
		0.998974,
		0.999873,
		0.999999,
		0.999994,
		0.999758,
		0.998506,
		0.99509,
		0.988257,
		0.976835,
		0.959841,
		0.936538,
		0.90646,
		0.869413,
		0.825463,
		0.774923,
		0.718333,
		0.656444,
		0.590198,
		0.520718,
		0.449289,
		0.377353,
		0.306492,
		0.238425,
		0.174994,
		0.118164,
		0.070013,
		0.032726,
		0.008592,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.004799,
		0.0249,
		0.058518,
		0.103402,
		0.1574,
		0.218463,
		0.28465,
		0.354133,
		0.425205,
		0.496282,
		0.565918,
		0.63281,
		0.69581,
		0.753936,
		0.806386,
		0.852555,
		0.892049,
		0.924704,
		0.950601,
		0.970081,
		0.983744,
		0.992444,
		0.997247,
		0.99935,
		0.999944,
		1,
		0.999979,
		0.999596,
		0.997978,
		0.993965,
		0.986346,
		0.974012,
		0.956054,
		0.931803,
		0.900855,
		0.863065,
		0.818536,
		0.767608,
		0.710833,
		0.648967,
		0.582946,
		0.51388,
		0.443032,
		0.371816,
		0.301781,
		0.234603,
		0.172081,
		0.116129,
		0.068768,
		0.032127,
		0.00843,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.000536,
		0.012775,
		0.039439,
		0.078295,
		0.127199,
		0.184108,
		0.247077,
		0.31427,
		0.383959,
		0.454537,
		0.524519,
		0.592557,
		0.657442,
		0.71812,
		0.773702,
		0.823476,
		0.866927,
		0.903749,
		0.93386,
		0.957423,
		0.974848,
		0.986798,
		0.994171,
		0.998049,
		0.999611,
		0.99998,
		1,
		0.999946,
		0.999373,
		0.997343,
		0.9927,
		0.984278,
		0.971035,
		0.952128,
		0.926957,
		0.895174,
		0.856678,
		0.811609,
		0.760326,
		0.703397,
		0.641577,
		0.575799,
		0.507156,
		0.436892,
		0.366393,
		0.297172,
		0.230869,
		0.169238,
		0.114144,
		0.067557,
		0.031545,
		0.008273,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.004622,
		0.024004,
		0.056468,
		0.099885,
		0.152218,
		0.211524,
		0.27596,
		0.343786,
		0.413374,
		0.483212,
		0.551909,
		0.618209,
		0.680995,
		0.739302,
		0.79233,
		0.839455,
		0.880246,
		0.914479,
		0.942154,
		0.963506,
		0.979014,
		0.989394,
		0.995579,
		0.998659,
		0.999783,
		0.999995,
		0.999999,
		0.999886,
		0.999079,
		0.996595,
		0.991293,
		0.982057,
		0.967907,
		0.948071,
		0.922008,
		0.889423,
		0.850259,
		0.804685,
		0.753082,
		0.696026,
		0.634275,
		0.568755,
		0.500544,
		0.430867,
		0.361079,
		0.292664,
		0.227222,
		0.166465,
		0.11221,
		0.066377,
		0.030978,
		0.008121,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.000516,
		0.012313,
		0.038051,
		0.075614,
		0.122976,
		0.178197,
		0.239433,
		0.304937,
		0.373065,
		0.442279,
		0.51116,
		0.578407,
		0.642853,
		0.703467,
		0.759371,
		0.809849,
		0.854361,
		0.892557,
		0.924291,
		0.94964,
		0.968908,
		0.982633,
		0.991581,
		0.996711,
		0.999112,
		0.99989,
		0.999999,
		0.999995,
		0.999791,
		0.998706,
		0.995728,
		0.989742,
		0.979683,
		0.964635,
		0.943889,
		0.916963,
		0.883611,
		0.843815,
		0.797771,
		0.745878,
		0.688723,
		0.627063,
		0.561815,
		0.494044,
		0.424954,
		0.355873,
		0.288253,
		0.223658,
		0.163757,
		0.110324,
		0.065227,
		0.030427,
		0.007973,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.004454,
		0.023155,
		0.054523,
		0.096544,
		0.147285,
		0.204903,
		0.267647,
		0.333861,
		0.40199,
		0.470589,
		0.538323,
		0.603979,
		0.666474,
		0.72486,
		0.778343,
		0.826287,
		0.868232,
		0.903905,
		0.933236,
		0.956371,
		0.973679,
		0.985756,
		0.993406,
		0.997607,
		0.999437,
		0.999951,
		1,
		0.999982,
		0.999651,
		0.998244,
		0.994738,
		0.988048,
		0.977161,
		0.961225,
		0.93959,
		0.91183,
		0.877744,
		0.83735,
		0.790869,
		0.738718,
		0.681489,
		0.619939,
		0.554977,
		0.487653,
		0.41915,
		0.350772,
		0.283937,
		0.220175,
		0.161115,
		0.108485,
		0.064107,
		0.02989,
		0.007828,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.000497,
		0.011876,
		0.036734,
		0.073068,
		0.118957,
		0.172561,
		0.232129,
		0.295997,
		0.362598,
		0.430464,
		0.498233,
		0.564656,
		0.628601,
		0.689066,
		0.745185,
		0.796242,
		0.841678,
		0.881109,
		0.914337,
		0.941362,
		0.962398,
		0.977873,
		0.988431,
		0.994912,
		0.998304,
		0.999663,
		0.999982,
		1,
		0.999953,
		0.999457,
		0.997688,
		0.993622,
		0.98621,
		0.974494,
		0.957681,
		0.935179,
		0.906615,
		0.871829,
		0.83087,
		0.783986,
		0.731605,
		0.674326,
		0.612904,
		0.54824,
		0.48137,
		0.413455,
		0.345773,
		0.279714,
		0.21677,
		0.158534,
		0.106691,
		0.063015,
		0.029367,
		0.007688,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.004296,
		0.02235,
		0.052677,
		0.093366,
		0.142584,
		0.198581,
		0.259691,
		0.324337,
		0.391034,
		0.4584,
		0.525152,
		0.59012,
		0.652255,
		0.710629,
		0.764456,
		0.813093,
		0.856057,
		0.893035,
		0.9239,
		0.948721,
		0.967772,
		0.981538,
		0.990705,
		0.99614,
		0.998834,
		0.999812,
		0.999996,
		0.999999,
		0.999901,
		0.9992,
		0.99703,
		0.992377,
		0.98423,
		0.971686,
		0.954009,
		0.930665,
		0.901325,
		0.865871,
		0.824381,
		0.777124,
		0.72454,
		0.667234,
		0.605958,
		0.541604,
		0.475193,
		0.407865,
		0.340874,
		0.27558,
		0.213443,
		0.156015,
		0.104941,
		0.061951,
		0.028857,
		0.007551,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.000479,
		0.011462,
		0.035484,
		0.070648,
		0.11513,
		0.167184,
		0.225145,
		0.287427,
		0.352538,
		0.419073,
		0.485726,
		0.551297,
		0.61469,
		0.674931,
		0.73117,
		0.78269,
		0.828924,
		0.869459,
		0.904051,
		0.932642,
		0.955361,
		0.972543,
		0.984721,
		0.992621,
		0.997128,
		0.999227,
		0.999905,
		0.999999,
		0.999996,
		0.999819,
		0.998873,
		0.996266,
		0.991001,
		0.982109,
		0.968741,
		0.950217,
		0.926053,
		0.895967,
		0.859876,
		0.817887,
		0.770286,
		0.717527,
		0.660215,
		0.599101,
		0.535067,
		0.469119,
		0.402378,
		0.336073,
		0.271535,
		0.210189,
		0.153554,
		0.103233,
		0.060913,
		0.028361,
		0.007418,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.004145,
		0.021587,
		0.050923,
		0.090341,
		0.138102,
		0.192542,
		0.252073,
		0.315194,
		0.380488,
		0.446628,
		0.512384,
		0.57663,
		0.638345,
		0.696627,
		0.750696,
		0.79991,
		0.843766,
		0.881923,
		0.914202,
		0.940607,
		0.961329,
		0.976757,
		0.987468,
		0.99422,
		0.99791,
		0.99951,
		0.999958,
		1,
		0.999984,
		0.999696,
		0.998468,
		0.995392,
		0.989494,
		0.97985,
		0.965663,
		0.946308,
		0.92135,
		0.890546,
		0.85385,
		0.811392,
		0.763477,
		0.710566,
		0.65327,
		0.592333,
		0.528629,
		0.463149,
		0.396993,
		0.331368,
		0.267574,
		0.207007,
		0.151149,
		0.101566,
		0.059902,
		0.027878,
		0.007289,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.000462,
		0.011069,
		0.034297,
		0.068345,
		0.111483,
		0.162051,
		0.218463,
		0.27921,
		0.342867,
		0.40809,
		0.473628,
		0.538323,
		0.601121,
		0.661072,
		0.717343,
		0.769224,
		0.816138,
		0.857652,
		0.893487,
		0.923529,
		0.947842,
		0.966673,
		0.980459,
		0.989821,
		0.995539,
		0.998518,
		0.999707,
		0.999985,
		1,
		0.999959,
		0.999527,
		0.997978,
		0.994403,
		0.987855,
		0.977455,
		0.962458,
		0.94229,
		0.916561,
		0.88507,
		0.847798,
		0.804901,
		0.756699,
		0.703661,
		0.646399,
		0.585653,
		0.522288,
		0.457278,
		0.391707,
		0.326755,
		0.263696,
		0.203896,
		0.148801,
		0.099939,
		0.058915,
		0.027406,
		0.007163,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.004003,
		0.020862,
		0.049255,
		0.087461,
		0.133826,
		0.186768,
		0.244776,
		0.306416,
		0.370333,
		0.435259,
		0.500011,
		0.563503,
		0.624748,
		0.682865,
		0.737086,
		0.78677,
		0.831403,
		0.870615,
		0.904191,
		0.932076,
		0.954391,
		0.971437,
		0.983695,
		0.991821,
		0.996615,
		0.998981,
		0.999836,
		0.999996,
		0.999999,
		0.999914,
		0.999302,
		0.997398,
		0.993298,
		0.986085,
		0.974928,
		0.959131,
		0.938168,
		0.911694,
		0.879542,
		0.841725,
		0.798418,
		0.749955,
		0.696813,
		0.639602,
		0.579062,
		0.516043,
		0.451507,
		0.386518,
		0.322233,
		0.2599,
		0.200852,
		0.146505,
		0.09835,
		0.057952,
		0.026947,
		0.00704,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.000447,
		0.010696,
		0.033168,
		0.066153,
		0.108006,
		0.157147,
		0.212067,
		0.271328,
		0.333566,
		0.397498,
		0.461923,
		0.525728,
		0.587893,
		0.647496,
		0.703721,
		0.755868,
		0.803355,
		0.845732,
		0.88269,
		0.914074,
		0.939885,
		0.960299,
		0.975665,
		0.986506,
		0.993506,
		0.997481,
		0.999324,
		0.999917,
		1,
		0.999996,
		0.999842,
		0.999014,
		0.996723,
		0.992073,
		0.984185,
		0.972271,
		0.955686,
		0.933948,
		0.906753,
		0.87397,
		0.835636,
		0.791947,
		0.743248,
		0.690022,
		0.632882,
		0.572558,
		0.509893,
		0.445833,
		0.381424,
		0.317799,
		0.256181,
		0.197875,
		0.144262,
		0.096799,
		0.057012,
		0.026499,
		0.00692,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.003868,
		0.020173,
		0.047667,
		0.084714,
		0.129743,
		0.181245,
		0.237781,
		0.297982,
		0.360554,
		0.424278,
		0.488022,
		0.550737,
		0.611467,
		0.669354,
		0.723646,
		0.773702,
		0.819002,
		0.859156,
		0.893914,
		0.923176,
		0.946999,
		0.965609,
		0.9794,
		0.988931,
		0.994912,
		0.998167,
		0.999572,
		0.999963,
		1,
		0.999986,
		0.999735,
		0.998657,
		0.995947,
		0.990728,
		0.982156,
		0.969489,
		0.952128,
		0.929636,
		0.901745,
		0.868358,
		0.829535,
		0.78549,
		0.736579,
		0.683291,
		0.626236,
		0.56614,
		0.503836,
		0.440255,
		0.376422,
		0.313452,
		0.25254,
		0.194962,
		0.142069,
		0.095283,
		0.056095,
		0.026062,
		0.006803,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000431,
		0.010342,
		0.032094,
		0.064064,
		0.104687,
		0.152459,
		0.205942,
		0.263763,
		0.324619,
		0.387283,
		0.450601,
		0.513503,
		0.575003,
		0.634207,
		0.690319,
		0.742645,
		0.790604,
		0.833736,
		0.871707,
		0.904322,
		0.931537,
		0.953459,
		0.970362,
		0.982681,
		0.991008,
		0.996073,
		0.9987,
		0.999743,
		0.999987,
		1,
		0.999965,
		0.999586,
		0.998225,
		0.995069,
		0.989263,
		0.980001,
		0.966586,
		0.948463,
		0.925236,
		0.896676,
		0.862711,
		0.823426,
		0.779051,
		0.729952,
		0.67662,
		0.619667,
		0.55981,
		0.497872,
		0.43477,
		0.371512,
		0.309189,
		0.248973,
		0.192112,
		0.139926,
		0.093803,
		0.0552,
		0.025636,
		0.006689,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.003739,
		0.019517,
		0.046154,
		0.082095,
		0.125842,
		0.17596,
		0.231074,
		0.289878,
		0.351133,
		0.413672,
		0.476406,
		0.538323,
		0.5985,
		0.656102,
		0.71039,
		0.76073,
		0.806596,
		0.847584,
		0.883417,
		0.913952,
		0.939196,
		0.959306,
		0.974599,
		0.985548,
		0.992774,
		0.99702,
		0.999106,
		0.999857,
		0.999997,
		1,
		0.999925,
		0.999388,
		0.997711,
		0.994084,
		0.987677,
		0.977722,
		0.963565,
		0.944695,
		0.920755,
		0.891549,
		0.857034,
		0.817314,
		0.772634,
		0.723368,
		0.670011,
		0.613173,
		0.553565,
		0.491999,
		0.429378,
		0.366691,
		0.305009,
		0.245479,
		0.189322,
		0.137829,
		0.092357,
		0.054327,
		0.02522,
		0.006578,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.000417,
		0.010005,
		0.031071,
		0.062072,
		0.101517,
		0.147975,
		0.200072,
		0.256499,
		0.31601,
		0.377429,
		0.439648,
		0.501639,
		0.562448,
		0.621209,
		0.677145,
		0.729573,
		0.777914,
		0.821699,
		0.860575,
		0.894319,
		0.92284,
		0.946192,
		0.964579,
		0.978359,
		0.988039,
		0.994262,
		0.997782,
		0.999407,
		0.999927,
		1,
		0.999997,
		0.999861,
		0.999134,
		0.997111,
		0.992991,
		0.98597,
		0.975321,
		0.960431,
		0.94083,
		0.916199,
		0.886372,
		0.851331,
		0.811201,
		0.76624,
		0.716828,
		0.663465,
		0.606755,
		0.547405,
		0.486216,
		0.424076,
		0.361957,
		0.300909,
		0.242056,
		0.186592,
		0.135779,
		0.090944,
		0.053473,
		0.024814,
		0.00647,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.003617,
		0.018893,
		0.044712,
		0.079594,
		0.122113,
		0.170898,
		0.224639,
		0.282087,
		0.342055,
		0.403426,
		0.465151,
		0.526257,
		0.585848,
		0.643115,
		0.697333,
		0.747875,
		0.794215,
		0.835936,
		0.872739,
		0.904447,
		0.931022,
		0.952563,
		0.969317,
		0.98168,
		0.990188,
		0.995504,
		0.998385,
		0.999624,
		0.999968,
		1,
		0.999988,
		0.999767,
		0.998818,
		0.996422,
		0.991788,
		0.984144,
		0.972801,
		0.957188,
		0.936872,
		0.911571,
		0.881147,
		0.845607,
		0.805092,
		0.759872,
		0.710335,
		0.656981,
		0.600413,
		0.54133,
		0.480522,
		0.418863,
		0.357308,
		0.296887,
		0.238701,
		0.183919,
		0.133774,
		0.089563,
		0.05264,
		0.024418,
		0.006364,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000403,
		0.009684,
		0.030097,
		0.060171,
		0.098489,
		0.143683,
		0.194445,
		0.249522,
		0.307723,
		0.36792,
		0.429052,
		0.490126,
		0.550223,
		0.608503,
		0.664208,
		0.716668,
		0.765307,
		0.809651,
		0.849334,
		0.884104,
		0.913836,
		0.938536,
		0.958348,
		0.973558,
		0.984597,
		0.992029,
		0.996529,
		0.998854,
		0.999775,
		0.999988,
		1,
		0.999969,
		0.999636,
		0.998435,
		0.995638,
		0.990474,
		0.9822,
		0.970166,
		0.95384,
		0.932827,
		0.906878,
		0.875881,
		0.839866,
		0.79899,
		0.753533,
		0.70389,
		0.650561,
		0.594146,
		0.535338,
		0.474914,
		0.413737,
		0.352743,
		0.292943,
		0.235414,
		0.181302,
		0.131812,
		0.088212,
		0.051826,
		0.024031,
		0.006261,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.003501,
		0.018298,
		0.043337,
		0.077206,
		0.118545,
		0.166048,
		0.218463,
		0.274594,
		0.333306,
		0.393526,
		0.454246,
		0.514529,
		0.573509,
		0.630396,
		0.684484,
		0.735156,
		0.781884,
		0.824244,
		0.861918,
		0.894703,
		0.922519,
		0.945417,
		0.963582,
		0.977338,
		0.987147,
		0.993593,
		0.997366,
		0.999212,
		0.999874,
		0.999997,
		1,
		0.999934,
		0.999461,
		0.997978,
		0.994759,
		0.989048,
		0.98014,
		0.967419,
		0.950392,
		0.9287,
		0.902123,
		0.870578,
		0.834112,
		0.792898,
		0.747226,
		0.697495,
		0.644206,
		0.587956,
		0.529428,
		0.469393,
		0.408697,
		0.34826,
		0.289073,
		0.232193,
		0.178739,
		0.129893,
		0.086892,
		0.05103,
		0.023653,
		0.006161,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.00039,
		0.009378,
		0.029167,
		0.058356,
		0.095593,
		0.139573,
		0.189047,
		0.242817,
		0.299743,
		0.358744,
		0.4188,
		0.478955,
		0.538323,
		0.596089,
		0.651515,
		0.703943,
		0.752804,
		0.79762,
		0.838015,
		0.873716,
		0.904566,
		0.930531,
		0.951701,
		0.968302,
		0.980692,
		0.989361,
		0.994912,
		0.998039,
		0.999477,
		0.999936,
		1,
		0.999997,
		0.999878,
		0.999236,
		0.997444,
		0.99378,
		0.987511,
		0.977965,
		0.964563,
		0.946848,
		0.924495,
		0.897313,
		0.865241,
		0.828347,
		0.786818,
		0.740951,
		0.69115,
		0.637915,
		0.58184,
		0.523601,
		0.463957,
		0.403741,
		0.343856,
		0.285276,
		0.229035,
		0.17623,
		0.128015,
		0.085601,
		0.050253,
		0.023285,
		0.006063,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.00339,
		0.017731,
		0.042023,
		0.074922,
		0.11513,
		0.161399,
		0.212531,
		0.267385,
		0.32487,
		0.383959,
		0.443682,
		0.503133,
		0.561477,
		0.617947,
		0.671853,
		0.722587,
		0.769624,
		0.812535,
		0.850988,
		0.884756,
		0.913726,
		0.937904,
		0.957423,
		0.972543,
		0.983655,
		0.991272,
		0.996013,
		0.998572,
		0.999668,
		0.999972,
		1,
		0.999989,
		0.999795,
		0.998956,
		0.996829,
		0.9927,
		0.985864,
		0.975678,
		0.961602,
		0.943213,
		0.920217,
		0.89245,
		0.859876,
		0.822577,
		0.780754,
		0.734711,
		0.684856,
		0.631689,
		0.575799,
		0.517854,
		0.458604,
		0.398867,
		0.339531,
		0.28155,
		0.225939,
		0.173771,
		0.126177,
		0.084339,
		0.049493,
		0.022924,
		0.005967,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.000378,
		0.009086,
		0.02828,
		0.056622,
		0.092821,
		0.135635,
		0.183866,
		0.236371,
		0.292057,
		0.349886,
		0.40888,
		0.468117,
		0.526742,
		0.583966,
		0.639069,
		0.691409,
		0.740422,
		0.78563,
		0.826649,
		0.863189,
		0.895068,
		0.922214,
		0.944673,
		0.962616,
		0.976338,
		0.986257,
		0.992908,
		0.996922,
		0.998987,
		0.999801,
		0.99999,
		1,
		0.999973,
		0.999679,
		0.998614,
		0.996128,
		0.991519,
		0.984106,
		0.973282,
		0.95854,
		0.939491,
		0.91587,
		0.887541,
		0.854485,
		0.816804,
		0.774707,
		0.728508,
		0.678615,
		0.625529,
		0.569832,
		0.512188,
		0.453333,
		0.394074,
		0.335282,
		0.277895,
		0.222905,
		0.171364,
		0.124377,
		0.083104,
		0.04875,
		0.022572,
		0.005873,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.003284,
		0.01719,
		0.040769,
		0.072738,
		0.111859,
		0.156939,
		0.206833,
		0.260446,
		0.316735,
		0.374712,
		0.433445,
		0.492061,
		0.54975,
		0.60577,
		0.659446,
		0.710181,
		0.757456,
		0.800837,
		0.839981,
		0.874642,
		0.90468,
		0.930061,
		0.950871,
		0.967315,
		0.97972,
		0.988531,
		0.9943,
		0.997663,
		0.999303,
		0.999889,
		0.999997,
		1,
		0.999942,
		0.999523,
		0.998208,
		0.99534,
		0.990235,
		0.98224,
		0.97078,
		0.955381,
		0.935686,
		0.91146,
		0.882588,
		0.849074,
		0.811031,
		0.768681,
		0.722343,
		0.672427,
		0.619433,
		0.563939,
		0.5066,
		0.448143,
		0.38936,
		0.331109,
		0.274307,
		0.219929,
		0.169005,
		0.122616,
		0.081895,
		0.048024,
		0.022228,
		0.005782,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.000366,
		0.008808,
		0.027432,
		0.054963,
		0.090168,
		0.131859,
		0.178891,
		0.230171,
		0.28465,
		0.341334,
		0.39928,
		0.457602,
		0.515474,
		0.572131,
		0.626874,
		0.679074,
		0.728175,
		0.773702,
		0.815262,
		0.852555,
		0.885375,
		0.913621,
		0.937299,
		0.95653,
		0.971552,
		0.982721,
		0.990506,
		0.995473,
		0.99826,
		0.999537,
		0.999944,
		1,
		0.999997,
		0.999892,
		0.999323,
		0.997731,
		0.994461,
		0.988848,
		0.980266,
		0.968173,
		0.952128,
		0.931803,
		0.90699,
		0.877597,
		0.843645,
		0.805262,
		0.762678,
		0.716218,
		0.666294,
		0.613403,
		0.55812,
		0.501091,
		0.443032,
		0.384724,
		0.327009,
		0.270786,
		0.217011,
		0.166693,
		0.120891,
		0.080713,
		0.047314,
		0.021892,
		0.005692,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.003183,
		0.016673,
		0.039569,
		0.070648,
		0.108724,
		0.152659,
		0.201355,
		0.253765,
		0.308888,
		0.365774,
		0.423526,
		0.481304,
		0.538323,
		0.593863,
		0.647268,
		0.69795,
		0.745396,
		0.789171,
		0.828924,
		0.864394,
		0.895415,
		0.921922,
		0.943959,
		0.961681,
		0.975358,
		0.985372,
		0.99221,
		0.996452,
		0.998733,
		0.999707,
		0.999975,
		1,
		0.999991,
		0.999819,
		0.999074,
		0.99718,
		0.99349,
		0.987358,
		0.978187,
		0.965467,
		0.948785,
		0.927846,
		0.902465,
		0.872571,
		0.838201,
		0.799498,
		0.756699,
		0.710133,
		0.660215,
		0.607439,
		0.552374,
		0.495659,
		0.438,
		0.380165,
		0.32298,
		0.267329,
		0.214149,
		0.164428,
		0.119202,
		0.079556,
		0.046619,
		0.021564,
		0.005605,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.004312,
		0.019046,
		0.043031,
		0.075048,
		0.113919,
		0.158507,
		0.207724,
		0.260527,
		0.315921,
		0.372963,
		0.430764,
		0.488488,
		0.545361,
		0.600669,
		0.653763,
		0.704067,
		0.751077,
		0.794366,
		0.833597,
		0.868517,
		0.898975,
		0.924919,
		0.946406,
		0.963605,
		0.976801,
		0.986389,
		0.99287,
		0.99683,
		0.998913,
		0.999767,
		0.999984,
		1,
		0.999984,
		0.999767,
		0.998913,
		0.99683,
		0.99287,
		0.986389,
		0.976801,
		0.963605,
		0.946406,
		0.924919,
		0.898975,
		0.868517,
		0.833597,
		0.794366,
		0.751077,
		0.704067,
		0.653763,
		0.600669,
		0.545361,
		0.488488,
		0.430764,
		0.372963,
		0.315921,
		0.260527,
		0.207724,
		0.158507,
		0.113919,
		0.075048,
		0.043031,
		0.019046,
		0.004312,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		0.002256,
		0.014668,
		0.03678,
		0.06733,
		0.105104,
		0.14893,
		0.197681,
		0.25028,
		0.3057,
		0.362964,
		0.421152,
		0.479399,
		0.536902,
		0.592919,
		0.646778,
		0.697877,
		0.745689,
		0.78977,
		0.82976,
		0.865395,
		0.896506,
		0.923032,
		0.945021,
		0.96264,
		0.976169,
		0.986009,
		0.992666,
		0.996737,
		0.99888,
		0.99976,
		0.999984,
		1,
		0.999984,
		0.99976,
		0.99888,
		0.996737,
		0.992666,
		0.986009,
		0.976169,
		0.96264,
		0.945021,
		0.923032,
		0.896506,
		0.865395,
		0.82976,
		0.78977,
		0.745689,
		0.697877,
		0.646778,
		0.592919,
		0.536902,
		0.479399,
		0.421152,
		0.362964,
		0.3057,
		0.25028,
		0.197681,
		0.14893,
		0.105104,
		0.06733,
		0.03678,
		0.014668,
		0.002256,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.000833,
		0.010771,
		0.030875,
		0.059846,
		0.096427,
		0.139409,
		0.187629,
		0.239972,
		0.295377,
		0.352834,
		0.411389,
		0.470147,
		0.528274,
		0.585003,
		0.639632,
		0.691535,
		0.740163,
		0.785049,
		0.825816,
		0.862181,
		0.893962,
		0.921086,
		0.943593,
		0.961642,
		0.975516,
		0.985615,
		0.992454,
		0.996641,
		0.998846,
		0.999753,
		0.999983,
		1,
		0.999983,
		0.999753,
		0.998846,
		0.996641,
		0.992454,
		0.985615,
		0.975516,
		0.961642,
		0.943593,
		0.921086,
		0.893962,
		0.862181,
		0.825816,
		0.785049,
		0.740163,
		0.691535,
		0.639632,
		0.585003,
		0.528274,
		0.470147,
		0.411389,
		0.352834,
		0.295377,
		0.239972,
		0.187629,
		0.139409,
		0.096427,
		0.059846,
		0.030875,
		0.010771,
		0.000833,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		9.5e-005,
		0.007397,
		0.025353,
		0.052623,
		0.08791,
		0.129963,
		0.17758,
		0.229612,
		0.284958,
		0.342574,
		0.401474,
		0.460729,
		0.519475,
		0.576915,
		0.63232,
		0.685037,
		0.734494,
		0.780201,
		0.821761,
		0.858873,
		0.891341,
		0.919079,
		0.942118,
		0.960612,
		0.97484,
		0.985208,
		0.992235,
		0.996541,
		0.998811,
		0.999745,
		0.999983,
		1,
		0.999983,
		0.999745,
		0.998811,
		0.996541,
		0.992235,
		0.985208,
		0.97484,
		0.960612,
		0.942118,
		0.919079,
		0.891341,
		0.858873,
		0.821761,
		0.780201,
		0.734494,
		0.685037,
		0.63232,
		0.576915,
		0.519475,
		0.460729,
		0.401474,
		0.342574,
		0.284958,
		0.229612,
		0.17758,
		0.129963,
		0.08791,
		0.052623,
		0.025353,
		0.007397,
		9.5e-005,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.004591,
		0.020251,
		0.045692,
		0.079575,
		0.120607,
		0.167548,
		0.219207,
		0.274448,
		0.332189,
		0.391409,
		0.451146,
		0.510503,
		0.568653,
		0.624839,
		0.67838,
		0.728678,
		0.775222,
		0.817592,
		0.855469,
		0.888641,
		0.917009,
		0.940595,
		0.959546,
		0.974141,
		0.984786,
		0.992008,
		0.996437,
		0.998774,
		0.999737,
		0.999982,
		1,
		0.999982,
		0.999737,
		0.998774,
		0.996437,
		0.992008,
		0.984786,
		0.974141,
		0.959546,
		0.940595,
		0.917009,
		0.888641,
		0.855469,
		0.817592,
		0.775222,
		0.728678,
		0.67838,
		0.624839,
		0.568653,
		0.510503,
		0.451146,
		0.391409,
		0.332189,
		0.274448,
		0.219207,
		0.167548,
		0.120607,
		0.079575,
		0.045692,
		0.020251,
		0.004591,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		0.002404,
		0.015609,
		0.039085,
		0.071448,
		0.111363,
		0.157547,
		0.208769,
		0.263854,
		0.321682,
		0.381194,
		0.441396,
		0.501356,
		0.560215,
		0.617185,
		0.671559,
		0.722712,
		0.770107,
		0.813304,
		0.851964,
		0.885858,
		0.914874,
		0.939022,
		0.958445,
		0.973418,
		0.984349,
		0.991772,
		0.996329,
		0.998737,
		0.999729,
		0.999982,
		1,
		0.999982,
		0.999729,
		0.998737,
		0.996329,
		0.991772,
		0.984349,
		0.973418,
		0.958445,
		0.939022,
		0.914874,
		0.885858,
		0.851964,
		0.813304,
		0.770107,
		0.722712,
		0.671559,
		0.617185,
		0.560215,
		0.501356,
		0.441396,
		0.381194,
		0.321682,
		0.263854,
		0.208769,
		0.157547,
		0.111363,
		0.071448,
		0.039085,
		0.015609,
		0.002404,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.000888,
		0.011471,
		0.032838,
		0.063557,
		0.102252,
		0.147593,
		0.198309,
		0.253183,
		0.311056,
		0.370832,
		0.431479,
		0.492031,
		0.551596,
		0.609355,
		0.66457,
		0.71659,
		0.764852,
		0.808893,
		0.848355,
		0.882989,
		0.91267,
		0.937398,
		0.957307,
		0.97267,
		0.983896,
		0.991528,
		0.996218,
		0.998697,
		0.99972,
		0.999981,
		1,
		0.999981,
		0.99972,
		0.998697,
		0.996218,
		0.991528,
		0.983896,
		0.97267,
		0.957307,
		0.937398,
		0.91267,
		0.882989,
		0.848355,
		0.808893,
		0.764852,
		0.71659,
		0.66457,
		0.609355,
		0.551596,
		0.492031,
		0.431479,
		0.370832,
		0.311056,
		0.253183,
		0.198309,
		0.147593,
		0.102252,
		0.063557,
		0.032838,
		0.011471,
		0.000888,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.000101,
		0.007884,
		0.026987,
		0.055933,
		0.093296,
		0.137703,
		0.187838,
		0.242442,
		0.300316,
		0.360324,
		0.421394,
		0.482527,
		0.542794,
		0.601344,
		0.65741,
		0.710309,
		0.759454,
		0.804357,
		0.844638,
		0.880032,
		0.910396,
		0.93572,
		0.95613,
		0.971895,
		0.983428,
		0.991275,
		0.996102,
		0.998656,
		0.999711,
		0.99998,
		1,
		0.99998,
		0.999711,
		0.998656,
		0.996102,
		0.991275,
		0.983428,
		0.971895,
		0.95613,
		0.93572,
		0.910396,
		0.880032,
		0.844638,
		0.804357,
		0.759454,
		0.710309,
		0.65741,
		0.601344,
		0.542794,
		0.482527,
		0.421394,
		0.360324,
		0.300316,
		0.242442,
		0.187838,
		0.137703,
		0.093296,
		0.055933,
		0.026987,
		0.007884,
		0.000101,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.004898,
		0.021574,
		0.048607,
		0.08452,
		0.127896,
		0.177371,
		0.231643,
		0.289469,
		0.349672,
		0.411143,
		0.472842,
		0.533806,
		0.59315,
		0.650073,
		0.703863,
		0.753907,
		0.79969,
		0.84081,
		0.876983,
		0.908049,
		0.933986,
		0.954912,
		0.971093,
		0.982941,
		0.991012,
		0.995981,
		0.998614,
		0.999702,
		0.99998,
		1,
		0.99998,
		0.999702,
		0.998614,
		0.995981,
		0.991012,
		0.982941,
		0.971093,
		0.954912,
		0.933986,
		0.908049,
		0.876983,
		0.84081,
		0.79969,
		0.753907,
		0.703863,
		0.650073,
		0.59315,
		0.533806,
		0.472842,
		0.411143,
		0.349672,
		0.289469,
		0.231643,
		0.177371,
		0.127896,
		0.08452,
		0.048607,
		0.021574,
		0.004898,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.002567,
		0.016644,
		0.041614,
		0.075953,
		0.118192,
		0.166922,
		0.220793,
		0.278519,
		0.33888,
		0.400724,
		0.462975,
		0.524629,
		0.584767,
		0.642555,
		0.69725,
		0.748208,
		0.794889,
		0.836868,
		0.873839,
		0.905627,
		0.932195,
		0.953653,
		0.970262,
		0.982438,
		0.990739,
		0.995856,
		0.998569,
		0.999692,
		0.999979,
		1,
		0.999979,
		0.999692,
		0.998569,
		0.995856,
		0.990739,
		0.982438,
		0.970262,
		0.953653,
		0.932195,
		0.905627,
		0.873839,
		0.836868,
		0.794889,
		0.748208,
		0.69725,
		0.642555,
		0.584767,
		0.524629,
		0.462975,
		0.400724,
		0.33888,
		0.278519,
		0.220793,
		0.166922,
		0.118192,
		0.075953,
		0.041614,
		0.016644,
		0.002567,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.000949,
		0.012242,
		0.034993,
		0.067623,
		0.108615,
		0.156508,
		0.209905,
		0.267474,
		0.327951,
		0.39014,
		0.452924,
		0.51526,
		0.576193,
		0.634852,
		0.690463,
		0.742351,
		0.789949,
		0.832806,
		0.870596,
		0.903125,
		0.930343,
		0.95235,
		0.969402,
		0.981915,
		0.990456,
		0.995726,
		0.998523,
		0.999682,
		0.999978,
		1,
		0.999978,
		0.999682,
		0.998523,
		0.995726,
		0.990456,
		0.981915,
		0.969402,
		0.95235,
		0.930343,
		0.903125,
		0.870596,
		0.832806,
		0.789949,
		0.742351,
		0.690463,
		0.634852,
		0.576193,
		0.51526,
		0.452924,
		0.39014,
		0.327951,
		0.267474,
		0.209905,
		0.156508,
		0.108615,
		0.067623,
		0.034993,
		0.012242,
		0.000949,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.000108,
		0.008421,
		0.028784,
		0.059563,
		0.099188,
		0.146146,
		0.198992,
		0.256343,
		0.316889,
		0.379391,
		0.442688,
		0.505698,
		0.567424,
		0.62696,
		0.683499,
		0.736332,
		0.784865,
		0.828621,
		0.867251,
		0.900541,
		0.928428,
		0.951001,
		0.96851,
		0.981373,
		0.990162,
		0.995591,
		0.998475,
		0.999671,
		0.999978,
		1,
		0.999978,
		0.999671,
		0.998475,
		0.995591,
		0.990162,
		0.981373,
		0.96851,
		0.951001,
		0.928428,
		0.900541,
		0.867251,
		0.828621,
		0.784865,
		0.736332,
		0.683499,
		0.62696,
		0.567424,
		0.505698,
		0.442688,
		0.379391,
		0.316889,
		0.256343,
		0.198992,
		0.146146,
		0.099188,
		0.059563,
		0.028784,
		0.008421,
		0.000108,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.005236,
		0.023032,
		0.051808,
		0.089938,
		0.135857,
		0.188066,
		0.245133,
		0.3057,
		0.36848,
		0.432268,
		0.495938,
		0.558455,
		0.618874,
		0.676351,
		0.730146,
		0.779633,
		0.824309,
		0.863799,
		0.897873,
		0.926448,
		0.949604,
		0.967586,
		0.980811,
		0.989857,
		0.99545,
		0.998425,
		0.99966,
		0.999977,
		1,
		0.999977,
		0.99966,
		0.998425,
		0.99545,
		0.989857,
		0.980811,
		0.967586,
		0.949604,
		0.926448,
		0.897873,
		0.863799,
		0.824309,
		0.779633,
		0.730146,
		0.676351,
		0.618874,
		0.558455,
		0.495938,
		0.432268,
		0.36848,
		0.3057,
		0.245133,
		0.188066,
		0.135857,
		0.089938,
		0.051808,
		0.023032,
		0.005236,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.002747,
		0.017784,
		0.044395,
		0.080895,
		0.125662,
		0.177144,
		0.233856,
		0.294389,
		0.357409,
		0.421662,
		0.48598,
		0.549284,
		0.61059,
		0.669016,
		0.723788,
		0.774249,
		0.819865,
		0.860238,
		0.895116,
		0.9244,
		0.948158,
		0.966629,
		0.980227,
		0.98954,
		0.995304,
		0.998373,
		0.999649,
		0.999976,
		1,
		0.999976,
		0.999649,
		0.998373,
		0.995304,
		0.98954,
		0.980227,
		0.966629,
		0.948158,
		0.9244,
		0.895116,
		0.860238,
		0.819865,
		0.774249,
		0.723788,
		0.669016,
		0.61059,
		0.549284,
		0.48598,
		0.421662,
		0.357409,
		0.294389,
		0.233856,
		0.177144,
		0.125662,
		0.080895,
		0.044395,
		0.017784,
		0.002747,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.001017,
		0.013093,
		0.037366,
		0.07209,
		0.115585,
		0.166242,
		0.222522,
		0.282963,
		0.34618,
		0.410871,
		0.475821,
		0.539907,
		0.602103,
		0.661489,
		0.717253,
		0.768706,
		0.815284,
		0.856563,
		0.892267,
		0.922282,
		0.946661,
		0.965636,
		0.979622,
		0.98921,
		0.995152,
		0.998319,
		0.999637,
		0.999975,
		1,
		0.999975,
		0.999637,
		0.998319,
		0.995152,
		0.98921,
		0.979622,
		0.965636,
		0.946661,
		0.922282,
		0.892267,
		0.856563,
		0.815284,
		0.768706,
		0.717253,
		0.661489,
		0.602103,
		0.539907,
		0.475821,
		0.410871,
		0.34618,
		0.282963,
		0.222522,
		0.166242,
		0.115585,
		0.07209,
		0.037366,
		0.013093,
		0.001017,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.000116,
		0.009016,
		0.030765,
		0.063557,
		0.105651,
		0.155379,
		0.211144,
		0.271429,
		0.334797,
		0.399895,
		0.465459,
		0.53032,
		0.593409,
		0.653763,
		0.710536,
		0.763,
		0.810562,
		0.852769,
		0.889324,
		0.92009,
		0.945109,
		0.964606,
		0.978993,
		0.988867,
		0.994994,
		0.998263,
		0.999624,
		0.999974,
		1,
		0.999974,
		0.999624,
		0.998263,
		0.994994,
		0.988867,
		0.978993,
		0.964606,
		0.945109,
		0.92009,
		0.889324,
		0.852769,
		0.810562,
		0.763,
		0.710536,
		0.653763,
		0.593409,
		0.53032,
		0.465459,
		0.399895,
		0.334797,
		0.271429,
		0.211144,
		0.155379,
		0.105651,
		0.063557,
		0.030765,
		0.009016,
		0.000116,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.005611,
		0.024641,
		0.055335,
		0.09589,
		0.144575,
		0.199736,
		0.259797,
		0.323265,
		0.388735,
		0.454892,
		0.52052,
		0.584502,
		0.645835,
		0.70363,
		0.757125,
		0.805693,
		0.848853,
		0.886281,
		0.917822,
		0.943502,
		0.963537,
		0.978339,
		0.988511,
		0.994829,
		0.998204,
		0.999611,
		0.999973,
		1,
		0.999973,
		0.999611,
		0.998204,
		0.994829,
		0.988511,
		0.978339,
		0.963537,
		0.943502,
		0.917822,
		0.886281,
		0.848853,
		0.805693,
		0.757125,
		0.70363,
		0.645835,
		0.584502,
		0.52052,
		0.454892,
		0.388735,
		0.323265,
		0.259797,
		0.199736,
		0.144575,
		0.09589,
		0.055335,
		0.024641,
		0.005611,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.002946,
		0.019046,
		0.047464,
		0.086332,
		0.133854,
		0.188314,
		0.248076,
		0.31159,
		0.377394,
		0.444121,
		0.510503,
		0.575379,
		0.637698,
		0.696531,
		0.751077,
		0.800673,
		0.844809,
		0.883135,
		0.915473,
		0.941836,
		0.962429,
		0.977661,
		0.98814,
		0.994657,
		0.998143,
		0.999598,
		0.999972,
		1,
		0.999972,
		0.999598,
		0.998143,
		0.994657,
		0.98814,
		0.977661,
		0.962429,
		0.941836,
		0.915473,
		0.883135,
		0.844809,
		0.800673,
		0.751077,
		0.696531,
		0.637698,
		0.575379,
		0.510503,
		0.444121,
		0.377394,
		0.31159,
		0.248076,
		0.188314,
		0.133854,
		0.086332,
		0.047464,
		0.019046,
		0.002946,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.001092,
		0.014037,
		0.039989,
		0.077011,
		0.123241,
		0.176896,
		0.236278,
		0.299776,
		0.365872,
		0.433143,
		0.500268,
		0.566035,
		0.629348,
		0.689233,
		0.744848,
		0.795497,
		0.840634,
		0.879882,
		0.913042,
		0.940109,
		0.961278,
		0.976956,
		0.987755,
		0.994478,
		0.998079,
		0.999584,
		0.999971,
		1,
		0.999971,
		0.999584,
		0.998079,
		0.994478,
		0.987755,
		0.976956,
		0.961278,
		0.940109,
		0.913042,
		0.879882,
		0.840634,
		0.795497,
		0.744848,
		0.689233,
		0.629348,
		0.566035,
		0.500268,
		0.433143,
		0.365872,
		0.299776,
		0.236278,
		0.176896,
		0.123241,
		0.077011,
		0.039989,
		0.014037,
		0.001092,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.000125,
		0.009675,
		0.032959,
		0.067965,
		0.112762,
		0.1655,
		0.224414,
		0.287832,
		0.354174,
		0.421959,
		0.489811,
		0.556466,
		0.620779,
		0.681729,
		0.738434,
		0.790157,
		0.836321,
		0.876517,
		0.910525,
		0.938318,
		0.960083,
		0.976223,
		0.987353,
		0.994292,
		0.998013,
		0.999569,
		0.99997,
		1,
		0.99997,
		0.999569,
		0.998013,
		0.994292,
		0.987353,
		0.976223,
		0.960083,
		0.938318,
		0.910525,
		0.876517,
		0.836321,
		0.790157,
		0.738434,
		0.681729,
		0.620779,
		0.556466,
		0.489811,
		0.421959,
		0.354174,
		0.287832,
		0.224414,
		0.1655,
		0.112762,
		0.067965,
		0.032959,
		0.009675,
		0.000125,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.006028,
		0.026425,
		0.059234,
		0.102448,
		0.154148,
		0.2125,
		0.275766,
		0.342303,
		0.410568,
		0.479129,
		0.546667,
		0.611985,
		0.674014,
		0.731828,
		0.78465,
		0.831866,
		0.873037,
		0.907917,
		0.936461,
		0.958843,
		0.97546,
		0.986936,
		0.994098,
		0.997943,
		0.999553,
		0.999969,
		1,
		0.999969,
		0.999553,
		0.997943,
		0.994098,
		0.986936,
		0.97546,
		0.958843,
		0.936461,
		0.907917,
		0.873037,
		0.831866,
		0.78465,
		0.731828,
		0.674014,
		0.611985,
		0.546667,
		0.479129,
		0.410568,
		0.342303,
		0.275766,
		0.2125,
		0.154148,
		0.102448,
		0.059234,
		0.026425,
		0.006028,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.003169,
		0.020447,
		0.050862,
		0.092332,
		0.142864,
		0.200552,
		0.263588,
		0.330263,
		0.398972,
		0.468221,
		0.536635,
		0.602961,
		0.666082,
		0.725025,
		0.778969,
		0.827263,
		0.869437,
		0.905215,
		0.934534,
		0.957554,
		0.974668,
		0.986501,
		0.993896,
		0.99787,
		0.999537,
		0.999968,
		1,
		0.999968,
		0.999537,
		0.99787,
		0.993896,
		0.986501,
		0.974668,
		0.957554,
		0.934534,
		0.905215,
		0.869437,
		0.827263,
		0.778969,
		0.725025,
		0.666082,
		0.602961,
		0.536635,
		0.468221,
		0.398972,
		0.330263,
		0.263588,
		0.200552,
		0.142864,
		0.092332,
		0.050862,
		0.020447,
		0.003169,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.001176,
		0.015085,
		0.042898,
		0.082451,
		0.131674,
		0.188586,
		0.251308,
		0.31806,
		0.387171,
		0.457085,
		0.526364,
		0.593702,
		0.657927,
		0.718017,
		0.773107,
		0.822507,
		0.865711,
		0.902416,
		0.932535,
		0.956215,
		0.973843,
		0.986047,
		0.993685,
		0.997795,
		0.99952,
		0.999967,
		1,
		0.999967,
		0.99952,
		0.997795,
		0.993685,
		0.986047,
		0.973843,
		0.956215,
		0.932535,
		0.902416,
		0.865711,
		0.822507,
		0.773107,
		0.718017,
		0.657927,
		0.593702,
		0.526364,
		0.457085,
		0.387171,
		0.31806,
		0.251308,
		0.188586,
		0.131674,
		0.082451,
		0.042898,
		0.015085,
		0.001176,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.000135,
		0.010409,
		0.035395,
		0.072845,
		0.120607,
		0.176624,
		0.238939,
		0.3057,
		0.375167,
		0.445718,
		0.515852,
		0.584202,
		0.649542,
		0.710798,
		0.767059,
		0.817592,
		0.861854,
		0.899514,
		0.93046,
		0.954824,
		0.972985,
		0.985575,
		0.993464,
		0.997715,
		0.999502,
		0.999966,
		1,
		0.999966,
		0.999502,
		0.997715,
		0.993464,
		0.985575,
		0.972985,
		0.954824,
		0.93046,
		0.899514,
		0.861854,
		0.817592,
		0.767059,
		0.710798,
		0.649542,
		0.584202,
		0.515852,
		0.445718,
		0.375167,
		0.3057,
		0.238939,
		0.176624,
		0.120607,
		0.072845,
		0.035395,
		0.010409,
		0.000135,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.006493,
		0.02841,
		0.063557,
		0.109696,
		0.164687,
		0.226495,
		0.293191,
		0.362964,
		0.43412,
		0.505093,
		0.574455,
		0.64092,
		0.703361,
		0.760817,
		0.812511,
		0.857862,
		0.896506,
		0.928306,
		0.953377,
		0.972091,
		0.985083,
		0.993234,
		0.997633,
		0.999484,
		0.999964,
		1,
		0.999964,
		0.999484,
		0.997633,
		0.993234,
		0.985083,
		0.972091,
		0.953377,
		0.928306,
		0.896506,
		0.857862,
		0.812511,
		0.760817,
		0.703361,
		0.64092,
		0.574455,
		0.505093,
		0.43412,
		0.362964,
		0.293191,
		0.226495,
		0.164687,
		0.109696,
		0.063557,
		0.02841,
		0.006493,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.003417,
		0.022008,
		0.054636,
		0.098976,
		0.1528,
		0.213992,
		0.280543,
		0.350565,
		0.422291,
		0.494086,
		0.564456,
		0.632056,
		0.695699,
		0.754375,
		0.807258,
		0.853729,
		0.893386,
		0.926069,
		0.951873,
		0.971161,
		0.984569,
		0.992994,
		0.997546,
		0.999464,
		0.999963,
		1,
		0.999963,
		0.999464,
		0.997546,
		0.992994,
		0.984569,
		0.971161,
		0.951873,
		0.926069,
		0.893386,
		0.853729,
		0.807258,
		0.754375,
		0.695699,
		0.632056,
		0.564456,
		0.494086,
		0.422291,
		0.350565,
		0.280543,
		0.213992,
		0.1528,
		0.098976,
		0.054636,
		0.022008,
		0.003417,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.001269,
		0.016256,
		0.046135,
		0.088485,
		0.140991,
		0.201449,
		0.267766,
		0.337974,
		0.41023,
		0.482827,
		0.554201,
		0.622942,
		0.687806,
		0.747726,
		0.801828,
		0.849448,
		0.890151,
		0.923746,
		0.950308,
		0.970192,
		0.984034,
		0.992743,
		0.997456,
		0.999444,
		0.999962,
		1,
		0.999962,
		0.999444,
		0.997456,
		0.992743,
		0.984034,
		0.970192,
		0.950308,
		0.923746,
		0.890151,
		0.849448,
		0.801828,
		0.747726,
		0.687806,
		0.622942,
		0.554201,
		0.482827,
		0.41023,
		0.337974,
		0.267766,
		0.201449,
		0.140991,
		0.088485,
		0.046135,
		0.016256,
		0.001269,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.000146,
		0.011231,
		0.03811,
		0.078267,
		0.129291,
		0.188885,
		0.254872,
		0.325198,
		0.397939,
		0.471312,
		0.543682,
		0.613572,
		0.679673,
		0.740862,
		0.796212,
		0.845015,
		0.886795,
		0.921332,
		0.94868,
		0.969182,
		0.983475,
		0.992481,
		0.997361,
		0.999423,
		0.99996,
		1,
		0.99996,
		0.999423,
		0.997361,
		0.992481,
		0.983475,
		0.969182,
		0.94868,
		0.921332,
		0.886795,
		0.845015,
		0.796212,
		0.740862,
		0.679673,
		0.613572,
		0.543682,
		0.471312,
		0.397939,
		0.325198,
		0.254872,
		0.188885,
		0.129291,
		0.078267,
		0.03811,
		0.011231,
		0.000146,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.007014,
		0.030627,
		0.06837,
		0.117733,
		0.176325,
		0.241875,
		0.312242,
		0.385418,
		0.45954,
		0.532897,
		0.603939,
		0.671294,
		0.733775,
		0.790404,
		0.840421,
		0.883312,
		0.918824,
		0.946986,
		0.968129,
		0.982892,
		0.992207,
		0.997262,
		0.9994,
		0.999959,
		1,
		0.999959,
		0.9994,
		0.997262,
		0.992207,
		0.982892,
		0.968129,
		0.946986,
		0.918824,
		0.883312,
		0.840421,
		0.790404,
		0.733775,
		0.671294,
		0.603939,
		0.532897,
		0.45954,
		0.385418,
		0.312242,
		0.241875,
		0.176325,
		0.117733,
		0.06837,
		0.030627,
		0.007014,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.003696,
		0.023754,
		0.058845,
		0.106356,
		0.163793,
		0.228792,
		0.299116,
		0.372671,
		0.447509,
		0.521839,
		0.594037,
		0.66266,
		0.726459,
		0.784396,
		0.835662,
		0.879698,
		0.916216,
		0.945222,
		0.967032,
		0.982283,
		0.99192,
		0.997158,
		0.999377,
		0.999957,
		1,
		0.999957,
		0.999377,
		0.997158,
		0.99192,
		0.982283,
		0.967032,
		0.945222,
		0.916216,
		0.879698,
		0.835662,
		0.784396,
		0.726459,
		0.66266,
		0.594037,
		0.521839,
		0.447509,
		0.372671,
		0.299116,
		0.228792,
		0.163793,
		0.106356,
		0.058845,
		0.023754,
		0.003696,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.001374,
		0.017568,
		0.049751,
		0.0952,
		0.151319,
		0.21564,
		0.28583,
		0.359701,
		0.435217,
		0.510503,
		0.583858,
		0.653763,
		0.718904,
		0.77818,
		0.830729,
		0.875946,
		0.913505,
		0.943385,
		0.965887,
		0.981647,
		0.991621,
		0.997049,
		0.999352,
		0.999955,
		1,
		0.999955,
		0.999352,
		0.997049,
		0.991621,
		0.981647,
		0.965887,
		0.943385,
		0.913505,
		0.875946,
		0.830729,
		0.77818,
		0.718904,
		0.653763,
		0.583858,
		0.510503,
		0.435217,
		0.359701,
		0.28583,
		0.21564,
		0.151319,
		0.0952,
		0.049751,
		0.017568,
		0.001374,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.000158,
		0.012153,
		0.04115,
		0.084314,
		0.138935,
		0.20244,
		0.272394,
		0.346512,
		0.422664,
		0.498887,
		0.573396,
		0.644597,
		0.711103,
		0.771749,
		0.825616,
		0.872049,
		0.910685,
		0.941471,
		0.964693,
		0.980982,
		0.991307,
		0.996935,
		0.999327,
		0.999953,
		1,
		0.999953,
		0.999327,
		0.996935,
		0.991307,
		0.980982,
		0.964693,
		0.941471,
		0.910685,
		0.872049,
		0.825616,
		0.771749,
		0.711103,
		0.644597,
		0.573396,
		0.498887,
		0.422664,
		0.346512,
		0.272394,
		0.20244,
		0.138935,
		0.084314,
		0.04115,
		0.012153,
		0.000158,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.0076,
		0.033113,
		0.073747,
		0.126677,
		0.189216,
		0.258824,
		0.333111,
		0.409849,
		0.486985,
		0.562644,
		0.635152,
		0.703046,
		0.765094,
		0.820315,
		0.868003,
		0.907751,
		0.939477,
		0.963446,
		0.980287,
		0.990978,
		0.996815,
		0.999299,
		0.999951,
		1,
		0.999951,
		0.999299,
		0.996815,
		0.990978,
		0.980287,
		0.963446,
		0.939477,
		0.907751,
		0.868003,
		0.820315,
		0.765094,
		0.703046,
		0.635152,
		0.562644,
		0.486985,
		0.409849,
		0.333111,
		0.258824,
		0.189216,
		0.126677,
		0.073747,
		0.033113,
		0.0076,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.00401,
		0.025717,
		0.063557,
		0.114585,
		0.175995,
		0.245133,
		0.319503,
		0.396775,
		0.474793,
		0.551596,
		0.62542,
		0.694725,
		0.758206,
		0.814818,
		0.863799,
		0.904698,
		0.937398,
		0.962145,
		0.97956,
		0.990634,
		0.99669,
		0.999271,
		0.999949,
		1,
		0.999949,
		0.999271,
		0.99669,
		0.990634,
		0.97956,
		0.962145,
		0.937398,
		0.904698,
		0.863799,
		0.814818,
		0.758206,
		0.694725,
		0.62542,
		0.551596,
		0.474793,
		0.396775,
		0.319503,
		0.245133,
		0.175995,
		0.114585,
		0.063557,
		0.025717,
		0.00401,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.001493,
		0.019046,
		0.053808,
		0.102704,
		0.162806,
		0.231342,
		0.3057,
		0.383441,
		0.46231,
		0.540245,
		0.615394,
		0.686131,
		0.751077,
		0.809117,
		0.859431,
		0.90152,
		0.935231,
		0.960786,
		0.978799,
		0.990274,
		0.996558,
		0.999241,
		0.999947,
		1,
		0.999947,
		0.999241,
		0.996558,
		0.990274,
		0.978799,
		0.960786,
		0.935231,
		0.90152,
		0.859431,
		0.809117,
		0.751077,
		0.686131,
		0.615394,
		0.540245,
		0.46231,
		0.383441,
		0.3057,
		0.231342,
		0.162806,
		0.102704,
		0.053808,
		0.019046,
		0.001493,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.000172,
		0.013194,
		0.044568,
		0.091084,
		0.149684,
		0.21747,
		0.29171,
		0.369853,
		0.449532,
		0.528585,
		0.605064,
		0.677255,
		0.743696,
		0.803203,
		0.854892,
		0.898211,
		0.93297,
		0.959365,
		0.978003,
		0.989896,
		0.996419,
		0.99921,
		0.999945,
		1,
		0.999945,
		0.99921,
		0.996419,
		0.989896,
		0.978003,
		0.959365,
		0.93297,
		0.898211,
		0.854892,
		0.803203,
		0.743696,
		0.677255,
		0.605064,
		0.528585,
		0.449532,
		0.369853,
		0.29171,
		0.21747,
		0.149684,
		0.091084,
		0.044568,
		0.013194,
		0.000172,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.008262,
		0.035914,
		0.079781,
		0.136666,
		0.203541,
		0.277548,
		0.356013,
		0.436458,
		0.516611,
		0.594422,
		0.668086,
		0.736055,
		0.797066,
		0.850172,
		0.894765,
		0.930611,
		0.957881,
		0.97717,
		0.989499,
		0.996274,
		0.999176,
		0.999943,
		1,
		0.999943,
		0.999176,
		0.996274,
		0.989499,
		0.97717,
		0.957881,
		0.930611,
		0.894765,
		0.850172,
		0.797066,
		0.736055,
		0.668086,
		0.594422,
		0.516611,
		0.436458,
		0.356013,
		0.277548,
		0.203541,
		0.136666,
		0.079781,
		0.035914,
		0.008262,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.004366,
		0.027933,
		0.068856,
		0.123796,
		0.189584,
		0.263228,
		0.341929,
		0.423087,
		0.504315,
		0.583461,
		0.658615,
		0.728142,
		0.790699,
		0.845265,
		0.891175,
		0.928149,
		0.956329,
		0.976297,
		0.989083,
		0.996121,
		0.999141,
		0.99994,
		1,
		0.99994,
		0.999141,
		0.996121,
		0.989083,
		0.976297,
		0.956329,
		0.928149,
		0.891175,
		0.845265,
		0.790699,
		0.728142,
		0.658615,
		0.583461,
		0.504315,
		0.423087,
		0.341929,
		0.263228,
		0.189584,
		0.123796,
		0.068856,
		0.027933,
		0.004366,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.001628,
		0.020718,
		0.05838,
		0.111122,
		0.175628,
		0.248768,
		0.327607,
		0.409418,
		0.491695,
		0.57217,
		0.648832,
		0.719949,
		0.78409,
		0.840161,
		0.887434,
		0.925579,
		0.954706,
		0.975383,
		0.988646,
		0.995961,
		0.999105,
		0.999937,
		1,
		0.999937,
		0.999105,
		0.995961,
		0.988646,
		0.975383,
		0.954706,
		0.925579,
		0.887434,
		0.840161,
		0.78409,
		0.719949,
		0.648832,
		0.57217,
		0.491695,
		0.409418,
		0.327607,
		0.248768,
		0.175628,
		0.111122,
		0.05838,
		0.020718,
		0.001628,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.000188,
		0.014374,
		0.048428,
		0.098696,
		0.16171,
		0.234189,
		0.313057,
		0.395453,
		0.478744,
		0.560543,
		0.638727,
		0.711463,
		0.77723,
		0.834852,
		0.883534,
		0.922895,
		0.953007,
		0.974424,
		0.988187,
		0.995792,
		0.999066,
		0.999935,
		1,
		0.999935,
		0.999066,
		0.995792,
		0.988187,
		0.974424,
		0.953007,
		0.922895,
		0.883534,
		0.834852,
		0.77723,
		0.711463,
		0.638727,
		0.560543,
		0.478744,
		0.395453,
		0.313057,
		0.234189,
		0.16171,
		0.098696,
		0.048428,
		0.014374,
		0.000188,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.009016,
		0.039085,
		0.086581,
		0.147869,
		0.219514,
		0.29829,
		0.381194,
		0.465459,
		0.54857,
		0.628289,
		0.702674,
		0.770107,
		0.829327,
		0.879467,
		0.92009,
		0.951229,
		0.973418,
		0.987705,
		0.995614,
		0.999025,
		0.999932,
		1,
		0.999932,
		0.999025,
		0.995614,
		0.987705,
		0.973418,
		0.951229,
		0.92009,
		0.879467,
		0.829327,
		0.770107,
		0.702674,
		0.628289,
		0.54857,
		0.465459,
		0.381194,
		0.29829,
		0.219514,
		0.147869,
		0.086581,
		0.039085,
		0.009016,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.004772,
		0.030448,
		0.074843,
		0.134151,
		0.204772,
		0.283322,
		0.366646,
		0.451836,
		0.536244,
		0.617508,
		0.69357,
		0.76271,
		0.823577,
		0.875226,
		0.917159,
		0.949367,
		0.972363,
		0.987199,
		0.995426,
		0.998982,
		0.999929,
		1,
		0.999929,
		0.998982,
		0.995426,
		0.987199,
		0.972363,
		0.949367,
		0.917159,
		0.875226,
		0.823577,
		0.76271,
		0.69357,
		0.617508,
		0.536244,
		0.451836,
		0.366646,
		0.283322,
		0.204772,
		0.134151,
		0.074843,
		0.030448,
		0.004772,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.001783,
		0.02262,
		0.063557,
		0.120607,
		0.189994,
		0.268167,
		0.351814,
		0.437873,
		0.523558,
		0.606372,
		0.684139,
		0.755028,
		0.817592,
		0.870801,
		0.914095,
		0.947416,
		0.971256,
		0.986666,
		0.995229,
		0.998936,
		0.999925,
		1,
		0.999925,
		0.998936,
		0.995229,
		0.986666,
		0.971256,
		0.947416,
		0.914095,
		0.870801,
		0.817592,
		0.755028,
		0.684139,
		0.606372,
		0.523558,
		0.437873,
		0.351814,
		0.268167,
		0.189994,
		0.120607,
		0.063557,
		0.02262,
		0.001783,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.000206,
		0.015719,
		0.05281,
		0.107296,
		0.175218,
		0.252848,
		0.336705,
		0.423569,
		0.510503,
		0.594872,
		0.67437,
		0.747049,
		0.811359,
		0.866184,
		0.91089,
		0.945372,
		0.970093,
		0.986105,
		0.99502,
		0.998888,
		0.999922,
		1,
		0.999922,
		0.998888,
		0.99502,
		0.986105,
		0.970093,
		0.945372,
		0.91089,
		0.866184,
		0.811359,
		0.747049,
		0.67437,
		0.594872,
		0.510503,
		0.423569,
		0.336705,
		0.252848,
		0.175218,
		0.107296,
		0.05281,
		0.015719,
		0.000206,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.009877,
		0.042694,
		0.094282,
		0.160485,
		0.237387,
		0.321329,
		0.408924,
		0.497074,
		0.582997,
		0.664249,
		0.738759,
		0.804867,
		0.861363,
		0.907537,
		0.943228,
		0.968871,
		0.985515,
		0.994801,
		0.998838,
		0.999918,
		1,
		0.999918,
		0.998838,
		0.994801,
		0.985515,
		0.968871,
		0.943228,
		0.907537,
		0.861363,
		0.804867,
		0.738759,
		0.664249,
		0.582997,
		0.497074,
		0.408924,
		0.321329,
		0.237387,
		0.160485,
		0.094282,
		0.042694,
		0.009877,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.005236,
		0.033317,
		0.08164,
		0.145843,
		0.221812,
		0.3057,
		0.393939,
		0.483264,
		0.570735,
		0.653763,
		0.730146,
		0.798105,
		0.856329,
		0.904028,
		0.94098,
		0.967586,
		0.984893,
		0.994569,
		0.998784,
		0.999914,
		1,
		0.999914,
		0.998784,
		0.994569,
		0.984893,
		0.967586,
		0.94098,
		0.904028,
		0.856329,
		0.798105,
		0.730146,
		0.653763,
		0.570735,
		0.483264,
		0.393939,
		0.3057,
		0.221812,
		0.145843,
		0.08164,
		0.033317,
		0.005236,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.00196,
		0.024796,
		0.069452,
		0.131346,
		0.206156,
		0.289832,
		0.378617,
		0.469068,
		0.558077,
		0.642901,
		0.721196,
		0.791059,
		0.851071,
		0.900354,
		0.938621,
		0.966236,
		0.984237,
		0.994324,
		0.998727,
		0.99991,
		1,
		0.99991,
		0.998727,
		0.994324,
		0.984237,
		0.966236,
		0.938621,
		0.900354,
		0.851071,
		0.791059,
		0.721196,
		0.642901,
		0.558077,
		0.469068,
		0.378617,
		0.289832,
		0.206156,
		0.131346,
		0.069452,
		0.024796,
		0.00196,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.000227,
		0.017263,
		0.057812,
		0.117059,
		0.190456,
		0.273744,
		0.362964,
		0.454482,
		0.545012,
		0.631647,
		0.711894,
		0.783716,
		0.845578,
		0.896506,
		0.936145,
		0.964815,
		0.983546,
		0.994065,
		0.998667,
		0.999906,
		1,
		0.999906,
		0.998667,
		0.994065,
		0.983546,
		0.964815,
		0.936145,
		0.896506,
		0.845578,
		0.783716,
		0.711894,
		0.631647,
		0.545012,
		0.454482,
		0.362964,
		0.273744,
		0.190456,
		0.117059,
		0.057812,
		0.017263,
		0.000227,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.010867,
		0.046826,
		0.10305,
		0.174757,
		0.25746,
		0.346987,
		0.439502,
		0.531529,
		0.619989,
		0.702226,
		0.776061,
		0.839836,
		0.892474,
		0.933544,
		0.963319,
		0.982817,
		0.993791,
		0.998603,
		0.999901,
		1,
		0.999901,
		0.998603,
		0.993791,
		0.982817,
		0.963319,
		0.933544,
		0.892474,
		0.839836,
		0.776061,
		0.702226,
		0.619989,
		0.531529,
		0.439502,
		0.346987,
		0.25746,
		0.174757,
		0.10305,
		0.046826,
		0.010867,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.005773,
		0.036611,
		0.0894,
		0.159107,
		0.241006,
		0.330696,
		0.424126,
		0.517621,
		0.607912,
		0.692176,
		0.76808,
		0.833833,
		0.888248,
		0.930812,
		0.961743,
		0.982047,
		0.993502,
		0.998535,
		0.999896,
		1,
		0.999896,
		0.998535,
		0.993502,
		0.982047,
		0.961743,
		0.930812,
		0.888248,
		0.833833,
		0.76808,
		0.692176,
		0.607912,
		0.517621,
		0.424126,
		0.330696,
		0.241006,
		0.159107,
		0.0894,
		0.036611,
		0.005773,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.002165,
		0.027301,
		0.076202,
		0.143566,
		0.224414,
		0.314104,
		0.408354,
		0.503276,
		0.595403,
		0.681729,
		0.759757,
		0.827555,
		0.883818,
		0.927939,
		0.960083,
		0.981234,
		0.993195,
		0.998463,
		0.999891,
		1,
		0.999891,
		0.998463,
		0.993195,
		0.981234,
		0.960083,
		0.927939,
		0.883818,
		0.827555,
		0.759757,
		0.681729,
		0.595403,
		0.503276,
		0.408354,
		0.314104,
		0.224414,
		0.143566,
		0.076202,
		0.027301,
		0.002165,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.000251,
		0.019046,
		0.063557,
		0.128201,
		0.207724,
		0.297228,
		0.392188,
		0.488488,
		0.582448,
		0.670868,
		0.751077,
		0.820988,
		0.87917,
		0.924919,
		0.958333,
		0.980375,
		0.99287,
		0.998387,
		0.999885,
		1,
		0.999885,
		0.998387,
		0.99287,
		0.980375,
		0.958333,
		0.924919,
		0.87917,
		0.820988,
		0.751077,
		0.670868,
		0.582448,
		0.488488,
		0.392188,
		0.297228,
		0.207724,
		0.128201,
		0.063557,
		0.019046,
		0.000251,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.012014,
		0.051585,
		0.113087,
		0.19098,
		0.280089,
		0.375633,
		0.473249,
		0.569033,
		0.659576,
		0.742021,
		0.814117,
		0.874294,
		0.921741,
		0.956487,
		0.979466,
		0.992526,
		0.998306,
		0.999879,
		1,
		0.999879,
		0.998306,
		0.992526,
		0.979466,
		0.956487,
		0.921741,
		0.874294,
		0.814117,
		0.742021,
		0.659576,
		0.569033,
		0.473249,
		0.375633,
		0.280089,
		0.19098,
		0.113087,
		0.051585,
		0.012014,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.006396,
		0.040417,
		0.098313,
		0.174234,
		0.262714,
		0.358694,
		0.457553,
		0.555144,
		0.647836,
		0.732572,
		0.806924,
		0.869175,
		0.918396,
		0.954539,
		0.978505,
		0.99216,
		0.998219,
		0.999872,
		1,
		0.999872,
		0.998219,
		0.99216,
		0.978505,
		0.954539,
		0.918396,
		0.869175,
		0.806924,
		0.732572,
		0.647836,
		0.555144,
		0.457553,
		0.358694,
		0.262714,
		0.174234,
		0.098313,
		0.040417,
		0.006396,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.002404,
		0.030205,
		0.083978,
		0.157547,
		0.245133,
		0.341382,
		0.441396,
		0.540768,
		0.635631,
		0.722712,
		0.799394,
		0.863799,
		0.914874,
		0.952482,
		0.977487,
		0.991772,
		0.998127,
		0.999866,
		1,
		0.999866,
		0.998127,
		0.991772,
		0.977487,
		0.952482,
		0.914874,
		0.863799,
		0.799394,
		0.722712,
		0.635631,
		0.540768,
		0.441396,
		0.341382,
		0.245133,
		0.157547,
		0.083978,
		0.030205,
		0.002404,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.000279,
		0.021119,
		0.070199,
		0.140991,
		0.227386,
		0.323712,
		0.424774,
		0.525892,
		0.622942,
		0.71242,
		0.791508,
		0.858152,
		0.911162,
		0.950308,
		0.976409,
		0.99136,
		0.998029,
		0.999858,
		1,
		0.999858,
		0.998029,
		0.99136,
		0.976409,
		0.950308,
		0.911162,
		0.858152,
		0.791508,
		0.71242,
		0.622942,
		0.525892,
		0.424774,
		0.323712,
		0.227386,
		0.140991,
		0.070199,
		0.021119,
		0.000279,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.013353,
		0.057105,
		0.124649,
		0.209516,
		0.3057,
		0.407688,
		0.510503,
		0.609751,
		0.701678,
		0.783246,
		0.852217,
		0.90725,
		0.948011,
		0.975265,
		0.990922,
		0.997925,
		0.999851,
		1,
		0.999851,
		0.997925,
		0.990922,
		0.975265,
		0.948011,
		0.90725,
		0.852217,
		0.783246,
		0.701678,
		0.609751,
		0.510503,
		0.407688,
		0.3057,
		0.209516,
		0.124649,
		0.057105,
		0.013353,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.007125,
		0.044848,
		0.108615,
		0.191578,
		0.287371,
		0.39014,
		0.494592,
		0.596039,
		0.690463,
		0.77459,
		0.845978,
		0.903125,
		0.94558,
		0.974052,
		0.990456,
		0.997814,
		0.999842,
		1,
		0.999842,
		0.997814,
		0.990456,
		0.974052,
		0.94558,
		0.903125,
		0.845978,
		0.77459,
		0.690463,
		0.596039,
		0.494592,
		0.39014,
		0.287371,
		0.191578,
		0.108615,
		0.044848,
		0.007125,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.002685,
		0.033598,
		0.093,
		0.173636,
		0.268753,
		0.372135,
		0.478147,
		0.581788,
		0.678754,
		0.765517,
		0.839416,
		0.898772,
		0.943008,
		0.972765,
		0.98996,
		0.997695,
		0.999833,
		1,
		0.999833,
		0.997695,
		0.98996,
		0.972765,
		0.943008,
		0.898772,
		0.839416,
		0.765517,
		0.678754,
		0.581788,
		0.478147,
		0.372135,
		0.268753,
		0.173636,
		0.093,
		0.033598,
		0.002685,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.000312,
		0.02355,
		0.077932,
		0.155766,
		0.249885,
		0.353683,
		0.461161,
		0.56698,
		0.666529,
		0.756005,
		0.832511,
		0.894177,
		0.940284,
		0.971397,
		0.989431,
		0.997568,
		0.999824,
		1,
		0.999824,
		0.997568,
		0.989431,
		0.971397,
		0.940284,
		0.894177,
		0.832511,
		0.756005,
		0.666529,
		0.56698,
		0.461161,
		0.353683,
		0.249885,
		0.155766,
		0.077932,
		0.02355,
		0.000312,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.014928,
		0.063557,
		0.138055,
		0.23081,
		0.334797,
		0.443626,
		0.551596,
		0.653763,
		0.74603,
		0.825244,
		0.889324,
		0.937398,
		0.969943,
		0.988867,
		0.997432,
		0.999813,
		1,
		0.999813,
		0.997432,
		0.988867,
		0.969943,
		0.937398,
		0.889324,
		0.825244,
		0.74603,
		0.653763,
		0.551596,
		0.443626,
		0.334797,
		0.23081,
		0.138055,
		0.063557,
		0.014928,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.007987,
		0.050046,
		0.120607,
		0.211582,
		0.315497,
		0.42554,
		0.535618,
		0.640434,
		0.735568,
		0.817592,
		0.884195,
		0.934338,
		0.968397,
		0.988265,
		0.997287,
		0.999802,
		1,
		0.999802,
		0.997287,
		0.988265,
		0.968397,
		0.934338,
		0.884195,
		0.817592,
		0.735568,
		0.640434,
		0.535618,
		0.42554,
		0.315497,
		0.211582,
		0.120607,
		0.050046,
		0.007987,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.003018,
		0.037594,
		0.103544,
		0.192269,
		0.295809,
		0.406901,
		0.519031,
		0.626516,
		0.724593,
		0.809531,
		0.878773,
		0.931091,
		0.96675,
		0.987623,
		0.997131,
		0.99979,
		1,
		0.99979,
		0.997131,
		0.987623,
		0.96675,
		0.931091,
		0.878773,
		0.809531,
		0.724593,
		0.626516,
		0.519031,
		0.406901,
		0.295809,
		0.192269,
		0.103544,
		0.037594,
		0.003018,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.000352,
		0.026425,
		0.087007,
		0.172947,
		0.275766,
		0.387712,
		0.501818,
		0.611985,
		0.713077,
		0.801037,
		0.873037,
		0.927644,
		0.964996,
		0.986936,
		0.996964,
		0.999778,
		1,
		0.999778,
		0.996964,
		0.986936,
		0.964996,
		0.927644,
		0.873037,
		0.801037,
		0.713077,
		0.611985,
		0.501818,
		0.387712,
		0.275766,
		0.172947,
		0.087007,
		0.026425,
		0.000352,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.0168,
		0.071162,
		0.153712,
		0.255412,
		0.36798,
		0.483964,
		0.596815,
		0.700991,
		0.792083,
		0.866967,
		0.923982,
		0.963126,
		0.986201,
		0.996784,
		0.999764,
		1,
		0.999764,
		0.996784,
		0.986201,
		0.963126,
		0.923982,
		0.866967,
		0.792083,
		0.700991,
		0.596815,
		0.483964,
		0.36798,
		0.255412,
		0.153712,
		0.071162,
		0.0168,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.009016,
		0.0562,
		0.134676,
		0.234798,
		0.34772,
		0.465459,
		0.58098,
		0.688306,
		0.782641,
		0.860539,
		0.92009,
		0.961131,
		0.985413,
		0.996591,
		0.999749,
		1,
		0.999749,
		0.996591,
		0.985413,
		0.961131,
		0.92009,
		0.860539,
		0.782641,
		0.688306,
		0.58098,
		0.465459,
		0.34772,
		0.234798,
		0.134676,
		0.0562,
		0.009016,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.003417,
		0.042346,
		0.11597,
		0.213992,
		0.32695,
		0.446291,
		0.564456,
		0.674989,
		0.77268,
		0.853728,
		0.915948,
		0.958999,
		0.984568,
		0.996382,
		0.999732,
		0.999999,
		0.999732,
		0.996382,
		0.984568,
		0.958999,
		0.915948,
		0.853728,
		0.77268,
		0.674989,
		0.564456,
		0.446291,
		0.32695,
		0.213992,
		0.11597,
		0.042346,
		0.003417,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.0004,
		0.02986,
		0.097751,
		0.193071,
		0.305694,
		0.42645,
		0.547208,
		0.661,
		0.762157,
		0.846495,
		0.911526,
		0.956706,
		0.983646,
		0.996142,
		0.999698,
		0.999982,
		0.999698,
		0.996142,
		0.983646,
		0.956706,
		0.911526,
		0.846495,
		0.762157,
		0.661,
		0.547208,
		0.42645,
		0.305694,
		0.193071,
		0.097751,
		0.02986,
		0.0004,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.019044,
		0.0802,
		0.172124,
		0.283974,
		0.405908,
		0.529181,
		0.646263,
		0.750991,
		0.838759,
		0.906745,
		0.95418,
		0.98258,
		0.995806,
		0.999584,
		0.999887,
		0.999584,
		0.995806,
		0.98258,
		0.95418,
		0.906745,
		0.838759,
		0.750991,
		0.646263,
		0.529181,
		0.405908,
		0.283974,
		0.172124,
		0.0802,
		0.019044,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.010251,
		0.063527,
		0.151247,
		0.261794,
		0.384603,
		0.51026,
		0.630629,
		0.739009,
		0.830333,
		0.901414,
		0.951227,
		0.981178,
		0.995184,
		0.9992,
		0.999523,
		0.9992,
		0.995184,
		0.981178,
		0.951227,
		0.901414,
		0.830333,
		0.739009,
		0.630629,
		0.51026,
		0.384603,
		0.261794,
		0.151247,
		0.063527,
		0.010251,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.003895,
		0.047983,
		0.130547,
		0.239137,
		0.362411,
		0.490235,
		0.61382,
		0.72588,
		0.820849,
		0.895138,
		0.94744,
		0.979026,
		0.993859,
		0.99813,
		0.998475,
		0.99813,
		0.993859,
		0.979026,
		0.94744,
		0.895138,
		0.820849,
		0.72588,
		0.61382,
		0.490235,
		0.362411,
		0.239137,
		0.130547,
		0.047983,
		0.003895,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.000457,
		0.033873,
		0.110155,
		0.215968,
		0.339142,
		0.468778,
		0.595385,
		0.711055,
		0.809681,
		0.887236,
		0.942099,
		0.975383,
		0.991083,
		0.995622,
		0.99599,
		0.995622,
		0.991083,
		0.975383,
		0.942099,
		0.887236,
		0.809681,
		0.711055,
		0.595385,
		0.468778,
		0.339142,
		0.215968,
		0.110155,
		0.033873,
		0.000457,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.021576,
		0.090256,
		0.192263,
		0.314561,
		0.445448,
		0.574699,
		0.693744,
		0.795905,
		0.876682,
		0.934111,
		0.969117,
		0.985706,
		0.990522,
		0.990914,
		0.990522,
		0.985706,
		0.969117,
		0.934111,
		0.876682,
		0.795905,
		0.693744,
		0.574699,
		0.445448,
		0.314561,
		0.192263,
		0.090256,
		0.021576,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.011555,
		0.071125,
		0.168059,
		0.28844,
		0.419747,
		0.551001,
		0.672954,
		0.778329,
		0.862132,
		0.922022,
		0.958709,
		0.976177,
		0.981272,
		0.981688,
		0.981272,
		0.976177,
		0.958709,
		0.922022,
		0.862132,
		0.778329,
		0.672954,
		0.551001,
		0.419747,
		0.28844,
		0.168059,
		0.071125,
		0.011555,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.004344,
		0.053154,
		0.143513,
		0.260641,
		0.391212,
		0.523504,
		0.64759,
		0.755596,
		0.842016,
		0.904114,
		0.942348,
		0.960643,
		0.966004,
		0.966443,
		0.966004,
		0.960643,
		0.942348,
		0.904114,
		0.842016,
		0.755596,
		0.64759,
		0.523504,
		0.391212,
		0.260641,
		0.143513,
		0.053154,
		0.004344,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.0005,
		0.036866,
		0.118961,
		0.231212,
		0.359546,
		0.491542,
		0.616635,
		0.726371,
		0.814743,
		0.878606,
		0.918136,
		0.937149,
		0.942748,
		0.943208,
		0.942748,
		0.937149,
		0.918136,
		0.878606,
		0.814743,
		0.726371,
		0.616635,
		0.491542,
		0.359546,
		0.231212,
		0.118961,
		0.036866,
		0.0005,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.022875,
		0.094942,
		0.20047,
		0.324747,
		0.454752,
		0.579355,
		0.689582,
		0.778956,
		0.84393,
		0.884372,
		0.903927,
		0.909715,
		0.910193,
		0.909715,
		0.903927,
		0.884372,
		0.84393,
		0.778956,
		0.689582,
		0.579355,
		0.454752,
		0.324747,
		0.20047,
		0.094942,
		0.022875,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.01182,
		0.072182,
		0.169039,
		0.287213,
		0.413227,
		0.535513,
		0.644671,
		0.733821,
		0.799039,
		0.83987,
		0.859722,
		0.865629,
		0.86612,
		0.865629,
		0.859722,
		0.83987,
		0.799039,
		0.733821,
		0.644671,
		0.535513,
		0.413227,
		0.287213,
		0.169039,
		0.072182,
		0.01182,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.004244,
		0.051515,
		0.137834,
		0.247779,
		0.367622,
		0.485526,
		0.591806,
		0.679276,
		0.743688,
		0.784259,
		0.8041,
		0.810035,
		0.810531,
		0.810035,
		0.8041,
		0.784259,
		0.743688,
		0.679276,
		0.591806,
		0.485526,
		0.367622,
		0.247779,
		0.137834,
		0.051515,
		0.004244,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.000462,
		0.033766,
		0.107965,
		0.207672,
		0.319156,
		0.430524,
		0.531985,
		0.616177,
		0.678608,
		0.718179,
		0.737649,
		0.743507,
		0.743999,
		0.743507,
		0.737649,
		0.718179,
		0.678608,
		0.616177,
		0.531985,
		0.430524,
		0.319156,
		0.207672,
		0.107965,
		0.033766,
		0.000462,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.019598,
		0.080587,
		0.168377,
		0.269507,
		0.372282,
		0.467005,
		0.546299,
		0.605532,
		0.643327,
		0.662039,
		0.667704,
		0.668181,
		0.667704,
		0.662039,
		0.643327,
		0.605532,
		0.546299,
		0.467005,
		0.372282,
		0.269507,
		0.168377,
		0.080587,
		0.019598,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.009378,
		0.056729,
		0.131439,
		0.220619,
		0.313028,
		0.399286,
		0.47218,
		0.527057,
		0.562317,
		0.57989,
		0.585242,
		0.585696,
		0.585242,
		0.57989,
		0.562317,
		0.527057,
		0.47218,
		0.399286,
		0.313028,
		0.220619,
		0.131439,
		0.056729,
		0.009378,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.003088,
		0.037121,
		0.09825,
		0.174444,
		0.255168,
		0.331586,
		0.396826,
		0.446348,
		0.4784,
		0.494485,
		0.499416,
		0.499836,
		0.499416,
		0.494485,
		0.4784,
		0.446348,
		0.396826,
		0.331586,
		0.255168,
		0.174444,
		0.09825,
		0.037121,
		0.003088,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.000305,
		0.02209,
		0.069857,
		0.132688,
		0.200987,
		0.26666,
		0.32335,
		0.366762,
		0.395076,
		0.409388,
		0.413805,
		0.414184,
		0.413805,
		0.409388,
		0.395076,
		0.366762,
		0.32335,
		0.26666,
		0.200987,
		0.132688,
		0.069857,
		0.02209,
		0.000305,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.011518,
		0.046838,
		0.096614,
		0.152377,
		0.206942,
		0.254614,
		0.291463,
		0.315694,
		0.328034,
		0.33187,
		0.332201,
		0.33187,
		0.328034,
		0.315694,
		0.291463,
		0.254614,
		0.206942,
		0.152377,
		0.096614,
		0.046838,
		0.011518,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.004893,
		0.029263,
		0.066924,
		0.110655,
		0.1543,
		0.192937,
		0.223105,
		0.243113,
		0.253384,
		0.256601,
		0.25688,
		0.256601,
		0.253384,
		0.243113,
		0.223105,
		0.192937,
		0.1543,
		0.110655,
		0.066924,
		0.029263,
		0.004893,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.001409,
		0.016747,
		0.043741,
		0.07648,
		0.109901,
		0.13992,
		0.163614,
		0.179472,
		0.18768,
		0.190271,
		0.190498,
		0.190271,
		0.18768,
		0.179472,
		0.163614,
		0.13992,
		0.109901,
		0.07648,
		0.043741,
		0.016747,
		0.001409,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.000119,
		0.008548,
		0.026672,
		0.049875,
		0.074187,
		0.096378,
		0.1141,
		0.126076,
		0.13233,
		0.13432,
		0.134496,
		0.13432,
		0.13233,
		0.126076,
		0.1141,
		0.096378,
		0.074187,
		0.049875,
		0.026672,
		0.008548,
		0.000119,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.003723,
		0.014934,
		0.030318,
		0.046936,
		0.062379,
		0.07487,
		0.083398,
		0.087893,
		0.089337,
		0.089465,
		0.089337,
		0.087893,
		0.083398,
		0.07487,
		0.062379,
		0.046936,
		0.030318,
		0.014934,
		0.003723,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		0.001272,
		0.007501,
		0.016879,
		0.027382,
		0.037343,
		0.045513,
		0.051152,
		0.054154,
		0.055126,
		0.055213,
		0.055126,
		0.054154,
		0.051152,
		0.045513,
		0.037343,
		0.027382,
		0.016879,
		0.007501,
		0.001272,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.000278,
		0.003262,
		0.008381,
		0.014372,
		0.020186,
		0.025028,
		0.02841,
		0.030229,
		0.030824,
		0.030877,
		0.030824,
		0.030229,
		0.02841,
		0.025028,
		0.020186,
		0.014372,
		0.008381,
		0.003262,
		0.000278,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		1.6e-005,
		0.00116,
		0.003558,
		0.006523,
		0.009478,
		0.01198,
		0.01375,
		0.014712,
		0.01503,
		0.015059,
		0.01503,
		0.014712,
		0.01375,
		0.01198,
		0.009478,
		0.006523,
		0.003558,
		0.00116,
		1.6e-005,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000304,
		0.0012,
		0.002388,
		0.003609,
		0.004662,
		0.005418,
		0.005833,
		0.005972,
		0.005985,
		0.005972,
		0.005833,
		0.005418,
		0.004662,
		0.003609,
		0.002388,
		0.0012,
		0.000304,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		4.8e-005,
		0.000278,
		0.000613,
		0.00097,
		0.001285,
		0.001514,
		0.001641,
		0.001685,
		0.001689,
		0.001685,
		0.001641,
		0.001514,
		0.001285,
		0.00097,
		0.000613,
		0.000278,
		4.8e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		3.1e-005,
		7.7e-005,
		0.000129,
		0.000175,
		0.00021,
		0.000229,
		0.000236,
		0.000237,
		0.000236,
		0.000229,
		0.00021,
		0.000175,
		0.000129,
		7.7e-005,
		3.1e-005,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		3.1e-005,
		7.7e-005,
		0.000129,
		0.000175,
		0.00021,
		0.000229,
		0.000236,
		0.000237,
		0.000236,
		0.000229,
		0.00021,
		0.000175,
		0.000129,
		7.7e-005,
		3.1e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		4.8e-005,
		0.000278,
		0.000613,
		0.00097,
		0.001285,
		0.001514,
		0.001641,
		0.001685,
		0.001689,
		0.001685,
		0.001641,
		0.001514,
		0.001285,
		0.00097,
		0.000613,
		0.000278,
		4.8e-005,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		0.000304,
		0.0012,
		0.002388,
		0.003609,
		0.004662,
		0.005418,
		0.005833,
		0.005972,
		0.005985,
		0.005972,
		0.005833,
		0.005418,
		0.004662,
		0.003609,
		0.002388,
		0.0012,
		0.000304,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		1.6e-005,
		0.00116,
		0.003558,
		0.006523,
		0.009478,
		0.01198,
		0.01375,
		0.014712,
		0.01503,
		0.015059,
		0.01503,
		0.014712,
		0.01375,
		0.01198,
		0.009478,
		0.006523,
		0.003558,
		0.00116,
		1.6e-005,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		0.000278,
		0.003262,
		0.008381,
		0.014372,
		0.020186,
		0.025028,
		0.02841,
		0.030229,
		0.030824,
		0.030877,
		0.030824,
		0.030229,
		0.02841,
		0.025028,
		0.020186,
		0.014372,
		0.008381,
		0.003262,
		0.000278,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.001272,
		0.007501,
		0.016879,
		0.027382,
		0.037343,
		0.045513,
		0.051152,
		0.054154,
		0.055126,
		0.055213,
		0.055126,
		0.054154,
		0.051152,
		0.045513,
		0.037343,
		0.027382,
		0.016879,
		0.007501,
		0.001272,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		0.003723,
		0.014934,
		0.030318,
		0.046936,
		0.062379,
		0.07487,
		0.083398,
		0.087893,
		0.089337,
		0.089465,
		0.089337,
		0.087893,
		0.083398,
		0.07487,
		0.062379,
		0.046936,
		0.030318,
		0.014934,
		0.003723,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.000119,
		0.008548,
		0.026672,
		0.049875,
		0.074187,
		0.096378,
		0.1141,
		0.126076,
		0.13233,
		0.13432,
		0.134496,
		0.13432,
		0.13233,
		0.126076,
		0.1141,
		0.096378,
		0.074187,
		0.049875,
		0.026672,
		0.008548,
		0.000119,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.001409,
		0.016747,
		0.043741,
		0.07648,
		0.109901,
		0.13992,
		0.163614,
		0.179472,
		0.18768,
		0.190271,
		0.190498,
		0.190271,
		0.18768,
		0.179472,
		0.163614,
		0.13992,
		0.109901,
		0.07648,
		0.043741,
		0.016747,
		0.001409,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.004893,
		0.029263,
		0.066924,
		0.110655,
		0.1543,
		0.192937,
		0.223105,
		0.243113,
		0.253384,
		0.256601,
		0.25688,
		0.256601,
		0.253384,
		0.243113,
		0.223105,
		0.192937,
		0.1543,
		0.110655,
		0.066924,
		0.029263,
		0.004893,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.011518,
		0.046838,
		0.096614,
		0.152377,
		0.206942,
		0.254614,
		0.291463,
		0.315694,
		0.328034,
		0.33187,
		0.332201,
		0.33187,
		0.328034,
		0.315694,
		0.291463,
		0.254614,
		0.206942,
		0.152377,
		0.096614,
		0.046838,
		0.011518,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.000305,
		0.02209,
		0.069857,
		0.132688,
		0.200987,
		0.26666,
		0.32335,
		0.366762,
		0.395076,
		0.409388,
		0.413805,
		0.414184,
		0.413805,
		0.409388,
		0.395076,
		0.366762,
		0.32335,
		0.26666,
		0.200987,
		0.132688,
		0.069857,
		0.02209,
		0.000305,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.003088,
		0.037121,
		0.09825,
		0.174444,
		0.255168,
		0.331586,
		0.396826,
		0.446348,
		0.4784,
		0.494485,
		0.499416,
		0.499836,
		0.499416,
		0.494485,
		0.4784,
		0.446348,
		0.396826,
		0.331586,
		0.255168,
		0.174444,
		0.09825,
		0.037121,
		0.003088,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.009378,
		0.056729,
		0.131439,
		0.220619,
		0.313028,
		0.399286,
		0.47218,
		0.527057,
		0.562317,
		0.57989,
		0.585242,
		0.585696,
		0.585242,
		0.57989,
		0.562317,
		0.527057,
		0.47218,
		0.399286,
		0.313028,
		0.220619,
		0.131439,
		0.056729,
		0.009378,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.019598,
		0.080587,
		0.168377,
		0.269507,
		0.372282,
		0.467005,
		0.546299,
		0.605532,
		0.643327,
		0.662039,
		0.667704,
		0.668181,
		0.667704,
		0.662039,
		0.643327,
		0.605532,
		0.546299,
		0.467005,
		0.372282,
		0.269507,
		0.168377,
		0.080587,
		0.019598,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.000462,
		0.033766,
		0.107965,
		0.207672,
		0.319156,
		0.430524,
		0.531985,
		0.616177,
		0.678608,
		0.718179,
		0.737649,
		0.743507,
		0.743999,
		0.743507,
		0.737649,
		0.718179,
		0.678608,
		0.616177,
		0.531985,
		0.430524,
		0.319156,
		0.207672,
		0.107965,
		0.033766,
		0.000462,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.004244,
		0.051515,
		0.137834,
		0.247779,
		0.367622,
		0.485526,
		0.591806,
		0.679276,
		0.743688,
		0.784259,
		0.8041,
		0.810035,
		0.810531,
		0.810035,
		0.8041,
		0.784259,
		0.743688,
		0.679276,
		0.591806,
		0.485526,
		0.367622,
		0.247779,
		0.137834,
		0.051515,
		0.004244,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.01182,
		0.072182,
		0.169039,
		0.287213,
		0.413227,
		0.535513,
		0.644671,
		0.733821,
		0.799039,
		0.83987,
		0.859722,
		0.865629,
		0.86612,
		0.865629,
		0.859722,
		0.83987,
		0.799039,
		0.733821,
		0.644671,
		0.535513,
		0.413227,
		0.287213,
		0.169039,
		0.072182,
		0.01182,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.022875,
		0.094942,
		0.20047,
		0.324747,
		0.454752,
		0.579355,
		0.689582,
		0.778956,
		0.84393,
		0.884372,
		0.903927,
		0.909715,
		0.910193,
		0.909715,
		0.903927,
		0.884372,
		0.84393,
		0.778956,
		0.689582,
		0.579355,
		0.454752,
		0.324747,
		0.20047,
		0.094942,
		0.022875,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.0005,
		0.036866,
		0.118961,
		0.231212,
		0.359546,
		0.491542,
		0.616635,
		0.726371,
		0.814743,
		0.878606,
		0.918136,
		0.937149,
		0.942748,
		0.943208,
		0.942748,
		0.937149,
		0.918136,
		0.878606,
		0.814743,
		0.726371,
		0.616635,
		0.491542,
		0.359546,
		0.231212,
		0.118961,
		0.036866,
		0.0005,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.004344,
		0.053154,
		0.143513,
		0.260641,
		0.391212,
		0.523504,
		0.64759,
		0.755596,
		0.842016,
		0.904114,
		0.942348,
		0.960643,
		0.966004,
		0.966443,
		0.966004,
		0.960643,
		0.942348,
		0.904114,
		0.842016,
		0.755596,
		0.64759,
		0.523504,
		0.391212,
		0.260641,
		0.143513,
		0.053154,
		0.004344,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.011555,
		0.071125,
		0.168059,
		0.28844,
		0.419747,
		0.551001,
		0.672954,
		0.778329,
		0.862132,
		0.922022,
		0.958709,
		0.976177,
		0.981272,
		0.981688,
		0.981272,
		0.976177,
		0.958709,
		0.922022,
		0.862132,
		0.778329,
		0.672954,
		0.551001,
		0.419747,
		0.28844,
		0.168059,
		0.071125,
		0.011555,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.021576,
		0.090256,
		0.192263,
		0.314561,
		0.445448,
		0.574699,
		0.693744,
		0.795905,
		0.876682,
		0.934111,
		0.969117,
		0.985706,
		0.990522,
		0.990914,
		0.990522,
		0.985706,
		0.969117,
		0.934111,
		0.876682,
		0.795905,
		0.693744,
		0.574699,
		0.445448,
		0.314561,
		0.192263,
		0.090256,
		0.021576,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.000457,
		0.033873,
		0.110155,
		0.215968,
		0.339142,
		0.468778,
		0.595385,
		0.711055,
		0.809681,
		0.887236,
		0.942099,
		0.975383,
		0.991083,
		0.995622,
		0.99599,
		0.995622,
		0.991083,
		0.975383,
		0.942099,
		0.887236,
		0.809681,
		0.711055,
		0.595385,
		0.468778,
		0.339142,
		0.215968,
		0.110155,
		0.033873,
		0.000457,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.003895,
		0.047983,
		0.130547,
		0.239137,
		0.362411,
		0.490235,
		0.61382,
		0.72588,
		0.820849,
		0.895138,
		0.94744,
		0.979026,
		0.993859,
		0.99813,
		0.998475,
		0.99813,
		0.993859,
		0.979026,
		0.94744,
		0.895138,
		0.820849,
		0.72588,
		0.61382,
		0.490235,
		0.362411,
		0.239137,
		0.130547,
		0.047983,
		0.003895,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.010251,
		0.063527,
		0.151247,
		0.261794,
		0.384603,
		0.51026,
		0.630629,
		0.739009,
		0.830333,
		0.901414,
		0.951227,
		0.981178,
		0.995184,
		0.9992,
		0.999523,
		0.9992,
		0.995184,
		0.981178,
		0.951227,
		0.901414,
		0.830333,
		0.739009,
		0.630629,
		0.51026,
		0.384603,
		0.261794,
		0.151247,
		0.063527,
		0.010251,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.019044,
		0.0802,
		0.172124,
		0.283974,
		0.405908,
		0.529181,
		0.646263,
		0.750991,
		0.838759,
		0.906745,
		0.95418,
		0.98258,
		0.995806,
		0.999584,
		0.999887,
		0.999584,
		0.995806,
		0.98258,
		0.95418,
		0.906745,
		0.838759,
		0.750991,
		0.646263,
		0.529181,
		0.405908,
		0.283974,
		0.172124,
		0.0802,
		0.019044,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.0004,
		0.02986,
		0.097751,
		0.193071,
		0.305694,
		0.42645,
		0.547208,
		0.661,
		0.762157,
		0.846495,
		0.911526,
		0.956706,
		0.983646,
		0.996142,
		0.999698,
		0.999982,
		0.999698,
		0.996142,
		0.983646,
		0.956706,
		0.911526,
		0.846495,
		0.762157,
		0.661,
		0.547208,
		0.42645,
		0.305694,
		0.193071,
		0.097751,
		0.02986,
		0.0004,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.003417,
		0.042346,
		0.11597,
		0.213992,
		0.32695,
		0.446291,
		0.564456,
		0.674989,
		0.77268,
		0.853728,
		0.915948,
		0.958999,
		0.984568,
		0.996382,
		0.999732,
		0.999999,
		0.999732,
		0.996382,
		0.984568,
		0.958999,
		0.915948,
		0.853728,
		0.77268,
		0.674989,
		0.564456,
		0.446291,
		0.32695,
		0.213992,
		0.11597,
		0.042346,
		0.003417,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.009016,
		0.0562,
		0.134676,
		0.234798,
		0.34772,
		0.465459,
		0.58098,
		0.688306,
		0.782641,
		0.860539,
		0.92009,
		0.961131,
		0.985413,
		0.996591,
		0.999749,
		1,
		0.999749,
		0.996591,
		0.985413,
		0.961131,
		0.92009,
		0.860539,
		0.782641,
		0.688306,
		0.58098,
		0.465459,
		0.34772,
		0.234798,
		0.134676,
		0.0562,
		0.009016,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.0168,
		0.071162,
		0.153712,
		0.255412,
		0.36798,
		0.483964,
		0.596815,
		0.700991,
		0.792083,
		0.866967,
		0.923982,
		0.963126,
		0.986201,
		0.996784,
		0.999764,
		1,
		0.999764,
		0.996784,
		0.986201,
		0.963126,
		0.923982,
		0.866967,
		0.792083,
		0.700991,
		0.596815,
		0.483964,
		0.36798,
		0.255412,
		0.153712,
		0.071162,
		0.0168,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.000352,
		0.026425,
		0.087007,
		0.172947,
		0.275766,
		0.387712,
		0.501818,
		0.611985,
		0.713077,
		0.801037,
		0.873037,
		0.927644,
		0.964996,
		0.986936,
		0.996964,
		0.999778,
		1,
		0.999778,
		0.996964,
		0.986936,
		0.964996,
		0.927644,
		0.873037,
		0.801037,
		0.713077,
		0.611985,
		0.501818,
		0.387712,
		0.275766,
		0.172947,
		0.087007,
		0.026425,
		0.000352,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.003018,
		0.037594,
		0.103544,
		0.192269,
		0.295809,
		0.406901,
		0.519031,
		0.626516,
		0.724593,
		0.809531,
		0.878773,
		0.931091,
		0.96675,
		0.987623,
		0.997131,
		0.99979,
		1,
		0.99979,
		0.997131,
		0.987623,
		0.96675,
		0.931091,
		0.878773,
		0.809531,
		0.724593,
		0.626516,
		0.519031,
		0.406901,
		0.295809,
		0.192269,
		0.103544,
		0.037594,
		0.003018,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.007987,
		0.050046,
		0.120607,
		0.211582,
		0.315497,
		0.42554,
		0.535618,
		0.640434,
		0.735568,
		0.817592,
		0.884195,
		0.934338,
		0.968397,
		0.988265,
		0.997287,
		0.999802,
		1,
		0.999802,
		0.997287,
		0.988265,
		0.968397,
		0.934338,
		0.884195,
		0.817592,
		0.735568,
		0.640434,
		0.535618,
		0.42554,
		0.315497,
		0.211582,
		0.120607,
		0.050046,
		0.007987,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.014928,
		0.063557,
		0.138055,
		0.23081,
		0.334797,
		0.443626,
		0.551596,
		0.653763,
		0.74603,
		0.825244,
		0.889324,
		0.937398,
		0.969943,
		0.988867,
		0.997432,
		0.999813,
		1,
		0.999813,
		0.997432,
		0.988867,
		0.969943,
		0.937398,
		0.889324,
		0.825244,
		0.74603,
		0.653763,
		0.551596,
		0.443626,
		0.334797,
		0.23081,
		0.138055,
		0.063557,
		0.014928,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.000312,
		0.02355,
		0.077932,
		0.155766,
		0.249885,
		0.353683,
		0.461161,
		0.56698,
		0.666529,
		0.756005,
		0.832511,
		0.894177,
		0.940284,
		0.971397,
		0.989431,
		0.997568,
		0.999824,
		1,
		0.999824,
		0.997568,
		0.989431,
		0.971397,
		0.940284,
		0.894177,
		0.832511,
		0.756005,
		0.666529,
		0.56698,
		0.461161,
		0.353683,
		0.249885,
		0.155766,
		0.077932,
		0.02355,
		0.000312,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.002685,
		0.033598,
		0.093,
		0.173636,
		0.268753,
		0.372135,
		0.478147,
		0.581788,
		0.678754,
		0.765517,
		0.839416,
		0.898772,
		0.943008,
		0.972765,
		0.98996,
		0.997695,
		0.999833,
		1,
		0.999833,
		0.997695,
		0.98996,
		0.972765,
		0.943008,
		0.898772,
		0.839416,
		0.765517,
		0.678754,
		0.581788,
		0.478147,
		0.372135,
		0.268753,
		0.173636,
		0.093,
		0.033598,
		0.002685,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.007125,
		0.044848,
		0.108615,
		0.191578,
		0.287371,
		0.39014,
		0.494592,
		0.596039,
		0.690463,
		0.77459,
		0.845978,
		0.903125,
		0.94558,
		0.974052,
		0.990456,
		0.997814,
		0.999842,
		1,
		0.999842,
		0.997814,
		0.990456,
		0.974052,
		0.94558,
		0.903125,
		0.845978,
		0.77459,
		0.690463,
		0.596039,
		0.494592,
		0.39014,
		0.287371,
		0.191578,
		0.108615,
		0.044848,
		0.007125,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.013353,
		0.057105,
		0.124649,
		0.209516,
		0.3057,
		0.407688,
		0.510503,
		0.609751,
		0.701678,
		0.783246,
		0.852217,
		0.90725,
		0.948011,
		0.975265,
		0.990922,
		0.997925,
		0.999851,
		1,
		0.999851,
		0.997925,
		0.990922,
		0.975265,
		0.948011,
		0.90725,
		0.852217,
		0.783246,
		0.701678,
		0.609751,
		0.510503,
		0.407688,
		0.3057,
		0.209516,
		0.124649,
		0.057105,
		0.013353,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.000279,
		0.021119,
		0.070199,
		0.140991,
		0.227386,
		0.323712,
		0.424774,
		0.525892,
		0.622942,
		0.71242,
		0.791508,
		0.858152,
		0.911162,
		0.950308,
		0.976409,
		0.99136,
		0.998029,
		0.999858,
		1,
		0.999858,
		0.998029,
		0.99136,
		0.976409,
		0.950308,
		0.911162,
		0.858152,
		0.791508,
		0.71242,
		0.622942,
		0.525892,
		0.424774,
		0.323712,
		0.227386,
		0.140991,
		0.070199,
		0.021119,
		0.000279,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.002404,
		0.030205,
		0.083978,
		0.157547,
		0.245133,
		0.341382,
		0.441396,
		0.540768,
		0.635631,
		0.722712,
		0.799394,
		0.863799,
		0.914874,
		0.952482,
		0.977487,
		0.991772,
		0.998127,
		0.999866,
		1,
		0.999866,
		0.998127,
		0.991772,
		0.977487,
		0.952482,
		0.914874,
		0.863799,
		0.799394,
		0.722712,
		0.635631,
		0.540768,
		0.441396,
		0.341382,
		0.245133,
		0.157547,
		0.083978,
		0.030205,
		0.002404,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.006396,
		0.040417,
		0.098313,
		0.174234,
		0.262714,
		0.358694,
		0.457553,
		0.555144,
		0.647836,
		0.732572,
		0.806924,
		0.869175,
		0.918396,
		0.954539,
		0.978505,
		0.99216,
		0.998219,
		0.999872,
		1,
		0.999872,
		0.998219,
		0.99216,
		0.978505,
		0.954539,
		0.918396,
		0.869175,
		0.806924,
		0.732572,
		0.647836,
		0.555144,
		0.457553,
		0.358694,
		0.262714,
		0.174234,
		0.098313,
		0.040417,
		0.006396,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.012014,
		0.051585,
		0.113087,
		0.19098,
		0.280089,
		0.375633,
		0.473249,
		0.569033,
		0.659576,
		0.742021,
		0.814117,
		0.874294,
		0.921741,
		0.956487,
		0.979466,
		0.992526,
		0.998306,
		0.999879,
		1,
		0.999879,
		0.998306,
		0.992526,
		0.979466,
		0.956487,
		0.921741,
		0.874294,
		0.814117,
		0.742021,
		0.659576,
		0.569033,
		0.473249,
		0.375633,
		0.280089,
		0.19098,
		0.113087,
		0.051585,
		0.012014,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.000251,
		0.019046,
		0.063557,
		0.128201,
		0.207724,
		0.297228,
		0.392188,
		0.488488,
		0.582448,
		0.670868,
		0.751077,
		0.820988,
		0.87917,
		0.924919,
		0.958333,
		0.980375,
		0.99287,
		0.998387,
		0.999885,
		1,
		0.999885,
		0.998387,
		0.99287,
		0.980375,
		0.958333,
		0.924919,
		0.87917,
		0.820988,
		0.751077,
		0.670868,
		0.582448,
		0.488488,
		0.392188,
		0.297228,
		0.207724,
		0.128201,
		0.063557,
		0.019046,
		0.000251,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.002165,
		0.027301,
		0.076202,
		0.143566,
		0.224414,
		0.314104,
		0.408354,
		0.503276,
		0.595403,
		0.681729,
		0.759757,
		0.827555,
		0.883818,
		0.927939,
		0.960083,
		0.981234,
		0.993195,
		0.998463,
		0.999891,
		1,
		0.999891,
		0.998463,
		0.993195,
		0.981234,
		0.960083,
		0.927939,
		0.883818,
		0.827555,
		0.759757,
		0.681729,
		0.595403,
		0.503276,
		0.408354,
		0.314104,
		0.224414,
		0.143566,
		0.076202,
		0.027301,
		0.002165,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.005773,
		0.036611,
		0.0894,
		0.159107,
		0.241006,
		0.330696,
		0.424126,
		0.517621,
		0.607912,
		0.692176,
		0.76808,
		0.833833,
		0.888248,
		0.930812,
		0.961743,
		0.982047,
		0.993502,
		0.998535,
		0.999896,
		1,
		0.999896,
		0.998535,
		0.993502,
		0.982047,
		0.961743,
		0.930812,
		0.888248,
		0.833833,
		0.76808,
		0.692176,
		0.607912,
		0.517621,
		0.424126,
		0.330696,
		0.241006,
		0.159107,
		0.0894,
		0.036611,
		0.005773,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.010867,
		0.046826,
		0.10305,
		0.174757,
		0.25746,
		0.346987,
		0.439502,
		0.531529,
		0.619989,
		0.702226,
		0.776061,
		0.839836,
		0.892474,
		0.933544,
		0.963319,
		0.982817,
		0.993791,
		0.998603,
		0.999901,
		1,
		0.999901,
		0.998603,
		0.993791,
		0.982817,
		0.963319,
		0.933544,
		0.892474,
		0.839836,
		0.776061,
		0.702226,
		0.619989,
		0.531529,
		0.439502,
		0.346987,
		0.25746,
		0.174757,
		0.10305,
		0.046826,
		0.010867,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.000227,
		0.017263,
		0.057812,
		0.117059,
		0.190456,
		0.273744,
		0.362964,
		0.454482,
		0.545012,
		0.631647,
		0.711894,
		0.783716,
		0.845578,
		0.896506,
		0.936145,
		0.964815,
		0.983546,
		0.994065,
		0.998667,
		0.999906,
		1,
		0.999906,
		0.998667,
		0.994065,
		0.983546,
		0.964815,
		0.936145,
		0.896506,
		0.845578,
		0.783716,
		0.711894,
		0.631647,
		0.545012,
		0.454482,
		0.362964,
		0.273744,
		0.190456,
		0.117059,
		0.057812,
		0.017263,
		0.000227,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.00196,
		0.024796,
		0.069452,
		0.131346,
		0.206156,
		0.289832,
		0.378617,
		0.469068,
		0.558077,
		0.642901,
		0.721196,
		0.791059,
		0.851071,
		0.900354,
		0.938621,
		0.966236,
		0.984237,
		0.994324,
		0.998727,
		0.99991,
		1,
		0.99991,
		0.998727,
		0.994324,
		0.984237,
		0.966236,
		0.938621,
		0.900354,
		0.851071,
		0.791059,
		0.721196,
		0.642901,
		0.558077,
		0.469068,
		0.378617,
		0.289832,
		0.206156,
		0.131346,
		0.069452,
		0.024796,
		0.00196,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.005236,
		0.033317,
		0.08164,
		0.145843,
		0.221812,
		0.3057,
		0.393939,
		0.483264,
		0.570735,
		0.653763,
		0.730146,
		0.798105,
		0.856329,
		0.904028,
		0.94098,
		0.967586,
		0.984893,
		0.994569,
		0.998784,
		0.999914,
		1,
		0.999914,
		0.998784,
		0.994569,
		0.984893,
		0.967586,
		0.94098,
		0.904028,
		0.856329,
		0.798105,
		0.730146,
		0.653763,
		0.570735,
		0.483264,
		0.393939,
		0.3057,
		0.221812,
		0.145843,
		0.08164,
		0.033317,
		0.005236,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.009877,
		0.042694,
		0.094282,
		0.160485,
		0.237387,
		0.321329,
		0.408924,
		0.497074,
		0.582997,
		0.664249,
		0.738759,
		0.804867,
		0.861363,
		0.907537,
		0.943228,
		0.968871,
		0.985515,
		0.994801,
		0.998838,
		0.999918,
		1,
		0.999918,
		0.998838,
		0.994801,
		0.985515,
		0.968871,
		0.943228,
		0.907537,
		0.861363,
		0.804867,
		0.738759,
		0.664249,
		0.582997,
		0.497074,
		0.408924,
		0.321329,
		0.237387,
		0.160485,
		0.094282,
		0.042694,
		0.009877,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000206,
		0.015719,
		0.05281,
		0.107296,
		0.175218,
		0.252848,
		0.336705,
		0.423569,
		0.510503,
		0.594872,
		0.67437,
		0.747049,
		0.811359,
		0.866184,
		0.91089,
		0.945372,
		0.970093,
		0.986105,
		0.99502,
		0.998888,
		0.999922,
		1,
		0.999922,
		0.998888,
		0.99502,
		0.986105,
		0.970093,
		0.945372,
		0.91089,
		0.866184,
		0.811359,
		0.747049,
		0.67437,
		0.594872,
		0.510503,
		0.423569,
		0.336705,
		0.252848,
		0.175218,
		0.107296,
		0.05281,
		0.015719,
		0.000206,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.001783,
		0.02262,
		0.063557,
		0.120607,
		0.189994,
		0.268167,
		0.351814,
		0.437873,
		0.523558,
		0.606372,
		0.684139,
		0.755028,
		0.817592,
		0.870801,
		0.914095,
		0.947416,
		0.971256,
		0.986666,
		0.995229,
		0.998936,
		0.999925,
		1,
		0.999925,
		0.998936,
		0.995229,
		0.986666,
		0.971256,
		0.947416,
		0.914095,
		0.870801,
		0.817592,
		0.755028,
		0.684139,
		0.606372,
		0.523558,
		0.437873,
		0.351814,
		0.268167,
		0.189994,
		0.120607,
		0.063557,
		0.02262,
		0.001783,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.004772,
		0.030448,
		0.074843,
		0.134151,
		0.204772,
		0.283322,
		0.366646,
		0.451836,
		0.536244,
		0.617508,
		0.69357,
		0.76271,
		0.823577,
		0.875226,
		0.917159,
		0.949367,
		0.972363,
		0.987199,
		0.995426,
		0.998982,
		0.999929,
		1,
		0.999929,
		0.998982,
		0.995426,
		0.987199,
		0.972363,
		0.949367,
		0.917159,
		0.875226,
		0.823577,
		0.76271,
		0.69357,
		0.617508,
		0.536244,
		0.451836,
		0.366646,
		0.283322,
		0.204772,
		0.134151,
		0.074843,
		0.030448,
		0.004772,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.009016,
		0.039085,
		0.086581,
		0.147869,
		0.219514,
		0.29829,
		0.381194,
		0.465459,
		0.54857,
		0.628289,
		0.702674,
		0.770107,
		0.829327,
		0.879467,
		0.92009,
		0.951229,
		0.973418,
		0.987705,
		0.995614,
		0.999025,
		0.999932,
		1,
		0.999932,
		0.999025,
		0.995614,
		0.987705,
		0.973418,
		0.951229,
		0.92009,
		0.879467,
		0.829327,
		0.770107,
		0.702674,
		0.628289,
		0.54857,
		0.465459,
		0.381194,
		0.29829,
		0.219514,
		0.147869,
		0.086581,
		0.039085,
		0.009016,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000188,
		0.014374,
		0.048428,
		0.098696,
		0.16171,
		0.234189,
		0.313057,
		0.395453,
		0.478744,
		0.560543,
		0.638727,
		0.711463,
		0.77723,
		0.834852,
		0.883534,
		0.922895,
		0.953007,
		0.974424,
		0.988187,
		0.995792,
		0.999066,
		0.999935,
		1,
		0.999935,
		0.999066,
		0.995792,
		0.988187,
		0.974424,
		0.953007,
		0.922895,
		0.883534,
		0.834852,
		0.77723,
		0.711463,
		0.638727,
		0.560543,
		0.478744,
		0.395453,
		0.313057,
		0.234189,
		0.16171,
		0.098696,
		0.048428,
		0.014374,
		0.000188,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.001628,
		0.020718,
		0.05838,
		0.111122,
		0.175628,
		0.248768,
		0.327607,
		0.409418,
		0.491695,
		0.57217,
		0.648832,
		0.719949,
		0.78409,
		0.840161,
		0.887434,
		0.925579,
		0.954706,
		0.975383,
		0.988646,
		0.995961,
		0.999105,
		0.999937,
		1,
		0.999937,
		0.999105,
		0.995961,
		0.988646,
		0.975383,
		0.954706,
		0.925579,
		0.887434,
		0.840161,
		0.78409,
		0.719949,
		0.648832,
		0.57217,
		0.491695,
		0.409418,
		0.327607,
		0.248768,
		0.175628,
		0.111122,
		0.05838,
		0.020718,
		0.001628,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.004366,
		0.027933,
		0.068856,
		0.123796,
		0.189584,
		0.263228,
		0.341929,
		0.423087,
		0.504315,
		0.583461,
		0.658615,
		0.728142,
		0.790699,
		0.845265,
		0.891175,
		0.928149,
		0.956329,
		0.976297,
		0.989083,
		0.996121,
		0.999141,
		0.99994,
		1,
		0.99994,
		0.999141,
		0.996121,
		0.989083,
		0.976297,
		0.956329,
		0.928149,
		0.891175,
		0.845265,
		0.790699,
		0.728142,
		0.658615,
		0.583461,
		0.504315,
		0.423087,
		0.341929,
		0.263228,
		0.189584,
		0.123796,
		0.068856,
		0.027933,
		0.004366,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.008262,
		0.035914,
		0.079781,
		0.136666,
		0.203541,
		0.277548,
		0.356013,
		0.436458,
		0.516611,
		0.594422,
		0.668086,
		0.736055,
		0.797066,
		0.850172,
		0.894765,
		0.930611,
		0.957881,
		0.97717,
		0.989499,
		0.996274,
		0.999176,
		0.999943,
		1,
		0.999943,
		0.999176,
		0.996274,
		0.989499,
		0.97717,
		0.957881,
		0.930611,
		0.894765,
		0.850172,
		0.797066,
		0.736055,
		0.668086,
		0.594422,
		0.516611,
		0.436458,
		0.356013,
		0.277548,
		0.203541,
		0.136666,
		0.079781,
		0.035914,
		0.008262,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000172,
		0.013194,
		0.044568,
		0.091084,
		0.149684,
		0.21747,
		0.29171,
		0.369853,
		0.449532,
		0.528585,
		0.605064,
		0.677255,
		0.743696,
		0.803203,
		0.854892,
		0.898211,
		0.93297,
		0.959365,
		0.978003,
		0.989896,
		0.996419,
		0.99921,
		0.999945,
		1,
		0.999945,
		0.99921,
		0.996419,
		0.989896,
		0.978003,
		0.959365,
		0.93297,
		0.898211,
		0.854892,
		0.803203,
		0.743696,
		0.677255,
		0.605064,
		0.528585,
		0.449532,
		0.369853,
		0.29171,
		0.21747,
		0.149684,
		0.091084,
		0.044568,
		0.013194,
		0.000172,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.001493,
		0.019046,
		0.053808,
		0.102704,
		0.162806,
		0.231342,
		0.3057,
		0.383441,
		0.46231,
		0.540245,
		0.615394,
		0.686131,
		0.751077,
		0.809117,
		0.859431,
		0.90152,
		0.935231,
		0.960786,
		0.978799,
		0.990274,
		0.996558,
		0.999241,
		0.999947,
		1,
		0.999947,
		0.999241,
		0.996558,
		0.990274,
		0.978799,
		0.960786,
		0.935231,
		0.90152,
		0.859431,
		0.809117,
		0.751077,
		0.686131,
		0.615394,
		0.540245,
		0.46231,
		0.383441,
		0.3057,
		0.231342,
		0.162806,
		0.102704,
		0.053808,
		0.019046,
		0.001493,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.00401,
		0.025717,
		0.063557,
		0.114585,
		0.175995,
		0.245133,
		0.319503,
		0.396775,
		0.474793,
		0.551596,
		0.62542,
		0.694725,
		0.758206,
		0.814818,
		0.863799,
		0.904698,
		0.937398,
		0.962145,
		0.97956,
		0.990634,
		0.99669,
		0.999271,
		0.999949,
		1,
		0.999949,
		0.999271,
		0.99669,
		0.990634,
		0.97956,
		0.962145,
		0.937398,
		0.904698,
		0.863799,
		0.814818,
		0.758206,
		0.694725,
		0.62542,
		0.551596,
		0.474793,
		0.396775,
		0.319503,
		0.245133,
		0.175995,
		0.114585,
		0.063557,
		0.025717,
		0.00401,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.0076,
		0.033113,
		0.073747,
		0.126677,
		0.189216,
		0.258824,
		0.333111,
		0.409849,
		0.486985,
		0.562644,
		0.635152,
		0.703046,
		0.765094,
		0.820315,
		0.868003,
		0.907751,
		0.939477,
		0.963446,
		0.980287,
		0.990978,
		0.996815,
		0.999299,
		0.999951,
		1,
		0.999951,
		0.999299,
		0.996815,
		0.990978,
		0.980287,
		0.963446,
		0.939477,
		0.907751,
		0.868003,
		0.820315,
		0.765094,
		0.703046,
		0.635152,
		0.562644,
		0.486985,
		0.409849,
		0.333111,
		0.258824,
		0.189216,
		0.126677,
		0.073747,
		0.033113,
		0.0076,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000158,
		0.012153,
		0.04115,
		0.084314,
		0.138935,
		0.20244,
		0.272394,
		0.346512,
		0.422664,
		0.498887,
		0.573396,
		0.644597,
		0.711103,
		0.771749,
		0.825616,
		0.872049,
		0.910685,
		0.941471,
		0.964693,
		0.980982,
		0.991307,
		0.996935,
		0.999327,
		0.999953,
		1,
		0.999953,
		0.999327,
		0.996935,
		0.991307,
		0.980982,
		0.964693,
		0.941471,
		0.910685,
		0.872049,
		0.825616,
		0.771749,
		0.711103,
		0.644597,
		0.573396,
		0.498887,
		0.422664,
		0.346512,
		0.272394,
		0.20244,
		0.138935,
		0.084314,
		0.04115,
		0.012153,
		0.000158,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.001374,
		0.017568,
		0.049751,
		0.0952,
		0.151319,
		0.21564,
		0.28583,
		0.359701,
		0.435217,
		0.510503,
		0.583858,
		0.653763,
		0.718904,
		0.77818,
		0.830729,
		0.875946,
		0.913505,
		0.943385,
		0.965887,
		0.981647,
		0.991621,
		0.997049,
		0.999352,
		0.999955,
		1,
		0.999955,
		0.999352,
		0.997049,
		0.991621,
		0.981647,
		0.965887,
		0.943385,
		0.913505,
		0.875946,
		0.830729,
		0.77818,
		0.718904,
		0.653763,
		0.583858,
		0.510503,
		0.435217,
		0.359701,
		0.28583,
		0.21564,
		0.151319,
		0.0952,
		0.049751,
		0.017568,
		0.001374,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.003696,
		0.023754,
		0.058845,
		0.106356,
		0.163793,
		0.228792,
		0.299116,
		0.372671,
		0.447509,
		0.521839,
		0.594037,
		0.66266,
		0.726459,
		0.784396,
		0.835662,
		0.879698,
		0.916216,
		0.945222,
		0.967032,
		0.982283,
		0.99192,
		0.997158,
		0.999377,
		0.999957,
		1,
		0.999957,
		0.999377,
		0.997158,
		0.99192,
		0.982283,
		0.967032,
		0.945222,
		0.916216,
		0.879698,
		0.835662,
		0.784396,
		0.726459,
		0.66266,
		0.594037,
		0.521839,
		0.447509,
		0.372671,
		0.299116,
		0.228792,
		0.163793,
		0.106356,
		0.058845,
		0.023754,
		0.003696,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.007014,
		0.030627,
		0.06837,
		0.117733,
		0.176325,
		0.241875,
		0.312242,
		0.385418,
		0.45954,
		0.532897,
		0.603939,
		0.671294,
		0.733775,
		0.790404,
		0.840421,
		0.883312,
		0.918824,
		0.946986,
		0.968129,
		0.982892,
		0.992207,
		0.997262,
		0.9994,
		0.999959,
		1,
		0.999959,
		0.9994,
		0.997262,
		0.992207,
		0.982892,
		0.968129,
		0.946986,
		0.918824,
		0.883312,
		0.840421,
		0.790404,
		0.733775,
		0.671294,
		0.603939,
		0.532897,
		0.45954,
		0.385418,
		0.312242,
		0.241875,
		0.176325,
		0.117733,
		0.06837,
		0.030627,
		0.007014,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.000146,
		0.011231,
		0.03811,
		0.078267,
		0.129291,
		0.188885,
		0.254872,
		0.325198,
		0.397939,
		0.471312,
		0.543682,
		0.613572,
		0.679673,
		0.740862,
		0.796212,
		0.845015,
		0.886795,
		0.921332,
		0.94868,
		0.969182,
		0.983475,
		0.992481,
		0.997361,
		0.999423,
		0.99996,
		1,
		0.99996,
		0.999423,
		0.997361,
		0.992481,
		0.983475,
		0.969182,
		0.94868,
		0.921332,
		0.886795,
		0.845015,
		0.796212,
		0.740862,
		0.679673,
		0.613572,
		0.543682,
		0.471312,
		0.397939,
		0.325198,
		0.254872,
		0.188885,
		0.129291,
		0.078267,
		0.03811,
		0.011231,
		0.000146,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.001269,
		0.016256,
		0.046135,
		0.088485,
		0.140991,
		0.201449,
		0.267766,
		0.337974,
		0.41023,
		0.482827,
		0.554201,
		0.622942,
		0.687806,
		0.747726,
		0.801828,
		0.849448,
		0.890151,
		0.923746,
		0.950308,
		0.970192,
		0.984034,
		0.992743,
		0.997456,
		0.999444,
		0.999962,
		1,
		0.999962,
		0.999444,
		0.997456,
		0.992743,
		0.984034,
		0.970192,
		0.950308,
		0.923746,
		0.890151,
		0.849448,
		0.801828,
		0.747726,
		0.687806,
		0.622942,
		0.554201,
		0.482827,
		0.41023,
		0.337974,
		0.267766,
		0.201449,
		0.140991,
		0.088485,
		0.046135,
		0.016256,
		0.001269,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.003417,
		0.022008,
		0.054636,
		0.098976,
		0.1528,
		0.213992,
		0.280543,
		0.350565,
		0.422291,
		0.494086,
		0.564456,
		0.632056,
		0.695699,
		0.754375,
		0.807258,
		0.853729,
		0.893386,
		0.926069,
		0.951873,
		0.971161,
		0.984569,
		0.992994,
		0.997546,
		0.999464,
		0.999963,
		1,
		0.999963,
		0.999464,
		0.997546,
		0.992994,
		0.984569,
		0.971161,
		0.951873,
		0.926069,
		0.893386,
		0.853729,
		0.807258,
		0.754375,
		0.695699,
		0.632056,
		0.564456,
		0.494086,
		0.422291,
		0.350565,
		0.280543,
		0.213992,
		0.1528,
		0.098976,
		0.054636,
		0.022008,
		0.003417,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.006493,
		0.02841,
		0.063557,
		0.109696,
		0.164687,
		0.226495,
		0.293191,
		0.362964,
		0.43412,
		0.505093,
		0.574455,
		0.64092,
		0.703361,
		0.760817,
		0.812511,
		0.857862,
		0.896506,
		0.928306,
		0.953377,
		0.972091,
		0.985083,
		0.993234,
		0.997633,
		0.999484,
		0.999964,
		1,
		0.999964,
		0.999484,
		0.997633,
		0.993234,
		0.985083,
		0.972091,
		0.953377,
		0.928306,
		0.896506,
		0.857862,
		0.812511,
		0.760817,
		0.703361,
		0.64092,
		0.574455,
		0.505093,
		0.43412,
		0.362964,
		0.293191,
		0.226495,
		0.164687,
		0.109696,
		0.063557,
		0.02841,
		0.006493,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.000135,
		0.010409,
		0.035395,
		0.072845,
		0.120607,
		0.176624,
		0.238939,
		0.3057,
		0.375167,
		0.445718,
		0.515852,
		0.584202,
		0.649542,
		0.710798,
		0.767059,
		0.817592,
		0.861854,
		0.899514,
		0.93046,
		0.954824,
		0.972985,
		0.985575,
		0.993464,
		0.997715,
		0.999502,
		0.999966,
		1,
		0.999966,
		0.999502,
		0.997715,
		0.993464,
		0.985575,
		0.972985,
		0.954824,
		0.93046,
		0.899514,
		0.861854,
		0.817592,
		0.767059,
		0.710798,
		0.649542,
		0.584202,
		0.515852,
		0.445718,
		0.375167,
		0.3057,
		0.238939,
		0.176624,
		0.120607,
		0.072845,
		0.035395,
		0.010409,
		0.000135,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.001176,
		0.015085,
		0.042898,
		0.082451,
		0.131674,
		0.188586,
		0.251308,
		0.31806,
		0.387171,
		0.457085,
		0.526364,
		0.593702,
		0.657927,
		0.718017,
		0.773107,
		0.822507,
		0.865711,
		0.902416,
		0.932535,
		0.956215,
		0.973843,
		0.986047,
		0.993685,
		0.997795,
		0.99952,
		0.999967,
		1,
		0.999967,
		0.99952,
		0.997795,
		0.993685,
		0.986047,
		0.973843,
		0.956215,
		0.932535,
		0.902416,
		0.865711,
		0.822507,
		0.773107,
		0.718017,
		0.657927,
		0.593702,
		0.526364,
		0.457085,
		0.387171,
		0.31806,
		0.251308,
		0.188586,
		0.131674,
		0.082451,
		0.042898,
		0.015085,
		0.001176,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.003169,
		0.020447,
		0.050862,
		0.092332,
		0.142864,
		0.200552,
		0.263588,
		0.330263,
		0.398972,
		0.468221,
		0.536635,
		0.602961,
		0.666082,
		0.725025,
		0.778969,
		0.827263,
		0.869437,
		0.905215,
		0.934534,
		0.957554,
		0.974668,
		0.986501,
		0.993896,
		0.99787,
		0.999537,
		0.999968,
		1,
		0.999968,
		0.999537,
		0.99787,
		0.993896,
		0.986501,
		0.974668,
		0.957554,
		0.934534,
		0.905215,
		0.869437,
		0.827263,
		0.778969,
		0.725025,
		0.666082,
		0.602961,
		0.536635,
		0.468221,
		0.398972,
		0.330263,
		0.263588,
		0.200552,
		0.142864,
		0.092332,
		0.050862,
		0.020447,
		0.003169,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.006028,
		0.026425,
		0.059234,
		0.102448,
		0.154148,
		0.2125,
		0.275766,
		0.342303,
		0.410568,
		0.479129,
		0.546667,
		0.611985,
		0.674014,
		0.731828,
		0.78465,
		0.831866,
		0.873037,
		0.907917,
		0.936461,
		0.958843,
		0.97546,
		0.986936,
		0.994098,
		0.997943,
		0.999553,
		0.999969,
		1,
		0.999969,
		0.999553,
		0.997943,
		0.994098,
		0.986936,
		0.97546,
		0.958843,
		0.936461,
		0.907917,
		0.873037,
		0.831866,
		0.78465,
		0.731828,
		0.674014,
		0.611985,
		0.546667,
		0.479129,
		0.410568,
		0.342303,
		0.275766,
		0.2125,
		0.154148,
		0.102448,
		0.059234,
		0.026425,
		0.006028,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.000125,
		0.009675,
		0.032959,
		0.067965,
		0.112762,
		0.1655,
		0.224414,
		0.287832,
		0.354174,
		0.421959,
		0.489811,
		0.556466,
		0.620779,
		0.681729,
		0.738434,
		0.790157,
		0.836321,
		0.876517,
		0.910525,
		0.938318,
		0.960083,
		0.976223,
		0.987353,
		0.994292,
		0.998013,
		0.999569,
		0.99997,
		1,
		0.99997,
		0.999569,
		0.998013,
		0.994292,
		0.987353,
		0.976223,
		0.960083,
		0.938318,
		0.910525,
		0.876517,
		0.836321,
		0.790157,
		0.738434,
		0.681729,
		0.620779,
		0.556466,
		0.489811,
		0.421959,
		0.354174,
		0.287832,
		0.224414,
		0.1655,
		0.112762,
		0.067965,
		0.032959,
		0.009675,
		0.000125,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.001092,
		0.014037,
		0.039989,
		0.077011,
		0.123241,
		0.176896,
		0.236278,
		0.299776,
		0.365872,
		0.433143,
		0.500268,
		0.566035,
		0.629348,
		0.689233,
		0.744848,
		0.795497,
		0.840634,
		0.879882,
		0.913042,
		0.940109,
		0.961278,
		0.976956,
		0.987755,
		0.994478,
		0.998079,
		0.999584,
		0.999971,
		1,
		0.999971,
		0.999584,
		0.998079,
		0.994478,
		0.987755,
		0.976956,
		0.961278,
		0.940109,
		0.913042,
		0.879882,
		0.840634,
		0.795497,
		0.744848,
		0.689233,
		0.629348,
		0.566035,
		0.500268,
		0.433143,
		0.365872,
		0.299776,
		0.236278,
		0.176896,
		0.123241,
		0.077011,
		0.039989,
		0.014037,
		0.001092,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.002946,
		0.019046,
		0.047464,
		0.086332,
		0.133854,
		0.188314,
		0.248076,
		0.31159,
		0.377394,
		0.444121,
		0.510503,
		0.575379,
		0.637698,
		0.696531,
		0.751077,
		0.800673,
		0.844809,
		0.883135,
		0.915473,
		0.941836,
		0.962429,
		0.977661,
		0.98814,
		0.994657,
		0.998143,
		0.999598,
		0.999972,
		1,
		0.999972,
		0.999598,
		0.998143,
		0.994657,
		0.98814,
		0.977661,
		0.962429,
		0.941836,
		0.915473,
		0.883135,
		0.844809,
		0.800673,
		0.751077,
		0.696531,
		0.637698,
		0.575379,
		0.510503,
		0.444121,
		0.377394,
		0.31159,
		0.248076,
		0.188314,
		0.133854,
		0.086332,
		0.047464,
		0.019046,
		0.002946,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.005611,
		0.024641,
		0.055335,
		0.09589,
		0.144575,
		0.199736,
		0.259797,
		0.323265,
		0.388735,
		0.454892,
		0.52052,
		0.584502,
		0.645835,
		0.70363,
		0.757125,
		0.805693,
		0.848853,
		0.886281,
		0.917822,
		0.943502,
		0.963537,
		0.978339,
		0.988511,
		0.994829,
		0.998204,
		0.999611,
		0.999973,
		1,
		0.999973,
		0.999611,
		0.998204,
		0.994829,
		0.988511,
		0.978339,
		0.963537,
		0.943502,
		0.917822,
		0.886281,
		0.848853,
		0.805693,
		0.757125,
		0.70363,
		0.645835,
		0.584502,
		0.52052,
		0.454892,
		0.388735,
		0.323265,
		0.259797,
		0.199736,
		0.144575,
		0.09589,
		0.055335,
		0.024641,
		0.005611,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.000116,
		0.009016,
		0.030765,
		0.063557,
		0.105651,
		0.155379,
		0.211144,
		0.271429,
		0.334797,
		0.399895,
		0.465459,
		0.53032,
		0.593409,
		0.653763,
		0.710536,
		0.763,
		0.810562,
		0.852769,
		0.889324,
		0.92009,
		0.945109,
		0.964606,
		0.978993,
		0.988867,
		0.994994,
		0.998263,
		0.999624,
		0.999974,
		1,
		0.999974,
		0.999624,
		0.998263,
		0.994994,
		0.988867,
		0.978993,
		0.964606,
		0.945109,
		0.92009,
		0.889324,
		0.852769,
		0.810562,
		0.763,
		0.710536,
		0.653763,
		0.593409,
		0.53032,
		0.465459,
		0.399895,
		0.334797,
		0.271429,
		0.211144,
		0.155379,
		0.105651,
		0.063557,
		0.030765,
		0.009016,
		0.000116,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.001017,
		0.013093,
		0.037366,
		0.07209,
		0.115585,
		0.166242,
		0.222522,
		0.282963,
		0.34618,
		0.410871,
		0.475821,
		0.539907,
		0.602103,
		0.661489,
		0.717253,
		0.768706,
		0.815284,
		0.856563,
		0.892267,
		0.922282,
		0.946661,
		0.965636,
		0.979622,
		0.98921,
		0.995152,
		0.998319,
		0.999637,
		0.999975,
		1,
		0.999975,
		0.999637,
		0.998319,
		0.995152,
		0.98921,
		0.979622,
		0.965636,
		0.946661,
		0.922282,
		0.892267,
		0.856563,
		0.815284,
		0.768706,
		0.717253,
		0.661489,
		0.602103,
		0.539907,
		0.475821,
		0.410871,
		0.34618,
		0.282963,
		0.222522,
		0.166242,
		0.115585,
		0.07209,
		0.037366,
		0.013093,
		0.001017,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.002747,
		0.017784,
		0.044395,
		0.080895,
		0.125662,
		0.177144,
		0.233856,
		0.294389,
		0.357409,
		0.421662,
		0.48598,
		0.549284,
		0.61059,
		0.669016,
		0.723788,
		0.774249,
		0.819865,
		0.860238,
		0.895116,
		0.9244,
		0.948158,
		0.966629,
		0.980227,
		0.98954,
		0.995304,
		0.998373,
		0.999649,
		0.999976,
		1,
		0.999976,
		0.999649,
		0.998373,
		0.995304,
		0.98954,
		0.980227,
		0.966629,
		0.948158,
		0.9244,
		0.895116,
		0.860238,
		0.819865,
		0.774249,
		0.723788,
		0.669016,
		0.61059,
		0.549284,
		0.48598,
		0.421662,
		0.357409,
		0.294389,
		0.233856,
		0.177144,
		0.125662,
		0.080895,
		0.044395,
		0.017784,
		0.002747,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.005236,
		0.023032,
		0.051808,
		0.089938,
		0.135857,
		0.188066,
		0.245133,
		0.3057,
		0.36848,
		0.432268,
		0.495938,
		0.558455,
		0.618874,
		0.676351,
		0.730146,
		0.779633,
		0.824309,
		0.863799,
		0.897873,
		0.926448,
		0.949604,
		0.967586,
		0.980811,
		0.989857,
		0.99545,
		0.998425,
		0.99966,
		0.999977,
		1,
		0.999977,
		0.99966,
		0.998425,
		0.99545,
		0.989857,
		0.980811,
		0.967586,
		0.949604,
		0.926448,
		0.897873,
		0.863799,
		0.824309,
		0.779633,
		0.730146,
		0.676351,
		0.618874,
		0.558455,
		0.495938,
		0.432268,
		0.36848,
		0.3057,
		0.245133,
		0.188066,
		0.135857,
		0.089938,
		0.051808,
		0.023032,
		0.005236,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000108,
		0.008421,
		0.028784,
		0.059563,
		0.099188,
		0.146146,
		0.198992,
		0.256343,
		0.316889,
		0.379391,
		0.442688,
		0.505698,
		0.567424,
		0.62696,
		0.683499,
		0.736332,
		0.784865,
		0.828621,
		0.867251,
		0.900541,
		0.928428,
		0.951001,
		0.96851,
		0.981373,
		0.990162,
		0.995591,
		0.998475,
		0.999671,
		0.999978,
		1,
		0.999978,
		0.999671,
		0.998475,
		0.995591,
		0.990162,
		0.981373,
		0.96851,
		0.951001,
		0.928428,
		0.900541,
		0.867251,
		0.828621,
		0.784865,
		0.736332,
		0.683499,
		0.62696,
		0.567424,
		0.505698,
		0.442688,
		0.379391,
		0.316889,
		0.256343,
		0.198992,
		0.146146,
		0.099188,
		0.059563,
		0.028784,
		0.008421,
		0.000108,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.000949,
		0.012242,
		0.034993,
		0.067623,
		0.108615,
		0.156508,
		0.209905,
		0.267474,
		0.327951,
		0.39014,
		0.452924,
		0.51526,
		0.576193,
		0.634852,
		0.690463,
		0.742351,
		0.789949,
		0.832806,
		0.870596,
		0.903125,
		0.930343,
		0.95235,
		0.969402,
		0.981915,
		0.990456,
		0.995726,
		0.998523,
		0.999682,
		0.999978,
		1,
		0.999978,
		0.999682,
		0.998523,
		0.995726,
		0.990456,
		0.981915,
		0.969402,
		0.95235,
		0.930343,
		0.903125,
		0.870596,
		0.832806,
		0.789949,
		0.742351,
		0.690463,
		0.634852,
		0.576193,
		0.51526,
		0.452924,
		0.39014,
		0.327951,
		0.267474,
		0.209905,
		0.156508,
		0.108615,
		0.067623,
		0.034993,
		0.012242,
		0.000949,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.002567,
		0.016644,
		0.041614,
		0.075953,
		0.118192,
		0.166922,
		0.220793,
		0.278519,
		0.33888,
		0.400724,
		0.462975,
		0.524629,
		0.584767,
		0.642555,
		0.69725,
		0.748208,
		0.794889,
		0.836868,
		0.873839,
		0.905627,
		0.932195,
		0.953653,
		0.970262,
		0.982438,
		0.990739,
		0.995856,
		0.998569,
		0.999692,
		0.999979,
		1,
		0.999979,
		0.999692,
		0.998569,
		0.995856,
		0.990739,
		0.982438,
		0.970262,
		0.953653,
		0.932195,
		0.905627,
		0.873839,
		0.836868,
		0.794889,
		0.748208,
		0.69725,
		0.642555,
		0.584767,
		0.524629,
		0.462975,
		0.400724,
		0.33888,
		0.278519,
		0.220793,
		0.166922,
		0.118192,
		0.075953,
		0.041614,
		0.016644,
		0.002567,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.004898,
		0.021574,
		0.048607,
		0.08452,
		0.127896,
		0.177371,
		0.231643,
		0.289469,
		0.349672,
		0.411143,
		0.472842,
		0.533806,
		0.59315,
		0.650073,
		0.703863,
		0.753907,
		0.79969,
		0.84081,
		0.876983,
		0.908049,
		0.933986,
		0.954912,
		0.971093,
		0.982941,
		0.991012,
		0.995981,
		0.998614,
		0.999702,
		0.99998,
		1,
		0.99998,
		0.999702,
		0.998614,
		0.995981,
		0.991012,
		0.982941,
		0.971093,
		0.954912,
		0.933986,
		0.908049,
		0.876983,
		0.84081,
		0.79969,
		0.753907,
		0.703863,
		0.650073,
		0.59315,
		0.533806,
		0.472842,
		0.411143,
		0.349672,
		0.289469,
		0.231643,
		0.177371,
		0.127896,
		0.08452,
		0.048607,
		0.021574,
		0.004898,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000101,
		0.007884,
		0.026987,
		0.055933,
		0.093296,
		0.137703,
		0.187838,
		0.242442,
		0.300316,
		0.360324,
		0.421394,
		0.482527,
		0.542794,
		0.601344,
		0.65741,
		0.710309,
		0.759454,
		0.804357,
		0.844638,
		0.880032,
		0.910396,
		0.93572,
		0.95613,
		0.971895,
		0.983428,
		0.991275,
		0.996102,
		0.998656,
		0.999711,
		0.99998,
		1,
		0.99998,
		0.999711,
		0.998656,
		0.996102,
		0.991275,
		0.983428,
		0.971895,
		0.95613,
		0.93572,
		0.910396,
		0.880032,
		0.844638,
		0.804357,
		0.759454,
		0.710309,
		0.65741,
		0.601344,
		0.542794,
		0.482527,
		0.421394,
		0.360324,
		0.300316,
		0.242442,
		0.187838,
		0.137703,
		0.093296,
		0.055933,
		0.026987,
		0.007884,
		0.000101,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.000888,
		0.011471,
		0.032838,
		0.063557,
		0.102252,
		0.147593,
		0.198309,
		0.253183,
		0.311056,
		0.370832,
		0.431479,
		0.492031,
		0.551596,
		0.609355,
		0.66457,
		0.71659,
		0.764852,
		0.808893,
		0.848355,
		0.882989,
		0.91267,
		0.937398,
		0.957307,
		0.97267,
		0.983896,
		0.991528,
		0.996218,
		0.998697,
		0.99972,
		0.999981,
		1,
		0.999981,
		0.99972,
		0.998697,
		0.996218,
		0.991528,
		0.983896,
		0.97267,
		0.957307,
		0.937398,
		0.91267,
		0.882989,
		0.848355,
		0.808893,
		0.764852,
		0.71659,
		0.66457,
		0.609355,
		0.551596,
		0.492031,
		0.431479,
		0.370832,
		0.311056,
		0.253183,
		0.198309,
		0.147593,
		0.102252,
		0.063557,
		0.032838,
		0.011471,
		0.000888,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.002404,
		0.015609,
		0.039085,
		0.071448,
		0.111363,
		0.157547,
		0.208769,
		0.263854,
		0.321682,
		0.381194,
		0.441396,
		0.501356,
		0.560215,
		0.617185,
		0.671559,
		0.722712,
		0.770107,
		0.813304,
		0.851964,
		0.885858,
		0.914874,
		0.939022,
		0.958445,
		0.973418,
		0.984349,
		0.991772,
		0.996329,
		0.998737,
		0.999729,
		0.999982,
		1,
		0.999982,
		0.999729,
		0.998737,
		0.996329,
		0.991772,
		0.984349,
		0.973418,
		0.958445,
		0.939022,
		0.914874,
		0.885858,
		0.851964,
		0.813304,
		0.770107,
		0.722712,
		0.671559,
		0.617185,
		0.560215,
		0.501356,
		0.441396,
		0.381194,
		0.321682,
		0.263854,
		0.208769,
		0.157547,
		0.111363,
		0.071448,
		0.039085,
		0.015609,
		0.002404,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		0.004591,
		0.020251,
		0.045692,
		0.079575,
		0.120607,
		0.167548,
		0.219207,
		0.274448,
		0.332189,
		0.391409,
		0.451146,
		0.510503,
		0.568653,
		0.624839,
		0.67838,
		0.728678,
		0.775222,
		0.817592,
		0.855469,
		0.888641,
		0.917009,
		0.940595,
		0.959546,
		0.974141,
		0.984786,
		0.992008,
		0.996437,
		0.998774,
		0.999737,
		0.999982,
		1,
		0.999982,
		0.999737,
		0.998774,
		0.996437,
		0.992008,
		0.984786,
		0.974141,
		0.959546,
		0.940595,
		0.917009,
		0.888641,
		0.855469,
		0.817592,
		0.775222,
		0.728678,
		0.67838,
		0.624839,
		0.568653,
		0.510503,
		0.451146,
		0.391409,
		0.332189,
		0.274448,
		0.219207,
		0.167548,
		0.120607,
		0.079575,
		0.045692,
		0.020251,
		0.004591,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		9.5e-005,
		0.007397,
		0.025353,
		0.052623,
		0.08791,
		0.129963,
		0.17758,
		0.229612,
		0.284958,
		0.342574,
		0.401474,
		0.460729,
		0.519475,
		0.576915,
		0.63232,
		0.685037,
		0.734494,
		0.780201,
		0.821761,
		0.858873,
		0.891341,
		0.919079,
		0.942118,
		0.960612,
		0.97484,
		0.985208,
		0.992235,
		0.996541,
		0.998811,
		0.999745,
		0.999983,
		1,
		0.999983,
		0.999745,
		0.998811,
		0.996541,
		0.992235,
		0.985208,
		0.97484,
		0.960612,
		0.942118,
		0.919079,
		0.891341,
		0.858873,
		0.821761,
		0.780201,
		0.734494,
		0.685037,
		0.63232,
		0.576915,
		0.519475,
		0.460729,
		0.401474,
		0.342574,
		0.284958,
		0.229612,
		0.17758,
		0.129963,
		0.08791,
		0.052623,
		0.025353,
		0.007397,
		9.5e-005,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		0.000833,
		0.010771,
		0.030875,
		0.059846,
		0.096427,
		0.139409,
		0.187629,
		0.239972,
		0.295377,
		0.352834,
		0.411389,
		0.470147,
		0.528274,
		0.585003,
		0.639632,
		0.691535,
		0.740163,
		0.785049,
		0.825816,
		0.862181,
		0.893962,
		0.921086,
		0.943593,
		0.961642,
		0.975516,
		0.985615,
		0.992454,
		0.996641,
		0.998846,
		0.999753,
		0.999983,
		1,
		0.999983,
		0.999753,
		0.998846,
		0.996641,
		0.992454,
		0.985615,
		0.975516,
		0.961642,
		0.943593,
		0.921086,
		0.893962,
		0.862181,
		0.825816,
		0.785049,
		0.740163,
		0.691535,
		0.639632,
		0.585003,
		0.528274,
		0.470147,
		0.411389,
		0.352834,
		0.295377,
		0.239972,
		0.187629,
		0.139409,
		0.096427,
		0.059846,
		0.030875,
		0.010771,
		0.000833,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.002256,
		0.014668,
		0.03678,
		0.06733,
		0.105104,
		0.14893,
		0.197681,
		0.25028,
		0.3057,
		0.362964,
		0.421152,
		0.479399,
		0.536902,
		0.592919,
		0.646778,
		0.697877,
		0.745689,
		0.78977,
		0.82976,
		0.865395,
		0.896506,
		0.923032,
		0.945021,
		0.96264,
		0.976169,
		0.986009,
		0.992666,
		0.996737,
		0.99888,
		0.99976,
		0.999984,
		1,
		0.999984,
		0.99976,
		0.99888,
		0.996737,
		0.992666,
		0.986009,
		0.976169,
		0.96264,
		0.945021,
		0.923032,
		0.896506,
		0.865395,
		0.82976,
		0.78977,
		0.745689,
		0.697877,
		0.646778,
		0.592919,
		0.536902,
		0.479399,
		0.421152,
		0.362964,
		0.3057,
		0.25028,
		0.197681,
		0.14893,
		0.105104,
		0.06733,
		0.03678,
		0.014668,
		0.002256,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		0.004312,
		0.019046,
		0.043031,
		0.075048,
		0.113919,
		0.158507,
		0.207724,
		0.260527,
		0.315921,
		0.372963,
		0.430764,
		0.488488,
		0.545361,
		0.600669,
		0.653763,
		0.704067,
		0.751077,
		0.794366,
		0.833597,
		0.868517,
		0.898975,
		0.924919,
		0.946406,
		0.963605,
		0.976801,
		0.986389,
		0.99287,
		0.99683,
		0.998913,
		0.999767,
		0.999984,
		1,
		0.999984,
		0.999767,
		0.998913,
		0.99683,
		0.99287,
		0.986389,
		0.976801,
		0.963605,
		0.946406,
		0.924919,
		0.898975,
		0.868517,
		0.833597,
		0.794366,
		0.751077,
		0.704067,
		0.653763,
		0.600669,
		0.545361,
		0.488488,
		0.430764,
		0.372963,
		0.315921,
		0.260527,
		0.207724,
		0.158507,
		0.113919,
		0.075048,
		0.043031,
		0.019046,
		0.004312,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.003183,
		0.016673,
		0.039569,
		0.070648,
		0.108724,
		0.152659,
		0.201355,
		0.253765,
		0.308888,
		0.365774,
		0.423526,
		0.481304,
		0.538323,
		0.593863,
		0.647268,
		0.69795,
		0.745396,
		0.789171,
		0.828924,
		0.864394,
		0.895415,
		0.921922,
		0.943959,
		0.961681,
		0.975358,
		0.985372,
		0.99221,
		0.996452,
		0.998733,
		0.999707,
		0.999975,
		1,
		0.999991,
		0.999819,
		0.999074,
		0.99718,
		0.99349,
		0.987358,
		0.978187,
		0.965467,
		0.948785,
		0.927846,
		0.902465,
		0.872571,
		0.838201,
		0.799498,
		0.756699,
		0.710133,
		0.660215,
		0.607439,
		0.552374,
		0.495659,
		0.438,
		0.380165,
		0.32298,
		0.267329,
		0.214149,
		0.164428,
		0.119202,
		0.079556,
		0.046619,
		0.021564,
		0.005605,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.000366,
		0.008808,
		0.027432,
		0.054963,
		0.090168,
		0.131859,
		0.178891,
		0.230171,
		0.28465,
		0.341334,
		0.39928,
		0.457602,
		0.515474,
		0.572131,
		0.626874,
		0.679074,
		0.728175,
		0.773702,
		0.815262,
		0.852555,
		0.885375,
		0.913621,
		0.937299,
		0.95653,
		0.971552,
		0.982721,
		0.990506,
		0.995473,
		0.99826,
		0.999537,
		0.999944,
		1,
		0.999997,
		0.999892,
		0.999323,
		0.997731,
		0.994461,
		0.988848,
		0.980266,
		0.968173,
		0.952128,
		0.931803,
		0.90699,
		0.877597,
		0.843645,
		0.805262,
		0.762678,
		0.716218,
		0.666294,
		0.613403,
		0.55812,
		0.501091,
		0.443032,
		0.384724,
		0.327009,
		0.270786,
		0.217011,
		0.166693,
		0.120891,
		0.080713,
		0.047314,
		0.021892,
		0.005692,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.003284,
		0.01719,
		0.040769,
		0.072738,
		0.111859,
		0.156939,
		0.206833,
		0.260446,
		0.316735,
		0.374712,
		0.433445,
		0.492061,
		0.54975,
		0.60577,
		0.659446,
		0.710181,
		0.757456,
		0.800837,
		0.839981,
		0.874642,
		0.90468,
		0.930061,
		0.950871,
		0.967315,
		0.97972,
		0.988531,
		0.9943,
		0.997663,
		0.999303,
		0.999889,
		0.999997,
		1,
		0.999942,
		0.999523,
		0.998208,
		0.99534,
		0.990235,
		0.98224,
		0.97078,
		0.955381,
		0.935686,
		0.91146,
		0.882588,
		0.849074,
		0.811031,
		0.768681,
		0.722343,
		0.672427,
		0.619433,
		0.563939,
		0.5066,
		0.448143,
		0.38936,
		0.331109,
		0.274307,
		0.219929,
		0.169005,
		0.122616,
		0.081895,
		0.048024,
		0.022228,
		0.005782,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.000378,
		0.009086,
		0.02828,
		0.056622,
		0.092821,
		0.135635,
		0.183866,
		0.236371,
		0.292057,
		0.349886,
		0.40888,
		0.468117,
		0.526742,
		0.583966,
		0.639069,
		0.691409,
		0.740422,
		0.78563,
		0.826649,
		0.863189,
		0.895068,
		0.922214,
		0.944673,
		0.962616,
		0.976338,
		0.986257,
		0.992908,
		0.996922,
		0.998987,
		0.999801,
		0.99999,
		1,
		0.999973,
		0.999679,
		0.998614,
		0.996128,
		0.991519,
		0.984106,
		0.973282,
		0.95854,
		0.939491,
		0.91587,
		0.887541,
		0.854485,
		0.816804,
		0.774707,
		0.728508,
		0.678615,
		0.625529,
		0.569832,
		0.512188,
		0.453333,
		0.394074,
		0.335282,
		0.277895,
		0.222905,
		0.171364,
		0.124377,
		0.083104,
		0.04875,
		0.022572,
		0.005873,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.00339,
		0.017731,
		0.042023,
		0.074922,
		0.11513,
		0.161399,
		0.212531,
		0.267385,
		0.32487,
		0.383959,
		0.443682,
		0.503133,
		0.561477,
		0.617947,
		0.671853,
		0.722587,
		0.769624,
		0.812535,
		0.850988,
		0.884756,
		0.913726,
		0.937904,
		0.957423,
		0.972543,
		0.983655,
		0.991272,
		0.996013,
		0.998572,
		0.999668,
		0.999972,
		1,
		0.999989,
		0.999795,
		0.998956,
		0.996829,
		0.9927,
		0.985864,
		0.975678,
		0.961602,
		0.943213,
		0.920217,
		0.89245,
		0.859876,
		0.822577,
		0.780754,
		0.734711,
		0.684856,
		0.631689,
		0.575799,
		0.517854,
		0.458604,
		0.398867,
		0.339531,
		0.28155,
		0.225939,
		0.173771,
		0.126177,
		0.084339,
		0.049493,
		0.022924,
		0.005967,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.00039,
		0.009378,
		0.029167,
		0.058356,
		0.095593,
		0.139573,
		0.189047,
		0.242817,
		0.299743,
		0.358744,
		0.4188,
		0.478955,
		0.538323,
		0.596089,
		0.651515,
		0.703943,
		0.752804,
		0.79762,
		0.838015,
		0.873716,
		0.904566,
		0.930531,
		0.951701,
		0.968302,
		0.980692,
		0.989361,
		0.994912,
		0.998039,
		0.999477,
		0.999936,
		1,
		0.999997,
		0.999878,
		0.999236,
		0.997444,
		0.99378,
		0.987511,
		0.977965,
		0.964563,
		0.946848,
		0.924495,
		0.897313,
		0.865241,
		0.828347,
		0.786818,
		0.740951,
		0.69115,
		0.637915,
		0.58184,
		0.523601,
		0.463957,
		0.403741,
		0.343856,
		0.285276,
		0.229035,
		0.17623,
		0.128015,
		0.085601,
		0.050253,
		0.023285,
		0.006063,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.003501,
		0.018298,
		0.043337,
		0.077206,
		0.118545,
		0.166048,
		0.218463,
		0.274594,
		0.333306,
		0.393526,
		0.454246,
		0.514529,
		0.573509,
		0.630396,
		0.684484,
		0.735156,
		0.781884,
		0.824244,
		0.861918,
		0.894703,
		0.922519,
		0.945417,
		0.963582,
		0.977338,
		0.987147,
		0.993593,
		0.997366,
		0.999212,
		0.999874,
		0.999997,
		1,
		0.999934,
		0.999461,
		0.997978,
		0.994759,
		0.989048,
		0.98014,
		0.967419,
		0.950392,
		0.9287,
		0.902123,
		0.870578,
		0.834112,
		0.792898,
		0.747226,
		0.697495,
		0.644206,
		0.587956,
		0.529428,
		0.469393,
		0.408697,
		0.34826,
		0.289073,
		0.232193,
		0.178739,
		0.129893,
		0.086892,
		0.05103,
		0.023653,
		0.006161,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.000403,
		0.009684,
		0.030097,
		0.060171,
		0.098489,
		0.143683,
		0.194445,
		0.249522,
		0.307723,
		0.36792,
		0.429052,
		0.490126,
		0.550223,
		0.608503,
		0.664208,
		0.716668,
		0.765307,
		0.809651,
		0.849334,
		0.884104,
		0.913836,
		0.938536,
		0.958348,
		0.973558,
		0.984597,
		0.992029,
		0.996529,
		0.998854,
		0.999775,
		0.999988,
		1,
		0.999969,
		0.999636,
		0.998435,
		0.995638,
		0.990474,
		0.9822,
		0.970166,
		0.95384,
		0.932827,
		0.906878,
		0.875881,
		0.839866,
		0.79899,
		0.753533,
		0.70389,
		0.650561,
		0.594146,
		0.535338,
		0.474914,
		0.413737,
		0.352743,
		0.292943,
		0.235414,
		0.181302,
		0.131812,
		0.088212,
		0.051826,
		0.024031,
		0.006261,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.003617,
		0.018893,
		0.044712,
		0.079594,
		0.122113,
		0.170898,
		0.224639,
		0.282087,
		0.342055,
		0.403426,
		0.465151,
		0.526257,
		0.585848,
		0.643115,
		0.697333,
		0.747875,
		0.794215,
		0.835936,
		0.872739,
		0.904447,
		0.931022,
		0.952563,
		0.969317,
		0.98168,
		0.990188,
		0.995504,
		0.998385,
		0.999624,
		0.999968,
		1,
		0.999988,
		0.999767,
		0.998818,
		0.996422,
		0.991788,
		0.984144,
		0.972801,
		0.957188,
		0.936872,
		0.911571,
		0.881147,
		0.845607,
		0.805092,
		0.759872,
		0.710335,
		0.656981,
		0.600413,
		0.54133,
		0.480522,
		0.418863,
		0.357308,
		0.296887,
		0.238701,
		0.183919,
		0.133774,
		0.089563,
		0.05264,
		0.024418,
		0.006364,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.000417,
		0.010005,
		0.031071,
		0.062072,
		0.101517,
		0.147975,
		0.200072,
		0.256499,
		0.31601,
		0.377429,
		0.439648,
		0.501639,
		0.562448,
		0.621209,
		0.677145,
		0.729573,
		0.777914,
		0.821699,
		0.860575,
		0.894319,
		0.92284,
		0.946192,
		0.964579,
		0.978359,
		0.988039,
		0.994262,
		0.997782,
		0.999407,
		0.999927,
		1,
		0.999997,
		0.999861,
		0.999134,
		0.997111,
		0.992991,
		0.98597,
		0.975321,
		0.960431,
		0.94083,
		0.916199,
		0.886372,
		0.851331,
		0.811201,
		0.76624,
		0.716828,
		0.663465,
		0.606755,
		0.547405,
		0.486216,
		0.424076,
		0.361957,
		0.300909,
		0.242056,
		0.186592,
		0.135779,
		0.090944,
		0.053473,
		0.024814,
		0.00647,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.003739,
		0.019517,
		0.046154,
		0.082095,
		0.125842,
		0.17596,
		0.231074,
		0.289878,
		0.351133,
		0.413672,
		0.476406,
		0.538323,
		0.5985,
		0.656102,
		0.71039,
		0.76073,
		0.806596,
		0.847584,
		0.883417,
		0.913952,
		0.939196,
		0.959306,
		0.974599,
		0.985548,
		0.992774,
		0.99702,
		0.999106,
		0.999857,
		0.999997,
		1,
		0.999925,
		0.999388,
		0.997711,
		0.994084,
		0.987677,
		0.977722,
		0.963565,
		0.944695,
		0.920755,
		0.891549,
		0.857034,
		0.817314,
		0.772634,
		0.723368,
		0.670011,
		0.613173,
		0.553565,
		0.491999,
		0.429378,
		0.366691,
		0.305009,
		0.245479,
		0.189322,
		0.137829,
		0.092357,
		0.054327,
		0.02522,
		0.006578,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.000431,
		0.010342,
		0.032094,
		0.064064,
		0.104687,
		0.152459,
		0.205942,
		0.263763,
		0.324619,
		0.387283,
		0.450601,
		0.513503,
		0.575003,
		0.634207,
		0.690319,
		0.742645,
		0.790604,
		0.833736,
		0.871707,
		0.904322,
		0.931537,
		0.953459,
		0.970362,
		0.982681,
		0.991008,
		0.996073,
		0.9987,
		0.999743,
		0.999987,
		1,
		0.999965,
		0.999586,
		0.998225,
		0.995069,
		0.989263,
		0.980001,
		0.966586,
		0.948463,
		0.925236,
		0.896676,
		0.862711,
		0.823426,
		0.779051,
		0.729952,
		0.67662,
		0.619667,
		0.55981,
		0.497872,
		0.43477,
		0.371512,
		0.309189,
		0.248973,
		0.192112,
		0.139926,
		0.093803,
		0.0552,
		0.025636,
		0.006689,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.003868,
		0.020173,
		0.047667,
		0.084714,
		0.129743,
		0.181245,
		0.237781,
		0.297982,
		0.360554,
		0.424278,
		0.488022,
		0.550737,
		0.611467,
		0.669354,
		0.723646,
		0.773702,
		0.819002,
		0.859156,
		0.893914,
		0.923176,
		0.946999,
		0.965609,
		0.9794,
		0.988931,
		0.994912,
		0.998167,
		0.999572,
		0.999963,
		1,
		0.999986,
		0.999735,
		0.998657,
		0.995947,
		0.990728,
		0.982156,
		0.969489,
		0.952128,
		0.929636,
		0.901745,
		0.868358,
		0.829535,
		0.78549,
		0.736579,
		0.683291,
		0.626236,
		0.56614,
		0.503836,
		0.440255,
		0.376422,
		0.313452,
		0.25254,
		0.194962,
		0.142069,
		0.095283,
		0.056095,
		0.026062,
		0.006803,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.000447,
		0.010696,
		0.033168,
		0.066153,
		0.108006,
		0.157147,
		0.212067,
		0.271328,
		0.333566,
		0.397498,
		0.461923,
		0.525728,
		0.587893,
		0.647496,
		0.703721,
		0.755868,
		0.803355,
		0.845732,
		0.88269,
		0.914074,
		0.939885,
		0.960299,
		0.975665,
		0.986506,
		0.993506,
		0.997481,
		0.999324,
		0.999917,
		1,
		0.999996,
		0.999842,
		0.999014,
		0.996723,
		0.992073,
		0.984185,
		0.972271,
		0.955686,
		0.933948,
		0.906753,
		0.87397,
		0.835636,
		0.791947,
		0.743248,
		0.690022,
		0.632882,
		0.572558,
		0.509893,
		0.445833,
		0.381424,
		0.317799,
		0.256181,
		0.197875,
		0.144262,
		0.096799,
		0.057012,
		0.026499,
		0.00692,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.004003,
		0.020862,
		0.049255,
		0.087461,
		0.133826,
		0.186768,
		0.244776,
		0.306416,
		0.370333,
		0.435259,
		0.500011,
		0.563503,
		0.624748,
		0.682865,
		0.737086,
		0.78677,
		0.831403,
		0.870615,
		0.904191,
		0.932076,
		0.954391,
		0.971437,
		0.983695,
		0.991821,
		0.996615,
		0.998981,
		0.999836,
		0.999996,
		0.999999,
		0.999914,
		0.999302,
		0.997398,
		0.993298,
		0.986085,
		0.974928,
		0.959131,
		0.938168,
		0.911694,
		0.879542,
		0.841725,
		0.798418,
		0.749955,
		0.696813,
		0.639602,
		0.579062,
		0.516043,
		0.451507,
		0.386518,
		0.322233,
		0.2599,
		0.200852,
		0.146505,
		0.09835,
		0.057952,
		0.026947,
		0.00704,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.000462,
		0.011069,
		0.034297,
		0.068345,
		0.111483,
		0.162051,
		0.218463,
		0.27921,
		0.342867,
		0.40809,
		0.473628,
		0.538323,
		0.601121,
		0.661072,
		0.717343,
		0.769224,
		0.816138,
		0.857652,
		0.893487,
		0.923529,
		0.947842,
		0.966673,
		0.980459,
		0.989821,
		0.995539,
		0.998518,
		0.999707,
		0.999985,
		1,
		0.999959,
		0.999527,
		0.997978,
		0.994403,
		0.987855,
		0.977455,
		0.962458,
		0.94229,
		0.916561,
		0.88507,
		0.847798,
		0.804901,
		0.756699,
		0.703661,
		0.646399,
		0.585653,
		0.522288,
		0.457278,
		0.391707,
		0.326755,
		0.263696,
		0.203896,
		0.148801,
		0.099939,
		0.058915,
		0.027406,
		0.007163,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.004145,
		0.021587,
		0.050923,
		0.090341,
		0.138102,
		0.192542,
		0.252073,
		0.315194,
		0.380488,
		0.446628,
		0.512384,
		0.57663,
		0.638345,
		0.696627,
		0.750696,
		0.79991,
		0.843766,
		0.881923,
		0.914202,
		0.940607,
		0.961329,
		0.976757,
		0.987468,
		0.99422,
		0.99791,
		0.99951,
		0.999958,
		1,
		0.999984,
		0.999696,
		0.998468,
		0.995392,
		0.989494,
		0.97985,
		0.965663,
		0.946308,
		0.92135,
		0.890546,
		0.85385,
		0.811392,
		0.763477,
		0.710566,
		0.65327,
		0.592333,
		0.528629,
		0.463149,
		0.396993,
		0.331368,
		0.267574,
		0.207007,
		0.151149,
		0.101566,
		0.059902,
		0.027878,
		0.007289,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.000479,
		0.011462,
		0.035484,
		0.070648,
		0.11513,
		0.167184,
		0.225145,
		0.287427,
		0.352538,
		0.419073,
		0.485726,
		0.551297,
		0.61469,
		0.674931,
		0.73117,
		0.78269,
		0.828924,
		0.869459,
		0.904051,
		0.932642,
		0.955361,
		0.972543,
		0.984721,
		0.992621,
		0.997128,
		0.999227,
		0.999905,
		0.999999,
		0.999996,
		0.999819,
		0.998873,
		0.996266,
		0.991001,
		0.982109,
		0.968741,
		0.950217,
		0.926053,
		0.895967,
		0.859876,
		0.817887,
		0.770286,
		0.717527,
		0.660215,
		0.599101,
		0.535067,
		0.469119,
		0.402378,
		0.336073,
		0.271535,
		0.210189,
		0.153554,
		0.103233,
		0.060913,
		0.028361,
		0.007418,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.004296,
		0.02235,
		0.052677,
		0.093366,
		0.142584,
		0.198581,
		0.259691,
		0.324337,
		0.391034,
		0.4584,
		0.525152,
		0.59012,
		0.652255,
		0.710629,
		0.764456,
		0.813093,
		0.856057,
		0.893035,
		0.9239,
		0.948721,
		0.967772,
		0.981538,
		0.990705,
		0.99614,
		0.998834,
		0.999812,
		0.999996,
		0.999999,
		0.999901,
		0.9992,
		0.99703,
		0.992377,
		0.98423,
		0.971686,
		0.954009,
		0.930665,
		0.901325,
		0.865871,
		0.824381,
		0.777124,
		0.72454,
		0.667234,
		0.605958,
		0.541604,
		0.475193,
		0.407865,
		0.340874,
		0.27558,
		0.213443,
		0.156015,
		0.104941,
		0.061951,
		0.028857,
		0.007551,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.000497,
		0.011876,
		0.036734,
		0.073068,
		0.118957,
		0.172561,
		0.232129,
		0.295997,
		0.362598,
		0.430464,
		0.498233,
		0.564656,
		0.628601,
		0.689066,
		0.745185,
		0.796242,
		0.841678,
		0.881109,
		0.914337,
		0.941362,
		0.962398,
		0.977873,
		0.988431,
		0.994912,
		0.998304,
		0.999663,
		0.999982,
		1,
		0.999953,
		0.999457,
		0.997688,
		0.993622,
		0.98621,
		0.974494,
		0.957681,
		0.935179,
		0.906615,
		0.871829,
		0.83087,
		0.783986,
		0.731605,
		0.674326,
		0.612904,
		0.54824,
		0.48137,
		0.413455,
		0.345773,
		0.279714,
		0.21677,
		0.158534,
		0.106691,
		0.063015,
		0.029367,
		0.007688,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.004454,
		0.023155,
		0.054523,
		0.096544,
		0.147285,
		0.204903,
		0.267647,
		0.333861,
		0.40199,
		0.470589,
		0.538323,
		0.603979,
		0.666474,
		0.72486,
		0.778343,
		0.826287,
		0.868232,
		0.903905,
		0.933236,
		0.956371,
		0.973679,
		0.985756,
		0.993406,
		0.997607,
		0.999437,
		0.999951,
		1,
		0.999982,
		0.999651,
		0.998244,
		0.994738,
		0.988048,
		0.977161,
		0.961225,
		0.93959,
		0.91183,
		0.877744,
		0.83735,
		0.790869,
		0.738718,
		0.681489,
		0.619939,
		0.554977,
		0.487653,
		0.41915,
		0.350772,
		0.283937,
		0.220175,
		0.161115,
		0.108485,
		0.064107,
		0.02989,
		0.007828,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.000516,
		0.012313,
		0.038051,
		0.075614,
		0.122976,
		0.178197,
		0.239433,
		0.304937,
		0.373065,
		0.442279,
		0.51116,
		0.578407,
		0.642853,
		0.703467,
		0.759371,
		0.809849,
		0.854361,
		0.892557,
		0.924291,
		0.94964,
		0.968908,
		0.982633,
		0.991581,
		0.996711,
		0.999112,
		0.99989,
		0.999999,
		0.999995,
		0.999791,
		0.998706,
		0.995728,
		0.989742,
		0.979683,
		0.964635,
		0.943889,
		0.916963,
		0.883611,
		0.843815,
		0.797771,
		0.745878,
		0.688723,
		0.627063,
		0.561815,
		0.494044,
		0.424954,
		0.355873,
		0.288253,
		0.223658,
		0.163757,
		0.110324,
		0.065227,
		0.030427,
		0.007973,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.004622,
		0.024004,
		0.056468,
		0.099885,
		0.152218,
		0.211524,
		0.27596,
		0.343786,
		0.413374,
		0.483212,
		0.551909,
		0.618209,
		0.680995,
		0.739302,
		0.79233,
		0.839455,
		0.880246,
		0.914479,
		0.942154,
		0.963506,
		0.979014,
		0.989394,
		0.995579,
		0.998659,
		0.999783,
		0.999995,
		0.999999,
		0.999886,
		0.999079,
		0.996595,
		0.991293,
		0.982057,
		0.967907,
		0.948071,
		0.922008,
		0.889423,
		0.850259,
		0.804685,
		0.753082,
		0.696026,
		0.634275,
		0.568755,
		0.500544,
		0.430867,
		0.361079,
		0.292664,
		0.227222,
		0.166465,
		0.11221,
		0.066377,
		0.030978,
		0.008121,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.000536,
		0.012775,
		0.039439,
		0.078295,
		0.127199,
		0.184108,
		0.247077,
		0.31427,
		0.383959,
		0.454537,
		0.524519,
		0.592557,
		0.657442,
		0.71812,
		0.773702,
		0.823476,
		0.866927,
		0.903749,
		0.93386,
		0.957423,
		0.974848,
		0.986798,
		0.994171,
		0.998049,
		0.999611,
		0.99998,
		1,
		0.999946,
		0.999373,
		0.997343,
		0.9927,
		0.984278,
		0.971035,
		0.952128,
		0.926957,
		0.895174,
		0.856678,
		0.811609,
		0.760326,
		0.703397,
		0.641577,
		0.575799,
		0.507156,
		0.436892,
		0.366393,
		0.297172,
		0.230869,
		0.169238,
		0.114144,
		0.067557,
		0.031545,
		0.008273,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.004799,
		0.0249,
		0.058518,
		0.103402,
		0.1574,
		0.218463,
		0.28465,
		0.354133,
		0.425205,
		0.496282,
		0.565918,
		0.63281,
		0.69581,
		0.753936,
		0.806386,
		0.852555,
		0.892049,
		0.924704,
		0.950601,
		0.970081,
		0.983744,
		0.992444,
		0.997247,
		0.99935,
		0.999944,
		1,
		0.999979,
		0.999596,
		0.997978,
		0.993965,
		0.986346,
		0.974012,
		0.956054,
		0.931803,
		0.900855,
		0.863065,
		0.818536,
		0.767608,
		0.710833,
		0.648967,
		0.582946,
		0.51388,
		0.443032,
		0.371816,
		0.301781,
		0.234603,
		0.172081,
		0.116129,
		0.068768,
		0.032127,
		0.00843,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.000557,
		0.013263,
		0.040905,
		0.081119,
		0.131641,
		0.190312,
		0.255081,
		0.324015,
		0.395301,
		0.467253,
		0.538323,
		0.607109,
		0.672365,
		0.73301,
		0.78815,
		0.837084,
		0.879329,
		0.91463,
		0.942986,
		0.964657,
		0.980179,
		0.990352,
		0.996216,
		0.998974,
		0.999873,
		0.999999,
		0.999994,
		0.999758,
		0.998506,
		0.99509,
		0.988257,
		0.976835,
		0.959841,
		0.936538,
		0.90646,
		0.869413,
		0.825463,
		0.774923,
		0.718333,
		0.656444,
		0.590198,
		0.520718,
		0.449289,
		0.377353,
		0.306492,
		0.238425,
		0.174994,
		0.118164,
		0.070013,
		0.032726,
		0.008592,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.004986,
		0.025848,
		0.060681,
		0.107106,
		0.162846,
		0.22574,
		0.29374,
		0.364925,
		0.437501,
		0.509814,
		0.580358,
		0.647781,
		0.710906,
		0.768736,
		0.820474,
		0.865538,
		0.903584,
		0.934518,
		0.958519,
		0.976047,
		0.987845,
		0.994912,
		0.998451,
		0.999749,
		0.999994,
		0.999999,
		0.999868,
		0.998935,
		0.996078,
		0.990011,
		0.979499,
		0.963484,
		0.941155,
		0.911982,
		0.875716,
		0.832383,
		0.782268,
		0.725895,
		0.664008,
		0.597555,
		0.527672,
		0.455665,
		0.383005,
		0.31131,
		0.242338,
		0.177981,
		0.120254,
		0.071291,
		0.033341,
		0.008758,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.000579,
		0.013779,
		0.042454,
		0.084099,
		0.136317,
		0.196828,
		0.263466,
		0.334196,
		0.407111,
		0.480445,
		0.552583,
		0.622067,
		0.687611,
		0.748116,
		0.802682,
		0.850628,
		0.89151,
		0.92514,
		0.951607,
		0.971292,
		0.984868,
		0.993291,
		0.997745,
		0.999548,
		0.999976,
		1,
		0.999937,
		0.999274,
		0.996932,
		0.991607,
		0.981999,
		0.966975,
		0.945646,
		0.917412,
		0.881966,
		0.839291,
		0.789639,
		0.733515,
		0.671658,
		0.605018,
		0.534743,
		0.462163,
		0.388776,
		0.316237,
		0.246346,
		0.181043,
		0.122398,
		0.072605,
		0.033974,
		0.008928,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.005185,
		0.02685,
		0.062966,
		0.11101,
		0.168575,
		0.233376,
		0.303253,
		0.376183,
		0.450283,
		0.523824,
		0.595234,
		0.663118,
		0.726267,
		0.783673,
		0.834551,
		0.878352,
		0.91479,
		0.943859,
		0.965851,
		0.981366,
		0.991302,
		0.996818,
		0.999246,
		0.999934,
		1,
		0.999975,
		0.999531,
		0.99766,
		0.993045,
		0.984332,
		0.970307,
		0.950003,
		0.922742,
		0.888157,
		0.84618,
		0.797031,
		0.741192,
		0.679392,
		0.612586,
		0.541933,
		0.468785,
		0.394668,
		0.321275,
		0.25045,
		0.184184,
		0.1246,
		0.073955,
		0.034625,
		0.009104,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.000602,
		0.014327,
		0.044093,
		0.087245,
		0.141244,
		0.203676,
		0.272257,
		0.344836,
		0.419412,
		0.494131,
		0.567308,
		0.637429,
		0.70317,
		0.763413,
		0.81726,
		0.864056,
		0.903409,
		0.935212,
		0.959663,
		0.977279,
		0.988892,
		0.995624,
		0.998808,
		0.999852,
		0.999999,
		0.999993,
		0.999718,
		0.998267,
		0.994328,
		0.986495,
		0.973475,
		0.954217,
		0.927963,
		0.894279,
		0.853044,
		0.804438,
		0.748921,
		0.68721,
		0.620261,
		0.549244,
		0.475533,
		0.400685,
		0.326429,
		0.254655,
		0.187405,
		0.126861,
		0.075343,
		0.035295,
		0.009285,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.005396,
		0.027912,
		0.065382,
		0.11513,
		0.174607,
		0.241394,
		0.313213,
		0.387931,
		0.463572,
		0.538323,
		0.610552,
		0.678813,
		0.741872,
		0.798713,
		0.848568,
		0.890936,
		0.925601,
		0.952661,
		0.972543,
		0.986004,
		0.994115,
		0.9982,
		0.999707,
		0.999993,
		0.999999,
		0.999846,
		0.998762,
		0.995458,
		0.988485,
		0.976473,
		0.95828,
		0.933068,
		0.900326,
		0.859876,
		0.811856,
		0.756699,
		0.695109,
		0.628041,
		0.556677,
		0.48241,
		0.406828,
		0.331701,
		0.258963,
		0.19071,
		0.129184,
		0.07677,
		0.035984,
		0.009472,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.000627,
		0.014907,
		0.045828,
		0.090569,
		0.146439,
		0.21088,
		0.281477,
		0.355962,
		0.432226,
		0.508329,
		0.582507,
		0.653193,
		0.719025,
		0.778869,
		0.831837,
		0.877309,
		0.91496,
		0.944778,
		0.967092,
		0.982575,
		0.992239,
		0.99738,
		0.999473,
		0.999972,
		1,
		0.999926,
		0.999153,
		0.996439,
		0.990301,
		0.979295,
		0.962186,
		0.938045,
		0.906287,
		0.866667,
		0.819279,
		0.764523,
		0.703088,
		0.635927,
		0.564232,
		0.489418,
		0.413102,
		0.337094,
		0.263377,
		0.194101,
		0.13157,
		0.078237,
		0.036694,
		0.009664,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.00562,
		0.029038,
		0.067939,
		0.119481,
		0.180962,
		0.249819,
		0.323646,
		0.400195,
		0.477387,
		0.553326,
		0.626313,
		0.694856,
		0.757694,
		0.813813,
		0.862471,
		0.903223,
		0.935944,
		0.960857,
		0.978541,
		0.989936,
		0.996301,
		0.999119,
		0.999923,
		1,
		0.999971,
		0.999452,
		0.997277,
		0.991944,
		0.981935,
		0.965925,
		0.942888,
		0.912153,
		0.873411,
		0.826699,
		0.772387,
		0.711144,
		0.643919,
		0.571912,
		0.496559,
		0.41951,
		0.342613,
		0.267901,
		0.197581,
		0.134022,
		0.079747,
		0.037425,
		0.009862,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.000654,
		0.015523,
		0.047667,
		0.094085,
		0.151921,
		0.218463,
		0.291153,
		0.367599,
		0.445578,
		0.523054,
		0.598188,
		0.669354,
		0.735156,
		0.794448,
		0.846361,
		0.890323,
		0.926089,
		0.953766,
		0.973833,
		0.987147,
		0.994912,
		0.998608,
		0.999826,
		0.999999,
		0.999992,
		0.999669,
		0.997978,
		0.993413,
		0.984391,
		0.969489,
		0.947585,
		0.917916,
		0.880097,
		0.834112,
		0.780288,
		0.719275,
		0.652015,
		0.579717,
		0.503836,
		0.426054,
		0.34826,
		0.272538,
		0.201154,
		0.136543,
		0.081301,
		0.038178,
		0.010067,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.005858,
		0.030233,
		0.070648,
		0.124081,
		0.187663,
		0.258679,
		0.334582,
		0.413,
		0.491749,
		0.568845,
		0.642517,
		0.71123,
		0.773702,
		0.828924,
		0.876195,
		0.91514,
		0.945745,
		0.968379,
		0.983803,
		0.993157,
		0.997897,
		0.999656,
		0.999992,
		0.999999,
		0.999819,
		0.998551,
		0.994712,
		0.986658,
		0.972873,
		0.952128,
		0.923565,
		0.886718,
		0.841508,
		0.78822,
		0.727478,
		0.660215,
		0.587648,
		0.511251,
		0.432738,
		0.354039,
		0.277292,
		0.204822,
		0.139134,
		0.0829,
		0.038954,
		0.010277,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.000682,
		0.016179,
		0.049618,
		0.097809,
		0.157713,
		0.226451,
		0.301315,
		0.379775,
		0.45949,
		0.538323,
		0.614355,
		0.685901,
		0.751535,
		0.810106,
		0.860772,
		0.903024,
		0.936718,
		0.962103,
		0.979834,
		0.990973,
		0.996937,
		0.999381,
		0.999967,
		1,
		0.999913,
		0.999006,
		0.995843,
		0.988734,
		0.976067,
		0.956507,
		0.92909,
		0.893264,
		0.84888,
		0.796177,
		0.735749,
		0.668518,
		0.595706,
		0.518807,
		0.439564,
		0.359954,
		0.282166,
		0.208589,
		0.141799,
		0.084547,
		0.039753,
		0.010495,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.006111,
		0.031504,
		0.073521,
		0.128949,
		0.194735,
		0.268,
		0.346048,
		0.426374,
		0.50668,
		0.584888,
		0.65916,
		0.727916,
		0.789856,
		0.843991,
		0.889668,
		0.926607,
		0.954925,
		0.975164,
		0.988294,
		0.995675,
		0.998965,
		0.999909,
		1,
		0.999966,
		0.999355,
		0.996812,
		0.990617,
		0.979066,
		0.960712,
		0.934481,
		0.899725,
		0.85622,
		0.804153,
		0.744084,
		0.676922,
		0.603892,
		0.526505,
		0.446537,
		0.366009,
		0.287165,
		0.212459,
		0.14454,
		0.086243,
		0.040578,
		0.010719,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.000713,
		0.016877,
		0.051692,
		0.101756,
		0.163836,
		0.234873,
		0.311993,
		0.392521,
		0.473988,
		0.554152,
		0.63101,
		0.702821,
		0.768129,
		0.82579,
		0.875,
		0.915332,
		0.946765,
		0.969716,
		0.985047,
		0.99405,
		0.998364,
		0.999794,
		0.999999,
		0.99999,
		0.99961,
		0.997627,
		0.992308,
		0.981865,
		0.964734,
		0.939726,
		0.906089,
		0.863519,
		0.812141,
		0.75248,
		0.685426,
		0.612205,
		0.534347,
		0.45366,
		0.372208,
		0.292293,
		0.216434,
		0.14736,
		0.08799,
		0.041428,
		0.010951,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.006382,
		0.032856,
		0.076573,
		0.134105,
		0.202206,
		0.277816,
		0.358077,
		0.440344,
		0.522198,
		0.601465,
		0.676234,
		0.744885,
		0.80611,
		0.858946,
		0.902812,
		0.937538,
		0.963404,
		0.981155,
		0.991996,
		0.997527,
		0.999593,
		0.99999,
		0.999999,
		0.999785,
		0.998294,
		0.993807,
		0.984457,
		0.968563,
		0.944814,
		0.912347,
		0.870768,
		0.820135,
		0.760931,
		0.694027,
		0.620647,
		0.542336,
		0.460936,
		0.378555,
		0.297552,
		0.220519,
		0.150262,
		0.089791,
		0.042305,
		0.01119,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.000745,
		0.017621,
		0.053897,
		0.105945,
		0.170317,
		0.243759,
		0.32322,
		0.405867,
		0.489096,
		0.570554,
		0.648151,
		0.720092,
		0.784898,
		0.841438,
		0.888967,
		0.927158,
		0.956143,
		0.976536,
		0.98944,
		0.996398,
		0.999268,
		0.999961,
		1,
		0.999897,
		0.998827,
		0.995118,
		0.986838,
		0.972191,
		0.949734,
		0.918486,
		0.877955,
		0.828126,
		0.769433,
		0.702724,
		0.629217,
		0.550474,
		0.468368,
		0.385053,
		0.302949,
		0.224718,
		0.15325,
		0.091646,
		0.043211,
		0.011437,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.006671,
		0.034297,
		0.079817,
		0.139573,
		0.210104,
		0.288158,
		0.370703,
		0.45494,
		0.538323,
		0.618579,
		0.693726,
		0.762104,
		0.822408,
		0.873716,
		0.915536,
		0.947842,
		0.971103,
		0.986304,
		0.994912,
		0.998776,
		0.999892,
		1,
		0.999959,
		0.999236,
		0.996245,
		0.989007,
		0.975609,
		0.954475,
		0.924495,
		0.88507,
		0.836105,
		0.77798,
		0.711512,
		0.637915,
		0.558763,
		0.475959,
		0.391707,
		0.308487,
		0.229035,
		0.156326,
		0.09356,
		0.044146,
		0.011692,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.00078,
		0.018415,
		0.056247,
		0.110396,
		0.177183,
		0.253142,
		0.33503,
		0.419845,
		0.504839,
		0.58754,
		0.665773,
		0.737688,
		0.801791,
		0.85698,
		0.902584,
		0.938407,
		0.964764,
		0.982503,
		0.992999,
		0.998064,
		0.999755,
		0.999998,
		0.999989,
		0.999536,
		0.997195,
		0.990961,
		0.97881,
		0.959025,
		0.930361,
		0.892101,
		0.844063,
		0.786565,
		0.72039,
		0.64674,
		0.567204,
		0.483714,
		0.398522,
		0.31417,
		0.233473,
		0.159495,
		0.095534,
		0.045111,
		0.011957,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.006979,
		0.035835,
		0.08327,
		0.145377,
		0.218463,
		0.299063,
		0.383959,
		0.470189,
		0.555074,
		0.636233,
		0.711615,
		0.77953,
		0.838682,
		0.888213,
		0.927745,
		0.957423,
		0.977948,
		0.990579,
		0.997073,
		0.999515,
		0.999988,
		0.999998,
		0.999744,
		0.997978,
		0.9927,
		0.981785,
		0.963372,
		0.93607,
		0.899037,
		0.85199,
		0.795181,
		0.729352,
		0.655692,
		0.575799,
		0.491635,
		0.405501,
		0.320005,
		0.238038,
		0.162759,
		0.09757,
		0.046108,
		0.01223,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.000817,
		0.019264,
		0.058753,
		0.11513,
		0.184465,
		0.263058,
		0.347461,
		0.434489,
		0.52124,
		0.605119,
		0.683863,
		0.755573,
		0.818748,
		0.872333,
		0.915755,
		0.94898,
		0.972543,
		0.987569,
		0.995735,
		0.999128,
		0.999953,
		1,
		0.999876,
		0.998605,
		0.994227,
		0.98453,
		0.967506,
		0.94161,
		0.905864,
		0.859876,
		0.80382,
		0.738394,
		0.664769,
		0.584549,
		0.499726,
		0.41265,
		0.325995,
		0.242735,
		0.166123,
		0.099671,
		0.047138,
		0.012512,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.00731,
		0.037478,
		0.086952,
		0.151546,
		0.227317,
		0.31057,
		0.397884,
		0.486123,
		0.572465,
		0.654422,
		0.729875,
		0.797109,
		0.854856,
		0.902341,
		0.939331,
		0.966185,
		0.983876,
		0.993974,
		0.998541,
		0.999871,
		1,
		0.999951,
		0.999088,
		0.995546,
		0.987038,
		0.971415,
		0.946968,
		0.912568,
		0.867708,
		0.812474,
		0.747512,
		0.673969,
		0.593455,
		0.50799,
		0.419973,
		0.332145,
		0.247567,
		0.16959,
		0.101841,
		0.048204,
		0.012805,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.000857,
		0.020173,
		0.061429,
		0.120172,
		0.192195,
		0.273547,
		0.360554,
		0.449834,
		0.538323,
		0.623297,
		0.702404,
		0.773702,
		0.835698,
		0.887401,
		0.928371,
		0.958769,
		0.9794,
		0.991705,
		0.997692,
		0.999707,
		0.999998,
		0.999986,
		0.999445,
		0.996662,
		0.989308,
		0.975089,
		0.952128,
		0.919137,
		0.875474,
		0.821134,
		0.756699,
		0.683291,
		0.602518,
		0.51643,
		0.427474,
		0.338462,
		0.25254,
		0.173165,
		0.104082,
		0.049305,
		0.013107,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.007665,
		0.039236,
		0.09088,
		0.15811,
		0.236704,
		0.32272,
		0.412516,
		0.502771,
		0.590511,
		0.673139,
		0.74847,
		0.814777,
		0.870838,
		0.91599,
		0.950184,
		0.974035,
		0.988837,
		0.996509,
		0.999419,
		0.999986,
		0.999998,
		0.999693,
		0.997585,
		0.991337,
		0.978519,
		0.957078,
		0.925553,
		0.883162,
		0.829789,
		0.76595,
		0.692731,
		0.61174,
		0.52505,
		0.435159,
		0.34495,
		0.257659,
		0.176852,
		0.106396,
		0.050445,
		0.013421,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.000899,
		0.021147,
		0.064291,
		0.12555,
		0.200411,
		0.28465,
		0.374349,
		0.465916,
		0.556109,
		0.642073,
		0.721369,
		0.792019,
		0.852555,
		0.902078,
		0.940314,
		0.967671,
		0.985268,
		0.994912,
		0.998953,
		0.999944,
		1,
		0.999851,
		0.998327,
		0.993125,
		0.981696,
		0.961803,
		0.931803,
		0.890756,
		0.838428,
		0.775256,
		0.702287,
		0.62112,
		0.533853,
		0.443032,
		0.351615,
		0.262931,
		0.180656,
		0.108788,
		0.051625,
		0.013746,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.008046,
		0.041121,
		0.09508,
		0.165103,
		0.246666,
		0.335557,
		0.427896,
		0.520162,
		0.609221,
		0.692366,
		0.767353,
		0.832457,
		0.886525,
		0.929041,
		0.960187,
		0.980889,
		0.992809,
		0.998248,
		0.999844,
		1,
		0.999941,
		0.998903,
		0.994676,
		0.984612,
		0.966289,
		0.93787,
		0.898242,
		0.84704,
		0.784611,
		0.711954,
		0.630658,
		0.542841,
		0.451099,
		0.358463,
		0.268359,
		0.184582,
		0.111261,
		0.052846,
		0.014083,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.000945,
		0.022194,
		0.067358,
		0.131291,
		0.209152,
		0.296414,
		0.388893,
		0.482769,
		0.574618,
		0.661443,
		0.740722,
		0.810454,
		0.869219,
		0.916242,
		0.951461,
		0.975582,
		0.990101,
		0.997227,
		0.999645,
		0.999998,
		0.999983,
		0.99933,
		0.995996,
		0.987261,
		0.970524,
		0.943738,
		0.905603,
		0.85561,
		0.794004,
		0.721729,
		0.640354,
		0.552017,
		0.459363,
		0.365499,
		0.273952,
		0.188634,
		0.113819,
		0.054111,
		0.014433,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.008457,
		0.043145,
		0.099574,
		0.172561,
		0.25725,
		0.349131,
		0.444067,
		0.538323,
		0.628601,
		0.712079,
		0.786465,
		0.850054,
		0.901795,
		0.941362,
		0.969224,
		0.986675,
		0.995804,
		0.999296,
		0.999982,
		0.999998,
		0.999627,
		0.997092,
		0.989641,
		0.974494,
		0.949389,
		0.912824,
		0.864126,
		0.803427,
		0.731605,
		0.650207,
		0.561384,
		0.467831,
		0.372731,
		0.279714,
		0.192819,
		0.116464,
		0.055422,
		0.014796,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.000995,
		0.023321,
		0.070648,
		0.137432,
		0.218463,
		0.308888,
		0.404232,
		0.500431,
		0.593863,
		0.681392,
		0.760415,
		0.828924,
		0.885575,
		0.929759,
		0.961681,
		0.982413,
		0.993882,
		0.998733,
		0.999931,
		1,
		0.999819,
		0.997978,
		0.991748,
		0.978187,
		0.954807,
		0.919885,
		0.872571,
		0.812868,
		0.741577,
		0.660215,
		0.570945,
		0.476508,
		0.380165,
		0.285652,
		0.197141,
		0.119202,
		0.056781,
		0.015172,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.0089,
		0.045322,
		0.104393,
		0.180527,
		0.268504,
		0.363491,
		0.461072,
		0.557281,
		0.648648,
		0.732241,
		0.805731,
		0.867459,
		0.916514,
		0.952816,
		0.977183,
		0.991352,
		0.997877,
		0.999809,
		1,
		0.999928,
		0.998669,
		0.993585,
		0.981594,
		0.959974,
		0.926769,
		0.880929,
		0.822315,
		0.751638,
		0.670377,
		0.580701,
		0.485397,
		0.387806,
		0.291774,
		0.201607,
		0.122037,
		0.05819,
		0.015563,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.001049,
		0.024536,
		0.074184,
		0.144007,
		0.228392,
		0.322126,
		0.420418,
		0.518938,
		0.613858,
		0.701897,
		0.780384,
		0.847326,
		0.901488,
		0.942482,
		0.970847,
		0.988091,
		0.996638,
		0.999567,
		0.999997,
		0.99998,
		0.999183,
		0.995156,
		0.984704,
		0.964874,
		0.933455,
		0.889182,
		0.831754,
		0.761779,
		0.68069,
		0.590655,
		0.494505,
		0.395662,
		0.298086,
		0.206223,
		0.124973,
		0.059653,
		0.01597,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.009378,
		0.047667,
		0.109566,
		0.189047,
		0.280485,
		0.378694,
		0.478955,
		0.577059,
		0.669354,
		0.752804,
		0.82506,
		0.884543,
		0.930531,
		0.963257,
		0.983968,
		0.994912,
		0.99914,
		0.999978,
		0.999997,
		0.999544,
		0.996468,
		0.987511,
		0.969489,
		0.939924,
		0.897313,
		0.841172,
		0.771993,
		0.69115,
		0.600807,
		0.503836,
		0.403741,
		0.304595,
		0.210995,
		0.128015,
		0.06117,
		0.016392,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.001107,
		0.025848,
		0.07799,
		0.151058,
		0.238994,
		0.336187,
		0.437501,
		0.538323,
		0.634605,
		0.722924,
		0.800552,
		0.865538,
		0.916807,
		0.954256,
		0.978837,
		0.992579,
		0.998451,
		0.999916,
		1,
		0.999777,
		0.997533,
		0.990011,
		0.973805,
		0.946156,
		0.905299,
		0.850552,
		0.782268,
		0.701753,
		0.611159,
		0.513395,
		0.412048,
		0.31131,
		0.215931,
		0.131168,
		0.062746,
		0.016832,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.009896,
		0.050198,
		0.11513,
		0.19817,
		0.293251,
		0.394798,
		0.497762,
		0.597675,
		0.690698,
		0.773702,
		0.84434,
		0.901155,
		0.943681,
		0.972543,
		0.989507,
		0.997403,
		0.999765,
		1,
		0.999911,
		0.998368,
		0.992201,
		0.977805,
		0.952128,
		0.913121,
		0.859876,
		0.792593,
		0.712494,
		0.621711,
		0.523188,
		0.420591,
		0.318238,
		0.221038,
		0.134438,
		0.064384,
		0.017289,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.00117,
		0.027267,
		0.082095,
		0.158632,
		0.250328,
		0.351133,
		0.455537,
		0.558619,
		0.656102,
		0.744423,
		0.820817,
		0.883417,
		0.931363,
		0.964919,
		0.985548,
		0.995886,
		0.999465,
		0.999997,
		0.999975,
		0.998994,
		0.994084,
		0.981477,
		0.957819,
		0.920755,
		0.869125,
		0.802956,
		0.723368,
		0.632464,
		0.533218,
		0.429378,
		0.325389,
		0.226323,
		0.137829,
		0.066086,
		0.017765,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.010458,
		0.052935,
		0.121124,
		0.207954,
		0.306869,
		0.411865,
		0.517538,
		0.61914,
		0.712649,
		0.794849,
		0.863435,
		0.917126,
		0.955789,
		0.980542,
		0.993771,
		0.998938,
		0.999973,
		0.999996,
		0.999435,
		0.995666,
		0.98481,
		0.963209,
		0.928179,
		0.87828,
		0.81334,
		0.734365,
		0.643416,
		0.54349,
		0.438416,
		0.332769,
		0.231794,
		0.141349,
		0.067857,
		0.018262,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.001239,
		0.028807,
		0.086531,
		0.166781,
		0.262459,
		0.367032,
		0.474581,
		0.579853,
		0.678333,
		0.766326,
		0.841057,
		0.900792,
		0.944967,
		0.974313,
		0.990912,
		0.998086,
		0.999895,
		1,
		0.999723,
		0.996959,
		0.987794,
		0.968276,
		0.935365,
		0.887317,
		0.823732,
		0.745479,
		0.654566,
		0.554008,
		0.447714,
		0.34039,
		0.237461,
		0.145004,
		0.069699,
		0.018779,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.011069,
		0.055902,
		0.127594,
		0.218463,
		0.321409,
		0.42996,
		0.538323,
		0.641458,
		0.735156,
		0.816138,
		0.882183,
		0.932261,
		0.966673,
		0.987147,
		0.99679,
		0.999707,
		1,
		0.999889,
		0.997978,
		0.990425,
		0.972999,
		0.94229,
		0.896212,
		0.834112,
		0.756699,
		0.665912,
		0.564777,
		0.457278,
		0.34826,
		0.243331,
		0.148801,
		0.071616,
		0.019319,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.001314,
		0.030481,
		0.091333,
		0.175563,
		0.27546,
		0.383959,
		0.494689,
		0.602046,
		0.701268,
		0.788542,
		0.861121,
		0.917472,
		0.957423,
		0.982294,
		0.994912,
		0.999333,
		0.999996,
		0.999968,
		0.998747,
		0.9927,
		0.97736,
		0.948926,
		0.90494,
		0.84446,
		0.768013,
		0.677451,
		0.575799,
		0.467118,
		0.356389,
		0.249415,
		0.152746,
		0.073614,
		0.019882,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.011736,
		0.059124,
		0.134589,
		0.229766,
		0.336951,
		0.449152,
		0.560159,
		0.664622,
		0.758149,
		0.837433,
		0.900395,
		0.946351,
		0.976158,
		0.992293,
		0.998674,
		0.999966,
		0.999995,
		0.999293,
		0.994625,
		0.981342,
		0.955247,
		0.913472,
		0.854755,
		0.779408,
		0.689177,
		0.587077,
		0.47724,
		0.364787,
		0.255722,
		0.156848,
		0.075696,
		0.02047,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.001396,
		0.032305,
		0.096544,
		0.185043,
		0.289411,
		0.40199,
		0.515919,
		0.625212,
		0.72486,
		0.810953,
		0.880827,
		0.933236,
		0.968525,
		0.988753,
		0.997607,
		0.999867,
		1,
		0.999651,
		0.996208,
		0.984931,
		0.961225,
		0.921781,
		0.864974,
		0.790869,
		0.701086,
		0.598615,
		0.487653,
		0.373465,
		0.262263,
		0.161115,
		0.077867,
		0.021084,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.012464,
		0.062632,
		0.142168,
		0.241943,
		0.353578,
		0.469512,
		0.583079,
		0.688606,
		0.781532,
		0.858565,
		0.91785,
		0.959167,
		0.984089,
		0.995985,
		0.999629,
		1,
		0.999859,
		0.997465,
		0.988116,
		0.966832,
		0.929834,
		0.875088,
		0.802379,
		0.71317,
		0.610413,
		0.498366,
		0.382435,
		0.269049,
		0.165555,
		0.080132,
		0.021727,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.001486,
		0.034297,
		0.102209,
		0.195295,
		0.304401,
		0.42121,
		0.538323,
		0.649351,
		0.749038,
		0.833411,
		0.899959,
		0.947842,
		0.978079,
		0.993633,
		0.999156,
		0.999994,
		0.999959,
		0.998419,
		0.99089,
		0.972044,
		0.937598,
		0.88507,
		0.813918,
		0.72542,
		0.622473,
		0.509384,
		0.391707,
		0.276092,
		0.170178,
		0.082496,
		0.022399,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.013263,
		0.06646,
		0.150396,
		0.255081,
		0.371384,
		0.491113,
		0.607109,
		0.713365,
		0.805178,
		0.879328,
		0.934294,
		0.970479,
		0.990351,
		0.99832,
		0.999955,
		0.999993,
		0.999102,
		0.993251,
		0.976834,
		0.94504,
		0.894886,
		0.825462,
		0.737825,
		0.634793,
		0.520717,
		0.401293,
		0.283404,
		0.174994,
		0.084966,
		0.023103,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.001584,
		0.036478,
		0.108382,
		0.2064,
		0.320521,
		0.441699,
		0.561943,
		0.674436,
		0.773688,
		0.85571,
		0.918247,
		0.961014,
		0.985898,
		0.996952,
		0.999812,
		0.999982,
		0.999536,
		0.995193,
		0.981167,
		0.952111,
		0.904491,
		0.836975,
		0.750362,
		0.647364,
		0.532363,
		0.411199,
		0.290993,
		0.180011,
		0.087547,
		0.023841,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.014139,
		0.07064,
		0.159331,
		0.269249,
		0.390425,
		0.513969,
		0.632191,
		0.73875,
		0.82883,
		0.899376,
		0.949345,
		0.979961,
		0.994799,
		0.999412,
		0.999887,
		0.999705,
		0.996665,
		0.98496,
		0.958714,
		0.91379,
		0.848375,
		0.762966,
		0.66014,
		0.544295,
		0.421411,
		0.298855,
		0.185227,
		0.090239,
		0.024612,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.001693,
		0.038856,
		0.115075,
		0.218359,
		0.337737,
		0.463351,
		0.586569,
		0.700132,
		0.798332,
		0.877246,
		0.935005,
		0.972079,
		0.991455,
		0.998441,
		0.999516,
		0.999471,
		0.997502,
		0.988014,
		0.96463,
		0.922566,
		0.859466,
		0.77547,
		0.672987,
		0.556411,
		0.431857,
		0.306942,
		0.190619,
		0.093032,
		0.025414,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.015085,
		0.075127,
		0.168853,
		0.284216,
		0.410314,
		0.537502,
		0.657532,
		0.763735,
		0.851254,
		0.917318,
		0.961557,
		0.986257,
		0.996322,
		0.998418,
		0.998467,
		0.997321,
		0.989904,
		0.969415,
		0.930382,
		0.86984,
		0.787513,
		0.685602,
		0.56847,
		0.442357,
		0.315133,
		0.196111,
		0.09589,
		0.026239,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.001807,
		0.041348,
		0.122028,
		0.230652,
		0.355202,
		0.484947,
		0.610585,
		0.724419,
		0.820585,
		0.895339,
		0.947385,
		0.978195,
		0.992111,
		0.99577,
		0.99599,
		0.995414,
		0.98988,
		0.972293,
		0.936471,
		0.878774,
		0.798444,
		0.697422,
		0.580011,
		0.452559,
		0.323181,
		0.201554,
		0.098742,
		0.027066,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.016032,
		0.079568,
		0.178148,
		0.298578,
		0.428982,
		0.558942,
		0.679669,
		0.784235,
		0.867848,
		0.928207,
		0.96586,
		0.984431,
		0.990301,
		0.990914,
		0.990679,
		0.986795,
		0.972083,
		0.939661,
		0.885148,
		0.807238,
		0.707552,
		0.590294,
		0.461886,
		0.330677,
		0.206694,
		0.101464,
		0.027862,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.001912,
		0.043618,
		0.128242,
		0.241376,
		0.369959,
		0.502414,
		0.628819,
		0.741127,
		0.833445,
		0.902391,
		0.947489,
		0.971493,
		0.980301,
		0.981678,
		0.98162,
		0.979108,
		0.967195,
		0.938357,
		0.887425,
		0.812471,
		0.714734,
		0.59826,
		0.46951,
		0.337025,
		0.211158,
		0.103873,
		0.028577,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.016784,
		0.082994,
		0.185047,
		0.308696,
		0.441194,
		0.571468,
		0.690328,
		0.790745,
		0.868202,
		0.921114,
		0.951225,
		0.963724,
		0.966371,
		0.966433,
		0.96498,
		0.955728,
		0.930634,
		0.883723,
		0.812377,
		0.717384,
		0.602567,
		0.474364,
		0.341453,
		0.21446,
		0.105728,
		0.029144,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.001978,
		0.04496,
		0.131646,
		0.246645,
		0.37608,
		0.507751,
		0.63134,
		0.73868,
		0.824096,
		0.884835,
		0.921517,
		0.938409,
		0.942932,
		0.943208,
		0.942488,
		0.935677,
		0.914431,
		0.871991,
		0.805002,
		0.713725,
		0.601687,
		0.475222,
		0.343066,
		0.216028,
		0.106744,
		0.029484,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.017013,
		0.083785,
		0.185967,
		0.308649,
		0.438588,
		0.564404,
		0.676837,
		0.769076,
		0.837182,
		0.88057,
		0.902408,
		0.909445,
		0.910193,
		0.909906,
		0.905228,
		0.887842,
		0.850289,
		0.788449,
		0.701997,
		0.594078,
		0.470829,
		0.340931,
		0.215263,
		0.10662,
		0.029512,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.001962,
		0.044403,
		0.12944,
		0.241294,
		0.365832,
		0.49074,
		0.605771,
		0.703054,
		0.777497,
		0.827286,
		0.854357,
		0.864494,
		0.866109,
		0.866039,
		0.863107,
		0.849464,
		0.817112,
		0.761188,
		0.680727,
		0.578414,
		0.460076,
		0.334215,
		0.211619,
		0.105077,
		0.029149,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.01635,
		0.080172,
		0.177071,
		0.29224,
		0.412633,
		0.527188,
		0.627109,
		0.706256,
		0.761617,
		0.793834,
		0.807497,
		0.810449,
		0.81052,
		0.808892,
		0.798725,
		0.771729,
		0.722367,
		0.649019,
		0.553828,
		0.442204,
		0.322316,
		0.204691,
		0.101903,
		0.028334,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.001825,
		0.041135,
		0.119331,
		0.221226,
		0.333305,
		0.443925,
		0.543575,
		0.625228,
		0.684786,
		0.721612,
		0.738959,
		0.743703,
		0.743999,
		0.743232,
		0.736123,
		0.714427,
		0.672067,
		0.606781,
		0.520121,
		0.416966,
		0.305002,
		0.1943,
		0.096994,
		0.027034,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.014591,
		0.071204,
		0.156413,
		0.256554,
		0.359697,
		0.455881,
		0.537419,
		0.599306,
		0.639732,
		0.660568,
		0.667436,
		0.668181,
		0.667894,
		0.663304,
		0.646618,
		0.611418,
		0.554862,
		0.477883,
		0.384738,
		0.282483,
		0.180546,
		0.090387,
		0.025256,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.001561,
		0.035026,
		0.101068,
		0.186232,
		0.278635,
		0.368169,
		0.446768,
		0.508764,
		0.551352,
		0.575095,
		0.584203,
		0.585686,
		0.585621,
		0.582942,
		0.570757,
		0.542556,
		0.495029,
		0.428498,
		0.346533,
		0.255441,
		0.163827,
		0.082265,
		0.023047,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.011854,
		0.057546,
		0.125657,
		0.204699,
		0.284745,
		0.35766,
		0.417399,
		0.460409,
		0.486112,
		0.497293,
		0.499767,
		0.499827,
		0.498456,
		0.490085,
		0.468408,
		0.429797,
		0.374015,
		0.303917,
		0.224966,
		0.144809,
		0.072946,
		0.020493,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.001204,
		0.026875,
		0.077097,
		0.141108,
		0.209488,
		0.274347,
		0.329566,
		0.371134,
		0.397574,
		0.410368,
		0.413956,
		0.414184,
		0.413595,
		0.408251,
		0.392355,
		0.36215,
		0.316924,
		0.25883,
		0.192443,
		0.124352,
		0.06285,
		0.017708,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.008587,
		0.041445,
		0.089903,
		0.145338,
		0.200397,
		0.249185,
		0.287533,
		0.313356,
		0.32705,
		0.331687,
		0.332201,
		0.332002,
		0.328883,
		0.317842,
		0.295193,
		0.259871,
		0.213377,
		0.159395,
		0.103418,
		0.052454,
		0.014824,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.000818,
		0.018148,
		0.051724,
		0.09396,
		0.138285,
		0.17929,
		0.212934,
		0.236823,
		0.25055,
		0.255969,
		0.256874,
		0.256834,
		0.255208,
		0.248009,
		0.231825,
		0.205375,
		0.169596,
		0.127325,
		0.082967,
		0.042238,
		0.011975,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.005387,
		0.025838,
		0.055636,
		0.089175,
		0.121741,
		0.149661,
		0.170509,
		0.183376,
		0.189147,
		0.19046,
		0.190493,
		0.18976,
		0.185408,
		0.174471,
		0.1556,
		0.129275,
		0.097571,
		0.063871,
		0.032644,
		0.009287,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.000471,
		0.010384,
		0.029382,
		0.052927,
		0.077134,
		0.098875,
		0.115924,
		0.127154,
		0.132767,
		0.134389,
		0.134495,
		0.134224,
		0.131824,
		0.124906,
		0.112184,
		0.09381,
		0.071206,
		0.046839,
		0.02404,
		0.006863,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.00278,
		0.013239,
		0.028272,
		0.044879,
		0.06058,
		0.073513,
		0.082561,
		0.087528,
		0.089266,
		0.089464,
		0.089387,
		0.088209,
		0.084171,
		0.076164,
		0.064131,
		0.048973,
		0.032382,
		0.016693,
		0.004784,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.000214,
		0.004677,
		0.013129,
		0.023426,
		0.033761,
		0.042718,
		0.049356,
		0.053314,
		0.054932,
		0.055211,
		0.055198,
		0.054701,
		0.05257,
		0.047945,
		0.040669,
		0.031258,
		0.020783,
		0.010764,
		0.003097,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.001059,
		0.005004,
		0.010587,
		0.016622,
		0.02215,
		0.026482,
		0.029267,
		0.030563,
		0.030868,
		0.030876,
		0.030705,
		0.029718,
		0.027329,
		0.023365,
		0.018083,
		0.012095,
		0.006296,
		0.001819,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		6.5e-005,
		0.001406,
		0.003911,
		0.006903,
		0.009823,
		0.012245,
		0.013913,
		0.014781,
		0.015042,
		0.015059,
		0.015015,
		0.014633,
		0.013573,
		0.011703,
		0.009125,
		0.006142,
		0.003214,
		0.000933,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000228,
		0.001066,
		0.002233,
		0.003462,
		0.004545,
		0.005342,
		0.005799,
		0.005965,
		0.005985,
		0.005977,
		0.005863,
		0.005488,
		0.004775,
		0.003753,
		0.002543,
		0.001339,
		0.00039,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		0.000175,
		0.000481,
		0.000838,
		0.001175,
		0.00144,
		0.001605,
		0.001676,
		0.001688,
		0.001688,
		0.001665,
		0.001573,
		0.001382,
		0.001096,
		0.000748,
		0.000396,
		0.000116,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		1e-005,
		4.7e-005,
		9.6e-005,
		0.000147,
		0.00019,
		0.000219,
		0.000233,
		0.000237,
		0.000237,
		0.000235,
		0.000224,
		0.000199,
		0.000159,
		0.000109,
		5.8e-005,
		1.7e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		1.7e-005,
		5.8e-005,
		0.000109,
		0.000159,
		0.000199,
		0.000224,
		0.000235,
		0.000237,
		0.000237,
		0.000233,
		0.000219,
		0.00019,
		0.000147,
		9.6e-005,
		4.7e-005,
		1e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		0.000116,
		0.000396,
		0.000748,
		0.001096,
		0.001382,
		0.001573,
		0.001665,
		0.001688,
		0.001688,
		0.001676,
		0.001605,
		0.00144,
		0.001175,
		0.000838,
		0.000481,
		0.000175,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		0.00039,
		0.001339,
		0.002543,
		0.003753,
		0.004775,
		0.005488,
		0.005863,
		0.005977,
		0.005985,
		0.005965,
		0.005799,
		0.005342,
		0.004545,
		0.003462,
		0.002233,
		0.001066,
		0.000228,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000933,
		0.003214,
		0.006142,
		0.009125,
		0.011703,
		0.013573,
		0.014633,
		0.015015,
		0.015059,
		0.015042,
		0.014781,
		0.013913,
		0.012245,
		0.009823,
		0.006903,
		0.003911,
		0.001406,
		6.5e-005,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.001819,
		0.006296,
		0.012095,
		0.018083,
		0.023365,
		0.027329,
		0.029718,
		0.030705,
		0.030876,
		0.030868,
		0.030563,
		0.029267,
		0.026482,
		0.02215,
		0.016622,
		0.010587,
		0.005004,
		0.001059,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.003097,
		0.010764,
		0.020783,
		0.031258,
		0.040669,
		0.047945,
		0.05257,
		0.054701,
		0.055198,
		0.055211,
		0.054932,
		0.053314,
		0.049356,
		0.042718,
		0.033761,
		0.023426,
		0.013129,
		0.004677,
		0.000214,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.004784,
		0.016693,
		0.032382,
		0.048973,
		0.064131,
		0.076164,
		0.084171,
		0.088209,
		0.089387,
		0.089464,
		0.089266,
		0.087528,
		0.082561,
		0.073513,
		0.06058,
		0.044879,
		0.028272,
		0.013239,
		0.00278,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.006863,
		0.02404,
		0.046839,
		0.071206,
		0.09381,
		0.112184,
		0.124906,
		0.131824,
		0.134224,
		0.134495,
		0.134389,
		0.132767,
		0.127154,
		0.115924,
		0.098875,
		0.077134,
		0.052927,
		0.029382,
		0.010384,
		0.000471,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.009287,
		0.032644,
		0.063871,
		0.097571,
		0.129275,
		0.1556,
		0.174471,
		0.185408,
		0.18976,
		0.190493,
		0.19046,
		0.189147,
		0.183376,
		0.170509,
		0.149661,
		0.121741,
		0.089175,
		0.055636,
		0.025838,
		0.005387,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.011975,
		0.042238,
		0.082967,
		0.127325,
		0.169596,
		0.205375,
		0.231825,
		0.248009,
		0.255208,
		0.256834,
		0.256874,
		0.255969,
		0.25055,
		0.236823,
		0.212934,
		0.17929,
		0.138285,
		0.09396,
		0.051724,
		0.018148,
		0.000818,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.014824,
		0.052454,
		0.103418,
		0.159395,
		0.213377,
		0.259871,
		0.295193,
		0.317842,
		0.328883,
		0.332002,
		0.332201,
		0.331687,
		0.32705,
		0.313356,
		0.287533,
		0.249185,
		0.200397,
		0.145338,
		0.089903,
		0.041445,
		0.008587,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.017708,
		0.06285,
		0.124352,
		0.192443,
		0.25883,
		0.316924,
		0.36215,
		0.392355,
		0.408251,
		0.413595,
		0.414184,
		0.413956,
		0.410368,
		0.397574,
		0.371134,
		0.329566,
		0.274347,
		0.209488,
		0.141108,
		0.077097,
		0.026875,
		0.001204,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.020493,
		0.072946,
		0.144809,
		0.224966,
		0.303917,
		0.374015,
		0.429797,
		0.468408,
		0.490085,
		0.498456,
		0.499827,
		0.499767,
		0.497293,
		0.486112,
		0.460409,
		0.417399,
		0.35766,
		0.284745,
		0.204699,
		0.125657,
		0.057546,
		0.011854,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.023047,
		0.082265,
		0.163827,
		0.255441,
		0.346533,
		0.428498,
		0.495029,
		0.542556,
		0.570757,
		0.582942,
		0.585621,
		0.585686,
		0.584203,
		0.575095,
		0.551352,
		0.508764,
		0.446768,
		0.368169,
		0.278635,
		0.186232,
		0.101068,
		0.035026,
		0.001561,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.025256,
		0.090387,
		0.180546,
		0.282483,
		0.384738,
		0.477883,
		0.554862,
		0.611418,
		0.646618,
		0.663304,
		0.667894,
		0.668181,
		0.667436,
		0.660568,
		0.639732,
		0.599306,
		0.537419,
		0.455881,
		0.359697,
		0.256554,
		0.156413,
		0.071204,
		0.014591,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.027034,
		0.096994,
		0.1943,
		0.305002,
		0.416966,
		0.520121,
		0.606781,
		0.672067,
		0.714427,
		0.736123,
		0.743232,
		0.743999,
		0.743703,
		0.738959,
		0.721612,
		0.684786,
		0.625228,
		0.543575,
		0.443925,
		0.333305,
		0.221226,
		0.119331,
		0.041135,
		0.001825,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.028334,
		0.101903,
		0.204691,
		0.322316,
		0.442204,
		0.553828,
		0.649019,
		0.722367,
		0.771729,
		0.798725,
		0.808892,
		0.81052,
		0.810449,
		0.807497,
		0.793834,
		0.761617,
		0.706256,
		0.627109,
		0.527188,
		0.412633,
		0.29224,
		0.177071,
		0.080172,
		0.01635,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.029149,
		0.105077,
		0.211619,
		0.334215,
		0.460076,
		0.578414,
		0.680727,
		0.761188,
		0.817112,
		0.849464,
		0.863107,
		0.866039,
		0.866109,
		0.864494,
		0.854357,
		0.827286,
		0.777497,
		0.703054,
		0.605771,
		0.49074,
		0.365832,
		0.241294,
		0.12944,
		0.044403,
		0.001962,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.029512,
		0.10662,
		0.215263,
		0.340931,
		0.470829,
		0.594078,
		0.701997,
		0.788449,
		0.850289,
		0.887842,
		0.905228,
		0.909906,
		0.910193,
		0.909445,
		0.902408,
		0.88057,
		0.837182,
		0.769076,
		0.676837,
		0.564404,
		0.438588,
		0.308649,
		0.185967,
		0.083785,
		0.017013,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.029484,
		0.106744,
		0.216028,
		0.343066,
		0.475222,
		0.601687,
		0.713725,
		0.805002,
		0.871991,
		0.914431,
		0.935677,
		0.942488,
		0.943208,
		0.942932,
		0.938409,
		0.921517,
		0.884835,
		0.824096,
		0.73868,
		0.63134,
		0.507751,
		0.37608,
		0.246645,
		0.131646,
		0.04496,
		0.001978,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.029144,
		0.105728,
		0.21446,
		0.341453,
		0.474364,
		0.602567,
		0.717384,
		0.812377,
		0.883723,
		0.930634,
		0.955728,
		0.96498,
		0.966433,
		0.966371,
		0.963724,
		0.951225,
		0.921114,
		0.868202,
		0.790745,
		0.690328,
		0.571468,
		0.441194,
		0.308696,
		0.185047,
		0.082994,
		0.016784,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.028577,
		0.103873,
		0.211158,
		0.337025,
		0.46951,
		0.59826,
		0.714734,
		0.812471,
		0.887425,
		0.938357,
		0.967195,
		0.979108,
		0.98162,
		0.981678,
		0.980301,
		0.971493,
		0.947489,
		0.902391,
		0.833445,
		0.741127,
		0.628819,
		0.502414,
		0.369959,
		0.241376,
		0.128242,
		0.043618,
		0.001912,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.027862,
		0.101464,
		0.206694,
		0.330677,
		0.461886,
		0.590294,
		0.707552,
		0.807238,
		0.885148,
		0.939661,
		0.972083,
		0.986795,
		0.990679,
		0.990914,
		0.990301,
		0.984431,
		0.96586,
		0.928207,
		0.867848,
		0.784235,
		0.679669,
		0.558942,
		0.428982,
		0.298578,
		0.178148,
		0.079568,
		0.016032,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.027066,
		0.098742,
		0.201554,
		0.323181,
		0.452559,
		0.580011,
		0.697422,
		0.798444,
		0.878774,
		0.936471,
		0.972293,
		0.98988,
		0.995414,
		0.99599,
		0.99577,
		0.992111,
		0.978195,
		0.947385,
		0.895339,
		0.820585,
		0.724419,
		0.610585,
		0.484947,
		0.355202,
		0.230652,
		0.122028,
		0.041348,
		0.001807,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.026239,
		0.09589,
		0.196111,
		0.315133,
		0.442357,
		0.56847,
		0.685602,
		0.787513,
		0.86984,
		0.930382,
		0.969415,
		0.989904,
		0.997321,
		0.998467,
		0.998418,
		0.996322,
		0.986257,
		0.961557,
		0.917318,
		0.851254,
		0.763735,
		0.657532,
		0.537502,
		0.410314,
		0.284216,
		0.168853,
		0.075127,
		0.015085,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.025414,
		0.093032,
		0.190619,
		0.306942,
		0.431857,
		0.556411,
		0.672987,
		0.77547,
		0.859466,
		0.922566,
		0.96463,
		0.988014,
		0.997502,
		0.999471,
		0.999516,
		0.998441,
		0.991455,
		0.972079,
		0.935005,
		0.877246,
		0.798332,
		0.700132,
		0.586569,
		0.463351,
		0.337737,
		0.218359,
		0.115075,
		0.038856,
		0.001693,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.024612,
		0.090239,
		0.185227,
		0.298855,
		0.421411,
		0.544295,
		0.66014,
		0.762966,
		0.848375,
		0.91379,
		0.958714,
		0.98496,
		0.996665,
		0.999705,
		0.999887,
		0.999412,
		0.994799,
		0.979961,
		0.949345,
		0.899376,
		0.82883,
		0.73875,
		0.632191,
		0.513969,
		0.390425,
		0.269249,
		0.159331,
		0.07064,
		0.014139,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.023841,
		0.087547,
		0.180011,
		0.290993,
		0.411199,
		0.532363,
		0.647364,
		0.750362,
		0.836975,
		0.904491,
		0.952111,
		0.981167,
		0.995193,
		0.999536,
		0.999982,
		0.999812,
		0.996952,
		0.985898,
		0.961014,
		0.918247,
		0.85571,
		0.773688,
		0.674436,
		0.561943,
		0.441699,
		0.320521,
		0.2064,
		0.108382,
		0.036478,
		0.001584,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.023103,
		0.084966,
		0.174994,
		0.283404,
		0.401293,
		0.520717,
		0.634793,
		0.737825,
		0.825462,
		0.894886,
		0.94504,
		0.976834,
		0.993251,
		0.999102,
		0.999993,
		0.999955,
		0.99832,
		0.990351,
		0.970479,
		0.934294,
		0.879328,
		0.805178,
		0.713365,
		0.607109,
		0.491113,
		0.371384,
		0.255081,
		0.150396,
		0.06646,
		0.013263,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.022399,
		0.082496,
		0.170178,
		0.276092,
		0.391707,
		0.509384,
		0.622473,
		0.72542,
		0.813918,
		0.88507,
		0.937598,
		0.972044,
		0.99089,
		0.998419,
		0.999959,
		0.999994,
		0.999156,
		0.993633,
		0.978079,
		0.947842,
		0.899959,
		0.833411,
		0.749038,
		0.649351,
		0.538323,
		0.42121,
		0.304401,
		0.195295,
		0.102209,
		0.034297,
		0.001486,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.021727,
		0.080132,
		0.165555,
		0.269049,
		0.382435,
		0.498366,
		0.610413,
		0.71317,
		0.802379,
		0.875088,
		0.929834,
		0.966832,
		0.988116,
		0.997465,
		0.999859,
		1,
		0.999629,
		0.995985,
		0.984089,
		0.959167,
		0.91785,
		0.858565,
		0.781532,
		0.688606,
		0.583079,
		0.469512,
		0.353578,
		0.241943,
		0.142168,
		0.062632,
		0.012464,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.021084,
		0.077867,
		0.161115,
		0.262263,
		0.373465,
		0.487653,
		0.598615,
		0.701086,
		0.790869,
		0.864974,
		0.921781,
		0.961225,
		0.984931,
		0.996208,
		0.999651,
		1,
		0.999867,
		0.997607,
		0.988753,
		0.968525,
		0.933236,
		0.880827,
		0.810953,
		0.72486,
		0.625212,
		0.515919,
		0.40199,
		0.289411,
		0.185043,
		0.096544,
		0.032305,
		0.001396,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.02047,
		0.075696,
		0.156848,
		0.255722,
		0.364787,
		0.47724,
		0.587077,
		0.689177,
		0.779408,
		0.854755,
		0.913472,
		0.955247,
		0.981342,
		0.994625,
		0.999293,
		0.999995,
		0.999966,
		0.998674,
		0.992293,
		0.976158,
		0.946351,
		0.900395,
		0.837433,
		0.758149,
		0.664622,
		0.560159,
		0.449152,
		0.336951,
		0.229766,
		0.134589,
		0.059124,
		0.011736,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.019882,
		0.073614,
		0.152746,
		0.249415,
		0.356389,
		0.467118,
		0.575799,
		0.677451,
		0.768013,
		0.84446,
		0.90494,
		0.948926,
		0.97736,
		0.9927,
		0.998747,
		0.999968,
		0.999996,
		0.999333,
		0.994912,
		0.982294,
		0.957423,
		0.917472,
		0.861121,
		0.788542,
		0.701268,
		0.602046,
		0.494689,
		0.383959,
		0.27546,
		0.175563,
		0.091333,
		0.030481,
		0.001314,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.019319,
		0.071616,
		0.148801,
		0.243331,
		0.34826,
		0.457278,
		0.564777,
		0.665912,
		0.756699,
		0.834112,
		0.896212,
		0.94229,
		0.972999,
		0.990425,
		0.997978,
		0.999889,
		1,
		0.999707,
		0.99679,
		0.987147,
		0.966673,
		0.932261,
		0.882183,
		0.816138,
		0.735156,
		0.641458,
		0.538323,
		0.42996,
		0.321409,
		0.218463,
		0.127594,
		0.055902,
		0.011069,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.018779,
		0.069699,
		0.145004,
		0.237461,
		0.34039,
		0.447714,
		0.554008,
		0.654566,
		0.745479,
		0.823732,
		0.887317,
		0.935365,
		0.968276,
		0.987794,
		0.996959,
		0.999723,
		1,
		0.999895,
		0.998086,
		0.990912,
		0.974313,
		0.944967,
		0.900792,
		0.841057,
		0.766326,
		0.678333,
		0.579853,
		0.474581,
		0.367032,
		0.262459,
		0.166781,
		0.086531,
		0.028807,
		0.001239,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.018262,
		0.067857,
		0.141349,
		0.231794,
		0.332769,
		0.438416,
		0.54349,
		0.643416,
		0.734365,
		0.81334,
		0.87828,
		0.928179,
		0.963209,
		0.98481,
		0.995666,
		0.999435,
		0.999996,
		0.999973,
		0.998938,
		0.993771,
		0.980542,
		0.955789,
		0.917126,
		0.863435,
		0.794849,
		0.712649,
		0.61914,
		0.517538,
		0.411865,
		0.306869,
		0.207954,
		0.121124,
		0.052935,
		0.010458,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.017765,
		0.066086,
		0.137829,
		0.226323,
		0.325389,
		0.429378,
		0.533218,
		0.632464,
		0.723368,
		0.802956,
		0.869125,
		0.920755,
		0.957819,
		0.981477,
		0.994084,
		0.998994,
		0.999975,
		0.999997,
		0.999465,
		0.995886,
		0.985548,
		0.964919,
		0.931363,
		0.883417,
		0.820817,
		0.744423,
		0.656102,
		0.558619,
		0.455537,
		0.351133,
		0.250328,
		0.158632,
		0.082095,
		0.027267,
		0.00117,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.017289,
		0.064384,
		0.134438,
		0.221038,
		0.318238,
		0.420591,
		0.523188,
		0.621711,
		0.712494,
		0.792593,
		0.859876,
		0.913121,
		0.952128,
		0.977805,
		0.992201,
		0.998368,
		0.999911,
		1,
		0.999765,
		0.997403,
		0.989507,
		0.972543,
		0.943681,
		0.901155,
		0.84434,
		0.773702,
		0.690698,
		0.597675,
		0.497762,
		0.394798,
		0.293251,
		0.19817,
		0.11513,
		0.050198,
		0.009896,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.016832,
		0.062746,
		0.131168,
		0.215931,
		0.31131,
		0.412048,
		0.513395,
		0.611159,
		0.701753,
		0.782268,
		0.850552,
		0.905299,
		0.946156,
		0.973805,
		0.990011,
		0.997533,
		0.999777,
		1,
		0.999916,
		0.998451,
		0.992579,
		0.978837,
		0.954256,
		0.916807,
		0.865538,
		0.800552,
		0.722924,
		0.634605,
		0.538323,
		0.437501,
		0.336187,
		0.238994,
		0.151058,
		0.07799,
		0.025848,
		0.001107,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.016392,
		0.06117,
		0.128015,
		0.210995,
		0.304595,
		0.403741,
		0.503836,
		0.600807,
		0.69115,
		0.771993,
		0.841172,
		0.897313,
		0.939924,
		0.969489,
		0.987511,
		0.996468,
		0.999544,
		0.999997,
		0.999978,
		0.99914,
		0.994912,
		0.983968,
		0.963257,
		0.930531,
		0.884543,
		0.82506,
		0.752804,
		0.669354,
		0.577059,
		0.478955,
		0.378694,
		0.280485,
		0.189047,
		0.109566,
		0.047667,
		0.009378,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.01597,
		0.059653,
		0.124973,
		0.206223,
		0.298086,
		0.395662,
		0.494505,
		0.590655,
		0.68069,
		0.761779,
		0.831754,
		0.889182,
		0.933455,
		0.964874,
		0.984704,
		0.995156,
		0.999183,
		0.99998,
		0.999997,
		0.999567,
		0.996638,
		0.988091,
		0.970847,
		0.942482,
		0.901488,
		0.847326,
		0.780384,
		0.701897,
		0.613858,
		0.518938,
		0.420418,
		0.322126,
		0.228392,
		0.144007,
		0.074184,
		0.024536,
		0.001049,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.015563,
		0.05819,
		0.122037,
		0.201607,
		0.291774,
		0.387806,
		0.485397,
		0.580701,
		0.670377,
		0.751638,
		0.822315,
		0.880929,
		0.926769,
		0.959974,
		0.981594,
		0.993585,
		0.998669,
		0.999928,
		1,
		0.999809,
		0.997877,
		0.991352,
		0.977183,
		0.952816,
		0.916514,
		0.867459,
		0.805731,
		0.732241,
		0.648648,
		0.557281,
		0.461072,
		0.363491,
		0.268504,
		0.180527,
		0.104393,
		0.045322,
		0.0089,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.015172,
		0.056781,
		0.119202,
		0.197141,
		0.285652,
		0.380165,
		0.476508,
		0.570945,
		0.660215,
		0.741577,
		0.812868,
		0.872571,
		0.919885,
		0.954807,
		0.978187,
		0.991748,
		0.997978,
		0.999819,
		1,
		0.999931,
		0.998733,
		0.993882,
		0.982413,
		0.961681,
		0.929759,
		0.885575,
		0.828924,
		0.760415,
		0.681392,
		0.593863,
		0.500431,
		0.404232,
		0.308888,
		0.218463,
		0.137432,
		0.070648,
		0.023321,
		0.000995,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.014796,
		0.055422,
		0.116464,
		0.192819,
		0.279714,
		0.372731,
		0.467831,
		0.561384,
		0.650207,
		0.731605,
		0.803427,
		0.864126,
		0.912824,
		0.949389,
		0.974494,
		0.989641,
		0.997092,
		0.999627,
		0.999998,
		0.999982,
		0.999296,
		0.995804,
		0.986675,
		0.969224,
		0.941362,
		0.901795,
		0.850054,
		0.786465,
		0.712079,
		0.628601,
		0.538323,
		0.444067,
		0.349131,
		0.25725,
		0.172561,
		0.099574,
		0.043145,
		0.008457,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.014433,
		0.054111,
		0.113819,
		0.188634,
		0.273952,
		0.365499,
		0.459363,
		0.552017,
		0.640354,
		0.721729,
		0.794004,
		0.85561,
		0.905603,
		0.943738,
		0.970524,
		0.987261,
		0.995996,
		0.99933,
		0.999983,
		0.999998,
		0.999645,
		0.997227,
		0.990101,
		0.975582,
		0.951461,
		0.916242,
		0.869219,
		0.810454,
		0.740722,
		0.661443,
		0.574618,
		0.482769,
		0.388893,
		0.296414,
		0.209152,
		0.131291,
		0.067358,
		0.022194,
		0.000945,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.014083,
		0.052846,
		0.111261,
		0.184582,
		0.268359,
		0.358463,
		0.451099,
		0.542841,
		0.630658,
		0.711954,
		0.784611,
		0.84704,
		0.898242,
		0.93787,
		0.966289,
		0.984612,
		0.994676,
		0.998903,
		0.999941,
		1,
		0.999844,
		0.998248,
		0.992809,
		0.980889,
		0.960187,
		0.929041,
		0.886525,
		0.832457,
		0.767353,
		0.692366,
		0.609221,
		0.520162,
		0.427896,
		0.335557,
		0.246666,
		0.165103,
		0.09508,
		0.041121,
		0.008046,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.013746,
		0.051625,
		0.108788,
		0.180656,
		0.262931,
		0.351615,
		0.443032,
		0.533853,
		0.62112,
		0.702287,
		0.775256,
		0.838428,
		0.890756,
		0.931803,
		0.961803,
		0.981696,
		0.993125,
		0.998327,
		0.999851,
		1,
		0.999944,
		0.998953,
		0.994912,
		0.985268,
		0.967671,
		0.940314,
		0.902078,
		0.852555,
		0.792019,
		0.721369,
		0.642073,
		0.556109,
		0.465916,
		0.374349,
		0.28465,
		0.200411,
		0.12555,
		0.064291,
		0.021147,
		0.000899,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.013421,
		0.050445,
		0.106396,
		0.176852,
		0.257659,
		0.34495,
		0.435159,
		0.52505,
		0.61174,
		0.692731,
		0.76595,
		0.829789,
		0.883162,
		0.925553,
		0.957078,
		0.978519,
		0.991337,
		0.997585,
		0.999693,
		0.999998,
		0.999986,
		0.999419,
		0.996509,
		0.988837,
		0.974035,
		0.950184,
		0.91599,
		0.870838,
		0.814777,
		0.74847,
		0.673139,
		0.590511,
		0.502771,
		0.412516,
		0.32272,
		0.236704,
		0.15811,
		0.09088,
		0.039236,
		0.007665,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.013107,
		0.049305,
		0.104082,
		0.173165,
		0.25254,
		0.338462,
		0.427474,
		0.51643,
		0.602518,
		0.683291,
		0.756699,
		0.821134,
		0.875474,
		0.919137,
		0.952128,
		0.975089,
		0.989308,
		0.996662,
		0.999445,
		0.999986,
		0.999998,
		0.999707,
		0.997692,
		0.991705,
		0.9794,
		0.958769,
		0.928371,
		0.887401,
		0.835698,
		0.773702,
		0.702404,
		0.623297,
		0.538323,
		0.449834,
		0.360554,
		0.273547,
		0.192195,
		0.120172,
		0.061429,
		0.020173,
		0.000857,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.012805,
		0.048204,
		0.101841,
		0.16959,
		0.247567,
		0.332145,
		0.419973,
		0.50799,
		0.593455,
		0.673969,
		0.747512,
		0.812474,
		0.867708,
		0.912568,
		0.946968,
		0.971415,
		0.987038,
		0.995546,
		0.999088,
		0.999951,
		1,
		0.999871,
		0.998541,
		0.993974,
		0.983876,
		0.966185,
		0.939331,
		0.902341,
		0.854856,
		0.797109,
		0.729875,
		0.654422,
		0.572465,
		0.486123,
		0.397884,
		0.31057,
		0.227317,
		0.151546,
		0.086952,
		0.037478,
		0.00731,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.012512,
		0.047138,
		0.099671,
		0.166123,
		0.242735,
		0.325995,
		0.41265,
		0.499726,
		0.584549,
		0.664769,
		0.738394,
		0.80382,
		0.859876,
		0.905864,
		0.94161,
		0.967506,
		0.98453,
		0.994227,
		0.998605,
		0.999876,
		1,
		0.999953,
		0.999128,
		0.995735,
		0.987569,
		0.972543,
		0.94898,
		0.915755,
		0.872333,
		0.818748,
		0.755573,
		0.683863,
		0.605119,
		0.52124,
		0.434489,
		0.347461,
		0.263058,
		0.184465,
		0.11513,
		0.058753,
		0.019264,
		0.000817,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.01223,
		0.046108,
		0.09757,
		0.162759,
		0.238038,
		0.320005,
		0.405501,
		0.491635,
		0.575799,
		0.655692,
		0.729352,
		0.795181,
		0.85199,
		0.899037,
		0.93607,
		0.963372,
		0.981785,
		0.9927,
		0.997978,
		0.999744,
		0.999998,
		0.999988,
		0.999515,
		0.997073,
		0.990579,
		0.977948,
		0.957423,
		0.927745,
		0.888213,
		0.838682,
		0.77953,
		0.711615,
		0.636233,
		0.555074,
		0.470189,
		0.383959,
		0.299063,
		0.218463,
		0.145377,
		0.08327,
		0.035835,
		0.006979,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.011957,
		0.045111,
		0.095534,
		0.159495,
		0.233473,
		0.31417,
		0.398522,
		0.483714,
		0.567204,
		0.64674,
		0.72039,
		0.786565,
		0.844063,
		0.892101,
		0.930361,
		0.959025,
		0.97881,
		0.990961,
		0.997195,
		0.999536,
		0.999989,
		0.999998,
		0.999755,
		0.998064,
		0.992999,
		0.982503,
		0.964764,
		0.938407,
		0.902584,
		0.85698,
		0.801791,
		0.737688,
		0.665773,
		0.58754,
		0.504839,
		0.419845,
		0.33503,
		0.253142,
		0.177183,
		0.110396,
		0.056247,
		0.018415,
		0.00078,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.011692,
		0.044146,
		0.09356,
		0.156326,
		0.229035,
		0.308487,
		0.391707,
		0.475959,
		0.558763,
		0.637915,
		0.711512,
		0.77798,
		0.836105,
		0.88507,
		0.924495,
		0.954475,
		0.975609,
		0.989007,
		0.996245,
		0.999236,
		0.999959,
		1,
		0.999892,
		0.998776,
		0.994912,
		0.986304,
		0.971103,
		0.947842,
		0.915536,
		0.873716,
		0.822408,
		0.762104,
		0.693726,
		0.618579,
		0.538323,
		0.45494,
		0.370703,
		0.288158,
		0.210104,
		0.139573,
		0.079817,
		0.034297,
		0.006671,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.011437,
		0.043211,
		0.091646,
		0.15325,
		0.224718,
		0.302949,
		0.385053,
		0.468368,
		0.550474,
		0.629217,
		0.702724,
		0.769433,
		0.828126,
		0.877955,
		0.918486,
		0.949734,
		0.972191,
		0.986838,
		0.995118,
		0.998827,
		0.999897,
		1,
		0.999961,
		0.999268,
		0.996398,
		0.98944,
		0.976536,
		0.956143,
		0.927158,
		0.888967,
		0.841438,
		0.784898,
		0.720092,
		0.648151,
		0.570554,
		0.489096,
		0.405867,
		0.32322,
		0.243759,
		0.170317,
		0.105945,
		0.053897,
		0.017621,
		0.000745,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.01119,
		0.042305,
		0.089791,
		0.150262,
		0.220519,
		0.297552,
		0.378555,
		0.460936,
		0.542336,
		0.620647,
		0.694027,
		0.760931,
		0.820135,
		0.870768,
		0.912347,
		0.944814,
		0.968563,
		0.984457,
		0.993807,
		0.998294,
		0.999785,
		0.999999,
		0.99999,
		0.999593,
		0.997527,
		0.991996,
		0.981155,
		0.963404,
		0.937538,
		0.902812,
		0.858946,
		0.80611,
		0.744885,
		0.676234,
		0.601465,
		0.522198,
		0.440344,
		0.358077,
		0.277816,
		0.202206,
		0.134105,
		0.076573,
		0.032856,
		0.006382,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.010951,
		0.041428,
		0.08799,
		0.14736,
		0.216434,
		0.292293,
		0.372208,
		0.45366,
		0.534347,
		0.612205,
		0.685426,
		0.75248,
		0.812141,
		0.863519,
		0.906089,
		0.939726,
		0.964734,
		0.981865,
		0.992308,
		0.997627,
		0.99961,
		0.99999,
		0.999999,
		0.999794,
		0.998364,
		0.99405,
		0.985047,
		0.969716,
		0.946765,
		0.915332,
		0.875,
		0.82579,
		0.768129,
		0.702821,
		0.63101,
		0.554152,
		0.473988,
		0.392521,
		0.311993,
		0.234873,
		0.163836,
		0.101756,
		0.051692,
		0.016877,
		0.000713,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.010719,
		0.040578,
		0.086243,
		0.14454,
		0.212459,
		0.287165,
		0.366009,
		0.446537,
		0.526505,
		0.603892,
		0.676922,
		0.744084,
		0.804153,
		0.85622,
		0.899725,
		0.934481,
		0.960712,
		0.979066,
		0.990617,
		0.996812,
		0.999355,
		0.999966,
		1,
		0.999909,
		0.998965,
		0.995675,
		0.988294,
		0.975164,
		0.954925,
		0.926607,
		0.889668,
		0.843991,
		0.789856,
		0.727916,
		0.65916,
		0.584888,
		0.50668,
		0.426374,
		0.346048,
		0.268,
		0.194735,
		0.128949,
		0.073521,
		0.031504,
		0.006111,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.010495,
		0.039753,
		0.084547,
		0.141799,
		0.208589,
		0.282166,
		0.359954,
		0.439564,
		0.518807,
		0.595706,
		0.668518,
		0.735749,
		0.796177,
		0.84888,
		0.893264,
		0.92909,
		0.956507,
		0.976067,
		0.988734,
		0.995843,
		0.999006,
		0.999913,
		1,
		0.999967,
		0.999381,
		0.996937,
		0.990973,
		0.979834,
		0.962103,
		0.936718,
		0.903024,
		0.860772,
		0.810106,
		0.751535,
		0.685901,
		0.614355,
		0.538323,
		0.45949,
		0.379775,
		0.301315,
		0.226451,
		0.157713,
		0.097809,
		0.049618,
		0.016179,
		0.000682,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.010277,
		0.038954,
		0.0829,
		0.139134,
		0.204822,
		0.277292,
		0.354039,
		0.432738,
		0.511251,
		0.587648,
		0.660215,
		0.727478,
		0.78822,
		0.841508,
		0.886718,
		0.923565,
		0.952128,
		0.972873,
		0.986658,
		0.994712,
		0.998551,
		0.999819,
		0.999999,
		0.999992,
		0.999656,
		0.997897,
		0.993157,
		0.983803,
		0.968379,
		0.945745,
		0.91514,
		0.876195,
		0.828924,
		0.773702,
		0.71123,
		0.642517,
		0.568845,
		0.491749,
		0.413,
		0.334582,
		0.258679,
		0.187663,
		0.124081,
		0.070648,
		0.030233,
		0.005858,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.010067,
		0.038178,
		0.081301,
		0.136543,
		0.201154,
		0.272538,
		0.34826,
		0.426054,
		0.503836,
		0.579717,
		0.652015,
		0.719275,
		0.780288,
		0.834112,
		0.880097,
		0.917916,
		0.947585,
		0.969489,
		0.984391,
		0.993413,
		0.997978,
		0.999669,
		0.999992,
		0.999999,
		0.999826,
		0.998608,
		0.994912,
		0.987147,
		0.973833,
		0.953766,
		0.926089,
		0.890323,
		0.846361,
		0.794448,
		0.735156,
		0.669354,
		0.598188,
		0.523054,
		0.445578,
		0.367599,
		0.291153,
		0.218463,
		0.151921,
		0.094085,
		0.047667,
		0.015523,
		0.000654,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.009862,
		0.037425,
		0.079747,
		0.134022,
		0.197581,
		0.267901,
		0.342613,
		0.41951,
		0.496559,
		0.571912,
		0.643919,
		0.711144,
		0.772387,
		0.826699,
		0.873411,
		0.912153,
		0.942888,
		0.965925,
		0.981935,
		0.991944,
		0.997277,
		0.999452,
		0.999971,
		1,
		0.999923,
		0.999119,
		0.996301,
		0.989936,
		0.978541,
		0.960857,
		0.935944,
		0.903223,
		0.862471,
		0.813813,
		0.757694,
		0.694856,
		0.626313,
		0.553326,
		0.477387,
		0.400195,
		0.323646,
		0.249819,
		0.180962,
		0.119481,
		0.067939,
		0.029038,
		0.00562,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.009664,
		0.036694,
		0.078237,
		0.13157,
		0.194101,
		0.263377,
		0.337094,
		0.413102,
		0.489418,
		0.564232,
		0.635927,
		0.703088,
		0.764523,
		0.819279,
		0.866667,
		0.906287,
		0.938045,
		0.962186,
		0.979295,
		0.990301,
		0.996439,
		0.999153,
		0.999926,
		1,
		0.999972,
		0.999473,
		0.99738,
		0.992239,
		0.982575,
		0.967092,
		0.944778,
		0.91496,
		0.877309,
		0.831837,
		0.778869,
		0.719025,
		0.653193,
		0.582507,
		0.508329,
		0.432226,
		0.355962,
		0.281477,
		0.21088,
		0.146439,
		0.090569,
		0.045828,
		0.014907,
		0.000627,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.009472,
		0.035984,
		0.07677,
		0.129184,
		0.19071,
		0.258963,
		0.331701,
		0.406828,
		0.48241,
		0.556677,
		0.628041,
		0.695109,
		0.756699,
		0.811856,
		0.859876,
		0.900326,
		0.933068,
		0.95828,
		0.976473,
		0.988485,
		0.995458,
		0.998762,
		0.999846,
		0.999999,
		0.999993,
		0.999707,
		0.9982,
		0.994115,
		0.986004,
		0.972543,
		0.952661,
		0.925601,
		0.890936,
		0.848568,
		0.798713,
		0.741872,
		0.678813,
		0.610552,
		0.538323,
		0.463572,
		0.387931,
		0.313213,
		0.241394,
		0.174607,
		0.11513,
		0.065382,
		0.027912,
		0.005396,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.009285,
		0.035295,
		0.075343,
		0.126861,
		0.187405,
		0.254655,
		0.326429,
		0.400685,
		0.475533,
		0.549244,
		0.620261,
		0.68721,
		0.748921,
		0.804438,
		0.853044,
		0.894279,
		0.927963,
		0.954217,
		0.973475,
		0.986495,
		0.994328,
		0.998267,
		0.999718,
		0.999993,
		0.999999,
		0.999852,
		0.998808,
		0.995624,
		0.988892,
		0.977279,
		0.959663,
		0.935212,
		0.903409,
		0.864056,
		0.81726,
		0.763413,
		0.70317,
		0.637429,
		0.567308,
		0.494131,
		0.419412,
		0.344836,
		0.272257,
		0.203676,
		0.141244,
		0.087245,
		0.044093,
		0.014327,
		0.000602,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.009104,
		0.034625,
		0.073955,
		0.1246,
		0.184184,
		0.25045,
		0.321275,
		0.394668,
		0.468785,
		0.541933,
		0.612586,
		0.679392,
		0.741192,
		0.797031,
		0.84618,
		0.888157,
		0.922742,
		0.950003,
		0.970307,
		0.984332,
		0.993045,
		0.99766,
		0.999531,
		0.999975,
		1,
		0.999934,
		0.999246,
		0.996818,
		0.991302,
		0.981366,
		0.965851,
		0.943859,
		0.91479,
		0.878352,
		0.834551,
		0.783673,
		0.726267,
		0.663118,
		0.595234,
		0.523824,
		0.450283,
		0.376183,
		0.303253,
		0.233376,
		0.168575,
		0.11101,
		0.062966,
		0.02685,
		0.005185,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.008928,
		0.033974,
		0.072605,
		0.122398,
		0.181043,
		0.246346,
		0.316237,
		0.388776,
		0.462163,
		0.534743,
		0.605018,
		0.671658,
		0.733515,
		0.789639,
		0.839291,
		0.881966,
		0.917412,
		0.945646,
		0.966975,
		0.981999,
		0.991607,
		0.996932,
		0.999274,
		0.999937,
		1,
		0.999976,
		0.999548,
		0.997745,
		0.993291,
		0.984868,
		0.971292,
		0.951607,
		0.92514,
		0.89151,
		0.850628,
		0.802682,
		0.748116,
		0.687611,
		0.622067,
		0.552583,
		0.480445,
		0.407111,
		0.334196,
		0.263466,
		0.196828,
		0.136317,
		0.084099,
		0.042454,
		0.013779,
		0.000579,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.008758,
		0.033341,
		0.071291,
		0.120254,
		0.177981,
		0.242338,
		0.31131,
		0.383005,
		0.455665,
		0.527672,
		0.597555,
		0.664008,
		0.725895,
		0.782268,
		0.832383,
		0.875716,
		0.911982,
		0.941155,
		0.963484,
		0.979499,
		0.990011,
		0.996078,
		0.998935,
		0.999868,
		0.999999,
		0.999994,
		0.999749,
		0.998451,
		0.994912,
		0.987845,
		0.976047,
		0.958519,
		0.934518,
		0.903584,
		0.865538,
		0.820474,
		0.768736,
		0.710906,
		0.647781,
		0.580358,
		0.509814,
		0.437501,
		0.364925,
		0.29374,
		0.22574,
		0.162846,
		0.107106,
		0.060681,
		0.025848,
		0.004986,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.008592,
		0.032726,
		0.070013,
		0.118164,
		0.174994,
		0.238425,
		0.306492,
		0.377353,
		0.449289,
		0.520718,
		0.590198,
		0.656444,
		0.718333,
		0.774923,
		0.825463,
		0.869413,
		0.90646,
		0.936538,
		0.959841,
		0.976835,
		0.988257,
		0.99509,
		0.998506,
		0.999758,
		0.999994,
		0.999999,
		0.999873,
		0.998974,
		0.996216,
		0.990352,
		0.980179,
		0.964657,
		0.942986,
		0.91463,
		0.879329,
		0.837084,
		0.78815,
		0.73301,
		0.672365,
		0.607109,
		0.538323,
		0.467253,
		0.395301,
		0.324015,
		0.255081,
		0.190312,
		0.131641,
		0.081119,
		0.040905,
		0.013263,
		0.000557,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.00843,
		0.032127,
		0.068768,
		0.116129,
		0.172081,
		0.234603,
		0.301781,
		0.371816,
		0.443032,
		0.51388,
		0.582946,
		0.648967,
		0.710833,
		0.767608,
		0.818536,
		0.863065,
		0.900855,
		0.931803,
		0.956054,
		0.974012,
		0.986346,
		0.993965,
		0.997978,
		0.999596,
		0.999979,
		1,
		0.999944,
		0.99935,
		0.997247,
		0.992444,
		0.983744,
		0.970081,
		0.950601,
		0.924704,
		0.892049,
		0.852555,
		0.806386,
		0.753936,
		0.69581,
		0.63281,
		0.565918,
		0.496282,
		0.425205,
		0.354133,
		0.28465,
		0.218463,
		0.1574,
		0.103402,
		0.058518,
		0.0249,
		0.004799,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.008273,
		0.031545,
		0.067557,
		0.114144,
		0.169238,
		0.230869,
		0.297172,
		0.366393,
		0.436892,
		0.507156,
		0.575799,
		0.641577,
		0.703397,
		0.760326,
		0.811609,
		0.856678,
		0.895174,
		0.926957,
		0.952128,
		0.971035,
		0.984278,
		0.9927,
		0.997343,
		0.999373,
		0.999946,
		1,
		0.99998,
		0.999611,
		0.998049,
		0.994171,
		0.986798,
		0.974848,
		0.957423,
		0.93386,
		0.903749,
		0.866927,
		0.823476,
		0.773702,
		0.71812,
		0.657442,
		0.592557,
		0.524519,
		0.454537,
		0.383959,
		0.31427,
		0.247077,
		0.184108,
		0.127199,
		0.078295,
		0.039439,
		0.012775,
		0.000536,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.008121,
		0.030978,
		0.066377,
		0.11221,
		0.166465,
		0.227222,
		0.292664,
		0.361079,
		0.430867,
		0.500544,
		0.568755,
		0.634275,
		0.696026,
		0.753082,
		0.804685,
		0.850259,
		0.889423,
		0.922008,
		0.948071,
		0.967907,
		0.982057,
		0.991293,
		0.996595,
		0.999079,
		0.999886,
		0.999999,
		0.999995,
		0.999783,
		0.998659,
		0.995579,
		0.989394,
		0.979014,
		0.963506,
		0.942154,
		0.914479,
		0.880246,
		0.839455,
		0.79233,
		0.739302,
		0.680995,
		0.618209,
		0.551909,
		0.483212,
		0.413374,
		0.343786,
		0.27596,
		0.211524,
		0.152218,
		0.099885,
		0.056468,
		0.024004,
		0.004622,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.007973,
		0.030427,
		0.065227,
		0.110324,
		0.163757,
		0.223658,
		0.288253,
		0.355873,
		0.424954,
		0.494044,
		0.561815,
		0.627063,
		0.688723,
		0.745878,
		0.797771,
		0.843815,
		0.883611,
		0.916963,
		0.943889,
		0.964635,
		0.979683,
		0.989742,
		0.995728,
		0.998706,
		0.999791,
		0.999995,
		0.999999,
		0.99989,
		0.999112,
		0.996711,
		0.991581,
		0.982633,
		0.968908,
		0.94964,
		0.924291,
		0.892557,
		0.854361,
		0.809849,
		0.759371,
		0.703467,
		0.642853,
		0.578407,
		0.51116,
		0.442279,
		0.373065,
		0.304937,
		0.239433,
		0.178197,
		0.122976,
		0.075614,
		0.038051,
		0.012313,
		0.000516,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.007828,
		0.02989,
		0.064107,
		0.108485,
		0.161115,
		0.220175,
		0.283937,
		0.350772,
		0.41915,
		0.487653,
		0.554977,
		0.619939,
		0.681489,
		0.738718,
		0.790869,
		0.83735,
		0.877744,
		0.91183,
		0.93959,
		0.961225,
		0.977161,
		0.988048,
		0.994738,
		0.998244,
		0.999651,
		0.999982,
		1,
		0.999951,
		0.999437,
		0.997607,
		0.993406,
		0.985756,
		0.973679,
		0.956371,
		0.933236,
		0.903905,
		0.868232,
		0.826287,
		0.778343,
		0.72486,
		0.666474,
		0.603979,
		0.538323,
		0.470589,
		0.40199,
		0.333861,
		0.267647,
		0.204903,
		0.147285,
		0.096544,
		0.054523,
		0.023155,
		0.004454,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.007688,
		0.029367,
		0.063015,
		0.106691,
		0.158534,
		0.21677,
		0.279714,
		0.345773,
		0.413455,
		0.48137,
		0.54824,
		0.612904,
		0.674326,
		0.731605,
		0.783986,
		0.83087,
		0.871829,
		0.906615,
		0.935179,
		0.957681,
		0.974494,
		0.98621,
		0.993622,
		0.997688,
		0.999457,
		0.999953,
		1,
		0.999982,
		0.999663,
		0.998304,
		0.994912,
		0.988431,
		0.977873,
		0.962398,
		0.941362,
		0.914337,
		0.881109,
		0.841678,
		0.796242,
		0.745185,
		0.689066,
		0.628601,
		0.564656,
		0.498233,
		0.430464,
		0.362598,
		0.295997,
		0.232129,
		0.172561,
		0.118957,
		0.073068,
		0.036734,
		0.011876,
		0.000497,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.007551,
		0.028857,
		0.061951,
		0.104941,
		0.156015,
		0.213443,
		0.27558,
		0.340874,
		0.407865,
		0.475193,
		0.541604,
		0.605958,
		0.667234,
		0.72454,
		0.777124,
		0.824381,
		0.865871,
		0.901325,
		0.930665,
		0.954009,
		0.971686,
		0.98423,
		0.992377,
		0.99703,
		0.9992,
		0.999901,
		0.999999,
		0.999996,
		0.999812,
		0.998834,
		0.99614,
		0.990705,
		0.981538,
		0.967772,
		0.948721,
		0.9239,
		0.893035,
		0.856057,
		0.813093,
		0.764456,
		0.710629,
		0.652255,
		0.59012,
		0.525152,
		0.4584,
		0.391034,
		0.324337,
		0.259691,
		0.198581,
		0.142584,
		0.093366,
		0.052677,
		0.02235,
		0.004296,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.007418,
		0.028361,
		0.060913,
		0.103233,
		0.153554,
		0.210189,
		0.271535,
		0.336073,
		0.402378,
		0.469119,
		0.535067,
		0.599101,
		0.660215,
		0.717527,
		0.770286,
		0.817887,
		0.859876,
		0.895967,
		0.926053,
		0.950217,
		0.968741,
		0.982109,
		0.991001,
		0.996266,
		0.998873,
		0.999819,
		0.999996,
		0.999999,
		0.999905,
		0.999227,
		0.997128,
		0.992621,
		0.984721,
		0.972543,
		0.955361,
		0.932642,
		0.904051,
		0.869459,
		0.828924,
		0.78269,
		0.73117,
		0.674931,
		0.61469,
		0.551297,
		0.485726,
		0.419073,
		0.352538,
		0.287427,
		0.225145,
		0.167184,
		0.11513,
		0.070648,
		0.035484,
		0.011462,
		0.000479,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.007289,
		0.027878,
		0.059902,
		0.101566,
		0.151149,
		0.207007,
		0.267574,
		0.331368,
		0.396993,
		0.463149,
		0.528629,
		0.592333,
		0.65327,
		0.710566,
		0.763477,
		0.811392,
		0.85385,
		0.890546,
		0.92135,
		0.946308,
		0.965663,
		0.97985,
		0.989494,
		0.995392,
		0.998468,
		0.999696,
		0.999984,
		1,
		0.999958,
		0.99951,
		0.99791,
		0.99422,
		0.987468,
		0.976757,
		0.961329,
		0.940607,
		0.914202,
		0.881923,
		0.843766,
		0.79991,
		0.750696,
		0.696627,
		0.638345,
		0.57663,
		0.512384,
		0.446628,
		0.380488,
		0.315194,
		0.252073,
		0.192542,
		0.138102,
		0.090341,
		0.050923,
		0.021587,
		0.004145,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.007163,
		0.027406,
		0.058915,
		0.099939,
		0.148801,
		0.203896,
		0.263696,
		0.326755,
		0.391707,
		0.457278,
		0.522288,
		0.585653,
		0.646399,
		0.703661,
		0.756699,
		0.804901,
		0.847798,
		0.88507,
		0.916561,
		0.94229,
		0.962458,
		0.977455,
		0.987855,
		0.994403,
		0.997978,
		0.999527,
		0.999959,
		1,
		0.999985,
		0.999707,
		0.998518,
		0.995539,
		0.989821,
		0.980459,
		0.966673,
		0.947842,
		0.923529,
		0.893487,
		0.857652,
		0.816138,
		0.769224,
		0.717343,
		0.661072,
		0.601121,
		0.538323,
		0.473628,
		0.40809,
		0.342867,
		0.27921,
		0.218463,
		0.162051,
		0.111483,
		0.068345,
		0.034297,
		0.011069,
		0.000462,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.00704,
		0.026947,
		0.057952,
		0.09835,
		0.146505,
		0.200852,
		0.2599,
		0.322233,
		0.386518,
		0.451507,
		0.516043,
		0.579062,
		0.639602,
		0.696813,
		0.749955,
		0.798418,
		0.841725,
		0.879542,
		0.911694,
		0.938168,
		0.959131,
		0.974928,
		0.986085,
		0.993298,
		0.997398,
		0.999302,
		0.999914,
		0.999999,
		0.999996,
		0.999836,
		0.998981,
		0.996615,
		0.991821,
		0.983695,
		0.971437,
		0.954391,
		0.932076,
		0.904191,
		0.870615,
		0.831403,
		0.78677,
		0.737086,
		0.682865,
		0.624748,
		0.563503,
		0.500011,
		0.435259,
		0.370333,
		0.306416,
		0.244776,
		0.186768,
		0.133826,
		0.087461,
		0.049255,
		0.020862,
		0.004003,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.00692,
		0.026499,
		0.057012,
		0.096799,
		0.144262,
		0.197875,
		0.256181,
		0.317799,
		0.381424,
		0.445833,
		0.509893,
		0.572558,
		0.632882,
		0.690022,
		0.743248,
		0.791947,
		0.835636,
		0.87397,
		0.906753,
		0.933948,
		0.955686,
		0.972271,
		0.984185,
		0.992073,
		0.996723,
		0.999014,
		0.999842,
		0.999996,
		1,
		0.999917,
		0.999324,
		0.997481,
		0.993506,
		0.986506,
		0.975665,
		0.960299,
		0.939885,
		0.914074,
		0.88269,
		0.845732,
		0.803355,
		0.755868,
		0.703721,
		0.647496,
		0.587893,
		0.525728,
		0.461923,
		0.397498,
		0.333566,
		0.271328,
		0.212067,
		0.157147,
		0.108006,
		0.066153,
		0.033168,
		0.010696,
		0.000447,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.006803,
		0.026062,
		0.056095,
		0.095283,
		0.142069,
		0.194962,
		0.25254,
		0.313452,
		0.376422,
		0.440255,
		0.503836,
		0.56614,
		0.626236,
		0.683291,
		0.736579,
		0.78549,
		0.829535,
		0.868358,
		0.901745,
		0.929636,
		0.952128,
		0.969489,
		0.982156,
		0.990728,
		0.995947,
		0.998657,
		0.999735,
		0.999986,
		1,
		0.999963,
		0.999572,
		0.998167,
		0.994912,
		0.988931,
		0.9794,
		0.965609,
		0.946999,
		0.923176,
		0.893914,
		0.859156,
		0.819002,
		0.773702,
		0.723646,
		0.669354,
		0.611467,
		0.550737,
		0.488022,
		0.424278,
		0.360554,
		0.297982,
		0.237781,
		0.181245,
		0.129743,
		0.084714,
		0.047667,
		0.020173,
		0.003868,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.006689,
		0.025636,
		0.0552,
		0.093803,
		0.139926,
		0.192112,
		0.248973,
		0.309189,
		0.371512,
		0.43477,
		0.497872,
		0.55981,
		0.619667,
		0.67662,
		0.729952,
		0.779051,
		0.823426,
		0.862711,
		0.896676,
		0.925236,
		0.948463,
		0.966586,
		0.980001,
		0.989263,
		0.995069,
		0.998225,
		0.999586,
		0.999965,
		1,
		0.999987,
		0.999743,
		0.9987,
		0.996073,
		0.991008,
		0.982681,
		0.970362,
		0.953459,
		0.931537,
		0.904322,
		0.871707,
		0.833736,
		0.790604,
		0.742645,
		0.690319,
		0.634207,
		0.575003,
		0.513503,
		0.450601,
		0.387283,
		0.324619,
		0.263763,
		0.205942,
		0.152459,
		0.104687,
		0.064064,
		0.032094,
		0.010342,
		0.000431,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.006578,
		0.02522,
		0.054327,
		0.092357,
		0.137829,
		0.189322,
		0.245479,
		0.305009,
		0.366691,
		0.429378,
		0.491999,
		0.553565,
		0.613173,
		0.670011,
		0.723368,
		0.772634,
		0.817314,
		0.857034,
		0.891549,
		0.920755,
		0.944695,
		0.963565,
		0.977722,
		0.987677,
		0.994084,
		0.997711,
		0.999388,
		0.999925,
		1,
		0.999997,
		0.999857,
		0.999106,
		0.99702,
		0.992774,
		0.985548,
		0.974599,
		0.959306,
		0.939196,
		0.913952,
		0.883417,
		0.847584,
		0.806596,
		0.76073,
		0.71039,
		0.656102,
		0.5985,
		0.538323,
		0.476406,
		0.413672,
		0.351133,
		0.289878,
		0.231074,
		0.17596,
		0.125842,
		0.082095,
		0.046154,
		0.019517,
		0.003739,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.00647,
		0.024814,
		0.053473,
		0.090944,
		0.135779,
		0.186592,
		0.242056,
		0.300909,
		0.361957,
		0.424076,
		0.486216,
		0.547405,
		0.606755,
		0.663465,
		0.716828,
		0.76624,
		0.811201,
		0.851331,
		0.886372,
		0.916199,
		0.94083,
		0.960431,
		0.975321,
		0.98597,
		0.992991,
		0.997111,
		0.999134,
		0.999861,
		0.999997,
		1,
		0.999927,
		0.999407,
		0.997782,
		0.994262,
		0.988039,
		0.978359,
		0.964579,
		0.946192,
		0.92284,
		0.894319,
		0.860575,
		0.821699,
		0.777914,
		0.729573,
		0.677145,
		0.621209,
		0.562448,
		0.501639,
		0.439648,
		0.377429,
		0.31601,
		0.256499,
		0.200072,
		0.147975,
		0.101517,
		0.062072,
		0.031071,
		0.010005,
		0.000417,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.006364,
		0.024418,
		0.05264,
		0.089563,
		0.133774,
		0.183919,
		0.238701,
		0.296887,
		0.357308,
		0.418863,
		0.480522,
		0.54133,
		0.600413,
		0.656981,
		0.710335,
		0.759872,
		0.805092,
		0.845607,
		0.881147,
		0.911571,
		0.936872,
		0.957188,
		0.972801,
		0.984144,
		0.991788,
		0.996422,
		0.998818,
		0.999767,
		0.999988,
		1,
		0.999968,
		0.999624,
		0.998385,
		0.995504,
		0.990188,
		0.98168,
		0.969317,
		0.952563,
		0.931022,
		0.904447,
		0.872739,
		0.835936,
		0.794215,
		0.747875,
		0.697333,
		0.643115,
		0.585848,
		0.526257,
		0.465151,
		0.403426,
		0.342055,
		0.282087,
		0.224639,
		0.170898,
		0.122113,
		0.079594,
		0.044712,
		0.018893,
		0.003617,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.006261,
		0.024031,
		0.051826,
		0.088212,
		0.131812,
		0.181302,
		0.235414,
		0.292943,
		0.352743,
		0.413737,
		0.474914,
		0.535338,
		0.594146,
		0.650561,
		0.70389,
		0.753533,
		0.79899,
		0.839866,
		0.875881,
		0.906878,
		0.932827,
		0.95384,
		0.970166,
		0.9822,
		0.990474,
		0.995638,
		0.998435,
		0.999636,
		0.999969,
		1,
		0.999988,
		0.999775,
		0.998854,
		0.996529,
		0.992029,
		0.984597,
		0.973558,
		0.958348,
		0.938536,
		0.913836,
		0.884104,
		0.849334,
		0.809651,
		0.765307,
		0.716668,
		0.664208,
		0.608503,
		0.550223,
		0.490126,
		0.429052,
		0.36792,
		0.307723,
		0.249522,
		0.194445,
		0.143683,
		0.098489,
		0.060171,
		0.030097,
		0.009684,
		0.000403,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.006161,
		0.023653,
		0.05103,
		0.086892,
		0.129893,
		0.178739,
		0.232193,
		0.289073,
		0.34826,
		0.408697,
		0.469393,
		0.529428,
		0.587956,
		0.644206,
		0.697495,
		0.747226,
		0.792898,
		0.834112,
		0.870578,
		0.902123,
		0.9287,
		0.950392,
		0.967419,
		0.98014,
		0.989048,
		0.994759,
		0.997978,
		0.999461,
		0.999934,
		1,
		0.999997,
		0.999874,
		0.999212,
		0.997366,
		0.993593,
		0.987147,
		0.977338,
		0.963582,
		0.945417,
		0.922519,
		0.894703,
		0.861918,
		0.824244,
		0.781884,
		0.735156,
		0.684484,
		0.630396,
		0.573509,
		0.514529,
		0.454246,
		0.393526,
		0.333306,
		0.274594,
		0.218463,
		0.166048,
		0.118545,
		0.077206,
		0.043337,
		0.018298,
		0.003501,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.006063,
		0.023285,
		0.050253,
		0.085601,
		0.128015,
		0.17623,
		0.229035,
		0.285276,
		0.343856,
		0.403741,
		0.463957,
		0.523601,
		0.58184,
		0.637915,
		0.69115,
		0.740951,
		0.786818,
		0.828347,
		0.865241,
		0.897313,
		0.924495,
		0.946848,
		0.964563,
		0.977965,
		0.987511,
		0.99378,
		0.997444,
		0.999236,
		0.999878,
		0.999997,
		1,
		0.999936,
		0.999477,
		0.998039,
		0.994912,
		0.989361,
		0.980692,
		0.968302,
		0.951701,
		0.930531,
		0.904566,
		0.873716,
		0.838015,
		0.79762,
		0.752804,
		0.703943,
		0.651515,
		0.596089,
		0.538323,
		0.478955,
		0.4188,
		0.358744,
		0.299743,
		0.242817,
		0.189047,
		0.139573,
		0.095593,
		0.058356,
		0.029167,
		0.009378,
		0.00039,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.005967,
		0.022924,
		0.049493,
		0.084339,
		0.126177,
		0.173771,
		0.225939,
		0.28155,
		0.339531,
		0.398867,
		0.458604,
		0.517854,
		0.575799,
		0.631689,
		0.684856,
		0.734711,
		0.780754,
		0.822577,
		0.859876,
		0.89245,
		0.920217,
		0.943213,
		0.961602,
		0.975678,
		0.985864,
		0.9927,
		0.996829,
		0.998956,
		0.999795,
		0.999989,
		1,
		0.999972,
		0.999668,
		0.998572,
		0.996013,
		0.991272,
		0.983655,
		0.972543,
		0.957423,
		0.937904,
		0.913726,
		0.884756,
		0.850988,
		0.812535,
		0.769624,
		0.722587,
		0.671853,
		0.617947,
		0.561477,
		0.503133,
		0.443682,
		0.383959,
		0.32487,
		0.267385,
		0.212531,
		0.161399,
		0.11513,
		0.074922,
		0.042023,
		0.017731,
		0.00339,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.005873,
		0.022572,
		0.04875,
		0.083104,
		0.124377,
		0.171364,
		0.222905,
		0.277895,
		0.335282,
		0.394074,
		0.453333,
		0.512188,
		0.569832,
		0.625529,
		0.678615,
		0.728508,
		0.774707,
		0.816804,
		0.854485,
		0.887541,
		0.91587,
		0.939491,
		0.95854,
		0.973282,
		0.984106,
		0.991519,
		0.996128,
		0.998614,
		0.999679,
		0.999973,
		1,
		0.99999,
		0.999801,
		0.998987,
		0.996922,
		0.992908,
		0.986257,
		0.976338,
		0.962616,
		0.944673,
		0.922214,
		0.895068,
		0.863189,
		0.826649,
		0.78563,
		0.740422,
		0.691409,
		0.639069,
		0.583966,
		0.526742,
		0.468117,
		0.40888,
		0.349886,
		0.292057,
		0.236371,
		0.183866,
		0.135635,
		0.092821,
		0.056622,
		0.02828,
		0.009086,
		0.000378,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.005782,
		0.022228,
		0.048024,
		0.081895,
		0.122616,
		0.169005,
		0.219929,
		0.274307,
		0.331109,
		0.38936,
		0.448143,
		0.5066,
		0.563939,
		0.619433,
		0.672427,
		0.722343,
		0.768681,
		0.811031,
		0.849074,
		0.882588,
		0.91146,
		0.935686,
		0.955381,
		0.97078,
		0.98224,
		0.990235,
		0.99534,
		0.998208,
		0.999523,
		0.999942,
		1,
		0.999997,
		0.999889,
		0.999303,
		0.997663,
		0.9943,
		0.988531,
		0.97972,
		0.967315,
		0.950871,
		0.930061,
		0.90468,
		0.874642,
		0.839981,
		0.800837,
		0.757456,
		0.710181,
		0.659446,
		0.60577,
		0.54975,
		0.492061,
		0.433445,
		0.374712,
		0.316735,
		0.260446,
		0.206833,
		0.156939,
		0.111859,
		0.072738,
		0.040769,
		0.01719,
		0.003284,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.005692,
		0.021892,
		0.047314,
		0.080713,
		0.120891,
		0.166693,
		0.217011,
		0.270786,
		0.327009,
		0.384724,
		0.443032,
		0.501091,
		0.55812,
		0.613403,
		0.666294,
		0.716218,
		0.762678,
		0.805262,
		0.843645,
		0.877597,
		0.90699,
		0.931803,
		0.952128,
		0.968173,
		0.980266,
		0.988848,
		0.994461,
		0.997731,
		0.999323,
		0.999892,
		0.999997,
		1,
		0.999944,
		0.999537,
		0.99826,
		0.995473,
		0.990506,
		0.982721,
		0.971552,
		0.95653,
		0.937299,
		0.913621,
		0.885375,
		0.852555,
		0.815262,
		0.773702,
		0.728175,
		0.679074,
		0.626874,
		0.572131,
		0.515474,
		0.457602,
		0.39928,
		0.341334,
		0.28465,
		0.230171,
		0.178891,
		0.131859,
		0.090168,
		0.054963,
		0.027432,
		0.008808,
		0.000366,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.005605,
		0.021564,
		0.046619,
		0.079556,
		0.119202,
		0.164428,
		0.214149,
		0.267329,
		0.32298,
		0.380165,
		0.438,
		0.495659,
		0.552374,
		0.607439,
		0.660215,
		0.710133,
		0.756699,
		0.799498,
		0.838201,
		0.872571,
		0.902465,
		0.927846,
		0.948785,
		0.965467,
		0.978187,
		0.987358,
		0.99349,
		0.99718,
		0.999074,
		0.999819,
		0.999991,
		1,
		0.999975,
		0.999707,
		0.998733,
		0.996452,
		0.99221,
		0.985372,
		0.975358,
		0.961681,
		0.943959,
		0.921922,
		0.895415,
		0.864394,
		0.828924,
		0.789171,
		0.745396,
		0.69795,
		0.647268,
		0.593863,
		0.538323,
		0.481304,
		0.423526,
		0.365774,
		0.308888,
		0.253765,
		0.201355,
		0.152659,
		0.108724,
		0.070648,
		0.039569,
		0.016673,
		0.003183,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.002221,
		0.014447,
		0.036235,
		0.066356,
		0.103621,
		0.146884,
		0.195044,
		0.247045,
		0.301882,
		0.358598,
		0.416289,
		0.474108,
		0.531263,
		0.587025,
		0.64073,
		0.691782,
		0.739658,
		0.783913,
		0.824185,
		0.860201,
		0.891784,
		0.918855,
		0.941445,
		0.959694,
		0.973858,
		0.984305,
		0.99151,
		0.996042,
		0.998532,
		0.999635,
		0.999962,
		1,
		0.999995,
		0.999861,
		0.999217,
		0.997501,
		0.994072,
		0.988279,
		0.979518,
		0.967266,
		0.951098,
		0.930704,
		0.905884,
		0.876554,
		0.842738,
		0.804566,
		0.762263,
		0.716147,
		0.666622,
		0.614173,
		0.559361,
		0.502815,
		0.445233,
		0.387376,
		0.330063,
		0.27417,
		0.220628,
		0.170419,
		0.124573,
		0.084169,
		0.05033,
		0.024223,
		0.00706,
		9.1e-005,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		0.004379,
		0.019337,
		0.043674,
		0.076143,
		0.115538,
		0.160699,
		0.210511,
		0.26391,
		0.31988,
		0.377459,
		0.435741,
		0.493875,
		0.551073,
		0.606611,
		0.659834,
		0.710157,
		0.757075,
		0.800163,
		0.839085,
		0.873601,
		0.903567,
		0.92895,
		0.949827,
		0.966392,
		0.978957,
		0.987949,
		0.993901,
		0.997428,
		0.999193,
		0.999857,
		0.999995,
		1,
		0.999961,
		0.999624,
		0.998488,
		0.995927,
		0.991269,
		0.983869,
		0.97315,
		0.958631,
		0.939942,
		0.916829,
		0.889157,
		0.856905,
		0.820162,
		0.779121,
		0.73407,
		0.685391,
		0.63355,
		0.579092,
		0.522638,
		0.464877,
		0.406567,
		0.348527,
		0.291635,
		0.236827,
		0.185091,
		0.137468,
		0.095049,
		0.058969,
		0.030412,
		0.010606,
		0.00082,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		9.3e-005,
		0.007282,
		0.024968,
		0.051842,
		0.086636,
		0.128128,
		0.175145,
		0.226559,
		0.281295,
		0.33833,
		0.396697,
		0.455485,
		0.513845,
		0.570991,
		0.626207,
		0.678846,
		0.72834,
		0.774201,
		0.816028,
		0.853514,
		0.886452,
		0.914741,
		0.938392,
		0.957534,
		0.972419,
		0.983418,
		0.991018,
		0.995807,
		0.998443,
		0.999612,
		0.99996,
		1,
		0.999995,
		0.999852,
		0.999169,
		0.997351,
		0.993724,
		0.987608,
		0.978376,
		0.965488,
		0.948514,
		0.927141,
		0.90118,
		0.870561,
		0.835329,
		0.79564,
		0.751752,
		0.704019,
		0.652886,
		0.598882,
		0.542616,
		0.484769,
		0.426093,
		0.367405,
		0.309587,
		0.253577,
		0.200371,
		0.151018,
		0.106619,
		0.068326,
		0.037337,
		0.014896,
		0.002292,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		0.000846,
		0.01094,
		0.031349,
		0.060743,
		0.097836,
		0.141391,
		0.190218,
		0.243179,
		0.299189,
		0.357216,
		0.416289,
		0.475496,
		0.533987,
		0.590984,
		0.645775,
		0.697728,
		0.746289,
		0.790993,
		0.831466,
		0.867431,
		0.89872,
		0.925275,
		0.947158,
		0.964555,
		0.977776,
		0.987255,
		0.993541,
		0.997272,
		0.999143,
		0.999848,
		0.999994,
		1,
		0.999959,
		0.9996,
		0.998396,
		0.995683,
		0.990759,
		0.982951,
		0.971663,
		0.956401,
		0.936792,
		0.912588,
		0.883666,
		0.850025,
		0.811778,
		0.769149,
		0.722464,
		0.672143,
		0.618696,
		0.562719,
		0.504881,
		0.44593,
		0.386681,
		0.328012,
		0.270869,
		0.216251,
		0.165218,
		0.118882,
		0.078406,
		0.045004,
		0.019939,
		0.004519,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.002366,
		0.015366,
		0.038489,
		0.070384,
		0.109748,
		0.155325,
		0.205913,
		0.260362,
		0.317576,
		0.376519,
		0.436213,
		0.495746,
		0.554272,
		0.611015,
		0.665277,
		0.716437,
		0.763962,
		0.80741,
		0.846434,
		0.880796,
		0.910368,
		0.935141,
		0.95523,
		0.970881,
		0.982468,
		0.99049,
		0.995554,
		0.998347,
		0.999588,
		0.999957,
		1,
		0.999994,
		0.999843,
		0.999117,
		0.997189,
		0.99335,
		0.986889,
		0.977154,
		0.96359,
		0.945759,
		0.92335,
		0.896184,
		0.864209,
		0.827492,
		0.78622,
		0.740684,
		0.69128,
		0.638497,
		0.582912,
		0.525185,
		0.466054,
		0.406331,
		0.346894,
		0.28869,
		0.232725,
		0.180066,
		0.131836,
		0.089212,
		0.053423,
		0.025748,
		0.007514,
		9.6e-005,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		0.004665,
		0.02057,
		0.046396,
		0.08077,
		0.122371,
		0.169927,
		0.222223,
		0.278096,
		0.336442,
		0.396218,
		0.456443,
		0.516206,
		0.574663,
		0.631048,
		0.684672,
		0.734932,
		0.781314,
		0.823405,
		0.86089,
		0.893571,
		0.921364,
		0.944313,
		0.962592,
		0.976511,
		0.98651,
		0.993153,
		0.997104,
		0.999089,
		0.999838,
		0.999994,
		1,
		0.999956,
		0.999575,
		0.998296,
		0.995421,
		0.990212,
		0.981968,
		0.970072,
		0.95402,
		0.933436,
		0.908079,
		0.877839,
		0.842738,
		0.802918,
		0.758636,
		0.710256,
		0.658245,
		0.60316,
		0.545648,
		0.486437,
		0.426333,
		0.366214,
		0.307026,
		0.249783,
		0.195557,
		0.145482,
		0.100747,
		0.062598,
		0.03233,
		0.01129,
		0.000874,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.0001,
		0.007758,
		0.026564,
		0.055077,
		0.091904,
		0.135705,
		0.185194,
		0.239139,
		0.296367,
		0.355767,
		0.416289,
		0.476952,
		0.536843,
		0.595126,
		0.651042,
		0.703917,
		0.753166,
		0.7983,
		0.838934,
		0.874792,
		0.905717,
		0.931675,
		0.952769,
		0.969235,
		0.98145,
		0.989923,
		0.995282,
		0.998243,
		0.999561,
		0.999955,
		1,
		0.999994,
		0.999833,
		0.999061,
		0.997015,
		0.992949,
		0.986118,
		0.975845,
		0.96156,
		0.942819,
		0.919314,
		0.890876,
		0.857473,
		0.8192,
		0.776274,
		0.729027,
		0.677898,
		0.623424,
		0.566235,
		0.507048,
		0.446662,
		0.385951,
		0.325863,
		0.267414,
		0.211684,
		0.159817,
		0.113014,
		0.072536,
		0.039695,
		0.015859,
		0.002443,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.000903,
		0.011657,
		0.033357,
		0.064539,
		0.103789,
		0.14975,
		0.201118,
		0.25665,
		0.315161,
		0.375532,
		0.436709,
		0.497709,
		0.557624,
		0.615622,
		0.670956,
		0.722968,
		0.771094,
		0.814873,
		0.853952,
		0.888097,
		0.917197,
		0.941275,
		0.960493,
		0.975156,
		0.985711,
		0.992736,
		0.996923,
		0.999031,
		0.999827,
		0.999994,
		1,
		0.999953,
		0.999547,
		0.998188,
		0.995138,
		0.989624,
		0.980913,
		0.968369,
		0.951476,
		0.929857,
		0.90328,
		0.871652,
		0.835018,
		0.793551,
		0.747548,
		0.697414,
		0.643665,
		0.586911,
		0.527856,
		0.46729,
		0.406083,
		0.345181,
		0.285604,
		0.228439,
		0.174837,
		0.126014,
		0.083242,
		0.047852,
		0.021232,
		0.004818,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.002525,
		0.016376,
		0.040959,
		0.074788,
		0.116428,
		0.164503,
		0.217694,
		0.274744,
		0.33446,
		0.395714,
		0.45745,
		0.518684,
		0.578511,
		0.636109,
		0.690745,
		0.741777,
		0.788668,
		0.830985,
		0.868415,
		0.900765,
		0.927979,
		0.950139,
		0.967473,
		0.980357,
		0.989313,
		0.994989,
		0.998131,
		0.999532,
		0.999952,
		1,
		0.999993,
		0.999822,
		0.999,
		0.996827,
		0.992516,
		0.985289,
		0.974442,
		0.959389,
		0.93968,
		0.915012,
		0.885232,
		0.850326,
		0.810421,
		0.76577,
		0.716747,
		0.66384,
		0.607637,
		0.548826,
		0.488188,
		0.426585,
		0.364963,
		0.304341,
		0.245813,
		0.190539,
		0.139745,
		0.094719,
		0.056809,
		0.02742,
		0.008014,
		0.000103,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.004979,
		0.021925,
		0.049379,
		0.085828,
		0.12982,
		0.179959,
		0.234912,
		0.293408,
		0.354245,
		0.416289,
		0.478481,
		0.53984,
		0.599465,
		0.656545,
		0.710363,
		0.760297,
		0.805839,
		0.846589,
		0.882275,
		0.912756,
		0.938031,
		0.958246,
		0.973702,
		0.984851,
		0.992288,
		0.996728,
		0.998968,
		0.999816,
		0.999993,
		1,
		0.99995,
		0.999517,
		0.998072,
		0.994834,
		0.988991,
		0.979781,
		0.966545,
		0.948756,
		0.926039,
		0.89817,
		0.865077,
		0.826833,
		0.783645,
		0.73585,
		0.683903,
		0.628371,
		0.569922,
		0.509323,
		0.44743,
		0.385184,
		0.323607,
		0.263794,
		0.206915,
		0.154206,
		0.10697,
		0.066571,
		0.034434,
		0.012042,
		0.000934,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.000107,
		0.008282,
		0.028318,
		0.058623,
		0.097664,
		0.143965,
		0.196114,
		0.252761,
		0.312626,
		0.374495,
		0.43723,
		0.499772,
		0.561141,
		0.620445,
		0.676885,
		0.729761,
		0.778478,
		0.822556,
		0.861636,
		0.895491,
		0.924033,
		0.947325,
		0.965583,
		0.979184,
		0.988657,
		0.994673,
		0.99801,
		0.999501,
		0.999948,
		1,
		0.999993,
		0.99981,
		0.998934,
		0.996625,
		0.99205,
		0.984397,
		0.972936,
		0.957064,
		0.936326,
		0.910426,
		0.879226,
		0.842738,
		0.801123,
		0.754672,
		0.703808,
		0.649069,
		0.591103,
		0.530661,
		0.468589,
		0.405822,
		0.343382,
		0.282369,
		0.223957,
		0.169394,
		0.119997,
		0.077145,
		0.042284,
		0.016918,
		0.00261,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.000966,
		0.012447,
		0.035564,
		0.0687,
		0.110298,
		0.158861,
		0.212959,
		0.27123,
		0.332378,
		0.395185,
		0.458508,
		0.521286,
		0.582546,
		0.641405,
		0.697079,
		0.748889,
		0.796268,
		0.83877,
		0.876079,
		0.908019,
		0.934562,
		0.955839,
		0.972142,
		0.983926,
		0.991804,
		0.996517,
		0.9989,
		0.999803,
		0.999993,
		1,
		0.999947,
		0.999485,
		0.997946,
		0.994505,
		0.98831,
		0.978564,
		0.964588,
		0.945845,
		0.921961,
		0.892725,
		0.858088,
		0.818151,
		0.773163,
		0.723506,
		0.669685,
		0.612327,
		0.552163,
		0.490028,
		0.42685,
		0.363647,
		0.301522,
		0.241656,
		0.185305,
		0.133799,
		0.088536,
		0.050979,
		0.022654,
		0.005149,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.0027,
		0.017488,
		0.043674,
		0.079615,
		0.12373,
		0.174504,
		0.230487,
		0.290302,
		0.352645,
		0.416289,
		0.48009,
		0.542987,
		0.604014,
		0.6623,
		0.717079,
		0.767695,
		0.813614,
		0.854428,
		0.889869,
		0.919818,
		0.944313,
		0.963556,
		0.977922,
		0.987949,
		0.994331,
		0.997879,
		0.999468,
		0.999945,
		1,
		0.999993,
		0.999797,
		0.998863,
		0.996405,
		0.991548,
		0.983438,
		0.971319,
		0.954571,
		0.932739,
		0.905532,
		0.872831,
		0.834679,
		0.79127,
		0.742944,
		0.690172,
		0.63355,
		0.573791,
		0.511714,
		0.448238,
		0.384378,
		0.321236,
		0.259999,
		0.201932,
		0.148377,
		0.100747,
		0.060526,
		0.029261,
		0.008564,
		0.00011,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.005327,
		0.023419,
		0.052658,
		0.091374,
		0.137962,
		0.190888,
		0.248685,
		0.309961,
		0.373403,
		0.437779,
		0.501941,
		0.564834,
		0.625499,
		0.68308,
		0.736831,
		0.786125,
		0.830462,
		0.869479,
		0.902962,
		0.930852,
		0.953258,
		0.970465,
		0.982931,
		0.991281,
		0.996289,
		0.998826,
		0.99979,
		0.999992,
		1,
		0.999943,
		0.99945,
		0.997809,
		0.99415,
		0.987575,
		0.977255,
		0.962487,
		0.942727,
		0.917603,
		0.88692,
		0.850653,
		0.808939,
		0.762069,
		0.710476,
		0.654724,
		0.595501,
		0.533608,
		0.469955,
		0.405548,
		0.34149,
		0.278972,
		0.219266,
		0.163726,
		0.113781,
		0.070933,
		0.036751,
		0.012873,
		0.000999,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.000114,
		0.008861,
		0.030251,
		0.062522,
		0.103978,
		0.152992,
		0.208006,
		0.26754,
		0.330188,
		0.394628,
		0.459622,
		0.524024,
		0.586783,
		0.646952,
		0.703692,
		0.75628,
		0.804122,
		0.846758,
		0.883873,
		0.915312,
		0.941085,
		0.961379,
		0.976563,
		0.987186,
		0.993962,
		0.997737,
		0.999431,
		0.999941,
		1,
		0.999992,
		0.999782,
		0.998786,
		0.996168,
		0.991005,
		0.982404,
		0.96958,
		0.951898,
		0.928899,
		0.900305,
		0.866019,
		0.826114,
		0.780826,
		0.730545,
		0.675798,
		0.617247,
		0.555671,
		0.491965,
		0.427129,
		0.362263,
		0.298559,
		0.237298,
		0.179843,
		0.127639,
		0.082205,
		0.045134,
		0.018088,
		0.002795,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.001035,
		0.01332,
		0.037997,
		0.073275,
		0.117431,
		0.168815,
		0.225849,
		0.287037,
		0.35096,
		0.416289,
		0.481783,
		0.546297,
		0.608789,
		0.668323,
		0.724081,
		0.77537,
		0.82163,
		0.862446,
		0.897559,
		0.926878,
		0.950488,
		0.968661,
		0.981857,
		0.990717,
		0.996042,
		0.998745,
		0.999775,
		0.999992,
		1,
		0.999939,
		0.999412,
		0.997662,
		0.993766,
		0.986782,
		0.975845,
		0.96023,
		0.939384,
		0.912942,
		0.880725,
		0.842738,
		0.799158,
		0.750323,
		0.69672,
		0.638979,
		0.577857,
		0.51423,
		0.449089,
		0.383529,
		0.318742,
		0.256014,
		0.196721,
		0.14232,
		0.094348,
		0.054421,
		0.024223,
		0.005514,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.002895,
		0.018718,
		0.046668,
		0.084923,
		0.131734,
		0.185427,
		0.244406,
		0.307157,
		0.372254,
		0.438357,
		0.504225,
		0.568718,
		0.630801,
		0.689557,
		0.744191,
		0.794042,
		0.838591,
		0.877473,
		0.91049,
		0.937623,
		0.959038,
		0.9751,
		0.986362,
		0.993563,
		0.997583,
		0.999391,
		0.999937,
		1,
		0.999991,
		0.999767,
		0.998703,
		0.995911,
		0.990418,
		0.981289,
		0.967708,
		0.949027,
		0.924786,
		0.89472,
		0.858756,
		0.817006,
		0.769751,
		0.717434,
		0.660647,
		0.60012,
		0.536711,
		0.471394,
		0.405259,
		0.339498,
		0.275401,
		0.214352,
		0.157821,
		0.107365,
		0.064618,
		0.031293,
		0.009174,
		0.000118,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.005711,
		0.02507,
		0.056273,
		0.09747,
		0.146884,
		0.20282,
		0.263662,
		0.327881,
		0.39404,
		0.460796,
		0.526907,
		0.591237,
		0.652767,
		0.710598,
		0.763962,
		0.812236,
		0.854945,
		0.891784,
		0.92262,
		0.947512,
		0.966719,
		0.980698,
		0.990107,
		0.995774,
		0.998658,
		0.999759,
		0.999991,
		1,
		0.999934,
		0.99937,
		0.997501,
		0.99335,
		0.985926,
		0.974325,
		0.957802,
		0.935798,
		0.907953,
		0.874111,
		0.834309,
		0.788769,
		0.737881,
		0.682196,
		0.622411,
		0.559361,
		0.494005,
		0.427424,
		0.360804,
		0.295441,
		0.232725,
		0.17414,
		0.121258,
		0.075735,
		0.039308,
		0.013791,
		0.001072,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.000123,
		0.009504,
		0.032389,
		0.066821,
		0.110919,
		0.16288,
		0.220985,
		0.283601,
		0.349184,
		0.416289,
		0.483569,
		0.549783,
		0.613805,
		0.674631,
		0.731385,
		0.783333,
		0.82989,
		0.870636,
		0.905327,
		0.933906,
		0.95652,
		0.973521,
		0.985471,
		0.99313,
		0.997416,
		0.999348,
		0.999932,
		1,
		0.999991,
		0.99975,
		0.998612,
		0.995632,
		0.989783,
		0.980084,
		0.965692,
		0.945942,
		0.920376,
		0.888746,
		0.851009,
		0.807315,
		0.758,
		0.703566,
		0.644676,
		0.582134,
		0.516882,
		0.449987,
		0.382633,
		0.316114,
		0.251827,
		0.191269,
		0.136026,
		0.087777,
		0.048281,
		0.019382,
		0.003,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.001112,
		0.014288,
		0.040688,
		0.07832,
		0.125272,
		0.179716,
		0.23991,
		0.304203,
		0.371041,
		0.438967,
		0.506634,
		0.572806,
		0.636368,
		0.696335,
		0.751857,
		0.802239,
		0.846941,
		0.885604,
		0.918052,
		0.944313,
		0.964625,
		0.979446,
		0.989445,
		0.995483,
		0.998563,
		0.999741,
		0.99999,
		1,
		0.999929,
		0.999325,
		0.997327,
		0.992899,
		0.984999,
		0.972685,
		0.955189,
		0.931946,
		0.902608,
		0.867043,
		0.825327,
		0.777728,
		0.724699,
		0.666857,
		0.604978,
		0.539979,
		0.472913,
		0.404954,
		0.337397,
		0.271644,
		0.2092,
		0.151669,
		0.100747,
		0.058222,
		0.025962,
		0.00592,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.003111,
		0.020082,
		0.049979,
		0.090775,
		0.140529,
		0.197385,
		0.259581,
		0.325448,
		0.39342,
		0.462036,
		0.529947,
		0.595924,
		0.658868,
		0.717815,
		0.771949,
		0.820615,
		0.863328,
		0.899793,
		0.929913,
		0.953807,
		0.971816,
		0.984507,
		0.99266,
		0.997234,
		0.999301,
		0.999927,
		1,
		0.99999,
		0.999732,
		0.998512,
		0.995329,
		0.989094,
		0.978782,
		0.963517,
		0.942622,
		0.915644,
		0.882352,
		0.842738,
		0.797,
		0.745529,
		0.688896,
		0.627839,
		0.563249,
		0.496159,
		0.427734,
		0.359265,
		0.292156,
		0.227922,
		0.168182,
		0.114651,
		0.069138,
		0.033544,
		0.009851,
		0.000127,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.00614,
		0.026902,
		0.060273,
		0.104192,
		0.156687,
		0.215878,
		0.27998,
		0.34731,
		0.416289,
		0.485454,
		0.553458,
		0.619082,
		0.681244,
		0.739008,
		0.791594,
		0.838394,
		0.878987,
		0.913148,
		0.940869,
		0.962366,
		0.978092,
		0.988728,
		0.995167,
		0.99846,
		0.999722,
		0.99999,
		1,
		0.999924,
		0.999276,
		0.997137,
		0.99241,
		0.983995,
		0.970913,
		0.952372,
		0.927805,
		0.896877,
		0.859486,
		0.815748,
		0.76599,
		0.710728,
		0.650658,
		0.586639,
		0.51968,
		0.450936,
		0.381687,
		0.313342,
		0.247423,
		0.185559,
		0.129487,
		0.081039,
		0.042142,
		0.014812,
		0.001154,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.000132,
		0.010218,
		0.034761,
		0.071577,
		0.118572,
		0.173742,
		0.235182,
		0.301087,
		0.369759,
		0.439612,
		0.509177,
		0.577116,
		0.642221,
		0.703432,
		0.759844,
		0.810721,
		0.855511,
		0.893857,
		0.925619,
		0.950882,
		0.969973,
		0.983462,
		0.992149,
		0.997036,
		0.999249,
		0.999921,
		1,
		0.999989,
		0.999712,
		0.998404,
		0.994999,
		0.988347,
		0.977373,
		0.961169,
		0.939048,
		0.910561,
		0.875503,
		0.833904,
		0.786014,
		0.732288,
		0.673374,
		0.610092,
		0.543429,
		0.474517,
		0.404632,
		0.335179,
		0.267685,
		0.203793,
		0.145256,
		0.09393,
		0.051768,
		0.020821,
		0.003228,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.001198,
		0.015366,
		0.043674,
		0.083901,
		0.133916,
		0.191686,
		0.255282,
		0.322879,
		0.392764,
		0.463346,
		0.533157,
		0.600863,
		0.665277,
		0.725363,
		0.780253,
		0.829262,
		0.871897,
		0.907879,
		0.937158,
		0.959925,
		0.976625,
		0.987949,
		0.994823,
		0.998347,
		0.999701,
		0.999989,
		1,
		0.999918,
		0.999222,
		0.99693,
		0.991877,
		0.982906,
		0.968996,
		0.949334,
		0.92335,
		0.890728,
		0.851398,
		0.805528,
		0.753505,
		0.695919,
		0.63355,
		0.567351,
		0.498434,
		0.428063,
		0.357638,
		0.28869,
		0.222873,
		0.161955,
		0.107814,
		0.062433,
		0.027894,
		0.006372,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.003352,
		0.021601,
		0.053654,
		0.09725,
		0.150223,
		0.210511,
		0.27616,
		0.345327,
		0.416289,
		0.487448,
		0.557338,
		0.62464,
		0.688183,
		0.746966,
		0.800163,
		0.847143,
		0.887485,
		0.920996,
		0.947725,
		0.967979,
		0.982327,
		0.991594,
		0.99682,
		0.999193,
		0.999915,
		1,
		0.999988,
		0.99969,
		0.998286,
		0.99464,
		0.987535,
		0.975845,
		0.958631,
		0.935195,
		0.905097,
		0.868161,
		0.824461,
		0.774307,
		0.718226,
		0.656948,
		0.591389,
		0.522638,
		0.451939,
		0.380687,
		0.310415,
		0.242783,
		0.179577,
		0.122695,
		0.074147,
		0.036046,
		0.010606,
		0.000137,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.006618,
		0.028941,
		0.064712,
		0.111627,
		0.167488,
		0.230203,
		0.297795,
		0.368403,
		0.440294,
		0.511868,
		0.581665,
		0.64838,
		0.710869,
		0.768167,
		0.819497,
		0.864293,
		0.902212,
		0.933156,
		0.957286,
		0.975034,
		0.987103,
		0.994448,
		0.998223,
		0.999678,
		0.999988,
		1,
		0.999912,
		0.999163,
		0.996704,
		0.991297,
		0.981724,
		0.96692,
		0.946052,
		0.918552,
		0.884124,
		0.842738,
		0.794618,
		0.74022,
		0.680218,
		0.615483,
		0.547074,
		0.476215,
		0.404291,
		0.332833,
		0.263508,
		0.198115,
		0.138573,
		0.086916,
		0.045292,
		0.015951,
		0.001245,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.000143,
		0.011016,
		0.037403,
		0.076857,
		0.127035,
		0.185705,
		0.250747,
		0.32016,
		0.39207,
		0.464734,
		0.536552,
		0.606074,
		0.672015,
		0.73326,
		0.788888,
		0.838179,
		0.880639,
		0.916014,
		0.944313,
		0.965817,
		0.981094,
		0.990988,
		0.996584,
		0.999131,
		0.999909,
		1,
		0.999988,
		0.999666,
		0.998157,
		0.994248,
		0.986651,
		0.974188,
		0.955885,
		0.931037,
		0.899218,
		0.860284,
		0.814362,
		0.761826,
		0.703287,
		0.639566,
		0.571684,
		0.500843,
		0.428411,
		0.355915,
		0.285028,
		0.217558,
		0.155444,
		0.100747,
		0.055645,
		0.022426,
		0.003484,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.001294,
		0.01657,
		0.047001,
		0.090096,
		0.143473,
		0.204865,
		0.272124,
		0.343228,
		0.416289,
		0.489559,
		0.561441,
		0.630499,
		0.69547,
		0.755278,
		0.809049,
		0.856131,
		0.896111,
		0.928835,
		0.954428,
		0.973306,
		0.986181,
		0.994038,
		0.998088,
		0.999653,
		0.999987,
		1,
		0.999905,
		0.999098,
		0.996457,
		0.990665,
		0.980438,
		0.964669,
		0.942504,
		0.913379,
		0.877025,
		0.833458,
		0.782964,
		0.726079,
		0.663568,
		0.596407,
		0.525768,
		0.453002,
		0.379627,
		0.307317,
		0.237891,
		0.173305,
		0.115644,
		0.067117,
		0.030049,
		0.006878,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.003623,
		0.023299,
		0.057749,
		0.104436,
		0.160938,
		0.224955,
		0.294311,
		0.366965,
		0.441017,
		0.514717,
		0.586474,
		0.654869,
		0.718669,
		0.776841,
		0.828569,
		0.873276,
		0.910641,
		0.940621,
		0.963472,
		0.979753,
		0.990328,
		0.996325,
		0.999064,
		0.999901,
		1,
		0.999987,
		0.999639,
		0.998016,
		0.993819,
		0.985689,
		0.972387,
		0.95291,
		0.926546,
		0.892886,
		0.851827,
		0.803552,
		0.748515,
		0.687413,
		0.621173,
		0.550931,
		0.478015,
		0.40393,
		0.330348,
		0.259096,
		0.192146,
		0.131607,
		0.079717,
		0.038838,
		0.011451,
		0.000148,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.007154,
		0.031222,
		0.069658,
		0.119879,
		0.179424,
		0.245958,
		0.317281,
		0.391333,
		0.466205,
		0.540147,
		0.61158,
		0.679106,
		0.741528,
		0.797864,
		0.847365,
		0.889537,
		0.924165,
		0.95133,
		0.971428,
		0.985175,
		0.99359,
		0.99794,
		0.999625,
		0.999986,
		1,
		0.999897,
		0.999027,
		0.996187,
		0.989975,
		0.979038,
		0.962225,
		0.938663,
		0.907796,
		0.869387,
		0.823506,
		0.770509,
		0.711022,
		0.645911,
		0.576269,
		0.503396,
		0.42878,
		0.354089,
		0.281153,
		0.211958,
		0.148635,
		0.093452,
		0.048808,
		0.017226,
		0.001347,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.000155,
		0.011912,
		0.040357,
		0.082738,
		0.136427,
		0.198921,
		0.267854,
		0.341002,
		0.416289,
		0.4918,
		0.565786,
		0.636686,
		0.70313,
		0.763962,
		0.81826,
		0.865352,
		0.904839,
		0.936624,
		0.960924,
		0.978292,
		0.989606,
		0.996042,
		0.998989,
		0.999893,
		1,
		0.999985,
		0.99961,
		0.99786,
		0.99335,
		0.984638,
		0.970427,
		0.949683,
		0.921688,
		0.886059,
		0.842738,
		0.791976,
		0.734311,
		0.670542,
		0.601712,
		0.529086,
		0.45413,
		0.378502,
		0.304035,
		0.232725,
		0.166724,
		0.108329,
		0.059973,
		0.024223,
		0.003771,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.001403,
		0.017921,
		0.050721,
		0.096998,
		0.154076,
		0.219416,
		0.290619,
		0.365439,
		0.441786,
		0.51774,
		0.591564,
		0.661713,
		0.726854,
		0.785882,
		0.837941,
		0.882447,
		0.919112,
		0.947967,
		0.969383,
		0.984077,
		0.993099,
		0.997777,
		0.999595,
		0.999985,
		1,
		0.999889,
		0.998949,
		0.99589,
		0.98922,
		0.977512,
		0.959568,
		0.934501,
		0.901765,
		0.861163,
		0.812825,
		0.757191,
		0.694984,
		0.627186,
		0.55502,
		0.479925,
		0.403546,
		0.327712,
		0.254429,
		0.185867,
		0.124349,
		0.072345,
		0.032464,
		0.007447,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.003928,
		0.025204,
		0.062328,
		0.112441,
		0.172822,
		0.240892,
		0.314225,
		0.39055,
		0.467769,
		0.543963,
		0.617404,
		0.686577,
		0.750188,
		0.807193,
		0.856815,
		0.898568,
		0.932289,
		0.958153,
		0.976697,
		0.988816,
		0.995731,
		0.998907,
		0.999884,
		1,
		0.999984,
		0.999578,
		0.997689,
		0.992836,
		0.98349,
		0.968292,
		0.946178,
		0.91643,
		0.878693,
		0.832964,
		0.779572,
		0.719149,
		0.65261,
		0.581127,
		0.506107,
		0.429172,
		0.352149,
		0.277047,
		0.206051,
		0.141513,
		0.085934,
		0.041967,
		0.012401,
		0.000161,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.007758,
		0.033782,
		0.07519,
		0.12907,
		0.192655,
		0.263329,
		0.338636,
		0.416289,
		0.49418,
		0.570395,
		0.643226,
		0.711187,
		0.773038,
		0.827802,
		0.874792,
		0.913638,
		0.944313,
		0.967154,
		0.982876,
		0.992561,
		0.997598,
		0.999561,
		0.999984,
		1,
		0.99988,
		0.998863,
		0.995564,
		0.988393,
		0.975845,
		0.956676,
		0.929985,
		0.895243,
		0.8523,
		0.801356,
		0.742944,
		0.677898,
		0.607332,
		0.532609,
		0.455329,
		0.377306,
		0.300551,
		0.227264,
		0.159817,
		0.100747,
		0.052748,
		0.01866,
		0.001462,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.000168,
		0.012921,
		0.043674,
		0.089317,
		0.146884,
		0.213563,
		0.286701,
		0.363815,
		0.442603,
		0.520954,
		0.596961,
		0.66894,
		0.735449,
		0.795304,
		0.84761,
		0.891784,
		0.927583,
		0.955135,
		0.974954,
		0.987949,
		0.995389,
		0.998817,
		0.999874,
		1,
		0.999983,
		0.999543,
		0.997501,
		0.992272,
		0.982234,
		0.965963,
		0.942367,
		0.910731,
		0.870737,
		0.822445,
		0.766271,
		0.702959,
		0.63355,
		0.559361,
		0.481957,
		0.403137,
		0.324909,
		0.249484,
		0.179254,
		0.11679,
		0.064824,
		0.026246,
		0.004095,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.001526,
		0.019444,
		0.054899,
		0.104717,
		0.165879,
		0.235528,
		0.310976,
		0.389717,
		0.469434,
		0.548017,
		0.623575,
		0.694455,
		0.759263,
		0.816885,
		0.86652,
		0.907703,
		0.940336,
		0.96472,
		0.981561,
		0.991969,
		0.9974,
		0.999524,
		0.999982,
		1,
		0.999869,
		0.998768,
		0.995205,
		0.987485,
		0.974023,
		0.953525,
		0.925079,
		0.888183,
		0.842738,
		0.78903,
		0.727694,
		0.659694,
		0.586284,
		0.508991,
		0.42959,
		0.350084,
		0.272688,
		0.199814,
		0.134062,
		0.078205,
		0.035182,
		0.008089,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.004273,
		0.027353,
		0.067471,
		0.121393,
		0.186045,
		0.258527,
		0.336118,
		0.416289,
		0.496716,
		0.575292,
		0.650148,
		0.71967,
		0.782523,
		0.837676,
		0.884434,
		0.922467,
		0.951842,
		0.973047,
		0.986998,
		0.995012,
		0.998717,
		0.999863,
		1,
		0.999981,
		0.999504,
		0.997294,
		0.991651,
		0.980856,
		0.963419,
		0.938217,
		0.904547,
		0.862134,
		0.811114,
		0.752002,
		0.685666,
		0.613292,
		0.536358,
		0.456608,
		0.376031,
		0.296847,
		0.221482,
		0.152563,
		0.092903,
		0.045488,
		0.013475,
		0.000176,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.008442,
		0.03667,
		0.081404,
		0.139346,
		0.207371,
		0.282534,
		0.362085,
		0.443475,
		0.524376,
		0.602692,
		0.676582,
		0.74448,
		0.805122,
		0.857572,
		0.901258,
		0.936004,
		0.962058,
		0.980117,
		0.991318,
		0.997182,
		0.999482,
		0.999981,
		1,
		0.999858,
		0.998663,
		0.994809,
		0.986488,
		0.972026,
		0.950084,
		0.919743,
		0.880531,
		0.832415,
		0.775775,
		0.711367,
		0.640295,
		0.563978,
		0.484124,
		0.402701,
		0.321925,
		0.244237,
		0.172287,
		0.108924,
		0.057184,
		0.02028,
		0.001593,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.000183,
		0.014064,
		0.047417,
		0.096707,
		0.158573,
		0.229837,
		0.307516,
		0.388827,
		0.47121,
		0.552335,
		0.630124,
		0.702773,
		0.768774,
		0.826947,
		0.876466,
		0.9169,
		0.948246,
		0.970957,
		0.985952,
		0.994596,
		0.998606,
		0.999851,
		1,
		0.99998,
		0.99946,
		0.997064,
		0.990967,
		0.979343,
		0.960634,
		0.933692,
		0.897829,
		0.852824,
		0.798899,
		0.736686,
		0.667192,
		0.591766,
		0.512066,
		0.430036,
		0.347883,
		0.268054,
		0.193222,
		0.126271,
		0.070284,
		0.028532,
		0.004462,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.001665,
		0.02117,
		0.059613,
		0.113386,
		0.179066,
		0.253422,
		0.333432,
		0.416289,
		0.499421,
		0.580504,
		0.657487,
		0.72861,
		0.792437,
		0.847883,
		0.894253,
		0.931276,
		0.959144,
		0.978531,
		0.990599,
		0.99694,
		0.999437,
		0.999979,
		1,
		0.999845,
		0.998546,
		0.994372,
		0.985389,
		0.969836,
		0.946324,
		0.913932,
		0.87223,
		0.82126,
		0.761512,
		0.693878,
		0.619625,
		0.540354,
		0.457972,
		0.374671,
		0.292901,
		0.215353,
		0.144944,
		0.084802,
		0.038254,
		0.008818,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.004665,
		0.029788,
		0.073275,
		0.131445,
		0.200812,
		0.278096,
		0.360237,
		0.444407,
		0.528027,
		0.608789,
		0.684672,
		0.753975,
		0.815347,
		0.867815,
		0.910833,
		0.944313,
		0.968661,
		0.984799,
		0.994136,
		0.998483,
		0.999838,
		1,
		0.999978,
		0.999412,
		0.99681,
		0.990212,
		0.977679,
		0.957583,
		0.928752,
		0.890522,
		0.842738,
		0.785723,
		0.720239,
		0.647454,
		0.568899,
		0.486437,
		0.402235,
		0.318742,
		0.238661,
		0.164939,
		0.100747,
		0.049471,
		0.014693,
		0.000192,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.00922,
		0.039943,
		0.088416,
		0.150882,
		0.223793,
		0.303823,
		0.387877,
		0.473108,
		0.556941,
		0.637084,
		0.711563,
		0.778747,
		0.837381,
		0.886629,
		0.926112,
		0.955947,
		0.976785,
		0.989805,
		0.996672,
		0.999385,
		0.999977,
		1,
		0.99983,
		0.998417,
		0.993888,
		0.984178,
		0.967429,
		0.942207,
		0.907596,
		0.863213,
		0.809195,
		0.746155,
		0.675142,
		0.597605,
		0.51535,
		0.430513,
		0.345531,
		0.263117,
		0.186245,
		0.118126,
		0.0622,
		0.022121,
		0.001742,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.000201,
		0.015366,
		0.05166,
		0.105044,
		0.171689,
		0.247986,
		0.330561,
		0.416289,
		0.502313,
		0.586062,
		0.665277,
		0.738037,
		0.802795,
		0.858415,
		0.904213,
		0.940003,
		0.966137,
		0.983525,
		0.993626,
		0.998347,
		0.999823,
		1,
		0.999976,
		0.999358,
		0.996527,
		0.989376,
		0.975845,
		0.954233,
		0.92335,
		0.882566,
		0.831801,
		0.771497,
		0.70257,
		0.626365,
		0.544621,
		0.459432,
		0.373216,
		0.28869,
		0.208847,
		0.13694,
		0.076461,
		0.031129,
		0.004882,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.001825,
		0.023136,
		0.064959,
		0.123166,
		0.193856,
		0.273359,
		0.358259,
		0.445405,
		0.531932,
		0.615286,
		0.693246,
		0.763962,
		0.825987,
		0.878322,
		0.920461,
		0.952436,
		0.974858,
		0.988925,
		0.996374,
		0.999328,
		0.999974,
		1,
		0.999814,
		0.998273,
		0.99335,
		0.982839,
		0.96478,
		0.937694,
		0.900677,
		0.85341,
		0.796135,
		0.72961,
		0.655065,
		0.574153,
		0.488913,
		0.401737,
		0.315337,
		0.232725,
		0.157186,
		0.092264,
		0.041747,
		0.00965,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.005114,
		0.032563,
		0.079857,
		0.142782,
		0.217363,
		0.299874,
		0.386858,
		0.475143,
		0.561865,
		0.644493,
		0.720862,
		0.789202,
		0.848188,
		0.89698,
		0.935275,
		0.963355,
		0.982117,
		0.99306,
		0.998194,
		0.999806,
		1,
		0.999973,
		0.999297,
		0.996213,
		0.98845,
		0.973821,
		0.950551,
		0.917436,
		0.87389,
		0.819929,
		0.756129,
		0.683581,
		0.603835,
		0.518865,
		0.431024,
		0.343012,
		0.25785,
		0.178855,
		0.10962,
		0.053998,
		0.016085,
		0.000211,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.010111,
		0.043674,
		0.096367,
		0.163887,
		0.242186,
		0.327485,
		0.416289,
		0.505413,
		0.592002,
		0.673559,
		0.747985,
		0.813614,
		0.869259,
		0.914268,
		0.948572,
		0.97273,
		0.987949,
		0.996042,
		0.999265,
		0.999972,
		1,
		0.999797,
		0.998111,
		0.992753,
		0.981357,
		0.961859,
		0.932739,
		0.893112,
		0.842738,
		0.781985,
		0.711778,
		0.63355,
		0.549189,
		0.460998,
		0.371655,
		0.284186,
		0.201932,
		0.128533,
		0.067904,
		0.024223,
		0.001913,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.000221,
		0.016857,
		0.056497,
		0.114498,
		0.186469,
		0.268292,
		0.356136,
		0.446476,
		0.536117,
		0.622222,
		0.702346,
		0.77447,
		0.837049,
		0.889065,
		0.930079,
		0.960286,
		0.980556,
		0.992429,
		0.998024,
		0.999787,
		1,
		0.999971,
		0.99923,
		0.995862,
		0.987421,
		0.971582,
		0.946495,
		0.91095,
		0.864419,
		0.807029,
		0.739516,
		0.663169,
		0.579774,
		0.491569,
		0.401201,
		0.311689,
		0.226396,
		0.149001,
		0.083483,
		0.034098,
		0.005363,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.002008,
		0.025389,
		0.071053,
		0.134253,
		0.210511,
		0.295641,
		0.385763,
		0.477327,
		0.56714,
		0.652395,
		0.730708,
		0.800163,
		0.859359,
		0.907473,
		0.944313,
		0.970373,
		0.986864,
		0.995671,
		0.999193,
		0.999969,
		1,
		0.999776,
		0.997931,
		0.992087,
		0.979712,
		0.958631,
		0.927288,
		0.884829,
		0.831109,
		0.766643,
		0.692551,
		0.610497,
		0.522638,
		0.431573,
		0.340308,
		0.252217,
		0.171018,
		0.100747,
		0.045739,
		0.010606,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.005631,
		0.035743,
		0.08736,
		0.155629,
		0.235987,
		0.324181,
		0.416289,
		0.508744,
		0.598361,
		0.682378,
		0.75849,
		0.824905,
		0.880394,
		0.924359,
		0.956891,
		0.978822,
		0.991726,
		0.997833,
		0.999765,
		1,
		0.999968,
		0.999154,
		0.99547,
		0.986276,
		0.9691,
		0.94202,
		0.903828,
		0.854068,
		0.793,
		0.721547,
		0.641225,
		0.55409,
		0.462682,
		0.369978,
		0.279358,
		0.19457,
		0.119705,
		0.059174,
		0.017685,
		0.000232,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.011138,
		0.047952,
		0.105431,
		0.178617,
		0.262862,
		0.353853,
		0.447628,
		0.540612,
		0.629643,
		0.712016,
		0.785527,
		0.848532,
		0.900005,
		0.939609,
		0.967759,
		0.985655,
		0.995256,
		0.999113,
		0.999966,
		1,
		0.999754,
		0.997729,
		0.991344,
		0.977883,
		0.95506,
		0.921283,
		0.875748,
		0.818422,
		0.749996,
		0.671812,
		0.585802,
		0.494427,
		0.400625,
		0.30777,
		0.219635,
		0.140358,
		0.074425,
		0.02664,
		0.002111,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.000244,
		0.018575,
		0.062044,
		0.125272,
		0.203198,
		0.291093,
		0.384583,
		0.47968,
		0.572806,
		0.660837,
		0.741144,
		0.811648,
		0.870879,
		0.918052,
		0.953131,
		0.976892,
		0.990939,
		0.997618,
		0.999741,
		1,
		0.999964,
		0.999069,
		0.99503,
		0.984999,
		0.966345,
		0.937074,
		0.895994,
		0.842738,
		0.777728,
		0.7021,
		0.617635,
		0.526696,
		0.432165,
		0.337397,
		0.246181,
		0.162702,
		0.091512,
		0.037511,
		0.00592,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.002221,
		0.027987,
		0.078042,
		0.146884,
		0.229348,
		0.320624,
		0.416289,
		0.512331,
		0.605186,
		0.691782,
		0.769588,
		0.836674,
		0.891784,
		0.934406,
		0.964853,
		0.984305,
		0.99479,
		0.999022,
		0.999962,
		1,
		0.999728,
		0.997501,
		0.990511,
		0.975845,
		0.951098,
		0.914656,
		0.865776,
		0.804566,
		0.731919,
		0.649438,
		0.559361,
		0.464497,
		0.368169,
		0.27417,
		0.186723,
		0.110446,
		0.05033,
		0.011711,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.006231,
		0.039412,
		0.095965,
		0.170259,
		0.257029,
		0.35139,
		0.448872,
		0.545454,
		0.637598,
		0.722303,
		0.797161,
		0.860425,
		0.911086,
		0.948956,
		0.974739,
		0.990058,
		0.997377,
		0.999714,
		0.999999,
		0.99996,
		0.998972,
		0.994536,
		0.98357,
		0.963278,
		0.931598,
		0.887363,
		0.830324,
		0.761088,
		0.681045,
		0.592281,
		0.497509,
		0.400003,
		0.30355,
		0.212399,
		0.131232,
		0.065127,
		0.019535,
		0.000257,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.012329,
		0.052888,
		0.115823,
		0.195381,
		0.286194,
		0.383309,
		0.482221,
		0.578906,
		0.669872,
		0.752213,
		0.823672,
		0.882719,
		0.928639,
		0.961616,
		0.982792,
		0.994266,
		0.99892,
		0.999958,
		0.999999,
		0.999699,
		0.997245,
		0.989577,
		0.973569,
		0.946697,
		0.907331,
		0.854812,
		0.789417,
		0.712278,
		0.6253,
		0.531073,
		0.432804,
		0.334256,
		0.2397,
		0.153869,
		0.081927,
		0.029437,
		0.002341,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.000272,
		0.02057,
		0.068445,
		0.137622,
		0.222223,
		0.316784,
		0.416289,
		0.516206,
		0.612528,
		0.701826,
		0.781314,
		0.848922,
		0.903379,
		0.944313,
		0.972331,
		0.989067,
		0.997104,
		0.999683,
		0.999999,
		0.999956,
		0.998863,
		0.99398,
		0.981968,
		0.959858,
		0.925522,
		0.877839,
		0.816702,
		0.742944,
		0.658245,
		0.565045,
		0.466459,
		0.366214,
		0.268583,
		0.178346,
		0.100747,
		0.041461,
		0.006567,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.00247,
		0.031004,
		0.086107,
		0.161357,
		0.250747,
		0.348725,
		0.450219,
		0.550683,
		0.646144,
		0.73326,
		0.809397,
		0.872709,
		0.922235,
		0.958001,
		0.981094,
		0.993675,
		0.998803,
		0.999954,
		0.999999,
		0.999666,
		0.996954,
		0.988525,
		0.971021,
		0.941795,
		0.899218,
		0.842738,
		0.772836,
		0.690927,
		0.599263,
		0.500843,
		0.39933,
		0.298992,
		0.204642,
		0.121599,
		0.055645,
		0.012998,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.006932,
		0.043674,
		0.105894,
		0.187012,
		0.280903,
		0.381929,
		0.484973,
		0.58549,
		0.679561,
		0.763962,
		0.836245,
		0.894835,
		0.939136,
		0.969633,
		0.987949,
		0.996795,
		0.999648,
		0.999999,
		0.999951,
		0.998739,
		0.99335,
		0.980168,
		0.956035,
		0.918768,
		0.867313,
		0.801739,
		0.723143,
		0.63355,
		0.535808,
		0.433497,
		0.330857,
		0.232725,
		0.144484,
		0.072021,
		0.02169,
		0.000287,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.013723,
		0.058623,
		0.127813,
		0.214561,
		0.312626,
		0.416289,
		0.520404,
		0.620445,
		0.71257,
		0.793707,
		0.861636,
		0.915109,
		0.953955,
		0.979184,
		0.993005,
		0.998671,
		0.999948,
		0.999999,
		0.999628,
		0.996625,
		0.987338,
		0.968161,
		0.936326,
		0.890217,
		0.829424,
		0.754672,
		0.667708,
		0.571191,
		0.468589,
		0.364093,
		0.262549,
		0.169394,
		0.090614,
		0.032698,
		0.00261,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.000304,
		0.022905,
		0.075884,
		0.151865,
		0.243966,
		0.345833,
		0.451682,
		0.556348,
		0.655346,
		0.744944,
		0.822257,
		0.885348,
		0.933353,
		0.9666,
		0.986687,
		0.996443,
		0.999607,
		0.999999,
		0.999945,
		0.998598,
		0.992637,
		0.978139,
		0.951752,
		0.911245,
		0.855661,
		0.785281,
		0.701521,
		0.606807,
		0.50446,
		0.398598,
		0.294054,
		0.19631,
		0.111441,
		0.046069,
		0.007327,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.002763,
		0.034535,
		0.095481,
		0.178037,
		0.275172,
		0.380428,
		0.487964,
		0.592618,
		0.689972,
		0.776439,
		0.849368,
		0.907163,
		0.949417,
		0.977027,
		0.992245,
		0.99852,
		0.999942,
		0.999999,
		0.999585,
		0.996249,
		0.985994,
		0.964943,
		0.930208,
		0.880212,
		0.814721,
		0.734755,
		0.642452,
		0.540946,
		0.43425,
		0.327165,
		0.2252,
		0.13451,
		0.061846,
		0.014509,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.007758,
		0.048665,
		0.117431,
		0.206304,
		0.308108,
		0.416289,
		0.524968,
		0.629005,
		0.724081,
		0.806797,
		0.874792,
		0.926878,
		0.963183,
		0.985256,
		0.996042,
		0.999561,
		0.999999,
		0.999939,
		0.998436,
		0.991827,
		0.975845,
		0.94694,
		0.902848,
		0.842738,
		0.767158,
		0.677898,
		0.577857,
		0.470906,
		0.361785,
		0.256014,
		0.159817,
		0.080064,
		0.024223,
		0.000322,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.015366,
		0.065341,
		0.14174,
		0.236626,
		0.342684,
		0.453277,
		0.562504,
		0.665277,
		0.757413,
		0.835752,
		0.898284,
		0.944313,
		0.974587,
		0.991379,
		0.998347,
		0.999935,
		0.999999,
		0.999535,
		0.99582,
		0.984468,
		0.961312,
		0.92335,
		0.86907,
		0.798462,
		0.712898,
		0.614981,
		0.5084,
		0.397801,
		0.28869,
		0.187344,
		0.100747,
		0.036531,
		0.002929,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.000342,
		0.025659,
		0.084596,
		0.168399,
		0.268945,
		0.37879,
		0.491227,
		0.600358,
		0.701178,
		0.789689,
		0.863025,
		0.91961,
		0.959323,
		0.983628,
		0.995583,
		0.999508,
		0.999999,
		0.999931,
		0.998251,
		0.990902,
		0.973246,
		0.941522,
		0.893455,
		0.828385,
		0.74718,
		0.652082,
		0.54654,
		0.435073,
		0.323142,
		0.217061,
		0.123912,
		0.051487,
		0.008227,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.003111,
		0.038704,
		0.106458,
		0.197385,
		0.303183,
		0.416289,
		0.529947,
		0.638287,
		0.73643,
		0.820615,
		0.888342,
		0.938558,
		0.971816,
		0.99039,
		0.998148,
		0.999927,
		0.999999,
		0.999478,
		0.995329,
		0.98273,
		0.957205,
		0.915644,
		0.856637,
		0.780455,
		0.688896,
		0.585111,
		0.473438,
		0.359265,
		0.248916,
		0.149561,
		0.069138,
		0.0163,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.008741,
		0.05456,
		0.130939,
		0.228658,
		0.339242,
		0.455022,
		0.569217,
		0.676022,
		0.770731,
		0.849883,
		0.911434,
		0.954949,
		0.981771,
		0.995056,
		0.999446,
		0.999999,
		0.999922,
		0.998037,
		0.989843,
		0.970291,
		0.935405,
		0.882926,
		0.812415,
		0.725136,
		0.623865,
		0.512705,
		0.396928,
		0.282841,
		0.177679,
		0.089522,
		0.027226,
		0.000363,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.017323,
		0.073275,
		0.158037,
		0.262156,
		0.376995,
		0.494799,
		0.608789,
		0.713267,
		0.803759,
		0.877185,
		0.932051,
		0.968661,
		0.989256,
		0.997917,
		0.999917,
		0.999999,
		0.999412,
		0.994763,
		0.980744,
		0.952544,
		0.906963,
		0.842738,
		0.760489,
		0.662527,
		0.552654,
		0.435974,
		0.318742,
		0.208236,
		0.112664,
		0.041077,
		0.00331,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.000387,
		0.028941,
		0.094888,
		0.187731,
		0.297794,
		0.416289,
		0.535398,
		0.648379,
		0.749695,
		0.835175,
		0.90221,
		0.949977,
		0.979644,
		0.994447,
		0.999373,
		0.999997,
		0.999911,
		0.997788,
		0.988624,
		0.966918,
		0.928478,
		0.871094,
		0.794617,
		0.70079,
		0.59303,
		0.476214,
		0.3565,
		0.241181,
		0.138572,
		0.057915,
		0.009304,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.003529,
		0.043674,
		0.11942,
		0.21998,
		0.335457,
		0.456932,
		0.576554,
		0.687664,
		0.784947,
		0.864619,
		0.924657,
		0.965039,
		0.987932,
		0.997633,
		0.999888,
		0.999981,
		0.999317,
		0.994091,
		0.978449,
		0.947222,
		0.897143,
		0.827155,
		0.738309,
		0.633539,
		0.51742,
		0.395962,
		0.276435,
		0.167239,
		0.077791,
		0.018444,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.009922,
		0.061582,
		0.146868,
		0.254699,
		0.374977,
		0.498671,
		0.617935,
		0.726248,
		0.818594,
		0.891682,
		0.944206,
		0.977092,
		0.993631,
		0.999178,
		0.999885,
		0.999787,
		0.997388,
		0.98711,
		0.962953,
		0.920512,
		0.857677,
		0.774666,
		0.673812,
		0.559298,
		0.436917,
		0.313873,
		0.198621,
		0.100736,
		0.03082,
		0.000414,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.019668,
		0.082699,
		0.177175,
		0.291735,
		0.416091,
		0.541136,
		0.659076,
		0.763598,
		0.85008,
		0.915849,
		0.960466,
		0.98597,
		0.996864,
		0.999416,
		0.999522,
		0.998768,
		0.992877,
		0.97538,
		0.940734,
		0.885637,
		0.80931,
		0.713346,
		0.601425,
		0.479047,
		0.353287,
		0.232614,
		0.126741,
		0.046503,
		0.003769,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.000442,
		0.032845,
		0.106998,
		0.21019,
		0.330792,
		0.458357,
		0.583747,
		0.699281,
		0.798942,
		0.878618,
		0.936403,
		0.9729,
		0.991411,
		0.997669,
		0.998473,
		0.998361,
		0.995644,
		0.984094,
		0.957169,
		0.910271,
		0.841453,
		0.751401,
		0.643164,
		0.52184,
		0.394307,
		0.268996,
		0.155717,
		0.065521,
		0.01059,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.004022,
		0.049465,
		0.134336,
		0.245582,
		0.371341,
		0.501051,
		0.625599,
		0.737504,
		0.831152,
		0.903084,
		0.952336,
		0.980739,
		0.992978,
		0.995868,
		0.995988,
		0.995131,
		0.988486,
		0.968915,
		0.930499,
		0.869955,
		0.786881,
		0.68353,
		0.564478,
		0.436307,
		0.30734,
		0.187433,
		0.087814,
		0.020955,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.011259,
		0.069425,
		0.164372,
		0.282748,
		0.412507,
		0.543038,
		0.66534,
		0.772234,
		0.858647,
		0.921945,
		0.962301,
		0.982956,
		0.99,
		0.990912,
		0.990785,
		0.987715,
		0.974769,
		0.944861,
		0.893244,
		0.818227,
		0.721083,
		0.605712,
		0.478279,
		0.346903,
		0.221416,
		0.113164,
		0.034862,
		0.000471,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.022135,
		0.092422,
		0.19647,
		0.320699,
		0.452958,
		0.582681,
		0.701088,
		0.801442,
		0.879373,
		0.93327,
		0.964651,
		0.978299,
		0.98155,
		0.981686,
		0.980718,
		0.973275,
		0.95155,
		0.909308,
		0.84338,
		0.753855,
		0.643772,
		0.51873,
		0.386524,
		0.256856,
		0.141106,
		0.052157,
		0.004255,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.000494,
		0.036449,
		0.11786,
		0.229604,
		0.357979,
		0.490845,
		0.617806,
		0.730447,
		0.822636,
		0.890903,
		0.934871,
		0.957597,
		0.96542,
		0.966441,
		0.966297,
		0.962873,
		0.948562,
		0.915816,
		0.859861,
		0.779382,
		0.676358,
		0.555639,
		0.424544,
		0.29251,
		0.170831,
		0.07245,
		0.011792,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.0044,
		0.053735,
		0.144773,
		0.262302,
		0.392648,
		0.523831,
		0.645788,
		0.750642,
		0.83306,
		0.890684,
		0.924584,
		0.939475,
		0.943055,
		0.943206,
		0.942136,
		0.933976,
		0.910389,
		0.864989,
		0.794878,
		0.700751,
		0.58651,
		0.458811,
		0.326664,
		0.201119,
		0.095026,
		0.022848,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.011958,
		0.073179,
		0.171783,
		0.292662,
		0.422357,
		0.549241,
		0.663773,
		0.758816,
		0.830044,
		0.876427,
		0.900656,
		0.909081,
		0.910191,
		0.910034,
		0.906329,
		0.890993,
		0.856264,
		0.797554,
		0.714073,
		0.608558,
		0.486799,
		0.35718,
		0.23032,
		0.118789,
		0.036891,
		0.000502,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.0226,
		0.093595,
		0.197141,
		0.318468,
		0.444555,
		0.564348,
		0.669042,
		0.75245,
		0.811468,
		0.84658,
		0.86217,
		0.865957,
		0.866118,
		0.864981,
		0.856393,
		0.83183,
		0.785067,
		0.713679,
		0.619035,
		0.505845,
		0.381656,
		0.256439,
		0.142261,
		0.053042,
		0.004361,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.000483,
		0.0354,
		0.113465,
		0.218859,
		0.337415,
		0.456801,
		0.566765,
		0.659459,
		0.729854,
		0.77625,
		0.800764,
		0.809382,
		0.810529,
		0.810366,
		0.806556,
		0.790953,
		0.756013,
		0.697635,
		0.615657,
		0.513512,
		0.397697,
		0.277308,
		0.163672,
		0.070064,
		0.011499,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.004055,
		0.049109,
		0.131066,
		0.234937,
		0.347429,
		0.45715,
		0.55488,
		0.633929,
		0.690604,
		0.724736,
		0.74007,
		0.743835,
		0.743996,
		0.742861,
		0.734367,
		0.710347,
		0.665163,
		0.597047,
		0.507996,
		0.403267,
		0.290865,
		0.181135,
		0.086449,
		0.020971,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.010267,
		0.062266,
		0.144681,
		0.243645,
		0.347001,
		0.444525,
		0.528227,
		0.592741,
		0.635826,
		0.658875,
		0.667076,
		0.668179,
		0.668022,
		0.664376,
		0.649616,
		0.616965,
		0.563102,
		0.488506,
		0.397048,
		0.295462,
		0.192897,
		0.100576,
		0.031536,
		0.000432,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.017889,
		0.073371,
		0.152846,
		0.243819,
		0.33549,
		0.418991,
		0.487707,
		0.537699,
		0.568211,
		0.582095,
		0.585544,
		0.585694,
		0.584648,
		0.57691,
		0.555299,
		0.515172,
		0.455512,
		0.378733,
		0.29017,
		0.197612,
		0.110928,
		0.04179,
		0.003467,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.000352,
		0.025572,
		0.08111,
		0.154586,
		0.235077,
		0.313298,
		0.381856,
		0.435589,
		0.472002,
		0.491749,
		0.498867,
		0.499835,
		0.499696,
		0.496513,
		0.483784,
		0.455999,
		0.4108,
		0.349164,
		0.274981,
		0.194576,
		0.116324,
		0.050357,
		0.008346,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.002672,
		0.032043,
		0.084551,
		0.149594,
		0.217932,
		0.281881,
		0.335565,
		0.375266,
		0.399855,
		0.411201,
		0.414057,
		0.414182,
		0.413311,
		0.406945,
		0.389406,
		0.3573,
		0.310294,
		0.250868,
		0.18387,
		0.116117,
		0.056098,
		0.013754,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.00605,
		0.036295,
		0.083299,
		0.138292,
		0.193752,
		0.24359,
		0.283404,
		0.310826,
		0.325921,
		0.33144,
		0.3322,
		0.33209,
		0.329606,
		0.319805,
		0.298721,
		0.26495,
		0.219692,
		0.166381,
		0.110298,
		0.058278,
		0.018485,
		0.000256,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.009308,
		0.037732,
		0.077548,
		0.121789,
		0.164591,
		0.201372,
		0.229082,
		0.246526,
		0.254699,
		0.256786,
		0.256879,
		0.256238,
		0.251619,
		0.239079,
		0.216483,
		0.183964,
		0.143689,
		0.099509,
		0.056675,
		0.021621,
		0.001814,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.000161,
		0.011571,
		0.036236,
		0.068049,
		0.101724,
		0.132916,
		0.158393,
		0.176263,
		0.186264,
		0.189978,
		0.190497,
		0.190422,
		0.188738,
		0.182192,
		0.168337,
		0.146521,
		0.11786,
		0.084949,
		0.0516,
		0.022647,
		0.003798,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.001043,
		0.012361,
		0.032163,
		0.055989,
		0.080042,
		0.101297,
		0.117656,
		0.128143,
		0.13314,
		0.134436,
		0.134495,
		0.134094,
		0.131247,
		0.123644,
		0.110179,
		0.091174,
		0.068196,
		0.043827,
		0.021495,
		0.005339,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001962,
		0.011614,
		0.02625,
		0.042807,
		0.058738,
		0.072094,
		0.081659,
		0.087112,
		0.089172,
		0.089464,
		0.089421,
		0.08848,
		0.08488,
		0.077394,
		0.065832,
		0.050988,
		0.03446,
		0.018512,
		0.005956,
		8.4e-005,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		0.002411,
		0.009635,
		0.019471,
		0.02998,
		0.039592,
		0.047176,
		0.052138,
		0.054548,
		0.055184,
		0.055213,
		0.055015,
		0.05363,
		0.049997,
		0.043688,
		0.03498,
		0.024749,
		0.014355,
		0.005563,
		0.000473,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		3.2e-005,
		0.002263,
		0.006974,
		0.012854,
		0.018798,
		0.023941,
		0.027714,
		0.02991,
		0.030755,
		0.030877,
		0.030859,
		0.030469,
		0.029006,
		0.026022,
		0.021513,
		0.015878,
		0.009842,
		0.004394,
		0.000748,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		0.000143,
		0.00167,
		0.004271,
		0.007282,
		0.01016,
		0.012498,
		0.014064,
		0.014841,
		0.01505,
		0.015059,
		0.014994,
		0.014543,
		0.013384,
		0.011415,
		0.008767,
		0.005762,
		0.00288,
		0.000727,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000161,
		0.000938,
		0.002079,
		0.003312,
		0.004423,
		0.005261,
		0.00576,
		0.005956,
		0.005985,
		0.005981,
		0.005889,
		0.005553,
		0.004882,
		0.003894,
		0.002699,
		0.001481,
		0.000485,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		9.1e-005,
		0.000355,
		0.000703,
		0.001055,
		0.001351,
		0.001555,
		0.001659,
		0.001687,
		0.001688,
		0.001679,
		0.001619,
		0.001466,
		0.001213,
		0.000882,
		0.000524,
		0.000207,
		1.8e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		2.1e-005,
		6.4e-005,
		0.000116,
		0.000165,
		0.000203,
		0.000226,
		0.000235,
		0.000237,
		0.000237,
		0.000232,
		0.000216,
		0.000185,
		0.000141,
		9e-005,
		4.1e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		4.1e-005,
		9e-005,
		0.000141,
		0.000185,
		0.000216,
		0.000232,
		0.000237,
		0.000237,
		0.000235,
		0.000226,
		0.000203,
		0.000165,
		0.000116,
		6.4e-005,
		2.1e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		1.8e-005,
		0.000207,
		0.000524,
		0.000882,
		0.001213,
		0.001466,
		0.001619,
		0.001679,
		0.001688,
		0.001687,
		0.001659,
		0.001555,
		0.001351,
		0.001055,
		0.000703,
		0.000355,
		9.1e-005,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		0.000485,
		0.001481,
		0.002699,
		0.003894,
		0.004882,
		0.005553,
		0.005889,
		0.005981,
		0.005985,
		0.005956,
		0.00576,
		0.005261,
		0.004423,
		0.003312,
		0.002079,
		0.000938,
		0.000161,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000727,
		0.00288,
		0.005762,
		0.008767,
		0.011415,
		0.013384,
		0.014543,
		0.014994,
		0.015059,
		0.01505,
		0.014841,
		0.014064,
		0.012498,
		0.01016,
		0.007282,
		0.004271,
		0.00167,
		0.000143,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		0.000748,
		0.004394,
		0.009842,
		0.015878,
		0.021513,
		0.026022,
		0.029006,
		0.030469,
		0.030859,
		0.030877,
		0.030755,
		0.02991,
		0.027714,
		0.023941,
		0.018798,
		0.012854,
		0.006974,
		0.002263,
		3.2e-005,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.000473,
		0.005563,
		0.014355,
		0.024749,
		0.03498,
		0.043688,
		0.049997,
		0.05363,
		0.055015,
		0.055213,
		0.055184,
		0.054548,
		0.052138,
		0.047176,
		0.039592,
		0.02998,
		0.019471,
		0.009635,
		0.002411,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		8.4e-005,
		0.005956,
		0.018512,
		0.03446,
		0.050988,
		0.065832,
		0.077394,
		0.08488,
		0.08848,
		0.089421,
		0.089464,
		0.089172,
		0.087112,
		0.081659,
		0.072094,
		0.058738,
		0.042807,
		0.02625,
		0.011614,
		0.001962,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.005339,
		0.021495,
		0.043827,
		0.068196,
		0.091174,
		0.110179,
		0.123644,
		0.131247,
		0.134094,
		0.134495,
		0.134436,
		0.13314,
		0.128143,
		0.117656,
		0.101297,
		0.080042,
		0.055989,
		0.032163,
		0.012361,
		0.001043,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.003798,
		0.022647,
		0.0516,
		0.084949,
		0.11786,
		0.146521,
		0.168337,
		0.182192,
		0.188738,
		0.190422,
		0.190497,
		0.189978,
		0.186264,
		0.176263,
		0.158393,
		0.132916,
		0.101724,
		0.068049,
		0.036236,
		0.011571,
		0.000161,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.001814,
		0.021621,
		0.056675,
		0.099509,
		0.143689,
		0.183964,
		0.216483,
		0.239079,
		0.251619,
		0.256238,
		0.256879,
		0.256786,
		0.254699,
		0.246526,
		0.229082,
		0.201372,
		0.164591,
		0.121789,
		0.077548,
		0.037732,
		0.009308,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.000256,
		0.018485,
		0.058278,
		0.110298,
		0.166381,
		0.219692,
		0.26495,
		0.298721,
		0.319805,
		0.329606,
		0.33209,
		0.3322,
		0.33144,
		0.325921,
		0.310826,
		0.283404,
		0.24359,
		0.193752,
		0.138292,
		0.083299,
		0.036295,
		0.00605,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.013754,
		0.056098,
		0.116117,
		0.18387,
		0.250868,
		0.310294,
		0.3573,
		0.389406,
		0.406945,
		0.413311,
		0.414182,
		0.414057,
		0.411201,
		0.399855,
		0.375266,
		0.335565,
		0.281881,
		0.217932,
		0.149594,
		0.084551,
		0.032043,
		0.002672,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.008346,
		0.050357,
		0.116324,
		0.194576,
		0.274981,
		0.349164,
		0.4108,
		0.455999,
		0.483784,
		0.496513,
		0.499696,
		0.499835,
		0.498867,
		0.491749,
		0.472002,
		0.435589,
		0.381856,
		0.313298,
		0.235077,
		0.154586,
		0.08111,
		0.025572,
		0.000352,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.003467,
		0.04179,
		0.110928,
		0.197612,
		0.29017,
		0.378733,
		0.455512,
		0.515172,
		0.555299,
		0.57691,
		0.584648,
		0.585694,
		0.585544,
		0.582095,
		0.568211,
		0.537699,
		0.487707,
		0.418991,
		0.33549,
		0.243819,
		0.152846,
		0.073371,
		0.017889,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.000432,
		0.031536,
		0.100576,
		0.192897,
		0.295462,
		0.397048,
		0.488506,
		0.563102,
		0.616965,
		0.649616,
		0.664376,
		0.668022,
		0.668179,
		0.667076,
		0.658875,
		0.635826,
		0.592741,
		0.528227,
		0.444525,
		0.347001,
		0.243645,
		0.144681,
		0.062266,
		0.010267,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.020971,
		0.086449,
		0.181135,
		0.290865,
		0.403267,
		0.507996,
		0.597047,
		0.665163,
		0.710347,
		0.734367,
		0.742861,
		0.743996,
		0.743835,
		0.74007,
		0.724736,
		0.690604,
		0.633929,
		0.55488,
		0.45715,
		0.347429,
		0.234937,
		0.131066,
		0.049109,
		0.004055,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.011499,
		0.070064,
		0.163672,
		0.277308,
		0.397697,
		0.513512,
		0.615657,
		0.697635,
		0.756013,
		0.790953,
		0.806556,
		0.810366,
		0.810529,
		0.809382,
		0.800764,
		0.77625,
		0.729854,
		0.659459,
		0.566765,
		0.456801,
		0.337415,
		0.218859,
		0.113465,
		0.0354,
		0.000483,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.004361,
		0.053042,
		0.142261,
		0.256439,
		0.381656,
		0.505845,
		0.619035,
		0.713679,
		0.785067,
		0.83183,
		0.856393,
		0.864981,
		0.866118,
		0.865957,
		0.86217,
		0.84658,
		0.811468,
		0.75245,
		0.669042,
		0.564348,
		0.444555,
		0.318468,
		0.197141,
		0.093595,
		0.0226,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.000502,
		0.036891,
		0.118789,
		0.23032,
		0.35718,
		0.486799,
		0.608558,
		0.714073,
		0.797554,
		0.856264,
		0.890993,
		0.906329,
		0.910034,
		0.910191,
		0.909081,
		0.900656,
		0.876427,
		0.830044,
		0.758816,
		0.663773,
		0.549241,
		0.422357,
		0.292662,
		0.171783,
		0.073179,
		0.011958,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.022848,
		0.095026,
		0.201119,
		0.326664,
		0.458811,
		0.58651,
		0.700751,
		0.794878,
		0.864989,
		0.910389,
		0.933976,
		0.942136,
		0.943206,
		0.943055,
		0.939475,
		0.924584,
		0.890684,
		0.83306,
		0.750642,
		0.645788,
		0.523831,
		0.392648,
		0.262302,
		0.144773,
		0.053735,
		0.0044,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.011792,
		0.07245,
		0.170831,
		0.29251,
		0.424544,
		0.555639,
		0.676358,
		0.779382,
		0.859861,
		0.915816,
		0.948562,
		0.962873,
		0.966297,
		0.966441,
		0.96542,
		0.957597,
		0.934871,
		0.890903,
		0.822636,
		0.730447,
		0.617806,
		0.490845,
		0.357979,
		0.229604,
		0.11786,
		0.036449,
		0.000494,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.004255,
		0.052157,
		0.141106,
		0.256856,
		0.386524,
		0.51873,
		0.643772,
		0.753855,
		0.84338,
		0.909308,
		0.95155,
		0.973275,
		0.980718,
		0.981686,
		0.98155,
		0.978299,
		0.964651,
		0.93327,
		0.879373,
		0.801442,
		0.701088,
		0.582681,
		0.452958,
		0.320699,
		0.19647,
		0.092422,
		0.022135,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.000471,
		0.034862,
		0.113164,
		0.221416,
		0.346903,
		0.478279,
		0.605712,
		0.721083,
		0.818227,
		0.893244,
		0.944861,
		0.974769,
		0.987715,
		0.990785,
		0.990912,
		0.99,
		0.982956,
		0.962301,
		0.921945,
		0.858647,
		0.772234,
		0.66534,
		0.543038,
		0.412507,
		0.282748,
		0.164372,
		0.069425,
		0.011259,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.020955,
		0.087814,
		0.187433,
		0.30734,
		0.436307,
		0.564478,
		0.68353,
		0.786881,
		0.869955,
		0.930499,
		0.968915,
		0.988486,
		0.995131,
		0.995988,
		0.995868,
		0.992978,
		0.980739,
		0.952336,
		0.903084,
		0.831152,
		0.737504,
		0.625599,
		0.501051,
		0.371341,
		0.245582,
		0.134336,
		0.049465,
		0.004022,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.01059,
		0.065521,
		0.155717,
		0.268996,
		0.394307,
		0.52184,
		0.643164,
		0.751401,
		0.841453,
		0.910271,
		0.957169,
		0.984094,
		0.995644,
		0.998361,
		0.998473,
		0.997669,
		0.991411,
		0.9729,
		0.936403,
		0.878618,
		0.798942,
		0.699281,
		0.583747,
		0.458357,
		0.330792,
		0.21019,
		0.106998,
		0.032845,
		0.000442,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.003769,
		0.046503,
		0.126741,
		0.232614,
		0.353287,
		0.479047,
		0.601425,
		0.713346,
		0.80931,
		0.885637,
		0.940734,
		0.97538,
		0.992877,
		0.998768,
		0.999522,
		0.999416,
		0.996864,
		0.98597,
		0.960466,
		0.915849,
		0.85008,
		0.763598,
		0.659076,
		0.541136,
		0.416091,
		0.291735,
		0.177175,
		0.082699,
		0.019668,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.000414,
		0.03082,
		0.100736,
		0.198621,
		0.313873,
		0.436917,
		0.559298,
		0.673812,
		0.774666,
		0.857677,
		0.920512,
		0.962953,
		0.98711,
		0.997388,
		0.999787,
		0.999885,
		0.999178,
		0.993631,
		0.977092,
		0.944206,
		0.891682,
		0.818594,
		0.726248,
		0.617935,
		0.498671,
		0.374977,
		0.254699,
		0.146868,
		0.061582,
		0.009922,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.018444,
		0.077791,
		0.167239,
		0.276435,
		0.395962,
		0.51742,
		0.633539,
		0.738309,
		0.827155,
		0.897143,
		0.947222,
		0.978449,
		0.994091,
		0.999317,
		0.999981,
		0.999888,
		0.997633,
		0.987932,
		0.965039,
		0.924657,
		0.864619,
		0.784947,
		0.687664,
		0.576554,
		0.456932,
		0.335457,
		0.21998,
		0.11942,
		0.043674,
		0.003529,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.009304,
		0.057915,
		0.138572,
		0.241181,
		0.3565,
		0.476214,
		0.59303,
		0.70079,
		0.794617,
		0.871094,
		0.928478,
		0.966918,
		0.988624,
		0.997788,
		0.999911,
		0.999997,
		0.999373,
		0.994447,
		0.979644,
		0.949977,
		0.90221,
		0.835175,
		0.749695,
		0.648379,
		0.535398,
		0.416289,
		0.297794,
		0.187731,
		0.094888,
		0.028941,
		0.000387,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.00331,
		0.041077,
		0.112664,
		0.208236,
		0.318742,
		0.435974,
		0.552654,
		0.662527,
		0.760489,
		0.842738,
		0.906963,
		0.952544,
		0.980744,
		0.994763,
		0.999412,
		0.999999,
		0.999917,
		0.997917,
		0.989256,
		0.968661,
		0.932051,
		0.877185,
		0.803759,
		0.713267,
		0.608789,
		0.494799,
		0.376995,
		0.262156,
		0.158037,
		0.073275,
		0.017323,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.000363,
		0.027226,
		0.089522,
		0.177679,
		0.282841,
		0.396928,
		0.512705,
		0.623865,
		0.725136,
		0.812415,
		0.882926,
		0.935405,
		0.970291,
		0.989843,
		0.998037,
		0.999922,
		0.999999,
		0.999446,
		0.995056,
		0.981771,
		0.954949,
		0.911434,
		0.849883,
		0.770731,
		0.676022,
		0.569217,
		0.455022,
		0.339242,
		0.228658,
		0.130939,
		0.05456,
		0.008741,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.0163,
		0.069138,
		0.149561,
		0.248916,
		0.359265,
		0.473438,
		0.585111,
		0.688896,
		0.780455,
		0.856637,
		0.915644,
		0.957205,
		0.98273,
		0.995329,
		0.999478,
		0.999999,
		0.999927,
		0.998148,
		0.99039,
		0.971816,
		0.938558,
		0.888342,
		0.820615,
		0.73643,
		0.638287,
		0.529947,
		0.416289,
		0.303183,
		0.197385,
		0.106458,
		0.038704,
		0.003111,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.008227,
		0.051487,
		0.123912,
		0.217061,
		0.323142,
		0.435073,
		0.54654,
		0.652082,
		0.74718,
		0.828385,
		0.893455,
		0.941522,
		0.973246,
		0.990902,
		0.998251,
		0.999931,
		0.999999,
		0.999508,
		0.995583,
		0.983628,
		0.959323,
		0.91961,
		0.863025,
		0.789689,
		0.701178,
		0.600358,
		0.491227,
		0.37879,
		0.268945,
		0.168399,
		0.084596,
		0.025659,
		0.000342,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.002929,
		0.036531,
		0.100747,
		0.187344,
		0.28869,
		0.397801,
		0.5084,
		0.614981,
		0.712898,
		0.798462,
		0.86907,
		0.92335,
		0.961312,
		0.984468,
		0.99582,
		0.999535,
		0.999999,
		0.999935,
		0.998347,
		0.991379,
		0.974587,
		0.944313,
		0.898284,
		0.835752,
		0.757413,
		0.665277,
		0.562504,
		0.453277,
		0.342684,
		0.236626,
		0.14174,
		0.065341,
		0.015366,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.000322,
		0.024223,
		0.080064,
		0.159817,
		0.256014,
		0.361785,
		0.470906,
		0.577857,
		0.677898,
		0.767158,
		0.842738,
		0.902848,
		0.94694,
		0.975845,
		0.991827,
		0.998436,
		0.999939,
		0.999999,
		0.999561,
		0.996042,
		0.985256,
		0.963183,
		0.926878,
		0.874792,
		0.806797,
		0.724081,
		0.629005,
		0.524968,
		0.416289,
		0.308108,
		0.206304,
		0.117431,
		0.048665,
		0.007758,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.014509,
		0.061846,
		0.13451,
		0.2252,
		0.327165,
		0.43425,
		0.540946,
		0.642452,
		0.734755,
		0.814721,
		0.880212,
		0.930208,
		0.964943,
		0.985994,
		0.996249,
		0.999585,
		0.999999,
		0.999942,
		0.99852,
		0.992245,
		0.977027,
		0.949417,
		0.907163,
		0.849368,
		0.776439,
		0.689972,
		0.592618,
		0.487964,
		0.380428,
		0.275172,
		0.178037,
		0.095481,
		0.034535,
		0.002763,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.007327,
		0.046069,
		0.111441,
		0.19631,
		0.294054,
		0.398598,
		0.50446,
		0.606807,
		0.701521,
		0.785281,
		0.855661,
		0.911245,
		0.951752,
		0.978139,
		0.992637,
		0.998598,
		0.999945,
		0.999999,
		0.999607,
		0.996443,
		0.986687,
		0.9666,
		0.933353,
		0.885348,
		0.822257,
		0.744944,
		0.655346,
		0.556348,
		0.451682,
		0.345833,
		0.243966,
		0.151865,
		0.075884,
		0.022905,
		0.000304,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.00261,
		0.032698,
		0.090614,
		0.169394,
		0.262549,
		0.364093,
		0.468589,
		0.571191,
		0.667708,
		0.754672,
		0.829424,
		0.890217,
		0.936326,
		0.968161,
		0.987338,
		0.996625,
		0.999628,
		0.999999,
		0.999948,
		0.998671,
		0.993005,
		0.979184,
		0.953955,
		0.915109,
		0.861636,
		0.793707,
		0.71257,
		0.620445,
		0.520404,
		0.416289,
		0.312626,
		0.214561,
		0.127813,
		0.058623,
		0.013723,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.000287,
		0.02169,
		0.072021,
		0.144484,
		0.232725,
		0.330857,
		0.433497,
		0.535808,
		0.63355,
		0.723143,
		0.801739,
		0.867313,
		0.918768,
		0.956035,
		0.980168,
		0.99335,
		0.998739,
		0.999951,
		0.999999,
		0.999648,
		0.996795,
		0.987949,
		0.969633,
		0.939136,
		0.894835,
		0.836245,
		0.763962,
		0.679561,
		0.58549,
		0.484973,
		0.381929,
		0.280903,
		0.187012,
		0.105894,
		0.043674,
		0.006932,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.012998,
		0.055645,
		0.121599,
		0.204642,
		0.298992,
		0.39933,
		0.500843,
		0.599263,
		0.690927,
		0.772836,
		0.842738,
		0.899218,
		0.941795,
		0.971021,
		0.988525,
		0.996954,
		0.999666,
		0.999999,
		0.999954,
		0.998803,
		0.993675,
		0.981094,
		0.958001,
		0.922235,
		0.872709,
		0.809397,
		0.73326,
		0.646144,
		0.550683,
		0.450219,
		0.348725,
		0.250747,
		0.161357,
		0.086107,
		0.031004,
		0.00247,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.006567,
		0.041461,
		0.100747,
		0.178346,
		0.268583,
		0.366214,
		0.466459,
		0.565045,
		0.658245,
		0.742944,
		0.816702,
		0.877839,
		0.925522,
		0.959858,
		0.981968,
		0.99398,
		0.998863,
		0.999956,
		0.999999,
		0.999683,
		0.997104,
		0.989067,
		0.972331,
		0.944313,
		0.903379,
		0.848922,
		0.781314,
		0.701826,
		0.612528,
		0.516206,
		0.416289,
		0.316784,
		0.222223,
		0.137622,
		0.068445,
		0.02057,
		0.000272,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.002341,
		0.029437,
		0.081927,
		0.153869,
		0.2397,
		0.334256,
		0.432804,
		0.531073,
		0.6253,
		0.712278,
		0.789417,
		0.854812,
		0.907331,
		0.946697,
		0.973569,
		0.989577,
		0.997245,
		0.999699,
		0.999999,
		0.999958,
		0.99892,
		0.994266,
		0.982792,
		0.961616,
		0.928639,
		0.882719,
		0.823672,
		0.752213,
		0.669872,
		0.578906,
		0.482221,
		0.383309,
		0.286194,
		0.195381,
		0.115823,
		0.052888,
		0.012329,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.000257,
		0.019535,
		0.065127,
		0.131232,
		0.212399,
		0.30355,
		0.400003,
		0.497509,
		0.592281,
		0.681045,
		0.761088,
		0.830324,
		0.887363,
		0.931598,
		0.963278,
		0.98357,
		0.994536,
		0.998972,
		0.99996,
		0.999999,
		0.999714,
		0.997377,
		0.990058,
		0.974739,
		0.948956,
		0.911086,
		0.860425,
		0.797161,
		0.722303,
		0.637598,
		0.545454,
		0.448872,
		0.35139,
		0.257029,
		0.170259,
		0.095965,
		0.039412,
		0.006231,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.011711,
		0.05033,
		0.110446,
		0.186723,
		0.27417,
		0.368169,
		0.464497,
		0.559361,
		0.649438,
		0.731919,
		0.804566,
		0.865776,
		0.914656,
		0.951098,
		0.975845,
		0.990511,
		0.997501,
		0.999728,
		1,
		0.999962,
		0.999022,
		0.99479,
		0.984305,
		0.964853,
		0.934406,
		0.891784,
		0.836674,
		0.769588,
		0.691782,
		0.605186,
		0.512331,
		0.416289,
		0.320624,
		0.229348,
		0.146884,
		0.078042,
		0.027987,
		0.002221,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.00592,
		0.037511,
		0.091512,
		0.162702,
		0.246181,
		0.337397,
		0.432165,
		0.526696,
		0.617635,
		0.7021,
		0.777728,
		0.842738,
		0.895994,
		0.937074,
		0.966345,
		0.984999,
		0.99503,
		0.999069,
		0.999964,
		1,
		0.999741,
		0.997618,
		0.990939,
		0.976892,
		0.953131,
		0.918052,
		0.870879,
		0.811648,
		0.741144,
		0.660837,
		0.572806,
		0.47968,
		0.384583,
		0.291093,
		0.203198,
		0.125272,
		0.062044,
		0.018575,
		0.000244,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.002111,
		0.02664,
		0.074425,
		0.140358,
		0.219635,
		0.30777,
		0.400625,
		0.494427,
		0.585802,
		0.671812,
		0.749996,
		0.818422,
		0.875748,
		0.921283,
		0.95506,
		0.977883,
		0.991344,
		0.997729,
		0.999754,
		1,
		0.999966,
		0.999113,
		0.995256,
		0.985655,
		0.967759,
		0.939609,
		0.900005,
		0.848532,
		0.785527,
		0.712016,
		0.629643,
		0.540612,
		0.447628,
		0.353853,
		0.262862,
		0.178617,
		0.105431,
		0.047952,
		0.011138,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.000232,
		0.017685,
		0.059174,
		0.119705,
		0.19457,
		0.279358,
		0.369978,
		0.462682,
		0.55409,
		0.641225,
		0.721547,
		0.793,
		0.854068,
		0.903828,
		0.94202,
		0.9691,
		0.986276,
		0.99547,
		0.999154,
		0.999968,
		1,
		0.999765,
		0.997833,
		0.991726,
		0.978822,
		0.956891,
		0.924359,
		0.880394,
		0.824905,
		0.75849,
		0.682378,
		0.598361,
		0.508744,
		0.416289,
		0.324181,
		0.235987,
		0.155629,
		0.08736,
		0.035743,
		0.005631,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.010606,
		0.045739,
		0.100747,
		0.171018,
		0.252217,
		0.340308,
		0.431573,
		0.522638,
		0.610497,
		0.692551,
		0.766643,
		0.831109,
		0.884829,
		0.927288,
		0.958631,
		0.979712,
		0.992087,
		0.997931,
		0.999776,
		1,
		0.999969,
		0.999193,
		0.995671,
		0.986864,
		0.970373,
		0.944313,
		0.907473,
		0.859359,
		0.800163,
		0.730708,
		0.652395,
		0.56714,
		0.477327,
		0.385763,
		0.295641,
		0.210511,
		0.134253,
		0.071053,
		0.025389,
		0.002008,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.005363,
		0.034098,
		0.083483,
		0.149001,
		0.226396,
		0.311689,
		0.401201,
		0.491569,
		0.579774,
		0.663169,
		0.739516,
		0.807029,
		0.864419,
		0.91095,
		0.946495,
		0.971582,
		0.987421,
		0.995862,
		0.99923,
		0.999971,
		1,
		0.999787,
		0.998024,
		0.992429,
		0.980556,
		0.960286,
		0.930079,
		0.889065,
		0.837049,
		0.77447,
		0.702346,
		0.622222,
		0.536117,
		0.446476,
		0.356136,
		0.268292,
		0.186469,
		0.114498,
		0.056497,
		0.016857,
		0.000221,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.001913,
		0.024223,
		0.067904,
		0.128533,
		0.201932,
		0.284186,
		0.371655,
		0.460998,
		0.549189,
		0.63355,
		0.711778,
		0.781985,
		0.842738,
		0.893112,
		0.932739,
		0.961859,
		0.981357,
		0.992753,
		0.998111,
		0.999797,
		1,
		0.999972,
		0.999265,
		0.996042,
		0.987949,
		0.97273,
		0.948572,
		0.914268,
		0.869259,
		0.813614,
		0.747985,
		0.673559,
		0.592002,
		0.505413,
		0.416289,
		0.327485,
		0.242186,
		0.163887,
		0.096367,
		0.043674,
		0.010111,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.000211,
		0.016085,
		0.053998,
		0.10962,
		0.178855,
		0.25785,
		0.343012,
		0.431024,
		0.518865,
		0.603835,
		0.683581,
		0.756129,
		0.819929,
		0.87389,
		0.917436,
		0.950551,
		0.973821,
		0.98845,
		0.996213,
		0.999297,
		0.999973,
		1,
		0.999806,
		0.998194,
		0.99306,
		0.982117,
		0.963355,
		0.935275,
		0.89698,
		0.848188,
		0.789202,
		0.720862,
		0.644493,
		0.561865,
		0.475143,
		0.386858,
		0.299874,
		0.217363,
		0.142782,
		0.079857,
		0.032563,
		0.005114,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.00965,
		0.041747,
		0.092264,
		0.157186,
		0.232725,
		0.315337,
		0.401737,
		0.488913,
		0.574153,
		0.655065,
		0.72961,
		0.796135,
		0.85341,
		0.900677,
		0.937694,
		0.96478,
		0.982839,
		0.99335,
		0.998273,
		0.999814,
		1,
		0.999974,
		0.999328,
		0.996374,
		0.988925,
		0.974858,
		0.952436,
		0.920461,
		0.878322,
		0.825987,
		0.763962,
		0.693246,
		0.615286,
		0.531932,
		0.445405,
		0.358259,
		0.273359,
		0.193856,
		0.123166,
		0.064959,
		0.023136,
		0.001825,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.004882,
		0.031129,
		0.076461,
		0.13694,
		0.208847,
		0.28869,
		0.373216,
		0.459432,
		0.544621,
		0.626365,
		0.70257,
		0.771497,
		0.831801,
		0.882566,
		0.92335,
		0.954233,
		0.975845,
		0.989376,
		0.996527,
		0.999358,
		0.999976,
		1,
		0.999823,
		0.998347,
		0.993626,
		0.983525,
		0.966137,
		0.940003,
		0.904213,
		0.858415,
		0.802795,
		0.738037,
		0.665277,
		0.586062,
		0.502313,
		0.416289,
		0.330561,
		0.247986,
		0.171689,
		0.105044,
		0.05166,
		0.015366,
		0.000201,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.001742,
		0.022121,
		0.0622,
		0.118126,
		0.186245,
		0.263117,
		0.345531,
		0.430513,
		0.51535,
		0.597605,
		0.675142,
		0.746155,
		0.809195,
		0.863213,
		0.907596,
		0.942207,
		0.967429,
		0.984178,
		0.993888,
		0.998417,
		0.99983,
		1,
		0.999977,
		0.999385,
		0.996672,
		0.989805,
		0.976785,
		0.955947,
		0.926112,
		0.886629,
		0.837381,
		0.778747,
		0.711563,
		0.637084,
		0.556941,
		0.473108,
		0.387877,
		0.303823,
		0.223793,
		0.150882,
		0.088416,
		0.039943,
		0.00922,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.000192,
		0.014693,
		0.049471,
		0.100747,
		0.164939,
		0.238661,
		0.318742,
		0.402235,
		0.486437,
		0.568899,
		0.647454,
		0.720239,
		0.785723,
		0.842738,
		0.890522,
		0.928752,
		0.957583,
		0.977679,
		0.990212,
		0.99681,
		0.999412,
		0.999978,
		1,
		0.999838,
		0.998483,
		0.994136,
		0.984799,
		0.968661,
		0.944313,
		0.910833,
		0.867815,
		0.815347,
		0.753975,
		0.684672,
		0.608789,
		0.528027,
		0.444407,
		0.360237,
		0.278096,
		0.200812,
		0.131445,
		0.073275,
		0.029788,
		0.004665,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.008818,
		0.038254,
		0.084802,
		0.144944,
		0.215353,
		0.292901,
		0.374671,
		0.457972,
		0.540354,
		0.619625,
		0.693878,
		0.761512,
		0.82126,
		0.87223,
		0.913932,
		0.946324,
		0.969836,
		0.985389,
		0.994372,
		0.998546,
		0.999845,
		1,
		0.999979,
		0.999437,
		0.99694,
		0.990599,
		0.978531,
		0.959144,
		0.931276,
		0.894253,
		0.847883,
		0.792437,
		0.72861,
		0.657487,
		0.580504,
		0.499421,
		0.416289,
		0.333432,
		0.253422,
		0.179066,
		0.113386,
		0.059613,
		0.02117,
		0.001665,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.004462,
		0.028532,
		0.070284,
		0.126271,
		0.193222,
		0.268054,
		0.347883,
		0.430036,
		0.512066,
		0.591766,
		0.667192,
		0.736686,
		0.798899,
		0.852824,
		0.897829,
		0.933692,
		0.960634,
		0.979343,
		0.990967,
		0.997064,
		0.99946,
		0.99998,
		1,
		0.999851,
		0.998606,
		0.994596,
		0.985952,
		0.970957,
		0.948246,
		0.9169,
		0.876466,
		0.826947,
		0.768774,
		0.702773,
		0.630124,
		0.552335,
		0.47121,
		0.388827,
		0.307516,
		0.229837,
		0.158573,
		0.096707,
		0.047417,
		0.014064,
		0.000183,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.001593,
		0.02028,
		0.057184,
		0.108924,
		0.172287,
		0.244237,
		0.321925,
		0.402701,
		0.484124,
		0.563978,
		0.640295,
		0.711367,
		0.775775,
		0.832415,
		0.880531,
		0.919743,
		0.950084,
		0.972026,
		0.986488,
		0.994809,
		0.998663,
		0.999858,
		1,
		0.999981,
		0.999482,
		0.997182,
		0.991318,
		0.980117,
		0.962058,
		0.936004,
		0.901258,
		0.857572,
		0.805122,
		0.74448,
		0.676582,
		0.602692,
		0.524376,
		0.443475,
		0.362085,
		0.282534,
		0.207371,
		0.139346,
		0.081404,
		0.03667,
		0.008442,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.000176,
		0.013475,
		0.045488,
		0.092903,
		0.152563,
		0.221482,
		0.296847,
		0.376031,
		0.456608,
		0.536358,
		0.613292,
		0.685666,
		0.752002,
		0.811114,
		0.862134,
		0.904547,
		0.938217,
		0.963419,
		0.980856,
		0.991651,
		0.997294,
		0.999504,
		0.999981,
		1,
		0.999863,
		0.998717,
		0.995012,
		0.986998,
		0.973047,
		0.951842,
		0.922467,
		0.884434,
		0.837676,
		0.782523,
		0.71967,
		0.650148,
		0.575292,
		0.496716,
		0.416289,
		0.336118,
		0.258527,
		0.186045,
		0.121393,
		0.067471,
		0.027353,
		0.004273,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.008089,
		0.035182,
		0.078205,
		0.134062,
		0.199814,
		0.272688,
		0.350084,
		0.42959,
		0.508991,
		0.586284,
		0.659694,
		0.727694,
		0.78903,
		0.842738,
		0.888183,
		0.925079,
		0.953525,
		0.974023,
		0.987485,
		0.995205,
		0.998768,
		0.999869,
		1,
		0.999982,
		0.999524,
		0.9974,
		0.991969,
		0.981561,
		0.96472,
		0.940336,
		0.907703,
		0.86652,
		0.816885,
		0.759263,
		0.694455,
		0.623575,
		0.548017,
		0.469434,
		0.389717,
		0.310976,
		0.235528,
		0.165879,
		0.104717,
		0.054899,
		0.019444,
		0.001526,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.004095,
		0.026246,
		0.064824,
		0.11679,
		0.179254,
		0.249484,
		0.324909,
		0.403137,
		0.481957,
		0.559361,
		0.63355,
		0.702959,
		0.766271,
		0.822445,
		0.870737,
		0.910731,
		0.942367,
		0.965963,
		0.982234,
		0.992272,
		0.997501,
		0.999543,
		0.999983,
		1,
		0.999874,
		0.998817,
		0.995389,
		0.987949,
		0.974954,
		0.955135,
		0.927583,
		0.891784,
		0.84761,
		0.795304,
		0.735449,
		0.66894,
		0.596961,
		0.520954,
		0.442603,
		0.363815,
		0.286701,
		0.213563,
		0.146884,
		0.089317,
		0.043674,
		0.012921,
		0.000168,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.001462,
		0.01866,
		0.052748,
		0.100747,
		0.159817,
		0.227264,
		0.300551,
		0.377306,
		0.455329,
		0.532609,
		0.607332,
		0.677898,
		0.742944,
		0.801356,
		0.8523,
		0.895243,
		0.929985,
		0.956676,
		0.975845,
		0.988393,
		0.995564,
		0.998863,
		0.99988,
		1,
		0.999984,
		0.999561,
		0.997598,
		0.992561,
		0.982876,
		0.967154,
		0.944313,
		0.913638,
		0.874792,
		0.827802,
		0.773038,
		0.711187,
		0.643226,
		0.570395,
		0.49418,
		0.416289,
		0.338636,
		0.263329,
		0.192655,
		0.12907,
		0.07519,
		0.033782,
		0.007758,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.000161,
		0.012401,
		0.041967,
		0.085934,
		0.141513,
		0.206051,
		0.277047,
		0.352149,
		0.429172,
		0.506107,
		0.581127,
		0.65261,
		0.719149,
		0.779572,
		0.832964,
		0.878693,
		0.91643,
		0.946178,
		0.968292,
		0.98349,
		0.992836,
		0.997689,
		0.999578,
		0.999984,
		1,
		0.999884,
		0.998907,
		0.995731,
		0.988816,
		0.976697,
		0.958153,
		0.932289,
		0.898568,
		0.856815,
		0.807193,
		0.750188,
		0.686577,
		0.617404,
		0.543963,
		0.467769,
		0.39055,
		0.314225,
		0.240892,
		0.172822,
		0.112441,
		0.062328,
		0.025204,
		0.003928,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.007447,
		0.032464,
		0.072345,
		0.124349,
		0.185867,
		0.254429,
		0.327712,
		0.403546,
		0.479925,
		0.55502,
		0.627186,
		0.694984,
		0.757191,
		0.812825,
		0.861163,
		0.901765,
		0.934501,
		0.959568,
		0.977512,
		0.98922,
		0.99589,
		0.998949,
		0.999889,
		1,
		0.999985,
		0.999595,
		0.997777,
		0.993099,
		0.984077,
		0.969383,
		0.947967,
		0.919112,
		0.882447,
		0.837941,
		0.785882,
		0.726854,
		0.661713,
		0.591564,
		0.51774,
		0.441786,
		0.365439,
		0.290619,
		0.219416,
		0.154076,
		0.096998,
		0.050721,
		0.017921,
		0.001403,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.003771,
		0.024223,
		0.059973,
		0.108329,
		0.166724,
		0.232725,
		0.304035,
		0.378502,
		0.45413,
		0.529086,
		0.601712,
		0.670542,
		0.734311,
		0.791976,
		0.842738,
		0.886059,
		0.921688,
		0.949683,
		0.970427,
		0.984638,
		0.99335,
		0.99786,
		0.99961,
		0.999985,
		1,
		0.999893,
		0.998989,
		0.996042,
		0.989606,
		0.978292,
		0.960924,
		0.936624,
		0.904839,
		0.865352,
		0.81826,
		0.763962,
		0.70313,
		0.636686,
		0.565786,
		0.4918,
		0.416289,
		0.341002,
		0.267854,
		0.198921,
		0.136427,
		0.082738,
		0.040357,
		0.011912,
		0.000155,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.001347,
		0.017226,
		0.048808,
		0.093452,
		0.148635,
		0.211958,
		0.281153,
		0.354089,
		0.42878,
		0.503396,
		0.576269,
		0.645911,
		0.711022,
		0.770509,
		0.823506,
		0.869387,
		0.907796,
		0.938663,
		0.962225,
		0.979038,
		0.989975,
		0.996187,
		0.999027,
		0.999897,
		1,
		0.999986,
		0.999625,
		0.99794,
		0.99359,
		0.985175,
		0.971428,
		0.95133,
		0.924165,
		0.889537,
		0.847365,
		0.797864,
		0.741528,
		0.679106,
		0.61158,
		0.540147,
		0.466205,
		0.391333,
		0.317281,
		0.245958,
		0.179424,
		0.119879,
		0.069658,
		0.031222,
		0.007154,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.000148,
		0.011451,
		0.038838,
		0.079717,
		0.131607,
		0.192146,
		0.259096,
		0.330348,
		0.40393,
		0.478015,
		0.550931,
		0.621173,
		0.687413,
		0.748515,
		0.803552,
		0.851827,
		0.892886,
		0.926546,
		0.95291,
		0.972387,
		0.985689,
		0.993819,
		0.998016,
		0.999639,
		0.999987,
		1,
		0.999901,
		0.999064,
		0.996325,
		0.990328,
		0.979753,
		0.963472,
		0.940621,
		0.910641,
		0.873276,
		0.828569,
		0.776841,
		0.718669,
		0.654869,
		0.586474,
		0.514717,
		0.441017,
		0.366965,
		0.294311,
		0.224955,
		0.160938,
		0.104436,
		0.057749,
		0.023299,
		0.003623,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.006878,
		0.030049,
		0.067117,
		0.115644,
		0.173305,
		0.237891,
		0.307317,
		0.379627,
		0.453002,
		0.525768,
		0.596407,
		0.663568,
		0.726079,
		0.782964,
		0.833458,
		0.877025,
		0.913379,
		0.942504,
		0.964669,
		0.980438,
		0.990665,
		0.996457,
		0.999098,
		0.999905,
		1,
		0.999987,
		0.999653,
		0.998088,
		0.994038,
		0.986181,
		0.973306,
		0.954428,
		0.928835,
		0.896111,
		0.856131,
		0.809049,
		0.755278,
		0.69547,
		0.630499,
		0.561441,
		0.489559,
		0.416289,
		0.343228,
		0.272124,
		0.204865,
		0.143473,
		0.090096,
		0.047001,
		0.01657,
		0.001294,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.003484,
		0.022426,
		0.055645,
		0.100747,
		0.155444,
		0.217558,
		0.285028,
		0.355915,
		0.428411,
		0.500843,
		0.571684,
		0.639566,
		0.703287,
		0.761826,
		0.814362,
		0.860284,
		0.899218,
		0.931037,
		0.955885,
		0.974188,
		0.986651,
		0.994248,
		0.998157,
		0.999666,
		0.999988,
		1,
		0.999909,
		0.999131,
		0.996584,
		0.990988,
		0.981094,
		0.965817,
		0.944313,
		0.916014,
		0.880639,
		0.838179,
		0.788888,
		0.73326,
		0.672015,
		0.606074,
		0.536552,
		0.464734,
		0.39207,
		0.32016,
		0.250747,
		0.185705,
		0.127035,
		0.076857,
		0.037403,
		0.011016,
		0.000143,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.001245,
		0.015951,
		0.045292,
		0.086916,
		0.138573,
		0.198115,
		0.263508,
		0.332833,
		0.404291,
		0.476215,
		0.547074,
		0.615483,
		0.680218,
		0.74022,
		0.794618,
		0.842738,
		0.884124,
		0.918552,
		0.946052,
		0.96692,
		0.981724,
		0.991297,
		0.996704,
		0.999163,
		0.999912,
		1,
		0.999988,
		0.999678,
		0.998223,
		0.994448,
		0.987103,
		0.975034,
		0.957286,
		0.933156,
		0.902212,
		0.864293,
		0.819497,
		0.768167,
		0.710869,
		0.64838,
		0.581665,
		0.511868,
		0.440294,
		0.368403,
		0.297795,
		0.230203,
		0.167488,
		0.111627,
		0.064712,
		0.028941,
		0.006618,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.000137,
		0.010606,
		0.036046,
		0.074147,
		0.122695,
		0.179577,
		0.242783,
		0.310415,
		0.380687,
		0.451939,
		0.522638,
		0.591389,
		0.656948,
		0.718226,
		0.774307,
		0.824461,
		0.868161,
		0.905097,
		0.935195,
		0.958631,
		0.975845,
		0.987535,
		0.99464,
		0.998286,
		0.99969,
		0.999988,
		1,
		0.999915,
		0.999193,
		0.99682,
		0.991594,
		0.982327,
		0.967979,
		0.947725,
		0.920996,
		0.887485,
		0.847143,
		0.800163,
		0.746966,
		0.688183,
		0.62464,
		0.557338,
		0.487448,
		0.416289,
		0.345327,
		0.27616,
		0.210511,
		0.150223,
		0.09725,
		0.053654,
		0.021601,
		0.003352,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.006372,
		0.027894,
		0.062433,
		0.107814,
		0.161955,
		0.222873,
		0.28869,
		0.357638,
		0.428063,
		0.498434,
		0.567351,
		0.63355,
		0.695919,
		0.753505,
		0.805528,
		0.851398,
		0.890728,
		0.92335,
		0.949334,
		0.968996,
		0.982906,
		0.991877,
		0.99693,
		0.999222,
		0.999918,
		1,
		0.999989,
		0.999701,
		0.998347,
		0.994823,
		0.987949,
		0.976625,
		0.959925,
		0.937158,
		0.907879,
		0.871897,
		0.829262,
		0.780253,
		0.725363,
		0.665277,
		0.600863,
		0.533157,
		0.463346,
		0.392764,
		0.322879,
		0.255282,
		0.191686,
		0.133916,
		0.083901,
		0.043674,
		0.015366,
		0.001198,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.003228,
		0.020821,
		0.051768,
		0.09393,
		0.145256,
		0.203793,
		0.267685,
		0.335179,
		0.404632,
		0.474517,
		0.543429,
		0.610092,
		0.673374,
		0.732288,
		0.786014,
		0.833904,
		0.875503,
		0.910561,
		0.939048,
		0.961169,
		0.977373,
		0.988347,
		0.994999,
		0.998404,
		0.999712,
		0.999989,
		1,
		0.999921,
		0.999249,
		0.997036,
		0.992149,
		0.983462,
		0.969973,
		0.950882,
		0.925619,
		0.893857,
		0.855511,
		0.810721,
		0.759844,
		0.703432,
		0.642221,
		0.577116,
		0.509177,
		0.439612,
		0.369759,
		0.301087,
		0.235182,
		0.173742,
		0.118572,
		0.071577,
		0.034761,
		0.010218,
		0.000132,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.001154,
		0.014812,
		0.042142,
		0.081039,
		0.129487,
		0.185559,
		0.247423,
		0.313342,
		0.381687,
		0.450936,
		0.51968,
		0.586639,
		0.650658,
		0.710728,
		0.76599,
		0.815748,
		0.859486,
		0.896877,
		0.927805,
		0.952372,
		0.970913,
		0.983995,
		0.99241,
		0.997137,
		0.999276,
		0.999924,
		1,
		0.99999,
		0.999722,
		0.99846,
		0.995167,
		0.988728,
		0.978092,
		0.962366,
		0.940869,
		0.913148,
		0.878987,
		0.838394,
		0.791594,
		0.739008,
		0.681244,
		0.619082,
		0.553458,
		0.485454,
		0.416289,
		0.34731,
		0.27998,
		0.215878,
		0.156687,
		0.104192,
		0.060273,
		0.026902,
		0.00614,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.000127,
		0.009851,
		0.033544,
		0.069138,
		0.114651,
		0.168182,
		0.227922,
		0.292156,
		0.359265,
		0.427734,
		0.496159,
		0.563249,
		0.627839,
		0.688896,
		0.745529,
		0.797,
		0.842738,
		0.882352,
		0.915644,
		0.942622,
		0.963517,
		0.978782,
		0.989094,
		0.995329,
		0.998512,
		0.999732,
		0.99999,
		1,
		0.999927,
		0.999301,
		0.997234,
		0.99266,
		0.984507,
		0.971816,
		0.953807,
		0.929913,
		0.899793,
		0.863328,
		0.820615,
		0.771949,
		0.717815,
		0.658868,
		0.595924,
		0.529947,
		0.462036,
		0.39342,
		0.325448,
		0.259581,
		0.197385,
		0.140529,
		0.090775,
		0.049979,
		0.020082,
		0.003111,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.00592,
		0.025962,
		0.058222,
		0.100747,
		0.151669,
		0.2092,
		0.271644,
		0.337397,
		0.404954,
		0.472913,
		0.539979,
		0.604978,
		0.666857,
		0.724699,
		0.777728,
		0.825327,
		0.867043,
		0.902608,
		0.931946,
		0.955189,
		0.972685,
		0.984999,
		0.992899,
		0.997327,
		0.999325,
		0.999929,
		1,
		0.99999,
		0.999741,
		0.998563,
		0.995483,
		0.989445,
		0.979446,
		0.964625,
		0.944313,
		0.918052,
		0.885604,
		0.846941,
		0.802239,
		0.751857,
		0.696335,
		0.636368,
		0.572806,
		0.506634,
		0.438967,
		0.371041,
		0.304203,
		0.23991,
		0.179716,
		0.125272,
		0.07832,
		0.040688,
		0.014288,
		0.001112,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.003,
		0.019382,
		0.048281,
		0.087777,
		0.136026,
		0.191269,
		0.251827,
		0.316114,
		0.382633,
		0.449987,
		0.516882,
		0.582134,
		0.644676,
		0.703566,
		0.758,
		0.807315,
		0.851009,
		0.888746,
		0.920376,
		0.945942,
		0.965692,
		0.980084,
		0.989783,
		0.995632,
		0.998612,
		0.99975,
		0.999991,
		1,
		0.999932,
		0.999348,
		0.997416,
		0.99313,
		0.985471,
		0.973521,
		0.95652,
		0.933906,
		0.905327,
		0.870636,
		0.82989,
		0.783333,
		0.731385,
		0.674631,
		0.613805,
		0.549783,
		0.483569,
		0.416289,
		0.349184,
		0.283601,
		0.220985,
		0.16288,
		0.110919,
		0.066821,
		0.032389,
		0.009504,
		0.000123,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.001072,
		0.013791,
		0.039308,
		0.075735,
		0.121258,
		0.17414,
		0.232725,
		0.295441,
		0.360804,
		0.427424,
		0.494005,
		0.559361,
		0.622411,
		0.682196,
		0.737881,
		0.788769,
		0.834309,
		0.874111,
		0.907953,
		0.935798,
		0.957802,
		0.974325,
		0.985926,
		0.99335,
		0.997501,
		0.99937,
		0.999934,
		1,
		0.999991,
		0.999759,
		0.998658,
		0.995774,
		0.990107,
		0.980698,
		0.966719,
		0.947512,
		0.92262,
		0.891784,
		0.854945,
		0.812236,
		0.763962,
		0.710598,
		0.652767,
		0.591237,
		0.526907,
		0.460796,
		0.39404,
		0.327881,
		0.263662,
		0.20282,
		0.146884,
		0.09747,
		0.056273,
		0.02507,
		0.005711,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.000118,
		0.009174,
		0.031293,
		0.064618,
		0.107365,
		0.157821,
		0.214352,
		0.275401,
		0.339498,
		0.405259,
		0.471394,
		0.536711,
		0.60012,
		0.660647,
		0.717434,
		0.769751,
		0.817006,
		0.858756,
		0.89472,
		0.924786,
		0.949027,
		0.967708,
		0.981289,
		0.990418,
		0.995911,
		0.998703,
		0.999767,
		0.999991,
		1,
		0.999937,
		0.999391,
		0.997583,
		0.993563,
		0.986362,
		0.9751,
		0.959038,
		0.937623,
		0.91049,
		0.877473,
		0.838591,
		0.794042,
		0.744191,
		0.689557,
		0.630801,
		0.568718,
		0.504225,
		0.438357,
		0.372254,
		0.307157,
		0.244406,
		0.185427,
		0.131734,
		0.084923,
		0.046668,
		0.018718,
		0.002895,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.005514,
		0.024223,
		0.054421,
		0.094348,
		0.14232,
		0.196721,
		0.256014,
		0.318742,
		0.383529,
		0.449089,
		0.51423,
		0.577857,
		0.638979,
		0.69672,
		0.750323,
		0.799158,
		0.842738,
		0.880725,
		0.912942,
		0.939384,
		0.96023,
		0.975845,
		0.986782,
		0.993766,
		0.997662,
		0.999412,
		0.999939,
		1,
		0.999992,
		0.999775,
		0.998745,
		0.996042,
		0.990717,
		0.981857,
		0.968661,
		0.950488,
		0.926878,
		0.897559,
		0.862446,
		0.82163,
		0.77537,
		0.724081,
		0.668323,
		0.608789,
		0.546297,
		0.481783,
		0.416289,
		0.35096,
		0.287037,
		0.225849,
		0.168815,
		0.117431,
		0.073275,
		0.037997,
		0.01332,
		0.001035,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.002795,
		0.018088,
		0.045134,
		0.082205,
		0.127639,
		0.179843,
		0.237298,
		0.298559,
		0.362263,
		0.427129,
		0.491965,
		0.555671,
		0.617247,
		0.675798,
		0.730545,
		0.780826,
		0.826114,
		0.866019,
		0.900305,
		0.928899,
		0.951898,
		0.96958,
		0.982404,
		0.991005,
		0.996168,
		0.998786,
		0.999782,
		0.999992,
		1,
		0.999941,
		0.999431,
		0.997737,
		0.993962,
		0.987186,
		0.976563,
		0.961379,
		0.941085,
		0.915312,
		0.883873,
		0.846758,
		0.804122,
		0.75628,
		0.703692,
		0.646952,
		0.586783,
		0.524024,
		0.459622,
		0.394628,
		0.330188,
		0.26754,
		0.208006,
		0.152992,
		0.103978,
		0.062522,
		0.030251,
		0.008861,
		0.000114,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.000999,
		0.012873,
		0.036751,
		0.070933,
		0.113781,
		0.163726,
		0.219266,
		0.278972,
		0.34149,
		0.405548,
		0.469955,
		0.533608,
		0.595501,
		0.654724,
		0.710476,
		0.762069,
		0.808939,
		0.850653,
		0.88692,
		0.917603,
		0.942727,
		0.962487,
		0.977255,
		0.987575,
		0.99415,
		0.997809,
		0.99945,
		0.999943,
		1,
		0.999992,
		0.99979,
		0.998826,
		0.996289,
		0.991281,
		0.982931,
		0.970465,
		0.953258,
		0.930852,
		0.902962,
		0.869479,
		0.830462,
		0.786125,
		0.736831,
		0.68308,
		0.625499,
		0.564834,
		0.501941,
		0.437779,
		0.373403,
		0.309961,
		0.248685,
		0.190888,
		0.137962,
		0.091374,
		0.052658,
		0.023419,
		0.005327,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.00011,
		0.008564,
		0.029261,
		0.060526,
		0.100747,
		0.148377,
		0.201932,
		0.259999,
		0.321236,
		0.384378,
		0.448238,
		0.511714,
		0.573791,
		0.63355,
		0.690172,
		0.742944,
		0.79127,
		0.834679,
		0.872831,
		0.905532,
		0.932739,
		0.954571,
		0.971319,
		0.983438,
		0.991548,
		0.996405,
		0.998863,
		0.999797,
		0.999993,
		1,
		0.999945,
		0.999468,
		0.997879,
		0.994331,
		0.987949,
		0.977922,
		0.963556,
		0.944313,
		0.919818,
		0.889869,
		0.854428,
		0.813614,
		0.767695,
		0.717079,
		0.6623,
		0.604014,
		0.542987,
		0.48009,
		0.416289,
		0.352645,
		0.290302,
		0.230487,
		0.174504,
		0.12373,
		0.079615,
		0.043674,
		0.017488,
		0.0027,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.005149,
		0.022654,
		0.050979,
		0.088536,
		0.133799,
		0.185305,
		0.241656,
		0.301522,
		0.363647,
		0.42685,
		0.490028,
		0.552163,
		0.612327,
		0.669685,
		0.723506,
		0.773163,
		0.818151,
		0.858088,
		0.892725,
		0.921961,
		0.945845,
		0.964588,
		0.978564,
		0.98831,
		0.994505,
		0.997946,
		0.999485,
		0.999947,
		1,
		0.999993,
		0.999803,
		0.9989,
		0.996517,
		0.991804,
		0.983926,
		0.972142,
		0.955839,
		0.934562,
		0.908019,
		0.876079,
		0.83877,
		0.796268,
		0.748889,
		0.697079,
		0.641405,
		0.582546,
		0.521286,
		0.458508,
		0.395185,
		0.332378,
		0.27123,
		0.212959,
		0.158861,
		0.110298,
		0.0687,
		0.035564,
		0.012447,
		0.000966,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.00261,
		0.016918,
		0.042284,
		0.077145,
		0.119997,
		0.169394,
		0.223957,
		0.282369,
		0.343382,
		0.405822,
		0.468589,
		0.530661,
		0.591103,
		0.649069,
		0.703808,
		0.754672,
		0.801123,
		0.842738,
		0.879226,
		0.910426,
		0.936326,
		0.957064,
		0.972936,
		0.984397,
		0.99205,
		0.996625,
		0.998934,
		0.99981,
		0.999993,
		1,
		0.999948,
		0.999501,
		0.99801,
		0.994673,
		0.988657,
		0.979184,
		0.965583,
		0.947325,
		0.924033,
		0.895491,
		0.861636,
		0.822556,
		0.778478,
		0.729761,
		0.676885,
		0.620445,
		0.561141,
		0.499772,
		0.43723,
		0.374495,
		0.312626,
		0.252761,
		0.196114,
		0.143965,
		0.097664,
		0.058623,
		0.028318,
		0.008282,
		0.000107,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.000934,
		0.012042,
		0.034434,
		0.066571,
		0.10697,
		0.154206,
		0.206915,
		0.263794,
		0.323607,
		0.385184,
		0.44743,
		0.509323,
		0.569922,
		0.628371,
		0.683903,
		0.73585,
		0.783645,
		0.826833,
		0.865077,
		0.89817,
		0.926039,
		0.948756,
		0.966545,
		0.979781,
		0.988991,
		0.994834,
		0.998072,
		0.999517,
		0.99995,
		1,
		0.999993,
		0.999816,
		0.998968,
		0.996728,
		0.992288,
		0.984851,
		0.973702,
		0.958246,
		0.938031,
		0.912756,
		0.882275,
		0.846589,
		0.805839,
		0.760297,
		0.710363,
		0.656545,
		0.599465,
		0.53984,
		0.478481,
		0.416289,
		0.354245,
		0.293408,
		0.234912,
		0.179959,
		0.12982,
		0.085828,
		0.049379,
		0.021925,
		0.004979,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.000103,
		0.008014,
		0.02742,
		0.056809,
		0.094719,
		0.139745,
		0.190539,
		0.245813,
		0.304341,
		0.364963,
		0.426585,
		0.488188,
		0.548826,
		0.607637,
		0.66384,
		0.716747,
		0.76577,
		0.810421,
		0.850326,
		0.885232,
		0.915012,
		0.93968,
		0.959389,
		0.974442,
		0.985289,
		0.992516,
		0.996827,
		0.999,
		0.999822,
		0.999993,
		1,
		0.999952,
		0.999532,
		0.998131,
		0.994989,
		0.989313,
		0.980357,
		0.967473,
		0.950139,
		0.927979,
		0.900765,
		0.868415,
		0.830985,
		0.788668,
		0.741777,
		0.690745,
		0.636109,
		0.578511,
		0.518684,
		0.45745,
		0.395714,
		0.33446,
		0.274744,
		0.217694,
		0.164503,
		0.116428,
		0.074788,
		0.040959,
		0.016376,
		0.002525,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.004818,
		0.021232,
		0.047852,
		0.083242,
		0.126014,
		0.174837,
		0.228439,
		0.285604,
		0.345181,
		0.406083,
		0.46729,
		0.527856,
		0.586911,
		0.643665,
		0.697414,
		0.747548,
		0.793551,
		0.835018,
		0.871652,
		0.90328,
		0.929857,
		0.951476,
		0.968369,
		0.980913,
		0.989624,
		0.995138,
		0.998188,
		0.999547,
		0.999953,
		1,
		0.999994,
		0.999827,
		0.999031,
		0.996923,
		0.992736,
		0.985711,
		0.975156,
		0.960493,
		0.941275,
		0.917197,
		0.888097,
		0.853952,
		0.814873,
		0.771094,
		0.722968,
		0.670956,
		0.615622,
		0.557624,
		0.497709,
		0.436709,
		0.375532,
		0.315161,
		0.25665,
		0.201118,
		0.14975,
		0.103789,
		0.064539,
		0.033357,
		0.011657,
		0.000903,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.002443,
		0.015859,
		0.039695,
		0.072536,
		0.113014,
		0.159817,
		0.211684,
		0.267414,
		0.325863,
		0.385951,
		0.446662,
		0.507048,
		0.566235,
		0.623424,
		0.677898,
		0.729027,
		0.776274,
		0.8192,
		0.857473,
		0.890876,
		0.919314,
		0.942819,
		0.96156,
		0.975845,
		0.986118,
		0.992949,
		0.997015,
		0.999061,
		0.999833,
		0.999994,
		1,
		0.999955,
		0.999561,
		0.998243,
		0.995282,
		0.989923,
		0.98145,
		0.969235,
		0.952769,
		0.931675,
		0.905717,
		0.874792,
		0.838934,
		0.7983,
		0.753166,
		0.703917,
		0.651042,
		0.595126,
		0.536843,
		0.476952,
		0.416289,
		0.355767,
		0.296367,
		0.239139,
		0.185194,
		0.135705,
		0.091904,
		0.055077,
		0.026564,
		0.007758,
		0.0001,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.000874,
		0.01129,
		0.03233,
		0.062598,
		0.100747,
		0.145482,
		0.195557,
		0.249783,
		0.307026,
		0.366214,
		0.426333,
		0.486437,
		0.545648,
		0.60316,
		0.658245,
		0.710256,
		0.758636,
		0.802918,
		0.842738,
		0.877839,
		0.908079,
		0.933436,
		0.95402,
		0.970072,
		0.981968,
		0.990212,
		0.995421,
		0.998296,
		0.999575,
		0.999956,
		1,
		0.999994,
		0.999838,
		0.999089,
		0.997104,
		0.993153,
		0.98651,
		0.976511,
		0.962592,
		0.944313,
		0.921364,
		0.893571,
		0.86089,
		0.823405,
		0.781314,
		0.734932,
		0.684672,
		0.631048,
		0.574663,
		0.516206,
		0.456443,
		0.396218,
		0.336442,
		0.278096,
		0.222223,
		0.169927,
		0.122371,
		0.08077,
		0.046396,
		0.02057,
		0.004665,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		9.6e-005,
		0.007514,
		0.025748,
		0.053423,
		0.089212,
		0.131836,
		0.180066,
		0.232725,
		0.28869,
		0.346894,
		0.406331,
		0.466054,
		0.525185,
		0.582912,
		0.638497,
		0.69128,
		0.740684,
		0.78622,
		0.827492,
		0.864209,
		0.896184,
		0.92335,
		0.945759,
		0.96359,
		0.977154,
		0.986889,
		0.99335,
		0.997189,
		0.999117,
		0.999843,
		0.999994,
		1,
		0.999957,
		0.999588,
		0.998347,
		0.995554,
		0.99049,
		0.982468,
		0.970881,
		0.95523,
		0.935141,
		0.910368,
		0.880796,
		0.846434,
		0.80741,
		0.763962,
		0.716437,
		0.665277,
		0.611015,
		0.554272,
		0.495746,
		0.436213,
		0.376519,
		0.317576,
		0.260362,
		0.205913,
		0.155325,
		0.109748,
		0.070384,
		0.038489,
		0.015366,
		0.002366,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.004519,
		0.019939,
		0.045004,
		0.078406,
		0.118882,
		0.165218,
		0.216251,
		0.270869,
		0.328012,
		0.386681,
		0.44593,
		0.504881,
		0.562719,
		0.618696,
		0.672143,
		0.722464,
		0.769149,
		0.811778,
		0.850025,
		0.883666,
		0.912588,
		0.936792,
		0.956401,
		0.971663,
		0.982951,
		0.990759,
		0.995683,
		0.998396,
		0.9996,
		0.999959,
		1,
		0.999994,
		0.999848,
		0.999143,
		0.997272,
		0.993541,
		0.987255,
		0.977776,
		0.964555,
		0.947158,
		0.925275,
		0.89872,
		0.867431,
		0.831466,
		0.790993,
		0.746289,
		0.697728,
		0.645775,
		0.590984,
		0.533987,
		0.475496,
		0.416289,
		0.357216,
		0.299189,
		0.243179,
		0.190218,
		0.141391,
		0.097836,
		0.060743,
		0.031349,
		0.01094,
		0.000846,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		0.002292,
		0.014896,
		0.037337,
		0.068326,
		0.106619,
		0.151018,
		0.200371,
		0.253577,
		0.309587,
		0.367405,
		0.426093,
		0.484769,
		0.542616,
		0.598882,
		0.652886,
		0.704019,
		0.751752,
		0.79564,
		0.835329,
		0.870561,
		0.90118,
		0.927141,
		0.948514,
		0.965488,
		0.978376,
		0.987608,
		0.993724,
		0.997351,
		0.999169,
		0.999852,
		0.999995,
		1,
		0.99996,
		0.999612,
		0.998443,
		0.995807,
		0.991018,
		0.983418,
		0.972419,
		0.957534,
		0.938392,
		0.914741,
		0.886452,
		0.853514,
		0.816028,
		0.774201,
		0.72834,
		0.678846,
		0.626207,
		0.570991,
		0.513845,
		0.455485,
		0.396697,
		0.33833,
		0.281295,
		0.226559,
		0.175145,
		0.128128,
		0.086636,
		0.051842,
		0.024968,
		0.007282,
		9.3e-005,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.00082,
		0.010606,
		0.030412,
		0.058969,
		0.095049,
		0.137468,
		0.185091,
		0.236827,
		0.291635,
		0.348527,
		0.406567,
		0.464877,
		0.522638,
		0.579092,
		0.63355,
		0.685391,
		0.73407,
		0.779121,
		0.820162,
		0.856905,
		0.889157,
		0.916829,
		0.939942,
		0.958631,
		0.97315,
		0.983869,
		0.991269,
		0.995927,
		0.998488,
		0.999624,
		0.999961,
		1,
		0.999995,
		0.999857,
		0.999193,
		0.997428,
		0.993901,
		0.987949,
		0.978957,
		0.966392,
		0.949827,
		0.92895,
		0.903567,
		0.873601,
		0.839085,
		0.800163,
		0.757075,
		0.710157,
		0.659834,
		0.606611,
		0.551073,
		0.493875,
		0.435741,
		0.377459,
		0.31988,
		0.26391,
		0.210511,
		0.160699,
		0.115538,
		0.076143,
		0.043674,
		0.019337,
		0.004379,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		9.1e-005,
		0.00706,
		0.024223,
		0.05033,
		0.084169,
		0.124573,
		0.170419,
		0.220628,
		0.27417,
		0.330063,
		0.387376,
		0.445233,
		0.502815,
		0.559361,
		0.614173,
		0.666622,
		0.716147,
		0.762263,
		0.804566,
		0.842738,
		0.876554,
		0.905884,
		0.930704,
		0.951098,
		0.967266,
		0.979518,
		0.988279,
		0.994072,
		0.997501,
		0.999217,
		0.999861,
		0.999995,
		1,
		0.999962,
		0.999635,
		0.998532,
		0.996042,
		0.99151,
		0.984305,
		0.973858,
		0.959694,
		0.941445,
		0.918855,
		0.891784,
		0.860201,
		0.824185,
		0.783913,
		0.739658,
		0.691782,
		0.64073,
		0.587025,
		0.531263,
		0.474108,
		0.416289,
		0.358598,
		0.301882,
		0.247045,
		0.195044,
		0.146884,
		0.103621,
		0.066356,
		0.036235,
		0.014447,
		0.002221,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.001428,
		0.01237,
		0.033031,
		0.062177,
		0.098612,
		0.141187,
		0.188793,
		0.24037,
		0.294906,
		0.351437,
		0.409054,
		0.466901,
		0.52418,
		0.580154,
		0.63415,
		0.685564,
		0.733863,
		0.778592,
		0.819379,
		0.855939,
		0.888082,
		0.915718,
		0.938862,
		0.957642,
		0.972299,
		0.983187,
		0.990768,
		0.995601,
		0.99831,
		0.999551,
		0.999945,
		1,
		0.999998,
		0.999896,
		0.999343,
		0.997796,
		0.994617,
		0.989154,
		0.980795,
		0.969004,
		0.953345,
		0.933492,
		0.909232,
		0.880467,
		0.847207,
		0.809569,
		0.767766,
		0.722107,
		0.672983,
		0.620871,
		0.566321,
		0.509955,
		0.452462,
		0.394596,
		0.337168,
		0.281048,
		0.227159,
		0.176478,
		0.130029,
		0.088884,
		0.054161,
		0.027023,
		0.008674,
		0.000361,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.001451,
		0.01256,
		0.033529,
		0.063092,
		0.100028,
		0.14316,
		0.191355,
		0.24353,
		0.298652,
		0.355737,
		0.413858,
		0.472143,
		0.529781,
		0.586022,
		0.640185,
		0.691659,
		0.739908,
		0.784477,
		0.824995,
		0.861185,
		0.892867,
		0.919966,
		0.942516,
		0.960668,
		0.974689,
		0.984964,
		0.991988,
		0.996348,
		0.998695,
		0.999698,
		0.999974,
		1,
		0.99999,
		0.999813,
		0.999046,
		0.997097,
		0.993303,
		0.987004,
		0.977591,
		0.964547,
		0.947458,
		0.926026,
		0.900072,
		0.869533,
		0.834457,
		0.794999,
		0.751412,
		0.704043,
		0.653328,
		0.599783,
		0.544,
		0.486644,
		0.428447,
		0.370205,
		0.312776,
		0.257074,
		0.204067,
		0.154777,
		0.110275,
		0.071682,
		0.040162,
		0.016928,
		0.003233,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.001474,
		0.012755,
		0.034038,
		0.064028,
		0.101474,
		0.145174,
		0.193969,
		0.246752,
		0.302467,
		0.360112,
		0.41874,
		0.477463,
		0.535457,
		0.59196,
		0.646281,
		0.697803,
		0.745987,
		0.790377,
		0.830607,
		0.866407,
		0.897607,
		0.924149,
		0.946088,
		0.963598,
		0.976975,
		0.986637,
		0.993109,
		0.997011,
		0.999017,
		0.999807,
		0.99999,
		1,
		0.999973,
		0.999688,
		0.998655,
		0.99624,
		0.991758,
		0.984543,
		0.973997,
		0.959621,
		0.941027,
		0.91795,
		0.890244,
		0.857883,
		0.820956,
		0.779657,
		0.73428,
		0.685216,
		0.63294,
		0.578013,
		0.52107,
		0.46282,
		0.40404,
		0.345573,
		0.288319,
		0.233241,
		0.181354,
		0.133727,
		0.09148,
		0.055783,
		0.027851,
		0.008946,
		0.000372,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.001497,
		0.012955,
		0.034559,
		0.064984,
		0.102952,
		0.14723,
		0.196635,
		0.250035,
		0.306352,
		0.364563,
		0.423701,
		0.482863,
		0.541209,
		0.597967,
		0.652437,
		0.703994,
		0.752097,
		0.796291,
		0.836212,
		0.8716,
		0.902298,
		0.928264,
		0.949573,
		0.966428,
		0.979155,
		0.988203,
		0.994133,
		0.997593,
		0.999281,
		0.999886,
		0.999997,
		1,
		0.99994,
		0.999508,
		0.998153,
		0.995203,
		0.989954,
		0.981742,
		0.969981,
		0.954194,
		0.934022,
		0.909232,
		0.879718,
		0.845492,
		0.806683,
		0.763528,
		0.716363,
		0.665622,
		0.611825,
		0.555576,
		0.497557,
		0.438523,
		0.379296,
		0.320766,
		0.263882,
		0.209654,
		0.159146,
		0.113477,
		0.073818,
		0.041389,
		0.017457,
		0.003336,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.001521,
		0.013159,
		0.035092,
		0.065963,
		0.104462,
		0.14933,
		0.199356,
		0.253383,
		0.310309,
		0.369091,
		0.428742,
		0.488343,
		0.547038,
		0.604044,
		0.658652,
		0.710231,
		0.758237,
		0.802214,
		0.841807,
		0.876762,
		0.906935,
		0.932304,
		0.952968,
		0.969156,
		0.981226,
		0.989663,
		0.99506,
		0.998097,
		0.999493,
		0.999938,
		1,
		0.999997,
		0.999882,
		0.999259,
		0.99752,
		0.993959,
		0.987863,
		0.97857,
		0.965511,
		0.948233,
		0.926409,
		0.899843,
		0.868468,
		0.832336,
		0.791619,
		0.746597,
		0.697651,
		0.645261,
		0.589991,
		0.532493,
		0.473495,
		0.413799,
		0.354276,
		0.295864,
		0.239562,
		0.18643,
		0.137583,
		0.094192,
		0.057479,
		0.028718,
		0.00923,
		0.000384,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.001546,
		0.013368,
		0.035637,
		0.066963,
		0.106005,
		0.151474,
		0.202132,
		0.256796,
		0.31434,
		0.373698,
		0.433865,
		0.493904,
		0.552945,
		0.610192,
		0.664927,
		0.716513,
		0.764404,
		0.808146,
		0.847388,
		0.881887,
		0.911514,
		0.936266,
		0.956268,
		0.971777,
		0.983187,
		0.991015,
		0.995893,
		0.998528,
		0.999658,
		0.999971,
		1,
		0.999989,
		0.999788,
		0.998923,
		0.996733,
		0.992484,
		0.985457,
		0.974996,
		0.960553,
		0.941704,
		0.918157,
		0.889753,
		0.856465,
		0.81839,
		0.775744,
		0.728851,
		0.678141,
		0.624138,
		0.567455,
		0.50879,
		0.448922,
		0.388702,
		0.32905,
		0.270954,
		0.215467,
		0.163699,
		0.116819,
		0.076051,
		0.042672,
		0.018011,
		0.003444,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.001571,
		0.013582,
		0.036195,
		0.067986,
		0.107582,
		0.153664,
		0.204965,
		0.260276,
		0.318445,
		0.378386,
		0.439072,
		0.499548,
		0.55893,
		0.61641,
		0.67126,
		0.722839,
		0.770596,
		0.814083,
		0.852952,
		0.886971,
		0.916031,
		0.940146,
		0.959469,
		0.974289,
		0.985035,
		0.992261,
		0.996633,
		0.998889,
		0.999782,
		0.999989,
		1,
		0.99997,
		0.999647,
		0.998482,
		0.995768,
		0.990749,
		0.982702,
		0.970985,
		0.955074,
		0.934575,
		0.909232,
		0.878929,
		0.843682,
		0.803632,
		0.759042,
		0.710282,
		0.657831,
		0.60226,
		0.544233,
		0.484498,
		0.423884,
		0.363292,
		0.303696,
		0.246136,
		0.191718,
		0.141606,
		0.097025,
		0.059254,
		0.029626,
		0.009529,
		0.000397,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.001597,
		0.013802,
		0.036766,
		0.069032,
		0.109195,
		0.155901,
		0.207857,
		0.263824,
		0.322628,
		0.383156,
		0.444363,
		0.505276,
		0.564995,
		0.622699,
		0.677651,
		0.729206,
		0.776812,
		0.820021,
		0.858494,
		0.892012,
		0.92048,
		0.943938,
		0.962567,
		0.976689,
		0.986769,
		0.9934,
		0.997284,
		0.999187,
		0.99987,
		0.999997,
		1,
		0.999932,
		0.999443,
		0.997915,
		0.994599,
		0.988724,
		0.979567,
		0.966504,
		0.949038,
		0.926809,
		0.899603,
		0.867344,
		0.830094,
		0.788042,
		0.741497,
		0.690882,
		0.636722,
		0.57964,
		0.520351,
		0.459655,
		0.398433,
		0.33764,
		0.278304,
		0.22152,
		0.168448,
		0.120309,
		0.078386,
		0.044016,
		0.018592,
		0.003558,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.001624,
		0.014026,
		0.037351,
		0.070103,
		0.110843,
		0.158187,
		0.210809,
		0.267443,
		0.326889,
		0.38801,
		0.449741,
		0.511089,
		0.571138,
		0.629057,
		0.6841,
		0.735615,
		0.783049,
		0.825958,
		0.864011,
		0.897002,
		0.924857,
		0.947639,
		0.965558,
		0.978974,
		0.988388,
		0.994434,
		0.99785,
		0.999426,
		0.99993,
		1,
		0.999997,
		0.999866,
		0.999161,
		0.997199,
		0.993199,
		0.986377,
		0.976017,
		0.961517,
		0.94241,
		0.918374,
		0.889236,
		0.854966,
		0.815675,
		0.771599,
		0.723099,
		0.670646,
		0.61482,
		0.556295,
		0.495839,
		0.434306,
		0.372632,
		0.311827,
		0.252975,
		0.197229,
		0.145806,
		0.099986,
		0.061111,
		0.030578,
		0.009842,
		0.00041,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.001651,
		0.014256,
		0.03795,
		0.071199,
		0.112529,
		0.160523,
		0.213823,
		0.271135,
		0.331231,
		0.39295,
		0.455206,
		0.516987,
		0.577362,
		0.635486,
		0.690605,
		0.742062,
		0.789303,
		0.831889,
		0.869498,
		0.901939,
		0.929157,
		0.951243,
		0.968439,
		0.98114,
		0.989891,
		0.995364,
		0.998334,
		0.999612,
		0.999967,
		1,
		0.999988,
		0.99976,
		0.99878,
		0.99631,
		0.991538,
		0.983674,
		0.972017,
		0.955987,
		0.935153,
		0.909232,
		0.878098,
		0.841768,
		0.800401,
		0.754286,
		0.703836,
		0.649575,
		0.592135,
		0.532247,
		0.470734,
		0.408505,
		0.346552,
		0.285944,
		0.227824,
		0.173402,
		0.123957,
		0.08083,
		0.045425,
		0.019201,
		0.003677,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.001679,
		0.014492,
		0.038563,
		0.07232,
		0.114253,
		0.16291,
		0.2169,
		0.2749,
		0.335654,
		0.397978,
		0.460761,
		0.522972,
		0.583667,
		0.641985,
		0.697165,
		0.748546,
		0.795573,
		0.837812,
		0.874952,
		0.906817,
		0.933375,
		0.954745,
		0.971204,
		0.983187,
		0.991278,
		0.996194,
		0.998741,
		0.999752,
		0.999987,
		1,
		0.999966,
		0.999599,
		0.99828,
		0.995219,
		0.989583,
		0.980581,
		0.967529,
		0.949876,
		0.927229,
		0.89935,
		0.866157,
		0.827721,
		0.78425,
		0.736089,
		0.683702,
		0.627672,
		0.568684,
		0.507526,
		0.44508,
		0.382312,
		0.320274,
		0.260094,
		0.202976,
		0.150192,
		0.103084,
		0.063057,
		0.031577,
		0.010171,
		0.000424,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.001708,
		0.014734,
		0.039191,
		0.073468,
		0.116017,
		0.165349,
		0.220042,
		0.278741,
		0.340162,
		0.403095,
		0.466406,
		0.529046,
		0.590052,
		0.648554,
		0.70378,
		0.755064,
		0.801855,
		0.843722,
		0.880367,
		0.911631,
		0.937506,
		0.958141,
		0.973851,
		0.985111,
		0.992549,
		0.996925,
		0.999076,
		0.999852,
		0.999996,
		1,
		0.999923,
		0.999367,
		0.997637,
		0.993899,
		0.987301,
		0.977061,
		0.962515,
		0.943146,
		0.918602,
		0.88869,
		0.853382,
		0.812798,
		0.767202,
		0.716994,
		0.662695,
		0.604944,
		0.544486,
		0.482168,
		0.418929,
		0.355799,
		0.29389,
		0.234393,
		0.178574,
		0.127771,
		0.083389,
		0.046901,
		0.019841,
		0.003803,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.001737,
		0.014982,
		0.039835,
		0.074644,
		0.117821,
		0.167843,
		0.223252,
		0.282661,
		0.344756,
		0.408303,
		0.472143,
		0.535208,
		0.596519,
		0.655192,
		0.710447,
		0.761615,
		0.808146,
		0.849616,
		0.885739,
		0.916375,
		0.941544,
		0.961427,
		0.976376,
		0.986911,
		0.993706,
		0.997561,
		0.999346,
		0.99992,
		1,
		0.999996,
		0.999847,
		0.999046,
		0.996826,
		0.992316,
		0.984657,
		0.973075,
		0.956935,
		0.935757,
		0.909232,
		0.877219,
		0.839741,
		0.796974,
		0.749239,
		0.696992,
		0.640815,
		0.581406,
		0.51957,
		0.456215,
		0.392345,
		0.32905,
		0.267506,
		0.208972,
		0.154777,
		0.106327,
		0.065096,
		0.032625,
		0.010517,
		0.000439,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.001767,
		0.015236,
		0.040494,
		0.075848,
		0.119668,
		0.170393,
		0.22653,
		0.28666,
		0.349439,
		0.413603,
		0.477974,
		0.541461,
		0.603067,
		0.661898,
		0.717165,
		0.768196,
		0.814442,
		0.855488,
		0.891062,
		0.921045,
		0.945484,
		0.964597,
		0.978775,
		0.988586,
		0.994749,
		0.998106,
		0.999557,
		0.999962,
		1,
		0.999986,
		0.999726,
		0.998613,
		0.995817,
		0.990437,
		0.981612,
		0.968585,
		0.950749,
		0.927669,
		0.899083,
		0.864902,
		0.825205,
		0.780225,
		0.730344,
		0.676077,
		0.618068,
		0.557076,
		0.493969,
		0.429721,
		0.365398,
		0.302157,
		0.241242,
		0.183976,
		0.131761,
		0.086071,
		0.048451,
		0.020513,
		0.003934,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.001798,
		0.015497,
		0.04117,
		0.077081,
		0.121558,
		0.173,
		0.229879,
		0.290741,
		0.354211,
		0.418999,
		0.483901,
		0.547804,
		0.609698,
		0.668673,
		0.723934,
		0.774805,
		0.820741,
		0.861335,
		0.896331,
		0.925634,
		0.949321,
		0.967646,
		0.981047,
		0.990135,
		0.995681,
		0.998566,
		0.999716,
		0.999985,
		1,
		0.999961,
		0.999542,
		0.998044,
		0.99458,
		0.988228,
		0.978128,
		0.963547,
		0.943915,
		0.918841,
		0.888115,
		0.851704,
		0.809744,
		0.762531,
		0.710506,
		0.654248,
		0.594464,
		0.531979,
		0.467727,
		0.402746,
		0.338171,
		0.275228,
		0.21523,
		0.159571,
		0.109724,
		0.067236,
		0.033726,
		0.01088,
		0.000454,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.00183,
		0.015765,
		0.041863,
		0.078344,
		0.123492,
		0.175666,
		0.2333,
		0.294906,
		0.359076,
		0.424491,
		0.489923,
		0.55424,
		0.61641,
		0.675515,
		0.73075,
		0.781437,
		0.827036,
		0.867151,
		0.901541,
		0.930138,
		0.95305,
		0.970571,
		0.983187,
		0.991558,
		0.996504,
		0.998946,
		0.999831,
		0.999996,
		0.999999,
		0.999911,
		0.999278,
		0.997312,
		0.993081,
		0.985648,
		0.974162,
		0.95792,
		0.93639,
		0.909232,
		0.87629,
		0.837591,
		0.793332,
		0.743871,
		0.689715,
		0.631507,
		0.570021,
		0.506149,
		0.440894,
		0.375363,
		0.310761,
		0.248386,
		0.189623,
		0.135939,
		0.088884,
		0.050078,
		0.02122,
		0.004073,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.001863,
		0.016039,
		0.042574,
		0.079638,
		0.125472,
		0.178394,
		0.236797,
		0.299157,
		0.364035,
		0.430082,
		0.496045,
		0.560768,
		0.623205,
		0.682423,
		0.737611,
		0.788092,
		0.833326,
		0.872931,
		0.906686,
		0.934549,
		0.956664,
		0.973368,
		0.985193,
		0.992856,
		0.997222,
		0.999253,
		0.999908,
		0.999999,
		0.999996,
		0.999825,
		0.99891,
		0.996388,
		0.991285,
		0.982658,
		0.969673,
		0.951659,
		0.928131,
		0.898801,
		0.863572,
		0.822533,
		0.775945,
		0.724231,
		0.667967,
		0.607863,
		0.544762,
		0.479627,
		0.413531,
		0.347655,
		0.283276,
		0.221767,
		0.164588,
		0.113285,
		0.069482,
		0.034883,
		0.011263,
		0.000471,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.001896,
		0.016321,
		0.043303,
		0.080964,
		0.1275,
		0.181184,
		0.24037,
		0.303496,
		0.369091,
		0.435774,
		0.502265,
		0.567389,
		0.630081,
		0.689396,
		0.744517,
		0.794763,
		0.839605,
		0.87867,
		0.91176,
		0.938862,
		0.960158,
		0.976031,
		0.987065,
		0.994029,
		0.997839,
		0.999493,
		0.999956,
		1,
		0.999984,
		0.999686,
		0.998415,
		0.995238,
		0.989154,
		0.979216,
		0.964615,
		0.944719,
		0.919093,
		0.887506,
		0.849923,
		0.806498,
		0.757559,
		0.703599,
		0.645261,
		0.583329,
		0.518718,
		0.452462,
		0.385711,
		0.319719,
		0.255841,
		0.195527,
		0.140317,
		0.091835,
		0.051789,
		0.021964,
		0.004219,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.001931,
		0.01661,
		0.044051,
		0.082324,
		0.129577,
		0.184039,
		0.244022,
		0.307926,
		0.374245,
		0.441568,
		0.508587,
		0.574105,
		0.63704,
		0.696434,
		0.751464,
		0.80145,
		0.845868,
		0.884362,
		0.916757,
		0.943071,
		0.963527,
		0.978558,
		0.9888,
		0.995079,
		0.998361,
		0.999675,
		0.999983,
		1,
		0.999955,
		0.999475,
		0.997765,
		0.993829,
		0.986646,
		0.975277,
		0.958943,
		0.937054,
		0.909232,
		0.875306,
		0.835308,
		0.789457,
		0.738155,
		0.681965,
		0.621603,
		0.557927,
		0.491928,
		0.424716,
		0.357518,
		0.291667,
		0.228598,
		0.169842,
		0.11702,
		0.071843,
		0.036101,
		0.011667,
		0.000488,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.001966,
		0.016907,
		0.044818,
		0.083718,
		0.131705,
		0.186961,
		0.247755,
		0.31245,
		0.379501,
		0.447466,
		0.515011,
		0.580916,
		0.64408,
		0.703534,
		0.758449,
		0.808146,
		0.85211,
		0.890001,
		0.92167,
		0.94717,
		0.966766,
		0.980944,
		0.990397,
		0.996009,
		0.998793,
		0.999805,
		0.999995,
		0.999999,
		0.999898,
		0.999172,
		0.996928,
		0.992122,
		0.983719,
		0.970795,
		0.952607,
		0.928617,
		0.898503,
		0.862161,
		0.819691,
		0.771385,
		0.717717,
		0.659326,
		0.597004,
		0.531686,
		0.464441,
		0.39646,
		0.32905,
		0.263625,
		0.201706,
		0.144906,
		0.094935,
		0.053588,
		0.022747,
		0.004374,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.002002,
		0.017212,
		0.045605,
		0.085147,
		0.133884,
		0.189951,
		0.251572,
		0.317068,
		0.384859,
		0.453471,
		0.521539,
		0.587822,
		0.651201,
		0.710695,
		0.76547,
		0.814849,
		0.858326,
		0.895582,
		0.926493,
		0.951151,
		0.96987,
		0.983187,
		0.991857,
		0.996822,
		0.999142,
		0.999894,
		0.999999,
		0.999995,
		0.999798,
		0.99875,
		0.995871,
		0.990077,
		0.980326,
		0.965721,
		0.94556,
		0.919359,
		0.886861,
		0.848031,
		0.803041,
		0.752258,
		0.696234,
		0.635684,
		0.571482,
		0.504643,
		0.436317,
		0.367779,
		0.300419,
		0.23574,
		0.175346,
		0.120941,
		0.074325,
		0.037384,
		0.012092,
		0.000506,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.00204,
		0.017525,
		0.046414,
		0.086613,
		0.136118,
		0.193012,
		0.255475,
		0.321785,
		0.390324,
		0.459584,
		0.528173,
		0.594824,
		0.658402,
		0.717915,
		0.772523,
		0.821553,
		0.864511,
		0.901097,
		0.931219,
		0.95501,
		0.972832,
		0.985283,
		0.993181,
		0.997523,
		0.999417,
		0.99995,
		1,
		0.999981,
		0.999638,
		0.998182,
		0.994557,
		0.987648,
		0.97642,
		0.960006,
		0.937751,
		0.909232,
		0.874262,
		0.832877,
		0.785326,
		0.732056,
		0.673697,
		0.611048,
		0.545064,
		0.476846,
		0.407628,
		0.338772,
		0.271757,
		0.208175,
		0.149721,
		0.098193,
		0.055483,
		0.023574,
		0.004537,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.002078,
		0.017847,
		0.047243,
		0.088117,
		0.138407,
		0.196146,
		0.259467,
		0.326602,
		0.395897,
		0.465807,
		0.534913,
		0.601923,
		0.665684,
		0.725192,
		0.779606,
		0.828254,
		0.870658,
		0.90654,
		0.935842,
		0.958739,
		0.97565,
		0.987232,
		0.994368,
		0.998117,
		0.999625,
		0.99998,
		1,
		0.999948,
		0.999396,
		0.997435,
		0.992945,
		0.984791,
		0.971951,
		0.953597,
		0.929128,
		0.898188,
		0.860662,
		0.816662,
		0.76652,
		0.710763,
		0.650105,
		0.585432,
		0.517785,
		0.448352,
		0.378457,
		0.309553,
		0.243212,
		0.181117,
		0.125061,
		0.076937,
		0.038736,
		0.012541,
		0.000525,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.002117,
		0.018178,
		0.048096,
		0.08966,
		0.140754,
		0.199356,
		0.263549,
		0.331523,
		0.40158,
		0.472143,
		0.541761,
		0.609118,
		0.673044,
		0.732524,
		0.786713,
		0.834948,
		0.876762,
		0.911904,
		0.940353,
		0.962333,
		0.978317,
		0.989031,
		0.995423,
		0.99861,
		0.999775,
		0.999995,
		0.999999,
		0.999882,
		0.999046,
		0.996474,
		0.990993,
		0.981456,
		0.966867,
		0.94644,
		0.919639,
		0.886177,
		0.846017,
		0.799352,
		0.746597,
		0.688367,
		0.625463,
		0.55886,
		0.48969,
		0.419232,
		0.348906,
		0.280256,
		0.214952,
		0.154777,
		0.101621,
		0.057479,
		0.024446,
		0.004709,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.002158,
		0.018518,
		0.048971,
		0.091244,
		0.14316,
		0.202643,
		0.267725,
		0.336549,
		0.407377,
		0.478594,
		0.548718,
		0.61641,
		0.680482,
		0.739908,
		0.793842,
		0.841627,
		0.882814,
		0.917181,
		0.944746,
		0.965784,
		0.980831,
		0.990679,
		0.996348,
		0.999011,
		0.999877,
		0.999999,
		0.999994,
		0.999767,
		0.99856,
		0.99526,
		0.988652,
		0.977591,
		0.961112,
		0.938482,
		0.909232,
		0.873152,
		0.830285,
		0.780913,
		0.725537,
		0.664862,
		0.599783,
		0.531365,
		0.460836,
		0.389573,
		0.319089,
		0.251033,
		0.187172,
		0.129392,
		0.079688,
		0.040162,
		0.013016,
		0.000546,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.0022,
		0.018867,
		0.049871,
		0.09287,
		0.145627,
		0.20601,
		0.271997,
		0.341684,
		0.413289,
		0.48516,
		0.555785,
		0.623799,
		0.687996,
		0.747342,
		0.800988,
		0.848287,
		0.88881,
		0.922365,
		0.949014,
		0.969087,
		0.983187,
		0.992177,
		0.997147,
		0.999326,
		0.999942,
		1,
		0.999978,
		0.999581,
		0.997904,
		0.99375,
		0.985874,
		0.973143,
		0.95463,
		0.929666,
		0.897854,
		0.859064,
		0.813428,
		0.761317,
		0.703324,
		0.64025,
		0.573084,
		0.502989,
		0.431293,
		0.359472,
		0.289144,
		0.222058,
		0.160089,
		0.105229,
		0.059585,
		0.025367,
		0.004891,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.002243,
		0.019227,
		0.050795,
		0.094539,
		0.148158,
		0.20946,
		0.276369,
		0.34693,
		0.419319,
		0.491846,
		0.562964,
		0.631286,
		0.695586,
		0.754822,
		0.808146,
		0.854921,
		0.89474,
		0.927446,
		0.953149,
		0.972236,
		0.985382,
		0.993526,
		0.997826,
		0.999565,
		0.999977,
		1,
		0.999939,
		0.9993,
		0.997042,
		0.991898,
		0.982605,
		0.968053,
		0.947363,
		0.919935,
		0.885451,
		0.843868,
		0.795408,
		0.740538,
		0.679948,
		0.614537,
		0.545395,
		0.473788,
		0.401146,
		0.32905,
		0.259225,
		0.193529,
		0.133949,
		0.082589,
		0.041669,
		0.013517,
		0.000567,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.002287,
		0.019597,
		0.051745,
		0.096253,
		0.150754,
		0.212995,
		0.280842,
		0.352291,
		0.425471,
		0.498652,
		0.570256,
		0.638868,
		0.703249,
		0.762345,
		0.815311,
		0.861523,
		0.900598,
		0.932418,
		0.957143,
		0.975226,
		0.987414,
		0.994726,
		0.998392,
		0.999739,
		0.999994,
		0.999999,
		0.999863,
		0.998895,
		0.995933,
		0.989654,
		0.97879,
		0.962262,
		0.939252,
		0.909232,
		0.87197,
		0.827516,
		0.77619,
		0.718555,
		0.655405,
		0.587741,
		0.516759,
		0.44383,
		0.370494,
		0.298442,
		0.229511,
		0.165674,
		0.109032,
		0.061808,
		0.026342,
		0.005084,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.002333,
		0.019978,
		0.052722,
		0.098015,
		0.153419,
		0.216617,
		0.285421,
		0.357769,
		0.431746,
		0.50558,
		0.577661,
		0.646547,
		0.710982,
		0.769907,
		0.822478,
		0.868085,
		0.906376,
		0.937272,
		0.960989,
		0.978051,
		0.989282,
		0.995782,
		0.998853,
		0.999857,
		0.999999,
		0.999993,
		0.999729,
		0.998331,
		0.994532,
		0.986964,
		0.974369,
		0.95571,
		0.930235,
		0.897499,
		0.85736,
		0.809968,
		0.755743,
		0.695352,
		0.629697,
		0.559887,
		0.487226,
		0.413199,
		0.339457,
		0.267809,
		0.200209,
		0.138748,
		0.08565,
		0.043262,
		0.014049,
		0.00059,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.00238,
		0.02037,
		0.053727,
		0.099824,
		0.156153,
		0.220331,
		0.290107,
		0.363368,
		0.438147,
		0.512632,
		0.58518,
		0.654322,
		0.718785,
		0.777504,
		0.82964,
		0.874601,
		0.912065,
		0.941999,
		0.964681,
		0.980706,
		0.990983,
		0.996697,
		0.999216,
		0.999932,
		1,
		0.999974,
		0.999512,
		0.997571,
		0.992789,
		0.983772,
		0.969281,
		0.948332,
		0.920249,
		0.884676,
		0.84157,
		0.791183,
		0.73404,
		0.670922,
		0.602837,
		0.531011,
		0.456863,
		0.381994,
		0.308175,
		0.237335,
		0.171552,
		0.113042,
		0.064157,
		0.027373,
		0.005289,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.002428,
		0.020774,
		0.054761,
		0.101684,
		0.15896,
		0.224139,
		0.294906,
		0.369091,
		0.444677,
		0.519811,
		0.592813,
		0.662191,
		0.726654,
		0.785132,
		0.836792,
		0.881063,
		0.917657,
		0.946591,
		0.968209,
		0.983187,
		0.99252,
		0.997478,
		0.999493,
		0.999973,
		1,
		0.999929,
		0.999185,
		0.996571,
		0.99065,
		0.980016,
		0.963459,
		0.940063,
		0.909232,
		0.870708,
		0.824552,
		0.771123,
		0.711062,
		0.645261,
		0.574848,
		0.501165,
		0.425753,
		0.350337,
		0.276811,
		0.207232,
		0.143807,
		0.088884,
		0.044948,
		0.014613,
		0.000615,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.002478,
		0.021189,
		0.055825,
		0.103595,
		0.161842,
		0.228043,
		0.299819,
		0.37494,
		0.451339,
		0.527118,
		0.600563,
		0.670154,
		0.734587,
		0.792785,
		0.843926,
		0.887461,
		0.923141,
		0.951039,
		0.971569,
		0.98549,
		0.993892,
		0.99813,
		0.999695,
		0.999993,
		0.999999,
		0.999839,
		0.998713,
		0.995285,
		0.988059,
		0.975631,
		0.956839,
		0.930835,
		0.897121,
		0.855538,
		0.806258,
		0.749757,
		0.686792,
		0.618377,
		0.545761,
		0.470412,
		0.393997,
		0.318369,
		0.245554,
		0.177742,
		0.117276,
		0.066642,
		0.028466,
		0.005506,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.00253,
		0.021618,
		0.05692,
		0.105561,
		0.164802,
		0.232048,
		0.30485,
		0.380919,
		0.458135,
		0.534555,
		0.608428,
		0.67821,
		0.742579,
		0.800458,
		0.851035,
		0.893788,
		0.928511,
		0.955335,
		0.974752,
		0.987613,
		0.995103,
		0.998662,
		0.999833,
		0.999999,
		0.999992,
		0.999682,
		0.998056,
		0.993658,
		0.984953,
		0.970552,
		0.949349,
		0.920582,
		0.883851,
		0.839109,
		0.786646,
		0.727057,
		0.661224,
		0.590287,
		0.515624,
		0.438833,
		0.361715,
		0.286256,
		0.214622,
		0.149143,
		0.092302,
		0.046734,
		0.015211,
		0.000641,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.002583,
		0.022059,
		0.058047,
		0.107582,
		0.167843,
		0.236155,
		0.310002,
		0.387032,
		0.465069,
		0.542124,
		0.61641,
		0.686356,
		0.750628,
		0.808146,
		0.858112,
		0.900035,
		0.933754,
		0.959469,
		0.977753,
		0.989554,
		0.996156,
		0.999083,
		0.99992,
		1,
		0.99997,
		0.999429,
		0.997169,
		0.991635,
		0.981267,
		0.964706,
		0.940917,
		0.909232,
		0.869359,
		0.82137,
		0.765677,
		0.703003,
		0.63436,
		0.561021,
		0.484498,
		0.406528,
		0.32905,
		0.254193,
		0.184267,
		0.121749,
		0.069274,
		0.029626,
		0.005737,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.002638,
		0.022514,
		0.059208,
		0.109662,
		0.170967,
		0.24037,
		0.315281,
		0.393283,
		0.472143,
		0.549827,
		0.624508,
		0.694592,
		0.758729,
		0.815842,
		0.865149,
		0.906191,
		0.938862,
		0.963433,
		0.980567,
		0.991312,
		0.997056,
		0.999406,
		0.999969,
		1,
		0.999917,
		0.999046,
		0.996003,
		0.989154,
		0.976929,
		0.95802,
		0.931471,
		0.896717,
		0.853585,
		0.80227,
		0.743316,
		0.67758,
		0.606211,
		0.53062,
		0.452462,
		0.373618,
		0.296172,
		0.222405,
		0.154777,
		0.09592,
		0.048628,
		0.015846,
		0.000668,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.002694,
		0.022984,
		0.060404,
		0.111802,
		0.174178,
		0.244695,
		0.320688,
		0.399673,
		0.47936,
		0.557664,
		0.632722,
		0.702914,
		0.766878,
		0.823539,
		0.872136,
		0.912247,
		0.943825,
		0.967218,
		0.983187,
		0.992887,
		0.997811,
		0.999641,
		0.999991,
		0.999999,
		0.999811,
		0.998492,
		0.994502,
		0.986146,
		0.971868,
		0.950418,
		0.920935,
		0.882967,
		0.836467,
		0.781763,
		0.719536,
		0.650784,
		0.5768,
		0.499142,
		0.419614,
		0.340247,
		0.26328,
		0.191151,
		0.12648,
		0.072063,
		0.030859,
		0.005983,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.002753,
		0.023468,
		0.061637,
		0.114005,
		0.177478,
		0.249134,
		0.326228,
		0.406208,
		0.486723,
		0.565638,
		0.641052,
		0.71132,
		0.77507,
		0.83123,
		0.879064,
		0.918192,
		0.948631,
		0.970816,
		0.98561,
		0.994282,
		0.998429,
		0.999803,
		0.999999,
		0.999991,
		0.999626,
		0.997721,
		0.992605,
		0.982541,
		0.966004,
		0.94182,
		0.909232,
		0.867913,
		0.817948,
		0.759807,
		0.694316,
		0.622621,
		0.546167,
		0.466665,
		0.386075,
		0.306588,
		0.230606,
		0.160731,
		0.099754,
		0.050639,
		0.016522,
		0.000697,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.002813,
		0.023968,
		0.062907,
		0.116272,
		0.180871,
		0.25369,
		0.331906,
		0.412891,
		0.494234,
		0.573749,
		0.649498,
		0.719808,
		0.783299,
		0.838908,
		0.885923,
		0.924016,
		0.953272,
		0.974219,
		0.987832,
		0.995499,
		0.998921,
		0.999905,
		1,
		0.999964,
		0.999327,
		0.996681,
		0.990244,
		0.978262,
		0.959257,
		0.932145,
		0.896287,
		0.851487,
		0.797974,
		0.736367,
		0.667644,
		0.593109,
		0.514364,
		0.433283,
		0.35199,
		0.272844,
		0.198419,
		0.13149,
		0.075024,
		0.032169,
		0.006244,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.002875,
		0.024483,
		0.064217,
		0.118608,
		0.18436,
		0.258369,
		0.337724,
		0.419726,
		0.501896,
		0.582,
		0.658057,
		0.728373,
		0.791559,
		0.846562,
		0.892703,
		0.929706,
		0.957735,
		0.97742,
		0.989852,
		0.996543,
		0.999299,
		0.999963,
		1,
		0.999901,
		0.998875,
		0.995313,
		0.987347,
		0.973229,
		0.951543,
		0.921311,
		0.882021,
		0.833622,
		0.776495,
		0.711414,
		0.63952,
		0.56228,
		0.481465,
		0.399117,
		0.317536,
		0.239256,
		0.16703,
		0.103819,
		0.052777,
		0.017243,
		0.000729,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.00294,
		0.025016,
		0.065569,
		0.121013,
		0.18795,
		0.263174,
		0.343689,
		0.426716,
		0.509713,
		0.59039,
		0.66673,
		0.737013,
		0.799844,
		0.854185,
		0.899393,
		0.935251,
		0.962012,
		0.980411,
		0.991668,
		0.997422,
		0.999575,
		0.99999,
		0.999999,
		0.999776,
		0.998222,
		0.993551,
		0.983837,
		0.967356,
		0.942774,
		0.909232,
		0.866359,
		0.814257,
		0.753466,
		0.684929,
		0.609954,
		0.530184,
		0.447562,
		0.364313,
		0.282919,
		0.2061,
		0.136799,
		0.07817,
		0.033565,
		0.006524,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.003007,
		0.025566,
		0.066963,
		0.123492,
		0.191643,
		0.268109,
		0.349803,
		0.433865,
		0.517687,
		0.598922,
		0.675515,
		0.745722,
		0.808146,
		0.861767,
		0.90598,
		0.94064,
		0.96609,
		0.983187,
		0.993282,
		0.998145,
		0.999766,
		0.999999,
		0.999989,
		0.999556,
		0.997312,
		0.991323,
		0.97963,
		0.960553,
		0.932861,
		0.895825,
		0.849227,
		0.793332,
		0.728851,
		0.656903,
		0.578973,
		0.496887,
		0.412775,
		0.32905,
		0.248386,
		0.1737,
		0.108136,
		0.055053,
		0.018011,
		0.000762,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.003075,
		0.026134,
		0.068402,
		0.126047,
		0.195443,
		0.273179,
		0.356071,
		0.441178,
		0.52582,
		0.607595,
		0.684409,
		0.754496,
		0.816458,
		0.869296,
		0.912454,
		0.94586,
		0.969962,
		0.985742,
		0.994696,
		0.998722,
		0.999887,
		1,
		0.999957,
		0.999202,
		0.996084,
		0.988552,
		0.974637,
		0.952728,
		0.921711,
		0.881004,
		0.830552,
		0.770794,
		0.702622,
		0.627338,
		0.546619,
		0.462481,
		0.37725,
		0.293538,
		0.214224,
		0.142431,
		0.081516,
		0.035053,
		0.006822,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.003147,
		0.026722,
		0.069887,
		0.128681,
		0.199356,
		0.278389,
		0.362499,
		0.448658,
		0.534114,
		0.61641,
		0.69341,
		0.763329,
		0.82477,
		0.876762,
		0.9188,
		0.950898,
		0.973615,
		0.988073,
		0.995914,
		0.999166,
		0.999956,
		1,
		0.999882,
		0.998665,
		0.994468,
		0.985151,
		0.968764,
		0.943785,
		0.909232,
		0.864686,
		0.810266,
		0.746597,
		0.67476,
		0.596255,
		0.512956,
		0.427082,
		0.341166,
		0.258031,
		0.18077,
		0.112726,
		0.057479,
		0.018832,
		0.000798,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.003221,
		0.027329,
		0.071421,
		0.131398,
		0.203384,
		0.283743,
		0.369091,
		0.45631,
		0.542574,
		0.625367,
		0.702516,
		0.772216,
		0.833075,
		0.884152,
		0.925006,
		0.955742,
		0.977043,
		0.990178,
		0.996943,
		0.999493,
		0.999988,
		0.999998,
		0.999732,
		0.997888,
		0.992386,
		0.98103,
		0.961912,
		0.933622,
		0.895329,
		0.846787,
		0.788304,
		0.720701,
		0.645261,
		0.563688,
		0.478069,
		0.390835,
		0.304739,
		0.222825,
		0.148414,
		0.085081,
		0.036642,
		0.007142,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.003297,
		0.027957,
		0.073006,
		0.1342,
		0.207533,
		0.289247,
		0.375851,
		0.464137,
		0.5512,
		0.634466,
		0.711723,
		0.781149,
		0.84136,
		0.891454,
		0.931058,
		0.960379,
		0.980236,
		0.992055,
		0.997793,
		0.99972,
		0.999998,
		0.999987,
		0.99947,
		0.996806,
		0.989754,
		0.976091,
		0.953979,
		0.922139,
		0.879909,
		0.827229,
		0.76461,
		0.693079,
		0.614133,
		0.529695,
		0.442072,
		0.353922,
		0.268228,
		0.188271,
		0.117611,
		0.060068,
		0.01971,
		0.000836,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.003376,
		0.028608,
		0.074644,
		0.137092,
		0.211807,
		0.294906,
		0.382786,
		0.472143,
		0.559995,
		0.643707,
		0.721027,
		0.79012,
		0.849616,
		0.898654,
		0.936941,
		0.964797,
		0.983187,
		0.993706,
		0.998474,
		0.999864,
		1,
		0.999949,
		0.999046,
		0.995346,
		0.986479,
		0.97023,
		0.944857,
		0.909232,
		0.862877,
		0.805937,
		0.739133,
		0.663715,
		0.581406,
		0.494356,
		0.405109,
		0.316562,
		0.231941,
		0.154777,
		0.088884,
		0.038342,
		0.007484,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.003458,
		0.029281,
		0.076337,
		0.140077,
		0.21621,
		0.300724,
		0.3899,
		0.480333,
		0.568961,
		0.653088,
		0.730424,
		0.799123,
		0.85783,
		0.905738,
		0.942641,
		0.968983,
		0.98589,
		0.995135,
		0.999001,
		0.999946,
		1,
		0.999858,
		0.998403,
		0.993424,
		0.982461,
		0.963338,
		0.934434,
		0.894796,
		0.844144,
		0.782841,
		0.711836,
		0.632611,
		0.547127,
		0.457781,
		0.367361,
		0.279019,
		0.19624,
		0.122817,
		0.062836,
		0.020651,
		0.000878,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.003544,
		0.029978,
		0.078089,
		0.14316,
		0.22075,
		0.306709,
		0.397198,
		0.488711,
		0.5781,
		0.662608,
		0.739908,
		0.808146,
		0.86599,
		0.912691,
		0.948142,
		0.972926,
		0.98834,
		0.996348,
		0.999391,
		0.999985,
		0.999998,
		0.999678,
		0.997473,
		0.990947,
		0.977591,
		0.955299,
		0.922598,
		0.878726,
		0.823621,
		0.757879,
		0.68269,
		0.599783,
		0.511372,
		0.42011,
		0.32905,
		0.24161,
		0.161551,
		0.092945,
		0.040162,
		0.007852,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.003632,
		0.0307,
		0.079901,
		0.146343,
		0.225429,
		0.312866,
		0.404687,
		0.497281,
		0.587413,
		0.672265,
		0.749475,
		0.817181,
		0.874082,
		0.919497,
		0.953429,
		0.976614,
		0.990536,
		0.997353,
		0.999662,
		0.999998,
		0.999984,
		0.999361,
		0.996177,
		0.987816,
		0.971756,
		0.945995,
		0.909232,
		0.860918,
		0.801227,
		0.731,
		0.651685,
		0.565272,
		0.474244,
		0.381525,
		0.290446,
		0.204713,
		0.128373,
		0.065798,
		0.021661,
		0.000922,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.003724,
		0.031449,
		0.081776,
		0.149633,
		0.230255,
		0.3192,
		0.412372,
		0.506046,
		0.596902,
		0.682058,
		0.759116,
		0.826216,
		0.882091,
		0.926139,
		0.958486,
		0.980037,
		0.992476,
		0.998163,
		0.999836,
		1,
		0.999938,
		0.99885,
		0.994427,
		0.983919,
		0.964833,
		0.935301,
		0.89422,
		0.841272,
		0.776885,
		0.702164,
		0.618827,
		0.529145,
		0.43588,
		0.342249,
		0.251877,
		0.168771,
		0.097288,
		0.042115,
		0.008248,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.003819,
		0.032226,
		0.083718,
		0.153034,
		0.235233,
		0.325719,
		0.420258,
		0.515011,
		0.606567,
		0.691982,
		0.768825,
		0.835238,
		0.890001,
		0.9326,
		0.963299,
		0.983187,
		0.994162,
		0.998793,
		0.999935,
		1,
		0.999827,
		0.998073,
		0.992122,
		0.979138,
		0.956695,
		0.92309,
		0.877444,
		0.819691,
		0.750529,
		0.671346,
		0.584147,
		0.491497,
		0.39646,
		0.302559,
		0.213733,
		0.13431,
		0.068974,
		0.022747,
		0.00097,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.003918,
		0.033031,
		0.085729,
		0.156549,
		0.24037,
		0.332429,
		0.428351,
		0.52418,
		0.61641,
		0.702034,
		0.778592,
		0.844236,
		0.897796,
		0.938862,
		0.967852,
		0.986055,
		0.995601,
		0.999261,
		0.999982,
		0.999998,
		0.999608,
		0.99695,
		0.989154,
		0.973345,
		0.947206,
		0.909232,
		0.858788,
		0.796084,
		0.722107,
		0.638541,
		0.547701,
		0.452462,
		0.356209,
		0.26279,
		0.176478,
		0.101941,
		0.044213,
		0.008674,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.00402,
		0.033867,
		0.087813,
		0.160186,
		0.245671,
		0.339336,
		0.436658,
		0.533556,
		0.62643,
		0.71221,
		0.788409,
		0.853193,
		0.905457,
		0.944907,
		0.97213,
		0.988638,
		0.996799,
		0.999588,
		0.999997,
		0.999981,
		0.999223,
		0.995385,
		0.985399,
		0.966404,
		0.936229,
		0.893596,
		0.838141,
		0.770369,
		0.691577,
		0.603767,
		0.509577,
		0.412216,
		0.315408,
		0.223347,
		0.140662,
		0.072383,
		0.023917,
		0.001021,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.004127,
		0.034735,
		0.089974,
		0.163949,
		0.251144,
		0.346449,
		0.445185,
		0.543144,
		0.636628,
		0.722505,
		0.798264,
		0.862095,
		0.912966,
		0.950715,
		0.97612,
		0.990931,
		0.997769,
		0.999799,
		1,
		0.999924,
		0.9986,
		0.99327,
		0.980727,
		0.958172,
		0.92362,
		0.87605,
		0.815394,
		0.742476,
		0.65892,
		0.567067,
		0.469901,
		0.370984,
		0.2744,
		0.184715,
		0.106933,
		0.046472,
		0.009134,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.004238,
		0.035637,
		0.092214,
		0.167843,
		0.256796,
		0.353774,
		0.453937,
		0.552945,
		0.647002,
		0.732912,
		0.808146,
		0.870925,
		0.920303,
		0.956268,
		0.97981,
		0.992935,
		0.998528,
		0.99992,
		1,
		0.999788,
		0.997655,
		0.990485,
		0.974996,
		0.948498,
		0.909232,
		0.856465,
		0.790449,
		0.712348,
		0.624138,
		0.528519,
		0.428844,
		0.32905,
		0.233606,
		0.14747,
		0.076051,
		0.025179,
		0.001077,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.004354,
		0.036574,
		0.094539,
		0.171875,
		0.262634,
		0.361318,
		0.462921,
		0.562964,
		0.657551,
		0.743424,
		0.818041,
		0.879663,
		0.927446,
		0.961546,
		0.983187,
		0.994653,
		0.999094,
		0.999977,
		0.999997,
		0.999519,
		0.996286,
		0.986896,
		0.968053,
		0.937225,
		0.892918,
		0.834714,
		0.763216,
		0.679948,
		0.587261,
		0.488241,
		0.38663,
		0.286766,
		0.193529,
		0.112297,
		0.048908,
		0.009632,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.004475,
		0.037549,
		0.096952,
		0.176052,
		0.268666,
		0.369091,
		0.472143,
		0.573204,
		0.668272,
		0.754033,
		0.827935,
		0.888291,
		0.934374,
		0.966531,
		0.986243,
		0.996092,
		0.999493,
		0.999997,
		0.999976,
		0.999046,
		0.994377,
		0.982358,
		0.959736,
		0.924192,
		0.874529,
		0.81068,
		0.733619,
		0.645261,
		0.548355,
		0.446396,
		0.343545,
		0.244566,
		0.154777,
		0.080003,
		0.026543,
		0.001138,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.0046,
		0.038563,
		0.099458,
		0.18038,
		0.2749,
		0.377099,
		0.48161,
		0.583667,
		0.679164,
		0.764729,
		0.837812,
		0.896786,
		0.941064,
		0.971204,
		0.988971,
		0.997265,
		0.999752,
		1,
		0.999906,
		0.99828,
		0.991798,
		0.976711,
		0.949876,
		0.909232,
		0.85392,
		0.78425,
		0.701601,
		0.608301,
		0.507526,
		0.403207,
		0.299949,
		0.202976,
		0.118071,
		0.051539,
		0.010171,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.004731,
		0.039618,
		0.102062,
		0.184867,
		0.281345,
		0.385352,
		0.491327,
		0.594354,
		0.690221,
		0.775502,
		0.847654,
		0.905127,
		0.947493,
		0.975548,
		0.991369,
		0.998186,
		0.9999,
		1,
		0.999738,
		0.997117,
		0.988401,
		0.969784,
		0.938296,
		0.892179,
		0.830949,
		0.755329,
		0.667127,
		0.569117,
		0.464929,
		0.358959,
		0.25629,
		0.162631,
		0.084269,
		0.028021,
		0.001204,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.004868,
		0.040718,
		0.104768,
		0.18952,
		0.288011,
		0.393858,
		0.501302,
		0.605268,
		0.701439,
		0.786339,
		0.85744,
		0.913289,
		0.953636,
		0.979547,
		0.993437,
		0.998879,
		0.999972,
		0.999996,
		0.999403,
		0.995431,
		0.984024,
		0.961395,
		0.924811,
		0.872864,
		0.805484,
		0.723836,
		0.630193,
		0.527802,
		0.420779,
		0.314019,
		0.213114,
		0.124297,
		0.054388,
		0.010757,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.00501,
		0.041863,
		0.107582,
		0.194346,
		0.294906,
		0.402626,
		0.511539,
		0.61641,
		0.712812,
		0.797225,
		0.867151,
		0.921246,
		0.959469,
		0.983187,
		0.99518,
		0.999369,
		0.999996,
		0.99997,
		0.998815,
		0.993081,
		0.978488,
		0.951351,
		0.909232,
		0.851122,
		0.777401,
		0.689715,
		0.590828,
		0.484498,
		0.375363,
		0.268846,
		0.171089,
		0.088884,
		0.029626,
		0.001275,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.005159,
		0.043058,
		0.110511,
		0.199356,
		0.30204,
		0.411665,
		0.522046,
		0.62778,
		0.724333,
		0.808146,
		0.876762,
		0.928972,
		0.964969,
		0.986456,
		0.996609,
		0.999689,
		1,
		0.999882,
		0.997863,
		0.989904,
		0.971601,
		0.939451,
		0.89137,
		0.826794,
		0.746597,
		0.652936,
		0.549108,
		0.439414,
		0.32905,
		0.22401,
		0.131022,
		0.057479,
		0.011395,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005315,
		0.044304,
		0.113559,
		0.204557,
		0.309424,
		0.420985,
		0.532828,
		0.639377,
		0.735993,
		0.819083,
		0.886248,
		0.936437,
		0.970112,
		0.989348,
		0.99774,
		0.999875,
		1,
		0.999671,
		0.996415,
		0.985719,
		0.963156,
		0.925483,
		0.871032,
		0.799732,
		0.712986,
		0.613507,
		0.505161,
		0.392831,
		0.282312,
		0.180211,
		0.093885,
		0.031373,
		0.001354,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.005478,
		0.045605,
		0.116734,
		0.209959,
		0.317068,
		0.430596,
		0.54389,
		0.651201,
		0.747781,
		0.830017,
		0.895582,
		0.943612,
		0.974876,
		0.991857,
		0.998595,
		0.999964,
		0.999995,
		0.999251,
		0.994317,
		0.980326,
		0.952932,
		0.909232,
		0.848031,
		0.769799,
		0.676513,
		0.571482,
		0.459182,
		0.345123,
		0.23574,
		0.138302,
		0.06084,
		0.012092,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005649,
		0.046964,
		0.120042,
		0.215573,
		0.324985,
		0.440506,
		0.555238,
		0.663248,
		0.759685,
		0.840925,
		0.904734,
		0.950468,
		0.97924,
		0.993987,
		0.999205,
		0.999995,
		0.999962,
		0.998511,
		0.99139,
		0.973507,
		0.940699,
		0.89048,
		0.822187,
		0.736881,
		0.63716,
		0.526971,
		0.411446,
		0.29677,
		0.190067,
		0.099316,
		0.033279,
		0.00144,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.005827,
		0.048385,
		0.123492,
		0.221409,
		0.333186,
		0.450728,
		0.566876,
		0.675515,
		0.771692,
		0.851782,
		0.913673,
		0.956972,
		0.983187,
		0.995745,
		0.999606,
		1,
		0.99985,
		0.997312,
		0.987433,
		0.965026,
		0.926215,
		0.869007,
		0.793332,
		0.700893,
		0.594955,
		0.480153,
		0.362328,
		0.248386,
		0.146196,
		0.064503,
		0.012854,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.006014,
		0.04987,
		0.127091,
		0.227479,
		0.341684,
		0.461269,
		0.578809,
		0.687995,
		0.783783,
		0.862561,
		0.922363,
		0.963094,
		0.9867,
		0.997146,
		0.999839,
		0.999998,
		0.99958,
		0.995487,
		0.982219,
		0.954629,
		0.909231,
		0.844598,
		0.761316,
		0.661781,
		0.549981,
		0.431293,
		0.312315,
		0.200736,
		0.105229,
		0.035363,
		0.001534,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.006211,
		0.051425,
		0.130845,
		0.233791,
		0.350485,
		0.472135,
		0.591031,
		0.700674,
		0.795928,
		0.873218,
		0.930757,
		0.968789,
		0.989757,
		0.998198,
		0.999936,
		0.999976,
		0.999028,
		0.992826,
		0.975486,
		0.942036,
		0.889481,
		0.817037,
		0.726004,
		0.619534,
		0.502393,
		0.380753,
		0.262037,
		0.154774,
		0.068505,
		0.013691,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.006416,
		0.053048,
		0.134756,
		0.240343,
		0.359581,
		0.483303,
		0.603506,
		0.713495,
		0.808054,
		0.883665,
		0.938756,
		0.973965,
		0.99229,
		0.998871,
		0.99988,
		0.999837,
		0.997988,
		0.989042,
		0.966903,
		0.926912,
		0.86666,
		0.786085,
		0.687268,
		0.574175,
		0.452411,
		0.329012,
		0.212284,
		0.111669,
		0.037644,
		0.001637,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.00663,
		0.054735,
		0.138805,
		0.247099,
		0.368915,
		0.494688,
		0.616116,
		0.726308,
		0.819979,
		0.893693,
		0.94614,
		0.97841,
		0.994116,
		0.999017,
		0.999523,
		0.99933,
		0.996096,
		0.983695,
		0.956001,
		0.908799,
		0.840368,
		0.751442,
		0.644953,
		0.525747,
		0.400334,
		0.276679,
		0.164044,
		0.072856,
		0.014606,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.006851,
		0.056465,
		0.142941,
		0.253963,
		0.378334,
		0.50608,
		0.628589,
		0.73878,
		0.831316,
		0.902878,
		0.952465,
		0.981687,
		0.994828,
		0.998268,
		0.998474,
		0.997934,
		0.992725,
		0.9761,
		0.942087,
		0.887051,
		0.810056,
		0.712711,
		0.598868,
		0.474322,
		0.346557,
		0.224539,
		0.118558,
		0.040101,
		0.001749,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.007072,
		0.058196,
		0.147054,
		0.260738,
		0.38754,
		0.517072,
		0.640413,
		0.750302,
		0.841376,
		0.910471,
		0.956957,
		0.983029,
		0.993699,
		0.99593,
		0.995982,
		0.994761,
		0.986889,
		0.965238,
		0.924176,
		0.860781,
		0.774998,
		0.66939,
		0.548801,
		0.420042,
		0.291638,
		0.173627,
		0.077397,
		0.015567,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.007286,
		0.059855,
		0.15096,
		0.267095,
		0.396042,
		0.527004,
		0.650766,
		0.75991,
		0.849081,
		0.91532,
		0.95843,
		0.981267,
		0.989607,
		0.990906,
		0.99085,
		0.988479,
		0.977186,
		0.949718,
		0.900971,
		0.828867,
		0.734326,
		0.620939,
		0.494607,
		0.36322,
		0.236406,
		0.125331,
		0.042546,
		0.001862,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.00748,
		0.061335,
		0.154387,
		0.272552,
		0.403127,
		0.534931,
		0.658487,
		0.766258,
		0.852944,
		0.915851,
		0.955287,
		0.974831,
		0.981036,
		0.981687,
		0.981438,
		0.977327,
		0.961826,
		0.927827,
		0.870942,
		0.790059,
		0.687158,
		0.566923,
		0.436364,
		0.304496,
		0.182095,
		0.081496,
		0.01645,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.007635,
		0.062493,
		0.156977,
		0.276483,
		0.407876,
		0.539646,
		0.662108,
		0.767663,
		0.851131,
		0.910151,
		0.945607,
		0.961852,
		0.966179,
		0.966443,
		0.965756,
		0.95923,
		0.938773,
		0.897699,
		0.832516,
		0.743188,
		0.632819,
		0.507231,
		0.374572,
		0.245006,
		0.130464,
		0.044463,
		0.001953,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.007731,
		0.06316,
		0.158311,
		0.278157,
		0.409225,
		0.539764,
		0.659966,
		0.76225,
		0.841631,
		0.896157,
		0.92735,
		0.940363,
		0.943133,
		0.943198,
		0.941677,
		0.932032,
		0.906002,
		0.857596,
		0.784374,
		0.687459,
		0.57112,
		0.442328,
		0.310362,
		0.186511,
		0.083836,
		0.016988,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.007746,
		0.063155,
		0.157944,
		0.276811,
		0.406079,
		0.533882,
		0.650402,
		0.748179,
		0.822512,
		0.871933,
		0.898653,
		0.908605,
		0.910182,
		0.910115,
		0.907247,
		0.893836,
		0.861859,
		0.806267,
		0.7258,
		0.622781,
		0.502643,
		0.373469,
		0.245614,
		0.131419,
		0.04498,
		0.001983,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.007659,
		0.062312,
		0.155471,
		0.271753,
		0.397462,
		0.520775,
		0.632006,
		0.723937,
		0.792248,
		0.836021,
		0.858176,
		0.865353,
		0.86612,
		0.865825,
		0.861048,
		0.843385,
		0.805445,
		0.743327,
		0.65702,
		0.550042,
		0.428931,
		0.302792,
		0.182939,
		0.08262,
		0.016812,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.007452,
		0.060501,
		0.150584,
		0.26248,
		0.382687,
		0.499619,
		0.603886,
		0.68864,
		0.750038,
		0.787765,
		0.805429,
		0.810233,
		0.810531,
		0.809757,
		0.80255,
		0.780424,
		0.736961,
		0.669547,
		0.579429,
		0.471247,
		0.352525,
		0.23323,
		0.125451,
		0.043137,
		0.00191,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.007118,
		0.057659,
		0.143141,
		0.24878,
		0.361507,
		0.470186,
		0.565889,
		0.642277,
		0.696065,
		0.727561,
		0.740998,
		0.743918,
		0.743988,
		0.742376,
		0.732363,
		0.705931,
		0.657896,
		0.586982,
		0.495623,
		0.389445,
		0.276767,
		0.168202,
		0.076359,
		0.015609,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.006658,
		0.053801,
		0.133206,
		0.230803,
		0.334211,
		0.43295,
		0.518731,
		0.585835,
		0.631602,
		0.656947,
		0.666607,
		0.66817,
		0.668102,
		0.665273,
		0.652329,
		0.622176,
		0.571013,
		0.498862,
		0.409197,
		0.308422,
		0.205406,
		0.111127,
		0.038407,
		0.001708,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.006083,
		0.049033,
		0.121059,
		0.209081,
		0.30164,
		0.389109,
		0.463986,
		0.52127,
		0.558952,
		0.578504,
		0.584989,
		0.585696,
		0.585423,
		0.581082,
		0.565401,
		0.532534,
		0.48009,
		0.40925,
		0.324316,
		0.232205,
		0.142042,
		0.064851,
		0.013323,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.005415,
		0.043541,
		0.107181,
		0.184485,
		0.265117,
		0.34047,
		0.403941,
		0.451312,
		0.481216,
		0.495581,
		0.499583,
		0.499836,
		0.499182,
		0.493212,
		0.475331,
		0.441107,
		0.389462,
		0.322525,
		0.24515,
		0.164471,
		0.089552,
		0.031123,
		0.001391,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.004685,
		0.037571,
		0.092199,
		0.158129,
		0.226305,
		0.289251,
		0.341342,
		0.379157,
		0.401925,
		0.4119,
		0.414121,
		0.414176,
		0.412943,
		0.405462,
		0.386226,
		0.352213,
		0.303466,
		0.242785,
		0.175283,
		0.108001,
		0.04962,
		0.01025,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.003928,
		0.031406,
		0.076818,
		0.131251,
		0.187018,
		0.237836,
		0.279078,
		0.308099,
		0.32464,
		0.331119,
		0.332194,
		0.332146,
		0.330213,
		0.321588,
		0.302048,
		0.269848,
		0.22588,
		0.17332,
		0.11724,
		0.064291,
		0.022482,
		0.00101,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.003178,
		0.025332,
		0.061746,
		0.105078,
		0.14903,
		0.188515,
		0.219874,
		0.241175,
		0.252561,
		0.256446,
		0.25688,
		0.256712,
		0.254094,
		0.244896,
		0.226175,
		0.197224,
		0.159489,
		0.116229,
		0.072196,
		0.033402,
		0.006942,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.002468,
		0.019612,
		0.047631,
		0.080714,
		0.113912,
		0.143273,
		0.166038,
		0.180892,
		0.188251,
		0.190361,
		0.190498,
		0.190146,
		0.18702,
		0.177929,
		0.161065,
		0.136466,
		0.105836,
		0.072255,
		0.039939,
		0.014062,
		0.000636,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.001827,
		0.014469,
		0.035006,
		0.059053,
		0.082904,
		0.103642,
		0.119294,
		0.129044,
		0.133455,
		0.134466,
		0.134491,
		0.133926,
		0.130593,
		0.122288,
		0.108085,
		0.088475,
		0.065163,
		0.040846,
		0.019045,
		0.003985,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001275,
		0.010067,
		0.024257,
		0.040725,
		0.056853,
		0.070615,
		0.080692,
		0.086641,
		0.08905,
		0.089462,
		0.089443,
		0.088707,
		0.085527,
		0.078559,
		0.067482,
		0.052975,
		0.036547,
		0.020382,
		0.007231,
		0.000329,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.000827,
		0.006505,
		0.015607,
		0.026068,
		0.036175,
		0.04462,
		0.050596,
		0.053909,
		0.055078,
		0.055213,
		0.05516,
		0.054366,
		0.051666,
		0.046365,
		0.038483,
		0.028687,
		0.018168,
		0.008546,
		0.001801,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000487,
		0.003813,
		0.009106,
		0.015128,
		0.020858,
		0.025537,
		0.02872,
		0.030358,
		0.030845,
		0.030877,
		0.030794,
		0.03008,
		0.028075,
		0.024496,
		0.019499,
		0.013614,
		0.007669,
		0.002744,
		0.000126,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.00025,
		0.001951,
		0.004637,
		0.007659,
		0.010488,
		0.012738,
		0.014202,
		0.014891,
		0.015054,
		0.015059,
		0.014966,
		0.014441,
		0.013182,
		0.011116,
		0.008402,
		0.005384,
		0.002558,
		0.000544,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000105,
		0.000815,
		0.001926,
		0.003161,
		0.004297,
		0.005174,
		0.005716,
		0.005944,
		0.005985,
		0.005983,
		0.005911,
		0.005612,
		0.004985,
		0.004032,
		0.002854,
		0.001626,
		0.000588,
		2.7e-005,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		3.1e-005,
		0.000242,
		0.000568,
		0.000927,
		0.001249,
		0.001491,
		0.001631,
		0.001682,
		0.001689,
		0.001686,
		0.001651,
		0.001535,
		0.001319,
		0.001013,
		0.000658,
		0.000316,
		6.8e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		3.6e-005,
		8.3e-005,
		0.000135,
		0.00018,
		0.000213,
		0.000231,
		0.000236,
		0.000237,
		0.000236,
		0.000228,
		0.000206,
		0.00017,
		0.000122,
		7.1e-005,
		2.6e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		2.6e-005,
		7.1e-005,
		0.000122,
		0.00017,
		0.000206,
		0.000228,
		0.000236,
		0.000237,
		0.000236,
		0.000231,
		0.000213,
		0.00018,
		0.000135,
		8.3e-005,
		3.6e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		6.8e-005,
		0.000316,
		0.000658,
		0.001013,
		0.001319,
		0.001535,
		0.001651,
		0.001686,
		0.001689,
		0.001682,
		0.001631,
		0.001491,
		0.001249,
		0.000927,
		0.000568,
		0.000242,
		3.1e-005,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		2.7e-005,
		0.000588,
		0.001626,
		0.002854,
		0.004032,
		0.004985,
		0.005612,
		0.005911,
		0.005983,
		0.005985,
		0.005944,
		0.005716,
		0.005174,
		0.004297,
		0.003161,
		0.001926,
		0.000815,
		0.000105,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000544,
		0.002558,
		0.005384,
		0.008402,
		0.011116,
		0.013182,
		0.014441,
		0.014966,
		0.015059,
		0.015054,
		0.014891,
		0.014202,
		0.012738,
		0.010488,
		0.007659,
		0.004637,
		0.001951,
		0.00025,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		0.000126,
		0.002744,
		0.007669,
		0.013614,
		0.019499,
		0.024496,
		0.028075,
		0.03008,
		0.030794,
		0.030877,
		0.030845,
		0.030358,
		0.02872,
		0.025537,
		0.020858,
		0.015128,
		0.009106,
		0.003813,
		0.000487,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.001801,
		0.008546,
		0.018168,
		0.028687,
		0.038483,
		0.046365,
		0.051666,
		0.054366,
		0.05516,
		0.055213,
		0.055078,
		0.053909,
		0.050596,
		0.04462,
		0.036175,
		0.026068,
		0.015607,
		0.006505,
		0.000827,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.000329,
		0.007231,
		0.020382,
		0.036547,
		0.052975,
		0.067482,
		0.078559,
		0.085527,
		0.088707,
		0.089443,
		0.089462,
		0.08905,
		0.086641,
		0.080692,
		0.070615,
		0.056853,
		0.040725,
		0.024257,
		0.010067,
		0.001275,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.003985,
		0.019045,
		0.040846,
		0.065163,
		0.088475,
		0.108085,
		0.122288,
		0.130593,
		0.133926,
		0.134491,
		0.134466,
		0.133455,
		0.129044,
		0.119294,
		0.103642,
		0.082904,
		0.059053,
		0.035006,
		0.014469,
		0.001827,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.000636,
		0.014062,
		0.039939,
		0.072255,
		0.105836,
		0.136466,
		0.161065,
		0.177929,
		0.18702,
		0.190146,
		0.190498,
		0.190361,
		0.188251,
		0.180892,
		0.166038,
		0.143273,
		0.113912,
		0.080714,
		0.047631,
		0.019612,
		0.002468,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.006942,
		0.033402,
		0.072196,
		0.116229,
		0.159489,
		0.197224,
		0.226175,
		0.244896,
		0.254094,
		0.256712,
		0.25688,
		0.256446,
		0.252561,
		0.241175,
		0.219874,
		0.188515,
		0.14903,
		0.105078,
		0.061746,
		0.025332,
		0.003178,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.00101,
		0.022482,
		0.064291,
		0.11724,
		0.17332,
		0.22588,
		0.269848,
		0.302048,
		0.321588,
		0.330213,
		0.332146,
		0.332194,
		0.331119,
		0.32464,
		0.308099,
		0.279078,
		0.237836,
		0.187018,
		0.131251,
		0.076818,
		0.031406,
		0.003928,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.01025,
		0.04962,
		0.108001,
		0.175283,
		0.242785,
		0.303466,
		0.352213,
		0.386226,
		0.405462,
		0.412943,
		0.414176,
		0.414121,
		0.4119,
		0.401925,
		0.379157,
		0.341342,
		0.289251,
		0.226305,
		0.158129,
		0.092199,
		0.037571,
		0.004685,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.001391,
		0.031123,
		0.089552,
		0.164471,
		0.24515,
		0.322525,
		0.389462,
		0.441107,
		0.475331,
		0.493212,
		0.499182,
		0.499836,
		0.499583,
		0.495581,
		0.481216,
		0.451312,
		0.403941,
		0.34047,
		0.265117,
		0.184485,
		0.107181,
		0.043541,
		0.005415,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.013323,
		0.064851,
		0.142042,
		0.232205,
		0.324316,
		0.40925,
		0.48009,
		0.532534,
		0.565401,
		0.581082,
		0.585423,
		0.585696,
		0.584989,
		0.578504,
		0.558952,
		0.52127,
		0.463986,
		0.389109,
		0.30164,
		0.209081,
		0.121059,
		0.049033,
		0.006083,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.001708,
		0.038407,
		0.111127,
		0.205406,
		0.308422,
		0.409197,
		0.498862,
		0.571013,
		0.622176,
		0.652329,
		0.665273,
		0.668102,
		0.66817,
		0.666607,
		0.656947,
		0.631602,
		0.585835,
		0.518731,
		0.43295,
		0.334211,
		0.230803,
		0.133206,
		0.053801,
		0.006658,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.015609,
		0.076359,
		0.168202,
		0.276767,
		0.389445,
		0.495623,
		0.586982,
		0.657896,
		0.705931,
		0.732363,
		0.742376,
		0.743988,
		0.743918,
		0.740998,
		0.727561,
		0.696065,
		0.642277,
		0.565889,
		0.470186,
		0.361507,
		0.24878,
		0.143141,
		0.057659,
		0.007118,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.00191,
		0.043137,
		0.125451,
		0.23323,
		0.352525,
		0.471247,
		0.579429,
		0.669547,
		0.736961,
		0.780424,
		0.80255,
		0.809757,
		0.810531,
		0.810233,
		0.805429,
		0.787765,
		0.750038,
		0.68864,
		0.603886,
		0.499619,
		0.382687,
		0.26248,
		0.150584,
		0.060501,
		0.007452,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.016812,
		0.08262,
		0.182939,
		0.302792,
		0.428931,
		0.550042,
		0.65702,
		0.743327,
		0.805445,
		0.843385,
		0.861048,
		0.865825,
		0.86612,
		0.865353,
		0.858176,
		0.836021,
		0.792248,
		0.723937,
		0.632006,
		0.520775,
		0.397462,
		0.271753,
		0.155471,
		0.062312,
		0.007659,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.001983,
		0.04498,
		0.131419,
		0.245614,
		0.373469,
		0.502643,
		0.622781,
		0.7258,
		0.806267,
		0.861859,
		0.893836,
		0.907247,
		0.910115,
		0.910182,
		0.908605,
		0.898653,
		0.871933,
		0.822512,
		0.748179,
		0.650402,
		0.533882,
		0.406079,
		0.276811,
		0.157944,
		0.063155,
		0.007746,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.016988,
		0.083836,
		0.186511,
		0.310362,
		0.442328,
		0.57112,
		0.687459,
		0.784374,
		0.857596,
		0.906002,
		0.932032,
		0.941677,
		0.943198,
		0.943133,
		0.940363,
		0.92735,
		0.896157,
		0.841631,
		0.76225,
		0.659966,
		0.539764,
		0.409225,
		0.278157,
		0.158311,
		0.06316,
		0.007731,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.001953,
		0.044463,
		0.130464,
		0.245006,
		0.374572,
		0.507231,
		0.632819,
		0.743188,
		0.832516,
		0.897699,
		0.938773,
		0.95923,
		0.965756,
		0.966443,
		0.966179,
		0.961852,
		0.945607,
		0.910151,
		0.851131,
		0.767663,
		0.662108,
		0.539646,
		0.407876,
		0.276483,
		0.156977,
		0.062493,
		0.007635,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.01645,
		0.081496,
		0.182095,
		0.304496,
		0.436364,
		0.566923,
		0.687158,
		0.790059,
		0.870942,
		0.927827,
		0.961826,
		0.977327,
		0.981438,
		0.981687,
		0.981036,
		0.974831,
		0.955287,
		0.915851,
		0.852944,
		0.766258,
		0.658487,
		0.534931,
		0.403127,
		0.272552,
		0.154387,
		0.061335,
		0.00748,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.001862,
		0.042546,
		0.125331,
		0.236406,
		0.36322,
		0.494607,
		0.620939,
		0.734326,
		0.828867,
		0.900971,
		0.949718,
		0.977186,
		0.988479,
		0.99085,
		0.990906,
		0.989607,
		0.981267,
		0.95843,
		0.91532,
		0.849081,
		0.75991,
		0.650766,
		0.527004,
		0.396042,
		0.267095,
		0.15096,
		0.059855,
		0.007286,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.015567,
		0.077397,
		0.173627,
		0.291638,
		0.420042,
		0.548801,
		0.66939,
		0.774998,
		0.860781,
		0.924176,
		0.965238,
		0.986889,
		0.994761,
		0.995982,
		0.99593,
		0.993699,
		0.983029,
		0.956957,
		0.910471,
		0.841376,
		0.750302,
		0.640413,
		0.517072,
		0.38754,
		0.260738,
		0.147054,
		0.058196,
		0.007072,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.001749,
		0.040101,
		0.118558,
		0.224539,
		0.346557,
		0.474322,
		0.598868,
		0.712711,
		0.810056,
		0.887051,
		0.942087,
		0.9761,
		0.992725,
		0.997934,
		0.998474,
		0.998268,
		0.994828,
		0.981687,
		0.952465,
		0.902878,
		0.831316,
		0.73878,
		0.628589,
		0.50608,
		0.378334,
		0.253963,
		0.142941,
		0.056465,
		0.006851,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.014606,
		0.072856,
		0.164044,
		0.276679,
		0.400334,
		0.525747,
		0.644953,
		0.751442,
		0.840368,
		0.908799,
		0.956001,
		0.983695,
		0.996096,
		0.99933,
		0.999523,
		0.999017,
		0.994116,
		0.97841,
		0.94614,
		0.893693,
		0.819979,
		0.726308,
		0.616116,
		0.494688,
		0.368915,
		0.247099,
		0.138805,
		0.054735,
		0.00663,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.001637,
		0.037644,
		0.111669,
		0.212284,
		0.329012,
		0.452411,
		0.574175,
		0.687268,
		0.786085,
		0.86666,
		0.926912,
		0.966903,
		0.989042,
		0.997988,
		0.999837,
		0.99988,
		0.998871,
		0.99229,
		0.973965,
		0.938756,
		0.883665,
		0.808054,
		0.713495,
		0.603506,
		0.483303,
		0.359581,
		0.240343,
		0.134756,
		0.053048,
		0.006416,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.013691,
		0.068505,
		0.154774,
		0.262037,
		0.380753,
		0.502393,
		0.619534,
		0.726004,
		0.817037,
		0.889481,
		0.942036,
		0.975486,
		0.992826,
		0.999028,
		0.999976,
		0.999936,
		0.998198,
		0.989757,
		0.968789,
		0.930757,
		0.873218,
		0.795928,
		0.700674,
		0.591031,
		0.472135,
		0.350485,
		0.233791,
		0.130845,
		0.051425,
		0.006211,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.001534,
		0.035363,
		0.105229,
		0.200736,
		0.312315,
		0.431293,
		0.549981,
		0.661781,
		0.761316,
		0.844598,
		0.909231,
		0.954629,
		0.982219,
		0.995487,
		0.99958,
		0.999998,
		0.999839,
		0.997146,
		0.9867,
		0.963094,
		0.922363,
		0.862561,
		0.783783,
		0.687995,
		0.578809,
		0.461269,
		0.341684,
		0.227479,
		0.127091,
		0.04987,
		0.006014,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.012854,
		0.064503,
		0.146196,
		0.248386,
		0.362328,
		0.480153,
		0.594955,
		0.700893,
		0.793332,
		0.869007,
		0.926215,
		0.965026,
		0.987433,
		0.997312,
		0.99985,
		1,
		0.999606,
		0.995745,
		0.983187,
		0.956972,
		0.913673,
		0.851782,
		0.771692,
		0.675515,
		0.566876,
		0.450728,
		0.333186,
		0.221409,
		0.123492,
		0.048385,
		0.005827,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.00144,
		0.033279,
		0.099316,
		0.190067,
		0.29677,
		0.411446,
		0.526971,
		0.63716,
		0.736881,
		0.822187,
		0.89048,
		0.940699,
		0.973507,
		0.99139,
		0.998511,
		0.999962,
		0.999995,
		0.999205,
		0.993987,
		0.97924,
		0.950468,
		0.904734,
		0.840925,
		0.759685,
		0.663248,
		0.555238,
		0.440506,
		0.324985,
		0.215573,
		0.120042,
		0.046964,
		0.005649,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.012092,
		0.06084,
		0.138302,
		0.23574,
		0.345123,
		0.459182,
		0.571482,
		0.676513,
		0.769799,
		0.848031,
		0.909232,
		0.952932,
		0.980326,
		0.994317,
		0.999251,
		0.999995,
		0.999964,
		0.998595,
		0.991857,
		0.974876,
		0.943612,
		0.895582,
		0.830017,
		0.747781,
		0.651201,
		0.54389,
		0.430596,
		0.317068,
		0.209959,
		0.116734,
		0.045605,
		0.005478,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.001354,
		0.031373,
		0.093885,
		0.180211,
		0.282312,
		0.392831,
		0.505161,
		0.613507,
		0.712986,
		0.799732,
		0.871032,
		0.925483,
		0.963156,
		0.985719,
		0.996415,
		0.999671,
		1,
		0.999875,
		0.99774,
		0.989348,
		0.970112,
		0.936437,
		0.886248,
		0.819083,
		0.735993,
		0.639377,
		0.532828,
		0.420985,
		0.309424,
		0.204557,
		0.113559,
		0.044304,
		0.005315,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.011395,
		0.057479,
		0.131022,
		0.22401,
		0.32905,
		0.439414,
		0.549108,
		0.652936,
		0.746597,
		0.826794,
		0.89137,
		0.939451,
		0.971601,
		0.989904,
		0.997863,
		0.999882,
		1,
		0.999689,
		0.996609,
		0.986456,
		0.964969,
		0.928972,
		0.876762,
		0.808146,
		0.724333,
		0.62778,
		0.522046,
		0.411665,
		0.30204,
		0.199356,
		0.110511,
		0.043058,
		0.005159,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.001275,
		0.029626,
		0.088884,
		0.171089,
		0.268846,
		0.375363,
		0.484498,
		0.590828,
		0.689715,
		0.777401,
		0.851122,
		0.909232,
		0.951351,
		0.978488,
		0.993081,
		0.998815,
		0.99997,
		0.999996,
		0.999369,
		0.99518,
		0.983187,
		0.959469,
		0.921246,
		0.867151,
		0.797225,
		0.712812,
		0.61641,
		0.511539,
		0.402626,
		0.294906,
		0.194346,
		0.107582,
		0.041863,
		0.00501,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.010757,
		0.054388,
		0.124297,
		0.213114,
		0.314019,
		0.420779,
		0.527802,
		0.630193,
		0.723836,
		0.805484,
		0.872864,
		0.924811,
		0.961395,
		0.984024,
		0.995431,
		0.999403,
		0.999996,
		0.999972,
		0.998879,
		0.993437,
		0.979547,
		0.953636,
		0.913289,
		0.85744,
		0.786339,
		0.701439,
		0.605268,
		0.501302,
		0.393858,
		0.288011,
		0.18952,
		0.104768,
		0.040718,
		0.004868,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.001204,
		0.028021,
		0.084269,
		0.162631,
		0.25629,
		0.358959,
		0.464929,
		0.569117,
		0.667127,
		0.755329,
		0.830949,
		0.892179,
		0.938296,
		0.969784,
		0.988401,
		0.997117,
		0.999738,
		1,
		0.9999,
		0.998186,
		0.991369,
		0.975548,
		0.947493,
		0.905127,
		0.847654,
		0.775502,
		0.690221,
		0.594354,
		0.491327,
		0.385352,
		0.281345,
		0.184867,
		0.102062,
		0.039618,
		0.004731,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.010171,
		0.051539,
		0.118071,
		0.202976,
		0.299949,
		0.403207,
		0.507526,
		0.608301,
		0.701601,
		0.78425,
		0.85392,
		0.909232,
		0.949876,
		0.976711,
		0.991798,
		0.99828,
		0.999906,
		1,
		0.999752,
		0.997265,
		0.988971,
		0.971204,
		0.941064,
		0.896786,
		0.837812,
		0.764729,
		0.679164,
		0.583667,
		0.48161,
		0.377099,
		0.2749,
		0.18038,
		0.099458,
		0.038563,
		0.0046,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.001138,
		0.026543,
		0.080003,
		0.154777,
		0.244566,
		0.343545,
		0.446396,
		0.548355,
		0.645261,
		0.733619,
		0.81068,
		0.874529,
		0.924192,
		0.959736,
		0.982358,
		0.994377,
		0.999046,
		0.999976,
		0.999997,
		0.999493,
		0.996092,
		0.986243,
		0.966531,
		0.934374,
		0.888291,
		0.827935,
		0.754033,
		0.668272,
		0.573204,
		0.472143,
		0.369091,
		0.268666,
		0.176052,
		0.096952,
		0.037549,
		0.004475,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.009632,
		0.048908,
		0.112297,
		0.193529,
		0.286766,
		0.38663,
		0.488241,
		0.587261,
		0.679948,
		0.763216,
		0.834714,
		0.892918,
		0.937225,
		0.968053,
		0.986896,
		0.996286,
		0.999519,
		0.999997,
		0.999977,
		0.999094,
		0.994653,
		0.983187,
		0.961546,
		0.927446,
		0.879663,
		0.818041,
		0.743424,
		0.657551,
		0.562964,
		0.462921,
		0.361318,
		0.262634,
		0.171875,
		0.094539,
		0.036574,
		0.004354,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.001077,
		0.025179,
		0.076051,
		0.14747,
		0.233606,
		0.32905,
		0.428844,
		0.528519,
		0.624138,
		0.712348,
		0.790449,
		0.856465,
		0.909232,
		0.948498,
		0.974996,
		0.990485,
		0.997655,
		0.999788,
		1,
		0.99992,
		0.998528,
		0.992935,
		0.97981,
		0.956268,
		0.920303,
		0.870925,
		0.808146,
		0.732912,
		0.647002,
		0.552945,
		0.453937,
		0.353774,
		0.256796,
		0.167843,
		0.092214,
		0.035637,
		0.004238,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.009134,
		0.046472,
		0.106933,
		0.184715,
		0.2744,
		0.370984,
		0.469901,
		0.567067,
		0.65892,
		0.742476,
		0.815394,
		0.87605,
		0.92362,
		0.958172,
		0.980727,
		0.99327,
		0.9986,
		0.999924,
		1,
		0.999799,
		0.997769,
		0.990931,
		0.97612,
		0.950715,
		0.912966,
		0.862095,
		0.798264,
		0.722505,
		0.636628,
		0.543144,
		0.445185,
		0.346449,
		0.251144,
		0.163949,
		0.089974,
		0.034735,
		0.004127,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.001021,
		0.023917,
		0.072383,
		0.140662,
		0.223347,
		0.315408,
		0.412216,
		0.509577,
		0.603767,
		0.691577,
		0.770369,
		0.838141,
		0.893596,
		0.936229,
		0.966404,
		0.985399,
		0.995385,
		0.999223,
		0.999981,
		0.999997,
		0.999588,
		0.996799,
		0.988638,
		0.97213,
		0.944907,
		0.905457,
		0.853193,
		0.788409,
		0.71221,
		0.62643,
		0.533556,
		0.436658,
		0.339336,
		0.245671,
		0.160186,
		0.087813,
		0.033867,
		0.00402,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.008674,
		0.044213,
		0.101941,
		0.176478,
		0.26279,
		0.356209,
		0.452462,
		0.547701,
		0.638541,
		0.722107,
		0.796084,
		0.858788,
		0.909232,
		0.947206,
		0.973345,
		0.989154,
		0.99695,
		0.999608,
		0.999998,
		0.999982,
		0.999261,
		0.995601,
		0.986055,
		0.967852,
		0.938862,
		0.897796,
		0.844236,
		0.778592,
		0.702034,
		0.61641,
		0.52418,
		0.428351,
		0.332429,
		0.24037,
		0.156549,
		0.085729,
		0.033031,
		0.003918,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.00097,
		0.022747,
		0.068974,
		0.13431,
		0.213733,
		0.302559,
		0.39646,
		0.491497,
		0.584147,
		0.671346,
		0.750529,
		0.819691,
		0.877444,
		0.92309,
		0.956695,
		0.979138,
		0.992122,
		0.998073,
		0.999827,
		1,
		0.999935,
		0.998793,
		0.994162,
		0.983187,
		0.963299,
		0.9326,
		0.890001,
		0.835238,
		0.768825,
		0.691982,
		0.606567,
		0.515011,
		0.420258,
		0.325719,
		0.235233,
		0.153034,
		0.083718,
		0.032226,
		0.003819,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.008248,
		0.042115,
		0.097288,
		0.168771,
		0.251877,
		0.342249,
		0.43588,
		0.529145,
		0.618827,
		0.702164,
		0.776885,
		0.841272,
		0.89422,
		0.935301,
		0.964833,
		0.983919,
		0.994427,
		0.99885,
		0.999938,
		1,
		0.999836,
		0.998163,
		0.992476,
		0.980037,
		0.958486,
		0.926139,
		0.882091,
		0.826216,
		0.759116,
		0.682058,
		0.596902,
		0.506046,
		0.412372,
		0.3192,
		0.230255,
		0.149633,
		0.081776,
		0.031449,
		0.003724,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.000922,
		0.021661,
		0.065798,
		0.128373,
		0.204713,
		0.290446,
		0.381525,
		0.474244,
		0.565272,
		0.651685,
		0.731,
		0.801227,
		0.860918,
		0.909232,
		0.945995,
		0.971756,
		0.987816,
		0.996177,
		0.999361,
		0.999984,
		0.999998,
		0.999662,
		0.997353,
		0.990536,
		0.976614,
		0.953429,
		0.919497,
		0.874082,
		0.817181,
		0.749475,
		0.672265,
		0.587413,
		0.497281,
		0.404687,
		0.312866,
		0.225429,
		0.146343,
		0.079901,
		0.0307,
		0.003632,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.007852,
		0.040162,
		0.092945,
		0.161551,
		0.24161,
		0.32905,
		0.42011,
		0.511372,
		0.599783,
		0.68269,
		0.757879,
		0.823621,
		0.878726,
		0.922598,
		0.955299,
		0.977591,
		0.990947,
		0.997473,
		0.999678,
		0.999998,
		0.999985,
		0.999391,
		0.996348,
		0.98834,
		0.972926,
		0.948142,
		0.912691,
		0.86599,
		0.808146,
		0.739908,
		0.662608,
		0.5781,
		0.488711,
		0.397198,
		0.306709,
		0.22075,
		0.14316,
		0.078089,
		0.029978,
		0.003544,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.000878,
		0.020651,
		0.062836,
		0.122817,
		0.19624,
		0.279019,
		0.367361,
		0.457781,
		0.547127,
		0.632611,
		0.711836,
		0.782841,
		0.844144,
		0.894796,
		0.934434,
		0.963338,
		0.982461,
		0.993424,
		0.998403,
		0.999858,
		1,
		0.999946,
		0.999001,
		0.995135,
		0.98589,
		0.968983,
		0.942641,
		0.905738,
		0.85783,
		0.799123,
		0.730424,
		0.653088,
		0.568961,
		0.480333,
		0.3899,
		0.300724,
		0.21621,
		0.140077,
		0.076337,
		0.029281,
		0.003458,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.007484,
		0.038342,
		0.088884,
		0.154777,
		0.231941,
		0.316562,
		0.405109,
		0.494356,
		0.581406,
		0.663715,
		0.739133,
		0.805937,
		0.862877,
		0.909232,
		0.944857,
		0.97023,
		0.986479,
		0.995346,
		0.999046,
		0.999949,
		1,
		0.999864,
		0.998474,
		0.993706,
		0.983187,
		0.964797,
		0.936941,
		0.898654,
		0.849616,
		0.79012,
		0.721027,
		0.643707,
		0.559995,
		0.472143,
		0.382786,
		0.294906,
		0.211807,
		0.137092,
		0.074644,
		0.028608,
		0.003376,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.000836,
		0.01971,
		0.060068,
		0.117611,
		0.188271,
		0.268228,
		0.353922,
		0.442072,
		0.529695,
		0.614133,
		0.693079,
		0.76461,
		0.827229,
		0.879909,
		0.922139,
		0.953979,
		0.976091,
		0.989754,
		0.996806,
		0.99947,
		0.999987,
		0.999998,
		0.99972,
		0.997793,
		0.992055,
		0.980236,
		0.960379,
		0.931058,
		0.891454,
		0.84136,
		0.781149,
		0.711723,
		0.634466,
		0.5512,
		0.464137,
		0.375851,
		0.289247,
		0.207533,
		0.1342,
		0.073006,
		0.027957,
		0.003297,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.007142,
		0.036642,
		0.085081,
		0.148414,
		0.222825,
		0.304739,
		0.390835,
		0.478069,
		0.563688,
		0.645261,
		0.720701,
		0.788304,
		0.846787,
		0.895329,
		0.933622,
		0.961912,
		0.98103,
		0.992386,
		0.997888,
		0.999732,
		0.999998,
		0.999988,
		0.999493,
		0.996943,
		0.990178,
		0.977043,
		0.955742,
		0.925006,
		0.884152,
		0.833075,
		0.772216,
		0.702516,
		0.625367,
		0.542574,
		0.45631,
		0.369091,
		0.283743,
		0.203384,
		0.131398,
		0.071421,
		0.027329,
		0.003221,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.000798,
		0.018832,
		0.057479,
		0.112726,
		0.18077,
		0.258031,
		0.341166,
		0.427082,
		0.512956,
		0.596255,
		0.67476,
		0.746597,
		0.810266,
		0.864686,
		0.909232,
		0.943785,
		0.968764,
		0.985151,
		0.994468,
		0.998665,
		0.999882,
		1,
		0.999956,
		0.999166,
		0.995914,
		0.988073,
		0.973615,
		0.950898,
		0.9188,
		0.876762,
		0.82477,
		0.763329,
		0.69341,
		0.61641,
		0.534114,
		0.448658,
		0.362499,
		0.278389,
		0.199356,
		0.128681,
		0.069887,
		0.026722,
		0.003147,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.006822,
		0.035053,
		0.081516,
		0.142431,
		0.214224,
		0.293538,
		0.37725,
		0.462481,
		0.546619,
		0.627338,
		0.702622,
		0.770794,
		0.830552,
		0.881004,
		0.921711,
		0.952728,
		0.974637,
		0.988552,
		0.996084,
		0.999202,
		0.999957,
		1,
		0.999887,
		0.998722,
		0.994696,
		0.985742,
		0.969962,
		0.94586,
		0.912454,
		0.869296,
		0.816458,
		0.754496,
		0.684409,
		0.607595,
		0.52582,
		0.441178,
		0.356071,
		0.273179,
		0.195443,
		0.126047,
		0.068402,
		0.026134,
		0.003075,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.000762,
		0.018011,
		0.055053,
		0.108136,
		0.1737,
		0.248386,
		0.32905,
		0.412775,
		0.496887,
		0.578973,
		0.656903,
		0.728851,
		0.793332,
		0.849227,
		0.895825,
		0.932861,
		0.960553,
		0.97963,
		0.991323,
		0.997312,
		0.999556,
		0.999989,
		0.999999,
		0.999766,
		0.998145,
		0.993282,
		0.983187,
		0.96609,
		0.94064,
		0.90598,
		0.861767,
		0.808146,
		0.745722,
		0.675515,
		0.598922,
		0.517687,
		0.433865,
		0.349803,
		0.268109,
		0.191643,
		0.123492,
		0.066963,
		0.025566,
		0.003007,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.006524,
		0.033565,
		0.07817,
		0.136799,
		0.2061,
		0.282919,
		0.364313,
		0.447562,
		0.530184,
		0.609954,
		0.684929,
		0.753466,
		0.814257,
		0.866359,
		0.909232,
		0.942774,
		0.967356,
		0.983837,
		0.993551,
		0.998222,
		0.999776,
		0.999999,
		0.99999,
		0.999575,
		0.997422,
		0.991668,
		0.980411,
		0.962012,
		0.935251,
		0.899393,
		0.854185,
		0.799844,
		0.737013,
		0.66673,
		0.59039,
		0.509713,
		0.426716,
		0.343689,
		0.263174,
		0.18795,
		0.121013,
		0.065569,
		0.025016,
		0.00294,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.000729,
		0.017243,
		0.052777,
		0.103819,
		0.16703,
		0.239256,
		0.317536,
		0.399117,
		0.481465,
		0.56228,
		0.63952,
		0.711414,
		0.776495,
		0.833622,
		0.882021,
		0.921311,
		0.951543,
		0.973229,
		0.987347,
		0.995313,
		0.998875,
		0.999901,
		1,
		0.999963,
		0.999299,
		0.996543,
		0.989852,
		0.97742,
		0.957735,
		0.929706,
		0.892703,
		0.846562,
		0.791559,
		0.728373,
		0.658057,
		0.582,
		0.501896,
		0.419726,
		0.337724,
		0.258369,
		0.18436,
		0.118608,
		0.064217,
		0.024483,
		0.002875,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.006244,
		0.032169,
		0.075024,
		0.13149,
		0.198419,
		0.272844,
		0.35199,
		0.433283,
		0.514364,
		0.593109,
		0.667644,
		0.736367,
		0.797974,
		0.851487,
		0.896287,
		0.932145,
		0.959257,
		0.978262,
		0.990244,
		0.996681,
		0.999327,
		0.999964,
		1,
		0.999905,
		0.998921,
		0.995499,
		0.987832,
		0.974219,
		0.953272,
		0.924016,
		0.885923,
		0.838908,
		0.783299,
		0.719808,
		0.649498,
		0.573749,
		0.494234,
		0.412891,
		0.331906,
		0.25369,
		0.180871,
		0.116272,
		0.062907,
		0.023968,
		0.002813,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.000697,
		0.016522,
		0.050639,
		0.099754,
		0.160731,
		0.230606,
		0.306588,
		0.386075,
		0.466665,
		0.546167,
		0.622621,
		0.694316,
		0.759807,
		0.817948,
		0.867913,
		0.909232,
		0.94182,
		0.966004,
		0.982541,
		0.992605,
		0.997721,
		0.999626,
		0.999991,
		0.999999,
		0.999803,
		0.998429,
		0.994282,
		0.98561,
		0.970816,
		0.948631,
		0.918192,
		0.879064,
		0.83123,
		0.77507,
		0.71132,
		0.641052,
		0.565638,
		0.486723,
		0.406208,
		0.326228,
		0.249134,
		0.177478,
		0.114005,
		0.061637,
		0.023468,
		0.002753,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.005983,
		0.030859,
		0.072063,
		0.12648,
		0.191151,
		0.26328,
		0.340247,
		0.419614,
		0.499142,
		0.5768,
		0.650784,
		0.719536,
		0.781763,
		0.836467,
		0.882967,
		0.920935,
		0.950418,
		0.971868,
		0.986146,
		0.994502,
		0.998492,
		0.999811,
		0.999999,
		0.999991,
		0.999641,
		0.997811,
		0.992887,
		0.983187,
		0.967218,
		0.943825,
		0.912247,
		0.872136,
		0.823539,
		0.766878,
		0.702914,
		0.632722,
		0.557664,
		0.47936,
		0.399673,
		0.320688,
		0.244695,
		0.174178,
		0.111802,
		0.060404,
		0.022984,
		0.002694,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.000668,
		0.015846,
		0.048628,
		0.09592,
		0.154777,
		0.222405,
		0.296172,
		0.373618,
		0.452462,
		0.53062,
		0.606211,
		0.67758,
		0.743316,
		0.80227,
		0.853585,
		0.896717,
		0.931471,
		0.95802,
		0.976929,
		0.989154,
		0.996003,
		0.999046,
		0.999917,
		1,
		0.999969,
		0.999406,
		0.997056,
		0.991312,
		0.980567,
		0.963433,
		0.938862,
		0.906191,
		0.865149,
		0.815842,
		0.758729,
		0.694592,
		0.624508,
		0.549827,
		0.472143,
		0.393283,
		0.315281,
		0.24037,
		0.170967,
		0.109662,
		0.059208,
		0.022514,
		0.002638,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.005737,
		0.029626,
		0.069274,
		0.121749,
		0.184267,
		0.254193,
		0.32905,
		0.406528,
		0.484498,
		0.561021,
		0.63436,
		0.703003,
		0.765677,
		0.82137,
		0.869359,
		0.909232,
		0.940917,
		0.964706,
		0.981267,
		0.991635,
		0.997169,
		0.999429,
		0.99997,
		1,
		0.99992,
		0.999083,
		0.996156,
		0.989554,
		0.977753,
		0.959469,
		0.933754,
		0.900035,
		0.858112,
		0.808146,
		0.750628,
		0.686356,
		0.61641,
		0.542124,
		0.465069,
		0.387032,
		0.310002,
		0.236155,
		0.167843,
		0.107582,
		0.058047,
		0.022059,
		0.002583,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.000641,
		0.015211,
		0.046734,
		0.092302,
		0.149143,
		0.214622,
		0.286256,
		0.361715,
		0.438833,
		0.515624,
		0.590287,
		0.661224,
		0.727057,
		0.786646,
		0.839109,
		0.883851,
		0.920582,
		0.949349,
		0.970552,
		0.984953,
		0.993658,
		0.998056,
		0.999682,
		0.999992,
		0.999999,
		0.999833,
		0.998662,
		0.995103,
		0.987613,
		0.974752,
		0.955335,
		0.928511,
		0.893788,
		0.851035,
		0.800458,
		0.742579,
		0.67821,
		0.608428,
		0.534555,
		0.458135,
		0.380919,
		0.30485,
		0.232048,
		0.164802,
		0.105561,
		0.05692,
		0.021618,
		0.00253,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.005506,
		0.028466,
		0.066642,
		0.117276,
		0.177742,
		0.245554,
		0.318369,
		0.393997,
		0.470412,
		0.545761,
		0.618377,
		0.686792,
		0.749757,
		0.806258,
		0.855538,
		0.897121,
		0.930835,
		0.956839,
		0.975631,
		0.988059,
		0.995285,
		0.998713,
		0.999839,
		0.999999,
		0.999993,
		0.999695,
		0.99813,
		0.993892,
		0.98549,
		0.971569,
		0.951039,
		0.923141,
		0.887461,
		0.843926,
		0.792785,
		0.734587,
		0.670154,
		0.600563,
		0.527118,
		0.451339,
		0.37494,
		0.299819,
		0.228043,
		0.161842,
		0.103595,
		0.055825,
		0.021189,
		0.002478,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.000615,
		0.014613,
		0.044948,
		0.088884,
		0.143807,
		0.207232,
		0.276811,
		0.350337,
		0.425753,
		0.501165,
		0.574848,
		0.645261,
		0.711062,
		0.771123,
		0.824552,
		0.870708,
		0.909232,
		0.940063,
		0.963459,
		0.980016,
		0.99065,
		0.996571,
		0.999185,
		0.999929,
		1,
		0.999973,
		0.999493,
		0.997478,
		0.99252,
		0.983187,
		0.968209,
		0.946591,
		0.917657,
		0.881063,
		0.836792,
		0.785132,
		0.726654,
		0.662191,
		0.592813,
		0.519811,
		0.444677,
		0.369091,
		0.294906,
		0.224139,
		0.15896,
		0.101684,
		0.054761,
		0.020774,
		0.002428,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.005289,
		0.027373,
		0.064157,
		0.113042,
		0.171552,
		0.237335,
		0.308175,
		0.381994,
		0.456863,
		0.531011,
		0.602837,
		0.670922,
		0.73404,
		0.791183,
		0.84157,
		0.884676,
		0.920249,
		0.948332,
		0.969281,
		0.983772,
		0.992789,
		0.997571,
		0.999512,
		0.999974,
		1,
		0.999932,
		0.999216,
		0.996697,
		0.990983,
		0.980706,
		0.964681,
		0.941999,
		0.912065,
		0.874601,
		0.82964,
		0.777504,
		0.718785,
		0.654322,
		0.58518,
		0.512632,
		0.438147,
		0.363368,
		0.290107,
		0.220331,
		0.156153,
		0.099824,
		0.053727,
		0.02037,
		0.00238,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.00059,
		0.014049,
		0.043262,
		0.08565,
		0.138748,
		0.200209,
		0.267809,
		0.339457,
		0.413199,
		0.487226,
		0.559887,
		0.629697,
		0.695352,
		0.755743,
		0.809968,
		0.85736,
		0.897499,
		0.930235,
		0.95571,
		0.974369,
		0.986964,
		0.994532,
		0.998331,
		0.999729,
		0.999993,
		0.999999,
		0.999857,
		0.998853,
		0.995782,
		0.989282,
		0.978051,
		0.960989,
		0.937272,
		0.906376,
		0.868085,
		0.822478,
		0.769907,
		0.710982,
		0.646547,
		0.577661,
		0.50558,
		0.431746,
		0.357769,
		0.285421,
		0.216617,
		0.153419,
		0.098015,
		0.052722,
		0.019978,
		0.002333,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.005084,
		0.026342,
		0.061808,
		0.109032,
		0.165674,
		0.229511,
		0.298442,
		0.370494,
		0.44383,
		0.516759,
		0.587741,
		0.655405,
		0.718555,
		0.77619,
		0.827516,
		0.87197,
		0.909232,
		0.939252,
		0.962262,
		0.97879,
		0.989654,
		0.995933,
		0.998895,
		0.999863,
		0.999999,
		0.999994,
		0.999739,
		0.998392,
		0.994726,
		0.987414,
		0.975226,
		0.957143,
		0.932418,
		0.900598,
		0.861523,
		0.815311,
		0.762345,
		0.703249,
		0.638868,
		0.570256,
		0.498652,
		0.425471,
		0.352291,
		0.280842,
		0.212995,
		0.150754,
		0.096253,
		0.051745,
		0.019597,
		0.002287,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.000567,
		0.013517,
		0.041669,
		0.082589,
		0.133949,
		0.193529,
		0.259225,
		0.32905,
		0.401146,
		0.473788,
		0.545395,
		0.614537,
		0.679948,
		0.740538,
		0.795408,
		0.843868,
		0.885451,
		0.919935,
		0.947363,
		0.968053,
		0.982605,
		0.991898,
		0.997042,
		0.9993,
		0.999939,
		1,
		0.999977,
		0.999565,
		0.997826,
		0.993526,
		0.985382,
		0.972236,
		0.953149,
		0.927446,
		0.89474,
		0.854921,
		0.808146,
		0.754822,
		0.695586,
		0.631286,
		0.562964,
		0.491846,
		0.419319,
		0.34693,
		0.276369,
		0.20946,
		0.148158,
		0.094539,
		0.050795,
		0.019227,
		0.002243,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.004891,
		0.025367,
		0.059585,
		0.105229,
		0.160089,
		0.222058,
		0.289144,
		0.359472,
		0.431293,
		0.502989,
		0.573084,
		0.64025,
		0.703324,
		0.761317,
		0.813428,
		0.859064,
		0.897854,
		0.929666,
		0.95463,
		0.973143,
		0.985874,
		0.99375,
		0.997904,
		0.999581,
		0.999978,
		1,
		0.999942,
		0.999326,
		0.997147,
		0.992177,
		0.983187,
		0.969087,
		0.949014,
		0.922365,
		0.88881,
		0.848287,
		0.800988,
		0.747342,
		0.687996,
		0.623799,
		0.555785,
		0.48516,
		0.413289,
		0.341684,
		0.271997,
		0.20601,
		0.145627,
		0.09287,
		0.049871,
		0.018867,
		0.0022,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.000546,
		0.013016,
		0.040162,
		0.079688,
		0.129392,
		0.187172,
		0.251033,
		0.319089,
		0.389573,
		0.460836,
		0.531365,
		0.599783,
		0.664862,
		0.725537,
		0.780913,
		0.830285,
		0.873152,
		0.909232,
		0.938482,
		0.961112,
		0.977591,
		0.988652,
		0.99526,
		0.99856,
		0.999767,
		0.999994,
		0.999999,
		0.999877,
		0.999011,
		0.996348,
		0.990679,
		0.980831,
		0.965784,
		0.944746,
		0.917181,
		0.882814,
		0.841627,
		0.793842,
		0.739908,
		0.680482,
		0.61641,
		0.548718,
		0.478594,
		0.407377,
		0.336549,
		0.267725,
		0.202643,
		0.14316,
		0.091244,
		0.048971,
		0.018518,
		0.002158,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.004709,
		0.024446,
		0.057479,
		0.101621,
		0.154777,
		0.214952,
		0.280256,
		0.348906,
		0.419232,
		0.48969,
		0.55886,
		0.625463,
		0.688367,
		0.746597,
		0.799352,
		0.846017,
		0.886177,
		0.919639,
		0.94644,
		0.966867,
		0.981456,
		0.990993,
		0.996474,
		0.999046,
		0.999882,
		0.999999,
		0.999995,
		0.999775,
		0.99861,
		0.995423,
		0.989031,
		0.978317,
		0.962333,
		0.940353,
		0.911904,
		0.876762,
		0.834948,
		0.786713,
		0.732524,
		0.673044,
		0.609118,
		0.541761,
		0.472143,
		0.40158,
		0.331523,
		0.263549,
		0.199356,
		0.140754,
		0.08966,
		0.048096,
		0.018178,
		0.002117,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.000525,
		0.012541,
		0.038736,
		0.076937,
		0.125061,
		0.181117,
		0.243212,
		0.309553,
		0.378457,
		0.448352,
		0.517785,
		0.585432,
		0.650105,
		0.710763,
		0.76652,
		0.816662,
		0.860662,
		0.898188,
		0.929128,
		0.953597,
		0.971951,
		0.984791,
		0.992945,
		0.997435,
		0.999396,
		0.999948,
		1,
		0.99998,
		0.999625,
		0.998117,
		0.994368,
		0.987232,
		0.97565,
		0.958739,
		0.935842,
		0.90654,
		0.870658,
		0.828254,
		0.779606,
		0.725192,
		0.665684,
		0.601923,
		0.534913,
		0.465807,
		0.395897,
		0.326602,
		0.259467,
		0.196146,
		0.138407,
		0.088117,
		0.047243,
		0.017847,
		0.002078,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.004537,
		0.023574,
		0.055483,
		0.098193,
		0.149721,
		0.208175,
		0.271757,
		0.338772,
		0.407628,
		0.476846,
		0.545064,
		0.611048,
		0.673697,
		0.732056,
		0.785326,
		0.832877,
		0.874262,
		0.909232,
		0.937751,
		0.960006,
		0.97642,
		0.987648,
		0.994557,
		0.998182,
		0.999638,
		0.999981,
		1,
		0.99995,
		0.999417,
		0.997523,
		0.993181,
		0.985283,
		0.972832,
		0.95501,
		0.931219,
		0.901097,
		0.864511,
		0.821553,
		0.772523,
		0.717915,
		0.658402,
		0.594824,
		0.528173,
		0.459584,
		0.390324,
		0.321785,
		0.255475,
		0.193012,
		0.136118,
		0.086613,
		0.046414,
		0.017525,
		0.00204,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.000506,
		0.012092,
		0.037384,
		0.074325,
		0.120941,
		0.175346,
		0.23574,
		0.300419,
		0.367779,
		0.436317,
		0.504643,
		0.571482,
		0.635684,
		0.696234,
		0.752258,
		0.803041,
		0.848031,
		0.886861,
		0.919359,
		0.94556,
		0.965721,
		0.980326,
		0.990077,
		0.995871,
		0.99875,
		0.999798,
		0.999995,
		0.999999,
		0.999894,
		0.999142,
		0.996822,
		0.991857,
		0.983187,
		0.96987,
		0.951151,
		0.926493,
		0.895582,
		0.858326,
		0.814849,
		0.76547,
		0.710695,
		0.651201,
		0.587822,
		0.521539,
		0.453471,
		0.384859,
		0.317068,
		0.251572,
		0.189951,
		0.133884,
		0.085147,
		0.045605,
		0.017212,
		0.002002,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.004374,
		0.022747,
		0.053588,
		0.094935,
		0.144906,
		0.201706,
		0.263625,
		0.32905,
		0.39646,
		0.464441,
		0.531686,
		0.597004,
		0.659326,
		0.717717,
		0.771385,
		0.819691,
		0.862161,
		0.898503,
		0.928617,
		0.952607,
		0.970795,
		0.983719,
		0.992122,
		0.996928,
		0.999172,
		0.999898,
		0.999999,
		0.999995,
		0.999805,
		0.998793,
		0.996009,
		0.990397,
		0.980944,
		0.966766,
		0.94717,
		0.92167,
		0.890001,
		0.85211,
		0.808146,
		0.758449,
		0.703534,
		0.64408,
		0.580916,
		0.515011,
		0.447466,
		0.379501,
		0.31245,
		0.247755,
		0.186961,
		0.131705,
		0.083718,
		0.044818,
		0.016907,
		0.001966,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.000488,
		0.011667,
		0.036101,
		0.071843,
		0.11702,
		0.169842,
		0.228598,
		0.291667,
		0.357518,
		0.424716,
		0.491928,
		0.557927,
		0.621603,
		0.681965,
		0.738155,
		0.789457,
		0.835308,
		0.875306,
		0.909232,
		0.937054,
		0.958943,
		0.975277,
		0.986646,
		0.993829,
		0.997765,
		0.999475,
		0.999955,
		1,
		0.999983,
		0.999675,
		0.998361,
		0.995079,
		0.9888,
		0.978558,
		0.963527,
		0.943071,
		0.916757,
		0.884362,
		0.845868,
		0.80145,
		0.751464,
		0.696434,
		0.63704,
		0.574105,
		0.508587,
		0.441568,
		0.374245,
		0.307926,
		0.244022,
		0.184039,
		0.129577,
		0.082324,
		0.044051,
		0.01661,
		0.001931,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.004219,
		0.021964,
		0.051789,
		0.091835,
		0.140317,
		0.195527,
		0.255841,
		0.319719,
		0.385711,
		0.452462,
		0.518718,
		0.583329,
		0.645261,
		0.703599,
		0.757559,
		0.806498,
		0.849923,
		0.887506,
		0.919093,
		0.944719,
		0.964615,
		0.979216,
		0.989154,
		0.995238,
		0.998415,
		0.999686,
		0.999984,
		1,
		0.999956,
		0.999493,
		0.997839,
		0.994029,
		0.987065,
		0.976031,
		0.960158,
		0.938862,
		0.91176,
		0.87867,
		0.839605,
		0.794763,
		0.744517,
		0.689396,
		0.630081,
		0.567389,
		0.502265,
		0.435774,
		0.369091,
		0.303496,
		0.24037,
		0.181184,
		0.1275,
		0.080964,
		0.043303,
		0.016321,
		0.001896,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.000471,
		0.011263,
		0.034883,
		0.069482,
		0.113285,
		0.164588,
		0.221767,
		0.283276,
		0.347655,
		0.413531,
		0.479627,
		0.544762,
		0.607863,
		0.667967,
		0.724231,
		0.775945,
		0.822533,
		0.863572,
		0.898801,
		0.928131,
		0.951659,
		0.969673,
		0.982658,
		0.991285,
		0.996388,
		0.99891,
		0.999825,
		0.999996,
		0.999999,
		0.999908,
		0.999253,
		0.997222,
		0.992856,
		0.985193,
		0.973368,
		0.956664,
		0.934549,
		0.906686,
		0.872931,
		0.833326,
		0.788092,
		0.737611,
		0.682423,
		0.623205,
		0.560768,
		0.496045,
		0.430082,
		0.364035,
		0.299157,
		0.236797,
		0.178394,
		0.125472,
		0.079638,
		0.042574,
		0.016039,
		0.001863,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.004073,
		0.02122,
		0.050078,
		0.088884,
		0.135939,
		0.189623,
		0.248386,
		0.310761,
		0.375363,
		0.440894,
		0.506149,
		0.570021,
		0.631507,
		0.689715,
		0.743871,
		0.793332,
		0.837591,
		0.87629,
		0.909232,
		0.93639,
		0.95792,
		0.974162,
		0.985648,
		0.993081,
		0.997312,
		0.999278,
		0.999911,
		0.999999,
		0.999996,
		0.999831,
		0.998946,
		0.996504,
		0.991558,
		0.983187,
		0.970571,
		0.95305,
		0.930138,
		0.901541,
		0.867151,
		0.827036,
		0.781437,
		0.73075,
		0.675515,
		0.61641,
		0.55424,
		0.489923,
		0.424491,
		0.359076,
		0.294906,
		0.2333,
		0.175666,
		0.123492,
		0.078344,
		0.041863,
		0.015765,
		0.00183,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.000454,
		0.01088,
		0.033726,
		0.067236,
		0.109724,
		0.159571,
		0.21523,
		0.275228,
		0.338171,
		0.402746,
		0.467727,
		0.531979,
		0.594464,
		0.654248,
		0.710506,
		0.762531,
		0.809744,
		0.851704,
		0.888115,
		0.918841,
		0.943915,
		0.963547,
		0.978128,
		0.988228,
		0.99458,
		0.998044,
		0.999542,
		0.999961,
		1,
		0.999985,
		0.999716,
		0.998566,
		0.995681,
		0.990135,
		0.981047,
		0.967646,
		0.949321,
		0.925634,
		0.896331,
		0.861335,
		0.820741,
		0.774805,
		0.723934,
		0.668673,
		0.609698,
		0.547804,
		0.483901,
		0.418999,
		0.354211,
		0.290741,
		0.229879,
		0.173,
		0.121558,
		0.077081,
		0.04117,
		0.015497,
		0.001798,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.003934,
		0.020513,
		0.048451,
		0.086071,
		0.131761,
		0.183976,
		0.241242,
		0.302157,
		0.365398,
		0.429721,
		0.493969,
		0.557076,
		0.618068,
		0.676077,
		0.730344,
		0.780225,
		0.825205,
		0.864902,
		0.899083,
		0.927669,
		0.950749,
		0.968585,
		0.981612,
		0.990437,
		0.995817,
		0.998613,
		0.999726,
		0.999986,
		1,
		0.999962,
		0.999557,
		0.998106,
		0.994749,
		0.988586,
		0.978775,
		0.964597,
		0.945484,
		0.921045,
		0.891062,
		0.855488,
		0.814442,
		0.768196,
		0.717165,
		0.661898,
		0.603067,
		0.541461,
		0.477974,
		0.413603,
		0.349439,
		0.28666,
		0.22653,
		0.170393,
		0.119668,
		0.075848,
		0.040494,
		0.015236,
		0.001767,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.000439,
		0.010517,
		0.032625,
		0.065096,
		0.106327,
		0.154777,
		0.208972,
		0.267506,
		0.32905,
		0.392345,
		0.456215,
		0.51957,
		0.581406,
		0.640815,
		0.696992,
		0.749239,
		0.796974,
		0.839741,
		0.877219,
		0.909232,
		0.935757,
		0.956935,
		0.973075,
		0.984657,
		0.992316,
		0.996826,
		0.999046,
		0.999847,
		0.999996,
		1,
		0.99992,
		0.999346,
		0.997561,
		0.993706,
		0.986911,
		0.976376,
		0.961427,
		0.941544,
		0.916375,
		0.885739,
		0.849616,
		0.808146,
		0.761615,
		0.710447,
		0.655192,
		0.596519,
		0.535208,
		0.472143,
		0.408303,
		0.344756,
		0.282661,
		0.223252,
		0.167843,
		0.117821,
		0.074644,
		0.039835,
		0.014982,
		0.001737,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.003803,
		0.019841,
		0.046901,
		0.083389,
		0.127771,
		0.178574,
		0.234393,
		0.29389,
		0.355799,
		0.418929,
		0.482168,
		0.544486,
		0.604944,
		0.662695,
		0.716994,
		0.767202,
		0.812798,
		0.853382,
		0.88869,
		0.918602,
		0.943146,
		0.962515,
		0.977061,
		0.987301,
		0.993899,
		0.997637,
		0.999367,
		0.999923,
		1,
		0.999996,
		0.999852,
		0.999076,
		0.996925,
		0.992549,
		0.985111,
		0.973851,
		0.958141,
		0.937506,
		0.911631,
		0.880367,
		0.843722,
		0.801855,
		0.755064,
		0.70378,
		0.648554,
		0.590052,
		0.529046,
		0.466406,
		0.403095,
		0.340162,
		0.278741,
		0.220042,
		0.165349,
		0.116017,
		0.073468,
		0.039191,
		0.014734,
		0.001708,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.000424,
		0.010171,
		0.031577,
		0.063057,
		0.103084,
		0.150192,
		0.202976,
		0.260094,
		0.320274,
		0.382312,
		0.44508,
		0.507526,
		0.568684,
		0.627672,
		0.683702,
		0.736089,
		0.78425,
		0.827721,
		0.866157,
		0.89935,
		0.927229,
		0.949876,
		0.967529,
		0.980581,
		0.989583,
		0.995219,
		0.99828,
		0.999599,
		0.999966,
		1,
		0.999987,
		0.999752,
		0.998741,
		0.996194,
		0.991278,
		0.983187,
		0.971204,
		0.954745,
		0.933375,
		0.906817,
		0.874952,
		0.837812,
		0.795573,
		0.748546,
		0.697165,
		0.641985,
		0.583667,
		0.522972,
		0.460761,
		0.397978,
		0.335654,
		0.2749,
		0.2169,
		0.16291,
		0.114253,
		0.07232,
		0.038563,
		0.014492,
		0.001679,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.003677,
		0.019201,
		0.045425,
		0.08083,
		0.123957,
		0.173402,
		0.227824,
		0.285944,
		0.346552,
		0.408505,
		0.470734,
		0.532247,
		0.592135,
		0.649575,
		0.703836,
		0.754286,
		0.800401,
		0.841768,
		0.878098,
		0.909232,
		0.935153,
		0.955987,
		0.972017,
		0.983674,
		0.991538,
		0.99631,
		0.99878,
		0.99976,
		0.999988,
		1,
		0.999967,
		0.999612,
		0.998334,
		0.995364,
		0.989891,
		0.98114,
		0.968439,
		0.951243,
		0.929157,
		0.901939,
		0.869498,
		0.831889,
		0.789303,
		0.742062,
		0.690605,
		0.635486,
		0.577362,
		0.516987,
		0.455206,
		0.39295,
		0.331231,
		0.271135,
		0.213823,
		0.160523,
		0.112529,
		0.071199,
		0.03795,
		0.014256,
		0.001651,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.00041,
		0.009842,
		0.030578,
		0.061111,
		0.099986,
		0.145806,
		0.197229,
		0.252975,
		0.311827,
		0.372632,
		0.434306,
		0.495839,
		0.556295,
		0.61482,
		0.670646,
		0.723099,
		0.771599,
		0.815675,
		0.854966,
		0.889236,
		0.918374,
		0.94241,
		0.961517,
		0.976017,
		0.986377,
		0.993199,
		0.997199,
		0.999161,
		0.999866,
		0.999997,
		1,
		0.99993,
		0.999426,
		0.99785,
		0.994434,
		0.988388,
		0.978974,
		0.965558,
		0.947639,
		0.924857,
		0.897002,
		0.864011,
		0.825958,
		0.783049,
		0.735615,
		0.6841,
		0.629057,
		0.571138,
		0.511089,
		0.449741,
		0.38801,
		0.326889,
		0.267443,
		0.210809,
		0.158187,
		0.110843,
		0.070103,
		0.037351,
		0.014026,
		0.001624,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.003558,
		0.018592,
		0.044016,
		0.078386,
		0.120309,
		0.168448,
		0.22152,
		0.278304,
		0.33764,
		0.398433,
		0.459655,
		0.520351,
		0.57964,
		0.636722,
		0.690882,
		0.741497,
		0.788042,
		0.830094,
		0.867344,
		0.899603,
		0.926809,
		0.949038,
		0.966504,
		0.979567,
		0.988724,
		0.994599,
		0.997915,
		0.999443,
		0.999932,
		1,
		0.999997,
		0.99987,
		0.999187,
		0.997284,
		0.9934,
		0.986769,
		0.976689,
		0.962567,
		0.943938,
		0.92048,
		0.892012,
		0.858494,
		0.820021,
		0.776812,
		0.729206,
		0.677651,
		0.622699,
		0.564995,
		0.505276,
		0.444363,
		0.383156,
		0.322628,
		0.263824,
		0.207857,
		0.155901,
		0.109195,
		0.069032,
		0.036766,
		0.013802,
		0.001597,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.000397,
		0.009529,
		0.029626,
		0.059254,
		0.097025,
		0.141606,
		0.191718,
		0.246136,
		0.303696,
		0.363292,
		0.423884,
		0.484498,
		0.544233,
		0.60226,
		0.657831,
		0.710282,
		0.759042,
		0.803632,
		0.843682,
		0.878929,
		0.909232,
		0.934575,
		0.955074,
		0.970985,
		0.982702,
		0.990749,
		0.995768,
		0.998482,
		0.999647,
		0.99997,
		1,
		0.999989,
		0.999782,
		0.998889,
		0.996633,
		0.992261,
		0.985035,
		0.974289,
		0.959469,
		0.940146,
		0.916031,
		0.886971,
		0.852952,
		0.814083,
		0.770596,
		0.722839,
		0.67126,
		0.61641,
		0.55893,
		0.499548,
		0.439072,
		0.378386,
		0.318445,
		0.260276,
		0.204965,
		0.153664,
		0.107582,
		0.067986,
		0.036195,
		0.013582,
		0.001571,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.003444,
		0.018011,
		0.042672,
		0.076051,
		0.116819,
		0.163699,
		0.215467,
		0.270954,
		0.32905,
		0.388702,
		0.448922,
		0.50879,
		0.567455,
		0.624138,
		0.678141,
		0.728851,
		0.775744,
		0.81839,
		0.856465,
		0.889753,
		0.918157,
		0.941704,
		0.960553,
		0.974996,
		0.985457,
		0.992484,
		0.996733,
		0.998923,
		0.999788,
		0.999989,
		1,
		0.999971,
		0.999658,
		0.998528,
		0.995893,
		0.991015,
		0.983187,
		0.971777,
		0.956268,
		0.936266,
		0.911514,
		0.881887,
		0.847388,
		0.808146,
		0.764404,
		0.716513,
		0.664927,
		0.610192,
		0.552945,
		0.493904,
		0.433865,
		0.373698,
		0.31434,
		0.256796,
		0.202132,
		0.151474,
		0.106005,
		0.066963,
		0.035637,
		0.013368,
		0.001546,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.000384,
		0.00923,
		0.028718,
		0.057479,
		0.094192,
		0.137583,
		0.18643,
		0.239562,
		0.295864,
		0.354276,
		0.413799,
		0.473495,
		0.532493,
		0.589991,
		0.645261,
		0.697651,
		0.746597,
		0.791619,
		0.832336,
		0.868468,
		0.899843,
		0.926409,
		0.948233,
		0.965511,
		0.97857,
		0.987863,
		0.993959,
		0.99752,
		0.999259,
		0.999882,
		0.999997,
		1,
		0.999938,
		0.999493,
		0.998097,
		0.99506,
		0.989663,
		0.981226,
		0.969156,
		0.952968,
		0.932304,
		0.906935,
		0.876762,
		0.841807,
		0.802214,
		0.758237,
		0.710231,
		0.658652,
		0.604044,
		0.547038,
		0.488343,
		0.428742,
		0.369091,
		0.310309,
		0.253383,
		0.199356,
		0.14933,
		0.104462,
		0.065963,
		0.035092,
		0.013159,
		0.001521,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.003336,
		0.017457,
		0.041389,
		0.073818,
		0.113477,
		0.159146,
		0.209654,
		0.263882,
		0.320766,
		0.379296,
		0.438523,
		0.497557,
		0.555576,
		0.611825,
		0.665622,
		0.716363,
		0.763528,
		0.806683,
		0.845492,
		0.879718,
		0.909232,
		0.934022,
		0.954194,
		0.969981,
		0.981742,
		0.989954,
		0.995203,
		0.998153,
		0.999508,
		0.99994,
		1,
		0.999997,
		0.999886,
		0.999281,
		0.997593,
		0.994133,
		0.988203,
		0.979155,
		0.966428,
		0.949573,
		0.928264,
		0.902298,
		0.8716,
		0.836212,
		0.796291,
		0.752097,
		0.703994,
		0.652437,
		0.597967,
		0.541209,
		0.482863,
		0.423701,
		0.364563,
		0.306352,
		0.250035,
		0.196635,
		0.14723,
		0.102952,
		0.064984,
		0.034559,
		0.012955,
		0.001497,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.000372,
		0.008946,
		0.027851,
		0.055783,
		0.09148,
		0.133727,
		0.181354,
		0.233241,
		0.288319,
		0.345573,
		0.40404,
		0.46282,
		0.52107,
		0.578013,
		0.63294,
		0.685216,
		0.73428,
		0.779657,
		0.820956,
		0.857883,
		0.890244,
		0.91795,
		0.941027,
		0.959621,
		0.973997,
		0.984543,
		0.991758,
		0.99624,
		0.998655,
		0.999688,
		0.999973,
		1,
		0.99999,
		0.999807,
		0.999017,
		0.997011,
		0.993109,
		0.986637,
		0.976975,
		0.963598,
		0.946088,
		0.924149,
		0.897607,
		0.866407,
		0.830607,
		0.790377,
		0.745987,
		0.697803,
		0.646281,
		0.59196,
		0.535457,
		0.477463,
		0.41874,
		0.360112,
		0.302467,
		0.246752,
		0.193969,
		0.145174,
		0.101474,
		0.064028,
		0.034038,
		0.012755,
		0.001474,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.003233,
		0.016928,
		0.040162,
		0.071682,
		0.110275,
		0.154777,
		0.204067,
		0.257074,
		0.312776,
		0.370205,
		0.428447,
		0.486644,
		0.544,
		0.599783,
		0.653328,
		0.704043,
		0.751412,
		0.794999,
		0.834457,
		0.869533,
		0.900072,
		0.926026,
		0.947458,
		0.964547,
		0.977591,
		0.987004,
		0.993303,
		0.997097,
		0.999046,
		0.999813,
		0.99999,
		1,
		0.999974,
		0.999698,
		0.998695,
		0.996348,
		0.991988,
		0.984964,
		0.974689,
		0.960668,
		0.942516,
		0.919966,
		0.892867,
		0.861185,
		0.824995,
		0.784477,
		0.739908,
		0.691659,
		0.640185,
		0.586022,
		0.529781,
		0.472143,
		0.413858,
		0.355737,
		0.298652,
		0.24353,
		0.191355,
		0.14316,
		0.100028,
		0.063092,
		0.033529,
		0.01256,
		0.001451,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.000361,
		0.008674,
		0.027023,
		0.054161,
		0.088884,
		0.130029,
		0.176478,
		0.227159,
		0.281048,
		0.337168,
		0.394596,
		0.452462,
		0.509955,
		0.566321,
		0.620871,
		0.672983,
		0.722107,
		0.767766,
		0.809569,
		0.847207,
		0.880467,
		0.909232,
		0.933492,
		0.953345,
		0.969004,
		0.980795,
		0.989154,
		0.994617,
		0.997796,
		0.999343,
		0.999896,
		0.999998,
		1,
		0.999945,
		0.999551,
		0.99831,
		0.995601,
		0.990768,
		0.983187,
		0.972299,
		0.957642,
		0.938862,
		0.915718,
		0.888082,
		0.855939,
		0.819379,
		0.778592,
		0.733863,
		0.685564,
		0.63415,
		0.580154,
		0.52418,
		0.466901,
		0.409054,
		0.351437,
		0.294906,
		0.24037,
		0.188793,
		0.141187,
		0.098612,
		0.062177,
		0.033031,
		0.01237,
		0.001428,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.000807,
		0.010445,
		0.029959,
		0.058111,
		0.093699,
		0.135567,
		0.182603,
		0.233741,
		0.28796,
		0.344293,
		0.401822,
		0.459685,
		0.517077,
		0.573252,
		0.62753,
		0.679298,
		0.728012,
		0.77321,
		0.814507,
		0.851607,
		0.88431,
		0.91251,
		0.936212,
		0.955526,
		0.970681,
		0.982017,
		0.989983,
		0.995126,
		0.998065,
		0.999454,
		0.999923,
		0.999999,
		0.999999,
		0.999923,
		0.999454,
		0.998065,
		0.995126,
		0.989983,
		0.982017,
		0.970681,
		0.955526,
		0.936212,
		0.91251,
		0.88431,
		0.851607,
		0.814507,
		0.77321,
		0.728012,
		0.679298,
		0.62753,
		0.573252,
		0.517077,
		0.459685,
		0.401822,
		0.344293,
		0.28796,
		0.233741,
		0.182603,
		0.135567,
		0.093699,
		0.058111,
		0.029959,
		0.010445,
		0.000807,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		9.2e-005,
		0.00717,
		0.024591,
		0.051077,
		0.085389,
		0.126332,
		0.172758,
		0.223565,
		0.2777,
		0.33416,
		0.391998,
		0.45032,
		0.508291,
		0.56514,
		0.620158,
		0.672708,
		0.722224,
		0.76822,
		0.810295,
		0.848135,
		0.881522,
		0.910341,
		0.934586,
		0.954362,
		0.969893,
		0.981521,
		0.9897,
		0.994985,
		0.998008,
		0.999438,
		0.999921,
		0.999999,
		0.999999,
		0.999921,
		0.999438,
		0.998008,
		0.994985,
		0.9897,
		0.981521,
		0.969893,
		0.954362,
		0.934586,
		0.910341,
		0.881522,
		0.848135,
		0.810295,
		0.76822,
		0.722224,
		0.672708,
		0.620158,
		0.56514,
		0.508291,
		0.45032,
		0.391998,
		0.33416,
		0.2777,
		0.223565,
		0.172758,
		0.126332,
		0.085389,
		0.051077,
		0.024591,
		0.00717,
		9.2e-005,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.004448,
		0.019635,
		0.044332,
		0.077262,
		0.117192,
		0.162935,
		0.213353,
		0.267356,
		0.32391,
		0.382031,
		0.440796,
		0.499339,
		0.556859,
		0.612621,
		0.66596,
		0.71629,
		0.7631,
		0.805968,
		0.844563,
		0.878652,
		0.908106,
		0.932909,
		0.95316,
		0.969079,
		0.981008,
		0.989406,
		0.994838,
		0.997948,
		0.999421,
		0.999919,
		0.999999,
		0.999999,
		0.999919,
		0.999421,
		0.997948,
		0.994838,
		0.989406,
		0.981008,
		0.969079,
		0.95316,
		0.932909,
		0.908106,
		0.878652,
		0.844563,
		0.805968,
		0.7631,
		0.71629,
		0.66596,
		0.612621,
		0.556859,
		0.499339,
		0.440796,
		0.382031,
		0.32391,
		0.267356,
		0.213353,
		0.162935,
		0.117192,
		0.077262,
		0.044332,
		0.019635,
		0.004448,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		0.002328,
		0.015128,
		0.037906,
		0.069344,
		0.108167,
		0.153149,
		0.203114,
		0.256937,
		0.313545,
		0.371923,
		0.431113,
		0.490217,
		0.548406,
		0.604914,
		0.659052,
		0.710206,
		0.757843,
		0.801521,
		0.840888,
		0.875696,
		0.905802,
		0.931179,
		0.951918,
		0.968237,
		0.980477,
		0.989102,
		0.994686,
		0.997886,
		0.999403,
		0.999916,
		0.999999,
		0.999999,
		0.999916,
		0.999403,
		0.997886,
		0.994686,
		0.989102,
		0.980477,
		0.968237,
		0.951918,
		0.931179,
		0.905802,
		0.875696,
		0.840888,
		0.801521,
		0.757843,
		0.710206,
		0.659052,
		0.604914,
		0.548406,
		0.490217,
		0.431113,
		0.371923,
		0.313545,
		0.256937,
		0.203114,
		0.153149,
		0.108167,
		0.069344,
		0.037906,
		0.015128,
		0.002328,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.00086,
		0.011113,
		0.031834,
		0.06166,
		0.099276,
		0.143415,
		0.19286,
		0.246449,
		0.303071,
		0.361676,
		0.42127,
		0.480925,
		0.539778,
		0.597036,
		0.651979,
		0.703969,
		0.752448,
		0.796951,
		0.837108,
		0.872652,
		0.903427,
		0.929393,
		0.950636,
		0.967366,
		0.979927,
		0.988787,
		0.994529,
		0.997822,
		0.999384,
		0.999913,
		0.999999,
		0.999999,
		0.999913,
		0.999384,
		0.997822,
		0.994529,
		0.988787,
		0.979927,
		0.967366,
		0.950636,
		0.929393,
		0.903427,
		0.872652,
		0.837108,
		0.796951,
		0.752448,
		0.703969,
		0.651979,
		0.597036,
		0.539778,
		0.480925,
		0.42127,
		0.361676,
		0.303071,
		0.246449,
		0.19286,
		0.143415,
		0.099276,
		0.06166,
		0.031834,
		0.011113,
		0.00086,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		9.8e-005,
		0.007634,
		0.026151,
		0.05424,
		0.090543,
		0.13375,
		0.182603,
		0.2359,
		0.292492,
		0.351291,
		0.411269,
		0.471461,
		0.530974,
		0.588982,
		0.644738,
		0.697574,
		0.746909,
		0.792253,
		0.833218,
		0.869516,
		0.900978,
		0.927551,
		0.949311,
		0.966466,
		0.979358,
		0.988461,
		0.994366,
		0.997755,
		0.999365,
		0.999911,
		0.999999,
		0.999999,
		0.999911,
		0.999365,
		0.997755,
		0.994366,
		0.988461,
		0.979358,
		0.966466,
		0.949311,
		0.927551,
		0.900978,
		0.869516,
		0.833218,
		0.792253,
		0.746909,
		0.697574,
		0.644738,
		0.588982,
		0.530974,
		0.471461,
		0.411269,
		0.351291,
		0.292492,
		0.2359,
		0.182603,
		0.13375,
		0.090543,
		0.05424,
		0.026151,
		0.007634,
		9.8e-005,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.004741,
		0.020897,
		0.047116,
		0.081992,
		0.124172,
		0.172357,
		0.2253,
		0.281814,
		0.340772,
		0.401108,
		0.461824,
		0.52199,
		0.580749,
		0.637323,
		0.691017,
		0.741222,
		0.787425,
		0.829215,
		0.866286,
		0.898452,
		0.925648,
		0.947941,
		0.965534,
		0.978769,
		0.988123,
		0.994196,
		0.997686,
		0.999345,
		0.999908,
		0.999999,
		0.999999,
		0.999908,
		0.999345,
		0.997686,
		0.994196,
		0.988123,
		0.978769,
		0.965534,
		0.947941,
		0.925648,
		0.898452,
		0.866286,
		0.829215,
		0.787425,
		0.741222,
		0.691017,
		0.637323,
		0.580749,
		0.52199,
		0.461824,
		0.401108,
		0.340772,
		0.281814,
		0.2253,
		0.172357,
		0.124172,
		0.081992,
		0.047116,
		0.020897,
		0.004741,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.002483,
		0.016114,
		0.04032,
		0.073649,
		0.114702,
		0.162135,
		0.214658,
		0.271043,
		0.330122,
		0.39079,
		0.452013,
		0.512824,
		0.572334,
		0.629732,
		0.684294,
		0.735383,
		0.782462,
		0.825095,
		0.862958,
		0.895847,
		0.923684,
		0.946526,
		0.964571,
		0.978159,
		0.987773,
		0.99402,
		0.997614,
		0.999324,
		0.999905,
		0.999999,
		0.999999,
		0.999905,
		0.999324,
		0.997614,
		0.99402,
		0.987773,
		0.978159,
		0.964571,
		0.946526,
		0.923684,
		0.895847,
		0.862958,
		0.825095,
		0.782462,
		0.735383,
		0.684294,
		0.629732,
		0.572334,
		0.512824,
		0.452013,
		0.39079,
		0.330122,
		0.271043,
		0.214658,
		0.162135,
		0.114702,
		0.073649,
		0.04032,
		0.016114,
		0.002483,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000918,
		0.011847,
		0.033889,
		0.065543,
		0.105362,
		0.151954,
		0.203986,
		0.260186,
		0.319344,
		0.380315,
		0.442026,
		0.503473,
		0.563732,
		0.62196,
		0.6774,
		0.729388,
		0.777359,
		0.820854,
		0.859528,
		0.893159,
		0.921655,
		0.945063,
		0.963574,
		0.977527,
		0.987409,
		0.993838,
		0.99754,
		0.999302,
		0.999902,
		0.999999,
		0.999999,
		0.999902,
		0.999302,
		0.99754,
		0.993838,
		0.987409,
		0.977527,
		0.963574,
		0.945063,
		0.921655,
		0.893159,
		0.859528,
		0.820854,
		0.777359,
		0.729388,
		0.6774,
		0.62196,
		0.563732,
		0.503473,
		0.442026,
		0.380315,
		0.319344,
		0.260186,
		0.203986,
		0.151954,
		0.105362,
		0.065543,
		0.033889,
		0.011847,
		0.000918,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.000105,
		0.008146,
		0.027864,
		0.057705,
		0.096175,
		0.141832,
		0.193297,
		0.249252,
		0.308444,
		0.369686,
		0.431863,
		0.493936,
		0.554942,
		0.614003,
		0.670332,
		0.723232,
		0.772112,
		0.816488,
		0.855993,
		0.890386,
		0.91956,
		0.94355,
		0.962541,
		0.976872,
		0.987032,
		0.993648,
		0.997462,
		0.99928,
		0.999898,
		0.999999,
		0.999999,
		0.999898,
		0.99928,
		0.997462,
		0.993648,
		0.987032,
		0.976872,
		0.962541,
		0.94355,
		0.91956,
		0.890386,
		0.855993,
		0.816488,
		0.772112,
		0.723232,
		0.670332,
		0.614003,
		0.554942,
		0.493936,
		0.431863,
		0.369686,
		0.308444,
		0.249252,
		0.193297,
		0.141832,
		0.096175,
		0.057705,
		0.027864,
		0.008146,
		0.000105,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.005063,
		0.022285,
		0.050169,
		0.087167,
		0.131787,
		0.182603,
		0.238249,
		0.297425,
		0.358903,
		0.421525,
		0.48421,
		0.545959,
		0.605858,
		0.663083,
		0.71691,
		0.766717,
		0.811993,
		0.85235,
		0.887524,
		0.917395,
		0.941985,
		0.961473,
		0.976194,
		0.986641,
		0.993452,
		0.997381,
		0.999256,
		0.999895,
		0.999999,
		0.999999,
		0.999895,
		0.999256,
		0.997381,
		0.993452,
		0.986641,
		0.976194,
		0.961473,
		0.941985,
		0.917395,
		0.887524,
		0.85235,
		0.811993,
		0.766717,
		0.71691,
		0.663083,
		0.605858,
		0.545959,
		0.48421,
		0.421525,
		0.358903,
		0.297425,
		0.238249,
		0.182603,
		0.131787,
		0.087167,
		0.050169,
		0.022285,
		0.005063,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.002655,
		0.0172,
		0.042971,
		0.078366,
		0.121842,
		0.171921,
		0.227188,
		0.286296,
		0.34797,
		0.41101,
		0.474293,
		0.53678,
		0.597518,
		0.655651,
		0.710418,
		0.761169,
		0.807365,
		0.848593,
		0.884571,
		0.915158,
		0.940367,
		0.960366,
		0.97549,
		0.986235,
		0.993247,
		0.997298,
		0.999232,
		0.999891,
		0.999999,
		0.999999,
		0.999891,
		0.999232,
		0.997298,
		0.993247,
		0.986235,
		0.97549,
		0.960366,
		0.940367,
		0.915158,
		0.884571,
		0.848593,
		0.807365,
		0.761169,
		0.710418,
		0.655651,
		0.597518,
		0.53678,
		0.474293,
		0.41101,
		0.34797,
		0.286296,
		0.227188,
		0.171921,
		0.121842,
		0.078366,
		0.042971,
		0.0172,
		0.002655,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000982,
		0.012657,
		0.03615,
		0.069803,
		0.112019,
		0.161266,
		0.216079,
		0.275061,
		0.336891,
		0.40032,
		0.464184,
		0.527402,
		0.588982,
		0.648029,
		0.703751,
		0.755463,
		0.802599,
		0.84472,
		0.881522,
		0.912847,
		0.938692,
		0.95922,
		0.974761,
		0.985814,
		0.993035,
		0.99721,
		0.999206,
		0.999888,
		0.999998,
		0.999998,
		0.999888,
		0.999206,
		0.99721,
		0.993035,
		0.985814,
		0.974761,
		0.95922,
		0.938692,
		0.912847,
		0.881522,
		0.84472,
		0.802599,
		0.755463,
		0.703751,
		0.648029,
		0.588982,
		0.527402,
		0.464184,
		0.40032,
		0.336891,
		0.275061,
		0.216079,
		0.161266,
		0.112019,
		0.069803,
		0.03615,
		0.012657,
		0.000982,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.000112,
		0.008711,
		0.02975,
		0.061512,
		0.102344,
		0.150658,
		0.204936,
		0.26373,
		0.325668,
		0.389456,
		0.453882,
		0.517822,
		0.580244,
		0.640214,
		0.696903,
		0.749594,
		0.79769,
		0.840726,
		0.878374,
		0.910457,
		0.936959,
		0.958032,
		0.974004,
		0.985376,
		0.992814,
		0.997119,
		0.99918,
		0.999884,
		0.999998,
		0.999998,
		0.999884,
		0.99918,
		0.997119,
		0.992814,
		0.985376,
		0.974004,
		0.958032,
		0.936959,
		0.910457,
		0.878374,
		0.840726,
		0.79769,
		0.749594,
		0.696903,
		0.640214,
		0.580244,
		0.517822,
		0.453882,
		0.389456,
		0.325668,
		0.26373,
		0.204936,
		0.150658,
		0.102344,
		0.061512,
		0.02975,
		0.008711,
		0.000112,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.005419,
		0.023816,
		0.053528,
		0.092843,
		0.140116,
		0.193772,
		0.252311,
		0.314308,
		0.378419,
		0.443386,
		0.508038,
		0.571301,
		0.632201,
		0.68987,
		0.743557,
		0.792634,
		0.836606,
		0.875123,
		0.907987,
		0.935165,
		0.956802,
		0.973219,
		0.984922,
		0.992584,
		0.997025,
		0.999152,
		0.99988,
		0.999998,
		0.999998,
		0.99988,
		0.999152,
		0.997025,
		0.992584,
		0.984922,
		0.973219,
		0.956802,
		0.935165,
		0.907987,
		0.875123,
		0.836606,
		0.792634,
		0.743557,
		0.68987,
		0.632201,
		0.571301,
		0.508038,
		0.443386,
		0.378419,
		0.314308,
		0.252311,
		0.193772,
		0.140116,
		0.092843,
		0.053528,
		0.023816,
		0.005419,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.002844,
		0.018399,
		0.045892,
		0.083548,
		0.129662,
		0.182603,
		0.240813,
		0.302814,
		0.367211,
		0.432694,
		0.498046,
		0.562148,
		0.623984,
		0.682646,
		0.737347,
		0.787425,
		0.832357,
		0.871765,
		0.905432,
		0.933308,
		0.955526,
		0.972405,
		0.98445,
		0.992346,
		0.996926,
		0.999123,
		0.999876,
		0.999998,
		0.999998,
		0.999876,
		0.999123,
		0.996926,
		0.992346,
		0.98445,
		0.972405,
		0.955526,
		0.933308,
		0.905432,
		0.871765,
		0.832357,
		0.787425,
		0.737347,
		0.682646,
		0.623984,
		0.562148,
		0.498046,
		0.432694,
		0.367211,
		0.302814,
		0.240813,
		0.182603,
		0.129662,
		0.083548,
		0.045892,
		0.018399,
		0.002844,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.001053,
		0.013553,
		0.038645,
		0.07449,
		0.119322,
		0.171447,
		0.229249,
		0.291195,
		0.355834,
		0.421806,
		0.487844,
		0.552782,
		0.615558,
		0.675225,
		0.730958,
		0.782059,
		0.827973,
		0.868297,
		0.90279,
		0.931385,
		0.954204,
		0.971559,
		0.983959,
		0.992097,
		0.996824,
		0.999093,
		0.999871,
		0.999998,
		0.999998,
		0.999871,
		0.999093,
		0.996824,
		0.992097,
		0.983959,
		0.971559,
		0.954204,
		0.931385,
		0.90279,
		0.868297,
		0.827973,
		0.782059,
		0.730958,
		0.675225,
		0.615558,
		0.552782,
		0.487844,
		0.421806,
		0.355834,
		0.291195,
		0.229249,
		0.171447,
		0.119322,
		0.07449,
		0.038645,
		0.013553,
		0.001053,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.00012,
		0.009337,
		0.031834,
		0.065706,
		0.109121,
		0.160321,
		0.217631,
		0.279458,
		0.344293,
		0.410723,
		0.47743,
		0.543198,
		0.606919,
		0.667604,
		0.724384,
		0.776529,
		0.82345,
		0.864714,
		0.900056,
		0.929393,
		0.952833,
		0.970681,
		0.983449,
		0.991839,
		0.996717,
		0.999062,
		0.999867,
		0.999998,
		0.999998,
		0.999867,
		0.999062,
		0.996717,
		0.991839,
		0.983449,
		0.970681,
		0.952833,
		0.929393,
		0.900056,
		0.864714,
		0.82345,
		0.776529,
		0.724384,
		0.667604,
		0.606919,
		0.543198,
		0.47743,
		0.410723,
		0.344293,
		0.279458,
		0.217631,
		0.160321,
		0.109121,
		0.065706,
		0.031834,
		0.009337,
		0.00012,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.005814,
		0.02551,
		0.057235,
		0.099088,
		0.149248,
		0.205974,
		0.26761,
		0.332591,
		0.399446,
		0.466802,
		0.533393,
		0.598062,
		0.659775,
		0.717621,
		0.770831,
		0.818782,
		0.861011,
		0.897228,
		0.92733,
		0.95141,
		0.96977,
		0.982919,
		0.991569,
		0.996606,
		0.999029,
		0.999862,
		0.999998,
		0.999998,
		0.999862,
		0.999029,
		0.996606,
		0.991569,
		0.982919,
		0.96977,
		0.95141,
		0.92733,
		0.897228,
		0.861011,
		0.818782,
		0.770831,
		0.717621,
		0.659775,
		0.598062,
		0.533393,
		0.466802,
		0.399446,
		0.332591,
		0.26761,
		0.205974,
		0.149248,
		0.099088,
		0.057235,
		0.02551,
		0.005814,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.003055,
		0.019728,
		0.049119,
		0.089257,
		0.13825,
		0.194292,
		0.255662,
		0.320733,
		0.387975,
		0.455958,
		0.523363,
		0.588982,
		0.651733,
		0.710662,
		0.764959,
		0.813964,
		0.857183,
		0.894301,
		0.925192,
		0.949934,
		0.968823,
		0.982367,
		0.991289,
		0.99649,
		0.998995,
		0.999857,
		0.999998,
		0.999998,
		0.999857,
		0.998995,
		0.99649,
		0.991289,
		0.982367,
		0.968823,
		0.949934,
		0.925192,
		0.894301,
		0.857183,
		0.813964,
		0.764959,
		0.710662,
		0.651733,
		0.588982,
		0.523363,
		0.455958,
		0.387975,
		0.320733,
		0.255662,
		0.194292,
		0.13825,
		0.089257,
		0.049119,
		0.019728,
		0.003055,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.001133,
		0.014547,
		0.041405,
		0.079662,
		0.127354,
		0.182603,
		0.243624,
		0.308725,
		0.376311,
		0.444897,
		0.513104,
		0.579674,
		0.643472,
		0.7035,
		0.758906,
		0.80899,
		0.853227,
		0.891271,
		0.922975,
		0.948403,
		0.967838,
		0.981794,
		0.990997,
		0.996368,
		0.99896,
		0.999852,
		0.999998,
		0.999998,
		0.999852,
		0.99896,
		0.996368,
		0.990997,
		0.981794,
		0.967838,
		0.948403,
		0.922975,
		0.891271,
		0.853227,
		0.80899,
		0.758906,
		0.7035,
		0.643472,
		0.579674,
		0.513104,
		0.444897,
		0.376311,
		0.308725,
		0.243624,
		0.182603,
		0.127354,
		0.079662,
		0.041405,
		0.014547,
		0.001133,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.00013,
		0.010032,
		0.034144,
		0.070342,
		0.116587,
		0.170928,
		0.231511,
		0.296573,
		0.364459,
		0.433618,
		0.502614,
		0.570132,
		0.634987,
		0.696131,
		0.752666,
		0.803856,
		0.849137,
		0.888134,
		0.920678,
		0.946813,
		0.966816,
		0.981196,
		0.990693,
		0.996242,
		0.998922,
		0.999846,
		0.999998,
		0.999998,
		0.999846,
		0.998922,
		0.996242,
		0.990693,
		0.981196,
		0.966816,
		0.946813,
		0.920678,
		0.888134,
		0.849137,
		0.803856,
		0.752666,
		0.696131,
		0.634987,
		0.570132,
		0.502614,
		0.433618,
		0.364459,
		0.296573,
		0.231511,
		0.170928,
		0.116587,
		0.070342,
		0.034144,
		0.010032,
		0.00013,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.006254,
		0.027391,
		0.061339,
		0.10598,
		0.159289,
		0.219335,
		0.284287,
		0.35242,
		0.42212,
		0.491888,
		0.560352,
		0.626271,
		0.688547,
		0.746234,
		0.798554,
		0.844907,
		0.884886,
		0.918295,
		0.945162,
		0.965752,
		0.980575,
		0.990375,
		0.99611,
		0.998883,
		0.999841,
		0.999998,
		0.999998,
		0.999841,
		0.998883,
		0.99611,
		0.990375,
		0.980575,
		0.965752,
		0.945162,
		0.918295,
		0.884886,
		0.844907,
		0.798554,
		0.746234,
		0.688547,
		0.626271,
		0.560352,
		0.491888,
		0.42212,
		0.35242,
		0.284287,
		0.219335,
		0.159289,
		0.10598,
		0.061339,
		0.027391,
		0.006254,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.003289,
		0.021206,
		0.052699,
		0.095568,
		0.147708,
		0.207112,
		0.271875,
		0.3402,
		0.410404,
		0.480925,
		0.55033,
		0.617319,
		0.680741,
		0.739603,
		0.79308,
		0.840533,
		0.881522,
		0.915824,
		0.943447,
		0.964646,
		0.979927,
		0.990045,
		0.995972,
		0.998843,
		0.999835,
		0.999998,
		0.999998,
		0.999835,
		0.998843,
		0.995972,
		0.990045,
		0.979927,
		0.964646,
		0.943447,
		0.915824,
		0.881522,
		0.840533,
		0.79308,
		0.739603,
		0.680741,
		0.617319,
		0.55033,
		0.480925,
		0.410404,
		0.3402,
		0.271875,
		0.207112,
		0.147708,
		0.095568,
		0.052699,
		0.021206,
		0.003289,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.001221,
		0.015654,
		0.044472,
		0.085389,
		0.136215,
		0.194862,
		0.259348,
		0.327802,
		0.39847,
		0.469722,
		0.540059,
		0.608124,
		0.672708,
		0.732765,
		0.787425,
		0.836008,
		0.878037,
		0.91326,
		0.941666,
		0.963495,
		0.979253,
		0.9897,
		0.995829,
		0.9988,
		0.999829,
		0.999998,
		0.999998,
		0.999829,
		0.9988,
		0.995829,
		0.9897,
		0.979253,
		0.963495,
		0.941666,
		0.91326,
		0.878037,
		0.836008,
		0.787425,
		0.732765,
		0.672708,
		0.608124,
		0.540059,
		0.469722,
		0.39847,
		0.327802,
		0.259348,
		0.194862,
		0.136215,
		0.085389,
		0.044472,
		0.015654,
		0.001221,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.00014,
		0.010808,
		0.036715,
		0.075484,
		0.124837,
		0.182603,
		0.246719,
		0.315234,
		0.38632,
		0.458277,
		0.529537,
		0.59868,
		0.664439,
		0.725714,
		0.781584,
		0.831326,
		0.874425,
		0.9106,
		0.939815,
		0.962298,
		0.97855,
		0.98934,
		0.995678,
		0.998756,
		0.999822,
		0.999998,
		0.999998,
		0.999822,
		0.998756,
		0.995678,
		0.98934,
		0.97855,
		0.962298,
		0.939815,
		0.9106,
		0.874425,
		0.831326,
		0.781584,
		0.725714,
		0.664439,
		0.59868,
		0.529537,
		0.458277,
		0.38632,
		0.315234,
		0.246719,
		0.182603,
		0.124837,
		0.075484,
		0.036715,
		0.010808,
		0.00014,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.006746,
		0.029488,
		0.065898,
		0.113609,
		0.17036,
		0.234001,
		0.302502,
		0.373956,
		0.446587,
		0.518758,
		0.588982,
		0.655929,
		0.718443,
		0.775551,
		0.826481,
		0.870682,
		0.907839,
		0.937891,
		0.961052,
		0.977817,
		0.988964,
		0.995521,
		0.998709,
		0.999815,
		0.999997,
		0.999997,
		0.999815,
		0.998709,
		0.995521,
		0.988964,
		0.977817,
		0.961052,
		0.937891,
		0.907839,
		0.870682,
		0.826481,
		0.775551,
		0.718443,
		0.655929,
		0.588982,
		0.518758,
		0.446587,
		0.373956,
		0.302502,
		0.234001,
		0.17036,
		0.113609,
		0.065898,
		0.029488,
		0.006746,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.003552,
		0.022856,
		0.056682,
		0.102567,
		0.158156,
		0.221211,
		0.289616,
		0.361381,
		0.434652,
		0.507719,
		0.579023,
		0.647171,
		0.710944,
		0.769316,
		0.821467,
		0.866803,
		0.904972,
		0.93589,
		0.959754,
		0.977053,
		0.988571,
		0.995357,
		0.99866,
		0.999808,
		0.999997,
		0.999997,
		0.999808,
		0.99866,
		0.995357,
		0.988571,
		0.977053,
		0.959754,
		0.93589,
		0.904972,
		0.866803,
		0.821467,
		0.769316,
		0.710944,
		0.647171,
		0.579023,
		0.507719,
		0.434652,
		0.361381,
		0.289616,
		0.221211,
		0.158156,
		0.102567,
		0.056682,
		0.022856,
		0.003552,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.00132,
		0.016893,
		0.047892,
		0.091751,
		0.14602,
		0.208367,
		0.276585,
		0.348599,
		0.422472,
		0.496415,
		0.568797,
		0.638156,
		0.70321,
		0.762874,
		0.816277,
		0.86278,
		0.901995,
		0.93381,
		0.958402,
		0.976257,
		0.988161,
		0.995185,
		0.998609,
		0.9998,
		0.999997,
		0.999997,
		0.9998,
		0.998609,
		0.995185,
		0.988161,
		0.976257,
		0.958402,
		0.93381,
		0.901995,
		0.86278,
		0.816277,
		0.762874,
		0.70321,
		0.638156,
		0.568797,
		0.496415,
		0.422472,
		0.348599,
		0.276585,
		0.208367,
		0.14602,
		0.091751,
		0.047892,
		0.016893,
		0.00132,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.000152,
		0.011678,
		0.039587,
		0.081207,
		0.133985,
		0.19549,
		0.263423,
		0.335615,
		0.410046,
		0.484843,
		0.558299,
		0.628878,
		0.695232,
		0.756216,
		0.810904,
		0.858608,
		0.898903,
		0.931646,
		0.956994,
		0.975425,
		0.987733,
		0.995005,
		0.998556,
		0.999793,
		0.999997,
		0.999997,
		0.999793,
		0.998556,
		0.995005,
		0.987733,
		0.975425,
		0.956994,
		0.931646,
		0.898903,
		0.858608,
		0.810904,
		0.756216,
		0.695232,
		0.628878,
		0.558299,
		0.484843,
		0.410046,
		0.335615,
		0.263423,
		0.19549,
		0.133985,
		0.081207,
		0.039587,
		0.011678,
		0.000152,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.007298,
		0.031834,
		0.070982,
		0.122084,
		0.182603,
		0.250141,
		0.322436,
		0.397375,
		0.473,
		0.547521,
		0.619329,
		0.687003,
		0.749335,
		0.80534,
		0.854281,
		0.895691,
		0.929393,
		0.955526,
		0.974557,
		0.987285,
		0.994817,
		0.998499,
		0.999784,
		0.999997,
		0.999997,
		0.999784,
		0.998499,
		0.994817,
		0.987285,
		0.974557,
		0.955526,
		0.929393,
		0.895691,
		0.854281,
		0.80534,
		0.749335,
		0.687003,
		0.619329,
		0.547521,
		0.473,
		0.397375,
		0.322436,
		0.250141,
		0.182603,
		0.122084,
		0.070982,
		0.031834,
		0.007298,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.003848,
		0.024706,
		0.061133,
		0.110356,
		0.169733,
		0.236757,
		0.30907,
		0.38446,
		0.460882,
		0.53646,
		0.609502,
		0.678515,
		0.742222,
		0.799577,
		0.849792,
		0.892353,
		0.927049,
		0.953996,
		0.973651,
		0.986817,
		0.99462,
		0.998441,
		0.999775,
		0.999997,
		0.999997,
		0.999775,
		0.998441,
		0.99462,
		0.986817,
		0.973651,
		0.953996,
		0.927049,
		0.892353,
		0.849792,
		0.799577,
		0.742222,
		0.678515,
		0.609502,
		0.53646,
		0.460882,
		0.38446,
		0.30907,
		0.236757,
		0.169733,
		0.110356,
		0.061133,
		0.024706,
		0.003848,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.001432,
		0.018285,
		0.05172,
		0.098846,
		0.156908,
		0.223289,
		0.295525,
		0.371306,
		0.448489,
		0.525108,
		0.599388,
		0.669758,
		0.734868,
		0.793609,
		0.845133,
		0.888882,
		0.924608,
		0.952401,
		0.972704,
		0.986327,
		0.994413,
		0.998379,
		0.999766,
		0.999997,
		0.999997,
		0.999766,
		0.998379,
		0.994413,
		0.986327,
		0.972704,
		0.952401,
		0.924608,
		0.888882,
		0.845133,
		0.793609,
		0.734868,
		0.669758,
		0.599388,
		0.525108,
		0.448489,
		0.371306,
		0.295525,
		0.223289,
		0.156908,
		0.098846,
		0.05172,
		0.018285,
		0.001432,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.000165,
		0.012657,
		0.042808,
		0.087601,
		0.144161,
		0.209758,
		0.281814,
		0.357915,
		0.435818,
		0.513462,
		0.588982,
		0.660725,
		0.727265,
		0.787425,
		0.840298,
		0.885274,
		0.922066,
		0.950736,
		0.971715,
		0.985814,
		0.994196,
		0.998314,
		0.999757,
		0.999997,
		0.999997,
		0.999757,
		0.998314,
		0.994196,
		0.985814,
		0.971715,
		0.950736,
		0.922066,
		0.885274,
		0.840298,
		0.787425,
		0.727265,
		0.660725,
		0.588982,
		0.513462,
		0.435818,
		0.357915,
		0.281814,
		0.209758,
		0.144161,
		0.087601,
		0.042808,
		0.012657,
		0.000165,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.007921,
		0.034471,
		0.076675,
		0.131531,
		0.196187,
		0.26795,
		0.344293,
		0.422868,
		0.501515,
		0.578275,
		0.651406,
		0.719404,
		0.781018,
		0.835279,
		0.881522,
		0.919418,
		0.948998,
		0.970681,
		0.985277,
		0.993969,
		0.998246,
		0.999747,
		0.999997,
		0.999997,
		0.999747,
		0.998246,
		0.993969,
		0.985277,
		0.970681,
		0.948998,
		0.919418,
		0.881522,
		0.835279,
		0.781018,
		0.719404,
		0.651406,
		0.578275,
		0.501515,
		0.422868,
		0.344293,
		0.26795,
		0.196187,
		0.131531,
		0.076675,
		0.034471,
		0.007921,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.004182,
		0.026791,
		0.066128,
		0.119058,
		0.182603,
		0.253947,
		0.330446,
		0.409641,
		0.489264,
		0.567259,
		0.641793,
		0.711275,
		0.774379,
		0.830067,
		0.877618,
		0.916658,
		0.947185,
		0.9696,
		0.984714,
		0.993731,
		0.998174,
		0.999736,
		0.999996,
		0.999996,
		0.999736,
		0.998174,
		0.993731,
		0.984714,
		0.9696,
		0.947185,
		0.916658,
		0.877618,
		0.830067,
		0.774379,
		0.711275,
		0.641793,
		0.567259,
		0.489264,
		0.409641,
		0.330446,
		0.253947,
		0.182603,
		0.119058,
		0.066128,
		0.026791,
		0.004182,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.001559,
		0.019856,
		0.056024,
		0.106789,
		0.169038,
		0.239825,
		0.316383,
		0.396135,
		0.476704,
		0.555928,
		0.631876,
		0.702868,
		0.767497,
		0.824654,
		0.873557,
		0.913781,
		0.945291,
		0.968469,
		0.984125,
		0.99348,
		0.998099,
		0.999725,
		0.999996,
		0.999996,
		0.999725,
		0.998099,
		0.99348,
		0.984125,
		0.968469,
		0.945291,
		0.913781,
		0.873557,
		0.824654,
		0.767497,
		0.702868,
		0.631876,
		0.555928,
		0.476704,
		0.396135,
		0.316383,
		0.239825,
		0.169038,
		0.106789,
		0.056024,
		0.019856,
		0.001559,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.000179,
		0.013765,
		0.046438,
		0.094776,
		0.155525,
		0.225603,
		0.302114,
		0.382355,
		0.463832,
		0.544274,
		0.621646,
		0.694173,
		0.760363,
		0.819031,
		0.869329,
		0.91078,
		0.943313,
		0.967285,
		0.983507,
		0.993217,
		0.998019,
		0.999713,
		0.999996,
		0.999996,
		0.999713,
		0.998019,
		0.993217,
		0.983507,
		0.967285,
		0.943313,
		0.91078,
		0.869329,
		0.819031,
		0.760363,
		0.694173,
		0.621646,
		0.544274,
		0.463832,
		0.382355,
		0.302114,
		0.225603,
		0.155525,
		0.094776,
		0.046438,
		0.013765,
		0.000179,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.008627,
		0.03745,
		0.083077,
		0.142104,
		0.211307,
		0.287651,
		0.368303,
		0.450646,
		0.53229,
		0.611093,
		0.68518,
		0.752968,
		0.813189,
		0.864928,
		0.907651,
		0.941245,
		0.966046,
		0.982859,
		0.992941,
		0.997936,
		0.9997,
		0.999996,
		0.999996,
		0.9997,
		0.997936,
		0.992941,
		0.982859,
		0.966046,
		0.941245,
		0.907651,
		0.864928,
		0.813189,
		0.752968,
		0.68518,
		0.611093,
		0.53229,
		0.450646,
		0.368303,
		0.287651,
		0.211307,
		0.142104,
		0.083077,
		0.03745,
		0.008627,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.004562,
		0.02915,
		0.071756,
		0.128819,
		0.196963,
		0.273008,
		0.353984,
		0.437142,
		0.519969,
		0.600208,
		0.675879,
		0.745299,
		0.807118,
		0.860344,
		0.904386,
		0.939084,
		0.964748,
		0.982179,
		0.992651,
		0.997848,
		0.999687,
		0.999996,
		0.999996,
		0.999687,
		0.997848,
		0.992651,
		0.982179,
		0.964748,
		0.939084,
		0.904386,
		0.860344,
		0.807118,
		0.745299,
		0.675879,
		0.600208,
		0.519969,
		0.437142,
		0.353984,
		0.273008,
		0.196963,
		0.128819,
		0.071756,
		0.02915,
		0.004562,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.001703,
		0.021638,
		0.060886,
		0.11572,
		0.182603,
		0.258204,
		0.339405,
		0.42332,
		0.507305,
		0.588982,
		0.666257,
		0.737347,
		0.800807,
		0.85557,
		0.900978,
		0.936823,
		0.963389,
		0.981465,
		0.992346,
		0.997755,
		0.999673,
		0.999995,
		0.999995,
		0.999673,
		0.997755,
		0.992346,
		0.981465,
		0.963389,
		0.936823,
		0.900978,
		0.85557,
		0.800807,
		0.737347,
		0.666257,
		0.588982,
		0.507305,
		0.42332,
		0.339405,
		0.258204,
		0.182603,
		0.11572,
		0.060886,
		0.021638,
		0.001703,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.000196,
		0.015024,
		0.050548,
		0.102863,
		0.168264,
		0.243258,
		0.324575,
		0.409179,
		0.494292,
		0.577405,
		0.656305,
		0.729099,
		0.794247,
		0.850596,
		0.89742,
		0.934459,
		0.961963,
		0.980715,
		0.992024,
		0.997658,
		0.999658,
		0.999995,
		0.999995,
		0.999658,
		0.997658,
		0.992024,
		0.980715,
		0.961963,
		0.934459,
		0.89742,
		0.850596,
		0.794247,
		0.729099,
		0.656305,
		0.577405,
		0.494292,
		0.409179,
		0.324575,
		0.243258,
		0.168264,
		0.102863,
		0.050548,
		0.015024,
		0.000196,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.009431,
		0.04083,
		0.090309,
		0.153986,
		0.228195,
		0.309504,
		0.394723,
		0.480925,
		0.565468,
		0.64601,
		0.720543,
		0.787425,
		0.845412,
		0.893703,
		0.931984,
		0.960468,
		0.979927,
		0.991686,
		0.997555,
		0.999643,
		0.999995,
		0.999995,
		0.999643,
		0.997555,
		0.991686,
		0.979927,
		0.960468,
		0.931984,
		0.893703,
		0.845412,
		0.787425,
		0.720543,
		0.64601,
		0.565468,
		0.480925,
		0.394723,
		0.309504,
		0.228195,
		0.153986,
		0.090309,
		0.04083,
		0.009431,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.004996,
		0.031834,
		0.078131,
		0.139816,
		0.213043,
		0.294205,
		0.379951,
		0.4672,
		0.553161,
		0.63536,
		0.711668,
		0.78033,
		0.840008,
		0.889821,
		0.929393,
		0.9589,
		0.979099,
		0.99133,
		0.997446,
		0.999626,
		0.999995,
		0.999995,
		0.999626,
		0.997446,
		0.99133,
		0.979099,
		0.9589,
		0.929393,
		0.889821,
		0.840008,
		0.78033,
		0.711668,
		0.63536,
		0.553161,
		0.4672,
		0.379951,
		0.294205,
		0.213043,
		0.139816,
		0.078131,
		0.031834,
		0.004996,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.001868,
		0.02367,
		0.066407,
		0.125807,
		0.197833,
		0.278696,
		0.364871,
		0.453111,
		0.540475,
		0.624345,
		0.702461,
		0.77295,
		0.834373,
		0.885764,
		0.92668,
		0.957254,
		0.978228,
		0.990954,
		0.997332,
		0.999609,
		0.999995,
		0.999995,
		0.999609,
		0.997332,
		0.990954,
		0.978228,
		0.957254,
		0.92668,
		0.885764,
		0.834373,
		0.77295,
		0.702461,
		0.624345,
		0.540475,
		0.453111,
		0.364871,
		0.278696,
		0.197833,
		0.125807,
		0.066407,
		0.02367,
		0.001868,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.000216,
		0.016464,
		0.055226,
		0.112019,
		0.182603,
		0.262996,
		0.349487,
		0.438658,
		0.527402,
		0.612952,
		0.692907,
		0.765272,
		0.828496,
		0.881522,
		0.923837,
		0.955526,
		0.977312,
		0.990558,
		0.99721,
		0.999591,
		0.999994,
		0.999994,
		0.999591,
		0.99721,
		0.990558,
		0.977312,
		0.955526,
		0.923837,
		0.881522,
		0.828496,
		0.765272,
		0.692907,
		0.612952,
		0.527402,
		0.438658,
		0.349487,
		0.262996,
		0.182603,
		0.112019,
		0.055226,
		0.016464,
		0.000216,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.010354,
		0.044689,
		0.098521,
		0.167396,
		0.247127,
		0.333809,
		0.423837,
		0.513933,
		0.601168,
		0.682994,
		0.757282,
		0.822365,
		0.877086,
		0.920857,
		0.953711,
		0.976347,
		0.990141,
		0.997082,
		0.999571,
		0.999994,
		0.999994,
		0.999571,
		0.997082,
		0.990141,
		0.976347,
		0.953711,
		0.920857,
		0.877086,
		0.822365,
		0.757282,
		0.682994,
		0.601168,
		0.513933,
		0.423837,
		0.333809,
		0.247127,
		0.167396,
		0.098521,
		0.044689,
		0.010354,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.005495,
		0.034906,
		0.085389,
		0.152262,
		0.231119,
		0.317848,
		0.40865,
		0.500061,
		0.588982,
		0.672708,
		0.748966,
		0.815966,
		0.872446,
		0.917733,
		0.951803,
		0.975331,
		0.9897,
		0.996946,
		0.999551,
		0.999994,
		0.999994,
		0.999551,
		0.996946,
		0.9897,
		0.975331,
		0.951803,
		0.917733,
		0.872446,
		0.815966,
		0.748966,
		0.672708,
		0.588982,
		0.500061,
		0.40865,
		0.317848,
		0.231119,
		0.152262,
		0.085389,
		0.034906,
		0.005495,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.002059,
		0.026003,
		0.07271,
		0.137255,
		0.215002,
		0.301618,
		0.393097,
		0.485779,
		0.576381,
		0.662033,
		0.74031,
		0.809287,
		0.86759,
		0.914455,
		0.949797,
		0.97426,
		0.989234,
		0.996803,
		0.999529,
		0.999993,
		0.999993,
		0.999529,
		0.996803,
		0.989234,
		0.97426,
		0.949797,
		0.914455,
		0.86759,
		0.809287,
		0.74031,
		0.662033,
		0.576381,
		0.485779,
		0.393097,
		0.301618,
		0.215002,
		0.137255,
		0.07271,
		0.026003,
		0.002059,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.000238,
		0.018122,
		0.060583,
		0.122442,
		0.198815,
		0.285138,
		0.377182,
		0.471079,
		0.563354,
		0.650955,
		0.731298,
		0.802314,
		0.862506,
		0.911016,
		0.947687,
		0.97313,
		0.988742,
		0.996651,
		0.999505,
		0.999993,
		0.999993,
		0.999505,
		0.996651,
		0.988742,
		0.97313,
		0.947687,
		0.911016,
		0.862506,
		0.802314,
		0.731298,
		0.650955,
		0.563354,
		0.471079,
		0.377182,
		0.285138,
		0.198815,
		0.122442,
		0.060583,
		0.018122,
		0.000238,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.011419,
		0.049119,
		0.107895,
		0.182603,
		0.26843,
		0.360911,
		0.455958,
		0.549888,
		0.63946,
		0.721915,
		0.795032,
		0.857183,
		0.907405,
		0.945467,
		0.971939,
		0.988221,
		0.99649,
		0.999481,
		0.999993,
		0.999993,
		0.999481,
		0.99649,
		0.988221,
		0.971939,
		0.945467,
		0.907405,
		0.857183,
		0.795032,
		0.721915,
		0.63946,
		0.549888,
		0.455958,
		0.360911,
		0.26843,
		0.182603,
		0.107895,
		0.049119,
		0.011419,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.006072,
		0.038444,
		0.093699,
		0.166417,
		0.25152,
		0.344293,
		0.440412,
		0.535972,
		0.62753,
		0.712144,
		0.787425,
		0.851607,
		0.903612,
		0.943128,
		0.970681,
		0.98767,
		0.996318,
		0.999454,
		0.999992,
		0.999992,
		0.999454,
		0.996318,
		0.98767,
		0.970681,
		0.943128,
		0.903612,
		0.851607,
		0.787425,
		0.712144,
		0.62753,
		0.535972,
		0.440412,
		0.344293,
		0.25152,
		0.166417,
		0.093699,
		0.038444,
		0.006072,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.00228,
		0.028698,
		0.079949,
		0.150317,
		0.234441,
		0.32734,
		0.424437,
		0.521595,
		0.615152,
		0.701966,
		0.779477,
		0.845765,
		0.899627,
		0.940665,
		0.969353,
		0.987087,
		0.996137,
		0.999426,
		0.999992,
		0.999992,
		0.999426,
		0.996137,
		0.987087,
		0.969353,
		0.940665,
		0.899627,
		0.845765,
		0.779477,
		0.701966,
		0.615152,
		0.521595,
		0.424437,
		0.32734,
		0.234441,
		0.150317,
		0.079949,
		0.028698,
		0.00228,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.000264,
		0.020042,
		0.066755,
		0.134371,
		0.21723,
		0.310066,
		0.408035,
		0.506746,
		0.602308,
		0.691366,
		0.771171,
		0.83964,
		0.895439,
		0.938069,
		0.96795,
		0.986469,
		0.995944,
		0.999397,
		0.999991,
		0.999991,
		0.999397,
		0.995944,
		0.986469,
		0.96795,
		0.938069,
		0.895439,
		0.83964,
		0.771171,
		0.691366,
		0.602308,
		0.506746,
		0.408035,
		0.310066,
		0.21723,
		0.134371,
		0.066755,
		0.020042,
		0.000264,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.012657,
		0.05424,
		0.118659,
		0.199933,
		0.292492,
		0.391206,
		0.491417,
		0.588982,
		0.680324,
		0.762488,
		0.833218,
		0.891034,
		0.935331,
		0.966466,
		0.985814,
		0.995739,
		0.999365,
		0.999991,
		0.999991,
		0.999365,
		0.995739,
		0.985814,
		0.966466,
		0.935331,
		0.891034,
		0.833218,
		0.762488,
		0.680324,
		0.588982,
		0.491417,
		0.391206,
		0.292492,
		0.199933,
		0.118659,
		0.05424,
		0.012657,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.006746,
		0.042546,
		0.103273,
		0.182603,
		0.274643,
		0.373956,
		0.475597,
		0.575159,
		0.668821,
		0.75341,
		0.826481,
		0.8864,
		0.932443,
		0.964896,
		0.985119,
		0.995521,
		0.999331,
		0.999991,
		0.999991,
		0.999331,
		0.995521,
		0.985119,
		0.964896,
		0.932443,
		0.8864,
		0.826481,
		0.75341,
		0.668821,
		0.575159,
		0.475597,
		0.373956,
		0.274643,
		0.182603,
		0.103273,
		0.042546,
		0.006746,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.002539,
		0.031834,
		0.088318,
		0.165303,
		0.25655,
		0.356293,
		0.459281,
		0.560823,
		0.656837,
		0.743917,
		0.819413,
		0.881522,
		0.929393,
		0.963233,
		0.984381,
		0.995289,
		0.999295,
		0.99999,
		0.99999,
		0.999295,
		0.995289,
		0.984381,
		0.963233,
		0.929393,
		0.881522,
		0.819413,
		0.743917,
		0.656837,
		0.560823,
		0.459281,
		0.356293,
		0.25655,
		0.165303,
		0.088318,
		0.031834,
		0.002539,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.000295,
		0.022285,
		0.073915,
		0.148107,
		0.238249,
		0.338228,
		0.442464,
		0.545959,
		0.644352,
		0.733987,
		0.811993,
		0.876386,
		0.926173,
		0.961473,
		0.983597,
		0.995042,
		0.999256,
		0.999989,
		0.999989,
		0.999256,
		0.995042,
		0.983597,
		0.961473,
		0.926173,
		0.876386,
		0.811993,
		0.733987,
		0.644352,
		0.545959,
		0.442464,
		0.338228,
		0.238249,
		0.148107,
		0.073915,
		0.022285,
		0.000295,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.014108,
		0.060202,
		0.131098,
		0.219786,
		0.319778,
		0.425141,
		0.53055,
		0.631345,
		0.723599,
		0.804203,
		0.870975,
		0.92277,
		0.959607,
		0.982764,
		0.994778,
		0.999215,
		0.999989,
		0.999989,
		0.999215,
		0.994778,
		0.982764,
		0.959607,
		0.92277,
		0.870975,
		0.804203,
		0.723599,
		0.631345,
		0.53055,
		0.425141,
		0.319778,
		0.219786,
		0.131098,
		0.060202,
		0.014108,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.007538,
		0.047341,
		0.114378,
		0.201216,
		0.300964,
		0.407312,
		0.514584,
		0.617796,
		0.712729,
		0.796022,
		0.865273,
		0.919171,
		0.957627,
		0.981877,
		0.994497,
		0.999171,
		0.999988,
		0.999988,
		0.999171,
		0.994497,
		0.981877,
		0.957627,
		0.919171,
		0.865273,
		0.796022,
		0.712729,
		0.617796,
		0.514584,
		0.407312,
		0.300964,
		0.201216,
		0.114378,
		0.047341,
		0.007538,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.002844,
		0.035512,
		0.098062,
		0.182603,
		0.281814,
		0.38898,
		0.498046,
		0.603682,
		0.701355,
		0.787425,
		0.85926,
		0.915365,
		0.955526,
		0.980933,
		0.994196,
		0.999123,
		0.999987,
		0.999987,
		0.999123,
		0.994196,
		0.980933,
		0.955526,
		0.915365,
		0.85926,
		0.787425,
		0.701355,
		0.603682,
		0.498046,
		0.38898,
		0.281814,
		0.182603,
		0.098062,
		0.035512,
		0.002844,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.000331,
		0.024926,
		0.082283,
		0.164025,
		0.262364,
		0.370151,
		0.480925,
		0.588982,
		0.68945,
		0.778391,
		0.852917,
		0.911335,
		0.953295,
		0.979927,
		0.993875,
		0.999073,
		0.999987,
		0.999987,
		0.999073,
		0.993875,
		0.979927,
		0.953295,
		0.911335,
		0.852917,
		0.778391,
		0.68945,
		0.588982,
		0.480925,
		0.370151,
		0.262364,
		0.164025,
		0.082283,
		0.024926,
		0.000331,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.015823,
		0.067199,
		0.145573,
		0.242658,
		0.350837,
		0.463212,
		0.573675,
		0.676989,
		0.768893,
		0.846224,
		0.907068,
		0.950924,
		0.978855,
		0.993531,
		0.999018,
		0.999986,
		0.999986,
		0.999018,
		0.993531,
		0.978855,
		0.950924,
		0.907068,
		0.846224,
		0.768893,
		0.676989,
		0.573675,
		0.463212,
		0.350837,
		0.242658,
		0.145573,
		0.067199,
		0.015823,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.008478,
		0.05299,
		0.127354,
		0.222749,
		0.331053,
		0.444897,
		0.557738,
		0.663946,
		0.758906,
		0.839157,
		0.902545,
		0.948403,
		0.97771,
		0.993163,
		0.99896,
		0.999985,
		0.999985,
		0.99896,
		0.993163,
		0.97771,
		0.948403,
		0.902545,
		0.839157,
		0.758906,
		0.663946,
		0.557738,
		0.444897,
		0.331053,
		0.222749,
		0.127354,
		0.05299,
		0.008478,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.003208,
		0.039864,
		0.109496,
		0.202704,
		0.310823,
		0.425977,
		0.541152,
		0.650291,
		0.7484,
		0.831692,
		0.897749,
		0.945719,
		0.976488,
		0.992769,
		0.998897,
		0.999984,
		0.999984,
		0.998897,
		0.992769,
		0.976488,
		0.945719,
		0.897749,
		0.831692,
		0.7484,
		0.650291,
		0.541152,
		0.425977,
		0.310823,
		0.202704,
		0.109496,
		0.039864,
		0.003208,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.000375,
		0.028064,
		0.092147,
		0.182603,
		0.290179,
		0.40645,
		0.523896,
		0.635997,
		0.737347,
		0.823803,
		0.892661,
		0.942861,
		0.975181,
		0.992346,
		0.998829,
		0.999983,
		0.999983,
		0.998829,
		0.992346,
		0.975181,
		0.942861,
		0.892661,
		0.823803,
		0.737347,
		0.635997,
		0.523896,
		0.40645,
		0.290179,
		0.182603,
		0.092147,
		0.028064,
		0.000375,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.01787,
		0.075483,
		0.162543,
		0.269161,
		0.38632,
		0.505951,
		0.621033,
		0.725713,
		0.815461,
		0.887259,
		0.939814,
		0.973782,
		0.99189,
		0.998754,
		0.99998,
		0.99998,
		0.998754,
		0.99189,
		0.973782,
		0.939814,
		0.887259,
		0.815461,
		0.725713,
		0.621033,
		0.505951,
		0.38632,
		0.269161,
		0.162543,
		0.075483,
		0.01787,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.009606,
		0.059709,
		0.142636,
		0.247816,
		0.365589,
		0.487293,
		0.605362,
		0.713457,
		0.806627,
		0.881506,
		0.936549,
		0.972268,
		0.991385,
		0.998659,
		0.999963,
		0.999963,
		0.998659,
		0.991385,
		0.972268,
		0.936549,
		0.881506,
		0.806627,
		0.713457,
		0.605362,
		0.487293,
		0.365589,
		0.247816,
		0.142636,
		0.059709,
		0.009606,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.003646,
		0.045061,
		0.123016,
		0.226198,
		0.344254,
		0.467878,
		0.588915,
		0.700499,
		0.797216,
		0.875324,
		0.932992,
		0.970571,
		0.990764,
		0.998478,
		0.999866,
		0.999866,
		0.998478,
		0.990764,
		0.970571,
		0.932992,
		0.875324,
		0.797216,
		0.700499,
		0.588915,
		0.467878,
		0.344254,
		0.226198,
		0.123016,
		0.045061,
		0.003646,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.000428,
		0.031819,
		0.103831,
		0.204353,
		0.322282,
		0.447615,
		0.57156,
		0.686676,
		0.78705,
		0.868522,
		0.92895,
		0.968498,
		0.989838,
		0.998023,
		0.999501,
		0.999501,
		0.998023,
		0.989838,
		0.968498,
		0.92895,
		0.868522,
		0.78705,
		0.686676,
		0.57156,
		0.447615,
		0.322282,
		0.204353,
		0.103831,
		0.031819,
		0.000428,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.020311,
		0.085259,
		0.182325,
		0.299601,
		0.426337,
		0.55305,
		0.671681,
		0.775775,
		0.860718,
		0.924021,
		0.965637,
		0.98819,
		0.996877,
		0.998451,
		0.998451,
		0.996877,
		0.98819,
		0.965637,
		0.924021,
		0.860718,
		0.775775,
		0.671681,
		0.55305,
		0.426337,
		0.299601,
		0.182325,
		0.085259,
		0.020311,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.01093,
		0.067515,
		0.160159,
		0.276097,
		0.40378,
		0.532994,
		0.655014,
		0.762802,
		0.851254,
		0.917501,
		0.961255,
		0.985074,
		0.994288,
		0.995965,
		0.995965,
		0.994288,
		0.985074,
		0.961255,
		0.917501,
		0.851254,
		0.762802,
		0.655014,
		0.532994,
		0.40378,
		0.276097,
		0.160159,
		0.067515,
		0.01093,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.004144,
		0.050887,
		0.137938,
		0.25164,
		0.379606,
		0.510855,
		0.635962,
		0.747271,
		0.839153,
		0.908329,
		0.954239,
		0.979348,
		0.989105,
		0.990887,
		0.990887,
		0.989105,
		0.979348,
		0.954239,
		0.908329,
		0.839153,
		0.747271,
		0.635962,
		0.510855,
		0.379606,
		0.25164,
		0.137938,
		0.050887,
		0.004144,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.000483,
		0.035748,
		0.115822,
		0.226135,
		0.353452,
		0.486002,
		0.613644,
		0.728083,
		0.82314,
		0.895097,
		0.943094,
		0.969472,
		0.97977,
		0.981659,
		0.981659,
		0.97977,
		0.969472,
		0.943094,
		0.895097,
		0.82314,
		0.728083,
		0.613644,
		0.486002,
		0.353452,
		0.226135,
		0.115822,
		0.035748,
		0.000483,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.022579,
		0.094096,
		0.199599,
		0.325018,
		0.457808,
		0.587116,
		0.704007,
		0.801745,
		0.876156,
		0.92605,
		0.953609,
		0.964421,
		0.966413,
		0.966413,
		0.964421,
		0.953609,
		0.92605,
		0.876156,
		0.801745,
		0.704007,
		0.587116,
		0.457808,
		0.325018,
		0.199599,
		0.094096,
		0.022579,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.011938,
		0.073204,
		0.172233,
		0.294186,
		0.425793,
		0.555533,
		0.67386,
		0.773497,
		0.84981,
		0.901261,
		0.929828,
		0.941091,
		0.943177,
		0.943177,
		0.941091,
		0.929828,
		0.901261,
		0.84981,
		0.773497,
		0.67386,
		0.555533,
		0.425793,
		0.294186,
		0.172233,
		0.073204,
		0.011938,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.00441,
		0.053745,
		0.14448,
		0.26112,
		0.389776,
		0.518344,
		0.636733,
		0.737171,
		0.814587,
		0.86708,
		0.896384,
		0.907999,
		0.91016,
		0.91016,
		0.907999,
		0.896384,
		0.86708,
		0.814587,
		0.737171,
		0.636733,
		0.518344,
		0.389776,
		0.26112,
		0.14448,
		0.053745,
		0.00441,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.000496,
		0.036428,
		0.117035,
		0.226346,
		0.35001,
		0.475478,
		0.592226,
		0.692067,
		0.769537,
		0.822383,
		0.852052,
		0.863875,
		0.866086,
		0.866086,
		0.863875,
		0.852052,
		0.822383,
		0.769537,
		0.692067,
		0.592226,
		0.475478,
		0.35001,
		0.226346,
		0.117035,
		0.036428,
		0.000496,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.021973,
		0.090795,
		0.190753,
		0.30725,
		0.427476,
		0.540631,
		0.638232,
		0.714501,
		0.766854,
		0.79642,
		0.80827,
		0.810497,
		0.810497,
		0.80827,
		0.79642,
		0.766854,
		0.714501,
		0.638232,
		0.540631,
		0.427476,
		0.30725,
		0.190753,
		0.090795,
		0.021973,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.01098,
		0.066752,
		0.155529,
		0.262731,
		0.375519,
		0.483015,
		0.576593,
		0.650267,
		0.701172,
		0.730099,
		0.741761,
		0.743965,
		0.743965,
		0.741761,
		0.730099,
		0.701172,
		0.650267,
		0.576593,
		0.483015,
		0.375519,
		0.262731,
		0.155529,
		0.066752,
		0.01098,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.003794,
		0.045838,
		0.122012,
		0.218049,
		0.321345,
		0.42117,
		0.50894,
		0.578592,
		0.627053,
		0.654769,
		0.666012,
		0.668148,
		0.668148,
		0.666012,
		0.654769,
		0.627053,
		0.578592,
		0.50894,
		0.42117,
		0.321345,
		0.218049,
		0.122012,
		0.045838,
		0.003794,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.000395,
		0.028769,
		0.091506,
		0.174964,
		0.267053,
		0.357431,
		0.43776,
		0.502049,
		0.547106,
		0.573048,
		0.58364,
		0.585664,
		0.585664,
		0.58364,
		0.573048,
		0.547106,
		0.502049,
		0.43776,
		0.357431,
		0.267053,
		0.174964,
		0.091506,
		0.028769,
		0.000395,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.015912,
		0.065084,
		0.135159,
		0.214834,
		0.294395,
		0.365947,
		0.423733,
		0.464545,
		0.48821,
		0.497936,
		0.499807,
		0.499807,
		0.497936,
		0.48821,
		0.464545,
		0.423733,
		0.365947,
		0.294395,
		0.214834,
		0.135159,
		0.065084,
		0.015912,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.007219,
		0.043437,
		0.100022,
		0.166698,
		0.234593,
		0.296449,
		0.346893,
		0.38281,
		0.403792,
		0.412477,
		0.414158,
		0.414158,
		0.412477,
		0.403792,
		0.38281,
		0.346893,
		0.296449,
		0.234593,
		0.166698,
		0.100022,
		0.043437,
		0.007219,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.002241,
		0.026795,
		0.070476,
		0.124229,
		0.180203,
		0.23193,
		0.274558,
		0.305174,
		0.323198,
		0.330713,
		0.332178,
		0.332178,
		0.330713,
		0.323198,
		0.305174,
		0.274558,
		0.23193,
		0.180203,
		0.124229,
		0.070476,
		0.026795,
		0.002241,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.000207,
		0.014927,
		0.046907,
		0.088442,
		0.132827,
		0.174498,
		0.209231,
		0.234406,
		0.249349,
		0.255628,
		0.256861,
		0.256861,
		0.255628,
		0.249349,
		0.234406,
		0.209231,
		0.174498,
		0.132827,
		0.088442,
		0.046907,
		0.014927,
		0.000207,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.007221,
		0.029175,
		0.05973,
		0.093385,
		0.125548,
		0.152688,
		0.172554,
		0.184447,
		0.189486,
		0.190482,
		0.190482,
		0.189486,
		0.184447,
		0.172554,
		0.152688,
		0.125548,
		0.093385,
		0.05973,
		0.029175,
		0.007221,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.002811,
		0.016701,
		0.037903,
		0.062113,
		0.085717,
		0.105905,
		0.120838,
		0.12986,
		0.133715,
		0.134483,
		0.134483,
		0.133715,
		0.12986,
		0.120838,
		0.105905,
		0.085717,
		0.062113,
		0.037903,
		0.016701,
		0.002811,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.000729,
		0.008604,
		0.022299,
		0.038636,
		0.054932,
		0.069077,
		0.079658,
		0.086113,
		0.088896,
		0.089455,
		0.089455,
		0.088896,
		0.086113,
		0.079658,
		0.069077,
		0.054932,
		0.038636,
		0.022299,
		0.008604,
		0.000729,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		5.4e-005,
		0.003854,
		0.01193,
		0.022103,
		0.03252,
		0.041711,
		0.048672,
		0.052962,
		0.054829,
		0.055207,
		0.055207,
		0.054829,
		0.052962,
		0.048672,
		0.041711,
		0.03252,
		0.022103,
		0.01193,
		0.003854,
		5.4e-005,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.001417,
		0.005638,
		0.011339,
		0.017358,
		0.022767,
		0.026918,
		0.029504,
		0.030641,
		0.030874,
		0.030874,
		0.030641,
		0.029504,
		0.026918,
		0.022767,
		0.017358,
		0.011339,
		0.005638,
		0.001417,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		0.000384,
		0.002248,
		0.005009,
		0.008033,
		0.010807,
		0.012966,
		0.014328,
		0.014932,
		0.015057,
		0.015057,
		0.014932,
		0.014328,
		0.012966,
		0.010807,
		0.008033,
		0.005009,
		0.002248,
		0.000384,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		6e-005,
		0.000698,
		0.001775,
		0.003008,
		0.004166,
		0.005082,
		0.005667,
		0.005929,
		0.005984,
		0.005984,
		0.005929,
		0.005667,
		0.005082,
		0.004166,
		0.003008,
		0.001775,
		0.000698,
		6e-005,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		0.000144,
		0.000438,
		0.000793,
		0.001136,
		0.001412,
		0.00159,
		0.001671,
		0.001688,
		0.001688,
		0.001671,
		0.00159,
		0.001412,
		0.001136,
		0.000793,
		0.000438,
		0.000144,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		1.3e-005,
		5.2e-005,
		0.000103,
		0.000153,
		0.000194,
		0.000222,
		0.000234,
		0.000237,
		0.000237,
		0.000234,
		0.000222,
		0.000194,
		0.000153,
		0.000103,
		5.2e-005,
		1.3e-005,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		1.3e-005,
		5.2e-005,
		0.000103,
		0.000153,
		0.000194,
		0.000222,
		0.000234,
		0.000237,
		0.000237,
		0.000234,
		0.000222,
		0.000194,
		0.000153,
		0.000103,
		5.2e-005,
		1.3e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		0.000144,
		0.000438,
		0.000793,
		0.001136,
		0.001412,
		0.00159,
		0.001671,
		0.001688,
		0.001688,
		0.001671,
		0.00159,
		0.001412,
		0.001136,
		0.000793,
		0.000438,
		0.000144,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		6e-005,
		0.000698,
		0.001775,
		0.003008,
		0.004166,
		0.005082,
		0.005667,
		0.005929,
		0.005984,
		0.005984,
		0.005929,
		0.005667,
		0.005082,
		0.004166,
		0.003008,
		0.001775,
		0.000698,
		6e-005,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000384,
		0.002248,
		0.005009,
		0.008033,
		0.010807,
		0.012966,
		0.014328,
		0.014932,
		0.015057,
		0.015057,
		0.014932,
		0.014328,
		0.012966,
		0.010807,
		0.008033,
		0.005009,
		0.002248,
		0.000384,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		0.001417,
		0.005638,
		0.011339,
		0.017358,
		0.022767,
		0.026918,
		0.029504,
		0.030641,
		0.030874,
		0.030874,
		0.030641,
		0.029504,
		0.026918,
		0.022767,
		0.017358,
		0.011339,
		0.005638,
		0.001417,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		5.4e-005,
		0.003854,
		0.01193,
		0.022103,
		0.03252,
		0.041711,
		0.048672,
		0.052962,
		0.054829,
		0.055207,
		0.055207,
		0.054829,
		0.052962,
		0.048672,
		0.041711,
		0.03252,
		0.022103,
		0.01193,
		0.003854,
		5.4e-005,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		0.000729,
		0.008604,
		0.022299,
		0.038636,
		0.054932,
		0.069077,
		0.079658,
		0.086113,
		0.088896,
		0.089455,
		0.089455,
		0.088896,
		0.086113,
		0.079658,
		0.069077,
		0.054932,
		0.038636,
		0.022299,
		0.008604,
		0.000729,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.002811,
		0.016701,
		0.037903,
		0.062113,
		0.085717,
		0.105905,
		0.120838,
		0.12986,
		0.133715,
		0.134483,
		0.134483,
		0.133715,
		0.12986,
		0.120838,
		0.105905,
		0.085717,
		0.062113,
		0.037903,
		0.016701,
		0.002811,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.007221,
		0.029175,
		0.05973,
		0.093385,
		0.125548,
		0.152688,
		0.172554,
		0.184447,
		0.189486,
		0.190482,
		0.190482,
		0.189486,
		0.184447,
		0.172554,
		0.152688,
		0.125548,
		0.093385,
		0.05973,
		0.029175,
		0.007221,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.000207,
		0.014927,
		0.046907,
		0.088442,
		0.132827,
		0.174498,
		0.209231,
		0.234406,
		0.249349,
		0.255628,
		0.256861,
		0.256861,
		0.255628,
		0.249349,
		0.234406,
		0.209231,
		0.174498,
		0.132827,
		0.088442,
		0.046907,
		0.014927,
		0.000207,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.002241,
		0.026795,
		0.070476,
		0.124229,
		0.180203,
		0.23193,
		0.274558,
		0.305174,
		0.323198,
		0.330713,
		0.332178,
		0.332178,
		0.330713,
		0.323198,
		0.305174,
		0.274558,
		0.23193,
		0.180203,
		0.124229,
		0.070476,
		0.026795,
		0.002241,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.007219,
		0.043437,
		0.100022,
		0.166698,
		0.234593,
		0.296449,
		0.346893,
		0.38281,
		0.403792,
		0.412477,
		0.414158,
		0.414158,
		0.412477,
		0.403792,
		0.38281,
		0.346893,
		0.296449,
		0.234593,
		0.166698,
		0.100022,
		0.043437,
		0.007219,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.015912,
		0.065084,
		0.135159,
		0.214834,
		0.294395,
		0.365947,
		0.423733,
		0.464545,
		0.48821,
		0.497936,
		0.499807,
		0.499807,
		0.497936,
		0.48821,
		0.464545,
		0.423733,
		0.365947,
		0.294395,
		0.214834,
		0.135159,
		0.065084,
		0.015912,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.000395,
		0.028769,
		0.091506,
		0.174964,
		0.267053,
		0.357431,
		0.43776,
		0.502049,
		0.547106,
		0.573048,
		0.58364,
		0.585664,
		0.585664,
		0.58364,
		0.573048,
		0.547106,
		0.502049,
		0.43776,
		0.357431,
		0.267053,
		0.174964,
		0.091506,
		0.028769,
		0.000395,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.003794,
		0.045838,
		0.122012,
		0.218049,
		0.321345,
		0.42117,
		0.50894,
		0.578592,
		0.627053,
		0.654769,
		0.666012,
		0.668148,
		0.668148,
		0.666012,
		0.654769,
		0.627053,
		0.578592,
		0.50894,
		0.42117,
		0.321345,
		0.218049,
		0.122012,
		0.045838,
		0.003794,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.01098,
		0.066752,
		0.155529,
		0.262731,
		0.375519,
		0.483015,
		0.576593,
		0.650267,
		0.701172,
		0.730099,
		0.741761,
		0.743965,
		0.743965,
		0.741761,
		0.730099,
		0.701172,
		0.650267,
		0.576593,
		0.483015,
		0.375519,
		0.262731,
		0.155529,
		0.066752,
		0.01098,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.021973,
		0.090795,
		0.190753,
		0.30725,
		0.427476,
		0.540631,
		0.638232,
		0.714501,
		0.766854,
		0.79642,
		0.80827,
		0.810497,
		0.810497,
		0.80827,
		0.79642,
		0.766854,
		0.714501,
		0.638232,
		0.540631,
		0.427476,
		0.30725,
		0.190753,
		0.090795,
		0.021973,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.000496,
		0.036428,
		0.117035,
		0.226346,
		0.35001,
		0.475478,
		0.592226,
		0.692067,
		0.769537,
		0.822383,
		0.852052,
		0.863875,
		0.866086,
		0.866086,
		0.863875,
		0.852052,
		0.822383,
		0.769537,
		0.692067,
		0.592226,
		0.475478,
		0.35001,
		0.226346,
		0.117035,
		0.036428,
		0.000496,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.00441,
		0.053745,
		0.14448,
		0.26112,
		0.389776,
		0.518344,
		0.636733,
		0.737171,
		0.814587,
		0.86708,
		0.896384,
		0.907999,
		0.91016,
		0.91016,
		0.907999,
		0.896384,
		0.86708,
		0.814587,
		0.737171,
		0.636733,
		0.518344,
		0.389776,
		0.26112,
		0.14448,
		0.053745,
		0.00441,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.011938,
		0.073204,
		0.172233,
		0.294186,
		0.425793,
		0.555533,
		0.67386,
		0.773497,
		0.84981,
		0.901261,
		0.929828,
		0.941091,
		0.943177,
		0.943177,
		0.941091,
		0.929828,
		0.901261,
		0.84981,
		0.773497,
		0.67386,
		0.555533,
		0.425793,
		0.294186,
		0.172233,
		0.073204,
		0.011938,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.022579,
		0.094096,
		0.199599,
		0.325018,
		0.457808,
		0.587116,
		0.704007,
		0.801745,
		0.876156,
		0.92605,
		0.953609,
		0.964421,
		0.966413,
		0.966413,
		0.964421,
		0.953609,
		0.92605,
		0.876156,
		0.801745,
		0.704007,
		0.587116,
		0.457808,
		0.325018,
		0.199599,
		0.094096,
		0.022579,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.000483,
		0.035748,
		0.115822,
		0.226135,
		0.353452,
		0.486002,
		0.613644,
		0.728083,
		0.82314,
		0.895097,
		0.943094,
		0.969472,
		0.97977,
		0.981659,
		0.981659,
		0.97977,
		0.969472,
		0.943094,
		0.895097,
		0.82314,
		0.728083,
		0.613644,
		0.486002,
		0.353452,
		0.226135,
		0.115822,
		0.035748,
		0.000483,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.004144,
		0.050887,
		0.137938,
		0.25164,
		0.379606,
		0.510855,
		0.635962,
		0.747271,
		0.839153,
		0.908329,
		0.954239,
		0.979348,
		0.989105,
		0.990887,
		0.990887,
		0.989105,
		0.979348,
		0.954239,
		0.908329,
		0.839153,
		0.747271,
		0.635962,
		0.510855,
		0.379606,
		0.25164,
		0.137938,
		0.050887,
		0.004144,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.01093,
		0.067515,
		0.160159,
		0.276097,
		0.40378,
		0.532994,
		0.655014,
		0.762802,
		0.851254,
		0.917501,
		0.961255,
		0.985074,
		0.994288,
		0.995965,
		0.995965,
		0.994288,
		0.985074,
		0.961255,
		0.917501,
		0.851254,
		0.762802,
		0.655014,
		0.532994,
		0.40378,
		0.276097,
		0.160159,
		0.067515,
		0.01093,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.020311,
		0.085259,
		0.182325,
		0.299601,
		0.426337,
		0.55305,
		0.671681,
		0.775775,
		0.860718,
		0.924021,
		0.965637,
		0.98819,
		0.996877,
		0.998451,
		0.998451,
		0.996877,
		0.98819,
		0.965637,
		0.924021,
		0.860718,
		0.775775,
		0.671681,
		0.55305,
		0.426337,
		0.299601,
		0.182325,
		0.085259,
		0.020311,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.000428,
		0.031819,
		0.103831,
		0.204353,
		0.322282,
		0.447615,
		0.57156,
		0.686676,
		0.78705,
		0.868522,
		0.92895,
		0.968498,
		0.989838,
		0.998023,
		0.999501,
		0.999501,
		0.998023,
		0.989838,
		0.968498,
		0.92895,
		0.868522,
		0.78705,
		0.686676,
		0.57156,
		0.447615,
		0.322282,
		0.204353,
		0.103831,
		0.031819,
		0.000428,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.003646,
		0.045061,
		0.123016,
		0.226198,
		0.344254,
		0.467878,
		0.588915,
		0.700499,
		0.797216,
		0.875324,
		0.932992,
		0.970571,
		0.990764,
		0.998478,
		0.999866,
		0.999866,
		0.998478,
		0.990764,
		0.970571,
		0.932992,
		0.875324,
		0.797216,
		0.700499,
		0.588915,
		0.467878,
		0.344254,
		0.226198,
		0.123016,
		0.045061,
		0.003646,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.009606,
		0.059709,
		0.142636,
		0.247816,
		0.365589,
		0.487293,
		0.605362,
		0.713457,
		0.806627,
		0.881506,
		0.936549,
		0.972268,
		0.991385,
		0.998659,
		0.999963,
		0.999963,
		0.998659,
		0.991385,
		0.972268,
		0.936549,
		0.881506,
		0.806627,
		0.713457,
		0.605362,
		0.487293,
		0.365589,
		0.247816,
		0.142636,
		0.059709,
		0.009606,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.01787,
		0.075483,
		0.162543,
		0.269161,
		0.38632,
		0.505951,
		0.621033,
		0.725713,
		0.815461,
		0.887259,
		0.939814,
		0.973782,
		0.99189,
		0.998754,
		0.99998,
		0.99998,
		0.998754,
		0.99189,
		0.973782,
		0.939814,
		0.887259,
		0.815461,
		0.725713,
		0.621033,
		0.505951,
		0.38632,
		0.269161,
		0.162543,
		0.075483,
		0.01787,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.000375,
		0.028064,
		0.092147,
		0.182603,
		0.290179,
		0.40645,
		0.523896,
		0.635997,
		0.737347,
		0.823803,
		0.892661,
		0.942861,
		0.975181,
		0.992346,
		0.998829,
		0.999983,
		0.999983,
		0.998829,
		0.992346,
		0.975181,
		0.942861,
		0.892661,
		0.823803,
		0.737347,
		0.635997,
		0.523896,
		0.40645,
		0.290179,
		0.182603,
		0.092147,
		0.028064,
		0.000375,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.003208,
		0.039864,
		0.109496,
		0.202704,
		0.310823,
		0.425977,
		0.541152,
		0.650291,
		0.7484,
		0.831692,
		0.897749,
		0.945719,
		0.976488,
		0.992769,
		0.998897,
		0.999984,
		0.999984,
		0.998897,
		0.992769,
		0.976488,
		0.945719,
		0.897749,
		0.831692,
		0.7484,
		0.650291,
		0.541152,
		0.425977,
		0.310823,
		0.202704,
		0.109496,
		0.039864,
		0.003208,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.008478,
		0.05299,
		0.127354,
		0.222749,
		0.331053,
		0.444897,
		0.557738,
		0.663946,
		0.758906,
		0.839157,
		0.902545,
		0.948403,
		0.97771,
		0.993163,
		0.99896,
		0.999985,
		0.999985,
		0.99896,
		0.993163,
		0.97771,
		0.948403,
		0.902545,
		0.839157,
		0.758906,
		0.663946,
		0.557738,
		0.444897,
		0.331053,
		0.222749,
		0.127354,
		0.05299,
		0.008478,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.015823,
		0.067199,
		0.145573,
		0.242658,
		0.350837,
		0.463212,
		0.573675,
		0.676989,
		0.768893,
		0.846224,
		0.907068,
		0.950924,
		0.978855,
		0.993531,
		0.999018,
		0.999986,
		0.999986,
		0.999018,
		0.993531,
		0.978855,
		0.950924,
		0.907068,
		0.846224,
		0.768893,
		0.676989,
		0.573675,
		0.463212,
		0.350837,
		0.242658,
		0.145573,
		0.067199,
		0.015823,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.000331,
		0.024926,
		0.082283,
		0.164025,
		0.262364,
		0.370151,
		0.480925,
		0.588982,
		0.68945,
		0.778391,
		0.852917,
		0.911335,
		0.953295,
		0.979927,
		0.993875,
		0.999073,
		0.999987,
		0.999987,
		0.999073,
		0.993875,
		0.979927,
		0.953295,
		0.911335,
		0.852917,
		0.778391,
		0.68945,
		0.588982,
		0.480925,
		0.370151,
		0.262364,
		0.164025,
		0.082283,
		0.024926,
		0.000331,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.002844,
		0.035512,
		0.098062,
		0.182603,
		0.281814,
		0.38898,
		0.498046,
		0.603682,
		0.701355,
		0.787425,
		0.85926,
		0.915365,
		0.955526,
		0.980933,
		0.994196,
		0.999123,
		0.999987,
		0.999987,
		0.999123,
		0.994196,
		0.980933,
		0.955526,
		0.915365,
		0.85926,
		0.787425,
		0.701355,
		0.603682,
		0.498046,
		0.38898,
		0.281814,
		0.182603,
		0.098062,
		0.035512,
		0.002844,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.007538,
		0.047341,
		0.114378,
		0.201216,
		0.300964,
		0.407312,
		0.514584,
		0.617796,
		0.712729,
		0.796022,
		0.865273,
		0.919171,
		0.957627,
		0.981877,
		0.994497,
		0.999171,
		0.999988,
		0.999988,
		0.999171,
		0.994497,
		0.981877,
		0.957627,
		0.919171,
		0.865273,
		0.796022,
		0.712729,
		0.617796,
		0.514584,
		0.407312,
		0.300964,
		0.201216,
		0.114378,
		0.047341,
		0.007538,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.014108,
		0.060202,
		0.131098,
		0.219786,
		0.319778,
		0.425141,
		0.53055,
		0.631345,
		0.723599,
		0.804203,
		0.870975,
		0.92277,
		0.959607,
		0.982764,
		0.994778,
		0.999215,
		0.999989,
		0.999989,
		0.999215,
		0.994778,
		0.982764,
		0.959607,
		0.92277,
		0.870975,
		0.804203,
		0.723599,
		0.631345,
		0.53055,
		0.425141,
		0.319778,
		0.219786,
		0.131098,
		0.060202,
		0.014108,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.000295,
		0.022285,
		0.073915,
		0.148107,
		0.238249,
		0.338228,
		0.442464,
		0.545959,
		0.644352,
		0.733987,
		0.811993,
		0.876386,
		0.926173,
		0.961473,
		0.983597,
		0.995042,
		0.999256,
		0.999989,
		0.999989,
		0.999256,
		0.995042,
		0.983597,
		0.961473,
		0.926173,
		0.876386,
		0.811993,
		0.733987,
		0.644352,
		0.545959,
		0.442464,
		0.338228,
		0.238249,
		0.148107,
		0.073915,
		0.022285,
		0.000295,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.002539,
		0.031834,
		0.088318,
		0.165303,
		0.25655,
		0.356293,
		0.459281,
		0.560823,
		0.656837,
		0.743917,
		0.819413,
		0.881522,
		0.929393,
		0.963233,
		0.984381,
		0.995289,
		0.999295,
		0.99999,
		0.99999,
		0.999295,
		0.995289,
		0.984381,
		0.963233,
		0.929393,
		0.881522,
		0.819413,
		0.743917,
		0.656837,
		0.560823,
		0.459281,
		0.356293,
		0.25655,
		0.165303,
		0.088318,
		0.031834,
		0.002539,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.006746,
		0.042546,
		0.103273,
		0.182603,
		0.274643,
		0.373956,
		0.475597,
		0.575159,
		0.668821,
		0.75341,
		0.826481,
		0.8864,
		0.932443,
		0.964896,
		0.985119,
		0.995521,
		0.999331,
		0.999991,
		0.999991,
		0.999331,
		0.995521,
		0.985119,
		0.964896,
		0.932443,
		0.8864,
		0.826481,
		0.75341,
		0.668821,
		0.575159,
		0.475597,
		0.373956,
		0.274643,
		0.182603,
		0.103273,
		0.042546,
		0.006746,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.012657,
		0.05424,
		0.118659,
		0.199933,
		0.292492,
		0.391206,
		0.491417,
		0.588982,
		0.680324,
		0.762488,
		0.833218,
		0.891034,
		0.935331,
		0.966466,
		0.985814,
		0.995739,
		0.999365,
		0.999991,
		0.999991,
		0.999365,
		0.995739,
		0.985814,
		0.966466,
		0.935331,
		0.891034,
		0.833218,
		0.762488,
		0.680324,
		0.588982,
		0.491417,
		0.391206,
		0.292492,
		0.199933,
		0.118659,
		0.05424,
		0.012657,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.000264,
		0.020042,
		0.066755,
		0.134371,
		0.21723,
		0.310066,
		0.408035,
		0.506746,
		0.602308,
		0.691366,
		0.771171,
		0.83964,
		0.895439,
		0.938069,
		0.96795,
		0.986469,
		0.995944,
		0.999397,
		0.999991,
		0.999991,
		0.999397,
		0.995944,
		0.986469,
		0.96795,
		0.938069,
		0.895439,
		0.83964,
		0.771171,
		0.691366,
		0.602308,
		0.506746,
		0.408035,
		0.310066,
		0.21723,
		0.134371,
		0.066755,
		0.020042,
		0.000264,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.00228,
		0.028698,
		0.079949,
		0.150317,
		0.234441,
		0.32734,
		0.424437,
		0.521595,
		0.615152,
		0.701966,
		0.779477,
		0.845765,
		0.899627,
		0.940665,
		0.969353,
		0.987087,
		0.996137,
		0.999426,
		0.999992,
		0.999992,
		0.999426,
		0.996137,
		0.987087,
		0.969353,
		0.940665,
		0.899627,
		0.845765,
		0.779477,
		0.701966,
		0.615152,
		0.521595,
		0.424437,
		0.32734,
		0.234441,
		0.150317,
		0.079949,
		0.028698,
		0.00228,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.006072,
		0.038444,
		0.093699,
		0.166417,
		0.25152,
		0.344293,
		0.440412,
		0.535972,
		0.62753,
		0.712144,
		0.787425,
		0.851607,
		0.903612,
		0.943128,
		0.970681,
		0.98767,
		0.996318,
		0.999454,
		0.999992,
		0.999992,
		0.999454,
		0.996318,
		0.98767,
		0.970681,
		0.943128,
		0.903612,
		0.851607,
		0.787425,
		0.712144,
		0.62753,
		0.535972,
		0.440412,
		0.344293,
		0.25152,
		0.166417,
		0.093699,
		0.038444,
		0.006072,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.011419,
		0.049119,
		0.107895,
		0.182603,
		0.26843,
		0.360911,
		0.455958,
		0.549888,
		0.63946,
		0.721915,
		0.795032,
		0.857183,
		0.907405,
		0.945467,
		0.971939,
		0.988221,
		0.99649,
		0.999481,
		0.999993,
		0.999993,
		0.999481,
		0.99649,
		0.988221,
		0.971939,
		0.945467,
		0.907405,
		0.857183,
		0.795032,
		0.721915,
		0.63946,
		0.549888,
		0.455958,
		0.360911,
		0.26843,
		0.182603,
		0.107895,
		0.049119,
		0.011419,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.000238,
		0.018122,
		0.060583,
		0.122442,
		0.198815,
		0.285138,
		0.377182,
		0.471079,
		0.563354,
		0.650955,
		0.731298,
		0.802314,
		0.862506,
		0.911016,
		0.947687,
		0.97313,
		0.988742,
		0.996651,
		0.999505,
		0.999993,
		0.999993,
		0.999505,
		0.996651,
		0.988742,
		0.97313,
		0.947687,
		0.911016,
		0.862506,
		0.802314,
		0.731298,
		0.650955,
		0.563354,
		0.471079,
		0.377182,
		0.285138,
		0.198815,
		0.122442,
		0.060583,
		0.018122,
		0.000238,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.002059,
		0.026003,
		0.07271,
		0.137255,
		0.215002,
		0.301618,
		0.393097,
		0.485779,
		0.576381,
		0.662033,
		0.74031,
		0.809287,
		0.86759,
		0.914455,
		0.949797,
		0.97426,
		0.989234,
		0.996803,
		0.999529,
		0.999993,
		0.999993,
		0.999529,
		0.996803,
		0.989234,
		0.97426,
		0.949797,
		0.914455,
		0.86759,
		0.809287,
		0.74031,
		0.662033,
		0.576381,
		0.485779,
		0.393097,
		0.301618,
		0.215002,
		0.137255,
		0.07271,
		0.026003,
		0.002059,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.005495,
		0.034906,
		0.085389,
		0.152262,
		0.231119,
		0.317848,
		0.40865,
		0.500061,
		0.588982,
		0.672708,
		0.748966,
		0.815966,
		0.872446,
		0.917733,
		0.951803,
		0.975331,
		0.9897,
		0.996946,
		0.999551,
		0.999994,
		0.999994,
		0.999551,
		0.996946,
		0.9897,
		0.975331,
		0.951803,
		0.917733,
		0.872446,
		0.815966,
		0.748966,
		0.672708,
		0.588982,
		0.500061,
		0.40865,
		0.317848,
		0.231119,
		0.152262,
		0.085389,
		0.034906,
		0.005495,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.010354,
		0.044689,
		0.098521,
		0.167396,
		0.247127,
		0.333809,
		0.423837,
		0.513933,
		0.601168,
		0.682994,
		0.757282,
		0.822365,
		0.877086,
		0.920857,
		0.953711,
		0.976347,
		0.990141,
		0.997082,
		0.999571,
		0.999994,
		0.999994,
		0.999571,
		0.997082,
		0.990141,
		0.976347,
		0.953711,
		0.920857,
		0.877086,
		0.822365,
		0.757282,
		0.682994,
		0.601168,
		0.513933,
		0.423837,
		0.333809,
		0.247127,
		0.167396,
		0.098521,
		0.044689,
		0.010354,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.000216,
		0.016464,
		0.055226,
		0.112019,
		0.182603,
		0.262996,
		0.349487,
		0.438658,
		0.527402,
		0.612952,
		0.692907,
		0.765272,
		0.828496,
		0.881522,
		0.923837,
		0.955526,
		0.977312,
		0.990558,
		0.99721,
		0.999591,
		0.999994,
		0.999994,
		0.999591,
		0.99721,
		0.990558,
		0.977312,
		0.955526,
		0.923837,
		0.881522,
		0.828496,
		0.765272,
		0.692907,
		0.612952,
		0.527402,
		0.438658,
		0.349487,
		0.262996,
		0.182603,
		0.112019,
		0.055226,
		0.016464,
		0.000216,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.001868,
		0.02367,
		0.066407,
		0.125807,
		0.197833,
		0.278696,
		0.364871,
		0.453111,
		0.540475,
		0.624345,
		0.702461,
		0.77295,
		0.834373,
		0.885764,
		0.92668,
		0.957254,
		0.978228,
		0.990954,
		0.997332,
		0.999609,
		0.999995,
		0.999995,
		0.999609,
		0.997332,
		0.990954,
		0.978228,
		0.957254,
		0.92668,
		0.885764,
		0.834373,
		0.77295,
		0.702461,
		0.624345,
		0.540475,
		0.453111,
		0.364871,
		0.278696,
		0.197833,
		0.125807,
		0.066407,
		0.02367,
		0.001868,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.004996,
		0.031834,
		0.078131,
		0.139816,
		0.213043,
		0.294205,
		0.379951,
		0.4672,
		0.553161,
		0.63536,
		0.711668,
		0.78033,
		0.840008,
		0.889821,
		0.929393,
		0.9589,
		0.979099,
		0.99133,
		0.997446,
		0.999626,
		0.999995,
		0.999995,
		0.999626,
		0.997446,
		0.99133,
		0.979099,
		0.9589,
		0.929393,
		0.889821,
		0.840008,
		0.78033,
		0.711668,
		0.63536,
		0.553161,
		0.4672,
		0.379951,
		0.294205,
		0.213043,
		0.139816,
		0.078131,
		0.031834,
		0.004996,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.009431,
		0.04083,
		0.090309,
		0.153986,
		0.228195,
		0.309504,
		0.394723,
		0.480925,
		0.565468,
		0.64601,
		0.720543,
		0.787425,
		0.845412,
		0.893703,
		0.931984,
		0.960468,
		0.979927,
		0.991686,
		0.997555,
		0.999643,
		0.999995,
		0.999995,
		0.999643,
		0.997555,
		0.991686,
		0.979927,
		0.960468,
		0.931984,
		0.893703,
		0.845412,
		0.787425,
		0.720543,
		0.64601,
		0.565468,
		0.480925,
		0.394723,
		0.309504,
		0.228195,
		0.153986,
		0.090309,
		0.04083,
		0.009431,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.000196,
		0.015024,
		0.050548,
		0.102863,
		0.168264,
		0.243258,
		0.324575,
		0.409179,
		0.494292,
		0.577405,
		0.656305,
		0.729099,
		0.794247,
		0.850596,
		0.89742,
		0.934459,
		0.961963,
		0.980715,
		0.992024,
		0.997658,
		0.999658,
		0.999995,
		0.999995,
		0.999658,
		0.997658,
		0.992024,
		0.980715,
		0.961963,
		0.934459,
		0.89742,
		0.850596,
		0.794247,
		0.729099,
		0.656305,
		0.577405,
		0.494292,
		0.409179,
		0.324575,
		0.243258,
		0.168264,
		0.102863,
		0.050548,
		0.015024,
		0.000196,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.001703,
		0.021638,
		0.060886,
		0.11572,
		0.182603,
		0.258204,
		0.339405,
		0.42332,
		0.507305,
		0.588982,
		0.666257,
		0.737347,
		0.800807,
		0.85557,
		0.900978,
		0.936823,
		0.963389,
		0.981465,
		0.992346,
		0.997755,
		0.999673,
		0.999995,
		0.999995,
		0.999673,
		0.997755,
		0.992346,
		0.981465,
		0.963389,
		0.936823,
		0.900978,
		0.85557,
		0.800807,
		0.737347,
		0.666257,
		0.588982,
		0.507305,
		0.42332,
		0.339405,
		0.258204,
		0.182603,
		0.11572,
		0.060886,
		0.021638,
		0.001703,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.004562,
		0.02915,
		0.071756,
		0.128819,
		0.196963,
		0.273008,
		0.353984,
		0.437142,
		0.519969,
		0.600208,
		0.675879,
		0.745299,
		0.807118,
		0.860344,
		0.904386,
		0.939084,
		0.964748,
		0.982179,
		0.992651,
		0.997848,
		0.999687,
		0.999996,
		0.999996,
		0.999687,
		0.997848,
		0.992651,
		0.982179,
		0.964748,
		0.939084,
		0.904386,
		0.860344,
		0.807118,
		0.745299,
		0.675879,
		0.600208,
		0.519969,
		0.437142,
		0.353984,
		0.273008,
		0.196963,
		0.128819,
		0.071756,
		0.02915,
		0.004562,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.008627,
		0.03745,
		0.083077,
		0.142104,
		0.211307,
		0.287651,
		0.368303,
		0.450646,
		0.53229,
		0.611093,
		0.68518,
		0.752968,
		0.813189,
		0.864928,
		0.907651,
		0.941245,
		0.966046,
		0.982859,
		0.992941,
		0.997936,
		0.9997,
		0.999996,
		0.999996,
		0.9997,
		0.997936,
		0.992941,
		0.982859,
		0.966046,
		0.941245,
		0.907651,
		0.864928,
		0.813189,
		0.752968,
		0.68518,
		0.611093,
		0.53229,
		0.450646,
		0.368303,
		0.287651,
		0.211307,
		0.142104,
		0.083077,
		0.03745,
		0.008627,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.000179,
		0.013765,
		0.046438,
		0.094776,
		0.155525,
		0.225603,
		0.302114,
		0.382355,
		0.463832,
		0.544274,
		0.621646,
		0.694173,
		0.760363,
		0.819031,
		0.869329,
		0.91078,
		0.943313,
		0.967285,
		0.983507,
		0.993217,
		0.998019,
		0.999713,
		0.999996,
		0.999996,
		0.999713,
		0.998019,
		0.993217,
		0.983507,
		0.967285,
		0.943313,
		0.91078,
		0.869329,
		0.819031,
		0.760363,
		0.694173,
		0.621646,
		0.544274,
		0.463832,
		0.382355,
		0.302114,
		0.225603,
		0.155525,
		0.094776,
		0.046438,
		0.013765,
		0.000179,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.001559,
		0.019856,
		0.056024,
		0.106789,
		0.169038,
		0.239825,
		0.316383,
		0.396135,
		0.476704,
		0.555928,
		0.631876,
		0.702868,
		0.767497,
		0.824654,
		0.873557,
		0.913781,
		0.945291,
		0.968469,
		0.984125,
		0.99348,
		0.998099,
		0.999725,
		0.999996,
		0.999996,
		0.999725,
		0.998099,
		0.99348,
		0.984125,
		0.968469,
		0.945291,
		0.913781,
		0.873557,
		0.824654,
		0.767497,
		0.702868,
		0.631876,
		0.555928,
		0.476704,
		0.396135,
		0.316383,
		0.239825,
		0.169038,
		0.106789,
		0.056024,
		0.019856,
		0.001559,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.004182,
		0.026791,
		0.066128,
		0.119058,
		0.182603,
		0.253947,
		0.330446,
		0.409641,
		0.489264,
		0.567259,
		0.641793,
		0.711275,
		0.774379,
		0.830067,
		0.877618,
		0.916658,
		0.947185,
		0.9696,
		0.984714,
		0.993731,
		0.998174,
		0.999736,
		0.999996,
		0.999996,
		0.999736,
		0.998174,
		0.993731,
		0.984714,
		0.9696,
		0.947185,
		0.916658,
		0.877618,
		0.830067,
		0.774379,
		0.711275,
		0.641793,
		0.567259,
		0.489264,
		0.409641,
		0.330446,
		0.253947,
		0.182603,
		0.119058,
		0.066128,
		0.026791,
		0.004182,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.007921,
		0.034471,
		0.076675,
		0.131531,
		0.196187,
		0.26795,
		0.344293,
		0.422868,
		0.501515,
		0.578275,
		0.651406,
		0.719404,
		0.781018,
		0.835279,
		0.881522,
		0.919418,
		0.948998,
		0.970681,
		0.985277,
		0.993969,
		0.998246,
		0.999747,
		0.999997,
		0.999997,
		0.999747,
		0.998246,
		0.993969,
		0.985277,
		0.970681,
		0.948998,
		0.919418,
		0.881522,
		0.835279,
		0.781018,
		0.719404,
		0.651406,
		0.578275,
		0.501515,
		0.422868,
		0.344293,
		0.26795,
		0.196187,
		0.131531,
		0.076675,
		0.034471,
		0.007921,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.000165,
		0.012657,
		0.042808,
		0.087601,
		0.144161,
		0.209758,
		0.281814,
		0.357915,
		0.435818,
		0.513462,
		0.588982,
		0.660725,
		0.727265,
		0.787425,
		0.840298,
		0.885274,
		0.922066,
		0.950736,
		0.971715,
		0.985814,
		0.994196,
		0.998314,
		0.999757,
		0.999997,
		0.999997,
		0.999757,
		0.998314,
		0.994196,
		0.985814,
		0.971715,
		0.950736,
		0.922066,
		0.885274,
		0.840298,
		0.787425,
		0.727265,
		0.660725,
		0.588982,
		0.513462,
		0.435818,
		0.357915,
		0.281814,
		0.209758,
		0.144161,
		0.087601,
		0.042808,
		0.012657,
		0.000165,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.001432,
		0.018285,
		0.05172,
		0.098846,
		0.156908,
		0.223289,
		0.295525,
		0.371306,
		0.448489,
		0.525108,
		0.599388,
		0.669758,
		0.734868,
		0.793609,
		0.845133,
		0.888882,
		0.924608,
		0.952401,
		0.972704,
		0.986327,
		0.994413,
		0.998379,
		0.999766,
		0.999997,
		0.999997,
		0.999766,
		0.998379,
		0.994413,
		0.986327,
		0.972704,
		0.952401,
		0.924608,
		0.888882,
		0.845133,
		0.793609,
		0.734868,
		0.669758,
		0.599388,
		0.525108,
		0.448489,
		0.371306,
		0.295525,
		0.223289,
		0.156908,
		0.098846,
		0.05172,
		0.018285,
		0.001432,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.003848,
		0.024706,
		0.061133,
		0.110356,
		0.169733,
		0.236757,
		0.30907,
		0.38446,
		0.460882,
		0.53646,
		0.609502,
		0.678515,
		0.742222,
		0.799577,
		0.849792,
		0.892353,
		0.927049,
		0.953996,
		0.973651,
		0.986817,
		0.99462,
		0.998441,
		0.999775,
		0.999997,
		0.999997,
		0.999775,
		0.998441,
		0.99462,
		0.986817,
		0.973651,
		0.953996,
		0.927049,
		0.892353,
		0.849792,
		0.799577,
		0.742222,
		0.678515,
		0.609502,
		0.53646,
		0.460882,
		0.38446,
		0.30907,
		0.236757,
		0.169733,
		0.110356,
		0.061133,
		0.024706,
		0.003848,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.007298,
		0.031834,
		0.070982,
		0.122084,
		0.182603,
		0.250141,
		0.322436,
		0.397375,
		0.473,
		0.547521,
		0.619329,
		0.687003,
		0.749335,
		0.80534,
		0.854281,
		0.895691,
		0.929393,
		0.955526,
		0.974557,
		0.987285,
		0.994817,
		0.998499,
		0.999784,
		0.999997,
		0.999997,
		0.999784,
		0.998499,
		0.994817,
		0.987285,
		0.974557,
		0.955526,
		0.929393,
		0.895691,
		0.854281,
		0.80534,
		0.749335,
		0.687003,
		0.619329,
		0.547521,
		0.473,
		0.397375,
		0.322436,
		0.250141,
		0.182603,
		0.122084,
		0.070982,
		0.031834,
		0.007298,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.000152,
		0.011678,
		0.039587,
		0.081207,
		0.133985,
		0.19549,
		0.263423,
		0.335615,
		0.410046,
		0.484843,
		0.558299,
		0.628878,
		0.695232,
		0.756216,
		0.810904,
		0.858608,
		0.898903,
		0.931646,
		0.956994,
		0.975425,
		0.987733,
		0.995005,
		0.998556,
		0.999793,
		0.999997,
		0.999997,
		0.999793,
		0.998556,
		0.995005,
		0.987733,
		0.975425,
		0.956994,
		0.931646,
		0.898903,
		0.858608,
		0.810904,
		0.756216,
		0.695232,
		0.628878,
		0.558299,
		0.484843,
		0.410046,
		0.335615,
		0.263423,
		0.19549,
		0.133985,
		0.081207,
		0.039587,
		0.011678,
		0.000152,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.00132,
		0.016893,
		0.047892,
		0.091751,
		0.14602,
		0.208367,
		0.276585,
		0.348599,
		0.422472,
		0.496415,
		0.568797,
		0.638156,
		0.70321,
		0.762874,
		0.816277,
		0.86278,
		0.901995,
		0.93381,
		0.958402,
		0.976257,
		0.988161,
		0.995185,
		0.998609,
		0.9998,
		0.999997,
		0.999997,
		0.9998,
		0.998609,
		0.995185,
		0.988161,
		0.976257,
		0.958402,
		0.93381,
		0.901995,
		0.86278,
		0.816277,
		0.762874,
		0.70321,
		0.638156,
		0.568797,
		0.496415,
		0.422472,
		0.348599,
		0.276585,
		0.208367,
		0.14602,
		0.091751,
		0.047892,
		0.016893,
		0.00132,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.003552,
		0.022856,
		0.056682,
		0.102567,
		0.158156,
		0.221211,
		0.289616,
		0.361381,
		0.434652,
		0.507719,
		0.579023,
		0.647171,
		0.710944,
		0.769316,
		0.821467,
		0.866803,
		0.904972,
		0.93589,
		0.959754,
		0.977053,
		0.988571,
		0.995357,
		0.99866,
		0.999808,
		0.999997,
		0.999997,
		0.999808,
		0.99866,
		0.995357,
		0.988571,
		0.977053,
		0.959754,
		0.93589,
		0.904972,
		0.866803,
		0.821467,
		0.769316,
		0.710944,
		0.647171,
		0.579023,
		0.507719,
		0.434652,
		0.361381,
		0.289616,
		0.221211,
		0.158156,
		0.102567,
		0.056682,
		0.022856,
		0.003552,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.006746,
		0.029488,
		0.065898,
		0.113609,
		0.17036,
		0.234001,
		0.302502,
		0.373956,
		0.446587,
		0.518758,
		0.588982,
		0.655929,
		0.718443,
		0.775551,
		0.826481,
		0.870682,
		0.907839,
		0.937891,
		0.961052,
		0.977817,
		0.988964,
		0.995521,
		0.998709,
		0.999815,
		0.999997,
		0.999997,
		0.999815,
		0.998709,
		0.995521,
		0.988964,
		0.977817,
		0.961052,
		0.937891,
		0.907839,
		0.870682,
		0.826481,
		0.775551,
		0.718443,
		0.655929,
		0.588982,
		0.518758,
		0.446587,
		0.373956,
		0.302502,
		0.234001,
		0.17036,
		0.113609,
		0.065898,
		0.029488,
		0.006746,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.00014,
		0.010808,
		0.036715,
		0.075484,
		0.124837,
		0.182603,
		0.246719,
		0.315234,
		0.38632,
		0.458277,
		0.529537,
		0.59868,
		0.664439,
		0.725714,
		0.781584,
		0.831326,
		0.874425,
		0.9106,
		0.939815,
		0.962298,
		0.97855,
		0.98934,
		0.995678,
		0.998756,
		0.999822,
		0.999998,
		0.999998,
		0.999822,
		0.998756,
		0.995678,
		0.98934,
		0.97855,
		0.962298,
		0.939815,
		0.9106,
		0.874425,
		0.831326,
		0.781584,
		0.725714,
		0.664439,
		0.59868,
		0.529537,
		0.458277,
		0.38632,
		0.315234,
		0.246719,
		0.182603,
		0.124837,
		0.075484,
		0.036715,
		0.010808,
		0.00014,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.001221,
		0.015654,
		0.044472,
		0.085389,
		0.136215,
		0.194862,
		0.259348,
		0.327802,
		0.39847,
		0.469722,
		0.540059,
		0.608124,
		0.672708,
		0.732765,
		0.787425,
		0.836008,
		0.878037,
		0.91326,
		0.941666,
		0.963495,
		0.979253,
		0.9897,
		0.995829,
		0.9988,
		0.999829,
		0.999998,
		0.999998,
		0.999829,
		0.9988,
		0.995829,
		0.9897,
		0.979253,
		0.963495,
		0.941666,
		0.91326,
		0.878037,
		0.836008,
		0.787425,
		0.732765,
		0.672708,
		0.608124,
		0.540059,
		0.469722,
		0.39847,
		0.327802,
		0.259348,
		0.194862,
		0.136215,
		0.085389,
		0.044472,
		0.015654,
		0.001221,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.003289,
		0.021206,
		0.052699,
		0.095568,
		0.147708,
		0.207112,
		0.271875,
		0.3402,
		0.410404,
		0.480925,
		0.55033,
		0.617319,
		0.680741,
		0.739603,
		0.79308,
		0.840533,
		0.881522,
		0.915824,
		0.943447,
		0.964646,
		0.979927,
		0.990045,
		0.995972,
		0.998843,
		0.999835,
		0.999998,
		0.999998,
		0.999835,
		0.998843,
		0.995972,
		0.990045,
		0.979927,
		0.964646,
		0.943447,
		0.915824,
		0.881522,
		0.840533,
		0.79308,
		0.739603,
		0.680741,
		0.617319,
		0.55033,
		0.480925,
		0.410404,
		0.3402,
		0.271875,
		0.207112,
		0.147708,
		0.095568,
		0.052699,
		0.021206,
		0.003289,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.006254,
		0.027391,
		0.061339,
		0.10598,
		0.159289,
		0.219335,
		0.284287,
		0.35242,
		0.42212,
		0.491888,
		0.560352,
		0.626271,
		0.688547,
		0.746234,
		0.798554,
		0.844907,
		0.884886,
		0.918295,
		0.945162,
		0.965752,
		0.980575,
		0.990375,
		0.99611,
		0.998883,
		0.999841,
		0.999998,
		0.999998,
		0.999841,
		0.998883,
		0.99611,
		0.990375,
		0.980575,
		0.965752,
		0.945162,
		0.918295,
		0.884886,
		0.844907,
		0.798554,
		0.746234,
		0.688547,
		0.626271,
		0.560352,
		0.491888,
		0.42212,
		0.35242,
		0.284287,
		0.219335,
		0.159289,
		0.10598,
		0.061339,
		0.027391,
		0.006254,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.00013,
		0.010032,
		0.034144,
		0.070342,
		0.116587,
		0.170928,
		0.231511,
		0.296573,
		0.364459,
		0.433618,
		0.502614,
		0.570132,
		0.634987,
		0.696131,
		0.752666,
		0.803856,
		0.849137,
		0.888134,
		0.920678,
		0.946813,
		0.966816,
		0.981196,
		0.990693,
		0.996242,
		0.998922,
		0.999846,
		0.999998,
		0.999998,
		0.999846,
		0.998922,
		0.996242,
		0.990693,
		0.981196,
		0.966816,
		0.946813,
		0.920678,
		0.888134,
		0.849137,
		0.803856,
		0.752666,
		0.696131,
		0.634987,
		0.570132,
		0.502614,
		0.433618,
		0.364459,
		0.296573,
		0.231511,
		0.170928,
		0.116587,
		0.070342,
		0.034144,
		0.010032,
		0.00013,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.001133,
		0.014547,
		0.041405,
		0.079662,
		0.127354,
		0.182603,
		0.243624,
		0.308725,
		0.376311,
		0.444897,
		0.513104,
		0.579674,
		0.643472,
		0.7035,
		0.758906,
		0.80899,
		0.853227,
		0.891271,
		0.922975,
		0.948403,
		0.967838,
		0.981794,
		0.990997,
		0.996368,
		0.99896,
		0.999852,
		0.999998,
		0.999998,
		0.999852,
		0.99896,
		0.996368,
		0.990997,
		0.981794,
		0.967838,
		0.948403,
		0.922975,
		0.891271,
		0.853227,
		0.80899,
		0.758906,
		0.7035,
		0.643472,
		0.579674,
		0.513104,
		0.444897,
		0.376311,
		0.308725,
		0.243624,
		0.182603,
		0.127354,
		0.079662,
		0.041405,
		0.014547,
		0.001133,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.003055,
		0.019728,
		0.049119,
		0.089257,
		0.13825,
		0.194292,
		0.255662,
		0.320733,
		0.387975,
		0.455958,
		0.523363,
		0.588982,
		0.651733,
		0.710662,
		0.764959,
		0.813964,
		0.857183,
		0.894301,
		0.925192,
		0.949934,
		0.968823,
		0.982367,
		0.991289,
		0.99649,
		0.998995,
		0.999857,
		0.999998,
		0.999998,
		0.999857,
		0.998995,
		0.99649,
		0.991289,
		0.982367,
		0.968823,
		0.949934,
		0.925192,
		0.894301,
		0.857183,
		0.813964,
		0.764959,
		0.710662,
		0.651733,
		0.588982,
		0.523363,
		0.455958,
		0.387975,
		0.320733,
		0.255662,
		0.194292,
		0.13825,
		0.089257,
		0.049119,
		0.019728,
		0.003055,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.005814,
		0.02551,
		0.057235,
		0.099088,
		0.149248,
		0.205974,
		0.26761,
		0.332591,
		0.399446,
		0.466802,
		0.533393,
		0.598062,
		0.659775,
		0.717621,
		0.770831,
		0.818782,
		0.861011,
		0.897228,
		0.92733,
		0.95141,
		0.96977,
		0.982919,
		0.991569,
		0.996606,
		0.999029,
		0.999862,
		0.999998,
		0.999998,
		0.999862,
		0.999029,
		0.996606,
		0.991569,
		0.982919,
		0.96977,
		0.95141,
		0.92733,
		0.897228,
		0.861011,
		0.818782,
		0.770831,
		0.717621,
		0.659775,
		0.598062,
		0.533393,
		0.466802,
		0.399446,
		0.332591,
		0.26761,
		0.205974,
		0.149248,
		0.099088,
		0.057235,
		0.02551,
		0.005814,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.00012,
		0.009337,
		0.031834,
		0.065706,
		0.109121,
		0.160321,
		0.217631,
		0.279458,
		0.344293,
		0.410723,
		0.47743,
		0.543198,
		0.606919,
		0.667604,
		0.724384,
		0.776529,
		0.82345,
		0.864714,
		0.900056,
		0.929393,
		0.952833,
		0.970681,
		0.983449,
		0.991839,
		0.996717,
		0.999062,
		0.999867,
		0.999998,
		0.999998,
		0.999867,
		0.999062,
		0.996717,
		0.991839,
		0.983449,
		0.970681,
		0.952833,
		0.929393,
		0.900056,
		0.864714,
		0.82345,
		0.776529,
		0.724384,
		0.667604,
		0.606919,
		0.543198,
		0.47743,
		0.410723,
		0.344293,
		0.279458,
		0.217631,
		0.160321,
		0.109121,
		0.065706,
		0.031834,
		0.009337,
		0.00012,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.001053,
		0.013553,
		0.038645,
		0.07449,
		0.119322,
		0.171447,
		0.229249,
		0.291195,
		0.355834,
		0.421806,
		0.487844,
		0.552782,
		0.615558,
		0.675225,
		0.730958,
		0.782059,
		0.827973,
		0.868297,
		0.90279,
		0.931385,
		0.954204,
		0.971559,
		0.983959,
		0.992097,
		0.996824,
		0.999093,
		0.999871,
		0.999998,
		0.999998,
		0.999871,
		0.999093,
		0.996824,
		0.992097,
		0.983959,
		0.971559,
		0.954204,
		0.931385,
		0.90279,
		0.868297,
		0.827973,
		0.782059,
		0.730958,
		0.675225,
		0.615558,
		0.552782,
		0.487844,
		0.421806,
		0.355834,
		0.291195,
		0.229249,
		0.171447,
		0.119322,
		0.07449,
		0.038645,
		0.013553,
		0.001053,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.002844,
		0.018399,
		0.045892,
		0.083548,
		0.129662,
		0.182603,
		0.240813,
		0.302814,
		0.367211,
		0.432694,
		0.498046,
		0.562148,
		0.623984,
		0.682646,
		0.737347,
		0.787425,
		0.832357,
		0.871765,
		0.905432,
		0.933308,
		0.955526,
		0.972405,
		0.98445,
		0.992346,
		0.996926,
		0.999123,
		0.999876,
		0.999998,
		0.999998,
		0.999876,
		0.999123,
		0.996926,
		0.992346,
		0.98445,
		0.972405,
		0.955526,
		0.933308,
		0.905432,
		0.871765,
		0.832357,
		0.787425,
		0.737347,
		0.682646,
		0.623984,
		0.562148,
		0.498046,
		0.432694,
		0.367211,
		0.302814,
		0.240813,
		0.182603,
		0.129662,
		0.083548,
		0.045892,
		0.018399,
		0.002844,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.005419,
		0.023816,
		0.053528,
		0.092843,
		0.140116,
		0.193772,
		0.252311,
		0.314308,
		0.378419,
		0.443386,
		0.508038,
		0.571301,
		0.632201,
		0.68987,
		0.743557,
		0.792634,
		0.836606,
		0.875123,
		0.907987,
		0.935165,
		0.956802,
		0.973219,
		0.984922,
		0.992584,
		0.997025,
		0.999152,
		0.99988,
		0.999998,
		0.999998,
		0.99988,
		0.999152,
		0.997025,
		0.992584,
		0.984922,
		0.973219,
		0.956802,
		0.935165,
		0.907987,
		0.875123,
		0.836606,
		0.792634,
		0.743557,
		0.68987,
		0.632201,
		0.571301,
		0.508038,
		0.443386,
		0.378419,
		0.314308,
		0.252311,
		0.193772,
		0.140116,
		0.092843,
		0.053528,
		0.023816,
		0.005419,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.000112,
		0.008711,
		0.02975,
		0.061512,
		0.102344,
		0.150658,
		0.204936,
		0.26373,
		0.325668,
		0.389456,
		0.453882,
		0.517822,
		0.580244,
		0.640214,
		0.696903,
		0.749594,
		0.79769,
		0.840726,
		0.878374,
		0.910457,
		0.936959,
		0.958032,
		0.974004,
		0.985376,
		0.992814,
		0.997119,
		0.99918,
		0.999884,
		0.999998,
		0.999998,
		0.999884,
		0.99918,
		0.997119,
		0.992814,
		0.985376,
		0.974004,
		0.958032,
		0.936959,
		0.910457,
		0.878374,
		0.840726,
		0.79769,
		0.749594,
		0.696903,
		0.640214,
		0.580244,
		0.517822,
		0.453882,
		0.389456,
		0.325668,
		0.26373,
		0.204936,
		0.150658,
		0.102344,
		0.061512,
		0.02975,
		0.008711,
		0.000112,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.000982,
		0.012657,
		0.03615,
		0.069803,
		0.112019,
		0.161266,
		0.216079,
		0.275061,
		0.336891,
		0.40032,
		0.464184,
		0.527402,
		0.588982,
		0.648029,
		0.703751,
		0.755463,
		0.802599,
		0.84472,
		0.881522,
		0.912847,
		0.938692,
		0.95922,
		0.974761,
		0.985814,
		0.993035,
		0.99721,
		0.999206,
		0.999888,
		0.999998,
		0.999998,
		0.999888,
		0.999206,
		0.99721,
		0.993035,
		0.985814,
		0.974761,
		0.95922,
		0.938692,
		0.912847,
		0.881522,
		0.84472,
		0.802599,
		0.755463,
		0.703751,
		0.648029,
		0.588982,
		0.527402,
		0.464184,
		0.40032,
		0.336891,
		0.275061,
		0.216079,
		0.161266,
		0.112019,
		0.069803,
		0.03615,
		0.012657,
		0.000982,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.002655,
		0.0172,
		0.042971,
		0.078366,
		0.121842,
		0.171921,
		0.227188,
		0.286296,
		0.34797,
		0.41101,
		0.474293,
		0.53678,
		0.597518,
		0.655651,
		0.710418,
		0.761169,
		0.807365,
		0.848593,
		0.884571,
		0.915158,
		0.940367,
		0.960366,
		0.97549,
		0.986235,
		0.993247,
		0.997298,
		0.999232,
		0.999891,
		0.999999,
		0.999999,
		0.999891,
		0.999232,
		0.997298,
		0.993247,
		0.986235,
		0.97549,
		0.960366,
		0.940367,
		0.915158,
		0.884571,
		0.848593,
		0.807365,
		0.761169,
		0.710418,
		0.655651,
		0.597518,
		0.53678,
		0.474293,
		0.41101,
		0.34797,
		0.286296,
		0.227188,
		0.171921,
		0.121842,
		0.078366,
		0.042971,
		0.0172,
		0.002655,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.005063,
		0.022285,
		0.050169,
		0.087167,
		0.131787,
		0.182603,
		0.238249,
		0.297425,
		0.358903,
		0.421525,
		0.48421,
		0.545959,
		0.605858,
		0.663083,
		0.71691,
		0.766717,
		0.811993,
		0.85235,
		0.887524,
		0.917395,
		0.941985,
		0.961473,
		0.976194,
		0.986641,
		0.993452,
		0.997381,
		0.999256,
		0.999895,
		0.999999,
		0.999999,
		0.999895,
		0.999256,
		0.997381,
		0.993452,
		0.986641,
		0.976194,
		0.961473,
		0.941985,
		0.917395,
		0.887524,
		0.85235,
		0.811993,
		0.766717,
		0.71691,
		0.663083,
		0.605858,
		0.545959,
		0.48421,
		0.421525,
		0.358903,
		0.297425,
		0.238249,
		0.182603,
		0.131787,
		0.087167,
		0.050169,
		0.022285,
		0.005063,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.000105,
		0.008146,
		0.027864,
		0.057705,
		0.096175,
		0.141832,
		0.193297,
		0.249252,
		0.308444,
		0.369686,
		0.431863,
		0.493936,
		0.554942,
		0.614003,
		0.670332,
		0.723232,
		0.772112,
		0.816488,
		0.855993,
		0.890386,
		0.91956,
		0.94355,
		0.962541,
		0.976872,
		0.987032,
		0.993648,
		0.997462,
		0.99928,
		0.999898,
		0.999999,
		0.999999,
		0.999898,
		0.99928,
		0.997462,
		0.993648,
		0.987032,
		0.976872,
		0.962541,
		0.94355,
		0.91956,
		0.890386,
		0.855993,
		0.816488,
		0.772112,
		0.723232,
		0.670332,
		0.614003,
		0.554942,
		0.493936,
		0.431863,
		0.369686,
		0.308444,
		0.249252,
		0.193297,
		0.141832,
		0.096175,
		0.057705,
		0.027864,
		0.008146,
		0.000105,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.000918,
		0.011847,
		0.033889,
		0.065543,
		0.105362,
		0.151954,
		0.203986,
		0.260186,
		0.319344,
		0.380315,
		0.442026,
		0.503473,
		0.563732,
		0.62196,
		0.6774,
		0.729388,
		0.777359,
		0.820854,
		0.859528,
		0.893159,
		0.921655,
		0.945063,
		0.963574,
		0.977527,
		0.987409,
		0.993838,
		0.99754,
		0.999302,
		0.999902,
		0.999999,
		0.999999,
		0.999902,
		0.999302,
		0.99754,
		0.993838,
		0.987409,
		0.977527,
		0.963574,
		0.945063,
		0.921655,
		0.893159,
		0.859528,
		0.820854,
		0.777359,
		0.729388,
		0.6774,
		0.62196,
		0.563732,
		0.503473,
		0.442026,
		0.380315,
		0.319344,
		0.260186,
		0.203986,
		0.151954,
		0.105362,
		0.065543,
		0.033889,
		0.011847,
		0.000918,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.002483,
		0.016114,
		0.04032,
		0.073649,
		0.114702,
		0.162135,
		0.214658,
		0.271043,
		0.330122,
		0.39079,
		0.452013,
		0.512824,
		0.572334,
		0.629732,
		0.684294,
		0.735383,
		0.782462,
		0.825095,
		0.862958,
		0.895847,
		0.923684,
		0.946526,
		0.964571,
		0.978159,
		0.987773,
		0.99402,
		0.997614,
		0.999324,
		0.999905,
		0.999999,
		0.999999,
		0.999905,
		0.999324,
		0.997614,
		0.99402,
		0.987773,
		0.978159,
		0.964571,
		0.946526,
		0.923684,
		0.895847,
		0.862958,
		0.825095,
		0.782462,
		0.735383,
		0.684294,
		0.629732,
		0.572334,
		0.512824,
		0.452013,
		0.39079,
		0.330122,
		0.271043,
		0.214658,
		0.162135,
		0.114702,
		0.073649,
		0.04032,
		0.016114,
		0.002483,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.004741,
		0.020897,
		0.047116,
		0.081992,
		0.124172,
		0.172357,
		0.2253,
		0.281814,
		0.340772,
		0.401108,
		0.461824,
		0.52199,
		0.580749,
		0.637323,
		0.691017,
		0.741222,
		0.787425,
		0.829215,
		0.866286,
		0.898452,
		0.925648,
		0.947941,
		0.965534,
		0.978769,
		0.988123,
		0.994196,
		0.997686,
		0.999345,
		0.999908,
		0.999999,
		0.999999,
		0.999908,
		0.999345,
		0.997686,
		0.994196,
		0.988123,
		0.978769,
		0.965534,
		0.947941,
		0.925648,
		0.898452,
		0.866286,
		0.829215,
		0.787425,
		0.741222,
		0.691017,
		0.637323,
		0.580749,
		0.52199,
		0.461824,
		0.401108,
		0.340772,
		0.281814,
		0.2253,
		0.172357,
		0.124172,
		0.081992,
		0.047116,
		0.020897,
		0.004741,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		9.8e-005,
		0.007634,
		0.026151,
		0.05424,
		0.090543,
		0.13375,
		0.182603,
		0.2359,
		0.292492,
		0.351291,
		0.411269,
		0.471461,
		0.530974,
		0.588982,
		0.644738,
		0.697574,
		0.746909,
		0.792253,
		0.833218,
		0.869516,
		0.900978,
		0.927551,
		0.949311,
		0.966466,
		0.979358,
		0.988461,
		0.994366,
		0.997755,
		0.999365,
		0.999911,
		0.999999,
		0.999999,
		0.999911,
		0.999365,
		0.997755,
		0.994366,
		0.988461,
		0.979358,
		0.966466,
		0.949311,
		0.927551,
		0.900978,
		0.869516,
		0.833218,
		0.792253,
		0.746909,
		0.697574,
		0.644738,
		0.588982,
		0.530974,
		0.471461,
		0.411269,
		0.351291,
		0.292492,
		0.2359,
		0.182603,
		0.13375,
		0.090543,
		0.05424,
		0.026151,
		0.007634,
		9.8e-005,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		0.00086,
		0.011113,
		0.031834,
		0.06166,
		0.099276,
		0.143415,
		0.19286,
		0.246449,
		0.303071,
		0.361676,
		0.42127,
		0.480925,
		0.539778,
		0.597036,
		0.651979,
		0.703969,
		0.752448,
		0.796951,
		0.837108,
		0.872652,
		0.903427,
		0.929393,
		0.950636,
		0.967366,
		0.979927,
		0.988787,
		0.994529,
		0.997822,
		0.999384,
		0.999913,
		0.999999,
		0.999999,
		0.999913,
		0.999384,
		0.997822,
		0.994529,
		0.988787,
		0.979927,
		0.967366,
		0.950636,
		0.929393,
		0.903427,
		0.872652,
		0.837108,
		0.796951,
		0.752448,
		0.703969,
		0.651979,
		0.597036,
		0.539778,
		0.480925,
		0.42127,
		0.361676,
		0.303071,
		0.246449,
		0.19286,
		0.143415,
		0.099276,
		0.06166,
		0.031834,
		0.011113,
		0.00086,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.002328,
		0.015128,
		0.037906,
		0.069344,
		0.108167,
		0.153149,
		0.203114,
		0.256937,
		0.313545,
		0.371923,
		0.431113,
		0.490217,
		0.548406,
		0.604914,
		0.659052,
		0.710206,
		0.757843,
		0.801521,
		0.840888,
		0.875696,
		0.905802,
		0.931179,
		0.951918,
		0.968237,
		0.980477,
		0.989102,
		0.994686,
		0.997886,
		0.999403,
		0.999916,
		0.999999,
		0.999999,
		0.999916,
		0.999403,
		0.997886,
		0.994686,
		0.989102,
		0.980477,
		0.968237,
		0.951918,
		0.931179,
		0.905802,
		0.875696,
		0.840888,
		0.801521,
		0.757843,
		0.710206,
		0.659052,
		0.604914,
		0.548406,
		0.490217,
		0.431113,
		0.371923,
		0.313545,
		0.256937,
		0.203114,
		0.153149,
		0.108167,
		0.069344,
		0.037906,
		0.015128,
		0.002328,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		0.004448,
		0.019635,
		0.044332,
		0.077262,
		0.117192,
		0.162935,
		0.213353,
		0.267356,
		0.32391,
		0.382031,
		0.440796,
		0.499339,
		0.556859,
		0.612621,
		0.66596,
		0.71629,
		0.7631,
		0.805968,
		0.844563,
		0.878652,
		0.908106,
		0.932909,
		0.95316,
		0.969079,
		0.981008,
		0.989406,
		0.994838,
		0.997948,
		0.999421,
		0.999919,
		0.999999,
		0.999999,
		0.999919,
		0.999421,
		0.997948,
		0.994838,
		0.989406,
		0.981008,
		0.969079,
		0.95316,
		0.932909,
		0.908106,
		0.878652,
		0.844563,
		0.805968,
		0.7631,
		0.71629,
		0.66596,
		0.612621,
		0.556859,
		0.499339,
		0.440796,
		0.382031,
		0.32391,
		0.267356,
		0.213353,
		0.162935,
		0.117192,
		0.077262,
		0.044332,
		0.019635,
		0.004448,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		9.2e-005,
		0.00717,
		0.024591,
		0.051077,
		0.085389,
		0.126332,
		0.172758,
		0.223565,
		0.2777,
		0.33416,
		0.391998,
		0.45032,
		0.508291,
		0.56514,
		0.620158,
		0.672708,
		0.722224,
		0.76822,
		0.810295,
		0.848135,
		0.881522,
		0.910341,
		0.934586,
		0.954362,
		0.969893,
		0.981521,
		0.9897,
		0.994985,
		0.998008,
		0.999438,
		0.999921,
		0.999999,
		0.999999,
		0.999921,
		0.999438,
		0.998008,
		0.994985,
		0.9897,
		0.981521,
		0.969893,
		0.954362,
		0.934586,
		0.910341,
		0.881522,
		0.848135,
		0.810295,
		0.76822,
		0.722224,
		0.672708,
		0.620158,
		0.56514,
		0.508291,
		0.45032,
		0.391998,
		0.33416,
		0.2777,
		0.223565,
		0.172758,
		0.126332,
		0.085389,
		0.051077,
		0.024591,
		0.00717,
		9.2e-005,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		0.000807,
		0.010445,
		0.029959,
		0.058111,
		0.093699,
		0.135567,
		0.182603,
		0.233741,
		0.28796,
		0.344293,
		0.401822,
		0.459685,
		0.517077,
		0.573252,
		0.62753,
		0.679298,
		0.728012,
		0.77321,
		0.814507,
		0.851607,
		0.88431,
		0.91251,
		0.936212,
		0.955526,
		0.970681,
		0.982017,
		0.989983,
		0.995126,
		0.998065,
		0.999454,
		0.999923,
		0.999999,
		0.999999,
		0.999923,
		0.999454,
		0.998065,
		0.995126,
		0.989983,
		0.982017,
		0.970681,
		0.955526,
		0.936212,
		0.91251,
		0.88431,
		0.851607,
		0.814507,
		0.77321,
		0.728012,
		0.679298,
		0.62753,
		0.573252,
		0.517077,
		0.459685,
		0.401822,
		0.344293,
		0.28796,
		0.233741,
		0.182603,
		0.135567,
		0.093699,
		0.058111,
		0.029959,
		0.010445,
		0.000807,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.000361,
		0.008674,
		0.027023,
		0.054161,
		0.088884,
		0.130029,
		0.176478,
		0.227159,
		0.281048,
		0.337168,
		0.394596,
		0.452462,
		0.509955,
		0.566321,
		0.620871,
		0.672983,
		0.722107,
		0.767766,
		0.809569,
		0.847207,
		0.880467,
		0.909232,
		0.933492,
		0.953345,
		0.969004,
		0.980795,
		0.989154,
		0.994617,
		0.997796,
		0.999343,
		0.999896,
		0.999998,
		1,
		0.999945,
		0.999551,
		0.99831,
		0.995601,
		0.990768,
		0.983187,
		0.972299,
		0.957642,
		0.938862,
		0.915718,
		0.888082,
		0.855939,
		0.819379,
		0.778592,
		0.733863,
		0.685564,
		0.63415,
		0.580154,
		0.52418,
		0.466901,
		0.409054,
		0.351437,
		0.294906,
		0.24037,
		0.188793,
		0.141187,
		0.098612,
		0.062177,
		0.033031,
		0.01237,
		0.001428,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.003233,
		0.016928,
		0.040162,
		0.071682,
		0.110275,
		0.154777,
		0.204067,
		0.257074,
		0.312776,
		0.370205,
		0.428447,
		0.486644,
		0.544,
		0.599783,
		0.653328,
		0.704043,
		0.751412,
		0.794999,
		0.834457,
		0.869533,
		0.900072,
		0.926026,
		0.947458,
		0.964547,
		0.977591,
		0.987004,
		0.993303,
		0.997097,
		0.999046,
		0.999813,
		0.99999,
		1,
		0.999974,
		0.999698,
		0.998695,
		0.996348,
		0.991988,
		0.984964,
		0.974689,
		0.960668,
		0.942516,
		0.919966,
		0.892867,
		0.861185,
		0.824995,
		0.784477,
		0.739908,
		0.691659,
		0.640185,
		0.586022,
		0.529781,
		0.472143,
		0.413858,
		0.355737,
		0.298652,
		0.24353,
		0.191355,
		0.14316,
		0.100028,
		0.063092,
		0.033529,
		0.01256,
		0.001451,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.000372,
		0.008946,
		0.027851,
		0.055783,
		0.09148,
		0.133727,
		0.181354,
		0.233241,
		0.288319,
		0.345573,
		0.40404,
		0.46282,
		0.52107,
		0.578013,
		0.63294,
		0.685216,
		0.73428,
		0.779657,
		0.820956,
		0.857883,
		0.890244,
		0.91795,
		0.941027,
		0.959621,
		0.973997,
		0.984543,
		0.991758,
		0.99624,
		0.998655,
		0.999688,
		0.999973,
		1,
		0.99999,
		0.999807,
		0.999017,
		0.997011,
		0.993109,
		0.986637,
		0.976975,
		0.963598,
		0.946088,
		0.924149,
		0.897607,
		0.866407,
		0.830607,
		0.790377,
		0.745987,
		0.697803,
		0.646281,
		0.59196,
		0.535457,
		0.477463,
		0.41874,
		0.360112,
		0.302467,
		0.246752,
		0.193969,
		0.145174,
		0.101474,
		0.064028,
		0.034038,
		0.012755,
		0.001474,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.003336,
		0.017457,
		0.041389,
		0.073818,
		0.113477,
		0.159146,
		0.209654,
		0.263882,
		0.320766,
		0.379296,
		0.438523,
		0.497557,
		0.555576,
		0.611825,
		0.665622,
		0.716363,
		0.763528,
		0.806683,
		0.845492,
		0.879718,
		0.909232,
		0.934022,
		0.954194,
		0.969981,
		0.981742,
		0.989954,
		0.995203,
		0.998153,
		0.999508,
		0.99994,
		1,
		0.999997,
		0.999886,
		0.999281,
		0.997593,
		0.994133,
		0.988203,
		0.979155,
		0.966428,
		0.949573,
		0.928264,
		0.902298,
		0.8716,
		0.836212,
		0.796291,
		0.752097,
		0.703994,
		0.652437,
		0.597967,
		0.541209,
		0.482863,
		0.423701,
		0.364563,
		0.306352,
		0.250035,
		0.196635,
		0.14723,
		0.102952,
		0.064984,
		0.034559,
		0.012955,
		0.001497,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.000384,
		0.00923,
		0.028718,
		0.057479,
		0.094192,
		0.137583,
		0.18643,
		0.239562,
		0.295864,
		0.354276,
		0.413799,
		0.473495,
		0.532493,
		0.589991,
		0.645261,
		0.697651,
		0.746597,
		0.791619,
		0.832336,
		0.868468,
		0.899843,
		0.926409,
		0.948233,
		0.965511,
		0.97857,
		0.987863,
		0.993959,
		0.99752,
		0.999259,
		0.999882,
		0.999997,
		1,
		0.999938,
		0.999493,
		0.998097,
		0.99506,
		0.989663,
		0.981226,
		0.969156,
		0.952968,
		0.932304,
		0.906935,
		0.876762,
		0.841807,
		0.802214,
		0.758237,
		0.710231,
		0.658652,
		0.604044,
		0.547038,
		0.488343,
		0.428742,
		0.369091,
		0.310309,
		0.253383,
		0.199356,
		0.14933,
		0.104462,
		0.065963,
		0.035092,
		0.013159,
		0.001521,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.003444,
		0.018011,
		0.042672,
		0.076051,
		0.116819,
		0.163699,
		0.215467,
		0.270954,
		0.32905,
		0.388702,
		0.448922,
		0.50879,
		0.567455,
		0.624138,
		0.678141,
		0.728851,
		0.775744,
		0.81839,
		0.856465,
		0.889753,
		0.918157,
		0.941704,
		0.960553,
		0.974996,
		0.985457,
		0.992484,
		0.996733,
		0.998923,
		0.999788,
		0.999989,
		1,
		0.999971,
		0.999658,
		0.998528,
		0.995893,
		0.991015,
		0.983187,
		0.971777,
		0.956268,
		0.936266,
		0.911514,
		0.881887,
		0.847388,
		0.808146,
		0.764404,
		0.716513,
		0.664927,
		0.610192,
		0.552945,
		0.493904,
		0.433865,
		0.373698,
		0.31434,
		0.256796,
		0.202132,
		0.151474,
		0.106005,
		0.066963,
		0.035637,
		0.013368,
		0.001546,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.000397,
		0.009529,
		0.029626,
		0.059254,
		0.097025,
		0.141606,
		0.191718,
		0.246136,
		0.303696,
		0.363292,
		0.423884,
		0.484498,
		0.544233,
		0.60226,
		0.657831,
		0.710282,
		0.759042,
		0.803632,
		0.843682,
		0.878929,
		0.909232,
		0.934575,
		0.955074,
		0.970985,
		0.982702,
		0.990749,
		0.995768,
		0.998482,
		0.999647,
		0.99997,
		1,
		0.999989,
		0.999782,
		0.998889,
		0.996633,
		0.992261,
		0.985035,
		0.974289,
		0.959469,
		0.940146,
		0.916031,
		0.886971,
		0.852952,
		0.814083,
		0.770596,
		0.722839,
		0.67126,
		0.61641,
		0.55893,
		0.499548,
		0.439072,
		0.378386,
		0.318445,
		0.260276,
		0.204965,
		0.153664,
		0.107582,
		0.067986,
		0.036195,
		0.013582,
		0.001571,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.003558,
		0.018592,
		0.044016,
		0.078386,
		0.120309,
		0.168448,
		0.22152,
		0.278304,
		0.33764,
		0.398433,
		0.459655,
		0.520351,
		0.57964,
		0.636722,
		0.690882,
		0.741497,
		0.788042,
		0.830094,
		0.867344,
		0.899603,
		0.926809,
		0.949038,
		0.966504,
		0.979567,
		0.988724,
		0.994599,
		0.997915,
		0.999443,
		0.999932,
		1,
		0.999997,
		0.99987,
		0.999187,
		0.997284,
		0.9934,
		0.986769,
		0.976689,
		0.962567,
		0.943938,
		0.92048,
		0.892012,
		0.858494,
		0.820021,
		0.776812,
		0.729206,
		0.677651,
		0.622699,
		0.564995,
		0.505276,
		0.444363,
		0.383156,
		0.322628,
		0.263824,
		0.207857,
		0.155901,
		0.109195,
		0.069032,
		0.036766,
		0.013802,
		0.001597,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.00041,
		0.009842,
		0.030578,
		0.061111,
		0.099986,
		0.145806,
		0.197229,
		0.252975,
		0.311827,
		0.372632,
		0.434306,
		0.495839,
		0.556295,
		0.61482,
		0.670646,
		0.723099,
		0.771599,
		0.815675,
		0.854966,
		0.889236,
		0.918374,
		0.94241,
		0.961517,
		0.976017,
		0.986377,
		0.993199,
		0.997199,
		0.999161,
		0.999866,
		0.999997,
		1,
		0.99993,
		0.999426,
		0.99785,
		0.994434,
		0.988388,
		0.978974,
		0.965558,
		0.947639,
		0.924857,
		0.897002,
		0.864011,
		0.825958,
		0.783049,
		0.735615,
		0.6841,
		0.629057,
		0.571138,
		0.511089,
		0.449741,
		0.38801,
		0.326889,
		0.267443,
		0.210809,
		0.158187,
		0.110843,
		0.070103,
		0.037351,
		0.014026,
		0.001624,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.003677,
		0.019201,
		0.045425,
		0.08083,
		0.123957,
		0.173402,
		0.227824,
		0.285944,
		0.346552,
		0.408505,
		0.470734,
		0.532247,
		0.592135,
		0.649575,
		0.703836,
		0.754286,
		0.800401,
		0.841768,
		0.878098,
		0.909232,
		0.935153,
		0.955987,
		0.972017,
		0.983674,
		0.991538,
		0.99631,
		0.99878,
		0.99976,
		0.999988,
		1,
		0.999967,
		0.999612,
		0.998334,
		0.995364,
		0.989891,
		0.98114,
		0.968439,
		0.951243,
		0.929157,
		0.901939,
		0.869498,
		0.831889,
		0.789303,
		0.742062,
		0.690605,
		0.635486,
		0.577362,
		0.516987,
		0.455206,
		0.39295,
		0.331231,
		0.271135,
		0.213823,
		0.160523,
		0.112529,
		0.071199,
		0.03795,
		0.014256,
		0.001651,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.000424,
		0.010171,
		0.031577,
		0.063057,
		0.103084,
		0.150192,
		0.202976,
		0.260094,
		0.320274,
		0.382312,
		0.44508,
		0.507526,
		0.568684,
		0.627672,
		0.683702,
		0.736089,
		0.78425,
		0.827721,
		0.866157,
		0.89935,
		0.927229,
		0.949876,
		0.967529,
		0.980581,
		0.989583,
		0.995219,
		0.99828,
		0.999599,
		0.999966,
		1,
		0.999987,
		0.999752,
		0.998741,
		0.996194,
		0.991278,
		0.983187,
		0.971204,
		0.954745,
		0.933375,
		0.906817,
		0.874952,
		0.837812,
		0.795573,
		0.748546,
		0.697165,
		0.641985,
		0.583667,
		0.522972,
		0.460761,
		0.397978,
		0.335654,
		0.2749,
		0.2169,
		0.16291,
		0.114253,
		0.07232,
		0.038563,
		0.014492,
		0.001679,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.003803,
		0.019841,
		0.046901,
		0.083389,
		0.127771,
		0.178574,
		0.234393,
		0.29389,
		0.355799,
		0.418929,
		0.482168,
		0.544486,
		0.604944,
		0.662695,
		0.716994,
		0.767202,
		0.812798,
		0.853382,
		0.88869,
		0.918602,
		0.943146,
		0.962515,
		0.977061,
		0.987301,
		0.993899,
		0.997637,
		0.999367,
		0.999923,
		1,
		0.999996,
		0.999852,
		0.999076,
		0.996925,
		0.992549,
		0.985111,
		0.973851,
		0.958141,
		0.937506,
		0.911631,
		0.880367,
		0.843722,
		0.801855,
		0.755064,
		0.70378,
		0.648554,
		0.590052,
		0.529046,
		0.466406,
		0.403095,
		0.340162,
		0.278741,
		0.220042,
		0.165349,
		0.116017,
		0.073468,
		0.039191,
		0.014734,
		0.001708,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000439,
		0.010517,
		0.032625,
		0.065096,
		0.106327,
		0.154777,
		0.208972,
		0.267506,
		0.32905,
		0.392345,
		0.456215,
		0.51957,
		0.581406,
		0.640815,
		0.696992,
		0.749239,
		0.796974,
		0.839741,
		0.877219,
		0.909232,
		0.935757,
		0.956935,
		0.973075,
		0.984657,
		0.992316,
		0.996826,
		0.999046,
		0.999847,
		0.999996,
		1,
		0.99992,
		0.999346,
		0.997561,
		0.993706,
		0.986911,
		0.976376,
		0.961427,
		0.941544,
		0.916375,
		0.885739,
		0.849616,
		0.808146,
		0.761615,
		0.710447,
		0.655192,
		0.596519,
		0.535208,
		0.472143,
		0.408303,
		0.344756,
		0.282661,
		0.223252,
		0.167843,
		0.117821,
		0.074644,
		0.039835,
		0.014982,
		0.001737,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.003934,
		0.020513,
		0.048451,
		0.086071,
		0.131761,
		0.183976,
		0.241242,
		0.302157,
		0.365398,
		0.429721,
		0.493969,
		0.557076,
		0.618068,
		0.676077,
		0.730344,
		0.780225,
		0.825205,
		0.864902,
		0.899083,
		0.927669,
		0.950749,
		0.968585,
		0.981612,
		0.990437,
		0.995817,
		0.998613,
		0.999726,
		0.999986,
		1,
		0.999962,
		0.999557,
		0.998106,
		0.994749,
		0.988586,
		0.978775,
		0.964597,
		0.945484,
		0.921045,
		0.891062,
		0.855488,
		0.814442,
		0.768196,
		0.717165,
		0.661898,
		0.603067,
		0.541461,
		0.477974,
		0.413603,
		0.349439,
		0.28666,
		0.22653,
		0.170393,
		0.119668,
		0.075848,
		0.040494,
		0.015236,
		0.001767,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.000454,
		0.01088,
		0.033726,
		0.067236,
		0.109724,
		0.159571,
		0.21523,
		0.275228,
		0.338171,
		0.402746,
		0.467727,
		0.531979,
		0.594464,
		0.654248,
		0.710506,
		0.762531,
		0.809744,
		0.851704,
		0.888115,
		0.918841,
		0.943915,
		0.963547,
		0.978128,
		0.988228,
		0.99458,
		0.998044,
		0.999542,
		0.999961,
		1,
		0.999985,
		0.999716,
		0.998566,
		0.995681,
		0.990135,
		0.981047,
		0.967646,
		0.949321,
		0.925634,
		0.896331,
		0.861335,
		0.820741,
		0.774805,
		0.723934,
		0.668673,
		0.609698,
		0.547804,
		0.483901,
		0.418999,
		0.354211,
		0.290741,
		0.229879,
		0.173,
		0.121558,
		0.077081,
		0.04117,
		0.015497,
		0.001798,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.004073,
		0.02122,
		0.050078,
		0.088884,
		0.135939,
		0.189623,
		0.248386,
		0.310761,
		0.375363,
		0.440894,
		0.506149,
		0.570021,
		0.631507,
		0.689715,
		0.743871,
		0.793332,
		0.837591,
		0.87629,
		0.909232,
		0.93639,
		0.95792,
		0.974162,
		0.985648,
		0.993081,
		0.997312,
		0.999278,
		0.999911,
		0.999999,
		0.999996,
		0.999831,
		0.998946,
		0.996504,
		0.991558,
		0.983187,
		0.970571,
		0.95305,
		0.930138,
		0.901541,
		0.867151,
		0.827036,
		0.781437,
		0.73075,
		0.675515,
		0.61641,
		0.55424,
		0.489923,
		0.424491,
		0.359076,
		0.294906,
		0.2333,
		0.175666,
		0.123492,
		0.078344,
		0.041863,
		0.015765,
		0.00183,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.000471,
		0.011263,
		0.034883,
		0.069482,
		0.113285,
		0.164588,
		0.221767,
		0.283276,
		0.347655,
		0.413531,
		0.479627,
		0.544762,
		0.607863,
		0.667967,
		0.724231,
		0.775945,
		0.822533,
		0.863572,
		0.898801,
		0.928131,
		0.951659,
		0.969673,
		0.982658,
		0.991285,
		0.996388,
		0.99891,
		0.999825,
		0.999996,
		0.999999,
		0.999908,
		0.999253,
		0.997222,
		0.992856,
		0.985193,
		0.973368,
		0.956664,
		0.934549,
		0.906686,
		0.872931,
		0.833326,
		0.788092,
		0.737611,
		0.682423,
		0.623205,
		0.560768,
		0.496045,
		0.430082,
		0.364035,
		0.299157,
		0.236797,
		0.178394,
		0.125472,
		0.079638,
		0.042574,
		0.016039,
		0.001863,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.004219,
		0.021964,
		0.051789,
		0.091835,
		0.140317,
		0.195527,
		0.255841,
		0.319719,
		0.385711,
		0.452462,
		0.518718,
		0.583329,
		0.645261,
		0.703599,
		0.757559,
		0.806498,
		0.849923,
		0.887506,
		0.919093,
		0.944719,
		0.964615,
		0.979216,
		0.989154,
		0.995238,
		0.998415,
		0.999686,
		0.999984,
		1,
		0.999956,
		0.999493,
		0.997839,
		0.994029,
		0.987065,
		0.976031,
		0.960158,
		0.938862,
		0.91176,
		0.87867,
		0.839605,
		0.794763,
		0.744517,
		0.689396,
		0.630081,
		0.567389,
		0.502265,
		0.435774,
		0.369091,
		0.303496,
		0.24037,
		0.181184,
		0.1275,
		0.080964,
		0.043303,
		0.016321,
		0.001896,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.000488,
		0.011667,
		0.036101,
		0.071843,
		0.11702,
		0.169842,
		0.228598,
		0.291667,
		0.357518,
		0.424716,
		0.491928,
		0.557927,
		0.621603,
		0.681965,
		0.738155,
		0.789457,
		0.835308,
		0.875306,
		0.909232,
		0.937054,
		0.958943,
		0.975277,
		0.986646,
		0.993829,
		0.997765,
		0.999475,
		0.999955,
		1,
		0.999983,
		0.999675,
		0.998361,
		0.995079,
		0.9888,
		0.978558,
		0.963527,
		0.943071,
		0.916757,
		0.884362,
		0.845868,
		0.80145,
		0.751464,
		0.696434,
		0.63704,
		0.574105,
		0.508587,
		0.441568,
		0.374245,
		0.307926,
		0.244022,
		0.184039,
		0.129577,
		0.082324,
		0.044051,
		0.01661,
		0.001931,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.004374,
		0.022747,
		0.053588,
		0.094935,
		0.144906,
		0.201706,
		0.263625,
		0.32905,
		0.39646,
		0.464441,
		0.531686,
		0.597004,
		0.659326,
		0.717717,
		0.771385,
		0.819691,
		0.862161,
		0.898503,
		0.928617,
		0.952607,
		0.970795,
		0.983719,
		0.992122,
		0.996928,
		0.999172,
		0.999898,
		0.999999,
		0.999995,
		0.999805,
		0.998793,
		0.996009,
		0.990397,
		0.980944,
		0.966766,
		0.94717,
		0.92167,
		0.890001,
		0.85211,
		0.808146,
		0.758449,
		0.703534,
		0.64408,
		0.580916,
		0.515011,
		0.447466,
		0.379501,
		0.31245,
		0.247755,
		0.186961,
		0.131705,
		0.083718,
		0.044818,
		0.016907,
		0.001966,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.000506,
		0.012092,
		0.037384,
		0.074325,
		0.120941,
		0.175346,
		0.23574,
		0.300419,
		0.367779,
		0.436317,
		0.504643,
		0.571482,
		0.635684,
		0.696234,
		0.752258,
		0.803041,
		0.848031,
		0.886861,
		0.919359,
		0.94556,
		0.965721,
		0.980326,
		0.990077,
		0.995871,
		0.99875,
		0.999798,
		0.999995,
		0.999999,
		0.999894,
		0.999142,
		0.996822,
		0.991857,
		0.983187,
		0.96987,
		0.951151,
		0.926493,
		0.895582,
		0.858326,
		0.814849,
		0.76547,
		0.710695,
		0.651201,
		0.587822,
		0.521539,
		0.453471,
		0.384859,
		0.317068,
		0.251572,
		0.189951,
		0.133884,
		0.085147,
		0.045605,
		0.017212,
		0.002002,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.004537,
		0.023574,
		0.055483,
		0.098193,
		0.149721,
		0.208175,
		0.271757,
		0.338772,
		0.407628,
		0.476846,
		0.545064,
		0.611048,
		0.673697,
		0.732056,
		0.785326,
		0.832877,
		0.874262,
		0.909232,
		0.937751,
		0.960006,
		0.97642,
		0.987648,
		0.994557,
		0.998182,
		0.999638,
		0.999981,
		1,
		0.99995,
		0.999417,
		0.997523,
		0.993181,
		0.985283,
		0.972832,
		0.95501,
		0.931219,
		0.901097,
		0.864511,
		0.821553,
		0.772523,
		0.717915,
		0.658402,
		0.594824,
		0.528173,
		0.459584,
		0.390324,
		0.321785,
		0.255475,
		0.193012,
		0.136118,
		0.086613,
		0.046414,
		0.017525,
		0.00204,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.000525,
		0.012541,
		0.038736,
		0.076937,
		0.125061,
		0.181117,
		0.243212,
		0.309553,
		0.378457,
		0.448352,
		0.517785,
		0.585432,
		0.650105,
		0.710763,
		0.76652,
		0.816662,
		0.860662,
		0.898188,
		0.929128,
		0.953597,
		0.971951,
		0.984791,
		0.992945,
		0.997435,
		0.999396,
		0.999948,
		1,
		0.99998,
		0.999625,
		0.998117,
		0.994368,
		0.987232,
		0.97565,
		0.958739,
		0.935842,
		0.90654,
		0.870658,
		0.828254,
		0.779606,
		0.725192,
		0.665684,
		0.601923,
		0.534913,
		0.465807,
		0.395897,
		0.326602,
		0.259467,
		0.196146,
		0.138407,
		0.088117,
		0.047243,
		0.017847,
		0.002078,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.004709,
		0.024446,
		0.057479,
		0.101621,
		0.154777,
		0.214952,
		0.280256,
		0.348906,
		0.419232,
		0.48969,
		0.55886,
		0.625463,
		0.688367,
		0.746597,
		0.799352,
		0.846017,
		0.886177,
		0.919639,
		0.94644,
		0.966867,
		0.981456,
		0.990993,
		0.996474,
		0.999046,
		0.999882,
		0.999999,
		0.999995,
		0.999775,
		0.99861,
		0.995423,
		0.989031,
		0.978317,
		0.962333,
		0.940353,
		0.911904,
		0.876762,
		0.834948,
		0.786713,
		0.732524,
		0.673044,
		0.609118,
		0.541761,
		0.472143,
		0.40158,
		0.331523,
		0.263549,
		0.199356,
		0.140754,
		0.08966,
		0.048096,
		0.018178,
		0.002117,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.000546,
		0.013016,
		0.040162,
		0.079688,
		0.129392,
		0.187172,
		0.251033,
		0.319089,
		0.389573,
		0.460836,
		0.531365,
		0.599783,
		0.664862,
		0.725537,
		0.780913,
		0.830285,
		0.873152,
		0.909232,
		0.938482,
		0.961112,
		0.977591,
		0.988652,
		0.99526,
		0.99856,
		0.999767,
		0.999994,
		0.999999,
		0.999877,
		0.999011,
		0.996348,
		0.990679,
		0.980831,
		0.965784,
		0.944746,
		0.917181,
		0.882814,
		0.841627,
		0.793842,
		0.739908,
		0.680482,
		0.61641,
		0.548718,
		0.478594,
		0.407377,
		0.336549,
		0.267725,
		0.202643,
		0.14316,
		0.091244,
		0.048971,
		0.018518,
		0.002158,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.004891,
		0.025367,
		0.059585,
		0.105229,
		0.160089,
		0.222058,
		0.289144,
		0.359472,
		0.431293,
		0.502989,
		0.573084,
		0.64025,
		0.703324,
		0.761317,
		0.813428,
		0.859064,
		0.897854,
		0.929666,
		0.95463,
		0.973143,
		0.985874,
		0.99375,
		0.997904,
		0.999581,
		0.999978,
		1,
		0.999942,
		0.999326,
		0.997147,
		0.992177,
		0.983187,
		0.969087,
		0.949014,
		0.922365,
		0.88881,
		0.848287,
		0.800988,
		0.747342,
		0.687996,
		0.623799,
		0.555785,
		0.48516,
		0.413289,
		0.341684,
		0.271997,
		0.20601,
		0.145627,
		0.09287,
		0.049871,
		0.018867,
		0.0022,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.000567,
		0.013517,
		0.041669,
		0.082589,
		0.133949,
		0.193529,
		0.259225,
		0.32905,
		0.401146,
		0.473788,
		0.545395,
		0.614537,
		0.679948,
		0.740538,
		0.795408,
		0.843868,
		0.885451,
		0.919935,
		0.947363,
		0.968053,
		0.982605,
		0.991898,
		0.997042,
		0.9993,
		0.999939,
		1,
		0.999977,
		0.999565,
		0.997826,
		0.993526,
		0.985382,
		0.972236,
		0.953149,
		0.927446,
		0.89474,
		0.854921,
		0.808146,
		0.754822,
		0.695586,
		0.631286,
		0.562964,
		0.491846,
		0.419319,
		0.34693,
		0.276369,
		0.20946,
		0.148158,
		0.094539,
		0.050795,
		0.019227,
		0.002243,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.005084,
		0.026342,
		0.061808,
		0.109032,
		0.165674,
		0.229511,
		0.298442,
		0.370494,
		0.44383,
		0.516759,
		0.587741,
		0.655405,
		0.718555,
		0.77619,
		0.827516,
		0.87197,
		0.909232,
		0.939252,
		0.962262,
		0.97879,
		0.989654,
		0.995933,
		0.998895,
		0.999863,
		0.999999,
		0.999994,
		0.999739,
		0.998392,
		0.994726,
		0.987414,
		0.975226,
		0.957143,
		0.932418,
		0.900598,
		0.861523,
		0.815311,
		0.762345,
		0.703249,
		0.638868,
		0.570256,
		0.498652,
		0.425471,
		0.352291,
		0.280842,
		0.212995,
		0.150754,
		0.096253,
		0.051745,
		0.019597,
		0.002287,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.00059,
		0.014049,
		0.043262,
		0.08565,
		0.138748,
		0.200209,
		0.267809,
		0.339457,
		0.413199,
		0.487226,
		0.559887,
		0.629697,
		0.695352,
		0.755743,
		0.809968,
		0.85736,
		0.897499,
		0.930235,
		0.95571,
		0.974369,
		0.986964,
		0.994532,
		0.998331,
		0.999729,
		0.999993,
		0.999999,
		0.999857,
		0.998853,
		0.995782,
		0.989282,
		0.978051,
		0.960989,
		0.937272,
		0.906376,
		0.868085,
		0.822478,
		0.769907,
		0.710982,
		0.646547,
		0.577661,
		0.50558,
		0.431746,
		0.357769,
		0.285421,
		0.216617,
		0.153419,
		0.098015,
		0.052722,
		0.019978,
		0.002333,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.005289,
		0.027373,
		0.064157,
		0.113042,
		0.171552,
		0.237335,
		0.308175,
		0.381994,
		0.456863,
		0.531011,
		0.602837,
		0.670922,
		0.73404,
		0.791183,
		0.84157,
		0.884676,
		0.920249,
		0.948332,
		0.969281,
		0.983772,
		0.992789,
		0.997571,
		0.999512,
		0.999974,
		1,
		0.999932,
		0.999216,
		0.996697,
		0.990983,
		0.980706,
		0.964681,
		0.941999,
		0.912065,
		0.874601,
		0.82964,
		0.777504,
		0.718785,
		0.654322,
		0.58518,
		0.512632,
		0.438147,
		0.363368,
		0.290107,
		0.220331,
		0.156153,
		0.099824,
		0.053727,
		0.02037,
		0.00238,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.000615,
		0.014613,
		0.044948,
		0.088884,
		0.143807,
		0.207232,
		0.276811,
		0.350337,
		0.425753,
		0.501165,
		0.574848,
		0.645261,
		0.711062,
		0.771123,
		0.824552,
		0.870708,
		0.909232,
		0.940063,
		0.963459,
		0.980016,
		0.99065,
		0.996571,
		0.999185,
		0.999929,
		1,
		0.999973,
		0.999493,
		0.997478,
		0.99252,
		0.983187,
		0.968209,
		0.946591,
		0.917657,
		0.881063,
		0.836792,
		0.785132,
		0.726654,
		0.662191,
		0.592813,
		0.519811,
		0.444677,
		0.369091,
		0.294906,
		0.224139,
		0.15896,
		0.101684,
		0.054761,
		0.020774,
		0.002428,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.005506,
		0.028466,
		0.066642,
		0.117276,
		0.177742,
		0.245554,
		0.318369,
		0.393997,
		0.470412,
		0.545761,
		0.618377,
		0.686792,
		0.749757,
		0.806258,
		0.855538,
		0.897121,
		0.930835,
		0.956839,
		0.975631,
		0.988059,
		0.995285,
		0.998713,
		0.999839,
		0.999999,
		0.999993,
		0.999695,
		0.99813,
		0.993892,
		0.98549,
		0.971569,
		0.951039,
		0.923141,
		0.887461,
		0.843926,
		0.792785,
		0.734587,
		0.670154,
		0.600563,
		0.527118,
		0.451339,
		0.37494,
		0.299819,
		0.228043,
		0.161842,
		0.103595,
		0.055825,
		0.021189,
		0.002478,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000641,
		0.015211,
		0.046734,
		0.092302,
		0.149143,
		0.214622,
		0.286256,
		0.361715,
		0.438833,
		0.515624,
		0.590287,
		0.661224,
		0.727057,
		0.786646,
		0.839109,
		0.883851,
		0.920582,
		0.949349,
		0.970552,
		0.984953,
		0.993658,
		0.998056,
		0.999682,
		0.999992,
		0.999999,
		0.999833,
		0.998662,
		0.995103,
		0.987613,
		0.974752,
		0.955335,
		0.928511,
		0.893788,
		0.851035,
		0.800458,
		0.742579,
		0.67821,
		0.608428,
		0.534555,
		0.458135,
		0.380919,
		0.30485,
		0.232048,
		0.164802,
		0.105561,
		0.05692,
		0.021618,
		0.00253,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.005737,
		0.029626,
		0.069274,
		0.121749,
		0.184267,
		0.254193,
		0.32905,
		0.406528,
		0.484498,
		0.561021,
		0.63436,
		0.703003,
		0.765677,
		0.82137,
		0.869359,
		0.909232,
		0.940917,
		0.964706,
		0.981267,
		0.991635,
		0.997169,
		0.999429,
		0.99997,
		1,
		0.99992,
		0.999083,
		0.996156,
		0.989554,
		0.977753,
		0.959469,
		0.933754,
		0.900035,
		0.858112,
		0.808146,
		0.750628,
		0.686356,
		0.61641,
		0.542124,
		0.465069,
		0.387032,
		0.310002,
		0.236155,
		0.167843,
		0.107582,
		0.058047,
		0.022059,
		0.002583,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.000668,
		0.015846,
		0.048628,
		0.09592,
		0.154777,
		0.222405,
		0.296172,
		0.373618,
		0.452462,
		0.53062,
		0.606211,
		0.67758,
		0.743316,
		0.80227,
		0.853585,
		0.896717,
		0.931471,
		0.95802,
		0.976929,
		0.989154,
		0.996003,
		0.999046,
		0.999917,
		1,
		0.999969,
		0.999406,
		0.997056,
		0.991312,
		0.980567,
		0.963433,
		0.938862,
		0.906191,
		0.865149,
		0.815842,
		0.758729,
		0.694592,
		0.624508,
		0.549827,
		0.472143,
		0.393283,
		0.315281,
		0.24037,
		0.170967,
		0.109662,
		0.059208,
		0.022514,
		0.002638,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.005983,
		0.030859,
		0.072063,
		0.12648,
		0.191151,
		0.26328,
		0.340247,
		0.419614,
		0.499142,
		0.5768,
		0.650784,
		0.719536,
		0.781763,
		0.836467,
		0.882967,
		0.920935,
		0.950418,
		0.971868,
		0.986146,
		0.994502,
		0.998492,
		0.999811,
		0.999999,
		0.999991,
		0.999641,
		0.997811,
		0.992887,
		0.983187,
		0.967218,
		0.943825,
		0.912247,
		0.872136,
		0.823539,
		0.766878,
		0.702914,
		0.632722,
		0.557664,
		0.47936,
		0.399673,
		0.320688,
		0.244695,
		0.174178,
		0.111802,
		0.060404,
		0.022984,
		0.002694,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000697,
		0.016522,
		0.050639,
		0.099754,
		0.160731,
		0.230606,
		0.306588,
		0.386075,
		0.466665,
		0.546167,
		0.622621,
		0.694316,
		0.759807,
		0.817948,
		0.867913,
		0.909232,
		0.94182,
		0.966004,
		0.982541,
		0.992605,
		0.997721,
		0.999626,
		0.999991,
		0.999999,
		0.999803,
		0.998429,
		0.994282,
		0.98561,
		0.970816,
		0.948631,
		0.918192,
		0.879064,
		0.83123,
		0.77507,
		0.71132,
		0.641052,
		0.565638,
		0.486723,
		0.406208,
		0.326228,
		0.249134,
		0.177478,
		0.114005,
		0.061637,
		0.023468,
		0.002753,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.006244,
		0.032169,
		0.075024,
		0.13149,
		0.198419,
		0.272844,
		0.35199,
		0.433283,
		0.514364,
		0.593109,
		0.667644,
		0.736367,
		0.797974,
		0.851487,
		0.896287,
		0.932145,
		0.959257,
		0.978262,
		0.990244,
		0.996681,
		0.999327,
		0.999964,
		1,
		0.999905,
		0.998921,
		0.995499,
		0.987832,
		0.974219,
		0.953272,
		0.924016,
		0.885923,
		0.838908,
		0.783299,
		0.719808,
		0.649498,
		0.573749,
		0.494234,
		0.412891,
		0.331906,
		0.25369,
		0.180871,
		0.116272,
		0.062907,
		0.023968,
		0.002813,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.000729,
		0.017243,
		0.052777,
		0.103819,
		0.16703,
		0.239256,
		0.317536,
		0.399117,
		0.481465,
		0.56228,
		0.63952,
		0.711414,
		0.776495,
		0.833622,
		0.882021,
		0.921311,
		0.951543,
		0.973229,
		0.987347,
		0.995313,
		0.998875,
		0.999901,
		1,
		0.999963,
		0.999299,
		0.996543,
		0.989852,
		0.97742,
		0.957735,
		0.929706,
		0.892703,
		0.846562,
		0.791559,
		0.728373,
		0.658057,
		0.582,
		0.501896,
		0.419726,
		0.337724,
		0.258369,
		0.18436,
		0.118608,
		0.064217,
		0.024483,
		0.002875,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.006524,
		0.033565,
		0.07817,
		0.136799,
		0.2061,
		0.282919,
		0.364313,
		0.447562,
		0.530184,
		0.609954,
		0.684929,
		0.753466,
		0.814257,
		0.866359,
		0.909232,
		0.942774,
		0.967356,
		0.983837,
		0.993551,
		0.998222,
		0.999776,
		0.999999,
		0.99999,
		0.999575,
		0.997422,
		0.991668,
		0.980411,
		0.962012,
		0.935251,
		0.899393,
		0.854185,
		0.799844,
		0.737013,
		0.66673,
		0.59039,
		0.509713,
		0.426716,
		0.343689,
		0.263174,
		0.18795,
		0.121013,
		0.065569,
		0.025016,
		0.00294,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000762,
		0.018011,
		0.055053,
		0.108136,
		0.1737,
		0.248386,
		0.32905,
		0.412775,
		0.496887,
		0.578973,
		0.656903,
		0.728851,
		0.793332,
		0.849227,
		0.895825,
		0.932861,
		0.960553,
		0.97963,
		0.991323,
		0.997312,
		0.999556,
		0.999989,
		0.999999,
		0.999766,
		0.998145,
		0.993282,
		0.983187,
		0.96609,
		0.94064,
		0.90598,
		0.861767,
		0.808146,
		0.745722,
		0.675515,
		0.598922,
		0.517687,
		0.433865,
		0.349803,
		0.268109,
		0.191643,
		0.123492,
		0.066963,
		0.025566,
		0.003007,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.006822,
		0.035053,
		0.081516,
		0.142431,
		0.214224,
		0.293538,
		0.37725,
		0.462481,
		0.546619,
		0.627338,
		0.702622,
		0.770794,
		0.830552,
		0.881004,
		0.921711,
		0.952728,
		0.974637,
		0.988552,
		0.996084,
		0.999202,
		0.999957,
		1,
		0.999887,
		0.998722,
		0.994696,
		0.985742,
		0.969962,
		0.94586,
		0.912454,
		0.869296,
		0.816458,
		0.754496,
		0.684409,
		0.607595,
		0.52582,
		0.441178,
		0.356071,
		0.273179,
		0.195443,
		0.126047,
		0.068402,
		0.026134,
		0.003075,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.000798,
		0.018832,
		0.057479,
		0.112726,
		0.18077,
		0.258031,
		0.341166,
		0.427082,
		0.512956,
		0.596255,
		0.67476,
		0.746597,
		0.810266,
		0.864686,
		0.909232,
		0.943785,
		0.968764,
		0.985151,
		0.994468,
		0.998665,
		0.999882,
		1,
		0.999956,
		0.999166,
		0.995914,
		0.988073,
		0.973615,
		0.950898,
		0.9188,
		0.876762,
		0.82477,
		0.763329,
		0.69341,
		0.61641,
		0.534114,
		0.448658,
		0.362499,
		0.278389,
		0.199356,
		0.128681,
		0.069887,
		0.026722,
		0.003147,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.007142,
		0.036642,
		0.085081,
		0.148414,
		0.222825,
		0.304739,
		0.390835,
		0.478069,
		0.563688,
		0.645261,
		0.720701,
		0.788304,
		0.846787,
		0.895329,
		0.933622,
		0.961912,
		0.98103,
		0.992386,
		0.997888,
		0.999732,
		0.999998,
		0.999988,
		0.999493,
		0.996943,
		0.990178,
		0.977043,
		0.955742,
		0.925006,
		0.884152,
		0.833075,
		0.772216,
		0.702516,
		0.625367,
		0.542574,
		0.45631,
		0.369091,
		0.283743,
		0.203384,
		0.131398,
		0.071421,
		0.027329,
		0.003221,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000836,
		0.01971,
		0.060068,
		0.117611,
		0.188271,
		0.268228,
		0.353922,
		0.442072,
		0.529695,
		0.614133,
		0.693079,
		0.76461,
		0.827229,
		0.879909,
		0.922139,
		0.953979,
		0.976091,
		0.989754,
		0.996806,
		0.99947,
		0.999987,
		0.999998,
		0.99972,
		0.997793,
		0.992055,
		0.980236,
		0.960379,
		0.931058,
		0.891454,
		0.84136,
		0.781149,
		0.711723,
		0.634466,
		0.5512,
		0.464137,
		0.375851,
		0.289247,
		0.207533,
		0.1342,
		0.073006,
		0.027957,
		0.003297,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.007484,
		0.038342,
		0.088884,
		0.154777,
		0.231941,
		0.316562,
		0.405109,
		0.494356,
		0.581406,
		0.663715,
		0.739133,
		0.805937,
		0.862877,
		0.909232,
		0.944857,
		0.97023,
		0.986479,
		0.995346,
		0.999046,
		0.999949,
		1,
		0.999864,
		0.998474,
		0.993706,
		0.983187,
		0.964797,
		0.936941,
		0.898654,
		0.849616,
		0.79012,
		0.721027,
		0.643707,
		0.559995,
		0.472143,
		0.382786,
		0.294906,
		0.211807,
		0.137092,
		0.074644,
		0.028608,
		0.003376,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.000878,
		0.020651,
		0.062836,
		0.122817,
		0.19624,
		0.279019,
		0.367361,
		0.457781,
		0.547127,
		0.632611,
		0.711836,
		0.782841,
		0.844144,
		0.894796,
		0.934434,
		0.963338,
		0.982461,
		0.993424,
		0.998403,
		0.999858,
		1,
		0.999946,
		0.999001,
		0.995135,
		0.98589,
		0.968983,
		0.942641,
		0.905738,
		0.85783,
		0.799123,
		0.730424,
		0.653088,
		0.568961,
		0.480333,
		0.3899,
		0.300724,
		0.21621,
		0.140077,
		0.076337,
		0.029281,
		0.003458,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.007852,
		0.040162,
		0.092945,
		0.161551,
		0.24161,
		0.32905,
		0.42011,
		0.511372,
		0.599783,
		0.68269,
		0.757879,
		0.823621,
		0.878726,
		0.922598,
		0.955299,
		0.977591,
		0.990947,
		0.997473,
		0.999678,
		0.999998,
		0.999985,
		0.999391,
		0.996348,
		0.98834,
		0.972926,
		0.948142,
		0.912691,
		0.86599,
		0.808146,
		0.739908,
		0.662608,
		0.5781,
		0.488711,
		0.397198,
		0.306709,
		0.22075,
		0.14316,
		0.078089,
		0.029978,
		0.003544,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.000922,
		0.021661,
		0.065798,
		0.128373,
		0.204713,
		0.290446,
		0.381525,
		0.474244,
		0.565272,
		0.651685,
		0.731,
		0.801227,
		0.860918,
		0.909232,
		0.945995,
		0.971756,
		0.987816,
		0.996177,
		0.999361,
		0.999984,
		0.999998,
		0.999662,
		0.997353,
		0.990536,
		0.976614,
		0.953429,
		0.919497,
		0.874082,
		0.817181,
		0.749475,
		0.672265,
		0.587413,
		0.497281,
		0.404687,
		0.312866,
		0.225429,
		0.146343,
		0.079901,
		0.0307,
		0.003632,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.008248,
		0.042115,
		0.097288,
		0.168771,
		0.251877,
		0.342249,
		0.43588,
		0.529145,
		0.618827,
		0.702164,
		0.776885,
		0.841272,
		0.89422,
		0.935301,
		0.964833,
		0.983919,
		0.994427,
		0.99885,
		0.999938,
		1,
		0.999836,
		0.998163,
		0.992476,
		0.980037,
		0.958486,
		0.926139,
		0.882091,
		0.826216,
		0.759116,
		0.682058,
		0.596902,
		0.506046,
		0.412372,
		0.3192,
		0.230255,
		0.149633,
		0.081776,
		0.031449,
		0.003724,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.00097,
		0.022747,
		0.068974,
		0.13431,
		0.213733,
		0.302559,
		0.39646,
		0.491497,
		0.584147,
		0.671346,
		0.750529,
		0.819691,
		0.877444,
		0.92309,
		0.956695,
		0.979138,
		0.992122,
		0.998073,
		0.999827,
		1,
		0.999935,
		0.998793,
		0.994162,
		0.983187,
		0.963299,
		0.9326,
		0.890001,
		0.835238,
		0.768825,
		0.691982,
		0.606567,
		0.515011,
		0.420258,
		0.325719,
		0.235233,
		0.153034,
		0.083718,
		0.032226,
		0.003819,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.008674,
		0.044213,
		0.101941,
		0.176478,
		0.26279,
		0.356209,
		0.452462,
		0.547701,
		0.638541,
		0.722107,
		0.796084,
		0.858788,
		0.909232,
		0.947206,
		0.973345,
		0.989154,
		0.99695,
		0.999608,
		0.999998,
		0.999982,
		0.999261,
		0.995601,
		0.986055,
		0.967852,
		0.938862,
		0.897796,
		0.844236,
		0.778592,
		0.702034,
		0.61641,
		0.52418,
		0.428351,
		0.332429,
		0.24037,
		0.156549,
		0.085729,
		0.033031,
		0.003918,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.001021,
		0.023917,
		0.072383,
		0.140662,
		0.223347,
		0.315408,
		0.412216,
		0.509577,
		0.603767,
		0.691577,
		0.770369,
		0.838141,
		0.893596,
		0.936229,
		0.966404,
		0.985399,
		0.995385,
		0.999223,
		0.999981,
		0.999997,
		0.999588,
		0.996799,
		0.988638,
		0.97213,
		0.944907,
		0.905457,
		0.853193,
		0.788409,
		0.71221,
		0.62643,
		0.533556,
		0.436658,
		0.339336,
		0.245671,
		0.160186,
		0.087813,
		0.033867,
		0.00402,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.009134,
		0.046472,
		0.106933,
		0.184715,
		0.2744,
		0.370984,
		0.469901,
		0.567067,
		0.65892,
		0.742476,
		0.815394,
		0.87605,
		0.92362,
		0.958172,
		0.980727,
		0.99327,
		0.9986,
		0.999924,
		1,
		0.999799,
		0.997769,
		0.990931,
		0.97612,
		0.950715,
		0.912966,
		0.862095,
		0.798264,
		0.722505,
		0.636628,
		0.543144,
		0.445185,
		0.346449,
		0.251144,
		0.163949,
		0.089974,
		0.034735,
		0.004127,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.001077,
		0.025179,
		0.076051,
		0.14747,
		0.233606,
		0.32905,
		0.428844,
		0.528519,
		0.624138,
		0.712348,
		0.790449,
		0.856465,
		0.909232,
		0.948498,
		0.974996,
		0.990485,
		0.997655,
		0.999788,
		1,
		0.99992,
		0.998528,
		0.992935,
		0.97981,
		0.956268,
		0.920303,
		0.870925,
		0.808146,
		0.732912,
		0.647002,
		0.552945,
		0.453937,
		0.353774,
		0.256796,
		0.167843,
		0.092214,
		0.035637,
		0.004238,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.009632,
		0.048908,
		0.112297,
		0.193529,
		0.286766,
		0.38663,
		0.488241,
		0.587261,
		0.679948,
		0.763216,
		0.834714,
		0.892918,
		0.937225,
		0.968053,
		0.986896,
		0.996286,
		0.999519,
		0.999997,
		0.999977,
		0.999094,
		0.994653,
		0.983187,
		0.961546,
		0.927446,
		0.879663,
		0.818041,
		0.743424,
		0.657551,
		0.562964,
		0.462921,
		0.361318,
		0.262634,
		0.171875,
		0.094539,
		0.036574,
		0.004354,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.001138,
		0.026543,
		0.080003,
		0.154777,
		0.244566,
		0.343545,
		0.446396,
		0.548355,
		0.645261,
		0.733619,
		0.81068,
		0.874529,
		0.924192,
		0.959736,
		0.982358,
		0.994377,
		0.999046,
		0.999976,
		0.999997,
		0.999493,
		0.996092,
		0.986243,
		0.966531,
		0.934374,
		0.888291,
		0.827935,
		0.754033,
		0.668272,
		0.573204,
		0.472143,
		0.369091,
		0.268666,
		0.176052,
		0.096952,
		0.037549,
		0.004475,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.010171,
		0.051539,
		0.118071,
		0.202976,
		0.299949,
		0.403207,
		0.507526,
		0.608301,
		0.701601,
		0.78425,
		0.85392,
		0.909232,
		0.949876,
		0.976711,
		0.991798,
		0.99828,
		0.999906,
		1,
		0.999752,
		0.997265,
		0.988971,
		0.971204,
		0.941064,
		0.896786,
		0.837812,
		0.764729,
		0.679164,
		0.583667,
		0.48161,
		0.377099,
		0.2749,
		0.18038,
		0.099458,
		0.038563,
		0.0046,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.001204,
		0.028021,
		0.084269,
		0.162631,
		0.25629,
		0.358959,
		0.464929,
		0.569117,
		0.667127,
		0.755329,
		0.830949,
		0.892179,
		0.938296,
		0.969784,
		0.988401,
		0.997117,
		0.999738,
		1,
		0.9999,
		0.998186,
		0.991369,
		0.975548,
		0.947493,
		0.905127,
		0.847654,
		0.775502,
		0.690221,
		0.594354,
		0.491327,
		0.385352,
		0.281345,
		0.184867,
		0.102062,
		0.039618,
		0.004731,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.010757,
		0.054388,
		0.124297,
		0.213114,
		0.314019,
		0.420779,
		0.527802,
		0.630193,
		0.723836,
		0.805484,
		0.872864,
		0.924811,
		0.961395,
		0.984024,
		0.995431,
		0.999403,
		0.999996,
		0.999972,
		0.998879,
		0.993437,
		0.979547,
		0.953636,
		0.913289,
		0.85744,
		0.786339,
		0.701439,
		0.605268,
		0.501302,
		0.393858,
		0.288011,
		0.18952,
		0.104768,
		0.040718,
		0.004868,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.001275,
		0.029626,
		0.088884,
		0.171089,
		0.268846,
		0.375363,
		0.484498,
		0.590828,
		0.689715,
		0.777401,
		0.851122,
		0.909232,
		0.951351,
		0.978488,
		0.993081,
		0.998815,
		0.99997,
		0.999996,
		0.999369,
		0.99518,
		0.983187,
		0.959469,
		0.921246,
		0.867151,
		0.797225,
		0.712812,
		0.61641,
		0.511539,
		0.402626,
		0.294906,
		0.194346,
		0.107582,
		0.041863,
		0.00501,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.011395,
		0.057479,
		0.131022,
		0.22401,
		0.32905,
		0.439414,
		0.549108,
		0.652936,
		0.746597,
		0.826794,
		0.89137,
		0.939451,
		0.971601,
		0.989904,
		0.997863,
		0.999882,
		1,
		0.999689,
		0.996609,
		0.986456,
		0.964969,
		0.928972,
		0.876762,
		0.808146,
		0.724333,
		0.62778,
		0.522046,
		0.411665,
		0.30204,
		0.199356,
		0.110511,
		0.043058,
		0.005159,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.001354,
		0.031373,
		0.093885,
		0.180211,
		0.282312,
		0.392831,
		0.505161,
		0.613507,
		0.712986,
		0.799732,
		0.871032,
		0.925483,
		0.963156,
		0.985719,
		0.996415,
		0.999671,
		1,
		0.999875,
		0.99774,
		0.989348,
		0.970112,
		0.936437,
		0.886248,
		0.819083,
		0.735993,
		0.639377,
		0.532828,
		0.420985,
		0.309424,
		0.204557,
		0.113559,
		0.044304,
		0.005315,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.012092,
		0.06084,
		0.138302,
		0.23574,
		0.345123,
		0.459182,
		0.571482,
		0.676513,
		0.769799,
		0.848031,
		0.909232,
		0.952932,
		0.980326,
		0.994317,
		0.999251,
		0.999995,
		0.999964,
		0.998595,
		0.991857,
		0.974876,
		0.943612,
		0.895582,
		0.830017,
		0.747781,
		0.651201,
		0.54389,
		0.430596,
		0.317068,
		0.209959,
		0.116734,
		0.045605,
		0.005478,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.00144,
		0.033279,
		0.099316,
		0.190067,
		0.29677,
		0.411446,
		0.526971,
		0.63716,
		0.736881,
		0.822187,
		0.89048,
		0.940699,
		0.973507,
		0.99139,
		0.998511,
		0.999962,
		0.999995,
		0.999205,
		0.993987,
		0.97924,
		0.950468,
		0.904734,
		0.840925,
		0.759685,
		0.663248,
		0.555238,
		0.440506,
		0.324985,
		0.215573,
		0.120042,
		0.046964,
		0.005649,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.012854,
		0.064503,
		0.146196,
		0.248386,
		0.362328,
		0.480153,
		0.594955,
		0.700893,
		0.793332,
		0.869007,
		0.926215,
		0.965026,
		0.987433,
		0.997312,
		0.99985,
		1,
		0.999606,
		0.995745,
		0.983187,
		0.956972,
		0.913673,
		0.851782,
		0.771692,
		0.675515,
		0.566876,
		0.450728,
		0.333186,
		0.221409,
		0.123492,
		0.048385,
		0.005827,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.001534,
		0.035363,
		0.105229,
		0.200736,
		0.312315,
		0.431293,
		0.549981,
		0.661781,
		0.761316,
		0.844598,
		0.909231,
		0.954629,
		0.982219,
		0.995487,
		0.99958,
		0.999998,
		0.999839,
		0.997146,
		0.9867,
		0.963094,
		0.922363,
		0.862561,
		0.783783,
		0.687995,
		0.578809,
		0.461269,
		0.341684,
		0.227479,
		0.127091,
		0.04987,
		0.006014,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.013691,
		0.068505,
		0.154774,
		0.262037,
		0.380753,
		0.502393,
		0.619534,
		0.726004,
		0.817037,
		0.889481,
		0.942036,
		0.975486,
		0.992826,
		0.999028,
		0.999976,
		0.999936,
		0.998198,
		0.989757,
		0.968789,
		0.930757,
		0.873218,
		0.795928,
		0.700674,
		0.591031,
		0.472135,
		0.350485,
		0.233791,
		0.130845,
		0.051425,
		0.006211,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.001637,
		0.037644,
		0.111669,
		0.212284,
		0.329012,
		0.452411,
		0.574175,
		0.687268,
		0.786085,
		0.86666,
		0.926912,
		0.966903,
		0.989042,
		0.997988,
		0.999837,
		0.99988,
		0.998871,
		0.99229,
		0.973965,
		0.938756,
		0.883665,
		0.808054,
		0.713495,
		0.603506,
		0.483303,
		0.359581,
		0.240343,
		0.134756,
		0.053048,
		0.006416,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.014606,
		0.072856,
		0.164044,
		0.276679,
		0.400334,
		0.525747,
		0.644953,
		0.751442,
		0.840368,
		0.908799,
		0.956001,
		0.983695,
		0.996096,
		0.99933,
		0.999523,
		0.999017,
		0.994116,
		0.97841,
		0.94614,
		0.893693,
		0.819979,
		0.726308,
		0.616116,
		0.494688,
		0.368915,
		0.247099,
		0.138805,
		0.054735,
		0.00663,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.001749,
		0.040101,
		0.118558,
		0.224539,
		0.346557,
		0.474322,
		0.598868,
		0.712711,
		0.810056,
		0.887051,
		0.942087,
		0.9761,
		0.992725,
		0.997934,
		0.998474,
		0.998268,
		0.994828,
		0.981687,
		0.952465,
		0.902878,
		0.831316,
		0.73878,
		0.628589,
		0.50608,
		0.378334,
		0.253963,
		0.142941,
		0.056465,
		0.006851,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.015567,
		0.077397,
		0.173627,
		0.291638,
		0.420042,
		0.548801,
		0.66939,
		0.774998,
		0.860781,
		0.924176,
		0.965238,
		0.986889,
		0.994761,
		0.995982,
		0.99593,
		0.993699,
		0.983029,
		0.956957,
		0.910471,
		0.841376,
		0.750302,
		0.640413,
		0.517072,
		0.38754,
		0.260738,
		0.147054,
		0.058196,
		0.007072,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.001862,
		0.042546,
		0.125331,
		0.236406,
		0.36322,
		0.494607,
		0.620939,
		0.734326,
		0.828867,
		0.900971,
		0.949718,
		0.977186,
		0.988479,
		0.99085,
		0.990906,
		0.989607,
		0.981267,
		0.95843,
		0.91532,
		0.849081,
		0.75991,
		0.650766,
		0.527004,
		0.396042,
		0.267095,
		0.15096,
		0.059855,
		0.007286,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.01645,
		0.081496,
		0.182095,
		0.304496,
		0.436364,
		0.566923,
		0.687158,
		0.790059,
		0.870942,
		0.927827,
		0.961826,
		0.977327,
		0.981438,
		0.981687,
		0.981036,
		0.974831,
		0.955287,
		0.915851,
		0.852944,
		0.766258,
		0.658487,
		0.534931,
		0.403127,
		0.272552,
		0.154387,
		0.061335,
		0.00748,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.001953,
		0.044463,
		0.130464,
		0.245006,
		0.374572,
		0.507231,
		0.632819,
		0.743188,
		0.832516,
		0.897699,
		0.938773,
		0.95923,
		0.965756,
		0.966443,
		0.966179,
		0.961852,
		0.945607,
		0.910151,
		0.851131,
		0.767663,
		0.662108,
		0.539646,
		0.407876,
		0.276483,
		0.156977,
		0.062493,
		0.007635,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.016988,
		0.083836,
		0.186511,
		0.310362,
		0.442328,
		0.57112,
		0.687459,
		0.784374,
		0.857596,
		0.906002,
		0.932032,
		0.941677,
		0.943198,
		0.943133,
		0.940363,
		0.92735,
		0.896157,
		0.841631,
		0.76225,
		0.659966,
		0.539764,
		0.409225,
		0.278157,
		0.158311,
		0.06316,
		0.007731,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.001983,
		0.04498,
		0.131419,
		0.245614,
		0.373469,
		0.502643,
		0.622781,
		0.7258,
		0.806267,
		0.861859,
		0.893836,
		0.907247,
		0.910115,
		0.910182,
		0.908605,
		0.898653,
		0.871933,
		0.822512,
		0.748179,
		0.650402,
		0.533882,
		0.406079,
		0.276811,
		0.157944,
		0.063155,
		0.007746,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.016812,
		0.08262,
		0.182939,
		0.302792,
		0.428931,
		0.550042,
		0.65702,
		0.743327,
		0.805445,
		0.843385,
		0.861048,
		0.865825,
		0.86612,
		0.865353,
		0.858176,
		0.836021,
		0.792248,
		0.723937,
		0.632006,
		0.520775,
		0.397462,
		0.271753,
		0.155471,
		0.062312,
		0.007659,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.00191,
		0.043137,
		0.125451,
		0.23323,
		0.352525,
		0.471247,
		0.579429,
		0.669547,
		0.736961,
		0.780424,
		0.80255,
		0.809757,
		0.810531,
		0.810233,
		0.805429,
		0.787765,
		0.750038,
		0.68864,
		0.603886,
		0.499619,
		0.382687,
		0.26248,
		0.150584,
		0.060501,
		0.007452,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.015609,
		0.076359,
		0.168202,
		0.276767,
		0.389445,
		0.495623,
		0.586982,
		0.657896,
		0.705931,
		0.732363,
		0.742376,
		0.743988,
		0.743918,
		0.740998,
		0.727561,
		0.696065,
		0.642277,
		0.565889,
		0.470186,
		0.361507,
		0.24878,
		0.143141,
		0.057659,
		0.007118,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.001708,
		0.038407,
		0.111127,
		0.205406,
		0.308422,
		0.409197,
		0.498862,
		0.571013,
		0.622176,
		0.652329,
		0.665273,
		0.668102,
		0.66817,
		0.666607,
		0.656947,
		0.631602,
		0.585835,
		0.518731,
		0.43295,
		0.334211,
		0.230803,
		0.133206,
		0.053801,
		0.006658,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.013323,
		0.064851,
		0.142042,
		0.232205,
		0.324316,
		0.40925,
		0.48009,
		0.532534,
		0.565401,
		0.581082,
		0.585423,
		0.585696,
		0.584989,
		0.578504,
		0.558952,
		0.52127,
		0.463986,
		0.389109,
		0.30164,
		0.209081,
		0.121059,
		0.049033,
		0.006083,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.001391,
		0.031123,
		0.089552,
		0.164471,
		0.24515,
		0.322525,
		0.389462,
		0.441107,
		0.475331,
		0.493212,
		0.499182,
		0.499836,
		0.499583,
		0.495581,
		0.481216,
		0.451312,
		0.403941,
		0.34047,
		0.265117,
		0.184485,
		0.107181,
		0.043541,
		0.005415,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.01025,
		0.04962,
		0.108001,
		0.175283,
		0.242785,
		0.303466,
		0.352213,
		0.386226,
		0.405462,
		0.412943,
		0.414176,
		0.414121,
		0.4119,
		0.401925,
		0.379157,
		0.341342,
		0.289251,
		0.226305,
		0.158129,
		0.092199,
		0.037571,
		0.004685,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.00101,
		0.022482,
		0.064291,
		0.11724,
		0.17332,
		0.22588,
		0.269848,
		0.302048,
		0.321588,
		0.330213,
		0.332146,
		0.332194,
		0.331119,
		0.32464,
		0.308099,
		0.279078,
		0.237836,
		0.187018,
		0.131251,
		0.076818,
		0.031406,
		0.003928,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.006942,
		0.033402,
		0.072196,
		0.116229,
		0.159489,
		0.197224,
		0.226175,
		0.244896,
		0.254094,
		0.256712,
		0.25688,
		0.256446,
		0.252561,
		0.241175,
		0.219874,
		0.188515,
		0.14903,
		0.105078,
		0.061746,
		0.025332,
		0.003178,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.000636,
		0.014062,
		0.039939,
		0.072255,
		0.105836,
		0.136466,
		0.161065,
		0.177929,
		0.18702,
		0.190146,
		0.190498,
		0.190361,
		0.188251,
		0.180892,
		0.166038,
		0.143273,
		0.113912,
		0.080714,
		0.047631,
		0.019612,
		0.002468,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.003985,
		0.019045,
		0.040846,
		0.065163,
		0.088475,
		0.108085,
		0.122288,
		0.130593,
		0.133926,
		0.134491,
		0.134466,
		0.133455,
		0.129044,
		0.119294,
		0.103642,
		0.082904,
		0.059053,
		0.035006,
		0.014469,
		0.001827,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.000329,
		0.007231,
		0.020382,
		0.036547,
		0.052975,
		0.067482,
		0.078559,
		0.085527,
		0.088707,
		0.089443,
		0.089462,
		0.08905,
		0.086641,
		0.080692,
		0.070615,
		0.056853,
		0.040725,
		0.024257,
		0.010067,
		0.001275,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.001801,
		0.008546,
		0.018168,
		0.028687,
		0.038483,
		0.046365,
		0.051666,
		0.054366,
		0.05516,
		0.055213,
		0.055078,
		0.053909,
		0.050596,
		0.04462,
		0.036175,
		0.026068,
		0.015607,
		0.006505,
		0.000827,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000126,
		0.002744,
		0.007669,
		0.013614,
		0.019499,
		0.024496,
		0.028075,
		0.03008,
		0.030794,
		0.030877,
		0.030845,
		0.030358,
		0.02872,
		0.025537,
		0.020858,
		0.015128,
		0.009106,
		0.003813,
		0.000487,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		0.000544,
		0.002558,
		0.005384,
		0.008402,
		0.011116,
		0.013182,
		0.014441,
		0.014966,
		0.015059,
		0.015054,
		0.014891,
		0.014202,
		0.012738,
		0.010488,
		0.007659,
		0.004637,
		0.001951,
		0.00025,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		2.7e-005,
		0.000588,
		0.001626,
		0.002854,
		0.004032,
		0.004985,
		0.005612,
		0.005911,
		0.005983,
		0.005985,
		0.005944,
		0.005716,
		0.005174,
		0.004297,
		0.003161,
		0.001926,
		0.000815,
		0.000105,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		6.8e-005,
		0.000316,
		0.000658,
		0.001013,
		0.001319,
		0.001535,
		0.001651,
		0.001686,
		0.001689,
		0.001682,
		0.001631,
		0.001491,
		0.001249,
		0.000927,
		0.000568,
		0.000242,
		3.1e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		2.6e-005,
		7.1e-005,
		0.000122,
		0.00017,
		0.000206,
		0.000228,
		0.000236,
		0.000237,
		0.000236,
		0.000231,
		0.000213,
		0.00018,
		0.000135,
		8.3e-005,
		3.6e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		3.6e-005,
		8.3e-005,
		0.000135,
		0.00018,
		0.000213,
		0.000231,
		0.000236,
		0.000237,
		0.000236,
		0.000228,
		0.000206,
		0.00017,
		0.000122,
		7.1e-005,
		2.6e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		3.1e-005,
		0.000242,
		0.000568,
		0.000927,
		0.001249,
		0.001491,
		0.001631,
		0.001682,
		0.001689,
		0.001686,
		0.001651,
		0.001535,
		0.001319,
		0.001013,
		0.000658,
		0.000316,
		6.8e-005,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		0.000105,
		0.000815,
		0.001926,
		0.003161,
		0.004297,
		0.005174,
		0.005716,
		0.005944,
		0.005985,
		0.005983,
		0.005911,
		0.005612,
		0.004985,
		0.004032,
		0.002854,
		0.001626,
		0.000588,
		2.7e-005,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.00025,
		0.001951,
		0.004637,
		0.007659,
		0.010488,
		0.012738,
		0.014202,
		0.014891,
		0.015054,
		0.015059,
		0.014966,
		0.014441,
		0.013182,
		0.011116,
		0.008402,
		0.005384,
		0.002558,
		0.000544,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.000487,
		0.003813,
		0.009106,
		0.015128,
		0.020858,
		0.025537,
		0.02872,
		0.030358,
		0.030845,
		0.030877,
		0.030794,
		0.03008,
		0.028075,
		0.024496,
		0.019499,
		0.013614,
		0.007669,
		0.002744,
		0.000126,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000827,
		0.006505,
		0.015607,
		0.026068,
		0.036175,
		0.04462,
		0.050596,
		0.053909,
		0.055078,
		0.055213,
		0.05516,
		0.054366,
		0.051666,
		0.046365,
		0.038483,
		0.028687,
		0.018168,
		0.008546,
		0.001801,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.001275,
		0.010067,
		0.024257,
		0.040725,
		0.056853,
		0.070615,
		0.080692,
		0.086641,
		0.08905,
		0.089462,
		0.089443,
		0.088707,
		0.085527,
		0.078559,
		0.067482,
		0.052975,
		0.036547,
		0.020382,
		0.007231,
		0.000329,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001827,
		0.014469,
		0.035006,
		0.059053,
		0.082904,
		0.103642,
		0.119294,
		0.129044,
		0.133455,
		0.134466,
		0.134491,
		0.133926,
		0.130593,
		0.122288,
		0.108085,
		0.088475,
		0.065163,
		0.040846,
		0.019045,
		0.003985,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.002468,
		0.019612,
		0.047631,
		0.080714,
		0.113912,
		0.143273,
		0.166038,
		0.180892,
		0.188251,
		0.190361,
		0.190498,
		0.190146,
		0.18702,
		0.177929,
		0.161065,
		0.136466,
		0.105836,
		0.072255,
		0.039939,
		0.014062,
		0.000636,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.003178,
		0.025332,
		0.061746,
		0.105078,
		0.14903,
		0.188515,
		0.219874,
		0.241175,
		0.252561,
		0.256446,
		0.25688,
		0.256712,
		0.254094,
		0.244896,
		0.226175,
		0.197224,
		0.159489,
		0.116229,
		0.072196,
		0.033402,
		0.006942,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.003928,
		0.031406,
		0.076818,
		0.131251,
		0.187018,
		0.237836,
		0.279078,
		0.308099,
		0.32464,
		0.331119,
		0.332194,
		0.332146,
		0.330213,
		0.321588,
		0.302048,
		0.269848,
		0.22588,
		0.17332,
		0.11724,
		0.064291,
		0.022482,
		0.00101,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.004685,
		0.037571,
		0.092199,
		0.158129,
		0.226305,
		0.289251,
		0.341342,
		0.379157,
		0.401925,
		0.4119,
		0.414121,
		0.414176,
		0.412943,
		0.405462,
		0.386226,
		0.352213,
		0.303466,
		0.242785,
		0.175283,
		0.108001,
		0.04962,
		0.01025,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.005415,
		0.043541,
		0.107181,
		0.184485,
		0.265117,
		0.34047,
		0.403941,
		0.451312,
		0.481216,
		0.495581,
		0.499583,
		0.499836,
		0.499182,
		0.493212,
		0.475331,
		0.441107,
		0.389462,
		0.322525,
		0.24515,
		0.164471,
		0.089552,
		0.031123,
		0.001391,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.006083,
		0.049033,
		0.121059,
		0.209081,
		0.30164,
		0.389109,
		0.463986,
		0.52127,
		0.558952,
		0.578504,
		0.584989,
		0.585696,
		0.585423,
		0.581082,
		0.565401,
		0.532534,
		0.48009,
		0.40925,
		0.324316,
		0.232205,
		0.142042,
		0.064851,
		0.013323,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.006658,
		0.053801,
		0.133206,
		0.230803,
		0.334211,
		0.43295,
		0.518731,
		0.585835,
		0.631602,
		0.656947,
		0.666607,
		0.66817,
		0.668102,
		0.665273,
		0.652329,
		0.622176,
		0.571013,
		0.498862,
		0.409197,
		0.308422,
		0.205406,
		0.111127,
		0.038407,
		0.001708,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.007118,
		0.057659,
		0.143141,
		0.24878,
		0.361507,
		0.470186,
		0.565889,
		0.642277,
		0.696065,
		0.727561,
		0.740998,
		0.743918,
		0.743988,
		0.742376,
		0.732363,
		0.705931,
		0.657896,
		0.586982,
		0.495623,
		0.389445,
		0.276767,
		0.168202,
		0.076359,
		0.015609,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.007452,
		0.060501,
		0.150584,
		0.26248,
		0.382687,
		0.499619,
		0.603886,
		0.68864,
		0.750038,
		0.787765,
		0.805429,
		0.810233,
		0.810531,
		0.809757,
		0.80255,
		0.780424,
		0.736961,
		0.669547,
		0.579429,
		0.471247,
		0.352525,
		0.23323,
		0.125451,
		0.043137,
		0.00191,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.007659,
		0.062312,
		0.155471,
		0.271753,
		0.397462,
		0.520775,
		0.632006,
		0.723937,
		0.792248,
		0.836021,
		0.858176,
		0.865353,
		0.86612,
		0.865825,
		0.861048,
		0.843385,
		0.805445,
		0.743327,
		0.65702,
		0.550042,
		0.428931,
		0.302792,
		0.182939,
		0.08262,
		0.016812,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.007746,
		0.063155,
		0.157944,
		0.276811,
		0.406079,
		0.533882,
		0.650402,
		0.748179,
		0.822512,
		0.871933,
		0.898653,
		0.908605,
		0.910182,
		0.910115,
		0.907247,
		0.893836,
		0.861859,
		0.806267,
		0.7258,
		0.622781,
		0.502643,
		0.373469,
		0.245614,
		0.131419,
		0.04498,
		0.001983,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.007731,
		0.06316,
		0.158311,
		0.278157,
		0.409225,
		0.539764,
		0.659966,
		0.76225,
		0.841631,
		0.896157,
		0.92735,
		0.940363,
		0.943133,
		0.943198,
		0.941677,
		0.932032,
		0.906002,
		0.857596,
		0.784374,
		0.687459,
		0.57112,
		0.442328,
		0.310362,
		0.186511,
		0.083836,
		0.016988,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.007635,
		0.062493,
		0.156977,
		0.276483,
		0.407876,
		0.539646,
		0.662108,
		0.767663,
		0.851131,
		0.910151,
		0.945607,
		0.961852,
		0.966179,
		0.966443,
		0.965756,
		0.95923,
		0.938773,
		0.897699,
		0.832516,
		0.743188,
		0.632819,
		0.507231,
		0.374572,
		0.245006,
		0.130464,
		0.044463,
		0.001953,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.00748,
		0.061335,
		0.154387,
		0.272552,
		0.403127,
		0.534931,
		0.658487,
		0.766258,
		0.852944,
		0.915851,
		0.955287,
		0.974831,
		0.981036,
		0.981687,
		0.981438,
		0.977327,
		0.961826,
		0.927827,
		0.870942,
		0.790059,
		0.687158,
		0.566923,
		0.436364,
		0.304496,
		0.182095,
		0.081496,
		0.01645,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.007286,
		0.059855,
		0.15096,
		0.267095,
		0.396042,
		0.527004,
		0.650766,
		0.75991,
		0.849081,
		0.91532,
		0.95843,
		0.981267,
		0.989607,
		0.990906,
		0.99085,
		0.988479,
		0.977186,
		0.949718,
		0.900971,
		0.828867,
		0.734326,
		0.620939,
		0.494607,
		0.36322,
		0.236406,
		0.125331,
		0.042546,
		0.001862,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.007072,
		0.058196,
		0.147054,
		0.260738,
		0.38754,
		0.517072,
		0.640413,
		0.750302,
		0.841376,
		0.910471,
		0.956957,
		0.983029,
		0.993699,
		0.99593,
		0.995982,
		0.994761,
		0.986889,
		0.965238,
		0.924176,
		0.860781,
		0.774998,
		0.66939,
		0.548801,
		0.420042,
		0.291638,
		0.173627,
		0.077397,
		0.015567,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.006851,
		0.056465,
		0.142941,
		0.253963,
		0.378334,
		0.50608,
		0.628589,
		0.73878,
		0.831316,
		0.902878,
		0.952465,
		0.981687,
		0.994828,
		0.998268,
		0.998474,
		0.997934,
		0.992725,
		0.9761,
		0.942087,
		0.887051,
		0.810056,
		0.712711,
		0.598868,
		0.474322,
		0.346557,
		0.224539,
		0.118558,
		0.040101,
		0.001749,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.00663,
		0.054735,
		0.138805,
		0.247099,
		0.368915,
		0.494688,
		0.616116,
		0.726308,
		0.819979,
		0.893693,
		0.94614,
		0.97841,
		0.994116,
		0.999017,
		0.999523,
		0.99933,
		0.996096,
		0.983695,
		0.956001,
		0.908799,
		0.840368,
		0.751442,
		0.644953,
		0.525747,
		0.400334,
		0.276679,
		0.164044,
		0.072856,
		0.014606,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.006416,
		0.053048,
		0.134756,
		0.240343,
		0.359581,
		0.483303,
		0.603506,
		0.713495,
		0.808054,
		0.883665,
		0.938756,
		0.973965,
		0.99229,
		0.998871,
		0.99988,
		0.999837,
		0.997988,
		0.989042,
		0.966903,
		0.926912,
		0.86666,
		0.786085,
		0.687268,
		0.574175,
		0.452411,
		0.329012,
		0.212284,
		0.111669,
		0.037644,
		0.001637,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.006211,
		0.051425,
		0.130845,
		0.233791,
		0.350485,
		0.472135,
		0.591031,
		0.700674,
		0.795928,
		0.873218,
		0.930757,
		0.968789,
		0.989757,
		0.998198,
		0.999936,
		0.999976,
		0.999028,
		0.992826,
		0.975486,
		0.942036,
		0.889481,
		0.817037,
		0.726004,
		0.619534,
		0.502393,
		0.380753,
		0.262037,
		0.154774,
		0.068505,
		0.013691,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.006014,
		0.04987,
		0.127091,
		0.227479,
		0.341684,
		0.461269,
		0.578809,
		0.687995,
		0.783783,
		0.862561,
		0.922363,
		0.963094,
		0.9867,
		0.997146,
		0.999839,
		0.999998,
		0.99958,
		0.995487,
		0.982219,
		0.954629,
		0.909231,
		0.844598,
		0.761316,
		0.661781,
		0.549981,
		0.431293,
		0.312315,
		0.200736,
		0.105229,
		0.035363,
		0.001534,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.005827,
		0.048385,
		0.123492,
		0.221409,
		0.333186,
		0.450728,
		0.566876,
		0.675515,
		0.771692,
		0.851782,
		0.913673,
		0.956972,
		0.983187,
		0.995745,
		0.999606,
		1,
		0.99985,
		0.997312,
		0.987433,
		0.965026,
		0.926215,
		0.869007,
		0.793332,
		0.700893,
		0.594955,
		0.480153,
		0.362328,
		0.248386,
		0.146196,
		0.064503,
		0.012854,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.005649,
		0.046964,
		0.120042,
		0.215573,
		0.324985,
		0.440506,
		0.555238,
		0.663248,
		0.759685,
		0.840925,
		0.904734,
		0.950468,
		0.97924,
		0.993987,
		0.999205,
		0.999995,
		0.999962,
		0.998511,
		0.99139,
		0.973507,
		0.940699,
		0.89048,
		0.822187,
		0.736881,
		0.63716,
		0.526971,
		0.411446,
		0.29677,
		0.190067,
		0.099316,
		0.033279,
		0.00144,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.005478,
		0.045605,
		0.116734,
		0.209959,
		0.317068,
		0.430596,
		0.54389,
		0.651201,
		0.747781,
		0.830017,
		0.895582,
		0.943612,
		0.974876,
		0.991857,
		0.998595,
		0.999964,
		0.999995,
		0.999251,
		0.994317,
		0.980326,
		0.952932,
		0.909232,
		0.848031,
		0.769799,
		0.676513,
		0.571482,
		0.459182,
		0.345123,
		0.23574,
		0.138302,
		0.06084,
		0.012092,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.005315,
		0.044304,
		0.113559,
		0.204557,
		0.309424,
		0.420985,
		0.532828,
		0.639377,
		0.735993,
		0.819083,
		0.886248,
		0.936437,
		0.970112,
		0.989348,
		0.99774,
		0.999875,
		1,
		0.999671,
		0.996415,
		0.985719,
		0.963156,
		0.925483,
		0.871032,
		0.799732,
		0.712986,
		0.613507,
		0.505161,
		0.392831,
		0.282312,
		0.180211,
		0.093885,
		0.031373,
		0.001354,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.005159,
		0.043058,
		0.110511,
		0.199356,
		0.30204,
		0.411665,
		0.522046,
		0.62778,
		0.724333,
		0.808146,
		0.876762,
		0.928972,
		0.964969,
		0.986456,
		0.996609,
		0.999689,
		1,
		0.999882,
		0.997863,
		0.989904,
		0.971601,
		0.939451,
		0.89137,
		0.826794,
		0.746597,
		0.652936,
		0.549108,
		0.439414,
		0.32905,
		0.22401,
		0.131022,
		0.057479,
		0.011395,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.00501,
		0.041863,
		0.107582,
		0.194346,
		0.294906,
		0.402626,
		0.511539,
		0.61641,
		0.712812,
		0.797225,
		0.867151,
		0.921246,
		0.959469,
		0.983187,
		0.99518,
		0.999369,
		0.999996,
		0.99997,
		0.998815,
		0.993081,
		0.978488,
		0.951351,
		0.909232,
		0.851122,
		0.777401,
		0.689715,
		0.590828,
		0.484498,
		0.375363,
		0.268846,
		0.171089,
		0.088884,
		0.029626,
		0.001275,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.004868,
		0.040718,
		0.104768,
		0.18952,
		0.288011,
		0.393858,
		0.501302,
		0.605268,
		0.701439,
		0.786339,
		0.85744,
		0.913289,
		0.953636,
		0.979547,
		0.993437,
		0.998879,
		0.999972,
		0.999996,
		0.999403,
		0.995431,
		0.984024,
		0.961395,
		0.924811,
		0.872864,
		0.805484,
		0.723836,
		0.630193,
		0.527802,
		0.420779,
		0.314019,
		0.213114,
		0.124297,
		0.054388,
		0.010757,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.004731,
		0.039618,
		0.102062,
		0.184867,
		0.281345,
		0.385352,
		0.491327,
		0.594354,
		0.690221,
		0.775502,
		0.847654,
		0.905127,
		0.947493,
		0.975548,
		0.991369,
		0.998186,
		0.9999,
		1,
		0.999738,
		0.997117,
		0.988401,
		0.969784,
		0.938296,
		0.892179,
		0.830949,
		0.755329,
		0.667127,
		0.569117,
		0.464929,
		0.358959,
		0.25629,
		0.162631,
		0.084269,
		0.028021,
		0.001204,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.0046,
		0.038563,
		0.099458,
		0.18038,
		0.2749,
		0.377099,
		0.48161,
		0.583667,
		0.679164,
		0.764729,
		0.837812,
		0.896786,
		0.941064,
		0.971204,
		0.988971,
		0.997265,
		0.999752,
		1,
		0.999906,
		0.99828,
		0.991798,
		0.976711,
		0.949876,
		0.909232,
		0.85392,
		0.78425,
		0.701601,
		0.608301,
		0.507526,
		0.403207,
		0.299949,
		0.202976,
		0.118071,
		0.051539,
		0.010171,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.004475,
		0.037549,
		0.096952,
		0.176052,
		0.268666,
		0.369091,
		0.472143,
		0.573204,
		0.668272,
		0.754033,
		0.827935,
		0.888291,
		0.934374,
		0.966531,
		0.986243,
		0.996092,
		0.999493,
		0.999997,
		0.999976,
		0.999046,
		0.994377,
		0.982358,
		0.959736,
		0.924192,
		0.874529,
		0.81068,
		0.733619,
		0.645261,
		0.548355,
		0.446396,
		0.343545,
		0.244566,
		0.154777,
		0.080003,
		0.026543,
		0.001138,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.004354,
		0.036574,
		0.094539,
		0.171875,
		0.262634,
		0.361318,
		0.462921,
		0.562964,
		0.657551,
		0.743424,
		0.818041,
		0.879663,
		0.927446,
		0.961546,
		0.983187,
		0.994653,
		0.999094,
		0.999977,
		0.999997,
		0.999519,
		0.996286,
		0.986896,
		0.968053,
		0.937225,
		0.892918,
		0.834714,
		0.763216,
		0.679948,
		0.587261,
		0.488241,
		0.38663,
		0.286766,
		0.193529,
		0.112297,
		0.048908,
		0.009632,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.004238,
		0.035637,
		0.092214,
		0.167843,
		0.256796,
		0.353774,
		0.453937,
		0.552945,
		0.647002,
		0.732912,
		0.808146,
		0.870925,
		0.920303,
		0.956268,
		0.97981,
		0.992935,
		0.998528,
		0.99992,
		1,
		0.999788,
		0.997655,
		0.990485,
		0.974996,
		0.948498,
		0.909232,
		0.856465,
		0.790449,
		0.712348,
		0.624138,
		0.528519,
		0.428844,
		0.32905,
		0.233606,
		0.14747,
		0.076051,
		0.025179,
		0.001077,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.004127,
		0.034735,
		0.089974,
		0.163949,
		0.251144,
		0.346449,
		0.445185,
		0.543144,
		0.636628,
		0.722505,
		0.798264,
		0.862095,
		0.912966,
		0.950715,
		0.97612,
		0.990931,
		0.997769,
		0.999799,
		1,
		0.999924,
		0.9986,
		0.99327,
		0.980727,
		0.958172,
		0.92362,
		0.87605,
		0.815394,
		0.742476,
		0.65892,
		0.567067,
		0.469901,
		0.370984,
		0.2744,
		0.184715,
		0.106933,
		0.046472,
		0.009134,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.00402,
		0.033867,
		0.087813,
		0.160186,
		0.245671,
		0.339336,
		0.436658,
		0.533556,
		0.62643,
		0.71221,
		0.788409,
		0.853193,
		0.905457,
		0.944907,
		0.97213,
		0.988638,
		0.996799,
		0.999588,
		0.999997,
		0.999981,
		0.999223,
		0.995385,
		0.985399,
		0.966404,
		0.936229,
		0.893596,
		0.838141,
		0.770369,
		0.691577,
		0.603767,
		0.509577,
		0.412216,
		0.315408,
		0.223347,
		0.140662,
		0.072383,
		0.023917,
		0.001021,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.003918,
		0.033031,
		0.085729,
		0.156549,
		0.24037,
		0.332429,
		0.428351,
		0.52418,
		0.61641,
		0.702034,
		0.778592,
		0.844236,
		0.897796,
		0.938862,
		0.967852,
		0.986055,
		0.995601,
		0.999261,
		0.999982,
		0.999998,
		0.999608,
		0.99695,
		0.989154,
		0.973345,
		0.947206,
		0.909232,
		0.858788,
		0.796084,
		0.722107,
		0.638541,
		0.547701,
		0.452462,
		0.356209,
		0.26279,
		0.176478,
		0.101941,
		0.044213,
		0.008674,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.003819,
		0.032226,
		0.083718,
		0.153034,
		0.235233,
		0.325719,
		0.420258,
		0.515011,
		0.606567,
		0.691982,
		0.768825,
		0.835238,
		0.890001,
		0.9326,
		0.963299,
		0.983187,
		0.994162,
		0.998793,
		0.999935,
		1,
		0.999827,
		0.998073,
		0.992122,
		0.979138,
		0.956695,
		0.92309,
		0.877444,
		0.819691,
		0.750529,
		0.671346,
		0.584147,
		0.491497,
		0.39646,
		0.302559,
		0.213733,
		0.13431,
		0.068974,
		0.022747,
		0.00097,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.003724,
		0.031449,
		0.081776,
		0.149633,
		0.230255,
		0.3192,
		0.412372,
		0.506046,
		0.596902,
		0.682058,
		0.759116,
		0.826216,
		0.882091,
		0.926139,
		0.958486,
		0.980037,
		0.992476,
		0.998163,
		0.999836,
		1,
		0.999938,
		0.99885,
		0.994427,
		0.983919,
		0.964833,
		0.935301,
		0.89422,
		0.841272,
		0.776885,
		0.702164,
		0.618827,
		0.529145,
		0.43588,
		0.342249,
		0.251877,
		0.168771,
		0.097288,
		0.042115,
		0.008248,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.003632,
		0.0307,
		0.079901,
		0.146343,
		0.225429,
		0.312866,
		0.404687,
		0.497281,
		0.587413,
		0.672265,
		0.749475,
		0.817181,
		0.874082,
		0.919497,
		0.953429,
		0.976614,
		0.990536,
		0.997353,
		0.999662,
		0.999998,
		0.999984,
		0.999361,
		0.996177,
		0.987816,
		0.971756,
		0.945995,
		0.909232,
		0.860918,
		0.801227,
		0.731,
		0.651685,
		0.565272,
		0.474244,
		0.381525,
		0.290446,
		0.204713,
		0.128373,
		0.065798,
		0.021661,
		0.000922,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.003544,
		0.029978,
		0.078089,
		0.14316,
		0.22075,
		0.306709,
		0.397198,
		0.488711,
		0.5781,
		0.662608,
		0.739908,
		0.808146,
		0.86599,
		0.912691,
		0.948142,
		0.972926,
		0.98834,
		0.996348,
		0.999391,
		0.999985,
		0.999998,
		0.999678,
		0.997473,
		0.990947,
		0.977591,
		0.955299,
		0.922598,
		0.878726,
		0.823621,
		0.757879,
		0.68269,
		0.599783,
		0.511372,
		0.42011,
		0.32905,
		0.24161,
		0.161551,
		0.092945,
		0.040162,
		0.007852,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.003458,
		0.029281,
		0.076337,
		0.140077,
		0.21621,
		0.300724,
		0.3899,
		0.480333,
		0.568961,
		0.653088,
		0.730424,
		0.799123,
		0.85783,
		0.905738,
		0.942641,
		0.968983,
		0.98589,
		0.995135,
		0.999001,
		0.999946,
		1,
		0.999858,
		0.998403,
		0.993424,
		0.982461,
		0.963338,
		0.934434,
		0.894796,
		0.844144,
		0.782841,
		0.711836,
		0.632611,
		0.547127,
		0.457781,
		0.367361,
		0.279019,
		0.19624,
		0.122817,
		0.062836,
		0.020651,
		0.000878,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.003376,
		0.028608,
		0.074644,
		0.137092,
		0.211807,
		0.294906,
		0.382786,
		0.472143,
		0.559995,
		0.643707,
		0.721027,
		0.79012,
		0.849616,
		0.898654,
		0.936941,
		0.964797,
		0.983187,
		0.993706,
		0.998474,
		0.999864,
		1,
		0.999949,
		0.999046,
		0.995346,
		0.986479,
		0.97023,
		0.944857,
		0.909232,
		0.862877,
		0.805937,
		0.739133,
		0.663715,
		0.581406,
		0.494356,
		0.405109,
		0.316562,
		0.231941,
		0.154777,
		0.088884,
		0.038342,
		0.007484,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.003297,
		0.027957,
		0.073006,
		0.1342,
		0.207533,
		0.289247,
		0.375851,
		0.464137,
		0.5512,
		0.634466,
		0.711723,
		0.781149,
		0.84136,
		0.891454,
		0.931058,
		0.960379,
		0.980236,
		0.992055,
		0.997793,
		0.99972,
		0.999998,
		0.999987,
		0.99947,
		0.996806,
		0.989754,
		0.976091,
		0.953979,
		0.922139,
		0.879909,
		0.827229,
		0.76461,
		0.693079,
		0.614133,
		0.529695,
		0.442072,
		0.353922,
		0.268228,
		0.188271,
		0.117611,
		0.060068,
		0.01971,
		0.000836,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.003221,
		0.027329,
		0.071421,
		0.131398,
		0.203384,
		0.283743,
		0.369091,
		0.45631,
		0.542574,
		0.625367,
		0.702516,
		0.772216,
		0.833075,
		0.884152,
		0.925006,
		0.955742,
		0.977043,
		0.990178,
		0.996943,
		0.999493,
		0.999988,
		0.999998,
		0.999732,
		0.997888,
		0.992386,
		0.98103,
		0.961912,
		0.933622,
		0.895329,
		0.846787,
		0.788304,
		0.720701,
		0.645261,
		0.563688,
		0.478069,
		0.390835,
		0.304739,
		0.222825,
		0.148414,
		0.085081,
		0.036642,
		0.007142,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.003147,
		0.026722,
		0.069887,
		0.128681,
		0.199356,
		0.278389,
		0.362499,
		0.448658,
		0.534114,
		0.61641,
		0.69341,
		0.763329,
		0.82477,
		0.876762,
		0.9188,
		0.950898,
		0.973615,
		0.988073,
		0.995914,
		0.999166,
		0.999956,
		1,
		0.999882,
		0.998665,
		0.994468,
		0.985151,
		0.968764,
		0.943785,
		0.909232,
		0.864686,
		0.810266,
		0.746597,
		0.67476,
		0.596255,
		0.512956,
		0.427082,
		0.341166,
		0.258031,
		0.18077,
		0.112726,
		0.057479,
		0.018832,
		0.000798,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.003075,
		0.026134,
		0.068402,
		0.126047,
		0.195443,
		0.273179,
		0.356071,
		0.441178,
		0.52582,
		0.607595,
		0.684409,
		0.754496,
		0.816458,
		0.869296,
		0.912454,
		0.94586,
		0.969962,
		0.985742,
		0.994696,
		0.998722,
		0.999887,
		1,
		0.999957,
		0.999202,
		0.996084,
		0.988552,
		0.974637,
		0.952728,
		0.921711,
		0.881004,
		0.830552,
		0.770794,
		0.702622,
		0.627338,
		0.546619,
		0.462481,
		0.37725,
		0.293538,
		0.214224,
		0.142431,
		0.081516,
		0.035053,
		0.006822,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.003007,
		0.025566,
		0.066963,
		0.123492,
		0.191643,
		0.268109,
		0.349803,
		0.433865,
		0.517687,
		0.598922,
		0.675515,
		0.745722,
		0.808146,
		0.861767,
		0.90598,
		0.94064,
		0.96609,
		0.983187,
		0.993282,
		0.998145,
		0.999766,
		0.999999,
		0.999989,
		0.999556,
		0.997312,
		0.991323,
		0.97963,
		0.960553,
		0.932861,
		0.895825,
		0.849227,
		0.793332,
		0.728851,
		0.656903,
		0.578973,
		0.496887,
		0.412775,
		0.32905,
		0.248386,
		0.1737,
		0.108136,
		0.055053,
		0.018011,
		0.000762,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.00294,
		0.025016,
		0.065569,
		0.121013,
		0.18795,
		0.263174,
		0.343689,
		0.426716,
		0.509713,
		0.59039,
		0.66673,
		0.737013,
		0.799844,
		0.854185,
		0.899393,
		0.935251,
		0.962012,
		0.980411,
		0.991668,
		0.997422,
		0.999575,
		0.99999,
		0.999999,
		0.999776,
		0.998222,
		0.993551,
		0.983837,
		0.967356,
		0.942774,
		0.909232,
		0.866359,
		0.814257,
		0.753466,
		0.684929,
		0.609954,
		0.530184,
		0.447562,
		0.364313,
		0.282919,
		0.2061,
		0.136799,
		0.07817,
		0.033565,
		0.006524,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.002875,
		0.024483,
		0.064217,
		0.118608,
		0.18436,
		0.258369,
		0.337724,
		0.419726,
		0.501896,
		0.582,
		0.658057,
		0.728373,
		0.791559,
		0.846562,
		0.892703,
		0.929706,
		0.957735,
		0.97742,
		0.989852,
		0.996543,
		0.999299,
		0.999963,
		1,
		0.999901,
		0.998875,
		0.995313,
		0.987347,
		0.973229,
		0.951543,
		0.921311,
		0.882021,
		0.833622,
		0.776495,
		0.711414,
		0.63952,
		0.56228,
		0.481465,
		0.399117,
		0.317536,
		0.239256,
		0.16703,
		0.103819,
		0.052777,
		0.017243,
		0.000729,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.002813,
		0.023968,
		0.062907,
		0.116272,
		0.180871,
		0.25369,
		0.331906,
		0.412891,
		0.494234,
		0.573749,
		0.649498,
		0.719808,
		0.783299,
		0.838908,
		0.885923,
		0.924016,
		0.953272,
		0.974219,
		0.987832,
		0.995499,
		0.998921,
		0.999905,
		1,
		0.999964,
		0.999327,
		0.996681,
		0.990244,
		0.978262,
		0.959257,
		0.932145,
		0.896287,
		0.851487,
		0.797974,
		0.736367,
		0.667644,
		0.593109,
		0.514364,
		0.433283,
		0.35199,
		0.272844,
		0.198419,
		0.13149,
		0.075024,
		0.032169,
		0.006244,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.002753,
		0.023468,
		0.061637,
		0.114005,
		0.177478,
		0.249134,
		0.326228,
		0.406208,
		0.486723,
		0.565638,
		0.641052,
		0.71132,
		0.77507,
		0.83123,
		0.879064,
		0.918192,
		0.948631,
		0.970816,
		0.98561,
		0.994282,
		0.998429,
		0.999803,
		0.999999,
		0.999991,
		0.999626,
		0.997721,
		0.992605,
		0.982541,
		0.966004,
		0.94182,
		0.909232,
		0.867913,
		0.817948,
		0.759807,
		0.694316,
		0.622621,
		0.546167,
		0.466665,
		0.386075,
		0.306588,
		0.230606,
		0.160731,
		0.099754,
		0.050639,
		0.016522,
		0.000697,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.002694,
		0.022984,
		0.060404,
		0.111802,
		0.174178,
		0.244695,
		0.320688,
		0.399673,
		0.47936,
		0.557664,
		0.632722,
		0.702914,
		0.766878,
		0.823539,
		0.872136,
		0.912247,
		0.943825,
		0.967218,
		0.983187,
		0.992887,
		0.997811,
		0.999641,
		0.999991,
		0.999999,
		0.999811,
		0.998492,
		0.994502,
		0.986146,
		0.971868,
		0.950418,
		0.920935,
		0.882967,
		0.836467,
		0.781763,
		0.719536,
		0.650784,
		0.5768,
		0.499142,
		0.419614,
		0.340247,
		0.26328,
		0.191151,
		0.12648,
		0.072063,
		0.030859,
		0.005983,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.002638,
		0.022514,
		0.059208,
		0.109662,
		0.170967,
		0.24037,
		0.315281,
		0.393283,
		0.472143,
		0.549827,
		0.624508,
		0.694592,
		0.758729,
		0.815842,
		0.865149,
		0.906191,
		0.938862,
		0.963433,
		0.980567,
		0.991312,
		0.997056,
		0.999406,
		0.999969,
		1,
		0.999917,
		0.999046,
		0.996003,
		0.989154,
		0.976929,
		0.95802,
		0.931471,
		0.896717,
		0.853585,
		0.80227,
		0.743316,
		0.67758,
		0.606211,
		0.53062,
		0.452462,
		0.373618,
		0.296172,
		0.222405,
		0.154777,
		0.09592,
		0.048628,
		0.015846,
		0.000668,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.002583,
		0.022059,
		0.058047,
		0.107582,
		0.167843,
		0.236155,
		0.310002,
		0.387032,
		0.465069,
		0.542124,
		0.61641,
		0.686356,
		0.750628,
		0.808146,
		0.858112,
		0.900035,
		0.933754,
		0.959469,
		0.977753,
		0.989554,
		0.996156,
		0.999083,
		0.99992,
		1,
		0.99997,
		0.999429,
		0.997169,
		0.991635,
		0.981267,
		0.964706,
		0.940917,
		0.909232,
		0.869359,
		0.82137,
		0.765677,
		0.703003,
		0.63436,
		0.561021,
		0.484498,
		0.406528,
		0.32905,
		0.254193,
		0.184267,
		0.121749,
		0.069274,
		0.029626,
		0.005737,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.00253,
		0.021618,
		0.05692,
		0.105561,
		0.164802,
		0.232048,
		0.30485,
		0.380919,
		0.458135,
		0.534555,
		0.608428,
		0.67821,
		0.742579,
		0.800458,
		0.851035,
		0.893788,
		0.928511,
		0.955335,
		0.974752,
		0.987613,
		0.995103,
		0.998662,
		0.999833,
		0.999999,
		0.999992,
		0.999682,
		0.998056,
		0.993658,
		0.984953,
		0.970552,
		0.949349,
		0.920582,
		0.883851,
		0.839109,
		0.786646,
		0.727057,
		0.661224,
		0.590287,
		0.515624,
		0.438833,
		0.361715,
		0.286256,
		0.214622,
		0.149143,
		0.092302,
		0.046734,
		0.015211,
		0.000641,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.002478,
		0.021189,
		0.055825,
		0.103595,
		0.161842,
		0.228043,
		0.299819,
		0.37494,
		0.451339,
		0.527118,
		0.600563,
		0.670154,
		0.734587,
		0.792785,
		0.843926,
		0.887461,
		0.923141,
		0.951039,
		0.971569,
		0.98549,
		0.993892,
		0.99813,
		0.999695,
		0.999993,
		0.999999,
		0.999839,
		0.998713,
		0.995285,
		0.988059,
		0.975631,
		0.956839,
		0.930835,
		0.897121,
		0.855538,
		0.806258,
		0.749757,
		0.686792,
		0.618377,
		0.545761,
		0.470412,
		0.393997,
		0.318369,
		0.245554,
		0.177742,
		0.117276,
		0.066642,
		0.028466,
		0.005506,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.002428,
		0.020774,
		0.054761,
		0.101684,
		0.15896,
		0.224139,
		0.294906,
		0.369091,
		0.444677,
		0.519811,
		0.592813,
		0.662191,
		0.726654,
		0.785132,
		0.836792,
		0.881063,
		0.917657,
		0.946591,
		0.968209,
		0.983187,
		0.99252,
		0.997478,
		0.999493,
		0.999973,
		1,
		0.999929,
		0.999185,
		0.996571,
		0.99065,
		0.980016,
		0.963459,
		0.940063,
		0.909232,
		0.870708,
		0.824552,
		0.771123,
		0.711062,
		0.645261,
		0.574848,
		0.501165,
		0.425753,
		0.350337,
		0.276811,
		0.207232,
		0.143807,
		0.088884,
		0.044948,
		0.014613,
		0.000615,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.00238,
		0.02037,
		0.053727,
		0.099824,
		0.156153,
		0.220331,
		0.290107,
		0.363368,
		0.438147,
		0.512632,
		0.58518,
		0.654322,
		0.718785,
		0.777504,
		0.82964,
		0.874601,
		0.912065,
		0.941999,
		0.964681,
		0.980706,
		0.990983,
		0.996697,
		0.999216,
		0.999932,
		1,
		0.999974,
		0.999512,
		0.997571,
		0.992789,
		0.983772,
		0.969281,
		0.948332,
		0.920249,
		0.884676,
		0.84157,
		0.791183,
		0.73404,
		0.670922,
		0.602837,
		0.531011,
		0.456863,
		0.381994,
		0.308175,
		0.237335,
		0.171552,
		0.113042,
		0.064157,
		0.027373,
		0.005289,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.002333,
		0.019978,
		0.052722,
		0.098015,
		0.153419,
		0.216617,
		0.285421,
		0.357769,
		0.431746,
		0.50558,
		0.577661,
		0.646547,
		0.710982,
		0.769907,
		0.822478,
		0.868085,
		0.906376,
		0.937272,
		0.960989,
		0.978051,
		0.989282,
		0.995782,
		0.998853,
		0.999857,
		0.999999,
		0.999993,
		0.999729,
		0.998331,
		0.994532,
		0.986964,
		0.974369,
		0.95571,
		0.930235,
		0.897499,
		0.85736,
		0.809968,
		0.755743,
		0.695352,
		0.629697,
		0.559887,
		0.487226,
		0.413199,
		0.339457,
		0.267809,
		0.200209,
		0.138748,
		0.08565,
		0.043262,
		0.014049,
		0.00059,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.002287,
		0.019597,
		0.051745,
		0.096253,
		0.150754,
		0.212995,
		0.280842,
		0.352291,
		0.425471,
		0.498652,
		0.570256,
		0.638868,
		0.703249,
		0.762345,
		0.815311,
		0.861523,
		0.900598,
		0.932418,
		0.957143,
		0.975226,
		0.987414,
		0.994726,
		0.998392,
		0.999739,
		0.999994,
		0.999999,
		0.999863,
		0.998895,
		0.995933,
		0.989654,
		0.97879,
		0.962262,
		0.939252,
		0.909232,
		0.87197,
		0.827516,
		0.77619,
		0.718555,
		0.655405,
		0.587741,
		0.516759,
		0.44383,
		0.370494,
		0.298442,
		0.229511,
		0.165674,
		0.109032,
		0.061808,
		0.026342,
		0.005084,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.002243,
		0.019227,
		0.050795,
		0.094539,
		0.148158,
		0.20946,
		0.276369,
		0.34693,
		0.419319,
		0.491846,
		0.562964,
		0.631286,
		0.695586,
		0.754822,
		0.808146,
		0.854921,
		0.89474,
		0.927446,
		0.953149,
		0.972236,
		0.985382,
		0.993526,
		0.997826,
		0.999565,
		0.999977,
		1,
		0.999939,
		0.9993,
		0.997042,
		0.991898,
		0.982605,
		0.968053,
		0.947363,
		0.919935,
		0.885451,
		0.843868,
		0.795408,
		0.740538,
		0.679948,
		0.614537,
		0.545395,
		0.473788,
		0.401146,
		0.32905,
		0.259225,
		0.193529,
		0.133949,
		0.082589,
		0.041669,
		0.013517,
		0.000567,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.0022,
		0.018867,
		0.049871,
		0.09287,
		0.145627,
		0.20601,
		0.271997,
		0.341684,
		0.413289,
		0.48516,
		0.555785,
		0.623799,
		0.687996,
		0.747342,
		0.800988,
		0.848287,
		0.88881,
		0.922365,
		0.949014,
		0.969087,
		0.983187,
		0.992177,
		0.997147,
		0.999326,
		0.999942,
		1,
		0.999978,
		0.999581,
		0.997904,
		0.99375,
		0.985874,
		0.973143,
		0.95463,
		0.929666,
		0.897854,
		0.859064,
		0.813428,
		0.761317,
		0.703324,
		0.64025,
		0.573084,
		0.502989,
		0.431293,
		0.359472,
		0.289144,
		0.222058,
		0.160089,
		0.105229,
		0.059585,
		0.025367,
		0.004891,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.002158,
		0.018518,
		0.048971,
		0.091244,
		0.14316,
		0.202643,
		0.267725,
		0.336549,
		0.407377,
		0.478594,
		0.548718,
		0.61641,
		0.680482,
		0.739908,
		0.793842,
		0.841627,
		0.882814,
		0.917181,
		0.944746,
		0.965784,
		0.980831,
		0.990679,
		0.996348,
		0.999011,
		0.999877,
		0.999999,
		0.999994,
		0.999767,
		0.99856,
		0.99526,
		0.988652,
		0.977591,
		0.961112,
		0.938482,
		0.909232,
		0.873152,
		0.830285,
		0.780913,
		0.725537,
		0.664862,
		0.599783,
		0.531365,
		0.460836,
		0.389573,
		0.319089,
		0.251033,
		0.187172,
		0.129392,
		0.079688,
		0.040162,
		0.013016,
		0.000546,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.002117,
		0.018178,
		0.048096,
		0.08966,
		0.140754,
		0.199356,
		0.263549,
		0.331523,
		0.40158,
		0.472143,
		0.541761,
		0.609118,
		0.673044,
		0.732524,
		0.786713,
		0.834948,
		0.876762,
		0.911904,
		0.940353,
		0.962333,
		0.978317,
		0.989031,
		0.995423,
		0.99861,
		0.999775,
		0.999995,
		0.999999,
		0.999882,
		0.999046,
		0.996474,
		0.990993,
		0.981456,
		0.966867,
		0.94644,
		0.919639,
		0.886177,
		0.846017,
		0.799352,
		0.746597,
		0.688367,
		0.625463,
		0.55886,
		0.48969,
		0.419232,
		0.348906,
		0.280256,
		0.214952,
		0.154777,
		0.101621,
		0.057479,
		0.024446,
		0.004709,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.002078,
		0.017847,
		0.047243,
		0.088117,
		0.138407,
		0.196146,
		0.259467,
		0.326602,
		0.395897,
		0.465807,
		0.534913,
		0.601923,
		0.665684,
		0.725192,
		0.779606,
		0.828254,
		0.870658,
		0.90654,
		0.935842,
		0.958739,
		0.97565,
		0.987232,
		0.994368,
		0.998117,
		0.999625,
		0.99998,
		1,
		0.999948,
		0.999396,
		0.997435,
		0.992945,
		0.984791,
		0.971951,
		0.953597,
		0.929128,
		0.898188,
		0.860662,
		0.816662,
		0.76652,
		0.710763,
		0.650105,
		0.585432,
		0.517785,
		0.448352,
		0.378457,
		0.309553,
		0.243212,
		0.181117,
		0.125061,
		0.076937,
		0.038736,
		0.012541,
		0.000525,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.00204,
		0.017525,
		0.046414,
		0.086613,
		0.136118,
		0.193012,
		0.255475,
		0.321785,
		0.390324,
		0.459584,
		0.528173,
		0.594824,
		0.658402,
		0.717915,
		0.772523,
		0.821553,
		0.864511,
		0.901097,
		0.931219,
		0.95501,
		0.972832,
		0.985283,
		0.993181,
		0.997523,
		0.999417,
		0.99995,
		1,
		0.999981,
		0.999638,
		0.998182,
		0.994557,
		0.987648,
		0.97642,
		0.960006,
		0.937751,
		0.909232,
		0.874262,
		0.832877,
		0.785326,
		0.732056,
		0.673697,
		0.611048,
		0.545064,
		0.476846,
		0.407628,
		0.338772,
		0.271757,
		0.208175,
		0.149721,
		0.098193,
		0.055483,
		0.023574,
		0.004537,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.002002,
		0.017212,
		0.045605,
		0.085147,
		0.133884,
		0.189951,
		0.251572,
		0.317068,
		0.384859,
		0.453471,
		0.521539,
		0.587822,
		0.651201,
		0.710695,
		0.76547,
		0.814849,
		0.858326,
		0.895582,
		0.926493,
		0.951151,
		0.96987,
		0.983187,
		0.991857,
		0.996822,
		0.999142,
		0.999894,
		0.999999,
		0.999995,
		0.999798,
		0.99875,
		0.995871,
		0.990077,
		0.980326,
		0.965721,
		0.94556,
		0.919359,
		0.886861,
		0.848031,
		0.803041,
		0.752258,
		0.696234,
		0.635684,
		0.571482,
		0.504643,
		0.436317,
		0.367779,
		0.300419,
		0.23574,
		0.175346,
		0.120941,
		0.074325,
		0.037384,
		0.012092,
		0.000506,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.001966,
		0.016907,
		0.044818,
		0.083718,
		0.131705,
		0.186961,
		0.247755,
		0.31245,
		0.379501,
		0.447466,
		0.515011,
		0.580916,
		0.64408,
		0.703534,
		0.758449,
		0.808146,
		0.85211,
		0.890001,
		0.92167,
		0.94717,
		0.966766,
		0.980944,
		0.990397,
		0.996009,
		0.998793,
		0.999805,
		0.999995,
		0.999999,
		0.999898,
		0.999172,
		0.996928,
		0.992122,
		0.983719,
		0.970795,
		0.952607,
		0.928617,
		0.898503,
		0.862161,
		0.819691,
		0.771385,
		0.717717,
		0.659326,
		0.597004,
		0.531686,
		0.464441,
		0.39646,
		0.32905,
		0.263625,
		0.201706,
		0.144906,
		0.094935,
		0.053588,
		0.022747,
		0.004374,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.001931,
		0.01661,
		0.044051,
		0.082324,
		0.129577,
		0.184039,
		0.244022,
		0.307926,
		0.374245,
		0.441568,
		0.508587,
		0.574105,
		0.63704,
		0.696434,
		0.751464,
		0.80145,
		0.845868,
		0.884362,
		0.916757,
		0.943071,
		0.963527,
		0.978558,
		0.9888,
		0.995079,
		0.998361,
		0.999675,
		0.999983,
		1,
		0.999955,
		0.999475,
		0.997765,
		0.993829,
		0.986646,
		0.975277,
		0.958943,
		0.937054,
		0.909232,
		0.875306,
		0.835308,
		0.789457,
		0.738155,
		0.681965,
		0.621603,
		0.557927,
		0.491928,
		0.424716,
		0.357518,
		0.291667,
		0.228598,
		0.169842,
		0.11702,
		0.071843,
		0.036101,
		0.011667,
		0.000488,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.001896,
		0.016321,
		0.043303,
		0.080964,
		0.1275,
		0.181184,
		0.24037,
		0.303496,
		0.369091,
		0.435774,
		0.502265,
		0.567389,
		0.630081,
		0.689396,
		0.744517,
		0.794763,
		0.839605,
		0.87867,
		0.91176,
		0.938862,
		0.960158,
		0.976031,
		0.987065,
		0.994029,
		0.997839,
		0.999493,
		0.999956,
		1,
		0.999984,
		0.999686,
		0.998415,
		0.995238,
		0.989154,
		0.979216,
		0.964615,
		0.944719,
		0.919093,
		0.887506,
		0.849923,
		0.806498,
		0.757559,
		0.703599,
		0.645261,
		0.583329,
		0.518718,
		0.452462,
		0.385711,
		0.319719,
		0.255841,
		0.195527,
		0.140317,
		0.091835,
		0.051789,
		0.021964,
		0.004219,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.001863,
		0.016039,
		0.042574,
		0.079638,
		0.125472,
		0.178394,
		0.236797,
		0.299157,
		0.364035,
		0.430082,
		0.496045,
		0.560768,
		0.623205,
		0.682423,
		0.737611,
		0.788092,
		0.833326,
		0.872931,
		0.906686,
		0.934549,
		0.956664,
		0.973368,
		0.985193,
		0.992856,
		0.997222,
		0.999253,
		0.999908,
		0.999999,
		0.999996,
		0.999825,
		0.99891,
		0.996388,
		0.991285,
		0.982658,
		0.969673,
		0.951659,
		0.928131,
		0.898801,
		0.863572,
		0.822533,
		0.775945,
		0.724231,
		0.667967,
		0.607863,
		0.544762,
		0.479627,
		0.413531,
		0.347655,
		0.283276,
		0.221767,
		0.164588,
		0.113285,
		0.069482,
		0.034883,
		0.011263,
		0.000471,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.00183,
		0.015765,
		0.041863,
		0.078344,
		0.123492,
		0.175666,
		0.2333,
		0.294906,
		0.359076,
		0.424491,
		0.489923,
		0.55424,
		0.61641,
		0.675515,
		0.73075,
		0.781437,
		0.827036,
		0.867151,
		0.901541,
		0.930138,
		0.95305,
		0.970571,
		0.983187,
		0.991558,
		0.996504,
		0.998946,
		0.999831,
		0.999996,
		0.999999,
		0.999911,
		0.999278,
		0.997312,
		0.993081,
		0.985648,
		0.974162,
		0.95792,
		0.93639,
		0.909232,
		0.87629,
		0.837591,
		0.793332,
		0.743871,
		0.689715,
		0.631507,
		0.570021,
		0.506149,
		0.440894,
		0.375363,
		0.310761,
		0.248386,
		0.189623,
		0.135939,
		0.088884,
		0.050078,
		0.02122,
		0.004073,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.001798,
		0.015497,
		0.04117,
		0.077081,
		0.121558,
		0.173,
		0.229879,
		0.290741,
		0.354211,
		0.418999,
		0.483901,
		0.547804,
		0.609698,
		0.668673,
		0.723934,
		0.774805,
		0.820741,
		0.861335,
		0.896331,
		0.925634,
		0.949321,
		0.967646,
		0.981047,
		0.990135,
		0.995681,
		0.998566,
		0.999716,
		0.999985,
		1,
		0.999961,
		0.999542,
		0.998044,
		0.99458,
		0.988228,
		0.978128,
		0.963547,
		0.943915,
		0.918841,
		0.888115,
		0.851704,
		0.809744,
		0.762531,
		0.710506,
		0.654248,
		0.594464,
		0.531979,
		0.467727,
		0.402746,
		0.338171,
		0.275228,
		0.21523,
		0.159571,
		0.109724,
		0.067236,
		0.033726,
		0.01088,
		0.000454,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.001767,
		0.015236,
		0.040494,
		0.075848,
		0.119668,
		0.170393,
		0.22653,
		0.28666,
		0.349439,
		0.413603,
		0.477974,
		0.541461,
		0.603067,
		0.661898,
		0.717165,
		0.768196,
		0.814442,
		0.855488,
		0.891062,
		0.921045,
		0.945484,
		0.964597,
		0.978775,
		0.988586,
		0.994749,
		0.998106,
		0.999557,
		0.999962,
		1,
		0.999986,
		0.999726,
		0.998613,
		0.995817,
		0.990437,
		0.981612,
		0.968585,
		0.950749,
		0.927669,
		0.899083,
		0.864902,
		0.825205,
		0.780225,
		0.730344,
		0.676077,
		0.618068,
		0.557076,
		0.493969,
		0.429721,
		0.365398,
		0.302157,
		0.241242,
		0.183976,
		0.131761,
		0.086071,
		0.048451,
		0.020513,
		0.003934,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.001737,
		0.014982,
		0.039835,
		0.074644,
		0.117821,
		0.167843,
		0.223252,
		0.282661,
		0.344756,
		0.408303,
		0.472143,
		0.535208,
		0.596519,
		0.655192,
		0.710447,
		0.761615,
		0.808146,
		0.849616,
		0.885739,
		0.916375,
		0.941544,
		0.961427,
		0.976376,
		0.986911,
		0.993706,
		0.997561,
		0.999346,
		0.99992,
		1,
		0.999996,
		0.999847,
		0.999046,
		0.996826,
		0.992316,
		0.984657,
		0.973075,
		0.956935,
		0.935757,
		0.909232,
		0.877219,
		0.839741,
		0.796974,
		0.749239,
		0.696992,
		0.640815,
		0.581406,
		0.51957,
		0.456215,
		0.392345,
		0.32905,
		0.267506,
		0.208972,
		0.154777,
		0.106327,
		0.065096,
		0.032625,
		0.010517,
		0.000439,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.001708,
		0.014734,
		0.039191,
		0.073468,
		0.116017,
		0.165349,
		0.220042,
		0.278741,
		0.340162,
		0.403095,
		0.466406,
		0.529046,
		0.590052,
		0.648554,
		0.70378,
		0.755064,
		0.801855,
		0.843722,
		0.880367,
		0.911631,
		0.937506,
		0.958141,
		0.973851,
		0.985111,
		0.992549,
		0.996925,
		0.999076,
		0.999852,
		0.999996,
		1,
		0.999923,
		0.999367,
		0.997637,
		0.993899,
		0.987301,
		0.977061,
		0.962515,
		0.943146,
		0.918602,
		0.88869,
		0.853382,
		0.812798,
		0.767202,
		0.716994,
		0.662695,
		0.604944,
		0.544486,
		0.482168,
		0.418929,
		0.355799,
		0.29389,
		0.234393,
		0.178574,
		0.127771,
		0.083389,
		0.046901,
		0.019841,
		0.003803,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.001679,
		0.014492,
		0.038563,
		0.07232,
		0.114253,
		0.16291,
		0.2169,
		0.2749,
		0.335654,
		0.397978,
		0.460761,
		0.522972,
		0.583667,
		0.641985,
		0.697165,
		0.748546,
		0.795573,
		0.837812,
		0.874952,
		0.906817,
		0.933375,
		0.954745,
		0.971204,
		0.983187,
		0.991278,
		0.996194,
		0.998741,
		0.999752,
		0.999987,
		1,
		0.999966,
		0.999599,
		0.99828,
		0.995219,
		0.989583,
		0.980581,
		0.967529,
		0.949876,
		0.927229,
		0.89935,
		0.866157,
		0.827721,
		0.78425,
		0.736089,
		0.683702,
		0.627672,
		0.568684,
		0.507526,
		0.44508,
		0.382312,
		0.320274,
		0.260094,
		0.202976,
		0.150192,
		0.103084,
		0.063057,
		0.031577,
		0.010171,
		0.000424,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.001651,
		0.014256,
		0.03795,
		0.071199,
		0.112529,
		0.160523,
		0.213823,
		0.271135,
		0.331231,
		0.39295,
		0.455206,
		0.516987,
		0.577362,
		0.635486,
		0.690605,
		0.742062,
		0.789303,
		0.831889,
		0.869498,
		0.901939,
		0.929157,
		0.951243,
		0.968439,
		0.98114,
		0.989891,
		0.995364,
		0.998334,
		0.999612,
		0.999967,
		1,
		0.999988,
		0.99976,
		0.99878,
		0.99631,
		0.991538,
		0.983674,
		0.972017,
		0.955987,
		0.935153,
		0.909232,
		0.878098,
		0.841768,
		0.800401,
		0.754286,
		0.703836,
		0.649575,
		0.592135,
		0.532247,
		0.470734,
		0.408505,
		0.346552,
		0.285944,
		0.227824,
		0.173402,
		0.123957,
		0.08083,
		0.045425,
		0.019201,
		0.003677,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.001624,
		0.014026,
		0.037351,
		0.070103,
		0.110843,
		0.158187,
		0.210809,
		0.267443,
		0.326889,
		0.38801,
		0.449741,
		0.511089,
		0.571138,
		0.629057,
		0.6841,
		0.735615,
		0.783049,
		0.825958,
		0.864011,
		0.897002,
		0.924857,
		0.947639,
		0.965558,
		0.978974,
		0.988388,
		0.994434,
		0.99785,
		0.999426,
		0.99993,
		1,
		0.999997,
		0.999866,
		0.999161,
		0.997199,
		0.993199,
		0.986377,
		0.976017,
		0.961517,
		0.94241,
		0.918374,
		0.889236,
		0.854966,
		0.815675,
		0.771599,
		0.723099,
		0.670646,
		0.61482,
		0.556295,
		0.495839,
		0.434306,
		0.372632,
		0.311827,
		0.252975,
		0.197229,
		0.145806,
		0.099986,
		0.061111,
		0.030578,
		0.009842,
		0.00041,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.001597,
		0.013802,
		0.036766,
		0.069032,
		0.109195,
		0.155901,
		0.207857,
		0.263824,
		0.322628,
		0.383156,
		0.444363,
		0.505276,
		0.564995,
		0.622699,
		0.677651,
		0.729206,
		0.776812,
		0.820021,
		0.858494,
		0.892012,
		0.92048,
		0.943938,
		0.962567,
		0.976689,
		0.986769,
		0.9934,
		0.997284,
		0.999187,
		0.99987,
		0.999997,
		1,
		0.999932,
		0.999443,
		0.997915,
		0.994599,
		0.988724,
		0.979567,
		0.966504,
		0.949038,
		0.926809,
		0.899603,
		0.867344,
		0.830094,
		0.788042,
		0.741497,
		0.690882,
		0.636722,
		0.57964,
		0.520351,
		0.459655,
		0.398433,
		0.33764,
		0.278304,
		0.22152,
		0.168448,
		0.120309,
		0.078386,
		0.044016,
		0.018592,
		0.003558,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.001571,
		0.013582,
		0.036195,
		0.067986,
		0.107582,
		0.153664,
		0.204965,
		0.260276,
		0.318445,
		0.378386,
		0.439072,
		0.499548,
		0.55893,
		0.61641,
		0.67126,
		0.722839,
		0.770596,
		0.814083,
		0.852952,
		0.886971,
		0.916031,
		0.940146,
		0.959469,
		0.974289,
		0.985035,
		0.992261,
		0.996633,
		0.998889,
		0.999782,
		0.999989,
		1,
		0.99997,
		0.999647,
		0.998482,
		0.995768,
		0.990749,
		0.982702,
		0.970985,
		0.955074,
		0.934575,
		0.909232,
		0.878929,
		0.843682,
		0.803632,
		0.759042,
		0.710282,
		0.657831,
		0.60226,
		0.544233,
		0.484498,
		0.423884,
		0.363292,
		0.303696,
		0.246136,
		0.191718,
		0.141606,
		0.097025,
		0.059254,
		0.029626,
		0.009529,
		0.000397,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.001546,
		0.013368,
		0.035637,
		0.066963,
		0.106005,
		0.151474,
		0.202132,
		0.256796,
		0.31434,
		0.373698,
		0.433865,
		0.493904,
		0.552945,
		0.610192,
		0.664927,
		0.716513,
		0.764404,
		0.808146,
		0.847388,
		0.881887,
		0.911514,
		0.936266,
		0.956268,
		0.971777,
		0.983187,
		0.991015,
		0.995893,
		0.998528,
		0.999658,
		0.999971,
		1,
		0.999989,
		0.999788,
		0.998923,
		0.996733,
		0.992484,
		0.985457,
		0.974996,
		0.960553,
		0.941704,
		0.918157,
		0.889753,
		0.856465,
		0.81839,
		0.775744,
		0.728851,
		0.678141,
		0.624138,
		0.567455,
		0.50879,
		0.448922,
		0.388702,
		0.32905,
		0.270954,
		0.215467,
		0.163699,
		0.116819,
		0.076051,
		0.042672,
		0.018011,
		0.003444,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.001521,
		0.013159,
		0.035092,
		0.065963,
		0.104462,
		0.14933,
		0.199356,
		0.253383,
		0.310309,
		0.369091,
		0.428742,
		0.488343,
		0.547038,
		0.604044,
		0.658652,
		0.710231,
		0.758237,
		0.802214,
		0.841807,
		0.876762,
		0.906935,
		0.932304,
		0.952968,
		0.969156,
		0.981226,
		0.989663,
		0.99506,
		0.998097,
		0.999493,
		0.999938,
		1,
		0.999997,
		0.999882,
		0.999259,
		0.99752,
		0.993959,
		0.987863,
		0.97857,
		0.965511,
		0.948233,
		0.926409,
		0.899843,
		0.868468,
		0.832336,
		0.791619,
		0.746597,
		0.697651,
		0.645261,
		0.589991,
		0.532493,
		0.473495,
		0.413799,
		0.354276,
		0.295864,
		0.239562,
		0.18643,
		0.137583,
		0.094192,
		0.057479,
		0.028718,
		0.00923,
		0.000384,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.001497,
		0.012955,
		0.034559,
		0.064984,
		0.102952,
		0.14723,
		0.196635,
		0.250035,
		0.306352,
		0.364563,
		0.423701,
		0.482863,
		0.541209,
		0.597967,
		0.652437,
		0.703994,
		0.752097,
		0.796291,
		0.836212,
		0.8716,
		0.902298,
		0.928264,
		0.949573,
		0.966428,
		0.979155,
		0.988203,
		0.994133,
		0.997593,
		0.999281,
		0.999886,
		0.999997,
		1,
		0.99994,
		0.999508,
		0.998153,
		0.995203,
		0.989954,
		0.981742,
		0.969981,
		0.954194,
		0.934022,
		0.909232,
		0.879718,
		0.845492,
		0.806683,
		0.763528,
		0.716363,
		0.665622,
		0.611825,
		0.555576,
		0.497557,
		0.438523,
		0.379296,
		0.320766,
		0.263882,
		0.209654,
		0.159146,
		0.113477,
		0.073818,
		0.041389,
		0.017457,
		0.003336,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.001474,
		0.012755,
		0.034038,
		0.064028,
		0.101474,
		0.145174,
		0.193969,
		0.246752,
		0.302467,
		0.360112,
		0.41874,
		0.477463,
		0.535457,
		0.59196,
		0.646281,
		0.697803,
		0.745987,
		0.790377,
		0.830607,
		0.866407,
		0.897607,
		0.924149,
		0.946088,
		0.963598,
		0.976975,
		0.986637,
		0.993109,
		0.997011,
		0.999017,
		0.999807,
		0.99999,
		1,
		0.999973,
		0.999688,
		0.998655,
		0.99624,
		0.991758,
		0.984543,
		0.973997,
		0.959621,
		0.941027,
		0.91795,
		0.890244,
		0.857883,
		0.820956,
		0.779657,
		0.73428,
		0.685216,
		0.63294,
		0.578013,
		0.52107,
		0.46282,
		0.40404,
		0.345573,
		0.288319,
		0.233241,
		0.181354,
		0.133727,
		0.09148,
		0.055783,
		0.027851,
		0.008946,
		0.000372,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.001451,
		0.01256,
		0.033529,
		0.063092,
		0.100028,
		0.14316,
		0.191355,
		0.24353,
		0.298652,
		0.355737,
		0.413858,
		0.472143,
		0.529781,
		0.586022,
		0.640185,
		0.691659,
		0.739908,
		0.784477,
		0.824995,
		0.861185,
		0.892867,
		0.919966,
		0.942516,
		0.960668,
		0.974689,
		0.984964,
		0.991988,
		0.996348,
		0.998695,
		0.999698,
		0.999974,
		1,
		0.99999,
		0.999813,
		0.999046,
		0.997097,
		0.993303,
		0.987004,
		0.977591,
		0.964547,
		0.947458,
		0.926026,
		0.900072,
		0.869533,
		0.834457,
		0.794999,
		0.751412,
		0.704043,
		0.653328,
		0.599783,
		0.544,
		0.486644,
		0.428447,
		0.370205,
		0.312776,
		0.257074,
		0.204067,
		0.154777,
		0.110275,
		0.071682,
		0.040162,
		0.016928,
		0.003233,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.001428,
		0.01237,
		0.033031,
		0.062177,
		0.098612,
		0.141187,
		0.188793,
		0.24037,
		0.294906,
		0.351437,
		0.409054,
		0.466901,
		0.52418,
		0.580154,
		0.63415,
		0.685564,
		0.733863,
		0.778592,
		0.819379,
		0.855939,
		0.888082,
		0.915718,
		0.938862,
		0.957642,
		0.972299,
		0.983187,
		0.990768,
		0.995601,
		0.99831,
		0.999551,
		0.999945,
		1,
		0.999998,
		0.999896,
		0.999343,
		0.997796,
		0.994617,
		0.989154,
		0.980795,
		0.969004,
		0.953345,
		0.933492,
		0.909232,
		0.880467,
		0.847207,
		0.809569,
		0.767766,
		0.722107,
		0.672983,
		0.620871,
		0.566321,
		0.509955,
		0.452462,
		0.394596,
		0.337168,
		0.281048,
		0.227159,
		0.176478,
		0.130029,
		0.088884,
		0.054161,
		0.027023,
		0.008674,
		0.000361,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		9.1e-005,
		0.00706,
		0.024223,
		0.05033,
		0.084169,
		0.124573,
		0.170419,
		0.220628,
		0.27417,
		0.330063,
		0.387376,
		0.445233,
		0.502815,
		0.559361,
		0.614173,
		0.666622,
		0.716147,
		0.762263,
		0.804566,
		0.842738,
		0.876554,
		0.905884,
		0.930704,
		0.951098,
		0.967266,
		0.979518,
		0.988279,
		0.994072,
		0.997501,
		0.999217,
		0.999861,
		0.999995,
		1,
		0.999962,
		0.999635,
		0.998532,
		0.996042,
		0.99151,
		0.984305,
		0.973858,
		0.959694,
		0.941445,
		0.918855,
		0.891784,
		0.860201,
		0.824185,
		0.783913,
		0.739658,
		0.691782,
		0.64073,
		0.587025,
		0.531263,
		0.474108,
		0.416289,
		0.358598,
		0.301882,
		0.247045,
		0.195044,
		0.146884,
		0.103621,
		0.066356,
		0.036235,
		0.014447,
		0.002221,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		0.00082,
		0.010606,
		0.030412,
		0.058969,
		0.095049,
		0.137468,
		0.185091,
		0.236827,
		0.291635,
		0.348527,
		0.406567,
		0.464877,
		0.522638,
		0.579092,
		0.63355,
		0.685391,
		0.73407,
		0.779121,
		0.820162,
		0.856905,
		0.889157,
		0.916829,
		0.939942,
		0.958631,
		0.97315,
		0.983869,
		0.991269,
		0.995927,
		0.998488,
		0.999624,
		0.999961,
		1,
		0.999995,
		0.999857,
		0.999193,
		0.997428,
		0.993901,
		0.987949,
		0.978957,
		0.966392,
		0.949827,
		0.92895,
		0.903567,
		0.873601,
		0.839085,
		0.800163,
		0.757075,
		0.710157,
		0.659834,
		0.606611,
		0.551073,
		0.493875,
		0.435741,
		0.377459,
		0.31988,
		0.26391,
		0.210511,
		0.160699,
		0.115538,
		0.076143,
		0.043674,
		0.019337,
		0.004379,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.002292,
		0.014896,
		0.037337,
		0.068326,
		0.106619,
		0.151018,
		0.200371,
		0.253577,
		0.309587,
		0.367405,
		0.426093,
		0.484769,
		0.542616,
		0.598882,
		0.652886,
		0.704019,
		0.751752,
		0.79564,
		0.835329,
		0.870561,
		0.90118,
		0.927141,
		0.948514,
		0.965488,
		0.978376,
		0.987608,
		0.993724,
		0.997351,
		0.999169,
		0.999852,
		0.999995,
		1,
		0.99996,
		0.999612,
		0.998443,
		0.995807,
		0.991018,
		0.983418,
		0.972419,
		0.957534,
		0.938392,
		0.914741,
		0.886452,
		0.853514,
		0.816028,
		0.774201,
		0.72834,
		0.678846,
		0.626207,
		0.570991,
		0.513845,
		0.455485,
		0.396697,
		0.33833,
		0.281295,
		0.226559,
		0.175145,
		0.128128,
		0.086636,
		0.051842,
		0.024968,
		0.007282,
		9.3e-005,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		0.004519,
		0.019939,
		0.045004,
		0.078406,
		0.118882,
		0.165218,
		0.216251,
		0.270869,
		0.328012,
		0.386681,
		0.44593,
		0.504881,
		0.562719,
		0.618696,
		0.672143,
		0.722464,
		0.769149,
		0.811778,
		0.850025,
		0.883666,
		0.912588,
		0.936792,
		0.956401,
		0.971663,
		0.982951,
		0.990759,
		0.995683,
		0.998396,
		0.9996,
		0.999959,
		1,
		0.999994,
		0.999848,
		0.999143,
		0.997272,
		0.993541,
		0.987255,
		0.977776,
		0.964555,
		0.947158,
		0.925275,
		0.89872,
		0.867431,
		0.831466,
		0.790993,
		0.746289,
		0.697728,
		0.645775,
		0.590984,
		0.533987,
		0.475496,
		0.416289,
		0.357216,
		0.299189,
		0.243179,
		0.190218,
		0.141391,
		0.097836,
		0.060743,
		0.031349,
		0.01094,
		0.000846,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		9.6e-005,
		0.007514,
		0.025748,
		0.053423,
		0.089212,
		0.131836,
		0.180066,
		0.232725,
		0.28869,
		0.346894,
		0.406331,
		0.466054,
		0.525185,
		0.582912,
		0.638497,
		0.69128,
		0.740684,
		0.78622,
		0.827492,
		0.864209,
		0.896184,
		0.92335,
		0.945759,
		0.96359,
		0.977154,
		0.986889,
		0.99335,
		0.997189,
		0.999117,
		0.999843,
		0.999994,
		1,
		0.999957,
		0.999588,
		0.998347,
		0.995554,
		0.99049,
		0.982468,
		0.970881,
		0.95523,
		0.935141,
		0.910368,
		0.880796,
		0.846434,
		0.80741,
		0.763962,
		0.716437,
		0.665277,
		0.611015,
		0.554272,
		0.495746,
		0.436213,
		0.376519,
		0.317576,
		0.260362,
		0.205913,
		0.155325,
		0.109748,
		0.070384,
		0.038489,
		0.015366,
		0.002366,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		0.000874,
		0.01129,
		0.03233,
		0.062598,
		0.100747,
		0.145482,
		0.195557,
		0.249783,
		0.307026,
		0.366214,
		0.426333,
		0.486437,
		0.545648,
		0.60316,
		0.658245,
		0.710256,
		0.758636,
		0.802918,
		0.842738,
		0.877839,
		0.908079,
		0.933436,
		0.95402,
		0.970072,
		0.981968,
		0.990212,
		0.995421,
		0.998296,
		0.999575,
		0.999956,
		1,
		0.999994,
		0.999838,
		0.999089,
		0.997104,
		0.993153,
		0.98651,
		0.976511,
		0.962592,
		0.944313,
		0.921364,
		0.893571,
		0.86089,
		0.823405,
		0.781314,
		0.734932,
		0.684672,
		0.631048,
		0.574663,
		0.516206,
		0.456443,
		0.396218,
		0.336442,
		0.278096,
		0.222223,
		0.169927,
		0.122371,
		0.08077,
		0.046396,
		0.02057,
		0.004665,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.002443,
		0.015859,
		0.039695,
		0.072536,
		0.113014,
		0.159817,
		0.211684,
		0.267414,
		0.325863,
		0.385951,
		0.446662,
		0.507048,
		0.566235,
		0.623424,
		0.677898,
		0.729027,
		0.776274,
		0.8192,
		0.857473,
		0.890876,
		0.919314,
		0.942819,
		0.96156,
		0.975845,
		0.986118,
		0.992949,
		0.997015,
		0.999061,
		0.999833,
		0.999994,
		1,
		0.999955,
		0.999561,
		0.998243,
		0.995282,
		0.989923,
		0.98145,
		0.969235,
		0.952769,
		0.931675,
		0.905717,
		0.874792,
		0.838934,
		0.7983,
		0.753166,
		0.703917,
		0.651042,
		0.595126,
		0.536843,
		0.476952,
		0.416289,
		0.355767,
		0.296367,
		0.239139,
		0.185194,
		0.135705,
		0.091904,
		0.055077,
		0.026564,
		0.007758,
		0.0001,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.004818,
		0.021232,
		0.047852,
		0.083242,
		0.126014,
		0.174837,
		0.228439,
		0.285604,
		0.345181,
		0.406083,
		0.46729,
		0.527856,
		0.586911,
		0.643665,
		0.697414,
		0.747548,
		0.793551,
		0.835018,
		0.871652,
		0.90328,
		0.929857,
		0.951476,
		0.968369,
		0.980913,
		0.989624,
		0.995138,
		0.998188,
		0.999547,
		0.999953,
		1,
		0.999994,
		0.999827,
		0.999031,
		0.996923,
		0.992736,
		0.985711,
		0.975156,
		0.960493,
		0.941275,
		0.917197,
		0.888097,
		0.853952,
		0.814873,
		0.771094,
		0.722968,
		0.670956,
		0.615622,
		0.557624,
		0.497709,
		0.436709,
		0.375532,
		0.315161,
		0.25665,
		0.201118,
		0.14975,
		0.103789,
		0.064539,
		0.033357,
		0.011657,
		0.000903,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000103,
		0.008014,
		0.02742,
		0.056809,
		0.094719,
		0.139745,
		0.190539,
		0.245813,
		0.304341,
		0.364963,
		0.426585,
		0.488188,
		0.548826,
		0.607637,
		0.66384,
		0.716747,
		0.76577,
		0.810421,
		0.850326,
		0.885232,
		0.915012,
		0.93968,
		0.959389,
		0.974442,
		0.985289,
		0.992516,
		0.996827,
		0.999,
		0.999822,
		0.999993,
		1,
		0.999952,
		0.999532,
		0.998131,
		0.994989,
		0.989313,
		0.980357,
		0.967473,
		0.950139,
		0.927979,
		0.900765,
		0.868415,
		0.830985,
		0.788668,
		0.741777,
		0.690745,
		0.636109,
		0.578511,
		0.518684,
		0.45745,
		0.395714,
		0.33446,
		0.274744,
		0.217694,
		0.164503,
		0.116428,
		0.074788,
		0.040959,
		0.016376,
		0.002525,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.000934,
		0.012042,
		0.034434,
		0.066571,
		0.10697,
		0.154206,
		0.206915,
		0.263794,
		0.323607,
		0.385184,
		0.44743,
		0.509323,
		0.569922,
		0.628371,
		0.683903,
		0.73585,
		0.783645,
		0.826833,
		0.865077,
		0.89817,
		0.926039,
		0.948756,
		0.966545,
		0.979781,
		0.988991,
		0.994834,
		0.998072,
		0.999517,
		0.99995,
		1,
		0.999993,
		0.999816,
		0.998968,
		0.996728,
		0.992288,
		0.984851,
		0.973702,
		0.958246,
		0.938031,
		0.912756,
		0.882275,
		0.846589,
		0.805839,
		0.760297,
		0.710363,
		0.656545,
		0.599465,
		0.53984,
		0.478481,
		0.416289,
		0.354245,
		0.293408,
		0.234912,
		0.179959,
		0.12982,
		0.085828,
		0.049379,
		0.021925,
		0.004979,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.00261,
		0.016918,
		0.042284,
		0.077145,
		0.119997,
		0.169394,
		0.223957,
		0.282369,
		0.343382,
		0.405822,
		0.468589,
		0.530661,
		0.591103,
		0.649069,
		0.703808,
		0.754672,
		0.801123,
		0.842738,
		0.879226,
		0.910426,
		0.936326,
		0.957064,
		0.972936,
		0.984397,
		0.99205,
		0.996625,
		0.998934,
		0.99981,
		0.999993,
		1,
		0.999948,
		0.999501,
		0.99801,
		0.994673,
		0.988657,
		0.979184,
		0.965583,
		0.947325,
		0.924033,
		0.895491,
		0.861636,
		0.822556,
		0.778478,
		0.729761,
		0.676885,
		0.620445,
		0.561141,
		0.499772,
		0.43723,
		0.374495,
		0.312626,
		0.252761,
		0.196114,
		0.143965,
		0.097664,
		0.058623,
		0.028318,
		0.008282,
		0.000107,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.005149,
		0.022654,
		0.050979,
		0.088536,
		0.133799,
		0.185305,
		0.241656,
		0.301522,
		0.363647,
		0.42685,
		0.490028,
		0.552163,
		0.612327,
		0.669685,
		0.723506,
		0.773163,
		0.818151,
		0.858088,
		0.892725,
		0.921961,
		0.945845,
		0.964588,
		0.978564,
		0.98831,
		0.994505,
		0.997946,
		0.999485,
		0.999947,
		1,
		0.999993,
		0.999803,
		0.9989,
		0.996517,
		0.991804,
		0.983926,
		0.972142,
		0.955839,
		0.934562,
		0.908019,
		0.876079,
		0.83877,
		0.796268,
		0.748889,
		0.697079,
		0.641405,
		0.582546,
		0.521286,
		0.458508,
		0.395185,
		0.332378,
		0.27123,
		0.212959,
		0.158861,
		0.110298,
		0.0687,
		0.035564,
		0.012447,
		0.000966,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.00011,
		0.008564,
		0.029261,
		0.060526,
		0.100747,
		0.148377,
		0.201932,
		0.259999,
		0.321236,
		0.384378,
		0.448238,
		0.511714,
		0.573791,
		0.63355,
		0.690172,
		0.742944,
		0.79127,
		0.834679,
		0.872831,
		0.905532,
		0.932739,
		0.954571,
		0.971319,
		0.983438,
		0.991548,
		0.996405,
		0.998863,
		0.999797,
		0.999993,
		1,
		0.999945,
		0.999468,
		0.997879,
		0.994331,
		0.987949,
		0.977922,
		0.963556,
		0.944313,
		0.919818,
		0.889869,
		0.854428,
		0.813614,
		0.767695,
		0.717079,
		0.6623,
		0.604014,
		0.542987,
		0.48009,
		0.416289,
		0.352645,
		0.290302,
		0.230487,
		0.174504,
		0.12373,
		0.079615,
		0.043674,
		0.017488,
		0.0027,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.000999,
		0.012873,
		0.036751,
		0.070933,
		0.113781,
		0.163726,
		0.219266,
		0.278972,
		0.34149,
		0.405548,
		0.469955,
		0.533608,
		0.595501,
		0.654724,
		0.710476,
		0.762069,
		0.808939,
		0.850653,
		0.88692,
		0.917603,
		0.942727,
		0.962487,
		0.977255,
		0.987575,
		0.99415,
		0.997809,
		0.99945,
		0.999943,
		1,
		0.999992,
		0.99979,
		0.998826,
		0.996289,
		0.991281,
		0.982931,
		0.970465,
		0.953258,
		0.930852,
		0.902962,
		0.869479,
		0.830462,
		0.786125,
		0.736831,
		0.68308,
		0.625499,
		0.564834,
		0.501941,
		0.437779,
		0.373403,
		0.309961,
		0.248685,
		0.190888,
		0.137962,
		0.091374,
		0.052658,
		0.023419,
		0.005327,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.002795,
		0.018088,
		0.045134,
		0.082205,
		0.127639,
		0.179843,
		0.237298,
		0.298559,
		0.362263,
		0.427129,
		0.491965,
		0.555671,
		0.617247,
		0.675798,
		0.730545,
		0.780826,
		0.826114,
		0.866019,
		0.900305,
		0.928899,
		0.951898,
		0.96958,
		0.982404,
		0.991005,
		0.996168,
		0.998786,
		0.999782,
		0.999992,
		1,
		0.999941,
		0.999431,
		0.997737,
		0.993962,
		0.987186,
		0.976563,
		0.961379,
		0.941085,
		0.915312,
		0.883873,
		0.846758,
		0.804122,
		0.75628,
		0.703692,
		0.646952,
		0.586783,
		0.524024,
		0.459622,
		0.394628,
		0.330188,
		0.26754,
		0.208006,
		0.152992,
		0.103978,
		0.062522,
		0.030251,
		0.008861,
		0.000114,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.005514,
		0.024223,
		0.054421,
		0.094348,
		0.14232,
		0.196721,
		0.256014,
		0.318742,
		0.383529,
		0.449089,
		0.51423,
		0.577857,
		0.638979,
		0.69672,
		0.750323,
		0.799158,
		0.842738,
		0.880725,
		0.912942,
		0.939384,
		0.96023,
		0.975845,
		0.986782,
		0.993766,
		0.997662,
		0.999412,
		0.999939,
		1,
		0.999992,
		0.999775,
		0.998745,
		0.996042,
		0.990717,
		0.981857,
		0.968661,
		0.950488,
		0.926878,
		0.897559,
		0.862446,
		0.82163,
		0.77537,
		0.724081,
		0.668323,
		0.608789,
		0.546297,
		0.481783,
		0.416289,
		0.35096,
		0.287037,
		0.225849,
		0.168815,
		0.117431,
		0.073275,
		0.037997,
		0.01332,
		0.001035,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.000118,
		0.009174,
		0.031293,
		0.064618,
		0.107365,
		0.157821,
		0.214352,
		0.275401,
		0.339498,
		0.405259,
		0.471394,
		0.536711,
		0.60012,
		0.660647,
		0.717434,
		0.769751,
		0.817006,
		0.858756,
		0.89472,
		0.924786,
		0.949027,
		0.967708,
		0.981289,
		0.990418,
		0.995911,
		0.998703,
		0.999767,
		0.999991,
		1,
		0.999937,
		0.999391,
		0.997583,
		0.993563,
		0.986362,
		0.9751,
		0.959038,
		0.937623,
		0.91049,
		0.877473,
		0.838591,
		0.794042,
		0.744191,
		0.689557,
		0.630801,
		0.568718,
		0.504225,
		0.438357,
		0.372254,
		0.307157,
		0.244406,
		0.185427,
		0.131734,
		0.084923,
		0.046668,
		0.018718,
		0.002895,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.001072,
		0.013791,
		0.039308,
		0.075735,
		0.121258,
		0.17414,
		0.232725,
		0.295441,
		0.360804,
		0.427424,
		0.494005,
		0.559361,
		0.622411,
		0.682196,
		0.737881,
		0.788769,
		0.834309,
		0.874111,
		0.907953,
		0.935798,
		0.957802,
		0.974325,
		0.985926,
		0.99335,
		0.997501,
		0.99937,
		0.999934,
		1,
		0.999991,
		0.999759,
		0.998658,
		0.995774,
		0.990107,
		0.980698,
		0.966719,
		0.947512,
		0.92262,
		0.891784,
		0.854945,
		0.812236,
		0.763962,
		0.710598,
		0.652767,
		0.591237,
		0.526907,
		0.460796,
		0.39404,
		0.327881,
		0.263662,
		0.20282,
		0.146884,
		0.09747,
		0.056273,
		0.02507,
		0.005711,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.003,
		0.019382,
		0.048281,
		0.087777,
		0.136026,
		0.191269,
		0.251827,
		0.316114,
		0.382633,
		0.449987,
		0.516882,
		0.582134,
		0.644676,
		0.703566,
		0.758,
		0.807315,
		0.851009,
		0.888746,
		0.920376,
		0.945942,
		0.965692,
		0.980084,
		0.989783,
		0.995632,
		0.998612,
		0.99975,
		0.999991,
		1,
		0.999932,
		0.999348,
		0.997416,
		0.99313,
		0.985471,
		0.973521,
		0.95652,
		0.933906,
		0.905327,
		0.870636,
		0.82989,
		0.783333,
		0.731385,
		0.674631,
		0.613805,
		0.549783,
		0.483569,
		0.416289,
		0.349184,
		0.283601,
		0.220985,
		0.16288,
		0.110919,
		0.066821,
		0.032389,
		0.009504,
		0.000123,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.00592,
		0.025962,
		0.058222,
		0.100747,
		0.151669,
		0.2092,
		0.271644,
		0.337397,
		0.404954,
		0.472913,
		0.539979,
		0.604978,
		0.666857,
		0.724699,
		0.777728,
		0.825327,
		0.867043,
		0.902608,
		0.931946,
		0.955189,
		0.972685,
		0.984999,
		0.992899,
		0.997327,
		0.999325,
		0.999929,
		1,
		0.99999,
		0.999741,
		0.998563,
		0.995483,
		0.989445,
		0.979446,
		0.964625,
		0.944313,
		0.918052,
		0.885604,
		0.846941,
		0.802239,
		0.751857,
		0.696335,
		0.636368,
		0.572806,
		0.506634,
		0.438967,
		0.371041,
		0.304203,
		0.23991,
		0.179716,
		0.125272,
		0.07832,
		0.040688,
		0.014288,
		0.001112,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.000127,
		0.009851,
		0.033544,
		0.069138,
		0.114651,
		0.168182,
		0.227922,
		0.292156,
		0.359265,
		0.427734,
		0.496159,
		0.563249,
		0.627839,
		0.688896,
		0.745529,
		0.797,
		0.842738,
		0.882352,
		0.915644,
		0.942622,
		0.963517,
		0.978782,
		0.989094,
		0.995329,
		0.998512,
		0.999732,
		0.99999,
		1,
		0.999927,
		0.999301,
		0.997234,
		0.99266,
		0.984507,
		0.971816,
		0.953807,
		0.929913,
		0.899793,
		0.863328,
		0.820615,
		0.771949,
		0.717815,
		0.658868,
		0.595924,
		0.529947,
		0.462036,
		0.39342,
		0.325448,
		0.259581,
		0.197385,
		0.140529,
		0.090775,
		0.049979,
		0.020082,
		0.003111,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.001154,
		0.014812,
		0.042142,
		0.081039,
		0.129487,
		0.185559,
		0.247423,
		0.313342,
		0.381687,
		0.450936,
		0.51968,
		0.586639,
		0.650658,
		0.710728,
		0.76599,
		0.815748,
		0.859486,
		0.896877,
		0.927805,
		0.952372,
		0.970913,
		0.983995,
		0.99241,
		0.997137,
		0.999276,
		0.999924,
		1,
		0.99999,
		0.999722,
		0.99846,
		0.995167,
		0.988728,
		0.978092,
		0.962366,
		0.940869,
		0.913148,
		0.878987,
		0.838394,
		0.791594,
		0.739008,
		0.681244,
		0.619082,
		0.553458,
		0.485454,
		0.416289,
		0.34731,
		0.27998,
		0.215878,
		0.156687,
		0.104192,
		0.060273,
		0.026902,
		0.00614,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.003228,
		0.020821,
		0.051768,
		0.09393,
		0.145256,
		0.203793,
		0.267685,
		0.335179,
		0.404632,
		0.474517,
		0.543429,
		0.610092,
		0.673374,
		0.732288,
		0.786014,
		0.833904,
		0.875503,
		0.910561,
		0.939048,
		0.961169,
		0.977373,
		0.988347,
		0.994999,
		0.998404,
		0.999712,
		0.999989,
		1,
		0.999921,
		0.999249,
		0.997036,
		0.992149,
		0.983462,
		0.969973,
		0.950882,
		0.925619,
		0.893857,
		0.855511,
		0.810721,
		0.759844,
		0.703432,
		0.642221,
		0.577116,
		0.509177,
		0.439612,
		0.369759,
		0.301087,
		0.235182,
		0.173742,
		0.118572,
		0.071577,
		0.034761,
		0.010218,
		0.000132,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.006372,
		0.027894,
		0.062433,
		0.107814,
		0.161955,
		0.222873,
		0.28869,
		0.357638,
		0.428063,
		0.498434,
		0.567351,
		0.63355,
		0.695919,
		0.753505,
		0.805528,
		0.851398,
		0.890728,
		0.92335,
		0.949334,
		0.968996,
		0.982906,
		0.991877,
		0.99693,
		0.999222,
		0.999918,
		1,
		0.999989,
		0.999701,
		0.998347,
		0.994823,
		0.987949,
		0.976625,
		0.959925,
		0.937158,
		0.907879,
		0.871897,
		0.829262,
		0.780253,
		0.725363,
		0.665277,
		0.600863,
		0.533157,
		0.463346,
		0.392764,
		0.322879,
		0.255282,
		0.191686,
		0.133916,
		0.083901,
		0.043674,
		0.015366,
		0.001198,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.000137,
		0.010606,
		0.036046,
		0.074147,
		0.122695,
		0.179577,
		0.242783,
		0.310415,
		0.380687,
		0.451939,
		0.522638,
		0.591389,
		0.656948,
		0.718226,
		0.774307,
		0.824461,
		0.868161,
		0.905097,
		0.935195,
		0.958631,
		0.975845,
		0.987535,
		0.99464,
		0.998286,
		0.99969,
		0.999988,
		1,
		0.999915,
		0.999193,
		0.99682,
		0.991594,
		0.982327,
		0.967979,
		0.947725,
		0.920996,
		0.887485,
		0.847143,
		0.800163,
		0.746966,
		0.688183,
		0.62464,
		0.557338,
		0.487448,
		0.416289,
		0.345327,
		0.27616,
		0.210511,
		0.150223,
		0.09725,
		0.053654,
		0.021601,
		0.003352,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.001245,
		0.015951,
		0.045292,
		0.086916,
		0.138573,
		0.198115,
		0.263508,
		0.332833,
		0.404291,
		0.476215,
		0.547074,
		0.615483,
		0.680218,
		0.74022,
		0.794618,
		0.842738,
		0.884124,
		0.918552,
		0.946052,
		0.96692,
		0.981724,
		0.991297,
		0.996704,
		0.999163,
		0.999912,
		1,
		0.999988,
		0.999678,
		0.998223,
		0.994448,
		0.987103,
		0.975034,
		0.957286,
		0.933156,
		0.902212,
		0.864293,
		0.819497,
		0.768167,
		0.710869,
		0.64838,
		0.581665,
		0.511868,
		0.440294,
		0.368403,
		0.297795,
		0.230203,
		0.167488,
		0.111627,
		0.064712,
		0.028941,
		0.006618,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.003484,
		0.022426,
		0.055645,
		0.100747,
		0.155444,
		0.217558,
		0.285028,
		0.355915,
		0.428411,
		0.500843,
		0.571684,
		0.639566,
		0.703287,
		0.761826,
		0.814362,
		0.860284,
		0.899218,
		0.931037,
		0.955885,
		0.974188,
		0.986651,
		0.994248,
		0.998157,
		0.999666,
		0.999988,
		1,
		0.999909,
		0.999131,
		0.996584,
		0.990988,
		0.981094,
		0.965817,
		0.944313,
		0.916014,
		0.880639,
		0.838179,
		0.788888,
		0.73326,
		0.672015,
		0.606074,
		0.536552,
		0.464734,
		0.39207,
		0.32016,
		0.250747,
		0.185705,
		0.127035,
		0.076857,
		0.037403,
		0.011016,
		0.000143,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.006878,
		0.030049,
		0.067117,
		0.115644,
		0.173305,
		0.237891,
		0.307317,
		0.379627,
		0.453002,
		0.525768,
		0.596407,
		0.663568,
		0.726079,
		0.782964,
		0.833458,
		0.877025,
		0.913379,
		0.942504,
		0.964669,
		0.980438,
		0.990665,
		0.996457,
		0.999098,
		0.999905,
		1,
		0.999987,
		0.999653,
		0.998088,
		0.994038,
		0.986181,
		0.973306,
		0.954428,
		0.928835,
		0.896111,
		0.856131,
		0.809049,
		0.755278,
		0.69547,
		0.630499,
		0.561441,
		0.489559,
		0.416289,
		0.343228,
		0.272124,
		0.204865,
		0.143473,
		0.090096,
		0.047001,
		0.01657,
		0.001294,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.000148,
		0.011451,
		0.038838,
		0.079717,
		0.131607,
		0.192146,
		0.259096,
		0.330348,
		0.40393,
		0.478015,
		0.550931,
		0.621173,
		0.687413,
		0.748515,
		0.803552,
		0.851827,
		0.892886,
		0.926546,
		0.95291,
		0.972387,
		0.985689,
		0.993819,
		0.998016,
		0.999639,
		0.999987,
		1,
		0.999901,
		0.999064,
		0.996325,
		0.990328,
		0.979753,
		0.963472,
		0.940621,
		0.910641,
		0.873276,
		0.828569,
		0.776841,
		0.718669,
		0.654869,
		0.586474,
		0.514717,
		0.441017,
		0.366965,
		0.294311,
		0.224955,
		0.160938,
		0.104436,
		0.057749,
		0.023299,
		0.003623,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.001347,
		0.017226,
		0.048808,
		0.093452,
		0.148635,
		0.211958,
		0.281153,
		0.354089,
		0.42878,
		0.503396,
		0.576269,
		0.645911,
		0.711022,
		0.770509,
		0.823506,
		0.869387,
		0.907796,
		0.938663,
		0.962225,
		0.979038,
		0.989975,
		0.996187,
		0.999027,
		0.999897,
		1,
		0.999986,
		0.999625,
		0.99794,
		0.99359,
		0.985175,
		0.971428,
		0.95133,
		0.924165,
		0.889537,
		0.847365,
		0.797864,
		0.741528,
		0.679106,
		0.61158,
		0.540147,
		0.466205,
		0.391333,
		0.317281,
		0.245958,
		0.179424,
		0.119879,
		0.069658,
		0.031222,
		0.007154,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.003771,
		0.024223,
		0.059973,
		0.108329,
		0.166724,
		0.232725,
		0.304035,
		0.378502,
		0.45413,
		0.529086,
		0.601712,
		0.670542,
		0.734311,
		0.791976,
		0.842738,
		0.886059,
		0.921688,
		0.949683,
		0.970427,
		0.984638,
		0.99335,
		0.99786,
		0.99961,
		0.999985,
		1,
		0.999893,
		0.998989,
		0.996042,
		0.989606,
		0.978292,
		0.960924,
		0.936624,
		0.904839,
		0.865352,
		0.81826,
		0.763962,
		0.70313,
		0.636686,
		0.565786,
		0.4918,
		0.416289,
		0.341002,
		0.267854,
		0.198921,
		0.136427,
		0.082738,
		0.040357,
		0.011912,
		0.000155,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.007447,
		0.032464,
		0.072345,
		0.124349,
		0.185867,
		0.254429,
		0.327712,
		0.403546,
		0.479925,
		0.55502,
		0.627186,
		0.694984,
		0.757191,
		0.812825,
		0.861163,
		0.901765,
		0.934501,
		0.959568,
		0.977512,
		0.98922,
		0.99589,
		0.998949,
		0.999889,
		1,
		0.999985,
		0.999595,
		0.997777,
		0.993099,
		0.984077,
		0.969383,
		0.947967,
		0.919112,
		0.882447,
		0.837941,
		0.785882,
		0.726854,
		0.661713,
		0.591564,
		0.51774,
		0.441786,
		0.365439,
		0.290619,
		0.219416,
		0.154076,
		0.096998,
		0.050721,
		0.017921,
		0.001403,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000161,
		0.012401,
		0.041967,
		0.085934,
		0.141513,
		0.206051,
		0.277047,
		0.352149,
		0.429172,
		0.506107,
		0.581127,
		0.65261,
		0.719149,
		0.779572,
		0.832964,
		0.878693,
		0.91643,
		0.946178,
		0.968292,
		0.98349,
		0.992836,
		0.997689,
		0.999578,
		0.999984,
		1,
		0.999884,
		0.998907,
		0.995731,
		0.988816,
		0.976697,
		0.958153,
		0.932289,
		0.898568,
		0.856815,
		0.807193,
		0.750188,
		0.686577,
		0.617404,
		0.543963,
		0.467769,
		0.39055,
		0.314225,
		0.240892,
		0.172822,
		0.112441,
		0.062328,
		0.025204,
		0.003928,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.001462,
		0.01866,
		0.052748,
		0.100747,
		0.159817,
		0.227264,
		0.300551,
		0.377306,
		0.455329,
		0.532609,
		0.607332,
		0.677898,
		0.742944,
		0.801356,
		0.8523,
		0.895243,
		0.929985,
		0.956676,
		0.975845,
		0.988393,
		0.995564,
		0.998863,
		0.99988,
		1,
		0.999984,
		0.999561,
		0.997598,
		0.992561,
		0.982876,
		0.967154,
		0.944313,
		0.913638,
		0.874792,
		0.827802,
		0.773038,
		0.711187,
		0.643226,
		0.570395,
		0.49418,
		0.416289,
		0.338636,
		0.263329,
		0.192655,
		0.12907,
		0.07519,
		0.033782,
		0.007758,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.004095,
		0.026246,
		0.064824,
		0.11679,
		0.179254,
		0.249484,
		0.324909,
		0.403137,
		0.481957,
		0.559361,
		0.63355,
		0.702959,
		0.766271,
		0.822445,
		0.870737,
		0.910731,
		0.942367,
		0.965963,
		0.982234,
		0.992272,
		0.997501,
		0.999543,
		0.999983,
		1,
		0.999874,
		0.998817,
		0.995389,
		0.987949,
		0.974954,
		0.955135,
		0.927583,
		0.891784,
		0.84761,
		0.795304,
		0.735449,
		0.66894,
		0.596961,
		0.520954,
		0.442603,
		0.363815,
		0.286701,
		0.213563,
		0.146884,
		0.089317,
		0.043674,
		0.012921,
		0.000168,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.008089,
		0.035182,
		0.078205,
		0.134062,
		0.199814,
		0.272688,
		0.350084,
		0.42959,
		0.508991,
		0.586284,
		0.659694,
		0.727694,
		0.78903,
		0.842738,
		0.888183,
		0.925079,
		0.953525,
		0.974023,
		0.987485,
		0.995205,
		0.998768,
		0.999869,
		1,
		0.999982,
		0.999524,
		0.9974,
		0.991969,
		0.981561,
		0.96472,
		0.940336,
		0.907703,
		0.86652,
		0.816885,
		0.759263,
		0.694455,
		0.623575,
		0.548017,
		0.469434,
		0.389717,
		0.310976,
		0.235528,
		0.165879,
		0.104717,
		0.054899,
		0.019444,
		0.001526,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000176,
		0.013475,
		0.045488,
		0.092903,
		0.152563,
		0.221482,
		0.296847,
		0.376031,
		0.456608,
		0.536358,
		0.613292,
		0.685666,
		0.752002,
		0.811114,
		0.862134,
		0.904547,
		0.938217,
		0.963419,
		0.980856,
		0.991651,
		0.997294,
		0.999504,
		0.999981,
		1,
		0.999863,
		0.998717,
		0.995012,
		0.986998,
		0.973047,
		0.951842,
		0.922467,
		0.884434,
		0.837676,
		0.782523,
		0.71967,
		0.650148,
		0.575292,
		0.496716,
		0.416289,
		0.336118,
		0.258527,
		0.186045,
		0.121393,
		0.067471,
		0.027353,
		0.004273,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.001593,
		0.02028,
		0.057184,
		0.108924,
		0.172287,
		0.244237,
		0.321925,
		0.402701,
		0.484124,
		0.563978,
		0.640295,
		0.711367,
		0.775775,
		0.832415,
		0.880531,
		0.919743,
		0.950084,
		0.972026,
		0.986488,
		0.994809,
		0.998663,
		0.999858,
		1,
		0.999981,
		0.999482,
		0.997182,
		0.991318,
		0.980117,
		0.962058,
		0.936004,
		0.901258,
		0.857572,
		0.805122,
		0.74448,
		0.676582,
		0.602692,
		0.524376,
		0.443475,
		0.362085,
		0.282534,
		0.207371,
		0.139346,
		0.081404,
		0.03667,
		0.008442,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.004462,
		0.028532,
		0.070284,
		0.126271,
		0.193222,
		0.268054,
		0.347883,
		0.430036,
		0.512066,
		0.591766,
		0.667192,
		0.736686,
		0.798899,
		0.852824,
		0.897829,
		0.933692,
		0.960634,
		0.979343,
		0.990967,
		0.997064,
		0.99946,
		0.99998,
		1,
		0.999851,
		0.998606,
		0.994596,
		0.985952,
		0.970957,
		0.948246,
		0.9169,
		0.876466,
		0.826947,
		0.768774,
		0.702773,
		0.630124,
		0.552335,
		0.47121,
		0.388827,
		0.307516,
		0.229837,
		0.158573,
		0.096707,
		0.047417,
		0.014064,
		0.000183,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.008818,
		0.038254,
		0.084802,
		0.144944,
		0.215353,
		0.292901,
		0.374671,
		0.457972,
		0.540354,
		0.619625,
		0.693878,
		0.761512,
		0.82126,
		0.87223,
		0.913932,
		0.946324,
		0.969836,
		0.985389,
		0.994372,
		0.998546,
		0.999845,
		1,
		0.999979,
		0.999437,
		0.99694,
		0.990599,
		0.978531,
		0.959144,
		0.931276,
		0.894253,
		0.847883,
		0.792437,
		0.72861,
		0.657487,
		0.580504,
		0.499421,
		0.416289,
		0.333432,
		0.253422,
		0.179066,
		0.113386,
		0.059613,
		0.02117,
		0.001665,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000192,
		0.014693,
		0.049471,
		0.100747,
		0.164939,
		0.238661,
		0.318742,
		0.402235,
		0.486437,
		0.568899,
		0.647454,
		0.720239,
		0.785723,
		0.842738,
		0.890522,
		0.928752,
		0.957583,
		0.977679,
		0.990212,
		0.99681,
		0.999412,
		0.999978,
		1,
		0.999838,
		0.998483,
		0.994136,
		0.984799,
		0.968661,
		0.944313,
		0.910833,
		0.867815,
		0.815347,
		0.753975,
		0.684672,
		0.608789,
		0.528027,
		0.444407,
		0.360237,
		0.278096,
		0.200812,
		0.131445,
		0.073275,
		0.029788,
		0.004665,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.001742,
		0.022121,
		0.0622,
		0.118126,
		0.186245,
		0.263117,
		0.345531,
		0.430513,
		0.51535,
		0.597605,
		0.675142,
		0.746155,
		0.809195,
		0.863213,
		0.907596,
		0.942207,
		0.967429,
		0.984178,
		0.993888,
		0.998417,
		0.99983,
		1,
		0.999977,
		0.999385,
		0.996672,
		0.989805,
		0.976785,
		0.955947,
		0.926112,
		0.886629,
		0.837381,
		0.778747,
		0.711563,
		0.637084,
		0.556941,
		0.473108,
		0.387877,
		0.303823,
		0.223793,
		0.150882,
		0.088416,
		0.039943,
		0.00922,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.004882,
		0.031129,
		0.076461,
		0.13694,
		0.208847,
		0.28869,
		0.373216,
		0.459432,
		0.544621,
		0.626365,
		0.70257,
		0.771497,
		0.831801,
		0.882566,
		0.92335,
		0.954233,
		0.975845,
		0.989376,
		0.996527,
		0.999358,
		0.999976,
		1,
		0.999823,
		0.998347,
		0.993626,
		0.983525,
		0.966137,
		0.940003,
		0.904213,
		0.858415,
		0.802795,
		0.738037,
		0.665277,
		0.586062,
		0.502313,
		0.416289,
		0.330561,
		0.247986,
		0.171689,
		0.105044,
		0.05166,
		0.015366,
		0.000201,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.00965,
		0.041747,
		0.092264,
		0.157186,
		0.232725,
		0.315337,
		0.401737,
		0.488913,
		0.574153,
		0.655065,
		0.72961,
		0.796135,
		0.85341,
		0.900677,
		0.937694,
		0.96478,
		0.982839,
		0.99335,
		0.998273,
		0.999814,
		1,
		0.999974,
		0.999328,
		0.996374,
		0.988925,
		0.974858,
		0.952436,
		0.920461,
		0.878322,
		0.825987,
		0.763962,
		0.693246,
		0.615286,
		0.531932,
		0.445405,
		0.358259,
		0.273359,
		0.193856,
		0.123166,
		0.064959,
		0.023136,
		0.001825,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000211,
		0.016085,
		0.053998,
		0.10962,
		0.178855,
		0.25785,
		0.343012,
		0.431024,
		0.518865,
		0.603835,
		0.683581,
		0.756129,
		0.819929,
		0.87389,
		0.917436,
		0.950551,
		0.973821,
		0.98845,
		0.996213,
		0.999297,
		0.999973,
		1,
		0.999806,
		0.998194,
		0.99306,
		0.982117,
		0.963355,
		0.935275,
		0.89698,
		0.848188,
		0.789202,
		0.720862,
		0.644493,
		0.561865,
		0.475143,
		0.386858,
		0.299874,
		0.217363,
		0.142782,
		0.079857,
		0.032563,
		0.005114,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.001913,
		0.024223,
		0.067904,
		0.128533,
		0.201932,
		0.284186,
		0.371655,
		0.460998,
		0.549189,
		0.63355,
		0.711778,
		0.781985,
		0.842738,
		0.893112,
		0.932739,
		0.961859,
		0.981357,
		0.992753,
		0.998111,
		0.999797,
		1,
		0.999972,
		0.999265,
		0.996042,
		0.987949,
		0.97273,
		0.948572,
		0.914268,
		0.869259,
		0.813614,
		0.747985,
		0.673559,
		0.592002,
		0.505413,
		0.416289,
		0.327485,
		0.242186,
		0.163887,
		0.096367,
		0.043674,
		0.010111,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.005363,
		0.034098,
		0.083483,
		0.149001,
		0.226396,
		0.311689,
		0.401201,
		0.491569,
		0.579774,
		0.663169,
		0.739516,
		0.807029,
		0.864419,
		0.91095,
		0.946495,
		0.971582,
		0.987421,
		0.995862,
		0.99923,
		0.999971,
		1,
		0.999787,
		0.998024,
		0.992429,
		0.980556,
		0.960286,
		0.930079,
		0.889065,
		0.837049,
		0.77447,
		0.702346,
		0.622222,
		0.536117,
		0.446476,
		0.356136,
		0.268292,
		0.186469,
		0.114498,
		0.056497,
		0.016857,
		0.000221,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.010606,
		0.045739,
		0.100747,
		0.171018,
		0.252217,
		0.340308,
		0.431573,
		0.522638,
		0.610497,
		0.692551,
		0.766643,
		0.831109,
		0.884829,
		0.927288,
		0.958631,
		0.979712,
		0.992087,
		0.997931,
		0.999776,
		1,
		0.999969,
		0.999193,
		0.995671,
		0.986864,
		0.970373,
		0.944313,
		0.907473,
		0.859359,
		0.800163,
		0.730708,
		0.652395,
		0.56714,
		0.477327,
		0.385763,
		0.295641,
		0.210511,
		0.134253,
		0.071053,
		0.025389,
		0.002008,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.000232,
		0.017685,
		0.059174,
		0.119705,
		0.19457,
		0.279358,
		0.369978,
		0.462682,
		0.55409,
		0.641225,
		0.721547,
		0.793,
		0.854068,
		0.903828,
		0.94202,
		0.9691,
		0.986276,
		0.99547,
		0.999154,
		0.999968,
		1,
		0.999765,
		0.997833,
		0.991726,
		0.978822,
		0.956891,
		0.924359,
		0.880394,
		0.824905,
		0.75849,
		0.682378,
		0.598361,
		0.508744,
		0.416289,
		0.324181,
		0.235987,
		0.155629,
		0.08736,
		0.035743,
		0.005631,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.002111,
		0.02664,
		0.074425,
		0.140358,
		0.219635,
		0.30777,
		0.400625,
		0.494427,
		0.585802,
		0.671812,
		0.749996,
		0.818422,
		0.875748,
		0.921283,
		0.95506,
		0.977883,
		0.991344,
		0.997729,
		0.999754,
		1,
		0.999966,
		0.999113,
		0.995256,
		0.985655,
		0.967759,
		0.939609,
		0.900005,
		0.848532,
		0.785527,
		0.712016,
		0.629643,
		0.540612,
		0.447628,
		0.353853,
		0.262862,
		0.178617,
		0.105431,
		0.047952,
		0.011138,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.00592,
		0.037511,
		0.091512,
		0.162702,
		0.246181,
		0.337397,
		0.432165,
		0.526696,
		0.617635,
		0.7021,
		0.777728,
		0.842738,
		0.895994,
		0.937074,
		0.966345,
		0.984999,
		0.99503,
		0.999069,
		0.999964,
		1,
		0.999741,
		0.997618,
		0.990939,
		0.976892,
		0.953131,
		0.918052,
		0.870879,
		0.811648,
		0.741144,
		0.660837,
		0.572806,
		0.47968,
		0.384583,
		0.291093,
		0.203198,
		0.125272,
		0.062044,
		0.018575,
		0.000244,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.011711,
		0.05033,
		0.110446,
		0.186723,
		0.27417,
		0.368169,
		0.464497,
		0.559361,
		0.649438,
		0.731919,
		0.804566,
		0.865776,
		0.914656,
		0.951098,
		0.975845,
		0.990511,
		0.997501,
		0.999728,
		1,
		0.999962,
		0.999022,
		0.99479,
		0.984305,
		0.964853,
		0.934406,
		0.891784,
		0.836674,
		0.769588,
		0.691782,
		0.605186,
		0.512331,
		0.416289,
		0.320624,
		0.229348,
		0.146884,
		0.078042,
		0.027987,
		0.002221,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.000257,
		0.019535,
		0.065127,
		0.131232,
		0.212399,
		0.30355,
		0.400003,
		0.497509,
		0.592281,
		0.681045,
		0.761088,
		0.830324,
		0.887363,
		0.931598,
		0.963278,
		0.98357,
		0.994536,
		0.998972,
		0.99996,
		0.999999,
		0.999714,
		0.997377,
		0.990058,
		0.974739,
		0.948956,
		0.911086,
		0.860425,
		0.797161,
		0.722303,
		0.637598,
		0.545454,
		0.448872,
		0.35139,
		0.257029,
		0.170259,
		0.095965,
		0.039412,
		0.006231,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.002341,
		0.029437,
		0.081927,
		0.153869,
		0.2397,
		0.334256,
		0.432804,
		0.531073,
		0.6253,
		0.712278,
		0.789417,
		0.854812,
		0.907331,
		0.946697,
		0.973569,
		0.989577,
		0.997245,
		0.999699,
		0.999999,
		0.999958,
		0.99892,
		0.994266,
		0.982792,
		0.961616,
		0.928639,
		0.882719,
		0.823672,
		0.752213,
		0.669872,
		0.578906,
		0.482221,
		0.383309,
		0.286194,
		0.195381,
		0.115823,
		0.052888,
		0.012329,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.006567,
		0.041461,
		0.100747,
		0.178346,
		0.268583,
		0.366214,
		0.466459,
		0.565045,
		0.658245,
		0.742944,
		0.816702,
		0.877839,
		0.925522,
		0.959858,
		0.981968,
		0.99398,
		0.998863,
		0.999956,
		0.999999,
		0.999683,
		0.997104,
		0.989067,
		0.972331,
		0.944313,
		0.903379,
		0.848922,
		0.781314,
		0.701826,
		0.612528,
		0.516206,
		0.416289,
		0.316784,
		0.222223,
		0.137622,
		0.068445,
		0.02057,
		0.000272,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.012998,
		0.055645,
		0.121599,
		0.204642,
		0.298992,
		0.39933,
		0.500843,
		0.599263,
		0.690927,
		0.772836,
		0.842738,
		0.899218,
		0.941795,
		0.971021,
		0.988525,
		0.996954,
		0.999666,
		0.999999,
		0.999954,
		0.998803,
		0.993675,
		0.981094,
		0.958001,
		0.922235,
		0.872709,
		0.809397,
		0.73326,
		0.646144,
		0.550683,
		0.450219,
		0.348725,
		0.250747,
		0.161357,
		0.086107,
		0.031004,
		0.00247,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.000287,
		0.02169,
		0.072021,
		0.144484,
		0.232725,
		0.330857,
		0.433497,
		0.535808,
		0.63355,
		0.723143,
		0.801739,
		0.867313,
		0.918768,
		0.956035,
		0.980168,
		0.99335,
		0.998739,
		0.999951,
		0.999999,
		0.999648,
		0.996795,
		0.987949,
		0.969633,
		0.939136,
		0.894835,
		0.836245,
		0.763962,
		0.679561,
		0.58549,
		0.484973,
		0.381929,
		0.280903,
		0.187012,
		0.105894,
		0.043674,
		0.006932,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.00261,
		0.032698,
		0.090614,
		0.169394,
		0.262549,
		0.364093,
		0.468589,
		0.571191,
		0.667708,
		0.754672,
		0.829424,
		0.890217,
		0.936326,
		0.968161,
		0.987338,
		0.996625,
		0.999628,
		0.999999,
		0.999948,
		0.998671,
		0.993005,
		0.979184,
		0.953955,
		0.915109,
		0.861636,
		0.793707,
		0.71257,
		0.620445,
		0.520404,
		0.416289,
		0.312626,
		0.214561,
		0.127813,
		0.058623,
		0.013723,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.007327,
		0.046069,
		0.111441,
		0.19631,
		0.294054,
		0.398598,
		0.50446,
		0.606807,
		0.701521,
		0.785281,
		0.855661,
		0.911245,
		0.951752,
		0.978139,
		0.992637,
		0.998598,
		0.999945,
		0.999999,
		0.999607,
		0.996443,
		0.986687,
		0.9666,
		0.933353,
		0.885348,
		0.822257,
		0.744944,
		0.655346,
		0.556348,
		0.451682,
		0.345833,
		0.243966,
		0.151865,
		0.075884,
		0.022905,
		0.000304,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.014509,
		0.061846,
		0.13451,
		0.2252,
		0.327165,
		0.43425,
		0.540946,
		0.642452,
		0.734755,
		0.814721,
		0.880212,
		0.930208,
		0.964943,
		0.985994,
		0.996249,
		0.999585,
		0.999999,
		0.999942,
		0.99852,
		0.992245,
		0.977027,
		0.949417,
		0.907163,
		0.849368,
		0.776439,
		0.689972,
		0.592618,
		0.487964,
		0.380428,
		0.275172,
		0.178037,
		0.095481,
		0.034535,
		0.002763,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.000322,
		0.024223,
		0.080064,
		0.159817,
		0.256014,
		0.361785,
		0.470906,
		0.577857,
		0.677898,
		0.767158,
		0.842738,
		0.902848,
		0.94694,
		0.975845,
		0.991827,
		0.998436,
		0.999939,
		0.999999,
		0.999561,
		0.996042,
		0.985256,
		0.963183,
		0.926878,
		0.874792,
		0.806797,
		0.724081,
		0.629005,
		0.524968,
		0.416289,
		0.308108,
		0.206304,
		0.117431,
		0.048665,
		0.007758,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.002929,
		0.036531,
		0.100747,
		0.187344,
		0.28869,
		0.397801,
		0.5084,
		0.614981,
		0.712898,
		0.798462,
		0.86907,
		0.92335,
		0.961312,
		0.984468,
		0.99582,
		0.999535,
		0.999999,
		0.999935,
		0.998347,
		0.991379,
		0.974587,
		0.944313,
		0.898284,
		0.835752,
		0.757413,
		0.665277,
		0.562504,
		0.453277,
		0.342684,
		0.236626,
		0.14174,
		0.065341,
		0.015366,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.008227,
		0.051487,
		0.123912,
		0.217061,
		0.323142,
		0.435073,
		0.54654,
		0.652082,
		0.74718,
		0.828385,
		0.893455,
		0.941522,
		0.973246,
		0.990902,
		0.998251,
		0.999931,
		0.999999,
		0.999508,
		0.995583,
		0.983628,
		0.959323,
		0.91961,
		0.863025,
		0.789689,
		0.701178,
		0.600358,
		0.491227,
		0.37879,
		0.268945,
		0.168399,
		0.084596,
		0.025659,
		0.000342,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.0163,
		0.069138,
		0.149561,
		0.248916,
		0.359265,
		0.473438,
		0.585111,
		0.688896,
		0.780455,
		0.856637,
		0.915644,
		0.957205,
		0.98273,
		0.995329,
		0.999478,
		0.999999,
		0.999927,
		0.998148,
		0.99039,
		0.971816,
		0.938558,
		0.888342,
		0.820615,
		0.73643,
		0.638287,
		0.529947,
		0.416289,
		0.303183,
		0.197385,
		0.106458,
		0.038704,
		0.003111,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.000363,
		0.027226,
		0.089522,
		0.177679,
		0.282841,
		0.396928,
		0.512705,
		0.623865,
		0.725136,
		0.812415,
		0.882926,
		0.935405,
		0.970291,
		0.989843,
		0.998037,
		0.999922,
		0.999999,
		0.999446,
		0.995056,
		0.981771,
		0.954949,
		0.911434,
		0.849883,
		0.770731,
		0.676022,
		0.569217,
		0.455022,
		0.339242,
		0.228658,
		0.130939,
		0.05456,
		0.008741,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.00331,
		0.041077,
		0.112664,
		0.208236,
		0.318742,
		0.435974,
		0.552654,
		0.662527,
		0.760489,
		0.842738,
		0.906963,
		0.952544,
		0.980744,
		0.994763,
		0.999412,
		0.999999,
		0.999917,
		0.997917,
		0.989256,
		0.968661,
		0.932051,
		0.877185,
		0.803759,
		0.713267,
		0.608789,
		0.494799,
		0.376995,
		0.262156,
		0.158037,
		0.073275,
		0.017323,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.009304,
		0.057915,
		0.138572,
		0.241181,
		0.3565,
		0.476214,
		0.59303,
		0.70079,
		0.794617,
		0.871094,
		0.928478,
		0.966918,
		0.988624,
		0.997788,
		0.999911,
		0.999997,
		0.999373,
		0.994447,
		0.979644,
		0.949977,
		0.90221,
		0.835175,
		0.749695,
		0.648379,
		0.535398,
		0.416289,
		0.297794,
		0.187731,
		0.094888,
		0.028941,
		0.000387,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.018444,
		0.077791,
		0.167239,
		0.276435,
		0.395962,
		0.51742,
		0.633539,
		0.738309,
		0.827155,
		0.897143,
		0.947222,
		0.978449,
		0.994091,
		0.999317,
		0.999981,
		0.999888,
		0.997633,
		0.987932,
		0.965039,
		0.924657,
		0.864619,
		0.784947,
		0.687664,
		0.576554,
		0.456932,
		0.335457,
		0.21998,
		0.11942,
		0.043674,
		0.003529,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.000414,
		0.03082,
		0.100736,
		0.198621,
		0.313873,
		0.436917,
		0.559298,
		0.673812,
		0.774666,
		0.857677,
		0.920512,
		0.962953,
		0.98711,
		0.997388,
		0.999787,
		0.999885,
		0.999178,
		0.993631,
		0.977092,
		0.944206,
		0.891682,
		0.818594,
		0.726248,
		0.617935,
		0.498671,
		0.374977,
		0.254699,
		0.146868,
		0.061582,
		0.009922,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.003769,
		0.046503,
		0.126741,
		0.232614,
		0.353287,
		0.479047,
		0.601425,
		0.713346,
		0.80931,
		0.885637,
		0.940734,
		0.97538,
		0.992877,
		0.998768,
		0.999522,
		0.999416,
		0.996864,
		0.98597,
		0.960466,
		0.915849,
		0.85008,
		0.763598,
		0.659076,
		0.541136,
		0.416091,
		0.291735,
		0.177175,
		0.082699,
		0.019668,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.01059,
		0.065521,
		0.155717,
		0.268996,
		0.394307,
		0.52184,
		0.643164,
		0.751401,
		0.841453,
		0.910271,
		0.957169,
		0.984094,
		0.995644,
		0.998361,
		0.998473,
		0.997669,
		0.991411,
		0.9729,
		0.936403,
		0.878618,
		0.798942,
		0.699281,
		0.583747,
		0.458357,
		0.330792,
		0.21019,
		0.106998,
		0.032845,
		0.000442,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.020955,
		0.087814,
		0.187433,
		0.30734,
		0.436307,
		0.564478,
		0.68353,
		0.786881,
		0.869955,
		0.930499,
		0.968915,
		0.988486,
		0.995131,
		0.995988,
		0.995868,
		0.992978,
		0.980739,
		0.952336,
		0.903084,
		0.831152,
		0.737504,
		0.625599,
		0.501051,
		0.371341,
		0.245582,
		0.134336,
		0.049465,
		0.004022,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.000471,
		0.034862,
		0.113164,
		0.221416,
		0.346903,
		0.478279,
		0.605712,
		0.721083,
		0.818227,
		0.893244,
		0.944861,
		0.974769,
		0.987715,
		0.990785,
		0.990912,
		0.99,
		0.982956,
		0.962301,
		0.921945,
		0.858647,
		0.772234,
		0.66534,
		0.543038,
		0.412507,
		0.282748,
		0.164372,
		0.069425,
		0.011259,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.004255,
		0.052157,
		0.141106,
		0.256856,
		0.386524,
		0.51873,
		0.643772,
		0.753855,
		0.84338,
		0.909308,
		0.95155,
		0.973275,
		0.980718,
		0.981686,
		0.98155,
		0.978299,
		0.964651,
		0.93327,
		0.879373,
		0.801442,
		0.701088,
		0.582681,
		0.452958,
		0.320699,
		0.19647,
		0.092422,
		0.022135,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.011792,
		0.07245,
		0.170831,
		0.29251,
		0.424544,
		0.555639,
		0.676358,
		0.779382,
		0.859861,
		0.915816,
		0.948562,
		0.962873,
		0.966297,
		0.966441,
		0.96542,
		0.957597,
		0.934871,
		0.890903,
		0.822636,
		0.730447,
		0.617806,
		0.490845,
		0.357979,
		0.229604,
		0.11786,
		0.036449,
		0.000494,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.022848,
		0.095026,
		0.201119,
		0.326664,
		0.458811,
		0.58651,
		0.700751,
		0.794878,
		0.864989,
		0.910389,
		0.933976,
		0.942136,
		0.943206,
		0.943055,
		0.939475,
		0.924584,
		0.890684,
		0.83306,
		0.750642,
		0.645788,
		0.523831,
		0.392648,
		0.262302,
		0.144773,
		0.053735,
		0.0044,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.000502,
		0.036891,
		0.118789,
		0.23032,
		0.35718,
		0.486799,
		0.608558,
		0.714073,
		0.797554,
		0.856264,
		0.890993,
		0.906329,
		0.910034,
		0.910191,
		0.909081,
		0.900656,
		0.876427,
		0.830044,
		0.758816,
		0.663773,
		0.549241,
		0.422357,
		0.292662,
		0.171783,
		0.073179,
		0.011958,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.004361,
		0.053042,
		0.142261,
		0.256439,
		0.381656,
		0.505845,
		0.619035,
		0.713679,
		0.785067,
		0.83183,
		0.856393,
		0.864981,
		0.866118,
		0.865957,
		0.86217,
		0.84658,
		0.811468,
		0.75245,
		0.669042,
		0.564348,
		0.444555,
		0.318468,
		0.197141,
		0.093595,
		0.0226,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.011499,
		0.070064,
		0.163672,
		0.277308,
		0.397697,
		0.513512,
		0.615657,
		0.697635,
		0.756013,
		0.790953,
		0.806556,
		0.810366,
		0.810529,
		0.809382,
		0.800764,
		0.77625,
		0.729854,
		0.659459,
		0.566765,
		0.456801,
		0.337415,
		0.218859,
		0.113465,
		0.0354,
		0.000483,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.020971,
		0.086449,
		0.181135,
		0.290865,
		0.403267,
		0.507996,
		0.597047,
		0.665163,
		0.710347,
		0.734367,
		0.742861,
		0.743996,
		0.743835,
		0.74007,
		0.724736,
		0.690604,
		0.633929,
		0.55488,
		0.45715,
		0.347429,
		0.234937,
		0.131066,
		0.049109,
		0.004055,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.000432,
		0.031536,
		0.100576,
		0.192897,
		0.295462,
		0.397048,
		0.488506,
		0.563102,
		0.616965,
		0.649616,
		0.664376,
		0.668022,
		0.668179,
		0.667076,
		0.658875,
		0.635826,
		0.592741,
		0.528227,
		0.444525,
		0.347001,
		0.243645,
		0.144681,
		0.062266,
		0.010267,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.003467,
		0.04179,
		0.110928,
		0.197612,
		0.29017,
		0.378733,
		0.455512,
		0.515172,
		0.555299,
		0.57691,
		0.584648,
		0.585694,
		0.585544,
		0.582095,
		0.568211,
		0.537699,
		0.487707,
		0.418991,
		0.33549,
		0.243819,
		0.152846,
		0.073371,
		0.017889,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.008346,
		0.050357,
		0.116324,
		0.194576,
		0.274981,
		0.349164,
		0.4108,
		0.455999,
		0.483784,
		0.496513,
		0.499696,
		0.499835,
		0.498867,
		0.491749,
		0.472002,
		0.435589,
		0.381856,
		0.313298,
		0.235077,
		0.154586,
		0.08111,
		0.025572,
		0.000352,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.013754,
		0.056098,
		0.116117,
		0.18387,
		0.250868,
		0.310294,
		0.3573,
		0.389406,
		0.406945,
		0.413311,
		0.414182,
		0.414057,
		0.411201,
		0.399855,
		0.375266,
		0.335565,
		0.281881,
		0.217932,
		0.149594,
		0.084551,
		0.032043,
		0.002672,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.000256,
		0.018485,
		0.058278,
		0.110298,
		0.166381,
		0.219692,
		0.26495,
		0.298721,
		0.319805,
		0.329606,
		0.33209,
		0.3322,
		0.33144,
		0.325921,
		0.310826,
		0.283404,
		0.24359,
		0.193752,
		0.138292,
		0.083299,
		0.036295,
		0.00605,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.001814,
		0.021621,
		0.056675,
		0.099509,
		0.143689,
		0.183964,
		0.216483,
		0.239079,
		0.251619,
		0.256238,
		0.256879,
		0.256786,
		0.254699,
		0.246526,
		0.229082,
		0.201372,
		0.164591,
		0.121789,
		0.077548,
		0.037732,
		0.009308,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.003798,
		0.022647,
		0.0516,
		0.084949,
		0.11786,
		0.146521,
		0.168337,
		0.182192,
		0.188738,
		0.190422,
		0.190497,
		0.189978,
		0.186264,
		0.176263,
		0.158393,
		0.132916,
		0.101724,
		0.068049,
		0.036236,
		0.011571,
		0.000161,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.005339,
		0.021495,
		0.043827,
		0.068196,
		0.091174,
		0.110179,
		0.123644,
		0.131247,
		0.134094,
		0.134495,
		0.134436,
		0.13314,
		0.128143,
		0.117656,
		0.101297,
		0.080042,
		0.055989,
		0.032163,
		0.012361,
		0.001043,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		8.4e-005,
		0.005956,
		0.018512,
		0.03446,
		0.050988,
		0.065832,
		0.077394,
		0.08488,
		0.08848,
		0.089421,
		0.089464,
		0.089172,
		0.087112,
		0.081659,
		0.072094,
		0.058738,
		0.042807,
		0.02625,
		0.011614,
		0.001962,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		0.000473,
		0.005563,
		0.014355,
		0.024749,
		0.03498,
		0.043688,
		0.049997,
		0.05363,
		0.055015,
		0.055213,
		0.055184,
		0.054548,
		0.052138,
		0.047176,
		0.039592,
		0.02998,
		0.019471,
		0.009635,
		0.002411,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.000748,
		0.004394,
		0.009842,
		0.015878,
		0.021513,
		0.026022,
		0.029006,
		0.030469,
		0.030859,
		0.030877,
		0.030755,
		0.02991,
		0.027714,
		0.023941,
		0.018798,
		0.012854,
		0.006974,
		0.002263,
		3.2e-005,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		0.000727,
		0.00288,
		0.005762,
		0.008767,
		0.011415,
		0.013384,
		0.014543,
		0.014994,
		0.015059,
		0.01505,
		0.014841,
		0.014064,
		0.012498,
		0.01016,
		0.007282,
		0.004271,
		0.00167,
		0.000143,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000485,
		0.001481,
		0.002699,
		0.003894,
		0.004882,
		0.005553,
		0.005889,
		0.005981,
		0.005985,
		0.005956,
		0.00576,
		0.005261,
		0.004423,
		0.003312,
		0.002079,
		0.000938,
		0.000161,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		1.8e-005,
		0.000207,
		0.000524,
		0.000882,
		0.001213,
		0.001466,
		0.001619,
		0.001679,
		0.001688,
		0.001687,
		0.001659,
		0.001555,
		0.001351,
		0.001055,
		0.000703,
		0.000355,
		9.1e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		4.1e-005,
		9e-005,
		0.000141,
		0.000185,
		0.000216,
		0.000232,
		0.000237,
		0.000237,
		0.000235,
		0.000226,
		0.000203,
		0.000165,
		0.000116,
		6.4e-005,
		2.1e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		2e-005,
		3.1e-005,
		4e-005,
		4.5e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.5e-005,
		4e-005,
		3.1e-005,
		2e-005,
		2.1e-005,
		6.4e-005,
		0.000116,
		0.000165,
		0.000203,
		0.000226,
		0.000235,
		0.000237,
		0.000237,
		0.000232,
		0.000216,
		0.000185,
		0.000141,
		9e-005,
		4.1e-005,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		9.1e-005,
		0.000355,
		0.000703,
		0.001055,
		0.001351,
		0.001555,
		0.001659,
		0.001687,
		0.001688,
		0.001679,
		0.001619,
		0.001466,
		0.001213,
		0.000882,
		0.000524,
		0.000207,
		1.8e-005,
		1.6e-005,
		0.000339,
		0.000936,
		0.001636,
		0.002303,
		0.002835,
		0.003177,
		0.003331,
		0.003363,
		0.003363,
		0.003331,
		0.003177,
		0.002835,
		0.002303,
		0.001636,
		0.000936,
		0.000339,
		1.6e-005,
		0.000161,
		0.000938,
		0.002079,
		0.003312,
		0.004423,
		0.005261,
		0.00576,
		0.005956,
		0.005985,
		0.005981,
		0.005889,
		0.005553,
		0.004882,
		0.003894,
		0.002699,
		0.001481,
		0.000485,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000143,
		0.00167,
		0.004271,
		0.007282,
		0.01016,
		0.012498,
		0.014064,
		0.014841,
		0.01505,
		0.015059,
		0.014994,
		0.014543,
		0.013384,
		0.011415,
		0.008767,
		0.005762,
		0.00288,
		0.000727,
		0.000774,
		0.003651,
		0.007705,
		0.012061,
		0.016015,
		0.01907,
		0.020985,
		0.021831,
		0.022005,
		0.022005,
		0.021831,
		0.020985,
		0.01907,
		0.016015,
		0.012061,
		0.007705,
		0.003651,
		0.000774,
		3.2e-005,
		0.002263,
		0.006974,
		0.012854,
		0.018798,
		0.023941,
		0.027714,
		0.02991,
		0.030755,
		0.030877,
		0.030859,
		0.030469,
		0.029006,
		0.026022,
		0.021513,
		0.015878,
		0.009842,
		0.004394,
		0.000748,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.002411,
		0.009635,
		0.019471,
		0.02998,
		0.039592,
		0.047176,
		0.052138,
		0.054548,
		0.055184,
		0.055213,
		0.055015,
		0.05363,
		0.049997,
		0.043688,
		0.03498,
		0.024749,
		0.014355,
		0.005563,
		0.000473,
		0.000268,
		0.005877,
		0.016531,
		0.029569,
		0.04274,
		0.054265,
		0.062938,
		0.068256,
		0.070559,
		0.071024,
		0.071024,
		0.070559,
		0.068256,
		0.062938,
		0.054265,
		0.04274,
		0.029569,
		0.016531,
		0.005877,
		0.000268,
		0.001962,
		0.011614,
		0.02625,
		0.042807,
		0.058738,
		0.072094,
		0.081659,
		0.087112,
		0.089172,
		0.089464,
		0.089421,
		0.08848,
		0.08488,
		0.077394,
		0.065832,
		0.050988,
		0.03446,
		0.018512,
		0.005956,
		8.4e-005,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001043,
		0.012361,
		0.032163,
		0.055989,
		0.080042,
		0.101297,
		0.117656,
		0.128143,
		0.13314,
		0.134436,
		0.134495,
		0.134094,
		0.131247,
		0.123644,
		0.110179,
		0.091174,
		0.068196,
		0.043827,
		0.021495,
		0.005339,
		0.004664,
		0.02233,
		0.047987,
		0.076739,
		0.104482,
		0.128047,
		0.145385,
		0.155812,
		0.160248,
		0.161129,
		0.161129,
		0.160248,
		0.155812,
		0.145385,
		0.128047,
		0.104482,
		0.076739,
		0.047987,
		0.02233,
		0.004664,
		0.000161,
		0.011571,
		0.036236,
		0.068049,
		0.101724,
		0.132916,
		0.158393,
		0.176263,
		0.186264,
		0.189978,
		0.190497,
		0.190422,
		0.188738,
		0.182192,
		0.168337,
		0.146521,
		0.11786,
		0.084949,
		0.0516,
		0.022647,
		0.003798,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.009308,
		0.037732,
		0.077548,
		0.121789,
		0.164591,
		0.201372,
		0.229082,
		0.246526,
		0.254699,
		0.256786,
		0.256879,
		0.256238,
		0.251619,
		0.239079,
		0.216483,
		0.183964,
		0.143689,
		0.099509,
		0.056675,
		0.021621,
		0.001814,
		0.000913,
		0.020295,
		0.057942,
		0.105461,
		0.155564,
		0.202223,
		0.240889,
		0.268784,
		0.285274,
		0.292176,
		0.293526,
		0.293526,
		0.292176,
		0.285274,
		0.268784,
		0.240889,
		0.202223,
		0.155564,
		0.105461,
		0.057942,
		0.020295,
		0.000913,
		0.00605,
		0.036295,
		0.083299,
		0.138292,
		0.193752,
		0.24359,
		0.283404,
		0.310826,
		0.325921,
		0.33144,
		0.3322,
		0.33209,
		0.329606,
		0.319805,
		0.298721,
		0.26495,
		0.219692,
		0.166381,
		0.110298,
		0.058278,
		0.018485,
		0.000256,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.002672,
		0.032043,
		0.084551,
		0.149594,
		0.217932,
		0.281881,
		0.335565,
		0.375266,
		0.399855,
		0.411201,
		0.414057,
		0.414182,
		0.413311,
		0.406945,
		0.389406,
		0.3573,
		0.310294,
		0.250868,
		0.18387,
		0.116117,
		0.056098,
		0.013754,
		0.011064,
		0.053638,
		0.116937,
		0.190145,
		0.263942,
		0.330732,
		0.38493,
		0.423361,
		0.445728,
		0.454953,
		0.456733,
		0.456733,
		0.454953,
		0.445728,
		0.423361,
		0.38493,
		0.330732,
		0.263942,
		0.190145,
		0.116937,
		0.053638,
		0.011064,
		0.000352,
		0.025572,
		0.08111,
		0.154586,
		0.235077,
		0.313298,
		0.381856,
		0.435589,
		0.472002,
		0.491749,
		0.498867,
		0.499835,
		0.499696,
		0.496513,
		0.483784,
		0.455999,
		0.4108,
		0.349164,
		0.274981,
		0.194576,
		0.116324,
		0.050357,
		0.008346,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.017889,
		0.073371,
		0.152846,
		0.243819,
		0.33549,
		0.418991,
		0.487707,
		0.537699,
		0.568211,
		0.582095,
		0.585544,
		0.585694,
		0.584648,
		0.57691,
		0.555299,
		0.515172,
		0.455512,
		0.378733,
		0.29017,
		0.197612,
		0.110928,
		0.04179,
		0.003467,
		0.001638,
		0.036791,
		0.106308,
		0.196197,
		0.294076,
		0.389377,
		0.473615,
		0.540744,
		0.587617,
		0.614514,
		0.625459,
		0.627545,
		0.627545,
		0.625459,
		0.614514,
		0.587617,
		0.540744,
		0.473615,
		0.389377,
		0.294076,
		0.196197,
		0.106308,
		0.036791,
		0.001638,
		0.010267,
		0.062266,
		0.144681,
		0.243645,
		0.347001,
		0.444525,
		0.528227,
		0.592741,
		0.635826,
		0.658875,
		0.667076,
		0.668179,
		0.668022,
		0.664376,
		0.649616,
		0.616965,
		0.563102,
		0.488506,
		0.397048,
		0.295462,
		0.192897,
		0.100576,
		0.031536,
		0.000432,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.004055,
		0.049109,
		0.131066,
		0.234937,
		0.347429,
		0.45715,
		0.55488,
		0.633929,
		0.690604,
		0.724736,
		0.74007,
		0.743835,
		0.743996,
		0.742861,
		0.734367,
		0.710347,
		0.665163,
		0.597047,
		0.507996,
		0.403267,
		0.290865,
		0.181135,
		0.086449,
		0.020971,
		0.016015,
		0.078437,
		0.173013,
		0.285117,
		0.401894,
		0.512479,
		0.608295,
		0.683445,
		0.735198,
		0.764515,
		0.776299,
		0.77852,
		0.77852,
		0.776299,
		0.764515,
		0.735198,
		0.683445,
		0.608295,
		0.512479,
		0.401894,
		0.285117,
		0.173013,
		0.078437,
		0.016015,
		0.000483,
		0.0354,
		0.113465,
		0.218859,
		0.337415,
		0.456801,
		0.566765,
		0.659459,
		0.729854,
		0.77625,
		0.800764,
		0.809382,
		0.810529,
		0.810366,
		0.806556,
		0.790953,
		0.756013,
		0.697635,
		0.615657,
		0.513512,
		0.397697,
		0.277308,
		0.163672,
		0.070064,
		0.011499,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.0226,
		0.093595,
		0.197141,
		0.318468,
		0.444555,
		0.564348,
		0.669042,
		0.75245,
		0.811468,
		0.84658,
		0.86217,
		0.865957,
		0.866118,
		0.864981,
		0.856393,
		0.83183,
		0.785067,
		0.713679,
		0.619035,
		0.505845,
		0.381656,
		0.256439,
		0.142261,
		0.053042,
		0.004361,
		0.001976,
		0.044773,
		0.130667,
		0.243899,
		0.370327,
		0.497602,
		0.615402,
		0.715733,
		0.793321,
		0.846087,
		0.875627,
		0.887367,
		0.889556,
		0.889556,
		0.887367,
		0.875627,
		0.846087,
		0.793321,
		0.715733,
		0.615402,
		0.497602,
		0.370327,
		0.243899,
		0.130667,
		0.044773,
		0.001976,
		0.011958,
		0.073179,
		0.171783,
		0.292662,
		0.422357,
		0.549241,
		0.663773,
		0.758816,
		0.830044,
		0.876427,
		0.900656,
		0.909081,
		0.910191,
		0.910034,
		0.906329,
		0.890993,
		0.856264,
		0.797554,
		0.714073,
		0.608558,
		0.486799,
		0.35718,
		0.23032,
		0.118789,
		0.036891,
		0.000502,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.0044,
		0.053735,
		0.144773,
		0.262302,
		0.392648,
		0.523831,
		0.645788,
		0.750642,
		0.83306,
		0.890684,
		0.924584,
		0.939475,
		0.943055,
		0.943206,
		0.942136,
		0.933976,
		0.910389,
		0.864989,
		0.794878,
		0.700751,
		0.58651,
		0.458811,
		0.326664,
		0.201119,
		0.095026,
		0.022848,
		0.016906,
		0.083512,
		0.185998,
		0.309899,
		0.442298,
		0.572,
		0.689759,
		0.788563,
		0.864009,
		0.914735,
		0.942826,
		0.953873,
		0.955914,
		0.955914,
		0.953873,
		0.942826,
		0.914735,
		0.864009,
		0.788563,
		0.689759,
		0.572,
		0.442298,
		0.309899,
		0.185998,
		0.083512,
		0.016906,
		0.000494,
		0.036449,
		0.11786,
		0.229604,
		0.357979,
		0.490845,
		0.617806,
		0.730447,
		0.822636,
		0.890903,
		0.934871,
		0.957597,
		0.96542,
		0.966441,
		0.966297,
		0.962873,
		0.948562,
		0.915816,
		0.859861,
		0.779382,
		0.676358,
		0.555639,
		0.424544,
		0.29251,
		0.170831,
		0.07245,
		0.011792,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.022135,
		0.092422,
		0.19647,
		0.320699,
		0.452958,
		0.582681,
		0.701088,
		0.801442,
		0.879373,
		0.93327,
		0.964651,
		0.978299,
		0.98155,
		0.981686,
		0.980718,
		0.973275,
		0.95155,
		0.909308,
		0.84338,
		0.753855,
		0.643772,
		0.51873,
		0.386524,
		0.256856,
		0.141106,
		0.052157,
		0.004255,
		0.001888,
		0.043104,
		0.126853,
		0.239021,
		0.366798,
		0.498808,
		0.62527,
		0.73821,
		0.831725,
		0.90232,
		0.949288,
		0.975038,
		0.985067,
		0.986903,
		0.986903,
		0.985067,
		0.975038,
		0.949288,
		0.90232,
		0.831725,
		0.73821,
		0.62527,
		0.498808,
		0.366798,
		0.239021,
		0.126853,
		0.043104,
		0.001888,
		0.011259,
		0.069425,
		0.164372,
		0.282748,
		0.412507,
		0.543038,
		0.66534,
		0.772234,
		0.858647,
		0.921945,
		0.962301,
		0.982956,
		0.99,
		0.990912,
		0.990785,
		0.987715,
		0.974769,
		0.944861,
		0.893244,
		0.818227,
		0.721083,
		0.605712,
		0.478279,
		0.346903,
		0.221416,
		0.113164,
		0.034862,
		0.000471,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.004022,
		0.049465,
		0.134336,
		0.245582,
		0.371341,
		0.501051,
		0.625599,
		0.737504,
		0.831152,
		0.903084,
		0.952336,
		0.980739,
		0.992978,
		0.995868,
		0.995988,
		0.995131,
		0.988486,
		0.968915,
		0.930499,
		0.869955,
		0.786881,
		0.68353,
		0.564478,
		0.436307,
		0.30734,
		0.187433,
		0.087814,
		0.020955,
		0.015327,
		0.076267,
		0.171255,
		0.287961,
		0.415238,
		0.543246,
		0.663597,
		0.769549,
		0.856247,
		0.921018,
		0.963696,
		0.986877,
		0.995825,
		0.99745,
		0.99745,
		0.995825,
		0.986877,
		0.963696,
		0.921018,
		0.856247,
		0.769549,
		0.663597,
		0.543246,
		0.415238,
		0.287961,
		0.171255,
		0.076267,
		0.015327,
		0.000442,
		0.032845,
		0.106998,
		0.21019,
		0.330792,
		0.458357,
		0.583747,
		0.699281,
		0.798942,
		0.878618,
		0.936403,
		0.9729,
		0.991411,
		0.997669,
		0.998473,
		0.998361,
		0.995644,
		0.984094,
		0.957169,
		0.910271,
		0.841453,
		0.751401,
		0.643164,
		0.52184,
		0.394307,
		0.268996,
		0.155717,
		0.065521,
		0.01059,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.019668,
		0.082699,
		0.177175,
		0.291735,
		0.416091,
		0.541136,
		0.659076,
		0.763598,
		0.85008,
		0.915849,
		0.960466,
		0.98597,
		0.996864,
		0.999416,
		0.999522,
		0.998768,
		0.992877,
		0.97538,
		0.940734,
		0.885637,
		0.80931,
		0.713346,
		0.601425,
		0.479047,
		0.353287,
		0.232614,
		0.126741,
		0.046503,
		0.003769,
		0.001665,
		0.038245,
		0.113359,
		0.215301,
		0.333351,
		0.457861,
		0.580364,
		0.69371,
		0.792245,
		0.872018,
		0.931051,
		0.969601,
		0.990359,
		0.998305,
		0.999737,
		0.999737,
		0.998305,
		0.990359,
		0.969601,
		0.931051,
		0.872018,
		0.792245,
		0.69371,
		0.580364,
		0.457861,
		0.333351,
		0.215301,
		0.113359,
		0.038245,
		0.001665,
		0.009922,
		0.061582,
		0.146868,
		0.254699,
		0.374977,
		0.498671,
		0.617935,
		0.726248,
		0.818594,
		0.891682,
		0.944206,
		0.977092,
		0.993631,
		0.999178,
		0.999885,
		0.999787,
		0.997388,
		0.98711,
		0.962953,
		0.920512,
		0.857677,
		0.774666,
		0.673812,
		0.559298,
		0.436917,
		0.313873,
		0.198621,
		0.100736,
		0.03082,
		0.000414,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.003529,
		0.043674,
		0.11942,
		0.21998,
		0.335457,
		0.456932,
		0.576554,
		0.687664,
		0.784947,
		0.864619,
		0.924657,
		0.965039,
		0.987932,
		0.997633,
		0.999888,
		0.999981,
		0.999317,
		0.994091,
		0.978449,
		0.947222,
		0.897143,
		0.827155,
		0.738309,
		0.633539,
		0.51742,
		0.395962,
		0.276435,
		0.167239,
		0.077791,
		0.018444,
		0.013474,
		0.067471,
		0.152562,
		0.258526,
		0.376029,
		0.496713,
		0.613289,
		0.719666,
		0.811109,
		0.884429,
		0.938211,
		0.973042,
		0.991646,
		0.998711,
		0.999976,
		0.999976,
		0.998711,
		0.991646,
		0.973042,
		0.938211,
		0.884429,
		0.811109,
		0.719666,
		0.613289,
		0.496713,
		0.376029,
		0.258526,
		0.152562,
		0.067471,
		0.013474,
		0.000387,
		0.028941,
		0.094888,
		0.187731,
		0.297794,
		0.416289,
		0.535398,
		0.648379,
		0.749695,
		0.835175,
		0.90221,
		0.949977,
		0.979644,
		0.994447,
		0.999373,
		0.999997,
		0.999911,
		0.997788,
		0.988624,
		0.966918,
		0.928478,
		0.871094,
		0.794617,
		0.70079,
		0.59303,
		0.476214,
		0.3565,
		0.241181,
		0.138572,
		0.057915,
		0.009304,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.017323,
		0.073275,
		0.158037,
		0.262156,
		0.376995,
		0.494799,
		0.608789,
		0.713267,
		0.803759,
		0.877185,
		0.932051,
		0.968661,
		0.989256,
		0.997917,
		0.999917,
		0.999999,
		0.999412,
		0.994763,
		0.980744,
		0.952544,
		0.906963,
		0.842738,
		0.760489,
		0.662527,
		0.552654,
		0.435974,
		0.318742,
		0.208236,
		0.112664,
		0.041077,
		0.00331,
		0.001462,
		0.033782,
		0.100747,
		0.192655,
		0.300551,
		0.416289,
		0.532609,
		0.643226,
		0.742944,
		0.827802,
		0.895243,
		0.944313,
		0.975845,
		0.992561,
		0.998863,
		0.999984,
		0.999984,
		0.998863,
		0.992561,
		0.975845,
		0.944313,
		0.895243,
		0.827802,
		0.742944,
		0.643226,
		0.532609,
		0.416289,
		0.300551,
		0.192655,
		0.100747,
		0.033782,
		0.001462,
		0.008741,
		0.05456,
		0.130939,
		0.228658,
		0.339242,
		0.455022,
		0.569217,
		0.676022,
		0.770731,
		0.849883,
		0.911434,
		0.954949,
		0.981771,
		0.995056,
		0.999446,
		0.999999,
		0.999922,
		0.998037,
		0.989843,
		0.970291,
		0.935405,
		0.882926,
		0.812415,
		0.725136,
		0.623865,
		0.512705,
		0.396928,
		0.282841,
		0.177679,
		0.089522,
		0.027226,
		0.000363,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.003111,
		0.038704,
		0.106458,
		0.197385,
		0.303183,
		0.416289,
		0.529947,
		0.638287,
		0.73643,
		0.820615,
		0.888342,
		0.938558,
		0.971816,
		0.99039,
		0.998148,
		0.999927,
		0.999999,
		0.999478,
		0.995329,
		0.98273,
		0.957205,
		0.915644,
		0.856637,
		0.780455,
		0.688896,
		0.585111,
		0.473438,
		0.359265,
		0.248916,
		0.149561,
		0.069138,
		0.0163,
		0.011912,
		0.059973,
		0.136427,
		0.232725,
		0.341002,
		0.45413,
		0.565786,
		0.670542,
		0.763962,
		0.842738,
		0.904839,
		0.949683,
		0.978292,
		0.99335,
		0.998989,
		0.999985,
		0.999985,
		0.998989,
		0.99335,
		0.978292,
		0.949683,
		0.904839,
		0.842738,
		0.763962,
		0.670542,
		0.565786,
		0.45413,
		0.341002,
		0.232725,
		0.136427,
		0.059973,
		0.011912,
		0.000342,
		0.025659,
		0.084596,
		0.168399,
		0.268945,
		0.37879,
		0.491227,
		0.600358,
		0.701178,
		0.789689,
		0.863025,
		0.91961,
		0.959323,
		0.983628,
		0.995583,
		0.999508,
		0.999999,
		0.999931,
		0.998251,
		0.990902,
		0.973246,
		0.941522,
		0.893455,
		0.828385,
		0.74718,
		0.652082,
		0.54654,
		0.435073,
		0.323142,
		0.217061,
		0.123912,
		0.051487,
		0.008227,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.015366,
		0.065341,
		0.14174,
		0.236626,
		0.342684,
		0.453277,
		0.562504,
		0.665277,
		0.757413,
		0.835752,
		0.898284,
		0.944313,
		0.974587,
		0.991379,
		0.998347,
		0.999935,
		0.999999,
		0.999535,
		0.99582,
		0.984468,
		0.961312,
		0.92335,
		0.86907,
		0.798462,
		0.712898,
		0.614981,
		0.5084,
		0.397801,
		0.28869,
		0.187344,
		0.100747,
		0.036531,
		0.002929,
		0.001294,
		0.030049,
		0.090096,
		0.173305,
		0.272124,
		0.379627,
		0.489559,
		0.596407,
		0.69547,
		0.782964,
		0.856131,
		0.913379,
		0.954428,
		0.980438,
		0.994038,
		0.999098,
		0.999987,
		0.999987,
		0.999098,
		0.994038,
		0.980438,
		0.954428,
		0.913379,
		0.856131,
		0.782964,
		0.69547,
		0.596407,
		0.489559,
		0.379627,
		0.272124,
		0.173305,
		0.090096,
		0.030049,
		0.001294,
		0.007758,
		0.048665,
		0.117431,
		0.206304,
		0.308108,
		0.416289,
		0.524968,
		0.629005,
		0.724081,
		0.806797,
		0.874792,
		0.926878,
		0.963183,
		0.985256,
		0.996042,
		0.999561,
		0.999999,
		0.999939,
		0.998436,
		0.991827,
		0.975845,
		0.94694,
		0.902848,
		0.842738,
		0.767158,
		0.677898,
		0.577857,
		0.470906,
		0.361785,
		0.256014,
		0.159817,
		0.080064,
		0.024223,
		0.000322,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.002763,
		0.034535,
		0.095481,
		0.178037,
		0.275172,
		0.380428,
		0.487964,
		0.592618,
		0.689972,
		0.776439,
		0.849368,
		0.907163,
		0.949417,
		0.977027,
		0.992245,
		0.99852,
		0.999942,
		0.999999,
		0.999585,
		0.996249,
		0.985994,
		0.964943,
		0.930208,
		0.880212,
		0.814721,
		0.734755,
		0.642452,
		0.540946,
		0.43425,
		0.327165,
		0.2252,
		0.13451,
		0.061846,
		0.014509,
		0.010606,
		0.053654,
		0.122695,
		0.210511,
		0.310415,
		0.416289,
		0.522638,
		0.62464,
		0.718226,
		0.800163,
		0.868161,
		0.920996,
		0.958631,
		0.982327,
		0.99464,
		0.999193,
		0.999988,
		0.999988,
		0.999193,
		0.99464,
		0.982327,
		0.958631,
		0.920996,
		0.868161,
		0.800163,
		0.718226,
		0.62464,
		0.522638,
		0.416289,
		0.310415,
		0.210511,
		0.122695,
		0.053654,
		0.010606,
		0.000304,
		0.022905,
		0.075884,
		0.151865,
		0.243966,
		0.345833,
		0.451682,
		0.556348,
		0.655346,
		0.744944,
		0.822257,
		0.885348,
		0.933353,
		0.9666,
		0.986687,
		0.996443,
		0.999607,
		0.999999,
		0.999945,
		0.998598,
		0.992637,
		0.978139,
		0.951752,
		0.911245,
		0.855661,
		0.785281,
		0.701521,
		0.606807,
		0.50446,
		0.398598,
		0.294054,
		0.19631,
		0.111441,
		0.046069,
		0.007327,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.013723,
		0.058623,
		0.127813,
		0.214561,
		0.312626,
		0.416289,
		0.520404,
		0.620445,
		0.71257,
		0.793707,
		0.861636,
		0.915109,
		0.953955,
		0.979184,
		0.993005,
		0.998671,
		0.999948,
		0.999999,
		0.999628,
		0.996625,
		0.987338,
		0.968161,
		0.936326,
		0.890217,
		0.829424,
		0.754672,
		0.667708,
		0.571191,
		0.468589,
		0.364093,
		0.262549,
		0.169394,
		0.090614,
		0.032698,
		0.00261,
		0.001154,
		0.026902,
		0.081039,
		0.156687,
		0.247423,
		0.34731,
		0.450936,
		0.553458,
		0.650658,
		0.739008,
		0.815748,
		0.878987,
		0.927805,
		0.962366,
		0.983995,
		0.995167,
		0.999276,
		0.99999,
		0.99999,
		0.999276,
		0.995167,
		0.983995,
		0.962366,
		0.927805,
		0.878987,
		0.815748,
		0.739008,
		0.650658,
		0.553458,
		0.450936,
		0.34731,
		0.247423,
		0.156687,
		0.081039,
		0.026902,
		0.001154,
		0.006932,
		0.043674,
		0.105894,
		0.187012,
		0.280903,
		0.381929,
		0.484973,
		0.58549,
		0.679561,
		0.763962,
		0.836245,
		0.894835,
		0.939136,
		0.969633,
		0.987949,
		0.996795,
		0.999648,
		0.999999,
		0.999951,
		0.998739,
		0.99335,
		0.980168,
		0.956035,
		0.918768,
		0.867313,
		0.801739,
		0.723143,
		0.63355,
		0.535808,
		0.433497,
		0.330857,
		0.232725,
		0.144484,
		0.072021,
		0.02169,
		0.000287,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.00247,
		0.031004,
		0.086107,
		0.161357,
		0.250747,
		0.348725,
		0.450219,
		0.550683,
		0.646144,
		0.73326,
		0.809397,
		0.872709,
		0.922235,
		0.958001,
		0.981094,
		0.993675,
		0.998803,
		0.999954,
		0.999999,
		0.999666,
		0.996954,
		0.988525,
		0.971021,
		0.941795,
		0.899218,
		0.842738,
		0.772836,
		0.690927,
		0.599263,
		0.500843,
		0.39933,
		0.298992,
		0.204642,
		0.121599,
		0.055645,
		0.012998,
		0.009504,
		0.048281,
		0.110919,
		0.191269,
		0.283601,
		0.382633,
		0.483569,
		0.582134,
		0.674631,
		0.758,
		0.82989,
		0.888746,
		0.933906,
		0.965692,
		0.985471,
		0.995632,
		0.999348,
		0.999991,
		0.999991,
		0.999348,
		0.995632,
		0.985471,
		0.965692,
		0.933906,
		0.888746,
		0.82989,
		0.758,
		0.674631,
		0.582134,
		0.483569,
		0.382633,
		0.283601,
		0.191269,
		0.110919,
		0.048281,
		0.009504,
		0.000272,
		0.02057,
		0.068445,
		0.137622,
		0.222223,
		0.316784,
		0.416289,
		0.516206,
		0.612528,
		0.701826,
		0.781314,
		0.848922,
		0.903379,
		0.944313,
		0.972331,
		0.989067,
		0.997104,
		0.999683,
		0.999999,
		0.999956,
		0.998863,
		0.99398,
		0.981968,
		0.959858,
		0.925522,
		0.877839,
		0.816702,
		0.742944,
		0.658245,
		0.565045,
		0.466459,
		0.366214,
		0.268583,
		0.178346,
		0.100747,
		0.041461,
		0.006567,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.012329,
		0.052888,
		0.115823,
		0.195381,
		0.286194,
		0.383309,
		0.482221,
		0.578906,
		0.669872,
		0.752213,
		0.823672,
		0.882719,
		0.928639,
		0.961616,
		0.982792,
		0.994266,
		0.99892,
		0.999958,
		0.999999,
		0.999699,
		0.997245,
		0.989577,
		0.973569,
		0.946697,
		0.907331,
		0.854812,
		0.789417,
		0.712278,
		0.6253,
		0.531073,
		0.432804,
		0.334256,
		0.2397,
		0.153869,
		0.081927,
		0.029437,
		0.002341,
		0.001035,
		0.024223,
		0.073275,
		0.14232,
		0.225849,
		0.318742,
		0.416289,
		0.51423,
		0.608789,
		0.69672,
		0.77537,
		0.842738,
		0.897559,
		0.939384,
		0.968661,
		0.986782,
		0.996042,
		0.999412,
		0.999992,
		0.999992,
		0.999412,
		0.996042,
		0.986782,
		0.968661,
		0.939384,
		0.897559,
		0.842738,
		0.77537,
		0.69672,
		0.608789,
		0.51423,
		0.416289,
		0.318742,
		0.225849,
		0.14232,
		0.073275,
		0.024223,
		0.001035,
		0.006231,
		0.039412,
		0.095965,
		0.170259,
		0.257029,
		0.35139,
		0.448872,
		0.545454,
		0.637598,
		0.722303,
		0.797161,
		0.860425,
		0.911086,
		0.948956,
		0.974739,
		0.990058,
		0.997377,
		0.999714,
		0.999999,
		0.99996,
		0.998972,
		0.994536,
		0.98357,
		0.963278,
		0.931598,
		0.887363,
		0.830324,
		0.761088,
		0.681045,
		0.592281,
		0.497509,
		0.400003,
		0.30355,
		0.212399,
		0.131232,
		0.065127,
		0.019535,
		0.000257,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.002221,
		0.027987,
		0.078042,
		0.146884,
		0.229348,
		0.320624,
		0.416289,
		0.512331,
		0.605186,
		0.691782,
		0.769588,
		0.836674,
		0.891784,
		0.934406,
		0.964853,
		0.984305,
		0.99479,
		0.999022,
		0.999962,
		1,
		0.999728,
		0.997501,
		0.990511,
		0.975845,
		0.951098,
		0.914656,
		0.865776,
		0.804566,
		0.731919,
		0.649438,
		0.559361,
		0.464497,
		0.368169,
		0.27417,
		0.186723,
		0.110446,
		0.05033,
		0.011711,
		0.008564,
		0.043674,
		0.100747,
		0.174504,
		0.259999,
		0.352645,
		0.448238,
		0.542987,
		0.63355,
		0.717079,
		0.79127,
		0.854428,
		0.905532,
		0.944313,
		0.971319,
		0.987949,
		0.996405,
		0.999468,
		0.999993,
		0.999993,
		0.999468,
		0.996405,
		0.987949,
		0.971319,
		0.944313,
		0.905532,
		0.854428,
		0.79127,
		0.717079,
		0.63355,
		0.542987,
		0.448238,
		0.352645,
		0.259999,
		0.174504,
		0.100747,
		0.043674,
		0.008564,
		0.000244,
		0.018575,
		0.062044,
		0.125272,
		0.203198,
		0.291093,
		0.384583,
		0.47968,
		0.572806,
		0.660837,
		0.741144,
		0.811648,
		0.870879,
		0.918052,
		0.953131,
		0.976892,
		0.990939,
		0.997618,
		0.999741,
		1,
		0.999964,
		0.999069,
		0.99503,
		0.984999,
		0.966345,
		0.937074,
		0.895994,
		0.842738,
		0.777728,
		0.7021,
		0.617635,
		0.526696,
		0.432165,
		0.337397,
		0.246181,
		0.162702,
		0.091512,
		0.037511,
		0.00592,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.011138,
		0.047952,
		0.105431,
		0.178617,
		0.262862,
		0.353853,
		0.447628,
		0.540612,
		0.629643,
		0.712016,
		0.785527,
		0.848532,
		0.900005,
		0.939609,
		0.967759,
		0.985655,
		0.995256,
		0.999113,
		0.999966,
		1,
		0.999754,
		0.997729,
		0.991344,
		0.977883,
		0.95506,
		0.921283,
		0.875748,
		0.818422,
		0.749996,
		0.671812,
		0.585802,
		0.494427,
		0.400625,
		0.30777,
		0.219635,
		0.140358,
		0.074425,
		0.02664,
		0.002111,
		0.000934,
		0.021925,
		0.066571,
		0.12982,
		0.206915,
		0.293408,
		0.385184,
		0.478481,
		0.569922,
		0.656545,
		0.73585,
		0.805839,
		0.865077,
		0.912756,
		0.948756,
		0.973702,
		0.988991,
		0.996728,
		0.999517,
		0.999993,
		0.999993,
		0.999517,
		0.996728,
		0.988991,
		0.973702,
		0.948756,
		0.912756,
		0.865077,
		0.805839,
		0.73585,
		0.656545,
		0.569922,
		0.478481,
		0.385184,
		0.293408,
		0.206915,
		0.12982,
		0.066571,
		0.021925,
		0.000934,
		0.005631,
		0.035743,
		0.08736,
		0.155629,
		0.235987,
		0.324181,
		0.416289,
		0.508744,
		0.598361,
		0.682378,
		0.75849,
		0.824905,
		0.880394,
		0.924359,
		0.956891,
		0.978822,
		0.991726,
		0.997833,
		0.999765,
		1,
		0.999968,
		0.999154,
		0.99547,
		0.986276,
		0.9691,
		0.94202,
		0.903828,
		0.854068,
		0.793,
		0.721547,
		0.641225,
		0.55409,
		0.462682,
		0.369978,
		0.279358,
		0.19457,
		0.119705,
		0.059174,
		0.017685,
		0.000232,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.002008,
		0.025389,
		0.071053,
		0.134253,
		0.210511,
		0.295641,
		0.385763,
		0.477327,
		0.56714,
		0.652395,
		0.730708,
		0.800163,
		0.859359,
		0.907473,
		0.944313,
		0.970373,
		0.986864,
		0.995671,
		0.999193,
		0.999969,
		1,
		0.999776,
		0.997931,
		0.992087,
		0.979712,
		0.958631,
		0.927288,
		0.884829,
		0.831109,
		0.766643,
		0.692551,
		0.610497,
		0.522638,
		0.431573,
		0.340308,
		0.252217,
		0.171018,
		0.100747,
		0.045739,
		0.010606,
		0.007758,
		0.039695,
		0.091904,
		0.159817,
		0.239139,
		0.325863,
		0.416289,
		0.507048,
		0.595126,
		0.677898,
		0.753166,
		0.8192,
		0.874792,
		0.919314,
		0.952769,
		0.975845,
		0.989923,
		0.997015,
		0.999561,
		0.999994,
		0.999994,
		0.999561,
		0.997015,
		0.989923,
		0.975845,
		0.952769,
		0.919314,
		0.874792,
		0.8192,
		0.753166,
		0.677898,
		0.595126,
		0.507048,
		0.416289,
		0.325863,
		0.239139,
		0.159817,
		0.091904,
		0.039695,
		0.007758,
		0.000221,
		0.016857,
		0.056497,
		0.114498,
		0.186469,
		0.268292,
		0.356136,
		0.446476,
		0.536117,
		0.622222,
		0.702346,
		0.77447,
		0.837049,
		0.889065,
		0.930079,
		0.960286,
		0.980556,
		0.992429,
		0.998024,
		0.999787,
		1,
		0.999971,
		0.99923,
		0.995862,
		0.987421,
		0.971582,
		0.946495,
		0.91095,
		0.864419,
		0.807029,
		0.739516,
		0.663169,
		0.579774,
		0.491569,
		0.401201,
		0.311689,
		0.226396,
		0.149001,
		0.083483,
		0.034098,
		0.005363,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.010111,
		0.043674,
		0.096367,
		0.163887,
		0.242186,
		0.327485,
		0.416289,
		0.505413,
		0.592002,
		0.673559,
		0.747985,
		0.813614,
		0.869259,
		0.914268,
		0.948572,
		0.97273,
		0.987949,
		0.996042,
		0.999265,
		0.999972,
		1,
		0.999797,
		0.998111,
		0.992753,
		0.981357,
		0.961859,
		0.932739,
		0.893112,
		0.842738,
		0.781985,
		0.711778,
		0.63355,
		0.549189,
		0.460998,
		0.371655,
		0.284186,
		0.201932,
		0.128533,
		0.067904,
		0.024223,
		0.001913,
		0.000846,
		0.019939,
		0.060743,
		0.118882,
		0.190218,
		0.270869,
		0.357216,
		0.44593,
		0.533987,
		0.618696,
		0.697728,
		0.769149,
		0.831466,
		0.883666,
		0.925275,
		0.956401,
		0.977776,
		0.990759,
		0.997272,
		0.9996,
		0.999994,
		0.999994,
		0.9996,
		0.997272,
		0.990759,
		0.977776,
		0.956401,
		0.925275,
		0.883666,
		0.831466,
		0.769149,
		0.697728,
		0.618696,
		0.533987,
		0.44593,
		0.357216,
		0.270869,
		0.190218,
		0.118882,
		0.060743,
		0.019939,
		0.000846,
		0.005114,
		0.032563,
		0.079857,
		0.142782,
		0.217363,
		0.299874,
		0.386858,
		0.475143,
		0.561865,
		0.644493,
		0.720862,
		0.789202,
		0.848188,
		0.89698,
		0.935275,
		0.963355,
		0.982117,
		0.99306,
		0.998194,
		0.999806,
		1,
		0.999973,
		0.999297,
		0.996213,
		0.98845,
		0.973821,
		0.950551,
		0.917436,
		0.87389,
		0.819929,
		0.756129,
		0.683581,
		0.603835,
		0.518865,
		0.431024,
		0.343012,
		0.25785,
		0.178855,
		0.10962,
		0.053998,
		0.016085,
		0.000211,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.001825,
		0.023136,
		0.064959,
		0.123166,
		0.193856,
		0.273359,
		0.358259,
		0.445405,
		0.531932,
		0.615286,
		0.693246,
		0.763962,
		0.825987,
		0.878322,
		0.920461,
		0.952436,
		0.974858,
		0.988925,
		0.996374,
		0.999328,
		0.999974,
		1,
		0.999814,
		0.998273,
		0.99335,
		0.982839,
		0.96478,
		0.937694,
		0.900677,
		0.85341,
		0.796135,
		0.72961,
		0.655065,
		0.574153,
		0.488913,
		0.401737,
		0.315337,
		0.232725,
		0.157186,
		0.092264,
		0.041747,
		0.00965,
		0.00706,
		0.036235,
		0.084169,
		0.146884,
		0.220628,
		0.301882,
		0.387376,
		0.474108,
		0.559361,
		0.64073,
		0.716147,
		0.783913,
		0.842738,
		0.891784,
		0.930704,
		0.959694,
		0.979518,
		0.99151,
		0.997501,
		0.999635,
		0.999995,
		0.999995,
		0.999635,
		0.997501,
		0.99151,
		0.979518,
		0.959694,
		0.930704,
		0.891784,
		0.842738,
		0.783913,
		0.716147,
		0.64073,
		0.559361,
		0.474108,
		0.387376,
		0.301882,
		0.220628,
		0.146884,
		0.084169,
		0.036235,
		0.00706,
		0.000201,
		0.015366,
		0.05166,
		0.105044,
		0.171689,
		0.247986,
		0.330561,
		0.416289,
		0.502313,
		0.586062,
		0.665277,
		0.738037,
		0.802795,
		0.858415,
		0.904213,
		0.940003,
		0.966137,
		0.983525,
		0.993626,
		0.998347,
		0.999823,
		1,
		0.999976,
		0.999358,
		0.996527,
		0.989376,
		0.975845,
		0.954233,
		0.92335,
		0.882566,
		0.831801,
		0.771497,
		0.70257,
		0.626365,
		0.544621,
		0.459432,
		0.373216,
		0.28869,
		0.208847,
		0.13694,
		0.076461,
		0.031129,
		0.004882,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.00922,
		0.039943,
		0.088416,
		0.150882,
		0.223793,
		0.303823,
		0.387877,
		0.473108,
		0.556941,
		0.637084,
		0.711563,
		0.778747,
		0.837381,
		0.886629,
		0.926112,
		0.955947,
		0.976785,
		0.989805,
		0.996672,
		0.999385,
		0.999977,
		1,
		0.99983,
		0.998417,
		0.993888,
		0.984178,
		0.967429,
		0.942207,
		0.907596,
		0.863213,
		0.809195,
		0.746155,
		0.675142,
		0.597605,
		0.51535,
		0.430513,
		0.345531,
		0.263117,
		0.186245,
		0.118126,
		0.0622,
		0.022121,
		0.001742,
		0.000771,
		0.018211,
		0.055645,
		0.109257,
		0.175429,
		0.250747,
		0.332021,
		0.416289,
		0.500843,
		0.583238,
		0.661323,
		0.73326,
		0.797559,
		0.853109,
		0.899218,
		0.935655,
		0.962684,
		0.981094,
		0.992187,
		0.997707,
		0.999666,
		0.999995,
		0.999995,
		0.999666,
		0.997707,
		0.992187,
		0.981094,
		0.962684,
		0.935655,
		0.899218,
		0.853109,
		0.797559,
		0.73326,
		0.661323,
		0.583238,
		0.500843,
		0.416289,
		0.332021,
		0.250747,
		0.175429,
		0.109257,
		0.055645,
		0.018211,
		0.000771,
		0.004665,
		0.029788,
		0.073275,
		0.131445,
		0.200812,
		0.278096,
		0.360237,
		0.444407,
		0.528027,
		0.608789,
		0.684672,
		0.753975,
		0.815347,
		0.867815,
		0.910833,
		0.944313,
		0.968661,
		0.984799,
		0.994136,
		0.998483,
		0.999838,
		1,
		0.999978,
		0.999412,
		0.99681,
		0.990212,
		0.977679,
		0.957583,
		0.928752,
		0.890522,
		0.842738,
		0.785723,
		0.720239,
		0.647454,
		0.568899,
		0.486437,
		0.402235,
		0.318742,
		0.238661,
		0.164939,
		0.100747,
		0.049471,
		0.014693,
		0.000192,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.001665,
		0.02117,
		0.059613,
		0.113386,
		0.179066,
		0.253422,
		0.333432,
		0.416289,
		0.499421,
		0.580504,
		0.657487,
		0.72861,
		0.792437,
		0.847883,
		0.894253,
		0.931276,
		0.959144,
		0.978531,
		0.990599,
		0.99694,
		0.999437,
		0.999979,
		1,
		0.999845,
		0.998546,
		0.994372,
		0.985389,
		0.969836,
		0.946324,
		0.913932,
		0.87223,
		0.82126,
		0.761512,
		0.693878,
		0.619625,
		0.540354,
		0.457972,
		0.374671,
		0.292901,
		0.215353,
		0.144944,
		0.084802,
		0.038254,
		0.008818,
		0.006452,
		0.033208,
		0.077365,
		0.135442,
		0.204139,
		0.280351,
		0.361176,
		0.443933,
		0.526172,
		0.605693,
		0.680569,
		0.749168,
		0.810183,
		0.862659,
		0.906036,
		0.940176,
		0.965405,
		0.982523,
		0.992798,
		0.997893,
		0.999694,
		0.999996,
		0.999996,
		0.999694,
		0.997893,
		0.992798,
		0.982523,
		0.965405,
		0.940176,
		0.906036,
		0.862659,
		0.810183,
		0.749168,
		0.680569,
		0.605693,
		0.526172,
		0.443933,
		0.361176,
		0.280351,
		0.204139,
		0.135442,
		0.077365,
		0.033208,
		0.006452,
		0.000183,
		0.014064,
		0.047417,
		0.096707,
		0.158573,
		0.229837,
		0.307516,
		0.388827,
		0.47121,
		0.552335,
		0.630124,
		0.702773,
		0.768774,
		0.826947,
		0.876466,
		0.9169,
		0.948246,
		0.970957,
		0.985952,
		0.994596,
		0.998606,
		0.999851,
		1,
		0.99998,
		0.99946,
		0.997064,
		0.990967,
		0.979343,
		0.960634,
		0.933692,
		0.897829,
		0.852824,
		0.798899,
		0.736686,
		0.667192,
		0.591766,
		0.512066,
		0.430036,
		0.347883,
		0.268054,
		0.193222,
		0.126271,
		0.070284,
		0.028532,
		0.004462,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.008442,
		0.03667,
		0.081404,
		0.139346,
		0.207371,
		0.282534,
		0.362085,
		0.443475,
		0.524376,
		0.602692,
		0.676582,
		0.74448,
		0.805122,
		0.857572,
		0.901258,
		0.936004,
		0.962058,
		0.980117,
		0.991318,
		0.997182,
		0.999482,
		0.999981,
		1,
		0.999858,
		0.998663,
		0.994809,
		0.986488,
		0.972026,
		0.950084,
		0.919743,
		0.880531,
		0.832415,
		0.775775,
		0.711367,
		0.640295,
		0.563978,
		0.484124,
		0.402701,
		0.321925,
		0.244237,
		0.172287,
		0.108924,
		0.057184,
		0.02028,
		0.001593,
		0.000705,
		0.016698,
		0.051161,
		0.100747,
		0.162273,
		0.232725,
		0.309274,
		0.389279,
		0.470308,
		0.550142,
		0.6268,
		0.698557,
		0.763962,
		0.821869,
		0.871464,
		0.912296,
		0.944313,
		0.967884,
		0.983819,
		0.99335,
		0.99806,
		0.999719,
		0.999996,
		0.999996,
		0.999719,
		0.99806,
		0.99335,
		0.983819,
		0.967884,
		0.944313,
		0.912296,
		0.871464,
		0.821869,
		0.763962,
		0.698557,
		0.6268,
		0.550142,
		0.470308,
		0.389279,
		0.309274,
		0.232725,
		0.162273,
		0.100747,
		0.051161,
		0.016698,
		0.000705,
		0.004273,
		0.027353,
		0.067471,
		0.121393,
		0.186045,
		0.258527,
		0.336118,
		0.416289,
		0.496716,
		0.575292,
		0.650148,
		0.71967,
		0.782523,
		0.837676,
		0.884434,
		0.922467,
		0.951842,
		0.973047,
		0.986998,
		0.995012,
		0.998717,
		0.999863,
		1,
		0.999981,
		0.999504,
		0.997294,
		0.991651,
		0.980856,
		0.963419,
		0.938217,
		0.904547,
		0.862134,
		0.811114,
		0.752002,
		0.685666,
		0.613292,
		0.536358,
		0.456608,
		0.376031,
		0.296847,
		0.221482,
		0.152563,
		0.092903,
		0.045488,
		0.013475,
		0.000176,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.001526,
		0.019444,
		0.054899,
		0.104717,
		0.165879,
		0.235528,
		0.310976,
		0.389717,
		0.469434,
		0.548017,
		0.623575,
		0.694455,
		0.759263,
		0.816885,
		0.86652,
		0.907703,
		0.940336,
		0.96472,
		0.981561,
		0.991969,
		0.9974,
		0.999524,
		0.999982,
		1,
		0.999869,
		0.998768,
		0.995205,
		0.987485,
		0.974023,
		0.953525,
		0.925079,
		0.888183,
		0.842738,
		0.78903,
		0.727694,
		0.659694,
		0.586284,
		0.508991,
		0.42959,
		0.350084,
		0.272688,
		0.199814,
		0.134062,
		0.078205,
		0.035182,
		0.008089,
		0.00592,
		0.030544,
		0.07135,
		0.125272,
		0.189395,
		0.260965,
		0.337397,
		0.416289,
		0.495428,
		0.572806,
		0.646637,
		0.715374,
		0.777728,
		0.832697,
		0.879589,
		0.918052,
		0.948101,
		0.970147,
		0.984999,
		0.993851,
		0.99821,
		0.999741,
		0.999996,
		0.999996,
		0.999741,
		0.99821,
		0.993851,
		0.984999,
		0.970147,
		0.948101,
		0.918052,
		0.879589,
		0.832697,
		0.777728,
		0.715374,
		0.646637,
		0.572806,
		0.495428,
		0.416289,
		0.337397,
		0.260965,
		0.189395,
		0.125272,
		0.07135,
		0.030544,
		0.00592,
		0.000168,
		0.012921,
		0.043674,
		0.089317,
		0.146884,
		0.213563,
		0.286701,
		0.363815,
		0.442603,
		0.520954,
		0.596961,
		0.66894,
		0.735449,
		0.795304,
		0.84761,
		0.891784,
		0.927583,
		0.955135,
		0.974954,
		0.987949,
		0.995389,
		0.998817,
		0.999874,
		1,
		0.999983,
		0.999543,
		0.997501,
		0.992272,
		0.982234,
		0.965963,
		0.942367,
		0.910731,
		0.870737,
		0.822445,
		0.766271,
		0.702959,
		0.63355,
		0.559361,
		0.481957,
		0.403137,
		0.324909,
		0.249484,
		0.179254,
		0.11679,
		0.064824,
		0.026246,
		0.004095,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.007758,
		0.033782,
		0.07519,
		0.12907,
		0.192655,
		0.263329,
		0.338636,
		0.416289,
		0.49418,
		0.570395,
		0.643226,
		0.711187,
		0.773038,
		0.827802,
		0.874792,
		0.913638,
		0.944313,
		0.967154,
		0.982876,
		0.992561,
		0.997598,
		0.999561,
		0.999984,
		1,
		0.99988,
		0.998863,
		0.995564,
		0.988393,
		0.975845,
		0.956676,
		0.929985,
		0.895243,
		0.8523,
		0.801356,
		0.742944,
		0.677898,
		0.607332,
		0.532609,
		0.455329,
		0.377306,
		0.300551,
		0.227264,
		0.159817,
		0.100747,
		0.052748,
		0.01866,
		0.001462,
		0.000647,
		0.015366,
		0.047197,
		0.093187,
		0.150522,
		0.21653,
		0.28869,
		0.36464,
		0.442188,
		0.519322,
		0.594222,
		0.665277,
		0.731099,
		0.790544,
		0.842738,
		0.887096,
		0.92335,
		0.951577,
		0.972215,
		0.986073,
		0.994306,
		0.998347,
		0.999762,
		0.999997,
		0.999997,
		0.999762,
		0.998347,
		0.994306,
		0.986073,
		0.972215,
		0.951577,
		0.92335,
		0.887096,
		0.842738,
		0.790544,
		0.731099,
		0.665277,
		0.594222,
		0.519322,
		0.442188,
		0.36464,
		0.28869,
		0.21653,
		0.150522,
		0.093187,
		0.047197,
		0.015366,
		0.000647,
		0.003928,
		0.025204,
		0.062328,
		0.112441,
		0.172822,
		0.240892,
		0.314225,
		0.39055,
		0.467769,
		0.543963,
		0.617404,
		0.686577,
		0.750188,
		0.807193,
		0.856815,
		0.898568,
		0.932289,
		0.958153,
		0.976697,
		0.988816,
		0.995731,
		0.998907,
		0.999884,
		1,
		0.999984,
		0.999578,
		0.997689,
		0.992836,
		0.98349,
		0.968292,
		0.946178,
		0.91643,
		0.878693,
		0.832964,
		0.779572,
		0.719149,
		0.65261,
		0.581127,
		0.506107,
		0.429172,
		0.352149,
		0.277047,
		0.206051,
		0.141513,
		0.085934,
		0.041967,
		0.012401,
		0.000161,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.001403,
		0.017921,
		0.050721,
		0.096998,
		0.154076,
		0.219416,
		0.290619,
		0.365439,
		0.441786,
		0.51774,
		0.591564,
		0.661713,
		0.726854,
		0.785882,
		0.837941,
		0.882447,
		0.919112,
		0.947967,
		0.969383,
		0.984077,
		0.993099,
		0.997777,
		0.999595,
		0.999985,
		1,
		0.999889,
		0.998949,
		0.99589,
		0.98922,
		0.977512,
		0.959568,
		0.934501,
		0.901765,
		0.861163,
		0.812825,
		0.757191,
		0.694984,
		0.627186,
		0.55502,
		0.479925,
		0.403546,
		0.327712,
		0.254429,
		0.185867,
		0.124349,
		0.072345,
		0.032464,
		0.007447,
		0.00545,
		0.028187,
		0.066007,
		0.116196,
		0.176164,
		0.243461,
		0.315776,
		0.390948,
		0.466975,
		0.542026,
		0.61445,
		0.682792,
		0.745808,
		0.802484,
		0.852057,
		0.894038,
		0.928233,
		0.954769,
		0.974109,
		0.987053,
		0.994719,
		0.99847,
		0.99978,
		0.999997,
		0.999997,
		0.99978,
		0.99847,
		0.994719,
		0.987053,
		0.974109,
		0.954769,
		0.928233,
		0.894038,
		0.852057,
		0.802484,
		0.745808,
		0.682792,
		0.61445,
		0.542026,
		0.466975,
		0.390948,
		0.315776,
		0.243461,
		0.176164,
		0.116196,
		0.066007,
		0.028187,
		0.00545,
		0.000155,
		0.011912,
		0.040357,
		0.082738,
		0.136427,
		0.198921,
		0.267854,
		0.341002,
		0.416289,
		0.4918,
		0.565786,
		0.636686,
		0.70313,
		0.763962,
		0.81826,
		0.865352,
		0.904839,
		0.936624,
		0.960924,
		0.978292,
		0.989606,
		0.996042,
		0.998989,
		0.999893,
		1,
		0.999985,
		0.99961,
		0.99786,
		0.99335,
		0.984638,
		0.970427,
		0.949683,
		0.921688,
		0.886059,
		0.842738,
		0.791976,
		0.734311,
		0.670542,
		0.601712,
		0.529086,
		0.45413,
		0.378502,
		0.304035,
		0.232725,
		0.166724,
		0.108329,
		0.059973,
		0.024223,
		0.003771,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.007154,
		0.031222,
		0.069658,
		0.119879,
		0.179424,
		0.245958,
		0.317281,
		0.391333,
		0.466205,
		0.540147,
		0.61158,
		0.679106,
		0.741528,
		0.797864,
		0.847365,
		0.889537,
		0.924165,
		0.95133,
		0.971428,
		0.985175,
		0.99359,
		0.99794,
		0.999625,
		0.999986,
		1,
		0.999897,
		0.999027,
		0.996187,
		0.989975,
		0.979038,
		0.962225,
		0.938663,
		0.907796,
		0.869387,
		0.823506,
		0.770509,
		0.711022,
		0.645911,
		0.576269,
		0.503396,
		0.42878,
		0.354089,
		0.281153,
		0.211958,
		0.148635,
		0.093452,
		0.048808,
		0.017226,
		0.001347,
		0.000596,
		0.014187,
		0.043674,
		0.086442,
		0.139988,
		0.201932,
		0.27002,
		0.342132,
		0.416289,
		0.490663,
		0.563582,
		0.63355,
		0.699252,
		0.759572,
		0.813614,
		0.860713,
		0.900464,
		0.932739,
		0.957705,
		0.975845,
		0.987949,
		0.995096,
		0.998583,
		0.999797,
		0.999997,
		0.999997,
		0.999797,
		0.998583,
		0.995096,
		0.987949,
		0.975845,
		0.957705,
		0.932739,
		0.900464,
		0.860713,
		0.813614,
		0.759572,
		0.699252,
		0.63355,
		0.563582,
		0.490663,
		0.416289,
		0.342132,
		0.27002,
		0.201932,
		0.139988,
		0.086442,
		0.043674,
		0.014187,
		0.000596,
		0.003623,
		0.023299,
		0.057749,
		0.104436,
		0.160938,
		0.224955,
		0.294311,
		0.366965,
		0.441017,
		0.514717,
		0.586474,
		0.654869,
		0.718669,
		0.776841,
		0.828569,
		0.873276,
		0.910641,
		0.940621,
		0.963472,
		0.979753,
		0.990328,
		0.996325,
		0.999064,
		0.999901,
		1,
		0.999987,
		0.999639,
		0.998016,
		0.993819,
		0.985689,
		0.972387,
		0.95291,
		0.926546,
		0.892886,
		0.851827,
		0.803552,
		0.748515,
		0.687413,
		0.621173,
		0.550931,
		0.478015,
		0.40393,
		0.330348,
		0.259096,
		0.192146,
		0.131607,
		0.079717,
		0.038838,
		0.011451,
		0.000148,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.001294,
		0.01657,
		0.047001,
		0.090096,
		0.143473,
		0.204865,
		0.272124,
		0.343228,
		0.416289,
		0.489559,
		0.561441,
		0.630499,
		0.69547,
		0.755278,
		0.809049,
		0.856131,
		0.896111,
		0.928835,
		0.954428,
		0.973306,
		0.986181,
		0.994038,
		0.998088,
		0.999653,
		0.999987,
		1,
		0.999905,
		0.999098,
		0.996457,
		0.990665,
		0.980438,
		0.964669,
		0.942504,
		0.913379,
		0.877025,
		0.833458,
		0.782964,
		0.726079,
		0.663568,
		0.596407,
		0.525768,
		0.453002,
		0.379627,
		0.307317,
		0.237891,
		0.173305,
		0.115644,
		0.067117,
		0.030049,
		0.006878,
		0.005035,
		0.026093,
		0.061241,
		0.108062,
		0.164251,
		0.227614,
		0.296078,
		0.367695,
		0.44065,
		0.513271,
		0.584036,
		0.651582,
		0.714722,
		0.772459,
		0.823996,
		0.86876,
		0.906419,
		0.9369,
		0.960409,
		0.977439,
		0.98877,
		0.99544,
		0.998685,
		0.999812,
		0.999997,
		0.999997,
		0.999812,
		0.998685,
		0.99544,
		0.98877,
		0.977439,
		0.960409,
		0.9369,
		0.906419,
		0.86876,
		0.823996,
		0.772459,
		0.714722,
		0.651582,
		0.584036,
		0.513271,
		0.44065,
		0.367695,
		0.296078,
		0.227614,
		0.164251,
		0.108062,
		0.061241,
		0.026093,
		0.005035,
		0.000143,
		0.011016,
		0.037403,
		0.076857,
		0.127035,
		0.185705,
		0.250747,
		0.32016,
		0.39207,
		0.464734,
		0.536552,
		0.606074,
		0.672015,
		0.73326,
		0.788888,
		0.838179,
		0.880639,
		0.916014,
		0.944313,
		0.965817,
		0.981094,
		0.990988,
		0.996584,
		0.999131,
		0.999909,
		1,
		0.999988,
		0.999666,
		0.998157,
		0.994248,
		0.986651,
		0.974188,
		0.955885,
		0.931037,
		0.899218,
		0.860284,
		0.814362,
		0.761826,
		0.703287,
		0.639566,
		0.571684,
		0.500843,
		0.428411,
		0.355915,
		0.285028,
		0.217558,
		0.155444,
		0.100747,
		0.055645,
		0.022426,
		0.003484,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.006618,
		0.028941,
		0.064712,
		0.111627,
		0.167488,
		0.230203,
		0.297795,
		0.368403,
		0.440294,
		0.511868,
		0.581665,
		0.64838,
		0.710869,
		0.768167,
		0.819497,
		0.864293,
		0.902212,
		0.933156,
		0.957286,
		0.975034,
		0.987103,
		0.994448,
		0.998223,
		0.999678,
		0.999988,
		1,
		0.999912,
		0.999163,
		0.996704,
		0.991297,
		0.981724,
		0.96692,
		0.946052,
		0.918552,
		0.884124,
		0.842738,
		0.794618,
		0.74022,
		0.680218,
		0.615483,
		0.547074,
		0.476215,
		0.404291,
		0.332833,
		0.263508,
		0.198115,
		0.138573,
		0.086916,
		0.045292,
		0.015951,
		0.001245,
		0.000551,
		0.013138,
		0.040531,
		0.080399,
		0.130509,
		0.188732,
		0.253045,
		0.321539,
		0.392422,
		0.46403,
		0.53483,
		0.603434,
		0.668603,
		0.729267,
		0.784529,
		0.833687,
		0.876247,
		0.911942,
		0.940749,
		0.962903,
		0.978905,
		0.989522,
		0.995754,
		0.998778,
		0.999825,
		0.999998,
		0.999998,
		0.999825,
		0.998778,
		0.995754,
		0.989522,
		0.978905,
		0.962903,
		0.940749,
		0.911942,
		0.876247,
		0.833687,
		0.784529,
		0.729267,
		0.668603,
		0.603434,
		0.53483,
		0.46403,
		0.392422,
		0.321539,
		0.253045,
		0.188732,
		0.130509,
		0.080399,
		0.040531,
		0.013138,
		0.000551,
		0.003352,
		0.021601,
		0.053654,
		0.09725,
		0.150223,
		0.210511,
		0.27616,
		0.345327,
		0.416289,
		0.487448,
		0.557338,
		0.62464,
		0.688183,
		0.746966,
		0.800163,
		0.847143,
		0.887485,
		0.920996,
		0.947725,
		0.967979,
		0.982327,
		0.991594,
		0.99682,
		0.999193,
		0.999915,
		1,
		0.999988,
		0.99969,
		0.998286,
		0.99464,
		0.987535,
		0.975845,
		0.958631,
		0.935195,
		0.905097,
		0.868161,
		0.824461,
		0.774307,
		0.718226,
		0.656948,
		0.591389,
		0.522638,
		0.451939,
		0.380687,
		0.310415,
		0.242783,
		0.179577,
		0.122695,
		0.074147,
		0.036046,
		0.010606,
		0.000137,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.001198,
		0.015366,
		0.043674,
		0.083901,
		0.133916,
		0.191686,
		0.255282,
		0.322879,
		0.392764,
		0.463346,
		0.533157,
		0.600863,
		0.665277,
		0.725363,
		0.780253,
		0.829262,
		0.871897,
		0.907879,
		0.937158,
		0.959925,
		0.976625,
		0.987949,
		0.994823,
		0.998347,
		0.999701,
		0.999989,
		1,
		0.999918,
		0.999222,
		0.99693,
		0.991877,
		0.982906,
		0.968996,
		0.949334,
		0.92335,
		0.890728,
		0.851398,
		0.805528,
		0.753505,
		0.695919,
		0.63355,
		0.567351,
		0.498434,
		0.428063,
		0.357638,
		0.28869,
		0.222873,
		0.161955,
		0.107814,
		0.062433,
		0.027894,
		0.006372,
		0.004665,
		0.024223,
		0.05697,
		0.100747,
		0.15349,
		0.213228,
		0.278096,
		0.346332,
		0.416289,
		0.486437,
		0.555372,
		0.621825,
		0.684672,
		0.742944,
		0.795839,
		0.842738,
		0.883219,
		0.917071,
		0.944313,
		0.965204,
		0.980254,
		0.990212,
		0.996042,
		0.998863,
		0.999838,
		0.999998,
		0.999998,
		0.999838,
		0.998863,
		0.996042,
		0.990212,
		0.980254,
		0.965204,
		0.944313,
		0.917071,
		0.883219,
		0.842738,
		0.795839,
		0.742944,
		0.684672,
		0.621825,
		0.555372,
		0.486437,
		0.416289,
		0.346332,
		0.278096,
		0.213228,
		0.15349,
		0.100747,
		0.05697,
		0.024223,
		0.004665,
		0.000132,
		0.010218,
		0.034761,
		0.071577,
		0.118572,
		0.173742,
		0.235182,
		0.301087,
		0.369759,
		0.439612,
		0.509177,
		0.577116,
		0.642221,
		0.703432,
		0.759844,
		0.810721,
		0.855511,
		0.893857,
		0.925619,
		0.950882,
		0.969973,
		0.983462,
		0.992149,
		0.997036,
		0.999249,
		0.999921,
		1,
		0.999989,
		0.999712,
		0.998404,
		0.994999,
		0.988347,
		0.977373,
		0.961169,
		0.939048,
		0.910561,
		0.875503,
		0.833904,
		0.786014,
		0.732288,
		0.673374,
		0.610092,
		0.543429,
		0.474517,
		0.404632,
		0.335179,
		0.267685,
		0.203793,
		0.145256,
		0.09393,
		0.051768,
		0.020821,
		0.003228,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.00614,
		0.026902,
		0.060273,
		0.104192,
		0.156687,
		0.215878,
		0.27998,
		0.34731,
		0.416289,
		0.485454,
		0.553458,
		0.619082,
		0.681244,
		0.739008,
		0.791594,
		0.838394,
		0.878987,
		0.913148,
		0.940869,
		0.962366,
		0.978092,
		0.988728,
		0.995167,
		0.99846,
		0.999722,
		0.99999,
		1,
		0.999924,
		0.999276,
		0.997137,
		0.99241,
		0.983995,
		0.970913,
		0.952372,
		0.927805,
		0.896877,
		0.859486,
		0.815748,
		0.76599,
		0.710728,
		0.650658,
		0.586639,
		0.51968,
		0.450936,
		0.381687,
		0.313342,
		0.247423,
		0.185559,
		0.129487,
		0.081039,
		0.042142,
		0.014812,
		0.001154,
		0.000511,
		0.012202,
		0.037715,
		0.074965,
		0.121952,
		0.176763,
		0.237576,
		0.302666,
		0.370409,
		0.439285,
		0.507888,
		0.574932,
		0.639258,
		0.699842,
		0.75581,
		0.806444,
		0.851199,
		0.889716,
		0.921837,
		0.947615,
		0.967332,
		0.981498,
		0.990846,
		0.996306,
		0.998941,
		0.999849,
		0.999998,
		0.999998,
		0.999849,
		0.998941,
		0.996306,
		0.990846,
		0.981498,
		0.967332,
		0.947615,
		0.921837,
		0.889716,
		0.851199,
		0.806444,
		0.75581,
		0.699842,
		0.639258,
		0.574932,
		0.507888,
		0.439285,
		0.370409,
		0.302666,
		0.237576,
		0.176763,
		0.121952,
		0.074965,
		0.037715,
		0.012202,
		0.000511,
		0.003111,
		0.020082,
		0.049979,
		0.090775,
		0.140529,
		0.197385,
		0.259581,
		0.325448,
		0.39342,
		0.462036,
		0.529947,
		0.595924,
		0.658868,
		0.717815,
		0.771949,
		0.820615,
		0.863328,
		0.899793,
		0.929913,
		0.953807,
		0.971816,
		0.984507,
		0.99266,
		0.997234,
		0.999301,
		0.999927,
		1,
		0.99999,
		0.999732,
		0.998512,
		0.995329,
		0.989094,
		0.978782,
		0.963517,
		0.942622,
		0.915644,
		0.882352,
		0.842738,
		0.797,
		0.745529,
		0.688896,
		0.627839,
		0.563249,
		0.496159,
		0.427734,
		0.359265,
		0.292156,
		0.227922,
		0.168182,
		0.114651,
		0.069138,
		0.033544,
		0.009851,
		0.000127,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.001112,
		0.014288,
		0.040688,
		0.07832,
		0.125272,
		0.179716,
		0.23991,
		0.304203,
		0.371041,
		0.438967,
		0.506634,
		0.572806,
		0.636368,
		0.696335,
		0.751857,
		0.802239,
		0.846941,
		0.885604,
		0.918052,
		0.944313,
		0.964625,
		0.979446,
		0.989445,
		0.995483,
		0.998563,
		0.999741,
		0.99999,
		1,
		0.999929,
		0.999325,
		0.997327,
		0.992899,
		0.984999,
		0.972685,
		0.955189,
		0.931946,
		0.902608,
		0.867043,
		0.825327,
		0.777728,
		0.724699,
		0.666857,
		0.604978,
		0.539979,
		0.472913,
		0.404954,
		0.337397,
		0.271644,
		0.2092,
		0.151669,
		0.100747,
		0.058222,
		0.025962,
		0.00592,
		0.004334,
		0.022547,
		0.05313,
		0.094146,
		0.143738,
		0.200135,
		0.261647,
		0.326681,
		0.393734,
		0.461407,
		0.528406,
		0.59355,
		0.655781,
		0.714166,
		0.767917,
		0.816392,
		0.859113,
		0.895777,
		0.92627,
		0.950679,
		0.969301,
		0.982646,
		0.991431,
		0.996548,
		0.999012,
		0.99986,
		0.999998,
		0.999998,
		0.99986,
		0.999012,
		0.996548,
		0.991431,
		0.982646,
		0.969301,
		0.950679,
		0.92627,
		0.895777,
		0.859113,
		0.816392,
		0.767917,
		0.714166,
		0.655781,
		0.59355,
		0.528406,
		0.461407,
		0.393734,
		0.326681,
		0.261647,
		0.200135,
		0.143738,
		0.094146,
		0.05313,
		0.022547,
		0.004334,
		0.000123,
		0.009504,
		0.032389,
		0.066821,
		0.110919,
		0.16288,
		0.220985,
		0.283601,
		0.349184,
		0.416289,
		0.483569,
		0.549783,
		0.613805,
		0.674631,
		0.731385,
		0.783333,
		0.82989,
		0.870636,
		0.905327,
		0.933906,
		0.95652,
		0.973521,
		0.985471,
		0.99313,
		0.997416,
		0.999348,
		0.999932,
		1,
		0.999991,
		0.99975,
		0.998612,
		0.995632,
		0.989783,
		0.980084,
		0.965692,
		0.945942,
		0.920376,
		0.888746,
		0.851009,
		0.807315,
		0.758,
		0.703566,
		0.644676,
		0.582134,
		0.516882,
		0.449987,
		0.382633,
		0.316114,
		0.251827,
		0.191269,
		0.136026,
		0.087777,
		0.048281,
		0.019382,
		0.003,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.005711,
		0.02507,
		0.056273,
		0.09747,
		0.146884,
		0.20282,
		0.263662,
		0.327881,
		0.39404,
		0.460796,
		0.526907,
		0.591237,
		0.652767,
		0.710598,
		0.763962,
		0.812236,
		0.854945,
		0.891784,
		0.92262,
		0.947512,
		0.966719,
		0.980698,
		0.990107,
		0.995774,
		0.998658,
		0.999759,
		0.999991,
		1,
		0.999934,
		0.99937,
		0.997501,
		0.99335,
		0.985926,
		0.974325,
		0.957802,
		0.935798,
		0.907953,
		0.874111,
		0.834309,
		0.788769,
		0.737881,
		0.682196,
		0.622411,
		0.559361,
		0.494005,
		0.427424,
		0.360804,
		0.295441,
		0.232725,
		0.17414,
		0.121258,
		0.075735,
		0.039308,
		0.013791,
		0.001072,
		0.000475,
		0.011362,
		0.035182,
		0.070061,
		0.114202,
		0.165879,
		0.223446,
		0.285341,
		0.350084,
		0.416289,
		0.482664,
		0.548017,
		0.611266,
		0.67144,
		0.727694,
		0.779315,
		0.825729,
		0.86652,
		0.901435,
		0.930398,
		0.953525,
		0.971124,
		0.983707,
		0.991969,
		0.996771,
		0.999078,
		0.999869,
		0.999998,
		0.999998,
		0.999869,
		0.999078,
		0.996771,
		0.991969,
		0.983707,
		0.971124,
		0.953525,
		0.930398,
		0.901435,
		0.86652,
		0.825729,
		0.779315,
		0.727694,
		0.67144,
		0.611266,
		0.548017,
		0.482664,
		0.416289,
		0.350084,
		0.285341,
		0.223446,
		0.165879,
		0.114202,
		0.070061,
		0.035182,
		0.011362,
		0.000475,
		0.002895,
		0.018718,
		0.046668,
		0.084923,
		0.131734,
		0.185427,
		0.244406,
		0.307157,
		0.372254,
		0.438357,
		0.504225,
		0.568718,
		0.630801,
		0.689557,
		0.744191,
		0.794042,
		0.838591,
		0.877473,
		0.91049,
		0.937623,
		0.959038,
		0.9751,
		0.986362,
		0.993563,
		0.997583,
		0.999391,
		0.999937,
		1,
		0.999991,
		0.999767,
		0.998703,
		0.995911,
		0.990418,
		0.981289,
		0.967708,
		0.949027,
		0.924786,
		0.89472,
		0.858756,
		0.817006,
		0.769751,
		0.717434,
		0.660647,
		0.60012,
		0.536711,
		0.471394,
		0.405259,
		0.339498,
		0.275401,
		0.214352,
		0.157821,
		0.107365,
		0.064618,
		0.031293,
		0.009174,
		0.000118,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.001035,
		0.01332,
		0.037997,
		0.073275,
		0.117431,
		0.168815,
		0.225849,
		0.287037,
		0.35096,
		0.416289,
		0.481783,
		0.546297,
		0.608789,
		0.668323,
		0.724081,
		0.77537,
		0.82163,
		0.862446,
		0.897559,
		0.926878,
		0.950488,
		0.968661,
		0.981857,
		0.990717,
		0.996042,
		0.998745,
		0.999775,
		0.999992,
		1,
		0.999939,
		0.999412,
		0.997662,
		0.993766,
		0.986782,
		0.975845,
		0.96023,
		0.939384,
		0.912942,
		0.880725,
		0.842738,
		0.799158,
		0.750323,
		0.69672,
		0.638979,
		0.577857,
		0.51423,
		0.449089,
		0.383529,
		0.318742,
		0.256014,
		0.196721,
		0.14232,
		0.094348,
		0.054421,
		0.024223,
		0.005514,
		0.004038,
		0.02104,
		0.049664,
		0.088168,
		0.134877,
		0.188187,
		0.246571,
		0.308577,
		0.372836,
		0.438064,
		0.503068,
		0.566751,
		0.628118,
		0.686282,
		0.740474,
		0.790049,
		0.834498,
		0.873458,
		0.90672,
		0.934245,
		0.95617,
		0.972816,
		0.984688,
		0.992466,
		0.996976,
		0.999138,
		0.999878,
		0.999998,
		0.999998,
		0.999878,
		0.999138,
		0.996976,
		0.992466,
		0.984688,
		0.972816,
		0.95617,
		0.934245,
		0.90672,
		0.873458,
		0.834498,
		0.790049,
		0.740474,
		0.686282,
		0.628118,
		0.566751,
		0.503068,
		0.438064,
		0.372836,
		0.308577,
		0.246571,
		0.188187,
		0.134877,
		0.088168,
		0.049664,
		0.02104,
		0.004038,
		0.000114,
		0.008861,
		0.030251,
		0.062522,
		0.103978,
		0.152992,
		0.208006,
		0.26754,
		0.330188,
		0.394628,
		0.459622,
		0.524024,
		0.586783,
		0.646952,
		0.703692,
		0.75628,
		0.804122,
		0.846758,
		0.883873,
		0.915312,
		0.941085,
		0.961379,
		0.976563,
		0.987186,
		0.993962,
		0.997737,
		0.999431,
		0.999941,
		1,
		0.999992,
		0.999782,
		0.998786,
		0.996168,
		0.991005,
		0.982404,
		0.96958,
		0.951898,
		0.928899,
		0.900305,
		0.866019,
		0.826114,
		0.780826,
		0.730545,
		0.675798,
		0.617247,
		0.555671,
		0.491965,
		0.427129,
		0.362263,
		0.298559,
		0.237298,
		0.179843,
		0.127639,
		0.082205,
		0.045134,
		0.018088,
		0.002795,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.005327,
		0.023419,
		0.052658,
		0.091374,
		0.137962,
		0.190888,
		0.248685,
		0.309961,
		0.373403,
		0.437779,
		0.501941,
		0.564834,
		0.625499,
		0.68308,
		0.736831,
		0.786125,
		0.830462,
		0.869479,
		0.902962,
		0.930852,
		0.953258,
		0.970465,
		0.982931,
		0.991281,
		0.996289,
		0.998826,
		0.99979,
		0.999992,
		1,
		0.999943,
		0.99945,
		0.997809,
		0.99415,
		0.987575,
		0.977255,
		0.962487,
		0.942727,
		0.917603,
		0.88692,
		0.850653,
		0.808939,
		0.762069,
		0.710476,
		0.654724,
		0.595501,
		0.533608,
		0.469955,
		0.405548,
		0.34149,
		0.278972,
		0.219266,
		0.163726,
		0.113781,
		0.070933,
		0.036751,
		0.012873,
		0.000999,
		0.000443,
		0.010606,
		0.032895,
		0.065621,
		0.107161,
		0.155955,
		0.210511,
		0.269407,
		0.331297,
		0.39491,
		0.459058,
		0.522638,
		0.584639,
		0.644146,
		0.70035,
		0.752549,
		0.800163,
		0.842738,
		0.87996,
		0.911662,
		0.937833,
		0.958631,
		0.974386,
		0.985597,
		0.992925,
		0.997165,
		0.999193,
		0.999886,
		0.999998,
		0.999998,
		0.999886,
		0.999193,
		0.997165,
		0.992925,
		0.985597,
		0.974386,
		0.958631,
		0.937833,
		0.911662,
		0.87996,
		0.842738,
		0.800163,
		0.752549,
		0.70035,
		0.644146,
		0.584639,
		0.522638,
		0.459058,
		0.39491,
		0.331297,
		0.269407,
		0.210511,
		0.155955,
		0.107161,
		0.065621,
		0.032895,
		0.010606,
		0.000443,
		0.0027,
		0.017488,
		0.043674,
		0.079615,
		0.12373,
		0.174504,
		0.230487,
		0.290302,
		0.352645,
		0.416289,
		0.48009,
		0.542987,
		0.604014,
		0.6623,
		0.717079,
		0.767695,
		0.813614,
		0.854428,
		0.889869,
		0.919818,
		0.944313,
		0.963556,
		0.977922,
		0.987949,
		0.994331,
		0.997879,
		0.999468,
		0.999945,
		1,
		0.999993,
		0.999797,
		0.998863,
		0.996405,
		0.991548,
		0.983438,
		0.971319,
		0.954571,
		0.932739,
		0.905532,
		0.872831,
		0.834679,
		0.79127,
		0.742944,
		0.690172,
		0.63355,
		0.573791,
		0.511714,
		0.448238,
		0.384378,
		0.321236,
		0.259999,
		0.201932,
		0.148377,
		0.100747,
		0.060526,
		0.029261,
		0.008564,
		0.00011,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000966,
		0.012447,
		0.035564,
		0.0687,
		0.110298,
		0.158861,
		0.212959,
		0.27123,
		0.332378,
		0.395185,
		0.458508,
		0.521286,
		0.582546,
		0.641405,
		0.697079,
		0.748889,
		0.796268,
		0.83877,
		0.876079,
		0.908019,
		0.934562,
		0.955839,
		0.972142,
		0.983926,
		0.991804,
		0.996517,
		0.9989,
		0.999803,
		0.999993,
		1,
		0.999947,
		0.999485,
		0.997946,
		0.994505,
		0.98831,
		0.978564,
		0.964588,
		0.945845,
		0.921961,
		0.892725,
		0.858088,
		0.818151,
		0.773163,
		0.723506,
		0.669685,
		0.612327,
		0.552163,
		0.490028,
		0.42685,
		0.363647,
		0.301522,
		0.241656,
		0.185305,
		0.133799,
		0.088536,
		0.050979,
		0.022654,
		0.005149,
		0.003771,
		0.019678,
		0.046525,
		0.082738,
		0.126801,
		0.17726,
		0.232725,
		0.291874,
		0.353455,
		0.416289,
		0.479275,
		0.541394,
		0.601712,
		0.65939,
		0.713686,
		0.763962,
		0.809696,
		0.850486,
		0.886059,
		0.916286,
		0.941183,
		0.960924,
		0.975845,
		0.98644,
		0.99335,
		0.99734,
		0.999244,
		0.999893,
		0.999999,
		0.999999,
		0.999893,
		0.999244,
		0.99734,
		0.99335,
		0.98644,
		0.975845,
		0.960924,
		0.941183,
		0.916286,
		0.886059,
		0.850486,
		0.809696,
		0.763962,
		0.713686,
		0.65939,
		0.601712,
		0.541394,
		0.479275,
		0.416289,
		0.353455,
		0.291874,
		0.232725,
		0.17726,
		0.126801,
		0.082738,
		0.046525,
		0.019678,
		0.003771,
		0.000107,
		0.008282,
		0.028318,
		0.058623,
		0.097664,
		0.143965,
		0.196114,
		0.252761,
		0.312626,
		0.374495,
		0.43723,
		0.499772,
		0.561141,
		0.620445,
		0.676885,
		0.729761,
		0.778478,
		0.822556,
		0.861636,
		0.895491,
		0.924033,
		0.947325,
		0.965583,
		0.979184,
		0.988657,
		0.994673,
		0.99801,
		0.999501,
		0.999948,
		1,
		0.999993,
		0.99981,
		0.998934,
		0.996625,
		0.99205,
		0.984397,
		0.972936,
		0.957064,
		0.936326,
		0.910426,
		0.879226,
		0.842738,
		0.801123,
		0.754672,
		0.703808,
		0.649069,
		0.591103,
		0.530661,
		0.468589,
		0.405822,
		0.343382,
		0.282369,
		0.223957,
		0.169394,
		0.119997,
		0.077145,
		0.042284,
		0.016918,
		0.00261,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.004979,
		0.021925,
		0.049379,
		0.085828,
		0.12982,
		0.179959,
		0.234912,
		0.293408,
		0.354245,
		0.416289,
		0.478481,
		0.53984,
		0.599465,
		0.656545,
		0.710363,
		0.760297,
		0.805839,
		0.846589,
		0.882275,
		0.912756,
		0.938031,
		0.958246,
		0.973702,
		0.984851,
		0.992288,
		0.996728,
		0.998968,
		0.999816,
		0.999993,
		1,
		0.99995,
		0.999517,
		0.998072,
		0.994834,
		0.988991,
		0.979781,
		0.966545,
		0.948756,
		0.926039,
		0.89817,
		0.865077,
		0.826833,
		0.783645,
		0.73585,
		0.683903,
		0.628371,
		0.569922,
		0.509323,
		0.44743,
		0.385184,
		0.323607,
		0.263794,
		0.206915,
		0.154206,
		0.10697,
		0.066571,
		0.034434,
		0.012042,
		0.000934,
		0.000414,
		0.009923,
		0.030823,
		0.061589,
		0.100747,
		0.146884,
		0.198643,
		0.254728,
		0.313909,
		0.37502,
		0.436966,
		0.498728,
		0.559361,
		0.618005,
		0.673888,
		0.726331,
		0.774754,
		0.818687,
		0.857774,
		0.891784,
		0.920616,
		0.944313,
		0.963062,
		0.977203,
		0.987222,
		0.993744,
		0.997501,
		0.999291,
		0.9999,
		0.999999,
		0.999999,
		0.9999,
		0.999291,
		0.997501,
		0.993744,
		0.987222,
		0.977203,
		0.963062,
		0.944313,
		0.920616,
		0.891784,
		0.857774,
		0.818687,
		0.774754,
		0.726331,
		0.673888,
		0.618005,
		0.559361,
		0.498728,
		0.436966,
		0.37502,
		0.313909,
		0.254728,
		0.198643,
		0.146884,
		0.100747,
		0.061589,
		0.030823,
		0.009923,
		0.000414,
		0.002525,
		0.016376,
		0.040959,
		0.074788,
		0.116428,
		0.164503,
		0.217694,
		0.274744,
		0.33446,
		0.395714,
		0.45745,
		0.518684,
		0.578511,
		0.636109,
		0.690745,
		0.741777,
		0.788668,
		0.830985,
		0.868415,
		0.900765,
		0.927979,
		0.950139,
		0.967473,
		0.980357,
		0.989313,
		0.994989,
		0.998131,
		0.999532,
		0.999952,
		1,
		0.999993,
		0.999822,
		0.999,
		0.996827,
		0.992516,
		0.985289,
		0.974442,
		0.959389,
		0.93968,
		0.915012,
		0.885232,
		0.850326,
		0.810421,
		0.76577,
		0.716747,
		0.66384,
		0.607637,
		0.548826,
		0.488188,
		0.426585,
		0.364963,
		0.304341,
		0.245813,
		0.190539,
		0.139745,
		0.094719,
		0.056809,
		0.02742,
		0.008014,
		0.000103,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000903,
		0.011657,
		0.033357,
		0.064539,
		0.103789,
		0.14975,
		0.201118,
		0.25665,
		0.315161,
		0.375532,
		0.436709,
		0.497709,
		0.557624,
		0.615622,
		0.670956,
		0.722968,
		0.771094,
		0.814873,
		0.853952,
		0.888097,
		0.917197,
		0.941275,
		0.960493,
		0.975156,
		0.985711,
		0.992736,
		0.996923,
		0.999031,
		0.999827,
		0.999994,
		1,
		0.999953,
		0.999547,
		0.998188,
		0.995138,
		0.989624,
		0.980913,
		0.968369,
		0.951476,
		0.929857,
		0.90328,
		0.871652,
		0.835018,
		0.793551,
		0.747548,
		0.697414,
		0.643665,
		0.586911,
		0.527856,
		0.46729,
		0.406083,
		0.345181,
		0.285604,
		0.228439,
		0.174837,
		0.126014,
		0.083242,
		0.047852,
		0.021232,
		0.004818,
		0.003529,
		0.018444,
		0.043674,
		0.077793,
		0.119423,
		0.167242,
		0.219983,
		0.27644,
		0.335463,
		0.395969,
		0.45694,
		0.51743,
		0.576564,
		0.63355,
		0.687676,
		0.738322,
		0.784961,
		0.827169,
		0.864634,
		0.897159,
		0.924674,
		0.947239,
		0.965057,
		0.978467,
		0.987949,
		0.994109,
		0.997651,
		0.999334,
		0.999906,
		0.999999,
		0.999999,
		0.999906,
		0.999334,
		0.997651,
		0.994109,
		0.987949,
		0.978467,
		0.965057,
		0.947239,
		0.924674,
		0.897159,
		0.864634,
		0.827169,
		0.784961,
		0.738322,
		0.687676,
		0.63355,
		0.576564,
		0.51743,
		0.45694,
		0.395969,
		0.335463,
		0.27644,
		0.219983,
		0.167242,
		0.119423,
		0.077793,
		0.043674,
		0.018444,
		0.003529,
		0.0001,
		0.007758,
		0.026564,
		0.055077,
		0.091904,
		0.135705,
		0.185194,
		0.239139,
		0.296367,
		0.355767,
		0.416289,
		0.476952,
		0.536843,
		0.595126,
		0.651042,
		0.703917,
		0.753166,
		0.7983,
		0.838934,
		0.874792,
		0.905717,
		0.931675,
		0.952769,
		0.969235,
		0.98145,
		0.989923,
		0.995282,
		0.998243,
		0.999561,
		0.999955,
		1,
		0.999994,
		0.999833,
		0.999061,
		0.997015,
		0.992949,
		0.986118,
		0.975845,
		0.96156,
		0.942819,
		0.919314,
		0.890876,
		0.857473,
		0.8192,
		0.776274,
		0.729027,
		0.677898,
		0.623424,
		0.566235,
		0.507048,
		0.446662,
		0.385951,
		0.325863,
		0.267414,
		0.211684,
		0.159817,
		0.113014,
		0.072536,
		0.039695,
		0.015859,
		0.002443,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.004665,
		0.02057,
		0.046396,
		0.08077,
		0.122371,
		0.169927,
		0.222223,
		0.278096,
		0.336442,
		0.396218,
		0.456443,
		0.516206,
		0.574663,
		0.631048,
		0.684672,
		0.734932,
		0.781314,
		0.823405,
		0.86089,
		0.893571,
		0.921364,
		0.944313,
		0.962592,
		0.976511,
		0.98651,
		0.993153,
		0.997104,
		0.999089,
		0.999838,
		0.999994,
		1,
		0.999956,
		0.999575,
		0.998296,
		0.995421,
		0.990212,
		0.981968,
		0.970072,
		0.95402,
		0.933436,
		0.908079,
		0.877839,
		0.842738,
		0.802918,
		0.758636,
		0.710256,
		0.658245,
		0.60316,
		0.545648,
		0.486437,
		0.426333,
		0.366214,
		0.307026,
		0.249783,
		0.195557,
		0.145482,
		0.100747,
		0.062598,
		0.03233,
		0.01129,
		0.000874,
		0.000387,
		0.009304,
		0.028941,
		0.057915,
		0.094888,
		0.138573,
		0.187731,
		0.241182,
		0.297795,
		0.3565,
		0.416289,
		0.476215,
		0.535398,
		0.593031,
		0.64838,
		0.700791,
		0.749696,
		0.794618,
		0.835177,
		0.871096,
		0.902212,
		0.928479,
		0.949978,
		0.96692,
		0.979645,
		0.988626,
		0.994448,
		0.997789,
		0.999375,
		0.999912,
		0.999999,
		0.999999,
		0.999912,
		0.999375,
		0.997789,
		0.994448,
		0.988626,
		0.979645,
		0.96692,
		0.949978,
		0.928479,
		0.902212,
		0.871096,
		0.835177,
		0.794618,
		0.749696,
		0.700791,
		0.64838,
		0.593031,
		0.535398,
		0.476215,
		0.416289,
		0.3565,
		0.297795,
		0.241182,
		0.187731,
		0.138573,
		0.094888,
		0.057915,
		0.028941,
		0.009304,
		0.000387,
		0.002366,
		0.015366,
		0.038489,
		0.070384,
		0.109748,
		0.155325,
		0.205913,
		0.260362,
		0.317576,
		0.376519,
		0.436213,
		0.495746,
		0.554272,
		0.611015,
		0.665277,
		0.716437,
		0.763962,
		0.80741,
		0.846434,
		0.880796,
		0.910368,
		0.935141,
		0.95523,
		0.970881,
		0.982468,
		0.99049,
		0.995554,
		0.998347,
		0.999588,
		0.999957,
		1,
		0.999994,
		0.999843,
		0.999117,
		0.997189,
		0.99335,
		0.986889,
		0.977154,
		0.96359,
		0.945759,
		0.92335,
		0.896184,
		0.864209,
		0.827492,
		0.78622,
		0.740684,
		0.69128,
		0.638497,
		0.582912,
		0.525185,
		0.466054,
		0.406331,
		0.346894,
		0.28869,
		0.232725,
		0.180066,
		0.131836,
		0.089212,
		0.053423,
		0.025748,
		0.007514,
		9.6e-005,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.000846,
		0.01094,
		0.031349,
		0.060743,
		0.097836,
		0.141391,
		0.190218,
		0.243179,
		0.299189,
		0.357216,
		0.416289,
		0.475496,
		0.533987,
		0.590984,
		0.645775,
		0.697728,
		0.746289,
		0.790993,
		0.831466,
		0.867431,
		0.89872,
		0.925275,
		0.947158,
		0.964555,
		0.977776,
		0.987255,
		0.993541,
		0.997272,
		0.999143,
		0.999848,
		0.999994,
		1,
		0.999959,
		0.9996,
		0.998396,
		0.995683,
		0.990759,
		0.982951,
		0.971663,
		0.956401,
		0.936792,
		0.912588,
		0.883666,
		0.850025,
		0.811778,
		0.769149,
		0.722464,
		0.672143,
		0.618696,
		0.562719,
		0.504881,
		0.44593,
		0.386681,
		0.328012,
		0.270869,
		0.216251,
		0.165218,
		0.118882,
		0.078406,
		0.045004,
		0.019939,
		0.004519,
		0.00331,
		0.017323,
		0.041077,
		0.073275,
		0.112664,
		0.158037,
		0.208236,
		0.262156,
		0.318742,
		0.376995,
		0.435974,
		0.494799,
		0.552654,
		0.608789,
		0.662527,
		0.713267,
		0.760489,
		0.803759,
		0.842738,
		0.877185,
		0.906963,
		0.932051,
		0.952544,
		0.968661,
		0.980745,
		0.989256,
		0.994763,
		0.997917,
		0.999412,
		0.999917,
		0.999999,
		0.999999,
		0.999917,
		0.999412,
		0.997917,
		0.994763,
		0.989256,
		0.980745,
		0.968661,
		0.952544,
		0.932051,
		0.906963,
		0.877185,
		0.842738,
		0.803759,
		0.760489,
		0.713267,
		0.662527,
		0.608789,
		0.552654,
		0.494799,
		0.435974,
		0.376995,
		0.318742,
		0.262156,
		0.208236,
		0.158037,
		0.112664,
		0.073275,
		0.041077,
		0.017323,
		0.00331,
		9.3e-005,
		0.007282,
		0.024968,
		0.051842,
		0.086636,
		0.128128,
		0.175145,
		0.226559,
		0.281295,
		0.33833,
		0.396697,
		0.455485,
		0.513845,
		0.570991,
		0.626207,
		0.678846,
		0.72834,
		0.774201,
		0.816028,
		0.853514,
		0.886452,
		0.914741,
		0.938392,
		0.957534,
		0.972419,
		0.983418,
		0.991018,
		0.995807,
		0.998443,
		0.999612,
		0.99996,
		1,
		0.999995,
		0.999852,
		0.999169,
		0.997351,
		0.993724,
		0.987608,
		0.978376,
		0.965488,
		0.948514,
		0.927141,
		0.90118,
		0.870561,
		0.835329,
		0.79564,
		0.751752,
		0.704019,
		0.652886,
		0.598882,
		0.542616,
		0.484769,
		0.426093,
		0.367405,
		0.309587,
		0.253577,
		0.200371,
		0.151018,
		0.106619,
		0.068326,
		0.037337,
		0.014896,
		0.002292,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.004379,
		0.019337,
		0.043674,
		0.076143,
		0.115538,
		0.160699,
		0.210511,
		0.26391,
		0.31988,
		0.377459,
		0.435741,
		0.493875,
		0.551073,
		0.606611,
		0.659834,
		0.710157,
		0.757075,
		0.800163,
		0.839085,
		0.873601,
		0.903567,
		0.92895,
		0.949827,
		0.966392,
		0.978957,
		0.987949,
		0.993901,
		0.997428,
		0.999193,
		0.999857,
		0.999995,
		1,
		0.999961,
		0.999624,
		0.998488,
		0.995927,
		0.991269,
		0.983869,
		0.97315,
		0.958631,
		0.939942,
		0.916829,
		0.889157,
		0.856905,
		0.820162,
		0.779121,
		0.73407,
		0.685391,
		0.63355,
		0.579092,
		0.522638,
		0.464877,
		0.406567,
		0.348527,
		0.291635,
		0.236827,
		0.185091,
		0.137468,
		0.095049,
		0.058969,
		0.030412,
		0.010606,
		0.00082,
		0.000363,
		0.008741,
		0.027226,
		0.05456,
		0.089522,
		0.130939,
		0.177679,
		0.228658,
		0.282841,
		0.339242,
		0.396928,
		0.455022,
		0.512705,
		0.569217,
		0.623865,
		0.676022,
		0.725136,
		0.770731,
		0.812415,
		0.849883,
		0.882926,
		0.911434,
		0.935405,
		0.954949,
		0.970291,
		0.981771,
		0.989843,
		0.995056,
		0.998037,
		0.999446,
		0.999922,
		0.999999,
		0.999999,
		0.999922,
		0.999446,
		0.998037,
		0.995056,
		0.989843,
		0.981771,
		0.970291,
		0.954949,
		0.935405,
		0.911434,
		0.882926,
		0.849883,
		0.812415,
		0.770731,
		0.725136,
		0.676022,
		0.623865,
		0.569217,
		0.512705,
		0.455022,
		0.396928,
		0.339242,
		0.282841,
		0.228658,
		0.177679,
		0.130939,
		0.089522,
		0.05456,
		0.027226,
		0.008741,
		0.000363,
		0.002221,
		0.014447,
		0.036235,
		0.066356,
		0.103621,
		0.146884,
		0.195044,
		0.247045,
		0.301882,
		0.358598,
		0.416289,
		0.474108,
		0.531263,
		0.587025,
		0.64073,
		0.691782,
		0.739658,
		0.783913,
		0.824185,
		0.860201,
		0.891784,
		0.918855,
		0.941445,
		0.959694,
		0.973858,
		0.984305,
		0.99151,
		0.996042,
		0.998532,
		0.999635,
		0.999962,
		1,
		0.999995,
		0.999861,
		0.999217,
		0.997501,
		0.994072,
		0.988279,
		0.979518,
		0.967266,
		0.951098,
		0.930704,
		0.905884,
		0.876554,
		0.842738,
		0.804566,
		0.762263,
		0.716147,
		0.666622,
		0.614173,
		0.559361,
		0.502815,
		0.445233,
		0.387376,
		0.330063,
		0.27417,
		0.220628,
		0.170419,
		0.124573,
		0.084169,
		0.05033,
		0.024223,
		0.00706,
		9.1e-005,
		0.005562,
		0.021402,
		0.046277,
		0.078987,
		0.118371,
		0.163312,
		0.212739,
		0.265625,
		0.320992,
		0.377913,
		0.435513,
		0.492972,
		0.549528,
		0.604481,
		0.657196,
		0.707107,
		0.753719,
		0.796619,
		0.835475,
		0.870045,
		0.900183,
		0.925841,
		0.947082,
		0.964076,
		0.977109,
		0.986574,
		0.992969,
		0.996876,
		0.998929,
		0.999771,
		0.999985,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.005605,
		0.021564,
		0.046619,
		0.079556,
		0.119202,
		0.164428,
		0.214149,
		0.267329,
		0.32298,
		0.380165,
		0.438,
		0.495659,
		0.552374,
		0.607439,
		0.660215,
		0.710133,
		0.756699,
		0.799498,
		0.838201,
		0.872571,
		0.902465,
		0.927846,
		0.948785,
		0.965467,
		0.978187,
		0.987358,
		0.99349,
		0.99718,
		0.999074,
		0.999819,
		0.999991,
		1,
		0.999975,
		0.999707,
		0.998733,
		0.996452,
		0.99221,
		0.985372,
		0.975358,
		0.961681,
		0.943959,
		0.921922,
		0.895415,
		0.864394,
		0.828924,
		0.789171,
		0.745396,
		0.69795,
		0.647268,
		0.593863,
		0.538323,
		0.481304,
		0.423526,
		0.365774,
		0.308888,
		0.253765,
		0.201355,
		0.152659,
		0.108724,
		0.070648,
		0.039569,
		0.016673,
		0.003183,
		0.005649,
		0.021727,
		0.046964,
		0.080132,
		0.120042,
		0.165555,
		0.215573,
		0.269049,
		0.324985,
		0.382435,
		0.440506,
		0.498366,
		0.555238,
		0.610413,
		0.663248,
		0.71317,
		0.759685,
		0.802379,
		0.840925,
		0.875088,
		0.904734,
		0.929834,
		0.950468,
		0.966832,
		0.97924,
		0.988116,
		0.993987,
		0.997465,
		0.999205,
		0.999859,
		0.999995,
		1,
		0.999962,
		0.999629,
		0.998511,
		0.995985,
		0.99139,
		0.984089,
		0.973507,
		0.959167,
		0.940699,
		0.91785,
		0.89048,
		0.858565,
		0.822187,
		0.781532,
		0.736881,
		0.688606,
		0.63716,
		0.583079,
		0.526971,
		0.469512,
		0.411446,
		0.353578,
		0.29677,
		0.241943,
		0.190067,
		0.142168,
		0.099316,
		0.062632,
		0.033279,
		0.012464,
		0.00144,
		0.005692,
		0.021892,
		0.047314,
		0.080713,
		0.120891,
		0.166693,
		0.217011,
		0.270786,
		0.327009,
		0.384724,
		0.443032,
		0.501091,
		0.55812,
		0.613403,
		0.666294,
		0.716218,
		0.762678,
		0.805262,
		0.843645,
		0.877597,
		0.90699,
		0.931803,
		0.952128,
		0.968173,
		0.980266,
		0.988848,
		0.994461,
		0.997731,
		0.999323,
		0.999892,
		0.999997,
		1,
		0.999944,
		0.999537,
		0.99826,
		0.995473,
		0.990506,
		0.982721,
		0.971552,
		0.95653,
		0.937299,
		0.913621,
		0.885375,
		0.852555,
		0.815262,
		0.773702,
		0.728175,
		0.679074,
		0.626874,
		0.572131,
		0.515474,
		0.457602,
		0.39928,
		0.341334,
		0.28465,
		0.230171,
		0.178891,
		0.131859,
		0.090168,
		0.054963,
		0.027432,
		0.008808,
		0.000366,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.005782,
		0.022228,
		0.048024,
		0.081895,
		0.122616,
		0.169005,
		0.219929,
		0.274307,
		0.331109,
		0.38936,
		0.448143,
		0.5066,
		0.563939,
		0.619433,
		0.672427,
		0.722343,
		0.768681,
		0.811031,
		0.849074,
		0.882588,
		0.91146,
		0.935686,
		0.955381,
		0.97078,
		0.98224,
		0.990235,
		0.99534,
		0.998208,
		0.999523,
		0.999942,
		1,
		0.999997,
		0.999889,
		0.999303,
		0.997663,
		0.9943,
		0.988531,
		0.97972,
		0.967315,
		0.950871,
		0.930061,
		0.90468,
		0.874642,
		0.839981,
		0.800837,
		0.757456,
		0.710181,
		0.659446,
		0.60577,
		0.54975,
		0.492061,
		0.433445,
		0.374712,
		0.316735,
		0.260446,
		0.206833,
		0.156939,
		0.111859,
		0.072738,
		0.040769,
		0.01719,
		0.003284,
		0.005827,
		0.022399,
		0.048385,
		0.082496,
		0.123492,
		0.170178,
		0.221409,
		0.276092,
		0.333186,
		0.391707,
		0.450728,
		0.509384,
		0.566876,
		0.622473,
		0.675515,
		0.72542,
		0.771692,
		0.813918,
		0.851782,
		0.88507,
		0.913673,
		0.937598,
		0.956972,
		0.972044,
		0.983187,
		0.99089,
		0.995745,
		0.99842,
		0.999606,
		0.999959,
		1,
		0.999995,
		0.99985,
		0.999156,
		0.997312,
		0.993633,
		0.987433,
		0.978079,
		0.965026,
		0.947842,
		0.926215,
		0.899959,
		0.869007,
		0.833411,
		0.793332,
		0.749038,
		0.700893,
		0.649351,
		0.594955,
		0.538323,
		0.480153,
		0.42121,
		0.362328,
		0.304401,
		0.248386,
		0.195295,
		0.146196,
		0.102209,
		0.064503,
		0.034297,
		0.012854,
		0.001486,
		0.005873,
		0.022572,
		0.04875,
		0.083104,
		0.124377,
		0.171364,
		0.222905,
		0.277895,
		0.335282,
		0.394074,
		0.453333,
		0.512188,
		0.569832,
		0.625529,
		0.678615,
		0.728508,
		0.774707,
		0.816804,
		0.854485,
		0.887541,
		0.91587,
		0.939491,
		0.95854,
		0.973282,
		0.984106,
		0.991519,
		0.996128,
		0.998614,
		0.999679,
		0.999973,
		1,
		0.99999,
		0.999801,
		0.998987,
		0.996922,
		0.992908,
		0.986257,
		0.976338,
		0.962616,
		0.944673,
		0.922214,
		0.895068,
		0.863189,
		0.826649,
		0.78563,
		0.740422,
		0.691409,
		0.639069,
		0.583966,
		0.526742,
		0.468117,
		0.40888,
		0.349886,
		0.292057,
		0.236371,
		0.183866,
		0.135635,
		0.092821,
		0.056622,
		0.02828,
		0.009086,
		0.000378,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.005967,
		0.022924,
		0.049493,
		0.084339,
		0.126177,
		0.173771,
		0.225939,
		0.28155,
		0.339531,
		0.398867,
		0.458604,
		0.517854,
		0.575799,
		0.631689,
		0.684856,
		0.734711,
		0.780754,
		0.822577,
		0.859876,
		0.89245,
		0.920217,
		0.943213,
		0.961602,
		0.975678,
		0.985864,
		0.9927,
		0.996829,
		0.998956,
		0.999795,
		0.999989,
		1,
		0.999972,
		0.999668,
		0.998572,
		0.996013,
		0.991272,
		0.983655,
		0.972543,
		0.957423,
		0.937904,
		0.913726,
		0.884756,
		0.850988,
		0.812535,
		0.769624,
		0.722587,
		0.671853,
		0.617947,
		0.561477,
		0.503133,
		0.443682,
		0.383959,
		0.32487,
		0.267385,
		0.212531,
		0.161399,
		0.11513,
		0.074922,
		0.042023,
		0.017731,
		0.00339,
		0.006014,
		0.023103,
		0.049871,
		0.084966,
		0.127091,
		0.174994,
		0.227479,
		0.283404,
		0.341684,
		0.401293,
		0.46127,
		0.520718,
		0.57881,
		0.634794,
		0.687996,
		0.737827,
		0.783784,
		0.825463,
		0.862562,
		0.894888,
		0.922365,
		0.945041,
		0.963095,
		0.976835,
		0.986701,
		0.993253,
		0.997147,
		0.999103,
		0.99984,
		0.999994,
		1,
		0.999957,
		0.999581,
		0.998321,
		0.995488,
		0.990352,
		0.98222,
		0.97048,
		0.95463,
		0.934295,
		0.909232,
		0.879329,
		0.8446,
		0.805179,
		0.761317,
		0.713366,
		0.661782,
		0.607109,
		0.549982,
		0.491113,
		0.431293,
		0.371384,
		0.312315,
		0.255081,
		0.200737,
		0.150396,
		0.105229,
		0.06646,
		0.035363,
		0.013263,
		0.001534,
		0.006063,
		0.023285,
		0.050253,
		0.085601,
		0.128015,
		0.17623,
		0.229035,
		0.285276,
		0.343856,
		0.403741,
		0.463957,
		0.523601,
		0.58184,
		0.637915,
		0.69115,
		0.740951,
		0.786818,
		0.828347,
		0.865241,
		0.897313,
		0.924495,
		0.946848,
		0.964563,
		0.977965,
		0.987511,
		0.99378,
		0.997444,
		0.999236,
		0.999878,
		0.999997,
		1,
		0.999936,
		0.999477,
		0.998039,
		0.994912,
		0.989361,
		0.980692,
		0.968302,
		0.951701,
		0.930531,
		0.904566,
		0.873716,
		0.838015,
		0.79762,
		0.752804,
		0.703943,
		0.651515,
		0.596089,
		0.538323,
		0.478955,
		0.4188,
		0.358744,
		0.299743,
		0.242817,
		0.189047,
		0.139573,
		0.095593,
		0.058356,
		0.029167,
		0.009378,
		0.00039,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.006161,
		0.023653,
		0.05103,
		0.086892,
		0.129893,
		0.178739,
		0.232193,
		0.289073,
		0.34826,
		0.408697,
		0.469393,
		0.529428,
		0.587956,
		0.644206,
		0.697495,
		0.747226,
		0.792898,
		0.834112,
		0.870578,
		0.902123,
		0.9287,
		0.950392,
		0.967419,
		0.98014,
		0.989048,
		0.994759,
		0.997978,
		0.999461,
		0.999934,
		1,
		0.999997,
		0.999874,
		0.999212,
		0.997366,
		0.993593,
		0.987147,
		0.977338,
		0.963582,
		0.945417,
		0.922519,
		0.894703,
		0.861918,
		0.824244,
		0.781884,
		0.735156,
		0.684484,
		0.630396,
		0.573509,
		0.514529,
		0.454246,
		0.393526,
		0.333306,
		0.274594,
		0.218463,
		0.166048,
		0.118545,
		0.077206,
		0.043337,
		0.018298,
		0.003501,
		0.006211,
		0.023841,
		0.051425,
		0.087549,
		0.130847,
		0.180014,
		0.233795,
		0.290998,
		0.350491,
		0.411206,
		0.472143,
		0.532373,
		0.591042,
		0.647376,
		0.700686,
		0.750376,
		0.795942,
		0.83699,
		0.873234,
		0.904508,
		0.930774,
		0.952128,
		0.968806,
		0.981184,
		0.989775,
		0.995211,
		0.998216,
		0.999554,
		0.999954,
		1,
		0.999994,
		0.99983,
		0.999046,
		0.99697,
		0.992844,
		0.985916,
		0.975504,
		0.961031,
		0.942053,
		0.918264,
		0.889497,
		0.855726,
		0.817052,
		0.773702,
		0.726017,
		0.674449,
		0.619546,
		0.561953,
		0.502402,
		0.441707,
		0.38076,
		0.320527,
		0.262041,
		0.206404,
		0.154777,
		0.108384,
		0.068506,
		0.036479,
		0.013691,
		0.001584,
		0.006261,
		0.024031,
		0.051826,
		0.088212,
		0.131812,
		0.181302,
		0.235414,
		0.292943,
		0.352743,
		0.413737,
		0.474914,
		0.535338,
		0.594146,
		0.650561,
		0.70389,
		0.753533,
		0.79899,
		0.839866,
		0.875881,
		0.906878,
		0.932827,
		0.95384,
		0.970166,
		0.9822,
		0.990474,
		0.995638,
		0.998435,
		0.999636,
		0.999969,
		1,
		0.999988,
		0.999775,
		0.998854,
		0.996529,
		0.992029,
		0.984597,
		0.973558,
		0.958348,
		0.938536,
		0.913836,
		0.884104,
		0.849334,
		0.809651,
		0.765307,
		0.716668,
		0.664208,
		0.608503,
		0.550223,
		0.490126,
		0.429052,
		0.36792,
		0.307723,
		0.249522,
		0.194445,
		0.143683,
		0.098489,
		0.060171,
		0.030097,
		0.009684,
		0.000403,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.006364,
		0.024418,
		0.05264,
		0.089563,
		0.133774,
		0.183919,
		0.238701,
		0.296887,
		0.357308,
		0.418863,
		0.480522,
		0.54133,
		0.600413,
		0.656981,
		0.710335,
		0.759872,
		0.805092,
		0.845607,
		0.881147,
		0.911571,
		0.936872,
		0.957188,
		0.972801,
		0.984144,
		0.991788,
		0.996422,
		0.998818,
		0.999767,
		0.999988,
		1,
		0.999968,
		0.999624,
		0.998385,
		0.995504,
		0.990188,
		0.98168,
		0.969317,
		0.952563,
		0.931022,
		0.904447,
		0.872739,
		0.835936,
		0.794215,
		0.747875,
		0.697333,
		0.643115,
		0.585848,
		0.526257,
		0.465151,
		0.403426,
		0.342055,
		0.282087,
		0.224639,
		0.170898,
		0.122113,
		0.079594,
		0.044712,
		0.018893,
		0.003617,
		0.006417,
		0.024615,
		0.053054,
		0.090249,
		0.134771,
		0.185248,
		0.24037,
		0.298888,
		0.359622,
		0.421459,
		0.483358,
		0.544357,
		0.603574,
		0.660215,
		0.713576,
		0.763052,
		0.808146,
		0.848471,
		0.883765,
		0.913893,
		0.938862,
		0.958822,
		0.974075,
		0.985072,
		0.992403,
		0.996778,
		0.998984,
		0.999819,
		0.999993,
		1,
		0.999951,
		0.999525,
		0.998102,
		0.994912,
		0.989154,
		0.980072,
		0.967013,
		0.949453,
		0.927017,
		0.899478,
		0.866759,
		0.828924,
		0.786174,
		0.738833,
		0.687346,
		0.632263,
		0.57424,
		0.514027,
		0.452462,
		0.390469,
		0.32905,
		0.26928,
		0.212308,
		0.159349,
		0.111681,
		0.070648,
		0.037648,
		0.014141,
		0.001638,
		0.00647,
		0.024814,
		0.053473,
		0.090944,
		0.135779,
		0.186592,
		0.242056,
		0.300909,
		0.361957,
		0.424076,
		0.486216,
		0.547405,
		0.606755,
		0.663465,
		0.716828,
		0.76624,
		0.811201,
		0.851331,
		0.886372,
		0.916199,
		0.94083,
		0.960431,
		0.975321,
		0.98597,
		0.992991,
		0.997111,
		0.999134,
		0.999861,
		0.999997,
		1,
		0.999927,
		0.999407,
		0.997782,
		0.994262,
		0.988039,
		0.978359,
		0.964579,
		0.946192,
		0.92284,
		0.894319,
		0.860575,
		0.821699,
		0.777914,
		0.729573,
		0.677145,
		0.621209,
		0.562448,
		0.501639,
		0.439648,
		0.377429,
		0.31601,
		0.256499,
		0.200072,
		0.147975,
		0.101517,
		0.062072,
		0.031071,
		0.010005,
		0.000417,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.006578,
		0.02522,
		0.054327,
		0.092357,
		0.137829,
		0.189322,
		0.245479,
		0.305009,
		0.366691,
		0.429378,
		0.491999,
		0.553565,
		0.613173,
		0.670011,
		0.723368,
		0.772634,
		0.817314,
		0.857034,
		0.891549,
		0.920755,
		0.944695,
		0.963565,
		0.977722,
		0.987677,
		0.994084,
		0.997711,
		0.999388,
		0.999925,
		1,
		0.999997,
		0.999857,
		0.999106,
		0.99702,
		0.992774,
		0.985548,
		0.974599,
		0.959306,
		0.939196,
		0.913952,
		0.883417,
		0.847584,
		0.806596,
		0.76073,
		0.71039,
		0.656102,
		0.5985,
		0.538323,
		0.476406,
		0.413672,
		0.351133,
		0.289878,
		0.231074,
		0.17596,
		0.125842,
		0.082095,
		0.046154,
		0.019517,
		0.003739,
		0.006633,
		0.025427,
		0.054761,
		0.093076,
		0.138872,
		0.19071,
		0.247217,
		0.307089,
		0.369091,
		0.432063,
		0.494924,
		0.556677,
		0.61641,
		0.673308,
		0.726654,
		0.77584,
		0.82037,
		0.859876,
		0.894119,
		0.923006,
		0.946591,
		0.96509,
		0.978877,
		0.988485,
		0.99459,
		0.997978,
		0.999493,
		0.999947,
		1,
		0.999993,
		0.999806,
		0.998917,
		0.996571,
		0.991928,
		0.984164,
		0.972543,
		0.956457,
		0.935451,
		0.909232,
		0.877665,
		0.840769,
		0.798713,
		0.751801,
		0.700466,
		0.645261,
		0.586849,
		0.525998,
		0.463572,
		0.400525,
		0.337898,
		0.276811,
		0.218463,
		0.164123,
		0.11513,
		0.072891,
		0.038875,
		0.014613,
		0.001693,
		0.006689,
		0.025636,
		0.0552,
		0.093803,
		0.139926,
		0.192112,
		0.248973,
		0.309189,
		0.371512,
		0.43477,
		0.497872,
		0.55981,
		0.619667,
		0.67662,
		0.729952,
		0.779051,
		0.823426,
		0.862711,
		0.896676,
		0.925236,
		0.948463,
		0.966586,
		0.980001,
		0.989263,
		0.995069,
		0.998225,
		0.999586,
		0.999965,
		1,
		0.999987,
		0.999743,
		0.9987,
		0.996073,
		0.991008,
		0.982681,
		0.970362,
		0.953459,
		0.931537,
		0.904322,
		0.871707,
		0.833736,
		0.790604,
		0.742645,
		0.690319,
		0.634207,
		0.575003,
		0.513503,
		0.450601,
		0.387283,
		0.324619,
		0.263763,
		0.205942,
		0.152459,
		0.104687,
		0.064064,
		0.032094,
		0.010342,
		0.000431,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.006803,
		0.026062,
		0.056095,
		0.095283,
		0.142069,
		0.194962,
		0.25254,
		0.313452,
		0.376422,
		0.440255,
		0.503836,
		0.56614,
		0.626236,
		0.683291,
		0.736579,
		0.78549,
		0.829535,
		0.868358,
		0.901745,
		0.929636,
		0.952128,
		0.969489,
		0.982156,
		0.990728,
		0.995947,
		0.998657,
		0.999735,
		0.999986,
		1,
		0.999963,
		0.999572,
		0.998167,
		0.994912,
		0.988931,
		0.9794,
		0.965609,
		0.946999,
		0.923176,
		0.893914,
		0.859156,
		0.819002,
		0.773702,
		0.723646,
		0.669354,
		0.611467,
		0.550737,
		0.488022,
		0.424278,
		0.360554,
		0.297982,
		0.237781,
		0.181245,
		0.129743,
		0.084714,
		0.047667,
		0.020173,
		0.003868,
		0.006861,
		0.026279,
		0.056551,
		0.096037,
		0.14316,
		0.196411,
		0.254351,
		0.315615,
		0.378911,
		0.443032,
		0.506853,
		0.569338,
		0.629549,
		0.686649,
		0.739908,
		0.788716,
		0.832587,
		0.871169,
		0.904257,
		0.931803,
		0.953921,
		0.970896,
		0.983187,
		0.991416,
		0.996348,
		0.998845,
		0.999793,
		0.999992,
		1,
		0.999944,
		0.999459,
		0.997845,
		0.994242,
		0.987764,
		0.977591,
		0.963026,
		0.943527,
		0.91872,
		0.888407,
		0.852555,
		0.811294,
		0.764902,
		0.7138,
		0.658536,
		0.599783,
		0.538323,
		0.475047,
		0.410941,
		0.347086,
		0.28465,
		0.224882,
		0.169111,
		0.118739,
		0.075242,
		0.040162,
		0.015108,
		0.001752,
		0.00692,
		0.026499,
		0.057012,
		0.096799,
		0.144262,
		0.197875,
		0.256181,
		0.317799,
		0.381424,
		0.445833,
		0.509893,
		0.572558,
		0.632882,
		0.690022,
		0.743248,
		0.791947,
		0.835636,
		0.87397,
		0.906753,
		0.933948,
		0.955686,
		0.972271,
		0.984185,
		0.992073,
		0.996723,
		0.999014,
		0.999842,
		0.999996,
		1,
		0.999917,
		0.999324,
		0.997481,
		0.993506,
		0.986506,
		0.975665,
		0.960299,
		0.939885,
		0.914074,
		0.88269,
		0.845732,
		0.803355,
		0.755868,
		0.703721,
		0.647496,
		0.587893,
		0.525728,
		0.461923,
		0.397498,
		0.333566,
		0.271328,
		0.212067,
		0.157147,
		0.108006,
		0.066153,
		0.033168,
		0.010696,
		0.000447,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.00704,
		0.026947,
		0.057952,
		0.09835,
		0.146505,
		0.200852,
		0.2599,
		0.322233,
		0.386518,
		0.451507,
		0.516043,
		0.579062,
		0.639602,
		0.696813,
		0.749955,
		0.798418,
		0.841725,
		0.879542,
		0.911694,
		0.938168,
		0.959131,
		0.974928,
		0.986085,
		0.993298,
		0.997398,
		0.999302,
		0.999914,
		0.999999,
		0.999996,
		0.999836,
		0.998981,
		0.996615,
		0.991821,
		0.983695,
		0.971437,
		0.954391,
		0.932076,
		0.904191,
		0.870615,
		0.831403,
		0.78677,
		0.737086,
		0.682865,
		0.624748,
		0.563503,
		0.500011,
		0.435259,
		0.370333,
		0.306416,
		0.244776,
		0.186768,
		0.133826,
		0.087461,
		0.049255,
		0.020862,
		0.004003,
		0.007101,
		0.027175,
		0.05843,
		0.09914,
		0.147646,
		0.202366,
		0.261788,
		0.324482,
		0.3891,
		0.454381,
		0.519153,
		0.582346,
		0.642991,
		0.70023,
		0.753323,
		0.801659,
		0.844764,
		0.882312,
		0.914137,
		0.940242,
		0.96081,
		0.976208,
		0.986986,
		0.993865,
		0.9977,
		0.999422,
		0.99994,
		1,
		0.999992,
		0.999779,
		0.998766,
		0.996106,
		0.990862,
		0.982133,
		0.969125,
		0.951199,
		0.927897,
		0.898944,
		0.864247,
		0.823889,
		0.778118,
		0.727336,
		0.672085,
		0.613044,
		0.55101,
		0.4869,
		0.421733,
		0.356632,
		0.292812,
		0.23158,
		0.174326,
		0.122519,
		0.077708,
		0.041515,
		0.01563,
		0.001814,
		0.007163,
		0.027406,
		0.058915,
		0.099939,
		0.148801,
		0.203896,
		0.263696,
		0.326755,
		0.391707,
		0.457278,
		0.522288,
		0.585653,
		0.646399,
		0.703661,
		0.756699,
		0.804901,
		0.847798,
		0.88507,
		0.916561,
		0.94229,
		0.962458,
		0.977455,
		0.987855,
		0.994403,
		0.997978,
		0.999527,
		0.999959,
		1,
		0.999985,
		0.999707,
		0.998518,
		0.995539,
		0.989821,
		0.980459,
		0.966673,
		0.947842,
		0.923529,
		0.893487,
		0.857652,
		0.816138,
		0.769224,
		0.717343,
		0.661072,
		0.601121,
		0.538323,
		0.473628,
		0.40809,
		0.342867,
		0.27921,
		0.218463,
		0.162051,
		0.111483,
		0.068345,
		0.034297,
		0.011069,
		0.000462,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.007289,
		0.027878,
		0.059902,
		0.101566,
		0.151149,
		0.207007,
		0.267574,
		0.331368,
		0.396993,
		0.463149,
		0.528629,
		0.592333,
		0.65327,
		0.710566,
		0.763477,
		0.811392,
		0.85385,
		0.890546,
		0.92135,
		0.946308,
		0.965663,
		0.97985,
		0.989494,
		0.995392,
		0.998468,
		0.999696,
		0.999984,
		1,
		0.999958,
		0.99951,
		0.99791,
		0.99422,
		0.987468,
		0.976757,
		0.961329,
		0.940607,
		0.914202,
		0.881923,
		0.843766,
		0.79991,
		0.750696,
		0.696627,
		0.638345,
		0.57663,
		0.512384,
		0.446628,
		0.380488,
		0.315194,
		0.252073,
		0.192542,
		0.138102,
		0.090341,
		0.050923,
		0.021587,
		0.004145,
		0.007353,
		0.028118,
		0.060404,
		0.102394,
		0.152345,
		0.208589,
		0.269544,
		0.333709,
		0.399673,
		0.466121,
		0.531836,
		0.595706,
		0.656733,
		0.71404,
		0.766878,
		0.814639,
		0.856866,
		0.893264,
		0.923712,
		0.948277,
		0.967218,
		0.980996,
		0.990264,
		0.995843,
		0.998681,
		0.999763,
		0.999991,
		1,
		0.999935,
		0.999381,
		0.997543,
		0.993458,
		0.986146,
		0.974716,
		0.958426,
		0.936718,
		0.909232,
		0.875806,
		0.836467,
		0.791425,
		0.741059,
		0.685901,
		0.626633,
		0.564067,
		0.499142,
		0.432915,
		0.366551,
		0.301315,
		0.238574,
		0.179781,
		0.12648,
		0.080297,
		0.042936,
		0.016179,
		0.001879,
		0.007418,
		0.028361,
		0.060913,
		0.103233,
		0.153554,
		0.210189,
		0.271535,
		0.336073,
		0.402378,
		0.469119,
		0.535067,
		0.599101,
		0.660215,
		0.717527,
		0.770286,
		0.817887,
		0.859876,
		0.895967,
		0.926053,
		0.950217,
		0.968741,
		0.982109,
		0.991001,
		0.996266,
		0.998873,
		0.999819,
		0.999996,
		0.999999,
		0.999905,
		0.999227,
		0.997128,
		0.992621,
		0.984721,
		0.972543,
		0.955361,
		0.932642,
		0.904051,
		0.869459,
		0.828924,
		0.78269,
		0.73117,
		0.674931,
		0.61469,
		0.551297,
		0.485726,
		0.419073,
		0.352538,
		0.287427,
		0.225145,
		0.167184,
		0.11513,
		0.070648,
		0.035484,
		0.011462,
		0.000479,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.007551,
		0.028857,
		0.061951,
		0.104941,
		0.156015,
		0.213443,
		0.27558,
		0.340874,
		0.407865,
		0.475193,
		0.541604,
		0.605958,
		0.667234,
		0.72454,
		0.777124,
		0.824381,
		0.865871,
		0.901325,
		0.930665,
		0.954009,
		0.971686,
		0.98423,
		0.992377,
		0.99703,
		0.9992,
		0.999901,
		0.999999,
		0.999996,
		0.999812,
		0.998834,
		0.99614,
		0.990705,
		0.981538,
		0.967772,
		0.948721,
		0.9239,
		0.893035,
		0.856057,
		0.813093,
		0.764456,
		0.710629,
		0.652255,
		0.59012,
		0.525152,
		0.4584,
		0.391034,
		0.324337,
		0.259691,
		0.198581,
		0.142584,
		0.093366,
		0.052677,
		0.02235,
		0.004296,
		0.007619,
		0.02911,
		0.06248,
		0.10581,
		0.157267,
		0.215097,
		0.277636,
		0.343311,
		0.410647,
		0.478268,
		0.54491,
		0.60942,
		0.670771,
		0.728066,
		0.780552,
		0.827626,
		0.868855,
		0.903979,
		0.932935,
		0.955861,
		0.973107,
		0.985237,
		0.993016,
		0.997372,
		0.999337,
		0.999931,
		1,
		0.999991,
		0.999746,
		0.998588,
		0.995558,
		0.989616,
		0.979768,
		0.965164,
		0.945135,
		0.919224,
		0.887188,
		0.848992,
		0.804797,
		0.754952,
		0.699976,
		0.640549,
		0.577498,
		0.511786,
		0.444504,
		0.37686,
		0.310176,
		0.245878,
		0.185491,
		0.130634,
		0.083016,
		0.044432,
		0.016757,
		0.001948,
		0.007688,
		0.029367,
		0.063015,
		0.106691,
		0.158534,
		0.21677,
		0.279714,
		0.345773,
		0.413455,
		0.48137,
		0.54824,
		0.612904,
		0.674326,
		0.731605,
		0.783986,
		0.83087,
		0.871829,
		0.906615,
		0.935179,
		0.957681,
		0.974494,
		0.98621,
		0.993622,
		0.997688,
		0.999457,
		0.999953,
		1,
		0.999982,
		0.999663,
		0.998304,
		0.994912,
		0.988431,
		0.977873,
		0.962398,
		0.941362,
		0.914337,
		0.881109,
		0.841678,
		0.796242,
		0.745185,
		0.689066,
		0.628601,
		0.564656,
		0.498233,
		0.430464,
		0.362598,
		0.295997,
		0.232129,
		0.172561,
		0.118957,
		0.073068,
		0.036734,
		0.011876,
		0.000497,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.007828,
		0.02989,
		0.064107,
		0.108485,
		0.161115,
		0.220175,
		0.283937,
		0.350772,
		0.41915,
		0.487653,
		0.554977,
		0.619939,
		0.681489,
		0.738718,
		0.790869,
		0.83735,
		0.877744,
		0.91183,
		0.93959,
		0.961225,
		0.977161,
		0.988048,
		0.994738,
		0.998244,
		0.999651,
		0.999982,
		1,
		0.999951,
		0.999437,
		0.997607,
		0.993406,
		0.985756,
		0.973679,
		0.956371,
		0.933236,
		0.903905,
		0.868232,
		0.826287,
		0.778343,
		0.72486,
		0.666474,
		0.603979,
		0.538323,
		0.470589,
		0.40199,
		0.333861,
		0.267647,
		0.204903,
		0.147285,
		0.096544,
		0.054523,
		0.023155,
		0.004454,
		0.0079,
		0.030156,
		0.064663,
		0.109399,
		0.162428,
		0.221906,
		0.286083,
		0.353309,
		0.422038,
		0.490835,
		0.558383,
		0.62349,
		0.685097,
		0.742293,
		0.794318,
		0.840584,
		0.880684,
		0.914407,
		0.941754,
		0.962947,
		0.97844,
		0.988913,
		0.995249,
		0.998486,
		0.999727,
		0.99999,
		1,
		0.999926,
		0.999289,
		0.997186,
		0.992536,
		0.984253,
		0.971369,
		0.953097,
		0.928869,
		0.898348,
		0.861423,
		0.818201,
		0.768993,
		0.714298,
		0.654791,
		0.591311,
		0.524843,
		0.456513,
		0.387578,
		0.319414,
		0.253513,
		0.191473,
		0.134994,
		0.085875,
		0.046007,
		0.017367,
		0.002021,
		0.007973,
		0.030427,
		0.065227,
		0.110324,
		0.163757,
		0.223658,
		0.288253,
		0.355873,
		0.424954,
		0.494044,
		0.561815,
		0.627063,
		0.688723,
		0.745878,
		0.797771,
		0.843815,
		0.883611,
		0.916963,
		0.943889,
		0.964635,
		0.979683,
		0.989742,
		0.995728,
		0.998706,
		0.999791,
		0.999995,
		0.999999,
		0.99989,
		0.999112,
		0.996711,
		0.991581,
		0.982633,
		0.968908,
		0.94964,
		0.924291,
		0.892557,
		0.854361,
		0.809849,
		0.759371,
		0.703467,
		0.642853,
		0.578407,
		0.51116,
		0.442279,
		0.373065,
		0.304937,
		0.239433,
		0.178197,
		0.122976,
		0.075614,
		0.038051,
		0.012313,
		0.000516,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.008121,
		0.030978,
		0.066377,
		0.11221,
		0.166465,
		0.227222,
		0.292664,
		0.361079,
		0.430867,
		0.500544,
		0.568755,
		0.634275,
		0.696026,
		0.753082,
		0.804685,
		0.850259,
		0.889423,
		0.922008,
		0.948071,
		0.967907,
		0.982057,
		0.991293,
		0.996595,
		0.999079,
		0.999886,
		0.999999,
		0.999995,
		0.999783,
		0.998659,
		0.995579,
		0.989394,
		0.979014,
		0.963506,
		0.942154,
		0.914479,
		0.880246,
		0.839455,
		0.79233,
		0.739302,
		0.680995,
		0.618209,
		0.551909,
		0.483212,
		0.413374,
		0.343786,
		0.27596,
		0.211524,
		0.152218,
		0.099885,
		0.056468,
		0.024004,
		0.004622,
		0.008197,
		0.031259,
		0.066963,
		0.113171,
		0.167843,
		0.229035,
		0.294906,
		0.363722,
		0.433865,
		0.503836,
		0.572264,
		0.637915,
		0.699703,
		0.756699,
		0.808146,
		0.853472,
		0.892307,
		0.924495,
		0.950115,
		0.969489,
		0.983187,
		0.992015,
		0.996984,
		0.999236,
		0.99992,
		1,
		0.999989,
		0.999707,
		0.998376,
		0.994912,
		0.98815,
		0.977002,
		0.960553,
		0.938112,
		0.909232,
		0.873716,
		0.831602,
		0.783156,
		0.728851,
		0.669354,
		0.605508,
		0.538323,
		0.468961,
		0.398725,
		0.32905,
		0.261496,
		0.197741,
		0.139573,
		0.088884,
		0.047667,
		0.018011,
		0.002098,
		0.008273,
		0.031545,
		0.067557,
		0.114144,
		0.169238,
		0.230869,
		0.297172,
		0.366393,
		0.436892,
		0.507156,
		0.575799,
		0.641577,
		0.703397,
		0.760326,
		0.811609,
		0.856678,
		0.895174,
		0.926957,
		0.952128,
		0.971035,
		0.984278,
		0.9927,
		0.997343,
		0.999373,
		0.999946,
		1,
		0.99998,
		0.999611,
		0.998049,
		0.994171,
		0.986798,
		0.974848,
		0.957423,
		0.93386,
		0.903749,
		0.866927,
		0.823476,
		0.773702,
		0.71812,
		0.657442,
		0.592557,
		0.524519,
		0.454537,
		0.383959,
		0.31427,
		0.247077,
		0.184108,
		0.127199,
		0.078295,
		0.039439,
		0.012775,
		0.000536,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.00843,
		0.032127,
		0.068768,
		0.116129,
		0.172081,
		0.234603,
		0.301781,
		0.371816,
		0.443032,
		0.51388,
		0.582946,
		0.648967,
		0.710833,
		0.767608,
		0.818536,
		0.863065,
		0.900855,
		0.931803,
		0.956054,
		0.974012,
		0.986346,
		0.993965,
		0.997978,
		0.999596,
		0.999979,
		1,
		0.999944,
		0.99935,
		0.997247,
		0.992444,
		0.983744,
		0.970081,
		0.950601,
		0.924704,
		0.892049,
		0.852555,
		0.806386,
		0.753936,
		0.69581,
		0.63281,
		0.565918,
		0.496282,
		0.425205,
		0.354133,
		0.28465,
		0.218463,
		0.1574,
		0.103402,
		0.058518,
		0.0249,
		0.004799,
		0.00851,
		0.032424,
		0.069387,
		0.11714,
		0.173528,
		0.236502,
		0.304124,
		0.37457,
		0.446146,
		0.517284,
		0.586559,
		0.652694,
		0.714575,
		0.771261,
		0.822,
		0.866244,
		0.903668,
		0.934185,
		0.957965,
		0.975444,
		0.987321,
		0.994545,
		0.998255,
		0.999684,
		0.999988,
		1,
		0.999914,
		0.999178,
		0.996763,
		0.991447,
		0.982029,
		0.967454,
		0.946896,
		0.919785,
		0.88582,
		0.84496,
		0.797413,
		0.743619,
		0.68423,
		0.620093,
		0.552238,
		0.481862,
		0.410318,
		0.339103,
		0.269849,
		0.204316,
		0.144385,
		0.092052,
		0.049418,
		0.018691,
		0.002179,
		0.008592,
		0.032726,
		0.070013,
		0.118164,
		0.174994,
		0.238425,
		0.306492,
		0.377353,
		0.449289,
		0.520718,
		0.590198,
		0.656444,
		0.718333,
		0.774923,
		0.825463,
		0.869413,
		0.90646,
		0.936538,
		0.959841,
		0.976835,
		0.988257,
		0.99509,
		0.998506,
		0.999758,
		0.999994,
		0.999999,
		0.999873,
		0.998974,
		0.996216,
		0.990352,
		0.980179,
		0.964657,
		0.942986,
		0.91463,
		0.879329,
		0.837084,
		0.78815,
		0.73301,
		0.672365,
		0.607109,
		0.538323,
		0.467253,
		0.395301,
		0.324015,
		0.255081,
		0.190312,
		0.131641,
		0.081119,
		0.040905,
		0.013263,
		0.000557,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.008758,
		0.033341,
		0.071291,
		0.120254,
		0.177981,
		0.242338,
		0.31131,
		0.383005,
		0.455665,
		0.527672,
		0.597555,
		0.664008,
		0.725895,
		0.782268,
		0.832383,
		0.875716,
		0.911982,
		0.941155,
		0.963484,
		0.979499,
		0.990011,
		0.996078,
		0.998935,
		0.999868,
		0.999999,
		0.999994,
		0.999749,
		0.998451,
		0.994912,
		0.987845,
		0.976047,
		0.958519,
		0.934518,
		0.903584,
		0.865538,
		0.820474,
		0.768736,
		0.710906,
		0.647781,
		0.580358,
		0.509814,
		0.437501,
		0.364925,
		0.29374,
		0.22574,
		0.162846,
		0.107106,
		0.060681,
		0.025848,
		0.004986,
		0.008842,
		0.033655,
		0.071944,
		0.121319,
		0.179502,
		0.24433,
		0.31376,
		0.385876,
		0.458899,
		0.531193,
		0.601273,
		0.667822,
		0.729698,
		0.785951,
		0.835839,
		0.878848,
		0.914709,
		0.943417,
		0.965249,
		0.98077,
		0.990828,
		0.996521,
		0.999115,
		0.999907,
		1,
		0.999987,
		0.99966,
		0.998123,
		0.994144,
		0.986419,
		0.973751,
		0.955164,
		0.929947,
		0.897679,
		0.858226,
		0.811728,
		0.758579,
		0.699408,
		0.635065,
		0.566596,
		0.495233,
		0.42238,
		0.349596,
		0.278592,
		0.211216,
		0.149448,
		0.095391,
		0.051267,
		0.019411,
		0.002265,
		0.008928,
		0.033974,
		0.072605,
		0.122398,
		0.181043,
		0.246346,
		0.316237,
		0.388776,
		0.462163,
		0.534743,
		0.605018,
		0.671658,
		0.733515,
		0.789639,
		0.839291,
		0.881966,
		0.917412,
		0.945646,
		0.966975,
		0.981999,
		0.991607,
		0.996932,
		0.999274,
		0.999937,
		1,
		0.999976,
		0.999548,
		0.997745,
		0.993291,
		0.984868,
		0.971292,
		0.951607,
		0.92514,
		0.89151,
		0.850628,
		0.802682,
		0.748116,
		0.687611,
		0.622067,
		0.552583,
		0.480445,
		0.407111,
		0.334196,
		0.263466,
		0.196828,
		0.136317,
		0.084099,
		0.042454,
		0.013779,
		0.000579,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.009104,
		0.034625,
		0.073955,
		0.1246,
		0.184184,
		0.25045,
		0.321275,
		0.394668,
		0.468785,
		0.541933,
		0.612586,
		0.679392,
		0.741192,
		0.797031,
		0.84618,
		0.888157,
		0.922742,
		0.950003,
		0.970307,
		0.984332,
		0.993045,
		0.99766,
		0.999531,
		0.999975,
		1,
		0.999934,
		0.999246,
		0.996818,
		0.991302,
		0.981366,
		0.965851,
		0.943859,
		0.91479,
		0.878352,
		0.834551,
		0.783673,
		0.726267,
		0.663118,
		0.595234,
		0.523824,
		0.450283,
		0.376183,
		0.303253,
		0.233376,
		0.168575,
		0.11101,
		0.062966,
		0.02685,
		0.005185,
		0.009194,
		0.034957,
		0.074644,
		0.125723,
		0.185784,
		0.25254,
		0.323837,
		0.397661,
		0.472143,
		0.545574,
		0.61641,
		0.683291,
		0.74505,
		0.800733,
		0.849616,
		0.891227,
		0.925367,
		0.952128,
		0.971912,
		0.985435,
		0.993706,
		0.997978,
		0.999632,
		0.999986,
		1,
		0.999899,
		0.999046,
		0.996257,
		0.990153,
		0.9794,
		0.962855,
		0.939652,
		0.909232,
		0.87135,
		0.82606,
		0.773702,
		0.714875,
		0.650423,
		0.581406,
		0.50909,
		0.43493,
		0.360554,
		0.28775,
		0.218463,
		0.154777,
		0.098913,
		0.053221,
		0.020173,
		0.002356,
		0.009285,
		0.035295,
		0.075343,
		0.126861,
		0.187405,
		0.254655,
		0.326429,
		0.400685,
		0.475533,
		0.549244,
		0.620261,
		0.68721,
		0.748921,
		0.804438,
		0.853044,
		0.894279,
		0.927963,
		0.954217,
		0.973475,
		0.986495,
		0.994328,
		0.998267,
		0.999718,
		0.999993,
		0.999999,
		0.999852,
		0.998808,
		0.995624,
		0.988892,
		0.977279,
		0.959663,
		0.935212,
		0.903409,
		0.864056,
		0.81726,
		0.763413,
		0.70317,
		0.637429,
		0.567308,
		0.494131,
		0.419412,
		0.344836,
		0.272257,
		0.203676,
		0.141244,
		0.087245,
		0.044093,
		0.014327,
		0.000602,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.009472,
		0.035984,
		0.07677,
		0.129184,
		0.19071,
		0.258963,
		0.331701,
		0.406828,
		0.48241,
		0.556677,
		0.628041,
		0.695109,
		0.756699,
		0.811856,
		0.859876,
		0.900326,
		0.933068,
		0.95828,
		0.976473,
		0.988485,
		0.995458,
		0.998762,
		0.999846,
		0.999999,
		0.999993,
		0.999707,
		0.9982,
		0.994115,
		0.986004,
		0.972543,
		0.952661,
		0.925601,
		0.890936,
		0.848568,
		0.798713,
		0.741872,
		0.678813,
		0.610552,
		0.538323,
		0.463572,
		0.387931,
		0.313213,
		0.241394,
		0.174607,
		0.11513,
		0.065382,
		0.027912,
		0.005396,
		0.009567,
		0.036336,
		0.077498,
		0.130369,
		0.192394,
		0.261156,
		0.334382,
		0.409949,
		0.485897,
		0.560439,
		0.631971,
		0.699089,
		0.760605,
		0.815567,
		0.863277,
		0.903317,
		0.935573,
		0.960253,
		0.977906,
		0.989415,
		0.995967,
		0.99897,
		0.999891,
		1,
		0.999985,
		0.999603,
		0.997819,
		0.993226,
		0.984361,
		0.969911,
		0.948834,
		0.920413,
		0.88427,
		0.840361,
		0.788955,
		0.730613,
		0.666161,
		0.596674,
		0.523449,
		0.447991,
		0.371999,
		0.297347,
		0.226079,
		0.160392,
		0.102633,
		0.055289,
		0.02098,
		0.002453,
		0.009664,
		0.036694,
		0.078237,
		0.13157,
		0.194101,
		0.263377,
		0.337094,
		0.413102,
		0.489418,
		0.564232,
		0.635927,
		0.703088,
		0.764523,
		0.819279,
		0.866667,
		0.906287,
		0.938045,
		0.962186,
		0.979295,
		0.990301,
		0.996439,
		0.999153,
		0.999926,
		1,
		0.999972,
		0.999473,
		0.99738,
		0.992239,
		0.982575,
		0.967092,
		0.944778,
		0.91496,
		0.877309,
		0.831837,
		0.778869,
		0.719025,
		0.653193,
		0.582507,
		0.508329,
		0.432226,
		0.355962,
		0.281477,
		0.21088,
		0.146439,
		0.090569,
		0.045828,
		0.014907,
		0.000627,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.009862,
		0.037425,
		0.079747,
		0.134022,
		0.197581,
		0.267901,
		0.342613,
		0.41951,
		0.496559,
		0.571912,
		0.643919,
		0.711144,
		0.772387,
		0.826699,
		0.873411,
		0.912153,
		0.942888,
		0.965925,
		0.981935,
		0.991944,
		0.997277,
		0.999452,
		0.999971,
		1,
		0.999923,
		0.999119,
		0.996301,
		0.989936,
		0.978541,
		0.960857,
		0.935944,
		0.903223,
		0.862471,
		0.813813,
		0.757694,
		0.694856,
		0.626313,
		0.553326,
		0.477387,
		0.400195,
		0.323646,
		0.249819,
		0.180962,
		0.119481,
		0.067939,
		0.029038,
		0.00562,
		0.009964,
		0.037798,
		0.080519,
		0.135274,
		0.199356,
		0.270205,
		0.34542,
		0.422765,
		0.50018,
		0.575799,
		0.647954,
		0.715201,
		0.776333,
		0.830407,
		0.876762,
		0.915048,
		0.945255,
		0.967729,
		0.983187,
		0.9927,
		0.997644,
		0.99957,
		0.999984,
		1,
		0.999882,
		0.998886,
		0.995649,
		0.988607,
		0.976276,
		0.957423,
		0.931149,
		0.896922,
		0.854579,
		0.804301,
		0.746597,
		0.682272,
		0.612405,
		0.538323,
		0.461585,
		0.383959,
		0.307411,
		0.234088,
		0.166312,
		0.106564,
		0.057479,
		0.021837,
		0.002556,
		0.010067,
		0.038178,
		0.081301,
		0.136543,
		0.201154,
		0.272538,
		0.34826,
		0.426054,
		0.503836,
		0.579717,
		0.652015,
		0.719275,
		0.780288,
		0.834112,
		0.880097,
		0.917916,
		0.947585,
		0.969489,
		0.984391,
		0.993413,
		0.997978,
		0.999669,
		0.999992,
		0.999999,
		0.999826,
		0.998608,
		0.994912,
		0.987147,
		0.973833,
		0.953766,
		0.926089,
		0.890323,
		0.846361,
		0.794448,
		0.735156,
		0.669354,
		0.598188,
		0.523054,
		0.445578,
		0.367599,
		0.291153,
		0.218463,
		0.151921,
		0.094085,
		0.047667,
		0.015523,
		0.000654,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.010277,
		0.038954,
		0.0829,
		0.139134,
		0.204822,
		0.277292,
		0.354039,
		0.432738,
		0.511251,
		0.587648,
		0.660215,
		0.727478,
		0.78822,
		0.841508,
		0.886718,
		0.923565,
		0.952128,
		0.972873,
		0.986658,
		0.994712,
		0.998551,
		0.999819,
		0.999999,
		0.999992,
		0.999656,
		0.997897,
		0.993157,
		0.983803,
		0.968379,
		0.945745,
		0.91514,
		0.876195,
		0.828924,
		0.773702,
		0.71123,
		0.642517,
		0.568845,
		0.491749,
		0.413,
		0.334582,
		0.258679,
		0.187663,
		0.124081,
		0.070648,
		0.030233,
		0.005858,
		0.010385,
		0.03935,
		0.083718,
		0.140457,
		0.206693,
		0.279714,
		0.35698,
		0.436133,
		0.515011,
		0.591661,
		0.664354,
		0.731605,
		0.792195,
		0.845197,
		0.890001,
		0.926344,
		0.954338,
		0.974494,
		0.98772,
		0.995298,
		0.998793,
		0.999872,
		1,
		0.999982,
		0.999533,
		0.997451,
		0.992122,
		0.981901,
		0.965348,
		0.941362,
		0.909232,
		0.868649,
		0.819691,
		0.762798,
		0.698742,
		0.628601,
		0.553728,
		0.475734,
		0.39646,
		0.317968,
		0.242518,
		0.172561,
		0.110724,
		0.059802,
		0.022747,
		0.002666,
		0.010495,
		0.039753,
		0.084547,
		0.141799,
		0.208589,
		0.282166,
		0.359954,
		0.439564,
		0.518807,
		0.595706,
		0.668518,
		0.735749,
		0.796177,
		0.84888,
		0.893264,
		0.92909,
		0.956507,
		0.976067,
		0.988734,
		0.995843,
		0.999006,
		0.999913,
		1,
		0.999967,
		0.999381,
		0.996937,
		0.990973,
		0.979834,
		0.962103,
		0.936718,
		0.903024,
		0.860772,
		0.810106,
		0.751535,
		0.685901,
		0.614355,
		0.538323,
		0.45949,
		0.379775,
		0.301315,
		0.226451,
		0.157713,
		0.097809,
		0.049618,
		0.016179,
		0.000682,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.010719,
		0.040578,
		0.086243,
		0.14454,
		0.212459,
		0.287165,
		0.366009,
		0.446537,
		0.526505,
		0.603892,
		0.676922,
		0.744084,
		0.804153,
		0.85622,
		0.899725,
		0.934481,
		0.960712,
		0.979066,
		0.990617,
		0.996812,
		0.999355,
		0.999966,
		1,
		0.999909,
		0.998965,
		0.995675,
		0.988294,
		0.975164,
		0.954925,
		0.926607,
		0.889668,
		0.843991,
		0.789856,
		0.727916,
		0.65916,
		0.584888,
		0.50668,
		0.426374,
		0.346048,
		0.268,
		0.194735,
		0.128949,
		0.073521,
		0.031504,
		0.006111,
		0.010834,
		0.041,
		0.08711,
		0.14594,
		0.214433,
		0.289713,
		0.369091,
		0.45008,
		0.530408,
		0.608032,
		0.681162,
		0.748275,
		0.808146,
		0.859876,
		0.90292,
		0.937122,
		0.962746,
		0.980491,
		0.991486,
		0.997238,
		0.999493,
		0.999981,
		1,
		0.999861,
		0.99869,
		0.994912,
		0.986746,
		0.972543,
		0.950973,
		0.92112,
		0.882503,
		0.835071,
		0.77918,
		0.715554,
		0.645261,
		0.569676,
		0.49046,
		0.409531,
		0.32905,
		0.251397,
		0.179163,
		0.11513,
		0.062267,
		0.023716,
		0.002783,
		0.010951,
		0.041428,
		0.08799,
		0.14736,
		0.216434,
		0.292293,
		0.372208,
		0.45366,
		0.534347,
		0.612205,
		0.685426,
		0.75248,
		0.812141,
		0.863519,
		0.906089,
		0.939726,
		0.964734,
		0.981865,
		0.992308,
		0.997627,
		0.99961,
		0.99999,
		0.999999,
		0.999794,
		0.998364,
		0.99405,
		0.985047,
		0.969716,
		0.946765,
		0.915332,
		0.875,
		0.82579,
		0.768129,
		0.702821,
		0.63101,
		0.554152,
		0.473988,
		0.392521,
		0.311993,
		0.234873,
		0.163836,
		0.101756,
		0.051692,
		0.016877,
		0.000713,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.01119,
		0.042305,
		0.089791,
		0.150262,
		0.220519,
		0.297552,
		0.378555,
		0.460936,
		0.542336,
		0.620647,
		0.694027,
		0.760931,
		0.820135,
		0.870768,
		0.912347,
		0.944814,
		0.968563,
		0.984457,
		0.993807,
		0.998294,
		0.999785,
		0.999999,
		0.99999,
		0.999593,
		0.997527,
		0.991996,
		0.981155,
		0.963404,
		0.937538,
		0.902812,
		0.858946,
		0.80611,
		0.744885,
		0.676234,
		0.601465,
		0.522198,
		0.440344,
		0.358077,
		0.277816,
		0.202206,
		0.134105,
		0.076573,
		0.032856,
		0.006382,
		0.011312,
		0.042755,
		0.090711,
		0.151745,
		0.222604,
		0.300233,
		0.381784,
		0.464632,
		0.546387,
		0.624916,
		0.698364,
		0.765177,
		0.824131,
		0.874369,
		0.915432,
		0.947296,
		0.970403,
		0.985674,
		0.994485,
		0.998577,
		0.999848,
		1,
		0.999979,
		0.999448,
		0.997003,
		0.990785,
		0.978942,
		0.959898,
		0.932498,
		0.89606,
		0.850378,
		0.795699,
		0.732684,
		0.66238,
		0.586177,
		0.505785,
		0.423201,
		0.340688,
		0.260756,
		0.186142,
		0.119801,
		0.064888,
		0.024747,
		0.002907,
		0.011437,
		0.043211,
		0.091646,
		0.15325,
		0.224718,
		0.302949,
		0.385053,
		0.468368,
		0.550474,
		0.629217,
		0.702724,
		0.769433,
		0.828126,
		0.877955,
		0.918486,
		0.949734,
		0.972191,
		0.986838,
		0.995118,
		0.998827,
		0.999897,
		1,
		0.999961,
		0.999268,
		0.996398,
		0.98944,
		0.976536,
		0.956143,
		0.927158,
		0.888967,
		0.841438,
		0.784898,
		0.720092,
		0.648151,
		0.570554,
		0.489096,
		0.405867,
		0.32322,
		0.243759,
		0.170317,
		0.105945,
		0.053897,
		0.017621,
		0.000745,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.011692,
		0.044146,
		0.09356,
		0.156326,
		0.229035,
		0.308487,
		0.391707,
		0.475959,
		0.558763,
		0.637915,
		0.711512,
		0.77798,
		0.836105,
		0.88507,
		0.924495,
		0.954475,
		0.975609,
		0.989007,
		0.996245,
		0.999236,
		0.999959,
		1,
		0.999892,
		0.998776,
		0.994912,
		0.986304,
		0.971103,
		0.947842,
		0.915536,
		0.873716,
		0.822408,
		0.762104,
		0.693726,
		0.618579,
		0.538323,
		0.45494,
		0.370703,
		0.288158,
		0.210104,
		0.139573,
		0.079817,
		0.034297,
		0.006671,
		0.011823,
		0.044624,
		0.094539,
		0.157899,
		0.231239,
		0.31131,
		0.395094,
		0.479816,
		0.562964,
		0.642312,
		0.71594,
		0.782268,
		0.840087,
		0.888597,
		0.927446,
		0.956775,
		0.977237,
		0.990011,
		0.996742,
		0.999399,
		0.999977,
		1,
		0.999834,
		0.998451,
		0.994013,
		0.984492,
		0.968053,
		0.943272,
		0.909232,
		0.865538,
		0.812301,
		0.750101,
		0.679948,
		0.603241,
		0.521733,
		0.437501,
		0.352917,
		0.270627,
		0.193529,
		0.12476,
		0.067676,
		0.025848,
		0.003041,
		0.011957,
		0.045111,
		0.095534,
		0.159495,
		0.233473,
		0.31417,
		0.398522,
		0.483714,
		0.567204,
		0.64674,
		0.72039,
		0.786565,
		0.844063,
		0.892101,
		0.930361,
		0.959025,
		0.97881,
		0.990961,
		0.997195,
		0.999536,
		0.999989,
		0.999998,
		0.999755,
		0.998064,
		0.992999,
		0.982503,
		0.964764,
		0.938407,
		0.902584,
		0.85698,
		0.801791,
		0.737688,
		0.665773,
		0.58754,
		0.504839,
		0.419845,
		0.33503,
		0.253142,
		0.177183,
		0.110396,
		0.056247,
		0.018415,
		0.00078,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.01223,
		0.046108,
		0.09757,
		0.162759,
		0.238038,
		0.320005,
		0.405501,
		0.491635,
		0.575799,
		0.655692,
		0.729352,
		0.795181,
		0.85199,
		0.899037,
		0.93607,
		0.963372,
		0.981785,
		0.9927,
		0.997978,
		0.999744,
		0.999998,
		0.999988,
		0.999515,
		0.997073,
		0.990579,
		0.977948,
		0.957423,
		0.927745,
		0.888213,
		0.838682,
		0.77953,
		0.711615,
		0.636233,
		0.555074,
		0.470189,
		0.383959,
		0.299063,
		0.218463,
		0.145377,
		0.08327,
		0.035835,
		0.006979,
		0.01237,
		0.046619,
		0.098612,
		0.164428,
		0.24037,
		0.32298,
		0.409054,
		0.495659,
		0.580154,
		0.660215,
		0.733863,
		0.799498,
		0.855939,
		0.902465,
		0.938862,
		0.965467,
		0.983187,
		0.99349,
		0.99831,
		0.999819,
		1,
		0.999975,
		0.999343,
		0.996452,
		0.989154,
		0.975358,
		0.953345,
		0.921922,
		0.880467,
		0.828924,
		0.767766,
		0.69795,
		0.620871,
		0.538323,
		0.452462,
		0.365774,
		0.281048,
		0.201355,
		0.130029,
		0.070648,
		0.027023,
		0.003183,
		0.012512,
		0.047138,
		0.099671,
		0.166123,
		0.242735,
		0.325995,
		0.41265,
		0.499726,
		0.584549,
		0.664769,
		0.738394,
		0.80382,
		0.859876,
		0.905864,
		0.94161,
		0.967506,
		0.98453,
		0.994227,
		0.998605,
		0.999876,
		1,
		0.999953,
		0.999128,
		0.995735,
		0.987569,
		0.972543,
		0.94898,
		0.915755,
		0.872333,
		0.818748,
		0.755573,
		0.683863,
		0.605119,
		0.52124,
		0.434489,
		0.347461,
		0.263058,
		0.184465,
		0.11513,
		0.058753,
		0.019264,
		0.000817,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.012805,
		0.048204,
		0.101841,
		0.16959,
		0.247567,
		0.332145,
		0.419973,
		0.50799,
		0.593455,
		0.673969,
		0.747512,
		0.812474,
		0.867708,
		0.912568,
		0.946968,
		0.971415,
		0.987038,
		0.995546,
		0.999088,
		0.999951,
		1,
		0.999871,
		0.998541,
		0.993974,
		0.983876,
		0.966185,
		0.939331,
		0.902341,
		0.854856,
		0.797109,
		0.729875,
		0.654422,
		0.572465,
		0.486123,
		0.397884,
		0.31057,
		0.227317,
		0.151546,
		0.086952,
		0.037478,
		0.00731,
		0.012955,
		0.04875,
		0.102952,
		0.171364,
		0.250035,
		0.335282,
		0.423701,
		0.512188,
		0.597967,
		0.678615,
		0.752097,
		0.816804,
		0.8716,
		0.91587,
		0.949573,
		0.973282,
		0.988203,
		0.996128,
		0.999281,
		0.999973,
		1,
		0.999801,
		0.998153,
		0.992908,
		0.981742,
		0.962616,
		0.934022,
		0.895068,
		0.845492,
		0.78563,
		0.716363,
		0.639069,
		0.555576,
		0.468117,
		0.379296,
		0.292057,
		0.209654,
		0.135635,
		0.073818,
		0.02828,
		0.003336,
		0.013107,
		0.049305,
		0.104082,
		0.173165,
		0.25254,
		0.338462,
		0.427474,
		0.51643,
		0.602518,
		0.683291,
		0.756699,
		0.821134,
		0.875474,
		0.919137,
		0.952128,
		0.975089,
		0.989308,
		0.996662,
		0.999445,
		0.999986,
		0.999998,
		0.999707,
		0.997692,
		0.991705,
		0.9794,
		0.958769,
		0.928371,
		0.887401,
		0.835698,
		0.773702,
		0.702404,
		0.623297,
		0.538323,
		0.449834,
		0.360554,
		0.273547,
		0.192195,
		0.120172,
		0.061429,
		0.020173,
		0.000857,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.013421,
		0.050445,
		0.106396,
		0.176852,
		0.257659,
		0.34495,
		0.435159,
		0.52505,
		0.61174,
		0.692731,
		0.76595,
		0.829789,
		0.883162,
		0.925553,
		0.957078,
		0.978519,
		0.991337,
		0.997585,
		0.999693,
		0.999998,
		0.999986,
		0.999419,
		0.996509,
		0.988837,
		0.974035,
		0.950184,
		0.91599,
		0.870838,
		0.814777,
		0.74847,
		0.673139,
		0.590511,
		0.502771,
		0.412516,
		0.32272,
		0.236704,
		0.15811,
		0.09088,
		0.039236,
		0.007665,
		0.013582,
		0.05103,
		0.107582,
		0.178739,
		0.260276,
		0.34826,
		0.439072,
		0.529428,
		0.61641,
		0.697495,
		0.770596,
		0.834112,
		0.886971,
		0.9287,
		0.959469,
		0.98014,
		0.992261,
		0.997978,
		0.999782,
		1,
		0.99997,
		0.999212,
		0.995768,
		0.987147,
		0.970985,
		0.945417,
		0.909232,
		0.861918,
		0.803632,
		0.735156,
		0.657831,
		0.573509,
		0.484498,
		0.393526,
		0.303696,
		0.218463,
		0.141606,
		0.077206,
		0.029626,
		0.003501,
		0.013746,
		0.051625,
		0.108788,
		0.180656,
		0.262931,
		0.351615,
		0.443032,
		0.533853,
		0.62112,
		0.702287,
		0.775256,
		0.838428,
		0.890756,
		0.931803,
		0.961803,
		0.981696,
		0.993125,
		0.998327,
		0.999851,
		1,
		0.999944,
		0.998953,
		0.994912,
		0.985268,
		0.967671,
		0.940314,
		0.902078,
		0.852555,
		0.792019,
		0.721369,
		0.642073,
		0.556109,
		0.465916,
		0.374349,
		0.28465,
		0.200411,
		0.12555,
		0.064291,
		0.021147,
		0.000899,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.014083,
		0.052846,
		0.111261,
		0.184582,
		0.268359,
		0.358463,
		0.451099,
		0.542841,
		0.630658,
		0.711954,
		0.784611,
		0.84704,
		0.898242,
		0.93787,
		0.966289,
		0.984612,
		0.994676,
		0.998903,
		0.999941,
		1,
		0.999844,
		0.998248,
		0.992809,
		0.980889,
		0.960187,
		0.929041,
		0.886525,
		0.832457,
		0.767353,
		0.692366,
		0.609221,
		0.520162,
		0.427896,
		0.335557,
		0.246666,
		0.165103,
		0.09508,
		0.041121,
		0.008046,
		0.014256,
		0.053473,
		0.112529,
		0.186592,
		0.271135,
		0.361957,
		0.455206,
		0.547405,
		0.635486,
		0.716828,
		0.789303,
		0.851331,
		0.901939,
		0.94083,
		0.968439,
		0.98597,
		0.995364,
		0.999134,
		0.999967,
		1,
		0.99976,
		0.997782,
		0.991538,
		0.978359,
		0.955987,
		0.92284,
		0.878098,
		0.821699,
		0.754286,
		0.677145,
		0.592135,
		0.501639,
		0.408505,
		0.31601,
		0.227824,
		0.147975,
		0.08083,
		0.031071,
		0.003677,
		0.014433,
		0.054111,
		0.113819,
		0.188634,
		0.273952,
		0.365499,
		0.459363,
		0.552017,
		0.640354,
		0.721729,
		0.794004,
		0.85561,
		0.905603,
		0.943738,
		0.970524,
		0.987261,
		0.995996,
		0.99933,
		0.999983,
		0.999998,
		0.999645,
		0.997227,
		0.990101,
		0.975582,
		0.951461,
		0.916242,
		0.869219,
		0.810454,
		0.740722,
		0.661443,
		0.574618,
		0.482769,
		0.388893,
		0.296414,
		0.209152,
		0.131291,
		0.067358,
		0.022194,
		0.000945,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.014796,
		0.055422,
		0.116464,
		0.192819,
		0.279714,
		0.372731,
		0.467831,
		0.561384,
		0.650207,
		0.731605,
		0.803427,
		0.864126,
		0.912824,
		0.949389,
		0.974494,
		0.989641,
		0.997092,
		0.999627,
		0.999998,
		0.999982,
		0.999296,
		0.995804,
		0.986675,
		0.969224,
		0.941362,
		0.901795,
		0.850054,
		0.786465,
		0.712079,
		0.628601,
		0.538323,
		0.444067,
		0.349131,
		0.25725,
		0.172561,
		0.099574,
		0.043145,
		0.008457,
		0.014982,
		0.056095,
		0.117821,
		0.194962,
		0.282661,
		0.376422,
		0.472143,
		0.56614,
		0.655192,
		0.736579,
		0.808146,
		0.868358,
		0.916375,
		0.952128,
		0.976376,
		0.990728,
		0.997561,
		0.999735,
		1,
		0.999963,
		0.999046,
		0.994912,
		0.984657,
		0.965609,
		0.935757,
		0.893914,
		0.839741,
		0.773702,
		0.696992,
		0.611467,
		0.51957,
		0.424278,
		0.32905,
		0.237781,
		0.154777,
		0.084714,
		0.032625,
		0.003868,
		0.015172,
		0.056781,
		0.119202,
		0.197141,
		0.285652,
		0.380165,
		0.476508,
		0.570945,
		0.660215,
		0.741577,
		0.812868,
		0.872571,
		0.919885,
		0.954807,
		0.978187,
		0.991748,
		0.997978,
		0.999819,
		1,
		0.999931,
		0.998733,
		0.993882,
		0.982413,
		0.961681,
		0.929759,
		0.885575,
		0.828924,
		0.760415,
		0.681392,
		0.593863,
		0.500431,
		0.404232,
		0.308888,
		0.218463,
		0.137432,
		0.070648,
		0.023321,
		0.000995,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.015563,
		0.05819,
		0.122037,
		0.201607,
		0.291774,
		0.387806,
		0.485397,
		0.580701,
		0.670377,
		0.751638,
		0.822315,
		0.880929,
		0.926769,
		0.959974,
		0.981594,
		0.993585,
		0.998669,
		0.999928,
		1,
		0.999809,
		0.997877,
		0.991352,
		0.977183,
		0.952816,
		0.916514,
		0.867459,
		0.805731,
		0.732241,
		0.648648,
		0.557281,
		0.461072,
		0.363491,
		0.268504,
		0.180527,
		0.104393,
		0.045322,
		0.0089,
		0.015765,
		0.058915,
		0.123492,
		0.203896,
		0.294906,
		0.391707,
		0.489923,
		0.585653,
		0.675515,
		0.756699,
		0.827036,
		0.88507,
		0.930138,
		0.962458,
		0.983187,
		0.994403,
		0.998946,
		0.999959,
		0.999999,
		0.999707,
		0.997312,
		0.989821,
		0.974162,
		0.947842,
		0.909232,
		0.857652,
		0.793332,
		0.717343,
		0.631507,
		0.538323,
		0.440894,
		0.342867,
		0.248386,
		0.162051,
		0.088884,
		0.034297,
		0.004073,
		0.01597,
		0.059653,
		0.124973,
		0.206223,
		0.298086,
		0.395662,
		0.494505,
		0.590655,
		0.68069,
		0.761779,
		0.831754,
		0.889182,
		0.933455,
		0.964874,
		0.984704,
		0.995156,
		0.999183,
		0.99998,
		0.999997,
		0.999567,
		0.996638,
		0.988091,
		0.970847,
		0.942482,
		0.901488,
		0.847326,
		0.780384,
		0.701897,
		0.613858,
		0.518938,
		0.420418,
		0.322126,
		0.228392,
		0.144007,
		0.074184,
		0.024536,
		0.001049,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.016392,
		0.06117,
		0.128015,
		0.210995,
		0.304595,
		0.403741,
		0.503836,
		0.600807,
		0.69115,
		0.771993,
		0.841172,
		0.897313,
		0.939924,
		0.969489,
		0.987511,
		0.996468,
		0.999544,
		0.999997,
		0.999978,
		0.99914,
		0.994912,
		0.983968,
		0.963257,
		0.930531,
		0.884543,
		0.82506,
		0.752804,
		0.669354,
		0.577059,
		0.478955,
		0.378694,
		0.280485,
		0.189047,
		0.109566,
		0.047667,
		0.009378,
		0.01661,
		0.061951,
		0.129577,
		0.213443,
		0.307926,
		0.407865,
		0.508587,
		0.605958,
		0.696434,
		0.777124,
		0.845868,
		0.901325,
		0.943071,
		0.971686,
		0.9888,
		0.99703,
		0.999675,
		0.999999,
		0.999955,
		0.998834,
		0.993829,
		0.981538,
		0.958943,
		0.9239,
		0.875306,
		0.813093,
		0.738155,
		0.652255,
		0.557927,
		0.4584,
		0.357518,
		0.259691,
		0.169842,
		0.093366,
		0.036101,
		0.004296,
		0.016832,
		0.062746,
		0.131168,
		0.215931,
		0.31131,
		0.412048,
		0.513395,
		0.611159,
		0.701753,
		0.782268,
		0.850552,
		0.905299,
		0.946156,
		0.973805,
		0.990011,
		0.997533,
		0.999777,
		1,
		0.999916,
		0.998451,
		0.992579,
		0.978837,
		0.954256,
		0.916807,
		0.865538,
		0.800552,
		0.722924,
		0.634605,
		0.538323,
		0.437501,
		0.336187,
		0.238994,
		0.151058,
		0.07799,
		0.025848,
		0.001107,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.017289,
		0.064384,
		0.134438,
		0.221038,
		0.318238,
		0.420591,
		0.523188,
		0.621711,
		0.712494,
		0.792593,
		0.859876,
		0.913121,
		0.952128,
		0.977805,
		0.992201,
		0.998368,
		0.999911,
		1,
		0.999765,
		0.997403,
		0.989507,
		0.972543,
		0.943681,
		0.901155,
		0.84434,
		0.773702,
		0.690698,
		0.597675,
		0.497762,
		0.394798,
		0.293251,
		0.19817,
		0.11513,
		0.050198,
		0.009896,
		0.017525,
		0.065227,
		0.136118,
		0.223658,
		0.321785,
		0.424954,
		0.528173,
		0.627063,
		0.717915,
		0.797771,
		0.864511,
		0.916963,
		0.95501,
		0.979683,
		0.993181,
		0.998706,
		0.99995,
		0.999999,
		0.999638,
		0.996711,
		0.987648,
		0.968908,
		0.937751,
		0.892557,
		0.832877,
		0.759371,
		0.673697,
		0.578407,
		0.476846,
		0.373065,
		0.271757,
		0.178197,
		0.098193,
		0.038051,
		0.004537,
		0.017765,
		0.066086,
		0.137829,
		0.226323,
		0.325389,
		0.429378,
		0.533218,
		0.632464,
		0.723368,
		0.802956,
		0.869125,
		0.920755,
		0.957819,
		0.981477,
		0.994084,
		0.998994,
		0.999975,
		0.999997,
		0.999465,
		0.995886,
		0.985548,
		0.964919,
		0.931363,
		0.883417,
		0.820817,
		0.744423,
		0.656102,
		0.558619,
		0.455537,
		0.351133,
		0.250328,
		0.158632,
		0.082095,
		0.027267,
		0.00117,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.018262,
		0.067857,
		0.141349,
		0.231794,
		0.332769,
		0.438416,
		0.54349,
		0.643416,
		0.734365,
		0.81334,
		0.87828,
		0.928179,
		0.963209,
		0.98481,
		0.995666,
		0.999435,
		0.999996,
		0.999973,
		0.998938,
		0.993771,
		0.980542,
		0.955789,
		0.917126,
		0.863435,
		0.794849,
		0.712649,
		0.61914,
		0.517538,
		0.411865,
		0.306869,
		0.207954,
		0.121124,
		0.052935,
		0.010458,
		0.018518,
		0.068768,
		0.14316,
		0.234603,
		0.336549,
		0.443032,
		0.548718,
		0.648967,
		0.739908,
		0.818536,
		0.882814,
		0.931803,
		0.965784,
		0.986346,
		0.996348,
		0.999596,
		0.999999,
		0.999944,
		0.99856,
		0.992444,
		0.977591,
		0.950601,
		0.909232,
		0.852555,
		0.780913,
		0.69581,
		0.599783,
		0.496282,
		0.389573,
		0.28465,
		0.187172,
		0.103402,
		0.040162,
		0.004799,
		0.018779,
		0.069699,
		0.145004,
		0.237461,
		0.34039,
		0.447714,
		0.554008,
		0.654566,
		0.745479,
		0.823732,
		0.887317,
		0.935365,
		0.968276,
		0.987794,
		0.996959,
		0.999723,
		1,
		0.999895,
		0.998086,
		0.990912,
		0.974313,
		0.944967,
		0.900792,
		0.841057,
		0.766326,
		0.678333,
		0.579853,
		0.474581,
		0.367032,
		0.262459,
		0.166781,
		0.086531,
		0.028807,
		0.001239,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.019319,
		0.071616,
		0.148801,
		0.243331,
		0.34826,
		0.457278,
		0.564777,
		0.665912,
		0.756699,
		0.834112,
		0.896212,
		0.94229,
		0.972999,
		0.990425,
		0.997978,
		0.999889,
		1,
		0.999707,
		0.99679,
		0.987147,
		0.966673,
		0.932261,
		0.882183,
		0.816138,
		0.735156,
		0.641458,
		0.538323,
		0.42996,
		0.321409,
		0.218463,
		0.127594,
		0.055902,
		0.011069,
		0.019597,
		0.072605,
		0.150754,
		0.246346,
		0.352291,
		0.462163,
		0.570256,
		0.671658,
		0.762345,
		0.839291,
		0.900598,
		0.945646,
		0.975226,
		0.991607,
		0.998392,
		0.999937,
		0.999999,
		0.999548,
		0.995933,
		0.984868,
		0.962262,
		0.92514,
		0.87197,
		0.802682,
		0.718555,
		0.622067,
		0.516759,
		0.407111,
		0.298442,
		0.196828,
		0.109032,
		0.042454,
		0.005084,
		0.019882,
		0.073614,
		0.152746,
		0.249415,
		0.356389,
		0.467118,
		0.575799,
		0.677451,
		0.768013,
		0.84446,
		0.90494,
		0.948926,
		0.97736,
		0.9927,
		0.998747,
		0.999968,
		0.999996,
		0.999333,
		0.994912,
		0.982294,
		0.957423,
		0.917472,
		0.861121,
		0.788542,
		0.701268,
		0.602046,
		0.494689,
		0.383959,
		0.27546,
		0.175563,
		0.091333,
		0.030481,
		0.001314,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.02047,
		0.075696,
		0.156848,
		0.255722,
		0.364787,
		0.47724,
		0.587077,
		0.689177,
		0.779408,
		0.854755,
		0.913472,
		0.955247,
		0.981342,
		0.994625,
		0.999293,
		0.999995,
		0.999966,
		0.998674,
		0.992293,
		0.976158,
		0.946351,
		0.900395,
		0.837433,
		0.758149,
		0.664622,
		0.560159,
		0.449152,
		0.336951,
		0.229766,
		0.134589,
		0.059124,
		0.011736,
		0.020774,
		0.07677,
		0.15896,
		0.258963,
		0.369091,
		0.48241,
		0.592813,
		0.695109,
		0.785132,
		0.859876,
		0.917657,
		0.95828,
		0.983187,
		0.995458,
		0.999493,
		0.999999,
		0.999929,
		0.9982,
		0.99065,
		0.972543,
		0.940063,
		0.890936,
		0.824552,
		0.741872,
		0.645261,
		0.538323,
		0.425753,
		0.313213,
		0.207232,
		0.11513,
		0.044948,
		0.005396,
		0.021084,
		0.077867,
		0.161115,
		0.262263,
		0.373465,
		0.487653,
		0.598615,
		0.701086,
		0.790869,
		0.864974,
		0.921781,
		0.961225,
		0.984931,
		0.996208,
		0.999651,
		1,
		0.999867,
		0.997607,
		0.988753,
		0.968525,
		0.933236,
		0.880827,
		0.810953,
		0.72486,
		0.625212,
		0.515919,
		0.40199,
		0.289411,
		0.185043,
		0.096544,
		0.032305,
		0.001396,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.021727,
		0.080132,
		0.165555,
		0.269049,
		0.382435,
		0.498366,
		0.610413,
		0.71317,
		0.802379,
		0.875088,
		0.929834,
		0.966832,
		0.988116,
		0.997465,
		0.999859,
		1,
		0.999629,
		0.995985,
		0.984089,
		0.959167,
		0.91785,
		0.858565,
		0.781532,
		0.688606,
		0.583079,
		0.469512,
		0.353578,
		0.241943,
		0.142168,
		0.062632,
		0.012464,
		0.022059,
		0.081301,
		0.167843,
		0.272538,
		0.387032,
		0.503836,
		0.61641,
		0.719275,
		0.808146,
		0.880097,
		0.933754,
		0.969489,
		0.989554,
		0.997978,
		0.99992,
		0.999999,
		0.999429,
		0.994912,
		0.981267,
		0.953766,
		0.909232,
		0.846361,
		0.765677,
		0.669354,
		0.561021,
		0.445578,
		0.32905,
		0.218463,
		0.121749,
		0.047667,
		0.005737,
		0.022399,
		0.082496,
		0.170178,
		0.276092,
		0.391707,
		0.509384,
		0.622473,
		0.72542,
		0.813918,
		0.88507,
		0.937598,
		0.972044,
		0.99089,
		0.998419,
		0.999959,
		0.999994,
		0.999156,
		0.993633,
		0.978079,
		0.947842,
		0.899959,
		0.833411,
		0.749038,
		0.649351,
		0.538323,
		0.42121,
		0.304401,
		0.195295,
		0.102209,
		0.034297,
		0.001486,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.023103,
		0.084966,
		0.174994,
		0.283404,
		0.401293,
		0.520717,
		0.634793,
		0.737825,
		0.825462,
		0.894886,
		0.94504,
		0.976834,
		0.993251,
		0.999102,
		0.999993,
		0.999955,
		0.99832,
		0.990351,
		0.970479,
		0.934294,
		0.879328,
		0.805178,
		0.713365,
		0.607109,
		0.491113,
		0.371384,
		0.255081,
		0.150396,
		0.06646,
		0.013263,
		0.023468,
		0.086243,
		0.177477,
		0.287164,
		0.406206,
		0.526502,
		0.641049,
		0.74408,
		0.831226,
		0.89972,
		0.948626,
		0.979061,
		0.994276,
		0.999349,
		0.999993,
		0.999904,
		0.997716,
		0.988288,
		0.965998,
		0.926602,
		0.867908,
		0.789852,
		0.694312,
		0.584885,
		0.466662,
		0.346046,
		0.230604,
		0.128948,
		0.050639,
		0.006111,
		0.023841,
		0.087547,
		0.180011,
		0.290993,
		0.411199,
		0.532363,
		0.647364,
		0.750362,
		0.836975,
		0.904491,
		0.952111,
		0.981167,
		0.995193,
		0.999536,
		0.999982,
		0.999812,
		0.996952,
		0.985898,
		0.961014,
		0.918247,
		0.85571,
		0.773688,
		0.674436,
		0.561943,
		0.441699,
		0.320521,
		0.2064,
		0.108382,
		0.036478,
		0.001584,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.024612,
		0.090239,
		0.185227,
		0.298855,
		0.421411,
		0.544295,
		0.66014,
		0.762966,
		0.848375,
		0.91379,
		0.958714,
		0.98496,
		0.996665,
		0.999705,
		0.999887,
		0.999412,
		0.994799,
		0.979961,
		0.949345,
		0.899376,
		0.82883,
		0.73875,
		0.632191,
		0.513969,
		0.390425,
		0.269249,
		0.159331,
		0.07064,
		0.014139,
		0.02501,
		0.091624,
		0.187904,
		0.302876,
		0.426612,
		0.550341,
		0.666569,
		0.769247,
		0.853978,
		0.918264,
		0.961779,
		0.986599,
		0.997181,
		0.999655,
		0.999756,
		0.999026,
		0.99331,
		0.9763,
		0.942546,
		0.888751,
		0.81406,
		0.719918,
		0.609807,
		0.488978,
		0.364225,
		0.243699,
		0.136765,
		0.053884,
		0.006522,
		0.025414,
		0.093032,
		0.190619,
		0.306942,
		0.431857,
		0.556411,
		0.672987,
		0.77547,
		0.859466,
		0.922566,
		0.96463,
		0.988014,
		0.997502,
		0.999471,
		0.999516,
		0.998441,
		0.991455,
		0.972079,
		0.935005,
		0.877246,
		0.798332,
		0.700132,
		0.586569,
		0.463351,
		0.337737,
		0.218359,
		0.115075,
		0.038856,
		0.001693,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.026239,
		0.09589,
		0.196111,
		0.315133,
		0.442357,
		0.56847,
		0.685602,
		0.787513,
		0.86984,
		0.930382,
		0.969415,
		0.989904,
		0.997321,
		0.998467,
		0.998418,
		0.996322,
		0.986257,
		0.961557,
		0.917318,
		0.851254,
		0.763735,
		0.657532,
		0.537502,
		0.410314,
		0.284216,
		0.168853,
		0.075127,
		0.015085,
		0.026654,
		0.097323,
		0.198852,
		0.319197,
		0.447525,
		0.574344,
		0.691658,
		0.793172,
		0.874547,
		0.933706,
		0.971156,
		0.990193,
		0.996642,
		0.997473,
		0.997356,
		0.994554,
		0.982663,
		0.955004,
		0.906936,
		0.836564,
		0.744711,
		0.634625,
		0.51166,
		0.382989,
		0.257379,
		0.14501,
		0.057334,
		0.006962,
		0.027066,
		0.098742,
		0.201554,
		0.323181,
		0.452559,
		0.580011,
		0.697422,
		0.798444,
		0.878774,
		0.936471,
		0.972293,
		0.98988,
		0.995414,
		0.99599,
		0.99577,
		0.992111,
		0.978195,
		0.947385,
		0.895339,
		0.820585,
		0.724419,
		0.610585,
		0.484947,
		0.355202,
		0.230652,
		0.122028,
		0.041348,
		0.001807,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.027862,
		0.101464,
		0.206694,
		0.330677,
		0.461886,
		0.590294,
		0.707552,
		0.807238,
		0.885148,
		0.939661,
		0.972083,
		0.986795,
		0.990679,
		0.990914,
		0.990301,
		0.984431,
		0.96586,
		0.928207,
		0.867848,
		0.784235,
		0.679669,
		0.558942,
		0.428982,
		0.298578,
		0.178148,
		0.079568,
		0.016032,
		0.028234,
		0.102721,
		0.209038,
		0.334039,
		0.465973,
		0.594644,
		0.711604,
		0.810403,
		0.88691,
		0.939685,
		0.970337,
		0.983637,
		0.986797,
		0.986929,
		0.985989,
		0.978745,
		0.95755,
		0.916238,
		0.8516,
		0.76359,
		0.655041,
		0.531288,
		0.399815,
		0.269972,
		0.152754,
		0.060626,
		0.007387,
		0.028577,
		0.103873,
		0.211158,
		0.337025,
		0.46951,
		0.59826,
		0.714734,
		0.812471,
		0.887425,
		0.938357,
		0.967195,
		0.979108,
		0.98162,
		0.981678,
		0.980301,
		0.971493,
		0.947489,
		0.902391,
		0.833445,
		0.741127,
		0.628819,
		0.502414,
		0.369959,
		0.241376,
		0.128242,
		0.043618,
		0.001912,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.029144,
		0.105728,
		0.21446,
		0.341453,
		0.474364,
		0.602567,
		0.717384,
		0.812377,
		0.883723,
		0.930634,
		0.955728,
		0.96498,
		0.966433,
		0.966371,
		0.963724,
		0.951225,
		0.921114,
		0.868202,
		0.790745,
		0.690328,
		0.571468,
		0.441194,
		0.308696,
		0.185047,
		0.082994,
		0.016784,
		0.029348,
		0.10636,
		0.215498,
		0.34267,
		0.475373,
		0.602874,
		0.716456,
		0.809723,
		0.878988,
		0.923719,
		0.946897,
		0.954896,
		0.955943,
		0.955795,
		0.95229,
		0.937675,
		0.904319,
		0.847468,
		0.765928,
		0.661863,
		0.540369,
		0.409045,
		0.277651,
		0.157829,
		0.062899,
		0.007692,
		0.029484,
		0.106744,
		0.216028,
		0.343066,
		0.475222,
		0.601687,
		0.713725,
		0.805002,
		0.871991,
		0.914431,
		0.935677,
		0.942488,
		0.943208,
		0.942932,
		0.938409,
		0.921517,
		0.884835,
		0.824096,
		0.73868,
		0.63134,
		0.507751,
		0.37608,
		0.246645,
		0.131646,
		0.04496,
		0.001978,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.029512,
		0.10662,
		0.215263,
		0.340931,
		0.470829,
		0.594078,
		0.701997,
		0.788449,
		0.850289,
		0.887842,
		0.905228,
		0.909906,
		0.910193,
		0.909445,
		0.902408,
		0.88057,
		0.837182,
		0.769076,
		0.676837,
		0.564404,
		0.438588,
		0.308649,
		0.185967,
		0.083785,
		0.017013,
		0.029384,
		0.106041,
		0.213831,
		0.338191,
		0.466305,
		0.58732,
		0.692627,
		0.77623,
		0.835202,
		0.870185,
		0.885676,
		0.889428,
		0.889587,
		0.888462,
		0.879941,
		0.855502,
		0.808844,
		0.737402,
		0.642378,
		0.528295,
		0.402506,
		0.274783,
		0.156993,
		0.062848,
		0.007716,
		0.029149,
		0.105077,
		0.211619,
		0.334215,
		0.460076,
		0.578414,
		0.680727,
		0.761188,
		0.817112,
		0.849464,
		0.863107,
		0.866039,
		0.866109,
		0.864494,
		0.854357,
		0.827286,
		0.777497,
		0.703054,
		0.605771,
		0.49074,
		0.365832,
		0.241294,
		0.12944,
		0.044403,
		0.001962,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.028334,
		0.101903,
		0.204691,
		0.322316,
		0.442204,
		0.553828,
		0.649019,
		0.722367,
		0.771729,
		0.798725,
		0.808892,
		0.81052,
		0.810449,
		0.807497,
		0.793834,
		0.761617,
		0.706256,
		0.627109,
		0.527188,
		0.412633,
		0.29224,
		0.177071,
		0.080172,
		0.01635,
		0.027745,
		0.099666,
		0.199929,
		0.314334,
		0.430498,
		0.538097,
		0.629185,
		0.698606,
		0.744506,
		0.768831,
		0.777408,
		0.778552,
		0.778389,
		0.774593,
		0.759087,
		0.724468,
		0.666804,
		0.586099,
		0.485925,
		0.372893,
		0.256183,
		0.147183,
		0.05921,
		0.007301,
		0.027034,
		0.096994,
		0.1943,
		0.305002,
		0.416966,
		0.520121,
		0.606781,
		0.672067,
		0.714427,
		0.736123,
		0.743232,
		0.743999,
		0.743703,
		0.738959,
		0.721612,
		0.684786,
		0.625228,
		0.543575,
		0.443925,
		0.333305,
		0.221226,
		0.119331,
		0.041135,
		0.001825,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.025256,
		0.090387,
		0.180546,
		0.282483,
		0.384738,
		0.477883,
		0.554862,
		0.611418,
		0.646618,
		0.663304,
		0.667894,
		0.668181,
		0.667436,
		0.660568,
		0.639732,
		0.599306,
		0.537419,
		0.455881,
		0.359697,
		0.256554,
		0.156413,
		0.071204,
		0.014591,
		0.024201,
		0.086499,
		0.172521,
		0.269468,
		0.366296,
		0.453969,
		0.525792,
		0.577846,
		0.609506,
		0.623865,
		0.627422,
		0.627576,
		0.626498,
		0.618508,
		0.596121,
		0.554413,
		0.492178,
		0.411755,
		0.318512,
		0.220363,
		0.127383,
		0.051521,
		0.006383,
		0.023047,
		0.082265,
		0.163827,
		0.255441,
		0.346533,
		0.428498,
		0.495029,
		0.542556,
		0.570757,
		0.582942,
		0.585621,
		0.585686,
		0.584203,
		0.575095,
		0.551352,
		0.508764,
		0.446768,
		0.368169,
		0.278635,
		0.186232,
		0.101068,
		0.035026,
		0.001561,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.020493,
		0.072946,
		0.144809,
		0.224966,
		0.303917,
		0.374015,
		0.429797,
		0.468408,
		0.490085,
		0.498456,
		0.499827,
		0.499767,
		0.497293,
		0.486112,
		0.460409,
		0.417399,
		0.35766,
		0.284745,
		0.204699,
		0.125657,
		0.057546,
		0.011854,
		0.019122,
		0.067966,
		0.134701,
		0.208867,
		0.281552,
		0.34563,
		0.39608,
		0.430401,
		0.449081,
		0.455837,
		0.456759,
		0.456627,
		0.4536,
		0.441538,
		0.415302,
		0.37278,
		0.315035,
		0.245885,
		0.171451,
		0.099785,
		0.040599,
		0.005056,
		0.017708,
		0.06285,
		0.124352,
		0.192443,
		0.25883,
		0.316924,
		0.36215,
		0.392355,
		0.408251,
		0.413595,
		0.414184,
		0.413956,
		0.410368,
		0.397574,
		0.371134,
		0.329566,
		0.274347,
		0.209488,
		0.141108,
		0.077097,
		0.026875,
		0.001204,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.014824,
		0.052454,
		0.103418,
		0.159395,
		0.213377,
		0.259871,
		0.295193,
		0.317842,
		0.328883,
		0.332002,
		0.332201,
		0.331687,
		0.32705,
		0.313356,
		0.287533,
		0.249185,
		0.200397,
		0.145338,
		0.089903,
		0.041445,
		0.008587,
		0.013387,
		0.047295,
		0.093074,
		0.143149,
		0.191158,
		0.232157,
		0.262895,
		0.282165,
		0.291157,
		0.293444,
		0.293546,
		0.292845,
		0.287772,
		0.273949,
		0.248937,
		0.212781,
		0.167757,
		0.118003,
		0.069201,
		0.02834,
		0.00355,
		0.011975,
		0.042238,
		0.082967,
		0.127325,
		0.169596,
		0.205375,
		0.231825,
		0.248009,
		0.255208,
		0.256834,
		0.256874,
		0.255969,
		0.25055,
		0.236823,
		0.212934,
		0.17929,
		0.138285,
		0.09396,
		0.051724,
		0.018148,
		0.000818,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.009287,
		0.032644,
		0.063871,
		0.097571,
		0.129275,
		0.1556,
		0.174471,
		0.185408,
		0.18976,
		0.190493,
		0.19046,
		0.189147,
		0.183376,
		0.170509,
		0.149661,
		0.121741,
		0.089175,
		0.055636,
		0.025838,
		0.005387,
		0.008036,
		0.028199,
		0.05506,
		0.08391,
		0.110866,
		0.133017,
		0.148632,
		0.15741,
		0.160683,
		0.161142,
		0.161075,
		0.159588,
		0.153832,
		0.141698,
		0.122681,
		0.097831,
		0.0695,
		0.041104,
		0.016957,
		0.002137,
		0.006863,
		0.02404,
		0.046839,
		0.071206,
		0.09381,
		0.112184,
		0.124906,
		0.131824,
		0.134224,
		0.134495,
		0.134389,
		0.132767,
		0.127154,
		0.115924,
		0.098875,
		0.077134,
		0.052927,
		0.029382,
		0.010384,
		0.000471,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.004784,
		0.016693,
		0.032382,
		0.048973,
		0.064131,
		0.076164,
		0.084171,
		0.088209,
		0.089387,
		0.089464,
		0.089266,
		0.087528,
		0.082561,
		0.073513,
		0.06058,
		0.044879,
		0.028272,
		0.013239,
		0.00278,
		0.00389,
		0.013546,
		0.026217,
		0.039541,
		0.051615,
		0.061078,
		0.067237,
		0.070213,
		0.070995,
		0.071031,
		0.070788,
		0.069081,
		0.064584,
		0.056733,
		0.045834,
		0.032928,
		0.019662,
		0.008178,
		0.001038,
		0.003097,
		0.010764,
		0.020783,
		0.031258,
		0.040669,
		0.047945,
		0.05257,
		0.054701,
		0.055198,
		0.055211,
		0.054932,
		0.053314,
		0.049356,
		0.042718,
		0.033761,
		0.023426,
		0.013129,
		0.004677,
		0.000214,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.001819,
		0.006296,
		0.012095,
		0.018083,
		0.023365,
		0.027329,
		0.029718,
		0.030705,
		0.030876,
		0.030868,
		0.030563,
		0.029267,
		0.026482,
		0.02215,
		0.016622,
		0.010587,
		0.005004,
		0.001059,
		0.001329,
		0.004591,
		0.008796,
		0.01311,
		0.016877,
		0.019659,
		0.021286,
		0.021916,
		0.022008,
		0.021994,
		0.021703,
		0.020615,
		0.018409,
		0.015095,
		0.010985,
		0.006631,
		0.002784,
		0.000356,
		0.000933,
		0.003214,
		0.006142,
		0.009125,
		0.011703,
		0.013573,
		0.014633,
		0.015015,
		0.015059,
		0.015042,
		0.014781,
		0.013913,
		0.012245,
		0.009823,
		0.006903,
		0.003911,
		0.001406,
		6.5e-005,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.00039,
		0.001339,
		0.002543,
		0.003753,
		0.004775,
		0.005488,
		0.005863,
		0.005977,
		0.005985,
		0.005965,
		0.005799,
		0.005342,
		0.004545,
		0.003462,
		0.002233,
		0.001066,
		0.000228,
		0.000225,
		0.00077,
		0.00146,
		0.002146,
		0.002719,
		0.00311,
		0.003307,
		0.003361,
		0.003364,
		0.003347,
		0.003231,
		0.002939,
		0.002452,
		0.001811,
		0.001107,
		0.00047,
		6.1e-005,
		0.000116,
		0.000396,
		0.000748,
		0.001096,
		0.001382,
		0.001573,
		0.001665,
		0.001688,
		0.001688,
		0.001676,
		0.001605,
		0.00144,
		0.001175,
		0.000838,
		0.000481,
		0.000175,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		1.7e-005,
		5.8e-005,
		0.000109,
		0.000159,
		0.000199,
		0.000224,
		0.000235,
		0.000237,
		0.000237,
		0.000233,
		0.000219,
		0.00019,
		0.000147,
		9.6e-005,
		4.7e-005,
		1e-005,
		1.2e-005,
		2.3e-005,
		3.3e-005,
		4.1e-005,
		4.6e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.4e-005,
		3.8e-005,
		2.8e-005,
		1.8e-005,
		1.8e-005,
		2.8e-005,
		3.8e-005,
		4.4e-005,
		4.8e-005,
		4.9e-005,
		4.9e-005,
		4.8e-005,
		4.6e-005,
		4.1e-005,
		3.3e-005,
		2.3e-005,
		1.2e-005,
		1e-005,
		4.7e-005,
		9.6e-005,
		0.000147,
		0.00019,
		0.000219,
		0.000233,
		0.000237,
		0.000237,
		0.000235,
		0.000224,
		0.000199,
		0.000159,
		0.000109,
		5.8e-005,
		1.7e-005,
		5.1e-005,
		0.000173,
		0.000326,
		0.000475,
		0.000596,
		0.000676,
		0.000712,
		0.000719,
		0.000719,
		0.000712,
		0.000676,
		0.000596,
		0.000475,
		0.000326,
		0.000173,
		5.1e-005,
		0.000175,
		0.000481,
		0.000838,
		0.001175,
		0.00144,
		0.001605,
		0.001676,
		0.001688,
		0.001688,
		0.001665,
		0.001573,
		0.001382,
		0.001096,
		0.000748,
		0.000396,
		0.000116,
		6.1e-005,
		0.00047,
		0.001107,
		0.001811,
		0.002452,
		0.002939,
		0.003231,
		0.003347,
		0.003364,
		0.003361,
		0.003307,
		0.00311,
		0.002719,
		0.002146,
		0.00146,
		0.00077,
		0.000225,
		0.000228,
		0.001066,
		0.002233,
		0.003462,
		0.004545,
		0.005342,
		0.005799,
		0.005965,
		0.005985,
		0.005977,
		0.005863,
		0.005488,
		0.004775,
		0.003753,
		0.002543,
		0.001339,
		0.00039,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		6.5e-005,
		0.001406,
		0.003911,
		0.006903,
		0.009823,
		0.012245,
		0.013913,
		0.014781,
		0.015042,
		0.015059,
		0.015015,
		0.014633,
		0.013573,
		0.011703,
		0.009125,
		0.006142,
		0.003214,
		0.000933,
		0.000356,
		0.002784,
		0.006631,
		0.010985,
		0.015095,
		0.018409,
		0.020615,
		0.021703,
		0.021994,
		0.022008,
		0.021916,
		0.021286,
		0.019659,
		0.016877,
		0.01311,
		0.008796,
		0.004591,
		0.001329,
		0.001059,
		0.005004,
		0.010587,
		0.016622,
		0.02215,
		0.026482,
		0.029267,
		0.030563,
		0.030868,
		0.030876,
		0.030705,
		0.029718,
		0.027329,
		0.023365,
		0.018083,
		0.012095,
		0.006296,
		0.001819,
		0.002407,
		0.008349,
		0.016081,
		0.024116,
		0.031269,
		0.036721,
		0.040099,
		0.041577,
		0.041877,
		0.041877,
		0.041577,
		0.040099,
		0.036721,
		0.031269,
		0.024116,
		0.016081,
		0.008349,
		0.002407,
		0.000214,
		0.004677,
		0.013129,
		0.023426,
		0.033761,
		0.042718,
		0.049356,
		0.053314,
		0.054932,
		0.055211,
		0.055198,
		0.054701,
		0.05257,
		0.047945,
		0.040669,
		0.031258,
		0.020783,
		0.010764,
		0.003097,
		0.001038,
		0.008178,
		0.019662,
		0.032928,
		0.045834,
		0.056733,
		0.064584,
		0.069081,
		0.070788,
		0.071031,
		0.070995,
		0.070213,
		0.067237,
		0.061078,
		0.051615,
		0.039541,
		0.026217,
		0.013546,
		0.00389,
		0.00278,
		0.013239,
		0.028272,
		0.044879,
		0.06058,
		0.073513,
		0.082561,
		0.087528,
		0.089266,
		0.089464,
		0.089387,
		0.088209,
		0.084171,
		0.076164,
		0.064131,
		0.048973,
		0.032382,
		0.016693,
		0.004784,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.000471,
		0.010384,
		0.029382,
		0.052927,
		0.077134,
		0.098875,
		0.115924,
		0.127154,
		0.132767,
		0.134389,
		0.134495,
		0.134224,
		0.131824,
		0.124906,
		0.112184,
		0.09381,
		0.071206,
		0.046839,
		0.02404,
		0.006863,
		0.002137,
		0.016957,
		0.041104,
		0.0695,
		0.097831,
		0.122681,
		0.141698,
		0.153832,
		0.159588,
		0.161075,
		0.161142,
		0.160683,
		0.15741,
		0.148632,
		0.133017,
		0.110866,
		0.08391,
		0.05506,
		0.028199,
		0.008036,
		0.005387,
		0.025838,
		0.055636,
		0.089175,
		0.121741,
		0.149661,
		0.170509,
		0.183376,
		0.189147,
		0.19046,
		0.190493,
		0.18976,
		0.185408,
		0.174471,
		0.1556,
		0.129275,
		0.097571,
		0.063871,
		0.032644,
		0.009287,
		0.010604,
		0.037339,
		0.073201,
		0.112084,
		0.148905,
		0.179781,
		0.202269,
		0.215674,
		0.22133,
		0.222444,
		0.222444,
		0.22133,
		0.215674,
		0.202269,
		0.179781,
		0.148905,
		0.112084,
		0.073201,
		0.037339,
		0.010604,
		0.000818,
		0.018148,
		0.051724,
		0.09396,
		0.138285,
		0.17929,
		0.212934,
		0.236823,
		0.25055,
		0.255969,
		0.256874,
		0.256834,
		0.255208,
		0.248009,
		0.231825,
		0.205375,
		0.169596,
		0.127325,
		0.082967,
		0.042238,
		0.011975,
		0.00355,
		0.02834,
		0.069201,
		0.118003,
		0.167757,
		0.212781,
		0.248937,
		0.273949,
		0.287772,
		0.292845,
		0.293546,
		0.293444,
		0.291157,
		0.282165,
		0.262895,
		0.232157,
		0.191158,
		0.143149,
		0.093074,
		0.047295,
		0.013387,
		0.008587,
		0.041445,
		0.089903,
		0.145338,
		0.200397,
		0.249185,
		0.287533,
		0.313356,
		0.32705,
		0.331687,
		0.332201,
		0.332002,
		0.328883,
		0.317842,
		0.295193,
		0.259871,
		0.213377,
		0.159395,
		0.103418,
		0.052454,
		0.014824,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.001204,
		0.026875,
		0.077097,
		0.141108,
		0.209488,
		0.274347,
		0.329566,
		0.371134,
		0.397574,
		0.410368,
		0.413956,
		0.414184,
		0.413595,
		0.408251,
		0.392355,
		0.36215,
		0.316924,
		0.25883,
		0.192443,
		0.124352,
		0.06285,
		0.017708,
		0.005056,
		0.040599,
		0.099785,
		0.171451,
		0.245885,
		0.315035,
		0.37278,
		0.415302,
		0.441538,
		0.4536,
		0.456627,
		0.456759,
		0.455837,
		0.449081,
		0.430401,
		0.39608,
		0.34563,
		0.281552,
		0.208867,
		0.134701,
		0.067966,
		0.019122,
		0.011854,
		0.057546,
		0.125657,
		0.204699,
		0.284745,
		0.35766,
		0.417399,
		0.460409,
		0.486112,
		0.497293,
		0.499767,
		0.499827,
		0.498456,
		0.490085,
		0.468408,
		0.429797,
		0.374015,
		0.303917,
		0.224966,
		0.144809,
		0.072946,
		0.020493,
		0.021807,
		0.077731,
		0.154555,
		0.240551,
		0.325662,
		0.401745,
		0.462908,
		0.505937,
		0.530799,
		0.540983,
		0.542935,
		0.542935,
		0.540983,
		0.530799,
		0.505937,
		0.462908,
		0.401745,
		0.325662,
		0.240551,
		0.154555,
		0.077731,
		0.021807,
		0.001561,
		0.035026,
		0.101068,
		0.186232,
		0.278635,
		0.368169,
		0.446768,
		0.508764,
		0.551352,
		0.575095,
		0.584203,
		0.585686,
		0.585621,
		0.582942,
		0.570757,
		0.542556,
		0.495029,
		0.428498,
		0.346533,
		0.255441,
		0.163827,
		0.082265,
		0.023047,
		0.006383,
		0.051521,
		0.127383,
		0.220363,
		0.318512,
		0.411755,
		0.492178,
		0.554413,
		0.596121,
		0.618508,
		0.626498,
		0.627576,
		0.627422,
		0.623865,
		0.609506,
		0.577846,
		0.525792,
		0.453969,
		0.366296,
		0.269468,
		0.172521,
		0.086499,
		0.024201,
		0.014591,
		0.071204,
		0.156413,
		0.256554,
		0.359697,
		0.455881,
		0.537419,
		0.599306,
		0.639732,
		0.660568,
		0.667436,
		0.668181,
		0.667894,
		0.663304,
		0.646618,
		0.611418,
		0.554862,
		0.477883,
		0.384738,
		0.282483,
		0.180546,
		0.090387,
		0.025256,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.001825,
		0.041135,
		0.119331,
		0.221226,
		0.333305,
		0.443925,
		0.543575,
		0.625228,
		0.684786,
		0.721612,
		0.738959,
		0.743703,
		0.743999,
		0.743232,
		0.736123,
		0.714427,
		0.672067,
		0.606781,
		0.520121,
		0.416966,
		0.305002,
		0.1943,
		0.096994,
		0.027034,
		0.007301,
		0.05921,
		0.147183,
		0.256183,
		0.372893,
		0.485925,
		0.586099,
		0.666804,
		0.724468,
		0.759087,
		0.774593,
		0.778389,
		0.778552,
		0.777408,
		0.768831,
		0.744506,
		0.698606,
		0.629185,
		0.538097,
		0.430498,
		0.314334,
		0.199929,
		0.099666,
		0.027745,
		0.01635,
		0.080172,
		0.177071,
		0.29224,
		0.412633,
		0.527188,
		0.627109,
		0.706256,
		0.761617,
		0.793834,
		0.807497,
		0.810449,
		0.81052,
		0.808892,
		0.798725,
		0.771729,
		0.722367,
		0.649019,
		0.553828,
		0.442204,
		0.322316,
		0.204691,
		0.101903,
		0.028334,
		0.028801,
		0.103703,
		0.208583,
		0.328938,
		0.45206,
		0.567267,
		0.666205,
		0.743241,
		0.795954,
		0.825636,
		0.837497,
		0.839721,
		0.839721,
		0.837497,
		0.825636,
		0.795954,
		0.743241,
		0.666205,
		0.567267,
		0.45206,
		0.328938,
		0.208583,
		0.103703,
		0.028801,
		0.001962,
		0.044403,
		0.12944,
		0.241294,
		0.365832,
		0.49074,
		0.605771,
		0.703054,
		0.777497,
		0.827286,
		0.854357,
		0.864494,
		0.866109,
		0.866039,
		0.863107,
		0.849464,
		0.817112,
		0.761188,
		0.680727,
		0.578414,
		0.460076,
		0.334215,
		0.211619,
		0.105077,
		0.029149,
		0.007716,
		0.062848,
		0.156993,
		0.274783,
		0.402506,
		0.528295,
		0.642378,
		0.737402,
		0.808844,
		0.855502,
		0.879941,
		0.888462,
		0.889587,
		0.889428,
		0.885676,
		0.870185,
		0.835202,
		0.77623,
		0.692627,
		0.58732,
		0.466305,
		0.338191,
		0.213831,
		0.106041,
		0.029384,
		0.017013,
		0.083785,
		0.185967,
		0.308649,
		0.438588,
		0.564404,
		0.676837,
		0.769076,
		0.837182,
		0.88057,
		0.902408,
		0.909445,
		0.910193,
		0.909906,
		0.905228,
		0.887842,
		0.850289,
		0.788449,
		0.701997,
		0.594078,
		0.470829,
		0.340931,
		0.215263,
		0.10662,
		0.029512,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.001978,
		0.04496,
		0.131646,
		0.246645,
		0.37608,
		0.507751,
		0.63134,
		0.73868,
		0.824096,
		0.884835,
		0.921517,
		0.938409,
		0.942932,
		0.943208,
		0.942488,
		0.935677,
		0.914431,
		0.871991,
		0.805002,
		0.713725,
		0.601687,
		0.475222,
		0.343066,
		0.216028,
		0.106744,
		0.029484,
		0.007692,
		0.062899,
		0.157829,
		0.277651,
		0.409045,
		0.540369,
		0.661863,
		0.765928,
		0.847468,
		0.904319,
		0.937675,
		0.95229,
		0.955795,
		0.955943,
		0.954896,
		0.946897,
		0.923719,
		0.878988,
		0.809723,
		0.716456,
		0.602874,
		0.475373,
		0.34267,
		0.215498,
		0.10636,
		0.029348,
		0.016784,
		0.082994,
		0.185047,
		0.308696,
		0.441194,
		0.571468,
		0.690328,
		0.790745,
		0.868202,
		0.921114,
		0.951225,
		0.963724,
		0.966371,
		0.966433,
		0.96498,
		0.955728,
		0.930634,
		0.883723,
		0.812377,
		0.717384,
		0.602567,
		0.474364,
		0.341453,
		0.21446,
		0.105728,
		0.029144,
		0.028884,
		0.104887,
		0.212988,
		0.339533,
		0.472358,
		0.600964,
		0.716734,
		0.81321,
		0.886448,
		0.935426,
		0.96241,
		0.97297,
		0.974912,
		0.974912,
		0.97297,
		0.96241,
		0.935426,
		0.886448,
		0.81321,
		0.716734,
		0.600964,
		0.472358,
		0.339533,
		0.212988,
		0.104887,
		0.028884,
		0.001912,
		0.043618,
		0.128242,
		0.241376,
		0.369959,
		0.502414,
		0.628819,
		0.741127,
		0.833445,
		0.902391,
		0.947489,
		0.971493,
		0.980301,
		0.981678,
		0.98162,
		0.979108,
		0.967195,
		0.938357,
		0.887425,
		0.812471,
		0.714734,
		0.59826,
		0.46951,
		0.337025,
		0.211158,
		0.103873,
		0.028577,
		0.007387,
		0.060626,
		0.152754,
		0.269972,
		0.399815,
		0.531288,
		0.655041,
		0.76359,
		0.8516,
		0.916238,
		0.95755,
		0.978745,
		0.985989,
		0.986929,
		0.986797,
		0.983637,
		0.970337,
		0.939685,
		0.88691,
		0.810403,
		0.711604,
		0.594644,
		0.465973,
		0.334039,
		0.209038,
		0.102721,
		0.028234,
		0.016032,
		0.079568,
		0.178148,
		0.298578,
		0.428982,
		0.558942,
		0.679669,
		0.784235,
		0.867848,
		0.928207,
		0.96586,
		0.984431,
		0.990301,
		0.990914,
		0.990679,
		0.986795,
		0.972083,
		0.939661,
		0.885148,
		0.807238,
		0.707552,
		0.590294,
		0.461886,
		0.330677,
		0.206694,
		0.101464,
		0.027862,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.001807,
		0.041348,
		0.122028,
		0.230652,
		0.355202,
		0.484947,
		0.610585,
		0.724419,
		0.820585,
		0.895339,
		0.947385,
		0.978195,
		0.992111,
		0.99577,
		0.99599,
		0.995414,
		0.98988,
		0.972293,
		0.936471,
		0.878774,
		0.798444,
		0.697422,
		0.580011,
		0.452559,
		0.323181,
		0.201554,
		0.098742,
		0.027066,
		0.006962,
		0.057334,
		0.14501,
		0.257379,
		0.382989,
		0.51166,
		0.634625,
		0.744711,
		0.836564,
		0.906936,
		0.955004,
		0.982663,
		0.994554,
		0.997356,
		0.997473,
		0.996642,
		0.990193,
		0.971156,
		0.933706,
		0.874547,
		0.793172,
		0.691658,
		0.574344,
		0.447525,
		0.319197,
		0.198852,
		0.097323,
		0.026654,
		0.015085,
		0.075127,
		0.168853,
		0.284216,
		0.410314,
		0.537502,
		0.657532,
		0.763735,
		0.851254,
		0.917318,
		0.961557,
		0.986257,
		0.996322,
		0.998418,
		0.998467,
		0.997321,
		0.989904,
		0.969415,
		0.930382,
		0.86984,
		0.787513,
		0.685602,
		0.56847,
		0.442357,
		0.315133,
		0.196111,
		0.09589,
		0.026239,
		0.025825,
		0.094456,
		0.19336,
		0.311037,
		0.437117,
		0.56247,
		0.679351,
		0.781582,
		0.864779,
		0.926633,
		0.967203,
		0.989142,
		0.997575,
		0.9991,
		0.9991,
		0.997575,
		0.989142,
		0.967203,
		0.926633,
		0.864779,
		0.781582,
		0.679351,
		0.56247,
		0.437117,
		0.311037,
		0.19336,
		0.094456,
		0.025825,
		0.001693,
		0.038856,
		0.115075,
		0.218359,
		0.337737,
		0.463351,
		0.586569,
		0.700132,
		0.798332,
		0.877246,
		0.935005,
		0.972079,
		0.991455,
		0.998441,
		0.999516,
		0.999471,
		0.997502,
		0.988014,
		0.96463,
		0.922566,
		0.859466,
		0.77547,
		0.672987,
		0.556411,
		0.431857,
		0.306942,
		0.190619,
		0.093032,
		0.025414,
		0.006522,
		0.053884,
		0.136765,
		0.243699,
		0.364225,
		0.488978,
		0.609807,
		0.719918,
		0.81406,
		0.888751,
		0.942546,
		0.9763,
		0.99331,
		0.999026,
		0.999756,
		0.999655,
		0.997181,
		0.986599,
		0.961779,
		0.918264,
		0.853978,
		0.769247,
		0.666569,
		0.550341,
		0.426612,
		0.302876,
		0.187904,
		0.091624,
		0.02501,
		0.014139,
		0.07064,
		0.159331,
		0.269249,
		0.390425,
		0.513969,
		0.632191,
		0.73875,
		0.82883,
		0.899376,
		0.949345,
		0.979961,
		0.994799,
		0.999412,
		0.999887,
		0.999705,
		0.996665,
		0.98496,
		0.958714,
		0.91379,
		0.848375,
		0.762966,
		0.66014,
		0.544295,
		0.421411,
		0.298855,
		0.185227,
		0.090239,
		0.024612,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.001584,
		0.036478,
		0.108382,
		0.2064,
		0.320521,
		0.441699,
		0.561943,
		0.674436,
		0.773688,
		0.85571,
		0.918247,
		0.961014,
		0.985898,
		0.996952,
		0.999812,
		0.999982,
		0.999536,
		0.995193,
		0.981167,
		0.952111,
		0.904491,
		0.836975,
		0.750362,
		0.647364,
		0.532363,
		0.411199,
		0.290993,
		0.180011,
		0.087547,
		0.023841,
		0.006111,
		0.050639,
		0.128948,
		0.230604,
		0.346046,
		0.466662,
		0.584885,
		0.694312,
		0.789852,
		0.867908,
		0.926602,
		0.965998,
		0.988288,
		0.997716,
		0.999904,
		0.999993,
		0.999349,
		0.994276,
		0.979061,
		0.948626,
		0.89972,
		0.831226,
		0.74408,
		0.641049,
		0.526502,
		0.406206,
		0.287164,
		0.177477,
		0.086243,
		0.023468,
		0.013263,
		0.06646,
		0.150396,
		0.255081,
		0.371384,
		0.491113,
		0.607109,
		0.713365,
		0.805178,
		0.879328,
		0.934294,
		0.970479,
		0.990351,
		0.99832,
		0.999955,
		0.999993,
		0.999102,
		0.993251,
		0.976834,
		0.94504,
		0.894886,
		0.825462,
		0.737825,
		0.634793,
		0.520717,
		0.401293,
		0.283404,
		0.174994,
		0.084966,
		0.023103,
		0.022747,
		0.083718,
		0.172561,
		0.279714,
		0.39646,
		0.515011,
		0.628601,
		0.731605,
		0.819691,
		0.890001,
		0.941362,
		0.974494,
		0.992122,
		0.998793,
		0.999982,
		0.999982,
		0.998793,
		0.992122,
		0.974494,
		0.941362,
		0.890001,
		0.819691,
		0.731605,
		0.628601,
		0.515011,
		0.39646,
		0.279714,
		0.172561,
		0.083718,
		0.022747,
		0.001486,
		0.034297,
		0.102209,
		0.195295,
		0.304401,
		0.42121,
		0.538323,
		0.649351,
		0.749038,
		0.833411,
		0.899959,
		0.947842,
		0.978079,
		0.993633,
		0.999156,
		0.999994,
		0.999959,
		0.998419,
		0.99089,
		0.972044,
		0.937598,
		0.88507,
		0.813918,
		0.72542,
		0.622473,
		0.509384,
		0.391707,
		0.276092,
		0.170178,
		0.082496,
		0.022399,
		0.005737,
		0.047667,
		0.121749,
		0.218463,
		0.32905,
		0.445578,
		0.561021,
		0.669354,
		0.765677,
		0.846361,
		0.909232,
		0.953766,
		0.981267,
		0.994912,
		0.999429,
		0.999999,
		0.99992,
		0.997978,
		0.989554,
		0.969489,
		0.933754,
		0.880097,
		0.808146,
		0.719275,
		0.61641,
		0.503836,
		0.387032,
		0.272538,
		0.167843,
		0.081301,
		0.022059,
		0.012464,
		0.062632,
		0.142168,
		0.241943,
		0.353578,
		0.469512,
		0.583079,
		0.688606,
		0.781532,
		0.858565,
		0.91785,
		0.959167,
		0.984089,
		0.995985,
		0.999629,
		1,
		0.999859,
		0.997465,
		0.988116,
		0.966832,
		0.929834,
		0.875088,
		0.802379,
		0.71317,
		0.610413,
		0.498366,
		0.382435,
		0.269049,
		0.165555,
		0.080132,
		0.021727,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.001396,
		0.032305,
		0.096544,
		0.185043,
		0.289411,
		0.40199,
		0.515919,
		0.625212,
		0.72486,
		0.810953,
		0.880827,
		0.933236,
		0.968525,
		0.988753,
		0.997607,
		0.999867,
		1,
		0.999651,
		0.996208,
		0.984931,
		0.961225,
		0.921781,
		0.864974,
		0.790869,
		0.701086,
		0.598615,
		0.487653,
		0.373465,
		0.262263,
		0.161115,
		0.077867,
		0.021084,
		0.005396,
		0.044948,
		0.11513,
		0.207232,
		0.313213,
		0.425753,
		0.538323,
		0.645261,
		0.741872,
		0.824552,
		0.890936,
		0.940063,
		0.972543,
		0.99065,
		0.9982,
		0.999929,
		0.999999,
		0.999493,
		0.995458,
		0.983187,
		0.95828,
		0.917657,
		0.859876,
		0.785132,
		0.695109,
		0.592813,
		0.48241,
		0.369091,
		0.258963,
		0.15896,
		0.07677,
		0.020774,
		0.011736,
		0.059124,
		0.134589,
		0.229766,
		0.336951,
		0.449152,
		0.560159,
		0.664622,
		0.758149,
		0.837433,
		0.900395,
		0.946351,
		0.976158,
		0.992293,
		0.998674,
		0.999966,
		0.999995,
		0.999293,
		0.994625,
		0.981342,
		0.955247,
		0.913472,
		0.854755,
		0.779408,
		0.689177,
		0.587077,
		0.47724,
		0.364787,
		0.255722,
		0.156848,
		0.075696,
		0.02047,
		0.020173,
		0.074644,
		0.154777,
		0.25254,
		0.360554,
		0.472143,
		0.581406,
		0.683291,
		0.773702,
		0.849616,
		0.909232,
		0.952128,
		0.9794,
		0.993706,
		0.999046,
		0.999986,
		0.999986,
		0.999046,
		0.993706,
		0.9794,
		0.952128,
		0.909232,
		0.849616,
		0.773702,
		0.683291,
		0.581406,
		0.472143,
		0.360554,
		0.25254,
		0.154777,
		0.074644,
		0.020173,
		0.001314,
		0.030481,
		0.091333,
		0.175563,
		0.27546,
		0.383959,
		0.494689,
		0.602046,
		0.701268,
		0.788542,
		0.861121,
		0.917472,
		0.957423,
		0.982294,
		0.994912,
		0.999333,
		0.999996,
		0.999968,
		0.998747,
		0.9927,
		0.97736,
		0.948926,
		0.90494,
		0.84446,
		0.768013,
		0.677451,
		0.575799,
		0.467118,
		0.356389,
		0.249415,
		0.152746,
		0.073614,
		0.019882,
		0.005084,
		0.042454,
		0.109032,
		0.196828,
		0.298442,
		0.407111,
		0.516759,
		0.622067,
		0.718555,
		0.802682,
		0.87197,
		0.92514,
		0.962262,
		0.984868,
		0.995933,
		0.999548,
		0.999999,
		0.999937,
		0.998392,
		0.991607,
		0.975226,
		0.945646,
		0.900598,
		0.839291,
		0.762345,
		0.671658,
		0.570256,
		0.462163,
		0.352291,
		0.246346,
		0.150754,
		0.072605,
		0.019597,
		0.011069,
		0.055902,
		0.127594,
		0.218463,
		0.321409,
		0.42996,
		0.538323,
		0.641458,
		0.735156,
		0.816138,
		0.882183,
		0.932261,
		0.966673,
		0.987147,
		0.99679,
		0.999707,
		1,
		0.999889,
		0.997978,
		0.990425,
		0.972999,
		0.94229,
		0.896212,
		0.834112,
		0.756699,
		0.665912,
		0.564777,
		0.457278,
		0.34826,
		0.243331,
		0.148801,
		0.071616,
		0.019319,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.001239,
		0.028807,
		0.086531,
		0.166781,
		0.262459,
		0.367032,
		0.474581,
		0.579853,
		0.678333,
		0.766326,
		0.841057,
		0.900792,
		0.944967,
		0.974313,
		0.990912,
		0.998086,
		0.999895,
		1,
		0.999723,
		0.996959,
		0.987794,
		0.968276,
		0.935365,
		0.887317,
		0.823732,
		0.745479,
		0.654566,
		0.554008,
		0.447714,
		0.34039,
		0.237461,
		0.145004,
		0.069699,
		0.018779,
		0.004799,
		0.040162,
		0.103402,
		0.187172,
		0.28465,
		0.389573,
		0.496282,
		0.599783,
		0.69581,
		0.780913,
		0.852555,
		0.909232,
		0.950601,
		0.977591,
		0.992444,
		0.99856,
		0.999944,
		0.999999,
		0.999596,
		0.996348,
		0.986346,
		0.965784,
		0.931803,
		0.882814,
		0.818536,
		0.739908,
		0.648967,
		0.548718,
		0.443032,
		0.336549,
		0.234603,
		0.14316,
		0.068768,
		0.018518,
		0.010458,
		0.052935,
		0.121124,
		0.207954,
		0.306869,
		0.411865,
		0.517538,
		0.61914,
		0.712649,
		0.794849,
		0.863435,
		0.917126,
		0.955789,
		0.980542,
		0.993771,
		0.998938,
		0.999973,
		0.999996,
		0.999435,
		0.995666,
		0.98481,
		0.963209,
		0.928179,
		0.87828,
		0.81334,
		0.734365,
		0.643416,
		0.54349,
		0.438416,
		0.332769,
		0.231794,
		0.141349,
		0.067857,
		0.018262,
		0.018011,
		0.066963,
		0.139573,
		0.229035,
		0.32905,
		0.433865,
		0.538323,
		0.637915,
		0.728851,
		0.808146,
		0.873716,
		0.924495,
		0.960553,
		0.983187,
		0.994912,
		0.999236,
		0.999989,
		0.999989,
		0.999236,
		0.994912,
		0.983187,
		0.960553,
		0.924495,
		0.873716,
		0.808146,
		0.728851,
		0.637915,
		0.538323,
		0.433865,
		0.32905,
		0.229035,
		0.139573,
		0.066963,
		0.018011,
		0.00117,
		0.027267,
		0.082095,
		0.158632,
		0.250328,
		0.351133,
		0.455537,
		0.558619,
		0.656102,
		0.744423,
		0.820817,
		0.883417,
		0.931363,
		0.964919,
		0.985548,
		0.995886,
		0.999465,
		0.999997,
		0.999975,
		0.998994,
		0.994084,
		0.981477,
		0.957819,
		0.920755,
		0.869125,
		0.802956,
		0.723368,
		0.632464,
		0.533218,
		0.429378,
		0.325389,
		0.226323,
		0.137829,
		0.066086,
		0.017765,
		0.004537,
		0.038051,
		0.098193,
		0.178197,
		0.271757,
		0.373065,
		0.476846,
		0.578407,
		0.673697,
		0.759371,
		0.832877,
		0.892557,
		0.937751,
		0.968908,
		0.987648,
		0.996711,
		0.999638,
		0.999999,
		0.99995,
		0.998706,
		0.993181,
		0.979683,
		0.95501,
		0.916963,
		0.864511,
		0.797771,
		0.717915,
		0.627063,
		0.528173,
		0.424954,
		0.321785,
		0.223658,
		0.136118,
		0.065227,
		0.017525,
		0.009896,
		0.050198,
		0.11513,
		0.19817,
		0.293251,
		0.394798,
		0.497762,
		0.597675,
		0.690698,
		0.773702,
		0.84434,
		0.901155,
		0.943681,
		0.972543,
		0.989507,
		0.997403,
		0.999765,
		1,
		0.999911,
		0.998368,
		0.992201,
		0.977805,
		0.952128,
		0.913121,
		0.859876,
		0.792593,
		0.712494,
		0.621711,
		0.523188,
		0.420591,
		0.318238,
		0.221038,
		0.134438,
		0.064384,
		0.017289,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.001107,
		0.025848,
		0.07799,
		0.151058,
		0.238994,
		0.336187,
		0.437501,
		0.538323,
		0.634605,
		0.722924,
		0.800552,
		0.865538,
		0.916807,
		0.954256,
		0.978837,
		0.992579,
		0.998451,
		0.999916,
		1,
		0.999777,
		0.997533,
		0.990011,
		0.973805,
		0.946156,
		0.905299,
		0.850552,
		0.782268,
		0.701753,
		0.611159,
		0.513395,
		0.412048,
		0.31131,
		0.215931,
		0.131168,
		0.062746,
		0.016832,
		0.004296,
		0.036101,
		0.093366,
		0.169842,
		0.259691,
		0.357518,
		0.4584,
		0.557927,
		0.652255,
		0.738155,
		0.813093,
		0.875306,
		0.9239,
		0.958943,
		0.981538,
		0.993829,
		0.998834,
		0.999955,
		0.999999,
		0.999675,
		0.99703,
		0.9888,
		0.971686,
		0.943071,
		0.901325,
		0.845868,
		0.777124,
		0.696434,
		0.605958,
		0.508587,
		0.407865,
		0.307926,
		0.213443,
		0.129577,
		0.061951,
		0.01661,
		0.009378,
		0.047667,
		0.109566,
		0.189047,
		0.280485,
		0.378694,
		0.478955,
		0.577059,
		0.669354,
		0.752804,
		0.82506,
		0.884543,
		0.930531,
		0.963257,
		0.983968,
		0.994912,
		0.99914,
		0.999978,
		0.999997,
		0.999544,
		0.996468,
		0.987511,
		0.969489,
		0.939924,
		0.897313,
		0.841172,
		0.771993,
		0.69115,
		0.600807,
		0.503836,
		0.403741,
		0.304595,
		0.210995,
		0.128015,
		0.06117,
		0.016392,
		0.016179,
		0.060404,
		0.12648,
		0.208589,
		0.301315,
		0.399673,
		0.499142,
		0.595706,
		0.685901,
		0.766878,
		0.836467,
		0.893264,
		0.936718,
		0.967218,
		0.986146,
		0.995843,
		0.999381,
		0.999991,
		0.999991,
		0.999381,
		0.995843,
		0.986146,
		0.967218,
		0.936718,
		0.893264,
		0.836467,
		0.766878,
		0.685901,
		0.595706,
		0.499142,
		0.399673,
		0.301315,
		0.208589,
		0.12648,
		0.060404,
		0.016179,
		0.001049,
		0.024536,
		0.074184,
		0.144007,
		0.228392,
		0.322126,
		0.420418,
		0.518938,
		0.613858,
		0.701897,
		0.780384,
		0.847326,
		0.901488,
		0.942482,
		0.970847,
		0.988091,
		0.996638,
		0.999567,
		0.999997,
		0.99998,
		0.999183,
		0.995156,
		0.984704,
		0.964874,
		0.933455,
		0.889182,
		0.831754,
		0.761779,
		0.68069,
		0.590655,
		0.494505,
		0.395662,
		0.298086,
		0.206223,
		0.124973,
		0.059653,
		0.01597,
		0.004073,
		0.034297,
		0.088884,
		0.162051,
		0.248386,
		0.342867,
		0.440894,
		0.538323,
		0.631507,
		0.717343,
		0.793332,
		0.857652,
		0.909232,
		0.947842,
		0.974162,
		0.989821,
		0.997312,
		0.999707,
		0.999999,
		0.999959,
		0.998946,
		0.994403,
		0.983187,
		0.962458,
		0.930138,
		0.88507,
		0.827036,
		0.756699,
		0.675515,
		0.585653,
		0.489923,
		0.391707,
		0.294906,
		0.203896,
		0.123492,
		0.058915,
		0.015765,
		0.0089,
		0.045322,
		0.104393,
		0.180527,
		0.268504,
		0.363491,
		0.461072,
		0.557281,
		0.648648,
		0.732241,
		0.805731,
		0.867459,
		0.916514,
		0.952816,
		0.977183,
		0.991352,
		0.997877,
		0.999809,
		1,
		0.999928,
		0.998669,
		0.993585,
		0.981594,
		0.959974,
		0.926769,
		0.880929,
		0.822315,
		0.751638,
		0.670377,
		0.580701,
		0.485397,
		0.387806,
		0.291774,
		0.201607,
		0.122037,
		0.05819,
		0.015563,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.000995,
		0.023321,
		0.070648,
		0.137432,
		0.218463,
		0.308888,
		0.404232,
		0.500431,
		0.593863,
		0.681392,
		0.760415,
		0.828924,
		0.885575,
		0.929759,
		0.961681,
		0.982413,
		0.993882,
		0.998733,
		0.999931,
		1,
		0.999819,
		0.997978,
		0.991748,
		0.978187,
		0.954807,
		0.919885,
		0.872571,
		0.812868,
		0.741577,
		0.660215,
		0.570945,
		0.476508,
		0.380165,
		0.285652,
		0.197141,
		0.119202,
		0.056781,
		0.015172,
		0.003868,
		0.032625,
		0.084714,
		0.154777,
		0.237781,
		0.32905,
		0.424278,
		0.51957,
		0.611467,
		0.696992,
		0.773702,
		0.839741,
		0.893914,
		0.935757,
		0.965609,
		0.984657,
		0.994912,
		0.999046,
		0.999963,
		1,
		0.999735,
		0.997561,
		0.990728,
		0.976376,
		0.952128,
		0.916375,
		0.868358,
		0.808146,
		0.736579,
		0.655192,
		0.56614,
		0.472143,
		0.376422,
		0.282661,
		0.194962,
		0.117821,
		0.056095,
		0.014982,
		0.008457,
		0.043145,
		0.099574,
		0.172561,
		0.25725,
		0.349131,
		0.444067,
		0.538323,
		0.628601,
		0.712079,
		0.786465,
		0.850054,
		0.901795,
		0.941362,
		0.969224,
		0.986675,
		0.995804,
		0.999296,
		0.999982,
		0.999998,
		0.999627,
		0.997092,
		0.989641,
		0.974494,
		0.949389,
		0.912824,
		0.864126,
		0.803427,
		0.731605,
		0.650207,
		0.561384,
		0.467831,
		0.372731,
		0.279714,
		0.192819,
		0.116464,
		0.055422,
		0.014796,
		0.014613,
		0.054761,
		0.11513,
		0.19071,
		0.276811,
		0.369091,
		0.463572,
		0.556677,
		0.645261,
		0.726654,
		0.798713,
		0.859876,
		0.909232,
		0.946591,
		0.972543,
		0.988485,
		0.996571,
		0.999493,
		0.999993,
		0.999993,
		0.999493,
		0.996571,
		0.988485,
		0.972543,
		0.946591,
		0.909232,
		0.859876,
		0.798713,
		0.726654,
		0.645261,
		0.556677,
		0.463572,
		0.369091,
		0.276811,
		0.19071,
		0.11513,
		0.054761,
		0.014613,
		0.000945,
		0.022194,
		0.067358,
		0.131291,
		0.209152,
		0.296414,
		0.388893,
		0.482769,
		0.574618,
		0.661443,
		0.740722,
		0.810454,
		0.869219,
		0.916242,
		0.951461,
		0.975582,
		0.990101,
		0.997227,
		0.999645,
		0.999998,
		0.999983,
		0.99933,
		0.995996,
		0.987261,
		0.970524,
		0.943738,
		0.905603,
		0.85561,
		0.794004,
		0.721729,
		0.640354,
		0.552017,
		0.459363,
		0.365499,
		0.273952,
		0.188634,
		0.113819,
		0.054111,
		0.014433,
		0.003677,
		0.031071,
		0.08083,
		0.147975,
		0.227824,
		0.31601,
		0.408505,
		0.501639,
		0.592135,
		0.677145,
		0.754286,
		0.821699,
		0.878098,
		0.92284,
		0.955987,
		0.978359,
		0.991538,
		0.997782,
		0.99976,
		1,
		0.999967,
		0.999134,
		0.995364,
		0.98597,
		0.968439,
		0.94083,
		0.901939,
		0.851331,
		0.789303,
		0.716828,
		0.635486,
		0.547405,
		0.455206,
		0.361957,
		0.271135,
		0.186592,
		0.112529,
		0.053473,
		0.014256,
		0.008046,
		0.041121,
		0.09508,
		0.165103,
		0.246666,
		0.335557,
		0.427896,
		0.520162,
		0.609221,
		0.692366,
		0.767353,
		0.832457,
		0.886525,
		0.929041,
		0.960187,
		0.980889,
		0.992809,
		0.998248,
		0.999844,
		1,
		0.999941,
		0.998903,
		0.994676,
		0.984612,
		0.966289,
		0.93787,
		0.898242,
		0.84704,
		0.784611,
		0.711954,
		0.630658,
		0.542841,
		0.451099,
		0.358463,
		0.268359,
		0.184582,
		0.111261,
		0.052846,
		0.014083,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.000899,
		0.021147,
		0.064291,
		0.12555,
		0.200411,
		0.28465,
		0.374349,
		0.465916,
		0.556109,
		0.642073,
		0.721369,
		0.792019,
		0.852555,
		0.902078,
		0.940314,
		0.967671,
		0.985268,
		0.994912,
		0.998953,
		0.999944,
		1,
		0.999851,
		0.998327,
		0.993125,
		0.981696,
		0.961803,
		0.931803,
		0.890756,
		0.838428,
		0.775256,
		0.702287,
		0.62112,
		0.533853,
		0.443032,
		0.351615,
		0.262931,
		0.180656,
		0.108788,
		0.051625,
		0.013746,
		0.003501,
		0.029626,
		0.077206,
		0.141606,
		0.218463,
		0.303696,
		0.393526,
		0.484498,
		0.573509,
		0.657831,
		0.735156,
		0.803632,
		0.861918,
		0.909232,
		0.945417,
		0.970985,
		0.987147,
		0.995768,
		0.999212,
		0.99997,
		1,
		0.999782,
		0.997978,
		0.992261,
		0.98014,
		0.959469,
		0.9287,
		0.886971,
		0.834112,
		0.770596,
		0.697495,
		0.61641,
		0.529428,
		0.439072,
		0.34826,
		0.260276,
		0.178739,
		0.107582,
		0.05103,
		0.013582,
		0.007665,
		0.039236,
		0.09088,
		0.15811,
		0.236704,
		0.32272,
		0.412516,
		0.502771,
		0.590511,
		0.673139,
		0.74847,
		0.814777,
		0.870838,
		0.91599,
		0.950184,
		0.974035,
		0.988837,
		0.996509,
		0.999419,
		0.999986,
		0.999998,
		0.999693,
		0.997585,
		0.991337,
		0.978519,
		0.957078,
		0.925553,
		0.883162,
		0.829789,
		0.76595,
		0.692731,
		0.61174,
		0.52505,
		0.435159,
		0.34495,
		0.257659,
		0.176852,
		0.106396,
		0.050445,
		0.013421,
		0.013263,
		0.049871,
		0.105229,
		0.174994,
		0.255081,
		0.341684,
		0.431293,
		0.520718,
		0.607109,
		0.687996,
		0.761317,
		0.825463,
		0.879329,
		0.922365,
		0.95463,
		0.976835,
		0.990352,
		0.997147,
		0.999581,
		0.999994,
		0.999994,
		0.999581,
		0.997147,
		0.990352,
		0.976835,
		0.95463,
		0.922365,
		0.879329,
		0.825463,
		0.761317,
		0.687996,
		0.607109,
		0.520718,
		0.431293,
		0.341684,
		0.255081,
		0.174994,
		0.105229,
		0.049871,
		0.013263,
		0.000857,
		0.020173,
		0.061429,
		0.120172,
		0.192195,
		0.273547,
		0.360554,
		0.449834,
		0.538323,
		0.623297,
		0.702404,
		0.773702,
		0.835698,
		0.887401,
		0.928371,
		0.958769,
		0.9794,
		0.991705,
		0.997692,
		0.999707,
		0.999998,
		0.999986,
		0.999445,
		0.996662,
		0.989308,
		0.975089,
		0.952128,
		0.919137,
		0.875474,
		0.821134,
		0.756699,
		0.683291,
		0.602518,
		0.51643,
		0.427474,
		0.338462,
		0.25254,
		0.173165,
		0.104082,
		0.049305,
		0.013107,
		0.003336,
		0.02828,
		0.073818,
		0.135635,
		0.209654,
		0.292057,
		0.379296,
		0.468117,
		0.555576,
		0.639069,
		0.716363,
		0.78563,
		0.845492,
		0.895068,
		0.934022,
		0.962616,
		0.981742,
		0.992908,
		0.998153,
		0.999801,
		1,
		0.999973,
		0.999281,
		0.996128,
		0.988203,
		0.973282,
		0.949573,
		0.91587,
		0.8716,
		0.816804,
		0.752097,
		0.678615,
		0.597967,
		0.512188,
		0.423701,
		0.335282,
		0.250035,
		0.171364,
		0.102952,
		0.04875,
		0.012955,
		0.00731,
		0.037478,
		0.086952,
		0.151546,
		0.227317,
		0.31057,
		0.397884,
		0.486123,
		0.572465,
		0.654422,
		0.729875,
		0.797109,
		0.854856,
		0.902341,
		0.939331,
		0.966185,
		0.983876,
		0.993974,
		0.998541,
		0.999871,
		1,
		0.999951,
		0.999088,
		0.995546,
		0.987038,
		0.971415,
		0.946968,
		0.912568,
		0.867708,
		0.812474,
		0.747512,
		0.673969,
		0.593455,
		0.50799,
		0.419973,
		0.332145,
		0.247567,
		0.16959,
		0.101841,
		0.048204,
		0.012805,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.000817,
		0.019264,
		0.058753,
		0.11513,
		0.184465,
		0.263058,
		0.347461,
		0.434489,
		0.52124,
		0.605119,
		0.683863,
		0.755573,
		0.818748,
		0.872333,
		0.915755,
		0.94898,
		0.972543,
		0.987569,
		0.995735,
		0.999128,
		0.999953,
		1,
		0.999876,
		0.998605,
		0.994227,
		0.98453,
		0.967506,
		0.94161,
		0.905864,
		0.859876,
		0.80382,
		0.738394,
		0.664769,
		0.584549,
		0.499726,
		0.41265,
		0.325995,
		0.242735,
		0.166123,
		0.099671,
		0.047138,
		0.012512,
		0.003183,
		0.027023,
		0.070648,
		0.130029,
		0.201355,
		0.281048,
		0.365774,
		0.452462,
		0.538323,
		0.620871,
		0.69795,
		0.767766,
		0.828924,
		0.880467,
		0.921922,
		0.953345,
		0.975358,
		0.989154,
		0.996452,
		0.999343,
		0.999975,
		1,
		0.999819,
		0.99831,
		0.99349,
		0.983187,
		0.965467,
		0.938862,
		0.902465,
		0.855939,
		0.799498,
		0.733863,
		0.660215,
		0.580154,
		0.495659,
		0.409054,
		0.32298,
		0.24037,
		0.164428,
		0.098612,
		0.046619,
		0.01237,
		0.006979,
		0.035835,
		0.08327,
		0.145377,
		0.218463,
		0.299063,
		0.383959,
		0.470189,
		0.555074,
		0.636233,
		0.711615,
		0.77953,
		0.838682,
		0.888213,
		0.927745,
		0.957423,
		0.977948,
		0.990579,
		0.997073,
		0.999515,
		0.999988,
		0.999998,
		0.999744,
		0.997978,
		0.9927,
		0.981785,
		0.963372,
		0.93607,
		0.899037,
		0.85199,
		0.795181,
		0.729352,
		0.655692,
		0.575799,
		0.491635,
		0.405501,
		0.320005,
		0.238038,
		0.162759,
		0.09757,
		0.046108,
		0.01223,
		0.012092,
		0.045605,
		0.096544,
		0.161115,
		0.23574,
		0.317068,
		0.40199,
		0.487653,
		0.571482,
		0.651201,
		0.72486,
		0.790869,
		0.848031,
		0.895582,
		0.933236,
		0.961225,
		0.980326,
		0.991857,
		0.997607,
		0.999651,
		0.999995,
		0.999995,
		0.999651,
		0.997607,
		0.991857,
		0.980326,
		0.961225,
		0.933236,
		0.895582,
		0.848031,
		0.790869,
		0.72486,
		0.651201,
		0.571482,
		0.487653,
		0.40199,
		0.317068,
		0.23574,
		0.161115,
		0.096544,
		0.045605,
		0.012092,
		0.00078,
		0.018415,
		0.056247,
		0.110396,
		0.177183,
		0.253142,
		0.33503,
		0.419845,
		0.504839,
		0.58754,
		0.665773,
		0.737688,
		0.801791,
		0.85698,
		0.902584,
		0.938407,
		0.964764,
		0.982503,
		0.992999,
		0.998064,
		0.999755,
		0.999998,
		0.999989,
		0.999536,
		0.997195,
		0.990961,
		0.97881,
		0.959025,
		0.930361,
		0.892101,
		0.844063,
		0.786565,
		0.72039,
		0.64674,
		0.567204,
		0.483714,
		0.398522,
		0.31417,
		0.233473,
		0.159495,
		0.095534,
		0.045111,
		0.011957,
		0.003041,
		0.025848,
		0.067676,
		0.12476,
		0.193529,
		0.270627,
		0.352917,
		0.437501,
		0.521733,
		0.603241,
		0.679948,
		0.750101,
		0.812301,
		0.865538,
		0.909232,
		0.943272,
		0.968053,
		0.984492,
		0.994013,
		0.998451,
		0.999834,
		1,
		0.999977,
		0.999399,
		0.996742,
		0.990011,
		0.977237,
		0.956775,
		0.927446,
		0.888597,
		0.840087,
		0.782268,
		0.71594,
		0.642312,
		0.562964,
		0.479816,
		0.395094,
		0.31131,
		0.231239,
		0.157899,
		0.094539,
		0.044624,
		0.011823,
		0.006671,
		0.034297,
		0.079817,
		0.139573,
		0.210104,
		0.288158,
		0.370703,
		0.45494,
		0.538323,
		0.618579,
		0.693726,
		0.762104,
		0.822408,
		0.873716,
		0.915536,
		0.947842,
		0.971103,
		0.986304,
		0.994912,
		0.998776,
		0.999892,
		1,
		0.999959,
		0.999236,
		0.996245,
		0.989007,
		0.975609,
		0.954475,
		0.924495,
		0.88507,
		0.836105,
		0.77798,
		0.711512,
		0.637915,
		0.558763,
		0.475959,
		0.391707,
		0.308487,
		0.229035,
		0.156326,
		0.09356,
		0.044146,
		0.011692,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.000745,
		0.017621,
		0.053897,
		0.105945,
		0.170317,
		0.243759,
		0.32322,
		0.405867,
		0.489096,
		0.570554,
		0.648151,
		0.720092,
		0.784898,
		0.841438,
		0.888967,
		0.927158,
		0.956143,
		0.976536,
		0.98944,
		0.996398,
		0.999268,
		0.999961,
		1,
		0.999897,
		0.998827,
		0.995118,
		0.986838,
		0.972191,
		0.949734,
		0.918486,
		0.877955,
		0.828126,
		0.769433,
		0.702724,
		0.629217,
		0.550474,
		0.468368,
		0.385053,
		0.302949,
		0.224718,
		0.15325,
		0.091646,
		0.043211,
		0.011437,
		0.002907,
		0.024747,
		0.064888,
		0.119801,
		0.186142,
		0.260756,
		0.340688,
		0.423201,
		0.505785,
		0.586177,
		0.66238,
		0.732684,
		0.795699,
		0.850378,
		0.89606,
		0.932498,
		0.959898,
		0.978942,
		0.990785,
		0.997003,
		0.999448,
		0.999979,
		1,
		0.999848,
		0.998577,
		0.994485,
		0.985674,
		0.970403,
		0.947296,
		0.915432,
		0.874369,
		0.824131,
		0.765177,
		0.698364,
		0.624916,
		0.546387,
		0.464632,
		0.381784,
		0.300233,
		0.222604,
		0.151745,
		0.090711,
		0.042755,
		0.011312,
		0.006382,
		0.032856,
		0.076573,
		0.134105,
		0.202206,
		0.277816,
		0.358077,
		0.440344,
		0.522198,
		0.601465,
		0.676234,
		0.744885,
		0.80611,
		0.858946,
		0.902812,
		0.937538,
		0.963404,
		0.981155,
		0.991996,
		0.997527,
		0.999593,
		0.99999,
		0.999999,
		0.999785,
		0.998294,
		0.993807,
		0.984457,
		0.968563,
		0.944814,
		0.912347,
		0.870768,
		0.820135,
		0.760931,
		0.694027,
		0.620647,
		0.542336,
		0.460936,
		0.378555,
		0.297552,
		0.220519,
		0.150262,
		0.089791,
		0.042305,
		0.01119,
		0.011069,
		0.041863,
		0.088884,
		0.148801,
		0.218463,
		0.294906,
		0.375363,
		0.457278,
		0.538323,
		0.61641,
		0.689715,
		0.756699,
		0.816138,
		0.867151,
		0.909232,
		0.94229,
		0.966673,
		0.983187,
		0.993081,
		0.997978,
		0.999707,
		0.999996,
		0.999996,
		0.999707,
		0.997978,
		0.993081,
		0.983187,
		0.966673,
		0.94229,
		0.909232,
		0.867151,
		0.816138,
		0.756699,
		0.689715,
		0.61641,
		0.538323,
		0.457278,
		0.375363,
		0.294906,
		0.218463,
		0.148801,
		0.088884,
		0.041863,
		0.011069,
		0.000713,
		0.016877,
		0.051692,
		0.101756,
		0.163836,
		0.234873,
		0.311993,
		0.392521,
		0.473988,
		0.554152,
		0.63101,
		0.702821,
		0.768129,
		0.82579,
		0.875,
		0.915332,
		0.946765,
		0.969716,
		0.985047,
		0.99405,
		0.998364,
		0.999794,
		0.999999,
		0.99999,
		0.99961,
		0.997627,
		0.992308,
		0.981865,
		0.964734,
		0.939726,
		0.906089,
		0.863519,
		0.812141,
		0.75248,
		0.685426,
		0.612205,
		0.534347,
		0.45366,
		0.372208,
		0.292293,
		0.216434,
		0.14736,
		0.08799,
		0.041428,
		0.010951,
		0.002783,
		0.023716,
		0.062267,
		0.11513,
		0.179163,
		0.251397,
		0.32905,
		0.409531,
		0.49046,
		0.569676,
		0.645261,
		0.715554,
		0.77918,
		0.835071,
		0.882503,
		0.92112,
		0.950973,
		0.972543,
		0.986746,
		0.994912,
		0.99869,
		0.999861,
		1,
		0.999981,
		0.999493,
		0.997238,
		0.991486,
		0.980491,
		0.962746,
		0.937122,
		0.90292,
		0.859876,
		0.808146,
		0.748275,
		0.681162,
		0.608032,
		0.530408,
		0.45008,
		0.369091,
		0.289713,
		0.214433,
		0.14594,
		0.08711,
		0.041,
		0.010834,
		0.006111,
		0.031504,
		0.073521,
		0.128949,
		0.194735,
		0.268,
		0.346048,
		0.426374,
		0.50668,
		0.584888,
		0.65916,
		0.727916,
		0.789856,
		0.843991,
		0.889668,
		0.926607,
		0.954925,
		0.975164,
		0.988294,
		0.995675,
		0.998965,
		0.999909,
		1,
		0.999966,
		0.999355,
		0.996812,
		0.990617,
		0.979066,
		0.960712,
		0.934481,
		0.899725,
		0.85622,
		0.804153,
		0.744084,
		0.676922,
		0.603892,
		0.526505,
		0.446537,
		0.366009,
		0.287165,
		0.212459,
		0.14454,
		0.086243,
		0.040578,
		0.010719,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000682,
		0.016179,
		0.049618,
		0.097809,
		0.157713,
		0.226451,
		0.301315,
		0.379775,
		0.45949,
		0.538323,
		0.614355,
		0.685901,
		0.751535,
		0.810106,
		0.860772,
		0.903024,
		0.936718,
		0.962103,
		0.979834,
		0.990973,
		0.996937,
		0.999381,
		0.999967,
		1,
		0.999913,
		0.999006,
		0.995843,
		0.988734,
		0.976067,
		0.956507,
		0.92909,
		0.893264,
		0.84888,
		0.796177,
		0.735749,
		0.668518,
		0.595706,
		0.518807,
		0.439564,
		0.359954,
		0.282166,
		0.208589,
		0.141799,
		0.084547,
		0.039753,
		0.010495,
		0.002666,
		0.022747,
		0.059802,
		0.110724,
		0.172561,
		0.242518,
		0.317968,
		0.39646,
		0.475734,
		0.553728,
		0.628601,
		0.698742,
		0.762798,
		0.819691,
		0.868649,
		0.909232,
		0.941362,
		0.965348,
		0.981901,
		0.992122,
		0.997451,
		0.999533,
		0.999982,
		1,
		0.999872,
		0.998793,
		0.995298,
		0.98772,
		0.974494,
		0.954338,
		0.926344,
		0.890001,
		0.845197,
		0.792195,
		0.731605,
		0.664354,
		0.591661,
		0.515011,
		0.436133,
		0.35698,
		0.279714,
		0.206693,
		0.140457,
		0.083718,
		0.03935,
		0.010385,
		0.005858,
		0.030233,
		0.070648,
		0.124081,
		0.187663,
		0.258679,
		0.334582,
		0.413,
		0.491749,
		0.568845,
		0.642517,
		0.71123,
		0.773702,
		0.828924,
		0.876195,
		0.91514,
		0.945745,
		0.968379,
		0.983803,
		0.993157,
		0.997897,
		0.999656,
		0.999992,
		0.999999,
		0.999819,
		0.998551,
		0.994712,
		0.986658,
		0.972873,
		0.952128,
		0.923565,
		0.886718,
		0.841508,
		0.78822,
		0.727478,
		0.660215,
		0.587648,
		0.511251,
		0.432738,
		0.354039,
		0.277292,
		0.204822,
		0.139134,
		0.0829,
		0.038954,
		0.010277,
		0.010171,
		0.038563,
		0.082095,
		0.137829,
		0.202976,
		0.2749,
		0.351133,
		0.429378,
		0.507526,
		0.583667,
		0.656102,
		0.723368,
		0.78425,
		0.837812,
		0.883417,
		0.920755,
		0.949876,
		0.971204,
		0.985548,
		0.994084,
		0.99828,
		0.999752,
		0.999997,
		0.999997,
		0.999752,
		0.99828,
		0.994084,
		0.985548,
		0.971204,
		0.949876,
		0.920755,
		0.883417,
		0.837812,
		0.78425,
		0.723368,
		0.656102,
		0.583667,
		0.507526,
		0.429378,
		0.351133,
		0.2749,
		0.202976,
		0.137829,
		0.082095,
		0.038563,
		0.010171,
		0.000654,
		0.015523,
		0.047667,
		0.094085,
		0.151921,
		0.218463,
		0.291153,
		0.367599,
		0.445578,
		0.523054,
		0.598188,
		0.669354,
		0.735156,
		0.794448,
		0.846361,
		0.890323,
		0.926089,
		0.953766,
		0.973833,
		0.987147,
		0.994912,
		0.998608,
		0.999826,
		0.999999,
		0.999992,
		0.999669,
		0.997978,
		0.993413,
		0.984391,
		0.969489,
		0.947585,
		0.917916,
		0.880097,
		0.834112,
		0.780288,
		0.719275,
		0.652015,
		0.579717,
		0.503836,
		0.426054,
		0.34826,
		0.272538,
		0.201154,
		0.136543,
		0.081301,
		0.038178,
		0.010067,
		0.002556,
		0.021837,
		0.057479,
		0.106564,
		0.166312,
		0.234088,
		0.307411,
		0.383959,
		0.461585,
		0.538323,
		0.612405,
		0.682272,
		0.746597,
		0.804301,
		0.854579,
		0.896922,
		0.931149,
		0.957423,
		0.976276,
		0.988607,
		0.995649,
		0.998886,
		0.999882,
		1,
		0.999984,
		0.99957,
		0.997644,
		0.9927,
		0.983187,
		0.967729,
		0.945255,
		0.915048,
		0.876762,
		0.830407,
		0.776333,
		0.715201,
		0.647954,
		0.575799,
		0.50018,
		0.422765,
		0.34542,
		0.270205,
		0.199356,
		0.135274,
		0.080519,
		0.037798,
		0.009964,
		0.00562,
		0.029038,
		0.067939,
		0.119481,
		0.180962,
		0.249819,
		0.323646,
		0.400195,
		0.477387,
		0.553326,
		0.626313,
		0.694856,
		0.757694,
		0.813813,
		0.862471,
		0.903223,
		0.935944,
		0.960857,
		0.978541,
		0.989936,
		0.996301,
		0.999119,
		0.999923,
		1,
		0.999971,
		0.999452,
		0.997277,
		0.991944,
		0.981935,
		0.965925,
		0.942888,
		0.912153,
		0.873411,
		0.826699,
		0.772387,
		0.711144,
		0.643919,
		0.571912,
		0.496559,
		0.41951,
		0.342613,
		0.267901,
		0.197581,
		0.134022,
		0.079747,
		0.037425,
		0.009862,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000627,
		0.014907,
		0.045828,
		0.090569,
		0.146439,
		0.21088,
		0.281477,
		0.355962,
		0.432226,
		0.508329,
		0.582507,
		0.653193,
		0.719025,
		0.778869,
		0.831837,
		0.877309,
		0.91496,
		0.944778,
		0.967092,
		0.982575,
		0.992239,
		0.99738,
		0.999473,
		0.999972,
		1,
		0.999926,
		0.999153,
		0.996439,
		0.990301,
		0.979295,
		0.962186,
		0.938045,
		0.906287,
		0.866667,
		0.819279,
		0.764523,
		0.703088,
		0.635927,
		0.564232,
		0.489418,
		0.413102,
		0.337094,
		0.263377,
		0.194101,
		0.13157,
		0.078237,
		0.036694,
		0.009664,
		0.002453,
		0.02098,
		0.055289,
		0.102633,
		0.160392,
		0.226079,
		0.297347,
		0.371999,
		0.447991,
		0.523449,
		0.596674,
		0.666161,
		0.730613,
		0.788955,
		0.840361,
		0.88427,
		0.920413,
		0.948834,
		0.969911,
		0.984361,
		0.993226,
		0.997819,
		0.999603,
		0.999985,
		1,
		0.999891,
		0.99897,
		0.995967,
		0.989415,
		0.977906,
		0.960253,
		0.935573,
		0.903317,
		0.863277,
		0.815567,
		0.760605,
		0.699089,
		0.631971,
		0.560439,
		0.485897,
		0.409949,
		0.334382,
		0.261156,
		0.192394,
		0.130369,
		0.077498,
		0.036336,
		0.009567,
		0.005396,
		0.027912,
		0.065382,
		0.11513,
		0.174607,
		0.241394,
		0.313213,
		0.387931,
		0.463572,
		0.538323,
		0.610552,
		0.678813,
		0.741872,
		0.798713,
		0.848568,
		0.890936,
		0.925601,
		0.952661,
		0.972543,
		0.986004,
		0.994115,
		0.9982,
		0.999707,
		0.999993,
		0.999999,
		0.999846,
		0.998762,
		0.995458,
		0.988485,
		0.976473,
		0.95828,
		0.933068,
		0.900326,
		0.859876,
		0.811856,
		0.756699,
		0.695109,
		0.628041,
		0.556677,
		0.48241,
		0.406828,
		0.331701,
		0.258963,
		0.19071,
		0.129184,
		0.07677,
		0.035984,
		0.009472,
		0.009378,
		0.035637,
		0.076051,
		0.128015,
		0.189047,
		0.256796,
		0.32905,
		0.403741,
		0.478955,
		0.552945,
		0.624138,
		0.69115,
		0.752804,
		0.808146,
		0.856465,
		0.897313,
		0.930531,
		0.956268,
		0.974996,
		0.987511,
		0.994912,
		0.998528,
		0.999788,
		0.999997,
		0.999997,
		0.999788,
		0.998528,
		0.994912,
		0.987511,
		0.974996,
		0.956268,
		0.930531,
		0.897313,
		0.856465,
		0.808146,
		0.752804,
		0.69115,
		0.624138,
		0.552945,
		0.478955,
		0.403741,
		0.32905,
		0.256796,
		0.189047,
		0.128015,
		0.076051,
		0.035637,
		0.009378,
		0.000602,
		0.014327,
		0.044093,
		0.087245,
		0.141244,
		0.203676,
		0.272257,
		0.344836,
		0.419412,
		0.494131,
		0.567308,
		0.637429,
		0.70317,
		0.763413,
		0.81726,
		0.864056,
		0.903409,
		0.935212,
		0.959663,
		0.977279,
		0.988892,
		0.995624,
		0.998808,
		0.999852,
		0.999999,
		0.999993,
		0.999718,
		0.998267,
		0.994328,
		0.986495,
		0.973475,
		0.954217,
		0.927963,
		0.894279,
		0.853044,
		0.804438,
		0.748921,
		0.68721,
		0.620261,
		0.549244,
		0.475533,
		0.400685,
		0.326429,
		0.254655,
		0.187405,
		0.126861,
		0.075343,
		0.035295,
		0.009285,
		0.002356,
		0.020173,
		0.053221,
		0.098913,
		0.154777,
		0.218463,
		0.28775,
		0.360554,
		0.43493,
		0.50909,
		0.581406,
		0.650423,
		0.714875,
		0.773702,
		0.82606,
		0.87135,
		0.909232,
		0.939652,
		0.962855,
		0.9794,
		0.990153,
		0.996257,
		0.999046,
		0.999899,
		1,
		0.999986,
		0.999632,
		0.997978,
		0.993706,
		0.985435,
		0.971912,
		0.952128,
		0.925367,
		0.891227,
		0.849616,
		0.800733,
		0.74505,
		0.683291,
		0.61641,
		0.545574,
		0.472143,
		0.397661,
		0.323837,
		0.25254,
		0.185784,
		0.125723,
		0.074644,
		0.034957,
		0.009194,
		0.005185,
		0.02685,
		0.062966,
		0.11101,
		0.168575,
		0.233376,
		0.303253,
		0.376183,
		0.450283,
		0.523824,
		0.595234,
		0.663118,
		0.726267,
		0.783673,
		0.834551,
		0.878352,
		0.91479,
		0.943859,
		0.965851,
		0.981366,
		0.991302,
		0.996818,
		0.999246,
		0.999934,
		1,
		0.999975,
		0.999531,
		0.99766,
		0.993045,
		0.984332,
		0.970307,
		0.950003,
		0.922742,
		0.888157,
		0.84618,
		0.797031,
		0.741192,
		0.679392,
		0.612586,
		0.541933,
		0.468785,
		0.394668,
		0.321275,
		0.25045,
		0.184184,
		0.1246,
		0.073955,
		0.034625,
		0.009104,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.000579,
		0.013779,
		0.042454,
		0.084099,
		0.136317,
		0.196828,
		0.263466,
		0.334196,
		0.407111,
		0.480445,
		0.552583,
		0.622067,
		0.687611,
		0.748116,
		0.802682,
		0.850628,
		0.89151,
		0.92514,
		0.951607,
		0.971292,
		0.984868,
		0.993291,
		0.997745,
		0.999548,
		0.999976,
		1,
		0.999937,
		0.999274,
		0.996932,
		0.991607,
		0.981999,
		0.966975,
		0.945646,
		0.917412,
		0.881966,
		0.839291,
		0.789639,
		0.733515,
		0.671658,
		0.605018,
		0.534743,
		0.462163,
		0.388776,
		0.316237,
		0.246346,
		0.181043,
		0.122398,
		0.072605,
		0.033974,
		0.008928,
		0.002265,
		0.019411,
		0.051267,
		0.095391,
		0.149448,
		0.211216,
		0.278592,
		0.349596,
		0.42238,
		0.495233,
		0.566596,
		0.635065,
		0.699408,
		0.758579,
		0.811728,
		0.858226,
		0.897679,
		0.929947,
		0.955164,
		0.973751,
		0.986419,
		0.994144,
		0.998123,
		0.99966,
		0.999987,
		1,
		0.999907,
		0.999115,
		0.996521,
		0.990828,
		0.98077,
		0.965249,
		0.943417,
		0.914709,
		0.878848,
		0.835839,
		0.785951,
		0.729698,
		0.667822,
		0.601273,
		0.531193,
		0.458899,
		0.385876,
		0.31376,
		0.24433,
		0.179502,
		0.121319,
		0.071944,
		0.033655,
		0.008842,
		0.004986,
		0.025848,
		0.060681,
		0.107106,
		0.162846,
		0.22574,
		0.29374,
		0.364925,
		0.437501,
		0.509814,
		0.580358,
		0.647781,
		0.710906,
		0.768736,
		0.820474,
		0.865538,
		0.903584,
		0.934518,
		0.958519,
		0.976047,
		0.987845,
		0.994912,
		0.998451,
		0.999749,
		0.999994,
		0.999999,
		0.999868,
		0.998935,
		0.996078,
		0.990011,
		0.979499,
		0.963484,
		0.941155,
		0.911982,
		0.875716,
		0.832383,
		0.782268,
		0.725895,
		0.664008,
		0.597555,
		0.527672,
		0.455665,
		0.383005,
		0.31131,
		0.242338,
		0.177981,
		0.120254,
		0.071291,
		0.033341,
		0.008758,
		0.008674,
		0.033031,
		0.070648,
		0.119202,
		0.176478,
		0.24037,
		0.308888,
		0.380165,
		0.452462,
		0.52418,
		0.593863,
		0.660215,
		0.722107,
		0.778592,
		0.828924,
		0.872571,
		0.909232,
		0.938862,
		0.961681,
		0.978187,
		0.989154,
		0.995601,
		0.998733,
		0.999819,
		0.999998,
		0.999998,
		0.999819,
		0.998733,
		0.995601,
		0.989154,
		0.978187,
		0.961681,
		0.938862,
		0.909232,
		0.872571,
		0.828924,
		0.778592,
		0.722107,
		0.660215,
		0.593863,
		0.52418,
		0.452462,
		0.380165,
		0.308888,
		0.24037,
		0.176478,
		0.119202,
		0.070648,
		0.033031,
		0.008674,
		0.000557,
		0.013263,
		0.040905,
		0.081119,
		0.131641,
		0.190312,
		0.255081,
		0.324015,
		0.395301,
		0.467253,
		0.538323,
		0.607109,
		0.672365,
		0.73301,
		0.78815,
		0.837084,
		0.879329,
		0.91463,
		0.942986,
		0.964657,
		0.980179,
		0.990352,
		0.996216,
		0.998974,
		0.999873,
		0.999999,
		0.999994,
		0.999758,
		0.998506,
		0.99509,
		0.988257,
		0.976835,
		0.959841,
		0.936538,
		0.90646,
		0.869413,
		0.825463,
		0.774923,
		0.718333,
		0.656444,
		0.590198,
		0.520718,
		0.449289,
		0.377353,
		0.306492,
		0.238425,
		0.174994,
		0.118164,
		0.070013,
		0.032726,
		0.008592,
		0.002179,
		0.018691,
		0.049418,
		0.092052,
		0.144385,
		0.204316,
		0.269849,
		0.339103,
		0.410318,
		0.481862,
		0.552238,
		0.620093,
		0.68423,
		0.743619,
		0.797413,
		0.84496,
		0.88582,
		0.919785,
		0.946896,
		0.967454,
		0.982029,
		0.991447,
		0.996763,
		0.999178,
		0.999914,
		1,
		0.999988,
		0.999684,
		0.998255,
		0.994545,
		0.987321,
		0.975444,
		0.957965,
		0.934185,
		0.903668,
		0.866244,
		0.822,
		0.771261,
		0.714575,
		0.652694,
		0.586559,
		0.517284,
		0.446146,
		0.37457,
		0.304124,
		0.236502,
		0.173528,
		0.11714,
		0.069387,
		0.032424,
		0.00851,
		0.004799,
		0.0249,
		0.058518,
		0.103402,
		0.1574,
		0.218463,
		0.28465,
		0.354133,
		0.425205,
		0.496282,
		0.565918,
		0.63281,
		0.69581,
		0.753936,
		0.806386,
		0.852555,
		0.892049,
		0.924704,
		0.950601,
		0.970081,
		0.983744,
		0.992444,
		0.997247,
		0.99935,
		0.999944,
		1,
		0.999979,
		0.999596,
		0.997978,
		0.993965,
		0.986346,
		0.974012,
		0.956054,
		0.931803,
		0.900855,
		0.863065,
		0.818536,
		0.767608,
		0.710833,
		0.648967,
		0.582946,
		0.51388,
		0.443032,
		0.371816,
		0.301781,
		0.234603,
		0.172081,
		0.116129,
		0.068768,
		0.032127,
		0.00843,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.000536,
		0.012775,
		0.039439,
		0.078295,
		0.127199,
		0.184108,
		0.247077,
		0.31427,
		0.383959,
		0.454537,
		0.524519,
		0.592557,
		0.657442,
		0.71812,
		0.773702,
		0.823476,
		0.866927,
		0.903749,
		0.93386,
		0.957423,
		0.974848,
		0.986798,
		0.994171,
		0.998049,
		0.999611,
		0.99998,
		1,
		0.999946,
		0.999373,
		0.997343,
		0.9927,
		0.984278,
		0.971035,
		0.952128,
		0.926957,
		0.895174,
		0.856678,
		0.811609,
		0.760326,
		0.703397,
		0.641577,
		0.575799,
		0.507156,
		0.436892,
		0.366393,
		0.297172,
		0.230869,
		0.169238,
		0.114144,
		0.067557,
		0.031545,
		0.008273,
		0.002098,
		0.018011,
		0.047667,
		0.088884,
		0.139573,
		0.197741,
		0.261496,
		0.32905,
		0.398725,
		0.468961,
		0.538323,
		0.605508,
		0.669354,
		0.728851,
		0.783156,
		0.831602,
		0.873716,
		0.909232,
		0.938112,
		0.960553,
		0.977002,
		0.98815,
		0.994912,
		0.998376,
		0.999707,
		0.999989,
		1,
		0.99992,
		0.999236,
		0.996984,
		0.992015,
		0.983187,
		0.969489,
		0.950115,
		0.924495,
		0.892307,
		0.853472,
		0.808146,
		0.756699,
		0.699703,
		0.637915,
		0.572264,
		0.503836,
		0.433865,
		0.363722,
		0.294906,
		0.229035,
		0.167843,
		0.113171,
		0.066963,
		0.031259,
		0.008197,
		0.004622,
		0.024004,
		0.056468,
		0.099885,
		0.152218,
		0.211524,
		0.27596,
		0.343786,
		0.413374,
		0.483212,
		0.551909,
		0.618209,
		0.680995,
		0.739302,
		0.79233,
		0.839455,
		0.880246,
		0.914479,
		0.942154,
		0.963506,
		0.979014,
		0.989394,
		0.995579,
		0.998659,
		0.999783,
		0.999995,
		0.999999,
		0.999886,
		0.999079,
		0.996595,
		0.991293,
		0.982057,
		0.967907,
		0.948071,
		0.922008,
		0.889423,
		0.850259,
		0.804685,
		0.753082,
		0.696026,
		0.634275,
		0.568755,
		0.500544,
		0.430867,
		0.361079,
		0.292664,
		0.227222,
		0.166465,
		0.11221,
		0.066377,
		0.030978,
		0.008121,
		0.008046,
		0.0307,
		0.065798,
		0.111261,
		0.165103,
		0.225429,
		0.290446,
		0.358463,
		0.427896,
		0.497281,
		0.565272,
		0.630658,
		0.692366,
		0.749475,
		0.801227,
		0.84704,
		0.886525,
		0.919497,
		0.945995,
		0.966289,
		0.980889,
		0.990536,
		0.996177,
		0.998903,
		0.999844,
		0.999998,
		0.999998,
		0.999844,
		0.998903,
		0.996177,
		0.990536,
		0.980889,
		0.966289,
		0.945995,
		0.919497,
		0.886525,
		0.84704,
		0.801227,
		0.749475,
		0.692366,
		0.630658,
		0.565272,
		0.497281,
		0.427896,
		0.358463,
		0.290446,
		0.225429,
		0.165103,
		0.111261,
		0.065798,
		0.0307,
		0.008046,
		0.000516,
		0.012313,
		0.038051,
		0.075614,
		0.122976,
		0.178197,
		0.239433,
		0.304937,
		0.373065,
		0.442279,
		0.51116,
		0.578407,
		0.642853,
		0.703467,
		0.759371,
		0.809849,
		0.854361,
		0.892557,
		0.924291,
		0.94964,
		0.968908,
		0.982633,
		0.991581,
		0.996711,
		0.999112,
		0.99989,
		0.999999,
		0.999995,
		0.999791,
		0.998706,
		0.995728,
		0.989742,
		0.979683,
		0.964635,
		0.943889,
		0.916963,
		0.883611,
		0.843815,
		0.797771,
		0.745878,
		0.688723,
		0.627063,
		0.561815,
		0.494044,
		0.424954,
		0.355873,
		0.288253,
		0.223658,
		0.163757,
		0.110324,
		0.065227,
		0.030427,
		0.007973,
		0.002021,
		0.017367,
		0.046007,
		0.085875,
		0.134994,
		0.191473,
		0.253513,
		0.319414,
		0.387578,
		0.456513,
		0.524843,
		0.591311,
		0.654791,
		0.714298,
		0.768993,
		0.818201,
		0.861423,
		0.898348,
		0.928869,
		0.953097,
		0.971369,
		0.984253,
		0.992536,
		0.997186,
		0.999289,
		0.999926,
		1,
		0.99999,
		0.999727,
		0.998486,
		0.995249,
		0.988913,
		0.97844,
		0.962947,
		0.941754,
		0.914407,
		0.880684,
		0.840584,
		0.794318,
		0.742293,
		0.685097,
		0.62349,
		0.558383,
		0.490835,
		0.422038,
		0.353309,
		0.286083,
		0.221906,
		0.162428,
		0.109399,
		0.064663,
		0.030156,
		0.0079,
		0.004454,
		0.023155,
		0.054523,
		0.096544,
		0.147285,
		0.204903,
		0.267647,
		0.333861,
		0.40199,
		0.470589,
		0.538323,
		0.603979,
		0.666474,
		0.72486,
		0.778343,
		0.826287,
		0.868232,
		0.903905,
		0.933236,
		0.956371,
		0.973679,
		0.985756,
		0.993406,
		0.997607,
		0.999437,
		0.999951,
		1,
		0.999982,
		0.999651,
		0.998244,
		0.994738,
		0.988048,
		0.977161,
		0.961225,
		0.93959,
		0.91183,
		0.877744,
		0.83735,
		0.790869,
		0.738718,
		0.681489,
		0.619939,
		0.554977,
		0.487653,
		0.41915,
		0.350772,
		0.283937,
		0.220175,
		0.161115,
		0.108485,
		0.064107,
		0.02989,
		0.007828,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.000497,
		0.011876,
		0.036734,
		0.073068,
		0.118957,
		0.172561,
		0.232129,
		0.295997,
		0.362598,
		0.430464,
		0.498233,
		0.564656,
		0.628601,
		0.689066,
		0.745185,
		0.796242,
		0.841678,
		0.881109,
		0.914337,
		0.941362,
		0.962398,
		0.977873,
		0.988431,
		0.994912,
		0.998304,
		0.999663,
		0.999982,
		1,
		0.999953,
		0.999457,
		0.997688,
		0.993622,
		0.98621,
		0.974494,
		0.957681,
		0.935179,
		0.906615,
		0.871829,
		0.83087,
		0.783986,
		0.731605,
		0.674326,
		0.612904,
		0.54824,
		0.48137,
		0.413455,
		0.345773,
		0.279714,
		0.21677,
		0.158534,
		0.106691,
		0.063015,
		0.029367,
		0.007688,
		0.001948,
		0.016757,
		0.044432,
		0.083016,
		0.130634,
		0.185491,
		0.245878,
		0.310176,
		0.37686,
		0.444504,
		0.511786,
		0.577498,
		0.640549,
		0.699976,
		0.754952,
		0.804797,
		0.848992,
		0.887188,
		0.919224,
		0.945135,
		0.965164,
		0.979768,
		0.989616,
		0.995558,
		0.998588,
		0.999746,
		0.999991,
		1,
		0.999931,
		0.999337,
		0.997372,
		0.993016,
		0.985237,
		0.973107,
		0.955861,
		0.932935,
		0.903979,
		0.868855,
		0.827626,
		0.780552,
		0.728066,
		0.670771,
		0.60942,
		0.54491,
		0.478268,
		0.410647,
		0.343311,
		0.277636,
		0.215097,
		0.157267,
		0.10581,
		0.06248,
		0.02911,
		0.007619,
		0.004296,
		0.02235,
		0.052677,
		0.093366,
		0.142584,
		0.198581,
		0.259691,
		0.324337,
		0.391034,
		0.4584,
		0.525152,
		0.59012,
		0.652255,
		0.710629,
		0.764456,
		0.813093,
		0.856057,
		0.893035,
		0.9239,
		0.948721,
		0.967772,
		0.981538,
		0.990705,
		0.99614,
		0.998834,
		0.999812,
		0.999996,
		0.999999,
		0.999901,
		0.9992,
		0.99703,
		0.992377,
		0.98423,
		0.971686,
		0.954009,
		0.930665,
		0.901325,
		0.865871,
		0.824381,
		0.777124,
		0.72454,
		0.667234,
		0.605958,
		0.541604,
		0.475193,
		0.407865,
		0.340874,
		0.27558,
		0.213443,
		0.156015,
		0.104941,
		0.061951,
		0.028857,
		0.007551,
		0.007484,
		0.028608,
		0.061429,
		0.104082,
		0.154777,
		0.211807,
		0.273547,
		0.338462,
		0.405109,
		0.472143,
		0.538323,
		0.602518,
		0.663715,
		0.721027,
		0.773702,
		0.821134,
		0.862877,
		0.898654,
		0.928371,
		0.952128,
		0.97023,
		0.983187,
		0.991705,
		0.996662,
		0.999046,
		0.999864,
		0.999998,
		0.999998,
		0.999864,
		0.999046,
		0.996662,
		0.991705,
		0.983187,
		0.97023,
		0.952128,
		0.928371,
		0.898654,
		0.862877,
		0.821134,
		0.773702,
		0.721027,
		0.663715,
		0.602518,
		0.538323,
		0.472143,
		0.405109,
		0.338462,
		0.273547,
		0.211807,
		0.154777,
		0.104082,
		0.061429,
		0.028608,
		0.007484,
		0.000479,
		0.011462,
		0.035484,
		0.070648,
		0.11513,
		0.167184,
		0.225145,
		0.287427,
		0.352538,
		0.419073,
		0.485726,
		0.551297,
		0.61469,
		0.674931,
		0.73117,
		0.78269,
		0.828924,
		0.869459,
		0.904051,
		0.932642,
		0.955361,
		0.972543,
		0.984721,
		0.992621,
		0.997128,
		0.999227,
		0.999905,
		0.999999,
		0.999996,
		0.999819,
		0.998873,
		0.996266,
		0.991001,
		0.982109,
		0.968741,
		0.950217,
		0.926053,
		0.895967,
		0.859876,
		0.817887,
		0.770286,
		0.717527,
		0.660215,
		0.599101,
		0.535067,
		0.469119,
		0.402378,
		0.336073,
		0.271535,
		0.210189,
		0.153554,
		0.103233,
		0.060913,
		0.028361,
		0.007418,
		0.001879,
		0.016179,
		0.042936,
		0.080297,
		0.12648,
		0.179781,
		0.238574,
		0.301315,
		0.366551,
		0.432915,
		0.499142,
		0.564067,
		0.626633,
		0.685901,
		0.741059,
		0.791425,
		0.836467,
		0.875806,
		0.909232,
		0.936718,
		0.958426,
		0.974716,
		0.986146,
		0.993458,
		0.997543,
		0.999381,
		0.999935,
		1,
		0.999991,
		0.999763,
		0.998681,
		0.995843,
		0.990264,
		0.980996,
		0.967218,
		0.948277,
		0.923712,
		0.893264,
		0.856866,
		0.814639,
		0.766878,
		0.71404,
		0.656733,
		0.595706,
		0.531836,
		0.466121,
		0.399673,
		0.333709,
		0.269544,
		0.208589,
		0.152345,
		0.102394,
		0.060404,
		0.028118,
		0.007353,
		0.004145,
		0.021587,
		0.050923,
		0.090341,
		0.138102,
		0.192542,
		0.252073,
		0.315194,
		0.380488,
		0.446628,
		0.512384,
		0.57663,
		0.638345,
		0.696627,
		0.750696,
		0.79991,
		0.843766,
		0.881923,
		0.914202,
		0.940607,
		0.961329,
		0.976757,
		0.987468,
		0.99422,
		0.99791,
		0.99951,
		0.999958,
		1,
		0.999984,
		0.999696,
		0.998468,
		0.995392,
		0.989494,
		0.97985,
		0.965663,
		0.946308,
		0.92135,
		0.890546,
		0.85385,
		0.811392,
		0.763477,
		0.710566,
		0.65327,
		0.592333,
		0.528629,
		0.463149,
		0.396993,
		0.331368,
		0.267574,
		0.207007,
		0.151149,
		0.101566,
		0.059902,
		0.027878,
		0.007289,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.000462,
		0.011069,
		0.034297,
		0.068345,
		0.111483,
		0.162051,
		0.218463,
		0.27921,
		0.342867,
		0.40809,
		0.473628,
		0.538323,
		0.601121,
		0.661072,
		0.717343,
		0.769224,
		0.816138,
		0.857652,
		0.893487,
		0.923529,
		0.947842,
		0.966673,
		0.980459,
		0.989821,
		0.995539,
		0.998518,
		0.999707,
		0.999985,
		1,
		0.999959,
		0.999527,
		0.997978,
		0.994403,
		0.987855,
		0.977455,
		0.962458,
		0.94229,
		0.916561,
		0.88507,
		0.847798,
		0.804901,
		0.756699,
		0.703661,
		0.646399,
		0.585653,
		0.522288,
		0.457278,
		0.391707,
		0.326755,
		0.263696,
		0.203896,
		0.148801,
		0.099939,
		0.058915,
		0.027406,
		0.007163,
		0.001814,
		0.01563,
		0.041515,
		0.077708,
		0.122519,
		0.174326,
		0.23158,
		0.292812,
		0.356632,
		0.421733,
		0.4869,
		0.55101,
		0.613044,
		0.672085,
		0.727336,
		0.778118,
		0.823889,
		0.864247,
		0.898944,
		0.927897,
		0.951199,
		0.969125,
		0.982133,
		0.990862,
		0.996106,
		0.998766,
		0.999779,
		0.999992,
		1,
		0.99994,
		0.999422,
		0.9977,
		0.993865,
		0.986986,
		0.976208,
		0.96081,
		0.940242,
		0.914137,
		0.882312,
		0.844764,
		0.801659,
		0.753323,
		0.70023,
		0.642991,
		0.582346,
		0.519153,
		0.454381,
		0.3891,
		0.324482,
		0.261788,
		0.202366,
		0.147646,
		0.09914,
		0.05843,
		0.027175,
		0.007101,
		0.004003,
		0.020862,
		0.049255,
		0.087461,
		0.133826,
		0.186768,
		0.244776,
		0.306416,
		0.370333,
		0.435259,
		0.500011,
		0.563503,
		0.624748,
		0.682865,
		0.737086,
		0.78677,
		0.831403,
		0.870615,
		0.904191,
		0.932076,
		0.954391,
		0.971437,
		0.983695,
		0.991821,
		0.996615,
		0.998981,
		0.999836,
		0.999996,
		0.999999,
		0.999914,
		0.999302,
		0.997398,
		0.993298,
		0.986085,
		0.974928,
		0.959131,
		0.938168,
		0.911694,
		0.879542,
		0.841725,
		0.798418,
		0.749955,
		0.696813,
		0.639602,
		0.579062,
		0.516043,
		0.451507,
		0.386518,
		0.322233,
		0.2599,
		0.200852,
		0.146505,
		0.09835,
		0.057952,
		0.026947,
		0.00704,
		0.006979,
		0.026722,
		0.057479,
		0.09757,
		0.145377,
		0.199356,
		0.258031,
		0.320005,
		0.383959,
		0.448658,
		0.512956,
		0.575799,
		0.636233,
		0.69341,
		0.746597,
		0.795181,
		0.838682,
		0.876762,
		0.909232,
		0.93607,
		0.957423,
		0.973615,
		0.985151,
		0.9927,
		0.997073,
		0.999166,
		0.999882,
		0.999998,
		0.999998,
		0.999882,
		0.999166,
		0.997073,
		0.9927,
		0.985151,
		0.973615,
		0.957423,
		0.93607,
		0.909232,
		0.876762,
		0.838682,
		0.795181,
		0.746597,
		0.69341,
		0.636233,
		0.575799,
		0.512956,
		0.448658,
		0.383959,
		0.320005,
		0.258031,
		0.199356,
		0.145377,
		0.09757,
		0.057479,
		0.026722,
		0.006979,
		0.000447,
		0.010696,
		0.033168,
		0.066153,
		0.108006,
		0.157147,
		0.212067,
		0.271328,
		0.333566,
		0.397498,
		0.461923,
		0.525728,
		0.587893,
		0.647496,
		0.703721,
		0.755868,
		0.803355,
		0.845732,
		0.88269,
		0.914074,
		0.939885,
		0.960299,
		0.975665,
		0.986506,
		0.993506,
		0.997481,
		0.999324,
		0.999917,
		1,
		0.999996,
		0.999842,
		0.999014,
		0.996723,
		0.992073,
		0.984185,
		0.972271,
		0.955686,
		0.933948,
		0.906753,
		0.87397,
		0.835636,
		0.791947,
		0.743248,
		0.690022,
		0.632882,
		0.572558,
		0.509893,
		0.445833,
		0.381424,
		0.317799,
		0.256181,
		0.197875,
		0.144262,
		0.096799,
		0.057012,
		0.026499,
		0.00692,
		0.001752,
		0.015108,
		0.040162,
		0.075242,
		0.118739,
		0.169111,
		0.224882,
		0.28465,
		0.347086,
		0.410941,
		0.475047,
		0.538323,
		0.599783,
		0.658536,
		0.7138,
		0.764902,
		0.811294,
		0.852555,
		0.888407,
		0.91872,
		0.943527,
		0.963026,
		0.977591,
		0.987764,
		0.994242,
		0.997845,
		0.999459,
		0.999944,
		1,
		0.999992,
		0.999793,
		0.998845,
		0.996348,
		0.991416,
		0.983187,
		0.970896,
		0.953921,
		0.931803,
		0.904257,
		0.871169,
		0.832587,
		0.788716,
		0.739908,
		0.686649,
		0.629549,
		0.569338,
		0.506853,
		0.443032,
		0.378911,
		0.315615,
		0.254351,
		0.196411,
		0.14316,
		0.096037,
		0.056551,
		0.026279,
		0.006861,
		0.003868,
		0.020173,
		0.047667,
		0.084714,
		0.129743,
		0.181245,
		0.237781,
		0.297982,
		0.360554,
		0.424278,
		0.488022,
		0.550737,
		0.611467,
		0.669354,
		0.723646,
		0.773702,
		0.819002,
		0.859156,
		0.893914,
		0.923176,
		0.946999,
		0.965609,
		0.9794,
		0.988931,
		0.994912,
		0.998167,
		0.999572,
		0.999963,
		1,
		0.999986,
		0.999735,
		0.998657,
		0.995947,
		0.990728,
		0.982156,
		0.969489,
		0.952128,
		0.929636,
		0.901745,
		0.868358,
		0.829535,
		0.78549,
		0.736579,
		0.683291,
		0.626236,
		0.56614,
		0.503836,
		0.440255,
		0.376422,
		0.313452,
		0.25254,
		0.194962,
		0.142069,
		0.095283,
		0.056095,
		0.026062,
		0.006803,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.000431,
		0.010342,
		0.032094,
		0.064064,
		0.104687,
		0.152459,
		0.205942,
		0.263763,
		0.324619,
		0.387283,
		0.450601,
		0.513503,
		0.575003,
		0.634207,
		0.690319,
		0.742645,
		0.790604,
		0.833736,
		0.871707,
		0.904322,
		0.931537,
		0.953459,
		0.970362,
		0.982681,
		0.991008,
		0.996073,
		0.9987,
		0.999743,
		0.999987,
		1,
		0.999965,
		0.999586,
		0.998225,
		0.995069,
		0.989263,
		0.980001,
		0.966586,
		0.948463,
		0.925236,
		0.896676,
		0.862711,
		0.823426,
		0.779051,
		0.729952,
		0.67662,
		0.619667,
		0.55981,
		0.497872,
		0.43477,
		0.371512,
		0.309189,
		0.248973,
		0.192112,
		0.139926,
		0.093803,
		0.0552,
		0.025636,
		0.006689,
		0.001693,
		0.014613,
		0.038875,
		0.072891,
		0.11513,
		0.164123,
		0.218463,
		0.276811,
		0.337898,
		0.400525,
		0.463572,
		0.525998,
		0.586849,
		0.645261,
		0.700466,
		0.751801,
		0.798713,
		0.840769,
		0.877665,
		0.909232,
		0.935451,
		0.956457,
		0.972543,
		0.984164,
		0.991928,
		0.996571,
		0.998917,
		0.999806,
		0.999993,
		1,
		0.999947,
		0.999493,
		0.997978,
		0.99459,
		0.988485,
		0.978877,
		0.96509,
		0.946591,
		0.923006,
		0.894119,
		0.859876,
		0.82037,
		0.77584,
		0.726654,
		0.673308,
		0.61641,
		0.556677,
		0.494924,
		0.432063,
		0.369091,
		0.307089,
		0.247217,
		0.19071,
		0.138872,
		0.093076,
		0.054761,
		0.025427,
		0.006633,
		0.003739,
		0.019517,
		0.046154,
		0.082095,
		0.125842,
		0.17596,
		0.231074,
		0.289878,
		0.351133,
		0.413672,
		0.476406,
		0.538323,
		0.5985,
		0.656102,
		0.71039,
		0.76073,
		0.806596,
		0.847584,
		0.883417,
		0.913952,
		0.939196,
		0.959306,
		0.974599,
		0.985548,
		0.992774,
		0.99702,
		0.999106,
		0.999857,
		0.999997,
		1,
		0.999925,
		0.999388,
		0.997711,
		0.994084,
		0.987677,
		0.977722,
		0.963565,
		0.944695,
		0.920755,
		0.891549,
		0.857034,
		0.817314,
		0.772634,
		0.723368,
		0.670011,
		0.613173,
		0.553565,
		0.491999,
		0.429378,
		0.366691,
		0.305009,
		0.245479,
		0.189322,
		0.137829,
		0.092357,
		0.054327,
		0.02522,
		0.006578,
		0.006524,
		0.025016,
		0.053897,
		0.091646,
		0.136799,
		0.18795,
		0.243759,
		0.302949,
		0.364313,
		0.426716,
		0.489096,
		0.550474,
		0.609954,
		0.66673,
		0.720092,
		0.769433,
		0.814257,
		0.854185,
		0.888967,
		0.918486,
		0.942774,
		0.962012,
		0.976536,
		0.986838,
		0.993551,
		0.997422,
		0.999268,
		0.999897,
		0.999999,
		0.999999,
		0.999897,
		0.999268,
		0.997422,
		0.993551,
		0.986838,
		0.976536,
		0.962012,
		0.942774,
		0.918486,
		0.888967,
		0.854185,
		0.814257,
		0.769433,
		0.720092,
		0.66673,
		0.609954,
		0.550474,
		0.489096,
		0.426716,
		0.364313,
		0.302949,
		0.243759,
		0.18795,
		0.136799,
		0.091646,
		0.053897,
		0.025016,
		0.006524,
		0.000417,
		0.010005,
		0.031071,
		0.062072,
		0.101517,
		0.147975,
		0.200072,
		0.256499,
		0.31601,
		0.377429,
		0.439648,
		0.501639,
		0.562448,
		0.621209,
		0.677145,
		0.729573,
		0.777914,
		0.821699,
		0.860575,
		0.894319,
		0.92284,
		0.946192,
		0.964579,
		0.978359,
		0.988039,
		0.994262,
		0.997782,
		0.999407,
		0.999927,
		1,
		0.999997,
		0.999861,
		0.999134,
		0.997111,
		0.992991,
		0.98597,
		0.975321,
		0.960431,
		0.94083,
		0.916199,
		0.886372,
		0.851331,
		0.811201,
		0.76624,
		0.716828,
		0.663465,
		0.606755,
		0.547405,
		0.486216,
		0.424076,
		0.361957,
		0.300909,
		0.242056,
		0.186592,
		0.135779,
		0.090944,
		0.053473,
		0.024814,
		0.00647,
		0.001638,
		0.014141,
		0.037648,
		0.070648,
		0.111681,
		0.159349,
		0.212308,
		0.26928,
		0.32905,
		0.390469,
		0.452462,
		0.514027,
		0.57424,
		0.632263,
		0.687346,
		0.738833,
		0.786174,
		0.828924,
		0.866759,
		0.899478,
		0.927017,
		0.949453,
		0.967013,
		0.980072,
		0.989154,
		0.994912,
		0.998102,
		0.999525,
		0.999951,
		1,
		0.999993,
		0.999819,
		0.998984,
		0.996778,
		0.992403,
		0.985072,
		0.974075,
		0.958822,
		0.938862,
		0.913893,
		0.883765,
		0.848471,
		0.808146,
		0.763052,
		0.713576,
		0.660215,
		0.603574,
		0.544357,
		0.483358,
		0.421459,
		0.359622,
		0.298888,
		0.24037,
		0.185248,
		0.134771,
		0.090249,
		0.053054,
		0.024615,
		0.006417,
		0.003617,
		0.018893,
		0.044712,
		0.079594,
		0.122113,
		0.170898,
		0.224639,
		0.282087,
		0.342055,
		0.403426,
		0.465151,
		0.526257,
		0.585848,
		0.643115,
		0.697333,
		0.747875,
		0.794215,
		0.835936,
		0.872739,
		0.904447,
		0.931022,
		0.952563,
		0.969317,
		0.98168,
		0.990188,
		0.995504,
		0.998385,
		0.999624,
		0.999968,
		1,
		0.999988,
		0.999767,
		0.998818,
		0.996422,
		0.991788,
		0.984144,
		0.972801,
		0.957188,
		0.936872,
		0.911571,
		0.881147,
		0.845607,
		0.805092,
		0.759872,
		0.710335,
		0.656981,
		0.600413,
		0.54133,
		0.480522,
		0.418863,
		0.357308,
		0.296887,
		0.238701,
		0.183919,
		0.133774,
		0.089563,
		0.05264,
		0.024418,
		0.006364,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.000403,
		0.009684,
		0.030097,
		0.060171,
		0.098489,
		0.143683,
		0.194445,
		0.249522,
		0.307723,
		0.36792,
		0.429052,
		0.490126,
		0.550223,
		0.608503,
		0.664208,
		0.716668,
		0.765307,
		0.809651,
		0.849334,
		0.884104,
		0.913836,
		0.938536,
		0.958348,
		0.973558,
		0.984597,
		0.992029,
		0.996529,
		0.998854,
		0.999775,
		0.999988,
		1,
		0.999969,
		0.999636,
		0.998435,
		0.995638,
		0.990474,
		0.9822,
		0.970166,
		0.95384,
		0.932827,
		0.906878,
		0.875881,
		0.839866,
		0.79899,
		0.753533,
		0.70389,
		0.650561,
		0.594146,
		0.535338,
		0.474914,
		0.413737,
		0.352743,
		0.292943,
		0.235414,
		0.181302,
		0.131812,
		0.088212,
		0.051826,
		0.024031,
		0.006261,
		0.001584,
		0.013691,
		0.036479,
		0.068506,
		0.108384,
		0.154777,
		0.206404,
		0.262041,
		0.320527,
		0.38076,
		0.441707,
		0.502402,
		0.561953,
		0.619546,
		0.674449,
		0.726017,
		0.773702,
		0.817052,
		0.855726,
		0.889497,
		0.918264,
		0.942053,
		0.961031,
		0.975504,
		0.985916,
		0.992844,
		0.99697,
		0.999046,
		0.99983,
		0.999994,
		1,
		0.999954,
		0.999554,
		0.998216,
		0.995211,
		0.989775,
		0.981184,
		0.968806,
		0.952128,
		0.930774,
		0.904508,
		0.873234,
		0.83699,
		0.795942,
		0.750376,
		0.700686,
		0.647376,
		0.591042,
		0.532373,
		0.472143,
		0.411206,
		0.350491,
		0.290998,
		0.233795,
		0.180014,
		0.130847,
		0.087549,
		0.051425,
		0.023841,
		0.006211,
		0.003501,
		0.018298,
		0.043337,
		0.077206,
		0.118545,
		0.166048,
		0.218463,
		0.274594,
		0.333306,
		0.393526,
		0.454246,
		0.514529,
		0.573509,
		0.630396,
		0.684484,
		0.735156,
		0.781884,
		0.824244,
		0.861918,
		0.894703,
		0.922519,
		0.945417,
		0.963582,
		0.977338,
		0.987147,
		0.993593,
		0.997366,
		0.999212,
		0.999874,
		0.999997,
		1,
		0.999934,
		0.999461,
		0.997978,
		0.994759,
		0.989048,
		0.98014,
		0.967419,
		0.950392,
		0.9287,
		0.902123,
		0.870578,
		0.834112,
		0.792898,
		0.747226,
		0.697495,
		0.644206,
		0.587956,
		0.529428,
		0.469393,
		0.408697,
		0.34826,
		0.289073,
		0.232193,
		0.178739,
		0.129893,
		0.086892,
		0.05103,
		0.023653,
		0.006161,
		0.006111,
		0.023468,
		0.050639,
		0.086243,
		0.128949,
		0.177478,
		0.230606,
		0.287165,
		0.346048,
		0.406208,
		0.466665,
		0.526505,
		0.584888,
		0.641052,
		0.694316,
		0.744084,
		0.789856,
		0.83123,
		0.867913,
		0.899725,
		0.926607,
		0.948631,
		0.966004,
		0.979066,
		0.988294,
		0.994282,
		0.997721,
		0.999355,
		0.999909,
		0.999999,
		0.999999,
		0.999909,
		0.999355,
		0.997721,
		0.994282,
		0.988294,
		0.979066,
		0.966004,
		0.948631,
		0.926607,
		0.899725,
		0.867913,
		0.83123,
		0.789856,
		0.744084,
		0.694316,
		0.641052,
		0.584888,
		0.526505,
		0.466665,
		0.406208,
		0.346048,
		0.287165,
		0.230606,
		0.177478,
		0.128949,
		0.086243,
		0.050639,
		0.023468,
		0.006111,
		0.00039,
		0.009378,
		0.029167,
		0.058356,
		0.095593,
		0.139573,
		0.189047,
		0.242817,
		0.299743,
		0.358744,
		0.4188,
		0.478955,
		0.538323,
		0.596089,
		0.651515,
		0.703943,
		0.752804,
		0.79762,
		0.838015,
		0.873716,
		0.904566,
		0.930531,
		0.951701,
		0.968302,
		0.980692,
		0.989361,
		0.994912,
		0.998039,
		0.999477,
		0.999936,
		1,
		0.999997,
		0.999878,
		0.999236,
		0.997444,
		0.99378,
		0.987511,
		0.977965,
		0.964563,
		0.946848,
		0.924495,
		0.897313,
		0.865241,
		0.828347,
		0.786818,
		0.740951,
		0.69115,
		0.637915,
		0.58184,
		0.523601,
		0.463957,
		0.403741,
		0.343856,
		0.285276,
		0.229035,
		0.17623,
		0.128015,
		0.085601,
		0.050253,
		0.023285,
		0.006063,
		0.001534,
		0.013263,
		0.035363,
		0.06646,
		0.105229,
		0.150396,
		0.200737,
		0.255081,
		0.312315,
		0.371384,
		0.431293,
		0.491113,
		0.549982,
		0.607109,
		0.661782,
		0.713366,
		0.761317,
		0.805179,
		0.8446,
		0.879329,
		0.909232,
		0.934295,
		0.95463,
		0.97048,
		0.98222,
		0.990352,
		0.995488,
		0.998321,
		0.999581,
		0.999957,
		1,
		0.999994,
		0.99984,
		0.999103,
		0.997147,
		0.993253,
		0.986701,
		0.976835,
		0.963095,
		0.945041,
		0.922365,
		0.894888,
		0.862562,
		0.825463,
		0.783784,
		0.737827,
		0.687996,
		0.634794,
		0.57881,
		0.520718,
		0.46127,
		0.401293,
		0.341684,
		0.283404,
		0.227479,
		0.174994,
		0.127091,
		0.084966,
		0.049871,
		0.023103,
		0.006014,
		0.00339,
		0.017731,
		0.042023,
		0.074922,
		0.11513,
		0.161399,
		0.212531,
		0.267385,
		0.32487,
		0.383959,
		0.443682,
		0.503133,
		0.561477,
		0.617947,
		0.671853,
		0.722587,
		0.769624,
		0.812535,
		0.850988,
		0.884756,
		0.913726,
		0.937904,
		0.957423,
		0.972543,
		0.983655,
		0.991272,
		0.996013,
		0.998572,
		0.999668,
		0.999972,
		1,
		0.999989,
		0.999795,
		0.998956,
		0.996829,
		0.9927,
		0.985864,
		0.975678,
		0.961602,
		0.943213,
		0.920217,
		0.89245,
		0.859876,
		0.822577,
		0.780754,
		0.734711,
		0.684856,
		0.631689,
		0.575799,
		0.517854,
		0.458604,
		0.398867,
		0.339531,
		0.28155,
		0.225939,
		0.173771,
		0.126177,
		0.084339,
		0.049493,
		0.022924,
		0.005967,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.000378,
		0.009086,
		0.02828,
		0.056622,
		0.092821,
		0.135635,
		0.183866,
		0.236371,
		0.292057,
		0.349886,
		0.40888,
		0.468117,
		0.526742,
		0.583966,
		0.639069,
		0.691409,
		0.740422,
		0.78563,
		0.826649,
		0.863189,
		0.895068,
		0.922214,
		0.944673,
		0.962616,
		0.976338,
		0.986257,
		0.992908,
		0.996922,
		0.998987,
		0.999801,
		0.99999,
		1,
		0.999973,
		0.999679,
		0.998614,
		0.996128,
		0.991519,
		0.984106,
		0.973282,
		0.95854,
		0.939491,
		0.91587,
		0.887541,
		0.854485,
		0.816804,
		0.774707,
		0.728508,
		0.678615,
		0.625529,
		0.569832,
		0.512188,
		0.453333,
		0.394074,
		0.335282,
		0.277895,
		0.222905,
		0.171364,
		0.124377,
		0.083104,
		0.04875,
		0.022572,
		0.005873,
		0.001486,
		0.012854,
		0.034297,
		0.064503,
		0.102209,
		0.146196,
		0.195295,
		0.248386,
		0.304401,
		0.362328,
		0.42121,
		0.480153,
		0.538323,
		0.594955,
		0.649351,
		0.700893,
		0.749038,
		0.793332,
		0.833411,
		0.869007,
		0.899959,
		0.926215,
		0.947842,
		0.965026,
		0.978079,
		0.987433,
		0.993633,
		0.997312,
		0.999156,
		0.99985,
		0.999995,
		1,
		0.999959,
		0.999606,
		0.99842,
		0.995745,
		0.99089,
		0.983187,
		0.972044,
		0.956972,
		0.937598,
		0.913673,
		0.88507,
		0.851782,
		0.813918,
		0.771692,
		0.72542,
		0.675515,
		0.622473,
		0.566876,
		0.509384,
		0.450728,
		0.391707,
		0.333186,
		0.276092,
		0.221409,
		0.170178,
		0.123492,
		0.082496,
		0.048385,
		0.022399,
		0.005827,
		0.003284,
		0.01719,
		0.040769,
		0.072738,
		0.111859,
		0.156939,
		0.206833,
		0.260446,
		0.316735,
		0.374712,
		0.433445,
		0.492061,
		0.54975,
		0.60577,
		0.659446,
		0.710181,
		0.757456,
		0.800837,
		0.839981,
		0.874642,
		0.90468,
		0.930061,
		0.950871,
		0.967315,
		0.97972,
		0.988531,
		0.9943,
		0.997663,
		0.999303,
		0.999889,
		0.999997,
		1,
		0.999942,
		0.999523,
		0.998208,
		0.99534,
		0.990235,
		0.98224,
		0.97078,
		0.955381,
		0.935686,
		0.91146,
		0.882588,
		0.849074,
		0.811031,
		0.768681,
		0.722343,
		0.672427,
		0.619433,
		0.563939,
		0.5066,
		0.448143,
		0.38936,
		0.331109,
		0.274307,
		0.219929,
		0.169005,
		0.122616,
		0.081895,
		0.048024,
		0.022228,
		0.005782,
		0.005737,
		0.022059,
		0.047667,
		0.081301,
		0.121749,
		0.167843,
		0.218463,
		0.272538,
		0.32905,
		0.387032,
		0.445578,
		0.503836,
		0.561021,
		0.61641,
		0.669354,
		0.719275,
		0.765677,
		0.808146,
		0.846361,
		0.880097,
		0.909232,
		0.933754,
		0.953766,
		0.969489,
		0.981267,
		0.989554,
		0.994912,
		0.997978,
		0.999429,
		0.99992,
		0.999999,
		0.999999,
		0.99992,
		0.999429,
		0.997978,
		0.994912,
		0.989554,
		0.981267,
		0.969489,
		0.953766,
		0.933754,
		0.909232,
		0.880097,
		0.846361,
		0.808146,
		0.765677,
		0.719275,
		0.669354,
		0.61641,
		0.561021,
		0.503836,
		0.445578,
		0.387032,
		0.32905,
		0.272538,
		0.218463,
		0.167843,
		0.121749,
		0.081301,
		0.047667,
		0.022059,
		0.005737,
		0.000366,
		0.008808,
		0.027432,
		0.054963,
		0.090168,
		0.131859,
		0.178891,
		0.230171,
		0.28465,
		0.341334,
		0.39928,
		0.457602,
		0.515474,
		0.572131,
		0.626874,
		0.679074,
		0.728175,
		0.773702,
		0.815262,
		0.852555,
		0.885375,
		0.913621,
		0.937299,
		0.95653,
		0.971552,
		0.982721,
		0.990506,
		0.995473,
		0.99826,
		0.999537,
		0.999944,
		1,
		0.999997,
		0.999892,
		0.999323,
		0.997731,
		0.994461,
		0.988848,
		0.980266,
		0.968173,
		0.952128,
		0.931803,
		0.90699,
		0.877597,
		0.843645,
		0.805262,
		0.762678,
		0.716218,
		0.666294,
		0.613403,
		0.55812,
		0.501091,
		0.443032,
		0.384724,
		0.327009,
		0.270786,
		0.217011,
		0.166693,
		0.120891,
		0.080713,
		0.047314,
		0.021892,
		0.005692,
		0.00144,
		0.012464,
		0.033279,
		0.062632,
		0.099316,
		0.142168,
		0.190067,
		0.241943,
		0.29677,
		0.353578,
		0.411446,
		0.469512,
		0.526971,
		0.583079,
		0.63716,
		0.688606,
		0.736881,
		0.781532,
		0.822187,
		0.858565,
		0.89048,
		0.91785,
		0.940699,
		0.959167,
		0.973507,
		0.984089,
		0.99139,
		0.995985,
		0.998511,
		0.999629,
		0.999962,
		1,
		0.999995,
		0.999859,
		0.999205,
		0.997465,
		0.993987,
		0.988116,
		0.97924,
		0.966832,
		0.950468,
		0.929834,
		0.904734,
		0.875088,
		0.840925,
		0.802379,
		0.759685,
		0.71317,
		0.663248,
		0.610413,
		0.555238,
		0.498366,
		0.440506,
		0.382435,
		0.324985,
		0.269049,
		0.215573,
		0.165555,
		0.120042,
		0.080132,
		0.046964,
		0.021727,
		0.005649,
		0.003183,
		0.016673,
		0.039569,
		0.070648,
		0.108724,
		0.152659,
		0.201355,
		0.253765,
		0.308888,
		0.365774,
		0.423526,
		0.481304,
		0.538323,
		0.593863,
		0.647268,
		0.69795,
		0.745396,
		0.789171,
		0.828924,
		0.864394,
		0.895415,
		0.921922,
		0.943959,
		0.961681,
		0.975358,
		0.985372,
		0.99221,
		0.996452,
		0.998733,
		0.999707,
		0.999975,
		1,
		0.999991,
		0.999819,
		0.999074,
		0.99718,
		0.99349,
		0.987358,
		0.978187,
		0.965467,
		0.948785,
		0.927846,
		0.902465,
		0.872571,
		0.838201,
		0.799498,
		0.756699,
		0.710133,
		0.660215,
		0.607439,
		0.552374,
		0.495659,
		0.438,
		0.380165,
		0.32298,
		0.267329,
		0.214149,
		0.164428,
		0.119202,
		0.079556,
		0.046619,
		0.021564,
		0.005605,
		1,
		0.999985,
		0.999771,
		0.998929,
		0.996876,
		0.992969,
		0.986574,
		0.977109,
		0.964076,
		0.947082,
		0.925841,
		0.900183,
		0.870045,
		0.835475,
		0.796619,
		0.753719,
		0.707107,
		0.657196,
		0.604481,
		0.549528,
		0.492972,
		0.435513,
		0.377913,
		0.320992,
		0.265625,
		0.212739,
		0.163312,
		0.118371,
		0.078987,
		0.046277,
		0.021402,
		0.005562,
		0.999985,
		1,
		0.999984,
		0.999767,
		0.998913,
		0.99683,
		0.99287,
		0.986389,
		0.976801,
		0.963605,
		0.946406,
		0.924919,
		0.898975,
		0.868517,
		0.833597,
		0.794366,
		0.751077,
		0.704067,
		0.653763,
		0.600669,
		0.545361,
		0.488488,
		0.430764,
		0.372963,
		0.315921,
		0.260527,
		0.207724,
		0.158507,
		0.113919,
		0.075048,
		0.043031,
		0.019046,
		0.004312,
		0.999771,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998897,
		0.996784,
		0.992769,
		0.986201,
		0.976488,
		0.963126,
		0.945719,
		0.923982,
		0.897749,
		0.866967,
		0.831692,
		0.792083,
		0.7484,
		0.700991,
		0.650291,
		0.596815,
		0.541152,
		0.483964,
		0.425977,
		0.36798,
		0.310823,
		0.255412,
		0.202704,
		0.153712,
		0.109496,
		0.071162,
		0.039864,
		0.0168,
		0.003208,
		0.998929,
		0.999767,
		0.999984,
		1,
		0.999984,
		0.99976,
		0.99888,
		0.996737,
		0.992666,
		0.986009,
		0.976169,
		0.96264,
		0.945021,
		0.923032,
		0.896506,
		0.865395,
		0.82976,
		0.78977,
		0.745689,
		0.697877,
		0.646778,
		0.592919,
		0.536902,
		0.479399,
		0.421152,
		0.362964,
		0.3057,
		0.25028,
		0.197681,
		0.14893,
		0.105104,
		0.06733,
		0.03678,
		0.014668,
		0.002256,
		0.996876,
		0.998913,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.999757,
		0.998863,
		0.99669,
		0.992561,
		0.985814,
		0.975845,
		0.962145,
		0.944313,
		0.922066,
		0.895243,
		0.863799,
		0.827802,
		0.787425,
		0.742944,
		0.694725,
		0.643226,
		0.588982,
		0.532609,
		0.474793,
		0.416289,
		0.357915,
		0.300551,
		0.245133,
		0.192655,
		0.144161,
		0.100747,
		0.063557,
		0.033782,
		0.012657,
		0.001462,
		0.992969,
		0.99683,
		0.998897,
		0.99976,
		0.999984,
		1,
		0.999983,
		0.999753,
		0.998846,
		0.996641,
		0.992454,
		0.985615,
		0.975516,
		0.961642,
		0.943593,
		0.921086,
		0.893962,
		0.862181,
		0.825816,
		0.785049,
		0.740163,
		0.691535,
		0.639632,
		0.585003,
		0.528274,
		0.470147,
		0.411389,
		0.352834,
		0.295377,
		0.239972,
		0.187629,
		0.139409,
		0.096427,
		0.059846,
		0.030875,
		0.010771,
		0.000833,
		0.986574,
		0.99287,
		0.996784,
		0.99888,
		0.999757,
		0.999983,
		1,
		0.999983,
		0.999749,
		0.998829,
		0.996591,
		0.992346,
		0.985413,
		0.975181,
		0.961131,
		0.942861,
		0.92009,
		0.892661,
		0.860539,
		0.823803,
		0.782641,
		0.737347,
		0.688306,
		0.635997,
		0.58098,
		0.523896,
		0.465459,
		0.40645,
		0.34772,
		0.290179,
		0.234798,
		0.182603,
		0.134676,
		0.092147,
		0.0562,
		0.028064,
		0.009016,
		0.000375,
		0.977109,
		0.986389,
		0.992769,
		0.996737,
		0.998863,
		0.999753,
		0.999983,
		1,
		0.999983,
		0.999745,
		0.998811,
		0.996541,
		0.992235,
		0.985208,
		0.97484,
		0.960612,
		0.942118,
		0.919079,
		0.891341,
		0.858873,
		0.821761,
		0.780201,
		0.734494,
		0.685037,
		0.63232,
		0.576915,
		0.519475,
		0.460729,
		0.401474,
		0.342574,
		0.284958,
		0.229612,
		0.17758,
		0.129963,
		0.08791,
		0.052623,
		0.025353,
		0.007397,
		9.5e-005,
		0.964076,
		0.976801,
		0.986201,
		0.992666,
		0.99669,
		0.998846,
		0.999749,
		0.999983,
		1,
		0.999982,
		0.999741,
		0.998793,
		0.99649,
		0.992122,
		0.984999,
		0.974494,
		0.960083,
		0.941362,
		0.918052,
		0.890001,
		0.857183,
		0.819691,
		0.777728,
		0.731605,
		0.681729,
		0.628601,
		0.572806,
		0.515011,
		0.455958,
		0.39646,
		0.337397,
		0.279714,
		0.224414,
		0.172561,
		0.125272,
		0.083718,
		0.049119,
		0.022747,
		0.00592,
		0.947082,
		0.963605,
		0.976488,
		0.986009,
		0.992561,
		0.996641,
		0.998829,
		0.999745,
		0.999982,
		1,
		0.999982,
		0.999737,
		0.998774,
		0.996437,
		0.992008,
		0.984786,
		0.974141,
		0.959546,
		0.940595,
		0.917009,
		0.888641,
		0.855469,
		0.817592,
		0.775222,
		0.728678,
		0.67838,
		0.624839,
		0.568653,
		0.510503,
		0.451146,
		0.391409,
		0.332189,
		0.274448,
		0.219207,
		0.167548,
		0.120607,
		0.079575,
		0.045692,
		0.020251,
		0.004591,
		0.925841,
		0.946406,
		0.963126,
		0.976169,
		0.985814,
		0.992454,
		0.996591,
		0.998811,
		0.999741,
		0.999982,
		1,
		0.999982,
		0.999733,
		0.998756,
		0.996384,
		0.991891,
		0.984569,
		0.973783,
		0.959001,
		0.939815,
		0.91595,
		0.88726,
		0.853729,
		0.815463,
		0.772681,
		0.725714,
		0.67499,
		0.621034,
		0.564456,
		0.505952,
		0.446292,
		0.38632,
		0.32695,
		0.269161,
		0.213992,
		0.162543,
		0.11597,
		0.075484,
		0.042346,
		0.01787,
		0.003417,
		0.900183,
		0.924919,
		0.945719,
		0.96264,
		0.975845,
		0.985615,
		0.992346,
		0.996541,
		0.998793,
		0.999737,
		0.999982,
		1,
		0.999982,
		0.999729,
		0.998737,
		0.996329,
		0.991772,
		0.984349,
		0.973418,
		0.958445,
		0.939022,
		0.914874,
		0.885858,
		0.851964,
		0.813304,
		0.770107,
		0.722712,
		0.671559,
		0.617185,
		0.560215,
		0.501356,
		0.441396,
		0.381194,
		0.321682,
		0.263854,
		0.208769,
		0.157547,
		0.111363,
		0.071448,
		0.039085,
		0.015609,
		0.002404,
		0.870045,
		0.898975,
		0.923982,
		0.945021,
		0.962145,
		0.975516,
		0.985413,
		0.992235,
		0.99649,
		0.998774,
		0.999733,
		0.999982,
		1,
		0.999981,
		0.999725,
		0.998717,
		0.996274,
		0.991651,
		0.984125,
		0.973047,
		0.957881,
		0.938217,
		0.913781,
		0.884434,
		0.850172,
		0.811114,
		0.767497,
		0.71967,
		0.668086,
		0.613292,
		0.555928,
		0.496716,
		0.436458,
		0.376031,
		0.316383,
		0.258527,
		0.203541,
		0.152563,
		0.106789,
		0.067471,
		0.035914,
		0.013475,
		0.001559,
		0.835475,
		0.868517,
		0.897749,
		0.923032,
		0.944313,
		0.961642,
		0.975181,
		0.985208,
		0.992122,
		0.996437,
		0.998756,
		0.999729,
		0.999981,
		1,
		0.999981,
		0.99972,
		0.998697,
		0.996218,
		0.991528,
		0.983896,
		0.97267,
		0.957307,
		0.937398,
		0.91267,
		0.882989,
		0.848355,
		0.808893,
		0.764852,
		0.71659,
		0.66457,
		0.609355,
		0.551596,
		0.492031,
		0.431479,
		0.370832,
		0.311056,
		0.253183,
		0.198309,
		0.147593,
		0.102252,
		0.063557,
		0.032838,
		0.011471,
		0.000888,
		0.796619,
		0.833597,
		0.866967,
		0.896506,
		0.922066,
		0.943593,
		0.961131,
		0.97484,
		0.984999,
		0.992008,
		0.996384,
		0.998737,
		0.999725,
		0.999981,
		1,
		0.999981,
		0.999716,
		0.998677,
		0.99616,
		0.991403,
		0.983664,
		0.972286,
		0.956724,
		0.936566,
		0.911542,
		0.881522,
		0.84651,
		0.806641,
		0.762171,
		0.713469,
		0.661012,
		0.605372,
		0.547218,
		0.487302,
		0.426457,
		0.365596,
		0.3057,
		0.247821,
		0.193074,
		0.142639,
		0.097753,
		0.05971,
		0.02986,
		0.009606,
		0.0004,
		0.753719,
		0.794366,
		0.831692,
		0.865395,
		0.895243,
		0.921086,
		0.942861,
		0.960612,
		0.974494,
		0.984786,
		0.991891,
		0.996329,
		0.998717,
		0.99972,
		0.999981,
		1,
		0.99998,
		0.999711,
		0.998656,
		0.996102,
		0.991275,
		0.983428,
		0.971895,
		0.95613,
		0.93572,
		0.910396,
		0.880032,
		0.844638,
		0.804357,
		0.759454,
		0.710309,
		0.65741,
		0.601344,
		0.542794,
		0.482527,
		0.421394,
		0.360324,
		0.300316,
		0.242442,
		0.187838,
		0.137703,
		0.093296,
		0.055933,
		0.026987,
		0.007884,
		0.000101,
		0.707107,
		0.751077,
		0.792083,
		0.82976,
		0.863799,
		0.893962,
		0.92009,
		0.942118,
		0.960083,
		0.974141,
		0.984569,
		0.991772,
		0.996274,
		0.998697,
		0.999716,
		0.99998,
		1,
		0.99998,
		0.999707,
		0.998635,
		0.996042,
		0.991144,
		0.983187,
		0.971498,
		0.955526,
		0.93486,
		0.909232,
		0.878519,
		0.842738,
		0.80204,
		0.756699,
		0.707107,
		0.653763,
		0.59727,
		0.538323,
		0.477707,
		0.416289,
		0.355015,
		0.294906,
		0.237049,
		0.182603,
		0.132788,
		0.088884,
		0.05223,
		0.024223,
		0.006313,
		0.657196,
		0.704067,
		0.7484,
		0.78977,
		0.827802,
		0.862181,
		0.892661,
		0.919079,
		0.941362,
		0.959546,
		0.973783,
		0.984349,
		0.991651,
		0.996218,
		0.998677,
		0.999711,
		0.99998,
		1,
		0.99998,
		0.999702,
		0.998614,
		0.995981,
		0.991012,
		0.982941,
		0.971093,
		0.954912,
		0.933986,
		0.908049,
		0.876983,
		0.84081,
		0.79969,
		0.753907,
		0.703863,
		0.650073,
		0.59315,
		0.533806,
		0.472842,
		0.411143,
		0.349672,
		0.289469,
		0.231643,
		0.177371,
		0.127896,
		0.08452,
		0.048607,
		0.021574,
		0.004898,
		0.604481,
		0.653763,
		0.700991,
		0.745689,
		0.787425,
		0.825816,
		0.860539,
		0.891341,
		0.918052,
		0.940595,
		0.959001,
		0.973418,
		0.984125,
		0.991528,
		0.99616,
		0.998656,
		0.999707,
		0.99998,
		1,
		0.999979,
		0.999697,
		0.998592,
		0.995919,
		0.990877,
		0.982692,
		0.970681,
		0.954288,
		0.933098,
		0.906848,
		0.875423,
		0.838854,
		0.797307,
		0.751077,
		0.700578,
		0.646337,
		0.588982,
		0.529241,
		0.467931,
		0.405954,
		0.344293,
		0.284006,
		0.226223,
		0.172143,
		0.12303,
		0.080209,
		0.045066,
		0.019046,
		0.003647,
		0.549528,
		0.600669,
		0.650291,
		0.697877,
		0.742944,
		0.785049,
		0.823803,
		0.858873,
		0.890001,
		0.917009,
		0.939815,
		0.958445,
		0.973047,
		0.983896,
		0.991403,
		0.996102,
		0.998635,
		0.999702,
		0.999979,
		1,
		0.999979,
		0.999692,
		0.998569,
		0.995856,
		0.990739,
		0.982438,
		0.970262,
		0.953653,
		0.932195,
		0.905627,
		0.873839,
		0.836868,
		0.794889,
		0.748208,
		0.69725,
		0.642555,
		0.584767,
		0.524629,
		0.462975,
		0.400724,
		0.33888,
		0.278519,
		0.220793,
		0.166922,
		0.118192,
		0.075953,
		0.041614,
		0.016644,
		0.002567,
		0.492972,
		0.545361,
		0.596815,
		0.646778,
		0.694725,
		0.740163,
		0.782641,
		0.821761,
		0.857183,
		0.888641,
		0.91595,
		0.939022,
		0.957881,
		0.97267,
		0.983664,
		0.991275,
		0.996042,
		0.998614,
		0.999697,
		0.999979,
		1,
		0.999979,
		0.999687,
		0.998546,
		0.995792,
		0.990599,
		0.982179,
		0.969836,
		0.953007,
		0.931276,
		0.904386,
		0.87223,
		0.834852,
		0.792437,
		0.745299,
		0.693878,
		0.638727,
		0.580504,
		0.519969,
		0.457972,
		0.395453,
		0.333432,
		0.273008,
		0.215353,
		0.16171,
		0.113386,
		0.071756,
		0.038254,
		0.014374,
		0.001665,
		0.435513,
		0.488488,
		0.541152,
		0.592919,
		0.643226,
		0.691535,
		0.737347,
		0.780201,
		0.819691,
		0.855469,
		0.88726,
		0.914874,
		0.938217,
		0.957307,
		0.972286,
		0.983428,
		0.991144,
		0.995981,
		0.998592,
		0.999692,
		0.999979,
		1,
		0.999978,
		0.999682,
		0.998523,
		0.995726,
		0.990456,
		0.981915,
		0.969402,
		0.95235,
		0.930343,
		0.903125,
		0.870596,
		0.832806,
		0.789949,
		0.742351,
		0.690463,
		0.634852,
		0.576193,
		0.51526,
		0.452924,
		0.39014,
		0.327951,
		0.267474,
		0.209905,
		0.156508,
		0.108615,
		0.067623,
		0.034993,
		0.012242,
		0.000949,
		0.377913,
		0.430764,
		0.483964,
		0.536902,
		0.588982,
		0.639632,
		0.688306,
		0.734494,
		0.777728,
		0.817592,
		0.853729,
		0.885858,
		0.913781,
		0.937398,
		0.956724,
		0.971895,
		0.983187,
		0.991012,
		0.995919,
		0.998569,
		0.999687,
		0.999978,
		1,
		0.999978,
		0.999677,
		0.998499,
		0.995659,
		0.99031,
		0.981647,
		0.96896,
		0.951681,
		0.929393,
		0.901844,
		0.868936,
		0.830729,
		0.787425,
		0.739362,
		0.687003,
		0.63093,
		0.571833,
		0.510503,
		0.447829,
		0.384786,
		0.322436,
		0.261919,
		0.204451,
		0.151319,
		0.103881,
		0.063557,
		0.031834,
		0.010256,
		0.000428,
		0.320992,
		0.372963,
		0.425977,
		0.479399,
		0.532609,
		0.585003,
		0.635997,
		0.685037,
		0.731605,
		0.775222,
		0.815463,
		0.851964,
		0.884434,
		0.91267,
		0.936566,
		0.95613,
		0.971498,
		0.982941,
		0.990877,
		0.995856,
		0.998546,
		0.999682,
		0.999978,
		1,
		0.999978,
		0.999671,
		0.998475,
		0.995591,
		0.990162,
		0.981373,
		0.96851,
		0.951001,
		0.928428,
		0.900541,
		0.867251,
		0.828621,
		0.784865,
		0.736332,
		0.683499,
		0.62696,
		0.567424,
		0.505698,
		0.442688,
		0.379391,
		0.316889,
		0.256343,
		0.198992,
		0.146146,
		0.099188,
		0.059563,
		0.028784,
		0.008421,
		0.000108,
		0.265625,
		0.315921,
		0.36798,
		0.421152,
		0.474793,
		0.528274,
		0.58098,
		0.63232,
		0.681729,
		0.728678,
		0.772681,
		0.813304,
		0.850172,
		0.882989,
		0.911542,
		0.93572,
		0.955526,
		0.971093,
		0.982692,
		0.990739,
		0.995792,
		0.998523,
		0.999677,
		0.999978,
		1,
		0.999977,
		0.999666,
		0.998451,
		0.995521,
		0.990011,
		0.981094,
		0.968053,
		0.950308,
		0.927446,
		0.899218,
		0.865538,
		0.826481,
		0.782268,
		0.73326,
		0.679948,
		0.622942,
		0.562964,
		0.500843,
		0.437501,
		0.373956,
		0.31131,
		0.250747,
		0.193529,
		0.140991,
		0.094539,
		0.055645,
		0.025848,
		0.006746,
		0.212739,
		0.260527,
		0.310823,
		0.362964,
		0.416289,
		0.470147,
		0.523896,
		0.576915,
		0.628601,
		0.67838,
		0.725714,
		0.770107,
		0.811114,
		0.848355,
		0.881522,
		0.910396,
		0.93486,
		0.954912,
		0.970681,
		0.982438,
		0.990599,
		0.995726,
		0.998499,
		0.999671,
		0.999977,
		1,
		0.999977,
		0.99966,
		0.998425,
		0.99545,
		0.989857,
		0.980811,
		0.967586,
		0.949604,
		0.926448,
		0.897873,
		0.863799,
		0.824309,
		0.779633,
		0.730146,
		0.676351,
		0.618874,
		0.558455,
		0.495938,
		0.432268,
		0.36848,
		0.3057,
		0.245133,
		0.188066,
		0.135857,
		0.089938,
		0.051808,
		0.023032,
		0.005236,
		0.163312,
		0.207724,
		0.255412,
		0.3057,
		0.357915,
		0.411389,
		0.465459,
		0.519475,
		0.572806,
		0.624839,
		0.67499,
		0.722712,
		0.767497,
		0.808893,
		0.84651,
		0.880032,
		0.909232,
		0.933986,
		0.954288,
		0.970262,
		0.982179,
		0.990456,
		0.995659,
		0.998475,
		0.999666,
		0.999977,
		1,
		0.999976,
		0.999655,
		0.9984,
		0.995378,
		0.9897,
		0.980522,
		0.967112,
		0.948887,
		0.925433,
		0.896506,
		0.862033,
		0.822103,
		0.77696,
		0.726989,
		0.672708,
		0.614757,
		0.553895,
		0.490984,
		0.426988,
		0.362964,
		0.300059,
		0.239503,
		0.182603,
		0.130747,
		0.085389,
		0.048056,
		0.020342,
		0.003901,
		0.118371,
		0.158507,
		0.202704,
		0.25028,
		0.300551,
		0.352834,
		0.40645,
		0.460729,
		0.515011,
		0.568653,
		0.621034,
		0.671559,
		0.71967,
		0.764852,
		0.806641,
		0.844638,
		0.878519,
		0.908049,
		0.933098,
		0.953653,
		0.969836,
		0.981915,
		0.99031,
		0.995591,
		0.998451,
		0.99966,
		0.999976,
		1,
		0.999976,
		0.999649,
		0.998373,
		0.995304,
		0.98954,
		0.980227,
		0.966629,
		0.948158,
		0.9244,
		0.895116,
		0.860238,
		0.819865,
		0.774249,
		0.723788,
		0.669016,
		0.61059,
		0.549284,
		0.48598,
		0.421662,
		0.357409,
		0.294389,
		0.233856,
		0.177144,
		0.125662,
		0.080895,
		0.044395,
		0.017784,
		0.002747,
		0.078987,
		0.113919,
		0.153712,
		0.197681,
		0.245133,
		0.295377,
		0.34772,
		0.401474,
		0.455958,
		0.510503,
		0.564456,
		0.617185,
		0.668086,
		0.71659,
		0.762171,
		0.804357,
		0.842738,
		0.876983,
		0.906848,
		0.932195,
		0.953007,
		0.969402,
		0.981647,
		0.990162,
		0.995521,
		0.998425,
		0.999655,
		0.999976,
		1,
		0.999976,
		0.999643,
		0.998347,
		0.995229,
		0.989376,
		0.979927,
		0.966137,
		0.947416,
		0.92335,
		0.893703,
		0.858415,
		0.817592,
		0.771497,
		0.720543,
		0.665277,
		0.606372,
		0.544621,
		0.480925,
		0.416289,
		0.351814,
		0.28869,
		0.228195,
		0.171689,
		0.120607,
		0.076461,
		0.04083,
		0.015366,
		0.001783,
		0.046277,
		0.075048,
		0.109496,
		0.14893,
		0.192655,
		0.239972,
		0.290179,
		0.342574,
		0.39646,
		0.451146,
		0.505952,
		0.560215,
		0.613292,
		0.66457,
		0.713469,
		0.759454,
		0.80204,
		0.84081,
		0.875423,
		0.905627,
		0.931276,
		0.95235,
		0.96896,
		0.981373,
		0.990011,
		0.99545,
		0.9984,
		0.999649,
		0.999976,
		1,
		0.999975,
		0.999637,
		0.998319,
		0.995152,
		0.98921,
		0.979622,
		0.965636,
		0.946661,
		0.922282,
		0.892267,
		0.856563,
		0.815284,
		0.768706,
		0.717253,
		0.661489,
		0.602103,
		0.539907,
		0.475821,
		0.410871,
		0.34618,
		0.282963,
		0.222522,
		0.166242,
		0.115585,
		0.07209,
		0.037366,
		0.013093,
		0.001017,
		0.021402,
		0.043031,
		0.071162,
		0.105104,
		0.144161,
		0.187629,
		0.234798,
		0.284958,
		0.337397,
		0.391409,
		0.446292,
		0.501356,
		0.555928,
		0.609355,
		0.661012,
		0.710309,
		0.756699,
		0.79969,
		0.838854,
		0.873839,
		0.904386,
		0.930343,
		0.951681,
		0.96851,
		0.981094,
		0.989857,
		0.995378,
		0.998373,
		0.999643,
		0.999975,
		1,
		0.999975,
		0.999631,
		0.998291,
		0.995074,
		0.98904,
		0.97931,
		0.965125,
		0.945892,
		0.921195,
		0.890808,
		0.854681,
		0.812941,
		0.765874,
		0.713918,
		0.657651,
		0.597782,
		0.535139,
		0.470665,
		0.405406,
		0.340507,
		0.277209,
		0.216838,
		0.160804,
		0.110599,
		0.067787,
		0.034009,
		0.010974,
		0.000458,
		0.005562,
		0.019046,
		0.039864,
		0.06733,
		0.100747,
		0.139409,
		0.182603,
		0.229612,
		0.279714,
		0.332189,
		0.38632,
		0.441396,
		0.496716,
		0.551596,
		0.605372,
		0.65741,
		0.707107,
		0.753907,
		0.797307,
		0.836868,
		0.87223,
		0.903125,
		0.929393,
		0.951001,
		0.968053,
		0.980811,
		0.9897,
		0.995304,
		0.998347,
		0.999637,
		0.999975,
		1,
		0.999974,
		0.999624,
		0.998263,
		0.994994,
		0.988867,
		0.978993,
		0.964606,
		0.945109,
		0.92009,
		0.889324,
		0.852769,
		0.810562,
		0.763,
		0.710536,
		0.653763,
		0.593409,
		0.53032,
		0.465459,
		0.399895,
		0.334797,
		0.271429,
		0.211144,
		0.155379,
		0.105651,
		0.063557,
		0.030765,
		0.009016,
		0.000116,
		0.004312,
		0.0168,
		0.03678,
		0.063557,
		0.096427,
		0.134676,
		0.17758,
		0.224414,
		0.274448,
		0.32695,
		0.381194,
		0.436458,
		0.492031,
		0.547218,
		0.601344,
		0.653763,
		0.703863,
		0.751077,
		0.794889,
		0.834852,
		0.870596,
		0.901844,
		0.928428,
		0.950308,
		0.967586,
		0.980522,
		0.98954,
		0.995229,
		0.998319,
		0.999631,
		0.999974,
		1,
		0.999974,
		0.999618,
		0.998234,
		0.994912,
		0.988691,
		0.978669,
		0.964076,
		0.944313,
		0.918966,
		0.887815,
		0.850827,
		0.808146,
		0.760084,
		0.707107,
		0.649825,
		0.588982,
		0.525446,
		0.460201,
		0.394338,
		0.32905,
		0.265625,
		0.205443,
		0.149968,
		0.100747,
		0.059405,
		0.027641,
		0.007225,
		0.003208,
		0.014668,
		0.033782,
		0.059846,
		0.092147,
		0.129963,
		0.172561,
		0.219207,
		0.269161,
		0.321682,
		0.376031,
		0.431479,
		0.487302,
		0.542794,
		0.59727,
		0.650073,
		0.700578,
		0.748208,
		0.792437,
		0.832806,
		0.868936,
		0.900541,
		0.927446,
		0.949604,
		0.967112,
		0.980227,
		0.989376,
		0.995152,
		0.998291,
		0.999624,
		0.999974,
		1,
		0.999973,
		0.999611,
		0.998204,
		0.994829,
		0.988511,
		0.978339,
		0.963537,
		0.943502,
		0.917822,
		0.886281,
		0.848853,
		0.805693,
		0.757125,
		0.70363,
		0.645835,
		0.584502,
		0.52052,
		0.454892,
		0.388735,
		0.323265,
		0.259797,
		0.199736,
		0.144575,
		0.09589,
		0.055335,
		0.024641,
		0.005611,
		0.002256,
		0.012657,
		0.030875,
		0.0562,
		0.08791,
		0.125272,
		0.167548,
		0.213992,
		0.263854,
		0.316383,
		0.370832,
		0.426457,
		0.482527,
		0.538323,
		0.59315,
		0.646337,
		0.69725,
		0.745299,
		0.789949,
		0.830729,
		0.867251,
		0.899218,
		0.926448,
		0.948887,
		0.966629,
		0.979927,
		0.98921,
		0.995074,
		0.998263,
		0.999618,
		0.999973,
		1,
		0.999973,
		0.999605,
		0.998174,
		0.994744,
		0.988327,
		0.978003,
		0.962988,
		0.942676,
		0.916658,
		0.884721,
		0.846847,
		0.803203,
		0.754123,
		0.700105,
		0.641793,
		0.579968,
		0.515539,
		0.449532,
		0.383087,
		0.317445,
		0.253947,
		0.194026,
		0.139203,
		0.091084,
		0.051353,
		0.021774,
		0.004182,
		0.001462,
		0.010771,
		0.028064,
		0.052623,
		0.083718,
		0.120607,
		0.162543,
		0.208769,
		0.258527,
		0.311056,
		0.365596,
		0.421394,
		0.477707,
		0.533806,
		0.588982,
		0.642555,
		0.693878,
		0.742351,
		0.787425,
		0.828621,
		0.865538,
		0.897873,
		0.925433,
		0.948158,
		0.966137,
		0.979622,
		0.98904,
		0.994994,
		0.998234,
		0.999611,
		0.999973,
		1,
		0.999972,
		0.999598,
		0.998143,
		0.994657,
		0.98814,
		0.977661,
		0.962429,
		0.941836,
		0.915473,
		0.883135,
		0.844809,
		0.800673,
		0.751077,
		0.696531,
		0.637698,
		0.575379,
		0.510503,
		0.444121,
		0.377394,
		0.31159,
		0.248076,
		0.188314,
		0.133854,
		0.086332,
		0.047464,
		0.019046,
		0.002946,
		0.000833,
		0.009016,
		0.025353,
		0.049119,
		0.079575,
		0.11597,
		0.157547,
		0.203541,
		0.253183,
		0.3057,
		0.360324,
		0.416289,
		0.472842,
		0.529241,
		0.584767,
		0.638727,
		0.690463,
		0.739362,
		0.784865,
		0.826481,
		0.863799,
		0.896506,
		0.9244,
		0.947416,
		0.965636,
		0.97931,
		0.988867,
		0.994912,
		0.998204,
		0.999605,
		0.999972,
		1,
		0.999972,
		0.999591,
		0.998111,
		0.994569,
		0.987949,
		0.977312,
		0.961859,
		0.94098,
		0.914268,
		0.881522,
		0.842738,
		0.798105,
		0.747985,
		0.692907,
		0.63355,
		0.570735,
		0.505413,
		0.438658,
		0.371655,
		0.3057,
		0.242186,
		0.182603,
		0.128533,
		0.08164,
		0.043674,
		0.016464,
		0.001913,
		0.000375,
		0.007397,
		0.022747,
		0.045692,
		0.075484,
		0.111363,
		0.152563,
		0.198309,
		0.247821,
		0.300316,
		0.355015,
		0.411143,
		0.467931,
		0.524629,
		0.580504,
		0.634852,
		0.687003,
		0.736332,
		0.782268,
		0.824309,
		0.862033,
		0.895116,
		0.92335,
		0.946661,
		0.965125,
		0.978993,
		0.988691,
		0.994829,
		0.998174,
		0.999598,
		0.999972,
		1,
		0.999971,
		0.999584,
		0.998079,
		0.994478,
		0.987755,
		0.976956,
		0.961278,
		0.940109,
		0.913042,
		0.879882,
		0.840634,
		0.795497,
		0.744848,
		0.689233,
		0.629348,
		0.566035,
		0.500268,
		0.433143,
		0.365872,
		0.299776,
		0.236278,
		0.176896,
		0.123241,
		0.077011,
		0.039989,
		0.014037,
		0.001092,
		9.5e-005,
		0.00592,
		0.020251,
		0.042346,
		0.071448,
		0.106789,
		0.147593,
		0.193074,
		0.242442,
		0.294906,
		0.349672,
		0.405954,
		0.462975,
		0.519969,
		0.576193,
		0.63093,
		0.683499,
		0.73326,
		0.779633,
		0.822103,
		0.860238,
		0.893703,
		0.922282,
		0.945892,
		0.964606,
		0.978669,
		0.988511,
		0.994744,
		0.998143,
		0.999591,
		0.999971,
		1,
		0.999971,
		0.999576,
		0.998046,
		0.994386,
		0.987556,
		0.976593,
		0.960686,
		0.939222,
		0.911794,
		0.878214,
		0.838495,
		0.792848,
		0.741665,
		0.685507,
		0.625091,
		0.561279,
		0.495067,
		0.427577,
		0.360045,
		0.29382,
		0.230353,
		0.171193,
		0.117983,
		0.072452,
		0.036415,
		0.011771,
		0.000492,
		0.004591,
		0.01787,
		0.039085,
		0.067471,
		0.102252,
		0.142639,
		0.187838,
		0.237049,
		0.289469,
		0.344293,
		0.400724,
		0.457972,
		0.51526,
		0.571833,
		0.62696,
		0.679948,
		0.730146,
		0.77696,
		0.819865,
		0.858415,
		0.892267,
		0.921195,
		0.945109,
		0.964076,
		0.978339,
		0.988327,
		0.994657,
		0.998111,
		0.999584,
		0.999971,
		1,
		0.99997,
		0.999569,
		0.998013,
		0.994292,
		0.987353,
		0.976223,
		0.960083,
		0.938318,
		0.910525,
		0.876517,
		0.836321,
		0.790157,
		0.738434,
		0.681729,
		0.620779,
		0.556466,
		0.489811,
		0.421959,
		0.354174,
		0.287832,
		0.224414,
		0.1655,
		0.112762,
		0.067965,
		0.032959,
		0.009675,
		0.000125,
		0.003417,
		0.015609,
		0.035914,
		0.063557,
		0.097753,
		0.137703,
		0.182603,
		0.231643,
		0.284006,
		0.33888,
		0.395453,
		0.452924,
		0.510503,
		0.567424,
		0.622942,
		0.676351,
		0.726989,
		0.774249,
		0.817592,
		0.856563,
		0.890808,
		0.92009,
		0.944313,
		0.963537,
		0.978003,
		0.98814,
		0.994569,
		0.998079,
		0.999576,
		0.99997,
		1,
		0.99997,
		0.999561,
		0.997978,
		0.994196,
		0.987147,
		0.975845,
		0.959469,
		0.937398,
		0.909232,
		0.874792,
		0.834112,
		0.787425,
		0.735156,
		0.677898,
		0.61641,
		0.551596,
		0.484498,
		0.416289,
		0.34826,
		0.281814,
		0.218463,
		0.159817,
		0.107582,
		0.063557,
		0.029626,
		0.007758,
		0.002404,
		0.013475,
		0.032838,
		0.05971,
		0.093296,
		0.132788,
		0.177371,
		0.226223,
		0.278519,
		0.333432,
		0.39014,
		0.447829,
		0.505698,
		0.562964,
		0.618874,
		0.672708,
		0.723788,
		0.771497,
		0.815284,
		0.854681,
		0.889324,
		0.918966,
		0.943502,
		0.962988,
		0.977661,
		0.987949,
		0.994478,
		0.998046,
		0.999569,
		0.99997,
		1,
		0.999969,
		0.999553,
		0.997943,
		0.994098,
		0.986936,
		0.97546,
		0.958843,
		0.936461,
		0.907917,
		0.873037,
		0.831866,
		0.78465,
		0.731828,
		0.674014,
		0.611985,
		0.546667,
		0.479129,
		0.410568,
		0.342303,
		0.275766,
		0.2125,
		0.154148,
		0.102448,
		0.059234,
		0.026425,
		0.006028,
		0.001559,
		0.011471,
		0.02986,
		0.055933,
		0.088884,
		0.127896,
		0.172143,
		0.220793,
		0.273008,
		0.327951,
		0.384786,
		0.442688,
		0.500843,
		0.558455,
		0.614757,
		0.669016,
		0.720543,
		0.768706,
		0.812941,
		0.852769,
		0.887815,
		0.917822,
		0.942676,
		0.962429,
		0.977312,
		0.987755,
		0.994386,
		0.998013,
		0.999561,
		0.999969,
		1,
		0.999969,
		0.999545,
		0.997907,
		0.993998,
		0.98672,
		0.975068,
		0.958205,
		0.935507,
		0.906578,
		0.871252,
		0.829583,
		0.781832,
		0.728452,
		0.670076,
		0.607502,
		0.541681,
		0.473704,
		0.404796,
		0.336303,
		0.269691,
		0.206529,
		0.148496,
		0.097363,
		0.055,
		0.023363,
		0.004495,
		0.000888,
		0.009606,
		0.026987,
		0.05223,
		0.08452,
		0.12303,
		0.166922,
		0.215353,
		0.267474,
		0.322436,
		0.379391,
		0.437501,
		0.495938,
		0.553895,
		0.61059,
		0.665277,
		0.717253,
		0.765874,
		0.810562,
		0.850827,
		0.886281,
		0.916658,
		0.941836,
		0.961859,
		0.976956,
		0.987556,
		0.994292,
		0.997978,
		0.999553,
		0.999969,
		1,
		0.999968,
		0.999537,
		0.99787,
		0.993896,
		0.986501,
		0.974668,
		0.957554,
		0.934534,
		0.905215,
		0.869437,
		0.827263,
		0.778969,
		0.725025,
		0.666082,
		0.602961,
		0.536635,
		0.468221,
		0.398972,
		0.330263,
		0.263588,
		0.200552,
		0.142864,
		0.092332,
		0.050862,
		0.020447,
		0.003169,
		0.0004,
		0.007884,
		0.024223,
		0.048607,
		0.080209,
		0.118192,
		0.16171,
		0.209905,
		0.261919,
		0.316889,
		0.373956,
		0.432268,
		0.490984,
		0.549284,
		0.606372,
		0.661489,
		0.713918,
		0.763,
		0.808146,
		0.848853,
		0.884721,
		0.915473,
		0.94098,
		0.961278,
		0.976593,
		0.987353,
		0.994196,
		0.997943,
		0.999545,
		0.999968,
		1,
		0.999968,
		0.999529,
		0.997833,
		0.993791,
		0.986276,
		0.97426,
		0.956891,
		0.933544,
		0.903828,
		0.86759,
		0.824905,
		0.776061,
		0.721547,
		0.662033,
		0.598361,
		0.531529,
		0.462682,
		0.393097,
		0.324181,
		0.25746,
		0.19457,
		0.137255,
		0.08736,
		0.046826,
		0.017685,
		0.002059,
		0.000101,
		0.006313,
		0.021574,
		0.045066,
		0.075953,
		0.113386,
		0.156508,
		0.204451,
		0.256343,
		0.31131,
		0.36848,
		0.426988,
		0.48598,
		0.544621,
		0.602103,
		0.657651,
		0.710536,
		0.760084,
		0.805693,
		0.846847,
		0.883135,
		0.914268,
		0.940109,
		0.960686,
		0.976223,
		0.987147,
		0.994098,
		0.997907,
		0.999537,
		0.999968,
		1,
		0.999967,
		0.99952,
		0.997795,
		0.993685,
		0.986047,
		0.973843,
		0.956215,
		0.932535,
		0.902416,
		0.865711,
		0.822507,
		0.773107,
		0.718017,
		0.657927,
		0.593702,
		0.526364,
		0.457085,
		0.387171,
		0.31806,
		0.251308,
		0.188586,
		0.131674,
		0.082451,
		0.042898,
		0.015085,
		0.001176,
		0.004898,
		0.019046,
		0.041614,
		0.071756,
		0.108615,
		0.151319,
		0.198992,
		0.250747,
		0.3057,
		0.362964,
		0.421662,
		0.480925,
		0.539907,
		0.597782,
		0.653763,
		0.707107,
		0.757125,
		0.803203,
		0.844809,
		0.881522,
		0.913042,
		0.939222,
		0.960083,
		0.975845,
		0.986936,
		0.993998,
		0.99787,
		0.999529,
		0.999967,
		1,
		0.999966,
		0.999511,
		0.997755,
		0.993576,
		0.985814,
		0.973418,
		0.955526,
		0.931507,
		0.900978,
		0.863799,
		0.820069,
		0.770107,
		0.714434,
		0.653763,
		0.588982,
		0.521138,
		0.45143,
		0.381194,
		0.311899,
		0.245133,
		0.182603,
		0.126123,
		0.077611,
		0.039085,
		0.012657,
		0.00053,
		0.003647,
		0.016644,
		0.038254,
		0.067623,
		0.103881,
		0.146146,
		0.193529,
		0.245133,
		0.300059,
		0.357409,
		0.416289,
		0.475821,
		0.535139,
		0.593409,
		0.649825,
		0.70363,
		0.754123,
		0.800673,
		0.842738,
		0.879882,
		0.911794,
		0.938318,
		0.959469,
		0.97546,
		0.98672,
		0.993896,
		0.997833,
		0.99952,
		0.999966,
		1,
		0.999966,
		0.999502,
		0.997715,
		0.993464,
		0.985575,
		0.972985,
		0.954824,
		0.93046,
		0.899514,
		0.861854,
		0.817592,
		0.767059,
		0.710798,
		0.649542,
		0.584202,
		0.515852,
		0.445718,
		0.375167,
		0.3057,
		0.238939,
		0.176624,
		0.120607,
		0.072845,
		0.035395,
		0.010409,
		0.000135,
		0.002567,
		0.014374,
		0.034993,
		0.063557,
		0.099188,
		0.140991,
		0.188066,
		0.239503,
		0.294389,
		0.351814,
		0.410871,
		0.470665,
		0.53032,
		0.588982,
		0.645835,
		0.700105,
		0.751077,
		0.798105,
		0.840634,
		0.878214,
		0.910525,
		0.937398,
		0.958843,
		0.975068,
		0.986501,
		0.993791,
		0.997795,
		0.999511,
		0.999966,
		1,
		0.999965,
		0.999493,
		0.997675,
		0.99335,
		0.985332,
		0.972543,
		0.954108,
		0.929393,
		0.898023,
		0.859876,
		0.815072,
		0.763962,
		0.707107,
		0.645261,
		0.579359,
		0.510503,
		0.439948,
		0.369091,
		0.299464,
		0.232725,
		0.170651,
		0.11513,
		0.068159,
		0.031834,
		0.008351,
		0.001665,
		0.012242,
		0.031834,
		0.059563,
		0.094539,
		0.135857,
		0.182603,
		0.233856,
		0.28869,
		0.34618,
		0.405406,
		0.465459,
		0.525446,
		0.584502,
		0.641793,
		0.696531,
		0.747985,
		0.795497,
		0.838495,
		0.876517,
		0.909232,
		0.936461,
		0.958205,
		0.974668,
		0.986276,
		0.993685,
		0.997755,
		0.999502,
		0.999965,
		1,
		0.999964,
		0.999484,
		0.997633,
		0.993234,
		0.985083,
		0.972091,
		0.953377,
		0.928306,
		0.896506,
		0.857862,
		0.812511,
		0.760817,
		0.703361,
		0.64092,
		0.574455,
		0.505093,
		0.43412,
		0.362964,
		0.293191,
		0.226495,
		0.164687,
		0.109696,
		0.063557,
		0.02841,
		0.006493,
		0.000949,
		0.010256,
		0.028784,
		0.055645,
		0.089938,
		0.130747,
		0.177144,
		0.228195,
		0.282963,
		0.340507,
		0.399895,
		0.460201,
		0.52052,
		0.579968,
		0.637698,
		0.692907,
		0.744848,
		0.792848,
		0.836321,
		0.874792,
		0.907917,
		0.935507,
		0.957554,
		0.97426,
		0.986047,
		0.993576,
		0.997715,
		0.999493,
		0.999964,
		1,
		0.999964,
		0.999474,
		0.99759,
		0.993115,
		0.984829,
		0.971631,
		0.952633,
		0.927198,
		0.89496,
		0.855814,
		0.809907,
		0.757621,
		0.699558,
		0.636519,
		0.569487,
		0.499621,
		0.428234,
		0.356789,
		0.286884,
		0.22025,
		0.158736,
		0.10431,
		0.059048,
		0.025132,
		0.004845,
		0.000428,
		0.008421,
		0.025848,
		0.051808,
		0.085389,
		0.125662,
		0.171689,
		0.222522,
		0.277209,
		0.334797,
		0.394338,
		0.454892,
		0.515539,
		0.575379,
		0.63355,
		0.689233,
		0.741665,
		0.790157,
		0.834112,
		0.873037,
		0.906578,
		0.934534,
		0.956891,
		0.973843,
		0.985814,
		0.993464,
		0.997675,
		0.999484,
		0.999964,
		1,
		0.999963,
		0.999464,
		0.997546,
		0.992994,
		0.984569,
		0.971161,
		0.951873,
		0.926069,
		0.893386,
		0.853729,
		0.807258,
		0.754375,
		0.695699,
		0.632056,
		0.564456,
		0.494086,
		0.422291,
		0.350565,
		0.280543,
		0.213992,
		0.1528,
		0.098976,
		0.054636,
		0.022008,
		0.003417,
		0.000108,
		0.006746,
		0.023032,
		0.048056,
		0.080895,
		0.120607,
		0.166242,
		0.216838,
		0.271429,
		0.32905,
		0.388735,
		0.449532,
		0.510503,
		0.570735,
		0.629348,
		0.685507,
		0.738434,
		0.787425,
		0.831866,
		0.871252,
		0.905215,
		0.933544,
		0.956215,
		0.973418,
		0.985575,
		0.99335,
		0.997633,
		0.999474,
		0.999963,
		1,
		0.999962,
		0.999454,
		0.997501,
		0.99287,
		0.984305,
		0.970681,
		0.951098,
		0.924919,
		0.891784,
		0.851607,
		0.804566,
		0.751077,
		0.691782,
		0.62753,
		0.559361,
		0.488488,
		0.416289,
		0.344293,
		0.27417,
		0.207724,
		0.146884,
		0.093699,
		0.05033,
		0.019046,
		0.002221,
		0.005236,
		0.020342,
		0.044395,
		0.076461,
		0.115585,
		0.160804,
		0.211144,
		0.265625,
		0.323265,
		0.383087,
		0.444121,
		0.505413,
		0.566035,
		0.625091,
		0.681729,
		0.735156,
		0.78465,
		0.829583,
		0.869437,
		0.903828,
		0.932535,
		0.955526,
		0.972985,
		0.985332,
		0.993234,
		0.99759,
		0.999464,
		0.999962,
		1,
		0.999962,
		0.999444,
		0.997456,
		0.992743,
		0.984034,
		0.970192,
		0.950308,
		0.923746,
		0.890151,
		0.849448,
		0.801828,
		0.747726,
		0.687806,
		0.622942,
		0.554201,
		0.482827,
		0.41023,
		0.337974,
		0.267766,
		0.201449,
		0.140991,
		0.088485,
		0.046135,
		0.016256,
		0.001269,
		0.003901,
		0.017784,
		0.04083,
		0.07209,
		0.110599,
		0.155379,
		0.205443,
		0.259797,
		0.317445,
		0.377394,
		0.438658,
		0.500268,
		0.561279,
		0.620779,
		0.677898,
		0.731828,
		0.781832,
		0.827263,
		0.86759,
		0.902416,
		0.931507,
		0.954824,
		0.972543,
		0.985083,
		0.993115,
		0.997546,
		0.999454,
		0.999962,
		1,
		0.999961,
		0.999434,
		0.997409,
		0.992614,
		0.983758,
		0.969692,
		0.949502,
		0.922551,
		0.888488,
		0.847251,
		0.799043,
		0.744321,
		0.68377,
		0.618289,
		0.548975,
		0.477101,
		0.404113,
		0.331609,
		0.261333,
		0.195168,
		0.135126,
		0.083339,
		0.042059,
		0.013648,
		0.000573,
		0.002747,
		0.015366,
		0.037366,
		0.067787,
		0.105651,
		0.149968,
		0.199736,
		0.253947,
		0.31159,
		0.371655,
		0.433143,
		0.495067,
		0.556466,
		0.61641,
		0.674014,
		0.728452,
		0.778969,
		0.824905,
		0.865711,
		0.900978,
		0.93046,
		0.954108,
		0.972091,
		0.984829,
		0.992994,
		0.997501,
		0.999444,
		0.999961,
		1,
		0.99996,
		0.999423,
		0.997361,
		0.992481,
		0.983475,
		0.969182,
		0.94868,
		0.921332,
		0.886795,
		0.845015,
		0.796212,
		0.740862,
		0.679673,
		0.613572,
		0.543682,
		0.471312,
		0.397939,
		0.325198,
		0.254872,
		0.188885,
		0.129291,
		0.078267,
		0.03811,
		0.011231,
		0.000146,
		0.001783,
		0.013093,
		0.034009,
		0.063557,
		0.100747,
		0.144575,
		0.194026,
		0.248076,
		0.3057,
		0.365872,
		0.427577,
		0.489811,
		0.551596,
		0.611985,
		0.670076,
		0.725025,
		0.776061,
		0.822507,
		0.863799,
		0.899514,
		0.929393,
		0.953377,
		0.971631,
		0.984569,
		0.99287,
		0.997456,
		0.999434,
		0.99996,
		1,
		0.999959,
		0.999412,
		0.997312,
		0.992346,
		0.983187,
		0.968661,
		0.947842,
		0.92009,
		0.88507,
		0.842738,
		0.793332,
		0.737347,
		0.675515,
		0.608789,
		0.538323,
		0.465459,
		0.391707,
		0.318742,
		0.248386,
		0.182603,
		0.123492,
		0.073275,
		0.034297,
		0.009016,
		0.001017,
		0.010974,
		0.030765,
		0.059405,
		0.09589,
		0.139203,
		0.188314,
		0.242186,
		0.299776,
		0.360045,
		0.421959,
		0.484498,
		0.546667,
		0.607502,
		0.666082,
		0.721547,
		0.773107,
		0.820069,
		0.861854,
		0.898023,
		0.928306,
		0.952633,
		0.971161,
		0.984305,
		0.992743,
		0.997409,
		0.999423,
		0.999959,
		1,
		0.999959,
		0.9994,
		0.997262,
		0.992207,
		0.982892,
		0.968129,
		0.946986,
		0.918824,
		0.883312,
		0.840421,
		0.790404,
		0.733775,
		0.671294,
		0.603939,
		0.532897,
		0.45954,
		0.385418,
		0.312242,
		0.241875,
		0.176325,
		0.117733,
		0.06837,
		0.030627,
		0.007014,
		0.000458,
		0.009016,
		0.027641,
		0.055335,
		0.091084,
		0.133854,
		0.182603,
		0.236278,
		0.29382,
		0.354174,
		0.416289,
		0.479129,
		0.541681,
		0.602961,
		0.662033,
		0.718017,
		0.770107,
		0.817592,
		0.859876,
		0.896506,
		0.927198,
		0.951873,
		0.970681,
		0.984034,
		0.992614,
		0.997361,
		0.999412,
		0.999959,
		1,
		0.999958,
		0.999389,
		0.99721,
		0.992065,
		0.982591,
		0.967586,
		0.946113,
		0.917532,
		0.881522,
		0.838063,
		0.787425,
		0.730146,
		0.667009,
		0.599022,
		0.527402,
		0.453557,
		0.379073,
		0.3057,
		0.235343,
		0.170054,
		0.112019,
		0.063557,
		0.02711,
		0.005236,
		0.000116,
		0.007225,
		0.024641,
		0.051353,
		0.086332,
		0.128533,
		0.176896,
		0.230353,
		0.287832,
		0.34826,
		0.410568,
		0.473704,
		0.536635,
		0.598361,
		0.657927,
		0.714434,
		0.767059,
		0.815072,
		0.857862,
		0.89496,
		0.926069,
		0.951098,
		0.970192,
		0.983758,
		0.992481,
		0.997312,
		0.9994,
		0.999958,
		1,
		0.999957,
		0.999377,
		0.997158,
		0.99192,
		0.982283,
		0.967032,
		0.945222,
		0.916216,
		0.879698,
		0.835662,
		0.784396,
		0.726459,
		0.66266,
		0.594037,
		0.521839,
		0.447509,
		0.372671,
		0.299116,
		0.228792,
		0.163793,
		0.106356,
		0.058845,
		0.023754,
		0.003696,
		0.005611,
		0.021774,
		0.047464,
		0.08164,
		0.123241,
		0.171193,
		0.224414,
		0.281814,
		0.342303,
		0.404796,
		0.468221,
		0.531529,
		0.593702,
		0.653763,
		0.710798,
		0.763962,
		0.812511,
		0.855814,
		0.893386,
		0.924919,
		0.950308,
		0.969692,
		0.983475,
		0.992346,
		0.997262,
		0.999389,
		0.999957,
		1,
		0.999956,
		0.999365,
		0.997104,
		0.991772,
		0.981968,
		0.966466,
		0.944313,
		0.914874,
		0.877839,
		0.833218,
		0.781314,
		0.722712,
		0.658245,
		0.588982,
		0.516206,
		0.441396,
		0.366214,
		0.292492,
		0.222223,
		0.157547,
		0.100747,
		0.05424,
		0.02057,
		0.002404,
		0.004182,
		0.019046,
		0.043674,
		0.077011,
		0.117983,
		0.1655,
		0.218463,
		0.275766,
		0.336303,
		0.398972,
		0.462682,
		0.526364,
		0.588982,
		0.649542,
		0.707107,
		0.760817,
		0.809907,
		0.853729,
		0.891784,
		0.923746,
		0.949502,
		0.969182,
		0.983187,
		0.992207,
		0.99721,
		0.999377,
		0.999956,
		1,
		0.999955,
		0.999352,
		0.997049,
		0.991621,
		0.981647,
		0.965887,
		0.943385,
		0.913505,
		0.875946,
		0.830729,
		0.77818,
		0.718904,
		0.653763,
		0.583858,
		0.510503,
		0.435217,
		0.359701,
		0.28583,
		0.21564,
		0.151319,
		0.0952,
		0.049751,
		0.017568,
		0.001374,
		0.002946,
		0.016464,
		0.039989,
		0.072452,
		0.112762,
		0.159817,
		0.2125,
		0.269691,
		0.330263,
		0.393097,
		0.457085,
		0.521138,
		0.584202,
		0.645261,
		0.703361,
		0.757621,
		0.807258,
		0.851607,
		0.890151,
		0.922551,
		0.94868,
		0.968661,
		0.982892,
		0.992065,
		0.997158,
		0.999365,
		0.999955,
		1,
		0.999954,
		0.99934,
		0.996992,
		0.991466,
		0.981318,
		0.965297,
		0.942438,
		0.912108,
		0.874016,
		0.828196,
		0.774992,
		0.715035,
		0.649214,
		0.578663,
		0.504731,
		0.428973,
		0.353134,
		0.27913,
		0.209044,
		0.145114,
		0.08972,
		0.045385,
		0.014759,
		0.000621,
		0.001913,
		0.014037,
		0.036415,
		0.067965,
		0.107582,
		0.154148,
		0.206529,
		0.263588,
		0.324181,
		0.387171,
		0.45143,
		0.515852,
		0.579359,
		0.64092,
		0.699558,
		0.754375,
		0.804566,
		0.849448,
		0.888488,
		0.921332,
		0.947842,
		0.968129,
		0.982591,
		0.99192,
		0.997104,
		0.999352,
		0.999954,
		1,
		0.999953,
		0.999327,
		0.996935,
		0.991307,
		0.980982,
		0.964693,
		0.941471,
		0.910685,
		0.872049,
		0.825616,
		0.771749,
		0.711103,
		0.644597,
		0.573396,
		0.498887,
		0.422664,
		0.346512,
		0.272394,
		0.20244,
		0.138935,
		0.084314,
		0.04115,
		0.012153,
		0.000158,
		0.001092,
		0.011771,
		0.032959,
		0.063557,
		0.102448,
		0.148496,
		0.200552,
		0.25746,
		0.31806,
		0.381194,
		0.445718,
		0.510503,
		0.574455,
		0.636519,
		0.695699,
		0.751077,
		0.801828,
		0.847251,
		0.886795,
		0.92009,
		0.946986,
		0.967586,
		0.982283,
		0.991772,
		0.997049,
		0.99934,
		0.999953,
		1,
		0.999952,
		0.999313,
		0.996876,
		0.991144,
		0.980638,
		0.964076,
		0.940484,
		0.909232,
		0.870045,
		0.82299,
		0.76845,
		0.707107,
		0.63991,
		0.568056,
		0.492972,
		0.416289,
		0.339837,
		0.265625,
		0.19583,
		0.132788,
		0.078987,
		0.037057,
		0.009762,
		0.000492,
		0.009675,
		0.029626,
		0.059234,
		0.097363,
		0.142864,
		0.19457,
		0.251308,
		0.311899,
		0.375167,
		0.439948,
		0.505093,
		0.569487,
		0.632056,
		0.691782,
		0.747726,
		0.799043,
		0.845015,
		0.88507,
		0.918824,
		0.946113,
		0.967032,
		0.981968,
		0.991621,
		0.996992,
		0.999327,
		0.999952,
		1,
		0.999951,
		0.999299,
		0.996815,
		0.990978,
		0.980287,
		0.963446,
		0.939477,
		0.907751,
		0.868003,
		0.820315,
		0.765094,
		0.703046,
		0.635152,
		0.562644,
		0.486985,
		0.409849,
		0.333111,
		0.258824,
		0.189216,
		0.126677,
		0.073747,
		0.033113,
		0.0076,
		0.000125,
		0.007758,
		0.026425,
		0.055,
		0.092332,
		0.137255,
		0.188586,
		0.245133,
		0.3057,
		0.369091,
		0.43412,
		0.499621,
		0.564456,
		0.62753,
		0.687806,
		0.744321,
		0.796212,
		0.842738,
		0.883312,
		0.917532,
		0.945222,
		0.966466,
		0.981647,
		0.991466,
		0.996935,
		0.999313,
		0.999951,
		1,
		0.99995,
		0.999285,
		0.996753,
		0.990808,
		0.979927,
		0.962803,
		0.938448,
		0.90624,
		0.865921,
		0.817592,
		0.76168,
		0.698919,
		0.630323,
		0.557157,
		0.480925,
		0.403344,
		0.326332,
		0.251992,
		0.182603,
		0.120607,
		0.068601,
		0.02933,
		0.005678,
		0.006028,
		0.023363,
		0.050862,
		0.08736,
		0.131674,
		0.182603,
		0.238939,
		0.299464,
		0.362964,
		0.428234,
		0.494086,
		0.559361,
		0.622942,
		0.68377,
		0.740862,
		0.793332,
		0.840421,
		0.881522,
		0.916216,
		0.944313,
		0.965887,
		0.981318,
		0.991307,
		0.996876,
		0.999299,
		0.99995,
		1,
		0.999949,
		0.999271,
		0.99669,
		0.990634,
		0.97956,
		0.962145,
		0.937398,
		0.904698,
		0.863799,
		0.814818,
		0.758206,
		0.694725,
		0.62542,
		0.551596,
		0.474793,
		0.396775,
		0.319503,
		0.245133,
		0.175995,
		0.114585,
		0.063557,
		0.025717,
		0.00401,
		0.004495,
		0.020447,
		0.046826,
		0.082451,
		0.126123,
		0.176624,
		0.232725,
		0.293191,
		0.356789,
		0.422291,
		0.488488,
		0.554201,
		0.618289,
		0.679673,
		0.737347,
		0.790404,
		0.838063,
		0.879698,
		0.914874,
		0.943385,
		0.965297,
		0.980982,
		0.991144,
		0.996815,
		0.999285,
		0.999949,
		1,
		0.999948,
		0.999256,
		0.996625,
		0.990456,
		0.979184,
		0.961473,
		0.936326,
		0.903125,
		0.861636,
		0.811993,
		0.754672,
		0.690463,
		0.620445,
		0.545959,
		0.468589,
		0.39014,
		0.312626,
		0.238249,
		0.169394,
		0.108615,
		0.058623,
		0.022285,
		0.00261,
		0.003169,
		0.017685,
		0.042898,
		0.077611,
		0.120607,
		0.170651,
		0.226495,
		0.286884,
		0.350565,
		0.416289,
		0.482827,
		0.548975,
		0.613572,
		0.675515,
		0.733775,
		0.787425,
		0.835662,
		0.877839,
		0.913505,
		0.942438,
		0.964693,
		0.980638,
		0.990978,
		0.996753,
		0.999271,
		0.999948,
		1,
		0.999947,
		0.999241,
		0.996558,
		0.990274,
		0.978799,
		0.960786,
		0.935231,
		0.90152,
		0.859431,
		0.809117,
		0.751077,
		0.686131,
		0.615394,
		0.540245,
		0.46231,
		0.383441,
		0.3057,
		0.231342,
		0.162806,
		0.102704,
		0.053808,
		0.019046,
		0.001493,
		0.002059,
		0.015085,
		0.039085,
		0.072845,
		0.11513,
		0.164687,
		0.22025,
		0.280543,
		0.344293,
		0.41023,
		0.477101,
		0.543682,
		0.608789,
		0.671294,
		0.730146,
		0.784396,
		0.833218,
		0.875946,
		0.912108,
		0.941471,
		0.964076,
		0.980287,
		0.990808,
		0.99669,
		0.999256,
		0.999947,
		1,
		0.999946,
		0.999225,
		0.99649,
		0.990087,
		0.978406,
		0.960083,
		0.934112,
		0.899882,
		0.857183,
		0.806187,
		0.747418,
		0.681729,
		0.610268,
		0.534454,
		0.455958,
		0.376679,
		0.298728,
		0.224414,
		0.156234,
		0.096858,
		0.049119,
		0.016011,
		0.000675,
		0.001176,
		0.012657,
		0.035395,
		0.068159,
		0.109696,
		0.158736,
		0.213992,
		0.27417,
		0.337974,
		0.404113,
		0.471312,
		0.538323,
		0.603939,
		0.667009,
		0.726459,
		0.781314,
		0.830729,
		0.874016,
		0.910685,
		0.940484,
		0.963446,
		0.979927,
		0.990634,
		0.996625,
		0.999241,
		0.999946,
		1,
		0.999945,
		0.99921,
		0.996419,
		0.989896,
		0.978003,
		0.959365,
		0.93297,
		0.898211,
		0.854892,
		0.803203,
		0.743696,
		0.677255,
		0.605064,
		0.528585,
		0.449532,
		0.369853,
		0.29171,
		0.21747,
		0.149684,
		0.091084,
		0.044568,
		0.013194,
		0.000172,
		0.00053,
		0.010409,
		0.031834,
		0.063557,
		0.10431,
		0.1528,
		0.207724,
		0.267766,
		0.331609,
		0.397939,
		0.465459,
		0.532897,
		0.599022,
		0.66266,
		0.722712,
		0.77818,
		0.828196,
		0.872049,
		0.909232,
		0.939477,
		0.962803,
		0.97956,
		0.990456,
		0.996558,
		0.999225,
		0.999945,
		1,
		0.999944,
		0.999193,
		0.996348,
		0.9897,
		0.977591,
		0.958631,
		0.931803,
		0.896506,
		0.852555,
		0.800163,
		0.739908,
		0.672708,
		0.599783,
		0.522638,
		0.443032,
		0.362964,
		0.28465,
		0.210511,
		0.14316,
		0.085389,
		0.040162,
		0.010606,
		0.000135,
		0.008351,
		0.02841,
		0.059048,
		0.098976,
		0.146884,
		0.201449,
		0.261333,
		0.325198,
		0.391707,
		0.45954,
		0.527402,
		0.594037,
		0.658245,
		0.718904,
		0.774992,
		0.825616,
		0.870045,
		0.907751,
		0.938448,
		0.962145,
		0.979184,
		0.990274,
		0.99649,
		0.99921,
		0.999944,
		1,
		0.999943,
		0.999176,
		0.996274,
		0.989499,
		0.97717,
		0.957881,
		0.930611,
		0.894765,
		0.850172,
		0.797066,
		0.736055,
		0.668086,
		0.594422,
		0.516611,
		0.436458,
		0.356013,
		0.277548,
		0.203541,
		0.136666,
		0.079781,
		0.035914,
		0.008262,
		0.006493,
		0.025132,
		0.054636,
		0.093699,
		0.140991,
		0.195168,
		0.254872,
		0.318742,
		0.385418,
		0.453557,
		0.521839,
		0.588982,
		0.653763,
		0.715035,
		0.771749,
		0.82299,
		0.868003,
		0.90624,
		0.937398,
		0.961473,
		0.978799,
		0.990087,
		0.996419,
		0.999193,
		0.999943,
		1,
		0.999941,
		0.999159,
		0.996199,
		0.989293,
		0.976739,
		0.957114,
		0.929393,
		0.892988,
		0.847743,
		0.793912,
		0.732133,
		0.663389,
		0.588982,
		0.510503,
		0.429809,
		0.349001,
		0.270407,
		0.196564,
		0.13021,
		0.074267,
		0.031834,
		0.006177,
		0.004845,
		0.022008,
		0.05033,
		0.088485,
		0.135126,
		0.188885,
		0.248386,
		0.312242,
		0.379073,
		0.447509,
		0.516206,
		0.583858,
		0.649214,
		0.711103,
		0.76845,
		0.820315,
		0.865921,
		0.904698,
		0.936326,
		0.960786,
		0.978406,
		0.989896,
		0.996348,
		0.999176,
		0.999941,
		1,
		0.99994,
		0.999141,
		0.996121,
		0.989083,
		0.976297,
		0.956329,
		0.928149,
		0.891175,
		0.845265,
		0.790699,
		0.728142,
		0.658615,
		0.583461,
		0.504315,
		0.423087,
		0.341929,
		0.263228,
		0.189584,
		0.123796,
		0.068856,
		0.027933,
		0.004366,
		0.003417,
		0.019046,
		0.046135,
		0.083339,
		0.129291,
		0.182603,
		0.241875,
		0.3057,
		0.372671,
		0.441396,
		0.510503,
		0.578663,
		0.644597,
		0.707107,
		0.765094,
		0.817592,
		0.863799,
		0.903125,
		0.935231,
		0.960083,
		0.978003,
		0.9897,
		0.996274,
		0.999159,
		0.99994,
		1,
		0.999939,
		0.999123,
		0.996042,
		0.988867,
		0.975845,
		0.955526,
		0.926878,
		0.889324,
		0.842738,
		0.787425,
		0.724081,
		0.653763,
		0.577857,
		0.498046,
		0.416289,
		0.334797,
		0.256014,
		0.182603,
		0.117431,
		0.063557,
		0.024223,
		0.002844,
		0.002221,
		0.016256,
		0.042059,
		0.078267,
		0.123492,
		0.176325,
		0.235343,
		0.299116,
		0.366214,
		0.435217,
		0.504731,
		0.573396,
		0.63991,
		0.703046,
		0.76168,
		0.814818,
		0.861636,
		0.90152,
		0.934112,
		0.959365,
		0.977591,
		0.989499,
		0.996199,
		0.999141,
		0.999939,
		1,
		0.999937,
		0.999105,
		0.995961,
		0.988646,
		0.975383,
		0.954706,
		0.925579,
		0.887434,
		0.840161,
		0.78409,
		0.719949,
		0.648832,
		0.57217,
		0.491695,
		0.409418,
		0.327607,
		0.248768,
		0.175628,
		0.111122,
		0.05838,
		0.020718,
		0.001628,
		0.001269,
		0.013648,
		0.03811,
		0.073275,
		0.117733,
		0.170054,
		0.228792,
		0.292492,
		0.359701,
		0.428973,
		0.498887,
		0.568056,
		0.635152,
		0.698919,
		0.758206,
		0.811993,
		0.859431,
		0.899882,
		0.93297,
		0.958631,
		0.97717,
		0.989293,
		0.996121,
		0.999123,
		0.999937,
		1,
		0.999936,
		0.999086,
		0.995877,
		0.98842,
		0.974909,
		0.953866,
		0.924251,
		0.885504,
		0.837533,
		0.780692,
		0.715743,
		0.643821,
		0.566399,
		0.485261,
		0.402472,
		0.32036,
		0.241492,
		0.168662,
		0.104874,
		0.053333,
		0.01743,
		0.000737,
		0.000573,
		0.011231,
		0.034297,
		0.06837,
		0.112019,
		0.163793,
		0.222223,
		0.28583,
		0.353134,
		0.422664,
		0.492972,
		0.562644,
		0.630323,
		0.694725,
		0.754672,
		0.809117,
		0.857183,
		0.898211,
		0.931803,
		0.957881,
		0.976739,
		0.989083,
		0.996042,
		0.999105,
		0.999936,
		1,
		0.999935,
		0.999066,
		0.995792,
		0.988187,
		0.974424,
		0.953007,
		0.922895,
		0.883534,
		0.834852,
		0.77723,
		0.711463,
		0.638727,
		0.560543,
		0.478744,
		0.395453,
		0.313057,
		0.234189,
		0.16171,
		0.098696,
		0.048428,
		0.014374,
		0.000188,
		0.000146,
		0.009016,
		0.030627,
		0.063557,
		0.106356,
		0.157547,
		0.21564,
		0.27913,
		0.346512,
		0.416289,
		0.486985,
		0.557157,
		0.62542,
		0.690463,
		0.751077,
		0.806187,
		0.854892,
		0.896506,
		0.930611,
		0.957114,
		0.976297,
		0.988867,
		0.995961,
		0.999086,
		0.999935,
		1,
		0.999933,
		0.999046,
		0.995704,
		0.987949,
		0.973927,
		0.952128,
		0.921508,
		0.881522,
		0.832117,
		0.773702,
		0.707107,
		0.63355,
		0.5546,
		0.472143,
		0.38836,
		0.3057,
		0.226862,
		0.154777,
		0.092596,
		0.043674,
		0.011564,
		0.007014,
		0.02711,
		0.058845,
		0.100747,
		0.151319,
		0.209044,
		0.272394,
		0.339837,
		0.409849,
		0.480925,
		0.551596,
		0.620445,
		0.686131,
		0.747418,
		0.803203,
		0.852555,
		0.894765,
		0.929393,
		0.956329,
		0.975845,
		0.988646,
		0.995877,
		0.999066,
		0.999933,
		1,
		0.999932,
		0.999025,
		0.995614,
		0.987705,
		0.973418,
		0.951229,
		0.92009,
		0.879467,
		0.829327,
		0.770107,
		0.702674,
		0.628289,
		0.54857,
		0.465459,
		0.381194,
		0.29829,
		0.219514,
		0.147869,
		0.086581,
		0.039085,
		0.009016,
		0.005236,
		0.023754,
		0.05424,
		0.0952,
		0.145114,
		0.20244,
		0.265625,
		0.333111,
		0.403344,
		0.474793,
		0.545959,
		0.615394,
		0.681729,
		0.743696,
		0.800163,
		0.850172,
		0.892988,
		0.928149,
		0.955526,
		0.975383,
		0.98842,
		0.995792,
		0.999046,
		0.999932,
		1,
		0.99993,
		0.999004,
		0.995521,
		0.987455,
		0.972897,
		0.950308,
		0.918641,
		0.877369,
		0.826481,
		0.766443,
		0.698162,
		0.622942,
		0.542452,
		0.45869,
		0.373956,
		0.29083,
		0.212149,
		0.140991,
		0.08066,
		0.034672,
		0.006746,
		0.003696,
		0.02057,
		0.049751,
		0.08972,
		0.138935,
		0.19583,
		0.258824,
		0.326332,
		0.396775,
		0.468589,
		0.540245,
		0.610268,
		0.677255,
		0.739908,
		0.797066,
		0.847743,
		0.891175,
		0.926878,
		0.954706,
		0.974909,
		0.988187,
		0.995704,
		0.999025,
		0.99993,
		1,
		0.999929,
		0.998982,
		0.995426,
		0.987199,
		0.972363,
		0.949367,
		0.917159,
		0.875226,
		0.823577,
		0.76271,
		0.69357,
		0.617508,
		0.536244,
		0.451836,
		0.366646,
		0.283322,
		0.204772,
		0.134151,
		0.074843,
		0.030448,
		0.004772,
		0.002404,
		0.017568,
		0.045385,
		0.084314,
		0.132788,
		0.189216,
		0.251992,
		0.319503,
		0.39014,
		0.46231,
		0.534454,
		0.605064,
		0.672708,
		0.736055,
		0.793912,
		0.845265,
		0.889324,
		0.925579,
		0.953866,
		0.974424,
		0.987949,
		0.995614,
		0.999004,
		0.999929,
		1,
		0.999927,
		0.99896,
		0.995329,
		0.986936,
		0.971816,
		0.948403,
		0.915644,
		0.873037,
		0.820615,
		0.758906,
		0.688896,
		0.611985,
		0.529947,
		0.444897,
		0.359265,
		0.275766,
		0.197385,
		0.127354,
		0.069138,
		0.026425,
		0.003111,
		0.001374,
		0.014759,
		0.04115,
		0.078987,
		0.126677,
		0.182603,
		0.245133,
		0.312626,
		0.383441,
		0.455958,
		0.528585,
		0.599783,
		0.668086,
		0.732133,
		0.790699,
		0.842738,
		0.887434,
		0.924251,
		0.953007,
		0.973927,
		0.987705,
		0.995521,
		0.998982,
		0.999927,
		1,
		0.999925,
		0.998936,
		0.995229,
		0.986666,
		0.971256,
		0.947416,
		0.914095,
		0.870801,
		0.817592,
		0.755028,
		0.684139,
		0.606372,
		0.523558,
		0.437873,
		0.351814,
		0.268167,
		0.189994,
		0.120607,
		0.063557,
		0.02262,
		0.001783,
		0.000621,
		0.012153,
		0.037057,
		0.073747,
		0.120607,
		0.175995,
		0.238249,
		0.3057,
		0.376679,
		0.449532,
		0.522638,
		0.594422,
		0.663389,
		0.728142,
		0.787425,
		0.840161,
		0.885504,
		0.922895,
		0.952128,
		0.973418,
		0.987455,
		0.995426,
		0.99896,
		0.999925,
		1,
		0.999923,
		0.998913,
		0.995126,
		0.986389,
		0.970681,
		0.946406,
		0.91251,
		0.868517,
		0.814507,
		0.751077,
		0.679298,
		0.600669,
		0.517077,
		0.430764,
		0.344293,
		0.260527,
		0.182603,
		0.113919,
		0.058111,
		0.019046,
		0.000807,
		0.000158,
		0.009762,
		0.033113,
		0.068601,
		0.114585,
		0.169394,
		0.231342,
		0.298728,
		0.369853,
		0.443032,
		0.516611,
		0.588982,
		0.658615,
		0.724081,
		0.78409,
		0.837533,
		0.883534,
		0.921508,
		0.951229,
		0.972897,
		0.987199,
		0.995329,
		0.998936,
		0.999923,
		1,
		0.999922,
		0.998888,
		0.99502,
		0.986105,
		0.970093,
		0.945372,
		0.91089,
		0.866184,
		0.811359,
		0.747049,
		0.67437,
		0.594872,
		0.510503,
		0.423569,
		0.336705,
		0.252848,
		0.175218,
		0.107296,
		0.05281,
		0.015719,
		0.000206,
		0.0076,
		0.02933,
		0.063557,
		0.108615,
		0.162806,
		0.224414,
		0.29171,
		0.362964,
		0.436458,
		0.510503,
		0.583461,
		0.653763,
		0.719949,
		0.780692,
		0.834852,
		0.881522,
		0.92009,
		0.950308,
		0.972363,
		0.986936,
		0.995229,
		0.998913,
		0.999922,
		1,
		0.99992,
		0.998863,
		0.994912,
		0.985814,
		0.969489,
		0.944313,
		0.909232,
		0.863799,
		0.808146,
		0.742944,
		0.669354,
		0.588982,
		0.503836,
		0.416289,
		0.32905,
		0.245133,
		0.167843,
		0.100747,
		0.047667,
		0.012657,
		0.005678,
		0.025717,
		0.058623,
		0.102704,
		0.156234,
		0.21747,
		0.28465,
		0.356013,
		0.429809,
		0.504315,
		0.577857,
		0.648832,
		0.715743,
		0.77723,
		0.832117,
		0.879467,
		0.918641,
		0.949367,
		0.971816,
		0.986666,
		0.995126,
		0.998888,
		0.99992,
		1,
		0.999918,
		0.998838,
		0.994801,
		0.985515,
		0.968871,
		0.943228,
		0.907537,
		0.861363,
		0.804867,
		0.738759,
		0.664249,
		0.582997,
		0.497074,
		0.408924,
		0.321329,
		0.237387,
		0.160485,
		0.094282,
		0.042694,
		0.009877,
		0.00401,
		0.022285,
		0.053808,
		0.096858,
		0.149684,
		0.210511,
		0.277548,
		0.349001,
		0.423087,
		0.498046,
		0.57217,
		0.643821,
		0.711463,
		0.773702,
		0.829327,
		0.877369,
		0.917159,
		0.948403,
		0.971256,
		0.986389,
		0.99502,
		0.998863,
		0.999918,
		1,
		0.999916,
		0.998811,
		0.994686,
		0.985208,
		0.968237,
		0.942118,
		0.905802,
		0.858873,
		0.801521,
		0.734494,
		0.659052,
		0.576915,
		0.490217,
		0.401474,
		0.313545,
		0.229612,
		0.153149,
		0.08791,
		0.037906,
		0.007397,
		0.00261,
		0.019046,
		0.049119,
		0.091084,
		0.14316,
		0.203541,
		0.270407,
		0.341929,
		0.416289,
		0.491695,
		0.566399,
		0.638727,
		0.707107,
		0.770107,
		0.826481,
		0.875226,
		0.915644,
		0.947416,
		0.970681,
		0.986105,
		0.994912,
		0.998838,
		0.999916,
		1,
		0.999914,
		0.998784,
		0.994569,
		0.984893,
		0.967586,
		0.94098,
		0.904028,
		0.856329,
		0.798105,
		0.730146,
		0.653763,
		0.570735,
		0.483264,
		0.393939,
		0.3057,
		0.221812,
		0.145843,
		0.08164,
		0.033317,
		0.005236,
		0.001493,
		0.016011,
		0.044568,
		0.085389,
		0.136666,
		0.196564,
		0.263228,
		0.334797,
		0.409418,
		0.485261,
		0.560543,
		0.63355,
		0.702674,
		0.766443,
		0.823577,
		0.873037,
		0.914095,
		0.946406,
		0.970093,
		0.985814,
		0.994801,
		0.998811,
		0.999914,
		1,
		0.999912,
		0.998756,
		0.994448,
		0.984569,
		0.96692,
		0.939815,
		0.902212,
		0.853729,
		0.794618,
		0.725714,
		0.64838,
		0.564456,
		0.476215,
		0.38632,
		0.297795,
		0.213992,
		0.138573,
		0.075484,
		0.028941,
		0.003417,
		0.000675,
		0.013194,
		0.040162,
		0.079781,
		0.13021,
		0.189584,
		0.256014,
		0.327607,
		0.402472,
		0.478744,
		0.5546,
		0.628289,
		0.698162,
		0.76271,
		0.820615,
		0.870801,
		0.91251,
		0.945372,
		0.969489,
		0.985515,
		0.994686,
		0.998784,
		0.999912,
		1,
		0.99991,
		0.998727,
		0.994324,
		0.984237,
		0.966236,
		0.938621,
		0.900354,
		0.851071,
		0.791059,
		0.721196,
		0.642901,
		0.558077,
		0.469068,
		0.378617,
		0.289832,
		0.206156,
		0.131346,
		0.069452,
		0.024796,
		0.00196,
		0.000172,
		0.010606,
		0.035914,
		0.074267,
		0.123796,
		0.182603,
		0.248768,
		0.32036,
		0.395453,
		0.472143,
		0.54857,
		0.622942,
		0.69357,
		0.758906,
		0.817592,
		0.868517,
		0.91089,
		0.944313,
		0.968871,
		0.985208,
		0.994569,
		0.998756,
		0.99991,
		1,
		0.999908,
		0.998697,
		0.994196,
		0.983896,
		0.965534,
		0.937398,
		0.898452,
		0.848355,
		0.787425,
		0.71659,
		0.637323,
		0.551596,
		0.461824,
		0.370832,
		0.281814,
		0.198309,
		0.124172,
		0.063557,
		0.020897,
		0.000888,
		0.008262,
		0.031834,
		0.068856,
		0.117431,
		0.175628,
		0.241492,
		0.313057,
		0.38836,
		0.465459,
		0.542452,
		0.617508,
		0.688896,
		0.755028,
		0.814507,
		0.866184,
		0.909232,
		0.943228,
		0.968237,
		0.984893,
		0.994448,
		0.998727,
		0.999908,
		1,
		0.999906,
		0.998667,
		0.994065,
		0.983546,
		0.964815,
		0.936145,
		0.896506,
		0.845578,
		0.783716,
		0.711894,
		0.631647,
		0.545012,
		0.454482,
		0.362964,
		0.273744,
		0.190456,
		0.117059,
		0.057812,
		0.017263,
		0.000227,
		0.006177,
		0.027933,
		0.063557,
		0.111122,
		0.168662,
		0.234189,
		0.3057,
		0.381194,
		0.45869,
		0.536244,
		0.611985,
		0.684139,
		0.751077,
		0.811359,
		0.863799,
		0.907537,
		0.942118,
		0.967586,
		0.984569,
		0.994324,
		0.998697,
		0.999906,
		1,
		0.999903,
		0.998635,
		0.99393,
		0.983187,
		0.964076,
		0.93486,
		0.894513,
		0.842738,
		0.779928,
		0.707107,
		0.625869,
		0.538323,
		0.447041,
		0.355015,
		0.265625,
		0.182603,
		0.110014,
		0.05223,
		0.013913,
		0.004366,
		0.024223,
		0.05838,
		0.104874,
		0.16171,
		0.226862,
		0.29829,
		0.373956,
		0.451836,
		0.529947,
		0.606372,
		0.679298,
		0.747049,
		0.808146,
		0.861363,
		0.905802,
		0.94098,
		0.96692,
		0.984237,
		0.994196,
		0.998667,
		0.999903,
		1,
		0.999901,
		0.998603,
		0.993791,
		0.982817,
		0.963319,
		0.933544,
		0.892474,
		0.839836,
		0.776061,
		0.702226,
		0.619989,
		0.531529,
		0.439502,
		0.346987,
		0.25746,
		0.174757,
		0.10305,
		0.046826,
		0.010867,
		0.002844,
		0.020718,
		0.053333,
		0.098696,
		0.154777,
		0.219514,
		0.29083,
		0.366646,
		0.444897,
		0.523558,
		0.600669,
		0.67437,
		0.742944,
		0.804867,
		0.858873,
		0.904028,
		0.939815,
		0.966236,
		0.983896,
		0.994065,
		0.998635,
		0.999901,
		1,
		0.999898,
		0.998569,
		0.993648,
		0.982438,
		0.962541,
		0.932195,
		0.890386,
		0.836868,
		0.772112,
		0.69725,
		0.614003,
		0.524629,
		0.431863,
		0.33888,
		0.249252,
		0.166922,
		0.096175,
		0.041614,
		0.008146,
		0.001628,
		0.01743,
		0.048428,
		0.092596,
		0.147869,
		0.212149,
		0.283322,
		0.359265,
		0.437873,
		0.517077,
		0.594872,
		0.669354,
		0.738759,
		0.801521,
		0.856329,
		0.902212,
		0.938621,
		0.965534,
		0.983546,
		0.99393,
		0.998603,
		0.999898,
		1,
		0.999896,
		0.998535,
		0.993502,
		0.982047,
		0.961743,
		0.930812,
		0.888248,
		0.833833,
		0.76808,
		0.692176,
		0.607912,
		0.517621,
		0.424126,
		0.330696,
		0.241006,
		0.159107,
		0.0894,
		0.036611,
		0.005773,
		0.000737,
		0.014374,
		0.043674,
		0.086581,
		0.140991,
		0.204772,
		0.275766,
		0.351814,
		0.430764,
		0.510503,
		0.588982,
		0.664249,
		0.734494,
		0.798105,
		0.853729,
		0.900354,
		0.937398,
		0.964815,
		0.983187,
		0.993791,
		0.998569,
		0.999896,
		1,
		0.999893,
		0.998499,
		0.99335,
		0.981647,
		0.960924,
		0.929393,
		0.886059,
		0.830729,
		0.763962,
		0.687003,
		0.601712,
		0.510503,
		0.416289,
		0.322436,
		0.232725,
		0.151319,
		0.082738,
		0.031834,
		0.003771,
		0.000188,
		0.011564,
		0.039085,
		0.08066,
		0.134151,
		0.197385,
		0.268167,
		0.344293,
		0.423569,
		0.503836,
		0.582997,
		0.659052,
		0.730146,
		0.794618,
		0.851071,
		0.898452,
		0.936145,
		0.964076,
		0.982817,
		0.993648,
		0.998535,
		0.999893,
		1,
		0.999891,
		0.998463,
		0.993195,
		0.981234,
		0.960083,
		0.927939,
		0.883818,
		0.827555,
		0.759757,
		0.681729,
		0.595403,
		0.503276,
		0.408354,
		0.314104,
		0.224414,
		0.143566,
		0.076202,
		0.027301,
		0.002165,
		0.009016,
		0.034672,
		0.074843,
		0.127354,
		0.189994,
		0.260527,
		0.336705,
		0.416289,
		0.497074,
		0.576915,
		0.653763,
		0.725714,
		0.791059,
		0.848355,
		0.896506,
		0.93486,
		0.963319,
		0.982438,
		0.993502,
		0.998499,
		0.999891,
		1,
		0.999888,
		0.998425,
		0.993035,
		0.980811,
		0.95922,
		0.926448,
		0.881522,
		0.824309,
		0.755463,
		0.676351,
		0.588982,
		0.495938,
		0.40032,
		0.3057,
		0.216079,
		0.135857,
		0.069803,
		0.023032,
		0.000982,
		0.006746,
		0.030448,
		0.069138,
		0.120607,
		0.182603,
		0.252848,
		0.32905,
		0.408924,
		0.490217,
		0.570735,
		0.64838,
		0.721196,
		0.787425,
		0.845578,
		0.894513,
		0.933544,
		0.962541,
		0.982047,
		0.99335,
		0.998463,
		0.999888,
		1,
		0.999885,
		0.998387,
		0.99287,
		0.980375,
		0.958333,
		0.924919,
		0.87917,
		0.820988,
		0.751077,
		0.670868,
		0.582448,
		0.488488,
		0.392188,
		0.297228,
		0.207724,
		0.128201,
		0.063557,
		0.019046,
		0.000251,
		0.004772,
		0.026425,
		0.063557,
		0.113919,
		0.175218,
		0.245133,
		0.321329,
		0.401474,
		0.483264,
		0.564456,
		0.642901,
		0.71659,
		0.783716,
		0.842738,
		0.892474,
		0.932195,
		0.961743,
		0.981647,
		0.993195,
		0.998425,
		0.999885,
		1,
		0.999882,
		0.998347,
		0.9927,
		0.979927,
		0.957423,
		0.92335,
		0.876762,
		0.817592,
		0.746597,
		0.665277,
		0.575799,
		0.480925,
		0.383959,
		0.28869,
		0.199356,
		0.120607,
		0.057479,
		0.015366,
		0.003111,
		0.02262,
		0.058111,
		0.107296,
		0.167843,
		0.237387,
		0.313545,
		0.393939,
		0.476215,
		0.558077,
		0.637323,
		0.711894,
		0.779928,
		0.839836,
		0.890386,
		0.930812,
		0.960924,
		0.981234,
		0.993035,
		0.998387,
		0.999882,
		1,
		0.999879,
		0.998306,
		0.992526,
		0.979466,
		0.956487,
		0.921741,
		0.874294,
		0.814117,
		0.742021,
		0.659576,
		0.569033,
		0.473249,
		0.375633,
		0.280089,
		0.19098,
		0.113087,
		0.051585,
		0.012014,
		0.001783,
		0.019046,
		0.05281,
		0.100747,
		0.160485,
		0.229612,
		0.3057,
		0.38632,
		0.469068,
		0.551596,
		0.631647,
		0.707107,
		0.776061,
		0.836868,
		0.888248,
		0.929393,
		0.960083,
		0.980811,
		0.99287,
		0.998347,
		0.999879,
		1,
		0.999876,
		0.998263,
		0.992346,
		0.978993,
		0.955526,
		0.92009,
		0.871765,
		0.810562,
		0.737347,
		0.653763,
		0.562148,
		0.465459,
		0.367211,
		0.271429,
		0.182603,
		0.105651,
		0.045892,
		0.009016,
		0.000807,
		0.015719,
		0.047667,
		0.094282,
		0.153149,
		0.221812,
		0.297795,
		0.378617,
		0.461824,
		0.545012,
		0.625869,
		0.702226,
		0.772112,
		0.833833,
		0.886059,
		0.927939,
		0.95922,
		0.980375,
		0.9927,
		0.998306,
		0.999876,
		1,
		0.999872,
		0.998219,
		0.99216,
		0.978505,
		0.954539,
		0.918396,
		0.869175,
		0.806924,
		0.732572,
		0.647836,
		0.555144,
		0.457553,
		0.358694,
		0.262714,
		0.174234,
		0.098313,
		0.040417,
		0.006396,
		0.000206,
		0.012657,
		0.042694,
		0.08791,
		0.145843,
		0.213992,
		0.289832,
		0.370832,
		0.454482,
		0.538323,
		0.619989,
		0.69725,
		0.76808,
		0.830729,
		0.883818,
		0.926448,
		0.958333,
		0.979927,
		0.992526,
		0.998263,
		0.999872,
		1,
		0.999869,
		0.998174,
		0.991969,
		0.978003,
		0.953525,
		0.916658,
		0.86652,
		0.803203,
		0.727694,
		0.641793,
		0.548017,
		0.449532,
		0.350084,
		0.253947,
		0.165879,
		0.091084,
		0.035182,
		0.004182,
		0.009877,
		0.037906,
		0.08164,
		0.138573,
		0.206156,
		0.281814,
		0.362964,
		0.447041,
		0.531529,
		0.614003,
		0.692176,
		0.763962,
		0.827555,
		0.881522,
		0.924919,
		0.957423,
		0.979466,
		0.992346,
		0.998219,
		0.999869,
		1,
		0.999866,
		0.998127,
		0.991772,
		0.977487,
		0.952482,
		0.914874,
		0.863799,
		0.799394,
		0.722712,
		0.635631,
		0.540768,
		0.441396,
		0.341382,
		0.245133,
		0.157547,
		0.083978,
		0.030205,
		0.002404,
		0.007397,
		0.033317,
		0.075484,
		0.131346,
		0.198309,
		0.273744,
		0.355015,
		0.439502,
		0.524629,
		0.607912,
		0.687003,
		0.759757,
		0.824309,
		0.87917,
		0.92335,
		0.956487,
		0.978993,
		0.99216,
		0.998174,
		0.999866,
		1,
		0.999862,
		0.998079,
		0.991569,
		0.976956,
		0.95141,
		0.913042,
		0.861011,
		0.795497,
		0.717621,
		0.629348,
		0.533393,
		0.433143,
		0.332591,
		0.236278,
		0.149248,
		0.077011,
		0.02551,
		0.001092,
		0.005236,
		0.028941,
		0.069452,
		0.124172,
		0.190456,
		0.265625,
		0.346987,
		0.431863,
		0.517621,
		0.601712,
		0.681729,
		0.755463,
		0.820988,
		0.876762,
		0.921741,
		0.955526,
		0.978505,
		0.991969,
		0.998127,
		0.999862,
		1,
		0.999858,
		0.998029,
		0.99136,
		0.976409,
		0.950308,
		0.911162,
		0.858152,
		0.791508,
		0.71242,
		0.622942,
		0.525892,
		0.424774,
		0.323712,
		0.227386,
		0.140991,
		0.070199,
		0.021119,
		0.000279,
		0.003417,
		0.024796,
		0.063557,
		0.117059,
		0.182603,
		0.25746,
		0.33888,
		0.424126,
		0.510503,
		0.595403,
		0.676351,
		0.751077,
		0.817592,
		0.874294,
		0.92009,
		0.954539,
		0.978003,
		0.991772,
		0.998079,
		0.999858,
		1,
		0.999854,
		0.997978,
		0.991144,
		0.975845,
		0.949176,
		0.909232,
		0.855222,
		0.787425,
		0.707107,
		0.61641,
		0.518262,
		0.416289,
		0.314747,
		0.218463,
		0.132788,
		0.063557,
		0.017058,
		0.00196,
		0.020897,
		0.057812,
		0.110014,
		0.174757,
		0.249252,
		0.330696,
		0.416289,
		0.503276,
		0.588982,
		0.670868,
		0.746597,
		0.814117,
		0.871765,
		0.918396,
		0.953525,
		0.977487,
		0.991569,
		0.998029,
		0.999854,
		1,
		0.999851,
		0.997925,
		0.990922,
		0.975265,
		0.948011,
		0.90725,
		0.852217,
		0.783246,
		0.701678,
		0.609751,
		0.510503,
		0.407688,
		0.3057,
		0.209516,
		0.124649,
		0.057105,
		0.013353,
		0.000888,
		0.017263,
		0.05223,
		0.10305,
		0.166922,
		0.241006,
		0.322436,
		0.408354,
		0.495938,
		0.582448,
		0.665277,
		0.742021,
		0.810562,
		0.869175,
		0.916658,
		0.952482,
		0.976956,
		0.99136,
		0.997978,
		0.999851,
		1,
		0.999846,
		0.99787,
		0.990693,
		0.974668,
		0.946813,
		0.905215,
		0.849137,
		0.778969,
		0.696131,
		0.602961,
		0.502614,
		0.398972,
		0.296573,
		0.200552,
		0.116587,
		0.050862,
		0.010032,
		0.000227,
		0.013913,
		0.046826,
		0.096175,
		0.159107,
		0.232725,
		0.314104,
		0.40032,
		0.488488,
		0.575799,
		0.659576,
		0.737347,
		0.806924,
		0.86652,
		0.914874,
		0.95141,
		0.976409,
		0.991144,
		0.997925,
		0.999846,
		1,
		0.999842,
		0.997814,
		0.990456,
		0.974052,
		0.94558,
		0.903125,
		0.845978,
		0.77459,
		0.690463,
		0.596039,
		0.494592,
		0.39014,
		0.287371,
		0.191578,
		0.108615,
		0.044848,
		0.007125,
		0.010867,
		0.041614,
		0.0894,
		0.151319,
		0.224414,
		0.3057,
		0.392188,
		0.480925,
		0.569033,
		0.653763,
		0.732572,
		0.803203,
		0.863799,
		0.913042,
		0.950308,
		0.975845,
		0.990922,
		0.99787,
		0.999842,
		1,
		0.999838,
		0.997755,
		0.990212,
		0.973418,
		0.944313,
		0.900978,
		0.842738,
		0.770107,
		0.684672,
		0.588982,
		0.486437,
		0.381194,
		0.278096,
		0.182603,
		0.100747,
		0.039085,
		0.004665,
		0.008146,
		0.036611,
		0.082738,
		0.143566,
		0.216079,
		0.297228,
		0.383959,
		0.473249,
		0.562148,
		0.647836,
		0.727694,
		0.799394,
		0.861011,
		0.911162,
		0.949176,
		0.975265,
		0.990693,
		0.997814,
		0.999838,
		1,
		0.999833,
		0.997695,
		0.98996,
		0.972765,
		0.943008,
		0.898772,
		0.839416,
		0.765517,
		0.678754,
		0.581788,
		0.478147,
		0.372135,
		0.268753,
		0.173636,
		0.093,
		0.033598,
		0.002685,
		0.005773,
		0.031834,
		0.076202,
		0.135857,
		0.207724,
		0.28869,
		0.375633,
		0.465459,
		0.555144,
		0.641793,
		0.722712,
		0.795497,
		0.858152,
		0.909232,
		0.948011,
		0.974668,
		0.990456,
		0.997755,
		0.999833,
		1,
		0.999829,
		0.997633,
		0.9897,
		0.972091,
		0.941666,
		0.896506,
		0.836008,
		0.760817,
		0.672708,
		0.574455,
		0.469722,
		0.362964,
		0.259348,
		0.164687,
		0.085389,
		0.02841,
		0.001221,
		0.003771,
		0.027301,
		0.069803,
		0.128201,
		0.199356,
		0.280089,
		0.367211,
		0.457553,
		0.548017,
		0.635631,
		0.717621,
		0.791508,
		0.855222,
		0.90725,
		0.946813,
		0.974052,
		0.990212,
		0.997695,
		0.999829,
		1,
		0.999824,
		0.997568,
		0.989431,
		0.971397,
		0.940284,
		0.894177,
		0.832511,
		0.756005,
		0.666529,
		0.56698,
		0.461161,
		0.353683,
		0.249885,
		0.155766,
		0.077932,
		0.02355,
		0.000312,
		0.002165,
		0.023032,
		0.063557,
		0.120607,
		0.19098,
		0.271429,
		0.358694,
		0.449532,
		0.540768,
		0.629348,
		0.71242,
		0.787425,
		0.852217,
		0.905215,
		0.94558,
		0.973418,
		0.98996,
		0.997633,
		0.999824,
		1,
		0.999819,
		0.997501,
		0.989154,
		0.970681,
		0.938862,
		0.891784,
		0.828924,
		0.751077,
		0.660215,
		0.559361,
		0.452462,
		0.344293,
		0.24037,
		0.146884,
		0.070648,
		0.019046,
		0.000982,
		0.019046,
		0.057479,
		0.113087,
		0.182603,
		0.262714,
		0.350084,
		0.441396,
		0.533393,
		0.622942,
		0.707107,
		0.783246,
		0.849137,
		0.903125,
		0.944313,
		0.972765,
		0.9897,
		0.997568,
		0.999819,
		1,
		0.999813,
		0.997432,
		0.988867,
		0.969943,
		0.937398,
		0.889324,
		0.825244,
		0.74603,
		0.653763,
		0.551596,
		0.443626,
		0.334797,
		0.23081,
		0.138055,
		0.063557,
		0.014928,
		0.000251,
		0.015366,
		0.051585,
		0.105651,
		0.174234,
		0.253947,
		0.341382,
		0.433143,
		0.525892,
		0.61641,
		0.701678,
		0.778969,
		0.845978,
		0.900978,
		0.943008,
		0.972091,
		0.989431,
		0.997501,
		0.999813,
		1,
		0.999808,
		0.997361,
		0.988571,
		0.969182,
		0.93589,
		0.886795,
		0.821467,
		0.740862,
		0.647171,
		0.543682,
		0.434652,
		0.325198,
		0.221211,
		0.129291,
		0.056682,
		0.011231,
		0.012014,
		0.045892,
		0.098313,
		0.165879,
		0.245133,
		0.332591,
		0.424774,
		0.518262,
		0.609751,
		0.696131,
		0.77459,
		0.842738,
		0.898772,
		0.941666,
		0.971397,
		0.989154,
		0.997432,
		0.999808,
		1,
		0.999802,
		0.997287,
		0.988265,
		0.968397,
		0.934338,
		0.884195,
		0.817592,
		0.735568,
		0.640434,
		0.535618,
		0.42554,
		0.315497,
		0.211582,
		0.120607,
		0.050046,
		0.007987,
		0.009016,
		0.040417,
		0.091084,
		0.157547,
		0.236278,
		0.323712,
		0.416289,
		0.510503,
		0.602961,
		0.690463,
		0.770107,
		0.839416,
		0.896506,
		0.940284,
		0.970681,
		0.988867,
		0.997361,
		0.999802,
		1,
		0.999797,
		0.99721,
		0.987949,
		0.967586,
		0.932739,
		0.881522,
		0.813614,
		0.730146,
		0.63355,
		0.527402,
		0.416289,
		0.3057,
		0.201932,
		0.112019,
		0.043674,
		0.005236,
		0.006396,
		0.035182,
		0.083978,
		0.149248,
		0.227386,
		0.314747,
		0.407688,
		0.502614,
		0.596039,
		0.684672,
		0.765517,
		0.836008,
		0.894177,
		0.938862,
		0.969943,
		0.988571,
		0.997287,
		0.999797,
		1,
		0.99979,
		0.997131,
		0.987623,
		0.96675,
		0.931091,
		0.878773,
		0.809531,
		0.724593,
		0.626516,
		0.519031,
		0.406901,
		0.295809,
		0.192269,
		0.103544,
		0.037594,
		0.003018,
		0.004182,
		0.030205,
		0.077011,
		0.140991,
		0.218463,
		0.3057,
		0.398972,
		0.494592,
		0.588982,
		0.678754,
		0.760817,
		0.832511,
		0.891784,
		0.937398,
		0.969182,
		0.988265,
		0.99721,
		0.99979,
		1,
		0.999784,
		0.997049,
		0.987285,
		0.965887,
		0.929393,
		0.875946,
		0.80534,
		0.718904,
		0.619329,
		0.510503,
		0.397375,
		0.28583,
		0.182603,
		0.0952,
		0.031834,
		0.001374,
		0.002404,
		0.02551,
		0.070199,
		0.132788,
		0.209516,
		0.296573,
		0.39014,
		0.486437,
		0.581788,
		0.672708,
		0.756005,
		0.828924,
		0.889324,
		0.93589,
		0.968397,
		0.987949,
		0.997131,
		0.999784,
		1,
		0.999778,
		0.996964,
		0.986936,
		0.964996,
		0.927644,
		0.873037,
		0.801037,
		0.713077,
		0.611985,
		0.501818,
		0.387712,
		0.275766,
		0.172947,
		0.087007,
		0.026425,
		0.000352,
		0.001092,
		0.021119,
		0.063557,
		0.124649,
		0.200552,
		0.287371,
		0.381194,
		0.478147,
		0.574455,
		0.666529,
		0.751077,
		0.825244,
		0.886795,
		0.934338,
		0.967586,
		0.987623,
		0.997049,
		0.999778,
		1,
		0.999771,
		0.996876,
		0.986574,
		0.964076,
		0.925841,
		0.870045,
		0.796619,
		0.707107,
		0.604481,
		0.492972,
		0.377913,
		0.265625,
		0.163312,
		0.078987,
		0.021402,
		0.000279,
		0.017058,
		0.057105,
		0.116587,
		0.191578,
		0.278096,
		0.372135,
		0.469722,
		0.56698,
		0.660215,
		0.74603,
		0.821467,
		0.884195,
		0.932739,
		0.96675,
		0.987285,
		0.996964,
		0.999771,
		1,
		0.999764,
		0.996784,
		0.986201,
		0.963126,
		0.923982,
		0.866967,
		0.792083,
		0.700991,
		0.596815,
		0.483964,
		0.36798,
		0.255412,
		0.153712,
		0.071162,
		0.0168,
		0.013353,
		0.050862,
		0.108615,
		0.182603,
		0.268753,
		0.362964,
		0.461161,
		0.559361,
		0.653763,
		0.740862,
		0.817592,
		0.881522,
		0.931091,
		0.965887,
		0.986936,
		0.996876,
		0.999764,
		1,
		0.999757,
		0.99669,
		0.985814,
		0.962145,
		0.922066,
		0.863799,
		0.787425,
		0.694725,
		0.588982,
		0.474793,
		0.357915,
		0.245133,
		0.144161,
		0.063557,
		0.012657,
		0.010032,
		0.044848,
		0.100747,
		0.173636,
		0.259348,
		0.353683,
		0.452462,
		0.551596,
		0.647171,
		0.735568,
		0.813614,
		0.878773,
		0.929393,
		0.964996,
		0.986574,
		0.996784,
		0.999757,
		1,
		0.999749,
		0.996591,
		0.985413,
		0.961131,
		0.92009,
		0.860539,
		0.782641,
		0.688306,
		0.58098,
		0.465459,
		0.34772,
		0.234798,
		0.134676,
		0.0562,
		0.009016,
		0.007125,
		0.039085,
		0.093,
		0.164687,
		0.249885,
		0.344293,
		0.443626,
		0.543682,
		0.640434,
		0.730146,
		0.809531,
		0.875946,
		0.927644,
		0.964076,
		0.986201,
		0.99669,
		0.999749,
		1,
		0.999741,
		0.996489,
		0.984999,
		0.960083,
		0.918052,
		0.857183,
		0.777728,
		0.681729,
		0.572806,
		0.455958,
		0.337397,
		0.224414,
		0.125272,
		0.049119,
		0.00592,
		0.004665,
		0.033598,
		0.085389,
		0.155766,
		0.24037,
		0.334797,
		0.434652,
		0.535618,
		0.63355,
		0.724593,
		0.80534,
		0.873037,
		0.925841,
		0.963126,
		0.985814,
		0.996591,
		0.999741,
		0.999999,
		0.999732,
		0.996382,
		0.984568,
		0.958999,
		0.915948,
		0.853728,
		0.77268,
		0.674989,
		0.564456,
		0.446291,
		0.32695,
		0.213992,
		0.11597,
		0.042346,
		0.003417,
		0.002685,
		0.02841,
		0.077932,
		0.146884,
		0.23081,
		0.325198,
		0.42554,
		0.527402,
		0.626516,
		0.718904,
		0.801037,
		0.870045,
		0.923982,
		0.962145,
		0.985413,
		0.996489,
		0.999732,
		0.999994,
		0.999719,
		0.996268,
		0.984119,
		0.957876,
		0.913775,
		0.850168,
		0.767493,
		0.668082,
		0.555925,
		0.436456,
		0.316381,
		0.20354,
		0.106789,
		0.035914,
		0.001559,
		0.001221,
		0.02355,
		0.070648,
		0.138055,
		0.221211,
		0.315497,
		0.416289,
		0.519031,
		0.619329,
		0.713077,
		0.796619,
		0.866967,
		0.922066,
		0.961131,
		0.984999,
		0.996382,
		0.999719,
		0.999982,
		0.999698,
		0.996142,
		0.983646,
		0.956706,
		0.911526,
		0.846495,
		0.762157,
		0.661,
		0.547208,
		0.42645,
		0.305694,
		0.193071,
		0.097751,
		0.02986,
		0.0004,
		0.000312,
		0.019046,
		0.063557,
		0.129291,
		0.211582,
		0.3057,
		0.406901,
		0.510503,
		0.611985,
		0.707107,
		0.792083,
		0.863799,
		0.92009,
		0.960083,
		0.984568,
		0.996268,
		0.999698,
		0.999952,
		0.999659,
		0.995994,
		0.983139,
		0.955481,
		0.909189,
		0.842698,
		0.756663,
		0.653732,
		0.538297,
		0.416269,
		0.294891,
		0.182595,
		0.088879,
		0.024222,
		0.014928,
		0.056682,
		0.120607,
		0.201932,
		0.295809,
		0.397375,
		0.501818,
		0.604481,
		0.700991,
		0.787425,
		0.860539,
		0.918052,
		0.958999,
		0.984119,
		0.996142,
		0.999659,
		0.999887,
		0.999584,
		0.995806,
		0.98258,
		0.95418,
		0.906745,
		0.838759,
		0.750991,
		0.646263,
		0.529181,
		0.405908,
		0.283974,
		0.172124,
		0.0802,
		0.019044,
		0.011231,
		0.050046,
		0.112019,
		0.192269,
		0.28583,
		0.387712,
		0.492972,
		0.596815,
		0.694725,
		0.782641,
		0.857183,
		0.915948,
		0.957876,
		0.983646,
		0.995994,
		0.999584,
		0.999758,
		0.999445,
		0.99555,
		0.981941,
		0.952776,
		0.904167,
		0.83465,
		0.745119,
		0.638572,
		0.519843,
		0.395357,
		0.272942,
		0.16167,
		0.071739,
		0.01437,
		0.007987,
		0.043674,
		0.103544,
		0.182603,
		0.275766,
		0.377913,
		0.483964,
		0.588982,
		0.688306,
		0.777728,
		0.853728,
		0.913775,
		0.956706,
		0.983139,
		0.995806,
		0.999445,
		0.999523,
		0.9992,
		0.995184,
		0.981178,
		0.951227,
		0.901414,
		0.830333,
		0.739009,
		0.630629,
		0.51026,
		0.384603,
		0.261794,
		0.151247,
		0.063527,
		0.010251,
		0.005236,
		0.037594,
		0.0952,
		0.172947,
		0.265625,
		0.36798,
		0.474793,
		0.58098,
		0.681729,
		0.77268,
		0.850168,
		0.911526,
		0.955481,
		0.98258,
		0.99555,
		0.9992,
		0.999123,
		0.998789,
		0.994648,
		0.980234,
		0.949475,
		0.898429,
		0.825756,
		0.732617,
		0.622395,
		0.500403,
		0.373628,
		0.250527,
		0.140868,
		0.055596,
		0.00674,
		0.003018,
		0.031834,
		0.087007,
		0.163312,
		0.255412,
		0.357915,
		0.465459,
		0.572806,
		0.674989,
		0.767493,
		0.846495,
		0.909189,
		0.95418,
		0.981941,
		0.995184,
		0.998789,
		0.998475,
		0.99813,
		0.993859,
		0.979026,
		0.94744,
		0.895138,
		0.820849,
		0.72588,
		0.61382,
		0.490235,
		0.362411,
		0.239137,
		0.130547,
		0.047983,
		0.003895,
		0.001374,
		0.026425,
		0.078987,
		0.153712,
		0.245133,
		0.34772,
		0.455958,
		0.564456,
		0.668082,
		0.762157,
		0.842698,
		0.906745,
		0.952776,
		0.981178,
		0.994648,
		0.99813,
		0.997474,
		0.997118,
		0.992715,
		0.977452,
		0.945023,
		0.891446,
		0.815526,
		0.718723,
		0.604841,
		0.479711,
		0.350925,
		0.227619,
		0.120303,
		0.040727,
		0.001778,
		0.000352,
		0.021402,
		0.071162,
		0.144161,
		0.234798,
		0.337397,
		0.446291,
		0.555925,
		0.661,
		0.756663,
		0.838759,
		0.904167,
		0.951227,
		0.980234,
		0.993859,
		0.997118,
		0.99599,
		0.995622,
		0.991083,
		0.975383,
		0.942099,
		0.887236,
		0.809681,
		0.711055,
		0.595385,
		0.468778,
		0.339142,
		0.215968,
		0.110155,
		0.033873,
		0.000457,
		0.0168,
		0.063557,
		0.134676,
		0.224414,
		0.32695,
		0.436456,
		0.547208,
		0.653732,
		0.750991,
		0.83465,
		0.901414,
		0.949475,
		0.979026,
		0.992715,
		0.995622,
		0.993865,
		0.993485,
		0.988808,
		0.972665,
		0.938519,
		0.882368,
		0.803188,
		0.702768,
		0.585368,
		0.457378,
		0.327031,
		0.204182,
		0.100129,
		0.027471,
		0.012657,
		0.0562,
		0.125272,
		0.213992,
		0.316381,
		0.42645,
		0.538297,
		0.646263,
		0.745119,
		0.830333,
		0.898429,
		0.94744,
		0.977452,
		0.991083,
		0.993485,
		0.990914,
		0.990522,
		0.985706,
		0.969117,
		0.934111,
		0.876682,
		0.795905,
		0.693744,
		0.574699,
		0.445448,
		0.314561,
		0.192263,
		0.090256,
		0.021576,
		0.009016,
		0.049119,
		0.11597,
		0.20354,
		0.305694,
		0.416269,
		0.529181,
		0.638572,
		0.739009,
		0.825756,
		0.895138,
		0.945023,
		0.975383,
		0.988808,
		0.990522,
		0.986931,
		0.986527,
		0.981571,
		0.964539,
		0.928683,
		0.870001,
		0.787675,
		0.683852,
		0.563276,
		0.432925,
		0.301705,
		0.180217,
		0.080573,
		0.016249,
		0.00592,
		0.042346,
		0.106789,
		0.193071,
		0.294891,
		0.405908,
		0.519843,
		0.630629,
		0.732617,
		0.820849,
		0.891446,
		0.942099,
		0.972665,
		0.985706,
		0.986527,
		0.981688,
		0.981272,
		0.976177,
		0.958709,
		0.922022,
		0.862132,
		0.778329,
		0.672954,
		0.551001,
		0.419747,
		0.28844,
		0.168059,
		0.071125,
		0.011555,
		0.003417,
		0.035914,
		0.097751,
		0.182595,
		0.283974,
		0.395357,
		0.51026,
		0.622395,
		0.72588,
		0.815526,
		0.887236,
		0.938519,
		0.969117,
		0.981571,
		0.981272,
		0.974941,
		0.974514,
		0.969283,
		0.951392,
		0.913908,
		0.852871,
		0.767694,
		0.660911,
		0.537774,
		0.405858,
		0.274752,
		0.155812,
		0.061965,
		0.007563,
		0.001559,
		0.02986,
		0.088879,
		0.172124,
		0.272942,
		0.384603,
		0.500403,
		0.61382,
		0.718723,
		0.809681,
		0.882368,
		0.934111,
		0.964539,
		0.976177,
		0.974514,
		0.966443,
		0.966004,
		0.960643,
		0.942348,
		0.904114,
		0.842016,
		0.755596,
		0.64759,
		0.523504,
		0.391212,
		0.260641,
		0.143513,
		0.053154,
		0.004344,
		0.0004,
		0.024222,
		0.0802,
		0.16167,
		0.261794,
		0.373628,
		0.490235,
		0.604841,
		0.711055,
		0.803188,
		0.876682,
		0.928683,
		0.958709,
		0.969283,
		0.966004,
		0.955945,
		0.955494,
		0.950009,
		0.931338,
		0.892416,
		0.829368,
		0.741871,
		0.632867,
		0.508113,
		0.375779,
		0.246117,
		0.131208,
		0.044763,
		0.001968,
		0.019044,
		0.071739,
		0.151247,
		0.250527,
		0.362411,
		0.479711,
		0.595385,
		0.702768,
		0.795905,
		0.870001,
		0.922022,
		0.951392,
		0.960643,
		0.955494,
		0.943208,
		0.942748,
		0.937149,
		0.918136,
		0.878606,
		0.814743,
		0.726371,
		0.616635,
		0.491542,
		0.359546,
		0.231212,
		0.118961,
		0.036866,
		0.0005,
		0.01437,
		0.063527,
		0.140868,
		0.239137,
		0.350925,
		0.468778,
		0.585368,
		0.693744,
		0.787675,
		0.862132,
		0.913908,
		0.942348,
		0.950009,
		0.942748,
		0.928019,
		0.927549,
		0.921848,
		0.902538,
		0.862495,
		0.797981,
		0.708972,
		0.598814,
		0.473757,
		0.342523,
		0.215973,
		0.106843,
		0.029542,
		0.010251,
		0.055596,
		0.130547,
		0.227619,
		0.339142,
		0.457378,
		0.574699,
		0.683852,
		0.778329,
		0.852871,
		0.904114,
		0.931338,
		0.937149,
		0.927549,
		0.910193,
		0.909715,
		0.903927,
		0.884372,
		0.84393,
		0.778956,
		0.689582,
		0.579355,
		0.454752,
		0.324747,
		0.20047,
		0.094942,
		0.022875,
		0.00674,
		0.047983,
		0.120303,
		0.215968,
		0.327031,
		0.445448,
		0.563276,
		0.672954,
		0.767694,
		0.842016,
		0.892416,
		0.918136,
		0.921848,
		0.909715,
		0.88959,
		0.889104,
		0.883247,
		0.863508,
		0.822798,
		0.757581,
		0.66815,
		0.558244,
		0.434554,
		0.30628,
		0.184789,
		0.083354,
		0.016943,
		0.003895,
		0.040727,
		0.110155,
		0.204182,
		0.314561,
		0.432925,
		0.551001,
		0.660911,
		0.755596,
		0.829368,
		0.878606,
		0.902538,
		0.903927,
		0.889104,
		0.86612,
		0.865629,
		0.859722,
		0.83987,
		0.799039,
		0.733821,
		0.644671,
		0.535513,
		0.413227,
		0.287213,
		0.169039,
		0.072182,
		0.01182,
		0.001778,
		0.033873,
		0.100129,
		0.192263,
		0.301705,
		0.419747,
		0.537774,
		0.64759,
		0.741871,
		0.814743,
		0.862495,
		0.884372,
		0.883247,
		0.865629,
		0.839755,
		0.839261,
		0.833327,
		0.813438,
		0.77265,
		0.707694,
		0.61919,
		0.511233,
		0.390871,
		0.267665,
		0.153342,
		0.061533,
		0.007571,
		0.000457,
		0.027471,
		0.090256,
		0.180217,
		0.28844,
		0.405858,
		0.523504,
		0.632867,
		0.726371,
		0.797981,
		0.84393,
		0.863508,
		0.859722,
		0.839261,
		0.810531,
		0.810035,
		0.8041,
		0.784259,
		0.743688,
		0.679276,
		0.591806,
		0.485526,
		0.367622,
		0.247779,
		0.137834,
		0.051515,
		0.004244,
		0.021576,
		0.080573,
		0.168059,
		0.274752,
		0.391212,
		0.508113,
		0.616635,
		0.708972,
		0.778956,
		0.822798,
		0.83987,
		0.833327,
		0.810035,
		0.778554,
		0.778059,
		0.772148,
		0.752446,
		0.712278,
		0.648705,
		0.56267,
		0.458554,
		0.34365,
		0.227721,
		0.122659,
		0.042229,
		0.001872,
		0.016249,
		0.071125,
		0.155812,
		0.260641,
		0.375779,
		0.491542,
		0.598814,
		0.689582,
		0.757581,
		0.799039,
		0.813438,
		0.8041,
		0.778059,
		0.743999,
		0.743507,
		0.737649,
		0.718179,
		0.678608,
		0.616177,
		0.531985,
		0.430524,
		0.319156,
		0.207672,
		0.107965,
		0.033766,
		0.000462,
		0.011555,
		0.061965,
		0.143513,
		0.246117,
		0.359546,
		0.473757,
		0.579355,
		0.66815,
		0.733821,
		0.77265,
		0.784259,
		0.772148,
		0.743507,
		0.707107,
		0.706621,
		0.700845,
		0.681705,
		0.642924,
		0.581942,
		0.5,
		0.401677,
		0.294361,
		0.187825,
		0.093895,
		0.026203,
		0.007563,
		0.053154,
		0.131208,
		0.231212,
		0.342523,
		0.454752,
		0.558244,
		0.644671,
		0.707694,
		0.743688,
		0.752446,
		0.737649,
		0.706621,
		0.668181,
		0.667704,
		0.662039,
		0.643327,
		0.605532,
		0.546299,
		0.467005,
		0.372282,
		0.269507,
		0.168377,
		0.080587,
		0.019598,
		0.004344,
		0.044763,
		0.118961,
		0.215973,
		0.324747,
		0.434554,
		0.535513,
		0.61919,
		0.679276,
		0.712278,
		0.718179,
		0.700845,
		0.667704,
		0.627578,
		0.627111,
		0.621588,
		0.603399,
		0.566781,
		0.509589,
		0.433319,
		0.342631,
		0.244843,
		0.14952,
		0.068164,
		0.013986,
		0.001968,
		0.036866,
		0.106843,
		0.20047,
		0.30628,
		0.413227,
		0.511233,
		0.591806,
		0.648705,
		0.678608,
		0.681705,
		0.662039,
		0.627111,
		0.585696,
		0.585242,
		0.57989,
		0.562317,
		0.527057,
		0.47218,
		0.399286,
		0.313028,
		0.220619,
		0.131439,
		0.056729,
		0.009378,
		0.0005,
		0.029542,
		0.094942,
		0.184789,
		0.287213,
		0.390871,
		0.485526,
		0.56267,
		0.616177,
		0.642924,
		0.643327,
		0.621588,
		0.585242,
		0.542966,
		0.542528,
		0.537373,
		0.520504,
		0.486772,
		0.434461,
		0.365257,
		0.283774,
		0.197077,
		0.1143,
		0.046363,
		0.005759,
		0.022875,
		0.083354,
		0.169039,
		0.267665,
		0.367622,
		0.458554,
		0.531985,
		0.581942,
		0.605532,
		0.603399,
		0.57989,
		0.542528,
		0.499836,
		0.499416,
		0.494485,
		0.4784,
		0.446348,
		0.396826,
		0.331586,
		0.255168,
		0.174444,
		0.09825,
		0.037121,
		0.003088,
		0.016943,
		0.072182,
		0.153342,
		0.247779,
		0.34365,
		0.430524,
		0.5,
		0.546299,
		0.566781,
		0.562317,
		0.537373,
		0.499416,
		0.456761,
		0.45636,
		0.451676,
		0.436447,
		0.406208,
		0.359665,
		0.298613,
		0.227488,
		0.152922,
		0.083406,
		0.029031,
		0.001299,
		0.01182,
		0.061533,
		0.137834,
		0.227721,
		0.319156,
		0.401677,
		0.467005,
		0.509589,
		0.527057,
		0.520504,
		0.494485,
		0.45636,
		0.414184,
		0.413805,
		0.409388,
		0.395076,
		0.366762,
		0.32335,
		0.26666,
		0.200987,
		0.132688,
		0.069857,
		0.02209,
		0.000305,
		0.007571,
		0.051515,
		0.122659,
		0.207672,
		0.294361,
		0.372282,
		0.433319,
		0.47218,
		0.486772,
		0.4784,
		0.451676,
		0.413805,
		0.372533,
		0.372177,
		0.368043,
		0.354699,
		0.328396,
		0.288229,
		0.236018,
		0.175889,
		0.113883,
		0.057659,
		0.01627,
		0.004244,
		0.042229,
		0.107965,
		0.187825,
		0.269507,
		0.342631,
		0.399286,
		0.434461,
		0.446348,
		0.436447,
		0.409388,
		0.372177,
		0.332201,
		0.33187,
		0.328034,
		0.315694,
		0.291463,
		0.254614,
		0.206942,
		0.152377,
		0.096614,
		0.046838,
		0.011518,
		0.001872,
		0.033766,
		0.093895,
		0.168377,
		0.244843,
		0.313028,
		0.365257,
		0.396826,
		0.406208,
		0.395076,
		0.368043,
		0.33187,
		0.293547,
		0.293242,
		0.289712,
		0.278401,
		0.256277,
		0.222774,
		0.179646,
		0.130598,
		0.08095,
		0.037384,
		0.007757,
		0.000462,
		0.026203,
		0.080587,
		0.14952,
		0.220619,
		0.283774,
		0.331586,
		0.359665,
		0.366762,
		0.354699,
		0.328034,
		0.293242,
		0.25688,
		0.256601,
		0.253384,
		0.243113,
		0.223105,
		0.192937,
		0.1543,
		0.110655,
		0.066924,
		0.029263,
		0.004893,
		0.019598,
		0.068164,
		0.131439,
		0.197077,
		0.255168,
		0.298613,
		0.32335,
		0.328396,
		0.315694,
		0.289712,
		0.256601,
		0.222461,
		0.222209,
		0.219306,
		0.210073,
		0.192162,
		0.165276,
		0.131026,
		0.092608,
		0.054533,
		0.022412,
		0.002816,
		0.013986,
		0.056729,
		0.1143,
		0.174444,
		0.227488,
		0.26666,
		0.288229,
		0.291463,
		0.278401,
		0.253384,
		0.222209,
		0.190498,
		0.190271,
		0.18768,
		0.179472,
		0.163614,
		0.13992,
		0.109901,
		0.07648,
		0.043741,
		0.016747,
		0.001409,
		0.009378,
		0.046363,
		0.09825,
		0.152922,
		0.200987,
		0.236018,
		0.254614,
		0.256277,
		0.243113,
		0.219306,
		0.190271,
		0.161143,
		0.160942,
		0.158656,
		0.151444,
		0.137572,
		0.116944,
		0.090958,
		0.062253,
		0.034483,
		0.012164,
		0.000551,
		0.005759,
		0.037121,
		0.083406,
		0.132688,
		0.175889,
		0.206942,
		0.222774,
		0.223105,
		0.210073,
		0.18768,
		0.160942,
		0.134496,
		0.13432,
		0.13233,
		0.126076,
		0.1141,
		0.096378,
		0.074187,
		0.049875,
		0.026672,
		0.008548,
		0.000119,
		0.003088,
		0.029031,
		0.069857,
		0.113883,
		0.152377,
		0.179646,
		0.192937,
		0.192162,
		0.179472,
		0.158656,
		0.13432,
		0.110603,
		0.110452,
		0.108744,
		0.103399,
		0.09321,
		0.078208,
		0.05954,
		0.039266,
		0.020197,
		0.005777,
		0.001299,
		0.02209,
		0.057659,
		0.096614,
		0.130598,
		0.1543,
		0.165276,
		0.163614,
		0.151444,
		0.13233,
		0.110452,
		0.089465,
		0.089337,
		0.087893,
		0.083398,
		0.07487,
		0.062379,
		0.046936,
		0.030318,
		0.014934,
		0.003723,
		0.000305,
		0.01627,
		0.046838,
		0.08095,
		0.110655,
		0.131026,
		0.13992,
		0.137572,
		0.126076,
		0.108744,
		0.089337,
		0.071031,
		0.070925,
		0.069727,
		0.066016,
		0.059008,
		0.048799,
		0.036262,
		0.022903,
		0.010748,
		0.002261,
		0.011518,
		0.037384,
		0.066924,
		0.092608,
		0.109901,
		0.116944,
		0.1141,
		0.103399,
		0.087893,
		0.070925,
		0.055213,
		0.055126,
		0.054154,
		0.051152,
		0.045513,
		0.037343,
		0.027382,
		0.016879,
		0.007501,
		0.001272,
		0.007757,
		0.029263,
		0.054533,
		0.07648,
		0.090958,
		0.096378,
		0.09321,
		0.083398,
		0.069727,
		0.055126,
		0.041882,
		0.041813,
		0.041042,
		0.038672,
		0.034243,
		0.027863,
		0.020142,
		0.012091,
		0.005051,
		0.000644,
		0.004893,
		0.022412,
		0.043741,
		0.062253,
		0.074187,
		0.078208,
		0.07487,
		0.066016,
		0.054154,
		0.041813,
		0.030877,
		0.030824,
		0.030229,
		0.02841,
		0.025028,
		0.020186,
		0.014372,
		0.008381,
		0.003262,
		0.000278,
		0.002816,
		0.016747,
		0.034483,
		0.049875,
		0.05954,
		0.062379,
		0.059008,
		0.051152,
		0.041042,
		0.030824,
		0.022008,
		0.021968,
		0.021524,
		0.020174,
		0.017677,
		0.014124,
		0.009893,
		0.005589,
		0.002005,
		9.2e-005,
		0.001409,
		0.012164,
		0.026672,
		0.039266,
		0.046936,
		0.048799,
		0.045513,
		0.038672,
		0.030229,
		0.021968,
		0.015059,
		0.01503,
		0.014712,
		0.01375,
		0.01198,
		0.009478,
		0.006523,
		0.003558,
		0.00116,
		1.6e-005,
		0.000551,
		0.008548,
		0.020197,
		0.030318,
		0.036262,
		0.037343,
		0.034243,
		0.02841,
		0.021524,
		0.01503,
		0.0098,
		0.00978,
		0.009563,
		0.00891,
		0.007717,
		0.006041,
		0.00408,
		0.002141,
		0.000623,
		0.000119,
		0.005777,
		0.014934,
		0.022903,
		0.027382,
		0.027863,
		0.025028,
		0.020174,
		0.014712,
		0.00978,
		0.005985,
		0.005972,
		0.005833,
		0.005418,
		0.004662,
		0.003609,
		0.002388,
		0.0012,
		0.000304,
		0.003723,
		0.010748,
		0.016879,
		0.020142,
		0.020186,
		0.017677,
		0.01375,
		0.009563,
		0.005972,
		0.003364,
		0.003356,
		0.003275,
		0.003031,
		0.002591,
		0.001981,
		0.001282,
		0.000614,
		0.000131,
		0.002261,
		0.007501,
		0.012091,
		0.014372,
		0.014124,
		0.01198,
		0.00891,
		0.005833,
		0.003356,
		0.001689,
		0.001685,
		0.001641,
		0.001514,
		0.001285,
		0.00097,
		0.000613,
		0.000278,
		4.8e-005,
		0.001272,
		0.005051,
		0.008381,
		0.009893,
		0.009478,
		0.007717,
		0.005418,
		0.003275,
		0.001685,
		0.000719,
		0.000718,
		0.000698,
		0.000642,
		0.00054,
		0.000402,
		0.000248,
		0.000106,
		1.4e-005,
		0.000644,
		0.003262,
		0.005589,
		0.006523,
		0.006041,
		0.004662,
		0.003031,
		0.001641,
		0.000718,
		0.000237,
		0.000236,
		0.000229,
		0.00021,
		0.000175,
		0.000129,
		7.7e-005,
		3.1e-005,
		0.000278,
		0.002005,
		0.003558,
		0.00408,
		0.003609,
		0.002591,
		0.001514,
		0.000698,
		0.000236,
		4.9e-005,
		4.8e-005,
		4.7e-005,
		4.3e-005,
		3.5e-005,
		2.6e-005,
		1.5e-005,
		9.2e-005,
		0.00116,
		0.002141,
		0.002388,
		0.001981,
		0.001285,
		0.000642,
		0.000229,
		4.8e-005,
		1.6e-005,
		0.000623,
		0.0012,
		0.001282,
		0.00097,
		0.00054,
		0.00021,
		4.7e-005,
		0.000304,
		0.000614,
		0.000613,
		0.000402,
		0.000175,
		4.3e-005,
		0.000131,
		0.000278,
		0.000248,
		0.000129,
		3.5e-005,
		4.8e-005,
		0.000106,
		7.7e-005,
		2.6e-005,
		1.4e-005,
		3.1e-005,
		1.5e-005,
		1.5e-005,
		3.1e-005,
		1.4e-005,
		2.6e-005,
		7.7e-005,
		0.000106,
		4.8e-005,
		3.5e-005,
		0.000129,
		0.000248,
		0.000278,
		0.000131,
		4.3e-005,
		0.000175,
		0.000402,
		0.000613,
		0.000614,
		0.000304,
		4.7e-005,
		0.00021,
		0.00054,
		0.00097,
		0.001282,
		0.0012,
		0.000623,
		1.6e-005,
		4.8e-005,
		0.000229,
		0.000642,
		0.001285,
		0.001981,
		0.002388,
		0.002141,
		0.00116,
		9.2e-005,
		1.5e-005,
		2.6e-005,
		3.5e-005,
		4.3e-005,
		4.7e-005,
		4.8e-005,
		4.9e-005,
		0.000236,
		0.000698,
		0.001514,
		0.002591,
		0.003609,
		0.00408,
		0.003558,
		0.002005,
		0.000278,
		3.1e-005,
		7.7e-005,
		0.000129,
		0.000175,
		0.00021,
		0.000229,
		0.000236,
		0.000237,
		0.000718,
		0.001641,
		0.003031,
		0.004662,
		0.006041,
		0.006523,
		0.005589,
		0.003262,
		0.000644,
		1.4e-005,
		0.000106,
		0.000248,
		0.000402,
		0.00054,
		0.000642,
		0.000698,
		0.000718,
		0.000719,
		0.001685,
		0.003275,
		0.005418,
		0.007717,
		0.009478,
		0.009893,
		0.008381,
		0.005051,
		0.001272,
		4.8e-005,
		0.000278,
		0.000613,
		0.00097,
		0.001285,
		0.001514,
		0.001641,
		0.001685,
		0.001689,
		0.003356,
		0.005833,
		0.00891,
		0.01198,
		0.014124,
		0.014372,
		0.012091,
		0.007501,
		0.002261,
		0.000131,
		0.000614,
		0.001282,
		0.001981,
		0.002591,
		0.003031,
		0.003275,
		0.003356,
		0.003364,
		0.005972,
		0.009563,
		0.01375,
		0.017677,
		0.020186,
		0.020142,
		0.016879,
		0.010748,
		0.003723,
		0.000304,
		0.0012,
		0.002388,
		0.003609,
		0.004662,
		0.005418,
		0.005833,
		0.005972,
		0.005985,
		0.00978,
		0.014712,
		0.020174,
		0.025028,
		0.027863,
		0.027382,
		0.022903,
		0.014934,
		0.005777,
		0.000119,
		0.000623,
		0.002141,
		0.00408,
		0.006041,
		0.007717,
		0.00891,
		0.009563,
		0.00978,
		0.0098,
		0.01503,
		0.021524,
		0.02841,
		0.034243,
		0.037343,
		0.036262,
		0.030318,
		0.020197,
		0.008548,
		0.000551,
		1.6e-005,
		0.00116,
		0.003558,
		0.006523,
		0.009478,
		0.01198,
		0.01375,
		0.014712,
		0.01503,
		0.015059,
		0.021968,
		0.030229,
		0.038672,
		0.045513,
		0.048799,
		0.046936,
		0.039266,
		0.026672,
		0.012164,
		0.001409,
		9.2e-005,
		0.002005,
		0.005589,
		0.009893,
		0.014124,
		0.017677,
		0.020174,
		0.021524,
		0.021968,
		0.022008,
		0.030824,
		0.041042,
		0.051152,
		0.059008,
		0.062379,
		0.05954,
		0.049875,
		0.034483,
		0.016747,
		0.002816,
		0.000278,
		0.003262,
		0.008381,
		0.014372,
		0.020186,
		0.025028,
		0.02841,
		0.030229,
		0.030824,
		0.030877,
		0.041813,
		0.054154,
		0.066016,
		0.07487,
		0.078208,
		0.074187,
		0.062253,
		0.043741,
		0.022412,
		0.004893,
		0.000644,
		0.005051,
		0.012091,
		0.020142,
		0.027863,
		0.034243,
		0.038672,
		0.041042,
		0.041813,
		0.041882,
		0.055126,
		0.069727,
		0.083398,
		0.09321,
		0.096378,
		0.090958,
		0.07648,
		0.054533,
		0.029263,
		0.007757,
		0.001272,
		0.007501,
		0.016879,
		0.027382,
		0.037343,
		0.045513,
		0.051152,
		0.054154,
		0.055126,
		0.055213,
		0.070925,
		0.087893,
		0.103399,
		0.1141,
		0.116944,
		0.109901,
		0.092608,
		0.066924,
		0.037384,
		0.011518,
		0.002261,
		0.010748,
		0.022903,
		0.036262,
		0.048799,
		0.059008,
		0.066016,
		0.069727,
		0.070925,
		0.071031,
		0.089337,
		0.108744,
		0.126076,
		0.137572,
		0.13992,
		0.131026,
		0.110655,
		0.08095,
		0.046838,
		0.01627,
		0.000305,
		0.003723,
		0.014934,
		0.030318,
		0.046936,
		0.062379,
		0.07487,
		0.083398,
		0.087893,
		0.089337,
		0.089465,
		0.110452,
		0.13233,
		0.151444,
		0.163614,
		0.165276,
		0.1543,
		0.130598,
		0.096614,
		0.057659,
		0.02209,
		0.001299,
		0.005777,
		0.020197,
		0.039266,
		0.05954,
		0.078208,
		0.09321,
		0.103399,
		0.108744,
		0.110452,
		0.110603,
		0.13432,
		0.158656,
		0.179472,
		0.192162,
		0.192937,
		0.179646,
		0.152377,
		0.113883,
		0.069857,
		0.029031,
		0.003088,
		0.000119,
		0.008548,
		0.026672,
		0.049875,
		0.074187,
		0.096378,
		0.1141,
		0.126076,
		0.13233,
		0.13432,
		0.134496,
		0.160942,
		0.18768,
		0.210073,
		0.223105,
		0.222774,
		0.206942,
		0.175889,
		0.132688,
		0.083406,
		0.037121,
		0.005759,
		0.000551,
		0.012164,
		0.034483,
		0.062253,
		0.090958,
		0.116944,
		0.137572,
		0.151444,
		0.158656,
		0.160942,
		0.161143,
		0.190271,
		0.219306,
		0.243113,
		0.256277,
		0.254614,
		0.236018,
		0.200987,
		0.152922,
		0.09825,
		0.046363,
		0.009378,
		0.001409,
		0.016747,
		0.043741,
		0.07648,
		0.109901,
		0.13992,
		0.163614,
		0.179472,
		0.18768,
		0.190271,
		0.190498,
		0.222209,
		0.253384,
		0.278401,
		0.291463,
		0.288229,
		0.26666,
		0.227488,
		0.174444,
		0.1143,
		0.056729,
		0.013986,
		0.002816,
		0.022412,
		0.054533,
		0.092608,
		0.131026,
		0.165276,
		0.192162,
		0.210073,
		0.219306,
		0.222209,
		0.222461,
		0.256601,
		0.289712,
		0.315694,
		0.328396,
		0.32335,
		0.298613,
		0.255168,
		0.197077,
		0.131439,
		0.068164,
		0.019598,
		0.004893,
		0.029263,
		0.066924,
		0.110655,
		0.1543,
		0.192937,
		0.223105,
		0.243113,
		0.253384,
		0.256601,
		0.25688,
		0.293242,
		0.328034,
		0.354699,
		0.366762,
		0.359665,
		0.331586,
		0.283774,
		0.220619,
		0.14952,
		0.080587,
		0.026203,
		0.000462,
		0.007757,
		0.037384,
		0.08095,
		0.130598,
		0.179646,
		0.222774,
		0.256277,
		0.278401,
		0.289712,
		0.293242,
		0.293547,
		0.33187,
		0.368043,
		0.395076,
		0.406208,
		0.396826,
		0.365257,
		0.313028,
		0.244843,
		0.168377,
		0.093895,
		0.033766,
		0.001872,
		0.011518,
		0.046838,
		0.096614,
		0.152377,
		0.206942,
		0.254614,
		0.291463,
		0.315694,
		0.328034,
		0.33187,
		0.332201,
		0.372177,
		0.409388,
		0.436447,
		0.446348,
		0.434461,
		0.399286,
		0.342631,
		0.269507,
		0.187825,
		0.107965,
		0.042229,
		0.004244,
		0.01627,
		0.057659,
		0.113883,
		0.175889,
		0.236018,
		0.288229,
		0.328396,
		0.354699,
		0.368043,
		0.372177,
		0.372533,
		0.413805,
		0.451676,
		0.4784,
		0.486772,
		0.47218,
		0.433319,
		0.372282,
		0.294361,
		0.207672,
		0.122659,
		0.051515,
		0.007571,
		0.000305,
		0.02209,
		0.069857,
		0.132688,
		0.200987,
		0.26666,
		0.32335,
		0.366762,
		0.395076,
		0.409388,
		0.413805,
		0.414184,
		0.45636,
		0.494485,
		0.520504,
		0.527057,
		0.509589,
		0.467005,
		0.401677,
		0.319156,
		0.227721,
		0.137834,
		0.061533,
		0.01182,
		0.001299,
		0.029031,
		0.083406,
		0.152922,
		0.227488,
		0.298613,
		0.359665,
		0.406208,
		0.436447,
		0.451676,
		0.45636,
		0.456761,
		0.499416,
		0.537373,
		0.562317,
		0.566781,
		0.546299,
		0.5,
		0.430524,
		0.34365,
		0.247779,
		0.153342,
		0.072182,
		0.016943,
		0.003088,
		0.037121,
		0.09825,
		0.174444,
		0.255168,
		0.331586,
		0.396826,
		0.446348,
		0.4784,
		0.494485,
		0.499416,
		0.499836,
		0.542528,
		0.57989,
		0.603399,
		0.605532,
		0.581942,
		0.531985,
		0.458554,
		0.367622,
		0.267665,
		0.169039,
		0.083354,
		0.022875,
		0.005759,
		0.046363,
		0.1143,
		0.197077,
		0.283774,
		0.365257,
		0.434461,
		0.486772,
		0.520504,
		0.537373,
		0.542528,
		0.542966,
		0.585242,
		0.621588,
		0.643327,
		0.642924,
		0.616177,
		0.56267,
		0.485526,
		0.390871,
		0.287213,
		0.184789,
		0.094942,
		0.029542,
		0.0005,
		0.009378,
		0.056729,
		0.131439,
		0.220619,
		0.313028,
		0.399286,
		0.47218,
		0.527057,
		0.562317,
		0.57989,
		0.585242,
		0.585696,
		0.627111,
		0.662039,
		0.681705,
		0.678608,
		0.648705,
		0.591806,
		0.511233,
		0.413227,
		0.30628,
		0.20047,
		0.106843,
		0.036866,
		0.001968,
		0.013986,
		0.068164,
		0.14952,
		0.244843,
		0.342631,
		0.433319,
		0.509589,
		0.566781,
		0.603399,
		0.621588,
		0.627111,
		0.627578,
		0.667704,
		0.700845,
		0.718179,
		0.712278,
		0.679276,
		0.61919,
		0.535513,
		0.434554,
		0.324747,
		0.215973,
		0.118961,
		0.044763,
		0.004344,
		0.019598,
		0.080587,
		0.168377,
		0.269507,
		0.372282,
		0.467005,
		0.546299,
		0.605532,
		0.643327,
		0.662039,
		0.667704,
		0.668181,
		0.706621,
		0.737649,
		0.752446,
		0.743688,
		0.707694,
		0.644671,
		0.558244,
		0.454752,
		0.342523,
		0.231212,
		0.131208,
		0.053154,
		0.007563,
		0.026203,
		0.093895,
		0.187825,
		0.294361,
		0.401677,
		0.5,
		0.581942,
		0.642924,
		0.681705,
		0.700845,
		0.706621,
		0.707107,
		0.743507,
		0.772148,
		0.784259,
		0.77265,
		0.733821,
		0.66815,
		0.579355,
		0.473757,
		0.359546,
		0.246117,
		0.143513,
		0.061965,
		0.011555,
		0.000462,
		0.033766,
		0.107965,
		0.207672,
		0.319156,
		0.430524,
		0.531985,
		0.616177,
		0.678608,
		0.718179,
		0.737649,
		0.743507,
		0.743999,
		0.778059,
		0.8041,
		0.813438,
		0.799039,
		0.757581,
		0.689582,
		0.598814,
		0.491542,
		0.375779,
		0.260641,
		0.155812,
		0.071125,
		0.016249,
		0.001872,
		0.042229,
		0.122659,
		0.227721,
		0.34365,
		0.458554,
		0.56267,
		0.648705,
		0.712278,
		0.752446,
		0.772148,
		0.778059,
		0.778554,
		0.810035,
		0.833327,
		0.83987,
		0.822798,
		0.778956,
		0.708972,
		0.616635,
		0.508113,
		0.391212,
		0.274752,
		0.168059,
		0.080573,
		0.021576,
		0.004244,
		0.051515,
		0.137834,
		0.247779,
		0.367622,
		0.485526,
		0.591806,
		0.679276,
		0.743688,
		0.784259,
		0.8041,
		0.810035,
		0.810531,
		0.839261,
		0.859722,
		0.863508,
		0.84393,
		0.797981,
		0.726371,
		0.632867,
		0.523504,
		0.405858,
		0.28844,
		0.180217,
		0.090256,
		0.027471,
		0.000457,
		0.007571,
		0.061533,
		0.153342,
		0.267665,
		0.390871,
		0.511233,
		0.61919,
		0.707694,
		0.77265,
		0.813438,
		0.833327,
		0.839261,
		0.839755,
		0.865629,
		0.883247,
		0.884372,
		0.862495,
		0.814743,
		0.741871,
		0.64759,
		0.537774,
		0.419747,
		0.301705,
		0.192263,
		0.100129,
		0.033873,
		0.001778,
		0.01182,
		0.072182,
		0.169039,
		0.287213,
		0.413227,
		0.535513,
		0.644671,
		0.733821,
		0.799039,
		0.83987,
		0.859722,
		0.865629,
		0.86612,
		0.889104,
		0.903927,
		0.902538,
		0.878606,
		0.829368,
		0.755596,
		0.660911,
		0.551001,
		0.432925,
		0.314561,
		0.204182,
		0.110155,
		0.040727,
		0.003895,
		0.016943,
		0.083354,
		0.184789,
		0.30628,
		0.434554,
		0.558244,
		0.66815,
		0.757581,
		0.822798,
		0.863508,
		0.883247,
		0.889104,
		0.88959,
		0.909715,
		0.921848,
		0.918136,
		0.892416,
		0.842016,
		0.767694,
		0.672954,
		0.563276,
		0.445448,
		0.327031,
		0.215968,
		0.120303,
		0.047983,
		0.00674,
		0.022875,
		0.094942,
		0.20047,
		0.324747,
		0.454752,
		0.579355,
		0.689582,
		0.778956,
		0.84393,
		0.884372,
		0.903927,
		0.909715,
		0.910193,
		0.927549,
		0.937149,
		0.931338,
		0.904114,
		0.852871,
		0.778329,
		0.683852,
		0.574699,
		0.457378,
		0.339142,
		0.227619,
		0.130547,
		0.055596,
		0.010251,
		0.029542,
		0.106843,
		0.215973,
		0.342523,
		0.473757,
		0.598814,
		0.708972,
		0.797981,
		0.862495,
		0.902538,
		0.921848,
		0.927549,
		0.928019,
		0.942748,
		0.950009,
		0.942348,
		0.913908,
		0.862132,
		0.787675,
		0.693744,
		0.585368,
		0.468778,
		0.350925,
		0.239137,
		0.140868,
		0.063527,
		0.01437,
		0.0005,
		0.036866,
		0.118961,
		0.231212,
		0.359546,
		0.491542,
		0.616635,
		0.726371,
		0.814743,
		0.878606,
		0.918136,
		0.937149,
		0.942748,
		0.943208,
		0.955494,
		0.960643,
		0.951392,
		0.922022,
		0.870001,
		0.795905,
		0.702768,
		0.595385,
		0.479711,
		0.362411,
		0.250527,
		0.151247,
		0.071739,
		0.019044,
		0.001968,
		0.044763,
		0.131208,
		0.246117,
		0.375779,
		0.508113,
		0.632867,
		0.741871,
		0.829368,
		0.892416,
		0.931338,
		0.950009,
		0.955494,
		0.955945,
		0.966004,
		0.969283,
		0.958709,
		0.928683,
		0.876682,
		0.803188,
		0.711055,
		0.604841,
		0.490235,
		0.373628,
		0.261794,
		0.16167,
		0.0802,
		0.024222,
		0.0004,
		0.004344,
		0.053154,
		0.143513,
		0.260641,
		0.391212,
		0.523504,
		0.64759,
		0.755596,
		0.842016,
		0.904114,
		0.942348,
		0.960643,
		0.966004,
		0.966443,
		0.974514,
		0.976177,
		0.964539,
		0.934111,
		0.882368,
		0.809681,
		0.718723,
		0.61382,
		0.500403,
		0.384603,
		0.272942,
		0.172124,
		0.088879,
		0.02986,
		0.001559,
		0.007563,
		0.061965,
		0.155812,
		0.274752,
		0.405858,
		0.537774,
		0.660911,
		0.767694,
		0.852871,
		0.913908,
		0.951392,
		0.969283,
		0.974514,
		0.974941,
		0.981272,
		0.981571,
		0.969117,
		0.938519,
		0.887236,
		0.815526,
		0.72588,
		0.622395,
		0.51026,
		0.395357,
		0.283974,
		0.182595,
		0.097751,
		0.035914,
		0.003417,
		0.011555,
		0.071125,
		0.168059,
		0.28844,
		0.419747,
		0.551001,
		0.672954,
		0.778329,
		0.862132,
		0.922022,
		0.958709,
		0.976177,
		0.981272,
		0.981688,
		0.986527,
		0.985706,
		0.972665,
		0.942099,
		0.891446,
		0.820849,
		0.732617,
		0.630629,
		0.519843,
		0.405908,
		0.294891,
		0.193071,
		0.106789,
		0.042346,
		0.00592,
		0.016249,
		0.080573,
		0.180217,
		0.301705,
		0.432925,
		0.563276,
		0.683852,
		0.787675,
		0.870001,
		0.928683,
		0.964539,
		0.981571,
		0.986527,
		0.986931,
		0.990522,
		0.988808,
		0.975383,
		0.945023,
		0.895138,
		0.825756,
		0.739009,
		0.638572,
		0.529181,
		0.416269,
		0.305694,
		0.20354,
		0.11597,
		0.049119,
		0.009016,
		0.021576,
		0.090256,
		0.192263,
		0.314561,
		0.445448,
		0.574699,
		0.693744,
		0.795905,
		0.876682,
		0.934111,
		0.969117,
		0.985706,
		0.990522,
		0.990914,
		0.993485,
		0.991083,
		0.977452,
		0.94744,
		0.898429,
		0.830333,
		0.745119,
		0.646263,
		0.538297,
		0.42645,
		0.316381,
		0.213992,
		0.125272,
		0.0562,
		0.012657,
		0.027471,
		0.100129,
		0.204182,
		0.327031,
		0.457378,
		0.585368,
		0.702768,
		0.803188,
		0.882368,
		0.938519,
		0.972665,
		0.988808,
		0.993485,
		0.993865,
		0.995622,
		0.992715,
		0.979026,
		0.949475,
		0.901414,
		0.83465,
		0.750991,
		0.653732,
		0.547208,
		0.436456,
		0.32695,
		0.224414,
		0.134676,
		0.063557,
		0.0168,
		0.000457,
		0.033873,
		0.110155,
		0.215968,
		0.339142,
		0.468778,
		0.595385,
		0.711055,
		0.809681,
		0.887236,
		0.942099,
		0.975383,
		0.991083,
		0.995622,
		0.99599,
		0.997118,
		0.993859,
		0.980234,
		0.951227,
		0.904167,
		0.838759,
		0.756663,
		0.661,
		0.555925,
		0.446291,
		0.337397,
		0.234798,
		0.144161,
		0.071162,
		0.021402,
		0.000352,
		0.001778,
		0.040727,
		0.120303,
		0.227619,
		0.350925,
		0.479711,
		0.604841,
		0.718723,
		0.815526,
		0.891446,
		0.945023,
		0.977452,
		0.992715,
		0.997118,
		0.997474,
		0.99813,
		0.994648,
		0.981178,
		0.952776,
		0.906745,
		0.842698,
		0.762157,
		0.668082,
		0.564456,
		0.455958,
		0.34772,
		0.245133,
		0.153712,
		0.078987,
		0.026425,
		0.001374,
		0.003895,
		0.047983,
		0.130547,
		0.239137,
		0.362411,
		0.490235,
		0.61382,
		0.72588,
		0.820849,
		0.895138,
		0.94744,
		0.979026,
		0.993859,
		0.99813,
		0.998475,
		0.998789,
		0.995184,
		0.981941,
		0.95418,
		0.909189,
		0.846495,
		0.767493,
		0.674989,
		0.572806,
		0.465459,
		0.357915,
		0.255412,
		0.163312,
		0.087007,
		0.031834,
		0.003018,
		0.00674,
		0.055596,
		0.140868,
		0.250527,
		0.373628,
		0.500403,
		0.622395,
		0.732617,
		0.825756,
		0.898429,
		0.949475,
		0.980234,
		0.994648,
		0.998789,
		0.999123,
		0.9992,
		0.99555,
		0.98258,
		0.955481,
		0.911526,
		0.850168,
		0.77268,
		0.681729,
		0.58098,
		0.474793,
		0.36798,
		0.265625,
		0.172947,
		0.0952,
		0.037594,
		0.005236,
		0.010251,
		0.063527,
		0.151247,
		0.261794,
		0.384603,
		0.51026,
		0.630629,
		0.739009,
		0.830333,
		0.901414,
		0.951227,
		0.981178,
		0.995184,
		0.9992,
		0.999523,
		0.999445,
		0.995806,
		0.983139,
		0.956706,
		0.913775,
		0.853728,
		0.777728,
		0.688306,
		0.588982,
		0.483964,
		0.377913,
		0.275766,
		0.182603,
		0.103544,
		0.043674,
		0.007987,
		0.01437,
		0.071739,
		0.16167,
		0.272942,
		0.395357,
		0.519843,
		0.638572,
		0.745119,
		0.83465,
		0.904167,
		0.952776,
		0.981941,
		0.99555,
		0.999445,
		0.999758,
		0.999584,
		0.995994,
		0.983646,
		0.957876,
		0.915948,
		0.857183,
		0.782641,
		0.694725,
		0.596815,
		0.492972,
		0.387712,
		0.28583,
		0.192269,
		0.112019,
		0.050046,
		0.011231,
		0.019044,
		0.0802,
		0.172124,
		0.283974,
		0.405908,
		0.529181,
		0.646263,
		0.750991,
		0.838759,
		0.906745,
		0.95418,
		0.98258,
		0.995806,
		0.999584,
		0.999887,
		0.999659,
		0.996142,
		0.984119,
		0.958999,
		0.918052,
		0.860539,
		0.787425,
		0.700991,
		0.604481,
		0.501818,
		0.397375,
		0.295809,
		0.201932,
		0.120607,
		0.056682,
		0.014928,
		0.024222,
		0.088879,
		0.182595,
		0.294891,
		0.416269,
		0.538297,
		0.653732,
		0.756663,
		0.842698,
		0.909189,
		0.955481,
		0.983139,
		0.995994,
		0.999659,
		0.999952,
		0.999698,
		0.996268,
		0.984568,
		0.960083,
		0.92009,
		0.863799,
		0.792083,
		0.707107,
		0.611985,
		0.510503,
		0.406901,
		0.3057,
		0.211582,
		0.129291,
		0.063557,
		0.019046,
		0.000312,
		0.0004,
		0.02986,
		0.097751,
		0.193071,
		0.305694,
		0.42645,
		0.547208,
		0.661,
		0.762157,
		0.846495,
		0.911526,
		0.956706,
		0.983646,
		0.996142,
		0.999698,
		0.999982,
		0.999719,
		0.996382,
		0.984999,
		0.961131,
		0.922066,
		0.866967,
		0.796619,
		0.713077,
		0.619329,
		0.519031,
		0.416289,
		0.315497,
		0.221211,
		0.138055,
		0.070648,
		0.02355,
		0.001221,
		0.001559,
		0.035914,
		0.106789,
		0.20354,
		0.316381,
		0.436456,
		0.555925,
		0.668082,
		0.767493,
		0.850168,
		0.913775,
		0.957876,
		0.984119,
		0.996268,
		0.999719,
		0.999994,
		0.999732,
		0.996489,
		0.985413,
		0.962145,
		0.923982,
		0.870045,
		0.801037,
		0.718904,
		0.626516,
		0.527402,
		0.42554,
		0.325198,
		0.23081,
		0.146884,
		0.077932,
		0.02841,
		0.002685,
		0.003417,
		0.042346,
		0.11597,
		0.213992,
		0.32695,
		0.446291,
		0.564456,
		0.674989,
		0.77268,
		0.853728,
		0.915948,
		0.958999,
		0.984568,
		0.996382,
		0.999732,
		0.999999,
		0.999741,
		0.996591,
		0.985814,
		0.963126,
		0.925841,
		0.873037,
		0.80534,
		0.724593,
		0.63355,
		0.535618,
		0.434652,
		0.334797,
		0.24037,
		0.155766,
		0.085389,
		0.033598,
		0.004665,
		0.00592,
		0.049119,
		0.125272,
		0.224414,
		0.337397,
		0.455958,
		0.572806,
		0.681729,
		0.777728,
		0.857183,
		0.918052,
		0.960083,
		0.984999,
		0.996489,
		0.999741,
		1,
		0.999749,
		0.99669,
		0.986201,
		0.964076,
		0.927644,
		0.875946,
		0.809531,
		0.730146,
		0.640434,
		0.543682,
		0.443626,
		0.344293,
		0.249885,
		0.164687,
		0.093,
		0.039085,
		0.007125,
		0.009016,
		0.0562,
		0.134676,
		0.234798,
		0.34772,
		0.465459,
		0.58098,
		0.688306,
		0.782641,
		0.860539,
		0.92009,
		0.961131,
		0.985413,
		0.996591,
		0.999749,
		1,
		0.999757,
		0.996784,
		0.986574,
		0.964996,
		0.929393,
		0.878773,
		0.813614,
		0.735568,
		0.647171,
		0.551596,
		0.452462,
		0.353683,
		0.259348,
		0.173636,
		0.100747,
		0.044848,
		0.010032,
		0.012657,
		0.063557,
		0.144161,
		0.245133,
		0.357915,
		0.474793,
		0.588982,
		0.694725,
		0.787425,
		0.863799,
		0.922066,
		0.962145,
		0.985814,
		0.99669,
		0.999757,
		1,
		0.999764,
		0.996876,
		0.986936,
		0.965887,
		0.931091,
		0.881522,
		0.817592,
		0.740862,
		0.653763,
		0.559361,
		0.461161,
		0.362964,
		0.268753,
		0.182603,
		0.108615,
		0.050862,
		0.013353,
		0.0168,
		0.071162,
		0.153712,
		0.255412,
		0.36798,
		0.483964,
		0.596815,
		0.700991,
		0.792083,
		0.866967,
		0.923982,
		0.963126,
		0.986201,
		0.996784,
		0.999764,
		1,
		0.999771,
		0.996964,
		0.987285,
		0.96675,
		0.932739,
		0.884195,
		0.821467,
		0.74603,
		0.660215,
		0.56698,
		0.469722,
		0.372135,
		0.278096,
		0.191578,
		0.116587,
		0.057105,
		0.017058,
		0.000279,
		0.021402,
		0.078987,
		0.163312,
		0.265625,
		0.377913,
		0.492972,
		0.604481,
		0.707107,
		0.796619,
		0.870045,
		0.925841,
		0.964076,
		0.986574,
		0.996876,
		0.999771,
		1,
		0.999778,
		0.997049,
		0.987623,
		0.967586,
		0.934338,
		0.886795,
		0.825244,
		0.751077,
		0.666529,
		0.574455,
		0.478147,
		0.381194,
		0.287371,
		0.200552,
		0.124649,
		0.063557,
		0.021119,
		0.001092,
		0.000352,
		0.026425,
		0.087007,
		0.172947,
		0.275766,
		0.387712,
		0.501818,
		0.611985,
		0.713077,
		0.801037,
		0.873037,
		0.927644,
		0.964996,
		0.986936,
		0.996964,
		0.999778,
		1,
		0.999784,
		0.997131,
		0.987949,
		0.968397,
		0.93589,
		0.889324,
		0.828924,
		0.756005,
		0.672708,
		0.581788,
		0.486437,
		0.39014,
		0.296573,
		0.209516,
		0.132788,
		0.070199,
		0.02551,
		0.002404,
		0.001374,
		0.031834,
		0.0952,
		0.182603,
		0.28583,
		0.397375,
		0.510503,
		0.619329,
		0.718904,
		0.80534,
		0.875946,
		0.929393,
		0.965887,
		0.987285,
		0.997049,
		0.999784,
		1,
		0.99979,
		0.99721,
		0.988265,
		0.969182,
		0.937398,
		0.891784,
		0.832511,
		0.760817,
		0.678754,
		0.588982,
		0.494592,
		0.398972,
		0.3057,
		0.218463,
		0.140991,
		0.077011,
		0.030205,
		0.004182,
		0.003018,
		0.037594,
		0.103544,
		0.192269,
		0.295809,
		0.406901,
		0.519031,
		0.626516,
		0.724593,
		0.809531,
		0.878773,
		0.931091,
		0.96675,
		0.987623,
		0.997131,
		0.99979,
		1,
		0.999797,
		0.997287,
		0.988571,
		0.969943,
		0.938862,
		0.894177,
		0.836008,
		0.765517,
		0.684672,
		0.596039,
		0.502614,
		0.407688,
		0.314747,
		0.227386,
		0.149248,
		0.083978,
		0.035182,
		0.006396,
		0.005236,
		0.043674,
		0.112019,
		0.201932,
		0.3057,
		0.416289,
		0.527402,
		0.63355,
		0.730146,
		0.813614,
		0.881522,
		0.932739,
		0.967586,
		0.987949,
		0.99721,
		0.999797,
		1,
		0.999802,
		0.997361,
		0.988867,
		0.970681,
		0.940284,
		0.896506,
		0.839416,
		0.770107,
		0.690463,
		0.602961,
		0.510503,
		0.416289,
		0.323712,
		0.236278,
		0.157547,
		0.091084,
		0.040417,
		0.009016,
		0.007987,
		0.050046,
		0.120607,
		0.211582,
		0.315497,
		0.42554,
		0.535618,
		0.640434,
		0.735568,
		0.817592,
		0.884195,
		0.934338,
		0.968397,
		0.988265,
		0.997287,
		0.999802,
		1,
		0.999808,
		0.997432,
		0.989154,
		0.971397,
		0.941666,
		0.898772,
		0.842738,
		0.77459,
		0.696131,
		0.609751,
		0.518262,
		0.424774,
		0.332591,
		0.245133,
		0.165879,
		0.098313,
		0.045892,
		0.012014,
		0.011231,
		0.056682,
		0.129291,
		0.221211,
		0.325198,
		0.434652,
		0.543682,
		0.647171,
		0.740862,
		0.821467,
		0.886795,
		0.93589,
		0.969182,
		0.988571,
		0.997361,
		0.999808,
		1,
		0.999813,
		0.997501,
		0.989431,
		0.972091,
		0.943008,
		0.900978,
		0.845978,
		0.778969,
		0.701678,
		0.61641,
		0.525892,
		0.433143,
		0.341382,
		0.253947,
		0.174234,
		0.105651,
		0.051585,
		0.015366,
		0.000251,
		0.014928,
		0.063557,
		0.138055,
		0.23081,
		0.334797,
		0.443626,
		0.551596,
		0.653763,
		0.74603,
		0.825244,
		0.889324,
		0.937398,
		0.969943,
		0.988867,
		0.997432,
		0.999813,
		1,
		0.999819,
		0.997568,
		0.9897,
		0.972765,
		0.944313,
		0.903125,
		0.849137,
		0.783246,
		0.707107,
		0.622942,
		0.533393,
		0.441396,
		0.350084,
		0.262714,
		0.182603,
		0.113087,
		0.057479,
		0.019046,
		0.000982,
		0.019046,
		0.070648,
		0.146884,
		0.24037,
		0.344293,
		0.452462,
		0.559361,
		0.660215,
		0.751077,
		0.828924,
		0.891784,
		0.938862,
		0.970681,
		0.989154,
		0.997501,
		0.999819,
		1,
		0.999824,
		0.997633,
		0.98996,
		0.973418,
		0.94558,
		0.905215,
		0.852217,
		0.787425,
		0.71242,
		0.629348,
		0.540768,
		0.449532,
		0.358694,
		0.271429,
		0.19098,
		0.120607,
		0.063557,
		0.023032,
		0.002165,
		0.000312,
		0.02355,
		0.077932,
		0.155766,
		0.249885,
		0.353683,
		0.461161,
		0.56698,
		0.666529,
		0.756005,
		0.832511,
		0.894177,
		0.940284,
		0.971397,
		0.989431,
		0.997568,
		0.999824,
		1,
		0.999829,
		0.997695,
		0.990212,
		0.974052,
		0.946813,
		0.90725,
		0.855222,
		0.791508,
		0.717621,
		0.635631,
		0.548017,
		0.457553,
		0.367211,
		0.280089,
		0.199356,
		0.128201,
		0.069803,
		0.027301,
		0.003771,
		0.001221,
		0.02841,
		0.085389,
		0.164687,
		0.259348,
		0.362964,
		0.469722,
		0.574455,
		0.672708,
		0.760817,
		0.836008,
		0.896506,
		0.941666,
		0.972091,
		0.9897,
		0.997633,
		0.999829,
		1,
		0.999833,
		0.997755,
		0.990456,
		0.974668,
		0.948011,
		0.909232,
		0.858152,
		0.795497,
		0.722712,
		0.641793,
		0.555144,
		0.465459,
		0.375633,
		0.28869,
		0.207724,
		0.135857,
		0.076202,
		0.031834,
		0.005773,
		0.002685,
		0.033598,
		0.093,
		0.173636,
		0.268753,
		0.372135,
		0.478147,
		0.581788,
		0.678754,
		0.765517,
		0.839416,
		0.898772,
		0.943008,
		0.972765,
		0.98996,
		0.997695,
		0.999833,
		1,
		0.999838,
		0.997814,
		0.990693,
		0.975265,
		0.949176,
		0.911162,
		0.861011,
		0.799394,
		0.727694,
		0.647836,
		0.562148,
		0.473249,
		0.383959,
		0.297228,
		0.216079,
		0.143566,
		0.082738,
		0.036611,
		0.008146,
		0.004665,
		0.039085,
		0.100747,
		0.182603,
		0.278096,
		0.381194,
		0.486437,
		0.588982,
		0.684672,
		0.770107,
		0.842738,
		0.900978,
		0.944313,
		0.973418,
		0.990212,
		0.997755,
		0.999838,
		1,
		0.999842,
		0.99787,
		0.990922,
		0.975845,
		0.950308,
		0.913042,
		0.863799,
		0.803203,
		0.732572,
		0.653763,
		0.569033,
		0.480925,
		0.392188,
		0.3057,
		0.224414,
		0.151319,
		0.0894,
		0.041614,
		0.010867,
		0.007125,
		0.044848,
		0.108615,
		0.191578,
		0.287371,
		0.39014,
		0.494592,
		0.596039,
		0.690463,
		0.77459,
		0.845978,
		0.903125,
		0.94558,
		0.974052,
		0.990456,
		0.997814,
		0.999842,
		1,
		0.999846,
		0.997925,
		0.991144,
		0.976409,
		0.95141,
		0.914874,
		0.86652,
		0.806924,
		0.737347,
		0.659576,
		0.575799,
		0.488488,
		0.40032,
		0.314104,
		0.232725,
		0.159107,
		0.096175,
		0.046826,
		0.013913,
		0.000227,
		0.010032,
		0.050862,
		0.116587,
		0.200552,
		0.296573,
		0.398972,
		0.502614,
		0.602961,
		0.696131,
		0.778969,
		0.849137,
		0.905215,
		0.946813,
		0.974668,
		0.990693,
		0.99787,
		0.999846,
		1,
		0.999851,
		0.997978,
		0.99136,
		0.976956,
		0.952482,
		0.916658,
		0.869175,
		0.810562,
		0.742021,
		0.665277,
		0.582448,
		0.495938,
		0.408354,
		0.322436,
		0.241006,
		0.166922,
		0.10305,
		0.05223,
		0.017263,
		0.000888,
		0.013353,
		0.057105,
		0.124649,
		0.209516,
		0.3057,
		0.407688,
		0.510503,
		0.609751,
		0.701678,
		0.783246,
		0.852217,
		0.90725,
		0.948011,
		0.975265,
		0.990922,
		0.997925,
		0.999851,
		1,
		0.999854,
		0.998029,
		0.991569,
		0.977487,
		0.953525,
		0.918396,
		0.871765,
		0.814117,
		0.746597,
		0.670868,
		0.588982,
		0.503276,
		0.416289,
		0.330696,
		0.249252,
		0.174757,
		0.110014,
		0.057812,
		0.020897,
		0.00196,
		0.017058,
		0.063557,
		0.132788,
		0.218463,
		0.314747,
		0.416289,
		0.518262,
		0.61641,
		0.707107,
		0.787425,
		0.855222,
		0.909232,
		0.949176,
		0.975845,
		0.991144,
		0.997978,
		0.999854,
		1,
		0.999858,
		0.998079,
		0.991772,
		0.978003,
		0.954539,
		0.92009,
		0.874294,
		0.817592,
		0.751077,
		0.676351,
		0.595403,
		0.510503,
		0.424126,
		0.33888,
		0.25746,
		0.182603,
		0.117059,
		0.063557,
		0.024796,
		0.003417,
		0.000279,
		0.021119,
		0.070199,
		0.140991,
		0.227386,
		0.323712,
		0.424774,
		0.525892,
		0.622942,
		0.71242,
		0.791508,
		0.858152,
		0.911162,
		0.950308,
		0.976409,
		0.99136,
		0.998029,
		0.999858,
		1,
		0.999862,
		0.998127,
		0.991969,
		0.978505,
		0.955526,
		0.921741,
		0.876762,
		0.820988,
		0.755463,
		0.681729,
		0.601712,
		0.517621,
		0.431863,
		0.346987,
		0.265625,
		0.190456,
		0.124172,
		0.069452,
		0.028941,
		0.005236,
		0.001092,
		0.02551,
		0.077011,
		0.149248,
		0.236278,
		0.332591,
		0.433143,
		0.533393,
		0.629348,
		0.717621,
		0.795497,
		0.861011,
		0.913042,
		0.95141,
		0.976956,
		0.991569,
		0.998079,
		0.999862,
		1,
		0.999866,
		0.998174,
		0.99216,
		0.978993,
		0.956487,
		0.92335,
		0.87917,
		0.824309,
		0.759757,
		0.687003,
		0.607912,
		0.524629,
		0.439502,
		0.355015,
		0.273744,
		0.198309,
		0.131346,
		0.075484,
		0.033317,
		0.007397,
		0.002404,
		0.030205,
		0.083978,
		0.157547,
		0.245133,
		0.341382,
		0.441396,
		0.540768,
		0.635631,
		0.722712,
		0.799394,
		0.863799,
		0.914874,
		0.952482,
		0.977487,
		0.991772,
		0.998127,
		0.999866,
		1,
		0.999869,
		0.998219,
		0.992346,
		0.979466,
		0.957423,
		0.924919,
		0.881522,
		0.827555,
		0.763962,
		0.692176,
		0.614003,
		0.531529,
		0.447041,
		0.362964,
		0.281814,
		0.206156,
		0.138573,
		0.08164,
		0.037906,
		0.009877,
		0.004182,
		0.035182,
		0.091084,
		0.165879,
		0.253947,
		0.350084,
		0.449532,
		0.548017,
		0.641793,
		0.727694,
		0.803203,
		0.86652,
		0.916658,
		0.953525,
		0.978003,
		0.991969,
		0.998174,
		0.999869,
		1,
		0.999872,
		0.998263,
		0.992526,
		0.979927,
		0.958333,
		0.926448,
		0.883818,
		0.830729,
		0.76808,
		0.69725,
		0.619989,
		0.538323,
		0.454482,
		0.370832,
		0.289832,
		0.213992,
		0.145843,
		0.08791,
		0.042694,
		0.012657,
		0.000206,
		0.006396,
		0.040417,
		0.098313,
		0.174234,
		0.262714,
		0.358694,
		0.457553,
		0.555144,
		0.647836,
		0.732572,
		0.806924,
		0.869175,
		0.918396,
		0.954539,
		0.978505,
		0.99216,
		0.998219,
		0.999872,
		1,
		0.999876,
		0.998306,
		0.9927,
		0.980375,
		0.95922,
		0.927939,
		0.886059,
		0.833833,
		0.772112,
		0.702226,
		0.625869,
		0.545012,
		0.461824,
		0.378617,
		0.297795,
		0.221812,
		0.153149,
		0.094282,
		0.047667,
		0.015719,
		0.000807,
		0.009016,
		0.045892,
		0.105651,
		0.182603,
		0.271429,
		0.367211,
		0.465459,
		0.562148,
		0.653763,
		0.737347,
		0.810562,
		0.871765,
		0.92009,
		0.955526,
		0.978993,
		0.992346,
		0.998263,
		0.999876,
		1,
		0.999879,
		0.998347,
		0.99287,
		0.980811,
		0.960083,
		0.929393,
		0.888248,
		0.836868,
		0.776061,
		0.707107,
		0.631647,
		0.551596,
		0.469068,
		0.38632,
		0.3057,
		0.229612,
		0.160485,
		0.100747,
		0.05281,
		0.019046,
		0.001783,
		0.012014,
		0.051585,
		0.113087,
		0.19098,
		0.280089,
		0.375633,
		0.473249,
		0.569033,
		0.659576,
		0.742021,
		0.814117,
		0.874294,
		0.921741,
		0.956487,
		0.979466,
		0.992526,
		0.998306,
		0.999879,
		1,
		0.999882,
		0.998387,
		0.993035,
		0.981234,
		0.960924,
		0.930812,
		0.890386,
		0.839836,
		0.779928,
		0.711894,
		0.637323,
		0.558077,
		0.476215,
		0.393939,
		0.313545,
		0.237387,
		0.167843,
		0.107296,
		0.058111,
		0.02262,
		0.003111,
		0.015366,
		0.057479,
		0.120607,
		0.199356,
		0.28869,
		0.383959,
		0.480925,
		0.575799,
		0.665277,
		0.746597,
		0.817592,
		0.876762,
		0.92335,
		0.957423,
		0.979927,
		0.9927,
		0.998347,
		0.999882,
		1,
		0.999885,
		0.998425,
		0.993195,
		0.981647,
		0.961743,
		0.932195,
		0.892474,
		0.842738,
		0.783716,
		0.71659,
		0.642901,
		0.564456,
		0.483264,
		0.401474,
		0.321329,
		0.245133,
		0.175218,
		0.113919,
		0.063557,
		0.026425,
		0.004772,
		0.000251,
		0.019046,
		0.063557,
		0.128201,
		0.207724,
		0.297228,
		0.392188,
		0.488488,
		0.582448,
		0.670868,
		0.751077,
		0.820988,
		0.87917,
		0.924919,
		0.958333,
		0.980375,
		0.99287,
		0.998387,
		0.999885,
		1,
		0.999888,
		0.998463,
		0.99335,
		0.982047,
		0.962541,
		0.933544,
		0.894513,
		0.845578,
		0.787425,
		0.721196,
		0.64838,
		0.570735,
		0.490217,
		0.408924,
		0.32905,
		0.252848,
		0.182603,
		0.120607,
		0.069138,
		0.030448,
		0.006746,
		0.000982,
		0.023032,
		0.069803,
		0.135857,
		0.216079,
		0.3057,
		0.40032,
		0.495938,
		0.588982,
		0.676351,
		0.755463,
		0.824309,
		0.881522,
		0.926448,
		0.95922,
		0.980811,
		0.993035,
		0.998425,
		0.999888,
		1,
		0.999891,
		0.998499,
		0.993502,
		0.982438,
		0.963319,
		0.93486,
		0.896506,
		0.848355,
		0.791059,
		0.725714,
		0.653763,
		0.576915,
		0.497074,
		0.416289,
		0.336705,
		0.260527,
		0.189994,
		0.127354,
		0.074843,
		0.034672,
		0.009016,
		0.002165,
		0.027301,
		0.076202,
		0.143566,
		0.224414,
		0.314104,
		0.408354,
		0.503276,
		0.595403,
		0.681729,
		0.759757,
		0.827555,
		0.883818,
		0.927939,
		0.960083,
		0.981234,
		0.993195,
		0.998463,
		0.999891,
		1,
		0.999893,
		0.998535,
		0.993648,
		0.982817,
		0.964076,
		0.936145,
		0.898452,
		0.851071,
		0.794618,
		0.730146,
		0.659052,
		0.582997,
		0.503836,
		0.423569,
		0.344293,
		0.268167,
		0.197385,
		0.134151,
		0.08066,
		0.039085,
		0.011564,
		0.000188,
		0.003771,
		0.031834,
		0.082738,
		0.151319,
		0.232725,
		0.322436,
		0.416289,
		0.510503,
		0.601712,
		0.687003,
		0.763962,
		0.830729,
		0.886059,
		0.929393,
		0.960924,
		0.981647,
		0.99335,
		0.998499,
		0.999893,
		1,
		0.999896,
		0.998569,
		0.993791,
		0.983187,
		0.964815,
		0.937398,
		0.900354,
		0.853729,
		0.798105,
		0.734494,
		0.664249,
		0.588982,
		0.510503,
		0.430764,
		0.351814,
		0.275766,
		0.204772,
		0.140991,
		0.086581,
		0.043674,
		0.014374,
		0.000737,
		0.005773,
		0.036611,
		0.0894,
		0.159107,
		0.241006,
		0.330696,
		0.424126,
		0.517621,
		0.607912,
		0.692176,
		0.76808,
		0.833833,
		0.888248,
		0.930812,
		0.961743,
		0.982047,
		0.993502,
		0.998535,
		0.999896,
		1,
		0.999898,
		0.998603,
		0.99393,
		0.983546,
		0.965534,
		0.938621,
		0.902212,
		0.856329,
		0.801521,
		0.738759,
		0.669354,
		0.594872,
		0.517077,
		0.437873,
		0.359265,
		0.283322,
		0.212149,
		0.147869,
		0.092596,
		0.048428,
		0.01743,
		0.001628,
		0.008146,
		0.041614,
		0.096175,
		0.166922,
		0.249252,
		0.33888,
		0.431863,
		0.524629,
		0.614003,
		0.69725,
		0.772112,
		0.836868,
		0.890386,
		0.932195,
		0.962541,
		0.982438,
		0.993648,
		0.998569,
		0.999898,
		1,
		0.999901,
		0.998635,
		0.994065,
		0.983896,
		0.966236,
		0.939815,
		0.904028,
		0.858873,
		0.804867,
		0.742944,
		0.67437,
		0.600669,
		0.523558,
		0.444897,
		0.366646,
		0.29083,
		0.219514,
		0.154777,
		0.098696,
		0.053333,
		0.020718,
		0.002844,
		0.010867,
		0.046826,
		0.10305,
		0.174757,
		0.25746,
		0.346987,
		0.439502,
		0.531529,
		0.619989,
		0.702226,
		0.776061,
		0.839836,
		0.892474,
		0.933544,
		0.963319,
		0.982817,
		0.993791,
		0.998603,
		0.999901,
		1,
		0.999903,
		0.998667,
		0.994196,
		0.984237,
		0.96692,
		0.94098,
		0.905802,
		0.861363,
		0.808146,
		0.747049,
		0.679298,
		0.606372,
		0.529947,
		0.451836,
		0.373956,
		0.29829,
		0.226862,
		0.16171,
		0.104874,
		0.05838,
		0.024223,
		0.004366,
		0.013913,
		0.05223,
		0.110014,
		0.182603,
		0.265625,
		0.355015,
		0.447041,
		0.538323,
		0.625869,
		0.707107,
		0.779928,
		0.842738,
		0.894513,
		0.93486,
		0.964076,
		0.983187,
		0.99393,
		0.998635,
		0.999903,
		1,
		0.999906,
		0.998697,
		0.994324,
		0.984569,
		0.967586,
		0.942118,
		0.907537,
		0.863799,
		0.811359,
		0.751077,
		0.684139,
		0.611985,
		0.536244,
		0.45869,
		0.381194,
		0.3057,
		0.234189,
		0.168662,
		0.111122,
		0.063557,
		0.027933,
		0.006177,
		0.000227,
		0.017263,
		0.057812,
		0.117059,
		0.190456,
		0.273744,
		0.362964,
		0.454482,
		0.545012,
		0.631647,
		0.711894,
		0.783716,
		0.845578,
		0.896506,
		0.936145,
		0.964815,
		0.983546,
		0.994065,
		0.998667,
		0.999906,
		1,
		0.999908,
		0.998727,
		0.994448,
		0.984893,
		0.968237,
		0.943228,
		0.909232,
		0.866184,
		0.814507,
		0.755028,
		0.688896,
		0.617508,
		0.542452,
		0.465459,
		0.38836,
		0.313057,
		0.241492,
		0.175628,
		0.117431,
		0.068856,
		0.031834,
		0.008262,
		0.000888,
		0.020897,
		0.063557,
		0.124172,
		0.198309,
		0.281814,
		0.370832,
		0.461824,
		0.551596,
		0.637323,
		0.71659,
		0.787425,
		0.848355,
		0.898452,
		0.937398,
		0.965534,
		0.983896,
		0.994196,
		0.998697,
		0.999908,
		1,
		0.99991,
		0.998756,
		0.994569,
		0.985208,
		0.968871,
		0.944313,
		0.91089,
		0.868517,
		0.817592,
		0.758906,
		0.69357,
		0.622942,
		0.54857,
		0.472143,
		0.395453,
		0.32036,
		0.248768,
		0.182603,
		0.123796,
		0.074267,
		0.035914,
		0.010606,
		0.000172,
		0.00196,
		0.024796,
		0.069452,
		0.131346,
		0.206156,
		0.289832,
		0.378617,
		0.469068,
		0.558077,
		0.642901,
		0.721196,
		0.791059,
		0.851071,
		0.900354,
		0.938621,
		0.966236,
		0.984237,
		0.994324,
		0.998727,
		0.99991,
		1,
		0.999912,
		0.998784,
		0.994686,
		0.985515,
		0.969489,
		0.945372,
		0.91251,
		0.870801,
		0.820615,
		0.76271,
		0.698162,
		0.628289,
		0.5546,
		0.478744,
		0.402472,
		0.327607,
		0.256014,
		0.189584,
		0.13021,
		0.079781,
		0.040162,
		0.013194,
		0.000675,
		0.003417,
		0.028941,
		0.075484,
		0.138573,
		0.213992,
		0.297795,
		0.38632,
		0.476215,
		0.564456,
		0.64838,
		0.725714,
		0.794618,
		0.853729,
		0.902212,
		0.939815,
		0.96692,
		0.984569,
		0.994448,
		0.998756,
		0.999912,
		1,
		0.999914,
		0.998811,
		0.994801,
		0.985814,
		0.970093,
		0.946406,
		0.914095,
		0.873037,
		0.823577,
		0.766443,
		0.702674,
		0.63355,
		0.560543,
		0.485261,
		0.409418,
		0.334797,
		0.263228,
		0.196564,
		0.136666,
		0.085389,
		0.044568,
		0.016011,
		0.001493,
		0.005236,
		0.033317,
		0.08164,
		0.145843,
		0.221812,
		0.3057,
		0.393939,
		0.483264,
		0.570735,
		0.653763,
		0.730146,
		0.798105,
		0.856329,
		0.904028,
		0.94098,
		0.967586,
		0.984893,
		0.994569,
		0.998784,
		0.999914,
		1,
		0.999916,
		0.998838,
		0.994912,
		0.986105,
		0.970681,
		0.947416,
		0.915644,
		0.875226,
		0.826481,
		0.770107,
		0.707107,
		0.638727,
		0.566399,
		0.491695,
		0.416289,
		0.341929,
		0.270407,
		0.203541,
		0.14316,
		0.091084,
		0.049119,
		0.019046,
		0.00261,
		0.007397,
		0.037906,
		0.08791,
		0.153149,
		0.229612,
		0.313545,
		0.401474,
		0.490217,
		0.576915,
		0.659052,
		0.734494,
		0.801521,
		0.858873,
		0.905802,
		0.942118,
		0.968237,
		0.985208,
		0.994686,
		0.998811,
		0.999916,
		1,
		0.999918,
		0.998863,
		0.99502,
		0.986389,
		0.971256,
		0.948403,
		0.917159,
		0.877369,
		0.829327,
		0.773702,
		0.711463,
		0.643821,
		0.57217,
		0.498046,
		0.423087,
		0.349001,
		0.277548,
		0.210511,
		0.149684,
		0.096858,
		0.053808,
		0.022285,
		0.00401,
		0.009877,
		0.042694,
		0.094282,
		0.160485,
		0.237387,
		0.321329,
		0.408924,
		0.497074,
		0.582997,
		0.664249,
		0.738759,
		0.804867,
		0.861363,
		0.907537,
		0.943228,
		0.968871,
		0.985515,
		0.994801,
		0.998838,
		0.999918,
		1,
		0.99992,
		0.998888,
		0.995126,
		0.986666,
		0.971816,
		0.949367,
		0.918641,
		0.879467,
		0.832117,
		0.77723,
		0.715743,
		0.648832,
		0.577857,
		0.504315,
		0.429809,
		0.356013,
		0.28465,
		0.21747,
		0.156234,
		0.102704,
		0.058623,
		0.025717,
		0.005678,
		0.012657,
		0.047667,
		0.100747,
		0.167843,
		0.245133,
		0.32905,
		0.416289,
		0.503836,
		0.588982,
		0.669354,
		0.742944,
		0.808146,
		0.863799,
		0.909232,
		0.944313,
		0.969489,
		0.985814,
		0.994912,
		0.998863,
		0.99992,
		1,
		0.999922,
		0.998913,
		0.995229,
		0.986936,
		0.972363,
		0.950308,
		0.92009,
		0.881522,
		0.834852,
		0.780692,
		0.719949,
		0.653763,
		0.583461,
		0.510503,
		0.436458,
		0.362964,
		0.29171,
		0.224414,
		0.162806,
		0.108615,
		0.063557,
		0.02933,
		0.0076,
		0.000206,
		0.015719,
		0.05281,
		0.107296,
		0.175218,
		0.252848,
		0.336705,
		0.423569,
		0.510503,
		0.594872,
		0.67437,
		0.747049,
		0.811359,
		0.866184,
		0.91089,
		0.945372,
		0.970093,
		0.986105,
		0.99502,
		0.998888,
		0.999922,
		1,
		0.999923,
		0.998936,
		0.995329,
		0.987199,
		0.972897,
		0.951229,
		0.921508,
		0.883534,
		0.837533,
		0.78409,
		0.724081,
		0.658615,
		0.588982,
		0.516611,
		0.443032,
		0.369853,
		0.298728,
		0.231342,
		0.169394,
		0.114585,
		0.068601,
		0.033113,
		0.009762,
		0.000158,
		0.000807,
		0.019046,
		0.058111,
		0.113919,
		0.182603,
		0.260527,
		0.344293,
		0.430764,
		0.517077,
		0.600669,
		0.679298,
		0.751077,
		0.814507,
		0.868517,
		0.91251,
		0.946406,
		0.970681,
		0.986389,
		0.995126,
		0.998913,
		0.999923,
		1,
		0.999925,
		0.99896,
		0.995426,
		0.987455,
		0.973418,
		0.952128,
		0.922895,
		0.885504,
		0.840161,
		0.787425,
		0.728142,
		0.663389,
		0.594422,
		0.522638,
		0.449532,
		0.376679,
		0.3057,
		0.238249,
		0.175995,
		0.120607,
		0.073747,
		0.037057,
		0.012153,
		0.000621,
		0.001783,
		0.02262,
		0.063557,
		0.120607,
		0.189994,
		0.268167,
		0.351814,
		0.437873,
		0.523558,
		0.606372,
		0.684139,
		0.755028,
		0.817592,
		0.870801,
		0.914095,
		0.947416,
		0.971256,
		0.986666,
		0.995229,
		0.998936,
		0.999925,
		1,
		0.999927,
		0.998982,
		0.995521,
		0.987705,
		0.973927,
		0.953007,
		0.924251,
		0.887434,
		0.842738,
		0.790699,
		0.732133,
		0.668086,
		0.599783,
		0.528585,
		0.455958,
		0.383441,
		0.312626,
		0.245133,
		0.182603,
		0.126677,
		0.078987,
		0.04115,
		0.014759,
		0.001374,
		0.003111,
		0.026425,
		0.069138,
		0.127354,
		0.197385,
		0.275766,
		0.359265,
		0.444897,
		0.529947,
		0.611985,
		0.688896,
		0.758906,
		0.820615,
		0.873037,
		0.915644,
		0.948403,
		0.971816,
		0.986936,
		0.995329,
		0.99896,
		0.999927,
		1,
		0.999929,
		0.999004,
		0.995614,
		0.987949,
		0.974424,
		0.953866,
		0.925579,
		0.889324,
		0.845265,
		0.793912,
		0.736055,
		0.672708,
		0.605064,
		0.534454,
		0.46231,
		0.39014,
		0.319503,
		0.251992,
		0.189216,
		0.132788,
		0.084314,
		0.045385,
		0.017568,
		0.002404,
		0.004772,
		0.030448,
		0.074843,
		0.134151,
		0.204772,
		0.283322,
		0.366646,
		0.451836,
		0.536244,
		0.617508,
		0.69357,
		0.76271,
		0.823577,
		0.875226,
		0.917159,
		0.949367,
		0.972363,
		0.987199,
		0.995426,
		0.998982,
		0.999929,
		1,
		0.99993,
		0.999025,
		0.995704,
		0.988187,
		0.974909,
		0.954706,
		0.926878,
		0.891175,
		0.847743,
		0.797066,
		0.739908,
		0.677255,
		0.610268,
		0.540245,
		0.468589,
		0.396775,
		0.326332,
		0.258824,
		0.19583,
		0.138935,
		0.08972,
		0.049751,
		0.02057,
		0.003696,
		0.006746,
		0.034672,
		0.08066,
		0.140991,
		0.212149,
		0.29083,
		0.373956,
		0.45869,
		0.542452,
		0.622942,
		0.698162,
		0.766443,
		0.826481,
		0.877369,
		0.918641,
		0.950308,
		0.972897,
		0.987455,
		0.995521,
		0.999004,
		0.99993,
		1,
		0.999932,
		0.999046,
		0.995792,
		0.98842,
		0.975383,
		0.955526,
		0.928149,
		0.892988,
		0.850172,
		0.800163,
		0.743696,
		0.681729,
		0.615394,
		0.545959,
		0.474793,
		0.403344,
		0.333111,
		0.265625,
		0.20244,
		0.145114,
		0.0952,
		0.05424,
		0.023754,
		0.005236,
		0.009016,
		0.039085,
		0.086581,
		0.147869,
		0.219514,
		0.29829,
		0.381194,
		0.465459,
		0.54857,
		0.628289,
		0.702674,
		0.770107,
		0.829327,
		0.879467,
		0.92009,
		0.951229,
		0.973418,
		0.987705,
		0.995614,
		0.999025,
		0.999932,
		1,
		0.999933,
		0.999066,
		0.995877,
		0.988646,
		0.975845,
		0.956329,
		0.929393,
		0.894765,
		0.852555,
		0.803203,
		0.747418,
		0.686131,
		0.620445,
		0.551596,
		0.480925,
		0.409849,
		0.339837,
		0.272394,
		0.209044,
		0.151319,
		0.100747,
		0.058845,
		0.02711,
		0.007014,
		0.011564,
		0.043674,
		0.092596,
		0.154777,
		0.226862,
		0.3057,
		0.38836,
		0.472143,
		0.5546,
		0.63355,
		0.707107,
		0.773702,
		0.832117,
		0.881522,
		0.921508,
		0.952128,
		0.973927,
		0.987949,
		0.995704,
		0.999046,
		0.999933,
		1,
		0.999935,
		0.999086,
		0.995961,
		0.988867,
		0.976297,
		0.957114,
		0.930611,
		0.896506,
		0.854892,
		0.806187,
		0.751077,
		0.690463,
		0.62542,
		0.557157,
		0.486985,
		0.416289,
		0.346512,
		0.27913,
		0.21564,
		0.157547,
		0.106356,
		0.063557,
		0.030627,
		0.009016,
		0.000146,
		0.000188,
		0.014374,
		0.048428,
		0.098696,
		0.16171,
		0.234189,
		0.313057,
		0.395453,
		0.478744,
		0.560543,
		0.638727,
		0.711463,
		0.77723,
		0.834852,
		0.883534,
		0.922895,
		0.953007,
		0.974424,
		0.988187,
		0.995792,
		0.999066,
		0.999935,
		1,
		0.999936,
		0.999105,
		0.996042,
		0.989083,
		0.976739,
		0.957881,
		0.931803,
		0.898211,
		0.857183,
		0.809117,
		0.754672,
		0.694725,
		0.630323,
		0.562644,
		0.492972,
		0.422664,
		0.353134,
		0.28583,
		0.222223,
		0.163793,
		0.112019,
		0.06837,
		0.034297,
		0.011231,
		0.000573,
		0.000737,
		0.01743,
		0.053333,
		0.104874,
		0.168662,
		0.241492,
		0.32036,
		0.402472,
		0.485261,
		0.566399,
		0.643821,
		0.715743,
		0.780692,
		0.837533,
		0.885504,
		0.924251,
		0.953866,
		0.974909,
		0.98842,
		0.995877,
		0.999086,
		0.999936,
		1,
		0.999937,
		0.999123,
		0.996121,
		0.989293,
		0.97717,
		0.958631,
		0.93297,
		0.899882,
		0.859431,
		0.811993,
		0.758206,
		0.698919,
		0.635152,
		0.568056,
		0.498887,
		0.428973,
		0.359701,
		0.292492,
		0.228792,
		0.170054,
		0.117733,
		0.073275,
		0.03811,
		0.013648,
		0.001269,
		0.001628,
		0.020718,
		0.05838,
		0.111122,
		0.175628,
		0.248768,
		0.327607,
		0.409418,
		0.491695,
		0.57217,
		0.648832,
		0.719949,
		0.78409,
		0.840161,
		0.887434,
		0.925579,
		0.954706,
		0.975383,
		0.988646,
		0.995961,
		0.999105,
		0.999937,
		1,
		0.999939,
		0.999141,
		0.996199,
		0.989499,
		0.977591,
		0.959365,
		0.934112,
		0.90152,
		0.861636,
		0.814818,
		0.76168,
		0.703046,
		0.63991,
		0.573396,
		0.504731,
		0.435217,
		0.366214,
		0.299116,
		0.235343,
		0.176325,
		0.123492,
		0.078267,
		0.042059,
		0.016256,
		0.002221,
		0.002844,
		0.024223,
		0.063557,
		0.117431,
		0.182603,
		0.256014,
		0.334797,
		0.416289,
		0.498046,
		0.577857,
		0.653763,
		0.724081,
		0.787425,
		0.842738,
		0.889324,
		0.926878,
		0.955526,
		0.975845,
		0.988867,
		0.996042,
		0.999123,
		0.999939,
		1,
		0.99994,
		0.999159,
		0.996274,
		0.9897,
		0.978003,
		0.960083,
		0.935231,
		0.903125,
		0.863799,
		0.817592,
		0.765094,
		0.707107,
		0.644597,
		0.578663,
		0.510503,
		0.441396,
		0.372671,
		0.3057,
		0.241875,
		0.182603,
		0.129291,
		0.083339,
		0.046135,
		0.019046,
		0.003417,
		0.004366,
		0.027933,
		0.068856,
		0.123796,
		0.189584,
		0.263228,
		0.341929,
		0.423087,
		0.504315,
		0.583461,
		0.658615,
		0.728142,
		0.790699,
		0.845265,
		0.891175,
		0.928149,
		0.956329,
		0.976297,
		0.989083,
		0.996121,
		0.999141,
		0.99994,
		1,
		0.999941,
		0.999176,
		0.996348,
		0.989896,
		0.978406,
		0.960786,
		0.936326,
		0.904698,
		0.865921,
		0.820315,
		0.76845,
		0.711103,
		0.649214,
		0.583858,
		0.516206,
		0.447509,
		0.379073,
		0.312242,
		0.248386,
		0.188885,
		0.135126,
		0.088485,
		0.05033,
		0.022008,
		0.004845,
		0.006177,
		0.031834,
		0.074267,
		0.13021,
		0.196564,
		0.270407,
		0.349001,
		0.429809,
		0.510503,
		0.588982,
		0.663389,
		0.732133,
		0.793912,
		0.847743,
		0.892988,
		0.929393,
		0.957114,
		0.976739,
		0.989293,
		0.996199,
		0.999159,
		0.999941,
		1,
		0.999943,
		0.999193,
		0.996419,
		0.990087,
		0.978799,
		0.961473,
		0.937398,
		0.90624,
		0.868003,
		0.82299,
		0.771749,
		0.715035,
		0.653763,
		0.588982,
		0.521839,
		0.453557,
		0.385418,
		0.318742,
		0.254872,
		0.195168,
		0.140991,
		0.093699,
		0.054636,
		0.025132,
		0.006493,
		0.008262,
		0.035914,
		0.079781,
		0.136666,
		0.203541,
		0.277548,
		0.356013,
		0.436458,
		0.516611,
		0.594422,
		0.668086,
		0.736055,
		0.797066,
		0.850172,
		0.894765,
		0.930611,
		0.957881,
		0.97717,
		0.989499,
		0.996274,
		0.999176,
		0.999943,
		1,
		0.999944,
		0.99921,
		0.99649,
		0.990274,
		0.979184,
		0.962145,
		0.938448,
		0.907751,
		0.870045,
		0.825616,
		0.774992,
		0.718904,
		0.658245,
		0.594037,
		0.527402,
		0.45954,
		0.391707,
		0.325198,
		0.261333,
		0.201449,
		0.146884,
		0.098976,
		0.059048,
		0.02841,
		0.008351,
		0.000135,
		0.010606,
		0.040162,
		0.085389,
		0.14316,
		0.210511,
		0.28465,
		0.362964,
		0.443032,
		0.522638,
		0.599783,
		0.672708,
		0.739908,
		0.800163,
		0.852555,
		0.896506,
		0.931803,
		0.958631,
		0.977591,
		0.9897,
		0.996348,
		0.999193,
		0.999944,
		1,
		0.999945,
		0.999225,
		0.996558,
		0.990456,
		0.97956,
		0.962803,
		0.939477,
		0.909232,
		0.872049,
		0.828196,
		0.77818,
		0.722712,
		0.66266,
		0.599022,
		0.532897,
		0.465459,
		0.397939,
		0.331609,
		0.267766,
		0.207724,
		0.1528,
		0.10431,
		0.063557,
		0.031834,
		0.010409,
		0.00053,
		0.000172,
		0.013194,
		0.044568,
		0.091084,
		0.149684,
		0.21747,
		0.29171,
		0.369853,
		0.449532,
		0.528585,
		0.605064,
		0.677255,
		0.743696,
		0.803203,
		0.854892,
		0.898211,
		0.93297,
		0.959365,
		0.978003,
		0.989896,
		0.996419,
		0.99921,
		0.999945,
		1,
		0.999946,
		0.999241,
		0.996625,
		0.990634,
		0.979927,
		0.963446,
		0.940484,
		0.910685,
		0.874016,
		0.830729,
		0.781314,
		0.726459,
		0.667009,
		0.603939,
		0.538323,
		0.471312,
		0.404113,
		0.337974,
		0.27417,
		0.213992,
		0.158736,
		0.109696,
		0.068159,
		0.035395,
		0.012657,
		0.001176,
		0.000675,
		0.016011,
		0.049119,
		0.096858,
		0.156234,
		0.224414,
		0.298728,
		0.376679,
		0.455958,
		0.534454,
		0.610268,
		0.681729,
		0.747418,
		0.806187,
		0.857183,
		0.899882,
		0.934112,
		0.960083,
		0.978406,
		0.990087,
		0.99649,
		0.999225,
		0.999946,
		1,
		0.999947,
		0.999256,
		0.99669,
		0.990808,
		0.980287,
		0.964076,
		0.941471,
		0.912108,
		0.875946,
		0.833218,
		0.784396,
		0.730146,
		0.671294,
		0.608789,
		0.543682,
		0.477101,
		0.41023,
		0.344293,
		0.280543,
		0.22025,
		0.164687,
		0.11513,
		0.072845,
		0.039085,
		0.015085,
		0.002059,
		0.001493,
		0.019046,
		0.053808,
		0.102704,
		0.162806,
		0.231342,
		0.3057,
		0.383441,
		0.46231,
		0.540245,
		0.615394,
		0.686131,
		0.751077,
		0.809117,
		0.859431,
		0.90152,
		0.935231,
		0.960786,
		0.978799,
		0.990274,
		0.996558,
		0.999241,
		0.999947,
		1,
		0.999948,
		0.999271,
		0.996753,
		0.990978,
		0.980638,
		0.964693,
		0.942438,
		0.913505,
		0.877839,
		0.835662,
		0.787425,
		0.733775,
		0.675515,
		0.613572,
		0.548975,
		0.482827,
		0.416289,
		0.350565,
		0.286884,
		0.226495,
		0.170651,
		0.120607,
		0.077611,
		0.042898,
		0.017685,
		0.003169,
		0.00261,
		0.022285,
		0.058623,
		0.108615,
		0.169394,
		0.238249,
		0.312626,
		0.39014,
		0.468589,
		0.545959,
		0.620445,
		0.690463,
		0.754672,
		0.811993,
		0.861636,
		0.903125,
		0.936326,
		0.961473,
		0.979184,
		0.990456,
		0.996625,
		0.999256,
		0.999948,
		1,
		0.999949,
		0.999285,
		0.996815,
		0.991144,
		0.980982,
		0.965297,
		0.943385,
		0.914874,
		0.879698,
		0.838063,
		0.790404,
		0.737347,
		0.679673,
		0.618289,
		0.554201,
		0.488488,
		0.422291,
		0.356789,
		0.293191,
		0.232725,
		0.176624,
		0.126123,
		0.082451,
		0.046826,
		0.020447,
		0.004495,
		0.00401,
		0.025717,
		0.063557,
		0.114585,
		0.175995,
		0.245133,
		0.319503,
		0.396775,
		0.474793,
		0.551596,
		0.62542,
		0.694725,
		0.758206,
		0.814818,
		0.863799,
		0.904698,
		0.937398,
		0.962145,
		0.97956,
		0.990634,
		0.99669,
		0.999271,
		0.999949,
		1,
		0.99995,
		0.999299,
		0.996876,
		0.991307,
		0.981318,
		0.965887,
		0.944313,
		0.916216,
		0.881522,
		0.840421,
		0.793332,
		0.740862,
		0.68377,
		0.622942,
		0.559361,
		0.494086,
		0.428234,
		0.362964,
		0.299464,
		0.238939,
		0.182603,
		0.131674,
		0.08736,
		0.050862,
		0.023363,
		0.006028,
		0.005678,
		0.02933,
		0.068601,
		0.120607,
		0.182603,
		0.251992,
		0.326332,
		0.403344,
		0.480925,
		0.557157,
		0.630323,
		0.698919,
		0.76168,
		0.817592,
		0.865921,
		0.90624,
		0.938448,
		0.962803,
		0.979927,
		0.990808,
		0.996753,
		0.999285,
		0.99995,
		1,
		0.999951,
		0.999313,
		0.996935,
		0.991466,
		0.981647,
		0.966466,
		0.945222,
		0.917532,
		0.883312,
		0.842738,
		0.796212,
		0.744321,
		0.687806,
		0.62753,
		0.564456,
		0.499621,
		0.43412,
		0.369091,
		0.3057,
		0.245133,
		0.188586,
		0.137255,
		0.092332,
		0.055,
		0.026425,
		0.007758,
		0.000125,
		0.0076,
		0.033113,
		0.073747,
		0.126677,
		0.189216,
		0.258824,
		0.333111,
		0.409849,
		0.486985,
		0.562644,
		0.635152,
		0.703046,
		0.765094,
		0.820315,
		0.868003,
		0.907751,
		0.939477,
		0.963446,
		0.980287,
		0.990978,
		0.996815,
		0.999299,
		0.999951,
		1,
		0.999952,
		0.999327,
		0.996992,
		0.991621,
		0.981968,
		0.967032,
		0.946113,
		0.918824,
		0.88507,
		0.845015,
		0.799043,
		0.747726,
		0.691782,
		0.632056,
		0.569487,
		0.505093,
		0.439948,
		0.375167,
		0.311899,
		0.251308,
		0.19457,
		0.142864,
		0.097363,
		0.059234,
		0.029626,
		0.009675,
		0.000492,
		0.009762,
		0.037057,
		0.078987,
		0.132788,
		0.19583,
		0.265625,
		0.339837,
		0.416289,
		0.492972,
		0.568056,
		0.63991,
		0.707107,
		0.76845,
		0.82299,
		0.870045,
		0.909232,
		0.940484,
		0.964076,
		0.980638,
		0.991144,
		0.996876,
		0.999313,
		0.999952,
		1,
		0.999953,
		0.99934,
		0.997049,
		0.991772,
		0.982283,
		0.967586,
		0.946986,
		0.92009,
		0.886795,
		0.847251,
		0.801828,
		0.751077,
		0.695699,
		0.636519,
		0.574455,
		0.510503,
		0.445718,
		0.381194,
		0.31806,
		0.25746,
		0.200552,
		0.148496,
		0.102448,
		0.063557,
		0.032959,
		0.011771,
		0.001092,
		0.000158,
		0.012153,
		0.04115,
		0.084314,
		0.138935,
		0.20244,
		0.272394,
		0.346512,
		0.422664,
		0.498887,
		0.573396,
		0.644597,
		0.711103,
		0.771749,
		0.825616,
		0.872049,
		0.910685,
		0.941471,
		0.964693,
		0.980982,
		0.991307,
		0.996935,
		0.999327,
		0.999953,
		1,
		0.999954,
		0.999352,
		0.997104,
		0.99192,
		0.982591,
		0.968129,
		0.947842,
		0.921332,
		0.888488,
		0.849448,
		0.804566,
		0.754375,
		0.699558,
		0.64092,
		0.579359,
		0.515852,
		0.45143,
		0.387171,
		0.324181,
		0.263588,
		0.206529,
		0.154148,
		0.107582,
		0.067965,
		0.036415,
		0.014037,
		0.001913,
		0.000621,
		0.014759,
		0.045385,
		0.08972,
		0.145114,
		0.209044,
		0.27913,
		0.353134,
		0.428973,
		0.504731,
		0.578663,
		0.649214,
		0.715035,
		0.774992,
		0.828196,
		0.874016,
		0.912108,
		0.942438,
		0.965297,
		0.981318,
		0.991466,
		0.996992,
		0.99934,
		0.999954,
		1,
		0.999955,
		0.999365,
		0.997158,
		0.992065,
		0.982892,
		0.968661,
		0.94868,
		0.922551,
		0.890151,
		0.851607,
		0.807258,
		0.757621,
		0.703361,
		0.645261,
		0.584202,
		0.521138,
		0.457085,
		0.393097,
		0.330263,
		0.269691,
		0.2125,
		0.159817,
		0.112762,
		0.072452,
		0.039989,
		0.016464,
		0.002946,
		0.001374,
		0.017568,
		0.049751,
		0.0952,
		0.151319,
		0.21564,
		0.28583,
		0.359701,
		0.435217,
		0.510503,
		0.583858,
		0.653763,
		0.718904,
		0.77818,
		0.830729,
		0.875946,
		0.913505,
		0.943385,
		0.965887,
		0.981647,
		0.991621,
		0.997049,
		0.999352,
		0.999955,
		1,
		0.999956,
		0.999377,
		0.99721,
		0.992207,
		0.983187,
		0.969182,
		0.949502,
		0.923746,
		0.891784,
		0.853729,
		0.809907,
		0.760817,
		0.707107,
		0.649542,
		0.588982,
		0.526364,
		0.462682,
		0.398972,
		0.336303,
		0.275766,
		0.218463,
		0.1655,
		0.117983,
		0.077011,
		0.043674,
		0.019046,
		0.004182,
		0.002404,
		0.02057,
		0.05424,
		0.100747,
		0.157547,
		0.222223,
		0.292492,
		0.366214,
		0.441396,
		0.516206,
		0.588982,
		0.658245,
		0.722712,
		0.781314,
		0.833218,
		0.877839,
		0.914874,
		0.944313,
		0.966466,
		0.981968,
		0.991772,
		0.997104,
		0.999365,
		0.999956,
		1,
		0.999957,
		0.999389,
		0.997262,
		0.992346,
		0.983475,
		0.969692,
		0.950308,
		0.924919,
		0.893386,
		0.855814,
		0.812511,
		0.763962,
		0.710798,
		0.653763,
		0.593702,
		0.531529,
		0.468221,
		0.404796,
		0.342303,
		0.281814,
		0.224414,
		0.171193,
		0.123241,
		0.08164,
		0.047464,
		0.021774,
		0.005611,
		0.003696,
		0.023754,
		0.058845,
		0.106356,
		0.163793,
		0.228792,
		0.299116,
		0.372671,
		0.447509,
		0.521839,
		0.594037,
		0.66266,
		0.726459,
		0.784396,
		0.835662,
		0.879698,
		0.916216,
		0.945222,
		0.967032,
		0.982283,
		0.99192,
		0.997158,
		0.999377,
		0.999957,
		1,
		0.999958,
		0.9994,
		0.997312,
		0.992481,
		0.983758,
		0.970192,
		0.951098,
		0.926069,
		0.89496,
		0.857862,
		0.815072,
		0.767059,
		0.714434,
		0.657927,
		0.598361,
		0.536635,
		0.473704,
		0.410568,
		0.34826,
		0.287832,
		0.230353,
		0.176896,
		0.128533,
		0.086332,
		0.051353,
		0.024641,
		0.007225,
		0.000116,
		0.005236,
		0.02711,
		0.063557,
		0.112019,
		0.170054,
		0.235343,
		0.3057,
		0.379073,
		0.453557,
		0.527402,
		0.599022,
		0.667009,
		0.730146,
		0.787425,
		0.838063,
		0.881522,
		0.917532,
		0.946113,
		0.967586,
		0.982591,
		0.992065,
		0.99721,
		0.999389,
		0.999958,
		1,
		0.999959,
		0.999412,
		0.997361,
		0.992614,
		0.984034,
		0.970681,
		0.951873,
		0.927198,
		0.896506,
		0.859876,
		0.817592,
		0.770107,
		0.718017,
		0.662033,
		0.602961,
		0.541681,
		0.479129,
		0.416289,
		0.354174,
		0.29382,
		0.236278,
		0.182603,
		0.133854,
		0.091084,
		0.055335,
		0.027641,
		0.009016,
		0.000458,
		0.007014,
		0.030627,
		0.06837,
		0.117733,
		0.176325,
		0.241875,
		0.312242,
		0.385418,
		0.45954,
		0.532897,
		0.603939,
		0.671294,
		0.733775,
		0.790404,
		0.840421,
		0.883312,
		0.918824,
		0.946986,
		0.968129,
		0.982892,
		0.992207,
		0.997262,
		0.9994,
		0.999959,
		1,
		0.999959,
		0.999423,
		0.997409,
		0.992743,
		0.984305,
		0.971161,
		0.952633,
		0.928306,
		0.898023,
		0.861854,
		0.820069,
		0.773107,
		0.721547,
		0.666082,
		0.607502,
		0.546667,
		0.484498,
		0.421959,
		0.360045,
		0.299776,
		0.242186,
		0.188314,
		0.139203,
		0.09589,
		0.059405,
		0.030765,
		0.010974,
		0.001017,
		0.009016,
		0.034297,
		0.073275,
		0.123492,
		0.182603,
		0.248386,
		0.318742,
		0.391707,
		0.465459,
		0.538323,
		0.608789,
		0.675515,
		0.737347,
		0.793332,
		0.842738,
		0.88507,
		0.92009,
		0.947842,
		0.968661,
		0.983187,
		0.992346,
		0.997312,
		0.999412,
		0.999959,
		1,
		0.99996,
		0.999434,
		0.997456,
		0.99287,
		0.984569,
		0.971631,
		0.953377,
		0.929393,
		0.899514,
		0.863799,
		0.822507,
		0.776061,
		0.725025,
		0.670076,
		0.611985,
		0.551596,
		0.489811,
		0.427577,
		0.365872,
		0.3057,
		0.248076,
		0.194026,
		0.144575,
		0.100747,
		0.063557,
		0.034009,
		0.013093,
		0.001783,
		0.000146,
		0.011231,
		0.03811,
		0.078267,
		0.129291,
		0.188885,
		0.254872,
		0.325198,
		0.397939,
		0.471312,
		0.543682,
		0.613572,
		0.679673,
		0.740862,
		0.796212,
		0.845015,
		0.886795,
		0.921332,
		0.94868,
		0.969182,
		0.983475,
		0.992481,
		0.997361,
		0.999423,
		0.99996,
		1,
		0.999961,
		0.999444,
		0.997501,
		0.992994,
		0.984829,
		0.972091,
		0.954108,
		0.93046,
		0.900978,
		0.865711,
		0.824905,
		0.778969,
		0.728452,
		0.674014,
		0.61641,
		0.556466,
		0.495067,
		0.433143,
		0.371655,
		0.31159,
		0.253947,
		0.199736,
		0.149968,
		0.105651,
		0.067787,
		0.037366,
		0.015366,
		0.002747,
		0.000573,
		0.013648,
		0.042059,
		0.083339,
		0.135126,
		0.195168,
		0.261333,
		0.331609,
		0.404113,
		0.477101,
		0.548975,
		0.618289,
		0.68377,
		0.744321,
		0.799043,
		0.847251,
		0.888488,
		0.922551,
		0.949502,
		0.969692,
		0.983758,
		0.992614,
		0.997409,
		0.999434,
		0.999961,
		1,
		0.999962,
		0.999454,
		0.997546,
		0.993115,
		0.985083,
		0.972543,
		0.954824,
		0.931507,
		0.902416,
		0.86759,
		0.827263,
		0.781832,
		0.731828,
		0.677898,
		0.620779,
		0.561279,
		0.500268,
		0.438658,
		0.377394,
		0.317445,
		0.259797,
		0.205443,
		0.155379,
		0.110599,
		0.07209,
		0.04083,
		0.017784,
		0.003901,
		0.001269,
		0.016256,
		0.046135,
		0.088485,
		0.140991,
		0.201449,
		0.267766,
		0.337974,
		0.41023,
		0.482827,
		0.554201,
		0.622942,
		0.687806,
		0.747726,
		0.801828,
		0.849448,
		0.890151,
		0.923746,
		0.950308,
		0.970192,
		0.984034,
		0.992743,
		0.997456,
		0.999444,
		0.999962,
		1,
		0.999962,
		0.999464,
		0.99759,
		0.993234,
		0.985332,
		0.972985,
		0.955526,
		0.932535,
		0.903828,
		0.869437,
		0.829583,
		0.78465,
		0.735156,
		0.681729,
		0.625091,
		0.566035,
		0.505413,
		0.444121,
		0.383087,
		0.323265,
		0.265625,
		0.211144,
		0.160804,
		0.115585,
		0.076461,
		0.044395,
		0.020342,
		0.005236,
		0.002221,
		0.019046,
		0.05033,
		0.093699,
		0.146884,
		0.207724,
		0.27417,
		0.344293,
		0.416289,
		0.488488,
		0.559361,
		0.62753,
		0.691782,
		0.751077,
		0.804566,
		0.851607,
		0.891784,
		0.924919,
		0.951098,
		0.970681,
		0.984305,
		0.99287,
		0.997501,
		0.999454,
		0.999962,
		1,
		0.999963,
		0.999474,
		0.997633,
		0.99335,
		0.985575,
		0.973418,
		0.956215,
		0.933544,
		0.905215,
		0.871252,
		0.831866,
		0.787425,
		0.738434,
		0.685507,
		0.629348,
		0.570735,
		0.510503,
		0.449532,
		0.388735,
		0.32905,
		0.271429,
		0.216838,
		0.166242,
		0.120607,
		0.080895,
		0.048056,
		0.023032,
		0.006746,
		0.000108,
		0.003417,
		0.022008,
		0.054636,
		0.098976,
		0.1528,
		0.213992,
		0.280543,
		0.350565,
		0.422291,
		0.494086,
		0.564456,
		0.632056,
		0.695699,
		0.754375,
		0.807258,
		0.853729,
		0.893386,
		0.926069,
		0.951873,
		0.971161,
		0.984569,
		0.992994,
		0.997546,
		0.999464,
		0.999963,
		1,
		0.999964,
		0.999484,
		0.997675,
		0.993464,
		0.985814,
		0.973843,
		0.956891,
		0.934534,
		0.906578,
		0.873037,
		0.834112,
		0.790157,
		0.741665,
		0.689233,
		0.63355,
		0.575379,
		0.515539,
		0.454892,
		0.394338,
		0.334797,
		0.277209,
		0.222522,
		0.171689,
		0.125662,
		0.085389,
		0.051808,
		0.025848,
		0.008421,
		0.000428,
		0.004845,
		0.025132,
		0.059048,
		0.10431,
		0.158736,
		0.22025,
		0.286884,
		0.356789,
		0.428234,
		0.499621,
		0.569487,
		0.636519,
		0.699558,
		0.757621,
		0.809907,
		0.855814,
		0.89496,
		0.927198,
		0.952633,
		0.971631,
		0.984829,
		0.993115,
		0.99759,
		0.999474,
		0.999964,
		1,
		0.999964,
		0.999493,
		0.997715,
		0.993576,
		0.986047,
		0.97426,
		0.957554,
		0.935507,
		0.907917,
		0.874792,
		0.836321,
		0.792848,
		0.744848,
		0.692907,
		0.637698,
		0.579968,
		0.52052,
		0.460201,
		0.399895,
		0.340507,
		0.282963,
		0.228195,
		0.177144,
		0.130747,
		0.089938,
		0.055645,
		0.028784,
		0.010256,
		0.000949,
		0.006493,
		0.02841,
		0.063557,
		0.109696,
		0.164687,
		0.226495,
		0.293191,
		0.362964,
		0.43412,
		0.505093,
		0.574455,
		0.64092,
		0.703361,
		0.760817,
		0.812511,
		0.857862,
		0.896506,
		0.928306,
		0.953377,
		0.972091,
		0.985083,
		0.993234,
		0.997633,
		0.999484,
		0.999964,
		1,
		0.999965,
		0.999502,
		0.997755,
		0.993685,
		0.986276,
		0.974668,
		0.958205,
		0.936461,
		0.909232,
		0.876517,
		0.838495,
		0.795497,
		0.747985,
		0.696531,
		0.641793,
		0.584502,
		0.525446,
		0.465459,
		0.405406,
		0.34618,
		0.28869,
		0.233856,
		0.182603,
		0.135857,
		0.094539,
		0.059563,
		0.031834,
		0.012242,
		0.001665,
		0.008351,
		0.031834,
		0.068159,
		0.11513,
		0.170651,
		0.232725,
		0.299464,
		0.369091,
		0.439948,
		0.510503,
		0.579359,
		0.645261,
		0.707107,
		0.763962,
		0.815072,
		0.859876,
		0.898023,
		0.929393,
		0.954108,
		0.972543,
		0.985332,
		0.99335,
		0.997675,
		0.999493,
		0.999965,
		1,
		0.999966,
		0.999511,
		0.997795,
		0.993791,
		0.986501,
		0.975068,
		0.958843,
		0.937398,
		0.910525,
		0.878214,
		0.840634,
		0.798105,
		0.751077,
		0.700105,
		0.645835,
		0.588982,
		0.53032,
		0.470665,
		0.410871,
		0.351814,
		0.294389,
		0.239503,
		0.188066,
		0.140991,
		0.099188,
		0.063557,
		0.034993,
		0.014374,
		0.002567,
		0.000135,
		0.010409,
		0.035395,
		0.072845,
		0.120607,
		0.176624,
		0.238939,
		0.3057,
		0.375167,
		0.445718,
		0.515852,
		0.584202,
		0.649542,
		0.710798,
		0.767059,
		0.817592,
		0.861854,
		0.899514,
		0.93046,
		0.954824,
		0.972985,
		0.985575,
		0.993464,
		0.997715,
		0.999502,
		0.999966,
		1,
		0.999966,
		0.99952,
		0.997833,
		0.993896,
		0.98672,
		0.97546,
		0.959469,
		0.938318,
		0.911794,
		0.879882,
		0.842738,
		0.800673,
		0.754123,
		0.70363,
		0.649825,
		0.593409,
		0.535139,
		0.475821,
		0.416289,
		0.357409,
		0.300059,
		0.245133,
		0.193529,
		0.146146,
		0.103881,
		0.067623,
		0.038254,
		0.016644,
		0.003647,
		0.00053,
		0.012657,
		0.039085,
		0.077611,
		0.126123,
		0.182603,
		0.245133,
		0.311899,
		0.381194,
		0.45143,
		0.521138,
		0.588982,
		0.653763,
		0.714434,
		0.770107,
		0.820069,
		0.863799,
		0.900978,
		0.931507,
		0.955526,
		0.973418,
		0.985814,
		0.993576,
		0.997755,
		0.999511,
		0.999966,
		1,
		0.999967,
		0.999529,
		0.99787,
		0.993998,
		0.986936,
		0.975845,
		0.960083,
		0.939222,
		0.913042,
		0.881522,
		0.844809,
		0.803203,
		0.757125,
		0.707107,
		0.653763,
		0.597782,
		0.539907,
		0.480925,
		0.421662,
		0.362964,
		0.3057,
		0.250747,
		0.198992,
		0.151319,
		0.108615,
		0.071756,
		0.041614,
		0.019046,
		0.004898,
		0.001176,
		0.015085,
		0.042898,
		0.082451,
		0.131674,
		0.188586,
		0.251308,
		0.31806,
		0.387171,
		0.457085,
		0.526364,
		0.593702,
		0.657927,
		0.718017,
		0.773107,
		0.822507,
		0.865711,
		0.902416,
		0.932535,
		0.956215,
		0.973843,
		0.986047,
		0.993685,
		0.997795,
		0.99952,
		0.999967,
		1,
		0.999968,
		0.999537,
		0.997907,
		0.994098,
		0.987147,
		0.976223,
		0.960686,
		0.940109,
		0.914268,
		0.883135,
		0.846847,
		0.805693,
		0.760084,
		0.710536,
		0.657651,
		0.602103,
		0.544621,
		0.48598,
		0.426988,
		0.36848,
		0.31131,
		0.256343,
		0.204451,
		0.156508,
		0.113386,
		0.075953,
		0.045066,
		0.021574,
		0.006313,
		0.000101,
		0.002059,
		0.017685,
		0.046826,
		0.08736,
		0.137255,
		0.19457,
		0.25746,
		0.324181,
		0.393097,
		0.462682,
		0.531529,
		0.598361,
		0.662033,
		0.721547,
		0.776061,
		0.824905,
		0.86759,
		0.903828,
		0.933544,
		0.956891,
		0.97426,
		0.986276,
		0.993791,
		0.997833,
		0.999529,
		0.999968,
		1,
		0.999968,
		0.999545,
		0.997943,
		0.994196,
		0.987353,
		0.976593,
		0.961278,
		0.94098,
		0.915473,
		0.884721,
		0.848853,
		0.808146,
		0.763,
		0.713918,
		0.661489,
		0.606372,
		0.549284,
		0.490984,
		0.432268,
		0.373956,
		0.316889,
		0.261919,
		0.209905,
		0.16171,
		0.118192,
		0.080209,
		0.048607,
		0.024223,
		0.007884,
		0.0004,
		0.003169,
		0.020447,
		0.050862,
		0.092332,
		0.142864,
		0.200552,
		0.263588,
		0.330263,
		0.398972,
		0.468221,
		0.536635,
		0.602961,
		0.666082,
		0.725025,
		0.778969,
		0.827263,
		0.869437,
		0.905215,
		0.934534,
		0.957554,
		0.974668,
		0.986501,
		0.993896,
		0.99787,
		0.999537,
		0.999968,
		1,
		0.999969,
		0.999553,
		0.997978,
		0.994292,
		0.987556,
		0.976956,
		0.961859,
		0.941836,
		0.916658,
		0.886281,
		0.850827,
		0.810562,
		0.765874,
		0.717253,
		0.665277,
		0.61059,
		0.553895,
		0.495938,
		0.437501,
		0.379391,
		0.322436,
		0.267474,
		0.215353,
		0.166922,
		0.12303,
		0.08452,
		0.05223,
		0.026987,
		0.009606,
		0.000888,
		0.004495,
		0.023363,
		0.055,
		0.097363,
		0.148496,
		0.206529,
		0.269691,
		0.336303,
		0.404796,
		0.473704,
		0.541681,
		0.607502,
		0.670076,
		0.728452,
		0.781832,
		0.829583,
		0.871252,
		0.906578,
		0.935507,
		0.958205,
		0.975068,
		0.98672,
		0.993998,
		0.997907,
		0.999545,
		0.999969,
		1,
		0.999969,
		0.999561,
		0.998013,
		0.994386,
		0.987755,
		0.977312,
		0.962429,
		0.942676,
		0.917822,
		0.887815,
		0.852769,
		0.812941,
		0.768706,
		0.720543,
		0.669016,
		0.614757,
		0.558455,
		0.500843,
		0.442688,
		0.384786,
		0.327951,
		0.273008,
		0.220793,
		0.172143,
		0.127896,
		0.088884,
		0.055933,
		0.02986,
		0.011471,
		0.001559,
		0.006028,
		0.026425,
		0.059234,
		0.102448,
		0.154148,
		0.2125,
		0.275766,
		0.342303,
		0.410568,
		0.479129,
		0.546667,
		0.611985,
		0.674014,
		0.731828,
		0.78465,
		0.831866,
		0.873037,
		0.907917,
		0.936461,
		0.958843,
		0.97546,
		0.986936,
		0.994098,
		0.997943,
		0.999553,
		0.999969,
		1,
		0.99997,
		0.999569,
		0.998046,
		0.994478,
		0.987949,
		0.977661,
		0.962988,
		0.943502,
		0.918966,
		0.889324,
		0.854681,
		0.815284,
		0.771497,
		0.723788,
		0.672708,
		0.618874,
		0.562964,
		0.505698,
		0.447829,
		0.39014,
		0.333432,
		0.278519,
		0.226223,
		0.177371,
		0.132788,
		0.093296,
		0.05971,
		0.032838,
		0.013475,
		0.002404,
		0.007758,
		0.029626,
		0.063557,
		0.107582,
		0.159817,
		0.218463,
		0.281814,
		0.34826,
		0.416289,
		0.484498,
		0.551596,
		0.61641,
		0.677898,
		0.735156,
		0.787425,
		0.834112,
		0.874792,
		0.909232,
		0.937398,
		0.959469,
		0.975845,
		0.987147,
		0.994196,
		0.997978,
		0.999561,
		0.99997,
		1,
		0.99997,
		0.999576,
		0.998079,
		0.994569,
		0.98814,
		0.978003,
		0.963537,
		0.944313,
		0.92009,
		0.890808,
		0.856563,
		0.817592,
		0.774249,
		0.726989,
		0.676351,
		0.622942,
		0.567424,
		0.510503,
		0.452924,
		0.395453,
		0.33888,
		0.284006,
		0.231643,
		0.182603,
		0.137703,
		0.097753,
		0.063557,
		0.035914,
		0.015609,
		0.003417,
		0.000125,
		0.009675,
		0.032959,
		0.067965,
		0.112762,
		0.1655,
		0.224414,
		0.287832,
		0.354174,
		0.421959,
		0.489811,
		0.556466,
		0.620779,
		0.681729,
		0.738434,
		0.790157,
		0.836321,
		0.876517,
		0.910525,
		0.938318,
		0.960083,
		0.976223,
		0.987353,
		0.994292,
		0.998013,
		0.999569,
		0.99997,
		1,
		0.999971,
		0.999584,
		0.998111,
		0.994657,
		0.988327,
		0.978339,
		0.964076,
		0.945109,
		0.921195,
		0.892267,
		0.858415,
		0.819865,
		0.77696,
		0.730146,
		0.679948,
		0.62696,
		0.571833,
		0.51526,
		0.457972,
		0.400724,
		0.344293,
		0.289469,
		0.237049,
		0.187838,
		0.142639,
		0.102252,
		0.067471,
		0.039085,
		0.01787,
		0.004591,
		0.000492,
		0.011771,
		0.036415,
		0.072452,
		0.117983,
		0.171193,
		0.230353,
		0.29382,
		0.360045,
		0.427577,
		0.495067,
		0.561279,
		0.625091,
		0.685507,
		0.741665,
		0.792848,
		0.838495,
		0.878214,
		0.911794,
		0.939222,
		0.960686,
		0.976593,
		0.987556,
		0.994386,
		0.998046,
		0.999576,
		0.999971,
		1,
		0.999971,
		0.999591,
		0.998143,
		0.994744,
		0.988511,
		0.978669,
		0.964606,
		0.945892,
		0.922282,
		0.893703,
		0.860238,
		0.822103,
		0.779633,
		0.73326,
		0.683499,
		0.63093,
		0.576193,
		0.519969,
		0.462975,
		0.405954,
		0.349672,
		0.294906,
		0.242442,
		0.193074,
		0.147593,
		0.106789,
		0.071448,
		0.042346,
		0.020251,
		0.00592,
		9.5e-005,
		0.001092,
		0.014037,
		0.039989,
		0.077011,
		0.123241,
		0.176896,
		0.236278,
		0.299776,
		0.365872,
		0.433143,
		0.500268,
		0.566035,
		0.629348,
		0.689233,
		0.744848,
		0.795497,
		0.840634,
		0.879882,
		0.913042,
		0.940109,
		0.961278,
		0.976956,
		0.987755,
		0.994478,
		0.998079,
		0.999584,
		0.999971,
		1,
		0.999972,
		0.999598,
		0.998174,
		0.994829,
		0.988691,
		0.978993,
		0.965125,
		0.946661,
		0.92335,
		0.895116,
		0.862033,
		0.824309,
		0.782268,
		0.736332,
		0.687003,
		0.634852,
		0.580504,
		0.524629,
		0.467931,
		0.411143,
		0.355015,
		0.300316,
		0.247821,
		0.198309,
		0.152563,
		0.111363,
		0.075484,
		0.045692,
		0.022747,
		0.007397,
		0.000375,
		0.001913,
		0.016464,
		0.043674,
		0.08164,
		0.128533,
		0.182603,
		0.242186,
		0.3057,
		0.371655,
		0.438658,
		0.505413,
		0.570735,
		0.63355,
		0.692907,
		0.747985,
		0.798105,
		0.842738,
		0.881522,
		0.914268,
		0.94098,
		0.961859,
		0.977312,
		0.987949,
		0.994569,
		0.998111,
		0.999591,
		0.999972,
		1,
		0.999972,
		0.999605,
		0.998204,
		0.994912,
		0.988867,
		0.97931,
		0.965636,
		0.947416,
		0.9244,
		0.896506,
		0.863799,
		0.826481,
		0.784865,
		0.739362,
		0.690463,
		0.638727,
		0.584767,
		0.529241,
		0.472842,
		0.416289,
		0.360324,
		0.3057,
		0.253183,
		0.203541,
		0.157547,
		0.11597,
		0.079575,
		0.049119,
		0.025353,
		0.009016,
		0.000833,
		0.002946,
		0.019046,
		0.047464,
		0.086332,
		0.133854,
		0.188314,
		0.248076,
		0.31159,
		0.377394,
		0.444121,
		0.510503,
		0.575379,
		0.637698,
		0.696531,
		0.751077,
		0.800673,
		0.844809,
		0.883135,
		0.915473,
		0.941836,
		0.962429,
		0.977661,
		0.98814,
		0.994657,
		0.998143,
		0.999598,
		0.999972,
		1,
		0.999973,
		0.999611,
		0.998234,
		0.994994,
		0.98904,
		0.979622,
		0.966137,
		0.948158,
		0.925433,
		0.897873,
		0.865538,
		0.828621,
		0.787425,
		0.742351,
		0.693878,
		0.642555,
		0.588982,
		0.533806,
		0.477707,
		0.421394,
		0.365596,
		0.311056,
		0.258527,
		0.208769,
		0.162543,
		0.120607,
		0.083718,
		0.052623,
		0.028064,
		0.010771,
		0.001462,
		0.004182,
		0.021774,
		0.051353,
		0.091084,
		0.139203,
		0.194026,
		0.253947,
		0.317445,
		0.383087,
		0.449532,
		0.515539,
		0.579968,
		0.641793,
		0.700105,
		0.754123,
		0.803203,
		0.846847,
		0.884721,
		0.916658,
		0.942676,
		0.962988,
		0.978003,
		0.988327,
		0.994744,
		0.998174,
		0.999605,
		0.999973,
		1,
		0.999973,
		0.999618,
		0.998263,
		0.995074,
		0.98921,
		0.979927,
		0.966629,
		0.948887,
		0.926448,
		0.899218,
		0.867251,
		0.830729,
		0.789949,
		0.745299,
		0.69725,
		0.646337,
		0.59315,
		0.538323,
		0.482527,
		0.426457,
		0.370832,
		0.316383,
		0.263854,
		0.213992,
		0.167548,
		0.125272,
		0.08791,
		0.0562,
		0.030875,
		0.012657,
		0.002256,
		0.005611,
		0.024641,
		0.055335,
		0.09589,
		0.144575,
		0.199736,
		0.259797,
		0.323265,
		0.388735,
		0.454892,
		0.52052,
		0.584502,
		0.645835,
		0.70363,
		0.757125,
		0.805693,
		0.848853,
		0.886281,
		0.917822,
		0.943502,
		0.963537,
		0.978339,
		0.988511,
		0.994829,
		0.998204,
		0.999611,
		0.999973,
		1,
		0.999974,
		0.999624,
		0.998291,
		0.995152,
		0.989376,
		0.980227,
		0.967112,
		0.949604,
		0.927446,
		0.900541,
		0.868936,
		0.832806,
		0.792437,
		0.748208,
		0.700578,
		0.650073,
		0.59727,
		0.542794,
		0.487302,
		0.431479,
		0.376031,
		0.321682,
		0.269161,
		0.219207,
		0.172561,
		0.129963,
		0.092147,
		0.059846,
		0.033782,
		0.014668,
		0.003208,
		0.007225,
		0.027641,
		0.059405,
		0.100747,
		0.149968,
		0.205443,
		0.265625,
		0.32905,
		0.394338,
		0.460201,
		0.525446,
		0.588982,
		0.649825,
		0.707107,
		0.760084,
		0.808146,
		0.850827,
		0.887815,
		0.918966,
		0.944313,
		0.964076,
		0.978669,
		0.988691,
		0.994912,
		0.998234,
		0.999618,
		0.999974,
		1,
		0.999974,
		0.999631,
		0.998319,
		0.995229,
		0.98954,
		0.980522,
		0.967586,
		0.950308,
		0.928428,
		0.901844,
		0.870596,
		0.834852,
		0.794889,
		0.751077,
		0.703863,
		0.653763,
		0.601344,
		0.547218,
		0.492031,
		0.436458,
		0.381194,
		0.32695,
		0.274448,
		0.224414,
		0.17758,
		0.134676,
		0.096427,
		0.063557,
		0.03678,
		0.0168,
		0.004312,
		0.000116,
		0.009016,
		0.030765,
		0.063557,
		0.105651,
		0.155379,
		0.211144,
		0.271429,
		0.334797,
		0.399895,
		0.465459,
		0.53032,
		0.593409,
		0.653763,
		0.710536,
		0.763,
		0.810562,
		0.852769,
		0.889324,
		0.92009,
		0.945109,
		0.964606,
		0.978993,
		0.988867,
		0.994994,
		0.998263,
		0.999624,
		0.999974,
		1,
		0.999975,
		0.999637,
		0.998347,
		0.995304,
		0.9897,
		0.980811,
		0.968053,
		0.951001,
		0.929393,
		0.903125,
		0.87223,
		0.836868,
		0.797307,
		0.753907,
		0.707107,
		0.65741,
		0.605372,
		0.551596,
		0.496716,
		0.441396,
		0.38632,
		0.332189,
		0.279714,
		0.229612,
		0.182603,
		0.139409,
		0.100747,
		0.06733,
		0.039864,
		0.019046,
		0.000458,
		0.010974,
		0.034009,
		0.067787,
		0.110599,
		0.160804,
		0.216838,
		0.277209,
		0.340507,
		0.405406,
		0.470665,
		0.535139,
		0.597782,
		0.657651,
		0.713918,
		0.765874,
		0.812941,
		0.854681,
		0.890808,
		0.921195,
		0.945892,
		0.965125,
		0.97931,
		0.98904,
		0.995074,
		0.998291,
		0.999631,
		0.999975,
		1,
		0.999975,
		0.999643,
		0.998373,
		0.995378,
		0.989857,
		0.981094,
		0.96851,
		0.951681,
		0.930343,
		0.904386,
		0.873839,
		0.838854,
		0.79969,
		0.756699,
		0.710309,
		0.661012,
		0.609355,
		0.555928,
		0.501356,
		0.446292,
		0.391409,
		0.337397,
		0.284958,
		0.234798,
		0.187629,
		0.144161,
		0.105104,
		0.071162,
		0.043031,
		0.001017,
		0.013093,
		0.037366,
		0.07209,
		0.115585,
		0.166242,
		0.222522,
		0.282963,
		0.34618,
		0.410871,
		0.475821,
		0.539907,
		0.602103,
		0.661489,
		0.717253,
		0.768706,
		0.815284,
		0.856563,
		0.892267,
		0.922282,
		0.946661,
		0.965636,
		0.979622,
		0.98921,
		0.995152,
		0.998319,
		0.999637,
		0.999975,
		1,
		0.999976,
		0.999649,
		0.9984,
		0.99545,
		0.990011,
		0.981373,
		0.96896,
		0.95235,
		0.931276,
		0.905627,
		0.875423,
		0.84081,
		0.80204,
		0.759454,
		0.713469,
		0.66457,
		0.613292,
		0.560215,
		0.505952,
		0.451146,
		0.39646,
		0.342574,
		0.290179,
		0.239972,
		0.192655,
		0.14893,
		0.109496,
		0.075048,
		0.001783,
		0.015366,
		0.04083,
		0.076461,
		0.120607,
		0.171689,
		0.228195,
		0.28869,
		0.351814,
		0.416289,
		0.480925,
		0.544621,
		0.606372,
		0.665277,
		0.720543,
		0.771497,
		0.817592,
		0.858415,
		0.893703,
		0.92335,
		0.947416,
		0.966137,
		0.979927,
		0.989376,
		0.995229,
		0.998347,
		0.999643,
		0.999976,
		1,
		0.999976,
		0.999655,
		0.998425,
		0.995521,
		0.990162,
		0.981647,
		0.969402,
		0.953007,
		0.932195,
		0.906848,
		0.876983,
		0.842738,
		0.804357,
		0.762171,
		0.71659,
		0.668086,
		0.617185,
		0.564456,
		0.510503,
		0.455958,
		0.401474,
		0.34772,
		0.295377,
		0.245133,
		0.197681,
		0.153712,
		0.113919,
		0.002747,
		0.017784,
		0.044395,
		0.080895,
		0.125662,
		0.177144,
		0.233856,
		0.294389,
		0.357409,
		0.421662,
		0.48598,
		0.549284,
		0.61059,
		0.669016,
		0.723788,
		0.774249,
		0.819865,
		0.860238,
		0.895116,
		0.9244,
		0.948158,
		0.966629,
		0.980227,
		0.98954,
		0.995304,
		0.998373,
		0.999649,
		0.999976,
		1,
		0.999976,
		0.99966,
		0.998451,
		0.995591,
		0.99031,
		0.981915,
		0.969836,
		0.953653,
		0.933098,
		0.908049,
		0.878519,
		0.844638,
		0.806641,
		0.764852,
		0.71967,
		0.671559,
		0.621034,
		0.568653,
		0.515011,
		0.460729,
		0.40645,
		0.352834,
		0.300551,
		0.25028,
		0.202704,
		0.158507,
		0.003901,
		0.020342,
		0.048056,
		0.085389,
		0.130747,
		0.182603,
		0.239503,
		0.300059,
		0.362964,
		0.426988,
		0.490984,
		0.553895,
		0.614757,
		0.672708,
		0.726989,
		0.77696,
		0.822103,
		0.862033,
		0.896506,
		0.925433,
		0.948887,
		0.967112,
		0.980522,
		0.9897,
		0.995378,
		0.9984,
		0.999655,
		0.999976,
		1,
		0.999977,
		0.999666,
		0.998475,
		0.995659,
		0.990456,
		0.982179,
		0.970262,
		0.954288,
		0.933986,
		0.909232,
		0.880032,
		0.84651,
		0.808893,
		0.767497,
		0.722712,
		0.67499,
		0.624839,
		0.572806,
		0.519475,
		0.465459,
		0.411389,
		0.357915,
		0.3057,
		0.255412,
		0.207724,
		0.005236,
		0.023032,
		0.051808,
		0.089938,
		0.135857,
		0.188066,
		0.245133,
		0.3057,
		0.36848,
		0.432268,
		0.495938,
		0.558455,
		0.618874,
		0.676351,
		0.730146,
		0.779633,
		0.824309,
		0.863799,
		0.897873,
		0.926448,
		0.949604,
		0.967586,
		0.980811,
		0.989857,
		0.99545,
		0.998425,
		0.99966,
		0.999977,
		1,
		0.999977,
		0.999671,
		0.998499,
		0.995726,
		0.990599,
		0.982438,
		0.970681,
		0.954912,
		0.93486,
		0.910396,
		0.881522,
		0.848355,
		0.811114,
		0.770107,
		0.725714,
		0.67838,
		0.628601,
		0.576915,
		0.523896,
		0.470147,
		0.416289,
		0.362964,
		0.310823,
		0.260527,
		0.006746,
		0.025848,
		0.055645,
		0.094539,
		0.140991,
		0.193529,
		0.250747,
		0.31131,
		0.373956,
		0.437501,
		0.500843,
		0.562964,
		0.622942,
		0.679948,
		0.73326,
		0.782268,
		0.826481,
		0.865538,
		0.899218,
		0.927446,
		0.950308,
		0.968053,
		0.981094,
		0.990011,
		0.995521,
		0.998451,
		0.999666,
		0.999977,
		1,
		0.999978,
		0.999677,
		0.998523,
		0.995792,
		0.990739,
		0.982692,
		0.971093,
		0.955526,
		0.93572,
		0.911542,
		0.882989,
		0.850172,
		0.813304,
		0.772681,
		0.728678,
		0.681729,
		0.63232,
		0.58098,
		0.528274,
		0.474793,
		0.421152,
		0.36798,
		0.315921,
		0.000108,
		0.008421,
		0.028784,
		0.059563,
		0.099188,
		0.146146,
		0.198992,
		0.256343,
		0.316889,
		0.379391,
		0.442688,
		0.505698,
		0.567424,
		0.62696,
		0.683499,
		0.736332,
		0.784865,
		0.828621,
		0.867251,
		0.900541,
		0.928428,
		0.951001,
		0.96851,
		0.981373,
		0.990162,
		0.995591,
		0.998475,
		0.999671,
		0.999978,
		1,
		0.999978,
		0.999682,
		0.998546,
		0.995856,
		0.990877,
		0.982941,
		0.971498,
		0.95613,
		0.936566,
		0.91267,
		0.884434,
		0.851964,
		0.815463,
		0.775222,
		0.731605,
		0.685037,
		0.635997,
		0.585003,
		0.532609,
		0.479399,
		0.425977,
		0.372963,
		0.000428,
		0.010256,
		0.031834,
		0.063557,
		0.103881,
		0.151319,
		0.204451,
		0.261919,
		0.322436,
		0.384786,
		0.447829,
		0.510503,
		0.571833,
		0.63093,
		0.687003,
		0.739362,
		0.787425,
		0.830729,
		0.868936,
		0.901844,
		0.929393,
		0.951681,
		0.96896,
		0.981647,
		0.99031,
		0.995659,
		0.998499,
		0.999677,
		0.999978,
		1,
		0.999978,
		0.999687,
		0.998569,
		0.995919,
		0.991012,
		0.983187,
		0.971895,
		0.956724,
		0.937398,
		0.913781,
		0.885858,
		0.853729,
		0.817592,
		0.777728,
		0.734494,
		0.688306,
		0.639632,
		0.588982,
		0.536902,
		0.483964,
		0.430764,
		0.000949,
		0.012242,
		0.034993,
		0.067623,
		0.108615,
		0.156508,
		0.209905,
		0.267474,
		0.327951,
		0.39014,
		0.452924,
		0.51526,
		0.576193,
		0.634852,
		0.690463,
		0.742351,
		0.789949,
		0.832806,
		0.870596,
		0.903125,
		0.930343,
		0.95235,
		0.969402,
		0.981915,
		0.990456,
		0.995726,
		0.998523,
		0.999682,
		0.999978,
		1,
		0.999979,
		0.999692,
		0.998592,
		0.995981,
		0.991144,
		0.983428,
		0.972286,
		0.957307,
		0.938217,
		0.914874,
		0.88726,
		0.855469,
		0.819691,
		0.780201,
		0.737347,
		0.691535,
		0.643226,
		0.592919,
		0.541152,
		0.488488,
		0.001665,
		0.014374,
		0.038254,
		0.071756,
		0.113386,
		0.16171,
		0.215353,
		0.273008,
		0.333432,
		0.395453,
		0.457972,
		0.519969,
		0.580504,
		0.638727,
		0.693878,
		0.745299,
		0.792437,
		0.834852,
		0.87223,
		0.904386,
		0.931276,
		0.953007,
		0.969836,
		0.982179,
		0.990599,
		0.995792,
		0.998546,
		0.999687,
		0.999979,
		1,
		0.999979,
		0.999697,
		0.998614,
		0.996042,
		0.991275,
		0.983664,
		0.97267,
		0.957881,
		0.939022,
		0.91595,
		0.888641,
		0.857183,
		0.821761,
		0.782641,
		0.740163,
		0.694725,
		0.646778,
		0.596815,
		0.545361,
		0.002567,
		0.016644,
		0.041614,
		0.075953,
		0.118192,
		0.166922,
		0.220793,
		0.278519,
		0.33888,
		0.400724,
		0.462975,
		0.524629,
		0.584767,
		0.642555,
		0.69725,
		0.748208,
		0.794889,
		0.836868,
		0.873839,
		0.905627,
		0.932195,
		0.953653,
		0.970262,
		0.982438,
		0.990739,
		0.995856,
		0.998569,
		0.999692,
		0.999979,
		1,
		0.999979,
		0.999702,
		0.998635,
		0.996102,
		0.991403,
		0.983896,
		0.973047,
		0.958445,
		0.939815,
		0.917009,
		0.890001,
		0.858873,
		0.823803,
		0.785049,
		0.742944,
		0.697877,
		0.650291,
		0.600669,
		0.003647,
		0.019046,
		0.045066,
		0.080209,
		0.12303,
		0.172143,
		0.226223,
		0.284006,
		0.344293,
		0.405954,
		0.467931,
		0.529241,
		0.588982,
		0.646337,
		0.700578,
		0.751077,
		0.797307,
		0.838854,
		0.875423,
		0.906848,
		0.933098,
		0.954288,
		0.970681,
		0.982692,
		0.990877,
		0.995919,
		0.998592,
		0.999697,
		0.999979,
		1,
		0.99998,
		0.999707,
		0.998656,
		0.99616,
		0.991528,
		0.984125,
		0.973418,
		0.959001,
		0.940595,
		0.918052,
		0.891341,
		0.860539,
		0.825816,
		0.787425,
		0.745689,
		0.700991,
		0.653763,
		0.004898,
		0.021574,
		0.048607,
		0.08452,
		0.127896,
		0.177371,
		0.231643,
		0.289469,
		0.349672,
		0.411143,
		0.472842,
		0.533806,
		0.59315,
		0.650073,
		0.703863,
		0.753907,
		0.79969,
		0.84081,
		0.876983,
		0.908049,
		0.933986,
		0.954912,
		0.971093,
		0.982941,
		0.991012,
		0.995981,
		0.998614,
		0.999702,
		0.99998,
		1,
		0.99998,
		0.999711,
		0.998677,
		0.996218,
		0.991651,
		0.984349,
		0.973783,
		0.959546,
		0.941362,
		0.919079,
		0.892661,
		0.862181,
		0.827802,
		0.78977,
		0.7484,
		0.704067,
		0.006313,
		0.024223,
		0.05223,
		0.088884,
		0.132788,
		0.182603,
		0.237049,
		0.294906,
		0.355015,
		0.416289,
		0.477707,
		0.538323,
		0.59727,
		0.653763,
		0.707107,
		0.756699,
		0.80204,
		0.842738,
		0.878519,
		0.909232,
		0.93486,
		0.955526,
		0.971498,
		0.983187,
		0.991144,
		0.996042,
		0.998635,
		0.999707,
		0.99998,
		1,
		0.99998,
		0.999716,
		0.998697,
		0.996274,
		0.991772,
		0.984569,
		0.974141,
		0.960083,
		0.942118,
		0.92009,
		0.893962,
		0.863799,
		0.82976,
		0.792083,
		0.751077,
		0.000101,
		0.007884,
		0.026987,
		0.055933,
		0.093296,
		0.137703,
		0.187838,
		0.242442,
		0.300316,
		0.360324,
		0.421394,
		0.482527,
		0.542794,
		0.601344,
		0.65741,
		0.710309,
		0.759454,
		0.804357,
		0.844638,
		0.880032,
		0.910396,
		0.93572,
		0.95613,
		0.971895,
		0.983428,
		0.991275,
		0.996102,
		0.998656,
		0.999711,
		0.99998,
		1,
		0.999981,
		0.99972,
		0.998717,
		0.996329,
		0.991891,
		0.984786,
		0.974494,
		0.960612,
		0.942861,
		0.921086,
		0.895243,
		0.865395,
		0.831692,
		0.794366,
		0.0004,
		0.009606,
		0.02986,
		0.05971,
		0.097753,
		0.142639,
		0.193074,
		0.247821,
		0.3057,
		0.365596,
		0.426457,
		0.487302,
		0.547218,
		0.605372,
		0.661012,
		0.713469,
		0.762171,
		0.806641,
		0.84651,
		0.881522,
		0.911542,
		0.936566,
		0.956724,
		0.972286,
		0.983664,
		0.991403,
		0.99616,
		0.998677,
		0.999716,
		0.999981,
		1,
		0.999981,
		0.999725,
		0.998737,
		0.996384,
		0.992008,
		0.984999,
		0.97484,
		0.961131,
		0.943593,
		0.922066,
		0.896506,
		0.866967,
		0.833597,
		0.000888,
		0.011471,
		0.032838,
		0.063557,
		0.102252,
		0.147593,
		0.198309,
		0.253183,
		0.311056,
		0.370832,
		0.431479,
		0.492031,
		0.551596,
		0.609355,
		0.66457,
		0.71659,
		0.764852,
		0.808893,
		0.848355,
		0.882989,
		0.91267,
		0.937398,
		0.957307,
		0.97267,
		0.983896,
		0.991528,
		0.996218,
		0.998697,
		0.99972,
		0.999981,
		1,
		0.999981,
		0.999729,
		0.998756,
		0.996437,
		0.992122,
		0.985208,
		0.975181,
		0.961642,
		0.944313,
		0.923032,
		0.897749,
		0.868517,
		0.001559,
		0.013475,
		0.035914,
		0.067471,
		0.106789,
		0.152563,
		0.203541,
		0.258527,
		0.316383,
		0.376031,
		0.436458,
		0.496716,
		0.555928,
		0.613292,
		0.668086,
		0.71967,
		0.767497,
		0.811114,
		0.850172,
		0.884434,
		0.913781,
		0.938217,
		0.957881,
		0.973047,
		0.984125,
		0.991651,
		0.996274,
		0.998717,
		0.999725,
		0.999981,
		1,
		0.999982,
		0.999733,
		0.998774,
		0.99649,
		0.992235,
		0.985413,
		0.975516,
		0.962145,
		0.945021,
		0.923982,
		0.898975,
		0.002404,
		0.015609,
		0.039085,
		0.071448,
		0.111363,
		0.157547,
		0.208769,
		0.263854,
		0.321682,
		0.381194,
		0.441396,
		0.501356,
		0.560215,
		0.617185,
		0.671559,
		0.722712,
		0.770107,
		0.813304,
		0.851964,
		0.885858,
		0.914874,
		0.939022,
		0.958445,
		0.973418,
		0.984349,
		0.991772,
		0.996329,
		0.998737,
		0.999729,
		0.999982,
		1,
		0.999982,
		0.999737,
		0.998793,
		0.996541,
		0.992346,
		0.985615,
		0.975845,
		0.96264,
		0.945719,
		0.924919,
		0.003417,
		0.01787,
		0.042346,
		0.075484,
		0.11597,
		0.162543,
		0.213992,
		0.269161,
		0.32695,
		0.38632,
		0.446292,
		0.505952,
		0.564456,
		0.621034,
		0.67499,
		0.725714,
		0.772681,
		0.815463,
		0.853729,
		0.88726,
		0.91595,
		0.939815,
		0.959001,
		0.973783,
		0.984569,
		0.991891,
		0.996384,
		0.998756,
		0.999733,
		0.999982,
		1,
		0.999982,
		0.999741,
		0.998811,
		0.996591,
		0.992454,
		0.985814,
		0.976169,
		0.963126,
		0.946406,
		0.004591,
		0.020251,
		0.045692,
		0.079575,
		0.120607,
		0.167548,
		0.219207,
		0.274448,
		0.332189,
		0.391409,
		0.451146,
		0.510503,
		0.568653,
		0.624839,
		0.67838,
		0.728678,
		0.775222,
		0.817592,
		0.855469,
		0.888641,
		0.917009,
		0.940595,
		0.959546,
		0.974141,
		0.984786,
		0.992008,
		0.996437,
		0.998774,
		0.999737,
		0.999982,
		1,
		0.999982,
		0.999745,
		0.998829,
		0.996641,
		0.992561,
		0.986009,
		0.976488,
		0.963605,
		0.00592,
		0.022747,
		0.049119,
		0.083718,
		0.125272,
		0.172561,
		0.224414,
		0.279714,
		0.337397,
		0.39646,
		0.455958,
		0.515011,
		0.572806,
		0.628601,
		0.681729,
		0.731605,
		0.777728,
		0.819691,
		0.857183,
		0.890001,
		0.918052,
		0.941362,
		0.960083,
		0.974494,
		0.984999,
		0.992122,
		0.99649,
		0.998793,
		0.999741,
		0.999982,
		1,
		0.999983,
		0.999749,
		0.998846,
		0.99669,
		0.992666,
		0.986201,
		0.976801,
		9.5e-005,
		0.007397,
		0.025353,
		0.052623,
		0.08791,
		0.129963,
		0.17758,
		0.229612,
		0.284958,
		0.342574,
		0.401474,
		0.460729,
		0.519475,
		0.576915,
		0.63232,
		0.685037,
		0.734494,
		0.780201,
		0.821761,
		0.858873,
		0.891341,
		0.919079,
		0.942118,
		0.960612,
		0.97484,
		0.985208,
		0.992235,
		0.996541,
		0.998811,
		0.999745,
		0.999983,
		1,
		0.999983,
		0.999753,
		0.998863,
		0.996737,
		0.992769,
		0.986389,
		0.000375,
		0.009016,
		0.028064,
		0.0562,
		0.092147,
		0.134676,
		0.182603,
		0.234798,
		0.290179,
		0.34772,
		0.40645,
		0.465459,
		0.523896,
		0.58098,
		0.635997,
		0.688306,
		0.737347,
		0.782641,
		0.823803,
		0.860539,
		0.892661,
		0.92009,
		0.942861,
		0.961131,
		0.975181,
		0.985413,
		0.992346,
		0.996591,
		0.998829,
		0.999749,
		0.999983,
		1,
		0.999983,
		0.999757,
		0.99888,
		0.996784,
		0.99287,
		0.000833,
		0.010771,
		0.030875,
		0.059846,
		0.096427,
		0.139409,
		0.187629,
		0.239972,
		0.295377,
		0.352834,
		0.411389,
		0.470147,
		0.528274,
		0.585003,
		0.639632,
		0.691535,
		0.740163,
		0.785049,
		0.825816,
		0.862181,
		0.893962,
		0.921086,
		0.943593,
		0.961642,
		0.975516,
		0.985615,
		0.992454,
		0.996641,
		0.998846,
		0.999753,
		0.999983,
		1,
		0.999984,
		0.99976,
		0.998897,
		0.99683,
		0.001462,
		0.012657,
		0.033782,
		0.063557,
		0.100747,
		0.144161,
		0.192655,
		0.245133,
		0.300551,
		0.357915,
		0.416289,
		0.474793,
		0.532609,
		0.588982,
		0.643226,
		0.694725,
		0.742944,
		0.787425,
		0.827802,
		0.863799,
		0.895243,
		0.922066,
		0.944313,
		0.962145,
		0.975845,
		0.985814,
		0.992561,
		0.99669,
		0.998863,
		0.999757,
		0.999984,
		1,
		0.999984,
		0.999764,
		0.998913,
		0.002256,
		0.014668,
		0.03678,
		0.06733,
		0.105104,
		0.14893,
		0.197681,
		0.25028,
		0.3057,
		0.362964,
		0.421152,
		0.479399,
		0.536902,
		0.592919,
		0.646778,
		0.697877,
		0.745689,
		0.78977,
		0.82976,
		0.865395,
		0.896506,
		0.923032,
		0.945021,
		0.96264,
		0.976169,
		0.986009,
		0.992666,
		0.996737,
		0.99888,
		0.99976,
		0.999984,
		1,
		0.999984,
		0.999767,
		0.003208,
		0.0168,
		0.039864,
		0.071162,
		0.109496,
		0.153712,
		0.202704,
		0.255412,
		0.310823,
		0.36798,
		0.425977,
		0.483964,
		0.541152,
		0.596815,
		0.650291,
		0.700991,
		0.7484,
		0.792083,
		0.831692,
		0.866967,
		0.897749,
		0.923982,
		0.945719,
		0.963126,
		0.976488,
		0.986201,
		0.992769,
		0.996784,
		0.998897,
		0.999764,
		0.999984,
		1,
		0.999984,
		0.004312,
		0.019046,
		0.043031,
		0.075048,
		0.113919,
		0.158507,
		0.207724,
		0.260527,
		0.315921,
		0.372963,
		0.430764,
		0.488488,
		0.545361,
		0.600669,
		0.653763,
		0.704067,
		0.751077,
		0.794366,
		0.833597,
		0.868517,
		0.898975,
		0.924919,
		0.946406,
		0.963605,
		0.976801,
		0.986389,
		0.99287,
		0.99683,
		0.998913,
		0.999767,
		0.999984,
		1,
		0.005562,
		0.021402,
		0.005605,
		0.046277,
		0.021564,
		0.005649,
		0.078987,
		0.046619,
		0.021727,
		0.005692,
		0.118371,
		0.079556,
		0.046964,
		0.021892,
		0.005737,
		0.163312,
		0.119202,
		0.080132,
		0.047314,
		0.022059,
		0.005782,
		0.212739,
		0.164428,
		0.120042,
		0.080713,
		0.047667,
		0.022228,
		0.005827,
		0.265625,
		0.214149,
		0.165555,
		0.120891,
		0.081301,
		0.048024,
		0.022399,
		0.005873,
		0.320992,
		0.267329,
		0.215573,
		0.166693,
		0.121749,
		0.081895,
		0.048385,
		0.022572,
		0.00592,
		0.377913,
		0.32298,
		0.269049,
		0.217011,
		0.167843,
		0.122616,
		0.082496,
		0.04875,
		0.022747,
		0.005967,
		0.435513,
		0.380165,
		0.324985,
		0.270786,
		0.218463,
		0.169005,
		0.123492,
		0.083104,
		0.049119,
		0.022924,
		0.006014,
		0.492972,
		0.438,
		0.382435,
		0.327009,
		0.272538,
		0.219929,
		0.170178,
		0.124377,
		0.083718,
		0.049493,
		0.023103,
		0.006063,
		0.549528,
		0.495659,
		0.440506,
		0.384724,
		0.32905,
		0.274307,
		0.221409,
		0.171364,
		0.125272,
		0.084339,
		0.049871,
		0.023285,
		0.006111,
		0.604481,
		0.552374,
		0.498366,
		0.443032,
		0.387032,
		0.331109,
		0.276092,
		0.222905,
		0.172561,
		0.126177,
		0.084966,
		0.050253,
		0.023468,
		0.006161,
		0.657196,
		0.607439,
		0.555238,
		0.501091,
		0.445578,
		0.38936,
		0.333186,
		0.277895,
		0.224414,
		0.173771,
		0.127091,
		0.085601,
		0.050639,
		0.023653,
		0.006211,
		0.707107,
		0.660215,
		0.610413,
		0.55812,
		0.503836,
		0.448143,
		0.391707,
		0.335282,
		0.279714,
		0.225939,
		0.174994,
		0.128015,
		0.086243,
		0.05103,
		0.023841,
		0.006261,
		0.753719,
		0.710133,
		0.663248,
		0.613403,
		0.561021,
		0.5066,
		0.450728,
		0.394074,
		0.337397,
		0.28155,
		0.227479,
		0.17623,
		0.128949,
		0.086892,
		0.051425,
		0.024031,
		0.006313,
		0.796619,
		0.756699,
		0.71317,
		0.666294,
		0.61641,
		0.563939,
		0.509384,
		0.453333,
		0.39646,
		0.339531,
		0.283404,
		0.229035,
		0.177478,
		0.129893,
		0.087549,
		0.051826,
		0.024223,
		0.006364,
		0.835475,
		0.799498,
		0.759685,
		0.716218,
		0.669354,
		0.619433,
		0.566876,
		0.512188,
		0.455958,
		0.398867,
		0.341684,
		0.285276,
		0.230606,
		0.178739,
		0.130847,
		0.088212,
		0.05223,
		0.024418,
		0.006417,
		0.870045,
		0.838201,
		0.802379,
		0.762678,
		0.719275,
		0.672427,
		0.622473,
		0.569832,
		0.515011,
		0.458604,
		0.401293,
		0.343856,
		0.287165,
		0.232193,
		0.180014,
		0.131812,
		0.088884,
		0.05264,
		0.024615,
		0.00647,
		0.900183,
		0.872571,
		0.840925,
		0.805262,
		0.765677,
		0.722343,
		0.675515,
		0.625529,
		0.572806,
		0.517854,
		0.46127,
		0.403741,
		0.346048,
		0.289073,
		0.233795,
		0.181302,
		0.132788,
		0.089563,
		0.053054,
		0.024814,
		0.006524,
		0.925841,
		0.902465,
		0.875088,
		0.843645,
		0.808146,
		0.768681,
		0.72542,
		0.678615,
		0.628601,
		0.575799,
		0.520718,
		0.463957,
		0.406208,
		0.34826,
		0.290998,
		0.235414,
		0.182603,
		0.133774,
		0.090249,
		0.053473,
		0.025016,
		0.006578,
		0.947082,
		0.927846,
		0.904734,
		0.877597,
		0.846361,
		0.811031,
		0.771692,
		0.728508,
		0.681729,
		0.631689,
		0.57881,
		0.523601,
		0.466665,
		0.408697,
		0.350491,
		0.292943,
		0.237049,
		0.183919,
		0.134771,
		0.090944,
		0.053897,
		0.02522,
		0.006633,
		0.964076,
		0.948785,
		0.929834,
		0.90699,
		0.880097,
		0.849074,
		0.813918,
		0.774707,
		0.731605,
		0.684856,
		0.634794,
		0.58184,
		0.526505,
		0.469393,
		0.411206,
		0.352743,
		0.294906,
		0.238701,
		0.185248,
		0.135779,
		0.091646,
		0.054327,
		0.025427,
		0.006689,
		0.977109,
		0.965467,
		0.950468,
		0.931803,
		0.909232,
		0.882588,
		0.851782,
		0.816804,
		0.777728,
		0.734711,
		0.687996,
		0.637915,
		0.584888,
		0.529428,
		0.472143,
		0.413737,
		0.355015,
		0.296887,
		0.24037,
		0.186592,
		0.136799,
		0.092357,
		0.054761,
		0.025636,
		0.006746,
		0.986574,
		0.978187,
		0.966832,
		0.952128,
		0.933754,
		0.91146,
		0.88507,
		0.854485,
		0.819691,
		0.780754,
		0.737827,
		0.69115,
		0.641052,
		0.587956,
		0.532373,
		0.474914,
		0.416289,
		0.357308,
		0.298888,
		0.242056,
		0.18795,
		0.137829,
		0.093076,
		0.0552,
		0.025848,
		0.006803,
		0.992969,
		0.987358,
		0.97924,
		0.968173,
		0.953766,
		0.935686,
		0.913673,
		0.887541,
		0.857183,
		0.822577,
		0.783784,
		0.740951,
		0.694316,
		0.644206,
		0.591042,
		0.535338,
		0.477707,
		0.418863,
		0.359622,
		0.300909,
		0.243759,
		0.189322,
		0.138872,
		0.093803,
		0.055645,
		0.026062,
		0.006861,
		0.996876,
		0.99349,
		0.988116,
		0.980266,
		0.969489,
		0.955381,
		0.937598,
		0.91587,
		0.890001,
		0.859876,
		0.825463,
		0.786818,
		0.744084,
		0.697495,
		0.647376,
		0.594146,
		0.538323,
		0.480522,
		0.421459,
		0.361957,
		0.302949,
		0.245479,
		0.19071,
		0.139926,
		0.094539,
		0.056095,
		0.026279,
		0.00692,
		0.998929,
		0.99718,
		0.993987,
		0.988848,
		0.981267,
		0.97078,
		0.956972,
		0.939491,
		0.918052,
		0.89245,
		0.862562,
		0.828347,
		0.789856,
		0.747226,
		0.700686,
		0.650561,
		0.59727,
		0.54133,
		0.483358,
		0.424076,
		0.364313,
		0.305009,
		0.247217,
		0.192112,
		0.140991,
		0.095283,
		0.056551,
		0.026499,
		0.006979,
		0.999771,
		0.999074,
		0.997465,
		0.994461,
		0.989554,
		0.98224,
		0.972044,
		0.95854,
		0.941362,
		0.920217,
		0.894888,
		0.865241,
		0.83123,
		0.792898,
		0.750376,
		0.70389,
		0.653763,
		0.600413,
		0.544357,
		0.486216,
		0.426716,
		0.366691,
		0.307089,
		0.248973,
		0.193529,
		0.142069,
		0.096037,
		0.057012,
		0.026722,
		0.00704,
		0.999985,
		0.999819,
		0.999205,
		0.997731,
		0.994912,
		0.990235,
		0.983187,
		0.973282,
		0.960083,
		0.943213,
		0.922365,
		0.897313,
		0.867913,
		0.834112,
		0.795942,
		0.753533,
		0.707107,
		0.656981,
		0.603574,
		0.547405,
		0.489096,
		0.429378,
		0.369091,
		0.309189,
		0.250747,
		0.194962,
		0.14316,
		0.096799,
		0.057479,
		0.026947,
		0.007101,
		1,
		0.999991,
		0.999859,
		0.999323,
		0.997978,
		0.99534,
		0.99089,
		0.984106,
		0.974494,
		0.961602,
		0.945041,
		0.924495,
		0.899725,
		0.870578,
		0.83699,
		0.79899,
		0.756699,
		0.710335,
		0.660215,
		0.606755,
		0.550474,
		0.491999,
		0.432063,
		0.371512,
		0.31131,
		0.25254,
		0.196411,
		0.144262,
		0.09757,
		0.057952,
		0.027175,
		0.007163,
		0.999985,
		1,
		0.999995,
		0.999892,
		0.999429,
		0.998208,
		0.995745,
		0.991519,
		0.984999,
		0.975678,
		0.963095,
		0.946848,
		0.926607,
		0.902123,
		0.873234,
		0.839866,
		0.80204,
		0.759872,
		0.713576,
		0.663465,
		0.609954,
		0.553565,
		0.494924,
		0.43477,
		0.373956,
		0.313452,
		0.254351,
		0.197875,
		0.145377,
		0.09835,
		0.05843,
		0.027406,
		0.007225,
		0.999771,
		0.999975,
		1,
		0.999997,
		0.99992,
		0.999523,
		0.99842,
		0.996128,
		0.992122,
		0.985864,
		0.976835,
		0.964563,
		0.948631,
		0.9287,
		0.904508,
		0.875881,
		0.842738,
		0.805092,
		0.763052,
		0.716828,
		0.66673,
		0.613173,
		0.556677,
		0.497872,
		0.437501,
		0.376422,
		0.315615,
		0.256181,
		0.199356,
		0.146505,
		0.09914,
		0.058915,
		0.027641,
		0.007289,
		0.998929,
		0.999707,
		0.999962,
		1,
		0.999999,
		0.999942,
		0.999606,
		0.998614,
		0.99649,
		0.9927,
		0.986701,
		0.977965,
		0.966004,
		0.950392,
		0.930774,
		0.906878,
		0.878519,
		0.845607,
		0.808146,
		0.76624,
		0.720092,
		0.670011,
		0.61641,
		0.55981,
		0.500843,
		0.440255,
		0.378911,
		0.317799,
		0.258031,
		0.200852,
		0.147646,
		0.099939,
		0.059405,
		0.027878,
		0.007353,
		0.996876,
		0.998733,
		0.999629,
		0.999944,
		0.999999,
		1,
		0.999959,
		0.999679,
		0.998793,
		0.996829,
		0.993253,
		0.987511,
		0.979066,
		0.967419,
		0.952128,
		0.932827,
		0.909232,
		0.881147,
		0.848471,
		0.811201,
		0.769433,
		0.723368,
		0.673308,
		0.619667,
		0.562964,
		0.503836,
		0.443032,
		0.381424,
		0.320005,
		0.2599,
		0.202366,
		0.148801,
		0.100747,
		0.059902,
		0.028118,
		0.007418,
		0.992969,
		0.996452,
		0.998511,
		0.999537,
		0.99992,
		0.999997,
		1,
		0.999973,
		0.999741,
		0.998956,
		0.997147,
		0.99378,
		0.988294,
		0.98014,
		0.968806,
		0.95384,
		0.93486,
		0.911571,
		0.883765,
		0.851331,
		0.814257,
		0.772634,
		0.726654,
		0.67662,
		0.622942,
		0.56614,
		0.506853,
		0.445833,
		0.383959,
		0.322233,
		0.261788,
		0.203896,
		0.149968,
		0.101566,
		0.060404,
		0.028361,
		0.007484,
		0.986574,
		0.99221,
		0.995985,
		0.99826,
		0.999429,
		0.999889,
		0.999995,
		1,
		0.999982,
		0.999795,
		0.999103,
		0.997444,
		0.994282,
		0.989048,
		0.981184,
		0.970166,
		0.955526,
		0.936872,
		0.913893,
		0.886372,
		0.854185,
		0.817314,
		0.77584,
		0.729952,
		0.679948,
		0.626236,
		0.569338,
		0.509893,
		0.448658,
		0.386518,
		0.324482,
		0.263696,
		0.205443,
		0.151149,
		0.102394,
		0.060913,
		0.028608,
		0.007551,
		0.977109,
		0.985372,
		0.99139,
		0.995473,
		0.997978,
		0.999303,
		0.99985,
		0.99999,
		1,
		0.999989,
		0.99984,
		0.999236,
		0.997721,
		0.994759,
		0.989775,
		0.9822,
		0.971498,
		0.957188,
		0.938862,
		0.916199,
		0.888967,
		0.857034,
		0.82037,
		0.779051,
		0.73326,
		0.683291,
		0.629549,
		0.572558,
		0.512956,
		0.451507,
		0.3891,
		0.326755,
		0.265625,
		0.207007,
		0.152345,
		0.103233,
		0.061429,
		0.028857,
		0.007619,
		0.964076,
		0.975358,
		0.984089,
		0.990506,
		0.994912,
		0.997663,
		0.999156,
		0.999801,
		0.999982,
		1,
		0.999994,
		0.999878,
		0.999355,
		0.997978,
		0.995211,
		0.990474,
		0.983187,
		0.972801,
		0.958822,
		0.94083,
		0.918486,
		0.891549,
		0.859876,
		0.823426,
		0.782268,
		0.736579,
		0.686649,
		0.632882,
		0.575799,
		0.516043,
		0.454381,
		0.391707,
		0.32905,
		0.267574,
		0.208589,
		0.153554,
		0.104082,
		0.061951,
		0.02911,
		0.007688,
		0.947082,
		0.961681,
		0.973507,
		0.982721,
		0.989554,
		0.9943,
		0.997312,
		0.998987,
		0.999741,
		0.999972,
		1,
		0.999997,
		0.999909,
		0.999461,
		0.998216,
		0.995638,
		0.991144,
		0.984144,
		0.974075,
		0.960431,
		0.942774,
		0.920755,
		0.894119,
		0.862711,
		0.826481,
		0.78549,
		0.739908,
		0.690022,
		0.636233,
		0.579062,
		0.519153,
		0.457278,
		0.394338,
		0.331368,
		0.269544,
		0.210189,
		0.154777,
		0.104941,
		0.06248,
		0.029367,
		0.007758,
		0.925841,
		0.943959,
		0.959167,
		0.971552,
		0.981267,
		0.988531,
		0.993633,
		0.996922,
		0.998793,
		0.999668,
		0.999957,
		1,
		0.999999,
		0.999934,
		0.999554,
		0.998435,
		0.996042,
		0.991788,
		0.985072,
		0.975321,
		0.962012,
		0.944695,
		0.923006,
		0.896676,
		0.865538,
		0.829535,
		0.788716,
		0.743248,
		0.69341,
		0.639602,
		0.582346,
		0.522288,
		0.460201,
		0.396993,
		0.333709,
		0.271535,
		0.211807,
		0.156015,
		0.10581,
		0.063015,
		0.029626,
		0.007828,
		0.900183,
		0.921922,
		0.940699,
		0.95653,
		0.969489,
		0.97972,
		0.987433,
		0.992908,
		0.99649,
		0.998572,
		0.999581,
		0.999936,
		0.999999,
		1,
		0.999954,
		0.999636,
		0.998635,
		0.996422,
		0.992403,
		0.98597,
		0.976536,
		0.963565,
		0.946591,
		0.925236,
		0.899218,
		0.868358,
		0.832587,
		0.791947,
		0.746597,
		0.696813,
		0.642991,
		0.585653,
		0.525446,
		0.463149,
		0.399673,
		0.336073,
		0.273547,
		0.213443,
		0.157267,
		0.106691,
		0.063557,
		0.02989,
		0.0079,
		0.870045,
		0.895415,
		0.91785,
		0.937299,
		0.953766,
		0.967315,
		0.978079,
		0.986257,
		0.992122,
		0.996013,
		0.998321,
		0.999477,
		0.999909,
		0.999997,
		1,
		0.999969,
		0.999707,
		0.998818,
		0.996778,
		0.992991,
		0.986838,
		0.977722,
		0.96509,
		0.948463,
		0.927446,
		0.901745,
		0.871169,
		0.835636,
		0.795181,
		0.749955,
		0.70023,
		0.646399,
		0.588982,
		0.528629,
		0.466121,
		0.402378,
		0.338462,
		0.27558,
		0.215097,
		0.158534,
		0.107582,
		0.064107,
		0.030156,
		0.007973,
		0.835475,
		0.864394,
		0.89048,
		0.913621,
		0.933754,
		0.950871,
		0.965026,
		0.976338,
		0.984999,
		0.991272,
		0.995488,
		0.998039,
		0.999355,
		0.999874,
		0.999994,
		1,
		0.99998,
		0.999767,
		0.998984,
		0.997111,
		0.993551,
		0.987677,
		0.978877,
		0.966586,
		0.950308,
		0.929636,
		0.904257,
		0.87397,
		0.838682,
		0.798418,
		0.753323,
		0.703661,
		0.649825,
		0.592333,
		0.531836,
		0.469119,
		0.405109,
		0.340874,
		0.277636,
		0.21677,
		0.159817,
		0.108485,
		0.064663,
		0.030427,
		0.008046,
		0.796619,
		0.828924,
		0.858565,
		0.885375,
		0.909232,
		0.930061,
		0.947842,
		0.962616,
		0.974494,
		0.983655,
		0.990352,
		0.994912,
		0.997721,
		0.999212,
		0.99983,
		0.999988,
		1,
		0.999988,
		0.999819,
		0.999134,
		0.997422,
		0.994084,
		0.988485,
		0.980001,
		0.968053,
		0.952128,
		0.931803,
		0.906753,
		0.876762,
		0.841725,
		0.801659,
		0.756699,
		0.707107,
		0.65327,
		0.595706,
		0.535067,
		0.472143,
		0.407865,
		0.343311,
		0.279714,
		0.218463,
		0.161115,
		0.109399,
		0.065227,
		0.0307,
		0.008121,
		0.753719,
		0.789171,
		0.822187,
		0.852555,
		0.880097,
		0.90468,
		0.926215,
		0.944673,
		0.960083,
		0.972543,
		0.98222,
		0.989361,
		0.994282,
		0.997366,
		0.999046,
		0.999775,
		0.99998,
		1,
		0.999993,
		0.999861,
		0.999268,
		0.997711,
		0.99459,
		0.989263,
		0.981094,
		0.969489,
		0.953921,
		0.933948,
		0.909232,
		0.879542,
		0.844764,
		0.804901,
		0.760084,
		0.710566,
		0.656733,
		0.599101,
		0.538323,
		0.475193,
		0.410647,
		0.345773,
		0.281814,
		0.220175,
		0.162428,
		0.110324,
		0.065798,
		0.030978,
		0.008197,
		0.707107,
		0.745396,
		0.781532,
		0.815262,
		0.846361,
		0.874642,
		0.899959,
		0.922214,
		0.941362,
		0.957423,
		0.97048,
		0.980692,
		0.988294,
		0.993593,
		0.99697,
		0.998854,
		0.999707,
		0.999968,
		1,
		0.999997,
		0.999897,
		0.999388,
		0.997978,
		0.995069,
		0.990011,
		0.982156,
		0.970896,
		0.955686,
		0.93607,
		0.911694,
		0.882312,
		0.847798,
		0.808146,
		0.763477,
		0.71404,
		0.660215,
		0.602518,
		0.541604,
		0.478268,
		0.413455,
		0.34826,
		0.283937,
		0.221906,
		0.163757,
		0.111261,
		0.066377,
		0.031259,
		0.008273,
		0.657196,
		0.69795,
		0.736881,
		0.773702,
		0.808146,
		0.839981,
		0.869007,
		0.895068,
		0.918052,
		0.937904,
		0.95463,
		0.968302,
		0.979066,
		0.987147,
		0.992844,
		0.996529,
		0.998635,
		0.999624,
		0.999951,
		1,
		0.999999,
		0.999925,
		0.999493,
		0.998225,
		0.995521,
		0.990728,
		0.983187,
		0.972271,
		0.957423,
		0.938168,
		0.914137,
		0.88507,
		0.850827,
		0.811392,
		0.766878,
		0.717527,
		0.663715,
		0.605958,
		0.54491,
		0.48137,
		0.416289,
		0.350772,
		0.286083,
		0.223658,
		0.165103,
		0.11221,
		0.066963,
		0.031545,
		0.008351,
		0.604481,
		0.647268,
		0.688606,
		0.728175,
		0.765677,
		0.800837,
		0.833411,
		0.863189,
		0.890001,
		0.913726,
		0.934295,
		0.951701,
		0.966004,
		0.977338,
		0.985916,
		0.992029,
		0.996042,
		0.998385,
		0.999525,
		0.999927,
		0.999999,
		1,
		0.999947,
		0.999586,
		0.998451,
		0.995947,
		0.991416,
		0.984185,
		0.973615,
		0.959131,
		0.940242,
		0.916561,
		0.887815,
		0.85385,
		0.814639,
		0.770286,
		0.721027,
		0.667234,
		0.60942,
		0.54824,
		0.484498,
		0.41915,
		0.353309,
		0.288253,
		0.225429,
		0.166465,
		0.113171,
		0.067557,
		0.031834,
		0.00843,
		0.549528,
		0.593863,
		0.63716,
		0.679074,
		0.719275,
		0.757456,
		0.793332,
		0.826649,
		0.857183,
		0.884756,
		0.909232,
		0.930531,
		0.948631,
		0.963582,
		0.975504,
		0.984597,
		0.991144,
		0.995504,
		0.998102,
		0.999407,
		0.999897,
		0.999997,
		1,
		0.999965,
		0.999666,
		0.998657,
		0.996348,
		0.992073,
		0.985151,
		0.974928,
		0.96081,
		0.94229,
		0.918966,
		0.890546,
		0.856866,
		0.817887,
		0.773702,
		0.72454,
		0.670771,
		0.612904,
		0.551596,
		0.487653,
		0.422038,
		0.355873,
		0.290446,
		0.227222,
		0.167843,
		0.114144,
		0.068159,
		0.032127,
		0.00851,
		0.492972,
		0.538323,
		0.583079,
		0.626874,
		0.669354,
		0.710181,
		0.749038,
		0.78563,
		0.819691,
		0.850988,
		0.879329,
		0.904566,
		0.926607,
		0.945417,
		0.961031,
		0.973558,
		0.983187,
		0.990188,
		0.994912,
		0.997782,
		0.999268,
		0.999857,
		0.999993,
		1,
		0.999977,
		0.999735,
		0.998845,
		0.996723,
		0.9927,
		0.986085,
		0.976208,
		0.962458,
		0.944313,
		0.92135,
		0.893264,
		0.859876,
		0.821134,
		0.777124,
		0.728066,
		0.674326,
		0.61641,
		0.554977,
		0.490835,
		0.424954,
		0.358463,
		0.292664,
		0.229035,
		0.169238,
		0.11513,
		0.068768,
		0.032424,
		0.008592,
		0.435513,
		0.481304,
		0.526971,
		0.572131,
		0.61641,
		0.659446,
		0.700893,
		0.740422,
		0.777728,
		0.812535,
		0.8446,
		0.873716,
		0.899725,
		0.922519,
		0.942053,
		0.958348,
		0.971498,
		0.98168,
		0.989154,
		0.994262,
		0.997422,
		0.999106,
		0.999806,
		0.999987,
		1,
		0.999986,
		0.999793,
		0.999014,
		0.997073,
		0.993298,
		0.986986,
		0.977455,
		0.964076,
		0.946308,
		0.923712,
		0.895967,
		0.862877,
		0.824381,
		0.780552,
		0.731605,
		0.677898,
		0.619939,
		0.558383,
		0.494044,
		0.427896,
		0.361079,
		0.294906,
		0.230869,
		0.170651,
		0.116129,
		0.069387,
		0.032726,
		0.008674,
		0.377913,
		0.423526,
		0.469512,
		0.515474,
		0.561021,
		0.60577,
		0.649351,
		0.691409,
		0.731605,
		0.769624,
		0.805179,
		0.838015,
		0.867913,
		0.894703,
		0.918264,
		0.938536,
		0.955526,
		0.969317,
		0.980072,
		0.988039,
		0.993551,
		0.99702,
		0.998917,
		0.999743,
		0.999977,
		1,
		0.999992,
		0.999842,
		0.999166,
		0.997398,
		0.993865,
		0.987855,
		0.978669,
		0.965663,
		0.948277,
		0.926053,
		0.898654,
		0.865871,
		0.827626,
		0.783986,
		0.735156,
		0.681489,
		0.62349,
		0.561815,
		0.497281,
		0.430867,
		0.363722,
		0.297172,
		0.232725,
		0.172081,
		0.11714,
		0.070013,
		0.033031,
		0.008758,
		0.320992,
		0.365774,
		0.411446,
		0.457602,
		0.503836,
		0.54975,
		0.594955,
		0.639069,
		0.681729,
		0.722587,
		0.761317,
		0.79762,
		0.83123,
		0.861918,
		0.889497,
		0.913836,
		0.93486,
		0.952563,
		0.967013,
		0.978359,
		0.986838,
		0.992774,
		0.996571,
		0.9987,
		0.999666,
		0.999963,
		1,
		0.999996,
		0.999882,
		0.999302,
		0.9977,
		0.994403,
		0.988691,
		0.97985,
		0.967218,
		0.950217,
		0.928371,
		0.901325,
		0.868855,
		0.83087,
		0.787425,
		0.738718,
		0.685097,
		0.627063,
		0.565272,
		0.500544,
		0.433865,
		0.366393,
		0.299464,
		0.234603,
		0.173528,
		0.118164,
		0.070648,
		0.033341,
		0.008842,
		0.265625,
		0.308888,
		0.353578,
		0.39928,
		0.445578,
		0.492061,
		0.538323,
		0.583966,
		0.628601,
		0.671853,
		0.713366,
		0.752804,
		0.789856,
		0.824244,
		0.855726,
		0.884104,
		0.909232,
		0.931022,
		0.949453,
		0.964579,
		0.976536,
		0.985548,
		0.991928,
		0.996073,
		0.998451,
		0.999572,
		0.999944,
		1,
		0.999998,
		0.999914,
		0.999422,
		0.997978,
		0.994912,
		0.989494,
		0.980996,
		0.968741,
		0.952128,
		0.930665,
		0.903979,
		0.871829,
		0.834112,
		0.790869,
		0.742293,
		0.688723,
		0.630658,
		0.568755,
		0.503836,
		0.436892,
		0.369091,
		0.301781,
		0.236502,
		0.174994,
		0.119202,
		0.071291,
		0.033655,
		0.008928,
		0.212739,
		0.253765,
		0.29677,
		0.341334,
		0.387032,
		0.433445,
		0.480153,
		0.526742,
		0.572806,
		0.617947,
		0.661782,
		0.703943,
		0.744084,
		0.781884,
		0.817052,
		0.849334,
		0.878519,
		0.904447,
		0.927017,
		0.946192,
		0.962012,
		0.974599,
		0.984164,
		0.991008,
		0.995521,
		0.998167,
		0.999459,
		0.999917,
		0.999998,
		0.999999,
		0.99994,
		0.999527,
		0.998234,
		0.995392,
		0.990264,
		0.982109,
		0.97023,
		0.954009,
		0.932935,
		0.906615,
		0.874792,
		0.83735,
		0.794318,
		0.745878,
		0.692366,
		0.634275,
		0.572264,
		0.507156,
		0.439948,
		0.371816,
		0.304124,
		0.238425,
		0.176478,
		0.120254,
		0.071944,
		0.033974,
		0.009016,
		0.163312,
		0.201355,
		0.241943,
		0.28465,
		0.32905,
		0.374712,
		0.42121,
		0.468117,
		0.515011,
		0.561477,
		0.607109,
		0.651515,
		0.694316,
		0.735156,
		0.773702,
		0.809651,
		0.842738,
		0.872739,
		0.899478,
		0.92284,
		0.942774,
		0.959306,
		0.972543,
		0.982681,
		0.990011,
		0.994912,
		0.997845,
		0.999324,
		0.999882,
		0.999996,
		1,
		0.999959,
		0.999618,
		0.998468,
		0.995843,
		0.991001,
		0.983187,
		0.971686,
		0.955861,
		0.935179,
		0.909232,
		0.877744,
		0.840584,
		0.797771,
		0.749475,
		0.696026,
		0.637915,
		0.575799,
		0.510503,
		0.443032,
		0.37457,
		0.306492,
		0.24037,
		0.177981,
		0.121319,
		0.072605,
		0.034297,
		0.009104,
		0.118371,
		0.152659,
		0.190067,
		0.230171,
		0.272538,
		0.316735,
		0.362328,
		0.40888,
		0.455958,
		0.503133,
		0.549982,
		0.596089,
		0.641052,
		0.684484,
		0.726017,
		0.765307,
		0.80204,
		0.835936,
		0.866759,
		0.894319,
		0.918486,
		0.939196,
		0.956457,
		0.970362,
		0.981094,
		0.988931,
		0.994242,
		0.997481,
		0.999166,
		0.999836,
		0.999992,
		1,
		0.999974,
		0.999696,
		0.998681,
		0.996266,
		0.991705,
		0.98423,
		0.973107,
		0.957681,
		0.937398,
		0.91183,
		0.880684,
		0.843815,
		0.801227,
		0.753082,
		0.699703,
		0.641577,
		0.579359,
		0.51388,
		0.446146,
		0.377353,
		0.308888,
		0.242338,
		0.179502,
		0.122398,
		0.073275,
		0.034625,
		0.009194,
		0.078987,
		0.108724,
		0.142168,
		0.178891,
		0.218463,
		0.260446,
		0.304401,
		0.349886,
		0.39646,
		0.443682,
		0.491113,
		0.538323,
		0.584888,
		0.630396,
		0.674449,
		0.716668,
		0.756699,
		0.794215,
		0.828924,
		0.860575,
		0.888967,
		0.913952,
		0.935451,
		0.953459,
		0.968053,
		0.9794,
		0.987764,
		0.993506,
		0.997073,
		0.998981,
		0.999779,
		0.999985,
		1,
		0.999984,
		0.999763,
		0.998873,
		0.996662,
		0.992377,
		0.985237,
		0.974494,
		0.959469,
		0.93959,
		0.914407,
		0.883611,
		0.84704,
		0.804685,
		0.756699,
		0.703397,
		0.645261,
		0.582946,
		0.517284,
		0.449289,
		0.380165,
		0.31131,
		0.24433,
		0.181043,
		0.123492,
		0.073955,
		0.034957,
		0.009285,
		0.046277,
		0.070648,
		0.099316,
		0.131859,
		0.167843,
		0.206833,
		0.248386,
		0.292057,
		0.337397,
		0.383959,
		0.431293,
		0.478955,
		0.526505,
		0.573509,
		0.619546,
		0.664208,
		0.707107,
		0.747875,
		0.786174,
		0.821699,
		0.854185,
		0.883417,
		0.909232,
		0.931537,
		0.950308,
		0.965609,
		0.977591,
		0.986506,
		0.9927,
		0.996615,
		0.998766,
		0.999707,
		0.999974,
		1,
		0.999991,
		0.999819,
		0.999046,
		0.99703,
		0.993016,
		0.98621,
		0.975845,
		0.961225,
		0.941754,
		0.916963,
		0.886525,
		0.850259,
		0.808146,
		0.760326,
		0.707107,
		0.648967,
		0.586559,
		0.520718,
		0.452462,
		0.383005,
		0.31376,
		0.246346,
		0.182603,
		0.1246,
		0.074644,
		0.035295,
		0.009378,
		0.021402,
		0.039569,
		0.062632,
		0.090168,
		0.121749,
		0.156939,
		0.195295,
		0.236371,
		0.279714,
		0.32487,
		0.371384,
		0.4188,
		0.466665,
		0.514529,
		0.561953,
		0.608503,
		0.653763,
		0.697333,
		0.738833,
		0.777914,
		0.814257,
		0.847584,
		0.877665,
		0.904322,
		0.927446,
		0.946999,
		0.963026,
		0.975665,
		0.985151,
		0.991821,
		0.996106,
		0.998518,
		0.999618,
		0.999958,
		1,
		0.999996,
		0.999864,
		0.9992,
		0.997372,
		0.993622,
		0.987147,
		0.977161,
		0.962947,
		0.943889,
		0.919497,
		0.889423,
		0.853472,
		0.811609,
		0.763962,
		0.710833,
		0.652694,
		0.590198,
		0.52418,
		0.455665,
		0.385876,
		0.316237,
		0.248386,
		0.184184,
		0.125723,
		0.075343,
		0.035637,
		0.009472,
		0.005562,
		0.016673,
		0.033279,
		0.054963,
		0.081301,
		0.111859,
		0.146196,
		0.183866,
		0.224414,
		0.267385,
		0.312315,
		0.358744,
		0.406208,
		0.454246,
		0.502402,
		0.550223,
		0.59727,
		0.643115,
		0.687346,
		0.729573,
		0.769433,
		0.806596,
		0.840769,
		0.871707,
		0.899218,
		0.923176,
		0.943527,
		0.960299,
		0.973615,
		0.983695,
		0.990862,
		0.995539,
		0.998234,
		0.99951,
		0.999935,
		0.999999,
		0.999998,
		0.999901,
		0.999337,
		0.997688,
		0.994196,
		0.988048,
		0.97844,
		0.964635,
		0.945995,
		0.922008,
		0.892307,
		0.856678,
		0.815072,
		0.767608,
		0.714575,
		0.656444,
		0.593863,
		0.527672,
		0.458899,
		0.388776,
		0.318742,
		0.25045,
		0.185784,
		0.126861,
		0.076051,
		0.035984,
		0.009567,
		0.003183,
		0.012464,
		0.027432,
		0.047667,
		0.072738,
		0.102209,
		0.135635,
		0.172561,
		0.212531,
		0.255081,
		0.299743,
		0.346048,
		0.393526,
		0.441707,
		0.490126,
		0.538323,
		0.585848,
		0.632263,
		0.677145,
		0.720092,
		0.76073,
		0.798713,
		0.833736,
		0.865538,
		0.893914,
		0.91872,
		0.939885,
		0.957423,
		0.971437,
		0.982133,
		0.989821,
		0.994912,
		0.99791,
		0.999381,
		0.999905,
		0.999998,
		0.999999,
		0.999931,
		0.999457,
		0.997978,
		0.994738,
		0.988913,
		0.979683,
		0.966289,
		0.948071,
		0.924495,
		0.895174,
		0.859876,
		0.818536,
		0.771261,
		0.718333,
		0.660215,
		0.597555,
		0.531193,
		0.462163,
		0.391707,
		0.321275,
		0.25254,
		0.187405,
		0.128015,
		0.07677,
		0.036336,
		0.009664,
		0.00144,
		0.008808,
		0.022059,
		0.040769,
		0.064503,
		0.092821,
		0.125272,
		0.161399,
		0.200737,
		0.242817,
		0.287165,
		0.333306,
		0.38076,
		0.429052,
		0.477707,
		0.526257,
		0.57424,
		0.621209,
		0.66673,
		0.71039,
		0.751801,
		0.790604,
		0.826481,
		0.859156,
		0.888407,
		0.914074,
		0.93607,
		0.954391,
		0.969125,
		0.980459,
		0.988691,
		0.99422,
		0.997543,
		0.999227,
		0.999864,
		0.999996,
		1,
		0.999953,
		0.999561,
		0.998244,
		0.995249,
		0.989742,
		0.980889,
		0.967907,
		0.950115,
		0.926957,
		0.898023,
		0.863065,
		0.822,
		0.774923,
		0.722107,
		0.664008,
		0.601273,
		0.534743,
		0.465459,
		0.394668,
		0.323837,
		0.254655,
		0.189047,
		0.129184,
		0.077498,
		0.036694,
		0.009762,
		0.000366,
		0.005737,
		0.01719,
		0.034297,
		0.056622,
		0.083718,
		0.11513,
		0.150396,
		0.189047,
		0.230606,
		0.274594,
		0.320527,
		0.36792,
		0.416289,
		0.465151,
		0.514027,
		0.562448,
		0.609954,
		0.656102,
		0.700466,
		0.742645,
		0.782268,
		0.819002,
		0.852555,
		0.88269,
		0.909232,
		0.932076,
		0.951199,
		0.966673,
		0.978669,
		0.987468,
		0.993458,
		0.997128,
		0.999046,
		0.999812,
		0.999991,
		1,
		0.99997,
		0.999651,
		0.998486,
		0.995728,
		0.990536,
		0.982057,
		0.969489,
		0.952128,
		0.929393,
		0.900855,
		0.866244,
		0.825463,
		0.778592,
		0.725895,
		0.667822,
		0.605018,
		0.538323,
		0.468785,
		0.397661,
		0.326429,
		0.256796,
		0.19071,
		0.130369,
		0.078237,
		0.037057,
		0.009862,
		0.003284,
		0.012854,
		0.02828,
		0.049119,
		0.074922,
		0.105229,
		0.139573,
		0.177478,
		0.218463,
		0.262041,
		0.307723,
		0.355015,
		0.403426,
		0.452462,
		0.501639,
		0.550474,
		0.5985,
		0.645261,
		0.690319,
		0.73326,
		0.773702,
		0.811294,
		0.845732,
		0.876762,
		0.904191,
		0.927897,
		0.947842,
		0.964076,
		0.976757,
		0.986146,
		0.992621,
		0.996662,
		0.998834,
		0.999746,
		0.999982,
		1,
		0.999982,
		0.999727,
		0.998706,
		0.996177,
		0.991293,
		0.983187,
		0.971035,
		0.954108,
		0.931803,
		0.903668,
		0.869413,
		0.828924,
		0.782268,
		0.729698,
		0.671658,
		0.608789,
		0.541933,
		0.472143,
		0.400685,
		0.32905,
		0.258963,
		0.192394,
		0.13157,
		0.078987,
		0.037425,
		0.009964,
		0.001486,
		0.009086,
		0.022747,
		0.042023,
		0.06646,
		0.095593,
		0.128949,
		0.166048,
		0.206404,
		0.249522,
		0.294906,
		0.342055,
		0.390469,
		0.439648,
		0.489096,
		0.538323,
		0.586849,
		0.634207,
		0.679948,
		0.723646,
		0.764902,
		0.803355,
		0.838682,
		0.870615,
		0.898944,
		0.923529,
		0.944313,
		0.961329,
		0.974716,
		0.984721,
		0.991705,
		0.99614,
		0.998588,
		0.999663,
		0.99997,
		1,
		0.99999,
		0.999791,
		0.998903,
		0.996595,
		0.992015,
		0.984278,
		0.972543,
		0.956054,
		0.934185,
		0.90646,
		0.872571,
		0.832383,
		0.785951,
		0.733515,
		0.675515,
		0.612586,
		0.545574,
		0.475533,
		0.403741,
		0.331701,
		0.261156,
		0.194101,
		0.132788,
		0.079747,
		0.037798,
		0.010067,
		0.000378,
		0.00592,
		0.017731,
		0.035363,
		0.058356,
		0.086243,
		0.118545,
		0.154777,
		0.194445,
		0.237049,
		0.282087,
		0.32905,
		0.377429,
		0.426716,
		0.476406,
		0.525998,
		0.575003,
		0.622942,
		0.669354,
		0.7138,
		0.755868,
		0.795181,
		0.831403,
		0.864247,
		0.893487,
		0.918966,
		0.940607,
		0.958426,
		0.972543,
		0.983187,
		0.990705,
		0.995558,
		0.998304,
		0.999561,
		0.999951,
		1,
		0.999995,
		0.999844,
		0.999079,
		0.996984,
		0.9927,
		0.985332,
		0.974012,
		0.957965,
		0.936538,
		0.909232,
		0.875716,
		0.835839,
		0.789639,
		0.737347,
		0.679392,
		0.61641,
		0.549244,
		0.478955,
		0.406828,
		0.334382,
		0.263377,
		0.19583,
		0.134022,
		0.080519,
		0.038178,
		0.010171,
		0.00339,
		0.013263,
		0.029167,
		0.050639,
		0.077206,
		0.108384,
		0.143683,
		0.182603,
		0.224639,
		0.26928,
		0.31601,
		0.364313,
		0.413672,
		0.463572,
		0.513503,
		0.562964,
		0.611467,
		0.658536,
		0.703721,
		0.746597,
		0.78677,
		0.823889,
		0.857652,
		0.887815,
		0.914202,
		0.936718,
		0.955361,
		0.97023,
		0.981538,
		0.989616,
		0.994912,
		0.997978,
		0.999437,
		0.999926,
		0.999999,
		0.999998,
		0.999886,
		0.999236,
		0.997343,
		0.99335,
		0.986346,
		0.975444,
		0.959841,
		0.938862,
		0.911982,
		0.878848,
		0.839291,
		0.793332,
		0.741192,
		0.683291,
		0.620261,
		0.552945,
		0.48241,
		0.409949,
		0.337094,
		0.265625,
		0.197581,
		0.135274,
		0.081301,
		0.038563,
		0.010277,
		0.001534,
		0.009378,
		0.023468,
		0.043337,
		0.068506,
		0.098489,
		0.132788,
		0.170898,
		0.212308,
		0.256499,
		0.302949,
		0.351133,
		0.400525,
		0.450601,
		0.500843,
		0.550737,
		0.599783,
		0.647496,
		0.69341,
		0.737086,
		0.778118,
		0.816138,
		0.850827,
		0.881923,
		0.909232,
		0.932642,
		0.952128,
		0.967772,
		0.979768,
		0.988431,
		0.994196,
		0.997607,
		0.999289,
		0.99989,
		0.999998,
		0.999999,
		0.99992,
		0.999373,
		0.997675,
		0.993965,
		0.987321,
		0.976835,
		0.961681,
		0.941155,
		0.914709,
		0.881966,
		0.842738,
		0.797031,
		0.74505,
		0.68721,
		0.624138,
		0.556677,
		0.485897,
		0.413102,
		0.339837,
		0.267901,
		0.199356,
		0.136543,
		0.082095,
		0.038954,
		0.010385,
		0.00039,
		0.006111,
		0.018298,
		0.036479,
		0.060171,
		0.088884,
		0.122113,
		0.159349,
		0.200072,
		0.243759,
		0.289878,
		0.337898,
		0.387283,
		0.437501,
		0.488022,
		0.538323,
		0.587893,
		0.636233,
		0.682865,
		0.727336,
		0.769224,
		0.808146,
		0.843766,
		0.875806,
		0.904051,
		0.928371,
		0.948721,
		0.965164,
		0.977873,
		0.987147,
		0.993406,
		0.997186,
		0.999112,
		0.999844,
		0.999995,
		1,
		0.999946,
		0.999493,
		0.997978,
		0.994545,
		0.988257,
		0.978187,
		0.963484,
		0.943417,
		0.917412,
		0.88507,
		0.84618,
		0.800733,
		0.748921,
		0.69115,
		0.628041,
		0.560439,
		0.489418,
		0.416289,
		0.342613,
		0.270205,
		0.201154,
		0.137829,
		0.0829,
		0.03935,
		0.010495,
		0.003501,
		0.013691,
		0.030097,
		0.05223,
		0.079594,
		0.111681,
		0.147975,
		0.18795,
		0.231074,
		0.276811,
		0.324619,
		0.373956,
		0.424278,
		0.475047,
		0.525728,
		0.575799,
		0.624748,
		0.672085,
		0.717343,
		0.760084,
		0.79991,
		0.836467,
		0.869459,
		0.898654,
		0.9239,
		0.945135,
		0.962398,
		0.975845,
		0.985756,
		0.992536,
		0.996711,
		0.998903,
		0.999783,
		0.999989,
		1,
		0.999965,
		0.999596,
		0.998255,
		0.99509,
		0.989154,
		0.979499,
		0.965249,
		0.945646,
		0.92009,
		0.888157,
		0.849616,
		0.804438,
		0.752804,
		0.695109,
		0.631971,
		0.564232,
		0.492972,
		0.41951,
		0.34542,
		0.272538,
		0.202976,
		0.139134,
		0.083718,
		0.039753,
		0.010606,
		0.001584,
		0.009684,
		0.024223,
		0.044712,
		0.070648,
		0.101517,
		0.136799,
		0.17596,
		0.218463,
		0.263763,
		0.31131,
		0.360554,
		0.410941,
		0.461923,
		0.512956,
		0.563503,
		0.613044,
		0.661072,
		0.707107,
		0.750696,
		0.791425,
		0.828924,
		0.862877,
		0.893035,
		0.919224,
		0.941362,
		0.959469,
		0.973679,
		0.984253,
		0.991581,
		0.996177,
		0.998659,
		0.999707,
		0.99998,
		1,
		0.999979,
		0.999684,
		0.998506,
		0.995601,
		0.990011,
		0.98077,
		0.966975,
		0.947842,
		0.922742,
		0.891227,
		0.853044,
		0.808146,
		0.756699,
		0.699089,
		0.635927,
		0.568056,
		0.496559,
		0.422765,
		0.34826,
		0.2749,
		0.204822,
		0.140457,
		0.084547,
		0.040162,
		0.010719,
		0.000403,
		0.006313,
		0.018893,
		0.037648,
		0.062072,
		0.091646,
		0.125842,
		0.164123,
		0.205942,
		0.250747,
		0.297982,
		0.347086,
		0.397498,
		0.448658,
		0.500011,
		0.55101,
		0.601121,
		0.649825,
		0.696627,
		0.741059,
		0.78269,
		0.821134,
		0.856057,
		0.887188,
		0.914337,
		0.937398,
		0.956371,
		0.971369,
		0.982633,
		0.990536,
		0.995579,
		0.998376,
		0.999611,
		0.999965,
		1,
		0.999988,
		0.999758,
		0.998733,
		0.996078,
		0.990828,
		0.981999,
		0.968661,
		0.950003,
		0.925367,
		0.894279,
		0.856465,
		0.811856,
		0.760605,
		0.703088,
		0.63991,
		0.571912,
		0.50018,
		0.426054,
		0.351133,
		0.277292,
		0.206693,
		0.141799,
		0.085389,
		0.040578,
		0.010834,
		0.003617,
		0.014141,
		0.031071,
		0.053897,
		0.082095,
		0.11513,
		0.152459,
		0.193529,
		0.237781,
		0.28465,
		0.333566,
		0.383959,
		0.435259,
		0.4869,
		0.538323,
		0.588982,
		0.638345,
		0.685901,
		0.73117,
		0.773702,
		0.813093,
		0.848992,
		0.881109,
		0.909232,
		0.933236,
		0.953097,
		0.968908,
		0.980889,
		0.989394,
		0.994912,
		0.998049,
		0.999493,
		0.999944,
		1,
		0.999994,
		0.999819,
		0.998935,
		0.996521,
		0.991607,
		0.983187,
		0.970307,
		0.952128,
		0.927963,
		0.897313,
		0.859876,
		0.815567,
		0.764523,
		0.707107,
		0.643919,
		0.575799,
		0.503836,
		0.429378,
		0.354039,
		0.279714,
		0.208589,
		0.14316,
		0.086243,
		0.041,
		0.010951,
		0.001638,
		0.010005,
		0.025016,
		0.046154,
		0.072891,
		0.104687,
		0.140991,
		0.181245,
		0.224882,
		0.271328,
		0.320005,
		0.370333,
		0.421733,
		0.473628,
		0.525446,
		0.57663,
		0.626633,
		0.674931,
		0.721027,
		0.764456,
		0.804797,
		0.841678,
		0.874792,
		0.903905,
		0.928869,
		0.94964,
		0.966289,
		0.979014,
		0.98815,
		0.994171,
		0.997675,
		0.99935,
		0.999914,
		0.999999,
		0.999998,
		0.999868,
		0.999115,
		0.996932,
		0.992346,
		0.984332,
		0.971912,
		0.954217,
		0.930531,
		0.900326,
		0.863277,
		0.819279,
		0.76845,
		0.711144,
		0.647954,
		0.579717,
		0.507526,
		0.432738,
		0.35698,
		0.282166,
		0.210511,
		0.14454,
		0.08711,
		0.041428,
		0.011069,
		0.000417,
		0.006524,
		0.019517,
		0.038875,
		0.064064,
		0.094539,
		0.129743,
		0.169111,
		0.212067,
		0.258031,
		0.306416,
		0.356632,
		0.40809,
		0.460201,
		0.512384,
		0.564067,
		0.61469,
		0.663715,
		0.710629,
		0.754952,
		0.796242,
		0.834112,
		0.868232,
		0.898348,
		0.924291,
		0.945995,
		0.963506,
		0.977002,
		0.986798,
		0.99335,
		0.997247,
		0.999178,
		0.999873,
		0.999998,
		0.999999,
		0.999907,
		0.999274,
		0.997312,
		0.993045,
		0.985435,
		0.973475,
		0.956268,
		0.933068,
		0.903317,
		0.866667,
		0.82299,
		0.772387,
		0.715201,
		0.652015,
		0.583667,
		0.511251,
		0.436133,
		0.359954,
		0.28465,
		0.212459,
		0.14594,
		0.08799,
		0.041863,
		0.01119,
		0.003739,
		0.014613,
		0.032094,
		0.055645,
		0.084714,
		0.118739,
		0.157147,
		0.199356,
		0.244776,
		0.292812,
		0.342867,
		0.394338,
		0.446628,
		0.499142,
		0.551297,
		0.602518,
		0.652255,
		0.699976,
		0.745185,
		0.787425,
		0.826287,
		0.861423,
		0.892557,
		0.919497,
		0.942154,
		0.960553,
		0.974848,
		0.985332,
		0.992444,
		0.996763,
		0.998974,
		0.999819,
		0.999994,
		1,
		0.999937,
		0.999412,
		0.99766,
		0.993706,
		0.986495,
		0.974996,
		0.95828,
		0.935573,
		0.906287,
		0.870045,
		0.826699,
		0.776333,
		0.719275,
		0.656102,
		0.587648,
		0.515011,
		0.439564,
		0.362964,
		0.287165,
		0.214433,
		0.14736,
		0.088884,
		0.042305,
		0.011312,
		0.001693,
		0.010342,
		0.025848,
		0.047667,
		0.075242,
		0.108006,
		0.145377,
		0.186768,
		0.23158,
		0.27921,
		0.32905,
		0.380488,
		0.432915,
		0.485726,
		0.538323,
		0.59012,
		0.640549,
		0.689066,
		0.735156,
		0.778343,
		0.818201,
		0.854361,
		0.886525,
		0.914479,
		0.938112,
		0.957423,
		0.972543,
		0.983744,
		0.991447,
		0.996216,
		0.998733,
		0.999749,
		0.999987,
		1,
		0.999959,
		0.999531,
		0.997978,
		0.994328,
		0.987511,
		0.976473,
		0.960253,
		0.938045,
		0.909232,
		0.873411,
		0.830407,
		0.780288,
		0.723368,
		0.660215,
		0.591661,
		0.518807,
		0.443032,
		0.366009,
		0.289713,
		0.216434,
		0.148801,
		0.089791,
		0.042755,
		0.011437,
		0.000431,
		0.006746,
		0.020173,
		0.040162,
		0.066153,
		0.09757,
		0.133826,
		0.174326,
		0.218463,
		0.265625,
		0.315194,
		0.366551,
		0.419073,
		0.472143,
		0.525152,
		0.577498,
		0.628601,
		0.677898,
		0.72486,
		0.768993,
		0.809849,
		0.84704,
		0.880246,
		0.909232,
		0.93386,
		0.954108,
		0.970081,
		0.982029,
		0.990352,
		0.995601,
		0.998451,
		0.99966,
		0.999976,
		1,
		0.999975,
		0.999632,
		0.998267,
		0.994912,
		0.988485,
		0.977906,
		0.962186,
		0.940484,
		0.912153,
		0.876762,
		0.834112,
		0.78425,
		0.727478,
		0.664354,
		0.595706,
		0.522638,
		0.446537,
		0.369091,
		0.292293,
		0.218463,
		0.150262,
		0.090711,
		0.043211,
		0.011564,
		0.003868,
		0.015108,
		0.033168,
		0.057479,
		0.087461,
		0.122519,
		0.162051,
		0.205443,
		0.252073,
		0.301315,
		0.352538,
		0.405109,
		0.4584,
		0.511786,
		0.564656,
		0.61641,
		0.666474,
		0.714298,
		0.759371,
		0.801227,
		0.839455,
		0.873716,
		0.903749,
		0.929393,
		0.950601,
		0.967454,
		0.980179,
		0.989154,
		0.994912,
		0.998123,
		0.999548,
		0.999959,
		1,
		0.999986,
		0.999718,
		0.998528,
		0.995458,
		0.989415,
		0.979295,
		0.964076,
		0.942888,
		0.915048,
		0.880097,
		0.837812,
		0.78822,
		0.731605,
		0.668518,
		0.599783,
		0.526505,
		0.45008,
		0.372208,
		0.294906,
		0.220519,
		0.151745,
		0.091646,
		0.043674,
		0.011692,
		0.001752,
		0.010696,
		0.026722,
		0.049255,
		0.077708,
		0.111483,
		0.149968,
		0.192542,
		0.238574,
		0.287427,
		0.338462,
		0.391034,
		0.444504,
		0.498233,
		0.551596,
		0.603979,
		0.654791,
		0.703467,
		0.749475,
		0.79233,
		0.831602,
		0.866927,
		0.898023,
		0.924704,
		0.946896,
		0.964657,
		0.978187,
		0.987845,
		0.994144,
		0.997745,
		0.999412,
		0.999934,
		1,
		0.999993,
		0.999788,
		0.998762,
		0.995967,
		0.990301,
		0.980638,
		0.965925,
		0.945255,
		0.917916,
		0.883417,
		0.841508,
		0.792195,
		0.735749,
		0.672708,
		0.603892,
		0.530408,
		0.45366,
		0.375363,
		0.297552,
		0.222604,
		0.15325,
		0.092596,
		0.044146,
		0.011823,
		0.000447,
		0.006979,
		0.020862,
		0.041515,
		0.068345,
		0.100747,
		0.138102,
		0.179781,
		0.225145,
		0.273547,
		0.324337,
		0.37686,
		0.430464,
		0.484498,
		0.538323,
		0.591311,
		0.642853,
		0.692366,
		0.739302,
		0.783156,
		0.823476,
		0.859876,
		0.892049,
		0.919785,
		0.942986,
		0.961681,
		0.976047,
		0.986419,
		0.993291,
		0.997312,
		0.999246,
		0.999899,
		0.999999,
		0.999997,
		0.999846,
		0.99897,
		0.996439,
		0.991144,
		0.981935,
		0.967729,
		0.947585,
		0.920755,
		0.886718,
		0.845197,
		0.796177,
		0.739908,
		0.676922,
		0.608032,
		0.534347,
		0.457278,
		0.378555,
		0.300233,
		0.224718,
		0.154777,
		0.09356,
		0.044624,
		0.011957,
		0.004003,
		0.01563,
		0.034297,
		0.059405,
		0.090341,
		0.12648,
		0.167184,
		0.211807,
		0.259691,
		0.310176,
		0.362598,
		0.416289,
		0.470589,
		0.524843,
		0.578407,
		0.630658,
		0.680995,
		0.728851,
		0.773702,
		0.815072,
		0.852555,
		0.88582,
		0.91463,
		0.938862,
		0.958519,
		0.973751,
		0.984868,
		0.992346,
		0.996818,
		0.999046,
		0.999852,
		0.999997,
		0.999999,
		0.999891,
		0.999153,
		0.996876,
		0.991944,
		0.983187,
		0.969489,
		0.949876,
		0.923565,
		0.890001,
		0.84888,
		0.800163,
		0.744084,
		0.681162,
		0.612205,
		0.538323,
		0.460936,
		0.381784,
		0.302949,
		0.226862,
		0.156326,
		0.094539,
		0.045111,
		0.012092,
		0.001814,
		0.011069,
		0.027641,
		0.050923,
		0.080297,
		0.11513,
		0.154777,
		0.198581,
		0.245878,
		0.295997,
		0.34826,
		0.40199,
		0.456513,
		0.51116,
		0.565272,
		0.618209,
		0.669354,
		0.71812,
		0.763962,
		0.806386,
		0.84496,
		0.879329,
		0.909232,
		0.934518,
		0.955164,
		0.971292,
		0.983187,
		0.991302,
		0.996257,
		0.998808,
		0.999788,
		0.999993,
		1,
		0.999926,
		0.999313,
		0.997277,
		0.9927,
		0.984391,
		0.971204,
		0.952128,
		0.926344,
		0.893264,
		0.852555,
		0.804153,
		0.748275,
		0.685426,
		0.61641,
		0.542336,
		0.464632,
		0.385053,
		0.3057,
		0.229035,
		0.157899,
		0.095534,
		0.045605,
		0.01223,
		0.000462,
		0.007225,
		0.021587,
		0.042936,
		0.070648,
		0.104082,
		0.142584,
		0.185491,
		0.232129,
		0.281814,
		0.333861,
		0.387578,
		0.442279,
		0.497281,
		0.551909,
		0.605508,
		0.657442,
		0.707107,
		0.753936,
		0.797413,
		0.837084,
		0.872571,
		0.903584,
		0.929947,
		0.951607,
		0.968661,
		0.981366,
		0.990153,
		0.995624,
		0.998528,
		0.999707,
		0.999985,
		1,
		0.999952,
		0.999452,
		0.997644,
		0.993413,
		0.985548,
		0.972873,
		0.954338,
		0.92909,
		0.896506,
		0.85622,
		0.808146,
		0.75248,
		0.689715,
		0.620647,
		0.546387,
		0.468368,
		0.38836,
		0.308487,
		0.231239,
		0.159495,
		0.096544,
		0.046108,
		0.01237,
		0.004145,
		0.016179,
		0.035484,
		0.061429,
		0.093366,
		0.130634,
		0.172561,
		0.218463,
		0.267647,
		0.319414,
		0.373065,
		0.427896,
		0.483212,
		0.538323,
		0.592557,
		0.645261,
		0.69581,
		0.743619,
		0.78815,
		0.828924,
		0.865538,
		0.897679,
		0.92514,
		0.947842,
		0.965851,
		0.9794,
		0.988892,
		0.994912,
		0.9982,
		0.999603,
		0.999972,
		1,
		0.999971,
		0.99957,
		0.997978,
		0.994084,
		0.986658,
		0.974494,
		0.956507,
		0.931803,
		0.899725,
		0.859876,
		0.812141,
		0.756699,
		0.694027,
		0.624916,
		0.550474,
		0.472143,
		0.391707,
		0.31131,
		0.233473,
		0.161115,
		0.09757,
		0.046619,
		0.012512,
		0.001879,
		0.011462,
		0.028608,
		0.052677,
		0.083016,
		0.118957,
		0.159817,
		0.204903,
		0.253513,
		0.304937,
		0.358463,
		0.413374,
		0.468961,
		0.524519,
		0.579359,
		0.63281,
		0.68423,
		0.73301,
		0.778592,
		0.820474,
		0.858226,
		0.89151,
		0.92009,
		0.943859,
		0.962855,
		0.977279,
		0.987511,
		0.994115,
		0.997819,
		0.999473,
		0.999952,
		1,
		0.999984,
		0.999669,
		0.99828,
		0.994712,
		0.98772,
		0.976067,
		0.958631,
		0.934481,
		0.90292,
		0.863519,
		0.816138,
		0.760931,
		0.698364,
		0.629217,
		0.5546,
		0.475959,
		0.395094,
		0.31417,
		0.23574,
		0.162759,
		0.098612,
		0.047138,
		0.012657,
		0.000479,
		0.007484,
		0.02235,
		0.044432,
		0.073068,
		0.107582,
		0.147285,
		0.191473,
		0.239433,
		0.290446,
		0.343786,
		0.398725,
		0.454537,
		0.510503,
		0.565918,
		0.620093,
		0.672365,
		0.722107,
		0.768736,
		0.811728,
		0.850628,
		0.88507,
		0.91479,
		0.939652,
		0.959663,
		0.974996,
		0.986004,
		0.993226,
		0.99738,
		0.999313,
		0.999923,
		1,
		0.999992,
		0.999752,
		0.998551,
		0.995298,
		0.988734,
		0.977591,
		0.960712,
		0.937122,
		0.906089,
		0.867151,
		0.820135,
		0.765177,
		0.702724,
		0.63355,
		0.558763,
		0.479816,
		0.398522,
		0.317068,
		0.238038,
		0.164428,
		0.099671,
		0.047667,
		0.012805,
		0.004296,
		0.016757,
		0.036734,
		0.063557,
		0.096544,
		0.134994,
		0.178197,
		0.225429,
		0.27596,
		0.32905,
		0.383959,
		0.439948,
		0.496282,
		0.552238,
		0.607109,
		0.660215,
		0.710906,
		0.758579,
		0.802682,
		0.842738,
		0.878352,
		0.909232,
		0.935212,
		0.956268,
		0.972543,
		0.984361,
		0.992239,
		0.996876,
		0.999119,
		0.999882,
		0.999999,
		0.999997,
		0.999819,
		0.998793,
		0.995843,
		0.9897,
		0.979066,
		0.962746,
		0.939726,
		0.909232,
		0.870768,
		0.824131,
		0.769433,
		0.707107,
		0.637915,
		0.562964,
		0.483714,
		0.40199,
		0.320005,
		0.24037,
		0.166123,
		0.100747,
		0.048204,
		0.012955,
		0.001948,
		0.011876,
		0.029626,
		0.054523,
		0.085875,
		0.122976,
		0.165103,
		0.211524,
		0.261496,
		0.31427,
		0.369091,
		0.425205,
		0.481862,
		0.538323,
		0.593863,
		0.647781,
		0.699408,
		0.748116,
		0.793332,
		0.834551,
		0.87135,
		0.903409,
		0.930531,
		0.952661,
		0.969911,
		0.982575,
		0.991144,
		0.996301,
		0.998886,
		0.999826,
		0.999997,
		0.999999,
		0.999872,
		0.999006,
		0.996348,
		0.990617,
		0.980491,
		0.964734,
		0.94229,
		0.912347,
		0.874369,
		0.828126,
		0.773702,
		0.711512,
		0.642312,
		0.567204,
		0.487653,
		0.405501,
		0.32298,
		0.242735,
		0.167843,
		0.101841,
		0.04875,
		0.013107,
		0.000497,
		0.007758,
		0.023155,
		0.046007,
		0.075614,
		0.111261,
		0.152218,
		0.197741,
		0.247077,
		0.299464,
		0.354133,
		0.410318,
		0.467253,
		0.52418,
		0.580358,
		0.635065,
		0.687611,
		0.737347,
		0.783673,
		0.82606,
		0.864056,
		0.897313,
		0.925601,
		0.948834,
		0.967092,
		0.980638,
		0.989936,
		0.995649,
		0.998608,
		0.999752,
		0.999992,
		1,
		0.999913,
		0.999193,
		0.996812,
		0.991486,
		0.981865,
		0.966673,
		0.944814,
		0.915432,
		0.877955,
		0.832117,
		0.77798,
		0.71594,
		0.64674,
		0.571482,
		0.491635,
		0.409054,
		0.325995,
		0.245133,
		0.16959,
		0.102952,
		0.049305,
		0.013263,
		0.004454,
		0.017367,
		0.038051,
		0.065798,
		0.099885,
		0.139573,
		0.184108,
		0.232725,
		0.28465,
		0.339103,
		0.395301,
		0.452462,
		0.509814,
		0.566596,
		0.622067,
		0.675515,
		0.726267,
		0.773702,
		0.81726,
		0.856465,
		0.890936,
		0.920413,
		0.944778,
		0.964076,
		0.978541,
		0.988607,
		0.994912,
		0.99828,
		0.999656,
		0.999982,
		1,
		0.999944,
		0.999355,
		0.997238,
		0.992308,
		0.983187,
		0.968563,
		0.947296,
		0.918486,
		0.881522,
		0.836105,
		0.782268,
		0.72039,
		0.651201,
		0.575799,
		0.495659,
		0.41265,
		0.32905,
		0.247567,
		0.171364,
		0.104082,
		0.049871,
		0.013421,
		0.002021,
		0.012313,
		0.0307,
		0.056468,
		0.088884,
		0.127199,
		0.170651,
		0.218463,
		0.269849,
		0.324015,
		0.380165,
		0.437501,
		0.495233,
		0.552583,
		0.608789,
		0.663118,
		0.714875,
		0.763413,
		0.808146,
		0.848568,
		0.88427,
		0.91496,
		0.940484,
		0.960857,
		0.976276,
		0.987147,
		0.994084,
		0.997897,
		0.999533,
		0.999967,
		1,
		0.999966,
		0.999493,
		0.997627,
		0.993081,
		0.984457,
		0.970403,
		0.949734,
		0.921508,
		0.88507,
		0.840087,
		0.786565,
		0.72486,
		0.655692,
		0.580154,
		0.499726,
		0.416289,
		0.332145,
		0.250035,
		0.173165,
		0.105229,
		0.050445,
		0.013582,
		0.000516,
		0.008046,
		0.024004,
		0.047667,
		0.078295,
		0.11513,
		0.1574,
		0.204316,
		0.255081,
		0.308888,
		0.364925,
		0.42238,
		0.480445,
		0.538323,
		0.595234,
		0.650423,
		0.70317,
		0.752804,
		0.798713,
		0.840361,
		0.877309,
		0.909232,
		0.935944,
		0.957423,
		0.973833,
		0.985548,
		0.993157,
		0.997451,
		0.999381,
		0.999944,
		1,
		0.999981,
		0.99961,
		0.997978,
		0.993807,
		0.985674,
		0.972191,
		0.952128,
		0.924495,
		0.888597,
		0.844063,
		0.790869,
		0.729352,
		0.660215,
		0.584549,
		0.503836,
		0.419973,
		0.335282,
		0.25254,
		0.174994,
		0.106396,
		0.05103,
		0.013746,
		0.004622,
		0.018011,
		0.039439,
		0.068159,
		0.103402,
		0.144385,
		0.190312,
		0.24037,
		0.29374,
		0.349596,
		0.407111,
		0.465459,
		0.523824,
		0.581406,
		0.637429,
		0.69115,
		0.741872,
		0.788955,
		0.831837,
		0.870045,
		0.903223,
		0.931149,
		0.953766,
		0.971204,
		0.983803,
		0.992122,
		0.996937,
		0.999193,
		0.999909,
		1,
		0.99999,
		0.999707,
		0.998294,
		0.994485,
		0.986838,
		0.973927,
		0.954475,
		0.927446,
		0.892101,
		0.848031,
		0.795181,
		0.733863,
		0.664769,
		0.588982,
		0.50799,
		0.423701,
		0.338462,
		0.255081,
		0.176852,
		0.107582,
		0.051625,
		0.013913,
		0.002098,
		0.012775,
		0.031834,
		0.058518,
		0.092052,
		0.131641,
		0.176478,
		0.22574,
		0.278592,
		0.334196,
		0.391707,
		0.450283,
		0.50909,
		0.567308,
		0.624138,
		0.678813,
		0.730613,
		0.778869,
		0.82299,
		0.862471,
		0.896922,
		0.926089,
		0.949876,
		0.968379,
		0.981901,
		0.990973,
		0.996348,
		0.998965,
		0.999861,
		0.999999,
		0.999996,
		0.999785,
		0.998577,
		0.995118,
		0.987949,
		0.975609,
		0.956775,
		0.930361,
		0.895582,
		0.85199,
		0.799498,
		0.738394,
		0.669354,
		0.593455,
		0.512188,
		0.427474,
		0.341684,
		0.257659,
		0.178739,
		0.108788,
		0.05223,
		0.014083,
		0.000536,
		0.008351,
		0.0249,
		0.049418,
		0.081119,
		0.119202,
		0.162846,
		0.211216,
		0.263466,
		0.318742,
		0.376183,
		0.43493,
		0.494131,
		0.552945,
		0.610552,
		0.666161,
		0.719025,
		0.76845,
		0.813813,
		0.854579,
		0.890323,
		0.920755,
		0.945745,
		0.965348,
		0.979834,
		0.9897,
		0.995675,
		0.99869,
		0.999794,
		0.999996,
		0.999999,
		0.999848,
		0.998827,
		0.995704,
		0.989007,
		0.977237,
		0.959025,
		0.933236,
		0.899037,
		0.855939,
		0.80382,
		0.742944,
		0.673969,
		0.597967,
		0.51643,
		0.431293,
		0.34495,
		0.260276,
		0.180656,
		0.110014,
		0.052846,
		0.014256,
		0.004799,
		0.018691,
		0.040905,
		0.070648,
		0.107106,
		0.149448,
		0.196828,
		0.248386,
		0.303253,
		0.360554,
		0.419412,
		0.478955,
		0.538323,
		0.596674,
		0.653193,
		0.707107,
		0.757694,
		0.804301,
		0.846361,
		0.883417,
		0.91514,
		0.941362,
		0.962103,
		0.977591,
		0.988294,
		0.994912,
		0.998364,
		0.999707,
		0.99999,
		1,
		0.999897,
		0.999046,
		0.996245,
		0.990011,
		0.97881,
		0.961225,
		0.93607,
		0.902465,
		0.859876,
		0.808146,
		0.747512,
		0.678615,
		0.602518,
		0.520718,
		0.435159,
		0.34826,
		0.262931,
		0.182603,
		0.111261,
		0.053473,
		0.014433,
		0.002179,
		0.013263,
		0.033031,
		0.060681,
		0.095391,
		0.136317,
		0.182603,
		0.233376,
		0.28775,
		0.344836,
		0.403741,
		0.463572,
		0.523449,
		0.582507,
		0.63991,
		0.694856,
		0.746597,
		0.794448,
		0.837812,
		0.876195,
		0.909232,
		0.936718,
		0.958631,
		0.975164,
		0.986746,
		0.99405,
		0.997978,
		0.999593,
		0.999979,
		1,
		0.999933,
		0.999236,
		0.996742,
		0.990961,
		0.980326,
		0.963372,
		0.938862,
		0.905864,
		0.863799,
		0.812474,
		0.752097,
		0.683291,
		0.607109,
		0.52505,
		0.439072,
		0.351615,
		0.265625,
		0.184582,
		0.112529,
		0.054111,
		0.014613,
		0.000557,
		0.008674,
		0.025848,
		0.051267,
		0.084099,
		0.123492,
		0.168575,
		0.218463,
		0.272257,
		0.32905,
		0.387931,
		0.447991,
		0.508329,
		0.568056,
		0.626313,
		0.682272,
		0.735156,
		0.78425,
		0.828924,
		0.868649,
		0.903024,
		0.931803,
		0.954925,
		0.972543,
		0.985047,
		0.993081,
		0.997527,
		0.999448,
		0.999961,
		1,
		0.999959,
		0.999399,
		0.997195,
		0.991857,
		0.981785,
		0.965467,
		0.94161,
		0.909232,
		0.867708,
		0.816804,
		0.756699,
		0.687996,
		0.61174,
		0.529428,
		0.443032,
		0.355015,
		0.268359,
		0.186592,
		0.113819,
		0.054761,
		0.014796,
		0.004986,
		0.019411,
		0.042454,
		0.073275,
		0.11101,
		0.154777,
		0.203676,
		0.256796,
		0.313213,
		0.371999,
		0.432226,
		0.492972,
		0.553326,
		0.612405,
		0.669354,
		0.723368,
		0.773702,
		0.819691,
		0.860772,
		0.896506,
		0.926607,
		0.950973,
		0.969716,
		0.983187,
		0.991996,
		0.997003,
		0.999268,
		0.999933,
		1,
		0.999977,
		0.999536,
		0.997607,
		0.9927,
		0.983187,
		0.967506,
		0.944313,
		0.912568,
		0.8716,
		0.821134,
		0.761317,
		0.692731,
		0.61641,
		0.533853,
		0.447041,
		0.358463,
		0.271135,
		0.188634,
		0.11513,
		0.055422,
		0.014982,
		0.002265,
		0.013779,
		0.034297,
		0.062966,
		0.098913,
		0.141244,
		0.189047,
		0.241394,
		0.297347,
		0.355962,
		0.416289,
		0.477387,
		0.538323,
		0.598188,
		0.656102,
		0.71123,
		0.762798,
		0.810106,
		0.852555,
		0.889668,
		0.92112,
		0.946765,
		0.966673,
		0.981155,
		0.990785,
		0.996398,
		0.999046,
		0.999892,
		1,
		0.999989,
		0.999651,
		0.997978,
		0.99349,
		0.98453,
		0.969489,
		0.946968,
		0.91587,
		0.875474,
		0.825463,
		0.76595,
		0.697495,
		0.62112,
		0.538323,
		0.451099,
		0.361957,
		0.273952,
		0.19071,
		0.116464,
		0.056095,
		0.015172,
		0.000579,
		0.009016,
		0.02685,
		0.053221,
		0.087245,
		0.128015,
		0.174607,
		0.226079,
		0.281477,
		0.339837,
		0.400195,
		0.461585,
		0.523054,
		0.583667,
		0.642517,
		0.698742,
		0.751535,
		0.800163,
		0.843991,
		0.882503,
		0.915332,
		0.94229,
		0.963404,
		0.978942,
		0.98944,
		0.995704,
		0.998776,
		0.999834,
		0.999998,
		0.999995,
		0.999744,
		0.99831,
		0.994227,
		0.985814,
		0.971415,
		0.949573,
		0.919137,
		0.879329,
		0.829789,
		0.770596,
		0.702287,
		0.625869,
		0.542841,
		0.455206,
		0.365499,
		0.276811,
		0.192819,
		0.117821,
		0.056781,
		0.015366,
		0.005185,
		0.020173,
		0.044093,
		0.076051,
		0.11513,
		0.160392,
		0.21088,
		0.265625,
		0.323646,
		0.383959,
		0.445578,
		0.507526,
		0.568845,
		0.628601,
		0.685901,
		0.739908,
		0.789856,
		0.835071,
		0.875,
		0.909232,
		0.937538,
		0.959898,
		0.976536,
		0.987949,
		0.994912,
		0.998451,
		0.999755,
		0.999995,
		0.999998,
		0.999819,
		0.998605,
		0.994912,
		0.987038,
		0.973282,
		0.952128,
		0.922365,
		0.883162,
		0.834112,
		0.775256,
		0.707107,
		0.630658,
		0.547405,
		0.459363,
		0.369091,
		0.279714,
		0.194962,
		0.119202,
		0.057479,
		0.015563,
		0.002356,
		0.014327,
		0.035637,
		0.065382,
		0.102633,
		0.146439,
		0.19583,
		0.249819,
		0.307411,
		0.367599,
		0.429378,
		0.491749,
		0.553728,
		0.614355,
		0.672708,
		0.727916,
		0.77918,
		0.82579,
		0.867151,
		0.902812,
		0.932498,
		0.956143,
		0.973927,
		0.986304,
		0.994013,
		0.998064,
		0.999651,
		0.999988,
		1,
		0.999876,
		0.998863,
		0.995546,
		0.988203,
		0.975089,
		0.95463,
		0.925553,
		0.886971,
		0.838428,
		0.779928,
		0.711954,
		0.635486,
		0.552017,
		0.463572,
		0.372731,
		0.282661,
		0.197141,
		0.120607,
		0.05819,
		0.015765,
		0.000602,
		0.009378,
		0.027912,
		0.055289,
		0.090569,
		0.132788,
		0.180962,
		0.234088,
		0.291153,
		0.351133,
		0.413,
		0.475734,
		0.538323,
		0.599783,
		0.65916,
		0.715554,
		0.768129,
		0.816138,
		0.858946,
		0.89606,
		0.927158,
		0.952128,
		0.971103,
		0.984492,
		0.992999,
		0.997607,
		0.999515,
		0.999975,
		1,
		0.99992,
		0.999088,
		0.996128,
		0.989308,
		0.976835,
		0.957078,
		0.9287,
		0.890756,
		0.842738,
		0.784611,
		0.716828,
		0.640354,
		0.556677,
		0.467831,
		0.376422,
		0.285652,
		0.199356,
		0.122037,
		0.058915,
		0.01597,
		0.005396,
		0.02098,
		0.045828,
		0.078987,
		0.119481,
		0.166312,
		0.218463,
		0.2749,
		0.334582,
		0.39646,
		0.45949,
		0.522638,
		0.584888,
		0.645261,
		0.702821,
		0.756699,
		0.80611,
		0.850378,
		0.888967,
		0.921508,
		0.947842,
		0.968053,
		0.982503,
		0.991857,
		0.997073,
		0.999343,
		0.999953,
		1,
		0.999951,
		0.999281,
		0.996662,
		0.990352,
		0.978519,
		0.959469,
		0.931803,
		0.894513,
		0.84704,
		0.789303,
		0.721729,
		0.645261,
		0.561384,
		0.472143,
		0.380165,
		0.28869,
		0.201607,
		0.123492,
		0.059653,
		0.016179,
		0.002453,
		0.014907,
		0.037057,
		0.067939,
		0.106564,
		0.151921,
		0.202976,
		0.258679,
		0.317968,
		0.379775,
		0.443032,
		0.50668,
		0.569676,
		0.63101,
		0.689715,
		0.744885,
		0.795699,
		0.841438,
		0.881522,
		0.915536,
		0.943272,
		0.964764,
		0.980326,
		0.990579,
		0.996452,
		0.999128,
		0.99992,
		1,
		0.999973,
		0.999445,
		0.997147,
		0.991337,
		0.98014,
		0.961803,
		0.93486,
		0.898242,
		0.851331,
		0.794004,
		0.726654,
		0.650207,
		0.56614,
		0.476508,
		0.383959,
		0.291774,
		0.203896,
		0.124973,
		0.060404,
		0.016392,
		0.000627,
		0.009762,
		0.029038,
		0.057479,
		0.094085,
		0.137829,
		0.187663,
		0.242518,
		0.301315,
		0.362964,
		0.426374,
		0.49046,
		0.554152,
		0.61641,
		0.676234,
		0.732684,
		0.784898,
		0.832117,
		0.873716,
		0.909232,
		0.938407,
		0.961225,
		0.977948,
		0.989154,
		0.995735,
		0.998863,
		0.999871,
		1,
		0.999986,
		0.999581,
		0.997585,
		0.992261,
		0.981696,
		0.964076,
		0.93787,
		0.901939,
		0.85561,
		0.798713,
		0.731605,
		0.655192,
		0.570945,
		0.480925,
		0.387806,
		0.294906,
		0.206223,
		0.12648,
		0.06117,
		0.01661,
		0.00562,
		0.021837,
		0.047667,
		0.082095,
		0.124081,
		0.172561,
		0.226451,
		0.28465,
		0.346048,
		0.409531,
		0.473988,
		0.538323,
		0.601465,
		0.66238,
		0.720092,
		0.773702,
		0.822408,
		0.865538,
		0.902584,
		0.933236,
		0.957423,
		0.975358,
		0.987569,
		0.994912,
		0.998541,
		0.999801,
		0.999998,
		0.999994,
		0.999693,
		0.997978,
		0.993125,
		0.983187,
		0.966289,
		0.94083,
		0.905603,
		0.859876,
		0.803427,
		0.736579,
		0.660215,
		0.575799,
		0.485397,
		0.391707,
		0.298086,
		0.208589,
		0.128015,
		0.061951,
		0.016832,
		0.002556,
		0.015523,
		0.038563,
		0.070648,
		0.110724,
		0.157713,
		0.210511,
		0.268,
		0.32905,
		0.392521,
		0.457278,
		0.522198,
		0.586177,
		0.648151,
		0.707107,
		0.762104,
		0.812301,
		0.85698,
		0.895582,
		0.927745,
		0.953345,
		0.972543,
		0.985814,
		0.993974,
		0.998153,
		0.999707,
		0.999994,
		0.999998,
		0.999782,
		0.998327,
		0.99393,
		0.984612,
		0.968439,
		0.943738,
		0.909232,
		0.864126,
		0.808146,
		0.741577,
		0.665277,
		0.580701,
		0.489923,
		0.395662,
		0.301315,
		0.210995,
		0.129577,
		0.062746,
		0.017058,
		0.000654,
		0.010171,
		0.030233,
		0.059802,
		0.097809,
		0.14316,
		0.194735,
		0.251397,
		0.311993,
		0.375363,
		0.440344,
		0.505785,
		0.570554,
		0.63355,
		0.693726,
		0.750101,
		0.801791,
		0.848031,
		0.888213,
		0.921922,
		0.94898,
		0.969489,
		0.983876,
		0.992908,
		0.997692,
		0.999581,
		0.999986,
		1,
		0.999851,
		0.998635,
		0.994676,
		0.98597,
		0.970524,
		0.946591,
		0.912824,
		0.868358,
		0.812868,
		0.746597,
		0.670377,
		0.585653,
		0.494505,
		0.399673,
		0.304595,
		0.213443,
		0.131168,
		0.063557,
		0.017289,
		0.005858,
		0.022747,
		0.049618,
		0.085389,
		0.128949,
		0.179163,
		0.234873,
		0.294906,
		0.358077,
		0.423201,
		0.489096,
		0.5546,
		0.618579,
		0.679948,
		0.737688,
		0.790869,
		0.838682,
		0.880467,
		0.915755,
		0.944313,
		0.966185,
		0.981742,
		0.991705,
		0.997147,
		0.999419,
		0.99997,
		1,
		0.999903,
		0.998903,
		0.995364,
		0.987261,
		0.972543,
		0.949389,
		0.916375,
		0.872571,
		0.817592,
		0.751638,
		0.675515,
		0.590655,
		0.499142,
		0.403741,
		0.307926,
		0.215931,
		0.132788,
		0.064384,
		0.017525,
		0.002666,
		0.016179,
		0.040162,
		0.073521,
		0.11513,
		0.163836,
		0.218463,
		0.277816,
		0.340688,
		0.405867,
		0.472143,
		0.538323,
		0.603241,
		0.665773,
		0.72486,
		0.77953,
		0.828924,
		0.872333,
		0.909232,
		0.939331,
		0.962616,
		0.9794,
		0.990352,
		0.996509,
		0.999212,
		0.999944,
		1,
		0.999941,
		0.999134,
		0.995996,
		0.988485,
		0.974494,
		0.952128,
		0.919885,
		0.876762,
		0.822315,
		0.756699,
		0.68069,
		0.595706,
		0.503836,
		0.407865,
		0.31131,
		0.218463,
		0.134438,
		0.065227,
		0.017765,
		0.000682,
		0.010606,
		0.031504,
		0.062267,
		0.101756,
		0.148801,
		0.202206,
		0.260756,
		0.32322,
		0.38836,
		0.45494,
		0.521733,
		0.58754,
		0.651201,
		0.711615,
		0.767766,
		0.818748,
		0.863799,
		0.902341,
		0.934022,
		0.958769,
		0.976835,
		0.988837,
		0.995768,
		0.998953,
		0.999903,
		1,
		0.999967,
		0.99933,
		0.996571,
		0.989641,
		0.976376,
		0.954807,
		0.92335,
		0.880929,
		0.827036,
		0.761779,
		0.685901,
		0.600807,
		0.508587,
		0.412048,
		0.314747,
		0.221038,
		0.136118,
		0.066086,
		0.018011,
		0.006111,
		0.023716,
		0.051692,
		0.088884,
		0.134105,
		0.186142,
		0.243759,
		0.3057,
		0.370703,
		0.437501,
		0.504839,
		0.571482,
		0.636233,
		0.69795,
		0.755573,
		0.808146,
		0.854856,
		0.895068,
		0.928371,
		0.95463,
		0.974035,
		0.987147,
		0.994912,
		0.998635,
		0.999844,
		1,
		0.999983,
		0.999493,
		0.997092,
		0.990728,
		0.978187,
		0.957423,
		0.926769,
		0.88507,
		0.831754,
		0.766878,
		0.69115,
		0.605958,
		0.513395,
		0.416289,
		0.318238,
		0.223658,
		0.137829,
		0.066963,
		0.018262,
		0.002783,
		0.016877,
		0.041863,
		0.076573,
		0.119801,
		0.170317,
		0.226862,
		0.288158,
		0.352917,
		0.419845,
		0.487653,
		0.555074,
		0.620871,
		0.683863,
		0.742944,
		0.797109,
		0.845492,
		0.887401,
		0.922365,
		0.950184,
		0.970985,
		0.985268,
		0.99393,
		0.998248,
		0.99976,
		0.999998,
		0.999993,
		0.999627,
		0.997561,
		0.991748,
		0.979927,
		0.959974,
		0.930138,
		0.889182,
		0.836467,
		0.771993,
		0.696434,
		0.611159,
		0.518262,
		0.420591,
		0.321785,
		0.226323,
		0.139573,
		0.067857,
		0.018518,
		0.000713,
		0.011069,
		0.032856,
		0.064888,
		0.105945,
		0.154777,
		0.210104,
		0.270627,
		0.33503,
		0.40199,
		0.470189,
		0.538323,
		0.605119,
		0.669354,
		0.729875,
		0.78563,
		0.835698,
		0.879329,
		0.91599,
		0.945417,
		0.967671,
		0.983187,
		0.992809,
		0.997782,
		0.999645,
		0.999993,
		0.999998,
		0.999735,
		0.997978,
		0.9927,
		0.981594,
		0.962458,
		0.933455,
		0.893264,
		0.841172,
		0.777124,
		0.701753,
		0.61641,
		0.523188,
		0.424954,
		0.325389,
		0.229035,
		0.141349,
		0.068768,
		0.018779,
		0.006382,
		0.024747,
		0.053897,
		0.092596,
		0.139573,
		0.193529,
		0.253142,
		0.317068,
		0.383959,
		0.452462,
		0.52124,
		0.588982,
		0.654422,
		0.716363,
		0.773702,
		0.825463,
		0.870838,
		0.909232,
		0.940314,
		0.964076,
		0.980889,
		0.991538,
		0.997227,
		0.999493,
		0.999982,
		1,
		0.999819,
		0.998347,
		0.993585,
		0.983187,
		0.964874,
		0.936718,
		0.897313,
		0.845868,
		0.782268,
		0.707107,
		0.621711,
		0.528173,
		0.429378,
		0.32905,
		0.231794,
		0.14316,
		0.069699,
		0.019046,
		0.002907,
		0.017621,
		0.043674,
		0.079817,
		0.12476,
		0.177183,
		0.23574,
		0.299063,
		0.365774,
		0.434489,
		0.503836,
		0.572465,
		0.639069,
		0.702404,
		0.761317,
		0.814777,
		0.861918,
		0.902078,
		0.93486,
		0.960187,
		0.978359,
		0.990101,
		0.996571,
		0.999296,
		0.999963,
		1,
		0.999882,
		0.998669,
		0.994403,
		0.984704,
		0.967218,
		0.939924,
		0.901325,
		0.850552,
		0.787425,
		0.712494,
		0.627063,
		0.533218,
		0.433865,
		0.332769,
		0.234603,
		0.145004,
		0.070648,
		0.019319,
		0.000745,
		0.011564,
		0.034297,
		0.067676,
		0.110396,
		0.161115,
		0.218463,
		0.281048,
		0.347461,
		0.416289,
		0.486123,
		0.555576,
		0.623297,
		0.687996,
		0.74847,
		0.803632,
		0.852555,
		0.894513,
		0.929041,
		0.955987,
		0.975582,
		0.988485,
		0.995804,
		0.999046,
		0.999931,
		1,
		0.999928,
		0.998946,
		0.995156,
		0.986146,
		0.969489,
		0.943071,
		0.905299,
		0.855222,
		0.792593,
		0.717915,
		0.632464,
		0.538323,
		0.438416,
		0.336549,
		0.237461,
		0.146884,
		0.071616,
		0.019597,
		0.006671,
		0.025848,
		0.056247,
		0.096544,
		0.145377,
		0.201355,
		0.263058,
		0.32905,
		0.397884,
		0.468117,
		0.538323,
		0.607109,
		0.673139,
		0.735156,
		0.792019,
		0.842738,
		0.886525,
		0.92284,
		0.951461,
		0.972543,
		0.986675,
		0.994912,
		0.998733,
		0.999882,
		1,
		0.999959,
		0.999183,
		0.995843,
		0.987511,
		0.971686,
		0.946156,
		0.909232,
		0.859876,
		0.797771,
		0.723368,
		0.637915,
		0.54349,
		0.443032,
		0.34039,
		0.24037,
		0.148801,
		0.072605,
		0.019882,
		0.003041,
		0.018415,
		0.045605,
		0.08327,
		0.130029,
		0.184465,
		0.245133,
		0.31057,
		0.379296,
		0.449834,
		0.520718,
		0.590511,
		0.657831,
		0.721369,
		0.779928,
		0.832457,
		0.878098,
		0.916242,
		0.946591,
		0.969224,
		0.984657,
		0.993882,
		0.998347,
		0.999809,
		0.999999,
		0.99998,
		0.999381,
		0.996468,
		0.9888,
		0.973805,
		0.949176,
		0.913121,
		0.864511,
		0.802956,
		0.728851,
		0.643416,
		0.548718,
		0.447714,
		0.344293,
		0.243331,
		0.150754,
		0.073614,
		0.020173,
		0.00078,
		0.012092,
		0.035835,
		0.070648,
		0.11513,
		0.167843,
		0.227317,
		0.292057,
		0.360554,
		0.431293,
		0.502771,
		0.573509,
		0.642073,
		0.707107,
		0.767353,
		0.821699,
		0.869219,
		0.909232,
		0.941362,
		0.965609,
		0.982413,
		0.9927,
		0.997877,
		0.999707,
		0.999997,
		0.999991,
		0.999544,
		0.99703,
		0.990011,
		0.975845,
		0.952128,
		0.916963,
		0.869125,
		0.808146,
		0.734365,
		0.648967,
		0.554008,
		0.452462,
		0.34826,
		0.246346,
		0.152746,
		0.074644,
		0.02047,
		0.006979,
		0.027023,
		0.058753,
		0.100747,
		0.151546,
		0.209654,
		0.273547,
		0.341684,
		0.412516,
		0.484498,
		0.556109,
		0.625869,
		0.692366,
		0.754286,
		0.810454,
		0.859876,
		0.901795,
		0.935757,
		0.961681,
		0.979927,
		0.991352,
		0.997312,
		0.999567,
		0.999991,
		0.999997,
		0.999675,
		0.997533,
		0.991144,
		0.977805,
		0.95501,
		0.920755,
		0.873716,
		0.81334,
		0.739908,
		0.654566,
		0.559361,
		0.457278,
		0.352291,
		0.249415,
		0.154777,
		0.075696,
		0.020774,
		0.003183,
		0.019264,
		0.047667,
		0.086952,
		0.135635,
		0.192195,
		0.255081,
		0.32272,
		0.393526,
		0.465916,
		0.538323,
		0.609221,
		0.677145,
		0.740722,
		0.798713,
		0.850054,
		0.893914,
		0.929759,
		0.957423,
		0.977183,
		0.989821,
		0.996638,
		0.999381,
		0.999978,
		0.999999,
		0.999777,
		0.997978,
		0.992201,
		0.979683,
		0.957819,
		0.924495,
		0.87828,
		0.818536,
		0.745479,
		0.660215,
		0.564777,
		0.462163,
		0.356389,
		0.25254,
		0.156848,
		0.07677,
		0.021084,
		0.000817,
		0.012657,
		0.037478,
		0.073818,
		0.120172,
		0.174994,
		0.236704,
		0.303696,
		0.374349,
		0.447041,
		0.520162,
		0.592135,
		0.661443,
		0.726654,
		0.786465,
		0.839741,
		0.885575,
		0.92335,
		0.952816,
		0.974162,
		0.988091,
		0.995843,
		0.99914,
		0.999955,
		1,
		0.999854,
		0.998368,
		0.993181,
		0.981477,
		0.960553,
		0.928179,
		0.882814,
		0.823732,
		0.751077,
		0.665912,
		0.570256,
		0.467118,
		0.360554,
		0.255722,
		0.15896,
		0.077867,
		0.021402,
		0.00731,
		0.02828,
		0.061429,
		0.105229,
		0.15811,
		0.218463,
		0.28465,
		0.355015,
		0.427896,
		0.501639,
		0.574618,
		0.645261,
		0.712079,
		0.773702,
		0.828924,
		0.876762,
		0.916514,
		0.947842,
		0.970847,
		0.986146,
		0.994912,
		0.998834,
		0.999916,
		1,
		0.999911,
		0.998706,
		0.994084,
		0.983187,
		0.963209,
		0.931803,
		0.887317,
		0.828924,
		0.756699,
		0.671658,
		0.575799,
		0.472143,
		0.364787,
		0.258963,
		0.161115,
		0.078987,
		0.021727,
		0.003336,
		0.020173,
		0.049871,
		0.09088,
		0.141606,
		0.200411,
		0.265625,
		0.335557,
		0.408505,
		0.482769,
		0.556677,
		0.628601,
		0.696992,
		0.760415,
		0.817592,
		0.867459,
		0.909232,
		0.942482,
		0.967218,
		0.983968,
		0.993829,
		0.998451,
		0.999854,
		1,
		0.99995,
		0.998994,
		0.994912,
		0.98481,
		0.965784,
		0.935365,
		0.891784,
		0.834112,
		0.762345,
		0.677451,
		0.581406,
		0.47724,
		0.369091,
		0.262263,
		0.163312,
		0.080132,
		0.022059,
		0.000857,
		0.013263,
		0.039236,
		0.077206,
		0.12555,
		0.182603,
		0.246666,
		0.31601,
		0.388893,
		0.463572,
		0.538323,
		0.611467,
		0.681392,
		0.746597,
		0.805731,
		0.857652,
		0.901488,
		0.936718,
		0.963257,
		0.981538,
		0.992579,
		0.997978,
		0.999765,
		0.999999,
		0.999975,
		0.999236,
		0.995666,
		0.986346,
		0.968276,
		0.938862,
		0.896212,
		0.839291,
		0.768013,
		0.683291,
		0.587077,
		0.48241,
		0.373465,
		0.265625,
		0.165555,
		0.081301,
		0.022399,
		0.007665,
		0.029626,
		0.064291,
		0.110014,
		0.165103,
		0.227824,
		0.296414,
		0.369091,
		0.444067,
		0.51957,
		0.593863,
		0.665277,
		0.732241,
		0.793332,
		0.847326,
		0.893264,
		0.930531,
		0.958943,
		0.978837,
		0.991144,
		0.997403,
		0.999638,
		0.999997,
		0.999989,
		0.999435,
		0.996348,
		0.987794,
		0.970681,
		0.94229,
		0.900598,
		0.84446,
		0.773702,
		0.689177,
		0.592813,
		0.487653,
		0.377913,
		0.269049,
		0.167843,
		0.082496,
		0.022747,
		0.003501,
		0.021147,
		0.05223,
		0.09508,
		0.147975,
		0.209152,
		0.276811,
		0.349131,
		0.424278,
		0.500431,
		0.575799,
		0.648648,
		0.717343,
		0.780384,
		0.836467,
		0.884543,
		0.9239,
		0.954256,
		0.975845,
		0.989507,
		0.996711,
		0.999465,
		0.999989,
		0.999996,
		0.999596,
		0.996959,
		0.989154,
		0.972999,
		0.945646,
		0.90494,
		0.849616,
		0.779408,
		0.695109,
		0.598615,
		0.492972,
		0.382435,
		0.272538,
		0.170178,
		0.083718,
		0.023103,
		0.000899,
		0.013913,
		0.041121,
		0.08083,
		0.131291,
		0.19071,
		0.25725,
		0.32905,
		0.404232,
		0.480925,
		0.557281,
		0.631507,
		0.701897,
		0.766878,
		0.82506,
		0.875306,
		0.916807,
		0.949176,
		0.972543,
		0.987648,
		0.995886,
		0.999236,
		0.999973,
		0.999999,
		0.999723,
		0.997501,
		0.990425,
		0.975226,
		0.948926,
		0.909232,
		0.854755,
		0.785132,
		0.701086,
		0.604481,
		0.498366,
		0.387032,
		0.276092,
		0.172561,
		0.084966,
		0.023468,
		0.008046,
		0.031071,
		0.067358,
		0.11513,
		0.172561,
		0.237781,
		0.308888,
		0.383959,
		0.461072,
		0.538323,
		0.613858,
		0.685901,
		0.752804,
		0.813093,
		0.865538,
		0.909232,
		0.943681,
		0.968908,
		0.985548,
		0.994912,
		0.998938,
		0.999944,
		1,
		0.999819,
		0.997978,
		0.991607,
		0.97736,
		0.952128,
		0.913472,
		0.859876,
		0.790869,
		0.707107,
		0.610413,
		0.503836,
		0.391707,
		0.279714,
		0.174994,
		0.086243,
		0.023841,
		0.003677,
		0.022194,
		0.054761,
		0.099574,
		0.154777,
		0.218463,
		0.28869,
		0.363491,
		0.440894,
		0.518938,
		0.595706,
		0.669354,
		0.738155,
		0.800552,
		0.855222,
		0.901155,
		0.937751,
		0.964919,
		0.983187,
		0.993771,
		0.99856,
		0.999895,
		1,
		0.999889,
		0.998392,
		0.9927,
		0.9794,
		0.955247,
		0.917657,
		0.864974,
		0.796619,
		0.71317,
		0.61641,
		0.509384,
		0.39646,
		0.283404,
		0.177477,
		0.087547,
		0.024222,
		0.000945,
		0.014613,
		0.043145,
		0.084714,
		0.137432,
		0.199356,
		0.268504,
		0.342867,
		0.420418,
		0.499142,
		0.577059,
		0.652255,
		0.722924,
		0.787425,
		0.84434,
		0.892557,
		0.931363,
		0.960553,
		0.980542,
		0.992444,
		0.998086,
		0.999819,
		1,
		0.999937,
		0.998747,
		0.993706,
		0.981342,
		0.95828,
		0.921781,
		0.870045,
		0.802379,
		0.719275,
		0.622473,
		0.515011,
		0.401293,
		0.287164,
		0.180011,
		0.088879,
		0.024612,
		0.008457,
		0.032625,
		0.070648,
		0.120607,
		0.180527,
		0.248386,
		0.322126,
		0.399673,
		0.478955,
		0.557927,
		0.634605,
		0.707107,
		0.773702,
		0.832877,
		0.883417,
		0.924495,
		0.955789,
		0.977591,
		0.990912,
		0.997501,
		0.999707,
		0.999999,
		0.999968,
		0.999046,
		0.994625,
		0.983187,
		0.961225,
		0.925841,
		0.875088,
		0.808146,
		0.72542,
		0.628601,
		0.520717,
		0.406206,
		0.290993,
		0.182595,
		0.090239,
		0.02501,
		0.003868,
		0.023321,
		0.057479,
		0.104393,
		0.162051,
		0.228392,
		0.301315,
		0.378694,
		0.4584,
		0.538323,
		0.61641,
		0.690698,
		0.759371,
		0.820817,
		0.873716,
		0.917126,
		0.950601,
		0.974313,
		0.989154,
		0.99679,
		0.999548,
		0.999996,
		0.999986,
		0.999293,
		0.995458,
		0.984931,
		0.964076,
		0.929834,
		0.880097,
		0.813918,
		0.731605,
		0.634793,
		0.526502,
		0.411199,
		0.294891,
		0.185227,
		0.091624,
		0.025414,
		0.000995,
		0.015366,
		0.045322,
		0.088884,
		0.144007,
		0.208589,
		0.280485,
		0.357518,
		0.437501,
		0.518262,
		0.597675,
		0.673697,
		0.744423,
		0.808146,
		0.863435,
		0.909232,
		0.944967,
		0.970681,
		0.987147,
		0.995933,
		0.999333,
		0.999986,
		0.999995,
		0.999493,
		0.996208,
		0.986574,
		0.966832,
		0.933754,
		0.88507,
		0.819691,
		0.737825,
		0.641049,
		0.532363,
		0.416269,
		0.298855,
		0.187904,
		0.093032,
		0.025825,
		0.0089,
		0.034297,
		0.074184,
		0.12648,
		0.189047,
		0.259691,
		0.336187,
		0.416289,
		0.497762,
		0.578407,
		0.656102,
		0.728851,
		0.794849,
		0.852555,
		0.900792,
		0.938862,
		0.966673,
		0.984868,
		0.994912,
		0.999046,
		0.999966,
		0.999999,
		0.999651,
		0.996876,
		0.988116,
		0.969489,
		0.937598,
		0.890001,
		0.825462,
		0.74408,
		0.647364,
		0.538297,
		0.421411,
		0.302876,
		0.190619,
		0.094456,
		0.026239,
		0.004073,
		0.024536,
		0.060404,
		0.109566,
		0.169842,
		0.238994,
		0.314747,
		0.394798,
		0.476846,
		0.558619,
		0.637915,
		0.712649,
		0.780913,
		0.841057,
		0.891784,
		0.932261,
		0.962262,
		0.982294,
		0.993706,
		0.998674,
		0.999929,
		1,
		0.999771,
		0.997465,
		0.989554,
		0.972044,
		0.941362,
		0.894886,
		0.831226,
		0.750362,
		0.653732,
		0.544295,
		0.426612,
		0.306942,
		0.19336,
		0.09589,
		0.026654,
		0.001049,
		0.016179,
		0.047667,
		0.093366,
		0.151058,
		0.218463,
		0.293251,
		0.373065,
		0.455537,
		0.538323,
		0.61914,
		0.69581,
		0.766326,
		0.828924,
		0.882183,
		0.92514,
		0.957423,
		0.9794,
		0.992293,
		0.9982,
		0.999867,
		1,
		0.999859,
		0.997978,
		0.99089,
		0.974494,
		0.94504,
		0.89972,
		0.836975,
		0.756663,
		0.66014,
		0.550341,
		0.431857,
		0.311037,
		0.196111,
		0.097323,
		0.027066,
		0.009378,
		0.036101,
		0.07799,
		0.132788,
		0.19817,
		0.271757,
		0.351133,
		0.433865,
		0.517538,
		0.599783,
		0.678333,
		0.751077,
		0.816138,
		0.87197,
		0.917472,
		0.952128,
		0.976158,
		0.99065,
		0.997607,
		0.999771,
		1,
		0.99992,
		0.998419,
		0.992122,
		0.976834,
		0.948626,
		0.904491,
		0.842698,
		0.762966,
		0.666569,
		0.556411,
		0.437117,
		0.315133,
		0.198852,
		0.098742,
		0.027471,
		0.004296,
		0.025848,
		0.063557,
		0.11513,
		0.178197,
		0.250328,
		0.32905,
		0.411865,
		0.496282,
		0.579853,
		0.660215,
		0.735156,
		0.802682,
		0.861121,
		0.909232,
		0.946351,
		0.972543,
		0.988753,
		0.996876,
		0.999629,
		0.999999,
		0.999959,
		0.998793,
		0.993251,
		0.979061,
		0.952111,
		0.909189,
		0.848375,
		0.769247,
		0.672987,
		0.56247,
		0.442357,
		0.319197,
		0.201554,
		0.100129,
		0.027862,
		0.001107,
		0.017058,
		0.050198,
		0.098193,
		0.158632,
		0.229035,
		0.306869,
		0.389573,
		0.474581,
		0.559361,
		0.641458,
		0.718555,
		0.788542,
		0.849616,
		0.900395,
		0.940063,
		0.968525,
		0.986574,
		0.995985,
		0.999429,
		0.999994,
		0.999982,
		0.999102,
		0.994276,
		0.981167,
		0.955481,
		0.91379,
		0.853978,
		0.77547,
		0.679351,
		0.56847,
		0.447525,
		0.323181,
		0.204182,
		0.101464,
		0.028234,
		0.009896,
		0.038051,
		0.082095,
		0.139573,
		0.207954,
		0.28465,
		0.367032,
		0.452462,
		0.538323,
		0.622067,
		0.701268,
		0.773702,
		0.837433,
		0.890936,
		0.933236,
		0.964076,
		0.984089,
		0.994912,
		0.999156,
		0.999982,
		0.999993,
		0.999349,
		0.995193,
		0.983139,
		0.958714,
		0.918264,
		0.859466,
		0.781582,
		0.685602,
		0.574344,
		0.452559,
		0.327031,
		0.206694,
		0.102721,
		0.028577,
		0.004537,
		0.027267,
		0.066963,
		0.121124,
		0.187172,
		0.262459,
		0.344293,
		0.42996,
		0.516759,
		0.602046,
		0.683291,
		0.758149,
		0.824552,
		0.880827,
		0.925841,
		0.959167,
		0.981267,
		0.993633,
		0.998793,
		0.999955,
		0.999993,
		0.999536,
		0.995994,
		0.98496,
		0.961779,
		0.922566,
		0.864779,
		0.787513,
		0.691658,
		0.580011,
		0.457378,
		0.330677,
		0.209038,
		0.103873,
		0.028884,
		0.00117,
		0.018011,
		0.052935,
		0.103402,
		0.166781,
		0.24037,
		0.321409,
		0.407111,
		0.494689,
		0.581406,
		0.664622,
		0.741872,
		0.810953,
		0.870045,
		0.91785,
		0.953766,
		0.978079,
		0.992122,
		0.99832,
		0.999904,
		0.999982,
		0.999659,
		0.996665,
		0.986599,
		0.96463,
		0.926633,
		0.86984,
		0.793172,
		0.697422,
		0.585368,
		0.461886,
		0.334039,
		0.211158,
		0.104887,
		0.029144,
		0.010458,
		0.040162,
		0.086531,
		0.146884,
		0.218463,
		0.298442,
		0.383959,
		0.472143,
		0.560159,
		0.645261,
		0.72486,
		0.796619,
		0.858565,
		0.909232,
		0.947842,
		0.974494,
		0.990351,
		0.997716,
		0.999812,
		0.999952,
		0.999705,
		0.997181,
		0.988014,
		0.967203,
		0.930382,
		0.874547,
		0.798444,
		0.702768,
		0.590294,
		0.465973,
		0.337025,
		0.212988,
		0.105728,
		0.029348,
		0.004799,
		0.028807,
		0.070648,
		0.127594,
		0.196828,
		0.27546,
		0.360554,
		0.449152,
		0.538323,
		0.625212,
		0.707107,
		0.781532,
		0.846361,
		0.899959,
		0.941362,
		0.970479,
		0.988288,
		0.996952,
		0.999659,
		0.999887,
		0.999655,
		0.997502,
		0.989142,
		0.969415,
		0.933706,
		0.878774,
		0.803188,
		0.707552,
		0.594644,
		0.46951,
		0.339533,
		0.21446,
		0.10636,
		0.029484,
		0.001239,
		0.019046,
		0.055902,
		0.109032,
		0.175563,
		0.25254,
		0.336951,
		0.425753,
		0.515919,
		0.604481,
		0.688606,
		0.765677,
		0.833411,
		0.890001,
		0.934294,
		0.965998,
		0.985898,
		0.995994,
		0.999412,
		0.999756,
		0.999471,
		0.997575,
		0.989904,
		0.971156,
		0.936471,
		0.882368,
		0.807238,
		0.711604,
		0.59826,
		0.472358,
		0.341453,
		0.215498,
		0.106744,
		0.029542,
		0.011069,
		0.042454,
		0.091333,
		0.154777,
		0.229766,
		0.313213,
		0.40199,
		0.492972,
		0.583079,
		0.669354,
		0.749038,
		0.819691,
		0.879328,
		0.926602,
		0.961014,
		0.983139,
		0.994799,
		0.999026,
		0.999516,
		0.9991,
		0.997321,
		0.990193,
		0.972293,
		0.938519,
		0.885148,
		0.810403,
		0.714734,
		0.600964,
		0.474364,
		0.34267,
		0.216028,
		0.106843,
		0.029512,
		0.005084,
		0.030481,
		0.074644,
		0.134589,
		0.207232,
		0.289411,
		0.377913,
		0.469512,
		0.561021,
		0.649351,
		0.731605,
		0.805178,
		0.867908,
		0.918247,
		0.955481,
		0.979961,
		0.99331,
		0.998441,
		0.9991,
		0.998467,
		0.996642,
		0.98988,
		0.972665,
		0.939661,
		0.88691,
		0.812471,
		0.716734,
		0.602567,
		0.475373,
		0.343066,
		0.215973,
		0.10662,
		0.029384,
		0.001314,
		0.020173,
		0.059124,
		0.11513,
		0.185043,
		0.265625,
		0.353578,
		0.445578,
		0.538323,
		0.628601,
		0.713365,
		0.789852,
		0.85571,
		0.909189,
		0.949345,
		0.9763,
		0.991455,
		0.997575,
		0.998418,
		0.997473,
		0.995414,
		0.988808,
		0.972083,
		0.939685,
		0.887425,
		0.81321,
		0.717384,
		0.602874,
		0.475222,
		0.342523,
		0.215263,
		0.106041,
		0.029149,
		0.011736,
		0.044948,
		0.096544,
		0.163312,
		0.241943,
		0.32905,
		0.42121,
		0.515011,
		0.607109,
		0.694312,
		0.773688,
		0.842698,
		0.899376,
		0.942546,
		0.972079,
		0.989142,
		0.996322,
		0.997356,
		0.99599,
		0.993485,
		0.986795,
		0.970337,
		0.938357,
		0.886448,
		0.812377,
		0.716456,
		0.601687,
		0.473757,
		0.340931,
		0.213831,
		0.105077,
		0.028801,
		0.005396,
		0.032305,
		0.078987,
		0.142168,
		0.218463,
		0.304401,
		0.39646,
		0.491113,
		0.584885,
		0.674436,
		0.756663,
		0.82883,
		0.888751,
		0.935005,
		0.967203,
		0.986257,
		0.994554,
		0.99577,
		0.993865,
		0.990679,
		0.983637,
		0.967195,
		0.935426,
		0.883723,
		0.809723,
		0.713725,
		0.598814,
		0.470829,
		0.338191,
		0.211619,
		0.103703,
		0.028334,
		0.001396,
		0.021402,
		0.062632,
		0.121749,
		0.195295,
		0.279714,
		0.371384,
		0.466662,
		0.561943,
		0.653732,
		0.73875,
		0.81406,
		0.877246,
		0.926633,
		0.961557,
		0.982663,
		0.992111,
		0.993485,
		0.990914,
		0.986797,
		0.979108,
		0.96241,
		0.930634,
		0.878988,
		0.805002,
		0.708972,
		0.594078,
		0.466305,
		0.334215,
		0.208583,
		0.101903,
		0.027745,
		0.012464,
		0.047667,
		0.102209,
		0.172561,
		0.255081,
		0.346046,
		0.441699,
		0.538297,
		0.632191,
		0.719918,
		0.798332,
		0.864779,
		0.917318,
		0.955004,
		0.978195,
		0.988808,
		0.990301,
		0.986929,
		0.98162,
		0.97297,
		0.955728,
		0.923719,
		0.871991,
		0.797981,
		0.701997,
		0.58732,
		0.460076,
		0.328938,
		0.204691,
		0.099666,
		0.027034,
		0.005737,
		0.034297,
		0.083718,
		0.150396,
		0.230604,
		0.320521,
		0.416269,
		0.513969,
		0.609807,
		0.700132,
		0.781582,
		0.851254,
		0.906936,
		0.947385,
		0.972665,
		0.984431,
		0.985989,
		0.981678,
		0.974912,
		0.96498,
		0.946897,
		0.914431,
		0.862495,
		0.788449,
		0.692627,
		0.578414,
		0.45206,
		0.322316,
		0.199929,
		0.096994,
		0.026203,
		0.001486,
		0.022747,
		0.06646,
		0.128948,
		0.2064,
		0.294891,
		0.390425,
		0.488978,
		0.586569,
		0.679351,
		0.763735,
		0.836564,
		0.895339,
		0.938519,
		0.96586,
		0.978745,
		0.980301,
		0.974912,
		0.966433,
		0.954896,
		0.935677,
		0.902538,
		0.850289,
		0.77623,
		0.680727,
		0.567267,
		0.442204,
		0.314334,
		0.1943,
		0.093895,
		0.025256,
		0.013263,
		0.050639,
		0.108382,
		0.182595,
		0.269249,
		0.364225,
		0.463351,
		0.56247,
		0.657532,
		0.744711,
		0.820585,
		0.882368,
		0.928207,
		0.95755,
		0.971493,
		0.97297,
		0.966371,
		0.955943,
		0.942488,
		0.921848,
		0.887842,
		0.835202,
		0.761188,
		0.666205,
		0.553828,
		0.430498,
		0.305002,
		0.187825,
		0.090387,
		0.024201,
		0.006111,
		0.036478,
		0.088879,
		0.159331,
		0.243699,
		0.337737,
		0.437117,
		0.537502,
		0.634625,
		0.724419,
		0.803188,
		0.867848,
		0.916238,
		0.947489,
		0.96241,
		0.963724,
		0.955795,
		0.943208,
		0.927549,
		0.905228,
		0.870185,
		0.817112,
		0.743241,
		0.649019,
		0.538097,
		0.416966,
		0.294361,
		0.180546,
		0.086499,
		0.023047,
		0.001584,
		0.024222,
		0.07064,
		0.136765,
		0.218359,
		0.311037,
		0.410314,
		0.51166,
		0.610585,
		0.702768,
		0.784235,
		0.8516,
		0.902391,
		0.935426,
		0.951225,
		0.95229,
		0.942932,
		0.928019,
		0.909906,
		0.885676,
		0.849464,
		0.795954,
		0.722367,
		0.629185,
		0.520121,
		0.401677,
		0.282483,
		0.172521,
		0.082265,
		0.021807,
		0.014139,
		0.053884,
		0.115075,
		0.19336,
		0.284216,
		0.382989,
		0.484947,
		0.585368,
		0.679669,
		0.76359,
		0.833445,
		0.886448,
		0.921114,
		0.937675,
		0.938409,
		0.927549,
		0.910193,
		0.889428,
		0.863107,
		0.825636,
		0.771729,
		0.698606,
		0.606781,
		0.5,
		0.384738,
		0.269468,
		0.163827,
		0.077731,
		0.020493,
		0.006522,
		0.038856,
		0.094456,
		0.168853,
		0.257379,
		0.355202,
		0.457378,
		0.558942,
		0.655041,
		0.741127,
		0.81321,
		0.868202,
		0.904319,
		0.921517,
		0.921848,
		0.909445,
		0.889587,
		0.866039,
		0.837497,
		0.798725,
		0.744506,
		0.672067,
		0.581942,
		0.477883,
		0.366296,
		0.255441,
		0.154555,
		0.072946,
		0.019122,
		0.001693,
		0.025825,
		0.075127,
		0.14501,
		0.230652,
		0.327031,
		0.428982,
		0.531288,
		0.628819,
		0.716734,
		0.790745,
		0.847468,
		0.884835,
		0.902538,
		0.902408,
		0.888462,
		0.866109,
		0.839721,
		0.808892,
		0.768831,
		0.714427,
		0.642924,
		0.554862,
		0.453969,
		0.346533,
		0.240551,
		0.144809,
		0.067966,
		0.017708,
		0.015085,
		0.057334,
		0.122028,
		0.204182,
		0.298578,
		0.399815,
		0.502414,
		0.600964,
		0.690328,
		0.765928,
		0.824096,
		0.862495,
		0.88057,
		0.879941,
		0.864494,
		0.839721,
		0.81052,
		0.777408,
		0.736123,
		0.681705,
		0.611418,
		0.525792,
		0.428498,
		0.325662,
		0.224966,
		0.134701,
		0.06285,
		0.01627,
		0.006962,
		0.041348,
		0.100129,
		0.178148,
		0.269972,
		0.369959,
		0.472358,
		0.571468,
		0.661863,
		0.73868,
		0.797981,
		0.837182,
		0.855502,
		0.854357,
		0.837497,
		0.810449,
		0.778552,
		0.743232,
		0.700845,
		0.646618,
		0.577846,
		0.495029,
		0.401745,
		0.303917,
		0.208867,
		0.124352,
		0.057659,
		0.014824,
		0.001807,
		0.027471,
		0.079568,
		0.152754,
		0.241376,
		0.339533,
		0.441194,
		0.540369,
		0.63134,
		0.708972,
		0.769076,
		0.808844,
		0.827286,
		0.825636,
		0.807497,
		0.778389,
		0.743999,
		0.706621,
		0.663304,
		0.609506,
		0.542556,
		0.462908,
		0.374015,
		0.281552,
		0.192443,
		0.113883,
		0.052454,
		0.013387,
		0.016032,
		0.060626,
		0.128242,
		0.212988,
		0.308696,
		0.409045,
		0.507751,
		0.598814,
		0.676837,
		0.737402,
		0.777497,
		0.795954,
		0.793834,
		0.774593,
		0.743703,
		0.707107,
		0.667894,
		0.623865,
		0.570757,
		0.505937,
		0.429797,
		0.34563,
		0.25883,
		0.175889,
		0.103418,
		0.047295,
		0.011975,
		0.007387,
		0.043618,
		0.104887,
		0.185047,
		0.277651,
		0.37608,
		0.473757,
		0.564404,
		0.642378,
		0.703054,
		0.743241,
		0.761617,
		0.759087,
		0.738959,
		0.706621,
		0.668181,
		0.627422,
		0.582942,
		0.530799,
		0.468408,
		0.39608,
		0.316924,
		0.236018,
		0.159395,
		0.093074,
		0.042238,
		0.010604,
		0.001912,
		0.028884,
		0.082994,
		0.157829,
		0.246645,
		0.342523,
		0.438588,
		0.528295,
		0.605771,
		0.666205,
		0.706256,
		0.724468,
		0.721612,
		0.700845,
		0.667436,
		0.627576,
		0.585621,
		0.540983,
		0.490085,
		0.430401,
		0.36215,
		0.288229,
		0.213377,
		0.143149,
		0.082967,
		0.037339,
		0.009287,
		0.016784,
		0.062899,
		0.131646,
		0.215973,
		0.308649,
		0.402506,
		0.49074,
		0.567267,
		0.627109,
		0.666804,
		0.684786,
		0.681705,
		0.660568,
		0.626498,
		0.585686,
		0.542935,
		0.498456,
		0.449081,
		0.392355,
		0.328396,
		0.259871,
		0.191158,
		0.127325,
		0.073201,
		0.032644,
		0.008036,
		0.007692,
		0.04496,
		0.106843,
		0.185967,
		0.274783,
		0.365832,
		0.45206,
		0.527188,
		0.586099,
		0.625228,
		0.642924,
		0.639732,
		0.618508,
		0.584203,
		0.542935,
		0.499827,
		0.455837,
		0.408251,
		0.354699,
		0.295193,
		0.232157,
		0.169596,
		0.112084,
		0.063871,
		0.028199,
		0.006863,
		0.001978,
		0.029542,
		0.083785,
		0.156993,
		0.241294,
		0.328938,
		0.412633,
		0.485925,
		0.543575,
		0.581942,
		0.599306,
		0.596121,
		0.575095,
		0.540983,
		0.499767,
		0.456759,
		0.413595,
		0.368043,
		0.317842,
		0.262895,
		0.205375,
		0.148905,
		0.097571,
		0.05506,
		0.02404,
		0.005777,
		0.017013,
		0.062848,
		0.12944,
		0.208583,
		0.29224,
		0.372893,
		0.443925,
		0.5,
		0.537419,
		0.554413,
		0.551352,
		0.530799,
		0.497293,
		0.456627,
		0.414184,
		0.372177,
		0.328883,
		0.282165,
		0.231825,
		0.179781,
		0.129275,
		0.08391,
		0.046839,
		0.020197,
		0.004784,
		0.007716,
		0.044403,
		0.103703,
		0.177071,
		0.256183,
		0.333305,
		0.401677,
		0.455881,
		0.492178,
		0.508764,
		0.505937,
		0.486112,
		0.4536,
		0.413956,
		0.372533,
		0.332002,
		0.291157,
		0.248009,
		0.202269,
		0.1556,
		0.110866,
		0.071206,
		0.039266,
		0.016693,
		0.00389,
		0.001962,
		0.028801,
		0.080172,
		0.147183,
		0.221226,
		0.294361,
		0.359697,
		0.411755,
		0.446768,
		0.462908,
		0.460409,
		0.441538,
		0.410368,
		0.372177,
		0.332201,
		0.293444,
		0.255208,
		0.215674,
		0.174471,
		0.133017,
		0.09381,
		0.05954,
		0.032382,
		0.013546,
		0.003097,
		0.01635,
		0.05921,
		0.119331,
		0.187825,
		0.256554,
		0.318512,
		0.368169,
		0.401745,
		0.417399,
		0.415302,
		0.397574,
		0.368043,
		0.331687,
		0.293546,
		0.256834,
		0.22133,
		0.185408,
		0.148632,
		0.112184,
		0.078208,
		0.048973,
		0.026217,
		0.010764,
		0.002407,
		0.007301,
		0.041135,
		0.093895,
		0.156413,
		0.220363,
		0.278635,
		0.325662,
		0.35766,
		0.37278,
		0.371134,
		0.354699,
		0.32705,
		0.292845,
		0.256874,
		0.222444,
		0.18976,
		0.15741,
		0.124906,
		0.09321,
		0.064131,
		0.039541,
		0.020783,
		0.008349,
		0.001819,
		0.001825,
		0.026203,
		0.071204,
		0.127383,
		0.186232,
		0.240551,
		0.284745,
		0.315035,
		0.329566,
		0.328396,
		0.313356,
		0.287772,
		0.255969,
		0.222444,
		0.190493,
		0.160683,
		0.131824,
		0.103399,
		0.076164,
		0.051615,
		0.031258,
		0.016081,
		0.006296,
		0.001329,
		0.014591,
		0.051521,
		0.101068,
		0.154555,
		0.204699,
		0.245885,
		0.274347,
		0.288229,
		0.287533,
		0.273949,
		0.25055,
		0.22133,
		0.19046,
		0.161142,
		0.134224,
		0.108744,
		0.084171,
		0.061078,
		0.040669,
		0.024116,
		0.012095,
		0.004591,
		0.000933,
		0.006383,
		0.035026,
		0.077731,
		0.125657,
		0.171451,
		0.209488,
		0.236018,
		0.249185,
		0.248937,
		0.236823,
		0.215674,
		0.189147,
		0.161075,
		0.134495,
		0.110452,
		0.088209,
		0.067237,
		0.047945,
		0.031269,
		0.018083,
		0.008796,
		0.003214,
		0.000623,
		0.001561,
		0.021807,
		0.057546,
		0.099785,
		0.141108,
		0.175889,
		0.200397,
		0.212781,
		0.212934,
		0.202269,
		0.183376,
		0.159588,
		0.134389,
		0.110603,
		0.089387,
		0.070213,
		0.05257,
		0.036721,
		0.023365,
		0.01311,
		0.006142,
		0.002141,
		0.00039,
		0.011854,
		0.040599,
		0.077097,
		0.113883,
		0.145338,
		0.167757,
		0.17929,
		0.179781,
		0.170509,
		0.153832,
		0.132767,
		0.110452,
		0.089464,
		0.070995,
		0.054701,
		0.040099,
		0.027329,
		0.016877,
		0.009125,
		0.00408,
		0.001339,
		0.000225,
		0.005056,
		0.026875,
		0.057659,
		0.089903,
		0.118003,
		0.138285,
		0.148905,
		0.149661,
		0.141698,
		0.127154,
		0.108744,
		0.089266,
		0.071031,
		0.055198,
		0.041577,
		0.029718,
		0.019659,
		0.011703,
		0.006041,
		0.002543,
		0.00077,
		0.000116,
		0.001204,
		0.01627,
		0.041445,
		0.069201,
		0.09396,
		0.112084,
		0.121741,
		0.122681,
		0.115924,
		0.103399,
		0.087528,
		0.070788,
		0.055211,
		0.041877,
		0.030705,
		0.021286,
		0.013573,
		0.007717,
		0.003753,
		0.00146,
		0.000396,
		5.1e-005,
		0.008587,
		0.02834,
		0.051724,
		0.073201,
		0.089175,
		0.097831,
		0.098875,
		0.09321,
		0.082561,
		0.069081,
		0.054932,
		0.041877,
		0.030876,
		0.021916,
		0.014633,
		0.00891,
		0.004775,
		0.002146,
		0.000748,
		0.000173,
		1.7e-005,
		0.00355,
		0.018148,
		0.037339,
		0.055636,
		0.0695,
		0.077134,
		0.078208,
		0.073513,
		0.064584,
		0.053314,
		0.041577,
		0.030868,
		0.022008,
		0.015015,
		0.009563,
		0.005488,
		0.002719,
		0.001096,
		0.000326,
		5.8e-005,
		0.000818,
		0.010604,
		0.025838,
		0.041104,
		0.052927,
		0.05954,
		0.06058,
		0.056733,
		0.049356,
		0.040099,
		0.030563,
		0.021994,
		0.015059,
		0.00978,
		0.005863,
		0.00311,
		0.001382,
		0.000475,
		0.000109,
		1.2e-005,
		0.005387,
		0.016957,
		0.029382,
		0.039266,
		0.044879,
		0.045834,
		0.042718,
		0.036721,
		0.029267,
		0.021703,
		0.015042,
		0.0098,
		0.005977,
		0.003307,
		0.001573,
		0.000596,
		0.000159,
		2.3e-005,
		0.002137,
		0.010384,
		0.020197,
		0.028272,
		0.032928,
		0.033761,
		0.031269,
		0.026482,
		0.020615,
		0.014781,
		0.00978,
		0.005985,
		0.003361,
		0.001665,
		0.000676,
		0.000199,
		3.3e-005,
		0.000471,
		0.005777,
		0.013239,
		0.019662,
		0.023426,
		0.024116,
		0.02215,
		0.018409,
		0.013913,
		0.009563,
		0.005965,
		0.003364,
		0.001688,
		0.000712,
		0.000224,
		4.1e-005,
		0.00278,
		0.008178,
		0.013129,
		0.016081,
		0.016622,
		0.015095,
		0.012245,
		0.00891,
		0.005799,
		0.003347,
		0.001688,
		0.000719,
		0.000235,
		4.6e-005,
		0.001038,
		0.004677,
		0.008349,
		0.010587,
		0.010985,
		0.009823,
		0.007717,
		0.005342,
		0.003231,
		0.001676,
		0.000719,
		0.000237,
		4.8e-005,
		0.000214,
		0.002407,
		0.005004,
		0.006631,
		0.006903,
		0.006041,
		0.004545,
		0.002939,
		0.001605,
		0.000712,
		0.000237,
		4.9e-005,
		0.001059,
		0.002784,
		0.003911,
		0.00408,
		0.003462,
		0.002452,
		0.00144,
		0.000676,
		0.000233,
		4.9e-005,
		0.000356,
		0.001406,
		0.002141,
		0.002233,
		0.001811,
		0.001175,
		0.000596,
		0.000219,
		4.8e-005,
		6.5e-005,
		0.000623,
		0.001066,
		0.001107,
		0.000838,
		0.000475,
		0.00019,
		4.4e-005,
		0.000228,
		0.00047,
		0.000481,
		0.000326,
		0.000147,
		3.8e-005,
		6.1e-005,
		0.000175,
		0.000173,
		9.6e-005,
		2.8e-005,
		5.1e-005,
		4.7e-005,
		1.8e-005,
		1e-005,
		1e-005,
		1.8e-005,
		4.7e-005,
		5.1e-005,
		2.8e-005,
		9.6e-005,
		0.000173,
		0.000175,
		6.1e-005,
		3.8e-005,
		0.000147,
		0.000326,
		0.000481,
		0.00047,
		0.000228,
		4.4e-005,
		0.00019,
		0.000475,
		0.000838,
		0.001107,
		0.001066,
		0.000623,
		6.5e-005,
		4.8e-005,
		0.000219,
		0.000596,
		0.001175,
		0.001811,
		0.002233,
		0.002141,
		0.001406,
		0.000356,
		4.9e-005,
		0.000233,
		0.000676,
		0.00144,
		0.002452,
		0.003462,
		0.00408,
		0.003911,
		0.002784,
		0.001059,
		4.9e-005,
		0.000237,
		0.000712,
		0.001605,
		0.002939,
		0.004545,
		0.006041,
		0.006903,
		0.006631,
		0.005004,
		0.002407,
		0.000214,
		4.8e-005,
		0.000237,
		0.000719,
		0.001676,
		0.003231,
		0.005342,
		0.007717,
		0.009823,
		0.010985,
		0.010587,
		0.008349,
		0.004677,
		0.001038,
		4.6e-005,
		0.000235,
		0.000719,
		0.001688,
		0.003347,
		0.005799,
		0.00891,
		0.012245,
		0.015095,
		0.016622,
		0.016081,
		0.013129,
		0.008178,
		0.00278,
		4.1e-005,
		0.000224,
		0.000712,
		0.001688,
		0.003364,
		0.005965,
		0.009563,
		0.013913,
		0.018409,
		0.02215,
		0.024116,
		0.023426,
		0.019662,
		0.013239,
		0.005777,
		0.000471,
		3.3e-005,
		0.000199,
		0.000676,
		0.001665,
		0.003361,
		0.005985,
		0.00978,
		0.014781,
		0.020615,
		0.026482,
		0.031269,
		0.033761,
		0.032928,
		0.028272,
		0.020197,
		0.010384,
		0.002137,
		2.3e-005,
		0.000159,
		0.000596,
		0.001573,
		0.003307,
		0.005977,
		0.0098,
		0.015042,
		0.021703,
		0.029267,
		0.036721,
		0.042718,
		0.045834,
		0.044879,
		0.039266,
		0.029382,
		0.016957,
		0.005387,
		1.2e-005,
		0.000109,
		0.000475,
		0.001382,
		0.00311,
		0.005863,
		0.00978,
		0.015059,
		0.021994,
		0.030563,
		0.040099,
		0.049356,
		0.056733,
		0.06058,
		0.05954,
		0.052927,
		0.041104,
		0.025838,
		0.010604,
		0.000818,
		5.8e-005,
		0.000326,
		0.001096,
		0.002719,
		0.005488,
		0.009563,
		0.015015,
		0.022008,
		0.030868,
		0.041577,
		0.053314,
		0.064584,
		0.073513,
		0.078208,
		0.077134,
		0.0695,
		0.055636,
		0.037339,
		0.018148,
		0.00355,
		1.7e-005,
		0.000173,
		0.000748,
		0.002146,
		0.004775,
		0.00891,
		0.014633,
		0.021916,
		0.030876,
		0.041877,
		0.054932,
		0.069081,
		0.082561,
		0.09321,
		0.098875,
		0.097831,
		0.089175,
		0.073201,
		0.051724,
		0.02834,
		0.008587,
		5.1e-005,
		0.000396,
		0.00146,
		0.003753,
		0.007717,
		0.013573,
		0.021286,
		0.030705,
		0.041877,
		0.055211,
		0.070788,
		0.087528,
		0.103399,
		0.115924,
		0.122681,
		0.121741,
		0.112084,
		0.09396,
		0.069201,
		0.041445,
		0.01627,
		0.001204,
		0.000116,
		0.00077,
		0.002543,
		0.006041,
		0.011703,
		0.019659,
		0.029718,
		0.041577,
		0.055198,
		0.071031,
		0.089266,
		0.108744,
		0.127154,
		0.141698,
		0.149661,
		0.148905,
		0.138285,
		0.118003,
		0.089903,
		0.057659,
		0.026875,
		0.005056,
		0.000225,
		0.001339,
		0.00408,
		0.009125,
		0.016877,
		0.027329,
		0.040099,
		0.054701,
		0.070995,
		0.089464,
		0.110452,
		0.132767,
		0.153832,
		0.170509,
		0.179781,
		0.17929,
		0.167757,
		0.145338,
		0.113883,
		0.077097,
		0.040599,
		0.011854,
		0.00039,
		0.002141,
		0.006142,
		0.01311,
		0.023365,
		0.036721,
		0.05257,
		0.070213,
		0.089387,
		0.110603,
		0.134389,
		0.159588,
		0.183376,
		0.202269,
		0.212934,
		0.212781,
		0.200397,
		0.175889,
		0.141108,
		0.099785,
		0.057546,
		0.021807,
		0.001561,
		0.000623,
		0.003214,
		0.008796,
		0.018083,
		0.031269,
		0.047945,
		0.067237,
		0.088209,
		0.110452,
		0.134495,
		0.161075,
		0.189147,
		0.215674,
		0.236823,
		0.248937,
		0.249185,
		0.236018,
		0.209488,
		0.171451,
		0.125657,
		0.077731,
		0.035026,
		0.006383,
		0.000933,
		0.004591,
		0.012095,
		0.024116,
		0.040669,
		0.061078,
		0.084171,
		0.108744,
		0.134224,
		0.161142,
		0.19046,
		0.22133,
		0.25055,
		0.273949,
		0.287533,
		0.288229,
		0.274347,
		0.245885,
		0.204699,
		0.154555,
		0.101068,
		0.051521,
		0.014591,
		0.001329,
		0.006296,
		0.016081,
		0.031258,
		0.051615,
		0.076164,
		0.103399,
		0.131824,
		0.160683,
		0.190493,
		0.222444,
		0.255969,
		0.287772,
		0.313356,
		0.328396,
		0.329566,
		0.315035,
		0.284745,
		0.240551,
		0.186232,
		0.127383,
		0.071204,
		0.026203,
		0.001825,
		0.001819,
		0.008349,
		0.020783,
		0.039541,
		0.064131,
		0.09321,
		0.124906,
		0.15741,
		0.18976,
		0.222444,
		0.256874,
		0.292845,
		0.32705,
		0.354699,
		0.371134,
		0.37278,
		0.35766,
		0.325662,
		0.278635,
		0.220363,
		0.156413,
		0.093895,
		0.041135,
		0.007301,
		0.002407,
		0.010764,
		0.026217,
		0.048973,
		0.078208,
		0.112184,
		0.148632,
		0.185408,
		0.22133,
		0.256834,
		0.293546,
		0.331687,
		0.368043,
		0.397574,
		0.415302,
		0.417399,
		0.401745,
		0.368169,
		0.318512,
		0.256554,
		0.187825,
		0.119331,
		0.05921,
		0.01635,
		0.003097,
		0.013546,
		0.032382,
		0.05954,
		0.09381,
		0.133017,
		0.174471,
		0.215674,
		0.255208,
		0.293444,
		0.332201,
		0.372177,
		0.410368,
		0.441538,
		0.460409,
		0.462908,
		0.446768,
		0.411755,
		0.359697,
		0.294361,
		0.221226,
		0.147183,
		0.080172,
		0.028801,
		0.001962,
		0.00389,
		0.016693,
		0.039266,
		0.071206,
		0.110866,
		0.1556,
		0.202269,
		0.248009,
		0.291157,
		0.332002,
		0.372533,
		0.413956,
		0.4536,
		0.486112,
		0.505937,
		0.508764,
		0.492178,
		0.455881,
		0.401677,
		0.333305,
		0.256183,
		0.177071,
		0.103703,
		0.044403,
		0.007716,
		0.004784,
		0.020197,
		0.046839,
		0.08391,
		0.129275,
		0.179781,
		0.231825,
		0.282165,
		0.328883,
		0.372177,
		0.414184,
		0.456627,
		0.497293,
		0.530799,
		0.551352,
		0.554413,
		0.537419,
		0.5,
		0.443925,
		0.372893,
		0.29224,
		0.208583,
		0.12944,
		0.062848,
		0.017013,
		0.005777,
		0.02404,
		0.05506,
		0.097571,
		0.148905,
		0.205375,
		0.262895,
		0.317842,
		0.368043,
		0.413595,
		0.456759,
		0.499767,
		0.540983,
		0.575095,
		0.596121,
		0.599306,
		0.581942,
		0.543575,
		0.485925,
		0.412633,
		0.328938,
		0.241294,
		0.156993,
		0.083785,
		0.029542,
		0.001978,
		0.006863,
		0.028199,
		0.063871,
		0.112084,
		0.169596,
		0.232157,
		0.295193,
		0.354699,
		0.408251,
		0.455837,
		0.499827,
		0.542935,
		0.584203,
		0.618508,
		0.639732,
		0.642924,
		0.625228,
		0.586099,
		0.527188,
		0.45206,
		0.365832,
		0.274783,
		0.185967,
		0.106843,
		0.04496,
		0.007692,
		0.008036,
		0.032644,
		0.073201,
		0.127325,
		0.191158,
		0.259871,
		0.328396,
		0.392355,
		0.449081,
		0.498456,
		0.542935,
		0.585686,
		0.626498,
		0.660568,
		0.681705,
		0.684786,
		0.666804,
		0.627109,
		0.567267,
		0.49074,
		0.402506,
		0.308649,
		0.215973,
		0.131646,
		0.062899,
		0.016784,
		0.009287,
		0.037339,
		0.082967,
		0.143149,
		0.213377,
		0.288229,
		0.36215,
		0.430401,
		0.490085,
		0.540983,
		0.585621,
		0.627576,
		0.667436,
		0.700845,
		0.721612,
		0.724468,
		0.706256,
		0.666205,
		0.605771,
		0.528295,
		0.438588,
		0.342523,
		0.246645,
		0.157829,
		0.082994,
		0.028884,
		0.001912,
		0.010604,
		0.042238,
		0.093074,
		0.159395,
		0.236018,
		0.316924,
		0.39608,
		0.468408,
		0.530799,
		0.582942,
		0.627422,
		0.668181,
		0.706621,
		0.738959,
		0.759087,
		0.761617,
		0.743241,
		0.703054,
		0.642378,
		0.564404,
		0.473757,
		0.37608,
		0.277651,
		0.185047,
		0.104887,
		0.043618,
		0.007387,
		0.011975,
		0.047295,
		0.103418,
		0.175889,
		0.25883,
		0.34563,
		0.429797,
		0.505937,
		0.570757,
		0.623865,
		0.667894,
		0.707107,
		0.743703,
		0.774593,
		0.793834,
		0.795954,
		0.777497,
		0.737402,
		0.676837,
		0.598814,
		0.507751,
		0.409045,
		0.308696,
		0.212988,
		0.128242,
		0.060626,
		0.016032,
		0.013387,
		0.052454,
		0.113883,
		0.192443,
		0.281552,
		0.374015,
		0.462908,
		0.542556,
		0.609506,
		0.663304,
		0.706621,
		0.743999,
		0.778389,
		0.807497,
		0.825636,
		0.827286,
		0.808844,
		0.769076,
		0.708972,
		0.63134,
		0.540369,
		0.441194,
		0.339533,
		0.241376,
		0.152754,
		0.079568,
		0.027471,
		0.001807,
		0.014824,
		0.057659,
		0.124352,
		0.208867,
		0.303917,
		0.401745,
		0.495029,
		0.577846,
		0.646618,
		0.700845,
		0.743232,
		0.778552,
		0.810449,
		0.837497,
		0.854357,
		0.855502,
		0.837182,
		0.797981,
		0.73868,
		0.661863,
		0.571468,
		0.472358,
		0.369959,
		0.269972,
		0.178148,
		0.100129,
		0.041348,
		0.006962,
		0.01627,
		0.06285,
		0.134701,
		0.224966,
		0.325662,
		0.428498,
		0.525792,
		0.611418,
		0.681705,
		0.736123,
		0.777408,
		0.81052,
		0.839721,
		0.864494,
		0.879941,
		0.88057,
		0.862495,
		0.824096,
		0.765928,
		0.690328,
		0.600964,
		0.502414,
		0.399815,
		0.298578,
		0.204182,
		0.122028,
		0.057334,
		0.015085,
		0.017708,
		0.067966,
		0.144809,
		0.240551,
		0.346533,
		0.453969,
		0.554862,
		0.642924,
		0.714427,
		0.768831,
		0.808892,
		0.839721,
		0.866109,
		0.888462,
		0.902408,
		0.902538,
		0.884835,
		0.847468,
		0.790745,
		0.716734,
		0.628819,
		0.531288,
		0.428982,
		0.327031,
		0.230652,
		0.14501,
		0.075127,
		0.025825,
		0.001693,
		0.019122,
		0.072946,
		0.154555,
		0.255441,
		0.366296,
		0.477883,
		0.581942,
		0.672067,
		0.744506,
		0.798725,
		0.837497,
		0.866039,
		0.889587,
		0.909445,
		0.921848,
		0.921517,
		0.904319,
		0.868202,
		0.81321,
		0.741127,
		0.655041,
		0.558942,
		0.457378,
		0.355202,
		0.257379,
		0.168853,
		0.094456,
		0.038856,
		0.006522,
		0.020493,
		0.077731,
		0.163827,
		0.269468,
		0.384738,
		0.5,
		0.606781,
		0.698606,
		0.771729,
		0.825636,
		0.863107,
		0.889428,
		0.910193,
		0.927549,
		0.938409,
		0.937675,
		0.921114,
		0.886448,
		0.833445,
		0.76359,
		0.679669,
		0.585368,
		0.484947,
		0.382989,
		0.284216,
		0.19336,
		0.115075,
		0.053884,
		0.014139,
		0.021807,
		0.082265,
		0.172521,
		0.282483,
		0.401677,
		0.520121,
		0.629185,
		0.722367,
		0.795954,
		0.849464,
		0.885676,
		0.909906,
		0.928019,
		0.942932,
		0.95229,
		0.951225,
		0.935426,
		0.902391,
		0.8516,
		0.784235,
		0.702768,
		0.610585,
		0.51166,
		0.410314,
		0.311037,
		0.218359,
		0.136765,
		0.07064,
		0.024222,
		0.001584,
		0.023047,
		0.086499,
		0.180546,
		0.294361,
		0.416966,
		0.538097,
		0.649019,
		0.743241,
		0.817112,
		0.870185,
		0.905228,
		0.927549,
		0.943208,
		0.955795,
		0.963724,
		0.96241,
		0.947489,
		0.916238,
		0.867848,
		0.803188,
		0.724419,
		0.634625,
		0.537502,
		0.437117,
		0.337737,
		0.243699,
		0.159331,
		0.088879,
		0.036478,
		0.006111,
		0.024201,
		0.090387,
		0.187825,
		0.305002,
		0.430498,
		0.553828,
		0.666205,
		0.761188,
		0.835202,
		0.887842,
		0.921848,
		0.942488,
		0.955943,
		0.966371,
		0.97297,
		0.971493,
		0.95755,
		0.928207,
		0.882368,
		0.820585,
		0.744711,
		0.657532,
		0.56247,
		0.463351,
		0.364225,
		0.269249,
		0.182595,
		0.108382,
		0.050639,
		0.013263,
		0.025256,
		0.093895,
		0.1943,
		0.314334,
		0.442204,
		0.567267,
		0.680727,
		0.77623,
		0.850289,
		0.902538,
		0.935677,
		0.954896,
		0.966433,
		0.974912,
		0.980301,
		0.978745,
		0.96586,
		0.938519,
		0.895339,
		0.836564,
		0.763735,
		0.679351,
		0.586569,
		0.488978,
		0.390425,
		0.294891,
		0.2064,
		0.128948,
		0.06646,
		0.022747,
		0.001486,
		0.026203,
		0.096994,
		0.199929,
		0.322316,
		0.45206,
		0.578414,
		0.692627,
		0.788449,
		0.862495,
		0.914431,
		0.946897,
		0.96498,
		0.974912,
		0.981678,
		0.985989,
		0.984431,
		0.972665,
		0.947385,
		0.906936,
		0.851254,
		0.781582,
		0.700132,
		0.609807,
		0.513969,
		0.416269,
		0.320521,
		0.230604,
		0.150396,
		0.083718,
		0.034297,
		0.005737,
		0.027034,
		0.099666,
		0.204691,
		0.328938,
		0.460076,
		0.58732,
		0.701997,
		0.797981,
		0.871991,
		0.923719,
		0.955728,
		0.97297,
		0.98162,
		0.986929,
		0.990301,
		0.988808,
		0.978195,
		0.955004,
		0.917318,
		0.864779,
		0.798332,
		0.719918,
		0.632191,
		0.538297,
		0.441699,
		0.346046,
		0.255081,
		0.172561,
		0.102209,
		0.047667,
		0.012464,
		0.027745,
		0.101903,
		0.208583,
		0.334215,
		0.466305,
		0.594078,
		0.708972,
		0.805002,
		0.878988,
		0.930634,
		0.96241,
		0.979108,
		0.986797,
		0.990914,
		0.993485,
		0.992111,
		0.982663,
		0.961557,
		0.926633,
		0.877246,
		0.81406,
		0.73875,
		0.653732,
		0.561943,
		0.466662,
		0.371384,
		0.279714,
		0.195295,
		0.121749,
		0.062632,
		0.021402,
		0.001396,
		0.028334,
		0.103703,
		0.211619,
		0.338191,
		0.470829,
		0.598814,
		0.713725,
		0.809723,
		0.883723,
		0.935426,
		0.967195,
		0.983637,
		0.990679,
		0.993865,
		0.99577,
		0.994554,
		0.986257,
		0.967203,
		0.935005,
		0.888751,
		0.82883,
		0.756663,
		0.674436,
		0.584885,
		0.491113,
		0.39646,
		0.304401,
		0.218463,
		0.142168,
		0.078987,
		0.032305,
		0.005396,
		0.028801,
		0.105077,
		0.213831,
		0.340931,
		0.473757,
		0.601687,
		0.716456,
		0.812377,
		0.886448,
		0.938357,
		0.970337,
		0.986795,
		0.993485,
		0.99599,
		0.997356,
		0.996322,
		0.989142,
		0.972079,
		0.942546,
		0.899376,
		0.842698,
		0.773688,
		0.694312,
		0.607109,
		0.515011,
		0.42121,
		0.32905,
		0.241943,
		0.163312,
		0.096544,
		0.044948,
		0.011736,
		0.029149,
		0.106041,
		0.215263,
		0.342523,
		0.475222,
		0.602874,
		0.717384,
		0.81321,
		0.887425,
		0.939685,
		0.972083,
		0.988808,
		0.995414,
		0.997473,
		0.998418,
		0.997575,
		0.991455,
		0.9763,
		0.949345,
		0.909189,
		0.85571,
		0.789852,
		0.713365,
		0.628601,
		0.538323,
		0.445578,
		0.353578,
		0.265625,
		0.185043,
		0.11513,
		0.059124,
		0.020173,
		0.001314,
		0.029384,
		0.10662,
		0.215973,
		0.343066,
		0.475373,
		0.602567,
		0.716734,
		0.812471,
		0.88691,
		0.939661,
		0.972665,
		0.98988,
		0.996642,
		0.998467,
		0.9991,
		0.998441,
		0.99331,
		0.979961,
		0.955481,
		0.918247,
		0.867908,
		0.805178,
		0.731605,
		0.649351,
		0.561021,
		0.469512,
		0.377913,
		0.289411,
		0.207232,
		0.134589,
		0.074644,
		0.030481,
		0.005084,
		0.029512,
		0.106843,
		0.216028,
		0.34267,
		0.474364,
		0.600964,
		0.714734,
		0.810403,
		0.885148,
		0.938519,
		0.972293,
		0.990193,
		0.997321,
		0.9991,
		0.999516,
		0.999026,
		0.994799,
		0.983139,
		0.961014,
		0.926602,
		0.879328,
		0.819691,
		0.749038,
		0.669354,
		0.583079,
		0.492972,
		0.40199,
		0.313213,
		0.229766,
		0.154777,
		0.091333,
		0.042454,
		0.011069,
		0.029542,
		0.106744,
		0.215498,
		0.341453,
		0.472358,
		0.59826,
		0.711604,
		0.807238,
		0.882368,
		0.936471,
		0.971156,
		0.989904,
		0.997575,
		0.999471,
		0.999756,
		0.999412,
		0.995994,
		0.985898,
		0.965998,
		0.934294,
		0.890001,
		0.833411,
		0.765677,
		0.688606,
		0.604481,
		0.515919,
		0.425753,
		0.336951,
		0.25254,
		0.175563,
		0.109032,
		0.055902,
		0.019046,
		0.001239,
		0.029484,
		0.10636,
		0.21446,
		0.339533,
		0.46951,
		0.594644,
		0.707552,
		0.803188,
		0.878774,
		0.933706,
		0.969415,
		0.989142,
		0.997502,
		0.999655,
		0.999887,
		0.999659,
		0.996952,
		0.988288,
		0.970479,
		0.941362,
		0.899959,
		0.846361,
		0.781532,
		0.707107,
		0.625212,
		0.538323,
		0.449152,
		0.360554,
		0.27546,
		0.196828,
		0.127594,
		0.070648,
		0.028807,
		0.004799,
		0.029348,
		0.105728,
		0.212988,
		0.337025,
		0.465973,
		0.590294,
		0.702768,
		0.798444,
		0.874547,
		0.930382,
		0.967203,
		0.988014,
		0.997181,
		0.999705,
		0.999952,
		0.999812,
		0.997716,
		0.990351,
		0.974494,
		0.947842,
		0.909232,
		0.858565,
		0.796619,
		0.72486,
		0.645261,
		0.560159,
		0.472143,
		0.383959,
		0.298442,
		0.218463,
		0.146884,
		0.086531,
		0.040162,
		0.010458,
		0.029144,
		0.104887,
		0.211158,
		0.334039,
		0.461886,
		0.585368,
		0.697422,
		0.793172,
		0.86984,
		0.926633,
		0.96463,
		0.986599,
		0.996665,
		0.999659,
		0.999982,
		0.999904,
		0.99832,
		0.992122,
		0.978079,
		0.953766,
		0.91785,
		0.870045,
		0.810953,
		0.741872,
		0.664622,
		0.581406,
		0.494689,
		0.407111,
		0.321409,
		0.24037,
		0.166781,
		0.103402,
		0.052935,
		0.018011,
		0.00117,
		0.028884,
		0.103873,
		0.209038,
		0.330677,
		0.457378,
		0.580011,
		0.691658,
		0.787513,
		0.864779,
		0.922566,
		0.961779,
		0.98496,
		0.995994,
		0.999536,
		0.999993,
		0.999955,
		0.998793,
		0.993633,
		0.981267,
		0.959167,
		0.925841,
		0.880827,
		0.824552,
		0.758149,
		0.683291,
		0.602046,
		0.516759,
		0.42996,
		0.344293,
		0.262459,
		0.187172,
		0.121124,
		0.066963,
		0.027267,
		0.004537,
		0.028577,
		0.102721,
		0.206694,
		0.327031,
		0.452559,
		0.574344,
		0.685602,
		0.781582,
		0.859466,
		0.918264,
		0.958714,
		0.983139,
		0.995193,
		0.999349,
		0.999993,
		0.999982,
		0.999156,
		0.994912,
		0.984089,
		0.964076,
		0.933236,
		0.890936,
		0.837433,
		0.773702,
		0.701268,
		0.622067,
		0.538323,
		0.452462,
		0.367032,
		0.28465,
		0.207954,
		0.139573,
		0.082095,
		0.038051,
		0.009896,
		0.028234,
		0.101464,
		0.204182,
		0.323181,
		0.447525,
		0.56847,
		0.679351,
		0.77547,
		0.853978,
		0.91379,
		0.955481,
		0.981167,
		0.994276,
		0.999102,
		0.999982,
		0.999994,
		0.999429,
		0.995985,
		0.986574,
		0.968525,
		0.940063,
		0.900395,
		0.849616,
		0.788542,
		0.718555,
		0.641458,
		0.559361,
		0.474581,
		0.389573,
		0.306869,
		0.229035,
		0.158632,
		0.098193,
		0.050198,
		0.017058,
		0.001107,
		0.027862,
		0.100129,
		0.201554,
		0.319197,
		0.442357,
		0.56247,
		0.672987,
		0.769247,
		0.848375,
		0.909189,
		0.952111,
		0.979061,
		0.993251,
		0.998793,
		0.999959,
		0.999999,
		0.999629,
		0.996876,
		0.988753,
		0.972543,
		0.946351,
		0.909232,
		0.861121,
		0.802682,
		0.735156,
		0.660215,
		0.579853,
		0.496282,
		0.411865,
		0.32905,
		0.250328,
		0.178197,
		0.11513,
		0.063557,
		0.025848,
		0.004296,
		0.027471,
		0.098742,
		0.198852,
		0.315133,
		0.437117,
		0.556411,
		0.666569,
		0.762966,
		0.842698,
		0.904491,
		0.948626,
		0.976834,
		0.992122,
		0.998419,
		0.99992,
		1,
		0.999771,
		0.997607,
		0.99065,
		0.976158,
		0.952128,
		0.917472,
		0.87197,
		0.816138,
		0.751077,
		0.678333,
		0.599783,
		0.517538,
		0.433865,
		0.351133,
		0.271757,
		0.19817,
		0.132788,
		0.07799,
		0.036101,
		0.009378,
		0.027066,
		0.097323,
		0.196111,
		0.311037,
		0.431857,
		0.550341,
		0.66014,
		0.756663,
		0.836975,
		0.89972,
		0.94504,
		0.974494,
		0.99089,
		0.997978,
		0.999859,
		1,
		0.999867,
		0.9982,
		0.992293,
		0.9794,
		0.957423,
		0.92514,
		0.882183,
		0.828924,
		0.766326,
		0.69581,
		0.61914,
		0.538323,
		0.455537,
		0.373065,
		0.293251,
		0.218463,
		0.151058,
		0.093366,
		0.047667,
		0.016179,
		0.001049,
		0.026654,
		0.09589,
		0.19336,
		0.306942,
		0.426612,
		0.544295,
		0.653732,
		0.750362,
		0.831226,
		0.894886,
		0.941362,
		0.972044,
		0.989554,
		0.997465,
		0.999771,
		1,
		0.999929,
		0.998674,
		0.993706,
		0.982294,
		0.962262,
		0.932261,
		0.891784,
		0.841057,
		0.780913,
		0.712649,
		0.637915,
		0.558619,
		0.476846,
		0.394798,
		0.314747,
		0.238994,
		0.169842,
		0.109566,
		0.060404,
		0.024536,
		0.004073,
		0.026239,
		0.094456,
		0.190619,
		0.302876,
		0.421411,
		0.538297,
		0.647364,
		0.74408,
		0.825462,
		0.890001,
		0.937598,
		0.969489,
		0.988116,
		0.996876,
		0.999651,
		0.999999,
		0.999966,
		0.999046,
		0.994912,
		0.984868,
		0.966673,
		0.938862,
		0.900792,
		0.852555,
		0.794849,
		0.728851,
		0.656102,
		0.578407,
		0.497762,
		0.416289,
		0.336187,
		0.259691,
		0.189047,
		0.12648,
		0.074184,
		0.034297,
		0.0089,
		0.025825,
		0.093032,
		0.187904,
		0.298855,
		0.416269,
		0.532363,
		0.641049,
		0.737825,
		0.819691,
		0.88507,
		0.933754,
		0.966832,
		0.986574,
		0.996208,
		0.999493,
		0.999995,
		0.999986,
		0.999333,
		0.995933,
		0.987147,
		0.970681,
		0.944967,
		0.909232,
		0.863435,
		0.808146,
		0.744423,
		0.673697,
		0.597675,
		0.518262,
		0.437501,
		0.357518,
		0.280485,
		0.208589,
		0.144007,
		0.088884,
		0.045322,
		0.015366,
		0.000995,
		0.025414,
		0.091624,
		0.185227,
		0.294891,
		0.411199,
		0.526502,
		0.634793,
		0.731605,
		0.813918,
		0.880097,
		0.929834,
		0.964076,
		0.984931,
		0.995458,
		0.999293,
		0.999986,
		0.999996,
		0.999548,
		0.99679,
		0.989154,
		0.974313,
		0.950601,
		0.917126,
		0.873716,
		0.820817,
		0.759371,
		0.690698,
		0.61641,
		0.538323,
		0.4584,
		0.378694,
		0.301315,
		0.228392,
		0.162051,
		0.104393,
		0.057479,
		0.023321,
		0.003868,
		0.02501,
		0.090239,
		0.182595,
		0.290993,
		0.406206,
		0.520717,
		0.628601,
		0.72542,
		0.808146,
		0.875088,
		0.925841,
		0.961225,
		0.983187,
		0.994625,
		0.999046,
		0.999968,
		0.999999,
		0.999707,
		0.997501,
		0.990912,
		0.977591,
		0.955789,
		0.924495,
		0.883417,
		0.832877,
		0.773702,
		0.707107,
		0.634605,
		0.557927,
		0.478955,
		0.399673,
		0.322126,
		0.248386,
		0.180527,
		0.120607,
		0.070648,
		0.032625,
		0.008457,
		0.024612,
		0.088879,
		0.180011,
		0.287164,
		0.401293,
		0.515011,
		0.622473,
		0.719275,
		0.802379,
		0.870045,
		0.921781,
		0.95828,
		0.981342,
		0.993706,
		0.998747,
		0.999937,
		1,
		0.999819,
		0.998086,
		0.992444,
		0.980542,
		0.960553,
		0.931363,
		0.892557,
		0.84434,
		0.787425,
		0.722924,
		0.652255,
		0.577059,
		0.499142,
		0.420418,
		0.342867,
		0.268504,
		0.199356,
		0.137432,
		0.084714,
		0.043145,
		0.014613,
		0.000945,
		0.024222,
		0.087547,
		0.177477,
		0.283404,
		0.39646,
		0.509384,
		0.61641,
		0.71317,
		0.796619,
		0.864974,
		0.917657,
		0.955247,
		0.9794,
		0.9927,
		0.998392,
		0.999889,
		1,
		0.999895,
		0.99856,
		0.993771,
		0.983187,
		0.964919,
		0.937751,
		0.901155,
		0.855222,
		0.800552,
		0.738155,
		0.669354,
		0.595706,
		0.518938,
		0.440894,
		0.363491,
		0.28869,
		0.218463,
		0.154777,
		0.099574,
		0.054761,
		0.022194,
		0.003677,
		0.023841,
		0.086243,
		0.174994,
		0.279714,
		0.391707,
		0.503836,
		0.610413,
		0.707107,
		0.790869,
		0.859876,
		0.913472,
		0.952128,
		0.97736,
		0.991607,
		0.997978,
		0.999819,
		1,
		0.999944,
		0.998938,
		0.994912,
		0.985548,
		0.968908,
		0.943681,
		0.909232,
		0.865538,
		0.813093,
		0.752804,
		0.685901,
		0.613858,
		0.538323,
		0.461072,
		0.383959,
		0.308888,
		0.237781,
		0.172561,
		0.11513,
		0.067358,
		0.031071,
		0.008046,
		0.023468,
		0.084966,
		0.172561,
		0.276092,
		0.387032,
		0.498366,
		0.604481,
		0.701086,
		0.785132,
		0.854755,
		0.909232,
		0.948926,
		0.975226,
		0.990425,
		0.997501,
		0.999723,
		0.999999,
		0.999973,
		0.999236,
		0.995886,
		0.987648,
		0.972543,
		0.949176,
		0.916807,
		0.875306,
		0.82506,
		0.766878,
		0.701897,
		0.631507,
		0.557281,
		0.480925,
		0.404232,
		0.32905,
		0.25725,
		0.19071,
		0.131291,
		0.08083,
		0.041121,
		0.013913,
		0.000899,
		0.023103,
		0.083718,
		0.170178,
		0.272538,
		0.382435,
		0.492972,
		0.598615,
		0.695109,
		0.779408,
		0.849616,
		0.90494,
		0.945646,
		0.972999,
		0.989154,
		0.996959,
		0.999596,
		0.999996,
		0.999989,
		0.999465,
		0.996711,
		0.989507,
		0.975845,
		0.954256,
		0.9239,
		0.884543,
		0.836467,
		0.780384,
		0.717343,
		0.648648,
		0.575799,
		0.500431,
		0.424278,
		0.349131,
		0.276811,
		0.209152,
		0.147975,
		0.09508,
		0.05223,
		0.021147,
		0.003501,
		0.022747,
		0.082496,
		0.167843,
		0.269049,
		0.377913,
		0.487653,
		0.592813,
		0.689177,
		0.773702,
		0.84446,
		0.900598,
		0.94229,
		0.970681,
		0.987794,
		0.996348,
		0.999435,
		0.999989,
		0.999997,
		0.999638,
		0.997403,
		0.991144,
		0.978837,
		0.958943,
		0.930531,
		0.893264,
		0.847326,
		0.793332,
		0.732241,
		0.665277,
		0.593863,
		0.51957,
		0.444067,
		0.369091,
		0.296414,
		0.227824,
		0.165103,
		0.110014,
		0.064291,
		0.029626,
		0.007665,
		0.022399,
		0.081301,
		0.165555,
		0.265625,
		0.373465,
		0.48241,
		0.587077,
		0.683291,
		0.768013,
		0.839291,
		0.896212,
		0.938862,
		0.968276,
		0.986346,
		0.995666,
		0.999236,
		0.999975,
		0.999999,
		0.999765,
		0.997978,
		0.992579,
		0.981538,
		0.963257,
		0.936718,
		0.901488,
		0.857652,
		0.805731,
		0.746597,
		0.681392,
		0.611467,
		0.538323,
		0.463572,
		0.388893,
		0.31601,
		0.246666,
		0.182603,
		0.12555,
		0.077206,
		0.039236,
		0.013263,
		0.000857,
		0.022059,
		0.080132,
		0.163312,
		0.262263,
		0.369091,
		0.47724,
		0.581406,
		0.677451,
		0.762345,
		0.834112,
		0.891784,
		0.935365,
		0.965784,
		0.98481,
		0.994912,
		0.998994,
		0.99995,
		1,
		0.999854,
		0.998451,
		0.993829,
		0.983968,
		0.967218,
		0.942482,
		0.909232,
		0.867459,
		0.817592,
		0.760415,
		0.696992,
		0.628601,
		0.556677,
		0.482769,
		0.408505,
		0.335557,
		0.265625,
		0.200411,
		0.141606,
		0.09088,
		0.049871,
		0.020173,
		0.003336,
		0.021727,
		0.078987,
		0.161115,
		0.258963,
		0.364787,
		0.472143,
		0.575799,
		0.671658,
		0.756699,
		0.828924,
		0.887317,
		0.931803,
		0.963209,
		0.983187,
		0.994084,
		0.998706,
		0.999911,
		1,
		0.999916,
		0.998834,
		0.994912,
		0.986146,
		0.970847,
		0.947842,
		0.916514,
		0.876762,
		0.828924,
		0.773702,
		0.712079,
		0.645261,
		0.574618,
		0.501639,
		0.427896,
		0.355015,
		0.28465,
		0.218463,
		0.15811,
		0.105229,
		0.061429,
		0.02828,
		0.00731,
		0.021402,
		0.077867,
		0.15896,
		0.255722,
		0.360554,
		0.467118,
		0.570256,
		0.665912,
		0.751077,
		0.823732,
		0.882814,
		0.928179,
		0.960553,
		0.981477,
		0.993181,
		0.998368,
		0.999854,
		1,
		0.999955,
		0.99914,
		0.995843,
		0.988091,
		0.974162,
		0.952816,
		0.92335,
		0.885575,
		0.839741,
		0.786465,
		0.726654,
		0.661443,
		0.592135,
		0.520162,
		0.447041,
		0.374349,
		0.303696,
		0.236704,
		0.174994,
		0.120172,
		0.073818,
		0.037478,
		0.012657,
		0.000817,
		0.021084,
		0.07677,
		0.156848,
		0.25254,
		0.356389,
		0.462163,
		0.564777,
		0.660215,
		0.745479,
		0.818536,
		0.87828,
		0.924495,
		0.957819,
		0.979683,
		0.992201,
		0.997978,
		0.999777,
		0.999999,
		0.999978,
		0.999381,
		0.996638,
		0.989821,
		0.977183,
		0.957423,
		0.929759,
		0.893914,
		0.850054,
		0.798713,
		0.740722,
		0.677145,
		0.609221,
		0.538323,
		0.465916,
		0.393526,
		0.32272,
		0.255081,
		0.192195,
		0.135635,
		0.086952,
		0.047667,
		0.019264,
		0.003183,
		0.020774,
		0.075696,
		0.154777,
		0.249415,
		0.352291,
		0.457278,
		0.559361,
		0.654566,
		0.739908,
		0.81334,
		0.873716,
		0.920755,
		0.95501,
		0.977805,
		0.991144,
		0.997533,
		0.999675,
		0.999997,
		0.999991,
		0.999567,
		0.997312,
		0.991352,
		0.979927,
		0.961681,
		0.935757,
		0.901795,
		0.859876,
		0.810454,
		0.754286,
		0.692366,
		0.625869,
		0.556109,
		0.484498,
		0.412516,
		0.341684,
		0.273547,
		0.209654,
		0.151546,
		0.100747,
		0.058753,
		0.027023,
		0.006979,
		0.02047,
		0.074644,
		0.152746,
		0.246346,
		0.34826,
		0.452462,
		0.554008,
		0.648967,
		0.734365,
		0.808146,
		0.869125,
		0.916963,
		0.952128,
		0.975845,
		0.990011,
		0.99703,
		0.999544,
		0.999991,
		0.999997,
		0.999707,
		0.997877,
		0.9927,
		0.982413,
		0.965609,
		0.941362,
		0.909232,
		0.869219,
		0.821699,
		0.767353,
		0.707107,
		0.642073,
		0.573509,
		0.502771,
		0.431293,
		0.360554,
		0.292057,
		0.227317,
		0.167843,
		0.11513,
		0.070648,
		0.035835,
		0.012092,
		0.00078,
		0.020173,
		0.073614,
		0.150754,
		0.243331,
		0.344293,
		0.447714,
		0.548718,
		0.643416,
		0.728851,
		0.802956,
		0.864511,
		0.913121,
		0.949176,
		0.973805,
		0.9888,
		0.996468,
		0.999381,
		0.99998,
		0.999999,
		0.999809,
		0.998347,
		0.993882,
		0.984657,
		0.969224,
		0.946591,
		0.916242,
		0.878098,
		0.832457,
		0.779928,
		0.721369,
		0.657831,
		0.590511,
		0.520718,
		0.449834,
		0.379296,
		0.31057,
		0.245133,
		0.184465,
		0.130029,
		0.08327,
		0.045605,
		0.018415,
		0.003041,
		0.019882,
		0.072605,
		0.148801,
		0.24037,
		0.34039,
		0.443032,
		0.54349,
		0.637915,
		0.723368,
		0.797771,
		0.859876,
		0.909232,
		0.946156,
		0.971686,
		0.987511,
		0.995843,
		0.999183,
		0.999959,
		1,
		0.999882,
		0.998733,
		0.994912,
		0.986675,
		0.972543,
		0.951461,
		0.92284,
		0.886525,
		0.842738,
		0.792019,
		0.735156,
		0.673139,
		0.607109,
		0.538323,
		0.468117,
		0.397884,
		0.32905,
		0.263058,
		0.201355,
		0.145377,
		0.096544,
		0.056247,
		0.025848,
		0.006671,
		0.019597,
		0.071616,
		0.146884,
		0.237461,
		0.336549,
		0.438416,
		0.538323,
		0.632464,
		0.717915,
		0.792593,
		0.855222,
		0.905299,
		0.943071,
		0.969489,
		0.986146,
		0.995156,
		0.998946,
		0.999928,
		1,
		0.999931,
		0.999046,
		0.995804,
		0.988485,
		0.975582,
		0.955987,
		0.929041,
		0.894513,
		0.852555,
		0.803632,
		0.74847,
		0.687996,
		0.623297,
		0.555576,
		0.486123,
		0.416289,
		0.347461,
		0.281048,
		0.218463,
		0.161115,
		0.110396,
		0.067676,
		0.034297,
		0.011564,
		0.000745,
		0.019319,
		0.070648,
		0.145004,
		0.234603,
		0.332769,
		0.433865,
		0.533218,
		0.627063,
		0.712494,
		0.787425,
		0.850552,
		0.901325,
		0.939924,
		0.967218,
		0.984704,
		0.994403,
		0.998669,
		0.999882,
		1,
		0.999963,
		0.999296,
		0.996571,
		0.990101,
		0.978359,
		0.960187,
		0.93486,
		0.902078,
		0.861918,
		0.814777,
		0.761317,
		0.702404,
		0.639069,
		0.572465,
		0.503836,
		0.434489,
		0.365774,
		0.299063,
		0.23574,
		0.177183,
		0.12476,
		0.079817,
		0.043674,
		0.017621,
		0.002907,
		0.019046,
		0.069699,
		0.14316,
		0.231794,
		0.32905,
		0.429378,
		0.528173,
		0.621711,
		0.707107,
		0.782268,
		0.845868,
		0.897313,
		0.936718,
		0.964874,
		0.983187,
		0.993585,
		0.998347,
		0.999819,
		1,
		0.999982,
		0.999493,
		0.997227,
		0.991538,
		0.980889,
		0.964076,
		0.940314,
		0.909232,
		0.870838,
		0.825463,
		0.773702,
		0.716363,
		0.654422,
		0.588982,
		0.52124,
		0.452462,
		0.383959,
		0.317068,
		0.253142,
		0.193529,
		0.139573,
		0.092596,
		0.053897,
		0.024747,
		0.006382,
		0.018779,
		0.068768,
		0.141349,
		0.229035,
		0.325389,
		0.424954,
		0.523188,
		0.61641,
		0.701753,
		0.777124,
		0.841172,
		0.893264,
		0.933455,
		0.962458,
		0.981594,
		0.9927,
		0.997978,
		0.999735,
		0.999998,
		0.999993,
		0.999645,
		0.997782,
		0.992809,
		0.983187,
		0.967671,
		0.945417,
		0.91599,
		0.879329,
		0.835698,
		0.78563,
		0.729875,
		0.669354,
		0.605119,
		0.538323,
		0.470189,
		0.40199,
		0.33503,
		0.270627,
		0.210104,
		0.154777,
		0.105945,
		0.064888,
		0.032856,
		0.011069,
		0.000713,
		0.018518,
		0.067857,
		0.139573,
		0.226323,
		0.321785,
		0.420591,
		0.518262,
		0.611159,
		0.696434,
		0.771993,
		0.836467,
		0.889182,
		0.930138,
		0.959974,
		0.979927,
		0.991748,
		0.997561,
		0.999627,
		0.999993,
		0.999998,
		0.99976,
		0.998248,
		0.99393,
		0.985268,
		0.970985,
		0.950184,
		0.922365,
		0.887401,
		0.845492,
		0.797109,
		0.742944,
		0.683863,
		0.620871,
		0.555074,
		0.487653,
		0.419845,
		0.352917,
		0.288158,
		0.226862,
		0.170317,
		0.119801,
		0.076573,
		0.041863,
		0.016877,
		0.002783,
		0.018262,
		0.066963,
		0.137829,
		0.223658,
		0.318238,
		0.416289,
		0.513395,
		0.605958,
		0.69115,
		0.766878,
		0.831754,
		0.88507,
		0.926769,
		0.957423,
		0.978187,
		0.990728,
		0.997092,
		0.999493,
		0.999983,
		1,
		0.999844,
		0.998635,
		0.994912,
		0.987147,
		0.974035,
		0.95463,
		0.928371,
		0.895068,
		0.854856,
		0.808146,
		0.755573,
		0.69795,
		0.636233,
		0.571482,
		0.504839,
		0.437501,
		0.370703,
		0.3057,
		0.243759,
		0.186142,
		0.134105,
		0.088884,
		0.051692,
		0.023716,
		0.006111,
		0.018011,
		0.066086,
		0.136118,
		0.221038,
		0.314747,
		0.412048,
		0.508587,
		0.600807,
		0.685901,
		0.761779,
		0.827036,
		0.880929,
		0.92335,
		0.954807,
		0.976376,
		0.989641,
		0.996571,
		0.99933,
		0.999967,
		1,
		0.999903,
		0.998953,
		0.995768,
		0.988837,
		0.976835,
		0.958769,
		0.934022,
		0.902341,
		0.863799,
		0.818748,
		0.767766,
		0.711615,
		0.651201,
		0.58754,
		0.521733,
		0.45494,
		0.38836,
		0.32322,
		0.260756,
		0.202206,
		0.148801,
		0.101756,
		0.062267,
		0.031504,
		0.010606,
		0.000682,
		0.017765,
		0.065227,
		0.134438,
		0.218463,
		0.31131,
		0.407865,
		0.503836,
		0.595706,
		0.68069,
		0.756699,
		0.822315,
		0.876762,
		0.919885,
		0.952128,
		0.974494,
		0.988485,
		0.995996,
		0.999134,
		0.999941,
		1,
		0.999944,
		0.999212,
		0.996509,
		0.990352,
		0.9794,
		0.962616,
		0.939331,
		0.909232,
		0.872333,
		0.828924,
		0.77953,
		0.72486,
		0.665773,
		0.603241,
		0.538323,
		0.472143,
		0.405867,
		0.340688,
		0.277816,
		0.218463,
		0.163836,
		0.11513,
		0.073521,
		0.040162,
		0.016179,
		0.002666,
		0.017525,
		0.064384,
		0.132788,
		0.215931,
		0.307926,
		0.403741,
		0.499142,
		0.590655,
		0.675515,
		0.751638,
		0.817592,
		0.872571,
		0.916375,
		0.949389,
		0.972543,
		0.987261,
		0.995364,
		0.998903,
		0.999903,
		1,
		0.99997,
		0.999419,
		0.997147,
		0.991705,
		0.981742,
		0.966185,
		0.944313,
		0.915755,
		0.880467,
		0.838682,
		0.790869,
		0.737688,
		0.679948,
		0.618579,
		0.5546,
		0.489096,
		0.423201,
		0.358077,
		0.294906,
		0.234873,
		0.179163,
		0.128949,
		0.085389,
		0.049618,
		0.022747,
		0.005858,
		0.017289,
		0.063557,
		0.131168,
		0.213443,
		0.304595,
		0.399673,
		0.494505,
		0.585653,
		0.670377,
		0.746597,
		0.812868,
		0.868358,
		0.912824,
		0.946591,
		0.970524,
		0.98597,
		0.994676,
		0.998635,
		0.999851,
		1,
		0.999986,
		0.999581,
		0.997692,
		0.992908,
		0.983876,
		0.969489,
		0.94898,
		0.921922,
		0.888213,
		0.848031,
		0.801791,
		0.750101,
		0.693726,
		0.63355,
		0.570554,
		0.505785,
		0.440344,
		0.375363,
		0.311993,
		0.251397,
		0.194735,
		0.14316,
		0.097809,
		0.059802,
		0.030233,
		0.010171,
		0.000654,
		0.017058,
		0.062746,
		0.129577,
		0.210995,
		0.301315,
		0.395662,
		0.489923,
		0.580701,
		0.665277,
		0.741577,
		0.808146,
		0.864126,
		0.909232,
		0.943738,
		0.968439,
		0.984612,
		0.99393,
		0.998327,
		0.999782,
		0.999998,
		0.999994,
		0.999707,
		0.998153,
		0.993974,
		0.985814,
		0.972543,
		0.953345,
		0.927745,
		0.895582,
		0.85698,
		0.812301,
		0.762104,
		0.707107,
		0.648151,
		0.586177,
		0.522198,
		0.457278,
		0.392521,
		0.32905,
		0.268,
		0.210511,
		0.157713,
		0.110724,
		0.070648,
		0.038563,
		0.015523,
		0.002556,
		0.016832,
		0.061951,
		0.128015,
		0.208589,
		0.298086,
		0.391707,
		0.485397,
		0.575799,
		0.660215,
		0.736579,
		0.803427,
		0.859876,
		0.905603,
		0.94083,
		0.966289,
		0.983187,
		0.993125,
		0.997978,
		0.999693,
		0.999994,
		0.999998,
		0.999801,
		0.998541,
		0.994912,
		0.987569,
		0.975358,
		0.957423,
		0.933236,
		0.902584,
		0.865538,
		0.822408,
		0.773702,
		0.720092,
		0.66238,
		0.601465,
		0.538323,
		0.473988,
		0.409531,
		0.346048,
		0.28465,
		0.226451,
		0.172561,
		0.124081,
		0.082095,
		0.047667,
		0.021837,
		0.00562,
		0.01661,
		0.06117,
		0.12648,
		0.206223,
		0.294906,
		0.387806,
		0.480925,
		0.570945,
		0.655192,
		0.731605,
		0.798713,
		0.85561,
		0.901939,
		0.93787,
		0.964076,
		0.981696,
		0.992261,
		0.997585,
		0.999581,
		0.999986,
		1,
		0.999871,
		0.998863,
		0.995735,
		0.989154,
		0.977948,
		0.961225,
		0.938407,
		0.909232,
		0.873716,
		0.832117,
		0.784898,
		0.732684,
		0.676234,
		0.61641,
		0.554152,
		0.49046,
		0.426374,
		0.362964,
		0.301315,
		0.242518,
		0.187663,
		0.137829,
		0.094085,
		0.057479,
		0.029038,
		0.009762,
		0.000627,
		0.016392,
		0.060404,
		0.124973,
		0.203896,
		0.291774,
		0.383959,
		0.476508,
		0.56614,
		0.650207,
		0.726654,
		0.794004,
		0.851331,
		0.898242,
		0.93486,
		0.961803,
		0.98014,
		0.991337,
		0.997147,
		0.999445,
		0.999973,
		1,
		0.99992,
		0.999128,
		0.996452,
		0.990579,
		0.980326,
		0.964764,
		0.943272,
		0.915536,
		0.881522,
		0.841438,
		0.795699,
		0.744885,
		0.689715,
		0.63101,
		0.569676,
		0.50668,
		0.443032,
		0.379775,
		0.317968,
		0.258679,
		0.202976,
		0.151921,
		0.106564,
		0.067939,
		0.037057,
		0.014907,
		0.002453,
		0.016179,
		0.059653,
		0.123492,
		0.201607,
		0.28869,
		0.380165,
		0.472143,
		0.561384,
		0.645261,
		0.721729,
		0.789303,
		0.84704,
		0.894513,
		0.931803,
		0.959469,
		0.978519,
		0.990352,
		0.996662,
		0.999281,
		0.999951,
		1,
		0.999953,
		0.999343,
		0.997073,
		0.991857,
		0.982503,
		0.968053,
		0.947842,
		0.921508,
		0.888967,
		0.850378,
		0.80611,
		0.756699,
		0.702821,
		0.645261,
		0.584888,
		0.522638,
		0.45949,
		0.39646,
		0.334582,
		0.2749,
		0.218463,
		0.166312,
		0.119481,
		0.078987,
		0.045828,
		0.02098,
		0.005396,
		0.01597,
		0.058915,
		0.122037,
		0.199356,
		0.285652,
		0.376422,
		0.467831,
		0.556677,
		0.640354,
		0.716828,
		0.784611,
		0.842738,
		0.890756,
		0.9287,
		0.957078,
		0.976835,
		0.989308,
		0.996128,
		0.999088,
		0.99992,
		1,
		0.999975,
		0.999515,
		0.997607,
		0.992999,
		0.984492,
		0.971103,
		0.952128,
		0.927158,
		0.89606,
		0.858946,
		0.816138,
		0.768129,
		0.715554,
		0.65916,
		0.599783,
		0.538323,
		0.475734,
		0.413,
		0.351133,
		0.291153,
		0.234088,
		0.180962,
		0.132788,
		0.090569,
		0.055289,
		0.027912,
		0.009378,
		0.000602,
		0.015765,
		0.05819,
		0.120607,
		0.197141,
		0.282661,
		0.372731,
		0.463572,
		0.552017,
		0.635486,
		0.711954,
		0.779928,
		0.838428,
		0.886971,
		0.925553,
		0.95463,
		0.975089,
		0.988203,
		0.995546,
		0.998863,
		0.999876,
		1,
		0.999988,
		0.999651,
		0.998064,
		0.994013,
		0.986304,
		0.973927,
		0.956143,
		0.932498,
		0.902812,
		0.867151,
		0.82579,
		0.77918,
		0.727916,
		0.672708,
		0.614355,
		0.553728,
		0.491749,
		0.429378,
		0.367599,
		0.307411,
		0.249819,
		0.19583,
		0.146439,
		0.102633,
		0.065382,
		0.035637,
		0.014327,
		0.002356,
		0.015563,
		0.057479,
		0.119202,
		0.194962,
		0.279714,
		0.369091,
		0.459363,
		0.547405,
		0.630658,
		0.707107,
		0.775256,
		0.834112,
		0.883162,
		0.922365,
		0.952128,
		0.973282,
		0.987038,
		0.994912,
		0.998605,
		0.999819,
		0.999998,
		0.999995,
		0.999755,
		0.998451,
		0.994912,
		0.987949,
		0.976536,
		0.959898,
		0.937538,
		0.909232,
		0.875,
		0.835071,
		0.789856,
		0.739908,
		0.685901,
		0.628601,
		0.568845,
		0.507526,
		0.445578,
		0.383959,
		0.323646,
		0.265625,
		0.21088,
		0.160392,
		0.11513,
		0.076051,
		0.044093,
		0.020173,
		0.005185,
		0.015366,
		0.056781,
		0.117821,
		0.192819,
		0.276811,
		0.365499,
		0.455206,
		0.542841,
		0.625869,
		0.702287,
		0.770596,
		0.829789,
		0.879329,
		0.919137,
		0.949573,
		0.971415,
		0.985814,
		0.994227,
		0.99831,
		0.999744,
		0.999995,
		0.999998,
		0.999834,
		0.998776,
		0.995704,
		0.98944,
		0.978942,
		0.963404,
		0.94229,
		0.915332,
		0.882503,
		0.843991,
		0.800163,
		0.751535,
		0.698742,
		0.642517,
		0.583667,
		0.523054,
		0.461585,
		0.400195,
		0.339837,
		0.281477,
		0.226079,
		0.174607,
		0.128015,
		0.087245,
		0.053221,
		0.02685,
		0.009016,
		0.000579,
		0.015172,
		0.056095,
		0.116464,
		0.19071,
		0.273952,
		0.361957,
		0.451099,
		0.538323,
		0.62112,
		0.697495,
		0.76595,
		0.825463,
		0.875474,
		0.91587,
		0.946968,
		0.969489,
		0.98453,
		0.99349,
		0.997978,
		0.999651,
		0.999989,
		1,
		0.999892,
		0.999046,
		0.996398,
		0.990785,
		0.981155,
		0.966673,
		0.946765,
		0.92112,
		0.889668,
		0.852555,
		0.810106,
		0.762798,
		0.71123,
		0.656102,
		0.598188,
		0.538323,
		0.477387,
		0.416289,
		0.355962,
		0.297347,
		0.241394,
		0.189047,
		0.141244,
		0.098913,
		0.062966,
		0.034297,
		0.013779,
		0.002265,
		0.014982,
		0.055422,
		0.11513,
		0.188634,
		0.271135,
		0.358463,
		0.447041,
		0.533853,
		0.61641,
		0.692731,
		0.761317,
		0.821134,
		0.8716,
		0.912568,
		0.944313,
		0.967506,
		0.983187,
		0.9927,
		0.997607,
		0.999536,
		0.999977,
		1,
		0.999933,
		0.999268,
		0.997003,
		0.991996,
		0.983187,
		0.969716,
		0.950973,
		0.926607,
		0.896506,
		0.860772,
		0.819691,
		0.773702,
		0.723368,
		0.669354,
		0.612405,
		0.553326,
		0.492972,
		0.432226,
		0.371999,
		0.313213,
		0.256796,
		0.203676,
		0.154777,
		0.11101,
		0.073275,
		0.042454,
		0.019411,
		0.004986,
		0.014796,
		0.054761,
		0.113819,
		0.186592,
		0.268359,
		0.355015,
		0.443032,
		0.529428,
		0.61174,
		0.687996,
		0.756699,
		0.816804,
		0.867708,
		0.909232,
		0.94161,
		0.965467,
		0.981785,
		0.991857,
		0.997195,
		0.999399,
		0.999959,
		1,
		0.999961,
		0.999448,
		0.997527,
		0.993081,
		0.985047,
		0.972543,
		0.954925,
		0.931803,
		0.903024,
		0.868649,
		0.828924,
		0.78425,
		0.735156,
		0.682272,
		0.626313,
		0.568056,
		0.508329,
		0.447991,
		0.387931,
		0.32905,
		0.272257,
		0.218463,
		0.168575,
		0.123492,
		0.084099,
		0.051267,
		0.025848,
		0.008674,
		0.000557,
		0.014613,
		0.054111,
		0.112529,
		0.184582,
		0.265625,
		0.351615,
		0.439072,
		0.52505,
		0.607109,
		0.683291,
		0.752097,
		0.812474,
		0.863799,
		0.905864,
		0.938862,
		0.963372,
		0.980326,
		0.990961,
		0.996742,
		0.999236,
		0.999933,
		1,
		0.999979,
		0.999593,
		0.997978,
		0.99405,
		0.986746,
		0.975164,
		0.958631,
		0.936718,
		0.909232,
		0.876195,
		0.837812,
		0.794448,
		0.746597,
		0.694856,
		0.63991,
		0.582507,
		0.523449,
		0.463572,
		0.403741,
		0.344836,
		0.28775,
		0.233376,
		0.182603,
		0.136317,
		0.095391,
		0.060681,
		0.033031,
		0.013263,
		0.002179,
		0.014433,
		0.053473,
		0.111261,
		0.182603,
		0.262931,
		0.34826,
		0.435159,
		0.520718,
		0.602518,
		0.678615,
		0.747512,
		0.808146,
		0.859876,
		0.902465,
		0.93607,
		0.961225,
		0.97881,
		0.990011,
		0.996245,
		0.999046,
		0.999897,
		1,
		0.99999,
		0.999707,
		0.998364,
		0.994912,
		0.988294,
		0.977591,
		0.962103,
		0.941362,
		0.91514,
		0.883417,
		0.846361,
		0.804301,
		0.757694,
		0.707107,
		0.653193,
		0.596674,
		0.538323,
		0.478955,
		0.419412,
		0.360554,
		0.303253,
		0.248386,
		0.196828,
		0.149448,
		0.107106,
		0.070648,
		0.040905,
		0.018691,
		0.004799,
		0.014256,
		0.052846,
		0.110014,
		0.180656,
		0.260276,
		0.34495,
		0.431293,
		0.51643,
		0.597967,
		0.673969,
		0.742944,
		0.80382,
		0.855939,
		0.899037,
		0.933236,
		0.959025,
		0.977237,
		0.989007,
		0.995704,
		0.998827,
		0.999848,
		0.999999,
		0.999996,
		0.999794,
		0.99869,
		0.995675,
		0.9897,
		0.979834,
		0.965348,
		0.945745,
		0.920755,
		0.890323,
		0.854579,
		0.813813,
		0.76845,
		0.719025,
		0.666161,
		0.610552,
		0.552945,
		0.494131,
		0.43493,
		0.376183,
		0.318742,
		0.263466,
		0.211216,
		0.162846,
		0.119202,
		0.081119,
		0.049418,
		0.0249,
		0.008351,
		0.000536,
		0.014083,
		0.05223,
		0.108788,
		0.178739,
		0.257659,
		0.341684,
		0.427474,
		0.512188,
		0.593455,
		0.669354,
		0.738394,
		0.799498,
		0.85199,
		0.895582,
		0.930361,
		0.956775,
		0.975609,
		0.987949,
		0.995118,
		0.998577,
		0.999785,
		0.999996,
		0.999999,
		0.999861,
		0.998965,
		0.996348,
		0.990973,
		0.981901,
		0.968379,
		0.949876,
		0.926089,
		0.896922,
		0.862471,
		0.82299,
		0.778869,
		0.730613,
		0.678813,
		0.624138,
		0.567308,
		0.50909,
		0.450283,
		0.391707,
		0.334196,
		0.278592,
		0.22574,
		0.176478,
		0.131641,
		0.092052,
		0.058518,
		0.031834,
		0.012775,
		0.002098,
		0.013913,
		0.051625,
		0.107582,
		0.176852,
		0.255081,
		0.338462,
		0.423701,
		0.50799,
		0.588982,
		0.664769,
		0.733863,
		0.795181,
		0.848031,
		0.892101,
		0.927446,
		0.954475,
		0.973927,
		0.986838,
		0.994485,
		0.998294,
		0.999707,
		0.99999,
		1,
		0.999909,
		0.999193,
		0.996937,
		0.992122,
		0.983803,
		0.971204,
		0.953766,
		0.931149,
		0.903223,
		0.870045,
		0.831837,
		0.788955,
		0.741872,
		0.69115,
		0.637429,
		0.581406,
		0.523824,
		0.465459,
		0.407111,
		0.349596,
		0.29374,
		0.24037,
		0.190312,
		0.144385,
		0.103402,
		0.068159,
		0.039439,
		0.018011,
		0.004622,
		0.013746,
		0.05103,
		0.106396,
		0.174994,
		0.25254,
		0.335282,
		0.419973,
		0.503836,
		0.584549,
		0.660215,
		0.729352,
		0.790869,
		0.844063,
		0.888597,
		0.924495,
		0.952128,
		0.972191,
		0.985674,
		0.993807,
		0.997978,
		0.99961,
		0.999981,
		1,
		0.999944,
		0.999381,
		0.997451,
		0.993157,
		0.985548,
		0.973833,
		0.957423,
		0.935944,
		0.909232,
		0.877309,
		0.840361,
		0.798713,
		0.752804,
		0.70317,
		0.650423,
		0.595234,
		0.538323,
		0.480445,
		0.42238,
		0.364925,
		0.308888,
		0.255081,
		0.204316,
		0.1574,
		0.11513,
		0.078295,
		0.047667,
		0.024004,
		0.008046,
		0.000516,
		0.013582,
		0.050445,
		0.105229,
		0.173165,
		0.250035,
		0.332145,
		0.416289,
		0.499726,
		0.580154,
		0.655692,
		0.72486,
		0.786565,
		0.840087,
		0.88507,
		0.921508,
		0.949734,
		0.970403,
		0.984457,
		0.993081,
		0.997627,
		0.999493,
		0.999966,
		1,
		0.999967,
		0.999533,
		0.997897,
		0.994084,
		0.987147,
		0.976276,
		0.960857,
		0.940484,
		0.91496,
		0.88427,
		0.848568,
		0.808146,
		0.763413,
		0.714875,
		0.663118,
		0.608789,
		0.552583,
		0.495233,
		0.437501,
		0.380165,
		0.324015,
		0.269849,
		0.218463,
		0.170651,
		0.127199,
		0.088884,
		0.056468,
		0.0307,
		0.012313,
		0.002021,
		0.013421,
		0.049871,
		0.104082,
		0.171364,
		0.247567,
		0.32905,
		0.41265,
		0.495659,
		0.575799,
		0.651201,
		0.72039,
		0.782268,
		0.836105,
		0.881522,
		0.918486,
		0.947296,
		0.968563,
		0.983187,
		0.992308,
		0.997238,
		0.999355,
		0.999944,
		1,
		0.999982,
		0.999656,
		0.99828,
		0.994912,
		0.988607,
		0.978541,
		0.964076,
		0.944778,
		0.920413,
		0.890936,
		0.856465,
		0.81726,
		0.773702,
		0.726267,
		0.675515,
		0.622067,
		0.566596,
		0.509814,
		0.452462,
		0.395301,
		0.339103,
		0.28465,
		0.232725,
		0.184108,
		0.139573,
		0.099885,
		0.065798,
		0.038051,
		0.017367,
		0.004454,
		0.013263,
		0.049305,
		0.102952,
		0.16959,
		0.245133,
		0.325995,
		0.409054,
		0.491635,
		0.571482,
		0.64674,
		0.71594,
		0.77798,
		0.832117,
		0.877955,
		0.915432,
		0.944814,
		0.966673,
		0.981865,
		0.991486,
		0.996812,
		0.999193,
		0.999913,
		1,
		0.999992,
		0.999752,
		0.998608,
		0.995649,
		0.989936,
		0.980638,
		0.967092,
		0.948834,
		0.925601,
		0.897313,
		0.864056,
		0.82606,
		0.783673,
		0.737347,
		0.687611,
		0.635065,
		0.580358,
		0.52418,
		0.467253,
		0.410318,
		0.354133,
		0.299464,
		0.247077,
		0.197741,
		0.152218,
		0.111261,
		0.075614,
		0.046007,
		0.023155,
		0.007758,
		0.000497,
		0.013107,
		0.04875,
		0.101841,
		0.167843,
		0.242735,
		0.32298,
		0.405501,
		0.487653,
		0.567204,
		0.642312,
		0.711512,
		0.773702,
		0.828126,
		0.874369,
		0.912347,
		0.94229,
		0.964734,
		0.980491,
		0.990617,
		0.996348,
		0.999006,
		0.999872,
		0.999999,
		0.999997,
		0.999826,
		0.998886,
		0.996301,
		0.991144,
		0.982575,
		0.969911,
		0.952661,
		0.930531,
		0.903409,
		0.87135,
		0.834551,
		0.793332,
		0.748116,
		0.699408,
		0.647781,
		0.593863,
		0.538323,
		0.481862,
		0.425205,
		0.369091,
		0.31427,
		0.261496,
		0.211524,
		0.165103,
		0.122976,
		0.085875,
		0.054523,
		0.029626,
		0.011876,
		0.001948,
		0.012955,
		0.048204,
		0.100747,
		0.166123,
		0.24037,
		0.320005,
		0.40199,
		0.483714,
		0.562964,
		0.637915,
		0.707107,
		0.769433,
		0.824131,
		0.870768,
		0.909232,
		0.939726,
		0.962746,
		0.979066,
		0.9897,
		0.995843,
		0.998793,
		0.999819,
		0.999997,
		0.999999,
		0.999882,
		0.999119,
		0.996876,
		0.992239,
		0.984361,
		0.972543,
		0.956268,
		0.935212,
		0.909232,
		0.878352,
		0.842738,
		0.802682,
		0.758579,
		0.710906,
		0.660215,
		0.607109,
		0.552238,
		0.496282,
		0.439948,
		0.383959,
		0.32905,
		0.27596,
		0.225429,
		0.178197,
		0.134994,
		0.096544,
		0.063557,
		0.036734,
		0.016757,
		0.004296,
		0.012805,
		0.047667,
		0.099671,
		0.164428,
		0.238038,
		0.317068,
		0.398522,
		0.479816,
		0.558763,
		0.63355,
		0.702724,
		0.765177,
		0.820135,
		0.867151,
		0.906089,
		0.937122,
		0.960712,
		0.977591,
		0.988734,
		0.995298,
		0.998551,
		0.999752,
		0.999992,
		1,
		0.999923,
		0.999313,
		0.99738,
		0.993226,
		0.986004,
		0.974996,
		0.959663,
		0.939652,
		0.91479,
		0.88507,
		0.850628,
		0.811728,
		0.768736,
		0.722107,
		0.672365,
		0.620093,
		0.565918,
		0.510503,
		0.454537,
		0.398725,
		0.343786,
		0.290446,
		0.239433,
		0.191473,
		0.147285,
		0.107582,
		0.073068,
		0.044432,
		0.02235,
		0.007484,
		0.000479,
		0.012657,
		0.047138,
		0.098612,
		0.162759,
		0.23574,
		0.31417,
		0.395094,
		0.475959,
		0.5546,
		0.629217,
		0.698364,
		0.760931,
		0.816138,
		0.863519,
		0.90292,
		0.934481,
		0.958631,
		0.976067,
		0.98772,
		0.994712,
		0.99828,
		0.999669,
		0.999984,
		1,
		0.999952,
		0.999473,
		0.997819,
		0.994115,
		0.987511,
		0.977279,
		0.962855,
		0.943859,
		0.92009,
		0.89151,
		0.858226,
		0.820474,
		0.778592,
		0.73301,
		0.68423,
		0.63281,
		0.579359,
		0.524519,
		0.468961,
		0.413374,
		0.358463,
		0.304937,
		0.253513,
		0.204903,
		0.159817,
		0.118957,
		0.083016,
		0.052677,
		0.028608,
		0.011462,
		0.001879,
		0.012512,
		0.046619,
		0.09757,
		0.161115,
		0.233473,
		0.31131,
		0.391707,
		0.472143,
		0.550474,
		0.624916,
		0.694027,
		0.756699,
		0.812141,
		0.859876,
		0.899725,
		0.931803,
		0.956507,
		0.974494,
		0.986658,
		0.994084,
		0.997978,
		0.99957,
		0.999971,
		1,
		0.999972,
		0.999603,
		0.9982,
		0.994912,
		0.988892,
		0.9794,
		0.965851,
		0.947842,
		0.92514,
		0.897679,
		0.865538,
		0.828924,
		0.78815,
		0.743619,
		0.69581,
		0.645261,
		0.592557,
		0.538323,
		0.483212,
		0.427896,
		0.373065,
		0.319414,
		0.267647,
		0.218463,
		0.172561,
		0.130634,
		0.093366,
		0.061429,
		0.035484,
		0.016179,
		0.004145,
		0.01237,
		0.046108,
		0.096544,
		0.159495,
		0.231239,
		0.308487,
		0.38836,
		0.468368,
		0.546387,
		0.620647,
		0.689715,
		0.75248,
		0.808146,
		0.85622,
		0.896506,
		0.92909,
		0.954338,
		0.972873,
		0.985548,
		0.993413,
		0.997644,
		0.999452,
		0.999952,
		1,
		0.999985,
		0.999707,
		0.998528,
		0.995624,
		0.990153,
		0.981366,
		0.968661,
		0.951607,
		0.929947,
		0.903584,
		0.872571,
		0.837084,
		0.797413,
		0.753936,
		0.707107,
		0.657442,
		0.605508,
		0.551909,
		0.497281,
		0.442279,
		0.387578,
		0.333861,
		0.281814,
		0.232129,
		0.185491,
		0.142584,
		0.104082,
		0.070648,
		0.042936,
		0.021587,
		0.007225,
		0.000462,
		0.01223,
		0.045605,
		0.095534,
		0.157899,
		0.229035,
		0.3057,
		0.385053,
		0.464632,
		0.542336,
		0.61641,
		0.685426,
		0.748275,
		0.804153,
		0.852555,
		0.893264,
		0.926344,
		0.952128,
		0.971204,
		0.984391,
		0.9927,
		0.997277,
		0.999313,
		0.999926,
		1,
		0.999993,
		0.999788,
		0.998808,
		0.996257,
		0.991302,
		0.983187,
		0.971292,
		0.955164,
		0.934518,
		0.909232,
		0.879329,
		0.84496,
		0.806386,
		0.763962,
		0.71812,
		0.669354,
		0.618209,
		0.565272,
		0.51116,
		0.456513,
		0.40199,
		0.34826,
		0.295997,
		0.245878,
		0.198581,
		0.154777,
		0.11513,
		0.080297,
		0.050923,
		0.027641,
		0.011069,
		0.001814,
		0.012092,
		0.045111,
		0.094539,
		0.156326,
		0.226862,
		0.302949,
		0.381784,
		0.460936,
		0.538323,
		0.612205,
		0.681162,
		0.744084,
		0.800163,
		0.84888,
		0.890001,
		0.923565,
		0.949876,
		0.969489,
		0.983187,
		0.991944,
		0.996876,
		0.999153,
		0.999891,
		0.999999,
		0.999997,
		0.999852,
		0.999046,
		0.996818,
		0.992346,
		0.984868,
		0.973751,
		0.958519,
		0.938862,
		0.91463,
		0.88582,
		0.852555,
		0.815072,
		0.773702,
		0.728851,
		0.680995,
		0.630658,
		0.578407,
		0.524843,
		0.470589,
		0.416289,
		0.362598,
		0.310176,
		0.259691,
		0.211807,
		0.167184,
		0.12648,
		0.090341,
		0.059405,
		0.034297,
		0.01563,
		0.004003,
		0.011957,
		0.044624,
		0.09356,
		0.154777,
		0.224718,
		0.300233,
		0.378555,
		0.457278,
		0.534347,
		0.608032,
		0.676922,
		0.739908,
		0.796177,
		0.845197,
		0.886718,
		0.920755,
		0.947585,
		0.967729,
		0.981935,
		0.991144,
		0.996439,
		0.99897,
		0.999846,
		0.999997,
		0.999999,
		0.999899,
		0.999246,
		0.997312,
		0.993291,
		0.986419,
		0.976047,
		0.961681,
		0.942986,
		0.919785,
		0.892049,
		0.859876,
		0.823476,
		0.783156,
		0.739302,
		0.692366,
		0.642853,
		0.591311,
		0.538323,
		0.484498,
		0.430464,
		0.37686,
		0.324337,
		0.273547,
		0.225145,
		0.179781,
		0.138102,
		0.100747,
		0.068345,
		0.041515,
		0.020862,
		0.006979,
		0.000447,
		0.011823,
		0.044146,
		0.092596,
		0.15325,
		0.222604,
		0.297552,
		0.375363,
		0.45366,
		0.530408,
		0.603892,
		0.672708,
		0.735749,
		0.792195,
		0.841508,
		0.883417,
		0.917916,
		0.945255,
		0.965925,
		0.980638,
		0.990301,
		0.995967,
		0.998762,
		0.999788,
		0.999993,
		1,
		0.999934,
		0.999412,
		0.997745,
		0.994144,
		0.987845,
		0.978187,
		0.964657,
		0.946896,
		0.924704,
		0.898023,
		0.866927,
		0.831602,
		0.79233,
		0.749475,
		0.703467,
		0.654791,
		0.603979,
		0.551596,
		0.498233,
		0.444504,
		0.391034,
		0.338462,
		0.287427,
		0.238574,
		0.192542,
		0.149968,
		0.111483,
		0.077708,
		0.049255,
		0.026722,
		0.010696,
		0.001752,
		0.011692,
		0.043674,
		0.091646,
		0.151745,
		0.220519,
		0.294906,
		0.372208,
		0.45008,
		0.526505,
		0.599783,
		0.668518,
		0.731605,
		0.78822,
		0.837812,
		0.880097,
		0.915048,
		0.942888,
		0.964076,
		0.979295,
		0.989415,
		0.995458,
		0.998528,
		0.999718,
		0.999986,
		1,
		0.999959,
		0.999548,
		0.998123,
		0.994912,
		0.989154,
		0.980179,
		0.967454,
		0.950601,
		0.929393,
		0.903749,
		0.873716,
		0.839455,
		0.801227,
		0.759371,
		0.714298,
		0.666474,
		0.61641,
		0.564656,
		0.511786,
		0.4584,
		0.405109,
		0.352538,
		0.301315,
		0.252073,
		0.205443,
		0.162051,
		0.122519,
		0.087461,
		0.057479,
		0.033168,
		0.015108,
		0.003868,
		0.011564,
		0.043211,
		0.090711,
		0.150262,
		0.218463,
		0.292293,
		0.369091,
		0.446537,
		0.522638,
		0.595706,
		0.664354,
		0.727478,
		0.78425,
		0.834112,
		0.876762,
		0.912153,
		0.940484,
		0.962186,
		0.977906,
		0.988485,
		0.994912,
		0.998267,
		0.999632,
		0.999975,
		1,
		0.999976,
		0.99966,
		0.998451,
		0.995601,
		0.990352,
		0.982029,
		0.970081,
		0.954108,
		0.93386,
		0.909232,
		0.880246,
		0.84704,
		0.809849,
		0.768993,
		0.72486,
		0.677898,
		0.628601,
		0.577498,
		0.525152,
		0.472143,
		0.419073,
		0.366551,
		0.315194,
		0.265625,
		0.218463,
		0.174326,
		0.133826,
		0.09757,
		0.066153,
		0.040162,
		0.020173,
		0.006746,
		0.000431,
		0.011437,
		0.042755,
		0.089791,
		0.148801,
		0.216434,
		0.289713,
		0.366009,
		0.443032,
		0.518807,
		0.591661,
		0.660215,
		0.723368,
		0.780288,
		0.830407,
		0.873411,
		0.909232,
		0.938045,
		0.960253,
		0.976473,
		0.987511,
		0.994328,
		0.997978,
		0.999531,
		0.999959,
		1,
		0.999987,
		0.999749,
		0.998733,
		0.996216,
		0.991447,
		0.983744,
		0.972543,
		0.957423,
		0.938112,
		0.914479,
		0.886525,
		0.854361,
		0.818201,
		0.778343,
		0.735156,
		0.689066,
		0.640549,
		0.59012,
		0.538323,
		0.485726,
		0.432915,
		0.380488,
		0.32905,
		0.27921,
		0.23158,
		0.186768,
		0.145377,
		0.108006,
		0.075242,
		0.047667,
		0.025848,
		0.010342,
		0.001693,
		0.011312,
		0.042305,
		0.088884,
		0.14736,
		0.214433,
		0.287165,
		0.362964,
		0.439564,
		0.515011,
		0.587648,
		0.656102,
		0.719275,
		0.776333,
		0.826699,
		0.870045,
		0.906287,
		0.935573,
		0.95828,
		0.974996,
		0.986495,
		0.993706,
		0.99766,
		0.999412,
		0.999937,
		1,
		0.999994,
		0.999819,
		0.998974,
		0.996763,
		0.992444,
		0.985332,
		0.974848,
		0.960553,
		0.942154,
		0.919497,
		0.892557,
		0.861423,
		0.826287,
		0.787425,
		0.745185,
		0.699976,
		0.652255,
		0.602518,
		0.551297,
		0.499142,
		0.446628,
		0.394338,
		0.342867,
		0.292812,
		0.244776,
		0.199356,
		0.157147,
		0.118739,
		0.084714,
		0.055645,
		0.032094,
		0.014613,
		0.003739,
		0.01119,
		0.041863,
		0.08799,
		0.14594,
		0.212459,
		0.28465,
		0.359954,
		0.436133,
		0.511251,
		0.583667,
		0.652015,
		0.715201,
		0.772387,
		0.82299,
		0.866667,
		0.903317,
		0.933068,
		0.956268,
		0.973475,
		0.985435,
		0.993045,
		0.997312,
		0.999274,
		0.999907,
		0.999999,
		0.999998,
		0.999873,
		0.999178,
		0.997247,
		0.99335,
		0.986798,
		0.977002,
		0.963506,
		0.945995,
		0.924291,
		0.898348,
		0.868232,
		0.834112,
		0.796242,
		0.754952,
		0.710629,
		0.663715,
		0.61469,
		0.564067,
		0.512384,
		0.460201,
		0.40809,
		0.356632,
		0.306416,
		0.258031,
		0.212067,
		0.169111,
		0.129743,
		0.094539,
		0.064064,
		0.038875,
		0.019517,
		0.006524,
		0.000417,
		0.011069,
		0.041428,
		0.08711,
		0.14454,
		0.210511,
		0.282166,
		0.35698,
		0.432738,
		0.507526,
		0.579717,
		0.647954,
		0.711144,
		0.76845,
		0.819279,
		0.863277,
		0.900326,
		0.930531,
		0.954217,
		0.971912,
		0.984332,
		0.992346,
		0.996932,
		0.999115,
		0.999868,
		0.999998,
		0.999999,
		0.999914,
		0.99935,
		0.997675,
		0.994171,
		0.98815,
		0.979014,
		0.966289,
		0.94964,
		0.928869,
		0.903905,
		0.874792,
		0.841678,
		0.804797,
		0.764456,
		0.721027,
		0.674931,
		0.626633,
		0.57663,
		0.525446,
		0.473628,
		0.421733,
		0.370333,
		0.320005,
		0.271328,
		0.224882,
		0.181245,
		0.140991,
		0.104687,
		0.072891,
		0.046154,
		0.025016,
		0.010005,
		0.001638,
		0.010951,
		0.041,
		0.086243,
		0.14316,
		0.208589,
		0.279714,
		0.354039,
		0.429378,
		0.503836,
		0.575799,
		0.643919,
		0.707107,
		0.764523,
		0.815567,
		0.859876,
		0.897313,
		0.927963,
		0.952128,
		0.970307,
		0.983187,
		0.991607,
		0.996521,
		0.998935,
		0.999819,
		0.999994,
		1,
		0.999944,
		0.999493,
		0.998049,
		0.994912,
		0.989394,
		0.980889,
		0.968908,
		0.953097,
		0.933236,
		0.909232,
		0.881109,
		0.848992,
		0.813093,
		0.773702,
		0.73117,
		0.685901,
		0.638345,
		0.588982,
		0.538323,
		0.4869,
		0.435259,
		0.383959,
		0.333566,
		0.28465,
		0.237781,
		0.193529,
		0.152459,
		0.11513,
		0.082095,
		0.053897,
		0.031071,
		0.014141,
		0.003617,
		0.010834,
		0.040578,
		0.085389,
		0.141799,
		0.206693,
		0.277292,
		0.351133,
		0.426054,
		0.50018,
		0.571912,
		0.63991,
		0.703088,
		0.760605,
		0.811856,
		0.856465,
		0.894279,
		0.925367,
		0.950003,
		0.968661,
		0.981999,
		0.990828,
		0.996078,
		0.998733,
		0.999758,
		0.999988,
		1,
		0.999965,
		0.999611,
		0.998376,
		0.995579,
		0.990536,
		0.982633,
		0.971369,
		0.956371,
		0.937398,
		0.914337,
		0.887188,
		0.856057,
		0.821134,
		0.78269,
		0.741059,
		0.696627,
		0.649825,
		0.601121,
		0.55101,
		0.500011,
		0.448658,
		0.397498,
		0.347086,
		0.297982,
		0.250747,
		0.205942,
		0.164123,
		0.125842,
		0.091646,
		0.062072,
		0.037648,
		0.018893,
		0.006313,
		0.000403,
		0.010719,
		0.040162,
		0.084547,
		0.140457,
		0.204822,
		0.2749,
		0.34826,
		0.422765,
		0.496559,
		0.568056,
		0.635927,
		0.699089,
		0.756699,
		0.808146,
		0.853044,
		0.891227,
		0.922742,
		0.947842,
		0.966975,
		0.98077,
		0.990011,
		0.995601,
		0.998506,
		0.999684,
		0.999979,
		1,
		0.99998,
		0.999707,
		0.998659,
		0.996177,
		0.991581,
		0.984253,
		0.973679,
		0.959469,
		0.941362,
		0.919224,
		0.893035,
		0.862877,
		0.828924,
		0.791425,
		0.750696,
		0.707107,
		0.661072,
		0.613044,
		0.563503,
		0.512956,
		0.461923,
		0.410941,
		0.360554,
		0.31131,
		0.263763,
		0.218463,
		0.17596,
		0.136799,
		0.101517,
		0.070648,
		0.044712,
		0.024223,
		0.009684,
		0.001584,
		0.010606,
		0.039753,
		0.083718,
		0.139134,
		0.202976,
		0.272538,
		0.34542,
		0.41951,
		0.492972,
		0.564232,
		0.631971,
		0.695109,
		0.752804,
		0.804438,
		0.849616,
		0.888157,
		0.92009,
		0.945646,
		0.965249,
		0.979499,
		0.989154,
		0.99509,
		0.998255,
		0.999596,
		0.999965,
		1,
		0.999989,
		0.999783,
		0.998903,
		0.996711,
		0.992536,
		0.985756,
		0.975845,
		0.962398,
		0.945135,
		0.9239,
		0.898654,
		0.869459,
		0.836467,
		0.79991,
		0.760084,
		0.717343,
		0.672085,
		0.624748,
		0.575799,
		0.525728,
		0.475047,
		0.424278,
		0.373956,
		0.324619,
		0.276811,
		0.231074,
		0.18795,
		0.147975,
		0.111681,
		0.079594,
		0.05223,
		0.030097,
		0.013691,
		0.003501,
		0.010495,
		0.03935,
		0.0829,
		0.137829,
		0.201154,
		0.270205,
		0.342613,
		0.416289,
		0.489418,
		0.560439,
		0.628041,
		0.69115,
		0.748921,
		0.800733,
		0.84618,
		0.88507,
		0.917412,
		0.943417,
		0.963484,
		0.978187,
		0.988257,
		0.994545,
		0.997978,
		0.999493,
		0.999946,
		1,
		0.999995,
		0.999844,
		0.999112,
		0.997186,
		0.993406,
		0.987147,
		0.977873,
		0.965164,
		0.948721,
		0.928371,
		0.904051,
		0.875806,
		0.843766,
		0.808146,
		0.769224,
		0.727336,
		0.682865,
		0.636233,
		0.587893,
		0.538323,
		0.488022,
		0.437501,
		0.387283,
		0.337898,
		0.289878,
		0.243759,
		0.200072,
		0.159349,
		0.122113,
		0.088884,
		0.060171,
		0.036479,
		0.018298,
		0.006111,
		0.00039,
		0.010385,
		0.038954,
		0.082095,
		0.136543,
		0.199356,
		0.267901,
		0.339837,
		0.413102,
		0.485897,
		0.556677,
		0.624138,
		0.68721,
		0.74505,
		0.797031,
		0.842738,
		0.881966,
		0.914709,
		0.941155,
		0.961681,
		0.976835,
		0.987321,
		0.993965,
		0.997675,
		0.999373,
		0.99992,
		0.999999,
		0.999998,
		0.99989,
		0.999289,
		0.997607,
		0.994196,
		0.988431,
		0.979768,
		0.967772,
		0.952128,
		0.932642,
		0.909232,
		0.881923,
		0.850827,
		0.816138,
		0.778118,
		0.737086,
		0.69341,
		0.647496,
		0.599783,
		0.550737,
		0.500843,
		0.450601,
		0.400525,
		0.351133,
		0.302949,
		0.256499,
		0.212308,
		0.170898,
		0.132788,
		0.098489,
		0.068506,
		0.043337,
		0.023468,
		0.009378,
		0.001534,
		0.010277,
		0.038563,
		0.081301,
		0.135274,
		0.197581,
		0.265625,
		0.337094,
		0.409949,
		0.48241,
		0.552945,
		0.620261,
		0.683291,
		0.741192,
		0.793332,
		0.839291,
		0.878848,
		0.911982,
		0.938862,
		0.959841,
		0.975444,
		0.986346,
		0.99335,
		0.997343,
		0.999236,
		0.999886,
		0.999998,
		0.999999,
		0.999926,
		0.999437,
		0.997978,
		0.994912,
		0.989616,
		0.981538,
		0.97023,
		0.955361,
		0.936718,
		0.914202,
		0.887815,
		0.857652,
		0.823889,
		0.78677,
		0.746597,
		0.703721,
		0.658536,
		0.611467,
		0.562964,
		0.513503,
		0.463572,
		0.413672,
		0.364313,
		0.31601,
		0.26928,
		0.224639,
		0.182603,
		0.143683,
		0.108384,
		0.077206,
		0.050639,
		0.029167,
		0.013263,
		0.00339,
		0.010171,
		0.038178,
		0.080519,
		0.134022,
		0.19583,
		0.263377,
		0.334382,
		0.406828,
		0.478955,
		0.549244,
		0.61641,
		0.679392,
		0.737347,
		0.789639,
		0.835839,
		0.875716,
		0.909232,
		0.936538,
		0.957965,
		0.974012,
		0.985332,
		0.9927,
		0.996984,
		0.999079,
		0.999844,
		0.999995,
		1,
		0.999951,
		0.999561,
		0.998304,
		0.995558,
		0.990705,
		0.983187,
		0.972543,
		0.958426,
		0.940607,
		0.918966,
		0.893487,
		0.864247,
		0.831403,
		0.795181,
		0.755868,
		0.7138,
		0.669354,
		0.622942,
		0.575003,
		0.525998,
		0.476406,
		0.426716,
		0.377429,
		0.32905,
		0.282087,
		0.237049,
		0.194445,
		0.154777,
		0.118545,
		0.086243,
		0.058356,
		0.035363,
		0.017731,
		0.00592,
		0.000378,
		0.010067,
		0.037798,
		0.079747,
		0.132788,
		0.194101,
		0.261156,
		0.331701,
		0.403741,
		0.475533,
		0.545574,
		0.612586,
		0.675515,
		0.733515,
		0.785951,
		0.832383,
		0.872571,
		0.90646,
		0.934185,
		0.956054,
		0.972543,
		0.984278,
		0.992015,
		0.996595,
		0.998903,
		0.999791,
		0.99999,
		1,
		0.99997,
		0.999663,
		0.998588,
		0.99614,
		0.991705,
		0.984721,
		0.974716,
		0.961329,
		0.944313,
		0.923529,
		0.898944,
		0.870615,
		0.838682,
		0.803355,
		0.764902,
		0.723646,
		0.679948,
		0.634207,
		0.586849,
		0.538323,
		0.489096,
		0.439648,
		0.390469,
		0.342055,
		0.294906,
		0.249522,
		0.206404,
		0.166048,
		0.128949,
		0.095593,
		0.06646,
		0.042023,
		0.022747,
		0.009086,
		0.001486,
		0.009964,
		0.037425,
		0.078987,
		0.13157,
		0.192394,
		0.258963,
		0.32905,
		0.400685,
		0.472143,
		0.541933,
		0.608789,
		0.671658,
		0.729698,
		0.782268,
		0.828924,
		0.869413,
		0.903668,
		0.931803,
		0.954108,
		0.971035,
		0.983187,
		0.991293,
		0.996177,
		0.998706,
		0.999727,
		0.999982,
		1,
		0.999982,
		0.999746,
		0.998834,
		0.996662,
		0.992621,
		0.986146,
		0.976757,
		0.964076,
		0.947842,
		0.927897,
		0.904191,
		0.876762,
		0.845732,
		0.811294,
		0.773702,
		0.73326,
		0.690319,
		0.645261,
		0.5985,
		0.550474,
		0.501639,
		0.452462,
		0.403426,
		0.355015,
		0.307723,
		0.262041,
		0.218463,
		0.177478,
		0.139573,
		0.105229,
		0.074922,
		0.049119,
		0.02828,
		0.012854,
		0.003284,
		0.009862,
		0.037057,
		0.078237,
		0.130369,
		0.19071,
		0.256796,
		0.326429,
		0.397661,
		0.468785,
		0.538323,
		0.605018,
		0.667822,
		0.725895,
		0.778592,
		0.825463,
		0.866244,
		0.900855,
		0.929393,
		0.952128,
		0.969489,
		0.982057,
		0.990536,
		0.995728,
		0.998486,
		0.999651,
		0.99997,
		1,
		0.999991,
		0.999812,
		0.999046,
		0.997128,
		0.993458,
		0.987468,
		0.978669,
		0.966673,
		0.951199,
		0.932076,
		0.909232,
		0.88269,
		0.852555,
		0.819002,
		0.782268,
		0.742645,
		0.700466,
		0.656102,
		0.609954,
		0.562448,
		0.514027,
		0.465151,
		0.416289,
		0.36792,
		0.320527,
		0.274594,
		0.230606,
		0.189047,
		0.150396,
		0.11513,
		0.083718,
		0.056622,
		0.034297,
		0.01719,
		0.005737,
		0.000366,
		0.009762,
		0.036694,
		0.077498,
		0.129184,
		0.189047,
		0.254655,
		0.323837,
		0.394668,
		0.465459,
		0.534743,
		0.601273,
		0.664008,
		0.722107,
		0.774923,
		0.822,
		0.863065,
		0.898023,
		0.926957,
		0.950115,
		0.967907,
		0.980889,
		0.989742,
		0.995249,
		0.998244,
		0.999561,
		0.999953,
		1,
		0.999996,
		0.999864,
		0.999227,
		0.997543,
		0.99422,
		0.988691,
		0.980459,
		0.969125,
		0.954391,
		0.93607,
		0.914074,
		0.888407,
		0.859156,
		0.826481,
		0.790604,
		0.751801,
		0.71039,
		0.66673,
		0.621209,
		0.57424,
		0.526257,
		0.477707,
		0.429052,
		0.38076,
		0.333306,
		0.287165,
		0.242817,
		0.200737,
		0.161399,
		0.125272,
		0.092821,
		0.064503,
		0.040769,
		0.022059,
		0.008808,
		0.00144,
		0.009664,
		0.036336,
		0.07677,
		0.128015,
		0.187405,
		0.25254,
		0.321275,
		0.391707,
		0.462163,
		0.531193,
		0.597555,
		0.660215,
		0.718333,
		0.771261,
		0.818536,
		0.859876,
		0.895174,
		0.924495,
		0.948071,
		0.966289,
		0.979683,
		0.988913,
		0.994738,
		0.997978,
		0.999457,
		0.999931,
		0.999999,
		0.999998,
		0.999905,
		0.999381,
		0.99791,
		0.994912,
		0.989821,
		0.982133,
		0.971437,
		0.957423,
		0.939885,
		0.91872,
		0.893914,
		0.865538,
		0.833736,
		0.798713,
		0.76073,
		0.720092,
		0.677145,
		0.632263,
		0.585848,
		0.538323,
		0.490126,
		0.441707,
		0.393526,
		0.346048,
		0.299743,
		0.255081,
		0.212531,
		0.172561,
		0.135635,
		0.102209,
		0.072738,
		0.047667,
		0.027432,
		0.012464,
		0.003183,
		0.009567,
		0.035984,
		0.076051,
		0.126861,
		0.185784,
		0.25045,
		0.318742,
		0.388776,
		0.458899,
		0.527672,
		0.593863,
		0.656444,
		0.714575,
		0.767608,
		0.815072,
		0.856678,
		0.892307,
		0.922008,
		0.945995,
		0.964635,
		0.97844,
		0.988048,
		0.994196,
		0.997688,
		0.999337,
		0.999901,
		0.999998,
		0.999999,
		0.999935,
		0.99951,
		0.998234,
		0.995539,
		0.990862,
		0.983695,
		0.973615,
		0.960299,
		0.943527,
		0.923176,
		0.899218,
		0.871707,
		0.840769,
		0.806596,
		0.769433,
		0.729573,
		0.687346,
		0.643115,
		0.59727,
		0.550223,
		0.502402,
		0.454246,
		0.406208,
		0.358744,
		0.312315,
		0.267385,
		0.224414,
		0.183866,
		0.146196,
		0.111859,
		0.081301,
		0.054963,
		0.033279,
		0.016673,
		0.009472,
		0.035637,
		0.075343,
		0.125723,
		0.184184,
		0.248386,
		0.316237,
		0.385876,
		0.455665,
		0.52418,
		0.590198,
		0.652694,
		0.710833,
		0.763962,
		0.811609,
		0.853472,
		0.889423,
		0.919497,
		0.943889,
		0.962947,
		0.977161,
		0.987147,
		0.993622,
		0.997372,
		0.9992,
		0.999864,
		0.999996,
		1,
		0.999958,
		0.999618,
		0.998518,
		0.996106,
		0.991821,
		0.985151,
		0.975665,
		0.963026,
		0.946999,
		0.927446,
		0.904322,
		0.877665,
		0.847584,
		0.814257,
		0.777914,
		0.738833,
		0.697333,
		0.653763,
		0.608503,
		0.561953,
		0.514529,
		0.466665,
		0.4188,
		0.371384,
		0.32487,
		0.279714,
		0.236371,
		0.195295,
		0.156939,
		0.121749,
		0.090168,
		0.062632,
		0.039569,
		0.009378,
		0.035295,
		0.074644,
		0.1246,
		0.182603,
		0.246346,
		0.31376,
		0.383005,
		0.452462,
		0.520718,
		0.586559,
		0.648967,
		0.707107,
		0.760326,
		0.808146,
		0.850259,
		0.886525,
		0.916963,
		0.941754,
		0.961225,
		0.975845,
		0.98621,
		0.993016,
		0.99703,
		0.999046,
		0.999819,
		0.999991,
		1,
		0.999974,
		0.999707,
		0.998766,
		0.996615,
		0.9927,
		0.986506,
		0.977591,
		0.965609,
		0.950308,
		0.931537,
		0.909232,
		0.883417,
		0.854185,
		0.821699,
		0.786174,
		0.747875,
		0.707107,
		0.664208,
		0.619546,
		0.573509,
		0.526505,
		0.478955,
		0.431293,
		0.383959,
		0.337397,
		0.292057,
		0.248386,
		0.206833,
		0.167843,
		0.131859,
		0.099316,
		0.070648,
		0.009285,
		0.034957,
		0.073955,
		0.123492,
		0.181043,
		0.24433,
		0.31131,
		0.380165,
		0.449289,
		0.517284,
		0.582946,
		0.645261,
		0.703397,
		0.756699,
		0.804685,
		0.84704,
		0.883611,
		0.914407,
		0.93959,
		0.959469,
		0.974494,
		0.985237,
		0.992377,
		0.996662,
		0.998873,
		0.999763,
		0.999984,
		1,
		0.999985,
		0.999779,
		0.998981,
		0.997073,
		0.993506,
		0.987764,
		0.9794,
		0.968053,
		0.953459,
		0.935451,
		0.913952,
		0.888967,
		0.860575,
		0.828924,
		0.794215,
		0.756699,
		0.716668,
		0.674449,
		0.630396,
		0.584888,
		0.538323,
		0.491113,
		0.443682,
		0.39646,
		0.349886,
		0.304401,
		0.260446,
		0.218463,
		0.178891,
		0.142168,
		0.108724,
		0.009194,
		0.034625,
		0.073275,
		0.122398,
		0.179502,
		0.242338,
		0.308888,
		0.377353,
		0.446146,
		0.51388,
		0.579359,
		0.641577,
		0.699703,
		0.753082,
		0.801227,
		0.843815,
		0.880684,
		0.91183,
		0.937398,
		0.957681,
		0.973107,
		0.98423,
		0.991705,
		0.996266,
		0.998681,
		0.999696,
		0.999974,
		1,
		0.999992,
		0.999836,
		0.999166,
		0.997481,
		0.994242,
		0.988931,
		0.981094,
		0.970362,
		0.956457,
		0.939196,
		0.918486,
		0.894319,
		0.866759,
		0.835936,
		0.80204,
		0.765307,
		0.726017,
		0.684484,
		0.641052,
		0.596089,
		0.549982,
		0.503133,
		0.455958,
		0.40888,
		0.362328,
		0.316735,
		0.272538,
		0.230171,
		0.190067,
		0.152659,
		0.009104,
		0.034297,
		0.072605,
		0.121319,
		0.177981,
		0.24037,
		0.306492,
		0.37457,
		0.443032,
		0.510503,
		0.575799,
		0.637915,
		0.696026,
		0.749475,
		0.797771,
		0.840584,
		0.877744,
		0.909232,
		0.935179,
		0.955861,
		0.971686,
		0.983187,
		0.991001,
		0.995843,
		0.998468,
		0.999618,
		0.999959,
		1,
		0.999996,
		0.999882,
		0.999324,
		0.997845,
		0.994912,
		0.990011,
		0.982681,
		0.972543,
		0.959306,
		0.942774,
		0.92284,
		0.899478,
		0.872739,
		0.842738,
		0.809651,
		0.773702,
		0.735156,
		0.694316,
		0.651515,
		0.607109,
		0.561477,
		0.515011,
		0.468117,
		0.42121,
		0.374712,
		0.32905,
		0.28465,
		0.241943,
		0.201355,
		0.009016,
		0.033974,
		0.071944,
		0.120254,
		0.176478,
		0.238425,
		0.304124,
		0.371816,
		0.439948,
		0.507156,
		0.572264,
		0.634275,
		0.692366,
		0.745878,
		0.794318,
		0.83735,
		0.874792,
		0.906615,
		0.932935,
		0.954009,
		0.97023,
		0.982109,
		0.990264,
		0.995392,
		0.998234,
		0.999527,
		0.99994,
		0.999999,
		0.999998,
		0.999917,
		0.999459,
		0.998167,
		0.995521,
		0.991008,
		0.984164,
		0.974599,
		0.962012,
		0.946192,
		0.927017,
		0.904447,
		0.878519,
		0.849334,
		0.817052,
		0.781884,
		0.744084,
		0.703943,
		0.661782,
		0.617947,
		0.572806,
		0.526742,
		0.480153,
		0.433445,
		0.387032,
		0.341334,
		0.29677,
		0.253765,
		0.008928,
		0.033655,
		0.071291,
		0.119202,
		0.174994,
		0.236502,
		0.301781,
		0.369091,
		0.436892,
		0.503836,
		0.568755,
		0.630658,
		0.688723,
		0.742293,
		0.790869,
		0.834112,
		0.871829,
		0.903979,
		0.930665,
		0.952128,
		0.968741,
		0.980996,
		0.989494,
		0.994912,
		0.997978,
		0.999422,
		0.999914,
		0.999998,
		1,
		0.999944,
		0.999572,
		0.998451,
		0.996073,
		0.991928,
		0.985548,
		0.976536,
		0.964579,
		0.949453,
		0.931022,
		0.909232,
		0.884104,
		0.855726,
		0.824244,
		0.789856,
		0.752804,
		0.713366,
		0.671853,
		0.628601,
		0.583966,
		0.538323,
		0.492061,
		0.445578,
		0.39928,
		0.353578,
		0.308888,
		0.008842,
		0.033341,
		0.070648,
		0.118164,
		0.173528,
		0.234603,
		0.299464,
		0.366393,
		0.433865,
		0.500544,
		0.565272,
		0.627063,
		0.685097,
		0.738718,
		0.787425,
		0.83087,
		0.868855,
		0.901325,
		0.928371,
		0.950217,
		0.967218,
		0.97985,
		0.988691,
		0.994403,
		0.9977,
		0.999302,
		0.999882,
		0.999996,
		1,
		0.999963,
		0.999666,
		0.9987,
		0.996571,
		0.992774,
		0.986838,
		0.978359,
		0.967013,
		0.952563,
		0.93486,
		0.913836,
		0.889497,
		0.861918,
		0.83123,
		0.79762,
		0.761317,
		0.722587,
		0.681729,
		0.639069,
		0.594955,
		0.54975,
		0.503836,
		0.457602,
		0.411446,
		0.365774,
		0.008758,
		0.033031,
		0.070013,
		0.11714,
		0.172081,
		0.232725,
		0.297172,
		0.363722,
		0.430867,
		0.497281,
		0.561815,
		0.62349,
		0.681489,
		0.735156,
		0.783986,
		0.827626,
		0.865871,
		0.898654,
		0.926053,
		0.948277,
		0.965663,
		0.978669,
		0.987855,
		0.993865,
		0.997398,
		0.999166,
		0.999842,
		0.999992,
		1,
		0.999977,
		0.999743,
		0.998917,
		0.99702,
		0.993551,
		0.988039,
		0.980072,
		0.969317,
		0.955526,
		0.938536,
		0.918264,
		0.894703,
		0.867913,
		0.838015,
		0.805179,
		0.769624,
		0.731605,
		0.691409,
		0.649351,
		0.60577,
		0.561021,
		0.515474,
		0.469512,
		0.423526,
		0.008674,
		0.032726,
		0.069387,
		0.116129,
		0.170651,
		0.230869,
		0.294906,
		0.361079,
		0.427896,
		0.494044,
		0.558383,
		0.619939,
		0.677898,
		0.731605,
		0.780552,
		0.824381,
		0.862877,
		0.895967,
		0.923712,
		0.946308,
		0.964076,
		0.977455,
		0.986986,
		0.993298,
		0.997073,
		0.999014,
		0.999793,
		0.999986,
		1,
		0.999987,
		0.999806,
		0.999106,
		0.997422,
		0.994262,
		0.989154,
		0.98168,
		0.971498,
		0.958348,
		0.942053,
		0.922519,
		0.899725,
		0.873716,
		0.8446,
		0.812535,
		0.777728,
		0.740422,
		0.700893,
		0.659446,
		0.61641,
		0.572131,
		0.526971,
		0.481304,
		0.008592,
		0.032424,
		0.068768,
		0.11513,
		0.169238,
		0.229035,
		0.292664,
		0.358463,
		0.424954,
		0.490835,
		0.554977,
		0.61641,
		0.674326,
		0.728066,
		0.777124,
		0.821134,
		0.859876,
		0.893264,
		0.92135,
		0.944313,
		0.962458,
		0.976208,
		0.986085,
		0.9927,
		0.996723,
		0.998845,
		0.999735,
		0.999977,
		1,
		0.999993,
		0.999857,
		0.999268,
		0.997782,
		0.994912,
		0.990188,
		0.983187,
		0.973558,
		0.961031,
		0.945417,
		0.926607,
		0.904566,
		0.879329,
		0.850988,
		0.819691,
		0.78563,
		0.749038,
		0.710181,
		0.669354,
		0.626874,
		0.583079,
		0.538323,
		0.00851,
		0.032127,
		0.068159,
		0.114144,
		0.167843,
		0.227222,
		0.290446,
		0.355873,
		0.422038,
		0.487653,
		0.551596,
		0.612904,
		0.670771,
		0.72454,
		0.773702,
		0.817887,
		0.856866,
		0.890546,
		0.918966,
		0.94229,
		0.96081,
		0.974928,
		0.985151,
		0.992073,
		0.996348,
		0.998657,
		0.999666,
		0.999965,
		1,
		0.999997,
		0.999897,
		0.999407,
		0.998102,
		0.995504,
		0.991144,
		0.984597,
		0.975504,
		0.963582,
		0.948631,
		0.930531,
		0.909232,
		0.884756,
		0.857183,
		0.826649,
		0.793332,
		0.757456,
		0.719275,
		0.679074,
		0.63716,
		0.593863,
		0.00843,
		0.031834,
		0.067557,
		0.113171,
		0.166465,
		0.225429,
		0.288253,
		0.353309,
		0.41915,
		0.484498,
		0.54824,
		0.60942,
		0.667234,
		0.721027,
		0.770286,
		0.814639,
		0.85385,
		0.887815,
		0.916561,
		0.940242,
		0.959131,
		0.973615,
		0.984185,
		0.991416,
		0.995947,
		0.998451,
		0.999586,
		0.999947,
		1,
		0.999999,
		0.999927,
		0.999525,
		0.998385,
		0.996042,
		0.992029,
		0.985916,
		0.977338,
		0.966004,
		0.951701,
		0.934295,
		0.913726,
		0.890001,
		0.863189,
		0.833411,
		0.800837,
		0.765677,
		0.728175,
		0.688606,
		0.647268,
		0.008351,
		0.031545,
		0.066963,
		0.11221,
		0.165103,
		0.223658,
		0.286083,
		0.350772,
		0.416289,
		0.48137,
		0.54491,
		0.605958,
		0.663715,
		0.717527,
		0.766878,
		0.811392,
		0.850827,
		0.88507,
		0.914137,
		0.938168,
		0.957423,
		0.972271,
		0.983187,
		0.990728,
		0.995521,
		0.998225,
		0.999493,
		0.999925,
		0.999999,
		1,
		0.999951,
		0.999624,
		0.998635,
		0.996529,
		0.992844,
		0.987147,
		0.979066,
		0.968302,
		0.95463,
		0.937904,
		0.918052,
		0.895068,
		0.869007,
		0.839981,
		0.808146,
		0.773702,
		0.736881,
		0.69795,
		0.008273,
		0.031259,
		0.066377,
		0.111261,
		0.163757,
		0.221906,
		0.283937,
		0.34826,
		0.413455,
		0.478268,
		0.541604,
		0.602518,
		0.660215,
		0.71404,
		0.763477,
		0.808146,
		0.847798,
		0.882312,
		0.911694,
		0.93607,
		0.955686,
		0.970896,
		0.982156,
		0.990011,
		0.995069,
		0.997978,
		0.999388,
		0.999897,
		0.999997,
		1,
		0.999968,
		0.999707,
		0.998854,
		0.99697,
		0.993593,
		0.988294,
		0.980692,
		0.97048,
		0.957423,
		0.941362,
		0.922214,
		0.899959,
		0.874642,
		0.846361,
		0.815262,
		0.781532,
		0.745396,
		0.008197,
		0.030978,
		0.065798,
		0.110324,
		0.162428,
		0.220175,
		0.281814,
		0.345773,
		0.410647,
		0.475193,
		0.538323,
		0.599101,
		0.656733,
		0.710566,
		0.760084,
		0.804901,
		0.844764,
		0.879542,
		0.909232,
		0.933948,
		0.953921,
		0.969489,
		0.981094,
		0.989263,
		0.99459,
		0.997711,
		0.999268,
		0.999861,
		0.999993,
		1,
		0.99998,
		0.999775,
		0.999046,
		0.997366,
		0.994282,
		0.989361,
		0.98222,
		0.972543,
		0.960083,
		0.944673,
		0.926215,
		0.90468,
		0.880097,
		0.852555,
		0.822187,
		0.789171,
		0.008121,
		0.0307,
		0.065227,
		0.109399,
		0.161115,
		0.218463,
		0.279714,
		0.343311,
		0.407865,
		0.472143,
		0.535067,
		0.595706,
		0.65327,
		0.707107,
		0.756699,
		0.801659,
		0.841725,
		0.876762,
		0.906753,
		0.931803,
		0.952128,
		0.968053,
		0.980001,
		0.988485,
		0.994084,
		0.997422,
		0.999134,
		0.999819,
		0.999988,
		1,
		0.999988,
		0.99983,
		0.999212,
		0.997721,
		0.994912,
		0.990352,
		0.983655,
		0.974494,
		0.962616,
		0.947842,
		0.930061,
		0.909232,
		0.885375,
		0.858565,
		0.828924,
		0.008046,
		0.030427,
		0.064663,
		0.108485,
		0.159817,
		0.21677,
		0.277636,
		0.340874,
		0.405109,
		0.469119,
		0.531836,
		0.592333,
		0.649825,
		0.703661,
		0.753323,
		0.798418,
		0.838682,
		0.87397,
		0.904257,
		0.929636,
		0.950308,
		0.966586,
		0.978877,
		0.987677,
		0.993551,
		0.997111,
		0.998984,
		0.999767,
		0.99998,
		1,
		0.999994,
		0.999874,
		0.999355,
		0.998039,
		0.995488,
		0.991272,
		0.984999,
		0.976338,
		0.965026,
		0.950871,
		0.933754,
		0.913621,
		0.89048,
		0.864394,
		0.007973,
		0.030156,
		0.064107,
		0.107582,
		0.158534,
		0.215097,
		0.27558,
		0.338462,
		0.402378,
		0.466121,
		0.528629,
		0.588982,
		0.646399,
		0.70023,
		0.749955,
		0.795181,
		0.835636,
		0.871169,
		0.901745,
		0.927446,
		0.948463,
		0.96509,
		0.977722,
		0.986838,
		0.992991,
		0.996778,
		0.998818,
		0.999707,
		0.999969,
		1,
		0.999997,
		0.999909,
		0.999477,
		0.998321,
		0.996013,
		0.992122,
		0.986257,
		0.978079,
		0.967315,
		0.953766,
		0.937299,
		0.91785,
		0.895415,
		0.0079,
		0.02989,
		0.063557,
		0.106691,
		0.157267,
		0.213443,
		0.273547,
		0.336073,
		0.399673,
		0.463149,
		0.525446,
		0.585653,
		0.642991,
		0.696813,
		0.746597,
		0.791947,
		0.832587,
		0.868358,
		0.899218,
		0.925236,
		0.946591,
		0.963565,
		0.976536,
		0.98597,
		0.992403,
		0.996422,
		0.998635,
		0.999636,
		0.999954,
		1,
		0.999999,
		0.999936,
		0.999581,
		0.998572,
		0.99649,
		0.992908,
		0.987433,
		0.97972,
		0.969489,
		0.95653,
		0.940699,
		0.921922,
		0.007828,
		0.029626,
		0.063015,
		0.10581,
		0.156015,
		0.211807,
		0.271535,
		0.333709,
		0.396993,
		0.460201,
		0.522288,
		0.582346,
		0.639602,
		0.69341,
		0.743248,
		0.788716,
		0.829535,
		0.865538,
		0.896676,
		0.923006,
		0.944695,
		0.962012,
		0.975321,
		0.985072,
		0.991788,
		0.996042,
		0.998435,
		0.999554,
		0.999934,
		0.999999,
		1,
		0.999957,
		0.999668,
		0.998793,
		0.996922,
		0.993633,
		0.988531,
		0.981267,
		0.971552,
		0.959167,
		0.943959,
		0.007758,
		0.029367,
		0.06248,
		0.104941,
		0.154777,
		0.210189,
		0.269544,
		0.331368,
		0.394338,
		0.457278,
		0.519153,
		0.579062,
		0.636233,
		0.690022,
		0.739908,
		0.78549,
		0.826481,
		0.862711,
		0.894119,
		0.920755,
		0.942774,
		0.960431,
		0.974075,
		0.984144,
		0.991144,
		0.995638,
		0.998216,
		0.999461,
		0.999909,
		0.999997,
		1,
		0.999972,
		0.999741,
		0.998987,
		0.997312,
		0.9943,
		0.989554,
		0.982721,
		0.973507,
		0.961681,
		0.007688,
		0.02911,
		0.061951,
		0.104082,
		0.153554,
		0.208589,
		0.267574,
		0.32905,
		0.391707,
		0.454381,
		0.516043,
		0.575799,
		0.632882,
		0.686649,
		0.736579,
		0.782268,
		0.823426,
		0.859876,
		0.891549,
		0.918486,
		0.94083,
		0.958822,
		0.972801,
		0.983187,
		0.990474,
		0.995211,
		0.997978,
		0.999355,
		0.999878,
		0.999994,
		1,
		0.999982,
		0.999801,
		0.999156,
		0.997663,
		0.994912,
		0.990506,
		0.984089,
		0.975358,
		0.007619,
		0.028857,
		0.061429,
		0.103233,
		0.152345,
		0.207007,
		0.265625,
		0.326755,
		0.3891,
		0.451507,
		0.512956,
		0.572558,
		0.629549,
		0.683291,
		0.73326,
		0.779051,
		0.82037,
		0.857034,
		0.888967,
		0.916199,
		0.938862,
		0.957188,
		0.971498,
		0.9822,
		0.989775,
		0.994759,
		0.997721,
		0.999236,
		0.99984,
		0.999989,
		1,
		0.99999,
		0.99985,
		0.999303,
		0.997978,
		0.995473,
		0.99139,
		0.985372,
		0.007551,
		0.028608,
		0.060913,
		0.102394,
		0.151149,
		0.205443,
		0.263696,
		0.324482,
		0.386518,
		0.448658,
		0.509893,
		0.569338,
		0.626236,
		0.679948,
		0.729952,
		0.77584,
		0.817314,
		0.854185,
		0.886372,
		0.913893,
		0.936872,
		0.955526,
		0.970166,
		0.981184,
		0.989048,
		0.994282,
		0.997444,
		0.999103,
		0.999795,
		0.999982,
		1,
		0.999995,
		0.999889,
		0.999429,
		0.99826,
		0.995985,
		0.99221,
		0.007484,
		0.028361,
		0.060404,
		0.101566,
		0.149968,
		0.203896,
		0.261788,
		0.322233,
		0.383959,
		0.445833,
		0.506853,
		0.56614,
		0.622942,
		0.67662,
		0.726654,
		0.772634,
		0.814257,
		0.851331,
		0.883765,
		0.911571,
		0.93486,
		0.95384,
		0.968806,
		0.98014,
		0.988294,
		0.99378,
		0.997147,
		0.998956,
		0.999741,
		0.999973,
		1,
		0.999997,
		0.99992,
		0.999537,
		0.998511,
		0.996452,
		0.007418,
		0.028118,
		0.059902,
		0.100747,
		0.148801,
		0.202366,
		0.2599,
		0.320005,
		0.381424,
		0.443032,
		0.503836,
		0.562964,
		0.619667,
		0.673308,
		0.723368,
		0.769433,
		0.811201,
		0.848471,
		0.881147,
		0.909232,
		0.932827,
		0.952128,
		0.967419,
		0.979066,
		0.987511,
		0.993253,
		0.996829,
		0.998793,
		0.999679,
		0.999959,
		1,
		0.999999,
		0.999944,
		0.999629,
		0.998733,
		0.007353,
		0.027878,
		0.059405,
		0.099939,
		0.147646,
		0.200852,
		0.258031,
		0.317799,
		0.378911,
		0.440255,
		0.500843,
		0.55981,
		0.61641,
		0.670011,
		0.720092,
		0.76624,
		0.808146,
		0.845607,
		0.878519,
		0.906878,
		0.930774,
		0.950392,
		0.966004,
		0.977965,
		0.986701,
		0.9927,
		0.99649,
		0.998614,
		0.999606,
		0.999942,
		0.999999,
		1,
		0.999962,
		0.999707,
		0.007289,
		0.027641,
		0.058915,
		0.09914,
		0.146505,
		0.199356,
		0.256181,
		0.315615,
		0.376422,
		0.437501,
		0.497872,
		0.556677,
		0.613173,
		0.66673,
		0.716828,
		0.763052,
		0.805092,
		0.842738,
		0.875881,
		0.904508,
		0.9287,
		0.948631,
		0.964563,
		0.976835,
		0.985864,
		0.992122,
		0.996128,
		0.99842,
		0.999523,
		0.99992,
		0.999997,
		1,
		0.999975,
		0.007225,
		0.027406,
		0.05843,
		0.09835,
		0.145377,
		0.197875,
		0.254351,
		0.313452,
		0.373956,
		0.43477,
		0.494924,
		0.553565,
		0.609954,
		0.663465,
		0.713576,
		0.759872,
		0.80204,
		0.839866,
		0.873234,
		0.902123,
		0.926607,
		0.946848,
		0.963095,
		0.975678,
		0.984999,
		0.991519,
		0.995745,
		0.998208,
		0.999429,
		0.999892,
		0.999995,
		1,
		0.007163,
		0.027175,
		0.057952,
		0.09757,
		0.144262,
		0.196411,
		0.25254,
		0.31131,
		0.371512,
		0.432063,
		0.491999,
		0.550474,
		0.606755,
		0.660215,
		0.710335,
		0.756699,
		0.79899,
		0.83699,
		0.870578,
		0.899725,
		0.924495,
		0.945041,
		0.961602,
		0.974494,
		0.984106,
		0.99089,
		0.99534,
		0.997978,
		0.999323,
		0.999859,
		0.999991,
		0.007101,
		0.026947,
		0.057479,
		0.096799,
		0.14316,
		0.194962,
		0.250747,
		0.309189,
		0.369091,
		0.429378,
		0.489096,
		0.547405,
		0.603574,
		0.656981,
		0.707107,
		0.753533,
		0.795942,
		0.834112,
		0.867913,
		0.897313,
		0.922365,
		0.943213,
		0.960083,
		0.973282,
		0.983187,
		0.990235,
		0.994912,
		0.997731,
		0.999205,
		0.999819,
		0.00704,
		0.026722,
		0.057012,
		0.096037,
		0.142069,
		0.193529,
		0.248973,
		0.307089,
		0.366691,
		0.426716,
		0.486216,
		0.544357,
		0.600413,
		0.653763,
		0.70389,
		0.750376,
		0.792898,
		0.83123,
		0.865241,
		0.894888,
		0.920217,
		0.941362,
		0.95854,
		0.972044,
		0.98224,
		0.989554,
		0.994461,
		0.997465,
		0.999074,
		0.006979,
		0.026499,
		0.056551,
		0.095283,
		0.140991,
		0.192112,
		0.247217,
		0.305009,
		0.364313,
		0.424076,
		0.483358,
		0.54133,
		0.59727,
		0.650561,
		0.700686,
		0.747226,
		0.789856,
		0.828347,
		0.862562,
		0.89245,
		0.918052,
		0.939491,
		0.956972,
		0.97078,
		0.981267,
		0.988848,
		0.993987,
		0.99718,
		0.00692,
		0.026279,
		0.056095,
		0.094539,
		0.139926,
		0.19071,
		0.245479,
		0.302949,
		0.361957,
		0.421459,
		0.480522,
		0.538323,
		0.594146,
		0.647376,
		0.697495,
		0.744084,
		0.786818,
		0.825463,
		0.859876,
		0.890001,
		0.91587,
		0.937598,
		0.955381,
		0.969489,
		0.980266,
		0.988116,
		0.99349,
		0.006861,
		0.026062,
		0.055645,
		0.093803,
		0.138872,
		0.189322,
		0.243759,
		0.300909,
		0.359622,
		0.418863,
		0.477707,
		0.535338,
		0.591042,
		0.644206,
		0.694316,
		0.740951,
		0.783784,
		0.822577,
		0.857183,
		0.887541,
		0.913673,
		0.935686,
		0.953766,
		0.968173,
		0.97924,
		0.987358,
		0.006803,
		0.025848,
		0.0552,
		0.093076,
		0.137829,
		0.18795,
		0.242056,
		0.298888,
		0.357308,
		0.416289,
		0.474914,
		0.532373,
		0.587956,
		0.641052,
		0.69115,
		0.737827,
		0.780754,
		0.819691,
		0.854485,
		0.88507,
		0.91146,
		0.933754,
		0.952128,
		0.966832,
		0.978187,
		0.006746,
		0.025636,
		0.054761,
		0.092357,
		0.136799,
		0.186592,
		0.24037,
		0.296887,
		0.355015,
		0.413737,
		0.472143,
		0.529428,
		0.584888,
		0.637915,
		0.687996,
		0.734711,
		0.777728,
		0.816804,
		0.851782,
		0.882588,
		0.909232,
		0.931803,
		0.950468,
		0.965467,
		0.006689,
		0.025427,
		0.054327,
		0.091646,
		0.135779,
		0.185248,
		0.238701,
		0.294906,
		0.352743,
		0.411206,
		0.469393,
		0.526505,
		0.58184,
		0.634794,
		0.684856,
		0.731605,
		0.774707,
		0.813918,
		0.849074,
		0.880097,
		0.90699,
		0.929834,
		0.948785,
		0.006633,
		0.02522,
		0.053897,
		0.090944,
		0.134771,
		0.183919,
		0.237049,
		0.292943,
		0.350491,
		0.408697,
		0.466665,
		0.523601,
		0.57881,
		0.631689,
		0.681729,
		0.728508,
		0.771692,
		0.811031,
		0.846361,
		0.877597,
		0.904734,
		0.927846,
		0.006578,
		0.025016,
		0.053473,
		0.090249,
		0.133774,
		0.182603,
		0.235414,
		0.290998,
		0.34826,
		0.406208,
		0.463957,
		0.520718,
		0.575799,
		0.628601,
		0.678615,
		0.72542,
		0.768681,
		0.808146,
		0.843645,
		0.875088,
		0.902465,
		0.006524,
		0.024814,
		0.053054,
		0.089563,
		0.132788,
		0.181302,
		0.233795,
		0.289073,
		0.346048,
		0.403741,
		0.46127,
		0.517854,
		0.572806,
		0.625529,
		0.675515,
		0.722343,
		0.765677,
		0.805262,
		0.840925,
		0.872571,
		0.00647,
		0.024615,
		0.05264,
		0.088884,
		0.131812,
		0.180014,
		0.232193,
		0.287165,
		0.343856,
		0.401293,
		0.458604,
		0.515011,
		0.569832,
		0.622473,
		0.672427,
		0.719275,
		0.762678,
		0.802379,
		0.838201,
		0.006417,
		0.024418,
		0.05223,
		0.088212,
		0.130847,
		0.178739,
		0.230606,
		0.285276,
		0.341684,
		0.398867,
		0.455958,
		0.512188,
		0.566876,
		0.619433,
		0.669354,
		0.716218,
		0.759685,
		0.799498,
		0.006364,
		0.024223,
		0.051826,
		0.087549,
		0.129893,
		0.177478,
		0.229035,
		0.283404,
		0.339531,
		0.39646,
		0.453333,
		0.509384,
		0.563939,
		0.61641,
		0.666294,
		0.71317,
		0.756699,
		0.006313,
		0.024031,
		0.051425,
		0.086892,
		0.128949,
		0.17623,
		0.227479,
		0.28155,
		0.337397,
		0.394074,
		0.450728,
		0.5066,
		0.561021,
		0.613403,
		0.663248,
		0.710133,
		0.006261,
		0.023841,
		0.05103,
		0.086243,
		0.128015,
		0.174994,
		0.225939,
		0.279714,
		0.335282,
		0.391707,
		0.448143,
		0.503836,
		0.55812,
		0.610413,
		0.660215,
		0.006211,
		0.023653,
		0.050639,
		0.085601,
		0.127091,
		0.173771,
		0.224414,
		0.277895,
		0.333186,
		0.38936,
		0.445578,
		0.501091,
		0.555238,
		0.607439,
		0.006161,
		0.023468,
		0.050253,
		0.084966,
		0.126177,
		0.172561,
		0.222905,
		0.276092,
		0.331109,
		0.387032,
		0.443032,
		0.498366,
		0.552374,
		0.006111,
		0.023285,
		0.049871,
		0.084339,
		0.125272,
		0.171364,
		0.221409,
		0.274307,
		0.32905,
		0.384724,
		0.440506,
		0.495659,
		0.006063,
		0.023103,
		0.049493,
		0.083718,
		0.124377,
		0.170178,
		0.219929,
		0.272538,
		0.327009,
		0.382435,
		0.438,
		0.006014,
		0.022924,
		0.049119,
		0.083104,
		0.123492,
		0.169005,
		0.218463,
		0.270786,
		0.324985,
		0.380165,
		0.005967,
		0.022747,
		0.04875,
		0.082496,
		0.122616,
		0.167843,
		0.217011,
		0.269049,
		0.32298,
		0.00592,
		0.022572,
		0.048385,
		0.081895,
		0.121749,
		0.166693,
		0.215573,
		0.267329,
		0.005873,
		0.022399,
		0.048024,
		0.081301,
		0.120891,
		0.165555,
		0.214149,
		0.005827,
		0.022228,
		0.047667,
		0.080713,
		0.120042,
		0.164428,
		0.005782,
		0.022059,
		0.047314,
		0.080132,
		0.119202,
		0.005737,
		0.021892,
		0.046964,
		0.079556,
		0.005692,
		0.021727,
		0.046619,
		0.005649,
		0.021564,
		0.005605,
		0.005562,
		9.1e-005,
		0.021402,
		0.00706,
		0.000363,
		0.046277,
		0.024223,
		0.008741,
		0.00082,
		0.078987,
		0.05033,
		0.027226,
		0.010606,
		0.001462,
		0.118371,
		0.084169,
		0.05456,
		0.030412,
		0.012657,
		0.002292,
		0.163312,
		0.124573,
		0.089522,
		0.058969,
		0.033782,
		0.014896,
		0.00331,
		0.212739,
		0.170419,
		0.130939,
		0.095049,
		0.063557,
		0.037337,
		0.017323,
		0.004519,
		0.265625,
		0.220628,
		0.177679,
		0.137468,
		0.100747,
		0.068326,
		0.041077,
		0.019939,
		0.00592,
		9.6e-005,
		0.320992,
		0.27417,
		0.228658,
		0.185091,
		0.144161,
		0.106619,
		0.073275,
		0.045004,
		0.022747,
		0.007514,
		0.000387,
		0.377913,
		0.330063,
		0.282841,
		0.236827,
		0.192655,
		0.151018,
		0.112664,
		0.078406,
		0.049119,
		0.025748,
		0.009304,
		0.000874,
		0.435513,
		0.387376,
		0.339242,
		0.291635,
		0.245133,
		0.200371,
		0.158037,
		0.118882,
		0.083718,
		0.053423,
		0.028941,
		0.01129,
		0.001559,
		0.492972,
		0.445233,
		0.396928,
		0.348527,
		0.300551,
		0.253577,
		0.208236,
		0.165218,
		0.125272,
		0.089212,
		0.057915,
		0.03233,
		0.013475,
		0.002443,
		0.549528,
		0.502815,
		0.455022,
		0.406567,
		0.357915,
		0.309587,
		0.262156,
		0.216251,
		0.172561,
		0.131836,
		0.094888,
		0.062598,
		0.035914,
		0.015859,
		0.003529,
		0.604481,
		0.559361,
		0.512705,
		0.464877,
		0.416289,
		0.367405,
		0.318742,
		0.270869,
		0.224414,
		0.180066,
		0.138573,
		0.100747,
		0.067471,
		0.039695,
		0.018444,
		0.004818,
		0.657196,
		0.614173,
		0.569217,
		0.522638,
		0.474793,
		0.426093,
		0.376995,
		0.328012,
		0.279714,
		0.232725,
		0.187731,
		0.145482,
		0.106789,
		0.072536,
		0.043674,
		0.021232,
		0.006313,
		0.000103,
		0.707107,
		0.666622,
		0.623865,
		0.579092,
		0.532609,
		0.484769,
		0.435974,
		0.386681,
		0.337397,
		0.28869,
		0.241182,
		0.195557,
		0.152563,
		0.113014,
		0.077793,
		0.047852,
		0.024223,
		0.008014,
		0.000414,
		0.753719,
		0.716147,
		0.676022,
		0.63355,
		0.588982,
		0.542616,
		0.494799,
		0.44593,
		0.39646,
		0.346894,
		0.297795,
		0.249783,
		0.203541,
		0.159817,
		0.119423,
		0.083242,
		0.05223,
		0.02742,
		0.009923,
		0.000934,
		0.796619,
		0.762263,
		0.725136,
		0.685391,
		0.643226,
		0.598882,
		0.552654,
		0.504881,
		0.455958,
		0.406331,
		0.3565,
		0.307026,
		0.258527,
		0.211684,
		0.167242,
		0.126014,
		0.088884,
		0.056809,
		0.030823,
		0.012042,
		0.001665,
		0.835475,
		0.804566,
		0.770731,
		0.73407,
		0.694725,
		0.652886,
		0.608789,
		0.562719,
		0.515011,
		0.466054,
		0.416289,
		0.366214,
		0.316383,
		0.267414,
		0.219983,
		0.174837,
		0.132788,
		0.094719,
		0.061589,
		0.034434,
		0.014374,
		0.00261,
		0.870045,
		0.842738,
		0.812415,
		0.779121,
		0.742944,
		0.704019,
		0.662527,
		0.618696,
		0.572806,
		0.525185,
		0.476215,
		0.426333,
		0.376031,
		0.325863,
		0.27644,
		0.228439,
		0.182603,
		0.139745,
		0.100747,
		0.066571,
		0.038254,
		0.016918,
		0.003771,
		0.900183,
		0.876554,
		0.849883,
		0.820162,
		0.787425,
		0.751752,
		0.713267,
		0.672143,
		0.628601,
		0.582912,
		0.535398,
		0.486437,
		0.436458,
		0.385951,
		0.335463,
		0.285604,
		0.237049,
		0.190539,
		0.146884,
		0.10697,
		0.071756,
		0.042284,
		0.019678,
		0.005149,
		0.925841,
		0.905884,
		0.882926,
		0.856905,
		0.827802,
		0.79564,
		0.760489,
		0.722464,
		0.681729,
		0.638497,
		0.593031,
		0.545648,
		0.496716,
		0.446662,
		0.395969,
		0.345181,
		0.294906,
		0.245813,
		0.198643,
		0.154206,
		0.113386,
		0.077145,
		0.046525,
		0.022654,
		0.006746,
		0.00011,
		0.947082,
		0.930704,
		0.911434,
		0.889157,
		0.863799,
		0.835329,
		0.803759,
		0.769149,
		0.731605,
		0.69128,
		0.64838,
		0.60316,
		0.555928,
		0.507048,
		0.45694,
		0.406083,
		0.355015,
		0.304341,
		0.254728,
		0.206915,
		0.16171,
		0.119997,
		0.082738,
		0.050979,
		0.025848,
		0.008564,
		0.000443,
		0.964076,
		0.951098,
		0.935405,
		0.916829,
		0.895243,
		0.870561,
		0.842738,
		0.811778,
		0.777728,
		0.740684,
		0.700791,
		0.658245,
		0.613292,
		0.566235,
		0.51743,
		0.46729,
		0.416289,
		0.364963,
		0.313909,
		0.263794,
		0.215353,
		0.169394,
		0.126801,
		0.088536,
		0.055645,
		0.029261,
		0.010606,
		0.000999,
		0.977109,
		0.967266,
		0.954949,
		0.939942,
		0.922066,
		0.90118,
		0.877185,
		0.850025,
		0.819691,
		0.78622,
		0.749696,
		0.710256,
		0.668086,
		0.623424,
		0.576564,
		0.527856,
		0.477707,
		0.426585,
		0.37502,
		0.323607,
		0.273008,
		0.223957,
		0.17726,
		0.133799,
		0.094539,
		0.060526,
		0.032895,
		0.012873,
		0.001783,
		0.986574,
		0.979518,
		0.970291,
		0.958631,
		0.944313,
		0.927141,
		0.906963,
		0.883666,
		0.857183,
		0.827492,
		0.794618,
		0.758636,
		0.71967,
		0.677898,
		0.63355,
		0.586911,
		0.538323,
		0.488188,
		0.436966,
		0.385184,
		0.333432,
		0.282369,
		0.232725,
		0.185305,
		0.140991,
		0.100747,
		0.065621,
		0.036751,
		0.015366,
		0.002795,
		0.992969,
		0.988279,
		0.981771,
		0.97315,
		0.962145,
		0.948514,
		0.932051,
		0.912588,
		0.890001,
		0.864209,
		0.835177,
		0.802918,
		0.767497,
		0.729027,
		0.687676,
		0.643665,
		0.59727,
		0.548826,
		0.498728,
		0.44743,
		0.395453,
		0.343382,
		0.291874,
		0.241656,
		0.193529,
		0.148377,
		0.107161,
		0.070933,
		0.04083,
		0.018088,
		0.004038,
		0.996876,
		0.994072,
		0.989843,
		0.983869,
		0.975845,
		0.965488,
		0.952544,
		0.936792,
		0.918052,
		0.896184,
		0.871096,
		0.842738,
		0.811114,
		0.776274,
		0.738322,
		0.697414,
		0.653763,
		0.607637,
		0.559361,
		0.509323,
		0.457972,
		0.405822,
		0.353455,
		0.301522,
		0.250747,
		0.201932,
		0.155955,
		0.113781,
		0.076461,
		0.045134,
		0.02104,
		0.005514,
		0.998929,
		0.997501,
		0.995056,
		0.991269,
		0.985814,
		0.978376,
		0.968661,
		0.956401,
		0.941362,
		0.92335,
		0.902212,
		0.877839,
		0.850172,
		0.8192,
		0.784961,
		0.747548,
		0.707107,
		0.66384,
		0.618005,
		0.569922,
		0.519969,
		0.468589,
		0.416289,
		0.363647,
		0.31131,
		0.259999,
		0.210511,
		0.163726,
		0.120607,
		0.082205,
		0.049664,
		0.024223,
		0.007225,
		0.000118,
		0.999771,
		0.999217,
		0.998037,
		0.995927,
		0.992561,
		0.987608,
		0.980745,
		0.971663,
		0.960083,
		0.945759,
		0.928479,
		0.908079,
		0.884434,
		0.857473,
		0.827169,
		0.793551,
		0.756699,
		0.716747,
		0.673888,
		0.628371,
		0.580504,
		0.530661,
		0.479275,
		0.42685,
		0.373956,
		0.321236,
		0.269407,
		0.219266,
		0.171689,
		0.127639,
		0.088168,
		0.054421,
		0.027641,
		0.009174,
		0.000475,
		0.999985,
		0.999861,
		0.999446,
		0.998488,
		0.99669,
		0.993724,
		0.989256,
		0.982951,
		0.974494,
		0.96359,
		0.949978,
		0.933436,
		0.913781,
		0.890876,
		0.864634,
		0.835018,
		0.80204,
		0.76577,
		0.726331,
		0.683903,
		0.638727,
		0.591103,
		0.541394,
		0.490028,
		0.437501,
		0.384378,
		0.331297,
		0.278972,
		0.228195,
		0.179843,
		0.134877,
		0.094348,
		0.059405,
		0.031293,
		0.011362,
		0.001072,
		1,
		0.999995,
		0.999922,
		0.999624,
		0.998863,
		0.997351,
		0.994763,
		0.990759,
		0.984999,
		0.977154,
		0.96692,
		0.95402,
		0.938217,
		0.919314,
		0.897159,
		0.871652,
		0.842738,
		0.810421,
		0.774754,
		0.73585,
		0.693878,
		0.649069,
		0.601712,
		0.552163,
		0.500843,
		0.448238,
		0.39491,
		0.34149,
		0.28869,
		0.237298,
		0.188187,
		0.14232,
		0.100747,
		0.064618,
		0.035182,
		0.013791,
		0.001913,
		0.999985,
		1,
		0.999999,
		0.999961,
		0.999757,
		0.999169,
		0.997917,
		0.995683,
		0.992122,
		0.986889,
		0.979645,
		0.970072,
		0.957881,
		0.942819,
		0.924674,
		0.90328,
		0.878519,
		0.850326,
		0.818687,
		0.783645,
		0.745299,
		0.703808,
		0.65939,
		0.612327,
		0.562964,
		0.511714,
		0.459058,
		0.405548,
		0.351814,
		0.298559,
		0.246571,
		0.196721,
		0.149968,
		0.107365,
		0.070061,
		0.039308,
		0.016464,
		0.003,
		0.999771,
		0.999962,
		0.999999,
		1,
		0.999984,
		0.999852,
		0.999412,
		0.998396,
		0.99649,
		0.99335,
		0.988626,
		0.981968,
		0.973047,
		0.96156,
		0.947239,
		0.929857,
		0.909232,
		0.885232,
		0.857774,
		0.826833,
		0.792437,
		0.754672,
		0.713686,
		0.669685,
		0.622942,
		0.573791,
		0.522638,
		0.469955,
		0.416289,
		0.362263,
		0.308577,
		0.256014,
		0.205443,
		0.157821,
		0.114202,
		0.075735,
		0.043674,
		0.019382,
		0.004334,
		0.998929,
		0.999635,
		0.999922,
		0.999995,
		1,
		0.999995,
		0.999917,
		0.9996,
		0.998793,
		0.997189,
		0.994448,
		0.990212,
		0.984125,
		0.975845,
		0.965057,
		0.951476,
		0.93486,
		0.915012,
		0.891784,
		0.865077,
		0.834852,
		0.801123,
		0.763962,
		0.723506,
		0.679948,
		0.63355,
		0.584639,
		0.533608,
		0.480925,
		0.427129,
		0.372836,
		0.318742,
		0.265625,
		0.214352,
		0.165879,
		0.121258,
		0.08164,
		0.048281,
		0.022547,
		0.00592,
		0.996876,
		0.998532,
		0.999446,
		0.999857,
		0.999984,
		1,
		0.999999,
		0.999959,
		0.999741,
		0.999117,
		0.997789,
		0.995421,
		0.991651,
		0.986118,
		0.978467,
		0.968369,
		0.955526,
		0.93968,
		0.920616,
		0.89817,
		0.87223,
		0.842738,
		0.809696,
		0.773163,
		0.73326,
		0.690172,
		0.644146,
		0.595501,
		0.544621,
		0.491965,
		0.438064,
		0.383529,
		0.32905,
		0.275401,
		0.223446,
		0.17414,
		0.128533,
		0.087777,
		0.05313,
		0.025962,
		0.007758,
		0.000127,
		0.992969,
		0.996042,
		0.998037,
		0.999193,
		0.999757,
		0.99996,
		0.999999,
		1,
		0.999982,
		0.999843,
		0.999375,
		0.998296,
		0.996274,
		0.992949,
		0.987949,
		0.980913,
		0.971498,
		0.959389,
		0.944313,
		0.926039,
		0.904386,
		0.879226,
		0.850486,
		0.818151,
		0.782268,
		0.742944,
		0.70035,
		0.654724,
		0.606372,
		0.555671,
		0.503068,
		0.449089,
		0.394338,
		0.339498,
		0.285341,
		0.232725,
		0.182603,
		0.136026,
		0.094146,
		0.058222,
		0.029626,
		0.009851,
		0.000511,
		0.986574,
		0.99151,
		0.995056,
		0.997428,
		0.998863,
		0.999612,
		0.999917,
		0.999994,
		1,
		0.999994,
		0.999912,
		0.999575,
		0.998717,
		0.997015,
		0.994109,
		0.989624,
		0.983187,
		0.974442,
		0.963062,
		0.948756,
		0.931276,
		0.910426,
		0.886059,
		0.858088,
		0.826481,
		0.79127,
		0.752549,
		0.710476,
		0.665277,
		0.617247,
		0.566751,
		0.51423,
		0.460201,
		0.405259,
		0.350084,
		0.295441,
		0.242186,
		0.191269,
		0.143738,
		0.100747,
		0.063557,
		0.033544,
		0.012202,
		0.001154,
		0.977109,
		0.984305,
		0.989843,
		0.993901,
		0.99669,
		0.998443,
		0.999412,
		0.999848,
		0.999982,
		1,
		0.999999,
		0.999956,
		0.999725,
		0.999061,
		0.997651,
		0.995138,
		0.991144,
		0.985289,
		0.977203,
		0.966545,
		0.953007,
		0.936326,
		0.916286,
		0.892725,
		0.865538,
		0.834679,
		0.800163,
		0.762069,
		0.720543,
		0.675798,
		0.628118,
		0.577857,
		0.525446,
		0.471394,
		0.416289,
		0.360804,
		0.3057,
		0.251827,
		0.200135,
		0.151669,
		0.107582,
		0.069138,
		0.037715,
		0.014812,
		0.002059,
		0.964076,
		0.973858,
		0.981771,
		0.987949,
		0.992561,
		0.995807,
		0.997917,
		0.999143,
		0.999741,
		0.999957,
		0.999999,
		1,
		0.999981,
		0.999833,
		0.999334,
		0.998188,
		0.996042,
		0.992516,
		0.987222,
		0.979781,
		0.969836,
		0.957064,
		0.941183,
		0.921961,
		0.899218,
		0.872831,
		0.842738,
		0.808939,
		0.771497,
		0.730545,
		0.686282,
		0.638979,
		0.588982,
		0.536711,
		0.482664,
		0.427424,
		0.371655,
		0.316114,
		0.261647,
		0.2092,
		0.159817,
		0.114651,
		0.074965,
		0.042142,
		0.017685,
		0.003228,
		0.947082,
		0.959694,
		0.970291,
		0.978957,
		0.985814,
		0.991018,
		0.994763,
		0.997272,
		0.998793,
		0.999588,
		0.999912,
		0.999994,
		1,
		0.999994,
		0.999906,
		0.999547,
		0.998635,
		0.996827,
		0.993744,
		0.988991,
		0.982179,
		0.972936,
		0.960924,
		0.945845,
		0.927446,
		0.905532,
		0.87996,
		0.850653,
		0.817592,
		0.780826,
		0.740474,
		0.69672,
		0.649825,
		0.60012,
		0.548017,
		0.494005,
		0.438658,
		0.382633,
		0.326681,
		0.271644,
		0.218463,
		0.168182,
		0.121952,
		0.081039,
		0.046826,
		0.020821,
		0.004665,
		0.925841,
		0.941445,
		0.954949,
		0.966392,
		0.975845,
		0.983418,
		0.989256,
		0.993541,
		0.99649,
		0.998347,
		0.999375,
		0.999838,
		0.999981,
		1,
		0.999999,
		0.999953,
		0.999707,
		0.999,
		0.997501,
		0.994834,
		0.990599,
		0.984397,
		0.975845,
		0.964588,
		0.950308,
		0.932739,
		0.911662,
		0.88692,
		0.858415,
		0.826114,
		0.790049,
		0.750323,
		0.707107,
		0.660647,
		0.611266,
		0.559361,
		0.505413,
		0.449987,
		0.393734,
		0.337397,
		0.281814,
		0.227922,
		0.176763,
		0.129487,
		0.08736,
		0.051768,
		0.024223,
		0.006372,
		0.900183,
		0.918855,
		0.935405,
		0.949827,
		0.962145,
		0.972419,
		0.980745,
		0.987255,
		0.992122,
		0.995554,
		0.997789,
		0.999089,
		0.999725,
		0.999955,
		0.999999,
		1,
		0.99998,
		0.999822,
		0.999291,
		0.998072,
		0.995792,
		0.99205,
		0.98644,
		0.978564,
		0.968053,
		0.954571,
		0.937833,
		0.917603,
		0.893703,
		0.866019,
		0.834498,
		0.799158,
		0.760084,
		0.717434,
		0.67144,
		0.622411,
		0.570735,
		0.516882,
		0.461407,
		0.404954,
		0.34826,
		0.292156,
		0.237576,
		0.185559,
		0.137255,
		0.09393,
		0.05697,
		0.027894,
		0.008351,
		0.000137,
		0.870045,
		0.891784,
		0.911434,
		0.92895,
		0.944313,
		0.957534,
		0.968661,
		0.977776,
		0.984999,
		0.99049,
		0.994448,
		0.997104,
		0.998717,
		0.999561,
		0.999906,
		0.999994,
		1,
		0.999993,
		0.9999,
		0.999517,
		0.998546,
		0.996625,
		0.99335,
		0.98831,
		0.981094,
		0.971319,
		0.958631,
		0.942727,
		0.92335,
		0.900305,
		0.873458,
		0.842738,
		0.808146,
		0.769751,
		0.727694,
		0.682196,
		0.63355,
		0.582134,
		0.528406,
		0.472913,
		0.416289,
		0.359265,
		0.302666,
		0.247423,
		0.19457,
		0.145256,
		0.100747,
		0.062433,
		0.031834,
		0.010606,
		0.000551,
		0.835475,
		0.860201,
		0.882926,
		0.903567,
		0.922066,
		0.938392,
		0.952544,
		0.964555,
		0.974494,
		0.982468,
		0.988626,
		0.993153,
		0.996274,
		0.998243,
		0.999334,
		0.999827,
		0.99998,
		1,
		0.999999,
		0.99995,
		0.999687,
		0.998934,
		0.99734,
		0.994505,
		0.990011,
		0.983438,
		0.974386,
		0.962487,
		0.947416,
		0.928899,
		0.90672,
		0.880725,
		0.850827,
		0.817006,
		0.779315,
		0.737881,
		0.692907,
		0.644676,
		0.59355,
		0.539979,
		0.484498,
		0.427734,
		0.370409,
		0.313342,
		0.25746,
		0.203793,
		0.15349,
		0.107814,
		0.068159,
		0.036046,
		0.013138,
		0.001245,
		0.796619,
		0.824185,
		0.849883,
		0.873601,
		0.895243,
		0.914741,
		0.932051,
		0.947158,
		0.960083,
		0.970881,
		0.979645,
		0.98651,
		0.991651,
		0.995282,
		0.997651,
		0.999031,
		0.999707,
		0.999952,
		0.999999,
		1,
		0.999979,
		0.99981,
		0.999244,
		0.997946,
		0.995521,
		0.991548,
		0.985597,
		0.977255,
		0.966137,
		0.951898,
		0.934245,
		0.912942,
		0.887815,
		0.858756,
		0.825729,
		0.788769,
		0.747985,
		0.703566,
		0.655781,
		0.604978,
		0.551596,
		0.496159,
		0.439285,
		0.381687,
		0.324181,
		0.267685,
		0.213228,
		0.161955,
		0.11513,
		0.074147,
		0.040531,
		0.015951,
		0.002221,
		0.753719,
		0.783913,
		0.812415,
		0.839085,
		0.863799,
		0.886452,
		0.906963,
		0.925275,
		0.941362,
		0.95523,
		0.96692,
		0.976511,
		0.984125,
		0.989923,
		0.994109,
		0.996923,
		0.998635,
		0.999532,
		0.9999,
		0.999993,
		1,
		0.999993,
		0.999893,
		0.999485,
		0.998451,
		0.996405,
		0.992925,
		0.987575,
		0.979927,
		0.96958,
		0.95617,
		0.939384,
		0.918966,
		0.89472,
		0.86652,
		0.834309,
		0.798105,
		0.758,
		0.714166,
		0.666857,
		0.61641,
		0.563249,
		0.507888,
		0.450936,
		0.393097,
		0.335179,
		0.278096,
		0.222873,
		0.170651,
		0.122695,
		0.080399,
		0.045292,
		0.019046,
		0.003484,
		0.707107,
		0.739658,
		0.770731,
		0.800163,
		0.827802,
		0.853514,
		0.877185,
		0.89872,
		0.918052,
		0.935141,
		0.949978,
		0.962592,
		0.973047,
		0.98145,
		0.987949,
		0.992736,
		0.996042,
		0.998131,
		0.999291,
		0.999816,
		0.999979,
		1,
		0.999999,
		0.999947,
		0.999666,
		0.998863,
		0.997165,
		0.99415,
		0.989376,
		0.982404,
		0.972816,
		0.96023,
		0.944313,
		0.924786,
		0.901435,
		0.874111,
		0.842738,
		0.807315,
		0.767917,
		0.724699,
		0.677898,
		0.627839,
		0.574932,
		0.51968,
		0.462682,
		0.404632,
		0.346332,
		0.28869,
		0.232725,
		0.179577,
		0.130509,
		0.086916,
		0.05033,
		0.022426,
		0.005035,
		0.657196,
		0.691782,
		0.725136,
		0.757075,
		0.787425,
		0.816028,
		0.842738,
		0.867431,
		0.890001,
		0.910368,
		0.928479,
		0.944313,
		0.957881,
		0.969235,
		0.978467,
		0.985711,
		0.991144,
		0.994989,
		0.997501,
		0.998968,
		0.999687,
		0.999948,
		0.999999,
		1,
		0.999977,
		0.999797,
		0.999193,
		0.997809,
		0.995229,
		0.991005,
		0.984688,
		0.975845,
		0.964076,
		0.949027,
		0.930398,
		0.907953,
		0.881522,
		0.851009,
		0.816392,
		0.777728,
		0.735156,
		0.688896,
		0.639258,
		0.586639,
		0.531529,
		0.474517,
		0.416289,
		0.357638,
		0.299464,
		0.242783,
		0.188732,
		0.138573,
		0.093699,
		0.055645,
		0.026093,
		0.006878,
		0.604481,
		0.64073,
		0.676022,
		0.710157,
		0.742944,
		0.774201,
		0.803759,
		0.831466,
		0.857183,
		0.880796,
		0.902212,
		0.921364,
		0.938217,
		0.952769,
		0.965057,
		0.975156,
		0.983187,
		0.989313,
		0.993744,
		0.996728,
		0.998546,
		0.999501,
		0.999893,
		0.999993,
		1,
		0.999993,
		0.999886,
		0.99945,
		0.998347,
		0.996168,
		0.992466,
		0.986782,
		0.978669,
		0.967708,
		0.953525,
		0.935798,
		0.914268,
		0.888746,
		0.859113,
		0.825327,
		0.787425,
		0.745529,
		0.699842,
		0.650658,
		0.598361,
		0.543429,
		0.486437,
		0.428063,
		0.369091,
		0.310415,
		0.253045,
		0.198115,
		0.146884,
		0.100747,
		0.061241,
		0.030049,
		0.009016,
		0.000148,
		0.549528,
		0.587025,
		0.623865,
		0.659834,
		0.694725,
		0.72834,
		0.760489,
		0.790993,
		0.819691,
		0.846434,
		0.871096,
		0.893571,
		0.913781,
		0.931675,
		0.947239,
		0.960493,
		0.971498,
		0.980357,
		0.987222,
		0.992288,
		0.995792,
		0.99801,
		0.999244,
		0.999803,
		0.999977,
		1,
		0.999998,
		0.999943,
		0.999643,
		0.998786,
		0.996976,
		0.993766,
		0.988691,
		0.981289,
		0.971124,
		0.957802,
		0.94098,
		0.920376,
		0.895777,
		0.867043,
		0.834112,
		0.797,
		0.75581,
		0.710728,
		0.662033,
		0.610092,
		0.555372,
		0.498434,
		0.439948,
		0.380687,
		0.321539,
		0.263508,
		0.207724,
		0.155444,
		0.108062,
		0.067117,
		0.034297,
		0.011451,
		0.000596,
		0.492972,
		0.531263,
		0.569217,
		0.606611,
		0.643226,
		0.678846,
		0.713267,
		0.746289,
		0.777728,
		0.80741,
		0.835177,
		0.86089,
		0.884434,
		0.905717,
		0.924674,
		0.941275,
		0.955526,
		0.967473,
		0.977203,
		0.984851,
		0.990599,
		0.994673,
		0.99734,
		0.9989,
		0.999666,
		0.999945,
		0.999998,
		1,
		0.999976,
		0.999782,
		0.999138,
		0.997662,
		0.994912,
		0.990418,
		0.983707,
		0.974325,
		0.961859,
		0.945942,
		0.92627,
		0.902608,
		0.874792,
		0.842738,
		0.806444,
		0.76599,
		0.721547,
		0.673374,
		0.621825,
		0.567351,
		0.510503,
		0.451939,
		0.392422,
		0.332833,
		0.27417,
		0.217558,
		0.164251,
		0.115644,
		0.073275,
		0.038838,
		0.014187,
		0.001347,
		0.435513,
		0.474108,
		0.512705,
		0.551073,
		0.588982,
		0.626207,
		0.662527,
		0.697728,
		0.731605,
		0.763962,
		0.794618,
		0.823405,
		0.850172,
		0.874792,
		0.897159,
		0.917197,
		0.93486,
		0.950139,
		0.963062,
		0.973702,
		0.982179,
		0.988657,
		0.99335,
		0.996517,
		0.998451,
		0.999468,
		0.999886,
		0.999992,
		1,
		0.999992,
		0.999878,
		0.999412,
		0.998234,
		0.995911,
		0.991969,
		0.985926,
		0.977312,
		0.965692,
		0.950679,
		0.931946,
		0.909232,
		0.882352,
		0.851199,
		0.815748,
		0.776061,
		0.732288,
		0.684672,
		0.63355,
		0.579359,
		0.522638,
		0.46403,
		0.404291,
		0.344293,
		0.285028,
		0.227614,
		0.173305,
		0.123492,
		0.079717,
		0.043674,
		0.017226,
		0.002404,
		0.377913,
		0.416289,
		0.455022,
		0.493875,
		0.532609,
		0.570991,
		0.608789,
		0.645775,
		0.681729,
		0.716437,
		0.749696,
		0.781314,
		0.811114,
		0.838934,
		0.864634,
		0.888097,
		0.909232,
		0.927979,
		0.944313,
		0.958246,
		0.969836,
		0.979184,
		0.98644,
		0.991804,
		0.995521,
		0.997879,
		0.999193,
		0.99979,
		0.999976,
		1,
		0.999998,
		0.999939,
		0.999618,
		0.998703,
		0.996771,
		0.99335,
		0.987949,
		0.980084,
		0.969301,
		0.955189,
		0.937398,
		0.915644,
		0.889716,
		0.859486,
		0.824905,
		0.786014,
		0.742944,
		0.695919,
		0.645261,
		0.591389,
		0.53483,
		0.476215,
		0.416289,
		0.355915,
		0.296078,
		0.237891,
		0.182603,
		0.131607,
		0.086442,
		0.048808,
		0.02057,
		0.003771,
		0.320992,
		0.358598,
		0.396928,
		0.435741,
		0.474793,
		0.513845,
		0.552654,
		0.590984,
		0.628601,
		0.665277,
		0.700791,
		0.734932,
		0.767497,
		0.7983,
		0.827169,
		0.853952,
		0.878519,
		0.900765,
		0.920616,
		0.938031,
		0.953007,
		0.965583,
		0.975845,
		0.983926,
		0.990011,
		0.994331,
		0.997165,
		0.998826,
		0.999643,
		0.999941,
		0.999998,
		1,
		0.999974,
		0.999767,
		0.999078,
		0.997501,
		0.994569,
		0.989783,
		0.982646,
		0.972685,
		0.959469,
		0.942622,
		0.921837,
		0.896877,
		0.86759,
		0.833904,
		0.795839,
		0.753505,
		0.707107,
		0.656948,
		0.603434,
		0.547074,
		0.488488,
		0.428411,
		0.367695,
		0.307317,
		0.248386,
		0.192146,
		0.139988,
		0.093452,
		0.05424,
		0.024223,
		0.00545,
		0.265625,
		0.301882,
		0.339242,
		0.377459,
		0.416289,
		0.455485,
		0.494799,
		0.533987,
		0.572806,
		0.611015,
		0.64838,
		0.684672,
		0.71967,
		0.753166,
		0.784961,
		0.814873,
		0.842738,
		0.868415,
		0.891784,
		0.912756,
		0.931276,
		0.947325,
		0.960924,
		0.972142,
		0.981094,
		0.987949,
		0.992925,
		0.996289,
		0.998347,
		0.999431,
		0.999878,
		0.999992,
		1,
		0.999991,
		0.999869,
		0.99937,
		0.998111,
		0.995632,
		0.991431,
		0.984999,
		0.975845,
		0.963517,
		0.947615,
		0.927805,
		0.903828,
		0.875503,
		0.842738,
		0.805528,
		0.763962,
		0.718226,
		0.668603,
		0.615483,
		0.559361,
		0.500843,
		0.44065,
		0.379627,
		0.318742,
		0.259096,
		0.201932,
		0.148635,
		0.100747,
		0.059973,
		0.028187,
		0.007447,
		0.212739,
		0.247045,
		0.282841,
		0.31988,
		0.357915,
		0.396697,
		0.435974,
		0.475496,
		0.515011,
		0.554272,
		0.593031,
		0.631048,
		0.668086,
		0.703917,
		0.738322,
		0.771094,
		0.80204,
		0.830985,
		0.857774,
		0.882275,
		0.904386,
		0.924033,
		0.941183,
		0.955839,
		0.968053,
		0.977922,
		0.985597,
		0.991281,
		0.995229,
		0.997737,
		0.999138,
		0.999775,
		0.999974,
		1,
		0.999998,
		0.999934,
		0.999591,
		0.998612,
		0.996548,
		0.992899,
		0.987147,
		0.978782,
		0.967332,
		0.952372,
		0.933544,
		0.910561,
		0.883219,
		0.851398,
		0.815072,
		0.774307,
		0.729267,
		0.680218,
		0.62753,
		0.571684,
		0.513271,
		0.453002,
		0.391707,
		0.330348,
		0.27002,
		0.211958,
		0.157547,
		0.108329,
		0.066007,
		0.032464,
		0.009762,
		0.000161,
		0.163312,
		0.195044,
		0.228658,
		0.26391,
		0.300551,
		0.33833,
		0.376995,
		0.416289,
		0.455958,
		0.495746,
		0.535398,
		0.574663,
		0.613292,
		0.651042,
		0.687676,
		0.722968,
		0.756699,
		0.788668,
		0.818687,
		0.846589,
		0.87223,
		0.895491,
		0.916286,
		0.934562,
		0.950308,
		0.963556,
		0.974386,
		0.982931,
		0.989376,
		0.993962,
		0.996976,
		0.998745,
		0.999618,
		0.999937,
		0.999998,
		1,
		0.999972,
		0.99975,
		0.999012,
		0.997327,
		0.994196,
		0.989094,
		0.981498,
		0.970913,
		0.956891,
		0.939048,
		0.917071,
		0.890728,
		0.859876,
		0.824461,
		0.784529,
		0.74022,
		0.691782,
		0.639566,
		0.584036,
		0.525768,
		0.465459,
		0.40393,
		0.342132,
		0.281153,
		0.222223,
		0.166724,
		0.116196,
		0.072345,
		0.037057,
		0.012401,
		0.000647,
		0.118371,
		0.146884,
		0.177679,
		0.210511,
		0.245133,
		0.281295,
		0.318742,
		0.357216,
		0.39646,
		0.436213,
		0.476215,
		0.516206,
		0.555928,
		0.595126,
		0.63355,
		0.670956,
		0.707107,
		0.741777,
		0.774754,
		0.805839,
		0.834852,
		0.861636,
		0.886059,
		0.908019,
		0.927446,
		0.944313,
		0.958631,
		0.970465,
		0.979927,
		0.987186,
		0.992466,
		0.996042,
		0.998234,
		0.999391,
		0.999869,
		0.999991,
		1,
		0.999991,
		0.99986,
		0.999325,
		0.997978,
		0.995329,
		0.990846,
		0.983995,
		0.97426,
		0.961169,
		0.944313,
		0.92335,
		0.898023,
		0.868161,
		0.833687,
		0.794618,
		0.751077,
		0.703287,
		0.651582,
		0.596407,
		0.538323,
		0.478015,
		0.416289,
		0.354089,
		0.292492,
		0.232725,
		0.176164,
		0.124349,
		0.078987,
		0.041967,
		0.015366,
		0.001462,
		0.078987,
		0.103621,
		0.130939,
		0.160699,
		0.192655,
		0.226559,
		0.262156,
		0.299189,
		0.337397,
		0.376519,
		0.416289,
		0.456443,
		0.496716,
		0.536843,
		0.576564,
		0.615622,
		0.653763,
		0.690745,
		0.726331,
		0.760297,
		0.792437,
		0.822556,
		0.850486,
		0.876079,
		0.899218,
		0.919818,
		0.937833,
		0.953258,
		0.966137,
		0.976563,
		0.984688,
		0.990717,
		0.994912,
		0.997583,
		0.999078,
		0.999759,
		0.999972,
		1,
		0.999998,
		0.999929,
		0.999561,
		0.998512,
		0.996306,
		0.99241,
		0.986276,
		0.977373,
		0.965204,
		0.949334,
		0.929393,
		0.905097,
		0.876247,
		0.842738,
		0.804566,
		0.761826,
		0.714722,
		0.663568,
		0.608789,
		0.550931,
		0.490663,
		0.42878,
		0.366214,
		0.304035,
		0.243461,
		0.185867,
		0.132788,
		0.085934,
		0.047197,
		0.01866,
		0.00261,
		0.046277,
		0.066356,
		0.089522,
		0.115538,
		0.144161,
		0.175145,
		0.208236,
		0.243179,
		0.279714,
		0.317576,
		0.3565,
		0.396218,
		0.436458,
		0.476952,
		0.51743,
		0.557624,
		0.59727,
		0.636109,
		0.673888,
		0.710363,
		0.745299,
		0.778478,
		0.809696,
		0.83877,
		0.865538,
		0.889869,
		0.911662,
		0.930852,
		0.947416,
		0.961379,
		0.972816,
		0.981857,
		0.988691,
		0.993563,
		0.996771,
		0.998658,
		0.999591,
		0.999932,
		0.999998,
		1,
		0.99997,
		0.999732,
		0.998941,
		0.997137,
		0.993791,
		0.988347,
		0.980254,
		0.968996,
		0.954108,
		0.935195,
		0.911942,
		0.884124,
		0.851607,
		0.814362,
		0.772459,
		0.726079,
		0.675515,
		0.621173,
		0.563582,
		0.503396,
		0.441396,
		0.378502,
		0.315776,
		0.254429,
		0.19583,
		0.141513,
		0.093187,
		0.052748,
		0.022285,
		0.004095,
		0.021402,
		0.036235,
		0.05456,
		0.076143,
		0.100747,
		0.128128,
		0.158037,
		0.190218,
		0.224414,
		0.260362,
		0.297795,
		0.336442,
		0.376031,
		0.416289,
		0.45694,
		0.497709,
		0.538323,
		0.578511,
		0.618005,
		0.656545,
		0.693878,
		0.729761,
		0.763962,
		0.796268,
		0.826481,
		0.854428,
		0.87996,
		0.902962,
		0.92335,
		0.941085,
		0.95617,
		0.968661,
		0.978669,
		0.986362,
		0.991969,
		0.995774,
		0.998111,
		0.999348,
		0.99986,
		0.99999,
		1,
		0.99999,
		0.999849,
		0.999276,
		0.997833,
		0.994999,
		0.990212,
		0.982906,
		0.972543,
		0.958631,
		0.940749,
		0.918552,
		0.891784,
		0.860284,
		0.823996,
		0.782964,
		0.737347,
		0.687413,
		0.63355,
		0.576269,
		0.516206,
		0.45413,
		0.390948,
		0.327712,
		0.265625,
		0.206051,
		0.150522,
		0.100747,
		0.058623,
		0.026246,
		0.00592,
		0.005562,
		0.014447,
		0.027226,
		0.043674,
		0.063557,
		0.086636,
		0.112664,
		0.141391,
		0.172561,
		0.205913,
		0.241182,
		0.278096,
		0.316383,
		0.355767,
		0.395969,
		0.436709,
		0.477707,
		0.518684,
		0.559361,
		0.599465,
		0.638727,
		0.676885,
		0.713686,
		0.748889,
		0.782268,
		0.813614,
		0.842738,
		0.869479,
		0.893703,
		0.915312,
		0.934245,
		0.950488,
		0.964076,
		0.9751,
		0.983707,
		0.990107,
		0.994569,
		0.997416,
		0.999012,
		0.999741,
		0.99997,
		1,
		0.999998,
		0.999924,
		0.999529,
		0.998404,
		0.996042,
		0.991877,
		0.985332,
		0.975845,
		0.962903,
		0.946052,
		0.924919,
		0.899218,
		0.86876,
		0.833458,
		0.793332,
		0.748515,
		0.699252,
		0.645911,
		0.588982,
		0.529086,
		0.466975,
		0.403546,
		0.339837,
		0.277047,
		0.21653,
		0.159817,
		0.108615,
		0.064824,
		0.030544,
		0.008089,
		0.002221,
		0.008741,
		0.019337,
		0.033782,
		0.051842,
		0.073275,
		0.097836,
		0.125272,
		0.155325,
		0.187731,
		0.222223,
		0.258527,
		0.296367,
		0.335463,
		0.375532,
		0.416289,
		0.45745,
		0.498728,
		0.53984,
		0.580504,
		0.620445,
		0.65939,
		0.697079,
		0.73326,
		0.767695,
		0.800163,
		0.830462,
		0.858415,
		0.883873,
		0.90672,
		0.926878,
		0.944313,
		0.959038,
		0.971124,
		0.980698,
		0.987949,
		0.99313,
		0.996548,
		0.998563,
		0.999561,
		0.999927,
		0.999998,
		1,
		0.999968,
		0.999712,
		0.998863,
		0.99693,
		0.99335,
		0.987535,
		0.978905,
		0.96692,
		0.951098,
		0.931037,
		0.906419,
		0.877025,
		0.842738,
		0.803552,
		0.759572,
		0.711022,
		0.658245,
		0.601712,
		0.542026,
		0.479925,
		0.416289,
		0.352149,
		0.28869,
		0.227264,
		0.169394,
		0.11679,
		0.07135,
		0.035182,
		0.010606,
		0.000176,
		0.000363,
		0.004379,
		0.012657,
		0.024968,
		0.041077,
		0.060743,
		0.083718,
		0.109748,
		0.138573,
		0.169927,
		0.203541,
		0.239139,
		0.27644,
		0.315161,
		0.355015,
		0.395714,
		0.436966,
		0.478481,
		0.519969,
		0.561141,
		0.601712,
		0.641405,
		0.679948,
		0.717079,
		0.752549,
		0.786125,
		0.817592,
		0.846758,
		0.873458,
		0.897559,
		0.918966,
		0.937623,
		0.953525,
		0.966719,
		0.977312,
		0.985471,
		0.991431,
		0.995483,
		0.997978,
		0.999301,
		0.999849,
		0.99999,
		1,
		0.999989,
		0.999838,
		0.999222,
		0.997675,
		0.99464,
		0.989522,
		0.981724,
		0.970681,
		0.955885,
		0.9369,
		0.913379,
		0.88507,
		0.851827,
		0.813614,
		0.770509,
		0.722712,
		0.670542,
		0.61445,
		0.55502,
		0.492972,
		0.429172,
		0.36464,
		0.300551,
		0.238249,
		0.179254,
		0.125272,
		0.078205,
		0.040162,
		0.013475,
		0.000705,
		0.001462,
		0.007282,
		0.017323,
		0.031349,
		0.049119,
		0.070384,
		0.094888,
		0.122371,
		0.152563,
		0.185194,
		0.219983,
		0.25665,
		0.294906,
		0.33446,
		0.37502,
		0.416289,
		0.457972,
		0.499772,
		0.541394,
		0.582546,
		0.622942,
		0.6623,
		0.70035,
		0.736831,
		0.771497,
		0.804122,
		0.834498,
		0.862446,
		0.887815,
		0.91049,
		0.930398,
		0.947512,
		0.961859,
		0.973521,
		0.982646,
		0.989445,
		0.994196,
		0.997234,
		0.998941,
		0.999722,
		0.999968,
		1,
		0.999998,
		0.999918,
		0.999493,
		0.998286,
		0.995754,
		0.991297,
		0.984305,
		0.974188,
		0.960409,
		0.942504,
		0.92009,
		0.892886,
		0.860713,
		0.823506,
		0.781314,
		0.734311,
		0.682792,
		0.627186,
		0.568056,
		0.506107,
		0.442188,
		0.377306,
		0.312626,
		0.249484,
		0.189395,
		0.134062,
		0.085389,
		0.045488,
		0.016698,
		0.001593,
		9.3e-005,
		0.00331,
		0.01094,
		0.022747,
		0.038489,
		0.057915,
		0.08077,
		0.106789,
		0.135705,
		0.167242,
		0.201118,
		0.237049,
		0.274744,
		0.313909,
		0.354245,
		0.395453,
		0.43723,
		0.479275,
		0.521286,
		0.562964,
		0.604014,
		0.644146,
		0.68308,
		0.720543,
		0.75628,
		0.790049,
		0.82163,
		0.850827,
		0.877473,
		0.901435,
		0.92262,
		0.94098,
		0.95652,
		0.969301,
		0.979446,
		0.987147,
		0.99266,
		0.996306,
		0.99846,
		0.999529,
		0.999921,
		0.999998,
		1,
		0.999965,
		0.99969,
		0.998778,
		0.996704,
		0.99287,
		0.986651,
		0.977439,
		0.964669,
		0.947842,
		0.926546,
		0.900464,
		0.869387,
		0.833218,
		0.791976,
		0.745808,
		0.694984,
		0.63991,
		0.581127,
		0.519322,
		0.455329,
		0.39014,
		0.324909,
		0.260965,
		0.199814,
		0.14316,
		0.092903,
		0.051161,
		0.02028,
		0.002844,
		0.000846,
		0.00592,
		0.015366,
		0.028941,
		0.046396,
		0.067471,
		0.091904,
		0.119423,
		0.14975,
		0.182603,
		0.217694,
		0.254728,
		0.293408,
		0.333432,
		0.374495,
		0.416289,
		0.458508,
		0.500843,
		0.542987,
		0.584639,
		0.625499,
		0.665277,
		0.703692,
		0.740474,
		0.77537,
		0.808146,
		0.838591,
		0.86652,
		0.891784,
		0.914268,
		0.933906,
		0.950679,
		0.964625,
		0.975845,
		0.984507,
		0.990846,
		0.995167,
		0.997833,
		0.999249,
		0.999838,
		0.999989,
		1,
		0.999988,
		0.999825,
		0.999163,
		0.997501,
		0.994248,
		0.98877,
		0.980438,
		0.968661,
		0.95291,
		0.932739,
		0.907796,
		0.877839,
		0.842738,
		0.802484,
		0.757191,
		0.707107,
		0.65261,
		0.594222,
		0.532609,
		0.468589,
		0.403137,
		0.337397,
		0.272688,
		0.210511,
		0.152563,
		0.100747,
		0.057184,
		0.024223,
		0.004462,
		0.002366,
		0.009304,
		0.02057,
		0.035914,
		0.055077,
		0.077793,
		0.103789,
		0.132788,
		0.164503,
		0.198643,
		0.234912,
		0.273008,
		0.312626,
		0.353455,
		0.395185,
		0.437501,
		0.48009,
		0.522638,
		0.564834,
		0.606372,
		0.646952,
		0.686282,
		0.724081,
		0.760084,
		0.794042,
		0.825729,
		0.854945,
		0.881522,
		0.905327,
		0.92627,
		0.944313,
		0.959469,
		0.971816,
		0.981498,
		0.988728,
		0.993791,
		0.997036,
		0.998863,
		0.999701,
		0.999965,
		1,
		0.999998,
		0.999912,
		0.999454,
		0.998157,
		0.99544,
		0.990665,
		0.983187,
		0.972387,
		0.957705,
		0.938663,
		0.914874,
		0.886059,
		0.852057,
		0.812825,
		0.76845,
		0.719149,
		0.665277,
		0.607332,
		0.545959,
		0.481957,
		0.416289,
		0.350084,
		0.28465,
		0.221482,
		0.162273,
		0.108924,
		0.063557,
		0.028532,
		0.006452,
		0.000387,
		0.004665,
		0.013475,
		0.026564,
		0.043674,
		0.064539,
		0.088884,
		0.116428,
		0.146884,
		0.179959,
		0.215353,
		0.252761,
		0.291874,
		0.332378,
		0.373956,
		0.416289,
		0.459058,
		0.501941,
		0.544621,
		0.586783,
		0.628118,
		0.668323,
		0.707107,
		0.744191,
		0.779315,
		0.812236,
		0.842738,
		0.870636,
		0.895777,
		0.918052,
		0.937398,
		0.953807,
		0.967332,
		0.978092,
		0.986276,
		0.992149,
		0.996042,
		0.998347,
		0.999493,
		0.999915,
		0.999998,
		1,
		0.999962,
		0.999666,
		0.998685,
		0.996457,
		0.992346,
		0.985689,
		0.975845,
		0.962225,
		0.944313,
		0.921688,
		0.894038,
		0.861163,
		0.82299,
		0.779572,
		0.731099,
		0.677898,
		0.620445,
		0.559361,
		0.495428,
		0.42959,
		0.362964,
		0.296847,
		0.232725,
		0.172287,
		0.117431,
		0.070284,
		0.033208,
		0.008818,
		0.001559,
		0.007758,
		0.018444,
		0.033357,
		0.05223,
		0.074788,
		0.100747,
		0.12982,
		0.16171,
		0.196114,
		0.232725,
		0.27123,
		0.31131,
		0.352645,
		0.39491,
		0.437779,
		0.480925,
		0.524024,
		0.566751,
		0.608789,
		0.649825,
		0.689557,
		0.727694,
		0.763962,
		0.798105,
		0.82989,
		0.859113,
		0.885604,
		0.909232,
		0.929913,
		0.947615,
		0.962366,
		0.97426,
		0.983462,
		0.990212,
		0.994823,
		0.997675,
		0.999193,
		0.999825,
		0.999988,
		1,
		0.999988,
		0.999812,
		0.999098,
		0.997312,
		0.993819,
		0.987949,
		0.979038,
		0.966466,
		0.949683,
		0.928233,
		0.901765,
		0.870045,
		0.832964,
		0.790544,
		0.742944,
		0.690463,
		0.63355,
		0.572806,
		0.508991,
		0.443032,
		0.376031,
		0.309274,
		0.244237,
		0.182603,
		0.126271,
		0.077365,
		0.038254,
		0.011564,
		0.000192,
		0.0001,
		0.003529,
		0.011657,
		0.024223,
		0.040959,
		0.061589,
		0.085828,
		0.113386,
		0.143965,
		0.17726,
		0.212959,
		0.250747,
		0.290302,
		0.331297,
		0.373403,
		0.416289,
		0.459622,
		0.503068,
		0.546297,
		0.588982,
		0.630801,
		0.67144,
		0.710598,
		0.747985,
		0.783333,
		0.816392,
		0.846941,
		0.874792,
		0.899793,
		0.921837,
		0.940869,
		0.956891,
		0.969973,
		0.980254,
		0.987949,
		0.99335,
		0.99682,
		0.998778,
		0.999678,
		0.999962,
		1,
		0.999997,
		0.999905,
		0.999412,
		0.998016,
		0.995096,
		0.989975,
		0.981968,
		0.970427,
		0.954769,
		0.934501,
		0.909232,
		0.878693,
		0.842738,
		0.801356,
		0.754672,
		0.702959,
		0.646637,
		0.586284,
		0.522638,
		0.456608,
		0.389279,
		0.321925,
		0.256014,
		0.193222,
		0.135442,
		0.084802,
		0.043674,
		0.014693,
		0.000771,
		0.000903,
		0.006313,
		0.016376,
		0.030823,
		0.049379,
		0.071756,
		0.097664,
		0.126801,
		0.158861,
		0.193529,
		0.230487,
		0.269407,
		0.309961,
		0.351814,
		0.394628,
		0.438064,
		0.481783,
		0.525446,
		0.568718,
		0.611266,
		0.652767,
		0.692907,
		0.731385,
		0.767917,
		0.802239,
		0.834112,
		0.863328,
		0.889716,
		0.913148,
		0.933544,
		0.950882,
		0.965204,
		0.976625,
		0.985332,
		0.991594,
		0.995754,
		0.998223,
		0.999454,
		0.999909,
		0.999997,
		1,
		0.999959,
		0.999639,
		0.998583,
		0.996187,
		0.991772,
		0.984638,
		0.974109,
		0.959568,
		0.940484,
		0.91643,
		0.887096,
		0.8523,
		0.811993,
		0.766271,
		0.715374,
		0.659694,
		0.599783,
		0.536358,
		0.470308,
		0.402701,
		0.334797,
		0.268054,
		0.204139,
		0.144944,
		0.092596,
		0.049471,
		0.018211,
		0.001742,
		0.002525,
		0.009923,
		0.021925,
		0.038254,
		0.058623,
		0.082738,
		0.110298,
		0.140991,
		0.174504,
		0.210511,
		0.248685,
		0.28869,
		0.330188,
		0.372836,
		0.416289,
		0.460201,
		0.504225,
		0.548017,
		0.591237,
		0.63355,
		0.674631,
		0.714166,
		0.751857,
		0.787425,
		0.820615,
		0.851199,
		0.878987,
		0.903828,
		0.925619,
		0.944313,
		0.959925,
		0.972543,
		0.982327,
		0.989522,
		0.994448,
		0.997501,
		0.999131,
		0.999812,
		0.999987,
		1,
		0.999987,
		0.999797,
		0.999027,
		0.997104,
		0.99335,
		0.987053,
		0.977512,
		0.964076,
		0.946178,
		0.92335,
		0.895243,
		0.861636,
		0.822445,
		0.777728,
		0.727694,
		0.672708,
		0.613292,
		0.550142,
		0.484124,
		0.416289,
		0.347883,
		0.280351,
		0.215353,
		0.154777,
		0.100747,
		0.055645,
		0.022121,
		0.003111,
		0.000414,
		0.004979,
		0.014374,
		0.028318,
		0.046525,
		0.0687,
		0.094539,
		0.12373,
		0.155955,
		0.190888,
		0.228195,
		0.26754,
		0.308577,
		0.35096,
		0.394338,
		0.438357,
		0.482664,
		0.526907,
		0.570735,
		0.613805,
		0.655781,
		0.696335,
		0.735156,
		0.771949,
		0.806444,
		0.838394,
		0.86759,
		0.893857,
		0.917071,
		0.937158,
		0.954108,
		0.967979,
		0.978905,
		0.987103,
		0.99287,
		0.996584,
		0.998685,
		0.999653,
		0.999959,
		1,
		0.999997,
		0.999897,
		0.999365,
		0.99786,
		0.994719,
		0.98922,
		0.980638,
		0.968292,
		0.951577,
		0.929985,
		0.903125,
		0.870737,
		0.832697,
		0.78903,
		0.739908,
		0.685666,
		0.6268,
		0.563978,
		0.498046,
		0.430036,
		0.361176,
		0.292901,
		0.226862,
		0.164939,
		0.109257,
		0.0622,
		0.026425,
		0.004882,
		0.001665,
		0.008282,
		0.019678,
		0.035564,
		0.055645,
		0.079615,
		0.107161,
		0.137962,
		0.171689,
		0.208006,
		0.246571,
		0.287037,
		0.32905,
		0.372254,
		0.416289,
		0.460796,
		0.505413,
		0.549783,
		0.59355,
		0.636368,
		0.677898,
		0.717815,
		0.75581,
		0.791594,
		0.824905,
		0.855511,
		0.883219,
		0.907879,
		0.929393,
		0.947725,
		0.962903,
		0.975034,
		0.984305,
		0.990988,
		0.99544,
		0.998088,
		0.999412,
		0.999901,
		0.999997,
		1,
		0.999956,
		0.99961,
		0.99847,
		0.99589,
		0.991144,
		0.98349,
		0.972215,
		0.956676,
		0.936326,
		0.910731,
		0.879589,
		0.842738,
		0.800163,
		0.752002,
		0.698557,
		0.640295,
		0.577857,
		0.512066,
		0.443933,
		0.374671,
		0.3057,
		0.238661,
		0.175429,
		0.118126,
		0.069138,
		0.031129,
		0.00706,
		0.000107,
		0.003771,
		0.012447,
		0.025848,
		0.043674,
		0.065621,
		0.091374,
		0.120607,
		0.152992,
		0.188187,
		0.225849,
		0.265625,
		0.307157,
		0.350084,
		0.39404,
		0.438658,
		0.483569,
		0.528406,
		0.572806,
		0.61641,
		0.658868,
		0.699842,
		0.739008,
		0.776061,
		0.810721,
		0.842738,
		0.871897,
		0.898023,
		0.920996,
		0.940749,
		0.957286,
		0.970681,
		0.981094,
		0.98877,
		0.994038,
		0.997312,
		0.999064,
		0.999797,
		0.999986,
		1,
		0.999985,
		0.99978,
		0.998949,
		0.996876,
		0.992836,
		0.986073,
		0.975845,
		0.961473,
		0.942367,
		0.918052,
		0.888183,
		0.852555,
		0.811114,
		0.763962,
		0.711367,
		0.653763,
		0.591766,
		0.526172,
		0.457972,
		0.38836,
		0.318742,
		0.250747,
		0.186245,
		0.127354,
		0.076461,
		0.036235,
		0.00965,
		0.000966,
		0.006746,
		0.017488,
		0.032895,
		0.052658,
		0.076461,
		0.103978,
		0.134877,
		0.168815,
		0.205443,
		0.244406,
		0.285341,
		0.327881,
		0.371655,
		0.416289,
		0.461407,
		0.506634,
		0.551596,
		0.595924,
		0.639258,
		0.681244,
		0.721547,
		0.759844,
		0.795839,
		0.829262,
		0.859876,
		0.887485,
		0.911942,
		0.933156,
		0.951098,
		0.965817,
		0.977439,
		0.986181,
		0.992346,
		0.996325,
		0.998583,
		0.999625,
		0.999956,
		1,
		0.999997,
		0.999889,
		0.999313,
		0.997689,
		0.994306,
		0.988393,
		0.979184,
		0.965963,
		0.948101,
		0.925079,
		0.896506,
		0.862134,
		0.821869,
		0.775775,
		0.724081,
		0.667192,
		0.605693,
		0.540354,
		0.472143,
		0.402235,
		0.332021,
		0.263117,
		0.197385,
		0.13694,
		0.084169,
		0.041747,
		0.012657,
		0.000211,
		0.0027,
		0.010606,
		0.023419,
		0.04083,
		0.062522,
		0.088168,
		0.117431,
		0.149968,
		0.185427,
		0.223446,
		0.263662,
		0.3057,
		0.349184,
		0.393734,
		0.438967,
		0.484498,
		0.529947,
		0.574932,
		0.619082,
		0.662033,
		0.703432,
		0.742944,
		0.780253,
		0.815072,
		0.847143,
		0.876247,
		0.902212,
		0.924919,
		0.944313,
		0.960409,
		0.973306,
		0.983187,
		0.990328,
		0.995096,
		0.99794,
		0.999365,
		0.999893,
		0.999997,
		1,
		0.999952,
		0.999578,
		0.998347,
		0.995564,
		0.990456,
		0.982234,
		0.970147,
		0.953525,
		0.931803,
		0.904547,
		0.871464,
		0.832415,
		0.787425,
		0.736686,
		0.680569,
		0.619625,
		0.5546,
		0.486437,
		0.416289,
		0.345531,
		0.275766,
		0.208847,
		0.146884,
		0.092264,
		0.047667,
		0.016085,
		0.000846,
		0.000443,
		0.005327,
		0.015366,
		0.030251,
		0.049664,
		0.073275,
		0.100747,
		0.131734,
		0.165879,
		0.20282,
		0.242186,
		0.283601,
		0.326681,
		0.371041,
		0.416289,
		0.462036,
		0.507888,
		0.553458,
		0.598361,
		0.642221,
		0.684672,
		0.725363,
		0.763962,
		0.800163,
		0.833687,
		0.864293,
		0.891784,
		0.916014,
		0.9369,
		0.954428,
		0.968661,
		0.979753,
		0.987949,
		0.99359,
		0.997104,
		0.998989,
		0.99978,
		0.999985,
		1,
		0.999984,
		0.999762,
		0.998863,
		0.996625,
		0.992272,
		0.984999,
		0.974023,
		0.958631,
		0.938217,
		0.912296,
		0.880531,
		0.842738,
		0.798899,
		0.749168,
		0.693878,
		0.63355,
		0.568899,
		0.500843,
		0.430513,
		0.359265,
		0.28869,
		0.220628,
		0.157186,
		0.100747,
		0.053998,
		0.019939,
		0.001913,
		0.001783,
		0.008861,
		0.02104,
		0.037997,
		0.059405,
		0.084923,
		0.114202,
		0.146884,
		0.182603,
		0.220985,
		0.261647,
		0.304203,
		0.34826,
		0.39342,
		0.439285,
		0.485454,
		0.531529,
		0.577116,
		0.621825,
		0.665277,
		0.707107,
		0.746966,
		0.784529,
		0.819497,
		0.851607,
		0.880639,
		0.906419,
		0.928835,
		0.947842,
		0.963472,
		0.975845,
		0.985175,
		0.991772,
		0.996042,
		0.99847,
		0.999595,
		0.999952,
		1,
		0.999997,
		0.99988,
		0.999256,
		0.997501,
		0.993851,
		0.987485,
		0.977591,
		0.963419,
		0.944313,
		0.919743,
		0.889324,
		0.852824,
		0.810183,
		0.761512,
		0.707107,
		0.647454,
		0.583238,
		0.51535,
		0.444897,
		0.373216,
		0.301882,
		0.232725,
		0.167843,
		0.10962,
		0.060743,
		0.024223,
		0.003417,
		0.000114,
		0.004038,
		0.01332,
		0.027641,
		0.046668,
		0.070061,
		0.09747,
		0.128533,
		0.16288,
		0.200135,
		0.23991,
		0.281814,
		0.325448,
		0.370409,
		0.416289,
		0.462682,
		0.509177,
		0.555372,
		0.600863,
		0.645261,
		0.688183,
		0.729267,
		0.768167,
		0.804566,
		0.838179,
		0.86876,
		0.896111,
		0.92009,
		0.940621,
		0.957705,
		0.971428,
		0.981968,
		0.989606,
		0.994719,
		0.997777,
		0.999313,
		0.999884,
		0.999997,
		1,
		0.999948,
		0.999543,
		0.99821,
		0.995205,
		0.9897,
		0.980856,
		0.967884,
		0.950084,
		0.926878,
		0.897829,
		0.862659,
		0.82126,
		0.773702,
		0.720239,
		0.661323,
		0.597605,
		0.529947,
		0.459432,
		0.387376,
		0.315337,
		0.245133,
		0.178855,
		0.118882,
		0.067904,
		0.028941,
		0.005363,
		0.001035,
		0.007225,
		0.018718,
		0.035182,
		0.056273,
		0.08164,
		0.110919,
		0.143738,
		0.179716,
		0.218463,
		0.259581,
		0.302666,
		0.34731,
		0.393097,
		0.439612,
		0.486437,
		0.533157,
		0.579359,
		0.62464,
		0.668603,
		0.710869,
		0.751077,
		0.788888,
		0.823996,
		0.856131,
		0.88507,
		0.910641,
		0.932739,
		0.95133,
		0.966466,
		0.978292,
		0.987053,
		0.993099,
		0.996876,
		0.998907,
		0.999762,
		0.999984,
		1,
		0.999983,
		0.999741,
		0.998768,
		0.996348,
		0.991651,
		0.983819,
		0.972026,
		0.955526,
		0.933692,
		0.906036,
		0.87223,
		0.832117,
		0.785723,
		0.73326,
		0.675142,
		0.611985,
		0.544621,
		0.474108,
		0.401737,
		0.32905,
		0.25785,
		0.190218,
		0.128533,
		0.075484,
		0.034098,
		0.007758,
		0.002895,
		0.011362,
		0.02507,
		0.043674,
		0.066821,
		0.094146,
		0.125272,
		0.159817,
		0.197385,
		0.237576,
		0.27998,
		0.324181,
		0.369759,
		0.416289,
		0.463346,
		0.510503,
		0.557338,
		0.603434,
		0.64838,
		0.691782,
		0.73326,
		0.772459,
		0.809049,
		0.842738,
		0.873276,
		0.900464,
		0.924165,
		0.944313,
		0.960924,
		0.974109,
		0.984077,
		0.991144,
		0.995731,
		0.998347,
		0.999561,
		0.999948,
		1,
		0.999996,
		0.999869,
		0.999193,
		0.997294,
		0.99335,
		0.986488,
		0.975845,
		0.960634,
		0.940176,
		0.913932,
		0.881522,
		0.842738,
		0.797559,
		0.746155,
		0.688896,
		0.626365,
		0.559361,
		0.488913,
		0.416289,
		0.343012,
		0.270869,
		0.201932,
		0.138573,
		0.083483,
		0.039695,
		0.010606,
		0.000475,
		0.005711,
		0.016464,
		0.032389,
		0.05313,
		0.07832,
		0.107582,
		0.140529,
		0.176763,
		0.215878,
		0.25746,
		0.301087,
		0.346332,
		0.392764,
		0.439948,
		0.487448,
		0.53483,
		0.581665,
		0.62753,
		0.672015,
		0.714722,
		0.755278,
		0.793332,
		0.828569,
		0.860713,
		0.889537,
		0.914874,
		0.936624,
		0.954769,
		0.969383,
		0.980638,
		0.988816,
		0.994306,
		0.997598,
		0.999256,
		0.999874,
		0.999996,
		1,
		0.999944,
		0.999504,
		0.99806,
		0.994809,
		0.988867,
		0.979343,
		0.965405,
		0.946324,
		0.921508,
		0.890522,
		0.853109,
		0.809195,
		0.758906,
		0.70257,
		0.64073,
		0.574153,
		0.503836,
		0.431024,
		0.357216,
		0.284186,
		0.213992,
		0.149001,
		0.091904,
		0.045739,
		0.013913,
		0.000232,
		0.001913,
		0.009504,
		0.022547,
		0.040688,
		0.063557,
		0.090775,
		0.121952,
		0.156687,
		0.19457,
		0.235182,
		0.278096,
		0.322879,
		0.369091,
		0.416289,
		0.46403,
		0.511868,
		0.559361,
		0.606074,
		0.651582,
		0.69547,
		0.737347,
		0.776841,
		0.813614,
		0.847365,
		0.877839,
		0.904839,
		0.928233,
		0.947967,
		0.964076,
		0.976697,
		0.986073,
		0.992561,
		0.996625,
		0.998817,
		0.999741,
		0.999982,
		1,
		0.999981,
		0.999719,
		0.998663,
		0.996042,
		0.990967,
		0.982523,
		0.969836,
		0.952128,
		0.928752,
		0.899218,
		0.863213,
		0.820615,
		0.771497,
		0.716147,
		0.655065,
		0.588982,
		0.518865,
		0.44593,
		0.371655,
		0.297795,
		0.226396,
		0.159817,
		0.100747,
		0.05223,
		0.017685,
		0.000934,
		0.000123,
		0.004334,
		0.014288,
		0.029626,
		0.049979,
		0.074965,
		0.104192,
		0.137255,
		0.173742,
		0.213228,
		0.255282,
		0.299464,
		0.345327,
		0.392422,
		0.440294,
		0.488488,
		0.536552,
		0.584036,
		0.630499,
		0.675515,
		0.718669,
		0.759572,
		0.797864,
		0.833218,
		0.865352,
		0.894038,
		0.919112,
		0.940484,
		0.958153,
		0.972215,
		0.982876,
		0.990456,
		0.995389,
		0.99821,
		0.999524,
		0.999944,
		1,
		0.999996,
		0.999858,
		0.999123,
		0.997064,
		0.992798,
		0.985389,
		0.973927,
		0.957583,
		0.935655,
		0.907596,
		0.873037,
		0.831801,
		0.783913,
		0.72961,
		0.669354,
		0.603835,
		0.533987,
		0.460998,
		0.38632,
		0.311689,
		0.239139,
		0.171018,
		0.110014,
		0.059174,
		0.021925,
		0.002111,
		0.001112,
		0.007758,
		0.020082,
		0.037715,
		0.060273,
		0.08736,
		0.118572,
		0.15349,
		0.191686,
		0.232725,
		0.27616,
		0.321539,
		0.368403,
		0.416289,
		0.464734,
		0.513271,
		0.561441,
		0.608789,
		0.654869,
		0.699252,
		0.741528,
		0.781314,
		0.81826,
		0.852057,
		0.882447,
		0.909232,
		0.932289,
		0.951577,
		0.967154,
		0.979184,
		0.987949,
		0.993851,
		0.9974,
		0.999193,
		0.999863,
		0.999996,
		1,
		0.999939,
		0.99946,
		0.997893,
		0.994372,
		0.987949,
		0.977679,
		0.962684,
		0.942207,
		0.915644,
		0.882566,
		0.842738,
		0.796135,
		0.742944,
		0.683581,
		0.618696,
		0.549189,
		0.476215,
		0.401201,
		0.325863,
		0.252217,
		0.182603,
		0.119705,
		0.066571,
		0.02664,
		0.003771,
		0.003111,
		0.012202,
		0.026902,
		0.046826,
		0.071577,
		0.100747,
		0.133916,
		0.170651,
		0.210511,
		0.253045,
		0.297795,
		0.344293,
		0.39207,
		0.44065,
		0.489559,
		0.538323,
		0.586474,
		0.63355,
		0.679106,
		0.722712,
		0.763962,
		0.802484,
		0.837941,
		0.870045,
		0.898568,
		0.92335,
		0.944313,
		0.961473,
		0.974954,
		0.984999,
		0.991969,
		0.996348,
		0.998717,
		0.999719,
		0.999981,
		1,
		0.99998,
		0.999694,
		0.998546,
		0.995704,
		0.990212,
		0.981094,
		0.967429,
		0.948403,
		0.92335,
		0.891784,
		0.85341,
		0.808146,
		0.756129,
		0.697728,
		0.63355,
		0.564456,
		0.491569,
		0.416289,
		0.340308,
		0.265625,
		0.19457,
		0.12982,
		0.074425,
		0.031834,
		0.00592,
		0.000511,
		0.00614,
		0.017685,
		0.034761,
		0.05697,
		0.083901,
		0.11513,
		0.150223,
		0.188732,
		0.230203,
		0.27417,
		0.32016,
		0.367695,
		0.416289,
		0.465459,
		0.514717,
		0.563582,
		0.61158,
		0.658245,
		0.70313,
		0.745808,
		0.785882,
		0.82299,
		0.856815,
		0.887096,
		0.913638,
		0.936326,
		0.955135,
		0.970147,
		0.981561,
		0.9897,
		0.995012,
		0.99806,
		0.999482,
		0.999939,
		1,
		0.999996,
		0.999845,
		0.999046,
		0.99681,
		0.992187,
		0.984178,
		0.971816,
		0.954233,
		0.930704,
		0.900677,
		0.863799,
		0.819929,
		0.769149,
		0.711778,
		0.64838,
		0.579774,
		0.507048,
		0.431573,
		0.355015,
		0.279358,
		0.206915,
		0.140358,
		0.082738,
		0.037511,
		0.008564,
		0.002059,
		0.010218,
		0.024223,
		0.043674,
		0.068159,
		0.09725,
		0.130509,
		0.167488,
		0.207724,
		0.250747,
		0.296078,
		0.343228,
		0.391707,
		0.441017,
		0.490663,
		0.540147,
		0.588982,
		0.636686,
		0.682792,
		0.726854,
		0.76845,
		0.807193,
		0.842738,
		0.874792,
		0.903125,
		0.927583,
		0.948101,
		0.96472,
		0.977591,
		0.986998,
		0.99335,
		0.997182,
		0.999123,
		0.999851,
		0.999996,
		1,
		0.999933,
		0.999412,
		0.997707,
		0.993888,
		0.986936,
		0.975845,
		0.959694,
		0.937694,
		0.909232,
		0.87389,
		0.831466,
		0.781985,
		0.725714,
		0.663169,
		0.595126,
		0.522638,
		0.447041,
		0.369978,
		0.293408,
		0.219635,
		0.151319,
		0.091512,
		0.043674,
		0.011711,
		0.000132,
		0.004665,
		0.015366,
		0.031834,
		0.053654,
		0.080399,
		0.111627,
		0.146884,
		0.185705,
		0.227614,
		0.272124,
		0.318742,
		0.366965,
		0.416289,
		0.466205,
		0.516206,
		0.565786,
		0.61445,
		0.661713,
		0.707107,
		0.750188,
		0.790544,
		0.827802,
		0.861636,
		0.891784,
		0.918052,
		0.940336,
		0.958631,
		0.973047,
		0.983819,
		0.991318,
		0.996042,
		0.998606,
		0.999694,
		0.999979,
		1,
		0.999978,
		0.999666,
		0.998417,
		0.995329,
		0.989376,
		0.979518,
		0.96478,
		0.944313,
		0.917436,
		0.883666,
		0.842738,
		0.794618,
		0.739516,
		0.677898,
		0.610497,
		0.538323,
		0.462682,
		0.385184,
		0.30777,
		0.232725,
		0.162702,
		0.100747,
		0.05033,
		0.015366,
		0.000257,
		0.001198,
		0.008351,
		0.021601,
		0.040531,
		0.064712,
		0.093699,
		0.127035,
		0.164251,
		0.204865,
		0.248386,
		0.294311,
		0.342132,
		0.391333,
		0.441396,
		0.4918,
		0.542026,
		0.591564,
		0.63991,
		0.686577,
		0.731099,
		0.773038,
		0.811993,
		0.84761,
		0.879589,
		0.907703,
		0.931803,
		0.951842,
		0.967884,
		0.980117,
		0.988867,
		0.994596,
		0.997893,
		0.999437,
		0.999933,
		1,
		0.999995,
		0.99983,
		0.99896,
		0.996527,
		0.99151,
		0.982839,
		0.969489,
		0.950551,
		0.925275,
		0.893112,
		0.853729,
		0.807029,
		0.753166,
		0.692551,
		0.625869,
		0.55409,
		0.478481,
		0.400625,
		0.322436,
		0.246181,
		0.174504,
		0.110446,
		0.057479,
		0.019535,
		0.001035,
		0.003352,
		0.013138,
		0.028941,
		0.05033,
		0.076857,
		0.108062,
		0.143473,
		0.182603,
		0.224955,
		0.27002,
		0.317281,
		0.366214,
		0.416289,
		0.466975,
		0.51774,
		0.568056,
		0.617404,
		0.665277,
		0.711187,
		0.754672,
		0.795304,
		0.832697,
		0.86652,
		0.896506,
		0.922467,
		0.944313,
		0.962058,
		0.975845,
		0.985952,
		0.992798,
		0.99694,
		0.999046,
		0.999838,
		0.999995,
		1,
		0.999927,
		0.999358,
		0.997501,
		0.99335,
		0.985814,
		0.973821,
		0.956401,
		0.932739,
		0.902212,
		0.864419,
		0.8192,
		0.766643,
		0.707107,
		0.641225,
		0.569922,
		0.494427,
		0.416289,
		0.337397,
		0.259999,
		0.186723,
		0.120607,
		0.065127,
		0.024223,
		0.002341,
		0.000551,
		0.006618,
		0.019046,
		0.037403,
		0.061241,
		0.090096,
		0.123492,
		0.160938,
		0.201932,
		0.245958,
		0.292492,
		0.341002,
		0.390948,
		0.441786,
		0.492972,
		0.543963,
		0.594222,
		0.643226,
		0.690463,
		0.735449,
		0.777728,
		0.816885,
		0.852555,
		0.884434,
		0.912296,
		0.936004,
		0.955526,
		0.970957,
		0.982523,
		0.990599,
		0.995704,
		0.998483,
		0.999666,
		0.999977,
		1,
		0.999976,
		0.999635,
		0.998273,
		0.994912,
		0.98845,
		0.977776,
		0.961859,
		0.939815,
		0.91095,
		0.874792,
		0.831109,
		0.779928,
		0.721547,
		0.656545,
		0.585802,
		0.510503,
		0.432165,
		0.352645,
		0.27417,
		0.199356,
		0.131232,
		0.073275,
		0.029437,
		0.004182,
		0.002221,
		0.011016,
		0.026093,
		0.047001,
		0.073275,
		0.104436,
		0.139988,
		0.179424,
		0.222223,
		0.267854,
		0.315776,
		0.365439,
		0.416289,
		0.467769,
		0.519322,
		0.570395,
		0.620445,
		0.66894,
		0.715374,
		0.759263,
		0.800163,
		0.837676,
		0.871464,
		0.901258,
		0.926878,
		0.948246,
		0.965405,
		0.978531,
		0.987949,
		0.994136,
		0.997707,
		0.999385,
		0.999927,
		1,
		0.999995,
		0.999814,
		0.998863,
		0.996213,
		0.990759,
		0.981357,
		0.96692,
		0.946495,
		0.919314,
		0.884829,
		0.842738,
		0.793,
		0.73585,
		0.671812,
		0.601712,
		0.526696,
		0.448238,
		0.368169,
		0.28869,
		0.212399,
		0.14232,
		0.081927,
		0.035182,
		0.006567,
		0.000143,
		0.005035,
		0.01657,
		0.034297,
		0.057749,
		0.086442,
		0.119879,
		0.157547,
		0.198921,
		0.243461,
		0.290619,
		0.339837,
		0.39055,
		0.442188,
		0.49418,
		0.545959,
		0.596961,
		0.646637,
		0.694455,
		0.739908,
		0.782523,
		0.821869,
		0.857572,
		0.889324,
		0.9169,
		0.940176,
		0.959144,
		0.973927,
		0.984799,
		0.992187,
		0.996672,
		0.99896,
		0.999823,
		0.999995,
		1,
		0.99992,
		0.999297,
		0.997272,
		0.992753,
		0.984569,
		0.971582,
		0.952769,
		0.927288,
		0.894513,
		0.854068,
		0.805839,
		0.749996,
		0.687003,
		0.617635,
		0.542987,
		0.464497,
		0.383959,
		0.30355,
		0.225849,
		0.153869,
		0.091084,
		0.041461,
		0.009504,
		0.001294,
		0.009016,
		0.023299,
		0.043674,
		0.069658,
		0.100747,
		0.136427,
		0.176164,
		0.219416,
		0.265625,
		0.314225,
		0.36464,
		0.416289,
		0.468589,
		0.520954,
		0.572806,
		0.623575,
		0.672708,
		0.71967,
		0.763962,
		0.805122,
		0.842738,
		0.876466,
		0.906036,
		0.931276,
		0.952128,
		0.968661,
		0.981094,
		0.989805,
		0.995329,
		0.998347,
		0.999635,
		0.999974,
		1,
		0.999973,
		0.9996,
		0.998111,
		0.994448,
		0.987421,
		0.975845,
		0.958631,
		0.93486,
		0.903828,
		0.865077,
		0.818422,
		0.763962,
		0.7021,
		0.63355,
		0.559361,
		0.480925,
		0.400003,
		0.318742,
		0.2397,
		0.165879,
		0.100747,
		0.048281,
		0.012998,
		0.003623,
		0.014187,
		0.031222,
		0.05424,
		0.082738,
		0.116196,
		0.154076,
		0.19583,
		0.240892,
		0.28869,
		0.338636,
		0.39014,
		0.442603,
		0.495428,
		0.548017,
		0.599783,
		0.650148,
		0.698557,
		0.74448,
		0.787425,
		0.826947,
		0.862659,
		0.894253,
		0.921508,
		0.944313,
		0.962684,
		0.976785,
		0.986936,
		0.993626,
		0.997501,
		0.999328,
		0.99992,
		1,
		0.999994,
		0.999797,
		0.998756,
		0.995862,
		0.989923,
		0.979712,
		0.964076,
		0.94202,
		0.912756,
		0.875748,
		0.830729,
		0.777728,
		0.717079,
		0.649438,
		0.575799,
		0.497509,
		0.416289,
		0.334256,
		0.253947,
		0.178346,
		0.110919,
		0.055645,
		0.017058,
		0.000287,
		0.000596,
		0.007154,
		0.02057,
		0.040357,
		0.066007,
		0.096998,
		0.132788,
		0.172822,
		0.21653,
		0.263329,
		0.312626,
		0.363815,
		0.416289,
		0.469434,
		0.522638,
		0.575292,
		0.6268,
		0.676582,
		0.724081,
		0.768774,
		0.810183,
		0.847883,
		0.881522,
		0.910833,
		0.935655,
		0.955947,
		0.971816,
		0.983525,
		0.99151,
		0.996374,
		0.998863,
		0.999806,
		0.999994,
		1,
		0.999912,
		0.99923,
		0.997015,
		0.992087,
		0.983187,
		0.9691,
		0.948756,
		0.921283,
		0.886059,
		0.842738,
		0.79127,
		0.731919,
		0.665277,
		0.592281,
		0.51423,
		0.432804,
		0.350084,
		0.268583,
		0.191269,
		0.121599,
		0.063557,
		0.02169,
		0.001154,
		0.002404,
		0.011912,
		0.028187,
		0.050721,
		0.078987,
		0.112441,
		0.150522,
		0.192655,
		0.238249,
		0.286701,
		0.337397,
		0.389717,
		0.443032,
		0.496716,
		0.550142,
		0.602692,
		0.653763,
		0.702773,
		0.749168,
		0.792437,
		0.832117,
		0.867815,
		0.899218,
		0.926112,
		0.948403,
		0.966137,
		0.979518,
		0.988925,
		0.994912,
		0.998194,
		0.9996,
		0.999972,
		1,
		0.999971,
		0.999561,
		0.997931,
		0.99393,
		0.986276,
		0.973702,
		0.95506,
		0.929393,
		0.895994,
		0.854428,
		0.804566,
		0.746597,
		0.681045,
		0.608789,
		0.531073,
		0.449532,
		0.366214,
		0.283601,
		0.204642,
		0.132788,
		0.072021,
		0.026902,
		0.00261,
		0.000155,
		0.00545,
		0.017921,
		0.037057,
		0.062328,
		0.093187,
		0.12907,
		0.169394,
		0.213563,
		0.260965,
		0.310976,
		0.362964,
		0.416289,
		0.470308,
		0.524376,
		0.577857,
		0.630124,
		0.680569,
		0.72861,
		0.773702,
		0.815347,
		0.853109,
		0.886629,
		0.915644,
		0.940003,
		0.959694,
		0.974858,
		0.985814,
		0.99306,
		0.997272,
		0.999265,
		0.999912,
		1,
		0.999994,
		0.999776,
		0.998635,
		0.99547,
		0.988991,
		0.977883,
		0.960924,
		0.937074,
		0.905532,
		0.865776,
		0.817592,
		0.761088,
		0.69672,
		0.6253,
		0.548017,
		0.466459,
		0.382633,
		0.298992,
		0.218463,
		0.144484,
		0.081039,
		0.032698,
		0.004665,
		0.001403,
		0.009762,
		0.025204,
		0.047197,
		0.07519,
		0.108615,
		0.146884,
		0.189395,
		0.235528,
		0.28465,
		0.336118,
		0.389279,
		0.443475,
		0.498046,
		0.552335,
		0.605693,
		0.657487,
		0.707107,
		0.753975,
		0.797559,
		0.837381,
		0.873037,
		0.904213,
		0.930704,
		0.952436,
		0.969489,
		0.982117,
		0.990759,
		0.996042,
		0.998756,
		0.999787,
		0.999994,
		1,
		0.999903,
		0.999154,
		0.996728,
		0.991344,
		0.981647,
		0.966345,
		0.944313,
		0.914656,
		0.876762,
		0.830324,
		0.77537,
		0.712278,
		0.641793,
		0.565045,
		0.483569,
		0.39933,
		0.314747,
		0.232725,
		0.156687,
		0.090614,
		0.039085,
		0.007327,
		0.003928,
		0.015366,
		0.033782,
		0.058623,
		0.089317,
		0.125272,
		0.165879,
		0.210511,
		0.258527,
		0.309274,
		0.362085,
		0.416289,
		0.47121,
		0.526172,
		0.580504,
		0.63355,
		0.684672,
		0.73326,
		0.778747,
		0.820615,
		0.858415,
		0.891784,
		0.920461,
		0.944313,
		0.963355,
		0.977776,
		0.987949,
		0.994448,
		0.998024,
		0.999561,
		0.999969,
		1,
		0.999968,
		0.999517,
		0.997729,
		0.99335,
		0.984999,
		0.971319,
		0.951098,
		0.92335,
		0.887363,
		0.842738,
		0.789417,
		0.727694,
		0.658245,
		0.582134,
		0.500843,
		0.416289,
		0.330857,
		0.247423,
		0.169394,
		0.100747,
		0.046069,
		0.010606,
		0.000647,
		0.007758,
		0.022285,
		0.043674,
		0.07135,
		0.104717,
		0.14316,
		0.186045,
		0.232725,
		0.282534,
		0.334797,
		0.388827,
		0.443933,
		0.499421,
		0.5546,
		0.608789,
		0.661323,
		0.711563,
		0.758906,
		0.802795,
		0.842738,
		0.878322,
		0.909232,
		0.935275,
		0.956401,
		0.97273,
		0.984569,
		0.992429,
		0.997015,
		0.999193,
		0.999903,
		1,
		0.999993,
		0.999754,
		0.998499,
		0.99503,
		0.987949,
		0.975845,
		0.957423,
		0.931598,
		0.897559,
		0.854812,
		0.803203,
		0.742944,
		0.674631,
		0.599263,
		0.518262,
		0.433497,
		0.34731,
		0.262549,
		0.182603,
		0.111441,
		0.053654,
		0.014509,
		0.00261,
		0.012921,
		0.030544,
		0.054899,
		0.085389,
		0.121393,
		0.162273,
		0.207371,
		0.256014,
		0.307516,
		0.361176,
		0.416289,
		0.472143,
		0.528027,
		0.583238,
		0.637084,
		0.688896,
		0.738037,
		0.783913,
		0.825987,
		0.863799,
		0.89698,
		0.925275,
		0.948572,
		0.96692,
		0.980556,
		0.989923,
		0.995671,
		0.998635,
		0.999765,
		0.999993,
		1,
		0.999893,
		0.999069,
		0.996405,
		0.990511,
		0.979927,
		0.963278,
		0.939384,
		0.907331,
		0.86652,
		0.816702,
		0.758,
		0.690927,
		0.61641,
		0.535808,
		0.450936,
		0.364093,
		0.278096,
		0.19631,
		0.122695,
		0.061846,
		0.019046,
		0.000322,
		0.000168,
		0.00592,
		0.019444,
		0.040162,
		0.067471,
		0.100747,
		0.139346,
		0.182603,
		0.229837,
		0.280351,
		0.333432,
		0.38836,
		0.444407,
		0.500843,
		0.556941,
		0.611985,
		0.665277,
		0.716147,
		0.763962,
		0.808146,
		0.848188,
		0.883666,
		0.914268,
		0.939815,
		0.960286,
		0.975845,
		0.986864,
		0.99393,
		0.997833,
		0.999517,
		0.999966,
		1,
		0.999964,
		0.999468,
		0.997501,
		0.9927,
		0.98357,
		0.968661,
		0.946697,
		0.916658,
		0.877839,
		0.82989,
		0.772836,
		0.707107,
		0.63355,
		0.553458,
		0.468589,
		0.381194,
		0.294054,
		0.210511,
		0.13451,
		0.070648,
		0.024223,
		0.001294,
		0.001526,
		0.010606,
		0.027353,
		0.051161,
		0.081404,
		0.117431,
		0.158573,
		0.204139,
		0.253422,
		0.3057,
		0.360237,
		0.416289,
		0.473108,
		0.529947,
		0.586062,
		0.64073,
		0.693246,
		0.742944,
		0.789202,
		0.831466,
		0.869259,
		0.902212,
		0.930079,
		0.952769,
		0.970373,
		0.983187,
		0.991726,
		0.996728,
		0.999113,
		0.999893,
		1,
		0.999993,
		0.999728,
		0.998347,
		0.994536,
		0.986782,
		0.973569,
		0.953525,
		0.925522,
		0.888746,
		0.842738,
		0.787425,
		0.723143,
		0.650658,
		0.571191,
		0.486437,
		0.398598,
		0.310415,
		0.2252,
		0.146884,
		0.080064,
		0.030049,
		0.002929,
		0.004273,
		0.016698,
		0.03667,
		0.063557,
		0.096707,
		0.135442,
		0.179066,
		0.226862,
		0.278096,
		0.332021,
		0.387877,
		0.444897,
		0.502313,
		0.559361,
		0.615286,
		0.669354,
		0.720862,
		0.769149,
		0.813614,
		0.853729,
		0.889065,
		0.919314,
		0.944313,
		0.964076,
		0.978822,
		0.988991,
		0.995256,
		0.998499,
		0.999741,
		0.999993,
		1,
		0.999882,
		0.998972,
		0.996042,
		0.989577,
		0.978003,
		0.959858,
		0.933906,
		0.899218,
		0.855222,
		0.801739,
		0.739008,
		0.667708,
		0.588982,
		0.50446,
		0.416289,
		0.327165,
		0.24037,
		0.159817,
		0.090096,
		0.036531,
		0.005236,
		0.000705,
		0.008442,
		0.024223,
		0.047417,
		0.077365,
		0.113386,
		0.154777,
		0.200812,
		0.250747,
		0.303823,
		0.359265,
		0.416289,
		0.474108,
		0.531932,
		0.588982,
		0.644493,
		0.697728,
		0.747985,
		0.794618,
		0.837049,
		0.874792,
		0.907473,
		0.93486,
		0.956891,
		0.973702,
		0.985655,
		0.99335,
		0.997618,
		0.999468,
		0.999962,
		1,
		0.99996,
		0.999412,
		0.997245,
		0.991969,
		0.981968,
		0.965692,
		0.941795,
		0.909232,
		0.867313,
		0.815748,
		0.754672,
		0.684672,
		0.606807,
		0.522638,
		0.43425,
		0.344293,
		0.256014,
		0.173305,
		0.100747,
		0.043674,
		0.008227,
		0.002844,
		0.014064,
		0.033208,
		0.059613,
		0.092596,
		0.131445,
		0.175429,
		0.223793,
		0.275766,
		0.330561,
		0.387376,
		0.445405,
		0.503836,
		0.561865,
		0.618696,
		0.673559,
		0.725714,
		0.77447,
		0.8192,
		0.859359,
		0.894513,
		0.924359,
		0.948756,
		0.967759,
		0.981647,
		0.990939,
		0.996405,
		0.999022,
		0.999882,
		0.999999,
		0.999992,
		0.999699,
		0.998174,
		0.99398,
		0.985471,
		0.971021,
		0.949176,
		0.918768,
		0.878987,
		0.829424,
		0.770107,
		0.701521,
		0.62464,
		0.540946,
		0.452462,
		0.361785,
		0.272124,
		0.187344,
		0.112019,
		0.051487,
		0.011912,
		0.000183,
		0.006452,
		0.02117,
		0.043674,
		0.073275,
		0.109257,
		0.150882,
		0.197385,
		0.247986,
		0.301882,
		0.358259,
		0.416289,
		0.475143,
		0.533987,
		0.592002,
		0.64838,
		0.702346,
		0.753166,
		0.800163,
		0.842738,
		0.880394,
		0.912756,
		0.939609,
		0.960924,
		0.976892,
		0.987949,
		0.99479,
		0.998347,
		0.999714,
		0.999992,
		0.999999,
		0.999869,
		0.998863,
		0.995632,
		0.988525,
		0.975845,
		0.956035,
		0.927805,
		0.890217,
		0.842738,
		0.785281,
		0.718226,
		0.642452,
		0.559361,
		0.470906,
		0.379627,
		0.28869,
		0.201932,
		0.123912,
		0.059973,
		0.0163,
		0.001665,
		0.011564,
		0.029788,
		0.055645,
		0.088416,
		0.127354,
		0.171689,
		0.220628,
		0.273359,
		0.32905,
		0.386858,
		0.44593,
		0.505413,
		0.564456,
		0.622222,
		0.677898,
		0.730708,
		0.779928,
		0.824905,
		0.865077,
		0.900005,
		0.929393,
		0.953131,
		0.971319,
		0.984305,
		0.9927,
		0.997377,
		0.999412,
		0.999958,
		1,
		0.999956,
		0.999348,
		0.996954,
		0.991144,
		0.980168,
		0.962366,
		0.936326,
		0.900978,
		0.855661,
		0.800163,
		0.734755,
		0.660215,
		0.577857,
		0.489559,
		0.397801,
		0.3057,
		0.217061,
		0.136427,
		0.069138,
		0.021402,
		0.000363,
		0.004665,
		0.018211,
		0.039943,
		0.069138,
		0.105044,
		0.146884,
		0.193856,
		0.245133,
		0.299874,
		0.357216,
		0.416289,
		0.476215,
		0.536117,
		0.595126,
		0.652395,
		0.707107,
		0.75849,
		0.805839,
		0.848532,
		0.886059,
		0.918052,
		0.944313,
		0.964853,
		0.979927,
		0.990058,
		0.996042,
		0.99892,
		0.999869,
		0.999999,
		0.999991,
		0.999666,
		0.997978,
		0.99335,
		0.983995,
		0.968161,
		0.944313,
		0.911245,
		0.868161,
		0.814721,
		0.751077,
		0.677898,
		0.596407,
		0.5084,
		0.416289,
		0.323142,
		0.232725,
		0.149561,
		0.078987,
		0.027226,
		0.001462,
		0.000771,
		0.00922,
		0.026425,
		0.05166,
		0.084169,
		0.123166,
		0.167843,
		0.217363,
		0.270869,
		0.327485,
		0.38632,
		0.446476,
		0.507048,
		0.56714,
		0.625869,
		0.682378,
		0.73585,
		0.785527,
		0.830729,
		0.870879,
		0.905532,
		0.934406,
		0.957423,
		0.974739,
		0.986782,
		0.994266,
		0.998174,
		0.999683,
		0.999991,
		0.999999,
		0.999854,
		0.998739,
		0.995167,
		0.987338,
		0.973418,
		0.951752,
		0.920996,
		0.880212,
		0.828924,
		0.767158,
		0.69547,
		0.614981,
		0.527402,
		0.435073,
		0.341002,
		0.248916,
		0.163312,
		0.089522,
		0.033782,
		0.00331,
		0.003111,
		0.015366,
		0.036235,
		0.064959,
		0.100747,
		0.142782,
		0.190218,
		0.242186,
		0.297795,
		0.356136,
		0.416289,
		0.477327,
		0.538323,
		0.598361,
		0.656545,
		0.712016,
		0.763962,
		0.811648,
		0.854428,
		0.891784,
		0.92335,
		0.948956,
		0.968661,
		0.982792,
		0.991969,
		0.997104,
		0.999348,
		0.999954,
		1,
		0.999951,
		0.999276,
		0.996625,
		0.990212,
		0.978139,
		0.958631,
		0.930208,
		0.891784,
		0.842738,
		0.782964,
		0.712898,
		0.63355,
		0.54654,
		0.45413,
		0.359265,
		0.265625,
		0.177679,
		0.100747,
		0.041077,
		0.00592,
		0.000201,
		0.00706,
		0.023136,
		0.047667,
		0.079857,
		0.118882,
		0.163887,
		0.213992,
		0.268292,
		0.325863,
		0.385763,
		0.447041,
		0.508744,
		0.569922,
		0.629643,
		0.687003,
		0.741144,
		0.79127,
		0.836674,
		0.876762,
		0.911086,
		0.939384,
		0.961616,
		0.978003,
		0.989067,
		0.995632,
		0.998803,
		0.999854,
		0.999999,
		0.99999,
		0.999628,
		0.997755,
		0.992637,
		0.982327,
		0.964943,
		0.938862,
		0.902848,
		0.856131,
		0.798462,
		0.730146,
		0.652082,
		0.565786,
		0.473438,
		0.377913,
		0.282841,
		0.192655,
		0.112664,
		0.049119,
		0.009304,
		0.001825,
		0.012657,
		0.032563,
		0.060743,
		0.096367,
		0.138573,
		0.186469,
		0.239139,
		0.295641,
		0.355015,
		0.416289,
		0.478481,
		0.540612,
		0.601712,
		0.660837,
		0.717079,
		0.769588,
		0.817592,
		0.860425,
		0.897559,
		0.928639,
		0.953525,
		0.972331,
		0.985471,
		0.993675,
		0.997978,
		0.999648,
		0.99999,
		0.999999,
		0.999838,
		0.998598,
		0.99464,
		0.985994,
		0.970681,
		0.94694,
		0.913379,
		0.86907,
		0.813614,
		0.74718,
		0.670542,
		0.585111,
		0.492972,
		0.396928,
		0.300551,
		0.208236,
		0.125272,
		0.057915,
		0.013474,
		0.005114,
		0.019939,
		0.043674,
		0.075484,
		0.114498,
		0.159817,
		0.210511,
		0.265625,
		0.324181,
		0.385184,
		0.447628,
		0.510503,
		0.572806,
		0.63355,
		0.691782,
		0.746597,
		0.797161,
		0.842738,
		0.882719,
		0.916658,
		0.944313,
		0.965692,
		0.981094,
		0.991144,
		0.996795,
		0.999276,
		0.999948,
		1,
		0.999945,
		0.999193,
		0.996249,
		0.989154,
		0.975845,
		0.954428,
		0.92335,
		0.881522,
		0.828385,
		0.763962,
		0.688896,
		0.604481,
		0.512705,
		0.416289,
		0.318742,
		0.224414,
		0.138572,
		0.067471,
		0.018444,
		0.000846,
		0.010111,
		0.028941,
		0.056497,
		0.091904,
		0.134253,
		0.182603,
		0.235987,
		0.293408,
		0.353853,
		0.416289,
		0.47968,
		0.542987,
		0.605186,
		0.665277,
		0.722303,
		0.77537,
		0.823672,
		0.86652,
		0.903379,
		0.933906,
		0.958001,
		0.975845,
		0.987949,
		0.995167,
		0.998671,
		0.999838,
		0.999999,
		0.999988,
		0.999585,
		0.997501,
		0.991827,
		0.980438,
		0.961312,
		0.932739,
		0.893455,
		0.842738,
		0.780455,
		0.707107,
		0.623865,
		0.532609,
		0.435974,
		0.337397,
		0.241181,
		0.152562,
		0.077791,
		0.024222,
		0.000414,
		0.003417,
		0.016857,
		0.039695,
		0.071053,
		0.110014,
		0.155629,
		0.206915,
		0.262862,
		0.322436,
		0.384583,
		0.448238,
		0.512331,
		0.575799,
		0.637598,
		0.69672,
		0.752213,
		0.803203,
		0.848922,
		0.888746,
		0.922235,
		0.949176,
		0.969633,
		0.983995,
		0.993005,
		0.997755,
		0.999607,
		0.999988,
		0.999999,
		0.999819,
		0.998436,
		0.994038,
		0.984468,
		0.967586,
		0.941522,
		0.904839,
		0.856637,
		0.796619,
		0.725136,
		0.643226,
		0.552654,
		0.455958,
		0.3565,
		0.258526,
		0.167239,
		0.088879,
		0.03082,
		0.001665,
		0.000221,
		0.007758,
		0.025389,
		0.05223,
		0.08736,
		0.12982,
		0.178617,
		0.232725,
		0.291093,
		0.352645,
		0.416289,
		0.480925,
		0.545454,
		0.608789,
		0.669872,
		0.727694,
		0.781314,
		0.82989,
		0.872709,
		0.909232,
		0.939136,
		0.962366,
		0.979184,
		0.990212,
		0.996443,
		0.999193,
		0.999942,
		1,
		0.999939,
		0.999098,
		0.99582,
		0.987949,
		0.973246,
		0.949683,
		0.915644,
		0.870045,
		0.812415,
		0.742944,
		0.662527,
		0.572806,
		0.476214,
		0.376029,
		0.276435,
		0.182595,
		0.100736,
		0.038245,
		0.003769,
		0.002008,
		0.013913,
		0.035743,
		0.066571,
		0.105431,
		0.151319,
		0.203198,
		0.259999,
		0.320624,
		0.383959,
		0.448872,
		0.51423,
		0.578906,
		0.641793,
		0.701826,
		0.758,
		0.809397,
		0.855222,
		0.894835,
		0.927805,
		0.953955,
		0.973418,
		0.986687,
		0.99464,
		0.99852,
		0.999819,
		0.999999,
		0.999987,
		0.999535,
		0.99721,
		0.990902,
		0.978292,
		0.957205,
		0.925841,
		0.882926,
		0.827802,
		0.760489,
		0.681729,
		0.59303,
		0.496713,
		0.395962,
		0.294891,
		0.198621,
		0.113359,
		0.046503,
		0.00674,
		0.005631,
		0.021925,
		0.047952,
		0.082738,
		0.125272,
		0.174504,
		0.229348,
		0.28869,
		0.35139,
		0.416289,
		0.482221,
		0.548017,
		0.612528,
		0.674631,
		0.73326,
		0.787425,
		0.836245,
		0.878987,
		0.915109,
		0.944313,
		0.9666,
		0.982327,
		0.992245,
		0.997501,
		0.999561,
		0.999987,
		0.999999,
		0.999797,
		0.998251,
		0.99335,
		0.98273,
		0.964076,
		0.935405,
		0.895243,
		0.842738,
		0.777728,
		0.70079,
		0.613289,
		0.51742,
		0.416269,
		0.313873,
		0.215301,
		0.126741,
		0.055596,
		0.01059,
		0.000934,
		0.011138,
		0.031834,
		0.062044,
		0.100747,
		0.146884,
		0.199356,
		0.257029,
		0.318742,
		0.383309,
		0.449532,
		0.516206,
		0.582134,
		0.646144,
		0.707107,
		0.763962,
		0.815748,
		0.861636,
		0.900978,
		0.933353,
		0.958631,
		0.977027,
		0.989154,
		0.996042,
		0.999098,
		0.999935,
		1,
		0.999931,
		0.998989,
		0.995329,
		0.986574,
		0.970291,
		0.944313,
		0.906963,
		0.857183,
		0.794617,
		0.719666,
		0.633539,
		0.538297,
		0.436917,
		0.333351,
		0.232614,
		0.140868,
		0.065521,
		0.015327,
		0.003771,
		0.018575,
		0.043674,
		0.078042,
		0.120607,
		0.170259,
		0.225849,
		0.286194,
		0.350084,
		0.416289,
		0.483569,
		0.550683,
		0.61641,
		0.679561,
		0.739008,
		0.793707,
		0.842738,
		0.885348,
		0.920996,
		0.949417,
		0.970681,
		0.985256,
		0.994038,
		0.998347,
		0.999797,
		0.999999,
		0.999985,
		0.999478,
		0.996876,
		0.989843,
		0.975845,
		0.952544,
		0.918052,
		0.871094,
		0.811109,
		0.738309,
		0.653732,
		0.559298,
		0.457861,
		0.353287,
		0.250527,
		0.155717,
		0.076267,
		0.020955,
		0.000244,
		0.008564,
		0.027987,
		0.057479,
		0.095965,
		0.14232,
		0.195381,
		0.253947,
		0.316784,
		0.382633,
		0.450219,
		0.518262,
		0.58549,
		0.650658,
		0.71257,
		0.770107,
		0.822257,
		0.868161,
		0.907163,
		0.938862,
		0.963183,
		0.980438,
		0.991379,
		0.99721,
		0.999508,
		0.999985,
		0.999999,
		0.999771,
		0.998037,
		0.992561,
		0.980744,
		0.960083,
		0.928478,
		0.884429,
		0.827155,
		0.756663,
		0.673812,
		0.580364,
		0.479047,
		0.373628,
		0.268996,
		0.171255,
		0.087814,
		0.027471,
		0.000471,
		0.002221,
		0.015366,
		0.039412,
		0.073275,
		0.115823,
		0.165879,
		0.222223,
		0.283601,
		0.348725,
		0.416289,
		0.484973,
		0.553458,
		0.620445,
		0.684672,
		0.744944,
		0.800163,
		0.849368,
		0.891784,
		0.926878,
		0.954428,
		0.974587,
		0.987949,
		0.995583,
		0.998989,
		0.999927,
		1,
		0.999922,
		0.998863,
		0.994763,
		0.984999,
		0.966918,
		0.938211,
		0.897143,
		0.842698,
		0.774666,
		0.69371,
		0.601425,
		0.500403,
		0.394307,
		0.287961,
		0.187433,
		0.100129,
		0.034862,
		0.001888,
		0.006231,
		0.024223,
		0.052888,
		0.091084,
		0.137622,
		0.191269,
		0.250747,
		0.314747,
		0.381929,
		0.450936,
		0.520404,
		0.588982,
		0.655346,
		0.718226,
		0.776439,
		0.828924,
		0.874792,
		0.913379,
		0.944313,
		0.967586,
		0.983628,
		0.99335,
		0.998148,
		0.999771,
		0.999999,
		0.999984,
		0.999412,
		0.996489,
		0.988624,
		0.973042,
		0.947222,
		0.909189,
		0.857677,
		0.792245,
		0.713346,
		0.622395,
		0.52184,
		0.415238,
		0.30734,
		0.204182,
		0.113164,
		0.043104,
		0.004255,
		0.001035,
		0.012329,
		0.035182,
		0.068445,
		0.110919,
		0.161357,
		0.218463,
		0.280903,
		0.34731,
		0.416289,
		0.486437,
		0.556348,
		0.62464,
		0.689972,
		0.751077,
		0.806797,
		0.856131,
		0.898284,
		0.932739,
		0.959323,
		0.978292,
		0.99039,
		0.996876,
		0.999446,
		0.999984,
		0.999999,
		0.999741,
		0.997788,
		0.991646,
		0.978449,
		0.955481,
		0.920512,
		0.872018,
		0.80931,
		0.732617,
		0.643164,
		0.543246,
		0.436307,
		0.327031,
		0.221416,
		0.126853,
		0.052157,
		0.007563,
		0.004182,
		0.02057,
		0.048281,
		0.086107,
		0.132788,
		0.187012,
		0.247423,
		0.312626,
		0.381194,
		0.451682,
		0.522638,
		0.592618,
		0.660215,
		0.724081,
		0.782964,
		0.835752,
		0.881522,
		0.91961,
		0.949683,
		0.971816,
		0.986574,
		0.995056,
		0.998863,
		0.999917,
		1,
		0.999911,
		0.998711,
		0.994091,
		0.983139,
		0.962953,
		0.931051,
		0.885637,
		0.825756,
		0.751401,
		0.663597,
		0.564478,
		0.457378,
		0.346903,
		0.239021,
		0.141106,
		0.061965,
		0.011792,
		0.000272,
		0.009504,
		0.031004,
		0.063557,
		0.105894,
		0.156687,
		0.214561,
		0.278096,
		0.345833,
		0.416289,
		0.487964,
		0.559361,
		0.629005,
		0.69547,
		0.757413,
		0.813614,
		0.863025,
		0.904839,
		0.938558,
		0.964076,
		0.981771,
		0.992561,
		0.997917,
		0.999741,
		0.999997,
		0.999976,
		0.999317,
		0.995994,
		0.98711,
		0.969601,
		0.940734,
		0.898429,
		0.841453,
		0.769549,
		0.68353,
		0.585368,
		0.478279,
		0.366798,
		0.256856,
		0.155812,
		0.07245,
		0.016906,
		0.00247,
		0.017058,
		0.043674,
		0.081039,
		0.127813,
		0.182603,
		0.243966,
		0.310415,
		0.380428,
		0.452462,
		0.524968,
		0.596407,
		0.665277,
		0.730146,
		0.789689,
		0.842738,
		0.888342,
		0.925841,
		0.954949,
		0.975845,
		0.989256,
		0.996489,
		0.999373,
		0.999976,
		0.999981,
		0.999659,
		0.997388,
		0.990359,
		0.97538,
		0.949475,
		0.910271,
		0.856247,
		0.786881,
		0.702768,
		0.605712,
		0.498808,
		0.386524,
		0.274752,
		0.170831,
		0.083512,
		0.022848,
		0.006932,
		0.026902,
		0.058623,
		0.100747,
		0.151865,
		0.210511,
		0.275172,
		0.344293,
		0.416289,
		0.489559,
		0.562504,
		0.63355,
		0.701178,
		0.763962,
		0.820615,
		0.870045,
		0.911434,
		0.944313,
		0.968661,
		0.984999,
		0.994447,
		0.998711,
		0.999888,
		0.999952,
		0.999787,
		0.998305,
		0.992877,
		0.980234,
		0.957169,
		0.921018,
		0.869955,
		0.803188,
		0.721083,
		0.62527,
		0.51873,
		0.405858,
		0.29251,
		0.185998,
		0.095026,
		0.029542,
		0.000502,
		0.001154,
		0.013723,
		0.039085,
		0.075884,
		0.122695,
		0.178037,
		0.24037,
		0.308108,
		0.379627,
		0.453277,
		0.527402,
		0.600358,
		0.670542,
		0.73643,
		0.796619,
		0.849883,
		0.895243,
		0.932051,
		0.960083,
		0.979644,
		0.991646,
		0.997633,
		0.999659,
		0.999885,
		0.999737,
		0.998768,
		0.994648,
		0.984094,
		0.963696,
		0.930499,
		0.882368,
		0.818227,
		0.73821,
		0.643772,
		0.537774,
		0.424544,
		0.309899,
		0.201119,
		0.106843,
		0.036891,
		0.001976,
		0.004665,
		0.022905,
		0.053654,
		0.095481,
		0.146884,
		0.206304,
		0.272124,
		0.342684,
		0.416289,
		0.491227,
		0.565786,
		0.638287,
		0.707107,
		0.770731,
		0.827802,
		0.877185,
		0.918052,
		0.949977,
		0.973042,
		0.987932,
		0.995994,
		0.999178,
		0.999737,
		0.999522,
		0.998789,
		0.995644,
		0.986877,
		0.968915,
		0.938519,
		0.893244,
		0.831725,
		0.753855,
		0.660911,
		0.555639,
		0.442298,
		0.326664,
		0.215973,
		0.118789,
		0.044773,
		0.004361,
		0.000304,
		0.010606,
		0.034535,
		0.070648,
		0.117431,
		0.173305,
		0.236626,
		0.3057,
		0.37879,
		0.45413,
		0.529947,
		0.604481,
		0.676022,
		0.742944,
		0.803759,
		0.857183,
		0.90221,
		0.938211,
		0.965039,
		0.983139,
		0.993631,
		0.998305,
		0.999416,
		0.999123,
		0.998361,
		0.995825,
		0.988486,
		0.972665,
		0.944861,
		0.90232,
		0.84338,
		0.767694,
		0.676358,
		0.572,
		0.458811,
		0.342523,
		0.23032,
		0.130667,
		0.053042,
		0.007571,
		0.002763,
		0.019046,
		0.048665,
		0.090096,
		0.14174,
		0.201932,
		0.268945,
		0.341002,
		0.416289,
		0.492972,
		0.569217,
		0.643226,
		0.713267,
		0.777728,
		0.835175,
		0.884429,
		0.924657,
		0.955481,
		0.977092,
		0.990359,
		0.996864,
		0.998789,
		0.998473,
		0.99745,
		0.995131,
		0.988808,
		0.974769,
		0.949288,
		0.909308,
		0.852871,
		0.779382,
		0.689759,
		0.58651,
		0.473757,
		0.35718,
		0.243899,
		0.142261,
		0.061533,
		0.011499,
		0.007758,
		0.030049,
		0.065341,
		0.112019,
		0.168399,
		0.232725,
		0.303183,
		0.377913,
		0.455022,
		0.532609,
		0.608789,
		0.681729,
		0.749695,
		0.811109,
		0.864619,
		0.909189,
		0.944206,
		0.969601,
		0.98597,
		0.994648,
		0.997669,
		0.99745,
		0.995988,
		0.993485,
		0.987715,
		0.975038,
		0.95155,
		0.913908,
		0.859861,
		0.788563,
		0.700751,
		0.598814,
		0.486799,
		0.370327,
		0.256439,
		0.153342,
		0.070064,
		0.016015,
		0.001294,
		0.015366,
		0.043674,
		0.084596,
		0.136427,
		0.197385,
		0.265625,
		0.339242,
		0.416289,
		0.494799,
		0.572806,
		0.648379,
		0.719666,
		0.784947,
		0.842698,
		0.891682,
		0.931051,
		0.960466,
		0.980234,
		0.991411,
		0.995825,
		0.995868,
		0.993865,
		0.990785,
		0.985067,
		0.973275,
		0.951392,
		0.915816,
		0.864009,
		0.794878,
		0.708972,
		0.608558,
		0.497602,
		0.381656,
		0.267665,
		0.163672,
		0.078437,
		0.020971,
		0.005236,
		0.025659,
		0.059973,
		0.106458,
		0.163312,
		0.228658,
		0.300551,
		0.376995,
		0.455958,
		0.535398,
		0.613289,
		0.687664,
		0.756663,
		0.818594,
		0.872018,
		0.915849,
		0.949475,
		0.9729,
		0.986877,
		0.992978,
		0.993485,
		0.990912,
		0.986903,
		0.980718,
		0.969283,
		0.948562,
		0.914735,
		0.864989,
		0.797981,
		0.714073,
		0.615402,
		0.505845,
		0.390871,
		0.277308,
		0.173013,
		0.086449,
		0.026203,
		0.000432,
		0.000342,
		0.011912,
		0.038704,
		0.078987,
		0.130939,
		0.192655,
		0.262156,
		0.337397,
		0.416289,
		0.496713,
		0.576554,
		0.653732,
		0.726248,
		0.792245,
		0.85008,
		0.898429,
		0.936403,
		0.963696,
		0.980739,
		0.988808,
		0.99,
		0.986903,
		0.981686,
		0.974514,
		0.962873,
		0.942826,
		0.910389,
		0.862495,
		0.797554,
		0.715733,
		0.619035,
		0.511233,
		0.397697,
		0.285117,
		0.181135,
		0.093895,
		0.031536,
		0.001638,
		0.003111,
		0.021402,
		0.05456,
		0.100747,
		0.158037,
		0.224414,
		0.297794,
		0.376029,
		0.456932,
		0.538297,
		0.617935,
		0.69371,
		0.763598,
		0.825756,
		0.878618,
		0.921018,
		0.952336,
		0.972665,
		0.982956,
		0.985067,
		0.98155,
		0.974941,
		0.966297,
		0.953873,
		0.933976,
		0.902538,
		0.856264,
		0.793321,
		0.713679,
		0.61919,
		0.513512,
		0.401894,
		0.290865,
		0.187825,
		0.100576,
		0.036791,
		0.003467,
		0.008741,
		0.033782,
		0.073275,
		0.125272,
		0.187731,
		0.258526,
		0.335457,
		0.416269,
		0.498671,
		0.580364,
		0.659076,
		0.732617,
		0.798942,
		0.856247,
		0.903084,
		0.938519,
		0.962301,
		0.975038,
		0.978299,
		0.974514,
		0.966441,
		0.955914,
		0.942136,
		0.921848,
		0.890993,
		0.846087,
		0.785067,
		0.707694,
		0.615657,
		0.512479,
		0.403267,
		0.294361,
		0.192897,
		0.106308,
		0.04179,
		0.005759,
		0.001462,
		0.017323,
		0.049119,
		0.094888,
		0.152562,
		0.21998,
		0.294891,
		0.374977,
		0.457861,
		0.541136,
		0.622395,
		0.699281,
		0.769549,
		0.831152,
		0.882368,
		0.921945,
		0.949288,
		0.964651,
		0.969283,
		0.96542,
		0.955914,
		0.943206,
		0.927549,
		0.906329,
		0.875627,
		0.83183,
		0.77265,
		0.697635,
		0.608295,
		0.507996,
		0.401677,
		0.295462,
		0.196197,
		0.110928,
		0.046363,
		0.008346,
		0.00592,
		0.028941,
		0.067471,
		0.11942,
		0.182595,
		0.254699,
		0.333351,
		0.416091,
		0.500403,
		0.583747,
		0.663597,
		0.737504,
		0.803188,
		0.858647,
		0.90232,
		0.93327,
		0.951392,
		0.957597,
		0.953873,
		0.943055,
		0.928019,
		0.910034,
		0.887367,
		0.856393,
		0.813438,
		0.756013,
		0.683445,
		0.597047,
		0.5,
		0.397048,
		0.294076,
		0.197612,
		0.1143,
		0.050357,
		0.011064,
		0.000387,
		0.013474,
		0.043674,
		0.088879,
		0.146868,
		0.215301,
		0.291735,
		0.373628,
		0.458357,
		0.543246,
		0.625599,
		0.702768,
		0.772234,
		0.831725,
		0.879373,
		0.913908,
		0.934871,
		0.942826,
		0.939475,
		0.927549,
		0.910191,
		0.889556,
		0.864981,
		0.833327,
		0.790953,
		0.735198,
		0.665163,
		0.581942,
		0.488506,
		0.389377,
		0.29017,
		0.197077,
		0.116324,
		0.053638,
		0.013754,
		0.003529,
		0.024222,
		0.061582,
		0.113359,
		0.177175,
		0.250527,
		0.330792,
		0.415238,
		0.501051,
		0.585368,
		0.66534,
		0.73821,
		0.801442,
		0.852871,
		0.890903,
		0.914735,
		0.924584,
		0.921848,
		0.909081,
		0.889556,
		0.866118,
		0.839261,
		0.806556,
		0.764515,
		0.710347,
		0.642924,
		0.563102,
		0.473615,
		0.378733,
		0.283774,
		0.194576,
		0.116937,
		0.056098,
		0.01627,
		0.000256,
		0.009922,
		0.038245,
		0.082699,
		0.140868,
		0.21019,
		0.287961,
		0.371341,
		0.457378,
		0.543038,
		0.62527,
		0.701088,
		0.767694,
		0.822636,
		0.864009,
		0.890684,
		0.902538,
		0.900656,
		0.887367,
		0.865957,
		0.839755,
		0.810366,
		0.776299,
		0.734367,
		0.681705,
		0.616965,
		0.540744,
		0.455512,
		0.365257,
		0.274981,
		0.190145,
		0.116117,
		0.057659,
		0.018485,
		0.000913,
		0.001665,
		0.019668,
		0.055596,
		0.106998,
		0.171255,
		0.245582,
		0.327031,
		0.412507,
		0.498808,
		0.582681,
		0.660911,
		0.730447,
		0.788563,
		0.83306,
		0.862495,
		0.876427,
		0.875627,
		0.86217,
		0.839261,
		0.810529,
		0.77852,
		0.742861,
		0.700845,
		0.649616,
		0.587617,
		0.515172,
		0.434461,
		0.349164,
		0.263942,
		0.18387,
		0.113883,
		0.058278,
		0.020295,
		0.001814,
		0.00674,
		0.032845,
		0.076267,
		0.134336,
		0.204182,
		0.282748,
		0.366798,
		0.452958,
		0.537774,
		0.617806,
		0.689759,
		0.750642,
		0.797981,
		0.830044,
		0.846087,
		0.84658,
		0.833327,
		0.809382,
		0.77852,
		0.743996,
		0.706621,
		0.664376,
		0.614514,
		0.555299,
		0.486772,
		0.4108,
		0.330732,
		0.250868,
		0.175889,
		0.110298,
		0.057942,
		0.021621,
		0.002816,
		0.000442,
		0.015327,
		0.049465,
		0.100129,
		0.164372,
		0.239021,
		0.320699,
		0.405858,
		0.490845,
		0.572,
		0.645788,
		0.708972,
		0.758816,
		0.793321,
		0.811468,
		0.813438,
		0.800764,
		0.776299,
		0.743835,
		0.707107,
		0.668022,
		0.625459,
		0.57691,
		0.520504,
		0.455999,
		0.38493,
		0.310294,
		0.236018,
		0.166381,
		0.105461,
		0.056675,
		0.022412,
		0.003798,
		0.004022,
		0.027471,
		0.069425,
		0.126853,
		0.19647,
		0.274752,
		0.357979,
		0.442298,
		0.523831,
		0.598814,
		0.663773,
		0.715733,
		0.75245,
		0.77265,
		0.77625,
		0.764515,
		0.74007,
		0.706621,
		0.668179,
		0.627545,
		0.584648,
		0.537373,
		0.483784,
		0.423361,
		0.3573,
		0.288229,
		0.219692,
		0.155564,
		0.099509,
		0.054533,
		0.022647,
		0.004664,
		0.011259,
		0.043104,
		0.092422,
		0.155812,
		0.229604,
		0.309899,
		0.392648,
		0.473757,
		0.549241,
		0.615402,
		0.669042,
		0.707694,
		0.729854,
		0.735198,
		0.724736,
		0.700845,
		0.667076,
		0.627545,
		0.585694,
		0.542528,
		0.496513,
		0.445728,
		0.389406,
		0.328396,
		0.26495,
		0.202223,
		0.143689,
		0.092608,
		0.0516,
		0.02233,
		0.005339,
		0.001888,
		0.022135,
		0.061965,
		0.11786,
		0.185998,
		0.262302,
		0.342523,
		0.422357,
		0.497602,
		0.564348,
		0.61919,
		0.659459,
		0.683445,
		0.690604,
		0.681705,
		0.658875,
		0.625459,
		0.585544,
		0.542966,
		0.499696,
		0.454953,
		0.406945,
		0.354699,
		0.298721,
		0.240889,
		0.183964,
		0.131026,
		0.084949,
		0.047987,
		0.021495,
		0.005777,
		8.4e-005,
		0.007563,
		0.036449,
		0.083512,
		0.144773,
		0.215973,
		0.292662,
		0.370327,
		0.444555,
		0.511233,
		0.566765,
		0.608295,
		0.633929,
		0.642924,
		0.635826,
		0.614514,
		0.582095,
		0.542528,
		0.499835,
		0.456733,
		0.413311,
		0.368043,
		0.319805,
		0.268784,
		0.216483,
		0.165276,
		0.11786,
		0.076739,
		0.043827,
		0.020197,
		0.005956,
		0.000268,
		0.000494,
		0.016906,
		0.053735,
		0.106843,
		0.171783,
		0.243899,
		0.318468,
		0.390871,
		0.456801,
		0.512479,
		0.55488,
		0.581942,
		0.592741,
		0.587617,
		0.568211,
		0.537373,
		0.498867,
		0.456733,
		0.414182,
		0.372177,
		0.329606,
		0.285274,
		0.239079,
		0.192162,
		0.146521,
		0.104482,
		0.068196,
		0.039266,
		0.018512,
		0.005877,
		0.000473,
		0.0044,
		0.029542,
		0.073179,
		0.130667,
		0.197141,
		0.267665,
		0.337415,
		0.401894,
		0.45715,
		0.5,
		0.528227,
		0.540744,
		0.537699,
		0.520504,
		0.491749,
		0.454953,
		0.414057,
		0.372533,
		0.33209,
		0.292176,
		0.251619,
		0.210073,
		0.168337,
		0.128047,
		0.091174,
		0.05954,
		0.03446,
		0.016531,
		0.005563,
		0.000644,
		0.011958,
		0.044773,
		0.093595,
		0.153342,
		0.218859,
		0.285117,
		0.347429,
		0.401677,
		0.444525,
		0.473615,
		0.487707,
		0.486772,
		0.472002,
		0.445728,
		0.411201,
		0.372177,
		0.3322,
		0.293526,
		0.256238,
		0.219306,
		0.182192,
		0.145385,
		0.110179,
		0.078208,
		0.050988,
		0.029569,
		0.014355,
		0.005051,
		0.000748,
		0.001976,
		0.0226,
		0.061533,
		0.113465,
		0.173013,
		0.234937,
		0.294361,
		0.347001,
		0.389377,
		0.418991,
		0.434461,
		0.435589,
		0.423361,
		0.399855,
		0.368043,
		0.33144,
		0.293526,
		0.256879,
		0.222209,
		0.188738,
		0.155812,
		0.123644,
		0.09321,
		0.065832,
		0.04274,
		0.024749,
		0.012091,
		0.004394,
		0.000774,
		0.007571,
		0.0354,
		0.078437,
		0.131066,
		0.187825,
		0.243645,
		0.294076,
		0.33549,
		0.365257,
		0.381856,
		0.38493,
		0.375266,
		0.354699,
		0.325921,
		0.292176,
		0.256786,
		0.222461,
		0.190422,
		0.160248,
		0.131247,
		0.103399,
		0.077394,
		0.054265,
		0.03498,
		0.020142,
		0.009842,
		0.003651,
		0.000727,
		0.000483,
		0.016015,
		0.049109,
		0.093895,
		0.144681,
		0.196197,
		0.243819,
		0.283774,
		0.313298,
		0.330732,
		0.335565,
		0.328396,
		0.310826,
		0.285274,
		0.254699,
		0.222209,
		0.190497,
		0.161129,
		0.134094,
		0.108744,
		0.08488,
		0.062938,
		0.043688,
		0.027863,
		0.015878,
		0.007705,
		0.00288,
		0.000623,
		0.004055,
		0.026203,
		0.062266,
		0.106308,
		0.152846,
		0.197077,
		0.235077,
		0.263942,
		0.281881,
		0.288229,
		0.283404,
		0.268784,
		0.246526,
		0.219306,
		0.189978,
		0.161129,
		0.134495,
		0.110452,
		0.08848,
		0.068256,
		0.049997,
		0.034243,
		0.021513,
		0.012061,
		0.005762,
		0.002141,
		0.000485,
		1.6e-005,
		0.010267,
		0.036791,
		0.073371,
		0.1143,
		0.154586,
		0.190145,
		0.217932,
		0.236018,
		0.24359,
		0.240889,
		0.229082,
		0.210073,
		0.186264,
		0.160248,
		0.134436,
		0.110603,
		0.089421,
		0.070559,
		0.05363,
		0.038672,
		0.026022,
		0.016015,
		0.008767,
		0.00408,
		0.001481,
		0.000339,
		1.8e-005,
		0.001638,
		0.017889,
		0.046363,
		0.08111,
		0.116937,
		0.149594,
		0.175889,
		0.193752,
		0.202223,
		0.201372,
		0.192162,
		0.176263,
		0.155812,
		0.13314,
		0.110452,
		0.089464,
		0.071024,
		0.055015,
		0.041042,
		0.029006,
		0.01907,
		0.011415,
		0.006041,
		0.002699,
		0.000936,
		0.000207,
		1.4e-005,
		0.005759,
		0.025572,
		0.053638,
		0.084551,
		0.113883,
		0.138292,
		0.155564,
		0.164591,
		0.165276,
		0.158393,
		0.145385,
		0.128143,
		0.108744,
		0.089172,
		0.071024,
		0.055213,
		0.041813,
		0.030469,
		0.020985,
		0.013384,
		0.007717,
		0.003894,
		0.001636,
		0.000524,
		0.000106,
		0.000352,
		0.011064,
		0.032043,
		0.057659,
		0.083299,
		0.105461,
		0.121789,
		0.131026,
		0.132916,
		0.128047,
		0.117656,
		0.103399,
		0.087112,
		0.070559,
		0.055184,
		0.041882,
		0.030859,
		0.021831,
		0.014543,
		0.00891,
		0.004882,
		0.002303,
		0.000882,
		0.000248,
		4.1e-005,
		0.002672,
		0.01627,
		0.036295,
		0.057942,
		0.077548,
		0.092608,
		0.101724,
		0.104482,
		0.101297,
		0.09321,
		0.081659,
		0.068256,
		0.054548,
		0.041813,
		0.030877,
		0.022005,
		0.014994,
		0.009563,
		0.005553,
		0.002835,
		0.001213,
		0.000402,
		9e-005,
		0.00605,
		0.020295,
		0.037732,
		0.054533,
		0.068049,
		0.076739,
		0.080042,
		0.078208,
		0.072094,
		0.062938,
		0.052138,
		0.041042,
		0.030755,
		0.022005,
		0.015059,
		0.00978,
		0.005889,
		0.003177,
		0.001466,
		0.00054,
		0.000141,
		2e-005,
		0.000913,
		0.009308,
		0.022412,
		0.036236,
		0.047987,
		0.055989,
		0.05954,
		0.058738,
		0.054265,
		0.047176,
		0.038672,
		0.02991,
		0.021831,
		0.01505,
		0.0098,
		0.005981,
		0.003331,
		0.001619,
		0.000642,
		0.000185,
		3.1e-005,
		0.002816,
		0.011571,
		0.02233,
		0.032163,
		0.039266,
		0.042807,
		0.04274,
		0.039592,
		0.034243,
		0.027714,
		0.020985,
		0.014841,
		0.00978,
		0.005985,
		0.003363,
		0.001679,
		0.000698,
		0.000216,
		4e-005,
		0.000161,
		0.004664,
		0.012361,
		0.020197,
		0.02625,
		0.029569,
		0.02998,
		0.027863,
		0.023941,
		0.01907,
		0.014064,
		0.009563,
		0.005956,
		0.003363,
		0.001688,
		0.000718,
		0.000232,
		4.5e-005,
		0.001043,
		0.005777,
		0.011614,
		0.016531,
		0.019471,
		0.020142,
		0.018798,
		0.016015,
		0.012498,
		0.00891,
		0.00576,
		0.003331,
		0.001687,
		0.000719,
		0.000237,
		4.8e-005,
		0.001962,
		0.005877,
		0.009635,
		0.012091,
		0.012854,
		0.012061,
		0.01016,
		0.007717,
		0.005261,
		0.003177,
		0.001659,
		0.000718,
		0.000237,
		4.9e-005,
		0.000268,
		0.002411,
		0.005051,
		0.006974,
		0.007705,
		0.007282,
		0.006041,
		0.004423,
		0.002835,
		0.001555,
		0.000698,
		0.000235,
		4.9e-005,
		0.000644,
		0.002263,
		0.003651,
		0.004271,
		0.00408,
		0.003312,
		0.002303,
		0.001351,
		0.000642,
		0.000226,
		4.8e-005,
		3.2e-005,
		0.000774,
		0.00167,
		0.002141,
		0.002079,
		0.001636,
		0.001055,
		0.00054,
		0.000203,
		4.5e-005,
		0.000143,
		0.000623,
		0.000938,
		0.000936,
		0.000703,
		0.000402,
		0.000165,
		4e-005,
		0.000161,
		0.000339,
		0.000355,
		0.000248,
		0.000116,
		3.1e-005,
		1.6e-005,
		9.1e-005,
		0.000106,
		6.4e-005,
		2e-005,
		1.4e-005,
		2.1e-005,
		2.1e-005,
		1.4e-005,
		2e-005,
		6.4e-005,
		0.000106,
		9.1e-005,
		1.6e-005,
		3.1e-005,
		0.000116,
		0.000248,
		0.000355,
		0.000339,
		0.000161,
		4e-005,
		0.000165,
		0.000402,
		0.000703,
		0.000936,
		0.000938,
		0.000623,
		0.000143,
		4.5e-005,
		0.000203,
		0.00054,
		0.001055,
		0.001636,
		0.002079,
		0.002141,
		0.00167,
		0.000774,
		3.2e-005,
		4.8e-005,
		0.000226,
		0.000642,
		0.001351,
		0.002303,
		0.003312,
		0.00408,
		0.004271,
		0.003651,
		0.002263,
		0.000644,
		4.9e-005,
		0.000235,
		0.000698,
		0.001555,
		0.002835,
		0.004423,
		0.006041,
		0.007282,
		0.007705,
		0.006974,
		0.005051,
		0.002411,
		0.000268,
		4.9e-005,
		0.000237,
		0.000718,
		0.001659,
		0.003177,
		0.005261,
		0.007717,
		0.01016,
		0.012061,
		0.012854,
		0.012091,
		0.009635,
		0.005877,
		0.001962,
		4.8e-005,
		0.000237,
		0.000719,
		0.001687,
		0.003331,
		0.00576,
		0.00891,
		0.012498,
		0.016015,
		0.018798,
		0.020142,
		0.019471,
		0.016531,
		0.011614,
		0.005777,
		0.001043,
		4.5e-005,
		0.000232,
		0.000718,
		0.001688,
		0.003363,
		0.005956,
		0.009563,
		0.014064,
		0.01907,
		0.023941,
		0.027863,
		0.02998,
		0.029569,
		0.02625,
		0.020197,
		0.012361,
		0.004664,
		0.000161,
		4e-005,
		0.000216,
		0.000698,
		0.001679,
		0.003363,
		0.005985,
		0.00978,
		0.014841,
		0.020985,
		0.027714,
		0.034243,
		0.039592,
		0.04274,
		0.042807,
		0.039266,
		0.032163,
		0.02233,
		0.011571,
		0.002816,
		3.1e-005,
		0.000185,
		0.000642,
		0.001619,
		0.003331,
		0.005981,
		0.0098,
		0.01505,
		0.021831,
		0.02991,
		0.038672,
		0.047176,
		0.054265,
		0.058738,
		0.05954,
		0.055989,
		0.047987,
		0.036236,
		0.022412,
		0.009308,
		0.000913,
		2e-005,
		0.000141,
		0.00054,
		0.001466,
		0.003177,
		0.005889,
		0.00978,
		0.015059,
		0.022005,
		0.030755,
		0.041042,
		0.052138,
		0.062938,
		0.072094,
		0.078208,
		0.080042,
		0.076739,
		0.068049,
		0.054533,
		0.037732,
		0.020295,
		0.00605,
		9e-005,
		0.000402,
		0.001213,
		0.002835,
		0.005553,
		0.009563,
		0.014994,
		0.022005,
		0.030877,
		0.041813,
		0.054548,
		0.068256,
		0.081659,
		0.09321,
		0.101297,
		0.104482,
		0.101724,
		0.092608,
		0.077548,
		0.057942,
		0.036295,
		0.01627,
		0.002672,
		4.1e-005,
		0.000248,
		0.000882,
		0.002303,
		0.004882,
		0.00891,
		0.014543,
		0.021831,
		0.030859,
		0.041882,
		0.055184,
		0.070559,
		0.087112,
		0.103399,
		0.117656,
		0.128047,
		0.132916,
		0.131026,
		0.121789,
		0.105461,
		0.083299,
		0.057659,
		0.032043,
		0.011064,
		0.000352,
		0.000106,
		0.000524,
		0.001636,
		0.003894,
		0.007717,
		0.013384,
		0.020985,
		0.030469,
		0.041813,
		0.055213,
		0.071024,
		0.089172,
		0.108744,
		0.128143,
		0.145385,
		0.158393,
		0.165276,
		0.164591,
		0.155564,
		0.138292,
		0.113883,
		0.084551,
		0.053638,
		0.025572,
		0.005759,
		1.4e-005,
		0.000207,
		0.000936,
		0.002699,
		0.006041,
		0.011415,
		0.01907,
		0.029006,
		0.041042,
		0.055015,
		0.071024,
		0.089464,
		0.110452,
		0.13314,
		0.155812,
		0.176263,
		0.192162,
		0.201372,
		0.202223,
		0.193752,
		0.175889,
		0.149594,
		0.116937,
		0.08111,
		0.046363,
		0.017889,
		0.001638,
		1.8e-005,
		0.000339,
		0.001481,
		0.00408,
		0.008767,
		0.016015,
		0.026022,
		0.038672,
		0.05363,
		0.070559,
		0.089421,
		0.110603,
		0.134436,
		0.160248,
		0.186264,
		0.210073,
		0.229082,
		0.240889,
		0.24359,
		0.236018,
		0.217932,
		0.190145,
		0.154586,
		0.1143,
		0.073371,
		0.036791,
		0.010267,
		1.6e-005,
		0.000485,
		0.002141,
		0.005762,
		0.012061,
		0.021513,
		0.034243,
		0.049997,
		0.068256,
		0.08848,
		0.110452,
		0.134495,
		0.161129,
		0.189978,
		0.219306,
		0.246526,
		0.268784,
		0.283404,
		0.288229,
		0.281881,
		0.263942,
		0.235077,
		0.197077,
		0.152846,
		0.106308,
		0.062266,
		0.026203,
		0.004055,
		0.000623,
		0.00288,
		0.007705,
		0.015878,
		0.027863,
		0.043688,
		0.062938,
		0.08488,
		0.108744,
		0.134094,
		0.161129,
		0.190497,
		0.222209,
		0.254699,
		0.285274,
		0.310826,
		0.328396,
		0.335565,
		0.330732,
		0.313298,
		0.283774,
		0.243819,
		0.196197,
		0.144681,
		0.093895,
		0.049109,
		0.016015,
		0.000483,
		0.000727,
		0.003651,
		0.009842,
		0.020142,
		0.03498,
		0.054265,
		0.077394,
		0.103399,
		0.131247,
		0.160248,
		0.190422,
		0.222461,
		0.256786,
		0.292176,
		0.325921,
		0.354699,
		0.375266,
		0.38493,
		0.381856,
		0.365257,
		0.33549,
		0.294076,
		0.243645,
		0.187825,
		0.131066,
		0.078437,
		0.0354,
		0.007571,
		0.000774,
		0.004394,
		0.012091,
		0.024749,
		0.04274,
		0.065832,
		0.09321,
		0.123644,
		0.155812,
		0.188738,
		0.222209,
		0.256879,
		0.293526,
		0.33144,
		0.368043,
		0.399855,
		0.423361,
		0.435589,
		0.434461,
		0.418991,
		0.389377,
		0.347001,
		0.294361,
		0.234937,
		0.173013,
		0.113465,
		0.061533,
		0.0226,
		0.001976,
		0.000748,
		0.005051,
		0.014355,
		0.029569,
		0.050988,
		0.078208,
		0.110179,
		0.145385,
		0.182192,
		0.219306,
		0.256238,
		0.293526,
		0.3322,
		0.372177,
		0.411201,
		0.445728,
		0.472002,
		0.486772,
		0.487707,
		0.473615,
		0.444525,
		0.401677,
		0.347429,
		0.285117,
		0.218859,
		0.153342,
		0.093595,
		0.044773,
		0.011958,
		0.000644,
		0.005563,
		0.016531,
		0.03446,
		0.05954,
		0.091174,
		0.128047,
		0.168337,
		0.210073,
		0.251619,
		0.292176,
		0.33209,
		0.372533,
		0.414057,
		0.454953,
		0.491749,
		0.520504,
		0.537699,
		0.540744,
		0.528227,
		0.5,
		0.45715,
		0.401894,
		0.337415,
		0.267665,
		0.197141,
		0.130667,
		0.073179,
		0.029542,
		0.0044,
		0.000473,
		0.005877,
		0.018512,
		0.039266,
		0.068196,
		0.104482,
		0.146521,
		0.192162,
		0.239079,
		0.285274,
		0.329606,
		0.372177,
		0.414182,
		0.456733,
		0.498867,
		0.537373,
		0.568211,
		0.587617,
		0.592741,
		0.581942,
		0.55488,
		0.512479,
		0.456801,
		0.390871,
		0.318468,
		0.243899,
		0.171783,
		0.106843,
		0.053735,
		0.016906,
		0.000494,
		0.000268,
		0.005956,
		0.020197,
		0.043827,
		0.076739,
		0.11786,
		0.165276,
		0.216483,
		0.268784,
		0.319805,
		0.368043,
		0.413311,
		0.456733,
		0.499835,
		0.542528,
		0.582095,
		0.614514,
		0.635826,
		0.642924,
		0.633929,
		0.608295,
		0.566765,
		0.511233,
		0.444555,
		0.370327,
		0.292662,
		0.215973,
		0.144773,
		0.083512,
		0.036449,
		0.007563,
		8.4e-005,
		0.005777,
		0.021495,
		0.047987,
		0.084949,
		0.131026,
		0.183964,
		0.240889,
		0.298721,
		0.354699,
		0.406945,
		0.454953,
		0.499696,
		0.542966,
		0.585544,
		0.625459,
		0.658875,
		0.681705,
		0.690604,
		0.683445,
		0.659459,
		0.61919,
		0.564348,
		0.497602,
		0.422357,
		0.342523,
		0.262302,
		0.185998,
		0.11786,
		0.061965,
		0.022135,
		0.001888,
		0.005339,
		0.02233,
		0.0516,
		0.092608,
		0.143689,
		0.202223,
		0.26495,
		0.328396,
		0.389406,
		0.445728,
		0.496513,
		0.542528,
		0.585694,
		0.627545,
		0.667076,
		0.700845,
		0.724736,
		0.735198,
		0.729854,
		0.707694,
		0.669042,
		0.615402,
		0.549241,
		0.473757,
		0.392648,
		0.309899,
		0.229604,
		0.155812,
		0.092422,
		0.043104,
		0.011259,
		0.004664,
		0.022647,
		0.054533,
		0.099509,
		0.155564,
		0.219692,
		0.288229,
		0.3573,
		0.423361,
		0.483784,
		0.537373,
		0.584648,
		0.627545,
		0.668179,
		0.706621,
		0.74007,
		0.764515,
		0.77625,
		0.77265,
		0.75245,
		0.715733,
		0.663773,
		0.598814,
		0.523831,
		0.442298,
		0.357979,
		0.274752,
		0.19647,
		0.126853,
		0.069425,
		0.027471,
		0.004022,
		0.003798,
		0.022412,
		0.056675,
		0.105461,
		0.166381,
		0.236018,
		0.310294,
		0.38493,
		0.455999,
		0.520504,
		0.57691,
		0.625459,
		0.668022,
		0.707107,
		0.743835,
		0.776299,
		0.800764,
		0.813438,
		0.811468,
		0.793321,
		0.758816,
		0.708972,
		0.645788,
		0.572,
		0.490845,
		0.405858,
		0.320699,
		0.239021,
		0.164372,
		0.100129,
		0.049465,
		0.015327,
		0.000442,
		0.002816,
		0.021621,
		0.057942,
		0.110298,
		0.175889,
		0.250868,
		0.330732,
		0.4108,
		0.486772,
		0.555299,
		0.614514,
		0.664376,
		0.706621,
		0.743996,
		0.77852,
		0.809382,
		0.833327,
		0.84658,
		0.846087,
		0.830044,
		0.797981,
		0.750642,
		0.689759,
		0.617806,
		0.537774,
		0.452958,
		0.366798,
		0.282748,
		0.204182,
		0.134336,
		0.076267,
		0.032845,
		0.00674,
		0.001814,
		0.020295,
		0.058278,
		0.113883,
		0.18387,
		0.263942,
		0.349164,
		0.434461,
		0.515172,
		0.587617,
		0.649616,
		0.700845,
		0.742861,
		0.77852,
		0.810529,
		0.839261,
		0.86217,
		0.875627,
		0.876427,
		0.862495,
		0.83306,
		0.788563,
		0.730447,
		0.660911,
		0.582681,
		0.498808,
		0.412507,
		0.327031,
		0.245582,
		0.171255,
		0.106998,
		0.055596,
		0.019668,
		0.001665,
		0.000913,
		0.018485,
		0.057659,
		0.116117,
		0.190145,
		0.274981,
		0.365257,
		0.455512,
		0.540744,
		0.616965,
		0.681705,
		0.734367,
		0.776299,
		0.810366,
		0.839755,
		0.865957,
		0.887367,
		0.900656,
		0.902538,
		0.890684,
		0.864009,
		0.822636,
		0.767694,
		0.701088,
		0.62527,
		0.543038,
		0.457378,
		0.371341,
		0.287961,
		0.21019,
		0.140868,
		0.082699,
		0.038245,
		0.009922,
		0.000256,
		0.01627,
		0.056098,
		0.116937,
		0.194576,
		0.283774,
		0.378733,
		0.473615,
		0.563102,
		0.642924,
		0.710347,
		0.764515,
		0.806556,
		0.839261,
		0.866118,
		0.889556,
		0.909081,
		0.921848,
		0.924584,
		0.914735,
		0.890903,
		0.852871,
		0.801442,
		0.73821,
		0.66534,
		0.585368,
		0.501051,
		0.415238,
		0.330792,
		0.250527,
		0.177175,
		0.113359,
		0.061582,
		0.024222,
		0.003529,
		0.013754,
		0.053638,
		0.116324,
		0.197077,
		0.29017,
		0.389377,
		0.488506,
		0.581942,
		0.665163,
		0.735198,
		0.790953,
		0.833327,
		0.864981,
		0.889556,
		0.910191,
		0.927549,
		0.939475,
		0.942826,
		0.934871,
		0.913908,
		0.879373,
		0.831725,
		0.772234,
		0.702768,
		0.625599,
		0.543246,
		0.458357,
		0.373628,
		0.291735,
		0.215301,
		0.146868,
		0.088879,
		0.043674,
		0.013474,
		0.000387,
		0.011064,
		0.050357,
		0.1143,
		0.197612,
		0.294076,
		0.397048,
		0.5,
		0.597047,
		0.683445,
		0.756013,
		0.813438,
		0.856393,
		0.887367,
		0.910034,
		0.928019,
		0.943055,
		0.953873,
		0.957597,
		0.951392,
		0.93327,
		0.90232,
		0.858647,
		0.803188,
		0.737504,
		0.663597,
		0.583747,
		0.500403,
		0.416091,
		0.333351,
		0.254699,
		0.182595,
		0.11942,
		0.067471,
		0.028941,
		0.00592,
		0.008346,
		0.046363,
		0.110928,
		0.196197,
		0.295462,
		0.401677,
		0.507996,
		0.608295,
		0.697635,
		0.77265,
		0.83183,
		0.875627,
		0.906329,
		0.927549,
		0.943206,
		0.955914,
		0.96542,
		0.969283,
		0.964651,
		0.949288,
		0.921945,
		0.882368,
		0.831152,
		0.769549,
		0.699281,
		0.622395,
		0.541136,
		0.457861,
		0.374977,
		0.294891,
		0.21998,
		0.152562,
		0.094888,
		0.049119,
		0.017323,
		0.001462,
		0.005759,
		0.04179,
		0.106308,
		0.192897,
		0.294361,
		0.403267,
		0.512479,
		0.615657,
		0.707694,
		0.785067,
		0.846087,
		0.890993,
		0.921848,
		0.942136,
		0.955914,
		0.966441,
		0.974514,
		0.978299,
		0.975038,
		0.962301,
		0.938519,
		0.903084,
		0.856247,
		0.798942,
		0.732617,
		0.659076,
		0.580364,
		0.498671,
		0.416269,
		0.335457,
		0.258526,
		0.187731,
		0.125272,
		0.073275,
		0.033782,
		0.008741,
		0.003467,
		0.036791,
		0.100576,
		0.187825,
		0.290865,
		0.401894,
		0.513512,
		0.61919,
		0.713679,
		0.793321,
		0.856264,
		0.902538,
		0.933976,
		0.953873,
		0.966297,
		0.974941,
		0.98155,
		0.985067,
		0.982956,
		0.972665,
		0.952336,
		0.921018,
		0.878618,
		0.825756,
		0.763598,
		0.69371,
		0.617935,
		0.538297,
		0.456932,
		0.376029,
		0.297794,
		0.224414,
		0.158037,
		0.100747,
		0.05456,
		0.021402,
		0.003111,
		0.001638,
		0.031536,
		0.093895,
		0.181135,
		0.285117,
		0.397697,
		0.511233,
		0.619035,
		0.715733,
		0.797554,
		0.862495,
		0.910389,
		0.942826,
		0.962873,
		0.974514,
		0.981686,
		0.986903,
		0.99,
		0.988808,
		0.980739,
		0.963696,
		0.936403,
		0.898429,
		0.85008,
		0.792245,
		0.726248,
		0.653732,
		0.576554,
		0.496713,
		0.416289,
		0.337397,
		0.262156,
		0.192655,
		0.130939,
		0.078987,
		0.038704,
		0.011912,
		0.000342,
		0.000432,
		0.026203,
		0.086449,
		0.173013,
		0.277308,
		0.390871,
		0.505845,
		0.615402,
		0.714073,
		0.797981,
		0.864989,
		0.914735,
		0.948562,
		0.969283,
		0.980718,
		0.986903,
		0.990912,
		0.993485,
		0.992978,
		0.986877,
		0.9729,
		0.949475,
		0.915849,
		0.872018,
		0.818594,
		0.756663,
		0.687664,
		0.613289,
		0.535398,
		0.455958,
		0.376995,
		0.300551,
		0.228658,
		0.163312,
		0.106458,
		0.059973,
		0.025659,
		0.005236,
		0.020971,
		0.078437,
		0.163672,
		0.267665,
		0.381656,
		0.497602,
		0.608558,
		0.708972,
		0.794878,
		0.864009,
		0.915816,
		0.951392,
		0.973275,
		0.985067,
		0.990785,
		0.993865,
		0.995868,
		0.995825,
		0.991411,
		0.980234,
		0.960466,
		0.931051,
		0.891682,
		0.842698,
		0.784947,
		0.719666,
		0.648379,
		0.572806,
		0.494799,
		0.416289,
		0.339242,
		0.265625,
		0.197385,
		0.136427,
		0.084596,
		0.043674,
		0.015366,
		0.001294,
		0.016015,
		0.070064,
		0.153342,
		0.256439,
		0.370327,
		0.486799,
		0.598814,
		0.700751,
		0.788563,
		0.859861,
		0.913908,
		0.95155,
		0.975038,
		0.987715,
		0.993485,
		0.995988,
		0.99745,
		0.997669,
		0.994648,
		0.98597,
		0.969601,
		0.944206,
		0.909189,
		0.864619,
		0.811109,
		0.749695,
		0.681729,
		0.608789,
		0.532609,
		0.455022,
		0.377913,
		0.303183,
		0.232725,
		0.168399,
		0.112019,
		0.065341,
		0.030049,
		0.007758,
		0.011499,
		0.061533,
		0.142261,
		0.243899,
		0.35718,
		0.473757,
		0.58651,
		0.689759,
		0.779382,
		0.852871,
		0.909308,
		0.949288,
		0.974769,
		0.988808,
		0.995131,
		0.99745,
		0.998473,
		0.998789,
		0.996864,
		0.990359,
		0.977092,
		0.955481,
		0.924657,
		0.884429,
		0.835175,
		0.777728,
		0.713267,
		0.643226,
		0.569217,
		0.492972,
		0.416289,
		0.341002,
		0.268945,
		0.201932,
		0.14174,
		0.090096,
		0.048665,
		0.019046,
		0.002763,
		0.007571,
		0.053042,
		0.130667,
		0.23032,
		0.342523,
		0.458811,
		0.572,
		0.676358,
		0.767694,
		0.84338,
		0.90232,
		0.944861,
		0.972665,
		0.988486,
		0.995825,
		0.998361,
		0.999123,
		0.999416,
		0.998305,
		0.993631,
		0.983139,
		0.965039,
		0.938211,
		0.90221,
		0.857183,
		0.803759,
		0.742944,
		0.676022,
		0.604481,
		0.529947,
		0.45413,
		0.37879,
		0.3057,
		0.236626,
		0.173305,
		0.117431,
		0.070648,
		0.034535,
		0.010606,
		0.000304,
		0.004361,
		0.044773,
		0.118789,
		0.215973,
		0.326664,
		0.442298,
		0.555639,
		0.660911,
		0.753855,
		0.831725,
		0.893244,
		0.938519,
		0.968915,
		0.986877,
		0.995644,
		0.998789,
		0.999522,
		0.999737,
		0.999178,
		0.995994,
		0.987932,
		0.973042,
		0.949977,
		0.918052,
		0.877185,
		0.827802,
		0.770731,
		0.707107,
		0.638287,
		0.565786,
		0.491227,
		0.416289,
		0.342684,
		0.272124,
		0.206304,
		0.146884,
		0.095481,
		0.053654,
		0.022905,
		0.004665,
		0.001976,
		0.036891,
		0.106843,
		0.201119,
		0.309899,
		0.424544,
		0.537774,
		0.643772,
		0.73821,
		0.818227,
		0.882368,
		0.930499,
		0.963696,
		0.984094,
		0.994648,
		0.998768,
		0.999737,
		0.999885,
		0.999659,
		0.997633,
		0.991646,
		0.979644,
		0.960083,
		0.932051,
		0.895243,
		0.849883,
		0.796619,
		0.73643,
		0.670542,
		0.600358,
		0.527402,
		0.453277,
		0.379627,
		0.308108,
		0.24037,
		0.178037,
		0.122695,
		0.075884,
		0.039085,
		0.013723,
		0.001154,
		0.000502,
		0.029542,
		0.095026,
		0.185998,
		0.29251,
		0.405858,
		0.51873,
		0.62527,
		0.721083,
		0.803188,
		0.869955,
		0.921018,
		0.957169,
		0.980234,
		0.992877,
		0.998305,
		0.999787,
		0.999952,
		0.999888,
		0.998711,
		0.994447,
		0.984999,
		0.968661,
		0.944313,
		0.911434,
		0.870045,
		0.820615,
		0.763962,
		0.701178,
		0.63355,
		0.562504,
		0.489559,
		0.416289,
		0.344293,
		0.275172,
		0.210511,
		0.151865,
		0.100747,
		0.058623,
		0.026902,
		0.006932,
		0.022848,
		0.083512,
		0.170831,
		0.274752,
		0.386524,
		0.498808,
		0.605712,
		0.702768,
		0.786881,
		0.856247,
		0.910271,
		0.949475,
		0.97538,
		0.990359,
		0.997388,
		0.999659,
		0.999981,
		0.999976,
		0.999373,
		0.996489,
		0.989256,
		0.975845,
		0.954949,
		0.925841,
		0.888342,
		0.842738,
		0.789689,
		0.730146,
		0.665277,
		0.596407,
		0.524968,
		0.452462,
		0.380428,
		0.310415,
		0.243966,
		0.182603,
		0.127813,
		0.081039,
		0.043674,
		0.017058,
		0.00247,
		0.016906,
		0.07245,
		0.155812,
		0.256856,
		0.366798,
		0.478279,
		0.585368,
		0.68353,
		0.769549,
		0.841453,
		0.898429,
		0.940734,
		0.969601,
		0.98711,
		0.995994,
		0.999317,
		0.999976,
		0.999997,
		0.999741,
		0.997917,
		0.992561,
		0.981771,
		0.964076,
		0.938558,
		0.904839,
		0.863025,
		0.813614,
		0.757413,
		0.69547,
		0.629005,
		0.559361,
		0.487964,
		0.416289,
		0.345833,
		0.278096,
		0.214561,
		0.156687,
		0.105894,
		0.063557,
		0.031004,
		0.009504,
		0.000272,
		0.011792,
		0.061965,
		0.141106,
		0.239021,
		0.346903,
		0.457378,
		0.564478,
		0.663597,
		0.751401,
		0.825756,
		0.885637,
		0.931051,
		0.962953,
		0.983139,
		0.994091,
		0.998711,
		0.999911,
		1,
		0.999917,
		0.998863,
		0.995056,
		0.986574,
		0.971816,
		0.949683,
		0.91961,
		0.881522,
		0.835752,
		0.782964,
		0.724081,
		0.660215,
		0.592618,
		0.522638,
		0.451682,
		0.381194,
		0.312626,
		0.247423,
		0.187012,
		0.132788,
		0.086107,
		0.048281,
		0.02057,
		0.004182,
		0.007563,
		0.052157,
		0.126853,
		0.221416,
		0.327031,
		0.436307,
		0.543246,
		0.643164,
		0.732617,
		0.80931,
		0.872018,
		0.920512,
		0.955481,
		0.978449,
		0.991646,
		0.997788,
		0.999741,
		0.999999,
		0.999984,
		0.999446,
		0.996876,
		0.99039,
		0.978292,
		0.959323,
		0.932739,
		0.898284,
		0.856131,
		0.806797,
		0.751077,
		0.689972,
		0.62464,
		0.556348,
		0.486437,
		0.416289,
		0.34731,
		0.280903,
		0.218463,
		0.161357,
		0.110919,
		0.068445,
		0.035182,
		0.012329,
		0.001035,
		0.004255,
		0.043104,
		0.113164,
		0.204182,
		0.30734,
		0.415238,
		0.52184,
		0.622395,
		0.713346,
		0.792245,
		0.857677,
		0.909189,
		0.947222,
		0.973042,
		0.988624,
		0.996489,
		0.999412,
		0.999984,
		0.999999,
		0.999771,
		0.998148,
		0.99335,
		0.983628,
		0.967586,
		0.944313,
		0.913379,
		0.874792,
		0.828924,
		0.776439,
		0.718226,
		0.655346,
		0.588982,
		0.520404,
		0.450936,
		0.381929,
		0.314747,
		0.250747,
		0.191269,
		0.137622,
		0.091084,
		0.052888,
		0.024223,
		0.006231,
		0.001888,
		0.034862,
		0.100129,
		0.187433,
		0.287961,
		0.394307,
		0.500403,
		0.601425,
		0.69371,
		0.774666,
		0.842698,
		0.897143,
		0.938211,
		0.966918,
		0.984999,
		0.994763,
		0.998863,
		0.999922,
		1,
		0.999927,
		0.998989,
		0.995583,
		0.987949,
		0.974587,
		0.954428,
		0.926878,
		0.891784,
		0.849368,
		0.800163,
		0.744944,
		0.684672,
		0.620445,
		0.553458,
		0.484973,
		0.416289,
		0.348725,
		0.283601,
		0.222223,
		0.165879,
		0.115823,
		0.073275,
		0.039412,
		0.015366,
		0.002221,
		0.000471,
		0.027471,
		0.087814,
		0.171255,
		0.268996,
		0.373628,
		0.479047,
		0.580364,
		0.673812,
		0.756663,
		0.827155,
		0.884429,
		0.928478,
		0.960083,
		0.980744,
		0.992561,
		0.998037,
		0.999771,
		0.999999,
		0.999985,
		0.999508,
		0.99721,
		0.991379,
		0.980438,
		0.963183,
		0.938862,
		0.907163,
		0.868161,
		0.822257,
		0.770107,
		0.71257,
		0.650658,
		0.58549,
		0.518262,
		0.450219,
		0.382633,
		0.316784,
		0.253947,
		0.195381,
		0.14232,
		0.095965,
		0.057479,
		0.027987,
		0.008564,
		0.000244,
		0.020955,
		0.076267,
		0.155717,
		0.250527,
		0.353287,
		0.457861,
		0.559298,
		0.653732,
		0.738309,
		0.811109,
		0.871094,
		0.918052,
		0.952544,
		0.975845,
		0.989843,
		0.996876,
		0.999478,
		0.999985,
		0.999999,
		0.999797,
		0.998347,
		0.994038,
		0.985256,
		0.970681,
		0.949417,
		0.920996,
		0.885348,
		0.842738,
		0.793707,
		0.739008,
		0.679561,
		0.61641,
		0.550683,
		0.483569,
		0.416289,
		0.350084,
		0.286194,
		0.225849,
		0.170259,
		0.120607,
		0.078042,
		0.043674,
		0.018575,
		0.003771,
		0.015327,
		0.065521,
		0.140868,
		0.232614,
		0.333351,
		0.436917,
		0.538297,
		0.633539,
		0.719666,
		0.794617,
		0.857183,
		0.906963,
		0.944313,
		0.970291,
		0.986574,
		0.995329,
		0.998989,
		0.999931,
		1,
		0.999935,
		0.999098,
		0.996042,
		0.989154,
		0.977027,
		0.958631,
		0.933353,
		0.900978,
		0.861636,
		0.815748,
		0.763962,
		0.707107,
		0.646144,
		0.582134,
		0.516206,
		0.449532,
		0.383309,
		0.318742,
		0.257029,
		0.199356,
		0.146884,
		0.100747,
		0.062044,
		0.031834,
		0.011138,
		0.000934,
		0.01059,
		0.055596,
		0.126741,
		0.215301,
		0.313873,
		0.416269,
		0.51742,
		0.613289,
		0.70079,
		0.777728,
		0.842738,
		0.895243,
		0.935405,
		0.964076,
		0.98273,
		0.99335,
		0.998251,
		0.999797,
		0.999999,
		0.999987,
		0.999561,
		0.997501,
		0.992245,
		0.982327,
		0.9666,
		0.944313,
		0.915109,
		0.878987,
		0.836245,
		0.787425,
		0.73326,
		0.674631,
		0.612528,
		0.548017,
		0.482221,
		0.416289,
		0.35139,
		0.28869,
		0.229348,
		0.174504,
		0.125272,
		0.082738,
		0.047952,
		0.021925,
		0.005631,
		0.00674,
		0.046503,
		0.113359,
		0.198621,
		0.294891,
		0.395962,
		0.496713,
		0.59303,
		0.681729,
		0.760489,
		0.827802,
		0.882926,
		0.925841,
		0.957205,
		0.978292,
		0.990902,
		0.99721,
		0.999535,
		0.999987,
		0.999999,
		0.999819,
		0.99852,
		0.99464,
		0.986687,
		0.973418,
		0.953955,
		0.927805,
		0.894835,
		0.855222,
		0.809397,
		0.758,
		0.701826,
		0.641793,
		0.578906,
		0.51423,
		0.448872,
		0.383959,
		0.320624,
		0.259999,
		0.203198,
		0.151319,
		0.105431,
		0.066571,
		0.035743,
		0.013913,
		0.002008,
		0.003769,
		0.038245,
		0.100736,
		0.182595,
		0.276435,
		0.376029,
		0.476214,
		0.572806,
		0.662527,
		0.742944,
		0.812415,
		0.870045,
		0.915644,
		0.949683,
		0.973246,
		0.987949,
		0.99582,
		0.999098,
		0.999939,
		1,
		0.999942,
		0.999193,
		0.996443,
		0.990212,
		0.979184,
		0.962366,
		0.939136,
		0.909232,
		0.872709,
		0.82989,
		0.781314,
		0.727694,
		0.669872,
		0.608789,
		0.545454,
		0.480925,
		0.416289,
		0.352645,
		0.291093,
		0.232725,
		0.178617,
		0.12982,
		0.08736,
		0.05223,
		0.025389,
		0.007758,
		0.000221,
		0.001665,
		0.03082,
		0.088879,
		0.167239,
		0.258526,
		0.3565,
		0.455958,
		0.552654,
		0.643226,
		0.725136,
		0.796619,
		0.856637,
		0.904839,
		0.941522,
		0.967586,
		0.984468,
		0.994038,
		0.998436,
		0.999819,
		0.999999,
		0.999988,
		0.999607,
		0.997755,
		0.993005,
		0.983995,
		0.969633,
		0.949176,
		0.922235,
		0.888746,
		0.848922,
		0.803203,
		0.752213,
		0.69672,
		0.637598,
		0.575799,
		0.512331,
		0.448238,
		0.384583,
		0.322436,
		0.262862,
		0.206915,
		0.155629,
		0.110014,
		0.071053,
		0.039695,
		0.016857,
		0.003417,
		0.000414,
		0.024222,
		0.077791,
		0.152562,
		0.241181,
		0.337397,
		0.435974,
		0.532609,
		0.623865,
		0.707107,
		0.780455,
		0.842738,
		0.893455,
		0.932739,
		0.961312,
		0.980438,
		0.991827,
		0.997501,
		0.999585,
		0.999988,
		0.999999,
		0.999838,
		0.998671,
		0.995167,
		0.987949,
		0.975845,
		0.958001,
		0.933906,
		0.903379,
		0.86652,
		0.823672,
		0.77537,
		0.722303,
		0.665277,
		0.605186,
		0.542987,
		0.47968,
		0.416289,
		0.353853,
		0.293408,
		0.235987,
		0.182603,
		0.134253,
		0.091904,
		0.056497,
		0.028941,
		0.010111,
		0.000846,
		0.018444,
		0.067471,
		0.138572,
		0.224414,
		0.318742,
		0.416289,
		0.512705,
		0.604481,
		0.688896,
		0.763962,
		0.828385,
		0.881522,
		0.92335,
		0.954428,
		0.975845,
		0.989154,
		0.996249,
		0.999193,
		0.999945,
		1,
		0.999948,
		0.999276,
		0.996795,
		0.991144,
		0.981094,
		0.965692,
		0.944313,
		0.916658,
		0.882719,
		0.842738,
		0.797161,
		0.746597,
		0.691782,
		0.63355,
		0.572806,
		0.510503,
		0.447628,
		0.385184,
		0.324181,
		0.265625,
		0.210511,
		0.159817,
		0.114498,
		0.075484,
		0.043674,
		0.019939,
		0.005114,
		0.013474,
		0.057915,
		0.125272,
		0.208236,
		0.300551,
		0.396928,
		0.492972,
		0.585111,
		0.670542,
		0.74718,
		0.813614,
		0.86907,
		0.913379,
		0.94694,
		0.970681,
		0.985994,
		0.99464,
		0.998598,
		0.999838,
		0.999999,
		0.99999,
		0.999648,
		0.997978,
		0.993675,
		0.985471,
		0.972331,
		0.953525,
		0.928639,
		0.897559,
		0.860425,
		0.817592,
		0.769588,
		0.717079,
		0.660837,
		0.601712,
		0.540612,
		0.478481,
		0.416289,
		0.355015,
		0.295641,
		0.239139,
		0.186469,
		0.138573,
		0.096367,
		0.060743,
		0.032563,
		0.012657,
		0.001825,
		0.009304,
		0.049119,
		0.112664,
		0.192655,
		0.282841,
		0.377913,
		0.473438,
		0.565786,
		0.652082,
		0.730146,
		0.798462,
		0.856131,
		0.902848,
		0.938862,
		0.964943,
		0.982327,
		0.992637,
		0.997755,
		0.999628,
		0.99999,
		0.999999,
		0.999854,
		0.998803,
		0.995632,
		0.989067,
		0.978003,
		0.961616,
		0.939384,
		0.911086,
		0.876762,
		0.836674,
		0.79127,
		0.741144,
		0.687003,
		0.629643,
		0.569922,
		0.508744,
		0.447041,
		0.385763,
		0.325863,
		0.268292,
		0.213992,
		0.163887,
		0.118882,
		0.079857,
		0.047667,
		0.023136,
		0.00706,
		0.000201,
		0.00592,
		0.041077,
		0.100747,
		0.177679,
		0.265625,
		0.359265,
		0.45413,
		0.54654,
		0.63355,
		0.712898,
		0.782964,
		0.842738,
		0.891784,
		0.930208,
		0.958631,
		0.978139,
		0.990212,
		0.996625,
		0.999276,
		0.999951,
		1,
		0.999954,
		0.999348,
		0.997104,
		0.991969,
		0.982792,
		0.968661,
		0.948956,
		0.92335,
		0.891784,
		0.854428,
		0.811648,
		0.763962,
		0.712016,
		0.656545,
		0.598361,
		0.538323,
		0.477327,
		0.416289,
		0.356136,
		0.297795,
		0.242186,
		0.190218,
		0.142782,
		0.100747,
		0.064959,
		0.036235,
		0.015366,
		0.003111,
		0.00331,
		0.033782,
		0.089522,
		0.163312,
		0.248916,
		0.341002,
		0.435073,
		0.527402,
		0.614981,
		0.69547,
		0.767158,
		0.828924,
		0.880212,
		0.920996,
		0.951752,
		0.973418,
		0.987338,
		0.995167,
		0.998739,
		0.999854,
		0.999999,
		0.999991,
		0.999683,
		0.998174,
		0.994266,
		0.986782,
		0.974739,
		0.957423,
		0.934406,
		0.905532,
		0.870879,
		0.830729,
		0.785527,
		0.73585,
		0.682378,
		0.625869,
		0.56714,
		0.507048,
		0.446476,
		0.38632,
		0.327485,
		0.270869,
		0.217363,
		0.167843,
		0.123166,
		0.084169,
		0.05166,
		0.026425,
		0.00922,
		0.000771,
		0.001462,
		0.027226,
		0.078987,
		0.149561,
		0.232725,
		0.323142,
		0.416289,
		0.5084,
		0.596407,
		0.677898,
		0.751077,
		0.814721,
		0.868161,
		0.911245,
		0.944313,
		0.968161,
		0.983995,
		0.99335,
		0.997978,
		0.999666,
		0.999991,
		0.999999,
		0.999869,
		0.99892,
		0.996042,
		0.990058,
		0.979927,
		0.964853,
		0.944313,
		0.918052,
		0.886059,
		0.848532,
		0.805839,
		0.75849,
		0.707107,
		0.652395,
		0.595126,
		0.536117,
		0.476215,
		0.416289,
		0.357216,
		0.299874,
		0.245133,
		0.193856,
		0.146884,
		0.105044,
		0.069138,
		0.039943,
		0.018211,
		0.004665,
		0.000363,
		0.021402,
		0.069138,
		0.136427,
		0.217061,
		0.3057,
		0.397801,
		0.489559,
		0.577857,
		0.660215,
		0.734755,
		0.800163,
		0.855661,
		0.900978,
		0.936326,
		0.962366,
		0.980168,
		0.991144,
		0.996954,
		0.999348,
		0.999956,
		1,
		0.999958,
		0.999412,
		0.997377,
		0.9927,
		0.984305,
		0.971319,
		0.953131,
		0.929393,
		0.900005,
		0.865077,
		0.824905,
		0.779928,
		0.730708,
		0.677898,
		0.622222,
		0.564456,
		0.505413,
		0.44593,
		0.386858,
		0.32905,
		0.273359,
		0.220628,
		0.171689,
		0.127354,
		0.088416,
		0.055645,
		0.029788,
		0.011564,
		0.001665,
		0.0163,
		0.059973,
		0.123912,
		0.201932,
		0.28869,
		0.379627,
		0.470906,
		0.559361,
		0.642452,
		0.718226,
		0.785281,
		0.842738,
		0.890217,
		0.927805,
		0.956035,
		0.975845,
		0.988525,
		0.995632,
		0.998863,
		0.999869,
		0.999999,
		0.999992,
		0.999714,
		0.998347,
		0.99479,
		0.987949,
		0.976892,
		0.960924,
		0.939609,
		0.912756,
		0.880394,
		0.842738,
		0.800163,
		0.753166,
		0.702346,
		0.64838,
		0.592002,
		0.533987,
		0.475143,
		0.416289,
		0.358259,
		0.301882,
		0.247986,
		0.197385,
		0.150882,
		0.109257,
		0.073275,
		0.043674,
		0.02117,
		0.006452,
		0.000183,
		0.011912,
		0.051487,
		0.112019,
		0.187344,
		0.272124,
		0.361785,
		0.452462,
		0.540946,
		0.62464,
		0.701521,
		0.770107,
		0.829424,
		0.878987,
		0.918768,
		0.949176,
		0.971021,
		0.985471,
		0.99398,
		0.998174,
		0.999699,
		0.999992,
		0.999999,
		0.999882,
		0.999022,
		0.996405,
		0.990939,
		0.981647,
		0.967759,
		0.948756,
		0.924359,
		0.894513,
		0.859359,
		0.8192,
		0.77447,
		0.725714,
		0.673559,
		0.618696,
		0.561865,
		0.503836,
		0.445405,
		0.387376,
		0.330561,
		0.275766,
		0.223793,
		0.175429,
		0.131445,
		0.092596,
		0.059613,
		0.033208,
		0.014064,
		0.002844,
		0.008227,
		0.043674,
		0.100747,
		0.173305,
		0.256014,
		0.344293,
		0.43425,
		0.522638,
		0.606807,
		0.684672,
		0.754672,
		0.815748,
		0.867313,
		0.909232,
		0.941795,
		0.965692,
		0.981968,
		0.991969,
		0.997245,
		0.999412,
		0.99996,
		1,
		0.999962,
		0.999468,
		0.997618,
		0.99335,
		0.985655,
		0.973702,
		0.956891,
		0.93486,
		0.907473,
		0.874792,
		0.837049,
		0.794618,
		0.747985,
		0.697728,
		0.644493,
		0.588982,
		0.531932,
		0.474108,
		0.416289,
		0.359265,
		0.303823,
		0.250747,
		0.200812,
		0.154777,
		0.113386,
		0.077365,
		0.047417,
		0.024223,
		0.008442,
		0.000705,
		0.005236,
		0.036531,
		0.090096,
		0.159817,
		0.24037,
		0.327165,
		0.416289,
		0.50446,
		0.588982,
		0.667708,
		0.739008,
		0.801739,
		0.855222,
		0.899218,
		0.933906,
		0.959858,
		0.978003,
		0.989577,
		0.996042,
		0.998972,
		0.999882,
		1,
		0.999993,
		0.999741,
		0.998499,
		0.995256,
		0.988991,
		0.978822,
		0.964076,
		0.944313,
		0.919314,
		0.889065,
		0.853729,
		0.813614,
		0.769149,
		0.720862,
		0.669354,
		0.615286,
		0.559361,
		0.502313,
		0.444897,
		0.387877,
		0.332021,
		0.278096,
		0.226862,
		0.179066,
		0.135442,
		0.096707,
		0.063557,
		0.03667,
		0.016698,
		0.004273,
		0.002929,
		0.030049,
		0.080064,
		0.146884,
		0.2252,
		0.310415,
		0.398598,
		0.486437,
		0.571191,
		0.650658,
		0.723143,
		0.787425,
		0.842738,
		0.888746,
		0.925522,
		0.953525,
		0.973569,
		0.986782,
		0.994536,
		0.998347,
		0.999728,
		0.999993,
		1,
		0.999893,
		0.999113,
		0.996728,
		0.991726,
		0.983187,
		0.970373,
		0.952769,
		0.930079,
		0.902212,
		0.869259,
		0.831466,
		0.789202,
		0.742944,
		0.693246,
		0.64073,
		0.586062,
		0.529947,
		0.473108,
		0.416289,
		0.360237,
		0.3057,
		0.253422,
		0.204139,
		0.158573,
		0.117431,
		0.081404,
		0.051161,
		0.027353,
		0.010606,
		0.001526,
		0.001294,
		0.024223,
		0.070648,
		0.13451,
		0.210511,
		0.294054,
		0.381194,
		0.468589,
		0.553458,
		0.63355,
		0.707107,
		0.772836,
		0.82989,
		0.877839,
		0.916658,
		0.946697,
		0.968661,
		0.98357,
		0.9927,
		0.997501,
		0.999468,
		0.999964,
		1,
		0.999966,
		0.999517,
		0.997833,
		0.99393,
		0.986864,
		0.975845,
		0.960286,
		0.939815,
		0.914268,
		0.883666,
		0.848188,
		0.808146,
		0.763962,
		0.716147,
		0.665277,
		0.611985,
		0.556941,
		0.500843,
		0.444407,
		0.38836,
		0.333432,
		0.280351,
		0.229837,
		0.182603,
		0.139346,
		0.100747,
		0.067471,
		0.040162,
		0.019444,
		0.00592,
		0.000168,
		0.000322,
		0.019046,
		0.061846,
		0.122695,
		0.19631,
		0.278096,
		0.364093,
		0.450936,
		0.535808,
		0.61641,
		0.690927,
		0.758,
		0.816702,
		0.86652,
		0.907331,
		0.939384,
		0.963278,
		0.979927,
		0.990511,
		0.996405,
		0.999069,
		0.999893,
		1,
		0.999993,
		0.999765,
		0.998635,
		0.995671,
		0.989923,
		0.980556,
		0.96692,
		0.948572,
		0.925275,
		0.89698,
		0.863799,
		0.825987,
		0.783913,
		0.738037,
		0.688896,
		0.637084,
		0.583238,
		0.528027,
		0.472143,
		0.416289,
		0.361176,
		0.307516,
		0.256014,
		0.207371,
		0.162273,
		0.121393,
		0.085389,
		0.054899,
		0.030544,
		0.012921,
		0.00261,
		0.014509,
		0.053654,
		0.111441,
		0.182603,
		0.262549,
		0.34731,
		0.433497,
		0.518262,
		0.599263,
		0.674631,
		0.742944,
		0.803203,
		0.854812,
		0.897559,
		0.931598,
		0.957423,
		0.975845,
		0.987949,
		0.99503,
		0.998499,
		0.999754,
		0.999993,
		1,
		0.999903,
		0.999193,
		0.997015,
		0.992429,
		0.984569,
		0.97273,
		0.956401,
		0.935275,
		0.909232,
		0.878322,
		0.842738,
		0.802795,
		0.758906,
		0.711563,
		0.661323,
		0.608789,
		0.5546,
		0.499421,
		0.443933,
		0.388827,
		0.334797,
		0.282534,
		0.232725,
		0.186045,
		0.14316,
		0.104717,
		0.07135,
		0.043674,
		0.022285,
		0.007758,
		0.000647,
		0.010606,
		0.046069,
		0.100747,
		0.169394,
		0.247423,
		0.330857,
		0.416289,
		0.500843,
		0.582134,
		0.658245,
		0.727694,
		0.789417,
		0.842738,
		0.887363,
		0.92335,
		0.951098,
		0.971319,
		0.984999,
		0.99335,
		0.997729,
		0.999517,
		0.999968,
		1,
		0.999969,
		0.999561,
		0.998024,
		0.994448,
		0.987949,
		0.977776,
		0.963355,
		0.944313,
		0.920461,
		0.891784,
		0.858415,
		0.820615,
		0.778747,
		0.73326,
		0.684672,
		0.63355,
		0.580504,
		0.526172,
		0.47121,
		0.416289,
		0.362085,
		0.309274,
		0.258527,
		0.210511,
		0.165879,
		0.125272,
		0.089317,
		0.058623,
		0.033782,
		0.015366,
		0.003928,
		0.007327,
		0.039085,
		0.090614,
		0.156687,
		0.232725,
		0.314747,
		0.39933,
		0.483569,
		0.565045,
		0.641793,
		0.712278,
		0.77537,
		0.830324,
		0.876762,
		0.914656,
		0.944313,
		0.966345,
		0.981647,
		0.991344,
		0.996728,
		0.999154,
		0.999903,
		1,
		0.999994,
		0.999787,
		0.998756,
		0.996042,
		0.990759,
		0.982117,
		0.969489,
		0.952436,
		0.930704,
		0.904213,
		0.873037,
		0.837381,
		0.797559,
		0.753975,
		0.707107,
		0.657487,
		0.605693,
		0.552335,
		0.498046,
		0.443475,
		0.389279,
		0.336118,
		0.28465,
		0.235528,
		0.189395,
		0.146884,
		0.108615,
		0.07519,
		0.047197,
		0.025204,
		0.009762,
		0.001403,
		0.004665,
		0.032698,
		0.081039,
		0.144484,
		0.218463,
		0.298992,
		0.382633,
		0.466459,
		0.548017,
		0.6253,
		0.69672,
		0.761088,
		0.817592,
		0.865776,
		0.905532,
		0.937074,
		0.960924,
		0.977883,
		0.988991,
		0.99547,
		0.998635,
		0.999776,
		0.999994,
		1,
		0.999912,
		0.999265,
		0.997272,
		0.99306,
		0.985814,
		0.974858,
		0.959694,
		0.940003,
		0.915644,
		0.886629,
		0.853109,
		0.815347,
		0.773702,
		0.72861,
		0.680569,
		0.630124,
		0.577857,
		0.524376,
		0.470308,
		0.416289,
		0.362964,
		0.310976,
		0.260965,
		0.213563,
		0.169394,
		0.12907,
		0.093187,
		0.062328,
		0.037057,
		0.017921,
		0.00545,
		0.000155,
		0.00261,
		0.026902,
		0.072021,
		0.132788,
		0.204642,
		0.283601,
		0.366214,
		0.449532,
		0.531073,
		0.608789,
		0.681045,
		0.746597,
		0.804566,
		0.854428,
		0.895994,
		0.929393,
		0.95506,
		0.973702,
		0.986276,
		0.99393,
		0.997931,
		0.999561,
		0.999971,
		1,
		0.999972,
		0.9996,
		0.998194,
		0.994912,
		0.988925,
		0.979518,
		0.966137,
		0.948403,
		0.926112,
		0.899218,
		0.867815,
		0.832117,
		0.792437,
		0.749168,
		0.702773,
		0.653763,
		0.602692,
		0.550142,
		0.496716,
		0.443032,
		0.389717,
		0.337397,
		0.286701,
		0.238249,
		0.192655,
		0.150522,
		0.112441,
		0.078987,
		0.050721,
		0.028187,
		0.011912,
		0.002404,
		0.001154,
		0.02169,
		0.063557,
		0.121599,
		0.191269,
		0.268583,
		0.350084,
		0.432804,
		0.51423,
		0.592281,
		0.665277,
		0.731919,
		0.79127,
		0.842738,
		0.886059,
		0.921283,
		0.948756,
		0.9691,
		0.983187,
		0.992087,
		0.997015,
		0.99923,
		0.999912,
		1,
		0.999994,
		0.999806,
		0.998863,
		0.996374,
		0.99151,
		0.983525,
		0.971816,
		0.955947,
		0.935655,
		0.910833,
		0.881522,
		0.847883,
		0.810183,
		0.768774,
		0.724081,
		0.676582,
		0.6268,
		0.575292,
		0.522638,
		0.469434,
		0.416289,
		0.363815,
		0.312626,
		0.263329,
		0.21653,
		0.172822,
		0.132788,
		0.096998,
		0.066007,
		0.040357,
		0.02057,
		0.007154,
		0.000596,
		0.000287,
		0.017058,
		0.055645,
		0.110919,
		0.178346,
		0.253947,
		0.334256,
		0.416289,
		0.497509,
		0.575799,
		0.649438,
		0.717079,
		0.777728,
		0.830729,
		0.875748,
		0.912756,
		0.94202,
		0.964076,
		0.979712,
		0.989923,
		0.995862,
		0.998756,
		0.999797,
		0.999994,
		1,
		0.99992,
		0.999328,
		0.997501,
		0.993626,
		0.986936,
		0.976785,
		0.962684,
		0.944313,
		0.921508,
		0.894253,
		0.862659,
		0.826947,
		0.787425,
		0.74448,
		0.698557,
		0.650148,
		0.599783,
		0.548017,
		0.495428,
		0.442603,
		0.39014,
		0.338636,
		0.28869,
		0.240892,
		0.19583,
		0.154076,
		0.116196,
		0.082738,
		0.05424,
		0.031222,
		0.014187,
		0.003623,
		0.012998,
		0.048281,
		0.100747,
		0.165879,
		0.2397,
		0.318742,
		0.400003,
		0.480925,
		0.559361,
		0.63355,
		0.7021,
		0.763962,
		0.818422,
		0.865077,
		0.903828,
		0.93486,
		0.958631,
		0.975845,
		0.987421,
		0.994448,
		0.998111,
		0.9996,
		0.999973,
		1,
		0.999974,
		0.999635,
		0.998347,
		0.995329,
		0.989805,
		0.981094,
		0.968661,
		0.952128,
		0.931276,
		0.906036,
		0.876466,
		0.842738,
		0.805122,
		0.763962,
		0.71967,
		0.672708,
		0.623575,
		0.572806,
		0.520954,
		0.468589,
		0.416289,
		0.36464,
		0.314225,
		0.265625,
		0.219416,
		0.176164,
		0.136427,
		0.100747,
		0.069658,
		0.043674,
		0.023299,
		0.009016,
		0.001294,
		0.009504,
		0.041461,
		0.091084,
		0.153869,
		0.225849,
		0.30355,
		0.383959,
		0.464497,
		0.542987,
		0.617635,
		0.687003,
		0.749996,
		0.805839,
		0.854068,
		0.894513,
		0.927288,
		0.952769,
		0.971582,
		0.984569,
		0.992753,
		0.997272,
		0.999297,
		0.99992,
		1,
		0.999995,
		0.999823,
		0.99896,
		0.996672,
		0.992187,
		0.984799,
		0.973927,
		0.959144,
		0.940176,
		0.9169,
		0.889324,
		0.857572,
		0.821869,
		0.782523,
		0.739908,
		0.694455,
		0.646637,
		0.596961,
		0.545959,
		0.49418,
		0.442188,
		0.39055,
		0.339837,
		0.290619,
		0.243461,
		0.198921,
		0.157547,
		0.119879,
		0.086442,
		0.057749,
		0.034297,
		0.01657,
		0.005035,
		0.000143,
		0.006567,
		0.035182,
		0.081927,
		0.14232,
		0.212399,
		0.28869,
		0.368169,
		0.448238,
		0.526696,
		0.601712,
		0.671812,
		0.73585,
		0.793,
		0.842738,
		0.884829,
		0.919314,
		0.946495,
		0.96692,
		0.981357,
		0.990759,
		0.996213,
		0.998863,
		0.999814,
		0.999995,
		1,
		0.999927,
		0.999385,
		0.997707,
		0.994136,
		0.987949,
		0.978531,
		0.965405,
		0.948246,
		0.926878,
		0.901258,
		0.871464,
		0.837676,
		0.800163,
		0.759263,
		0.715374,
		0.66894,
		0.620445,
		0.570395,
		0.519322,
		0.467769,
		0.416289,
		0.365439,
		0.315776,
		0.267854,
		0.222223,
		0.179424,
		0.139988,
		0.104436,
		0.073275,
		0.047001,
		0.026093,
		0.011016,
		0.002221,
		0.004182,
		0.029437,
		0.073275,
		0.131232,
		0.199356,
		0.27417,
		0.352645,
		0.432165,
		0.510503,
		0.585802,
		0.656545,
		0.721547,
		0.779928,
		0.831109,
		0.874792,
		0.91095,
		0.939815,
		0.961859,
		0.977776,
		0.98845,
		0.994912,
		0.998273,
		0.999635,
		0.999976,
		1,
		0.999977,
		0.999666,
		0.998483,
		0.995704,
		0.990599,
		0.982523,
		0.970957,
		0.955526,
		0.936004,
		0.912296,
		0.884434,
		0.852555,
		0.816885,
		0.777728,
		0.735449,
		0.690463,
		0.643226,
		0.594222,
		0.543963,
		0.492972,
		0.441786,
		0.390948,
		0.341002,
		0.292492,
		0.245958,
		0.201932,
		0.160938,
		0.123492,
		0.090096,
		0.061241,
		0.037403,
		0.019046,
		0.006618,
		0.000551,
		0.002341,
		0.024223,
		0.065127,
		0.120607,
		0.186723,
		0.259999,
		0.337397,
		0.416289,
		0.494427,
		0.569922,
		0.641225,
		0.707107,
		0.766643,
		0.8192,
		0.864419,
		0.902212,
		0.932739,
		0.956401,
		0.973821,
		0.985814,
		0.99335,
		0.997501,
		0.999358,
		0.999927,
		1,
		0.999995,
		0.999838,
		0.999046,
		0.99694,
		0.992798,
		0.985952,
		0.975845,
		0.962058,
		0.944313,
		0.922467,
		0.896506,
		0.86652,
		0.832697,
		0.795304,
		0.754672,
		0.711187,
		0.665277,
		0.617404,
		0.568056,
		0.51774,
		0.466975,
		0.416289,
		0.366214,
		0.317281,
		0.27002,
		0.224955,
		0.182603,
		0.143473,
		0.108062,
		0.076857,
		0.05033,
		0.028941,
		0.013138,
		0.003352,
		0.001035,
		0.019535,
		0.057479,
		0.110446,
		0.174504,
		0.246181,
		0.322436,
		0.400625,
		0.478481,
		0.55409,
		0.625869,
		0.692551,
		0.753166,
		0.807029,
		0.853729,
		0.893112,
		0.925275,
		0.950551,
		0.969489,
		0.982839,
		0.99151,
		0.996527,
		0.99896,
		0.99983,
		0.999995,
		1,
		0.999933,
		0.999437,
		0.997893,
		0.994596,
		0.988867,
		0.980117,
		0.967884,
		0.951842,
		0.931803,
		0.907703,
		0.879589,
		0.84761,
		0.811993,
		0.773038,
		0.731099,
		0.686577,
		0.63991,
		0.591564,
		0.542026,
		0.4918,
		0.441396,
		0.391333,
		0.342132,
		0.294311,
		0.248386,
		0.204865,
		0.164251,
		0.127035,
		0.093699,
		0.064712,
		0.040531,
		0.021601,
		0.008351,
		0.001198,
		0.000257,
		0.015366,
		0.05033,
		0.100747,
		0.162702,
		0.232725,
		0.30777,
		0.385184,
		0.462682,
		0.538323,
		0.610497,
		0.677898,
		0.739516,
		0.794618,
		0.842738,
		0.883666,
		0.917436,
		0.944313,
		0.96478,
		0.979518,
		0.989376,
		0.995329,
		0.998417,
		0.999666,
		0.999978,
		1,
		0.999979,
		0.999694,
		0.998606,
		0.996042,
		0.991318,
		0.983819,
		0.973047,
		0.958631,
		0.940336,
		0.918052,
		0.891784,
		0.861636,
		0.827802,
		0.790544,
		0.750188,
		0.707107,
		0.661713,
		0.61445,
		0.565786,
		0.516206,
		0.466205,
		0.416289,
		0.366965,
		0.318742,
		0.272124,
		0.227614,
		0.185705,
		0.146884,
		0.111627,
		0.080399,
		0.053654,
		0.031834,
		0.015366,
		0.004665,
		0.000132,
		0.011711,
		0.043674,
		0.091512,
		0.151319,
		0.219635,
		0.293408,
		0.369978,
		0.447041,
		0.522638,
		0.595126,
		0.663169,
		0.725714,
		0.781985,
		0.831466,
		0.87389,
		0.909232,
		0.937694,
		0.959694,
		0.975845,
		0.986936,
		0.993888,
		0.997707,
		0.999412,
		0.999933,
		1,
		0.999996,
		0.999851,
		0.999123,
		0.997182,
		0.99335,
		0.986998,
		0.977591,
		0.96472,
		0.948101,
		0.927583,
		0.903125,
		0.874792,
		0.842738,
		0.807193,
		0.76845,
		0.726854,
		0.682792,
		0.636686,
		0.588982,
		0.540147,
		0.490663,
		0.441017,
		0.391707,
		0.343228,
		0.296078,
		0.250747,
		0.207724,
		0.167488,
		0.130509,
		0.09725,
		0.068159,
		0.043674,
		0.024223,
		0.010218,
		0.002059,
		0.008564,
		0.037511,
		0.082738,
		0.140358,
		0.206915,
		0.279358,
		0.355015,
		0.431573,
		0.507048,
		0.579774,
		0.64838,
		0.711778,
		0.769149,
		0.819929,
		0.863799,
		0.900677,
		0.930704,
		0.954233,
		0.971816,
		0.984178,
		0.992187,
		0.99681,
		0.999046,
		0.999845,
		0.999996,
		1,
		0.999939,
		0.999482,
		0.99806,
		0.995012,
		0.9897,
		0.981561,
		0.970147,
		0.955135,
		0.936326,
		0.913638,
		0.887096,
		0.856815,
		0.82299,
		0.785882,
		0.745808,
		0.70313,
		0.658245,
		0.61158,
		0.563582,
		0.514717,
		0.465459,
		0.416289,
		0.367695,
		0.32016,
		0.27417,
		0.230203,
		0.188732,
		0.150223,
		0.11513,
		0.083901,
		0.05697,
		0.034761,
		0.017685,
		0.00614,
		0.000511,
		0.00592,
		0.031834,
		0.074425,
		0.12982,
		0.19457,
		0.265625,
		0.340308,
		0.416289,
		0.491569,
		0.564456,
		0.63355,
		0.697728,
		0.756129,
		0.808146,
		0.85341,
		0.891784,
		0.92335,
		0.948403,
		0.967429,
		0.981094,
		0.990212,
		0.995704,
		0.998546,
		0.999694,
		0.99998,
		1,
		0.999981,
		0.999719,
		0.998717,
		0.996348,
		0.991969,
		0.984999,
		0.974954,
		0.961473,
		0.944313,
		0.92335,
		0.898568,
		0.870045,
		0.837941,
		0.802484,
		0.763962,
		0.722712,
		0.679106,
		0.63355,
		0.586474,
		0.538323,
		0.489559,
		0.44065,
		0.39207,
		0.344293,
		0.297795,
		0.253045,
		0.210511,
		0.170651,
		0.133916,
		0.100747,
		0.071577,
		0.046826,
		0.026902,
		0.012202,
		0.003111,
		0.003771,
		0.02664,
		0.066571,
		0.119705,
		0.182603,
		0.252217,
		0.325863,
		0.401201,
		0.476215,
		0.549189,
		0.618696,
		0.683581,
		0.742944,
		0.796135,
		0.842738,
		0.882566,
		0.915644,
		0.942207,
		0.962684,
		0.977679,
		0.987949,
		0.994372,
		0.997893,
		0.99946,
		0.999939,
		1,
		0.999996,
		0.999863,
		0.999193,
		0.9974,
		0.993851,
		0.987949,
		0.979184,
		0.967154,
		0.951577,
		0.932289,
		0.909232,
		0.882447,
		0.852057,
		0.81826,
		0.781314,
		0.741528,
		0.699252,
		0.654869,
		0.608789,
		0.561441,
		0.513271,
		0.464734,
		0.416289,
		0.368403,
		0.321539,
		0.27616,
		0.232725,
		0.191686,
		0.15349,
		0.118572,
		0.08736,
		0.060273,
		0.037715,
		0.020082,
		0.007758,
		0.001112,
		0.002111,
		0.021925,
		0.059174,
		0.110014,
		0.171018,
		0.239139,
		0.311689,
		0.38632,
		0.460998,
		0.533987,
		0.603835,
		0.669354,
		0.72961,
		0.783913,
		0.831801,
		0.873037,
		0.907596,
		0.935655,
		0.957583,
		0.973927,
		0.985389,
		0.992798,
		0.997064,
		0.999123,
		0.999858,
		0.999996,
		1,
		0.999944,
		0.999524,
		0.99821,
		0.995389,
		0.990456,
		0.982876,
		0.972215,
		0.958153,
		0.940484,
		0.919112,
		0.894038,
		0.865352,
		0.833218,
		0.797864,
		0.759572,
		0.718669,
		0.675515,
		0.630499,
		0.584036,
		0.536552,
		0.488488,
		0.440294,
		0.392422,
		0.345327,
		0.299464,
		0.255282,
		0.213228,
		0.173742,
		0.137255,
		0.104192,
		0.074965,
		0.049979,
		0.029626,
		0.014288,
		0.004334,
		0.000123,
		0.000934,
		0.017685,
		0.05223,
		0.100747,
		0.159817,
		0.226396,
		0.297795,
		0.371655,
		0.44593,
		0.518865,
		0.588982,
		0.655065,
		0.716147,
		0.771497,
		0.820615,
		0.863213,
		0.899218,
		0.928752,
		0.952128,
		0.969836,
		0.982523,
		0.990967,
		0.996042,
		0.998663,
		0.999719,
		0.999981,
		1,
		0.999982,
		0.999741,
		0.998817,
		0.996625,
		0.992561,
		0.986073,
		0.976697,
		0.964076,
		0.947967,
		0.928233,
		0.904839,
		0.877839,
		0.847365,
		0.813614,
		0.776841,
		0.737347,
		0.69547,
		0.651582,
		0.606074,
		0.559361,
		0.511868,
		0.46403,
		0.416289,
		0.369091,
		0.322879,
		0.278096,
		0.235182,
		0.19457,
		0.156687,
		0.121952,
		0.090775,
		0.063557,
		0.040688,
		0.022547,
		0.009504,
		0.001913,
		0.000232,
		0.013913,
		0.045739,
		0.091904,
		0.149001,
		0.213992,
		0.284186,
		0.357216,
		0.431024,
		0.503836,
		0.574153,
		0.64073,
		0.70257,
		0.758906,
		0.809195,
		0.853109,
		0.890522,
		0.921508,
		0.946324,
		0.965405,
		0.979343,
		0.988867,
		0.994809,
		0.99806,
		0.999504,
		0.999944,
		1,
		0.999996,
		0.999874,
		0.999256,
		0.997598,
		0.994306,
		0.988816,
		0.980638,
		0.969383,
		0.954769,
		0.936624,
		0.914874,
		0.889537,
		0.860713,
		0.828569,
		0.793332,
		0.755278,
		0.714722,
		0.672015,
		0.62753,
		0.581665,
		0.53483,
		0.487448,
		0.439948,
		0.392764,
		0.346332,
		0.301087,
		0.25746,
		0.215878,
		0.176763,
		0.140529,
		0.107582,
		0.07832,
		0.05313,
		0.032389,
		0.016464,
		0.005711,
		0.000475,
		0.010606,
		0.039695,
		0.083483,
		0.138573,
		0.201932,
		0.270869,
		0.343012,
		0.416289,
		0.488913,
		0.559361,
		0.626365,
		0.688896,
		0.746155,
		0.797559,
		0.842738,
		0.881522,
		0.913932,
		0.940176,
		0.960634,
		0.975845,
		0.986488,
		0.99335,
		0.997294,
		0.999193,
		0.999869,
		0.999996,
		1,
		0.999948,
		0.999561,
		0.998347,
		0.995731,
		0.991144,
		0.984077,
		0.974109,
		0.960924,
		0.944313,
		0.924165,
		0.900464,
		0.873276,
		0.842738,
		0.809049,
		0.772459,
		0.73326,
		0.691782,
		0.64838,
		0.603434,
		0.557338,
		0.510503,
		0.463346,
		0.416289,
		0.369759,
		0.324181,
		0.27998,
		0.237576,
		0.197385,
		0.159817,
		0.125272,
		0.094146,
		0.066821,
		0.043674,
		0.02507,
		0.011362,
		0.002895,
		0.007758,
		0.034098,
		0.075484,
		0.128533,
		0.190218,
		0.25785,
		0.32905,
		0.401737,
		0.474108,
		0.544621,
		0.611985,
		0.675142,
		0.73326,
		0.785723,
		0.832117,
		0.87223,
		0.906036,
		0.933692,
		0.955526,
		0.972026,
		0.983819,
		0.991651,
		0.996348,
		0.998768,
		0.999741,
		0.999983,
		1,
		0.999984,
		0.999762,
		0.998907,
		0.996876,
		0.993099,
		0.987053,
		0.978292,
		0.966466,
		0.95133,
		0.932739,
		0.910641,
		0.88507,
		0.856131,
		0.823996,
		0.788888,
		0.751077,
		0.710869,
		0.668603,
		0.62464,
		0.579359,
		0.533157,
		0.486437,
		0.439612,
		0.393097,
		0.34731,
		0.302666,
		0.259581,
		0.218463,
		0.179716,
		0.143738,
		0.110919,
		0.08164,
		0.056273,
		0.035182,
		0.018718,
		0.007225,
		0.001035,
		0.005363,
		0.028941,
		0.067904,
		0.118882,
		0.178855,
		0.245133,
		0.315337,
		0.387376,
		0.459432,
		0.529947,
		0.597605,
		0.661323,
		0.720239,
		0.773702,
		0.82126,
		0.862659,
		0.897829,
		0.926878,
		0.950084,
		0.967884,
		0.980856,
		0.9897,
		0.995205,
		0.99821,
		0.999543,
		0.999948,
		1,
		0.999997,
		0.999884,
		0.999313,
		0.997777,
		0.994719,
		0.989606,
		0.981968,
		0.971428,
		0.957705,
		0.940621,
		0.92009,
		0.896111,
		0.86876,
		0.838179,
		0.804566,
		0.768167,
		0.729267,
		0.688183,
		0.645261,
		0.600863,
		0.555372,
		0.509177,
		0.462682,
		0.416289,
		0.370409,
		0.325448,
		0.281814,
		0.23991,
		0.200135,
		0.16288,
		0.128533,
		0.09747,
		0.070061,
		0.046668,
		0.027641,
		0.01332,
		0.004038,
		0.000114,
		0.003417,
		0.024223,
		0.060743,
		0.10962,
		0.167843,
		0.232725,
		0.301882,
		0.373216,
		0.444897,
		0.51535,
		0.583238,
		0.647454,
		0.707107,
		0.761512,
		0.810183,
		0.852824,
		0.889324,
		0.919743,
		0.944313,
		0.963419,
		0.977591,
		0.987485,
		0.993851,
		0.997501,
		0.999256,
		0.99988,
		0.999997,
		1,
		0.999952,
		0.999595,
		0.99847,
		0.996042,
		0.991772,
		0.985175,
		0.975845,
		0.963472,
		0.947842,
		0.928835,
		0.906419,
		0.880639,
		0.851607,
		0.819497,
		0.784529,
		0.746966,
		0.707107,
		0.665277,
		0.621825,
		0.577116,
		0.531529,
		0.485454,
		0.439285,
		0.39342,
		0.34826,
		0.304203,
		0.261647,
		0.220985,
		0.182603,
		0.146884,
		0.114202,
		0.084923,
		0.059405,
		0.037997,
		0.02104,
		0.008861,
		0.001783,
		0.001913,
		0.019939,
		0.053998,
		0.100747,
		0.157186,
		0.220628,
		0.28869,
		0.359265,
		0.430513,
		0.500843,
		0.568899,
		0.63355,
		0.693878,
		0.749168,
		0.798899,
		0.842738,
		0.880531,
		0.912296,
		0.938217,
		0.958631,
		0.974023,
		0.984999,
		0.992272,
		0.996625,
		0.998863,
		0.999762,
		0.999984,
		1,
		0.999985,
		0.99978,
		0.998989,
		0.997104,
		0.99359,
		0.987949,
		0.979753,
		0.968661,
		0.954428,
		0.9369,
		0.916014,
		0.891784,
		0.864293,
		0.833687,
		0.800163,
		0.763962,
		0.725363,
		0.684672,
		0.642221,
		0.598361,
		0.553458,
		0.507888,
		0.462036,
		0.416289,
		0.371041,
		0.326681,
		0.283601,
		0.242186,
		0.20282,
		0.165879,
		0.131734,
		0.100747,
		0.073275,
		0.049664,
		0.030251,
		0.015366,
		0.005327,
		0.000443,
		0.000846,
		0.016085,
		0.047667,
		0.092264,
		0.146884,
		0.208847,
		0.275766,
		0.345531,
		0.416289,
		0.486437,
		0.5546,
		0.619625,
		0.680569,
		0.736686,
		0.787425,
		0.832415,
		0.871464,
		0.904547,
		0.931803,
		0.953525,
		0.970147,
		0.982234,
		0.990456,
		0.995564,
		0.998347,
		0.999578,
		0.999952,
		1,
		0.999997,
		0.999893,
		0.999365,
		0.99794,
		0.995096,
		0.990328,
		0.983187,
		0.973306,
		0.960409,
		0.944313,
		0.924919,
		0.902212,
		0.876247,
		0.847143,
		0.815072,
		0.780253,
		0.742944,
		0.703432,
		0.662033,
		0.619082,
		0.574932,
		0.529947,
		0.484498,
		0.438967,
		0.393734,
		0.349184,
		0.3057,
		0.263662,
		0.223446,
		0.185427,
		0.149968,
		0.117431,
		0.088168,
		0.062522,
		0.04083,
		0.023419,
		0.010606,
		0.0027,
		0.000211,
		0.012657,
		0.041747,
		0.084169,
		0.13694,
		0.197385,
		0.263117,
		0.332021,
		0.402235,
		0.472143,
		0.540354,
		0.605693,
		0.667192,
		0.724081,
		0.775775,
		0.821869,
		0.862134,
		0.896506,
		0.925079,
		0.948101,
		0.965963,
		0.979184,
		0.988393,
		0.994306,
		0.997689,
		0.999313,
		0.999889,
		0.999997,
		1,
		0.999956,
		0.999625,
		0.998583,
		0.996325,
		0.992346,
		0.986181,
		0.977439,
		0.965817,
		0.951098,
		0.933156,
		0.911942,
		0.887485,
		0.859876,
		0.829262,
		0.795839,
		0.759844,
		0.721547,
		0.681244,
		0.639258,
		0.595924,
		0.551596,
		0.506634,
		0.461407,
		0.416289,
		0.371655,
		0.327881,
		0.285341,
		0.244406,
		0.205443,
		0.168815,
		0.134877,
		0.103978,
		0.076461,
		0.052658,
		0.032895,
		0.017488,
		0.006746,
		0.000966,
		0.00965,
		0.036235,
		0.076461,
		0.127354,
		0.186245,
		0.250747,
		0.318742,
		0.38836,
		0.457972,
		0.526172,
		0.591766,
		0.653763,
		0.711367,
		0.763962,
		0.811114,
		0.852555,
		0.888183,
		0.918052,
		0.942367,
		0.961473,
		0.975845,
		0.986073,
		0.992836,
		0.996876,
		0.998949,
		0.99978,
		0.999985,
		1,
		0.999986,
		0.999797,
		0.999064,
		0.997312,
		0.994038,
		0.98877,
		0.981094,
		0.970681,
		0.957286,
		0.940749,
		0.920996,
		0.898023,
		0.871897,
		0.842738,
		0.810721,
		0.776061,
		0.739008,
		0.699842,
		0.658868,
		0.61641,
		0.572806,
		0.528406,
		0.483569,
		0.438658,
		0.39404,
		0.350084,
		0.307157,
		0.265625,
		0.225849,
		0.188187,
		0.152992,
		0.120607,
		0.091374,
		0.065621,
		0.043674,
		0.025848,
		0.012447,
		0.003771,
		0.000107,
		0.00706,
		0.031129,
		0.069138,
		0.118126,
		0.175429,
		0.238661,
		0.3057,
		0.374671,
		0.443933,
		0.512066,
		0.577857,
		0.640295,
		0.698557,
		0.752002,
		0.800163,
		0.842738,
		0.879589,
		0.910731,
		0.936326,
		0.956676,
		0.972215,
		0.98349,
		0.991144,
		0.99589,
		0.99847,
		0.99961,
		0.999956,
		1,
		0.999997,
		0.999901,
		0.999412,
		0.998088,
		0.99544,
		0.990988,
		0.984305,
		0.975034,
		0.962903,
		0.947725,
		0.929393,
		0.907879,
		0.883219,
		0.855511,
		0.824905,
		0.791594,
		0.75581,
		0.717815,
		0.677898,
		0.636368,
		0.59355,
		0.549783,
		0.505413,
		0.460796,
		0.416289,
		0.372254,
		0.32905,
		0.287037,
		0.246571,
		0.208006,
		0.171689,
		0.137962,
		0.107161,
		0.079615,
		0.055645,
		0.035564,
		0.019678,
		0.008282,
		0.001665,
		0.004882,
		0.026425,
		0.0622,
		0.109257,
		0.164939,
		0.226862,
		0.292901,
		0.361176,
		0.430036,
		0.498046,
		0.563978,
		0.6268,
		0.685666,
		0.739908,
		0.78903,
		0.832697,
		0.870737,
		0.903125,
		0.929985,
		0.951577,
		0.968292,
		0.980638,
		0.98922,
		0.994719,
		0.99786,
		0.999365,
		0.999897,
		0.999997,
		1,
		0.999959,
		0.999653,
		0.998685,
		0.996584,
		0.99287,
		0.987103,
		0.978905,
		0.967979,
		0.954108,
		0.937158,
		0.917071,
		0.893857,
		0.86759,
		0.838394,
		0.806444,
		0.771949,
		0.735156,
		0.696335,
		0.655781,
		0.613805,
		0.570735,
		0.526907,
		0.482664,
		0.438357,
		0.394338,
		0.35096,
		0.308577,
		0.26754,
		0.228195,
		0.190888,
		0.155955,
		0.12373,
		0.094539,
		0.0687,
		0.046525,
		0.028318,
		0.014374,
		0.004979,
		0.000414,
		0.003111,
		0.022121,
		0.055645,
		0.100747,
		0.154777,
		0.215353,
		0.280351,
		0.347883,
		0.416289,
		0.484124,
		0.550142,
		0.613292,
		0.672708,
		0.727694,
		0.777728,
		0.822445,
		0.861636,
		0.895243,
		0.92335,
		0.946178,
		0.964076,
		0.977512,
		0.987053,
		0.99335,
		0.997104,
		0.999027,
		0.999797,
		0.999987,
		1,
		0.999987,
		0.999812,
		0.999131,
		0.997501,
		0.994448,
		0.989522,
		0.982327,
		0.972543,
		0.959925,
		0.944313,
		0.925619,
		0.903828,
		0.878987,
		0.851199,
		0.820615,
		0.787425,
		0.751857,
		0.714166,
		0.674631,
		0.63355,
		0.591237,
		0.548017,
		0.504225,
		0.460201,
		0.416289,
		0.372836,
		0.330188,
		0.28869,
		0.248685,
		0.210511,
		0.174504,
		0.140991,
		0.110298,
		0.082738,
		0.058623,
		0.038254,
		0.021925,
		0.009923,
		0.002525,
		0.001742,
		0.018211,
		0.049471,
		0.092596,
		0.144944,
		0.204139,
		0.268054,
		0.334797,
		0.402701,
		0.470308,
		0.536358,
		0.599783,
		0.659694,
		0.715374,
		0.766271,
		0.811993,
		0.8523,
		0.887096,
		0.91643,
		0.940484,
		0.959568,
		0.974109,
		0.984638,
		0.991772,
		0.996187,
		0.998583,
		0.999639,
		0.999959,
		1,
		0.999997,
		0.999909,
		0.999454,
		0.998223,
		0.995754,
		0.991594,
		0.985332,
		0.976625,
		0.965204,
		0.950882,
		0.933544,
		0.913148,
		0.889716,
		0.863328,
		0.834112,
		0.802239,
		0.767917,
		0.731385,
		0.692907,
		0.652767,
		0.611266,
		0.568718,
		0.525446,
		0.481783,
		0.438064,
		0.394628,
		0.351814,
		0.309961,
		0.269407,
		0.230487,
		0.193529,
		0.158861,
		0.126801,
		0.097664,
		0.071756,
		0.049379,
		0.030823,
		0.016376,
		0.006313,
		0.000903,
		0.000771,
		0.014693,
		0.043674,
		0.084802,
		0.135442,
		0.193222,
		0.256014,
		0.321925,
		0.389279,
		0.456608,
		0.522638,
		0.586284,
		0.646637,
		0.702959,
		0.754672,
		0.801356,
		0.842738,
		0.878693,
		0.909232,
		0.934501,
		0.954769,
		0.970427,
		0.981968,
		0.989975,
		0.995096,
		0.998016,
		0.999412,
		0.999905,
		0.999997,
		1,
		0.999962,
		0.999678,
		0.998778,
		0.99682,
		0.99335,
		0.987949,
		0.980254,
		0.969973,
		0.956891,
		0.940869,
		0.921837,
		0.899793,
		0.874792,
		0.846941,
		0.816392,
		0.783333,
		0.747985,
		0.710598,
		0.67144,
		0.630801,
		0.588982,
		0.546297,
		0.503068,
		0.459622,
		0.416289,
		0.373403,
		0.331297,
		0.290302,
		0.250747,
		0.212959,
		0.17726,
		0.143965,
		0.113386,
		0.085828,
		0.061589,
		0.040959,
		0.024223,
		0.011657,
		0.003529,
		0.0001,
		0.000192,
		0.011564,
		0.038254,
		0.077365,
		0.126271,
		0.182603,
		0.244237,
		0.309274,
		0.376031,
		0.443032,
		0.508991,
		0.572806,
		0.63355,
		0.690463,
		0.742944,
		0.790544,
		0.832964,
		0.870045,
		0.901765,
		0.928233,
		0.949683,
		0.966466,
		0.979038,
		0.987949,
		0.993819,
		0.997312,
		0.999098,
		0.999812,
		0.999988,
		1,
		0.999988,
		0.999825,
		0.999193,
		0.997675,
		0.994823,
		0.990212,
		0.983462,
		0.97426,
		0.962366,
		0.947615,
		0.929913,
		0.909232,
		0.885604,
		0.859113,
		0.82989,
		0.798105,
		0.763962,
		0.727694,
		0.689557,
		0.649825,
		0.608789,
		0.566751,
		0.524024,
		0.480925,
		0.437779,
		0.39491,
		0.352645,
		0.31131,
		0.27123,
		0.232725,
		0.196114,
		0.16171,
		0.12982,
		0.100747,
		0.074788,
		0.05223,
		0.033357,
		0.018444,
		0.007758,
		0.001559,
		0.008818,
		0.033208,
		0.070284,
		0.117431,
		0.172287,
		0.232725,
		0.296847,
		0.362964,
		0.42959,
		0.495428,
		0.559361,
		0.620445,
		0.677898,
		0.731099,
		0.779572,
		0.82299,
		0.861163,
		0.894038,
		0.921688,
		0.944313,
		0.962225,
		0.975845,
		0.985689,
		0.992346,
		0.996457,
		0.998685,
		0.999666,
		0.999962,
		1,
		0.999998,
		0.999915,
		0.999493,
		0.998347,
		0.996042,
		0.992149,
		0.986276,
		0.978092,
		0.967332,
		0.953807,
		0.937398,
		0.918052,
		0.895777,
		0.870636,
		0.842738,
		0.812236,
		0.779315,
		0.744191,
		0.707107,
		0.668323,
		0.628118,
		0.586783,
		0.544621,
		0.501941,
		0.459058,
		0.416289,
		0.373956,
		0.332378,
		0.291874,
		0.252761,
		0.215353,
		0.179959,
		0.146884,
		0.116428,
		0.088884,
		0.064539,
		0.043674,
		0.026564,
		0.013475,
		0.004665,
		0.000387,
		0.006452,
		0.028532,
		0.063557,
		0.108924,
		0.162273,
		0.221482,
		0.28465,
		0.350084,
		0.416289,
		0.481957,
		0.545959,
		0.607332,
		0.665277,
		0.719149,
		0.76845,
		0.812825,
		0.852057,
		0.886059,
		0.914874,
		0.938663,
		0.957705,
		0.972387,
		0.983187,
		0.990665,
		0.99544,
		0.998157,
		0.999454,
		0.999912,
		0.999998,
		1,
		0.999965,
		0.999701,
		0.998863,
		0.997036,
		0.993791,
		0.988728,
		0.981498,
		0.971816,
		0.959469,
		0.944313,
		0.92627,
		0.905327,
		0.881522,
		0.854945,
		0.825729,
		0.794042,
		0.760084,
		0.724081,
		0.686282,
		0.646952,
		0.606372,
		0.564834,
		0.522638,
		0.48009,
		0.437501,
		0.395185,
		0.353455,
		0.312626,
		0.273008,
		0.234912,
		0.198643,
		0.164503,
		0.132788,
		0.103789,
		0.077793,
		0.055077,
		0.035914,
		0.02057,
		0.009304,
		0.002366,
		0.004462,
		0.024223,
		0.057184,
		0.100747,
		0.152563,
		0.210511,
		0.272688,
		0.337397,
		0.403137,
		0.468589,
		0.532609,
		0.594222,
		0.65261,
		0.707107,
		0.757191,
		0.802484,
		0.842738,
		0.877839,
		0.907796,
		0.932739,
		0.95291,
		0.968661,
		0.980438,
		0.98877,
		0.994248,
		0.997501,
		0.999163,
		0.999825,
		0.999988,
		1,
		0.999989,
		0.999838,
		0.999249,
		0.997833,
		0.995167,
		0.990846,
		0.984507,
		0.975845,
		0.964625,
		0.950679,
		0.933906,
		0.914268,
		0.891784,
		0.86652,
		0.838591,
		0.808146,
		0.77537,
		0.740474,
		0.703692,
		0.665277,
		0.625499,
		0.584639,
		0.542987,
		0.500843,
		0.458508,
		0.416289,
		0.374495,
		0.333432,
		0.293408,
		0.254728,
		0.217694,
		0.182603,
		0.14975,
		0.119423,
		0.091904,
		0.067471,
		0.046396,
		0.028941,
		0.015366,
		0.00592,
		0.000846,
		0.002844,
		0.02028,
		0.051161,
		0.092903,
		0.14316,
		0.199814,
		0.260965,
		0.324909,
		0.39014,
		0.455329,
		0.519322,
		0.581127,
		0.63991,
		0.694984,
		0.745808,
		0.791976,
		0.833218,
		0.869387,
		0.900464,
		0.926546,
		0.947842,
		0.964669,
		0.977439,
		0.986651,
		0.99287,
		0.996704,
		0.998778,
		0.99969,
		0.999965,
		1,
		0.999998,
		0.999921,
		0.999529,
		0.99846,
		0.996306,
		0.99266,
		0.987147,
		0.979446,
		0.969301,
		0.95652,
		0.94098,
		0.92262,
		0.901435,
		0.877473,
		0.850827,
		0.82163,
		0.790049,
		0.75628,
		0.720543,
		0.68308,
		0.644146,
		0.604014,
		0.562964,
		0.521286,
		0.479275,
		0.43723,
		0.395453,
		0.354245,
		0.313909,
		0.274744,
		0.237049,
		0.201118,
		0.167242,
		0.135705,
		0.106789,
		0.08077,
		0.057915,
		0.038489,
		0.022747,
		0.01094,
		0.00331,
		9.3e-005,
		0.001593,
		0.016698,
		0.045488,
		0.085389,
		0.134062,
		0.189395,
		0.249484,
		0.312626,
		0.377306,
		0.442188,
		0.506107,
		0.568056,
		0.627186,
		0.682792,
		0.734311,
		0.781314,
		0.823506,
		0.860713,
		0.892886,
		0.92009,
		0.942504,
		0.960409,
		0.974188,
		0.984305,
		0.991297,
		0.995754,
		0.998286,
		0.999493,
		0.999918,
		0.999998,
		1,
		0.999968,
		0.999722,
		0.998941,
		0.997234,
		0.994196,
		0.989445,
		0.982646,
		0.973521,
		0.961859,
		0.947512,
		0.930398,
		0.91049,
		0.887815,
		0.862446,
		0.834498,
		0.804122,
		0.771497,
		0.736831,
		0.70035,
		0.6623,
		0.622942,
		0.582546,
		0.541394,
		0.499772,
		0.457972,
		0.416289,
		0.37502,
		0.33446,
		0.294906,
		0.25665,
		0.219983,
		0.185194,
		0.152563,
		0.122371,
		0.094888,
		0.070384,
		0.049119,
		0.031349,
		0.017323,
		0.007282,
		0.001462,
		0.000705,
		0.013475,
		0.040162,
		0.078205,
		0.125272,
		0.179254,
		0.238249,
		0.300551,
		0.36464,
		0.429172,
		0.492972,
		0.55502,
		0.61445,
		0.670542,
		0.722712,
		0.770509,
		0.813614,
		0.851827,
		0.88507,
		0.913379,
		0.9369,
		0.955885,
		0.970681,
		0.981724,
		0.989522,
		0.99464,
		0.997675,
		0.999222,
		0.999838,
		0.999989,
		1,
		0.99999,
		0.999849,
		0.999301,
		0.997978,
		0.995483,
		0.991431,
		0.985471,
		0.977312,
		0.966719,
		0.953525,
		0.937623,
		0.918966,
		0.897559,
		0.873458,
		0.846758,
		0.817592,
		0.786125,
		0.752549,
		0.717079,
		0.679948,
		0.641405,
		0.601712,
		0.561141,
		0.519969,
		0.478481,
		0.436966,
		0.395714,
		0.355015,
		0.315161,
		0.27644,
		0.239139,
		0.203541,
		0.169927,
		0.138573,
		0.109748,
		0.083718,
		0.060743,
		0.041077,
		0.024968,
		0.012657,
		0.004379,
		0.000363,
		0.000176,
		0.010606,
		0.035182,
		0.07135,
		0.11679,
		0.169394,
		0.227264,
		0.28869,
		0.352149,
		0.416289,
		0.479925,
		0.542026,
		0.601712,
		0.658245,
		0.711022,
		0.759572,
		0.803552,
		0.842738,
		0.877025,
		0.906419,
		0.931037,
		0.951098,
		0.96692,
		0.978905,
		0.987535,
		0.99335,
		0.99693,
		0.998863,
		0.999712,
		0.999968,
		1,
		0.999998,
		0.999927,
		0.999561,
		0.998563,
		0.996548,
		0.99313,
		0.987949,
		0.980698,
		0.971124,
		0.959038,
		0.944313,
		0.926878,
		0.90672,
		0.883873,
		0.858415,
		0.830462,
		0.800163,
		0.767695,
		0.73326,
		0.697079,
		0.65939,
		0.620445,
		0.580504,
		0.53984,
		0.498728,
		0.45745,
		0.416289,
		0.375532,
		0.335463,
		0.296367,
		0.258527,
		0.222223,
		0.187731,
		0.155325,
		0.125272,
		0.097836,
		0.073275,
		0.051842,
		0.033782,
		0.019337,
		0.008741,
		0.002221,
		0.008089,
		0.030544,
		0.064824,
		0.108615,
		0.159817,
		0.21653,
		0.277047,
		0.339837,
		0.403546,
		0.466975,
		0.529086,
		0.588982,
		0.645911,
		0.699252,
		0.748515,
		0.793332,
		0.833458,
		0.86876,
		0.899218,
		0.924919,
		0.946052,
		0.962903,
		0.975845,
		0.985332,
		0.991877,
		0.996042,
		0.998404,
		0.999529,
		0.999924,
		0.999998,
		1,
		0.99997,
		0.999741,
		0.999012,
		0.997416,
		0.994569,
		0.990107,
		0.983707,
		0.9751,
		0.964076,
		0.950488,
		0.934245,
		0.915312,
		0.893703,
		0.869479,
		0.842738,
		0.813614,
		0.782268,
		0.748889,
		0.713686,
		0.676885,
		0.638727,
		0.599465,
		0.559361,
		0.518684,
		0.477707,
		0.436709,
		0.395969,
		0.355767,
		0.316383,
		0.278096,
		0.241182,
		0.205913,
		0.172561,
		0.141391,
		0.112664,
		0.086636,
		0.063557,
		0.043674,
		0.027226,
		0.014447,
		0.00592,
		0.026246,
		0.058623,
		0.100747,
		0.150522,
		0.206051,
		0.265625,
		0.327712,
		0.390948,
		0.45413,
		0.516206,
		0.576269,
		0.63355,
		0.687413,
		0.737347,
		0.782964,
		0.823996,
		0.860284,
		0.891784,
		0.918552,
		0.940749,
		0.958631,
		0.972543,
		0.982906,
		0.990212,
		0.994999,
		0.997833,
		0.999276,
		0.999849,
		0.99999,
		1,
		0.99999,
		0.99986,
		0.999348,
		0.998111,
		0.995774,
		0.991969,
		0.986362,
		0.978669,
		0.968661,
		0.95617,
		0.941085,
		0.92335,
		0.902962,
		0.87996,
		0.854428,
		0.826481,
		0.796268,
		0.763962,
		0.729761,
		0.693878,
		0.656545,
		0.618005,
		0.578511,
		0.538323,
		0.497709,
		0.45694,
		0.416289,
		0.376031,
		0.336442,
		0.297795,
		0.260362,
		0.224414,
		0.190218,
		0.158037,
		0.128128,
		0.100747,
		0.076143,
		0.05456,
		0.036235,
		0.004095,
		0.022285,
		0.052748,
		0.093187,
		0.141513,
		0.19583,
		0.254429,
		0.315776,
		0.378502,
		0.441396,
		0.503396,
		0.563582,
		0.621173,
		0.675515,
		0.726079,
		0.772459,
		0.814362,
		0.851607,
		0.884124,
		0.911942,
		0.935195,
		0.954108,
		0.968996,
		0.980254,
		0.988347,
		0.993791,
		0.997137,
		0.998941,
		0.999732,
		0.99997,
		1,
		0.999998,
		0.999932,
		0.999591,
		0.998658,
		0.996771,
		0.993563,
		0.988691,
		0.981857,
		0.972816,
		0.961379,
		0.947416,
		0.930852,
		0.911662,
		0.889869,
		0.865538,
		0.83877,
		0.809696,
		0.778478,
		0.745299,
		0.710363,
		0.673888,
		0.636109,
		0.59727,
		0.557624,
		0.51743,
		0.476952,
		0.436458,
		0.396218,
		0.3565,
		0.317576,
		0.279714,
		0.243179,
		0.208236,
		0.175145,
		0.144161,
		0.115538,
		0.089522,
		0.066356,
		0.00261,
		0.01866,
		0.047197,
		0.085934,
		0.132788,
		0.185867,
		0.243461,
		0.304035,
		0.366214,
		0.42878,
		0.490663,
		0.550931,
		0.608789,
		0.663568,
		0.714722,
		0.761826,
		0.804566,
		0.842738,
		0.876247,
		0.905097,
		0.929393,
		0.949334,
		0.965204,
		0.977373,
		0.986276,
		0.99241,
		0.996306,
		0.998512,
		0.999561,
		0.999929,
		0.999998,
		1,
		0.999972,
		0.999759,
		0.999078,
		0.997583,
		0.994912,
		0.990717,
		0.984688,
		0.976563,
		0.966137,
		0.953258,
		0.937833,
		0.919818,
		0.899218,
		0.876079,
		0.850486,
		0.822556,
		0.792437,
		0.760297,
		0.726331,
		0.690745,
		0.653763,
		0.615622,
		0.576564,
		0.536843,
		0.496716,
		0.456443,
		0.416289,
		0.376519,
		0.337397,
		0.299189,
		0.262156,
		0.226559,
		0.192655,
		0.160699,
		0.130939,
		0.103621,
		0.001462,
		0.015366,
		0.041967,
		0.078987,
		0.124349,
		0.176164,
		0.232725,
		0.292492,
		0.354089,
		0.416289,
		0.478015,
		0.538323,
		0.596407,
		0.651582,
		0.703287,
		0.751077,
		0.794618,
		0.833687,
		0.868161,
		0.898023,
		0.92335,
		0.944313,
		0.961169,
		0.97426,
		0.983995,
		0.990846,
		0.995329,
		0.997978,
		0.999325,
		0.99986,
		0.999991,
		1,
		0.999991,
		0.999869,
		0.999391,
		0.998234,
		0.996042,
		0.992466,
		0.987186,
		0.979927,
		0.970465,
		0.958631,
		0.944313,
		0.927446,
		0.908019,
		0.886059,
		0.861636,
		0.834852,
		0.805839,
		0.774754,
		0.741777,
		0.707107,
		0.670956,
		0.63355,
		0.595126,
		0.555928,
		0.516206,
		0.476215,
		0.436213,
		0.39646,
		0.357216,
		0.318742,
		0.281295,
		0.245133,
		0.210511,
		0.177679,
		0.146884,
		0.000647,
		0.012401,
		0.037057,
		0.072345,
		0.116196,
		0.166724,
		0.222223,
		0.281153,
		0.342132,
		0.40393,
		0.465459,
		0.525768,
		0.584036,
		0.639566,
		0.691782,
		0.74022,
		0.784529,
		0.824461,
		0.859876,
		0.890728,
		0.917071,
		0.939048,
		0.956891,
		0.970913,
		0.981498,
		0.989094,
		0.994196,
		0.997327,
		0.999012,
		0.99975,
		0.999972,
		1,
		0.999998,
		0.999937,
		0.999618,
		0.998745,
		0.996976,
		0.993962,
		0.989376,
		0.982931,
		0.974386,
		0.963556,
		0.950308,
		0.934562,
		0.916286,
		0.895491,
		0.87223,
		0.846589,
		0.818687,
		0.788668,
		0.756699,
		0.722968,
		0.687676,
		0.651042,
		0.613292,
		0.574663,
		0.535398,
		0.495746,
		0.455958,
		0.416289,
		0.376995,
		0.33833,
		0.300551,
		0.26391,
		0.228658,
		0.195044,
		0.000161,
		0.009762,
		0.032464,
		0.066007,
		0.108329,
		0.157547,
		0.211958,
		0.27002,
		0.330348,
		0.391707,
		0.453002,
		0.513271,
		0.571684,
		0.62753,
		0.680218,
		0.729267,
		0.774307,
		0.815072,
		0.851398,
		0.883219,
		0.910561,
		0.933544,
		0.952372,
		0.967332,
		0.978782,
		0.987147,
		0.992899,
		0.996548,
		0.998612,
		0.999591,
		0.999934,
		0.999998,
		1,
		0.999974,
		0.999775,
		0.999138,
		0.997737,
		0.995229,
		0.991281,
		0.985597,
		0.977922,
		0.968053,
		0.955839,
		0.941183,
		0.924033,
		0.904386,
		0.882275,
		0.857774,
		0.830985,
		0.80204,
		0.771094,
		0.738322,
		0.703917,
		0.668086,
		0.631048,
		0.593031,
		0.554272,
		0.515011,
		0.475496,
		0.435974,
		0.396697,
		0.357915,
		0.31988,
		0.282841,
		0.247045,
		0.007447,
		0.028187,
		0.059973,
		0.100747,
		0.148635,
		0.201932,
		0.259096,
		0.318742,
		0.379627,
		0.44065,
		0.500843,
		0.559361,
		0.615483,
		0.668603,
		0.718226,
		0.763962,
		0.805528,
		0.842738,
		0.875503,
		0.903828,
		0.927805,
		0.947615,
		0.963517,
		0.975845,
		0.984999,
		0.991431,
		0.995632,
		0.998111,
		0.99937,
		0.999869,
		0.999991,
		1,
		0.999992,
		0.999878,
		0.999431,
		0.998347,
		0.996289,
		0.992925,
		0.987949,
		0.981094,
		0.972142,
		0.960924,
		0.947325,
		0.931276,
		0.912756,
		0.891784,
		0.868415,
		0.842738,
		0.814873,
		0.784961,
		0.753166,
		0.71967,
		0.684672,
		0.64838,
		0.611015,
		0.572806,
		0.533987,
		0.494799,
		0.455485,
		0.416289,
		0.377459,
		0.339242,
		0.301882,
		0.00545,
		0.024223,
		0.05424,
		0.093452,
		0.139988,
		0.192146,
		0.248386,
		0.307317,
		0.367695,
		0.428411,
		0.488488,
		0.547074,
		0.603434,
		0.656948,
		0.707107,
		0.753505,
		0.795839,
		0.833904,
		0.86759,
		0.896877,
		0.921837,
		0.942622,
		0.959469,
		0.972685,
		0.982646,
		0.989783,
		0.994569,
		0.997501,
		0.999078,
		0.999767,
		0.999974,
		1,
		0.999998,
		0.999941,
		0.999643,
		0.998826,
		0.997165,
		0.994331,
		0.990011,
		0.983926,
		0.975845,
		0.965583,
		0.953007,
		0.938031,
		0.920616,
		0.900765,
		0.878519,
		0.853952,
		0.827169,
		0.7983,
		0.767497,
		0.734932,
		0.700791,
		0.665277,
		0.628601,
		0.590984,
		0.552654,
		0.513845,
		0.474793,
		0.435741,
		0.396928,
		0.358598,
		0.003771,
		0.02057,
		0.048808,
		0.086442,
		0.131607,
		0.182603,
		0.237891,
		0.296078,
		0.355915,
		0.416289,
		0.476215,
		0.53483,
		0.591389,
		0.645261,
		0.695919,
		0.742944,
		0.786014,
		0.824905,
		0.859486,
		0.889716,
		0.915644,
		0.937398,
		0.955189,
		0.969301,
		0.980084,
		0.987949,
		0.99335,
		0.996771,
		0.998703,
		0.999618,
		0.999939,
		0.999998,
		1,
		0.999976,
		0.99979,
		0.999193,
		0.997879,
		0.995521,
		0.991804,
		0.98644,
		0.979184,
		0.969836,
		0.958246,
		0.944313,
		0.927979,
		0.909232,
		0.888097,
		0.864634,
		0.838934,
		0.811114,
		0.781314,
		0.749696,
		0.716437,
		0.681729,
		0.645775,
		0.608789,
		0.570991,
		0.532609,
		0.493875,
		0.455022,
		0.416289,
		0.002404,
		0.017226,
		0.043674,
		0.079717,
		0.123492,
		0.173305,
		0.227614,
		0.285028,
		0.344293,
		0.404291,
		0.46403,
		0.522638,
		0.579359,
		0.63355,
		0.684672,
		0.732288,
		0.776061,
		0.815748,
		0.851199,
		0.882352,
		0.909232,
		0.931946,
		0.950679,
		0.965692,
		0.977312,
		0.985926,
		0.991969,
		0.995911,
		0.998234,
		0.999412,
		0.999878,
		0.999992,
		1,
		0.999992,
		0.999886,
		0.999468,
		0.998451,
		0.996517,
		0.99335,
		0.988657,
		0.982179,
		0.973702,
		0.963062,
		0.950139,
		0.93486,
		0.917197,
		0.897159,
		0.874792,
		0.850172,
		0.823405,
		0.794618,
		0.763962,
		0.731605,
		0.697728,
		0.662527,
		0.626207,
		0.588982,
		0.551073,
		0.512705,
		0.474108,
		0.001347,
		0.014187,
		0.038838,
		0.073275,
		0.115644,
		0.164251,
		0.217558,
		0.27417,
		0.332833,
		0.392422,
		0.451939,
		0.510503,
		0.567351,
		0.621825,
		0.673374,
		0.721547,
		0.76599,
		0.806444,
		0.842738,
		0.874792,
		0.902608,
		0.92627,
		0.945942,
		0.961859,
		0.974325,
		0.983707,
		0.990418,
		0.994912,
		0.997662,
		0.999138,
		0.999782,
		0.999976,
		1,
		0.999998,
		0.999945,
		0.999666,
		0.9989,
		0.99734,
		0.994673,
		0.990599,
		0.984851,
		0.977203,
		0.967473,
		0.955526,
		0.941275,
		0.924674,
		0.905717,
		0.884434,
		0.86089,
		0.835177,
		0.80741,
		0.777728,
		0.746289,
		0.713267,
		0.678846,
		0.643226,
		0.606611,
		0.569217,
		0.531263,
		0.000596,
		0.011451,
		0.034297,
		0.067117,
		0.108062,
		0.155444,
		0.207724,
		0.263508,
		0.321539,
		0.380687,
		0.439948,
		0.498434,
		0.555372,
		0.610092,
		0.662033,
		0.710728,
		0.75581,
		0.797,
		0.834112,
		0.867043,
		0.895777,
		0.920376,
		0.94098,
		0.957802,
		0.971124,
		0.981289,
		0.988691,
		0.993766,
		0.996976,
		0.998786,
		0.999643,
		0.999943,
		0.999998,
		1,
		0.999977,
		0.999803,
		0.999244,
		0.99801,
		0.995792,
		0.992288,
		0.987222,
		0.980357,
		0.971498,
		0.960493,
		0.947239,
		0.931675,
		0.913781,
		0.893571,
		0.871096,
		0.846434,
		0.819691,
		0.790993,
		0.760489,
		0.72834,
		0.694725,
		0.659834,
		0.623865,
		0.587025,
		0.000148,
		0.009016,
		0.030049,
		0.061241,
		0.100747,
		0.146884,
		0.198115,
		0.253045,
		0.310415,
		0.369091,
		0.428063,
		0.486437,
		0.543429,
		0.598361,
		0.650658,
		0.699842,
		0.745529,
		0.787425,
		0.825327,
		0.859113,
		0.888746,
		0.914268,
		0.935798,
		0.953525,
		0.967708,
		0.978669,
		0.986782,
		0.992466,
		0.996168,
		0.998347,
		0.99945,
		0.999886,
		0.999993,
		1,
		0.999993,
		0.999893,
		0.999501,
		0.998546,
		0.996728,
		0.993744,
		0.989313,
		0.983187,
		0.975156,
		0.965057,
		0.952769,
		0.938217,
		0.921364,
		0.902212,
		0.880796,
		0.857183,
		0.831466,
		0.803759,
		0.774201,
		0.742944,
		0.710157,
		0.676022,
		0.64073,
		0.006878,
		0.026093,
		0.055645,
		0.093699,
		0.138573,
		0.188732,
		0.242783,
		0.299464,
		0.357638,
		0.416289,
		0.474517,
		0.531529,
		0.586639,
		0.639258,
		0.688896,
		0.735156,
		0.777728,
		0.816392,
		0.851009,
		0.881522,
		0.907953,
		0.930398,
		0.949027,
		0.964076,
		0.975845,
		0.984688,
		0.991005,
		0.995229,
		0.997809,
		0.999193,
		0.999797,
		0.999977,
		1,
		0.999999,
		0.999948,
		0.999687,
		0.998968,
		0.997501,
		0.994989,
		0.991144,
		0.985711,
		0.978467,
		0.969235,
		0.957881,
		0.944313,
		0.928479,
		0.910368,
		0.890001,
		0.867431,
		0.842738,
		0.816028,
		0.787425,
		0.757075,
		0.725136,
		0.691782,
		0.005035,
		0.022426,
		0.05033,
		0.086916,
		0.130509,
		0.179577,
		0.232725,
		0.28869,
		0.346332,
		0.404632,
		0.462682,
		0.51968,
		0.574932,
		0.627839,
		0.677898,
		0.724699,
		0.767917,
		0.807315,
		0.842738,
		0.874111,
		0.901435,
		0.924786,
		0.944313,
		0.96023,
		0.972816,
		0.982404,
		0.989376,
		0.99415,
		0.997165,
		0.998863,
		0.999666,
		0.999947,
		0.999999,
		1,
		0.999979,
		0.999816,
		0.999291,
		0.998131,
		0.996042,
		0.992736,
		0.987949,
		0.98145,
		0.973047,
		0.962592,
		0.949978,
		0.935141,
		0.918052,
		0.89872,
		0.877185,
		0.853514,
		0.827802,
		0.800163,
		0.770731,
		0.739658,
		0.003484,
		0.019046,
		0.045292,
		0.080399,
		0.122695,
		0.170651,
		0.222873,
		0.278096,
		0.335179,
		0.393097,
		0.450936,
		0.507888,
		0.563249,
		0.61641,
		0.666857,
		0.714166,
		0.758,
		0.798105,
		0.834309,
		0.86652,
		0.89472,
		0.918966,
		0.939384,
		0.95617,
		0.96958,
		0.979927,
		0.987575,
		0.992925,
		0.996405,
		0.998451,
		0.999485,
		0.999893,
		0.999993,
		1,
		0.999993,
		0.9999,
		0.999532,
		0.998635,
		0.996923,
		0.994109,
		0.989923,
		0.984125,
		0.976511,
		0.96692,
		0.95523,
		0.941362,
		0.925275,
		0.906963,
		0.886452,
		0.863799,
		0.839085,
		0.812415,
		0.783913,
		0.002221,
		0.015951,
		0.040531,
		0.074147,
		0.11513,
		0.161955,
		0.213228,
		0.267685,
		0.324181,
		0.381687,
		0.439285,
		0.496159,
		0.551596,
		0.604978,
		0.655781,
		0.703566,
		0.747985,
		0.788769,
		0.825729,
		0.858756,
		0.887815,
		0.912942,
		0.934245,
		0.951898,
		0.966137,
		0.977255,
		0.985597,
		0.991548,
		0.995521,
		0.997946,
		0.999244,
		0.99981,
		0.999979,
		1,
		0.999999,
		0.999952,
		0.999707,
		0.999031,
		0.997651,
		0.995282,
		0.991651,
		0.98651,
		0.979645,
		0.970881,
		0.960083,
		0.947158,
		0.932051,
		0.914741,
		0.895243,
		0.873601,
		0.849883,
		0.824185,
		0.001245,
		0.013138,
		0.036046,
		0.068159,
		0.107814,
		0.15349,
		0.203793,
		0.25746,
		0.313342,
		0.370409,
		0.427734,
		0.484498,
		0.539979,
		0.59355,
		0.644676,
		0.692907,
		0.737881,
		0.779315,
		0.817006,
		0.850827,
		0.880725,
		0.90672,
		0.928899,
		0.947416,
		0.962487,
		0.974386,
		0.983438,
		0.990011,
		0.994505,
		0.99734,
		0.998934,
		0.999687,
		0.99995,
		0.999999,
		1,
		0.99998,
		0.999827,
		0.999334,
		0.998243,
		0.996274,
		0.993153,
		0.988626,
		0.982468,
		0.974494,
		0.964555,
		0.952544,
		0.938392,
		0.922066,
		0.903567,
		0.882926,
		0.860201,
		0.000551,
		0.010606,
		0.031834,
		0.062433,
		0.100747,
		0.145256,
		0.19457,
		0.247423,
		0.302666,
		0.359265,
		0.416289,
		0.472913,
		0.528406,
		0.582134,
		0.63355,
		0.682196,
		0.727694,
		0.769751,
		0.808146,
		0.842738,
		0.873458,
		0.900305,
		0.92335,
		0.942727,
		0.958631,
		0.971319,
		0.981094,
		0.98831,
		0.99335,
		0.996625,
		0.998546,
		0.999517,
		0.9999,
		0.999993,
		1,
		0.999994,
		0.999906,
		0.999561,
		0.998717,
		0.997104,
		0.994448,
		0.99049,
		0.984999,
		0.977776,
		0.968661,
		0.957534,
		0.944313,
		0.92895,
		0.911434,
		0.891784,
		0.000137,
		0.008351,
		0.027894,
		0.05697,
		0.09393,
		0.137255,
		0.185559,
		0.237576,
		0.292156,
		0.34826,
		0.404954,
		0.461407,
		0.516882,
		0.570735,
		0.622411,
		0.67144,
		0.717434,
		0.760084,
		0.799158,
		0.834498,
		0.866019,
		0.893703,
		0.917603,
		0.937833,
		0.954571,
		0.968053,
		0.978564,
		0.98644,
		0.99205,
		0.995792,
		0.998072,
		0.999291,
		0.999822,
		0.99998,
		1,
		0.999999,
		0.999955,
		0.999725,
		0.999089,
		0.997789,
		0.995554,
		0.992122,
		0.987255,
		0.980745,
		0.972419,
		0.962145,
		0.949827,
		0.935405,
		0.918855,
		0.006372,
		0.024223,
		0.051768,
		0.08736,
		0.129487,
		0.176763,
		0.227922,
		0.281814,
		0.337397,
		0.393734,
		0.449987,
		0.505413,
		0.559361,
		0.611266,
		0.660647,
		0.707107,
		0.750323,
		0.790049,
		0.826114,
		0.858415,
		0.88692,
		0.911662,
		0.932739,
		0.950308,
		0.964588,
		0.975845,
		0.984397,
		0.990599,
		0.994834,
		0.997501,
		0.999,
		0.999707,
		0.999953,
		0.999999,
		1,
		0.999981,
		0.999838,
		0.999375,
		0.998347,
		0.99649,
		0.993541,
		0.989256,
		0.983418,
		0.975845,
		0.966392,
		0.954949,
		0.941445,
		0.004665,
		0.020821,
		0.046826,
		0.081039,
		0.121952,
		0.168182,
		0.218463,
		0.271644,
		0.326681,
		0.382633,
		0.438658,
		0.494005,
		0.548017,
		0.60012,
		0.649825,
		0.69672,
		0.740474,
		0.780826,
		0.817592,
		0.850653,
		0.87996,
		0.905532,
		0.927446,
		0.945845,
		0.960924,
		0.972936,
		0.982179,
		0.988991,
		0.993744,
		0.996827,
		0.998635,
		0.999547,
		0.999906,
		0.999994,
		1,
		0.999994,
		0.999912,
		0.999588,
		0.998793,
		0.997272,
		0.994763,
		0.991018,
		0.985814,
		0.978957,
		0.970291,
		0.959694,
		0.003228,
		0.017685,
		0.042142,
		0.074965,
		0.114651,
		0.159817,
		0.2092,
		0.261647,
		0.316114,
		0.371655,
		0.427424,
		0.482664,
		0.536711,
		0.588982,
		0.638979,
		0.686282,
		0.730545,
		0.771497,
		0.808939,
		0.842738,
		0.872831,
		0.899218,
		0.921961,
		0.941183,
		0.957064,
		0.969836,
		0.979781,
		0.987222,
		0.992516,
		0.996042,
		0.998188,
		0.999334,
		0.999833,
		0.999981,
		1,
		0.999999,
		0.999957,
		0.999741,
		0.999143,
		0.997917,
		0.995807,
		0.992561,
		0.987949,
		0.981771,
		0.973858,
		0.002059,
		0.014812,
		0.037715,
		0.069138,
		0.107582,
		0.151669,
		0.200135,
		0.251827,
		0.3057,
		0.360804,
		0.416289,
		0.471394,
		0.525446,
		0.577857,
		0.628118,
		0.675798,
		0.720543,
		0.762069,
		0.800163,
		0.834679,
		0.865538,
		0.892725,
		0.916286,
		0.936326,
		0.953007,
		0.966545,
		0.977203,
		0.985289,
		0.991144,
		0.995138,
		0.997651,
		0.999061,
		0.999725,
		0.999956,
		0.999999,
		1,
		0.999982,
		0.999848,
		0.999412,
		0.998443,
		0.99669,
		0.993901,
		0.989843,
		0.984305,
		0.001154,
		0.012202,
		0.033544,
		0.063557,
		0.100747,
		0.143738,
		0.191269,
		0.242186,
		0.295441,
		0.350084,
		0.405259,
		0.460201,
		0.51423,
		0.566751,
		0.617247,
		0.665277,
		0.710476,
		0.752549,
		0.79127,
		0.826481,
		0.858088,
		0.886059,
		0.910426,
		0.931276,
		0.948756,
		0.963062,
		0.974442,
		0.983187,
		0.989624,
		0.994109,
		0.997015,
		0.998717,
		0.999575,
		0.999912,
		0.999994,
		1,
		0.999994,
		0.999917,
		0.999612,
		0.998863,
		0.997428,
		0.995056,
		0.99151,
		0.000511,
		0.009851,
		0.029626,
		0.058222,
		0.094146,
		0.136026,
		0.182603,
		0.232725,
		0.285341,
		0.339498,
		0.394338,
		0.449089,
		0.503068,
		0.555671,
		0.606372,
		0.654724,
		0.70035,
		0.742944,
		0.782268,
		0.818151,
		0.850486,
		0.879226,
		0.904386,
		0.926039,
		0.944313,
		0.959389,
		0.971498,
		0.980913,
		0.987949,
		0.992949,
		0.996274,
		0.998296,
		0.999375,
		0.999843,
		0.999982,
		1,
		0.999999,
		0.99996,
		0.999757,
		0.999193,
		0.998037,
		0.996042,
		0.000127,
		0.007758,
		0.025962,
		0.05313,
		0.087777,
		0.128533,
		0.17414,
		0.223446,
		0.275401,
		0.32905,
		0.383529,
		0.438064,
		0.491965,
		0.544621,
		0.595501,
		0.644146,
		0.690172,
		0.73326,
		0.773163,
		0.809696,
		0.842738,
		0.87223,
		0.89817,
		0.920616,
		0.93968,
		0.955526,
		0.968369,
		0.978467,
		0.986118,
		0.991651,
		0.995421,
		0.997789,
		0.999117,
		0.999741,
		0.999959,
		0.999999,
		1,
		0.999984,
		0.999857,
		0.999446,
		0.998532,
		0.00592,
		0.022547,
		0.048281,
		0.08164,
		0.121258,
		0.165879,
		0.214352,
		0.265625,
		0.318742,
		0.372836,
		0.427129,
		0.480925,
		0.533608,
		0.584639,
		0.63355,
		0.679948,
		0.723506,
		0.763962,
		0.801123,
		0.834852,
		0.865077,
		0.891784,
		0.915012,
		0.93486,
		0.951476,
		0.965057,
		0.975845,
		0.984125,
		0.990212,
		0.994448,
		0.997189,
		0.998793,
		0.9996,
		0.999917,
		0.999995,
		1,
		0.999995,
		0.999922,
		0.999635,
		0.004334,
		0.019382,
		0.043674,
		0.075735,
		0.114202,
		0.157821,
		0.205443,
		0.256014,
		0.308577,
		0.362263,
		0.416289,
		0.469955,
		0.522638,
		0.573791,
		0.622942,
		0.669685,
		0.713686,
		0.754672,
		0.792437,
		0.826833,
		0.857774,
		0.885232,
		0.909232,
		0.929857,
		0.947239,
		0.96156,
		0.973047,
		0.981968,
		0.988626,
		0.99335,
		0.99649,
		0.998396,
		0.999412,
		0.999852,
		0.999984,
		1,
		0.999999,
		0.999962,
		0.003,
		0.016464,
		0.039308,
		0.070061,
		0.107365,
		0.149968,
		0.196721,
		0.246571,
		0.298559,
		0.351814,
		0.405548,
		0.459058,
		0.511714,
		0.562964,
		0.612327,
		0.65939,
		0.703808,
		0.745299,
		0.783645,
		0.818687,
		0.850326,
		0.878519,
		0.90328,
		0.924674,
		0.942819,
		0.957881,
		0.970072,
		0.979645,
		0.986889,
		0.992122,
		0.995683,
		0.997917,
		0.999169,
		0.999757,
		0.999961,
		0.999999,
		1,
		0.001913,
		0.013791,
		0.035182,
		0.064618,
		0.100747,
		0.14232,
		0.188187,
		0.237298,
		0.28869,
		0.34149,
		0.39491,
		0.448238,
		0.500843,
		0.552163,
		0.601712,
		0.649069,
		0.693878,
		0.73585,
		0.774754,
		0.810421,
		0.842738,
		0.871652,
		0.897159,
		0.919314,
		0.938217,
		0.95402,
		0.96692,
		0.977154,
		0.984999,
		0.990759,
		0.994763,
		0.997351,
		0.998863,
		0.999624,
		0.999922,
		0.999995,
		0.001072,
		0.011362,
		0.031293,
		0.059405,
		0.094348,
		0.134877,
		0.179843,
		0.228195,
		0.278972,
		0.331297,
		0.384378,
		0.437501,
		0.490028,
		0.541394,
		0.591103,
		0.638727,
		0.683903,
		0.726331,
		0.76577,
		0.80204,
		0.835018,
		0.864634,
		0.890876,
		0.913781,
		0.933436,
		0.949978,
		0.96359,
		0.974494,
		0.982951,
		0.989256,
		0.993724,
		0.99669,
		0.998488,
		0.999446,
		0.999861,
		0.000475,
		0.009174,
		0.027641,
		0.054421,
		0.088168,
		0.127639,
		0.171689,
		0.219266,
		0.269407,
		0.321236,
		0.373956,
		0.42685,
		0.479275,
		0.530661,
		0.580504,
		0.628371,
		0.673888,
		0.716747,
		0.756699,
		0.793551,
		0.827169,
		0.857473,
		0.884434,
		0.908079,
		0.928479,
		0.945759,
		0.960083,
		0.971663,
		0.980745,
		0.987608,
		0.992561,
		0.995927,
		0.998037,
		0.999217,
		0.000118,
		0.007225,
		0.024223,
		0.049664,
		0.082205,
		0.120607,
		0.163726,
		0.210511,
		0.259999,
		0.31131,
		0.363647,
		0.416289,
		0.468589,
		0.519969,
		0.569922,
		0.618005,
		0.66384,
		0.707107,
		0.747548,
		0.784961,
		0.8192,
		0.850172,
		0.877839,
		0.902212,
		0.92335,
		0.941362,
		0.956401,
		0.968661,
		0.978376,
		0.985814,
		0.991269,
		0.995056,
		0.997501,
		0.005514,
		0.02104,
		0.045134,
		0.076461,
		0.113781,
		0.155955,
		0.201932,
		0.250747,
		0.301522,
		0.353455,
		0.405822,
		0.457972,
		0.509323,
		0.559361,
		0.607637,
		0.653763,
		0.697414,
		0.738322,
		0.776274,
		0.811114,
		0.842738,
		0.871096,
		0.896184,
		0.918052,
		0.936792,
		0.952544,
		0.965488,
		0.975845,
		0.983869,
		0.989843,
		0.994072,
		0.004038,
		0.018088,
		0.04083,
		0.070933,
		0.107161,
		0.148377,
		0.193529,
		0.241656,
		0.291874,
		0.343382,
		0.395453,
		0.44743,
		0.498728,
		0.548826,
		0.59727,
		0.643665,
		0.687676,
		0.729027,
		0.767497,
		0.802918,
		0.835177,
		0.864209,
		0.890001,
		0.912588,
		0.932051,
		0.948514,
		0.962145,
		0.97315,
		0.981771,
		0.988279,
		0.002795,
		0.015366,
		0.036751,
		0.065621,
		0.100747,
		0.140991,
		0.185305,
		0.232725,
		0.282369,
		0.333432,
		0.385184,
		0.436966,
		0.488188,
		0.538323,
		0.586911,
		0.63355,
		0.677898,
		0.71967,
		0.758636,
		0.794618,
		0.827492,
		0.857183,
		0.883666,
		0.906963,
		0.927141,
		0.944313,
		0.958631,
		0.970291,
		0.979518,
		0.001783,
		0.012873,
		0.032895,
		0.060526,
		0.094539,
		0.133799,
		0.17726,
		0.223957,
		0.273008,
		0.323607,
		0.37502,
		0.426585,
		0.477707,
		0.527856,
		0.576564,
		0.623424,
		0.668086,
		0.710256,
		0.749696,
		0.78622,
		0.819691,
		0.850025,
		0.877185,
		0.90118,
		0.922066,
		0.939942,
		0.954949,
		0.967266,
		0.000999,
		0.010606,
		0.029261,
		0.055645,
		0.088536,
		0.126801,
		0.169394,
		0.215353,
		0.263794,
		0.313909,
		0.364963,
		0.416289,
		0.46729,
		0.51743,
		0.566235,
		0.613292,
		0.658245,
		0.700791,
		0.740684,
		0.777728,
		0.811778,
		0.842738,
		0.870561,
		0.895243,
		0.916829,
		0.935405,
		0.951098,
		0.000443,
		0.008564,
		0.025848,
		0.050979,
		0.082738,
		0.119997,
		0.16171,
		0.206915,
		0.254728,
		0.304341,
		0.355015,
		0.406083,
		0.45694,
		0.507048,
		0.555928,
		0.60316,
		0.64838,
		0.69128,
		0.731605,
		0.769149,
		0.803759,
		0.835329,
		0.863799,
		0.889157,
		0.911434,
		0.930704,
		0.00011,
		0.006746,
		0.022654,
		0.046525,
		0.077145,
		0.113386,
		0.154206,
		0.198643,
		0.245813,
		0.294906,
		0.345181,
		0.395969,
		0.446662,
		0.496716,
		0.545648,
		0.593031,
		0.638497,
		0.681729,
		0.722464,
		0.760489,
		0.79564,
		0.827802,
		0.856905,
		0.882926,
		0.905884,
		0.005149,
		0.019678,
		0.042284,
		0.071756,
		0.10697,
		0.146884,
		0.190539,
		0.237049,
		0.285604,
		0.335463,
		0.385951,
		0.436458,
		0.486437,
		0.535398,
		0.582912,
		0.628601,
		0.672143,
		0.713267,
		0.751752,
		0.787425,
		0.820162,
		0.849883,
		0.876554,
		0.003771,
		0.016918,
		0.038254,
		0.066571,
		0.100747,
		0.139745,
		0.182603,
		0.228439,
		0.27644,
		0.325863,
		0.376031,
		0.426333,
		0.476215,
		0.525185,
		0.572806,
		0.618696,
		0.662527,
		0.704019,
		0.742944,
		0.779121,
		0.812415,
		0.842738,
		0.00261,
		0.014374,
		0.034434,
		0.061589,
		0.094719,
		0.132788,
		0.174837,
		0.219983,
		0.267414,
		0.316383,
		0.366214,
		0.416289,
		0.466054,
		0.515011,
		0.562719,
		0.608789,
		0.652886,
		0.694725,
		0.73407,
		0.770731,
		0.804566,
		0.001665,
		0.012042,
		0.030823,
		0.056809,
		0.088884,
		0.126014,
		0.167242,
		0.211684,
		0.258527,
		0.307026,
		0.3565,
		0.406331,
		0.455958,
		0.504881,
		0.552654,
		0.598882,
		0.643226,
		0.685391,
		0.725136,
		0.762263,
		0.000934,
		0.009923,
		0.02742,
		0.05223,
		0.083242,
		0.119423,
		0.159817,
		0.203541,
		0.249783,
		0.297795,
		0.346894,
		0.39646,
		0.44593,
		0.494799,
		0.542616,
		0.588982,
		0.63355,
		0.676022,
		0.716147,
		0.000414,
		0.008014,
		0.024223,
		0.047852,
		0.077793,
		0.113014,
		0.152563,
		0.195557,
		0.241182,
		0.28869,
		0.337397,
		0.386681,
		0.435974,
		0.484769,
		0.532609,
		0.579092,
		0.623865,
		0.666622,
		0.000103,
		0.006313,
		0.021232,
		0.043674,
		0.072536,
		0.106789,
		0.145482,
		0.187731,
		0.232725,
		0.279714,
		0.328012,
		0.376995,
		0.426093,
		0.474793,
		0.522638,
		0.569217,
		0.614173,
		0.004818,
		0.018444,
		0.039695,
		0.067471,
		0.100747,
		0.138573,
		0.180066,
		0.224414,
		0.270869,
		0.318742,
		0.367405,
		0.416289,
		0.464877,
		0.512705,
		0.559361,
		0.003529,
		0.015859,
		0.035914,
		0.062598,
		0.094888,
		0.131836,
		0.172561,
		0.216251,
		0.262156,
		0.309587,
		0.357915,
		0.406567,
		0.455022,
		0.502815,
		0.002443,
		0.013475,
		0.03233,
		0.057915,
		0.089212,
		0.125272,
		0.165218,
		0.208236,
		0.253577,
		0.300551,
		0.348527,
		0.396928,
		0.445233,
		0.001559,
		0.01129,
		0.028941,
		0.053423,
		0.083718,
		0.118882,
		0.158037,
		0.200371,
		0.245133,
		0.291635,
		0.339242,
		0.387376,
		0.000874,
		0.009304,
		0.025748,
		0.049119,
		0.078406,
		0.112664,
		0.151018,
		0.192655,
		0.236827,
		0.282841,
		0.330063,
		0.000387,
		0.007514,
		0.022747,
		0.045004,
		0.073275,
		0.106619,
		0.144161,
		0.185091,
		0.228658,
		0.27417,
		9.6e-005,
		0.00592,
		0.019939,
		0.041077,
		0.068326,
		0.100747,
		0.137468,
		0.177679,
		0.220628,
		0.004519,
		0.017323,
		0.037337,
		0.063557,
		0.095049,
		0.130939,
		0.170419,
		0.00331,
		0.014896,
		0.033782,
		0.058969,
		0.089522,
		0.124573,
		0.002292,
		0.012657,
		0.030412,
		0.05456,
		0.084169,
		0.001462,
		0.010606,
		0.027226,
		0.05033,
		0.00082,
		0.008741,
		0.024223,
		0.000363,
		0.00706,
		9.1e-005,
		0.005562,
		0.000361,
		0.021402,
		0.008674,
		0.00144,
		0.046277,
		0.027023,
		0.012464,
		0.003233,
		0.078987,
		0.054161,
		0.033279,
		0.016928,
		0.005737,
		0.000372,
		0.118371,
		0.088884,
		0.062632,
		0.040162,
		0.022059,
		0.008946,
		0.001486,
		0.163312,
		0.130029,
		0.099316,
		0.071682,
		0.047667,
		0.027851,
		0.012854,
		0.003336,
		0.212739,
		0.176478,
		0.142168,
		0.110275,
		0.081301,
		0.055783,
		0.034297,
		0.017457,
		0.00592,
		0.000384,
		0.265625,
		0.227159,
		0.190067,
		0.154777,
		0.121749,
		0.09148,
		0.064503,
		0.041389,
		0.022747,
		0.00923,
		0.001534,
		0.320992,
		0.281048,
		0.241943,
		0.204067,
		0.167843,
		0.133727,
		0.102209,
		0.073818,
		0.049119,
		0.028718,
		0.013263,
		0.003444,
		0.377913,
		0.337168,
		0.29677,
		0.257074,
		0.218463,
		0.181354,
		0.146196,
		0.113477,
		0.083718,
		0.057479,
		0.035363,
		0.018011,
		0.006111,
		0.000397,
		0.435513,
		0.394596,
		0.353578,
		0.312776,
		0.272538,
		0.233241,
		0.195295,
		0.159146,
		0.125272,
		0.094192,
		0.06646,
		0.042672,
		0.023468,
		0.009529,
		0.001584,
		0.492972,
		0.452462,
		0.411446,
		0.370205,
		0.32905,
		0.288319,
		0.248386,
		0.209654,
		0.172561,
		0.137583,
		0.105229,
		0.076051,
		0.050639,
		0.029626,
		0.013691,
		0.003558,
		0.549528,
		0.509955,
		0.469512,
		0.428447,
		0.387032,
		0.345573,
		0.304401,
		0.263882,
		0.224414,
		0.18643,
		0.150396,
		0.116819,
		0.086243,
		0.059254,
		0.036479,
		0.018592,
		0.006313,
		0.00041,
		0.604481,
		0.566321,
		0.526971,
		0.486644,
		0.445578,
		0.40404,
		0.362328,
		0.320766,
		0.279714,
		0.239562,
		0.200737,
		0.163699,
		0.128949,
		0.097025,
		0.068506,
		0.044016,
		0.024223,
		0.009842,
		0.001638,
		0.657196,
		0.620871,
		0.583079,
		0.544,
		0.503836,
		0.46282,
		0.42121,
		0.379296,
		0.337397,
		0.295864,
		0.255081,
		0.215467,
		0.177478,
		0.141606,
		0.108384,
		0.078386,
		0.05223,
		0.030578,
		0.014141,
		0.003677,
		0.707107,
		0.672983,
		0.63716,
		0.599783,
		0.561021,
		0.52107,
		0.480153,
		0.438523,
		0.39646,
		0.354276,
		0.312315,
		0.270954,
		0.230606,
		0.191718,
		0.154777,
		0.120309,
		0.088884,
		0.061111,
		0.037648,
		0.019201,
		0.006524,
		0.000424,
		0.753719,
		0.722107,
		0.688606,
		0.653328,
		0.61641,
		0.578013,
		0.538323,
		0.497557,
		0.455958,
		0.413799,
		0.371384,
		0.32905,
		0.287165,
		0.246136,
		0.206404,
		0.168448,
		0.132788,
		0.099986,
		0.070648,
		0.045425,
		0.025016,
		0.010171,
		0.001693,
		0.796619,
		0.767766,
		0.736881,
		0.704043,
		0.669354,
		0.63294,
		0.594955,
		0.555576,
		0.515011,
		0.473495,
		0.431293,
		0.388702,
		0.346048,
		0.303696,
		0.262041,
		0.22152,
		0.182603,
		0.145806,
		0.111681,
		0.08083,
		0.053897,
		0.031577,
		0.014613,
		0.003803,
		0.835475,
		0.809569,
		0.781532,
		0.751412,
		0.719275,
		0.685216,
		0.649351,
		0.611825,
		0.572806,
		0.532493,
		0.491113,
		0.448922,
		0.406208,
		0.363292,
		0.320527,
		0.278304,
		0.237049,
		0.197229,
		0.159349,
		0.123957,
		0.091646,
		0.063057,
		0.038875,
		0.019841,
		0.006746,
		0.000439,
		0.870045,
		0.847207,
		0.822187,
		0.794999,
		0.765677,
		0.73428,
		0.700893,
		0.665622,
		0.628601,
		0.589991,
		0.549982,
		0.50879,
		0.466665,
		0.423884,
		0.38076,
		0.33764,
		0.294906,
		0.252975,
		0.212308,
		0.173402,
		0.136799,
		0.103084,
		0.072891,
		0.046901,
		0.025848,
		0.010517,
		0.001752,
		0.900183,
		0.880467,
		0.858565,
		0.834457,
		0.808146,
		0.779657,
		0.749038,
		0.716363,
		0.681729,
		0.645261,
		0.607109,
		0.567455,
		0.526505,
		0.484498,
		0.441707,
		0.398433,
		0.355015,
		0.311827,
		0.26928,
		0.227824,
		0.18795,
		0.150192,
		0.11513,
		0.083389,
		0.055645,
		0.032625,
		0.015108,
		0.003934,
		0.925841,
		0.909232,
		0.89048,
		0.869533,
		0.846361,
		0.820956,
		0.793332,
		0.763528,
		0.731605,
		0.697651,
		0.661782,
		0.624138,
		0.584888,
		0.544233,
		0.502402,
		0.459655,
		0.416289,
		0.372632,
		0.32905,
		0.285944,
		0.243759,
		0.202976,
		0.164123,
		0.127771,
		0.094539,
		0.065096,
		0.040162,
		0.020513,
		0.006979,
		0.000454,
		0.947082,
		0.933492,
		0.91785,
		0.900072,
		0.880097,
		0.857883,
		0.833411,
		0.806683,
		0.777728,
		0.746597,
		0.713366,
		0.678141,
		0.641052,
		0.60226,
		0.561953,
		0.520351,
		0.477707,
		0.434306,
		0.390469,
		0.346552,
		0.302949,
		0.260094,
		0.218463,
		0.178574,
		0.140991,
		0.106327,
		0.075242,
		0.048451,
		0.026722,
		0.01088,
		0.001814,
		0.964076,
		0.953345,
		0.940699,
		0.926026,
		0.909232,
		0.890244,
		0.869007,
		0.845492,
		0.819691,
		0.791619,
		0.761317,
		0.728851,
		0.694316,
		0.657831,
		0.619546,
		0.57964,
		0.538323,
		0.495839,
		0.452462,
		0.408505,
		0.364313,
		0.320274,
		0.276811,
		0.234393,
		0.193529,
		0.154777,
		0.118739,
		0.086071,
		0.057479,
		0.033726,
		0.01563,
		0.004073,
		0.977109,
		0.969004,
		0.959167,
		0.947458,
		0.933754,
		0.91795,
		0.899959,
		0.879718,
		0.857183,
		0.832336,
		0.805179,
		0.775744,
		0.744084,
		0.710282,
		0.674449,
		0.636722,
		0.59727,
		0.556295,
		0.514027,
		0.470734,
		0.426716,
		0.382312,
		0.337898,
		0.29389,
		0.250747,
		0.208972,
		0.169111,
		0.131761,
		0.09757,
		0.067236,
		0.041515,
		0.02122,
		0.007225,
		0.000471,
		0.986574,
		0.980795,
		0.973507,
		0.964547,
		0.953766,
		0.941027,
		0.926215,
		0.909232,
		0.890001,
		0.868468,
		0.8446,
		0.81839,
		0.789856,
		0.759042,
		0.726017,
		0.690882,
		0.653763,
		0.61482,
		0.57424,
		0.532247,
		0.489096,
		0.44508,
		0.400525,
		0.355799,
		0.31131,
		0.267506,
		0.224882,
		0.183976,
		0.145377,
		0.109724,
		0.077708,
		0.050078,
		0.027641,
		0.011263,
		0.001879,
		0.992969,
		0.989154,
		0.984089,
		0.977591,
		0.969489,
		0.959621,
		0.947842,
		0.934022,
		0.918052,
		0.899843,
		0.879329,
		0.856465,
		0.83123,
		0.803632,
		0.773702,
		0.741497,
		0.707107,
		0.670646,
		0.632263,
		0.592135,
		0.550474,
		0.507526,
		0.463572,
		0.418929,
		0.373956,
		0.32905,
		0.28465,
		0.241242,
		0.199356,
		0.159571,
		0.122519,
		0.088884,
		0.059405,
		0.034883,
		0.016179,
		0.004219,
		0.996876,
		0.994617,
		0.99139,
		0.987004,
		0.981267,
		0.973997,
		0.965026,
		0.954194,
		0.941362,
		0.926409,
		0.909232,
		0.889753,
		0.867913,
		0.843682,
		0.817052,
		0.788042,
		0.756699,
		0.723099,
		0.687346,
		0.649575,
		0.609954,
		0.568684,
		0.525998,
		0.482168,
		0.437501,
		0.392345,
		0.347086,
		0.302157,
		0.258031,
		0.21523,
		0.174326,
		0.135939,
		0.100747,
		0.069482,
		0.042936,
		0.021964,
		0.007484,
		0.000488,
		0.998929,
		0.997796,
		0.995985,
		0.993303,
		0.989554,
		0.984543,
		0.978079,
		0.969981,
		0.960083,
		0.948233,
		0.934295,
		0.918157,
		0.899725,
		0.878929,
		0.855726,
		0.830094,
		0.80204,
		0.771599,
		0.738833,
		0.703836,
		0.66673,
		0.627672,
		0.586849,
		0.544486,
		0.500843,
		0.456215,
		0.410941,
		0.365398,
		0.320005,
		0.275228,
		0.23158,
		0.189623,
		0.149968,
		0.113285,
		0.080297,
		0.051789,
		0.028608,
		0.011667,
		0.001948,
		0.999771,
		0.999343,
		0.998511,
		0.997097,
		0.994912,
		0.991758,
		0.987433,
		0.981742,
		0.974494,
		0.965511,
		0.95463,
		0.941704,
		0.926607,
		0.909232,
		0.889497,
		0.867344,
		0.842738,
		0.815675,
		0.786174,
		0.754286,
		0.720092,
		0.683702,
		0.645261,
		0.604944,
		0.562964,
		0.51957,
		0.475047,
		0.429721,
		0.383959,
		0.338171,
		0.292812,
		0.248386,
		0.205443,
		0.164588,
		0.12648,
		0.091835,
		0.061429,
		0.036101,
		0.016757,
		0.004374,
		0.999985,
		0.999896,
		0.999629,
		0.999046,
		0.997978,
		0.99624,
		0.993633,
		0.989954,
		0.984999,
		0.97857,
		0.97048,
		0.960553,
		0.948631,
		0.934575,
		0.918264,
		0.899603,
		0.878519,
		0.854966,
		0.828924,
		0.800401,
		0.769433,
		0.736089,
		0.700466,
		0.662695,
		0.622942,
		0.581406,
		0.538323,
		0.493969,
		0.448658,
		0.402746,
		0.356632,
		0.310761,
		0.265625,
		0.221767,
		0.179781,
		0.140317,
		0.104082,
		0.071843,
		0.044432,
		0.022747,
		0.007758,
		0.000506,
		1,
		0.999998,
		0.999962,
		0.999813,
		0.999429,
		0.998655,
		0.997312,
		0.995203,
		0.992122,
		0.987863,
		0.98222,
		0.974996,
		0.966004,
		0.955074,
		0.942053,
		0.926809,
		0.909232,
		0.889236,
		0.866759,
		0.841768,
		0.814257,
		0.78425,
		0.751801,
		0.716994,
		0.679948,
		0.640815,
		0.599783,
		0.557076,
		0.512956,
		0.467727,
		0.421733,
		0.375363,
		0.32905,
		0.283276,
		0.238574,
		0.195527,
		0.154777,
		0.11702,
		0.083016,
		0.053588,
		0.029626,
		0.012092,
		0.002021,
		0.999985,
		1,
		1,
		0.99999,
		0.99992,
		0.999688,
		0.999156,
		0.998153,
		0.99649,
		0.993959,
		0.990352,
		0.985457,
		0.979066,
		0.970985,
		0.961031,
		0.949038,
		0.93486,
		0.918374,
		0.899478,
		0.878098,
		0.854185,
		0.827721,
		0.798713,
		0.767202,
		0.73326,
		0.696992,
		0.658536,
		0.618068,
		0.575799,
		0.531979,
		0.4869,
		0.440894,
		0.394338,
		0.347655,
		0.301315,
		0.255841,
		0.211807,
		0.169842,
		0.130634,
		0.094935,
		0.063557,
		0.037384,
		0.017367,
		0.004537,
		0.999771,
		0.999945,
		0.999995,
		1,
		0.999999,
		0.999973,
		0.99985,
		0.999508,
		0.998793,
		0.99752,
		0.995488,
		0.992484,
		0.988294,
		0.982702,
		0.975504,
		0.966504,
		0.955526,
		0.94241,
		0.927017,
		0.909232,
		0.888967,
		0.866157,
		0.840769,
		0.812798,
		0.782268,
		0.749239,
		0.7138,
		0.676077,
		0.636233,
		0.594464,
		0.55101,
		0.506149,
		0.460201,
		0.413531,
		0.366551,
		0.319719,
		0.273547,
		0.228598,
		0.185491,
		0.144906,
		0.107582,
		0.074325,
		0.046007,
		0.023574,
		0.008046,
		0.000525,
		0.998929,
		0.999551,
		0.999859,
		0.999974,
		0.999999,
		1,
		0.999995,
		0.99994,
		0.999741,
		0.999259,
		0.998321,
		0.996733,
		0.994282,
		0.990749,
		0.985916,
		0.979567,
		0.971498,
		0.961517,
		0.949453,
		0.935153,
		0.918486,
		0.89935,
		0.877665,
		0.853382,
		0.826481,
		0.796974,
		0.764902,
		0.730344,
		0.69341,
		0.654248,
		0.613044,
		0.570021,
		0.525446,
		0.479627,
		0.432915,
		0.385711,
		0.338462,
		0.291667,
		0.245878,
		0.201706,
		0.159817,
		0.120941,
		0.085875,
		0.055483,
		0.0307,
		0.012541,
		0.002098,
		0.996876,
		0.99831,
		0.999205,
		0.999698,
		0.99992,
		0.99999,
		1,
		1,
		0.999982,
		0.999882,
		0.999581,
		0.998923,
		0.997721,
		0.995768,
		0.992844,
		0.988724,
		0.983187,
		0.976017,
		0.967013,
		0.955987,
		0.942774,
		0.927229,
		0.909232,
		0.88869,
		0.865538,
		0.839741,
		0.811294,
		0.780225,
		0.746597,
		0.710506,
		0.672085,
		0.631507,
		0.588982,
		0.544762,
		0.499142,
		0.452462,
		0.405109,
		0.357518,
		0.310176,
		0.263625,
		0.218463,
		0.175346,
		0.134994,
		0.098193,
		0.065798,
		0.038736,
		0.018011,
		0.004709,
		0.992969,
		0.995601,
		0.997465,
		0.998695,
		0.999429,
		0.999807,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999957,
		0.999788,
		0.999355,
		0.998482,
		0.99697,
		0.994599,
		0.991144,
		0.986377,
		0.980072,
		0.972017,
		0.962012,
		0.949876,
		0.935451,
		0.918602,
		0.899218,
		0.877219,
		0.852555,
		0.825205,
		0.795181,
		0.762531,
		0.727336,
		0.689715,
		0.649825,
		0.607863,
		0.564067,
		0.518718,
		0.472143,
		0.424716,
		0.37686,
		0.32905,
		0.281814,
		0.23574,
		0.191473,
		0.149721,
		0.111261,
		0.076937,
		0.047667,
		0.024446,
		0.008351,
		0.000546,
		0.986574,
		0.990768,
		0.993987,
		0.996348,
		0.997978,
		0.999017,
		0.999606,
		0.999886,
		0.999982,
		1,
		1,
		0.999989,
		0.999909,
		0.999647,
		0.999046,
		0.997915,
		0.996042,
		0.993199,
		0.989154,
		0.983674,
		0.976536,
		0.967529,
		0.956457,
		0.943146,
		0.927446,
		0.909232,
		0.888407,
		0.864902,
		0.838682,
		0.809744,
		0.778118,
		0.743871,
		0.707107,
		0.667967,
		0.626633,
		0.583329,
		0.538323,
		0.491928,
		0.444504,
		0.39646,
		0.34826,
		0.300419,
		0.253513,
		0.208175,
		0.165103,
		0.125061,
		0.088884,
		0.057479,
		0.031834,
		0.013016,
		0.002179,
		0.977109,
		0.983187,
		0.988116,
		0.991988,
		0.994912,
		0.997011,
		0.99842,
		0.999281,
		0.999741,
		0.999938,
		0.999994,
		1,
		0.999999,
		0.99997,
		0.99983,
		0.999443,
		0.998635,
		0.997199,
		0.994912,
		0.991538,
		0.986838,
		0.980581,
		0.972543,
		0.962515,
		0.950308,
		0.935757,
		0.91872,
		0.899083,
		0.876762,
		0.851704,
		0.823889,
		0.793332,
		0.760084,
		0.724231,
		0.685901,
		0.645261,
		0.602518,
		0.557927,
		0.511786,
		0.464441,
		0.416289,
		0.367779,
		0.319414,
		0.271757,
		0.225429,
		0.181117,
		0.139573,
		0.101621,
		0.068159,
		0.040162,
		0.018691,
		0.004891,
		0.964076,
		0.972299,
		0.97924,
		0.984964,
		0.989554,
		0.993109,
		0.995745,
		0.997593,
		0.998793,
		0.999493,
		0.99984,
		0.999971,
		0.999999,
		1,
		0.999994,
		0.999932,
		0.999707,
		0.999161,
		0.998102,
		0.99631,
		0.993551,
		0.989583,
		0.984164,
		0.977061,
		0.968053,
		0.956935,
		0.943527,
		0.927669,
		0.909232,
		0.888115,
		0.864247,
		0.837591,
		0.808146,
		0.775945,
		0.741059,
		0.703599,
		0.663715,
		0.621603,
		0.577498,
		0.531686,
		0.484498,
		0.436317,
		0.387578,
		0.338772,
		0.290446,
		0.243212,
		0.197741,
		0.154777,
		0.11513,
		0.079688,
		0.049418,
		0.025367,
		0.008674,
		0.000567,
		0.947082,
		0.957642,
		0.966832,
		0.974689,
		0.981267,
		0.986637,
		0.99089,
		0.994133,
		0.99649,
		0.998097,
		0.999103,
		0.999658,
		0.999909,
		0.999989,
		1,
		1,
		0.99998,
		0.999866,
		0.999525,
		0.99878,
		0.997422,
		0.995219,
		0.991928,
		0.987301,
		0.981094,
		0.973075,
		0.963026,
		0.950749,
		0.93607,
		0.918841,
		0.898944,
		0.87629,
		0.850827,
		0.822533,
		0.791425,
		0.757559,
		0.721027,
		0.681965,
		0.640549,
		0.597004,
		0.551596,
		0.504643,
		0.456513,
		0.407628,
		0.358463,
		0.309553,
		0.261496,
		0.214952,
		0.170651,
		0.129392,
		0.092052,
		0.059585,
		0.033031,
		0.013517,
		0.002265,
		0.925841,
		0.938862,
		0.950468,
		0.960668,
		0.969489,
		0.976975,
		0.983187,
		0.988203,
		0.992122,
		0.99506,
		0.997147,
		0.998528,
		0.999355,
		0.999782,
		0.999954,
		0.999997,
		1,
		0.999997,
		0.999951,
		0.99976,
		0.999268,
		0.99828,
		0.996571,
		0.993899,
		0.990011,
		0.984657,
		0.977591,
		0.968585,
		0.957423,
		0.943915,
		0.927897,
		0.909232,
		0.887815,
		0.863572,
		0.836467,
		0.806498,
		0.773702,
		0.738155,
		0.699976,
		0.659326,
		0.61641,
		0.571482,
		0.524843,
		0.476846,
		0.427896,
		0.378457,
		0.32905,
		0.280256,
		0.232725,
		0.187172,
		0.144385,
		0.105229,
		0.070648,
		0.041669,
		0.019411,
		0.005084,
		0.900183,
		0.915718,
		0.929834,
		0.942516,
		0.953766,
		0.963598,
		0.972044,
		0.979155,
		0.984999,
		0.989663,
		0.993253,
		0.995893,
		0.997721,
		0.998889,
		0.999554,
		0.99987,
		0.99998,
		1,
		1,
		0.999988,
		0.999897,
		0.999599,
		0.998917,
		0.997637,
		0.995521,
		0.992316,
		0.987764,
		0.981612,
		0.973615,
		0.963547,
		0.951199,
		0.93639,
		0.918966,
		0.898801,
		0.875806,
		0.849923,
		0.821134,
		0.789457,
		0.754952,
		0.717717,
		0.677898,
		0.635684,
		0.591311,
		0.545064,
		0.497281,
		0.448352,
		0.398725,
		0.348906,
		0.299464,
		0.251033,
		0.204316,
		0.160089,
		0.119202,
		0.082589,
		0.051267,
		0.026342,
		0.009016,
		0.00059,
		0.870045,
		0.888082,
		0.904734,
		0.919966,
		0.933754,
		0.946088,
		0.956972,
		0.966428,
		0.974494,
		0.981226,
		0.986701,
		0.991015,
		0.994282,
		0.996633,
		0.998216,
		0.999187,
		0.999707,
		0.99993,
		0.999993,
		1,
		0.999999,
		0.999966,
		0.999806,
		0.999367,
		0.998451,
		0.996826,
		0.994242,
		0.990437,
		0.985151,
		0.978128,
		0.969125,
		0.95792,
		0.944313,
		0.928131,
		0.909232,
		0.887506,
		0.862877,
		0.835308,
		0.804797,
		0.771385,
		0.735156,
		0.696234,
		0.654791,
		0.611048,
		0.565272,
		0.517785,
		0.468961,
		0.419232,
		0.369091,
		0.319089,
		0.269849,
		0.222058,
		0.176478,
		0.133949,
		0.095391,
		0.061808,
		0.034297,
		0.014049,
		0.002356,
		0.835475,
		0.855939,
		0.875088,
		0.892867,
		0.909232,
		0.924149,
		0.937598,
		0.949573,
		0.960083,
		0.969156,
		0.976835,
		0.983187,
		0.988294,
		0.992261,
		0.995211,
		0.997284,
		0.998635,
		0.999426,
		0.999819,
		0.999967,
		0.999999,
		1,
		0.999993,
		0.999923,
		0.999666,
		0.999046,
		0.997845,
		0.995817,
		0.9927,
		0.988228,
		0.982133,
		0.974162,
		0.964076,
		0.951659,
		0.936718,
		0.919093,
		0.898654,
		0.875306,
		0.848992,
		0.819691,
		0.787425,
		0.752258,
		0.714298,
		0.673697,
		0.630658,
		0.585432,
		0.538323,
		0.48969,
		0.439948,
		0.389573,
		0.339103,
		0.289144,
		0.24037,
		0.193529,
		0.149448,
		0.109032,
		0.073275,
		0.043262,
		0.020173,
		0.005289,
		0.796619,
		0.819379,
		0.840925,
		0.861185,
		0.880097,
		0.897607,
		0.913673,
		0.928264,
		0.941362,
		0.952968,
		0.963095,
		0.971777,
		0.979066,
		0.985035,
		0.989775,
		0.9934,
		0.996042,
		0.99785,
		0.998984,
		0.999612,
		0.999897,
		0.999987,
		1,
		1,
		0.999977,
		0.999847,
		0.999459,
		0.998613,
		0.997073,
		0.99458,
		0.990862,
		0.985648,
		0.978669,
		0.969673,
		0.958426,
		0.944719,
		0.928371,
		0.909232,
		0.887188,
		0.862161,
		0.834112,
		0.803041,
		0.768993,
		0.732056,
		0.692366,
		0.650105,
		0.605508,
		0.55886,
		0.510503,
		0.460836,
		0.410318,
		0.359472,
		0.308888,
		0.259225,
		0.211216,
		0.165674,
		0.123492,
		0.08565,
		0.053221,
		0.027373,
		0.009378,
		0.000615,
		0.753719,
		0.778592,
		0.802379,
		0.824995,
		0.846361,
		0.866407,
		0.88507,
		0.902298,
		0.918052,
		0.932304,
		0.945041,
		0.956268,
		0.966004,
		0.974289,
		0.981184,
		0.986769,
		0.991144,
		0.994434,
		0.996778,
		0.998334,
		0.999268,
		0.999752,
		0.999947,
		0.999996,
		1,
		0.999996,
		0.999944,
		0.999726,
		0.999166,
		0.998044,
		0.996106,
		0.993081,
		0.988691,
		0.982658,
		0.974716,
		0.964615,
		0.952128,
		0.937054,
		0.919224,
		0.898503,
		0.874792,
		0.848031,
		0.818201,
		0.785326,
		0.749475,
		0.710763,
		0.669354,
		0.625463,
		0.579359,
		0.531365,
		0.481862,
		0.431293,
		0.380165,
		0.32905,
		0.278592,
		0.229511,
		0.182603,
		0.138748,
		0.098913,
		0.064157,
		0.035637,
		0.014613,
		0.002453,
		0.707107,
		0.733863,
		0.759685,
		0.784477,
		0.808146,
		0.830607,
		0.851782,
		0.8716,
		0.890001,
		0.906935,
		0.922365,
		0.936266,
		0.948631,
		0.959469,
		0.968806,
		0.976689,
		0.983187,
		0.988388,
		0.992403,
		0.995364,
		0.997422,
		0.998741,
		0.999493,
		0.999852,
		0.999977,
		1,
		1,
		0.999986,
		0.999882,
		0.999542,
		0.998766,
		0.997312,
		0.994912,
		0.991285,
		0.986146,
		0.979216,
		0.97023,
		0.958943,
		0.945135,
		0.928617,
		0.909232,
		0.886861,
		0.861423,
		0.832877,
		0.801227,
		0.76652,
		0.728851,
		0.688367,
		0.645261,
		0.599783,
		0.552238,
		0.502989,
		0.452462,
		0.401146,
		0.349596,
		0.298442,
		0.248386,
		0.200209,
		0.154777,
		0.113042,
		0.076051,
		0.044948,
		0.02098,
		0.005506,
		0.657196,
		0.685564,
		0.71317,
		0.739908,
		0.765677,
		0.790377,
		0.813918,
		0.836212,
		0.857183,
		0.876762,
		0.894888,
		0.911514,
		0.926607,
		0.940146,
		0.952128,
		0.962567,
		0.971498,
		0.978974,
		0.985072,
		0.989891,
		0.993551,
		0.996194,
		0.997978,
		0.999076,
		0.999666,
		0.99992,
		0.999992,
		1,
		0.999998,
		0.999961,
		0.999779,
		0.999278,
		0.998234,
		0.996388,
		0.993458,
		0.989154,
		0.983187,
		0.975277,
		0.965164,
		0.952607,
		0.937398,
		0.919359,
		0.898348,
		0.874262,
		0.84704,
		0.816662,
		0.783156,
		0.746597,
		0.707107,
		0.664862,
		0.620093,
		0.573084,
		0.52418,
		0.473788,
		0.42238,
		0.370494,
		0.318742,
		0.267809,
		0.218463,
		0.171552,
		0.128015,
		0.088884,
		0.055289,
		0.028466,
		0.009762,
		0.000641,
		0.604481,
		0.63415,
		0.663248,
		0.691659,
		0.719275,
		0.745987,
		0.771692,
		0.796291,
		0.819691,
		0.841807,
		0.862562,
		0.881887,
		0.899725,
		0.916031,
		0.930774,
		0.943938,
		0.955526,
		0.965558,
		0.974075,
		0.98114,
		0.986838,
		0.991278,
		0.99459,
		0.996925,
		0.998451,
		0.999346,
		0.999793,
		0.999962,
		0.999998,
		1,
		0.999992,
		0.999911,
		0.999618,
		0.99891,
		0.997543,
		0.995238,
		0.991705,
		0.986646,
		0.979768,
		0.970795,
		0.959469,
		0.94556,
		0.928869,
		0.909232,
		0.886525,
		0.860662,
		0.831602,
		0.799352,
		0.763962,
		0.725537,
		0.68423,
		0.64025,
		0.593863,
		0.545395,
		0.495233,
		0.44383,
		0.391707,
		0.339457,
		0.28775,
		0.237335,
		0.189047,
		0.143807,
		0.102633,
		0.066642,
		0.037057,
		0.015211,
		0.002556,
		0.549528,
		0.580154,
		0.610413,
		0.640185,
		0.669354,
		0.697803,
		0.72542,
		0.752097,
		0.777728,
		0.802214,
		0.825463,
		0.847388,
		0.867913,
		0.886971,
		0.904508,
		0.92048,
		0.93486,
		0.947639,
		0.958822,
		0.968439,
		0.976536,
		0.983187,
		0.988485,
		0.992549,
		0.995521,
		0.997561,
		0.998845,
		0.999557,
		0.999882,
		0.999985,
		1,
		0.999999,
		0.999974,
		0.999825,
		0.999381,
		0.998415,
		0.996662,
		0.993829,
		0.989616,
		0.983719,
		0.975845,
		0.965721,
		0.953097,
		0.937751,
		0.919497,
		0.898188,
		0.873716,
		0.846017,
		0.815072,
		0.780913,
		0.743619,
		0.703324,
		0.660215,
		0.614537,
		0.566596,
		0.516759,
		0.465459,
		0.413199,
		0.360554,
		0.308175,
		0.256796,
		0.207232,
		0.160392,
		0.117276,
		0.078987,
		0.046734,
		0.021837,
		0.005737,
		0.492972,
		0.52418,
		0.555238,
		0.586022,
		0.61641,
		0.646281,
		0.675515,
		0.703994,
		0.731605,
		0.758237,
		0.783784,
		0.808146,
		0.83123,
		0.852952,
		0.873234,
		0.892012,
		0.909232,
		0.924857,
		0.938862,
		0.951243,
		0.962012,
		0.971204,
		0.978877,
		0.985111,
		0.990011,
		0.993706,
		0.996348,
		0.998106,
		0.999166,
		0.999716,
		0.99994,
		0.999996,
		1,
		0.999996,
		0.999935,
		0.999686,
		0.999046,
		0.997765,
		0.995558,
		0.992122,
		0.987147,
		0.980326,
		0.971369,
		0.960006,
		0.945995,
		0.929128,
		0.909232,
		0.886177,
		0.859876,
		0.830285,
		0.797413,
		0.761317,
		0.722107,
		0.679948,
		0.635065,
		0.587741,
		0.538323,
		0.487226,
		0.43493,
		0.381994,
		0.32905,
		0.276811,
		0.226079,
		0.177742,
		0.132788,
		0.092302,
		0.057479,
		0.029626,
		0.010171,
		0.000668,
		0.435513,
		0.466901,
		0.498366,
		0.529781,
		0.561021,
		0.59196,
		0.622473,
		0.652437,
		0.681729,
		0.710231,
		0.737827,
		0.764404,
		0.789856,
		0.814083,
		0.83699,
		0.858494,
		0.878519,
		0.897002,
		0.913893,
		0.929157,
		0.942774,
		0.954745,
		0.96509,
		0.973851,
		0.981094,
		0.986911,
		0.991416,
		0.994749,
		0.997073,
		0.998566,
		0.999422,
		0.999831,
		0.999974,
		0.999999,
		1,
		0.999984,
		0.999864,
		0.999475,
		0.998588,
		0.996928,
		0.994196,
		0.990077,
		0.984253,
		0.97642,
		0.966289,
		0.953597,
		0.938112,
		0.919639,
		0.898023,
		0.873152,
		0.84496,
		0.813428,
		0.778592,
		0.740538,
		0.699408,
		0.655405,
		0.608789,
		0.559887,
		0.50909,
		0.456863,
		0.403741,
		0.350337,
		0.297347,
		0.245554,
		0.19583,
		0.149143,
		0.106564,
		0.069274,
		0.038563,
		0.015846,
		0.002666,
		0.377913,
		0.409054,
		0.440506,
		0.472143,
		0.503836,
		0.535457,
		0.566876,
		0.597967,
		0.628601,
		0.658652,
		0.687996,
		0.716513,
		0.744084,
		0.770596,
		0.795942,
		0.820021,
		0.842738,
		0.864011,
		0.883765,
		0.901939,
		0.918486,
		0.933375,
		0.946591,
		0.958141,
		0.968053,
		0.976376,
		0.983187,
		0.988586,
		0.9927,
		0.995681,
		0.9977,
		0.998946,
		0.999618,
		0.999908,
		0.999991,
		1,
		0.999998,
		0.999955,
		0.999746,
		0.999172,
		0.997978,
		0.995871,
		0.992536,
		0.987648,
		0.980889,
		0.971951,
		0.960553,
		0.94644,
		0.929393,
		0.909232,
		0.88582,
		0.859064,
		0.828924,
		0.795408,
		0.758579,
		0.718555,
		0.675515,
		0.629697,
		0.581406,
		0.531011,
		0.478955,
		0.425753,
		0.371999,
		0.318369,
		0.265625,
		0.214622,
		0.166312,
		0.121749,
		0.082095,
		0.048628,
		0.022747,
		0.005983,
		0.320992,
		0.351437,
		0.382435,
		0.413858,
		0.445578,
		0.477463,
		0.509384,
		0.541209,
		0.572806,
		0.604044,
		0.634794,
		0.664927,
		0.694316,
		0.722839,
		0.750376,
		0.776812,
		0.80204,
		0.825958,
		0.848471,
		0.869498,
		0.888967,
		0.906817,
		0.923006,
		0.937506,
		0.950308,
		0.961427,
		0.970896,
		0.978775,
		0.985151,
		0.990135,
		0.993865,
		0.996504,
		0.998234,
		0.999253,
		0.999763,
		0.999956,
		0.999998,
		1,
		0.999991,
		0.999898,
		0.999561,
		0.99875,
		0.997186,
		0.994557,
		0.990536,
		0.984791,
		0.977002,
		0.966867,
		0.954108,
		0.938482,
		0.919785,
		0.897854,
		0.872571,
		0.843868,
		0.811728,
		0.77619,
		0.737347,
		0.695352,
		0.650423,
		0.602837,
		0.552945,
		0.501165,
		0.447991,
		0.393997,
		0.339837,
		0.286256,
		0.234088,
		0.184267,
		0.137829,
		0.09592,
		0.059802,
		0.030859,
		0.010606,
		0.000697,
		0.265625,
		0.294906,
		0.324985,
		0.355737,
		0.387032,
		0.41874,
		0.450728,
		0.482863,
		0.515011,
		0.547038,
		0.57881,
		0.610192,
		0.641052,
		0.67126,
		0.700686,
		0.729206,
		0.756699,
		0.783049,
		0.808146,
		0.831889,
		0.854185,
		0.874952,
		0.894119,
		0.911631,
		0.927446,
		0.941544,
		0.953921,
		0.964597,
		0.973615,
		0.981047,
		0.986986,
		0.991558,
		0.994912,
		0.997222,
		0.998681,
		0.999493,
		0.999864,
		0.999983,
		1,
		0.999999,
		0.99997,
		0.999798,
		0.999289,
		0.998182,
		0.996177,
		0.992945,
		0.98815,
		0.981456,
		0.972543,
		0.961112,
		0.946896,
		0.929666,
		0.909232,
		0.885451,
		0.858226,
		0.827516,
		0.793332,
		0.755743,
		0.714875,
		0.670922,
		0.624138,
		0.574848,
		0.523449,
		0.470412,
		0.416289,
		0.361715,
		0.307411,
		0.254193,
		0.202976,
		0.154777,
		0.110724,
		0.072063,
		0.040162,
		0.016522,
		0.002783,
		0.212739,
		0.24037,
		0.269049,
		0.298652,
		0.32905,
		0.360112,
		0.391707,
		0.423701,
		0.455958,
		0.488343,
		0.520718,
		0.552945,
		0.584888,
		0.61641,
		0.647376,
		0.677651,
		0.707107,
		0.735615,
		0.763052,
		0.789303,
		0.814257,
		0.837812,
		0.859876,
		0.880367,
		0.899218,
		0.916375,
		0.931803,
		0.945484,
		0.957423,
		0.967646,
		0.976208,
		0.983187,
		0.988691,
		0.992856,
		0.995843,
		0.997839,
		0.999046,
		0.999675,
		0.999931,
		0.999995,
		1,
		0.999995,
		0.999926,
		0.999638,
		0.998903,
		0.997435,
		0.994912,
		0.990993,
		0.985332,
		0.977591,
		0.967454,
		0.95463,
		0.938862,
		0.919935,
		0.897679,
		0.87197,
		0.842738,
		0.809968,
		0.773702,
		0.73404,
		0.69115,
		0.645261,
		0.596674,
		0.545761,
		0.492972,
		0.438833,
		0.383959,
		0.32905,
		0.2749,
		0.222405,
		0.172561,
		0.12648,
		0.085389,
		0.050639,
		0.023716,
		0.006244,
		0.163312,
		0.188793,
		0.215573,
		0.24353,
		0.272538,
		0.302467,
		0.333186,
		0.364563,
		0.39646,
		0.428742,
		0.46127,
		0.493904,
		0.526505,
		0.55893,
		0.591042,
		0.622699,
		0.653763,
		0.6841,
		0.713576,
		0.742062,
		0.769433,
		0.795573,
		0.82037,
		0.843722,
		0.865538,
		0.885739,
		0.904257,
		0.921045,
		0.93607,
		0.949321,
		0.96081,
		0.970571,
		0.978669,
		0.985193,
		0.990264,
		0.994029,
		0.996662,
		0.998361,
		0.999337,
		0.999805,
		0.99997,
		0.999999,
		1,
		0.999981,
		0.999844,
		0.999396,
		0.998376,
		0.996474,
		0.99335,
		0.988652,
		0.982029,
		0.973143,
		0.961681,
		0.947363,
		0.929947,
		0.909232,
		0.88507,
		0.85736,
		0.82606,
		0.791183,
		0.752804,
		0.711062,
		0.666161,
		0.618377,
		0.568056,
		0.515624,
		0.461585,
		0.406528,
		0.351133,
		0.296172,
		0.242518,
		0.191151,
		0.14316,
		0.099754,
		0.062267,
		0.032169,
		0.011069,
		0.000729,
		0.118371,
		0.141187,
		0.165555,
		0.191355,
		0.218463,
		0.246752,
		0.276092,
		0.306352,
		0.337397,
		0.369091,
		0.401293,
		0.433865,
		0.466665,
		0.499548,
		0.532373,
		0.564995,
		0.59727,
		0.629057,
		0.660215,
		0.690605,
		0.720092,
		0.748546,
		0.77584,
		0.801855,
		0.826481,
		0.849616,
		0.871169,
		0.891062,
		0.909232,
		0.925634,
		0.940242,
		0.95305,
		0.964076,
		0.973368,
		0.980996,
		0.987065,
		0.991705,
		0.995079,
		0.997372,
		0.998793,
		0.999561,
		0.999894,
		0.99999,
		1,
		0.999998,
		0.999948,
		0.999707,
		0.999046,
		0.997675,
		0.99526,
		0.991447,
		0.985874,
		0.978187,
		0.968053,
		0.955164,
		0.939252,
		0.92009,
		0.897499,
		0.87135,
		0.84157,
		0.808146,
		0.771123,
		0.730613,
		0.686792,
		0.63991,
		0.590287,
		0.538323,
		0.484498,
		0.429378,
		0.373618,
		0.317968,
		0.26328,
		0.210511,
		0.160731,
		0.11513,
		0.075024,
		0.041863,
		0.017243,
		0.002907,
		0.078987,
		0.098612,
		0.120042,
		0.14316,
		0.167843,
		0.193969,
		0.221409,
		0.250035,
		0.279714,
		0.310309,
		0.341684,
		0.373698,
		0.406208,
		0.439072,
		0.472143,
		0.505276,
		0.538323,
		0.571138,
		0.603574,
		0.635486,
		0.66673,
		0.697165,
		0.726654,
		0.755064,
		0.782268,
		0.808146,
		0.832587,
		0.855488,
		0.876762,
		0.896331,
		0.914137,
		0.930138,
		0.944313,
		0.956664,
		0.967218,
		0.976031,
		0.983187,
		0.9888,
		0.993016,
		0.996009,
		0.997978,
		0.999142,
		0.999727,
		0.99995,
		0.999998,
		1,
		0.999989,
		0.999882,
		0.999493,
		0.99856,
		0.996763,
		0.99375,
		0.989154,
		0.982605,
		0.973751,
		0.962262,
		0.947842,
		0.930235,
		0.909232,
		0.884676,
		0.856465,
		0.824552,
		0.788955,
		0.749757,
		0.707107,
		0.661224,
		0.612405,
		0.561021,
		0.507526,
		0.452462,
		0.39646,
		0.340247,
		0.28465,
		0.230606,
		0.179163,
		0.13149,
		0.088884,
		0.052777,
		0.024747,
		0.006524,
		0.046277,
		0.062177,
		0.080132,
		0.100028,
		0.121749,
		0.145174,
		0.170178,
		0.196635,
		0.224414,
		0.253383,
		0.283404,
		0.31434,
		0.346048,
		0.378386,
		0.411206,
		0.444363,
		0.477707,
		0.511089,
		0.544357,
		0.577362,
		0.609954,
		0.641985,
		0.673308,
		0.70378,
		0.73326,
		0.761615,
		0.788716,
		0.814442,
		0.838682,
		0.861335,
		0.882312,
		0.901541,
		0.918966,
		0.934549,
		0.948277,
		0.960158,
		0.97023,
		0.978558,
		0.985237,
		0.990397,
		0.994196,
		0.996822,
		0.998486,
		0.999417,
		0.999844,
		0.99998,
		1,
		0.999999,
		0.999965,
		0.999767,
		0.999178,
		0.997904,
		0.995601,
		0.991898,
		0.986419,
		0.97879,
		0.968661,
		0.95571,
		0.939652,
		0.920249,
		0.897313,
		0.870708,
		0.840361,
		0.806258,
		0.76845,
		0.727057,
		0.682272,
		0.63436,
		0.583667,
		0.53062,
		0.475734,
		0.419614,
		0.362964,
		0.306588,
		0.251397,
		0.198419,
		0.148801,
		0.103819,
		0.064888,
		0.033565,
		0.011564,
		0.000762,
		0.021402,
		0.033031,
		0.046964,
		0.063092,
		0.081301,
		0.101474,
		0.123492,
		0.14723,
		0.172561,
		0.199356,
		0.227479,
		0.256796,
		0.287165,
		0.318445,
		0.350491,
		0.383156,
		0.416289,
		0.449741,
		0.483358,
		0.516987,
		0.550474,
		0.583667,
		0.61641,
		0.648554,
		0.679948,
		0.710447,
		0.739908,
		0.768196,
		0.795181,
		0.820741,
		0.844764,
		0.867151,
		0.887815,
		0.906686,
		0.923712,
		0.938862,
		0.952128,
		0.963527,
		0.973107,
		0.980944,
		0.987147,
		0.991857,
		0.995249,
		0.997523,
		0.998903,
		0.999625,
		0.99992,
		0.999995,
		1,
		0.999994,
		0.999914,
		0.999581,
		0.998733,
		0.997042,
		0.994144,
		0.989654,
		0.983187,
		0.974369,
		0.962855,
		0.948332,
		0.930531,
		0.909232,
		0.88427,
		0.855538,
		0.82299,
		0.786646,
		0.746597,
		0.703003,
		0.656102,
		0.606211,
		0.553728,
		0.499142,
		0.443032,
		0.386075,
		0.32905,
		0.272844,
		0.218463,
		0.16703,
		0.119801,
		0.07817,
		0.043674,
		0.018011,
		0.003041,
		0.005562,
		0.01237,
		0.021727,
		0.033529,
		0.047667,
		0.064028,
		0.082496,
		0.102952,
		0.125272,
		0.14933,
		0.174994,
		0.202132,
		0.230606,
		0.260276,
		0.290998,
		0.322628,
		0.355015,
		0.38801,
		0.421459,
		0.455206,
		0.489096,
		0.522972,
		0.556677,
		0.590052,
		0.622942,
		0.655192,
		0.686649,
		0.717165,
		0.746597,
		0.774805,
		0.801659,
		0.827036,
		0.850827,
		0.872931,
		0.893264,
		0.91176,
		0.928371,
		0.943071,
		0.955861,
		0.966766,
		0.975845,
		0.983187,
		0.988913,
		0.993181,
		0.996177,
		0.998117,
		0.999236,
		0.999775,
		0.999965,
		0.999999,
		1,
		0.999978,
		0.999819,
		0.9993,
		0.998123,
		0.995933,
		0.992346,
		0.986964,
		0.9794,
		0.969281,
		0.956268,
		0.940063,
		0.920413,
		0.897121,
		0.870045,
		0.839109,
		0.804301,
		0.765677,
		0.723368,
		0.67758,
		0.628601,
		0.5768,
		0.522638,
		0.466665,
		0.409531,
		0.35199,
		0.294906,
		0.239256,
		0.186142,
		0.136799,
		0.092596,
		0.055053,
		0.025848,
		0.006822,
		0.001428,
		0.005649,
		0.01256,
		0.022059,
		0.034038,
		0.048385,
		0.064984,
		0.083718,
		0.104462,
		0.127091,
		0.151474,
		0.177478,
		0.204965,
		0.233795,
		0.263824,
		0.294906,
		0.326889,
		0.359622,
		0.39295,
		0.426716,
		0.460761,
		0.494924,
		0.529046,
		0.562964,
		0.596519,
		0.629549,
		0.661898,
		0.69341,
		0.723934,
		0.753323,
		0.781437,
		0.808146,
		0.833326,
		0.856866,
		0.87867,
		0.898654,
		0.916757,
		0.932935,
		0.94717,
		0.959469,
		0.96987,
		0.97844,
		0.985283,
		0.990536,
		0.994368,
		0.996984,
		0.99861,
		0.999493,
		0.999877,
		0.999988,
		1,
		0.999998,
		0.999939,
		0.99966,
		0.998895,
		0.997312,
		0.994532,
		0.990153,
		0.983772,
		0.974996,
		0.963459,
		0.948834,
		0.930835,
		0.909232,
		0.883851,
		0.854579,
		0.82137,
		0.78425,
		0.743316,
		0.698742,
		0.650784,
		0.599783,
		0.546167,
		0.49046,
		0.433283,
		0.375363,
		0.317536,
		0.260756,
		0.2061,
		0.154777,
		0.108136,
		0.067676,
		0.035053,
		0.012092,
		0.000798,
		0.001451,
		0.005737,
		0.012755,
		0.022399,
		0.034559,
		0.049119,
		0.065963,
		0.084966,
		0.106005,
		0.128949,
		0.153664,
		0.180014,
		0.207857,
		0.237049,
		0.267443,
		0.298888,
		0.331231,
		0.364313,
		0.397978,
		0.432063,
		0.466406,
		0.500843,
		0.535208,
		0.569338,
		0.603067,
		0.636233,
		0.668673,
		0.70023,
		0.73075,
		0.760084,
		0.788092,
		0.814639,
		0.839605,
		0.862877,
		0.884362,
		0.903979,
		0.92167,
		0.937398,
		0.951151,
		0.962947,
		0.972832,
		0.980889,
		0.987232,
		0.992015,
		0.995423,
		0.997675,
		0.999011,
		0.999684,
		0.999942,
		0.999998,
		1,
		0.999987,
		0.999863,
		0.999412,
		0.998331,
		0.996257,
		0.992789,
		0.987511,
		0.980016,
		0.969911,
		0.956839,
		0.940484,
		0.920582,
		0.896922,
		0.869359,
		0.837812,
		0.80227,
		0.762798,
		0.719536,
		0.672708,
		0.622621,
		0.569676,
		0.514364,
		0.457278,
		0.399117,
		0.340688,
		0.282919,
		0.226862,
		0.1737,
		0.12476,
		0.081516,
		0.045605,
		0.018832,
		0.003183,
		0.001474,
		0.005827,
		0.012955,
		0.022747,
		0.035092,
		0.049871,
		0.066963,
		0.086243,
		0.107582,
		0.130847,
		0.155901,
		0.182603,
		0.210809,
		0.24037,
		0.271135,
		0.302949,
		0.335654,
		0.369091,
		0.403095,
		0.437501,
		0.472143,
		0.506853,
		0.541461,
		0.575799,
		0.609698,
		0.642991,
		0.675515,
		0.707107,
		0.737611,
		0.766878,
		0.794763,
		0.821134,
		0.845868,
		0.868855,
		0.890001,
		0.909232,
		0.926493,
		0.941754,
		0.95501,
		0.966289,
		0.97565,
		0.983187,
		0.989031,
		0.99335,
		0.996348,
		0.998255,
		0.999326,
		0.999819,
		0.999977,
		1,
		0.999999,
		0.999959,
		0.999729,
		0.999046,
		0.997571,
		0.994912,
		0.99065,
		0.984361,
		0.975631,
		0.964076,
		0.949349,
		0.931149,
		0.909232,
		0.883417,
		0.853585,
		0.819691,
		0.781763,
		0.739908,
		0.694316,
		0.645261,
		0.593109,
		0.538323,
		0.481465,
		0.423201,
		0.364313,
		0.3057,
		0.248386,
		0.193529,
		0.142431,
		0.096544,
		0.057479,
		0.027023,
		0.007142,
		0.001497,
		0.00592,
		0.013159,
		0.023103,
		0.035637,
		0.050639,
		0.067986,
		0.087549,
		0.109195,
		0.132788,
		0.158187,
		0.185248,
		0.213823,
		0.243759,
		0.2749,
		0.307089,
		0.340162,
		0.373956,
		0.408303,
		0.443032,
		0.477974,
		0.512956,
		0.547804,
		0.582346,
		0.61641,
		0.649825,
		0.682423,
		0.71404,
		0.744517,
		0.773702,
		0.80145,
		0.827626,
		0.85211,
		0.874792,
		0.895582,
		0.914407,
		0.931219,
		0.945995,
		0.958739,
		0.969489,
		0.978317,
		0.985332,
		0.990679,
		0.994545,
		0.997147,
		0.998733,
		0.999565,
		0.999907,
		0.999994,
		1,
		0.999993,
		0.999899,
		0.999512,
		0.998528,
		0.996571,
		0.993226,
		0.988059,
		0.980638,
		0.970552,
		0.957423,
		0.940917,
		0.920755,
		0.896717,
		0.868649,
		0.836467,
		0.800163,
		0.759807,
		0.715554,
		0.667644,
		0.61641,
		0.56228,
		0.505785,
		0.447562,
		0.38836,
		0.32905,
		0.270627,
		0.214224,
		0.161115,
		0.112726,
		0.070648,
		0.036642,
		0.012657,
		0.000836,
		0.001521,
		0.006014,
		0.013368,
		0.023468,
		0.036195,
		0.051425,
		0.069032,
		0.088884,
		0.110843,
		0.134771,
		0.160523,
		0.18795,
		0.2169,
		0.247217,
		0.278741,
		0.31131,
		0.344756,
		0.378911,
		0.413603,
		0.448658,
		0.483901,
		0.519153,
		0.55424,
		0.588982,
		0.623205,
		0.656733,
		0.689396,
		0.721027,
		0.751464,
		0.780552,
		0.808146,
		0.834112,
		0.858326,
		0.880684,
		0.901097,
		0.919497,
		0.935842,
		0.950115,
		0.962333,
		0.972543,
		0.980831,
		0.987321,
		0.992177,
		0.995601,
		0.997826,
		0.999115,
		0.999739,
		0.999959,
		0.999999,
		1,
		0.999974,
		0.999788,
		0.999185,
		0.997819,
		0.995285,
		0.991144,
		0.984953,
		0.976276,
		0.964706,
		0.949876,
		0.931471,
		0.909232,
		0.882967,
		0.852555,
		0.817948,
		0.77918,
		0.736367,
		0.689715,
		0.63952,
		0.586177,
		0.530184,
		0.472143,
		0.412775,
		0.352917,
		0.293538,
		0.23574,
		0.18077,
		0.130029,
		0.085081,
		0.047667,
		0.01971,
		0.003336,
		0.001546,
		0.006111,
		0.013582,
		0.023841,
		0.036766,
		0.05223,
		0.070103,
		0.090249,
		0.112529,
		0.136799,
		0.16291,
		0.19071,
		0.220042,
		0.250747,
		0.282661,
		0.315615,
		0.349439,
		0.383959,
		0.418999,
		0.454381,
		0.489923,
		0.525446,
		0.560768,
		0.595706,
		0.630081,
		0.663715,
		0.696434,
		0.728066,
		0.758449,
		0.787425,
		0.814849,
		0.840584,
		0.864511,
		0.886525,
		0.90654,
		0.924495,
		0.940353,
		0.954108,
		0.965784,
		0.975444,
		0.983187,
		0.989154,
		0.993526,
		0.996521,
		0.998392,
		0.999412,
		0.999857,
		0.999986,
		1,
		0.999997,
		0.999929,
		0.999603,
		0.998713,
		0.996876,
		0.993658,
		0.988607,
		0.981267,
		0.971204,
		0.95802,
		0.941362,
		0.920935,
		0.896506,
		0.867913,
		0.835071,
		0.797974,
		0.756699,
		0.711414,
		0.66238,
		0.609954,
		0.5546,
		0.496887,
		0.437501,
		0.37725,
		0.317068,
		0.258031,
		0.201355,
		0.148414,
		0.100747,
		0.060068,
		0.02828,
		0.007484,
		0.001571,
		0.006211,
		0.013802,
		0.024223,
		0.037351,
		0.053054,
		0.071199,
		0.091646,
		0.114253,
		0.138872,
		0.165349,
		0.193529,
		0.223252,
		0.254351,
		0.28666,
		0.320005,
		0.354211,
		0.3891,
		0.424491,
		0.460201,
		0.496045,
		0.531836,
		0.567389,
		0.602518,
		0.63704,
		0.670771,
		0.703534,
		0.735156,
		0.76547,
		0.794318,
		0.821553,
		0.84704,
		0.870658,
		0.892307,
		0.911904,
		0.929393,
		0.944746,
		0.957965,
		0.969087,
		0.978187,
		0.985382,
		0.990828,
		0.994726,
		0.997312,
		0.998853,
		0.999632,
		0.999932,
		0.999997,
		1,
		0.999985,
		0.999839,
		0.999313,
		0.998056,
		0.995649,
		0.991635,
		0.985548,
		0.976929,
		0.965348,
		0.950418,
		0.931803,
		0.909232,
		0.882503,
		0.851487,
		0.816138,
		0.776495,
		0.732684,
		0.684929,
		0.63355,
		0.578973,
		0.521733,
		0.462481,
		0.40199,
		0.341166,
		0.281048,
		0.222825,
		0.167843,
		0.117611,
		0.073818,
		0.038342,
		0.013263,
		0.000878,
		0.001597,
		0.006313,
		0.014026,
		0.024615,
		0.03795,
		0.053897,
		0.07232,
		0.093076,
		0.116017,
		0.140991,
		0.167843,
		0.196411,
		0.22653,
		0.258031,
		0.290741,
		0.324482,
		0.359076,
		0.394338,
		0.430082,
		0.466121,
		0.502265,
		0.538323,
		0.574105,
		0.60942,
		0.64408,
		0.677898,
		0.710695,
		0.742293,
		0.772523,
		0.801227,
		0.828254,
		0.853472,
		0.876762,
		0.898023,
		0.917181,
		0.934185,
		0.949014,
		0.961681,
		0.972236,
		0.98077,
		0.987414,
		0.992346,
		0.995782,
		0.997978,
		0.999216,
		0.999788,
		0.999973,
		1,
		0.999999,
		0.999952,
		0.999682,
		0.998886,
		0.997169,
		0.994084,
		0.989154,
		0.981901,
		0.971868,
		0.958631,
		0.94182,
		0.92112,
		0.896287,
		0.867151,
		0.833622,
		0.795699,
		0.753466,
		0.707107,
		0.656903,
		0.603241,
		0.546619,
		0.487653,
		0.427082,
		0.365774,
		0.304739,
		0.245133,
		0.188271,
		0.135635,
		0.088884,
		0.049871,
		0.020651,
		0.003501,
		0.001624,
		0.006417,
		0.014256,
		0.025016,
		0.038563,
		0.054761,
		0.073468,
		0.094539,
		0.117821,
		0.14316,
		0.170393,
		0.199356,
		0.229879,
		0.261788,
		0.294906,
		0.32905,
		0.364035,
		0.399673,
		0.435774,
		0.472143,
		0.508587,
		0.54491,
		0.580916,
		0.61641,
		0.651201,
		0.685097,
		0.717915,
		0.749475,
		0.779606,
		0.808146,
		0.834948,
		0.859876,
		0.882814,
		0.903668,
		0.922365,
		0.938862,
		0.953149,
		0.965249,
		0.975226,
		0.983187,
		0.989282,
		0.993706,
		0.996697,
		0.998528,
		0.999493,
		0.999891,
		0.999993,
		1,
		0.999992,
		0.999882,
		0.999429,
		0.99828,
		0.996003,
		0.992122,
		0.986146,
		0.977591,
		0.966004,
		0.950973,
		0.932145,
		0.909232,
		0.882021,
		0.850378,
		0.814257,
		0.773702,
		0.728851,
		0.679948,
		0.627338,
		0.571482,
		0.512956,
		0.452462,
		0.390835,
		0.32905,
		0.268228,
		0.209654,
		0.154777,
		0.105229,
		0.062836,
		0.029626,
		0.007852,
		0.001651,
		0.006524,
		0.014492,
		0.025427,
		0.039191,
		0.055645,
		0.074644,
		0.096037,
		0.119668,
		0.145377,
		0.173,
		0.202366,
		0.2333,
		0.265625,
		0.299157,
		0.333709,
		0.369091,
		0.405109,
		0.441568,
		0.478268,
		0.515011,
		0.551596,
		0.587822,
		0.62349,
		0.658402,
		0.692366,
		0.725192,
		0.756699,
		0.786713,
		0.815072,
		0.841627,
		0.866244,
		0.88881,
		0.909232,
		0.927446,
		0.943417,
		0.957143,
		0.968661,
		0.978051,
		0.985435,
		0.990983,
		0.994912,
		0.997478,
		0.99897,
		0.999695,
		0.999952,
		0.999999,
		1,
		0.99997,
		0.999752,
		0.999046,
		0.997451,
		0.994502,
		0.9897,
		0.982541,
		0.972543,
		0.959257,
		0.94229,
		0.921311,
		0.89606,
		0.866359,
		0.832117,
		0.793332,
		0.750101,
		0.702622,
		0.651201,
		0.596255,
		0.538323,
		0.478069,
		0.416289,
		0.353922,
		0.292057,
		0.231941,
		0.174994,
		0.122817,
		0.077206,
		0.040162,
		0.013913,
		0.000922,
		0.001679,
		0.006633,
		0.014734,
		0.025848,
		0.039835,
		0.056551,
		0.075848,
		0.09757,
		0.121558,
		0.147646,
		0.175666,
		0.205443,
		0.236797,
		0.269544,
		0.303496,
		0.338462,
		0.374245,
		0.410647,
		0.447466,
		0.484498,
		0.521539,
		0.558383,
		0.594824,
		0.630658,
		0.665684,
		0.699703,
		0.732524,
		0.763962,
		0.793842,
		0.822,
		0.848287,
		0.872571,
		0.89474,
		0.914709,
		0.932418,
		0.947842,
		0.960989,
		0.971912,
		0.980706,
		0.987511,
		0.99252,
		0.995967,
		0.99813,
		0.999313,
		0.999833,
		0.999984,
		1,
		0.999997,
		0.999917,
		0.999533,
		0.998492,
		0.996348,
		0.992605,
		0.986746,
		0.978262,
		0.966673,
		0.951543,
		0.932498,
		0.909232,
		0.881522,
		0.849227,
		0.812301,
		0.770794,
		0.72486,
		0.67476,
		0.620871,
		0.563688,
		0.503836,
		0.442072,
		0.379296,
		0.316562,
		0.255081,
		0.19624,
		0.141606,
		0.092945,
		0.05223,
		0.021661,
		0.003677,
		0.001708,
		0.006746,
		0.014982,
		0.026279,
		0.040494,
		0.057479,
		0.077081,
		0.09914,
		0.123492,
		0.149968,
		0.178394,
		0.208589,
		0.24037,
		0.273547,
		0.307926,
		0.343311,
		0.379501,
		0.416289,
		0.453471,
		0.490835,
		0.528173,
		0.565272,
		0.601923,
		0.637915,
		0.673044,
		0.707107,
		0.739908,
		0.771261,
		0.800988,
		0.828924,
		0.854921,
		0.878848,
		0.900598,
		0.92009,
		0.937272,
		0.952128,
		0.964681,
		0.974996,
		0.983187,
		0.989415,
		0.993892,
		0.996876,
		0.998662,
		0.99957,
		0.99992,
		0.999997,
		1,
		0.999982,
		0.999811,
		0.999193,
		0.997721,
		0.994912,
		0.990244,
		0.983187,
		0.973229,
		0.959898,
		0.942774,
		0.921508,
		0.895825,
		0.865538,
		0.830552,
		0.790869,
		0.746597,
		0.69795,
		0.645261,
		0.588982,
		0.529695,
		0.468117,
		0.405109,
		0.341684,
		0.279019,
		0.218463,
		0.161551,
		0.110014,
		0.065798,
		0.031071,
		0.008248,
		0.001737,
		0.006861,
		0.015236,
		0.026722,
		0.04117,
		0.05843,
		0.078344,
		0.100747,
		0.125472,
		0.152345,
		0.181184,
		0.211807,
		0.244022,
		0.277636,
		0.31245,
		0.34826,
		0.384859,
		0.422038,
		0.459584,
		0.497281,
		0.534913,
		0.572264,
		0.609118,
		0.645261,
		0.680482,
		0.714575,
		0.747342,
		0.778592,
		0.808146,
		0.835839,
		0.861523,
		0.88507,
		0.906376,
		0.925367,
		0.941999,
		0.956268,
		0.968209,
		0.977906,
		0.98549,
		0.991144,
		0.995103,
		0.997644,
		0.999083,
		0.999752,
		0.999969,
		1,
		0.999999,
		0.999944,
		0.999626,
		0.99869,
		0.996681,
		0.993081,
		0.987347,
		0.978942,
		0.967356,
		0.952128,
		0.932861,
		0.909232,
		0.881004,
		0.848031,
		0.810266,
		0.767766,
		0.720701,
		0.669354,
		0.614133,
		0.555576,
		0.494356,
		0.431293,
		0.367361,
		0.303696,
		0.24161,
		0.182603,
		0.128373,
		0.08083,
		0.042115,
		0.014613,
		0.00097,
		0.001767,
		0.006979,
		0.015497,
		0.027175,
		0.041863,
		0.059405,
		0.079638,
		0.102394,
		0.1275,
		0.154777,
		0.184039,
		0.215097,
		0.247755,
		0.281814,
		0.317068,
		0.353309,
		0.390324,
		0.427896,
		0.465807,
		0.503836,
		0.541761,
		0.579359,
		0.61641,
		0.652694,
		0.687996,
		0.722107,
		0.754822,
		0.785951,
		0.815311,
		0.842738,
		0.868085,
		0.891227,
		0.912065,
		0.930531,
		0.946591,
		0.960253,
		0.971569,
		0.980638,
		0.987613,
		0.9927,
		0.996156,
		0.99828,
		0.999406,
		0.999872,
		0.999991,
		1,
		0.999991,
		0.999861,
		0.999327,
		0.997978,
		0.995313,
		0.990785,
		0.983837,
		0.973927,
		0.960553,
		0.943272,
		0.921711,
		0.895582,
		0.864686,
		0.828924,
		0.788304,
		0.742944,
		0.693079,
		0.639069,
		0.581406,
		0.520718,
		0.457781,
		0.393526,
		0.32905,
		0.265625,
		0.204713,
		0.147975,
		0.097288,
		0.054761,
		0.022747,
		0.003868,
		0.001798,
		0.007101,
		0.015765,
		0.027641,
		0.042574,
		0.060404,
		0.080964,
		0.104082,
		0.129577,
		0.157267,
		0.186961,
		0.218463,
		0.251572,
		0.286083,
		0.321785,
		0.358463,
		0.395897,
		0.433865,
		0.472143,
		0.510503,
		0.548718,
		0.586559,
		0.623799,
		0.660215,
		0.695586,
		0.729698,
		0.762345,
		0.793332,
		0.822478,
		0.849616,
		0.874601,
		0.897313,
		0.917657,
		0.935573,
		0.951039,
		0.964076,
		0.974752,
		0.983187,
		0.989554,
		0.994084,
		0.997056,
		0.998793,
		0.999641,
		0.999944,
		0.999999,
		1,
		0.999964,
		0.999707,
		0.998875,
		0.997003,
		0.993551,
		0.987949,
		0.97963,
		0.968053,
		0.952728,
		0.933236,
		0.909232,
		0.880467,
		0.846787,
		0.808146,
		0.76461,
		0.716363,
		0.663715,
		0.607109,
		0.547127,
		0.484498,
		0.42011,
		0.355015,
		0.290446,
		0.227824,
		0.168771,
		0.11513,
		0.068974,
		0.032625,
		0.008674,
		0.00183,
		0.007225,
		0.016039,
		0.028118,
		0.043303,
		0.061429,
		0.082324,
		0.10581,
		0.131705,
		0.159817,
		0.189951,
		0.221906,
		0.255475,
		0.290446,
		0.326602,
		0.363722,
		0.40158,
		0.439948,
		0.478594,
		0.517284,
		0.555785,
		0.593863,
		0.631286,
		0.667822,
		0.703249,
		0.737347,
		0.769907,
		0.800733,
		0.82964,
		0.856465,
		0.881063,
		0.903317,
		0.923141,
		0.940484,
		0.955335,
		0.967729,
		0.977753,
		0.985548,
		0.991312,
		0.995298,
		0.997811,
		0.999193,
		0.999803,
		0.999981,
		1,
		0.999996,
		0.999901,
		0.999448,
		0.998222,
		0.995704,
		0.991323,
		0.984492,
		0.974637,
		0.961225,
		0.943785,
		0.921922,
		0.895329,
		0.863799,
		0.827229,
		0.78563,
		0.739133,
		0.687996,
		0.632611,
		0.573509,
		0.511372,
		0.447041,
		0.381525,
		0.31601,
		0.251877,
		0.19071,
		0.13431,
		0.084714,
		0.044213,
		0.015366,
		0.001021,
		0.001863,
		0.007353,
		0.016321,
		0.028608,
		0.044051,
		0.06248,
		0.083718,
		0.107582,
		0.133884,
		0.162428,
		0.193012,
		0.225429,
		0.259467,
		0.294906,
		0.331523,
		0.369091,
		0.407377,
		0.446146,
		0.48516,
		0.52418,
		0.562964,
		0.601273,
		0.638868,
		0.675515,
		0.710982,
		0.74505,
		0.777504,
		0.808146,
		0.836792,
		0.863277,
		0.887461,
		0.909232,
		0.928511,
		0.945255,
		0.959469,
		0.971204,
		0.980567,
		0.98772,
		0.992887,
		0.996348,
		0.998429,
		0.999493,
		0.999905,
		0.999996,
		1,
		0.999979,
		0.999776,
		0.999046,
		0.997312,
		0.994013,
		0.988552,
		0.980326,
		0.968764,
		0.953345,
		0.933622,
		0.909232,
		0.879909,
		0.845492,
		0.805937,
		0.761317,
		0.711836,
		0.657831,
		0.599783,
		0.538323,
		0.474244,
		0.408505,
		0.342249,
		0.276811,
		0.213733,
		0.154777,
		0.101941,
		0.057479,
		0.023917,
		0.004073,
		0.001896,
		0.007484,
		0.01661,
		0.02911,
		0.044818,
		0.063557,
		0.085147,
		0.109399,
		0.136118,
		0.165103,
		0.196146,
		0.229035,
		0.263549,
		0.299464,
		0.336549,
		0.37457,
		0.413289,
		0.452462,
		0.491846,
		0.531193,
		0.570256,
		0.608789,
		0.646547,
		0.683291,
		0.718785,
		0.752804,
		0.785132,
		0.815567,
		0.843926,
		0.870045,
		0.893788,
		0.915048,
		0.933754,
		0.949876,
		0.963433,
		0.974494,
		0.983187,
		0.9897,
		0.994282,
		0.997238,
		0.998921,
		0.999707,
		0.999963,
		1,
		0.999999,
		0.999933,
		0.999556,
		0.998451,
		0.996084,
		0.991857,
		0.985151,
		0.975358,
		0.961912,
		0.944313,
		0.922139,
		0.895068,
		0.862877,
		0.825463,
		0.782841,
		0.735156,
		0.68269,
		0.625869,
		0.565272,
		0.501639,
		0.43588,
		0.369091,
		0.302559,
		0.237781,
		0.176478,
		0.120607,
		0.072383,
		0.034297,
		0.009134,
		0.001931,
		0.007619,
		0.016907,
		0.029626,
		0.045605,
		0.064663,
		0.086613,
		0.111261,
		0.138407,
		0.167843,
		0.199356,
		0.232725,
		0.267725,
		0.304124,
		0.341684,
		0.380165,
		0.419319,
		0.458899,
		0.498652,
		0.538323,
		0.577661,
		0.61641,
		0.654322,
		0.69115,
		0.726654,
		0.760605,
		0.792785,
		0.82299,
		0.851035,
		0.876762,
		0.900035,
		0.920755,
		0.938862,
		0.954338,
		0.967218,
		0.977591,
		0.98561,
		0.991486,
		0.995499,
		0.997978,
		0.999299,
		0.999848,
		0.99999,
		1,
		0.999989,
		0.999834,
		0.999202,
		0.997607,
		0.994468,
		0.989154,
		0.98103,
		0.969489,
		0.953979,
		0.934022,
		0.909232,
		0.879329,
		0.844144,
		0.803632,
		0.757879,
		0.707107,
		0.651685,
		0.592135,
		0.529145,
		0.463572,
		0.39646,
		0.32905,
		0.26279,
		0.199356,
		0.140662,
		0.088884,
		0.046472,
		0.016179,
		0.001077,
		0.001966,
		0.007758,
		0.017212,
		0.030156,
		0.046414,
		0.065798,
		0.088117,
		0.113171,
		0.140754,
		0.170651,
		0.202643,
		0.236502,
		0.271997,
		0.308888,
		0.34693,
		0.385876,
		0.425471,
		0.465459,
		0.50558,
		0.545574,
		0.58518,
		0.624138,
		0.662191,
		0.699089,
		0.734587,
		0.76845,
		0.800458,
		0.830407,
		0.858112,
		0.883417,
		0.906191,
		0.926344,
		0.943825,
		0.958631,
		0.970816,
		0.980491,
		0.987832,
		0.993081,
		0.996543,
		0.998577,
		0.999575,
		0.999933,
		0.999999,
		1,
		0.999957,
		0.999651,
		0.998665,
		0.996452,
		0.992386,
		0.985814,
		0.976091,
		0.962616,
		0.944857,
		0.922365,
		0.894796,
		0.861918,
		0.823621,
		0.779928,
		0.731,
		0.677145,
		0.618827,
		0.556677,
		0.491497,
		0.424278,
		0.356209,
		0.28869,
		0.223347,
		0.162051,
		0.106933,
		0.060404,
		0.025179,
		0.004296,
		0.002002,
		0.0079,
		0.017525,
		0.0307,
		0.047243,
		0.066963,
		0.08966,
		0.11513,
		0.14316,
		0.173528,
		0.20601,
		0.24037,
		0.276369,
		0.31376,
		0.352291,
		0.391707,
		0.431746,
		0.472143,
		0.512632,
		0.552945,
		0.592813,
		0.631971,
		0.670154,
		0.707107,
		0.742579,
		0.776333,
		0.808146,
		0.837812,
		0.865149,
		0.890001,
		0.912247,
		0.931803,
		0.948631,
		0.962746,
		0.974219,
		0.983187,
		0.989852,
		0.994485,
		0.997422,
		0.999046,
		0.999766,
		0.999977,
		1,
		0.999995,
		0.999882,
		0.999343,
		0.997888,
		0.994912,
		0.989754,
		0.981742,
		0.97023,
		0.95463,
		0.934434,
		0.909232,
		0.878726,
		0.842738,
		0.801227,
		0.754286,
		0.702164,
		0.645261,
		0.584147,
		0.51957,
		0.452462,
		0.383959,
		0.315408,
		0.248386,
		0.184715,
		0.12648,
		0.076051,
		0.036101,
		0.009632,
		0.00204,
		0.008046,
		0.017847,
		0.031259,
		0.048096,
		0.068159,
		0.091244,
		0.11714,
		0.145627,
		0.176478,
		0.20946,
		0.24433,
		0.280842,
		0.318742,
		0.357769,
		0.397661,
		0.438147,
		0.478955,
		0.519811,
		0.560439,
		0.600563,
		0.63991,
		0.67821,
		0.715201,
		0.750628,
		0.78425,
		0.815842,
		0.845197,
		0.872136,
		0.896506,
		0.918192,
		0.937122,
		0.953272,
		0.966673,
		0.97742,
		0.985674,
		0.991668,
		0.995704,
		0.998145,
		0.999399,
		0.999887,
		0.999995,
		1,
		0.999975,
		0.999732,
		0.998863,
		0.996806,
		0.992908,
		0.986479,
		0.976835,
		0.963338,
		0.945417,
		0.922598,
		0.894513,
		0.860918,
		0.821699,
		0.776885,
		0.726654,
		0.671346,
		0.611467,
		0.547701,
		0.480925,
		0.412216,
		0.342867,
		0.2744,
		0.208589,
		0.14747,
		0.093366,
		0.048908,
		0.017058,
		0.001138,
		0.002078,
		0.008197,
		0.018178,
		0.031834,
		0.048971,
		0.069387,
		0.09287,
		0.119202,
		0.148158,
		0.179502,
		0.212995,
		0.248386,
		0.285421,
		0.323837,
		0.363368,
		0.403741,
		0.444677,
		0.485897,
		0.527118,
		0.568056,
		0.608428,
		0.647954,
		0.686356,
		0.723368,
		0.758729,
		0.792195,
		0.823539,
		0.852555,
		0.879064,
		0.90292,
		0.924016,
		0.94229,
		0.957735,
		0.970403,
		0.980411,
		0.987949,
		0.993282,
		0.996742,
		0.998722,
		0.999651,
		0.999956,
		1,
		0.999998,
		0.99992,
		0.99947,
		0.998153,
		0.995346,
		0.990352,
		0.982461,
		0.970985,
		0.955299,
		0.93486,
		0.909232,
		0.878098,
		0.841272,
		0.798713,
		0.750529,
		0.696992,
		0.638541,
		0.575799,
		0.509577,
		0.440894,
		0.370984,
		0.301315,
		0.233606,
		0.169842,
		0.112297,
		0.063557,
		0.026543,
		0.004537,
		0.002117,
		0.008351,
		0.018518,
		0.032424,
		0.049871,
		0.070648,
		0.094539,
		0.121319,
		0.150754,
		0.182603,
		0.216617,
		0.25254,
		0.290107,
		0.32905,
		0.369091,
		0.409949,
		0.451339,
		0.492972,
		0.534555,
		0.575799,
		0.61641,
		0.656102,
		0.694592,
		0.731605,
		0.766878,
		0.800163,
		0.83123,
		0.859876,
		0.885923,
		0.909232,
		0.929706,
		0.947296,
		0.962012,
		0.973927,
		0.983187,
		0.990011,
		0.994696,
		0.997607,
		0.999166,
		0.999819,
		0.999988,
		1,
		0.999987,
		0.999801,
		0.999046,
		0.997147,
		0.993424,
		0.987147,
		0.977591,
		0.964076,
		0.945995,
		0.92284,
		0.89422,
		0.859876,
		0.819691,
		0.773702,
		0.722107,
		0.665277,
		0.603767,
		0.538323,
		0.469901,
		0.399673,
		0.32905,
		0.259691,
		0.193529,
		0.132788,
		0.080003,
		0.038051,
		0.010171,
		0.002158,
		0.00851,
		0.018867,
		0.033031,
		0.050795,
		0.071944,
		0.096253,
		0.123492,
		0.153419,
		0.185784,
		0.220331,
		0.256796,
		0.294906,
		0.334382,
		0.37494,
		0.416289,
		0.458135,
		0.50018,
		0.542124,
		0.583667,
		0.624508,
		0.664354,
		0.702914,
		0.739908,
		0.77507,
		0.808146,
		0.838908,
		0.867151,
		0.892703,
		0.915432,
		0.935251,
		0.952128,
		0.96609,
		0.977237,
		0.985742,
		0.991857,
		0.995914,
		0.99831,
		0.999493,
		0.99992,
		0.999998,
		1,
		0.999949,
		0.999581,
		0.998403,
		0.995768,
		0.990947,
		0.983187,
		0.971756,
		0.955987,
		0.935301,
		0.909232,
		0.877444,
		0.839741,
		0.796084,
		0.746597,
		0.691577,
		0.631507,
		0.567067,
		0.499142,
		0.428844,
		0.357518,
		0.286766,
		0.218463,
		0.154777,
		0.098193,
		0.051539,
		0.018011,
		0.001204,
		0.0022,
		0.008674,
		0.019227,
		0.033655,
		0.051745,
		0.073275,
		0.098015,
		0.125723,
		0.156153,
		0.189047,
		0.224139,
		0.261156,
		0.299819,
		0.339837,
		0.380919,
		0.422765,
		0.465069,
		0.507526,
		0.549827,
		0.591661,
		0.632722,
		0.672708,
		0.71132,
		0.748275,
		0.783299,
		0.816138,
		0.846562,
		0.874369,
		0.899393,
		0.921508,
		0.94064,
		0.956775,
		0.969962,
		0.980326,
		0.988073,
		0.99349,
		0.996943,
		0.998863,
		0.99972,
		0.999973,
		1,
		0.999994,
		0.999858,
		0.999212,
		0.997473,
		0.99393,
		0.987816,
		0.978359,
		0.964833,
		0.946591,
		0.92309,
		0.893914,
		0.858788,
		0.817592,
		0.770369,
		0.717343,
		0.65892,
		0.595706,
		0.528519,
		0.4584,
		0.38663,
		0.314747,
		0.244566,
		0.178197,
		0.118071,
		0.066963,
		0.028021,
		0.004799,
		0.002243,
		0.008842,
		0.019597,
		0.034297,
		0.052722,
		0.074644,
		0.099824,
		0.128015,
		0.15896,
		0.192394,
		0.228043,
		0.265625,
		0.30485,
		0.34542,
		0.387032,
		0.429378,
		0.472143,
		0.515011,
		0.557664,
		0.599783,
		0.641052,
		0.681162,
		0.719808,
		0.756699,
		0.791559,
		0.824131,
		0.854185,
		0.881522,
		0.90598,
		0.927446,
		0.94586,
		0.961225,
		0.973615,
		0.983187,
		0.990178,
		0.994912,
		0.997793,
		0.999281,
		0.999864,
		0.999994,
		1,
		0.99997,
		0.999678,
		0.998635,
		0.996177,
		0.991538,
		0.983919,
		0.972543,
		0.956695,
		0.935757,
		0.909232,
		0.876762,
		0.838141,
		0.793332,
		0.742476,
		0.685901,
		0.624138,
		0.557927,
		0.488241,
		0.416289,
		0.343545,
		0.271757,
		0.202976,
		0.139573,
		0.084269,
		0.040162,
		0.010757,
		0.002287,
		0.009016,
		0.019978,
		0.034957,
		0.053727,
		0.076051,
		0.101684,
		0.130369,
		0.161842,
		0.19583,
		0.232048,
		0.270205,
		0.310002,
		0.351133,
		0.393283,
		0.436133,
		0.47936,
		0.522638,
		0.565638,
		0.608032,
		0.649498,
		0.689715,
		0.728373,
		0.765177,
		0.799844,
		0.832117,
		0.861767,
		0.888597,
		0.912454,
		0.933236,
		0.950898,
		0.965467,
		0.977043,
		0.985814,
		0.992055,
		0.996128,
		0.998474,
		0.999581,
		0.999946,
		1,
		0.999998,
		0.999903,
		0.999361,
		0.997782,
		0.994427,
		0.988485,
		0.979138,
		0.965609,
		0.947206,
		0.92335,
		0.893596,
		0.857652,
		0.815394,
		0.766878,
		0.712348,
		0.652255,
		0.587261,
		0.518262,
		0.446396,
		0.373065,
		0.299949,
		0.229035,
		0.162631,
		0.103402,
		0.054388,
		0.019046,
		0.001275,
		0.002333,
		0.009194,
		0.02037,
		0.035637,
		0.054761,
		0.077498,
		0.103595,
		0.132788,
		0.164802,
		0.199356,
		0.236155,
		0.2749,
		0.315281,
		0.35698,
		0.399673,
		0.443032,
		0.486723,
		0.530408,
		0.573749,
		0.61641,
		0.658057,
		0.698364,
		0.737013,
		0.773702,
		0.808146,
		0.840087,
		0.869296,
		0.895582,
		0.9188,
		0.938862,
		0.955742,
		0.969489,
		0.980236,
		0.988203,
		0.993706,
		0.997147,
		0.999001,
		0.999782,
		0.999985,
		1,
		0.999984,
		0.99976,
		0.99885,
		0.996571,
		0.992122,
		0.984657,
		0.973345,
		0.957423,
		0.936229,
		0.909232,
		0.87605,
		0.836467,
		0.790449,
		0.738155,
		0.679948,
		0.61641,
		0.548355,
		0.476846,
		0.403207,
		0.32905,
		0.25629,
		0.187172,
		0.124297,
		0.070648,
		0.029626,
		0.005084,
		0.00238,
		0.009378,
		0.020774,
		0.036336,
		0.055825,
		0.078987,
		0.105561,
		0.135274,
		0.167843,
		0.202976,
		0.24037,
		0.279714,
		0.320688,
		0.362964,
		0.406208,
		0.45008,
		0.494234,
		0.538323,
		0.582,
		0.624916,
		0.66673,
		0.707107,
		0.745722,
		0.782268,
		0.816458,
		0.848031,
		0.876762,
		0.902465,
		0.925006,
		0.944313,
		0.960379,
		0.973282,
		0.983187,
		0.990352,
		0.995135,
		0.997978,
		0.999391,
		0.999903,
		0.999998,
		1,
		0.999938,
		0.999493,
		0.998073,
		0.994912,
		0.989154,
		0.979927,
		0.966404,
		0.947842,
		0.92362,
		0.893264,
		0.856465,
		0.813093,
		0.763216,
		0.707107,
		0.645261,
		0.578407,
		0.507526,
		0.433865,
		0.358959,
		0.28465,
		0.213114,
		0.146884,
		0.088884,
		0.042454,
		0.011395,
		0.002428,
		0.009567,
		0.021189,
		0.037057,
		0.05692,
		0.080519,
		0.107582,
		0.137829,
		0.170967,
		0.206693,
		0.244695,
		0.28465,
		0.326228,
		0.369091,
		0.412891,
		0.457278,
		0.501896,
		0.546387,
		0.59039,
		0.63355,
		0.675515,
		0.71594,
		0.754496,
		0.790869,
		0.82477,
		0.855939,
		0.884152,
		0.909232,
		0.931058,
		0.949573,
		0.964797,
		0.976835,
		0.98589,
		0.992261,
		0.996348,
		0.998635,
		0.999662,
		0.999967,
		1,
		0.999993,
		0.999827,
		0.999046,
		0.99695,
		0.9927,
		0.985399,
		0.974162,
		0.958172,
		0.936718,
		0.909232,
		0.875306,
		0.834714,
		0.787425,
		0.733619,
		0.673697,
		0.608301,
		0.538323,
		0.464929,
		0.389573,
		0.314019,
		0.24037,
		0.171089,
		0.109032,
		0.057479,
		0.020173,
		0.001354,
		0.002478,
		0.009762,
		0.021618,
		0.037798,
		0.058047,
		0.082095,
		0.109662,
		0.140457,
		0.174178,
		0.210511,
		0.249134,
		0.289713,
		0.331906,
		0.375363,
		0.419726,
		0.464632,
		0.509713,
		0.5546,
		0.598922,
		0.642312,
		0.684409,
		0.72486,
		0.763329,
		0.799498,
		0.833075,
		0.863799,
		0.891454,
		0.91587,
		0.936941,
		0.95463,
		0.968983,
		0.98014,
		0.98834,
		0.99393,
		0.997353,
		0.999134,
		0.999836,
		0.999993,
		1,
		0.999963,
		0.999608,
		0.998347,
		0.995385,
		0.989821,
		0.980727,
		0.967218,
		0.948498,
		0.9239,
		0.892918,
		0.855222,
		0.81068,
		0.759371,
		0.701601,
		0.637915,
		0.569117,
		0.496282,
		0.420779,
		0.344293,
		0.268846,
		0.196828,
		0.131022,
		0.074644,
		0.031373,
		0.005396,
		0.00253,
		0.009964,
		0.022059,
		0.038563,
		0.059208,
		0.083718,
		0.111802,
		0.14316,
		0.177478,
		0.214433,
		0.25369,
		0.294906,
		0.337724,
		0.381784,
		0.426716,
		0.472143,
		0.517687,
		0.562964,
		0.607595,
		0.651201,
		0.69341,
		0.733863,
		0.772216,
		0.808146,
		0.84136,
		0.8716,
		0.898654,
		0.922365,
		0.942641,
		0.959469,
		0.972926,
		0.983187,
		0.990536,
		0.995364,
		0.998163,
		0.999493,
		0.999935,
		1,
		0.999998,
		0.999882,
		0.999223,
		0.997312,
		0.99327,
		0.986146,
		0.974996,
		0.958943,
		0.937225,
		0.909232,
		0.874529,
		0.832877,
		0.78425,
		0.728851,
		0.667127,
		0.599783,
		0.527802,
		0.452462,
		0.375363,
		0.298442,
		0.22401,
		0.154777,
		0.093885,
		0.044948,
		0.012092,
		0.002583,
		0.010171,
		0.022514,
		0.03935,
		0.060404,
		0.085389,
		0.114005,
		0.14594,
		0.180871,
		0.218463,
		0.258369,
		0.300233,
		0.343689,
		0.38836,
		0.433865,
		0.479816,
		0.52582,
		0.571482,
		0.61641,
		0.660215,
		0.702516,
		0.742944,
		0.781149,
		0.816804,
		0.849616,
		0.879329,
		0.905738,
		0.9287,
		0.948142,
		0.964076,
		0.976614,
		0.98597,
		0.992476,
		0.996571,
		0.998793,
		0.999735,
		0.999982,
		1,
		0.999981,
		0.999707,
		0.9986,
		0.995843,
		0.990485,
		0.981538,
		0.968053,
		0.949176,
		0.924192,
		0.892557,
		0.85392,
		0.808146,
		0.755329,
		0.69581,
		0.630193,
		0.559361,
		0.484498,
		0.407111,
		0.32905,
		0.25254,
		0.180211,
		0.11513,
		0.06084,
		0.021402,
		0.00144,
		0.002638,
		0.010385,
		0.022984,
		0.040162,
		0.061637,
		0.08711,
		0.116272,
		0.148801,
		0.18436,
		0.222604,
		0.263174,
		0.3057,
		0.349803,
		0.395094,
		0.441178,
		0.487653,
		0.534114,
		0.580154,
		0.625367,
		0.669354,
		0.711723,
		0.752097,
		0.79012,
		0.825463,
		0.85783,
		0.886971,
		0.912691,
		0.93486,
		0.953429,
		0.968439,
		0.980037,
		0.988485,
		0.994162,
		0.997561,
		0.999261,
		0.999882,
		0.999997,
		0.999999,
		0.999924,
		0.999381,
		0.997655,
		0.993829,
		0.986896,
		0.975845,
		0.959736,
		0.937751,
		0.909232,
		0.873716,
		0.830949,
		0.780913,
		0.723836,
		0.660215,
		0.590828,
		0.516759,
		0.439414,
		0.360554,
		0.282312,
		0.207232,
		0.138302,
		0.078987,
		0.033279,
		0.005737,
		0.002694,
		0.010606,
		0.023468,
		0.041,
		0.062907,
		0.088884,
		0.118608,
		0.151745,
		0.18795,
		0.226862,
		0.268109,
		0.31131,
		0.356071,
		0.40199,
		0.448658,
		0.495659,
		0.542574,
		0.588982,
		0.634466,
		0.678615,
		0.721027,
		0.761317,
		0.799123,
		0.834112,
		0.86599,
		0.894513,
		0.919497,
		0.94083,
		0.958486,
		0.972543,
		0.983187,
		0.990728,
		0.995601,
		0.998347,
		0.999588,
		0.999959,
		1,
		0.999991,
		0.999788,
		0.998834,
		0.996286,
		0.991144,
		0.982358,
		0.968908,
		0.949876,
		0.924495,
		0.892179,
		0.852555,
		0.805484,
		0.751077,
		0.689715,
		0.622067,
		0.549108,
		0.472143,
		0.392831,
		0.313213,
		0.23574,
		0.163312,
		0.099316,
		0.047667,
		0.012854,
		0.002753,
		0.010834,
		0.023968,
		0.041863,
		0.064217,
		0.090711,
		0.121013,
		0.154777,
		0.191643,
		0.231239,
		0.273179,
		0.317068,
		0.362499,
		0.409054,
		0.45631,
		0.503836,
		0.5512,
		0.597967,
		0.643707,
		0.687996,
		0.730424,
		0.770596,
		0.808146,
		0.842738,
		0.874082,
		0.901939,
		0.926139,
		0.946591,
		0.963299,
		0.976376,
		0.986055,
		0.9927,
		0.996799,
		0.998946,
		0.999799,
		0.999991,
		1,
		0.999955,
		0.999519,
		0.997978,
		0.994377,
		0.987648,
		0.976711,
		0.960553,
		0.938296,
		0.909232,
		0.872864,
		0.828924,
		0.777401,
		0.718555,
		0.652936,
		0.581406,
		0.505161,
		0.425753,
		0.345123,
		0.265625,
		0.190067,
		0.121749,
		0.064503,
		0.022747,
		0.001534,
		0.002813,
		0.011069,
		0.024483,
		0.042755,
		0.065569,
		0.092596,
		0.123492,
		0.157899,
		0.195443,
		0.23574,
		0.278389,
		0.32298,
		0.369091,
		0.416289,
		0.464137,
		0.512188,
		0.559995,
		0.607109,
		0.653088,
		0.697495,
		0.739908,
		0.779928,
		0.817181,
		0.851331,
		0.882091,
		0.909232,
		0.9326,
		0.952128,
		0.967852,
		0.979927,
		0.988638,
		0.994403,
		0.997769,
		0.999381,
		0.99992,
		0.999999,
		0.999997,
		0.999854,
		0.999046,
		0.996711,
		0.991798,
		0.983187,
		0.969784,
		0.950601,
		0.924811,
		0.891784,
		0.851122,
		0.802682,
		0.746597,
		0.683291,
		0.613507,
		0.538323,
		0.459182,
		0.377913,
		0.29677,
		0.218463,
		0.146196,
		0.083718,
		0.035363,
		0.006111,
		0.002875,
		0.011312,
		0.025016,
		0.043674,
		0.066963,
		0.094539,
		0.126047,
		0.161115,
		0.199356,
		0.24037,
		0.283743,
		0.32905,
		0.375851,
		0.423701,
		0.472143,
		0.520718,
		0.568961,
		0.61641,
		0.662608,
		0.707107,
		0.749475,
		0.789303,
		0.826216,
		0.859876,
		0.890001,
		0.916375,
		0.938862,
		0.957423,
		0.97213,
		0.983187,
		0.990931,
		0.995843,
		0.998528,
		0.999675,
		0.999977,
		1,
		0.999976,
		0.999638,
		0.99828,
		0.994912,
		0.988401,
		0.977591,
		0.961395,
		0.938862,
		0.909232,
		0.87197,
		0.826794,
		0.773702,
		0.712986,
		0.645261,
		0.571482,
		0.492972,
		0.411446,
		0.32905,
		0.248386,
		0.172561,
		0.105229,
		0.050639,
		0.013691,
		0.00294,
		0.011564,
		0.025566,
		0.044624,
		0.068402,
		0.096544,
		0.128681,
		0.164428,
		0.203384,
		0.245133,
		0.289247,
		0.335282,
		0.382786,
		0.431293,
		0.480333,
		0.529428,
		0.5781,
		0.625869,
		0.672265,
		0.716828,
		0.759116,
		0.798713,
		0.835238,
		0.868358,
		0.897796,
		0.92335,
		0.944907,
		0.962458,
		0.97612,
		0.986146,
		0.992935,
		0.99703,
		0.999094,
		0.999854,
		0.999997,
		0.999999,
		0.999906,
		0.999236,
		0.997117,
		0.992444,
		0.984024,
		0.970681,
		0.951351,
		0.92514,
		0.89137,
		0.849616,
		0.799732,
		0.741872,
		0.676513,
		0.604481,
		0.526971,
		0.445578,
		0.362328,
		0.279714,
		0.200736,
		0.128948,
		0.068505,
		0.024222,
		0.001637,
		0.003007,
		0.011823,
		0.026134,
		0.045605,
		0.069887,
		0.098612,
		0.131398,
		0.167843,
		0.207533,
		0.250035,
		0.294906,
		0.341684,
		0.3899,
		0.439072,
		0.488711,
		0.538323,
		0.587413,
		0.635486,
		0.682058,
		0.726654,
		0.768825,
		0.808146,
		0.844236,
		0.876762,
		0.905457,
		0.930138,
		0.950715,
		0.967218,
		0.97981,
		0.9888,
		0.994653,
		0.997978,
		0.999493,
		0.99995,
		1,
		0.999989,
		0.999738,
		0.99856,
		0.995431,
		0.989154,
		0.978488,
		0.962262,
		0.939451,
		0.909232,
		0.871032,
		0.824552,
		0.769799,
		0.707107,
		0.63716,
		0.561021,
		0.480153,
		0.39646,
		0.312315,
		0.230604,
		0.154774,
		0.088879,
		0.037644,
		0.006522,
		0.003075,
		0.012092,
		0.026722,
		0.046619,
		0.071421,
		0.100747,
		0.1342,
		0.171364,
		0.211807,
		0.255081,
		0.300724,
		0.34826,
		0.397198,
		0.447041,
		0.497281,
		0.547405,
		0.596902,
		0.645261,
		0.691982,
		0.736579,
		0.778592,
		0.817592,
		0.853193,
		0.88507,
		0.912966,
		0.936718,
		0.956268,
		0.971686,
		0.983187,
		0.991144,
		0.996092,
		0.998706,
		0.999752,
		0.999989,
		1,
		0.999944,
		0.999403,
		0.997501,
		0.993081,
		0.984868,
		0.971601,
		0.952128,
		0.925483,
		0.890936,
		0.848031,
		0.796619,
		0.736881,
		0.669354,
		0.594955,
		0.515011,
		0.431293,
		0.346046,
		0.262037,
		0.182595,
		0.111669,
		0.053884,
		0.014606,
		0.003147,
		0.01237,
		0.027329,
		0.047667,
		0.073006,
		0.102952,
		0.137092,
		0.174994,
		0.21621,
		0.260276,
		0.306709,
		0.355015,
		0.404687,
		0.455206,
		0.506046,
		0.556677,
		0.606567,
		0.655192,
		0.702034,
		0.746597,
		0.788409,
		0.827036,
		0.862095,
		0.893264,
		0.920303,
		0.943071,
		0.961546,
		0.975845,
		0.986243,
		0.993181,
		0.997265,
		0.999236,
		0.9999,
		0.999999,
		0.999996,
		0.999819,
		0.998815,
		0.995933,
		0.989904,
		0.9794,
		0.963156,
		0.940063,
		0.909232,
		0.870045,
		0.822187,
		0.765677,
		0.700893,
		0.628601,
		0.549981,
		0.466662,
		0.380753,
		0.294891,
		0.212284,
		0.136765,
		0.072856,
		0.025825,
		0.001749,
		0.003221,
		0.012657,
		0.027957,
		0.04875,
		0.074644,
		0.105229,
		0.140077,
		0.178739,
		0.22075,
		0.265625,
		0.312866,
		0.361957,
		0.412372,
		0.463572,
		0.515011,
		0.56614,
		0.61641,
		0.665277,
		0.71221,
		0.756699,
		0.798264,
		0.836467,
		0.870925,
		0.901325,
		0.927446,
		0.949176,
		0.966531,
		0.979683,
		0.988971,
		0.994912,
		0.998186,
		0.999596,
		0.999972,
		1,
		0.99997,
		0.999548,
		0.997863,
		0.993706,
		0.985719,
		0.972543,
		0.952932,
		0.925841,
		0.89048,
		0.846361,
		0.793332,
		0.731605,
		0.661781,
		0.584885,
		0.502393,
		0.416269,
		0.329012,
		0.243699,
		0.164044,
		0.094456,
		0.040101,
		0.006962,
		0.003297,
		0.012955,
		0.028608,
		0.049871,
		0.076337,
		0.107582,
		0.14316,
		0.182603,
		0.225429,
		0.271135,
		0.3192,
		0.369091,
		0.420258,
		0.472143,
		0.52418,
		0.575799,
		0.62643,
		0.675515,
		0.722505,
		0.766878,
		0.808146,
		0.845868,
		0.879663,
		0.909232,
		0.934374,
		0.95501,
		0.971204,
		0.983187,
		0.991369,
		0.996348,
		0.998879,
		0.999819,
		0.999996,
		0.999999,
		0.999882,
		0.999046,
		0.996415,
		0.99065,
		0.980326,
		0.964076,
		0.940699,
		0.909232,
		0.869007,
		0.819691,
		0.761316,
		0.694312,
		0.619534,
		0.538297,
		0.452411,
		0.364225,
		0.276679,
		0.19336,
		0.118558,
		0.057334,
		0.015567,
		0.003376,
		0.013263,
		0.029281,
		0.05103,
		0.078089,
		0.110014,
		0.146343,
		0.186592,
		0.230255,
		0.276811,
		0.325719,
		0.376422,
		0.428351,
		0.480925,
		0.533556,
		0.585653,
		0.636628,
		0.685901,
		0.732912,
		0.777124,
		0.818041,
		0.855222,
		0.888291,
		0.916963,
		0.941064,
		0.960553,
		0.975548,
		0.986346,
		0.993437,
		0.997501,
		0.999369,
		0.999937,
		1,
		0.999986,
		0.999671,
		0.9982,
		0.994317,
		0.986574,
		0.973507,
		0.953766,
		0.926215,
		0.890001,
		0.844598,
		0.789852,
		0.726004,
		0.653732,
		0.574175,
		0.488978,
		0.400334,
		0.311037,
		0.224539,
		0.14501,
		0.077397,
		0.027471,
		0.001862,
		0.003458,
		0.013582,
		0.029978,
		0.05223,
		0.079901,
		0.112529,
		0.149633,
		0.19071,
		0.235233,
		0.282661,
		0.332429,
		0.383959,
		0.436658,
		0.489923,
		0.543144,
		0.595706,
		0.647002,
		0.696434,
		0.743424,
		0.787425,
		0.827935,
		0.864511,
		0.896786,
		0.924495,
		0.947493,
		0.965784,
		0.979547,
		0.989154,
		0.99518,
		0.998392,
		0.999689,
		0.999986,
		1,
		0.999929,
		0.999251,
		0.996876,
		0.99139,
		0.981267,
		0.965026,
		0.941362,
		0.909231,
		0.867908,
		0.817037,
		0.756663,
		0.687268,
		0.609807,
		0.525747,
		0.437117,
		0.346557,
		0.257379,
		0.173627,
		0.100129,
		0.042546,
		0.007387,
		0.003544,
		0.013913,
		0.0307,
		0.053473,
		0.081776,
		0.11513,
		0.153034,
		0.194962,
		0.24037,
		0.28869,
		0.339336,
		0.391707,
		0.445185,
		0.499142,
		0.552945,
		0.605958,
		0.657551,
		0.707107,
		0.754033,
		0.797771,
		0.837812,
		0.873716,
		0.905127,
		0.931803,
		0.953636,
		0.970681,
		0.983187,
		0.991607,
		0.996609,
		0.999046,
		0.999875,
		0.999999,
		0.999995,
		0.999771,
		0.998511,
		0.994912,
		0.987433,
		0.974494,
		0.954629,
		0.926602,
		0.889481,
		0.842698,
		0.786085,
		0.719918,
		0.644953,
		0.56247,
		0.474322,
		0.382989,
		0.291638,
		0.204182,
		0.125331,
		0.060626,
		0.01645,
		0.003632,
		0.014256,
		0.031449,
		0.054761,
		0.083718,
		0.117821,
		0.156549,
		0.199356,
		0.245671,
		0.294906,
		0.346449,
		0.399673,
		0.453937,
		0.508587,
		0.562964,
		0.61641,
		0.668272,
		0.717915,
		0.764729,
		0.808146,
		0.847654,
		0.882814,
		0.913289,
		0.938862,
		0.959469,
		0.975226,
		0.986456,
		0.993706,
		0.99774,
		0.999493,
		0.999964,
		1,
		0.999962,
		0.999429,
		0.997312,
		0.992122,
		0.982219,
		0.965998,
		0.942036,
		0.909189,
		0.86666,
		0.81406,
		0.751442,
		0.679351,
		0.598868,
		0.51166,
		0.420042,
		0.327031,
		0.236406,
		0.152754,
		0.081496,
		0.028884,
		0.001953,
		0.003724,
		0.014613,
		0.032226,
		0.056095,
		0.085729,
		0.120607,
		0.160186,
		0.203896,
		0.251144,
		0.301315,
		0.353774,
		0.407865,
		0.462921,
		0.518262,
		0.573204,
		0.627063,
		0.679164,
		0.728851,
		0.775502,
		0.818536,
		0.85744,
		0.891784,
		0.921246,
		0.945646,
		0.964969,
		0.9794,
		0.989348,
		0.995458,
		0.998595,
		0.999771,
		0.999995,
		0.999999,
		0.99985,
		0.998793,
		0.995487,
		0.988288,
		0.975486,
		0.955481,
		0.926912,
		0.888751,
		0.840368,
		0.781582,
		0.712711,
		0.634625,
		0.548801,
		0.457378,
		0.36322,
		0.269972,
		0.182095,
		0.104887,
		0.044463,
		0.007692,
		0.003819,
		0.014982,
		0.033031,
		0.057479,
		0.087813,
		0.123492,
		0.163949,
		0.208589,
		0.256796,
		0.307926,
		0.361318,
		0.416289,
		0.472143,
		0.528173,
		0.583667,
		0.637915,
		0.690221,
		0.739908,
		0.786339,
		0.828924,
		0.867151,
		0.900598,
		0.928972,
		0.952128,
		0.970112,
		0.983187,
		0.991857,
		0.996876,
		0.999205,
		0.99992,
		1,
		0.999982,
		0.99958,
		0.997716,
		0.992826,
		0.983139,
		0.966903,
		0.942546,
		0.908799,
		0.864779,
		0.810056,
		0.744711,
		0.66939,
		0.585368,
		0.494607,
		0.399815,
		0.304496,
		0.212988,
		0.130464,
		0.062899,
		0.016988,
		0.003918,
		0.015366,
		0.033867,
		0.058915,
		0.089974,
		0.12648,
		0.167843,
		0.213443,
		0.262634,
		0.314747,
		0.369091,
		0.424954,
		0.48161,
		0.538323,
		0.594354,
		0.648967,
		0.701439,
		0.751077,
		0.797225,
		0.839291,
		0.876762,
		0.909232,
		0.936437,
		0.95828,
		0.974876,
		0.986574,
		0.993987,
		0.997978,
		0.999606,
		0.999982,
		0.999998,
		0.999904,
		0.999028,
		0.995994,
		0.989042,
		0.9763,
		0.956001,
		0.926633,
		0.887051,
		0.836564,
		0.774998,
		0.702768,
		0.620939,
		0.531288,
		0.436364,
		0.339533,
		0.245006,
		0.157829,
		0.083836,
		0.029542,
		0.001983,
		0.00402,
		0.015765,
		0.034735,
		0.060404,
		0.092214,
		0.129577,
		0.171875,
		0.218463,
		0.268666,
		0.321785,
		0.377099,
		0.433865,
		0.491327,
		0.548718,
		0.605268,
		0.660215,
		0.712812,
		0.762345,
		0.808146,
		0.849616,
		0.886248,
		0.917657,
		0.943612,
		0.964076,
		0.97924,
		0.989554,
		0.995745,
		0.998793,
		0.999839,
		0.999993,
		0.999976,
		0.999659,
		0.997988,
		0.99331,
		0.983695,
		0.967203,
		0.942087,
		0.906936,
		0.860781,
		0.803188,
		0.734326,
		0.655041,
		0.566923,
		0.472358,
		0.374572,
		0.277651,
		0.186511,
		0.106843,
		0.04498,
		0.007716,
		0.004127,
		0.016179,
		0.035637,
		0.061951,
		0.094539,
		0.132788,
		0.176052,
		0.223658,
		0.2749,
		0.32905,
		0.385352,
		0.443032,
		0.501302,
		0.559361,
		0.61641,
		0.671658,
		0.724333,
		0.773702,
		0.819083,
		0.859876,
		0.895582,
		0.925841,
		0.950468,
		0.969489,
		0.983187,
		0.992122,
		0.997146,
		0.999349,
		0.999936,
		0.999952,
		0.999837,
		0.999026,
		0.996096,
		0.989142,
		0.9761,
		0.955004,
		0.924176,
		0.882368,
		0.828867,
		0.76359,
		0.687158,
		0.600964,
		0.507231,
		0.409045,
		0.310362,
		0.215973,
		0.131419,
		0.062848,
		0.016812,
		0.004238,
		0.01661,
		0.036574,
		0.063557,
		0.096952,
		0.136118,
		0.18038,
		0.229035,
		0.281345,
		0.336549,
		0.393858,
		0.452462,
		0.511539,
		0.570256,
		0.62778,
		0.683291,
		0.735993,
		0.785132,
		0.830017,
		0.870045,
		0.904734,
		0.933754,
		0.956972,
		0.974494,
		0.9867,
		0.994276,
		0.998198,
		0.999659,
		0.99988,
		0.999756,
		0.99933,
		0.997575,
		0.992725,
		0.982663,
		0.965238,
		0.938519,
		0.900971,
		0.8516,
		0.790059,
		0.716734,
		0.632819,
		0.540369,
		0.442328,
		0.342523,
		0.245614,
		0.156993,
		0.08262,
		0.028801,
		0.00191,
		0.004354,
		0.017058,
		0.037549,
		0.065227,
		0.099458,
		0.139573,
		0.184867,
		0.234603,
		0.288011,
		0.344293,
		0.402626,
		0.462163,
		0.522046,
		0.581406,
		0.639377,
		0.695109,
		0.747781,
		0.796619,
		0.840925,
		0.880097,
		0.913673,
		0.941362,
		0.963094,
		0.979061,
		0.989757,
		0.995994,
		0.998871,
		0.999655,
		0.999523,
		0.9991,
		0.997934,
		0.994554,
		0.986889,
		0.972665,
		0.949718,
		0.916238,
		0.870942,
		0.81321,
		0.743188,
		0.661863,
		0.57112,
		0.473757,
		0.373469,
		0.274783,
		0.182939,
		0.103703,
		0.043137,
		0.007301,
		0.004475,
		0.017525,
		0.038563,
		0.066963,
		0.102062,
		0.14316,
		0.18952,
		0.24037,
		0.294906,
		0.352291,
		0.411665,
		0.472143,
		0.532828,
		0.592813,
		0.651201,
		0.707107,
		0.759685,
		0.808146,
		0.851782,
		0.890001,
		0.922363,
		0.948626,
		0.968789,
		0.983139,
		0.99229,
		0.997181,
		0.999017,
		0.9991,
		0.998474,
		0.997356,
		0.994761,
		0.988808,
		0.977186,
		0.95755,
		0.927827,
		0.886448,
		0.832516,
		0.765928,
		0.687459,
		0.598814,
		0.502643,
		0.402506,
		0.302792,
		0.208583,
		0.125451,
		0.05921,
		0.015609,
		0.0046,
		0.018011,
		0.039618,
		0.068768,
		0.104768,
		0.146884,
		0.194346,
		0.246346,
		0.30204,
		0.360554,
		0.420985,
		0.48241,
		0.54389,
		0.604481,
		0.663248,
		0.719275,
		0.771692,
		0.819691,
		0.862561,
		0.89972,
		0.930757,
		0.955481,
		0.973965,
		0.986599,
		0.994116,
		0.997575,
		0.998268,
		0.997473,
		0.995982,
		0.993485,
		0.988479,
		0.978745,
		0.961826,
		0.935426,
		0.897699,
		0.847468,
		0.784374,
		0.708972,
		0.622781,
		0.528295,
		0.428931,
		0.328938,
		0.23323,
		0.147183,
		0.076359,
		0.026203,
		0.001708,
		0.004731,
		0.018518,
		0.040718,
		0.070648,
		0.107582,
		0.150754,
		0.199356,
		0.25254,
		0.309424,
		0.369091,
		0.430596,
		0.492972,
		0.555238,
		0.61641,
		0.675515,
		0.731605,
		0.783783,
		0.831226,
		0.873218,
		0.909189,
		0.938756,
		0.961779,
		0.97841,
		0.989142,
		0.994828,
		0.996642,
		0.99593,
		0.993865,
		0.99085,
		0.985989,
		0.977327,
		0.96241,
		0.938773,
		0.904319,
		0.857596,
		0.797981,
		0.7258,
		0.642378,
		0.550042,
		0.45206,
		0.352525,
		0.256183,
		0.168202,
		0.093895,
		0.038407,
		0.006383,
		0.004868,
		0.019046,
		0.041863,
		0.072605,
		0.110511,
		0.154777,
		0.204557,
		0.258963,
		0.317068,
		0.377913,
		0.440506,
		0.503836,
		0.566876,
		0.628601,
		0.687995,
		0.74408,
		0.795928,
		0.842698,
		0.883665,
		0.918264,
		0.94614,
		0.967203,
		0.981687,
		0.990193,
		0.993699,
		0.993485,
		0.990906,
		0.986929,
		0.981438,
		0.97297,
		0.95923,
		0.937675,
		0.906002,
		0.862495,
		0.806267,
		0.737402,
		0.65702,
		0.567267,
		0.471247,
		0.372893,
		0.276767,
		0.187825,
		0.111127,
		0.051521,
		0.013323,
		0.00501,
		0.019597,
		0.043058,
		0.074644,
		0.113559,
		0.15896,
		0.209959,
		0.265625,
		0.324985,
		0.387032,
		0.450728,
		0.515011,
		0.578809,
		0.641049,
		0.700674,
		0.756663,
		0.808054,
		0.853978,
		0.893693,
		0.926633,
		0.952465,
		0.971156,
		0.983029,
		0.988808,
		0.989607,
		0.986797,
		0.981687,
		0.974912,
		0.965756,
		0.95229,
		0.932032,
		0.902538,
		0.861859,
		0.808844,
		0.743327,
		0.666205,
		0.579429,
		0.485925,
		0.389445,
		0.294361,
		0.205406,
		0.127383,
		0.064851,
		0.021807,
		0.001391,
		0.005159,
		0.020173,
		0.044304,
		0.07677,
		0.116734,
		0.163312,
		0.215573,
		0.272538,
		0.333186,
		0.39646,
		0.461269,
		0.526502,
		0.591031,
		0.653732,
		0.713495,
		0.769247,
		0.819979,
		0.864779,
		0.902878,
		0.933706,
		0.956957,
		0.972665,
		0.981267,
		0.983637,
		0.981036,
		0.974912,
		0.966443,
		0.955795,
		0.941677,
		0.921848,
		0.893836,
		0.855502,
		0.805445,
		0.743241,
		0.669547,
		0.586099,
		0.495623,
		0.401677,
		0.308422,
		0.220363,
		0.142042,
		0.077731,
		0.031123,
		0.005056,
		0.005315,
		0.020774,
		0.045605,
		0.078987,
		0.120042,
		0.167843,
		0.221409,
		0.279714,
		0.341684,
		0.406206,
		0.472135,
		0.538297,
		0.603506,
		0.666569,
		0.726308,
		0.781582,
		0.831316,
		0.874547,
		0.910471,
		0.938519,
		0.95843,
		0.970337,
		0.974831,
		0.97297,
		0.966179,
		0.955943,
		0.943198,
		0.927549,
		0.907247,
		0.879941,
		0.843385,
		0.795954,
		0.736961,
		0.666804,
		0.586982,
		0.5,
		0.409197,
		0.318512,
		0.232205,
		0.154555,
		0.089552,
		0.040599,
		0.01025,
		0.005478,
		0.021402,
		0.046964,
		0.081301,
		0.123492,
		0.172561,
		0.227479,
		0.287164,
		0.350485,
		0.416269,
		0.483303,
		0.550341,
		0.616116,
		0.679351,
		0.73878,
		0.793172,
		0.841376,
		0.882368,
		0.91532,
		0.939685,
		0.955287,
		0.96241,
		0.961852,
		0.954896,
		0.943133,
		0.928019,
		0.910115,
		0.888462,
		0.861048,
		0.825636,
		0.780424,
		0.724468,
		0.657896,
		0.581942,
		0.498862,
		0.411755,
		0.324316,
		0.240551,
		0.164471,
		0.099785,
		0.04962,
		0.01627,
		0.00101,
		0.005649,
		0.022059,
		0.048385,
		0.083718,
		0.127091,
		0.177477,
		0.233791,
		0.294891,
		0.359581,
		0.426612,
		0.494688,
		0.56247,
		0.628589,
		0.691658,
		0.750302,
		0.803188,
		0.849081,
		0.88691,
		0.915851,
		0.935426,
		0.945607,
		0.946897,
		0.940363,
		0.927549,
		0.910182,
		0.889587,
		0.865825,
		0.837497,
		0.80255,
		0.759087,
		0.705931,
		0.642924,
		0.571013,
		0.492178,
		0.40925,
		0.325662,
		0.24515,
		0.171451,
		0.108001,
		0.057659,
		0.022482,
		0.00355,
		0.005827,
		0.022747,
		0.04987,
		0.086243,
		0.130845,
		0.182595,
		0.240343,
		0.302876,
		0.368915,
		0.437117,
		0.50608,
		0.574344,
		0.640413,
		0.702768,
		0.75991,
		0.810403,
		0.852944,
		0.886448,
		0.910151,
		0.923719,
		0.92735,
		0.921848,
		0.908605,
		0.889428,
		0.86612,
		0.839721,
		0.809757,
		0.774593,
		0.732363,
		0.681705,
		0.622176,
		0.554413,
		0.48009,
		0.401745,
		0.322525,
		0.245885,
		0.175283,
		0.113883,
		0.064291,
		0.02834,
		0.006942,
		0.006014,
		0.023468,
		0.051425,
		0.088879,
		0.134756,
		0.187904,
		0.247099,
		0.311037,
		0.378334,
		0.447525,
		0.517072,
		0.585368,
		0.650766,
		0.711604,
		0.766258,
		0.81321,
		0.851131,
		0.878988,
		0.896157,
		0.902538,
		0.898653,
		0.885676,
		0.865353,
		0.839721,
		0.810531,
		0.778389,
		0.742376,
		0.700845,
		0.652329,
		0.596121,
		0.532534,
		0.462908,
		0.389462,
		0.315035,
		0.242785,
		0.175889,
		0.11724,
		0.069201,
		0.033402,
		0.010604,
		0.000636,
		0.006211,
		0.024222,
		0.053048,
		0.091624,
		0.138805,
		0.19336,
		0.253963,
		0.319197,
		0.38754,
		0.457378,
		0.527004,
		0.594644,
		0.658487,
		0.716734,
		0.767663,
		0.809723,
		0.841631,
		0.862495,
		0.871933,
		0.870185,
		0.858176,
		0.837497,
		0.810233,
		0.778552,
		0.743988,
		0.706621,
		0.665273,
		0.618508,
		0.565401,
		0.505937,
		0.441107,
		0.37278,
		0.303466,
		0.236018,
		0.17332,
		0.118003,
		0.072196,
		0.037339,
		0.014062,
		0.002137,
		0.006416,
		0.02501,
		0.054735,
		0.094456,
		0.142941,
		0.198852,
		0.260738,
		0.327031,
		0.396042,
		0.465973,
		0.534931,
		0.600964,
		0.662108,
		0.716456,
		0.76225,
		0.797981,
		0.822512,
		0.835202,
		0.836021,
		0.825636,
		0.805429,
		0.777408,
		0.743918,
		0.707107,
		0.668102,
		0.626498,
		0.581082,
		0.530799,
		0.475331,
		0.415302,
		0.352213,
		0.288229,
		0.22588,
		0.167757,
		0.116229,
		0.073201,
		0.039939,
		0.016957,
		0.003985,
		0.00663,
		0.025825,
		0.056465,
		0.097323,
		0.147054,
		0.204182,
		0.267095,
		0.334039,
		0.403127,
		0.472358,
		0.539646,
		0.602874,
		0.659966,
		0.708972,
		0.748179,
		0.77623,
		0.792248,
		0.795954,
		0.787765,
		0.768831,
		0.740998,
		0.706621,
		0.66817,
		0.627576,
		0.585423,
		0.540983,
		0.493212,
		0.441538,
		0.386226,
		0.328396,
		0.269848,
		0.212781,
		0.159489,
		0.112084,
		0.072255,
		0.041104,
		0.019045,
		0.005777,
		0.000329,
		0.006851,
		0.026654,
		0.058196,
		0.100129,
		0.15096,
		0.209038,
		0.272552,
		0.339533,
		0.407876,
		0.475373,
		0.539764,
		0.598814,
		0.650402,
		0.692627,
		0.723937,
		0.743241,
		0.750038,
		0.744506,
		0.727561,
		0.700845,
		0.666607,
		0.627422,
		0.585696,
		0.542935,
		0.499182,
		0.4536,
		0.405462,
		0.354699,
		0.302048,
		0.248937,
		0.197224,
		0.148905,
		0.105836,
		0.0695,
		0.040846,
		0.020197,
		0.007231,
		0.001038,
		0.007072,
		0.027471,
		0.059855,
		0.102721,
		0.154387,
		0.212988,
		0.276483,
		0.34267,
		0.409225,
		0.473757,
		0.533882,
		0.58732,
		0.632006,
		0.666205,
		0.68864,
		0.698606,
		0.696065,
		0.681705,
		0.656947,
		0.623865,
		0.584989,
		0.542935,
		0.499836,
		0.456627,
		0.412943,
		0.368043,
		0.321588,
		0.273949,
		0.226175,
		0.179781,
		0.136466,
		0.097831,
		0.065163,
		0.039266,
		0.020382,
		0.008178,
		0.001801,
		0.007286,
		0.028234,
		0.061335,
		0.104887,
		0.156977,
		0.215498,
		0.278157,
		0.342523,
		0.406079,
		0.466305,
		0.520775,
		0.567267,
		0.603886,
		0.629185,
		0.642277,
		0.642924,
		0.631602,
		0.609506,
		0.578504,
		0.540983,
		0.499583,
		0.456759,
		0.414176,
		0.372177,
		0.330213,
		0.287772,
		0.244896,
		0.202269,
		0.161065,
		0.122681,
		0.088475,
		0.05954,
		0.036547,
		0.019662,
		0.008546,
		0.002407,
		0.000126,
		0.00748,
		0.028884,
		0.062493,
		0.10636,
		0.158311,
		0.215973,
		0.276811,
		0.338191,
		0.397462,
		0.45206,
		0.499619,
		0.538097,
		0.565889,
		0.581942,
		0.585835,
		0.577846,
		0.558952,
		0.530799,
		0.495581,
		0.455837,
		0.414121,
		0.372533,
		0.332146,
		0.292845,
		0.254094,
		0.215674,
		0.177929,
		0.141698,
		0.108085,
		0.078208,
		0.052975,
		0.032928,
		0.018168,
		0.008349,
		0.002744,
		0.000356,
		0.007635,
		0.029348,
		0.06316,
		0.106843,
		0.157944,
		0.213831,
		0.271753,
		0.328938,
		0.382687,
		0.430498,
		0.470186,
		0.5,
		0.518731,
		0.525792,
		0.52127,
		0.505937,
		0.481216,
		0.449081,
		0.4119,
		0.372177,
		0.332194,
		0.293546,
		0.256712,
		0.22133,
		0.18702,
		0.153832,
		0.122288,
		0.09321,
		0.067482,
		0.045834,
		0.028687,
		0.016081,
		0.007669,
		0.002784,
		0.000544,
		0.007731,
		0.029542,
		0.063155,
		0.106041,
		0.155471,
		0.208583,
		0.26248,
		0.314334,
		0.361507,
		0.401677,
		0.43295,
		0.453969,
		0.463986,
		0.462908,
		0.451312,
		0.430401,
		0.401925,
		0.368043,
		0.331119,
		0.293444,
		0.25688,
		0.222444,
		0.190146,
		0.159588,
		0.130593,
		0.103399,
		0.078559,
		0.056733,
		0.038483,
		0.024116,
		0.013614,
		0.006631,
		0.002558,
		0.000623,
		2.7e-005,
		0.007746,
		0.029384,
		0.062312,
		0.103703,
		0.150584,
		0.199929,
		0.24878,
		0.294361,
		0.334211,
		0.366296,
		0.389109,
		0.401745,
		0.403941,
		0.39608,
		0.379157,
		0.354699,
		0.32464,
		0.291157,
		0.256446,
		0.222444,
		0.190498,
		0.161075,
		0.133926,
		0.108744,
		0.085527,
		0.064584,
		0.046365,
		0.031269,
		0.019499,
		0.010985,
		0.005384,
		0.002141,
		0.000588,
		6.1e-005,
		0.007659,
		0.028801,
		0.060501,
		0.099666,
		0.143141,
		0.187825,
		0.230803,
		0.269468,
		0.30164,
		0.325662,
		0.34047,
		0.34563,
		0.341342,
		0.328396,
		0.308099,
		0.282165,
		0.252561,
		0.22133,
		0.190361,
		0.161142,
		0.134491,
		0.110452,
		0.088707,
		0.069081,
		0.051666,
		0.036721,
		0.024496,
		0.015095,
		0.008402,
		0.00408,
		0.001626,
		0.00047,
		6.8e-005,
		0.007452,
		0.027745,
		0.057659,
		0.093895,
		0.133206,
		0.172521,
		0.209081,
		0.240551,
		0.265117,
		0.281552,
		0.289251,
		0.288229,
		0.279078,
		0.262895,
		0.241175,
		0.215674,
		0.188251,
		0.160683,
		0.134466,
		0.110603,
		0.089443,
		0.070788,
		0.054366,
		0.040099,
		0.028075,
		0.018409,
		0.011116,
		0.006041,
		0.002854,
		0.001107,
		0.000316,
		5.1e-005,
		0.007118,
		0.026203,
		0.053801,
		0.086499,
		0.121059,
		0.154555,
		0.184485,
		0.208867,
		0.226305,
		0.236018,
		0.237836,
		0.232157,
		0.219874,
		0.202269,
		0.180892,
		0.15741,
		0.133455,
		0.110452,
		0.089462,
		0.071031,
		0.05516,
		0.041577,
		0.03008,
		0.020615,
		0.013182,
		0.007717,
		0.004032,
		0.001811,
		0.000658,
		0.000173,
		2.6e-005,
		0.006658,
		0.024201,
		0.049033,
		0.077731,
		0.107181,
		0.134701,
		0.158129,
		0.175889,
		0.187018,
		0.191158,
		0.188515,
		0.179781,
		0.166038,
		0.148632,
		0.129044,
		0.108744,
		0.08905,
		0.070995,
		0.055213,
		0.041877,
		0.030794,
		0.021703,
		0.014441,
		0.00891,
		0.004985,
		0.002452,
		0.001013,
		0.000326,
		7.1e-005,
		0.006083,
		0.021807,
		0.043541,
		0.067966,
		0.092199,
		0.113883,
		0.131251,
		0.143149,
		0.14903,
		0.148905,
		0.143273,
		0.133017,
		0.119294,
		0.103399,
		0.086641,
		0.070213,
		0.055078,
		0.041877,
		0.030877,
		0.021994,
		0.014966,
		0.009563,
		0.005612,
		0.002939,
		0.001319,
		0.000475,
		0.000122,
		1.8e-005,
		0.005415,
		0.019122,
		0.037571,
		0.057659,
		0.076818,
		0.093074,
		0.105078,
		0.112084,
		0.113912,
		0.110866,
		0.103642,
		0.09321,
		0.080692,
		0.067237,
		0.053909,
		0.041577,
		0.030845,
		0.022008,
		0.015059,
		0.00978,
		0.005911,
		0.003231,
		0.001535,
		0.000596,
		0.00017,
		2.8e-005,
		0.004685,
		0.01627,
		0.031406,
		0.047295,
		0.061746,
		0.073201,
		0.080714,
		0.08391,
		0.082904,
		0.078208,
		0.070615,
		0.061078,
		0.050596,
		0.040099,
		0.030358,
		0.021916,
		0.015054,
		0.0098,
		0.005983,
		0.003347,
		0.001651,
		0.000676,
		0.000206,
		3.8e-005,
		0.003928,
		0.013387,
		0.025332,
		0.037339,
		0.047631,
		0.05506,
		0.059053,
		0.05954,
		0.056853,
		0.051615,
		0.04462,
		0.036721,
		0.02872,
		0.021286,
		0.014891,
		0.00978,
		0.005985,
		0.003364,
		0.001686,
		0.000712,
		0.000228,
		4.4e-005,
		0.003178,
		0.010604,
		0.019612,
		0.028199,
		0.035006,
		0.039266,
		0.040725,
		0.039541,
		0.036175,
		0.031269,
		0.025537,
		0.019659,
		0.014202,
		0.009563,
		0.005944,
		0.003361,
		0.001689,
		0.000719,
		0.000236,
		4.8e-005,
		0.002468,
		0.008036,
		0.014469,
		0.020197,
		0.024257,
		0.026217,
		0.026068,
		0.024116,
		0.020858,
		0.016877,
		0.012738,
		0.00891,
		0.005716,
		0.003307,
		0.001682,
		0.000719,
		0.000237,
		4.9e-005,
		0.001827,
		0.005777,
		0.010067,
		0.013546,
		0.015607,
		0.016081,
		0.015128,
		0.01311,
		0.010488,
		0.007717,
		0.005174,
		0.00311,
		0.001631,
		0.000712,
		0.000236,
		4.9e-005,
		0.001275,
		0.00389,
		0.006505,
		0.008349,
		0.009106,
		0.008796,
		0.007659,
		0.006041,
		0.004297,
		0.002719,
		0.001491,
		0.000676,
		0.000231,
		4.8e-005,
		0.000827,
		0.002407,
		0.003813,
		0.004591,
		0.004637,
		0.00408,
		0.003161,
		0.002146,
		0.001249,
		0.000596,
		0.000213,
		4.6e-005,
		0.000487,
		0.001329,
		0.001951,
		0.002141,
		0.001926,
		0.00146,
		0.000927,
		0.000475,
		0.00018,
		4.1e-005,
		0.00025,
		0.000623,
		0.000815,
		0.00077,
		0.000568,
		0.000326,
		0.000135,
		3.3e-005,
		0.000105,
		0.000225,
		0.000242,
		0.000173,
		8.3e-005,
		2.3e-005,
		3.1e-005,
		5.1e-005,
		3.6e-005,
		1.2e-005,
		1.2e-005,
		3.6e-005,
		5.1e-005,
		3.1e-005,
		2.3e-005,
		8.3e-005,
		0.000173,
		0.000242,
		0.000225,
		0.000105,
		3.3e-005,
		0.000135,
		0.000326,
		0.000568,
		0.00077,
		0.000815,
		0.000623,
		0.00025,
		4.1e-005,
		0.00018,
		0.000475,
		0.000927,
		0.00146,
		0.001926,
		0.002141,
		0.001951,
		0.001329,
		0.000487,
		4.6e-005,
		0.000213,
		0.000596,
		0.001249,
		0.002146,
		0.003161,
		0.00408,
		0.004637,
		0.004591,
		0.003813,
		0.002407,
		0.000827,
		4.8e-005,
		0.000231,
		0.000676,
		0.001491,
		0.002719,
		0.004297,
		0.006041,
		0.007659,
		0.008796,
		0.009106,
		0.008349,
		0.006505,
		0.00389,
		0.001275,
		4.9e-005,
		0.000236,
		0.000712,
		0.001631,
		0.00311,
		0.005174,
		0.007717,
		0.010488,
		0.01311,
		0.015128,
		0.016081,
		0.015607,
		0.013546,
		0.010067,
		0.005777,
		0.001827,
		4.9e-005,
		0.000237,
		0.000719,
		0.001682,
		0.003307,
		0.005716,
		0.00891,
		0.012738,
		0.016877,
		0.020858,
		0.024116,
		0.026068,
		0.026217,
		0.024257,
		0.020197,
		0.014469,
		0.008036,
		0.002468,
		4.8e-005,
		0.000236,
		0.000719,
		0.001689,
		0.003361,
		0.005944,
		0.009563,
		0.014202,
		0.019659,
		0.025537,
		0.031269,
		0.036175,
		0.039541,
		0.040725,
		0.039266,
		0.035006,
		0.028199,
		0.019612,
		0.010604,
		0.003178,
		4.4e-005,
		0.000228,
		0.000712,
		0.001686,
		0.003364,
		0.005985,
		0.00978,
		0.014891,
		0.021286,
		0.02872,
		0.036721,
		0.04462,
		0.051615,
		0.056853,
		0.05954,
		0.059053,
		0.05506,
		0.047631,
		0.037339,
		0.025332,
		0.013387,
		0.003928,
		3.8e-005,
		0.000206,
		0.000676,
		0.001651,
		0.003347,
		0.005983,
		0.0098,
		0.015054,
		0.021916,
		0.030358,
		0.040099,
		0.050596,
		0.061078,
		0.070615,
		0.078208,
		0.082904,
		0.08391,
		0.080714,
		0.073201,
		0.061746,
		0.047295,
		0.031406,
		0.01627,
		0.004685,
		2.8e-005,
		0.00017,
		0.000596,
		0.001535,
		0.003231,
		0.005911,
		0.00978,
		0.015059,
		0.022008,
		0.030845,
		0.041577,
		0.053909,
		0.067237,
		0.080692,
		0.09321,
		0.103642,
		0.110866,
		0.113912,
		0.112084,
		0.105078,
		0.093074,
		0.076818,
		0.057659,
		0.037571,
		0.019122,
		0.005415,
		1.8e-005,
		0.000122,
		0.000475,
		0.001319,
		0.002939,
		0.005612,
		0.009563,
		0.014966,
		0.021994,
		0.030877,
		0.041877,
		0.055078,
		0.070213,
		0.086641,
		0.103399,
		0.119294,
		0.133017,
		0.143273,
		0.148905,
		0.14903,
		0.143149,
		0.131251,
		0.113883,
		0.092199,
		0.067966,
		0.043541,
		0.021807,
		0.006083,
		7.1e-005,
		0.000326,
		0.001013,
		0.002452,
		0.004985,
		0.00891,
		0.014441,
		0.021703,
		0.030794,
		0.041877,
		0.055213,
		0.070995,
		0.08905,
		0.108744,
		0.129044,
		0.148632,
		0.166038,
		0.179781,
		0.188515,
		0.191158,
		0.187018,
		0.175889,
		0.158129,
		0.134701,
		0.107181,
		0.077731,
		0.049033,
		0.024201,
		0.006658,
		2.6e-005,
		0.000173,
		0.000658,
		0.001811,
		0.004032,
		0.007717,
		0.013182,
		0.020615,
		0.03008,
		0.041577,
		0.05516,
		0.071031,
		0.089462,
		0.110452,
		0.133455,
		0.15741,
		0.180892,
		0.202269,
		0.219874,
		0.232157,
		0.237836,
		0.236018,
		0.226305,
		0.208867,
		0.184485,
		0.154555,
		0.121059,
		0.086499,
		0.053801,
		0.026203,
		0.007118,
		5.1e-005,
		0.000316,
		0.001107,
		0.002854,
		0.006041,
		0.011116,
		0.018409,
		0.028075,
		0.040099,
		0.054366,
		0.070788,
		0.089443,
		0.110603,
		0.134466,
		0.160683,
		0.188251,
		0.215674,
		0.241175,
		0.262895,
		0.279078,
		0.288229,
		0.289251,
		0.281552,
		0.265117,
		0.240551,
		0.209081,
		0.172521,
		0.133206,
		0.093895,
		0.057659,
		0.027745,
		0.007452,
		6.8e-005,
		0.00047,
		0.001626,
		0.00408,
		0.008402,
		0.015095,
		0.024496,
		0.036721,
		0.051666,
		0.069081,
		0.088707,
		0.110452,
		0.134491,
		0.161142,
		0.190361,
		0.22133,
		0.252561,
		0.282165,
		0.308099,
		0.328396,
		0.341342,
		0.34563,
		0.34047,
		0.325662,
		0.30164,
		0.269468,
		0.230803,
		0.187825,
		0.143141,
		0.099666,
		0.060501,
		0.028801,
		0.007659,
		6.1e-005,
		0.000588,
		0.002141,
		0.005384,
		0.010985,
		0.019499,
		0.031269,
		0.046365,
		0.064584,
		0.085527,
		0.108744,
		0.133926,
		0.161075,
		0.190498,
		0.222444,
		0.256446,
		0.291157,
		0.32464,
		0.354699,
		0.379157,
		0.39608,
		0.403941,
		0.401745,
		0.389109,
		0.366296,
		0.334211,
		0.294361,
		0.24878,
		0.199929,
		0.150584,
		0.103703,
		0.062312,
		0.029384,
		0.007746,
		2.7e-005,
		0.000623,
		0.002558,
		0.006631,
		0.013614,
		0.024116,
		0.038483,
		0.056733,
		0.078559,
		0.103399,
		0.130593,
		0.159588,
		0.190146,
		0.222444,
		0.25688,
		0.293444,
		0.331119,
		0.368043,
		0.401925,
		0.430401,
		0.451312,
		0.462908,
		0.463986,
		0.453969,
		0.43295,
		0.401677,
		0.361507,
		0.314334,
		0.26248,
		0.208583,
		0.155471,
		0.106041,
		0.063155,
		0.029542,
		0.007731,
		0.000544,
		0.002784,
		0.007669,
		0.016081,
		0.028687,
		0.045834,
		0.067482,
		0.09321,
		0.122288,
		0.153832,
		0.18702,
		0.22133,
		0.256712,
		0.293546,
		0.332194,
		0.372177,
		0.4119,
		0.449081,
		0.481216,
		0.505937,
		0.52127,
		0.525792,
		0.518731,
		0.5,
		0.470186,
		0.430498,
		0.382687,
		0.328938,
		0.271753,
		0.213831,
		0.157944,
		0.106843,
		0.06316,
		0.029348,
		0.007635,
		0.000356,
		0.002744,
		0.008349,
		0.018168,
		0.032928,
		0.052975,
		0.078208,
		0.108085,
		0.141698,
		0.177929,
		0.215674,
		0.254094,
		0.292845,
		0.332146,
		0.372533,
		0.414121,
		0.455837,
		0.495581,
		0.530799,
		0.558952,
		0.577846,
		0.585835,
		0.581942,
		0.565889,
		0.538097,
		0.499619,
		0.45206,
		0.397462,
		0.338191,
		0.276811,
		0.215973,
		0.158311,
		0.10636,
		0.062493,
		0.028884,
		0.00748,
		0.000126,
		0.002407,
		0.008546,
		0.019662,
		0.036547,
		0.05954,
		0.088475,
		0.122681,
		0.161065,
		0.202269,
		0.244896,
		0.287772,
		0.330213,
		0.372177,
		0.414176,
		0.456759,
		0.499583,
		0.540983,
		0.578504,
		0.609506,
		0.631602,
		0.642924,
		0.642277,
		0.629185,
		0.603886,
		0.567267,
		0.520775,
		0.466305,
		0.406079,
		0.342523,
		0.278157,
		0.215498,
		0.156977,
		0.104887,
		0.061335,
		0.028234,
		0.007286,
		0.001801,
		0.008178,
		0.020382,
		0.039266,
		0.065163,
		0.097831,
		0.136466,
		0.179781,
		0.226175,
		0.273949,
		0.321588,
		0.368043,
		0.412943,
		0.456627,
		0.499836,
		0.542935,
		0.584989,
		0.623865,
		0.656947,
		0.681705,
		0.696065,
		0.698606,
		0.68864,
		0.666205,
		0.632006,
		0.58732,
		0.533882,
		0.473757,
		0.409225,
		0.34267,
		0.276483,
		0.212988,
		0.154387,
		0.102721,
		0.059855,
		0.027471,
		0.007072,
		0.001038,
		0.007231,
		0.020197,
		0.040846,
		0.0695,
		0.105836,
		0.148905,
		0.197224,
		0.248937,
		0.302048,
		0.354699,
		0.405462,
		0.4536,
		0.499182,
		0.542935,
		0.585696,
		0.627422,
		0.666607,
		0.700845,
		0.727561,
		0.744506,
		0.750038,
		0.743241,
		0.723937,
		0.692627,
		0.650402,
		0.598814,
		0.539764,
		0.475373,
		0.407876,
		0.339533,
		0.272552,
		0.209038,
		0.15096,
		0.100129,
		0.058196,
		0.026654,
		0.006851,
		0.000329,
		0.005777,
		0.019045,
		0.041104,
		0.072255,
		0.112084,
		0.159489,
		0.212781,
		0.269848,
		0.328396,
		0.386226,
		0.441538,
		0.493212,
		0.540983,
		0.585423,
		0.627576,
		0.66817,
		0.706621,
		0.740998,
		0.768831,
		0.787765,
		0.795954,
		0.792248,
		0.77623,
		0.748179,
		0.708972,
		0.659966,
		0.602874,
		0.539646,
		0.472358,
		0.403127,
		0.334039,
		0.267095,
		0.204182,
		0.147054,
		0.097323,
		0.056465,
		0.025825,
		0.00663,
		0.003985,
		0.016957,
		0.039939,
		0.073201,
		0.116229,
		0.167757,
		0.22588,
		0.288229,
		0.352213,
		0.415302,
		0.475331,
		0.530799,
		0.581082,
		0.626498,
		0.668102,
		0.707107,
		0.743918,
		0.777408,
		0.805429,
		0.825636,
		0.836021,
		0.835202,
		0.822512,
		0.797981,
		0.76225,
		0.716456,
		0.662108,
		0.600964,
		0.534931,
		0.465973,
		0.396042,
		0.327031,
		0.260738,
		0.198852,
		0.142941,
		0.094456,
		0.054735,
		0.02501,
		0.006416,
		0.002137,
		0.014062,
		0.037339,
		0.072196,
		0.118003,
		0.17332,
		0.236018,
		0.303466,
		0.37278,
		0.441107,
		0.505937,
		0.565401,
		0.618508,
		0.665273,
		0.706621,
		0.743988,
		0.778552,
		0.810233,
		0.837497,
		0.858176,
		0.870185,
		0.871933,
		0.862495,
		0.841631,
		0.809723,
		0.767663,
		0.716734,
		0.658487,
		0.594644,
		0.527004,
		0.457378,
		0.38754,
		0.319197,
		0.253963,
		0.19336,
		0.138805,
		0.091624,
		0.053048,
		0.024222,
		0.006211,
		0.000636,
		0.010604,
		0.033402,
		0.069201,
		0.11724,
		0.175889,
		0.242785,
		0.315035,
		0.389462,
		0.462908,
		0.532534,
		0.596121,
		0.652329,
		0.700845,
		0.742376,
		0.778389,
		0.810531,
		0.839721,
		0.865353,
		0.885676,
		0.898653,
		0.902538,
		0.896157,
		0.878988,
		0.851131,
		0.81321,
		0.766258,
		0.711604,
		0.650766,
		0.585368,
		0.517072,
		0.447525,
		0.378334,
		0.311037,
		0.247099,
		0.187904,
		0.134756,
		0.088879,
		0.051425,
		0.023468,
		0.006014,
		0.006942,
		0.02834,
		0.064291,
		0.113883,
		0.175283,
		0.245885,
		0.322525,
		0.401745,
		0.48009,
		0.554413,
		0.622176,
		0.681705,
		0.732363,
		0.774593,
		0.809757,
		0.839721,
		0.86612,
		0.889428,
		0.908605,
		0.921848,
		0.92735,
		0.923719,
		0.910151,
		0.886448,
		0.852944,
		0.810403,
		0.75991,
		0.702768,
		0.640413,
		0.574344,
		0.50608,
		0.437117,
		0.368915,
		0.302876,
		0.240343,
		0.182595,
		0.130845,
		0.086243,
		0.04987,
		0.022747,
		0.005827,
		0.00355,
		0.022482,
		0.057659,
		0.108001,
		0.171451,
		0.24515,
		0.325662,
		0.40925,
		0.492178,
		0.571013,
		0.642924,
		0.705931,
		0.759087,
		0.80255,
		0.837497,
		0.865825,
		0.889587,
		0.910182,
		0.927549,
		0.940363,
		0.946897,
		0.945607,
		0.935426,
		0.915851,
		0.88691,
		0.849081,
		0.803188,
		0.750302,
		0.691658,
		0.628589,
		0.56247,
		0.494688,
		0.426612,
		0.359581,
		0.294891,
		0.233791,
		0.177477,
		0.127091,
		0.083718,
		0.048385,
		0.022059,
		0.005649,
		0.00101,
		0.01627,
		0.04962,
		0.099785,
		0.164471,
		0.240551,
		0.324316,
		0.411755,
		0.498862,
		0.581942,
		0.657896,
		0.724468,
		0.780424,
		0.825636,
		0.861048,
		0.888462,
		0.910115,
		0.928019,
		0.943133,
		0.954896,
		0.961852,
		0.96241,
		0.955287,
		0.939685,
		0.91532,
		0.882368,
		0.841376,
		0.793172,
		0.73878,
		0.679351,
		0.616116,
		0.550341,
		0.483303,
		0.416269,
		0.350485,
		0.287164,
		0.227479,
		0.172561,
		0.123492,
		0.081301,
		0.046964,
		0.021402,
		0.005478,
		0.01025,
		0.040599,
		0.089552,
		0.154555,
		0.232205,
		0.318512,
		0.409197,
		0.5,
		0.586982,
		0.666804,
		0.736961,
		0.795954,
		0.843385,
		0.879941,
		0.907247,
		0.927549,
		0.943198,
		0.955943,
		0.966179,
		0.97297,
		0.974831,
		0.970337,
		0.95843,
		0.938519,
		0.910471,
		0.874547,
		0.831316,
		0.781582,
		0.726308,
		0.666569,
		0.603506,
		0.538297,
		0.472135,
		0.406206,
		0.341684,
		0.279714,
		0.221409,
		0.167843,
		0.120042,
		0.078987,
		0.045605,
		0.020774,
		0.005315,
		0.005056,
		0.031123,
		0.077731,
		0.142042,
		0.220363,
		0.308422,
		0.401677,
		0.495623,
		0.586099,
		0.669547,
		0.743241,
		0.805445,
		0.855502,
		0.893836,
		0.921848,
		0.941677,
		0.955795,
		0.966443,
		0.974912,
		0.981036,
		0.983637,
		0.981267,
		0.972665,
		0.956957,
		0.933706,
		0.902878,
		0.864779,
		0.819979,
		0.769247,
		0.713495,
		0.653732,
		0.591031,
		0.526502,
		0.461269,
		0.39646,
		0.333186,
		0.272538,
		0.215573,
		0.163312,
		0.116734,
		0.07677,
		0.044304,
		0.020173,
		0.005159,
		0.001391,
		0.021807,
		0.064851,
		0.127383,
		0.205406,
		0.294361,
		0.389445,
		0.485925,
		0.579429,
		0.666205,
		0.743327,
		0.808844,
		0.861859,
		0.902538,
		0.932032,
		0.95229,
		0.965756,
		0.974912,
		0.981687,
		0.986797,
		0.989607,
		0.988808,
		0.983029,
		0.971156,
		0.952465,
		0.926633,
		0.893693,
		0.853978,
		0.808054,
		0.756663,
		0.700674,
		0.641049,
		0.578809,
		0.515011,
		0.450728,
		0.387032,
		0.324985,
		0.265625,
		0.209959,
		0.15896,
		0.113559,
		0.074644,
		0.043058,
		0.019597,
		0.00501,
		0.013323,
		0.051521,
		0.111127,
		0.187825,
		0.276767,
		0.372893,
		0.471247,
		0.567267,
		0.65702,
		0.737402,
		0.806267,
		0.862495,
		0.906002,
		0.937675,
		0.95923,
		0.97297,
		0.981438,
		0.986929,
		0.990906,
		0.993485,
		0.993699,
		0.990193,
		0.981687,
		0.967203,
		0.94614,
		0.918264,
		0.883665,
		0.842698,
		0.795928,
		0.74408,
		0.687995,
		0.628601,
		0.566876,
		0.503836,
		0.440506,
		0.377913,
		0.317068,
		0.258963,
		0.204557,
		0.154777,
		0.110511,
		0.072605,
		0.041863,
		0.019046,
		0.004868,
		0.006383,
		0.038407,
		0.093895,
		0.168202,
		0.256183,
		0.352525,
		0.45206,
		0.550042,
		0.642378,
		0.7258,
		0.797981,
		0.857596,
		0.904319,
		0.938773,
		0.96241,
		0.977327,
		0.985989,
		0.99085,
		0.993865,
		0.99593,
		0.996642,
		0.994828,
		0.989142,
		0.97841,
		0.961779,
		0.938756,
		0.909189,
		0.873218,
		0.831226,
		0.783783,
		0.731605,
		0.675515,
		0.61641,
		0.555238,
		0.492972,
		0.430596,
		0.369091,
		0.309424,
		0.25254,
		0.199356,
		0.150754,
		0.107582,
		0.070648,
		0.040718,
		0.018518,
		0.004731,
		0.001708,
		0.026203,
		0.076359,
		0.147183,
		0.23323,
		0.328938,
		0.428931,
		0.528295,
		0.622781,
		0.708972,
		0.784374,
		0.847468,
		0.897699,
		0.935426,
		0.961826,
		0.978745,
		0.988479,
		0.993485,
		0.995982,
		0.997473,
		0.998268,
		0.997575,
		0.994116,
		0.986599,
		0.973965,
		0.955481,
		0.930757,
		0.89972,
		0.862561,
		0.819691,
		0.771692,
		0.719275,
		0.663248,
		0.604481,
		0.54389,
		0.48241,
		0.420985,
		0.360554,
		0.30204,
		0.246346,
		0.194346,
		0.146884,
		0.104768,
		0.068768,
		0.039618,
		0.018011,
		0.0046,
		0.015609,
		0.05921,
		0.125451,
		0.208583,
		0.302792,
		0.402506,
		0.502643,
		0.598814,
		0.687459,
		0.765928,
		0.832516,
		0.886448,
		0.927827,
		0.95755,
		0.977186,
		0.988808,
		0.994761,
		0.997356,
		0.998474,
		0.9991,
		0.999017,
		0.997181,
		0.99229,
		0.983139,
		0.968789,
		0.948626,
		0.922363,
		0.890001,
		0.851782,
		0.808146,
		0.759685,
		0.707107,
		0.651201,
		0.592813,
		0.532828,
		0.472143,
		0.411665,
		0.352291,
		0.294906,
		0.24037,
		0.18952,
		0.14316,
		0.102062,
		0.066963,
		0.038563,
		0.017525,
		0.004475,
		0.007301,
		0.043137,
		0.103703,
		0.182939,
		0.274783,
		0.373469,
		0.473757,
		0.57112,
		0.661863,
		0.743188,
		0.81321,
		0.870942,
		0.916238,
		0.949718,
		0.972665,
		0.986889,
		0.994554,
		0.997934,
		0.9991,
		0.999523,
		0.999655,
		0.998871,
		0.995994,
		0.989757,
		0.979061,
		0.963094,
		0.941362,
		0.913673,
		0.880097,
		0.840925,
		0.796619,
		0.747781,
		0.695109,
		0.639377,
		0.581406,
		0.522046,
		0.462163,
		0.402626,
		0.344293,
		0.288011,
		0.234603,
		0.184867,
		0.139573,
		0.099458,
		0.065227,
		0.037549,
		0.017058,
		0.004354,
		0.00191,
		0.028801,
		0.08262,
		0.156993,
		0.245614,
		0.342523,
		0.442328,
		0.540369,
		0.632819,
		0.716734,
		0.790059,
		0.8516,
		0.900971,
		0.938519,
		0.965238,
		0.982663,
		0.992725,
		0.997575,
		0.99933,
		0.999756,
		0.99988,
		0.999659,
		0.998198,
		0.994276,
		0.9867,
		0.974494,
		0.956972,
		0.933754,
		0.904734,
		0.870045,
		0.830017,
		0.785132,
		0.735993,
		0.683291,
		0.62778,
		0.570256,
		0.511539,
		0.452462,
		0.393858,
		0.336549,
		0.281345,
		0.229035,
		0.18038,
		0.136118,
		0.096952,
		0.063557,
		0.036574,
		0.01661,
		0.004238,
		0.016812,
		0.062848,
		0.131419,
		0.215973,
		0.310362,
		0.409045,
		0.507231,
		0.600964,
		0.687158,
		0.76359,
		0.828867,
		0.882368,
		0.924176,
		0.955004,
		0.9761,
		0.989142,
		0.996096,
		0.999026,
		0.999837,
		0.999952,
		0.999936,
		0.999349,
		0.997146,
		0.992122,
		0.983187,
		0.969489,
		0.950468,
		0.925841,
		0.895582,
		0.859876,
		0.819083,
		0.773702,
		0.724333,
		0.671658,
		0.61641,
		0.559361,
		0.501302,
		0.443032,
		0.385352,
		0.32905,
		0.2749,
		0.223658,
		0.176052,
		0.132788,
		0.094539,
		0.061951,
		0.035637,
		0.016179,
		0.004127,
		0.007716,
		0.04498,
		0.106843,
		0.186511,
		0.277651,
		0.374572,
		0.472358,
		0.566923,
		0.655041,
		0.734326,
		0.803188,
		0.860781,
		0.906936,
		0.942087,
		0.967203,
		0.983695,
		0.99331,
		0.997988,
		0.999659,
		0.999976,
		0.999993,
		0.999839,
		0.998793,
		0.995745,
		0.989554,
		0.97924,
		0.964076,
		0.943612,
		0.917657,
		0.886248,
		0.849616,
		0.808146,
		0.762345,
		0.712812,
		0.660215,
		0.605268,
		0.548718,
		0.491327,
		0.433865,
		0.377099,
		0.321785,
		0.268666,
		0.218463,
		0.171875,
		0.129577,
		0.092214,
		0.060404,
		0.034735,
		0.015765,
		0.00402,
		0.001983,
		0.029542,
		0.083836,
		0.157829,
		0.245006,
		0.339533,
		0.436364,
		0.531288,
		0.620939,
		0.702768,
		0.774998,
		0.836564,
		0.887051,
		0.926633,
		0.956001,
		0.9763,
		0.989042,
		0.995994,
		0.999028,
		0.999904,
		0.999998,
		0.999982,
		0.999606,
		0.997978,
		0.993987,
		0.986574,
		0.974876,
		0.95828,
		0.936437,
		0.909232,
		0.876762,
		0.839291,
		0.797225,
		0.751077,
		0.701439,
		0.648967,
		0.594354,
		0.538323,
		0.48161,
		0.424954,
		0.369091,
		0.314747,
		0.262634,
		0.213443,
		0.167843,
		0.12648,
		0.089974,
		0.058915,
		0.033867,
		0.015366,
		0.003918,
		0.016988,
		0.062899,
		0.130464,
		0.212988,
		0.304496,
		0.399815,
		0.494607,
		0.585368,
		0.66939,
		0.744711,
		0.810056,
		0.864779,
		0.908799,
		0.942546,
		0.966903,
		0.983139,
		0.992826,
		0.997716,
		0.99958,
		0.999982,
		1,
		0.99992,
		0.999205,
		0.996876,
		0.991857,
		0.983187,
		0.970112,
		0.952128,
		0.928972,
		0.900598,
		0.867151,
		0.828924,
		0.786339,
		0.739908,
		0.690221,
		0.637915,
		0.583667,
		0.528173,
		0.472143,
		0.416289,
		0.361318,
		0.307926,
		0.256796,
		0.208589,
		0.163949,
		0.123492,
		0.087813,
		0.057479,
		0.033031,
		0.014982,
		0.003819,
		0.007692,
		0.044463,
		0.104887,
		0.182095,
		0.269972,
		0.36322,
		0.457378,
		0.548801,
		0.634625,
		0.712711,
		0.781582,
		0.840368,
		0.888751,
		0.926912,
		0.955481,
		0.975486,
		0.988288,
		0.995487,
		0.998793,
		0.99985,
		0.999999,
		0.999995,
		0.999771,
		0.998595,
		0.995458,
		0.989348,
		0.9794,
		0.964969,
		0.945646,
		0.921246,
		0.891784,
		0.85744,
		0.818536,
		0.775502,
		0.728851,
		0.679164,
		0.627063,
		0.573204,
		0.518262,
		0.462921,
		0.407865,
		0.353774,
		0.301315,
		0.251144,
		0.203896,
		0.160186,
		0.120607,
		0.085729,
		0.056095,
		0.032226,
		0.014613,
		0.003724,
		0.001953,
		0.028884,
		0.081496,
		0.152754,
		0.236406,
		0.327031,
		0.420042,
		0.51166,
		0.598868,
		0.679351,
		0.751442,
		0.81406,
		0.86666,
		0.909189,
		0.942036,
		0.965998,
		0.982219,
		0.992122,
		0.997312,
		0.999429,
		0.999962,
		1,
		0.999964,
		0.999493,
		0.99774,
		0.993706,
		0.986456,
		0.975226,
		0.959469,
		0.938862,
		0.913289,
		0.882814,
		0.847654,
		0.808146,
		0.764729,
		0.717915,
		0.668272,
		0.61641,
		0.562964,
		0.508587,
		0.453937,
		0.399673,
		0.346449,
		0.294906,
		0.245671,
		0.199356,
		0.156549,
		0.117821,
		0.083718,
		0.054761,
		0.031449,
		0.014256,
		0.003632,
		0.01645,
		0.060626,
		0.125331,
		0.204182,
		0.291638,
		0.382989,
		0.474322,
		0.56247,
		0.644953,
		0.719918,
		0.786085,
		0.842698,
		0.889481,
		0.926602,
		0.954629,
		0.974494,
		0.987433,
		0.994912,
		0.998511,
		0.999771,
		0.999995,
		0.999999,
		0.999875,
		0.999046,
		0.996609,
		0.991607,
		0.983187,
		0.970681,
		0.953636,
		0.931803,
		0.905127,
		0.873716,
		0.837812,
		0.797771,
		0.754033,
		0.707107,
		0.657551,
		0.605958,
		0.552945,
		0.499142,
		0.445185,
		0.391707,
		0.339336,
		0.28869,
		0.24037,
		0.194962,
		0.153034,
		0.11513,
		0.081776,
		0.053473,
		0.0307,
		0.013913,
		0.003544,
		0.007387,
		0.042546,
		0.100129,
		0.173627,
		0.257379,
		0.346557,
		0.437117,
		0.525747,
		0.609807,
		0.687268,
		0.756663,
		0.817037,
		0.867908,
		0.909231,
		0.941362,
		0.965026,
		0.981267,
		0.99139,
		0.996876,
		0.999251,
		0.999929,
		1,
		0.999986,
		0.999689,
		0.998392,
		0.99518,
		0.989154,
		0.979547,
		0.965784,
		0.947493,
		0.924495,
		0.896786,
		0.864511,
		0.827935,
		0.787425,
		0.743424,
		0.696434,
		0.647002,
		0.595706,
		0.543144,
		0.489923,
		0.436658,
		0.383959,
		0.332429,
		0.282661,
		0.235233,
		0.19071,
		0.149633,
		0.112529,
		0.079901,
		0.05223,
		0.029978,
		0.013582,
		0.003458,
		0.001862,
		0.027471,
		0.077397,
		0.14501,
		0.224539,
		0.311037,
		0.400334,
		0.488978,
		0.574175,
		0.653732,
		0.726004,
		0.789852,
		0.844598,
		0.890001,
		0.926215,
		0.953766,
		0.973507,
		0.986574,
		0.994317,
		0.9982,
		0.999671,
		0.999986,
		1,
		0.999937,
		0.999369,
		0.997501,
		0.993437,
		0.986346,
		0.975548,
		0.960553,
		0.941064,
		0.916963,
		0.888291,
		0.855222,
		0.818041,
		0.777124,
		0.732912,
		0.685901,
		0.636628,
		0.585653,
		0.533556,
		0.480925,
		0.428351,
		0.376422,
		0.325719,
		0.276811,
		0.230255,
		0.186592,
		0.146343,
		0.110014,
		0.078089,
		0.05103,
		0.029281,
		0.013263,
		0.003376,
		0.015567,
		0.057334,
		0.118558,
		0.19336,
		0.276679,
		0.364225,
		0.452411,
		0.538297,
		0.619534,
		0.694312,
		0.761316,
		0.819691,
		0.869007,
		0.909232,
		0.940699,
		0.964076,
		0.980326,
		0.99065,
		0.996415,
		0.999046,
		0.999882,
		0.999999,
		0.999996,
		0.999819,
		0.998879,
		0.996348,
		0.991369,
		0.983187,
		0.971204,
		0.95501,
		0.934374,
		0.909232,
		0.879663,
		0.845868,
		0.808146,
		0.766878,
		0.722505,
		0.675515,
		0.62643,
		0.575799,
		0.52418,
		0.472143,
		0.420258,
		0.369091,
		0.3192,
		0.271135,
		0.225429,
		0.182603,
		0.14316,
		0.107582,
		0.076337,
		0.049871,
		0.028608,
		0.012955,
		0.003297,
		0.006962,
		0.040101,
		0.094456,
		0.164044,
		0.243699,
		0.329012,
		0.416269,
		0.502393,
		0.584885,
		0.661781,
		0.731605,
		0.793332,
		0.846361,
		0.89048,
		0.925841,
		0.952932,
		0.972543,
		0.985719,
		0.993706,
		0.997863,
		0.999548,
		0.99997,
		1,
		0.999972,
		0.999596,
		0.998186,
		0.994912,
		0.988971,
		0.979683,
		0.966531,
		0.949176,
		0.927446,
		0.901325,
		0.870925,
		0.836467,
		0.798264,
		0.756699,
		0.71221,
		0.665277,
		0.61641,
		0.56614,
		0.515011,
		0.463572,
		0.412372,
		0.361957,
		0.312866,
		0.265625,
		0.22075,
		0.178739,
		0.140077,
		0.105229,
		0.074644,
		0.04875,
		0.027957,
		0.012657,
		0.003221,
		0.001749,
		0.025825,
		0.072856,
		0.136765,
		0.212284,
		0.294891,
		0.380753,
		0.466662,
		0.549981,
		0.628601,
		0.700893,
		0.765677,
		0.822187,
		0.870045,
		0.909232,
		0.940063,
		0.963156,
		0.9794,
		0.989904,
		0.995933,
		0.998815,
		0.999819,
		0.999996,
		0.999999,
		0.9999,
		0.999236,
		0.997265,
		0.993181,
		0.986243,
		0.975845,
		0.961546,
		0.943071,
		0.920303,
		0.893264,
		0.862095,
		0.827036,
		0.788409,
		0.746597,
		0.702034,
		0.655192,
		0.606567,
		0.556677,
		0.506046,
		0.455206,
		0.404687,
		0.355015,
		0.306709,
		0.260276,
		0.21621,
		0.174994,
		0.137092,
		0.102952,
		0.073006,
		0.047667,
		0.027329,
		0.01237,
		0.003147,
		0.014606,
		0.053884,
		0.111669,
		0.182595,
		0.262037,
		0.346046,
		0.431293,
		0.515011,
		0.594955,
		0.669354,
		0.736881,
		0.796619,
		0.848031,
		0.890936,
		0.925483,
		0.952128,
		0.971601,
		0.984868,
		0.993081,
		0.997501,
		0.999403,
		0.999944,
		1,
		0.999989,
		0.999752,
		0.998706,
		0.996092,
		0.991144,
		0.983187,
		0.971686,
		0.956268,
		0.936718,
		0.912966,
		0.88507,
		0.853193,
		0.817592,
		0.778592,
		0.736579,
		0.691982,
		0.645261,
		0.596902,
		0.547405,
		0.497281,
		0.447041,
		0.397198,
		0.34826,
		0.300724,
		0.255081,
		0.211807,
		0.171364,
		0.1342,
		0.100747,
		0.071421,
		0.046619,
		0.026722,
		0.012092,
		0.003075,
		0.006522,
		0.037644,
		0.088879,
		0.154774,
		0.230604,
		0.312315,
		0.39646,
		0.480153,
		0.561021,
		0.63716,
		0.707107,
		0.769799,
		0.824552,
		0.871032,
		0.909232,
		0.939451,
		0.962262,
		0.978488,
		0.989154,
		0.995431,
		0.99856,
		0.999738,
		0.999989,
		1,
		0.99995,
		0.999493,
		0.997978,
		0.994653,
		0.9888,
		0.97981,
		0.967218,
		0.950715,
		0.930138,
		0.905457,
		0.876762,
		0.844236,
		0.808146,
		0.768825,
		0.726654,
		0.682058,
		0.635486,
		0.587413,
		0.538323,
		0.488711,
		0.439072,
		0.3899,
		0.341684,
		0.294906,
		0.250035,
		0.207533,
		0.167843,
		0.131398,
		0.098612,
		0.069887,
		0.045605,
		0.026134,
		0.011823,
		0.003007,
		0.001637,
		0.024222,
		0.068505,
		0.128948,
		0.200736,
		0.279714,
		0.362328,
		0.445578,
		0.526971,
		0.604481,
		0.676513,
		0.741872,
		0.799732,
		0.849616,
		0.89137,
		0.92514,
		0.951351,
		0.970681,
		0.984024,
		0.992444,
		0.997117,
		0.999236,
		0.999906,
		0.999999,
		0.999997,
		0.999854,
		0.999094,
		0.99703,
		0.992935,
		0.986146,
		0.97612,
		0.962458,
		0.944907,
		0.92335,
		0.897796,
		0.868358,
		0.835238,
		0.798713,
		0.759116,
		0.716828,
		0.672265,
		0.625869,
		0.5781,
		0.529428,
		0.480333,
		0.431293,
		0.382786,
		0.335282,
		0.289247,
		0.245133,
		0.203384,
		0.164428,
		0.128681,
		0.096544,
		0.068402,
		0.044624,
		0.025566,
		0.011564,
		0.00294,
		0.013691,
		0.050639,
		0.105229,
		0.172561,
		0.248386,
		0.32905,
		0.411446,
		0.492972,
		0.571482,
		0.645261,
		0.712986,
		0.773702,
		0.826794,
		0.87197,
		0.909232,
		0.938862,
		0.961395,
		0.977591,
		0.988401,
		0.994912,
		0.99828,
		0.999638,
		0.999976,
		1,
		0.999977,
		0.999675,
		0.998528,
		0.995843,
		0.990931,
		0.983187,
		0.97213,
		0.957423,
		0.938862,
		0.916375,
		0.890001,
		0.859876,
		0.826216,
		0.789303,
		0.749475,
		0.707107,
		0.662608,
		0.61641,
		0.568961,
		0.520718,
		0.472143,
		0.423701,
		0.375851,
		0.32905,
		0.283743,
		0.24037,
		0.199356,
		0.161115,
		0.126047,
		0.094539,
		0.066963,
		0.043674,
		0.025016,
		0.011312,
		0.002875,
		0.006111,
		0.035363,
		0.083718,
		0.146196,
		0.218463,
		0.29677,
		0.377913,
		0.459182,
		0.538323,
		0.613507,
		0.683291,
		0.746597,
		0.802682,
		0.851122,
		0.891784,
		0.924811,
		0.950601,
		0.969784,
		0.983187,
		0.991798,
		0.996711,
		0.999046,
		0.999854,
		0.999997,
		0.999999,
		0.99992,
		0.999381,
		0.997769,
		0.994403,
		0.988638,
		0.979927,
		0.967852,
		0.952128,
		0.9326,
		0.909232,
		0.882091,
		0.851331,
		0.817181,
		0.779928,
		0.739908,
		0.697495,
		0.653088,
		0.607109,
		0.559995,
		0.512188,
		0.464137,
		0.416289,
		0.369091,
		0.32298,
		0.278389,
		0.23574,
		0.195443,
		0.157899,
		0.123492,
		0.092596,
		0.065569,
		0.042755,
		0.024483,
		0.011069,
		0.002813,
		0.001534,
		0.022747,
		0.064503,
		0.121749,
		0.190067,
		0.265625,
		0.345123,
		0.425753,
		0.505161,
		0.581406,
		0.652936,
		0.718555,
		0.777401,
		0.828924,
		0.872864,
		0.909232,
		0.938296,
		0.960553,
		0.976711,
		0.987648,
		0.994377,
		0.997978,
		0.999519,
		0.999955,
		1,
		0.999991,
		0.999799,
		0.998946,
		0.996799,
		0.9927,
		0.986055,
		0.976376,
		0.963299,
		0.946591,
		0.926139,
		0.901939,
		0.874082,
		0.842738,
		0.808146,
		0.770596,
		0.730424,
		0.687996,
		0.643707,
		0.597967,
		0.5512,
		0.503836,
		0.45631,
		0.409054,
		0.362499,
		0.317068,
		0.273179,
		0.231239,
		0.191643,
		0.154777,
		0.121013,
		0.090711,
		0.064217,
		0.041863,
		0.023968,
		0.010834,
		0.002753,
		0.012854,
		0.047667,
		0.099316,
		0.163312,
		0.23574,
		0.313213,
		0.392831,
		0.472143,
		0.549108,
		0.622067,
		0.689715,
		0.751077,
		0.805484,
		0.852555,
		0.892179,
		0.924495,
		0.949876,
		0.968908,
		0.982358,
		0.991144,
		0.996286,
		0.998834,
		0.999788,
		0.999991,
		1,
		0.999959,
		0.999588,
		0.998347,
		0.995601,
		0.990728,
		0.983187,
		0.972543,
		0.958486,
		0.94083,
		0.919497,
		0.894513,
		0.86599,
		0.834112,
		0.799123,
		0.761317,
		0.721027,
		0.678615,
		0.634466,
		0.588982,
		0.542574,
		0.495659,
		0.448658,
		0.40199,
		0.356071,
		0.31131,
		0.268109,
		0.226862,
		0.18795,
		0.151745,
		0.118608,
		0.088884,
		0.062907,
		0.041,
		0.023468,
		0.010606,
		0.002694,
		0.005737,
		0.033279,
		0.078987,
		0.138302,
		0.207232,
		0.282312,
		0.360554,
		0.439414,
		0.516759,
		0.590828,
		0.660215,
		0.723836,
		0.780913,
		0.830949,
		0.873716,
		0.909232,
		0.937751,
		0.959736,
		0.975845,
		0.986896,
		0.993829,
		0.997655,
		0.999381,
		0.999924,
		0.999999,
		0.999997,
		0.999882,
		0.999261,
		0.997561,
		0.994162,
		0.988485,
		0.980037,
		0.968439,
		0.953429,
		0.93486,
		0.912691,
		0.886971,
		0.85783,
		0.825463,
		0.79012,
		0.752097,
		0.711723,
		0.669354,
		0.625367,
		0.580154,
		0.534114,
		0.487653,
		0.441178,
		0.395094,
		0.349803,
		0.3057,
		0.263174,
		0.222604,
		0.18436,
		0.148801,
		0.116272,
		0.08711,
		0.061637,
		0.040162,
		0.022984,
		0.010385,
		0.002638,
		0.00144,
		0.021402,
		0.06084,
		0.11513,
		0.180211,
		0.25254,
		0.32905,
		0.407111,
		0.484498,
		0.559361,
		0.630193,
		0.69581,
		0.755329,
		0.808146,
		0.85392,
		0.892557,
		0.924192,
		0.949176,
		0.968053,
		0.981538,
		0.990485,
		0.995843,
		0.9986,
		0.999707,
		0.999981,
		1,
		0.999982,
		0.999735,
		0.998793,
		0.996571,
		0.992476,
		0.98597,
		0.976614,
		0.964076,
		0.948142,
		0.9287,
		0.905738,
		0.879329,
		0.849616,
		0.816804,
		0.781149,
		0.742944,
		0.702516,
		0.660215,
		0.61641,
		0.571482,
		0.52582,
		0.479816,
		0.433865,
		0.38836,
		0.343689,
		0.300233,
		0.258369,
		0.218463,
		0.180871,
		0.14594,
		0.114005,
		0.085389,
		0.060404,
		0.03935,
		0.022514,
		0.010171,
		0.002583,
		0.012092,
		0.044948,
		0.093885,
		0.154777,
		0.22401,
		0.298442,
		0.375363,
		0.452462,
		0.527802,
		0.599783,
		0.667127,
		0.728851,
		0.78425,
		0.832877,
		0.874529,
		0.909232,
		0.937225,
		0.958943,
		0.974996,
		0.986146,
		0.99327,
		0.997312,
		0.999223,
		0.999882,
		0.999998,
		1,
		0.999935,
		0.999493,
		0.998163,
		0.995364,
		0.990536,
		0.983187,
		0.972926,
		0.959469,
		0.942641,
		0.922365,
		0.898654,
		0.8716,
		0.84136,
		0.808146,
		0.772216,
		0.733863,
		0.69341,
		0.651201,
		0.607595,
		0.562964,
		0.517687,
		0.472143,
		0.426716,
		0.381784,
		0.337724,
		0.294906,
		0.25369,
		0.214433,
		0.177478,
		0.14316,
		0.111802,
		0.083718,
		0.059208,
		0.038563,
		0.022059,
		0.009964,
		0.00253,
		0.005396,
		0.031373,
		0.074644,
		0.131022,
		0.196828,
		0.268846,
		0.344293,
		0.420779,
		0.496282,
		0.569117,
		0.637915,
		0.701601,
		0.759371,
		0.81068,
		0.855222,
		0.892918,
		0.9239,
		0.948498,
		0.967218,
		0.980727,
		0.989821,
		0.995385,
		0.998347,
		0.999608,
		0.999963,
		1,
		0.999993,
		0.999836,
		0.999134,
		0.997353,
		0.99393,
		0.98834,
		0.98014,
		0.968983,
		0.95463,
		0.936941,
		0.91587,
		0.891454,
		0.863799,
		0.833075,
		0.799498,
		0.763329,
		0.72486,
		0.684409,
		0.642312,
		0.598922,
		0.5546,
		0.509713,
		0.464632,
		0.419726,
		0.375363,
		0.331906,
		0.289713,
		0.249134,
		0.210511,
		0.174178,
		0.140457,
		0.109662,
		0.082095,
		0.058047,
		0.037798,
		0.021618,
		0.009762,
		0.002478,
		0.001354,
		0.020173,
		0.057479,
		0.109032,
		0.171089,
		0.24037,
		0.314019,
		0.389573,
		0.464929,
		0.538323,
		0.608301,
		0.673697,
		0.733619,
		0.787425,
		0.834714,
		0.875306,
		0.909232,
		0.936718,
		0.958172,
		0.974162,
		0.985399,
		0.9927,
		0.99695,
		0.999046,
		0.999827,
		0.999993,
		1,
		0.999967,
		0.999662,
		0.998635,
		0.996348,
		0.992261,
		0.98589,
		0.976835,
		0.964797,
		0.949573,
		0.931058,
		0.909232,
		0.884152,
		0.855939,
		0.82477,
		0.790869,
		0.754496,
		0.71594,
		0.675515,
		0.63355,
		0.59039,
		0.546387,
		0.501896,
		0.457278,
		0.412891,
		0.369091,
		0.326228,
		0.28465,
		0.244695,
		0.206693,
		0.170967,
		0.137829,
		0.107582,
		0.080519,
		0.05692,
		0.037057,
		0.021189,
		0.009567,
		0.002428,
		0.011395,
		0.042454,
		0.088884,
		0.146884,
		0.213114,
		0.28465,
		0.358959,
		0.433865,
		0.507526,
		0.578407,
		0.645261,
		0.707107,
		0.763216,
		0.813093,
		0.856465,
		0.893264,
		0.92362,
		0.947842,
		0.966404,
		0.979927,
		0.989154,
		0.994912,
		0.998073,
		0.999493,
		0.999938,
		1,
		0.999998,
		0.999903,
		0.999391,
		0.997978,
		0.995135,
		0.990352,
		0.983187,
		0.973282,
		0.960379,
		0.944313,
		0.925006,
		0.902465,
		0.876762,
		0.848031,
		0.816458,
		0.782268,
		0.745722,
		0.707107,
		0.66673,
		0.624916,
		0.582,
		0.538323,
		0.494234,
		0.45008,
		0.406208,
		0.362964,
		0.320688,
		0.279714,
		0.24037,
		0.202976,
		0.167843,
		0.135274,
		0.105561,
		0.078987,
		0.055825,
		0.036336,
		0.020774,
		0.009378,
		0.00238,
		0.005084,
		0.029626,
		0.070648,
		0.124297,
		0.187172,
		0.25629,
		0.32905,
		0.403207,
		0.476846,
		0.548355,
		0.61641,
		0.679948,
		0.738155,
		0.790449,
		0.836467,
		0.87605,
		0.909232,
		0.936229,
		0.957423,
		0.973345,
		0.984657,
		0.992122,
		0.996571,
		0.99885,
		0.99976,
		0.999984,
		1,
		0.999985,
		0.999782,
		0.999001,
		0.997147,
		0.993706,
		0.988203,
		0.980236,
		0.969489,
		0.955742,
		0.938862,
		0.9188,
		0.895582,
		0.869296,
		0.840087,
		0.808146,
		0.773702,
		0.737013,
		0.698364,
		0.658057,
		0.61641,
		0.573749,
		0.530408,
		0.486723,
		0.443032,
		0.399673,
		0.35698,
		0.315281,
		0.2749,
		0.236155,
		0.199356,
		0.164802,
		0.132788,
		0.103595,
		0.077498,
		0.054761,
		0.035637,
		0.02037,
		0.009194,
		0.002333,
		0.001275,
		0.019046,
		0.054388,
		0.103402,
		0.162631,
		0.229035,
		0.299949,
		0.373065,
		0.446396,
		0.518262,
		0.587261,
		0.652255,
		0.712348,
		0.766878,
		0.815394,
		0.857652,
		0.893596,
		0.92335,
		0.947206,
		0.965609,
		0.979138,
		0.988485,
		0.994427,
		0.997782,
		0.999361,
		0.999903,
		0.999998,
		1,
		0.999946,
		0.999581,
		0.998474,
		0.996128,
		0.992055,
		0.985814,
		0.977043,
		0.965467,
		0.950898,
		0.933236,
		0.912454,
		0.888597,
		0.861767,
		0.832117,
		0.799844,
		0.765177,
		0.728373,
		0.689715,
		0.649498,
		0.608032,
		0.565638,
		0.522638,
		0.47936,
		0.436133,
		0.393283,
		0.351133,
		0.310002,
		0.270205,
		0.232048,
		0.19583,
		0.161842,
		0.130369,
		0.101684,
		0.076051,
		0.053727,
		0.034957,
		0.019978,
		0.009016,
		0.002287,
		0.010757,
		0.040162,
		0.084269,
		0.139573,
		0.202976,
		0.271757,
		0.343545,
		0.416289,
		0.488241,
		0.557927,
		0.624138,
		0.685901,
		0.742476,
		0.793332,
		0.838141,
		0.876762,
		0.909232,
		0.935757,
		0.956695,
		0.972543,
		0.983919,
		0.991538,
		0.996177,
		0.998635,
		0.999678,
		0.99997,
		1,
		0.999994,
		0.999864,
		0.999281,
		0.997793,
		0.994912,
		0.990178,
		0.983187,
		0.973615,
		0.961225,
		0.94586,
		0.927446,
		0.90598,
		0.881522,
		0.854185,
		0.824131,
		0.791559,
		0.756699,
		0.719808,
		0.681162,
		0.641052,
		0.599783,
		0.557664,
		0.515011,
		0.472143,
		0.429378,
		0.387032,
		0.34542,
		0.30485,
		0.265625,
		0.228043,
		0.192394,
		0.15896,
		0.128015,
		0.099824,
		0.074644,
		0.052722,
		0.034297,
		0.019597,
		0.008842,
		0.002243,
		0.004799,
		0.028021,
		0.066963,
		0.118071,
		0.178197,
		0.244566,
		0.314747,
		0.38663,
		0.4584,
		0.528519,
		0.595706,
		0.65892,
		0.717343,
		0.770369,
		0.817592,
		0.858788,
		0.893914,
		0.92309,
		0.946591,
		0.964833,
		0.978359,
		0.987816,
		0.99393,
		0.997473,
		0.999212,
		0.999858,
		0.999994,
		1,
		0.999973,
		0.99972,
		0.998863,
		0.996943,
		0.99349,
		0.988073,
		0.980326,
		0.969962,
		0.956775,
		0.94064,
		0.921508,
		0.899393,
		0.874369,
		0.846562,
		0.816138,
		0.783299,
		0.748275,
		0.71132,
		0.672708,
		0.632722,
		0.591661,
		0.549827,
		0.507526,
		0.465069,
		0.422765,
		0.380919,
		0.339837,
		0.299819,
		0.261156,
		0.224139,
		0.189047,
		0.156153,
		0.125723,
		0.098015,
		0.073275,
		0.051745,
		0.033655,
		0.019227,
		0.008674,
		0.0022,
		0.001204,
		0.018011,
		0.051539,
		0.098193,
		0.154777,
		0.218463,
		0.286766,
		0.357518,
		0.428844,
		0.499142,
		0.567067,
		0.631507,
		0.691577,
		0.746597,
		0.796084,
		0.839741,
		0.877444,
		0.909232,
		0.935301,
		0.955987,
		0.971756,
		0.983187,
		0.990947,
		0.995768,
		0.998403,
		0.999581,
		0.999949,
		1,
		0.999998,
		0.99992,
		0.999493,
		0.99831,
		0.995914,
		0.991857,
		0.985742,
		0.977237,
		0.96609,
		0.952128,
		0.935251,
		0.915432,
		0.892703,
		0.867151,
		0.838908,
		0.808146,
		0.77507,
		0.739908,
		0.702914,
		0.664354,
		0.624508,
		0.583667,
		0.542124,
		0.50018,
		0.458135,
		0.416289,
		0.37494,
		0.334382,
		0.294906,
		0.256796,
		0.220331,
		0.185784,
		0.153419,
		0.123492,
		0.096253,
		0.071944,
		0.050795,
		0.033031,
		0.018867,
		0.00851,
		0.002158,
		0.010171,
		0.038051,
		0.080003,
		0.132788,
		0.193529,
		0.259691,
		0.32905,
		0.399673,
		0.469901,
		0.538323,
		0.603767,
		0.665277,
		0.722107,
		0.773702,
		0.819691,
		0.859876,
		0.89422,
		0.92284,
		0.945995,
		0.964076,
		0.977591,
		0.987147,
		0.993424,
		0.997147,
		0.999046,
		0.999801,
		0.999987,
		1,
		0.999988,
		0.999819,
		0.999166,
		0.997607,
		0.994696,
		0.990011,
		0.983187,
		0.973927,
		0.962012,
		0.947296,
		0.929706,
		0.909232,
		0.885923,
		0.859876,
		0.83123,
		0.800163,
		0.766878,
		0.731605,
		0.694592,
		0.656102,
		0.61641,
		0.575799,
		0.534555,
		0.492972,
		0.451339,
		0.409949,
		0.369091,
		0.32905,
		0.290107,
		0.25254,
		0.216617,
		0.182603,
		0.150754,
		0.121319,
		0.094539,
		0.070648,
		0.049871,
		0.032424,
		0.018518,
		0.008351,
		0.002117,
		0.004537,
		0.026543,
		0.063557,
		0.112297,
		0.169842,
		0.233606,
		0.301315,
		0.370984,
		0.440894,
		0.509577,
		0.575799,
		0.638541,
		0.696992,
		0.750529,
		0.798713,
		0.841272,
		0.878098,
		0.909232,
		0.93486,
		0.955299,
		0.970985,
		0.982461,
		0.990352,
		0.995346,
		0.998153,
		0.99947,
		0.99992,
		0.999998,
		1,
		0.999956,
		0.999651,
		0.998722,
		0.996742,
		0.993282,
		0.987949,
		0.980411,
		0.970403,
		0.957735,
		0.94229,
		0.924016,
		0.90292,
		0.879064,
		0.852555,
		0.823539,
		0.792195,
		0.758729,
		0.723368,
		0.686356,
		0.647954,
		0.608428,
		0.568056,
		0.527118,
		0.485897,
		0.444677,
		0.403741,
		0.363368,
		0.323837,
		0.285421,
		0.248386,
		0.212995,
		0.179502,
		0.148158,
		0.119202,
		0.09287,
		0.069387,
		0.048971,
		0.031834,
		0.018178,
		0.008197,
		0.002078,
		0.001138,
		0.017058,
		0.048908,
		0.093366,
		0.14747,
		0.208589,
		0.2744,
		0.342867,
		0.412216,
		0.480925,
		0.547701,
		0.611467,
		0.671346,
		0.726654,
		0.776885,
		0.821699,
		0.860918,
		0.894513,
		0.922598,
		0.945417,
		0.963338,
		0.976835,
		0.986479,
		0.992908,
		0.996806,
		0.998863,
		0.999732,
		0.999975,
		1,
		0.999995,
		0.999887,
		0.999399,
		0.998145,
		0.995704,
		0.991668,
		0.985674,
		0.97742,
		0.966673,
		0.953272,
		0.937122,
		0.918192,
		0.896506,
		0.872136,
		0.845197,
		0.815842,
		0.78425,
		0.750628,
		0.715201,
		0.67821,
		0.63991,
		0.600563,
		0.560439,
		0.519811,
		0.478955,
		0.438147,
		0.397661,
		0.357769,
		0.318742,
		0.280842,
		0.24433,
		0.20946,
		0.176478,
		0.145627,
		0.11714,
		0.091244,
		0.068159,
		0.048096,
		0.031259,
		0.017847,
		0.008046,
		0.00204,
		0.009632,
		0.036101,
		0.076051,
		0.12648,
		0.184715,
		0.248386,
		0.315408,
		0.383959,
		0.452462,
		0.51957,
		0.584147,
		0.645261,
		0.702164,
		0.754286,
		0.801227,
		0.842738,
		0.878726,
		0.909232,
		0.934434,
		0.95463,
		0.97023,
		0.981742,
		0.989754,
		0.994912,
		0.997888,
		0.999343,
		0.999882,
		0.999995,
		1,
		0.999977,
		0.999766,
		0.999046,
		0.997422,
		0.994485,
		0.989852,
		0.983187,
		0.974219,
		0.962746,
		0.948631,
		0.931803,
		0.912247,
		0.890001,
		0.865149,
		0.837812,
		0.808146,
		0.776333,
		0.742579,
		0.707107,
		0.670154,
		0.631971,
		0.592813,
		0.552945,
		0.512632,
		0.472143,
		0.431746,
		0.391707,
		0.352291,
		0.31376,
		0.276369,
		0.24037,
		0.20601,
		0.173528,
		0.14316,
		0.11513,
		0.08966,
		0.066963,
		0.047243,
		0.0307,
		0.017525,
		0.0079,
		0.002002,
		0.004296,
		0.025179,
		0.060404,
		0.106933,
		0.162051,
		0.223347,
		0.28869,
		0.356209,
		0.424278,
		0.491497,
		0.556677,
		0.618827,
		0.677145,
		0.731,
		0.779928,
		0.823621,
		0.861918,
		0.894796,
		0.922365,
		0.944857,
		0.962616,
		0.976091,
		0.985814,
		0.992386,
		0.996452,
		0.998665,
		0.999651,
		0.999957,
		1,
		0.999999,
		0.999933,
		0.999575,
		0.998577,
		0.996543,
		0.993081,
		0.987832,
		0.980491,
		0.970816,
		0.958631,
		0.943825,
		0.926344,
		0.906191,
		0.883417,
		0.858112,
		0.830407,
		0.800458,
		0.76845,
		0.734587,
		0.699089,
		0.662191,
		0.624138,
		0.58518,
		0.545574,
		0.50558,
		0.465459,
		0.425471,
		0.385876,
		0.34693,
		0.308888,
		0.271997,
		0.236502,
		0.202643,
		0.170651,
		0.140754,
		0.113171,
		0.088117,
		0.065798,
		0.046414,
		0.030156,
		0.017212,
		0.007758,
		0.001966,
		0.001077,
		0.016179,
		0.046472,
		0.088884,
		0.140662,
		0.199356,
		0.26279,
		0.32905,
		0.39646,
		0.463572,
		0.529145,
		0.592135,
		0.651685,
		0.707107,
		0.757879,
		0.803632,
		0.844144,
		0.879329,
		0.909232,
		0.934022,
		0.953979,
		0.969489,
		0.98103,
		0.989154,
		0.994468,
		0.997607,
		0.999202,
		0.999834,
		0.999989,
		1,
		0.99999,
		0.999848,
		0.999299,
		0.997978,
		0.995499,
		0.991486,
		0.98561,
		0.977591,
		0.967218,
		0.954338,
		0.938862,
		0.920755,
		0.900035,
		0.876762,
		0.851035,
		0.82299,
		0.792785,
		0.760605,
		0.726654,
		0.69115,
		0.654322,
		0.61641,
		0.577661,
		0.538323,
		0.498652,
		0.458899,
		0.419319,
		0.380165,
		0.341684,
		0.304124,
		0.267725,
		0.232725,
		0.199356,
		0.167843,
		0.138407,
		0.111261,
		0.086613,
		0.064663,
		0.045605,
		0.029626,
		0.016907,
		0.007619,
		0.001931,
		0.009134,
		0.034297,
		0.072383,
		0.120607,
		0.176478,
		0.237781,
		0.302559,
		0.369091,
		0.43588,
		0.501639,
		0.565272,
		0.625869,
		0.68269,
		0.735156,
		0.782841,
		0.825463,
		0.862877,
		0.895068,
		0.922139,
		0.944313,
		0.961912,
		0.975358,
		0.985151,
		0.991857,
		0.996084,
		0.998451,
		0.999556,
		0.999933,
		0.999999,
		1,
		0.999963,
		0.999707,
		0.998921,
		0.997238,
		0.994282,
		0.9897,
		0.983187,
		0.974494,
		0.963433,
		0.949876,
		0.933754,
		0.915048,
		0.893788,
		0.870045,
		0.843926,
		0.815567,
		0.785132,
		0.752804,
		0.718785,
		0.683291,
		0.646547,
		0.608789,
		0.570256,
		0.531193,
		0.491846,
		0.452462,
		0.413289,
		0.37457,
		0.336549,
		0.299464,
		0.263549,
		0.229035,
		0.196146,
		0.165103,
		0.136118,
		0.109399,
		0.085147,
		0.063557,
		0.044818,
		0.02911,
		0.01661,
		0.007484,
		0.001896,
		0.004073,
		0.023917,
		0.057479,
		0.101941,
		0.154777,
		0.213733,
		0.276811,
		0.342249,
		0.408505,
		0.474244,
		0.538323,
		0.599783,
		0.657831,
		0.711836,
		0.761317,
		0.805937,
		0.845492,
		0.879909,
		0.909232,
		0.933622,
		0.953345,
		0.968764,
		0.980326,
		0.988552,
		0.994013,
		0.997312,
		0.999046,
		0.999776,
		0.999979,
		1,
		0.999996,
		0.999905,
		0.999493,
		0.998429,
		0.996348,
		0.992887,
		0.98772,
		0.980567,
		0.971204,
		0.959469,
		0.945255,
		0.928511,
		0.909232,
		0.887461,
		0.863277,
		0.836792,
		0.808146,
		0.777504,
		0.74505,
		0.710982,
		0.675515,
		0.638868,
		0.601273,
		0.562964,
		0.52418,
		0.48516,
		0.446146,
		0.407377,
		0.369091,
		0.331523,
		0.294906,
		0.259467,
		0.225429,
		0.193012,
		0.162428,
		0.133884,
		0.107582,
		0.083718,
		0.06248,
		0.044051,
		0.028608,
		0.016321,
		0.007353,
		0.001863,
		0.001021,
		0.015366,
		0.044213,
		0.084714,
		0.13431,
		0.19071,
		0.251877,
		0.31601,
		0.381525,
		0.447041,
		0.511372,
		0.573509,
		0.632611,
		0.687996,
		0.739133,
		0.78563,
		0.827229,
		0.863799,
		0.895329,
		0.921922,
		0.943785,
		0.961225,
		0.974637,
		0.984492,
		0.991323,
		0.995704,
		0.998222,
		0.999448,
		0.999901,
		0.999996,
		1,
		0.999981,
		0.999803,
		0.999193,
		0.997811,
		0.995298,
		0.991312,
		0.985548,
		0.977753,
		0.967729,
		0.955335,
		0.940484,
		0.923141,
		0.903317,
		0.881063,
		0.856465,
		0.82964,
		0.800733,
		0.769907,
		0.737347,
		0.703249,
		0.667822,
		0.631286,
		0.593863,
		0.555785,
		0.517284,
		0.478594,
		0.439948,
		0.40158,
		0.363722,
		0.326602,
		0.290446,
		0.255475,
		0.221906,
		0.189951,
		0.159817,
		0.131705,
		0.10581,
		0.082324,
		0.061429,
		0.043303,
		0.028118,
		0.016039,
		0.007225,
		0.00183,
		0.008674,
		0.032625,
		0.068974,
		0.11513,
		0.168771,
		0.227824,
		0.290446,
		0.355015,
		0.42011,
		0.484498,
		0.547127,
		0.607109,
		0.663715,
		0.716363,
		0.76461,
		0.808146,
		0.846787,
		0.880467,
		0.909232,
		0.933236,
		0.952728,
		0.968053,
		0.97963,
		0.987949,
		0.993551,
		0.997003,
		0.998875,
		0.999707,
		0.999964,
		1,
		0.999999,
		0.999944,
		0.999641,
		0.998793,
		0.997056,
		0.994084,
		0.989554,
		0.983187,
		0.974752,
		0.964076,
		0.951039,
		0.935573,
		0.917657,
		0.897313,
		0.874601,
		0.849616,
		0.822478,
		0.793332,
		0.762345,
		0.729698,
		0.695586,
		0.660215,
		0.623799,
		0.586559,
		0.548718,
		0.510503,
		0.472143,
		0.433865,
		0.395897,
		0.358463,
		0.321785,
		0.286083,
		0.251572,
		0.218463,
		0.186961,
		0.157267,
		0.129577,
		0.104082,
		0.080964,
		0.060404,
		0.042574,
		0.027641,
		0.015765,
		0.007101,
		0.001798,
		0.003868,
		0.022747,
		0.054761,
		0.097288,
		0.147975,
		0.204713,
		0.265625,
		0.32905,
		0.393526,
		0.457781,
		0.520718,
		0.581406,
		0.639069,
		0.693079,
		0.742944,
		0.788304,
		0.828924,
		0.864686,
		0.895582,
		0.921711,
		0.943272,
		0.960553,
		0.973927,
		0.983837,
		0.990785,
		0.995313,
		0.997978,
		0.999327,
		0.999861,
		0.999991,
		1,
		0.999991,
		0.999872,
		0.999406,
		0.99828,
		0.996156,
		0.9927,
		0.987613,
		0.980638,
		0.971569,
		0.960253,
		0.946591,
		0.930531,
		0.912065,
		0.891227,
		0.868085,
		0.842738,
		0.815311,
		0.785951,
		0.754822,
		0.722107,
		0.687996,
		0.652694,
		0.61641,
		0.579359,
		0.541761,
		0.503836,
		0.465807,
		0.427896,
		0.390324,
		0.353309,
		0.317068,
		0.281814,
		0.247755,
		0.215097,
		0.184039,
		0.154777,
		0.1275,
		0.102394,
		0.079638,
		0.059405,
		0.041863,
		0.027175,
		0.015497,
		0.006979,
		0.001767,
		0.00097,
		0.014613,
		0.042115,
		0.08083,
		0.128373,
		0.182603,
		0.24161,
		0.303696,
		0.367361,
		0.431293,
		0.494356,
		0.555576,
		0.614133,
		0.669354,
		0.720701,
		0.767766,
		0.810266,
		0.848031,
		0.881004,
		0.909232,
		0.932861,
		0.952128,
		0.967356,
		0.978942,
		0.987347,
		0.993081,
		0.996681,
		0.99869,
		0.999626,
		0.999944,
		0.999999,
		1,
		0.999969,
		0.999752,
		0.999083,
		0.997644,
		0.995103,
		0.991144,
		0.98549,
		0.977906,
		0.968209,
		0.956268,
		0.941999,
		0.925367,
		0.906376,
		0.88507,
		0.861523,
		0.835839,
		0.808146,
		0.778592,
		0.747342,
		0.714575,
		0.680482,
		0.645261,
		0.609118,
		0.572264,
		0.534913,
		0.497281,
		0.459584,
		0.422038,
		0.384859,
		0.34826,
		0.31245,
		0.277636,
		0.244022,
		0.211807,
		0.181184,
		0.152345,
		0.125472,
		0.100747,
		0.078344,
		0.05843,
		0.04117,
		0.026722,
		0.015236,
		0.006861,
		0.001737,
		0.008248,
		0.031071,
		0.065798,
		0.110014,
		0.161551,
		0.218463,
		0.279019,
		0.341684,
		0.405109,
		0.468117,
		0.529695,
		0.588982,
		0.645261,
		0.69795,
		0.746597,
		0.790869,
		0.830552,
		0.865538,
		0.895825,
		0.921508,
		0.942774,
		0.959898,
		0.973229,
		0.983187,
		0.990244,
		0.994912,
		0.997721,
		0.999193,
		0.999811,
		0.999982,
		1,
		0.999997,
		0.99992,
		0.99957,
		0.998662,
		0.996876,
		0.993892,
		0.989415,
		0.983187,
		0.974996,
		0.964681,
		0.952128,
		0.937272,
		0.92009,
		0.900598,
		0.878848,
		0.854921,
		0.828924,
		0.800988,
		0.771261,
		0.739908,
		0.707107,
		0.673044,
		0.637915,
		0.601923,
		0.565272,
		0.528173,
		0.490835,
		0.453471,
		0.416289,
		0.379501,
		0.343311,
		0.307926,
		0.273547,
		0.24037,
		0.208589,
		0.178394,
		0.149968,
		0.123492,
		0.09914,
		0.077081,
		0.057479,
		0.040494,
		0.026279,
		0.014982,
		0.006746,
		0.001708,
		0.003677,
		0.021661,
		0.05223,
		0.092945,
		0.141606,
		0.19624,
		0.255081,
		0.316562,
		0.379296,
		0.442072,
		0.503836,
		0.563688,
		0.620871,
		0.67476,
		0.72486,
		0.770794,
		0.812301,
		0.849227,
		0.881522,
		0.909232,
		0.932498,
		0.951543,
		0.966673,
		0.978262,
		0.986746,
		0.992605,
		0.996348,
		0.998492,
		0.999533,
		0.999917,
		0.999997,
		1,
		0.999984,
		0.999833,
		0.999313,
		0.99813,
		0.995967,
		0.99252,
		0.987511,
		0.980706,
		0.971912,
		0.960989,
		0.947842,
		0.932418,
		0.914709,
		0.89474,
		0.872571,
		0.848287,
		0.822,
		0.793842,
		0.763962,
		0.732524,
		0.699703,
		0.665684,
		0.630658,
		0.594824,
		0.558383,
		0.521539,
		0.484498,
		0.447466,
		0.410647,
		0.374245,
		0.338462,
		0.303496,
		0.269544,
		0.236797,
		0.205443,
		0.175666,
		0.147646,
		0.121558,
		0.09757,
		0.075848,
		0.056551,
		0.039835,
		0.025848,
		0.014734,
		0.006633,
		0.001679,
		0.000922,
		0.013913,
		0.040162,
		0.077206,
		0.122817,
		0.174994,
		0.231941,
		0.292057,
		0.353922,
		0.416289,
		0.478069,
		0.538323,
		0.596255,
		0.651201,
		0.702622,
		0.750101,
		0.793332,
		0.832117,
		0.866359,
		0.89606,
		0.921311,
		0.94229,
		0.959257,
		0.972543,
		0.982541,
		0.9897,
		0.994502,
		0.997451,
		0.999046,
		0.999752,
		0.99997,
		1,
		0.999999,
		0.999952,
		0.999695,
		0.99897,
		0.997478,
		0.994912,
		0.990983,
		0.985435,
		0.978051,
		0.968661,
		0.957143,
		0.943417,
		0.927446,
		0.909232,
		0.88881,
		0.866244,
		0.841627,
		0.815072,
		0.786713,
		0.756699,
		0.725192,
		0.692366,
		0.658402,
		0.62349,
		0.587822,
		0.551596,
		0.515011,
		0.478268,
		0.441568,
		0.405109,
		0.369091,
		0.333709,
		0.299157,
		0.265625,
		0.2333,
		0.202366,
		0.173,
		0.145377,
		0.119668,
		0.096037,
		0.074644,
		0.055645,
		0.039191,
		0.025427,
		0.014492,
		0.006524,
		0.001651,
		0.007852,
		0.029626,
		0.062836,
		0.105229,
		0.154777,
		0.209654,
		0.268228,
		0.32905,
		0.390835,
		0.452462,
		0.512956,
		0.571482,
		0.627338,
		0.679948,
		0.728851,
		0.773702,
		0.814257,
		0.850378,
		0.882021,
		0.909232,
		0.932145,
		0.950973,
		0.966004,
		0.977591,
		0.986146,
		0.992122,
		0.996003,
		0.99828,
		0.999429,
		0.999882,
		0.999992,
		1,
		0.999993,
		0.999891,
		0.999493,
		0.998528,
		0.996697,
		0.993706,
		0.989282,
		0.983187,
		0.975226,
		0.965249,
		0.953149,
		0.938862,
		0.922365,
		0.903668,
		0.882814,
		0.859876,
		0.834948,
		0.808146,
		0.779606,
		0.749475,
		0.717915,
		0.685097,
		0.651201,
		0.61641,
		0.580916,
		0.54491,
		0.508587,
		0.472143,
		0.435774,
		0.399673,
		0.364035,
		0.32905,
		0.294906,
		0.261788,
		0.229879,
		0.199356,
		0.170393,
		0.14316,
		0.117821,
		0.094539,
		0.073468,
		0.054761,
		0.038563,
		0.025016,
		0.014256,
		0.006417,
		0.001624,
		0.003501,
		0.020651,
		0.049871,
		0.088884,
		0.135635,
		0.188271,
		0.245133,
		0.304739,
		0.365774,
		0.427082,
		0.487653,
		0.546619,
		0.603241,
		0.656903,
		0.707107,
		0.753466,
		0.795699,
		0.833622,
		0.867151,
		0.896287,
		0.92112,
		0.94182,
		0.958631,
		0.971868,
		0.981901,
		0.989154,
		0.994084,
		0.997169,
		0.998886,
		0.999682,
		0.999952,
		0.999999,
		1,
		0.999973,
		0.999788,
		0.999216,
		0.997978,
		0.995782,
		0.992346,
		0.987414,
		0.98077,
		0.972236,
		0.961681,
		0.949014,
		0.934185,
		0.917181,
		0.898023,
		0.876762,
		0.853472,
		0.828254,
		0.801227,
		0.772523,
		0.742293,
		0.710695,
		0.677898,
		0.64408,
		0.60942,
		0.574105,
		0.538323,
		0.502265,
		0.466121,
		0.430082,
		0.394338,
		0.359076,
		0.324482,
		0.290741,
		0.258031,
		0.22653,
		0.196411,
		0.167843,
		0.140991,
		0.116017,
		0.093076,
		0.07232,
		0.053897,
		0.03795,
		0.024615,
		0.014026,
		0.006313,
		0.001597,
		0.000878,
		0.013263,
		0.038342,
		0.073818,
		0.117611,
		0.167843,
		0.222825,
		0.281048,
		0.341166,
		0.40199,
		0.462481,
		0.521733,
		0.578973,
		0.63355,
		0.684929,
		0.732684,
		0.776495,
		0.816138,
		0.851487,
		0.882503,
		0.909232,
		0.931803,
		0.950418,
		0.965348,
		0.976929,
		0.985548,
		0.991635,
		0.995649,
		0.998056,
		0.999313,
		0.999839,
		0.999985,
		1,
		0.999997,
		0.999932,
		0.999632,
		0.998853,
		0.997312,
		0.994726,
		0.990828,
		0.985382,
		0.978187,
		0.969087,
		0.957965,
		0.944746,
		0.929393,
		0.911904,
		0.892307,
		0.870658,
		0.84704,
		0.821553,
		0.794318,
		0.76547,
		0.735156,
		0.703534,
		0.670771,
		0.63704,
		0.602518,
		0.567389,
		0.531836,
		0.496045,
		0.460201,
		0.424491,
		0.3891,
		0.354211,
		0.320005,
		0.28666,
		0.254351,
		0.223252,
		0.193529,
		0.165349,
		0.138872,
		0.114253,
		0.091646,
		0.071199,
		0.053054,
		0.037351,
		0.024223,
		0.013802,
		0.006211,
		0.001571,
		0.007484,
		0.02828,
		0.060068,
		0.100747,
		0.148414,
		0.201355,
		0.258031,
		0.317068,
		0.37725,
		0.437501,
		0.496887,
		0.5546,
		0.609954,
		0.66238,
		0.711414,
		0.756699,
		0.797974,
		0.835071,
		0.867913,
		0.896506,
		0.920935,
		0.941362,
		0.95802,
		0.971204,
		0.981267,
		0.988607,
		0.993658,
		0.996876,
		0.998713,
		0.999603,
		0.999929,
		0.999997,
		1,
		0.999986,
		0.999857,
		0.999412,
		0.998392,
		0.996521,
		0.993526,
		0.989154,
		0.983187,
		0.975444,
		0.965784,
		0.954108,
		0.940353,
		0.924495,
		0.90654,
		0.886525,
		0.864511,
		0.840584,
		0.814849,
		0.787425,
		0.758449,
		0.728066,
		0.696434,
		0.663715,
		0.630081,
		0.595706,
		0.560768,
		0.525446,
		0.489923,
		0.454381,
		0.418999,
		0.383959,
		0.349439,
		0.315615,
		0.282661,
		0.250747,
		0.220042,
		0.19071,
		0.16291,
		0.136799,
		0.112529,
		0.090249,
		0.070103,
		0.05223,
		0.036766,
		0.023841,
		0.013582,
		0.006111,
		0.001546,
		0.003336,
		0.01971,
		0.047667,
		0.085081,
		0.130029,
		0.18077,
		0.23574,
		0.293538,
		0.352917,
		0.412775,
		0.472143,
		0.530184,
		0.586177,
		0.63952,
		0.689715,
		0.736367,
		0.77918,
		0.817948,
		0.852555,
		0.882967,
		0.909232,
		0.931471,
		0.949876,
		0.964706,
		0.976276,
		0.984953,
		0.991144,
		0.995285,
		0.997819,
		0.999185,
		0.999788,
		0.999974,
		1,
		0.999999,
		0.999959,
		0.999739,
		0.999115,
		0.997826,
		0.995601,
		0.992177,
		0.987321,
		0.980831,
		0.972543,
		0.962333,
		0.950115,
		0.935842,
		0.919497,
		0.901097,
		0.880684,
		0.858326,
		0.834112,
		0.808146,
		0.780552,
		0.751464,
		0.721027,
		0.689396,
		0.656733,
		0.623205,
		0.588982,
		0.55424,
		0.519153,
		0.483901,
		0.448658,
		0.413603,
		0.378911,
		0.344756,
		0.31131,
		0.278741,
		0.247217,
		0.2169,
		0.18795,
		0.160523,
		0.134771,
		0.110843,
		0.088884,
		0.069032,
		0.051425,
		0.036195,
		0.023468,
		0.013368,
		0.006014,
		0.001521,
		0.000836,
		0.012657,
		0.036642,
		0.070648,
		0.112726,
		0.161115,
		0.214224,
		0.270627,
		0.32905,
		0.38836,
		0.447562,
		0.505785,
		0.56228,
		0.61641,
		0.667644,
		0.715554,
		0.759807,
		0.800163,
		0.836467,
		0.868649,
		0.896717,
		0.920755,
		0.940917,
		0.957423,
		0.970552,
		0.980638,
		0.988059,
		0.993226,
		0.996571,
		0.998528,
		0.999512,
		0.999899,
		0.999993,
		1,
		0.999994,
		0.999907,
		0.999565,
		0.998733,
		0.997147,
		0.994545,
		0.990679,
		0.985332,
		0.978317,
		0.969489,
		0.958739,
		0.945995,
		0.931219,
		0.914407,
		0.895582,
		0.874792,
		0.85211,
		0.827626,
		0.80145,
		0.773702,
		0.744517,
		0.71404,
		0.682423,
		0.649825,
		0.61641,
		0.582346,
		0.547804,
		0.512956,
		0.477974,
		0.443032,
		0.408303,
		0.373956,
		0.340162,
		0.307089,
		0.2749,
		0.243759,
		0.213823,
		0.185248,
		0.158187,
		0.132788,
		0.109195,
		0.087549,
		0.067986,
		0.050639,
		0.035637,
		0.023103,
		0.013159,
		0.00592,
		0.001497,
		0.007142,
		0.027023,
		0.057479,
		0.096544,
		0.142431,
		0.193529,
		0.248386,
		0.3057,
		0.364313,
		0.423201,
		0.481465,
		0.538323,
		0.593109,
		0.645261,
		0.694316,
		0.739908,
		0.781763,
		0.819691,
		0.853585,
		0.883417,
		0.909232,
		0.931149,
		0.949349,
		0.964076,
		0.975631,
		0.984361,
		0.99065,
		0.994912,
		0.997571,
		0.999046,
		0.999729,
		0.999959,
		0.999999,
		1,
		0.999977,
		0.999819,
		0.999326,
		0.998255,
		0.996348,
		0.99335,
		0.989031,
		0.983187,
		0.97565,
		0.966289,
		0.95501,
		0.941754,
		0.926493,
		0.909232,
		0.890001,
		0.868855,
		0.845868,
		0.821134,
		0.794763,
		0.766878,
		0.737611,
		0.707107,
		0.675515,
		0.642991,
		0.609698,
		0.575799,
		0.541461,
		0.506853,
		0.472143,
		0.437501,
		0.403095,
		0.369091,
		0.335654,
		0.302949,
		0.271135,
		0.24037,
		0.210809,
		0.182603,
		0.155901,
		0.130847,
		0.107582,
		0.086243,
		0.066963,
		0.049871,
		0.035092,
		0.022747,
		0.012955,
		0.005827,
		0.001474,
		0.003183,
		0.018832,
		0.045605,
		0.081516,
		0.12476,
		0.1737,
		0.226862,
		0.282919,
		0.340688,
		0.399117,
		0.457278,
		0.514364,
		0.569676,
		0.622621,
		0.672708,
		0.719536,
		0.762798,
		0.80227,
		0.837812,
		0.869359,
		0.896922,
		0.920582,
		0.940484,
		0.956839,
		0.969911,
		0.980016,
		0.987511,
		0.992789,
		0.996257,
		0.998331,
		0.999412,
		0.999863,
		0.999987,
		1,
		0.999998,
		0.999942,
		0.999684,
		0.999011,
		0.997675,
		0.995423,
		0.992015,
		0.987232,
		0.980889,
		0.972832,
		0.962947,
		0.951151,
		0.937398,
		0.92167,
		0.903979,
		0.884362,
		0.862877,
		0.839605,
		0.814639,
		0.788092,
		0.760084,
		0.73075,
		0.70023,
		0.668673,
		0.636233,
		0.603067,
		0.569338,
		0.535208,
		0.500843,
		0.466406,
		0.432063,
		0.397978,
		0.364313,
		0.331231,
		0.298888,
		0.267443,
		0.237049,
		0.207857,
		0.180014,
		0.153664,
		0.128949,
		0.106005,
		0.084966,
		0.065963,
		0.049119,
		0.034559,
		0.022399,
		0.012755,
		0.005737,
		0.001451,
		0.000798,
		0.012092,
		0.035053,
		0.067676,
		0.108136,
		0.154777,
		0.2061,
		0.260756,
		0.317536,
		0.375363,
		0.433283,
		0.49046,
		0.546167,
		0.599783,
		0.650784,
		0.698742,
		0.743316,
		0.78425,
		0.82137,
		0.854579,
		0.883851,
		0.909232,
		0.930835,
		0.948834,
		0.963459,
		0.974996,
		0.983772,
		0.990153,
		0.994532,
		0.997312,
		0.998895,
		0.99966,
		0.999939,
		0.999998,
		1,
		0.999988,
		0.999877,
		0.999493,
		0.99861,
		0.996984,
		0.994368,
		0.990536,
		0.985283,
		0.97844,
		0.96987,
		0.959469,
		0.94717,
		0.932935,
		0.916757,
		0.898654,
		0.87867,
		0.856866,
		0.833326,
		0.808146,
		0.781437,
		0.753323,
		0.723934,
		0.69341,
		0.661898,
		0.629549,
		0.596519,
		0.562964,
		0.529046,
		0.494924,
		0.460761,
		0.426716,
		0.39295,
		0.359622,
		0.326889,
		0.294906,
		0.263824,
		0.233795,
		0.204965,
		0.177478,
		0.151474,
		0.127091,
		0.104462,
		0.083718,
		0.064984,
		0.048385,
		0.034038,
		0.022059,
		0.01256,
		0.005649,
		0.001428,
		0.006822,
		0.025848,
		0.055053,
		0.092596,
		0.136799,
		0.186142,
		0.239256,
		0.294906,
		0.35199,
		0.409531,
		0.466665,
		0.522638,
		0.5768,
		0.628601,
		0.67758,
		0.723368,
		0.765677,
		0.804301,
		0.839109,
		0.870045,
		0.897121,
		0.920413,
		0.940063,
		0.956268,
		0.969281,
		0.9794,
		0.986964,
		0.992346,
		0.995933,
		0.998123,
		0.9993,
		0.999819,
		0.999978,
		1,
		0.999999,
		0.999965,
		0.999775,
		0.999236,
		0.998117,
		0.996177,
		0.993181,
		0.988913,
		0.983187,
		0.975845,
		0.966766,
		0.955861,
		0.943071,
		0.928371,
		0.91176,
		0.893264,
		0.872931,
		0.850827,
		0.827036,
		0.801659,
		0.774805,
		0.746597,
		0.717165,
		0.686649,
		0.655192,
		0.622942,
		0.590052,
		0.556677,
		0.522972,
		0.489096,
		0.455206,
		0.421459,
		0.38801,
		0.355015,
		0.322628,
		0.290998,
		0.260276,
		0.230606,
		0.202132,
		0.174994,
		0.14933,
		0.125272,
		0.102952,
		0.082496,
		0.064028,
		0.047667,
		0.033529,
		0.021727,
		0.01237,
		0.003041,
		0.018011,
		0.043674,
		0.07817,
		0.119801,
		0.16703,
		0.218463,
		0.272844,
		0.32905,
		0.386075,
		0.443032,
		0.499142,
		0.553728,
		0.606211,
		0.656102,
		0.703003,
		0.746597,
		0.786646,
		0.82299,
		0.855538,
		0.88427,
		0.909232,
		0.930531,
		0.948332,
		0.962855,
		0.974369,
		0.983187,
		0.989654,
		0.994144,
		0.997042,
		0.998733,
		0.999581,
		0.999914,
		0.999994,
		1,
		0.999995,
		0.99992,
		0.999625,
		0.998903,
		0.997523,
		0.995249,
		0.991857,
		0.987147,
		0.980944,
		0.973107,
		0.963527,
		0.952128,
		0.938862,
		0.923712,
		0.906686,
		0.887815,
		0.867151,
		0.844764,
		0.820741,
		0.795181,
		0.768196,
		0.739908,
		0.710447,
		0.679948,
		0.648554,
		0.61641,
		0.583667,
		0.550474,
		0.516987,
		0.483358,
		0.449741,
		0.416289,
		0.383156,
		0.350491,
		0.318445,
		0.287165,
		0.256796,
		0.227479,
		0.199356,
		0.172561,
		0.14723,
		0.123492,
		0.101474,
		0.081301,
		0.063092,
		0.046964,
		0.033031,
		0.000762,
		0.011564,
		0.033565,
		0.064888,
		0.103819,
		0.148801,
		0.198419,
		0.251397,
		0.306588,
		0.362964,
		0.419614,
		0.475734,
		0.53062,
		0.583667,
		0.63436,
		0.682272,
		0.727057,
		0.76845,
		0.806258,
		0.840361,
		0.870708,
		0.897313,
		0.920249,
		0.939652,
		0.95571,
		0.968661,
		0.97879,
		0.986419,
		0.991898,
		0.995601,
		0.997904,
		0.999178,
		0.999767,
		0.999965,
		0.999999,
		1,
		0.99998,
		0.999844,
		0.999417,
		0.998486,
		0.996822,
		0.994196,
		0.990397,
		0.985237,
		0.978558,
		0.97023,
		0.960158,
		0.948277,
		0.934549,
		0.918966,
		0.901541,
		0.882312,
		0.861335,
		0.838682,
		0.814442,
		0.788716,
		0.761615,
		0.73326,
		0.70378,
		0.673308,
		0.641985,
		0.609954,
		0.577362,
		0.544357,
		0.511089,
		0.477707,
		0.444363,
		0.411206,
		0.378386,
		0.346048,
		0.31434,
		0.283404,
		0.253383,
		0.224414,
		0.196635,
		0.170178,
		0.145174,
		0.121749,
		0.100028,
		0.080132,
		0.062177,
		0.006524,
		0.024747,
		0.052777,
		0.088884,
		0.13149,
		0.179163,
		0.230606,
		0.28465,
		0.340247,
		0.39646,
		0.452462,
		0.507526,
		0.561021,
		0.612405,
		0.661224,
		0.707107,
		0.749757,
		0.788955,
		0.824552,
		0.856465,
		0.884676,
		0.909232,
		0.930235,
		0.947842,
		0.962262,
		0.973751,
		0.982605,
		0.989154,
		0.99375,
		0.996763,
		0.99856,
		0.999493,
		0.999882,
		0.999989,
		1,
		0.999998,
		0.99995,
		0.999727,
		0.999142,
		0.997978,
		0.996009,
		0.993016,
		0.9888,
		0.983187,
		0.976031,
		0.967218,
		0.956664,
		0.944313,
		0.930138,
		0.914137,
		0.896331,
		0.876762,
		0.855488,
		0.832587,
		0.808146,
		0.782268,
		0.755064,
		0.726654,
		0.697165,
		0.66673,
		0.635486,
		0.603574,
		0.571138,
		0.538323,
		0.505276,
		0.472143,
		0.439072,
		0.406208,
		0.373698,
		0.341684,
		0.310309,
		0.279714,
		0.250035,
		0.221409,
		0.193969,
		0.167843,
		0.14316,
		0.120042,
		0.098612,
		0.002907,
		0.017243,
		0.041863,
		0.075024,
		0.11513,
		0.160731,
		0.210511,
		0.26328,
		0.317968,
		0.373618,
		0.429378,
		0.484498,
		0.538323,
		0.590287,
		0.63991,
		0.686792,
		0.730613,
		0.771123,
		0.808146,
		0.84157,
		0.87135,
		0.897499,
		0.92009,
		0.939252,
		0.955164,
		0.968053,
		0.978187,
		0.985874,
		0.991447,
		0.99526,
		0.997675,
		0.999046,
		0.999707,
		0.999948,
		0.999998,
		1,
		0.99999,
		0.999894,
		0.999561,
		0.998793,
		0.997372,
		0.995079,
		0.991705,
		0.987065,
		0.980996,
		0.973368,
		0.964076,
		0.95305,
		0.940242,
		0.925634,
		0.909232,
		0.891062,
		0.871169,
		0.849616,
		0.826481,
		0.801855,
		0.77584,
		0.748546,
		0.720092,
		0.690605,
		0.660215,
		0.629057,
		0.59727,
		0.564995,
		0.532373,
		0.499548,
		0.466665,
		0.433865,
		0.401293,
		0.369091,
		0.337397,
		0.306352,
		0.276092,
		0.246752,
		0.218463,
		0.191355,
		0.165555,
		0.141187,
		0.000729,
		0.011069,
		0.032169,
		0.062267,
		0.099754,
		0.14316,
		0.191151,
		0.242518,
		0.296172,
		0.351133,
		0.406528,
		0.461585,
		0.515624,
		0.568056,
		0.618377,
		0.666161,
		0.711062,
		0.752804,
		0.791183,
		0.82606,
		0.85736,
		0.88507,
		0.909232,
		0.929947,
		0.947363,
		0.961681,
		0.973143,
		0.982029,
		0.988652,
		0.99335,
		0.996474,
		0.998376,
		0.999396,
		0.999844,
		0.999981,
		1,
		0.999999,
		0.99997,
		0.999805,
		0.999337,
		0.998361,
		0.996662,
		0.994029,
		0.990264,
		0.985193,
		0.978669,
		0.970571,
		0.96081,
		0.949321,
		0.93607,
		0.921045,
		0.904257,
		0.885739,
		0.865538,
		0.843722,
		0.82037,
		0.795573,
		0.769433,
		0.742062,
		0.713576,
		0.6841,
		0.653763,
		0.622699,
		0.591042,
		0.55893,
		0.526505,
		0.493904,
		0.46127,
		0.428742,
		0.39646,
		0.364563,
		0.333186,
		0.302467,
		0.272538,
		0.24353,
		0.215573,
		0.188793,
		0.006244,
		0.023716,
		0.050639,
		0.085389,
		0.12648,
		0.172561,
		0.222405,
		0.2749,
		0.32905,
		0.383959,
		0.438833,
		0.492972,
		0.545761,
		0.596674,
		0.645261,
		0.69115,
		0.73404,
		0.773702,
		0.809968,
		0.842738,
		0.87197,
		0.897679,
		0.919935,
		0.938862,
		0.95463,
		0.967454,
		0.977591,
		0.985332,
		0.990993,
		0.994912,
		0.997435,
		0.998903,
		0.999638,
		0.999926,
		0.999995,
		1,
		0.999995,
		0.999931,
		0.999675,
		0.999046,
		0.997839,
		0.995843,
		0.992856,
		0.988691,
		0.983187,
		0.976208,
		0.967646,
		0.957423,
		0.945484,
		0.931803,
		0.916375,
		0.899218,
		0.880367,
		0.859876,
		0.837812,
		0.814257,
		0.789303,
		0.763052,
		0.735615,
		0.707107,
		0.677651,
		0.647376,
		0.61641,
		0.584888,
		0.552945,
		0.520718,
		0.488343,
		0.455958,
		0.423701,
		0.391707,
		0.360112,
		0.32905,
		0.298652,
		0.269049,
		0.24037,
		0.002783,
		0.016522,
		0.040162,
		0.072063,
		0.110724,
		0.154777,
		0.202976,
		0.254193,
		0.307411,
		0.361715,
		0.416289,
		0.470412,
		0.523449,
		0.574848,
		0.624138,
		0.670922,
		0.714875,
		0.755743,
		0.793332,
		0.827516,
		0.858226,
		0.885451,
		0.909232,
		0.929666,
		0.946896,
		0.961112,
		0.972543,
		0.981456,
		0.98815,
		0.992945,
		0.996177,
		0.998182,
		0.999289,
		0.999798,
		0.99997,
		0.999999,
		1,
		0.999983,
		0.999864,
		0.999493,
		0.998681,
		0.997222,
		0.994912,
		0.991558,
		0.986986,
		0.981047,
		0.973615,
		0.964597,
		0.953921,
		0.941544,
		0.927446,
		0.911631,
		0.894119,
		0.874952,
		0.854185,
		0.831889,
		0.808146,
		0.783049,
		0.756699,
		0.729206,
		0.700686,
		0.67126,
		0.641052,
		0.610192,
		0.57881,
		0.547038,
		0.515011,
		0.482863,
		0.450728,
		0.41874,
		0.387032,
		0.355737,
		0.324985,
		0.294906,
		0.000697,
		0.010606,
		0.030859,
		0.059802,
		0.09592,
		0.137829,
		0.184267,
		0.234088,
		0.286256,
		0.339837,
		0.393997,
		0.447991,
		0.501165,
		0.552945,
		0.602837,
		0.650423,
		0.695352,
		0.737347,
		0.77619,
		0.811728,
		0.843868,
		0.872571,
		0.897854,
		0.919785,
		0.938482,
		0.954108,
		0.966867,
		0.977002,
		0.984791,
		0.990536,
		0.994557,
		0.997186,
		0.99875,
		0.999561,
		0.999898,
		0.999991,
		1,
		0.999998,
		0.999956,
		0.999763,
		0.999253,
		0.998234,
		0.996504,
		0.993865,
		0.990135,
		0.985151,
		0.978775,
		0.970896,
		0.961427,
		0.950308,
		0.937506,
		0.923006,
		0.906817,
		0.888967,
		0.869498,
		0.848471,
		0.825958,
		0.80204,
		0.776812,
		0.750376,
		0.722839,
		0.694316,
		0.664927,
		0.634794,
		0.604044,
		0.572806,
		0.541209,
		0.509384,
		0.477463,
		0.445578,
		0.413858,
		0.382435,
		0.351437,
		0.005983,
		0.022747,
		0.048628,
		0.082095,
		0.121749,
		0.166312,
		0.214622,
		0.265625,
		0.318369,
		0.371999,
		0.425753,
		0.478955,
		0.531011,
		0.581406,
		0.629697,
		0.675515,
		0.718555,
		0.758579,
		0.795408,
		0.828924,
		0.859064,
		0.88582,
		0.909232,
		0.929393,
		0.94644,
		0.960553,
		0.971951,
		0.980889,
		0.987648,
		0.992536,
		0.995871,
		0.997978,
		0.999172,
		0.999746,
		0.999955,
		0.999998,
		1,
		0.999991,
		0.999908,
		0.999618,
		0.998946,
		0.9977,
		0.995681,
		0.9927,
		0.988586,
		0.983187,
		0.976376,
		0.968053,
		0.958141,
		0.946591,
		0.933375,
		0.918486,
		0.901939,
		0.883765,
		0.864011,
		0.842738,
		0.820021,
		0.795942,
		0.770596,
		0.744084,
		0.716513,
		0.687996,
		0.658652,
		0.628601,
		0.597967,
		0.566876,
		0.535457,
		0.503836,
		0.472143,
		0.440506,
		0.409054,
		0.002666,
		0.015846,
		0.038563,
		0.069274,
		0.106564,
		0.149143,
		0.19583,
		0.245554,
		0.297347,
		0.350337,
		0.403741,
		0.456863,
		0.50909,
		0.559887,
		0.608789,
		0.655405,
		0.699408,
		0.740538,
		0.778592,
		0.813428,
		0.84496,
		0.873152,
		0.898023,
		0.919639,
		0.938112,
		0.953597,
		0.966289,
		0.97642,
		0.984253,
		0.990077,
		0.994196,
		0.996928,
		0.998588,
		0.999475,
		0.999864,
		0.999984,
		1,
		0.999999,
		0.999974,
		0.999831,
		0.999422,
		0.998566,
		0.997073,
		0.994749,
		0.991416,
		0.986911,
		0.981094,
		0.973851,
		0.96509,
		0.954745,
		0.942774,
		0.929157,
		0.913893,
		0.897002,
		0.878519,
		0.858494,
		0.83699,
		0.814083,
		0.789856,
		0.764404,
		0.737827,
		0.710231,
		0.681729,
		0.652437,
		0.622473,
		0.59196,
		0.561021,
		0.529781,
		0.498366,
		0.466901,
		0.000668,
		0.010171,
		0.029626,
		0.057479,
		0.092302,
		0.132788,
		0.177742,
		0.226079,
		0.276811,
		0.32905,
		0.381994,
		0.43493,
		0.487226,
		0.538323,
		0.587741,
		0.635065,
		0.679948,
		0.722107,
		0.761317,
		0.797413,
		0.830285,
		0.859876,
		0.886177,
		0.909232,
		0.929128,
		0.945995,
		0.960006,
		0.971369,
		0.980326,
		0.987147,
		0.992122,
		0.995558,
		0.997765,
		0.999046,
		0.999686,
		0.999935,
		0.999996,
		1,
		0.999996,
		0.99994,
		0.999716,
		0.999166,
		0.998106,
		0.996348,
		0.993706,
		0.990011,
		0.985111,
		0.978877,
		0.971204,
		0.962012,
		0.951243,
		0.938862,
		0.924857,
		0.909232,
		0.892012,
		0.873234,
		0.852952,
		0.83123,
		0.808146,
		0.783784,
		0.758237,
		0.731605,
		0.703994,
		0.675515,
		0.646281,
		0.61641,
		0.586022,
		0.555238,
		0.52418,
		0.005737,
		0.021837,
		0.046734,
		0.078987,
		0.117276,
		0.160392,
		0.207232,
		0.256796,
		0.308175,
		0.360554,
		0.413199,
		0.465459,
		0.516759,
		0.566596,
		0.614537,
		0.660215,
		0.703324,
		0.743619,
		0.780913,
		0.815072,
		0.846017,
		0.873716,
		0.898188,
		0.919497,
		0.937751,
		0.953097,
		0.965721,
		0.975845,
		0.983719,
		0.989616,
		0.993829,
		0.996662,
		0.998415,
		0.999381,
		0.999825,
		0.999974,
		0.999999,
		1,
		0.999985,
		0.999882,
		0.999557,
		0.998845,
		0.997561,
		0.995521,
		0.992549,
		0.988485,
		0.983187,
		0.976536,
		0.968439,
		0.958822,
		0.947639,
		0.93486,
		0.92048,
		0.904508,
		0.886971,
		0.867913,
		0.847388,
		0.825463,
		0.802214,
		0.777728,
		0.752097,
		0.72542,
		0.697803,
		0.669354,
		0.640185,
		0.610413,
		0.580154,
		0.002556,
		0.015211,
		0.037057,
		0.066642,
		0.102633,
		0.143807,
		0.189047,
		0.237335,
		0.28775,
		0.339457,
		0.391707,
		0.44383,
		0.495233,
		0.545395,
		0.593863,
		0.64025,
		0.68423,
		0.725537,
		0.763962,
		0.799352,
		0.831602,
		0.860662,
		0.886525,
		0.909232,
		0.928869,
		0.94556,
		0.959469,
		0.970795,
		0.979768,
		0.986646,
		0.991705,
		0.995238,
		0.997543,
		0.99891,
		0.999618,
		0.999911,
		0.999992,
		1,
		0.999998,
		0.999962,
		0.999793,
		0.999346,
		0.998451,
		0.996925,
		0.99459,
		0.991278,
		0.986838,
		0.98114,
		0.974075,
		0.965558,
		0.955526,
		0.943938,
		0.930774,
		0.916031,
		0.899725,
		0.881887,
		0.862562,
		0.841807,
		0.819691,
		0.796291,
		0.771692,
		0.745987,
		0.719275,
		0.691659,
		0.663248,
		0.63415,
		0.000641,
		0.009762,
		0.028466,
		0.055289,
		0.088884,
		0.128015,
		0.171552,
		0.218463,
		0.267809,
		0.318742,
		0.370494,
		0.42238,
		0.473788,
		0.52418,
		0.573084,
		0.620093,
		0.664862,
		0.707107,
		0.746597,
		0.783156,
		0.816662,
		0.84704,
		0.874262,
		0.898348,
		0.919359,
		0.937398,
		0.952607,
		0.965164,
		0.975277,
		0.983187,
		0.989154,
		0.993458,
		0.996388,
		0.998234,
		0.999278,
		0.999779,
		0.999961,
		0.999998,
		1,
		0.999992,
		0.99992,
		0.999666,
		0.999076,
		0.997978,
		0.996194,
		0.993551,
		0.989891,
		0.985072,
		0.978974,
		0.971498,
		0.962567,
		0.952128,
		0.940146,
		0.926607,
		0.911514,
		0.894888,
		0.876762,
		0.857183,
		0.836212,
		0.813918,
		0.790377,
		0.765677,
		0.739908,
		0.71317,
		0.685564,
		0.005506,
		0.02098,
		0.044948,
		0.076051,
		0.113042,
		0.154777,
		0.200209,
		0.248386,
		0.298442,
		0.349596,
		0.401146,
		0.452462,
		0.502989,
		0.552238,
		0.599783,
		0.645261,
		0.688367,
		0.728851,
		0.76652,
		0.801227,
		0.832877,
		0.861423,
		0.886861,
		0.909232,
		0.928617,
		0.945135,
		0.958943,
		0.97023,
		0.979216,
		0.986146,
		0.991285,
		0.994912,
		0.997312,
		0.998766,
		0.999542,
		0.999882,
		0.999986,
		1,
		1,
		0.999977,
		0.999852,
		0.999493,
		0.998741,
		0.997422,
		0.995364,
		0.992403,
		0.988388,
		0.983187,
		0.976689,
		0.968806,
		0.959469,
		0.948631,
		0.936266,
		0.922365,
		0.906935,
		0.890001,
		0.8716,
		0.851782,
		0.830607,
		0.808146,
		0.784477,
		0.759685,
		0.733863,
		0.002453,
		0.014613,
		0.035637,
		0.064157,
		0.098913,
		0.138748,
		0.182603,
		0.229511,
		0.278592,
		0.32905,
		0.380165,
		0.431293,
		0.481862,
		0.531365,
		0.579359,
		0.625463,
		0.669354,
		0.710763,
		0.749475,
		0.785326,
		0.818201,
		0.848031,
		0.874792,
		0.898503,
		0.919224,
		0.937054,
		0.952128,
		0.964615,
		0.974716,
		0.982658,
		0.988691,
		0.993081,
		0.996106,
		0.998044,
		0.999166,
		0.999726,
		0.999944,
		0.999996,
		1,
		0.999996,
		0.999947,
		0.999752,
		0.999268,
		0.998334,
		0.996778,
		0.994434,
		0.991144,
		0.986769,
		0.981184,
		0.974289,
		0.966004,
		0.956268,
		0.945041,
		0.932304,
		0.918052,
		0.902298,
		0.88507,
		0.866407,
		0.846361,
		0.824995,
		0.802379,
		0.778592,
		0.000615,
		0.009378,
		0.027373,
		0.053221,
		0.08565,
		0.123492,
		0.165674,
		0.211216,
		0.259225,
		0.308888,
		0.359472,
		0.410318,
		0.460836,
		0.510503,
		0.55886,
		0.605508,
		0.650105,
		0.692366,
		0.732056,
		0.768993,
		0.803041,
		0.834112,
		0.862161,
		0.887188,
		0.909232,
		0.928371,
		0.944719,
		0.958426,
		0.969673,
		0.978669,
		0.985648,
		0.990862,
		0.99458,
		0.997073,
		0.998613,
		0.999459,
		0.999847,
		0.999977,
		1,
		1,
		0.999987,
		0.999897,
		0.999612,
		0.998984,
		0.99785,
		0.996042,
		0.9934,
		0.989775,
		0.985035,
		0.979066,
		0.971777,
		0.963095,
		0.952968,
		0.941362,
		0.928264,
		0.913673,
		0.897607,
		0.880097,
		0.861185,
		0.840925,
		0.819379,
		0.005289,
		0.020173,
		0.043262,
		0.073275,
		0.109032,
		0.149448,
		0.193529,
		0.24037,
		0.289144,
		0.339103,
		0.389573,
		0.439948,
		0.48969,
		0.538323,
		0.585432,
		0.630658,
		0.673697,
		0.714298,
		0.752258,
		0.787425,
		0.819691,
		0.848992,
		0.875306,
		0.898654,
		0.919093,
		0.936718,
		0.951659,
		0.964076,
		0.974162,
		0.982133,
		0.988228,
		0.9927,
		0.995817,
		0.997845,
		0.999046,
		0.999666,
		0.999923,
		0.999993,
		1,
		0.999999,
		0.999967,
		0.999819,
		0.999426,
		0.998635,
		0.997284,
		0.995211,
		0.992261,
		0.988294,
		0.983187,
		0.976835,
		0.969156,
		0.960083,
		0.949573,
		0.937598,
		0.924149,
		0.909232,
		0.892867,
		0.875088,
		0.855939,
		0.002356,
		0.014049,
		0.034297,
		0.061808,
		0.095391,
		0.133949,
		0.176478,
		0.222058,
		0.269849,
		0.319089,
		0.369091,
		0.419232,
		0.468961,
		0.517785,
		0.565272,
		0.611048,
		0.654791,
		0.696234,
		0.735156,
		0.771385,
		0.804797,
		0.835308,
		0.862877,
		0.887506,
		0.909232,
		0.928131,
		0.944313,
		0.95792,
		0.969125,
		0.978128,
		0.985151,
		0.990437,
		0.994242,
		0.996826,
		0.998451,
		0.999367,
		0.999806,
		0.999966,
		0.999999,
		1,
		0.999993,
		0.99993,
		0.999707,
		0.999187,
		0.998216,
		0.996633,
		0.994282,
		0.991015,
		0.986701,
		0.981226,
		0.974494,
		0.966428,
		0.956972,
		0.946088,
		0.933754,
		0.919966,
		0.904734,
		0.888082,
		0.00059,
		0.009016,
		0.026342,
		0.051267,
		0.082589,
		0.119202,
		0.160089,
		0.204316,
		0.251033,
		0.299464,
		0.348906,
		0.398725,
		0.448352,
		0.497281,
		0.545064,
		0.591311,
		0.635684,
		0.677898,
		0.717717,
		0.754952,
		0.789457,
		0.821134,
		0.849923,
		0.875806,
		0.898801,
		0.918966,
		0.93639,
		0.951199,
		0.963547,
		0.973615,
		0.981612,
		0.987764,
		0.992316,
		0.995521,
		0.997637,
		0.998917,
		0.999599,
		0.999897,
		0.999988,
		1,
		1,
		0.99998,
		0.99987,
		0.999554,
		0.998889,
		0.997721,
		0.995893,
		0.993253,
		0.989663,
		0.984999,
		0.979155,
		0.972044,
		0.963598,
		0.953766,
		0.942516,
		0.929834,
		0.915718,
		0.005084,
		0.019411,
		0.041669,
		0.070648,
		0.105229,
		0.144385,
		0.187172,
		0.232725,
		0.280256,
		0.32905,
		0.378457,
		0.427896,
		0.476846,
		0.524843,
		0.571482,
		0.61641,
		0.659326,
		0.699976,
		0.738155,
		0.773702,
		0.806498,
		0.836467,
		0.863572,
		0.887815,
		0.909232,
		0.927897,
		0.943915,
		0.957423,
		0.968585,
		0.977591,
		0.984657,
		0.990011,
		0.993899,
		0.996571,
		0.99828,
		0.999268,
		0.99976,
		0.999951,
		0.999997,
		1,
		0.999997,
		0.999954,
		0.999782,
		0.999355,
		0.998528,
		0.997147,
		0.99506,
		0.992122,
		0.988203,
		0.983187,
		0.976975,
		0.969489,
		0.960668,
		0.950468,
		0.938862,
		0.002265,
		0.013517,
		0.033031,
		0.059585,
		0.092052,
		0.129392,
		0.170651,
		0.214952,
		0.261496,
		0.309553,
		0.358463,
		0.407628,
		0.456513,
		0.504643,
		0.551596,
		0.597004,
		0.640549,
		0.681965,
		0.721027,
		0.757559,
		0.791425,
		0.822533,
		0.850827,
		0.87629,
		0.898944,
		0.918841,
		0.93607,
		0.950749,
		0.963026,
		0.973075,
		0.981094,
		0.987301,
		0.991928,
		0.995219,
		0.997422,
		0.99878,
		0.999525,
		0.999866,
		0.99998,
		1,
		1,
		0.999989,
		0.999909,
		0.999658,
		0.999103,
		0.998097,
		0.99649,
		0.994133,
		0.99089,
		0.986637,
		0.981267,
		0.974689,
		0.966832,
		0.957642,
		0.000567,
		0.008674,
		0.025367,
		0.049418,
		0.079688,
		0.11513,
		0.154777,
		0.197741,
		0.243212,
		0.290446,
		0.338772,
		0.387578,
		0.436317,
		0.484498,
		0.531686,
		0.577498,
		0.621603,
		0.663715,
		0.703599,
		0.741059,
		0.775945,
		0.808146,
		0.837591,
		0.864247,
		0.888115,
		0.909232,
		0.927669,
		0.943527,
		0.956935,
		0.968053,
		0.977061,
		0.984164,
		0.989583,
		0.993551,
		0.99631,
		0.998102,
		0.999161,
		0.999707,
		0.999932,
		0.999994,
		1,
		0.999999,
		0.999971,
		0.99984,
		0.999493,
		0.998793,
		0.997593,
		0.995745,
		0.993109,
		0.989554,
		0.984964,
		0.97924,
		0.972299,
		0.004891,
		0.018691,
		0.040162,
		0.068159,
		0.101621,
		0.139573,
		0.181117,
		0.225429,
		0.271757,
		0.319414,
		0.367779,
		0.416289,
		0.464441,
		0.511786,
		0.557927,
		0.602518,
		0.645261,
		0.685901,
		0.724231,
		0.760084,
		0.793332,
		0.823889,
		0.851704,
		0.876762,
		0.899083,
		0.91872,
		0.935757,
		0.950308,
		0.962515,
		0.972543,
		0.980581,
		0.986838,
		0.991538,
		0.994912,
		0.997199,
		0.998635,
		0.999443,
		0.99983,
		0.99997,
		0.999999,
		1,
		0.999994,
		0.999938,
		0.999741,
		0.999281,
		0.99842,
		0.997011,
		0.994912,
		0.991988,
		0.988116,
		0.983187,
		0.002179,
		0.013016,
		0.031834,
		0.057479,
		0.088884,
		0.125061,
		0.165103,
		0.208175,
		0.253513,
		0.300419,
		0.34826,
		0.39646,
		0.444504,
		0.491928,
		0.538323,
		0.583329,
		0.626633,
		0.667967,
		0.707107,
		0.743871,
		0.778118,
		0.809744,
		0.838682,
		0.864902,
		0.888407,
		0.909232,
		0.927446,
		0.943146,
		0.956457,
		0.967529,
		0.976536,
		0.983674,
		0.989154,
		0.993199,
		0.996042,
		0.997915,
		0.999046,
		0.999647,
		0.999909,
		0.999989,
		1,
		1,
		0.999982,
		0.999886,
		0.999606,
		0.999017,
		0.997978,
		0.996348,
		0.993987,
		0.990768,
		0.000546,
		0.008351,
		0.024446,
		0.047667,
		0.076937,
		0.111261,
		0.149721,
		0.191473,
		0.23574,
		0.281814,
		0.32905,
		0.37686,
		0.424716,
		0.472143,
		0.518718,
		0.564067,
		0.607863,
		0.649825,
		0.689715,
		0.727336,
		0.762531,
		0.795181,
		0.825205,
		0.852555,
		0.877219,
		0.899218,
		0.918602,
		0.935451,
		0.949876,
		0.962012,
		0.972017,
		0.980072,
		0.986377,
		0.991144,
		0.994599,
		0.99697,
		0.998482,
		0.999355,
		0.999788,
		0.999957,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.999807,
		0.999429,
		0.998695,
		0.997465,
		0.995601,
		0.004709,
		0.018011,
		0.038736,
		0.065798,
		0.098193,
		0.134994,
		0.175346,
		0.218463,
		0.263625,
		0.310176,
		0.357518,
		0.405109,
		0.452462,
		0.499142,
		0.544762,
		0.588982,
		0.631507,
		0.672085,
		0.710506,
		0.746597,
		0.780225,
		0.811294,
		0.839741,
		0.865538,
		0.88869,
		0.909232,
		0.927229,
		0.942774,
		0.955987,
		0.967013,
		0.976017,
		0.983187,
		0.988724,
		0.992844,
		0.995768,
		0.997721,
		0.998923,
		0.999581,
		0.999882,
		0.999982,
		1,
		1,
		0.99999,
		0.99992,
		0.999698,
		0.999205,
		0.99831,
		0.002098,
		0.012541,
		0.0307,
		0.055483,
		0.085875,
		0.120941,
		0.159817,
		0.201706,
		0.245878,
		0.291667,
		0.338462,
		0.385711,
		0.432915,
		0.479627,
		0.525446,
		0.570021,
		0.613044,
		0.654248,
		0.69341,
		0.730344,
		0.764902,
		0.796974,
		0.826481,
		0.853382,
		0.877665,
		0.89935,
		0.918486,
		0.935153,
		0.949453,
		0.961517,
		0.971498,
		0.979567,
		0.985916,
		0.990749,
		0.994282,
		0.996733,
		0.998321,
		0.999259,
		0.999741,
		0.99994,
		0.999995,
		1,
		0.999999,
		0.999974,
		0.999859,
		0.999551,
		0.000525,
		0.008046,
		0.023574,
		0.046007,
		0.074325,
		0.107582,
		0.144906,
		0.185491,
		0.228598,
		0.273547,
		0.319719,
		0.366551,
		0.413531,
		0.460201,
		0.506149,
		0.55101,
		0.594464,
		0.636233,
		0.676077,
		0.7138,
		0.749239,
		0.782268,
		0.812798,
		0.840769,
		0.866157,
		0.888967,
		0.909232,
		0.927017,
		0.94241,
		0.955526,
		0.966504,
		0.975504,
		0.982702,
		0.988294,
		0.992484,
		0.995488,
		0.99752,
		0.998793,
		0.999508,
		0.99985,
		0.999973,
		0.999999,
		1,
		0.999995,
		0.999945,
		0.004537,
		0.017367,
		0.037384,
		0.063557,
		0.094935,
		0.130634,
		0.169842,
		0.211807,
		0.255841,
		0.301315,
		0.347655,
		0.394338,
		0.440894,
		0.4869,
		0.531979,
		0.575799,
		0.618068,
		0.658536,
		0.696992,
		0.73326,
		0.767202,
		0.798713,
		0.827721,
		0.854185,
		0.878098,
		0.899478,
		0.918374,
		0.93486,
		0.949038,
		0.961031,
		0.970985,
		0.979066,
		0.985457,
		0.990352,
		0.993959,
		0.99649,
		0.998153,
		0.999156,
		0.999688,
		0.99992,
		0.99999,
		1,
		1,
		0.002021,
		0.012092,
		0.029626,
		0.053588,
		0.083016,
		0.11702,
		0.154777,
		0.195527,
		0.238574,
		0.283276,
		0.32905,
		0.375363,
		0.421733,
		0.467727,
		0.512956,
		0.557076,
		0.599783,
		0.640815,
		0.679948,
		0.716994,
		0.751801,
		0.78425,
		0.814257,
		0.841768,
		0.866759,
		0.889236,
		0.909232,
		0.926809,
		0.942053,
		0.955074,
		0.966004,
		0.974996,
		0.98222,
		0.987863,
		0.992122,
		0.995203,
		0.997312,
		0.998655,
		0.999429,
		0.999813,
		0.999962,
		0.999998,
		0.000506,
		0.007758,
		0.022747,
		0.044432,
		0.071843,
		0.104082,
		0.140317,
		0.179781,
		0.221767,
		0.265625,
		0.310761,
		0.356632,
		0.402746,
		0.448658,
		0.493969,
		0.538323,
		0.581406,
		0.622942,
		0.662695,
		0.700466,
		0.736089,
		0.769433,
		0.800401,
		0.828924,
		0.854966,
		0.878519,
		0.899603,
		0.918264,
		0.934575,
		0.948631,
		0.960553,
		0.97048,
		0.97857,
		0.984999,
		0.989954,
		0.993633,
		0.99624,
		0.997978,
		0.999046,
		0.999629,
		0.999896,
		0.004374,
		0.016757,
		0.036101,
		0.061429,
		0.091835,
		0.12648,
		0.164588,
		0.205443,
		0.248386,
		0.292812,
		0.338171,
		0.383959,
		0.429721,
		0.475047,
		0.51957,
		0.562964,
		0.604944,
		0.645261,
		0.683702,
		0.720092,
		0.754286,
		0.786174,
		0.815675,
		0.842738,
		0.867344,
		0.889497,
		0.909232,
		0.926607,
		0.941704,
		0.95463,
		0.965511,
		0.974494,
		0.981742,
		0.987433,
		0.991758,
		0.994912,
		0.997097,
		0.998511,
		0.999343,
		0.001948,
		0.011667,
		0.028608,
		0.051789,
		0.080297,
		0.113285,
		0.149968,
		0.189623,
		0.23158,
		0.275228,
		0.320005,
		0.365398,
		0.410941,
		0.456215,
		0.500843,
		0.544486,
		0.586849,
		0.627672,
		0.66673,
		0.703836,
		0.738833,
		0.771599,
		0.80204,
		0.830094,
		0.855726,
		0.878929,
		0.899725,
		0.918157,
		0.934295,
		0.948233,
		0.960083,
		0.969981,
		0.978079,
		0.984543,
		0.989554,
		0.993303,
		0.995985,
		0.997796,
		0.000488,
		0.007484,
		0.021964,
		0.042936,
		0.069482,
		0.100747,
		0.135939,
		0.174326,
		0.21523,
		0.258031,
		0.302157,
		0.347086,
		0.392345,
		0.437501,
		0.482168,
		0.525998,
		0.568684,
		0.609954,
		0.649575,
		0.687346,
		0.723099,
		0.756699,
		0.788042,
		0.817052,
		0.843682,
		0.867913,
		0.889753,
		0.909232,
		0.926409,
		0.941362,
		0.954194,
		0.965026,
		0.973997,
		0.981267,
		0.987004,
		0.99139,
		0.994617,
		0.004219,
		0.016179,
		0.034883,
		0.059405,
		0.088884,
		0.122519,
		0.159571,
		0.199356,
		0.241242,
		0.28465,
		0.32905,
		0.373956,
		0.418929,
		0.463572,
		0.507526,
		0.550474,
		0.592135,
		0.632263,
		0.670646,
		0.707107,
		0.741497,
		0.773702,
		0.803632,
		0.83123,
		0.856465,
		0.879329,
		0.899843,
		0.918052,
		0.934022,
		0.947842,
		0.959621,
		0.969489,
		0.977591,
		0.984089,
		0.989154,
		0.001879,
		0.011263,
		0.027641,
		0.050078,
		0.077708,
		0.109724,
		0.145377,
		0.183976,
		0.224882,
		0.267506,
		0.31131,
		0.355799,
		0.400525,
		0.44508,
		0.489096,
		0.532247,
		0.57424,
		0.61482,
		0.653763,
		0.690882,
		0.726017,
		0.759042,
		0.789856,
		0.81839,
		0.8446,
		0.868468,
		0.890001,
		0.909232,
		0.926215,
		0.941027,
		0.953766,
		0.964547,
		0.973507,
		0.980795,
		0.000471,
		0.007225,
		0.02122,
		0.041515,
		0.067236,
		0.09757,
		0.131761,
		0.169111,
		0.208972,
		0.250747,
		0.29389,
		0.337898,
		0.382312,
		0.426716,
		0.470734,
		0.514027,
		0.556295,
		0.59727,
		0.636722,
		0.674449,
		0.710282,
		0.744084,
		0.775744,
		0.805179,
		0.832336,
		0.857183,
		0.879718,
		0.899959,
		0.91795,
		0.933754,
		0.947458,
		0.959167,
		0.969004,
		0.004073,
		0.01563,
		0.033726,
		0.057479,
		0.086071,
		0.118739,
		0.154777,
		0.193529,
		0.234393,
		0.276811,
		0.320274,
		0.364313,
		0.408505,
		0.452462,
		0.495839,
		0.538323,
		0.57964,
		0.619546,
		0.657831,
		0.694316,
		0.728851,
		0.761317,
		0.791619,
		0.819691,
		0.845492,
		0.869007,
		0.890244,
		0.909232,
		0.926026,
		0.940699,
		0.953345,
		0.001814,
		0.01088,
		0.026722,
		0.048451,
		0.075242,
		0.106327,
		0.140991,
		0.178574,
		0.218463,
		0.260094,
		0.302949,
		0.346552,
		0.390469,
		0.434306,
		0.477707,
		0.520351,
		0.561953,
		0.60226,
		0.641052,
		0.678141,
		0.713366,
		0.746597,
		0.777728,
		0.806683,
		0.833411,
		0.857883,
		0.880097,
		0.900072,
		0.91785,
		0.933492,
		0.000454,
		0.006979,
		0.020513,
		0.040162,
		0.065096,
		0.094539,
		0.127771,
		0.164123,
		0.202976,
		0.243759,
		0.285944,
		0.32905,
		0.372632,
		0.416289,
		0.459655,
		0.502402,
		0.544233,
		0.584888,
		0.624138,
		0.661782,
		0.697651,
		0.731605,
		0.763528,
		0.793332,
		0.820956,
		0.846361,
		0.869533,
		0.89048,
		0.909232,
		0.003934,
		0.015108,
		0.032625,
		0.055645,
		0.083389,
		0.11513,
		0.150192,
		0.18795,
		0.227824,
		0.26928,
		0.311827,
		0.355015,
		0.398433,
		0.441707,
		0.484498,
		0.526505,
		0.567455,
		0.607109,
		0.645261,
		0.681729,
		0.716363,
		0.749038,
		0.779657,
		0.808146,
		0.834457,
		0.858565,
		0.880467,
		0.001752,
		0.010517,
		0.025848,
		0.046901,
		0.072891,
		0.103084,
		0.136799,
		0.173402,
		0.212308,
		0.252975,
		0.294906,
		0.33764,
		0.38076,
		0.423884,
		0.466665,
		0.50879,
		0.549982,
		0.589991,
		0.628601,
		0.665622,
		0.700893,
		0.73428,
		0.765677,
		0.794999,
		0.822187,
		0.847207,
		0.000439,
		0.006746,
		0.019841,
		0.038875,
		0.063057,
		0.091646,
		0.123957,
		0.159349,
		0.197229,
		0.237049,
		0.278304,
		0.320527,
		0.363292,
		0.406208,
		0.448922,
		0.491113,
		0.532493,
		0.572806,
		0.611825,
		0.649351,
		0.685216,
		0.719275,
		0.751412,
		0.781532,
		0.809569,
		0.003803,
		0.014613,
		0.031577,
		0.053897,
		0.08083,
		0.111681,
		0.145806,
		0.182603,
		0.22152,
		0.262041,
		0.303696,
		0.346048,
		0.388702,
		0.431293,
		0.473495,
		0.515011,
		0.555576,
		0.594955,
		0.63294,
		0.669354,
		0.704043,
		0.736881,
		0.767766,
		0.001693,
		0.010171,
		0.025016,
		0.045425,
		0.070648,
		0.099986,
		0.132788,
		0.168448,
		0.206404,
		0.246136,
		0.287165,
		0.32905,
		0.371384,
		0.413799,
		0.455958,
		0.497557,
		0.538323,
		0.578013,
		0.61641,
		0.653328,
		0.688606,
		0.722107,
		0.000424,
		0.006524,
		0.019201,
		0.037648,
		0.061111,
		0.088884,
		0.120309,
		0.154777,
		0.191718,
		0.230606,
		0.270954,
		0.312315,
		0.354276,
		0.39646,
		0.438523,
		0.480153,
		0.52107,
		0.561021,
		0.599783,
		0.63716,
		0.672983,
		0.003677,
		0.014141,
		0.030578,
		0.05223,
		0.078386,
		0.108384,
		0.141606,
		0.177478,
		0.215467,
		0.255081,
		0.295864,
		0.337397,
		0.379296,
		0.42121,
		0.46282,
		0.503836,
		0.544,
		0.583079,
		0.620871,
		0.001638,
		0.009842,
		0.024223,
		0.044016,
		0.068506,
		0.097025,
		0.128949,
		0.163699,
		0.200737,
		0.239562,
		0.279714,
		0.320766,
		0.362328,
		0.40404,
		0.445578,
		0.486644,
		0.526971,
		0.566321,
		0.00041,
		0.006313,
		0.018592,
		0.036479,
		0.059254,
		0.086243,
		0.116819,
		0.150396,
		0.18643,
		0.224414,
		0.263882,
		0.304401,
		0.345573,
		0.387032,
		0.428447,
		0.469512,
		0.509955,
		0.003558,
		0.013691,
		0.029626,
		0.050639,
		0.076051,
		0.105229,
		0.137583,
		0.172561,
		0.209654,
		0.248386,
		0.288319,
		0.32905,
		0.370205,
		0.411446,
		0.452462,
		0.001584,
		0.009529,
		0.023468,
		0.042672,
		0.06646,
		0.094192,
		0.125272,
		0.159146,
		0.195295,
		0.233241,
		0.272538,
		0.312776,
		0.353578,
		0.394596,
		0.000397,
		0.006111,
		0.018011,
		0.035363,
		0.057479,
		0.083718,
		0.113477,
		0.146196,
		0.181354,
		0.218463,
		0.257074,
		0.29677,
		0.337168,
		0.003444,
		0.013263,
		0.028718,
		0.049119,
		0.073818,
		0.102209,
		0.133727,
		0.167843,
		0.204067,
		0.241943,
		0.281048,
		0.001534,
		0.00923,
		0.022747,
		0.041389,
		0.064503,
		0.09148,
		0.121749,
		0.154777,
		0.190067,
		0.227159,
		0.000384,
		0.00592,
		0.017457,
		0.034297,
		0.055783,
		0.081301,
		0.110275,
		0.142168,
		0.176478,
		0.003336,
		0.012854,
		0.027851,
		0.047667,
		0.071682,
		0.099316,
		0.130029,
		0.001486,
		0.008946,
		0.022059,
		0.040162,
		0.062632,
		0.088884,
		0.000372,
		0.005737,
		0.016928,
		0.033279,
		0.054161,
		0.003233,
		0.012464,
		0.027023,
		0.00144,
		0.008674,
		0.000361,
		0.005562,
		0.000807,
		0.021402,
		0.010445,
		0.003208,
		9.2e-005,
		0.046277,
		0.029959,
		0.0168,
		0.00717,
		0.001462,
		0.078987,
		0.058111,
		0.039864,
		0.024591,
		0.012657,
		0.004448,
		0.000375,
		0.118371,
		0.093699,
		0.071162,
		0.051077,
		0.033782,
		0.019635,
		0.009016,
		0.002328,
		0.163312,
		0.135567,
		0.109496,
		0.085389,
		0.063557,
		0.044332,
		0.028064,
		0.015128,
		0.00592,
		0.00086,
		0.212739,
		0.182603,
		0.153712,
		0.126332,
		0.100747,
		0.077262,
		0.0562,
		0.037906,
		0.022747,
		0.011113,
		0.003417,
		9.8e-005,
		0.265625,
		0.233741,
		0.202704,
		0.172758,
		0.144161,
		0.117192,
		0.092147,
		0.069344,
		0.049119,
		0.031834,
		0.01787,
		0.007634,
		0.001559,
		0.320992,
		0.28796,
		0.255412,
		0.223565,
		0.192655,
		0.162935,
		0.134676,
		0.108167,
		0.083718,
		0.06166,
		0.042346,
		0.026151,
		0.013475,
		0.004741,
		0.0004,
		0.377913,
		0.344293,
		0.310823,
		0.2777,
		0.245133,
		0.213353,
		0.182603,
		0.153149,
		0.125272,
		0.099276,
		0.075484,
		0.05424,
		0.035914,
		0.020897,
		0.009606,
		0.002483,
		0.435513,
		0.401822,
		0.36798,
		0.33416,
		0.300551,
		0.267356,
		0.234798,
		0.203114,
		0.172561,
		0.143415,
		0.11597,
		0.090543,
		0.067471,
		0.047116,
		0.02986,
		0.016114,
		0.006313,
		0.000918,
		0.492972,
		0.459685,
		0.425977,
		0.391998,
		0.357915,
		0.32391,
		0.290179,
		0.256937,
		0.224414,
		0.19286,
		0.162543,
		0.13375,
		0.106789,
		0.081992,
		0.05971,
		0.04032,
		0.024223,
		0.011847,
		0.003647,
		0.000105,
		0.549528,
		0.517077,
		0.483964,
		0.45032,
		0.416289,
		0.382031,
		0.34772,
		0.313545,
		0.279714,
		0.246449,
		0.213992,
		0.182603,
		0.152563,
		0.124172,
		0.097753,
		0.073649,
		0.05223,
		0.033889,
		0.019046,
		0.008146,
		0.001665,
		0.604481,
		0.573252,
		0.541152,
		0.508291,
		0.474793,
		0.440796,
		0.40645,
		0.371923,
		0.337397,
		0.303071,
		0.269161,
		0.2359,
		0.203541,
		0.172357,
		0.142639,
		0.114702,
		0.088884,
		0.065543,
		0.045066,
		0.027864,
		0.014374,
		0.005063,
		0.000428,
		0.657196,
		0.62753,
		0.596815,
		0.56514,
		0.532609,
		0.499339,
		0.465459,
		0.431113,
		0.39646,
		0.361676,
		0.32695,
		0.292492,
		0.258527,
		0.2253,
		0.193074,
		0.162135,
		0.132788,
		0.105362,
		0.080209,
		0.057705,
		0.038254,
		0.022285,
		0.010256,
		0.002655,
		0.707107,
		0.679298,
		0.650291,
		0.620158,
		0.588982,
		0.556859,
		0.523896,
		0.490217,
		0.455958,
		0.42127,
		0.38632,
		0.351291,
		0.316383,
		0.281814,
		0.247821,
		0.214658,
		0.182603,
		0.151954,
		0.12303,
		0.096175,
		0.071756,
		0.050169,
		0.031834,
		0.0172,
		0.006746,
		0.000982,
		0.753719,
		0.728012,
		0.700991,
		0.672708,
		0.643226,
		0.612621,
		0.58098,
		0.548406,
		0.515011,
		0.480925,
		0.446292,
		0.411269,
		0.376031,
		0.340772,
		0.3057,
		0.271043,
		0.237049,
		0.203986,
		0.172143,
		0.141832,
		0.113386,
		0.087167,
		0.063557,
		0.042971,
		0.025848,
		0.012657,
		0.003901,
		0.000112,
		0.796619,
		0.77321,
		0.7484,
		0.722224,
		0.694725,
		0.66596,
		0.635997,
		0.604914,
		0.572806,
		0.539778,
		0.505952,
		0.471461,
		0.436458,
		0.401108,
		0.365596,
		0.330122,
		0.294906,
		0.260186,
		0.226223,
		0.193297,
		0.16171,
		0.131787,
		0.103881,
		0.078366,
		0.055645,
		0.03615,
		0.020342,
		0.008711,
		0.001783,
		0.835475,
		0.814507,
		0.792083,
		0.76822,
		0.742944,
		0.71629,
		0.688306,
		0.659052,
		0.628601,
		0.597036,
		0.564456,
		0.530974,
		0.496716,
		0.461824,
		0.426457,
		0.39079,
		0.355015,
		0.319344,
		0.284006,
		0.249252,
		0.215353,
		0.182603,
		0.151319,
		0.121842,
		0.094539,
		0.069803,
		0.048056,
		0.02975,
		0.015366,
		0.005419,
		0.000458,
		0.870045,
		0.851607,
		0.831692,
		0.810295,
		0.787425,
		0.7631,
		0.737347,
		0.710206,
		0.681729,
		0.651979,
		0.621034,
		0.588982,
		0.555928,
		0.52199,
		0.487302,
		0.452013,
		0.416289,
		0.380315,
		0.344293,
		0.308444,
		0.273008,
		0.238249,
		0.204451,
		0.171921,
		0.140991,
		0.112019,
		0.085389,
		0.061512,
		0.04083,
		0.023816,
		0.010974,
		0.002844,
		0.900183,
		0.88431,
		0.866967,
		0.848135,
		0.827802,
		0.805968,
		0.782641,
		0.757843,
		0.731605,
		0.703969,
		0.67499,
		0.644738,
		0.613292,
		0.580749,
		0.547218,
		0.512824,
		0.477707,
		0.442026,
		0.405954,
		0.369686,
		0.333432,
		0.297425,
		0.261919,
		0.227188,
		0.193529,
		0.161266,
		0.130747,
		0.102344,
		0.076461,
		0.053528,
		0.034009,
		0.018399,
		0.007225,
		0.001053,
		0.925841,
		0.91251,
		0.897749,
		0.881522,
		0.863799,
		0.844563,
		0.823803,
		0.801521,
		0.777728,
		0.752448,
		0.725714,
		0.697574,
		0.668086,
		0.637323,
		0.605372,
		0.572334,
		0.538323,
		0.503473,
		0.467931,
		0.431863,
		0.395453,
		0.358903,
		0.322436,
		0.286296,
		0.250747,
		0.216079,
		0.182603,
		0.150658,
		0.120607,
		0.092843,
		0.067787,
		0.045892,
		0.027641,
		0.013553,
		0.004182,
		0.00012,
		0.947082,
		0.936212,
		0.923982,
		0.910341,
		0.895243,
		0.878652,
		0.860539,
		0.840888,
		0.819691,
		0.796951,
		0.772681,
		0.746909,
		0.71967,
		0.691017,
		0.661012,
		0.629732,
		0.59727,
		0.563732,
		0.529241,
		0.493936,
		0.457972,
		0.421525,
		0.384786,
		0.34797,
		0.31131,
		0.275061,
		0.239503,
		0.204936,
		0.171689,
		0.140116,
		0.110599,
		0.083548,
		0.059405,
		0.038645,
		0.021774,
		0.009337,
		0.001913,
		0.964076,
		0.955526,
		0.945719,
		0.934586,
		0.922066,
		0.908106,
		0.892661,
		0.875696,
		0.857183,
		0.837108,
		0.815463,
		0.792253,
		0.767497,
		0.741222,
		0.713469,
		0.684294,
		0.653763,
		0.62196,
		0.588982,
		0.554942,
		0.519969,
		0.48421,
		0.447829,
		0.41101,
		0.373956,
		0.336891,
		0.300059,
		0.26373,
		0.228195,
		0.193772,
		0.160804,
		0.129662,
		0.100747,
		0.07449,
		0.051353,
		0.031834,
		0.016464,
		0.005814,
		0.000492,
		0.977109,
		0.970681,
		0.963126,
		0.954362,
		0.944313,
		0.932909,
		0.92009,
		0.905802,
		0.890001,
		0.872652,
		0.853729,
		0.833218,
		0.811114,
		0.787425,
		0.762171,
		0.735383,
		0.707107,
		0.6774,
		0.646337,
		0.614003,
		0.580504,
		0.545959,
		0.510503,
		0.474293,
		0.437501,
		0.40032,
		0.362964,
		0.325668,
		0.28869,
		0.252311,
		0.216838,
		0.182603,
		0.149968,
		0.119322,
		0.091084,
		0.065706,
		0.043674,
		0.02551,
		0.011771,
		0.003055,
		0.986574,
		0.982017,
		0.976488,
		0.969893,
		0.962145,
		0.95316,
		0.942861,
		0.931179,
		0.918052,
		0.903427,
		0.88726,
		0.869516,
		0.850172,
		0.829215,
		0.806641,
		0.782462,
		0.756699,
		0.729388,
		0.700578,
		0.670332,
		0.638727,
		0.605858,
		0.571833,
		0.53678,
		0.500843,
		0.464184,
		0.426988,
		0.389456,
		0.351814,
		0.314308,
		0.277209,
		0.240813,
		0.205443,
		0.171447,
		0.139203,
		0.109121,
		0.08164,
		0.057235,
		0.036415,
		0.019728,
		0.007758,
		0.001133,
		0.992969,
		0.989983,
		0.986201,
		0.981521,
		0.975845,
		0.969079,
		0.961131,
		0.951918,
		0.941362,
		0.929393,
		0.91595,
		0.900978,
		0.884434,
		0.866286,
		0.84651,
		0.825095,
		0.80204,
		0.777359,
		0.751077,
		0.723232,
		0.693878,
		0.663083,
		0.63093,
		0.597518,
		0.562964,
		0.527402,
		0.490984,
		0.453882,
		0.416289,
		0.378419,
		0.340507,
		0.302814,
		0.265625,
		0.229249,
		0.194026,
		0.160321,
		0.128533,
		0.099088,
		0.072452,
		0.049119,
		0.029626,
		0.014547,
		0.004495,
		0.00013,
		0.996876,
		0.995126,
		0.992769,
		0.9897,
		0.985814,
		0.981008,
		0.975181,
		0.968237,
		0.960083,
		0.950636,
		0.939815,
		0.927551,
		0.913781,
		0.898452,
		0.881522,
		0.862958,
		0.842738,
		0.820854,
		0.797307,
		0.772112,
		0.745299,
		0.71691,
		0.687003,
		0.655651,
		0.622942,
		0.588982,
		0.553895,
		0.517822,
		0.480925,
		0.443386,
		0.405406,
		0.367211,
		0.32905,
		0.291195,
		0.253947,
		0.217631,
		0.182603,
		0.149248,
		0.117983,
		0.089257,
		0.063557,
		0.041405,
		0.023363,
		0.010032,
		0.002059,
		0.998929,
		0.998065,
		0.996784,
		0.994985,
		0.992561,
		0.989406,
		0.985413,
		0.980477,
		0.974494,
		0.967366,
		0.959001,
		0.949311,
		0.938217,
		0.925648,
		0.911542,
		0.895847,
		0.878519,
		0.859528,
		0.838854,
		0.816488,
		0.792437,
		0.766717,
		0.739362,
		0.710418,
		0.679948,
		0.648029,
		0.614757,
		0.580244,
		0.544621,
		0.508038,
		0.470665,
		0.432694,
		0.394338,
		0.355834,
		0.317445,
		0.279458,
		0.242186,
		0.205974,
		0.171193,
		0.13825,
		0.107582,
		0.079662,
		0.055,
		0.034144,
		0.017685,
		0.006254,
		0.00053,
		0.999771,
		0.999454,
		0.998897,
		0.998008,
		0.99669,
		0.994838,
		0.992346,
		0.989102,
		0.984999,
		0.979927,
		0.973783,
		0.966466,
		0.957881,
		0.947941,
		0.936566,
		0.923684,
		0.909232,
		0.893159,
		0.875423,
		0.855993,
		0.834852,
		0.811993,
		0.787425,
		0.761169,
		0.73326,
		0.703751,
		0.672708,
		0.640214,
		0.606372,
		0.571301,
		0.535139,
		0.498046,
		0.460201,
		0.421806,
		0.383087,
		0.344293,
		0.3057,
		0.26761,
		0.230353,
		0.194292,
		0.159817,
		0.127354,
		0.097363,
		0.070342,
		0.046826,
		0.027391,
		0.012657,
		0.003289,
		0.999985,
		0.999923,
		0.999764,
		0.999438,
		0.998863,
		0.997948,
		0.996591,
		0.994686,
		0.992122,
		0.988787,
		0.984569,
		0.979358,
		0.973047,
		0.965534,
		0.956724,
		0.946526,
		0.93486,
		0.921655,
		0.906848,
		0.890386,
		0.87223,
		0.85235,
		0.830729,
		0.807365,
		0.782268,
		0.755463,
		0.726989,
		0.696903,
		0.665277,
		0.632201,
		0.597782,
		0.562148,
		0.525446,
		0.487844,
		0.449532,
		0.410723,
		0.371655,
		0.332591,
		0.29382,
		0.255662,
		0.218463,
		0.182603,
		0.148496,
		0.116587,
		0.08736,
		0.061339,
		0.039085,
		0.021206,
		0.008351,
		0.001221,
		1,
		0.999999,
		0.999984,
		0.999921,
		0.999757,
		0.999421,
		0.998829,
		0.997886,
		0.99649,
		0.994529,
		0.991891,
		0.988461,
		0.984125,
		0.978769,
		0.972286,
		0.964571,
		0.955526,
		0.945063,
		0.933098,
		0.91956,
		0.904386,
		0.887524,
		0.868936,
		0.848593,
		0.826481,
		0.802599,
		0.77696,
		0.749594,
		0.720543,
		0.68987,
		0.657651,
		0.623984,
		0.588982,
		0.552782,
		0.515539,
		0.47743,
		0.438658,
		0.399446,
		0.360045,
		0.320733,
		0.281814,
		0.243624,
		0.206529,
		0.170928,
		0.137255,
		0.10598,
		0.077611,
		0.052699,
		0.031834,
		0.015654,
		0.004845,
		0.00014,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999919,
		0.999749,
		0.999403,
		0.998793,
		0.997822,
		0.996384,
		0.994366,
		0.991651,
		0.988123,
		0.983664,
		0.978159,
		0.971498,
		0.963574,
		0.954288,
		0.94355,
		0.931276,
		0.917395,
		0.901844,
		0.884571,
		0.865538,
		0.84472,
		0.822103,
		0.79769,
		0.771497,
		0.743557,
		0.713918,
		0.682646,
		0.649825,
		0.615558,
		0.579968,
		0.543198,
		0.505413,
		0.466802,
		0.427577,
		0.387975,
		0.34826,
		0.308725,
		0.269691,
		0.231511,
		0.19457,
		0.159289,
		0.126123,
		0.095568,
		0.068159,
		0.044472,
		0.025132,
		0.010808,
		0.002221,
		0.999771,
		0.999923,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999983,
		0.999916,
		0.999741,
		0.999384,
		0.998756,
		0.997755,
		0.996274,
		0.994196,
		0.991403,
		0.987773,
		0.983187,
		0.977527,
		0.970681,
		0.962541,
		0.953007,
		0.941985,
		0.929393,
		0.915158,
		0.899218,
		0.881522,
		0.862033,
		0.840726,
		0.817592,
		0.792634,
		0.765874,
		0.737347,
		0.707107,
		0.675225,
		0.641793,
		0.606919,
		0.570735,
		0.533393,
		0.495067,
		0.455958,
		0.416289,
		0.376311,
		0.336303,
		0.296573,
		0.25746,
		0.219335,
		0.182603,
		0.147708,
		0.11513,
		0.085389,
		0.059048,
		0.036715,
		0.019046,
		0.006746,
		0.000573,
		0.998929,
		0.999454,
		0.999764,
		0.999921,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999913,
		0.999733,
		0.999365,
		0.998717,
		0.997686,
		0.99616,
		0.99402,
		0.991144,
		0.987409,
		0.982692,
		0.976872,
		0.969836,
		0.961473,
		0.951681,
		0.940367,
		0.927446,
		0.912847,
		0.896506,
		0.878374,
		0.858415,
		0.836606,
		0.812941,
		0.787425,
		0.760084,
		0.730958,
		0.700105,
		0.667604,
		0.63355,
		0.598062,
		0.561279,
		0.523363,
		0.484498,
		0.444897,
		0.404796,
		0.364459,
		0.324181,
		0.284287,
		0.245133,
		0.207112,
		0.170651,
		0.136215,
		0.10431,
		0.075484,
		0.05033,
		0.029488,
		0.013648,
		0.003552,
		0.996876,
		0.998065,
		0.998897,
		0.999438,
		0.999757,
		0.999919,
		0.999983,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999911,
		0.999725,
		0.999345,
		0.998677,
		0.997614,
		0.996042,
		0.993838,
		0.990877,
		0.987032,
		0.982179,
		0.976194,
		0.96896,
		0.960366,
		0.950308,
		0.938692,
		0.925433,
		0.910457,
		0.893703,
		0.875123,
		0.854681,
		0.832357,
		0.808146,
		0.782059,
		0.754123,
		0.724384,
		0.692907,
		0.659775,
		0.625091,
		0.588982,
		0.551596,
		0.513104,
		0.473704,
		0.433618,
		0.393097,
		0.35242,
		0.311899,
		0.271875,
		0.232725,
		0.194862,
		0.158736,
		0.124837,
		0.093699,
		0.065898,
		0.042059,
		0.022856,
		0.009016,
		0.00132,
		0.992969,
		0.995126,
		0.996784,
		0.998008,
		0.998863,
		0.999421,
		0.999749,
		0.999916,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999908,
		0.999716,
		0.999324,
		0.998635,
		0.99754,
		0.995919,
		0.993648,
		0.990599,
		0.986641,
		0.981647,
		0.97549,
		0.968053,
		0.95922,
		0.948887,
		0.936959,
		0.92335,
		0.907987,
		0.890808,
		0.871765,
		0.850827,
		0.827973,
		0.803203,
		0.776529,
		0.747985,
		0.717621,
		0.685507,
		0.651733,
		0.61641,
		0.579674,
		0.541681,
		0.502614,
		0.462682,
		0.42212,
		0.381194,
		0.3402,
		0.299464,
		0.259348,
		0.22025,
		0.182603,
		0.146884,
		0.113609,
		0.083339,
		0.056682,
		0.034297,
		0.016893,
		0.005236,
		0.000152,
		0.986574,
		0.989983,
		0.992769,
		0.994985,
		0.99669,
		0.997948,
		0.998829,
		0.999403,
		0.999741,
		0.999913,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999905,
		0.999707,
		0.999302,
		0.998592,
		0.997462,
		0.995792,
		0.993452,
		0.99031,
		0.986235,
		0.981094,
		0.974761,
		0.967112,
		0.958032,
		0.947416,
		0.935165,
		0.921195,
		0.905432,
		0.887815,
		0.868297,
		0.846847,
		0.82345,
		0.798105,
		0.770831,
		0.741665,
		0.710662,
		0.677898,
		0.643472,
		0.607502,
		0.570132,
		0.531529,
		0.491888,
		0.45143,
		0.410404,
		0.369091,
		0.327802,
		0.286884,
		0.246719,
		0.207724,
		0.17036,
		0.135126,
		0.102567,
		0.073275,
		0.047892,
		0.02711,
		0.011678,
		0.002404,
		0.977109,
		0.982017,
		0.986201,
		0.9897,
		0.992561,
		0.994838,
		0.996591,
		0.997886,
		0.998793,
		0.999384,
		0.999733,
		0.999911,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.99998,
		0.999902,
		0.999697,
		0.99928,
		0.998546,
		0.997381,
		0.995659,
		0.993247,
		0.990011,
		0.985814,
		0.980522,
		0.974004,
		0.966137,
		0.956802,
		0.945892,
		0.933308,
		0.918966,
		0.90279,
		0.884721,
		0.864714,
		0.842738,
		0.818782,
		0.792848,
		0.764959,
		0.735156,
		0.7035,
		0.670076,
		0.634987,
		0.598361,
		0.560352,
		0.521138,
		0.480925,
		0.439948,
		0.39847,
		0.356789,
		0.315234,
		0.27417,
		0.234001,
		0.195168,
		0.158156,
		0.123492,
		0.091751,
		0.063557,
		0.039587,
		0.02057,
		0.007298,
		0.000621,
		0.964076,
		0.970681,
		0.976488,
		0.981521,
		0.985814,
		0.989406,
		0.992346,
		0.994686,
		0.99649,
		0.997822,
		0.998756,
		0.999365,
		0.999725,
		0.999908,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999898,
		0.999687,
		0.999256,
		0.998499,
		0.997298,
		0.995521,
		0.993035,
		0.9897,
		0.985376,
		0.979927,
		0.973219,
		0.965125,
		0.955526,
		0.944313,
		0.931385,
		0.916658,
		0.900056,
		0.881522,
		0.861011,
		0.838495,
		0.813964,
		0.787425,
		0.758906,
		0.728452,
		0.696131,
		0.662033,
		0.626271,
		0.588982,
		0.55033,
		0.510503,
		0.469722,
		0.428234,
		0.38632,
		0.344293,
		0.302502,
		0.261333,
		0.221211,
		0.182603,
		0.14602,
		0.112019,
		0.081207,
		0.05424,
		0.031834,
		0.014759,
		0.003848,
		0.947082,
		0.955526,
		0.963126,
		0.969893,
		0.975845,
		0.981008,
		0.985413,
		0.989102,
		0.992122,
		0.994529,
		0.996384,
		0.997755,
		0.998717,
		0.999345,
		0.999716,
		0.999905,
		0.99998,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999895,
		0.999677,
		0.999232,
		0.998451,
		0.99721,
		0.995378,
		0.992814,
		0.989376,
		0.984922,
		0.97931,
		0.972405,
		0.964076,
		0.954204,
		0.942676,
		0.929393,
		0.914268,
		0.897228,
		0.878214,
		0.857183,
		0.834112,
		0.80899,
		0.781832,
		0.752666,
		0.721547,
		0.688547,
		0.653763,
		0.617319,
		0.579359,
		0.540059,
		0.499621,
		0.458277,
		0.416289,
		0.373956,
		0.331609,
		0.289616,
		0.248386,
		0.208367,
		0.170054,
		0.133985,
		0.100747,
		0.070982,
		0.045385,
		0.024706,
		0.009762,
		0.001432,
		0.925841,
		0.936212,
		0.945719,
		0.954362,
		0.962145,
		0.969079,
		0.975181,
		0.980477,
		0.984999,
		0.988787,
		0.991891,
		0.994366,
		0.996274,
		0.997686,
		0.998677,
		0.999324,
		0.999707,
		0.999902,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999978,
		0.999891,
		0.999666,
		0.999206,
		0.9984,
		0.997119,
		0.995229,
		0.992584,
		0.98904,
		0.98445,
		0.978669,
		0.971559,
		0.962988,
		0.952833,
		0.94098,
		0.92733,
		0.911794,
		0.894301,
		0.874792,
		0.853227,
		0.829583,
		0.803856,
		0.776061,
		0.746234,
		0.714434,
		0.680741,
		0.645261,
		0.608124,
		0.569487,
		0.529537,
		0.488488,
		0.446587,
		0.404113,
		0.361381,
		0.318742,
		0.276585,
		0.235343,
		0.19549,
		0.157547,
		0.122084,
		0.08972,
		0.061133,
		0.037057,
		0.018285,
		0.005678,
		0.000165,
		0.900183,
		0.91251,
		0.923982,
		0.934586,
		0.944313,
		0.95316,
		0.961131,
		0.968237,
		0.974494,
		0.979927,
		0.984569,
		0.988461,
		0.991651,
		0.994196,
		0.99616,
		0.997614,
		0.998635,
		0.999302,
		0.999697,
		0.999898,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999977,
		0.999888,
		0.999655,
		0.99918,
		0.998347,
		0.997025,
		0.995074,
		0.992346,
		0.988691,
		0.983959,
		0.978003,
		0.970681,
		0.961859,
		0.95141,
		0.939222,
		0.925192,
		0.909232,
		0.891271,
		0.871252,
		0.849137,
		0.824905,
		0.798554,
		0.770107,
		0.739603,
		0.707107,
		0.672708,
		0.636519,
		0.59868,
		0.559361,
		0.518758,
		0.477101,
		0.434652,
		0.391707,
		0.348599,
		0.3057,
		0.263423,
		0.222223,
		0.182603,
		0.145114,
		0.110356,
		0.078987,
		0.05172,
		0.02933,
		0.012657,
		0.00261,
		0.870045,
		0.88431,
		0.897749,
		0.910341,
		0.922066,
		0.932909,
		0.942861,
		0.951918,
		0.960083,
		0.967366,
		0.973783,
		0.979358,
		0.984125,
		0.988123,
		0.991403,
		0.99402,
		0.996042,
		0.99754,
		0.998592,
		0.99928,
		0.999687,
		0.999895,
		0.999978,
		0.999999,
		1,
		0.999998,
		0.999976,
		0.999884,
		0.999643,
		0.999152,
		0.998291,
		0.996926,
		0.994912,
		0.992097,
		0.988327,
		0.983449,
		0.977312,
		0.96977,
		0.960686,
		0.949934,
		0.937398,
		0.922975,
		0.906578,
		0.888134,
		0.86759,
		0.844907,
		0.820069,
		0.79308,
		0.763962,
		0.732765,
		0.699558,
		0.664439,
		0.62753,
		0.588982,
		0.548975,
		0.507719,
		0.465459,
		0.422472,
		0.379073,
		0.335615,
		0.292492,
		0.250141,
		0.209044,
		0.169733,
		0.132788,
		0.098846,
		0.068601,
		0.042808,
		0.022285,
		0.007921,
		0.000675,
		0.835475,
		0.851607,
		0.866967,
		0.881522,
		0.895243,
		0.908106,
		0.92009,
		0.931179,
		0.941362,
		0.950636,
		0.959001,
		0.966466,
		0.973047,
		0.978769,
		0.983664,
		0.987773,
		0.991144,
		0.993838,
		0.995919,
		0.997462,
		0.998546,
		0.999256,
		0.999677,
		0.999891,
		0.999977,
		0.999998,
		1,
		0.999998,
		0.999976,
		0.99988,
		0.999631,
		0.999123,
		0.998234,
		0.996824,
		0.994744,
		0.991839,
		0.987949,
		0.982919,
		0.976593,
		0.968823,
		0.959469,
		0.948403,
		0.935507,
		0.920678,
		0.903828,
		0.884886,
		0.863799,
		0.840533,
		0.815072,
		0.787425,
		0.757621,
		0.725714,
		0.691782,
		0.655929,
		0.618289,
		0.579023,
		0.538323,
		0.496415,
		0.453557,
		0.410046,
		0.366214,
		0.322436,
		0.27913,
		0.236757,
		0.19583,
		0.156908,
		0.120607,
		0.087601,
		0.058623,
		0.034471,
		0.016011,
		0.004182,
		0.796619,
		0.814507,
		0.831692,
		0.848135,
		0.863799,
		0.878652,
		0.892661,
		0.905802,
		0.918052,
		0.929393,
		0.939815,
		0.949311,
		0.957881,
		0.965534,
		0.972286,
		0.978159,
		0.983187,
		0.987409,
		0.990877,
		0.993648,
		0.995792,
		0.997381,
		0.998499,
		0.999232,
		0.999666,
		0.999888,
		0.999976,
		0.999998,
		1,
		0.999998,
		0.999975,
		0.999876,
		0.999618,
		0.999093,
		0.998174,
		0.996717,
		0.994569,
		0.991569,
		0.987556,
		0.982367,
		0.975845,
		0.967838,
		0.958205,
		0.946813,
		0.933544,
		0.918295,
		0.900978,
		0.881522,
		0.859876,
		0.836008,
		0.809907,
		0.781584,
		0.751077,
		0.718443,
		0.68377,
		0.647171,
		0.608789,
		0.568797,
		0.527402,
		0.484843,
		0.441396,
		0.397375,
		0.353134,
		0.30907,
		0.265625,
		0.223289,
		0.182603,
		0.144161,
		0.108615,
		0.076675,
		0.049119,
		0.026791,
		0.010606,
		0.001559,
		0.753719,
		0.77321,
		0.792083,
		0.810295,
		0.827802,
		0.844563,
		0.860539,
		0.875696,
		0.890001,
		0.903427,
		0.91595,
		0.927551,
		0.938217,
		0.947941,
		0.956724,
		0.964571,
		0.971498,
		0.977527,
		0.982692,
		0.987032,
		0.990599,
		0.993452,
		0.995659,
		0.997298,
		0.998451,
		0.999206,
		0.999655,
		0.999884,
		0.999976,
		0.999998,
		1,
		0.999998,
		0.999974,
		0.999871,
		0.999605,
		0.999062,
		0.998111,
		0.996606,
		0.994386,
		0.991289,
		0.987147,
		0.981794,
		0.975068,
		0.966816,
		0.956891,
		0.945162,
		0.931507,
		0.915824,
		0.898023,
		0.878037,
		0.855814,
		0.831326,
		0.804566,
		0.775551,
		0.744321,
		0.710944,
		0.675515,
		0.638156,
		0.599022,
		0.558299,
		0.516206,
		0.473,
		0.428973,
		0.38446,
		0.339837,
		0.295525,
		0.251992,
		0.209758,
		0.169394,
		0.131531,
		0.096858,
		0.066128,
		0.040162,
		0.019856,
		0.006177,
		0.000179,
		0.707107,
		0.728012,
		0.7484,
		0.76822,
		0.787425,
		0.805968,
		0.823803,
		0.840888,
		0.857183,
		0.872652,
		0.88726,
		0.900978,
		0.913781,
		0.925648,
		0.936566,
		0.946526,
		0.955526,
		0.963574,
		0.970681,
		0.976872,
		0.982179,
		0.986641,
		0.99031,
		0.993247,
		0.995521,
		0.99721,
		0.9984,
		0.99918,
		0.999643,
		0.99988,
		0.999975,
		0.999998,
		1,
		0.999998,
		0.999973,
		0.999867,
		0.999591,
		0.999029,
		0.998046,
		0.99649,
		0.994196,
		0.990997,
		0.98672,
		0.981196,
		0.97426,
		0.965752,
		0.955526,
		0.943447,
		0.929393,
		0.91326,
		0.89496,
		0.874425,
		0.851607,
		0.826481,
		0.799043,
		0.769316,
		0.737347,
		0.70321,
		0.667009,
		0.628878,
		0.588982,
		0.547521,
		0.504731,
		0.460882,
		0.416289,
		0.371306,
		0.326332,
		0.281814,
		0.238249,
		0.196187,
		0.156234,
		0.119058,
		0.085389,
		0.056024,
		0.031834,
		0.013765,
		0.002844,
		0.657196,
		0.679298,
		0.700991,
		0.722224,
		0.742944,
		0.7631,
		0.782641,
		0.801521,
		0.819691,
		0.837108,
		0.853729,
		0.869516,
		0.884434,
		0.898452,
		0.911542,
		0.923684,
		0.93486,
		0.945063,
		0.954288,
		0.962541,
		0.969836,
		0.976194,
		0.981647,
		0.986235,
		0.990011,
		0.993035,
		0.995378,
		0.997119,
		0.998347,
		0.999152,
		0.999631,
		0.999876,
		0.999974,
		0.999998,
		1,
		0.999998,
		0.999972,
		0.999862,
		0.999576,
		0.998995,
		0.997978,
		0.996368,
		0.993998,
		0.990693,
		0.986276,
		0.980575,
		0.973418,
		0.964646,
		0.954108,
		0.941666,
		0.927198,
		0.9106,
		0.891784,
		0.870682,
		0.847251,
		0.821467,
		0.793332,
		0.762874,
		0.730146,
		0.695232,
		0.658245,
		0.619329,
		0.578663,
		0.53646,
		0.492972,
		0.448489,
		0.403344,
		0.357915,
		0.312626,
		0.26795,
		0.224414,
		0.182603,
		0.14316,
		0.106789,
		0.074267,
		0.046438,
		0.024223,
		0.008627,
		0.000737,
		0.604481,
		0.62753,
		0.650291,
		0.672708,
		0.694725,
		0.71629,
		0.737347,
		0.757843,
		0.777728,
		0.796951,
		0.815463,
		0.833218,
		0.850172,
		0.866286,
		0.881522,
		0.895847,
		0.909232,
		0.921655,
		0.933098,
		0.94355,
		0.953007,
		0.961473,
		0.96896,
		0.97549,
		0.981094,
		0.985814,
		0.9897,
		0.992814,
		0.995229,
		0.997025,
		0.998291,
		0.999123,
		0.999618,
		0.999871,
		0.999973,
		0.999998,
		1,
		0.999998,
		0.999971,
		0.999857,
		0.999561,
		0.99896,
		0.997907,
		0.996242,
		0.993791,
		0.990375,
		0.985814,
		0.979927,
		0.972543,
		0.963495,
		0.952633,
		0.939815,
		0.924919,
		0.907839,
		0.888488,
		0.866803,
		0.842738,
		0.816277,
		0.787425,
		0.756216,
		0.722712,
		0.687003,
		0.649214,
		0.609502,
		0.568056,
		0.525108,
		0.480925,
		0.435818,
		0.39014,
		0.344293,
		0.298728,
		0.253947,
		0.210511,
		0.169038,
		0.13021,
		0.094776,
		0.063557,
		0.03745,
		0.01743,
		0.004562,
		0.549528,
		0.573252,
		0.596815,
		0.620158,
		0.643226,
		0.66596,
		0.688306,
		0.710206,
		0.731605,
		0.752448,
		0.772681,
		0.792253,
		0.811114,
		0.829215,
		0.84651,
		0.862958,
		0.878519,
		0.893159,
		0.906848,
		0.91956,
		0.931276,
		0.941985,
		0.951681,
		0.960366,
		0.968053,
		0.974761,
		0.980522,
		0.985376,
		0.989376,
		0.992584,
		0.995074,
		0.996926,
		0.998234,
		0.999093,
		0.999605,
		0.999867,
		0.999972,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999852,
		0.999545,
		0.998922,
		0.997833,
		0.99611,
		0.993576,
		0.990045,
		0.985332,
		0.979253,
		0.971631,
		0.962298,
		0.951098,
		0.937891,
		0.922551,
		0.904972,
		0.88507,
		0.86278,
		0.838063,
		0.810904,
		0.781314,
		0.749335,
		0.715035,
		0.678515,
		0.63991,
		0.599388,
		0.557157,
		0.513462,
		0.468589,
		0.422868,
		0.376679,
		0.330446,
		0.28465,
		0.239825,
		0.196564,
		0.155525,
		0.117431,
		0.083077,
		0.053333,
		0.02915,
		0.011564,
		0.001703,
		0.492972,
		0.517077,
		0.541152,
		0.56514,
		0.588982,
		0.612621,
		0.635997,
		0.659052,
		0.681729,
		0.703969,
		0.725714,
		0.746909,
		0.767497,
		0.787425,
		0.806641,
		0.825095,
		0.842738,
		0.859528,
		0.875423,
		0.890386,
		0.904386,
		0.917395,
		0.929393,
		0.940367,
		0.950308,
		0.95922,
		0.967112,
		0.974004,
		0.979927,
		0.984922,
		0.98904,
		0.992346,
		0.994912,
		0.996824,
		0.998174,
		0.999062,
		0.999591,
		0.999862,
		0.999971,
		0.999998,
		1,
		0.999998,
		0.999969,
		0.999846,
		0.999529,
		0.998883,
		0.997755,
		0.995972,
		0.99335,
		0.9897,
		0.984829,
		0.97855,
		0.970681,
		0.961052,
		0.949502,
		0.93589,
		0.92009,
		0.901995,
		0.881522,
		0.858608,
		0.833218,
		0.80534,
		0.774992,
		0.742222,
		0.707107,
		0.669758,
		0.630323,
		0.588982,
		0.545959,
		0.501515,
		0.455958,
		0.409641,
		0.362964,
		0.316383,
		0.270407,
		0.225603,
		0.182603,
		0.142104,
		0.104874,
		0.071756,
		0.043674,
		0.021638,
		0.006746,
		0.000196,
		0.435513,
		0.459685,
		0.483964,
		0.508291,
		0.532609,
		0.556859,
		0.58098,
		0.604914,
		0.628601,
		0.651979,
		0.67499,
		0.697574,
		0.71967,
		0.741222,
		0.762171,
		0.782462,
		0.80204,
		0.820854,
		0.838854,
		0.855993,
		0.87223,
		0.887524,
		0.901844,
		0.915158,
		0.927446,
		0.938692,
		0.948887,
		0.958032,
		0.966137,
		0.973219,
		0.97931,
		0.98445,
		0.988691,
		0.992097,
		0.994744,
		0.996717,
		0.998111,
		0.999029,
		0.999576,
		0.999857,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999968,
		0.999841,
		0.999511,
		0.998843,
		0.997675,
		0.995829,
		0.993115,
		0.98934,
		0.984305,
		0.977817,
		0.969692,
		0.959754,
		0.947842,
		0.93381,
		0.917532,
		0.898903,
		0.877839,
		0.854281,
		0.828196,
		0.799577,
		0.76845,
		0.734868,
		0.698919,
		0.660725,
		0.620445,
		0.578275,
		0.534454,
		0.489264,
		0.443032,
		0.396135,
		0.349001,
		0.302114,
		0.256014,
		0.211307,
		0.168662,
		0.128819,
		0.092596,
		0.060886,
		0.034672,
		0.015024,
		0.003111,
		0.377913,
		0.401822,
		0.425977,
		0.45032,
		0.474793,
		0.499339,
		0.523896,
		0.548406,
		0.572806,
		0.597036,
		0.621034,
		0.644738,
		0.668086,
		0.691017,
		0.713469,
		0.735383,
		0.756699,
		0.777359,
		0.797307,
		0.816488,
		0.834852,
		0.85235,
		0.868936,
		0.884571,
		0.899218,
		0.912847,
		0.925433,
		0.936959,
		0.947416,
		0.956802,
		0.965125,
		0.972405,
		0.978669,
		0.983959,
		0.988327,
		0.991839,
		0.994569,
		0.996606,
		0.998046,
		0.998995,
		0.999561,
		0.999852,
		0.999969,
		0.999998,
		1,
		0.999998,
		0.999966,
		0.999835,
		0.999493,
		0.9988,
		0.99759,
		0.995678,
		0.99287,
		0.988964,
		0.983758,
		0.977053,
		0.968661,
		0.958402,
		0.946113,
		0.931646,
		0.914874,
		0.895691,
		0.874016,
		0.849792,
		0.82299,
		0.793609,
		0.76168,
		0.727265,
		0.690463,
		0.651406,
		0.610268,
		0.567259,
		0.522638,
		0.476704,
		0.429809,
		0.382355,
		0.334797,
		0.287651,
		0.241492,
		0.196963,
		0.154777,
		0.11572,
		0.08066,
		0.050548,
		0.026425,
		0.009431,
		0.000807,
		0.320992,
		0.344293,
		0.36798,
		0.391998,
		0.416289,
		0.440796,
		0.465459,
		0.490217,
		0.515011,
		0.539778,
		0.564456,
		0.588982,
		0.613292,
		0.637323,
		0.661012,
		0.684294,
		0.707107,
		0.729388,
		0.751077,
		0.772112,
		0.792437,
		0.811993,
		0.830729,
		0.848593,
		0.865538,
		0.881522,
		0.896506,
		0.910457,
		0.92335,
		0.935165,
		0.945892,
		0.955526,
		0.964076,
		0.971559,
		0.978003,
		0.983449,
		0.987949,
		0.991569,
		0.994386,
		0.99649,
		0.997978,
		0.99896,
		0.999545,
		0.999846,
		0.999968,
		0.999998,
		1,
		0.999998,
		0.999965,
		0.999829,
		0.999474,
		0.998756,
		0.997501,
		0.995521,
		0.992614,
		0.988571,
		0.983187,
		0.976257,
		0.967586,
		0.956994,
		0.944313,
		0.929393,
		0.912108,
		0.892353,
		0.870045,
		0.845133,
		0.817592,
		0.787425,
		0.754672,
		0.719404,
		0.681729,
		0.641793,
		0.599783,
		0.555928,
		0.510503,
		0.463832,
		0.416289,
		0.368303,
		0.32036,
		0.273008,
		0.226862,
		0.182603,
		0.140991,
		0.102863,
		0.069138,
		0.04083,
		0.019046,
		0.004996,
		0.265625,
		0.28796,
		0.310823,
		0.33416,
		0.357915,
		0.382031,
		0.40645,
		0.431113,
		0.455958,
		0.480925,
		0.505952,
		0.530974,
		0.555928,
		0.580749,
		0.605372,
		0.629732,
		0.653763,
		0.6774,
		0.700578,
		0.723232,
		0.745299,
		0.766717,
		0.787425,
		0.807365,
		0.826481,
		0.84472,
		0.862033,
		0.878374,
		0.893703,
		0.907987,
		0.921195,
		0.933308,
		0.944313,
		0.954204,
		0.962988,
		0.970681,
		0.977312,
		0.982919,
		0.987556,
		0.991289,
		0.994196,
		0.996368,
		0.997907,
		0.998922,
		0.999529,
		0.999841,
		0.999966,
		0.999998,
		1,
		0.999998,
		0.999964,
		0.999822,
		0.999454,
		0.998709,
		0.997409,
		0.995357,
		0.992346,
		0.988161,
		0.982591,
		0.975425,
		0.966466,
		0.955526,
		0.942438,
		0.927049,
		0.909232,
		0.888882,
		0.865921,
		0.840298,
		0.811993,
		0.781018,
		0.747418,
		0.711275,
		0.672708,
		0.631876,
		0.588982,
		0.544274,
		0.498046,
		0.450646,
		0.402472,
		0.353984,
		0.3057,
		0.258204,
		0.212149,
		0.168264,
		0.127354,
		0.090309,
		0.058111,
		0.031834,
		0.012657,
		0.001868,
		0.212739,
		0.233741,
		0.255412,
		0.2777,
		0.300551,
		0.32391,
		0.34772,
		0.371923,
		0.39646,
		0.42127,
		0.446292,
		0.471461,
		0.496716,
		0.52199,
		0.547218,
		0.572334,
		0.59727,
		0.62196,
		0.646337,
		0.670332,
		0.693878,
		0.71691,
		0.739362,
		0.761169,
		0.782268,
		0.802599,
		0.822103,
		0.840726,
		0.858415,
		0.875123,
		0.890808,
		0.905432,
		0.918966,
		0.931385,
		0.942676,
		0.952833,
		0.961859,
		0.96977,
		0.976593,
		0.982367,
		0.987147,
		0.990997,
		0.993998,
		0.996242,
		0.997833,
		0.998883,
		0.999511,
		0.999835,
		0.999965,
		0.999998,
		1,
		0.999998,
		0.999962,
		0.999815,
		0.999434,
		0.99866,
		0.997312,
		0.995185,
		0.992065,
		0.987733,
		0.981968,
		0.974557,
		0.965297,
		0.953996,
		0.940484,
		0.924608,
		0.90624,
		0.885274,
		0.861636,
		0.835279,
		0.806187,
		0.774379,
		0.739908,
		0.702868,
		0.663389,
		0.621646,
		0.577857,
		0.53229,
		0.485261,
		0.437142,
		0.38836,
		0.339405,
		0.29083,
		0.243258,
		0.197385,
		0.153986,
		0.113919,
		0.078131,
		0.047667,
		0.02367,
		0.007397,
		0.000216,
		0.163312,
		0.182603,
		0.202704,
		0.223565,
		0.245133,
		0.267356,
		0.290179,
		0.313545,
		0.337397,
		0.361676,
		0.38632,
		0.411269,
		0.436458,
		0.461824,
		0.487302,
		0.512824,
		0.538323,
		0.563732,
		0.588982,
		0.614003,
		0.638727,
		0.663083,
		0.687003,
		0.710418,
		0.73326,
		0.755463,
		0.77696,
		0.79769,
		0.817592,
		0.836606,
		0.854681,
		0.871765,
		0.887815,
		0.90279,
		0.916658,
		0.929393,
		0.94098,
		0.95141,
		0.960686,
		0.968823,
		0.975845,
		0.981794,
		0.98672,
		0.990693,
		0.993791,
		0.99611,
		0.997755,
		0.998843,
		0.999493,
		0.999829,
		0.999964,
		0.999998,
		1,
		0.999997,
		0.999961,
		0.999808,
		0.999412,
		0.998609,
		0.99721,
		0.995005,
		0.991772,
		0.987285,
		0.981318,
		0.973651,
		0.964076,
		0.952401,
		0.938448,
		0.922066,
		0.903125,
		0.881522,
		0.857183,
		0.830067,
		0.800163,
		0.767497,
		0.732133,
		0.694173,
		0.653763,
		0.611093,
		0.566399,
		0.519969,
		0.472143,
		0.42332,
		0.373956,
		0.324575,
		0.275766,
		0.228195,
		0.182603,
		0.139816,
		0.100747,
		0.066407,
		0.037906,
		0.016464,
		0.003417,
		0.118371,
		0.135567,
		0.153712,
		0.172758,
		0.192655,
		0.213353,
		0.234798,
		0.256937,
		0.279714,
		0.303071,
		0.32695,
		0.351291,
		0.376031,
		0.401108,
		0.426457,
		0.452013,
		0.477707,
		0.503473,
		0.529241,
		0.554942,
		0.580504,
		0.605858,
		0.63093,
		0.655651,
		0.679948,
		0.703751,
		0.726989,
		0.749594,
		0.771497,
		0.792634,
		0.812941,
		0.832357,
		0.850827,
		0.868297,
		0.884721,
		0.900056,
		0.914268,
		0.92733,
		0.939222,
		0.949934,
		0.959469,
		0.967838,
		0.975068,
		0.981196,
		0.986276,
		0.990375,
		0.993576,
		0.995972,
		0.997675,
		0.9988,
		0.999474,
		0.999822,
		0.999962,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.9998,
		0.999389,
		0.998556,
		0.997104,
		0.994817,
		0.991466,
		0.986817,
		0.980638,
		0.972704,
		0.962803,
		0.950736,
		0.936326,
		0.919418,
		0.899882,
		0.877618,
		0.852555,
		0.824654,
		0.793912,
		0.760363,
		0.724081,
		0.68518,
		0.643821,
		0.600208,
		0.5546,
		0.507305,
		0.45869,
		0.409179,
		0.359265,
		0.309504,
		0.260527,
		0.213043,
		0.167843,
		0.125807,
		0.08791,
		0.055226,
		0.028941,
		0.010354,
		0.000888,
		0.078987,
		0.093699,
		0.109496,
		0.126332,
		0.144161,
		0.162935,
		0.182603,
		0.203114,
		0.224414,
		0.246449,
		0.269161,
		0.292492,
		0.316383,
		0.340772,
		0.365596,
		0.39079,
		0.416289,
		0.442026,
		0.467931,
		0.493936,
		0.519969,
		0.545959,
		0.571833,
		0.597518,
		0.622942,
		0.648029,
		0.672708,
		0.696903,
		0.720543,
		0.743557,
		0.765874,
		0.787425,
		0.808146,
		0.827973,
		0.846847,
		0.864714,
		0.881522,
		0.897228,
		0.911794,
		0.925192,
		0.937398,
		0.948403,
		0.958205,
		0.966816,
		0.97426,
		0.980575,
		0.985814,
		0.990045,
		0.99335,
		0.995829,
		0.99759,
		0.998756,
		0.999454,
		0.999815,
		0.999961,
		0.999997,
		1,
		0.999997,
		0.999958,
		0.999793,
		0.999365,
		0.998499,
		0.996992,
		0.99462,
		0.991144,
		0.986327,
		0.979927,
		0.971715,
		0.961473,
		0.948998,
		0.934112,
		0.916658,
		0.896506,
		0.873557,
		0.847743,
		0.819031,
		0.787425,
		0.752968,
		0.715743,
		0.675879,
		0.63355,
		0.588982,
		0.542452,
		0.494292,
		0.444897,
		0.394723,
		0.344293,
		0.294205,
		0.245133,
		0.197833,
		0.153149,
		0.112019,
		0.075484,
		0.044689,
		0.020897,
		0.005495,
		0.046277,
		0.058111,
		0.071162,
		0.085389,
		0.100747,
		0.117192,
		0.134676,
		0.153149,
		0.172561,
		0.19286,
		0.213992,
		0.2359,
		0.258527,
		0.281814,
		0.3057,
		0.330122,
		0.355015,
		0.380315,
		0.405954,
		0.431863,
		0.457972,
		0.48421,
		0.510503,
		0.53678,
		0.562964,
		0.588982,
		0.614757,
		0.640214,
		0.665277,
		0.68987,
		0.713918,
		0.737347,
		0.760084,
		0.782059,
		0.803203,
		0.82345,
		0.842738,
		0.861011,
		0.878214,
		0.894301,
		0.909232,
		0.922975,
		0.935507,
		0.946813,
		0.956891,
		0.965752,
		0.973418,
		0.979927,
		0.985332,
		0.9897,
		0.993115,
		0.995678,
		0.997501,
		0.998709,
		0.999434,
		0.999808,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999956,
		0.999784,
		0.99934,
		0.998441,
		0.996876,
		0.994413,
		0.990808,
		0.985814,
		0.979184,
		0.970681,
		0.960083,
		0.947185,
		0.931803,
		0.913781,
		0.892988,
		0.869329,
		0.842738,
		0.813189,
		0.780692,
		0.745299,
		0.707107,
		0.666257,
		0.622942,
		0.577405,
		0.529947,
		0.480925,
		0.430764,
		0.379951,
		0.32905,
		0.278696,
		0.229612,
		0.182603,
		0.138573,
		0.098521,
		0.063557,
		0.034906,
		0.013913,
		0.002059,
		0.021402,
		0.029959,
		0.039864,
		0.051077,
		0.063557,
		0.077262,
		0.092147,
		0.108167,
		0.125272,
		0.143415,
		0.162543,
		0.182603,
		0.203541,
		0.2253,
		0.247821,
		0.271043,
		0.294906,
		0.319344,
		0.344293,
		0.369686,
		0.395453,
		0.421525,
		0.447829,
		0.474293,
		0.500843,
		0.527402,
		0.553895,
		0.580244,
		0.606372,
		0.632201,
		0.657651,
		0.682646,
		0.707107,
		0.730958,
		0.754123,
		0.776529,
		0.798105,
		0.818782,
		0.838495,
		0.857183,
		0.874792,
		0.891271,
		0.906578,
		0.920678,
		0.933544,
		0.945162,
		0.955526,
		0.964646,
		0.972543,
		0.979253,
		0.984829,
		0.98934,
		0.99287,
		0.995521,
		0.997409,
		0.99866,
		0.999412,
		0.9998,
		0.999958,
		0.999997,
		1,
		0.999997,
		0.999954,
		0.999775,
		0.999313,
		0.998379,
		0.996753,
		0.994196,
		0.990456,
		0.985277,
		0.978406,
		0.9696,
		0.958631,
		0.945291,
		0.929393,
		0.91078,
		0.889324,
		0.864928,
		0.837533,
		0.807118,
		0.773702,
		0.737347,
		0.698162,
		0.656305,
		0.611985,
		0.565468,
		0.517077,
		0.4672,
		0.416289,
		0.364871,
		0.313545,
		0.262996,
		0.213992,
		0.167396,
		0.124172,
		0.085389,
		0.05223,
		0.026003,
		0.008146,
		0.000238,
		0.005562,
		0.010445,
		0.0168,
		0.024591,
		0.033782,
		0.044332,
		0.0562,
		0.069344,
		0.083718,
		0.099276,
		0.11597,
		0.13375,
		0.152563,
		0.172357,
		0.193074,
		0.214658,
		0.237049,
		0.260186,
		0.284006,
		0.308444,
		0.333432,
		0.358903,
		0.384786,
		0.41101,
		0.437501,
		0.464184,
		0.490984,
		0.517822,
		0.544621,
		0.571301,
		0.597782,
		0.623984,
		0.649825,
		0.675225,
		0.700105,
		0.724384,
		0.747985,
		0.770831,
		0.792848,
		0.813964,
		0.834112,
		0.853227,
		0.871252,
		0.888134,
		0.903828,
		0.918295,
		0.931507,
		0.943447,
		0.954108,
		0.963495,
		0.971631,
		0.97855,
		0.984305,
		0.988964,
		0.992614,
		0.995357,
		0.997312,
		0.998609,
		0.999389,
		0.999793,
		0.999956,
		0.999997,
		1,
		0.999997,
		0.999952,
		0.999766,
		0.999285,
		0.998314,
		0.996625,
		0.993969,
		0.990087,
		0.984714,
		0.977591,
		0.968469,
		0.957114,
		0.943313,
		0.926878,
		0.907651,
		0.885504,
		0.860344,
		0.832117,
		0.800807,
		0.766443,
		0.729099,
		0.688896,
		0.64601,
		0.600669,
		0.553161,
		0.503836,
		0.453111,
		0.401474,
		0.349487,
		0.297795,
		0.247127,
		0.198309,
		0.152262,
		0.110014,
		0.07271,
		0.041614,
		0.018122,
		0.003771,
		0.000807,
		0.003208,
		0.00717,
		0.012657,
		0.019635,
		0.028064,
		0.037906,
		0.049119,
		0.06166,
		0.075484,
		0.090543,
		0.106789,
		0.124172,
		0.142639,
		0.162135,
		0.182603,
		0.203986,
		0.226223,
		0.249252,
		0.273008,
		0.297425,
		0.322436,
		0.34797,
		0.373956,
		0.40032,
		0.426988,
		0.453882,
		0.480925,
		0.508038,
		0.535139,
		0.562148,
		0.588982,
		0.615558,
		0.641793,
		0.667604,
		0.692907,
		0.717621,
		0.741665,
		0.764959,
		0.787425,
		0.80899,
		0.829583,
		0.849137,
		0.86759,
		0.884886,
		0.900978,
		0.915824,
		0.929393,
		0.941666,
		0.952633,
		0.962298,
		0.970681,
		0.977817,
		0.983758,
		0.988571,
		0.992346,
		0.995185,
		0.99721,
		0.998556,
		0.999365,
		0.999784,
		0.999954,
		0.999997,
		1,
		0.999997,
		0.99995,
		0.999757,
		0.999256,
		0.998246,
		0.99649,
		0.993731,
		0.9897,
		0.984125,
		0.976739,
		0.967285,
		0.955526,
		0.941245,
		0.924251,
		0.904386,
		0.881522,
		0.85557,
		0.826481,
		0.794247,
		0.758906,
		0.720543,
		0.679298,
		0.63536,
		0.588982,
		0.540475,
		0.490217,
		0.438658,
		0.38632,
		0.333809,
		0.281814,
		0.231119,
		0.182603,
		0.137255,
		0.096175,
		0.060583,
		0.031834,
		0.011419,
		0.000982,
		9.2e-005,
		0.001462,
		0.004448,
		0.009016,
		0.015128,
		0.022747,
		0.031834,
		0.042346,
		0.05424,
		0.067471,
		0.081992,
		0.097753,
		0.114702,
		0.132788,
		0.151954,
		0.172143,
		0.193297,
		0.215353,
		0.238249,
		0.261919,
		0.286296,
		0.31131,
		0.336891,
		0.362964,
		0.389456,
		0.416289,
		0.443386,
		0.470665,
		0.498046,
		0.525446,
		0.552782,
		0.579968,
		0.606919,
		0.63355,
		0.659775,
		0.685507,
		0.710662,
		0.735156,
		0.758906,
		0.781832,
		0.803856,
		0.824905,
		0.844907,
		0.863799,
		0.881522,
		0.898023,
		0.91326,
		0.927198,
		0.939815,
		0.951098,
		0.961052,
		0.969692,
		0.977053,
		0.983187,
		0.988161,
		0.992065,
		0.995005,
		0.997104,
		0.998499,
		0.99934,
		0.999775,
		0.999952,
		0.999997,
		1,
		0.999997,
		0.999948,
		0.999747,
		0.999225,
		0.998174,
		0.996348,
		0.99348,
		0.989293,
		0.983507,
		0.975845,
		0.966046,
		0.953866,
		0.939084,
		0.921508,
		0.900978,
		0.877369,
		0.850596,
		0.820615,
		0.787425,
		0.751077,
		0.711668,
		0.669354,
		0.624345,
		0.576915,
		0.527402,
		0.476215,
		0.423837,
		0.370832,
		0.317848,
		0.265625,
		0.215002,
		0.166922,
		0.122442,
		0.082738,
		0.049119,
		0.023032,
		0.006072,
		0.000375,
		0.002328,
		0.00592,
		0.011113,
		0.01787,
		0.026151,
		0.035914,
		0.047116,
		0.05971,
		0.073649,
		0.088884,
		0.105362,
		0.12303,
		0.141832,
		0.16171,
		0.182603,
		0.204451,
		0.227188,
		0.250747,
		0.275061,
		0.300059,
		0.325668,
		0.351814,
		0.378419,
		0.405406,
		0.432694,
		0.460201,
		0.487844,
		0.515539,
		0.543198,
		0.570735,
		0.598062,
		0.625091,
		0.651733,
		0.677898,
		0.7035,
		0.728452,
		0.752666,
		0.776061,
		0.798554,
		0.820069,
		0.840533,
		0.859876,
		0.878037,
		0.89496,
		0.9106,
		0.924919,
		0.937891,
		0.949502,
		0.959754,
		0.968661,
		0.976257,
		0.982591,
		0.987733,
		0.991772,
		0.994817,
		0.996992,
		0.998441,
		0.999313,
		0.999766,
		0.99995,
		0.999997,
		1,
		0.999997,
		0.999946,
		0.999736,
		0.999193,
		0.998099,
		0.996199,
		0.993217,
		0.988867,
		0.982859,
		0.974909,
		0.964748,
		0.952128,
		0.936823,
		0.918641,
		0.89742,
		0.873037,
		0.845412,
		0.814507,
		0.78033,
		0.742944,
		0.702461,
		0.659052,
		0.612952,
		0.564456,
		0.513933,
		0.461824,
		0.40865,
		0.355015,
		0.301618,
		0.249252,
		0.198815,
		0.151319,
		0.107895,
		0.069803,
		0.038444,
		0.015366,
		0.00228,
		0.00086,
		0.003417,
		0.007634,
		0.013475,
		0.020897,
		0.02986,
		0.04032,
		0.05223,
		0.065543,
		0.080209,
		0.096175,
		0.113386,
		0.131787,
		0.151319,
		0.171921,
		0.193529,
		0.216079,
		0.239503,
		0.26373,
		0.28869,
		0.314308,
		0.340507,
		0.367211,
		0.394338,
		0.421806,
		0.449532,
		0.47743,
		0.505413,
		0.533393,
		0.561279,
		0.588982,
		0.61641,
		0.643472,
		0.670076,
		0.696131,
		0.721547,
		0.746234,
		0.770107,
		0.79308,
		0.815072,
		0.836008,
		0.855814,
		0.874425,
		0.891784,
		0.907839,
		0.922551,
		0.93589,
		0.947842,
		0.958402,
		0.967586,
		0.975425,
		0.981968,
		0.987285,
		0.991466,
		0.99462,
		0.996876,
		0.998379,
		0.999285,
		0.999757,
		0.999948,
		0.999997,
		1,
		0.999996,
		0.999944,
		0.999725,
		0.999159,
		0.998019,
		0.996042,
		0.992941,
		0.98842,
		0.982179,
		0.973927,
		0.963389,
		0.950308,
		0.934459,
		0.915644,
		0.893703,
		0.868517,
		0.840008,
		0.808146,
		0.77295,
		0.734494,
		0.692907,
		0.64838,
		0.601168,
		0.551596,
		0.500061,
		0.447041,
		0.393097,
		0.33888,
		0.285138,
		0.232725,
		0.182603,
		0.135857,
		0.093699,
		0.057479,
		0.028698,
		0.009016,
		0.000264,
		9.8e-005,
		0.001559,
		0.004741,
		0.009606,
		0.016114,
		0.024223,
		0.033889,
		0.045066,
		0.057705,
		0.071756,
		0.087167,
		0.103881,
		0.121842,
		0.140991,
		0.161266,
		0.182603,
		0.204936,
		0.228195,
		0.252311,
		0.277209,
		0.302814,
		0.32905,
		0.355834,
		0.383087,
		0.410723,
		0.438658,
		0.466802,
		0.495067,
		0.523363,
		0.551596,
		0.579674,
		0.607502,
		0.634987,
		0.662033,
		0.688547,
		0.714434,
		0.739603,
		0.763962,
		0.787425,
		0.809907,
		0.831326,
		0.851607,
		0.870682,
		0.888488,
		0.904972,
		0.92009,
		0.93381,
		0.946113,
		0.956994,
		0.966466,
		0.974557,
		0.981318,
		0.986817,
		0.991144,
		0.994413,
		0.996753,
		0.998314,
		0.999256,
		0.999747,
		0.999946,
		0.999996,
		1,
		0.999996,
		0.999941,
		0.999713,
		0.999123,
		0.997936,
		0.995877,
		0.992651,
		0.987949,
		0.981465,
		0.972897,
		0.961963,
		0.948403,
		0.931984,
		0.91251,
		0.889821,
		0.863799,
		0.834373,
		0.801521,
		0.765272,
		0.725714,
		0.682994,
		0.637323,
		0.588982,
		0.538323,
		0.485779,
		0.431863,
		0.377182,
		0.322436,
		0.26843,
		0.216079,
		0.166417,
		0.120607,
		0.079949,
		0.045892,
		0.020042,
		0.004182,
		0.0004,
		0.002483,
		0.006313,
		0.011847,
		0.019046,
		0.027864,
		0.038254,
		0.050169,
		0.063557,
		0.078366,
		0.094539,
		0.112019,
		0.130747,
		0.150658,
		0.171689,
		0.193772,
		0.216838,
		0.240813,
		0.265625,
		0.291195,
		0.317445,
		0.344293,
		0.371655,
		0.399446,
		0.427577,
		0.455958,
		0.484498,
		0.513104,
		0.541681,
		0.570132,
		0.598361,
		0.626271,
		0.653763,
		0.680741,
		0.707107,
		0.732765,
		0.757621,
		0.781584,
		0.804566,
		0.826481,
		0.847251,
		0.866803,
		0.88507,
		0.901995,
		0.917532,
		0.931646,
		0.944313,
		0.955526,
		0.965297,
		0.973651,
		0.980638,
		0.986327,
		0.990808,
		0.994196,
		0.996625,
		0.998246,
		0.999225,
		0.999736,
		0.999944,
		0.999996,
		1,
		0.999996,
		0.999939,
		0.9997,
		0.999086,
		0.997848,
		0.995704,
		0.992346,
		0.987455,
		0.980715,
		0.971816,
		0.960468,
		0.946406,
		0.929393,
		0.909232,
		0.885764,
		0.858873,
		0.828496,
		0.794618,
		0.757282,
		0.71659,
		0.672708,
		0.625869,
		0.576381,
		0.524629,
		0.471079,
		0.416289,
		0.360911,
		0.3057,
		0.25152,
		0.199356,
		0.150317,
		0.105651,
		0.066755,
		0.035182,
		0.012657,
		0.001092,
		0.000918,
		0.003647,
		0.008146,
		0.014374,
		0.022285,
		0.031834,
		0.042971,
		0.055645,
		0.069803,
		0.085389,
		0.102344,
		0.120607,
		0.140116,
		0.160804,
		0.182603,
		0.205443,
		0.229249,
		0.253947,
		0.279458,
		0.3057,
		0.332591,
		0.360045,
		0.387975,
		0.416289,
		0.444897,
		0.473704,
		0.502614,
		0.531529,
		0.560352,
		0.588982,
		0.617319,
		0.645261,
		0.672708,
		0.699558,
		0.725714,
		0.751077,
		0.775551,
		0.799043,
		0.821467,
		0.842738,
		0.86278,
		0.881522,
		0.898903,
		0.914874,
		0.929393,
		0.942438,
		0.953996,
		0.964076,
		0.972704,
		0.979927,
		0.985814,
		0.990456,
		0.993969,
		0.99649,
		0.998174,
		0.999193,
		0.999725,
		0.999941,
		0.999996,
		1,
		0.999996,
		0.999936,
		0.999687,
		0.999046,
		0.997755,
		0.995521,
		0.992024,
		0.986936,
		0.979927,
		0.970681,
		0.9589,
		0.944313,
		0.92668,
		0.905802,
		0.881522,
		0.853729,
		0.822365,
		0.787425,
		0.748966,
		0.707107,
		0.662033,
		0.614003,
		0.563354,
		0.510503,
		0.455958,
		0.40032,
		0.344293,
		0.28869,
		0.234441,
		0.182603,
		0.134371,
		0.091084,
		0.05424,
		0.02551,
		0.006746,
		0.000105,
		0.001665,
		0.005063,
		0.010256,
		0.0172,
		0.025848,
		0.03615,
		0.048056,
		0.061512,
		0.076461,
		0.092843,
		0.110599,
		0.129662,
		0.149968,
		0.171447,
		0.194026,
		0.217631,
		0.242186,
		0.26761,
		0.29382,
		0.320733,
		0.34826,
		0.376311,
		0.404796,
		0.433618,
		0.462682,
		0.491888,
		0.521138,
		0.55033,
		0.579359,
		0.608124,
		0.636519,
		0.664439,
		0.691782,
		0.718443,
		0.744321,
		0.769316,
		0.793332,
		0.816277,
		0.838063,
		0.858608,
		0.877839,
		0.895691,
		0.912108,
		0.927049,
		0.940484,
		0.952401,
		0.962803,
		0.971715,
		0.979184,
		0.985277,
		0.990087,
		0.993731,
		0.996348,
		0.998099,
		0.999159,
		0.999713,
		0.999939,
		0.999996,
		1,
		0.999996,
		0.999933,
		0.999673,
		0.999004,
		0.997658,
		0.995329,
		0.991686,
		0.986389,
		0.979099,
		0.969489,
		0.957254,
		0.942118,
		0.923837,
		0.902212,
		0.877086,
		0.848355,
		0.815966,
		0.779928,
		0.74031,
		0.69725,
		0.650955,
		0.601712,
		0.549888,
		0.495938,
		0.440412,
		0.383959,
		0.32734,
		0.271429,
		0.21723,
		0.165879,
		0.118659,
		0.077011,
		0.042546,
		0.017058,
		0.002539,
		0.000428,
		0.002655,
		0.006746,
		0.012657,
		0.020342,
		0.02975,
		0.04083,
		0.053528,
		0.067787,
		0.083548,
		0.100747,
		0.119322,
		0.139203,
		0.160321,
		0.182603,
		0.205974,
		0.230353,
		0.255662,
		0.281814,
		0.308725,
		0.336303,
		0.364459,
		0.393097,
		0.42212,
		0.45143,
		0.480925,
		0.510503,
		0.540059,
		0.569487,
		0.59868,
		0.62753,
		0.655929,
		0.68377,
		0.710944,
		0.737347,
		0.762874,
		0.787425,
		0.810904,
		0.833218,
		0.854281,
		0.874016,
		0.892353,
		0.909232,
		0.924608,
		0.938448,
		0.950736,
		0.961473,
		0.970681,
		0.978406,
		0.984714,
		0.9897,
		0.99348,
		0.996199,
		0.998019,
		0.999123,
		0.9997,
		0.999936,
		0.999996,
		1,
		0.999995,
		0.99993,
		0.999658,
		0.99896,
		0.997555,
		0.995126,
		0.99133,
		0.985814,
		0.978228,
		0.968237,
		0.955526,
		0.939815,
		0.920857,
		0.898452,
		0.872446,
		0.842738,
		0.809287,
		0.772112,
		0.731298,
		0.687003,
		0.63946,
		0.588982,
		0.535972,
		0.480925,
		0.424437,
		0.367211,
		0.310066,
		0.253947,
		0.199933,
		0.149248,
		0.103273,
		0.063557,
		0.031834,
		0.010032,
		0.000295,
		0.000982,
		0.003901,
		0.008711,
		0.015366,
		0.023816,
		0.034009,
		0.045892,
		0.059405,
		0.07449,
		0.091084,
		0.109121,
		0.128533,
		0.149248,
		0.171193,
		0.194292,
		0.218463,
		0.243624,
		0.269691,
		0.296573,
		0.324181,
		0.35242,
		0.381194,
		0.410404,
		0.439948,
		0.469722,
		0.499621,
		0.529537,
		0.559361,
		0.588982,
		0.618289,
		0.647171,
		0.675515,
		0.70321,
		0.730146,
		0.756216,
		0.781314,
		0.80534,
		0.828196,
		0.849792,
		0.870045,
		0.888882,
		0.90624,
		0.922066,
		0.936326,
		0.948998,
		0.960083,
		0.9696,
		0.977591,
		0.984125,
		0.989293,
		0.993217,
		0.996042,
		0.997936,
		0.999086,
		0.999687,
		0.999933,
		0.999995,
		1,
		0.999995,
		0.999927,
		0.999643,
		0.998913,
		0.997446,
		0.994912,
		0.990954,
		0.985208,
		0.977312,
		0.96692,
		0.953711,
		0.937398,
		0.917733,
		0.894513,
		0.86759,
		0.836868,
		0.802314,
		0.763962,
		0.721915,
		0.676351,
		0.62753,
		0.575799,
		0.521595,
		0.465459,
		0.408035,
		0.350084,
		0.292492,
		0.236278,
		0.182603,
		0.132788,
		0.088318,
		0.050862,
		0.022285,
		0.004665,
		0.000112,
		0.001783,
		0.005419,
		0.010974,
		0.018399,
		0.027641,
		0.038645,
		0.051353,
		0.065706,
		0.08164,
		0.099088,
		0.117983,
		0.13825,
		0.159817,
		0.182603,
		0.206529,
		0.231511,
		0.25746,
		0.284287,
		0.311899,
		0.3402,
		0.369091,
		0.39847,
		0.428234,
		0.458277,
		0.488488,
		0.518758,
		0.548975,
		0.579023,
		0.608789,
		0.638156,
		0.667009,
		0.695232,
		0.722712,
		0.749335,
		0.774992,
		0.799577,
		0.82299,
		0.845133,
		0.865921,
		0.885274,
		0.903125,
		0.919418,
		0.934112,
		0.947185,
		0.958631,
		0.968469,
		0.976739,
		0.983507,
		0.988867,
		0.992941,
		0.995877,
		0.997848,
		0.999046,
		0.999673,
		0.99993,
		0.999995,
		1,
		0.999995,
		0.999923,
		0.999626,
		0.998863,
		0.997332,
		0.994686,
		0.990558,
		0.984569,
		0.976347,
		0.965534,
		0.951803,
		0.93486,
		0.914455,
		0.890386,
		0.862506,
		0.830729,
		0.795032,
		0.755463,
		0.712144,
		0.665277,
		0.615152,
		0.562148,
		0.506746,
		0.449532,
		0.391206,
		0.332591,
		0.274643,
		0.218463,
		0.165303,
		0.116587,
		0.073915,
		0.039085,
		0.014108,
		0.001221,
		0.000458,
		0.002844,
		0.007225,
		0.013553,
		0.021774,
		0.031834,
		0.043674,
		0.057235,
		0.072452,
		0.089257,
		0.107582,
		0.127354,
		0.148496,
		0.170928,
		0.19457,
		0.219335,
		0.245133,
		0.271875,
		0.299464,
		0.327802,
		0.356789,
		0.38632,
		0.416289,
		0.446587,
		0.477101,
		0.507719,
		0.538323,
		0.568797,
		0.599022,
		0.628878,
		0.658245,
		0.687003,
		0.715035,
		0.742222,
		0.76845,
		0.793609,
		0.817592,
		0.840298,
		0.861636,
		0.881522,
		0.899882,
		0.916658,
		0.931803,
		0.945291,
		0.957114,
		0.967285,
		0.975845,
		0.982859,
		0.98842,
		0.992651,
		0.995704,
		0.997755,
		0.999004,
		0.999658,
		0.999927,
		0.999995,
		1,
		0.999995,
		0.99992,
		0.999609,
		0.998811,
		0.99721,
		0.994448,
		0.990141,
		0.983896,
		0.975331,
		0.964076,
		0.949797,
		0.932195,
		0.911016,
		0.886059,
		0.857183,
		0.824309,
		0.787425,
		0.746597,
		0.701966,
		0.653763,
		0.602308,
		0.548017,
		0.491417,
		0.433143,
		0.373956,
		0.314747,
		0.25655,
		0.200552,
		0.148107,
		0.100747,
		0.060202,
		0.02841,
		0.007538,
		0.001053,
		0.004182,
		0.009337,
		0.016464,
		0.02551,
		0.036415,
		0.049119,
		0.063557,
		0.079662,
		0.097363,
		0.116587,
		0.137255,
		0.159289,
		0.182603,
		0.207112,
		0.232725,
		0.259348,
		0.286884,
		0.315234,
		0.344293,
		0.373956,
		0.404113,
		0.434652,
		0.465459,
		0.496415,
		0.527402,
		0.558299,
		0.588982,
		0.619329,
		0.649214,
		0.678515,
		0.707107,
		0.734868,
		0.76168,
		0.787425,
		0.811993,
		0.835279,
		0.857183,
		0.877618,
		0.896506,
		0.913781,
		0.929393,
		0.943313,
		0.955526,
		0.966046,
		0.974909,
		0.982179,
		0.987949,
		0.992346,
		0.995521,
		0.997658,
		0.99896,
		0.999643,
		0.999923,
		0.999995,
		1,
		0.999995,
		0.999916,
		0.999591,
		0.998756,
		0.997082,
		0.994196,
		0.9897,
		0.983187,
		0.97426,
		0.962541,
		0.947687,
		0.929393,
		0.907405,
		0.881522,
		0.851607,
		0.817592,
		0.779477,
		0.737347,
		0.691366,
		0.641793,
		0.588982,
		0.533393,
		0.475597,
		0.416289,
		0.356293,
		0.296573,
		0.238249,
		0.182603,
		0.131098,
		0.085389,
		0.047341,
		0.019046,
		0.002844,
		0.00012,
		0.001913,
		0.005814,
		0.011771,
		0.019728,
		0.029626,
		0.041405,
		0.055,
		0.070342,
		0.08736,
		0.10598,
		0.126123,
		0.147708,
		0.170651,
		0.194862,
		0.22025,
		0.246719,
		0.27417,
		0.302502,
		0.331609,
		0.361381,
		0.391707,
		0.422472,
		0.453557,
		0.484843,
		0.516206,
		0.547521,
		0.578663,
		0.609502,
		0.63991,
		0.669758,
		0.698919,
		0.727265,
		0.754672,
		0.781018,
		0.806187,
		0.830067,
		0.852555,
		0.873557,
		0.892988,
		0.91078,
		0.926878,
		0.941245,
		0.953866,
		0.964748,
		0.973927,
		0.981465,
		0.987455,
		0.992024,
		0.995329,
		0.997555,
		0.998913,
		0.999626,
		0.99992,
		0.999995,
		1,
		0.999994,
		0.999912,
		0.999571,
		0.998697,
		0.996946,
		0.99393,
		0.989234,
		0.982438,
		0.97313,
		0.960924,
		0.945467,
		0.926448,
		0.903612,
		0.876762,
		0.845765,
		0.810562,
		0.771171,
		0.727694,
		0.680324,
		0.629348,
		0.575159,
		0.518262,
		0.459281,
		0.398972,
		0.338228,
		0.278096,
		0.219786,
		0.164687,
		0.114378,
		0.070648,
		0.035512,
		0.011231,
		0.000331,
		0.000492,
		0.003055,
		0.007758,
		0.014547,
		0.023363,
		0.034144,
		0.046826,
		0.061339,
		0.077611,
		0.095568,
		0.11513,
		0.136215,
		0.158736,
		0.182603,
		0.207724,
		0.234001,
		0.261333,
		0.289616,
		0.318742,
		0.348599,
		0.379073,
		0.410046,
		0.441396,
		0.473,
		0.504731,
		0.53646,
		0.568056,
		0.599388,
		0.630323,
		0.660725,
		0.690463,
		0.719404,
		0.747418,
		0.774379,
		0.800163,
		0.824654,
		0.847743,
		0.869329,
		0.889324,
		0.907651,
		0.924251,
		0.939084,
		0.952128,
		0.963389,
		0.972897,
		0.980715,
		0.986936,
		0.991686,
		0.995126,
		0.997446,
		0.998863,
		0.999609,
		0.999916,
		0.999994,
		1,
		0.999994,
		0.999908,
		0.999551,
		0.998635,
		0.996803,
		0.993648,
		0.988742,
		0.981647,
		0.971939,
		0.95922,
		0.943128,
		0.92335,
		0.899627,
		0.871765,
		0.83964,
		0.803203,
		0.762488,
		0.717621,
		0.668821,
		0.61641,
		0.560823,
		0.502614,
		0.442464,
		0.381194,
		0.319778,
		0.259348,
		0.201216,
		0.146884,
		0.098062,
		0.056682,
		0.024926,
		0.005236,
		0.001133,
		0.004495,
		0.010032,
		0.017685,
		0.027391,
		0.039085,
		0.052699,
		0.068159,
		0.085389,
		0.10431,
		0.124837,
		0.146884,
		0.17036,
		0.195168,
		0.221211,
		0.248386,
		0.276585,
		0.3057,
		0.335615,
		0.366214,
		0.397375,
		0.428973,
		0.460882,
		0.492972,
		0.525108,
		0.557157,
		0.588982,
		0.620445,
		0.651406,
		0.681729,
		0.711275,
		0.739908,
		0.767497,
		0.793912,
		0.819031,
		0.842738,
		0.864928,
		0.885504,
		0.904386,
		0.921508,
		0.936823,
		0.950308,
		0.961963,
		0.971816,
		0.979927,
		0.986389,
		0.99133,
		0.994912,
		0.997332,
		0.998811,
		0.999591,
		0.999912,
		0.999994,
		1,
		0.999994,
		0.999903,
		0.999529,
		0.998569,
		0.996651,
		0.99335,
		0.988221,
		0.980811,
		0.970681,
		0.957423,
		0.940665,
		0.92009,
		0.895439,
		0.86652,
		0.833218,
		0.795497,
		0.75341,
		0.707107,
		0.656837,
		0.602961,
		0.545959,
		0.486437,
		0.425141,
		0.362964,
		0.300964,
		0.24037,
		0.182603,
		0.129291,
		0.082283,
		0.043674,
		0.015823,
		0.001374,
		0.00013,
		0.002059,
		0.006254,
		0.012657,
		0.021206,
		0.031834,
		0.044472,
		0.059048,
		0.075484,
		0.093699,
		0.113609,
		0.135126,
		0.158156,
		0.182603,
		0.208367,
		0.235343,
		0.263423,
		0.292492,
		0.322436,
		0.353134,
		0.38446,
		0.416289,
		0.448489,
		0.480925,
		0.513462,
		0.545959,
		0.578275,
		0.610268,
		0.641793,
		0.672708,
		0.702868,
		0.732133,
		0.760363,
		0.787425,
		0.813189,
		0.837533,
		0.860344,
		0.881522,
		0.900978,
		0.918641,
		0.934459,
		0.948403,
		0.960468,
		0.970681,
		0.979099,
		0.985814,
		0.990954,
		0.994686,
		0.99721,
		0.998756,
		0.999571,
		0.999908,
		0.999994,
		1,
		0.999993,
		0.999898,
		0.999505,
		0.998499,
		0.99649,
		0.993035,
		0.98767,
		0.979927,
		0.969353,
		0.955526,
		0.938069,
		0.916658,
		0.891034,
		0.861011,
		0.826481,
		0.787425,
		0.743917,
		0.696131,
		0.644352,
		0.588982,
		0.53055,
		0.469722,
		0.407312,
		0.344293,
		0.281814,
		0.221211,
		0.164025,
		0.112019,
		0.067199,
		0.031834,
		0.008478,
		0.00053,
		0.003289,
		0.008351,
		0.015654,
		0.025132,
		0.036715,
		0.05033,
		0.065898,
		0.083339,
		0.102567,
		0.123492,
		0.14602,
		0.170054,
		0.19549,
		0.222223,
		0.250141,
		0.27913,
		0.30907,
		0.339837,
		0.371306,
		0.403344,
		0.435818,
		0.468589,
		0.501515,
		0.534454,
		0.567259,
		0.599783,
		0.631876,
		0.663389,
		0.694173,
		0.724081,
		0.752968,
		0.780692,
		0.807118,
		0.832117,
		0.85557,
		0.877369,
		0.89742,
		0.915644,
		0.931984,
		0.946406,
		0.9589,
		0.969489,
		0.978228,
		0.985208,
		0.990558,
		0.994448,
		0.997082,
		0.998697,
		0.999551,
		0.999903,
		0.999993,
		1,
		0.999993,
		0.999893,
		0.999481,
		0.998425,
		0.996318,
		0.9927,
		0.987087,
		0.978993,
		0.96795,
		0.953525,
		0.935331,
		0.913042,
		0.8864,
		0.855222,
		0.819413,
		0.778969,
		0.733987,
		0.684672,
		0.631345,
		0.574455,
		0.514584,
		0.452462,
		0.38898,
		0.325198,
		0.262364,
		0.201932,
		0.145573,
		0.0952,
		0.05299,
		0.021402,
		0.003208,
		0.001221,
		0.004845,
		0.010808,
		0.019046,
		0.029488,
		0.042059,
		0.056682,
		0.073275,
		0.091751,
		0.112019,
		0.133985,
		0.157547,
		0.182603,
		0.209044,
		0.236757,
		0.265625,
		0.295525,
		0.326332,
		0.357915,
		0.39014,
		0.422868,
		0.455958,
		0.489264,
		0.522638,
		0.555928,
		0.588982,
		0.621646,
		0.653763,
		0.68518,
		0.715743,
		0.745299,
		0.773702,
		0.800807,
		0.826481,
		0.850596,
		0.873037,
		0.893703,
		0.91251,
		0.929393,
		0.944313,
		0.957254,
		0.968237,
		0.977312,
		0.984569,
		0.990141,
		0.994196,
		0.996946,
		0.998635,
		0.999529,
		0.999898,
		0.999993,
		1,
		0.999993,
		0.999888,
		0.999454,
		0.998347,
		0.996137,
		0.992346,
		0.986469,
		0.978003,
		0.966466,
		0.95141,
		0.932443,
		0.909232,
		0.881522,
		0.849137,
		0.811993,
		0.770107,
		0.723599,
		0.672708,
		0.617796,
		0.559361,
		0.498046,
		0.434652,
		0.370151,
		0.3057,
		0.242658,
		0.182603,
		0.127354,
		0.078987,
		0.039864,
		0.012657,
		0.000375,
		0.00014,
		0.002221,
		0.006746,
		0.013648,
		0.022856,
		0.034297,
		0.047892,
		0.063557,
		0.081207,
		0.100747,
		0.122084,
		0.145114,
		0.169733,
		0.19583,
		0.223289,
		0.251992,
		0.281814,
		0.312626,
		0.344293,
		0.376679,
		0.409641,
		0.443032,
		0.476704,
		0.510503,
		0.544274,
		0.577857,
		0.611093,
		0.643821,
		0.675879,
		0.707107,
		0.737347,
		0.766443,
		0.794247,
		0.820615,
		0.845412,
		0.868517,
		0.889821,
		0.909232,
		0.92668,
		0.942118,
		0.955526,
		0.96692,
		0.976347,
		0.983896,
		0.9897,
		0.99393,
		0.996803,
		0.998569,
		0.999505,
		0.999893,
		0.999993,
		1,
		0.999992,
		0.999882,
		0.999426,
		0.998263,
		0.995944,
		0.991969,
		0.985814,
		0.976956,
		0.964896,
		0.949176,
		0.929393,
		0.905215,
		0.876386,
		0.842738,
		0.804203,
		0.760817,
		0.712729,
		0.660215,
		0.603682,
		0.543682,
		0.480925,
		0.416289,
		0.350837,
		0.28583,
		0.222749,
		0.163312,
		0.109496,
		0.063557,
		0.028064,
		0.00592,
		0.000573,
		0.003552,
		0.009016,
		0.016893,
		0.02711,
		0.039587,
		0.05424,
		0.070982,
		0.08972,
		0.110356,
		0.132788,
		0.156908,
		0.182603,
		0.209758,
		0.238249,
		0.26795,
		0.298728,
		0.330446,
		0.362964,
		0.396135,
		0.429809,
		0.463832,
		0.498046,
		0.53229,
		0.566399,
		0.600208,
		0.63355,
		0.666257,
		0.698162,
		0.729099,
		0.758906,
		0.787425,
		0.814507,
		0.840008,
		0.863799,
		0.885764,
		0.905802,
		0.923837,
		0.939815,
		0.953711,
		0.965534,
		0.975331,
		0.983187,
		0.989234,
		0.993648,
		0.996651,
		0.998499,
		0.999481,
		0.999888,
		0.999992,
		1,
		0.999992,
		0.999876,
		0.999397,
		0.998174,
		0.995739,
		0.991569,
		0.985119,
		0.975845,
		0.963233,
		0.946813,
		0.926173,
		0.900978,
		0.870975,
		0.836008,
		0.796022,
		0.751077,
		0.701355,
		0.647171,
		0.588982,
		0.527402,
		0.463212,
		0.397375,
		0.331053,
		0.265625,
		0.202704,
		0.144161,
		0.092147,
		0.049119,
		0.01787,
		0.001559,
		0.00132,
		0.005236,
		0.011678,
		0.02057,
		0.031834,
		0.045385,
		0.061133,
		0.078987,
		0.098846,
		0.120607,
		0.144161,
		0.169394,
		0.196187,
		0.224414,
		0.253947,
		0.28465,
		0.316383,
		0.349001,
		0.382355,
		0.416289,
		0.450646,
		0.485261,
		0.519969,
		0.5546,
		0.588982,
		0.622942,
		0.656305,
		0.688896,
		0.720543,
		0.751077,
		0.78033,
		0.808146,
		0.834373,
		0.858873,
		0.881522,
		0.902212,
		0.920857,
		0.937398,
		0.951803,
		0.964076,
		0.97426,
		0.982438,
		0.988742,
		0.99335,
		0.99649,
		0.998425,
		0.999454,
		0.999882,
		0.999992,
		1,
		0.999991,
		0.999869,
		0.999365,
		0.998079,
		0.995521,
		0.991144,
		0.984381,
		0.974668,
		0.961473,
		0.944313,
		0.92277,
		0.896506,
		0.865273,
		0.828924,
		0.787425,
		0.740862,
		0.68945,
		0.63355,
		0.573675,
		0.510503,
		0.444897,
		0.377913,
		0.310823,
		0.245133,
		0.182603,
		0.125272,
		0.075483,
		0.035914,
		0.009606,
		0.000152,
		0.002404,
		0.007298,
		0.014759,
		0.024706,
		0.037057,
		0.05172,
		0.068601,
		0.087601,
		0.108615,
		0.131531,
		0.156234,
		0.182603,
		0.210511,
		0.239825,
		0.270407,
		0.302114,
		0.334797,
		0.368303,
		0.402472,
		0.437142,
		0.472143,
		0.507305,
		0.542452,
		0.577405,
		0.611985,
		0.64601,
		0.679298,
		0.711668,
		0.742944,
		0.77295,
		0.801521,
		0.828496,
		0.853729,
		0.877086,
		0.898452,
		0.917733,
		0.93486,
		0.949797,
		0.962541,
		0.97313,
		0.981647,
		0.988221,
		0.993035,
		0.996318,
		0.998347,
		0.999426,
		0.999876,
		0.999991,
		1,
		0.999991,
		0.999862,
		0.999331,
		0.997978,
		0.995289,
		0.990693,
		0.983597,
		0.973418,
		0.959607,
		0.941666,
		0.919171,
		0.891784,
		0.85926,
		0.821467,
		0.778391,
		0.730146,
		0.676989,
		0.619329,
		0.557738,
		0.492972,
		0.425977,
		0.357915,
		0.290179,
		0.224414,
		0.162543,
		0.106789,
		0.059709,
		0.024222,
		0.003646,
		0.000621,
		0.003848,
		0.009762,
		0.018285,
		0.02933,
		0.042808,
		0.058623,
		0.076675,
		0.096858,
		0.119058,
		0.14316,
		0.169038,
		0.196564,
		0.225603,
		0.256014,
		0.287651,
		0.32036,
		0.353984,
		0.38836,
		0.42332,
		0.45869,
		0.494292,
		0.529947,
		0.565468,
		0.600669,
		0.63536,
		0.669354,
		0.702461,
		0.734494,
		0.765272,
		0.794618,
		0.822365,
		0.848355,
		0.872446,
		0.894513,
		0.914455,
		0.932195,
		0.947687,
		0.960924,
		0.971939,
		0.980811,
		0.98767,
		0.9927,
		0.996137,
		0.998263,
		0.999397,
		0.999869,
		0.999991,
		1,
		0.999991,
		0.999854,
		0.999295,
		0.99787,
		0.995042,
		0.990212,
		0.982764,
		0.972091,
		0.957627,
		0.938862,
		0.915365,
		0.886795,
		0.852917,
		0.813614,
		0.768893,
		0.718904,
		0.663946,
		0.604481,
		0.541152,
		0.474793,
		0.40645,
		0.337397,
		0.269161,
		0.20354,
		0.142636,
		0.088879,
		0.045061,
		0.01437,
		0.000428,
		0.001432,
		0.005678,
		0.012657,
		0.022285,
		0.034471,
		0.049119,
		0.066128,
		0.085389,
		0.106789,
		0.13021,
		0.155525,
		0.182603,
		0.211307,
		0.241492,
		0.273008,
		0.3057,
		0.339405,
		0.373956,
		0.409179,
		0.444897,
		0.480925,
		0.517077,
		0.553161,
		0.588982,
		0.624345,
		0.659052,
		0.692907,
		0.725714,
		0.757282,
		0.787425,
		0.815966,
		0.842738,
		0.86759,
		0.890386,
		0.911016,
		0.929393,
		0.945467,
		0.95922,
		0.970681,
		0.979927,
		0.987087,
		0.992346,
		0.995944,
		0.998174,
		0.999365,
		0.999862,
		0.999991,
		1,
		0.99999,
		0.999846,
		0.999256,
		0.997755,
		0.994778,
		0.9897,
		0.981877,
		0.970681,
		0.955526,
		0.93589,
		0.911335,
		0.881522,
		0.846224,
		0.80534,
		0.758906,
		0.707107,
		0.650291,
		0.588982,
		0.523896,
		0.455958,
		0.38632,
		0.316381,
		0.247816,
		0.182595,
		0.123016,
		0.071739,
		0.031819,
		0.00674,
		0.000165,
		0.00261,
		0.007921,
		0.016011,
		0.026791,
		0.040162,
		0.056024,
		0.074267,
		0.094776,
		0.117431,
		0.142104,
		0.168662,
		0.196963,
		0.226862,
		0.258204,
		0.29083,
		0.324575,
		0.359265,
		0.394723,
		0.430764,
		0.4672,
		0.503836,
		0.540475,
		0.576915,
		0.612952,
		0.64838,
		0.682994,
		0.71659,
		0.748966,
		0.779928,
		0.809287,
		0.836868,
		0.862506,
		0.886059,
		0.907405,
		0.926448,
		0.943128,
		0.957423,
		0.969353,
		0.978993,
		0.986469,
		0.991969,
		0.995739,
		0.998079,
		0.999331,
		0.999854,
		0.99999,
		1,
		0.999989,
		0.999838,
		0.999215,
		0.997633,
		0.994497,
		0.989154,
		0.980933,
		0.969182,
		0.953295,
		0.932739,
		0.907068,
		0.875946,
		0.839157,
		0.796619,
		0.7484,
		0.694725,
		0.635997,
		0.572806,
		0.505951,
		0.436456,
		0.365589,
		0.294891,
		0.226198,
		0.16167,
		0.103831,
		0.055596,
		0.020311,
		0.001778,
		0.000675,
		0.004182,
		0.010606,
		0.019856,
		0.031834,
		0.046438,
		0.063557,
		0.083077,
		0.104874,
		0.128819,
		0.154777,
		0.182603,
		0.212149,
		0.243258,
		0.275766,
		0.309504,
		0.344293,
		0.379951,
		0.416289,
		0.453111,
		0.490217,
		0.527402,
		0.564456,
		0.601168,
		0.637323,
		0.672708,
		0.707107,
		0.74031,
		0.772112,
		0.802314,
		0.830729,
		0.857183,
		0.881522,
		0.903612,
		0.92335,
		0.940665,
		0.955526,
		0.96795,
		0.978003,
		0.985814,
		0.991569,
		0.995521,
		0.997978,
		0.999295,
		0.999846,
		0.999989,
		1,
		0.999989,
		0.999829,
		0.999171,
		0.997501,
		0.994196,
		0.988571,
		0.979927,
		0.967586,
		0.950924,
		0.929393,
		0.902545,
		0.870045,
		0.831692,
		0.787425,
		0.737347,
		0.681729,
		0.621033,
		0.555925,
		0.487293,
		0.416269,
		0.344254,
		0.272942,
		0.204353,
		0.140868,
		0.085259,
		0.040727,
		0.01093,
		0.001559,
		0.006177,
		0.013765,
		0.024223,
		0.03745,
		0.053333,
		0.071756,
		0.092596,
		0.11572,
		0.140991,
		0.168264,
		0.197385,
		0.228195,
		0.260527,
		0.294205,
		0.32905,
		0.364871,
		0.401474,
		0.438658,
		0.476215,
		0.513933,
		0.551596,
		0.588982,
		0.625869,
		0.662033,
		0.69725,
		0.731298,
		0.763962,
		0.795032,
		0.824309,
		0.851607,
		0.876762,
		0.899627,
		0.92009,
		0.938069,
		0.953525,
		0.966466,
		0.976956,
		0.985119,
		0.991144,
		0.995289,
		0.99787,
		0.999256,
		0.999838,
		0.999989,
		1,
		0.999988,
		0.999819,
		0.999123,
		0.997361,
		0.993875,
		0.987949,
		0.978855,
		0.965887,
		0.948403,
		0.925841,
		0.897749,
		0.863799,
		0.823803,
		0.777728,
		0.725713,
		0.668082,
		0.605362,
		0.538297,
		0.467878,
		0.395357,
		0.322282,
		0.250527,
		0.182325,
		0.120303,
		0.067515,
		0.027471,
		0.004144,
		0.000179,
		0.002844,
		0.008627,
		0.01743,
		0.02915,
		0.043674,
		0.060886,
		0.08066,
		0.102863,
		0.127354,
		0.153986,
		0.182603,
		0.213043,
		0.245133,
		0.278696,
		0.313545,
		0.349487,
		0.38632,
		0.423837,
		0.461824,
		0.500061,
		0.538323,
		0.576381,
		0.614003,
		0.650955,
		0.687003,
		0.721915,
		0.755463,
		0.787425,
		0.817592,
		0.845765,
		0.871765,
		0.895439,
		0.916658,
		0.935331,
		0.95141,
		0.964896,
		0.975845,
		0.984381,
		0.990693,
		0.995042,
		0.997755,
		0.999215,
		0.999829,
		0.999988,
		1,
		0.999987,
		0.999808,
		0.999073,
		0.99721,
		0.993531,
		0.987285,
		0.97771,
		0.964076,
		0.945719,
		0.922066,
		0.892661,
		0.857183,
		0.815461,
		0.767493,
		0.713457,
		0.653732,
		0.588915,
		0.519843,
		0.447615,
		0.373628,
		0.299601,
		0.227619,
		0.160159,
		0.100129,
		0.050887,
		0.016249,
		0.000483,
		0.000737,
		0.004562,
		0.011564,
		0.021638,
		0.034672,
		0.050548,
		0.069138,
		0.090309,
		0.113919,
		0.139816,
		0.167843,
		0.197833,
		0.229612,
		0.262996,
		0.297795,
		0.333809,
		0.370832,
		0.40865,
		0.447041,
		0.485779,
		0.524629,
		0.563354,
		0.601712,
		0.63946,
		0.676351,
		0.712144,
		0.746597,
		0.779477,
		0.810562,
		0.83964,
		0.86652,
		0.891034,
		0.913042,
		0.932443,
		0.949176,
		0.963233,
		0.974668,
		0.983597,
		0.990212,
		0.994778,
		0.997633,
		0.999171,
		0.999819,
		0.999987,
		1,
		0.999987,
		0.999797,
		0.999018,
		0.997049,
		0.993163,
		0.986574,
		0.976488,
		0.962145,
		0.942861,
		0.918052,
		0.887259,
		0.850168,
		0.806627,
		0.756663,
		0.700499,
		0.638572,
		0.57156,
		0.500403,
		0.426337,
		0.350925,
		0.276097,
		0.204182,
		0.137938,
		0.080573,
		0.035748,
		0.007563,
		0.001703,
		0.006746,
		0.015024,
		0.026425,
		0.04083,
		0.058111,
		0.078131,
		0.100747,
		0.125807,
		0.153149,
		0.182603,
		0.213992,
		0.247127,
		0.281814,
		0.317848,
		0.355015,
		0.393097,
		0.431863,
		0.471079,
		0.510503,
		0.549888,
		0.588982,
		0.62753,
		0.665277,
		0.701966,
		0.737347,
		0.771171,
		0.803203,
		0.833218,
		0.861011,
		0.8864,
		0.909232,
		0.929393,
		0.946813,
		0.961473,
		0.973418,
		0.982764,
		0.9897,
		0.994497,
		0.997501,
		0.999123,
		0.999808,
		0.999987,
		1,
		0.999986,
		0.999784,
		0.99896,
		0.996876,
		0.992769,
		0.985814,
		0.975181,
		0.960083,
		0.939814,
		0.913775,
		0.881506,
		0.842698,
		0.797216,
		0.745119,
		0.686676,
		0.622395,
		0.55305,
		0.479711,
		0.40378,
		0.327031,
		0.25164,
		0.180217,
		0.115822,
		0.061965,
		0.022579,
		0.001968,
		0.000196,
		0.003111,
		0.009431,
		0.019046,
		0.031834,
		0.047667,
		0.066407,
		0.08791,
		0.112019,
		0.138573,
		0.167396,
		0.198309,
		0.231119,
		0.265625,
		0.301618,
		0.33888,
		0.377182,
		0.416289,
		0.455958,
		0.495938,
		0.535972,
		0.575799,
		0.615152,
		0.653763,
		0.691366,
		0.727694,
		0.762488,
		0.795497,
		0.826481,
		0.855222,
		0.881522,
		0.905215,
		0.926173,
		0.944313,
		0.959607,
		0.972091,
		0.981877,
		0.989154,
		0.994196,
		0.997361,
		0.999073,
		0.999797,
		0.999986,
		1,
		0.999985,
		0.999771,
		0.998897,
		0.99669,
		0.992346,
		0.984999,
		0.973782,
		0.957876,
		0.936549,
		0.909189,
		0.875324,
		0.83465,
		0.78705,
		0.732617,
		0.671681,
		0.604841,
		0.532994,
		0.457378,
		0.379606,
		0.301705,
		0.226135,
		0.155812,
		0.094096,
		0.044763,
		0.011938,
		0.000807,
		0.004996,
		0.012657,
		0.02367,
		0.037906,
		0.055226,
		0.075484,
		0.098521,
		0.124172,
		0.152262,
		0.182603,
		0.215002,
		0.249252,
		0.285138,
		0.322436,
		0.360911,
		0.40032,
		0.440412,
		0.480925,
		0.521595,
		0.562148,
		0.602308,
		0.641793,
		0.680324,
		0.717621,
		0.75341,
		0.787425,
		0.819413,
		0.849137,
		0.876386,
		0.900978,
		0.92277,
		0.941666,
		0.957627,
		0.970681,
		0.980933,
		0.988571,
		0.993875,
		0.99721,
		0.999018,
		0.999784,
		0.999985,
		1,
		0.999984,
		0.999757,
		0.998829,
		0.996489,
		0.99189,
		0.984119,
		0.972268,
		0.955481,
		0.932992,
		0.904167,
		0.868522,
		0.825756,
		0.775775,
		0.718723,
		0.655014,
		0.585368,
		0.510855,
		0.432925,
		0.353452,
		0.274752,
		0.199599,
		0.131208,
		0.073204,
		0.029542,
		0.00441,
		0.001868,
		0.007397,
		0.016464,
		0.028941,
		0.044689,
		0.063557,
		0.085389,
		0.110014,
		0.137255,
		0.166922,
		0.198815,
		0.232725,
		0.26843,
		0.3057,
		0.344293,
		0.383959,
		0.424437,
		0.465459,
		0.506746,
		0.548017,
		0.588982,
		0.629348,
		0.668821,
		0.707107,
		0.743917,
		0.778969,
		0.811993,
		0.842738,
		0.870975,
		0.896506,
		0.919171,
		0.938862,
		0.955526,
		0.969182,
		0.979927,
		0.987949,
		0.993531,
		0.997049,
		0.99896,
		0.999771,
		0.999984,
		1,
		0.999983,
		0.999741,
		0.998754,
		0.996268,
		0.991385,
		0.983139,
		0.970571,
		0.952776,
		0.92895,
		0.898429,
		0.860718,
		0.815526,
		0.762802,
		0.702768,
		0.635962,
		0.563276,
		0.486002,
		0.405858,
		0.325018,
		0.246117,
		0.172233,
		0.106843,
		0.053745,
		0.016943,
		0.000496,
		0.000216,
		0.003417,
		0.010354,
		0.020897,
		0.034906,
		0.05223,
		0.07271,
		0.096175,
		0.122442,
		0.151319,
		0.182603,
		0.216079,
		0.25152,
		0.28869,
		0.32734,
		0.367211,
		0.408035,
		0.449532,
		0.491417,
		0.533393,
		0.575159,
		0.61641,
		0.656837,
		0.696131,
		0.733987,
		0.770107,
		0.804203,
		0.836008,
		0.865273,
		0.891784,
		0.915365,
		0.93589,
		0.953295,
		0.967586,
		0.978855,
		0.987285,
		0.993163,
		0.996876,
		0.998897,
		0.999757,
		0.999983,
		1,
		0.99998,
		0.999719,
		0.998659,
		0.995994,
		0.990764,
		0.981941,
		0.968498,
		0.949475,
		0.924021,
		0.891446,
		0.851254,
		0.803188,
		0.747271,
		0.683852,
		0.613644,
		0.537774,
		0.457808,
		0.375779,
		0.294186,
		0.215973,
		0.14448,
		0.083354,
		0.036428,
		0.007571,
		0.000888,
		0.005495,
		0.013913,
		0.026003,
		0.041614,
		0.060583,
		0.082738,
		0.107895,
		0.135857,
		0.166417,
		0.199356,
		0.234441,
		0.271429,
		0.310066,
		0.350084,
		0.391206,
		0.433143,
		0.475597,
		0.518262,
		0.560823,
		0.602961,
		0.644352,
		0.684672,
		0.723599,
		0.760817,
		0.796022,
		0.828924,
		0.85926,
		0.886795,
		0.911335,
		0.932739,
		0.950924,
		0.965887,
		0.97771,
		0.986574,
		0.992769,
		0.99669,
		0.998829,
		0.999741,
		0.99998,
		0.999994,
		0.999963,
		0.999659,
		0.998478,
		0.99555,
		0.989838,
		0.980234,
		0.965637,
		0.945023,
		0.917501,
		0.882368,
		0.839153,
		0.787675,
		0.728083,
		0.660911,
		0.587116,
		0.508113,
		0.425793,
		0.342523,
		0.26112,
		0.184789,
		0.117035,
		0.061533,
		0.021973,
		0.001872,
		0.002059,
		0.008146,
		0.018122,
		0.031834,
		0.049119,
		0.069803,
		0.093699,
		0.120607,
		0.150317,
		0.182603,
		0.21723,
		0.253947,
		0.292492,
		0.332591,
		0.373956,
		0.416289,
		0.459281,
		0.502614,
		0.545959,
		0.588982,
		0.631345,
		0.672708,
		0.712729,
		0.751077,
		0.787425,
		0.821467,
		0.852917,
		0.881522,
		0.907068,
		0.929393,
		0.948403,
		0.964076,
		0.976488,
		0.985814,
		0.992346,
		0.996489,
		0.998754,
		0.999719,
		0.999963,
		0.999952,
		0.999866,
		0.999445,
		0.998023,
		0.994648,
		0.98819,
		0.977452,
		0.961255,
		0.938519,
		0.908329,
		0.870001,
		0.82314,
		0.767694,
		0.704007,
		0.632867,
		0.555533,
		0.473757,
		0.389776,
		0.30628,
		0.226346,
		0.153342,
		0.090795,
		0.042229,
		0.01098,
		0.000238,
		0.003771,
		0.011419,
		0.023032,
		0.038444,
		0.057479,
		0.079949,
		0.105651,
		0.134371,
		0.165879,
		0.199933,
		0.236278,
		0.274643,
		0.314747,
		0.356293,
		0.398972,
		0.442464,
		0.486437,
		0.53055,
		0.574455,
		0.617796,
		0.660215,
		0.701355,
		0.740862,
		0.778391,
		0.813614,
		0.846224,
		0.875946,
		0.902545,
		0.925841,
		0.945719,
		0.962145,
		0.975181,
		0.984999,
		0.99189,
		0.996268,
		0.998659,
		0.999659,
		0.999866,
		0.999758,
		0.999501,
		0.998789,
		0.996877,
		0.992715,
		0.985074,
		0.972665,
		0.954239,
		0.928683,
		0.895097,
		0.852871,
		0.801745,
		0.741871,
		0.67386,
		0.598814,
		0.518344,
		0.434554,
		0.35001,
		0.267665,
		0.190753,
		0.122659,
		0.066752,
		0.026203,
		0.003794,
		0.000982,
		0.006072,
		0.015366,
		0.028698,
		0.045892,
		0.066755,
		0.091084,
		0.118659,
		0.149248,
		0.182603,
		0.218463,
		0.25655,
		0.296573,
		0.338228,
		0.381194,
		0.425141,
		0.469722,
		0.514584,
		0.559361,
		0.603682,
		0.647171,
		0.68945,
		0.730146,
		0.768893,
		0.80534,
		0.839157,
		0.870045,
		0.897749,
		0.922066,
		0.942861,
		0.960083,
		0.973782,
		0.984119,
		0.991385,
		0.995994,
		0.998478,
		0.999445,
		0.999501,
		0.999123,
		0.998451,
		0.997118,
		0.994288,
		0.988808,
		0.979348,
		0.964539,
		0.943094,
		0.913908,
		0.876156,
		0.829368,
		0.773497,
		0.708972,
		0.636733,
		0.558244,
		0.475478,
		0.390871,
		0.30725,
		0.227721,
		0.155529,
		0.093895,
		0.045838,
		0.013986,
		0.000395,
		0.00228,
		0.009016,
		0.020042,
		0.035182,
		0.05424,
		0.077011,
		0.103273,
		0.132788,
		0.165303,
		0.200552,
		0.238249,
		0.278096,
		0.319778,
		0.362964,
		0.407312,
		0.452462,
		0.498046,
		0.543682,
		0.588982,
		0.63355,
		0.676989,
		0.718904,
		0.758906,
		0.796619,
		0.831692,
		0.863799,
		0.892661,
		0.918052,
		0.939814,
		0.957876,
		0.972268,
		0.983139,
		0.990764,
		0.99555,
		0.998023,
		0.998789,
		0.998451,
		0.997474,
		0.995965,
		0.993485,
		0.989105,
		0.981571,
		0.969472,
		0.951392,
		0.92605,
		0.892416,
		0.84981,
		0.797981,
		0.737171,
		0.66815,
		0.592226,
		0.511233,
		0.427476,
		0.34365,
		0.262731,
		0.187825,
		0.122012,
		0.068164,
		0.028769,
		0.005759,
		0.000264,
		0.004182,
		0.012657,
		0.02551,
		0.042546,
		0.063557,
		0.088318,
		0.116587,
		0.148107,
		0.182603,
		0.219786,
		0.259348,
		0.300964,
		0.344293,
		0.38898,
		0.434652,
		0.480925,
		0.527402,
		0.573675,
		0.619329,
		0.663946,
		0.707107,
		0.7484,
		0.787425,
		0.823803,
		0.857183,
		0.887259,
		0.913775,
		0.936549,
		0.955481,
		0.970571,
		0.981941,
		0.989838,
		0.994648,
		0.996877,
		0.997118,
		0.995965,
		0.993865,
		0.990887,
		0.986527,
		0.97977,
		0.969283,
		0.953609,
		0.931338,
		0.901261,
		0.862495,
		0.814587,
		0.757581,
		0.692067,
		0.61919,
		0.540631,
		0.458554,
		0.375519,
		0.294361,
		0.218049,
		0.14952,
		0.091506,
		0.046363,
		0.015912,
		0.001299,
		0.001092,
		0.006746,
		0.017058,
		0.031834,
		0.050862,
		0.073915,
		0.100747,
		0.131098,
		0.164687,
		0.201216,
		0.24037,
		0.281814,
		0.325198,
		0.370151,
		0.416289,
		0.463212,
		0.510503,
		0.557738,
		0.604481,
		0.650291,
		0.694725,
		0.737347,
		0.777728,
		0.815461,
		0.850168,
		0.881506,
		0.909189,
		0.932992,
		0.952776,
		0.968498,
		0.980234,
		0.98819,
		0.992715,
		0.994288,
		0.993485,
		0.990887,
		0.986931,
		0.981659,
		0.974514,
		0.964421,
		0.950009,
		0.929828,
		0.902538,
		0.86708,
		0.822798,
		0.769537,
		0.707694,
		0.638232,
		0.56267,
		0.483015,
		0.401677,
		0.321345,
		0.244843,
		0.174964,
		0.1143,
		0.065084,
		0.029031,
		0.007219,
		0.002539,
		0.010032,
		0.022285,
		0.039085,
		0.060202,
		0.085389,
		0.114378,
		0.146884,
		0.182603,
		0.221211,
		0.262364,
		0.3057,
		0.350837,
		0.397375,
		0.444897,
		0.492972,
		0.541152,
		0.588982,
		0.635997,
		0.681729,
		0.725713,
		0.767493,
		0.806627,
		0.842698,
		0.875324,
		0.904167,
		0.92895,
		0.949475,
		0.965637,
		0.977452,
		0.985074,
		0.988808,
		0.989105,
		0.986527,
		0.981659,
		0.974941,
		0.966413,
		0.955494,
		0.941091,
		0.921848,
		0.896384,
		0.863508,
		0.822383,
		0.77265,
		0.714501,
		0.648705,
		0.576593,
		0.5,
		0.42117,
		0.342631,
		0.267053,
		0.197077,
		0.135159,
		0.083406,
		0.043437,
		0.01627,
		0.002241,
		0.000295,
		0.004665,
		0.014108,
		0.02841,
		0.047341,
		0.070648,
		0.098062,
		0.129291,
		0.164025,
		0.201932,
		0.242658,
		0.28583,
		0.331053,
		0.377913,
		0.425977,
		0.474793,
		0.523896,
		0.572806,
		0.621033,
		0.668082,
		0.713457,
		0.756663,
		0.797216,
		0.83465,
		0.868522,
		0.898429,
		0.924021,
		0.945023,
		0.961255,
		0.972665,
		0.979348,
		0.981571,
		0.97977,
		0.974514,
		0.966413,
		0.955945,
		0.943177,
		0.927549,
		0.907999,
		0.883247,
		0.852052,
		0.813438,
		0.766854,
		0.712278,
		0.650267,
		0.581942,
		0.50894,
		0.433319,
		0.357431,
		0.283774,
		0.214834,
		0.152922,
		0.100022,
		0.057659,
		0.026795,
		0.007757,
		0.000207,
		0.001221,
		0.007538,
		0.019046,
		0.035512,
		0.056682,
		0.082283,
		0.112019,
		0.145573,
		0.182603,
		0.222749,
		0.265625,
		0.310823,
		0.357915,
		0.40645,
		0.455958,
		0.505951,
		0.555925,
		0.605362,
		0.653732,
		0.700499,
		0.745119,
		0.78705,
		0.825756,
		0.860718,
		0.891446,
		0.917501,
		0.938519,
		0.954239,
		0.964539,
		0.969472,
		0.969283,
		0.964421,
		0.955494,
		0.943177,
		0.928019,
		0.91016,
		0.889104,
		0.863875,
		0.833327,
		0.79642,
		0.752446,
		0.701172,
		0.642924,
		0.578592,
		0.509589,
		0.43776,
		0.365257,
		0.294395,
		0.227488,
		0.166698,
		0.113883,
		0.070476,
		0.037384,
		0.014927,
		0.002816,
		0.002844,
		0.011231,
		0.024926,
		0.043674,
		0.067199,
		0.0952,
		0.127354,
		0.163312,
		0.202704,
		0.245133,
		0.290179,
		0.337397,
		0.38632,
		0.436456,
		0.487293,
		0.538297,
		0.588915,
		0.638572,
		0.686676,
		0.732617,
		0.775775,
		0.815526,
		0.851254,
		0.882368,
		0.908329,
		0.928683,
		0.943094,
		0.951392,
		0.953609,
		0.950009,
		0.941091,
		0.927549,
		0.91016,
		0.88959,
		0.866086,
		0.839261,
		0.80827,
		0.772148,
		0.730099,
		0.681705,
		0.627053,
		0.566781,
		0.502049,
		0.434461,
		0.365947,
		0.298613,
		0.234593,
		0.175889,
		0.124229,
		0.08095,
		0.046907,
		0.022412,
		0.007221,
		0.000551,
		0.000331,
		0.005236,
		0.015823,
		0.031834,
		0.05299,
		0.078987,
		0.109496,
		0.144161,
		0.182603,
		0.224414,
		0.269161,
		0.316381,
		0.365589,
		0.416269,
		0.467878,
		0.519843,
		0.57156,
		0.622395,
		0.671681,
		0.718723,
		0.762802,
		0.803188,
		0.839153,
		0.870001,
		0.895097,
		0.913908,
		0.92605,
		0.931338,
		0.929828,
		0.921848,
		0.907999,
		0.889104,
		0.866086,
		0.839755,
		0.810497,
		0.778059,
		0.741761,
		0.700845,
		0.654769,
		0.603399,
		0.547106,
		0.486772,
		0.423733,
		0.359665,
		0.296449,
		0.236018,
		0.180203,
		0.130598,
		0.088442,
		0.054533,
		0.029175,
		0.012164,
		0.002811,
		0.001374,
		0.008478,
		0.021402,
		0.039864,
		0.063557,
		0.092147,
		0.125272,
		0.162543,
		0.20354,
		0.247816,
		0.294891,
		0.344254,
		0.395357,
		0.447615,
		0.500403,
		0.55305,
		0.604841,
		0.655014,
		0.702768,
		0.747271,
		0.787675,
		0.82314,
		0.852871,
		0.876156,
		0.892416,
		0.901261,
		0.902538,
		0.896384,
		0.883247,
		0.863875,
		0.839261,
		0.810497,
		0.778554,
		0.743965,
		0.706621,
		0.666012,
		0.621588,
		0.573048,
		0.520504,
		0.464545,
		0.406208,
		0.346893,
		0.288229,
		0.23193,
		0.179646,
		0.132827,
		0.092608,
		0.05973,
		0.034483,
		0.016701,
		0.005777,
		0.000729,
		0.003208,
		0.012657,
		0.028064,
		0.049119,
		0.075483,
		0.106789,
		0.142636,
		0.182595,
		0.226198,
		0.272942,
		0.322282,
		0.373628,
		0.426337,
		0.479711,
		0.532994,
		0.585368,
		0.635962,
		0.683852,
		0.728083,
		0.767694,
		0.801745,
		0.829368,
		0.84981,
		0.862495,
		0.86708,
		0.863508,
		0.852052,
		0.833327,
		0.80827,
		0.778059,
		0.743965,
		0.707107,
		0.668148,
		0.627111,
		0.58364,
		0.537373,
		0.48821,
		0.436447,
		0.38281,
		0.328396,
		0.274558,
		0.222774,
		0.174498,
		0.131026,
		0.093385,
		0.062253,
		0.037903,
		0.020197,
		0.008604,
		0.002261,
		5.4e-005,
		0.000375,
		0.00592,
		0.01787,
		0.035914,
		0.059709,
		0.088879,
		0.123016,
		0.16167,
		0.204353,
		0.250527,
		0.299601,
		0.350925,
		0.40378,
		0.457378,
		0.510855,
		0.563276,
		0.613644,
		0.660911,
		0.704007,
		0.741871,
		0.773497,
		0.797981,
		0.814587,
		0.822798,
		0.822383,
		0.813438,
		0.79642,
		0.772148,
		0.741761,
		0.706621,
		0.668148,
		0.627578,
		0.585664,
		0.542528,
		0.497936,
		0.451676,
		0.403792,
		0.354699,
		0.305174,
		0.256277,
		0.209231,
		0.165276,
		0.125548,
		0.090958,
		0.062113,
		0.039266,
		0.022299,
		0.010748,
		0.003854,
		0.000644,
		0.001559,
		0.009606,
		0.024222,
		0.045061,
		0.071739,
		0.103831,
		0.140868,
		0.182325,
		0.227619,
		0.276097,
		0.327031,
		0.379606,
		0.432925,
		0.486002,
		0.537774,
		0.587116,
		0.632867,
		0.67386,
		0.708972,
		0.737171,
		0.757581,
		0.769537,
		0.77265,
		0.766854,
		0.752446,
		0.730099,
		0.700845,
		0.666012,
		0.627111,
		0.585664,
		0.542966,
		0.499807,
		0.45636,
		0.412477,
		0.368043,
		0.323198,
		0.278401,
		0.234406,
		0.192162,
		0.152688,
		0.116944,
		0.085717,
		0.05954,
		0.038636,
		0.022903,
		0.01193,
		0.005051,
		0.001417,
		9.2e-005,
		0.003646,
		0.01437,
		0.031819,
		0.055596,
		0.085259,
		0.120303,
		0.160159,
		0.204182,
		0.25164,
		0.301705,
		0.353452,
		0.405858,
		0.457808,
		0.508113,
		0.555533,
		0.598814,
		0.636733,
		0.66815,
		0.692067,
		0.707694,
		0.714501,
		0.712278,
		0.701172,
		0.681705,
		0.654769,
		0.621588,
		0.58364,
		0.542528,
		0.499807,
		0.456761,
		0.414158,
		0.372177,
		0.330713,
		0.289712,
		0.249349,
		0.210073,
		0.172554,
		0.137572,
		0.105905,
		0.078208,
		0.054932,
		0.036262,
		0.022103,
		0.012091,
		0.005638,
		0.002005,
		0.000384,
		0.000428,
		0.00674,
		0.020311,
		0.040727,
		0.067515,
		0.100129,
		0.137938,
		0.180217,
		0.226135,
		0.274752,
		0.325018,
		0.375779,
		0.425793,
		0.473757,
		0.518344,
		0.558244,
		0.592226,
		0.61919,
		0.638232,
		0.648705,
		0.650267,
		0.642924,
		0.627053,
		0.603399,
		0.573048,
		0.537373,
		0.497936,
		0.45636,
		0.414158,
		0.372533,
		0.332178,
		0.293242,
		0.255628,
		0.219306,
		0.184447,
		0.151444,
		0.120838,
		0.09321,
		0.069077,
		0.048799,
		0.03252,
		0.020142,
		0.011339,
		0.005589,
		0.002248,
		0.000623,
		6e-005,
		0.001778,
		0.01093,
		0.027471,
		0.050887,
		0.080573,
		0.115822,
		0.155812,
		0.199599,
		0.246117,
		0.294186,
		0.342523,
		0.389776,
		0.434554,
		0.475478,
		0.511233,
		0.540631,
		0.56267,
		0.576593,
		0.581942,
		0.578592,
		0.566781,
		0.547106,
		0.520504,
		0.48821,
		0.451676,
		0.412477,
		0.372177,
		0.332178,
		0.293547,
		0.256861,
		0.222209,
		0.189486,
		0.158656,
		0.12986,
		0.103399,
		0.079658,
		0.059008,
		0.041711,
		0.027863,
		0.017358,
		0.009893,
		0.005009,
		0.002141,
		0.000698,
		0.000131,
		0.004144,
		0.016249,
		0.035748,
		0.061965,
		0.094096,
		0.131208,
		0.172233,
		0.215973,
		0.26112,
		0.30628,
		0.35001,
		0.390871,
		0.427476,
		0.458554,
		0.483015,
		0.5,
		0.50894,
		0.509589,
		0.502049,
		0.486772,
		0.464545,
		0.436447,
		0.403792,
		0.368043,
		0.330713,
		0.293242,
		0.256861,
		0.222461,
		0.190482,
		0.160942,
		0.133715,
		0.108744,
		0.086113,
		0.066016,
		0.048672,
		0.034243,
		0.022767,
		0.014124,
		0.008033,
		0.00408,
		0.001775,
		0.000614,
		0.000144,
		1.4e-005,
		0.000483,
		0.007563,
		0.022579,
		0.044763,
		0.073204,
		0.106843,
		0.14448,
		0.184789,
		0.226346,
		0.267665,
		0.30725,
		0.34365,
		0.375519,
		0.401677,
		0.42117,
		0.433319,
		0.43776,
		0.434461,
		0.423733,
		0.406208,
		0.38281,
		0.354699,
		0.323198,
		0.289712,
		0.255628,
		0.222209,
		0.190482,
		0.161143,
		0.134483,
		0.110452,
		0.088896,
		0.069727,
		0.052962,
		0.038672,
		0.026918,
		0.017677,
		0.010807,
		0.006041,
		0.003008,
		0.001282,
		0.000438,
		0.000106,
		1.3e-005,
		0.001968,
		0.011938,
		0.029542,
		0.053745,
		0.083354,
		0.117035,
		0.153342,
		0.190753,
		0.227721,
		0.262731,
		0.294361,
		0.321345,
		0.342631,
		0.357431,
		0.365257,
		0.365947,
		0.359665,
		0.346893,
		0.328396,
		0.305174,
		0.278401,
		0.249349,
		0.219306,
		0.189486,
		0.160942,
		0.134483,
		0.110603,
		0.089455,
		0.070925,
		0.054829,
		0.041042,
		0.029504,
		0.020174,
		0.012966,
		0.007717,
		0.004166,
		0.001981,
		0.000793,
		0.000248,
		5.2e-005,
		0.00441,
		0.016943,
		0.036428,
		0.061533,
		0.090795,
		0.122659,
		0.155529,
		0.187825,
		0.218049,
		0.244843,
		0.267053,
		0.283774,
		0.294395,
		0.298613,
		0.296449,
		0.288229,
		0.274558,
		0.256277,
		0.234406,
		0.210073,
		0.184447,
		0.158656,
		0.133715,
		0.110452,
		0.089455,
		0.071031,
		0.055207,
		0.041813,
		0.030641,
		0.021524,
		0.014328,
		0.00891,
		0.005082,
		0.002591,
		0.001136,
		0.000402,
		0.000103,
		1.5e-005,
		0.000496,
		0.007571,
		0.021973,
		0.042229,
		0.066752,
		0.093895,
		0.122012,
		0.14952,
		0.174964,
		0.197077,
		0.214834,
		0.227488,
		0.234593,
		0.236018,
		0.23193,
		0.222774,
		0.209231,
		0.192162,
		0.172554,
		0.151444,
		0.12986,
		0.108744,
		0.088896,
		0.070925,
		0.055207,
		0.041882,
		0.030874,
		0.021968,
		0.014932,
		0.009563,
		0.005667,
		0.003031,
		0.001412,
		0.00054,
		0.000153,
		2.6e-005,
		0.001872,
		0.01098,
		0.026203,
		0.045838,
		0.068164,
		0.091506,
		0.1143,
		0.135159,
		0.152922,
		0.166698,
		0.175889,
		0.180203,
		0.179646,
		0.174498,
		0.165276,
		0.152688,
		0.137572,
		0.120838,
		0.103399,
		0.086113,
		0.069727,
		0.054829,
		0.041813,
		0.030874,
		0.022008,
		0.015057,
		0.00978,
		0.005929,
		0.003275,
		0.00159,
		0.000642,
		0.000194,
		3.5e-005,
		0.003794,
		0.013986,
		0.028769,
		0.046363,
		0.065084,
		0.083406,
		0.100022,
		0.113883,
		0.124229,
		0.130598,
		0.132827,
		0.131026,
		0.125548,
		0.116944,
		0.105905,
		0.09321,
		0.079658,
		0.066016,
		0.052962,
		0.041042,
		0.030641,
		0.021968,
		0.015057,
		0.0098,
		0.005984,
		0.003356,
		0.001671,
		0.000698,
		0.000222,
		4.3e-005,
		0.000395,
		0.005759,
		0.015912,
		0.029031,
		0.043437,
		0.057659,
		0.070476,
		0.08095,
		0.088442,
		0.092608,
		0.093385,
		0.090958,
		0.085717,
		0.078208,
		0.069077,
		0.059008,
		0.048672,
		0.038672,
		0.029504,
		0.021524,
		0.014932,
		0.00978,
		0.005984,
		0.003364,
		0.001688,
		0.000718,
		0.000234,
		4.7e-005,
		0.001299,
		0.007219,
		0.01627,
		0.026795,
		0.037384,
		0.046907,
		0.054533,
		0.05973,
		0.062253,
		0.062113,
		0.05954,
		0.054932,
		0.048799,
		0.041711,
		0.034243,
		0.026918,
		0.020174,
		0.014328,
		0.009563,
		0.005929,
		0.003356,
		0.001688,
		0.000719,
		0.000237,
		4.8e-005,
		0.002241,
		0.007757,
		0.014927,
		0.022412,
		0.029175,
		0.034483,
		0.037903,
		0.039266,
		0.038636,
		0.036262,
		0.03252,
		0.027863,
		0.022767,
		0.017677,
		0.012966,
		0.00891,
		0.005667,
		0.003275,
		0.001671,
		0.000718,
		0.000237,
		4.9e-005,
		0.000207,
		0.002816,
		0.007221,
		0.012164,
		0.016701,
		0.020197,
		0.022299,
		0.022903,
		0.022103,
		0.020142,
		0.017358,
		0.014124,
		0.010807,
		0.007717,
		0.005082,
		0.003031,
		0.00159,
		0.000698,
		0.000234,
		4.8e-005,
		0.000551,
		0.002811,
		0.005777,
		0.008604,
		0.010748,
		0.01193,
		0.012091,
		0.011339,
		0.009893,
		0.008033,
		0.006041,
		0.004166,
		0.002591,
		0.001412,
		0.000642,
		0.000222,
		4.7e-005,
		0.000729,
		0.002261,
		0.003854,
		0.005051,
		0.005638,
		0.005589,
		0.005009,
		0.00408,
		0.003008,
		0.001981,
		0.001136,
		0.00054,
		0.000194,
		4.3e-005,
		5.4e-005,
		0.000644,
		0.001417,
		0.002005,
		0.002248,
		0.002141,
		0.001775,
		0.001282,
		0.000793,
		0.000402,
		0.000153,
		3.5e-005,
		9.2e-005,
		0.000384,
		0.000623,
		0.000698,
		0.000614,
		0.000438,
		0.000248,
		0.000103,
		2.6e-005,
		6e-005,
		0.000131,
		0.000144,
		0.000106,
		5.2e-005,
		1.5e-005,
		1.4e-005,
		1.3e-005,
		1.3e-005,
		1.4e-005,
		1.5e-005,
		5.2e-005,
		0.000106,
		0.000144,
		0.000131,
		6e-005,
		2.6e-005,
		0.000103,
		0.000248,
		0.000438,
		0.000614,
		0.000698,
		0.000623,
		0.000384,
		9.2e-005,
		3.5e-005,
		0.000153,
		0.000402,
		0.000793,
		0.001282,
		0.001775,
		0.002141,
		0.002248,
		0.002005,
		0.001417,
		0.000644,
		5.4e-005,
		4.3e-005,
		0.000194,
		0.00054,
		0.001136,
		0.001981,
		0.003008,
		0.00408,
		0.005009,
		0.005589,
		0.005638,
		0.005051,
		0.003854,
		0.002261,
		0.000729,
		4.7e-005,
		0.000222,
		0.000642,
		0.001412,
		0.002591,
		0.004166,
		0.006041,
		0.008033,
		0.009893,
		0.011339,
		0.012091,
		0.01193,
		0.010748,
		0.008604,
		0.005777,
		0.002811,
		0.000551,
		4.8e-005,
		0.000234,
		0.000698,
		0.00159,
		0.003031,
		0.005082,
		0.007717,
		0.010807,
		0.014124,
		0.017358,
		0.020142,
		0.022103,
		0.022903,
		0.022299,
		0.020197,
		0.016701,
		0.012164,
		0.007221,
		0.002816,
		0.000207,
		4.9e-005,
		0.000237,
		0.000718,
		0.001671,
		0.003275,
		0.005667,
		0.00891,
		0.012966,
		0.017677,
		0.022767,
		0.027863,
		0.03252,
		0.036262,
		0.038636,
		0.039266,
		0.037903,
		0.034483,
		0.029175,
		0.022412,
		0.014927,
		0.007757,
		0.002241,
		4.8e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003356,
		0.005929,
		0.009563,
		0.014328,
		0.020174,
		0.026918,
		0.034243,
		0.041711,
		0.048799,
		0.054932,
		0.05954,
		0.062113,
		0.062253,
		0.05973,
		0.054533,
		0.046907,
		0.037384,
		0.026795,
		0.01627,
		0.007219,
		0.001299,
		4.7e-005,
		0.000234,
		0.000718,
		0.001688,
		0.003364,
		0.005984,
		0.00978,
		0.014932,
		0.021524,
		0.029504,
		0.038672,
		0.048672,
		0.059008,
		0.069077,
		0.078208,
		0.085717,
		0.090958,
		0.093385,
		0.092608,
		0.088442,
		0.08095,
		0.070476,
		0.057659,
		0.043437,
		0.029031,
		0.015912,
		0.005759,
		0.000395,
		4.3e-005,
		0.000222,
		0.000698,
		0.001671,
		0.003356,
		0.005984,
		0.0098,
		0.015057,
		0.021968,
		0.030641,
		0.041042,
		0.052962,
		0.066016,
		0.079658,
		0.09321,
		0.105905,
		0.116944,
		0.125548,
		0.131026,
		0.132827,
		0.130598,
		0.124229,
		0.113883,
		0.100022,
		0.083406,
		0.065084,
		0.046363,
		0.028769,
		0.013986,
		0.003794,
		3.5e-005,
		0.000194,
		0.000642,
		0.00159,
		0.003275,
		0.005929,
		0.00978,
		0.015057,
		0.022008,
		0.030874,
		0.041813,
		0.054829,
		0.069727,
		0.086113,
		0.103399,
		0.120838,
		0.137572,
		0.152688,
		0.165276,
		0.174498,
		0.179646,
		0.180203,
		0.175889,
		0.166698,
		0.152922,
		0.135159,
		0.1143,
		0.091506,
		0.068164,
		0.045838,
		0.026203,
		0.01098,
		0.001872,
		2.6e-005,
		0.000153,
		0.00054,
		0.001412,
		0.003031,
		0.005667,
		0.009563,
		0.014932,
		0.021968,
		0.030874,
		0.041882,
		0.055207,
		0.070925,
		0.088896,
		0.108744,
		0.12986,
		0.151444,
		0.172554,
		0.192162,
		0.209231,
		0.222774,
		0.23193,
		0.236018,
		0.234593,
		0.227488,
		0.214834,
		0.197077,
		0.174964,
		0.14952,
		0.122012,
		0.093895,
		0.066752,
		0.042229,
		0.021973,
		0.007571,
		0.000496,
		1.5e-005,
		0.000103,
		0.000402,
		0.001136,
		0.002591,
		0.005082,
		0.00891,
		0.014328,
		0.021524,
		0.030641,
		0.041813,
		0.055207,
		0.071031,
		0.089455,
		0.110452,
		0.133715,
		0.158656,
		0.184447,
		0.210073,
		0.234406,
		0.256277,
		0.274558,
		0.288229,
		0.296449,
		0.298613,
		0.294395,
		0.283774,
		0.267053,
		0.244843,
		0.218049,
		0.187825,
		0.155529,
		0.122659,
		0.090795,
		0.061533,
		0.036428,
		0.016943,
		0.00441,
		5.2e-005,
		0.000248,
		0.000793,
		0.001981,
		0.004166,
		0.007717,
		0.012966,
		0.020174,
		0.029504,
		0.041042,
		0.054829,
		0.070925,
		0.089455,
		0.110603,
		0.134483,
		0.160942,
		0.189486,
		0.219306,
		0.249349,
		0.278401,
		0.305174,
		0.328396,
		0.346893,
		0.359665,
		0.365947,
		0.365257,
		0.357431,
		0.342631,
		0.321345,
		0.294361,
		0.262731,
		0.227721,
		0.190753,
		0.153342,
		0.117035,
		0.083354,
		0.053745,
		0.029542,
		0.011938,
		0.001968,
		1.3e-005,
		0.000106,
		0.000438,
		0.001282,
		0.003008,
		0.006041,
		0.010807,
		0.017677,
		0.026918,
		0.038672,
		0.052962,
		0.069727,
		0.088896,
		0.110452,
		0.134483,
		0.161143,
		0.190482,
		0.222209,
		0.255628,
		0.289712,
		0.323198,
		0.354699,
		0.38281,
		0.406208,
		0.423733,
		0.434461,
		0.43776,
		0.433319,
		0.42117,
		0.401677,
		0.375519,
		0.34365,
		0.30725,
		0.267665,
		0.226346,
		0.184789,
		0.14448,
		0.106843,
		0.073204,
		0.044763,
		0.022579,
		0.007563,
		0.000483,
		1.4e-005,
		0.000144,
		0.000614,
		0.001775,
		0.00408,
		0.008033,
		0.014124,
		0.022767,
		0.034243,
		0.048672,
		0.066016,
		0.086113,
		0.108744,
		0.133715,
		0.160942,
		0.190482,
		0.222461,
		0.256861,
		0.293242,
		0.330713,
		0.368043,
		0.403792,
		0.436447,
		0.464545,
		0.486772,
		0.502049,
		0.509589,
		0.50894,
		0.5,
		0.483015,
		0.458554,
		0.427476,
		0.390871,
		0.35001,
		0.30628,
		0.26112,
		0.215973,
		0.172233,
		0.131208,
		0.094096,
		0.061965,
		0.035748,
		0.016249,
		0.004144,
		0.000131,
		0.000698,
		0.002141,
		0.005009,
		0.009893,
		0.017358,
		0.027863,
		0.041711,
		0.059008,
		0.079658,
		0.103399,
		0.12986,
		0.158656,
		0.189486,
		0.222209,
		0.256861,
		0.293547,
		0.332178,
		0.372177,
		0.412477,
		0.451676,
		0.48821,
		0.520504,
		0.547106,
		0.566781,
		0.578592,
		0.581942,
		0.576593,
		0.56267,
		0.540631,
		0.511233,
		0.475478,
		0.434554,
		0.389776,
		0.342523,
		0.294186,
		0.246117,
		0.199599,
		0.155812,
		0.115822,
		0.080573,
		0.050887,
		0.027471,
		0.01093,
		0.001778,
		6e-005,
		0.000623,
		0.002248,
		0.005589,
		0.011339,
		0.020142,
		0.03252,
		0.048799,
		0.069077,
		0.09321,
		0.120838,
		0.151444,
		0.184447,
		0.219306,
		0.255628,
		0.293242,
		0.332178,
		0.372533,
		0.414158,
		0.45636,
		0.497936,
		0.537373,
		0.573048,
		0.603399,
		0.627053,
		0.642924,
		0.650267,
		0.648705,
		0.638232,
		0.61919,
		0.592226,
		0.558244,
		0.518344,
		0.473757,
		0.425793,
		0.375779,
		0.325018,
		0.274752,
		0.226135,
		0.180217,
		0.137938,
		0.100129,
		0.067515,
		0.040727,
		0.020311,
		0.00674,
		0.000428,
		0.000384,
		0.002005,
		0.005638,
		0.012091,
		0.022103,
		0.036262,
		0.054932,
		0.078208,
		0.105905,
		0.137572,
		0.172554,
		0.210073,
		0.249349,
		0.289712,
		0.330713,
		0.372177,
		0.414158,
		0.456761,
		0.499807,
		0.542528,
		0.58364,
		0.621588,
		0.654769,
		0.681705,
		0.701172,
		0.712278,
		0.714501,
		0.707694,
		0.692067,
		0.66815,
		0.636733,
		0.598814,
		0.555533,
		0.508113,
		0.457808,
		0.405858,
		0.353452,
		0.301705,
		0.25164,
		0.204182,
		0.160159,
		0.120303,
		0.085259,
		0.055596,
		0.031819,
		0.01437,
		0.003646,
		9.2e-005,
		0.001417,
		0.005051,
		0.01193,
		0.022903,
		0.038636,
		0.05954,
		0.085717,
		0.116944,
		0.152688,
		0.192162,
		0.234406,
		0.278401,
		0.323198,
		0.368043,
		0.412477,
		0.45636,
		0.499807,
		0.542966,
		0.585664,
		0.627111,
		0.666012,
		0.700845,
		0.730099,
		0.752446,
		0.766854,
		0.77265,
		0.769537,
		0.757581,
		0.737171,
		0.708972,
		0.67386,
		0.632867,
		0.587116,
		0.537774,
		0.486002,
		0.432925,
		0.379606,
		0.327031,
		0.276097,
		0.227619,
		0.182325,
		0.140868,
		0.103831,
		0.071739,
		0.045061,
		0.024222,
		0.009606,
		0.001559,
		0.000644,
		0.003854,
		0.010748,
		0.022299,
		0.039266,
		0.062113,
		0.090958,
		0.125548,
		0.165276,
		0.209231,
		0.256277,
		0.305174,
		0.354699,
		0.403792,
		0.451676,
		0.497936,
		0.542528,
		0.585664,
		0.627578,
		0.668148,
		0.706621,
		0.741761,
		0.772148,
		0.79642,
		0.813438,
		0.822383,
		0.822798,
		0.814587,
		0.797981,
		0.773497,
		0.741871,
		0.704007,
		0.660911,
		0.613644,
		0.563276,
		0.510855,
		0.457378,
		0.40378,
		0.350925,
		0.299601,
		0.250527,
		0.204353,
		0.16167,
		0.123016,
		0.088879,
		0.059709,
		0.035914,
		0.01787,
		0.00592,
		0.000375,
		5.4e-005,
		0.002261,
		0.008604,
		0.020197,
		0.037903,
		0.062253,
		0.093385,
		0.131026,
		0.174498,
		0.222774,
		0.274558,
		0.328396,
		0.38281,
		0.436447,
		0.48821,
		0.537373,
		0.58364,
		0.627111,
		0.668148,
		0.707107,
		0.743965,
		0.778059,
		0.80827,
		0.833327,
		0.852052,
		0.863508,
		0.86708,
		0.862495,
		0.84981,
		0.829368,
		0.801745,
		0.767694,
		0.728083,
		0.683852,
		0.635962,
		0.585368,
		0.532994,
		0.479711,
		0.426337,
		0.373628,
		0.322282,
		0.272942,
		0.226198,
		0.182595,
		0.142636,
		0.106789,
		0.075483,
		0.049119,
		0.028064,
		0.012657,
		0.003208,
		0.000729,
		0.005777,
		0.016701,
		0.034483,
		0.05973,
		0.092608,
		0.132827,
		0.179646,
		0.23193,
		0.288229,
		0.346893,
		0.406208,
		0.464545,
		0.520504,
		0.573048,
		0.621588,
		0.666012,
		0.706621,
		0.743965,
		0.778554,
		0.810497,
		0.839261,
		0.863875,
		0.883247,
		0.896384,
		0.902538,
		0.901261,
		0.892416,
		0.876156,
		0.852871,
		0.82314,
		0.787675,
		0.747271,
		0.702768,
		0.655014,
		0.604841,
		0.55305,
		0.500403,
		0.447615,
		0.395357,
		0.344254,
		0.294891,
		0.247816,
		0.20354,
		0.162543,
		0.125272,
		0.092147,
		0.063557,
		0.039864,
		0.021402,
		0.008478,
		0.001374,
		0.002811,
		0.012164,
		0.029175,
		0.054533,
		0.088442,
		0.130598,
		0.180203,
		0.236018,
		0.296449,
		0.359665,
		0.423733,
		0.486772,
		0.547106,
		0.603399,
		0.654769,
		0.700845,
		0.741761,
		0.778059,
		0.810497,
		0.839755,
		0.866086,
		0.889104,
		0.907999,
		0.921848,
		0.929828,
		0.931338,
		0.92605,
		0.913908,
		0.895097,
		0.870001,
		0.839153,
		0.803188,
		0.762802,
		0.718723,
		0.671681,
		0.622395,
		0.57156,
		0.519843,
		0.467878,
		0.416269,
		0.365589,
		0.316381,
		0.269161,
		0.224414,
		0.182603,
		0.144161,
		0.109496,
		0.078987,
		0.05299,
		0.031834,
		0.015823,
		0.005236,
		0.000331,
		0.000551,
		0.007221,
		0.022412,
		0.046907,
		0.08095,
		0.124229,
		0.175889,
		0.234593,
		0.298613,
		0.365947,
		0.434461,
		0.502049,
		0.566781,
		0.627053,
		0.681705,
		0.730099,
		0.772148,
		0.80827,
		0.839261,
		0.866086,
		0.88959,
		0.91016,
		0.927549,
		0.941091,
		0.950009,
		0.953609,
		0.951392,
		0.943094,
		0.928683,
		0.908329,
		0.882368,
		0.851254,
		0.815526,
		0.775775,
		0.732617,
		0.686676,
		0.638572,
		0.588915,
		0.538297,
		0.487293,
		0.436456,
		0.38632,
		0.337397,
		0.290179,
		0.245133,
		0.202704,
		0.163312,
		0.127354,
		0.0952,
		0.067199,
		0.043674,
		0.024926,
		0.011231,
		0.002844,
		0.002816,
		0.014927,
		0.037384,
		0.070476,
		0.113883,
		0.166698,
		0.227488,
		0.294395,
		0.365257,
		0.43776,
		0.509589,
		0.578592,
		0.642924,
		0.701172,
		0.752446,
		0.79642,
		0.833327,
		0.863875,
		0.889104,
		0.91016,
		0.928019,
		0.943177,
		0.955494,
		0.964421,
		0.969283,
		0.969472,
		0.964539,
		0.954239,
		0.938519,
		0.917501,
		0.891446,
		0.860718,
		0.825756,
		0.78705,
		0.745119,
		0.700499,
		0.653732,
		0.605362,
		0.555925,
		0.505951,
		0.455958,
		0.40645,
		0.357915,
		0.310823,
		0.265625,
		0.222749,
		0.182603,
		0.145573,
		0.112019,
		0.082283,
		0.056682,
		0.035512,
		0.019046,
		0.007538,
		0.001221,
		0.000207,
		0.007757,
		0.026795,
		0.057659,
		0.100022,
		0.152922,
		0.214834,
		0.283774,
		0.357431,
		0.433319,
		0.50894,
		0.581942,
		0.650267,
		0.712278,
		0.766854,
		0.813438,
		0.852052,
		0.883247,
		0.907999,
		0.927549,
		0.943177,
		0.955945,
		0.966413,
		0.974514,
		0.97977,
		0.981571,
		0.979348,
		0.972665,
		0.961255,
		0.945023,
		0.924021,
		0.898429,
		0.868522,
		0.83465,
		0.797216,
		0.756663,
		0.713457,
		0.668082,
		0.621033,
		0.572806,
		0.523896,
		0.474793,
		0.425977,
		0.377913,
		0.331053,
		0.28583,
		0.242658,
		0.201932,
		0.164025,
		0.129291,
		0.098062,
		0.070648,
		0.047341,
		0.02841,
		0.014108,
		0.004665,
		0.000295,
		0.002241,
		0.01627,
		0.043437,
		0.083406,
		0.135159,
		0.197077,
		0.267053,
		0.342631,
		0.42117,
		0.5,
		0.576593,
		0.648705,
		0.714501,
		0.77265,
		0.822383,
		0.863508,
		0.896384,
		0.921848,
		0.941091,
		0.955494,
		0.966413,
		0.974941,
		0.981659,
		0.986527,
		0.989105,
		0.988808,
		0.985074,
		0.977452,
		0.965637,
		0.949475,
		0.92895,
		0.904167,
		0.875324,
		0.842698,
		0.806627,
		0.767493,
		0.725713,
		0.681729,
		0.635997,
		0.588982,
		0.541152,
		0.492972,
		0.444897,
		0.397375,
		0.350837,
		0.3057,
		0.262364,
		0.221211,
		0.182603,
		0.146884,
		0.114378,
		0.085389,
		0.060202,
		0.039085,
		0.022285,
		0.010032,
		0.002539,
		0.007219,
		0.029031,
		0.065084,
		0.1143,
		0.174964,
		0.244843,
		0.321345,
		0.401677,
		0.483015,
		0.56267,
		0.638232,
		0.707694,
		0.769537,
		0.822798,
		0.86708,
		0.902538,
		0.929828,
		0.950009,
		0.964421,
		0.974514,
		0.981659,
		0.986931,
		0.990887,
		0.993485,
		0.994288,
		0.992715,
		0.98819,
		0.980234,
		0.968498,
		0.952776,
		0.932992,
		0.909189,
		0.881506,
		0.850168,
		0.815461,
		0.777728,
		0.737347,
		0.694725,
		0.650291,
		0.604481,
		0.557738,
		0.510503,
		0.463212,
		0.416289,
		0.370151,
		0.325198,
		0.281814,
		0.24037,
		0.201216,
		0.164687,
		0.131098,
		0.100747,
		0.073915,
		0.050862,
		0.031834,
		0.017058,
		0.006746,
		0.001092,
		0.001299,
		0.015912,
		0.046363,
		0.091506,
		0.14952,
		0.218049,
		0.294361,
		0.375519,
		0.458554,
		0.540631,
		0.61919,
		0.692067,
		0.757581,
		0.814587,
		0.862495,
		0.901261,
		0.931338,
		0.953609,
		0.969283,
		0.97977,
		0.986527,
		0.990887,
		0.993865,
		0.995965,
		0.997118,
		0.996877,
		0.994648,
		0.989838,
		0.981941,
		0.970571,
		0.955481,
		0.936549,
		0.913775,
		0.887259,
		0.857183,
		0.823803,
		0.787425,
		0.7484,
		0.707107,
		0.663946,
		0.619329,
		0.573675,
		0.527402,
		0.480925,
		0.434652,
		0.38898,
		0.344293,
		0.300964,
		0.259348,
		0.219786,
		0.182603,
		0.148107,
		0.116587,
		0.088318,
		0.063557,
		0.042546,
		0.02551,
		0.012657,
		0.004182,
		0.000264,
		0.005759,
		0.028769,
		0.068164,
		0.122012,
		0.187825,
		0.262731,
		0.34365,
		0.427476,
		0.511233,
		0.592226,
		0.66815,
		0.737171,
		0.797981,
		0.84981,
		0.892416,
		0.92605,
		0.951392,
		0.969472,
		0.981571,
		0.989105,
		0.993485,
		0.995965,
		0.997474,
		0.998451,
		0.998789,
		0.998023,
		0.99555,
		0.990764,
		0.983139,
		0.972268,
		0.957876,
		0.939814,
		0.918052,
		0.892661,
		0.863799,
		0.831692,
		0.796619,
		0.758906,
		0.718904,
		0.676989,
		0.63355,
		0.588982,
		0.543682,
		0.498046,
		0.452462,
		0.407312,
		0.362964,
		0.319778,
		0.278096,
		0.238249,
		0.200552,
		0.165303,
		0.132788,
		0.103273,
		0.077011,
		0.05424,
		0.035182,
		0.020042,
		0.009016,
		0.00228,
		0.000395,
		0.013986,
		0.045838,
		0.093895,
		0.155529,
		0.227721,
		0.30725,
		0.390871,
		0.475478,
		0.558244,
		0.636733,
		0.708972,
		0.773497,
		0.829368,
		0.876156,
		0.913908,
		0.943094,
		0.964539,
		0.979348,
		0.988808,
		0.994288,
		0.997118,
		0.998451,
		0.999123,
		0.999501,
		0.999445,
		0.998478,
		0.995994,
		0.991385,
		0.984119,
		0.973782,
		0.960083,
		0.942861,
		0.922066,
		0.897749,
		0.870045,
		0.839157,
		0.80534,
		0.768893,
		0.730146,
		0.68945,
		0.647171,
		0.603682,
		0.559361,
		0.514584,
		0.469722,
		0.425141,
		0.381194,
		0.338228,
		0.296573,
		0.25655,
		0.218463,
		0.182603,
		0.149248,
		0.118659,
		0.091084,
		0.066755,
		0.045892,
		0.028698,
		0.015366,
		0.006072,
		0.000982,
		0.003794,
		0.026203,
		0.066752,
		0.122659,
		0.190753,
		0.267665,
		0.35001,
		0.434554,
		0.518344,
		0.598814,
		0.67386,
		0.741871,
		0.801745,
		0.852871,
		0.895097,
		0.928683,
		0.954239,
		0.972665,
		0.985074,
		0.992715,
		0.996877,
		0.998789,
		0.999501,
		0.999758,
		0.999866,
		0.999659,
		0.998659,
		0.996268,
		0.99189,
		0.984999,
		0.975181,
		0.962145,
		0.945719,
		0.925841,
		0.902545,
		0.875946,
		0.846224,
		0.813614,
		0.778391,
		0.740862,
		0.701355,
		0.660215,
		0.617796,
		0.574455,
		0.53055,
		0.486437,
		0.442464,
		0.398972,
		0.356293,
		0.314747,
		0.274643,
		0.236278,
		0.199933,
		0.165879,
		0.134371,
		0.105651,
		0.079949,
		0.057479,
		0.038444,
		0.023032,
		0.011419,
		0.003771,
		0.000238,
		0.01098,
		0.042229,
		0.090795,
		0.153342,
		0.226346,
		0.30628,
		0.389776,
		0.473757,
		0.555533,
		0.632867,
		0.704007,
		0.767694,
		0.82314,
		0.870001,
		0.908329,
		0.938519,
		0.961255,
		0.977452,
		0.98819,
		0.994648,
		0.998023,
		0.999445,
		0.999866,
		0.999952,
		0.999963,
		0.999719,
		0.998754,
		0.996489,
		0.992346,
		0.985814,
		0.976488,
		0.964076,
		0.948403,
		0.929393,
		0.907068,
		0.881522,
		0.852917,
		0.821467,
		0.787425,
		0.751077,
		0.712729,
		0.672708,
		0.631345,
		0.588982,
		0.545959,
		0.502614,
		0.459281,
		0.416289,
		0.373956,
		0.332591,
		0.292492,
		0.253947,
		0.21723,
		0.182603,
		0.150317,
		0.120607,
		0.093699,
		0.069803,
		0.049119,
		0.031834,
		0.018122,
		0.008146,
		0.002059,
		0.001872,
		0.021973,
		0.061533,
		0.117035,
		0.184789,
		0.26112,
		0.342523,
		0.425793,
		0.508113,
		0.587116,
		0.660911,
		0.728083,
		0.787675,
		0.839153,
		0.882368,
		0.917501,
		0.945023,
		0.965637,
		0.980234,
		0.989838,
		0.99555,
		0.998478,
		0.999659,
		0.999963,
		0.999994,
		0.99998,
		0.999741,
		0.998829,
		0.99669,
		0.992769,
		0.986574,
		0.97771,
		0.965887,
		0.950924,
		0.932739,
		0.911335,
		0.886795,
		0.85926,
		0.828924,
		0.796022,
		0.760817,
		0.723599,
		0.684672,
		0.644352,
		0.602961,
		0.560823,
		0.518262,
		0.475597,
		0.433143,
		0.391206,
		0.350084,
		0.310066,
		0.271429,
		0.234441,
		0.199356,
		0.166417,
		0.135857,
		0.107895,
		0.082738,
		0.060583,
		0.041614,
		0.026003,
		0.013913,
		0.005495,
		0.000888,
		0.007571,
		0.036428,
		0.083354,
		0.14448,
		0.215973,
		0.294186,
		0.375779,
		0.457808,
		0.537774,
		0.613644,
		0.683852,
		0.747271,
		0.803188,
		0.851254,
		0.891446,
		0.924021,
		0.949475,
		0.968498,
		0.981941,
		0.990764,
		0.995994,
		0.998659,
		0.999719,
		0.99998,
		1,
		0.999983,
		0.999757,
		0.998897,
		0.996876,
		0.993163,
		0.987285,
		0.978855,
		0.967586,
		0.953295,
		0.93589,
		0.915365,
		0.891784,
		0.865273,
		0.836008,
		0.804203,
		0.770107,
		0.733987,
		0.696131,
		0.656837,
		0.61641,
		0.575159,
		0.533393,
		0.491417,
		0.449532,
		0.408035,
		0.367211,
		0.32734,
		0.28869,
		0.25152,
		0.216079,
		0.182603,
		0.151319,
		0.122442,
		0.096175,
		0.07271,
		0.05223,
		0.034906,
		0.020897,
		0.010354,
		0.003417,
		0.000216,
		0.000496,
		0.016943,
		0.053745,
		0.106843,
		0.172233,
		0.246117,
		0.325018,
		0.405858,
		0.486002,
		0.563276,
		0.635962,
		0.702768,
		0.762802,
		0.815526,
		0.860718,
		0.898429,
		0.92895,
		0.952776,
		0.970571,
		0.983139,
		0.991385,
		0.996268,
		0.998754,
		0.999741,
		0.999983,
		1,
		0.999984,
		0.999771,
		0.99896,
		0.997049,
		0.993531,
		0.987949,
		0.979927,
		0.969182,
		0.955526,
		0.938862,
		0.919171,
		0.896506,
		0.870975,
		0.842738,
		0.811993,
		0.778969,
		0.743917,
		0.707107,
		0.668821,
		0.629348,
		0.588982,
		0.548017,
		0.506746,
		0.465459,
		0.424437,
		0.383959,
		0.344293,
		0.3057,
		0.26843,
		0.232725,
		0.198815,
		0.166922,
		0.137255,
		0.110014,
		0.085389,
		0.063557,
		0.044689,
		0.028941,
		0.016464,
		0.007397,
		0.001868,
		0.00441,
		0.029542,
		0.073204,
		0.131208,
		0.199599,
		0.274752,
		0.353452,
		0.432925,
		0.510855,
		0.585368,
		0.655014,
		0.718723,
		0.775775,
		0.825756,
		0.868522,
		0.904167,
		0.932992,
		0.955481,
		0.972268,
		0.984119,
		0.99189,
		0.996489,
		0.998829,
		0.999757,
		0.999984,
		1,
		0.999985,
		0.999784,
		0.999018,
		0.99721,
		0.993875,
		0.988571,
		0.980933,
		0.970681,
		0.957627,
		0.941666,
		0.92277,
		0.900978,
		0.876386,
		0.849137,
		0.819413,
		0.787425,
		0.75341,
		0.717621,
		0.680324,
		0.641793,
		0.602308,
		0.562148,
		0.521595,
		0.480925,
		0.440412,
		0.40032,
		0.360911,
		0.322436,
		0.285138,
		0.249252,
		0.215002,
		0.182603,
		0.152262,
		0.124172,
		0.098521,
		0.075484,
		0.055226,
		0.037906,
		0.02367,
		0.012657,
		0.004996,
		0.000807,
		0.011938,
		0.044763,
		0.094096,
		0.155812,
		0.226135,
		0.301705,
		0.379606,
		0.457378,
		0.532994,
		0.604841,
		0.671681,
		0.732617,
		0.78705,
		0.83465,
		0.875324,
		0.909189,
		0.936549,
		0.957876,
		0.973782,
		0.984999,
		0.992346,
		0.99669,
		0.998897,
		0.999771,
		0.999985,
		1,
		0.999986,
		0.999797,
		0.999073,
		0.997361,
		0.994196,
		0.989154,
		0.981877,
		0.972091,
		0.959607,
		0.944313,
		0.926173,
		0.905215,
		0.881522,
		0.855222,
		0.826481,
		0.795497,
		0.762488,
		0.727694,
		0.691366,
		0.653763,
		0.615152,
		0.575799,
		0.535972,
		0.495938,
		0.455958,
		0.416289,
		0.377182,
		0.33888,
		0.301618,
		0.265625,
		0.231119,
		0.198309,
		0.167396,
		0.138573,
		0.112019,
		0.08791,
		0.066407,
		0.047667,
		0.031834,
		0.019046,
		0.009431,
		0.003111,
		0.000196,
		0.001968,
		0.022579,
		0.061965,
		0.115822,
		0.180217,
		0.25164,
		0.327031,
		0.40378,
		0.479711,
		0.55305,
		0.622395,
		0.686676,
		0.745119,
		0.797216,
		0.842698,
		0.881506,
		0.913775,
		0.939814,
		0.960083,
		0.975181,
		0.985814,
		0.992769,
		0.996876,
		0.99896,
		0.999784,
		0.999986,
		1,
		0.999987,
		0.999808,
		0.999123,
		0.997501,
		0.994497,
		0.9897,
		0.982764,
		0.973418,
		0.961473,
		0.946813,
		0.929393,
		0.909232,
		0.8864,
		0.861011,
		0.833218,
		0.803203,
		0.771171,
		0.737347,
		0.701966,
		0.665277,
		0.62753,
		0.588982,
		0.549888,
		0.510503,
		0.471079,
		0.431863,
		0.393097,
		0.355015,
		0.317848,
		0.281814,
		0.247127,
		0.213992,
		0.182603,
		0.153149,
		0.125807,
		0.100747,
		0.078131,
		0.058111,
		0.04083,
		0.026425,
		0.015024,
		0.006746,
		0.001703,
		0.007563,
		0.035748,
		0.080573,
		0.137938,
		0.204182,
		0.276097,
		0.350925,
		0.426337,
		0.500403,
		0.57156,
		0.638572,
		0.700499,
		0.756663,
		0.806627,
		0.850168,
		0.887259,
		0.918052,
		0.942861,
		0.962145,
		0.976488,
		0.986574,
		0.993163,
		0.997049,
		0.999018,
		0.999797,
		0.999987,
		1,
		0.999987,
		0.999819,
		0.999171,
		0.997633,
		0.994778,
		0.990212,
		0.983597,
		0.974668,
		0.963233,
		0.949176,
		0.932443,
		0.913042,
		0.891034,
		0.86652,
		0.83964,
		0.810562,
		0.779477,
		0.746597,
		0.712144,
		0.676351,
		0.63946,
		0.601712,
		0.563354,
		0.524629,
		0.485779,
		0.447041,
		0.40865,
		0.370832,
		0.333809,
		0.297795,
		0.262996,
		0.229612,
		0.197833,
		0.167843,
		0.139816,
		0.113919,
		0.090309,
		0.069138,
		0.050548,
		0.034672,
		0.021638,
		0.011564,
		0.004562,
		0.000737,
		0.000483,
		0.016249,
		0.050887,
		0.100129,
		0.160159,
		0.227619,
		0.299601,
		0.373628,
		0.447615,
		0.519843,
		0.588915,
		0.653732,
		0.713457,
		0.767493,
		0.815461,
		0.857183,
		0.892661,
		0.922066,
		0.945719,
		0.964076,
		0.97771,
		0.987285,
		0.993531,
		0.99721,
		0.999073,
		0.999808,
		0.999987,
		1,
		0.999988,
		0.999829,
		0.999215,
		0.997755,
		0.995042,
		0.990693,
		0.984381,
		0.975845,
		0.964896,
		0.95141,
		0.935331,
		0.916658,
		0.895439,
		0.871765,
		0.845765,
		0.817592,
		0.787425,
		0.755463,
		0.721915,
		0.687003,
		0.650955,
		0.614003,
		0.576381,
		0.538323,
		0.500061,
		0.461824,
		0.423837,
		0.38632,
		0.349487,
		0.313545,
		0.278696,
		0.245133,
		0.213043,
		0.182603,
		0.153986,
		0.127354,
		0.102863,
		0.08066,
		0.060886,
		0.043674,
		0.02915,
		0.01743,
		0.008627,
		0.002844,
		0.000179,
		0.004144,
		0.027471,
		0.067515,
		0.120303,
		0.182325,
		0.250527,
		0.322282,
		0.395357,
		0.467878,
		0.538297,
		0.605362,
		0.668082,
		0.725713,
		0.777728,
		0.823803,
		0.863799,
		0.897749,
		0.925841,
		0.948403,
		0.965887,
		0.978855,
		0.987949,
		0.993875,
		0.997361,
		0.999123,
		0.999819,
		0.999988,
		1,
		0.999989,
		0.999838,
		0.999256,
		0.99787,
		0.995289,
		0.991144,
		0.985119,
		0.976956,
		0.966466,
		0.953525,
		0.938069,
		0.92009,
		0.899627,
		0.876762,
		0.851607,
		0.824309,
		0.795032,
		0.763962,
		0.731298,
		0.69725,
		0.662033,
		0.625869,
		0.588982,
		0.551596,
		0.513933,
		0.476215,
		0.438658,
		0.401474,
		0.364871,
		0.32905,
		0.294205,
		0.260527,
		0.228195,
		0.197385,
		0.168264,
		0.140991,
		0.11572,
		0.092596,
		0.071756,
		0.053333,
		0.03745,
		0.024223,
		0.013765,
		0.006177,
		0.001559,
		0.01093,
		0.040727,
		0.085259,
		0.140868,
		0.204353,
		0.272942,
		0.344254,
		0.416269,
		0.487293,
		0.555925,
		0.621033,
		0.681729,
		0.737347,
		0.787425,
		0.831692,
		0.870045,
		0.902545,
		0.929393,
		0.950924,
		0.967586,
		0.979927,
		0.988571,
		0.994196,
		0.997501,
		0.999171,
		0.999829,
		0.999989,
		1,
		0.999989,
		0.999846,
		0.999295,
		0.997978,
		0.995521,
		0.991569,
		0.985814,
		0.978003,
		0.96795,
		0.955526,
		0.940665,
		0.92335,
		0.903612,
		0.881522,
		0.857183,
		0.830729,
		0.802314,
		0.772112,
		0.74031,
		0.707107,
		0.672708,
		0.637323,
		0.601168,
		0.564456,
		0.527402,
		0.490217,
		0.453111,
		0.416289,
		0.379951,
		0.344293,
		0.309504,
		0.275766,
		0.243258,
		0.212149,
		0.182603,
		0.154777,
		0.128819,
		0.104874,
		0.083077,
		0.063557,
		0.046438,
		0.031834,
		0.019856,
		0.010606,
		0.004182,
		0.000675,
		0.001778,
		0.020311,
		0.055596,
		0.103831,
		0.16167,
		0.226198,
		0.294891,
		0.365589,
		0.436456,
		0.505951,
		0.572806,
		0.635997,
		0.694725,
		0.7484,
		0.796619,
		0.839157,
		0.875946,
		0.907068,
		0.932739,
		0.953295,
		0.969182,
		0.980933,
		0.989154,
		0.994497,
		0.997633,
		0.999215,
		0.999838,
		0.999989,
		1,
		0.99999,
		0.999854,
		0.999331,
		0.998079,
		0.995739,
		0.991969,
		0.986469,
		0.978993,
		0.969353,
		0.957423,
		0.943128,
		0.926448,
		0.907405,
		0.886059,
		0.862506,
		0.836868,
		0.809287,
		0.779928,
		0.748966,
		0.71659,
		0.682994,
		0.64838,
		0.612952,
		0.576915,
		0.540475,
		0.503836,
		0.4672,
		0.430764,
		0.394723,
		0.359265,
		0.324575,
		0.29083,
		0.258204,
		0.226862,
		0.196963,
		0.168662,
		0.142104,
		0.117431,
		0.094776,
		0.074267,
		0.056024,
		0.040162,
		0.026791,
		0.016011,
		0.007921,
		0.00261,
		0.000165,
		0.00674,
		0.031819,
		0.071739,
		0.123016,
		0.182595,
		0.247816,
		0.316381,
		0.38632,
		0.455958,
		0.523896,
		0.588982,
		0.650291,
		0.707107,
		0.758906,
		0.80534,
		0.846224,
		0.881522,
		0.911335,
		0.93589,
		0.955526,
		0.970681,
		0.981877,
		0.9897,
		0.994778,
		0.997755,
		0.999256,
		0.999846,
		0.99999,
		1,
		0.999991,
		0.999862,
		0.999365,
		0.998174,
		0.995944,
		0.992346,
		0.987087,
		0.979927,
		0.970681,
		0.95922,
		0.945467,
		0.929393,
		0.911016,
		0.890386,
		0.86759,
		0.842738,
		0.815966,
		0.787425,
		0.757282,
		0.725714,
		0.692907,
		0.659052,
		0.624345,
		0.588982,
		0.553161,
		0.517077,
		0.480925,
		0.444897,
		0.409179,
		0.373956,
		0.339405,
		0.3057,
		0.273008,
		0.241492,
		0.211307,
		0.182603,
		0.155525,
		0.13021,
		0.106789,
		0.085389,
		0.066128,
		0.049119,
		0.034471,
		0.022285,
		0.012657,
		0.005678,
		0.001432,
		0.000428,
		0.01437,
		0.045061,
		0.088879,
		0.142636,
		0.20354,
		0.269161,
		0.337397,
		0.40645,
		0.474793,
		0.541152,
		0.604481,
		0.663946,
		0.718904,
		0.768893,
		0.813614,
		0.852917,
		0.886795,
		0.915365,
		0.938862,
		0.957627,
		0.972091,
		0.982764,
		0.990212,
		0.995042,
		0.99787,
		0.999295,
		0.999854,
		0.999991,
		1,
		0.999991,
		0.999869,
		0.999397,
		0.998263,
		0.996137,
		0.9927,
		0.98767,
		0.980811,
		0.971939,
		0.960924,
		0.947687,
		0.932195,
		0.914455,
		0.894513,
		0.872446,
		0.848355,
		0.822365,
		0.794618,
		0.765272,
		0.734494,
		0.702461,
		0.669354,
		0.63536,
		0.600669,
		0.565468,
		0.529947,
		0.494292,
		0.45869,
		0.42332,
		0.38836,
		0.353984,
		0.32036,
		0.287651,
		0.256014,
		0.225603,
		0.196564,
		0.169038,
		0.14316,
		0.119058,
		0.096858,
		0.076675,
		0.058623,
		0.042808,
		0.02933,
		0.018285,
		0.009762,
		0.003848,
		0.000621,
		0.003646,
		0.024222,
		0.059709,
		0.106789,
		0.162543,
		0.224414,
		0.290179,
		0.357915,
		0.425977,
		0.492972,
		0.557738,
		0.619329,
		0.676989,
		0.730146,
		0.778391,
		0.821467,
		0.85926,
		0.891784,
		0.919171,
		0.941666,
		0.959607,
		0.973418,
		0.983597,
		0.990693,
		0.995289,
		0.997978,
		0.999331,
		0.999862,
		0.999991,
		1,
		0.999991,
		0.999876,
		0.999426,
		0.998347,
		0.996318,
		0.993035,
		0.988221,
		0.981647,
		0.97313,
		0.962541,
		0.949797,
		0.93486,
		0.917733,
		0.898452,
		0.877086,
		0.853729,
		0.828496,
		0.801521,
		0.77295,
		0.742944,
		0.711668,
		0.679298,
		0.64601,
		0.611985,
		0.577405,
		0.542452,
		0.507305,
		0.472143,
		0.437142,
		0.402472,
		0.368303,
		0.334797,
		0.302114,
		0.270407,
		0.239825,
		0.210511,
		0.182603,
		0.156234,
		0.131531,
		0.108615,
		0.087601,
		0.068601,
		0.05172,
		0.037057,
		0.024706,
		0.014759,
		0.007298,
		0.002404,
		0.000152,
		0.009606,
		0.035914,
		0.075483,
		0.125272,
		0.182603,
		0.245133,
		0.310823,
		0.377913,
		0.444897,
		0.510503,
		0.573675,
		0.63355,
		0.68945,
		0.740862,
		0.787425,
		0.828924,
		0.865273,
		0.896506,
		0.92277,
		0.944313,
		0.961473,
		0.974668,
		0.984381,
		0.991144,
		0.995521,
		0.998079,
		0.999365,
		0.999869,
		0.999991,
		1,
		0.999992,
		0.999882,
		0.999454,
		0.998425,
		0.99649,
		0.99335,
		0.988742,
		0.982438,
		0.97426,
		0.964076,
		0.951803,
		0.937398,
		0.920857,
		0.902212,
		0.881522,
		0.858873,
		0.834373,
		0.808146,
		0.78033,
		0.751077,
		0.720543,
		0.688896,
		0.656305,
		0.622942,
		0.588982,
		0.5546,
		0.519969,
		0.485261,
		0.450646,
		0.416289,
		0.382355,
		0.349001,
		0.316383,
		0.28465,
		0.253947,
		0.224414,
		0.196187,
		0.169394,
		0.144161,
		0.120607,
		0.098846,
		0.078987,
		0.061133,
		0.045385,
		0.031834,
		0.02057,
		0.011678,
		0.005236,
		0.00132,
		0.001559,
		0.01787,
		0.049119,
		0.092147,
		0.144161,
		0.202704,
		0.265625,
		0.331053,
		0.397375,
		0.463212,
		0.527402,
		0.588982,
		0.647171,
		0.701355,
		0.751077,
		0.796022,
		0.836008,
		0.870975,
		0.900978,
		0.926173,
		0.946813,
		0.963233,
		0.975845,
		0.985119,
		0.991569,
		0.995739,
		0.998174,
		0.999397,
		0.999876,
		0.999992,
		1,
		0.999992,
		0.999888,
		0.999481,
		0.998499,
		0.996651,
		0.993648,
		0.989234,
		0.983187,
		0.975331,
		0.965534,
		0.953711,
		0.939815,
		0.923837,
		0.905802,
		0.885764,
		0.863799,
		0.840008,
		0.814507,
		0.787425,
		0.758906,
		0.729099,
		0.698162,
		0.666257,
		0.63355,
		0.600208,
		0.566399,
		0.53229,
		0.498046,
		0.463832,
		0.429809,
		0.396135,
		0.362964,
		0.330446,
		0.298728,
		0.26795,
		0.238249,
		0.209758,
		0.182603,
		0.156908,
		0.132788,
		0.110356,
		0.08972,
		0.070982,
		0.05424,
		0.039587,
		0.02711,
		0.016893,
		0.009016,
		0.003552,
		0.000573,
		0.00592,
		0.028064,
		0.063557,
		0.109496,
		0.163312,
		0.222749,
		0.28583,
		0.350837,
		0.416289,
		0.480925,
		0.543682,
		0.603682,
		0.660215,
		0.712729,
		0.760817,
		0.804203,
		0.842738,
		0.876386,
		0.905215,
		0.929393,
		0.949176,
		0.964896,
		0.976956,
		0.985814,
		0.991969,
		0.995944,
		0.998263,
		0.999426,
		0.999882,
		0.999992,
		1,
		0.999993,
		0.999893,
		0.999505,
		0.998569,
		0.996803,
		0.99393,
		0.9897,
		0.983896,
		0.976347,
		0.96692,
		0.955526,
		0.942118,
		0.92668,
		0.909232,
		0.889821,
		0.868517,
		0.845412,
		0.820615,
		0.794247,
		0.766443,
		0.737347,
		0.707107,
		0.675879,
		0.643821,
		0.611093,
		0.577857,
		0.544274,
		0.510503,
		0.476704,
		0.443032,
		0.409641,
		0.376679,
		0.344293,
		0.312626,
		0.281814,
		0.251992,
		0.223289,
		0.19583,
		0.169733,
		0.145114,
		0.122084,
		0.100747,
		0.081207,
		0.063557,
		0.047892,
		0.034297,
		0.022856,
		0.013648,
		0.006746,
		0.002221,
		0.00014,
		0.000375,
		0.012657,
		0.039864,
		0.078987,
		0.127354,
		0.182603,
		0.242658,
		0.3057,
		0.370151,
		0.434652,
		0.498046,
		0.559361,
		0.617796,
		0.672708,
		0.723599,
		0.770107,
		0.811993,
		0.849137,
		0.881522,
		0.909232,
		0.932443,
		0.95141,
		0.966466,
		0.978003,
		0.986469,
		0.992346,
		0.996137,
		0.998347,
		0.999454,
		0.999888,
		0.999993,
		1,
		0.999993,
		0.999898,
		0.999529,
		0.998635,
		0.996946,
		0.994196,
		0.990141,
		0.984569,
		0.977312,
		0.968237,
		0.957254,
		0.944313,
		0.929393,
		0.91251,
		0.893703,
		0.873037,
		0.850596,
		0.826481,
		0.800807,
		0.773702,
		0.745299,
		0.715743,
		0.68518,
		0.653763,
		0.621646,
		0.588982,
		0.555928,
		0.522638,
		0.489264,
		0.455958,
		0.422868,
		0.39014,
		0.357915,
		0.326332,
		0.295525,
		0.265625,
		0.236757,
		0.209044,
		0.182603,
		0.157547,
		0.133985,
		0.112019,
		0.091751,
		0.073275,
		0.056682,
		0.042059,
		0.029488,
		0.019046,
		0.010808,
		0.004845,
		0.001221,
		0.003208,
		0.021402,
		0.05299,
		0.0952,
		0.145573,
		0.201932,
		0.262364,
		0.325198,
		0.38898,
		0.452462,
		0.514584,
		0.574455,
		0.631345,
		0.684672,
		0.733987,
		0.778969,
		0.819413,
		0.855222,
		0.8864,
		0.913042,
		0.935331,
		0.953525,
		0.96795,
		0.978993,
		0.987087,
		0.9927,
		0.996318,
		0.998425,
		0.999481,
		0.999893,
		0.999993,
		1,
		0.999993,
		0.999903,
		0.999551,
		0.998697,
		0.997082,
		0.994448,
		0.990558,
		0.985208,
		0.978228,
		0.969489,
		0.9589,
		0.946406,
		0.931984,
		0.915644,
		0.89742,
		0.877369,
		0.85557,
		0.832117,
		0.807118,
		0.780692,
		0.752968,
		0.724081,
		0.694173,
		0.663389,
		0.631876,
		0.599783,
		0.567259,
		0.534454,
		0.501515,
		0.468589,
		0.435818,
		0.403344,
		0.371306,
		0.339837,
		0.30907,
		0.27913,
		0.250141,
		0.222223,
		0.19549,
		0.170054,
		0.14602,
		0.123492,
		0.102567,
		0.083339,
		0.065898,
		0.05033,
		0.036715,
		0.025132,
		0.015654,
		0.008351,
		0.003289,
		0.00053,
		0.008478,
		0.031834,
		0.067199,
		0.112019,
		0.164025,
		0.221211,
		0.281814,
		0.344293,
		0.407312,
		0.469722,
		0.53055,
		0.588982,
		0.644352,
		0.696131,
		0.743917,
		0.787425,
		0.826481,
		0.861011,
		0.891034,
		0.916658,
		0.938069,
		0.955526,
		0.969353,
		0.979927,
		0.98767,
		0.993035,
		0.99649,
		0.998499,
		0.999505,
		0.999898,
		0.999993,
		1,
		0.999994,
		0.999908,
		0.999571,
		0.998756,
		0.99721,
		0.994686,
		0.990954,
		0.985814,
		0.979099,
		0.970681,
		0.960468,
		0.948403,
		0.934459,
		0.918641,
		0.900978,
		0.881522,
		0.860344,
		0.837533,
		0.813189,
		0.787425,
		0.760363,
		0.732133,
		0.702868,
		0.672708,
		0.641793,
		0.610268,
		0.578275,
		0.545959,
		0.513462,
		0.480925,
		0.448489,
		0.416289,
		0.38446,
		0.353134,
		0.322436,
		0.292492,
		0.263423,
		0.235343,
		0.208367,
		0.182603,
		0.158156,
		0.135126,
		0.113609,
		0.093699,
		0.075484,
		0.059048,
		0.044472,
		0.031834,
		0.021206,
		0.012657,
		0.006254,
		0.002059,
		0.00013,
		0.001374,
		0.015823,
		0.043674,
		0.082283,
		0.129291,
		0.182603,
		0.24037,
		0.300964,
		0.362964,
		0.425141,
		0.486437,
		0.545959,
		0.602961,
		0.656837,
		0.707107,
		0.75341,
		0.795497,
		0.833218,
		0.86652,
		0.895439,
		0.92009,
		0.940665,
		0.957423,
		0.970681,
		0.980811,
		0.988221,
		0.99335,
		0.996651,
		0.998569,
		0.999529,
		0.999903,
		0.999994,
		1,
		0.999994,
		0.999912,
		0.999591,
		0.998811,
		0.997332,
		0.994912,
		0.99133,
		0.986389,
		0.979927,
		0.971816,
		0.961963,
		0.950308,
		0.936823,
		0.921508,
		0.904386,
		0.885504,
		0.864928,
		0.842738,
		0.819031,
		0.793912,
		0.767497,
		0.739908,
		0.711275,
		0.681729,
		0.651406,
		0.620445,
		0.588982,
		0.557157,
		0.525108,
		0.492972,
		0.460882,
		0.428973,
		0.397375,
		0.366214,
		0.335615,
		0.3057,
		0.276585,
		0.248386,
		0.221211,
		0.195168,
		0.17036,
		0.146884,
		0.124837,
		0.10431,
		0.085389,
		0.068159,
		0.052699,
		0.039085,
		0.027391,
		0.017685,
		0.010032,
		0.004495,
		0.001133,
		0.005236,
		0.024926,
		0.056682,
		0.098062,
		0.146884,
		0.201216,
		0.259348,
		0.319778,
		0.381194,
		0.442464,
		0.502614,
		0.560823,
		0.61641,
		0.668821,
		0.717621,
		0.762488,
		0.803203,
		0.83964,
		0.871765,
		0.899627,
		0.92335,
		0.943128,
		0.95922,
		0.971939,
		0.981647,
		0.988742,
		0.993648,
		0.996803,
		0.998635,
		0.999551,
		0.999908,
		0.999994,
		1,
		0.999994,
		0.999916,
		0.999609,
		0.998863,
		0.997446,
		0.995126,
		0.991686,
		0.986936,
		0.980715,
		0.972897,
		0.963389,
		0.952128,
		0.939084,
		0.924251,
		0.907651,
		0.889324,
		0.869329,
		0.847743,
		0.824654,
		0.800163,
		0.774379,
		0.747418,
		0.719404,
		0.690463,
		0.660725,
		0.630323,
		0.599388,
		0.568056,
		0.53646,
		0.504731,
		0.473,
		0.441396,
		0.410046,
		0.379073,
		0.348599,
		0.318742,
		0.289616,
		0.261333,
		0.234001,
		0.207724,
		0.182603,
		0.158736,
		0.136215,
		0.11513,
		0.095568,
		0.077611,
		0.061339,
		0.046826,
		0.034144,
		0.023363,
		0.014547,
		0.007758,
		0.003055,
		0.000492,
		0.000331,
		0.011231,
		0.035512,
		0.070648,
		0.114378,
		0.164687,
		0.219786,
		0.278096,
		0.338228,
		0.398972,
		0.459281,
		0.518262,
		0.575159,
		0.629348,
		0.680324,
		0.727694,
		0.771171,
		0.810562,
		0.845765,
		0.876762,
		0.903612,
		0.926448,
		0.945467,
		0.960924,
		0.97313,
		0.982438,
		0.989234,
		0.99393,
		0.996946,
		0.998697,
		0.999571,
		0.999912,
		0.999994,
		1,
		0.999995,
		0.99992,
		0.999626,
		0.998913,
		0.997555,
		0.995329,
		0.992024,
		0.987455,
		0.981465,
		0.973927,
		0.964748,
		0.953866,
		0.941245,
		0.926878,
		0.91078,
		0.892988,
		0.873557,
		0.852555,
		0.830067,
		0.806187,
		0.781018,
		0.754672,
		0.727265,
		0.698919,
		0.669758,
		0.63991,
		0.609502,
		0.578663,
		0.547521,
		0.516206,
		0.484843,
		0.453557,
		0.422472,
		0.391707,
		0.361381,
		0.331609,
		0.302502,
		0.27417,
		0.246719,
		0.22025,
		0.194862,
		0.170651,
		0.147708,
		0.126123,
		0.10598,
		0.08736,
		0.070342,
		0.055,
		0.041405,
		0.029626,
		0.019728,
		0.011771,
		0.005814,
		0.001913,
		0.00012,
		0.002844,
		0.019046,
		0.047341,
		0.085389,
		0.131098,
		0.182603,
		0.238249,
		0.296573,
		0.356293,
		0.416289,
		0.475597,
		0.533393,
		0.588982,
		0.641793,
		0.691366,
		0.737347,
		0.779477,
		0.817592,
		0.851607,
		0.881522,
		0.907405,
		0.929393,
		0.947687,
		0.962541,
		0.97426,
		0.983187,
		0.9897,
		0.994196,
		0.997082,
		0.998756,
		0.999591,
		0.999916,
		0.999995,
		1,
		0.999995,
		0.999923,
		0.999643,
		0.99896,
		0.997658,
		0.995521,
		0.992346,
		0.987949,
		0.982179,
		0.974909,
		0.966046,
		0.955526,
		0.943313,
		0.929393,
		0.913781,
		0.896506,
		0.877618,
		0.857183,
		0.835279,
		0.811993,
		0.787425,
		0.76168,
		0.734868,
		0.707107,
		0.678515,
		0.649214,
		0.619329,
		0.588982,
		0.558299,
		0.527402,
		0.496415,
		0.465459,
		0.434652,
		0.404113,
		0.373956,
		0.344293,
		0.315234,
		0.286884,
		0.259348,
		0.232725,
		0.207112,
		0.182603,
		0.159289,
		0.137255,
		0.116587,
		0.097363,
		0.079662,
		0.063557,
		0.049119,
		0.036415,
		0.02551,
		0.016464,
		0.009337,
		0.004182,
		0.001053,
		0.007538,
		0.02841,
		0.060202,
		0.100747,
		0.148107,
		0.200552,
		0.25655,
		0.314747,
		0.373956,
		0.433143,
		0.491417,
		0.548017,
		0.602308,
		0.653763,
		0.701966,
		0.746597,
		0.787425,
		0.824309,
		0.857183,
		0.886059,
		0.911016,
		0.932195,
		0.949797,
		0.964076,
		0.975331,
		0.983896,
		0.990141,
		0.994448,
		0.99721,
		0.998811,
		0.999609,
		0.99992,
		0.999995,
		1,
		0.999995,
		0.999927,
		0.999658,
		0.999004,
		0.997755,
		0.995704,
		0.992651,
		0.98842,
		0.982859,
		0.975845,
		0.967285,
		0.957114,
		0.945291,
		0.931803,
		0.916658,
		0.899882,
		0.881522,
		0.861636,
		0.840298,
		0.817592,
		0.793609,
		0.76845,
		0.742222,
		0.715035,
		0.687003,
		0.658245,
		0.628878,
		0.599022,
		0.568797,
		0.538323,
		0.507719,
		0.477101,
		0.446587,
		0.416289,
		0.38632,
		0.356789,
		0.327802,
		0.299464,
		0.271875,
		0.245133,
		0.219335,
		0.19457,
		0.170928,
		0.148496,
		0.127354,
		0.107582,
		0.089257,
		0.072452,
		0.057235,
		0.043674,
		0.031834,
		0.021774,
		0.013553,
		0.007225,
		0.002844,
		0.000458,
		0.001221,
		0.014108,
		0.039085,
		0.073915,
		0.116587,
		0.165303,
		0.218463,
		0.274643,
		0.332591,
		0.391206,
		0.449532,
		0.506746,
		0.562148,
		0.615152,
		0.665277,
		0.712144,
		0.755463,
		0.795032,
		0.830729,
		0.862506,
		0.890386,
		0.914455,
		0.93486,
		0.951803,
		0.965534,
		0.976347,
		0.984569,
		0.990558,
		0.994686,
		0.997332,
		0.998863,
		0.999626,
		0.999923,
		0.999995,
		1,
		0.999995,
		0.99993,
		0.999673,
		0.999046,
		0.997848,
		0.995877,
		0.992941,
		0.988867,
		0.983507,
		0.976739,
		0.968469,
		0.958631,
		0.947185,
		0.934112,
		0.919418,
		0.903125,
		0.885274,
		0.865921,
		0.845133,
		0.82299,
		0.799577,
		0.774992,
		0.749335,
		0.722712,
		0.695232,
		0.667009,
		0.638156,
		0.608789,
		0.579023,
		0.548975,
		0.518758,
		0.488488,
		0.458277,
		0.428234,
		0.39847,
		0.369091,
		0.3402,
		0.311899,
		0.284287,
		0.25746,
		0.231511,
		0.206529,
		0.182603,
		0.159817,
		0.13825,
		0.117983,
		0.099088,
		0.08164,
		0.065706,
		0.051353,
		0.038645,
		0.027641,
		0.018399,
		0.010974,
		0.005419,
		0.001783,
		0.000112,
		0.004665,
		0.022285,
		0.050862,
		0.088318,
		0.132788,
		0.182603,
		0.236278,
		0.292492,
		0.350084,
		0.408035,
		0.465459,
		0.521595,
		0.575799,
		0.62753,
		0.676351,
		0.721915,
		0.763962,
		0.802314,
		0.836868,
		0.86759,
		0.894513,
		0.917733,
		0.937398,
		0.953711,
		0.96692,
		0.977312,
		0.985208,
		0.990954,
		0.994912,
		0.997446,
		0.998913,
		0.999643,
		0.999927,
		0.999995,
		1,
		0.999995,
		0.999933,
		0.999687,
		0.999086,
		0.997936,
		0.996042,
		0.993217,
		0.989293,
		0.984125,
		0.977591,
		0.9696,
		0.960083,
		0.948998,
		0.936326,
		0.922066,
		0.90624,
		0.888882,
		0.870045,
		0.849792,
		0.828196,
		0.80534,
		0.781314,
		0.756216,
		0.730146,
		0.70321,
		0.675515,
		0.647171,
		0.618289,
		0.588982,
		0.559361,
		0.529537,
		0.499621,
		0.469722,
		0.439948,
		0.410404,
		0.381194,
		0.35242,
		0.324181,
		0.296573,
		0.269691,
		0.243624,
		0.218463,
		0.194292,
		0.171193,
		0.149248,
		0.128533,
		0.109121,
		0.091084,
		0.07449,
		0.059405,
		0.045892,
		0.034009,
		0.023816,
		0.015366,
		0.008711,
		0.003901,
		0.000982,
		0.000295,
		0.010032,
		0.031834,
		0.063557,
		0.103273,
		0.149248,
		0.199933,
		0.253947,
		0.310066,
		0.367211,
		0.424437,
		0.480925,
		0.535972,
		0.588982,
		0.63946,
		0.687003,
		0.731298,
		0.772112,
		0.809287,
		0.842738,
		0.872446,
		0.898452,
		0.920857,
		0.939815,
		0.955526,
		0.968237,
		0.978228,
		0.985814,
		0.99133,
		0.995126,
		0.997555,
		0.99896,
		0.999658,
		0.99993,
		0.999995,
		1,
		0.999996,
		0.999936,
		0.9997,
		0.999123,
		0.998019,
		0.996199,
		0.99348,
		0.9897,
		0.984714,
		0.978406,
		0.970681,
		0.961473,
		0.950736,
		0.938448,
		0.924608,
		0.909232,
		0.892353,
		0.874016,
		0.854281,
		0.833218,
		0.810904,
		0.787425,
		0.762874,
		0.737347,
		0.710944,
		0.68377,
		0.655929,
		0.62753,
		0.59868,
		0.569487,
		0.540059,
		0.510503,
		0.480925,
		0.45143,
		0.42212,
		0.393097,
		0.364459,
		0.336303,
		0.308725,
		0.281814,
		0.255662,
		0.230353,
		0.205974,
		0.182603,
		0.160321,
		0.139203,
		0.119322,
		0.100747,
		0.083548,
		0.067787,
		0.053528,
		0.04083,
		0.02975,
		0.020342,
		0.012657,
		0.006746,
		0.002655,
		0.000428,
		0.002539,
		0.017058,
		0.042546,
		0.077011,
		0.118659,
		0.165879,
		0.21723,
		0.271429,
		0.32734,
		0.383959,
		0.440412,
		0.495938,
		0.549888,
		0.601712,
		0.650955,
		0.69725,
		0.74031,
		0.779928,
		0.815966,
		0.848355,
		0.877086,
		0.902212,
		0.923837,
		0.942118,
		0.957254,
		0.969489,
		0.979099,
		0.986389,
		0.991686,
		0.995329,
		0.997658,
		0.999004,
		0.999673,
		0.999933,
		0.999996,
		1,
		0.999996,
		0.999939,
		0.999713,
		0.999159,
		0.998099,
		0.996348,
		0.993731,
		0.990087,
		0.985277,
		0.979184,
		0.971715,
		0.962803,
		0.952401,
		0.940484,
		0.927049,
		0.912108,
		0.895691,
		0.877839,
		0.858608,
		0.838063,
		0.816277,
		0.793332,
		0.769316,
		0.744321,
		0.718443,
		0.691782,
		0.664439,
		0.636519,
		0.608124,
		0.579359,
		0.55033,
		0.521138,
		0.491888,
		0.462682,
		0.433618,
		0.404796,
		0.376311,
		0.34826,
		0.320733,
		0.29382,
		0.26761,
		0.242186,
		0.217631,
		0.194026,
		0.171447,
		0.149968,
		0.129662,
		0.110599,
		0.092843,
		0.076461,
		0.061512,
		0.048056,
		0.03615,
		0.025848,
		0.0172,
		0.010256,
		0.005063,
		0.001665,
		0.000105,
		0.006746,
		0.02551,
		0.05424,
		0.091084,
		0.134371,
		0.182603,
		0.234441,
		0.28869,
		0.344293,
		0.40032,
		0.455958,
		0.510503,
		0.563354,
		0.614003,
		0.662033,
		0.707107,
		0.748966,
		0.787425,
		0.822365,
		0.853729,
		0.881522,
		0.905802,
		0.92668,
		0.944313,
		0.9589,
		0.970681,
		0.979927,
		0.986936,
		0.992024,
		0.995521,
		0.997755,
		0.999046,
		0.999687,
		0.999936,
		0.999996,
		1,
		0.999996,
		0.999941,
		0.999725,
		0.999193,
		0.998174,
		0.99649,
		0.993969,
		0.990456,
		0.985814,
		0.979927,
		0.972704,
		0.964076,
		0.953996,
		0.942438,
		0.929393,
		0.914874,
		0.898903,
		0.881522,
		0.86278,
		0.842738,
		0.821467,
		0.799043,
		0.775551,
		0.751077,
		0.725714,
		0.699558,
		0.672708,
		0.645261,
		0.617319,
		0.588982,
		0.560352,
		0.531529,
		0.502614,
		0.473704,
		0.444897,
		0.416289,
		0.387975,
		0.360045,
		0.332591,
		0.3057,
		0.279458,
		0.253947,
		0.229249,
		0.205443,
		0.182603,
		0.160804,
		0.140116,
		0.120607,
		0.102344,
		0.085389,
		0.069803,
		0.055645,
		0.042971,
		0.031834,
		0.022285,
		0.014374,
		0.008146,
		0.003647,
		0.000918,
		0.001092,
		0.012657,
		0.035182,
		0.066755,
		0.105651,
		0.150317,
		0.199356,
		0.25152,
		0.3057,
		0.360911,
		0.416289,
		0.471079,
		0.524629,
		0.576381,
		0.625869,
		0.672708,
		0.71659,
		0.757282,
		0.794618,
		0.828496,
		0.858873,
		0.885764,
		0.909232,
		0.929393,
		0.946406,
		0.960468,
		0.971816,
		0.980715,
		0.987455,
		0.992346,
		0.995704,
		0.997848,
		0.999086,
		0.9997,
		0.999939,
		0.999996,
		1,
		0.999996,
		0.999944,
		0.999736,
		0.999225,
		0.998246,
		0.996625,
		0.994196,
		0.990808,
		0.986327,
		0.980638,
		0.973651,
		0.965297,
		0.955526,
		0.944313,
		0.931646,
		0.917532,
		0.901995,
		0.88507,
		0.866803,
		0.847251,
		0.826481,
		0.804566,
		0.781584,
		0.757621,
		0.732765,
		0.707107,
		0.680741,
		0.653763,
		0.626271,
		0.598361,
		0.570132,
		0.541681,
		0.513104,
		0.484498,
		0.455958,
		0.427577,
		0.399446,
		0.371655,
		0.344293,
		0.317445,
		0.291195,
		0.265625,
		0.240813,
		0.216838,
		0.193772,
		0.171689,
		0.150658,
		0.130747,
		0.112019,
		0.094539,
		0.078366,
		0.063557,
		0.050169,
		0.038254,
		0.027864,
		0.019046,
		0.011847,
		0.006313,
		0.002483,
		0.0004,
		0.004182,
		0.020042,
		0.045892,
		0.079949,
		0.120607,
		0.166417,
		0.216079,
		0.26843,
		0.322436,
		0.377182,
		0.431863,
		0.485779,
		0.538323,
		0.588982,
		0.637323,
		0.682994,
		0.725714,
		0.765272,
		0.801521,
		0.834373,
		0.863799,
		0.889821,
		0.91251,
		0.931984,
		0.948403,
		0.961963,
		0.972897,
		0.981465,
		0.987949,
		0.992651,
		0.995877,
		0.997936,
		0.999123,
		0.999713,
		0.999941,
		0.999996,
		1,
		0.999996,
		0.999946,
		0.999747,
		0.999256,
		0.998314,
		0.996753,
		0.994413,
		0.991144,
		0.986817,
		0.981318,
		0.974557,
		0.966466,
		0.956994,
		0.946113,
		0.93381,
		0.92009,
		0.904972,
		0.888488,
		0.870682,
		0.851607,
		0.831326,
		0.809907,
		0.787425,
		0.763962,
		0.739603,
		0.714434,
		0.688547,
		0.662033,
		0.634987,
		0.607502,
		0.579674,
		0.551596,
		0.523363,
		0.495067,
		0.466802,
		0.438658,
		0.410723,
		0.383087,
		0.355834,
		0.32905,
		0.302814,
		0.277209,
		0.252311,
		0.228195,
		0.204936,
		0.182603,
		0.161266,
		0.140991,
		0.121842,
		0.103881,
		0.087167,
		0.071756,
		0.057705,
		0.045066,
		0.033889,
		0.024223,
		0.016114,
		0.009606,
		0.004741,
		0.001559,
		9.8e-005,
		0.000264,
		0.009016,
		0.028698,
		0.057479,
		0.093699,
		0.135857,
		0.182603,
		0.232725,
		0.285138,
		0.33888,
		0.393097,
		0.447041,
		0.500061,
		0.551596,
		0.601168,
		0.64838,
		0.692907,
		0.734494,
		0.77295,
		0.808146,
		0.840008,
		0.868517,
		0.893703,
		0.915644,
		0.934459,
		0.950308,
		0.963389,
		0.973927,
		0.982179,
		0.98842,
		0.992941,
		0.996042,
		0.998019,
		0.999159,
		0.999725,
		0.999944,
		0.999996,
		1,
		0.999997,
		0.999948,
		0.999757,
		0.999285,
		0.998379,
		0.996876,
		0.99462,
		0.991466,
		0.987285,
		0.981968,
		0.975425,
		0.967586,
		0.958402,
		0.947842,
		0.93589,
		0.922551,
		0.907839,
		0.891784,
		0.874425,
		0.855814,
		0.836008,
		0.815072,
		0.79308,
		0.770107,
		0.746234,
		0.721547,
		0.696131,
		0.670076,
		0.643472,
		0.61641,
		0.588982,
		0.561279,
		0.533393,
		0.505413,
		0.47743,
		0.449532,
		0.421806,
		0.394338,
		0.367211,
		0.340507,
		0.314308,
		0.28869,
		0.26373,
		0.239503,
		0.216079,
		0.193529,
		0.171921,
		0.151319,
		0.131787,
		0.113386,
		0.096175,
		0.080209,
		0.065543,
		0.05223,
		0.04032,
		0.02986,
		0.020897,
		0.013475,
		0.007634,
		0.003417,
		0.00086,
		0.00228,
		0.015366,
		0.038444,
		0.069803,
		0.107895,
		0.151319,
		0.198815,
		0.249252,
		0.301618,
		0.355015,
		0.40865,
		0.461824,
		0.513933,
		0.564456,
		0.612952,
		0.659052,
		0.702461,
		0.742944,
		0.78033,
		0.814507,
		0.845412,
		0.873037,
		0.89742,
		0.918641,
		0.936823,
		0.952128,
		0.964748,
		0.974909,
		0.982859,
		0.988867,
		0.993217,
		0.996199,
		0.998099,
		0.999193,
		0.999736,
		0.999946,
		0.999997,
		1,
		0.999997,
		0.99995,
		0.999766,
		0.999313,
		0.998441,
		0.996992,
		0.994817,
		0.991772,
		0.987733,
		0.982591,
		0.976257,
		0.968661,
		0.959754,
		0.949502,
		0.937891,
		0.924919,
		0.9106,
		0.89496,
		0.878037,
		0.859876,
		0.840533,
		0.820069,
		0.798554,
		0.776061,
		0.752666,
		0.728452,
		0.7035,
		0.677898,
		0.651733,
		0.625091,
		0.598062,
		0.570735,
		0.543198,
		0.515539,
		0.487844,
		0.460201,
		0.432694,
		0.405406,
		0.378419,
		0.351814,
		0.325668,
		0.300059,
		0.275061,
		0.250747,
		0.227188,
		0.204451,
		0.182603,
		0.16171,
		0.141832,
		0.12303,
		0.105362,
		0.088884,
		0.073649,
		0.05971,
		0.047116,
		0.035914,
		0.026151,
		0.01787,
		0.011113,
		0.00592,
		0.002328,
		0.000375,
		0.006072,
		0.023032,
		0.049119,
		0.082738,
		0.122442,
		0.166922,
		0.215002,
		0.265625,
		0.317848,
		0.370832,
		0.423837,
		0.476215,
		0.527402,
		0.576915,
		0.624345,
		0.669354,
		0.711668,
		0.751077,
		0.787425,
		0.820615,
		0.850596,
		0.877369,
		0.900978,
		0.921508,
		0.939084,
		0.953866,
		0.966046,
		0.975845,
		0.983507,
		0.989293,
		0.99348,
		0.996348,
		0.998174,
		0.999225,
		0.999747,
		0.999948,
		0.999997,
		1,
		0.999997,
		0.999952,
		0.999775,
		0.99934,
		0.998499,
		0.997104,
		0.995005,
		0.992065,
		0.988161,
		0.983187,
		0.977053,
		0.969692,
		0.961052,
		0.951098,
		0.939815,
		0.927198,
		0.91326,
		0.898023,
		0.881522,
		0.863799,
		0.844907,
		0.824905,
		0.803856,
		0.781832,
		0.758906,
		0.735156,
		0.710662,
		0.685507,
		0.659775,
		0.63355,
		0.606919,
		0.579968,
		0.552782,
		0.525446,
		0.498046,
		0.470665,
		0.443386,
		0.416289,
		0.389456,
		0.362964,
		0.336891,
		0.31131,
		0.286296,
		0.261919,
		0.238249,
		0.215353,
		0.193297,
		0.172143,
		0.151954,
		0.132788,
		0.114702,
		0.097753,
		0.081992,
		0.067471,
		0.05424,
		0.042346,
		0.031834,
		0.022747,
		0.015128,
		0.009016,
		0.004448,
		0.001462,
		9.2e-005,
		0.000982,
		0.011419,
		0.031834,
		0.060583,
		0.096175,
		0.137255,
		0.182603,
		0.231119,
		0.281814,
		0.333809,
		0.38632,
		0.438658,
		0.490217,
		0.540475,
		0.588982,
		0.63536,
		0.679298,
		0.720543,
		0.758906,
		0.794247,
		0.826481,
		0.85557,
		0.881522,
		0.904386,
		0.924251,
		0.941245,
		0.955526,
		0.967285,
		0.976739,
		0.984125,
		0.9897,
		0.993731,
		0.99649,
		0.998246,
		0.999256,
		0.999757,
		0.99995,
		0.999997,
		1,
		0.999997,
		0.999954,
		0.999784,
		0.999365,
		0.998556,
		0.99721,
		0.995185,
		0.992346,
		0.988571,
		0.983758,
		0.977817,
		0.970681,
		0.962298,
		0.952633,
		0.941666,
		0.929393,
		0.915824,
		0.900978,
		0.884886,
		0.86759,
		0.849137,
		0.829583,
		0.80899,
		0.787425,
		0.764959,
		0.741665,
		0.717621,
		0.692907,
		0.667604,
		0.641793,
		0.615558,
		0.588982,
		0.562148,
		0.535139,
		0.508038,
		0.480925,
		0.453882,
		0.426988,
		0.40032,
		0.373956,
		0.34797,
		0.322436,
		0.297425,
		0.273008,
		0.249252,
		0.226223,
		0.203986,
		0.182603,
		0.162135,
		0.142639,
		0.124172,
		0.106789,
		0.090543,
		0.075484,
		0.06166,
		0.049119,
		0.037906,
		0.028064,
		0.019635,
		0.012657,
		0.00717,
		0.003208,
		0.000807,
		0.003771,
		0.018122,
		0.041614,
		0.07271,
		0.110014,
		0.152262,
		0.198309,
		0.247127,
		0.297795,
		0.349487,
		0.401474,
		0.453111,
		0.503836,
		0.553161,
		0.600669,
		0.64601,
		0.688896,
		0.729099,
		0.766443,
		0.800807,
		0.832117,
		0.860344,
		0.885504,
		0.907651,
		0.926878,
		0.943313,
		0.957114,
		0.968469,
		0.977591,
		0.984714,
		0.990087,
		0.993969,
		0.996625,
		0.998314,
		0.999285,
		0.999766,
		0.999952,
		0.999997,
		1,
		0.999997,
		0.999956,
		0.999793,
		0.999389,
		0.998609,
		0.997312,
		0.995357,
		0.992614,
		0.988964,
		0.984305,
		0.97855,
		0.971631,
		0.963495,
		0.954108,
		0.943447,
		0.931507,
		0.918295,
		0.903828,
		0.888134,
		0.871252,
		0.853227,
		0.834112,
		0.813964,
		0.792848,
		0.770831,
		0.747985,
		0.724384,
		0.700105,
		0.675225,
		0.649825,
		0.623984,
		0.597782,
		0.571301,
		0.544621,
		0.517822,
		0.490984,
		0.464184,
		0.437501,
		0.41101,
		0.384786,
		0.358903,
		0.333432,
		0.308444,
		0.284006,
		0.260186,
		0.237049,
		0.214658,
		0.193074,
		0.172357,
		0.152563,
		0.13375,
		0.11597,
		0.099276,
		0.083718,
		0.069344,
		0.0562,
		0.044332,
		0.033782,
		0.024591,
		0.0168,
		0.010445,
		0.000238,
		0.008146,
		0.026003,
		0.05223,
		0.085389,
		0.124172,
		0.167396,
		0.213992,
		0.262996,
		0.313545,
		0.364871,
		0.416289,
		0.4672,
		0.517077,
		0.565468,
		0.611985,
		0.656305,
		0.698162,
		0.737347,
		0.773702,
		0.807118,
		0.837533,
		0.864928,
		0.889324,
		0.91078,
		0.929393,
		0.945291,
		0.958631,
		0.9696,
		0.978406,
		0.985277,
		0.990456,
		0.994196,
		0.996753,
		0.998379,
		0.999313,
		0.999775,
		0.999954,
		0.999997,
		1,
		0.999997,
		0.999958,
		0.9998,
		0.999412,
		0.99866,
		0.997409,
		0.995521,
		0.99287,
		0.98934,
		0.984829,
		0.979253,
		0.972543,
		0.964646,
		0.955526,
		0.945162,
		0.933544,
		0.920678,
		0.906578,
		0.891271,
		0.874792,
		0.857183,
		0.838495,
		0.818782,
		0.798105,
		0.776529,
		0.754123,
		0.730958,
		0.707107,
		0.682646,
		0.657651,
		0.632201,
		0.606372,
		0.580244,
		0.553895,
		0.527402,
		0.500843,
		0.474293,
		0.447829,
		0.421525,
		0.395453,
		0.369686,
		0.344293,
		0.319344,
		0.294906,
		0.271043,
		0.247821,
		0.2253,
		0.203541,
		0.182603,
		0.162543,
		0.143415,
		0.125272,
		0.108167,
		0.092147,
		0.077262,
		0.063557,
		0.051077,
		0.039864,
		0.029959,
		0.002059,
		0.013913,
		0.034906,
		0.063557,
		0.098521,
		0.138573,
		0.182603,
		0.229612,
		0.278696,
		0.32905,
		0.379951,
		0.430764,
		0.480925,
		0.529947,
		0.577405,
		0.622942,
		0.666257,
		0.707107,
		0.745299,
		0.780692,
		0.813189,
		0.842738,
		0.869329,
		0.892988,
		0.913781,
		0.931803,
		0.947185,
		0.960083,
		0.970681,
		0.979184,
		0.985814,
		0.990808,
		0.994413,
		0.996876,
		0.998441,
		0.99934,
		0.999784,
		0.999956,
		0.999997,
		1,
		0.999997,
		0.999959,
		0.999808,
		0.999434,
		0.998709,
		0.997501,
		0.995678,
		0.993115,
		0.9897,
		0.985332,
		0.979927,
		0.973418,
		0.965752,
		0.956891,
		0.946813,
		0.935507,
		0.922975,
		0.909232,
		0.894301,
		0.878214,
		0.861011,
		0.842738,
		0.82345,
		0.803203,
		0.782059,
		0.760084,
		0.737347,
		0.713918,
		0.68987,
		0.665277,
		0.640214,
		0.614757,
		0.588982,
		0.562964,
		0.53678,
		0.510503,
		0.48421,
		0.457972,
		0.431863,
		0.405954,
		0.380315,
		0.355015,
		0.330122,
		0.3057,
		0.281814,
		0.258527,
		0.2359,
		0.213992,
		0.19286,
		0.172561,
		0.153149,
		0.134676,
		0.117192,
		0.100747,
		0.085389,
		0.071162,
		0.058111,
		0.005495,
		0.020897,
		0.044689,
		0.075484,
		0.112019,
		0.153149,
		0.197833,
		0.245133,
		0.294205,
		0.344293,
		0.394723,
		0.444897,
		0.494292,
		0.542452,
		0.588982,
		0.63355,
		0.675879,
		0.715743,
		0.752968,
		0.787425,
		0.819031,
		0.847743,
		0.873557,
		0.896506,
		0.916658,
		0.934112,
		0.948998,
		0.961473,
		0.971715,
		0.979927,
		0.986327,
		0.991144,
		0.99462,
		0.996992,
		0.998499,
		0.999365,
		0.999793,
		0.999958,
		0.999997,
		1,
		0.999997,
		0.999961,
		0.999815,
		0.999454,
		0.998756,
		0.99759,
		0.995829,
		0.99335,
		0.990045,
		0.985814,
		0.980575,
		0.97426,
		0.966816,
		0.958205,
		0.948403,
		0.937398,
		0.925192,
		0.911794,
		0.897228,
		0.881522,
		0.864714,
		0.846847,
		0.827973,
		0.808146,
		0.787425,
		0.765874,
		0.743557,
		0.720543,
		0.696903,
		0.672708,
		0.648029,
		0.622942,
		0.597518,
		0.571833,
		0.545959,
		0.519969,
		0.493936,
		0.467931,
		0.442026,
		0.416289,
		0.39079,
		0.365596,
		0.340772,
		0.316383,
		0.292492,
		0.269161,
		0.246449,
		0.224414,
		0.203114,
		0.182603,
		0.162935,
		0.144161,
		0.126332,
		0.109496,
		0.093699,
		0.000888,
		0.010354,
		0.028941,
		0.055226,
		0.08791,
		0.125807,
		0.167843,
		0.213043,
		0.260527,
		0.309504,
		0.359265,
		0.409179,
		0.45869,
		0.507305,
		0.5546,
		0.600208,
		0.643821,
		0.68518,
		0.724081,
		0.760363,
		0.793912,
		0.824654,
		0.852555,
		0.877618,
		0.899882,
		0.919418,
		0.936326,
		0.950736,
		0.962803,
		0.972704,
		0.980638,
		0.986817,
		0.991466,
		0.994817,
		0.997104,
		0.998556,
		0.999389,
		0.9998,
		0.999959,
		0.999997,
		1,
		0.999997,
		0.999962,
		0.999822,
		0.999474,
		0.9988,
		0.997675,
		0.995972,
		0.993576,
		0.990375,
		0.986276,
		0.981196,
		0.975068,
		0.967838,
		0.959469,
		0.949934,
		0.939222,
		0.92733,
		0.914268,
		0.900056,
		0.884721,
		0.868297,
		0.850827,
		0.832357,
		0.812941,
		0.792634,
		0.771497,
		0.749594,
		0.726989,
		0.703751,
		0.679948,
		0.655651,
		0.63093,
		0.605858,
		0.580504,
		0.554942,
		0.529241,
		0.503473,
		0.477707,
		0.452013,
		0.426457,
		0.401108,
		0.376031,
		0.351291,
		0.32695,
		0.303071,
		0.279714,
		0.256937,
		0.234798,
		0.213353,
		0.192655,
		0.172758,
		0.153712,
		0.135567,
		0.003417,
		0.016464,
		0.037906,
		0.066407,
		0.100747,
		0.139816,
		0.182603,
		0.228195,
		0.275766,
		0.324575,
		0.373956,
		0.42332,
		0.472143,
		0.519969,
		0.566399,
		0.611093,
		0.653763,
		0.694173,
		0.732133,
		0.767497,
		0.800163,
		0.830067,
		0.857183,
		0.881522,
		0.903125,
		0.922066,
		0.938448,
		0.952401,
		0.964076,
		0.973651,
		0.981318,
		0.987285,
		0.991772,
		0.995005,
		0.99721,
		0.998609,
		0.999412,
		0.999808,
		0.999961,
		0.999997,
		1,
		0.999998,
		0.999964,
		0.999829,
		0.999493,
		0.998843,
		0.997755,
		0.99611,
		0.993791,
		0.990693,
		0.98672,
		0.981794,
		0.975845,
		0.968823,
		0.960686,
		0.95141,
		0.94098,
		0.929393,
		0.916658,
		0.90279,
		0.887815,
		0.871765,
		0.854681,
		0.836606,
		0.817592,
		0.79769,
		0.77696,
		0.755463,
		0.73326,
		0.710418,
		0.687003,
		0.663083,
		0.638727,
		0.614003,
		0.588982,
		0.563732,
		0.538323,
		0.512824,
		0.487302,
		0.461824,
		0.436458,
		0.411269,
		0.38632,
		0.361676,
		0.337397,
		0.313545,
		0.290179,
		0.267356,
		0.245133,
		0.223565,
		0.202704,
		0.182603,
		0.000216,
		0.007397,
		0.02367,
		0.047667,
		0.078131,
		0.113919,
		0.153986,
		0.197385,
		0.243258,
		0.29083,
		0.339405,
		0.38836,
		0.437142,
		0.485261,
		0.53229,
		0.577857,
		0.621646,
		0.663389,
		0.702868,
		0.739908,
		0.774379,
		0.806187,
		0.835279,
		0.861636,
		0.885274,
		0.90624,
		0.924608,
		0.940484,
		0.953996,
		0.965297,
		0.974557,
		0.981968,
		0.987733,
		0.992065,
		0.995185,
		0.997312,
		0.99866,
		0.999434,
		0.999815,
		0.999962,
		0.999998,
		1,
		0.999998,
		0.999965,
		0.999835,
		0.999511,
		0.998883,
		0.997833,
		0.996242,
		0.993998,
		0.990997,
		0.987147,
		0.982367,
		0.976593,
		0.96977,
		0.961859,
		0.952833,
		0.942676,
		0.931385,
		0.918966,
		0.905432,
		0.890808,
		0.875123,
		0.858415,
		0.840726,
		0.822103,
		0.802599,
		0.782268,
		0.761169,
		0.739362,
		0.71691,
		0.693878,
		0.670332,
		0.646337,
		0.62196,
		0.59727,
		0.572334,
		0.547218,
		0.52199,
		0.496716,
		0.471461,
		0.446292,
		0.42127,
		0.39646,
		0.371923,
		0.34772,
		0.32391,
		0.300551,
		0.2777,
		0.255412,
		0.233741,
		0.001868,
		0.012657,
		0.031834,
		0.058111,
		0.090309,
		0.127354,
		0.168264,
		0.212149,
		0.258204,
		0.3057,
		0.353984,
		0.402472,
		0.450646,
		0.498046,
		0.544274,
		0.588982,
		0.631876,
		0.672708,
		0.711275,
		0.747418,
		0.781018,
		0.811993,
		0.840298,
		0.865921,
		0.888882,
		0.909232,
		0.927049,
		0.942438,
		0.955526,
		0.966466,
		0.975425,
		0.982591,
		0.988161,
		0.992346,
		0.995357,
		0.997409,
		0.998709,
		0.999454,
		0.999822,
		0.999964,
		0.999998,
		1,
		0.999998,
		0.999966,
		0.999841,
		0.999529,
		0.998922,
		0.997907,
		0.996368,
		0.994196,
		0.991289,
		0.987556,
		0.982919,
		0.977312,
		0.970681,
		0.962988,
		0.954204,
		0.944313,
		0.933308,
		0.921195,
		0.907987,
		0.893703,
		0.878374,
		0.862033,
		0.84472,
		0.826481,
		0.807365,
		0.787425,
		0.766717,
		0.745299,
		0.723232,
		0.700578,
		0.6774,
		0.653763,
		0.629732,
		0.605372,
		0.580749,
		0.555928,
		0.530974,
		0.505952,
		0.480925,
		0.455958,
		0.431113,
		0.40645,
		0.382031,
		0.357915,
		0.33416,
		0.310823,
		0.28796,
		0.004996,
		0.019046,
		0.04083,
		0.069138,
		0.102863,
		0.140991,
		0.182603,
		0.226862,
		0.273008,
		0.32036,
		0.368303,
		0.416289,
		0.463832,
		0.510503,
		0.555928,
		0.599783,
		0.641793,
		0.681729,
		0.719404,
		0.754672,
		0.787425,
		0.817592,
		0.845133,
		0.870045,
		0.892353,
		0.912108,
		0.929393,
		0.944313,
		0.956994,
		0.967586,
		0.976257,
		0.983187,
		0.988571,
		0.992614,
		0.995521,
		0.997501,
		0.998756,
		0.999474,
		0.999829,
		0.999965,
		0.999998,
		1,
		0.999998,
		0.999968,
		0.999846,
		0.999545,
		0.99896,
		0.997978,
		0.99649,
		0.994386,
		0.991569,
		0.987949,
		0.983449,
		0.978003,
		0.971559,
		0.964076,
		0.955526,
		0.945892,
		0.935165,
		0.92335,
		0.910457,
		0.896506,
		0.881522,
		0.865538,
		0.848593,
		0.830729,
		0.811993,
		0.792437,
		0.772112,
		0.751077,
		0.729388,
		0.707107,
		0.684294,
		0.661012,
		0.637323,
		0.613292,
		0.588982,
		0.564456,
		0.539778,
		0.515011,
		0.490217,
		0.465459,
		0.440796,
		0.416289,
		0.391998,
		0.36798,
		0.344293,
		0.000807,
		0.009431,
		0.026425,
		0.050548,
		0.08066,
		0.11572,
		0.154777,
		0.196963,
		0.241492,
		0.287651,
		0.334797,
		0.382355,
		0.429809,
		0.476704,
		0.522638,
		0.567259,
		0.610268,
		0.651406,
		0.690463,
		0.727265,
		0.76168,
		0.793609,
		0.82299,
		0.849792,
		0.874016,
		0.895691,
		0.914874,
		0.931646,
		0.946113,
		0.958402,
		0.968661,
		0.977053,
		0.983758,
		0.988964,
		0.99287,
		0.995678,
		0.99759,
		0.9988,
		0.999493,
		0.999835,
		0.999966,
		0.999998,
		1,
		0.999998,
		0.999969,
		0.999852,
		0.999561,
		0.998995,
		0.998046,
		0.996606,
		0.994569,
		0.991839,
		0.988327,
		0.983959,
		0.978669,
		0.972405,
		0.965125,
		0.956802,
		0.947416,
		0.936959,
		0.925433,
		0.912847,
		0.899218,
		0.884571,
		0.868936,
		0.85235,
		0.834852,
		0.816488,
		0.797307,
		0.777359,
		0.756699,
		0.735383,
		0.713469,
		0.691017,
		0.668086,
		0.644738,
		0.621034,
		0.597036,
		0.572806,
		0.548406,
		0.523896,
		0.499339,
		0.474793,
		0.45032,
		0.425977,
		0.401822,
		0.003111,
		0.015024,
		0.034672,
		0.060886,
		0.092596,
		0.128819,
		0.168662,
		0.211307,
		0.256014,
		0.302114,
		0.349001,
		0.396135,
		0.443032,
		0.489264,
		0.534454,
		0.578275,
		0.620445,
		0.660725,
		0.698919,
		0.734868,
		0.76845,
		0.799577,
		0.828196,
		0.854281,
		0.877839,
		0.898903,
		0.917532,
		0.93381,
		0.947842,
		0.959754,
		0.969692,
		0.977817,
		0.984305,
		0.98934,
		0.993115,
		0.995829,
		0.997675,
		0.998843,
		0.999511,
		0.999841,
		0.999968,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999857,
		0.999576,
		0.999029,
		0.998111,
		0.996717,
		0.994744,
		0.992097,
		0.988691,
		0.98445,
		0.97931,
		0.973219,
		0.966137,
		0.958032,
		0.948887,
		0.938692,
		0.927446,
		0.915158,
		0.901844,
		0.887524,
		0.87223,
		0.855993,
		0.838854,
		0.820854,
		0.80204,
		0.782462,
		0.762171,
		0.741222,
		0.71967,
		0.697574,
		0.67499,
		0.651979,
		0.628601,
		0.604914,
		0.58098,
		0.556859,
		0.532609,
		0.508291,
		0.483964,
		0.459685,
		0.000196,
		0.006746,
		0.021638,
		0.043674,
		0.071756,
		0.104874,
		0.142104,
		0.182603,
		0.225603,
		0.270407,
		0.316383,
		0.362964,
		0.409641,
		0.455958,
		0.501515,
		0.545959,
		0.588982,
		0.630323,
		0.669758,
		0.707107,
		0.742222,
		0.774992,
		0.80534,
		0.833218,
		0.858608,
		0.881522,
		0.901995,
		0.92009,
		0.93589,
		0.949502,
		0.961052,
		0.970681,
		0.97855,
		0.984829,
		0.9897,
		0.99335,
		0.995972,
		0.997755,
		0.998883,
		0.999529,
		0.999846,
		0.999969,
		0.999998,
		1,
		0.999998,
		0.999971,
		0.999862,
		0.999591,
		0.999062,
		0.998174,
		0.996824,
		0.994912,
		0.992346,
		0.98904,
		0.984922,
		0.979927,
		0.974004,
		0.967112,
		0.95922,
		0.950308,
		0.940367,
		0.929393,
		0.917395,
		0.904386,
		0.890386,
		0.875423,
		0.859528,
		0.842738,
		0.825095,
		0.806641,
		0.787425,
		0.767497,
		0.746909,
		0.725714,
		0.703969,
		0.681729,
		0.659052,
		0.635997,
		0.612621,
		0.588982,
		0.56514,
		0.541152,
		0.517077,
		0.001703,
		0.011564,
		0.02915,
		0.053333,
		0.083077,
		0.117431,
		0.155525,
		0.196564,
		0.239825,
		0.28465,
		0.330446,
		0.376679,
		0.422868,
		0.468589,
		0.513462,
		0.557157,
		0.599388,
		0.63991,
		0.678515,
		0.715035,
		0.749335,
		0.781314,
		0.810904,
		0.838063,
		0.86278,
		0.88507,
		0.904972,
		0.922551,
		0.937891,
		0.951098,
		0.962298,
		0.971631,
		0.979253,
		0.985332,
		0.990045,
		0.993576,
		0.99611,
		0.997833,
		0.998922,
		0.999545,
		0.999852,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999972,
		0.999867,
		0.999605,
		0.999093,
		0.998234,
		0.996926,
		0.995074,
		0.992584,
		0.989376,
		0.985376,
		0.980522,
		0.974761,
		0.968053,
		0.960366,
		0.951681,
		0.941985,
		0.931276,
		0.91956,
		0.906848,
		0.893159,
		0.878519,
		0.862958,
		0.84651,
		0.829215,
		0.811114,
		0.792253,
		0.772681,
		0.752448,
		0.731605,
		0.710206,
		0.688306,
		0.66596,
		0.643226,
		0.620158,
		0.596815,
		0.573252,
		0.004562,
		0.01743,
		0.03745,
		0.063557,
		0.094776,
		0.13021,
		0.169038,
		0.210511,
		0.253947,
		0.298728,
		0.344293,
		0.39014,
		0.435818,
		0.480925,
		0.525108,
		0.568056,
		0.609502,
		0.649214,
		0.687003,
		0.722712,
		0.756216,
		0.787425,
		0.816277,
		0.842738,
		0.866803,
		0.888488,
		0.907839,
		0.924919,
		0.939815,
		0.952633,
		0.963495,
		0.972543,
		0.979927,
		0.985814,
		0.990375,
		0.993791,
		0.996242,
		0.997907,
		0.99896,
		0.999561,
		0.999857,
		0.999971,
		0.999998,
		1,
		0.999998,
		0.999973,
		0.999871,
		0.999618,
		0.999123,
		0.998291,
		0.997025,
		0.995229,
		0.992814,
		0.9897,
		0.985814,
		0.981094,
		0.97549,
		0.96896,
		0.961473,
		0.953007,
		0.94355,
		0.933098,
		0.921655,
		0.909232,
		0.895847,
		0.881522,
		0.866286,
		0.850172,
		0.833218,
		0.815463,
		0.796951,
		0.777728,
		0.757843,
		0.737347,
		0.71629,
		0.694725,
		0.672708,
		0.650291,
		0.62753,
		0.000737,
		0.008627,
		0.024223,
		0.046438,
		0.074267,
		0.106789,
		0.14316,
		0.182603,
		0.224414,
		0.26795,
		0.312626,
		0.357915,
		0.403344,
		0.448489,
		0.492972,
		0.53646,
		0.578663,
		0.619329,
		0.658245,
		0.695232,
		0.730146,
		0.762874,
		0.793332,
		0.821467,
		0.847251,
		0.870682,
		0.891784,
		0.9106,
		0.927198,
		0.941666,
		0.954108,
		0.964646,
		0.973418,
		0.980575,
		0.986276,
		0.990693,
		0.993998,
		0.996368,
		0.997978,
		0.998995,
		0.999576,
		0.999862,
		0.999972,
		0.999998,
		1,
		0.999998,
		0.999974,
		0.999876,
		0.999631,
		0.999152,
		0.998347,
		0.997119,
		0.995378,
		0.993035,
		0.990011,
		0.986235,
		0.981647,
		0.976194,
		0.969836,
		0.962541,
		0.954288,
		0.945063,
		0.93486,
		0.923684,
		0.911542,
		0.898452,
		0.884434,
		0.869516,
		0.853729,
		0.837108,
		0.819691,
		0.801521,
		0.782641,
		0.7631,
		0.742944,
		0.722224,
		0.700991,
		0.679298,
		0.002844,
		0.013765,
		0.031834,
		0.056024,
		0.085389,
		0.119058,
		0.156234,
		0.196187,
		0.238249,
		0.281814,
		0.326332,
		0.371306,
		0.416289,
		0.460882,
		0.504731,
		0.547521,
		0.588982,
		0.628878,
		0.667009,
		0.70321,
		0.737347,
		0.769316,
		0.799043,
		0.826481,
		0.851607,
		0.874425,
		0.89496,
		0.91326,
		0.929393,
		0.943447,
		0.955526,
		0.965752,
		0.97426,
		0.981196,
		0.98672,
		0.990997,
		0.994196,
		0.99649,
		0.998046,
		0.999029,
		0.999591,
		0.999867,
		0.999973,
		0.999998,
		1,
		0.999998,
		0.999975,
		0.99988,
		0.999643,
		0.99918,
		0.9984,
		0.99721,
		0.995521,
		0.993247,
		0.99031,
		0.986641,
		0.982179,
		0.976872,
		0.970681,
		0.963574,
		0.955526,
		0.946526,
		0.936566,
		0.925648,
		0.913781,
		0.900978,
		0.88726,
		0.872652,
		0.857183,
		0.840888,
		0.823803,
		0.805968,
		0.787425,
		0.76822,
		0.7484,
		0.728012,
		0.000179,
		0.006177,
		0.019856,
		0.040162,
		0.066128,
		0.096858,
		0.131531,
		0.169394,
		0.209758,
		0.251992,
		0.295525,
		0.339837,
		0.38446,
		0.428973,
		0.473,
		0.516206,
		0.558299,
		0.599022,
		0.638156,
		0.675515,
		0.710944,
		0.744321,
		0.775551,
		0.804566,
		0.831326,
		0.855814,
		0.878037,
		0.898023,
		0.915824,
		0.931507,
		0.945162,
		0.956891,
		0.966816,
		0.975068,
		0.981794,
		0.987147,
		0.991289,
		0.994386,
		0.996606,
		0.998111,
		0.999062,
		0.999605,
		0.999871,
		0.999974,
		0.999998,
		1,
		0.999998,
		0.999976,
		0.999884,
		0.999655,
		0.999206,
		0.998451,
		0.997298,
		0.995659,
		0.993452,
		0.990599,
		0.987032,
		0.982692,
		0.977527,
		0.971498,
		0.964571,
		0.956724,
		0.947941,
		0.938217,
		0.927551,
		0.91595,
		0.903427,
		0.890001,
		0.875696,
		0.860539,
		0.844563,
		0.827802,
		0.810295,
		0.792083,
		0.77321,
		0.001559,
		0.010606,
		0.026791,
		0.049119,
		0.076675,
		0.108615,
		0.144161,
		0.182603,
		0.223289,
		0.265625,
		0.30907,
		0.353134,
		0.397375,
		0.441396,
		0.484843,
		0.527402,
		0.568797,
		0.608789,
		0.647171,
		0.68377,
		0.718443,
		0.751077,
		0.781584,
		0.809907,
		0.836008,
		0.859876,
		0.881522,
		0.900978,
		0.918295,
		0.933544,
		0.946813,
		0.958205,
		0.967838,
		0.975845,
		0.982367,
		0.987556,
		0.991569,
		0.994569,
		0.996717,
		0.998174,
		0.999093,
		0.999618,
		0.999876,
		0.999975,
		0.999998,
		1,
		0.999998,
		0.999976,
		0.999888,
		0.999666,
		0.999232,
		0.998499,
		0.997381,
		0.995792,
		0.993648,
		0.990877,
		0.987409,
		0.983187,
		0.978159,
		0.972286,
		0.965534,
		0.957881,
		0.949311,
		0.939815,
		0.929393,
		0.918052,
		0.905802,
		0.892661,
		0.878652,
		0.863799,
		0.848135,
		0.831692,
		0.814507,
		0.004182,
		0.016011,
		0.034471,
		0.058623,
		0.087601,
		0.120607,
		0.156908,
		0.19583,
		0.236757,
		0.27913,
		0.322436,
		0.366214,
		0.410046,
		0.453557,
		0.496415,
		0.538323,
		0.579023,
		0.618289,
		0.655929,
		0.691782,
		0.725714,
		0.757621,
		0.787425,
		0.815072,
		0.840533,
		0.863799,
		0.884886,
		0.903828,
		0.920678,
		0.935507,
		0.948403,
		0.959469,
		0.968823,
		0.976593,
		0.982919,
		0.987949,
		0.991839,
		0.994744,
		0.996824,
		0.998234,
		0.999123,
		0.999631,
		0.99988,
		0.999976,
		0.999998,
		1,
		0.999998,
		0.999977,
		0.999891,
		0.999677,
		0.999256,
		0.998546,
		0.997462,
		0.995919,
		0.993838,
		0.991144,
		0.987773,
		0.983664,
		0.978769,
		0.973047,
		0.966466,
		0.959001,
		0.950636,
		0.941362,
		0.931179,
		0.92009,
		0.908106,
		0.895243,
		0.881522,
		0.866967,
		0.851607,
		0.000675,
		0.007921,
		0.022285,
		0.042808,
		0.068601,
		0.098846,
		0.132788,
		0.169733,
		0.209044,
		0.250141,
		0.292492,
		0.335615,
		0.379073,
		0.422472,
		0.465459,
		0.507719,
		0.548975,
		0.588982,
		0.62753,
		0.664439,
		0.699558,
		0.732765,
		0.763962,
		0.79308,
		0.820069,
		0.844907,
		0.86759,
		0.888134,
		0.906578,
		0.922975,
		0.937398,
		0.949934,
		0.960686,
		0.96977,
		0.977312,
		0.983449,
		0.988327,
		0.992097,
		0.994912,
		0.996926,
		0.998291,
		0.999152,
		0.999643,
		0.999884,
		0.999976,
		0.999998,
		1,
		0.999999,
		0.999978,
		0.999895,
		0.999687,
		0.99928,
		0.998592,
		0.99754,
		0.996042,
		0.99402,
		0.991403,
		0.988123,
		0.984125,
		0.979358,
		0.973783,
		0.967366,
		0.960083,
		0.951918,
		0.942861,
		0.932909,
		0.922066,
		0.910341,
		0.897749,
		0.88431,
		0.00261,
		0.012657,
		0.02933,
		0.05172,
		0.078987,
		0.110356,
		0.145114,
		0.182603,
		0.222223,
		0.263423,
		0.3057,
		0.348599,
		0.391707,
		0.434652,
		0.477101,
		0.518758,
		0.559361,
		0.59868,
		0.636519,
		0.672708,
		0.707107,
		0.739603,
		0.770107,
		0.798554,
		0.824905,
		0.849137,
		0.871252,
		0.891271,
		0.909232,
		0.925192,
		0.939222,
		0.95141,
		0.961859,
		0.970681,
		0.978003,
		0.983959,
		0.988691,
		0.992346,
		0.995074,
		0.997025,
		0.998347,
		0.99918,
		0.999655,
		0.999888,
		0.999977,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999898,
		0.999697,
		0.999302,
		0.998635,
		0.997614,
		0.99616,
		0.994196,
		0.991651,
		0.988461,
		0.984569,
		0.979927,
		0.974494,
		0.968237,
		0.961131,
		0.95316,
		0.944313,
		0.934586,
		0.923982,
		0.91251,
		0.000165,
		0.005678,
		0.018285,
		0.037057,
		0.061133,
		0.08972,
		0.122084,
		0.157547,
		0.19549,
		0.235343,
		0.276585,
		0.318742,
		0.361381,
		0.404113,
		0.446587,
		0.488488,
		0.529537,
		0.569487,
		0.608124,
		0.645261,
		0.680741,
		0.714434,
		0.746234,
		0.776061,
		0.803856,
		0.829583,
		0.853227,
		0.874792,
		0.894301,
		0.911794,
		0.92733,
		0.94098,
		0.952833,
		0.962988,
		0.971559,
		0.978669,
		0.98445,
		0.98904,
		0.992584,
		0.995229,
		0.997119,
		0.9984,
		0.999206,
		0.999666,
		0.999891,
		0.999978,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999902,
		0.999707,
		0.999324,
		0.998677,
		0.997686,
		0.996274,
		0.994366,
		0.991891,
		0.988787,
		0.984999,
		0.980477,
		0.975181,
		0.969079,
		0.962145,
		0.954362,
		0.945719,
		0.936212,
		0.001432,
		0.009762,
		0.024706,
		0.045385,
		0.070982,
		0.100747,
		0.133985,
		0.170054,
		0.208367,
		0.248386,
		0.289616,
		0.331609,
		0.373956,
		0.416289,
		0.458277,
		0.499621,
		0.540059,
		0.579359,
		0.617319,
		0.653763,
		0.688547,
		0.721547,
		0.752666,
		0.781832,
		0.80899,
		0.834112,
		0.857183,
		0.878214,
		0.897228,
		0.914268,
		0.929393,
		0.942676,
		0.954204,
		0.964076,
		0.972405,
		0.97931,
		0.984922,
		0.989376,
		0.992814,
		0.995378,
		0.99721,
		0.998451,
		0.999232,
		0.999677,
		0.999895,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.99998,
		0.999905,
		0.999716,
		0.999345,
		0.998717,
		0.997755,
		0.996384,
		0.994529,
		0.992122,
		0.989102,
		0.985413,
		0.981008,
		0.975845,
		0.969893,
		0.963126,
		0.955526,
		0.003848,
		0.014759,
		0.031834,
		0.05424,
		0.081207,
		0.112019,
		0.14602,
		0.182603,
		0.221211,
		0.261333,
		0.302502,
		0.344293,
		0.38632,
		0.428234,
		0.469722,
		0.510503,
		0.55033,
		0.588982,
		0.626271,
		0.662033,
		0.696131,
		0.728452,
		0.758906,
		0.787425,
		0.813964,
		0.838495,
		0.861011,
		0.881522,
		0.900056,
		0.916658,
		0.931385,
		0.944313,
		0.955526,
		0.965125,
		0.973219,
		0.979927,
		0.985376,
		0.9897,
		0.993035,
		0.995521,
		0.997298,
		0.998499,
		0.999256,
		0.999687,
		0.999898,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999908,
		0.999725,
		0.999365,
		0.998756,
		0.997822,
		0.99649,
		0.994686,
		0.992346,
		0.989406,
		0.985814,
		0.981521,
		0.976488,
		0.970681,
		0.000621,
		0.007298,
		0.02057,
		0.039587,
		0.063557,
		0.091751,
		0.123492,
		0.158156,
		0.195168,
		0.234001,
		0.27417,
		0.315234,
		0.356789,
		0.39847,
		0.439948,
		0.480925,
		0.521138,
		0.560352,
		0.598361,
		0.634987,
		0.670076,
		0.7035,
		0.735156,
		0.764959,
		0.792848,
		0.818782,
		0.842738,
		0.864714,
		0.884721,
		0.90279,
		0.918966,
		0.933308,
		0.945892,
		0.956802,
		0.966137,
		0.974004,
		0.980522,
		0.985814,
		0.990011,
		0.993247,
		0.995659,
		0.997381,
		0.998546,
		0.99928,
		0.999697,
		0.999902,
		0.99998,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999911,
		0.999733,
		0.999384,
		0.998793,
		0.997886,
		0.996591,
		0.994838,
		0.992561,
		0.9897,
		0.986201,
		0.982017,
		0.002404,
		0.011678,
		0.02711,
		0.047892,
		0.073275,
		0.102567,
		0.135126,
		0.17036,
		0.207724,
		0.246719,
		0.286884,
		0.327802,
		0.369091,
		0.410404,
		0.45143,
		0.491888,
		0.531529,
		0.570132,
		0.607502,
		0.643472,
		0.677898,
		0.710662,
		0.741665,
		0.770831,
		0.798105,
		0.82345,
		0.846847,
		0.868297,
		0.887815,
		0.905432,
		0.921195,
		0.935165,
		0.947416,
		0.958032,
		0.967112,
		0.974761,
		0.981094,
		0.986235,
		0.99031,
		0.993452,
		0.995792,
		0.997462,
		0.998592,
		0.999302,
		0.999707,
		0.999905,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999913,
		0.999741,
		0.999403,
		0.998829,
		0.997948,
		0.99669,
		0.994985,
		0.992769,
		0.989983,
		0.000152,
		0.005236,
		0.016893,
		0.034297,
		0.056682,
		0.083339,
		0.113609,
		0.146884,
		0.182603,
		0.22025,
		0.259348,
		0.299464,
		0.3402,
		0.381194,
		0.42212,
		0.462682,
		0.502614,
		0.541681,
		0.579674,
		0.61641,
		0.651733,
		0.685507,
		0.717621,
		0.747985,
		0.776529,
		0.803203,
		0.827973,
		0.850827,
		0.871765,
		0.890808,
		0.907987,
		0.92335,
		0.936959,
		0.948887,
		0.95922,
		0.968053,
		0.97549,
		0.981647,
		0.986641,
		0.990599,
		0.993648,
		0.995919,
		0.99754,
		0.998635,
		0.999324,
		0.999716,
		0.999908,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999916,
		0.999749,
		0.999421,
		0.998863,
		0.998008,
		0.996784,
		0.995126,
		0.00132,
		0.009016,
		0.022856,
		0.042059,
		0.065898,
		0.093699,
		0.124837,
		0.158736,
		0.194862,
		0.232725,
		0.271875,
		0.311899,
		0.35242,
		0.393097,
		0.433618,
		0.473704,
		0.513104,
		0.551596,
		0.588982,
		0.625091,
		0.659775,
		0.692907,
		0.724384,
		0.754123,
		0.782059,
		0.808146,
		0.832357,
		0.854681,
		0.875123,
		0.893703,
		0.910457,
		0.925433,
		0.938692,
		0.950308,
		0.960366,
		0.96896,
		0.976194,
		0.982179,
		0.987032,
		0.990877,
		0.993838,
		0.996042,
		0.997614,
		0.998677,
		0.999345,
		0.999725,
		0.999911,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999983,
		0.999919,
		0.999757,
		0.999438,
		0.998897,
		0.998065,
		0.003552,
		0.013648,
		0.029488,
		0.05033,
		0.075484,
		0.10431,
		0.136215,
		0.170651,
		0.207112,
		0.245133,
		0.284287,
		0.324181,
		0.364459,
		0.404796,
		0.444897,
		0.484498,
		0.523363,
		0.561279,
		0.598062,
		0.63355,
		0.667604,
		0.700105,
		0.730958,
		0.760084,
		0.787425,
		0.812941,
		0.836606,
		0.858415,
		0.878374,
		0.896506,
		0.912847,
		0.927446,
		0.940367,
		0.951681,
		0.961473,
		0.969836,
		0.976872,
		0.982692,
		0.987409,
		0.991144,
		0.99402,
		0.99616,
		0.997686,
		0.998717,
		0.999365,
		0.999733,
		0.999913,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999921,
		0.999764,
		0.999454,
		0.000573,
		0.006746,
		0.019046,
		0.036715,
		0.059048,
		0.085389,
		0.11513,
		0.147708,
		0.182603,
		0.219335,
		0.25746,
		0.296573,
		0.336303,
		0.376311,
		0.416289,
		0.455958,
		0.495067,
		0.533393,
		0.570735,
		0.606919,
		0.641793,
		0.675225,
		0.707107,
		0.737347,
		0.765874,
		0.792634,
		0.817592,
		0.840726,
		0.862033,
		0.881522,
		0.899218,
		0.915158,
		0.929393,
		0.941985,
		0.953007,
		0.962541,
		0.970681,
		0.977527,
		0.983187,
		0.987773,
		0.991403,
		0.994196,
		0.996274,
		0.997755,
		0.998756,
		0.999384,
		0.999741,
		0.999916,
		0.999983,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999923,
		0.002221,
		0.010808,
		0.025132,
		0.044472,
		0.068159,
		0.095568,
		0.126123,
		0.159289,
		0.19457,
		0.231511,
		0.269691,
		0.308725,
		0.34826,
		0.387975,
		0.427577,
		0.466802,
		0.505413,
		0.543198,
		0.579968,
		0.615558,
		0.649825,
		0.682646,
		0.713918,
		0.743557,
		0.771497,
		0.79769,
		0.822103,
		0.84472,
		0.865538,
		0.884571,
		0.901844,
		0.917395,
		0.931276,
		0.94355,
		0.954288,
		0.963574,
		0.971498,
		0.978159,
		0.983664,
		0.988123,
		0.991651,
		0.994366,
		0.996384,
		0.997822,
		0.998793,
		0.999403,
		0.999749,
		0.999919,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.00014,
		0.004845,
		0.015654,
		0.031834,
		0.052699,
		0.077611,
		0.10598,
		0.137255,
		0.170928,
		0.206529,
		0.243624,
		0.281814,
		0.320733,
		0.360045,
		0.399446,
		0.438658,
		0.47743,
		0.515539,
		0.552782,
		0.588982,
		0.623984,
		0.657651,
		0.68987,
		0.720543,
		0.749594,
		0.77696,
		0.802599,
		0.826481,
		0.848593,
		0.868936,
		0.887524,
		0.904386,
		0.91956,
		0.933098,
		0.945063,
		0.955526,
		0.964571,
		0.972286,
		0.978769,
		0.984125,
		0.988461,
		0.991891,
		0.994529,
		0.99649,
		0.997886,
		0.998829,
		0.999421,
		0.999757,
		0.999921,
		0.999984,
		0.999999,
		0.001221,
		0.008351,
		0.021206,
		0.039085,
		0.061339,
		0.08736,
		0.116587,
		0.148496,
		0.182603,
		0.218463,
		0.255662,
		0.29382,
		0.332591,
		0.371655,
		0.410723,
		0.449532,
		0.487844,
		0.525446,
		0.562148,
		0.597782,
		0.632201,
		0.665277,
		0.696903,
		0.726989,
		0.755463,
		0.782268,
		0.807365,
		0.830729,
		0.85235,
		0.87223,
		0.890386,
		0.906848,
		0.921655,
		0.93486,
		0.946526,
		0.956724,
		0.965534,
		0.973047,
		0.979358,
		0.984569,
		0.988787,
		0.992122,
		0.994686,
		0.996591,
		0.997948,
		0.998863,
		0.999438,
		0.999764,
		0.999923,
		0.003289,
		0.012657,
		0.027391,
		0.046826,
		0.070342,
		0.097363,
		0.127354,
		0.159817,
		0.194292,
		0.230353,
		0.26761,
		0.3057,
		0.344293,
		0.383087,
		0.421806,
		0.460201,
		0.498046,
		0.535139,
		0.571301,
		0.606372,
		0.640214,
		0.672708,
		0.703751,
		0.73326,
		0.761169,
		0.787425,
		0.811993,
		0.834852,
		0.855993,
		0.875423,
		0.893159,
		0.909232,
		0.923684,
		0.936566,
		0.947941,
		0.957881,
		0.966466,
		0.973783,
		0.979927,
		0.984999,
		0.989102,
		0.992346,
		0.994838,
		0.99669,
		0.998008,
		0.998897,
		0.999454,
		0.00053,
		0.006254,
		0.017685,
		0.034144,
		0.055,
		0.079662,
		0.107582,
		0.13825,
		0.171193,
		0.205974,
		0.242186,
		0.279458,
		0.317445,
		0.355834,
		0.394338,
		0.432694,
		0.470665,
		0.508038,
		0.544621,
		0.580244,
		0.614757,
		0.648029,
		0.679948,
		0.710418,
		0.739362,
		0.766717,
		0.792437,
		0.816488,
		0.838854,
		0.859528,
		0.878519,
		0.895847,
		0.911542,
		0.925648,
		0.938217,
		0.949311,
		0.959001,
		0.967366,
		0.974494,
		0.980477,
		0.985413,
		0.989406,
		0.992561,
		0.994985,
		0.996784,
		0.998065,
		0.002059,
		0.010032,
		0.023363,
		0.041405,
		0.063557,
		0.089257,
		0.117983,
		0.149248,
		0.182603,
		0.217631,
		0.253947,
		0.291195,
		0.32905,
		0.367211,
		0.405406,
		0.443386,
		0.480925,
		0.517822,
		0.553895,
		0.588982,
		0.622942,
		0.655651,
		0.687003,
		0.71691,
		0.745299,
		0.772112,
		0.797307,
		0.820854,
		0.842738,
		0.862958,
		0.881522,
		0.898452,
		0.913781,
		0.927551,
		0.939815,
		0.950636,
		0.960083,
		0.968237,
		0.975181,
		0.981008,
		0.985814,
		0.9897,
		0.992769,
		0.995126,
		0.00013,
		0.004495,
		0.014547,
		0.029626,
		0.049119,
		0.072452,
		0.099088,
		0.128533,
		0.160321,
		0.194026,
		0.229249,
		0.265625,
		0.302814,
		0.340507,
		0.378419,
		0.416289,
		0.453882,
		0.490984,
		0.527402,
		0.562964,
		0.597518,
		0.63093,
		0.663083,
		0.693878,
		0.723232,
		0.751077,
		0.777359,
		0.80204,
		0.825095,
		0.84651,
		0.866286,
		0.884434,
		0.900978,
		0.91595,
		0.929393,
		0.941362,
		0.951918,
		0.961131,
		0.969079,
		0.975845,
		0.981521,
		0.986201,
		0.989983,
		0.001133,
		0.007758,
		0.019728,
		0.036415,
		0.057235,
		0.08164,
		0.109121,
		0.139203,
		0.171447,
		0.205443,
		0.240813,
		0.277209,
		0.314308,
		0.351814,
		0.389456,
		0.426988,
		0.464184,
		0.500843,
		0.53678,
		0.571833,
		0.605858,
		0.638727,
		0.670332,
		0.700578,
		0.729388,
		0.756699,
		0.782462,
		0.806641,
		0.829215,
		0.850172,
		0.869516,
		0.88726,
		0.903427,
		0.918052,
		0.931179,
		0.942861,
		0.95316,
		0.962145,
		0.969893,
		0.976488,
		0.982017,
		0.003055,
		0.011771,
		0.02551,
		0.043674,
		0.065706,
		0.091084,
		0.119322,
		0.149968,
		0.182603,
		0.216838,
		0.252311,
		0.28869,
		0.325668,
		0.362964,
		0.40032,
		0.437501,
		0.474293,
		0.510503,
		0.545959,
		0.580504,
		0.614003,
		0.646337,
		0.6774,
		0.707107,
		0.735383,
		0.762171,
		0.787425,
		0.811114,
		0.833218,
		0.853729,
		0.872652,
		0.890001,
		0.905802,
		0.92009,
		0.932909,
		0.944313,
		0.954362,
		0.963126,
		0.970681,
		0.000492,
		0.005814,
		0.016464,
		0.031834,
		0.051353,
		0.07449,
		0.100747,
		0.129662,
		0.160804,
		0.193772,
		0.228195,
		0.26373,
		0.300059,
		0.336891,
		0.373956,
		0.41101,
		0.447829,
		0.48421,
		0.519969,
		0.554942,
		0.588982,
		0.62196,
		0.653763,
		0.684294,
		0.713469,
		0.741222,
		0.767497,
		0.792253,
		0.815463,
		0.837108,
		0.857183,
		0.875696,
		0.892661,
		0.908106,
		0.922066,
		0.934586,
		0.945719,
		0.955526,
		0.001913,
		0.009337,
		0.021774,
		0.038645,
		0.059405,
		0.083548,
		0.110599,
		0.140116,
		0.171689,
		0.204936,
		0.239503,
		0.275061,
		0.31131,
		0.34797,
		0.384786,
		0.421525,
		0.457972,
		0.493936,
		0.529241,
		0.563732,
		0.59727,
		0.629732,
		0.661012,
		0.691017,
		0.71967,
		0.746909,
		0.772681,
		0.796951,
		0.819691,
		0.840888,
		0.860539,
		0.878652,
		0.895243,
		0.910341,
		0.923982,
		0.936212,
		0.00012,
		0.004182,
		0.013553,
		0.027641,
		0.045892,
		0.067787,
		0.092843,
		0.120607,
		0.150658,
		0.182603,
		0.216079,
		0.250747,
		0.286296,
		0.322436,
		0.358903,
		0.395453,
		0.431863,
		0.467931,
		0.503473,
		0.538323,
		0.572334,
		0.605372,
		0.637323,
		0.668086,
		0.697574,
		0.725714,
		0.752448,
		0.777728,
		0.801521,
		0.823803,
		0.844563,
		0.863799,
		0.881522,
		0.897749,
		0.91251,
		0.001053,
		0.007225,
		0.018399,
		0.034009,
		0.053528,
		0.076461,
		0.102344,
		0.130747,
		0.161266,
		0.193529,
		0.227188,
		0.261919,
		0.297425,
		0.333432,
		0.369686,
		0.405954,
		0.442026,
		0.477707,
		0.512824,
		0.547218,
		0.580749,
		0.613292,
		0.644738,
		0.67499,
		0.703969,
		0.731605,
		0.757843,
		0.782641,
		0.805968,
		0.827802,
		0.848135,
		0.866967,
		0.88431,
		0.002844,
		0.010974,
		0.023816,
		0.04083,
		0.061512,
		0.085389,
		0.112019,
		0.140991,
		0.171921,
		0.204451,
		0.238249,
		0.273008,
		0.308444,
		0.344293,
		0.380315,
		0.416289,
		0.452013,
		0.487302,
		0.52199,
		0.555928,
		0.588982,
		0.621034,
		0.651979,
		0.681729,
		0.710206,
		0.737347,
		0.7631,
		0.787425,
		0.810295,
		0.831692,
		0.851607,
		0.000458,
		0.005419,
		0.015366,
		0.02975,
		0.048056,
		0.069803,
		0.094539,
		0.121842,
		0.151319,
		0.182603,
		0.215353,
		0.249252,
		0.284006,
		0.319344,
		0.355015,
		0.39079,
		0.426457,
		0.461824,
		0.496716,
		0.530974,
		0.564456,
		0.597036,
		0.628601,
		0.659052,
		0.688306,
		0.71629,
		0.742944,
		0.76822,
		0.792083,
		0.814507,
		0.001783,
		0.008711,
		0.020342,
		0.03615,
		0.055645,
		0.078366,
		0.103881,
		0.131787,
		0.16171,
		0.193297,
		0.226223,
		0.260186,
		0.294906,
		0.330122,
		0.365596,
		0.401108,
		0.436458,
		0.471461,
		0.505952,
		0.539778,
		0.572806,
		0.604914,
		0.635997,
		0.66596,
		0.694725,
		0.722224,
		0.7484,
		0.77321,
		0.000112,
		0.003901,
		0.012657,
		0.025848,
		0.042971,
		0.063557,
		0.087167,
		0.113386,
		0.141832,
		0.172143,
		0.203986,
		0.237049,
		0.271043,
		0.3057,
		0.340772,
		0.376031,
		0.411269,
		0.446292,
		0.480925,
		0.515011,
		0.548406,
		0.58098,
		0.612621,
		0.643226,
		0.672708,
		0.700991,
		0.728012,
		0.000982,
		0.006746,
		0.0172,
		0.031834,
		0.050169,
		0.071756,
		0.096175,
		0.12303,
		0.151954,
		0.182603,
		0.214658,
		0.247821,
		0.281814,
		0.316383,
		0.351291,
		0.38632,
		0.42127,
		0.455958,
		0.490217,
		0.523896,
		0.556859,
		0.588982,
		0.620158,
		0.650291,
		0.679298,
		0.002655,
		0.010256,
		0.022285,
		0.038254,
		0.057705,
		0.080209,
		0.105362,
		0.132788,
		0.162135,
		0.193074,
		0.2253,
		0.258527,
		0.292492,
		0.32695,
		0.361676,
		0.39646,
		0.431113,
		0.465459,
		0.499339,
		0.532609,
		0.56514,
		0.596815,
		0.62753,
		0.000428,
		0.005063,
		0.014374,
		0.027864,
		0.045066,
		0.065543,
		0.088884,
		0.114702,
		0.142639,
		0.172357,
		0.203541,
		0.2359,
		0.269161,
		0.303071,
		0.337397,
		0.371923,
		0.40645,
		0.440796,
		0.474793,
		0.508291,
		0.541152,
		0.573252,
		0.001665,
		0.008146,
		0.019046,
		0.033889,
		0.05223,
		0.073649,
		0.097753,
		0.124172,
		0.152563,
		0.182603,
		0.213992,
		0.246449,
		0.279714,
		0.313545,
		0.34772,
		0.382031,
		0.416289,
		0.45032,
		0.483964,
		0.517077,
		0.000105,
		0.003647,
		0.011847,
		0.024223,
		0.04032,
		0.05971,
		0.081992,
		0.106789,
		0.13375,
		0.162543,
		0.19286,
		0.224414,
		0.256937,
		0.290179,
		0.32391,
		0.357915,
		0.391998,
		0.425977,
		0.459685,
		0.000918,
		0.006313,
		0.016114,
		0.02986,
		0.047116,
		0.067471,
		0.090543,
		0.11597,
		0.143415,
		0.172561,
		0.203114,
		0.234798,
		0.267356,
		0.300551,
		0.33416,
		0.36798,
		0.401822,
		0.002483,
		0.009606,
		0.020897,
		0.035914,
		0.05424,
		0.075484,
		0.099276,
		0.125272,
		0.153149,
		0.182603,
		0.213353,
		0.245133,
		0.2777,
		0.310823,
		0.344293,
		0.0004,
		0.004741,
		0.013475,
		0.026151,
		0.042346,
		0.06166,
		0.083718,
		0.108167,
		0.134676,
		0.162935,
		0.192655,
		0.223565,
		0.255412,
		0.28796,
		0.001559,
		0.007634,
		0.01787,
		0.031834,
		0.049119,
		0.069344,
		0.092147,
		0.117192,
		0.144161,
		0.172758,
		0.202704,
		0.233741,
		9.8e-005,
		0.003417,
		0.011113,
		0.022747,
		0.037906,
		0.0562,
		0.077262,
		0.100747,
		0.126332,
		0.153712,
		0.182603,
		0.00086,
		0.00592,
		0.015128,
		0.028064,
		0.044332,
		0.063557,
		0.085389,
		0.109496,
		0.135567,
		0.002328,
		0.009016,
		0.019635,
		0.033782,
		0.051077,
		0.071162,
		0.093699,
		0.000375,
		0.004448,
		0.012657,
		0.024591,
		0.039864,
		0.058111,
		0.001462,
		0.00717,
		0.0168,
		0.029959,
		9.2e-005,
		0.003208,
		0.010445,
		0.000807,
		0.005562,
		0.001428,
		0.021402,
		0.01237,
		0.005649,
		0.001451,
		0.046277,
		0.033031,
		0.021727,
		0.01256,
		0.005737,
		0.001474,
		0.078987,
		0.062177,
		0.046964,
		0.033529,
		0.022059,
		0.012755,
		0.005827,
		0.001497,
		0.118371,
		0.098612,
		0.080132,
		0.063092,
		0.047667,
		0.034038,
		0.022399,
		0.012955,
		0.00592,
		0.001521,
		0.163312,
		0.141187,
		0.120042,
		0.100028,
		0.081301,
		0.064028,
		0.048385,
		0.034559,
		0.022747,
		0.013159,
		0.006014,
		0.001546,
		0.212739,
		0.188793,
		0.165555,
		0.14316,
		0.121749,
		0.101474,
		0.082496,
		0.064984,
		0.049119,
		0.035092,
		0.023103,
		0.013368,
		0.006111,
		0.001571,
		0.265625,
		0.24037,
		0.215573,
		0.191355,
		0.167843,
		0.145174,
		0.123492,
		0.102952,
		0.083718,
		0.065963,
		0.049871,
		0.035637,
		0.023468,
		0.013582,
		0.006211,
		0.001597,
		0.320992,
		0.294906,
		0.269049,
		0.24353,
		0.218463,
		0.193969,
		0.170178,
		0.14723,
		0.125272,
		0.104462,
		0.084966,
		0.066963,
		0.050639,
		0.036195,
		0.023841,
		0.013802,
		0.006313,
		0.001624,
		0.377913,
		0.351437,
		0.324985,
		0.298652,
		0.272538,
		0.246752,
		0.221409,
		0.196635,
		0.172561,
		0.14933,
		0.127091,
		0.106005,
		0.086243,
		0.067986,
		0.051425,
		0.036766,
		0.024223,
		0.014026,
		0.006417,
		0.001651,
		0.435513,
		0.409054,
		0.382435,
		0.355737,
		0.32905,
		0.302467,
		0.276092,
		0.250035,
		0.224414,
		0.199356,
		0.174994,
		0.151474,
		0.128949,
		0.107582,
		0.087549,
		0.069032,
		0.05223,
		0.037351,
		0.024615,
		0.014256,
		0.006524,
		0.001679,
		0.492972,
		0.466901,
		0.440506,
		0.413858,
		0.387032,
		0.360112,
		0.333186,
		0.306352,
		0.279714,
		0.253383,
		0.227479,
		0.202132,
		0.177478,
		0.153664,
		0.130847,
		0.109195,
		0.088884,
		0.070103,
		0.053054,
		0.03795,
		0.025016,
		0.014492,
		0.006633,
		0.001708,
		0.549528,
		0.52418,
		0.498366,
		0.472143,
		0.445578,
		0.41874,
		0.391707,
		0.364563,
		0.337397,
		0.310309,
		0.283404,
		0.256796,
		0.230606,
		0.204965,
		0.180014,
		0.155901,
		0.132788,
		0.110843,
		0.090249,
		0.071199,
		0.053897,
		0.038563,
		0.025427,
		0.014734,
		0.006746,
		0.001737,
		0.604481,
		0.580154,
		0.555238,
		0.529781,
		0.503836,
		0.477463,
		0.450728,
		0.423701,
		0.39646,
		0.369091,
		0.341684,
		0.31434,
		0.287165,
		0.260276,
		0.233795,
		0.207857,
		0.182603,
		0.158187,
		0.134771,
		0.112529,
		0.091646,
		0.07232,
		0.054761,
		0.039191,
		0.025848,
		0.014982,
		0.006861,
		0.001767,
		0.657196,
		0.63415,
		0.610413,
		0.586022,
		0.561021,
		0.535457,
		0.509384,
		0.482863,
		0.455958,
		0.428742,
		0.401293,
		0.373698,
		0.346048,
		0.318445,
		0.290998,
		0.263824,
		0.237049,
		0.210809,
		0.185248,
		0.160523,
		0.136799,
		0.114253,
		0.093076,
		0.073468,
		0.055645,
		0.039835,
		0.026279,
		0.015236,
		0.006979,
		0.001798,
		0.707107,
		0.685564,
		0.663248,
		0.640185,
		0.61641,
		0.59196,
		0.566876,
		0.541209,
		0.515011,
		0.488343,
		0.46127,
		0.433865,
		0.406208,
		0.378386,
		0.350491,
		0.322628,
		0.294906,
		0.267443,
		0.24037,
		0.213823,
		0.18795,
		0.16291,
		0.138872,
		0.116017,
		0.094539,
		0.074644,
		0.056551,
		0.040494,
		0.026722,
		0.015497,
		0.007101,
		0.00183,
		0.753719,
		0.733863,
		0.71317,
		0.691659,
		0.669354,
		0.646281,
		0.622473,
		0.597967,
		0.572806,
		0.547038,
		0.520718,
		0.493904,
		0.466665,
		0.439072,
		0.411206,
		0.383156,
		0.355015,
		0.326889,
		0.298888,
		0.271135,
		0.243759,
		0.2169,
		0.19071,
		0.165349,
		0.140991,
		0.117821,
		0.096037,
		0.075848,
		0.057479,
		0.04117,
		0.027175,
		0.015765,
		0.007225,
		0.001863,
		0.796619,
		0.778592,
		0.759685,
		0.739908,
		0.719275,
		0.697803,
		0.675515,
		0.652437,
		0.628601,
		0.604044,
		0.57881,
		0.552945,
		0.526505,
		0.499548,
		0.472143,
		0.444363,
		0.416289,
		0.38801,
		0.359622,
		0.331231,
		0.302949,
		0.2749,
		0.247217,
		0.220042,
		0.193529,
		0.167843,
		0.14316,
		0.119668,
		0.09757,
		0.077081,
		0.05843,
		0.041863,
		0.027641,
		0.016039,
		0.007353,
		0.001896,
		0.835475,
		0.819379,
		0.802379,
		0.784477,
		0.765677,
		0.745987,
		0.72542,
		0.703994,
		0.681729,
		0.658652,
		0.634794,
		0.610192,
		0.584888,
		0.55893,
		0.532373,
		0.505276,
		0.477707,
		0.449741,
		0.421459,
		0.39295,
		0.364313,
		0.335654,
		0.307089,
		0.278741,
		0.250747,
		0.223252,
		0.196411,
		0.170393,
		0.145377,
		0.121558,
		0.09914,
		0.078344,
		0.059405,
		0.042574,
		0.028118,
		0.016321,
		0.007484,
		0.001931,
		0.870045,
		0.855939,
		0.840925,
		0.824995,
		0.808146,
		0.790377,
		0.771692,
		0.752097,
		0.731605,
		0.710231,
		0.687996,
		0.664927,
		0.641052,
		0.61641,
		0.591042,
		0.564995,
		0.538323,
		0.511089,
		0.483358,
		0.455206,
		0.426716,
		0.397978,
		0.369091,
		0.340162,
		0.31131,
		0.282661,
		0.254351,
		0.22653,
		0.199356,
		0.173,
		0.147646,
		0.123492,
		0.100747,
		0.079638,
		0.060404,
		0.043303,
		0.028608,
		0.01661,
		0.007619,
		0.001966,
		0.900183,
		0.888082,
		0.875088,
		0.861185,
		0.846361,
		0.830607,
		0.813918,
		0.796291,
		0.777728,
		0.758237,
		0.737827,
		0.716513,
		0.694316,
		0.67126,
		0.647376,
		0.622699,
		0.59727,
		0.571138,
		0.544357,
		0.516987,
		0.489096,
		0.460761,
		0.432063,
		0.403095,
		0.373956,
		0.344756,
		0.315615,
		0.28666,
		0.258031,
		0.229879,
		0.202366,
		0.175666,
		0.149968,
		0.125472,
		0.102394,
		0.080964,
		0.061429,
		0.044051,
		0.02911,
		0.016907,
		0.007758,
		0.002002,
		0.925841,
		0.915718,
		0.904734,
		0.892867,
		0.880097,
		0.866407,
		0.851782,
		0.836212,
		0.819691,
		0.802214,
		0.783784,
		0.764404,
		0.744084,
		0.722839,
		0.700686,
		0.677651,
		0.653763,
		0.629057,
		0.603574,
		0.577362,
		0.550474,
		0.522972,
		0.494924,
		0.466406,
		0.437501,
		0.408303,
		0.378911,
		0.349439,
		0.320005,
		0.290741,
		0.261788,
		0.2333,
		0.205443,
		0.178394,
		0.152345,
		0.1275,
		0.104082,
		0.082324,
		0.06248,
		0.044818,
		0.029626,
		0.017212,
		0.0079,
		0.00204,
		0.947082,
		0.938862,
		0.929834,
		0.919966,
		0.909232,
		0.897607,
		0.88507,
		0.8716,
		0.857183,
		0.841807,
		0.825463,
		0.808146,
		0.789856,
		0.770596,
		0.750376,
		0.729206,
		0.707107,
		0.6841,
		0.660215,
		0.635486,
		0.609954,
		0.583667,
		0.556677,
		0.529046,
		0.500843,
		0.472143,
		0.443032,
		0.413603,
		0.383959,
		0.354211,
		0.324482,
		0.294906,
		0.265625,
		0.236797,
		0.208589,
		0.181184,
		0.154777,
		0.129577,
		0.10581,
		0.083718,
		0.063557,
		0.045605,
		0.030156,
		0.017525,
		0.008046,
		0.002078,
		0.964076,
		0.957642,
		0.950468,
		0.942516,
		0.933754,
		0.924149,
		0.913673,
		0.902298,
		0.890001,
		0.876762,
		0.862562,
		0.847388,
		0.83123,
		0.814083,
		0.795942,
		0.776812,
		0.756699,
		0.735615,
		0.713576,
		0.690605,
		0.66673,
		0.641985,
		0.61641,
		0.590052,
		0.562964,
		0.535208,
		0.506853,
		0.477974,
		0.448658,
		0.418999,
		0.3891,
		0.359076,
		0.32905,
		0.299157,
		0.269544,
		0.24037,
		0.211807,
		0.184039,
		0.157267,
		0.131705,
		0.107582,
		0.085147,
		0.064663,
		0.046414,
		0.0307,
		0.017847,
		0.008197,
		0.002117,
		0.977109,
		0.972299,
		0.966832,
		0.960668,
		0.953766,
		0.946088,
		0.937598,
		0.928264,
		0.918052,
		0.906935,
		0.894888,
		0.881887,
		0.867913,
		0.852952,
		0.83699,
		0.820021,
		0.80204,
		0.783049,
		0.763052,
		0.742062,
		0.720092,
		0.697165,
		0.673308,
		0.648554,
		0.622942,
		0.596519,
		0.569338,
		0.541461,
		0.512956,
		0.483901,
		0.454381,
		0.424491,
		0.394338,
		0.364035,
		0.333709,
		0.303496,
		0.273547,
		0.244022,
		0.215097,
		0.186961,
		0.159817,
		0.133884,
		0.109399,
		0.086613,
		0.065798,
		0.047243,
		0.031259,
		0.018178,
		0.008351,
		0.002158,
		0.986574,
		0.983187,
		0.97924,
		0.974689,
		0.969489,
		0.963598,
		0.956972,
		0.949573,
		0.941362,
		0.932304,
		0.922365,
		0.911514,
		0.899725,
		0.886971,
		0.873234,
		0.858494,
		0.842738,
		0.825958,
		0.808146,
		0.789303,
		0.769433,
		0.748546,
		0.726654,
		0.70378,
		0.679948,
		0.655192,
		0.629549,
		0.603067,
		0.575799,
		0.547804,
		0.519153,
		0.489923,
		0.460201,
		0.430082,
		0.399673,
		0.369091,
		0.338462,
		0.307926,
		0.277636,
		0.247755,
		0.218463,
		0.189951,
		0.162428,
		0.136118,
		0.111261,
		0.088117,
		0.066963,
		0.048096,
		0.031834,
		0.018518,
		0.00851,
		0.0022,
		0.992969,
		0.990768,
		0.988116,
		0.984964,
		0.981267,
		0.976975,
		0.972044,
		0.966428,
		0.960083,
		0.952968,
		0.945041,
		0.936266,
		0.926607,
		0.916031,
		0.904508,
		0.892012,
		0.878519,
		0.864011,
		0.848471,
		0.831889,
		0.814257,
		0.795573,
		0.77584,
		0.755064,
		0.73326,
		0.710447,
		0.686649,
		0.661898,
		0.636233,
		0.609698,
		0.582346,
		0.55424,
		0.525446,
		0.496045,
		0.466121,
		0.435774,
		0.405109,
		0.374245,
		0.343311,
		0.31245,
		0.281814,
		0.251572,
		0.221906,
		0.193012,
		0.165103,
		0.138407,
		0.113171,
		0.08966,
		0.068159,
		0.048971,
		0.032424,
		0.018867,
		0.008674,
		0.002243,
		0.996876,
		0.995601,
		0.993987,
		0.991988,
		0.989554,
		0.986637,
		0.983187,
		0.979155,
		0.974494,
		0.969156,
		0.963095,
		0.956268,
		0.948631,
		0.940146,
		0.930774,
		0.92048,
		0.909232,
		0.897002,
		0.883765,
		0.869498,
		0.854185,
		0.837812,
		0.82037,
		0.801855,
		0.782268,
		0.761615,
		0.739908,
		0.717165,
		0.69341,
		0.668673,
		0.642991,
		0.61641,
		0.588982,
		0.560768,
		0.531836,
		0.502265,
		0.472143,
		0.441568,
		0.410647,
		0.379501,
		0.34826,
		0.317068,
		0.286083,
		0.255475,
		0.225429,
		0.196146,
		0.167843,
		0.140754,
		0.11513,
		0.091244,
		0.069387,
		0.049871,
		0.033031,
		0.019227,
		0.008842,
		0.002287,
		0.998929,
		0.99831,
		0.997465,
		0.996348,
		0.994912,
		0.993109,
		0.99089,
		0.988203,
		0.984999,
		0.981226,
		0.976835,
		0.971777,
		0.966004,
		0.959469,
		0.952128,
		0.943938,
		0.93486,
		0.924857,
		0.913893,
		0.901939,
		0.888967,
		0.874952,
		0.859876,
		0.843722,
		0.826481,
		0.808146,
		0.788716,
		0.768196,
		0.746597,
		0.723934,
		0.70023,
		0.675515,
		0.649825,
		0.623205,
		0.595706,
		0.567389,
		0.538323,
		0.508587,
		0.478268,
		0.447466,
		0.416289,
		0.384859,
		0.353309,
		0.321785,
		0.290446,
		0.259467,
		0.229035,
		0.199356,
		0.170651,
		0.14316,
		0.11714,
		0.09287,
		0.070648,
		0.050795,
		0.033655,
		0.019597,
		0.009016,
		0.002333,
		0.999771,
		0.999551,
		0.999205,
		0.998695,
		0.997978,
		0.997011,
		0.995745,
		0.994133,
		0.992122,
		0.989663,
		0.986701,
		0.983187,
		0.979066,
		0.974289,
		0.968806,
		0.962567,
		0.955526,
		0.947639,
		0.938862,
		0.929157,
		0.918486,
		0.906817,
		0.894119,
		0.880367,
		0.865538,
		0.849616,
		0.832587,
		0.814442,
		0.795181,
		0.774805,
		0.753323,
		0.73075,
		0.707107,
		0.682423,
		0.656733,
		0.630081,
		0.602518,
		0.574105,
		0.54491,
		0.515011,
		0.484498,
		0.453471,
		0.422038,
		0.390324,
		0.358463,
		0.326602,
		0.294906,
		0.263549,
		0.232725,
		0.202643,
		0.173528,
		0.145627,
		0.119202,
		0.094539,
		0.071944,
		0.051745,
		0.034297,
		0.019978,
		0.009194,
		0.00238,
		0.999985,
		0.999945,
		0.999859,
		0.999698,
		0.999429,
		0.999017,
		0.99842,
		0.997593,
		0.99649,
		0.99506,
		0.993253,
		0.991015,
		0.988294,
		0.985035,
		0.981184,
		0.976689,
		0.971498,
		0.965558,
		0.958822,
		0.951243,
		0.942774,
		0.933375,
		0.923006,
		0.911631,
		0.899218,
		0.885739,
		0.871169,
		0.855488,
		0.838682,
		0.820741,
		0.801659,
		0.781437,
		0.760084,
		0.737611,
		0.71404,
		0.689396,
		0.663715,
		0.63704,
		0.60942,
		0.580916,
		0.551596,
		0.521539,
		0.490835,
		0.459584,
		0.427896,
		0.395897,
		0.363722,
		0.331523,
		0.299464,
		0.267725,
		0.236502,
		0.20601,
		0.176478,
		0.148158,
		0.121319,
		0.096253,
		0.073275,
		0.052722,
		0.034957,
		0.02037,
		0.009378,
		0.002428,
		1,
		1,
		0.999995,
		0.999974,
		0.99992,
		0.999807,
		0.999606,
		0.999281,
		0.998793,
		0.998097,
		0.997147,
		0.995893,
		0.994282,
		0.992261,
		0.989775,
		0.986769,
		0.983187,
		0.978974,
		0.974075,
		0.968439,
		0.962012,
		0.954745,
		0.946591,
		0.937506,
		0.927446,
		0.916375,
		0.904257,
		0.891062,
		0.876762,
		0.861335,
		0.844764,
		0.827036,
		0.808146,
		0.788092,
		0.766878,
		0.744517,
		0.721027,
		0.696434,
		0.670771,
		0.64408,
		0.61641,
		0.587822,
		0.558383,
		0.528173,
		0.497281,
		0.465807,
		0.433865,
		0.40158,
		0.369091,
		0.336549,
		0.304124,
		0.271997,
		0.24037,
		0.20946,
		0.179502,
		0.150754,
		0.123492,
		0.098015,
		0.074644,
		0.053727,
		0.035637,
		0.020774,
		0.009567,
		0.002478,
		0.999985,
		0.999998,
		1,
		1,
		0.999999,
		0.99999,
		0.999959,
		0.999886,
		0.999741,
		0.999493,
		0.999103,
		0.998528,
		0.997721,
		0.996633,
		0.995211,
		0.9934,
		0.991144,
		0.988388,
		0.985072,
		0.98114,
		0.976536,
		0.971204,
		0.96509,
		0.958141,
		0.950308,
		0.941544,
		0.931803,
		0.921045,
		0.909232,
		0.896331,
		0.882312,
		0.867151,
		0.850827,
		0.833326,
		0.814639,
		0.794763,
		0.773702,
		0.751464,
		0.728066,
		0.703534,
		0.677898,
		0.651201,
		0.62349,
		0.594824,
		0.565272,
		0.534913,
		0.503836,
		0.472143,
		0.439948,
		0.407377,
		0.37457,
		0.341684,
		0.308888,
		0.276369,
		0.24433,
		0.212995,
		0.182603,
		0.153419,
		0.125723,
		0.099824,
		0.076051,
		0.054761,
		0.036336,
		0.021189,
		0.009762,
		0.00253,
		0.999771,
		0.999896,
		0.999962,
		0.99999,
		0.999999,
		1,
		1,
		0.999997,
		0.999982,
		0.999938,
		0.99984,
		0.999658,
		0.999355,
		0.998889,
		0.998216,
		0.997284,
		0.996042,
		0.994434,
		0.992403,
		0.989891,
		0.986838,
		0.983187,
		0.978877,
		0.973851,
		0.968053,
		0.961427,
		0.953921,
		0.945484,
		0.93607,
		0.925634,
		0.914137,
		0.901541,
		0.887815,
		0.872931,
		0.856866,
		0.839605,
		0.821134,
		0.80145,
		0.780552,
		0.758449,
		0.735156,
		0.710695,
		0.685097,
		0.658402,
		0.630658,
		0.601923,
		0.572264,
		0.541761,
		0.510503,
		0.478594,
		0.446146,
		0.413289,
		0.380165,
		0.34693,
		0.31376,
		0.280842,
		0.248386,
		0.216617,
		0.185784,
		0.156153,
		0.128015,
		0.101684,
		0.077498,
		0.055825,
		0.037057,
		0.021618,
		0.009964,
		0.002583,
		0.998929,
		0.999343,
		0.999629,
		0.999813,
		0.99992,
		0.999973,
		0.999995,
		1,
		1,
		1,
		0.999994,
		0.999971,
		0.999909,
		0.999782,
		0.999554,
		0.999187,
		0.998635,
		0.99785,
		0.996778,
		0.995364,
		0.993551,
		0.991278,
		0.988485,
		0.985111,
		0.981094,
		0.976376,
		0.970896,
		0.964597,
		0.957423,
		0.949321,
		0.940242,
		0.930138,
		0.918966,
		0.906686,
		0.893264,
		0.87867,
		0.862877,
		0.845868,
		0.827626,
		0.808146,
		0.787425,
		0.76547,
		0.742293,
		0.717915,
		0.692366,
		0.665684,
		0.637915,
		0.609118,
		0.579359,
		0.548718,
		0.517284,
		0.48516,
		0.452462,
		0.419319,
		0.385876,
		0.352291,
		0.318742,
		0.285421,
		0.25254,
		0.220331,
		0.189047,
		0.15896,
		0.130369,
		0.103595,
		0.078987,
		0.05692,
		0.037798,
		0.022059,
		0.010171,
		0.002638,
		0.996876,
		0.997796,
		0.998511,
		0.999046,
		0.999429,
		0.999688,
		0.99985,
		0.99994,
		0.999982,
		0.999997,
		1,
		1,
		0.999999,
		0.999989,
		0.999954,
		0.99987,
		0.999707,
		0.999426,
		0.998984,
		0.998334,
		0.997422,
		0.996194,
		0.99459,
		0.992549,
		0.990011,
		0.986911,
		0.983187,
		0.978775,
		0.973615,
		0.967646,
		0.96081,
		0.95305,
		0.944313,
		0.934549,
		0.923712,
		0.91176,
		0.898654,
		0.884362,
		0.868855,
		0.85211,
		0.834112,
		0.814849,
		0.794318,
		0.772523,
		0.749475,
		0.725192,
		0.699703,
		0.673044,
		0.645261,
		0.61641,
		0.586559,
		0.555785,
		0.52418,
		0.491846,
		0.458899,
		0.425471,
		0.391707,
		0.357769,
		0.323837,
		0.290107,
		0.256796,
		0.224139,
		0.192394,
		0.161842,
		0.132788,
		0.105561,
		0.080519,
		0.058047,
		0.038563,
		0.022514,
		0.010385,
		0.002694,
		0.992969,
		0.994617,
		0.995985,
		0.997097,
		0.997978,
		0.998655,
		0.999156,
		0.999508,
		0.999741,
		0.999882,
		0.999957,
		0.999989,
		0.999999,
		1,
		1,
		0.999997,
		0.99998,
		0.99993,
		0.999819,
		0.999612,
		0.999268,
		0.998741,
		0.997978,
		0.996925,
		0.995521,
		0.993706,
		0.991416,
		0.988586,
		0.985151,
		0.981047,
		0.976208,
		0.970571,
		0.964076,
		0.956664,
		0.948277,
		0.938862,
		0.928371,
		0.916757,
		0.903979,
		0.890001,
		0.874792,
		0.858326,
		0.840584,
		0.821553,
		0.801227,
		0.779606,
		0.756699,
		0.732524,
		0.707107,
		0.680482,
		0.652694,
		0.623799,
		0.593863,
		0.562964,
		0.531193,
		0.498652,
		0.465459,
		0.431746,
		0.397661,
		0.363368,
		0.32905,
		0.294906,
		0.261156,
		0.228043,
		0.19583,
		0.164802,
		0.135274,
		0.107582,
		0.082095,
		0.059208,
		0.03935,
		0.022984,
		0.010606,
		0.002753,
		0.986574,
		0.989154,
		0.99139,
		0.993303,
		0.994912,
		0.99624,
		0.997312,
		0.998153,
		0.998793,
		0.999259,
		0.999581,
		0.999788,
		0.999909,
		0.99997,
		0.999994,
		1,
		1,
		1,
		0.999993,
		0.999967,
		0.999897,
		0.999752,
		0.999493,
		0.999076,
		0.998451,
		0.997561,
		0.996348,
		0.994749,
		0.9927,
		0.990135,
		0.986986,
		0.983187,
		0.978669,
		0.973368,
		0.967218,
		0.960158,
		0.952128,
		0.943071,
		0.932935,
		0.92167,
		0.909232,
		0.895582,
		0.880684,
		0.864511,
		0.84704,
		0.828254,
		0.808146,
		0.786713,
		0.763962,
		0.739908,
		0.714575,
		0.687996,
		0.660215,
		0.631286,
		0.601273,
		0.570256,
		0.538323,
		0.50558,
		0.472143,
		0.438147,
		0.403741,
		0.369091,
		0.334382,
		0.299819,
		0.265625,
		0.232048,
		0.199356,
		0.167843,
		0.137829,
		0.109662,
		0.083718,
		0.060404,
		0.040162,
		0.023468,
		0.010834,
		0.002813,
		0.977109,
		0.980795,
		0.984089,
		0.987004,
		0.989554,
		0.991758,
		0.993633,
		0.995203,
		0.99649,
		0.99752,
		0.998321,
		0.998923,
		0.999355,
		0.999647,
		0.99983,
		0.999932,
		0.99998,
		0.999997,
		1,
		1,
		0.999999,
		0.999987,
		0.999947,
		0.999852,
		0.999666,
		0.999346,
		0.998845,
		0.998106,
		0.997073,
		0.995681,
		0.993865,
		0.991558,
		0.988691,
		0.985193,
		0.980996,
		0.976031,
		0.97023,
		0.963527,
		0.955861,
		0.94717,
		0.937398,
		0.926493,
		0.914407,
		0.901097,
		0.886525,
		0.870658,
		0.853472,
		0.834948,
		0.815072,
		0.793842,
		0.771261,
		0.747342,
		0.722107,
		0.695586,
		0.667822,
		0.638868,
		0.608789,
		0.577661,
		0.545574,
		0.512632,
		0.478955,
		0.444677,
		0.409949,
		0.37494,
		0.339837,
		0.30485,
		0.270205,
		0.236155,
		0.202976,
		0.170967,
		0.140457,
		0.111802,
		0.085389,
		0.061637,
		0.041,
		0.023968,
		0.011069,
		0.002875,
		0.964076,
		0.969004,
		0.973507,
		0.977591,
		0.981267,
		0.984543,
		0.987433,
		0.989954,
		0.992122,
		0.993959,
		0.995488,
		0.996733,
		0.997721,
		0.998482,
		0.999046,
		0.999443,
		0.999707,
		0.999866,
		0.999951,
		0.999988,
		0.999999,
		1,
		1,
		0.999996,
		0.999977,
		0.99992,
		0.999793,
		0.999557,
		0.999166,
		0.998566,
		0.9977,
		0.996504,
		0.994912,
		0.992856,
		0.990264,
		0.987065,
		0.983187,
		0.978558,
		0.973107,
		0.966766,
		0.959469,
		0.951151,
		0.941754,
		0.931219,
		0.919497,
		0.90654,
		0.892307,
		0.876762,
		0.859876,
		0.841627,
		0.822,
		0.800988,
		0.778592,
		0.754822,
		0.729698,
		0.703249,
		0.675515,
		0.646547,
		0.61641,
		0.58518,
		0.552945,
		0.519811,
		0.485897,
		0.451339,
		0.416289,
		0.380919,
		0.34542,
		0.310002,
		0.2749,
		0.24037,
		0.206693,
		0.174178,
		0.14316,
		0.114005,
		0.08711,
		0.062907,
		0.041863,
		0.024483,
		0.011312,
		0.00294,
		0.947082,
		0.953345,
		0.959167,
		0.964547,
		0.969489,
		0.973997,
		0.978079,
		0.981742,
		0.984999,
		0.987863,
		0.990352,
		0.992484,
		0.994282,
		0.995768,
		0.99697,
		0.997915,
		0.998635,
		0.999161,
		0.999525,
		0.99976,
		0.999897,
		0.999966,
		0.999993,
		1,
		1,
		1,
		0.999992,
		0.999962,
		0.999882,
		0.999716,
		0.999422,
		0.998946,
		0.998234,
		0.997222,
		0.995843,
		0.994029,
		0.991705,
		0.9888,
		0.985237,
		0.980944,
		0.975845,
		0.96987,
		0.962947,
		0.95501,
		0.945995,
		0.935842,
		0.924495,
		0.911904,
		0.898023,
		0.882814,
		0.866244,
		0.848287,
		0.828924,
		0.808146,
		0.785951,
		0.762345,
		0.737347,
		0.710982,
		0.683291,
		0.654322,
		0.624138,
		0.592813,
		0.560439,
		0.527118,
		0.492972,
		0.458135,
		0.422765,
		0.387032,
		0.351133,
		0.315281,
		0.279714,
		0.244695,
		0.210511,
		0.177478,
		0.14594,
		0.116272,
		0.088884,
		0.064217,
		0.042755,
		0.025016,
		0.011564,
		0.003007,
		0.925841,
		0.933492,
		0.940699,
		0.947458,
		0.953766,
		0.959621,
		0.965026,
		0.969981,
		0.974494,
		0.97857,
		0.98222,
		0.985457,
		0.988294,
		0.990749,
		0.992844,
		0.994599,
		0.996042,
		0.997199,
		0.998102,
		0.99878,
		0.999268,
		0.999599,
		0.999806,
		0.999923,
		0.999977,
		0.999996,
		1,
		1,
		0.999998,
		0.999985,
		0.99994,
		0.999831,
		0.999618,
		0.999253,
		0.998681,
		0.997839,
		0.996662,
		0.995079,
		0.993016,
		0.990397,
		0.987147,
		0.983187,
		0.97844,
		0.972832,
		0.966289,
		0.958739,
		0.950115,
		0.940353,
		0.929393,
		0.917181,
		0.903668,
		0.88881,
		0.872571,
		0.854921,
		0.835839,
		0.815311,
		0.793332,
		0.769907,
		0.74505,
		0.718785,
		0.69115,
		0.662191,
		0.631971,
		0.600563,
		0.568056,
		0.534555,
		0.50018,
		0.465069,
		0.429378,
		0.393283,
		0.35698,
		0.320688,
		0.28465,
		0.249134,
		0.214433,
		0.180871,
		0.148801,
		0.118608,
		0.090711,
		0.065569,
		0.043674,
		0.025566,
		0.011823,
		0.003075,
		0.900183,
		0.909232,
		0.91785,
		0.926026,
		0.933754,
		0.941027,
		0.947842,
		0.954194,
		0.960083,
		0.965511,
		0.97048,
		0.974996,
		0.979066,
		0.982702,
		0.985916,
		0.988724,
		0.991144,
		0.993199,
		0.994912,
		0.99631,
		0.997422,
		0.99828,
		0.998917,
		0.999367,
		0.999666,
		0.999847,
		0.999944,
		0.999986,
		0.999998,
		1,
		1,
		0.999996,
		0.999974,
		0.999908,
		0.999763,
		0.999493,
		0.999046,
		0.998361,
		0.997372,
		0.996009,
		0.994196,
		0.991857,
		0.988913,
		0.985283,
		0.980889,
		0.97565,
		0.969489,
		0.962333,
		0.954108,
		0.944746,
		0.934185,
		0.922365,
		0.909232,
		0.89474,
		0.878848,
		0.861523,
		0.842738,
		0.822478,
		0.800733,
		0.777504,
		0.752804,
		0.726654,
		0.699089,
		0.670154,
		0.63991,
		0.608428,
		0.575799,
		0.542124,
		0.507526,
		0.472143,
		0.436133,
		0.399673,
		0.362964,
		0.326228,
		0.289713,
		0.25369,
		0.218463,
		0.18436,
		0.151745,
		0.121013,
		0.092596,
		0.066963,
		0.044624,
		0.026134,
		0.012092,
		0.003147,
		0.870045,
		0.880467,
		0.89048,
		0.900072,
		0.909232,
		0.91795,
		0.926215,
		0.934022,
		0.941362,
		0.948233,
		0.95463,
		0.960553,
		0.966004,
		0.970985,
		0.975504,
		0.979567,
		0.983187,
		0.986377,
		0.989154,
		0.991538,
		0.993551,
		0.995219,
		0.996571,
		0.997637,
		0.998451,
		0.999046,
		0.999459,
		0.999726,
		0.999882,
		0.999961,
		0.999992,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999956,
		0.999864,
		0.999675,
		0.999337,
		0.998793,
		0.997978,
		0.996822,
		0.995249,
		0.993181,
		0.990536,
		0.987232,
		0.983187,
		0.978317,
		0.972543,
		0.965784,
		0.957965,
		0.949014,
		0.938862,
		0.927446,
		0.914709,
		0.900598,
		0.88507,
		0.868085,
		0.849616,
		0.82964,
		0.808146,
		0.785132,
		0.760605,
		0.734587,
		0.707107,
		0.67821,
		0.647954,
		0.61641,
		0.583667,
		0.549827,
		0.515011,
		0.47936,
		0.443032,
		0.406208,
		0.369091,
		0.331906,
		0.294906,
		0.258369,
		0.222604,
		0.18795,
		0.154777,
		0.123492,
		0.094539,
		0.068402,
		0.045605,
		0.026722,
		0.01237,
		0.003221,
		0.835475,
		0.847207,
		0.858565,
		0.869533,
		0.880097,
		0.890244,
		0.899959,
		0.909232,
		0.918052,
		0.926409,
		0.934295,
		0.941704,
		0.948631,
		0.955074,
		0.961031,
		0.966504,
		0.971498,
		0.976017,
		0.980072,
		0.983674,
		0.986838,
		0.989583,
		0.991928,
		0.993899,
		0.995521,
		0.996826,
		0.997845,
		0.998613,
		0.999166,
		0.999542,
		0.999779,
		0.999911,
		0.999974,
		0.999996,
		1,
		1,
		0.999998,
		0.999983,
		0.999931,
		0.999805,
		0.999561,
		0.999142,
		0.998486,
		0.997523,
		0.996177,
		0.994368,
		0.992015,
		0.989031,
		0.985332,
		0.980831,
		0.975444,
		0.969087,
		0.961681,
		0.953149,
		0.943417,
		0.932418,
		0.92009,
		0.906376,
		0.891227,
		0.874601,
		0.856465,
		0.836792,
		0.815567,
		0.792785,
		0.76845,
		0.742579,
		0.715201,
		0.686356,
		0.656102,
		0.624508,
		0.591661,
		0.557664,
		0.522638,
		0.486723,
		0.45008,
		0.412891,
		0.375363,
		0.337724,
		0.300233,
		0.263174,
		0.226862,
		0.191643,
		0.157899,
		0.126047,
		0.096544,
		0.069887,
		0.046619,
		0.027329,
		0.012657,
		0.003297,
		0.796619,
		0.809569,
		0.822187,
		0.834457,
		0.846361,
		0.857883,
		0.869007,
		0.879718,
		0.890001,
		0.899843,
		0.909232,
		0.918157,
		0.926607,
		0.934575,
		0.942053,
		0.949038,
		0.955526,
		0.961517,
		0.967013,
		0.972017,
		0.976536,
		0.980581,
		0.984164,
		0.987301,
		0.990011,
		0.992316,
		0.994242,
		0.995817,
		0.997073,
		0.998044,
		0.998766,
		0.999278,
		0.999618,
		0.999825,
		0.999935,
		0.999984,
		0.999998,
		1,
		1,
		0.999995,
		0.99997,
		0.999894,
		0.999727,
		0.999417,
		0.998903,
		0.998117,
		0.996984,
		0.995423,
		0.99335,
		0.990679,
		0.987321,
		0.983187,
		0.978187,
		0.972236,
		0.965249,
		0.957143,
		0.947842,
		0.937272,
		0.925367,
		0.912065,
		0.897313,
		0.881063,
		0.863277,
		0.843926,
		0.82299,
		0.800458,
		0.776333,
		0.750628,
		0.723368,
		0.694592,
		0.664354,
		0.632722,
		0.599783,
		0.565638,
		0.530408,
		0.494234,
		0.457278,
		0.419726,
		0.381784,
		0.343689,
		0.3057,
		0.268109,
		0.231239,
		0.195443,
		0.161115,
		0.128681,
		0.098612,
		0.071421,
		0.047667,
		0.027957,
		0.012955,
		0.003376,
		0.753719,
		0.767766,
		0.781532,
		0.794999,
		0.808146,
		0.820956,
		0.833411,
		0.845492,
		0.857183,
		0.868468,
		0.879329,
		0.889753,
		0.899725,
		0.909232,
		0.918264,
		0.926809,
		0.93486,
		0.94241,
		0.949453,
		0.955987,
		0.962012,
		0.967529,
		0.972543,
		0.977061,
		0.981094,
		0.984657,
		0.987764,
		0.990437,
		0.9927,
		0.99458,
		0.996106,
		0.997312,
		0.998234,
		0.99891,
		0.999381,
		0.999686,
		0.999864,
		0.999955,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.99999,
		0.99995,
		0.999844,
		0.999625,
		0.999236,
		0.99861,
		0.997675,
		0.996348,
		0.994545,
		0.992177,
		0.989154,
		0.985382,
		0.98077,
		0.975226,
		0.968661,
		0.960989,
		0.952128,
		0.941999,
		0.930531,
		0.917657,
		0.903317,
		0.887461,
		0.870045,
		0.851035,
		0.830407,
		0.808146,
		0.78425,
		0.758729,
		0.731605,
		0.702914,
		0.672708,
		0.641052,
		0.608032,
		0.573749,
		0.538323,
		0.501896,
		0.464632,
		0.426716,
		0.38836,
		0.349803,
		0.31131,
		0.273179,
		0.23574,
		0.199356,
		0.164428,
		0.131398,
		0.100747,
		0.073006,
		0.04875,
		0.028608,
		0.013263,
		0.003458,
		0.707107,
		0.722107,
		0.736881,
		0.751412,
		0.765677,
		0.779657,
		0.793332,
		0.806683,
		0.819691,
		0.832336,
		0.8446,
		0.856465,
		0.867913,
		0.878929,
		0.889497,
		0.899603,
		0.909232,
		0.918374,
		0.927017,
		0.935153,
		0.942774,
		0.949876,
		0.956457,
		0.962515,
		0.968053,
		0.973075,
		0.977591,
		0.981612,
		0.985151,
		0.988228,
		0.990862,
		0.993081,
		0.994912,
		0.996388,
		0.997543,
		0.998415,
		0.999046,
		0.999475,
		0.999746,
		0.999898,
		0.99997,
		0.999995,
		1,
		1,
		0.999998,
		0.99998,
		0.99992,
		0.999775,
		0.999493,
		0.999011,
		0.998255,
		0.997147,
		0.995601,
		0.993526,
		0.990828,
		0.987414,
		0.983187,
		0.978051,
		0.971912,
		0.964681,
		0.956268,
		0.946591,
		0.935573,
		0.923141,
		0.909232,
		0.893788,
		0.876762,
		0.858112,
		0.837812,
		0.815842,
		0.792195,
		0.766878,
		0.739908,
		0.71132,
		0.681162,
		0.649498,
		0.61641,
		0.582,
		0.546387,
		0.509713,
		0.472143,
		0.433865,
		0.395094,
		0.356071,
		0.317068,
		0.278389,
		0.24037,
		0.203384,
		0.167843,
		0.1342,
		0.102952,
		0.074644,
		0.049871,
		0.029281,
		0.013582,
		0.003544,
		0.657196,
		0.672983,
		0.688606,
		0.704043,
		0.719275,
		0.73428,
		0.749038,
		0.763528,
		0.777728,
		0.791619,
		0.805179,
		0.81839,
		0.83123,
		0.843682,
		0.855726,
		0.867344,
		0.878519,
		0.889236,
		0.899478,
		0.909232,
		0.918486,
		0.927229,
		0.935451,
		0.943146,
		0.950308,
		0.956935,
		0.963026,
		0.968585,
		0.973615,
		0.978128,
		0.982133,
		0.985648,
		0.988691,
		0.991285,
		0.993458,
		0.995238,
		0.996662,
		0.997765,
		0.998588,
		0.999172,
		0.999561,
		0.999798,
		0.999926,
		0.999981,
		0.999998,
		1,
		1,
		0.999995,
		0.999965,
		0.999877,
		0.999684,
		0.999326,
		0.998733,
		0.997826,
		0.996521,
		0.994726,
		0.992346,
		0.989282,
		0.985435,
		0.980706,
		0.974996,
		0.968209,
		0.960253,
		0.951039,
		0.940484,
		0.928511,
		0.915048,
		0.900035,
		0.883417,
		0.865149,
		0.845197,
		0.823539,
		0.800163,
		0.77507,
		0.748275,
		0.719808,
		0.689715,
		0.658057,
		0.624916,
		0.59039,
		0.5546,
		0.517687,
		0.479816,
		0.441178,
		0.40199,
		0.362499,
		0.32298,
		0.283743,
		0.245133,
		0.207533,
		0.171364,
		0.137092,
		0.105229,
		0.076337,
		0.05103,
		0.029978,
		0.013913,
		0.003632,
		0.604481,
		0.620871,
		0.63716,
		0.653328,
		0.669354,
		0.685216,
		0.700893,
		0.716363,
		0.731605,
		0.746597,
		0.761317,
		0.775744,
		0.789856,
		0.803632,
		0.817052,
		0.830094,
		0.842738,
		0.854966,
		0.866759,
		0.878098,
		0.888967,
		0.89935,
		0.909232,
		0.918602,
		0.927446,
		0.935757,
		0.943527,
		0.950749,
		0.957423,
		0.963547,
		0.969125,
		0.974162,
		0.978669,
		0.982658,
		0.986146,
		0.989154,
		0.991705,
		0.993829,
		0.995558,
		0.996928,
		0.997978,
		0.99875,
		0.999289,
		0.999638,
		0.999844,
		0.999948,
		0.999989,
		0.999999,
		1,
		0.999999,
		0.999988,
		0.999942,
		0.999819,
		0.999565,
		0.999115,
		0.998392,
		0.997312,
		0.995782,
		0.993706,
		0.990983,
		0.987511,
		0.983187,
		0.977906,
		0.971569,
		0.964076,
		0.955335,
		0.945255,
		0.933754,
		0.920755,
		0.906191,
		0.890001,
		0.872136,
		0.852555,
		0.83123,
		0.808146,
		0.783299,
		0.756699,
		0.728373,
		0.698364,
		0.66673,
		0.63355,
		0.598922,
		0.562964,
		0.52582,
		0.487653,
		0.448658,
		0.409054,
		0.369091,
		0.32905,
		0.289247,
		0.250035,
		0.211807,
		0.174994,
		0.140077,
		0.107582,
		0.078089,
		0.05223,
		0.0307,
		0.014256,
		0.003724,
		0.549528,
		0.566321,
		0.583079,
		0.599783,
		0.61641,
		0.63294,
		0.649351,
		0.665622,
		0.681729,
		0.697651,
		0.713366,
		0.728851,
		0.744084,
		0.759042,
		0.773702,
		0.788042,
		0.80204,
		0.815675,
		0.828924,
		0.841768,
		0.854185,
		0.866157,
		0.877665,
		0.88869,
		0.899218,
		0.909232,
		0.91872,
		0.927669,
		0.93607,
		0.943915,
		0.951199,
		0.95792,
		0.964076,
		0.969673,
		0.974716,
		0.979216,
		0.983187,
		0.986646,
		0.989616,
		0.992122,
		0.994196,
		0.995871,
		0.997186,
		0.998182,
		0.998903,
		0.999396,
		0.999707,
		0.999882,
		0.999965,
		0.999994,
		1,
		1,
		0.999998,
		0.999977,
		0.999907,
		0.999739,
		0.999412,
		0.998853,
		0.997978,
		0.996697,
		0.994912,
		0.99252,
		0.989415,
		0.98549,
		0.980638,
		0.974752,
		0.967729,
		0.959469,
		0.949876,
		0.938862,
		0.926344,
		0.912247,
		0.896506,
		0.879064,
		0.859876,
		0.838908,
		0.816138,
		0.791559,
		0.765177,
		0.737013,
		0.707107,
		0.675515,
		0.642312,
		0.607595,
		0.571482,
		0.534114,
		0.495659,
		0.45631,
		0.416289,
		0.375851,
		0.335282,
		0.294906,
		0.255081,
		0.21621,
		0.178739,
		0.14316,
		0.110014,
		0.079901,
		0.053473,
		0.031449,
		0.014613,
		0.003819,
		0.492972,
		0.509955,
		0.526971,
		0.544,
		0.561021,
		0.578013,
		0.594955,
		0.611825,
		0.628601,
		0.645261,
		0.661782,
		0.678141,
		0.694316,
		0.710282,
		0.726017,
		0.741497,
		0.756699,
		0.771599,
		0.786174,
		0.800401,
		0.814257,
		0.827721,
		0.840769,
		0.853382,
		0.865538,
		0.877219,
		0.888407,
		0.899083,
		0.909232,
		0.918841,
		0.927897,
		0.93639,
		0.944313,
		0.951659,
		0.958426,
		0.964615,
		0.97023,
		0.975277,
		0.979768,
		0.983719,
		0.987147,
		0.990077,
		0.992536,
		0.994557,
		0.996177,
		0.997435,
		0.998376,
		0.999046,
		0.999493,
		0.999767,
		0.999914,
		0.999978,
		0.999998,
		1,
		1,
		0.999994,
		0.999959,
		0.999857,
		0.999632,
		0.999216,
		0.998528,
		0.997478,
		0.995967,
		0.993892,
		0.991144,
		0.987613,
		0.983187,
		0.977753,
		0.971204,
		0.963433,
		0.954338,
		0.943825,
		0.931803,
		0.918192,
		0.90292,
		0.885923,
		0.867151,
		0.846562,
		0.824131,
		0.799844,
		0.773702,
		0.745722,
		0.71594,
		0.684409,
		0.651201,
		0.61641,
		0.580154,
		0.542574,
		0.503836,
		0.464137,
		0.423701,
		0.382786,
		0.341684,
		0.300724,
		0.260276,
		0.22075,
		0.182603,
		0.146343,
		0.112529,
		0.081776,
		0.054761,
		0.032226,
		0.014982,
		0.003918,
		0.435513,
		0.452462,
		0.469512,
		0.486644,
		0.503836,
		0.52107,
		0.538323,
		0.555576,
		0.572806,
		0.589991,
		0.607109,
		0.624138,
		0.641052,
		0.657831,
		0.674449,
		0.690882,
		0.707107,
		0.723099,
		0.738833,
		0.754286,
		0.769433,
		0.78425,
		0.798713,
		0.812798,
		0.826481,
		0.839741,
		0.852555,
		0.864902,
		0.876762,
		0.888115,
		0.898944,
		0.909232,
		0.918966,
		0.928131,
		0.936718,
		0.944719,
		0.952128,
		0.958943,
		0.965164,
		0.970795,
		0.975845,
		0.980326,
		0.984253,
		0.987648,
		0.990536,
		0.992945,
		0.994912,
		0.996474,
		0.997675,
		0.99856,
		0.999178,
		0.999581,
		0.999819,
		0.999939,
		0.999987,
		0.999999,
		1,
		0.999999,
		0.999986,
		0.999932,
		0.999788,
		0.999493,
		0.99897,
		0.99813,
		0.996876,
		0.995103,
		0.9927,
		0.989554,
		0.985548,
		0.980567,
		0.974494,
		0.967218,
		0.958631,
		0.948631,
		0.937122,
		0.924016,
		0.909232,
		0.892703,
		0.874369,
		0.854185,
		0.832117,
		0.808146,
		0.782268,
		0.754496,
		0.72486,
		0.69341,
		0.660215,
		0.625367,
		0.588982,
		0.5512,
		0.512188,
		0.472143,
		0.431293,
		0.3899,
		0.34826,
		0.306709,
		0.265625,
		0.225429,
		0.186592,
		0.149633,
		0.11513,
		0.083718,
		0.056095,
		0.033031,
		0.015366,
		0.00402,
		0.377913,
		0.394596,
		0.411446,
		0.428447,
		0.445578,
		0.46282,
		0.480153,
		0.497557,
		0.515011,
		0.532493,
		0.549982,
		0.567455,
		0.584888,
		0.60226,
		0.619546,
		0.636722,
		0.653763,
		0.670646,
		0.687346,
		0.703836,
		0.720092,
		0.736089,
		0.751801,
		0.767202,
		0.782268,
		0.796974,
		0.811294,
		0.825205,
		0.838682,
		0.851704,
		0.864247,
		0.87629,
		0.887815,
		0.898801,
		0.909232,
		0.919093,
		0.928371,
		0.937054,
		0.945135,
		0.952607,
		0.959469,
		0.965721,
		0.971369,
		0.97642,
		0.980889,
		0.984791,
		0.98815,
		0.990993,
		0.99335,
		0.99526,
		0.996763,
		0.997904,
		0.998733,
		0.9993,
		0.99966,
		0.999863,
		0.999959,
		0.999993,
		1,
		1,
		0.999997,
		0.999973,
		0.999891,
		0.999695,
		0.999313,
		0.998662,
		0.997644,
		0.996156,
		0.994084,
		0.991312,
		0.98772,
		0.983187,
		0.977591,
		0.970816,
		0.962746,
		0.953272,
		0.94229,
		0.929706,
		0.915432,
		0.899393,
		0.881522,
		0.861767,
		0.840087,
		0.816458,
		0.790869,
		0.763329,
		0.733863,
		0.702516,
		0.669354,
		0.634466,
		0.597967,
		0.559995,
		0.520718,
		0.480333,
		0.439072,
		0.397198,
		0.355015,
		0.312866,
		0.271135,
		0.230255,
		0.19071,
		0.153034,
		0.117821,
		0.085729,
		0.057479,
		0.033867,
		0.015765,
		0.004127,
		0.320992,
		0.337168,
		0.353578,
		0.370205,
		0.387032,
		0.40404,
		0.42121,
		0.438523,
		0.455958,
		0.473495,
		0.491113,
		0.50879,
		0.526505,
		0.544233,
		0.561953,
		0.57964,
		0.59727,
		0.61482,
		0.632263,
		0.649575,
		0.66673,
		0.683702,
		0.700466,
		0.716994,
		0.73326,
		0.749239,
		0.764902,
		0.780225,
		0.795181,
		0.809744,
		0.823889,
		0.837591,
		0.850827,
		0.863572,
		0.875806,
		0.887506,
		0.898654,
		0.909232,
		0.919224,
		0.928617,
		0.937398,
		0.94556,
		0.953097,
		0.960006,
		0.966289,
		0.971951,
		0.977002,
		0.981456,
		0.985332,
		0.988652,
		0.991447,
		0.99375,
		0.995601,
		0.997042,
		0.998123,
		0.998895,
		0.999412,
		0.999729,
		0.999899,
		0.999974,
		0.999997,
		1,
		1,
		0.999993,
		0.999952,
		0.999833,
		0.99957,
		0.999083,
		0.99828,
		0.997056,
		0.995298,
		0.992887,
		0.9897,
		0.98561,
		0.980491,
		0.974219,
		0.966673,
		0.957735,
		0.947296,
		0.935251,
		0.921508,
		0.90598,
		0.888597,
		0.869296,
		0.848031,
		0.82477,
		0.799498,
		0.772216,
		0.742944,
		0.711723,
		0.678615,
		0.643707,
		0.607109,
		0.568961,
		0.529428,
		0.488711,
		0.447041,
		0.404687,
		0.361957,
		0.3192,
		0.276811,
		0.235233,
		0.194962,
		0.156549,
		0.120607,
		0.087813,
		0.058915,
		0.034735,
		0.016179,
		0.004238,
		0.265625,
		0.281048,
		0.29677,
		0.312776,
		0.32905,
		0.345573,
		0.362328,
		0.379296,
		0.39646,
		0.413799,
		0.431293,
		0.448922,
		0.466665,
		0.484498,
		0.502402,
		0.520351,
		0.538323,
		0.556295,
		0.57424,
		0.592135,
		0.609954,
		0.627672,
		0.645261,
		0.662695,
		0.679948,
		0.696992,
		0.7138,
		0.730344,
		0.746597,
		0.762531,
		0.778118,
		0.793332,
		0.808146,
		0.822533,
		0.836467,
		0.849923,
		0.862877,
		0.875306,
		0.887188,
		0.898503,
		0.909232,
		0.919359,
		0.928869,
		0.937751,
		0.945995,
		0.953597,
		0.960553,
		0.966867,
		0.972543,
		0.977591,
		0.982029,
		0.985874,
		0.989154,
		0.991898,
		0.994144,
		0.995933,
		0.997312,
		0.998331,
		0.999046,
		0.999512,
		0.999788,
		0.999929,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.99992,
		0.999752,
		0.999406,
		0.998793,
		0.997811,
		0.996348,
		0.994282,
		0.991486,
		0.987832,
		0.983187,
		0.97742,
		0.970403,
		0.962012,
		0.952128,
		0.94064,
		0.927446,
		0.912454,
		0.895582,
		0.876762,
		0.855939,
		0.833075,
		0.808146,
		0.781149,
		0.752097,
		0.721027,
		0.687996,
		0.653088,
		0.61641,
		0.5781,
		0.538323,
		0.497281,
		0.455206,
		0.412372,
		0.369091,
		0.325719,
		0.282661,
		0.24037,
		0.199356,
		0.160186,
		0.123492,
		0.089974,
		0.060404,
		0.035637,
		0.01661,
		0.004354,
		0.212739,
		0.227159,
		0.241943,
		0.257074,
		0.272538,
		0.288319,
		0.304401,
		0.320766,
		0.337397,
		0.354276,
		0.371384,
		0.388702,
		0.406208,
		0.423884,
		0.441707,
		0.459655,
		0.477707,
		0.495839,
		0.514027,
		0.532247,
		0.550474,
		0.568684,
		0.586849,
		0.604944,
		0.622942,
		0.640815,
		0.658536,
		0.676077,
		0.69341,
		0.710506,
		0.727336,
		0.743871,
		0.760084,
		0.775945,
		0.791425,
		0.806498,
		0.821134,
		0.835308,
		0.848992,
		0.862161,
		0.874792,
		0.886861,
		0.898348,
		0.909232,
		0.919497,
		0.929128,
		0.938112,
		0.94644,
		0.954108,
		0.961112,
		0.967454,
		0.973143,
		0.978187,
		0.982605,
		0.986419,
		0.989654,
		0.992346,
		0.994532,
		0.996257,
		0.997571,
		0.998528,
		0.999185,
		0.999603,
		0.999839,
		0.999952,
		0.999992,
		1,
		1,
		0.999997,
		0.999969,
		0.999872,
		0.999641,
		0.999193,
		0.998429,
		0.997238,
		0.995499,
		0.993081,
		0.989852,
		0.985674,
		0.980411,
		0.973927,
		0.96609,
		0.956775,
		0.94586,
		0.933236,
		0.9188,
		0.902465,
		0.884152,
		0.863799,
		0.84136,
		0.816804,
		0.79012,
		0.761317,
		0.730424,
		0.697495,
		0.662608,
		0.625869,
		0.587413,
		0.547405,
		0.506046,
		0.463572,
		0.420258,
		0.376422,
		0.332429,
		0.28869,
		0.245671,
		0.203896,
		0.163949,
		0.12648,
		0.092214,
		0.061951,
		0.036574,
		0.017058,
		0.004475,
		0.163312,
		0.176478,
		0.190067,
		0.204067,
		0.218463,
		0.233241,
		0.248386,
		0.263882,
		0.279714,
		0.295864,
		0.312315,
		0.32905,
		0.346048,
		0.363292,
		0.38076,
		0.398433,
		0.416289,
		0.434306,
		0.452462,
		0.470734,
		0.489096,
		0.507526,
		0.525998,
		0.544486,
		0.562964,
		0.581406,
		0.599783,
		0.618068,
		0.636233,
		0.654248,
		0.672085,
		0.689715,
		0.707107,
		0.724231,
		0.741059,
		0.757559,
		0.773702,
		0.789457,
		0.804797,
		0.819691,
		0.834112,
		0.848031,
		0.861423,
		0.874262,
		0.886525,
		0.898188,
		0.909232,
		0.919639,
		0.929393,
		0.938482,
		0.946896,
		0.95463,
		0.961681,
		0.968053,
		0.973751,
		0.97879,
		0.983187,
		0.986964,
		0.990153,
		0.992789,
		0.994912,
		0.996571,
		0.997819,
		0.998713,
		0.999313,
		0.999682,
		0.999882,
		0.99997,
		0.999997,
		1,
		1,
		0.999991,
		0.999944,
		0.999803,
		0.999493,
		0.998921,
		0.997978,
		0.996543,
		0.994485,
		0.991668,
		0.987949,
		0.983187,
		0.977237,
		0.969962,
		0.961225,
		0.950898,
		0.938862,
		0.925006,
		0.909232,
		0.891454,
		0.8716,
		0.849616,
		0.825463,
		0.799123,
		0.770596,
		0.739908,
		0.707107,
		0.672265,
		0.635486,
		0.596902,
		0.556677,
		0.515011,
		0.472143,
		0.428351,
		0.383959,
		0.339336,
		0.294906,
		0.251144,
		0.208589,
		0.167843,
		0.129577,
		0.094539,
		0.063557,
		0.037549,
		0.017525,
		0.0046,
		0.118371,
		0.130029,
		0.142168,
		0.154777,
		0.167843,
		0.181354,
		0.195295,
		0.209654,
		0.224414,
		0.239562,
		0.255081,
		0.270954,
		0.287165,
		0.303696,
		0.320527,
		0.33764,
		0.355015,
		0.372632,
		0.390469,
		0.408505,
		0.426716,
		0.44508,
		0.463572,
		0.482168,
		0.500843,
		0.51957,
		0.538323,
		0.557076,
		0.575799,
		0.594464,
		0.613044,
		0.631507,
		0.649825,
		0.667967,
		0.685901,
		0.703599,
		0.721027,
		0.738155,
		0.754952,
		0.771385,
		0.787425,
		0.803041,
		0.818201,
		0.832877,
		0.84704,
		0.860662,
		0.873716,
		0.886177,
		0.898023,
		0.909232,
		0.919785,
		0.929666,
		0.938862,
		0.947363,
		0.955164,
		0.962262,
		0.968661,
		0.974369,
		0.9794,
		0.983772,
		0.987511,
		0.99065,
		0.993226,
		0.995285,
		0.996876,
		0.998056,
		0.998886,
		0.999429,
		0.999752,
		0.999917,
		0.999982,
		0.999999,
		1,
		0.999999,
		0.999981,
		0.999905,
		0.999707,
		0.999299,
		0.998577,
		0.997422,
		0.995704,
		0.993282,
		0.990011,
		0.985742,
		0.980326,
		0.973615,
		0.965467,
		0.955742,
		0.944313,
		0.931058,
		0.91587,
		0.898654,
		0.879329,
		0.85783,
		0.834112,
		0.808146,
		0.779928,
		0.749475,
		0.716828,
		0.682058,
		0.645261,
		0.606567,
		0.56614,
		0.52418,
		0.480925,
		0.436658,
		0.391707,
		0.346449,
		0.301315,
		0.256796,
		0.213443,
		0.171875,
		0.132788,
		0.096952,
		0.065227,
		0.038563,
		0.018011,
		0.004731,
		0.078987,
		0.088884,
		0.099316,
		0.110275,
		0.121749,
		0.133727,
		0.146196,
		0.159146,
		0.172561,
		0.18643,
		0.200737,
		0.215467,
		0.230606,
		0.246136,
		0.262041,
		0.278304,
		0.294906,
		0.311827,
		0.32905,
		0.346552,
		0.364313,
		0.382312,
		0.400525,
		0.418929,
		0.437501,
		0.456215,
		0.475047,
		0.493969,
		0.512956,
		0.531979,
		0.55101,
		0.570021,
		0.588982,
		0.607863,
		0.626633,
		0.645261,
		0.663715,
		0.681965,
		0.699976,
		0.717717,
		0.735156,
		0.752258,
		0.768993,
		0.785326,
		0.801227,
		0.816662,
		0.831602,
		0.846017,
		0.859876,
		0.873152,
		0.88582,
		0.897854,
		0.909232,
		0.919935,
		0.929947,
		0.939252,
		0.947842,
		0.95571,
		0.962855,
		0.969281,
		0.974996,
		0.980016,
		0.984361,
		0.988059,
		0.991144,
		0.993658,
		0.995649,
		0.997169,
		0.99828,
		0.999046,
		0.999533,
		0.999811,
		0.999944,
		0.999991,
		1,
		1,
		0.999996,
		0.999963,
		0.999848,
		0.999575,
		0.999046,
		0.998145,
		0.996742,
		0.994696,
		0.991857,
		0.988073,
		0.983187,
		0.977043,
		0.969489,
		0.960379,
		0.949573,
		0.936941,
		0.922365,
		0.905738,
		0.886971,
		0.86599,
		0.842738,
		0.817181,
		0.789303,
		0.759116,
		0.726654,
		0.691982,
		0.655192,
		0.61641,
		0.575799,
		0.533556,
		0.489923,
		0.445185,
		0.399673,
		0.353774,
		0.307926,
		0.262634,
		0.218463,
		0.176052,
		0.136118,
		0.099458,
		0.066963,
		0.039618,
		0.018518,
		0.004868,
		0.046277,
		0.054161,
		0.062632,
		0.071682,
		0.081301,
		0.09148,
		0.102209,
		0.113477,
		0.125272,
		0.137583,
		0.150396,
		0.163699,
		0.177478,
		0.191718,
		0.206404,
		0.22152,
		0.237049,
		0.252975,
		0.26928,
		0.285944,
		0.302949,
		0.320274,
		0.337898,
		0.355799,
		0.373956,
		0.392345,
		0.410941,
		0.429721,
		0.448658,
		0.467727,
		0.4869,
		0.506149,
		0.525446,
		0.544762,
		0.564067,
		0.583329,
		0.602518,
		0.621603,
		0.640549,
		0.659326,
		0.677898,
		0.696234,
		0.714298,
		0.732056,
		0.749475,
		0.76652,
		0.783156,
		0.799352,
		0.815072,
		0.830285,
		0.84496,
		0.859064,
		0.872571,
		0.885451,
		0.897679,
		0.909232,
		0.92009,
		0.930235,
		0.939652,
		0.948332,
		0.956268,
		0.963459,
		0.969911,
		0.975631,
		0.980638,
		0.984953,
		0.988607,
		0.991635,
		0.994084,
		0.996003,
		0.997451,
		0.998492,
		0.999193,
		0.999626,
		0.999861,
		0.999964,
		0.999996,
		1,
		1,
		0.99999,
		0.999933,
		0.999766,
		0.999399,
		0.998722,
		0.997607,
		0.995914,
		0.99349,
		0.990178,
		0.985814,
		0.980236,
		0.973282,
		0.964797,
		0.95463,
		0.942641,
		0.9287,
		0.912691,
		0.894513,
		0.874082,
		0.851331,
		0.826216,
		0.798713,
		0.768825,
		0.736579,
		0.702034,
		0.665277,
		0.62643,
		0.585653,
		0.543144,
		0.499142,
		0.453937,
		0.407865,
		0.361318,
		0.314747,
		0.268666,
		0.223658,
		0.18038,
		0.139573,
		0.102062,
		0.068768,
		0.040718,
		0.019046,
		0.00501,
		0.021402,
		0.027023,
		0.033279,
		0.040162,
		0.047667,
		0.055783,
		0.064503,
		0.073818,
		0.083718,
		0.094192,
		0.105229,
		0.116819,
		0.128949,
		0.141606,
		0.154777,
		0.168448,
		0.182603,
		0.197229,
		0.212308,
		0.227824,
		0.243759,
		0.260094,
		0.276811,
		0.29389,
		0.31131,
		0.32905,
		0.347086,
		0.365398,
		0.383959,
		0.402746,
		0.421733,
		0.440894,
		0.460201,
		0.479627,
		0.499142,
		0.518718,
		0.538323,
		0.557927,
		0.577498,
		0.597004,
		0.61641,
		0.635684,
		0.654791,
		0.673697,
		0.692366,
		0.710763,
		0.728851,
		0.746597,
		0.763962,
		0.780913,
		0.797413,
		0.813428,
		0.828924,
		0.843868,
		0.858226,
		0.87197,
		0.88507,
		0.897499,
		0.909232,
		0.920249,
		0.930531,
		0.940063,
		0.948834,
		0.956839,
		0.964076,
		0.970552,
		0.976276,
		0.981267,
		0.985548,
		0.989154,
		0.992122,
		0.994502,
		0.996348,
		0.997721,
		0.99869,
		0.999327,
		0.999707,
		0.999901,
		0.999979,
		0.999999,
		1,
		0.999999,
		0.999977,
		0.999887,
		0.999651,
		0.999166,
		0.99831,
		0.996943,
		0.994912,
		0.992055,
		0.988203,
		0.983187,
		0.976835,
		0.968983,
		0.959469,
		0.948142,
		0.93486,
		0.919497,
		0.901939,
		0.882091,
		0.859876,
		0.835238,
		0.808146,
		0.778592,
		0.746597,
		0.71221,
		0.675515,
		0.636628,
		0.595706,
		0.552945,
		0.508587,
		0.462921,
		0.416289,
		0.369091,
		0.321785,
		0.2749,
		0.229035,
		0.184867,
		0.14316,
		0.104768,
		0.070648,
		0.041863,
		0.019597,
		0.005159,
		0.005562,
		0.008674,
		0.012464,
		0.016928,
		0.022059,
		0.027851,
		0.034297,
		0.041389,
		0.049119,
		0.057479,
		0.06646,
		0.076051,
		0.086243,
		0.097025,
		0.108384,
		0.120309,
		0.132788,
		0.145806,
		0.159349,
		0.173402,
		0.18795,
		0.202976,
		0.218463,
		0.234393,
		0.250747,
		0.267506,
		0.28465,
		0.302157,
		0.320005,
		0.338171,
		0.356632,
		0.375363,
		0.394338,
		0.413531,
		0.432915,
		0.452462,
		0.472143,
		0.491928,
		0.511786,
		0.531686,
		0.551596,
		0.571482,
		0.591311,
		0.611048,
		0.630658,
		0.650105,
		0.669354,
		0.688367,
		0.707107,
		0.725537,
		0.743619,
		0.761317,
		0.778592,
		0.795408,
		0.811728,
		0.827516,
		0.842738,
		0.85736,
		0.87135,
		0.884676,
		0.897313,
		0.909232,
		0.920413,
		0.930835,
		0.940484,
		0.949349,
		0.957423,
		0.964706,
		0.971204,
		0.976929,
		0.981901,
		0.986146,
		0.9897,
		0.992605,
		0.994912,
		0.996681,
		0.997978,
		0.998875,
		0.999448,
		0.999776,
		0.999933,
		0.999989,
		1,
		1,
		0.999995,
		0.999956,
		0.999819,
		0.999493,
		0.998863,
		0.997793,
		0.996128,
		0.993706,
		0.990352,
		0.98589,
		0.98014,
		0.972926,
		0.964076,
		0.953429,
		0.94083,
		0.926139,
		0.909232,
		0.890001,
		0.868358,
		0.844236,
		0.817592,
		0.788409,
		0.756699,
		0.722505,
		0.685901,
		0.647002,
		0.605958,
		0.562964,
		0.518262,
		0.472143,
		0.424954,
		0.377099,
		0.32905,
		0.281345,
		0.234603,
		0.18952,
		0.146884,
		0.107582,
		0.072605,
		0.043058,
		0.020173,
		0.005315,
		0.000361,
		0.00144,
		0.003233,
		0.005737,
		0.008946,
		0.012854,
		0.017457,
		0.022747,
		0.028718,
		0.035363,
		0.042672,
		0.050639,
		0.059254,
		0.068506,
		0.078386,
		0.088884,
		0.099986,
		0.111681,
		0.123957,
		0.136799,
		0.150192,
		0.164123,
		0.178574,
		0.193529,
		0.208972,
		0.224882,
		0.241242,
		0.258031,
		0.275228,
		0.292812,
		0.310761,
		0.32905,
		0.347655,
		0.366551,
		0.385711,
		0.405109,
		0.424716,
		0.444504,
		0.464441,
		0.484498,
		0.504643,
		0.524843,
		0.545064,
		0.565272,
		0.585432,
		0.605508,
		0.625463,
		0.645261,
		0.664862,
		0.68423,
		0.703324,
		0.722107,
		0.740538,
		0.758579,
		0.77619,
		0.793332,
		0.809968,
		0.82606,
		0.84157,
		0.856465,
		0.870708,
		0.88427,
		0.897121,
		0.909232,
		0.920582,
		0.931149,
		0.940917,
		0.949876,
		0.95802,
		0.965348,
		0.971868,
		0.977591,
		0.982541,
		0.986746,
		0.990244,
		0.993081,
		0.995313,
		0.997003,
		0.998222,
		0.999046,
		0.999556,
		0.999834,
		0.999957,
		0.999995,
		1,
		1,
		0.999988,
		0.99992,
		0.99972,
		0.999281,
		0.998474,
		0.997147,
		0.995135,
		0.992261,
		0.98834,
		0.983187,
		0.976614,
		0.968439,
		0.958486,
		0.946591,
		0.9326,
		0.916375,
		0.897796,
		0.876762,
		0.853193,
		0.827036,
		0.798264,
		0.766878,
		0.732912,
		0.696434,
		0.657551,
		0.61641,
		0.573204,
		0.528173,
		0.48161,
		0.433865,
		0.385352,
		0.336549,
		0.288011,
		0.24037,
		0.194346,
		0.150754,
		0.110511,
		0.074644,
		0.044304,
		0.020774,
		0.005478,
		0.000372,
		0.001486,
		0.003336,
		0.00592,
		0.00923,
		0.013263,
		0.018011,
		0.023468,
		0.029626,
		0.036479,
		0.044016,
		0.05223,
		0.061111,
		0.070648,
		0.08083,
		0.091646,
		0.103084,
		0.11513,
		0.127771,
		0.140991,
		0.154777,
		0.169111,
		0.183976,
		0.199356,
		0.21523,
		0.23158,
		0.248386,
		0.265625,
		0.283276,
		0.301315,
		0.319719,
		0.338462,
		0.357518,
		0.37686,
		0.39646,
		0.416289,
		0.436317,
		0.456513,
		0.476846,
		0.497281,
		0.517785,
		0.538323,
		0.55886,
		0.579359,
		0.599783,
		0.620093,
		0.64025,
		0.660215,
		0.679948,
		0.699408,
		0.718555,
		0.737347,
		0.755743,
		0.773702,
		0.791183,
		0.808146,
		0.824552,
		0.840361,
		0.855538,
		0.870045,
		0.883851,
		0.896922,
		0.909232,
		0.920755,
		0.931471,
		0.941362,
		0.950418,
		0.958631,
		0.966004,
		0.972543,
		0.978262,
		0.983187,
		0.987347,
		0.990785,
		0.993551,
		0.995704,
		0.997312,
		0.998451,
		0.999202,
		0.999651,
		0.999882,
		0.999975,
		0.999998,
		1,
		0.999998,
		0.999973,
		0.999864,
		0.999581,
		0.999001,
		0.997978,
		0.996348,
		0.99393,
		0.990536,
		0.98597,
		0.980037,
		0.972543,
		0.963299,
		0.952128,
		0.938862,
		0.92335,
		0.905457,
		0.88507,
		0.862095,
		0.836467,
		0.808146,
		0.777124,
		0.743424,
		0.707107,
		0.668272,
		0.627063,
		0.583667,
		0.538323,
		0.491327,
		0.443032,
		0.393858,
		0.344293,
		0.294906,
		0.246346,
		0.199356,
		0.154777,
		0.113559,
		0.07677,
		0.045605,
		0.021402,
		0.005649,
		0.000384,
		0.001534,
		0.003444,
		0.006111,
		0.009529,
		0.013691,
		0.018592,
		0.024223,
		0.030578,
		0.037648,
		0.045425,
		0.053897,
		0.063057,
		0.072891,
		0.083389,
		0.094539,
		0.106327,
		0.118739,
		0.131761,
		0.145377,
		0.159571,
		0.174326,
		0.189623,
		0.205443,
		0.221767,
		0.238574,
		0.255841,
		0.273547,
		0.291667,
		0.310176,
		0.32905,
		0.34826,
		0.367779,
		0.387578,
		0.407628,
		0.427896,
		0.448352,
		0.468961,
		0.48969,
		0.510503,
		0.531365,
		0.552238,
		0.573084,
		0.593863,
		0.614537,
		0.635065,
		0.655405,
		0.675515,
		0.695352,
		0.714875,
		0.73404,
		0.752804,
		0.771123,
		0.788955,
		0.806258,
		0.82299,
		0.839109,
		0.854579,
		0.869359,
		0.883417,
		0.896717,
		0.909232,
		0.920935,
		0.931803,
		0.94182,
		0.950973,
		0.959257,
		0.966673,
		0.973229,
		0.978942,
		0.983837,
		0.987949,
		0.991323,
		0.994013,
		0.996084,
		0.997607,
		0.998665,
		0.999343,
		0.999732,
		0.99992,
		0.999987,
		1,
		1,
		0.999994,
		0.999946,
		0.999782,
		0.999391,
		0.998635,
		0.997353,
		0.995364,
		0.992476,
		0.988485,
		0.983187,
		0.976376,
		0.967852,
		0.957423,
		0.944907,
		0.930138,
		0.912966,
		0.893264,
		0.870925,
		0.845868,
		0.818041,
		0.787425,
		0.754033,
		0.717915,
		0.679164,
		0.637915,
		0.594354,
		0.548718,
		0.501302,
		0.452462,
		0.402626,
		0.352291,
		0.30204,
		0.25254,
		0.204557,
		0.15896,
		0.116734,
		0.078987,
		0.046964,
		0.022059,
		0.005827,
		0.000397,
		0.001584,
		0.003558,
		0.006313,
		0.009842,
		0.014141,
		0.019201,
		0.025016,
		0.031577,
		0.038875,
		0.046901,
		0.055645,
		0.065096,
		0.075242,
		0.086071,
		0.09757,
		0.109724,
		0.122519,
		0.135939,
		0.149968,
		0.164588,
		0.179781,
		0.195527,
		0.211807,
		0.228598,
		0.245878,
		0.263625,
		0.281814,
		0.300419,
		0.319414,
		0.338772,
		0.358463,
		0.378457,
		0.398725,
		0.419232,
		0.439948,
		0.460836,
		0.481862,
		0.502989,
		0.52418,
		0.545395,
		0.566596,
		0.587741,
		0.608789,
		0.629697,
		0.650423,
		0.670922,
		0.69115,
		0.711062,
		0.730613,
		0.749757,
		0.76845,
		0.786646,
		0.804301,
		0.82137,
		0.837812,
		0.853585,
		0.868649,
		0.882967,
		0.896506,
		0.909232,
		0.92112,
		0.932145,
		0.94229,
		0.951543,
		0.959898,
		0.967356,
		0.973927,
		0.97963,
		0.984492,
		0.988552,
		0.991857,
		0.994468,
		0.996452,
		0.997888,
		0.998863,
		0.99947,
		0.999801,
		0.999949,
		0.999994,
		1,
		1,
		0.999985,
		0.999903,
		0.999662,
		0.999134,
		0.998163,
		0.996571,
		0.994162,
		0.990728,
		0.986055,
		0.979927,
		0.97213,
		0.962458,
		0.950715,
		0.936718,
		0.920303,
		0.901325,
		0.879663,
		0.855222,
		0.827935,
		0.797771,
		0.764729,
		0.728851,
		0.690221,
		0.648967,
		0.605268,
		0.559361,
		0.511539,
		0.462163,
		0.411665,
		0.360554,
		0.309424,
		0.258963,
		0.209959,
		0.163312,
		0.120042,
		0.081301,
		0.048385,
		0.022747,
		0.006014,
		0.00041,
		0.001638,
		0.003677,
		0.006524,
		0.010171,
		0.014613,
		0.019841,
		0.025848,
		0.032625,
		0.040162,
		0.048451,
		0.057479,
		0.067236,
		0.077708,
		0.088884,
		0.100747,
		0.113285,
		0.12648,
		0.140317,
		0.154777,
		0.169842,
		0.185491,
		0.201706,
		0.218463,
		0.23574,
		0.253513,
		0.271757,
		0.290446,
		0.309553,
		0.32905,
		0.348906,
		0.369091,
		0.389573,
		0.410318,
		0.431293,
		0.452462,
		0.473788,
		0.495233,
		0.516759,
		0.538323,
		0.559887,
		0.581406,
		0.602837,
		0.624138,
		0.645261,
		0.666161,
		0.686792,
		0.707107,
		0.727057,
		0.746597,
		0.765677,
		0.78425,
		0.80227,
		0.819691,
		0.836467,
		0.852555,
		0.867913,
		0.882503,
		0.896287,
		0.909232,
		0.921311,
		0.932498,
		0.942774,
		0.952128,
		0.960553,
		0.968053,
		0.974637,
		0.980326,
		0.985151,
		0.989154,
		0.992386,
		0.994912,
		0.996806,
		0.998153,
		0.999046,
		0.999581,
		0.999858,
		0.99997,
		0.999998,
		1,
		0.999998,
		0.999967,
		0.999836,
		0.999493,
		0.998793,
		0.997561,
		0.995601,
		0.9927,
		0.988638,
		0.983187,
		0.97612,
		0.967218,
		0.956268,
		0.943071,
		0.927446,
		0.909232,
		0.888291,
		0.864511,
		0.837812,
		0.808146,
		0.775502,
		0.739908,
		0.701439,
		0.660215,
		0.61641,
		0.570256,
		0.522046,
		0.472143,
		0.420985,
		0.369091,
		0.317068,
		0.265625,
		0.215573,
		0.167843,
		0.123492,
		0.083718,
		0.04987,
		0.023468,
		0.006211,
		0.000424,
		0.001693,
		0.003803,
		0.006746,
		0.010517,
		0.015108,
		0.020513,
		0.026722,
		0.033726,
		0.041515,
		0.050078,
		0.059405,
		0.069482,
		0.080297,
		0.091835,
		0.104082,
		0.11702,
		0.130634,
		0.144906,
		0.159817,
		0.175346,
		0.191473,
		0.208175,
		0.225429,
		0.243212,
		0.261496,
		0.280256,
		0.299464,
		0.319089,
		0.339103,
		0.359472,
		0.380165,
		0.401146,
		0.42238,
		0.44383,
		0.465459,
		0.487226,
		0.50909,
		0.531011,
		0.552945,
		0.574848,
		0.596674,
		0.618377,
		0.63991,
		0.661224,
		0.682272,
		0.703003,
		0.723368,
		0.743316,
		0.762798,
		0.781763,
		0.800163,
		0.817948,
		0.835071,
		0.851487,
		0.867151,
		0.882021,
		0.89606,
		0.909232,
		0.921508,
		0.932861,
		0.943272,
		0.952728,
		0.961225,
		0.968764,
		0.975358,
		0.98103,
		0.985814,
		0.989754,
		0.992908,
		0.995346,
		0.997147,
		0.998403,
		0.999212,
		0.999678,
		0.999903,
		0.999984,
		1,
		1,
		0.999993,
		0.999935,
		0.999735,
		0.999261,
		0.998347,
		0.996799,
		0.994403,
		0.990931,
		0.986146,
		0.97981,
		0.971686,
		0.961546,
		0.949176,
		0.934374,
		0.916963,
		0.896786,
		0.873716,
		0.847654,
		0.818536,
		0.786339,
		0.751077,
		0.712812,
		0.671658,
		0.62778,
		0.581406,
		0.532828,
		0.48241,
		0.430596,
		0.377913,
		0.324985,
		0.272538,
		0.221409,
		0.172561,
		0.127091,
		0.086243,
		0.051425,
		0.024222,
		0.006416,
		0.000439,
		0.001752,
		0.003934,
		0.006979,
		0.01088,
		0.01563,
		0.02122,
		0.027641,
		0.034883,
		0.042936,
		0.051789,
		0.061429,
		0.071843,
		0.083016,
		0.094935,
		0.107582,
		0.120941,
		0.134994,
		0.149721,
		0.165103,
		0.181117,
		0.197741,
		0.214952,
		0.232725,
		0.251033,
		0.269849,
		0.289144,
		0.308888,
		0.32905,
		0.349596,
		0.370494,
		0.391707,
		0.413199,
		0.43493,
		0.456863,
		0.478955,
		0.501165,
		0.523449,
		0.545761,
		0.568056,
		0.590287,
		0.612405,
		0.63436,
		0.656102,
		0.67758,
		0.698742,
		0.719536,
		0.739908,
		0.759807,
		0.77918,
		0.797974,
		0.816138,
		0.833622,
		0.850378,
		0.866359,
		0.881522,
		0.895825,
		0.909232,
		0.921711,
		0.933236,
		0.943785,
		0.953345,
		0.961912,
		0.969489,
		0.976091,
		0.981742,
		0.986479,
		0.990352,
		0.993424,
		0.995768,
		0.997473,
		0.998635,
		0.999361,
		0.99976,
		0.999938,
		0.999993,
		1,
		1,
		0.999982,
		0.999882,
		0.999588,
		0.998946,
		0.997769,
		0.995843,
		0.992935,
		0.9888,
		0.983187,
		0.975845,
		0.966531,
		0.95501,
		0.941064,
		0.924495,
		0.905127,
		0.882814,
		0.85744,
		0.828924,
		0.797225,
		0.762345,
		0.724333,
		0.683291,
		0.639377,
		0.592813,
		0.54389,
		0.492972,
		0.440506,
		0.387032,
		0.333186,
		0.279714,
		0.227479,
		0.177477,
		0.130845,
		0.088879,
		0.053048,
		0.02501,
		0.00663,
		0.000454,
		0.001814,
		0.004073,
		0.007225,
		0.011263,
		0.016179,
		0.021964,
		0.028608,
		0.036101,
		0.044432,
		0.053588,
		0.063557,
		0.074325,
		0.085875,
		0.098193,
		0.111261,
		0.125061,
		0.139573,
		0.154777,
		0.170651,
		0.187172,
		0.204316,
		0.222058,
		0.24037,
		0.259225,
		0.278592,
		0.298442,
		0.318742,
		0.339457,
		0.360554,
		0.381994,
		0.403741,
		0.425753,
		0.447991,
		0.470412,
		0.492972,
		0.515624,
		0.538323,
		0.561021,
		0.583667,
		0.606211,
		0.628601,
		0.650784,
		0.672708,
		0.694316,
		0.715554,
		0.736367,
		0.756699,
		0.776495,
		0.795699,
		0.814257,
		0.832117,
		0.849227,
		0.865538,
		0.881004,
		0.895582,
		0.909232,
		0.921922,
		0.933622,
		0.944313,
		0.953979,
		0.962616,
		0.97023,
		0.976835,
		0.982461,
		0.987147,
		0.990947,
		0.99393,
		0.996177,
		0.997782,
		0.99885,
		0.999493,
		0.999827,
		0.999963,
		0.999998,
		1,
		0.999997,
		0.999959,
		0.999799,
		0.999381,
		0.998528,
		0.99703,
		0.994653,
		0.991144,
		0.986243,
		0.979683,
		0.971204,
		0.960553,
		0.947493,
		0.931803,
		0.913289,
		0.891784,
		0.867151,
		0.839291,
		0.808146,
		0.773702,
		0.735993,
		0.695109,
		0.651201,
		0.604481,
		0.555238,
		0.503836,
		0.450728,
		0.39646,
		0.341684,
		0.287164,
		0.233791,
		0.182595,
		0.134756,
		0.091624,
		0.054735,
		0.025825,
		0.006851,
		0.000471,
		0.001879,
		0.004219,
		0.007484,
		0.011667,
		0.016757,
		0.022747,
		0.029626,
		0.037384,
		0.046007,
		0.055483,
		0.065798,
		0.076937,
		0.088884,
		0.101621,
		0.11513,
		0.129392,
		0.144385,
		0.160089,
		0.176478,
		0.193529,
		0.211216,
		0.229511,
		0.248386,
		0.267809,
		0.28775,
		0.308175,
		0.32905,
		0.350337,
		0.371999,
		0.393997,
		0.416289,
		0.438833,
		0.461585,
		0.484498,
		0.507526,
		0.53062,
		0.553728,
		0.5768,
		0.599783,
		0.622621,
		0.645261,
		0.667644,
		0.689715,
		0.711414,
		0.732684,
		0.753466,
		0.773702,
		0.793332,
		0.812301,
		0.830552,
		0.848031,
		0.864686,
		0.880467,
		0.895329,
		0.909232,
		0.922139,
		0.934022,
		0.944857,
		0.95463,
		0.963338,
		0.970985,
		0.977591,
		0.983187,
		0.987816,
		0.991538,
		0.994427,
		0.996571,
		0.998073,
		0.999046,
		0.999608,
		0.999882,
		0.999981,
		0.999999,
		1,
		0.999991,
		0.99992,
		0.999675,
		0.999094,
		0.997978,
		0.996092,
		0.993181,
		0.988971,
		0.983187,
		0.975548,
		0.965784,
		0.953636,
		0.938862,
		0.921246,
		0.900598,
		0.876762,
		0.849616,
		0.819083,
		0.785132,
		0.747781,
		0.707107,
		0.663248,
		0.61641,
		0.566876,
		0.515011,
		0.461269,
		0.406206,
		0.350485,
		0.294891,
		0.240343,
		0.187904,
		0.138805,
		0.094456,
		0.056465,
		0.026654,
		0.007072,
		0.000488,
		0.001948,
		0.004374,
		0.007758,
		0.012092,
		0.017367,
		0.023574,
		0.0307,
		0.038736,
		0.047667,
		0.057479,
		0.068159,
		0.079688,
		0.092052,
		0.105229,
		0.119202,
		0.133949,
		0.149448,
		0.165674,
		0.182603,
		0.200209,
		0.218463,
		0.237335,
		0.256796,
		0.276811,
		0.297347,
		0.318369,
		0.339837,
		0.361715,
		0.383959,
		0.406528,
		0.429378,
		0.452462,
		0.475734,
		0.499142,
		0.522638,
		0.546167,
		0.569676,
		0.593109,
		0.61641,
		0.63952,
		0.66238,
		0.684929,
		0.707107,
		0.728851,
		0.750101,
		0.770794,
		0.790869,
		0.810266,
		0.828924,
		0.846787,
		0.863799,
		0.879909,
		0.895068,
		0.909232,
		0.922365,
		0.934434,
		0.945417,
		0.955299,
		0.964076,
		0.971756,
		0.978359,
		0.983919,
		0.988485,
		0.992122,
		0.994912,
		0.99695,
		0.998347,
		0.999223,
		0.999707,
		0.999924,
		0.999991,
		1,
		0.999999,
		0.999977,
		0.999854,
		0.999493,
		0.998706,
		0.997265,
		0.994912,
		0.991369,
		0.986346,
		0.979547,
		0.970681,
		0.959469,
		0.945646,
		0.928972,
		0.909232,
		0.886248,
		0.859876,
		0.830017,
		0.796619,
		0.759685,
		0.719275,
		0.675515,
		0.628601,
		0.578809,
		0.526502,
		0.472135,
		0.416269,
		0.359581,
		0.302876,
		0.247099,
		0.19336,
		0.142941,
		0.097323,
		0.058196,
		0.027471,
		0.007286,
		0.000506,
		0.002021,
		0.004537,
		0.008046,
		0.012541,
		0.018011,
		0.024446,
		0.031834,
		0.040162,
		0.049418,
		0.059585,
		0.070648,
		0.082589,
		0.095391,
		0.109032,
		0.123492,
		0.138748,
		0.154777,
		0.171552,
		0.189047,
		0.207232,
		0.226079,
		0.245554,
		0.265625,
		0.286256,
		0.307411,
		0.32905,
		0.351133,
		0.373618,
		0.39646,
		0.419614,
		0.443032,
		0.466665,
		0.49046,
		0.514364,
		0.538323,
		0.56228,
		0.586177,
		0.609954,
		0.63355,
		0.656903,
		0.679948,
		0.702622,
		0.72486,
		0.746597,
		0.767766,
		0.788304,
		0.808146,
		0.827229,
		0.845492,
		0.862877,
		0.879329,
		0.894796,
		0.909232,
		0.922598,
		0.93486,
		0.945995,
		0.955987,
		0.964833,
		0.972543,
		0.979138,
		0.984657,
		0.989154,
		0.9927,
		0.995385,
		0.997312,
		0.9986,
		0.999381,
		0.999788,
		0.999955,
		0.999997,
		1,
		0.999997,
		0.99995,
		0.999752,
		0.999236,
		0.998186,
		0.996348,
		0.993437,
		0.989154,
		0.983187,
		0.975226,
		0.964969,
		0.952128,
		0.936437,
		0.917657,
		0.895582,
		0.870045,
		0.840925,
		0.808146,
		0.771692,
		0.731605,
		0.687995,
		0.641049,
		0.591031,
		0.538297,
		0.483303,
		0.426612,
		0.368915,
		0.311037,
		0.253963,
		0.198852,
		0.147054,
		0.100129,
		0.059855,
		0.028234,
		0.00748,
		0.000525,
		0.002098,
		0.004709,
		0.008351,
		0.013016,
		0.018691,
		0.025367,
		0.033031,
		0.041669,
		0.051267,
		0.061808,
		0.073275,
		0.08565,
		0.098913,
		0.113042,
		0.128015,
		0.143807,
		0.160392,
		0.177742,
		0.19583,
		0.214622,
		0.234088,
		0.254193,
		0.2749,
		0.296172,
		0.317968,
		0.340247,
		0.362964,
		0.386075,
		0.409531,
		0.433283,
		0.457278,
		0.481465,
		0.505785,
		0.530184,
		0.5546,
		0.578973,
		0.603241,
		0.627338,
		0.651201,
		0.67476,
		0.69795,
		0.720701,
		0.742944,
		0.76461,
		0.78563,
		0.805937,
		0.825463,
		0.844144,
		0.861918,
		0.878726,
		0.894513,
		0.909232,
		0.92284,
		0.935301,
		0.946591,
		0.956695,
		0.965609,
		0.973345,
		0.979927,
		0.985399,
		0.989821,
		0.99327,
		0.995843,
		0.997655,
		0.998834,
		0.999519,
		0.999854,
		0.999976,
		0.999999,
		1,
		0.999989,
		0.9999,
		0.999596,
		0.998879,
		0.997501,
		0.99518,
		0.991607,
		0.986456,
		0.9794,
		0.970112,
		0.95828,
		0.943612,
		0.925841,
		0.904734,
		0.880097,
		0.851782,
		0.819691,
		0.783783,
		0.74408,
		0.700674,
		0.653732,
		0.603506,
		0.550341,
		0.494688,
		0.437117,
		0.378334,
		0.319197,
		0.260738,
		0.204182,
		0.15096,
		0.102721,
		0.061335,
		0.028884,
		0.007635,
		0.000546,
		0.002179,
		0.004891,
		0.008674,
		0.013517,
		0.019411,
		0.026342,
		0.034297,
		0.043262,
		0.053221,
		0.064157,
		0.076051,
		0.088884,
		0.102633,
		0.117276,
		0.132788,
		0.149143,
		0.166312,
		0.184267,
		0.202976,
		0.222405,
		0.242518,
		0.26328,
		0.28465,
		0.306588,
		0.32905,
		0.35199,
		0.375363,
		0.399117,
		0.423201,
		0.447562,
		0.472143,
		0.496887,
		0.521733,
		0.546619,
		0.571482,
		0.596255,
		0.620871,
		0.645261,
		0.669354,
		0.693079,
		0.716363,
		0.739133,
		0.761317,
		0.782841,
		0.803632,
		0.823621,
		0.842738,
		0.860918,
		0.878098,
		0.89422,
		0.909232,
		0.92309,
		0.935757,
		0.947206,
		0.957423,
		0.966404,
		0.974162,
		0.980727,
		0.986146,
		0.990485,
		0.993829,
		0.996286,
		0.997978,
		0.999046,
		0.999638,
		0.999906,
		0.999989,
		1,
		0.999999,
		0.999972,
		0.999819,
		0.999369,
		0.998392,
		0.996609,
		0.993706,
		0.989348,
		0.983187,
		0.974876,
		0.964076,
		0.950468,
		0.933754,
		0.913673,
		0.890001,
		0.862561,
		0.831226,
		0.795928,
		0.756663,
		0.713495,
		0.666569,
		0.616116,
		0.56247,
		0.50608,
		0.447525,
		0.38754,
		0.327031,
		0.267095,
		0.209038,
		0.154387,
		0.104887,
		0.062493,
		0.029348,
		0.007731,
		0.000567,
		0.002265,
		0.005084,
		0.009016,
		0.014049,
		0.020173,
		0.027373,
		0.035637,
		0.044948,
		0.055289,
		0.066642,
		0.078987,
		0.092302,
		0.106564,
		0.121749,
		0.137829,
		0.154777,
		0.172561,
		0.191151,
		0.210511,
		0.230606,
		0.251397,
		0.272844,
		0.294906,
		0.317536,
		0.340688,
		0.364313,
		0.38836,
		0.412775,
		0.437501,
		0.462481,
		0.487653,
		0.512956,
		0.538323,
		0.563688,
		0.588982,
		0.614133,
		0.639069,
		0.663715,
		0.687996,
		0.711836,
		0.735156,
		0.757879,
		0.779928,
		0.801227,
		0.821699,
		0.841272,
		0.859876,
		0.877444,
		0.893914,
		0.909232,
		0.92335,
		0.936229,
		0.947842,
		0.958172,
		0.967218,
		0.974996,
		0.981538,
		0.986896,
		0.991144,
		0.994377,
		0.996711,
		0.99828,
		0.999236,
		0.999738,
		0.999944,
		0.999996,
		1,
		0.999996,
		0.999937,
		0.999689,
		0.999046,
		0.99774,
		0.995458,
		0.991857,
		0.986574,
		0.97924,
		0.969489,
		0.956972,
		0.941362,
		0.922363,
		0.89972,
		0.873218,
		0.842698,
		0.808054,
		0.769247,
		0.726308,
		0.679351,
		0.628589,
		0.574344,
		0.517072,
		0.457378,
		0.396042,
		0.334039,
		0.272552,
		0.212988,
		0.156977,
		0.10636,
		0.06316,
		0.029542,
		0.007746,
		0.00059,
		0.002356,
		0.005289,
		0.009378,
		0.014613,
		0.02098,
		0.028466,
		0.037057,
		0.046734,
		0.057479,
		0.069274,
		0.082095,
		0.09592,
		0.110724,
		0.12648,
		0.14316,
		0.160731,
		0.179163,
		0.198419,
		0.218463,
		0.239256,
		0.260756,
		0.282919,
		0.3057,
		0.32905,
		0.352917,
		0.37725,
		0.40199,
		0.427082,
		0.452462,
		0.478069,
		0.503836,
		0.529695,
		0.555576,
		0.581406,
		0.607109,
		0.632611,
		0.657831,
		0.68269,
		0.707107,
		0.731,
		0.754286,
		0.776885,
		0.798713,
		0.819691,
		0.839741,
		0.858788,
		0.876762,
		0.893596,
		0.909232,
		0.92362,
		0.936718,
		0.948498,
		0.958943,
		0.968053,
		0.975845,
		0.982358,
		0.987648,
		0.991798,
		0.994912,
		0.997117,
		0.99856,
		0.999403,
		0.999819,
		0.99997,
		0.999999,
		1,
		0.999986,
		0.999875,
		0.999493,
		0.998595,
		0.996876,
		0.993987,
		0.989554,
		0.983187,
		0.974494,
		0.963094,
		0.948626,
		0.930757,
		0.909189,
		0.883665,
		0.853978,
		0.819979,
		0.781582,
		0.73878,
		0.691658,
		0.640413,
		0.585368,
		0.527004,
		0.465973,
		0.403127,
		0.339533,
		0.276483,
		0.215498,
		0.158311,
		0.106843,
		0.063155,
		0.029384,
		0.007659,
		0.000615,
		0.002453,
		0.005506,
		0.009762,
		0.015211,
		0.021837,
		0.029626,
		0.038563,
		0.048628,
		0.059802,
		0.072063,
		0.085389,
		0.099754,
		0.11513,
		0.13149,
		0.148801,
		0.16703,
		0.186142,
		0.2061,
		0.226862,
		0.248386,
		0.270627,
		0.293538,
		0.317068,
		0.341166,
		0.365774,
		0.390835,
		0.416289,
		0.442072,
		0.468117,
		0.494356,
		0.520718,
		0.547127,
		0.573509,
		0.599783,
		0.625869,
		0.651685,
		0.677145,
		0.702164,
		0.726654,
		0.750529,
		0.773702,
		0.796084,
		0.817592,
		0.838141,
		0.857652,
		0.87605,
		0.893264,
		0.909232,
		0.9239,
		0.937225,
		0.949176,
		0.959736,
		0.968908,
		0.976711,
		0.983187,
		0.988401,
		0.992444,
		0.995431,
		0.997501,
		0.998815,
		0.999548,
		0.999882,
		0.999986,
		1,
		0.999999,
		0.999964,
		0.999771,
		0.999205,
		0.997978,
		0.995745,
		0.992122,
		0.9867,
		0.979061,
		0.968789,
		0.955481,
		0.938756,
		0.918264,
		0.893693,
		0.864779,
		0.831316,
		0.793172,
		0.750302,
		0.702768,
		0.650766,
		0.594644,
		0.534931,
		0.472358,
		0.407876,
		0.34267,
		0.278157,
		0.215973,
		0.157944,
		0.106041,
		0.062312,
		0.028801,
		0.007452,
		0.000641,
		0.002556,
		0.005737,
		0.010171,
		0.015846,
		0.022747,
		0.030859,
		0.040162,
		0.050639,
		0.062267,
		0.075024,
		0.088884,
		0.103819,
		0.119801,
		0.136799,
		0.154777,
		0.1737,
		0.193529,
		0.214224,
		0.23574,
		0.258031,
		0.281048,
		0.304739,
		0.32905,
		0.353922,
		0.379296,
		0.405109,
		0.431293,
		0.457781,
		0.484498,
		0.511372,
		0.538323,
		0.565272,
		0.592135,
		0.618827,
		0.645261,
		0.671346,
		0.696992,
		0.722107,
		0.746597,
		0.770369,
		0.793332,
		0.815394,
		0.836467,
		0.856465,
		0.875306,
		0.892918,
		0.909232,
		0.924192,
		0.937751,
		0.949876,
		0.960553,
		0.969784,
		0.977591,
		0.984024,
		0.989154,
		0.993081,
		0.995933,
		0.997863,
		0.999046,
		0.999671,
		0.999929,
		0.999995,
		1,
		0.999995,
		0.99992,
		0.999606,
		0.998793,
		0.997146,
		0.994276,
		0.989757,
		0.983139,
		0.973965,
		0.961779,
		0.94614,
		0.926633,
		0.902878,
		0.874547,
		0.841376,
		0.803188,
		0.75991,
		0.711604,
		0.658487,
		0.600964,
		0.539646,
		0.475373,
		0.409225,
		0.342523,
		0.276811,
		0.213831,
		0.155471,
		0.103703,
		0.060501,
		0.027745,
		0.007118,
		0.000668,
		0.002666,
		0.005983,
		0.010606,
		0.016522,
		0.023716,
		0.032169,
		0.041863,
		0.052777,
		0.064888,
		0.07817,
		0.092596,
		0.108136,
		0.12476,
		0.142431,
		0.161115,
		0.18077,
		0.201355,
		0.222825,
		0.245133,
		0.268228,
		0.292057,
		0.316562,
		0.341684,
		0.367361,
		0.393526,
		0.42011,
		0.447041,
		0.474244,
		0.501639,
		0.529145,
		0.556677,
		0.584147,
		0.611467,
		0.638541,
		0.665277,
		0.691577,
		0.717343,
		0.742476,
		0.766878,
		0.790449,
		0.813093,
		0.834714,
		0.855222,
		0.874529,
		0.892557,
		0.909232,
		0.924495,
		0.938296,
		0.950601,
		0.961395,
		0.970681,
		0.978488,
		0.984868,
		0.989904,
		0.993706,
		0.996415,
		0.9982,
		0.999251,
		0.999771,
		0.999962,
		0.999999,
		1,
		0.999982,
		0.999839,
		0.999349,
		0.998198,
		0.995994,
		0.99229,
		0.986599,
		0.97841,
		0.967203,
		0.952465,
		0.933706,
		0.910471,
		0.882368,
		0.849081,
		0.810403,
		0.766258,
		0.716734,
		0.662108,
		0.602874,
		0.539764,
		0.473757,
		0.406079,
		0.338191,
		0.271753,
		0.208583,
		0.150584,
		0.099666,
		0.057659,
		0.026203,
		0.006658,
		0.000697,
		0.002783,
		0.006244,
		0.011069,
		0.017243,
		0.024747,
		0.033565,
		0.043674,
		0.055053,
		0.067676,
		0.081516,
		0.096544,
		0.112726,
		0.130029,
		0.148414,
		0.167843,
		0.188271,
		0.209654,
		0.231941,
		0.255081,
		0.279019,
		0.303696,
		0.32905,
		0.355015,
		0.381525,
		0.408505,
		0.43588,
		0.463572,
		0.491497,
		0.51957,
		0.547701,
		0.575799,
		0.603767,
		0.631507,
		0.65892,
		0.685901,
		0.712348,
		0.738155,
		0.763216,
		0.787425,
		0.81068,
		0.832877,
		0.85392,
		0.873716,
		0.892179,
		0.909232,
		0.924811,
		0.938862,
		0.951351,
		0.962262,
		0.971601,
		0.9794,
		0.985719,
		0.99065,
		0.994317,
		0.996876,
		0.998511,
		0.999429,
		0.99985,
		0.999982,
		0.999998,
		0.999993,
		0.999936,
		0.999659,
		0.998871,
		0.997181,
		0.994116,
		0.989142,
		0.981687,
		0.971156,
		0.956957,
		0.938519,
		0.91532,
		0.88691,
		0.852944,
		0.81321,
		0.767663,
		0.716456,
		0.659966,
		0.598814,
		0.533882,
		0.466305,
		0.397462,
		0.328938,
		0.26248,
		0.199929,
		0.143141,
		0.093895,
		0.053801,
		0.024201,
		0.006083,
		0.000729,
		0.002907,
		0.006524,
		0.011564,
		0.018011,
		0.025848,
		0.035053,
		0.045605,
		0.057479,
		0.070648,
		0.085081,
		0.100747,
		0.117611,
		0.135635,
		0.154777,
		0.174994,
		0.19624,
		0.218463,
		0.24161,
		0.265625,
		0.290446,
		0.31601,
		0.342249,
		0.369091,
		0.39646,
		0.424278,
		0.452462,
		0.480925,
		0.509577,
		0.538323,
		0.567067,
		0.595706,
		0.624138,
		0.652255,
		0.679948,
		0.707107,
		0.733619,
		0.759371,
		0.78425,
		0.808146,
		0.830949,
		0.852555,
		0.872864,
		0.891784,
		0.909232,
		0.92514,
		0.939451,
		0.952128,
		0.963156,
		0.972543,
		0.980326,
		0.986574,
		0.99139,
		0.994912,
		0.997312,
		0.998793,
		0.99958,
		0.999904,
		0.999976,
		0.999952,
		0.99988,
		0.999655,
		0.999017,
		0.997575,
		0.994828,
		0.990193,
		0.983029,
		0.972665,
		0.95843,
		0.939685,
		0.915851,
		0.886448,
		0.851131,
		0.809723,
		0.76225,
		0.708972,
		0.650402,
		0.58732,
		0.520775,
		0.45206,
		0.382687,
		0.314334,
		0.24878,
		0.187825,
		0.133206,
		0.086499,
		0.049033,
		0.021807,
		0.005415,
		0.000762,
		0.003041,
		0.006822,
		0.012092,
		0.018832,
		0.027023,
		0.036642,
		0.047667,
		0.060068,
		0.073818,
		0.088884,
		0.105229,
		0.122817,
		0.141606,
		0.161551,
		0.182603,
		0.204713,
		0.227824,
		0.251877,
		0.276811,
		0.302559,
		0.32905,
		0.356209,
		0.383959,
		0.412216,
		0.440894,
		0.469901,
		0.499142,
		0.528519,
		0.557927,
		0.587261,
		0.61641,
		0.645261,
		0.673697,
		0.701601,
		0.728851,
		0.755329,
		0.780913,
		0.805484,
		0.828924,
		0.851122,
		0.87197,
		0.89137,
		0.909232,
		0.925483,
		0.940063,
		0.952932,
		0.964076,
		0.973507,
		0.981267,
		0.987433,
		0.992122,
		0.995487,
		0.997716,
		0.999028,
		0.999659,
		0.999837,
		0.999756,
		0.999523,
		0.9991,
		0.998268,
		0.996642,
		0.993699,
		0.988808,
		0.981267,
		0.970337,
		0.955287,
		0.935426,
		0.910151,
		0.878988,
		0.841631,
		0.797981,
		0.748179,
		0.692627,
		0.632006,
		0.567267,
		0.499619,
		0.430498,
		0.361507,
		0.294361,
		0.230803,
		0.172521,
		0.121059,
		0.077731,
		0.043541,
		0.019122,
		0.004685,
		0.000798,
		0.003183,
		0.007142,
		0.012657,
		0.01971,
		0.02828,
		0.038342,
		0.049871,
		0.062836,
		0.077206,
		0.092945,
		0.110014,
		0.128373,
		0.147975,
		0.168771,
		0.19071,
		0.213733,
		0.237781,
		0.26279,
		0.28869,
		0.315408,
		0.342867,
		0.370984,
		0.399673,
		0.428844,
		0.4584,
		0.488241,
		0.518262,
		0.548355,
		0.578407,
		0.608301,
		0.637915,
		0.667127,
		0.69581,
		0.723836,
		0.751077,
		0.777401,
		0.802682,
		0.826794,
		0.849616,
		0.871032,
		0.890936,
		0.909232,
		0.925841,
		0.940699,
		0.953766,
		0.965026,
		0.974494,
		0.982219,
		0.988288,
		0.992826,
		0.995994,
		0.997988,
		0.999026,
		0.99933,
		0.9991,
		0.998474,
		0.997473,
		0.99593,
		0.993485,
		0.989607,
		0.983637,
		0.974831,
		0.96241,
		0.945607,
		0.923719,
		0.896157,
		0.862495,
		0.822512,
		0.77623,
		0.723937,
		0.666205,
		0.603886,
		0.538097,
		0.470186,
		0.401677,
		0.334211,
		0.269468,
		0.209081,
		0.154555,
		0.107181,
		0.067966,
		0.037571,
		0.01627,
		0.003928,
		0.000836,
		0.003336,
		0.007484,
		0.013263,
		0.020651,
		0.029626,
		0.040162,
		0.05223,
		0.065798,
		0.08083,
		0.097288,
		0.11513,
		0.13431,
		0.154777,
		0.176478,
		0.199356,
		0.223347,
		0.248386,
		0.2744,
		0.301315,
		0.32905,
		0.357518,
		0.38663,
		0.416289,
		0.446396,
		0.476846,
		0.507526,
		0.538323,
		0.569117,
		0.599783,
		0.630193,
		0.660215,
		0.689715,
		0.718555,
		0.746597,
		0.773702,
		0.799732,
		0.824552,
		0.848031,
		0.870045,
		0.89048,
		0.909232,
		0.926215,
		0.941362,
		0.954629,
		0.965998,
		0.975486,
		0.983139,
		0.989042,
		0.99331,
		0.996096,
		0.997575,
		0.997934,
		0.997356,
		0.995982,
		0.993865,
		0.990906,
		0.986797,
		0.981036,
		0.97297,
		0.961852,
		0.946897,
		0.92735,
		0.902538,
		0.871933,
		0.835202,
		0.792248,
		0.743241,
		0.68864,
		0.629185,
		0.565889,
		0.5,
		0.43295,
		0.366296,
		0.30164,
		0.240551,
		0.184485,
		0.134701,
		0.092199,
		0.057659,
		0.031406,
		0.013387,
		0.003178,
		0.000878,
		0.003501,
		0.007852,
		0.013913,
		0.021661,
		0.031071,
		0.042115,
		0.054761,
		0.068974,
		0.084714,
		0.101941,
		0.120607,
		0.140662,
		0.162051,
		0.184715,
		0.208589,
		0.233606,
		0.259691,
		0.286766,
		0.314747,
		0.343545,
		0.373065,
		0.403207,
		0.433865,
		0.464929,
		0.496282,
		0.527802,
		0.559361,
		0.590828,
		0.622067,
		0.652936,
		0.683291,
		0.712986,
		0.741872,
		0.769799,
		0.796619,
		0.822187,
		0.846361,
		0.869007,
		0.890001,
		0.909231,
		0.926602,
		0.942036,
		0.955481,
		0.966903,
		0.9763,
		0.983695,
		0.989142,
		0.992725,
		0.994554,
		0.994761,
		0.993485,
		0.99085,
		0.986929,
		0.981687,
		0.974912,
		0.966179,
		0.954896,
		0.940363,
		0.921848,
		0.898653,
		0.870185,
		0.836021,
		0.795954,
		0.750038,
		0.698606,
		0.642277,
		0.581942,
		0.518731,
		0.453969,
		0.389109,
		0.325662,
		0.265117,
		0.208867,
		0.158129,
		0.113883,
		0.076818,
		0.047295,
		0.025332,
		0.010604,
		0.002468,
		0.000922,
		0.003677,
		0.008248,
		0.014613,
		0.022747,
		0.032625,
		0.044213,
		0.057479,
		0.072383,
		0.088884,
		0.106933,
		0.12648,
		0.14747,
		0.169842,
		0.193529,
		0.218463,
		0.244566,
		0.271757,
		0.299949,
		0.32905,
		0.358959,
		0.389573,
		0.420779,
		0.452462,
		0.484498,
		0.516759,
		0.549108,
		0.581406,
		0.613507,
		0.645261,
		0.676513,
		0.707107,
		0.736881,
		0.765677,
		0.793332,
		0.819691,
		0.844598,
		0.867908,
		0.889481,
		0.909189,
		0.926912,
		0.942546,
		0.956001,
		0.967203,
		0.9761,
		0.982663,
		0.986889,
		0.988808,
		0.988479,
		0.985989,
		0.981438,
		0.974912,
		0.966443,
		0.955943,
		0.943133,
		0.927549,
		0.908605,
		0.885676,
		0.858176,
		0.825636,
		0.787765,
		0.744506,
		0.696065,
		0.642924,
		0.585835,
		0.525792,
		0.463986,
		0.401745,
		0.34047,
		0.281552,
		0.226305,
		0.175889,
		0.131251,
		0.093074,
		0.061746,
		0.037339,
		0.019612,
		0.008036,
		0.001827,
		0.00097,
		0.003868,
		0.008674,
		0.015366,
		0.023917,
		0.034297,
		0.046472,
		0.060404,
		0.076051,
		0.093366,
		0.112297,
		0.132788,
		0.154777,
		0.178197,
		0.202976,
		0.229035,
		0.25629,
		0.28465,
		0.314019,
		0.344293,
		0.375363,
		0.407111,
		0.439414,
		0.472143,
		0.505161,
		0.538323,
		0.571482,
		0.604481,
		0.63716,
		0.669354,
		0.700893,
		0.731605,
		0.761316,
		0.789852,
		0.817037,
		0.842698,
		0.86666,
		0.888751,
		0.908799,
		0.926633,
		0.942087,
		0.955004,
		0.965238,
		0.972665,
		0.977186,
		0.978745,
		0.977327,
		0.97297,
		0.965756,
		0.955795,
		0.943198,
		0.928019,
		0.910182,
		0.889428,
		0.865353,
		0.837497,
		0.805429,
		0.768831,
		0.727561,
		0.681705,
		0.631602,
		0.577846,
		0.52127,
		0.462908,
		0.403941,
		0.34563,
		0.289251,
		0.236018,
		0.187018,
		0.143149,
		0.105078,
		0.073201,
		0.047631,
		0.028199,
		0.014469,
		0.005777,
		0.001275,
		0.001021,
		0.004073,
		0.009134,
		0.016179,
		0.025179,
		0.036101,
		0.048908,
		0.063557,
		0.080003,
		0.098193,
		0.118071,
		0.139573,
		0.162631,
		0.187172,
		0.213114,
		0.24037,
		0.268846,
		0.298442,
		0.32905,
		0.360554,
		0.392831,
		0.425753,
		0.459182,
		0.492972,
		0.526971,
		0.561021,
		0.594955,
		0.628601,
		0.661781,
		0.694312,
		0.726004,
		0.756663,
		0.786085,
		0.81406,
		0.840368,
		0.864779,
		0.887051,
		0.906936,
		0.924176,
		0.938519,
		0.949718,
		0.95755,
		0.961826,
		0.96241,
		0.95923,
		0.95229,
		0.941677,
		0.927549,
		0.910115,
		0.889587,
		0.86612,
		0.839721,
		0.810233,
		0.777408,
		0.740998,
		0.700845,
		0.656947,
		0.609506,
		0.558952,
		0.505937,
		0.451312,
		0.39608,
		0.341342,
		0.288229,
		0.237836,
		0.191158,
		0.14903,
		0.112084,
		0.080714,
		0.05506,
		0.035006,
		0.020197,
		0.010067,
		0.00389,
		0.000827,
		0.001077,
		0.004296,
		0.009632,
		0.017058,
		0.026543,
		0.038051,
		0.051539,
		0.066963,
		0.084269,
		0.103402,
		0.124297,
		0.146884,
		0.171089,
		0.196828,
		0.22401,
		0.25254,
		0.282312,
		0.313213,
		0.345123,
		0.377913,
		0.411446,
		0.445578,
		0.480153,
		0.515011,
		0.549981,
		0.584885,
		0.619534,
		0.653732,
		0.687268,
		0.719918,
		0.751442,
		0.781582,
		0.810056,
		0.836564,
		0.860781,
		0.882368,
		0.900971,
		0.916238,
		0.927827,
		0.935426,
		0.938773,
		0.937675,
		0.932032,
		0.921848,
		0.907247,
		0.888462,
		0.865825,
		0.839721,
		0.810531,
		0.778552,
		0.743918,
		0.706621,
		0.666607,
		0.623865,
		0.578504,
		0.530799,
		0.481216,
		0.430401,
		0.379157,
		0.328396,
		0.279078,
		0.232157,
		0.188515,
		0.148905,
		0.113912,
		0.08391,
		0.059053,
		0.039266,
		0.024257,
		0.013546,
		0.006505,
		0.002407,
		0.000487,
		0.001138,
		0.004537,
		0.010171,
		0.018011,
		0.028021,
		0.040162,
		0.054388,
		0.070648,
		0.088884,
		0.109032,
		0.131022,
		0.154777,
		0.180211,
		0.207232,
		0.23574,
		0.265625,
		0.29677,
		0.32905,
		0.362328,
		0.39646,
		0.431293,
		0.466662,
		0.502393,
		0.538297,
		0.574175,
		0.609807,
		0.644953,
		0.679351,
		0.712711,
		0.744711,
		0.774998,
		0.803188,
		0.828867,
		0.8516,
		0.870942,
		0.886448,
		0.897699,
		0.904319,
		0.906002,
		0.902538,
		0.893836,
		0.879941,
		0.861048,
		0.837497,
		0.809757,
		0.778389,
		0.743988,
		0.707107,
		0.66817,
		0.627422,
		0.584989,
		0.540983,
		0.495581,
		0.449081,
		0.401925,
		0.354699,
		0.308099,
		0.262895,
		0.219874,
		0.179781,
		0.143273,
		0.110866,
		0.082904,
		0.05954,
		0.040725,
		0.026217,
		0.015607,
		0.008349,
		0.003813,
		0.001329,
		0.00025,
		0.001204,
		0.004799,
		0.010757,
		0.019046,
		0.029626,
		0.042454,
		0.057479,
		0.074644,
		0.093885,
		0.11513,
		0.138302,
		0.163312,
		0.190067,
		0.218463,
		0.248386,
		0.279714,
		0.312315,
		0.346046,
		0.380753,
		0.416269,
		0.452411,
		0.488978,
		0.525747,
		0.56247,
		0.598868,
		0.634625,
		0.66939,
		0.702768,
		0.734326,
		0.76359,
		0.790059,
		0.81321,
		0.832516,
		0.847468,
		0.857596,
		0.862495,
		0.861859,
		0.855502,
		0.843385,
		0.825636,
		0.80255,
		0.774593,
		0.742376,
		0.706621,
		0.668102,
		0.627576,
		0.585696,
		0.542935,
		0.499583,
		0.455837,
		0.4119,
		0.368043,
		0.32464,
		0.282165,
		0.241175,
		0.202269,
		0.166038,
		0.133017,
		0.103642,
		0.078208,
		0.056853,
		0.039541,
		0.026068,
		0.016081,
		0.009106,
		0.004591,
		0.001951,
		0.000623,
		0.000105,
		0.001275,
		0.005084,
		0.011395,
		0.020173,
		0.031373,
		0.044948,
		0.06084,
		0.078987,
		0.099316,
		0.121749,
		0.146196,
		0.172561,
		0.200736,
		0.230604,
		0.262037,
		0.294891,
		0.329012,
		0.364225,
		0.400334,
		0.437117,
		0.474322,
		0.51166,
		0.548801,
		0.585368,
		0.620939,
		0.655041,
		0.687158,
		0.716734,
		0.743188,
		0.765928,
		0.784374,
		0.797981,
		0.806267,
		0.808844,
		0.805445,
		0.795954,
		0.780424,
		0.759087,
		0.732363,
		0.700845,
		0.665273,
		0.626498,
		0.585423,
		0.542935,
		0.499836,
		0.456759,
		0.414121,
		0.372177,
		0.331119,
		0.291157,
		0.252561,
		0.215674,
		0.180892,
		0.148632,
		0.119294,
		0.09321,
		0.070615,
		0.051615,
		0.036175,
		0.024116,
		0.015128,
		0.008796,
		0.004637,
		0.002141,
		0.000815,
		0.000225,
		3.1e-005,
		0.001354,
		0.005396,
		0.012092,
		0.021402,
		0.033279,
		0.047667,
		0.064503,
		0.083718,
		0.105229,
		0.128948,
		0.154774,
		0.182595,
		0.212284,
		0.243699,
		0.276679,
		0.311037,
		0.346557,
		0.382989,
		0.420042,
		0.457378,
		0.494607,
		0.531288,
		0.566923,
		0.600964,
		0.632819,
		0.661863,
		0.687459,
		0.708972,
		0.7258,
		0.737402,
		0.743327,
		0.743241,
		0.736961,
		0.724468,
		0.705931,
		0.681705,
		0.652329,
		0.618508,
		0.581082,
		0.540983,
		0.499182,
		0.456627,
		0.414176,
		0.372533,
		0.332194,
		0.293444,
		0.256446,
		0.22133,
		0.188251,
		0.15741,
		0.129044,
		0.103399,
		0.080692,
		0.061078,
		0.04462,
		0.031269,
		0.020858,
		0.01311,
		0.007659,
		0.00408,
		0.001926,
		0.00077,
		0.000242,
		5.1e-005,
		0.00144,
		0.005737,
		0.012854,
		0.022747,
		0.035363,
		0.050639,
		0.068505,
		0.088879,
		0.111669,
		0.136765,
		0.164044,
		0.19336,
		0.224539,
		0.257379,
		0.291638,
		0.327031,
		0.36322,
		0.399815,
		0.436364,
		0.472358,
		0.507231,
		0.540369,
		0.57112,
		0.598814,
		0.622781,
		0.642378,
		0.65702,
		0.666205,
		0.669547,
		0.666804,
		0.657896,
		0.642924,
		0.622176,
		0.596121,
		0.565401,
		0.530799,
		0.493212,
		0.4536,
		0.412943,
		0.372177,
		0.332146,
		0.293546,
		0.25688,
		0.222444,
		0.190361,
		0.160683,
		0.133455,
		0.108744,
		0.086641,
		0.067237,
		0.050596,
		0.036721,
		0.025537,
		0.016877,
		0.010488,
		0.006041,
		0.003161,
		0.00146,
		0.000568,
		0.000173,
		3.6e-005,
		0.001534,
		0.006111,
		0.013691,
		0.024222,
		0.037644,
		0.053884,
		0.072856,
		0.094456,
		0.118558,
		0.14501,
		0.173627,
		0.204182,
		0.236406,
		0.269972,
		0.304496,
		0.339533,
		0.374572,
		0.409045,
		0.442328,
		0.473757,
		0.502643,
		0.528295,
		0.550042,
		0.567267,
		0.579429,
		0.586099,
		0.586982,
		0.581942,
		0.571013,
		0.554413,
		0.532534,
		0.505937,
		0.475331,
		0.441538,
		0.405462,
		0.368043,
		0.330213,
		0.292845,
		0.256712,
		0.222444,
		0.190498,
		0.161142,
		0.134466,
		0.110452,
		0.08905,
		0.070213,
		0.053909,
		0.040099,
		0.02872,
		0.019659,
		0.012738,
		0.007717,
		0.004297,
		0.002146,
		0.000927,
		0.000326,
		8.3e-005,
		1.2e-005,
		0.001637,
		0.006522,
		0.014606,
		0.025825,
		0.040101,
		0.057334,
		0.077397,
		0.100129,
		0.125331,
		0.152754,
		0.182095,
		0.212988,
		0.245006,
		0.277651,
		0.310362,
		0.342523,
		0.373469,
		0.402506,
		0.428931,
		0.45206,
		0.471247,
		0.485925,
		0.495623,
		0.5,
		0.498862,
		0.492178,
		0.48009,
		0.462908,
		0.441107,
		0.415302,
		0.386226,
		0.354699,
		0.321588,
		0.287772,
		0.254094,
		0.22133,
		0.190146,
		0.161075,
		0.134491,
		0.110603,
		0.089462,
		0.070995,
		0.055078,
		0.041577,
		0.030358,
		0.021286,
		0.014202,
		0.00891,
		0.005174,
		0.002719,
		0.001249,
		0.000475,
		0.000135,
		2.3e-005,
		0.001749,
		0.006962,
		0.015567,
		0.027471,
		0.042546,
		0.060626,
		0.081496,
		0.104887,
		0.130464,
		0.157829,
		0.186511,
		0.215973,
		0.245614,
		0.274783,
		0.302792,
		0.328938,
		0.352525,
		0.372893,
		0.389445,
		0.401677,
		0.409197,
		0.411755,
		0.40925,
		0.401745,
		0.389462,
		0.37278,
		0.352213,
		0.328396,
		0.302048,
		0.273949,
		0.244896,
		0.215674,
		0.18702,
		0.159588,
		0.133926,
		0.110452,
		0.089443,
		0.071031,
		0.055213,
		0.041877,
		0.030845,
		0.021916,
		0.014891,
		0.009563,
		0.005716,
		0.00311,
		0.001491,
		0.000596,
		0.00018,
		3.3e-005,
		0.001862,
		0.007387,
		0.01645,
		0.028884,
		0.044463,
		0.062899,
		0.083836,
		0.106843,
		0.131419,
		0.156993,
		0.182939,
		0.208583,
		0.23323,
		0.256183,
		0.276767,
		0.294361,
		0.308422,
		0.318512,
		0.324316,
		0.325662,
		0.322525,
		0.315035,
		0.303466,
		0.288229,
		0.269848,
		0.248937,
		0.226175,
		0.202269,
		0.177929,
		0.153832,
		0.130593,
		0.108744,
		0.088707,
		0.070788,
		0.05516,
		0.041877,
		0.030877,
		0.022008,
		0.015054,
		0.00978,
		0.005944,
		0.003307,
		0.001631,
		0.000676,
		0.000213,
		4.1e-005,
		0.001953,
		0.007692,
		0.016988,
		0.029542,
		0.04498,
		0.062848,
		0.08262,
		0.103703,
		0.125451,
		0.147183,
		0.168202,
		0.187825,
		0.205406,
		0.220363,
		0.232205,
		0.240551,
		0.24515,
		0.245885,
		0.242785,
		0.236018,
		0.22588,
		0.212781,
		0.197224,
		0.179781,
		0.161065,
		0.141698,
		0.122288,
		0.103399,
		0.085527,
		0.069081,
		0.054366,
		0.041577,
		0.030794,
		0.021994,
		0.015059,
		0.0098,
		0.005985,
		0.003361,
		0.001682,
		0.000712,
		0.000231,
		4.6e-005,
		0.001983,
		0.007716,
		0.016812,
		0.028801,
		0.043137,
		0.05921,
		0.076359,
		0.093895,
		0.111127,
		0.127383,
		0.142042,
		0.154555,
		0.164471,
		0.171451,
		0.175283,
		0.175889,
		0.17332,
		0.167757,
		0.159489,
		0.148905,
		0.136466,
		0.122681,
		0.108085,
		0.09321,
		0.078559,
		0.064584,
		0.051666,
		0.040099,
		0.03008,
		0.021703,
		0.014966,
		0.00978,
		0.005983,
		0.003364,
		0.001689,
		0.000719,
		0.000236,
		4.8e-005,
		0.00191,
		0.007301,
		0.015609,
		0.026203,
		0.038407,
		0.051521,
		0.064851,
		0.077731,
		0.089552,
		0.099785,
		0.108001,
		0.113883,
		0.11724,
		0.118003,
		0.116229,
		0.112084,
		0.105836,
		0.097831,
		0.088475,
		0.078208,
		0.067482,
		0.056733,
		0.046365,
		0.036721,
		0.028075,
		0.020615,
		0.014441,
		0.009563,
		0.005911,
		0.003347,
		0.001686,
		0.000719,
		0.000237,
		4.9e-005,
		0.001708,
		0.006383,
		0.013323,
		0.021807,
		0.031123,
		0.040599,
		0.04962,
		0.057659,
		0.064291,
		0.069201,
		0.072196,
		0.073201,
		0.072255,
		0.0695,
		0.065163,
		0.05954,
		0.052975,
		0.045834,
		0.038483,
		0.031269,
		0.024496,
		0.018409,
		0.013182,
		0.00891,
		0.005612,
		0.003231,
		0.001651,
		0.000712,
		0.000236,
		4.9e-005,
		0.001391,
		0.005056,
		0.01025,
		0.01627,
		0.022482,
		0.02834,
		0.033402,
		0.037339,
		0.039939,
		0.041104,
		0.040846,
		0.039266,
		0.036547,
		0.032928,
		0.028687,
		0.024116,
		0.019499,
		0.015095,
		0.011116,
		0.007717,
		0.004985,
		0.002939,
		0.001535,
		0.000676,
		0.000228,
		4.8e-005,
		0.00101,
		0.00355,
		0.006942,
		0.010604,
		0.014062,
		0.016957,
		0.019045,
		0.020197,
		0.020382,
		0.019662,
		0.018168,
		0.016081,
		0.013614,
		0.010985,
		0.008402,
		0.006041,
		0.004032,
		0.002452,
		0.001319,
		0.000596,
		0.000206,
		4.4e-005,
		0.000636,
		0.002137,
		0.003985,
		0.005777,
		0.007231,
		0.008178,
		0.008546,
		0.008349,
		0.007669,
		0.006631,
		0.005384,
		0.00408,
		0.002854,
		0.001811,
		0.001013,
		0.000475,
		0.00017,
		3.8e-005,
		0.000329,
		0.001038,
		0.001801,
		0.002407,
		0.002744,
		0.002784,
		0.002558,
		0.002141,
		0.001626,
		0.001107,
		0.000658,
		0.000326,
		0.000122,
		2.8e-005,
		0.000126,
		0.000356,
		0.000544,
		0.000623,
		0.000588,
		0.00047,
		0.000316,
		0.000173,
		7.1e-005,
		1.8e-005,
		2.7e-005,
		6.1e-005,
		6.8e-005,
		5.1e-005,
		2.6e-005,
		2.6e-005,
		5.1e-005,
		6.8e-005,
		6.1e-005,
		2.7e-005,
		1.8e-005,
		7.1e-005,
		0.000173,
		0.000316,
		0.00047,
		0.000588,
		0.000623,
		0.000544,
		0.000356,
		0.000126,
		2.8e-005,
		0.000122,
		0.000326,
		0.000658,
		0.001107,
		0.001626,
		0.002141,
		0.002558,
		0.002784,
		0.002744,
		0.002407,
		0.001801,
		0.001038,
		0.000329,
		3.8e-005,
		0.00017,
		0.000475,
		0.001013,
		0.001811,
		0.002854,
		0.00408,
		0.005384,
		0.006631,
		0.007669,
		0.008349,
		0.008546,
		0.008178,
		0.007231,
		0.005777,
		0.003985,
		0.002137,
		0.000636,
		4.4e-005,
		0.000206,
		0.000596,
		0.001319,
		0.002452,
		0.004032,
		0.006041,
		0.008402,
		0.010985,
		0.013614,
		0.016081,
		0.018168,
		0.019662,
		0.020382,
		0.020197,
		0.019045,
		0.016957,
		0.014062,
		0.010604,
		0.006942,
		0.00355,
		0.00101,
		4.8e-005,
		0.000228,
		0.000676,
		0.001535,
		0.002939,
		0.004985,
		0.007717,
		0.011116,
		0.015095,
		0.019499,
		0.024116,
		0.028687,
		0.032928,
		0.036547,
		0.039266,
		0.040846,
		0.041104,
		0.039939,
		0.037339,
		0.033402,
		0.02834,
		0.022482,
		0.01627,
		0.01025,
		0.005056,
		0.001391,
		4.9e-005,
		0.000236,
		0.000712,
		0.001651,
		0.003231,
		0.005612,
		0.00891,
		0.013182,
		0.018409,
		0.024496,
		0.031269,
		0.038483,
		0.045834,
		0.052975,
		0.05954,
		0.065163,
		0.0695,
		0.072255,
		0.073201,
		0.072196,
		0.069201,
		0.064291,
		0.057659,
		0.04962,
		0.040599,
		0.031123,
		0.021807,
		0.013323,
		0.006383,
		0.001708,
		4.9e-005,
		0.000237,
		0.000719,
		0.001686,
		0.003347,
		0.005911,
		0.009563,
		0.014441,
		0.020615,
		0.028075,
		0.036721,
		0.046365,
		0.056733,
		0.067482,
		0.078208,
		0.088475,
		0.097831,
		0.105836,
		0.112084,
		0.116229,
		0.118003,
		0.11724,
		0.113883,
		0.108001,
		0.099785,
		0.089552,
		0.077731,
		0.064851,
		0.051521,
		0.038407,
		0.026203,
		0.015609,
		0.007301,
		0.00191,
		4.8e-005,
		0.000236,
		0.000719,
		0.001689,
		0.003364,
		0.005983,
		0.00978,
		0.014966,
		0.021703,
		0.03008,
		0.040099,
		0.051666,
		0.064584,
		0.078559,
		0.09321,
		0.108085,
		0.122681,
		0.136466,
		0.148905,
		0.159489,
		0.167757,
		0.17332,
		0.175889,
		0.175283,
		0.171451,
		0.164471,
		0.154555,
		0.142042,
		0.127383,
		0.111127,
		0.093895,
		0.076359,
		0.05921,
		0.043137,
		0.028801,
		0.016812,
		0.007716,
		0.001983,
		4.6e-005,
		0.000231,
		0.000712,
		0.001682,
		0.003361,
		0.005985,
		0.0098,
		0.015059,
		0.021994,
		0.030794,
		0.041577,
		0.054366,
		0.069081,
		0.085527,
		0.103399,
		0.122288,
		0.141698,
		0.161065,
		0.179781,
		0.197224,
		0.212781,
		0.22588,
		0.236018,
		0.242785,
		0.245885,
		0.24515,
		0.240551,
		0.232205,
		0.220363,
		0.205406,
		0.187825,
		0.168202,
		0.147183,
		0.125451,
		0.103703,
		0.08262,
		0.062848,
		0.04498,
		0.029542,
		0.016988,
		0.007692,
		0.001953,
		4.1e-005,
		0.000213,
		0.000676,
		0.001631,
		0.003307,
		0.005944,
		0.00978,
		0.015054,
		0.022008,
		0.030877,
		0.041877,
		0.05516,
		0.070788,
		0.088707,
		0.108744,
		0.130593,
		0.153832,
		0.177929,
		0.202269,
		0.226175,
		0.248937,
		0.269848,
		0.288229,
		0.303466,
		0.315035,
		0.322525,
		0.325662,
		0.324316,
		0.318512,
		0.308422,
		0.294361,
		0.276767,
		0.256183,
		0.23323,
		0.208583,
		0.182939,
		0.156993,
		0.131419,
		0.106843,
		0.083836,
		0.062899,
		0.044463,
		0.028884,
		0.01645,
		0.007387,
		0.001862,
		3.3e-005,
		0.00018,
		0.000596,
		0.001491,
		0.00311,
		0.005716,
		0.009563,
		0.014891,
		0.021916,
		0.030845,
		0.041877,
		0.055213,
		0.071031,
		0.089443,
		0.110452,
		0.133926,
		0.159588,
		0.18702,
		0.215674,
		0.244896,
		0.273949,
		0.302048,
		0.328396,
		0.352213,
		0.37278,
		0.389462,
		0.401745,
		0.40925,
		0.411755,
		0.409197,
		0.401677,
		0.389445,
		0.372893,
		0.352525,
		0.328938,
		0.302792,
		0.274783,
		0.245614,
		0.215973,
		0.186511,
		0.157829,
		0.130464,
		0.104887,
		0.081496,
		0.060626,
		0.042546,
		0.027471,
		0.015567,
		0.006962,
		0.001749,
		2.3e-005,
		0.000135,
		0.000475,
		0.001249,
		0.002719,
		0.005174,
		0.00891,
		0.014202,
		0.021286,
		0.030358,
		0.041577,
		0.055078,
		0.070995,
		0.089462,
		0.110603,
		0.134491,
		0.161075,
		0.190146,
		0.22133,
		0.254094,
		0.287772,
		0.321588,
		0.354699,
		0.386226,
		0.415302,
		0.441107,
		0.462908,
		0.48009,
		0.492178,
		0.498862,
		0.5,
		0.495623,
		0.485925,
		0.471247,
		0.45206,
		0.428931,
		0.402506,
		0.373469,
		0.342523,
		0.310362,
		0.277651,
		0.245006,
		0.212988,
		0.182095,
		0.152754,
		0.125331,
		0.100129,
		0.077397,
		0.057334,
		0.040101,
		0.025825,
		0.014606,
		0.006522,
		0.001637,
		1.2e-005,
		8.3e-005,
		0.000326,
		0.000927,
		0.002146,
		0.004297,
		0.007717,
		0.012738,
		0.019659,
		0.02872,
		0.040099,
		0.053909,
		0.070213,
		0.08905,
		0.110452,
		0.134466,
		0.161142,
		0.190498,
		0.222444,
		0.256712,
		0.292845,
		0.330213,
		0.368043,
		0.405462,
		0.441538,
		0.475331,
		0.505937,
		0.532534,
		0.554413,
		0.571013,
		0.581942,
		0.586982,
		0.586099,
		0.579429,
		0.567267,
		0.550042,
		0.528295,
		0.502643,
		0.473757,
		0.442328,
		0.409045,
		0.374572,
		0.339533,
		0.304496,
		0.269972,
		0.236406,
		0.204182,
		0.173627,
		0.14501,
		0.118558,
		0.094456,
		0.072856,
		0.053884,
		0.037644,
		0.024222,
		0.013691,
		0.006111,
		0.001534,
		3.6e-005,
		0.000173,
		0.000568,
		0.00146,
		0.003161,
		0.006041,
		0.010488,
		0.016877,
		0.025537,
		0.036721,
		0.050596,
		0.067237,
		0.086641,
		0.108744,
		0.133455,
		0.160683,
		0.190361,
		0.222444,
		0.25688,
		0.293546,
		0.332146,
		0.372177,
		0.412943,
		0.4536,
		0.493212,
		0.530799,
		0.565401,
		0.596121,
		0.622176,
		0.642924,
		0.657896,
		0.666804,
		0.669547,
		0.666205,
		0.65702,
		0.642378,
		0.622781,
		0.598814,
		0.57112,
		0.540369,
		0.507231,
		0.472358,
		0.436364,
		0.399815,
		0.36322,
		0.327031,
		0.291638,
		0.257379,
		0.224539,
		0.19336,
		0.164044,
		0.136765,
		0.111669,
		0.088879,
		0.068505,
		0.050639,
		0.035363,
		0.022747,
		0.012854,
		0.005737,
		0.00144,
		5.1e-005,
		0.000242,
		0.00077,
		0.001926,
		0.00408,
		0.007659,
		0.01311,
		0.020858,
		0.031269,
		0.04462,
		0.061078,
		0.080692,
		0.103399,
		0.129044,
		0.15741,
		0.188251,
		0.22133,
		0.256446,
		0.293444,
		0.332194,
		0.372533,
		0.414176,
		0.456627,
		0.499182,
		0.540983,
		0.581082,
		0.618508,
		0.652329,
		0.681705,
		0.705931,
		0.724468,
		0.736961,
		0.743241,
		0.743327,
		0.737402,
		0.7258,
		0.708972,
		0.687459,
		0.661863,
		0.632819,
		0.600964,
		0.566923,
		0.531288,
		0.494607,
		0.457378,
		0.420042,
		0.382989,
		0.346557,
		0.311037,
		0.276679,
		0.243699,
		0.212284,
		0.182595,
		0.154774,
		0.128948,
		0.105229,
		0.083718,
		0.064503,
		0.047667,
		0.033279,
		0.021402,
		0.012092,
		0.005396,
		0.001354,
		3.1e-005,
		0.000225,
		0.000815,
		0.002141,
		0.004637,
		0.008796,
		0.015128,
		0.024116,
		0.036175,
		0.051615,
		0.070615,
		0.09321,
		0.119294,
		0.148632,
		0.180892,
		0.215674,
		0.252561,
		0.291157,
		0.331119,
		0.372177,
		0.414121,
		0.456759,
		0.499836,
		0.542935,
		0.585423,
		0.626498,
		0.665273,
		0.700845,
		0.732363,
		0.759087,
		0.780424,
		0.795954,
		0.805445,
		0.808844,
		0.806267,
		0.797981,
		0.784374,
		0.765928,
		0.743188,
		0.716734,
		0.687158,
		0.655041,
		0.620939,
		0.585368,
		0.548801,
		0.51166,
		0.474322,
		0.437117,
		0.400334,
		0.364225,
		0.329012,
		0.294891,
		0.262037,
		0.230604,
		0.200736,
		0.172561,
		0.146196,
		0.121749,
		0.099316,
		0.078987,
		0.06084,
		0.044948,
		0.031373,
		0.020173,
		0.011395,
		0.005084,
		0.001275,
		0.000105,
		0.000623,
		0.001951,
		0.004591,
		0.009106,
		0.016081,
		0.026068,
		0.039541,
		0.056853,
		0.078208,
		0.103642,
		0.133017,
		0.166038,
		0.202269,
		0.241175,
		0.282165,
		0.32464,
		0.368043,
		0.4119,
		0.455837,
		0.499583,
		0.542935,
		0.585696,
		0.627576,
		0.668102,
		0.706621,
		0.742376,
		0.774593,
		0.80255,
		0.825636,
		0.843385,
		0.855502,
		0.861859,
		0.862495,
		0.857596,
		0.847468,
		0.832516,
		0.81321,
		0.790059,
		0.76359,
		0.734326,
		0.702768,
		0.66939,
		0.634625,
		0.598868,
		0.56247,
		0.525747,
		0.488978,
		0.452411,
		0.416269,
		0.380753,
		0.346046,
		0.312315,
		0.279714,
		0.248386,
		0.218463,
		0.190067,
		0.163312,
		0.138302,
		0.11513,
		0.093885,
		0.074644,
		0.057479,
		0.042454,
		0.029626,
		0.019046,
		0.010757,
		0.004799,
		0.001204,
		0.00025,
		0.001329,
		0.003813,
		0.008349,
		0.015607,
		0.026217,
		0.040725,
		0.05954,
		0.082904,
		0.110866,
		0.143273,
		0.179781,
		0.219874,
		0.262895,
		0.308099,
		0.354699,
		0.401925,
		0.449081,
		0.495581,
		0.540983,
		0.584989,
		0.627422,
		0.66817,
		0.707107,
		0.743988,
		0.778389,
		0.809757,
		0.837497,
		0.861048,
		0.879941,
		0.893836,
		0.902538,
		0.906002,
		0.904319,
		0.897699,
		0.886448,
		0.870942,
		0.8516,
		0.828867,
		0.803188,
		0.774998,
		0.744711,
		0.712711,
		0.679351,
		0.644953,
		0.609807,
		0.574175,
		0.538297,
		0.502393,
		0.466662,
		0.431293,
		0.39646,
		0.362328,
		0.32905,
		0.29677,
		0.265625,
		0.23574,
		0.207232,
		0.180211,
		0.154777,
		0.131022,
		0.109032,
		0.088884,
		0.070648,
		0.054388,
		0.040162,
		0.028021,
		0.018011,
		0.010171,
		0.004537,
		0.001138,
		0.000487,
		0.002407,
		0.006505,
		0.013546,
		0.024257,
		0.039266,
		0.059053,
		0.08391,
		0.113912,
		0.148905,
		0.188515,
		0.232157,
		0.279078,
		0.328396,
		0.379157,
		0.430401,
		0.481216,
		0.530799,
		0.578504,
		0.623865,
		0.666607,
		0.706621,
		0.743918,
		0.778552,
		0.810531,
		0.839721,
		0.865825,
		0.888462,
		0.907247,
		0.921848,
		0.932032,
		0.937675,
		0.938773,
		0.935426,
		0.927827,
		0.916238,
		0.900971,
		0.882368,
		0.860781,
		0.836564,
		0.810056,
		0.781582,
		0.751442,
		0.719918,
		0.687268,
		0.653732,
		0.619534,
		0.584885,
		0.549981,
		0.515011,
		0.480153,
		0.445578,
		0.411446,
		0.377913,
		0.345123,
		0.313213,
		0.282312,
		0.25254,
		0.22401,
		0.196828,
		0.171089,
		0.146884,
		0.124297,
		0.103402,
		0.084269,
		0.066963,
		0.051539,
		0.038051,
		0.026543,
		0.017058,
		0.009632,
		0.004296,
		0.001077,
		0.000827,
		0.00389,
		0.010067,
		0.020197,
		0.035006,
		0.05506,
		0.080714,
		0.112084,
		0.14903,
		0.191158,
		0.237836,
		0.288229,
		0.341342,
		0.39608,
		0.451312,
		0.505937,
		0.558952,
		0.609506,
		0.656947,
		0.700845,
		0.740998,
		0.777408,
		0.810233,
		0.839721,
		0.86612,
		0.889587,
		0.910115,
		0.927549,
		0.941677,
		0.95229,
		0.95923,
		0.96241,
		0.961826,
		0.95755,
		0.949718,
		0.938519,
		0.924176,
		0.906936,
		0.887051,
		0.864779,
		0.840368,
		0.81406,
		0.786085,
		0.756663,
		0.726004,
		0.694312,
		0.661781,
		0.628601,
		0.594955,
		0.561021,
		0.526971,
		0.492972,
		0.459182,
		0.425753,
		0.392831,
		0.360554,
		0.32905,
		0.298442,
		0.268846,
		0.24037,
		0.213114,
		0.187172,
		0.162631,
		0.139573,
		0.118071,
		0.098193,
		0.080003,
		0.063557,
		0.048908,
		0.036101,
		0.025179,
		0.016179,
		0.009134,
		0.004073,
		0.001021,
		0.001275,
		0.005777,
		0.014469,
		0.028199,
		0.047631,
		0.073201,
		0.105078,
		0.143149,
		0.187018,
		0.236018,
		0.289251,
		0.34563,
		0.403941,
		0.462908,
		0.52127,
		0.577846,
		0.631602,
		0.681705,
		0.727561,
		0.768831,
		0.805429,
		0.837497,
		0.865353,
		0.889428,
		0.910182,
		0.928019,
		0.943198,
		0.955795,
		0.965756,
		0.97297,
		0.977327,
		0.978745,
		0.977186,
		0.972665,
		0.965238,
		0.955004,
		0.942087,
		0.926633,
		0.908799,
		0.888751,
		0.86666,
		0.842698,
		0.817037,
		0.789852,
		0.761316,
		0.731605,
		0.700893,
		0.669354,
		0.63716,
		0.604481,
		0.571482,
		0.538323,
		0.505161,
		0.472143,
		0.439414,
		0.407111,
		0.375363,
		0.344293,
		0.314019,
		0.28465,
		0.25629,
		0.229035,
		0.202976,
		0.178197,
		0.154777,
		0.132788,
		0.112297,
		0.093366,
		0.076051,
		0.060404,
		0.046472,
		0.034297,
		0.023917,
		0.015366,
		0.008674,
		0.003868,
		0.00097,
		0.001827,
		0.008036,
		0.019612,
		0.037339,
		0.061746,
		0.093074,
		0.131251,
		0.175889,
		0.226305,
		0.281552,
		0.34047,
		0.401745,
		0.463986,
		0.525792,
		0.585835,
		0.642924,
		0.696065,
		0.744506,
		0.787765,
		0.825636,
		0.858176,
		0.885676,
		0.908605,
		0.927549,
		0.943133,
		0.955943,
		0.966443,
		0.974912,
		0.981438,
		0.985989,
		0.988479,
		0.988808,
		0.986889,
		0.982663,
		0.9761,
		0.967203,
		0.956001,
		0.942546,
		0.926912,
		0.909189,
		0.889481,
		0.867908,
		0.844598,
		0.819691,
		0.793332,
		0.765677,
		0.736881,
		0.707107,
		0.676513,
		0.645261,
		0.613507,
		0.581406,
		0.549108,
		0.516759,
		0.484498,
		0.452462,
		0.420779,
		0.389573,
		0.358959,
		0.32905,
		0.299949,
		0.271757,
		0.244566,
		0.218463,
		0.193529,
		0.169842,
		0.14747,
		0.12648,
		0.106933,
		0.088884,
		0.072383,
		0.057479,
		0.044213,
		0.032625,
		0.022747,
		0.014613,
		0.008248,
		0.003677,
		0.000922,
		0.002468,
		0.010604,
		0.025332,
		0.047295,
		0.076818,
		0.113883,
		0.158129,
		0.208867,
		0.265117,
		0.325662,
		0.389109,
		0.453969,
		0.518731,
		0.581942,
		0.642277,
		0.698606,
		0.750038,
		0.795954,
		0.836021,
		0.870185,
		0.898653,
		0.921848,
		0.940363,
		0.954896,
		0.966179,
		0.974912,
		0.981687,
		0.986929,
		0.99085,
		0.993485,
		0.994761,
		0.994554,
		0.992725,
		0.989142,
		0.983695,
		0.9763,
		0.966903,
		0.955481,
		0.942036,
		0.926602,
		0.909231,
		0.890001,
		0.869007,
		0.846361,
		0.822187,
		0.796619,
		0.769799,
		0.741872,
		0.712986,
		0.683291,
		0.652936,
		0.622067,
		0.590828,
		0.559361,
		0.527802,
		0.496282,
		0.464929,
		0.433865,
		0.403207,
		0.373065,
		0.343545,
		0.314747,
		0.286766,
		0.259691,
		0.233606,
		0.208589,
		0.184715,
		0.162051,
		0.140662,
		0.120607,
		0.101941,
		0.084714,
		0.068974,
		0.054761,
		0.042115,
		0.031071,
		0.021661,
		0.013913,
		0.007852,
		0.003501,
		0.000878,
		0.003178,
		0.013387,
		0.031406,
		0.057659,
		0.092199,
		0.134701,
		0.184485,
		0.240551,
		0.30164,
		0.366296,
		0.43295,
		0.5,
		0.565889,
		0.629185,
		0.68864,
		0.743241,
		0.792248,
		0.835202,
		0.871933,
		0.902538,
		0.92735,
		0.946897,
		0.961852,
		0.97297,
		0.981036,
		0.986797,
		0.990906,
		0.993865,
		0.995982,
		0.997356,
		0.997934,
		0.997575,
		0.996096,
		0.99331,
		0.989042,
		0.983139,
		0.975486,
		0.965998,
		0.954629,
		0.941362,
		0.926215,
		0.909232,
		0.89048,
		0.870045,
		0.848031,
		0.824552,
		0.799732,
		0.773702,
		0.746597,
		0.718555,
		0.689715,
		0.660215,
		0.630193,
		0.599783,
		0.569117,
		0.538323,
		0.507526,
		0.476846,
		0.446396,
		0.416289,
		0.38663,
		0.357518,
		0.32905,
		0.301315,
		0.2744,
		0.248386,
		0.223347,
		0.199356,
		0.176478,
		0.154777,
		0.13431,
		0.11513,
		0.097288,
		0.08083,
		0.065798,
		0.05223,
		0.040162,
		0.029626,
		0.020651,
		0.013263,
		0.007484,
		0.003336,
		0.000836,
		0.003928,
		0.01627,
		0.037571,
		0.067966,
		0.107181,
		0.154555,
		0.209081,
		0.269468,
		0.334211,
		0.401677,
		0.470186,
		0.538097,
		0.603886,
		0.666205,
		0.723937,
		0.77623,
		0.822512,
		0.862495,
		0.896157,
		0.923719,
		0.945607,
		0.96241,
		0.974831,
		0.983637,
		0.989607,
		0.993485,
		0.99593,
		0.997473,
		0.998474,
		0.9991,
		0.99933,
		0.999026,
		0.997988,
		0.995994,
		0.992826,
		0.988288,
		0.982219,
		0.974494,
		0.965026,
		0.953766,
		0.940699,
		0.925841,
		0.909232,
		0.890936,
		0.871032,
		0.849616,
		0.826794,
		0.802682,
		0.777401,
		0.751077,
		0.723836,
		0.69581,
		0.667127,
		0.637915,
		0.608301,
		0.578407,
		0.548355,
		0.518262,
		0.488241,
		0.4584,
		0.428844,
		0.399673,
		0.370984,
		0.342867,
		0.315408,
		0.28869,
		0.26279,
		0.237781,
		0.213733,
		0.19071,
		0.168771,
		0.147975,
		0.128373,
		0.110014,
		0.092945,
		0.077206,
		0.062836,
		0.049871,
		0.038342,
		0.02828,
		0.01971,
		0.012657,
		0.007142,
		0.003183,
		0.000798,
		0.004685,
		0.019122,
		0.043541,
		0.077731,
		0.121059,
		0.172521,
		0.230803,
		0.294361,
		0.361507,
		0.430498,
		0.499619,
		0.567267,
		0.632006,
		0.692627,
		0.748179,
		0.797981,
		0.841631,
		0.878988,
		0.910151,
		0.935426,
		0.955287,
		0.970337,
		0.981267,
		0.988808,
		0.993699,
		0.996642,
		0.998268,
		0.9991,
		0.999523,
		0.999756,
		0.999837,
		0.999659,
		0.999028,
		0.997716,
		0.995487,
		0.992122,
		0.987433,
		0.981267,
		0.973507,
		0.964076,
		0.952932,
		0.940063,
		0.925483,
		0.909232,
		0.89137,
		0.87197,
		0.851122,
		0.828924,
		0.805484,
		0.780913,
		0.755329,
		0.728851,
		0.701601,
		0.673697,
		0.645261,
		0.61641,
		0.587261,
		0.557927,
		0.528519,
		0.499142,
		0.469901,
		0.440894,
		0.412216,
		0.383959,
		0.356209,
		0.32905,
		0.302559,
		0.276811,
		0.251877,
		0.227824,
		0.204713,
		0.182603,
		0.161551,
		0.141606,
		0.122817,
		0.105229,
		0.088884,
		0.073818,
		0.060068,
		0.047667,
		0.036642,
		0.027023,
		0.018832,
		0.012092,
		0.006822,
		0.003041,
		0.000762,
		0.005415,
		0.021807,
		0.049033,
		0.086499,
		0.133206,
		0.187825,
		0.24878,
		0.314334,
		0.382687,
		0.45206,
		0.520775,
		0.58732,
		0.650402,
		0.708972,
		0.76225,
		0.809723,
		0.851131,
		0.886448,
		0.915851,
		0.939685,
		0.95843,
		0.972665,
		0.983029,
		0.990193,
		0.994828,
		0.997575,
		0.999017,
		0.999655,
		0.99988,
		0.999952,
		0.999976,
		0.999904,
		0.99958,
		0.998793,
		0.997312,
		0.994912,
		0.99139,
		0.986574,
		0.980326,
		0.972543,
		0.963156,
		0.952128,
		0.939451,
		0.92514,
		0.909232,
		0.891784,
		0.872864,
		0.852555,
		0.830949,
		0.808146,
		0.78425,
		0.759371,
		0.733619,
		0.707107,
		0.679948,
		0.652255,
		0.624138,
		0.595706,
		0.567067,
		0.538323,
		0.509577,
		0.480925,
		0.452462,
		0.424278,
		0.39646,
		0.369091,
		0.342249,
		0.31601,
		0.290446,
		0.265625,
		0.24161,
		0.218463,
		0.19624,
		0.174994,
		0.154777,
		0.135635,
		0.117611,
		0.100747,
		0.085081,
		0.070648,
		0.057479,
		0.045605,
		0.035053,
		0.025848,
		0.018011,
		0.011564,
		0.006524,
		0.002907,
		0.000729,
		0.006083,
		0.024201,
		0.053801,
		0.093895,
		0.143141,
		0.199929,
		0.26248,
		0.328938,
		0.397462,
		0.466305,
		0.533882,
		0.598814,
		0.659966,
		0.716456,
		0.767663,
		0.81321,
		0.852944,
		0.88691,
		0.91532,
		0.938519,
		0.956957,
		0.971156,
		0.981687,
		0.989142,
		0.994116,
		0.997181,
		0.998871,
		0.999659,
		0.999936,
		0.999993,
		0.999998,
		0.999982,
		0.99985,
		0.999429,
		0.998511,
		0.996876,
		0.994317,
		0.99065,
		0.985719,
		0.9794,
		0.971601,
		0.962262,
		0.951351,
		0.938862,
		0.924811,
		0.909232,
		0.892179,
		0.873716,
		0.85392,
		0.832877,
		0.81068,
		0.787425,
		0.763216,
		0.738155,
		0.712348,
		0.685901,
		0.65892,
		0.631507,
		0.603767,
		0.575799,
		0.547701,
		0.51957,
		0.491497,
		0.463572,
		0.43588,
		0.408505,
		0.381525,
		0.355015,
		0.32905,
		0.303696,
		0.279019,
		0.255081,
		0.231941,
		0.209654,
		0.188271,
		0.167843,
		0.148414,
		0.130029,
		0.112726,
		0.096544,
		0.081516,
		0.067676,
		0.055053,
		0.043674,
		0.033565,
		0.024747,
		0.017243,
		0.011069,
		0.006244,
		0.002783,
		0.000697,
		0.006658,
		0.026203,
		0.057659,
		0.099666,
		0.150584,
		0.208583,
		0.271753,
		0.338191,
		0.406079,
		0.473757,
		0.539764,
		0.602874,
		0.662108,
		0.716734,
		0.766258,
		0.810403,
		0.849081,
		0.882368,
		0.910471,
		0.933706,
		0.952465,
		0.967203,
		0.97841,
		0.986599,
		0.99229,
		0.995994,
		0.998198,
		0.999349,
		0.999839,
		0.999982,
		1,
		0.999999,
		0.999962,
		0.999771,
		0.999251,
		0.9982,
		0.996415,
		0.993706,
		0.989904,
		0.984868,
		0.978488,
		0.970681,
		0.961395,
		0.950601,
		0.938296,
		0.924495,
		0.909232,
		0.892557,
		0.874529,
		0.855222,
		0.834714,
		0.813093,
		0.790449,
		0.766878,
		0.742476,
		0.717343,
		0.691577,
		0.665277,
		0.638541,
		0.611467,
		0.584147,
		0.556677,
		0.529145,
		0.501639,
		0.474244,
		0.447041,
		0.42011,
		0.393526,
		0.367361,
		0.341684,
		0.316562,
		0.292057,
		0.268228,
		0.245133,
		0.222825,
		0.201355,
		0.18077,
		0.161115,
		0.142431,
		0.12476,
		0.108136,
		0.092596,
		0.07817,
		0.064888,
		0.052777,
		0.041863,
		0.032169,
		0.023716,
		0.016522,
		0.010606,
		0.005983,
		0.002666,
		0.000668,
		0.007118,
		0.027745,
		0.060501,
		0.103703,
		0.155471,
		0.213831,
		0.276811,
		0.342523,
		0.409225,
		0.475373,
		0.539646,
		0.600964,
		0.658487,
		0.711604,
		0.75991,
		0.803188,
		0.841376,
		0.874547,
		0.902878,
		0.926633,
		0.94614,
		0.961779,
		0.973965,
		0.983139,
		0.989757,
		0.994276,
		0.997146,
		0.998793,
		0.999606,
		0.99992,
		0.999995,
		1,
		0.999995,
		0.999929,
		0.999671,
		0.999046,
		0.997863,
		0.995933,
		0.993081,
		0.989154,
		0.984024,
		0.977591,
		0.969784,
		0.960553,
		0.949876,
		0.937751,
		0.924192,
		0.909232,
		0.892918,
		0.875306,
		0.856465,
		0.836467,
		0.815394,
		0.793332,
		0.770369,
		0.746597,
		0.722107,
		0.696992,
		0.671346,
		0.645261,
		0.618827,
		0.592135,
		0.565272,
		0.538323,
		0.511372,
		0.484498,
		0.457781,
		0.431293,
		0.405109,
		0.379296,
		0.353922,
		0.32905,
		0.304739,
		0.281048,
		0.258031,
		0.23574,
		0.214224,
		0.193529,
		0.1737,
		0.154777,
		0.136799,
		0.119801,
		0.103819,
		0.088884,
		0.075024,
		0.062267,
		0.050639,
		0.040162,
		0.030859,
		0.022747,
		0.015846,
		0.010171,
		0.005737,
		0.002556,
		0.000641,
		0.007452,
		0.028801,
		0.062312,
		0.106041,
		0.157944,
		0.215973,
		0.278157,
		0.34267,
		0.407876,
		0.472358,
		0.534931,
		0.594644,
		0.650766,
		0.702768,
		0.750302,
		0.793172,
		0.831316,
		0.864779,
		0.893693,
		0.918264,
		0.938756,
		0.955481,
		0.968789,
		0.979061,
		0.9867,
		0.992122,
		0.995745,
		0.997978,
		0.999205,
		0.999771,
		0.999964,
		0.999999,
		1,
		0.999986,
		0.999882,
		0.999548,
		0.998815,
		0.997501,
		0.995431,
		0.992444,
		0.988401,
		0.983187,
		0.976711,
		0.968908,
		0.959736,
		0.949176,
		0.937225,
		0.9239,
		0.909232,
		0.893264,
		0.87605,
		0.857652,
		0.838141,
		0.817592,
		0.796084,
		0.773702,
		0.750529,
		0.726654,
		0.702164,
		0.677145,
		0.651685,
		0.625869,
		0.599783,
		0.573509,
		0.547127,
		0.520718,
		0.494356,
		0.468117,
		0.442072,
		0.416289,
		0.390835,
		0.365774,
		0.341166,
		0.317068,
		0.293538,
		0.270627,
		0.248386,
		0.226862,
		0.2061,
		0.186142,
		0.16703,
		0.148801,
		0.13149,
		0.11513,
		0.099754,
		0.085389,
		0.072063,
		0.059802,
		0.048628,
		0.038563,
		0.029626,
		0.021837,
		0.015211,
		0.009762,
		0.005506,
		0.002453,
		0.000615,
		0.007659,
		0.029384,
		0.063155,
		0.106843,
		0.158311,
		0.215498,
		0.276483,
		0.339533,
		0.403127,
		0.465973,
		0.527004,
		0.585368,
		0.640413,
		0.691658,
		0.73878,
		0.781582,
		0.819979,
		0.853978,
		0.883665,
		0.909189,
		0.930757,
		0.948626,
		0.963094,
		0.974494,
		0.983187,
		0.989554,
		0.993987,
		0.996876,
		0.998595,
		0.999493,
		0.999875,
		0.999986,
		1,
		0.999999,
		0.99997,
		0.999819,
		0.999403,
		0.99856,
		0.997117,
		0.994912,
		0.991798,
		0.987648,
		0.982358,
		0.975845,
		0.968053,
		0.958943,
		0.948498,
		0.936718,
		0.92362,
		0.909232,
		0.893596,
		0.876762,
		0.858788,
		0.839741,
		0.819691,
		0.798713,
		0.776885,
		0.754286,
		0.731,
		0.707107,
		0.68269,
		0.657831,
		0.632611,
		0.607109,
		0.581406,
		0.555576,
		0.529695,
		0.503836,
		0.478069,
		0.452462,
		0.427082,
		0.40199,
		0.37725,
		0.352917,
		0.32905,
		0.3057,
		0.282919,
		0.260756,
		0.239256,
		0.218463,
		0.198419,
		0.179163,
		0.160731,
		0.14316,
		0.12648,
		0.110724,
		0.09592,
		0.082095,
		0.069274,
		0.057479,
		0.046734,
		0.037057,
		0.028466,
		0.02098,
		0.014613,
		0.009378,
		0.005289,
		0.002356,
		0.00059,
		0.007746,
		0.029542,
		0.06316,
		0.10636,
		0.156977,
		0.212988,
		0.272552,
		0.334039,
		0.396042,
		0.457378,
		0.517072,
		0.574344,
		0.628589,
		0.679351,
		0.726308,
		0.769247,
		0.808054,
		0.842698,
		0.873218,
		0.89972,
		0.922363,
		0.941362,
		0.956972,
		0.969489,
		0.97924,
		0.986574,
		0.991857,
		0.995458,
		0.99774,
		0.999046,
		0.999689,
		0.999937,
		0.999996,
		1,
		0.999996,
		0.999944,
		0.999738,
		0.999236,
		0.99828,
		0.996711,
		0.994377,
		0.991144,
		0.986896,
		0.981538,
		0.974996,
		0.967218,
		0.958172,
		0.947842,
		0.936229,
		0.92335,
		0.909232,
		0.893914,
		0.877444,
		0.859876,
		0.841272,
		0.821699,
		0.801227,
		0.779928,
		0.757879,
		0.735156,
		0.711836,
		0.687996,
		0.663715,
		0.639069,
		0.614133,
		0.588982,
		0.563688,
		0.538323,
		0.512956,
		0.487653,
		0.462481,
		0.437501,
		0.412775,
		0.38836,
		0.364313,
		0.340688,
		0.317536,
		0.294906,
		0.272844,
		0.251397,
		0.230606,
		0.210511,
		0.191151,
		0.172561,
		0.154777,
		0.137829,
		0.121749,
		0.106564,
		0.092302,
		0.078987,
		0.066642,
		0.055289,
		0.044948,
		0.035637,
		0.027373,
		0.020173,
		0.014049,
		0.009016,
		0.005084,
		0.002265,
		0.000567,
		0.007731,
		0.029348,
		0.062493,
		0.104887,
		0.154387,
		0.209038,
		0.267095,
		0.327031,
		0.38754,
		0.447525,
		0.50608,
		0.56247,
		0.616116,
		0.666569,
		0.713495,
		0.756663,
		0.795928,
		0.831226,
		0.862561,
		0.890001,
		0.913673,
		0.933754,
		0.950468,
		0.964076,
		0.974876,
		0.983187,
		0.989348,
		0.993706,
		0.996609,
		0.998392,
		0.999369,
		0.999819,
		0.999972,
		0.999999,
		1,
		0.999989,
		0.999906,
		0.999638,
		0.999046,
		0.997978,
		0.996286,
		0.993829,
		0.990485,
		0.986146,
		0.980727,
		0.974162,
		0.966404,
		0.957423,
		0.947206,
		0.935757,
		0.92309,
		0.909232,
		0.89422,
		0.878098,
		0.860918,
		0.842738,
		0.823621,
		0.803632,
		0.782841,
		0.761317,
		0.739133,
		0.716363,
		0.693079,
		0.669354,
		0.645261,
		0.620871,
		0.596255,
		0.571482,
		0.546619,
		0.521733,
		0.496887,
		0.472143,
		0.447562,
		0.423201,
		0.399117,
		0.375363,
		0.35199,
		0.32905,
		0.306588,
		0.28465,
		0.26328,
		0.242518,
		0.222405,
		0.202976,
		0.184267,
		0.166312,
		0.149143,
		0.132788,
		0.117276,
		0.102633,
		0.088884,
		0.076051,
		0.064157,
		0.053221,
		0.043262,
		0.034297,
		0.026342,
		0.019411,
		0.013517,
		0.008674,
		0.004891,
		0.002179,
		0.000546,
		0.007635,
		0.028884,
		0.061335,
		0.102721,
		0.15096,
		0.204182,
		0.260738,
		0.319197,
		0.378334,
		0.437117,
		0.494688,
		0.550341,
		0.603506,
		0.653732,
		0.700674,
		0.74408,
		0.783783,
		0.819691,
		0.851782,
		0.880097,
		0.904734,
		0.925841,
		0.943612,
		0.95828,
		0.970112,
		0.9794,
		0.986456,
		0.991607,
		0.99518,
		0.997501,
		0.998879,
		0.999596,
		0.9999,
		0.999989,
		1,
		0.999999,
		0.999976,
		0.999854,
		0.999519,
		0.998834,
		0.997655,
		0.995843,
		0.99327,
		0.989821,
		0.985399,
		0.979927,
		0.973345,
		0.965609,
		0.956695,
		0.946591,
		0.935301,
		0.92284,
		0.909232,
		0.894513,
		0.878726,
		0.861918,
		0.844144,
		0.825463,
		0.805937,
		0.78563,
		0.76461,
		0.742944,
		0.720701,
		0.69795,
		0.67476,
		0.651201,
		0.627338,
		0.603241,
		0.578973,
		0.5546,
		0.530184,
		0.505785,
		0.481465,
		0.457278,
		0.433283,
		0.409531,
		0.386075,
		0.362964,
		0.340247,
		0.317968,
		0.296172,
		0.2749,
		0.254193,
		0.234088,
		0.214622,
		0.19583,
		0.177742,
		0.160392,
		0.143807,
		0.128015,
		0.113042,
		0.098913,
		0.08565,
		0.073275,
		0.061808,
		0.051267,
		0.041669,
		0.033031,
		0.025367,
		0.018691,
		0.013016,
		0.008351,
		0.004709,
		0.002098,
		0.000525,
		0.00748,
		0.028234,
		0.059855,
		0.100129,
		0.147054,
		0.198852,
		0.253963,
		0.311037,
		0.368915,
		0.426612,
		0.483303,
		0.538297,
		0.591031,
		0.641049,
		0.687995,
		0.731605,
		0.771692,
		0.808146,
		0.840925,
		0.870045,
		0.895582,
		0.917657,
		0.936437,
		0.952128,
		0.964969,
		0.975226,
		0.983187,
		0.989154,
		0.993437,
		0.996348,
		0.998186,
		0.999236,
		0.999752,
		0.99995,
		0.999997,
		1,
		0.999997,
		0.999955,
		0.999788,
		0.999381,
		0.9986,
		0.997312,
		0.995385,
		0.9927,
		0.989154,
		0.984657,
		0.979138,
		0.972543,
		0.964833,
		0.955987,
		0.945995,
		0.93486,
		0.922598,
		0.909232,
		0.894796,
		0.879329,
		0.862877,
		0.845492,
		0.827229,
		0.808146,
		0.788304,
		0.767766,
		0.746597,
		0.72486,
		0.702622,
		0.679948,
		0.656903,
		0.63355,
		0.609954,
		0.586177,
		0.56228,
		0.538323,
		0.514364,
		0.49046,
		0.466665,
		0.443032,
		0.419614,
		0.39646,
		0.373618,
		0.351133,
		0.32905,
		0.307411,
		0.286256,
		0.265625,
		0.245554,
		0.226079,
		0.207232,
		0.189047,
		0.171552,
		0.154777,
		0.138748,
		0.123492,
		0.109032,
		0.095391,
		0.082589,
		0.070648,
		0.059585,
		0.049418,
		0.040162,
		0.031834,
		0.024446,
		0.018011,
		0.012541,
		0.008046,
		0.004537,
		0.002021,
		0.000506,
		0.007286,
		0.027471,
		0.058196,
		0.097323,
		0.142941,
		0.19336,
		0.247099,
		0.302876,
		0.359581,
		0.416269,
		0.472135,
		0.526502,
		0.578809,
		0.628601,
		0.675515,
		0.719275,
		0.759685,
		0.796619,
		0.830017,
		0.859876,
		0.886248,
		0.909232,
		0.928972,
		0.945646,
		0.959469,
		0.970681,
		0.979547,
		0.986346,
		0.991369,
		0.994912,
		0.997265,
		0.998706,
		0.999493,
		0.999854,
		0.999977,
		0.999999,
		1,
		0.999991,
		0.999924,
		0.999707,
		0.999223,
		0.998347,
		0.99695,
		0.994912,
		0.992122,
		0.988485,
		0.983919,
		0.978359,
		0.971756,
		0.964076,
		0.955299,
		0.945417,
		0.934434,
		0.922365,
		0.909232,
		0.895068,
		0.879909,
		0.863799,
		0.846787,
		0.828924,
		0.810266,
		0.790869,
		0.770794,
		0.750101,
		0.728851,
		0.707107,
		0.684929,
		0.66238,
		0.63952,
		0.61641,
		0.593109,
		0.569676,
		0.546167,
		0.522638,
		0.499142,
		0.475734,
		0.452462,
		0.429378,
		0.406528,
		0.383959,
		0.361715,
		0.339837,
		0.318369,
		0.297347,
		0.276811,
		0.256796,
		0.237335,
		0.218463,
		0.200209,
		0.182603,
		0.165674,
		0.149448,
		0.133949,
		0.119202,
		0.105229,
		0.092052,
		0.079688,
		0.068159,
		0.057479,
		0.047667,
		0.038736,
		0.0307,
		0.023574,
		0.017367,
		0.012092,
		0.007758,
		0.004374,
		0.001948,
		0.000488,
		0.007072,
		0.026654,
		0.056465,
		0.094456,
		0.138805,
		0.187904,
		0.240343,
		0.294891,
		0.350485,
		0.406206,
		0.461269,
		0.515011,
		0.566876,
		0.61641,
		0.663248,
		0.707107,
		0.747781,
		0.785132,
		0.819083,
		0.849616,
		0.876762,
		0.900598,
		0.921246,
		0.938862,
		0.953636,
		0.965784,
		0.975548,
		0.983187,
		0.988971,
		0.993181,
		0.996092,
		0.997978,
		0.999094,
		0.999675,
		0.99992,
		0.999991,
		1,
		0.999999,
		0.999981,
		0.999882,
		0.999608,
		0.999046,
		0.998073,
		0.996571,
		0.994427,
		0.991538,
		0.987816,
		0.983187,
		0.977591,
		0.970985,
		0.963338,
		0.95463,
		0.944857,
		0.934022,
		0.922139,
		0.909232,
		0.895329,
		0.880467,
		0.864686,
		0.848031,
		0.830552,
		0.812301,
		0.793332,
		0.773702,
		0.753466,
		0.732684,
		0.711414,
		0.689715,
		0.667644,
		0.645261,
		0.622621,
		0.599783,
		0.5768,
		0.553728,
		0.53062,
		0.507526,
		0.484498,
		0.461585,
		0.438833,
		0.416289,
		0.393997,
		0.371999,
		0.350337,
		0.32905,
		0.308175,
		0.28775,
		0.267809,
		0.248386,
		0.229511,
		0.211216,
		0.193529,
		0.176478,
		0.160089,
		0.144385,
		0.129392,
		0.11513,
		0.101621,
		0.088884,
		0.076937,
		0.065798,
		0.055483,
		0.046007,
		0.037384,
		0.029626,
		0.022747,
		0.016757,
		0.011667,
		0.007484,
		0.004219,
		0.001879,
		0.000471,
		0.006851,
		0.025825,
		0.054735,
		0.091624,
		0.134756,
		0.182595,
		0.233791,
		0.287164,
		0.341684,
		0.39646,
		0.450728,
		0.503836,
		0.555238,
		0.604481,
		0.651201,
		0.695109,
		0.735993,
		0.773702,
		0.808146,
		0.839291,
		0.867151,
		0.891784,
		0.913289,
		0.931803,
		0.947493,
		0.960553,
		0.971204,
		0.979683,
		0.986243,
		0.991144,
		0.994653,
		0.99703,
		0.998528,
		0.999381,
		0.999799,
		0.999959,
		0.999997,
		1,
		0.999998,
		0.999963,
		0.999827,
		0.999493,
		0.99885,
		0.997782,
		0.996177,
		0.99393,
		0.990947,
		0.987147,
		0.982461,
		0.976835,
		0.97023,
		0.962616,
		0.953979,
		0.944313,
		0.933622,
		0.921922,
		0.909232,
		0.895582,
		0.881004,
		0.865538,
		0.849227,
		0.832117,
		0.814257,
		0.795699,
		0.776495,
		0.756699,
		0.736367,
		0.715554,
		0.694316,
		0.672708,
		0.650784,
		0.628601,
		0.606211,
		0.583667,
		0.561021,
		0.538323,
		0.515624,
		0.492972,
		0.470412,
		0.447991,
		0.425753,
		0.403741,
		0.381994,
		0.360554,
		0.339457,
		0.318742,
		0.298442,
		0.278592,
		0.259225,
		0.24037,
		0.222058,
		0.204316,
		0.187172,
		0.170651,
		0.154777,
		0.139573,
		0.125061,
		0.111261,
		0.098193,
		0.085875,
		0.074325,
		0.063557,
		0.053588,
		0.044432,
		0.036101,
		0.028608,
		0.021964,
		0.016179,
		0.011263,
		0.007225,
		0.004073,
		0.001814,
		0.000454,
		0.00663,
		0.02501,
		0.053048,
		0.088879,
		0.130845,
		0.177477,
		0.227479,
		0.279714,
		0.333186,
		0.387032,
		0.440506,
		0.492972,
		0.54389,
		0.592813,
		0.639377,
		0.683291,
		0.724333,
		0.762345,
		0.797225,
		0.828924,
		0.85744,
		0.882814,
		0.905127,
		0.924495,
		0.941064,
		0.95501,
		0.966531,
		0.975845,
		0.983187,
		0.9888,
		0.992935,
		0.995843,
		0.997769,
		0.998946,
		0.999588,
		0.999882,
		0.999982,
		1,
		1,
		0.999993,
		0.999938,
		0.99976,
		0.999361,
		0.998635,
		0.997473,
		0.995768,
		0.993424,
		0.990352,
		0.986479,
		0.981742,
		0.976091,
		0.969489,
		0.961912,
		0.953345,
		0.943785,
		0.933236,
		0.921711,
		0.909232,
		0.895825,
		0.881522,
		0.866359,
		0.850378,
		0.833622,
		0.816138,
		0.797974,
		0.77918,
		0.759807,
		0.739908,
		0.719536,
		0.698742,
		0.67758,
		0.656102,
		0.63436,
		0.612405,
		0.590287,
		0.568056,
		0.545761,
		0.523449,
		0.501165,
		0.478955,
		0.456863,
		0.43493,
		0.413199,
		0.391707,
		0.370494,
		0.349596,
		0.32905,
		0.308888,
		0.289144,
		0.269849,
		0.251033,
		0.232725,
		0.214952,
		0.197741,
		0.181117,
		0.165103,
		0.149721,
		0.134994,
		0.120941,
		0.107582,
		0.094935,
		0.083016,
		0.071843,
		0.061429,
		0.051789,
		0.042936,
		0.034883,
		0.027641,
		0.02122,
		0.01563,
		0.01088,
		0.006979,
		0.003934,
		0.001752,
		0.000439,
		0.006416,
		0.024222,
		0.051425,
		0.086243,
		0.127091,
		0.172561,
		0.221409,
		0.272538,
		0.324985,
		0.377913,
		0.430596,
		0.48241,
		0.532828,
		0.581406,
		0.62778,
		0.671658,
		0.712812,
		0.751077,
		0.786339,
		0.818536,
		0.847654,
		0.873716,
		0.896786,
		0.916963,
		0.934374,
		0.949176,
		0.961546,
		0.971686,
		0.97981,
		0.986146,
		0.990931,
		0.994403,
		0.996799,
		0.998347,
		0.999261,
		0.999735,
		0.999935,
		0.999993,
		1,
		1,
		0.999984,
		0.999903,
		0.999678,
		0.999212,
		0.998403,
		0.997147,
		0.995346,
		0.992908,
		0.989754,
		0.985814,
		0.98103,
		0.975358,
		0.968764,
		0.961225,
		0.952728,
		0.943272,
		0.932861,
		0.921508,
		0.909232,
		0.89606,
		0.882021,
		0.867151,
		0.851487,
		0.835071,
		0.817948,
		0.800163,
		0.781763,
		0.762798,
		0.743316,
		0.723368,
		0.703003,
		0.682272,
		0.661224,
		0.63991,
		0.618377,
		0.596674,
		0.574848,
		0.552945,
		0.531011,
		0.50909,
		0.487226,
		0.465459,
		0.44383,
		0.42238,
		0.401146,
		0.380165,
		0.359472,
		0.339103,
		0.319089,
		0.299464,
		0.280256,
		0.261496,
		0.243212,
		0.225429,
		0.208175,
		0.191473,
		0.175346,
		0.159817,
		0.144906,
		0.130634,
		0.11702,
		0.104082,
		0.091835,
		0.080297,
		0.069482,
		0.059405,
		0.050078,
		0.041515,
		0.033726,
		0.026722,
		0.020513,
		0.015108,
		0.010517,
		0.006746,
		0.003803,
		0.001693,
		0.000424,
		0.006211,
		0.023468,
		0.04987,
		0.083718,
		0.123492,
		0.167843,
		0.215573,
		0.265625,
		0.317068,
		0.369091,
		0.420985,
		0.472143,
		0.522046,
		0.570256,
		0.61641,
		0.660215,
		0.701439,
		0.739908,
		0.775502,
		0.808146,
		0.837812,
		0.864511,
		0.888291,
		0.909232,
		0.927446,
		0.943071,
		0.956268,
		0.967218,
		0.97612,
		0.983187,
		0.988638,
		0.9927,
		0.995601,
		0.997561,
		0.998793,
		0.999493,
		0.999836,
		0.999967,
		0.999998,
		1,
		0.999998,
		0.99997,
		0.999858,
		0.999581,
		0.999046,
		0.998153,
		0.996806,
		0.994912,
		0.992386,
		0.989154,
		0.985151,
		0.980326,
		0.974637,
		0.968053,
		0.960553,
		0.952128,
		0.942774,
		0.932498,
		0.921311,
		0.909232,
		0.896287,
		0.882503,
		0.867913,
		0.852555,
		0.836467,
		0.819691,
		0.80227,
		0.78425,
		0.765677,
		0.746597,
		0.727057,
		0.707107,
		0.686792,
		0.666161,
		0.645261,
		0.624138,
		0.602837,
		0.581406,
		0.559887,
		0.538323,
		0.516759,
		0.495233,
		0.473788,
		0.452462,
		0.431293,
		0.410318,
		0.389573,
		0.369091,
		0.348906,
		0.32905,
		0.309553,
		0.290446,
		0.271757,
		0.253513,
		0.23574,
		0.218463,
		0.201706,
		0.185491,
		0.169842,
		0.154777,
		0.140317,
		0.12648,
		0.113285,
		0.100747,
		0.088884,
		0.077708,
		0.067236,
		0.057479,
		0.048451,
		0.040162,
		0.032625,
		0.025848,
		0.019841,
		0.014613,
		0.010171,
		0.006524,
		0.003677,
		0.001638,
		0.00041,
		0.006014,
		0.022747,
		0.048385,
		0.081301,
		0.120042,
		0.163312,
		0.209959,
		0.258963,
		0.309424,
		0.360554,
		0.411665,
		0.462163,
		0.511539,
		0.559361,
		0.605268,
		0.648967,
		0.690221,
		0.728851,
		0.764729,
		0.797771,
		0.827935,
		0.855222,
		0.879663,
		0.901325,
		0.920303,
		0.936718,
		0.950715,
		0.962458,
		0.97213,
		0.979927,
		0.986055,
		0.990728,
		0.994162,
		0.996571,
		0.998163,
		0.999134,
		0.999662,
		0.999903,
		0.999985,
		1,
		1,
		0.999994,
		0.999949,
		0.999801,
		0.99947,
		0.998863,
		0.997888,
		0.996452,
		0.994468,
		0.991857,
		0.988552,
		0.984492,
		0.97963,
		0.973927,
		0.967356,
		0.959898,
		0.951543,
		0.94229,
		0.932145,
		0.92112,
		0.909232,
		0.896506,
		0.882967,
		0.868649,
		0.853585,
		0.837812,
		0.82137,
		0.804301,
		0.786646,
		0.76845,
		0.749757,
		0.730613,
		0.711062,
		0.69115,
		0.670922,
		0.650423,
		0.629697,
		0.608789,
		0.587741,
		0.566596,
		0.545395,
		0.52418,
		0.502989,
		0.481862,
		0.460836,
		0.439948,
		0.419232,
		0.398725,
		0.378457,
		0.358463,
		0.338772,
		0.319414,
		0.300419,
		0.281814,
		0.263625,
		0.245878,
		0.228598,
		0.211807,
		0.195527,
		0.179781,
		0.164588,
		0.149968,
		0.135939,
		0.122519,
		0.109724,
		0.09757,
		0.086071,
		0.075242,
		0.065096,
		0.055645,
		0.046901,
		0.038875,
		0.031577,
		0.025016,
		0.019201,
		0.014141,
		0.009842,
		0.006313,
		0.003558,
		0.001584,
		0.000397,
		0.005827,
		0.022059,
		0.046964,
		0.078987,
		0.116734,
		0.15896,
		0.204557,
		0.25254,
		0.30204,
		0.352291,
		0.402626,
		0.452462,
		0.501302,
		0.548718,
		0.594354,
		0.637915,
		0.679164,
		0.717915,
		0.754033,
		0.787425,
		0.818041,
		0.845868,
		0.870925,
		0.893264,
		0.912966,
		0.930138,
		0.944907,
		0.957423,
		0.967852,
		0.976376,
		0.983187,
		0.988485,
		0.992476,
		0.995364,
		0.997353,
		0.998635,
		0.999391,
		0.999782,
		0.999946,
		0.999994,
		1,
		1,
		0.999987,
		0.99992,
		0.999732,
		0.999343,
		0.998665,
		0.997607,
		0.996084,
		0.994013,
		0.991323,
		0.987949,
		0.983837,
		0.978942,
		0.973229,
		0.966673,
		0.959257,
		0.950973,
		0.94182,
		0.931803,
		0.920935,
		0.909232,
		0.896717,
		0.883417,
		0.869359,
		0.854579,
		0.839109,
		0.82299,
		0.806258,
		0.788955,
		0.771123,
		0.752804,
		0.73404,
		0.714875,
		0.695352,
		0.675515,
		0.655405,
		0.635065,
		0.614537,
		0.593863,
		0.573084,
		0.552238,
		0.531365,
		0.510503,
		0.48969,
		0.468961,
		0.448352,
		0.427896,
		0.407628,
		0.387578,
		0.367779,
		0.34826,
		0.32905,
		0.310176,
		0.291667,
		0.273547,
		0.255841,
		0.238574,
		0.221767,
		0.205443,
		0.189623,
		0.174326,
		0.159571,
		0.145377,
		0.131761,
		0.118739,
		0.106327,
		0.094539,
		0.083389,
		0.072891,
		0.063057,
		0.053897,
		0.045425,
		0.037648,
		0.030578,
		0.024223,
		0.018592,
		0.013691,
		0.009529,
		0.006111,
		0.003444,
		0.001534,
		0.000384,
		0.005649,
		0.021402,
		0.045605,
		0.07677,
		0.113559,
		0.154777,
		0.199356,
		0.246346,
		0.294906,
		0.344293,
		0.393858,
		0.443032,
		0.491327,
		0.538323,
		0.583667,
		0.627063,
		0.668272,
		0.707107,
		0.743424,
		0.777124,
		0.808146,
		0.836467,
		0.862095,
		0.88507,
		0.905457,
		0.92335,
		0.938862,
		0.952128,
		0.963299,
		0.972543,
		0.980037,
		0.98597,
		0.990536,
		0.99393,
		0.996348,
		0.997978,
		0.999001,
		0.999581,
		0.999864,
		0.999973,
		0.999998,
		1,
		0.999998,
		0.999975,
		0.999882,
		0.999651,
		0.999202,
		0.998451,
		0.997312,
		0.995704,
		0.993551,
		0.990785,
		0.987347,
		0.983187,
		0.978262,
		0.972543,
		0.966004,
		0.958631,
		0.950418,
		0.941362,
		0.931471,
		0.920755,
		0.909232,
		0.896922,
		0.883851,
		0.870045,
		0.855538,
		0.840361,
		0.824552,
		0.808146,
		0.791183,
		0.773702,
		0.755743,
		0.737347,
		0.718555,
		0.699408,
		0.679948,
		0.660215,
		0.64025,
		0.620093,
		0.599783,
		0.579359,
		0.55886,
		0.538323,
		0.517785,
		0.497281,
		0.476846,
		0.456513,
		0.436317,
		0.416289,
		0.39646,
		0.37686,
		0.357518,
		0.338462,
		0.319719,
		0.301315,
		0.283276,
		0.265625,
		0.248386,
		0.23158,
		0.21523,
		0.199356,
		0.183976,
		0.169111,
		0.154777,
		0.140991,
		0.127771,
		0.11513,
		0.103084,
		0.091646,
		0.08083,
		0.070648,
		0.061111,
		0.05223,
		0.044016,
		0.036479,
		0.029626,
		0.023468,
		0.018011,
		0.013263,
		0.00923,
		0.00592,
		0.003336,
		0.001486,
		0.000372,
		0.005478,
		0.020774,
		0.044304,
		0.074644,
		0.110511,
		0.150754,
		0.194346,
		0.24037,
		0.288011,
		0.336549,
		0.385352,
		0.433865,
		0.48161,
		0.528173,
		0.573204,
		0.61641,
		0.657551,
		0.696434,
		0.732912,
		0.766878,
		0.798264,
		0.827036,
		0.853193,
		0.876762,
		0.897796,
		0.916375,
		0.9326,
		0.946591,
		0.958486,
		0.968439,
		0.976614,
		0.983187,
		0.98834,
		0.992261,
		0.995135,
		0.997147,
		0.998474,
		0.999281,
		0.99972,
		0.99992,
		0.999988,
		1,
		1,
		0.999995,
		0.999957,
		0.999834,
		0.999556,
		0.999046,
		0.998222,
		0.997003,
		0.995313,
		0.993081,
		0.990244,
		0.986746,
		0.982541,
		0.977591,
		0.971868,
		0.965348,
		0.95802,
		0.949876,
		0.940917,
		0.931149,
		0.920582,
		0.909232,
		0.897121,
		0.88427,
		0.870708,
		0.856465,
		0.84157,
		0.82606,
		0.809968,
		0.793332,
		0.77619,
		0.758579,
		0.740538,
		0.722107,
		0.703324,
		0.68423,
		0.664862,
		0.645261,
		0.625463,
		0.605508,
		0.585432,
		0.565272,
		0.545064,
		0.524843,
		0.504643,
		0.484498,
		0.464441,
		0.444504,
		0.424716,
		0.405109,
		0.385711,
		0.366551,
		0.347655,
		0.32905,
		0.310761,
		0.292812,
		0.275228,
		0.258031,
		0.241242,
		0.224882,
		0.208972,
		0.193529,
		0.178574,
		0.164123,
		0.150192,
		0.136799,
		0.123957,
		0.111681,
		0.099986,
		0.088884,
		0.078386,
		0.068506,
		0.059254,
		0.050639,
		0.042672,
		0.035363,
		0.028718,
		0.022747,
		0.017457,
		0.012854,
		0.008946,
		0.005737,
		0.003233,
		0.00144,
		0.000361,
		0.005315,
		0.020173,
		0.043058,
		0.072605,
		0.107582,
		0.146884,
		0.18952,
		0.234603,
		0.281345,
		0.32905,
		0.377099,
		0.424954,
		0.472143,
		0.518262,
		0.562964,
		0.605958,
		0.647002,
		0.685901,
		0.722505,
		0.756699,
		0.788409,
		0.817592,
		0.844236,
		0.868358,
		0.890001,
		0.909232,
		0.926139,
		0.94083,
		0.953429,
		0.964076,
		0.972926,
		0.98014,
		0.98589,
		0.990352,
		0.993706,
		0.996128,
		0.997793,
		0.998863,
		0.999493,
		0.999819,
		0.999956,
		0.999995,
		1,
		1,
		0.999989,
		0.999933,
		0.999776,
		0.999448,
		0.998875,
		0.997978,
		0.996681,
		0.994912,
		0.992605,
		0.9897,
		0.986146,
		0.981901,
		0.976929,
		0.971204,
		0.964706,
		0.957423,
		0.949349,
		0.940484,
		0.930835,
		0.920413,
		0.909232,
		0.897313,
		0.884676,
		0.87135,
		0.85736,
		0.842738,
		0.827516,
		0.811728,
		0.795408,
		0.778592,
		0.761317,
		0.743619,
		0.725537,
		0.707107,
		0.688367,
		0.669354,
		0.650105,
		0.630658,
		0.611048,
		0.591311,
		0.571482,
		0.551596,
		0.531686,
		0.511786,
		0.491928,
		0.472143,
		0.452462,
		0.432915,
		0.413531,
		0.394338,
		0.375363,
		0.356632,
		0.338171,
		0.320005,
		0.302157,
		0.28465,
		0.267506,
		0.250747,
		0.234393,
		0.218463,
		0.202976,
		0.18795,
		0.173402,
		0.159349,
		0.145806,
		0.132788,
		0.120309,
		0.108384,
		0.097025,
		0.086243,
		0.076051,
		0.06646,
		0.057479,
		0.049119,
		0.041389,
		0.034297,
		0.027851,
		0.022059,
		0.016928,
		0.012464,
		0.008674,
		0.005159,
		0.019597,
		0.041863,
		0.070648,
		0.104768,
		0.14316,
		0.184867,
		0.229035,
		0.2749,
		0.321785,
		0.369091,
		0.416289,
		0.462921,
		0.508587,
		0.552945,
		0.595706,
		0.636628,
		0.675515,
		0.71221,
		0.746597,
		0.778592,
		0.808146,
		0.835238,
		0.859876,
		0.882091,
		0.901939,
		0.919497,
		0.93486,
		0.948142,
		0.959469,
		0.968983,
		0.976835,
		0.983187,
		0.988203,
		0.992055,
		0.994912,
		0.996943,
		0.99831,
		0.999166,
		0.999651,
		0.999887,
		0.999977,
		0.999999,
		1,
		0.999999,
		0.999979,
		0.999901,
		0.999707,
		0.999327,
		0.99869,
		0.997721,
		0.996348,
		0.994502,
		0.992122,
		0.989154,
		0.985548,
		0.981267,
		0.976276,
		0.970552,
		0.964076,
		0.956839,
		0.948834,
		0.940063,
		0.930531,
		0.920249,
		0.909232,
		0.897499,
		0.88507,
		0.87197,
		0.858226,
		0.843868,
		0.828924,
		0.813428,
		0.797413,
		0.780913,
		0.763962,
		0.746597,
		0.728851,
		0.710763,
		0.692366,
		0.673697,
		0.654791,
		0.635684,
		0.61641,
		0.597004,
		0.577498,
		0.557927,
		0.538323,
		0.518718,
		0.499142,
		0.479627,
		0.460201,
		0.440894,
		0.421733,
		0.402746,
		0.383959,
		0.365398,
		0.347086,
		0.32905,
		0.31131,
		0.29389,
		0.276811,
		0.260094,
		0.243759,
		0.227824,
		0.212308,
		0.197229,
		0.182603,
		0.168448,
		0.154777,
		0.141606,
		0.128949,
		0.116819,
		0.105229,
		0.094192,
		0.083718,
		0.073818,
		0.064503,
		0.055783,
		0.047667,
		0.040162,
		0.033279,
		0.027023,
		0.00501,
		0.019046,
		0.040718,
		0.068768,
		0.102062,
		0.139573,
		0.18038,
		0.223658,
		0.268666,
		0.314747,
		0.361318,
		0.407865,
		0.453937,
		0.499142,
		0.543144,
		0.585653,
		0.62643,
		0.665277,
		0.702034,
		0.736579,
		0.768825,
		0.798713,
		0.826216,
		0.851331,
		0.874082,
		0.894513,
		0.912691,
		0.9287,
		0.942641,
		0.95463,
		0.964797,
		0.973282,
		0.980236,
		0.985814,
		0.990178,
		0.99349,
		0.995914,
		0.997607,
		0.998722,
		0.999399,
		0.999766,
		0.999933,
		0.99999,
		1,
		1,
		0.999996,
		0.999964,
		0.999861,
		0.999626,
		0.999193,
		0.998492,
		0.997451,
		0.996003,
		0.994084,
		0.991635,
		0.988607,
		0.984953,
		0.980638,
		0.975631,
		0.969911,
		0.963459,
		0.956268,
		0.948332,
		0.939652,
		0.930235,
		0.92009,
		0.909232,
		0.897679,
		0.885451,
		0.872571,
		0.859064,
		0.84496,
		0.830285,
		0.815072,
		0.799352,
		0.783156,
		0.76652,
		0.749475,
		0.732056,
		0.714298,
		0.696234,
		0.677898,
		0.659326,
		0.640549,
		0.621603,
		0.602518,
		0.583329,
		0.564067,
		0.544762,
		0.525446,
		0.506149,
		0.4869,
		0.467727,
		0.448658,
		0.429721,
		0.410941,
		0.392345,
		0.373956,
		0.355799,
		0.337898,
		0.320274,
		0.302949,
		0.285944,
		0.26928,
		0.252975,
		0.237049,
		0.22152,
		0.206404,
		0.191718,
		0.177478,
		0.163699,
		0.150396,
		0.137583,
		0.125272,
		0.113477,
		0.102209,
		0.09148,
		0.081301,
		0.071682,
		0.062632,
		0.054161,
		0.004868,
		0.018518,
		0.039618,
		0.066963,
		0.099458,
		0.136118,
		0.176052,
		0.218463,
		0.262634,
		0.307926,
		0.353774,
		0.399673,
		0.445185,
		0.489923,
		0.533556,
		0.575799,
		0.61641,
		0.655192,
		0.691982,
		0.726654,
		0.759116,
		0.789303,
		0.817181,
		0.842738,
		0.86599,
		0.886971,
		0.905738,
		0.922365,
		0.936941,
		0.949573,
		0.960379,
		0.969489,
		0.977043,
		0.983187,
		0.988073,
		0.991857,
		0.994696,
		0.996742,
		0.998145,
		0.999046,
		0.999575,
		0.999848,
		0.999963,
		0.999996,
		1,
		1,
		0.999991,
		0.999944,
		0.999811,
		0.999533,
		0.999046,
		0.99828,
		0.997169,
		0.995649,
		0.993658,
		0.991144,
		0.988059,
		0.984361,
		0.980016,
		0.974996,
		0.969281,
		0.962855,
		0.95571,
		0.947842,
		0.939252,
		0.929947,
		0.919935,
		0.909232,
		0.897854,
		0.88582,
		0.873152,
		0.859876,
		0.846017,
		0.831602,
		0.816662,
		0.801227,
		0.785326,
		0.768993,
		0.752258,
		0.735156,
		0.717717,
		0.699976,
		0.681965,
		0.663715,
		0.645261,
		0.626633,
		0.607863,
		0.588982,
		0.570021,
		0.55101,
		0.531979,
		0.512956,
		0.493969,
		0.475047,
		0.456215,
		0.437501,
		0.418929,
		0.400525,
		0.382312,
		0.364313,
		0.346552,
		0.32905,
		0.311827,
		0.294906,
		0.278304,
		0.262041,
		0.246136,
		0.230606,
		0.215467,
		0.200737,
		0.18643,
		0.172561,
		0.159146,
		0.146196,
		0.133727,
		0.121749,
		0.110275,
		0.099316,
		0.088884,
		0.004731,
		0.018011,
		0.038563,
		0.065227,
		0.096952,
		0.132788,
		0.171875,
		0.213443,
		0.256796,
		0.301315,
		0.346449,
		0.391707,
		0.436658,
		0.480925,
		0.52418,
		0.56614,
		0.606567,
		0.645261,
		0.682058,
		0.716828,
		0.749475,
		0.779928,
		0.808146,
		0.834112,
		0.85783,
		0.879329,
		0.898654,
		0.91587,
		0.931058,
		0.944313,
		0.955742,
		0.965467,
		0.973615,
		0.980326,
		0.985742,
		0.990011,
		0.993282,
		0.995704,
		0.997422,
		0.998577,
		0.999299,
		0.999707,
		0.999905,
		0.999981,
		0.999999,
		1,
		0.999999,
		0.999982,
		0.999917,
		0.999752,
		0.999429,
		0.998886,
		0.998056,
		0.996876,
		0.995285,
		0.993226,
		0.99065,
		0.987511,
		0.983772,
		0.9794,
		0.974369,
		0.968661,
		0.962262,
		0.955164,
		0.947363,
		0.938862,
		0.929666,
		0.919785,
		0.909232,
		0.898023,
		0.886177,
		0.873716,
		0.860662,
		0.84704,
		0.832877,
		0.818201,
		0.803041,
		0.787425,
		0.771385,
		0.754952,
		0.738155,
		0.721027,
		0.703599,
		0.685901,
		0.667967,
		0.649825,
		0.631507,
		0.613044,
		0.594464,
		0.575799,
		0.557076,
		0.538323,
		0.51957,
		0.500843,
		0.482168,
		0.463572,
		0.44508,
		0.426716,
		0.408505,
		0.390469,
		0.372632,
		0.355015,
		0.33764,
		0.320527,
		0.303696,
		0.287165,
		0.270954,
		0.255081,
		0.239562,
		0.224414,
		0.209654,
		0.195295,
		0.181354,
		0.167843,
		0.154777,
		0.142168,
		0.130029,
		0.0046,
		0.017525,
		0.037549,
		0.063557,
		0.094539,
		0.129577,
		0.167843,
		0.208589,
		0.251144,
		0.294906,
		0.339336,
		0.383959,
		0.428351,
		0.472143,
		0.515011,
		0.556677,
		0.596902,
		0.635486,
		0.672265,
		0.707107,
		0.739908,
		0.770596,
		0.799123,
		0.825463,
		0.849616,
		0.8716,
		0.891454,
		0.909232,
		0.925006,
		0.938862,
		0.950898,
		0.961225,
		0.969962,
		0.977237,
		0.983187,
		0.987949,
		0.991668,
		0.994485,
		0.996543,
		0.997978,
		0.998921,
		0.999493,
		0.999803,
		0.999944,
		0.999991,
		1,
		1,
		0.999997,
		0.99997,
		0.999882,
		0.999682,
		0.999313,
		0.998713,
		0.997819,
		0.996571,
		0.994912,
		0.992789,
		0.990153,
		0.986964,
		0.983187,
		0.97879,
		0.973751,
		0.968053,
		0.961681,
		0.95463,
		0.946896,
		0.938482,
		0.929393,
		0.919639,
		0.909232,
		0.898188,
		0.886525,
		0.874262,
		0.861423,
		0.848031,
		0.834112,
		0.819691,
		0.804797,
		0.789457,
		0.773702,
		0.757559,
		0.741059,
		0.724231,
		0.707107,
		0.689715,
		0.672085,
		0.654248,
		0.636233,
		0.618068,
		0.599783,
		0.581406,
		0.562964,
		0.544486,
		0.525998,
		0.507526,
		0.489096,
		0.470734,
		0.452462,
		0.434306,
		0.416289,
		0.398433,
		0.38076,
		0.363292,
		0.346048,
		0.32905,
		0.312315,
		0.295864,
		0.279714,
		0.263882,
		0.248386,
		0.233241,
		0.218463,
		0.204067,
		0.190067,
		0.176478,
		0.004475,
		0.017058,
		0.036574,
		0.061951,
		0.092214,
		0.12648,
		0.163949,
		0.203896,
		0.245671,
		0.28869,
		0.332429,
		0.376422,
		0.420258,
		0.463572,
		0.506046,
		0.547405,
		0.587413,
		0.625869,
		0.662608,
		0.697495,
		0.730424,
		0.761317,
		0.79012,
		0.816804,
		0.84136,
		0.863799,
		0.884152,
		0.902465,
		0.9188,
		0.933236,
		0.94586,
		0.956775,
		0.96609,
		0.973927,
		0.980411,
		0.985674,
		0.989852,
		0.993081,
		0.995499,
		0.997238,
		0.998429,
		0.999193,
		0.999641,
		0.999872,
		0.999969,
		0.999997,
		1,
		1,
		0.999992,
		0.999952,
		0.999839,
		0.999603,
		0.999185,
		0.998528,
		0.997571,
		0.996257,
		0.994532,
		0.992346,
		0.989654,
		0.986419,
		0.982605,
		0.978187,
		0.973143,
		0.967454,
		0.961112,
		0.954108,
		0.94644,
		0.938112,
		0.929128,
		0.919497,
		0.909232,
		0.898348,
		0.886861,
		0.874792,
		0.862161,
		0.848992,
		0.835308,
		0.821134,
		0.806498,
		0.791425,
		0.775945,
		0.760084,
		0.743871,
		0.727336,
		0.710506,
		0.69341,
		0.676077,
		0.658536,
		0.640815,
		0.622942,
		0.604944,
		0.586849,
		0.568684,
		0.550474,
		0.532247,
		0.514027,
		0.495839,
		0.477707,
		0.459655,
		0.441707,
		0.423884,
		0.406208,
		0.388702,
		0.371384,
		0.354276,
		0.337397,
		0.320766,
		0.304401,
		0.288319,
		0.272538,
		0.257074,
		0.241943,
		0.227159,
		0.004354,
		0.01661,
		0.035637,
		0.060404,
		0.089974,
		0.123492,
		0.160186,
		0.199356,
		0.24037,
		0.282661,
		0.325719,
		0.369091,
		0.412372,
		0.455206,
		0.497281,
		0.538323,
		0.5781,
		0.61641,
		0.653088,
		0.687996,
		0.721027,
		0.752097,
		0.781149,
		0.808146,
		0.833075,
		0.855939,
		0.876762,
		0.895582,
		0.912454,
		0.927446,
		0.94064,
		0.952128,
		0.962012,
		0.970403,
		0.97742,
		0.983187,
		0.987832,
		0.991486,
		0.994282,
		0.996348,
		0.997811,
		0.998793,
		0.999406,
		0.999752,
		0.99992,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999985,
		0.999929,
		0.999788,
		0.999512,
		0.999046,
		0.998331,
		0.997312,
		0.995933,
		0.994144,
		0.991898,
		0.989154,
		0.985874,
		0.982029,
		0.977591,
		0.972543,
		0.966867,
		0.960553,
		0.953597,
		0.945995,
		0.937751,
		0.928869,
		0.919359,
		0.909232,
		0.898503,
		0.887188,
		0.875306,
		0.862877,
		0.849923,
		0.836467,
		0.822533,
		0.808146,
		0.793332,
		0.778118,
		0.762531,
		0.746597,
		0.730344,
		0.7138,
		0.696992,
		0.679948,
		0.662695,
		0.645261,
		0.627672,
		0.609954,
		0.592135,
		0.57424,
		0.556295,
		0.538323,
		0.520351,
		0.502402,
		0.484498,
		0.466665,
		0.448922,
		0.431293,
		0.413799,
		0.39646,
		0.379296,
		0.362328,
		0.345573,
		0.32905,
		0.312776,
		0.29677,
		0.281048,
		0.004238,
		0.016179,
		0.034735,
		0.058915,
		0.087813,
		0.120607,
		0.156549,
		0.194962,
		0.235233,
		0.276811,
		0.3192,
		0.361957,
		0.404687,
		0.447041,
		0.488711,
		0.529428,
		0.568961,
		0.607109,
		0.643707,
		0.678615,
		0.711723,
		0.742944,
		0.772216,
		0.799498,
		0.82477,
		0.848031,
		0.869296,
		0.888597,
		0.90598,
		0.921508,
		0.935251,
		0.947296,
		0.957735,
		0.966673,
		0.974219,
		0.980491,
		0.98561,
		0.9897,
		0.992887,
		0.995298,
		0.997056,
		0.99828,
		0.999083,
		0.99957,
		0.999833,
		0.999952,
		0.999993,
		1,
		1,
		0.999997,
		0.999974,
		0.999899,
		0.999729,
		0.999412,
		0.998895,
		0.998123,
		0.997042,
		0.995601,
		0.99375,
		0.991447,
		0.988652,
		0.985332,
		0.981456,
		0.977002,
		0.971951,
		0.966289,
		0.960006,
		0.953097,
		0.94556,
		0.937398,
		0.928617,
		0.919224,
		0.909232,
		0.898654,
		0.887506,
		0.875806,
		0.863572,
		0.850827,
		0.837591,
		0.823889,
		0.809744,
		0.795181,
		0.780225,
		0.764902,
		0.749239,
		0.73326,
		0.716994,
		0.700466,
		0.683702,
		0.66673,
		0.649575,
		0.632263,
		0.61482,
		0.59727,
		0.57964,
		0.561953,
		0.544233,
		0.526505,
		0.50879,
		0.491113,
		0.473495,
		0.455958,
		0.438523,
		0.42121,
		0.40404,
		0.387032,
		0.370205,
		0.353578,
		0.337168,
		0.004127,
		0.015765,
		0.033867,
		0.057479,
		0.085729,
		0.117821,
		0.153034,
		0.19071,
		0.230255,
		0.271135,
		0.312866,
		0.355015,
		0.397198,
		0.439072,
		0.480333,
		0.520718,
		0.559995,
		0.597967,
		0.634466,
		0.669354,
		0.702516,
		0.733863,
		0.763329,
		0.790869,
		0.816458,
		0.840087,
		0.861767,
		0.881522,
		0.899393,
		0.915432,
		0.929706,
		0.94229,
		0.953272,
		0.962746,
		0.970816,
		0.977591,
		0.983187,
		0.98772,
		0.991312,
		0.994084,
		0.996156,
		0.997644,
		0.998662,
		0.999313,
		0.999695,
		0.999891,
		0.999973,
		0.999997,
		1,
		1,
		0.999993,
		0.999959,
		0.999863,
		0.99966,
		0.9993,
		0.998733,
		0.997904,
		0.996763,
		0.99526,
		0.99335,
		0.990993,
		0.98815,
		0.984791,
		0.980889,
		0.97642,
		0.971369,
		0.965721,
		0.959469,
		0.952607,
		0.945135,
		0.937054,
		0.928371,
		0.919093,
		0.909232,
		0.898801,
		0.887815,
		0.87629,
		0.864247,
		0.851704,
		0.838682,
		0.825205,
		0.811294,
		0.796974,
		0.782268,
		0.767202,
		0.751801,
		0.736089,
		0.720092,
		0.703836,
		0.687346,
		0.670646,
		0.653763,
		0.636722,
		0.619546,
		0.60226,
		0.584888,
		0.567455,
		0.549982,
		0.532493,
		0.515011,
		0.497557,
		0.480153,
		0.46282,
		0.445578,
		0.428447,
		0.411446,
		0.394596,
		0.00402,
		0.015366,
		0.033031,
		0.056095,
		0.083718,
		0.11513,
		0.149633,
		0.186592,
		0.225429,
		0.265625,
		0.306709,
		0.34826,
		0.3899,
		0.431293,
		0.472143,
		0.512188,
		0.5512,
		0.588982,
		0.625367,
		0.660215,
		0.69341,
		0.72486,
		0.754496,
		0.782268,
		0.808146,
		0.832117,
		0.854185,
		0.874369,
		0.892703,
		0.909232,
		0.924016,
		0.937122,
		0.948631,
		0.958631,
		0.967218,
		0.974494,
		0.980567,
		0.985548,
		0.989554,
		0.9927,
		0.995103,
		0.996876,
		0.99813,
		0.99897,
		0.999493,
		0.999788,
		0.999932,
		0.999986,
		0.999999,
		1,
		0.999999,
		0.999987,
		0.999939,
		0.999819,
		0.999581,
		0.999178,
		0.99856,
		0.997675,
		0.996474,
		0.994912,
		0.992945,
		0.990536,
		0.987648,
		0.984253,
		0.980326,
		0.975845,
		0.970795,
		0.965164,
		0.958943,
		0.952128,
		0.944719,
		0.936718,
		0.928131,
		0.918966,
		0.909232,
		0.898944,
		0.888115,
		0.876762,
		0.864902,
		0.852555,
		0.839741,
		0.826481,
		0.812798,
		0.798713,
		0.78425,
		0.769433,
		0.754286,
		0.738833,
		0.723099,
		0.707107,
		0.690882,
		0.674449,
		0.657831,
		0.641052,
		0.624138,
		0.607109,
		0.589991,
		0.572806,
		0.555576,
		0.538323,
		0.52107,
		0.503836,
		0.486644,
		0.469512,
		0.452462,
		0.003918,
		0.014982,
		0.032226,
		0.054761,
		0.081776,
		0.112529,
		0.146343,
		0.182603,
		0.22075,
		0.260276,
		0.300724,
		0.341684,
		0.382786,
		0.423701,
		0.464137,
		0.503836,
		0.542574,
		0.580154,
		0.61641,
		0.651201,
		0.684409,
		0.71594,
		0.745722,
		0.773702,
		0.799844,
		0.824131,
		0.846562,
		0.867151,
		0.885923,
		0.90292,
		0.918192,
		0.931803,
		0.943825,
		0.954338,
		0.963433,
		0.971204,
		0.977753,
		0.983187,
		0.987613,
		0.991144,
		0.993892,
		0.995967,
		0.997478,
		0.998528,
		0.999216,
		0.999632,
		0.999857,
		0.999959,
		0.999994,
		1,
		1,
		0.999998,
		0.999978,
		0.999914,
		0.999767,
		0.999493,
		0.999046,
		0.998376,
		0.997435,
		0.996177,
		0.994557,
		0.992536,
		0.990077,
		0.987147,
		0.983719,
		0.979768,
		0.975277,
		0.97023,
		0.964615,
		0.958426,
		0.951659,
		0.944313,
		0.93639,
		0.927897,
		0.918841,
		0.909232,
		0.899083,
		0.888407,
		0.877219,
		0.865538,
		0.853382,
		0.840769,
		0.827721,
		0.814257,
		0.800401,
		0.786174,
		0.771599,
		0.756699,
		0.741497,
		0.726017,
		0.710282,
		0.694316,
		0.678141,
		0.661782,
		0.645261,
		0.628601,
		0.611825,
		0.594955,
		0.578013,
		0.561021,
		0.544,
		0.526971,
		0.509955,
		0.003819,
		0.014613,
		0.031449,
		0.053473,
		0.079901,
		0.110014,
		0.14316,
		0.178739,
		0.21621,
		0.255081,
		0.294906,
		0.335282,
		0.375851,
		0.416289,
		0.45631,
		0.495659,
		0.534114,
		0.571482,
		0.607595,
		0.642312,
		0.675515,
		0.707107,
		0.737013,
		0.765177,
		0.791559,
		0.816138,
		0.838908,
		0.859876,
		0.879064,
		0.896506,
		0.912247,
		0.926344,
		0.938862,
		0.949876,
		0.959469,
		0.967729,
		0.974752,
		0.980638,
		0.98549,
		0.989415,
		0.99252,
		0.994912,
		0.996697,
		0.997978,
		0.998853,
		0.999412,
		0.999739,
		0.999907,
		0.999977,
		0.999998,
		1,
		1,
		0.999994,
		0.999965,
		0.999882,
		0.999707,
		0.999396,
		0.998903,
		0.998182,
		0.997186,
		0.995871,
		0.994196,
		0.992122,
		0.989616,
		0.986646,
		0.983187,
		0.979216,
		0.974716,
		0.969673,
		0.964076,
		0.95792,
		0.951199,
		0.943915,
		0.93607,
		0.927669,
		0.91872,
		0.909232,
		0.899218,
		0.88869,
		0.877665,
		0.866157,
		0.854185,
		0.841768,
		0.828924,
		0.815675,
		0.80204,
		0.788042,
		0.773702,
		0.759042,
		0.744084,
		0.728851,
		0.713366,
		0.697651,
		0.681729,
		0.665622,
		0.649351,
		0.63294,
		0.61641,
		0.599783,
		0.583079,
		0.566321,
		0.003724,
		0.014256,
		0.0307,
		0.05223,
		0.078089,
		0.107582,
		0.140077,
		0.174994,
		0.211807,
		0.250035,
		0.289247,
		0.32905,
		0.369091,
		0.409054,
		0.448658,
		0.487653,
		0.52582,
		0.562964,
		0.598922,
		0.63355,
		0.66673,
		0.698364,
		0.728373,
		0.756699,
		0.783299,
		0.808146,
		0.83123,
		0.852555,
		0.872136,
		0.890001,
		0.906191,
		0.920755,
		0.933754,
		0.945255,
		0.955335,
		0.964076,
		0.971569,
		0.977906,
		0.983187,
		0.987511,
		0.990983,
		0.993706,
		0.995782,
		0.997312,
		0.998392,
		0.999115,
		0.999565,
		0.999819,
		0.999942,
		0.999988,
		0.999999,
		1,
		0.999999,
		0.999989,
		0.999948,
		0.999844,
		0.999638,
		0.999289,
		0.99875,
		0.997978,
		0.996928,
		0.995558,
		0.993829,
		0.991705,
		0.989154,
		0.986146,
		0.982658,
		0.978669,
		0.974162,
		0.969125,
		0.963547,
		0.957423,
		0.950749,
		0.943527,
		0.935757,
		0.927446,
		0.918602,
		0.909232,
		0.89935,
		0.888967,
		0.878098,
		0.866759,
		0.854966,
		0.842738,
		0.830094,
		0.817052,
		0.803632,
		0.789856,
		0.775744,
		0.761317,
		0.746597,
		0.731605,
		0.716363,
		0.700893,
		0.685216,
		0.669354,
		0.653328,
		0.63716,
		0.620871,
		0.003632,
		0.013913,
		0.029978,
		0.05103,
		0.076337,
		0.105229,
		0.137092,
		0.171364,
		0.207533,
		0.245133,
		0.283743,
		0.32298,
		0.362499,
		0.40199,
		0.441178,
		0.479816,
		0.517687,
		0.5546,
		0.59039,
		0.624916,
		0.658057,
		0.689715,
		0.719808,
		0.748275,
		0.77507,
		0.800163,
		0.823539,
		0.845197,
		0.865149,
		0.883417,
		0.900035,
		0.915048,
		0.928511,
		0.940484,
		0.951039,
		0.960253,
		0.968209,
		0.974996,
		0.980706,
		0.985435,
		0.989282,
		0.992346,
		0.994726,
		0.996521,
		0.997826,
		0.998733,
		0.999326,
		0.999684,
		0.999877,
		0.999965,
		0.999995,
		1,
		1,
		0.999998,
		0.999981,
		0.999926,
		0.999798,
		0.999561,
		0.999172,
		0.998588,
		0.997765,
		0.996662,
		0.995238,
		0.993458,
		0.991285,
		0.988691,
		0.985648,
		0.982133,
		0.978128,
		0.973615,
		0.968585,
		0.963026,
		0.956935,
		0.950308,
		0.943146,
		0.935451,
		0.927229,
		0.918486,
		0.909232,
		0.899478,
		0.889236,
		0.878519,
		0.867344,
		0.855726,
		0.843682,
		0.83123,
		0.81839,
		0.805179,
		0.791619,
		0.777728,
		0.763528,
		0.749038,
		0.73428,
		0.719275,
		0.704043,
		0.688606,
		0.672983,
		0.003544,
		0.013582,
		0.029281,
		0.049871,
		0.074644,
		0.102952,
		0.1342,
		0.167843,
		0.203384,
		0.24037,
		0.278389,
		0.317068,
		0.356071,
		0.395094,
		0.433865,
		0.472143,
		0.509713,
		0.546387,
		0.582,
		0.61641,
		0.649498,
		0.681162,
		0.71132,
		0.739908,
		0.766878,
		0.792195,
		0.815842,
		0.837812,
		0.858112,
		0.876762,
		0.893788,
		0.909232,
		0.923141,
		0.935573,
		0.946591,
		0.956268,
		0.964681,
		0.971912,
		0.978051,
		0.983187,
		0.987414,
		0.990828,
		0.993526,
		0.995601,
		0.997147,
		0.998255,
		0.999011,
		0.999493,
		0.999775,
		0.99992,
		0.99998,
		0.999998,
		1,
		1,
		0.999995,
		0.99997,
		0.999898,
		0.999746,
		0.999475,
		0.999046,
		0.998415,
		0.997543,
		0.996388,
		0.994912,
		0.993081,
		0.990862,
		0.988228,
		0.985151,
		0.981612,
		0.977591,
		0.973075,
		0.968053,
		0.962515,
		0.956457,
		0.949876,
		0.942774,
		0.935153,
		0.927017,
		0.918374,
		0.909232,
		0.899603,
		0.889497,
		0.878929,
		0.867913,
		0.856465,
		0.8446,
		0.832336,
		0.819691,
		0.806683,
		0.793332,
		0.779657,
		0.765677,
		0.751412,
		0.736881,
		0.722107,
		0.003458,
		0.013263,
		0.028608,
		0.04875,
		0.073006,
		0.100747,
		0.131398,
		0.164428,
		0.199356,
		0.23574,
		0.273179,
		0.31131,
		0.349803,
		0.38836,
		0.426716,
		0.464632,
		0.501896,
		0.538323,
		0.573749,
		0.608032,
		0.641052,
		0.672708,
		0.702914,
		0.731605,
		0.758729,
		0.78425,
		0.808146,
		0.830407,
		0.851035,
		0.870045,
		0.887461,
		0.903317,
		0.917657,
		0.930531,
		0.941999,
		0.952128,
		0.960989,
		0.968661,
		0.975226,
		0.98077,
		0.985382,
		0.989154,
		0.992177,
		0.994545,
		0.996348,
		0.997675,
		0.99861,
		0.999236,
		0.999625,
		0.999844,
		0.99995,
		0.99999,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999955,
		0.999864,
		0.999686,
		0.999381,
		0.99891,
		0.998234,
		0.997312,
		0.996106,
		0.99458,
		0.9927,
		0.990437,
		0.987764,
		0.984657,
		0.981094,
		0.977061,
		0.972543,
		0.967529,
		0.962012,
		0.955987,
		0.949453,
		0.94241,
		0.93486,
		0.926809,
		0.918264,
		0.909232,
		0.899725,
		0.889753,
		0.879329,
		0.868468,
		0.857183,
		0.845492,
		0.833411,
		0.820956,
		0.808146,
		0.794999,
		0.781532,
		0.767766,
		0.003376,
		0.012955,
		0.027957,
		0.047667,
		0.071421,
		0.098612,
		0.128681,
		0.161115,
		0.195443,
		0.231239,
		0.268109,
		0.3057,
		0.343689,
		0.381784,
		0.419726,
		0.457278,
		0.494234,
		0.530408,
		0.565638,
		0.599783,
		0.632722,
		0.664354,
		0.694592,
		0.723368,
		0.750628,
		0.776333,
		0.800458,
		0.82299,
		0.843926,
		0.863277,
		0.881063,
		0.897313,
		0.912065,
		0.925367,
		0.937272,
		0.947842,
		0.957143,
		0.965249,
		0.972236,
		0.978187,
		0.983187,
		0.987321,
		0.990679,
		0.99335,
		0.995423,
		0.996984,
		0.998117,
		0.998903,
		0.999417,
		0.999727,
		0.999894,
		0.99997,
		0.999995,
		1,
		1,
		0.999998,
		0.999984,
		0.999935,
		0.999825,
		0.999618,
		0.999278,
		0.998766,
		0.998044,
		0.997073,
		0.995817,
		0.994242,
		0.992316,
		0.990011,
		0.987301,
		0.984164,
		0.980581,
		0.976536,
		0.972017,
		0.967013,
		0.961517,
		0.955526,
		0.949038,
		0.942053,
		0.934575,
		0.926607,
		0.918157,
		0.909232,
		0.899843,
		0.890001,
		0.879718,
		0.869007,
		0.857883,
		0.846361,
		0.834457,
		0.822187,
		0.809569,
		0.003297,
		0.012657,
		0.027329,
		0.046619,
		0.069887,
		0.096544,
		0.126047,
		0.157899,
		0.191643,
		0.226862,
		0.263174,
		0.300233,
		0.337724,
		0.375363,
		0.412891,
		0.45008,
		0.486723,
		0.522638,
		0.557664,
		0.591661,
		0.624508,
		0.656102,
		0.686356,
		0.715201,
		0.742579,
		0.76845,
		0.792785,
		0.815567,
		0.836792,
		0.856465,
		0.874601,
		0.891227,
		0.906376,
		0.92009,
		0.932418,
		0.943417,
		0.953149,
		0.961681,
		0.969087,
		0.975444,
		0.980831,
		0.985332,
		0.989031,
		0.992015,
		0.994368,
		0.996177,
		0.997523,
		0.998486,
		0.999142,
		0.999561,
		0.999805,
		0.999931,
		0.999983,
		0.999998,
		1,
		1,
		0.999996,
		0.999974,
		0.999911,
		0.999779,
		0.999542,
		0.999166,
		0.998613,
		0.997845,
		0.996826,
		0.995521,
		0.993899,
		0.991928,
		0.989583,
		0.986838,
		0.983674,
		0.980072,
		0.976017,
		0.971498,
		0.966504,
		0.961031,
		0.955074,
		0.948631,
		0.941704,
		0.934295,
		0.926409,
		0.918052,
		0.909232,
		0.899959,
		0.890244,
		0.880097,
		0.869533,
		0.858565,
		0.847207,
		0.003221,
		0.01237,
		0.026722,
		0.045605,
		0.068402,
		0.094539,
		0.123492,
		0.154777,
		0.18795,
		0.222604,
		0.258369,
		0.294906,
		0.331906,
		0.369091,
		0.406208,
		0.443032,
		0.47936,
		0.515011,
		0.549827,
		0.583667,
		0.61641,
		0.647954,
		0.67821,
		0.707107,
		0.734587,
		0.760605,
		0.785132,
		0.808146,
		0.82964,
		0.849616,
		0.868085,
		0.88507,
		0.900598,
		0.914709,
		0.927446,
		0.938862,
		0.949014,
		0.957965,
		0.965784,
		0.972543,
		0.978317,
		0.983187,
		0.987232,
		0.990536,
		0.993181,
		0.995249,
		0.996822,
		0.997978,
		0.998793,
		0.999337,
		0.999675,
		0.999864,
		0.999956,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.999992,
		0.999961,
		0.999882,
		0.999726,
		0.999459,
		0.999046,
		0.998451,
		0.997637,
		0.996571,
		0.995219,
		0.993551,
		0.991538,
		0.989154,
		0.986377,
		0.983187,
		0.979567,
		0.975504,
		0.970985,
		0.966004,
		0.960553,
		0.95463,
		0.948233,
		0.941362,
		0.934022,
		0.926215,
		0.91795,
		0.909232,
		0.900072,
		0.89048,
		0.880467,
		0.003147,
		0.012092,
		0.026134,
		0.044624,
		0.066963,
		0.092596,
		0.121013,
		0.151745,
		0.18436,
		0.218463,
		0.25369,
		0.289713,
		0.326228,
		0.362964,
		0.399673,
		0.436133,
		0.472143,
		0.507526,
		0.542124,
		0.575799,
		0.608428,
		0.63991,
		0.670154,
		0.699089,
		0.726654,
		0.752804,
		0.777504,
		0.800733,
		0.822478,
		0.842738,
		0.861523,
		0.878848,
		0.89474,
		0.909232,
		0.922365,
		0.934185,
		0.944746,
		0.954108,
		0.962333,
		0.969489,
		0.97565,
		0.980889,
		0.985283,
		0.988913,
		0.991857,
		0.994196,
		0.996009,
		0.997372,
		0.998361,
		0.999046,
		0.999493,
		0.999763,
		0.999908,
		0.999974,
		0.999996,
		1,
		1,
		0.999998,
		0.999986,
		0.999944,
		0.999847,
		0.999666,
		0.999367,
		0.998917,
		0.99828,
		0.997422,
		0.99631,
		0.994912,
		0.993199,
		0.991144,
		0.988724,
		0.985916,
		0.982702,
		0.979066,
		0.974996,
		0.97048,
		0.965511,
		0.960083,
		0.954194,
		0.947842,
		0.941027,
		0.933754,
		0.926026,
		0.91785,
		0.909232,
		0.003075,
		0.011823,
		0.025566,
		0.043674,
		0.065569,
		0.090711,
		0.118608,
		0.148801,
		0.180871,
		0.214433,
		0.249134,
		0.28465,
		0.320688,
		0.35698,
		0.393283,
		0.429378,
		0.465069,
		0.50018,
		0.534555,
		0.568056,
		0.600563,
		0.631971,
		0.662191,
		0.69115,
		0.718785,
		0.74505,
		0.769907,
		0.793332,
		0.815311,
		0.835839,
		0.854921,
		0.872571,
		0.88881,
		0.903668,
		0.917181,
		0.929393,
		0.940353,
		0.950115,
		0.958739,
		0.966289,
		0.972832,
		0.97844,
		0.983187,
		0.987147,
		0.990397,
		0.993016,
		0.995079,
		0.996662,
		0.997839,
		0.998681,
		0.999253,
		0.999618,
		0.999831,
		0.99994,
		0.999985,
		0.999998,
		1,
		1,
		0.999996,
		0.999977,
		0.999923,
		0.999806,
		0.999599,
		0.999268,
		0.99878,
		0.998102,
		0.997199,
		0.996042,
		0.994599,
		0.992844,
		0.990749,
		0.988294,
		0.985457,
		0.98222,
		0.97857,
		0.974494,
		0.969981,
		0.965026,
		0.959621,
		0.953766,
		0.947458,
		0.940699,
		0.933492,
		0.003007,
		0.011564,
		0.025016,
		0.042755,
		0.064217,
		0.088884,
		0.116272,
		0.14594,
		0.177478,
		0.210511,
		0.244695,
		0.279714,
		0.315281,
		0.351133,
		0.387032,
		0.422765,
		0.458135,
		0.492972,
		0.527118,
		0.560439,
		0.592813,
		0.624138,
		0.654322,
		0.683291,
		0.710982,
		0.737347,
		0.762345,
		0.785951,
		0.808146,
		0.828924,
		0.848287,
		0.866244,
		0.882814,
		0.898023,
		0.911904,
		0.924495,
		0.935842,
		0.945995,
		0.95501,
		0.962947,
		0.96987,
		0.975845,
		0.980944,
		0.985237,
		0.9888,
		0.991705,
		0.994029,
		0.995843,
		0.997222,
		0.998234,
		0.998946,
		0.999422,
		0.999716,
		0.999882,
		0.999962,
		0.999992,
		1,
		1,
		1,
		0.999993,
		0.999966,
		0.999897,
		0.99976,
		0.999525,
		0.999161,
		0.998635,
		0.997915,
		0.99697,
		0.995768,
		0.994282,
		0.992484,
		0.990352,
		0.987863,
		0.984999,
		0.981742,
		0.978079,
		0.973997,
		0.969489,
		0.964547,
		0.959167,
		0.953345,
		0.00294,
		0.011312,
		0.024483,
		0.041863,
		0.062907,
		0.08711,
		0.114005,
		0.14316,
		0.174178,
		0.206693,
		0.24037,
		0.2749,
		0.310002,
		0.34542,
		0.380919,
		0.416289,
		0.451339,
		0.485897,
		0.519811,
		0.552945,
		0.58518,
		0.61641,
		0.646547,
		0.675515,
		0.703249,
		0.729698,
		0.754822,
		0.778592,
		0.800988,
		0.822,
		0.841627,
		0.859876,
		0.876762,
		0.892307,
		0.90654,
		0.919497,
		0.931219,
		0.941754,
		0.951151,
		0.959469,
		0.966766,
		0.973107,
		0.978558,
		0.983187,
		0.987065,
		0.990264,
		0.992856,
		0.994912,
		0.996504,
		0.9977,
		0.998566,
		0.999166,
		0.999557,
		0.999793,
		0.99992,
		0.999977,
		0.999996,
		1,
		1,
		0.999999,
		0.999988,
		0.999951,
		0.999866,
		0.999707,
		0.999443,
		0.999046,
		0.998482,
		0.997721,
		0.996733,
		0.995488,
		0.993959,
		0.992122,
		0.989954,
		0.987433,
		0.984543,
		0.981267,
		0.977591,
		0.973507,
		0.969004,
		0.002875,
		0.011069,
		0.023968,
		0.041,
		0.061637,
		0.085389,
		0.111802,
		0.140457,
		0.170967,
		0.202976,
		0.236155,
		0.270205,
		0.30485,
		0.339837,
		0.37494,
		0.409949,
		0.444677,
		0.478955,
		0.512632,
		0.545574,
		0.577661,
		0.608789,
		0.638868,
		0.667822,
		0.695586,
		0.722107,
		0.747342,
		0.771261,
		0.793842,
		0.815072,
		0.834948,
		0.853472,
		0.870658,
		0.886525,
		0.901097,
		0.914407,
		0.926493,
		0.937398,
		0.94717,
		0.955861,
		0.963527,
		0.97023,
		0.976031,
		0.980996,
		0.985193,
		0.988691,
		0.991558,
		0.993865,
		0.995681,
		0.997073,
		0.998106,
		0.998845,
		0.999346,
		0.999666,
		0.999852,
		0.999947,
		0.999987,
		0.999999,
		1,
		1,
		0.999997,
		0.99998,
		0.999932,
		0.99983,
		0.999647,
		0.999355,
		0.998923,
		0.998321,
		0.99752,
		0.99649,
		0.995203,
		0.993633,
		0.991758,
		0.989554,
		0.987004,
		0.984089,
		0.980795,
		0.002813,
		0.010834,
		0.023468,
		0.040162,
		0.060404,
		0.083718,
		0.109662,
		0.137829,
		0.167843,
		0.199356,
		0.232048,
		0.265625,
		0.299819,
		0.334382,
		0.369091,
		0.403741,
		0.438147,
		0.472143,
		0.50558,
		0.538323,
		0.570256,
		0.601273,
		0.631286,
		0.660215,
		0.687996,
		0.714575,
		0.739908,
		0.763962,
		0.786713,
		0.808146,
		0.828254,
		0.84704,
		0.864511,
		0.880684,
		0.895582,
		0.909232,
		0.92167,
		0.932935,
		0.943071,
		0.952128,
		0.960158,
		0.967218,
		0.973368,
		0.978669,
		0.983187,
		0.986986,
		0.990135,
		0.9927,
		0.994749,
		0.996348,
		0.997561,
		0.998451,
		0.999076,
		0.999493,
		0.999752,
		0.999897,
		0.999967,
		0.999993,
		1,
		1,
		1,
		0.999994,
		0.99997,
		0.999909,
		0.999788,
		0.999581,
		0.999259,
		0.998793,
		0.998153,
		0.997312,
		0.99624,
		0.994912,
		0.993303,
		0.99139,
		0.989154,
		0.002753,
		0.010606,
		0.022984,
		0.03935,
		0.059208,
		0.082095,
		0.107582,
		0.135274,
		0.164802,
		0.19583,
		0.228043,
		0.261156,
		0.294906,
		0.32905,
		0.363368,
		0.397661,
		0.431746,
		0.465459,
		0.498652,
		0.531193,
		0.562964,
		0.593863,
		0.623799,
		0.652694,
		0.680482,
		0.707107,
		0.732524,
		0.756699,
		0.779606,
		0.801227,
		0.821553,
		0.840584,
		0.858326,
		0.874792,
		0.890001,
		0.903979,
		0.916757,
		0.928371,
		0.938862,
		0.948277,
		0.956664,
		0.964076,
		0.970571,
		0.976208,
		0.981047,
		0.985151,
		0.988586,
		0.991416,
		0.993706,
		0.995521,
		0.996925,
		0.997978,
		0.998741,
		0.999268,
		0.999612,
		0.999819,
		0.99993,
		0.99998,
		0.999997,
		1,
		1,
		0.999999,
		0.999989,
		0.999957,
		0.999882,
		0.999741,
		0.999508,
		0.999156,
		0.998655,
		0.997978,
		0.997097,
		0.995985,
		0.994617,
		0.002694,
		0.010385,
		0.022514,
		0.038563,
		0.058047,
		0.080519,
		0.105561,
		0.132788,
		0.161842,
		0.192394,
		0.224139,
		0.256796,
		0.290107,
		0.323837,
		0.357769,
		0.391707,
		0.425471,
		0.458899,
		0.491846,
		0.52418,
		0.555785,
		0.586559,
		0.61641,
		0.645261,
		0.673044,
		0.699703,
		0.725192,
		0.749475,
		0.772523,
		0.794318,
		0.814849,
		0.834112,
		0.85211,
		0.868855,
		0.884362,
		0.898654,
		0.91176,
		0.923712,
		0.934549,
		0.944313,
		0.95305,
		0.96081,
		0.967646,
		0.973615,
		0.978775,
		0.983187,
		0.986911,
		0.990011,
		0.992549,
		0.99459,
		0.996194,
		0.997422,
		0.998334,
		0.998984,
		0.999426,
		0.999707,
		0.99987,
		0.999954,
		0.999989,
		0.999999,
		1,
		1,
		0.999997,
		0.999982,
		0.99994,
		0.99985,
		0.999688,
		0.999429,
		0.999046,
		0.998511,
		0.997796,
		0.002638,
		0.010171,
		0.022059,
		0.037798,
		0.05692,
		0.078987,
		0.103595,
		0.130369,
		0.15896,
		0.189047,
		0.220331,
		0.25254,
		0.285421,
		0.318742,
		0.352291,
		0.385876,
		0.419319,
		0.452462,
		0.48516,
		0.517284,
		0.548718,
		0.579359,
		0.609118,
		0.637915,
		0.665684,
		0.692366,
		0.717915,
		0.742293,
		0.76547,
		0.787425,
		0.808146,
		0.827626,
		0.845868,
		0.862877,
		0.87867,
		0.893264,
		0.906686,
		0.918966,
		0.930138,
		0.940242,
		0.949321,
		0.957423,
		0.964597,
		0.970896,
		0.976376,
		0.981094,
		0.985111,
		0.988485,
		0.991278,
		0.993551,
		0.995364,
		0.996778,
		0.99785,
		0.998635,
		0.999187,
		0.999554,
		0.999782,
		0.999909,
		0.999971,
		0.999994,
		1,
		1,
		1,
		0.999995,
		0.999973,
		0.99992,
		0.999813,
		0.999629,
		0.999343,
		0.002583,
		0.009964,
		0.021618,
		0.037057,
		0.055825,
		0.077498,
		0.101684,
		0.128015,
		0.156153,
		0.185784,
		0.216617,
		0.248386,
		0.280842,
		0.31376,
		0.34693,
		0.380165,
		0.413289,
		0.446146,
		0.478594,
		0.510503,
		0.541761,
		0.572264,
		0.601923,
		0.630658,
		0.658402,
		0.685097,
		0.710695,
		0.735156,
		0.758449,
		0.780552,
		0.80145,
		0.821134,
		0.839605,
		0.856866,
		0.872931,
		0.887815,
		0.901541,
		0.914137,
		0.925634,
		0.93607,
		0.945484,
		0.953921,
		0.961427,
		0.968053,
		0.973851,
		0.978877,
		0.983187,
		0.986838,
		0.989891,
		0.992403,
		0.994434,
		0.996042,
		0.997284,
		0.998216,
		0.998889,
		0.999355,
		0.999658,
		0.99984,
		0.999938,
		0.999982,
		0.999997,
		1,
		1,
		0.999999,
		0.99999,
		0.999962,
		0.999896,
		0.00253,
		0.009762,
		0.021189,
		0.036336,
		0.054761,
		0.076051,
		0.099824,
		0.125723,
		0.153419,
		0.182603,
		0.212995,
		0.24433,
		0.276369,
		0.308888,
		0.341684,
		0.37457,
		0.407377,
		0.439948,
		0.472143,
		0.503836,
		0.534913,
		0.565272,
		0.594824,
		0.62349,
		0.651201,
		0.677898,
		0.703534,
		0.728066,
		0.751464,
		0.773702,
		0.794763,
		0.814639,
		0.833326,
		0.850827,
		0.867151,
		0.882312,
		0.896331,
		0.909232,
		0.921045,
		0.931803,
		0.941544,
		0.950308,
		0.958141,
		0.96509,
		0.971204,
		0.976536,
		0.98114,
		0.985072,
		0.988388,
		0.991144,
		0.9934,
		0.995211,
		0.996633,
		0.997721,
		0.998528,
		0.999103,
		0.999493,
		0.999741,
		0.999886,
		0.999959,
		0.99999,
		0.999999,
		1,
		1,
		0.999998,
		0.002478,
		0.009567,
		0.020774,
		0.035637,
		0.053727,
		0.074644,
		0.098015,
		0.123492,
		0.150754,
		0.179502,
		0.20946,
		0.24037,
		0.271997,
		0.304124,
		0.336549,
		0.369091,
		0.40158,
		0.433865,
		0.465807,
		0.497281,
		0.528173,
		0.558383,
		0.587822,
		0.61641,
		0.64408,
		0.670771,
		0.696434,
		0.721027,
		0.744517,
		0.766878,
		0.788092,
		0.808146,
		0.827036,
		0.844764,
		0.861335,
		0.876762,
		0.891062,
		0.904257,
		0.916375,
		0.927446,
		0.937506,
		0.946591,
		0.954745,
		0.962012,
		0.968439,
		0.974075,
		0.978974,
		0.983187,
		0.986769,
		0.989775,
		0.992261,
		0.994282,
		0.995893,
		0.997147,
		0.998097,
		0.998793,
		0.999281,
		0.999606,
		0.999807,
		0.99992,
		0.999974,
		0.999995,
		1,
		0.002428,
		0.009378,
		0.02037,
		0.034957,
		0.052722,
		0.073275,
		0.096253,
		0.121319,
		0.148158,
		0.176478,
		0.20601,
		0.236502,
		0.267725,
		0.299464,
		0.331523,
		0.363722,
		0.395897,
		0.427896,
		0.459584,
		0.490835,
		0.521539,
		0.551596,
		0.580916,
		0.60942,
		0.63704,
		0.663715,
		0.689396,
		0.71404,
		0.737611,
		0.760084,
		0.781437,
		0.801659,
		0.820741,
		0.838682,
		0.855488,
		0.871169,
		0.885739,
		0.899218,
		0.911631,
		0.923006,
		0.933375,
		0.942774,
		0.951243,
		0.958822,
		0.965558,
		0.971498,
		0.976689,
		0.981184,
		0.985035,
		0.988294,
		0.991015,
		0.993253,
		0.99506,
		0.99649,
		0.997593,
		0.99842,
		0.999017,
		0.999429,
		0.999698,
		0.999859,
		0.999945,
		0.00238,
		0.009194,
		0.019978,
		0.034297,
		0.051745,
		0.071944,
		0.094539,
		0.119202,
		0.145627,
		0.173528,
		0.202643,
		0.232725,
		0.263549,
		0.294906,
		0.326602,
		0.358463,
		0.390324,
		0.422038,
		0.453471,
		0.484498,
		0.515011,
		0.54491,
		0.574105,
		0.602518,
		0.630081,
		0.656733,
		0.682423,
		0.707107,
		0.73075,
		0.753323,
		0.774805,
		0.795181,
		0.814442,
		0.832587,
		0.849616,
		0.865538,
		0.880367,
		0.894119,
		0.906817,
		0.918486,
		0.929157,
		0.938862,
		0.947639,
		0.955526,
		0.962567,
		0.968806,
		0.974289,
		0.979066,
		0.983187,
		0.986701,
		0.989663,
		0.992122,
		0.994133,
		0.995745,
		0.997011,
		0.997978,
		0.998695,
		0.999205,
		0.999551,
		0.002333,
		0.009016,
		0.019597,
		0.033655,
		0.050795,
		0.070648,
		0.09287,
		0.11714,
		0.14316,
		0.170651,
		0.199356,
		0.229035,
		0.259467,
		0.290446,
		0.321785,
		0.353309,
		0.384859,
		0.416289,
		0.447466,
		0.478268,
		0.508587,
		0.538323,
		0.567389,
		0.595706,
		0.623205,
		0.649825,
		0.675515,
		0.70023,
		0.723934,
		0.746597,
		0.768196,
		0.788716,
		0.808146,
		0.826481,
		0.843722,
		0.859876,
		0.874952,
		0.888967,
		0.901939,
		0.913893,
		0.924857,
		0.93486,
		0.943938,
		0.952128,
		0.959469,
		0.966004,
		0.971777,
		0.976835,
		0.981226,
		0.984999,
		0.988203,
		0.99089,
		0.993109,
		0.994912,
		0.996348,
		0.997465,
		0.99831,
		0.002287,
		0.008842,
		0.019227,
		0.033031,
		0.049871,
		0.069387,
		0.091244,
		0.11513,
		0.140754,
		0.167843,
		0.196146,
		0.225429,
		0.255475,
		0.286083,
		0.317068,
		0.34826,
		0.379501,
		0.410647,
		0.441568,
		0.472143,
		0.502265,
		0.531836,
		0.560768,
		0.588982,
		0.61641,
		0.642991,
		0.668673,
		0.69341,
		0.717165,
		0.739908,
		0.761615,
		0.782268,
		0.801855,
		0.82037,
		0.837812,
		0.854185,
		0.869498,
		0.883765,
		0.897002,
		0.909232,
		0.92048,
		0.930774,
		0.940146,
		0.948631,
		0.956268,
		0.963095,
		0.969156,
		0.974494,
		0.979155,
		0.983187,
		0.986637,
		0.989554,
		0.991988,
		0.993987,
		0.995601,
		0.002243,
		0.008674,
		0.018867,
		0.032424,
		0.048971,
		0.068159,
		0.08966,
		0.113171,
		0.138407,
		0.165103,
		0.193012,
		0.221906,
		0.251572,
		0.281814,
		0.31245,
		0.343311,
		0.374245,
		0.405109,
		0.435774,
		0.466121,
		0.496045,
		0.525446,
		0.55424,
		0.582346,
		0.609698,
		0.636233,
		0.661898,
		0.686649,
		0.710447,
		0.73326,
		0.755064,
		0.77584,
		0.795573,
		0.814257,
		0.831889,
		0.848471,
		0.864011,
		0.878519,
		0.892012,
		0.904508,
		0.916031,
		0.926607,
		0.936266,
		0.945041,
		0.952968,
		0.960083,
		0.966428,
		0.972044,
		0.976975,
		0.981267,
		0.984964,
		0.988116,
		0.990768,
		0.0022,
		0.00851,
		0.018518,
		0.031834,
		0.048096,
		0.066963,
		0.088117,
		0.111261,
		0.136118,
		0.162428,
		0.189951,
		0.218463,
		0.247755,
		0.277636,
		0.307926,
		0.338462,
		0.369091,
		0.399673,
		0.430082,
		0.460201,
		0.489923,
		0.519153,
		0.547804,
		0.575799,
		0.603067,
		0.629549,
		0.655192,
		0.679948,
		0.70378,
		0.726654,
		0.748546,
		0.769433,
		0.789303,
		0.808146,
		0.825958,
		0.842738,
		0.858494,
		0.873234,
		0.886971,
		0.899725,
		0.911514,
		0.922365,
		0.932304,
		0.941362,
		0.949573,
		0.956972,
		0.963598,
		0.969489,
		0.974689,
		0.97924,
		0.983187,
		0.002158,
		0.008351,
		0.018178,
		0.031259,
		0.047243,
		0.065798,
		0.086613,
		0.109399,
		0.133884,
		0.159817,
		0.186961,
		0.215097,
		0.244022,
		0.273547,
		0.303496,
		0.333709,
		0.364035,
		0.394338,
		0.424491,
		0.454381,
		0.483901,
		0.512956,
		0.541461,
		0.569338,
		0.596519,
		0.622942,
		0.648554,
		0.673308,
		0.697165,
		0.720092,
		0.742062,
		0.763052,
		0.783049,
		0.80204,
		0.820021,
		0.83699,
		0.852952,
		0.867913,
		0.881887,
		0.894888,
		0.906935,
		0.918052,
		0.928264,
		0.937598,
		0.946088,
		0.953766,
		0.960668,
		0.966832,
		0.972299,
		0.002117,
		0.008197,
		0.017847,
		0.0307,
		0.046414,
		0.064663,
		0.085147,
		0.107582,
		0.131705,
		0.157267,
		0.184039,
		0.211807,
		0.24037,
		0.269544,
		0.299157,
		0.32905,
		0.359076,
		0.3891,
		0.418999,
		0.448658,
		0.477974,
		0.506853,
		0.535208,
		0.562964,
		0.590052,
		0.61641,
		0.641985,
		0.66673,
		0.690605,
		0.713576,
		0.735615,
		0.756699,
		0.776812,
		0.795942,
		0.814083,
		0.83123,
		0.847388,
		0.862562,
		0.876762,
		0.890001,
		0.902298,
		0.913673,
		0.924149,
		0.933754,
		0.942516,
		0.950468,
		0.957642,
		0.002078,
		0.008046,
		0.017525,
		0.030156,
		0.045605,
		0.063557,
		0.083718,
		0.10581,
		0.129577,
		0.154777,
		0.181184,
		0.208589,
		0.236797,
		0.265625,
		0.294906,
		0.324482,
		0.354211,
		0.383959,
		0.413603,
		0.443032,
		0.472143,
		0.500843,
		0.529046,
		0.556677,
		0.583667,
		0.609954,
		0.635486,
		0.660215,
		0.6841,
		0.707107,
		0.729206,
		0.750376,
		0.770596,
		0.789856,
		0.808146,
		0.825463,
		0.841807,
		0.857183,
		0.8716,
		0.88507,
		0.897607,
		0.909232,
		0.919966,
		0.929834,
		0.938862,
		0.00204,
		0.0079,
		0.017212,
		0.029626,
		0.044818,
		0.06248,
		0.082324,
		0.104082,
		0.1275,
		0.152345,
		0.178394,
		0.205443,
		0.2333,
		0.261788,
		0.290741,
		0.320005,
		0.349439,
		0.378911,
		0.408303,
		0.437501,
		0.466406,
		0.494924,
		0.522972,
		0.550474,
		0.577362,
		0.603574,
		0.629057,
		0.653763,
		0.677651,
		0.700686,
		0.722839,
		0.744084,
		0.764404,
		0.783784,
		0.802214,
		0.819691,
		0.836212,
		0.851782,
		0.866407,
		0.880097,
		0.892867,
		0.904734,
		0.915718,
		0.002002,
		0.007758,
		0.016907,
		0.02911,
		0.044051,
		0.061429,
		0.080964,
		0.102394,
		0.125472,
		0.149968,
		0.175666,
		0.202366,
		0.229879,
		0.258031,
		0.28666,
		0.315615,
		0.344756,
		0.373956,
		0.403095,
		0.432063,
		0.460761,
		0.489096,
		0.516987,
		0.544357,
		0.571138,
		0.59727,
		0.622699,
		0.647376,
		0.67126,
		0.694316,
		0.716513,
		0.737827,
		0.758237,
		0.777728,
		0.796291,
		0.813918,
		0.830607,
		0.846361,
		0.861185,
		0.875088,
		0.888082,
		0.001966,
		0.007619,
		0.01661,
		0.028608,
		0.043303,
		0.060404,
		0.079638,
		0.100747,
		0.123492,
		0.147646,
		0.173,
		0.199356,
		0.22653,
		0.254351,
		0.282661,
		0.31131,
		0.340162,
		0.369091,
		0.397978,
		0.426716,
		0.455206,
		0.483358,
		0.511089,
		0.538323,
		0.564995,
		0.591042,
		0.61641,
		0.641052,
		0.664927,
		0.687996,
		0.710231,
		0.731605,
		0.752097,
		0.771692,
		0.790377,
		0.808146,
		0.824995,
		0.840925,
		0.855939,
		0.001931,
		0.007484,
		0.016321,
		0.028118,
		0.042574,
		0.059405,
		0.078344,
		0.09914,
		0.121558,
		0.145377,
		0.170393,
		0.196411,
		0.223252,
		0.250747,
		0.278741,
		0.307089,
		0.335654,
		0.364313,
		0.39295,
		0.421459,
		0.449741,
		0.477707,
		0.505276,
		0.532373,
		0.55893,
		0.584888,
		0.610192,
		0.634794,
		0.658652,
		0.681729,
		0.703994,
		0.72542,
		0.745987,
		0.765677,
		0.784477,
		0.802379,
		0.819379,
		0.001896,
		0.007353,
		0.016039,
		0.027641,
		0.041863,
		0.05843,
		0.077081,
		0.09757,
		0.119668,
		0.14316,
		0.167843,
		0.193529,
		0.220042,
		0.247217,
		0.2749,
		0.302949,
		0.331231,
		0.359622,
		0.38801,
		0.416289,
		0.444363,
		0.472143,
		0.499548,
		0.526505,
		0.552945,
		0.57881,
		0.604044,
		0.628601,
		0.652437,
		0.675515,
		0.697803,
		0.719275,
		0.739908,
		0.759685,
		0.778592,
		0.001863,
		0.007225,
		0.015765,
		0.027175,
		0.04117,
		0.057479,
		0.075848,
		0.096037,
		0.117821,
		0.140991,
		0.165349,
		0.19071,
		0.2169,
		0.243759,
		0.271135,
		0.298888,
		0.326889,
		0.355015,
		0.383156,
		0.411206,
		0.439072,
		0.466665,
		0.493904,
		0.520718,
		0.547038,
		0.572806,
		0.597967,
		0.622473,
		0.646281,
		0.669354,
		0.691659,
		0.71317,
		0.733863,
		0.00183,
		0.007101,
		0.015497,
		0.026722,
		0.040494,
		0.056551,
		0.074644,
		0.094539,
		0.116017,
		0.138872,
		0.16291,
		0.18795,
		0.213823,
		0.24037,
		0.267443,
		0.294906,
		0.322628,
		0.350491,
		0.378386,
		0.406208,
		0.433865,
		0.46127,
		0.488343,
		0.515011,
		0.541209,
		0.566876,
		0.59196,
		0.61641,
		0.640185,
		0.663248,
		0.685564,
		0.001798,
		0.006979,
		0.015236,
		0.026279,
		0.039835,
		0.055645,
		0.073468,
		0.093076,
		0.114253,
		0.136799,
		0.160523,
		0.185248,
		0.210809,
		0.237049,
		0.263824,
		0.290998,
		0.318445,
		0.346048,
		0.373698,
		0.401293,
		0.428742,
		0.455958,
		0.482863,
		0.509384,
		0.535457,
		0.561021,
		0.586022,
		0.610413,
		0.63415,
		0.001767,
		0.006861,
		0.014982,
		0.025848,
		0.039191,
		0.054761,
		0.07232,
		0.091646,
		0.112529,
		0.134771,
		0.158187,
		0.182603,
		0.207857,
		0.233795,
		0.260276,
		0.287165,
		0.31434,
		0.341684,
		0.369091,
		0.39646,
		0.423701,
		0.450728,
		0.477463,
		0.503836,
		0.529781,
		0.555238,
		0.580154,
		0.001737,
		0.006746,
		0.014734,
		0.025427,
		0.038563,
		0.053897,
		0.071199,
		0.090249,
		0.110843,
		0.132788,
		0.155901,
		0.180014,
		0.204965,
		0.230606,
		0.256796,
		0.283404,
		0.310309,
		0.337397,
		0.364563,
		0.391707,
		0.41874,
		0.445578,
		0.472143,
		0.498366,
		0.52418,
		0.001708,
		0.006633,
		0.014492,
		0.025016,
		0.03795,
		0.053054,
		0.070103,
		0.088884,
		0.109195,
		0.130847,
		0.153664,
		0.177478,
		0.202132,
		0.227479,
		0.253383,
		0.279714,
		0.306352,
		0.333186,
		0.360112,
		0.387032,
		0.413858,
		0.440506,
		0.466901,
		0.001679,
		0.006524,
		0.014256,
		0.024615,
		0.037351,
		0.05223,
		0.069032,
		0.087549,
		0.107582,
		0.128949,
		0.151474,
		0.174994,
		0.199356,
		0.224414,
		0.250035,
		0.276092,
		0.302467,
		0.32905,
		0.355737,
		0.382435,
		0.409054,
		0.001651,
		0.006417,
		0.014026,
		0.024223,
		0.036766,
		0.051425,
		0.067986,
		0.086243,
		0.106005,
		0.127091,
		0.14933,
		0.172561,
		0.196635,
		0.221409,
		0.246752,
		0.272538,
		0.298652,
		0.324985,
		0.351437,
		0.001624,
		0.006313,
		0.013802,
		0.023841,
		0.036195,
		0.050639,
		0.066963,
		0.084966,
		0.104462,
		0.125272,
		0.14723,
		0.170178,
		0.193969,
		0.218463,
		0.24353,
		0.269049,
		0.294906,
		0.001597,
		0.006211,
		0.013582,
		0.023468,
		0.035637,
		0.049871,
		0.065963,
		0.083718,
		0.102952,
		0.123492,
		0.145174,
		0.167843,
		0.191355,
		0.215573,
		0.24037,
		0.001571,
		0.006111,
		0.013368,
		0.023103,
		0.035092,
		0.049119,
		0.064984,
		0.082496,
		0.101474,
		0.121749,
		0.14316,
		0.165555,
		0.188793,
		0.001546,
		0.006014,
		0.013159,
		0.022747,
		0.034559,
		0.048385,
		0.064028,
		0.081301,
		0.100028,
		0.120042,
		0.141187,
		0.001521,
		0.00592,
		0.012955,
		0.022399,
		0.034038,
		0.047667,
		0.063092,
		0.080132,
		0.098612,
		0.001497,
		0.005827,
		0.012755,
		0.022059,
		0.033529,
		0.046964,
		0.062177,
		0.001474,
		0.005737,
		0.01256,
		0.021727,
		0.033031,
		0.001451,
		0.005649,
		0.01237,
		0.001428,
		0.005562,
		0.002221,
		0.000363,
		0.021402,
		0.014447,
		0.008741,
		0.004379,
		0.001462,
		9.3e-005,
		0.046277,
		0.036235,
		0.027226,
		0.019337,
		0.012657,
		0.007282,
		0.00331,
		0.000846,
		0.078987,
		0.066356,
		0.05456,
		0.043674,
		0.033782,
		0.024968,
		0.017323,
		0.01094,
		0.00592,
		0.002366,
		0.000387,
		0.118371,
		0.103621,
		0.089522,
		0.076143,
		0.063557,
		0.051842,
		0.041077,
		0.031349,
		0.022747,
		0.015366,
		0.009304,
		0.004665,
		0.001559,
		0.0001,
		0.163312,
		0.146884,
		0.130939,
		0.115538,
		0.100747,
		0.086636,
		0.073275,
		0.060743,
		0.049119,
		0.038489,
		0.028941,
		0.02057,
		0.013475,
		0.007758,
		0.003529,
		0.000903,
		0.212739,
		0.195044,
		0.177679,
		0.160699,
		0.144161,
		0.128128,
		0.112664,
		0.097836,
		0.083718,
		0.070384,
		0.057915,
		0.046396,
		0.035914,
		0.026564,
		0.018444,
		0.011657,
		0.006313,
		0.002525,
		0.000414,
		0.265625,
		0.247045,
		0.228658,
		0.210511,
		0.192655,
		0.175145,
		0.158037,
		0.141391,
		0.125272,
		0.109748,
		0.094888,
		0.08077,
		0.067471,
		0.055077,
		0.043674,
		0.033357,
		0.024223,
		0.016376,
		0.009923,
		0.004979,
		0.001665,
		0.000107,
		0.320992,
		0.301882,
		0.282841,
		0.26391,
		0.245133,
		0.226559,
		0.208236,
		0.190218,
		0.172561,
		0.155325,
		0.138573,
		0.122371,
		0.106789,
		0.091904,
		0.077793,
		0.064539,
		0.05223,
		0.040959,
		0.030823,
		0.021925,
		0.014374,
		0.008282,
		0.003771,
		0.000966,
		0.377913,
		0.358598,
		0.339242,
		0.31988,
		0.300551,
		0.281295,
		0.262156,
		0.243179,
		0.224414,
		0.205913,
		0.187731,
		0.169927,
		0.152563,
		0.135705,
		0.119423,
		0.103789,
		0.088884,
		0.074788,
		0.061589,
		0.049379,
		0.038254,
		0.028318,
		0.019678,
		0.012447,
		0.006746,
		0.0027,
		0.000443,
		0.435513,
		0.416289,
		0.396928,
		0.377459,
		0.357915,
		0.33833,
		0.318742,
		0.299189,
		0.279714,
		0.260362,
		0.241182,
		0.222223,
		0.203541,
		0.185194,
		0.167242,
		0.14975,
		0.132788,
		0.116428,
		0.100747,
		0.085828,
		0.071756,
		0.058623,
		0.046525,
		0.035564,
		0.025848,
		0.017488,
		0.010606,
		0.005327,
		0.001783,
		0.000114,
		0.492972,
		0.474108,
		0.455022,
		0.435741,
		0.416289,
		0.396697,
		0.376995,
		0.357216,
		0.337397,
		0.317576,
		0.297795,
		0.278096,
		0.258527,
		0.239139,
		0.219983,
		0.201118,
		0.182603,
		0.164503,
		0.146884,
		0.12982,
		0.113386,
		0.097664,
		0.082738,
		0.0687,
		0.055645,
		0.043674,
		0.032895,
		0.023419,
		0.015366,
		0.008861,
		0.004038,
		0.001035,
		0.549528,
		0.531263,
		0.512705,
		0.493875,
		0.474793,
		0.455485,
		0.435974,
		0.416289,
		0.39646,
		0.376519,
		0.3565,
		0.336442,
		0.316383,
		0.296367,
		0.27644,
		0.25665,
		0.237049,
		0.217694,
		0.198643,
		0.179959,
		0.16171,
		0.143965,
		0.126801,
		0.110298,
		0.094539,
		0.079615,
		0.065621,
		0.052658,
		0.04083,
		0.030251,
		0.02104,
		0.01332,
		0.007225,
		0.002895,
		0.000475,
		0.604481,
		0.587025,
		0.569217,
		0.551073,
		0.532609,
		0.513845,
		0.494799,
		0.475496,
		0.455958,
		0.436213,
		0.416289,
		0.396218,
		0.376031,
		0.355767,
		0.335463,
		0.315161,
		0.294906,
		0.274744,
		0.254728,
		0.234912,
		0.215353,
		0.196114,
		0.17726,
		0.158861,
		0.140991,
		0.12373,
		0.107161,
		0.091374,
		0.076461,
		0.062522,
		0.049664,
		0.037997,
		0.027641,
		0.018718,
		0.011362,
		0.005711,
		0.001913,
		0.000123,
		0.657196,
		0.64073,
		0.623865,
		0.606611,
		0.588982,
		0.570991,
		0.552654,
		0.533987,
		0.515011,
		0.495746,
		0.476215,
		0.456443,
		0.436458,
		0.416289,
		0.395969,
		0.375532,
		0.355015,
		0.33446,
		0.313909,
		0.293408,
		0.273008,
		0.252761,
		0.232725,
		0.212959,
		0.193529,
		0.174504,
		0.155955,
		0.137962,
		0.120607,
		0.103978,
		0.088168,
		0.073275,
		0.059405,
		0.046668,
		0.035182,
		0.02507,
		0.016464,
		0.009504,
		0.004334,
		0.001112,
		0.707107,
		0.691782,
		0.676022,
		0.659834,
		0.643226,
		0.626207,
		0.608789,
		0.590984,
		0.572806,
		0.554272,
		0.535398,
		0.516206,
		0.496716,
		0.476952,
		0.45694,
		0.436709,
		0.416289,
		0.395714,
		0.37502,
		0.354245,
		0.333432,
		0.312626,
		0.291874,
		0.27123,
		0.250747,
		0.230487,
		0.210511,
		0.190888,
		0.171689,
		0.152992,
		0.134877,
		0.117431,
		0.100747,
		0.084923,
		0.070061,
		0.056273,
		0.043674,
		0.032389,
		0.022547,
		0.014288,
		0.007758,
		0.003111,
		0.000511,
		0.753719,
		0.739658,
		0.725136,
		0.710157,
		0.694725,
		0.678846,
		0.662527,
		0.645775,
		0.628601,
		0.611015,
		0.593031,
		0.574663,
		0.555928,
		0.536843,
		0.51743,
		0.497709,
		0.477707,
		0.45745,
		0.436966,
		0.416289,
		0.395453,
		0.374495,
		0.353455,
		0.332378,
		0.31131,
		0.290302,
		0.269407,
		0.248685,
		0.228195,
		0.208006,
		0.188187,
		0.168815,
		0.149968,
		0.131734,
		0.114202,
		0.09747,
		0.08164,
		0.066821,
		0.05313,
		0.040688,
		0.029626,
		0.020082,
		0.012202,
		0.00614,
		0.002059,
		0.000132,
		0.796619,
		0.783913,
		0.770731,
		0.757075,
		0.742944,
		0.72834,
		0.713267,
		0.697728,
		0.681729,
		0.665277,
		0.64838,
		0.631048,
		0.613292,
		0.595126,
		0.576564,
		0.557624,
		0.538323,
		0.518684,
		0.498728,
		0.478481,
		0.457972,
		0.43723,
		0.416289,
		0.395185,
		0.373956,
		0.352645,
		0.331297,
		0.309961,
		0.28869,
		0.26754,
		0.246571,
		0.225849,
		0.205443,
		0.185427,
		0.165879,
		0.146884,
		0.128533,
		0.110919,
		0.094146,
		0.07832,
		0.063557,
		0.049979,
		0.037715,
		0.026902,
		0.017685,
		0.010218,
		0.004665,
		0.001198,
		0.835475,
		0.824185,
		0.812415,
		0.800163,
		0.787425,
		0.774201,
		0.760489,
		0.746289,
		0.731605,
		0.716437,
		0.700791,
		0.684672,
		0.668086,
		0.651042,
		0.63355,
		0.615622,
		0.59727,
		0.578511,
		0.559361,
		0.53984,
		0.519969,
		0.499772,
		0.479275,
		0.458508,
		0.437501,
		0.416289,
		0.39491,
		0.373403,
		0.351814,
		0.330188,
		0.308577,
		0.287037,
		0.265625,
		0.244406,
		0.223446,
		0.20282,
		0.182603,
		0.16288,
		0.143738,
		0.125272,
		0.107582,
		0.090775,
		0.074965,
		0.060273,
		0.046826,
		0.034761,
		0.024223,
		0.015366,
		0.008351,
		0.003352,
		0.000551,
		0.870045,
		0.860201,
		0.849883,
		0.839085,
		0.827802,
		0.816028,
		0.803759,
		0.790993,
		0.777728,
		0.763962,
		0.749696,
		0.734932,
		0.71967,
		0.703917,
		0.687676,
		0.670956,
		0.653763,
		0.636109,
		0.618005,
		0.599465,
		0.580504,
		0.561141,
		0.541394,
		0.521286,
		0.500843,
		0.48009,
		0.459058,
		0.437779,
		0.416289,
		0.394628,
		0.372836,
		0.35096,
		0.32905,
		0.307157,
		0.285341,
		0.263662,
		0.242186,
		0.220985,
		0.200135,
		0.179716,
		0.159817,
		0.140529,
		0.121952,
		0.104192,
		0.08736,
		0.071577,
		0.05697,
		0.043674,
		0.031834,
		0.021601,
		0.013138,
		0.006618,
		0.002221,
		0.000143,
		0.900183,
		0.891784,
		0.882926,
		0.873601,
		0.863799,
		0.853514,
		0.842738,
		0.831466,
		0.819691,
		0.80741,
		0.794618,
		0.781314,
		0.767497,
		0.753166,
		0.738322,
		0.722968,
		0.707107,
		0.690745,
		0.673888,
		0.656545,
		0.638727,
		0.620445,
		0.601712,
		0.582546,
		0.562964,
		0.542987,
		0.522638,
		0.501941,
		0.480925,
		0.459622,
		0.438064,
		0.416289,
		0.394338,
		0.372254,
		0.350084,
		0.327881,
		0.3057,
		0.283601,
		0.261647,
		0.23991,
		0.218463,
		0.197385,
		0.176763,
		0.156687,
		0.137255,
		0.118572,
		0.100747,
		0.083901,
		0.068159,
		0.053654,
		0.040531,
		0.028941,
		0.019046,
		0.011016,
		0.005035,
		0.001294,
		0.925841,
		0.918855,
		0.911434,
		0.903567,
		0.895243,
		0.886452,
		0.877185,
		0.867431,
		0.857183,
		0.846434,
		0.835177,
		0.823405,
		0.811114,
		0.7983,
		0.784961,
		0.771094,
		0.756699,
		0.741777,
		0.726331,
		0.710363,
		0.693878,
		0.676885,
		0.65939,
		0.641405,
		0.622942,
		0.604014,
		0.584639,
		0.564834,
		0.544621,
		0.524024,
		0.503068,
		0.481783,
		0.460201,
		0.438357,
		0.416289,
		0.39404,
		0.371655,
		0.349184,
		0.326681,
		0.304203,
		0.281814,
		0.259581,
		0.237576,
		0.215878,
		0.19457,
		0.173742,
		0.15349,
		0.133916,
		0.11513,
		0.09725,
		0.080399,
		0.064712,
		0.05033,
		0.037403,
		0.026093,
		0.01657,
		0.009016,
		0.003623,
		0.000596,
		0.947082,
		0.941445,
		0.935405,
		0.92895,
		0.922066,
		0.914741,
		0.906963,
		0.89872,
		0.890001,
		0.880796,
		0.871096,
		0.86089,
		0.850172,
		0.838934,
		0.827169,
		0.814873,
		0.80204,
		0.788668,
		0.774754,
		0.760297,
		0.745299,
		0.729761,
		0.713686,
		0.697079,
		0.679948,
		0.6623,
		0.644146,
		0.625499,
		0.606372,
		0.586783,
		0.566751,
		0.546297,
		0.525446,
		0.504225,
		0.482664,
		0.460796,
		0.438658,
		0.416289,
		0.393734,
		0.371041,
		0.34826,
		0.325448,
		0.302666,
		0.27998,
		0.25746,
		0.235182,
		0.213228,
		0.191686,
		0.170651,
		0.150223,
		0.130509,
		0.111627,
		0.093699,
		0.076857,
		0.061241,
		0.047001,
		0.034297,
		0.023299,
		0.014187,
		0.007154,
		0.002404,
		0.000155,
		0.964076,
		0.959694,
		0.954949,
		0.949827,
		0.944313,
		0.938392,
		0.932051,
		0.925275,
		0.918052,
		0.910368,
		0.902212,
		0.893571,
		0.884434,
		0.874792,
		0.864634,
		0.853952,
		0.842738,
		0.830985,
		0.818687,
		0.805839,
		0.792437,
		0.778478,
		0.763962,
		0.748889,
		0.73326,
		0.717079,
		0.70035,
		0.68308,
		0.665277,
		0.646952,
		0.628118,
		0.608789,
		0.588982,
		0.568718,
		0.548017,
		0.526907,
		0.505413,
		0.483569,
		0.461407,
		0.438967,
		0.416289,
		0.39342,
		0.370409,
		0.34731,
		0.324181,
		0.301087,
		0.278096,
		0.255282,
		0.232725,
		0.210511,
		0.188732,
		0.167488,
		0.146884,
		0.127035,
		0.108062,
		0.090096,
		0.073275,
		0.057749,
		0.043674,
		0.031222,
		0.02057,
		0.011912,
		0.00545,
		0.001403,
		0.977109,
		0.973858,
		0.970291,
		0.966392,
		0.962145,
		0.957534,
		0.952544,
		0.947158,
		0.941362,
		0.935141,
		0.928479,
		0.921364,
		0.913781,
		0.905717,
		0.897159,
		0.888097,
		0.878519,
		0.868415,
		0.857774,
		0.846589,
		0.834852,
		0.822556,
		0.809696,
		0.796268,
		0.782268,
		0.767695,
		0.752549,
		0.736831,
		0.720543,
		0.703692,
		0.686282,
		0.668323,
		0.649825,
		0.630801,
		0.611266,
		0.591237,
		0.570735,
		0.549783,
		0.528406,
		0.506634,
		0.484498,
		0.462036,
		0.439285,
		0.416289,
		0.393097,
		0.369759,
		0.346332,
		0.322879,
		0.299464,
		0.27616,
		0.253045,
		0.230203,
		0.207724,
		0.185705,
		0.164251,
		0.143473,
		0.123492,
		0.104436,
		0.086442,
		0.069658,
		0.05424,
		0.040357,
		0.028187,
		0.017921,
		0.009762,
		0.003928,
		0.000647,
		0.986574,
		0.984305,
		0.981771,
		0.978957,
		0.975845,
		0.972419,
		0.968661,
		0.964555,
		0.960083,
		0.95523,
		0.949978,
		0.944313,
		0.938217,
		0.931675,
		0.924674,
		0.917197,
		0.909232,
		0.900765,
		0.891784,
		0.882275,
		0.87223,
		0.861636,
		0.850486,
		0.83877,
		0.826481,
		0.813614,
		0.800163,
		0.786125,
		0.771497,
		0.75628,
		0.740474,
		0.724081,
		0.707107,
		0.689557,
		0.67144,
		0.652767,
		0.63355,
		0.613805,
		0.59355,
		0.572806,
		0.551596,
		0.529947,
		0.507888,
		0.485454,
		0.462682,
		0.439612,
		0.416289,
		0.392764,
		0.369091,
		0.345327,
		0.321539,
		0.297795,
		0.27417,
		0.250747,
		0.227614,
		0.204865,
		0.182603,
		0.160938,
		0.139988,
		0.119879,
		0.100747,
		0.082738,
		0.066007,
		0.050721,
		0.037057,
		0.025204,
		0.015366,
		0.007758,
		0.00261,
		0.000168,
		0.992969,
		0.99151,
		0.989843,
		0.987949,
		0.985814,
		0.983418,
		0.980745,
		0.977776,
		0.974494,
		0.970881,
		0.96692,
		0.962592,
		0.957881,
		0.952769,
		0.947239,
		0.941275,
		0.93486,
		0.927979,
		0.920616,
		0.912756,
		0.904386,
		0.895491,
		0.886059,
		0.876079,
		0.865538,
		0.854428,
		0.842738,
		0.830462,
		0.817592,
		0.804122,
		0.790049,
		0.77537,
		0.760084,
		0.744191,
		0.727694,
		0.710598,
		0.692907,
		0.674631,
		0.655781,
		0.636368,
		0.61641,
		0.595924,
		0.574932,
		0.553458,
		0.531529,
		0.509177,
		0.486437,
		0.463346,
		0.439948,
		0.416289,
		0.392422,
		0.368403,
		0.344293,
		0.32016,
		0.296078,
		0.272124,
		0.248386,
		0.224955,
		0.201932,
		0.179424,
		0.157547,
		0.136427,
		0.116196,
		0.096998,
		0.078987,
		0.062328,
		0.047197,
		0.033782,
		0.022285,
		0.012921,
		0.00592,
		0.001526,
		0.996876,
		0.996042,
		0.995056,
		0.993901,
		0.992561,
		0.991018,
		0.989256,
		0.987255,
		0.984999,
		0.982468,
		0.979645,
		0.976511,
		0.973047,
		0.969235,
		0.965057,
		0.960493,
		0.955526,
		0.950139,
		0.944313,
		0.938031,
		0.931276,
		0.924033,
		0.916286,
		0.908019,
		0.899218,
		0.889869,
		0.87996,
		0.869479,
		0.858415,
		0.846758,
		0.834498,
		0.82163,
		0.808146,
		0.794042,
		0.779315,
		0.763962,
		0.747985,
		0.731385,
		0.714166,
		0.696335,
		0.677898,
		0.658868,
		0.639258,
		0.619082,
		0.598361,
		0.577116,
		0.555372,
		0.533157,
		0.510503,
		0.487448,
		0.46403,
		0.440294,
		0.416289,
		0.39207,
		0.367695,
		0.343228,
		0.318742,
		0.294311,
		0.27002,
		0.245958,
		0.222223,
		0.198921,
		0.176164,
		0.154076,
		0.132788,
		0.112441,
		0.093187,
		0.07519,
		0.058623,
		0.043674,
		0.030544,
		0.019444,
		0.010606,
		0.004273,
		0.000705,
		0.998929,
		0.998532,
		0.998037,
		0.997428,
		0.99669,
		0.995807,
		0.994763,
		0.993541,
		0.992122,
		0.99049,
		0.988626,
		0.98651,
		0.984125,
		0.98145,
		0.978467,
		0.975156,
		0.971498,
		0.967473,
		0.963062,
		0.958246,
		0.953007,
		0.947325,
		0.941183,
		0.934562,
		0.927446,
		0.919818,
		0.911662,
		0.902962,
		0.893703,
		0.883873,
		0.873458,
		0.862446,
		0.850827,
		0.838591,
		0.825729,
		0.812236,
		0.798105,
		0.783333,
		0.767917,
		0.751857,
		0.735156,
		0.717815,
		0.699842,
		0.681244,
		0.662033,
		0.642221,
		0.621825,
		0.600863,
		0.579359,
		0.557338,
		0.53483,
		0.511868,
		0.488488,
		0.464734,
		0.44065,
		0.416289,
		0.391707,
		0.366965,
		0.342132,
		0.317281,
		0.292492,
		0.267854,
		0.243461,
		0.219416,
		0.19583,
		0.172822,
		0.150522,
		0.12907,
		0.108615,
		0.089317,
		0.07135,
		0.054899,
		0.040162,
		0.027353,
		0.016698,
		0.008442,
		0.002844,
		0.000183,
		0.999771,
		0.999635,
		0.999446,
		0.999193,
		0.998863,
		0.998443,
		0.997917,
		0.997272,
		0.99649,
		0.995554,
		0.994448,
		0.993153,
		0.991651,
		0.989923,
		0.987949,
		0.985711,
		0.983187,
		0.980357,
		0.977203,
		0.973702,
		0.969836,
		0.965583,
		0.960924,
		0.955839,
		0.950308,
		0.944313,
		0.937833,
		0.930852,
		0.92335,
		0.915312,
		0.90672,
		0.897559,
		0.887815,
		0.877473,
		0.86652,
		0.854945,
		0.842738,
		0.82989,
		0.816392,
		0.802239,
		0.787425,
		0.771949,
		0.75581,
		0.739008,
		0.721547,
		0.703432,
		0.684672,
		0.665277,
		0.645261,
		0.62464,
		0.603434,
		0.581665,
		0.559361,
		0.536552,
		0.513271,
		0.489559,
		0.465459,
		0.441017,
		0.416289,
		0.391333,
		0.366214,
		0.341002,
		0.315776,
		0.290619,
		0.265625,
		0.240892,
		0.21653,
		0.192655,
		0.169394,
		0.146884,
		0.125272,
		0.104717,
		0.085389,
		0.067471,
		0.051161,
		0.03667,
		0.024223,
		0.014064,
		0.006452,
		0.001665,
		0.999985,
		0.999962,
		0.999922,
		0.999857,
		0.999757,
		0.999612,
		0.999412,
		0.999143,
		0.998793,
		0.998347,
		0.997789,
		0.997104,
		0.996274,
		0.995282,
		0.994109,
		0.992736,
		0.991144,
		0.989313,
		0.987222,
		0.984851,
		0.982179,
		0.979184,
		0.975845,
		0.972142,
		0.968053,
		0.963556,
		0.958631,
		0.953258,
		0.947416,
		0.941085,
		0.934245,
		0.926878,
		0.918966,
		0.91049,
		0.901435,
		0.891784,
		0.881522,
		0.870636,
		0.859113,
		0.846941,
		0.834112,
		0.820615,
		0.806444,
		0.791594,
		0.776061,
		0.759844,
		0.742944,
		0.725363,
		0.707107,
		0.688183,
		0.668603,
		0.64838,
		0.62753,
		0.606074,
		0.584036,
		0.561441,
		0.538323,
		0.514717,
		0.490663,
		0.466205,
		0.441396,
		0.416289,
		0.390948,
		0.365439,
		0.339837,
		0.314225,
		0.28869,
		0.263329,
		0.238249,
		0.213563,
		0.189395,
		0.165879,
		0.14316,
		0.121393,
		0.100747,
		0.081404,
		0.063557,
		0.047417,
		0.033208,
		0.02117,
		0.011564,
		0.004665,
		0.000771,
		1,
		1,
		0.999999,
		0.999995,
		0.999984,
		0.99996,
		0.999917,
		0.999848,
		0.999741,
		0.999588,
		0.999375,
		0.999089,
		0.998717,
		0.998243,
		0.997651,
		0.996923,
		0.996042,
		0.994989,
		0.993744,
		0.992288,
		0.990599,
		0.988657,
		0.98644,
		0.983926,
		0.981094,
		0.977922,
		0.974386,
		0.970465,
		0.966137,
		0.961379,
		0.95617,
		0.950488,
		0.944313,
		0.937623,
		0.930398,
		0.92262,
		0.914268,
		0.905327,
		0.895777,
		0.885604,
		0.874792,
		0.863328,
		0.851199,
		0.838394,
		0.824905,
		0.810721,
		0.795839,
		0.780253,
		0.763962,
		0.746966,
		0.729267,
		0.710869,
		0.691782,
		0.672015,
		0.651582,
		0.630499,
		0.608789,
		0.586474,
		0.563582,
		0.540147,
		0.516206,
		0.4918,
		0.466975,
		0.441786,
		0.416289,
		0.39055,
		0.36464,
		0.338636,
		0.312626,
		0.286701,
		0.260965,
		0.235528,
		0.210511,
		0.186045,
		0.162273,
		0.139346,
		0.117431,
		0.096707,
		0.077365,
		0.059613,
		0.043674,
		0.029788,
		0.018211,
		0.00922,
		0.003111,
		0.000201,
		0.999985,
		0.999995,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999957,
		0.999912,
		0.999838,
		0.999725,
		0.999561,
		0.999334,
		0.999031,
		0.998635,
		0.998131,
		0.997501,
		0.996728,
		0.995792,
		0.994673,
		0.99335,
		0.991804,
		0.990011,
		0.987949,
		0.985597,
		0.982931,
		0.979927,
		0.976563,
		0.972816,
		0.968661,
		0.964076,
		0.959038,
		0.953525,
		0.947512,
		0.94098,
		0.933906,
		0.92627,
		0.918052,
		0.909232,
		0.899793,
		0.889716,
		0.878987,
		0.86759,
		0.855511,
		0.842738,
		0.829262,
		0.815072,
		0.800163,
		0.784529,
		0.768167,
		0.751077,
		0.73326,
		0.714722,
		0.69547,
		0.675515,
		0.654869,
		0.63355,
		0.61158,
		0.588982,
		0.565786,
		0.542026,
		0.51774,
		0.492972,
		0.467769,
		0.442188,
		0.416289,
		0.39014,
		0.363815,
		0.337397,
		0.310976,
		0.28465,
		0.258527,
		0.232725,
		0.207371,
		0.182603,
		0.158573,
		0.135442,
		0.113386,
		0.092596,
		0.073275,
		0.055645,
		0.039943,
		0.026425,
		0.015366,
		0.00706,
		0.001825,
		0.999771,
		0.999861,
		0.999922,
		0.999961,
		0.999984,
		0.999995,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999981,
		0.999955,
		0.999906,
		0.999827,
		0.999707,
		0.999532,
		0.999291,
		0.998968,
		0.998546,
		0.99801,
		0.99734,
		0.996517,
		0.995521,
		0.994331,
		0.992925,
		0.991281,
		0.989376,
		0.987186,
		0.984688,
		0.981857,
		0.978669,
		0.9751,
		0.971124,
		0.966719,
		0.961859,
		0.95652,
		0.950679,
		0.944313,
		0.937398,
		0.929913,
		0.921837,
		0.913148,
		0.903828,
		0.893857,
		0.883219,
		0.871897,
		0.859876,
		0.847143,
		0.833687,
		0.819497,
		0.804566,
		0.788888,
		0.772459,
		0.755278,
		0.737347,
		0.718669,
		0.699252,
		0.679106,
		0.658245,
		0.636686,
		0.61445,
		0.591564,
		0.568056,
		0.543963,
		0.519322,
		0.49418,
		0.468589,
		0.442603,
		0.416289,
		0.389717,
		0.362964,
		0.336118,
		0.309274,
		0.282534,
		0.256014,
		0.229837,
		0.204139,
		0.179066,
		0.154777,
		0.131445,
		0.109257,
		0.088416,
		0.069138,
		0.05166,
		0.036235,
		0.023136,
		0.012657,
		0.005114,
		0.000846,
		0.998929,
		0.999217,
		0.999446,
		0.999624,
		0.999757,
		0.999852,
		0.999917,
		0.999959,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.99998,
		0.999952,
		0.9999,
		0.999816,
		0.999687,
		0.999501,
		0.999244,
		0.9989,
		0.998451,
		0.997879,
		0.997165,
		0.996289,
		0.995229,
		0.993962,
		0.992466,
		0.990717,
		0.988691,
		0.986362,
		0.983707,
		0.980698,
		0.977312,
		0.973521,
		0.969301,
		0.964625,
		0.959469,
		0.953807,
		0.947615,
		0.940869,
		0.933544,
		0.925619,
		0.917071,
		0.907879,
		0.898023,
		0.887485,
		0.876247,
		0.864293,
		0.851607,
		0.838179,
		0.823996,
		0.809049,
		0.793332,
		0.776841,
		0.759572,
		0.741528,
		0.722712,
		0.70313,
		0.682792,
		0.661713,
		0.63991,
		0.617404,
		0.594222,
		0.570395,
		0.545959,
		0.520954,
		0.495428,
		0.469434,
		0.443032,
		0.416289,
		0.389279,
		0.362085,
		0.334797,
		0.307516,
		0.280351,
		0.253422,
		0.226862,
		0.200812,
		0.175429,
		0.150882,
		0.127354,
		0.105044,
		0.084169,
		0.064959,
		0.047667,
		0.032563,
		0.019939,
		0.010111,
		0.003417,
		0.000221,
		0.996876,
		0.997501,
		0.998037,
		0.998488,
		0.998863,
		0.999169,
		0.999412,
		0.9996,
		0.999741,
		0.999843,
		0.999912,
		0.999956,
		0.999981,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.999979,
		0.999948,
		0.999893,
		0.999803,
		0.999666,
		0.999468,
		0.999193,
		0.998826,
		0.998347,
		0.997737,
		0.996976,
		0.996042,
		0.994912,
		0.993563,
		0.991969,
		0.990107,
		0.987949,
		0.985471,
		0.982646,
		0.979446,
		0.975845,
		0.971816,
		0.967332,
		0.962366,
		0.956891,
		0.950882,
		0.944313,
		0.937158,
		0.929393,
		0.920996,
		0.911942,
		0.902212,
		0.891784,
		0.880639,
		0.86876,
		0.856131,
		0.842738,
		0.828569,
		0.813614,
		0.797864,
		0.781314,
		0.763962,
		0.745808,
		0.726854,
		0.707107,
		0.686577,
		0.665277,
		0.643226,
		0.620445,
		0.596961,
		0.572806,
		0.548017,
		0.522638,
		0.496716,
		0.470308,
		0.443475,
		0.416289,
		0.388827,
		0.361176,
		0.333432,
		0.3057,
		0.278096,
		0.250747,
		0.223793,
		0.197385,
		0.171689,
		0.146884,
		0.123166,
		0.100747,
		0.079857,
		0.060743,
		0.043674,
		0.028941,
		0.016857,
		0.007758,
		0.002008,
		0.992969,
		0.994072,
		0.995056,
		0.995927,
		0.99669,
		0.997351,
		0.997917,
		0.998396,
		0.998793,
		0.999117,
		0.999375,
		0.999575,
		0.999725,
		0.999833,
		0.999906,
		0.999953,
		0.99998,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.999977,
		0.999945,
		0.999886,
		0.99979,
		0.999643,
		0.999431,
		0.999138,
		0.998745,
		0.998234,
		0.997583,
		0.996771,
		0.995774,
		0.994569,
		0.99313,
		0.991431,
		0.989445,
		0.987147,
		0.984507,
		0.981498,
		0.978092,
		0.97426,
		0.969973,
		0.965204,
		0.959925,
		0.954108,
		0.947725,
		0.940749,
		0.933156,
		0.924919,
		0.916014,
		0.906419,
		0.896111,
		0.88507,
		0.873276,
		0.860713,
		0.847365,
		0.833218,
		0.81826,
		0.802484,
		0.785882,
		0.76845,
		0.750188,
		0.731099,
		0.711187,
		0.690463,
		0.66894,
		0.646637,
		0.623575,
		0.599783,
		0.575292,
		0.550142,
		0.524376,
		0.498046,
		0.47121,
		0.443933,
		0.416289,
		0.38836,
		0.360237,
		0.332021,
		0.303823,
		0.275766,
		0.247986,
		0.220628,
		0.193856,
		0.167843,
		0.142782,
		0.118882,
		0.096367,
		0.075484,
		0.056497,
		0.039695,
		0.025389,
		0.013913,
		0.005631,
		0.000934,
		0.986574,
		0.988279,
		0.989843,
		0.991269,
		0.992561,
		0.993724,
		0.994763,
		0.995683,
		0.99649,
		0.997189,
		0.997789,
		0.998296,
		0.998717,
		0.999061,
		0.999334,
		0.999547,
		0.999707,
		0.999822,
		0.9999,
		0.99995,
		0.999979,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999998,
		0.999992,
		0.999976,
		0.999941,
		0.999878,
		0.999775,
		0.999618,
		0.999391,
		0.999078,
		0.998658,
		0.998111,
		0.997416,
		0.996548,
		0.995483,
		0.994196,
		0.99266,
		0.990846,
		0.988728,
		0.986276,
		0.983462,
		0.980254,
		0.976625,
		0.972543,
		0.967979,
		0.962903,
		0.957286,
		0.951098,
		0.944313,
		0.9369,
		0.928835,
		0.92009,
		0.910641,
		0.900464,
		0.889537,
		0.877839,
		0.865352,
		0.852057,
		0.837941,
		0.82299,
		0.807193,
		0.790544,
		0.773038,
		0.754672,
		0.735449,
		0.715374,
		0.694455,
		0.672708,
		0.650148,
		0.6268,
		0.602692,
		0.577857,
		0.552335,
		0.526172,
		0.499421,
		0.472143,
		0.444407,
		0.416289,
		0.387877,
		0.359265,
		0.330561,
		0.301882,
		0.273359,
		0.245133,
		0.217363,
		0.190218,
		0.163887,
		0.138573,
		0.114498,
		0.091904,
		0.071053,
		0.05223,
		0.035743,
		0.021925,
		0.011138,
		0.003771,
		0.000244,
		0.977109,
		0.979518,
		0.981771,
		0.983869,
		0.985814,
		0.987608,
		0.989256,
		0.990759,
		0.992122,
		0.99335,
		0.994448,
		0.995421,
		0.996274,
		0.997015,
		0.997651,
		0.998188,
		0.998635,
		0.999,
		0.999291,
		0.999517,
		0.999687,
		0.99981,
		0.999893,
		0.999947,
		0.999977,
		0.999993,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999992,
		0.999974,
		0.999937,
		0.999869,
		0.999759,
		0.999591,
		0.999348,
		0.999012,
		0.998563,
		0.997978,
		0.997234,
		0.996306,
		0.995167,
		0.993791,
		0.992149,
		0.990212,
		0.987949,
		0.985332,
		0.982327,
		0.978905,
		0.975034,
		0.970681,
		0.965817,
		0.960409,
		0.954428,
		0.947842,
		0.940621,
		0.932739,
		0.924165,
		0.914874,
		0.904839,
		0.894038,
		0.882447,
		0.870045,
		0.856815,
		0.842738,
		0.827802,
		0.811993,
		0.795304,
		0.777728,
		0.759263,
		0.739908,
		0.71967,
		0.698557,
		0.676582,
		0.653763,
		0.630124,
		0.605693,
		0.580504,
		0.5546,
		0.528027,
		0.500843,
		0.473108,
		0.444897,
		0.416289,
		0.387376,
		0.358259,
		0.32905,
		0.299874,
		0.270869,
		0.242186,
		0.213992,
		0.186469,
		0.159817,
		0.134253,
		0.110014,
		0.08736,
		0.066571,
		0.047952,
		0.031834,
		0.018575,
		0.008564,
		0.002221,
		0.964076,
		0.967266,
		0.970291,
		0.97315,
		0.975845,
		0.978376,
		0.980745,
		0.982951,
		0.984999,
		0.986889,
		0.988626,
		0.990212,
		0.991651,
		0.992949,
		0.994109,
		0.995138,
		0.996042,
		0.996827,
		0.997501,
		0.998072,
		0.998546,
		0.998934,
		0.999244,
		0.999485,
		0.999666,
		0.999797,
		0.999886,
		0.999943,
		0.999976,
		0.999992,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999991,
		0.999972,
		0.999932,
		0.99986,
		0.999741,
		0.999561,
		0.999301,
		0.998941,
		0.99846,
		0.997833,
		0.997036,
		0.996042,
		0.994823,
		0.99335,
		0.991594,
		0.989522,
		0.987103,
		0.984305,
		0.981094,
		0.977439,
		0.973306,
		0.968661,
		0.963472,
		0.957705,
		0.95133,
		0.944313,
		0.936624,
		0.928233,
		0.919112,
		0.909232,
		0.898568,
		0.887096,
		0.874792,
		0.861636,
		0.84761,
		0.832697,
		0.816885,
		0.800163,
		0.782523,
		0.763962,
		0.74448,
		0.724081,
		0.702773,
		0.680569,
		0.657487,
		0.63355,
		0.608789,
		0.583238,
		0.556941,
		0.529947,
		0.502313,
		0.474108,
		0.445405,
		0.416289,
		0.386858,
		0.357216,
		0.327485,
		0.297795,
		0.268292,
		0.239139,
		0.210511,
		0.182603,
		0.155629,
		0.12982,
		0.105431,
		0.082738,
		0.062044,
		0.043674,
		0.027987,
		0.015366,
		0.006231,
		0.001035,
		0.947082,
		0.951098,
		0.954949,
		0.958631,
		0.962145,
		0.965488,
		0.968661,
		0.971663,
		0.974494,
		0.977154,
		0.979645,
		0.981968,
		0.984125,
		0.986118,
		0.987949,
		0.989624,
		0.991144,
		0.992516,
		0.993744,
		0.994834,
		0.995792,
		0.996625,
		0.99734,
		0.997946,
		0.998451,
		0.998863,
		0.999193,
		0.99945,
		0.999643,
		0.999782,
		0.999878,
		0.999939,
		0.999974,
		0.999991,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.99999,
		0.99997,
		0.999927,
		0.999849,
		0.999722,
		0.999529,
		0.999249,
		0.998863,
		0.998347,
		0.997675,
		0.99682,
		0.995754,
		0.994448,
		0.99287,
		0.990988,
		0.98877,
		0.986181,
		0.983187,
		0.979753,
		0.975845,
		0.971428,
		0.966466,
		0.960924,
		0.954769,
		0.947967,
		0.940484,
		0.932289,
		0.92335,
		0.913638,
		0.903125,
		0.891784,
		0.879589,
		0.86652,
		0.852555,
		0.837676,
		0.821869,
		0.805122,
		0.787425,
		0.768774,
		0.749168,
		0.72861,
		0.707107,
		0.684672,
		0.661323,
		0.637084,
		0.611985,
		0.586062,
		0.559361,
		0.531932,
		0.503836,
		0.475143,
		0.44593,
		0.416289,
		0.38632,
		0.356136,
		0.325863,
		0.295641,
		0.265625,
		0.235987,
		0.206915,
		0.178617,
		0.151319,
		0.125272,
		0.100747,
		0.078042,
		0.057479,
		0.039412,
		0.024223,
		0.012329,
		0.004182,
		0.000272,
		0.925841,
		0.930704,
		0.935405,
		0.939942,
		0.944313,
		0.948514,
		0.952544,
		0.956401,
		0.960083,
		0.96359,
		0.96692,
		0.970072,
		0.973047,
		0.975845,
		0.978467,
		0.980913,
		0.983187,
		0.985289,
		0.987222,
		0.988991,
		0.990599,
		0.99205,
		0.99335,
		0.994505,
		0.995521,
		0.996405,
		0.997165,
		0.997809,
		0.998347,
		0.998786,
		0.999138,
		0.999412,
		0.999618,
		0.999767,
		0.999869,
		0.999934,
		0.999972,
		0.999991,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.99999,
		0.999968,
		0.999921,
		0.999838,
		0.999701,
		0.999493,
		0.999193,
		0.998778,
		0.998223,
		0.997501,
		0.996584,
		0.99544,
		0.994038,
		0.992346,
		0.990328,
		0.987949,
		0.985175,
		0.981968,
		0.978292,
		0.974109,
		0.969383,
		0.964076,
		0.958153,
		0.951577,
		0.944313,
		0.936326,
		0.927583,
		0.918052,
		0.907703,
		0.896506,
		0.884434,
		0.871464,
		0.857572,
		0.842738,
		0.826947,
		0.810183,
		0.792437,
		0.773702,
		0.753975,
		0.73326,
		0.711563,
		0.688896,
		0.665277,
		0.64073,
		0.615286,
		0.588982,
		0.561865,
		0.533987,
		0.505413,
		0.476215,
		0.446476,
		0.416289,
		0.385763,
		0.355015,
		0.324181,
		0.293408,
		0.262862,
		0.232725,
		0.203198,
		0.174504,
		0.146884,
		0.120607,
		0.095965,
		0.073275,
		0.052888,
		0.035182,
		0.02057,
		0.009504,
		0.00247,
		0.900183,
		0.905884,
		0.911434,
		0.916829,
		0.922066,
		0.927141,
		0.932051,
		0.936792,
		0.941362,
		0.945759,
		0.949978,
		0.95402,
		0.957881,
		0.96156,
		0.965057,
		0.968369,
		0.971498,
		0.974442,
		0.977203,
		0.979781,
		0.982179,
		0.984397,
		0.98644,
		0.98831,
		0.990011,
		0.991548,
		0.992925,
		0.99415,
		0.995229,
		0.996168,
		0.996976,
		0.997662,
		0.998234,
		0.998703,
		0.999078,
		0.99937,
		0.999591,
		0.99975,
		0.99986,
		0.999929,
		0.99997,
		0.99999,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999989,
		0.999965,
		0.999915,
		0.999825,
		0.999678,
		0.999454,
		0.999131,
		0.998685,
		0.998088,
		0.997312,
		0.996325,
		0.995096,
		0.99359,
		0.991772,
		0.989606,
		0.987053,
		0.984077,
		0.980638,
		0.976697,
		0.972215,
		0.967154,
		0.961473,
		0.955135,
		0.948101,
		0.940336,
		0.931803,
		0.922467,
		0.912296,
		0.901258,
		0.889324,
		0.876466,
		0.862659,
		0.847883,
		0.832117,
		0.815347,
		0.797559,
		0.778747,
		0.758906,
		0.738037,
		0.716147,
		0.693246,
		0.669354,
		0.644493,
		0.618696,
		0.592002,
		0.564456,
		0.536117,
		0.507048,
		0.477327,
		0.447041,
		0.416289,
		0.385184,
		0.353853,
		0.322436,
		0.291093,
		0.259999,
		0.229348,
		0.199356,
		0.170259,
		0.14232,
		0.115823,
		0.091084,
		0.068445,
		0.048281,
		0.031004,
		0.017058,
		0.006932,
		0.001154,
		0.870045,
		0.876554,
		0.882926,
		0.889157,
		0.895243,
		0.90118,
		0.906963,
		0.912588,
		0.918052,
		0.92335,
		0.928479,
		0.933436,
		0.938217,
		0.942819,
		0.947239,
		0.951476,
		0.955526,
		0.959389,
		0.963062,
		0.966545,
		0.969836,
		0.972936,
		0.975845,
		0.978564,
		0.981094,
		0.983438,
		0.985597,
		0.987575,
		0.989376,
		0.991005,
		0.992466,
		0.993766,
		0.994912,
		0.995911,
		0.996771,
		0.997501,
		0.998111,
		0.998612,
		0.999012,
		0.999325,
		0.999561,
		0.999732,
		0.999849,
		0.999924,
		0.999968,
		0.999989,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999988,
		0.999962,
		0.999909,
		0.999812,
		0.999653,
		0.999412,
		0.999064,
		0.998583,
		0.99794,
		0.997104,
		0.996042,
		0.994719,
		0.993099,
		0.991144,
		0.988816,
		0.986073,
		0.982876,
		0.979184,
		0.974954,
		0.970147,
		0.96472,
		0.958631,
		0.951842,
		0.944313,
		0.936004,
		0.926878,
		0.9169,
		0.906036,
		0.894253,
		0.881522,
		0.867815,
		0.853109,
		0.837381,
		0.820615,
		0.802795,
		0.783913,
		0.763962,
		0.742944,
		0.720862,
		0.697728,
		0.673559,
		0.64838,
		0.622222,
		0.595126,
		0.56714,
		0.538323,
		0.508744,
		0.478481,
		0.447628,
		0.416289,
		0.384583,
		0.352645,
		0.320624,
		0.28869,
		0.257029,
		0.225849,
		0.195381,
		0.165879,
		0.137622,
		0.110919,
		0.086107,
		0.063557,
		0.043674,
		0.026902,
		0.013723,
		0.004665,
		0.000304,
		0.835475,
		0.842738,
		0.849883,
		0.856905,
		0.863799,
		0.870561,
		0.877185,
		0.883666,
		0.890001,
		0.896184,
		0.902212,
		0.908079,
		0.913781,
		0.919314,
		0.924674,
		0.929857,
		0.93486,
		0.93968,
		0.944313,
		0.948756,
		0.953007,
		0.957064,
		0.960924,
		0.964588,
		0.968053,
		0.971319,
		0.974386,
		0.977255,
		0.979927,
		0.982404,
		0.984688,
		0.986782,
		0.988691,
		0.990418,
		0.991969,
		0.99335,
		0.994569,
		0.995632,
		0.996548,
		0.997327,
		0.997978,
		0.998512,
		0.998941,
		0.999276,
		0.999529,
		0.999712,
		0.999838,
		0.999918,
		0.999965,
		0.999988,
		0.999998,
		1,
		1,
		1,
		0.999997,
		0.999987,
		0.999959,
		0.999901,
		0.999797,
		0.999625,
		0.999365,
		0.998989,
		0.99847,
		0.997777,
		0.996876,
		0.995731,
		0.994306,
		0.992561,
		0.990456,
		0.987949,
		0.984999,
		0.981561,
		0.977591,
		0.973047,
		0.967884,
		0.962058,
		0.955526,
		0.948246,
		0.940176,
		0.931276,
		0.921508,
		0.910833,
		0.899218,
		0.886629,
		0.873037,
		0.858415,
		0.842738,
		0.825987,
		0.808146,
		0.789202,
		0.769149,
		0.747985,
		0.725714,
		0.702346,
		0.677898,
		0.652395,
		0.625869,
		0.598361,
		0.569922,
		0.540612,
		0.510503,
		0.47968,
		0.448238,
		0.416289,
		0.383959,
		0.35139,
		0.318742,
		0.286194,
		0.253947,
		0.222223,
		0.191269,
		0.161357,
		0.132788,
		0.105894,
		0.081039,
		0.058623,
		0.039085,
		0.022905,
		0.010606,
		0.002763,
		0.796619,
		0.804566,
		0.812415,
		0.820162,
		0.827802,
		0.835329,
		0.842738,
		0.850025,
		0.857183,
		0.864209,
		0.871096,
		0.877839,
		0.884434,
		0.890876,
		0.897159,
		0.90328,
		0.909232,
		0.915012,
		0.920616,
		0.926039,
		0.931276,
		0.936326,
		0.941183,
		0.945845,
		0.950308,
		0.954571,
		0.958631,
		0.962487,
		0.966137,
		0.96958,
		0.972816,
		0.975845,
		0.978669,
		0.981289,
		0.983707,
		0.985926,
		0.987949,
		0.989783,
		0.991431,
		0.992899,
		0.994196,
		0.995329,
		0.996306,
		0.997137,
		0.997833,
		0.998404,
		0.998863,
		0.999222,
		0.999493,
		0.99969,
		0.999825,
		0.999912,
		0.999962,
		0.999988,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999986,
		0.999956,
		0.999893,
		0.99978,
		0.999595,
		0.999313,
		0.998907,
		0.998347,
		0.997598,
		0.996625,
		0.995389,
		0.993851,
		0.991969,
		0.9897,
		0.986998,
		0.983819,
		0.980117,
		0.975845,
		0.970957,
		0.965405,
		0.959144,
		0.952128,
		0.944313,
		0.935655,
		0.926112,
		0.915644,
		0.904213,
		0.891784,
		0.878322,
		0.863799,
		0.848188,
		0.831466,
		0.813614,
		0.794618,
		0.77447,
		0.753166,
		0.730708,
		0.707107,
		0.682378,
		0.656545,
		0.629643,
		0.601712,
		0.572806,
		0.542987,
		0.512331,
		0.480925,
		0.448872,
		0.416289,
		0.383309,
		0.350084,
		0.316784,
		0.283601,
		0.250747,
		0.218463,
		0.187012,
		0.156687,
		0.127813,
		0.100747,
		0.075884,
		0.053654,
		0.034535,
		0.019046,
		0.007758,
		0.001294,
		0.753719,
		0.762263,
		0.770731,
		0.779121,
		0.787425,
		0.79564,
		0.803759,
		0.811778,
		0.819691,
		0.827492,
		0.835177,
		0.842738,
		0.850172,
		0.857473,
		0.864634,
		0.871652,
		0.878519,
		0.885232,
		0.891784,
		0.89817,
		0.904386,
		0.910426,
		0.916286,
		0.921961,
		0.927446,
		0.932739,
		0.937833,
		0.942727,
		0.947416,
		0.951898,
		0.95617,
		0.96023,
		0.964076,
		0.967708,
		0.971124,
		0.974325,
		0.977312,
		0.980084,
		0.982646,
		0.984999,
		0.987147,
		0.989094,
		0.990846,
		0.99241,
		0.993791,
		0.994999,
		0.996042,
		0.99693,
		0.997675,
		0.998286,
		0.998778,
		0.999163,
		0.999454,
		0.999666,
		0.999812,
		0.999905,
		0.999959,
		0.999987,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999985,
		0.999952,
		0.999884,
		0.999762,
		0.999561,
		0.999256,
		0.998817,
		0.99821,
		0.9974,
		0.996348,
		0.995012,
		0.99335,
		0.991318,
		0.988867,
		0.985952,
		0.982523,
		0.978531,
		0.973927,
		0.968661,
		0.962684,
		0.955947,
		0.948403,
		0.940003,
		0.930704,
		0.920461,
		0.909232,
		0.89698,
		0.883666,
		0.869259,
		0.853729,
		0.837049,
		0.8192,
		0.800163,
		0.779928,
		0.75849,
		0.73585,
		0.712016,
		0.687003,
		0.660837,
		0.63355,
		0.605186,
		0.575799,
		0.545454,
		0.51423,
		0.482221,
		0.449532,
		0.416289,
		0.382633,
		0.348725,
		0.314747,
		0.280903,
		0.247423,
		0.214561,
		0.182603,
		0.151865,
		0.122695,
		0.095481,
		0.070648,
		0.048665,
		0.030049,
		0.015366,
		0.005236,
		0.000342,
		0.707107,
		0.716147,
		0.725136,
		0.73407,
		0.742944,
		0.751752,
		0.760489,
		0.769149,
		0.777728,
		0.78622,
		0.794618,
		0.802918,
		0.811114,
		0.8192,
		0.827169,
		0.835018,
		0.842738,
		0.850326,
		0.857774,
		0.865077,
		0.87223,
		0.879226,
		0.886059,
		0.892725,
		0.899218,
		0.905532,
		0.911662,
		0.917603,
		0.92335,
		0.928899,
		0.934245,
		0.939384,
		0.944313,
		0.949027,
		0.953525,
		0.957802,
		0.961859,
		0.965692,
		0.969301,
		0.972685,
		0.975845,
		0.978782,
		0.981498,
		0.983995,
		0.986276,
		0.988347,
		0.990212,
		0.991877,
		0.99335,
		0.99464,
		0.995754,
		0.996704,
		0.997501,
		0.998157,
		0.998685,
		0.999098,
		0.999412,
		0.999639,
		0.999797,
		0.999897,
		0.999956,
		0.999985,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999984,
		0.999948,
		0.999874,
		0.999741,
		0.999524,
		0.999193,
		0.998717,
		0.99806,
		0.997182,
		0.996042,
		0.994596,
		0.992798,
		0.990599,
		0.987949,
		0.984799,
		0.981094,
		0.976785,
		0.971816,
		0.966137,
		0.959694,
		0.952436,
		0.944313,
		0.935275,
		0.925275,
		0.914268,
		0.902212,
		0.889065,
		0.874792,
		0.859359,
		0.842738,
		0.824905,
		0.805839,
		0.785527,
		0.763962,
		0.741144,
		0.717079,
		0.691782,
		0.665277,
		0.637598,
		0.608789,
		0.578906,
		0.548017,
		0.516206,
		0.483569,
		0.450219,
		0.416289,
		0.381929,
		0.34731,
		0.312626,
		0.278096,
		0.243966,
		0.210511,
		0.178037,
		0.146884,
		0.117431,
		0.090096,
		0.065341,
		0.043674,
		0.025659,
		0.011912,
		0.003111,
		0.657196,
		0.666622,
		0.676022,
		0.685391,
		0.694725,
		0.704019,
		0.713267,
		0.722464,
		0.731605,
		0.740684,
		0.749696,
		0.758636,
		0.767497,
		0.776274,
		0.784961,
		0.793551,
		0.80204,
		0.810421,
		0.818687,
		0.826833,
		0.834852,
		0.842738,
		0.850486,
		0.858088,
		0.865538,
		0.872831,
		0.87996,
		0.88692,
		0.893703,
		0.900305,
		0.90672,
		0.912942,
		0.918966,
		0.924786,
		0.930398,
		0.935798,
		0.94098,
		0.945942,
		0.950679,
		0.955189,
		0.959469,
		0.963517,
		0.967332,
		0.970913,
		0.97426,
		0.977373,
		0.980254,
		0.982906,
		0.985332,
		0.987535,
		0.989522,
		0.991297,
		0.99287,
		0.994248,
		0.99544,
		0.996457,
		0.997312,
		0.998016,
		0.998583,
		0.999027,
		0.999365,
		0.99961,
		0.99978,
		0.999889,
		0.999952,
		0.999984,
		0.999997,
		1,
		1,
		1,
		0.999996,
		0.999982,
		0.999944,
		0.999863,
		0.999719,
		0.999482,
		0.999123,
		0.998606,
		0.997893,
		0.99694,
		0.995704,
		0.994136,
		0.992187,
		0.989805,
		0.986936,
		0.983525,
		0.979518,
		0.974858,
		0.969489,
		0.963355,
		0.956401,
		0.948572,
		0.939815,
		0.930079,
		0.919314,
		0.907473,
		0.894513,
		0.880394,
		0.865077,
		0.848532,
		0.830729,
		0.811648,
		0.79127,
		0.769588,
		0.746597,
		0.722303,
		0.69672,
		0.669872,
		0.641793,
		0.612528,
		0.582134,
		0.550683,
		0.518262,
		0.484973,
		0.450936,
		0.416289,
		0.381194,
		0.345833,
		0.310415,
		0.275172,
		0.24037,
		0.206304,
		0.173305,
		0.14174,
		0.112019,
		0.084596,
		0.059973,
		0.038704,
		0.021402,
		0.008741,
		0.001462,
		0.604481,
		0.614173,
		0.623865,
		0.63355,
		0.643226,
		0.652886,
		0.662527,
		0.672143,
		0.681729,
		0.69128,
		0.700791,
		0.710256,
		0.71967,
		0.729027,
		0.738322,
		0.747548,
		0.756699,
		0.76577,
		0.774754,
		0.783645,
		0.792437,
		0.801123,
		0.809696,
		0.818151,
		0.826481,
		0.834679,
		0.842738,
		0.850653,
		0.858415,
		0.866019,
		0.873458,
		0.880725,
		0.887815,
		0.89472,
		0.901435,
		0.907953,
		0.914268,
		0.920376,
		0.92627,
		0.931946,
		0.937398,
		0.942622,
		0.947615,
		0.952372,
		0.956891,
		0.961169,
		0.965204,
		0.968996,
		0.972543,
		0.975845,
		0.978905,
		0.981724,
		0.984305,
		0.986651,
		0.98877,
		0.990665,
		0.992346,
		0.993819,
		0.995096,
		0.996187,
		0.997104,
		0.99786,
		0.99847,
		0.998949,
		0.999313,
		0.999578,
		0.999762,
		0.99988,
		0.999948,
		0.999983,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999981,
		0.999939,
		0.999851,
		0.999694,
		0.999437,
		0.999046,
		0.998483,
		0.997707,
		0.996672,
		0.995329,
		0.993626,
		0.99151,
		0.988925,
		0.985814,
		0.982117,
		0.977776,
		0.97273,
		0.96692,
		0.960286,
		0.952769,
		0.944313,
		0.93486,
		0.924359,
		0.912756,
		0.900005,
		0.886059,
		0.870879,
		0.854428,
		0.836674,
		0.817592,
		0.797161,
		0.77537,
		0.752213,
		0.727694,
		0.701826,
		0.674631,
		0.646144,
		0.61641,
		0.58549,
		0.553458,
		0.520404,
		0.486437,
		0.451682,
		0.416289,
		0.380428,
		0.344293,
		0.308108,
		0.272124,
		0.236626,
		0.201932,
		0.168399,
		0.136427,
		0.106458,
		0.078987,
		0.05456,
		0.033782,
		0.017323,
		0.00592,
		0.000387,
		0.549528,
		0.559361,
		0.569217,
		0.579092,
		0.588982,
		0.598882,
		0.608789,
		0.618696,
		0.628601,
		0.638497,
		0.64838,
		0.658245,
		0.668086,
		0.677898,
		0.687676,
		0.697414,
		0.707107,
		0.716747,
		0.726331,
		0.73585,
		0.745299,
		0.754672,
		0.763962,
		0.773163,
		0.782268,
		0.79127,
		0.800163,
		0.808939,
		0.817592,
		0.826114,
		0.834498,
		0.842738,
		0.850827,
		0.858756,
		0.86652,
		0.874111,
		0.881522,
		0.888746,
		0.895777,
		0.902608,
		0.909232,
		0.915644,
		0.921837,
		0.927805,
		0.933544,
		0.939048,
		0.944313,
		0.949334,
		0.954108,
		0.958631,
		0.962903,
		0.96692,
		0.970681,
		0.974188,
		0.977439,
		0.980438,
		0.983187,
		0.985689,
		0.987949,
		0.989975,
		0.991772,
		0.99335,
		0.994719,
		0.99589,
		0.996876,
		0.997689,
		0.998347,
		0.998863,
		0.999256,
		0.999543,
		0.999741,
		0.999869,
		0.999944,
		0.999981,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999979,
		0.999933,
		0.999838,
		0.999666,
		0.999385,
		0.99896,
		0.998347,
		0.997501,
		0.996374,
		0.994912,
		0.99306,
		0.990759,
		0.987949,
		0.984569,
		0.980556,
		0.975845,
		0.970373,
		0.964076,
		0.956891,
		0.948756,
		0.939609,
		0.929393,
		0.918052,
		0.905532,
		0.891784,
		0.876762,
		0.860425,
		0.842738,
		0.823672,
		0.803203,
		0.781314,
		0.758,
		0.73326,
		0.707107,
		0.679561,
		0.650658,
		0.620445,
		0.588982,
		0.556348,
		0.522638,
		0.487964,
		0.452462,
		0.416289,
		0.379627,
		0.342684,
		0.3057,
		0.268945,
		0.232725,
		0.197385,
		0.163312,
		0.130939,
		0.100747,
		0.073275,
		0.049119,
		0.028941,
		0.013474,
		0.003529,
		0.492972,
		0.502815,
		0.512705,
		0.522638,
		0.532609,
		0.542616,
		0.552654,
		0.562719,
		0.572806,
		0.582912,
		0.593031,
		0.60316,
		0.613292,
		0.623424,
		0.63355,
		0.643665,
		0.653763,
		0.66384,
		0.673888,
		0.683903,
		0.693878,
		0.703808,
		0.713686,
		0.723506,
		0.73326,
		0.742944,
		0.752549,
		0.762069,
		0.771497,
		0.780826,
		0.790049,
		0.799158,
		0.808146,
		0.817006,
		0.825729,
		0.834309,
		0.842738,
		0.851009,
		0.859113,
		0.867043,
		0.874792,
		0.882352,
		0.889716,
		0.896877,
		0.903828,
		0.910561,
		0.917071,
		0.92335,
		0.929393,
		0.935195,
		0.940749,
		0.946052,
		0.951098,
		0.955885,
		0.960409,
		0.964669,
		0.968661,
		0.972387,
		0.975845,
		0.979038,
		0.981968,
		0.984638,
		0.987053,
		0.98922,
		0.991144,
		0.992836,
		0.994306,
		0.995564,
		0.996625,
		0.997501,
		0.99821,
		0.998768,
		0.999193,
		0.999504,
		0.999719,
		0.999858,
		0.999939,
		0.99998,
		0.999996,
		1,
		1,
		1,
		0.999995,
		0.999977,
		0.999927,
		0.999823,
		0.999635,
		0.999328,
		0.998863,
		0.998194,
		0.997272,
		0.996042,
		0.994448,
		0.992429,
		0.989923,
		0.986864,
		0.983187,
		0.978822,
		0.973702,
		0.967759,
		0.960924,
		0.953131,
		0.944313,
		0.934406,
		0.92335,
		0.911086,
		0.897559,
		0.882719,
		0.86652,
		0.848922,
		0.82989,
		0.809397,
		0.787425,
		0.763962,
		0.739008,
		0.71257,
		0.684672,
		0.655346,
		0.62464,
		0.592618,
		0.559361,
		0.524968,
		0.489559,
		0.453277,
		0.416289,
		0.37879,
		0.341002,
		0.303183,
		0.265625,
		0.228658,
		0.192655,
		0.158037,
		0.125272,
		0.094888,
		0.067471,
		0.043674,
		0.024222,
		0.009922,
		0.001665,
		0.435513,
		0.445233,
		0.455022,
		0.464877,
		0.474793,
		0.484769,
		0.494799,
		0.504881,
		0.515011,
		0.525185,
		0.535398,
		0.545648,
		0.555928,
		0.566235,
		0.576564,
		0.586911,
		0.59727,
		0.607637,
		0.618005,
		0.628371,
		0.638727,
		0.649069,
		0.65939,
		0.669685,
		0.679948,
		0.690172,
		0.70035,
		0.710476,
		0.720543,
		0.730545,
		0.740474,
		0.750323,
		0.760084,
		0.769751,
		0.779315,
		0.788769,
		0.798105,
		0.807315,
		0.816392,
		0.825327,
		0.834112,
		0.842738,
		0.851199,
		0.859486,
		0.86759,
		0.875503,
		0.883219,
		0.890728,
		0.898023,
		0.905097,
		0.911942,
		0.918552,
		0.924919,
		0.931037,
		0.9369,
		0.942504,
		0.947842,
		0.95291,
		0.957705,
		0.962225,
		0.966466,
		0.970427,
		0.974109,
		0.977512,
		0.980638,
		0.98349,
		0.986073,
		0.988393,
		0.990456,
		0.992272,
		0.993851,
		0.995205,
		0.996348,
		0.997294,
		0.99806,
		0.998663,
		0.999123,
		0.99946,
		0.999694,
		0.999845,
		0.999933,
		0.999978,
		0.999995,
		1,
		1,
		1,
		0.999995,
		0.999974,
		0.99992,
		0.999806,
		0.9996,
		0.999265,
		0.998756,
		0.998024,
		0.997015,
		0.995671,
		0.99393,
		0.991726,
		0.988991,
		0.985655,
		0.981647,
		0.976892,
		0.971319,
		0.964853,
		0.957423,
		0.948956,
		0.939384,
		0.928639,
		0.916658,
		0.903379,
		0.888746,
		0.872709,
		0.855222,
		0.836245,
		0.815748,
		0.793707,
		0.770107,
		0.744944,
		0.718226,
		0.689972,
		0.660215,
		0.629005,
		0.596407,
		0.562504,
		0.527402,
		0.491227,
		0.45413,
		0.416289,
		0.377913,
		0.339242,
		0.300551,
		0.262156,
		0.224414,
		0.187731,
		0.152562,
		0.11942,
		0.088879,
		0.061582,
		0.038245,
		0.019668,
		0.00674,
		0.000442,
		0.377913,
		0.387376,
		0.396928,
		0.406567,
		0.416289,
		0.426093,
		0.435974,
		0.44593,
		0.455958,
		0.466054,
		0.476215,
		0.486437,
		0.496716,
		0.507048,
		0.51743,
		0.527856,
		0.538323,
		0.548826,
		0.559361,
		0.569922,
		0.580504,
		0.591103,
		0.601712,
		0.612327,
		0.622942,
		0.63355,
		0.644146,
		0.654724,
		0.665277,
		0.675798,
		0.686282,
		0.69672,
		0.707107,
		0.717434,
		0.727694,
		0.737881,
		0.747985,
		0.758,
		0.767917,
		0.777728,
		0.787425,
		0.797,
		0.806444,
		0.815748,
		0.824905,
		0.833904,
		0.842738,
		0.851398,
		0.859876,
		0.868161,
		0.876247,
		0.884124,
		0.891784,
		0.899218,
		0.906419,
		0.913379,
		0.92009,
		0.926546,
		0.932739,
		0.938663,
		0.944313,
		0.949683,
		0.954769,
		0.959568,
		0.964076,
		0.968292,
		0.972215,
		0.975845,
		0.979184,
		0.982234,
		0.984999,
		0.987485,
		0.9897,
		0.991651,
		0.99335,
		0.994809,
		0.996042,
		0.997064,
		0.997893,
		0.998546,
		0.999046,
		0.999412,
		0.999666,
		0.99983,
		0.999927,
		0.999976,
		0.999995,
		1,
		1,
		1,
		0.999994,
		0.999972,
		0.999912,
		0.999787,
		0.999561,
		0.999193,
		0.998635,
		0.997833,
		0.996728,
		0.995256,
		0.99335,
		0.990939,
		0.987949,
		0.984305,
		0.979927,
		0.974739,
		0.968661,
		0.961616,
		0.953525,
		0.944313,
		0.933906,
		0.922235,
		0.909232,
		0.894835,
		0.878987,
		0.861636,
		0.842738,
		0.822257,
		0.800163,
		0.776439,
		0.751077,
		0.724081,
		0.69547,
		0.665277,
		0.63355,
		0.600358,
		0.565786,
		0.529947,
		0.492972,
		0.455022,
		0.416289,
		0.376995,
		0.337397,
		0.297794,
		0.258526,
		0.21998,
		0.182595,
		0.146868,
		0.113359,
		0.082699,
		0.055596,
		0.032845,
		0.015327,
		0.004022,
		0.320992,
		0.330063,
		0.339242,
		0.348527,
		0.357915,
		0.367405,
		0.376995,
		0.386681,
		0.39646,
		0.406331,
		0.416289,
		0.426333,
		0.436458,
		0.446662,
		0.45694,
		0.46729,
		0.477707,
		0.488188,
		0.498728,
		0.509323,
		0.519969,
		0.530661,
		0.541394,
		0.552163,
		0.562964,
		0.573791,
		0.584639,
		0.595501,
		0.606372,
		0.617247,
		0.628118,
		0.638979,
		0.649825,
		0.660647,
		0.67144,
		0.682196,
		0.692907,
		0.703566,
		0.714166,
		0.724699,
		0.735156,
		0.745529,
		0.75581,
		0.76599,
		0.776061,
		0.786014,
		0.795839,
		0.805528,
		0.815072,
		0.824461,
		0.833687,
		0.842738,
		0.851607,
		0.860284,
		0.86876,
		0.877025,
		0.88507,
		0.892886,
		0.900464,
		0.907796,
		0.914874,
		0.921688,
		0.928233,
		0.934501,
		0.940484,
		0.946178,
		0.951577,
		0.956676,
		0.961473,
		0.965963,
		0.970147,
		0.974023,
		0.977591,
		0.980856,
		0.983819,
		0.986488,
		0.988867,
		0.990967,
		0.992798,
		0.994372,
		0.995704,
		0.99681,
		0.997707,
		0.998417,
		0.99896,
		0.999358,
		0.999635,
		0.999814,
		0.99992,
		0.999973,
		0.999994,
		1,
		1,
		1,
		0.999994,
		0.999969,
		0.999903,
		0.999765,
		0.999517,
		0.999113,
		0.998499,
		0.997618,
		0.996405,
		0.99479,
		0.9927,
		0.990058,
		0.986782,
		0.982792,
		0.978003,
		0.972331,
		0.965692,
		0.958001,
		0.949176,
		0.939136,
		0.927805,
		0.915109,
		0.900978,
		0.885348,
		0.868161,
		0.849368,
		0.828924,
		0.806797,
		0.782964,
		0.757413,
		0.730146,
		0.701178,
		0.670542,
		0.638287,
		0.604481,
		0.569217,
		0.532609,
		0.494799,
		0.455958,
		0.416289,
		0.376029,
		0.335457,
		0.294891,
		0.254699,
		0.215301,
		0.177175,
		0.140868,
		0.106998,
		0.076267,
		0.049465,
		0.027471,
		0.011259,
		0.001888,
		0.265625,
		0.27417,
		0.282841,
		0.291635,
		0.300551,
		0.309587,
		0.318742,
		0.328012,
		0.337397,
		0.346894,
		0.3565,
		0.366214,
		0.376031,
		0.385951,
		0.395969,
		0.406083,
		0.416289,
		0.426585,
		0.436966,
		0.44743,
		0.457972,
		0.468589,
		0.479275,
		0.490028,
		0.500843,
		0.511714,
		0.522638,
		0.533608,
		0.544621,
		0.555671,
		0.566751,
		0.577857,
		0.588982,
		0.60012,
		0.611266,
		0.622411,
		0.63355,
		0.644676,
		0.655781,
		0.666857,
		0.677898,
		0.688896,
		0.699842,
		0.710728,
		0.721547,
		0.732288,
		0.742944,
		0.753505,
		0.763962,
		0.774307,
		0.784529,
		0.794618,
		0.804566,
		0.814362,
		0.823996,
		0.833458,
		0.842738,
		0.851827,
		0.860713,
		0.869387,
		0.877839,
		0.886059,
		0.894038,
		0.901765,
		0.909232,
		0.91643,
		0.92335,
		0.929985,
		0.936326,
		0.942367,
		0.948101,
		0.953525,
		0.958631,
		0.963419,
		0.967884,
		0.972026,
		0.975845,
		0.979343,
		0.982523,
		0.985389,
		0.987949,
		0.990212,
		0.992187,
		0.993888,
		0.995329,
		0.996527,
		0.997501,
		0.998273,
		0.998863,
		0.999297,
		0.9996,
		0.999797,
		0.999912,
		0.999971,
		0.999994,
		1,
		1,
		1,
		0.999993,
		0.999966,
		0.999893,
		0.999741,
		0.999468,
		0.999022,
		0.998347,
		0.997377,
		0.996042,
		0.994266,
		0.991969,
		0.989067,
		0.985471,
		0.981094,
		0.975845,
		0.969633,
		0.962366,
		0.953955,
		0.944313,
		0.933353,
		0.920996,
		0.907163,
		0.891784,
		0.874792,
		0.856131,
		0.835752,
		0.813614,
		0.789689,
		0.763962,
		0.73643,
		0.707107,
		0.676022,
		0.643226,
		0.608789,
		0.572806,
		0.535398,
		0.496713,
		0.456932,
		0.416269,
		0.374977,
		0.333351,
		0.291735,
		0.250527,
		0.21019,
		0.171255,
		0.134336,
		0.100129,
		0.069425,
		0.043104,
		0.022135,
		0.007563,
		0.000494,
		0.212739,
		0.220628,
		0.228658,
		0.236827,
		0.245133,
		0.253577,
		0.262156,
		0.270869,
		0.279714,
		0.28869,
		0.297795,
		0.307026,
		0.316383,
		0.325863,
		0.335463,
		0.345181,
		0.355015,
		0.364963,
		0.37502,
		0.385184,
		0.395453,
		0.405822,
		0.416289,
		0.42685,
		0.437501,
		0.448238,
		0.459058,
		0.469955,
		0.480925,
		0.491965,
		0.503068,
		0.51423,
		0.525446,
		0.536711,
		0.548017,
		0.559361,
		0.570735,
		0.582134,
		0.59355,
		0.604978,
		0.61641,
		0.627839,
		0.639258,
		0.650658,
		0.662033,
		0.673374,
		0.684672,
		0.695919,
		0.707107,
		0.718226,
		0.729267,
		0.74022,
		0.751077,
		0.761826,
		0.772459,
		0.782964,
		0.793332,
		0.803552,
		0.813614,
		0.823506,
		0.833218,
		0.842738,
		0.852057,
		0.861163,
		0.870045,
		0.878693,
		0.887096,
		0.895243,
		0.903125,
		0.910731,
		0.918052,
		0.925079,
		0.931803,
		0.938217,
		0.944313,
		0.950084,
		0.955526,
		0.960634,
		0.965405,
		0.969836,
		0.973927,
		0.977679,
		0.981094,
		0.984178,
		0.986936,
		0.989376,
		0.99151,
		0.99335,
		0.994912,
		0.996213,
		0.997272,
		0.998111,
		0.998756,
		0.99923,
		0.999561,
		0.999776,
		0.999903,
		0.999968,
		0.999993,
		1,
		1,
		1,
		0.999993,
		0.999962,
		0.999882,
		0.999714,
		0.999412,
		0.99892,
		0.998174,
		0.997104,
		0.995632,
		0.993675,
		0.991144,
		0.987949,
		0.983995,
		0.979184,
		0.973418,
		0.9666,
		0.958631,
		0.949417,
		0.938862,
		0.926878,
		0.913379,
		0.898284,
		0.881522,
		0.863025,
		0.842738,
		0.820615,
		0.796619,
		0.770731,
		0.742944,
		0.713267,
		0.681729,
		0.648379,
		0.613289,
		0.576554,
		0.538297,
		0.498671,
		0.457861,
		0.416091,
		0.373628,
		0.330792,
		0.287961,
		0.245582,
		0.204182,
		0.164372,
		0.126853,
		0.092422,
		0.061965,
		0.036449,
		0.016906,
		0.0044,
		0.163312,
		0.170419,
		0.177679,
		0.185091,
		0.192655,
		0.200371,
		0.208236,
		0.216251,
		0.224414,
		0.232725,
		0.241182,
		0.249783,
		0.258527,
		0.267414,
		0.27644,
		0.285604,
		0.294906,
		0.304341,
		0.313909,
		0.323607,
		0.333432,
		0.343382,
		0.353455,
		0.363647,
		0.373956,
		0.384378,
		0.39491,
		0.405548,
		0.416289,
		0.427129,
		0.438064,
		0.449089,
		0.460201,
		0.471394,
		0.482664,
		0.494005,
		0.505413,
		0.516882,
		0.528406,
		0.539979,
		0.551596,
		0.563249,
		0.574932,
		0.586639,
		0.598361,
		0.610092,
		0.621825,
		0.63355,
		0.645261,
		0.656948,
		0.668603,
		0.680218,
		0.691782,
		0.703287,
		0.714722,
		0.726079,
		0.737347,
		0.748515,
		0.759572,
		0.770509,
		0.781314,
		0.791976,
		0.802484,
		0.812825,
		0.82299,
		0.832964,
		0.842738,
		0.8523,
		0.861636,
		0.870737,
		0.879589,
		0.888183,
		0.896506,
		0.904547,
		0.912296,
		0.919743,
		0.926878,
		0.933692,
		0.940176,
		0.946324,
		0.952128,
		0.957583,
		0.962684,
		0.967429,
		0.971816,
		0.975845,
		0.979518,
		0.982839,
		0.985814,
		0.98845,
		0.990759,
		0.992753,
		0.994448,
		0.995862,
		0.997015,
		0.997931,
		0.998635,
		0.999154,
		0.999517,
		0.999754,
		0.999893,
		0.999964,
		0.999993,
		1,
		1,
		0.999999,
		0.999992,
		0.999958,
		0.999869,
		0.999683,
		0.999348,
		0.998803,
		0.997978,
		0.996795,
		0.995167,
		0.993005,
		0.990212,
		0.986687,
		0.982327,
		0.977027,
		0.970681,
		0.963183,
		0.954428,
		0.944313,
		0.932739,
		0.91961,
		0.904839,
		0.888342,
		0.870045,
		0.849883,
		0.827802,
		0.803759,
		0.777728,
		0.749695,
		0.719666,
		0.687664,
		0.653732,
		0.617935,
		0.580364,
		0.541136,
		0.500403,
		0.458357,
		0.415238,
		0.371341,
		0.327031,
		0.282748,
		0.239021,
		0.19647,
		0.155812,
		0.11786,
		0.083512,
		0.053735,
		0.029542,
		0.011958,
		0.001976,
		0.118371,
		0.124573,
		0.130939,
		0.137468,
		0.144161,
		0.151018,
		0.158037,
		0.165218,
		0.172561,
		0.180066,
		0.187731,
		0.195557,
		0.203541,
		0.211684,
		0.219983,
		0.228439,
		0.237049,
		0.245813,
		0.254728,
		0.263794,
		0.273008,
		0.282369,
		0.291874,
		0.301522,
		0.31131,
		0.321236,
		0.331297,
		0.34149,
		0.351814,
		0.362263,
		0.372836,
		0.383529,
		0.394338,
		0.405259,
		0.416289,
		0.427424,
		0.438658,
		0.449987,
		0.461407,
		0.472913,
		0.484498,
		0.496159,
		0.507888,
		0.51968,
		0.531529,
		0.543429,
		0.555372,
		0.567351,
		0.579359,
		0.591389,
		0.603434,
		0.615483,
		0.62753,
		0.639566,
		0.651582,
		0.663568,
		0.675515,
		0.687413,
		0.699252,
		0.711022,
		0.722712,
		0.734311,
		0.745808,
		0.757191,
		0.76845,
		0.779572,
		0.790544,
		0.801356,
		0.811993,
		0.822445,
		0.832697,
		0.842738,
		0.852555,
		0.862134,
		0.871464,
		0.880531,
		0.889324,
		0.897829,
		0.906036,
		0.913932,
		0.921508,
		0.928752,
		0.935655,
		0.942207,
		0.948403,
		0.954233,
		0.959694,
		0.96478,
		0.969489,
		0.973821,
		0.977776,
		0.981357,
		0.984569,
		0.987421,
		0.989923,
		0.992087,
		0.99393,
		0.99547,
		0.996728,
		0.997729,
		0.998499,
		0.999069,
		0.999468,
		0.999728,
		0.999882,
		0.99996,
		0.999992,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999954,
		0.999854,
		0.999648,
		0.999276,
		0.998671,
		0.997755,
		0.996443,
		0.99464,
		0.992245,
		0.989154,
		0.985256,
		0.980438,
		0.974587,
		0.967586,
		0.959323,
		0.949683,
		0.938558,
		0.925841,
		0.911434,
		0.895243,
		0.877185,
		0.857183,
		0.835175,
		0.811109,
		0.784947,
		0.756663,
		0.726248,
		0.69371,
		0.659076,
		0.622395,
		0.583747,
		0.543246,
		0.501051,
		0.457378,
		0.412507,
		0.366798,
		0.320699,
		0.274752,
		0.229604,
		0.185998,
		0.144773,
		0.106843,
		0.073179,
		0.044773,
		0.0226,
		0.007571,
		0.000483,
		0.078987,
		0.084169,
		0.089522,
		0.095049,
		0.100747,
		0.106619,
		0.112664,
		0.118882,
		0.125272,
		0.131836,
		0.138573,
		0.145482,
		0.152563,
		0.159817,
		0.167242,
		0.174837,
		0.182603,
		0.190539,
		0.198643,
		0.206915,
		0.215353,
		0.223957,
		0.232725,
		0.241656,
		0.250747,
		0.259999,
		0.269407,
		0.278972,
		0.28869,
		0.298559,
		0.308577,
		0.318742,
		0.32905,
		0.339498,
		0.350084,
		0.360804,
		0.371655,
		0.382633,
		0.393734,
		0.404954,
		0.416289,
		0.427734,
		0.439285,
		0.450936,
		0.462682,
		0.474517,
		0.486437,
		0.498434,
		0.510503,
		0.522638,
		0.53483,
		0.547074,
		0.559361,
		0.571684,
		0.584036,
		0.596407,
		0.608789,
		0.621173,
		0.63355,
		0.645911,
		0.658245,
		0.670542,
		0.682792,
		0.694984,
		0.707107,
		0.719149,
		0.731099,
		0.742944,
		0.754672,
		0.766271,
		0.777728,
		0.78903,
		0.800163,
		0.811114,
		0.821869,
		0.832415,
		0.842738,
		0.852824,
		0.862659,
		0.87223,
		0.881522,
		0.890522,
		0.899218,
		0.907596,
		0.915644,
		0.92335,
		0.930704,
		0.937694,
		0.944313,
		0.950551,
		0.956401,
		0.961859,
		0.96692,
		0.971582,
		0.975845,
		0.979712,
		0.983187,
		0.986276,
		0.988991,
		0.991344,
		0.99335,
		0.99503,
		0.996405,
		0.997501,
		0.998347,
		0.998972,
		0.999412,
		0.999699,
		0.999869,
		0.999956,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.99999,
		0.999948,
		0.999838,
		0.999607,
		0.999193,
		0.99852,
		0.997501,
		0.996042,
		0.994038,
		0.991379,
		0.987949,
		0.983628,
		0.978292,
		0.971816,
		0.964076,
		0.954949,
		0.944313,
		0.932051,
		0.918052,
		0.90221,
		0.884429,
		0.864619,
		0.842698,
		0.818594,
		0.792245,
		0.763598,
		0.732617,
		0.699281,
		0.663597,
		0.625599,
		0.585368,
		0.543038,
		0.498808,
		0.452958,
		0.405858,
		0.357979,
		0.309899,
		0.262302,
		0.215973,
		0.171783,
		0.130667,
		0.093595,
		0.061533,
		0.0354,
		0.016015,
		0.004055,
		0.046277,
		0.05033,
		0.05456,
		0.058969,
		0.063557,
		0.068326,
		0.073275,
		0.078406,
		0.083718,
		0.089212,
		0.094888,
		0.100747,
		0.106789,
		0.113014,
		0.119423,
		0.126014,
		0.132788,
		0.139745,
		0.146884,
		0.154206,
		0.16171,
		0.169394,
		0.17726,
		0.185305,
		0.193529,
		0.201932,
		0.210511,
		0.219266,
		0.228195,
		0.237298,
		0.246571,
		0.256014,
		0.265625,
		0.275401,
		0.285341,
		0.295441,
		0.3057,
		0.316114,
		0.326681,
		0.337397,
		0.34826,
		0.359265,
		0.370409,
		0.381687,
		0.393097,
		0.404632,
		0.416289,
		0.428063,
		0.439948,
		0.451939,
		0.46403,
		0.476215,
		0.488488,
		0.500843,
		0.513271,
		0.525768,
		0.538323,
		0.550931,
		0.563582,
		0.576269,
		0.588982,
		0.601712,
		0.61445,
		0.627186,
		0.63991,
		0.65261,
		0.665277,
		0.677898,
		0.690463,
		0.702959,
		0.715374,
		0.727694,
		0.739908,
		0.752002,
		0.763962,
		0.775775,
		0.787425,
		0.798899,
		0.810183,
		0.82126,
		0.832117,
		0.842738,
		0.853109,
		0.863213,
		0.873037,
		0.882566,
		0.891784,
		0.900677,
		0.909232,
		0.917436,
		0.925275,
		0.932739,
		0.939815,
		0.946495,
		0.952769,
		0.958631,
		0.964076,
		0.9691,
		0.973702,
		0.977883,
		0.981647,
		0.984999,
		0.987949,
		0.990511,
		0.9927,
		0.994536,
		0.996042,
		0.997245,
		0.998174,
		0.998863,
		0.999348,
		0.999666,
		0.999854,
		0.999951,
		0.99999,
		0.999999,
		1,
		0.999999,
		0.999988,
		0.999942,
		0.999819,
		0.999561,
		0.999098,
		0.998347,
		0.99721,
		0.995583,
		0.99335,
		0.99039,
		0.986574,
		0.981771,
		0.975845,
		0.968661,
		0.960083,
		0.949977,
		0.938211,
		0.924657,
		0.909189,
		0.891682,
		0.872018,
		0.85008,
		0.825756,
		0.798942,
		0.769549,
		0.737504,
		0.702768,
		0.66534,
		0.62527,
		0.582681,
		0.537774,
		0.490845,
		0.442298,
		0.392648,
		0.342523,
		0.292662,
		0.243899,
		0.197141,
		0.153342,
		0.113465,
		0.078437,
		0.049109,
		0.026203,
		0.010267,
		0.001638,
		0.021402,
		0.024223,
		0.027226,
		0.030412,
		0.033782,
		0.037337,
		0.041077,
		0.045004,
		0.049119,
		0.053423,
		0.057915,
		0.062598,
		0.067471,
		0.072536,
		0.077793,
		0.083242,
		0.088884,
		0.094719,
		0.100747,
		0.10697,
		0.113386,
		0.119997,
		0.126801,
		0.133799,
		0.140991,
		0.148377,
		0.155955,
		0.163726,
		0.171689,
		0.179843,
		0.188187,
		0.196721,
		0.205443,
		0.214352,
		0.223446,
		0.232725,
		0.242186,
		0.251827,
		0.261647,
		0.271644,
		0.281814,
		0.292156,
		0.302666,
		0.313342,
		0.324181,
		0.335179,
		0.346332,
		0.357638,
		0.369091,
		0.380687,
		0.392422,
		0.404291,
		0.416289,
		0.428411,
		0.44065,
		0.453002,
		0.465459,
		0.478015,
		0.490663,
		0.503396,
		0.516206,
		0.529086,
		0.542026,
		0.55502,
		0.568056,
		0.581127,
		0.594222,
		0.607332,
		0.620445,
		0.63355,
		0.646637,
		0.659694,
		0.672708,
		0.685666,
		0.698557,
		0.711367,
		0.724081,
		0.736686,
		0.749168,
		0.761512,
		0.773702,
		0.785723,
		0.797559,
		0.809195,
		0.820615,
		0.831801,
		0.842738,
		0.85341,
		0.863799,
		0.87389,
		0.883666,
		0.893112,
		0.902212,
		0.91095,
		0.919314,
		0.927288,
		0.93486,
		0.94202,
		0.948756,
		0.95506,
		0.960924,
		0.966345,
		0.971319,
		0.975845,
		0.979927,
		0.98357,
		0.986782,
		0.989577,
		0.991969,
		0.99398,
		0.995632,
		0.996954,
		0.997978,
		0.998739,
		0.999276,
		0.999628,
		0.999838,
		0.999945,
		0.999988,
		0.999999,
		1,
		0.999999,
		0.999987,
		0.999935,
		0.999797,
		0.999508,
		0.998989,
		0.998148,
		0.996876,
		0.995056,
		0.992561,
		0.989256,
		0.984999,
		0.979644,
		0.973042,
		0.965039,
		0.955481,
		0.944206,
		0.931051,
		0.915849,
		0.898429,
		0.878618,
		0.856247,
		0.831152,
		0.803188,
		0.772234,
		0.73821,
		0.701088,
		0.660911,
		0.617806,
		0.572,
		0.523831,
		0.473757,
		0.422357,
		0.370327,
		0.318468,
		0.267665,
		0.218859,
		0.173013,
		0.131066,
		0.093895,
		0.062266,
		0.036791,
		0.017889,
		0.005759,
		0.000352,
		0.005562,
		0.00706,
		0.008741,
		0.010606,
		0.012657,
		0.014896,
		0.017323,
		0.019939,
		0.022747,
		0.025748,
		0.028941,
		0.03233,
		0.035914,
		0.039695,
		0.043674,
		0.047852,
		0.05223,
		0.056809,
		0.061589,
		0.066571,
		0.071756,
		0.077145,
		0.082738,
		0.088536,
		0.094539,
		0.100747,
		0.107161,
		0.113781,
		0.120607,
		0.127639,
		0.134877,
		0.14232,
		0.149968,
		0.157821,
		0.165879,
		0.17414,
		0.182603,
		0.191269,
		0.200135,
		0.2092,
		0.218463,
		0.227922,
		0.237576,
		0.247423,
		0.25746,
		0.267685,
		0.278096,
		0.28869,
		0.299464,
		0.310415,
		0.321539,
		0.332833,
		0.344293,
		0.355915,
		0.367695,
		0.379627,
		0.391707,
		0.40393,
		0.416289,
		0.42878,
		0.441396,
		0.45413,
		0.466975,
		0.479925,
		0.492972,
		0.506107,
		0.519322,
		0.532609,
		0.545959,
		0.559361,
		0.572806,
		0.586284,
		0.599783,
		0.613292,
		0.6268,
		0.640295,
		0.653763,
		0.667192,
		0.680569,
		0.693878,
		0.707107,
		0.720239,
		0.73326,
		0.746155,
		0.758906,
		0.771497,
		0.783913,
		0.796135,
		0.808146,
		0.819929,
		0.831466,
		0.842738,
		0.853729,
		0.864419,
		0.874792,
		0.884829,
		0.894513,
		0.903828,
		0.912756,
		0.921283,
		0.929393,
		0.937074,
		0.944313,
		0.951098,
		0.957423,
		0.963278,
		0.968661,
		0.973569,
		0.978003,
		0.981968,
		0.985471,
		0.988525,
		0.991144,
		0.99335,
		0.995167,
		0.996625,
		0.997755,
		0.998598,
		0.999193,
		0.999585,
		0.999819,
		0.999939,
		0.999987,
		0.999999,
		1,
		0.999999,
		0.999985,
		0.999927,
		0.999771,
		0.999446,
		0.998863,
		0.997917,
		0.996489,
		0.994447,
		0.991646,
		0.987932,
		0.983139,
		0.977092,
		0.969601,
		0.960466,
		0.949475,
		0.936403,
		0.921018,
		0.903084,
		0.882368,
		0.858647,
		0.831725,
		0.801442,
		0.767694,
		0.730447,
		0.689759,
		0.645788,
		0.598814,
		0.549241,
		0.497602,
		0.444555,
		0.390871,
		0.337415,
		0.285117,
		0.234937,
		0.187825,
		0.144681,
		0.106308,
		0.073371,
		0.046363,
		0.025572,
		0.011064,
		0.002672,
		9.1e-005,
		0.000363,
		0.00082,
		0.001462,
		0.002292,
		0.00331,
		0.004519,
		0.00592,
		0.007514,
		0.009304,
		0.01129,
		0.013475,
		0.015859,
		0.018444,
		0.021232,
		0.024223,
		0.02742,
		0.030823,
		0.034434,
		0.038254,
		0.042284,
		0.046525,
		0.050979,
		0.055645,
		0.060526,
		0.065621,
		0.070933,
		0.076461,
		0.082205,
		0.088168,
		0.094348,
		0.100747,
		0.107365,
		0.114202,
		0.121258,
		0.128533,
		0.136026,
		0.143738,
		0.151669,
		0.159817,
		0.168182,
		0.176763,
		0.185559,
		0.19457,
		0.203793,
		0.213228,
		0.222873,
		0.232725,
		0.242783,
		0.253045,
		0.263508,
		0.27417,
		0.285028,
		0.296078,
		0.307317,
		0.318742,
		0.330348,
		0.342132,
		0.354089,
		0.366214,
		0.378502,
		0.390948,
		0.403546,
		0.416289,
		0.429172,
		0.442188,
		0.455329,
		0.468589,
		0.481957,
		0.495428,
		0.508991,
		0.522638,
		0.536358,
		0.550142,
		0.563978,
		0.577857,
		0.591766,
		0.605693,
		0.619625,
		0.63355,
		0.647454,
		0.661323,
		0.675142,
		0.688896,
		0.70257,
		0.716147,
		0.72961,
		0.742944,
		0.756129,
		0.769149,
		0.781985,
		0.794618,
		0.807029,
		0.8192,
		0.831109,
		0.842738,
		0.854068,
		0.865077,
		0.875748,
		0.886059,
		0.895994,
		0.905532,
		0.914656,
		0.92335,
		0.931598,
		0.939384,
		0.946697,
		0.953525,
		0.959858,
		0.965692,
		0.971021,
		0.975845,
		0.980168,
		0.983995,
		0.987338,
		0.990212,
		0.992637,
		0.99464,
		0.996249,
		0.997501,
		0.998436,
		0.999098,
		0.999535,
		0.999797,
		0.999931,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999984,
		0.999917,
		0.999741,
		0.999373,
		0.998711,
		0.997633,
		0.995994,
		0.993631,
		0.990359,
		0.98597,
		0.980234,
		0.9729,
		0.963696,
		0.952336,
		0.938519,
		0.921945,
		0.90232,
		0.879373,
		0.852871,
		0.822636,
		0.788563,
		0.750642,
		0.708972,
		0.663773,
		0.615402,
		0.564348,
		0.511233,
		0.456801,
		0.401894,
		0.347429,
		0.294361,
		0.243645,
		0.196197,
		0.152846,
		0.1143,
		0.08111,
		0.053638,
		0.032043,
		0.01627,
		0.00605,
		0.000913,
		9.6e-005,
		0.000387,
		0.000874,
		0.001559,
		0.002443,
		0.003529,
		0.004818,
		0.006313,
		0.008014,
		0.009923,
		0.012042,
		0.014374,
		0.016918,
		0.019678,
		0.022654,
		0.025848,
		0.029261,
		0.032895,
		0.036751,
		0.04083,
		0.045134,
		0.049664,
		0.054421,
		0.059405,
		0.064618,
		0.070061,
		0.075735,
		0.08164,
		0.087777,
		0.094146,
		0.100747,
		0.107582,
		0.114651,
		0.121952,
		0.129487,
		0.137255,
		0.145256,
		0.15349,
		0.161955,
		0.170651,
		0.179577,
		0.188732,
		0.198115,
		0.207724,
		0.217558,
		0.227614,
		0.237891,
		0.248386,
		0.259096,
		0.27002,
		0.281153,
		0.292492,
		0.304035,
		0.315776,
		0.327712,
		0.339837,
		0.352149,
		0.36464,
		0.377306,
		0.39014,
		0.403137,
		0.416289,
		0.42959,
		0.443032,
		0.456608,
		0.470308,
		0.484124,
		0.498046,
		0.512066,
		0.526172,
		0.540354,
		0.5546,
		0.568899,
		0.583238,
		0.597605,
		0.611985,
		0.626365,
		0.64073,
		0.655065,
		0.669354,
		0.683581,
		0.697728,
		0.711778,
		0.725714,
		0.739516,
		0.753166,
		0.766643,
		0.779928,
		0.793,
		0.805839,
		0.818422,
		0.830729,
		0.842738,
		0.854428,
		0.865776,
		0.876762,
		0.887363,
		0.897559,
		0.907331,
		0.916658,
		0.925522,
		0.933906,
		0.941795,
		0.949176,
		0.956035,
		0.962366,
		0.968161,
		0.973418,
		0.978139,
		0.982327,
		0.985994,
		0.989154,
		0.991827,
		0.994038,
		0.99582,
		0.99721,
		0.998251,
		0.998989,
		0.999478,
		0.999771,
		0.999922,
		0.999984,
		0.999999,
		1,
		0.999997,
		0.999976,
		0.999888,
		0.999659,
		0.999178,
		0.998305,
		0.996864,
		0.994648,
		0.991411,
		0.986877,
		0.980739,
		0.972665,
		0.962301,
		0.949288,
		0.93327,
		0.913908,
		0.890903,
		0.864009,
		0.83306,
		0.797981,
		0.758816,
		0.715733,
		0.669042,
		0.61919,
		0.566765,
		0.512479,
		0.45715,
		0.401677,
		0.347001,
		0.294076,
		0.243819,
		0.197077,
		0.154586,
		0.116937,
		0.084551,
		0.057659,
		0.036295,
		0.020295,
		0.009308,
		0.002816,
		0.000161,
		0.000103,
		0.000414,
		0.000934,
		0.001665,
		0.00261,
		0.003771,
		0.005149,
		0.006746,
		0.008564,
		0.010606,
		0.012873,
		0.015366,
		0.018088,
		0.02104,
		0.024223,
		0.027641,
		0.031293,
		0.035182,
		0.039308,
		0.043674,
		0.048281,
		0.05313,
		0.058222,
		0.063557,
		0.069138,
		0.074965,
		0.081039,
		0.08736,
		0.09393,
		0.100747,
		0.107814,
		0.11513,
		0.122695,
		0.130509,
		0.138573,
		0.146884,
		0.155444,
		0.164251,
		0.173305,
		0.182603,
		0.192146,
		0.201932,
		0.211958,
		0.222223,
		0.232725,
		0.243461,
		0.254429,
		0.265625,
		0.277047,
		0.28869,
		0.300551,
		0.312626,
		0.324909,
		0.337397,
		0.350084,
		0.362964,
		0.376031,
		0.389279,
		0.402701,
		0.416289,
		0.430036,
		0.443933,
		0.457972,
		0.472143,
		0.486437,
		0.500843,
		0.51535,
		0.529947,
		0.544621,
		0.559361,
		0.574153,
		0.588982,
		0.603835,
		0.618696,
		0.63355,
		0.64838,
		0.663169,
		0.677898,
		0.692551,
		0.707107,
		0.721547,
		0.73585,
		0.749996,
		0.763962,
		0.777728,
		0.79127,
		0.804566,
		0.817592,
		0.830324,
		0.842738,
		0.854812,
		0.86652,
		0.877839,
		0.888746,
		0.899218,
		0.909232,
		0.918768,
		0.927805,
		0.936326,
		0.944313,
		0.951752,
		0.958631,
		0.964943,
		0.970681,
		0.975845,
		0.980438,
		0.984468,
		0.987949,
		0.990902,
		0.99335,
		0.995329,
		0.996876,
		0.998037,
		0.998863,
		0.999412,
		0.999741,
		0.999911,
		0.999976,
		0.999981,
		0.999952,
		0.999885,
		0.999737,
		0.999416,
		0.998789,
		0.997669,
		0.995825,
		0.992978,
		0.988808,
		0.982956,
		0.975038,
		0.964651,
		0.951392,
		0.934871,
		0.914735,
		0.890684,
		0.862495,
		0.830044,
		0.793321,
		0.75245,
		0.707694,
		0.659459,
		0.608295,
		0.55488,
		0.5,
		0.444525,
		0.389377,
		0.33549,
		0.283774,
		0.235077,
		0.190145,
		0.149594,
		0.113883,
		0.083299,
		0.057942,
		0.037732,
		0.022412,
		0.011571,
		0.004664,
		0.001043,
		0.00011,
		0.000443,
		0.000999,
		0.001783,
		0.002795,
		0.004038,
		0.005514,
		0.007225,
		0.009174,
		0.011362,
		0.013791,
		0.016464,
		0.019382,
		0.022547,
		0.025962,
		0.029626,
		0.033544,
		0.037715,
		0.042142,
		0.046826,
		0.051768,
		0.05697,
		0.062433,
		0.068159,
		0.074147,
		0.080399,
		0.086916,
		0.093699,
		0.100747,
		0.108062,
		0.115644,
		0.123492,
		0.131607,
		0.139988,
		0.148635,
		0.157547,
		0.166724,
		0.176164,
		0.185867,
		0.19583,
		0.206051,
		0.21653,
		0.227264,
		0.238249,
		0.249484,
		0.260965,
		0.272688,
		0.28465,
		0.296847,
		0.309274,
		0.321925,
		0.334797,
		0.347883,
		0.361176,
		0.374671,
		0.38836,
		0.402235,
		0.416289,
		0.430513,
		0.444897,
		0.459432,
		0.474108,
		0.488913,
		0.503836,
		0.518865,
		0.533987,
		0.549189,
		0.564456,
		0.579774,
		0.595126,
		0.610497,
		0.625869,
		0.641225,
		0.656545,
		0.671812,
		0.687003,
		0.7021,
		0.717079,
		0.731919,
		0.746597,
		0.761088,
		0.77537,
		0.789417,
		0.803203,
		0.816702,
		0.82989,
		0.842738,
		0.855222,
		0.867313,
		0.878987,
		0.890217,
		0.900978,
		0.911245,
		0.920996,
		0.930208,
		0.938862,
		0.94694,
		0.954428,
		0.961312,
		0.967586,
		0.973246,
		0.978292,
		0.98273,
		0.986574,
		0.989843,
		0.992561,
		0.994763,
		0.996489,
		0.997788,
		0.998711,
		0.999317,
		0.999659,
		0.999787,
		0.999737,
		0.999522,
		0.999123,
		0.998473,
		0.99745,
		0.995868,
		0.993485,
		0.99,
		0.985067,
		0.978299,
		0.969283,
		0.957597,
		0.942826,
		0.924584,
		0.902538,
		0.876427,
		0.846087,
		0.811468,
		0.77265,
		0.729854,
		0.683445,
		0.633929,
		0.581942,
		0.528227,
		0.473615,
		0.418991,
		0.365257,
		0.313298,
		0.263942,
		0.217932,
		0.175889,
		0.138292,
		0.105461,
		0.077548,
		0.054533,
		0.036236,
		0.02233,
		0.012361,
		0.005777,
		0.001962,
		0.000268,
		0.000118,
		0.000475,
		0.001072,
		0.001913,
		0.003,
		0.004334,
		0.00592,
		0.007758,
		0.009851,
		0.012202,
		0.014812,
		0.017685,
		0.020821,
		0.024223,
		0.027894,
		0.031834,
		0.036046,
		0.040531,
		0.045292,
		0.05033,
		0.055645,
		0.061241,
		0.067117,
		0.073275,
		0.079717,
		0.086442,
		0.093452,
		0.100747,
		0.108329,
		0.116196,
		0.124349,
		0.132788,
		0.141513,
		0.150522,
		0.159817,
		0.169394,
		0.179254,
		0.189395,
		0.199814,
		0.210511,
		0.221482,
		0.232725,
		0.244237,
		0.256014,
		0.268054,
		0.280351,
		0.292901,
		0.3057,
		0.318742,
		0.332021,
		0.345531,
		0.359265,
		0.373216,
		0.387376,
		0.401737,
		0.416289,
		0.431024,
		0.44593,
		0.460998,
		0.476215,
		0.491569,
		0.507048,
		0.522638,
		0.538323,
		0.55409,
		0.569922,
		0.585802,
		0.601712,
		0.617635,
		0.63355,
		0.649438,
		0.665277,
		0.681045,
		0.69672,
		0.712278,
		0.727694,
		0.742944,
		0.758,
		0.772836,
		0.787425,
		0.801739,
		0.815748,
		0.829424,
		0.842738,
		0.855661,
		0.868161,
		0.880212,
		0.891784,
		0.902848,
		0.913379,
		0.92335,
		0.932739,
		0.941522,
		0.949683,
		0.957205,
		0.964076,
		0.970291,
		0.975845,
		0.980744,
		0.984999,
		0.988624,
		0.991646,
		0.994091,
		0.995994,
		0.997388,
		0.998305,
		0.998768,
		0.998789,
		0.998361,
		0.99745,
		0.995988,
		0.993865,
		0.990912,
		0.986903,
		0.98155,
		0.974514,
		0.96542,
		0.953873,
		0.939475,
		0.921848,
		0.900656,
		0.875627,
		0.84658,
		0.813438,
		0.77625,
		0.735198,
		0.690604,
		0.642924,
		0.592741,
		0.540744,
		0.487707,
		0.434461,
		0.381856,
		0.330732,
		0.281881,
		0.236018,
		0.193752,
		0.155564,
		0.121789,
		0.092608,
		0.068049,
		0.047987,
		0.032163,
		0.020197,
		0.011614,
		0.005877,
		0.002411,
		0.000644,
		3.2e-005,
		0.000127,
		0.000511,
		0.001154,
		0.002059,
		0.003228,
		0.004665,
		0.006372,
		0.008351,
		0.010606,
		0.013138,
		0.015951,
		0.019046,
		0.022426,
		0.026093,
		0.030049,
		0.034297,
		0.038838,
		0.043674,
		0.048808,
		0.05424,
		0.059973,
		0.066007,
		0.072345,
		0.078987,
		0.085934,
		0.093187,
		0.100747,
		0.108615,
		0.11679,
		0.125272,
		0.134062,
		0.14316,
		0.152563,
		0.162273,
		0.172287,
		0.182603,
		0.193222,
		0.204139,
		0.215353,
		0.226862,
		0.238661,
		0.250747,
		0.263117,
		0.275766,
		0.28869,
		0.301882,
		0.315337,
		0.32905,
		0.343012,
		0.357216,
		0.371655,
		0.38632,
		0.401201,
		0.416289,
		0.431573,
		0.447041,
		0.462682,
		0.478481,
		0.494427,
		0.510503,
		0.526696,
		0.542987,
		0.559361,
		0.575799,
		0.592281,
		0.608789,
		0.6253,
		0.641793,
		0.658245,
		0.674631,
		0.690927,
		0.707107,
		0.723143,
		0.739008,
		0.754672,
		0.770107,
		0.785281,
		0.800163,
		0.814721,
		0.828924,
		0.842738,
		0.856131,
		0.86907,
		0.881522,
		0.893455,
		0.904839,
		0.915644,
		0.925841,
		0.935405,
		0.944313,
		0.952544,
		0.960083,
		0.966918,
		0.973042,
		0.978449,
		0.983139,
		0.98711,
		0.990359,
		0.992877,
		0.994648,
		0.995644,
		0.995825,
		0.995131,
		0.993485,
		0.990785,
		0.986903,
		0.981686,
		0.974941,
		0.966441,
		0.955914,
		0.943055,
		0.927549,
		0.909081,
		0.887367,
		0.86217,
		0.833327,
		0.800764,
		0.764515,
		0.724736,
		0.681705,
		0.635826,
		0.587617,
		0.537699,
		0.486772,
		0.435589,
		0.38493,
		0.335565,
		0.288229,
		0.24359,
		0.202223,
		0.164591,
		0.131026,
		0.101724,
		0.076739,
		0.055989,
		0.039266,
		0.02625,
		0.016531,
		0.009635,
		0.005051,
		0.002263,
		0.000774,
		0.000143,
		0.000137,
		0.000551,
		0.001245,
		0.002221,
		0.003484,
		0.005035,
		0.006878,
		0.009016,
		0.011451,
		0.014187,
		0.017226,
		0.02057,
		0.024223,
		0.028187,
		0.032464,
		0.037057,
		0.041967,
		0.047197,
		0.052748,
		0.058623,
		0.064824,
		0.07135,
		0.078205,
		0.085389,
		0.092903,
		0.100747,
		0.108924,
		0.117431,
		0.126271,
		0.135442,
		0.144944,
		0.154777,
		0.164939,
		0.175429,
		0.186245,
		0.197385,
		0.208847,
		0.220628,
		0.232725,
		0.245133,
		0.25785,
		0.270869,
		0.284186,
		0.297795,
		0.311689,
		0.325863,
		0.340308,
		0.355015,
		0.369978,
		0.385184,
		0.400625,
		0.416289,
		0.432165,
		0.448238,
		0.464497,
		0.480925,
		0.497509,
		0.51423,
		0.531073,
		0.548017,
		0.565045,
		0.582134,
		0.599263,
		0.61641,
		0.63355,
		0.650658,
		0.667708,
		0.684672,
		0.701521,
		0.718226,
		0.734755,
		0.751077,
		0.767158,
		0.782964,
		0.798462,
		0.813614,
		0.828385,
		0.842738,
		0.856637,
		0.870045,
		0.882926,
		0.895243,
		0.906963,
		0.918052,
		0.928478,
		0.938211,
		0.947222,
		0.955481,
		0.962953,
		0.969601,
		0.97538,
		0.980234,
		0.984094,
		0.986877,
		0.988486,
		0.988808,
		0.987715,
		0.985067,
		0.980718,
		0.974514,
		0.966297,
		0.955914,
		0.943206,
		0.928019,
		0.910191,
		0.889556,
		0.865957,
		0.839261,
		0.809382,
		0.776299,
		0.74007,
		0.700845,
		0.658875,
		0.614514,
		0.568211,
		0.520504,
		0.472002,
		0.423361,
		0.375266,
		0.328396,
		0.283404,
		0.240889,
		0.201372,
		0.165276,
		0.132916,
		0.104482,
		0.080042,
		0.05954,
		0.042807,
		0.029569,
		0.019471,
		0.012091,
		0.006974,
		0.003651,
		0.00167,
		0.000623,
		0.000161,
		1.6e-005,
		0.000148,
		0.000596,
		0.001347,
		0.002404,
		0.003771,
		0.00545,
		0.007447,
		0.009762,
		0.012401,
		0.015366,
		0.01866,
		0.022285,
		0.026246,
		0.030544,
		0.035182,
		0.040162,
		0.045488,
		0.051161,
		0.057184,
		0.063557,
		0.070284,
		0.077365,
		0.084802,
		0.092596,
		0.100747,
		0.109257,
		0.118126,
		0.127354,
		0.13694,
		0.146884,
		0.157186,
		0.167843,
		0.178855,
		0.190218,
		0.201932,
		0.213992,
		0.226396,
		0.239139,
		0.252217,
		0.265625,
		0.279358,
		0.293408,
		0.30777,
		0.322436,
		0.337397,
		0.352645,
		0.368169,
		0.383959,
		0.400003,
		0.416289,
		0.432804,
		0.449532,
		0.466459,
		0.483569,
		0.500843,
		0.518262,
		0.535808,
		0.553458,
		0.571191,
		0.588982,
		0.606807,
		0.62464,
		0.642452,
		0.660215,
		0.677898,
		0.69547,
		0.712898,
		0.730146,
		0.74718,
		0.763962,
		0.780455,
		0.796619,
		0.812415,
		0.827802,
		0.842738,
		0.857183,
		0.871094,
		0.884429,
		0.897143,
		0.909189,
		0.920512,
		0.931051,
		0.940734,
		0.949475,
		0.957169,
		0.963696,
		0.968915,
		0.972665,
		0.974769,
		0.975038,
		0.973275,
		0.969283,
		0.962873,
		0.953873,
		0.942136,
		0.927549,
		0.910034,
		0.889556,
		0.866118,
		0.839755,
		0.810529,
		0.77852,
		0.743835,
		0.706621,
		0.667076,
		0.625459,
		0.582095,
		0.537373,
		0.491749,
		0.445728,
		0.399855,
		0.354699,
		0.310826,
		0.268784,
		0.229082,
		0.192162,
		0.158393,
		0.128047,
		0.101297,
		0.078208,
		0.058738,
		0.04274,
		0.02998,
		0.020142,
		0.012854,
		0.007705,
		0.004271,
		0.002141,
		0.000938,
		0.000339,
		9.1e-005,
		1.4e-005,
		0.000161,
		0.000647,
		0.001462,
		0.00261,
		0.004095,
		0.00592,
		0.008089,
		0.010606,
		0.013475,
		0.016698,
		0.02028,
		0.024223,
		0.028532,
		0.033208,
		0.038254,
		0.043674,
		0.049471,
		0.055645,
		0.0622,
		0.069138,
		0.076461,
		0.084169,
		0.092264,
		0.100747,
		0.10962,
		0.118882,
		0.128533,
		0.138573,
		0.149001,
		0.159817,
		0.171018,
		0.182603,
		0.19457,
		0.206915,
		0.219635,
		0.232725,
		0.246181,
		0.259999,
		0.27417,
		0.28869,
		0.30355,
		0.318742,
		0.334256,
		0.350084,
		0.366214,
		0.382633,
		0.39933,
		0.416289,
		0.433497,
		0.450936,
		0.468589,
		0.486437,
		0.50446,
		0.522638,
		0.540946,
		0.559361,
		0.577857,
		0.596407,
		0.614981,
		0.63355,
		0.652082,
		0.670542,
		0.688896,
		0.707107,
		0.725136,
		0.742944,
		0.760489,
		0.777728,
		0.794617,
		0.811109,
		0.827155,
		0.842698,
		0.857677,
		0.872018,
		0.885637,
		0.898429,
		0.910271,
		0.921018,
		0.930499,
		0.938519,
		0.944861,
		0.949288,
		0.95155,
		0.951392,
		0.948562,
		0.942826,
		0.933976,
		0.921848,
		0.906329,
		0.887367,
		0.864981,
		0.839261,
		0.810366,
		0.77852,
		0.743996,
		0.707107,
		0.668179,
		0.627545,
		0.585544,
		0.542528,
		0.498867,
		0.454953,
		0.411201,
		0.368043,
		0.325921,
		0.285274,
		0.246526,
		0.210073,
		0.176263,
		0.145385,
		0.117656,
		0.09321,
		0.072094,
		0.054265,
		0.039592,
		0.027863,
		0.018798,
		0.012061,
		0.007282,
		0.00408,
		0.002079,
		0.000936,
		0.000355,
		0.000106,
		2.1e-005,
		0.000176,
		0.000705,
		0.001593,
		0.002844,
		0.004462,
		0.006452,
		0.008818,
		0.011564,
		0.014693,
		0.018211,
		0.022121,
		0.026425,
		0.031129,
		0.036235,
		0.041747,
		0.047667,
		0.053998,
		0.060743,
		0.067904,
		0.075484,
		0.083483,
		0.091904,
		0.100747,
		0.110014,
		0.119705,
		0.12982,
		0.140358,
		0.151319,
		0.162702,
		0.174504,
		0.186723,
		0.199356,
		0.212399,
		0.225849,
		0.2397,
		0.253947,
		0.268583,
		0.283601,
		0.298992,
		0.314747,
		0.330857,
		0.34731,
		0.364093,
		0.381194,
		0.398598,
		0.416289,
		0.43425,
		0.452462,
		0.470906,
		0.489559,
		0.5084,
		0.527402,
		0.54654,
		0.565786,
		0.585111,
		0.604481,
		0.623865,
		0.643226,
		0.662527,
		0.681729,
		0.70079,
		0.719666,
		0.738309,
		0.756663,
		0.774666,
		0.792245,
		0.80931,
		0.825756,
		0.841453,
		0.856247,
		0.869955,
		0.882368,
		0.893244,
		0.90232,
		0.909308,
		0.913908,
		0.915816,
		0.914735,
		0.910389,
		0.902538,
		0.890993,
		0.875627,
		0.856393,
		0.833327,
		0.806556,
		0.776299,
		0.742861,
		0.706621,
		0.668022,
		0.627545,
		0.585694,
		0.542966,
		0.499835,
		0.456733,
		0.414057,
		0.372177,
		0.33144,
		0.292176,
		0.254699,
		0.219306,
		0.186264,
		0.155812,
		0.128143,
		0.103399,
		0.081659,
		0.062938,
		0.047176,
		0.034243,
		0.023941,
		0.016015,
		0.01016,
		0.006041,
		0.003312,
		0.001636,
		0.000703,
		0.000248,
		6.4e-005,
		0.000192,
		0.000771,
		0.001742,
		0.003111,
		0.004882,
		0.00706,
		0.00965,
		0.012657,
		0.016085,
		0.019939,
		0.024223,
		0.028941,
		0.034098,
		0.039695,
		0.045739,
		0.05223,
		0.059174,
		0.066571,
		0.074425,
		0.082738,
		0.091512,
		0.100747,
		0.110446,
		0.120607,
		0.131232,
		0.14232,
		0.153869,
		0.165879,
		0.178346,
		0.191269,
		0.204642,
		0.218463,
		0.232725,
		0.247423,
		0.262549,
		0.278096,
		0.294054,
		0.310415,
		0.327165,
		0.344293,
		0.361785,
		0.379627,
		0.397801,
		0.416289,
		0.435073,
		0.45413,
		0.473438,
		0.492972,
		0.512705,
		0.532609,
		0.552654,
		0.572806,
		0.59303,
		0.613289,
		0.633539,
		0.653732,
		0.673812,
		0.69371,
		0.713346,
		0.732617,
		0.751401,
		0.769549,
		0.786881,
		0.803188,
		0.818227,
		0.831725,
		0.84338,
		0.852871,
		0.859861,
		0.864009,
		0.864989,
		0.862495,
		0.856264,
		0.846087,
		0.83183,
		0.813438,
		0.790953,
		0.764515,
		0.734367,
		0.700845,
		0.664376,
		0.625459,
		0.584648,
		0.542528,
		0.499696,
		0.456733,
		0.414182,
		0.372533,
		0.3322,
		0.293526,
		0.256786,
		0.222209,
		0.189978,
		0.160248,
		0.13314,
		0.108744,
		0.087112,
		0.068256,
		0.052138,
		0.038672,
		0.027714,
		0.01907,
		0.012498,
		0.007717,
		0.004423,
		0.002303,
		0.001055,
		0.000402,
		0.000116,
		2e-005,
		0.000211,
		0.000846,
		0.001913,
		0.003417,
		0.005363,
		0.007758,
		0.010606,
		0.013913,
		0.017685,
		0.021925,
		0.02664,
		0.031834,
		0.037511,
		0.043674,
		0.05033,
		0.057479,
		0.065127,
		0.073275,
		0.081927,
		0.091084,
		0.100747,
		0.110919,
		0.121599,
		0.132788,
		0.144484,
		0.156687,
		0.169394,
		0.182603,
		0.19631,
		0.210511,
		0.2252,
		0.24037,
		0.256014,
		0.272124,
		0.28869,
		0.3057,
		0.323142,
		0.341002,
		0.359265,
		0.377913,
		0.396928,
		0.416289,
		0.435974,
		0.455958,
		0.476214,
		0.496713,
		0.51742,
		0.538297,
		0.559298,
		0.580364,
		0.601425,
		0.622395,
		0.643164,
		0.663597,
		0.68353,
		0.702768,
		0.721083,
		0.73821,
		0.753855,
		0.767694,
		0.779382,
		0.788563,
		0.794878,
		0.797981,
		0.797554,
		0.793321,
		0.785067,
		0.77265,
		0.756013,
		0.735198,
		0.710347,
		0.681705,
		0.649616,
		0.614514,
		0.57691,
		0.537373,
		0.496513,
		0.454953,
		0.413311,
		0.372177,
		0.33209,
		0.293526,
		0.256879,
		0.222461,
		0.190497,
		0.161129,
		0.134436,
		0.110452,
		0.089172,
		0.070559,
		0.054548,
		0.041042,
		0.02991,
		0.020985,
		0.014064,
		0.00891,
		0.005261,
		0.002835,
		0.001351,
		0.00054,
		0.000165,
		3.1e-005,
		0.000232,
		0.000934,
		0.002111,
		0.003771,
		0.00592,
		0.008564,
		0.011711,
		0.015366,
		0.019535,
		0.024223,
		0.029437,
		0.035182,
		0.041461,
		0.048281,
		0.055645,
		0.063557,
		0.072021,
		0.081039,
		0.090614,
		0.100747,
		0.111441,
		0.122695,
		0.13451,
		0.146884,
		0.159817,
		0.173305,
		0.187344,
		0.201932,
		0.217061,
		0.232725,
		0.248916,
		0.265625,
		0.282841,
		0.300551,
		0.318742,
		0.337397,
		0.3565,
		0.376029,
		0.395962,
		0.416269,
		0.436917,
		0.457861,
		0.479047,
		0.500403,
		0.52184,
		0.543246,
		0.564478,
		0.585368,
		0.605712,
		0.62527,
		0.643772,
		0.660911,
		0.676358,
		0.689759,
		0.700751,
		0.708972,
		0.714073,
		0.715733,
		0.713679,
		0.707694,
		0.697635,
		0.683445,
		0.665163,
		0.642924,
		0.616965,
		0.587617,
		0.555299,
		0.520504,
		0.483784,
		0.445728,
		0.406945,
		0.368043,
		0.329606,
		0.292176,
		0.256238,
		0.222209,
		0.190422,
		0.161129,
		0.134495,
		0.110603,
		0.089464,
		0.071024,
		0.055184,
		0.041813,
		0.030755,
		0.021831,
		0.014841,
		0.009563,
		0.00576,
		0.003177,
		0.001555,
		0.000642,
		0.000203,
		4e-005,
		0.000257,
		0.001035,
		0.002341,
		0.004182,
		0.006567,
		0.009504,
		0.012998,
		0.017058,
		0.02169,
		0.026902,
		0.032698,
		0.039085,
		0.046069,
		0.053654,
		0.061846,
		0.070648,
		0.080064,
		0.090096,
		0.100747,
		0.112019,
		0.123912,
		0.136427,
		0.149561,
		0.163312,
		0.177679,
		0.192655,
		0.208236,
		0.224414,
		0.241181,
		0.258526,
		0.276435,
		0.294891,
		0.313873,
		0.333351,
		0.353287,
		0.373628,
		0.394307,
		0.415238,
		0.436307,
		0.457378,
		0.478279,
		0.498808,
		0.51873,
		0.537774,
		0.555639,
		0.572,
		0.58651,
		0.598814,
		0.608558,
		0.615402,
		0.619035,
		0.61919,
		0.615657,
		0.608295,
		0.597047,
		0.581942,
		0.563102,
		0.540744,
		0.515172,
		0.486772,
		0.455999,
		0.423361,
		0.389406,
		0.354699,
		0.319805,
		0.285274,
		0.251619,
		0.219306,
		0.188738,
		0.160248,
		0.134094,
		0.110452,
		0.089421,
		0.071024,
		0.055213,
		0.041882,
		0.030877,
		0.022005,
		0.01505,
		0.00978,
		0.005956,
		0.003331,
		0.001659,
		0.000698,
		0.000226,
		4.5e-005,
		0.000287,
		0.001154,
		0.00261,
		0.004665,
		0.007327,
		0.010606,
		0.014509,
		0.019046,
		0.024223,
		0.030049,
		0.036531,
		0.043674,
		0.051487,
		0.059973,
		0.069138,
		0.078987,
		0.089522,
		0.100747,
		0.112664,
		0.125272,
		0.138572,
		0.152562,
		0.167239,
		0.182595,
		0.198621,
		0.215301,
		0.232614,
		0.250527,
		0.268996,
		0.287961,
		0.30734,
		0.327031,
		0.346903,
		0.366798,
		0.386524,
		0.405858,
		0.424544,
		0.442298,
		0.458811,
		0.473757,
		0.486799,
		0.497602,
		0.505845,
		0.511233,
		0.513512,
		0.512479,
		0.507996,
		0.5,
		0.488506,
		0.473615,
		0.455512,
		0.434461,
		0.4108,
		0.38493,
		0.3573,
		0.328396,
		0.298721,
		0.268784,
		0.239079,
		0.210073,
		0.182192,
		0.155812,
		0.131247,
		0.108744,
		0.08848,
		0.070559,
		0.055015,
		0.041813,
		0.030859,
		0.022005,
		0.015059,
		0.0098,
		0.005985,
		0.003363,
		0.001687,
		0.000718,
		0.000235,
		4.8e-005,
		0.000322,
		0.001294,
		0.002929,
		0.005236,
		0.008227,
		0.011912,
		0.0163,
		0.021402,
		0.027226,
		0.033782,
		0.041077,
		0.049119,
		0.057915,
		0.067471,
		0.077791,
		0.088879,
		0.100736,
		0.113359,
		0.126741,
		0.140868,
		0.155717,
		0.171255,
		0.187433,
		0.204182,
		0.221416,
		0.239021,
		0.256856,
		0.274752,
		0.29251,
		0.309899,
		0.326664,
		0.342523,
		0.35718,
		0.370327,
		0.381656,
		0.390871,
		0.397697,
		0.401894,
		0.403267,
		0.401677,
		0.397048,
		0.389377,
		0.378733,
		0.365257,
		0.349164,
		0.330732,
		0.310294,
		0.288229,
		0.26495,
		0.240889,
		0.216483,
		0.192162,
		0.168337,
		0.145385,
		0.123644,
		0.103399,
		0.08488,
		0.068256,
		0.05363,
		0.041042,
		0.030469,
		0.021831,
		0.014994,
		0.00978,
		0.005981,
		0.003363,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.000363,
		0.001462,
		0.00331,
		0.00592,
		0.009304,
		0.013474,
		0.018444,
		0.024222,
		0.03082,
		0.038245,
		0.046503,
		0.055596,
		0.065521,
		0.076267,
		0.087814,
		0.100129,
		0.113164,
		0.126853,
		0.141106,
		0.155812,
		0.170831,
		0.185998,
		0.201119,
		0.215973,
		0.23032,
		0.243899,
		0.256439,
		0.267665,
		0.277308,
		0.285117,
		0.290865,
		0.294361,
		0.295462,
		0.294076,
		0.29017,
		0.283774,
		0.274981,
		0.263942,
		0.250868,
		0.236018,
		0.219692,
		0.202223,
		0.183964,
		0.165276,
		0.146521,
		0.128047,
		0.110179,
		0.09321,
		0.077394,
		0.062938,
		0.049997,
		0.038672,
		0.029006,
		0.020985,
		0.014543,
		0.009563,
		0.005889,
		0.003331,
		0.001679,
		0.000718,
		0.000237,
		4.9e-005,
		0.000414,
		0.001665,
		0.003769,
		0.00674,
		0.01059,
		0.015327,
		0.020955,
		0.027471,
		0.034862,
		0.043104,
		0.052157,
		0.061965,
		0.07245,
		0.083512,
		0.095026,
		0.106843,
		0.118789,
		0.130667,
		0.142261,
		0.153342,
		0.163672,
		0.173013,
		0.181135,
		0.187825,
		0.192897,
		0.196197,
		0.197612,
		0.197077,
		0.194576,
		0.190145,
		0.18387,
		0.175889,
		0.166381,
		0.155564,
		0.143689,
		0.131026,
		0.11786,
		0.104482,
		0.091174,
		0.078208,
		0.065832,
		0.054265,
		0.043688,
		0.034243,
		0.026022,
		0.01907,
		0.013384,
		0.00891,
		0.005553,
		0.003177,
		0.001619,
		0.000698,
		0.000232,
		4.8e-005,
		0.000471,
		0.001888,
		0.004255,
		0.007563,
		0.011792,
		0.016906,
		0.022848,
		0.029542,
		0.036891,
		0.044773,
		0.053042,
		0.061533,
		0.070064,
		0.078437,
		0.086449,
		0.093895,
		0.100576,
		0.106308,
		0.110928,
		0.1143,
		0.116324,
		0.116937,
		0.116117,
		0.113883,
		0.110298,
		0.105461,
		0.099509,
		0.092608,
		0.084949,
		0.076739,
		0.068196,
		0.05954,
		0.050988,
		0.04274,
		0.03498,
		0.027863,
		0.021513,
		0.016015,
		0.011415,
		0.007717,
		0.004882,
		0.002835,
		0.001466,
		0.000642,
		0.000216,
		4.5e-005,
		0.000502,
		0.001976,
		0.004361,
		0.007571,
		0.011499,
		0.016015,
		0.020971,
		0.026203,
		0.031536,
		0.036791,
		0.04179,
		0.046363,
		0.050357,
		0.053638,
		0.056098,
		0.057659,
		0.058278,
		0.057942,
		0.056675,
		0.054533,
		0.0516,
		0.047987,
		0.043827,
		0.039266,
		0.03446,
		0.029569,
		0.024749,
		0.020142,
		0.015878,
		0.012061,
		0.008767,
		0.006041,
		0.003894,
		0.002303,
		0.001213,
		0.00054,
		0.000185,
		4e-005,
		0.000432,
		0.001638,
		0.003467,
		0.005759,
		0.008346,
		0.011064,
		0.013754,
		0.01627,
		0.018485,
		0.020295,
		0.021621,
		0.022412,
		0.022647,
		0.02233,
		0.021495,
		0.020197,
		0.018512,
		0.016531,
		0.014355,
		0.012091,
		0.009842,
		0.007705,
		0.005762,
		0.00408,
		0.002699,
		0.001636,
		0.000882,
		0.000402,
		0.000141,
		3.1e-005,
		0.000256,
		0.000913,
		0.001814,
		0.002816,
		0.003798,
		0.004664,
		0.005339,
		0.005777,
		0.005956,
		0.005877,
		0.005563,
		0.005051,
		0.004394,
		0.003651,
		0.00288,
		0.002141,
		0.001481,
		0.000936,
		0.000524,
		0.000248,
		9e-005,
		2e-005,
		8.4e-005,
		0.000268,
		0.000473,
		0.000644,
		0.000748,
		0.000774,
		0.000727,
		0.000623,
		0.000485,
		0.000339,
		0.000207,
		0.000106,
		4.1e-005,
		1.6e-005,
		1.8e-005,
		1.4e-005,
		1.4e-005,
		1.8e-005,
		1.6e-005,
		4.1e-005,
		0.000106,
		0.000207,
		0.000339,
		0.000485,
		0.000623,
		0.000727,
		0.000774,
		0.000748,
		0.000644,
		0.000473,
		0.000268,
		8.4e-005,
		2e-005,
		9e-005,
		0.000248,
		0.000524,
		0.000936,
		0.001481,
		0.002141,
		0.00288,
		0.003651,
		0.004394,
		0.005051,
		0.005563,
		0.005877,
		0.005956,
		0.005777,
		0.005339,
		0.004664,
		0.003798,
		0.002816,
		0.001814,
		0.000913,
		0.000256,
		3.1e-005,
		0.000141,
		0.000402,
		0.000882,
		0.001636,
		0.002699,
		0.00408,
		0.005762,
		0.007705,
		0.009842,
		0.012091,
		0.014355,
		0.016531,
		0.018512,
		0.020197,
		0.021495,
		0.02233,
		0.022647,
		0.022412,
		0.021621,
		0.020295,
		0.018485,
		0.01627,
		0.013754,
		0.011064,
		0.008346,
		0.005759,
		0.003467,
		0.001638,
		0.000432,
		4e-005,
		0.000185,
		0.00054,
		0.001213,
		0.002303,
		0.003894,
		0.006041,
		0.008767,
		0.012061,
		0.015878,
		0.020142,
		0.024749,
		0.029569,
		0.03446,
		0.039266,
		0.043827,
		0.047987,
		0.0516,
		0.054533,
		0.056675,
		0.057942,
		0.058278,
		0.057659,
		0.056098,
		0.053638,
		0.050357,
		0.046363,
		0.04179,
		0.036791,
		0.031536,
		0.026203,
		0.020971,
		0.016015,
		0.011499,
		0.007571,
		0.004361,
		0.001976,
		0.000502,
		4.5e-005,
		0.000216,
		0.000642,
		0.001466,
		0.002835,
		0.004882,
		0.007717,
		0.011415,
		0.016015,
		0.021513,
		0.027863,
		0.03498,
		0.04274,
		0.050988,
		0.05954,
		0.068196,
		0.076739,
		0.084949,
		0.092608,
		0.099509,
		0.105461,
		0.110298,
		0.113883,
		0.116117,
		0.116937,
		0.116324,
		0.1143,
		0.110928,
		0.106308,
		0.100576,
		0.093895,
		0.086449,
		0.078437,
		0.070064,
		0.061533,
		0.053042,
		0.044773,
		0.036891,
		0.029542,
		0.022848,
		0.016906,
		0.011792,
		0.007563,
		0.004255,
		0.001888,
		0.000471,
		4.8e-005,
		0.000232,
		0.000698,
		0.001619,
		0.003177,
		0.005553,
		0.00891,
		0.013384,
		0.01907,
		0.026022,
		0.034243,
		0.043688,
		0.054265,
		0.065832,
		0.078208,
		0.091174,
		0.104482,
		0.11786,
		0.131026,
		0.143689,
		0.155564,
		0.166381,
		0.175889,
		0.18387,
		0.190145,
		0.194576,
		0.197077,
		0.197612,
		0.196197,
		0.192897,
		0.187825,
		0.181135,
		0.173013,
		0.163672,
		0.153342,
		0.142261,
		0.130667,
		0.118789,
		0.106843,
		0.095026,
		0.083512,
		0.07245,
		0.061965,
		0.052157,
		0.043104,
		0.034862,
		0.027471,
		0.020955,
		0.015327,
		0.01059,
		0.00674,
		0.003769,
		0.001665,
		0.000414,
		4.9e-005,
		0.000237,
		0.000718,
		0.001679,
		0.003331,
		0.005889,
		0.009563,
		0.014543,
		0.020985,
		0.029006,
		0.038672,
		0.049997,
		0.062938,
		0.077394,
		0.09321,
		0.110179,
		0.128047,
		0.146521,
		0.165276,
		0.183964,
		0.202223,
		0.219692,
		0.236018,
		0.250868,
		0.263942,
		0.274981,
		0.283774,
		0.29017,
		0.294076,
		0.295462,
		0.294361,
		0.290865,
		0.285117,
		0.277308,
		0.267665,
		0.256439,
		0.243899,
		0.23032,
		0.215973,
		0.201119,
		0.185998,
		0.170831,
		0.155812,
		0.141106,
		0.126853,
		0.113164,
		0.100129,
		0.087814,
		0.076267,
		0.065521,
		0.055596,
		0.046503,
		0.038245,
		0.03082,
		0.024222,
		0.018444,
		0.013474,
		0.009304,
		0.00592,
		0.00331,
		0.001462,
		0.000363,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003363,
		0.005981,
		0.00978,
		0.014994,
		0.021831,
		0.030469,
		0.041042,
		0.05363,
		0.068256,
		0.08488,
		0.103399,
		0.123644,
		0.145385,
		0.168337,
		0.192162,
		0.216483,
		0.240889,
		0.26495,
		0.288229,
		0.310294,
		0.330732,
		0.349164,
		0.365257,
		0.378733,
		0.389377,
		0.397048,
		0.401677,
		0.403267,
		0.401894,
		0.397697,
		0.390871,
		0.381656,
		0.370327,
		0.35718,
		0.342523,
		0.326664,
		0.309899,
		0.29251,
		0.274752,
		0.256856,
		0.239021,
		0.221416,
		0.204182,
		0.187433,
		0.171255,
		0.155717,
		0.140868,
		0.126741,
		0.113359,
		0.100736,
		0.088879,
		0.077791,
		0.067471,
		0.057915,
		0.049119,
		0.041077,
		0.033782,
		0.027226,
		0.021402,
		0.0163,
		0.011912,
		0.008227,
		0.005236,
		0.002929,
		0.001294,
		0.000322,
		4.8e-005,
		0.000235,
		0.000718,
		0.001687,
		0.003363,
		0.005985,
		0.0098,
		0.015059,
		0.022005,
		0.030859,
		0.041813,
		0.055015,
		0.070559,
		0.08848,
		0.108744,
		0.131247,
		0.155812,
		0.182192,
		0.210073,
		0.239079,
		0.268784,
		0.298721,
		0.328396,
		0.3573,
		0.38493,
		0.4108,
		0.434461,
		0.455512,
		0.473615,
		0.488506,
		0.5,
		0.507996,
		0.512479,
		0.513512,
		0.511233,
		0.505845,
		0.497602,
		0.486799,
		0.473757,
		0.458811,
		0.442298,
		0.424544,
		0.405858,
		0.386524,
		0.366798,
		0.346903,
		0.327031,
		0.30734,
		0.287961,
		0.268996,
		0.250527,
		0.232614,
		0.215301,
		0.198621,
		0.182595,
		0.167239,
		0.152562,
		0.138572,
		0.125272,
		0.112664,
		0.100747,
		0.089522,
		0.078987,
		0.069138,
		0.059973,
		0.051487,
		0.043674,
		0.036531,
		0.030049,
		0.024223,
		0.019046,
		0.014509,
		0.010606,
		0.007327,
		0.004665,
		0.00261,
		0.001154,
		0.000287,
		4.5e-005,
		0.000226,
		0.000698,
		0.001659,
		0.003331,
		0.005956,
		0.00978,
		0.01505,
		0.022005,
		0.030877,
		0.041882,
		0.055213,
		0.071024,
		0.089421,
		0.110452,
		0.134094,
		0.160248,
		0.188738,
		0.219306,
		0.251619,
		0.285274,
		0.319805,
		0.354699,
		0.389406,
		0.423361,
		0.455999,
		0.486772,
		0.515172,
		0.540744,
		0.563102,
		0.581942,
		0.597047,
		0.608295,
		0.615657,
		0.61919,
		0.619035,
		0.615402,
		0.608558,
		0.598814,
		0.58651,
		0.572,
		0.555639,
		0.537774,
		0.51873,
		0.498808,
		0.478279,
		0.457378,
		0.436307,
		0.415238,
		0.394307,
		0.373628,
		0.353287,
		0.333351,
		0.313873,
		0.294891,
		0.276435,
		0.258526,
		0.241181,
		0.224414,
		0.208236,
		0.192655,
		0.177679,
		0.163312,
		0.149561,
		0.136427,
		0.123912,
		0.112019,
		0.100747,
		0.090096,
		0.080064,
		0.070648,
		0.061846,
		0.053654,
		0.046069,
		0.039085,
		0.032698,
		0.026902,
		0.02169,
		0.017058,
		0.012998,
		0.009504,
		0.006567,
		0.004182,
		0.002341,
		0.001035,
		0.000257,
		4e-005,
		0.000203,
		0.000642,
		0.001555,
		0.003177,
		0.00576,
		0.009563,
		0.014841,
		0.021831,
		0.030755,
		0.041813,
		0.055184,
		0.071024,
		0.089464,
		0.110603,
		0.134495,
		0.161129,
		0.190422,
		0.222209,
		0.256238,
		0.292176,
		0.329606,
		0.368043,
		0.406945,
		0.445728,
		0.483784,
		0.520504,
		0.555299,
		0.587617,
		0.616965,
		0.642924,
		0.665163,
		0.683445,
		0.697635,
		0.707694,
		0.713679,
		0.715733,
		0.714073,
		0.708972,
		0.700751,
		0.689759,
		0.676358,
		0.660911,
		0.643772,
		0.62527,
		0.605712,
		0.585368,
		0.564478,
		0.543246,
		0.52184,
		0.500403,
		0.479047,
		0.457861,
		0.436917,
		0.416269,
		0.395962,
		0.376029,
		0.3565,
		0.337397,
		0.318742,
		0.300551,
		0.282841,
		0.265625,
		0.248916,
		0.232725,
		0.217061,
		0.201932,
		0.187344,
		0.173305,
		0.159817,
		0.146884,
		0.13451,
		0.122695,
		0.111441,
		0.100747,
		0.090614,
		0.081039,
		0.072021,
		0.063557,
		0.055645,
		0.048281,
		0.041461,
		0.035182,
		0.029437,
		0.024223,
		0.019535,
		0.015366,
		0.011711,
		0.008564,
		0.00592,
		0.003771,
		0.002111,
		0.000934,
		0.000232,
		3.1e-005,
		0.000165,
		0.00054,
		0.001351,
		0.002835,
		0.005261,
		0.00891,
		0.014064,
		0.020985,
		0.02991,
		0.041042,
		0.054548,
		0.070559,
		0.089172,
		0.110452,
		0.134436,
		0.161129,
		0.190497,
		0.222461,
		0.256879,
		0.293526,
		0.33209,
		0.372177,
		0.413311,
		0.454953,
		0.496513,
		0.537373,
		0.57691,
		0.614514,
		0.649616,
		0.681705,
		0.710347,
		0.735198,
		0.756013,
		0.77265,
		0.785067,
		0.793321,
		0.797554,
		0.797981,
		0.794878,
		0.788563,
		0.779382,
		0.767694,
		0.753855,
		0.73821,
		0.721083,
		0.702768,
		0.68353,
		0.663597,
		0.643164,
		0.622395,
		0.601425,
		0.580364,
		0.559298,
		0.538297,
		0.51742,
		0.496713,
		0.476214,
		0.455958,
		0.435974,
		0.416289,
		0.396928,
		0.377913,
		0.359265,
		0.341002,
		0.323142,
		0.3057,
		0.28869,
		0.272124,
		0.256014,
		0.24037,
		0.2252,
		0.210511,
		0.19631,
		0.182603,
		0.169394,
		0.156687,
		0.144484,
		0.132788,
		0.121599,
		0.110919,
		0.100747,
		0.091084,
		0.081927,
		0.073275,
		0.065127,
		0.057479,
		0.05033,
		0.043674,
		0.037511,
		0.031834,
		0.02664,
		0.021925,
		0.017685,
		0.013913,
		0.010606,
		0.007758,
		0.005363,
		0.003417,
		0.001913,
		0.000846,
		0.000211,
		2e-005,
		0.000116,
		0.000402,
		0.001055,
		0.002303,
		0.004423,
		0.007717,
		0.012498,
		0.01907,
		0.027714,
		0.038672,
		0.052138,
		0.068256,
		0.087112,
		0.108744,
		0.13314,
		0.160248,
		0.189978,
		0.222209,
		0.256786,
		0.293526,
		0.3322,
		0.372533,
		0.414182,
		0.456733,
		0.499696,
		0.542528,
		0.584648,
		0.625459,
		0.664376,
		0.700845,
		0.734367,
		0.764515,
		0.790953,
		0.813438,
		0.83183,
		0.846087,
		0.856264,
		0.862495,
		0.864989,
		0.864009,
		0.859861,
		0.852871,
		0.84338,
		0.831725,
		0.818227,
		0.803188,
		0.786881,
		0.769549,
		0.751401,
		0.732617,
		0.713346,
		0.69371,
		0.673812,
		0.653732,
		0.633539,
		0.613289,
		0.59303,
		0.572806,
		0.552654,
		0.532609,
		0.512705,
		0.492972,
		0.473438,
		0.45413,
		0.435073,
		0.416289,
		0.397801,
		0.379627,
		0.361785,
		0.344293,
		0.327165,
		0.310415,
		0.294054,
		0.278096,
		0.262549,
		0.247423,
		0.232725,
		0.218463,
		0.204642,
		0.191269,
		0.178346,
		0.165879,
		0.153869,
		0.14232,
		0.131232,
		0.120607,
		0.110446,
		0.100747,
		0.091512,
		0.082738,
		0.074425,
		0.066571,
		0.059174,
		0.05223,
		0.045739,
		0.039695,
		0.034098,
		0.028941,
		0.024223,
		0.019939,
		0.016085,
		0.012657,
		0.00965,
		0.00706,
		0.004882,
		0.003111,
		0.001742,
		0.000771,
		0.000192,
		6.4e-005,
		0.000248,
		0.000703,
		0.001636,
		0.003312,
		0.006041,
		0.01016,
		0.016015,
		0.023941,
		0.034243,
		0.047176,
		0.062938,
		0.081659,
		0.103399,
		0.128143,
		0.155812,
		0.186264,
		0.219306,
		0.254699,
		0.292176,
		0.33144,
		0.372177,
		0.414057,
		0.456733,
		0.499835,
		0.542966,
		0.585694,
		0.627545,
		0.668022,
		0.706621,
		0.742861,
		0.776299,
		0.806556,
		0.833327,
		0.856393,
		0.875627,
		0.890993,
		0.902538,
		0.910389,
		0.914735,
		0.915816,
		0.913908,
		0.909308,
		0.90232,
		0.893244,
		0.882368,
		0.869955,
		0.856247,
		0.841453,
		0.825756,
		0.80931,
		0.792245,
		0.774666,
		0.756663,
		0.738309,
		0.719666,
		0.70079,
		0.681729,
		0.662527,
		0.643226,
		0.623865,
		0.604481,
		0.585111,
		0.565786,
		0.54654,
		0.527402,
		0.5084,
		0.489559,
		0.470906,
		0.452462,
		0.43425,
		0.416289,
		0.398598,
		0.381194,
		0.364093,
		0.34731,
		0.330857,
		0.314747,
		0.298992,
		0.283601,
		0.268583,
		0.253947,
		0.2397,
		0.225849,
		0.212399,
		0.199356,
		0.186723,
		0.174504,
		0.162702,
		0.151319,
		0.140358,
		0.12982,
		0.119705,
		0.110014,
		0.100747,
		0.091904,
		0.083483,
		0.075484,
		0.067904,
		0.060743,
		0.053998,
		0.047667,
		0.041747,
		0.036235,
		0.031129,
		0.026425,
		0.022121,
		0.018211,
		0.014693,
		0.011564,
		0.008818,
		0.006452,
		0.004462,
		0.002844,
		0.001593,
		0.000705,
		0.000176,
		2.1e-005,
		0.000106,
		0.000355,
		0.000936,
		0.002079,
		0.00408,
		0.007282,
		0.012061,
		0.018798,
		0.027863,
		0.039592,
		0.054265,
		0.072094,
		0.09321,
		0.117656,
		0.145385,
		0.176263,
		0.210073,
		0.246526,
		0.285274,
		0.325921,
		0.368043,
		0.411201,
		0.454953,
		0.498867,
		0.542528,
		0.585544,
		0.627545,
		0.668179,
		0.707107,
		0.743996,
		0.77852,
		0.810366,
		0.839261,
		0.864981,
		0.887367,
		0.906329,
		0.921848,
		0.933976,
		0.942826,
		0.948562,
		0.951392,
		0.95155,
		0.949288,
		0.944861,
		0.938519,
		0.930499,
		0.921018,
		0.910271,
		0.898429,
		0.885637,
		0.872018,
		0.857677,
		0.842698,
		0.827155,
		0.811109,
		0.794617,
		0.777728,
		0.760489,
		0.742944,
		0.725136,
		0.707107,
		0.688896,
		0.670542,
		0.652082,
		0.63355,
		0.614981,
		0.596407,
		0.577857,
		0.559361,
		0.540946,
		0.522638,
		0.50446,
		0.486437,
		0.468589,
		0.450936,
		0.433497,
		0.416289,
		0.39933,
		0.382633,
		0.366214,
		0.350084,
		0.334256,
		0.318742,
		0.30355,
		0.28869,
		0.27417,
		0.259999,
		0.246181,
		0.232725,
		0.219635,
		0.206915,
		0.19457,
		0.182603,
		0.171018,
		0.159817,
		0.149001,
		0.138573,
		0.128533,
		0.118882,
		0.10962,
		0.100747,
		0.092264,
		0.084169,
		0.076461,
		0.069138,
		0.0622,
		0.055645,
		0.049471,
		0.043674,
		0.038254,
		0.033208,
		0.028532,
		0.024223,
		0.02028,
		0.016698,
		0.013475,
		0.010606,
		0.008089,
		0.00592,
		0.004095,
		0.00261,
		0.001462,
		0.000647,
		0.000161,
		1.4e-005,
		9.1e-005,
		0.000339,
		0.000938,
		0.002141,
		0.004271,
		0.007705,
		0.012854,
		0.020142,
		0.02998,
		0.04274,
		0.058738,
		0.078208,
		0.101297,
		0.128047,
		0.158393,
		0.192162,
		0.229082,
		0.268784,
		0.310826,
		0.354699,
		0.399855,
		0.445728,
		0.491749,
		0.537373,
		0.582095,
		0.625459,
		0.667076,
		0.706621,
		0.743835,
		0.77852,
		0.810529,
		0.839755,
		0.866118,
		0.889556,
		0.910034,
		0.927549,
		0.942136,
		0.953873,
		0.962873,
		0.969283,
		0.973275,
		0.975038,
		0.974769,
		0.972665,
		0.968915,
		0.963696,
		0.957169,
		0.949475,
		0.940734,
		0.931051,
		0.920512,
		0.909189,
		0.897143,
		0.884429,
		0.871094,
		0.857183,
		0.842738,
		0.827802,
		0.812415,
		0.796619,
		0.780455,
		0.763962,
		0.74718,
		0.730146,
		0.712898,
		0.69547,
		0.677898,
		0.660215,
		0.642452,
		0.62464,
		0.606807,
		0.588982,
		0.571191,
		0.553458,
		0.535808,
		0.518262,
		0.500843,
		0.483569,
		0.466459,
		0.449532,
		0.432804,
		0.416289,
		0.400003,
		0.383959,
		0.368169,
		0.352645,
		0.337397,
		0.322436,
		0.30777,
		0.293408,
		0.279358,
		0.265625,
		0.252217,
		0.239139,
		0.226396,
		0.213992,
		0.201932,
		0.190218,
		0.178855,
		0.167843,
		0.157186,
		0.146884,
		0.13694,
		0.127354,
		0.118126,
		0.109257,
		0.100747,
		0.092596,
		0.084802,
		0.077365,
		0.070284,
		0.063557,
		0.057184,
		0.051161,
		0.045488,
		0.040162,
		0.035182,
		0.030544,
		0.026246,
		0.022285,
		0.01866,
		0.015366,
		0.012401,
		0.009762,
		0.007447,
		0.00545,
		0.003771,
		0.002404,
		0.001347,
		0.000596,
		0.000148,
		1.6e-005,
		0.000161,
		0.000623,
		0.00167,
		0.003651,
		0.006974,
		0.012091,
		0.019471,
		0.029569,
		0.042807,
		0.05954,
		0.080042,
		0.104482,
		0.132916,
		0.165276,
		0.201372,
		0.240889,
		0.283404,
		0.328396,
		0.375266,
		0.423361,
		0.472002,
		0.520504,
		0.568211,
		0.614514,
		0.658875,
		0.700845,
		0.74007,
		0.776299,
		0.809382,
		0.839261,
		0.865957,
		0.889556,
		0.910191,
		0.928019,
		0.943206,
		0.955914,
		0.966297,
		0.974514,
		0.980718,
		0.985067,
		0.987715,
		0.988808,
		0.988486,
		0.986877,
		0.984094,
		0.980234,
		0.97538,
		0.969601,
		0.962953,
		0.955481,
		0.947222,
		0.938211,
		0.928478,
		0.918052,
		0.906963,
		0.895243,
		0.882926,
		0.870045,
		0.856637,
		0.842738,
		0.828385,
		0.813614,
		0.798462,
		0.782964,
		0.767158,
		0.751077,
		0.734755,
		0.718226,
		0.701521,
		0.684672,
		0.667708,
		0.650658,
		0.63355,
		0.61641,
		0.599263,
		0.582134,
		0.565045,
		0.548017,
		0.531073,
		0.51423,
		0.497509,
		0.480925,
		0.464497,
		0.448238,
		0.432165,
		0.416289,
		0.400625,
		0.385184,
		0.369978,
		0.355015,
		0.340308,
		0.325863,
		0.311689,
		0.297795,
		0.284186,
		0.270869,
		0.25785,
		0.245133,
		0.232725,
		0.220628,
		0.208847,
		0.197385,
		0.186245,
		0.175429,
		0.164939,
		0.154777,
		0.144944,
		0.135442,
		0.126271,
		0.117431,
		0.108924,
		0.100747,
		0.092903,
		0.085389,
		0.078205,
		0.07135,
		0.064824,
		0.058623,
		0.052748,
		0.047197,
		0.041967,
		0.037057,
		0.032464,
		0.028187,
		0.024223,
		0.02057,
		0.017226,
		0.014187,
		0.011451,
		0.009016,
		0.006878,
		0.005035,
		0.003484,
		0.002221,
		0.001245,
		0.000551,
		0.000137,
		0.000143,
		0.000774,
		0.002263,
		0.005051,
		0.009635,
		0.016531,
		0.02625,
		0.039266,
		0.055989,
		0.076739,
		0.101724,
		0.131026,
		0.164591,
		0.202223,
		0.24359,
		0.288229,
		0.335565,
		0.38493,
		0.435589,
		0.486772,
		0.537699,
		0.587617,
		0.635826,
		0.681705,
		0.724736,
		0.764515,
		0.800764,
		0.833327,
		0.86217,
		0.887367,
		0.909081,
		0.927549,
		0.943055,
		0.955914,
		0.966441,
		0.974941,
		0.981686,
		0.986903,
		0.990785,
		0.993485,
		0.995131,
		0.995825,
		0.995644,
		0.994648,
		0.992877,
		0.990359,
		0.98711,
		0.983139,
		0.978449,
		0.973042,
		0.966918,
		0.960083,
		0.952544,
		0.944313,
		0.935405,
		0.925841,
		0.915644,
		0.904839,
		0.893455,
		0.881522,
		0.86907,
		0.856131,
		0.842738,
		0.828924,
		0.814721,
		0.800163,
		0.785281,
		0.770107,
		0.754672,
		0.739008,
		0.723143,
		0.707107,
		0.690927,
		0.674631,
		0.658245,
		0.641793,
		0.6253,
		0.608789,
		0.592281,
		0.575799,
		0.559361,
		0.542987,
		0.526696,
		0.510503,
		0.494427,
		0.478481,
		0.462682,
		0.447041,
		0.431573,
		0.416289,
		0.401201,
		0.38632,
		0.371655,
		0.357216,
		0.343012,
		0.32905,
		0.315337,
		0.301882,
		0.28869,
		0.275766,
		0.263117,
		0.250747,
		0.238661,
		0.226862,
		0.215353,
		0.204139,
		0.193222,
		0.182603,
		0.172287,
		0.162273,
		0.152563,
		0.14316,
		0.134062,
		0.125272,
		0.11679,
		0.108615,
		0.100747,
		0.093187,
		0.085934,
		0.078987,
		0.072345,
		0.066007,
		0.059973,
		0.05424,
		0.048808,
		0.043674,
		0.038838,
		0.034297,
		0.030049,
		0.026093,
		0.022426,
		0.019046,
		0.015951,
		0.013138,
		0.010606,
		0.008351,
		0.006372,
		0.004665,
		0.003228,
		0.002059,
		0.001154,
		0.000511,
		0.000127,
		3.2e-005,
		0.000644,
		0.002411,
		0.005877,
		0.011614,
		0.020197,
		0.032163,
		0.047987,
		0.068049,
		0.092608,
		0.121789,
		0.155564,
		0.193752,
		0.236018,
		0.281881,
		0.330732,
		0.381856,
		0.434461,
		0.487707,
		0.540744,
		0.592741,
		0.642924,
		0.690604,
		0.735198,
		0.77625,
		0.813438,
		0.84658,
		0.875627,
		0.900656,
		0.921848,
		0.939475,
		0.953873,
		0.96542,
		0.974514,
		0.98155,
		0.986903,
		0.990912,
		0.993865,
		0.995988,
		0.99745,
		0.998361,
		0.998789,
		0.998768,
		0.998305,
		0.997388,
		0.995994,
		0.994091,
		0.991646,
		0.988624,
		0.984999,
		0.980744,
		0.975845,
		0.970291,
		0.964076,
		0.957205,
		0.949683,
		0.941522,
		0.932739,
		0.92335,
		0.913379,
		0.902848,
		0.891784,
		0.880212,
		0.868161,
		0.855661,
		0.842738,
		0.829424,
		0.815748,
		0.801739,
		0.787425,
		0.772836,
		0.758,
		0.742944,
		0.727694,
		0.712278,
		0.69672,
		0.681045,
		0.665277,
		0.649438,
		0.63355,
		0.617635,
		0.601712,
		0.585802,
		0.569922,
		0.55409,
		0.538323,
		0.522638,
		0.507048,
		0.491569,
		0.476215,
		0.460998,
		0.44593,
		0.431024,
		0.416289,
		0.401737,
		0.387376,
		0.373216,
		0.359265,
		0.345531,
		0.332021,
		0.318742,
		0.3057,
		0.292901,
		0.280351,
		0.268054,
		0.256014,
		0.244237,
		0.232725,
		0.221482,
		0.210511,
		0.199814,
		0.189395,
		0.179254,
		0.169394,
		0.159817,
		0.150522,
		0.141513,
		0.132788,
		0.124349,
		0.116196,
		0.108329,
		0.100747,
		0.093452,
		0.086442,
		0.079717,
		0.073275,
		0.067117,
		0.061241,
		0.055645,
		0.05033,
		0.045292,
		0.040531,
		0.036046,
		0.031834,
		0.027894,
		0.024223,
		0.020821,
		0.017685,
		0.014812,
		0.012202,
		0.009851,
		0.007758,
		0.00592,
		0.004334,
		0.003,
		0.001913,
		0.001072,
		0.000475,
		0.000118,
		0.000268,
		0.001962,
		0.005777,
		0.012361,
		0.02233,
		0.036236,
		0.054533,
		0.077548,
		0.105461,
		0.138292,
		0.175889,
		0.217932,
		0.263942,
		0.313298,
		0.365257,
		0.418991,
		0.473615,
		0.528227,
		0.581942,
		0.633929,
		0.683445,
		0.729854,
		0.77265,
		0.811468,
		0.846087,
		0.876427,
		0.902538,
		0.924584,
		0.942826,
		0.957597,
		0.969283,
		0.978299,
		0.985067,
		0.99,
		0.993485,
		0.995868,
		0.99745,
		0.998473,
		0.999123,
		0.999522,
		0.999737,
		0.999787,
		0.999659,
		0.999317,
		0.998711,
		0.997788,
		0.996489,
		0.994763,
		0.992561,
		0.989843,
		0.986574,
		0.98273,
		0.978292,
		0.973246,
		0.967586,
		0.961312,
		0.954428,
		0.94694,
		0.938862,
		0.930208,
		0.920996,
		0.911245,
		0.900978,
		0.890217,
		0.878987,
		0.867313,
		0.855222,
		0.842738,
		0.82989,
		0.816702,
		0.803203,
		0.789417,
		0.77537,
		0.761088,
		0.746597,
		0.731919,
		0.717079,
		0.7021,
		0.687003,
		0.671812,
		0.656545,
		0.641225,
		0.625869,
		0.610497,
		0.595126,
		0.579774,
		0.564456,
		0.549189,
		0.533987,
		0.518865,
		0.503836,
		0.488913,
		0.474108,
		0.459432,
		0.444897,
		0.430513,
		0.416289,
		0.402235,
		0.38836,
		0.374671,
		0.361176,
		0.347883,
		0.334797,
		0.321925,
		0.309274,
		0.296847,
		0.28465,
		0.272688,
		0.260965,
		0.249484,
		0.238249,
		0.227264,
		0.21653,
		0.206051,
		0.19583,
		0.185867,
		0.176164,
		0.166724,
		0.157547,
		0.148635,
		0.139988,
		0.131607,
		0.123492,
		0.115644,
		0.108062,
		0.100747,
		0.093699,
		0.086916,
		0.080399,
		0.074147,
		0.068159,
		0.062433,
		0.05697,
		0.051768,
		0.046826,
		0.042142,
		0.037715,
		0.033544,
		0.029626,
		0.025962,
		0.022547,
		0.019382,
		0.016464,
		0.013791,
		0.011362,
		0.009174,
		0.007225,
		0.005514,
		0.004038,
		0.002795,
		0.001783,
		0.000999,
		0.000443,
		0.00011,
		0.001043,
		0.004664,
		0.011571,
		0.022412,
		0.037732,
		0.057942,
		0.083299,
		0.113883,
		0.149594,
		0.190145,
		0.235077,
		0.283774,
		0.33549,
		0.389377,
		0.444525,
		0.5,
		0.55488,
		0.608295,
		0.659459,
		0.707694,
		0.75245,
		0.793321,
		0.830044,
		0.862495,
		0.890684,
		0.914735,
		0.934871,
		0.951392,
		0.964651,
		0.975038,
		0.982956,
		0.988808,
		0.992978,
		0.995825,
		0.997669,
		0.998789,
		0.999416,
		0.999737,
		0.999885,
		0.999952,
		0.999981,
		0.999976,
		0.999911,
		0.999741,
		0.999412,
		0.998863,
		0.998037,
		0.996876,
		0.995329,
		0.99335,
		0.990902,
		0.987949,
		0.984468,
		0.980438,
		0.975845,
		0.970681,
		0.964943,
		0.958631,
		0.951752,
		0.944313,
		0.936326,
		0.927805,
		0.918768,
		0.909232,
		0.899218,
		0.888746,
		0.877839,
		0.86652,
		0.854812,
		0.842738,
		0.830324,
		0.817592,
		0.804566,
		0.79127,
		0.777728,
		0.763962,
		0.749996,
		0.73585,
		0.721547,
		0.707107,
		0.692551,
		0.677898,
		0.663169,
		0.64838,
		0.63355,
		0.618696,
		0.603835,
		0.588982,
		0.574153,
		0.559361,
		0.544621,
		0.529947,
		0.51535,
		0.500843,
		0.486437,
		0.472143,
		0.457972,
		0.443933,
		0.430036,
		0.416289,
		0.402701,
		0.389279,
		0.376031,
		0.362964,
		0.350084,
		0.337397,
		0.324909,
		0.312626,
		0.300551,
		0.28869,
		0.277047,
		0.265625,
		0.254429,
		0.243461,
		0.232725,
		0.222223,
		0.211958,
		0.201932,
		0.192146,
		0.182603,
		0.173305,
		0.164251,
		0.155444,
		0.146884,
		0.138573,
		0.130509,
		0.122695,
		0.11513,
		0.107814,
		0.100747,
		0.09393,
		0.08736,
		0.081039,
		0.074965,
		0.069138,
		0.063557,
		0.058222,
		0.05313,
		0.048281,
		0.043674,
		0.039308,
		0.035182,
		0.031293,
		0.027641,
		0.024223,
		0.02104,
		0.018088,
		0.015366,
		0.012873,
		0.010606,
		0.008564,
		0.006746,
		0.005149,
		0.003771,
		0.00261,
		0.001665,
		0.000934,
		0.000414,
		0.000103,
		0.000161,
		0.002816,
		0.009308,
		0.020295,
		0.036295,
		0.057659,
		0.084551,
		0.116937,
		0.154586,
		0.197077,
		0.243819,
		0.294076,
		0.347001,
		0.401677,
		0.45715,
		0.512479,
		0.566765,
		0.61919,
		0.669042,
		0.715733,
		0.758816,
		0.797981,
		0.83306,
		0.864009,
		0.890903,
		0.913908,
		0.93327,
		0.949288,
		0.962301,
		0.972665,
		0.980739,
		0.986877,
		0.991411,
		0.994648,
		0.996864,
		0.998305,
		0.999178,
		0.999659,
		0.999888,
		0.999976,
		0.999997,
		1,
		0.999999,
		0.999984,
		0.999922,
		0.999771,
		0.999478,
		0.998989,
		0.998251,
		0.99721,
		0.99582,
		0.994038,
		0.991827,
		0.989154,
		0.985994,
		0.982327,
		0.978139,
		0.973418,
		0.968161,
		0.962366,
		0.956035,
		0.949176,
		0.941795,
		0.933906,
		0.925522,
		0.916658,
		0.907331,
		0.897559,
		0.887363,
		0.876762,
		0.865776,
		0.854428,
		0.842738,
		0.830729,
		0.818422,
		0.805839,
		0.793,
		0.779928,
		0.766643,
		0.753166,
		0.739516,
		0.725714,
		0.711778,
		0.697728,
		0.683581,
		0.669354,
		0.655065,
		0.64073,
		0.626365,
		0.611985,
		0.597605,
		0.583238,
		0.568899,
		0.5546,
		0.540354,
		0.526172,
		0.512066,
		0.498046,
		0.484124,
		0.470308,
		0.456608,
		0.443032,
		0.42959,
		0.416289,
		0.403137,
		0.39014,
		0.377306,
		0.36464,
		0.352149,
		0.339837,
		0.327712,
		0.315776,
		0.304035,
		0.292492,
		0.281153,
		0.27002,
		0.259096,
		0.248386,
		0.237891,
		0.227614,
		0.217558,
		0.207724,
		0.198115,
		0.188732,
		0.179577,
		0.170651,
		0.161955,
		0.15349,
		0.145256,
		0.137255,
		0.129487,
		0.121952,
		0.114651,
		0.107582,
		0.100747,
		0.094146,
		0.087777,
		0.08164,
		0.075735,
		0.070061,
		0.064618,
		0.059405,
		0.054421,
		0.049664,
		0.045134,
		0.04083,
		0.036751,
		0.032895,
		0.029261,
		0.025848,
		0.022654,
		0.019678,
		0.016918,
		0.014374,
		0.012042,
		0.009923,
		0.008014,
		0.006313,
		0.004818,
		0.003529,
		0.002443,
		0.001559,
		0.000874,
		0.000387,
		9.6e-005,
		0.000913,
		0.00605,
		0.01627,
		0.032043,
		0.053638,
		0.08111,
		0.1143,
		0.152846,
		0.196197,
		0.243645,
		0.294361,
		0.347429,
		0.401894,
		0.456801,
		0.511233,
		0.564348,
		0.615402,
		0.663773,
		0.708972,
		0.750642,
		0.788563,
		0.822636,
		0.852871,
		0.879373,
		0.90232,
		0.921945,
		0.938519,
		0.952336,
		0.963696,
		0.9729,
		0.980234,
		0.98597,
		0.990359,
		0.993631,
		0.995994,
		0.997633,
		0.998711,
		0.999373,
		0.999741,
		0.999917,
		0.999984,
		0.999999,
		1,
		0.999999,
		0.999985,
		0.999931,
		0.999797,
		0.999535,
		0.999098,
		0.998436,
		0.997501,
		0.996249,
		0.99464,
		0.992637,
		0.990212,
		0.987338,
		0.983995,
		0.980168,
		0.975845,
		0.971021,
		0.965692,
		0.959858,
		0.953525,
		0.946697,
		0.939384,
		0.931598,
		0.92335,
		0.914656,
		0.905532,
		0.895994,
		0.886059,
		0.875748,
		0.865077,
		0.854068,
		0.842738,
		0.831109,
		0.8192,
		0.807029,
		0.794618,
		0.781985,
		0.769149,
		0.756129,
		0.742944,
		0.72961,
		0.716147,
		0.70257,
		0.688896,
		0.675142,
		0.661323,
		0.647454,
		0.63355,
		0.619625,
		0.605693,
		0.591766,
		0.577857,
		0.563978,
		0.550142,
		0.536358,
		0.522638,
		0.508991,
		0.495428,
		0.481957,
		0.468589,
		0.455329,
		0.442188,
		0.429172,
		0.416289,
		0.403546,
		0.390948,
		0.378502,
		0.366214,
		0.354089,
		0.342132,
		0.330348,
		0.318742,
		0.307317,
		0.296078,
		0.285028,
		0.27417,
		0.263508,
		0.253045,
		0.242783,
		0.232725,
		0.222873,
		0.213228,
		0.203793,
		0.19457,
		0.185559,
		0.176763,
		0.168182,
		0.159817,
		0.151669,
		0.143738,
		0.136026,
		0.128533,
		0.121258,
		0.114202,
		0.107365,
		0.100747,
		0.094348,
		0.088168,
		0.082205,
		0.076461,
		0.070933,
		0.065621,
		0.060526,
		0.055645,
		0.050979,
		0.046525,
		0.042284,
		0.038254,
		0.034434,
		0.030823,
		0.02742,
		0.024223,
		0.021232,
		0.018444,
		0.015859,
		0.013475,
		0.01129,
		0.009304,
		0.007514,
		0.00592,
		0.004519,
		0.00331,
		0.002292,
		0.001462,
		0.00082,
		0.000363,
		9.1e-005,
		0.002672,
		0.011064,
		0.025572,
		0.046363,
		0.073371,
		0.106308,
		0.144681,
		0.187825,
		0.234937,
		0.285117,
		0.337415,
		0.390871,
		0.444555,
		0.497602,
		0.549241,
		0.598814,
		0.645788,
		0.689759,
		0.730447,
		0.767694,
		0.801442,
		0.831725,
		0.858647,
		0.882368,
		0.903084,
		0.921018,
		0.936403,
		0.949475,
		0.960466,
		0.969601,
		0.977092,
		0.983139,
		0.987932,
		0.991646,
		0.994447,
		0.996489,
		0.997917,
		0.998863,
		0.999446,
		0.999771,
		0.999927,
		0.999985,
		0.999999,
		1,
		0.999999,
		0.999987,
		0.999939,
		0.999819,
		0.999585,
		0.999193,
		0.998598,
		0.997755,
		0.996625,
		0.995167,
		0.99335,
		0.991144,
		0.988525,
		0.985471,
		0.981968,
		0.978003,
		0.973569,
		0.968661,
		0.963278,
		0.957423,
		0.951098,
		0.944313,
		0.937074,
		0.929393,
		0.921283,
		0.912756,
		0.903828,
		0.894513,
		0.884829,
		0.874792,
		0.864419,
		0.853729,
		0.842738,
		0.831466,
		0.819929,
		0.808146,
		0.796135,
		0.783913,
		0.771497,
		0.758906,
		0.746155,
		0.73326,
		0.720239,
		0.707107,
		0.693878,
		0.680569,
		0.667192,
		0.653763,
		0.640295,
		0.6268,
		0.613292,
		0.599783,
		0.586284,
		0.572806,
		0.559361,
		0.545959,
		0.532609,
		0.519322,
		0.506107,
		0.492972,
		0.479925,
		0.466975,
		0.45413,
		0.441396,
		0.42878,
		0.416289,
		0.40393,
		0.391707,
		0.379627,
		0.367695,
		0.355915,
		0.344293,
		0.332833,
		0.321539,
		0.310415,
		0.299464,
		0.28869,
		0.278096,
		0.267685,
		0.25746,
		0.247423,
		0.237576,
		0.227922,
		0.218463,
		0.2092,
		0.200135,
		0.191269,
		0.182603,
		0.17414,
		0.165879,
		0.157821,
		0.149968,
		0.14232,
		0.134877,
		0.127639,
		0.120607,
		0.113781,
		0.107161,
		0.100747,
		0.094539,
		0.088536,
		0.082738,
		0.077145,
		0.071756,
		0.066571,
		0.061589,
		0.056809,
		0.05223,
		0.047852,
		0.043674,
		0.039695,
		0.035914,
		0.03233,
		0.028941,
		0.025748,
		0.022747,
		0.019939,
		0.017323,
		0.014896,
		0.012657,
		0.010606,
		0.008741,
		0.00706,
		0.000352,
		0.005759,
		0.017889,
		0.036791,
		0.062266,
		0.093895,
		0.131066,
		0.173013,
		0.218859,
		0.267665,
		0.318468,
		0.370327,
		0.422357,
		0.473757,
		0.523831,
		0.572,
		0.617806,
		0.660911,
		0.701088,
		0.73821,
		0.772234,
		0.803188,
		0.831152,
		0.856247,
		0.878618,
		0.898429,
		0.915849,
		0.931051,
		0.944206,
		0.955481,
		0.965039,
		0.973042,
		0.979644,
		0.984999,
		0.989256,
		0.992561,
		0.995056,
		0.996876,
		0.998148,
		0.998989,
		0.999508,
		0.999797,
		0.999935,
		0.999987,
		0.999999,
		1,
		0.999999,
		0.999988,
		0.999945,
		0.999838,
		0.999628,
		0.999276,
		0.998739,
		0.997978,
		0.996954,
		0.995632,
		0.99398,
		0.991969,
		0.989577,
		0.986782,
		0.98357,
		0.979927,
		0.975845,
		0.971319,
		0.966345,
		0.960924,
		0.95506,
		0.948756,
		0.94202,
		0.93486,
		0.927288,
		0.919314,
		0.91095,
		0.902212,
		0.893112,
		0.883666,
		0.87389,
		0.863799,
		0.85341,
		0.842738,
		0.831801,
		0.820615,
		0.809195,
		0.797559,
		0.785723,
		0.773702,
		0.761512,
		0.749168,
		0.736686,
		0.724081,
		0.711367,
		0.698557,
		0.685666,
		0.672708,
		0.659694,
		0.646637,
		0.63355,
		0.620445,
		0.607332,
		0.594222,
		0.581127,
		0.568056,
		0.55502,
		0.542026,
		0.529086,
		0.516206,
		0.503396,
		0.490663,
		0.478015,
		0.465459,
		0.453002,
		0.44065,
		0.428411,
		0.416289,
		0.404291,
		0.392422,
		0.380687,
		0.369091,
		0.357638,
		0.346332,
		0.335179,
		0.324181,
		0.313342,
		0.302666,
		0.292156,
		0.281814,
		0.271644,
		0.261647,
		0.251827,
		0.242186,
		0.232725,
		0.223446,
		0.214352,
		0.205443,
		0.196721,
		0.188187,
		0.179843,
		0.171689,
		0.163726,
		0.155955,
		0.148377,
		0.140991,
		0.133799,
		0.126801,
		0.119997,
		0.113386,
		0.10697,
		0.100747,
		0.094719,
		0.088884,
		0.083242,
		0.077793,
		0.072536,
		0.067471,
		0.062598,
		0.057915,
		0.053423,
		0.049119,
		0.045004,
		0.041077,
		0.037337,
		0.033782,
		0.030412,
		0.027226,
		0.024223,
		0.001638,
		0.010267,
		0.026203,
		0.049109,
		0.078437,
		0.113465,
		0.153342,
		0.197141,
		0.243899,
		0.292662,
		0.342523,
		0.392648,
		0.442298,
		0.490845,
		0.537774,
		0.582681,
		0.62527,
		0.66534,
		0.702768,
		0.737504,
		0.769549,
		0.798942,
		0.825756,
		0.85008,
		0.872018,
		0.891682,
		0.909189,
		0.924657,
		0.938211,
		0.949977,
		0.960083,
		0.968661,
		0.975845,
		0.981771,
		0.986574,
		0.99039,
		0.99335,
		0.995583,
		0.99721,
		0.998347,
		0.999098,
		0.999561,
		0.999819,
		0.999942,
		0.999988,
		0.999999,
		1,
		0.999999,
		0.99999,
		0.999951,
		0.999854,
		0.999666,
		0.999348,
		0.998863,
		0.998174,
		0.997245,
		0.996042,
		0.994536,
		0.9927,
		0.990511,
		0.987949,
		0.984999,
		0.981647,
		0.977883,
		0.973702,
		0.9691,
		0.964076,
		0.958631,
		0.952769,
		0.946495,
		0.939815,
		0.932739,
		0.925275,
		0.917436,
		0.909232,
		0.900677,
		0.891784,
		0.882566,
		0.873037,
		0.863213,
		0.853109,
		0.842738,
		0.832117,
		0.82126,
		0.810183,
		0.798899,
		0.787425,
		0.775775,
		0.763962,
		0.752002,
		0.739908,
		0.727694,
		0.715374,
		0.702959,
		0.690463,
		0.677898,
		0.665277,
		0.65261,
		0.63991,
		0.627186,
		0.61445,
		0.601712,
		0.588982,
		0.576269,
		0.563582,
		0.550931,
		0.538323,
		0.525768,
		0.513271,
		0.500843,
		0.488488,
		0.476215,
		0.46403,
		0.451939,
		0.439948,
		0.428063,
		0.416289,
		0.404632,
		0.393097,
		0.381687,
		0.370409,
		0.359265,
		0.34826,
		0.337397,
		0.326681,
		0.316114,
		0.3057,
		0.295441,
		0.285341,
		0.275401,
		0.265625,
		0.256014,
		0.246571,
		0.237298,
		0.228195,
		0.219266,
		0.210511,
		0.201932,
		0.193529,
		0.185305,
		0.17726,
		0.169394,
		0.16171,
		0.154206,
		0.146884,
		0.139745,
		0.132788,
		0.126014,
		0.119423,
		0.113014,
		0.106789,
		0.100747,
		0.094888,
		0.089212,
		0.083718,
		0.078406,
		0.073275,
		0.068326,
		0.063557,
		0.058969,
		0.05456,
		0.05033,
		0.004055,
		0.016015,
		0.0354,
		0.061533,
		0.093595,
		0.130667,
		0.171783,
		0.215973,
		0.262302,
		0.309899,
		0.357979,
		0.405858,
		0.452958,
		0.498808,
		0.543038,
		0.585368,
		0.625599,
		0.663597,
		0.699281,
		0.732617,
		0.763598,
		0.792245,
		0.818594,
		0.842698,
		0.864619,
		0.884429,
		0.90221,
		0.918052,
		0.932051,
		0.944313,
		0.954949,
		0.964076,
		0.971816,
		0.978292,
		0.983628,
		0.987949,
		0.991379,
		0.994038,
		0.996042,
		0.997501,
		0.99852,
		0.999193,
		0.999607,
		0.999838,
		0.999948,
		0.99999,
		0.999999,
		1,
		0.999999,
		0.999991,
		0.999956,
		0.999869,
		0.999699,
		0.999412,
		0.998972,
		0.998347,
		0.997501,
		0.996405,
		0.99503,
		0.99335,
		0.991344,
		0.988991,
		0.986276,
		0.983187,
		0.979712,
		0.975845,
		0.971582,
		0.96692,
		0.961859,
		0.956401,
		0.950551,
		0.944313,
		0.937694,
		0.930704,
		0.92335,
		0.915644,
		0.907596,
		0.899218,
		0.890522,
		0.881522,
		0.87223,
		0.862659,
		0.852824,
		0.842738,
		0.832415,
		0.821869,
		0.811114,
		0.800163,
		0.78903,
		0.777728,
		0.766271,
		0.754672,
		0.742944,
		0.731099,
		0.719149,
		0.707107,
		0.694984,
		0.682792,
		0.670542,
		0.658245,
		0.645911,
		0.63355,
		0.621173,
		0.608789,
		0.596407,
		0.584036,
		0.571684,
		0.559361,
		0.547074,
		0.53483,
		0.522638,
		0.510503,
		0.498434,
		0.486437,
		0.474517,
		0.462682,
		0.450936,
		0.439285,
		0.427734,
		0.416289,
		0.404954,
		0.393734,
		0.382633,
		0.371655,
		0.360804,
		0.350084,
		0.339498,
		0.32905,
		0.318742,
		0.308577,
		0.298559,
		0.28869,
		0.278972,
		0.269407,
		0.259999,
		0.250747,
		0.241656,
		0.232725,
		0.223957,
		0.215353,
		0.206915,
		0.198643,
		0.190539,
		0.182603,
		0.174837,
		0.167242,
		0.159817,
		0.152563,
		0.145482,
		0.138573,
		0.131836,
		0.125272,
		0.118882,
		0.112664,
		0.106619,
		0.100747,
		0.095049,
		0.089522,
		0.084169,
		0.000483,
		0.007571,
		0.0226,
		0.044773,
		0.073179,
		0.106843,
		0.144773,
		0.185998,
		0.229604,
		0.274752,
		0.320699,
		0.366798,
		0.412507,
		0.457378,
		0.501051,
		0.543246,
		0.583747,
		0.622395,
		0.659076,
		0.69371,
		0.726248,
		0.756663,
		0.784947,
		0.811109,
		0.835175,
		0.857183,
		0.877185,
		0.895243,
		0.911434,
		0.925841,
		0.938558,
		0.949683,
		0.959323,
		0.967586,
		0.974587,
		0.980438,
		0.985256,
		0.989154,
		0.992245,
		0.99464,
		0.996443,
		0.997755,
		0.998671,
		0.999276,
		0.999648,
		0.999854,
		0.999954,
		0.999991,
		0.999999,
		1,
		0.999999,
		0.999992,
		0.99996,
		0.999882,
		0.999728,
		0.999468,
		0.999069,
		0.998499,
		0.997729,
		0.996728,
		0.99547,
		0.99393,
		0.992087,
		0.989923,
		0.987421,
		0.984569,
		0.981357,
		0.977776,
		0.973821,
		0.969489,
		0.96478,
		0.959694,
		0.954233,
		0.948403,
		0.942207,
		0.935655,
		0.928752,
		0.921508,
		0.913932,
		0.906036,
		0.897829,
		0.889324,
		0.880531,
		0.871464,
		0.862134,
		0.852555,
		0.842738,
		0.832697,
		0.822445,
		0.811993,
		0.801356,
		0.790544,
		0.779572,
		0.76845,
		0.757191,
		0.745808,
		0.734311,
		0.722712,
		0.711022,
		0.699252,
		0.687413,
		0.675515,
		0.663568,
		0.651582,
		0.639566,
		0.62753,
		0.615483,
		0.603434,
		0.591389,
		0.579359,
		0.567351,
		0.555372,
		0.543429,
		0.531529,
		0.51968,
		0.507888,
		0.496159,
		0.484498,
		0.472913,
		0.461407,
		0.449987,
		0.438658,
		0.427424,
		0.416289,
		0.405259,
		0.394338,
		0.383529,
		0.372836,
		0.362263,
		0.351814,
		0.34149,
		0.331297,
		0.321236,
		0.31131,
		0.301522,
		0.291874,
		0.282369,
		0.273008,
		0.263794,
		0.254728,
		0.245813,
		0.237049,
		0.228439,
		0.219983,
		0.211684,
		0.203541,
		0.195557,
		0.187731,
		0.180066,
		0.172561,
		0.165218,
		0.158037,
		0.151018,
		0.144161,
		0.137468,
		0.130939,
		0.124573,
		0.001976,
		0.011958,
		0.029542,
		0.053735,
		0.083512,
		0.11786,
		0.155812,
		0.19647,
		0.239021,
		0.282748,
		0.327031,
		0.371341,
		0.415238,
		0.458357,
		0.500403,
		0.541136,
		0.580364,
		0.617935,
		0.653732,
		0.687664,
		0.719666,
		0.749695,
		0.777728,
		0.803759,
		0.827802,
		0.849883,
		0.870045,
		0.888342,
		0.904839,
		0.91961,
		0.932739,
		0.944313,
		0.954428,
		0.963183,
		0.970681,
		0.977027,
		0.982327,
		0.986687,
		0.990212,
		0.993005,
		0.995167,
		0.996795,
		0.997978,
		0.998803,
		0.999348,
		0.999683,
		0.999869,
		0.999958,
		0.999992,
		0.999999,
		1,
		1,
		0.999993,
		0.999964,
		0.999893,
		0.999754,
		0.999517,
		0.999154,
		0.998635,
		0.997931,
		0.997015,
		0.995862,
		0.994448,
		0.992753,
		0.990759,
		0.98845,
		0.985814,
		0.982839,
		0.979518,
		0.975845,
		0.971816,
		0.967429,
		0.962684,
		0.957583,
		0.952128,
		0.946324,
		0.940176,
		0.933692,
		0.926878,
		0.919743,
		0.912296,
		0.904547,
		0.896506,
		0.888183,
		0.879589,
		0.870737,
		0.861636,
		0.8523,
		0.842738,
		0.832964,
		0.82299,
		0.812825,
		0.802484,
		0.791976,
		0.781314,
		0.770509,
		0.759572,
		0.748515,
		0.737347,
		0.726079,
		0.714722,
		0.703287,
		0.691782,
		0.680218,
		0.668603,
		0.656948,
		0.645261,
		0.63355,
		0.621825,
		0.610092,
		0.598361,
		0.586639,
		0.574932,
		0.563249,
		0.551596,
		0.539979,
		0.528406,
		0.516882,
		0.505413,
		0.494005,
		0.482664,
		0.471394,
		0.460201,
		0.449089,
		0.438064,
		0.427129,
		0.416289,
		0.405548,
		0.39491,
		0.384378,
		0.373956,
		0.363647,
		0.353455,
		0.343382,
		0.333432,
		0.323607,
		0.313909,
		0.304341,
		0.294906,
		0.285604,
		0.27644,
		0.267414,
		0.258527,
		0.249783,
		0.241182,
		0.232725,
		0.224414,
		0.216251,
		0.208236,
		0.200371,
		0.192655,
		0.185091,
		0.177679,
		0.170419,
		0.0044,
		0.016906,
		0.036449,
		0.061965,
		0.092422,
		0.126853,
		0.164372,
		0.204182,
		0.245582,
		0.287961,
		0.330792,
		0.373628,
		0.416091,
		0.457861,
		0.498671,
		0.538297,
		0.576554,
		0.613289,
		0.648379,
		0.681729,
		0.713267,
		0.742944,
		0.770731,
		0.796619,
		0.820615,
		0.842738,
		0.863025,
		0.881522,
		0.898284,
		0.913379,
		0.926878,
		0.938862,
		0.949417,
		0.958631,
		0.9666,
		0.973418,
		0.979184,
		0.983995,
		0.987949,
		0.991144,
		0.993675,
		0.995632,
		0.997104,
		0.998174,
		0.99892,
		0.999412,
		0.999714,
		0.999882,
		0.999962,
		0.999993,
		1,
		1,
		1,
		0.999993,
		0.999968,
		0.999903,
		0.999776,
		0.999561,
		0.99923,
		0.998756,
		0.998111,
		0.997272,
		0.996213,
		0.994912,
		0.99335,
		0.99151,
		0.989376,
		0.986936,
		0.984178,
		0.981094,
		0.977679,
		0.973927,
		0.969836,
		0.965405,
		0.960634,
		0.955526,
		0.950084,
		0.944313,
		0.938217,
		0.931803,
		0.925079,
		0.918052,
		0.910731,
		0.903125,
		0.895243,
		0.887096,
		0.878693,
		0.870045,
		0.861163,
		0.852057,
		0.842738,
		0.833218,
		0.823506,
		0.813614,
		0.803552,
		0.793332,
		0.782964,
		0.772459,
		0.761826,
		0.751077,
		0.74022,
		0.729267,
		0.718226,
		0.707107,
		0.695919,
		0.684672,
		0.673374,
		0.662033,
		0.650658,
		0.639258,
		0.627839,
		0.61641,
		0.604978,
		0.59355,
		0.582134,
		0.570735,
		0.559361,
		0.548017,
		0.536711,
		0.525446,
		0.51423,
		0.503068,
		0.491965,
		0.480925,
		0.469955,
		0.459058,
		0.448238,
		0.437501,
		0.42685,
		0.416289,
		0.405822,
		0.395453,
		0.385184,
		0.37502,
		0.364963,
		0.355015,
		0.345181,
		0.335463,
		0.325863,
		0.316383,
		0.307026,
		0.297795,
		0.28869,
		0.279714,
		0.270869,
		0.262156,
		0.253577,
		0.245133,
		0.236827,
		0.228658,
		0.220628,
		0.000494,
		0.007563,
		0.022135,
		0.043104,
		0.069425,
		0.100129,
		0.134336,
		0.171255,
		0.21019,
		0.250527,
		0.291735,
		0.333351,
		0.374977,
		0.416269,
		0.456932,
		0.496713,
		0.535398,
		0.572806,
		0.608789,
		0.643226,
		0.676022,
		0.707107,
		0.73643,
		0.763962,
		0.789689,
		0.813614,
		0.835752,
		0.856131,
		0.874792,
		0.891784,
		0.907163,
		0.920996,
		0.933353,
		0.944313,
		0.953955,
		0.962366,
		0.969633,
		0.975845,
		0.981094,
		0.985471,
		0.989067,
		0.991969,
		0.994266,
		0.996042,
		0.997377,
		0.998347,
		0.999022,
		0.999468,
		0.999741,
		0.999893,
		0.999966,
		0.999993,
		1,
		1,
		1,
		0.999994,
		0.999971,
		0.999912,
		0.999797,
		0.9996,
		0.999297,
		0.998863,
		0.998273,
		0.997501,
		0.996527,
		0.995329,
		0.993888,
		0.992187,
		0.990212,
		0.987949,
		0.985389,
		0.982523,
		0.979343,
		0.975845,
		0.972026,
		0.967884,
		0.963419,
		0.958631,
		0.953525,
		0.948101,
		0.942367,
		0.936326,
		0.929985,
		0.92335,
		0.91643,
		0.909232,
		0.901765,
		0.894038,
		0.886059,
		0.877839,
		0.869387,
		0.860713,
		0.851827,
		0.842738,
		0.833458,
		0.823996,
		0.814362,
		0.804566,
		0.794618,
		0.784529,
		0.774307,
		0.763962,
		0.753505,
		0.742944,
		0.732288,
		0.721547,
		0.710728,
		0.699842,
		0.688896,
		0.677898,
		0.666857,
		0.655781,
		0.644676,
		0.63355,
		0.622411,
		0.611266,
		0.60012,
		0.588982,
		0.577857,
		0.566751,
		0.555671,
		0.544621,
		0.533608,
		0.522638,
		0.511714,
		0.500843,
		0.490028,
		0.479275,
		0.468589,
		0.457972,
		0.44743,
		0.436966,
		0.426585,
		0.416289,
		0.406083,
		0.395969,
		0.385951,
		0.376031,
		0.366214,
		0.3565,
		0.346894,
		0.337397,
		0.328012,
		0.318742,
		0.309587,
		0.300551,
		0.291635,
		0.282841,
		0.27417,
		0.001888,
		0.011259,
		0.027471,
		0.049465,
		0.076267,
		0.106998,
		0.140868,
		0.177175,
		0.215301,
		0.254699,
		0.294891,
		0.335457,
		0.376029,
		0.416289,
		0.455958,
		0.494799,
		0.532609,
		0.569217,
		0.604481,
		0.638287,
		0.670542,
		0.701178,
		0.730146,
		0.757413,
		0.782964,
		0.806797,
		0.828924,
		0.849368,
		0.868161,
		0.885348,
		0.900978,
		0.915109,
		0.927805,
		0.939136,
		0.949176,
		0.958001,
		0.965692,
		0.972331,
		0.978003,
		0.982792,
		0.986782,
		0.990058,
		0.9927,
		0.99479,
		0.996405,
		0.997618,
		0.998499,
		0.999113,
		0.999517,
		0.999765,
		0.999903,
		0.999969,
		0.999994,
		1,
		1,
		1,
		0.999994,
		0.999973,
		0.99992,
		0.999814,
		0.999635,
		0.999358,
		0.99896,
		0.998417,
		0.997707,
		0.99681,
		0.995704,
		0.994372,
		0.992798,
		0.990967,
		0.988867,
		0.986488,
		0.983819,
		0.980856,
		0.977591,
		0.974023,
		0.970147,
		0.965963,
		0.961473,
		0.956676,
		0.951577,
		0.946178,
		0.940484,
		0.934501,
		0.928233,
		0.921688,
		0.914874,
		0.907796,
		0.900464,
		0.892886,
		0.88507,
		0.877025,
		0.86876,
		0.860284,
		0.851607,
		0.842738,
		0.833687,
		0.824461,
		0.815072,
		0.805528,
		0.795839,
		0.786014,
		0.776061,
		0.76599,
		0.75581,
		0.745529,
		0.735156,
		0.724699,
		0.714166,
		0.703566,
		0.692907,
		0.682196,
		0.67144,
		0.660647,
		0.649825,
		0.638979,
		0.628118,
		0.617247,
		0.606372,
		0.595501,
		0.584639,
		0.573791,
		0.562964,
		0.552163,
		0.541394,
		0.530661,
		0.519969,
		0.509323,
		0.498728,
		0.488188,
		0.477707,
		0.46729,
		0.45694,
		0.446662,
		0.436458,
		0.426333,
		0.416289,
		0.406331,
		0.39646,
		0.386681,
		0.376995,
		0.367405,
		0.357915,
		0.348527,
		0.339242,
		0.330063,
		0.004022,
		0.015327,
		0.032845,
		0.055596,
		0.082699,
		0.113359,
		0.146868,
		0.182595,
		0.21998,
		0.258526,
		0.297794,
		0.337397,
		0.376995,
		0.416289,
		0.455022,
		0.492972,
		0.529947,
		0.565786,
		0.600358,
		0.63355,
		0.665277,
		0.69547,
		0.724081,
		0.751077,
		0.776439,
		0.800163,
		0.822257,
		0.842738,
		0.861636,
		0.878987,
		0.894835,
		0.909232,
		0.922235,
		0.933906,
		0.944313,
		0.953525,
		0.961616,
		0.968661,
		0.974739,
		0.979927,
		0.984305,
		0.987949,
		0.990939,
		0.99335,
		0.995256,
		0.996728,
		0.997833,
		0.998635,
		0.999193,
		0.999561,
		0.999787,
		0.999912,
		0.999972,
		0.999994,
		1,
		1,
		1,
		0.999995,
		0.999976,
		0.999927,
		0.99983,
		0.999666,
		0.999412,
		0.999046,
		0.998546,
		0.997893,
		0.997064,
		0.996042,
		0.994809,
		0.99335,
		0.991651,
		0.9897,
		0.987485,
		0.984999,
		0.982234,
		0.979184,
		0.975845,
		0.972215,
		0.968292,
		0.964076,
		0.959568,
		0.954769,
		0.949683,
		0.944313,
		0.938663,
		0.932739,
		0.926546,
		0.92009,
		0.913379,
		0.906419,
		0.899218,
		0.891784,
		0.884124,
		0.876247,
		0.868161,
		0.859876,
		0.851398,
		0.842738,
		0.833904,
		0.824905,
		0.815748,
		0.806444,
		0.797,
		0.787425,
		0.777728,
		0.767917,
		0.758,
		0.747985,
		0.737881,
		0.727694,
		0.717434,
		0.707107,
		0.69672,
		0.686282,
		0.675798,
		0.665277,
		0.654724,
		0.644146,
		0.63355,
		0.622942,
		0.612327,
		0.601712,
		0.591103,
		0.580504,
		0.569922,
		0.559361,
		0.548826,
		0.538323,
		0.527856,
		0.51743,
		0.507048,
		0.496716,
		0.486437,
		0.476215,
		0.466054,
		0.455958,
		0.44593,
		0.435974,
		0.426093,
		0.416289,
		0.406567,
		0.396928,
		0.387376,
		0.000442,
		0.00674,
		0.019668,
		0.038245,
		0.061582,
		0.088879,
		0.11942,
		0.152562,
		0.187731,
		0.224414,
		0.262156,
		0.300551,
		0.339242,
		0.377913,
		0.416289,
		0.45413,
		0.491227,
		0.527402,
		0.562504,
		0.596407,
		0.629005,
		0.660215,
		0.689972,
		0.718226,
		0.744944,
		0.770107,
		0.793707,
		0.815748,
		0.836245,
		0.855222,
		0.872709,
		0.888746,
		0.903379,
		0.916658,
		0.928639,
		0.939384,
		0.948956,
		0.957423,
		0.964853,
		0.971319,
		0.976892,
		0.981647,
		0.985655,
		0.988991,
		0.991726,
		0.99393,
		0.995671,
		0.997015,
		0.998024,
		0.998756,
		0.999265,
		0.9996,
		0.999806,
		0.99992,
		0.999974,
		0.999995,
		1,
		1,
		1,
		0.999995,
		0.999978,
		0.999933,
		0.999845,
		0.999694,
		0.99946,
		0.999123,
		0.998663,
		0.99806,
		0.997294,
		0.996348,
		0.995205,
		0.993851,
		0.992272,
		0.990456,
		0.988393,
		0.986073,
		0.98349,
		0.980638,
		0.977512,
		0.974109,
		0.970427,
		0.966466,
		0.962225,
		0.957705,
		0.95291,
		0.947842,
		0.942504,
		0.9369,
		0.931037,
		0.924919,
		0.918552,
		0.911942,
		0.905097,
		0.898023,
		0.890728,
		0.883219,
		0.875503,
		0.86759,
		0.859486,
		0.851199,
		0.842738,
		0.834112,
		0.825327,
		0.816392,
		0.807315,
		0.798105,
		0.788769,
		0.779315,
		0.769751,
		0.760084,
		0.750323,
		0.740474,
		0.730545,
		0.720543,
		0.710476,
		0.70035,
		0.690172,
		0.679948,
		0.669685,
		0.65939,
		0.649069,
		0.638727,
		0.628371,
		0.618005,
		0.607637,
		0.59727,
		0.586911,
		0.576564,
		0.566235,
		0.555928,
		0.545648,
		0.535398,
		0.525185,
		0.515011,
		0.504881,
		0.494799,
		0.484769,
		0.474793,
		0.464877,
		0.455022,
		0.445233,
		0.001665,
		0.009922,
		0.024222,
		0.043674,
		0.067471,
		0.094888,
		0.125272,
		0.158037,
		0.192655,
		0.228658,
		0.265625,
		0.303183,
		0.341002,
		0.37879,
		0.416289,
		0.453277,
		0.489559,
		0.524968,
		0.559361,
		0.592618,
		0.62464,
		0.655346,
		0.684672,
		0.71257,
		0.739008,
		0.763962,
		0.787425,
		0.809397,
		0.82989,
		0.848922,
		0.86652,
		0.882719,
		0.897559,
		0.911086,
		0.92335,
		0.934406,
		0.944313,
		0.953131,
		0.960924,
		0.967759,
		0.973702,
		0.978822,
		0.983187,
		0.986864,
		0.989923,
		0.992429,
		0.994448,
		0.996042,
		0.997272,
		0.998194,
		0.998863,
		0.999328,
		0.999635,
		0.999823,
		0.999927,
		0.999977,
		0.999995,
		1,
		1,
		1,
		0.999996,
		0.99998,
		0.999939,
		0.999858,
		0.999719,
		0.999504,
		0.999193,
		0.998768,
		0.99821,
		0.997501,
		0.996625,
		0.995564,
		0.994306,
		0.992836,
		0.991144,
		0.98922,
		0.987053,
		0.984638,
		0.981968,
		0.979038,
		0.975845,
		0.972387,
		0.968661,
		0.964669,
		0.960409,
		0.955885,
		0.951098,
		0.946052,
		0.940749,
		0.935195,
		0.929393,
		0.92335,
		0.917071,
		0.910561,
		0.903828,
		0.896877,
		0.889716,
		0.882352,
		0.874792,
		0.867043,
		0.859113,
		0.851009,
		0.842738,
		0.834309,
		0.825729,
		0.817006,
		0.808146,
		0.799158,
		0.790049,
		0.780826,
		0.771497,
		0.762069,
		0.752549,
		0.742944,
		0.73326,
		0.723506,
		0.713686,
		0.703808,
		0.693878,
		0.683903,
		0.673888,
		0.66384,
		0.653763,
		0.643665,
		0.63355,
		0.623424,
		0.613292,
		0.60316,
		0.593031,
		0.582912,
		0.572806,
		0.562719,
		0.552654,
		0.542616,
		0.532609,
		0.522638,
		0.512705,
		0.502815,
		0.003529,
		0.013474,
		0.028941,
		0.049119,
		0.073275,
		0.100747,
		0.130939,
		0.163312,
		0.197385,
		0.232725,
		0.268945,
		0.3057,
		0.342684,
		0.379627,
		0.416289,
		0.452462,
		0.487964,
		0.522638,
		0.556348,
		0.588982,
		0.620445,
		0.650658,
		0.679561,
		0.707107,
		0.73326,
		0.758,
		0.781314,
		0.803203,
		0.823672,
		0.842738,
		0.860425,
		0.876762,
		0.891784,
		0.905532,
		0.918052,
		0.929393,
		0.939609,
		0.948756,
		0.956891,
		0.964076,
		0.970373,
		0.975845,
		0.980556,
		0.984569,
		0.987949,
		0.990759,
		0.99306,
		0.994912,
		0.996374,
		0.997501,
		0.998347,
		0.99896,
		0.999385,
		0.999666,
		0.999838,
		0.999933,
		0.999979,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999981,
		0.999944,
		0.999869,
		0.999741,
		0.999543,
		0.999256,
		0.998863,
		0.998347,
		0.997689,
		0.996876,
		0.99589,
		0.994719,
		0.99335,
		0.991772,
		0.989975,
		0.987949,
		0.985689,
		0.983187,
		0.980438,
		0.977439,
		0.974188,
		0.970681,
		0.96692,
		0.962903,
		0.958631,
		0.954108,
		0.949334,
		0.944313,
		0.939048,
		0.933544,
		0.927805,
		0.921837,
		0.915644,
		0.909232,
		0.902608,
		0.895777,
		0.888746,
		0.881522,
		0.874111,
		0.86652,
		0.858756,
		0.850827,
		0.842738,
		0.834498,
		0.826114,
		0.817592,
		0.808939,
		0.800163,
		0.79127,
		0.782268,
		0.773163,
		0.763962,
		0.754672,
		0.745299,
		0.73585,
		0.726331,
		0.716747,
		0.707107,
		0.697414,
		0.687676,
		0.677898,
		0.668086,
		0.658245,
		0.64838,
		0.638497,
		0.628601,
		0.618696,
		0.608789,
		0.598882,
		0.588982,
		0.579092,
		0.569217,
		0.559361,
		0.000387,
		0.00592,
		0.017323,
		0.033782,
		0.05456,
		0.078987,
		0.106458,
		0.136427,
		0.168399,
		0.201932,
		0.236626,
		0.272124,
		0.308108,
		0.344293,
		0.380428,
		0.416289,
		0.451682,
		0.486437,
		0.520404,
		0.553458,
		0.58549,
		0.61641,
		0.646144,
		0.674631,
		0.701826,
		0.727694,
		0.752213,
		0.77537,
		0.797161,
		0.817592,
		0.836674,
		0.854428,
		0.870879,
		0.886059,
		0.900005,
		0.912756,
		0.924359,
		0.93486,
		0.944313,
		0.952769,
		0.960286,
		0.96692,
		0.97273,
		0.977776,
		0.982117,
		0.985814,
		0.988925,
		0.99151,
		0.993626,
		0.995329,
		0.996672,
		0.997707,
		0.998483,
		0.999046,
		0.999437,
		0.999694,
		0.999851,
		0.999939,
		0.999981,
		0.999996,
		1,
		1,
		1,
		0.999996,
		0.999983,
		0.999948,
		0.99988,
		0.999762,
		0.999578,
		0.999313,
		0.998949,
		0.99847,
		0.99786,
		0.997104,
		0.996187,
		0.995096,
		0.993819,
		0.992346,
		0.990665,
		0.98877,
		0.986651,
		0.984305,
		0.981724,
		0.978905,
		0.975845,
		0.972543,
		0.968996,
		0.965204,
		0.961169,
		0.956891,
		0.952372,
		0.947615,
		0.942622,
		0.937398,
		0.931946,
		0.92627,
		0.920376,
		0.914268,
		0.907953,
		0.901435,
		0.89472,
		0.887815,
		0.880725,
		0.873458,
		0.866019,
		0.858415,
		0.850653,
		0.842738,
		0.834679,
		0.826481,
		0.818151,
		0.809696,
		0.801123,
		0.792437,
		0.783645,
		0.774754,
		0.76577,
		0.756699,
		0.747548,
		0.738322,
		0.729027,
		0.71967,
		0.710256,
		0.700791,
		0.69128,
		0.681729,
		0.672143,
		0.662527,
		0.652886,
		0.643226,
		0.63355,
		0.623865,
		0.614173,
		0.001462,
		0.008741,
		0.021402,
		0.038704,
		0.059973,
		0.084596,
		0.112019,
		0.14174,
		0.173305,
		0.206304,
		0.24037,
		0.275172,
		0.310415,
		0.345833,
		0.381194,
		0.416289,
		0.450936,
		0.484973,
		0.518262,
		0.550683,
		0.582134,
		0.612528,
		0.641793,
		0.669872,
		0.69672,
		0.722303,
		0.746597,
		0.769588,
		0.79127,
		0.811648,
		0.830729,
		0.848532,
		0.865077,
		0.880394,
		0.894513,
		0.907473,
		0.919314,
		0.930079,
		0.939815,
		0.948572,
		0.956401,
		0.963355,
		0.969489,
		0.974858,
		0.979518,
		0.983525,
		0.986936,
		0.989805,
		0.992187,
		0.994136,
		0.995704,
		0.99694,
		0.997893,
		0.998606,
		0.999123,
		0.999482,
		0.999719,
		0.999863,
		0.999944,
		0.999982,
		0.999996,
		1,
		1,
		1,
		0.999997,
		0.999984,
		0.999952,
		0.999889,
		0.99978,
		0.99961,
		0.999365,
		0.999027,
		0.998583,
		0.998016,
		0.997312,
		0.996457,
		0.99544,
		0.994248,
		0.99287,
		0.991297,
		0.989522,
		0.987535,
		0.985332,
		0.982906,
		0.980254,
		0.977373,
		0.97426,
		0.970913,
		0.967332,
		0.963517,
		0.959469,
		0.955189,
		0.950679,
		0.945942,
		0.94098,
		0.935798,
		0.930398,
		0.924786,
		0.918966,
		0.912942,
		0.90672,
		0.900305,
		0.893703,
		0.88692,
		0.87996,
		0.872831,
		0.865538,
		0.858088,
		0.850486,
		0.842738,
		0.834852,
		0.826833,
		0.818687,
		0.810421,
		0.80204,
		0.793551,
		0.784961,
		0.776274,
		0.767497,
		0.758636,
		0.749696,
		0.740684,
		0.731605,
		0.722464,
		0.713267,
		0.704019,
		0.694725,
		0.685391,
		0.676022,
		0.666622,
		0.003111,
		0.011912,
		0.025659,
		0.043674,
		0.065341,
		0.090096,
		0.117431,
		0.146884,
		0.178037,
		0.210511,
		0.243966,
		0.278096,
		0.312626,
		0.34731,
		0.381929,
		0.416289,
		0.450219,
		0.483569,
		0.516206,
		0.548017,
		0.578906,
		0.608789,
		0.637598,
		0.665277,
		0.691782,
		0.717079,
		0.741144,
		0.763962,
		0.785527,
		0.805839,
		0.824905,
		0.842738,
		0.859359,
		0.874792,
		0.889065,
		0.902212,
		0.914268,
		0.925275,
		0.935275,
		0.944313,
		0.952436,
		0.959694,
		0.966137,
		0.971816,
		0.976785,
		0.981094,
		0.984799,
		0.987949,
		0.990599,
		0.992798,
		0.994596,
		0.996042,
		0.997182,
		0.99806,
		0.998717,
		0.999193,
		0.999524,
		0.999741,
		0.999874,
		0.999948,
		0.999984,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999985,
		0.999956,
		0.999897,
		0.999797,
		0.999639,
		0.999412,
		0.999098,
		0.998685,
		0.998157,
		0.997501,
		0.996704,
		0.995754,
		0.99464,
		0.99335,
		0.991877,
		0.990212,
		0.988347,
		0.986276,
		0.983995,
		0.981498,
		0.978782,
		0.975845,
		0.972685,
		0.969301,
		0.965692,
		0.961859,
		0.957802,
		0.953525,
		0.949027,
		0.944313,
		0.939384,
		0.934245,
		0.928899,
		0.92335,
		0.917603,
		0.911662,
		0.905532,
		0.899218,
		0.892725,
		0.886059,
		0.879226,
		0.87223,
		0.865077,
		0.857774,
		0.850326,
		0.842738,
		0.835018,
		0.827169,
		0.8192,
		0.811114,
		0.802918,
		0.794618,
		0.78622,
		0.777728,
		0.769149,
		0.760489,
		0.751752,
		0.742944,
		0.73407,
		0.725136,
		0.716147,
		0.000342,
		0.005236,
		0.015366,
		0.030049,
		0.048665,
		0.070648,
		0.095481,
		0.122695,
		0.151865,
		0.182603,
		0.214561,
		0.247423,
		0.280903,
		0.314747,
		0.348725,
		0.382633,
		0.416289,
		0.449532,
		0.482221,
		0.51423,
		0.545454,
		0.575799,
		0.605186,
		0.63355,
		0.660837,
		0.687003,
		0.712016,
		0.73585,
		0.75849,
		0.779928,
		0.800163,
		0.8192,
		0.837049,
		0.853729,
		0.869259,
		0.883666,
		0.89698,
		0.909232,
		0.920461,
		0.930704,
		0.940003,
		0.948403,
		0.955947,
		0.962684,
		0.968661,
		0.973927,
		0.978531,
		0.982523,
		0.985952,
		0.988867,
		0.991318,
		0.99335,
		0.995012,
		0.996348,
		0.9974,
		0.99821,
		0.998817,
		0.999256,
		0.999561,
		0.999762,
		0.999884,
		0.999952,
		0.999985,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999987,
		0.999959,
		0.999905,
		0.999812,
		0.999666,
		0.999454,
		0.999163,
		0.998778,
		0.998286,
		0.997675,
		0.99693,
		0.996042,
		0.994999,
		0.993791,
		0.99241,
		0.990846,
		0.989094,
		0.987147,
		0.984999,
		0.982646,
		0.980084,
		0.977312,
		0.974325,
		0.971124,
		0.967708,
		0.964076,
		0.96023,
		0.95617,
		0.951898,
		0.947416,
		0.942727,
		0.937833,
		0.932739,
		0.927446,
		0.921961,
		0.916286,
		0.910426,
		0.904386,
		0.89817,
		0.891784,
		0.885232,
		0.878519,
		0.871652,
		0.864634,
		0.857473,
		0.850172,
		0.842738,
		0.835177,
		0.827492,
		0.819691,
		0.811778,
		0.803759,
		0.79564,
		0.787425,
		0.779121,
		0.770731,
		0.762263,
		0.001294,
		0.007758,
		0.019046,
		0.034535,
		0.053654,
		0.075884,
		0.100747,
		0.127813,
		0.156687,
		0.187012,
		0.218463,
		0.250747,
		0.283601,
		0.316784,
		0.350084,
		0.383309,
		0.416289,
		0.448872,
		0.480925,
		0.512331,
		0.542987,
		0.572806,
		0.601712,
		0.629643,
		0.656545,
		0.682378,
		0.707107,
		0.730708,
		0.753166,
		0.77447,
		0.794618,
		0.813614,
		0.831466,
		0.848188,
		0.863799,
		0.878322,
		0.891784,
		0.904213,
		0.915644,
		0.926112,
		0.935655,
		0.944313,
		0.952128,
		0.959144,
		0.965405,
		0.970957,
		0.975845,
		0.980117,
		0.983819,
		0.986998,
		0.9897,
		0.991969,
		0.993851,
		0.995389,
		0.996625,
		0.997598,
		0.998347,
		0.998907,
		0.999313,
		0.999595,
		0.99978,
		0.999893,
		0.999956,
		0.999986,
		0.999997,
		1,
		1,
		1,
		0.999997,
		0.999988,
		0.999962,
		0.999912,
		0.999825,
		0.99969,
		0.999493,
		0.999222,
		0.998863,
		0.998404,
		0.997833,
		0.997137,
		0.996306,
		0.995329,
		0.994196,
		0.992899,
		0.991431,
		0.989783,
		0.987949,
		0.985926,
		0.983707,
		0.981289,
		0.978669,
		0.975845,
		0.972816,
		0.96958,
		0.966137,
		0.962487,
		0.958631,
		0.954571,
		0.950308,
		0.945845,
		0.941183,
		0.936326,
		0.931276,
		0.926039,
		0.920616,
		0.915012,
		0.909232,
		0.90328,
		0.897159,
		0.890876,
		0.884434,
		0.877839,
		0.871096,
		0.864209,
		0.857183,
		0.850025,
		0.842738,
		0.835329,
		0.827802,
		0.820162,
		0.812415,
		0.804566,
		0.002763,
		0.010606,
		0.022905,
		0.039085,
		0.058623,
		0.081039,
		0.105894,
		0.132788,
		0.161357,
		0.191269,
		0.222223,
		0.253947,
		0.286194,
		0.318742,
		0.35139,
		0.383959,
		0.416289,
		0.448238,
		0.47968,
		0.510503,
		0.540612,
		0.569922,
		0.598361,
		0.625869,
		0.652395,
		0.677898,
		0.702346,
		0.725714,
		0.747985,
		0.769149,
		0.789202,
		0.808146,
		0.825987,
		0.842738,
		0.858415,
		0.873037,
		0.886629,
		0.899218,
		0.910833,
		0.921508,
		0.931276,
		0.940176,
		0.948246,
		0.955526,
		0.962058,
		0.967884,
		0.973047,
		0.977591,
		0.981561,
		0.984999,
		0.987949,
		0.990456,
		0.992561,
		0.994306,
		0.995731,
		0.996876,
		0.997777,
		0.99847,
		0.998989,
		0.999365,
		0.999625,
		0.999797,
		0.999901,
		0.999959,
		0.999987,
		0.999997,
		1,
		1,
		1,
		0.999998,
		0.999988,
		0.999965,
		0.999918,
		0.999838,
		0.999712,
		0.999529,
		0.999276,
		0.998941,
		0.998512,
		0.997978,
		0.997327,
		0.996548,
		0.995632,
		0.994569,
		0.99335,
		0.991969,
		0.990418,
		0.988691,
		0.986782,
		0.984688,
		0.982404,
		0.979927,
		0.977255,
		0.974386,
		0.971319,
		0.968053,
		0.964588,
		0.960924,
		0.957064,
		0.953007,
		0.948756,
		0.944313,
		0.93968,
		0.93486,
		0.929857,
		0.924674,
		0.919314,
		0.913781,
		0.908079,
		0.902212,
		0.896184,
		0.890001,
		0.883666,
		0.877185,
		0.870561,
		0.863799,
		0.856905,
		0.849883,
		0.842738,
		0.000304,
		0.004665,
		0.013723,
		0.026902,
		0.043674,
		0.063557,
		0.086107,
		0.110919,
		0.137622,
		0.165879,
		0.195381,
		0.225849,
		0.257029,
		0.28869,
		0.320624,
		0.352645,
		0.384583,
		0.416289,
		0.447628,
		0.478481,
		0.508744,
		0.538323,
		0.56714,
		0.595126,
		0.622222,
		0.64838,
		0.673559,
		0.697728,
		0.720862,
		0.742944,
		0.763962,
		0.783913,
		0.802795,
		0.820615,
		0.837381,
		0.853109,
		0.867815,
		0.881522,
		0.894253,
		0.906036,
		0.9169,
		0.926878,
		0.936004,
		0.944313,
		0.951842,
		0.958631,
		0.96472,
		0.970147,
		0.974954,
		0.979184,
		0.982876,
		0.986073,
		0.988816,
		0.991144,
		0.993099,
		0.994719,
		0.996042,
		0.997104,
		0.99794,
		0.998583,
		0.999064,
		0.999412,
		0.999653,
		0.999812,
		0.999909,
		0.999962,
		0.999988,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999989,
		0.999968,
		0.999924,
		0.999849,
		0.999732,
		0.999561,
		0.999325,
		0.999012,
		0.998612,
		0.998111,
		0.997501,
		0.996771,
		0.995911,
		0.994912,
		0.993766,
		0.992466,
		0.991005,
		0.989376,
		0.987575,
		0.985597,
		0.983438,
		0.981094,
		0.978564,
		0.975845,
		0.972936,
		0.969836,
		0.966545,
		0.963062,
		0.959389,
		0.955526,
		0.951476,
		0.947239,
		0.942819,
		0.938217,
		0.933436,
		0.928479,
		0.92335,
		0.918052,
		0.912588,
		0.906963,
		0.90118,
		0.895243,
		0.889157,
		0.882926,
		0.876554,
		0.001154,
		0.006932,
		0.017058,
		0.031004,
		0.048281,
		0.068445,
		0.091084,
		0.115823,
		0.14232,
		0.170259,
		0.199356,
		0.229348,
		0.259999,
		0.291093,
		0.322436,
		0.353853,
		0.385184,
		0.416289,
		0.447041,
		0.477327,
		0.507048,
		0.536117,
		0.564456,
		0.592002,
		0.618696,
		0.644493,
		0.669354,
		0.693246,
		0.716147,
		0.738037,
		0.758906,
		0.778747,
		0.797559,
		0.815347,
		0.832117,
		0.847883,
		0.862659,
		0.876466,
		0.889324,
		0.901258,
		0.912296,
		0.922467,
		0.931803,
		0.940336,
		0.948101,
		0.955135,
		0.961473,
		0.967154,
		0.972215,
		0.976697,
		0.980638,
		0.984077,
		0.987053,
		0.989606,
		0.991772,
		0.99359,
		0.995096,
		0.996325,
		0.997312,
		0.998088,
		0.998685,
		0.999131,
		0.999454,
		0.999678,
		0.999825,
		0.999915,
		0.999965,
		0.999989,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.99999,
		0.99997,
		0.999929,
		0.99986,
		0.99975,
		0.999591,
		0.99937,
		0.999078,
		0.998703,
		0.998234,
		0.997662,
		0.996976,
		0.996168,
		0.995229,
		0.99415,
		0.992925,
		0.991548,
		0.990011,
		0.98831,
		0.98644,
		0.984397,
		0.982179,
		0.979781,
		0.977203,
		0.974442,
		0.971498,
		0.968369,
		0.965057,
		0.96156,
		0.957881,
		0.95402,
		0.949978,
		0.945759,
		0.941362,
		0.936792,
		0.932051,
		0.927141,
		0.922066,
		0.916829,
		0.911434,
		0.905884,
		0.00247,
		0.009504,
		0.02057,
		0.035182,
		0.052888,
		0.073275,
		0.095965,
		0.120607,
		0.146884,
		0.174504,
		0.203198,
		0.232725,
		0.262862,
		0.293408,
		0.324181,
		0.355015,
		0.385763,
		0.416289,
		0.446476,
		0.476215,
		0.505413,
		0.533987,
		0.561865,
		0.588982,
		0.615286,
		0.64073,
		0.665277,
		0.688896,
		0.711563,
		0.73326,
		0.753975,
		0.773702,
		0.792437,
		0.810183,
		0.826947,
		0.842738,
		0.857572,
		0.871464,
		0.884434,
		0.896506,
		0.907703,
		0.918052,
		0.927583,
		0.936326,
		0.944313,
		0.951577,
		0.958153,
		0.964076,
		0.969383,
		0.974109,
		0.978292,
		0.981968,
		0.985175,
		0.987949,
		0.990328,
		0.992346,
		0.994038,
		0.99544,
		0.996584,
		0.997501,
		0.998223,
		0.998778,
		0.999193,
		0.999493,
		0.999701,
		0.999838,
		0.999921,
		0.999968,
		0.99999,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999991,
		0.999972,
		0.999934,
		0.999869,
		0.999767,
		0.999618,
		0.999412,
		0.999138,
		0.998786,
		0.998347,
		0.997809,
		0.997165,
		0.996405,
		0.995521,
		0.994505,
		0.99335,
		0.99205,
		0.990599,
		0.988991,
		0.987222,
		0.985289,
		0.983187,
		0.980913,
		0.978467,
		0.975845,
		0.973047,
		0.970072,
		0.96692,
		0.96359,
		0.960083,
		0.956401,
		0.952544,
		0.948514,
		0.944313,
		0.939942,
		0.935405,
		0.930704,
		0.000272,
		0.004182,
		0.012329,
		0.024223,
		0.039412,
		0.057479,
		0.078042,
		0.100747,
		0.125272,
		0.151319,
		0.178617,
		0.206915,
		0.235987,
		0.265625,
		0.295641,
		0.325863,
		0.356136,
		0.38632,
		0.416289,
		0.44593,
		0.475143,
		0.503836,
		0.531932,
		0.559361,
		0.586062,
		0.611985,
		0.637084,
		0.661323,
		0.684672,
		0.707107,
		0.72861,
		0.749168,
		0.768774,
		0.787425,
		0.805122,
		0.821869,
		0.837676,
		0.852555,
		0.86652,
		0.879589,
		0.891784,
		0.903125,
		0.913638,
		0.92335,
		0.932289,
		0.940484,
		0.947967,
		0.954769,
		0.960924,
		0.966466,
		0.971428,
		0.975845,
		0.979753,
		0.983187,
		0.986181,
		0.98877,
		0.990988,
		0.99287,
		0.994448,
		0.995754,
		0.99682,
		0.997675,
		0.998347,
		0.998863,
		0.999249,
		0.999529,
		0.999722,
		0.999849,
		0.999927,
		0.99997,
		0.99999,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999991,
		0.999974,
		0.999939,
		0.999878,
		0.999782,
		0.999643,
		0.99945,
		0.999193,
		0.998863,
		0.998451,
		0.997946,
		0.99734,
		0.996625,
		0.995792,
		0.994834,
		0.993744,
		0.992516,
		0.991144,
		0.989624,
		0.987949,
		0.986118,
		0.984125,
		0.981968,
		0.979645,
		0.977154,
		0.974494,
		0.971663,
		0.968661,
		0.965488,
		0.962145,
		0.958631,
		0.954949,
		0.951098,
		0.001035,
		0.006231,
		0.015366,
		0.027987,
		0.043674,
		0.062044,
		0.082738,
		0.105431,
		0.12982,
		0.155629,
		0.182603,
		0.210511,
		0.239139,
		0.268292,
		0.297795,
		0.327485,
		0.357216,
		0.386858,
		0.416289,
		0.445405,
		0.474108,
		0.502313,
		0.529947,
		0.556941,
		0.583238,
		0.608789,
		0.63355,
		0.657487,
		0.680569,
		0.702773,
		0.724081,
		0.74448,
		0.763962,
		0.782523,
		0.800163,
		0.816885,
		0.832697,
		0.84761,
		0.861636,
		0.874792,
		0.887096,
		0.898568,
		0.909232,
		0.919112,
		0.928233,
		0.936624,
		0.944313,
		0.95133,
		0.957705,
		0.963472,
		0.968661,
		0.973306,
		0.977439,
		0.981094,
		0.984305,
		0.987103,
		0.989522,
		0.991594,
		0.99335,
		0.994823,
		0.996042,
		0.997036,
		0.997833,
		0.99846,
		0.998941,
		0.999301,
		0.999561,
		0.999741,
		0.99986,
		0.999932,
		0.999972,
		0.999991,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999992,
		0.999976,
		0.999943,
		0.999886,
		0.999797,
		0.999666,
		0.999485,
		0.999244,
		0.998934,
		0.998546,
		0.998072,
		0.997501,
		0.996827,
		0.996042,
		0.995138,
		0.994109,
		0.992949,
		0.991651,
		0.990212,
		0.988626,
		0.986889,
		0.984999,
		0.982951,
		0.980745,
		0.978376,
		0.975845,
		0.97315,
		0.970291,
		0.967266,
		0.002221,
		0.008564,
		0.018575,
		0.031834,
		0.047952,
		0.066571,
		0.08736,
		0.110014,
		0.134253,
		0.159817,
		0.186469,
		0.213992,
		0.242186,
		0.270869,
		0.299874,
		0.32905,
		0.358259,
		0.387376,
		0.416289,
		0.444897,
		0.473108,
		0.500843,
		0.528027,
		0.5546,
		0.580504,
		0.605693,
		0.630124,
		0.653763,
		0.676582,
		0.698557,
		0.71967,
		0.739908,
		0.759263,
		0.777728,
		0.795304,
		0.811993,
		0.827802,
		0.842738,
		0.856815,
		0.870045,
		0.882447,
		0.894038,
		0.904839,
		0.914874,
		0.924165,
		0.932739,
		0.940621,
		0.947842,
		0.954428,
		0.960409,
		0.965817,
		0.970681,
		0.975034,
		0.978905,
		0.982327,
		0.985332,
		0.987949,
		0.990212,
		0.992149,
		0.993791,
		0.995167,
		0.996306,
		0.997234,
		0.997978,
		0.998563,
		0.999012,
		0.999348,
		0.999591,
		0.999759,
		0.999869,
		0.999937,
		0.999974,
		0.999992,
		0.999998,
		1,
		1,
		1,
		0.999998,
		0.999993,
		0.999977,
		0.999947,
		0.999893,
		0.99981,
		0.999687,
		0.999517,
		0.999291,
		0.999,
		0.998635,
		0.998188,
		0.997651,
		0.997015,
		0.996274,
		0.995421,
		0.994448,
		0.99335,
		0.992122,
		0.990759,
		0.989256,
		0.987608,
		0.985814,
		0.983869,
		0.981771,
		0.979518,
		0.000244,
		0.003771,
		0.011138,
		0.021925,
		0.035743,
		0.05223,
		0.071053,
		0.091904,
		0.114498,
		0.138573,
		0.163887,
		0.190218,
		0.217363,
		0.245133,
		0.273359,
		0.301882,
		0.330561,
		0.359265,
		0.387877,
		0.416289,
		0.444407,
		0.472143,
		0.499421,
		0.526172,
		0.552335,
		0.577857,
		0.602692,
		0.6268,
		0.650148,
		0.672708,
		0.694455,
		0.715374,
		0.735449,
		0.754672,
		0.773038,
		0.790544,
		0.807193,
		0.82299,
		0.837941,
		0.852057,
		0.865352,
		0.877839,
		0.889537,
		0.900464,
		0.910641,
		0.92009,
		0.928835,
		0.9369,
		0.944313,
		0.951098,
		0.957286,
		0.962903,
		0.967979,
		0.972543,
		0.976625,
		0.980254,
		0.983462,
		0.986276,
		0.988728,
		0.990846,
		0.99266,
		0.994196,
		0.995483,
		0.996548,
		0.997416,
		0.998111,
		0.998658,
		0.999078,
		0.999391,
		0.999618,
		0.999775,
		0.999878,
		0.999941,
		0.999976,
		0.999992,
		0.999998,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.999979,
		0.99995,
		0.9999,
		0.999822,
		0.999707,
		0.999547,
		0.999334,
		0.999061,
		0.998717,
		0.998296,
		0.997789,
		0.997189,
		0.99649,
		0.995683,
		0.994763,
		0.993724,
		0.992561,
		0.991269,
		0.989843,
		0.988279,
		0.000934,
		0.005631,
		0.013913,
		0.025389,
		0.039695,
		0.056497,
		0.075484,
		0.096367,
		0.118882,
		0.142782,
		0.167843,
		0.193856,
		0.220628,
		0.247986,
		0.275766,
		0.303823,
		0.332021,
		0.360237,
		0.38836,
		0.416289,
		0.443933,
		0.47121,
		0.498046,
		0.524376,
		0.550142,
		0.575292,
		0.599783,
		0.623575,
		0.646637,
		0.66894,
		0.690463,
		0.711187,
		0.731099,
		0.750188,
		0.76845,
		0.785882,
		0.802484,
		0.81826,
		0.833218,
		0.847365,
		0.860713,
		0.873276,
		0.88507,
		0.896111,
		0.906419,
		0.916014,
		0.924919,
		0.933156,
		0.940749,
		0.947725,
		0.954108,
		0.959925,
		0.965204,
		0.969973,
		0.97426,
		0.978092,
		0.981498,
		0.984507,
		0.987147,
		0.989445,
		0.991431,
		0.99313,
		0.994569,
		0.995774,
		0.996771,
		0.997583,
		0.998234,
		0.998745,
		0.999138,
		0.999431,
		0.999643,
		0.99979,
		0.999886,
		0.999945,
		0.999977,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999993,
		0.99998,
		0.999953,
		0.999906,
		0.999833,
		0.999725,
		0.999575,
		0.999375,
		0.999117,
		0.998793,
		0.998396,
		0.997917,
		0.997351,
		0.99669,
		0.995927,
		0.995056,
		0.994072,
		0.002008,
		0.007758,
		0.016857,
		0.028941,
		0.043674,
		0.060743,
		0.079857,
		0.100747,
		0.123166,
		0.146884,
		0.171689,
		0.197385,
		0.223793,
		0.250747,
		0.278096,
		0.3057,
		0.333432,
		0.361176,
		0.388827,
		0.416289,
		0.443475,
		0.470308,
		0.496716,
		0.522638,
		0.548017,
		0.572806,
		0.596961,
		0.620445,
		0.643226,
		0.665277,
		0.686577,
		0.707107,
		0.726854,
		0.745808,
		0.763962,
		0.781314,
		0.797864,
		0.813614,
		0.828569,
		0.842738,
		0.856131,
		0.86876,
		0.880639,
		0.891784,
		0.902212,
		0.911942,
		0.920996,
		0.929393,
		0.937158,
		0.944313,
		0.950882,
		0.956891,
		0.962366,
		0.967332,
		0.971816,
		0.975845,
		0.979446,
		0.982646,
		0.985471,
		0.987949,
		0.990107,
		0.991969,
		0.993563,
		0.994912,
		0.996042,
		0.996976,
		0.997737,
		0.998347,
		0.998826,
		0.999193,
		0.999468,
		0.999666,
		0.999803,
		0.999893,
		0.999948,
		0.999979,
		0.999993,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999981,
		0.999956,
		0.999912,
		0.999843,
		0.999741,
		0.9996,
		0.999412,
		0.999169,
		0.998863,
		0.998488,
		0.998037,
		0.997501,
		0.000221,
		0.003417,
		0.010111,
		0.019939,
		0.032563,
		0.047667,
		0.064959,
		0.084169,
		0.105044,
		0.127354,
		0.150882,
		0.175429,
		0.200812,
		0.226862,
		0.253422,
		0.280351,
		0.307516,
		0.334797,
		0.362085,
		0.389279,
		0.416289,
		0.443032,
		0.469434,
		0.495428,
		0.520954,
		0.545959,
		0.570395,
		0.594222,
		0.617404,
		0.63991,
		0.661713,
		0.682792,
		0.70313,
		0.722712,
		0.741528,
		0.759572,
		0.776841,
		0.793332,
		0.809049,
		0.823996,
		0.838179,
		0.851607,
		0.864293,
		0.876247,
		0.887485,
		0.898023,
		0.907879,
		0.917071,
		0.925619,
		0.933544,
		0.940869,
		0.947615,
		0.953807,
		0.959469,
		0.964625,
		0.969301,
		0.973521,
		0.977312,
		0.980698,
		0.983707,
		0.986362,
		0.988691,
		0.990717,
		0.992466,
		0.993962,
		0.995229,
		0.996289,
		0.997165,
		0.997879,
		0.998451,
		0.9989,
		0.999244,
		0.999501,
		0.999687,
		0.999816,
		0.9999,
		0.999952,
		0.99998,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999959,
		0.999917,
		0.999852,
		0.999757,
		0.999624,
		0.999446,
		0.999217,
		0.000846,
		0.005114,
		0.012657,
		0.023136,
		0.036235,
		0.05166,
		0.069138,
		0.088416,
		0.109257,
		0.131445,
		0.154777,
		0.179066,
		0.204139,
		0.229837,
		0.256014,
		0.282534,
		0.309274,
		0.336118,
		0.362964,
		0.389717,
		0.416289,
		0.442603,
		0.468589,
		0.49418,
		0.519322,
		0.543963,
		0.568056,
		0.591564,
		0.61445,
		0.636686,
		0.658245,
		0.679106,
		0.699252,
		0.718669,
		0.737347,
		0.755278,
		0.772459,
		0.788888,
		0.804566,
		0.819497,
		0.833687,
		0.847143,
		0.859876,
		0.871897,
		0.883219,
		0.893857,
		0.903828,
		0.913148,
		0.921837,
		0.929913,
		0.937398,
		0.944313,
		0.950679,
		0.95652,
		0.961859,
		0.966719,
		0.971124,
		0.9751,
		0.978669,
		0.981857,
		0.984688,
		0.987186,
		0.989376,
		0.991281,
		0.992925,
		0.994331,
		0.995521,
		0.996517,
		0.99734,
		0.99801,
		0.998546,
		0.998968,
		0.999291,
		0.999532,
		0.999707,
		0.999827,
		0.999906,
		0.999955,
		0.999981,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999995,
		0.999984,
		0.999961,
		0.999922,
		0.999861,
		0.001825,
		0.00706,
		0.015366,
		0.026425,
		0.039943,
		0.055645,
		0.073275,
		0.092596,
		0.113386,
		0.135442,
		0.158573,
		0.182603,
		0.207371,
		0.232725,
		0.258527,
		0.28465,
		0.310976,
		0.337397,
		0.363815,
		0.39014,
		0.416289,
		0.442188,
		0.467769,
		0.492972,
		0.51774,
		0.542026,
		0.565786,
		0.588982,
		0.61158,
		0.63355,
		0.654869,
		0.675515,
		0.69547,
		0.714722,
		0.73326,
		0.751077,
		0.768167,
		0.784529,
		0.800163,
		0.815072,
		0.829262,
		0.842738,
		0.855511,
		0.86759,
		0.878987,
		0.889716,
		0.899793,
		0.909232,
		0.918052,
		0.92627,
		0.933906,
		0.94098,
		0.947512,
		0.953525,
		0.959038,
		0.964076,
		0.968661,
		0.972816,
		0.976563,
		0.979927,
		0.982931,
		0.985597,
		0.987949,
		0.990011,
		0.991804,
		0.99335,
		0.994673,
		0.995792,
		0.996728,
		0.997501,
		0.998131,
		0.998635,
		0.999031,
		0.999334,
		0.999561,
		0.999725,
		0.999838,
		0.999912,
		0.999957,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999995,
		0.000201,
		0.003111,
		0.00922,
		0.018211,
		0.029788,
		0.043674,
		0.059613,
		0.077365,
		0.096707,
		0.117431,
		0.139346,
		0.162273,
		0.186045,
		0.210511,
		0.235528,
		0.260965,
		0.286701,
		0.312626,
		0.338636,
		0.36464,
		0.39055,
		0.416289,
		0.441786,
		0.466975,
		0.4918,
		0.516206,
		0.540147,
		0.563582,
		0.586474,
		0.608789,
		0.630499,
		0.651582,
		0.672015,
		0.691782,
		0.710869,
		0.729267,
		0.746966,
		0.763962,
		0.780253,
		0.795839,
		0.810721,
		0.824905,
		0.838394,
		0.851199,
		0.863328,
		0.874792,
		0.885604,
		0.895777,
		0.905327,
		0.914268,
		0.92262,
		0.930398,
		0.937623,
		0.944313,
		0.950488,
		0.95617,
		0.961379,
		0.966137,
		0.970465,
		0.974386,
		0.977922,
		0.981094,
		0.983926,
		0.98644,
		0.988657,
		0.990599,
		0.992288,
		0.993744,
		0.994989,
		0.996042,
		0.996923,
		0.997651,
		0.998243,
		0.998717,
		0.999089,
		0.999375,
		0.999588,
		0.999741,
		0.999848,
		0.999917,
		0.99996,
		0.999984,
		0.999995,
		0.999999,
		1,
		0.000771,
		0.004665,
		0.011564,
		0.02117,
		0.033208,
		0.047417,
		0.063557,
		0.081404,
		0.100747,
		0.121393,
		0.14316,
		0.165879,
		0.189395,
		0.213563,
		0.238249,
		0.263329,
		0.28869,
		0.314225,
		0.339837,
		0.365439,
		0.390948,
		0.416289,
		0.441396,
		0.466205,
		0.490663,
		0.514717,
		0.538323,
		0.561441,
		0.584036,
		0.606074,
		0.62753,
		0.64838,
		0.668603,
		0.688183,
		0.707107,
		0.725363,
		0.742944,
		0.759844,
		0.776061,
		0.791594,
		0.806444,
		0.820615,
		0.834112,
		0.846941,
		0.859113,
		0.870636,
		0.881522,
		0.891784,
		0.901435,
		0.91049,
		0.918966,
		0.926878,
		0.934245,
		0.941085,
		0.947416,
		0.953258,
		0.958631,
		0.963556,
		0.968053,
		0.972142,
		0.975845,
		0.979184,
		0.982179,
		0.984851,
		0.987222,
		0.989313,
		0.991144,
		0.992736,
		0.994109,
		0.995282,
		0.996274,
		0.997104,
		0.997789,
		0.998347,
		0.998793,
		0.999143,
		0.999412,
		0.999612,
		0.999757,
		0.999857,
		0.999922,
		0.999962,
		0.001665,
		0.006452,
		0.014064,
		0.024223,
		0.03667,
		0.051161,
		0.067471,
		0.085389,
		0.104717,
		0.125272,
		0.146884,
		0.169394,
		0.192655,
		0.21653,
		0.240892,
		0.265625,
		0.290619,
		0.315776,
		0.341002,
		0.366214,
		0.391333,
		0.416289,
		0.441017,
		0.465459,
		0.489559,
		0.513271,
		0.536552,
		0.559361,
		0.581665,
		0.603434,
		0.62464,
		0.645261,
		0.665277,
		0.684672,
		0.703432,
		0.721547,
		0.739008,
		0.75581,
		0.771949,
		0.787425,
		0.802239,
		0.816392,
		0.82989,
		0.842738,
		0.854945,
		0.86652,
		0.877473,
		0.887815,
		0.897559,
		0.90672,
		0.915312,
		0.92335,
		0.930852,
		0.937833,
		0.944313,
		0.950308,
		0.955839,
		0.960924,
		0.965583,
		0.969836,
		0.973702,
		0.977203,
		0.980357,
		0.983187,
		0.985711,
		0.987949,
		0.989923,
		0.991651,
		0.993153,
		0.994448,
		0.995554,
		0.99649,
		0.997272,
		0.997917,
		0.998443,
		0.998863,
		0.999193,
		0.999446,
		0.999635,
		0.000183,
		0.002844,
		0.008442,
		0.016698,
		0.027353,
		0.040162,
		0.054899,
		0.07135,
		0.089317,
		0.108615,
		0.12907,
		0.150522,
		0.172822,
		0.19583,
		0.219416,
		0.243461,
		0.267854,
		0.292492,
		0.317281,
		0.342132,
		0.366965,
		0.391707,
		0.416289,
		0.44065,
		0.464734,
		0.488488,
		0.511868,
		0.53483,
		0.557338,
		0.579359,
		0.600863,
		0.621825,
		0.642221,
		0.662033,
		0.681244,
		0.699842,
		0.717815,
		0.735156,
		0.751857,
		0.767917,
		0.783333,
		0.798105,
		0.812236,
		0.825729,
		0.838591,
		0.850827,
		0.862446,
		0.873458,
		0.883873,
		0.893703,
		0.902962,
		0.911662,
		0.919818,
		0.927446,
		0.934562,
		0.941183,
		0.947325,
		0.953007,
		0.958246,
		0.963062,
		0.967473,
		0.971498,
		0.975156,
		0.978467,
		0.98145,
		0.984125,
		0.98651,
		0.988626,
		0.99049,
		0.992122,
		0.993541,
		0.994763,
		0.995807,
		0.99669,
		0.997428,
		0.998037,
		0.998532,
		0.000705,
		0.004273,
		0.010606,
		0.019444,
		0.030544,
		0.043674,
		0.058623,
		0.07519,
		0.093187,
		0.112441,
		0.132788,
		0.154076,
		0.176164,
		0.198921,
		0.222223,
		0.245958,
		0.27002,
		0.294311,
		0.318742,
		0.343228,
		0.367695,
		0.39207,
		0.416289,
		0.440294,
		0.46403,
		0.487448,
		0.510503,
		0.533157,
		0.555372,
		0.577116,
		0.598361,
		0.619082,
		0.639258,
		0.658868,
		0.677898,
		0.696335,
		0.714166,
		0.731385,
		0.747985,
		0.763962,
		0.779315,
		0.794042,
		0.808146,
		0.82163,
		0.834498,
		0.846758,
		0.858415,
		0.869479,
		0.87996,
		0.889869,
		0.899218,
		0.908019,
		0.916286,
		0.924033,
		0.931276,
		0.938031,
		0.944313,
		0.950139,
		0.955526,
		0.960493,
		0.965057,
		0.969235,
		0.973047,
		0.976511,
		0.979645,
		0.982468,
		0.984999,
		0.987255,
		0.989256,
		0.991018,
		0.992561,
		0.993901,
		0.995056,
		0.996042,
		0.001526,
		0.00592,
		0.012921,
		0.022285,
		0.033782,
		0.047197,
		0.062328,
		0.078987,
		0.096998,
		0.116196,
		0.136427,
		0.157547,
		0.179424,
		0.201932,
		0.224955,
		0.248386,
		0.272124,
		0.296078,
		0.32016,
		0.344293,
		0.368403,
		0.392422,
		0.416289,
		0.439948,
		0.463346,
		0.486437,
		0.509177,
		0.531529,
		0.553458,
		0.574932,
		0.595924,
		0.61641,
		0.636368,
		0.655781,
		0.674631,
		0.692907,
		0.710598,
		0.727694,
		0.744191,
		0.760084,
		0.77537,
		0.790049,
		0.804122,
		0.817592,
		0.830462,
		0.842738,
		0.854428,
		0.865538,
		0.876079,
		0.886059,
		0.895491,
		0.904386,
		0.912756,
		0.920616,
		0.927979,
		0.93486,
		0.941275,
		0.947239,
		0.952769,
		0.957881,
		0.962592,
		0.96692,
		0.970881,
		0.974494,
		0.977776,
		0.980745,
		0.983418,
		0.985814,
		0.987949,
		0.989843,
		0.99151,
		0.000168,
		0.00261,
		0.007758,
		0.015366,
		0.025204,
		0.037057,
		0.050721,
		0.066007,
		0.082738,
		0.100747,
		0.119879,
		0.139988,
		0.160938,
		0.182603,
		0.204865,
		0.227614,
		0.250747,
		0.27417,
		0.297795,
		0.321539,
		0.345327,
		0.369091,
		0.392764,
		0.416289,
		0.439612,
		0.462682,
		0.485454,
		0.507888,
		0.529947,
		0.551596,
		0.572806,
		0.59355,
		0.613805,
		0.63355,
		0.652767,
		0.67144,
		0.689557,
		0.707107,
		0.724081,
		0.740474,
		0.75628,
		0.771497,
		0.786125,
		0.800163,
		0.813614,
		0.826481,
		0.83877,
		0.850486,
		0.861636,
		0.87223,
		0.882275,
		0.891784,
		0.900765,
		0.909232,
		0.917197,
		0.924674,
		0.931675,
		0.938217,
		0.944313,
		0.949978,
		0.95523,
		0.960083,
		0.964555,
		0.968661,
		0.972419,
		0.975845,
		0.978957,
		0.981771,
		0.984305,
		0.000647,
		0.003928,
		0.009762,
		0.017921,
		0.028187,
		0.040357,
		0.05424,
		0.069658,
		0.086442,
		0.104436,
		0.123492,
		0.143473,
		0.164251,
		0.185705,
		0.207724,
		0.230203,
		0.253045,
		0.27616,
		0.299464,
		0.322879,
		0.346332,
		0.369759,
		0.393097,
		0.416289,
		0.439285,
		0.462036,
		0.484498,
		0.506634,
		0.528406,
		0.549783,
		0.570735,
		0.591237,
		0.611266,
		0.630801,
		0.649825,
		0.668323,
		0.686282,
		0.703692,
		0.720543,
		0.736831,
		0.752549,
		0.767695,
		0.782268,
		0.796268,
		0.809696,
		0.822556,
		0.834852,
		0.846589,
		0.857774,
		0.868415,
		0.878519,
		0.888097,
		0.897159,
		0.905717,
		0.913781,
		0.921364,
		0.928479,
		0.935141,
		0.941362,
		0.947158,
		0.952544,
		0.957534,
		0.962145,
		0.966392,
		0.970291,
		0.973858,
		0.001403,
		0.00545,
		0.011912,
		0.02057,
		0.031222,
		0.043674,
		0.057749,
		0.073275,
		0.090096,
		0.108062,
		0.127035,
		0.146884,
		0.167488,
		0.188732,
		0.210511,
		0.232725,
		0.255282,
		0.278096,
		0.301087,
		0.324181,
		0.34731,
		0.370409,
		0.39342,
		0.416289,
		0.438967,
		0.461407,
		0.483569,
		0.505413,
		0.526907,
		0.548017,
		0.568718,
		0.588982,
		0.608789,
		0.628118,
		0.646952,
		0.665277,
		0.68308,
		0.70035,
		0.717079,
		0.73326,
		0.748889,
		0.763962,
		0.778478,
		0.792437,
		0.805839,
		0.818687,
		0.830985,
		0.842738,
		0.853952,
		0.864634,
		0.874792,
		0.884434,
		0.893571,
		0.902212,
		0.910368,
		0.918052,
		0.925275,
		0.932051,
		0.938392,
		0.944313,
		0.949827,
		0.954949,
		0.959694,
		0.000155,
		0.002404,
		0.007154,
		0.014187,
		0.023299,
		0.034297,
		0.047001,
		0.061241,
		0.076857,
		0.093699,
		0.111627,
		0.130509,
		0.150223,
		0.170651,
		0.191686,
		0.213228,
		0.235182,
		0.25746,
		0.27998,
		0.302666,
		0.325448,
		0.34826,
		0.371041,
		0.393734,
		0.416289,
		0.438658,
		0.460796,
		0.482664,
		0.504225,
		0.525446,
		0.546297,
		0.566751,
		0.586783,
		0.606372,
		0.625499,
		0.644146,
		0.6623,
		0.679948,
		0.697079,
		0.713686,
		0.729761,
		0.745299,
		0.760297,
		0.774754,
		0.788668,
		0.80204,
		0.814873,
		0.827169,
		0.838934,
		0.850172,
		0.86089,
		0.871096,
		0.880796,
		0.890001,
		0.89872,
		0.906963,
		0.914741,
		0.922066,
		0.92895,
		0.935405,
		0.941445,
		0.000596,
		0.003623,
		0.009016,
		0.01657,
		0.026093,
		0.037403,
		0.05033,
		0.064712,
		0.080399,
		0.09725,
		0.11513,
		0.133916,
		0.15349,
		0.173742,
		0.19457,
		0.215878,
		0.237576,
		0.259581,
		0.281814,
		0.304203,
		0.326681,
		0.349184,
		0.371655,
		0.39404,
		0.416289,
		0.438357,
		0.460201,
		0.481783,
		0.503068,
		0.524024,
		0.544621,
		0.564834,
		0.584639,
		0.604014,
		0.622942,
		0.641405,
		0.65939,
		0.676885,
		0.693878,
		0.710363,
		0.726331,
		0.741777,
		0.756699,
		0.771094,
		0.784961,
		0.7983,
		0.811114,
		0.823405,
		0.835177,
		0.846434,
		0.857183,
		0.867431,
		0.877185,
		0.886452,
		0.895243,
		0.903567,
		0.911434,
		0.918855,
		0.001294,
		0.005035,
		0.011016,
		0.019046,
		0.028941,
		0.040531,
		0.053654,
		0.068159,
		0.083901,
		0.100747,
		0.118572,
		0.137255,
		0.156687,
		0.176763,
		0.197385,
		0.218463,
		0.23991,
		0.261647,
		0.283601,
		0.3057,
		0.327881,
		0.350084,
		0.372254,
		0.394338,
		0.416289,
		0.438064,
		0.459622,
		0.480925,
		0.501941,
		0.522638,
		0.542987,
		0.562964,
		0.582546,
		0.601712,
		0.620445,
		0.638727,
		0.656545,
		0.673888,
		0.690745,
		0.707107,
		0.722968,
		0.738322,
		0.753166,
		0.767497,
		0.781314,
		0.794618,
		0.80741,
		0.819691,
		0.831466,
		0.842738,
		0.853514,
		0.863799,
		0.873601,
		0.882926,
		0.891784,
		0.000143,
		0.002221,
		0.006618,
		0.013138,
		0.021601,
		0.031834,
		0.043674,
		0.05697,
		0.071577,
		0.08736,
		0.104192,
		0.121952,
		0.140529,
		0.159817,
		0.179716,
		0.200135,
		0.220985,
		0.242186,
		0.263662,
		0.285341,
		0.307157,
		0.32905,
		0.35096,
		0.372836,
		0.394628,
		0.416289,
		0.437779,
		0.459058,
		0.48009,
		0.500843,
		0.521286,
		0.541394,
		0.561141,
		0.580504,
		0.599465,
		0.618005,
		0.636109,
		0.653763,
		0.670956,
		0.687676,
		0.703917,
		0.71967,
		0.734932,
		0.749696,
		0.763962,
		0.777728,
		0.790993,
		0.803759,
		0.816028,
		0.827802,
		0.839085,
		0.849883,
		0.860201,
		0.000551,
		0.003352,
		0.008351,
		0.015366,
		0.024223,
		0.034761,
		0.046826,
		0.060273,
		0.074965,
		0.090775,
		0.107582,
		0.125272,
		0.143738,
		0.16288,
		0.182603,
		0.20282,
		0.223446,
		0.244406,
		0.265625,
		0.287037,
		0.308577,
		0.330188,
		0.351814,
		0.373403,
		0.39491,
		0.416289,
		0.437501,
		0.458508,
		0.479275,
		0.499772,
		0.519969,
		0.53984,
		0.559361,
		0.578511,
		0.59727,
		0.615622,
		0.63355,
		0.651042,
		0.668086,
		0.684672,
		0.700791,
		0.716437,
		0.731605,
		0.746289,
		0.760489,
		0.774201,
		0.787425,
		0.800163,
		0.812415,
		0.824185,
		0.001198,
		0.004665,
		0.010218,
		0.017685,
		0.026902,
		0.037715,
		0.049979,
		0.063557,
		0.07832,
		0.094146,
		0.110919,
		0.128533,
		0.146884,
		0.165879,
		0.185427,
		0.205443,
		0.225849,
		0.246571,
		0.26754,
		0.28869,
		0.309961,
		0.331297,
		0.352645,
		0.373956,
		0.395185,
		0.416289,
		0.43723,
		0.457972,
		0.478481,
		0.498728,
		0.518684,
		0.538323,
		0.557624,
		0.576564,
		0.595126,
		0.613292,
		0.631048,
		0.64838,
		0.665277,
		0.681729,
		0.697728,
		0.713267,
		0.72834,
		0.742944,
		0.757075,
		0.770731,
		0.783913,
		0.000132,
		0.002059,
		0.00614,
		0.012202,
		0.020082,
		0.029626,
		0.040688,
		0.05313,
		0.066821,
		0.08164,
		0.09747,
		0.114202,
		0.131734,
		0.149968,
		0.168815,
		0.188187,
		0.208006,
		0.228195,
		0.248685,
		0.269407,
		0.290302,
		0.31131,
		0.332378,
		0.353455,
		0.374495,
		0.395453,
		0.416289,
		0.436966,
		0.45745,
		0.477707,
		0.497709,
		0.51743,
		0.536843,
		0.555928,
		0.574663,
		0.593031,
		0.611015,
		0.628601,
		0.645775,
		0.662527,
		0.678846,
		0.694725,
		0.710157,
		0.725136,
		0.739658,
		0.000511,
		0.003111,
		0.007758,
		0.014288,
		0.022547,
		0.032389,
		0.043674,
		0.056273,
		0.070061,
		0.084923,
		0.100747,
		0.117431,
		0.134877,
		0.152992,
		0.171689,
		0.190888,
		0.210511,
		0.230487,
		0.250747,
		0.27123,
		0.291874,
		0.312626,
		0.333432,
		0.354245,
		0.37502,
		0.395714,
		0.416289,
		0.436709,
		0.45694,
		0.476952,
		0.496716,
		0.516206,
		0.535398,
		0.554272,
		0.572806,
		0.590984,
		0.608789,
		0.626207,
		0.643226,
		0.659834,
		0.676022,
		0.691782,
		0.001112,
		0.004334,
		0.009504,
		0.016464,
		0.02507,
		0.035182,
		0.046668,
		0.059405,
		0.073275,
		0.088168,
		0.103978,
		0.120607,
		0.137962,
		0.155955,
		0.174504,
		0.193529,
		0.212959,
		0.232725,
		0.252761,
		0.273008,
		0.293408,
		0.313909,
		0.33446,
		0.355015,
		0.375532,
		0.395969,
		0.416289,
		0.436458,
		0.456443,
		0.476215,
		0.495746,
		0.515011,
		0.533987,
		0.552654,
		0.570991,
		0.588982,
		0.606611,
		0.623865,
		0.64073,
		0.000123,
		0.001913,
		0.005711,
		0.011362,
		0.018718,
		0.027641,
		0.037997,
		0.049664,
		0.062522,
		0.076461,
		0.091374,
		0.107161,
		0.12373,
		0.140991,
		0.158861,
		0.17726,
		0.196114,
		0.215353,
		0.234912,
		0.254728,
		0.274744,
		0.294906,
		0.315161,
		0.335463,
		0.355767,
		0.376031,
		0.396218,
		0.416289,
		0.436213,
		0.455958,
		0.475496,
		0.494799,
		0.513845,
		0.532609,
		0.551073,
		0.569217,
		0.587025,
		0.000475,
		0.002895,
		0.007225,
		0.01332,
		0.02104,
		0.030251,
		0.04083,
		0.052658,
		0.065621,
		0.079615,
		0.094539,
		0.110298,
		0.126801,
		0.143965,
		0.16171,
		0.179959,
		0.198643,
		0.217694,
		0.237049,
		0.25665,
		0.27644,
		0.296367,
		0.316383,
		0.336442,
		0.3565,
		0.376519,
		0.39646,
		0.416289,
		0.435974,
		0.455485,
		0.474793,
		0.493875,
		0.512705,
		0.531263,
		0.001035,
		0.004038,
		0.008861,
		0.015366,
		0.023419,
		0.032895,
		0.043674,
		0.055645,
		0.0687,
		0.082738,
		0.097664,
		0.113386,
		0.12982,
		0.146884,
		0.164503,
		0.182603,
		0.201118,
		0.219983,
		0.239139,
		0.258527,
		0.278096,
		0.297795,
		0.317576,
		0.337397,
		0.357216,
		0.376995,
		0.396697,
		0.416289,
		0.435741,
		0.455022,
		0.474108,
		0.000114,
		0.001783,
		0.005327,
		0.010606,
		0.017488,
		0.025848,
		0.035564,
		0.046525,
		0.058623,
		0.071756,
		0.085828,
		0.100747,
		0.116428,
		0.132788,
		0.14975,
		0.167242,
		0.185194,
		0.203541,
		0.222223,
		0.241182,
		0.260362,
		0.279714,
		0.299189,
		0.318742,
		0.33833,
		0.357915,
		0.377459,
		0.396928,
		0.416289,
		0.000443,
		0.0027,
		0.006746,
		0.012447,
		0.019678,
		0.028318,
		0.038254,
		0.049379,
		0.061589,
		0.074788,
		0.088884,
		0.103789,
		0.119423,
		0.135705,
		0.152563,
		0.169927,
		0.187731,
		0.205913,
		0.224414,
		0.243179,
		0.262156,
		0.281295,
		0.300551,
		0.31988,
		0.339242,
		0.358598,
		0.000966,
		0.003771,
		0.008282,
		0.014374,
		0.021925,
		0.030823,
		0.040959,
		0.05223,
		0.064539,
		0.077793,
		0.091904,
		0.106789,
		0.122371,
		0.138573,
		0.155325,
		0.172561,
		0.190218,
		0.208236,
		0.226559,
		0.245133,
		0.26391,
		0.282841,
		0.301882,
		0.000107,
		0.001665,
		0.004979,
		0.009923,
		0.016376,
		0.024223,
		0.033357,
		0.043674,
		0.055077,
		0.067471,
		0.08077,
		0.094888,
		0.109748,
		0.125272,
		0.141391,
		0.158037,
		0.175145,
		0.192655,
		0.210511,
		0.228658,
		0.247045,
		0.000414,
		0.002525,
		0.006313,
		0.011657,
		0.018444,
		0.026564,
		0.035914,
		0.046396,
		0.057915,
		0.070384,
		0.083718,
		0.097836,
		0.112664,
		0.128128,
		0.144161,
		0.160699,
		0.177679,
		0.195044,
		0.000903,
		0.003529,
		0.007758,
		0.013475,
		0.02057,
		0.028941,
		0.038489,
		0.049119,
		0.060743,
		0.073275,
		0.086636,
		0.100747,
		0.115538,
		0.130939,
		0.146884,
		0.0001,
		0.001559,
		0.004665,
		0.009304,
		0.015366,
		0.022747,
		0.031349,
		0.041077,
		0.051842,
		0.063557,
		0.076143,
		0.089522,
		0.103621,
		0.000387,
		0.002366,
		0.00592,
		0.01094,
		0.017323,
		0.024968,
		0.033782,
		0.043674,
		0.05456,
		0.066356,
		0.000846,
		0.00331,
		0.007282,
		0.012657,
		0.019337,
		0.027226,
		0.036235,
		9.3e-005,
		0.001462,
		0.004379,
		0.008741,
		0.014447,
		0.000363,
		0.002221,
		0.005562,
		0.003183,
		0.00144,
		0.000366,
		0.021402,
		0.016673,
		0.012464,
		0.008808,
		0.005737,
		0.003284,
		0.001486,
		0.000378,
		0.046277,
		0.039569,
		0.033279,
		0.027432,
		0.022059,
		0.01719,
		0.012854,
		0.009086,
		0.00592,
		0.00339,
		0.001534,
		0.00039,
		0.078987,
		0.070648,
		0.062632,
		0.054963,
		0.047667,
		0.040769,
		0.034297,
		0.02828,
		0.022747,
		0.017731,
		0.013263,
		0.009378,
		0.006111,
		0.003501,
		0.001584,
		0.000403,
		0.118371,
		0.108724,
		0.099316,
		0.090168,
		0.081301,
		0.072738,
		0.064503,
		0.056622,
		0.049119,
		0.042023,
		0.035363,
		0.029167,
		0.023468,
		0.018298,
		0.013691,
		0.009684,
		0.006313,
		0.003617,
		0.001638,
		0.000417,
		0.163312,
		0.152659,
		0.142168,
		0.131859,
		0.121749,
		0.111859,
		0.102209,
		0.092821,
		0.083718,
		0.074922,
		0.06646,
		0.058356,
		0.050639,
		0.043337,
		0.036479,
		0.030097,
		0.024223,
		0.018893,
		0.014141,
		0.010005,
		0.006524,
		0.003739,
		0.001693,
		0.000431,
		0.212739,
		0.201355,
		0.190067,
		0.178891,
		0.167843,
		0.156939,
		0.146196,
		0.135635,
		0.125272,
		0.11513,
		0.105229,
		0.095593,
		0.086243,
		0.077206,
		0.068506,
		0.060171,
		0.05223,
		0.044712,
		0.037648,
		0.031071,
		0.025016,
		0.019517,
		0.014613,
		0.010342,
		0.006746,
		0.003868,
		0.001752,
		0.000447,
		0.265625,
		0.253765,
		0.241943,
		0.230171,
		0.218463,
		0.206833,
		0.195295,
		0.183866,
		0.172561,
		0.161399,
		0.150396,
		0.139573,
		0.128949,
		0.118545,
		0.108384,
		0.098489,
		0.088884,
		0.079594,
		0.070648,
		0.062072,
		0.053897,
		0.046154,
		0.038875,
		0.032094,
		0.025848,
		0.020173,
		0.015108,
		0.010696,
		0.006979,
		0.004003,
		0.001814,
		0.000462,
		0.320992,
		0.308888,
		0.29677,
		0.28465,
		0.272538,
		0.260446,
		0.248386,
		0.236371,
		0.224414,
		0.212531,
		0.200737,
		0.189047,
		0.177478,
		0.166048,
		0.154777,
		0.143683,
		0.132788,
		0.122113,
		0.111681,
		0.101517,
		0.091646,
		0.082095,
		0.072891,
		0.064064,
		0.055645,
		0.047667,
		0.040162,
		0.033168,
		0.026722,
		0.020862,
		0.01563,
		0.011069,
		0.007225,
		0.004145,
		0.001879,
		0.000479,
		0.377913,
		0.365774,
		0.353578,
		0.341334,
		0.32905,
		0.316735,
		0.304401,
		0.292057,
		0.279714,
		0.267385,
		0.255081,
		0.242817,
		0.230606,
		0.218463,
		0.206404,
		0.194445,
		0.182603,
		0.170898,
		0.159349,
		0.147975,
		0.136799,
		0.125842,
		0.11513,
		0.104687,
		0.094539,
		0.084714,
		0.075242,
		0.066153,
		0.057479,
		0.049255,
		0.041515,
		0.034297,
		0.027641,
		0.021587,
		0.016179,
		0.011462,
		0.007484,
		0.004296,
		0.001948,
		0.000497,
		0.435513,
		0.423526,
		0.411446,
		0.39928,
		0.387032,
		0.374712,
		0.362328,
		0.349886,
		0.337397,
		0.32487,
		0.312315,
		0.299743,
		0.287165,
		0.274594,
		0.262041,
		0.249522,
		0.237049,
		0.224639,
		0.212308,
		0.200072,
		0.18795,
		0.17596,
		0.164123,
		0.152459,
		0.140991,
		0.129743,
		0.118739,
		0.108006,
		0.09757,
		0.087461,
		0.077708,
		0.068345,
		0.059405,
		0.050923,
		0.042936,
		0.035484,
		0.028608,
		0.02235,
		0.016757,
		0.011876,
		0.007758,
		0.004454,
		0.002021,
		0.000516,
		0.492972,
		0.481304,
		0.469512,
		0.457602,
		0.445578,
		0.433445,
		0.42121,
		0.40888,
		0.39646,
		0.383959,
		0.371384,
		0.358744,
		0.346048,
		0.333306,
		0.320527,
		0.307723,
		0.294906,
		0.282087,
		0.26928,
		0.256499,
		0.243759,
		0.231074,
		0.218463,
		0.205942,
		0.193529,
		0.181245,
		0.169111,
		0.157147,
		0.145377,
		0.133826,
		0.122519,
		0.111483,
		0.100747,
		0.090341,
		0.080297,
		0.070648,
		0.061429,
		0.052677,
		0.044432,
		0.036734,
		0.029626,
		0.023155,
		0.017367,
		0.012313,
		0.008046,
		0.004622,
		0.002098,
		0.000536,
		0.549528,
		0.538323,
		0.526971,
		0.515474,
		0.503836,
		0.492061,
		0.480153,
		0.468117,
		0.455958,
		0.443682,
		0.431293,
		0.4188,
		0.406208,
		0.393526,
		0.38076,
		0.36792,
		0.355015,
		0.342055,
		0.32905,
		0.31601,
		0.302949,
		0.289878,
		0.276811,
		0.263763,
		0.250747,
		0.237781,
		0.224882,
		0.212067,
		0.199356,
		0.186768,
		0.174326,
		0.162051,
		0.149968,
		0.138102,
		0.12648,
		0.11513,
		0.104082,
		0.093366,
		0.083016,
		0.073068,
		0.063557,
		0.054523,
		0.046007,
		0.038051,
		0.0307,
		0.024004,
		0.018011,
		0.012775,
		0.008351,
		0.004799,
		0.002179,
		0.000557,
		0.604481,
		0.593863,
		0.583079,
		0.572131,
		0.561021,
		0.54975,
		0.538323,
		0.526742,
		0.515011,
		0.503133,
		0.491113,
		0.478955,
		0.466665,
		0.454246,
		0.441707,
		0.429052,
		0.416289,
		0.403426,
		0.390469,
		0.377429,
		0.364313,
		0.351133,
		0.337898,
		0.324619,
		0.31131,
		0.297982,
		0.28465,
		0.271328,
		0.258031,
		0.244776,
		0.23158,
		0.218463,
		0.205443,
		0.192542,
		0.179781,
		0.167184,
		0.154777,
		0.142584,
		0.130634,
		0.118957,
		0.107582,
		0.096544,
		0.085875,
		0.075614,
		0.065798,
		0.056468,
		0.047667,
		0.039439,
		0.031834,
		0.0249,
		0.018691,
		0.013263,
		0.008674,
		0.004986,
		0.002265,
		0.000579,
		0.657196,
		0.647268,
		0.63716,
		0.626874,
		0.61641,
		0.60577,
		0.594955,
		0.583966,
		0.572806,
		0.561477,
		0.549982,
		0.538323,
		0.526505,
		0.514529,
		0.502402,
		0.490126,
		0.477707,
		0.465151,
		0.452462,
		0.439648,
		0.426716,
		0.413672,
		0.400525,
		0.387283,
		0.373956,
		0.360554,
		0.347086,
		0.333566,
		0.320005,
		0.306416,
		0.292812,
		0.27921,
		0.265625,
		0.252073,
		0.238574,
		0.225145,
		0.211807,
		0.198581,
		0.185491,
		0.172561,
		0.159817,
		0.147285,
		0.134994,
		0.122976,
		0.111261,
		0.099885,
		0.088884,
		0.078295,
		0.068159,
		0.058518,
		0.049418,
		0.040905,
		0.033031,
		0.025848,
		0.019411,
		0.013779,
		0.009016,
		0.005185,
		0.002356,
		0.000602,
		0.707107,
		0.69795,
		0.688606,
		0.679074,
		0.669354,
		0.659446,
		0.649351,
		0.639069,
		0.628601,
		0.617947,
		0.607109,
		0.596089,
		0.584888,
		0.573509,
		0.561953,
		0.550223,
		0.538323,
		0.526257,
		0.514027,
		0.501639,
		0.489096,
		0.476406,
		0.463572,
		0.450601,
		0.437501,
		0.424278,
		0.410941,
		0.397498,
		0.383959,
		0.370333,
		0.356632,
		0.342867,
		0.32905,
		0.315194,
		0.301315,
		0.287427,
		0.273547,
		0.259691,
		0.245878,
		0.232129,
		0.218463,
		0.204903,
		0.191473,
		0.178197,
		0.165103,
		0.152218,
		0.139573,
		0.127199,
		0.11513,
		0.103402,
		0.092052,
		0.081119,
		0.070648,
		0.060681,
		0.051267,
		0.042454,
		0.034297,
		0.02685,
		0.020173,
		0.014327,
		0.009378,
		0.005396,
		0.002453,
		0.000627,
		0.753719,
		0.745396,
		0.736881,
		0.728175,
		0.719275,
		0.710181,
		0.700893,
		0.691409,
		0.681729,
		0.671853,
		0.661782,
		0.651515,
		0.641052,
		0.630396,
		0.619546,
		0.608503,
		0.59727,
		0.585848,
		0.57424,
		0.562448,
		0.550474,
		0.538323,
		0.525998,
		0.513503,
		0.500843,
		0.488022,
		0.475047,
		0.461923,
		0.448658,
		0.435259,
		0.421733,
		0.40809,
		0.394338,
		0.380488,
		0.366551,
		0.352538,
		0.338462,
		0.324337,
		0.310176,
		0.295997,
		0.281814,
		0.267647,
		0.253513,
		0.239433,
		0.225429,
		0.211524,
		0.197741,
		0.184108,
		0.170651,
		0.1574,
		0.144385,
		0.131641,
		0.119202,
		0.107106,
		0.095391,
		0.084099,
		0.073275,
		0.062966,
		0.053221,
		0.044093,
		0.035637,
		0.027912,
		0.02098,
		0.014907,
		0.009762,
		0.00562,
		0.002556,
		0.000654,
		0.796619,
		0.789171,
		0.781532,
		0.773702,
		0.765677,
		0.757456,
		0.749038,
		0.740422,
		0.731605,
		0.722587,
		0.713366,
		0.703943,
		0.694316,
		0.684484,
		0.674449,
		0.664208,
		0.653763,
		0.643115,
		0.632263,
		0.621209,
		0.609954,
		0.5985,
		0.586849,
		0.575003,
		0.562964,
		0.550737,
		0.538323,
		0.525728,
		0.512956,
		0.500011,
		0.4869,
		0.473628,
		0.460201,
		0.446628,
		0.432915,
		0.419073,
		0.405109,
		0.391034,
		0.37686,
		0.362598,
		0.34826,
		0.333861,
		0.319414,
		0.304937,
		0.290446,
		0.27596,
		0.261496,
		0.247077,
		0.232725,
		0.218463,
		0.204316,
		0.190312,
		0.176478,
		0.162846,
		0.149448,
		0.136317,
		0.123492,
		0.11101,
		0.098913,
		0.087245,
		0.076051,
		0.065382,
		0.055289,
		0.045828,
		0.037057,
		0.029038,
		0.021837,
		0.015523,
		0.010171,
		0.005858,
		0.002666,
		0.000682,
		0.835475,
		0.828924,
		0.822187,
		0.815262,
		0.808146,
		0.800837,
		0.793332,
		0.78563,
		0.777728,
		0.769624,
		0.761317,
		0.752804,
		0.744084,
		0.735156,
		0.726017,
		0.716668,
		0.707107,
		0.697333,
		0.687346,
		0.677145,
		0.66673,
		0.656102,
		0.645261,
		0.634207,
		0.622942,
		0.611467,
		0.599783,
		0.587893,
		0.575799,
		0.563503,
		0.55101,
		0.538323,
		0.525446,
		0.512384,
		0.499142,
		0.485726,
		0.472143,
		0.4584,
		0.444504,
		0.430464,
		0.416289,
		0.40199,
		0.387578,
		0.373065,
		0.358463,
		0.343786,
		0.32905,
		0.31427,
		0.299464,
		0.28465,
		0.269849,
		0.255081,
		0.24037,
		0.22574,
		0.211216,
		0.196828,
		0.182603,
		0.168575,
		0.154777,
		0.141244,
		0.128015,
		0.11513,
		0.102633,
		0.090569,
		0.078987,
		0.067939,
		0.057479,
		0.047667,
		0.038563,
		0.030233,
		0.022747,
		0.016179,
		0.010606,
		0.006111,
		0.002783,
		0.000713,
		0.870045,
		0.864394,
		0.858565,
		0.852555,
		0.846361,
		0.839981,
		0.833411,
		0.826649,
		0.819691,
		0.812535,
		0.805179,
		0.79762,
		0.789856,
		0.781884,
		0.773702,
		0.765307,
		0.756699,
		0.747875,
		0.738833,
		0.729573,
		0.720092,
		0.71039,
		0.700466,
		0.690319,
		0.679948,
		0.669354,
		0.658536,
		0.647496,
		0.636233,
		0.624748,
		0.613044,
		0.601121,
		0.588982,
		0.57663,
		0.564067,
		0.551297,
		0.538323,
		0.525152,
		0.511786,
		0.498233,
		0.484498,
		0.470589,
		0.456513,
		0.442279,
		0.427896,
		0.413374,
		0.398725,
		0.383959,
		0.369091,
		0.354133,
		0.339103,
		0.324015,
		0.308888,
		0.29374,
		0.278592,
		0.263466,
		0.248386,
		0.233376,
		0.218463,
		0.203676,
		0.189047,
		0.174607,
		0.160392,
		0.146439,
		0.132788,
		0.119481,
		0.106564,
		0.094085,
		0.082095,
		0.070648,
		0.059802,
		0.049618,
		0.040162,
		0.031504,
		0.023716,
		0.016877,
		0.011069,
		0.006382,
		0.002907,
		0.000745,
		0.900183,
		0.895415,
		0.89048,
		0.885375,
		0.880097,
		0.874642,
		0.869007,
		0.863189,
		0.857183,
		0.850988,
		0.8446,
		0.838015,
		0.83123,
		0.824244,
		0.817052,
		0.809651,
		0.80204,
		0.794215,
		0.786174,
		0.777914,
		0.769433,
		0.76073,
		0.751801,
		0.742645,
		0.73326,
		0.723646,
		0.7138,
		0.703721,
		0.69341,
		0.682865,
		0.672085,
		0.661072,
		0.649825,
		0.638345,
		0.626633,
		0.61469,
		0.602518,
		0.59012,
		0.577498,
		0.564656,
		0.551596,
		0.538323,
		0.524843,
		0.51116,
		0.497281,
		0.483212,
		0.468961,
		0.454537,
		0.439948,
		0.425205,
		0.410318,
		0.395301,
		0.380165,
		0.364925,
		0.349596,
		0.334196,
		0.318742,
		0.303253,
		0.28775,
		0.272257,
		0.256796,
		0.241394,
		0.226079,
		0.21088,
		0.19583,
		0.180962,
		0.166312,
		0.151921,
		0.137829,
		0.124081,
		0.110724,
		0.097809,
		0.085389,
		0.073521,
		0.062267,
		0.051692,
		0.041863,
		0.032856,
		0.024747,
		0.017621,
		0.011564,
		0.006671,
		0.003041,
		0.00078,
		0.925841,
		0.921922,
		0.91785,
		0.913621,
		0.909232,
		0.90468,
		0.899959,
		0.895068,
		0.890001,
		0.884756,
		0.879329,
		0.873716,
		0.867913,
		0.861918,
		0.855726,
		0.849334,
		0.842738,
		0.835936,
		0.828924,
		0.821699,
		0.814257,
		0.806596,
		0.798713,
		0.790604,
		0.782268,
		0.773702,
		0.764902,
		0.755868,
		0.746597,
		0.737086,
		0.727336,
		0.717343,
		0.707107,
		0.696627,
		0.685901,
		0.674931,
		0.663715,
		0.652255,
		0.640549,
		0.628601,
		0.61641,
		0.603979,
		0.591311,
		0.578407,
		0.565272,
		0.551909,
		0.538323,
		0.524519,
		0.510503,
		0.496282,
		0.481862,
		0.467253,
		0.452462,
		0.437501,
		0.42238,
		0.407111,
		0.391707,
		0.376183,
		0.360554,
		0.344836,
		0.32905,
		0.313213,
		0.297347,
		0.281477,
		0.265625,
		0.249819,
		0.234088,
		0.218463,
		0.202976,
		0.187663,
		0.172561,
		0.157713,
		0.14316,
		0.128949,
		0.11513,
		0.101756,
		0.088884,
		0.076573,
		0.064888,
		0.053897,
		0.043674,
		0.034297,
		0.025848,
		0.018415,
		0.012092,
		0.006979,
		0.003183,
		0.000817,
		0.947082,
		0.943959,
		0.940699,
		0.937299,
		0.933754,
		0.930061,
		0.926215,
		0.922214,
		0.918052,
		0.913726,
		0.909232,
		0.904566,
		0.899725,
		0.894703,
		0.889497,
		0.884104,
		0.878519,
		0.872739,
		0.866759,
		0.860575,
		0.854185,
		0.847584,
		0.840769,
		0.833736,
		0.826481,
		0.819002,
		0.811294,
		0.803355,
		0.795181,
		0.78677,
		0.778118,
		0.769224,
		0.760084,
		0.750696,
		0.741059,
		0.73117,
		0.721027,
		0.710629,
		0.699976,
		0.689066,
		0.677898,
		0.666474,
		0.654791,
		0.642853,
		0.630658,
		0.618209,
		0.605508,
		0.592557,
		0.579359,
		0.565918,
		0.552238,
		0.538323,
		0.52418,
		0.509814,
		0.495233,
		0.480445,
		0.465459,
		0.450283,
		0.43493,
		0.419412,
		0.403741,
		0.387931,
		0.371999,
		0.355962,
		0.339837,
		0.323646,
		0.307411,
		0.291153,
		0.2749,
		0.258679,
		0.242518,
		0.226451,
		0.210511,
		0.194735,
		0.179163,
		0.163836,
		0.148801,
		0.134105,
		0.119801,
		0.105945,
		0.092596,
		0.079817,
		0.067676,
		0.056247,
		0.045605,
		0.035835,
		0.027023,
		0.019264,
		0.012657,
		0.00731,
		0.003336,
		0.000857,
		0.964076,
		0.961681,
		0.959167,
		0.95653,
		0.953766,
		0.950871,
		0.947842,
		0.944673,
		0.941362,
		0.937904,
		0.934295,
		0.930531,
		0.926607,
		0.922519,
		0.918264,
		0.913836,
		0.909232,
		0.904447,
		0.899478,
		0.894319,
		0.888967,
		0.883417,
		0.877665,
		0.871707,
		0.865538,
		0.859156,
		0.852555,
		0.845732,
		0.838682,
		0.831403,
		0.823889,
		0.816138,
		0.808146,
		0.79991,
		0.791425,
		0.78269,
		0.773702,
		0.764456,
		0.754952,
		0.745185,
		0.735156,
		0.72486,
		0.714298,
		0.703467,
		0.692366,
		0.680995,
		0.669354,
		0.657442,
		0.645261,
		0.63281,
		0.620093,
		0.607109,
		0.593863,
		0.580358,
		0.566596,
		0.552583,
		0.538323,
		0.523824,
		0.50909,
		0.494131,
		0.478955,
		0.463572,
		0.447991,
		0.432226,
		0.416289,
		0.400195,
		0.383959,
		0.367599,
		0.351133,
		0.334582,
		0.317968,
		0.301315,
		0.28465,
		0.268,
		0.251397,
		0.234873,
		0.218463,
		0.202206,
		0.186142,
		0.170317,
		0.154777,
		0.139573,
		0.12476,
		0.110396,
		0.096544,
		0.08327,
		0.070648,
		0.058753,
		0.047667,
		0.037478,
		0.02828,
		0.020173,
		0.013263,
		0.007665,
		0.003501,
		0.000899,
		0.977109,
		0.975358,
		0.973507,
		0.971552,
		0.969489,
		0.967315,
		0.965026,
		0.962616,
		0.960083,
		0.957423,
		0.95463,
		0.951701,
		0.948631,
		0.945417,
		0.942053,
		0.938536,
		0.93486,
		0.931022,
		0.927017,
		0.92284,
		0.918486,
		0.913952,
		0.909232,
		0.904322,
		0.899218,
		0.893914,
		0.888407,
		0.88269,
		0.876762,
		0.870615,
		0.864247,
		0.857652,
		0.850827,
		0.843766,
		0.836467,
		0.828924,
		0.821134,
		0.813093,
		0.804797,
		0.796242,
		0.787425,
		0.778343,
		0.768993,
		0.759371,
		0.749475,
		0.739302,
		0.728851,
		0.71812,
		0.707107,
		0.69581,
		0.68423,
		0.672365,
		0.660215,
		0.647781,
		0.635065,
		0.622067,
		0.608789,
		0.595234,
		0.581406,
		0.567308,
		0.552945,
		0.538323,
		0.523449,
		0.508329,
		0.492972,
		0.477387,
		0.461585,
		0.445578,
		0.429378,
		0.413,
		0.39646,
		0.379775,
		0.362964,
		0.346048,
		0.32905,
		0.311993,
		0.294906,
		0.277816,
		0.260756,
		0.243759,
		0.226862,
		0.210104,
		0.193529,
		0.177183,
		0.161115,
		0.145377,
		0.130029,
		0.11513,
		0.100747,
		0.086952,
		0.073818,
		0.061429,
		0.049871,
		0.039236,
		0.029626,
		0.021147,
		0.013913,
		0.008046,
		0.003677,
		0.000945,
		0.986574,
		0.985372,
		0.984089,
		0.982721,
		0.981267,
		0.97972,
		0.978079,
		0.976338,
		0.974494,
		0.972543,
		0.97048,
		0.968302,
		0.966004,
		0.963582,
		0.961031,
		0.958348,
		0.955526,
		0.952563,
		0.949453,
		0.946192,
		0.942774,
		0.939196,
		0.935451,
		0.931537,
		0.927446,
		0.923176,
		0.91872,
		0.914074,
		0.909232,
		0.904191,
		0.898944,
		0.893487,
		0.887815,
		0.881923,
		0.875806,
		0.869459,
		0.862877,
		0.856057,
		0.848992,
		0.841678,
		0.834112,
		0.826287,
		0.818201,
		0.809849,
		0.801227,
		0.79233,
		0.783156,
		0.773702,
		0.763962,
		0.753936,
		0.743619,
		0.73301,
		0.722107,
		0.710906,
		0.699408,
		0.687611,
		0.675515,
		0.663118,
		0.650423,
		0.637429,
		0.624138,
		0.610552,
		0.596674,
		0.582507,
		0.568056,
		0.553326,
		0.538323,
		0.523054,
		0.507526,
		0.491749,
		0.475734,
		0.45949,
		0.443032,
		0.426374,
		0.409531,
		0.392521,
		0.375363,
		0.358077,
		0.340688,
		0.32322,
		0.3057,
		0.288158,
		0.270627,
		0.253142,
		0.23574,
		0.218463,
		0.201355,
		0.184465,
		0.167843,
		0.151546,
		0.135635,
		0.120172,
		0.105229,
		0.09088,
		0.077206,
		0.064291,
		0.05223,
		0.041121,
		0.031071,
		0.022194,
		0.014613,
		0.008457,
		0.003868,
		0.000995,
		0.992969,
		0.99221,
		0.99139,
		0.990506,
		0.989554,
		0.988531,
		0.987433,
		0.986257,
		0.984999,
		0.983655,
		0.98222,
		0.980692,
		0.979066,
		0.977338,
		0.975504,
		0.973558,
		0.971498,
		0.969317,
		0.967013,
		0.964579,
		0.962012,
		0.959306,
		0.956457,
		0.953459,
		0.950308,
		0.946999,
		0.943527,
		0.939885,
		0.93607,
		0.932076,
		0.927897,
		0.923529,
		0.918966,
		0.914202,
		0.909232,
		0.904051,
		0.898654,
		0.893035,
		0.887188,
		0.881109,
		0.874792,
		0.868232,
		0.861423,
		0.854361,
		0.84704,
		0.839455,
		0.831602,
		0.823476,
		0.815072,
		0.806386,
		0.797413,
		0.78815,
		0.778592,
		0.768736,
		0.758579,
		0.748116,
		0.737347,
		0.726267,
		0.714875,
		0.70317,
		0.69115,
		0.678813,
		0.666161,
		0.653193,
		0.63991,
		0.626313,
		0.612405,
		0.598188,
		0.583667,
		0.568845,
		0.553728,
		0.538323,
		0.522638,
		0.50668,
		0.49046,
		0.473988,
		0.457278,
		0.440344,
		0.423201,
		0.405867,
		0.38836,
		0.370703,
		0.352917,
		0.33503,
		0.317068,
		0.299063,
		0.281048,
		0.263058,
		0.245133,
		0.227317,
		0.209654,
		0.192195,
		0.174994,
		0.15811,
		0.141606,
		0.12555,
		0.110014,
		0.09508,
		0.08083,
		0.067358,
		0.054761,
		0.043145,
		0.032625,
		0.023321,
		0.015366,
		0.0089,
		0.004073,
		0.001049,
		0.996876,
		0.996452,
		0.995985,
		0.995473,
		0.994912,
		0.9943,
		0.993633,
		0.992908,
		0.992122,
		0.991272,
		0.990352,
		0.989361,
		0.988294,
		0.987147,
		0.985916,
		0.984597,
		0.983187,
		0.98168,
		0.980072,
		0.978359,
		0.976536,
		0.974599,
		0.972543,
		0.970362,
		0.968053,
		0.965609,
		0.963026,
		0.960299,
		0.957423,
		0.954391,
		0.951199,
		0.947842,
		0.944313,
		0.940607,
		0.936718,
		0.932642,
		0.928371,
		0.9239,
		0.919224,
		0.914337,
		0.909232,
		0.903905,
		0.898348,
		0.892557,
		0.886525,
		0.880246,
		0.873716,
		0.866927,
		0.859876,
		0.852555,
		0.84496,
		0.837084,
		0.828924,
		0.820474,
		0.811728,
		0.802682,
		0.793332,
		0.783673,
		0.773702,
		0.763413,
		0.752804,
		0.741872,
		0.730613,
		0.719025,
		0.707107,
		0.694856,
		0.682272,
		0.669354,
		0.656102,
		0.642517,
		0.628601,
		0.614355,
		0.599783,
		0.584888,
		0.569676,
		0.554152,
		0.538323,
		0.522198,
		0.505785,
		0.489096,
		0.472143,
		0.45494,
		0.437501,
		0.419845,
		0.40199,
		0.383959,
		0.365774,
		0.347461,
		0.32905,
		0.31057,
		0.292057,
		0.273547,
		0.255081,
		0.236704,
		0.218463,
		0.200411,
		0.182603,
		0.165103,
		0.147975,
		0.131291,
		0.11513,
		0.099574,
		0.084714,
		0.070648,
		0.057479,
		0.045322,
		0.034297,
		0.024536,
		0.016179,
		0.009378,
		0.004296,
		0.001107,
		0.998929,
		0.998733,
		0.998511,
		0.99826,
		0.997978,
		0.997663,
		0.997312,
		0.996922,
		0.99649,
		0.996013,
		0.995488,
		0.994912,
		0.994282,
		0.993593,
		0.992844,
		0.992029,
		0.991144,
		0.990188,
		0.989154,
		0.988039,
		0.986838,
		0.985548,
		0.984164,
		0.982681,
		0.981094,
		0.9794,
		0.977591,
		0.975665,
		0.973615,
		0.971437,
		0.969125,
		0.966673,
		0.964076,
		0.961329,
		0.958426,
		0.955361,
		0.952128,
		0.948721,
		0.945135,
		0.941362,
		0.937398,
		0.933236,
		0.928869,
		0.924291,
		0.919497,
		0.914479,
		0.909232,
		0.903749,
		0.898023,
		0.892049,
		0.88582,
		0.879329,
		0.872571,
		0.865538,
		0.858226,
		0.850628,
		0.842738,
		0.834551,
		0.82606,
		0.81726,
		0.808146,
		0.798713,
		0.788955,
		0.778869,
		0.76845,
		0.757694,
		0.746597,
		0.735156,
		0.723368,
		0.71123,
		0.698742,
		0.685901,
		0.672708,
		0.65916,
		0.645261,
		0.63101,
		0.61641,
		0.601465,
		0.586177,
		0.570554,
		0.5546,
		0.538323,
		0.521733,
		0.504839,
		0.487653,
		0.470189,
		0.452462,
		0.434489,
		0.416289,
		0.397884,
		0.379296,
		0.360554,
		0.341684,
		0.32272,
		0.303696,
		0.28465,
		0.265625,
		0.246666,
		0.227824,
		0.209152,
		0.19071,
		0.172561,
		0.154777,
		0.137432,
		0.120607,
		0.104393,
		0.088884,
		0.074184,
		0.060404,
		0.047667,
		0.036101,
		0.025848,
		0.017058,
		0.009896,
		0.004537,
		0.00117,
		0.999771,
		0.999707,
		0.999629,
		0.999537,
		0.999429,
		0.999303,
		0.999156,
		0.998987,
		0.998793,
		0.998572,
		0.998321,
		0.998039,
		0.997721,
		0.997366,
		0.99697,
		0.996529,
		0.996042,
		0.995504,
		0.994912,
		0.994262,
		0.993551,
		0.992774,
		0.991928,
		0.991008,
		0.990011,
		0.988931,
		0.987764,
		0.986506,
		0.985151,
		0.983695,
		0.982133,
		0.980459,
		0.978669,
		0.976757,
		0.974716,
		0.972543,
		0.97023,
		0.967772,
		0.965164,
		0.962398,
		0.959469,
		0.956371,
		0.953097,
		0.94964,
		0.945995,
		0.942154,
		0.938112,
		0.93386,
		0.929393,
		0.924704,
		0.919785,
		0.91463,
		0.909232,
		0.903584,
		0.897679,
		0.89151,
		0.88507,
		0.878352,
		0.87135,
		0.864056,
		0.856465,
		0.848568,
		0.840361,
		0.831837,
		0.82299,
		0.813813,
		0.804301,
		0.794448,
		0.78425,
		0.773702,
		0.762798,
		0.751535,
		0.739908,
		0.727916,
		0.715554,
		0.702821,
		0.689715,
		0.676234,
		0.66238,
		0.648151,
		0.63355,
		0.618579,
		0.603241,
		0.58754,
		0.571482,
		0.555074,
		0.538323,
		0.52124,
		0.503836,
		0.486123,
		0.468117,
		0.449834,
		0.431293,
		0.412516,
		0.393526,
		0.374349,
		0.355015,
		0.335557,
		0.31601,
		0.296414,
		0.276811,
		0.25725,
		0.237781,
		0.218463,
		0.199356,
		0.180527,
		0.162051,
		0.144007,
		0.12648,
		0.109566,
		0.093366,
		0.07799,
		0.063557,
		0.050198,
		0.038051,
		0.027267,
		0.018011,
		0.010458,
		0.004799,
		0.001239,
		0.999985,
		0.999975,
		0.999962,
		0.999944,
		0.99992,
		0.999889,
		0.99985,
		0.999801,
		0.999741,
		0.999668,
		0.999581,
		0.999477,
		0.999355,
		0.999212,
		0.999046,
		0.998854,
		0.998635,
		0.998385,
		0.998102,
		0.997782,
		0.997422,
		0.99702,
		0.996571,
		0.996073,
		0.995521,
		0.994912,
		0.994242,
		0.993506,
		0.9927,
		0.991821,
		0.990862,
		0.989821,
		0.988691,
		0.987468,
		0.986146,
		0.984721,
		0.983187,
		0.981538,
		0.979768,
		0.977873,
		0.975845,
		0.973679,
		0.971369,
		0.968908,
		0.966289,
		0.963506,
		0.960553,
		0.957423,
		0.954108,
		0.950601,
		0.946896,
		0.942986,
		0.938862,
		0.934518,
		0.929947,
		0.92514,
		0.92009,
		0.91479,
		0.909232,
		0.903409,
		0.897313,
		0.890936,
		0.88427,
		0.877309,
		0.870045,
		0.862471,
		0.854579,
		0.846361,
		0.837812,
		0.828924,
		0.819691,
		0.810106,
		0.800163,
		0.789856,
		0.77918,
		0.768129,
		0.756699,
		0.744885,
		0.732684,
		0.720092,
		0.707107,
		0.693726,
		0.679948,
		0.665773,
		0.651201,
		0.636233,
		0.620871,
		0.605119,
		0.588982,
		0.572465,
		0.555576,
		0.538323,
		0.520718,
		0.502771,
		0.484498,
		0.465916,
		0.447041,
		0.427896,
		0.408505,
		0.388893,
		0.369091,
		0.349131,
		0.32905,
		0.308888,
		0.28869,
		0.268504,
		0.248386,
		0.228392,
		0.208589,
		0.189047,
		0.169842,
		0.151058,
		0.132788,
		0.11513,
		0.098193,
		0.082095,
		0.066963,
		0.052935,
		0.040162,
		0.028807,
		0.019046,
		0.011069,
		0.005084,
		0.001314,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999995,
		0.99999,
		0.999982,
		0.999972,
		0.999957,
		0.999936,
		0.999909,
		0.999874,
		0.99983,
		0.999775,
		0.999707,
		0.999624,
		0.999525,
		0.999407,
		0.999268,
		0.999106,
		0.998917,
		0.9987,
		0.998451,
		0.998167,
		0.997845,
		0.997481,
		0.997073,
		0.996615,
		0.996106,
		0.995539,
		0.994912,
		0.99422,
		0.993458,
		0.992621,
		0.991705,
		0.990705,
		0.989616,
		0.988431,
		0.987147,
		0.985756,
		0.984253,
		0.982633,
		0.980889,
		0.979014,
		0.977002,
		0.974848,
		0.972543,
		0.970081,
		0.967454,
		0.964657,
		0.961681,
		0.958519,
		0.955164,
		0.951607,
		0.947842,
		0.943859,
		0.939652,
		0.935212,
		0.930531,
		0.925601,
		0.920413,
		0.91496,
		0.909232,
		0.903223,
		0.896922,
		0.890323,
		0.883417,
		0.876195,
		0.868649,
		0.860772,
		0.852555,
		0.843991,
		0.835071,
		0.82579,
		0.816138,
		0.80611,
		0.795699,
		0.784898,
		0.773702,
		0.762104,
		0.750101,
		0.737688,
		0.72486,
		0.711615,
		0.69795,
		0.683863,
		0.669354,
		0.654422,
		0.639069,
		0.623297,
		0.607109,
		0.590511,
		0.573509,
		0.556109,
		0.538323,
		0.520162,
		0.501639,
		0.482769,
		0.463572,
		0.444067,
		0.424278,
		0.404232,
		0.383959,
		0.363491,
		0.342867,
		0.322126,
		0.301315,
		0.280485,
		0.259691,
		0.238994,
		0.218463,
		0.19817,
		0.178197,
		0.158632,
		0.139573,
		0.121124,
		0.103402,
		0.086531,
		0.070648,
		0.055902,
		0.042454,
		0.030481,
		0.020173,
		0.011736,
		0.005396,
		0.001396,
		0.999985,
		0.999991,
		0.999995,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999994,
		0.999988,
		0.99998,
		0.999968,
		0.999951,
		0.999927,
		0.999897,
		0.999857,
		0.999806,
		0.999743,
		0.999666,
		0.999572,
		0.999459,
		0.999324,
		0.999166,
		0.998981,
		0.998766,
		0.998518,
		0.998234,
		0.99791,
		0.997543,
		0.997128,
		0.996662,
		0.99614,
		0.995558,
		0.994912,
		0.994196,
		0.993406,
		0.992536,
		0.991581,
		0.990536,
		0.989394,
		0.98815,
		0.986798,
		0.985332,
		0.983744,
		0.982029,
		0.980179,
		0.978187,
		0.976047,
		0.973751,
		0.971292,
		0.968661,
		0.965851,
		0.962855,
		0.959663,
		0.956268,
		0.952661,
		0.948834,
		0.944778,
		0.940484,
		0.935944,
		0.931149,
		0.926089,
		0.920755,
		0.91514,
		0.909232,
		0.903024,
		0.896506,
		0.889668,
		0.882503,
		0.875,
		0.867151,
		0.858946,
		0.850378,
		0.841438,
		0.832117,
		0.822408,
		0.812301,
		0.801791,
		0.790869,
		0.77953,
		0.767766,
		0.755573,
		0.742944,
		0.729875,
		0.716363,
		0.702404,
		0.687996,
		0.673139,
		0.657831,
		0.642073,
		0.625869,
		0.609221,
		0.592135,
		0.574618,
		0.556677,
		0.538323,
		0.51957,
		0.500431,
		0.480925,
		0.461072,
		0.440894,
		0.420418,
		0.399673,
		0.378694,
		0.357518,
		0.336187,
		0.314747,
		0.293251,
		0.271757,
		0.250328,
		0.229035,
		0.207954,
		0.187172,
		0.166781,
		0.146884,
		0.127594,
		0.109032,
		0.091333,
		0.074644,
		0.059124,
		0.044948,
		0.032305,
		0.021402,
		0.012464,
		0.005737,
		0.001486,
		0.999771,
		0.999819,
		0.999859,
		0.999892,
		0.99992,
		0.999942,
		0.999959,
		0.999973,
		0.999982,
		0.999989,
		0.999994,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999987,
		0.999977,
		0.999963,
		0.999944,
		0.999917,
		0.999882,
		0.999836,
		0.999779,
		0.999707,
		0.999618,
		0.99951,
		0.999381,
		0.999227,
		0.999046,
		0.998834,
		0.998588,
		0.998304,
		0.997978,
		0.997607,
		0.997186,
		0.996711,
		0.996177,
		0.995579,
		0.994912,
		0.994171,
		0.99335,
		0.992444,
		0.991447,
		0.990352,
		0.989154,
		0.987845,
		0.986419,
		0.984868,
		0.983187,
		0.981366,
		0.9794,
		0.977279,
		0.974996,
		0.972543,
		0.969911,
		0.967092,
		0.964076,
		0.960857,
		0.957423,
		0.953766,
		0.949876,
		0.945745,
		0.941362,
		0.936718,
		0.931803,
		0.926607,
		0.92112,
		0.915332,
		0.909232,
		0.902812,
		0.89606,
		0.888967,
		0.881522,
		0.873716,
		0.865538,
		0.85698,
		0.848031,
		0.838682,
		0.828924,
		0.818748,
		0.808146,
		0.797109,
		0.78563,
		0.773702,
		0.761317,
		0.74847,
		0.735156,
		0.721369,
		0.707107,
		0.692366,
		0.677145,
		0.661443,
		0.645261,
		0.628601,
		0.611467,
		0.593863,
		0.575799,
		0.557281,
		0.538323,
		0.518938,
		0.499142,
		0.478955,
		0.4584,
		0.437501,
		0.416289,
		0.394798,
		0.373065,
		0.351133,
		0.32905,
		0.306869,
		0.28465,
		0.262459,
		0.24037,
		0.218463,
		0.196828,
		0.175563,
		0.154777,
		0.134589,
		0.11513,
		0.096544,
		0.078987,
		0.062632,
		0.047667,
		0.034297,
		0.022747,
		0.013263,
		0.006111,
		0.001584,
		0.998929,
		0.999074,
		0.999205,
		0.999323,
		0.999429,
		0.999523,
		0.999606,
		0.999679,
		0.999741,
		0.999795,
		0.99984,
		0.999878,
		0.999909,
		0.999934,
		0.999954,
		0.999969,
		0.99998,
		0.999988,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999996,
		0.999992,
		0.999985,
		0.999974,
		0.999958,
		0.999935,
		0.999905,
		0.999864,
		0.999812,
		0.999746,
		0.999663,
		0.999561,
		0.999437,
		0.999289,
		0.999112,
		0.998903,
		0.998659,
		0.998376,
		0.998049,
		0.997675,
		0.997247,
		0.996763,
		0.996216,
		0.995601,
		0.994912,
		0.994144,
		0.993291,
		0.992346,
		0.991302,
		0.990153,
		0.988892,
		0.987511,
		0.986004,
		0.984361,
		0.982575,
		0.980638,
		0.978541,
		0.976276,
		0.973833,
		0.971204,
		0.968379,
		0.965348,
		0.962103,
		0.958631,
		0.954925,
		0.950973,
		0.946765,
		0.94229,
		0.937538,
		0.932498,
		0.927158,
		0.921508,
		0.915536,
		0.909232,
		0.902584,
		0.895582,
		0.888213,
		0.880467,
		0.872333,
		0.863799,
		0.854856,
		0.845492,
		0.835698,
		0.825463,
		0.814777,
		0.803632,
		0.792019,
		0.779928,
		0.767353,
		0.754286,
		0.740722,
		0.726654,
		0.712079,
		0.696992,
		0.681392,
		0.665277,
		0.648648,
		0.631507,
		0.613858,
		0.595706,
		0.577059,
		0.557927,
		0.538323,
		0.518262,
		0.497762,
		0.476846,
		0.455537,
		0.433865,
		0.411865,
		0.389573,
		0.367032,
		0.344293,
		0.321409,
		0.298442,
		0.27546,
		0.25254,
		0.229766,
		0.207232,
		0.185043,
		0.163312,
		0.142168,
		0.121749,
		0.102209,
		0.083718,
		0.06646,
		0.050639,
		0.036478,
		0.024222,
		0.014139,
		0.006522,
		0.001693,
		0.996876,
		0.99718,
		0.997465,
		0.997731,
		0.997978,
		0.998208,
		0.99842,
		0.998614,
		0.998793,
		0.998956,
		0.999103,
		0.999236,
		0.999355,
		0.999461,
		0.999554,
		0.999636,
		0.999707,
		0.999767,
		0.999819,
		0.999861,
		0.999897,
		0.999925,
		0.999947,
		0.999965,
		0.999977,
		0.999986,
		0.999992,
		0.999996,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999996,
		0.999991,
		0.999982,
		0.99997,
		0.999951,
		0.999926,
		0.99989,
		0.999844,
		0.999783,
		0.999707,
		0.999611,
		0.999493,
		0.99935,
		0.999178,
		0.998974,
		0.998733,
		0.998451,
		0.998123,
		0.997745,
		0.997312,
		0.996818,
		0.996257,
		0.995624,
		0.994912,
		0.994115,
		0.993226,
		0.992239,
		0.991144,
		0.989936,
		0.988607,
		0.987147,
		0.985548,
		0.983803,
		0.981901,
		0.979834,
		0.977591,
		0.975164,
		0.972543,
		0.969716,
		0.966673,
		0.963404,
		0.959898,
		0.956143,
		0.952128,
		0.947842,
		0.943272,
		0.938407,
		0.933236,
		0.927745,
		0.921922,
		0.915755,
		0.909232,
		0.902341,
		0.895068,
		0.887401,
		0.879329,
		0.870838,
		0.861918,
		0.852555,
		0.842738,
		0.832457,
		0.821699,
		0.810454,
		0.798713,
		0.786465,
		0.773702,
		0.760415,
		0.746597,
		0.732241,
		0.717343,
		0.701897,
		0.685901,
		0.669354,
		0.652255,
		0.634605,
		0.61641,
		0.597675,
		0.578407,
		0.558619,
		0.538323,
		0.517538,
		0.496282,
		0.474581,
		0.452462,
		0.42996,
		0.407111,
		0.383959,
		0.360554,
		0.336951,
		0.313213,
		0.289411,
		0.265625,
		0.241943,
		0.218463,
		0.195295,
		0.172561,
		0.150396,
		0.128948,
		0.108382,
		0.088879,
		0.07064,
		0.053884,
		0.038856,
		0.025825,
		0.015085,
		0.006962,
		0.001807,
		0.992969,
		0.99349,
		0.993987,
		0.994461,
		0.994912,
		0.99534,
		0.995745,
		0.996128,
		0.99649,
		0.996829,
		0.997147,
		0.997444,
		0.997721,
		0.997978,
		0.998216,
		0.998435,
		0.998635,
		0.998818,
		0.998984,
		0.999134,
		0.999268,
		0.999388,
		0.999493,
		0.999586,
		0.999666,
		0.999735,
		0.999793,
		0.999842,
		0.999882,
		0.999914,
		0.99994,
		0.999959,
		0.999974,
		0.999984,
		0.999991,
		0.999996,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999995,
		0.999989,
		0.99998,
		0.999965,
		0.999944,
		0.999914,
		0.999873,
		0.999819,
		0.999749,
		0.99966,
		0.999548,
		0.999412,
		0.999246,
		0.999046,
		0.998808,
		0.998528,
		0.9982,
		0.997819,
		0.99738,
		0.996876,
		0.996301,
		0.995649,
		0.994912,
		0.994084,
		0.993157,
		0.992122,
		0.990973,
		0.9897,
		0.988294,
		0.986746,
		0.985047,
		0.983187,
		0.981155,
		0.978942,
		0.976536,
		0.973927,
		0.971103,
		0.968053,
		0.964764,
		0.961225,
		0.957423,
		0.953345,
		0.94898,
		0.944313,
		0.939331,
		0.934022,
		0.928371,
		0.922365,
		0.91599,
		0.909232,
		0.902078,
		0.894513,
		0.886525,
		0.878098,
		0.869219,
		0.859876,
		0.850054,
		0.839741,
		0.828924,
		0.817592,
		0.805731,
		0.793332,
		0.780384,
		0.766878,
		0.752804,
		0.738155,
		0.722924,
		0.707107,
		0.690698,
		0.673697,
		0.656102,
		0.637915,
		0.61914,
		0.599783,
		0.579853,
		0.559361,
		0.538323,
		0.516759,
		0.494689,
		0.472143,
		0.449152,
		0.425753,
		0.40199,
		0.377913,
		0.353578,
		0.32905,
		0.304401,
		0.279714,
		0.255081,
		0.230604,
		0.2064,
		0.182595,
		0.159331,
		0.136765,
		0.115075,
		0.094456,
		0.075127,
		0.057334,
		0.041348,
		0.027471,
		0.016032,
		0.007387,
		0.001912,
		0.986574,
		0.987358,
		0.988116,
		0.988848,
		0.989554,
		0.990235,
		0.99089,
		0.991519,
		0.992122,
		0.9927,
		0.993253,
		0.99378,
		0.994282,
		0.994759,
		0.995211,
		0.995638,
		0.996042,
		0.996422,
		0.996778,
		0.997111,
		0.997422,
		0.997711,
		0.997978,
		0.998225,
		0.998451,
		0.998657,
		0.998845,
		0.999014,
		0.999166,
		0.999302,
		0.999422,
		0.999527,
		0.999618,
		0.999696,
		0.999763,
		0.999819,
		0.999864,
		0.999901,
		0.999931,
		0.999953,
		0.99997,
		0.999982,
		0.99999,
		0.999995,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999994,
		0.999987,
		0.999976,
		0.999959,
		0.999934,
		0.999899,
		0.999852,
		0.999788,
		0.999707,
		0.999603,
		0.999473,
		0.999313,
		0.999119,
		0.998886,
		0.998608,
		0.99828,
		0.997897,
		0.997451,
		0.996937,
		0.996348,
		0.995675,
		0.994912,
		0.99405,
		0.993081,
		0.991996,
		0.990785,
		0.98944,
		0.987949,
		0.986304,
		0.984492,
		0.982503,
		0.980326,
		0.977948,
		0.975358,
		0.972543,
		0.969489,
		0.966185,
		0.962616,
		0.958769,
		0.95463,
		0.950184,
		0.945417,
		0.940314,
		0.93486,
		0.929041,
		0.92284,
		0.916242,
		0.909232,
		0.901795,
		0.893914,
		0.885575,
		0.876762,
		0.867459,
		0.857652,
		0.847326,
		0.836467,
		0.82506,
		0.813093,
		0.800552,
		0.787425,
		0.773702,
		0.759371,
		0.744423,
		0.728851,
		0.712649,
		0.69581,
		0.678333,
		0.660215,
		0.641458,
		0.622067,
		0.602046,
		0.581406,
		0.560159,
		0.538323,
		0.515919,
		0.492972,
		0.469512,
		0.445578,
		0.42121,
		0.39646,
		0.371384,
		0.346046,
		0.320521,
		0.294891,
		0.269249,
		0.243699,
		0.218359,
		0.19336,
		0.168853,
		0.14501,
		0.122028,
		0.100129,
		0.079568,
		0.060626,
		0.043618,
		0.028884,
		0.016784,
		0.007692,
		0.001978,
		0.977109,
		0.978187,
		0.97924,
		0.980266,
		0.981267,
		0.98224,
		0.983187,
		0.984106,
		0.984999,
		0.985864,
		0.986701,
		0.987511,
		0.988294,
		0.989048,
		0.989775,
		0.990474,
		0.991144,
		0.991788,
		0.992403,
		0.992991,
		0.993551,
		0.994084,
		0.99459,
		0.995069,
		0.995521,
		0.995947,
		0.996348,
		0.996723,
		0.997073,
		0.997398,
		0.9977,
		0.997978,
		0.998234,
		0.998468,
		0.998681,
		0.998873,
		0.999046,
		0.9992,
		0.999337,
		0.999457,
		0.999561,
		0.999651,
		0.999727,
		0.999791,
		0.999844,
		0.999886,
		0.99992,
		0.999946,
		0.999965,
		0.999979,
		0.999988,
		0.999994,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999985,
		0.999972,
		0.999952,
		0.999923,
		0.999882,
		0.999826,
		0.999752,
		0.999656,
		0.999533,
		0.999381,
		0.999193,
		0.998965,
		0.99869,
		0.998364,
		0.997978,
		0.997527,
		0.997003,
		0.996398,
		0.995704,
		0.994912,
		0.994013,
		0.992999,
		0.991857,
		0.990579,
		0.989154,
		0.987569,
		0.985814,
		0.983876,
		0.981742,
		0.9794,
		0.976835,
		0.974035,
		0.970985,
		0.967671,
		0.964076,
		0.960187,
		0.955987,
		0.951461,
		0.946591,
		0.941362,
		0.935757,
		0.929759,
		0.92335,
		0.916514,
		0.909232,
		0.901488,
		0.893264,
		0.884543,
		0.875306,
		0.865538,
		0.855222,
		0.84434,
		0.832877,
		0.820817,
		0.808146,
		0.794849,
		0.780913,
		0.766326,
		0.751077,
		0.735156,
		0.718555,
		0.701268,
		0.683291,
		0.664622,
		0.645261,
		0.625212,
		0.604481,
		0.583079,
		0.561021,
		0.538323,
		0.515011,
		0.491113,
		0.466662,
		0.441699,
		0.416269,
		0.390425,
		0.364225,
		0.337737,
		0.311037,
		0.284216,
		0.257379,
		0.230652,
		0.204182,
		0.178148,
		0.152754,
		0.128242,
		0.104887,
		0.082994,
		0.062899,
		0.04496,
		0.029542,
		0.017013,
		0.007716,
		0.001962,
		0.964076,
		0.965467,
		0.966832,
		0.968173,
		0.969489,
		0.97078,
		0.972044,
		0.973282,
		0.974494,
		0.975678,
		0.976835,
		0.977965,
		0.979066,
		0.98014,
		0.981184,
		0.9822,
		0.983187,
		0.984144,
		0.985072,
		0.98597,
		0.986838,
		0.987677,
		0.988485,
		0.989263,
		0.990011,
		0.990728,
		0.991416,
		0.992073,
		0.9927,
		0.993298,
		0.993865,
		0.994403,
		0.994912,
		0.995392,
		0.995843,
		0.996266,
		0.996662,
		0.99703,
		0.997372,
		0.997688,
		0.997978,
		0.998244,
		0.998486,
		0.998706,
		0.998903,
		0.999079,
		0.999236,
		0.999373,
		0.999493,
		0.999596,
		0.999684,
		0.999758,
		0.999819,
		0.999868,
		0.999907,
		0.999937,
		0.999959,
		0.999975,
		0.999986,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999992,
		0.999982,
		0.999967,
		0.999944,
		0.999909,
		0.999861,
		0.999794,
		0.999707,
		0.999593,
		0.999448,
		0.999268,
		0.999046,
		0.998776,
		0.998451,
		0.998064,
		0.997607,
		0.997073,
		0.996452,
		0.995735,
		0.994912,
		0.993974,
		0.992908,
		0.991705,
		0.990352,
		0.988837,
		0.987147,
		0.985268,
		0.983187,
		0.980889,
		0.978359,
		0.975582,
		0.972543,
		0.969224,
		0.965609,
		0.961681,
		0.957423,
		0.952816,
		0.947842,
		0.942482,
		0.936718,
		0.930531,
		0.9239,
		0.916807,
		0.909232,
		0.901155,
		0.892557,
		0.883417,
		0.873716,
		0.863435,
		0.852555,
		0.841057,
		0.828924,
		0.816138,
		0.802682,
		0.788542,
		0.773702,
		0.758149,
		0.741872,
		0.72486,
		0.707107,
		0.688606,
		0.669354,
		0.649351,
		0.628601,
		0.607109,
		0.584885,
		0.561943,
		0.538297,
		0.513969,
		0.488978,
		0.463351,
		0.437117,
		0.410314,
		0.382989,
		0.355202,
		0.327031,
		0.298578,
		0.269972,
		0.241376,
		0.212988,
		0.185047,
		0.157829,
		0.131646,
		0.106843,
		0.083785,
		0.062848,
		0.044403,
		0.028801,
		0.01635,
		0.007301,
		0.001825,
		0.947082,
		0.948785,
		0.950468,
		0.952128,
		0.953766,
		0.955381,
		0.956972,
		0.95854,
		0.960083,
		0.961602,
		0.963095,
		0.964563,
		0.966004,
		0.967419,
		0.968806,
		0.970166,
		0.971498,
		0.972801,
		0.974075,
		0.975321,
		0.976536,
		0.977722,
		0.978877,
		0.980001,
		0.981094,
		0.982156,
		0.983187,
		0.984185,
		0.985151,
		0.986085,
		0.986986,
		0.987855,
		0.988691,
		0.989494,
		0.990264,
		0.991001,
		0.991705,
		0.992377,
		0.993016,
		0.993622,
		0.994196,
		0.994738,
		0.995249,
		0.995728,
		0.996177,
		0.996595,
		0.996984,
		0.997343,
		0.997675,
		0.997978,
		0.998255,
		0.998506,
		0.998733,
		0.998935,
		0.999115,
		0.999274,
		0.999412,
		0.999531,
		0.999632,
		0.999718,
		0.999788,
		0.999846,
		0.999891,
		0.999926,
		0.999952,
		0.999971,
		0.999984,
		0.999992,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.99999,
		0.999979,
		0.999961,
		0.999933,
		0.999892,
		0.999834,
		0.999755,
		0.999651,
		0.999515,
		0.999343,
		0.999128,
		0.998863,
		0.998541,
		0.998153,
		0.997692,
		0.997147,
		0.996509,
		0.995768,
		0.994912,
		0.99393,
		0.992809,
		0.991538,
		0.990101,
		0.988485,
		0.986675,
		0.984657,
		0.982413,
		0.979927,
		0.977183,
		0.974162,
		0.970847,
		0.967218,
		0.963257,
		0.958943,
		0.954256,
		0.949176,
		0.943681,
		0.937751,
		0.931363,
		0.924495,
		0.917126,
		0.909232,
		0.900792,
		0.891784,
		0.882183,
		0.87197,
		0.861121,
		0.849616,
		0.837433,
		0.824552,
		0.810953,
		0.796619,
		0.781532,
		0.765677,
		0.749038,
		0.731605,
		0.713365,
		0.694312,
		0.674436,
		0.653732,
		0.632191,
		0.609807,
		0.586569,
		0.56247,
		0.537502,
		0.51166,
		0.484947,
		0.457378,
		0.428982,
		0.399815,
		0.369959,
		0.339533,
		0.308696,
		0.277651,
		0.246645,
		0.215973,
		0.185967,
		0.156993,
		0.12944,
		0.103703,
		0.080172,
		0.05921,
		0.041135,
		0.026203,
		0.014591,
		0.006383,
		0.001561,
		0.925841,
		0.927846,
		0.929834,
		0.931803,
		0.933754,
		0.935686,
		0.937598,
		0.939491,
		0.941362,
		0.943213,
		0.945041,
		0.946848,
		0.948631,
		0.950392,
		0.952128,
		0.95384,
		0.955526,
		0.957188,
		0.958822,
		0.960431,
		0.962012,
		0.963565,
		0.96509,
		0.966586,
		0.968053,
		0.969489,
		0.970896,
		0.972271,
		0.973615,
		0.974928,
		0.976208,
		0.977455,
		0.978669,
		0.97985,
		0.980996,
		0.982109,
		0.983187,
		0.98423,
		0.985237,
		0.98621,
		0.987147,
		0.988048,
		0.988913,
		0.989742,
		0.990536,
		0.991293,
		0.992015,
		0.9927,
		0.99335,
		0.993965,
		0.994545,
		0.99509,
		0.995601,
		0.996078,
		0.996521,
		0.996932,
		0.997312,
		0.99766,
		0.997978,
		0.998267,
		0.998528,
		0.998762,
		0.99897,
		0.999153,
		0.999313,
		0.999452,
		0.99957,
		0.999669,
		0.999752,
		0.999819,
		0.999872,
		0.999913,
		0.999944,
		0.999966,
		0.999981,
		0.99999,
		0.999996,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999995,
		0.999988,
		0.999975,
		0.999953,
		0.99992,
		0.999871,
		0.999801,
		0.999707,
		0.999581,
		0.999419,
		0.999212,
		0.998953,
		0.998635,
		0.998248,
		0.997782,
		0.997227,
		0.996571,
		0.995804,
		0.994912,
		0.993882,
		0.9927,
		0.991352,
		0.989821,
		0.988091,
		0.986146,
		0.983968,
		0.981538,
		0.978837,
		0.975845,
		0.972543,
		0.968908,
		0.964919,
		0.960553,
		0.955789,
		0.950601,
		0.944967,
		0.938862,
		0.932261,
		0.92514,
		0.917472,
		0.909232,
		0.900395,
		0.890936,
		0.880827,
		0.870045,
		0.858565,
		0.846361,
		0.833411,
		0.819691,
		0.805178,
		0.789852,
		0.773688,
		0.756663,
		0.73875,
		0.719918,
		0.700132,
		0.679351,
		0.657532,
		0.634625,
		0.610585,
		0.585368,
		0.558942,
		0.531288,
		0.502414,
		0.472358,
		0.441194,
		0.409045,
		0.37608,
		0.342523,
		0.308649,
		0.274783,
		0.241294,
		0.208583,
		0.177071,
		0.147183,
		0.119331,
		0.093895,
		0.071204,
		0.051521,
		0.035026,
		0.021807,
		0.011854,
		0.005056,
		0.001204,
		0.900183,
		0.902465,
		0.904734,
		0.90699,
		0.909232,
		0.91146,
		0.913673,
		0.91587,
		0.918052,
		0.920217,
		0.922365,
		0.924495,
		0.926607,
		0.9287,
		0.930774,
		0.932827,
		0.93486,
		0.936872,
		0.938862,
		0.94083,
		0.942774,
		0.944695,
		0.946591,
		0.948463,
		0.950308,
		0.952128,
		0.953921,
		0.955686,
		0.957423,
		0.959131,
		0.96081,
		0.962458,
		0.964076,
		0.965663,
		0.967218,
		0.968741,
		0.97023,
		0.971686,
		0.973107,
		0.974494,
		0.975845,
		0.977161,
		0.97844,
		0.979683,
		0.980889,
		0.982057,
		0.983187,
		0.984278,
		0.985332,
		0.986346,
		0.987321,
		0.988257,
		0.989154,
		0.990011,
		0.990828,
		0.991607,
		0.992346,
		0.993045,
		0.993706,
		0.994328,
		0.994912,
		0.995458,
		0.995967,
		0.996439,
		0.996876,
		0.997277,
		0.997644,
		0.997978,
		0.99828,
		0.998551,
		0.998793,
		0.999006,
		0.999193,
		0.999355,
		0.999493,
		0.99961,
		0.999707,
		0.999785,
		0.999848,
		0.999897,
		0.999933,
		0.999959,
		0.999977,
		0.999989,
		0.999995,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999994,
		0.999986,
		0.99997,
		0.999944,
		0.999903,
		0.999844,
		0.99976,
		0.999645,
		0.999493,
		0.999296,
		0.999046,
		0.998733,
		0.998347,
		0.997877,
		0.997312,
		0.996638,
		0.995843,
		0.994912,
		0.993829,
		0.992579,
		0.991144,
		0.989507,
		0.987648,
		0.985548,
		0.983187,
		0.980542,
		0.977591,
		0.974313,
		0.970681,
		0.966673,
		0.962262,
		0.957423,
		0.952128,
		0.946351,
		0.940063,
		0.933236,
		0.925841,
		0.91785,
		0.909232,
		0.899959,
		0.890001,
		0.879328,
		0.867908,
		0.85571,
		0.842698,
		0.82883,
		0.81406,
		0.798332,
		0.781582,
		0.763735,
		0.744711,
		0.724419,
		0.702768,
		0.679669,
		0.655041,
		0.628819,
		0.600964,
		0.571468,
		0.540369,
		0.507751,
		0.473757,
		0.438588,
		0.402506,
		0.365832,
		0.328938,
		0.29224,
		0.256183,
		0.221226,
		0.187825,
		0.156413,
		0.127383,
		0.101068,
		0.077731,
		0.057546,
		0.040599,
		0.026875,
		0.01627,
		0.008587,
		0.00355,
		0.000818,
		0.870045,
		0.872571,
		0.875088,
		0.877597,
		0.880097,
		0.882588,
		0.88507,
		0.887541,
		0.890001,
		0.89245,
		0.894888,
		0.897313,
		0.899725,
		0.902123,
		0.904508,
		0.906878,
		0.909232,
		0.911571,
		0.913893,
		0.916199,
		0.918486,
		0.920755,
		0.923006,
		0.925236,
		0.927446,
		0.929636,
		0.931803,
		0.933948,
		0.93607,
		0.938168,
		0.940242,
		0.94229,
		0.944313,
		0.946308,
		0.948277,
		0.950217,
		0.952128,
		0.954009,
		0.955861,
		0.957681,
		0.959469,
		0.961225,
		0.962947,
		0.964635,
		0.966289,
		0.967907,
		0.969489,
		0.971035,
		0.972543,
		0.974012,
		0.975444,
		0.976835,
		0.978187,
		0.979499,
		0.98077,
		0.981999,
		0.983187,
		0.984332,
		0.985435,
		0.986495,
		0.987511,
		0.988485,
		0.989415,
		0.990301,
		0.991144,
		0.991944,
		0.9927,
		0.993413,
		0.994084,
		0.994712,
		0.995298,
		0.995843,
		0.996348,
		0.996812,
		0.997238,
		0.997627,
		0.997978,
		0.998294,
		0.998577,
		0.998827,
		0.999046,
		0.999236,
		0.999399,
		0.999536,
		0.999651,
		0.999744,
		0.999819,
		0.999876,
		0.99992,
		0.999951,
		0.999973,
		0.999986,
		0.999994,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999993,
		0.999982,
		0.999963,
		0.999931,
		0.999882,
		0.999809,
		0.999707,
		0.999567,
		0.999381,
		0.99914,
		0.998834,
		0.998451,
		0.997978,
		0.997403,
		0.996711,
		0.995886,
		0.994912,
		0.993771,
		0.992444,
		0.990912,
		0.989154,
		0.987147,
		0.984868,
		0.982294,
		0.9794,
		0.976158,
		0.972543,
		0.968525,
		0.964076,
		0.959167,
		0.953766,
		0.947842,
		0.941362,
		0.934294,
		0.926602,
		0.918247,
		0.909189,
		0.899376,
		0.888751,
		0.877246,
		0.864779,
		0.851254,
		0.836564,
		0.820585,
		0.803188,
		0.784235,
		0.76359,
		0.741127,
		0.716734,
		0.690328,
		0.661863,
		0.63134,
		0.598814,
		0.564404,
		0.528295,
		0.49074,
		0.45206,
		0.412633,
		0.372893,
		0.333305,
		0.294361,
		0.256554,
		0.220363,
		0.186232,
		0.154555,
		0.125657,
		0.099785,
		0.077097,
		0.057659,
		0.041445,
		0.02834,
		0.018148,
		0.010604,
		0.005387,
		0.002137,
		0.000471,
		0.835475,
		0.838201,
		0.840925,
		0.843645,
		0.846361,
		0.849074,
		0.851782,
		0.854485,
		0.857183,
		0.859876,
		0.862562,
		0.865241,
		0.867913,
		0.870578,
		0.873234,
		0.875881,
		0.878519,
		0.881147,
		0.883765,
		0.886372,
		0.888967,
		0.891549,
		0.894119,
		0.896676,
		0.899218,
		0.901745,
		0.904257,
		0.906753,
		0.909232,
		0.911694,
		0.914137,
		0.916561,
		0.918966,
		0.92135,
		0.923712,
		0.926053,
		0.928371,
		0.930665,
		0.932935,
		0.935179,
		0.937398,
		0.93959,
		0.941754,
		0.943889,
		0.945995,
		0.948071,
		0.950115,
		0.952128,
		0.954108,
		0.956054,
		0.957965,
		0.959841,
		0.961681,
		0.963484,
		0.965249,
		0.966975,
		0.968661,
		0.970307,
		0.971912,
		0.973475,
		0.974996,
		0.976473,
		0.977906,
		0.979295,
		0.980638,
		0.981935,
		0.983187,
		0.984391,
		0.985548,
		0.986658,
		0.98772,
		0.988734,
		0.9897,
		0.990617,
		0.991486,
		0.992308,
		0.993081,
		0.993807,
		0.994485,
		0.995118,
		0.995704,
		0.996245,
		0.996742,
		0.997195,
		0.997607,
		0.997978,
		0.99831,
		0.998605,
		0.998863,
		0.999088,
		0.999281,
		0.999445,
		0.999581,
		0.999693,
		0.999782,
		0.999851,
		0.999903,
		0.999941,
		0.999967,
		0.999983,
		0.999993,
		0.999998,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999991,
		0.999978,
		0.999955,
		0.999916,
		0.999854,
		0.999765,
		0.999638,
		0.999465,
		0.999236,
		0.998938,
		0.99856,
		0.998086,
		0.997501,
		0.99679,
		0.995933,
		0.994912,
		0.993706,
		0.992293,
		0.99065,
		0.988753,
		0.986574,
		0.984089,
		0.981267,
		0.978079,
		0.974494,
		0.970479,
		0.965998,
		0.961014,
		0.955481,
		0.949345,
		0.942546,
		0.935005,
		0.926633,
		0.917318,
		0.906936,
		0.895339,
		0.882368,
		0.867848,
		0.8516,
		0.833445,
		0.81321,
		0.790745,
		0.765928,
		0.73868,
		0.708972,
		0.676837,
		0.642378,
		0.605771,
		0.567267,
		0.527188,
		0.485925,
		0.443925,
		0.401677,
		0.359697,
		0.318512,
		0.278635,
		0.240551,
		0.204699,
		0.171451,
		0.141108,
		0.113883,
		0.089903,
		0.069201,
		0.051724,
		0.037339,
		0.025838,
		0.016957,
		0.010384,
		0.005777,
		0.00278,
		0.001038,
		0.000214,
		0.796619,
		0.799498,
		0.802379,
		0.805262,
		0.808146,
		0.811031,
		0.813918,
		0.816804,
		0.819691,
		0.822577,
		0.825463,
		0.828347,
		0.83123,
		0.834112,
		0.83699,
		0.839866,
		0.842738,
		0.845607,
		0.848471,
		0.851331,
		0.854185,
		0.857034,
		0.859876,
		0.862711,
		0.865538,
		0.868358,
		0.871169,
		0.87397,
		0.876762,
		0.879542,
		0.882312,
		0.88507,
		0.887815,
		0.890546,
		0.893264,
		0.895967,
		0.898654,
		0.901325,
		0.903979,
		0.906615,
		0.909232,
		0.91183,
		0.914407,
		0.916963,
		0.919497,
		0.922008,
		0.924495,
		0.926957,
		0.929393,
		0.931803,
		0.934185,
		0.936538,
		0.938862,
		0.941155,
		0.943417,
		0.945646,
		0.947842,
		0.950003,
		0.952128,
		0.954217,
		0.956268,
		0.95828,
		0.960253,
		0.962186,
		0.964076,
		0.965925,
		0.967729,
		0.969489,
		0.971204,
		0.972873,
		0.974494,
		0.976067,
		0.977591,
		0.979066,
		0.980491,
		0.981865,
		0.983187,
		0.984457,
		0.985674,
		0.986838,
		0.987949,
		0.989007,
		0.990011,
		0.990961,
		0.991857,
		0.9927,
		0.99349,
		0.994227,
		0.994912,
		0.995546,
		0.996128,
		0.996662,
		0.997147,
		0.997585,
		0.997978,
		0.998327,
		0.998635,
		0.998903,
		0.999134,
		0.99933,
		0.999493,
		0.999627,
		0.999735,
		0.999819,
		0.999882,
		0.999928,
		0.999959,
		0.99998,
		0.999991,
		0.999997,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999989,
		0.999973,
		0.999944,
		0.999895,
		0.999819,
		0.999707,
		0.999548,
		0.999333,
		0.999046,
		0.998674,
		0.9982,
		0.997607,
		0.996876,
		0.995985,
		0.994912,
		0.993633,
		0.992122,
		0.990351,
		0.988288,
		0.985898,
		0.983139,
		0.979961,
		0.9763,
		0.972079,
		0.967203,
		0.961557,
		0.955004,
		0.947385,
		0.938519,
		0.928207,
		0.916238,
		0.902391,
		0.886448,
		0.868202,
		0.847468,
		0.824096,
		0.797981,
		0.769076,
		0.737402,
		0.703054,
		0.666205,
		0.627109,
		0.586099,
		0.543575,
		0.5,
		0.455881,
		0.411755,
		0.368169,
		0.325662,
		0.284745,
		0.245885,
		0.209488,
		0.175889,
		0.145338,
		0.118003,
		0.09396,
		0.073201,
		0.055636,
		0.041104,
		0.029382,
		0.020197,
		0.013239,
		0.008178,
		0.004677,
		0.002407,
		0.001059,
		0.000356,
		6.5e-005,
		0.753719,
		0.756699,
		0.759685,
		0.762678,
		0.765677,
		0.768681,
		0.771692,
		0.774707,
		0.777728,
		0.780754,
		0.783784,
		0.786818,
		0.789856,
		0.792898,
		0.795942,
		0.79899,
		0.80204,
		0.805092,
		0.808146,
		0.811201,
		0.814257,
		0.817314,
		0.82037,
		0.823426,
		0.826481,
		0.829535,
		0.832587,
		0.835636,
		0.838682,
		0.841725,
		0.844764,
		0.847798,
		0.850827,
		0.85385,
		0.856866,
		0.859876,
		0.862877,
		0.865871,
		0.868855,
		0.871829,
		0.874792,
		0.877744,
		0.880684,
		0.883611,
		0.886525,
		0.889423,
		0.892307,
		0.895174,
		0.898023,
		0.900855,
		0.903668,
		0.90646,
		0.909232,
		0.911982,
		0.914709,
		0.917412,
		0.92009,
		0.922742,
		0.925367,
		0.927963,
		0.930531,
		0.933068,
		0.935573,
		0.938045,
		0.940484,
		0.942888,
		0.945255,
		0.947585,
		0.949876,
		0.952128,
		0.954338,
		0.956507,
		0.958631,
		0.960712,
		0.962746,
		0.964734,
		0.966673,
		0.968563,
		0.970403,
		0.972191,
		0.973927,
		0.975609,
		0.977237,
		0.97881,
		0.980326,
		0.981785,
		0.983187,
		0.98453,
		0.985814,
		0.987038,
		0.988203,
		0.989308,
		0.990352,
		0.991337,
		0.992261,
		0.993125,
		0.99393,
		0.994676,
		0.995364,
		0.995996,
		0.996571,
		0.997092,
		0.997561,
		0.997978,
		0.998347,
		0.998669,
		0.998946,
		0.999183,
		0.999381,
		0.999544,
		0.999675,
		0.999777,
		0.999854,
		0.999911,
		0.99995,
		0.999975,
		0.999989,
		0.999996,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.999986,
		0.999966,
		0.999929,
		0.999867,
		0.999771,
		0.999629,
		0.999429,
		0.999156,
		0.998793,
		0.99832,
		0.997716,
		0.996952,
		0.995994,
		0.994799,
		0.99331,
		0.991455,
		0.989142,
		0.986257,
		0.982663,
		0.978195,
		0.972665,
		0.96586,
		0.95755,
		0.947489,
		0.935426,
		0.921114,
		0.904319,
		0.884835,
		0.862495,
		0.837182,
		0.808844,
		0.777497,
		0.743241,
		0.706256,
		0.666804,
		0.625228,
		0.581942,
		0.537419,
		0.492178,
		0.446768,
		0.401745,
		0.35766,
		0.315035,
		0.274347,
		0.236018,
		0.200397,
		0.167757,
		0.138285,
		0.112084,
		0.089175,
		0.0695,
		0.052927,
		0.039266,
		0.028272,
		0.019662,
		0.013129,
		0.008349,
		0.005004,
		0.002784,
		0.001406,
		0.000623,
		0.000228,
		6.1e-005,
		0.707107,
		0.710133,
		0.71317,
		0.716218,
		0.719275,
		0.722343,
		0.72542,
		0.728508,
		0.731605,
		0.734711,
		0.737827,
		0.740951,
		0.744084,
		0.747226,
		0.750376,
		0.753533,
		0.756699,
		0.759872,
		0.763052,
		0.76624,
		0.769433,
		0.772634,
		0.77584,
		0.779051,
		0.782268,
		0.78549,
		0.788716,
		0.791947,
		0.795181,
		0.798418,
		0.801659,
		0.804901,
		0.808146,
		0.811392,
		0.814639,
		0.817887,
		0.821134,
		0.824381,
		0.827626,
		0.83087,
		0.834112,
		0.83735,
		0.840584,
		0.843815,
		0.84704,
		0.850259,
		0.853472,
		0.856678,
		0.859876,
		0.863065,
		0.866244,
		0.869413,
		0.872571,
		0.875716,
		0.878848,
		0.881966,
		0.88507,
		0.888157,
		0.891227,
		0.894279,
		0.897313,
		0.900326,
		0.903317,
		0.906287,
		0.909232,
		0.912153,
		0.915048,
		0.917916,
		0.920755,
		0.923565,
		0.926344,
		0.92909,
		0.931803,
		0.934481,
		0.937122,
		0.939726,
		0.94229,
		0.944814,
		0.947296,
		0.949734,
		0.952128,
		0.954475,
		0.956775,
		0.959025,
		0.961225,
		0.963372,
		0.965467,
		0.967506,
		0.969489,
		0.971415,
		0.973282,
		0.975089,
		0.976835,
		0.978519,
		0.98014,
		0.981696,
		0.983187,
		0.984612,
		0.98597,
		0.987261,
		0.988485,
		0.989641,
		0.990728,
		0.991748,
		0.9927,
		0.993585,
		0.994403,
		0.995156,
		0.995843,
		0.996468,
		0.99703,
		0.997533,
		0.997978,
		0.998368,
		0.998706,
		0.998994,
		0.999236,
		0.999435,
		0.999596,
		0.999723,
		0.999819,
		0.999889,
		0.999937,
		0.999968,
		0.999986,
		0.999995,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999994,
		0.999982,
		0.999955,
		0.999904,
		0.999812,
		0.999659,
		0.999412,
		0.999026,
		0.998441,
		0.997575,
		0.996322,
		0.994554,
		0.992111,
		0.988808,
		0.984431,
		0.978745,
		0.971493,
		0.96241,
		0.951225,
		0.937675,
		0.921517,
		0.902538,
		0.88057,
		0.855502,
		0.827286,
		0.795954,
		0.761617,
		0.724468,
		0.684786,
		0.642924,
		0.599306,
		0.554413,
		0.508764,
		0.462908,
		0.417399,
		0.37278,
		0.329566,
		0.288229,
		0.249185,
		0.212781,
		0.17929,
		0.148905,
		0.121741,
		0.097831,
		0.077134,
		0.05954,
		0.044879,
		0.032928,
		0.023426,
		0.016081,
		0.010587,
		0.006631,
		0.003911,
		0.002141,
		0.001066,
		0.00047,
		0.000175,
		5.1e-005,
		1e-005,
		0.657196,
		0.660215,
		0.663248,
		0.666294,
		0.669354,
		0.672427,
		0.675515,
		0.678615,
		0.681729,
		0.684856,
		0.687996,
		0.69115,
		0.694316,
		0.697495,
		0.700686,
		0.70389,
		0.707107,
		0.710335,
		0.713576,
		0.716828,
		0.720092,
		0.723368,
		0.726654,
		0.729952,
		0.73326,
		0.736579,
		0.739908,
		0.743248,
		0.746597,
		0.749955,
		0.753323,
		0.756699,
		0.760084,
		0.763477,
		0.766878,
		0.770286,
		0.773702,
		0.777124,
		0.780552,
		0.783986,
		0.787425,
		0.790869,
		0.794318,
		0.797771,
		0.801227,
		0.804685,
		0.808146,
		0.811609,
		0.815072,
		0.818536,
		0.822,
		0.825463,
		0.828924,
		0.832383,
		0.835839,
		0.839291,
		0.842738,
		0.84618,
		0.849616,
		0.853044,
		0.856465,
		0.859876,
		0.863277,
		0.866667,
		0.870045,
		0.873411,
		0.876762,
		0.880097,
		0.883417,
		0.886718,
		0.890001,
		0.893264,
		0.896506,
		0.899725,
		0.90292,
		0.906089,
		0.909232,
		0.912347,
		0.915432,
		0.918486,
		0.921508,
		0.924495,
		0.927446,
		0.930361,
		0.933236,
		0.93607,
		0.938862,
		0.94161,
		0.944313,
		0.946968,
		0.949573,
		0.952128,
		0.95463,
		0.957078,
		0.959469,
		0.961803,
		0.964076,
		0.966289,
		0.968439,
		0.970524,
		0.972543,
		0.974494,
		0.976376,
		0.978187,
		0.979927,
		0.981594,
		0.983187,
		0.984704,
		0.986146,
		0.987511,
		0.9888,
		0.990011,
		0.991144,
		0.992201,
		0.993181,
		0.994084,
		0.994912,
		0.995666,
		0.996348,
		0.996959,
		0.997501,
		0.997978,
		0.998392,
		0.998747,
		0.999046,
		0.999293,
		0.999493,
		0.999651,
		0.999771,
		0.999859,
		0.99992,
		0.999959,
		0.999982,
		0.999993,
		0.999993,
		0.999982,
		0.999952,
		0.999887,
		0.999756,
		0.999516,
		0.9991,
		0.998418,
		0.997356,
		0.99577,
		0.993485,
		0.990301,
		0.985989,
		0.980301,
		0.97297,
		0.963724,
		0.95229,
		0.938409,
		0.921848,
		0.902408,
		0.879941,
		0.854357,
		0.825636,
		0.793834,
		0.759087,
		0.721612,
		0.681705,
		0.639732,
		0.596121,
		0.551352,
		0.505937,
		0.460409,
		0.415302,
		0.371134,
		0.328396,
		0.287533,
		0.248937,
		0.212934,
		0.179781,
		0.149661,
		0.122681,
		0.098875,
		0.078208,
		0.06058,
		0.045834,
		0.033761,
		0.024116,
		0.016622,
		0.010985,
		0.006903,
		0.00408,
		0.002233,
		0.001107,
		0.000481,
		0.000173,
		4.7e-005,
		0.604481,
		0.607439,
		0.610413,
		0.613403,
		0.61641,
		0.619433,
		0.622473,
		0.625529,
		0.628601,
		0.631689,
		0.634794,
		0.637915,
		0.641052,
		0.644206,
		0.647376,
		0.650561,
		0.653763,
		0.656981,
		0.660215,
		0.663465,
		0.66673,
		0.670011,
		0.673308,
		0.67662,
		0.679948,
		0.683291,
		0.686649,
		0.690022,
		0.69341,
		0.696813,
		0.70023,
		0.703661,
		0.707107,
		0.710566,
		0.71404,
		0.717527,
		0.721027,
		0.72454,
		0.728066,
		0.731605,
		0.735156,
		0.738718,
		0.742293,
		0.745878,
		0.749475,
		0.753082,
		0.756699,
		0.760326,
		0.763962,
		0.767608,
		0.771261,
		0.774923,
		0.778592,
		0.782268,
		0.785951,
		0.789639,
		0.793332,
		0.797031,
		0.800733,
		0.804438,
		0.808146,
		0.811856,
		0.815567,
		0.819279,
		0.82299,
		0.826699,
		0.830407,
		0.834112,
		0.837812,
		0.841508,
		0.845197,
		0.84888,
		0.852555,
		0.85622,
		0.859876,
		0.863519,
		0.867151,
		0.870768,
		0.874369,
		0.877955,
		0.881522,
		0.88507,
		0.888597,
		0.892101,
		0.895582,
		0.899037,
		0.902465,
		0.905864,
		0.909232,
		0.912568,
		0.91587,
		0.919137,
		0.922365,
		0.925553,
		0.9287,
		0.931803,
		0.93486,
		0.93787,
		0.94083,
		0.943738,
		0.946591,
		0.949389,
		0.952128,
		0.954807,
		0.957423,
		0.959974,
		0.962458,
		0.964874,
		0.967218,
		0.969489,
		0.971686,
		0.973805,
		0.975845,
		0.977805,
		0.979683,
		0.981477,
		0.983187,
		0.98481,
		0.986346,
		0.987794,
		0.989154,
		0.990425,
		0.991607,
		0.9927,
		0.993706,
		0.994625,
		0.995458,
		0.996208,
		0.996876,
		0.997465,
		0.997978,
		0.998419,
		0.998793,
		0.999102,
		0.999349,
		0.999536,
		0.999659,
		0.999705,
		0.999655,
		0.999471,
		0.9991,
		0.998467,
		0.997473,
		0.99599,
		0.993865,
		0.990914,
		0.986929,
		0.981678,
		0.974912,
		0.966371,
		0.955795,
		0.942932,
		0.927549,
		0.909445,
		0.888462,
		0.864494,
		0.837497,
		0.807497,
		0.774593,
		0.738959,
		0.700845,
		0.660568,
		0.618508,
		0.575095,
		0.530799,
		0.486112,
		0.441538,
		0.397574,
		0.354699,
		0.313356,
		0.273949,
		0.236823,
		0.202269,
		0.170509,
		0.141698,
		0.115924,
		0.09321,
		0.073513,
		0.056733,
		0.042718,
		0.031269,
		0.02215,
		0.015095,
		0.009823,
		0.006041,
		0.003462,
		0.001811,
		0.000838,
		0.000326,
		9.6e-005,
		1.8e-005,
		0.549528,
		0.552374,
		0.555238,
		0.55812,
		0.561021,
		0.563939,
		0.566876,
		0.569832,
		0.572806,
		0.575799,
		0.57881,
		0.58184,
		0.584888,
		0.587956,
		0.591042,
		0.594146,
		0.59727,
		0.600413,
		0.603574,
		0.606755,
		0.609954,
		0.613173,
		0.61641,
		0.619667,
		0.622942,
		0.626236,
		0.629549,
		0.632882,
		0.636233,
		0.639602,
		0.642991,
		0.646399,
		0.649825,
		0.65327,
		0.656733,
		0.660215,
		0.663715,
		0.667234,
		0.670771,
		0.674326,
		0.677898,
		0.681489,
		0.685097,
		0.688723,
		0.692366,
		0.696026,
		0.699703,
		0.703397,
		0.707107,
		0.710833,
		0.714575,
		0.718333,
		0.722107,
		0.725895,
		0.729698,
		0.733515,
		0.737347,
		0.741192,
		0.74505,
		0.748921,
		0.752804,
		0.756699,
		0.760605,
		0.764523,
		0.76845,
		0.772387,
		0.776333,
		0.780288,
		0.78425,
		0.78822,
		0.792195,
		0.796177,
		0.800163,
		0.804153,
		0.808146,
		0.812141,
		0.816138,
		0.820135,
		0.824131,
		0.828126,
		0.832117,
		0.836105,
		0.840087,
		0.844063,
		0.848031,
		0.85199,
		0.855939,
		0.859876,
		0.863799,
		0.867708,
		0.8716,
		0.875474,
		0.879329,
		0.883162,
		0.886971,
		0.890756,
		0.894513,
		0.898242,
		0.901939,
		0.905603,
		0.909232,
		0.912824,
		0.916375,
		0.919885,
		0.92335,
		0.926769,
		0.930138,
		0.933455,
		0.936718,
		0.939924,
		0.943071,
		0.946156,
		0.949176,
		0.952128,
		0.95501,
		0.957819,
		0.960553,
		0.963209,
		0.965784,
		0.968276,
		0.970681,
		0.972999,
		0.975226,
		0.97736,
		0.9794,
		0.981342,
		0.983187,
		0.984931,
		0.986574,
		0.988116,
		0.989554,
		0.99089,
		0.992122,
		0.993251,
		0.994276,
		0.995193,
		0.995994,
		0.996665,
		0.997181,
		0.997502,
		0.997575,
		0.997321,
		0.996642,
		0.995414,
		0.993485,
		0.990679,
		0.986797,
		0.98162,
		0.974912,
		0.966433,
		0.955943,
		0.943208,
		0.928019,
		0.910193,
		0.889587,
		0.866109,
		0.839721,
		0.810449,
		0.778389,
		0.743703,
		0.706621,
		0.667436,
		0.626498,
		0.584203,
		0.540983,
		0.497293,
		0.4536,
		0.410368,
		0.368043,
		0.32705,
		0.287772,
		0.25055,
		0.215674,
		0.183376,
		0.153832,
		0.127154,
		0.103399,
		0.082561,
		0.064584,
		0.049356,
		0.036721,
		0.026482,
		0.018409,
		0.012245,
		0.007717,
		0.004545,
		0.002452,
		0.001175,
		0.000475,
		0.000147,
		2.8e-005,
		0.492972,
		0.495659,
		0.498366,
		0.501091,
		0.503836,
		0.5066,
		0.509384,
		0.512188,
		0.515011,
		0.517854,
		0.520718,
		0.523601,
		0.526505,
		0.529428,
		0.532373,
		0.535338,
		0.538323,
		0.54133,
		0.544357,
		0.547405,
		0.550474,
		0.553565,
		0.556677,
		0.55981,
		0.562964,
		0.56614,
		0.569338,
		0.572558,
		0.575799,
		0.579062,
		0.582346,
		0.585653,
		0.588982,
		0.592333,
		0.595706,
		0.599101,
		0.602518,
		0.605958,
		0.60942,
		0.612904,
		0.61641,
		0.619939,
		0.62349,
		0.627063,
		0.630658,
		0.634275,
		0.637915,
		0.641577,
		0.645261,
		0.648967,
		0.652694,
		0.656444,
		0.660215,
		0.664008,
		0.667822,
		0.671658,
		0.675515,
		0.679392,
		0.683291,
		0.68721,
		0.69115,
		0.695109,
		0.699089,
		0.703088,
		0.707107,
		0.711144,
		0.715201,
		0.719275,
		0.723368,
		0.727478,
		0.731605,
		0.735749,
		0.739908,
		0.744084,
		0.748275,
		0.75248,
		0.756699,
		0.760931,
		0.765177,
		0.769433,
		0.773702,
		0.77798,
		0.782268,
		0.786565,
		0.790869,
		0.795181,
		0.799498,
		0.80382,
		0.808146,
		0.812474,
		0.816804,
		0.821134,
		0.825463,
		0.829789,
		0.834112,
		0.838428,
		0.842738,
		0.84704,
		0.851331,
		0.85561,
		0.859876,
		0.864126,
		0.868358,
		0.872571,
		0.876762,
		0.880929,
		0.88507,
		0.889182,
		0.893264,
		0.897313,
		0.901325,
		0.905299,
		0.909232,
		0.913121,
		0.916963,
		0.920755,
		0.924495,
		0.928179,
		0.931803,
		0.935365,
		0.938862,
		0.94229,
		0.945646,
		0.948926,
		0.952128,
		0.955247,
		0.95828,
		0.961225,
		0.964076,
		0.966832,
		0.969489,
		0.972044,
		0.974494,
		0.976834,
		0.979061,
		0.981167,
		0.983139,
		0.98496,
		0.986599,
		0.988014,
		0.989142,
		0.989904,
		0.990193,
		0.98988,
		0.988808,
		0.986795,
		0.983637,
		0.979108,
		0.97297,
		0.96498,
		0.954896,
		0.942488,
		0.927549,
		0.909906,
		0.889428,
		0.866039,
		0.839721,
		0.81052,
		0.778552,
		0.743999,
		0.707107,
		0.668181,
		0.627576,
		0.585686,
		0.542935,
		0.499767,
		0.456627,
		0.413956,
		0.372177,
		0.331687,
		0.292845,
		0.255969,
		0.22133,
		0.189147,
		0.159588,
		0.132767,
		0.108744,
		0.087528,
		0.069081,
		0.053314,
		0.040099,
		0.029267,
		0.020615,
		0.013913,
		0.00891,
		0.005342,
		0.002939,
		0.00144,
		0.000596,
		0.00019,
		3.8e-005,
		0.435513,
		0.438,
		0.440506,
		0.443032,
		0.445578,
		0.448143,
		0.450728,
		0.453333,
		0.455958,
		0.458604,
		0.46127,
		0.463957,
		0.466665,
		0.469393,
		0.472143,
		0.474914,
		0.477707,
		0.480522,
		0.483358,
		0.486216,
		0.489096,
		0.491999,
		0.494924,
		0.497872,
		0.500843,
		0.503836,
		0.506853,
		0.509893,
		0.512956,
		0.516043,
		0.519153,
		0.522288,
		0.525446,
		0.528629,
		0.531836,
		0.535067,
		0.538323,
		0.541604,
		0.54491,
		0.54824,
		0.551596,
		0.554977,
		0.558383,
		0.561815,
		0.565272,
		0.568755,
		0.572264,
		0.575799,
		0.579359,
		0.582946,
		0.586559,
		0.590198,
		0.593863,
		0.597555,
		0.601273,
		0.605018,
		0.608789,
		0.612586,
		0.61641,
		0.620261,
		0.624138,
		0.628041,
		0.631971,
		0.635927,
		0.63991,
		0.643919,
		0.647954,
		0.652015,
		0.656102,
		0.660215,
		0.664354,
		0.668518,
		0.672708,
		0.676922,
		0.681162,
		0.685426,
		0.689715,
		0.694027,
		0.698364,
		0.702724,
		0.707107,
		0.711512,
		0.71594,
		0.72039,
		0.72486,
		0.729352,
		0.733863,
		0.738394,
		0.742944,
		0.747512,
		0.752097,
		0.756699,
		0.761317,
		0.76595,
		0.770596,
		0.775256,
		0.779928,
		0.784611,
		0.789303,
		0.794004,
		0.798713,
		0.803427,
		0.808146,
		0.812868,
		0.817592,
		0.822315,
		0.827036,
		0.831754,
		0.836467,
		0.841172,
		0.845868,
		0.850552,
		0.855222,
		0.859876,
		0.864511,
		0.869125,
		0.873716,
		0.87828,
		0.882814,
		0.887317,
		0.891784,
		0.896212,
		0.900598,
		0.90494,
		0.909232,
		0.913472,
		0.917657,
		0.921781,
		0.925841,
		0.929834,
		0.933754,
		0.937598,
		0.941362,
		0.94504,
		0.948626,
		0.952111,
		0.955481,
		0.958714,
		0.961779,
		0.96463,
		0.967203,
		0.969415,
		0.971156,
		0.972293,
		0.972665,
		0.972083,
		0.970337,
		0.967195,
		0.96241,
		0.955728,
		0.946897,
		0.935677,
		0.921848,
		0.905228,
		0.885676,
		0.863107,
		0.837497,
		0.808892,
		0.777408,
		0.743232,
		0.706621,
		0.667894,
		0.627422,
		0.585621,
		0.542935,
		0.499827,
		0.456759,
		0.414184,
		0.372533,
		0.332201,
		0.293546,
		0.256874,
		0.222444,
		0.19046,
		0.161075,
		0.134389,
		0.110452,
		0.089266,
		0.070788,
		0.054932,
		0.041577,
		0.030563,
		0.021703,
		0.014781,
		0.009563,
		0.005799,
		0.003231,
		0.001605,
		0.000676,
		0.000219,
		4.4e-005,
		0.377913,
		0.380165,
		0.382435,
		0.384724,
		0.387032,
		0.38936,
		0.391707,
		0.394074,
		0.39646,
		0.398867,
		0.401293,
		0.403741,
		0.406208,
		0.408697,
		0.411206,
		0.413737,
		0.416289,
		0.418863,
		0.421459,
		0.424076,
		0.426716,
		0.429378,
		0.432063,
		0.43477,
		0.437501,
		0.440255,
		0.443032,
		0.445833,
		0.448658,
		0.451507,
		0.454381,
		0.457278,
		0.460201,
		0.463149,
		0.466121,
		0.469119,
		0.472143,
		0.475193,
		0.478268,
		0.48137,
		0.484498,
		0.487653,
		0.490835,
		0.494044,
		0.497281,
		0.500544,
		0.503836,
		0.507156,
		0.510503,
		0.51388,
		0.517284,
		0.520718,
		0.52418,
		0.527672,
		0.531193,
		0.534743,
		0.538323,
		0.541933,
		0.545574,
		0.549244,
		0.552945,
		0.556677,
		0.560439,
		0.564232,
		0.568056,
		0.571912,
		0.575799,
		0.579717,
		0.583667,
		0.587648,
		0.591661,
		0.595706,
		0.599783,
		0.603892,
		0.608032,
		0.612205,
		0.61641,
		0.620647,
		0.624916,
		0.629217,
		0.63355,
		0.637915,
		0.642312,
		0.64674,
		0.651201,
		0.655692,
		0.660215,
		0.664769,
		0.669354,
		0.673969,
		0.678615,
		0.683291,
		0.687996,
		0.692731,
		0.697495,
		0.702287,
		0.707107,
		0.711954,
		0.716828,
		0.721729,
		0.726654,
		0.731605,
		0.736579,
		0.741577,
		0.746597,
		0.751638,
		0.756699,
		0.761779,
		0.766878,
		0.771993,
		0.777124,
		0.782268,
		0.787425,
		0.792593,
		0.797771,
		0.802956,
		0.808146,
		0.81334,
		0.818536,
		0.823732,
		0.828924,
		0.834112,
		0.839291,
		0.84446,
		0.849616,
		0.854755,
		0.859876,
		0.864974,
		0.870045,
		0.875088,
		0.880097,
		0.88507,
		0.890001,
		0.894886,
		0.89972,
		0.904491,
		0.909189,
		0.91379,
		0.918264,
		0.922566,
		0.926633,
		0.930382,
		0.933706,
		0.936471,
		0.938519,
		0.939661,
		0.939685,
		0.938357,
		0.935426,
		0.930634,
		0.923719,
		0.914431,
		0.902538,
		0.887842,
		0.870185,
		0.849464,
		0.825636,
		0.798725,
		0.768831,
		0.736123,
		0.700845,
		0.663304,
		0.623865,
		0.582942,
		0.540983,
		0.498456,
		0.455837,
		0.413595,
		0.372177,
		0.332002,
		0.293444,
		0.256834,
		0.222444,
		0.190493,
		0.161142,
		0.134495,
		0.110603,
		0.089464,
		0.071031,
		0.055211,
		0.041877,
		0.030868,
		0.021994,
		0.015042,
		0.00978,
		0.005965,
		0.003347,
		0.001676,
		0.000712,
		0.000233,
		4.8e-005,
		0.320992,
		0.32298,
		0.324985,
		0.327009,
		0.32905,
		0.331109,
		0.333186,
		0.335282,
		0.337397,
		0.339531,
		0.341684,
		0.343856,
		0.346048,
		0.34826,
		0.350491,
		0.352743,
		0.355015,
		0.357308,
		0.359622,
		0.361957,
		0.364313,
		0.366691,
		0.369091,
		0.371512,
		0.373956,
		0.376422,
		0.378911,
		0.381424,
		0.383959,
		0.386518,
		0.3891,
		0.391707,
		0.394338,
		0.396993,
		0.399673,
		0.402378,
		0.405109,
		0.407865,
		0.410647,
		0.413455,
		0.416289,
		0.41915,
		0.422038,
		0.424954,
		0.427896,
		0.430867,
		0.433865,
		0.436892,
		0.439948,
		0.443032,
		0.446146,
		0.449289,
		0.452462,
		0.455665,
		0.458899,
		0.462163,
		0.465459,
		0.468785,
		0.472143,
		0.475533,
		0.478955,
		0.48241,
		0.485897,
		0.489418,
		0.492972,
		0.496559,
		0.50018,
		0.503836,
		0.507526,
		0.511251,
		0.515011,
		0.518807,
		0.522638,
		0.526505,
		0.530408,
		0.534347,
		0.538323,
		0.542336,
		0.546387,
		0.550474,
		0.5546,
		0.558763,
		0.562964,
		0.567204,
		0.571482,
		0.575799,
		0.580154,
		0.584549,
		0.588982,
		0.593455,
		0.597967,
		0.602518,
		0.607109,
		0.61174,
		0.61641,
		0.62112,
		0.625869,
		0.630658,
		0.635486,
		0.640354,
		0.645261,
		0.650207,
		0.655192,
		0.660215,
		0.665277,
		0.670377,
		0.675515,
		0.68069,
		0.685901,
		0.69115,
		0.696434,
		0.701753,
		0.707107,
		0.712494,
		0.717915,
		0.723368,
		0.728851,
		0.734365,
		0.739908,
		0.745479,
		0.751077,
		0.756699,
		0.762345,
		0.768013,
		0.773702,
		0.779408,
		0.785132,
		0.790869,
		0.796619,
		0.802379,
		0.808146,
		0.813918,
		0.819691,
		0.825462,
		0.831226,
		0.836975,
		0.842698,
		0.848375,
		0.853978,
		0.859466,
		0.864779,
		0.86984,
		0.874547,
		0.878774,
		0.882368,
		0.885148,
		0.88691,
		0.887425,
		0.886448,
		0.883723,
		0.878988,
		0.871991,
		0.862495,
		0.850289,
		0.835202,
		0.817112,
		0.795954,
		0.771729,
		0.744506,
		0.714427,
		0.681705,
		0.646618,
		0.609506,
		0.570757,
		0.530799,
		0.490085,
		0.449081,
		0.408251,
		0.368043,
		0.328883,
		0.291157,
		0.255208,
		0.22133,
		0.18976,
		0.160683,
		0.134224,
		0.110452,
		0.089387,
		0.070995,
		0.055198,
		0.041877,
		0.030876,
		0.022008,
		0.015059,
		0.0098,
		0.005985,
		0.003364,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.265625,
		0.267329,
		0.269049,
		0.270786,
		0.272538,
		0.274307,
		0.276092,
		0.277895,
		0.279714,
		0.28155,
		0.283404,
		0.285276,
		0.287165,
		0.289073,
		0.290998,
		0.292943,
		0.294906,
		0.296887,
		0.298888,
		0.300909,
		0.302949,
		0.305009,
		0.307089,
		0.309189,
		0.31131,
		0.313452,
		0.315615,
		0.317799,
		0.320005,
		0.322233,
		0.324482,
		0.326755,
		0.32905,
		0.331368,
		0.333709,
		0.336073,
		0.338462,
		0.340874,
		0.343311,
		0.345773,
		0.34826,
		0.350772,
		0.353309,
		0.355873,
		0.358463,
		0.361079,
		0.363722,
		0.366393,
		0.369091,
		0.371816,
		0.37457,
		0.377353,
		0.380165,
		0.383005,
		0.385876,
		0.388776,
		0.391707,
		0.394668,
		0.397661,
		0.400685,
		0.403741,
		0.406828,
		0.409949,
		0.413102,
		0.416289,
		0.41951,
		0.422765,
		0.426054,
		0.429378,
		0.432738,
		0.436133,
		0.439564,
		0.443032,
		0.446537,
		0.45008,
		0.45366,
		0.457278,
		0.460936,
		0.464632,
		0.468368,
		0.472143,
		0.475959,
		0.479816,
		0.483714,
		0.487653,
		0.491635,
		0.495659,
		0.499726,
		0.503836,
		0.50799,
		0.512188,
		0.51643,
		0.520718,
		0.52505,
		0.529428,
		0.533853,
		0.538323,
		0.542841,
		0.547405,
		0.552017,
		0.556677,
		0.561384,
		0.56614,
		0.570945,
		0.575799,
		0.580701,
		0.585653,
		0.590655,
		0.595706,
		0.600807,
		0.605958,
		0.611159,
		0.61641,
		0.621711,
		0.627063,
		0.632464,
		0.637915,
		0.643416,
		0.648967,
		0.654566,
		0.660215,
		0.665912,
		0.671658,
		0.677451,
		0.683291,
		0.689177,
		0.695109,
		0.701086,
		0.707107,
		0.71317,
		0.719275,
		0.72542,
		0.731605,
		0.737825,
		0.74408,
		0.750362,
		0.756663,
		0.762966,
		0.769247,
		0.77547,
		0.781582,
		0.787513,
		0.793172,
		0.798444,
		0.803188,
		0.807238,
		0.810403,
		0.812471,
		0.81321,
		0.812377,
		0.809723,
		0.805002,
		0.797981,
		0.788449,
		0.77623,
		0.761188,
		0.743241,
		0.722367,
		0.698606,
		0.672067,
		0.642924,
		0.611418,
		0.577846,
		0.542556,
		0.505937,
		0.468408,
		0.430401,
		0.392355,
		0.354699,
		0.317842,
		0.282165,
		0.248009,
		0.215674,
		0.185408,
		0.15741,
		0.131824,
		0.108744,
		0.088209,
		0.070213,
		0.054701,
		0.041577,
		0.030705,
		0.021916,
		0.015015,
		0.00978,
		0.005977,
		0.003361,
		0.001688,
		0.000719,
		0.000237,
		4.9e-005,
		0.212739,
		0.214149,
		0.215573,
		0.217011,
		0.218463,
		0.219929,
		0.221409,
		0.222905,
		0.224414,
		0.225939,
		0.227479,
		0.229035,
		0.230606,
		0.232193,
		0.233795,
		0.235414,
		0.237049,
		0.238701,
		0.24037,
		0.242056,
		0.243759,
		0.245479,
		0.247217,
		0.248973,
		0.250747,
		0.25254,
		0.254351,
		0.256181,
		0.258031,
		0.2599,
		0.261788,
		0.263696,
		0.265625,
		0.267574,
		0.269544,
		0.271535,
		0.273547,
		0.27558,
		0.277636,
		0.279714,
		0.281814,
		0.283937,
		0.286083,
		0.288253,
		0.290446,
		0.292664,
		0.294906,
		0.297172,
		0.299464,
		0.301781,
		0.304124,
		0.306492,
		0.308888,
		0.31131,
		0.31376,
		0.316237,
		0.318742,
		0.321275,
		0.323837,
		0.326429,
		0.32905,
		0.331701,
		0.334382,
		0.337094,
		0.339837,
		0.342613,
		0.34542,
		0.34826,
		0.351133,
		0.354039,
		0.35698,
		0.359954,
		0.362964,
		0.366009,
		0.369091,
		0.372208,
		0.375363,
		0.378555,
		0.381784,
		0.385053,
		0.38836,
		0.391707,
		0.395094,
		0.398522,
		0.40199,
		0.405501,
		0.409054,
		0.41265,
		0.416289,
		0.419973,
		0.423701,
		0.427474,
		0.431293,
		0.435159,
		0.439072,
		0.443032,
		0.447041,
		0.451099,
		0.455206,
		0.459363,
		0.463572,
		0.467831,
		0.472143,
		0.476508,
		0.480925,
		0.485397,
		0.489923,
		0.494505,
		0.499142,
		0.503836,
		0.508587,
		0.513395,
		0.518262,
		0.523188,
		0.528173,
		0.533218,
		0.538323,
		0.54349,
		0.548718,
		0.554008,
		0.559361,
		0.564777,
		0.570256,
		0.575799,
		0.581406,
		0.587077,
		0.592813,
		0.598615,
		0.604481,
		0.610413,
		0.61641,
		0.622473,
		0.628601,
		0.634793,
		0.641049,
		0.647364,
		0.653732,
		0.66014,
		0.666569,
		0.672987,
		0.679351,
		0.685602,
		0.691658,
		0.697422,
		0.702768,
		0.707552,
		0.711604,
		0.714734,
		0.716734,
		0.717384,
		0.716456,
		0.713725,
		0.708972,
		0.701997,
		0.692627,
		0.680727,
		0.666205,
		0.649019,
		0.629185,
		0.606781,
		0.581942,
		0.554862,
		0.525792,
		0.495029,
		0.462908,
		0.429797,
		0.39608,
		0.36215,
		0.328396,
		0.295193,
		0.262895,
		0.231825,
		0.202269,
		0.174471,
		0.148632,
		0.124906,
		0.103399,
		0.084171,
		0.067237,
		0.05257,
		0.040099,
		0.029718,
		0.021286,
		0.014633,
		0.009563,
		0.005863,
		0.003307,
		0.001665,
		0.000712,
		0.000235,
		4.8e-005,
		0.163312,
		0.164428,
		0.165555,
		0.166693,
		0.167843,
		0.169005,
		0.170178,
		0.171364,
		0.172561,
		0.173771,
		0.174994,
		0.17623,
		0.177478,
		0.178739,
		0.180014,
		0.181302,
		0.182603,
		0.183919,
		0.185248,
		0.186592,
		0.18795,
		0.189322,
		0.19071,
		0.192112,
		0.193529,
		0.194962,
		0.196411,
		0.197875,
		0.199356,
		0.200852,
		0.202366,
		0.203896,
		0.205443,
		0.207007,
		0.208589,
		0.210189,
		0.211807,
		0.213443,
		0.215097,
		0.21677,
		0.218463,
		0.220175,
		0.221906,
		0.223658,
		0.225429,
		0.227222,
		0.229035,
		0.230869,
		0.232725,
		0.234603,
		0.236502,
		0.238425,
		0.24037,
		0.242338,
		0.24433,
		0.246346,
		0.248386,
		0.25045,
		0.25254,
		0.254655,
		0.256796,
		0.258963,
		0.261156,
		0.263377,
		0.265625,
		0.267901,
		0.270205,
		0.272538,
		0.2749,
		0.277292,
		0.279714,
		0.282166,
		0.28465,
		0.287165,
		0.289713,
		0.292293,
		0.294906,
		0.297552,
		0.300233,
		0.302949,
		0.3057,
		0.308487,
		0.31131,
		0.31417,
		0.317068,
		0.320005,
		0.32298,
		0.325995,
		0.32905,
		0.332145,
		0.335282,
		0.338462,
		0.341684,
		0.34495,
		0.34826,
		0.351615,
		0.355015,
		0.358463,
		0.361957,
		0.365499,
		0.369091,
		0.372731,
		0.376422,
		0.380165,
		0.383959,
		0.387806,
		0.391707,
		0.395662,
		0.399673,
		0.403741,
		0.407865,
		0.412048,
		0.416289,
		0.420591,
		0.424954,
		0.429378,
		0.433865,
		0.438416,
		0.443032,
		0.447714,
		0.452462,
		0.457278,
		0.462163,
		0.467118,
		0.472143,
		0.47724,
		0.48241,
		0.487653,
		0.492972,
		0.498366,
		0.503836,
		0.509384,
		0.515011,
		0.520717,
		0.526502,
		0.532363,
		0.538297,
		0.544295,
		0.550341,
		0.556411,
		0.56247,
		0.56847,
		0.574344,
		0.580011,
		0.585368,
		0.590294,
		0.594644,
		0.59826,
		0.600964,
		0.602567,
		0.602874,
		0.601687,
		0.598814,
		0.594078,
		0.58732,
		0.578414,
		0.567267,
		0.553828,
		0.538097,
		0.520121,
		0.5,
		0.477883,
		0.453969,
		0.428498,
		0.401745,
		0.374015,
		0.34563,
		0.316924,
		0.288229,
		0.259871,
		0.232157,
		0.205375,
		0.179781,
		0.1556,
		0.133017,
		0.112184,
		0.09321,
		0.076164,
		0.061078,
		0.047945,
		0.036721,
		0.027329,
		0.019659,
		0.013573,
		0.00891,
		0.005488,
		0.00311,
		0.001573,
		0.000676,
		0.000224,
		4.6e-005,
		0.118371,
		0.119202,
		0.120042,
		0.120891,
		0.121749,
		0.122616,
		0.123492,
		0.124377,
		0.125272,
		0.126177,
		0.127091,
		0.128015,
		0.128949,
		0.129893,
		0.130847,
		0.131812,
		0.132788,
		0.133774,
		0.134771,
		0.135779,
		0.136799,
		0.137829,
		0.138872,
		0.139926,
		0.140991,
		0.142069,
		0.14316,
		0.144262,
		0.145377,
		0.146505,
		0.147646,
		0.148801,
		0.149968,
		0.151149,
		0.152345,
		0.153554,
		0.154777,
		0.156015,
		0.157267,
		0.158534,
		0.159817,
		0.161115,
		0.162428,
		0.163757,
		0.165103,
		0.166465,
		0.167843,
		0.169238,
		0.170651,
		0.172081,
		0.173528,
		0.174994,
		0.176478,
		0.177981,
		0.179502,
		0.181043,
		0.182603,
		0.184184,
		0.185784,
		0.187405,
		0.189047,
		0.19071,
		0.192394,
		0.194101,
		0.19583,
		0.197581,
		0.199356,
		0.201154,
		0.202976,
		0.204822,
		0.206693,
		0.208589,
		0.210511,
		0.212459,
		0.214433,
		0.216434,
		0.218463,
		0.220519,
		0.222604,
		0.224718,
		0.226862,
		0.229035,
		0.231239,
		0.233473,
		0.23574,
		0.238038,
		0.24037,
		0.242735,
		0.245133,
		0.247567,
		0.250035,
		0.25254,
		0.255081,
		0.257659,
		0.260276,
		0.262931,
		0.265625,
		0.268359,
		0.271135,
		0.273952,
		0.276811,
		0.279714,
		0.282661,
		0.285652,
		0.28869,
		0.291774,
		0.294906,
		0.298086,
		0.301315,
		0.304595,
		0.307926,
		0.31131,
		0.314747,
		0.318238,
		0.321785,
		0.325389,
		0.32905,
		0.332769,
		0.336549,
		0.34039,
		0.344293,
		0.34826,
		0.352291,
		0.356389,
		0.360554,
		0.364787,
		0.369091,
		0.373465,
		0.377913,
		0.382435,
		0.387032,
		0.391707,
		0.39646,
		0.401293,
		0.406206,
		0.411199,
		0.416269,
		0.421411,
		0.426612,
		0.431857,
		0.437117,
		0.442357,
		0.447525,
		0.452559,
		0.457378,
		0.461886,
		0.465973,
		0.46951,
		0.472358,
		0.474364,
		0.475373,
		0.475222,
		0.473757,
		0.470829,
		0.466305,
		0.460076,
		0.45206,
		0.442204,
		0.430498,
		0.416966,
		0.401677,
		0.384738,
		0.366296,
		0.346533,
		0.325662,
		0.303917,
		0.281552,
		0.25883,
		0.236018,
		0.213377,
		0.191158,
		0.169596,
		0.148905,
		0.129275,
		0.110866,
		0.09381,
		0.078208,
		0.064131,
		0.051615,
		0.040669,
		0.031269,
		0.023365,
		0.016877,
		0.011703,
		0.007717,
		0.004775,
		0.002719,
		0.001382,
		0.000596,
		0.000199,
		4.1e-005,
		0.078987,
		0.079556,
		0.080132,
		0.080713,
		0.081301,
		0.081895,
		0.082496,
		0.083104,
		0.083718,
		0.084339,
		0.084966,
		0.085601,
		0.086243,
		0.086892,
		0.087549,
		0.088212,
		0.088884,
		0.089563,
		0.090249,
		0.090944,
		0.091646,
		0.092357,
		0.093076,
		0.093803,
		0.094539,
		0.095283,
		0.096037,
		0.096799,
		0.09757,
		0.09835,
		0.09914,
		0.099939,
		0.100747,
		0.101566,
		0.102394,
		0.103233,
		0.104082,
		0.104941,
		0.10581,
		0.106691,
		0.107582,
		0.108485,
		0.109399,
		0.110324,
		0.111261,
		0.11221,
		0.113171,
		0.114144,
		0.11513,
		0.116129,
		0.11714,
		0.118164,
		0.119202,
		0.120254,
		0.121319,
		0.122398,
		0.123492,
		0.1246,
		0.125723,
		0.126861,
		0.128015,
		0.129184,
		0.130369,
		0.13157,
		0.132788,
		0.134022,
		0.135274,
		0.136543,
		0.137829,
		0.139134,
		0.140457,
		0.141799,
		0.14316,
		0.14454,
		0.14594,
		0.14736,
		0.148801,
		0.150262,
		0.151745,
		0.15325,
		0.154777,
		0.156326,
		0.157899,
		0.159495,
		0.161115,
		0.162759,
		0.164428,
		0.166123,
		0.167843,
		0.16959,
		0.171364,
		0.173165,
		0.174994,
		0.176852,
		0.178739,
		0.180656,
		0.182603,
		0.184582,
		0.186592,
		0.188634,
		0.19071,
		0.192819,
		0.194962,
		0.197141,
		0.199356,
		0.201607,
		0.203896,
		0.206223,
		0.208589,
		0.210995,
		0.213443,
		0.215931,
		0.218463,
		0.221038,
		0.223658,
		0.226323,
		0.229035,
		0.231794,
		0.234603,
		0.237461,
		0.24037,
		0.243331,
		0.246346,
		0.249415,
		0.25254,
		0.255722,
		0.258963,
		0.262263,
		0.265625,
		0.269049,
		0.272538,
		0.276092,
		0.279714,
		0.283404,
		0.287164,
		0.290993,
		0.294891,
		0.298855,
		0.302876,
		0.306942,
		0.311037,
		0.315133,
		0.319197,
		0.323181,
		0.327031,
		0.330677,
		0.334039,
		0.337025,
		0.339533,
		0.341453,
		0.34267,
		0.343066,
		0.342523,
		0.340931,
		0.338191,
		0.334215,
		0.328938,
		0.322316,
		0.314334,
		0.305002,
		0.294361,
		0.282483,
		0.269468,
		0.255441,
		0.240551,
		0.224966,
		0.208867,
		0.192443,
		0.175889,
		0.159395,
		0.143149,
		0.127325,
		0.112084,
		0.097571,
		0.08391,
		0.071206,
		0.05954,
		0.048973,
		0.039541,
		0.031258,
		0.024116,
		0.018083,
		0.01311,
		0.009125,
		0.006041,
		0.003753,
		0.002146,
		0.001096,
		0.000475,
		0.000159,
		3.3e-005,
		0.046277,
		0.046619,
		0.046964,
		0.047314,
		0.047667,
		0.048024,
		0.048385,
		0.04875,
		0.049119,
		0.049493,
		0.049871,
		0.050253,
		0.050639,
		0.05103,
		0.051425,
		0.051826,
		0.05223,
		0.05264,
		0.053054,
		0.053473,
		0.053897,
		0.054327,
		0.054761,
		0.0552,
		0.055645,
		0.056095,
		0.056551,
		0.057012,
		0.057479,
		0.057952,
		0.05843,
		0.058915,
		0.059405,
		0.059902,
		0.060404,
		0.060913,
		0.061429,
		0.061951,
		0.06248,
		0.063015,
		0.063557,
		0.064107,
		0.064663,
		0.065227,
		0.065798,
		0.066377,
		0.066963,
		0.067557,
		0.068159,
		0.068768,
		0.069387,
		0.070013,
		0.070648,
		0.071291,
		0.071944,
		0.072605,
		0.073275,
		0.073955,
		0.074644,
		0.075343,
		0.076051,
		0.07677,
		0.077498,
		0.078237,
		0.078987,
		0.079747,
		0.080519,
		0.081301,
		0.082095,
		0.0829,
		0.083718,
		0.084547,
		0.085389,
		0.086243,
		0.08711,
		0.08799,
		0.088884,
		0.089791,
		0.090711,
		0.091646,
		0.092596,
		0.09356,
		0.094539,
		0.095534,
		0.096544,
		0.09757,
		0.098612,
		0.099671,
		0.100747,
		0.101841,
		0.102952,
		0.104082,
		0.105229,
		0.106396,
		0.107582,
		0.108788,
		0.110014,
		0.111261,
		0.112529,
		0.113819,
		0.11513,
		0.116464,
		0.117821,
		0.119202,
		0.120607,
		0.122037,
		0.123492,
		0.124973,
		0.12648,
		0.128015,
		0.129577,
		0.131168,
		0.132788,
		0.134438,
		0.136118,
		0.137829,
		0.139573,
		0.141349,
		0.14316,
		0.145004,
		0.146884,
		0.148801,
		0.150754,
		0.152746,
		0.154777,
		0.156848,
		0.15896,
		0.161115,
		0.163312,
		0.165555,
		0.167843,
		0.170178,
		0.172561,
		0.174994,
		0.177477,
		0.180011,
		0.182595,
		0.185227,
		0.187904,
		0.190619,
		0.19336,
		0.196111,
		0.198852,
		0.201554,
		0.204182,
		0.206694,
		0.209038,
		0.211158,
		0.212988,
		0.21446,
		0.215498,
		0.216028,
		0.215973,
		0.215263,
		0.213831,
		0.211619,
		0.208583,
		0.204691,
		0.199929,
		0.1943,
		0.187825,
		0.180546,
		0.172521,
		0.163827,
		0.154555,
		0.144809,
		0.134701,
		0.124352,
		0.113883,
		0.103418,
		0.093074,
		0.082967,
		0.073201,
		0.063871,
		0.05506,
		0.046839,
		0.039266,
		0.032382,
		0.026217,
		0.020783,
		0.016081,
		0.012095,
		0.008796,
		0.006142,
		0.00408,
		0.002543,
		0.00146,
		0.000748,
		0.000326,
		0.000109,
		2.3e-005,
		0.021402,
		0.021564,
		0.021727,
		0.021892,
		0.022059,
		0.022228,
		0.022399,
		0.022572,
		0.022747,
		0.022924,
		0.023103,
		0.023285,
		0.023468,
		0.023653,
		0.023841,
		0.024031,
		0.024223,
		0.024418,
		0.024615,
		0.024814,
		0.025016,
		0.02522,
		0.025427,
		0.025636,
		0.025848,
		0.026062,
		0.026279,
		0.026499,
		0.026722,
		0.026947,
		0.027175,
		0.027406,
		0.027641,
		0.027878,
		0.028118,
		0.028361,
		0.028608,
		0.028857,
		0.02911,
		0.029367,
		0.029626,
		0.02989,
		0.030156,
		0.030427,
		0.0307,
		0.030978,
		0.031259,
		0.031545,
		0.031834,
		0.032127,
		0.032424,
		0.032726,
		0.033031,
		0.033341,
		0.033655,
		0.033974,
		0.034297,
		0.034625,
		0.034957,
		0.035295,
		0.035637,
		0.035984,
		0.036336,
		0.036694,
		0.037057,
		0.037425,
		0.037798,
		0.038178,
		0.038563,
		0.038954,
		0.03935,
		0.039753,
		0.040162,
		0.040578,
		0.041,
		0.041428,
		0.041863,
		0.042305,
		0.042755,
		0.043211,
		0.043674,
		0.044146,
		0.044624,
		0.045111,
		0.045605,
		0.046108,
		0.046619,
		0.047138,
		0.047667,
		0.048204,
		0.04875,
		0.049305,
		0.049871,
		0.050445,
		0.05103,
		0.051625,
		0.05223,
		0.052846,
		0.053473,
		0.054111,
		0.054761,
		0.055422,
		0.056095,
		0.056781,
		0.057479,
		0.05819,
		0.058915,
		0.059653,
		0.060404,
		0.06117,
		0.061951,
		0.062746,
		0.063557,
		0.064384,
		0.065227,
		0.066086,
		0.066963,
		0.067857,
		0.068768,
		0.069699,
		0.070648,
		0.071616,
		0.072605,
		0.073614,
		0.074644,
		0.075696,
		0.07677,
		0.077867,
		0.078987,
		0.080132,
		0.081301,
		0.082496,
		0.083718,
		0.084966,
		0.086243,
		0.087547,
		0.088879,
		0.090239,
		0.091624,
		0.093032,
		0.094456,
		0.09589,
		0.097323,
		0.098742,
		0.100129,
		0.101464,
		0.102721,
		0.103873,
		0.104887,
		0.105728,
		0.10636,
		0.106744,
		0.106843,
		0.10662,
		0.106041,
		0.105077,
		0.103703,
		0.101903,
		0.099666,
		0.096994,
		0.093895,
		0.090387,
		0.086499,
		0.082265,
		0.077731,
		0.072946,
		0.067966,
		0.06285,
		0.057659,
		0.052454,
		0.047295,
		0.042238,
		0.037339,
		0.032644,
		0.028199,
		0.02404,
		0.020197,
		0.016693,
		0.013546,
		0.010764,
		0.008349,
		0.006296,
		0.004591,
		0.003214,
		0.002141,
		0.001339,
		0.00077,
		0.000396,
		0.000173,
		5.8e-005,
		1.2e-005,
		0.005562,
		0.005605,
		0.005649,
		0.005692,
		0.005737,
		0.005782,
		0.005827,
		0.005873,
		0.00592,
		0.005967,
		0.006014,
		0.006063,
		0.006111,
		0.006161,
		0.006211,
		0.006261,
		0.006313,
		0.006364,
		0.006417,
		0.00647,
		0.006524,
		0.006578,
		0.006633,
		0.006689,
		0.006746,
		0.006803,
		0.006861,
		0.00692,
		0.006979,
		0.00704,
		0.007101,
		0.007163,
		0.007225,
		0.007289,
		0.007353,
		0.007418,
		0.007484,
		0.007551,
		0.007619,
		0.007688,
		0.007758,
		0.007828,
		0.0079,
		0.007973,
		0.008046,
		0.008121,
		0.008197,
		0.008273,
		0.008351,
		0.00843,
		0.00851,
		0.008592,
		0.008674,
		0.008758,
		0.008842,
		0.008928,
		0.009016,
		0.009104,
		0.009194,
		0.009285,
		0.009378,
		0.009472,
		0.009567,
		0.009664,
		0.009762,
		0.009862,
		0.009964,
		0.010067,
		0.010171,
		0.010277,
		0.010385,
		0.010495,
		0.010606,
		0.010719,
		0.010834,
		0.010951,
		0.011069,
		0.01119,
		0.011312,
		0.011437,
		0.011564,
		0.011692,
		0.011823,
		0.011957,
		0.012092,
		0.01223,
		0.01237,
		0.012512,
		0.012657,
		0.012805,
		0.012955,
		0.013107,
		0.013263,
		0.013421,
		0.013582,
		0.013746,
		0.013913,
		0.014083,
		0.014256,
		0.014433,
		0.014613,
		0.014796,
		0.014982,
		0.015172,
		0.015366,
		0.015563,
		0.015765,
		0.01597,
		0.016179,
		0.016392,
		0.01661,
		0.016832,
		0.017058,
		0.017289,
		0.017525,
		0.017765,
		0.018011,
		0.018262,
		0.018518,
		0.018779,
		0.019046,
		0.019319,
		0.019597,
		0.019882,
		0.020173,
		0.02047,
		0.020774,
		0.021084,
		0.021402,
		0.021727,
		0.022059,
		0.022399,
		0.022747,
		0.023103,
		0.023468,
		0.023841,
		0.024222,
		0.024612,
		0.02501,
		0.025414,
		0.025825,
		0.026239,
		0.026654,
		0.027066,
		0.027471,
		0.027862,
		0.028234,
		0.028577,
		0.028884,
		0.029144,
		0.029348,
		0.029484,
		0.029542,
		0.029512,
		0.029384,
		0.029149,
		0.028801,
		0.028334,
		0.027745,
		0.027034,
		0.026203,
		0.025256,
		0.024201,
		0.023047,
		0.021807,
		0.020493,
		0.019122,
		0.017708,
		0.01627,
		0.014824,
		0.013387,
		0.011975,
		0.010604,
		0.009287,
		0.008036,
		0.006863,
		0.005777,
		0.004784,
		0.00389,
		0.003097,
		0.002407,
		0.001819,
		0.001329,
		0.000933,
		0.000623,
		0.00039,
		0.000225,
		0.000116,
		5.1e-005,
		1.7e-005,
		1.7e-005,
		5.1e-005,
		0.000116,
		0.000225,
		0.00039,
		0.000623,
		0.000933,
		0.001329,
		0.001819,
		0.002407,
		0.003097,
		0.00389,
		0.004784,
		0.005777,
		0.006863,
		0.008036,
		0.009287,
		0.010604,
		0.011975,
		0.013387,
		0.014824,
		0.01627,
		0.017708,
		0.019122,
		0.020493,
		0.021807,
		0.023047,
		0.024201,
		0.025256,
		0.026203,
		0.027034,
		0.027745,
		0.028334,
		0.028801,
		0.029149,
		0.029384,
		0.029512,
		0.029542,
		0.029484,
		0.029348,
		0.029144,
		0.028884,
		0.028577,
		0.028234,
		0.027862,
		0.027471,
		0.027066,
		0.026654,
		0.026239,
		0.025825,
		0.025414,
		0.02501,
		0.024612,
		0.024222,
		0.023841,
		0.023468,
		0.023103,
		0.022747,
		0.022399,
		0.022059,
		0.021727,
		0.021402,
		0.021084,
		0.020774,
		0.02047,
		0.020173,
		0.019882,
		0.019597,
		0.019319,
		0.019046,
		0.018779,
		0.018518,
		0.018262,
		0.018011,
		0.017765,
		0.017525,
		0.017289,
		0.017058,
		0.016832,
		0.01661,
		0.016392,
		0.016179,
		0.01597,
		0.015765,
		0.015563,
		0.015366,
		0.015172,
		0.014982,
		0.014796,
		0.014613,
		0.014433,
		0.014256,
		0.014083,
		0.013913,
		0.013746,
		0.013582,
		0.013421,
		0.013263,
		0.013107,
		0.012955,
		0.012805,
		0.012657,
		0.012512,
		0.01237,
		0.01223,
		0.012092,
		0.011957,
		0.011823,
		0.011692,
		0.011564,
		0.011437,
		0.011312,
		0.01119,
		0.011069,
		0.010951,
		0.010834,
		0.010719,
		0.010606,
		0.010495,
		0.010385,
		0.010277,
		0.010171,
		0.010067,
		0.009964,
		0.009862,
		0.009762,
		0.009664,
		0.009567,
		0.009472,
		0.009378,
		0.009285,
		0.009194,
		0.009104,
		0.009016,
		0.008928,
		0.008842,
		0.008758,
		0.008674,
		0.008592,
		0.00851,
		0.00843,
		0.008351,
		0.008273,
		0.008197,
		0.008121,
		0.008046,
		0.007973,
		0.0079,
		0.007828,
		0.007758,
		0.007688,
		0.007619,
		0.007551,
		0.007484,
		0.007418,
		0.007353,
		0.007289,
		0.007225,
		0.007163,
		0.007101,
		0.00704,
		0.006979,
		0.00692,
		0.006861,
		0.006803,
		0.006746,
		0.006689,
		0.006633,
		0.006578,
		0.006524,
		0.00647,
		0.006417,
		0.006364,
		0.006313,
		0.006261,
		0.006211,
		0.006161,
		0.006111,
		0.006063,
		0.006014,
		0.005967,
		0.00592,
		0.005873,
		0.005827,
		0.005782,
		0.005737,
		0.005692,
		0.005649,
		0.005605,
		1.2e-005,
		5.8e-005,
		0.000173,
		0.000396,
		0.00077,
		0.001339,
		0.002141,
		0.003214,
		0.004591,
		0.006296,
		0.008349,
		0.010764,
		0.013546,
		0.016693,
		0.020197,
		0.02404,
		0.028199,
		0.032644,
		0.037339,
		0.042238,
		0.047295,
		0.052454,
		0.057659,
		0.06285,
		0.067966,
		0.072946,
		0.077731,
		0.082265,
		0.086499,
		0.090387,
		0.093895,
		0.096994,
		0.099666,
		0.101903,
		0.103703,
		0.105077,
		0.106041,
		0.10662,
		0.106843,
		0.106744,
		0.10636,
		0.105728,
		0.104887,
		0.103873,
		0.102721,
		0.101464,
		0.100129,
		0.098742,
		0.097323,
		0.09589,
		0.094456,
		0.093032,
		0.091624,
		0.090239,
		0.088879,
		0.087547,
		0.086243,
		0.084966,
		0.083718,
		0.082496,
		0.081301,
		0.080132,
		0.078987,
		0.077867,
		0.07677,
		0.075696,
		0.074644,
		0.073614,
		0.072605,
		0.071616,
		0.070648,
		0.069699,
		0.068768,
		0.067857,
		0.066963,
		0.066086,
		0.065227,
		0.064384,
		0.063557,
		0.062746,
		0.061951,
		0.06117,
		0.060404,
		0.059653,
		0.058915,
		0.05819,
		0.057479,
		0.056781,
		0.056095,
		0.055422,
		0.054761,
		0.054111,
		0.053473,
		0.052846,
		0.05223,
		0.051625,
		0.05103,
		0.050445,
		0.049871,
		0.049305,
		0.04875,
		0.048204,
		0.047667,
		0.047138,
		0.046619,
		0.046108,
		0.045605,
		0.045111,
		0.044624,
		0.044146,
		0.043674,
		0.043211,
		0.042755,
		0.042305,
		0.041863,
		0.041428,
		0.041,
		0.040578,
		0.040162,
		0.039753,
		0.03935,
		0.038954,
		0.038563,
		0.038178,
		0.037798,
		0.037425,
		0.037057,
		0.036694,
		0.036336,
		0.035984,
		0.035637,
		0.035295,
		0.034957,
		0.034625,
		0.034297,
		0.033974,
		0.033655,
		0.033341,
		0.033031,
		0.032726,
		0.032424,
		0.032127,
		0.031834,
		0.031545,
		0.031259,
		0.030978,
		0.0307,
		0.030427,
		0.030156,
		0.02989,
		0.029626,
		0.029367,
		0.02911,
		0.028857,
		0.028608,
		0.028361,
		0.028118,
		0.027878,
		0.027641,
		0.027406,
		0.027175,
		0.026947,
		0.026722,
		0.026499,
		0.026279,
		0.026062,
		0.025848,
		0.025636,
		0.025427,
		0.02522,
		0.025016,
		0.024814,
		0.024615,
		0.024418,
		0.024223,
		0.024031,
		0.023841,
		0.023653,
		0.023468,
		0.023285,
		0.023103,
		0.022924,
		0.022747,
		0.022572,
		0.022399,
		0.022228,
		0.022059,
		0.021892,
		0.021727,
		0.021564,
		2.3e-005,
		0.000109,
		0.000326,
		0.000748,
		0.00146,
		0.002543,
		0.00408,
		0.006142,
		0.008796,
		0.012095,
		0.016081,
		0.020783,
		0.026217,
		0.032382,
		0.039266,
		0.046839,
		0.05506,
		0.063871,
		0.073201,
		0.082967,
		0.093074,
		0.103418,
		0.113883,
		0.124352,
		0.134701,
		0.144809,
		0.154555,
		0.163827,
		0.172521,
		0.180546,
		0.187825,
		0.1943,
		0.199929,
		0.204691,
		0.208583,
		0.211619,
		0.213831,
		0.215263,
		0.215973,
		0.216028,
		0.215498,
		0.21446,
		0.212988,
		0.211158,
		0.209038,
		0.206694,
		0.204182,
		0.201554,
		0.198852,
		0.196111,
		0.19336,
		0.190619,
		0.187904,
		0.185227,
		0.182595,
		0.180011,
		0.177477,
		0.174994,
		0.172561,
		0.170178,
		0.167843,
		0.165555,
		0.163312,
		0.161115,
		0.15896,
		0.156848,
		0.154777,
		0.152746,
		0.150754,
		0.148801,
		0.146884,
		0.145004,
		0.14316,
		0.141349,
		0.139573,
		0.137829,
		0.136118,
		0.134438,
		0.132788,
		0.131168,
		0.129577,
		0.128015,
		0.12648,
		0.124973,
		0.123492,
		0.122037,
		0.120607,
		0.119202,
		0.117821,
		0.116464,
		0.11513,
		0.113819,
		0.112529,
		0.111261,
		0.110014,
		0.108788,
		0.107582,
		0.106396,
		0.105229,
		0.104082,
		0.102952,
		0.101841,
		0.100747,
		0.099671,
		0.098612,
		0.09757,
		0.096544,
		0.095534,
		0.094539,
		0.09356,
		0.092596,
		0.091646,
		0.090711,
		0.089791,
		0.088884,
		0.08799,
		0.08711,
		0.086243,
		0.085389,
		0.084547,
		0.083718,
		0.0829,
		0.082095,
		0.081301,
		0.080519,
		0.079747,
		0.078987,
		0.078237,
		0.077498,
		0.07677,
		0.076051,
		0.075343,
		0.074644,
		0.073955,
		0.073275,
		0.072605,
		0.071944,
		0.071291,
		0.070648,
		0.070013,
		0.069387,
		0.068768,
		0.068159,
		0.067557,
		0.066963,
		0.066377,
		0.065798,
		0.065227,
		0.064663,
		0.064107,
		0.063557,
		0.063015,
		0.06248,
		0.061951,
		0.061429,
		0.060913,
		0.060404,
		0.059902,
		0.059405,
		0.058915,
		0.05843,
		0.057952,
		0.057479,
		0.057012,
		0.056551,
		0.056095,
		0.055645,
		0.0552,
		0.054761,
		0.054327,
		0.053897,
		0.053473,
		0.053054,
		0.05264,
		0.05223,
		0.051826,
		0.051425,
		0.05103,
		0.050639,
		0.050253,
		0.049871,
		0.049493,
		0.049119,
		0.04875,
		0.048385,
		0.048024,
		0.047667,
		0.047314,
		0.046964,
		0.046619,
		3.3e-005,
		0.000159,
		0.000475,
		0.001096,
		0.002146,
		0.003753,
		0.006041,
		0.009125,
		0.01311,
		0.018083,
		0.024116,
		0.031258,
		0.039541,
		0.048973,
		0.05954,
		0.071206,
		0.08391,
		0.097571,
		0.112084,
		0.127325,
		0.143149,
		0.159395,
		0.175889,
		0.192443,
		0.208867,
		0.224966,
		0.240551,
		0.255441,
		0.269468,
		0.282483,
		0.294361,
		0.305002,
		0.314334,
		0.322316,
		0.328938,
		0.334215,
		0.338191,
		0.340931,
		0.342523,
		0.343066,
		0.34267,
		0.341453,
		0.339533,
		0.337025,
		0.334039,
		0.330677,
		0.327031,
		0.323181,
		0.319197,
		0.315133,
		0.311037,
		0.306942,
		0.302876,
		0.298855,
		0.294891,
		0.290993,
		0.287164,
		0.283404,
		0.279714,
		0.276092,
		0.272538,
		0.269049,
		0.265625,
		0.262263,
		0.258963,
		0.255722,
		0.25254,
		0.249415,
		0.246346,
		0.243331,
		0.24037,
		0.237461,
		0.234603,
		0.231794,
		0.229035,
		0.226323,
		0.223658,
		0.221038,
		0.218463,
		0.215931,
		0.213443,
		0.210995,
		0.208589,
		0.206223,
		0.203896,
		0.201607,
		0.199356,
		0.197141,
		0.194962,
		0.192819,
		0.19071,
		0.188634,
		0.186592,
		0.184582,
		0.182603,
		0.180656,
		0.178739,
		0.176852,
		0.174994,
		0.173165,
		0.171364,
		0.16959,
		0.167843,
		0.166123,
		0.164428,
		0.162759,
		0.161115,
		0.159495,
		0.157899,
		0.156326,
		0.154777,
		0.15325,
		0.151745,
		0.150262,
		0.148801,
		0.14736,
		0.14594,
		0.14454,
		0.14316,
		0.141799,
		0.140457,
		0.139134,
		0.137829,
		0.136543,
		0.135274,
		0.134022,
		0.132788,
		0.13157,
		0.130369,
		0.129184,
		0.128015,
		0.126861,
		0.125723,
		0.1246,
		0.123492,
		0.122398,
		0.121319,
		0.120254,
		0.119202,
		0.118164,
		0.11714,
		0.116129,
		0.11513,
		0.114144,
		0.113171,
		0.11221,
		0.111261,
		0.110324,
		0.109399,
		0.108485,
		0.107582,
		0.106691,
		0.10581,
		0.104941,
		0.104082,
		0.103233,
		0.102394,
		0.101566,
		0.100747,
		0.099939,
		0.09914,
		0.09835,
		0.09757,
		0.096799,
		0.096037,
		0.095283,
		0.094539,
		0.093803,
		0.093076,
		0.092357,
		0.091646,
		0.090944,
		0.090249,
		0.089563,
		0.088884,
		0.088212,
		0.087549,
		0.086892,
		0.086243,
		0.085601,
		0.084966,
		0.084339,
		0.083718,
		0.083104,
		0.082496,
		0.081895,
		0.081301,
		0.080713,
		0.080132,
		0.079556,
		4.1e-005,
		0.000199,
		0.000596,
		0.001382,
		0.002719,
		0.004775,
		0.007717,
		0.011703,
		0.016877,
		0.023365,
		0.031269,
		0.040669,
		0.051615,
		0.064131,
		0.078208,
		0.09381,
		0.110866,
		0.129275,
		0.148905,
		0.169596,
		0.191158,
		0.213377,
		0.236018,
		0.25883,
		0.281552,
		0.303917,
		0.325662,
		0.346533,
		0.366296,
		0.384738,
		0.401677,
		0.416966,
		0.430498,
		0.442204,
		0.45206,
		0.460076,
		0.466305,
		0.470829,
		0.473757,
		0.475222,
		0.475373,
		0.474364,
		0.472358,
		0.46951,
		0.465973,
		0.461886,
		0.457378,
		0.452559,
		0.447525,
		0.442357,
		0.437117,
		0.431857,
		0.426612,
		0.421411,
		0.416269,
		0.411199,
		0.406206,
		0.401293,
		0.39646,
		0.391707,
		0.387032,
		0.382435,
		0.377913,
		0.373465,
		0.369091,
		0.364787,
		0.360554,
		0.356389,
		0.352291,
		0.34826,
		0.344293,
		0.34039,
		0.336549,
		0.332769,
		0.32905,
		0.325389,
		0.321785,
		0.318238,
		0.314747,
		0.31131,
		0.307926,
		0.304595,
		0.301315,
		0.298086,
		0.294906,
		0.291774,
		0.28869,
		0.285652,
		0.282661,
		0.279714,
		0.276811,
		0.273952,
		0.271135,
		0.268359,
		0.265625,
		0.262931,
		0.260276,
		0.257659,
		0.255081,
		0.25254,
		0.250035,
		0.247567,
		0.245133,
		0.242735,
		0.24037,
		0.238038,
		0.23574,
		0.233473,
		0.231239,
		0.229035,
		0.226862,
		0.224718,
		0.222604,
		0.220519,
		0.218463,
		0.216434,
		0.214433,
		0.212459,
		0.210511,
		0.208589,
		0.206693,
		0.204822,
		0.202976,
		0.201154,
		0.199356,
		0.197581,
		0.19583,
		0.194101,
		0.192394,
		0.19071,
		0.189047,
		0.187405,
		0.185784,
		0.184184,
		0.182603,
		0.181043,
		0.179502,
		0.177981,
		0.176478,
		0.174994,
		0.173528,
		0.172081,
		0.170651,
		0.169238,
		0.167843,
		0.166465,
		0.165103,
		0.163757,
		0.162428,
		0.161115,
		0.159817,
		0.158534,
		0.157267,
		0.156015,
		0.154777,
		0.153554,
		0.152345,
		0.151149,
		0.149968,
		0.148801,
		0.147646,
		0.146505,
		0.145377,
		0.144262,
		0.14316,
		0.142069,
		0.140991,
		0.139926,
		0.138872,
		0.137829,
		0.136799,
		0.135779,
		0.134771,
		0.133774,
		0.132788,
		0.131812,
		0.130847,
		0.129893,
		0.128949,
		0.128015,
		0.127091,
		0.126177,
		0.125272,
		0.124377,
		0.123492,
		0.122616,
		0.121749,
		0.120891,
		0.120042,
		0.119202,
		4.6e-005,
		0.000224,
		0.000676,
		0.001573,
		0.00311,
		0.005488,
		0.00891,
		0.013573,
		0.019659,
		0.027329,
		0.036721,
		0.047945,
		0.061078,
		0.076164,
		0.09321,
		0.112184,
		0.133017,
		0.1556,
		0.179781,
		0.205375,
		0.232157,
		0.259871,
		0.288229,
		0.316924,
		0.34563,
		0.374015,
		0.401745,
		0.428498,
		0.453969,
		0.477883,
		0.5,
		0.520121,
		0.538097,
		0.553828,
		0.567267,
		0.578414,
		0.58732,
		0.594078,
		0.598814,
		0.601687,
		0.602874,
		0.602567,
		0.600964,
		0.59826,
		0.594644,
		0.590294,
		0.585368,
		0.580011,
		0.574344,
		0.56847,
		0.56247,
		0.556411,
		0.550341,
		0.544295,
		0.538297,
		0.532363,
		0.526502,
		0.520717,
		0.515011,
		0.509384,
		0.503836,
		0.498366,
		0.492972,
		0.487653,
		0.48241,
		0.47724,
		0.472143,
		0.467118,
		0.462163,
		0.457278,
		0.452462,
		0.447714,
		0.443032,
		0.438416,
		0.433865,
		0.429378,
		0.424954,
		0.420591,
		0.416289,
		0.412048,
		0.407865,
		0.403741,
		0.399673,
		0.395662,
		0.391707,
		0.387806,
		0.383959,
		0.380165,
		0.376422,
		0.372731,
		0.369091,
		0.365499,
		0.361957,
		0.358463,
		0.355015,
		0.351615,
		0.34826,
		0.34495,
		0.341684,
		0.338462,
		0.335282,
		0.332145,
		0.32905,
		0.325995,
		0.32298,
		0.320005,
		0.317068,
		0.31417,
		0.31131,
		0.308487,
		0.3057,
		0.302949,
		0.300233,
		0.297552,
		0.294906,
		0.292293,
		0.289713,
		0.287165,
		0.28465,
		0.282166,
		0.279714,
		0.277292,
		0.2749,
		0.272538,
		0.270205,
		0.267901,
		0.265625,
		0.263377,
		0.261156,
		0.258963,
		0.256796,
		0.254655,
		0.25254,
		0.25045,
		0.248386,
		0.246346,
		0.24433,
		0.242338,
		0.24037,
		0.238425,
		0.236502,
		0.234603,
		0.232725,
		0.230869,
		0.229035,
		0.227222,
		0.225429,
		0.223658,
		0.221906,
		0.220175,
		0.218463,
		0.21677,
		0.215097,
		0.213443,
		0.211807,
		0.210189,
		0.208589,
		0.207007,
		0.205443,
		0.203896,
		0.202366,
		0.200852,
		0.199356,
		0.197875,
		0.196411,
		0.194962,
		0.193529,
		0.192112,
		0.19071,
		0.189322,
		0.18795,
		0.186592,
		0.185248,
		0.183919,
		0.182603,
		0.181302,
		0.180014,
		0.178739,
		0.177478,
		0.17623,
		0.174994,
		0.173771,
		0.172561,
		0.171364,
		0.170178,
		0.169005,
		0.167843,
		0.166693,
		0.165555,
		0.164428,
		4.8e-005,
		0.000235,
		0.000712,
		0.001665,
		0.003307,
		0.005863,
		0.009563,
		0.014633,
		0.021286,
		0.029718,
		0.040099,
		0.05257,
		0.067237,
		0.084171,
		0.103399,
		0.124906,
		0.148632,
		0.174471,
		0.202269,
		0.231825,
		0.262895,
		0.295193,
		0.328396,
		0.36215,
		0.39608,
		0.429797,
		0.462908,
		0.495029,
		0.525792,
		0.554862,
		0.581942,
		0.606781,
		0.629185,
		0.649019,
		0.666205,
		0.680727,
		0.692627,
		0.701997,
		0.708972,
		0.713725,
		0.716456,
		0.717384,
		0.716734,
		0.714734,
		0.711604,
		0.707552,
		0.702768,
		0.697422,
		0.691658,
		0.685602,
		0.679351,
		0.672987,
		0.666569,
		0.66014,
		0.653732,
		0.647364,
		0.641049,
		0.634793,
		0.628601,
		0.622473,
		0.61641,
		0.610413,
		0.604481,
		0.598615,
		0.592813,
		0.587077,
		0.581406,
		0.575799,
		0.570256,
		0.564777,
		0.559361,
		0.554008,
		0.548718,
		0.54349,
		0.538323,
		0.533218,
		0.528173,
		0.523188,
		0.518262,
		0.513395,
		0.508587,
		0.503836,
		0.499142,
		0.494505,
		0.489923,
		0.485397,
		0.480925,
		0.476508,
		0.472143,
		0.467831,
		0.463572,
		0.459363,
		0.455206,
		0.451099,
		0.447041,
		0.443032,
		0.439072,
		0.435159,
		0.431293,
		0.427474,
		0.423701,
		0.419973,
		0.416289,
		0.41265,
		0.409054,
		0.405501,
		0.40199,
		0.398522,
		0.395094,
		0.391707,
		0.38836,
		0.385053,
		0.381784,
		0.378555,
		0.375363,
		0.372208,
		0.369091,
		0.366009,
		0.362964,
		0.359954,
		0.35698,
		0.354039,
		0.351133,
		0.34826,
		0.34542,
		0.342613,
		0.339837,
		0.337094,
		0.334382,
		0.331701,
		0.32905,
		0.326429,
		0.323837,
		0.321275,
		0.318742,
		0.316237,
		0.31376,
		0.31131,
		0.308888,
		0.306492,
		0.304124,
		0.301781,
		0.299464,
		0.297172,
		0.294906,
		0.292664,
		0.290446,
		0.288253,
		0.286083,
		0.283937,
		0.281814,
		0.279714,
		0.277636,
		0.27558,
		0.273547,
		0.271535,
		0.269544,
		0.267574,
		0.265625,
		0.263696,
		0.261788,
		0.2599,
		0.258031,
		0.256181,
		0.254351,
		0.25254,
		0.250747,
		0.248973,
		0.247217,
		0.245479,
		0.243759,
		0.242056,
		0.24037,
		0.238701,
		0.237049,
		0.235414,
		0.233795,
		0.232193,
		0.230606,
		0.229035,
		0.227479,
		0.225939,
		0.224414,
		0.222905,
		0.221409,
		0.219929,
		0.218463,
		0.217011,
		0.215573,
		0.214149,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003361,
		0.005977,
		0.00978,
		0.015015,
		0.021916,
		0.030705,
		0.041577,
		0.054701,
		0.070213,
		0.088209,
		0.108744,
		0.131824,
		0.15741,
		0.185408,
		0.215674,
		0.248009,
		0.282165,
		0.317842,
		0.354699,
		0.392355,
		0.430401,
		0.468408,
		0.505937,
		0.542556,
		0.577846,
		0.611418,
		0.642924,
		0.672067,
		0.698606,
		0.722367,
		0.743241,
		0.761188,
		0.77623,
		0.788449,
		0.797981,
		0.805002,
		0.809723,
		0.812377,
		0.81321,
		0.812471,
		0.810403,
		0.807238,
		0.803188,
		0.798444,
		0.793172,
		0.787513,
		0.781582,
		0.77547,
		0.769247,
		0.762966,
		0.756663,
		0.750362,
		0.74408,
		0.737825,
		0.731605,
		0.72542,
		0.719275,
		0.71317,
		0.707107,
		0.701086,
		0.695109,
		0.689177,
		0.683291,
		0.677451,
		0.671658,
		0.665912,
		0.660215,
		0.654566,
		0.648967,
		0.643416,
		0.637915,
		0.632464,
		0.627063,
		0.621711,
		0.61641,
		0.611159,
		0.605958,
		0.600807,
		0.595706,
		0.590655,
		0.585653,
		0.580701,
		0.575799,
		0.570945,
		0.56614,
		0.561384,
		0.556677,
		0.552017,
		0.547405,
		0.542841,
		0.538323,
		0.533853,
		0.529428,
		0.52505,
		0.520718,
		0.51643,
		0.512188,
		0.50799,
		0.503836,
		0.499726,
		0.495659,
		0.491635,
		0.487653,
		0.483714,
		0.479816,
		0.475959,
		0.472143,
		0.468368,
		0.464632,
		0.460936,
		0.457278,
		0.45366,
		0.45008,
		0.446537,
		0.443032,
		0.439564,
		0.436133,
		0.432738,
		0.429378,
		0.426054,
		0.422765,
		0.41951,
		0.416289,
		0.413102,
		0.409949,
		0.406828,
		0.403741,
		0.400685,
		0.397661,
		0.394668,
		0.391707,
		0.388776,
		0.385876,
		0.383005,
		0.380165,
		0.377353,
		0.37457,
		0.371816,
		0.369091,
		0.366393,
		0.363722,
		0.361079,
		0.358463,
		0.355873,
		0.353309,
		0.350772,
		0.34826,
		0.345773,
		0.343311,
		0.340874,
		0.338462,
		0.336073,
		0.333709,
		0.331368,
		0.32905,
		0.326755,
		0.324482,
		0.322233,
		0.320005,
		0.317799,
		0.315615,
		0.313452,
		0.31131,
		0.309189,
		0.307089,
		0.305009,
		0.302949,
		0.300909,
		0.298888,
		0.296887,
		0.294906,
		0.292943,
		0.290998,
		0.289073,
		0.287165,
		0.285276,
		0.283404,
		0.28155,
		0.279714,
		0.277895,
		0.276092,
		0.274307,
		0.272538,
		0.270786,
		0.269049,
		0.267329,
		4.9e-005,
		0.000237,
		0.000719,
		0.001688,
		0.003364,
		0.005985,
		0.0098,
		0.015059,
		0.022008,
		0.030876,
		0.041877,
		0.055198,
		0.070995,
		0.089387,
		0.110452,
		0.134224,
		0.160683,
		0.18976,
		0.22133,
		0.255208,
		0.291157,
		0.328883,
		0.368043,
		0.408251,
		0.449081,
		0.490085,
		0.530799,
		0.570757,
		0.609506,
		0.646618,
		0.681705,
		0.714427,
		0.744506,
		0.771729,
		0.795954,
		0.817112,
		0.835202,
		0.850289,
		0.862495,
		0.871991,
		0.878988,
		0.883723,
		0.886448,
		0.887425,
		0.88691,
		0.885148,
		0.882368,
		0.878774,
		0.874547,
		0.86984,
		0.864779,
		0.859466,
		0.853978,
		0.848375,
		0.842698,
		0.836975,
		0.831226,
		0.825462,
		0.819691,
		0.813918,
		0.808146,
		0.802379,
		0.796619,
		0.790869,
		0.785132,
		0.779408,
		0.773702,
		0.768013,
		0.762345,
		0.756699,
		0.751077,
		0.745479,
		0.739908,
		0.734365,
		0.728851,
		0.723368,
		0.717915,
		0.712494,
		0.707107,
		0.701753,
		0.696434,
		0.69115,
		0.685901,
		0.68069,
		0.675515,
		0.670377,
		0.665277,
		0.660215,
		0.655192,
		0.650207,
		0.645261,
		0.640354,
		0.635486,
		0.630658,
		0.625869,
		0.62112,
		0.61641,
		0.61174,
		0.607109,
		0.602518,
		0.597967,
		0.593455,
		0.588982,
		0.584549,
		0.580154,
		0.575799,
		0.571482,
		0.567204,
		0.562964,
		0.558763,
		0.5546,
		0.550474,
		0.546387,
		0.542336,
		0.538323,
		0.534347,
		0.530408,
		0.526505,
		0.522638,
		0.518807,
		0.515011,
		0.511251,
		0.507526,
		0.503836,
		0.50018,
		0.496559,
		0.492972,
		0.489418,
		0.485897,
		0.48241,
		0.478955,
		0.475533,
		0.472143,
		0.468785,
		0.465459,
		0.462163,
		0.458899,
		0.455665,
		0.452462,
		0.449289,
		0.446146,
		0.443032,
		0.439948,
		0.436892,
		0.433865,
		0.430867,
		0.427896,
		0.424954,
		0.422038,
		0.41915,
		0.416289,
		0.413455,
		0.410647,
		0.407865,
		0.405109,
		0.402378,
		0.399673,
		0.396993,
		0.394338,
		0.391707,
		0.3891,
		0.386518,
		0.383959,
		0.381424,
		0.378911,
		0.376422,
		0.373956,
		0.371512,
		0.369091,
		0.366691,
		0.364313,
		0.361957,
		0.359622,
		0.357308,
		0.355015,
		0.352743,
		0.350491,
		0.34826,
		0.346048,
		0.343856,
		0.341684,
		0.339531,
		0.337397,
		0.335282,
		0.333186,
		0.331109,
		0.32905,
		0.327009,
		0.324985,
		0.32298,
		4.8e-005,
		0.000233,
		0.000712,
		0.001676,
		0.003347,
		0.005965,
		0.00978,
		0.015042,
		0.021994,
		0.030868,
		0.041877,
		0.055211,
		0.071031,
		0.089464,
		0.110603,
		0.134495,
		0.161142,
		0.190493,
		0.222444,
		0.256834,
		0.293444,
		0.332002,
		0.372177,
		0.413595,
		0.455837,
		0.498456,
		0.540983,
		0.582942,
		0.623865,
		0.663304,
		0.700845,
		0.736123,
		0.768831,
		0.798725,
		0.825636,
		0.849464,
		0.870185,
		0.887842,
		0.902538,
		0.914431,
		0.923719,
		0.930634,
		0.935426,
		0.938357,
		0.939685,
		0.939661,
		0.938519,
		0.936471,
		0.933706,
		0.930382,
		0.926633,
		0.922566,
		0.918264,
		0.91379,
		0.909189,
		0.904491,
		0.89972,
		0.894886,
		0.890001,
		0.88507,
		0.880097,
		0.875088,
		0.870045,
		0.864974,
		0.859876,
		0.854755,
		0.849616,
		0.84446,
		0.839291,
		0.834112,
		0.828924,
		0.823732,
		0.818536,
		0.81334,
		0.808146,
		0.802956,
		0.797771,
		0.792593,
		0.787425,
		0.782268,
		0.777124,
		0.771993,
		0.766878,
		0.761779,
		0.756699,
		0.751638,
		0.746597,
		0.741577,
		0.736579,
		0.731605,
		0.726654,
		0.721729,
		0.716828,
		0.711954,
		0.707107,
		0.702287,
		0.697495,
		0.692731,
		0.687996,
		0.683291,
		0.678615,
		0.673969,
		0.669354,
		0.664769,
		0.660215,
		0.655692,
		0.651201,
		0.64674,
		0.642312,
		0.637915,
		0.63355,
		0.629217,
		0.624916,
		0.620647,
		0.61641,
		0.612205,
		0.608032,
		0.603892,
		0.599783,
		0.595706,
		0.591661,
		0.587648,
		0.583667,
		0.579717,
		0.575799,
		0.571912,
		0.568056,
		0.564232,
		0.560439,
		0.556677,
		0.552945,
		0.549244,
		0.545574,
		0.541933,
		0.538323,
		0.534743,
		0.531193,
		0.527672,
		0.52418,
		0.520718,
		0.517284,
		0.51388,
		0.510503,
		0.507156,
		0.503836,
		0.500544,
		0.497281,
		0.494044,
		0.490835,
		0.487653,
		0.484498,
		0.48137,
		0.478268,
		0.475193,
		0.472143,
		0.469119,
		0.466121,
		0.463149,
		0.460201,
		0.457278,
		0.454381,
		0.451507,
		0.448658,
		0.445833,
		0.443032,
		0.440255,
		0.437501,
		0.43477,
		0.432063,
		0.429378,
		0.426716,
		0.424076,
		0.421459,
		0.418863,
		0.416289,
		0.413737,
		0.411206,
		0.408697,
		0.406208,
		0.403741,
		0.401293,
		0.398867,
		0.39646,
		0.394074,
		0.391707,
		0.38936,
		0.387032,
		0.384724,
		0.382435,
		0.380165,
		4.4e-005,
		0.000219,
		0.000676,
		0.001605,
		0.003231,
		0.005799,
		0.009563,
		0.014781,
		0.021703,
		0.030563,
		0.041577,
		0.054932,
		0.070788,
		0.089266,
		0.110452,
		0.134389,
		0.161075,
		0.19046,
		0.222444,
		0.256874,
		0.293546,
		0.332201,
		0.372533,
		0.414184,
		0.456759,
		0.499827,
		0.542935,
		0.585621,
		0.627422,
		0.667894,
		0.706621,
		0.743232,
		0.777408,
		0.808892,
		0.837497,
		0.863107,
		0.885676,
		0.905228,
		0.921848,
		0.935677,
		0.946897,
		0.955728,
		0.96241,
		0.967195,
		0.970337,
		0.972083,
		0.972665,
		0.972293,
		0.971156,
		0.969415,
		0.967203,
		0.96463,
		0.961779,
		0.958714,
		0.955481,
		0.952111,
		0.948626,
		0.94504,
		0.941362,
		0.937598,
		0.933754,
		0.929834,
		0.925841,
		0.921781,
		0.917657,
		0.913472,
		0.909232,
		0.90494,
		0.900598,
		0.896212,
		0.891784,
		0.887317,
		0.882814,
		0.87828,
		0.873716,
		0.869125,
		0.864511,
		0.859876,
		0.855222,
		0.850552,
		0.845868,
		0.841172,
		0.836467,
		0.831754,
		0.827036,
		0.822315,
		0.817592,
		0.812868,
		0.808146,
		0.803427,
		0.798713,
		0.794004,
		0.789303,
		0.784611,
		0.779928,
		0.775256,
		0.770596,
		0.76595,
		0.761317,
		0.756699,
		0.752097,
		0.747512,
		0.742944,
		0.738394,
		0.733863,
		0.729352,
		0.72486,
		0.72039,
		0.71594,
		0.711512,
		0.707107,
		0.702724,
		0.698364,
		0.694027,
		0.689715,
		0.685426,
		0.681162,
		0.676922,
		0.672708,
		0.668518,
		0.664354,
		0.660215,
		0.656102,
		0.652015,
		0.647954,
		0.643919,
		0.63991,
		0.635927,
		0.631971,
		0.628041,
		0.624138,
		0.620261,
		0.61641,
		0.612586,
		0.608789,
		0.605018,
		0.601273,
		0.597555,
		0.593863,
		0.590198,
		0.586559,
		0.582946,
		0.579359,
		0.575799,
		0.572264,
		0.568755,
		0.565272,
		0.561815,
		0.558383,
		0.554977,
		0.551596,
		0.54824,
		0.54491,
		0.541604,
		0.538323,
		0.535067,
		0.531836,
		0.528629,
		0.525446,
		0.522288,
		0.519153,
		0.516043,
		0.512956,
		0.509893,
		0.506853,
		0.503836,
		0.500843,
		0.497872,
		0.494924,
		0.491999,
		0.489096,
		0.486216,
		0.483358,
		0.480522,
		0.477707,
		0.474914,
		0.472143,
		0.469393,
		0.466665,
		0.463957,
		0.46127,
		0.458604,
		0.455958,
		0.453333,
		0.450728,
		0.448143,
		0.445578,
		0.443032,
		0.440506,
		0.438,
		3.8e-005,
		0.00019,
		0.000596,
		0.00144,
		0.002939,
		0.005342,
		0.00891,
		0.013913,
		0.020615,
		0.029267,
		0.040099,
		0.053314,
		0.069081,
		0.087528,
		0.108744,
		0.132767,
		0.159588,
		0.189147,
		0.22133,
		0.255969,
		0.292845,
		0.331687,
		0.372177,
		0.413956,
		0.456627,
		0.499767,
		0.542935,
		0.585686,
		0.627576,
		0.668181,
		0.707107,
		0.743999,
		0.778552,
		0.81052,
		0.839721,
		0.866039,
		0.889428,
		0.909906,
		0.927549,
		0.942488,
		0.954896,
		0.96498,
		0.97297,
		0.979108,
		0.983637,
		0.986795,
		0.988808,
		0.98988,
		0.990193,
		0.989904,
		0.989142,
		0.988014,
		0.986599,
		0.98496,
		0.983139,
		0.981167,
		0.979061,
		0.976834,
		0.974494,
		0.972044,
		0.969489,
		0.966832,
		0.964076,
		0.961225,
		0.95828,
		0.955247,
		0.952128,
		0.948926,
		0.945646,
		0.94229,
		0.938862,
		0.935365,
		0.931803,
		0.928179,
		0.924495,
		0.920755,
		0.916963,
		0.913121,
		0.909232,
		0.905299,
		0.901325,
		0.897313,
		0.893264,
		0.889182,
		0.88507,
		0.880929,
		0.876762,
		0.872571,
		0.868358,
		0.864126,
		0.859876,
		0.85561,
		0.851331,
		0.84704,
		0.842738,
		0.838428,
		0.834112,
		0.829789,
		0.825463,
		0.821134,
		0.816804,
		0.812474,
		0.808146,
		0.80382,
		0.799498,
		0.795181,
		0.790869,
		0.786565,
		0.782268,
		0.77798,
		0.773702,
		0.769433,
		0.765177,
		0.760931,
		0.756699,
		0.75248,
		0.748275,
		0.744084,
		0.739908,
		0.735749,
		0.731605,
		0.727478,
		0.723368,
		0.719275,
		0.715201,
		0.711144,
		0.707107,
		0.703088,
		0.699089,
		0.695109,
		0.69115,
		0.68721,
		0.683291,
		0.679392,
		0.675515,
		0.671658,
		0.667822,
		0.664008,
		0.660215,
		0.656444,
		0.652694,
		0.648967,
		0.645261,
		0.641577,
		0.637915,
		0.634275,
		0.630658,
		0.627063,
		0.62349,
		0.619939,
		0.61641,
		0.612904,
		0.60942,
		0.605958,
		0.602518,
		0.599101,
		0.595706,
		0.592333,
		0.588982,
		0.585653,
		0.582346,
		0.579062,
		0.575799,
		0.572558,
		0.569338,
		0.56614,
		0.562964,
		0.55981,
		0.556677,
		0.553565,
		0.550474,
		0.547405,
		0.544357,
		0.54133,
		0.538323,
		0.535338,
		0.532373,
		0.529428,
		0.526505,
		0.523601,
		0.520718,
		0.517854,
		0.515011,
		0.512188,
		0.509384,
		0.5066,
		0.503836,
		0.501091,
		0.498366,
		0.495659,
		2.8e-005,
		0.000147,
		0.000475,
		0.001175,
		0.002452,
		0.004545,
		0.007717,
		0.012245,
		0.018409,
		0.026482,
		0.036721,
		0.049356,
		0.064584,
		0.082561,
		0.103399,
		0.127154,
		0.153832,
		0.183376,
		0.215674,
		0.25055,
		0.287772,
		0.32705,
		0.368043,
		0.410368,
		0.4536,
		0.497293,
		0.540983,
		0.584203,
		0.626498,
		0.667436,
		0.706621,
		0.743703,
		0.778389,
		0.810449,
		0.839721,
		0.866109,
		0.889587,
		0.910193,
		0.928019,
		0.943208,
		0.955943,
		0.966433,
		0.974912,
		0.98162,
		0.986797,
		0.990679,
		0.993485,
		0.995414,
		0.996642,
		0.997321,
		0.997575,
		0.997502,
		0.997181,
		0.996665,
		0.995994,
		0.995193,
		0.994276,
		0.993251,
		0.992122,
		0.99089,
		0.989554,
		0.988116,
		0.986574,
		0.984931,
		0.983187,
		0.981342,
		0.9794,
		0.97736,
		0.975226,
		0.972999,
		0.970681,
		0.968276,
		0.965784,
		0.963209,
		0.960553,
		0.957819,
		0.95501,
		0.952128,
		0.949176,
		0.946156,
		0.943071,
		0.939924,
		0.936718,
		0.933455,
		0.930138,
		0.926769,
		0.92335,
		0.919885,
		0.916375,
		0.912824,
		0.909232,
		0.905603,
		0.901939,
		0.898242,
		0.894513,
		0.890756,
		0.886971,
		0.883162,
		0.879329,
		0.875474,
		0.8716,
		0.867708,
		0.863799,
		0.859876,
		0.855939,
		0.85199,
		0.848031,
		0.844063,
		0.840087,
		0.836105,
		0.832117,
		0.828126,
		0.824131,
		0.820135,
		0.816138,
		0.812141,
		0.808146,
		0.804153,
		0.800163,
		0.796177,
		0.792195,
		0.78822,
		0.78425,
		0.780288,
		0.776333,
		0.772387,
		0.76845,
		0.764523,
		0.760605,
		0.756699,
		0.752804,
		0.748921,
		0.74505,
		0.741192,
		0.737347,
		0.733515,
		0.729698,
		0.725895,
		0.722107,
		0.718333,
		0.714575,
		0.710833,
		0.707107,
		0.703397,
		0.699703,
		0.696026,
		0.692366,
		0.688723,
		0.685097,
		0.681489,
		0.677898,
		0.674326,
		0.670771,
		0.667234,
		0.663715,
		0.660215,
		0.656733,
		0.65327,
		0.649825,
		0.646399,
		0.642991,
		0.639602,
		0.636233,
		0.632882,
		0.629549,
		0.626236,
		0.622942,
		0.619667,
		0.61641,
		0.613173,
		0.609954,
		0.606755,
		0.603574,
		0.600413,
		0.59727,
		0.594146,
		0.591042,
		0.587956,
		0.584888,
		0.58184,
		0.57881,
		0.575799,
		0.572806,
		0.569832,
		0.566876,
		0.563939,
		0.561021,
		0.55812,
		0.555238,
		0.552374,
		1.8e-005,
		9.6e-005,
		0.000326,
		0.000838,
		0.001811,
		0.003462,
		0.006041,
		0.009823,
		0.015095,
		0.02215,
		0.031269,
		0.042718,
		0.056733,
		0.073513,
		0.09321,
		0.115924,
		0.141698,
		0.170509,
		0.202269,
		0.236823,
		0.273949,
		0.313356,
		0.354699,
		0.397574,
		0.441538,
		0.486112,
		0.530799,
		0.575095,
		0.618508,
		0.660568,
		0.700845,
		0.738959,
		0.774593,
		0.807497,
		0.837497,
		0.864494,
		0.888462,
		0.909445,
		0.927549,
		0.942932,
		0.955795,
		0.966371,
		0.974912,
		0.981678,
		0.986929,
		0.990914,
		0.993865,
		0.99599,
		0.997473,
		0.998467,
		0.9991,
		0.999471,
		0.999655,
		0.999705,
		0.999659,
		0.999536,
		0.999349,
		0.999102,
		0.998793,
		0.998419,
		0.997978,
		0.997465,
		0.996876,
		0.996208,
		0.995458,
		0.994625,
		0.993706,
		0.9927,
		0.991607,
		0.990425,
		0.989154,
		0.987794,
		0.986346,
		0.98481,
		0.983187,
		0.981477,
		0.979683,
		0.977805,
		0.975845,
		0.973805,
		0.971686,
		0.969489,
		0.967218,
		0.964874,
		0.962458,
		0.959974,
		0.957423,
		0.954807,
		0.952128,
		0.949389,
		0.946591,
		0.943738,
		0.94083,
		0.93787,
		0.93486,
		0.931803,
		0.9287,
		0.925553,
		0.922365,
		0.919137,
		0.91587,
		0.912568,
		0.909232,
		0.905864,
		0.902465,
		0.899037,
		0.895582,
		0.892101,
		0.888597,
		0.88507,
		0.881522,
		0.877955,
		0.874369,
		0.870768,
		0.867151,
		0.863519,
		0.859876,
		0.85622,
		0.852555,
		0.84888,
		0.845197,
		0.841508,
		0.837812,
		0.834112,
		0.830407,
		0.826699,
		0.82299,
		0.819279,
		0.815567,
		0.811856,
		0.808146,
		0.804438,
		0.800733,
		0.797031,
		0.793332,
		0.789639,
		0.785951,
		0.782268,
		0.778592,
		0.774923,
		0.771261,
		0.767608,
		0.763962,
		0.760326,
		0.756699,
		0.753082,
		0.749475,
		0.745878,
		0.742293,
		0.738718,
		0.735156,
		0.731605,
		0.728066,
		0.72454,
		0.721027,
		0.717527,
		0.71404,
		0.710566,
		0.707107,
		0.703661,
		0.70023,
		0.696813,
		0.69341,
		0.690022,
		0.686649,
		0.683291,
		0.679948,
		0.67662,
		0.673308,
		0.670011,
		0.66673,
		0.663465,
		0.660215,
		0.656981,
		0.653763,
		0.650561,
		0.647376,
		0.644206,
		0.641052,
		0.637915,
		0.634794,
		0.631689,
		0.628601,
		0.625529,
		0.622473,
		0.619433,
		0.61641,
		0.613403,
		0.610413,
		0.607439,
		4.7e-005,
		0.000173,
		0.000481,
		0.001107,
		0.002233,
		0.00408,
		0.006903,
		0.010985,
		0.016622,
		0.024116,
		0.033761,
		0.045834,
		0.06058,
		0.078208,
		0.098875,
		0.122681,
		0.149661,
		0.179781,
		0.212934,
		0.248937,
		0.287533,
		0.328396,
		0.371134,
		0.415302,
		0.460409,
		0.505937,
		0.551352,
		0.596121,
		0.639732,
		0.681705,
		0.721612,
		0.759087,
		0.793834,
		0.825636,
		0.854357,
		0.879941,
		0.902408,
		0.921848,
		0.938409,
		0.95229,
		0.963724,
		0.97297,
		0.980301,
		0.985989,
		0.990301,
		0.993485,
		0.99577,
		0.997356,
		0.998418,
		0.9991,
		0.999516,
		0.999756,
		0.999887,
		0.999952,
		0.999982,
		0.999993,
		0.999993,
		0.999982,
		0.999959,
		0.99992,
		0.999859,
		0.999771,
		0.999651,
		0.999493,
		0.999293,
		0.999046,
		0.998747,
		0.998392,
		0.997978,
		0.997501,
		0.996959,
		0.996348,
		0.995666,
		0.994912,
		0.994084,
		0.993181,
		0.992201,
		0.991144,
		0.990011,
		0.9888,
		0.987511,
		0.986146,
		0.984704,
		0.983187,
		0.981594,
		0.979927,
		0.978187,
		0.976376,
		0.974494,
		0.972543,
		0.970524,
		0.968439,
		0.966289,
		0.964076,
		0.961803,
		0.959469,
		0.957078,
		0.95463,
		0.952128,
		0.949573,
		0.946968,
		0.944313,
		0.94161,
		0.938862,
		0.93607,
		0.933236,
		0.930361,
		0.927446,
		0.924495,
		0.921508,
		0.918486,
		0.915432,
		0.912347,
		0.909232,
		0.906089,
		0.90292,
		0.899725,
		0.896506,
		0.893264,
		0.890001,
		0.886718,
		0.883417,
		0.880097,
		0.876762,
		0.873411,
		0.870045,
		0.866667,
		0.863277,
		0.859876,
		0.856465,
		0.853044,
		0.849616,
		0.84618,
		0.842738,
		0.839291,
		0.835839,
		0.832383,
		0.828924,
		0.825463,
		0.822,
		0.818536,
		0.815072,
		0.811609,
		0.808146,
		0.804685,
		0.801227,
		0.797771,
		0.794318,
		0.790869,
		0.787425,
		0.783986,
		0.780552,
		0.777124,
		0.773702,
		0.770286,
		0.766878,
		0.763477,
		0.760084,
		0.756699,
		0.753323,
		0.749955,
		0.746597,
		0.743248,
		0.739908,
		0.736579,
		0.73326,
		0.729952,
		0.726654,
		0.723368,
		0.720092,
		0.716828,
		0.713576,
		0.710335,
		0.707107,
		0.70389,
		0.700686,
		0.697495,
		0.694316,
		0.69115,
		0.687996,
		0.684856,
		0.681729,
		0.678615,
		0.675515,
		0.672427,
		0.669354,
		0.666294,
		0.663248,
		0.660215,
		1e-005,
		5.1e-005,
		0.000175,
		0.00047,
		0.001066,
		0.002141,
		0.003911,
		0.006631,
		0.010587,
		0.016081,
		0.023426,
		0.032928,
		0.044879,
		0.05954,
		0.077134,
		0.097831,
		0.121741,
		0.148905,
		0.17929,
		0.212781,
		0.249185,
		0.288229,
		0.329566,
		0.37278,
		0.417399,
		0.462908,
		0.508764,
		0.554413,
		0.599306,
		0.642924,
		0.684786,
		0.724468,
		0.761617,
		0.795954,
		0.827286,
		0.855502,
		0.88057,
		0.902538,
		0.921517,
		0.937675,
		0.951225,
		0.96241,
		0.971493,
		0.978745,
		0.984431,
		0.988808,
		0.992111,
		0.994554,
		0.996322,
		0.997575,
		0.998441,
		0.999026,
		0.999412,
		0.999659,
		0.999812,
		0.999904,
		0.999955,
		0.999982,
		0.999994,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999995,
		0.999986,
		0.999968,
		0.999937,
		0.999889,
		0.999819,
		0.999723,
		0.999596,
		0.999435,
		0.999236,
		0.998994,
		0.998706,
		0.998368,
		0.997978,
		0.997533,
		0.99703,
		0.996468,
		0.995843,
		0.995156,
		0.994403,
		0.993585,
		0.9927,
		0.991748,
		0.990728,
		0.989641,
		0.988485,
		0.987261,
		0.98597,
		0.984612,
		0.983187,
		0.981696,
		0.98014,
		0.978519,
		0.976835,
		0.975089,
		0.973282,
		0.971415,
		0.969489,
		0.967506,
		0.965467,
		0.963372,
		0.961225,
		0.959025,
		0.956775,
		0.954475,
		0.952128,
		0.949734,
		0.947296,
		0.944814,
		0.94229,
		0.939726,
		0.937122,
		0.934481,
		0.931803,
		0.92909,
		0.926344,
		0.923565,
		0.920755,
		0.917916,
		0.915048,
		0.912153,
		0.909232,
		0.906287,
		0.903317,
		0.900326,
		0.897313,
		0.894279,
		0.891227,
		0.888157,
		0.88507,
		0.881966,
		0.878848,
		0.875716,
		0.872571,
		0.869413,
		0.866244,
		0.863065,
		0.859876,
		0.856678,
		0.853472,
		0.850259,
		0.84704,
		0.843815,
		0.840584,
		0.83735,
		0.834112,
		0.83087,
		0.827626,
		0.824381,
		0.821134,
		0.817887,
		0.814639,
		0.811392,
		0.808146,
		0.804901,
		0.801659,
		0.798418,
		0.795181,
		0.791947,
		0.788716,
		0.78549,
		0.782268,
		0.779051,
		0.77584,
		0.772634,
		0.769433,
		0.76624,
		0.763052,
		0.759872,
		0.756699,
		0.753533,
		0.750376,
		0.747226,
		0.744084,
		0.740951,
		0.737827,
		0.734711,
		0.731605,
		0.728508,
		0.72542,
		0.722343,
		0.719275,
		0.716218,
		0.71317,
		0.710133,
		6.1e-005,
		0.000228,
		0.000623,
		0.001406,
		0.002784,
		0.005004,
		0.008349,
		0.013129,
		0.019662,
		0.028272,
		0.039266,
		0.052927,
		0.0695,
		0.089175,
		0.112084,
		0.138285,
		0.167757,
		0.200397,
		0.236018,
		0.274347,
		0.315035,
		0.35766,
		0.401745,
		0.446768,
		0.492178,
		0.537419,
		0.581942,
		0.625228,
		0.666804,
		0.706256,
		0.743241,
		0.777497,
		0.808844,
		0.837182,
		0.862495,
		0.884835,
		0.904319,
		0.921114,
		0.935426,
		0.947489,
		0.95755,
		0.96586,
		0.972665,
		0.978195,
		0.982663,
		0.986257,
		0.989142,
		0.991455,
		0.99331,
		0.994799,
		0.995994,
		0.996952,
		0.997716,
		0.99832,
		0.998793,
		0.999156,
		0.999429,
		0.999629,
		0.999771,
		0.999867,
		0.999929,
		0.999966,
		0.999986,
		0.999996,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.999989,
		0.999975,
		0.99995,
		0.999911,
		0.999854,
		0.999777,
		0.999675,
		0.999544,
		0.999381,
		0.999183,
		0.998946,
		0.998669,
		0.998347,
		0.997978,
		0.997561,
		0.997092,
		0.996571,
		0.995996,
		0.995364,
		0.994676,
		0.99393,
		0.993125,
		0.992261,
		0.991337,
		0.990352,
		0.989308,
		0.988203,
		0.987038,
		0.985814,
		0.98453,
		0.983187,
		0.981785,
		0.980326,
		0.97881,
		0.977237,
		0.975609,
		0.973927,
		0.972191,
		0.970403,
		0.968563,
		0.966673,
		0.964734,
		0.962746,
		0.960712,
		0.958631,
		0.956507,
		0.954338,
		0.952128,
		0.949876,
		0.947585,
		0.945255,
		0.942888,
		0.940484,
		0.938045,
		0.935573,
		0.933068,
		0.930531,
		0.927963,
		0.925367,
		0.922742,
		0.92009,
		0.917412,
		0.914709,
		0.911982,
		0.909232,
		0.90646,
		0.903668,
		0.900855,
		0.898023,
		0.895174,
		0.892307,
		0.889423,
		0.886525,
		0.883611,
		0.880684,
		0.877744,
		0.874792,
		0.871829,
		0.868855,
		0.865871,
		0.862877,
		0.859876,
		0.856866,
		0.85385,
		0.850827,
		0.847798,
		0.844764,
		0.841725,
		0.838682,
		0.835636,
		0.832587,
		0.829535,
		0.826481,
		0.823426,
		0.82037,
		0.817314,
		0.814257,
		0.811201,
		0.808146,
		0.805092,
		0.80204,
		0.79899,
		0.795942,
		0.792898,
		0.789856,
		0.786818,
		0.783784,
		0.780754,
		0.777728,
		0.774707,
		0.771692,
		0.768681,
		0.765677,
		0.762678,
		0.759685,
		0.756699,
		6.5e-005,
		0.000356,
		0.001059,
		0.002407,
		0.004677,
		0.008178,
		0.013239,
		0.020197,
		0.029382,
		0.041104,
		0.055636,
		0.073201,
		0.09396,
		0.118003,
		0.145338,
		0.175889,
		0.209488,
		0.245885,
		0.284745,
		0.325662,
		0.368169,
		0.411755,
		0.455881,
		0.5,
		0.543575,
		0.586099,
		0.627109,
		0.666205,
		0.703054,
		0.737402,
		0.769076,
		0.797981,
		0.824096,
		0.847468,
		0.868202,
		0.886448,
		0.902391,
		0.916238,
		0.928207,
		0.938519,
		0.947385,
		0.955004,
		0.961557,
		0.967203,
		0.972079,
		0.9763,
		0.979961,
		0.983139,
		0.985898,
		0.988288,
		0.990351,
		0.992122,
		0.993633,
		0.994912,
		0.995985,
		0.996876,
		0.997607,
		0.9982,
		0.998674,
		0.999046,
		0.999333,
		0.999548,
		0.999707,
		0.999819,
		0.999895,
		0.999944,
		0.999973,
		0.999989,
		0.999997,
		0.999999,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999991,
		0.99998,
		0.999959,
		0.999928,
		0.999882,
		0.999819,
		0.999735,
		0.999627,
		0.999493,
		0.99933,
		0.999134,
		0.998903,
		0.998635,
		0.998327,
		0.997978,
		0.997585,
		0.997147,
		0.996662,
		0.996128,
		0.995546,
		0.994912,
		0.994227,
		0.99349,
		0.9927,
		0.991857,
		0.990961,
		0.990011,
		0.989007,
		0.987949,
		0.986838,
		0.985674,
		0.984457,
		0.983187,
		0.981865,
		0.980491,
		0.979066,
		0.977591,
		0.976067,
		0.974494,
		0.972873,
		0.971204,
		0.969489,
		0.967729,
		0.965925,
		0.964076,
		0.962186,
		0.960253,
		0.95828,
		0.956268,
		0.954217,
		0.952128,
		0.950003,
		0.947842,
		0.945646,
		0.943417,
		0.941155,
		0.938862,
		0.936538,
		0.934185,
		0.931803,
		0.929393,
		0.926957,
		0.924495,
		0.922008,
		0.919497,
		0.916963,
		0.914407,
		0.91183,
		0.909232,
		0.906615,
		0.903979,
		0.901325,
		0.898654,
		0.895967,
		0.893264,
		0.890546,
		0.887815,
		0.88507,
		0.882312,
		0.879542,
		0.876762,
		0.87397,
		0.871169,
		0.868358,
		0.865538,
		0.862711,
		0.859876,
		0.857034,
		0.854185,
		0.851331,
		0.848471,
		0.845607,
		0.842738,
		0.839866,
		0.83699,
		0.834112,
		0.83123,
		0.828347,
		0.825463,
		0.822577,
		0.819691,
		0.816804,
		0.813918,
		0.811031,
		0.808146,
		0.805262,
		0.802379,
		0.799498,
		0.000214,
		0.001038,
		0.00278,
		0.005777,
		0.010384,
		0.016957,
		0.025838,
		0.037339,
		0.051724,
		0.069201,
		0.089903,
		0.113883,
		0.141108,
		0.171451,
		0.204699,
		0.240551,
		0.278635,
		0.318512,
		0.359697,
		0.401677,
		0.443925,
		0.485925,
		0.527188,
		0.567267,
		0.605771,
		0.642378,
		0.676837,
		0.708972,
		0.73868,
		0.765928,
		0.790745,
		0.81321,
		0.833445,
		0.8516,
		0.867848,
		0.882368,
		0.895339,
		0.906936,
		0.917318,
		0.926633,
		0.935005,
		0.942546,
		0.949345,
		0.955481,
		0.961014,
		0.965998,
		0.970479,
		0.974494,
		0.978079,
		0.981267,
		0.984089,
		0.986574,
		0.988753,
		0.99065,
		0.992293,
		0.993706,
		0.994912,
		0.995933,
		0.99679,
		0.997501,
		0.998086,
		0.99856,
		0.998938,
		0.999236,
		0.999465,
		0.999638,
		0.999765,
		0.999854,
		0.999916,
		0.999955,
		0.999978,
		0.999991,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		0.999998,
		0.999993,
		0.999983,
		0.999967,
		0.999941,
		0.999903,
		0.999851,
		0.999782,
		0.999693,
		0.999581,
		0.999445,
		0.999281,
		0.999088,
		0.998863,
		0.998605,
		0.99831,
		0.997978,
		0.997607,
		0.997195,
		0.996742,
		0.996245,
		0.995704,
		0.995118,
		0.994485,
		0.993807,
		0.993081,
		0.992308,
		0.991486,
		0.990617,
		0.9897,
		0.988734,
		0.98772,
		0.986658,
		0.985548,
		0.984391,
		0.983187,
		0.981935,
		0.980638,
		0.979295,
		0.977906,
		0.976473,
		0.974996,
		0.973475,
		0.971912,
		0.970307,
		0.968661,
		0.966975,
		0.965249,
		0.963484,
		0.961681,
		0.959841,
		0.957965,
		0.956054,
		0.954108,
		0.952128,
		0.950115,
		0.948071,
		0.945995,
		0.943889,
		0.941754,
		0.93959,
		0.937398,
		0.935179,
		0.932935,
		0.930665,
		0.928371,
		0.926053,
		0.923712,
		0.92135,
		0.918966,
		0.916561,
		0.914137,
		0.911694,
		0.909232,
		0.906753,
		0.904257,
		0.901745,
		0.899218,
		0.896676,
		0.894119,
		0.891549,
		0.888967,
		0.886372,
		0.883765,
		0.881147,
		0.878519,
		0.875881,
		0.873234,
		0.870578,
		0.867913,
		0.865241,
		0.862562,
		0.859876,
		0.857183,
		0.854485,
		0.851782,
		0.849074,
		0.846361,
		0.843645,
		0.840925,
		0.838201,
		0.000471,
		0.002137,
		0.005387,
		0.010604,
		0.018148,
		0.02834,
		0.041445,
		0.057659,
		0.077097,
		0.099785,
		0.125657,
		0.154555,
		0.186232,
		0.220363,
		0.256554,
		0.294361,
		0.333305,
		0.372893,
		0.412633,
		0.45206,
		0.49074,
		0.528295,
		0.564404,
		0.598814,
		0.63134,
		0.661863,
		0.690328,
		0.716734,
		0.741127,
		0.76359,
		0.784235,
		0.803188,
		0.820585,
		0.836564,
		0.851254,
		0.864779,
		0.877246,
		0.888751,
		0.899376,
		0.909189,
		0.918247,
		0.926602,
		0.934294,
		0.941362,
		0.947842,
		0.953766,
		0.959167,
		0.964076,
		0.968525,
		0.972543,
		0.976158,
		0.9794,
		0.982294,
		0.984868,
		0.987147,
		0.989154,
		0.990912,
		0.992444,
		0.993771,
		0.994912,
		0.995886,
		0.996711,
		0.997403,
		0.997978,
		0.998451,
		0.998834,
		0.99914,
		0.999381,
		0.999567,
		0.999707,
		0.999809,
		0.999882,
		0.999931,
		0.999963,
		0.999982,
		0.999993,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999994,
		0.999986,
		0.999973,
		0.999951,
		0.99992,
		0.999876,
		0.999819,
		0.999744,
		0.999651,
		0.999536,
		0.999399,
		0.999236,
		0.999046,
		0.998827,
		0.998577,
		0.998294,
		0.997978,
		0.997627,
		0.997238,
		0.996812,
		0.996348,
		0.995843,
		0.995298,
		0.994712,
		0.994084,
		0.993413,
		0.9927,
		0.991944,
		0.991144,
		0.990301,
		0.989415,
		0.988485,
		0.987511,
		0.986495,
		0.985435,
		0.984332,
		0.983187,
		0.981999,
		0.98077,
		0.979499,
		0.978187,
		0.976835,
		0.975444,
		0.974012,
		0.972543,
		0.971035,
		0.969489,
		0.967907,
		0.966289,
		0.964635,
		0.962947,
		0.961225,
		0.959469,
		0.957681,
		0.955861,
		0.954009,
		0.952128,
		0.950217,
		0.948277,
		0.946308,
		0.944313,
		0.94229,
		0.940242,
		0.938168,
		0.93607,
		0.933948,
		0.931803,
		0.929636,
		0.927446,
		0.925236,
		0.923006,
		0.920755,
		0.918486,
		0.916199,
		0.913893,
		0.911571,
		0.909232,
		0.906878,
		0.904508,
		0.902123,
		0.899725,
		0.897313,
		0.894888,
		0.89245,
		0.890001,
		0.887541,
		0.88507,
		0.882588,
		0.880097,
		0.877597,
		0.875088,
		0.872571,
		0.000818,
		0.00355,
		0.008587,
		0.01627,
		0.026875,
		0.040599,
		0.057546,
		0.077731,
		0.101068,
		0.127383,
		0.156413,
		0.187825,
		0.221226,
		0.256183,
		0.29224,
		0.328938,
		0.365832,
		0.402506,
		0.438588,
		0.473757,
		0.507751,
		0.540369,
		0.571468,
		0.600964,
		0.628819,
		0.655041,
		0.679669,
		0.702768,
		0.724419,
		0.744711,
		0.763735,
		0.781582,
		0.798332,
		0.81406,
		0.82883,
		0.842698,
		0.85571,
		0.867908,
		0.879328,
		0.890001,
		0.899959,
		0.909232,
		0.91785,
		0.925841,
		0.933236,
		0.940063,
		0.946351,
		0.952128,
		0.957423,
		0.962262,
		0.966673,
		0.970681,
		0.974313,
		0.977591,
		0.980542,
		0.983187,
		0.985548,
		0.987648,
		0.989507,
		0.991144,
		0.992579,
		0.993829,
		0.994912,
		0.995843,
		0.996638,
		0.997312,
		0.997877,
		0.998347,
		0.998733,
		0.999046,
		0.999296,
		0.999493,
		0.999645,
		0.99976,
		0.999844,
		0.999903,
		0.999944,
		0.99997,
		0.999986,
		0.999994,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999998,
		0.999995,
		0.999989,
		0.999977,
		0.999959,
		0.999933,
		0.999897,
		0.999848,
		0.999785,
		0.999707,
		0.99961,
		0.999493,
		0.999355,
		0.999193,
		0.999006,
		0.998793,
		0.998551,
		0.99828,
		0.997978,
		0.997644,
		0.997277,
		0.996876,
		0.996439,
		0.995967,
		0.995458,
		0.994912,
		0.994328,
		0.993706,
		0.993045,
		0.992346,
		0.991607,
		0.990828,
		0.990011,
		0.989154,
		0.988257,
		0.987321,
		0.986346,
		0.985332,
		0.984278,
		0.983187,
		0.982057,
		0.980889,
		0.979683,
		0.97844,
		0.977161,
		0.975845,
		0.974494,
		0.973107,
		0.971686,
		0.97023,
		0.968741,
		0.967218,
		0.965663,
		0.964076,
		0.962458,
		0.96081,
		0.959131,
		0.957423,
		0.955686,
		0.953921,
		0.952128,
		0.950308,
		0.948463,
		0.946591,
		0.944695,
		0.942774,
		0.94083,
		0.938862,
		0.936872,
		0.93486,
		0.932827,
		0.930774,
		0.9287,
		0.926607,
		0.924495,
		0.922365,
		0.920217,
		0.918052,
		0.91587,
		0.913673,
		0.91146,
		0.909232,
		0.90699,
		0.904734,
		0.902465,
		0.001204,
		0.005056,
		0.011854,
		0.021807,
		0.035026,
		0.051521,
		0.071204,
		0.093895,
		0.119331,
		0.147183,
		0.177071,
		0.208583,
		0.241294,
		0.274783,
		0.308649,
		0.342523,
		0.37608,
		0.409045,
		0.441194,
		0.472358,
		0.502414,
		0.531288,
		0.558942,
		0.585368,
		0.610585,
		0.634625,
		0.657532,
		0.679351,
		0.700132,
		0.719918,
		0.73875,
		0.756663,
		0.773688,
		0.789852,
		0.805178,
		0.819691,
		0.833411,
		0.846361,
		0.858565,
		0.870045,
		0.880827,
		0.890936,
		0.900395,
		0.909232,
		0.917472,
		0.92514,
		0.932261,
		0.938862,
		0.944967,
		0.950601,
		0.955789,
		0.960553,
		0.964919,
		0.968908,
		0.972543,
		0.975845,
		0.978837,
		0.981538,
		0.983968,
		0.986146,
		0.988091,
		0.989821,
		0.991352,
		0.9927,
		0.993882,
		0.994912,
		0.995804,
		0.996571,
		0.997227,
		0.997782,
		0.998248,
		0.998635,
		0.998953,
		0.999212,
		0.999419,
		0.999581,
		0.999707,
		0.999801,
		0.999871,
		0.99992,
		0.999953,
		0.999975,
		0.999988,
		0.999995,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999996,
		0.99999,
		0.999981,
		0.999966,
		0.999944,
		0.999913,
		0.999872,
		0.999819,
		0.999752,
		0.999669,
		0.99957,
		0.999452,
		0.999313,
		0.999153,
		0.99897,
		0.998762,
		0.998528,
		0.998267,
		0.997978,
		0.99766,
		0.997312,
		0.996932,
		0.996521,
		0.996078,
		0.995601,
		0.99509,
		0.994545,
		0.993965,
		0.99335,
		0.9927,
		0.992015,
		0.991293,
		0.990536,
		0.989742,
		0.988913,
		0.988048,
		0.987147,
		0.98621,
		0.985237,
		0.98423,
		0.983187,
		0.982109,
		0.980996,
		0.97985,
		0.978669,
		0.977455,
		0.976208,
		0.974928,
		0.973615,
		0.972271,
		0.970896,
		0.969489,
		0.968053,
		0.966586,
		0.96509,
		0.963565,
		0.962012,
		0.960431,
		0.958822,
		0.957188,
		0.955526,
		0.95384,
		0.952128,
		0.950392,
		0.948631,
		0.946848,
		0.945041,
		0.943213,
		0.941362,
		0.939491,
		0.937598,
		0.935686,
		0.933754,
		0.931803,
		0.929834,
		0.927846,
		0.001561,
		0.006383,
		0.014591,
		0.026203,
		0.041135,
		0.05921,
		0.080172,
		0.103703,
		0.12944,
		0.156993,
		0.185967,
		0.215973,
		0.246645,
		0.277651,
		0.308696,
		0.339533,
		0.369959,
		0.399815,
		0.428982,
		0.457378,
		0.484947,
		0.51166,
		0.537502,
		0.56247,
		0.586569,
		0.609807,
		0.632191,
		0.653732,
		0.674436,
		0.694312,
		0.713365,
		0.731605,
		0.749038,
		0.765677,
		0.781532,
		0.796619,
		0.810953,
		0.824552,
		0.837433,
		0.849616,
		0.861121,
		0.87197,
		0.882183,
		0.891784,
		0.900792,
		0.909232,
		0.917126,
		0.924495,
		0.931363,
		0.937751,
		0.943681,
		0.949176,
		0.954256,
		0.958943,
		0.963257,
		0.967218,
		0.970847,
		0.974162,
		0.977183,
		0.979927,
		0.982413,
		0.984657,
		0.986675,
		0.988485,
		0.990101,
		0.991538,
		0.992809,
		0.99393,
		0.994912,
		0.995768,
		0.996509,
		0.997147,
		0.997692,
		0.998153,
		0.998541,
		0.998863,
		0.999128,
		0.999343,
		0.999515,
		0.999651,
		0.999755,
		0.999834,
		0.999892,
		0.999933,
		0.999961,
		0.999979,
		0.99999,
		0.999996,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999992,
		0.999984,
		0.999971,
		0.999952,
		0.999926,
		0.999891,
		0.999846,
		0.999788,
		0.999718,
		0.999632,
		0.999531,
		0.999412,
		0.999274,
		0.999115,
		0.998935,
		0.998733,
		0.998506,
		0.998255,
		0.997978,
		0.997675,
		0.997343,
		0.996984,
		0.996595,
		0.996177,
		0.995728,
		0.995249,
		0.994738,
		0.994196,
		0.993622,
		0.993016,
		0.992377,
		0.991705,
		0.991001,
		0.990264,
		0.989494,
		0.988691,
		0.987855,
		0.986986,
		0.986085,
		0.985151,
		0.984185,
		0.983187,
		0.982156,
		0.981094,
		0.980001,
		0.978877,
		0.977722,
		0.976536,
		0.975321,
		0.974075,
		0.972801,
		0.971498,
		0.970166,
		0.968806,
		0.967419,
		0.966004,
		0.964563,
		0.963095,
		0.961602,
		0.960083,
		0.95854,
		0.956972,
		0.955381,
		0.953766,
		0.952128,
		0.950468,
		0.948785,
		0.001825,
		0.007301,
		0.01635,
		0.028801,
		0.044403,
		0.062848,
		0.083785,
		0.106843,
		0.131646,
		0.157829,
		0.185047,
		0.212988,
		0.241376,
		0.269972,
		0.298578,
		0.327031,
		0.355202,
		0.382989,
		0.410314,
		0.437117,
		0.463351,
		0.488978,
		0.513969,
		0.538297,
		0.561943,
		0.584885,
		0.607109,
		0.628601,
		0.649351,
		0.669354,
		0.688606,
		0.707107,
		0.72486,
		0.741872,
		0.758149,
		0.773702,
		0.788542,
		0.802682,
		0.816138,
		0.828924,
		0.841057,
		0.852555,
		0.863435,
		0.873716,
		0.883417,
		0.892557,
		0.901155,
		0.909232,
		0.916807,
		0.9239,
		0.930531,
		0.936718,
		0.942482,
		0.947842,
		0.952816,
		0.957423,
		0.961681,
		0.965609,
		0.969224,
		0.972543,
		0.975582,
		0.978359,
		0.980889,
		0.983187,
		0.985268,
		0.987147,
		0.988837,
		0.990352,
		0.991705,
		0.992908,
		0.993974,
		0.994912,
		0.995735,
		0.996452,
		0.997073,
		0.997607,
		0.998064,
		0.998451,
		0.998776,
		0.999046,
		0.999268,
		0.999448,
		0.999593,
		0.999707,
		0.999794,
		0.999861,
		0.999909,
		0.999944,
		0.999967,
		0.999982,
		0.999992,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999986,
		0.999975,
		0.999959,
		0.999937,
		0.999907,
		0.999868,
		0.999819,
		0.999758,
		0.999684,
		0.999596,
		0.999493,
		0.999373,
		0.999236,
		0.999079,
		0.998903,
		0.998706,
		0.998486,
		0.998244,
		0.997978,
		0.997688,
		0.997372,
		0.99703,
		0.996662,
		0.996266,
		0.995843,
		0.995392,
		0.994912,
		0.994403,
		0.993865,
		0.993298,
		0.9927,
		0.992073,
		0.991416,
		0.990728,
		0.990011,
		0.989263,
		0.988485,
		0.987677,
		0.986838,
		0.98597,
		0.985072,
		0.984144,
		0.983187,
		0.9822,
		0.981184,
		0.98014,
		0.979066,
		0.977965,
		0.976835,
		0.975678,
		0.974494,
		0.973282,
		0.972044,
		0.97078,
		0.969489,
		0.968173,
		0.966832,
		0.965467,
		0.001962,
		0.007716,
		0.017013,
		0.029542,
		0.04496,
		0.062899,
		0.082994,
		0.104887,
		0.128242,
		0.152754,
		0.178148,
		0.204182,
		0.230652,
		0.257379,
		0.284216,
		0.311037,
		0.337737,
		0.364225,
		0.390425,
		0.416269,
		0.441699,
		0.466662,
		0.491113,
		0.515011,
		0.538323,
		0.561021,
		0.583079,
		0.604481,
		0.625212,
		0.645261,
		0.664622,
		0.683291,
		0.701268,
		0.718555,
		0.735156,
		0.751077,
		0.766326,
		0.780913,
		0.794849,
		0.808146,
		0.820817,
		0.832877,
		0.84434,
		0.855222,
		0.865538,
		0.875306,
		0.884543,
		0.893264,
		0.901488,
		0.909232,
		0.916514,
		0.92335,
		0.929759,
		0.935757,
		0.941362,
		0.946591,
		0.951461,
		0.955987,
		0.960187,
		0.964076,
		0.967671,
		0.970985,
		0.974035,
		0.976835,
		0.9794,
		0.981742,
		0.983876,
		0.985814,
		0.987569,
		0.989154,
		0.990579,
		0.991857,
		0.992999,
		0.994013,
		0.994912,
		0.995704,
		0.996398,
		0.997003,
		0.997527,
		0.997978,
		0.998364,
		0.99869,
		0.998965,
		0.999193,
		0.999381,
		0.999533,
		0.999656,
		0.999752,
		0.999826,
		0.999882,
		0.999923,
		0.999952,
		0.999972,
		0.999985,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999994,
		0.999988,
		0.999979,
		0.999965,
		0.999946,
		0.99992,
		0.999886,
		0.999844,
		0.999791,
		0.999727,
		0.999651,
		0.999561,
		0.999457,
		0.999337,
		0.9992,
		0.999046,
		0.998873,
		0.998681,
		0.998468,
		0.998234,
		0.997978,
		0.9977,
		0.997398,
		0.997073,
		0.996723,
		0.996348,
		0.995947,
		0.995521,
		0.995069,
		0.99459,
		0.994084,
		0.993551,
		0.992991,
		0.992403,
		0.991788,
		0.991144,
		0.990474,
		0.989775,
		0.989048,
		0.988294,
		0.987511,
		0.986701,
		0.985864,
		0.984999,
		0.984106,
		0.983187,
		0.98224,
		0.981267,
		0.980266,
		0.97924,
		0.978187,
		0.001978,
		0.007692,
		0.016784,
		0.028884,
		0.043618,
		0.060626,
		0.079568,
		0.100129,
		0.122028,
		0.14501,
		0.168853,
		0.19336,
		0.218359,
		0.243699,
		0.269249,
		0.294891,
		0.320521,
		0.346046,
		0.371384,
		0.39646,
		0.42121,
		0.445578,
		0.469512,
		0.492972,
		0.515919,
		0.538323,
		0.560159,
		0.581406,
		0.602046,
		0.622067,
		0.641458,
		0.660215,
		0.678333,
		0.69581,
		0.712649,
		0.728851,
		0.744423,
		0.759371,
		0.773702,
		0.787425,
		0.800552,
		0.813093,
		0.82506,
		0.836467,
		0.847326,
		0.857652,
		0.867459,
		0.876762,
		0.885575,
		0.893914,
		0.901795,
		0.909232,
		0.916242,
		0.92284,
		0.929041,
		0.93486,
		0.940314,
		0.945417,
		0.950184,
		0.95463,
		0.958769,
		0.962616,
		0.966185,
		0.969489,
		0.972543,
		0.975358,
		0.977948,
		0.980326,
		0.982503,
		0.984492,
		0.986304,
		0.987949,
		0.98944,
		0.990785,
		0.991996,
		0.993081,
		0.99405,
		0.994912,
		0.995675,
		0.996348,
		0.996937,
		0.997451,
		0.997897,
		0.99828,
		0.998608,
		0.998886,
		0.999119,
		0.999313,
		0.999473,
		0.999603,
		0.999707,
		0.999788,
		0.999852,
		0.999899,
		0.999934,
		0.999959,
		0.999976,
		0.999987,
		0.999994,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999995,
		0.99999,
		0.999982,
		0.99997,
		0.999953,
		0.999931,
		0.999901,
		0.999864,
		0.999819,
		0.999763,
		0.999696,
		0.999618,
		0.999527,
		0.999422,
		0.999302,
		0.999166,
		0.999014,
		0.998845,
		0.998657,
		0.998451,
		0.998225,
		0.997978,
		0.997711,
		0.997422,
		0.997111,
		0.996778,
		0.996422,
		0.996042,
		0.995638,
		0.995211,
		0.994759,
		0.994282,
		0.99378,
		0.993253,
		0.9927,
		0.992122,
		0.991519,
		0.99089,
		0.990235,
		0.989554,
		0.988848,
		0.988116,
		0.987358,
		0.001912,
		0.007387,
		0.016032,
		0.027471,
		0.041348,
		0.057334,
		0.075127,
		0.094456,
		0.115075,
		0.136765,
		0.159331,
		0.182595,
		0.2064,
		0.230604,
		0.255081,
		0.279714,
		0.304401,
		0.32905,
		0.353578,
		0.377913,
		0.40199,
		0.425753,
		0.449152,
		0.472143,
		0.494689,
		0.516759,
		0.538323,
		0.559361,
		0.579853,
		0.599783,
		0.61914,
		0.637915,
		0.656102,
		0.673697,
		0.690698,
		0.707107,
		0.722924,
		0.738155,
		0.752804,
		0.766878,
		0.780384,
		0.793332,
		0.805731,
		0.817592,
		0.828924,
		0.839741,
		0.850054,
		0.859876,
		0.869219,
		0.878098,
		0.886525,
		0.894513,
		0.902078,
		0.909232,
		0.91599,
		0.922365,
		0.928371,
		0.934022,
		0.939331,
		0.944313,
		0.94898,
		0.953345,
		0.957423,
		0.961225,
		0.964764,
		0.968053,
		0.971103,
		0.973927,
		0.976536,
		0.978942,
		0.981155,
		0.983187,
		0.985047,
		0.986746,
		0.988294,
		0.9897,
		0.990973,
		0.992122,
		0.993157,
		0.994084,
		0.994912,
		0.995649,
		0.996301,
		0.996876,
		0.99738,
		0.997819,
		0.9982,
		0.998528,
		0.998808,
		0.999046,
		0.999246,
		0.999412,
		0.999548,
		0.99966,
		0.999749,
		0.999819,
		0.999873,
		0.999914,
		0.999944,
		0.999965,
		0.99998,
		0.999989,
		0.999995,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999996,
		0.999991,
		0.999984,
		0.999974,
		0.999959,
		0.99994,
		0.999914,
		0.999882,
		0.999842,
		0.999793,
		0.999735,
		0.999666,
		0.999586,
		0.999493,
		0.999388,
		0.999268,
		0.999134,
		0.998984,
		0.998818,
		0.998635,
		0.998435,
		0.998216,
		0.997978,
		0.997721,
		0.997444,
		0.997147,
		0.996829,
		0.99649,
		0.996128,
		0.995745,
		0.99534,
		0.994912,
		0.994461,
		0.993987,
		0.99349,
		0.001807,
		0.006962,
		0.015085,
		0.025825,
		0.038856,
		0.053884,
		0.07064,
		0.088879,
		0.108382,
		0.128948,
		0.150396,
		0.172561,
		0.195295,
		0.218463,
		0.241943,
		0.265625,
		0.289411,
		0.313213,
		0.336951,
		0.360554,
		0.383959,
		0.407111,
		0.42996,
		0.452462,
		0.474581,
		0.496282,
		0.517538,
		0.538323,
		0.558619,
		0.578407,
		0.597675,
		0.61641,
		0.634605,
		0.652255,
		0.669354,
		0.685901,
		0.701897,
		0.717343,
		0.732241,
		0.746597,
		0.760415,
		0.773702,
		0.786465,
		0.798713,
		0.810454,
		0.821699,
		0.832457,
		0.842738,
		0.852555,
		0.861918,
		0.870838,
		0.879329,
		0.887401,
		0.895068,
		0.902341,
		0.909232,
		0.915755,
		0.921922,
		0.927745,
		0.933236,
		0.938407,
		0.943272,
		0.947842,
		0.952128,
		0.956143,
		0.959898,
		0.963404,
		0.966673,
		0.969716,
		0.972543,
		0.975164,
		0.977591,
		0.979834,
		0.981901,
		0.983803,
		0.985548,
		0.987147,
		0.988607,
		0.989936,
		0.991144,
		0.992239,
		0.993226,
		0.994115,
		0.994912,
		0.995624,
		0.996257,
		0.996818,
		0.997312,
		0.997745,
		0.998123,
		0.998451,
		0.998733,
		0.998974,
		0.999178,
		0.99935,
		0.999493,
		0.999611,
		0.999707,
		0.999783,
		0.999844,
		0.99989,
		0.999926,
		0.999951,
		0.99997,
		0.999982,
		0.999991,
		0.999996,
		0.999998,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999998,
		0.999996,
		0.999992,
		0.999986,
		0.999977,
		0.999965,
		0.999947,
		0.999925,
		0.999897,
		0.999861,
		0.999819,
		0.999767,
		0.999707,
		0.999636,
		0.999554,
		0.999461,
		0.999355,
		0.999236,
		0.999103,
		0.998956,
		0.998793,
		0.998614,
		0.99842,
		0.998208,
		0.997978,
		0.997731,
		0.997465,
		0.99718,
		0.001693,
		0.006522,
		0.014139,
		0.024222,
		0.036478,
		0.050639,
		0.06646,
		0.083718,
		0.102209,
		0.121749,
		0.142168,
		0.163312,
		0.185043,
		0.207232,
		0.229766,
		0.25254,
		0.27546,
		0.298442,
		0.321409,
		0.344293,
		0.367032,
		0.389573,
		0.411865,
		0.433865,
		0.455537,
		0.476846,
		0.497762,
		0.518262,
		0.538323,
		0.557927,
		0.577059,
		0.595706,
		0.613858,
		0.631507,
		0.648648,
		0.665277,
		0.681392,
		0.696992,
		0.712079,
		0.726654,
		0.740722,
		0.754286,
		0.767353,
		0.779928,
		0.792019,
		0.803632,
		0.814777,
		0.825463,
		0.835698,
		0.845492,
		0.854856,
		0.863799,
		0.872333,
		0.880467,
		0.888213,
		0.895582,
		0.902584,
		0.909232,
		0.915536,
		0.921508,
		0.927158,
		0.932498,
		0.937538,
		0.94229,
		0.946765,
		0.950973,
		0.954925,
		0.958631,
		0.962103,
		0.965348,
		0.968379,
		0.971204,
		0.973833,
		0.976276,
		0.978541,
		0.980638,
		0.982575,
		0.984361,
		0.986004,
		0.987511,
		0.988892,
		0.990153,
		0.991302,
		0.992346,
		0.993291,
		0.994144,
		0.994912,
		0.995601,
		0.996216,
		0.996763,
		0.997247,
		0.997675,
		0.998049,
		0.998376,
		0.998659,
		0.998903,
		0.999112,
		0.999289,
		0.999437,
		0.999561,
		0.999663,
		0.999746,
		0.999812,
		0.999864,
		0.999905,
		0.999935,
		0.999958,
		0.999974,
		0.999985,
		0.999992,
		0.999996,
		0.999998,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999993,
		0.999988,
		0.99998,
		0.999969,
		0.999954,
		0.999934,
		0.999909,
		0.999878,
		0.99984,
		0.999795,
		0.999741,
		0.999679,
		0.999606,
		0.999523,
		0.999429,
		0.999323,
		0.999205,
		0.999074,
		0.001584,
		0.006111,
		0.013263,
		0.022747,
		0.034297,
		0.047667,
		0.062632,
		0.078987,
		0.096544,
		0.11513,
		0.134589,
		0.154777,
		0.175563,
		0.196828,
		0.218463,
		0.24037,
		0.262459,
		0.28465,
		0.306869,
		0.32905,
		0.351133,
		0.373065,
		0.394798,
		0.416289,
		0.437501,
		0.4584,
		0.478955,
		0.499142,
		0.518938,
		0.538323,
		0.557281,
		0.575799,
		0.593863,
		0.611467,
		0.628601,
		0.645261,
		0.661443,
		0.677145,
		0.692366,
		0.707107,
		0.721369,
		0.735156,
		0.74847,
		0.761317,
		0.773702,
		0.78563,
		0.797109,
		0.808146,
		0.818748,
		0.828924,
		0.838682,
		0.848031,
		0.85698,
		0.865538,
		0.873716,
		0.881522,
		0.888967,
		0.89606,
		0.902812,
		0.909232,
		0.915332,
		0.92112,
		0.926607,
		0.931803,
		0.936718,
		0.941362,
		0.945745,
		0.949876,
		0.953766,
		0.957423,
		0.960857,
		0.964076,
		0.967092,
		0.969911,
		0.972543,
		0.974996,
		0.977279,
		0.9794,
		0.981366,
		0.983187,
		0.984868,
		0.986419,
		0.987845,
		0.989154,
		0.990352,
		0.991447,
		0.992444,
		0.99335,
		0.994171,
		0.994912,
		0.995579,
		0.996177,
		0.996711,
		0.997186,
		0.997607,
		0.997978,
		0.998304,
		0.998588,
		0.998834,
		0.999046,
		0.999227,
		0.999381,
		0.99951,
		0.999618,
		0.999707,
		0.999779,
		0.999836,
		0.999882,
		0.999917,
		0.999944,
		0.999963,
		0.999977,
		0.999987,
		0.999993,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999994,
		0.999989,
		0.999982,
		0.999973,
		0.999959,
		0.999942,
		0.99992,
		0.999892,
		0.999859,
		0.999819,
		0.001486,
		0.005737,
		0.012464,
		0.021402,
		0.032305,
		0.044948,
		0.059124,
		0.074644,
		0.091333,
		0.109032,
		0.127594,
		0.146884,
		0.166781,
		0.187172,
		0.207954,
		0.229035,
		0.250328,
		0.271757,
		0.293251,
		0.314747,
		0.336187,
		0.357518,
		0.378694,
		0.399673,
		0.420418,
		0.440894,
		0.461072,
		0.480925,
		0.500431,
		0.51957,
		0.538323,
		0.556677,
		0.574618,
		0.592135,
		0.609221,
		0.625869,
		0.642073,
		0.657831,
		0.673139,
		0.687996,
		0.702404,
		0.716363,
		0.729875,
		0.742944,
		0.755573,
		0.767766,
		0.77953,
		0.790869,
		0.801791,
		0.812301,
		0.822408,
		0.832117,
		0.841438,
		0.850378,
		0.858946,
		0.867151,
		0.875,
		0.882503,
		0.889668,
		0.896506,
		0.903024,
		0.909232,
		0.91514,
		0.920755,
		0.926089,
		0.931149,
		0.935944,
		0.940484,
		0.944778,
		0.948834,
		0.952661,
		0.956268,
		0.959663,
		0.962855,
		0.965851,
		0.968661,
		0.971292,
		0.973751,
		0.976047,
		0.978187,
		0.980179,
		0.982029,
		0.983744,
		0.985332,
		0.986798,
		0.98815,
		0.989394,
		0.990536,
		0.991581,
		0.992536,
		0.993406,
		0.994196,
		0.994912,
		0.995558,
		0.99614,
		0.996662,
		0.997128,
		0.997543,
		0.99791,
		0.998234,
		0.998518,
		0.998766,
		0.998981,
		0.999166,
		0.999324,
		0.999459,
		0.999572,
		0.999666,
		0.999743,
		0.999806,
		0.999857,
		0.999897,
		0.999927,
		0.999951,
		0.999968,
		0.99998,
		0.999988,
		0.999994,
		0.999997,
		0.999999,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		0.999999,
		0.999997,
		0.999995,
		0.999991,
		0.001396,
		0.005396,
		0.011736,
		0.020173,
		0.030481,
		0.042454,
		0.055902,
		0.070648,
		0.086531,
		0.103402,
		0.121124,
		0.139573,
		0.158632,
		0.178197,
		0.19817,
		0.218463,
		0.238994,
		0.259691,
		0.280485,
		0.301315,
		0.322126,
		0.342867,
		0.363491,
		0.383959,
		0.404232,
		0.424278,
		0.444067,
		0.463572,
		0.482769,
		0.501639,
		0.520162,
		0.538323,
		0.556109,
		0.573509,
		0.590511,
		0.607109,
		0.623297,
		0.639069,
		0.654422,
		0.669354,
		0.683863,
		0.69795,
		0.711615,
		0.72486,
		0.737688,
		0.750101,
		0.762104,
		0.773702,
		0.784898,
		0.795699,
		0.80611,
		0.816138,
		0.82579,
		0.835071,
		0.843991,
		0.852555,
		0.860772,
		0.868649,
		0.876195,
		0.883417,
		0.890323,
		0.896922,
		0.903223,
		0.909232,
		0.91496,
		0.920413,
		0.925601,
		0.930531,
		0.935212,
		0.939652,
		0.943859,
		0.947842,
		0.951607,
		0.955164,
		0.958519,
		0.961681,
		0.964657,
		0.967454,
		0.970081,
		0.972543,
		0.974848,
		0.977002,
		0.979014,
		0.980889,
		0.982633,
		0.984253,
		0.985756,
		0.987147,
		0.988431,
		0.989616,
		0.990705,
		0.991705,
		0.992621,
		0.993458,
		0.99422,
		0.994912,
		0.995539,
		0.996106,
		0.996615,
		0.997073,
		0.997481,
		0.997845,
		0.998167,
		0.998451,
		0.9987,
		0.998917,
		0.999106,
		0.999268,
		0.999407,
		0.999525,
		0.999624,
		0.999707,
		0.999775,
		0.99983,
		0.999874,
		0.999909,
		0.999936,
		0.999957,
		0.999972,
		0.999982,
		0.99999,
		0.999995,
		0.999997,
		0.999999,
		1,
		1,
		1,
		0.001314,
		0.005084,
		0.011069,
		0.019046,
		0.028807,
		0.040162,
		0.052935,
		0.066963,
		0.082095,
		0.098193,
		0.11513,
		0.132788,
		0.151058,
		0.169842,
		0.189047,
		0.208589,
		0.228392,
		0.248386,
		0.268504,
		0.28869,
		0.308888,
		0.32905,
		0.349131,
		0.369091,
		0.388893,
		0.408505,
		0.427896,
		0.447041,
		0.465916,
		0.484498,
		0.502771,
		0.520718,
		0.538323,
		0.555576,
		0.572465,
		0.588982,
		0.605119,
		0.620871,
		0.636233,
		0.651201,
		0.665773,
		0.679948,
		0.693726,
		0.707107,
		0.720092,
		0.732684,
		0.744885,
		0.756699,
		0.768129,
		0.77918,
		0.789856,
		0.800163,
		0.810106,
		0.819691,
		0.828924,
		0.837812,
		0.846361,
		0.854579,
		0.862471,
		0.870045,
		0.877309,
		0.88427,
		0.890936,
		0.897313,
		0.903409,
		0.909232,
		0.91479,
		0.92009,
		0.92514,
		0.929947,
		0.934518,
		0.938862,
		0.942986,
		0.946896,
		0.950601,
		0.954108,
		0.957423,
		0.960553,
		0.963506,
		0.966289,
		0.968908,
		0.971369,
		0.973679,
		0.975845,
		0.977873,
		0.979768,
		0.981538,
		0.983187,
		0.984721,
		0.986146,
		0.987468,
		0.988691,
		0.989821,
		0.990862,
		0.991821,
		0.9927,
		0.993506,
		0.994242,
		0.994912,
		0.995521,
		0.996073,
		0.996571,
		0.99702,
		0.997422,
		0.997782,
		0.998102,
		0.998385,
		0.998635,
		0.998854,
		0.999046,
		0.999212,
		0.999355,
		0.999477,
		0.999581,
		0.999668,
		0.999741,
		0.999801,
		0.99985,
		0.999889,
		0.99992,
		0.999944,
		0.999962,
		0.999975,
		0.001239,
		0.004799,
		0.010458,
		0.018011,
		0.027267,
		0.038051,
		0.050198,
		0.063557,
		0.07799,
		0.093366,
		0.109566,
		0.12648,
		0.144007,
		0.162051,
		0.180527,
		0.199356,
		0.218463,
		0.237781,
		0.25725,
		0.276811,
		0.296414,
		0.31601,
		0.335557,
		0.355015,
		0.374349,
		0.393526,
		0.412516,
		0.431293,
		0.449834,
		0.468117,
		0.486123,
		0.503836,
		0.52124,
		0.538323,
		0.555074,
		0.571482,
		0.58754,
		0.603241,
		0.618579,
		0.63355,
		0.648151,
		0.66238,
		0.676234,
		0.689715,
		0.702821,
		0.715554,
		0.727916,
		0.739908,
		0.751535,
		0.762798,
		0.773702,
		0.78425,
		0.794448,
		0.804301,
		0.813813,
		0.82299,
		0.831837,
		0.840361,
		0.848568,
		0.856465,
		0.864056,
		0.87135,
		0.878352,
		0.88507,
		0.89151,
		0.897679,
		0.903584,
		0.909232,
		0.91463,
		0.919785,
		0.924704,
		0.929393,
		0.93386,
		0.938112,
		0.942154,
		0.945995,
		0.94964,
		0.953097,
		0.956371,
		0.959469,
		0.962398,
		0.965164,
		0.967772,
		0.97023,
		0.972543,
		0.974716,
		0.976757,
		0.978669,
		0.980459,
		0.982133,
		0.983695,
		0.985151,
		0.986506,
		0.987764,
		0.988931,
		0.990011,
		0.991008,
		0.991928,
		0.992774,
		0.993551,
		0.994262,
		0.994912,
		0.995504,
		0.996042,
		0.996529,
		0.99697,
		0.997366,
		0.997721,
		0.998039,
		0.998321,
		0.998572,
		0.998793,
		0.998987,
		0.999156,
		0.999303,
		0.999429,
		0.999537,
		0.999629,
		0.999707,
		0.00117,
		0.004537,
		0.009896,
		0.017058,
		0.025848,
		0.036101,
		0.047667,
		0.060404,
		0.074184,
		0.088884,
		0.104393,
		0.120607,
		0.137432,
		0.154777,
		0.172561,
		0.19071,
		0.209152,
		0.227824,
		0.246666,
		0.265625,
		0.28465,
		0.303696,
		0.32272,
		0.341684,
		0.360554,
		0.379296,
		0.397884,
		0.416289,
		0.434489,
		0.452462,
		0.470189,
		0.487653,
		0.504839,
		0.521733,
		0.538323,
		0.5546,
		0.570554,
		0.586177,
		0.601465,
		0.61641,
		0.63101,
		0.645261,
		0.65916,
		0.672708,
		0.685901,
		0.698742,
		0.71123,
		0.723368,
		0.735156,
		0.746597,
		0.757694,
		0.76845,
		0.778869,
		0.788955,
		0.798713,
		0.808146,
		0.81726,
		0.82606,
		0.834551,
		0.842738,
		0.850628,
		0.858226,
		0.865538,
		0.872571,
		0.879329,
		0.88582,
		0.892049,
		0.898023,
		0.903749,
		0.909232,
		0.914479,
		0.919497,
		0.924291,
		0.928869,
		0.933236,
		0.937398,
		0.941362,
		0.945135,
		0.948721,
		0.952128,
		0.955361,
		0.958426,
		0.961329,
		0.964076,
		0.966673,
		0.969125,
		0.971437,
		0.973615,
		0.975665,
		0.977591,
		0.9794,
		0.981094,
		0.982681,
		0.984164,
		0.985548,
		0.986838,
		0.988039,
		0.989154,
		0.990188,
		0.991144,
		0.992029,
		0.992844,
		0.993593,
		0.994282,
		0.994912,
		0.995488,
		0.996013,
		0.99649,
		0.996922,
		0.997312,
		0.997663,
		0.997978,
		0.99826,
		0.998511,
		0.998733,
		0.001107,
		0.004296,
		0.009378,
		0.016179,
		0.024536,
		0.034297,
		0.045322,
		0.057479,
		0.070648,
		0.084714,
		0.099574,
		0.11513,
		0.131291,
		0.147975,
		0.165103,
		0.182603,
		0.200411,
		0.218463,
		0.236704,
		0.255081,
		0.273547,
		0.292057,
		0.31057,
		0.32905,
		0.347461,
		0.365774,
		0.383959,
		0.40199,
		0.419845,
		0.437501,
		0.45494,
		0.472143,
		0.489096,
		0.505785,
		0.522198,
		0.538323,
		0.554152,
		0.569676,
		0.584888,
		0.599783,
		0.614355,
		0.628601,
		0.642517,
		0.656102,
		0.669354,
		0.682272,
		0.694856,
		0.707107,
		0.719025,
		0.730613,
		0.741872,
		0.752804,
		0.763413,
		0.773702,
		0.783673,
		0.793332,
		0.802682,
		0.811728,
		0.820474,
		0.828924,
		0.837084,
		0.84496,
		0.852555,
		0.859876,
		0.866927,
		0.873716,
		0.880246,
		0.886525,
		0.892557,
		0.898348,
		0.903905,
		0.909232,
		0.914337,
		0.919224,
		0.9239,
		0.928371,
		0.932642,
		0.936718,
		0.940607,
		0.944313,
		0.947842,
		0.951199,
		0.954391,
		0.957423,
		0.960299,
		0.963026,
		0.965609,
		0.968053,
		0.970362,
		0.972543,
		0.974599,
		0.976536,
		0.978359,
		0.980072,
		0.98168,
		0.983187,
		0.984597,
		0.985916,
		0.987147,
		0.988294,
		0.989361,
		0.990352,
		0.991272,
		0.992122,
		0.992908,
		0.993633,
		0.9943,
		0.994912,
		0.995473,
		0.995985,
		0.996452,
		0.001049,
		0.004073,
		0.0089,
		0.015366,
		0.023321,
		0.032625,
		0.043145,
		0.054761,
		0.067358,
		0.08083,
		0.09508,
		0.110014,
		0.12555,
		0.141606,
		0.15811,
		0.174994,
		0.192195,
		0.209654,
		0.227317,
		0.245133,
		0.263058,
		0.281048,
		0.299063,
		0.317068,
		0.33503,
		0.352917,
		0.370703,
		0.38836,
		0.405867,
		0.423201,
		0.440344,
		0.457278,
		0.473988,
		0.49046,
		0.50668,
		0.522638,
		0.538323,
		0.553728,
		0.568845,
		0.583667,
		0.598188,
		0.612405,
		0.626313,
		0.63991,
		0.653193,
		0.666161,
		0.678813,
		0.69115,
		0.70317,
		0.714875,
		0.726267,
		0.737347,
		0.748116,
		0.758579,
		0.768736,
		0.778592,
		0.78815,
		0.797413,
		0.806386,
		0.815072,
		0.823476,
		0.831602,
		0.839455,
		0.84704,
		0.854361,
		0.861423,
		0.868232,
		0.874792,
		0.881109,
		0.887188,
		0.893035,
		0.898654,
		0.904051,
		0.909232,
		0.914202,
		0.918966,
		0.923529,
		0.927897,
		0.932076,
		0.93607,
		0.939885,
		0.943527,
		0.946999,
		0.950308,
		0.953459,
		0.956457,
		0.959306,
		0.962012,
		0.964579,
		0.967013,
		0.969317,
		0.971498,
		0.973558,
		0.975504,
		0.977338,
		0.979066,
		0.980692,
		0.98222,
		0.983655,
		0.984999,
		0.986257,
		0.987433,
		0.988531,
		0.989554,
		0.990506,
		0.99139,
		0.99221,
		0.000995,
		0.003868,
		0.008457,
		0.014613,
		0.022194,
		0.031071,
		0.041121,
		0.05223,
		0.064291,
		0.077206,
		0.09088,
		0.105229,
		0.120172,
		0.135635,
		0.151546,
		0.167843,
		0.184465,
		0.201355,
		0.218463,
		0.23574,
		0.253142,
		0.270627,
		0.288158,
		0.3057,
		0.32322,
		0.340688,
		0.358077,
		0.375363,
		0.392521,
		0.409531,
		0.426374,
		0.443032,
		0.45949,
		0.475734,
		0.491749,
		0.507526,
		0.523054,
		0.538323,
		0.553326,
		0.568056,
		0.582507,
		0.596674,
		0.610552,
		0.624138,
		0.637429,
		0.650423,
		0.663118,
		0.675515,
		0.687611,
		0.699408,
		0.710906,
		0.722107,
		0.73301,
		0.743619,
		0.753936,
		0.763962,
		0.773702,
		0.783156,
		0.79233,
		0.801227,
		0.809849,
		0.818201,
		0.826287,
		0.834112,
		0.841678,
		0.848992,
		0.856057,
		0.862877,
		0.869459,
		0.875806,
		0.881923,
		0.887815,
		0.893487,
		0.898944,
		0.904191,
		0.909232,
		0.914074,
		0.91872,
		0.923176,
		0.927446,
		0.931537,
		0.935451,
		0.939196,
		0.942774,
		0.946192,
		0.949453,
		0.952563,
		0.955526,
		0.958348,
		0.961031,
		0.963582,
		0.966004,
		0.968302,
		0.97048,
		0.972543,
		0.974494,
		0.976338,
		0.978079,
		0.97972,
		0.981267,
		0.982721,
		0.984089,
		0.985372,
		0.000945,
		0.003677,
		0.008046,
		0.013913,
		0.021147,
		0.029626,
		0.039236,
		0.049871,
		0.061429,
		0.073818,
		0.086952,
		0.100747,
		0.11513,
		0.130029,
		0.145377,
		0.161115,
		0.177183,
		0.193529,
		0.210104,
		0.226862,
		0.243759,
		0.260756,
		0.277816,
		0.294906,
		0.311993,
		0.32905,
		0.346048,
		0.362964,
		0.379775,
		0.39646,
		0.413,
		0.429378,
		0.445578,
		0.461585,
		0.477387,
		0.492972,
		0.508329,
		0.523449,
		0.538323,
		0.552945,
		0.567308,
		0.581406,
		0.595234,
		0.608789,
		0.622067,
		0.635065,
		0.647781,
		0.660215,
		0.672365,
		0.68423,
		0.69581,
		0.707107,
		0.71812,
		0.728851,
		0.739302,
		0.749475,
		0.759371,
		0.768993,
		0.778343,
		0.787425,
		0.796242,
		0.804797,
		0.813093,
		0.821134,
		0.828924,
		0.836467,
		0.843766,
		0.850827,
		0.857652,
		0.864247,
		0.870615,
		0.876762,
		0.88269,
		0.888407,
		0.893914,
		0.899218,
		0.904322,
		0.909232,
		0.913952,
		0.918486,
		0.92284,
		0.927017,
		0.931022,
		0.93486,
		0.938536,
		0.942053,
		0.945417,
		0.948631,
		0.951701,
		0.95463,
		0.957423,
		0.960083,
		0.962616,
		0.965026,
		0.967315,
		0.969489,
		0.971552,
		0.973507,
		0.975358,
		0.000899,
		0.003501,
		0.007665,
		0.013263,
		0.020173,
		0.02828,
		0.037478,
		0.047667,
		0.058753,
		0.070648,
		0.08327,
		0.096544,
		0.110396,
		0.12476,
		0.139573,
		0.154777,
		0.170317,
		0.186142,
		0.202206,
		0.218463,
		0.234873,
		0.251397,
		0.268,
		0.28465,
		0.301315,
		0.317968,
		0.334582,
		0.351133,
		0.367599,
		0.383959,
		0.400195,
		0.416289,
		0.432226,
		0.447991,
		0.463572,
		0.478955,
		0.494131,
		0.50909,
		0.523824,
		0.538323,
		0.552583,
		0.566596,
		0.580358,
		0.593863,
		0.607109,
		0.620093,
		0.63281,
		0.645261,
		0.657442,
		0.669354,
		0.680995,
		0.692366,
		0.703467,
		0.714298,
		0.72486,
		0.735156,
		0.745185,
		0.754952,
		0.764456,
		0.773702,
		0.78269,
		0.791425,
		0.79991,
		0.808146,
		0.816138,
		0.823889,
		0.831403,
		0.838682,
		0.845732,
		0.852555,
		0.859156,
		0.865538,
		0.871707,
		0.877665,
		0.883417,
		0.888967,
		0.894319,
		0.899478,
		0.904447,
		0.909232,
		0.913836,
		0.918264,
		0.922519,
		0.926607,
		0.930531,
		0.934295,
		0.937904,
		0.941362,
		0.944673,
		0.947842,
		0.950871,
		0.953766,
		0.95653,
		0.959167,
		0.961681,
		0.000857,
		0.003336,
		0.00731,
		0.012657,
		0.019264,
		0.027023,
		0.035835,
		0.045605,
		0.056247,
		0.067676,
		0.079817,
		0.092596,
		0.105945,
		0.119801,
		0.134105,
		0.148801,
		0.163836,
		0.179163,
		0.194735,
		0.210511,
		0.226451,
		0.242518,
		0.258679,
		0.2749,
		0.291153,
		0.307411,
		0.323646,
		0.339837,
		0.355962,
		0.371999,
		0.387931,
		0.403741,
		0.419412,
		0.43493,
		0.450283,
		0.465459,
		0.480445,
		0.495233,
		0.509814,
		0.52418,
		0.538323,
		0.552238,
		0.565918,
		0.579359,
		0.592557,
		0.605508,
		0.618209,
		0.630658,
		0.642853,
		0.654791,
		0.666474,
		0.677898,
		0.689066,
		0.699976,
		0.710629,
		0.721027,
		0.73117,
		0.741059,
		0.750696,
		0.760084,
		0.769224,
		0.778118,
		0.78677,
		0.795181,
		0.803355,
		0.811294,
		0.819002,
		0.826481,
		0.833736,
		0.840769,
		0.847584,
		0.854185,
		0.860575,
		0.866759,
		0.872739,
		0.878519,
		0.884104,
		0.889497,
		0.894703,
		0.899725,
		0.904566,
		0.909232,
		0.913726,
		0.918052,
		0.922214,
		0.926215,
		0.930061,
		0.933754,
		0.937299,
		0.940699,
		0.943959,
		0.000817,
		0.003183,
		0.006979,
		0.012092,
		0.018415,
		0.025848,
		0.034297,
		0.043674,
		0.053897,
		0.064888,
		0.076573,
		0.088884,
		0.101756,
		0.11513,
		0.128949,
		0.14316,
		0.157713,
		0.172561,
		0.187663,
		0.202976,
		0.218463,
		0.234088,
		0.249819,
		0.265625,
		0.281477,
		0.297347,
		0.313213,
		0.32905,
		0.344836,
		0.360554,
		0.376183,
		0.391707,
		0.407111,
		0.42238,
		0.437501,
		0.452462,
		0.467253,
		0.481862,
		0.496282,
		0.510503,
		0.524519,
		0.538323,
		0.551909,
		0.565272,
		0.578407,
		0.591311,
		0.603979,
		0.61641,
		0.628601,
		0.640549,
		0.652255,
		0.663715,
		0.674931,
		0.685901,
		0.696627,
		0.707107,
		0.717343,
		0.727336,
		0.737086,
		0.746597,
		0.755868,
		0.764902,
		0.773702,
		0.782268,
		0.790604,
		0.798713,
		0.806596,
		0.814257,
		0.821699,
		0.828924,
		0.835936,
		0.842738,
		0.849334,
		0.855726,
		0.861918,
		0.867913,
		0.873716,
		0.879329,
		0.884756,
		0.890001,
		0.895068,
		0.899959,
		0.90468,
		0.909232,
		0.913621,
		0.91785,
		0.921922,
		0.00078,
		0.003041,
		0.006671,
		0.011564,
		0.017621,
		0.024747,
		0.032856,
		0.041863,
		0.051692,
		0.062267,
		0.073521,
		0.085389,
		0.097809,
		0.110724,
		0.124081,
		0.137829,
		0.151921,
		0.166312,
		0.180962,
		0.19583,
		0.21088,
		0.226079,
		0.241394,
		0.256796,
		0.272257,
		0.28775,
		0.303253,
		0.318742,
		0.334196,
		0.349596,
		0.364925,
		0.380165,
		0.395301,
		0.410318,
		0.425205,
		0.439948,
		0.454537,
		0.468961,
		0.483212,
		0.497281,
		0.51116,
		0.524843,
		0.538323,
		0.551596,
		0.564656,
		0.577498,
		0.59012,
		0.602518,
		0.61469,
		0.626633,
		0.638345,
		0.649825,
		0.661072,
		0.672085,
		0.682865,
		0.69341,
		0.703721,
		0.7138,
		0.723646,
		0.73326,
		0.742645,
		0.751801,
		0.76073,
		0.769433,
		0.777914,
		0.786174,
		0.794215,
		0.80204,
		0.809651,
		0.817052,
		0.824244,
		0.83123,
		0.838015,
		0.8446,
		0.850988,
		0.857183,
		0.863189,
		0.869007,
		0.874642,
		0.880097,
		0.885375,
		0.89048,
		0.895415,
		0.000745,
		0.002907,
		0.006382,
		0.011069,
		0.016877,
		0.023716,
		0.031504,
		0.040162,
		0.049618,
		0.059802,
		0.070648,
		0.082095,
		0.094085,
		0.106564,
		0.119481,
		0.132788,
		0.146439,
		0.160392,
		0.174607,
		0.189047,
		0.203676,
		0.218463,
		0.233376,
		0.248386,
		0.263466,
		0.278592,
		0.29374,
		0.308888,
		0.324015,
		0.339103,
		0.354133,
		0.369091,
		0.383959,
		0.398725,
		0.413374,
		0.427896,
		0.442279,
		0.456513,
		0.470589,
		0.484498,
		0.498233,
		0.511786,
		0.525152,
		0.538323,
		0.551297,
		0.564067,
		0.57663,
		0.588982,
		0.601121,
		0.613044,
		0.624748,
		0.636233,
		0.647496,
		0.658536,
		0.669354,
		0.679948,
		0.690319,
		0.700466,
		0.71039,
		0.720092,
		0.729573,
		0.738833,
		0.747875,
		0.756699,
		0.765307,
		0.773702,
		0.781884,
		0.789856,
		0.79762,
		0.805179,
		0.812535,
		0.819691,
		0.826649,
		0.833411,
		0.839981,
		0.846361,
		0.852555,
		0.858565,
		0.864394,
		0.000713,
		0.002783,
		0.006111,
		0.010606,
		0.016179,
		0.022747,
		0.030233,
		0.038563,
		0.047667,
		0.057479,
		0.067939,
		0.078987,
		0.090569,
		0.102633,
		0.11513,
		0.128015,
		0.141244,
		0.154777,
		0.168575,
		0.182603,
		0.196828,
		0.211216,
		0.22574,
		0.24037,
		0.255081,
		0.269849,
		0.28465,
		0.299464,
		0.31427,
		0.32905,
		0.343786,
		0.358463,
		0.373065,
		0.387578,
		0.40199,
		0.416289,
		0.430464,
		0.444504,
		0.4584,
		0.472143,
		0.485726,
		0.499142,
		0.512384,
		0.525446,
		0.538323,
		0.55101,
		0.563503,
		0.575799,
		0.587893,
		0.599783,
		0.611467,
		0.622942,
		0.634207,
		0.645261,
		0.656102,
		0.66673,
		0.677145,
		0.687346,
		0.697333,
		0.707107,
		0.716668,
		0.726017,
		0.735156,
		0.744084,
		0.752804,
		0.761317,
		0.769624,
		0.777728,
		0.78563,
		0.793332,
		0.800837,
		0.808146,
		0.815262,
		0.822187,
		0.828924,
		0.000682,
		0.002666,
		0.005858,
		0.010171,
		0.015523,
		0.021837,
		0.029038,
		0.037057,
		0.045828,
		0.055289,
		0.065382,
		0.076051,
		0.087245,
		0.098913,
		0.11101,
		0.123492,
		0.136317,
		0.149448,
		0.162846,
		0.176478,
		0.190312,
		0.204316,
		0.218463,
		0.232725,
		0.247077,
		0.261496,
		0.27596,
		0.290446,
		0.304937,
		0.319414,
		0.333861,
		0.34826,
		0.362598,
		0.37686,
		0.391034,
		0.405109,
		0.419073,
		0.432915,
		0.446628,
		0.460201,
		0.473628,
		0.4869,
		0.500011,
		0.512956,
		0.525728,
		0.538323,
		0.550737,
		0.562964,
		0.575003,
		0.586849,
		0.5985,
		0.609954,
		0.621209,
		0.632263,
		0.643115,
		0.653763,
		0.664208,
		0.674449,
		0.684484,
		0.694316,
		0.703943,
		0.713366,
		0.722587,
		0.731605,
		0.740422,
		0.749038,
		0.757456,
		0.765677,
		0.773702,
		0.781532,
		0.789171,
		0.000654,
		0.002556,
		0.00562,
		0.009762,
		0.014907,
		0.02098,
		0.027912,
		0.035637,
		0.044093,
		0.053221,
		0.062966,
		0.073275,
		0.084099,
		0.095391,
		0.107106,
		0.119202,
		0.131641,
		0.144385,
		0.1574,
		0.170651,
		0.184108,
		0.197741,
		0.211524,
		0.225429,
		0.239433,
		0.253513,
		0.267647,
		0.281814,
		0.295997,
		0.310176,
		0.324337,
		0.338462,
		0.352538,
		0.366551,
		0.380488,
		0.394338,
		0.40809,
		0.421733,
		0.435259,
		0.448658,
		0.461923,
		0.475047,
		0.488022,
		0.500843,
		0.513503,
		0.525998,
		0.538323,
		0.550474,
		0.562448,
		0.57424,
		0.585848,
		0.59727,
		0.608503,
		0.619546,
		0.630396,
		0.641052,
		0.651515,
		0.661782,
		0.671853,
		0.681729,
		0.691409,
		0.700893,
		0.710181,
		0.719275,
		0.728175,
		0.736881,
		0.745396,
		0.000627,
		0.002453,
		0.005396,
		0.009378,
		0.014327,
		0.020173,
		0.02685,
		0.034297,
		0.042454,
		0.051267,
		0.060681,
		0.070648,
		0.081119,
		0.092052,
		0.103402,
		0.11513,
		0.127199,
		0.139573,
		0.152218,
		0.165103,
		0.178197,
		0.191473,
		0.204903,
		0.218463,
		0.232129,
		0.245878,
		0.259691,
		0.273547,
		0.287427,
		0.301315,
		0.315194,
		0.32905,
		0.342867,
		0.356632,
		0.370333,
		0.383959,
		0.397498,
		0.410941,
		0.424278,
		0.437501,
		0.450601,
		0.463572,
		0.476406,
		0.489096,
		0.501639,
		0.514027,
		0.526257,
		0.538323,
		0.550223,
		0.561953,
		0.573509,
		0.584888,
		0.596089,
		0.607109,
		0.617947,
		0.628601,
		0.639069,
		0.649351,
		0.659446,
		0.669354,
		0.679074,
		0.688606,
		0.69795,
		0.000602,
		0.002356,
		0.005185,
		0.009016,
		0.013779,
		0.019411,
		0.025848,
		0.033031,
		0.040905,
		0.049418,
		0.058518,
		0.068159,
		0.078295,
		0.088884,
		0.099885,
		0.111261,
		0.122976,
		0.134994,
		0.147285,
		0.159817,
		0.172561,
		0.185491,
		0.198581,
		0.211807,
		0.225145,
		0.238574,
		0.252073,
		0.265625,
		0.27921,
		0.292812,
		0.306416,
		0.320005,
		0.333566,
		0.347086,
		0.360554,
		0.373956,
		0.387283,
		0.400525,
		0.413672,
		0.426716,
		0.439648,
		0.452462,
		0.465151,
		0.477707,
		0.490126,
		0.502402,
		0.514529,
		0.526505,
		0.538323,
		0.549982,
		0.561477,
		0.572806,
		0.583966,
		0.594955,
		0.60577,
		0.61641,
		0.626874,
		0.63716,
		0.647268,
		0.000579,
		0.002265,
		0.004986,
		0.008674,
		0.013263,
		0.018691,
		0.0249,
		0.031834,
		0.039439,
		0.047667,
		0.056468,
		0.065798,
		0.075614,
		0.085875,
		0.096544,
		0.107582,
		0.118957,
		0.130634,
		0.142584,
		0.154777,
		0.167184,
		0.179781,
		0.192542,
		0.205443,
		0.218463,
		0.23158,
		0.244776,
		0.258031,
		0.271328,
		0.28465,
		0.297982,
		0.31131,
		0.324619,
		0.337898,
		0.351133,
		0.364313,
		0.377429,
		0.390469,
		0.403426,
		0.416289,
		0.429052,
		0.441707,
		0.454246,
		0.466665,
		0.478955,
		0.491113,
		0.503133,
		0.515011,
		0.526742,
		0.538323,
		0.54975,
		0.561021,
		0.572131,
		0.583079,
		0.593863,
		0.000557,
		0.002179,
		0.004799,
		0.008351,
		0.012775,
		0.018011,
		0.024004,
		0.0307,
		0.038051,
		0.046007,
		0.054523,
		0.063557,
		0.073068,
		0.083016,
		0.093366,
		0.104082,
		0.11513,
		0.12648,
		0.138102,
		0.149968,
		0.162051,
		0.174326,
		0.186768,
		0.199356,
		0.212067,
		0.224882,
		0.237781,
		0.250747,
		0.263763,
		0.276811,
		0.289878,
		0.302949,
		0.31601,
		0.32905,
		0.342055,
		0.355015,
		0.36792,
		0.38076,
		0.393526,
		0.406208,
		0.4188,
		0.431293,
		0.443682,
		0.455958,
		0.468117,
		0.480153,
		0.492061,
		0.503836,
		0.515474,
		0.526971,
		0.538323,
		0.000536,
		0.002098,
		0.004622,
		0.008046,
		0.012313,
		0.017367,
		0.023155,
		0.029626,
		0.036734,
		0.044432,
		0.052677,
		0.061429,
		0.070648,
		0.080297,
		0.090341,
		0.100747,
		0.111483,
		0.122519,
		0.133826,
		0.145377,
		0.157147,
		0.169111,
		0.181245,
		0.193529,
		0.205942,
		0.218463,
		0.231074,
		0.243759,
		0.256499,
		0.26928,
		0.282087,
		0.294906,
		0.307723,
		0.320527,
		0.333306,
		0.346048,
		0.358744,
		0.371384,
		0.383959,
		0.39646,
		0.40888,
		0.42121,
		0.433445,
		0.445578,
		0.457602,
		0.469512,
		0.481304,
		0.000516,
		0.002021,
		0.004454,
		0.007758,
		0.011876,
		0.016757,
		0.02235,
		0.028608,
		0.035484,
		0.042936,
		0.050923,
		0.059405,
		0.068345,
		0.077708,
		0.087461,
		0.09757,
		0.108006,
		0.118739,
		0.129743,
		0.140991,
		0.152459,
		0.164123,
		0.17596,
		0.18795,
		0.200072,
		0.212308,
		0.224639,
		0.237049,
		0.249522,
		0.262041,
		0.274594,
		0.287165,
		0.299743,
		0.312315,
		0.32487,
		0.337397,
		0.349886,
		0.362328,
		0.374712,
		0.387032,
		0.39928,
		0.411446,
		0.423526,
		0.000497,
		0.001948,
		0.004296,
		0.007484,
		0.011462,
		0.016179,
		0.021587,
		0.027641,
		0.034297,
		0.041515,
		0.049255,
		0.057479,
		0.066153,
		0.075242,
		0.084714,
		0.094539,
		0.104687,
		0.11513,
		0.125842,
		0.136799,
		0.147975,
		0.159349,
		0.170898,
		0.182603,
		0.194445,
		0.206404,
		0.218463,
		0.230606,
		0.242817,
		0.255081,
		0.267385,
		0.279714,
		0.292057,
		0.304401,
		0.316735,
		0.32905,
		0.341334,
		0.353578,
		0.365774,
		0.000479,
		0.001879,
		0.004145,
		0.007225,
		0.011069,
		0.01563,
		0.020862,
		0.026722,
		0.033168,
		0.040162,
		0.047667,
		0.055645,
		0.064064,
		0.072891,
		0.082095,
		0.091646,
		0.101517,
		0.111681,
		0.122113,
		0.132788,
		0.143683,
		0.154777,
		0.166048,
		0.177478,
		0.189047,
		0.200737,
		0.212531,
		0.224414,
		0.236371,
		0.248386,
		0.260446,
		0.272538,
		0.28465,
		0.29677,
		0.308888,
		0.000462,
		0.001814,
		0.004003,
		0.006979,
		0.010696,
		0.015108,
		0.020173,
		0.025848,
		0.032094,
		0.038875,
		0.046154,
		0.053897,
		0.062072,
		0.070648,
		0.079594,
		0.088884,
		0.098489,
		0.108384,
		0.118545,
		0.128949,
		0.139573,
		0.150396,
		0.161399,
		0.172561,
		0.183866,
		0.195295,
		0.206833,
		0.218463,
		0.230171,
		0.241943,
		0.253765,
		0.000447,
		0.001752,
		0.003868,
		0.006746,
		0.010342,
		0.014613,
		0.019517,
		0.025016,
		0.031071,
		0.037648,
		0.044712,
		0.05223,
		0.060171,
		0.068506,
		0.077206,
		0.086243,
		0.095593,
		0.105229,
		0.11513,
		0.125272,
		0.135635,
		0.146196,
		0.156939,
		0.167843,
		0.178891,
		0.190067,
		0.201355,
		0.000431,
		0.001693,
		0.003739,
		0.006524,
		0.010005,
		0.014141,
		0.018893,
		0.024223,
		0.030097,
		0.036479,
		0.043337,
		0.050639,
		0.058356,
		0.06646,
		0.074922,
		0.083718,
		0.092821,
		0.102209,
		0.111859,
		0.121749,
		0.131859,
		0.142168,
		0.152659,
		0.000417,
		0.001638,
		0.003617,
		0.006313,
		0.009684,
		0.013691,
		0.018298,
		0.023468,
		0.029167,
		0.035363,
		0.042023,
		0.049119,
		0.056622,
		0.064503,
		0.072738,
		0.081301,
		0.090168,
		0.099316,
		0.108724,
		0.000403,
		0.001584,
		0.003501,
		0.006111,
		0.009378,
		0.013263,
		0.017731,
		0.022747,
		0.02828,
		0.034297,
		0.040769,
		0.047667,
		0.054963,
		0.062632,
		0.070648,
		0.00039,
		0.001534,
		0.00339,
		0.00592,
		0.009086,
		0.012854,
		0.01719,
		0.022059,
		0.027432,
		0.033279,
		0.039569,
		0.000378,
		0.001486,
		0.003284,
		0.005737,
		0.008808,
		0.012464,
		0.016673,
		0.000366,
		0.00144,
		0.003183,
	};

	float coeff_pos[616952][3] = {
		{ 256, 256, 1 },
		{ 256, 257, 1 },
		{ 256, 258, 1 },
		{ 257, 256, 1 },
		{ 257, 257, 1 },
		{ 257, 258, 1 },
		{ 258, 256, 1 },
		{ 258, 257, 1 },
		{ 258, 258, 1 },
		{ 251, 250, 2 },
		{ 251, 264, 2 },
		{ 252, 250, 2 },
		{ 252, 251, 2 },
		{ 252, 263, 2 },
		{ 252, 264, 2 },
		{ 253, 250, 2 },
		{ 253, 251, 2 },
		{ 253, 252, 2 },
		{ 253, 262, 2 },
		{ 253, 263, 2 },
		{ 253, 264, 2 },
		{ 254, 250, 2 },
		{ 254, 251, 2 },
		{ 254, 252, 2 },
		{ 254, 253, 2 },
		{ 254, 261, 2 },
		{ 254, 262, 2 },
		{ 254, 263, 2 },
		{ 254, 264, 2 },
		{ 255, 250, 2 },
		{ 255, 251, 2 },
		{ 255, 252, 2 },
		{ 255, 253, 2 },
		{ 255, 254, 2 },
		{ 255, 260, 2 },
		{ 255, 261, 2 },
		{ 255, 262, 2 },
		{ 255, 263, 2 },
		{ 255, 264, 2 },
		{ 256, 250, 2 },
		{ 256, 251, 2 },
		{ 256, 252, 2 },
		{ 256, 253, 2 },
		{ 256, 254, 2 },
		{ 256, 255, 2 },
		{ 256, 259, 2 },
		{ 256, 260, 2 },
		{ 256, 261, 2 },
		{ 256, 262, 2 },
		{ 256, 263, 2 },
		{ 256, 264, 2 },
		{ 257, 250, 2 },
		{ 257, 251, 2 },
		{ 257, 252, 2 },
		{ 257, 253, 2 },
		{ 257, 254, 2 },
		{ 257, 255, 2 },
		{ 257, 259, 2 },
		{ 257, 260, 2 },
		{ 257, 261, 2 },
		{ 257, 262, 2 },
		{ 257, 263, 2 },
		{ 257, 264, 2 },
		{ 258, 250, 2 },
		{ 258, 251, 2 },
		{ 258, 252, 2 },
		{ 258, 253, 2 },
		{ 258, 254, 2 },
		{ 258, 255, 2 },
		{ 258, 259, 2 },
		{ 258, 260, 2 },
		{ 258, 261, 2 },
		{ 258, 262, 2 },
		{ 258, 263, 2 },
		{ 258, 264, 2 },
		{ 259, 250, 2 },
		{ 259, 251, 2 },
		{ 259, 252, 2 },
		{ 259, 253, 2 },
		{ 259, 254, 2 },
		{ 259, 260, 2 },
		{ 259, 261, 2 },
		{ 259, 262, 2 },
		{ 259, 263, 2 },
		{ 259, 264, 2 },
		{ 260, 250, 2 },
		{ 260, 251, 2 },
		{ 260, 252, 2 },
		{ 260, 253, 2 },
		{ 260, 261, 2 },
		{ 260, 262, 2 },
		{ 260, 263, 2 },
		{ 260, 264, 2 },
		{ 261, 250, 2 },
		{ 261, 251, 2 },
		{ 261, 252, 2 },
		{ 261, 262, 2 },
		{ 261, 263, 2 },
		{ 261, 264, 2 },
		{ 262, 250, 2 },
		{ 262, 251, 2 },
		{ 262, 263, 2 },
		{ 262, 264, 2 },
		{ 263, 250, 2 },
		{ 263, 264, 2 },
		{ 250, 258, 3 },
		{ 250, 259, 3 },
		{ 250, 260, 3 },
		{ 250, 261, 3 },
		{ 250, 262, 3 },
		{ 250, 263, 3 },
		{ 250, 264, 3 },
		{ 251, 258, 3 },
		{ 251, 259, 3 },
		{ 251, 260, 3 },
		{ 251, 261, 3 },
		{ 251, 262, 3 },
		{ 251, 263, 3 },
		{ 251, 264, 3 },
		{ 252, 258, 3 },
		{ 252, 259, 3 },
		{ 252, 260, 3 },
		{ 252, 261, 3 },
		{ 252, 262, 3 },
		{ 252, 263, 3 },
		{ 252, 264, 3 },
		{ 253, 258, 3 },
		{ 253, 259, 3 },
		{ 253, 260, 3 },
		{ 253, 261, 3 },
		{ 253, 262, 3 },
		{ 253, 263, 3 },
		{ 253, 264, 3 },
		{ 254, 258, 3 },
		{ 254, 259, 3 },
		{ 254, 260, 3 },
		{ 254, 261, 3 },
		{ 254, 262, 3 },
		{ 254, 263, 3 },
		{ 254, 264, 3 },
		{ 255, 258, 3 },
		{ 255, 259, 3 },
		{ 255, 260, 3 },
		{ 255, 261, 3 },
		{ 255, 262, 3 },
		{ 255, 263, 3 },
		{ 255, 264, 3 },
		{ 256, 259, 3 },
		{ 256, 260, 3 },
		{ 256, 261, 3 },
		{ 256, 262, 3 },
		{ 256, 263, 3 },
		{ 256, 264, 3 },
		{ 258, 250, 3 },
		{ 258, 251, 3 },
		{ 258, 252, 3 },
		{ 258, 253, 3 },
		{ 258, 254, 3 },
		{ 258, 255, 3 },
		{ 259, 250, 3 },
		{ 259, 251, 3 },
		{ 259, 252, 3 },
		{ 259, 253, 3 },
		{ 259, 254, 3 },
		{ 259, 255, 3 },
		{ 259, 256, 3 },
		{ 260, 250, 3 },
		{ 260, 251, 3 },
		{ 260, 252, 3 },
		{ 260, 253, 3 },
		{ 260, 254, 3 },
		{ 260, 255, 3 },
		{ 260, 256, 3 },
		{ 261, 250, 3 },
		{ 261, 251, 3 },
		{ 261, 252, 3 },
		{ 261, 253, 3 },
		{ 261, 254, 3 },
		{ 261, 255, 3 },
		{ 261, 256, 3 },
		{ 262, 250, 3 },
		{ 262, 251, 3 },
		{ 262, 252, 3 },
		{ 262, 253, 3 },
		{ 262, 254, 3 },
		{ 262, 255, 3 },
		{ 262, 256, 3 },
		{ 263, 250, 3 },
		{ 263, 251, 3 },
		{ 263, 252, 3 },
		{ 263, 253, 3 },
		{ 263, 254, 3 },
		{ 263, 255, 3 },
		{ 263, 256, 3 },
		{ 264, 250, 3 },
		{ 264, 251, 3 },
		{ 264, 252, 3 },
		{ 264, 253, 3 },
		{ 264, 254, 3 },
		{ 264, 255, 3 },
		{ 264, 256, 3 },
		{ 250, 251, 4 },
		{ 250, 252, 4 },
		{ 250, 253, 4 },
		{ 250, 254, 4 },
		{ 250, 255, 4 },
		{ 250, 256, 4 },
		{ 250, 257, 4 },
		{ 250, 258, 4 },
		{ 250, 259, 4 },
		{ 250, 260, 4 },
		{ 250, 261, 4 },
		{ 250, 262, 4 },
		{ 250, 263, 4 },
		{ 251, 252, 4 },
		{ 251, 253, 4 },
		{ 251, 254, 4 },
		{ 251, 255, 4 },
		{ 251, 256, 4 },
		{ 251, 257, 4 },
		{ 251, 258, 4 },
		{ 251, 259, 4 },
		{ 251, 260, 4 },
		{ 251, 261, 4 },
		{ 251, 262, 4 },
		{ 252, 253, 4 },
		{ 252, 254, 4 },
		{ 252, 255, 4 },
		{ 252, 256, 4 },
		{ 252, 257, 4 },
		{ 252, 258, 4 },
		{ 252, 259, 4 },
		{ 252, 260, 4 },
		{ 252, 261, 4 },
		{ 253, 254, 4 },
		{ 253, 255, 4 },
		{ 253, 256, 4 },
		{ 253, 257, 4 },
		{ 253, 258, 4 },
		{ 253, 259, 4 },
		{ 253, 260, 4 },
		{ 254, 255, 4 },
		{ 254, 256, 4 },
		{ 254, 257, 4 },
		{ 254, 258, 4 },
		{ 254, 259, 4 },
		{ 255, 256, 4 },
		{ 255, 257, 4 },
		{ 255, 258, 4 },
		{ 259, 256, 4 },
		{ 259, 257, 4 },
		{ 259, 258, 4 },
		{ 260, 255, 4 },
		{ 260, 256, 4 },
		{ 260, 257, 4 },
		{ 260, 258, 4 },
		{ 260, 259, 4 },
		{ 261, 254, 4 },
		{ 261, 255, 4 },
		{ 261, 256, 4 },
		{ 261, 257, 4 },
		{ 261, 258, 4 },
		{ 261, 259, 4 },
		{ 261, 260, 4 },
		{ 262, 253, 4 },
		{ 262, 254, 4 },
		{ 262, 255, 4 },
		{ 262, 256, 4 },
		{ 262, 257, 4 },
		{ 262, 258, 4 },
		{ 262, 259, 4 },
		{ 262, 260, 4 },
		{ 262, 261, 4 },
		{ 263, 252, 4 },
		{ 263, 253, 4 },
		{ 263, 254, 4 },
		{ 263, 255, 4 },
		{ 263, 256, 4 },
		{ 263, 257, 4 },
		{ 263, 258, 4 },
		{ 263, 259, 4 },
		{ 263, 260, 4 },
		{ 263, 261, 4 },
		{ 263, 262, 4 },
		{ 264, 251, 4 },
		{ 264, 252, 4 },
		{ 264, 253, 4 },
		{ 264, 254, 4 },
		{ 264, 255, 4 },
		{ 264, 256, 4 },
		{ 264, 257, 4 },
		{ 264, 258, 4 },
		{ 264, 259, 4 },
		{ 264, 260, 4 },
		{ 264, 261, 4 },
		{ 264, 262, 4 },
		{ 264, 263, 4 },
		{ 250, 250, 5 },
		{ 250, 251, 5 },
		{ 250, 252, 5 },
		{ 250, 253, 5 },
		{ 250, 254, 5 },
		{ 250, 255, 5 },
		{ 250, 256, 5 },
		{ 251, 250, 5 },
		{ 251, 251, 5 },
		{ 251, 252, 5 },
		{ 251, 253, 5 },
		{ 251, 254, 5 },
		{ 251, 255, 5 },
		{ 251, 256, 5 },
		{ 252, 250, 5 },
		{ 252, 251, 5 },
		{ 252, 252, 5 },
		{ 252, 253, 5 },
		{ 252, 254, 5 },
		{ 252, 255, 5 },
		{ 252, 256, 5 },
		{ 253, 250, 5 },
		{ 253, 251, 5 },
		{ 253, 252, 5 },
		{ 253, 253, 5 },
		{ 253, 254, 5 },
		{ 253, 255, 5 },
		{ 253, 256, 5 },
		{ 254, 250, 5 },
		{ 254, 251, 5 },
		{ 254, 252, 5 },
		{ 254, 253, 5 },
		{ 254, 254, 5 },
		{ 254, 255, 5 },
		{ 254, 256, 5 },
		{ 255, 250, 5 },
		{ 255, 251, 5 },
		{ 255, 252, 5 },
		{ 255, 253, 5 },
		{ 255, 254, 5 },
		{ 255, 255, 5 },
		{ 255, 256, 5 },
		{ 256, 250, 5 },
		{ 256, 251, 5 },
		{ 256, 252, 5 },
		{ 256, 253, 5 },
		{ 256, 254, 5 },
		{ 256, 255, 5 },
		{ 258, 259, 5 },
		{ 258, 260, 5 },
		{ 258, 261, 5 },
		{ 258, 262, 5 },
		{ 258, 263, 5 },
		{ 258, 264, 5 },
		{ 259, 258, 5 },
		{ 259, 259, 5 },
		{ 259, 260, 5 },
		{ 259, 261, 5 },
		{ 259, 262, 5 },
		{ 259, 263, 5 },
		{ 259, 264, 5 },
		{ 260, 258, 5 },
		{ 260, 259, 5 },
		{ 260, 260, 5 },
		{ 260, 261, 5 },
		{ 260, 262, 5 },
		{ 260, 263, 5 },
		{ 260, 264, 5 },
		{ 261, 258, 5 },
		{ 261, 259, 5 },
		{ 261, 260, 5 },
		{ 261, 261, 5 },
		{ 261, 262, 5 },
		{ 261, 263, 5 },
		{ 261, 264, 5 },
		{ 262, 258, 5 },
		{ 262, 259, 5 },
		{ 262, 260, 5 },
		{ 262, 261, 5 },
		{ 262, 262, 5 },
		{ 262, 263, 5 },
		{ 262, 264, 5 },
		{ 263, 258, 5 },
		{ 263, 259, 5 },
		{ 263, 260, 5 },
		{ 263, 261, 5 },
		{ 263, 262, 5 },
		{ 263, 263, 5 },
		{ 263, 264, 5 },
		{ 264, 258, 5 },
		{ 264, 259, 5 },
		{ 264, 260, 5 },
		{ 264, 261, 5 },
		{ 264, 262, 5 },
		{ 264, 263, 5 },
		{ 264, 264, 5 },
		{ 243, 226, 6 },
		{ 243, 227, 6 },
		{ 243, 228, 6 },
		{ 243, 286, 6 },
		{ 243, 287, 6 },
		{ 243, 288, 6 },
		{ 244, 226, 6 },
		{ 244, 227, 6 },
		{ 244, 228, 6 },
		{ 244, 229, 6 },
		{ 244, 230, 6 },
		{ 244, 284, 6 },
		{ 244, 285, 6 },
		{ 244, 286, 6 },
		{ 244, 287, 6 },
		{ 244, 288, 6 },
		{ 245, 226, 6 },
		{ 245, 227, 6 },
		{ 245, 228, 6 },
		{ 245, 229, 6 },
		{ 245, 230, 6 },
		{ 245, 231, 6 },
		{ 245, 232, 6 },
		{ 245, 282, 6 },
		{ 245, 283, 6 },
		{ 245, 284, 6 },
		{ 245, 285, 6 },
		{ 245, 286, 6 },
		{ 245, 287, 6 },
		{ 245, 288, 6 },
		{ 246, 226, 6 },
		{ 246, 227, 6 },
		{ 246, 228, 6 },
		{ 246, 229, 6 },
		{ 246, 230, 6 },
		{ 246, 231, 6 },
		{ 246, 232, 6 },
		{ 246, 233, 6 },
		{ 246, 234, 6 },
		{ 246, 280, 6 },
		{ 246, 281, 6 },
		{ 246, 282, 6 },
		{ 246, 283, 6 },
		{ 246, 284, 6 },
		{ 246, 285, 6 },
		{ 246, 286, 6 },
		{ 246, 287, 6 },
		{ 246, 288, 6 },
		{ 247, 226, 6 },
		{ 247, 227, 6 },
		{ 247, 228, 6 },
		{ 247, 229, 6 },
		{ 247, 230, 6 },
		{ 247, 231, 6 },
		{ 247, 232, 6 },
		{ 247, 233, 6 },
		{ 247, 234, 6 },
		{ 247, 235, 6 },
		{ 247, 236, 6 },
		{ 247, 278, 6 },
		{ 247, 279, 6 },
		{ 247, 280, 6 },
		{ 247, 281, 6 },
		{ 247, 282, 6 },
		{ 247, 283, 6 },
		{ 247, 284, 6 },
		{ 247, 285, 6 },
		{ 247, 286, 6 },
		{ 247, 287, 6 },
		{ 247, 288, 6 },
		{ 248, 226, 6 },
		{ 248, 227, 6 },
		{ 248, 228, 6 },
		{ 248, 229, 6 },
		{ 248, 230, 6 },
		{ 248, 231, 6 },
		{ 248, 232, 6 },
		{ 248, 233, 6 },
		{ 248, 234, 6 },
		{ 248, 235, 6 },
		{ 248, 236, 6 },
		{ 248, 237, 6 },
		{ 248, 238, 6 },
		{ 248, 276, 6 },
		{ 248, 277, 6 },
		{ 248, 278, 6 },
		{ 248, 279, 6 },
		{ 248, 280, 6 },
		{ 248, 281, 6 },
		{ 248, 282, 6 },
		{ 248, 283, 6 },
		{ 248, 284, 6 },
		{ 248, 285, 6 },
		{ 248, 286, 6 },
		{ 248, 287, 6 },
		{ 248, 288, 6 },
		{ 249, 226, 6 },
		{ 249, 227, 6 },
		{ 249, 228, 6 },
		{ 249, 229, 6 },
		{ 249, 230, 6 },
		{ 249, 231, 6 },
		{ 249, 232, 6 },
		{ 249, 233, 6 },
		{ 249, 234, 6 },
		{ 249, 235, 6 },
		{ 249, 236, 6 },
		{ 249, 237, 6 },
		{ 249, 238, 6 },
		{ 249, 239, 6 },
		{ 249, 240, 6 },
		{ 249, 274, 6 },
		{ 249, 275, 6 },
		{ 249, 276, 6 },
		{ 249, 277, 6 },
		{ 249, 278, 6 },
		{ 249, 279, 6 },
		{ 249, 280, 6 },
		{ 249, 281, 6 },
		{ 249, 282, 6 },
		{ 249, 283, 6 },
		{ 249, 284, 6 },
		{ 249, 285, 6 },
		{ 249, 286, 6 },
		{ 249, 287, 6 },
		{ 249, 288, 6 },
		{ 250, 226, 6 },
		{ 250, 227, 6 },
		{ 250, 228, 6 },
		{ 250, 229, 6 },
		{ 250, 230, 6 },
		{ 250, 231, 6 },
		{ 250, 232, 6 },
		{ 250, 233, 6 },
		{ 250, 234, 6 },
		{ 250, 235, 6 },
		{ 250, 236, 6 },
		{ 250, 237, 6 },
		{ 250, 238, 6 },
		{ 250, 239, 6 },
		{ 250, 240, 6 },
		{ 250, 241, 6 },
		{ 250, 242, 6 },
		{ 250, 272, 6 },
		{ 250, 273, 6 },
		{ 250, 274, 6 },
		{ 250, 275, 6 },
		{ 250, 276, 6 },
		{ 250, 277, 6 },
		{ 250, 278, 6 },
		{ 250, 279, 6 },
		{ 250, 280, 6 },
		{ 250, 281, 6 },
		{ 250, 282, 6 },
		{ 250, 283, 6 },
		{ 250, 284, 6 },
		{ 250, 285, 6 },
		{ 250, 286, 6 },
		{ 250, 287, 6 },
		{ 250, 288, 6 },
		{ 251, 226, 6 },
		{ 251, 227, 6 },
		{ 251, 228, 6 },
		{ 251, 229, 6 },
		{ 251, 230, 6 },
		{ 251, 231, 6 },
		{ 251, 232, 6 },
		{ 251, 233, 6 },
		{ 251, 234, 6 },
		{ 251, 235, 6 },
		{ 251, 236, 6 },
		{ 251, 237, 6 },
		{ 251, 238, 6 },
		{ 251, 239, 6 },
		{ 251, 240, 6 },
		{ 251, 241, 6 },
		{ 251, 242, 6 },
		{ 251, 243, 6 },
		{ 251, 244, 6 },
		{ 251, 270, 6 },
		{ 251, 271, 6 },
		{ 251, 272, 6 },
		{ 251, 273, 6 },
		{ 251, 274, 6 },
		{ 251, 275, 6 },
		{ 251, 276, 6 },
		{ 251, 277, 6 },
		{ 251, 278, 6 },
		{ 251, 279, 6 },
		{ 251, 280, 6 },
		{ 251, 281, 6 },
		{ 251, 282, 6 },
		{ 251, 283, 6 },
		{ 251, 284, 6 },
		{ 251, 285, 6 },
		{ 251, 286, 6 },
		{ 251, 287, 6 },
		{ 251, 288, 6 },
		{ 252, 226, 6 },
		{ 252, 227, 6 },
		{ 252, 228, 6 },
		{ 252, 229, 6 },
		{ 252, 230, 6 },
		{ 252, 231, 6 },
		{ 252, 232, 6 },
		{ 252, 233, 6 },
		{ 252, 234, 6 },
		{ 252, 235, 6 },
		{ 252, 236, 6 },
		{ 252, 237, 6 },
		{ 252, 238, 6 },
		{ 252, 239, 6 },
		{ 252, 240, 6 },
		{ 252, 241, 6 },
		{ 252, 242, 6 },
		{ 252, 243, 6 },
		{ 252, 244, 6 },
		{ 252, 245, 6 },
		{ 252, 246, 6 },
		{ 252, 268, 6 },
		{ 252, 269, 6 },
		{ 252, 270, 6 },
		{ 252, 271, 6 },
		{ 252, 272, 6 },
		{ 252, 273, 6 },
		{ 252, 274, 6 },
		{ 252, 275, 6 },
		{ 252, 276, 6 },
		{ 252, 277, 6 },
		{ 252, 278, 6 },
		{ 252, 279, 6 },
		{ 252, 280, 6 },
		{ 252, 281, 6 },
		{ 252, 282, 6 },
		{ 252, 283, 6 },
		{ 252, 284, 6 },
		{ 252, 285, 6 },
		{ 252, 286, 6 },
		{ 252, 287, 6 },
		{ 252, 288, 6 },
		{ 253, 226, 6 },
		{ 253, 227, 6 },
		{ 253, 228, 6 },
		{ 253, 229, 6 },
		{ 253, 230, 6 },
		{ 253, 231, 6 },
		{ 253, 232, 6 },
		{ 253, 233, 6 },
		{ 253, 234, 6 },
		{ 253, 235, 6 },
		{ 253, 236, 6 },
		{ 253, 237, 6 },
		{ 253, 238, 6 },
		{ 253, 239, 6 },
		{ 253, 240, 6 },
		{ 253, 241, 6 },
		{ 253, 242, 6 },
		{ 253, 243, 6 },
		{ 253, 244, 6 },
		{ 253, 245, 6 },
		{ 253, 246, 6 },
		{ 253, 247, 6 },
		{ 253, 248, 6 },
		{ 253, 266, 6 },
		{ 253, 267, 6 },
		{ 253, 268, 6 },
		{ 253, 269, 6 },
		{ 253, 270, 6 },
		{ 253, 271, 6 },
		{ 253, 272, 6 },
		{ 253, 273, 6 },
		{ 253, 274, 6 },
		{ 253, 275, 6 },
		{ 253, 276, 6 },
		{ 253, 277, 6 },
		{ 253, 278, 6 },
		{ 253, 279, 6 },
		{ 253, 280, 6 },
		{ 253, 281, 6 },
		{ 253, 282, 6 },
		{ 253, 283, 6 },
		{ 253, 284, 6 },
		{ 253, 285, 6 },
		{ 253, 286, 6 },
		{ 253, 287, 6 },
		{ 253, 288, 6 },
		{ 254, 226, 6 },
		{ 254, 227, 6 },
		{ 254, 228, 6 },
		{ 254, 229, 6 },
		{ 254, 230, 6 },
		{ 254, 231, 6 },
		{ 254, 232, 6 },
		{ 254, 233, 6 },
		{ 254, 234, 6 },
		{ 254, 235, 6 },
		{ 254, 236, 6 },
		{ 254, 237, 6 },
		{ 254, 238, 6 },
		{ 254, 239, 6 },
		{ 254, 240, 6 },
		{ 254, 241, 6 },
		{ 254, 242, 6 },
		{ 254, 243, 6 },
		{ 254, 244, 6 },
		{ 254, 245, 6 },
		{ 254, 246, 6 },
		{ 254, 247, 6 },
		{ 254, 248, 6 },
		{ 254, 249, 6 },
		{ 254, 250, 6 },
		{ 254, 264, 6 },
		{ 254, 265, 6 },
		{ 254, 266, 6 },
		{ 254, 267, 6 },
		{ 254, 268, 6 },
		{ 254, 269, 6 },
		{ 254, 270, 6 },
		{ 254, 271, 6 },
		{ 254, 272, 6 },
		{ 254, 273, 6 },
		{ 254, 274, 6 },
		{ 254, 275, 6 },
		{ 254, 276, 6 },
		{ 254, 277, 6 },
		{ 254, 278, 6 },
		{ 254, 279, 6 },
		{ 254, 280, 6 },
		{ 254, 281, 6 },
		{ 254, 282, 6 },
		{ 254, 283, 6 },
		{ 254, 284, 6 },
		{ 254, 285, 6 },
		{ 254, 286, 6 },
		{ 254, 287, 6 },
		{ 254, 288, 6 },
		{ 255, 226, 6 },
		{ 255, 227, 6 },
		{ 255, 228, 6 },
		{ 255, 229, 6 },
		{ 255, 230, 6 },
		{ 255, 231, 6 },
		{ 255, 232, 6 },
		{ 255, 233, 6 },
		{ 255, 234, 6 },
		{ 255, 235, 6 },
		{ 255, 236, 6 },
		{ 255, 237, 6 },
		{ 255, 238, 6 },
		{ 255, 239, 6 },
		{ 255, 240, 6 },
		{ 255, 241, 6 },
		{ 255, 242, 6 },
		{ 255, 243, 6 },
		{ 255, 244, 6 },
		{ 255, 245, 6 },
		{ 255, 246, 6 },
		{ 255, 247, 6 },
		{ 255, 248, 6 },
		{ 255, 249, 6 },
		{ 255, 250, 6 },
		{ 255, 251, 6 },
		{ 255, 252, 6 },
		{ 255, 262, 6 },
		{ 255, 263, 6 },
		{ 255, 264, 6 },
		{ 255, 265, 6 },
		{ 255, 266, 6 },
		{ 255, 267, 6 },
		{ 255, 268, 6 },
		{ 255, 269, 6 },
		{ 255, 270, 6 },
		{ 255, 271, 6 },
		{ 255, 272, 6 },
		{ 255, 273, 6 },
		{ 255, 274, 6 },
		{ 255, 275, 6 },
		{ 255, 276, 6 },
		{ 255, 277, 6 },
		{ 255, 278, 6 },
		{ 255, 279, 6 },
		{ 255, 280, 6 },
		{ 255, 281, 6 },
		{ 255, 282, 6 },
		{ 255, 283, 6 },
		{ 255, 284, 6 },
		{ 255, 285, 6 },
		{ 255, 286, 6 },
		{ 255, 287, 6 },
		{ 255, 288, 6 },
		{ 256, 226, 6 },
		{ 256, 227, 6 },
		{ 256, 228, 6 },
		{ 256, 229, 6 },
		{ 256, 230, 6 },
		{ 256, 231, 6 },
		{ 256, 232, 6 },
		{ 256, 233, 6 },
		{ 256, 234, 6 },
		{ 256, 235, 6 },
		{ 256, 236, 6 },
		{ 256, 237, 6 },
		{ 256, 238, 6 },
		{ 256, 239, 6 },
		{ 256, 240, 6 },
		{ 256, 241, 6 },
		{ 256, 242, 6 },
		{ 256, 243, 6 },
		{ 256, 244, 6 },
		{ 256, 245, 6 },
		{ 256, 246, 6 },
		{ 256, 247, 6 },
		{ 256, 248, 6 },
		{ 256, 249, 6 },
		{ 256, 250, 6 },
		{ 256, 251, 6 },
		{ 256, 252, 6 },
		{ 256, 262, 6 },
		{ 256, 263, 6 },
		{ 256, 264, 6 },
		{ 256, 265, 6 },
		{ 256, 266, 6 },
		{ 256, 267, 6 },
		{ 256, 268, 6 },
		{ 256, 269, 6 },
		{ 256, 270, 6 },
		{ 256, 271, 6 },
		{ 256, 272, 6 },
		{ 256, 273, 6 },
		{ 256, 274, 6 },
		{ 256, 275, 6 },
		{ 256, 276, 6 },
		{ 256, 277, 6 },
		{ 256, 278, 6 },
		{ 256, 279, 6 },
		{ 256, 280, 6 },
		{ 256, 281, 6 },
		{ 256, 282, 6 },
		{ 256, 283, 6 },
		{ 256, 284, 6 },
		{ 256, 285, 6 },
		{ 256, 286, 6 },
		{ 256, 287, 6 },
		{ 256, 288, 6 },
		{ 257, 226, 6 },
		{ 257, 227, 6 },
		{ 257, 228, 6 },
		{ 257, 229, 6 },
		{ 257, 230, 6 },
		{ 257, 231, 6 },
		{ 257, 232, 6 },
		{ 257, 233, 6 },
		{ 257, 234, 6 },
		{ 257, 235, 6 },
		{ 257, 236, 6 },
		{ 257, 237, 6 },
		{ 257, 238, 6 },
		{ 257, 239, 6 },
		{ 257, 240, 6 },
		{ 257, 241, 6 },
		{ 257, 242, 6 },
		{ 257, 243, 6 },
		{ 257, 244, 6 },
		{ 257, 245, 6 },
		{ 257, 246, 6 },
		{ 257, 247, 6 },
		{ 257, 248, 6 },
		{ 257, 249, 6 },
		{ 257, 250, 6 },
		{ 257, 251, 6 },
		{ 257, 252, 6 },
		{ 257, 262, 6 },
		{ 257, 263, 6 },
		{ 257, 264, 6 },
		{ 257, 265, 6 },
		{ 257, 266, 6 },
		{ 257, 267, 6 },
		{ 257, 268, 6 },
		{ 257, 269, 6 },
		{ 257, 270, 6 },
		{ 257, 271, 6 },
		{ 257, 272, 6 },
		{ 257, 273, 6 },
		{ 257, 274, 6 },
		{ 257, 275, 6 },
		{ 257, 276, 6 },
		{ 257, 277, 6 },
		{ 257, 278, 6 },
		{ 257, 279, 6 },
		{ 257, 280, 6 },
		{ 257, 281, 6 },
		{ 257, 282, 6 },
		{ 257, 283, 6 },
		{ 257, 284, 6 },
		{ 257, 285, 6 },
		{ 257, 286, 6 },
		{ 257, 287, 6 },
		{ 257, 288, 6 },
		{ 258, 226, 6 },
		{ 258, 227, 6 },
		{ 258, 228, 6 },
		{ 258, 229, 6 },
		{ 258, 230, 6 },
		{ 258, 231, 6 },
		{ 258, 232, 6 },
		{ 258, 233, 6 },
		{ 258, 234, 6 },
		{ 258, 235, 6 },
		{ 258, 236, 6 },
		{ 258, 237, 6 },
		{ 258, 238, 6 },
		{ 258, 239, 6 },
		{ 258, 240, 6 },
		{ 258, 241, 6 },
		{ 258, 242, 6 },
		{ 258, 243, 6 },
		{ 258, 244, 6 },
		{ 258, 245, 6 },
		{ 258, 246, 6 },
		{ 258, 247, 6 },
		{ 258, 248, 6 },
		{ 258, 249, 6 },
		{ 258, 250, 6 },
		{ 258, 251, 6 },
		{ 258, 252, 6 },
		{ 258, 262, 6 },
		{ 258, 263, 6 },
		{ 258, 264, 6 },
		{ 258, 265, 6 },
		{ 258, 266, 6 },
		{ 258, 267, 6 },
		{ 258, 268, 6 },
		{ 258, 269, 6 },
		{ 258, 270, 6 },
		{ 258, 271, 6 },
		{ 258, 272, 6 },
		{ 258, 273, 6 },
		{ 258, 274, 6 },
		{ 258, 275, 6 },
		{ 258, 276, 6 },
		{ 258, 277, 6 },
		{ 258, 278, 6 },
		{ 258, 279, 6 },
		{ 258, 280, 6 },
		{ 258, 281, 6 },
		{ 258, 282, 6 },
		{ 258, 283, 6 },
		{ 258, 284, 6 },
		{ 258, 285, 6 },
		{ 258, 286, 6 },
		{ 258, 287, 6 },
		{ 258, 288, 6 },
		{ 259, 226, 6 },
		{ 259, 227, 6 },
		{ 259, 228, 6 },
		{ 259, 229, 6 },
		{ 259, 230, 6 },
		{ 259, 231, 6 },
		{ 259, 232, 6 },
		{ 259, 233, 6 },
		{ 259, 234, 6 },
		{ 259, 235, 6 },
		{ 259, 236, 6 },
		{ 259, 237, 6 },
		{ 259, 238, 6 },
		{ 259, 239, 6 },
		{ 259, 240, 6 },
		{ 259, 241, 6 },
		{ 259, 242, 6 },
		{ 259, 243, 6 },
		{ 259, 244, 6 },
		{ 259, 245, 6 },
		{ 259, 246, 6 },
		{ 259, 247, 6 },
		{ 259, 248, 6 },
		{ 259, 249, 6 },
		{ 259, 250, 6 },
		{ 259, 251, 6 },
		{ 259, 252, 6 },
		{ 259, 262, 6 },
		{ 259, 263, 6 },
		{ 259, 264, 6 },
		{ 259, 265, 6 },
		{ 259, 266, 6 },
		{ 259, 267, 6 },
		{ 259, 268, 6 },
		{ 259, 269, 6 },
		{ 259, 270, 6 },
		{ 259, 271, 6 },
		{ 259, 272, 6 },
		{ 259, 273, 6 },
		{ 259, 274, 6 },
		{ 259, 275, 6 },
		{ 259, 276, 6 },
		{ 259, 277, 6 },
		{ 259, 278, 6 },
		{ 259, 279, 6 },
		{ 259, 280, 6 },
		{ 259, 281, 6 },
		{ 259, 282, 6 },
		{ 259, 283, 6 },
		{ 259, 284, 6 },
		{ 259, 285, 6 },
		{ 259, 286, 6 },
		{ 259, 287, 6 },
		{ 259, 288, 6 },
		{ 260, 226, 6 },
		{ 260, 227, 6 },
		{ 260, 228, 6 },
		{ 260, 229, 6 },
		{ 260, 230, 6 },
		{ 260, 231, 6 },
		{ 260, 232, 6 },
		{ 260, 233, 6 },
		{ 260, 234, 6 },
		{ 260, 235, 6 },
		{ 260, 236, 6 },
		{ 260, 237, 6 },
		{ 260, 238, 6 },
		{ 260, 239, 6 },
		{ 260, 240, 6 },
		{ 260, 241, 6 },
		{ 260, 242, 6 },
		{ 260, 243, 6 },
		{ 260, 244, 6 },
		{ 260, 245, 6 },
		{ 260, 246, 6 },
		{ 260, 247, 6 },
		{ 260, 248, 6 },
		{ 260, 249, 6 },
		{ 260, 250, 6 },
		{ 260, 264, 6 },
		{ 260, 265, 6 },
		{ 260, 266, 6 },
		{ 260, 267, 6 },
		{ 260, 268, 6 },
		{ 260, 269, 6 },
		{ 260, 270, 6 },
		{ 260, 271, 6 },
		{ 260, 272, 6 },
		{ 260, 273, 6 },
		{ 260, 274, 6 },
		{ 260, 275, 6 },
		{ 260, 276, 6 },
		{ 260, 277, 6 },
		{ 260, 278, 6 },
		{ 260, 279, 6 },
		{ 260, 280, 6 },
		{ 260, 281, 6 },
		{ 260, 282, 6 },
		{ 260, 283, 6 },
		{ 260, 284, 6 },
		{ 260, 285, 6 },
		{ 260, 286, 6 },
		{ 260, 287, 6 },
		{ 260, 288, 6 },
		{ 261, 226, 6 },
		{ 261, 227, 6 },
		{ 261, 228, 6 },
		{ 261, 229, 6 },
		{ 261, 230, 6 },
		{ 261, 231, 6 },
		{ 261, 232, 6 },
		{ 261, 233, 6 },
		{ 261, 234, 6 },
		{ 261, 235, 6 },
		{ 261, 236, 6 },
		{ 261, 237, 6 },
		{ 261, 238, 6 },
		{ 261, 239, 6 },
		{ 261, 240, 6 },
		{ 261, 241, 6 },
		{ 261, 242, 6 },
		{ 261, 243, 6 },
		{ 261, 244, 6 },
		{ 261, 245, 6 },
		{ 261, 246, 6 },
		{ 261, 247, 6 },
		{ 261, 248, 6 },
		{ 261, 266, 6 },
		{ 261, 267, 6 },
		{ 261, 268, 6 },
		{ 261, 269, 6 },
		{ 261, 270, 6 },
		{ 261, 271, 6 },
		{ 261, 272, 6 },
		{ 261, 273, 6 },
		{ 261, 274, 6 },
		{ 261, 275, 6 },
		{ 261, 276, 6 },
		{ 261, 277, 6 },
		{ 261, 278, 6 },
		{ 261, 279, 6 },
		{ 261, 280, 6 },
		{ 261, 281, 6 },
		{ 261, 282, 6 },
		{ 261, 283, 6 },
		{ 261, 284, 6 },
		{ 261, 285, 6 },
		{ 261, 286, 6 },
		{ 261, 287, 6 },
		{ 261, 288, 6 },
		{ 262, 226, 6 },
		{ 262, 227, 6 },
		{ 262, 228, 6 },
		{ 262, 229, 6 },
		{ 262, 230, 6 },
		{ 262, 231, 6 },
		{ 262, 232, 6 },
		{ 262, 233, 6 },
		{ 262, 234, 6 },
		{ 262, 235, 6 },
		{ 262, 236, 6 },
		{ 262, 237, 6 },
		{ 262, 238, 6 },
		{ 262, 239, 6 },
		{ 262, 240, 6 },
		{ 262, 241, 6 },
		{ 262, 242, 6 },
		{ 262, 243, 6 },
		{ 262, 244, 6 },
		{ 262, 245, 6 },
		{ 262, 246, 6 },
		{ 262, 268, 6 },
		{ 262, 269, 6 },
		{ 262, 270, 6 },
		{ 262, 271, 6 },
		{ 262, 272, 6 },
		{ 262, 273, 6 },
		{ 262, 274, 6 },
		{ 262, 275, 6 },
		{ 262, 276, 6 },
		{ 262, 277, 6 },
		{ 262, 278, 6 },
		{ 262, 279, 6 },
		{ 262, 280, 6 },
		{ 262, 281, 6 },
		{ 262, 282, 6 },
		{ 262, 283, 6 },
		{ 262, 284, 6 },
		{ 262, 285, 6 },
		{ 262, 286, 6 },
		{ 262, 287, 6 },
		{ 262, 288, 6 },
		{ 263, 226, 6 },
		{ 263, 227, 6 },
		{ 263, 228, 6 },
		{ 263, 229, 6 },
		{ 263, 230, 6 },
		{ 263, 231, 6 },
		{ 263, 232, 6 },
		{ 263, 233, 6 },
		{ 263, 234, 6 },
		{ 263, 235, 6 },
		{ 263, 236, 6 },
		{ 263, 237, 6 },
		{ 263, 238, 6 },
		{ 263, 239, 6 },
		{ 263, 240, 6 },
		{ 263, 241, 6 },
		{ 263, 242, 6 },
		{ 263, 243, 6 },
		{ 263, 244, 6 },
		{ 263, 270, 6 },
		{ 263, 271, 6 },
		{ 263, 272, 6 },
		{ 263, 273, 6 },
		{ 263, 274, 6 },
		{ 263, 275, 6 },
		{ 263, 276, 6 },
		{ 263, 277, 6 },
		{ 263, 278, 6 },
		{ 263, 279, 6 },
		{ 263, 280, 6 },
		{ 263, 281, 6 },
		{ 263, 282, 6 },
		{ 263, 283, 6 },
		{ 263, 284, 6 },
		{ 263, 285, 6 },
		{ 263, 286, 6 },
		{ 263, 287, 6 },
		{ 263, 288, 6 },
		{ 264, 226, 6 },
		{ 264, 227, 6 },
		{ 264, 228, 6 },
		{ 264, 229, 6 },
		{ 264, 230, 6 },
		{ 264, 231, 6 },
		{ 264, 232, 6 },
		{ 264, 233, 6 },
		{ 264, 234, 6 },
		{ 264, 235, 6 },
		{ 264, 236, 6 },
		{ 264, 237, 6 },
		{ 264, 238, 6 },
		{ 264, 239, 6 },
		{ 264, 240, 6 },
		{ 264, 241, 6 },
		{ 264, 242, 6 },
		{ 264, 272, 6 },
		{ 264, 273, 6 },
		{ 264, 274, 6 },
		{ 264, 275, 6 },
		{ 264, 276, 6 },
		{ 264, 277, 6 },
		{ 264, 278, 6 },
		{ 264, 279, 6 },
		{ 264, 280, 6 },
		{ 264, 281, 6 },
		{ 264, 282, 6 },
		{ 264, 283, 6 },
		{ 264, 284, 6 },
		{ 264, 285, 6 },
		{ 264, 286, 6 },
		{ 264, 287, 6 },
		{ 264, 288, 6 },
		{ 265, 226, 6 },
		{ 265, 227, 6 },
		{ 265, 228, 6 },
		{ 265, 229, 6 },
		{ 265, 230, 6 },
		{ 265, 231, 6 },
		{ 265, 232, 6 },
		{ 265, 233, 6 },
		{ 265, 234, 6 },
		{ 265, 235, 6 },
		{ 265, 236, 6 },
		{ 265, 237, 6 },
		{ 265, 238, 6 },
		{ 265, 239, 6 },
		{ 265, 240, 6 },
		{ 265, 274, 6 },
		{ 265, 275, 6 },
		{ 265, 276, 6 },
		{ 265, 277, 6 },
		{ 265, 278, 6 },
		{ 265, 279, 6 },
		{ 265, 280, 6 },
		{ 265, 281, 6 },
		{ 265, 282, 6 },
		{ 265, 283, 6 },
		{ 265, 284, 6 },
		{ 265, 285, 6 },
		{ 265, 286, 6 },
		{ 265, 287, 6 },
		{ 265, 288, 6 },
		{ 266, 226, 6 },
		{ 266, 227, 6 },
		{ 266, 228, 6 },
		{ 266, 229, 6 },
		{ 266, 230, 6 },
		{ 266, 231, 6 },
		{ 266, 232, 6 },
		{ 266, 233, 6 },
		{ 266, 234, 6 },
		{ 266, 235, 6 },
		{ 266, 236, 6 },
		{ 266, 237, 6 },
		{ 266, 238, 6 },
		{ 266, 276, 6 },
		{ 266, 277, 6 },
		{ 266, 278, 6 },
		{ 266, 279, 6 },
		{ 266, 280, 6 },
		{ 266, 281, 6 },
		{ 266, 282, 6 },
		{ 266, 283, 6 },
		{ 266, 284, 6 },
		{ 266, 285, 6 },
		{ 266, 286, 6 },
		{ 266, 287, 6 },
		{ 266, 288, 6 },
		{ 267, 226, 6 },
		{ 267, 227, 6 },
		{ 267, 228, 6 },
		{ 267, 229, 6 },
		{ 267, 230, 6 },
		{ 267, 231, 6 },
		{ 267, 232, 6 },
		{ 267, 233, 6 },
		{ 267, 234, 6 },
		{ 267, 235, 6 },
		{ 267, 236, 6 },
		{ 267, 278, 6 },
		{ 267, 279, 6 },
		{ 267, 280, 6 },
		{ 267, 281, 6 },
		{ 267, 282, 6 },
		{ 267, 283, 6 },
		{ 267, 284, 6 },
		{ 267, 285, 6 },
		{ 267, 286, 6 },
		{ 267, 287, 6 },
		{ 267, 288, 6 },
		{ 268, 226, 6 },
		{ 268, 227, 6 },
		{ 268, 228, 6 },
		{ 268, 229, 6 },
		{ 268, 230, 6 },
		{ 268, 231, 6 },
		{ 268, 232, 6 },
		{ 268, 233, 6 },
		{ 268, 234, 6 },
		{ 268, 280, 6 },
		{ 268, 281, 6 },
		{ 268, 282, 6 },
		{ 268, 283, 6 },
		{ 268, 284, 6 },
		{ 268, 285, 6 },
		{ 268, 286, 6 },
		{ 268, 287, 6 },
		{ 268, 288, 6 },
		{ 269, 226, 6 },
		{ 269, 227, 6 },
		{ 269, 228, 6 },
		{ 269, 229, 6 },
		{ 269, 230, 6 },
		{ 269, 231, 6 },
		{ 269, 232, 6 },
		{ 269, 282, 6 },
		{ 269, 283, 6 },
		{ 269, 284, 6 },
		{ 269, 285, 6 },
		{ 269, 286, 6 },
		{ 269, 287, 6 },
		{ 269, 288, 6 },
		{ 270, 226, 6 },
		{ 270, 227, 6 },
		{ 270, 228, 6 },
		{ 270, 229, 6 },
		{ 270, 230, 6 },
		{ 270, 284, 6 },
		{ 270, 285, 6 },
		{ 270, 286, 6 },
		{ 270, 287, 6 },
		{ 270, 288, 6 },
		{ 271, 226, 6 },
		{ 271, 227, 6 },
		{ 271, 228, 6 },
		{ 271, 286, 6 },
		{ 271, 287, 6 },
		{ 271, 288, 6 },
		{ 227, 288, 7 },
		{ 228, 287, 7 },
		{ 228, 288, 7 },
		{ 229, 286, 7 },
		{ 229, 287, 7 },
		{ 229, 288, 7 },
		{ 230, 285, 7 },
		{ 230, 286, 7 },
		{ 230, 287, 7 },
		{ 230, 288, 7 },
		{ 231, 284, 7 },
		{ 231, 285, 7 },
		{ 231, 286, 7 },
		{ 231, 287, 7 },
		{ 231, 288, 7 },
		{ 232, 283, 7 },
		{ 232, 284, 7 },
		{ 232, 285, 7 },
		{ 232, 286, 7 },
		{ 232, 287, 7 },
		{ 232, 288, 7 },
		{ 233, 282, 7 },
		{ 233, 283, 7 },
		{ 233, 284, 7 },
		{ 233, 285, 7 },
		{ 233, 286, 7 },
		{ 233, 287, 7 },
		{ 233, 288, 7 },
		{ 234, 281, 7 },
		{ 234, 282, 7 },
		{ 234, 283, 7 },
		{ 234, 284, 7 },
		{ 234, 285, 7 },
		{ 234, 286, 7 },
		{ 234, 287, 7 },
		{ 234, 288, 7 },
		{ 235, 280, 7 },
		{ 235, 281, 7 },
		{ 235, 282, 7 },
		{ 235, 283, 7 },
		{ 235, 284, 7 },
		{ 235, 285, 7 },
		{ 235, 286, 7 },
		{ 235, 287, 7 },
		{ 235, 288, 7 },
		{ 236, 279, 7 },
		{ 236, 280, 7 },
		{ 236, 281, 7 },
		{ 236, 282, 7 },
		{ 236, 283, 7 },
		{ 236, 284, 7 },
		{ 236, 285, 7 },
		{ 236, 286, 7 },
		{ 236, 287, 7 },
		{ 236, 288, 7 },
		{ 237, 278, 7 },
		{ 237, 279, 7 },
		{ 237, 280, 7 },
		{ 237, 281, 7 },
		{ 237, 282, 7 },
		{ 237, 283, 7 },
		{ 237, 284, 7 },
		{ 237, 285, 7 },
		{ 237, 286, 7 },
		{ 237, 287, 7 },
		{ 237, 288, 7 },
		{ 238, 277, 7 },
		{ 238, 278, 7 },
		{ 238, 279, 7 },
		{ 238, 280, 7 },
		{ 238, 281, 7 },
		{ 238, 282, 7 },
		{ 238, 283, 7 },
		{ 238, 284, 7 },
		{ 238, 285, 7 },
		{ 238, 286, 7 },
		{ 238, 287, 7 },
		{ 238, 288, 7 },
		{ 239, 276, 7 },
		{ 239, 277, 7 },
		{ 239, 278, 7 },
		{ 239, 279, 7 },
		{ 239, 280, 7 },
		{ 239, 281, 7 },
		{ 239, 282, 7 },
		{ 239, 283, 7 },
		{ 239, 284, 7 },
		{ 239, 285, 7 },
		{ 239, 286, 7 },
		{ 239, 287, 7 },
		{ 239, 288, 7 },
		{ 240, 275, 7 },
		{ 240, 276, 7 },
		{ 240, 277, 7 },
		{ 240, 278, 7 },
		{ 240, 279, 7 },
		{ 240, 280, 7 },
		{ 240, 281, 7 },
		{ 240, 282, 7 },
		{ 240, 283, 7 },
		{ 240, 284, 7 },
		{ 240, 285, 7 },
		{ 240, 286, 7 },
		{ 240, 287, 7 },
		{ 240, 288, 7 },
		{ 241, 274, 7 },
		{ 241, 275, 7 },
		{ 241, 276, 7 },
		{ 241, 277, 7 },
		{ 241, 278, 7 },
		{ 241, 279, 7 },
		{ 241, 280, 7 },
		{ 241, 281, 7 },
		{ 241, 282, 7 },
		{ 241, 283, 7 },
		{ 241, 284, 7 },
		{ 241, 285, 7 },
		{ 241, 286, 7 },
		{ 241, 287, 7 },
		{ 241, 288, 7 },
		{ 242, 273, 7 },
		{ 242, 274, 7 },
		{ 242, 275, 7 },
		{ 242, 276, 7 },
		{ 242, 277, 7 },
		{ 242, 278, 7 },
		{ 242, 279, 7 },
		{ 242, 280, 7 },
		{ 242, 281, 7 },
		{ 242, 282, 7 },
		{ 242, 283, 7 },
		{ 242, 284, 7 },
		{ 242, 285, 7 },
		{ 242, 286, 7 },
		{ 242, 287, 7 },
		{ 242, 288, 7 },
		{ 243, 272, 7 },
		{ 243, 273, 7 },
		{ 243, 274, 7 },
		{ 243, 275, 7 },
		{ 243, 276, 7 },
		{ 243, 277, 7 },
		{ 243, 278, 7 },
		{ 243, 279, 7 },
		{ 243, 280, 7 },
		{ 243, 281, 7 },
		{ 243, 282, 7 },
		{ 243, 283, 7 },
		{ 243, 284, 7 },
		{ 243, 285, 7 },
		{ 243, 286, 7 },
		{ 243, 287, 7 },
		{ 243, 288, 7 },
		{ 244, 271, 7 },
		{ 244, 272, 7 },
		{ 244, 273, 7 },
		{ 244, 274, 7 },
		{ 244, 275, 7 },
		{ 244, 276, 7 },
		{ 244, 277, 7 },
		{ 244, 278, 7 },
		{ 244, 279, 7 },
		{ 244, 280, 7 },
		{ 244, 281, 7 },
		{ 244, 282, 7 },
		{ 244, 283, 7 },
		{ 244, 284, 7 },
		{ 244, 285, 7 },
		{ 244, 286, 7 },
		{ 244, 287, 7 },
		{ 244, 288, 7 },
		{ 245, 270, 7 },
		{ 245, 271, 7 },
		{ 245, 272, 7 },
		{ 245, 273, 7 },
		{ 245, 274, 7 },
		{ 245, 275, 7 },
		{ 245, 276, 7 },
		{ 245, 277, 7 },
		{ 245, 278, 7 },
		{ 245, 279, 7 },
		{ 245, 280, 7 },
		{ 245, 281, 7 },
		{ 245, 282, 7 },
		{ 245, 283, 7 },
		{ 245, 284, 7 },
		{ 245, 285, 7 },
		{ 245, 286, 7 },
		{ 245, 287, 7 },
		{ 245, 288, 7 },
		{ 246, 269, 7 },
		{ 246, 270, 7 },
		{ 246, 271, 7 },
		{ 246, 272, 7 },
		{ 246, 273, 7 },
		{ 246, 274, 7 },
		{ 246, 275, 7 },
		{ 246, 276, 7 },
		{ 246, 277, 7 },
		{ 246, 278, 7 },
		{ 246, 279, 7 },
		{ 246, 280, 7 },
		{ 246, 281, 7 },
		{ 246, 282, 7 },
		{ 246, 283, 7 },
		{ 246, 284, 7 },
		{ 246, 285, 7 },
		{ 246, 286, 7 },
		{ 246, 287, 7 },
		{ 246, 288, 7 },
		{ 247, 268, 7 },
		{ 247, 269, 7 },
		{ 247, 270, 7 },
		{ 247, 271, 7 },
		{ 247, 272, 7 },
		{ 247, 273, 7 },
		{ 247, 274, 7 },
		{ 247, 275, 7 },
		{ 247, 276, 7 },
		{ 247, 277, 7 },
		{ 247, 278, 7 },
		{ 247, 279, 7 },
		{ 247, 280, 7 },
		{ 247, 281, 7 },
		{ 247, 282, 7 },
		{ 247, 283, 7 },
		{ 247, 284, 7 },
		{ 247, 285, 7 },
		{ 247, 286, 7 },
		{ 247, 287, 7 },
		{ 247, 288, 7 },
		{ 248, 267, 7 },
		{ 248, 268, 7 },
		{ 248, 269, 7 },
		{ 248, 270, 7 },
		{ 248, 271, 7 },
		{ 248, 272, 7 },
		{ 248, 273, 7 },
		{ 248, 274, 7 },
		{ 248, 275, 7 },
		{ 248, 276, 7 },
		{ 248, 277, 7 },
		{ 248, 278, 7 },
		{ 248, 279, 7 },
		{ 248, 280, 7 },
		{ 248, 281, 7 },
		{ 248, 282, 7 },
		{ 248, 283, 7 },
		{ 248, 284, 7 },
		{ 248, 285, 7 },
		{ 248, 286, 7 },
		{ 248, 287, 7 },
		{ 248, 288, 7 },
		{ 249, 266, 7 },
		{ 249, 267, 7 },
		{ 249, 268, 7 },
		{ 249, 269, 7 },
		{ 249, 270, 7 },
		{ 249, 271, 7 },
		{ 249, 272, 7 },
		{ 249, 273, 7 },
		{ 249, 274, 7 },
		{ 249, 275, 7 },
		{ 249, 276, 7 },
		{ 249, 277, 7 },
		{ 249, 278, 7 },
		{ 249, 279, 7 },
		{ 249, 280, 7 },
		{ 249, 281, 7 },
		{ 249, 282, 7 },
		{ 249, 283, 7 },
		{ 249, 284, 7 },
		{ 249, 285, 7 },
		{ 249, 286, 7 },
		{ 249, 287, 7 },
		{ 249, 288, 7 },
		{ 250, 265, 7 },
		{ 250, 266, 7 },
		{ 250, 267, 7 },
		{ 250, 268, 7 },
		{ 250, 269, 7 },
		{ 250, 270, 7 },
		{ 250, 271, 7 },
		{ 250, 272, 7 },
		{ 250, 273, 7 },
		{ 250, 274, 7 },
		{ 250, 275, 7 },
		{ 250, 276, 7 },
		{ 250, 277, 7 },
		{ 250, 278, 7 },
		{ 250, 279, 7 },
		{ 250, 280, 7 },
		{ 250, 281, 7 },
		{ 250, 282, 7 },
		{ 250, 283, 7 },
		{ 250, 284, 7 },
		{ 250, 285, 7 },
		{ 250, 286, 7 },
		{ 250, 287, 7 },
		{ 250, 288, 7 },
		{ 251, 264, 7 },
		{ 251, 265, 7 },
		{ 251, 266, 7 },
		{ 251, 267, 7 },
		{ 251, 268, 7 },
		{ 251, 269, 7 },
		{ 251, 270, 7 },
		{ 251, 271, 7 },
		{ 251, 272, 7 },
		{ 251, 273, 7 },
		{ 251, 274, 7 },
		{ 251, 275, 7 },
		{ 251, 276, 7 },
		{ 251, 277, 7 },
		{ 251, 278, 7 },
		{ 251, 279, 7 },
		{ 251, 280, 7 },
		{ 251, 281, 7 },
		{ 251, 282, 7 },
		{ 251, 283, 7 },
		{ 251, 284, 7 },
		{ 251, 285, 7 },
		{ 251, 286, 7 },
		{ 251, 287, 7 },
		{ 251, 288, 7 },
		{ 252, 263, 7 },
		{ 252, 264, 7 },
		{ 252, 265, 7 },
		{ 252, 266, 7 },
		{ 252, 267, 7 },
		{ 252, 268, 7 },
		{ 252, 269, 7 },
		{ 252, 270, 7 },
		{ 252, 271, 7 },
		{ 252, 272, 7 },
		{ 252, 273, 7 },
		{ 252, 274, 7 },
		{ 252, 275, 7 },
		{ 252, 276, 7 },
		{ 252, 277, 7 },
		{ 252, 278, 7 },
		{ 252, 279, 7 },
		{ 252, 280, 7 },
		{ 252, 281, 7 },
		{ 252, 282, 7 },
		{ 252, 283, 7 },
		{ 252, 284, 7 },
		{ 252, 285, 7 },
		{ 252, 286, 7 },
		{ 252, 287, 7 },
		{ 252, 288, 7 },
		{ 253, 262, 7 },
		{ 253, 263, 7 },
		{ 253, 264, 7 },
		{ 253, 265, 7 },
		{ 253, 266, 7 },
		{ 253, 267, 7 },
		{ 253, 268, 7 },
		{ 253, 269, 7 },
		{ 253, 270, 7 },
		{ 253, 271, 7 },
		{ 253, 272, 7 },
		{ 253, 273, 7 },
		{ 253, 274, 7 },
		{ 253, 275, 7 },
		{ 253, 276, 7 },
		{ 253, 277, 7 },
		{ 253, 278, 7 },
		{ 253, 279, 7 },
		{ 253, 280, 7 },
		{ 253, 281, 7 },
		{ 253, 282, 7 },
		{ 253, 283, 7 },
		{ 253, 284, 7 },
		{ 253, 285, 7 },
		{ 253, 286, 7 },
		{ 253, 287, 7 },
		{ 253, 288, 7 },
		{ 254, 262, 7 },
		{ 254, 263, 7 },
		{ 254, 264, 7 },
		{ 254, 265, 7 },
		{ 254, 266, 7 },
		{ 254, 267, 7 },
		{ 254, 268, 7 },
		{ 254, 269, 7 },
		{ 254, 270, 7 },
		{ 254, 271, 7 },
		{ 254, 272, 7 },
		{ 254, 273, 7 },
		{ 254, 274, 7 },
		{ 254, 275, 7 },
		{ 254, 276, 7 },
		{ 254, 277, 7 },
		{ 254, 278, 7 },
		{ 254, 279, 7 },
		{ 254, 280, 7 },
		{ 254, 281, 7 },
		{ 254, 282, 7 },
		{ 254, 283, 7 },
		{ 254, 284, 7 },
		{ 254, 285, 7 },
		{ 254, 286, 7 },
		{ 254, 287, 7 },
		{ 254, 288, 7 },
		{ 255, 262, 7 },
		{ 255, 263, 7 },
		{ 255, 264, 7 },
		{ 255, 265, 7 },
		{ 255, 266, 7 },
		{ 255, 267, 7 },
		{ 255, 268, 7 },
		{ 255, 269, 7 },
		{ 255, 270, 7 },
		{ 255, 271, 7 },
		{ 255, 272, 7 },
		{ 255, 273, 7 },
		{ 255, 274, 7 },
		{ 255, 275, 7 },
		{ 255, 276, 7 },
		{ 255, 277, 7 },
		{ 255, 278, 7 },
		{ 255, 279, 7 },
		{ 255, 280, 7 },
		{ 255, 281, 7 },
		{ 255, 282, 7 },
		{ 255, 283, 7 },
		{ 255, 284, 7 },
		{ 255, 285, 7 },
		{ 255, 286, 7 },
		{ 255, 287, 7 },
		{ 255, 288, 7 },
		{ 256, 262, 7 },
		{ 256, 263, 7 },
		{ 256, 264, 7 },
		{ 256, 265, 7 },
		{ 256, 266, 7 },
		{ 256, 267, 7 },
		{ 256, 268, 7 },
		{ 256, 269, 7 },
		{ 256, 270, 7 },
		{ 256, 271, 7 },
		{ 256, 272, 7 },
		{ 256, 273, 7 },
		{ 256, 274, 7 },
		{ 256, 275, 7 },
		{ 256, 276, 7 },
		{ 256, 277, 7 },
		{ 256, 278, 7 },
		{ 256, 279, 7 },
		{ 256, 280, 7 },
		{ 256, 281, 7 },
		{ 256, 282, 7 },
		{ 256, 283, 7 },
		{ 256, 284, 7 },
		{ 256, 285, 7 },
		{ 256, 286, 7 },
		{ 256, 287, 7 },
		{ 256, 288, 7 },
		{ 258, 226, 7 },
		{ 258, 227, 7 },
		{ 258, 228, 7 },
		{ 258, 229, 7 },
		{ 258, 230, 7 },
		{ 258, 231, 7 },
		{ 258, 232, 7 },
		{ 258, 233, 7 },
		{ 258, 234, 7 },
		{ 258, 235, 7 },
		{ 258, 236, 7 },
		{ 258, 237, 7 },
		{ 258, 238, 7 },
		{ 258, 239, 7 },
		{ 258, 240, 7 },
		{ 258, 241, 7 },
		{ 258, 242, 7 },
		{ 258, 243, 7 },
		{ 258, 244, 7 },
		{ 258, 245, 7 },
		{ 258, 246, 7 },
		{ 258, 247, 7 },
		{ 258, 248, 7 },
		{ 258, 249, 7 },
		{ 258, 250, 7 },
		{ 258, 251, 7 },
		{ 258, 252, 7 },
		{ 259, 226, 7 },
		{ 259, 227, 7 },
		{ 259, 228, 7 },
		{ 259, 229, 7 },
		{ 259, 230, 7 },
		{ 259, 231, 7 },
		{ 259, 232, 7 },
		{ 259, 233, 7 },
		{ 259, 234, 7 },
		{ 259, 235, 7 },
		{ 259, 236, 7 },
		{ 259, 237, 7 },
		{ 259, 238, 7 },
		{ 259, 239, 7 },
		{ 259, 240, 7 },
		{ 259, 241, 7 },
		{ 259, 242, 7 },
		{ 259, 243, 7 },
		{ 259, 244, 7 },
		{ 259, 245, 7 },
		{ 259, 246, 7 },
		{ 259, 247, 7 },
		{ 259, 248, 7 },
		{ 259, 249, 7 },
		{ 259, 250, 7 },
		{ 259, 251, 7 },
		{ 259, 252, 7 },
		{ 260, 226, 7 },
		{ 260, 227, 7 },
		{ 260, 228, 7 },
		{ 260, 229, 7 },
		{ 260, 230, 7 },
		{ 260, 231, 7 },
		{ 260, 232, 7 },
		{ 260, 233, 7 },
		{ 260, 234, 7 },
		{ 260, 235, 7 },
		{ 260, 236, 7 },
		{ 260, 237, 7 },
		{ 260, 238, 7 },
		{ 260, 239, 7 },
		{ 260, 240, 7 },
		{ 260, 241, 7 },
		{ 260, 242, 7 },
		{ 260, 243, 7 },
		{ 260, 244, 7 },
		{ 260, 245, 7 },
		{ 260, 246, 7 },
		{ 260, 247, 7 },
		{ 260, 248, 7 },
		{ 260, 249, 7 },
		{ 260, 250, 7 },
		{ 260, 251, 7 },
		{ 260, 252, 7 },
		{ 261, 226, 7 },
		{ 261, 227, 7 },
		{ 261, 228, 7 },
		{ 261, 229, 7 },
		{ 261, 230, 7 },
		{ 261, 231, 7 },
		{ 261, 232, 7 },
		{ 261, 233, 7 },
		{ 261, 234, 7 },
		{ 261, 235, 7 },
		{ 261, 236, 7 },
		{ 261, 237, 7 },
		{ 261, 238, 7 },
		{ 261, 239, 7 },
		{ 261, 240, 7 },
		{ 261, 241, 7 },
		{ 261, 242, 7 },
		{ 261, 243, 7 },
		{ 261, 244, 7 },
		{ 261, 245, 7 },
		{ 261, 246, 7 },
		{ 261, 247, 7 },
		{ 261, 248, 7 },
		{ 261, 249, 7 },
		{ 261, 250, 7 },
		{ 261, 251, 7 },
		{ 261, 252, 7 },
		{ 262, 226, 7 },
		{ 262, 227, 7 },
		{ 262, 228, 7 },
		{ 262, 229, 7 },
		{ 262, 230, 7 },
		{ 262, 231, 7 },
		{ 262, 232, 7 },
		{ 262, 233, 7 },
		{ 262, 234, 7 },
		{ 262, 235, 7 },
		{ 262, 236, 7 },
		{ 262, 237, 7 },
		{ 262, 238, 7 },
		{ 262, 239, 7 },
		{ 262, 240, 7 },
		{ 262, 241, 7 },
		{ 262, 242, 7 },
		{ 262, 243, 7 },
		{ 262, 244, 7 },
		{ 262, 245, 7 },
		{ 262, 246, 7 },
		{ 262, 247, 7 },
		{ 262, 248, 7 },
		{ 262, 249, 7 },
		{ 262, 250, 7 },
		{ 262, 251, 7 },
		{ 263, 226, 7 },
		{ 263, 227, 7 },
		{ 263, 228, 7 },
		{ 263, 229, 7 },
		{ 263, 230, 7 },
		{ 263, 231, 7 },
		{ 263, 232, 7 },
		{ 263, 233, 7 },
		{ 263, 234, 7 },
		{ 263, 235, 7 },
		{ 263, 236, 7 },
		{ 263, 237, 7 },
		{ 263, 238, 7 },
		{ 263, 239, 7 },
		{ 263, 240, 7 },
		{ 263, 241, 7 },
		{ 263, 242, 7 },
		{ 263, 243, 7 },
		{ 263, 244, 7 },
		{ 263, 245, 7 },
		{ 263, 246, 7 },
		{ 263, 247, 7 },
		{ 263, 248, 7 },
		{ 263, 249, 7 },
		{ 263, 250, 7 },
		{ 264, 226, 7 },
		{ 264, 227, 7 },
		{ 264, 228, 7 },
		{ 264, 229, 7 },
		{ 264, 230, 7 },
		{ 264, 231, 7 },
		{ 264, 232, 7 },
		{ 264, 233, 7 },
		{ 264, 234, 7 },
		{ 264, 235, 7 },
		{ 264, 236, 7 },
		{ 264, 237, 7 },
		{ 264, 238, 7 },
		{ 264, 239, 7 },
		{ 264, 240, 7 },
		{ 264, 241, 7 },
		{ 264, 242, 7 },
		{ 264, 243, 7 },
		{ 264, 244, 7 },
		{ 264, 245, 7 },
		{ 264, 246, 7 },
		{ 264, 247, 7 },
		{ 264, 248, 7 },
		{ 264, 249, 7 },
		{ 265, 226, 7 },
		{ 265, 227, 7 },
		{ 265, 228, 7 },
		{ 265, 229, 7 },
		{ 265, 230, 7 },
		{ 265, 231, 7 },
		{ 265, 232, 7 },
		{ 265, 233, 7 },
		{ 265, 234, 7 },
		{ 265, 235, 7 },
		{ 265, 236, 7 },
		{ 265, 237, 7 },
		{ 265, 238, 7 },
		{ 265, 239, 7 },
		{ 265, 240, 7 },
		{ 265, 241, 7 },
		{ 265, 242, 7 },
		{ 265, 243, 7 },
		{ 265, 244, 7 },
		{ 265, 245, 7 },
		{ 265, 246, 7 },
		{ 265, 247, 7 },
		{ 265, 248, 7 },
		{ 266, 226, 7 },
		{ 266, 227, 7 },
		{ 266, 228, 7 },
		{ 266, 229, 7 },
		{ 266, 230, 7 },
		{ 266, 231, 7 },
		{ 266, 232, 7 },
		{ 266, 233, 7 },
		{ 266, 234, 7 },
		{ 266, 235, 7 },
		{ 266, 236, 7 },
		{ 266, 237, 7 },
		{ 266, 238, 7 },
		{ 266, 239, 7 },
		{ 266, 240, 7 },
		{ 266, 241, 7 },
		{ 266, 242, 7 },
		{ 266, 243, 7 },
		{ 266, 244, 7 },
		{ 266, 245, 7 },
		{ 266, 246, 7 },
		{ 266, 247, 7 },
		{ 267, 226, 7 },
		{ 267, 227, 7 },
		{ 267, 228, 7 },
		{ 267, 229, 7 },
		{ 267, 230, 7 },
		{ 267, 231, 7 },
		{ 267, 232, 7 },
		{ 267, 233, 7 },
		{ 267, 234, 7 },
		{ 267, 235, 7 },
		{ 267, 236, 7 },
		{ 267, 237, 7 },
		{ 267, 238, 7 },
		{ 267, 239, 7 },
		{ 267, 240, 7 },
		{ 267, 241, 7 },
		{ 267, 242, 7 },
		{ 267, 243, 7 },
		{ 267, 244, 7 },
		{ 267, 245, 7 },
		{ 267, 246, 7 },
		{ 268, 226, 7 },
		{ 268, 227, 7 },
		{ 268, 228, 7 },
		{ 268, 229, 7 },
		{ 268, 230, 7 },
		{ 268, 231, 7 },
		{ 268, 232, 7 },
		{ 268, 233, 7 },
		{ 268, 234, 7 },
		{ 268, 235, 7 },
		{ 268, 236, 7 },
		{ 268, 237, 7 },
		{ 268, 238, 7 },
		{ 268, 239, 7 },
		{ 268, 240, 7 },
		{ 268, 241, 7 },
		{ 268, 242, 7 },
		{ 268, 243, 7 },
		{ 268, 244, 7 },
		{ 268, 245, 7 },
		{ 269, 226, 7 },
		{ 269, 227, 7 },
		{ 269, 228, 7 },
		{ 269, 229, 7 },
		{ 269, 230, 7 },
		{ 269, 231, 7 },
		{ 269, 232, 7 },
		{ 269, 233, 7 },
		{ 269, 234, 7 },
		{ 269, 235, 7 },
		{ 269, 236, 7 },
		{ 269, 237, 7 },
		{ 269, 238, 7 },
		{ 269, 239, 7 },
		{ 269, 240, 7 },
		{ 269, 241, 7 },
		{ 269, 242, 7 },
		{ 269, 243, 7 },
		{ 269, 244, 7 },
		{ 270, 226, 7 },
		{ 270, 227, 7 },
		{ 270, 228, 7 },
		{ 270, 229, 7 },
		{ 270, 230, 7 },
		{ 270, 231, 7 },
		{ 270, 232, 7 },
		{ 270, 233, 7 },
		{ 270, 234, 7 },
		{ 270, 235, 7 },
		{ 270, 236, 7 },
		{ 270, 237, 7 },
		{ 270, 238, 7 },
		{ 270, 239, 7 },
		{ 270, 240, 7 },
		{ 270, 241, 7 },
		{ 270, 242, 7 },
		{ 270, 243, 7 },
		{ 271, 226, 7 },
		{ 271, 227, 7 },
		{ 271, 228, 7 },
		{ 271, 229, 7 },
		{ 271, 230, 7 },
		{ 271, 231, 7 },
		{ 271, 232, 7 },
		{ 271, 233, 7 },
		{ 271, 234, 7 },
		{ 271, 235, 7 },
		{ 271, 236, 7 },
		{ 271, 237, 7 },
		{ 271, 238, 7 },
		{ 271, 239, 7 },
		{ 271, 240, 7 },
		{ 271, 241, 7 },
		{ 271, 242, 7 },
		{ 272, 226, 7 },
		{ 272, 227, 7 },
		{ 272, 228, 7 },
		{ 272, 229, 7 },
		{ 272, 230, 7 },
		{ 272, 231, 7 },
		{ 272, 232, 7 },
		{ 272, 233, 7 },
		{ 272, 234, 7 },
		{ 272, 235, 7 },
		{ 272, 236, 7 },
		{ 272, 237, 7 },
		{ 272, 238, 7 },
		{ 272, 239, 7 },
		{ 272, 240, 7 },
		{ 272, 241, 7 },
		{ 273, 226, 7 },
		{ 273, 227, 7 },
		{ 273, 228, 7 },
		{ 273, 229, 7 },
		{ 273, 230, 7 },
		{ 273, 231, 7 },
		{ 273, 232, 7 },
		{ 273, 233, 7 },
		{ 273, 234, 7 },
		{ 273, 235, 7 },
		{ 273, 236, 7 },
		{ 273, 237, 7 },
		{ 273, 238, 7 },
		{ 273, 239, 7 },
		{ 273, 240, 7 },
		{ 274, 226, 7 },
		{ 274, 227, 7 },
		{ 274, 228, 7 },
		{ 274, 229, 7 },
		{ 274, 230, 7 },
		{ 274, 231, 7 },
		{ 274, 232, 7 },
		{ 274, 233, 7 },
		{ 274, 234, 7 },
		{ 274, 235, 7 },
		{ 274, 236, 7 },
		{ 274, 237, 7 },
		{ 274, 238, 7 },
		{ 274, 239, 7 },
		{ 275, 226, 7 },
		{ 275, 227, 7 },
		{ 275, 228, 7 },
		{ 275, 229, 7 },
		{ 275, 230, 7 },
		{ 275, 231, 7 },
		{ 275, 232, 7 },
		{ 275, 233, 7 },
		{ 275, 234, 7 },
		{ 275, 235, 7 },
		{ 275, 236, 7 },
		{ 275, 237, 7 },
		{ 275, 238, 7 },
		{ 276, 226, 7 },
		{ 276, 227, 7 },
		{ 276, 228, 7 },
		{ 276, 229, 7 },
		{ 276, 230, 7 },
		{ 276, 231, 7 },
		{ 276, 232, 7 },
		{ 276, 233, 7 },
		{ 276, 234, 7 },
		{ 276, 235, 7 },
		{ 276, 236, 7 },
		{ 276, 237, 7 },
		{ 277, 226, 7 },
		{ 277, 227, 7 },
		{ 277, 228, 7 },
		{ 277, 229, 7 },
		{ 277, 230, 7 },
		{ 277, 231, 7 },
		{ 277, 232, 7 },
		{ 277, 233, 7 },
		{ 277, 234, 7 },
		{ 277, 235, 7 },
		{ 277, 236, 7 },
		{ 278, 226, 7 },
		{ 278, 227, 7 },
		{ 278, 228, 7 },
		{ 278, 229, 7 },
		{ 278, 230, 7 },
		{ 278, 231, 7 },
		{ 278, 232, 7 },
		{ 278, 233, 7 },
		{ 278, 234, 7 },
		{ 278, 235, 7 },
		{ 279, 226, 7 },
		{ 279, 227, 7 },
		{ 279, 228, 7 },
		{ 279, 229, 7 },
		{ 279, 230, 7 },
		{ 279, 231, 7 },
		{ 279, 232, 7 },
		{ 279, 233, 7 },
		{ 279, 234, 7 },
		{ 280, 226, 7 },
		{ 280, 227, 7 },
		{ 280, 228, 7 },
		{ 280, 229, 7 },
		{ 280, 230, 7 },
		{ 280, 231, 7 },
		{ 280, 232, 7 },
		{ 280, 233, 7 },
		{ 281, 226, 7 },
		{ 281, 227, 7 },
		{ 281, 228, 7 },
		{ 281, 229, 7 },
		{ 281, 230, 7 },
		{ 281, 231, 7 },
		{ 281, 232, 7 },
		{ 282, 226, 7 },
		{ 282, 227, 7 },
		{ 282, 228, 7 },
		{ 282, 229, 7 },
		{ 282, 230, 7 },
		{ 282, 231, 7 },
		{ 283, 226, 7 },
		{ 283, 227, 7 },
		{ 283, 228, 7 },
		{ 283, 229, 7 },
		{ 283, 230, 7 },
		{ 284, 226, 7 },
		{ 284, 227, 7 },
		{ 284, 228, 7 },
		{ 284, 229, 7 },
		{ 285, 226, 7 },
		{ 285, 227, 7 },
		{ 285, 228, 7 },
		{ 286, 226, 7 },
		{ 286, 227, 7 },
		{ 287, 226, 7 },
		{ 226, 274, 8 },
		{ 226, 275, 8 },
		{ 226, 276, 8 },
		{ 226, 277, 8 },
		{ 226, 278, 8 },
		{ 226, 279, 8 },
		{ 226, 280, 8 },
		{ 226, 281, 8 },
		{ 226, 282, 8 },
		{ 226, 283, 8 },
		{ 226, 284, 8 },
		{ 226, 285, 8 },
		{ 226, 286, 8 },
		{ 226, 287, 8 },
		{ 226, 288, 8 },
		{ 227, 273, 8 },
		{ 227, 274, 8 },
		{ 227, 275, 8 },
		{ 227, 276, 8 },
		{ 227, 277, 8 },
		{ 227, 278, 8 },
		{ 227, 279, 8 },
		{ 227, 280, 8 },
		{ 227, 281, 8 },
		{ 227, 282, 8 },
		{ 227, 283, 8 },
		{ 227, 284, 8 },
		{ 227, 285, 8 },
		{ 227, 286, 8 },
		{ 227, 287, 8 },
		{ 227, 288, 8 },
		{ 228, 272, 8 },
		{ 228, 273, 8 },
		{ 228, 274, 8 },
		{ 228, 275, 8 },
		{ 228, 276, 8 },
		{ 228, 277, 8 },
		{ 228, 278, 8 },
		{ 228, 279, 8 },
		{ 228, 280, 8 },
		{ 228, 281, 8 },
		{ 228, 282, 8 },
		{ 228, 283, 8 },
		{ 228, 284, 8 },
		{ 228, 285, 8 },
		{ 228, 286, 8 },
		{ 228, 287, 8 },
		{ 228, 288, 8 },
		{ 229, 272, 8 },
		{ 229, 273, 8 },
		{ 229, 274, 8 },
		{ 229, 275, 8 },
		{ 229, 276, 8 },
		{ 229, 277, 8 },
		{ 229, 278, 8 },
		{ 229, 279, 8 },
		{ 229, 280, 8 },
		{ 229, 281, 8 },
		{ 229, 282, 8 },
		{ 229, 283, 8 },
		{ 229, 284, 8 },
		{ 229, 285, 8 },
		{ 229, 286, 8 },
		{ 229, 287, 8 },
		{ 229, 288, 8 },
		{ 230, 271, 8 },
		{ 230, 272, 8 },
		{ 230, 273, 8 },
		{ 230, 274, 8 },
		{ 230, 275, 8 },
		{ 230, 276, 8 },
		{ 230, 277, 8 },
		{ 230, 278, 8 },
		{ 230, 279, 8 },
		{ 230, 280, 8 },
		{ 230, 281, 8 },
		{ 230, 282, 8 },
		{ 230, 283, 8 },
		{ 230, 284, 8 },
		{ 230, 285, 8 },
		{ 230, 286, 8 },
		{ 230, 287, 8 },
		{ 230, 288, 8 },
		{ 231, 271, 8 },
		{ 231, 272, 8 },
		{ 231, 273, 8 },
		{ 231, 274, 8 },
		{ 231, 275, 8 },
		{ 231, 276, 8 },
		{ 231, 277, 8 },
		{ 231, 278, 8 },
		{ 231, 279, 8 },
		{ 231, 280, 8 },
		{ 231, 281, 8 },
		{ 231, 282, 8 },
		{ 231, 283, 8 },
		{ 231, 284, 8 },
		{ 231, 285, 8 },
		{ 231, 286, 8 },
		{ 231, 287, 8 },
		{ 231, 288, 8 },
		{ 232, 270, 8 },
		{ 232, 271, 8 },
		{ 232, 272, 8 },
		{ 232, 273, 8 },
		{ 232, 274, 8 },
		{ 232, 275, 8 },
		{ 232, 276, 8 },
		{ 232, 277, 8 },
		{ 232, 278, 8 },
		{ 232, 279, 8 },
		{ 232, 280, 8 },
		{ 232, 281, 8 },
		{ 232, 282, 8 },
		{ 232, 283, 8 },
		{ 232, 284, 8 },
		{ 232, 285, 8 },
		{ 232, 286, 8 },
		{ 232, 287, 8 },
		{ 232, 288, 8 },
		{ 233, 270, 8 },
		{ 233, 271, 8 },
		{ 233, 272, 8 },
		{ 233, 273, 8 },
		{ 233, 274, 8 },
		{ 233, 275, 8 },
		{ 233, 276, 8 },
		{ 233, 277, 8 },
		{ 233, 278, 8 },
		{ 233, 279, 8 },
		{ 233, 280, 8 },
		{ 233, 281, 8 },
		{ 233, 282, 8 },
		{ 233, 283, 8 },
		{ 233, 284, 8 },
		{ 233, 285, 8 },
		{ 233, 286, 8 },
		{ 233, 287, 8 },
		{ 233, 288, 8 },
		{ 234, 269, 8 },
		{ 234, 270, 8 },
		{ 234, 271, 8 },
		{ 234, 272, 8 },
		{ 234, 273, 8 },
		{ 234, 274, 8 },
		{ 234, 275, 8 },
		{ 234, 276, 8 },
		{ 234, 277, 8 },
		{ 234, 278, 8 },
		{ 234, 279, 8 },
		{ 234, 280, 8 },
		{ 234, 281, 8 },
		{ 234, 282, 8 },
		{ 234, 283, 8 },
		{ 234, 284, 8 },
		{ 234, 285, 8 },
		{ 234, 286, 8 },
		{ 234, 287, 8 },
		{ 234, 288, 8 },
		{ 235, 269, 8 },
		{ 235, 270, 8 },
		{ 235, 271, 8 },
		{ 235, 272, 8 },
		{ 235, 273, 8 },
		{ 235, 274, 8 },
		{ 235, 275, 8 },
		{ 235, 276, 8 },
		{ 235, 277, 8 },
		{ 235, 278, 8 },
		{ 235, 279, 8 },
		{ 235, 280, 8 },
		{ 235, 281, 8 },
		{ 235, 282, 8 },
		{ 235, 283, 8 },
		{ 235, 284, 8 },
		{ 235, 285, 8 },
		{ 235, 286, 8 },
		{ 235, 287, 8 },
		{ 235, 288, 8 },
		{ 236, 268, 8 },
		{ 236, 269, 8 },
		{ 236, 270, 8 },
		{ 236, 271, 8 },
		{ 236, 272, 8 },
		{ 236, 273, 8 },
		{ 236, 274, 8 },
		{ 236, 275, 8 },
		{ 236, 276, 8 },
		{ 236, 277, 8 },
		{ 236, 278, 8 },
		{ 236, 279, 8 },
		{ 236, 280, 8 },
		{ 236, 281, 8 },
		{ 236, 282, 8 },
		{ 236, 283, 8 },
		{ 236, 284, 8 },
		{ 236, 285, 8 },
		{ 236, 286, 8 },
		{ 236, 287, 8 },
		{ 236, 288, 8 },
		{ 237, 268, 8 },
		{ 237, 269, 8 },
		{ 237, 270, 8 },
		{ 237, 271, 8 },
		{ 237, 272, 8 },
		{ 237, 273, 8 },
		{ 237, 274, 8 },
		{ 237, 275, 8 },
		{ 237, 276, 8 },
		{ 237, 277, 8 },
		{ 237, 278, 8 },
		{ 237, 279, 8 },
		{ 237, 280, 8 },
		{ 237, 281, 8 },
		{ 237, 282, 8 },
		{ 237, 283, 8 },
		{ 237, 284, 8 },
		{ 237, 285, 8 },
		{ 237, 286, 8 },
		{ 237, 287, 8 },
		{ 237, 288, 8 },
		{ 238, 267, 8 },
		{ 238, 268, 8 },
		{ 238, 269, 8 },
		{ 238, 270, 8 },
		{ 238, 271, 8 },
		{ 238, 272, 8 },
		{ 238, 273, 8 },
		{ 238, 274, 8 },
		{ 238, 275, 8 },
		{ 238, 276, 8 },
		{ 238, 277, 8 },
		{ 238, 278, 8 },
		{ 238, 279, 8 },
		{ 238, 280, 8 },
		{ 238, 281, 8 },
		{ 238, 282, 8 },
		{ 238, 283, 8 },
		{ 238, 284, 8 },
		{ 238, 285, 8 },
		{ 238, 286, 8 },
		{ 238, 287, 8 },
		{ 238, 288, 8 },
		{ 239, 267, 8 },
		{ 239, 268, 8 },
		{ 239, 269, 8 },
		{ 239, 270, 8 },
		{ 239, 271, 8 },
		{ 239, 272, 8 },
		{ 239, 273, 8 },
		{ 239, 274, 8 },
		{ 239, 275, 8 },
		{ 239, 276, 8 },
		{ 239, 277, 8 },
		{ 239, 278, 8 },
		{ 239, 279, 8 },
		{ 239, 280, 8 },
		{ 239, 281, 8 },
		{ 239, 282, 8 },
		{ 239, 283, 8 },
		{ 239, 284, 8 },
		{ 239, 285, 8 },
		{ 239, 286, 8 },
		{ 239, 287, 8 },
		{ 239, 288, 8 },
		{ 240, 266, 8 },
		{ 240, 267, 8 },
		{ 240, 268, 8 },
		{ 240, 269, 8 },
		{ 240, 270, 8 },
		{ 240, 271, 8 },
		{ 240, 272, 8 },
		{ 240, 273, 8 },
		{ 240, 274, 8 },
		{ 240, 275, 8 },
		{ 240, 276, 8 },
		{ 240, 277, 8 },
		{ 240, 278, 8 },
		{ 240, 279, 8 },
		{ 240, 280, 8 },
		{ 240, 281, 8 },
		{ 240, 282, 8 },
		{ 240, 283, 8 },
		{ 240, 284, 8 },
		{ 240, 285, 8 },
		{ 240, 286, 8 },
		{ 240, 287, 8 },
		{ 240, 288, 8 },
		{ 241, 266, 8 },
		{ 241, 267, 8 },
		{ 241, 268, 8 },
		{ 241, 269, 8 },
		{ 241, 270, 8 },
		{ 241, 271, 8 },
		{ 241, 272, 8 },
		{ 241, 273, 8 },
		{ 241, 274, 8 },
		{ 241, 275, 8 },
		{ 241, 276, 8 },
		{ 241, 277, 8 },
		{ 241, 278, 8 },
		{ 241, 279, 8 },
		{ 241, 280, 8 },
		{ 241, 281, 8 },
		{ 241, 282, 8 },
		{ 241, 283, 8 },
		{ 241, 284, 8 },
		{ 241, 285, 8 },
		{ 241, 286, 8 },
		{ 241, 287, 8 },
		{ 242, 265, 8 },
		{ 242, 266, 8 },
		{ 242, 267, 8 },
		{ 242, 268, 8 },
		{ 242, 269, 8 },
		{ 242, 270, 8 },
		{ 242, 271, 8 },
		{ 242, 272, 8 },
		{ 242, 273, 8 },
		{ 242, 274, 8 },
		{ 242, 275, 8 },
		{ 242, 276, 8 },
		{ 242, 277, 8 },
		{ 242, 278, 8 },
		{ 242, 279, 8 },
		{ 242, 280, 8 },
		{ 242, 281, 8 },
		{ 242, 282, 8 },
		{ 242, 283, 8 },
		{ 242, 284, 8 },
		{ 242, 285, 8 },
		{ 242, 286, 8 },
		{ 243, 265, 8 },
		{ 243, 266, 8 },
		{ 243, 267, 8 },
		{ 243, 268, 8 },
		{ 243, 269, 8 },
		{ 243, 270, 8 },
		{ 243, 271, 8 },
		{ 243, 272, 8 },
		{ 243, 273, 8 },
		{ 243, 274, 8 },
		{ 243, 275, 8 },
		{ 243, 276, 8 },
		{ 243, 277, 8 },
		{ 243, 278, 8 },
		{ 243, 279, 8 },
		{ 243, 280, 8 },
		{ 243, 281, 8 },
		{ 243, 282, 8 },
		{ 243, 283, 8 },
		{ 243, 284, 8 },
		{ 244, 264, 8 },
		{ 244, 265, 8 },
		{ 244, 266, 8 },
		{ 244, 267, 8 },
		{ 244, 268, 8 },
		{ 244, 269, 8 },
		{ 244, 270, 8 },
		{ 244, 271, 8 },
		{ 244, 272, 8 },
		{ 244, 273, 8 },
		{ 244, 274, 8 },
		{ 244, 275, 8 },
		{ 244, 276, 8 },
		{ 244, 277, 8 },
		{ 244, 278, 8 },
		{ 244, 279, 8 },
		{ 244, 280, 8 },
		{ 244, 281, 8 },
		{ 244, 282, 8 },
		{ 245, 264, 8 },
		{ 245, 265, 8 },
		{ 245, 266, 8 },
		{ 245, 267, 8 },
		{ 245, 268, 8 },
		{ 245, 269, 8 },
		{ 245, 270, 8 },
		{ 245, 271, 8 },
		{ 245, 272, 8 },
		{ 245, 273, 8 },
		{ 245, 274, 8 },
		{ 245, 275, 8 },
		{ 245, 276, 8 },
		{ 245, 277, 8 },
		{ 245, 278, 8 },
		{ 245, 279, 8 },
		{ 245, 280, 8 },
		{ 246, 263, 8 },
		{ 246, 264, 8 },
		{ 246, 265, 8 },
		{ 246, 266, 8 },
		{ 246, 267, 8 },
		{ 246, 268, 8 },
		{ 246, 269, 8 },
		{ 246, 270, 8 },
		{ 246, 271, 8 },
		{ 246, 272, 8 },
		{ 246, 273, 8 },
		{ 246, 274, 8 },
		{ 246, 275, 8 },
		{ 246, 276, 8 },
		{ 246, 277, 8 },
		{ 246, 278, 8 },
		{ 247, 263, 8 },
		{ 247, 264, 8 },
		{ 247, 265, 8 },
		{ 247, 266, 8 },
		{ 247, 267, 8 },
		{ 247, 268, 8 },
		{ 247, 269, 8 },
		{ 247, 270, 8 },
		{ 247, 271, 8 },
		{ 247, 272, 8 },
		{ 247, 273, 8 },
		{ 247, 274, 8 },
		{ 247, 275, 8 },
		{ 247, 276, 8 },
		{ 248, 262, 8 },
		{ 248, 263, 8 },
		{ 248, 264, 8 },
		{ 248, 265, 8 },
		{ 248, 266, 8 },
		{ 248, 267, 8 },
		{ 248, 268, 8 },
		{ 248, 269, 8 },
		{ 248, 270, 8 },
		{ 248, 271, 8 },
		{ 248, 272, 8 },
		{ 248, 273, 8 },
		{ 248, 274, 8 },
		{ 249, 262, 8 },
		{ 249, 263, 8 },
		{ 249, 264, 8 },
		{ 249, 265, 8 },
		{ 249, 266, 8 },
		{ 249, 267, 8 },
		{ 249, 268, 8 },
		{ 249, 269, 8 },
		{ 249, 270, 8 },
		{ 249, 271, 8 },
		{ 249, 272, 8 },
		{ 250, 261, 8 },
		{ 250, 262, 8 },
		{ 250, 263, 8 },
		{ 250, 264, 8 },
		{ 250, 265, 8 },
		{ 250, 266, 8 },
		{ 250, 267, 8 },
		{ 250, 268, 8 },
		{ 250, 269, 8 },
		{ 250, 270, 8 },
		{ 251, 261, 8 },
		{ 251, 262, 8 },
		{ 251, 263, 8 },
		{ 251, 264, 8 },
		{ 251, 265, 8 },
		{ 251, 266, 8 },
		{ 251, 267, 8 },
		{ 251, 268, 8 },
		{ 252, 260, 8 },
		{ 252, 261, 8 },
		{ 252, 262, 8 },
		{ 252, 263, 8 },
		{ 252, 264, 8 },
		{ 252, 265, 8 },
		{ 252, 266, 8 },
		{ 253, 262, 8 },
		{ 253, 263, 8 },
		{ 253, 264, 8 },
		{ 254, 262, 8 },
		{ 260, 252, 8 },
		{ 261, 250, 8 },
		{ 261, 251, 8 },
		{ 261, 252, 8 },
		{ 262, 248, 8 },
		{ 262, 249, 8 },
		{ 262, 250, 8 },
		{ 262, 251, 8 },
		{ 262, 252, 8 },
		{ 262, 253, 8 },
		{ 262, 254, 8 },
		{ 263, 246, 8 },
		{ 263, 247, 8 },
		{ 263, 248, 8 },
		{ 263, 249, 8 },
		{ 263, 250, 8 },
		{ 263, 251, 8 },
		{ 263, 252, 8 },
		{ 263, 253, 8 },
		{ 264, 244, 8 },
		{ 264, 245, 8 },
		{ 264, 246, 8 },
		{ 264, 247, 8 },
		{ 264, 248, 8 },
		{ 264, 249, 8 },
		{ 264, 250, 8 },
		{ 264, 251, 8 },
		{ 264, 252, 8 },
		{ 264, 253, 8 },
		{ 265, 242, 8 },
		{ 265, 243, 8 },
		{ 265, 244, 8 },
		{ 265, 245, 8 },
		{ 265, 246, 8 },
		{ 265, 247, 8 },
		{ 265, 248, 8 },
		{ 265, 249, 8 },
		{ 265, 250, 8 },
		{ 265, 251, 8 },
		{ 265, 252, 8 },
		{ 266, 240, 8 },
		{ 266, 241, 8 },
		{ 266, 242, 8 },
		{ 266, 243, 8 },
		{ 266, 244, 8 },
		{ 266, 245, 8 },
		{ 266, 246, 8 },
		{ 266, 247, 8 },
		{ 266, 248, 8 },
		{ 266, 249, 8 },
		{ 266, 250, 8 },
		{ 266, 251, 8 },
		{ 266, 252, 8 },
		{ 267, 238, 8 },
		{ 267, 239, 8 },
		{ 267, 240, 8 },
		{ 267, 241, 8 },
		{ 267, 242, 8 },
		{ 267, 243, 8 },
		{ 267, 244, 8 },
		{ 267, 245, 8 },
		{ 267, 246, 8 },
		{ 267, 247, 8 },
		{ 267, 248, 8 },
		{ 267, 249, 8 },
		{ 267, 250, 8 },
		{ 267, 251, 8 },
		{ 268, 236, 8 },
		{ 268, 237, 8 },
		{ 268, 238, 8 },
		{ 268, 239, 8 },
		{ 268, 240, 8 },
		{ 268, 241, 8 },
		{ 268, 242, 8 },
		{ 268, 243, 8 },
		{ 268, 244, 8 },
		{ 268, 245, 8 },
		{ 268, 246, 8 },
		{ 268, 247, 8 },
		{ 268, 248, 8 },
		{ 268, 249, 8 },
		{ 268, 250, 8 },
		{ 268, 251, 8 },
		{ 269, 234, 8 },
		{ 269, 235, 8 },
		{ 269, 236, 8 },
		{ 269, 237, 8 },
		{ 269, 238, 8 },
		{ 269, 239, 8 },
		{ 269, 240, 8 },
		{ 269, 241, 8 },
		{ 269, 242, 8 },
		{ 269, 243, 8 },
		{ 269, 244, 8 },
		{ 269, 245, 8 },
		{ 269, 246, 8 },
		{ 269, 247, 8 },
		{ 269, 248, 8 },
		{ 269, 249, 8 },
		{ 269, 250, 8 },
		{ 270, 232, 8 },
		{ 270, 233, 8 },
		{ 270, 234, 8 },
		{ 270, 235, 8 },
		{ 270, 236, 8 },
		{ 270, 237, 8 },
		{ 270, 238, 8 },
		{ 270, 239, 8 },
		{ 270, 240, 8 },
		{ 270, 241, 8 },
		{ 270, 242, 8 },
		{ 270, 243, 8 },
		{ 270, 244, 8 },
		{ 270, 245, 8 },
		{ 270, 246, 8 },
		{ 270, 247, 8 },
		{ 270, 248, 8 },
		{ 270, 249, 8 },
		{ 270, 250, 8 },
		{ 271, 230, 8 },
		{ 271, 231, 8 },
		{ 271, 232, 8 },
		{ 271, 233, 8 },
		{ 271, 234, 8 },
		{ 271, 235, 8 },
		{ 271, 236, 8 },
		{ 271, 237, 8 },
		{ 271, 238, 8 },
		{ 271, 239, 8 },
		{ 271, 240, 8 },
		{ 271, 241, 8 },
		{ 271, 242, 8 },
		{ 271, 243, 8 },
		{ 271, 244, 8 },
		{ 271, 245, 8 },
		{ 271, 246, 8 },
		{ 271, 247, 8 },
		{ 271, 248, 8 },
		{ 271, 249, 8 },
		{ 272, 228, 8 },
		{ 272, 229, 8 },
		{ 272, 230, 8 },
		{ 272, 231, 8 },
		{ 272, 232, 8 },
		{ 272, 233, 8 },
		{ 272, 234, 8 },
		{ 272, 235, 8 },
		{ 272, 236, 8 },
		{ 272, 237, 8 },
		{ 272, 238, 8 },
		{ 272, 239, 8 },
		{ 272, 240, 8 },
		{ 272, 241, 8 },
		{ 272, 242, 8 },
		{ 272, 243, 8 },
		{ 272, 244, 8 },
		{ 272, 245, 8 },
		{ 272, 246, 8 },
		{ 272, 247, 8 },
		{ 272, 248, 8 },
		{ 272, 249, 8 },
		{ 273, 227, 8 },
		{ 273, 228, 8 },
		{ 273, 229, 8 },
		{ 273, 230, 8 },
		{ 273, 231, 8 },
		{ 273, 232, 8 },
		{ 273, 233, 8 },
		{ 273, 234, 8 },
		{ 273, 235, 8 },
		{ 273, 236, 8 },
		{ 273, 237, 8 },
		{ 273, 238, 8 },
		{ 273, 239, 8 },
		{ 273, 240, 8 },
		{ 273, 241, 8 },
		{ 273, 242, 8 },
		{ 273, 243, 8 },
		{ 273, 244, 8 },
		{ 273, 245, 8 },
		{ 273, 246, 8 },
		{ 273, 247, 8 },
		{ 273, 248, 8 },
		{ 274, 226, 8 },
		{ 274, 227, 8 },
		{ 274, 228, 8 },
		{ 274, 229, 8 },
		{ 274, 230, 8 },
		{ 274, 231, 8 },
		{ 274, 232, 8 },
		{ 274, 233, 8 },
		{ 274, 234, 8 },
		{ 274, 235, 8 },
		{ 274, 236, 8 },
		{ 274, 237, 8 },
		{ 274, 238, 8 },
		{ 274, 239, 8 },
		{ 274, 240, 8 },
		{ 274, 241, 8 },
		{ 274, 242, 8 },
		{ 274, 243, 8 },
		{ 274, 244, 8 },
		{ 274, 245, 8 },
		{ 274, 246, 8 },
		{ 274, 247, 8 },
		{ 274, 248, 8 },
		{ 275, 226, 8 },
		{ 275, 227, 8 },
		{ 275, 228, 8 },
		{ 275, 229, 8 },
		{ 275, 230, 8 },
		{ 275, 231, 8 },
		{ 275, 232, 8 },
		{ 275, 233, 8 },
		{ 275, 234, 8 },
		{ 275, 235, 8 },
		{ 275, 236, 8 },
		{ 275, 237, 8 },
		{ 275, 238, 8 },
		{ 275, 239, 8 },
		{ 275, 240, 8 },
		{ 275, 241, 8 },
		{ 275, 242, 8 },
		{ 275, 243, 8 },
		{ 275, 244, 8 },
		{ 275, 245, 8 },
		{ 275, 246, 8 },
		{ 275, 247, 8 },
		{ 276, 226, 8 },
		{ 276, 227, 8 },
		{ 276, 228, 8 },
		{ 276, 229, 8 },
		{ 276, 230, 8 },
		{ 276, 231, 8 },
		{ 276, 232, 8 },
		{ 276, 233, 8 },
		{ 276, 234, 8 },
		{ 276, 235, 8 },
		{ 276, 236, 8 },
		{ 276, 237, 8 },
		{ 276, 238, 8 },
		{ 276, 239, 8 },
		{ 276, 240, 8 },
		{ 276, 241, 8 },
		{ 276, 242, 8 },
		{ 276, 243, 8 },
		{ 276, 244, 8 },
		{ 276, 245, 8 },
		{ 276, 246, 8 },
		{ 276, 247, 8 },
		{ 277, 226, 8 },
		{ 277, 227, 8 },
		{ 277, 228, 8 },
		{ 277, 229, 8 },
		{ 277, 230, 8 },
		{ 277, 231, 8 },
		{ 277, 232, 8 },
		{ 277, 233, 8 },
		{ 277, 234, 8 },
		{ 277, 235, 8 },
		{ 277, 236, 8 },
		{ 277, 237, 8 },
		{ 277, 238, 8 },
		{ 277, 239, 8 },
		{ 277, 240, 8 },
		{ 277, 241, 8 },
		{ 277, 242, 8 },
		{ 277, 243, 8 },
		{ 277, 244, 8 },
		{ 277, 245, 8 },
		{ 277, 246, 8 },
		{ 278, 226, 8 },
		{ 278, 227, 8 },
		{ 278, 228, 8 },
		{ 278, 229, 8 },
		{ 278, 230, 8 },
		{ 278, 231, 8 },
		{ 278, 232, 8 },
		{ 278, 233, 8 },
		{ 278, 234, 8 },
		{ 278, 235, 8 },
		{ 278, 236, 8 },
		{ 278, 237, 8 },
		{ 278, 238, 8 },
		{ 278, 239, 8 },
		{ 278, 240, 8 },
		{ 278, 241, 8 },
		{ 278, 242, 8 },
		{ 278, 243, 8 },
		{ 278, 244, 8 },
		{ 278, 245, 8 },
		{ 278, 246, 8 },
		{ 279, 226, 8 },
		{ 279, 227, 8 },
		{ 279, 228, 8 },
		{ 279, 229, 8 },
		{ 279, 230, 8 },
		{ 279, 231, 8 },
		{ 279, 232, 8 },
		{ 279, 233, 8 },
		{ 279, 234, 8 },
		{ 279, 235, 8 },
		{ 279, 236, 8 },
		{ 279, 237, 8 },
		{ 279, 238, 8 },
		{ 279, 239, 8 },
		{ 279, 240, 8 },
		{ 279, 241, 8 },
		{ 279, 242, 8 },
		{ 279, 243, 8 },
		{ 279, 244, 8 },
		{ 279, 245, 8 },
		{ 280, 226, 8 },
		{ 280, 227, 8 },
		{ 280, 228, 8 },
		{ 280, 229, 8 },
		{ 280, 230, 8 },
		{ 280, 231, 8 },
		{ 280, 232, 8 },
		{ 280, 233, 8 },
		{ 280, 234, 8 },
		{ 280, 235, 8 },
		{ 280, 236, 8 },
		{ 280, 237, 8 },
		{ 280, 238, 8 },
		{ 280, 239, 8 },
		{ 280, 240, 8 },
		{ 280, 241, 8 },
		{ 280, 242, 8 },
		{ 280, 243, 8 },
		{ 280, 244, 8 },
		{ 280, 245, 8 },
		{ 281, 226, 8 },
		{ 281, 227, 8 },
		{ 281, 228, 8 },
		{ 281, 229, 8 },
		{ 281, 230, 8 },
		{ 281, 231, 8 },
		{ 281, 232, 8 },
		{ 281, 233, 8 },
		{ 281, 234, 8 },
		{ 281, 235, 8 },
		{ 281, 236, 8 },
		{ 281, 237, 8 },
		{ 281, 238, 8 },
		{ 281, 239, 8 },
		{ 281, 240, 8 },
		{ 281, 241, 8 },
		{ 281, 242, 8 },
		{ 281, 243, 8 },
		{ 281, 244, 8 },
		{ 282, 226, 8 },
		{ 282, 227, 8 },
		{ 282, 228, 8 },
		{ 282, 229, 8 },
		{ 282, 230, 8 },
		{ 282, 231, 8 },
		{ 282, 232, 8 },
		{ 282, 233, 8 },
		{ 282, 234, 8 },
		{ 282, 235, 8 },
		{ 282, 236, 8 },
		{ 282, 237, 8 },
		{ 282, 238, 8 },
		{ 282, 239, 8 },
		{ 282, 240, 8 },
		{ 282, 241, 8 },
		{ 282, 242, 8 },
		{ 282, 243, 8 },
		{ 282, 244, 8 },
		{ 283, 226, 8 },
		{ 283, 227, 8 },
		{ 283, 228, 8 },
		{ 283, 229, 8 },
		{ 283, 230, 8 },
		{ 283, 231, 8 },
		{ 283, 232, 8 },
		{ 283, 233, 8 },
		{ 283, 234, 8 },
		{ 283, 235, 8 },
		{ 283, 236, 8 },
		{ 283, 237, 8 },
		{ 283, 238, 8 },
		{ 283, 239, 8 },
		{ 283, 240, 8 },
		{ 283, 241, 8 },
		{ 283, 242, 8 },
		{ 283, 243, 8 },
		{ 284, 226, 8 },
		{ 284, 227, 8 },
		{ 284, 228, 8 },
		{ 284, 229, 8 },
		{ 284, 230, 8 },
		{ 284, 231, 8 },
		{ 284, 232, 8 },
		{ 284, 233, 8 },
		{ 284, 234, 8 },
		{ 284, 235, 8 },
		{ 284, 236, 8 },
		{ 284, 237, 8 },
		{ 284, 238, 8 },
		{ 284, 239, 8 },
		{ 284, 240, 8 },
		{ 284, 241, 8 },
		{ 284, 242, 8 },
		{ 284, 243, 8 },
		{ 285, 226, 8 },
		{ 285, 227, 8 },
		{ 285, 228, 8 },
		{ 285, 229, 8 },
		{ 285, 230, 8 },
		{ 285, 231, 8 },
		{ 285, 232, 8 },
		{ 285, 233, 8 },
		{ 285, 234, 8 },
		{ 285, 235, 8 },
		{ 285, 236, 8 },
		{ 285, 237, 8 },
		{ 285, 238, 8 },
		{ 285, 239, 8 },
		{ 285, 240, 8 },
		{ 285, 241, 8 },
		{ 285, 242, 8 },
		{ 286, 226, 8 },
		{ 286, 227, 8 },
		{ 286, 228, 8 },
		{ 286, 229, 8 },
		{ 286, 230, 8 },
		{ 286, 231, 8 },
		{ 286, 232, 8 },
		{ 286, 233, 8 },
		{ 286, 234, 8 },
		{ 286, 235, 8 },
		{ 286, 236, 8 },
		{ 286, 237, 8 },
		{ 286, 238, 8 },
		{ 286, 239, 8 },
		{ 286, 240, 8 },
		{ 286, 241, 8 },
		{ 286, 242, 8 },
		{ 287, 226, 8 },
		{ 287, 227, 8 },
		{ 287, 228, 8 },
		{ 287, 229, 8 },
		{ 287, 230, 8 },
		{ 287, 231, 8 },
		{ 287, 232, 8 },
		{ 287, 233, 8 },
		{ 287, 234, 8 },
		{ 287, 235, 8 },
		{ 287, 236, 8 },
		{ 287, 237, 8 },
		{ 287, 238, 8 },
		{ 287, 239, 8 },
		{ 287, 240, 8 },
		{ 287, 241, 8 },
		{ 288, 226, 8 },
		{ 288, 227, 8 },
		{ 288, 228, 8 },
		{ 288, 229, 8 },
		{ 288, 230, 8 },
		{ 288, 231, 8 },
		{ 288, 232, 8 },
		{ 288, 233, 8 },
		{ 288, 234, 8 },
		{ 288, 235, 8 },
		{ 288, 236, 8 },
		{ 288, 237, 8 },
		{ 288, 238, 8 },
		{ 288, 239, 8 },
		{ 288, 240, 8 },
		{ 226, 258, 9 },
		{ 226, 259, 9 },
		{ 226, 260, 9 },
		{ 226, 261, 9 },
		{ 226, 262, 9 },
		{ 226, 263, 9 },
		{ 226, 264, 9 },
		{ 226, 265, 9 },
		{ 226, 266, 9 },
		{ 226, 267, 9 },
		{ 226, 268, 9 },
		{ 226, 269, 9 },
		{ 226, 270, 9 },
		{ 226, 271, 9 },
		{ 226, 272, 9 },
		{ 226, 273, 9 },
		{ 226, 274, 9 },
		{ 226, 275, 9 },
		{ 226, 276, 9 },
		{ 226, 277, 9 },
		{ 226, 278, 9 },
		{ 226, 279, 9 },
		{ 226, 280, 9 },
		{ 226, 281, 9 },
		{ 226, 282, 9 },
		{ 226, 283, 9 },
		{ 226, 284, 9 },
		{ 226, 285, 9 },
		{ 226, 286, 9 },
		{ 226, 287, 9 },
		{ 227, 258, 9 },
		{ 227, 259, 9 },
		{ 227, 260, 9 },
		{ 227, 261, 9 },
		{ 227, 262, 9 },
		{ 227, 263, 9 },
		{ 227, 264, 9 },
		{ 227, 265, 9 },
		{ 227, 266, 9 },
		{ 227, 267, 9 },
		{ 227, 268, 9 },
		{ 227, 269, 9 },
		{ 227, 270, 9 },
		{ 227, 271, 9 },
		{ 227, 272, 9 },
		{ 227, 273, 9 },
		{ 227, 274, 9 },
		{ 227, 275, 9 },
		{ 227, 276, 9 },
		{ 227, 277, 9 },
		{ 227, 278, 9 },
		{ 227, 279, 9 },
		{ 227, 280, 9 },
		{ 227, 281, 9 },
		{ 227, 282, 9 },
		{ 227, 283, 9 },
		{ 227, 284, 9 },
		{ 227, 285, 9 },
		{ 227, 286, 9 },
		{ 228, 258, 9 },
		{ 228, 259, 9 },
		{ 228, 260, 9 },
		{ 228, 261, 9 },
		{ 228, 262, 9 },
		{ 228, 263, 9 },
		{ 228, 264, 9 },
		{ 228, 265, 9 },
		{ 228, 266, 9 },
		{ 228, 267, 9 },
		{ 228, 268, 9 },
		{ 228, 269, 9 },
		{ 228, 270, 9 },
		{ 228, 271, 9 },
		{ 228, 272, 9 },
		{ 228, 273, 9 },
		{ 228, 274, 9 },
		{ 228, 275, 9 },
		{ 228, 276, 9 },
		{ 228, 277, 9 },
		{ 228, 278, 9 },
		{ 228, 279, 9 },
		{ 228, 280, 9 },
		{ 228, 281, 9 },
		{ 228, 282, 9 },
		{ 228, 283, 9 },
		{ 228, 284, 9 },
		{ 228, 285, 9 },
		{ 229, 258, 9 },
		{ 229, 259, 9 },
		{ 229, 260, 9 },
		{ 229, 261, 9 },
		{ 229, 262, 9 },
		{ 229, 263, 9 },
		{ 229, 264, 9 },
		{ 229, 265, 9 },
		{ 229, 266, 9 },
		{ 229, 267, 9 },
		{ 229, 268, 9 },
		{ 229, 269, 9 },
		{ 229, 270, 9 },
		{ 229, 271, 9 },
		{ 229, 272, 9 },
		{ 229, 273, 9 },
		{ 229, 274, 9 },
		{ 229, 275, 9 },
		{ 229, 276, 9 },
		{ 229, 277, 9 },
		{ 229, 278, 9 },
		{ 229, 279, 9 },
		{ 229, 280, 9 },
		{ 229, 281, 9 },
		{ 229, 282, 9 },
		{ 229, 283, 9 },
		{ 229, 284, 9 },
		{ 230, 258, 9 },
		{ 230, 259, 9 },
		{ 230, 260, 9 },
		{ 230, 261, 9 },
		{ 230, 262, 9 },
		{ 230, 263, 9 },
		{ 230, 264, 9 },
		{ 230, 265, 9 },
		{ 230, 266, 9 },
		{ 230, 267, 9 },
		{ 230, 268, 9 },
		{ 230, 269, 9 },
		{ 230, 270, 9 },
		{ 230, 271, 9 },
		{ 230, 272, 9 },
		{ 230, 273, 9 },
		{ 230, 274, 9 },
		{ 230, 275, 9 },
		{ 230, 276, 9 },
		{ 230, 277, 9 },
		{ 230, 278, 9 },
		{ 230, 279, 9 },
		{ 230, 280, 9 },
		{ 230, 281, 9 },
		{ 230, 282, 9 },
		{ 230, 283, 9 },
		{ 231, 258, 9 },
		{ 231, 259, 9 },
		{ 231, 260, 9 },
		{ 231, 261, 9 },
		{ 231, 262, 9 },
		{ 231, 263, 9 },
		{ 231, 264, 9 },
		{ 231, 265, 9 },
		{ 231, 266, 9 },
		{ 231, 267, 9 },
		{ 231, 268, 9 },
		{ 231, 269, 9 },
		{ 231, 270, 9 },
		{ 231, 271, 9 },
		{ 231, 272, 9 },
		{ 231, 273, 9 },
		{ 231, 274, 9 },
		{ 231, 275, 9 },
		{ 231, 276, 9 },
		{ 231, 277, 9 },
		{ 231, 278, 9 },
		{ 231, 279, 9 },
		{ 231, 280, 9 },
		{ 231, 281, 9 },
		{ 231, 282, 9 },
		{ 232, 258, 9 },
		{ 232, 259, 9 },
		{ 232, 260, 9 },
		{ 232, 261, 9 },
		{ 232, 262, 9 },
		{ 232, 263, 9 },
		{ 232, 264, 9 },
		{ 232, 265, 9 },
		{ 232, 266, 9 },
		{ 232, 267, 9 },
		{ 232, 268, 9 },
		{ 232, 269, 9 },
		{ 232, 270, 9 },
		{ 232, 271, 9 },
		{ 232, 272, 9 },
		{ 232, 273, 9 },
		{ 232, 274, 9 },
		{ 232, 275, 9 },
		{ 232, 276, 9 },
		{ 232, 277, 9 },
		{ 232, 278, 9 },
		{ 232, 279, 9 },
		{ 232, 280, 9 },
		{ 232, 281, 9 },
		{ 233, 258, 9 },
		{ 233, 259, 9 },
		{ 233, 260, 9 },
		{ 233, 261, 9 },
		{ 233, 262, 9 },
		{ 233, 263, 9 },
		{ 233, 264, 9 },
		{ 233, 265, 9 },
		{ 233, 266, 9 },
		{ 233, 267, 9 },
		{ 233, 268, 9 },
		{ 233, 269, 9 },
		{ 233, 270, 9 },
		{ 233, 271, 9 },
		{ 233, 272, 9 },
		{ 233, 273, 9 },
		{ 233, 274, 9 },
		{ 233, 275, 9 },
		{ 233, 276, 9 },
		{ 233, 277, 9 },
		{ 233, 278, 9 },
		{ 233, 279, 9 },
		{ 233, 280, 9 },
		{ 234, 258, 9 },
		{ 234, 259, 9 },
		{ 234, 260, 9 },
		{ 234, 261, 9 },
		{ 234, 262, 9 },
		{ 234, 263, 9 },
		{ 234, 264, 9 },
		{ 234, 265, 9 },
		{ 234, 266, 9 },
		{ 234, 267, 9 },
		{ 234, 268, 9 },
		{ 234, 269, 9 },
		{ 234, 270, 9 },
		{ 234, 271, 9 },
		{ 234, 272, 9 },
		{ 234, 273, 9 },
		{ 234, 274, 9 },
		{ 234, 275, 9 },
		{ 234, 276, 9 },
		{ 234, 277, 9 },
		{ 234, 278, 9 },
		{ 234, 279, 9 },
		{ 235, 258, 9 },
		{ 235, 259, 9 },
		{ 235, 260, 9 },
		{ 235, 261, 9 },
		{ 235, 262, 9 },
		{ 235, 263, 9 },
		{ 235, 264, 9 },
		{ 235, 265, 9 },
		{ 235, 266, 9 },
		{ 235, 267, 9 },
		{ 235, 268, 9 },
		{ 235, 269, 9 },
		{ 235, 270, 9 },
		{ 235, 271, 9 },
		{ 235, 272, 9 },
		{ 235, 273, 9 },
		{ 235, 274, 9 },
		{ 235, 275, 9 },
		{ 235, 276, 9 },
		{ 235, 277, 9 },
		{ 235, 278, 9 },
		{ 236, 258, 9 },
		{ 236, 259, 9 },
		{ 236, 260, 9 },
		{ 236, 261, 9 },
		{ 236, 262, 9 },
		{ 236, 263, 9 },
		{ 236, 264, 9 },
		{ 236, 265, 9 },
		{ 236, 266, 9 },
		{ 236, 267, 9 },
		{ 236, 268, 9 },
		{ 236, 269, 9 },
		{ 236, 270, 9 },
		{ 236, 271, 9 },
		{ 236, 272, 9 },
		{ 236, 273, 9 },
		{ 236, 274, 9 },
		{ 236, 275, 9 },
		{ 236, 276, 9 },
		{ 236, 277, 9 },
		{ 237, 258, 9 },
		{ 237, 259, 9 },
		{ 237, 260, 9 },
		{ 237, 261, 9 },
		{ 237, 262, 9 },
		{ 237, 263, 9 },
		{ 237, 264, 9 },
		{ 237, 265, 9 },
		{ 237, 266, 9 },
		{ 237, 267, 9 },
		{ 237, 268, 9 },
		{ 237, 269, 9 },
		{ 237, 270, 9 },
		{ 237, 271, 9 },
		{ 237, 272, 9 },
		{ 237, 273, 9 },
		{ 237, 274, 9 },
		{ 237, 275, 9 },
		{ 237, 276, 9 },
		{ 238, 258, 9 },
		{ 238, 259, 9 },
		{ 238, 260, 9 },
		{ 238, 261, 9 },
		{ 238, 262, 9 },
		{ 238, 263, 9 },
		{ 238, 264, 9 },
		{ 238, 265, 9 },
		{ 238, 266, 9 },
		{ 238, 267, 9 },
		{ 238, 268, 9 },
		{ 238, 269, 9 },
		{ 238, 270, 9 },
		{ 238, 271, 9 },
		{ 238, 272, 9 },
		{ 238, 273, 9 },
		{ 238, 274, 9 },
		{ 238, 275, 9 },
		{ 239, 258, 9 },
		{ 239, 259, 9 },
		{ 239, 260, 9 },
		{ 239, 261, 9 },
		{ 239, 262, 9 },
		{ 239, 263, 9 },
		{ 239, 264, 9 },
		{ 239, 265, 9 },
		{ 239, 266, 9 },
		{ 239, 267, 9 },
		{ 239, 268, 9 },
		{ 239, 269, 9 },
		{ 239, 270, 9 },
		{ 239, 271, 9 },
		{ 239, 272, 9 },
		{ 239, 273, 9 },
		{ 239, 274, 9 },
		{ 240, 258, 9 },
		{ 240, 259, 9 },
		{ 240, 260, 9 },
		{ 240, 261, 9 },
		{ 240, 262, 9 },
		{ 240, 263, 9 },
		{ 240, 264, 9 },
		{ 240, 265, 9 },
		{ 240, 266, 9 },
		{ 240, 267, 9 },
		{ 240, 268, 9 },
		{ 240, 269, 9 },
		{ 240, 270, 9 },
		{ 240, 271, 9 },
		{ 240, 272, 9 },
		{ 240, 273, 9 },
		{ 241, 258, 9 },
		{ 241, 259, 9 },
		{ 241, 260, 9 },
		{ 241, 261, 9 },
		{ 241, 262, 9 },
		{ 241, 263, 9 },
		{ 241, 264, 9 },
		{ 241, 265, 9 },
		{ 241, 266, 9 },
		{ 241, 267, 9 },
		{ 241, 268, 9 },
		{ 241, 269, 9 },
		{ 241, 270, 9 },
		{ 241, 271, 9 },
		{ 241, 272, 9 },
		{ 242, 258, 9 },
		{ 242, 259, 9 },
		{ 242, 260, 9 },
		{ 242, 261, 9 },
		{ 242, 262, 9 },
		{ 242, 263, 9 },
		{ 242, 264, 9 },
		{ 242, 265, 9 },
		{ 242, 266, 9 },
		{ 242, 267, 9 },
		{ 242, 268, 9 },
		{ 242, 269, 9 },
		{ 242, 270, 9 },
		{ 242, 271, 9 },
		{ 243, 258, 9 },
		{ 243, 259, 9 },
		{ 243, 260, 9 },
		{ 243, 261, 9 },
		{ 243, 262, 9 },
		{ 243, 263, 9 },
		{ 243, 264, 9 },
		{ 243, 265, 9 },
		{ 243, 266, 9 },
		{ 243, 267, 9 },
		{ 243, 268, 9 },
		{ 243, 269, 9 },
		{ 243, 270, 9 },
		{ 244, 258, 9 },
		{ 244, 259, 9 },
		{ 244, 260, 9 },
		{ 244, 261, 9 },
		{ 244, 262, 9 },
		{ 244, 263, 9 },
		{ 244, 264, 9 },
		{ 244, 265, 9 },
		{ 244, 266, 9 },
		{ 244, 267, 9 },
		{ 244, 268, 9 },
		{ 244, 269, 9 },
		{ 245, 258, 9 },
		{ 245, 259, 9 },
		{ 245, 260, 9 },
		{ 245, 261, 9 },
		{ 245, 262, 9 },
		{ 245, 263, 9 },
		{ 245, 264, 9 },
		{ 245, 265, 9 },
		{ 245, 266, 9 },
		{ 245, 267, 9 },
		{ 245, 268, 9 },
		{ 246, 258, 9 },
		{ 246, 259, 9 },
		{ 246, 260, 9 },
		{ 246, 261, 9 },
		{ 246, 262, 9 },
		{ 246, 263, 9 },
		{ 246, 264, 9 },
		{ 246, 265, 9 },
		{ 246, 266, 9 },
		{ 246, 267, 9 },
		{ 247, 258, 9 },
		{ 247, 259, 9 },
		{ 247, 260, 9 },
		{ 247, 261, 9 },
		{ 247, 262, 9 },
		{ 247, 263, 9 },
		{ 247, 264, 9 },
		{ 247, 265, 9 },
		{ 247, 266, 9 },
		{ 248, 258, 9 },
		{ 248, 259, 9 },
		{ 248, 260, 9 },
		{ 248, 261, 9 },
		{ 248, 262, 9 },
		{ 248, 263, 9 },
		{ 248, 264, 9 },
		{ 248, 265, 9 },
		{ 249, 258, 9 },
		{ 249, 259, 9 },
		{ 249, 260, 9 },
		{ 249, 261, 9 },
		{ 249, 262, 9 },
		{ 249, 263, 9 },
		{ 249, 264, 9 },
		{ 250, 258, 9 },
		{ 250, 259, 9 },
		{ 250, 260, 9 },
		{ 250, 261, 9 },
		{ 250, 262, 9 },
		{ 250, 263, 9 },
		{ 251, 258, 9 },
		{ 251, 259, 9 },
		{ 251, 260, 9 },
		{ 251, 261, 9 },
		{ 251, 262, 9 },
		{ 252, 258, 9 },
		{ 252, 259, 9 },
		{ 252, 260, 9 },
		{ 252, 261, 9 },
		{ 262, 253, 9 },
		{ 262, 254, 9 },
		{ 262, 255, 9 },
		{ 262, 256, 9 },
		{ 263, 252, 9 },
		{ 263, 253, 9 },
		{ 263, 254, 9 },
		{ 263, 255, 9 },
		{ 263, 256, 9 },
		{ 264, 251, 9 },
		{ 264, 252, 9 },
		{ 264, 253, 9 },
		{ 264, 254, 9 },
		{ 264, 255, 9 },
		{ 264, 256, 9 },
		{ 265, 250, 9 },
		{ 265, 251, 9 },
		{ 265, 252, 9 },
		{ 265, 253, 9 },
		{ 265, 254, 9 },
		{ 265, 255, 9 },
		{ 265, 256, 9 },
		{ 266, 249, 9 },
		{ 266, 250, 9 },
		{ 266, 251, 9 },
		{ 266, 252, 9 },
		{ 266, 253, 9 },
		{ 266, 254, 9 },
		{ 266, 255, 9 },
		{ 266, 256, 9 },
		{ 267, 248, 9 },
		{ 267, 249, 9 },
		{ 267, 250, 9 },
		{ 267, 251, 9 },
		{ 267, 252, 9 },
		{ 267, 253, 9 },
		{ 267, 254, 9 },
		{ 267, 255, 9 },
		{ 267, 256, 9 },
		{ 268, 247, 9 },
		{ 268, 248, 9 },
		{ 268, 249, 9 },
		{ 268, 250, 9 },
		{ 268, 251, 9 },
		{ 268, 252, 9 },
		{ 268, 253, 9 },
		{ 268, 254, 9 },
		{ 268, 255, 9 },
		{ 268, 256, 9 },
		{ 269, 246, 9 },
		{ 269, 247, 9 },
		{ 269, 248, 9 },
		{ 269, 249, 9 },
		{ 269, 250, 9 },
		{ 269, 251, 9 },
		{ 269, 252, 9 },
		{ 269, 253, 9 },
		{ 269, 254, 9 },
		{ 269, 255, 9 },
		{ 269, 256, 9 },
		{ 270, 245, 9 },
		{ 270, 246, 9 },
		{ 270, 247, 9 },
		{ 270, 248, 9 },
		{ 270, 249, 9 },
		{ 270, 250, 9 },
		{ 270, 251, 9 },
		{ 270, 252, 9 },
		{ 270, 253, 9 },
		{ 270, 254, 9 },
		{ 270, 255, 9 },
		{ 270, 256, 9 },
		{ 271, 244, 9 },
		{ 271, 245, 9 },
		{ 271, 246, 9 },
		{ 271, 247, 9 },
		{ 271, 248, 9 },
		{ 271, 249, 9 },
		{ 271, 250, 9 },
		{ 271, 251, 9 },
		{ 271, 252, 9 },
		{ 271, 253, 9 },
		{ 271, 254, 9 },
		{ 271, 255, 9 },
		{ 271, 256, 9 },
		{ 272, 243, 9 },
		{ 272, 244, 9 },
		{ 272, 245, 9 },
		{ 272, 246, 9 },
		{ 272, 247, 9 },
		{ 272, 248, 9 },
		{ 272, 249, 9 },
		{ 272, 250, 9 },
		{ 272, 251, 9 },
		{ 272, 252, 9 },
		{ 272, 253, 9 },
		{ 272, 254, 9 },
		{ 272, 255, 9 },
		{ 272, 256, 9 },
		{ 273, 242, 9 },
		{ 273, 243, 9 },
		{ 273, 244, 9 },
		{ 273, 245, 9 },
		{ 273, 246, 9 },
		{ 273, 247, 9 },
		{ 273, 248, 9 },
		{ 273, 249, 9 },
		{ 273, 250, 9 },
		{ 273, 251, 9 },
		{ 273, 252, 9 },
		{ 273, 253, 9 },
		{ 273, 254, 9 },
		{ 273, 255, 9 },
		{ 273, 256, 9 },
		{ 274, 241, 9 },
		{ 274, 242, 9 },
		{ 274, 243, 9 },
		{ 274, 244, 9 },
		{ 274, 245, 9 },
		{ 274, 246, 9 },
		{ 274, 247, 9 },
		{ 274, 248, 9 },
		{ 274, 249, 9 },
		{ 274, 250, 9 },
		{ 274, 251, 9 },
		{ 274, 252, 9 },
		{ 274, 253, 9 },
		{ 274, 254, 9 },
		{ 274, 255, 9 },
		{ 274, 256, 9 },
		{ 275, 240, 9 },
		{ 275, 241, 9 },
		{ 275, 242, 9 },
		{ 275, 243, 9 },
		{ 275, 244, 9 },
		{ 275, 245, 9 },
		{ 275, 246, 9 },
		{ 275, 247, 9 },
		{ 275, 248, 9 },
		{ 275, 249, 9 },
		{ 275, 250, 9 },
		{ 275, 251, 9 },
		{ 275, 252, 9 },
		{ 275, 253, 9 },
		{ 275, 254, 9 },
		{ 275, 255, 9 },
		{ 275, 256, 9 },
		{ 276, 239, 9 },
		{ 276, 240, 9 },
		{ 276, 241, 9 },
		{ 276, 242, 9 },
		{ 276, 243, 9 },
		{ 276, 244, 9 },
		{ 276, 245, 9 },
		{ 276, 246, 9 },
		{ 276, 247, 9 },
		{ 276, 248, 9 },
		{ 276, 249, 9 },
		{ 276, 250, 9 },
		{ 276, 251, 9 },
		{ 276, 252, 9 },
		{ 276, 253, 9 },
		{ 276, 254, 9 },
		{ 276, 255, 9 },
		{ 276, 256, 9 },
		{ 277, 238, 9 },
		{ 277, 239, 9 },
		{ 277, 240, 9 },
		{ 277, 241, 9 },
		{ 277, 242, 9 },
		{ 277, 243, 9 },
		{ 277, 244, 9 },
		{ 277, 245, 9 },
		{ 277, 246, 9 },
		{ 277, 247, 9 },
		{ 277, 248, 9 },
		{ 277, 249, 9 },
		{ 277, 250, 9 },
		{ 277, 251, 9 },
		{ 277, 252, 9 },
		{ 277, 253, 9 },
		{ 277, 254, 9 },
		{ 277, 255, 9 },
		{ 277, 256, 9 },
		{ 278, 237, 9 },
		{ 278, 238, 9 },
		{ 278, 239, 9 },
		{ 278, 240, 9 },
		{ 278, 241, 9 },
		{ 278, 242, 9 },
		{ 278, 243, 9 },
		{ 278, 244, 9 },
		{ 278, 245, 9 },
		{ 278, 246, 9 },
		{ 278, 247, 9 },
		{ 278, 248, 9 },
		{ 278, 249, 9 },
		{ 278, 250, 9 },
		{ 278, 251, 9 },
		{ 278, 252, 9 },
		{ 278, 253, 9 },
		{ 278, 254, 9 },
		{ 278, 255, 9 },
		{ 278, 256, 9 },
		{ 279, 236, 9 },
		{ 279, 237, 9 },
		{ 279, 238, 9 },
		{ 279, 239, 9 },
		{ 279, 240, 9 },
		{ 279, 241, 9 },
		{ 279, 242, 9 },
		{ 279, 243, 9 },
		{ 279, 244, 9 },
		{ 279, 245, 9 },
		{ 279, 246, 9 },
		{ 279, 247, 9 },
		{ 279, 248, 9 },
		{ 279, 249, 9 },
		{ 279, 250, 9 },
		{ 279, 251, 9 },
		{ 279, 252, 9 },
		{ 279, 253, 9 },
		{ 279, 254, 9 },
		{ 279, 255, 9 },
		{ 279, 256, 9 },
		{ 280, 235, 9 },
		{ 280, 236, 9 },
		{ 280, 237, 9 },
		{ 280, 238, 9 },
		{ 280, 239, 9 },
		{ 280, 240, 9 },
		{ 280, 241, 9 },
		{ 280, 242, 9 },
		{ 280, 243, 9 },
		{ 280, 244, 9 },
		{ 280, 245, 9 },
		{ 280, 246, 9 },
		{ 280, 247, 9 },
		{ 280, 248, 9 },
		{ 280, 249, 9 },
		{ 280, 250, 9 },
		{ 280, 251, 9 },
		{ 280, 252, 9 },
		{ 280, 253, 9 },
		{ 280, 254, 9 },
		{ 280, 255, 9 },
		{ 280, 256, 9 },
		{ 281, 234, 9 },
		{ 281, 235, 9 },
		{ 281, 236, 9 },
		{ 281, 237, 9 },
		{ 281, 238, 9 },
		{ 281, 239, 9 },
		{ 281, 240, 9 },
		{ 281, 241, 9 },
		{ 281, 242, 9 },
		{ 281, 243, 9 },
		{ 281, 244, 9 },
		{ 281, 245, 9 },
		{ 281, 246, 9 },
		{ 281, 247, 9 },
		{ 281, 248, 9 },
		{ 281, 249, 9 },
		{ 281, 250, 9 },
		{ 281, 251, 9 },
		{ 281, 252, 9 },
		{ 281, 253, 9 },
		{ 281, 254, 9 },
		{ 281, 255, 9 },
		{ 281, 256, 9 },
		{ 282, 233, 9 },
		{ 282, 234, 9 },
		{ 282, 235, 9 },
		{ 282, 236, 9 },
		{ 282, 237, 9 },
		{ 282, 238, 9 },
		{ 282, 239, 9 },
		{ 282, 240, 9 },
		{ 282, 241, 9 },
		{ 282, 242, 9 },
		{ 282, 243, 9 },
		{ 282, 244, 9 },
		{ 282, 245, 9 },
		{ 282, 246, 9 },
		{ 282, 247, 9 },
		{ 282, 248, 9 },
		{ 282, 249, 9 },
		{ 282, 250, 9 },
		{ 282, 251, 9 },
		{ 282, 252, 9 },
		{ 282, 253, 9 },
		{ 282, 254, 9 },
		{ 282, 255, 9 },
		{ 282, 256, 9 },
		{ 283, 232, 9 },
		{ 283, 233, 9 },
		{ 283, 234, 9 },
		{ 283, 235, 9 },
		{ 283, 236, 9 },
		{ 283, 237, 9 },
		{ 283, 238, 9 },
		{ 283, 239, 9 },
		{ 283, 240, 9 },
		{ 283, 241, 9 },
		{ 283, 242, 9 },
		{ 283, 243, 9 },
		{ 283, 244, 9 },
		{ 283, 245, 9 },
		{ 283, 246, 9 },
		{ 283, 247, 9 },
		{ 283, 248, 9 },
		{ 283, 249, 9 },
		{ 283, 250, 9 },
		{ 283, 251, 9 },
		{ 283, 252, 9 },
		{ 283, 253, 9 },
		{ 283, 254, 9 },
		{ 283, 255, 9 },
		{ 283, 256, 9 },
		{ 284, 231, 9 },
		{ 284, 232, 9 },
		{ 284, 233, 9 },
		{ 284, 234, 9 },
		{ 284, 235, 9 },
		{ 284, 236, 9 },
		{ 284, 237, 9 },
		{ 284, 238, 9 },
		{ 284, 239, 9 },
		{ 284, 240, 9 },
		{ 284, 241, 9 },
		{ 284, 242, 9 },
		{ 284, 243, 9 },
		{ 284, 244, 9 },
		{ 284, 245, 9 },
		{ 284, 246, 9 },
		{ 284, 247, 9 },
		{ 284, 248, 9 },
		{ 284, 249, 9 },
		{ 284, 250, 9 },
		{ 284, 251, 9 },
		{ 284, 252, 9 },
		{ 284, 253, 9 },
		{ 284, 254, 9 },
		{ 284, 255, 9 },
		{ 284, 256, 9 },
		{ 285, 230, 9 },
		{ 285, 231, 9 },
		{ 285, 232, 9 },
		{ 285, 233, 9 },
		{ 285, 234, 9 },
		{ 285, 235, 9 },
		{ 285, 236, 9 },
		{ 285, 237, 9 },
		{ 285, 238, 9 },
		{ 285, 239, 9 },
		{ 285, 240, 9 },
		{ 285, 241, 9 },
		{ 285, 242, 9 },
		{ 285, 243, 9 },
		{ 285, 244, 9 },
		{ 285, 245, 9 },
		{ 285, 246, 9 },
		{ 285, 247, 9 },
		{ 285, 248, 9 },
		{ 285, 249, 9 },
		{ 285, 250, 9 },
		{ 285, 251, 9 },
		{ 285, 252, 9 },
		{ 285, 253, 9 },
		{ 285, 254, 9 },
		{ 285, 255, 9 },
		{ 285, 256, 9 },
		{ 286, 229, 9 },
		{ 286, 230, 9 },
		{ 286, 231, 9 },
		{ 286, 232, 9 },
		{ 286, 233, 9 },
		{ 286, 234, 9 },
		{ 286, 235, 9 },
		{ 286, 236, 9 },
		{ 286, 237, 9 },
		{ 286, 238, 9 },
		{ 286, 239, 9 },
		{ 286, 240, 9 },
		{ 286, 241, 9 },
		{ 286, 242, 9 },
		{ 286, 243, 9 },
		{ 286, 244, 9 },
		{ 286, 245, 9 },
		{ 286, 246, 9 },
		{ 286, 247, 9 },
		{ 286, 248, 9 },
		{ 286, 249, 9 },
		{ 286, 250, 9 },
		{ 286, 251, 9 },
		{ 286, 252, 9 },
		{ 286, 253, 9 },
		{ 286, 254, 9 },
		{ 286, 255, 9 },
		{ 286, 256, 9 },
		{ 287, 228, 9 },
		{ 287, 229, 9 },
		{ 287, 230, 9 },
		{ 287, 231, 9 },
		{ 287, 232, 9 },
		{ 287, 233, 9 },
		{ 287, 234, 9 },
		{ 287, 235, 9 },
		{ 287, 236, 9 },
		{ 287, 237, 9 },
		{ 287, 238, 9 },
		{ 287, 239, 9 },
		{ 287, 240, 9 },
		{ 287, 241, 9 },
		{ 287, 242, 9 },
		{ 287, 243, 9 },
		{ 287, 244, 9 },
		{ 287, 245, 9 },
		{ 287, 246, 9 },
		{ 287, 247, 9 },
		{ 287, 248, 9 },
		{ 287, 249, 9 },
		{ 287, 250, 9 },
		{ 287, 251, 9 },
		{ 287, 252, 9 },
		{ 287, 253, 9 },
		{ 287, 254, 9 },
		{ 287, 255, 9 },
		{ 287, 256, 9 },
		{ 288, 227, 9 },
		{ 288, 228, 9 },
		{ 288, 229, 9 },
		{ 288, 230, 9 },
		{ 288, 231, 9 },
		{ 288, 232, 9 },
		{ 288, 233, 9 },
		{ 288, 234, 9 },
		{ 288, 235, 9 },
		{ 288, 236, 9 },
		{ 288, 237, 9 },
		{ 288, 238, 9 },
		{ 288, 239, 9 },
		{ 288, 240, 9 },
		{ 288, 241, 9 },
		{ 288, 242, 9 },
		{ 288, 243, 9 },
		{ 288, 244, 9 },
		{ 288, 245, 9 },
		{ 288, 246, 9 },
		{ 288, 247, 9 },
		{ 288, 248, 9 },
		{ 288, 249, 9 },
		{ 288, 250, 9 },
		{ 288, 251, 9 },
		{ 288, 252, 9 },
		{ 288, 253, 9 },
		{ 288, 254, 9 },
		{ 288, 255, 9 },
		{ 288, 256, 9 },
		{ 226, 243, 10 },
		{ 226, 244, 10 },
		{ 226, 245, 10 },
		{ 226, 246, 10 },
		{ 226, 247, 10 },
		{ 226, 248, 10 },
		{ 226, 249, 10 },
		{ 226, 250, 10 },
		{ 226, 251, 10 },
		{ 226, 252, 10 },
		{ 226, 253, 10 },
		{ 226, 254, 10 },
		{ 226, 255, 10 },
		{ 226, 256, 10 },
		{ 226, 257, 10 },
		{ 226, 258, 10 },
		{ 226, 259, 10 },
		{ 226, 260, 10 },
		{ 226, 261, 10 },
		{ 226, 262, 10 },
		{ 226, 263, 10 },
		{ 226, 264, 10 },
		{ 226, 265, 10 },
		{ 226, 266, 10 },
		{ 226, 267, 10 },
		{ 226, 268, 10 },
		{ 226, 269, 10 },
		{ 226, 270, 10 },
		{ 226, 271, 10 },
		{ 227, 243, 10 },
		{ 227, 244, 10 },
		{ 227, 245, 10 },
		{ 227, 246, 10 },
		{ 227, 247, 10 },
		{ 227, 248, 10 },
		{ 227, 249, 10 },
		{ 227, 250, 10 },
		{ 227, 251, 10 },
		{ 227, 252, 10 },
		{ 227, 253, 10 },
		{ 227, 254, 10 },
		{ 227, 255, 10 },
		{ 227, 256, 10 },
		{ 227, 257, 10 },
		{ 227, 258, 10 },
		{ 227, 259, 10 },
		{ 227, 260, 10 },
		{ 227, 261, 10 },
		{ 227, 262, 10 },
		{ 227, 263, 10 },
		{ 227, 264, 10 },
		{ 227, 265, 10 },
		{ 227, 266, 10 },
		{ 227, 267, 10 },
		{ 227, 268, 10 },
		{ 227, 269, 10 },
		{ 227, 270, 10 },
		{ 227, 271, 10 },
		{ 228, 243, 10 },
		{ 228, 244, 10 },
		{ 228, 245, 10 },
		{ 228, 246, 10 },
		{ 228, 247, 10 },
		{ 228, 248, 10 },
		{ 228, 249, 10 },
		{ 228, 250, 10 },
		{ 228, 251, 10 },
		{ 228, 252, 10 },
		{ 228, 253, 10 },
		{ 228, 254, 10 },
		{ 228, 255, 10 },
		{ 228, 256, 10 },
		{ 228, 257, 10 },
		{ 228, 258, 10 },
		{ 228, 259, 10 },
		{ 228, 260, 10 },
		{ 228, 261, 10 },
		{ 228, 262, 10 },
		{ 228, 263, 10 },
		{ 228, 264, 10 },
		{ 228, 265, 10 },
		{ 228, 266, 10 },
		{ 228, 267, 10 },
		{ 228, 268, 10 },
		{ 228, 269, 10 },
		{ 228, 270, 10 },
		{ 228, 271, 10 },
		{ 229, 244, 10 },
		{ 229, 245, 10 },
		{ 229, 246, 10 },
		{ 229, 247, 10 },
		{ 229, 248, 10 },
		{ 229, 249, 10 },
		{ 229, 250, 10 },
		{ 229, 251, 10 },
		{ 229, 252, 10 },
		{ 229, 253, 10 },
		{ 229, 254, 10 },
		{ 229, 255, 10 },
		{ 229, 256, 10 },
		{ 229, 257, 10 },
		{ 229, 258, 10 },
		{ 229, 259, 10 },
		{ 229, 260, 10 },
		{ 229, 261, 10 },
		{ 229, 262, 10 },
		{ 229, 263, 10 },
		{ 229, 264, 10 },
		{ 229, 265, 10 },
		{ 229, 266, 10 },
		{ 229, 267, 10 },
		{ 229, 268, 10 },
		{ 229, 269, 10 },
		{ 229, 270, 10 },
		{ 230, 244, 10 },
		{ 230, 245, 10 },
		{ 230, 246, 10 },
		{ 230, 247, 10 },
		{ 230, 248, 10 },
		{ 230, 249, 10 },
		{ 230, 250, 10 },
		{ 230, 251, 10 },
		{ 230, 252, 10 },
		{ 230, 253, 10 },
		{ 230, 254, 10 },
		{ 230, 255, 10 },
		{ 230, 256, 10 },
		{ 230, 257, 10 },
		{ 230, 258, 10 },
		{ 230, 259, 10 },
		{ 230, 260, 10 },
		{ 230, 261, 10 },
		{ 230, 262, 10 },
		{ 230, 263, 10 },
		{ 230, 264, 10 },
		{ 230, 265, 10 },
		{ 230, 266, 10 },
		{ 230, 267, 10 },
		{ 230, 268, 10 },
		{ 230, 269, 10 },
		{ 230, 270, 10 },
		{ 231, 245, 10 },
		{ 231, 246, 10 },
		{ 231, 247, 10 },
		{ 231, 248, 10 },
		{ 231, 249, 10 },
		{ 231, 250, 10 },
		{ 231, 251, 10 },
		{ 231, 252, 10 },
		{ 231, 253, 10 },
		{ 231, 254, 10 },
		{ 231, 255, 10 },
		{ 231, 256, 10 },
		{ 231, 257, 10 },
		{ 231, 258, 10 },
		{ 231, 259, 10 },
		{ 231, 260, 10 },
		{ 231, 261, 10 },
		{ 231, 262, 10 },
		{ 231, 263, 10 },
		{ 231, 264, 10 },
		{ 231, 265, 10 },
		{ 231, 266, 10 },
		{ 231, 267, 10 },
		{ 231, 268, 10 },
		{ 231, 269, 10 },
		{ 232, 245, 10 },
		{ 232, 246, 10 },
		{ 232, 247, 10 },
		{ 232, 248, 10 },
		{ 232, 249, 10 },
		{ 232, 250, 10 },
		{ 232, 251, 10 },
		{ 232, 252, 10 },
		{ 232, 253, 10 },
		{ 232, 254, 10 },
		{ 232, 255, 10 },
		{ 232, 256, 10 },
		{ 232, 257, 10 },
		{ 232, 258, 10 },
		{ 232, 259, 10 },
		{ 232, 260, 10 },
		{ 232, 261, 10 },
		{ 232, 262, 10 },
		{ 232, 263, 10 },
		{ 232, 264, 10 },
		{ 232, 265, 10 },
		{ 232, 266, 10 },
		{ 232, 267, 10 },
		{ 232, 268, 10 },
		{ 232, 269, 10 },
		{ 233, 246, 10 },
		{ 233, 247, 10 },
		{ 233, 248, 10 },
		{ 233, 249, 10 },
		{ 233, 250, 10 },
		{ 233, 251, 10 },
		{ 233, 252, 10 },
		{ 233, 253, 10 },
		{ 233, 254, 10 },
		{ 233, 255, 10 },
		{ 233, 256, 10 },
		{ 233, 257, 10 },
		{ 233, 258, 10 },
		{ 233, 259, 10 },
		{ 233, 260, 10 },
		{ 233, 261, 10 },
		{ 233, 262, 10 },
		{ 233, 263, 10 },
		{ 233, 264, 10 },
		{ 233, 265, 10 },
		{ 233, 266, 10 },
		{ 233, 267, 10 },
		{ 233, 268, 10 },
		{ 234, 246, 10 },
		{ 234, 247, 10 },
		{ 234, 248, 10 },
		{ 234, 249, 10 },
		{ 234, 250, 10 },
		{ 234, 251, 10 },
		{ 234, 252, 10 },
		{ 234, 253, 10 },
		{ 234, 254, 10 },
		{ 234, 255, 10 },
		{ 234, 256, 10 },
		{ 234, 257, 10 },
		{ 234, 258, 10 },
		{ 234, 259, 10 },
		{ 234, 260, 10 },
		{ 234, 261, 10 },
		{ 234, 262, 10 },
		{ 234, 263, 10 },
		{ 234, 264, 10 },
		{ 234, 265, 10 },
		{ 234, 266, 10 },
		{ 234, 267, 10 },
		{ 234, 268, 10 },
		{ 235, 247, 10 },
		{ 235, 248, 10 },
		{ 235, 249, 10 },
		{ 235, 250, 10 },
		{ 235, 251, 10 },
		{ 235, 252, 10 },
		{ 235, 253, 10 },
		{ 235, 254, 10 },
		{ 235, 255, 10 },
		{ 235, 256, 10 },
		{ 235, 257, 10 },
		{ 235, 258, 10 },
		{ 235, 259, 10 },
		{ 235, 260, 10 },
		{ 235, 261, 10 },
		{ 235, 262, 10 },
		{ 235, 263, 10 },
		{ 235, 264, 10 },
		{ 235, 265, 10 },
		{ 235, 266, 10 },
		{ 235, 267, 10 },
		{ 236, 247, 10 },
		{ 236, 248, 10 },
		{ 236, 249, 10 },
		{ 236, 250, 10 },
		{ 236, 251, 10 },
		{ 236, 252, 10 },
		{ 236, 253, 10 },
		{ 236, 254, 10 },
		{ 236, 255, 10 },
		{ 236, 256, 10 },
		{ 236, 257, 10 },
		{ 236, 258, 10 },
		{ 236, 259, 10 },
		{ 236, 260, 10 },
		{ 236, 261, 10 },
		{ 236, 262, 10 },
		{ 236, 263, 10 },
		{ 236, 264, 10 },
		{ 236, 265, 10 },
		{ 236, 266, 10 },
		{ 236, 267, 10 },
		{ 237, 248, 10 },
		{ 237, 249, 10 },
		{ 237, 250, 10 },
		{ 237, 251, 10 },
		{ 237, 252, 10 },
		{ 237, 253, 10 },
		{ 237, 254, 10 },
		{ 237, 255, 10 },
		{ 237, 256, 10 },
		{ 237, 257, 10 },
		{ 237, 258, 10 },
		{ 237, 259, 10 },
		{ 237, 260, 10 },
		{ 237, 261, 10 },
		{ 237, 262, 10 },
		{ 237, 263, 10 },
		{ 237, 264, 10 },
		{ 237, 265, 10 },
		{ 237, 266, 10 },
		{ 238, 248, 10 },
		{ 238, 249, 10 },
		{ 238, 250, 10 },
		{ 238, 251, 10 },
		{ 238, 252, 10 },
		{ 238, 253, 10 },
		{ 238, 254, 10 },
		{ 238, 255, 10 },
		{ 238, 256, 10 },
		{ 238, 257, 10 },
		{ 238, 258, 10 },
		{ 238, 259, 10 },
		{ 238, 260, 10 },
		{ 238, 261, 10 },
		{ 238, 262, 10 },
		{ 238, 263, 10 },
		{ 238, 264, 10 },
		{ 238, 265, 10 },
		{ 238, 266, 10 },
		{ 239, 249, 10 },
		{ 239, 250, 10 },
		{ 239, 251, 10 },
		{ 239, 252, 10 },
		{ 239, 253, 10 },
		{ 239, 254, 10 },
		{ 239, 255, 10 },
		{ 239, 256, 10 },
		{ 239, 257, 10 },
		{ 239, 258, 10 },
		{ 239, 259, 10 },
		{ 239, 260, 10 },
		{ 239, 261, 10 },
		{ 239, 262, 10 },
		{ 239, 263, 10 },
		{ 239, 264, 10 },
		{ 239, 265, 10 },
		{ 240, 249, 10 },
		{ 240, 250, 10 },
		{ 240, 251, 10 },
		{ 240, 252, 10 },
		{ 240, 253, 10 },
		{ 240, 254, 10 },
		{ 240, 255, 10 },
		{ 240, 256, 10 },
		{ 240, 257, 10 },
		{ 240, 258, 10 },
		{ 240, 259, 10 },
		{ 240, 260, 10 },
		{ 240, 261, 10 },
		{ 240, 262, 10 },
		{ 240, 263, 10 },
		{ 240, 264, 10 },
		{ 240, 265, 10 },
		{ 241, 250, 10 },
		{ 241, 251, 10 },
		{ 241, 252, 10 },
		{ 241, 253, 10 },
		{ 241, 254, 10 },
		{ 241, 255, 10 },
		{ 241, 256, 10 },
		{ 241, 257, 10 },
		{ 241, 258, 10 },
		{ 241, 259, 10 },
		{ 241, 260, 10 },
		{ 241, 261, 10 },
		{ 241, 262, 10 },
		{ 241, 263, 10 },
		{ 241, 264, 10 },
		{ 242, 250, 10 },
		{ 242, 251, 10 },
		{ 242, 252, 10 },
		{ 242, 253, 10 },
		{ 242, 254, 10 },
		{ 242, 255, 10 },
		{ 242, 256, 10 },
		{ 242, 257, 10 },
		{ 242, 258, 10 },
		{ 242, 259, 10 },
		{ 242, 260, 10 },
		{ 242, 261, 10 },
		{ 242, 262, 10 },
		{ 242, 263, 10 },
		{ 242, 264, 10 },
		{ 243, 251, 10 },
		{ 243, 252, 10 },
		{ 243, 253, 10 },
		{ 243, 254, 10 },
		{ 243, 255, 10 },
		{ 243, 256, 10 },
		{ 243, 257, 10 },
		{ 243, 258, 10 },
		{ 243, 259, 10 },
		{ 243, 260, 10 },
		{ 243, 261, 10 },
		{ 243, 262, 10 },
		{ 243, 263, 10 },
		{ 244, 251, 10 },
		{ 244, 252, 10 },
		{ 244, 253, 10 },
		{ 244, 254, 10 },
		{ 244, 255, 10 },
		{ 244, 256, 10 },
		{ 244, 257, 10 },
		{ 244, 258, 10 },
		{ 244, 259, 10 },
		{ 244, 260, 10 },
		{ 244, 261, 10 },
		{ 244, 262, 10 },
		{ 244, 263, 10 },
		{ 245, 252, 10 },
		{ 245, 253, 10 },
		{ 245, 254, 10 },
		{ 245, 255, 10 },
		{ 245, 256, 10 },
		{ 245, 257, 10 },
		{ 245, 258, 10 },
		{ 245, 259, 10 },
		{ 245, 260, 10 },
		{ 245, 261, 10 },
		{ 245, 262, 10 },
		{ 246, 252, 10 },
		{ 246, 253, 10 },
		{ 246, 254, 10 },
		{ 246, 255, 10 },
		{ 246, 256, 10 },
		{ 246, 257, 10 },
		{ 246, 258, 10 },
		{ 246, 259, 10 },
		{ 246, 260, 10 },
		{ 246, 261, 10 },
		{ 246, 262, 10 },
		{ 247, 253, 10 },
		{ 247, 254, 10 },
		{ 247, 255, 10 },
		{ 247, 256, 10 },
		{ 247, 257, 10 },
		{ 247, 258, 10 },
		{ 247, 259, 10 },
		{ 247, 260, 10 },
		{ 247, 261, 10 },
		{ 248, 253, 10 },
		{ 248, 254, 10 },
		{ 248, 255, 10 },
		{ 248, 256, 10 },
		{ 248, 257, 10 },
		{ 248, 258, 10 },
		{ 248, 259, 10 },
		{ 248, 260, 10 },
		{ 248, 261, 10 },
		{ 249, 254, 10 },
		{ 249, 255, 10 },
		{ 249, 256, 10 },
		{ 249, 257, 10 },
		{ 249, 258, 10 },
		{ 249, 259, 10 },
		{ 249, 260, 10 },
		{ 250, 254, 10 },
		{ 250, 255, 10 },
		{ 250, 256, 10 },
		{ 250, 257, 10 },
		{ 250, 258, 10 },
		{ 250, 259, 10 },
		{ 250, 260, 10 },
		{ 251, 255, 10 },
		{ 251, 256, 10 },
		{ 251, 257, 10 },
		{ 251, 258, 10 },
		{ 251, 259, 10 },
		{ 252, 255, 10 },
		{ 252, 256, 10 },
		{ 252, 257, 10 },
		{ 252, 258, 10 },
		{ 252, 259, 10 },
		{ 262, 255, 10 },
		{ 262, 256, 10 },
		{ 262, 257, 10 },
		{ 262, 258, 10 },
		{ 262, 259, 10 },
		{ 263, 255, 10 },
		{ 263, 256, 10 },
		{ 263, 257, 10 },
		{ 263, 258, 10 },
		{ 263, 259, 10 },
		{ 264, 254, 10 },
		{ 264, 255, 10 },
		{ 264, 256, 10 },
		{ 264, 257, 10 },
		{ 264, 258, 10 },
		{ 264, 259, 10 },
		{ 264, 260, 10 },
		{ 265, 254, 10 },
		{ 265, 255, 10 },
		{ 265, 256, 10 },
		{ 265, 257, 10 },
		{ 265, 258, 10 },
		{ 265, 259, 10 },
		{ 265, 260, 10 },
		{ 266, 253, 10 },
		{ 266, 254, 10 },
		{ 266, 255, 10 },
		{ 266, 256, 10 },
		{ 266, 257, 10 },
		{ 266, 258, 10 },
		{ 266, 259, 10 },
		{ 266, 260, 10 },
		{ 266, 261, 10 },
		{ 267, 253, 10 },
		{ 267, 254, 10 },
		{ 267, 255, 10 },
		{ 267, 256, 10 },
		{ 267, 257, 10 },
		{ 267, 258, 10 },
		{ 267, 259, 10 },
		{ 267, 260, 10 },
		{ 267, 261, 10 },
		{ 268, 252, 10 },
		{ 268, 253, 10 },
		{ 268, 254, 10 },
		{ 268, 255, 10 },
		{ 268, 256, 10 },
		{ 268, 257, 10 },
		{ 268, 258, 10 },
		{ 268, 259, 10 },
		{ 268, 260, 10 },
		{ 268, 261, 10 },
		{ 268, 262, 10 },
		{ 269, 252, 10 },
		{ 269, 253, 10 },
		{ 269, 254, 10 },
		{ 269, 255, 10 },
		{ 269, 256, 10 },
		{ 269, 257, 10 },
		{ 269, 258, 10 },
		{ 269, 259, 10 },
		{ 269, 260, 10 },
		{ 269, 261, 10 },
		{ 269, 262, 10 },
		{ 270, 251, 10 },
		{ 270, 252, 10 },
		{ 270, 253, 10 },
		{ 270, 254, 10 },
		{ 270, 255, 10 },
		{ 270, 256, 10 },
		{ 270, 257, 10 },
		{ 270, 258, 10 },
		{ 270, 259, 10 },
		{ 270, 260, 10 },
		{ 270, 261, 10 },
		{ 270, 262, 10 },
		{ 270, 263, 10 },
		{ 271, 251, 10 },
		{ 271, 252, 10 },
		{ 271, 253, 10 },
		{ 271, 254, 10 },
		{ 271, 255, 10 },
		{ 271, 256, 10 },
		{ 271, 257, 10 },
		{ 271, 258, 10 },
		{ 271, 259, 10 },
		{ 271, 260, 10 },
		{ 271, 261, 10 },
		{ 271, 262, 10 },
		{ 271, 263, 10 },
		{ 272, 250, 10 },
		{ 272, 251, 10 },
		{ 272, 252, 10 },
		{ 272, 253, 10 },
		{ 272, 254, 10 },
		{ 272, 255, 10 },
		{ 272, 256, 10 },
		{ 272, 257, 10 },
		{ 272, 258, 10 },
		{ 272, 259, 10 },
		{ 272, 260, 10 },
		{ 272, 261, 10 },
		{ 272, 262, 10 },
		{ 272, 263, 10 },
		{ 272, 264, 10 },
		{ 273, 250, 10 },
		{ 273, 251, 10 },
		{ 273, 252, 10 },
		{ 273, 253, 10 },
		{ 273, 254, 10 },
		{ 273, 255, 10 },
		{ 273, 256, 10 },
		{ 273, 257, 10 },
		{ 273, 258, 10 },
		{ 273, 259, 10 },
		{ 273, 260, 10 },
		{ 273, 261, 10 },
		{ 273, 262, 10 },
		{ 273, 263, 10 },
		{ 273, 264, 10 },
		{ 274, 249, 10 },
		{ 274, 250, 10 },
		{ 274, 251, 10 },
		{ 274, 252, 10 },
		{ 274, 253, 10 },
		{ 274, 254, 10 },
		{ 274, 255, 10 },
		{ 274, 256, 10 },
		{ 274, 257, 10 },
		{ 274, 258, 10 },
		{ 274, 259, 10 },
		{ 274, 260, 10 },
		{ 274, 261, 10 },
		{ 274, 262, 10 },
		{ 274, 263, 10 },
		{ 274, 264, 10 },
		{ 274, 265, 10 },
		{ 275, 249, 10 },
		{ 275, 250, 10 },
		{ 275, 251, 10 },
		{ 275, 252, 10 },
		{ 275, 253, 10 },
		{ 275, 254, 10 },
		{ 275, 255, 10 },
		{ 275, 256, 10 },
		{ 275, 257, 10 },
		{ 275, 258, 10 },
		{ 275, 259, 10 },
		{ 275, 260, 10 },
		{ 275, 261, 10 },
		{ 275, 262, 10 },
		{ 275, 263, 10 },
		{ 275, 264, 10 },
		{ 275, 265, 10 },
		{ 276, 248, 10 },
		{ 276, 249, 10 },
		{ 276, 250, 10 },
		{ 276, 251, 10 },
		{ 276, 252, 10 },
		{ 276, 253, 10 },
		{ 276, 254, 10 },
		{ 276, 255, 10 },
		{ 276, 256, 10 },
		{ 276, 257, 10 },
		{ 276, 258, 10 },
		{ 276, 259, 10 },
		{ 276, 260, 10 },
		{ 276, 261, 10 },
		{ 276, 262, 10 },
		{ 276, 263, 10 },
		{ 276, 264, 10 },
		{ 276, 265, 10 },
		{ 276, 266, 10 },
		{ 277, 248, 10 },
		{ 277, 249, 10 },
		{ 277, 250, 10 },
		{ 277, 251, 10 },
		{ 277, 252, 10 },
		{ 277, 253, 10 },
		{ 277, 254, 10 },
		{ 277, 255, 10 },
		{ 277, 256, 10 },
		{ 277, 257, 10 },
		{ 277, 258, 10 },
		{ 277, 259, 10 },
		{ 277, 260, 10 },
		{ 277, 261, 10 },
		{ 277, 262, 10 },
		{ 277, 263, 10 },
		{ 277, 264, 10 },
		{ 277, 265, 10 },
		{ 277, 266, 10 },
		{ 278, 247, 10 },
		{ 278, 248, 10 },
		{ 278, 249, 10 },
		{ 278, 250, 10 },
		{ 278, 251, 10 },
		{ 278, 252, 10 },
		{ 278, 253, 10 },
		{ 278, 254, 10 },
		{ 278, 255, 10 },
		{ 278, 256, 10 },
		{ 278, 257, 10 },
		{ 278, 258, 10 },
		{ 278, 259, 10 },
		{ 278, 260, 10 },
		{ 278, 261, 10 },
		{ 278, 262, 10 },
		{ 278, 263, 10 },
		{ 278, 264, 10 },
		{ 278, 265, 10 },
		{ 278, 266, 10 },
		{ 278, 267, 10 },
		{ 279, 247, 10 },
		{ 279, 248, 10 },
		{ 279, 249, 10 },
		{ 279, 250, 10 },
		{ 279, 251, 10 },
		{ 279, 252, 10 },
		{ 279, 253, 10 },
		{ 279, 254, 10 },
		{ 279, 255, 10 },
		{ 279, 256, 10 },
		{ 279, 257, 10 },
		{ 279, 258, 10 },
		{ 279, 259, 10 },
		{ 279, 260, 10 },
		{ 279, 261, 10 },
		{ 279, 262, 10 },
		{ 279, 263, 10 },
		{ 279, 264, 10 },
		{ 279, 265, 10 },
		{ 279, 266, 10 },
		{ 279, 267, 10 },
		{ 280, 246, 10 },
		{ 280, 247, 10 },
		{ 280, 248, 10 },
		{ 280, 249, 10 },
		{ 280, 250, 10 },
		{ 280, 251, 10 },
		{ 280, 252, 10 },
		{ 280, 253, 10 },
		{ 280, 254, 10 },
		{ 280, 255, 10 },
		{ 280, 256, 10 },
		{ 280, 257, 10 },
		{ 280, 258, 10 },
		{ 280, 259, 10 },
		{ 280, 260, 10 },
		{ 280, 261, 10 },
		{ 280, 262, 10 },
		{ 280, 263, 10 },
		{ 280, 264, 10 },
		{ 280, 265, 10 },
		{ 280, 266, 10 },
		{ 280, 267, 10 },
		{ 280, 268, 10 },
		{ 281, 246, 10 },
		{ 281, 247, 10 },
		{ 281, 248, 10 },
		{ 281, 249, 10 },
		{ 281, 250, 10 },
		{ 281, 251, 10 },
		{ 281, 252, 10 },
		{ 281, 253, 10 },
		{ 281, 254, 10 },
		{ 281, 255, 10 },
		{ 281, 256, 10 },
		{ 281, 257, 10 },
		{ 281, 258, 10 },
		{ 281, 259, 10 },
		{ 281, 260, 10 },
		{ 281, 261, 10 },
		{ 281, 262, 10 },
		{ 281, 263, 10 },
		{ 281, 264, 10 },
		{ 281, 265, 10 },
		{ 281, 266, 10 },
		{ 281, 267, 10 },
		{ 281, 268, 10 },
		{ 282, 245, 10 },
		{ 282, 246, 10 },
		{ 282, 247, 10 },
		{ 282, 248, 10 },
		{ 282, 249, 10 },
		{ 282, 250, 10 },
		{ 282, 251, 10 },
		{ 282, 252, 10 },
		{ 282, 253, 10 },
		{ 282, 254, 10 },
		{ 282, 255, 10 },
		{ 282, 256, 10 },
		{ 282, 257, 10 },
		{ 282, 258, 10 },
		{ 282, 259, 10 },
		{ 282, 260, 10 },
		{ 282, 261, 10 },
		{ 282, 262, 10 },
		{ 282, 263, 10 },
		{ 282, 264, 10 },
		{ 282, 265, 10 },
		{ 282, 266, 10 },
		{ 282, 267, 10 },
		{ 282, 268, 10 },
		{ 282, 269, 10 },
		{ 283, 245, 10 },
		{ 283, 246, 10 },
		{ 283, 247, 10 },
		{ 283, 248, 10 },
		{ 283, 249, 10 },
		{ 283, 250, 10 },
		{ 283, 251, 10 },
		{ 283, 252, 10 },
		{ 283, 253, 10 },
		{ 283, 254, 10 },
		{ 283, 255, 10 },
		{ 283, 256, 10 },
		{ 283, 257, 10 },
		{ 283, 258, 10 },
		{ 283, 259, 10 },
		{ 283, 260, 10 },
		{ 283, 261, 10 },
		{ 283, 262, 10 },
		{ 283, 263, 10 },
		{ 283, 264, 10 },
		{ 283, 265, 10 },
		{ 283, 266, 10 },
		{ 283, 267, 10 },
		{ 283, 268, 10 },
		{ 283, 269, 10 },
		{ 284, 244, 10 },
		{ 284, 245, 10 },
		{ 284, 246, 10 },
		{ 284, 247, 10 },
		{ 284, 248, 10 },
		{ 284, 249, 10 },
		{ 284, 250, 10 },
		{ 284, 251, 10 },
		{ 284, 252, 10 },
		{ 284, 253, 10 },
		{ 284, 254, 10 },
		{ 284, 255, 10 },
		{ 284, 256, 10 },
		{ 284, 257, 10 },
		{ 284, 258, 10 },
		{ 284, 259, 10 },
		{ 284, 260, 10 },
		{ 284, 261, 10 },
		{ 284, 262, 10 },
		{ 284, 263, 10 },
		{ 284, 264, 10 },
		{ 284, 265, 10 },
		{ 284, 266, 10 },
		{ 284, 267, 10 },
		{ 284, 268, 10 },
		{ 284, 269, 10 },
		{ 284, 270, 10 },
		{ 285, 244, 10 },
		{ 285, 245, 10 },
		{ 285, 246, 10 },
		{ 285, 247, 10 },
		{ 285, 248, 10 },
		{ 285, 249, 10 },
		{ 285, 250, 10 },
		{ 285, 251, 10 },
		{ 285, 252, 10 },
		{ 285, 253, 10 },
		{ 285, 254, 10 },
		{ 285, 255, 10 },
		{ 285, 256, 10 },
		{ 285, 257, 10 },
		{ 285, 258, 10 },
		{ 285, 259, 10 },
		{ 285, 260, 10 },
		{ 285, 261, 10 },
		{ 285, 262, 10 },
		{ 285, 263, 10 },
		{ 285, 264, 10 },
		{ 285, 265, 10 },
		{ 285, 266, 10 },
		{ 285, 267, 10 },
		{ 285, 268, 10 },
		{ 285, 269, 10 },
		{ 285, 270, 10 },
		{ 286, 243, 10 },
		{ 286, 244, 10 },
		{ 286, 245, 10 },
		{ 286, 246, 10 },
		{ 286, 247, 10 },
		{ 286, 248, 10 },
		{ 286, 249, 10 },
		{ 286, 250, 10 },
		{ 286, 251, 10 },
		{ 286, 252, 10 },
		{ 286, 253, 10 },
		{ 286, 254, 10 },
		{ 286, 255, 10 },
		{ 286, 256, 10 },
		{ 286, 257, 10 },
		{ 286, 258, 10 },
		{ 286, 259, 10 },
		{ 286, 260, 10 },
		{ 286, 261, 10 },
		{ 286, 262, 10 },
		{ 286, 263, 10 },
		{ 286, 264, 10 },
		{ 286, 265, 10 },
		{ 286, 266, 10 },
		{ 286, 267, 10 },
		{ 286, 268, 10 },
		{ 286, 269, 10 },
		{ 286, 270, 10 },
		{ 286, 271, 10 },
		{ 287, 243, 10 },
		{ 287, 244, 10 },
		{ 287, 245, 10 },
		{ 287, 246, 10 },
		{ 287, 247, 10 },
		{ 287, 248, 10 },
		{ 287, 249, 10 },
		{ 287, 250, 10 },
		{ 287, 251, 10 },
		{ 287, 252, 10 },
		{ 287, 253, 10 },
		{ 287, 254, 10 },
		{ 287, 255, 10 },
		{ 287, 256, 10 },
		{ 287, 257, 10 },
		{ 287, 258, 10 },
		{ 287, 259, 10 },
		{ 287, 260, 10 },
		{ 287, 261, 10 },
		{ 287, 262, 10 },
		{ 287, 263, 10 },
		{ 287, 264, 10 },
		{ 287, 265, 10 },
		{ 287, 266, 10 },
		{ 287, 267, 10 },
		{ 287, 268, 10 },
		{ 287, 269, 10 },
		{ 287, 270, 10 },
		{ 287, 271, 10 },
		{ 288, 243, 10 },
		{ 288, 244, 10 },
		{ 288, 245, 10 },
		{ 288, 246, 10 },
		{ 288, 247, 10 },
		{ 288, 248, 10 },
		{ 288, 249, 10 },
		{ 288, 250, 10 },
		{ 288, 251, 10 },
		{ 288, 252, 10 },
		{ 288, 253, 10 },
		{ 288, 254, 10 },
		{ 288, 255, 10 },
		{ 288, 256, 10 },
		{ 288, 257, 10 },
		{ 288, 258, 10 },
		{ 288, 259, 10 },
		{ 288, 260, 10 },
		{ 288, 261, 10 },
		{ 288, 262, 10 },
		{ 288, 263, 10 },
		{ 288, 264, 10 },
		{ 288, 265, 10 },
		{ 288, 266, 10 },
		{ 288, 267, 10 },
		{ 288, 268, 10 },
		{ 288, 269, 10 },
		{ 288, 270, 10 },
		{ 288, 271, 10 },
		{ 226, 227, 11 },
		{ 226, 228, 11 },
		{ 226, 229, 11 },
		{ 226, 230, 11 },
		{ 226, 231, 11 },
		{ 226, 232, 11 },
		{ 226, 233, 11 },
		{ 226, 234, 11 },
		{ 226, 235, 11 },
		{ 226, 236, 11 },
		{ 226, 237, 11 },
		{ 226, 238, 11 },
		{ 226, 239, 11 },
		{ 226, 240, 11 },
		{ 226, 241, 11 },
		{ 226, 242, 11 },
		{ 226, 243, 11 },
		{ 226, 244, 11 },
		{ 226, 245, 11 },
		{ 226, 246, 11 },
		{ 226, 247, 11 },
		{ 226, 248, 11 },
		{ 226, 249, 11 },
		{ 226, 250, 11 },
		{ 226, 251, 11 },
		{ 226, 252, 11 },
		{ 226, 253, 11 },
		{ 226, 254, 11 },
		{ 226, 255, 11 },
		{ 226, 256, 11 },
		{ 227, 228, 11 },
		{ 227, 229, 11 },
		{ 227, 230, 11 },
		{ 227, 231, 11 },
		{ 227, 232, 11 },
		{ 227, 233, 11 },
		{ 227, 234, 11 },
		{ 227, 235, 11 },
		{ 227, 236, 11 },
		{ 227, 237, 11 },
		{ 227, 238, 11 },
		{ 227, 239, 11 },
		{ 227, 240, 11 },
		{ 227, 241, 11 },
		{ 227, 242, 11 },
		{ 227, 243, 11 },
		{ 227, 244, 11 },
		{ 227, 245, 11 },
		{ 227, 246, 11 },
		{ 227, 247, 11 },
		{ 227, 248, 11 },
		{ 227, 249, 11 },
		{ 227, 250, 11 },
		{ 227, 251, 11 },
		{ 227, 252, 11 },
		{ 227, 253, 11 },
		{ 227, 254, 11 },
		{ 227, 255, 11 },
		{ 227, 256, 11 },
		{ 228, 229, 11 },
		{ 228, 230, 11 },
		{ 228, 231, 11 },
		{ 228, 232, 11 },
		{ 228, 233, 11 },
		{ 228, 234, 11 },
		{ 228, 235, 11 },
		{ 228, 236, 11 },
		{ 228, 237, 11 },
		{ 228, 238, 11 },
		{ 228, 239, 11 },
		{ 228, 240, 11 },
		{ 228, 241, 11 },
		{ 228, 242, 11 },
		{ 228, 243, 11 },
		{ 228, 244, 11 },
		{ 228, 245, 11 },
		{ 228, 246, 11 },
		{ 228, 247, 11 },
		{ 228, 248, 11 },
		{ 228, 249, 11 },
		{ 228, 250, 11 },
		{ 228, 251, 11 },
		{ 228, 252, 11 },
		{ 228, 253, 11 },
		{ 228, 254, 11 },
		{ 228, 255, 11 },
		{ 228, 256, 11 },
		{ 229, 230, 11 },
		{ 229, 231, 11 },
		{ 229, 232, 11 },
		{ 229, 233, 11 },
		{ 229, 234, 11 },
		{ 229, 235, 11 },
		{ 229, 236, 11 },
		{ 229, 237, 11 },
		{ 229, 238, 11 },
		{ 229, 239, 11 },
		{ 229, 240, 11 },
		{ 229, 241, 11 },
		{ 229, 242, 11 },
		{ 229, 243, 11 },
		{ 229, 244, 11 },
		{ 229, 245, 11 },
		{ 229, 246, 11 },
		{ 229, 247, 11 },
		{ 229, 248, 11 },
		{ 229, 249, 11 },
		{ 229, 250, 11 },
		{ 229, 251, 11 },
		{ 229, 252, 11 },
		{ 229, 253, 11 },
		{ 229, 254, 11 },
		{ 229, 255, 11 },
		{ 229, 256, 11 },
		{ 230, 231, 11 },
		{ 230, 232, 11 },
		{ 230, 233, 11 },
		{ 230, 234, 11 },
		{ 230, 235, 11 },
		{ 230, 236, 11 },
		{ 230, 237, 11 },
		{ 230, 238, 11 },
		{ 230, 239, 11 },
		{ 230, 240, 11 },
		{ 230, 241, 11 },
		{ 230, 242, 11 },
		{ 230, 243, 11 },
		{ 230, 244, 11 },
		{ 230, 245, 11 },
		{ 230, 246, 11 },
		{ 230, 247, 11 },
		{ 230, 248, 11 },
		{ 230, 249, 11 },
		{ 230, 250, 11 },
		{ 230, 251, 11 },
		{ 230, 252, 11 },
		{ 230, 253, 11 },
		{ 230, 254, 11 },
		{ 230, 255, 11 },
		{ 230, 256, 11 },
		{ 231, 232, 11 },
		{ 231, 233, 11 },
		{ 231, 234, 11 },
		{ 231, 235, 11 },
		{ 231, 236, 11 },
		{ 231, 237, 11 },
		{ 231, 238, 11 },
		{ 231, 239, 11 },
		{ 231, 240, 11 },
		{ 231, 241, 11 },
		{ 231, 242, 11 },
		{ 231, 243, 11 },
		{ 231, 244, 11 },
		{ 231, 245, 11 },
		{ 231, 246, 11 },
		{ 231, 247, 11 },
		{ 231, 248, 11 },
		{ 231, 249, 11 },
		{ 231, 250, 11 },
		{ 231, 251, 11 },
		{ 231, 252, 11 },
		{ 231, 253, 11 },
		{ 231, 254, 11 },
		{ 231, 255, 11 },
		{ 231, 256, 11 },
		{ 232, 233, 11 },
		{ 232, 234, 11 },
		{ 232, 235, 11 },
		{ 232, 236, 11 },
		{ 232, 237, 11 },
		{ 232, 238, 11 },
		{ 232, 239, 11 },
		{ 232, 240, 11 },
		{ 232, 241, 11 },
		{ 232, 242, 11 },
		{ 232, 243, 11 },
		{ 232, 244, 11 },
		{ 232, 245, 11 },
		{ 232, 246, 11 },
		{ 232, 247, 11 },
		{ 232, 248, 11 },
		{ 232, 249, 11 },
		{ 232, 250, 11 },
		{ 232, 251, 11 },
		{ 232, 252, 11 },
		{ 232, 253, 11 },
		{ 232, 254, 11 },
		{ 232, 255, 11 },
		{ 232, 256, 11 },
		{ 233, 234, 11 },
		{ 233, 235, 11 },
		{ 233, 236, 11 },
		{ 233, 237, 11 },
		{ 233, 238, 11 },
		{ 233, 239, 11 },
		{ 233, 240, 11 },
		{ 233, 241, 11 },
		{ 233, 242, 11 },
		{ 233, 243, 11 },
		{ 233, 244, 11 },
		{ 233, 245, 11 },
		{ 233, 246, 11 },
		{ 233, 247, 11 },
		{ 233, 248, 11 },
		{ 233, 249, 11 },
		{ 233, 250, 11 },
		{ 233, 251, 11 },
		{ 233, 252, 11 },
		{ 233, 253, 11 },
		{ 233, 254, 11 },
		{ 233, 255, 11 },
		{ 233, 256, 11 },
		{ 234, 235, 11 },
		{ 234, 236, 11 },
		{ 234, 237, 11 },
		{ 234, 238, 11 },
		{ 234, 239, 11 },
		{ 234, 240, 11 },
		{ 234, 241, 11 },
		{ 234, 242, 11 },
		{ 234, 243, 11 },
		{ 234, 244, 11 },
		{ 234, 245, 11 },
		{ 234, 246, 11 },
		{ 234, 247, 11 },
		{ 234, 248, 11 },
		{ 234, 249, 11 },
		{ 234, 250, 11 },
		{ 234, 251, 11 },
		{ 234, 252, 11 },
		{ 234, 253, 11 },
		{ 234, 254, 11 },
		{ 234, 255, 11 },
		{ 234, 256, 11 },
		{ 235, 236, 11 },
		{ 235, 237, 11 },
		{ 235, 238, 11 },
		{ 235, 239, 11 },
		{ 235, 240, 11 },
		{ 235, 241, 11 },
		{ 235, 242, 11 },
		{ 235, 243, 11 },
		{ 235, 244, 11 },
		{ 235, 245, 11 },
		{ 235, 246, 11 },
		{ 235, 247, 11 },
		{ 235, 248, 11 },
		{ 235, 249, 11 },
		{ 235, 250, 11 },
		{ 235, 251, 11 },
		{ 235, 252, 11 },
		{ 235, 253, 11 },
		{ 235, 254, 11 },
		{ 235, 255, 11 },
		{ 235, 256, 11 },
		{ 236, 237, 11 },
		{ 236, 238, 11 },
		{ 236, 239, 11 },
		{ 236, 240, 11 },
		{ 236, 241, 11 },
		{ 236, 242, 11 },
		{ 236, 243, 11 },
		{ 236, 244, 11 },
		{ 236, 245, 11 },
		{ 236, 246, 11 },
		{ 236, 247, 11 },
		{ 236, 248, 11 },
		{ 236, 249, 11 },
		{ 236, 250, 11 },
		{ 236, 251, 11 },
		{ 236, 252, 11 },
		{ 236, 253, 11 },
		{ 236, 254, 11 },
		{ 236, 255, 11 },
		{ 236, 256, 11 },
		{ 237, 238, 11 },
		{ 237, 239, 11 },
		{ 237, 240, 11 },
		{ 237, 241, 11 },
		{ 237, 242, 11 },
		{ 237, 243, 11 },
		{ 237, 244, 11 },
		{ 237, 245, 11 },
		{ 237, 246, 11 },
		{ 237, 247, 11 },
		{ 237, 248, 11 },
		{ 237, 249, 11 },
		{ 237, 250, 11 },
		{ 237, 251, 11 },
		{ 237, 252, 11 },
		{ 237, 253, 11 },
		{ 237, 254, 11 },
		{ 237, 255, 11 },
		{ 237, 256, 11 },
		{ 238, 239, 11 },
		{ 238, 240, 11 },
		{ 238, 241, 11 },
		{ 238, 242, 11 },
		{ 238, 243, 11 },
		{ 238, 244, 11 },
		{ 238, 245, 11 },
		{ 238, 246, 11 },
		{ 238, 247, 11 },
		{ 238, 248, 11 },
		{ 238, 249, 11 },
		{ 238, 250, 11 },
		{ 238, 251, 11 },
		{ 238, 252, 11 },
		{ 238, 253, 11 },
		{ 238, 254, 11 },
		{ 238, 255, 11 },
		{ 238, 256, 11 },
		{ 239, 240, 11 },
		{ 239, 241, 11 },
		{ 239, 242, 11 },
		{ 239, 243, 11 },
		{ 239, 244, 11 },
		{ 239, 245, 11 },
		{ 239, 246, 11 },
		{ 239, 247, 11 },
		{ 239, 248, 11 },
		{ 239, 249, 11 },
		{ 239, 250, 11 },
		{ 239, 251, 11 },
		{ 239, 252, 11 },
		{ 239, 253, 11 },
		{ 239, 254, 11 },
		{ 239, 255, 11 },
		{ 239, 256, 11 },
		{ 240, 241, 11 },
		{ 240, 242, 11 },
		{ 240, 243, 11 },
		{ 240, 244, 11 },
		{ 240, 245, 11 },
		{ 240, 246, 11 },
		{ 240, 247, 11 },
		{ 240, 248, 11 },
		{ 240, 249, 11 },
		{ 240, 250, 11 },
		{ 240, 251, 11 },
		{ 240, 252, 11 },
		{ 240, 253, 11 },
		{ 240, 254, 11 },
		{ 240, 255, 11 },
		{ 240, 256, 11 },
		{ 241, 242, 11 },
		{ 241, 243, 11 },
		{ 241, 244, 11 },
		{ 241, 245, 11 },
		{ 241, 246, 11 },
		{ 241, 247, 11 },
		{ 241, 248, 11 },
		{ 241, 249, 11 },
		{ 241, 250, 11 },
		{ 241, 251, 11 },
		{ 241, 252, 11 },
		{ 241, 253, 11 },
		{ 241, 254, 11 },
		{ 241, 255, 11 },
		{ 241, 256, 11 },
		{ 242, 243, 11 },
		{ 242, 244, 11 },
		{ 242, 245, 11 },
		{ 242, 246, 11 },
		{ 242, 247, 11 },
		{ 242, 248, 11 },
		{ 242, 249, 11 },
		{ 242, 250, 11 },
		{ 242, 251, 11 },
		{ 242, 252, 11 },
		{ 242, 253, 11 },
		{ 242, 254, 11 },
		{ 242, 255, 11 },
		{ 242, 256, 11 },
		{ 243, 244, 11 },
		{ 243, 245, 11 },
		{ 243, 246, 11 },
		{ 243, 247, 11 },
		{ 243, 248, 11 },
		{ 243, 249, 11 },
		{ 243, 250, 11 },
		{ 243, 251, 11 },
		{ 243, 252, 11 },
		{ 243, 253, 11 },
		{ 243, 254, 11 },
		{ 243, 255, 11 },
		{ 243, 256, 11 },
		{ 244, 245, 11 },
		{ 244, 246, 11 },
		{ 244, 247, 11 },
		{ 244, 248, 11 },
		{ 244, 249, 11 },
		{ 244, 250, 11 },
		{ 244, 251, 11 },
		{ 244, 252, 11 },
		{ 244, 253, 11 },
		{ 244, 254, 11 },
		{ 244, 255, 11 },
		{ 244, 256, 11 },
		{ 245, 246, 11 },
		{ 245, 247, 11 },
		{ 245, 248, 11 },
		{ 245, 249, 11 },
		{ 245, 250, 11 },
		{ 245, 251, 11 },
		{ 245, 252, 11 },
		{ 245, 253, 11 },
		{ 245, 254, 11 },
		{ 245, 255, 11 },
		{ 245, 256, 11 },
		{ 246, 247, 11 },
		{ 246, 248, 11 },
		{ 246, 249, 11 },
		{ 246, 250, 11 },
		{ 246, 251, 11 },
		{ 246, 252, 11 },
		{ 246, 253, 11 },
		{ 246, 254, 11 },
		{ 246, 255, 11 },
		{ 246, 256, 11 },
		{ 247, 248, 11 },
		{ 247, 249, 11 },
		{ 247, 250, 11 },
		{ 247, 251, 11 },
		{ 247, 252, 11 },
		{ 247, 253, 11 },
		{ 247, 254, 11 },
		{ 247, 255, 11 },
		{ 247, 256, 11 },
		{ 248, 249, 11 },
		{ 248, 250, 11 },
		{ 248, 251, 11 },
		{ 248, 252, 11 },
		{ 248, 253, 11 },
		{ 248, 254, 11 },
		{ 248, 255, 11 },
		{ 248, 256, 11 },
		{ 249, 250, 11 },
		{ 249, 251, 11 },
		{ 249, 252, 11 },
		{ 249, 253, 11 },
		{ 249, 254, 11 },
		{ 249, 255, 11 },
		{ 249, 256, 11 },
		{ 250, 251, 11 },
		{ 250, 252, 11 },
		{ 250, 253, 11 },
		{ 250, 254, 11 },
		{ 250, 255, 11 },
		{ 250, 256, 11 },
		{ 251, 252, 11 },
		{ 251, 253, 11 },
		{ 251, 254, 11 },
		{ 251, 255, 11 },
		{ 251, 256, 11 },
		{ 252, 253, 11 },
		{ 252, 254, 11 },
		{ 252, 255, 11 },
		{ 252, 256, 11 },
		{ 262, 258, 11 },
		{ 262, 259, 11 },
		{ 262, 260, 11 },
		{ 262, 261, 11 },
		{ 263, 258, 11 },
		{ 263, 259, 11 },
		{ 263, 260, 11 },
		{ 263, 261, 11 },
		{ 263, 262, 11 },
		{ 264, 258, 11 },
		{ 264, 259, 11 },
		{ 264, 260, 11 },
		{ 264, 261, 11 },
		{ 264, 262, 11 },
		{ 264, 263, 11 },
		{ 265, 258, 11 },
		{ 265, 259, 11 },
		{ 265, 260, 11 },
		{ 265, 261, 11 },
		{ 265, 262, 11 },
		{ 265, 263, 11 },
		{ 265, 264, 11 },
		{ 266, 258, 11 },
		{ 266, 259, 11 },
		{ 266, 260, 11 },
		{ 266, 261, 11 },
		{ 266, 262, 11 },
		{ 266, 263, 11 },
		{ 266, 264, 11 },
		{ 266, 265, 11 },
		{ 267, 258, 11 },
		{ 267, 259, 11 },
		{ 267, 260, 11 },
		{ 267, 261, 11 },
		{ 267, 262, 11 },
		{ 267, 263, 11 },
		{ 267, 264, 11 },
		{ 267, 265, 11 },
		{ 267, 266, 11 },
		{ 268, 258, 11 },
		{ 268, 259, 11 },
		{ 268, 260, 11 },
		{ 268, 261, 11 },
		{ 268, 262, 11 },
		{ 268, 263, 11 },
		{ 268, 264, 11 },
		{ 268, 265, 11 },
		{ 268, 266, 11 },
		{ 268, 267, 11 },
		{ 269, 258, 11 },
		{ 269, 259, 11 },
		{ 269, 260, 11 },
		{ 269, 261, 11 },
		{ 269, 262, 11 },
		{ 269, 263, 11 },
		{ 269, 264, 11 },
		{ 269, 265, 11 },
		{ 269, 266, 11 },
		{ 269, 267, 11 },
		{ 269, 268, 11 },
		{ 270, 258, 11 },
		{ 270, 259, 11 },
		{ 270, 260, 11 },
		{ 270, 261, 11 },
		{ 270, 262, 11 },
		{ 270, 263, 11 },
		{ 270, 264, 11 },
		{ 270, 265, 11 },
		{ 270, 266, 11 },
		{ 270, 267, 11 },
		{ 270, 268, 11 },
		{ 270, 269, 11 },
		{ 271, 258, 11 },
		{ 271, 259, 11 },
		{ 271, 260, 11 },
		{ 271, 261, 11 },
		{ 271, 262, 11 },
		{ 271, 263, 11 },
		{ 271, 264, 11 },
		{ 271, 265, 11 },
		{ 271, 266, 11 },
		{ 271, 267, 11 },
		{ 271, 268, 11 },
		{ 271, 269, 11 },
		{ 271, 270, 11 },
		{ 272, 258, 11 },
		{ 272, 259, 11 },
		{ 272, 260, 11 },
		{ 272, 261, 11 },
		{ 272, 262, 11 },
		{ 272, 263, 11 },
		{ 272, 264, 11 },
		{ 272, 265, 11 },
		{ 272, 266, 11 },
		{ 272, 267, 11 },
		{ 272, 268, 11 },
		{ 272, 269, 11 },
		{ 272, 270, 11 },
		{ 272, 271, 11 },
		{ 273, 258, 11 },
		{ 273, 259, 11 },
		{ 273, 260, 11 },
		{ 273, 261, 11 },
		{ 273, 262, 11 },
		{ 273, 263, 11 },
		{ 273, 264, 11 },
		{ 273, 265, 11 },
		{ 273, 266, 11 },
		{ 273, 267, 11 },
		{ 273, 268, 11 },
		{ 273, 269, 11 },
		{ 273, 270, 11 },
		{ 273, 271, 11 },
		{ 273, 272, 11 },
		{ 274, 258, 11 },
		{ 274, 259, 11 },
		{ 274, 260, 11 },
		{ 274, 261, 11 },
		{ 274, 262, 11 },
		{ 274, 263, 11 },
		{ 274, 264, 11 },
		{ 274, 265, 11 },
		{ 274, 266, 11 },
		{ 274, 267, 11 },
		{ 274, 268, 11 },
		{ 274, 269, 11 },
		{ 274, 270, 11 },
		{ 274, 271, 11 },
		{ 274, 272, 11 },
		{ 274, 273, 11 },
		{ 275, 258, 11 },
		{ 275, 259, 11 },
		{ 275, 260, 11 },
		{ 275, 261, 11 },
		{ 275, 262, 11 },
		{ 275, 263, 11 },
		{ 275, 264, 11 },
		{ 275, 265, 11 },
		{ 275, 266, 11 },
		{ 275, 267, 11 },
		{ 275, 268, 11 },
		{ 275, 269, 11 },
		{ 275, 270, 11 },
		{ 275, 271, 11 },
		{ 275, 272, 11 },
		{ 275, 273, 11 },
		{ 275, 274, 11 },
		{ 276, 258, 11 },
		{ 276, 259, 11 },
		{ 276, 260, 11 },
		{ 276, 261, 11 },
		{ 276, 262, 11 },
		{ 276, 263, 11 },
		{ 276, 264, 11 },
		{ 276, 265, 11 },
		{ 276, 266, 11 },
		{ 276, 267, 11 },
		{ 276, 268, 11 },
		{ 276, 269, 11 },
		{ 276, 270, 11 },
		{ 276, 271, 11 },
		{ 276, 272, 11 },
		{ 276, 273, 11 },
		{ 276, 274, 11 },
		{ 276, 275, 11 },
		{ 277, 258, 11 },
		{ 277, 259, 11 },
		{ 277, 260, 11 },
		{ 277, 261, 11 },
		{ 277, 262, 11 },
		{ 277, 263, 11 },
		{ 277, 264, 11 },
		{ 277, 265, 11 },
		{ 277, 266, 11 },
		{ 277, 267, 11 },
		{ 277, 268, 11 },
		{ 277, 269, 11 },
		{ 277, 270, 11 },
		{ 277, 271, 11 },
		{ 277, 272, 11 },
		{ 277, 273, 11 },
		{ 277, 274, 11 },
		{ 277, 275, 11 },
		{ 277, 276, 11 },
		{ 278, 258, 11 },
		{ 278, 259, 11 },
		{ 278, 260, 11 },
		{ 278, 261, 11 },
		{ 278, 262, 11 },
		{ 278, 263, 11 },
		{ 278, 264, 11 },
		{ 278, 265, 11 },
		{ 278, 266, 11 },
		{ 278, 267, 11 },
		{ 278, 268, 11 },
		{ 278, 269, 11 },
		{ 278, 270, 11 },
		{ 278, 271, 11 },
		{ 278, 272, 11 },
		{ 278, 273, 11 },
		{ 278, 274, 11 },
		{ 278, 275, 11 },
		{ 278, 276, 11 },
		{ 278, 277, 11 },
		{ 279, 258, 11 },
		{ 279, 259, 11 },
		{ 279, 260, 11 },
		{ 279, 261, 11 },
		{ 279, 262, 11 },
		{ 279, 263, 11 },
		{ 279, 264, 11 },
		{ 279, 265, 11 },
		{ 279, 266, 11 },
		{ 279, 267, 11 },
		{ 279, 268, 11 },
		{ 279, 269, 11 },
		{ 279, 270, 11 },
		{ 279, 271, 11 },
		{ 279, 272, 11 },
		{ 279, 273, 11 },
		{ 279, 274, 11 },
		{ 279, 275, 11 },
		{ 279, 276, 11 },
		{ 279, 277, 11 },
		{ 279, 278, 11 },
		{ 280, 258, 11 },
		{ 280, 259, 11 },
		{ 280, 260, 11 },
		{ 280, 261, 11 },
		{ 280, 262, 11 },
		{ 280, 263, 11 },
		{ 280, 264, 11 },
		{ 280, 265, 11 },
		{ 280, 266, 11 },
		{ 280, 267, 11 },
		{ 280, 268, 11 },
		{ 280, 269, 11 },
		{ 280, 270, 11 },
		{ 280, 271, 11 },
		{ 280, 272, 11 },
		{ 280, 273, 11 },
		{ 280, 274, 11 },
		{ 280, 275, 11 },
		{ 280, 276, 11 },
		{ 280, 277, 11 },
		{ 280, 278, 11 },
		{ 280, 279, 11 },
		{ 281, 258, 11 },
		{ 281, 259, 11 },
		{ 281, 260, 11 },
		{ 281, 261, 11 },
		{ 281, 262, 11 },
		{ 281, 263, 11 },
		{ 281, 264, 11 },
		{ 281, 265, 11 },
		{ 281, 266, 11 },
		{ 281, 267, 11 },
		{ 281, 268, 11 },
		{ 281, 269, 11 },
		{ 281, 270, 11 },
		{ 281, 271, 11 },
		{ 281, 272, 11 },
		{ 281, 273, 11 },
		{ 281, 274, 11 },
		{ 281, 275, 11 },
		{ 281, 276, 11 },
		{ 281, 277, 11 },
		{ 281, 278, 11 },
		{ 281, 279, 11 },
		{ 281, 280, 11 },
		{ 282, 258, 11 },
		{ 282, 259, 11 },
		{ 282, 260, 11 },
		{ 282, 261, 11 },
		{ 282, 262, 11 },
		{ 282, 263, 11 },
		{ 282, 264, 11 },
		{ 282, 265, 11 },
		{ 282, 266, 11 },
		{ 282, 267, 11 },
		{ 282, 268, 11 },
		{ 282, 269, 11 },
		{ 282, 270, 11 },
		{ 282, 271, 11 },
		{ 282, 272, 11 },
		{ 282, 273, 11 },
		{ 282, 274, 11 },
		{ 282, 275, 11 },
		{ 282, 276, 11 },
		{ 282, 277, 11 },
		{ 282, 278, 11 },
		{ 282, 279, 11 },
		{ 282, 280, 11 },
		{ 282, 281, 11 },
		{ 283, 258, 11 },
		{ 283, 259, 11 },
		{ 283, 260, 11 },
		{ 283, 261, 11 },
		{ 283, 262, 11 },
		{ 283, 263, 11 },
		{ 283, 264, 11 },
		{ 283, 265, 11 },
		{ 283, 266, 11 },
		{ 283, 267, 11 },
		{ 283, 268, 11 },
		{ 283, 269, 11 },
		{ 283, 270, 11 },
		{ 283, 271, 11 },
		{ 283, 272, 11 },
		{ 283, 273, 11 },
		{ 283, 274, 11 },
		{ 283, 275, 11 },
		{ 283, 276, 11 },
		{ 283, 277, 11 },
		{ 283, 278, 11 },
		{ 283, 279, 11 },
		{ 283, 280, 11 },
		{ 283, 281, 11 },
		{ 283, 282, 11 },
		{ 284, 258, 11 },
		{ 284, 259, 11 },
		{ 284, 260, 11 },
		{ 284, 261, 11 },
		{ 284, 262, 11 },
		{ 284, 263, 11 },
		{ 284, 264, 11 },
		{ 284, 265, 11 },
		{ 284, 266, 11 },
		{ 284, 267, 11 },
		{ 284, 268, 11 },
		{ 284, 269, 11 },
		{ 284, 270, 11 },
		{ 284, 271, 11 },
		{ 284, 272, 11 },
		{ 284, 273, 11 },
		{ 284, 274, 11 },
		{ 284, 275, 11 },
		{ 284, 276, 11 },
		{ 284, 277, 11 },
		{ 284, 278, 11 },
		{ 284, 279, 11 },
		{ 284, 280, 11 },
		{ 284, 281, 11 },
		{ 284, 282, 11 },
		{ 284, 283, 11 },
		{ 285, 258, 11 },
		{ 285, 259, 11 },
		{ 285, 260, 11 },
		{ 285, 261, 11 },
		{ 285, 262, 11 },
		{ 285, 263, 11 },
		{ 285, 264, 11 },
		{ 285, 265, 11 },
		{ 285, 266, 11 },
		{ 285, 267, 11 },
		{ 285, 268, 11 },
		{ 285, 269, 11 },
		{ 285, 270, 11 },
		{ 285, 271, 11 },
		{ 285, 272, 11 },
		{ 285, 273, 11 },
		{ 285, 274, 11 },
		{ 285, 275, 11 },
		{ 285, 276, 11 },
		{ 285, 277, 11 },
		{ 285, 278, 11 },
		{ 285, 279, 11 },
		{ 285, 280, 11 },
		{ 285, 281, 11 },
		{ 285, 282, 11 },
		{ 285, 283, 11 },
		{ 285, 284, 11 },
		{ 286, 258, 11 },
		{ 286, 259, 11 },
		{ 286, 260, 11 },
		{ 286, 261, 11 },
		{ 286, 262, 11 },
		{ 286, 263, 11 },
		{ 286, 264, 11 },
		{ 286, 265, 11 },
		{ 286, 266, 11 },
		{ 286, 267, 11 },
		{ 286, 268, 11 },
		{ 286, 269, 11 },
		{ 286, 270, 11 },
		{ 286, 271, 11 },
		{ 286, 272, 11 },
		{ 286, 273, 11 },
		{ 286, 274, 11 },
		{ 286, 275, 11 },
		{ 286, 276, 11 },
		{ 286, 277, 11 },
		{ 286, 278, 11 },
		{ 286, 279, 11 },
		{ 286, 280, 11 },
		{ 286, 281, 11 },
		{ 286, 282, 11 },
		{ 286, 283, 11 },
		{ 286, 284, 11 },
		{ 286, 285, 11 },
		{ 287, 258, 11 },
		{ 287, 259, 11 },
		{ 287, 260, 11 },
		{ 287, 261, 11 },
		{ 287, 262, 11 },
		{ 287, 263, 11 },
		{ 287, 264, 11 },
		{ 287, 265, 11 },
		{ 287, 266, 11 },
		{ 287, 267, 11 },
		{ 287, 268, 11 },
		{ 287, 269, 11 },
		{ 287, 270, 11 },
		{ 287, 271, 11 },
		{ 287, 272, 11 },
		{ 287, 273, 11 },
		{ 287, 274, 11 },
		{ 287, 275, 11 },
		{ 287, 276, 11 },
		{ 287, 277, 11 },
		{ 287, 278, 11 },
		{ 287, 279, 11 },
		{ 287, 280, 11 },
		{ 287, 281, 11 },
		{ 287, 282, 11 },
		{ 287, 283, 11 },
		{ 287, 284, 11 },
		{ 287, 285, 11 },
		{ 287, 286, 11 },
		{ 288, 258, 11 },
		{ 288, 259, 11 },
		{ 288, 260, 11 },
		{ 288, 261, 11 },
		{ 288, 262, 11 },
		{ 288, 263, 11 },
		{ 288, 264, 11 },
		{ 288, 265, 11 },
		{ 288, 266, 11 },
		{ 288, 267, 11 },
		{ 288, 268, 11 },
		{ 288, 269, 11 },
		{ 288, 270, 11 },
		{ 288, 271, 11 },
		{ 288, 272, 11 },
		{ 288, 273, 11 },
		{ 288, 274, 11 },
		{ 288, 275, 11 },
		{ 288, 276, 11 },
		{ 288, 277, 11 },
		{ 288, 278, 11 },
		{ 288, 279, 11 },
		{ 288, 280, 11 },
		{ 288, 281, 11 },
		{ 288, 282, 11 },
		{ 288, 283, 11 },
		{ 288, 284, 11 },
		{ 288, 285, 11 },
		{ 288, 286, 11 },
		{ 288, 287, 11 },
		{ 226, 226, 12 },
		{ 226, 227, 12 },
		{ 226, 228, 12 },
		{ 226, 229, 12 },
		{ 226, 230, 12 },
		{ 226, 231, 12 },
		{ 226, 232, 12 },
		{ 226, 233, 12 },
		{ 226, 234, 12 },
		{ 226, 235, 12 },
		{ 226, 236, 12 },
		{ 226, 237, 12 },
		{ 226, 238, 12 },
		{ 226, 239, 12 },
		{ 226, 240, 12 },
		{ 227, 226, 12 },
		{ 227, 227, 12 },
		{ 227, 228, 12 },
		{ 227, 229, 12 },
		{ 227, 230, 12 },
		{ 227, 231, 12 },
		{ 227, 232, 12 },
		{ 227, 233, 12 },
		{ 227, 234, 12 },
		{ 227, 235, 12 },
		{ 227, 236, 12 },
		{ 227, 237, 12 },
		{ 227, 238, 12 },
		{ 227, 239, 12 },
		{ 227, 240, 12 },
		{ 227, 241, 12 },
		{ 228, 226, 12 },
		{ 228, 227, 12 },
		{ 228, 228, 12 },
		{ 228, 229, 12 },
		{ 228, 230, 12 },
		{ 228, 231, 12 },
		{ 228, 232, 12 },
		{ 228, 233, 12 },
		{ 228, 234, 12 },
		{ 228, 235, 12 },
		{ 228, 236, 12 },
		{ 228, 237, 12 },
		{ 228, 238, 12 },
		{ 228, 239, 12 },
		{ 228, 240, 12 },
		{ 228, 241, 12 },
		{ 228, 242, 12 },
		{ 229, 226, 12 },
		{ 229, 227, 12 },
		{ 229, 228, 12 },
		{ 229, 229, 12 },
		{ 229, 230, 12 },
		{ 229, 231, 12 },
		{ 229, 232, 12 },
		{ 229, 233, 12 },
		{ 229, 234, 12 },
		{ 229, 235, 12 },
		{ 229, 236, 12 },
		{ 229, 237, 12 },
		{ 229, 238, 12 },
		{ 229, 239, 12 },
		{ 229, 240, 12 },
		{ 229, 241, 12 },
		{ 229, 242, 12 },
		{ 230, 226, 12 },
		{ 230, 227, 12 },
		{ 230, 228, 12 },
		{ 230, 229, 12 },
		{ 230, 230, 12 },
		{ 230, 231, 12 },
		{ 230, 232, 12 },
		{ 230, 233, 12 },
		{ 230, 234, 12 },
		{ 230, 235, 12 },
		{ 230, 236, 12 },
		{ 230, 237, 12 },
		{ 230, 238, 12 },
		{ 230, 239, 12 },
		{ 230, 240, 12 },
		{ 230, 241, 12 },
		{ 230, 242, 12 },
		{ 230, 243, 12 },
		{ 231, 226, 12 },
		{ 231, 227, 12 },
		{ 231, 228, 12 },
		{ 231, 229, 12 },
		{ 231, 230, 12 },
		{ 231, 231, 12 },
		{ 231, 232, 12 },
		{ 231, 233, 12 },
		{ 231, 234, 12 },
		{ 231, 235, 12 },
		{ 231, 236, 12 },
		{ 231, 237, 12 },
		{ 231, 238, 12 },
		{ 231, 239, 12 },
		{ 231, 240, 12 },
		{ 231, 241, 12 },
		{ 231, 242, 12 },
		{ 231, 243, 12 },
		{ 232, 226, 12 },
		{ 232, 227, 12 },
		{ 232, 228, 12 },
		{ 232, 229, 12 },
		{ 232, 230, 12 },
		{ 232, 231, 12 },
		{ 232, 232, 12 },
		{ 232, 233, 12 },
		{ 232, 234, 12 },
		{ 232, 235, 12 },
		{ 232, 236, 12 },
		{ 232, 237, 12 },
		{ 232, 238, 12 },
		{ 232, 239, 12 },
		{ 232, 240, 12 },
		{ 232, 241, 12 },
		{ 232, 242, 12 },
		{ 232, 243, 12 },
		{ 232, 244, 12 },
		{ 233, 226, 12 },
		{ 233, 227, 12 },
		{ 233, 228, 12 },
		{ 233, 229, 12 },
		{ 233, 230, 12 },
		{ 233, 231, 12 },
		{ 233, 232, 12 },
		{ 233, 233, 12 },
		{ 233, 234, 12 },
		{ 233, 235, 12 },
		{ 233, 236, 12 },
		{ 233, 237, 12 },
		{ 233, 238, 12 },
		{ 233, 239, 12 },
		{ 233, 240, 12 },
		{ 233, 241, 12 },
		{ 233, 242, 12 },
		{ 233, 243, 12 },
		{ 233, 244, 12 },
		{ 234, 226, 12 },
		{ 234, 227, 12 },
		{ 234, 228, 12 },
		{ 234, 229, 12 },
		{ 234, 230, 12 },
		{ 234, 231, 12 },
		{ 234, 232, 12 },
		{ 234, 233, 12 },
		{ 234, 234, 12 },
		{ 234, 235, 12 },
		{ 234, 236, 12 },
		{ 234, 237, 12 },
		{ 234, 238, 12 },
		{ 234, 239, 12 },
		{ 234, 240, 12 },
		{ 234, 241, 12 },
		{ 234, 242, 12 },
		{ 234, 243, 12 },
		{ 234, 244, 12 },
		{ 234, 245, 12 },
		{ 235, 226, 12 },
		{ 235, 227, 12 },
		{ 235, 228, 12 },
		{ 235, 229, 12 },
		{ 235, 230, 12 },
		{ 235, 231, 12 },
		{ 235, 232, 12 },
		{ 235, 233, 12 },
		{ 235, 234, 12 },
		{ 235, 235, 12 },
		{ 235, 236, 12 },
		{ 235, 237, 12 },
		{ 235, 238, 12 },
		{ 235, 239, 12 },
		{ 235, 240, 12 },
		{ 235, 241, 12 },
		{ 235, 242, 12 },
		{ 235, 243, 12 },
		{ 235, 244, 12 },
		{ 235, 245, 12 },
		{ 236, 226, 12 },
		{ 236, 227, 12 },
		{ 236, 228, 12 },
		{ 236, 229, 12 },
		{ 236, 230, 12 },
		{ 236, 231, 12 },
		{ 236, 232, 12 },
		{ 236, 233, 12 },
		{ 236, 234, 12 },
		{ 236, 235, 12 },
		{ 236, 236, 12 },
		{ 236, 237, 12 },
		{ 236, 238, 12 },
		{ 236, 239, 12 },
		{ 236, 240, 12 },
		{ 236, 241, 12 },
		{ 236, 242, 12 },
		{ 236, 243, 12 },
		{ 236, 244, 12 },
		{ 236, 245, 12 },
		{ 236, 246, 12 },
		{ 237, 226, 12 },
		{ 237, 227, 12 },
		{ 237, 228, 12 },
		{ 237, 229, 12 },
		{ 237, 230, 12 },
		{ 237, 231, 12 },
		{ 237, 232, 12 },
		{ 237, 233, 12 },
		{ 237, 234, 12 },
		{ 237, 235, 12 },
		{ 237, 236, 12 },
		{ 237, 237, 12 },
		{ 237, 238, 12 },
		{ 237, 239, 12 },
		{ 237, 240, 12 },
		{ 237, 241, 12 },
		{ 237, 242, 12 },
		{ 237, 243, 12 },
		{ 237, 244, 12 },
		{ 237, 245, 12 },
		{ 237, 246, 12 },
		{ 238, 226, 12 },
		{ 238, 227, 12 },
		{ 238, 228, 12 },
		{ 238, 229, 12 },
		{ 238, 230, 12 },
		{ 238, 231, 12 },
		{ 238, 232, 12 },
		{ 238, 233, 12 },
		{ 238, 234, 12 },
		{ 238, 235, 12 },
		{ 238, 236, 12 },
		{ 238, 237, 12 },
		{ 238, 238, 12 },
		{ 238, 239, 12 },
		{ 238, 240, 12 },
		{ 238, 241, 12 },
		{ 238, 242, 12 },
		{ 238, 243, 12 },
		{ 238, 244, 12 },
		{ 238, 245, 12 },
		{ 238, 246, 12 },
		{ 238, 247, 12 },
		{ 239, 226, 12 },
		{ 239, 227, 12 },
		{ 239, 228, 12 },
		{ 239, 229, 12 },
		{ 239, 230, 12 },
		{ 239, 231, 12 },
		{ 239, 232, 12 },
		{ 239, 233, 12 },
		{ 239, 234, 12 },
		{ 239, 235, 12 },
		{ 239, 236, 12 },
		{ 239, 237, 12 },
		{ 239, 238, 12 },
		{ 239, 239, 12 },
		{ 239, 240, 12 },
		{ 239, 241, 12 },
		{ 239, 242, 12 },
		{ 239, 243, 12 },
		{ 239, 244, 12 },
		{ 239, 245, 12 },
		{ 239, 246, 12 },
		{ 239, 247, 12 },
		{ 240, 226, 12 },
		{ 240, 227, 12 },
		{ 240, 228, 12 },
		{ 240, 229, 12 },
		{ 240, 230, 12 },
		{ 240, 231, 12 },
		{ 240, 232, 12 },
		{ 240, 233, 12 },
		{ 240, 234, 12 },
		{ 240, 235, 12 },
		{ 240, 236, 12 },
		{ 240, 237, 12 },
		{ 240, 238, 12 },
		{ 240, 239, 12 },
		{ 240, 240, 12 },
		{ 240, 241, 12 },
		{ 240, 242, 12 },
		{ 240, 243, 12 },
		{ 240, 244, 12 },
		{ 240, 245, 12 },
		{ 240, 246, 12 },
		{ 240, 247, 12 },
		{ 240, 248, 12 },
		{ 241, 227, 12 },
		{ 241, 228, 12 },
		{ 241, 229, 12 },
		{ 241, 230, 12 },
		{ 241, 231, 12 },
		{ 241, 232, 12 },
		{ 241, 233, 12 },
		{ 241, 234, 12 },
		{ 241, 235, 12 },
		{ 241, 236, 12 },
		{ 241, 237, 12 },
		{ 241, 238, 12 },
		{ 241, 239, 12 },
		{ 241, 240, 12 },
		{ 241, 241, 12 },
		{ 241, 242, 12 },
		{ 241, 243, 12 },
		{ 241, 244, 12 },
		{ 241, 245, 12 },
		{ 241, 246, 12 },
		{ 241, 247, 12 },
		{ 241, 248, 12 },
		{ 242, 228, 12 },
		{ 242, 229, 12 },
		{ 242, 230, 12 },
		{ 242, 231, 12 },
		{ 242, 232, 12 },
		{ 242, 233, 12 },
		{ 242, 234, 12 },
		{ 242, 235, 12 },
		{ 242, 236, 12 },
		{ 242, 237, 12 },
		{ 242, 238, 12 },
		{ 242, 239, 12 },
		{ 242, 240, 12 },
		{ 242, 241, 12 },
		{ 242, 242, 12 },
		{ 242, 243, 12 },
		{ 242, 244, 12 },
		{ 242, 245, 12 },
		{ 242, 246, 12 },
		{ 242, 247, 12 },
		{ 242, 248, 12 },
		{ 242, 249, 12 },
		{ 243, 230, 12 },
		{ 243, 231, 12 },
		{ 243, 232, 12 },
		{ 243, 233, 12 },
		{ 243, 234, 12 },
		{ 243, 235, 12 },
		{ 243, 236, 12 },
		{ 243, 237, 12 },
		{ 243, 238, 12 },
		{ 243, 239, 12 },
		{ 243, 240, 12 },
		{ 243, 241, 12 },
		{ 243, 242, 12 },
		{ 243, 243, 12 },
		{ 243, 244, 12 },
		{ 243, 245, 12 },
		{ 243, 246, 12 },
		{ 243, 247, 12 },
		{ 243, 248, 12 },
		{ 243, 249, 12 },
		{ 244, 232, 12 },
		{ 244, 233, 12 },
		{ 244, 234, 12 },
		{ 244, 235, 12 },
		{ 244, 236, 12 },
		{ 244, 237, 12 },
		{ 244, 238, 12 },
		{ 244, 239, 12 },
		{ 244, 240, 12 },
		{ 244, 241, 12 },
		{ 244, 242, 12 },
		{ 244, 243, 12 },
		{ 244, 244, 12 },
		{ 244, 245, 12 },
		{ 244, 246, 12 },
		{ 244, 247, 12 },
		{ 244, 248, 12 },
		{ 244, 249, 12 },
		{ 244, 250, 12 },
		{ 245, 234, 12 },
		{ 245, 235, 12 },
		{ 245, 236, 12 },
		{ 245, 237, 12 },
		{ 245, 238, 12 },
		{ 245, 239, 12 },
		{ 245, 240, 12 },
		{ 245, 241, 12 },
		{ 245, 242, 12 },
		{ 245, 243, 12 },
		{ 245, 244, 12 },
		{ 245, 245, 12 },
		{ 245, 246, 12 },
		{ 245, 247, 12 },
		{ 245, 248, 12 },
		{ 245, 249, 12 },
		{ 245, 250, 12 },
		{ 246, 236, 12 },
		{ 246, 237, 12 },
		{ 246, 238, 12 },
		{ 246, 239, 12 },
		{ 246, 240, 12 },
		{ 246, 241, 12 },
		{ 246, 242, 12 },
		{ 246, 243, 12 },
		{ 246, 244, 12 },
		{ 246, 245, 12 },
		{ 246, 246, 12 },
		{ 246, 247, 12 },
		{ 246, 248, 12 },
		{ 246, 249, 12 },
		{ 246, 250, 12 },
		{ 246, 251, 12 },
		{ 247, 238, 12 },
		{ 247, 239, 12 },
		{ 247, 240, 12 },
		{ 247, 241, 12 },
		{ 247, 242, 12 },
		{ 247, 243, 12 },
		{ 247, 244, 12 },
		{ 247, 245, 12 },
		{ 247, 246, 12 },
		{ 247, 247, 12 },
		{ 247, 248, 12 },
		{ 247, 249, 12 },
		{ 247, 250, 12 },
		{ 247, 251, 12 },
		{ 248, 240, 12 },
		{ 248, 241, 12 },
		{ 248, 242, 12 },
		{ 248, 243, 12 },
		{ 248, 244, 12 },
		{ 248, 245, 12 },
		{ 248, 246, 12 },
		{ 248, 247, 12 },
		{ 248, 248, 12 },
		{ 248, 249, 12 },
		{ 248, 250, 12 },
		{ 248, 251, 12 },
		{ 248, 252, 12 },
		{ 249, 242, 12 },
		{ 249, 243, 12 },
		{ 249, 244, 12 },
		{ 249, 245, 12 },
		{ 249, 246, 12 },
		{ 249, 247, 12 },
		{ 249, 248, 12 },
		{ 249, 249, 12 },
		{ 249, 250, 12 },
		{ 249, 251, 12 },
		{ 249, 252, 12 },
		{ 250, 244, 12 },
		{ 250, 245, 12 },
		{ 250, 246, 12 },
		{ 250, 247, 12 },
		{ 250, 248, 12 },
		{ 250, 249, 12 },
		{ 250, 250, 12 },
		{ 250, 251, 12 },
		{ 250, 252, 12 },
		{ 250, 253, 12 },
		{ 251, 246, 12 },
		{ 251, 247, 12 },
		{ 251, 248, 12 },
		{ 251, 249, 12 },
		{ 251, 250, 12 },
		{ 251, 251, 12 },
		{ 251, 252, 12 },
		{ 251, 253, 12 },
		{ 252, 248, 12 },
		{ 252, 249, 12 },
		{ 252, 250, 12 },
		{ 252, 251, 12 },
		{ 252, 252, 12 },
		{ 252, 253, 12 },
		{ 252, 254, 12 },
		{ 253, 250, 12 },
		{ 253, 251, 12 },
		{ 253, 252, 12 },
		{ 254, 252, 12 },
		{ 260, 262, 12 },
		{ 261, 262, 12 },
		{ 261, 263, 12 },
		{ 261, 264, 12 },
		{ 262, 260, 12 },
		{ 262, 261, 12 },
		{ 262, 262, 12 },
		{ 262, 263, 12 },
		{ 262, 264, 12 },
		{ 262, 265, 12 },
		{ 262, 266, 12 },
		{ 263, 261, 12 },
		{ 263, 262, 12 },
		{ 263, 263, 12 },
		{ 263, 264, 12 },
		{ 263, 265, 12 },
		{ 263, 266, 12 },
		{ 263, 267, 12 },
		{ 263, 268, 12 },
		{ 264, 261, 12 },
		{ 264, 262, 12 },
		{ 264, 263, 12 },
		{ 264, 264, 12 },
		{ 264, 265, 12 },
		{ 264, 266, 12 },
		{ 264, 267, 12 },
		{ 264, 268, 12 },
		{ 264, 269, 12 },
		{ 264, 270, 12 },
		{ 265, 262, 12 },
		{ 265, 263, 12 },
		{ 265, 264, 12 },
		{ 265, 265, 12 },
		{ 265, 266, 12 },
		{ 265, 267, 12 },
		{ 265, 268, 12 },
		{ 265, 269, 12 },
		{ 265, 270, 12 },
		{ 265, 271, 12 },
		{ 265, 272, 12 },
		{ 266, 262, 12 },
		{ 266, 263, 12 },
		{ 266, 264, 12 },
		{ 266, 265, 12 },
		{ 266, 266, 12 },
		{ 266, 267, 12 },
		{ 266, 268, 12 },
		{ 266, 269, 12 },
		{ 266, 270, 12 },
		{ 266, 271, 12 },
		{ 266, 272, 12 },
		{ 266, 273, 12 },
		{ 266, 274, 12 },
		{ 267, 263, 12 },
		{ 267, 264, 12 },
		{ 267, 265, 12 },
		{ 267, 266, 12 },
		{ 267, 267, 12 },
		{ 267, 268, 12 },
		{ 267, 269, 12 },
		{ 267, 270, 12 },
		{ 267, 271, 12 },
		{ 267, 272, 12 },
		{ 267, 273, 12 },
		{ 267, 274, 12 },
		{ 267, 275, 12 },
		{ 267, 276, 12 },
		{ 268, 263, 12 },
		{ 268, 264, 12 },
		{ 268, 265, 12 },
		{ 268, 266, 12 },
		{ 268, 267, 12 },
		{ 268, 268, 12 },
		{ 268, 269, 12 },
		{ 268, 270, 12 },
		{ 268, 271, 12 },
		{ 268, 272, 12 },
		{ 268, 273, 12 },
		{ 268, 274, 12 },
		{ 268, 275, 12 },
		{ 268, 276, 12 },
		{ 268, 277, 12 },
		{ 268, 278, 12 },
		{ 269, 264, 12 },
		{ 269, 265, 12 },
		{ 269, 266, 12 },
		{ 269, 267, 12 },
		{ 269, 268, 12 },
		{ 269, 269, 12 },
		{ 269, 270, 12 },
		{ 269, 271, 12 },
		{ 269, 272, 12 },
		{ 269, 273, 12 },
		{ 269, 274, 12 },
		{ 269, 275, 12 },
		{ 269, 276, 12 },
		{ 269, 277, 12 },
		{ 269, 278, 12 },
		{ 269, 279, 12 },
		{ 269, 280, 12 },
		{ 270, 264, 12 },
		{ 270, 265, 12 },
		{ 270, 266, 12 },
		{ 270, 267, 12 },
		{ 270, 268, 12 },
		{ 270, 269, 12 },
		{ 270, 270, 12 },
		{ 270, 271, 12 },
		{ 270, 272, 12 },
		{ 270, 273, 12 },
		{ 270, 274, 12 },
		{ 270, 275, 12 },
		{ 270, 276, 12 },
		{ 270, 277, 12 },
		{ 270, 278, 12 },
		{ 270, 279, 12 },
		{ 270, 280, 12 },
		{ 270, 281, 12 },
		{ 270, 282, 12 },
		{ 271, 265, 12 },
		{ 271, 266, 12 },
		{ 271, 267, 12 },
		{ 271, 268, 12 },
		{ 271, 269, 12 },
		{ 271, 270, 12 },
		{ 271, 271, 12 },
		{ 271, 272, 12 },
		{ 271, 273, 12 },
		{ 271, 274, 12 },
		{ 271, 275, 12 },
		{ 271, 276, 12 },
		{ 271, 277, 12 },
		{ 271, 278, 12 },
		{ 271, 279, 12 },
		{ 271, 280, 12 },
		{ 271, 281, 12 },
		{ 271, 282, 12 },
		{ 271, 283, 12 },
		{ 271, 284, 12 },
		{ 272, 265, 12 },
		{ 272, 266, 12 },
		{ 272, 267, 12 },
		{ 272, 268, 12 },
		{ 272, 269, 12 },
		{ 272, 270, 12 },
		{ 272, 271, 12 },
		{ 272, 272, 12 },
		{ 272, 273, 12 },
		{ 272, 274, 12 },
		{ 272, 275, 12 },
		{ 272, 276, 12 },
		{ 272, 277, 12 },
		{ 272, 278, 12 },
		{ 272, 279, 12 },
		{ 272, 280, 12 },
		{ 272, 281, 12 },
		{ 272, 282, 12 },
		{ 272, 283, 12 },
		{ 272, 284, 12 },
		{ 272, 285, 12 },
		{ 272, 286, 12 },
		{ 273, 266, 12 },
		{ 273, 267, 12 },
		{ 273, 268, 12 },
		{ 273, 269, 12 },
		{ 273, 270, 12 },
		{ 273, 271, 12 },
		{ 273, 272, 12 },
		{ 273, 273, 12 },
		{ 273, 274, 12 },
		{ 273, 275, 12 },
		{ 273, 276, 12 },
		{ 273, 277, 12 },
		{ 273, 278, 12 },
		{ 273, 279, 12 },
		{ 273, 280, 12 },
		{ 273, 281, 12 },
		{ 273, 282, 12 },
		{ 273, 283, 12 },
		{ 273, 284, 12 },
		{ 273, 285, 12 },
		{ 273, 286, 12 },
		{ 273, 287, 12 },
		{ 274, 266, 12 },
		{ 274, 267, 12 },
		{ 274, 268, 12 },
		{ 274, 269, 12 },
		{ 274, 270, 12 },
		{ 274, 271, 12 },
		{ 274, 272, 12 },
		{ 274, 273, 12 },
		{ 274, 274, 12 },
		{ 274, 275, 12 },
		{ 274, 276, 12 },
		{ 274, 277, 12 },
		{ 274, 278, 12 },
		{ 274, 279, 12 },
		{ 274, 280, 12 },
		{ 274, 281, 12 },
		{ 274, 282, 12 },
		{ 274, 283, 12 },
		{ 274, 284, 12 },
		{ 274, 285, 12 },
		{ 274, 286, 12 },
		{ 274, 287, 12 },
		{ 274, 288, 12 },
		{ 275, 267, 12 },
		{ 275, 268, 12 },
		{ 275, 269, 12 },
		{ 275, 270, 12 },
		{ 275, 271, 12 },
		{ 275, 272, 12 },
		{ 275, 273, 12 },
		{ 275, 274, 12 },
		{ 275, 275, 12 },
		{ 275, 276, 12 },
		{ 275, 277, 12 },
		{ 275, 278, 12 },
		{ 275, 279, 12 },
		{ 275, 280, 12 },
		{ 275, 281, 12 },
		{ 275, 282, 12 },
		{ 275, 283, 12 },
		{ 275, 284, 12 },
		{ 275, 285, 12 },
		{ 275, 286, 12 },
		{ 275, 287, 12 },
		{ 275, 288, 12 },
		{ 276, 267, 12 },
		{ 276, 268, 12 },
		{ 276, 269, 12 },
		{ 276, 270, 12 },
		{ 276, 271, 12 },
		{ 276, 272, 12 },
		{ 276, 273, 12 },
		{ 276, 274, 12 },
		{ 276, 275, 12 },
		{ 276, 276, 12 },
		{ 276, 277, 12 },
		{ 276, 278, 12 },
		{ 276, 279, 12 },
		{ 276, 280, 12 },
		{ 276, 281, 12 },
		{ 276, 282, 12 },
		{ 276, 283, 12 },
		{ 276, 284, 12 },
		{ 276, 285, 12 },
		{ 276, 286, 12 },
		{ 276, 287, 12 },
		{ 276, 288, 12 },
		{ 277, 268, 12 },
		{ 277, 269, 12 },
		{ 277, 270, 12 },
		{ 277, 271, 12 },
		{ 277, 272, 12 },
		{ 277, 273, 12 },
		{ 277, 274, 12 },
		{ 277, 275, 12 },
		{ 277, 276, 12 },
		{ 277, 277, 12 },
		{ 277, 278, 12 },
		{ 277, 279, 12 },
		{ 277, 280, 12 },
		{ 277, 281, 12 },
		{ 277, 282, 12 },
		{ 277, 283, 12 },
		{ 277, 284, 12 },
		{ 277, 285, 12 },
		{ 277, 286, 12 },
		{ 277, 287, 12 },
		{ 277, 288, 12 },
		{ 278, 268, 12 },
		{ 278, 269, 12 },
		{ 278, 270, 12 },
		{ 278, 271, 12 },
		{ 278, 272, 12 },
		{ 278, 273, 12 },
		{ 278, 274, 12 },
		{ 278, 275, 12 },
		{ 278, 276, 12 },
		{ 278, 277, 12 },
		{ 278, 278, 12 },
		{ 278, 279, 12 },
		{ 278, 280, 12 },
		{ 278, 281, 12 },
		{ 278, 282, 12 },
		{ 278, 283, 12 },
		{ 278, 284, 12 },
		{ 278, 285, 12 },
		{ 278, 286, 12 },
		{ 278, 287, 12 },
		{ 278, 288, 12 },
		{ 279, 269, 12 },
		{ 279, 270, 12 },
		{ 279, 271, 12 },
		{ 279, 272, 12 },
		{ 279, 273, 12 },
		{ 279, 274, 12 },
		{ 279, 275, 12 },
		{ 279, 276, 12 },
		{ 279, 277, 12 },
		{ 279, 278, 12 },
		{ 279, 279, 12 },
		{ 279, 280, 12 },
		{ 279, 281, 12 },
		{ 279, 282, 12 },
		{ 279, 283, 12 },
		{ 279, 284, 12 },
		{ 279, 285, 12 },
		{ 279, 286, 12 },
		{ 279, 287, 12 },
		{ 279, 288, 12 },
		{ 280, 269, 12 },
		{ 280, 270, 12 },
		{ 280, 271, 12 },
		{ 280, 272, 12 },
		{ 280, 273, 12 },
		{ 280, 274, 12 },
		{ 280, 275, 12 },
		{ 280, 276, 12 },
		{ 280, 277, 12 },
		{ 280, 278, 12 },
		{ 280, 279, 12 },
		{ 280, 280, 12 },
		{ 280, 281, 12 },
		{ 280, 282, 12 },
		{ 280, 283, 12 },
		{ 280, 284, 12 },
		{ 280, 285, 12 },
		{ 280, 286, 12 },
		{ 280, 287, 12 },
		{ 280, 288, 12 },
		{ 281, 270, 12 },
		{ 281, 271, 12 },
		{ 281, 272, 12 },
		{ 281, 273, 12 },
		{ 281, 274, 12 },
		{ 281, 275, 12 },
		{ 281, 276, 12 },
		{ 281, 277, 12 },
		{ 281, 278, 12 },
		{ 281, 279, 12 },
		{ 281, 280, 12 },
		{ 281, 281, 12 },
		{ 281, 282, 12 },
		{ 281, 283, 12 },
		{ 281, 284, 12 },
		{ 281, 285, 12 },
		{ 281, 286, 12 },
		{ 281, 287, 12 },
		{ 281, 288, 12 },
		{ 282, 270, 12 },
		{ 282, 271, 12 },
		{ 282, 272, 12 },
		{ 282, 273, 12 },
		{ 282, 274, 12 },
		{ 282, 275, 12 },
		{ 282, 276, 12 },
		{ 282, 277, 12 },
		{ 282, 278, 12 },
		{ 282, 279, 12 },
		{ 282, 280, 12 },
		{ 282, 281, 12 },
		{ 282, 282, 12 },
		{ 282, 283, 12 },
		{ 282, 284, 12 },
		{ 282, 285, 12 },
		{ 282, 286, 12 },
		{ 282, 287, 12 },
		{ 282, 288, 12 },
		{ 283, 271, 12 },
		{ 283, 272, 12 },
		{ 283, 273, 12 },
		{ 283, 274, 12 },
		{ 283, 275, 12 },
		{ 283, 276, 12 },
		{ 283, 277, 12 },
		{ 283, 278, 12 },
		{ 283, 279, 12 },
		{ 283, 280, 12 },
		{ 283, 281, 12 },
		{ 283, 282, 12 },
		{ 283, 283, 12 },
		{ 283, 284, 12 },
		{ 283, 285, 12 },
		{ 283, 286, 12 },
		{ 283, 287, 12 },
		{ 283, 288, 12 },
		{ 284, 271, 12 },
		{ 284, 272, 12 },
		{ 284, 273, 12 },
		{ 284, 274, 12 },
		{ 284, 275, 12 },
		{ 284, 276, 12 },
		{ 284, 277, 12 },
		{ 284, 278, 12 },
		{ 284, 279, 12 },
		{ 284, 280, 12 },
		{ 284, 281, 12 },
		{ 284, 282, 12 },
		{ 284, 283, 12 },
		{ 284, 284, 12 },
		{ 284, 285, 12 },
		{ 284, 286, 12 },
		{ 284, 287, 12 },
		{ 284, 288, 12 },
		{ 285, 272, 12 },
		{ 285, 273, 12 },
		{ 285, 274, 12 },
		{ 285, 275, 12 },
		{ 285, 276, 12 },
		{ 285, 277, 12 },
		{ 285, 278, 12 },
		{ 285, 279, 12 },
		{ 285, 280, 12 },
		{ 285, 281, 12 },
		{ 285, 282, 12 },
		{ 285, 283, 12 },
		{ 285, 284, 12 },
		{ 285, 285, 12 },
		{ 285, 286, 12 },
		{ 285, 287, 12 },
		{ 285, 288, 12 },
		{ 286, 272, 12 },
		{ 286, 273, 12 },
		{ 286, 274, 12 },
		{ 286, 275, 12 },
		{ 286, 276, 12 },
		{ 286, 277, 12 },
		{ 286, 278, 12 },
		{ 286, 279, 12 },
		{ 286, 280, 12 },
		{ 286, 281, 12 },
		{ 286, 282, 12 },
		{ 286, 283, 12 },
		{ 286, 284, 12 },
		{ 286, 285, 12 },
		{ 286, 286, 12 },
		{ 286, 287, 12 },
		{ 286, 288, 12 },
		{ 287, 273, 12 },
		{ 287, 274, 12 },
		{ 287, 275, 12 },
		{ 287, 276, 12 },
		{ 287, 277, 12 },
		{ 287, 278, 12 },
		{ 287, 279, 12 },
		{ 287, 280, 12 },
		{ 287, 281, 12 },
		{ 287, 282, 12 },
		{ 287, 283, 12 },
		{ 287, 284, 12 },
		{ 287, 285, 12 },
		{ 287, 286, 12 },
		{ 287, 287, 12 },
		{ 287, 288, 12 },
		{ 288, 274, 12 },
		{ 288, 275, 12 },
		{ 288, 276, 12 },
		{ 288, 277, 12 },
		{ 288, 278, 12 },
		{ 288, 279, 12 },
		{ 288, 280, 12 },
		{ 288, 281, 12 },
		{ 288, 282, 12 },
		{ 288, 283, 12 },
		{ 288, 284, 12 },
		{ 288, 285, 12 },
		{ 288, 286, 12 },
		{ 288, 287, 12 },
		{ 288, 288, 12 },
		{ 227, 226, 13 },
		{ 228, 226, 13 },
		{ 228, 227, 13 },
		{ 229, 226, 13 },
		{ 229, 227, 13 },
		{ 229, 228, 13 },
		{ 230, 226, 13 },
		{ 230, 227, 13 },
		{ 230, 228, 13 },
		{ 230, 229, 13 },
		{ 231, 226, 13 },
		{ 231, 227, 13 },
		{ 231, 228, 13 },
		{ 231, 229, 13 },
		{ 231, 230, 13 },
		{ 232, 226, 13 },
		{ 232, 227, 13 },
		{ 232, 228, 13 },
		{ 232, 229, 13 },
		{ 232, 230, 13 },
		{ 232, 231, 13 },
		{ 233, 226, 13 },
		{ 233, 227, 13 },
		{ 233, 228, 13 },
		{ 233, 229, 13 },
		{ 233, 230, 13 },
		{ 233, 231, 13 },
		{ 233, 232, 13 },
		{ 234, 226, 13 },
		{ 234, 227, 13 },
		{ 234, 228, 13 },
		{ 234, 229, 13 },
		{ 234, 230, 13 },
		{ 234, 231, 13 },
		{ 234, 232, 13 },
		{ 234, 233, 13 },
		{ 235, 226, 13 },
		{ 235, 227, 13 },
		{ 235, 228, 13 },
		{ 235, 229, 13 },
		{ 235, 230, 13 },
		{ 235, 231, 13 },
		{ 235, 232, 13 },
		{ 235, 233, 13 },
		{ 235, 234, 13 },
		{ 236, 226, 13 },
		{ 236, 227, 13 },
		{ 236, 228, 13 },
		{ 236, 229, 13 },
		{ 236, 230, 13 },
		{ 236, 231, 13 },
		{ 236, 232, 13 },
		{ 236, 233, 13 },
		{ 236, 234, 13 },
		{ 236, 235, 13 },
		{ 237, 226, 13 },
		{ 237, 227, 13 },
		{ 237, 228, 13 },
		{ 237, 229, 13 },
		{ 237, 230, 13 },
		{ 237, 231, 13 },
		{ 237, 232, 13 },
		{ 237, 233, 13 },
		{ 237, 234, 13 },
		{ 237, 235, 13 },
		{ 237, 236, 13 },
		{ 238, 226, 13 },
		{ 238, 227, 13 },
		{ 238, 228, 13 },
		{ 238, 229, 13 },
		{ 238, 230, 13 },
		{ 238, 231, 13 },
		{ 238, 232, 13 },
		{ 238, 233, 13 },
		{ 238, 234, 13 },
		{ 238, 235, 13 },
		{ 238, 236, 13 },
		{ 238, 237, 13 },
		{ 239, 226, 13 },
		{ 239, 227, 13 },
		{ 239, 228, 13 },
		{ 239, 229, 13 },
		{ 239, 230, 13 },
		{ 239, 231, 13 },
		{ 239, 232, 13 },
		{ 239, 233, 13 },
		{ 239, 234, 13 },
		{ 239, 235, 13 },
		{ 239, 236, 13 },
		{ 239, 237, 13 },
		{ 239, 238, 13 },
		{ 240, 226, 13 },
		{ 240, 227, 13 },
		{ 240, 228, 13 },
		{ 240, 229, 13 },
		{ 240, 230, 13 },
		{ 240, 231, 13 },
		{ 240, 232, 13 },
		{ 240, 233, 13 },
		{ 240, 234, 13 },
		{ 240, 235, 13 },
		{ 240, 236, 13 },
		{ 240, 237, 13 },
		{ 240, 238, 13 },
		{ 240, 239, 13 },
		{ 241, 226, 13 },
		{ 241, 227, 13 },
		{ 241, 228, 13 },
		{ 241, 229, 13 },
		{ 241, 230, 13 },
		{ 241, 231, 13 },
		{ 241, 232, 13 },
		{ 241, 233, 13 },
		{ 241, 234, 13 },
		{ 241, 235, 13 },
		{ 241, 236, 13 },
		{ 241, 237, 13 },
		{ 241, 238, 13 },
		{ 241, 239, 13 },
		{ 241, 240, 13 },
		{ 242, 226, 13 },
		{ 242, 227, 13 },
		{ 242, 228, 13 },
		{ 242, 229, 13 },
		{ 242, 230, 13 },
		{ 242, 231, 13 },
		{ 242, 232, 13 },
		{ 242, 233, 13 },
		{ 242, 234, 13 },
		{ 242, 235, 13 },
		{ 242, 236, 13 },
		{ 242, 237, 13 },
		{ 242, 238, 13 },
		{ 242, 239, 13 },
		{ 242, 240, 13 },
		{ 242, 241, 13 },
		{ 243, 226, 13 },
		{ 243, 227, 13 },
		{ 243, 228, 13 },
		{ 243, 229, 13 },
		{ 243, 230, 13 },
		{ 243, 231, 13 },
		{ 243, 232, 13 },
		{ 243, 233, 13 },
		{ 243, 234, 13 },
		{ 243, 235, 13 },
		{ 243, 236, 13 },
		{ 243, 237, 13 },
		{ 243, 238, 13 },
		{ 243, 239, 13 },
		{ 243, 240, 13 },
		{ 243, 241, 13 },
		{ 243, 242, 13 },
		{ 244, 226, 13 },
		{ 244, 227, 13 },
		{ 244, 228, 13 },
		{ 244, 229, 13 },
		{ 244, 230, 13 },
		{ 244, 231, 13 },
		{ 244, 232, 13 },
		{ 244, 233, 13 },
		{ 244, 234, 13 },
		{ 244, 235, 13 },
		{ 244, 236, 13 },
		{ 244, 237, 13 },
		{ 244, 238, 13 },
		{ 244, 239, 13 },
		{ 244, 240, 13 },
		{ 244, 241, 13 },
		{ 244, 242, 13 },
		{ 244, 243, 13 },
		{ 245, 226, 13 },
		{ 245, 227, 13 },
		{ 245, 228, 13 },
		{ 245, 229, 13 },
		{ 245, 230, 13 },
		{ 245, 231, 13 },
		{ 245, 232, 13 },
		{ 245, 233, 13 },
		{ 245, 234, 13 },
		{ 245, 235, 13 },
		{ 245, 236, 13 },
		{ 245, 237, 13 },
		{ 245, 238, 13 },
		{ 245, 239, 13 },
		{ 245, 240, 13 },
		{ 245, 241, 13 },
		{ 245, 242, 13 },
		{ 245, 243, 13 },
		{ 245, 244, 13 },
		{ 246, 226, 13 },
		{ 246, 227, 13 },
		{ 246, 228, 13 },
		{ 246, 229, 13 },
		{ 246, 230, 13 },
		{ 246, 231, 13 },
		{ 246, 232, 13 },
		{ 246, 233, 13 },
		{ 246, 234, 13 },
		{ 246, 235, 13 },
		{ 246, 236, 13 },
		{ 246, 237, 13 },
		{ 246, 238, 13 },
		{ 246, 239, 13 },
		{ 246, 240, 13 },
		{ 246, 241, 13 },
		{ 246, 242, 13 },
		{ 246, 243, 13 },
		{ 246, 244, 13 },
		{ 246, 245, 13 },
		{ 247, 226, 13 },
		{ 247, 227, 13 },
		{ 247, 228, 13 },
		{ 247, 229, 13 },
		{ 247, 230, 13 },
		{ 247, 231, 13 },
		{ 247, 232, 13 },
		{ 247, 233, 13 },
		{ 247, 234, 13 },
		{ 247, 235, 13 },
		{ 247, 236, 13 },
		{ 247, 237, 13 },
		{ 247, 238, 13 },
		{ 247, 239, 13 },
		{ 247, 240, 13 },
		{ 247, 241, 13 },
		{ 247, 242, 13 },
		{ 247, 243, 13 },
		{ 247, 244, 13 },
		{ 247, 245, 13 },
		{ 247, 246, 13 },
		{ 248, 226, 13 },
		{ 248, 227, 13 },
		{ 248, 228, 13 },
		{ 248, 229, 13 },
		{ 248, 230, 13 },
		{ 248, 231, 13 },
		{ 248, 232, 13 },
		{ 248, 233, 13 },
		{ 248, 234, 13 },
		{ 248, 235, 13 },
		{ 248, 236, 13 },
		{ 248, 237, 13 },
		{ 248, 238, 13 },
		{ 248, 239, 13 },
		{ 248, 240, 13 },
		{ 248, 241, 13 },
		{ 248, 242, 13 },
		{ 248, 243, 13 },
		{ 248, 244, 13 },
		{ 248, 245, 13 },
		{ 248, 246, 13 },
		{ 248, 247, 13 },
		{ 249, 226, 13 },
		{ 249, 227, 13 },
		{ 249, 228, 13 },
		{ 249, 229, 13 },
		{ 249, 230, 13 },
		{ 249, 231, 13 },
		{ 249, 232, 13 },
		{ 249, 233, 13 },
		{ 249, 234, 13 },
		{ 249, 235, 13 },
		{ 249, 236, 13 },
		{ 249, 237, 13 },
		{ 249, 238, 13 },
		{ 249, 239, 13 },
		{ 249, 240, 13 },
		{ 249, 241, 13 },
		{ 249, 242, 13 },
		{ 249, 243, 13 },
		{ 249, 244, 13 },
		{ 249, 245, 13 },
		{ 249, 246, 13 },
		{ 249, 247, 13 },
		{ 249, 248, 13 },
		{ 250, 226, 13 },
		{ 250, 227, 13 },
		{ 250, 228, 13 },
		{ 250, 229, 13 },
		{ 250, 230, 13 },
		{ 250, 231, 13 },
		{ 250, 232, 13 },
		{ 250, 233, 13 },
		{ 250, 234, 13 },
		{ 250, 235, 13 },
		{ 250, 236, 13 },
		{ 250, 237, 13 },
		{ 250, 238, 13 },
		{ 250, 239, 13 },
		{ 250, 240, 13 },
		{ 250, 241, 13 },
		{ 250, 242, 13 },
		{ 250, 243, 13 },
		{ 250, 244, 13 },
		{ 250, 245, 13 },
		{ 250, 246, 13 },
		{ 250, 247, 13 },
		{ 250, 248, 13 },
		{ 250, 249, 13 },
		{ 251, 226, 13 },
		{ 251, 227, 13 },
		{ 251, 228, 13 },
		{ 251, 229, 13 },
		{ 251, 230, 13 },
		{ 251, 231, 13 },
		{ 251, 232, 13 },
		{ 251, 233, 13 },
		{ 251, 234, 13 },
		{ 251, 235, 13 },
		{ 251, 236, 13 },
		{ 251, 237, 13 },
		{ 251, 238, 13 },
		{ 251, 239, 13 },
		{ 251, 240, 13 },
		{ 251, 241, 13 },
		{ 251, 242, 13 },
		{ 251, 243, 13 },
		{ 251, 244, 13 },
		{ 251, 245, 13 },
		{ 251, 246, 13 },
		{ 251, 247, 13 },
		{ 251, 248, 13 },
		{ 251, 249, 13 },
		{ 251, 250, 13 },
		{ 252, 226, 13 },
		{ 252, 227, 13 },
		{ 252, 228, 13 },
		{ 252, 229, 13 },
		{ 252, 230, 13 },
		{ 252, 231, 13 },
		{ 252, 232, 13 },
		{ 252, 233, 13 },
		{ 252, 234, 13 },
		{ 252, 235, 13 },
		{ 252, 236, 13 },
		{ 252, 237, 13 },
		{ 252, 238, 13 },
		{ 252, 239, 13 },
		{ 252, 240, 13 },
		{ 252, 241, 13 },
		{ 252, 242, 13 },
		{ 252, 243, 13 },
		{ 252, 244, 13 },
		{ 252, 245, 13 },
		{ 252, 246, 13 },
		{ 252, 247, 13 },
		{ 252, 248, 13 },
		{ 252, 249, 13 },
		{ 252, 250, 13 },
		{ 252, 251, 13 },
		{ 253, 226, 13 },
		{ 253, 227, 13 },
		{ 253, 228, 13 },
		{ 253, 229, 13 },
		{ 253, 230, 13 },
		{ 253, 231, 13 },
		{ 253, 232, 13 },
		{ 253, 233, 13 },
		{ 253, 234, 13 },
		{ 253, 235, 13 },
		{ 253, 236, 13 },
		{ 253, 237, 13 },
		{ 253, 238, 13 },
		{ 253, 239, 13 },
		{ 253, 240, 13 },
		{ 253, 241, 13 },
		{ 253, 242, 13 },
		{ 253, 243, 13 },
		{ 253, 244, 13 },
		{ 253, 245, 13 },
		{ 253, 246, 13 },
		{ 253, 247, 13 },
		{ 253, 248, 13 },
		{ 253, 249, 13 },
		{ 253, 250, 13 },
		{ 253, 251, 13 },
		{ 253, 252, 13 },
		{ 254, 226, 13 },
		{ 254, 227, 13 },
		{ 254, 228, 13 },
		{ 254, 229, 13 },
		{ 254, 230, 13 },
		{ 254, 231, 13 },
		{ 254, 232, 13 },
		{ 254, 233, 13 },
		{ 254, 234, 13 },
		{ 254, 235, 13 },
		{ 254, 236, 13 },
		{ 254, 237, 13 },
		{ 254, 238, 13 },
		{ 254, 239, 13 },
		{ 254, 240, 13 },
		{ 254, 241, 13 },
		{ 254, 242, 13 },
		{ 254, 243, 13 },
		{ 254, 244, 13 },
		{ 254, 245, 13 },
		{ 254, 246, 13 },
		{ 254, 247, 13 },
		{ 254, 248, 13 },
		{ 254, 249, 13 },
		{ 254, 250, 13 },
		{ 254, 251, 13 },
		{ 254, 252, 13 },
		{ 255, 226, 13 },
		{ 255, 227, 13 },
		{ 255, 228, 13 },
		{ 255, 229, 13 },
		{ 255, 230, 13 },
		{ 255, 231, 13 },
		{ 255, 232, 13 },
		{ 255, 233, 13 },
		{ 255, 234, 13 },
		{ 255, 235, 13 },
		{ 255, 236, 13 },
		{ 255, 237, 13 },
		{ 255, 238, 13 },
		{ 255, 239, 13 },
		{ 255, 240, 13 },
		{ 255, 241, 13 },
		{ 255, 242, 13 },
		{ 255, 243, 13 },
		{ 255, 244, 13 },
		{ 255, 245, 13 },
		{ 255, 246, 13 },
		{ 255, 247, 13 },
		{ 255, 248, 13 },
		{ 255, 249, 13 },
		{ 255, 250, 13 },
		{ 255, 251, 13 },
		{ 255, 252, 13 },
		{ 256, 226, 13 },
		{ 256, 227, 13 },
		{ 256, 228, 13 },
		{ 256, 229, 13 },
		{ 256, 230, 13 },
		{ 256, 231, 13 },
		{ 256, 232, 13 },
		{ 256, 233, 13 },
		{ 256, 234, 13 },
		{ 256, 235, 13 },
		{ 256, 236, 13 },
		{ 256, 237, 13 },
		{ 256, 238, 13 },
		{ 256, 239, 13 },
		{ 256, 240, 13 },
		{ 256, 241, 13 },
		{ 256, 242, 13 },
		{ 256, 243, 13 },
		{ 256, 244, 13 },
		{ 256, 245, 13 },
		{ 256, 246, 13 },
		{ 256, 247, 13 },
		{ 256, 248, 13 },
		{ 256, 249, 13 },
		{ 256, 250, 13 },
		{ 256, 251, 13 },
		{ 256, 252, 13 },
		{ 258, 262, 13 },
		{ 258, 263, 13 },
		{ 258, 264, 13 },
		{ 258, 265, 13 },
		{ 258, 266, 13 },
		{ 258, 267, 13 },
		{ 258, 268, 13 },
		{ 258, 269, 13 },
		{ 258, 270, 13 },
		{ 258, 271, 13 },
		{ 258, 272, 13 },
		{ 258, 273, 13 },
		{ 258, 274, 13 },
		{ 258, 275, 13 },
		{ 258, 276, 13 },
		{ 258, 277, 13 },
		{ 258, 278, 13 },
		{ 258, 279, 13 },
		{ 258, 280, 13 },
		{ 258, 281, 13 },
		{ 258, 282, 13 },
		{ 258, 283, 13 },
		{ 258, 284, 13 },
		{ 258, 285, 13 },
		{ 258, 286, 13 },
		{ 258, 287, 13 },
		{ 258, 288, 13 },
		{ 259, 262, 13 },
		{ 259, 263, 13 },
		{ 259, 264, 13 },
		{ 259, 265, 13 },
		{ 259, 266, 13 },
		{ 259, 267, 13 },
		{ 259, 268, 13 },
		{ 259, 269, 13 },
		{ 259, 270, 13 },
		{ 259, 271, 13 },
		{ 259, 272, 13 },
		{ 259, 273, 13 },
		{ 259, 274, 13 },
		{ 259, 275, 13 },
		{ 259, 276, 13 },
		{ 259, 277, 13 },
		{ 259, 278, 13 },
		{ 259, 279, 13 },
		{ 259, 280, 13 },
		{ 259, 281, 13 },
		{ 259, 282, 13 },
		{ 259, 283, 13 },
		{ 259, 284, 13 },
		{ 259, 285, 13 },
		{ 259, 286, 13 },
		{ 259, 287, 13 },
		{ 259, 288, 13 },
		{ 260, 262, 13 },
		{ 260, 263, 13 },
		{ 260, 264, 13 },
		{ 260, 265, 13 },
		{ 260, 266, 13 },
		{ 260, 267, 13 },
		{ 260, 268, 13 },
		{ 260, 269, 13 },
		{ 260, 270, 13 },
		{ 260, 271, 13 },
		{ 260, 272, 13 },
		{ 260, 273, 13 },
		{ 260, 274, 13 },
		{ 260, 275, 13 },
		{ 260, 276, 13 },
		{ 260, 277, 13 },
		{ 260, 278, 13 },
		{ 260, 279, 13 },
		{ 260, 280, 13 },
		{ 260, 281, 13 },
		{ 260, 282, 13 },
		{ 260, 283, 13 },
		{ 260, 284, 13 },
		{ 260, 285, 13 },
		{ 260, 286, 13 },
		{ 260, 287, 13 },
		{ 260, 288, 13 },
		{ 261, 262, 13 },
		{ 261, 263, 13 },
		{ 261, 264, 13 },
		{ 261, 265, 13 },
		{ 261, 266, 13 },
		{ 261, 267, 13 },
		{ 261, 268, 13 },
		{ 261, 269, 13 },
		{ 261, 270, 13 },
		{ 261, 271, 13 },
		{ 261, 272, 13 },
		{ 261, 273, 13 },
		{ 261, 274, 13 },
		{ 261, 275, 13 },
		{ 261, 276, 13 },
		{ 261, 277, 13 },
		{ 261, 278, 13 },
		{ 261, 279, 13 },
		{ 261, 280, 13 },
		{ 261, 281, 13 },
		{ 261, 282, 13 },
		{ 261, 283, 13 },
		{ 261, 284, 13 },
		{ 261, 285, 13 },
		{ 261, 286, 13 },
		{ 261, 287, 13 },
		{ 261, 288, 13 },
		{ 262, 263, 13 },
		{ 262, 264, 13 },
		{ 262, 265, 13 },
		{ 262, 266, 13 },
		{ 262, 267, 13 },
		{ 262, 268, 13 },
		{ 262, 269, 13 },
		{ 262, 270, 13 },
		{ 262, 271, 13 },
		{ 262, 272, 13 },
		{ 262, 273, 13 },
		{ 262, 274, 13 },
		{ 262, 275, 13 },
		{ 262, 276, 13 },
		{ 262, 277, 13 },
		{ 262, 278, 13 },
		{ 262, 279, 13 },
		{ 262, 280, 13 },
		{ 262, 281, 13 },
		{ 262, 282, 13 },
		{ 262, 283, 13 },
		{ 262, 284, 13 },
		{ 262, 285, 13 },
		{ 262, 286, 13 },
		{ 262, 287, 13 },
		{ 262, 288, 13 },
		{ 263, 264, 13 },
		{ 263, 265, 13 },
		{ 263, 266, 13 },
		{ 263, 267, 13 },
		{ 263, 268, 13 },
		{ 263, 269, 13 },
		{ 263, 270, 13 },
		{ 263, 271, 13 },
		{ 263, 272, 13 },
		{ 263, 273, 13 },
		{ 263, 274, 13 },
		{ 263, 275, 13 },
		{ 263, 276, 13 },
		{ 263, 277, 13 },
		{ 263, 278, 13 },
		{ 263, 279, 13 },
		{ 263, 280, 13 },
		{ 263, 281, 13 },
		{ 263, 282, 13 },
		{ 263, 283, 13 },
		{ 263, 284, 13 },
		{ 263, 285, 13 },
		{ 263, 286, 13 },
		{ 263, 287, 13 },
		{ 263, 288, 13 },
		{ 264, 265, 13 },
		{ 264, 266, 13 },
		{ 264, 267, 13 },
		{ 264, 268, 13 },
		{ 264, 269, 13 },
		{ 264, 270, 13 },
		{ 264, 271, 13 },
		{ 264, 272, 13 },
		{ 264, 273, 13 },
		{ 264, 274, 13 },
		{ 264, 275, 13 },
		{ 264, 276, 13 },
		{ 264, 277, 13 },
		{ 264, 278, 13 },
		{ 264, 279, 13 },
		{ 264, 280, 13 },
		{ 264, 281, 13 },
		{ 264, 282, 13 },
		{ 264, 283, 13 },
		{ 264, 284, 13 },
		{ 264, 285, 13 },
		{ 264, 286, 13 },
		{ 264, 287, 13 },
		{ 264, 288, 13 },
		{ 265, 266, 13 },
		{ 265, 267, 13 },
		{ 265, 268, 13 },
		{ 265, 269, 13 },
		{ 265, 270, 13 },
		{ 265, 271, 13 },
		{ 265, 272, 13 },
		{ 265, 273, 13 },
		{ 265, 274, 13 },
		{ 265, 275, 13 },
		{ 265, 276, 13 },
		{ 265, 277, 13 },
		{ 265, 278, 13 },
		{ 265, 279, 13 },
		{ 265, 280, 13 },
		{ 265, 281, 13 },
		{ 265, 282, 13 },
		{ 265, 283, 13 },
		{ 265, 284, 13 },
		{ 265, 285, 13 },
		{ 265, 286, 13 },
		{ 265, 287, 13 },
		{ 265, 288, 13 },
		{ 266, 267, 13 },
		{ 266, 268, 13 },
		{ 266, 269, 13 },
		{ 266, 270, 13 },
		{ 266, 271, 13 },
		{ 266, 272, 13 },
		{ 266, 273, 13 },
		{ 266, 274, 13 },
		{ 266, 275, 13 },
		{ 266, 276, 13 },
		{ 266, 277, 13 },
		{ 266, 278, 13 },
		{ 266, 279, 13 },
		{ 266, 280, 13 },
		{ 266, 281, 13 },
		{ 266, 282, 13 },
		{ 266, 283, 13 },
		{ 266, 284, 13 },
		{ 266, 285, 13 },
		{ 266, 286, 13 },
		{ 266, 287, 13 },
		{ 266, 288, 13 },
		{ 267, 268, 13 },
		{ 267, 269, 13 },
		{ 267, 270, 13 },
		{ 267, 271, 13 },
		{ 267, 272, 13 },
		{ 267, 273, 13 },
		{ 267, 274, 13 },
		{ 267, 275, 13 },
		{ 267, 276, 13 },
		{ 267, 277, 13 },
		{ 267, 278, 13 },
		{ 267, 279, 13 },
		{ 267, 280, 13 },
		{ 267, 281, 13 },
		{ 267, 282, 13 },
		{ 267, 283, 13 },
		{ 267, 284, 13 },
		{ 267, 285, 13 },
		{ 267, 286, 13 },
		{ 267, 287, 13 },
		{ 267, 288, 13 },
		{ 268, 269, 13 },
		{ 268, 270, 13 },
		{ 268, 271, 13 },
		{ 268, 272, 13 },
		{ 268, 273, 13 },
		{ 268, 274, 13 },
		{ 268, 275, 13 },
		{ 268, 276, 13 },
		{ 268, 277, 13 },
		{ 268, 278, 13 },
		{ 268, 279, 13 },
		{ 268, 280, 13 },
		{ 268, 281, 13 },
		{ 268, 282, 13 },
		{ 268, 283, 13 },
		{ 268, 284, 13 },
		{ 268, 285, 13 },
		{ 268, 286, 13 },
		{ 268, 287, 13 },
		{ 268, 288, 13 },
		{ 269, 270, 13 },
		{ 269, 271, 13 },
		{ 269, 272, 13 },
		{ 269, 273, 13 },
		{ 269, 274, 13 },
		{ 269, 275, 13 },
		{ 269, 276, 13 },
		{ 269, 277, 13 },
		{ 269, 278, 13 },
		{ 269, 279, 13 },
		{ 269, 280, 13 },
		{ 269, 281, 13 },
		{ 269, 282, 13 },
		{ 269, 283, 13 },
		{ 269, 284, 13 },
		{ 269, 285, 13 },
		{ 269, 286, 13 },
		{ 269, 287, 13 },
		{ 269, 288, 13 },
		{ 270, 271, 13 },
		{ 270, 272, 13 },
		{ 270, 273, 13 },
		{ 270, 274, 13 },
		{ 270, 275, 13 },
		{ 270, 276, 13 },
		{ 270, 277, 13 },
		{ 270, 278, 13 },
		{ 270, 279, 13 },
		{ 270, 280, 13 },
		{ 270, 281, 13 },
		{ 270, 282, 13 },
		{ 270, 283, 13 },
		{ 270, 284, 13 },
		{ 270, 285, 13 },
		{ 270, 286, 13 },
		{ 270, 287, 13 },
		{ 270, 288, 13 },
		{ 271, 272, 13 },
		{ 271, 273, 13 },
		{ 271, 274, 13 },
		{ 271, 275, 13 },
		{ 271, 276, 13 },
		{ 271, 277, 13 },
		{ 271, 278, 13 },
		{ 271, 279, 13 },
		{ 271, 280, 13 },
		{ 271, 281, 13 },
		{ 271, 282, 13 },
		{ 271, 283, 13 },
		{ 271, 284, 13 },
		{ 271, 285, 13 },
		{ 271, 286, 13 },
		{ 271, 287, 13 },
		{ 271, 288, 13 },
		{ 272, 273, 13 },
		{ 272, 274, 13 },
		{ 272, 275, 13 },
		{ 272, 276, 13 },
		{ 272, 277, 13 },
		{ 272, 278, 13 },
		{ 272, 279, 13 },
		{ 272, 280, 13 },
		{ 272, 281, 13 },
		{ 272, 282, 13 },
		{ 272, 283, 13 },
		{ 272, 284, 13 },
		{ 272, 285, 13 },
		{ 272, 286, 13 },
		{ 272, 287, 13 },
		{ 272, 288, 13 },
		{ 273, 274, 13 },
		{ 273, 275, 13 },
		{ 273, 276, 13 },
		{ 273, 277, 13 },
		{ 273, 278, 13 },
		{ 273, 279, 13 },
		{ 273, 280, 13 },
		{ 273, 281, 13 },
		{ 273, 282, 13 },
		{ 273, 283, 13 },
		{ 273, 284, 13 },
		{ 273, 285, 13 },
		{ 273, 286, 13 },
		{ 273, 287, 13 },
		{ 273, 288, 13 },
		{ 274, 275, 13 },
		{ 274, 276, 13 },
		{ 274, 277, 13 },
		{ 274, 278, 13 },
		{ 274, 279, 13 },
		{ 274, 280, 13 },
		{ 274, 281, 13 },
		{ 274, 282, 13 },
		{ 274, 283, 13 },
		{ 274, 284, 13 },
		{ 274, 285, 13 },
		{ 274, 286, 13 },
		{ 274, 287, 13 },
		{ 274, 288, 13 },
		{ 275, 276, 13 },
		{ 275, 277, 13 },
		{ 275, 278, 13 },
		{ 275, 279, 13 },
		{ 275, 280, 13 },
		{ 275, 281, 13 },
		{ 275, 282, 13 },
		{ 275, 283, 13 },
		{ 275, 284, 13 },
		{ 275, 285, 13 },
		{ 275, 286, 13 },
		{ 275, 287, 13 },
		{ 275, 288, 13 },
		{ 276, 277, 13 },
		{ 276, 278, 13 },
		{ 276, 279, 13 },
		{ 276, 280, 13 },
		{ 276, 281, 13 },
		{ 276, 282, 13 },
		{ 276, 283, 13 },
		{ 276, 284, 13 },
		{ 276, 285, 13 },
		{ 276, 286, 13 },
		{ 276, 287, 13 },
		{ 276, 288, 13 },
		{ 277, 278, 13 },
		{ 277, 279, 13 },
		{ 277, 280, 13 },
		{ 277, 281, 13 },
		{ 277, 282, 13 },
		{ 277, 283, 13 },
		{ 277, 284, 13 },
		{ 277, 285, 13 },
		{ 277, 286, 13 },
		{ 277, 287, 13 },
		{ 277, 288, 13 },
		{ 278, 279, 13 },
		{ 278, 280, 13 },
		{ 278, 281, 13 },
		{ 278, 282, 13 },
		{ 278, 283, 13 },
		{ 278, 284, 13 },
		{ 278, 285, 13 },
		{ 278, 286, 13 },
		{ 278, 287, 13 },
		{ 278, 288, 13 },
		{ 279, 280, 13 },
		{ 279, 281, 13 },
		{ 279, 282, 13 },
		{ 279, 283, 13 },
		{ 279, 284, 13 },
		{ 279, 285, 13 },
		{ 279, 286, 13 },
		{ 279, 287, 13 },
		{ 279, 288, 13 },
		{ 280, 281, 13 },
		{ 280, 282, 13 },
		{ 280, 283, 13 },
		{ 280, 284, 13 },
		{ 280, 285, 13 },
		{ 280, 286, 13 },
		{ 280, 287, 13 },
		{ 280, 288, 13 },
		{ 281, 282, 13 },
		{ 281, 283, 13 },
		{ 281, 284, 13 },
		{ 281, 285, 13 },
		{ 281, 286, 13 },
		{ 281, 287, 13 },
		{ 281, 288, 13 },
		{ 282, 283, 13 },
		{ 282, 284, 13 },
		{ 282, 285, 13 },
		{ 282, 286, 13 },
		{ 282, 287, 13 },
		{ 282, 288, 13 },
		{ 283, 284, 13 },
		{ 283, 285, 13 },
		{ 283, 286, 13 },
		{ 283, 287, 13 },
		{ 283, 288, 13 },
		{ 284, 285, 13 },
		{ 284, 286, 13 },
		{ 284, 287, 13 },
		{ 284, 288, 13 },
		{ 285, 286, 13 },
		{ 285, 287, 13 },
		{ 285, 288, 13 },
		{ 286, 287, 13 },
		{ 286, 288, 13 },
		{ 287, 288, 13 },
		{ 228, 133, 14 },
		{ 228, 134, 14 },
		{ 228, 135, 14 },
		{ 228, 136, 14 },
		{ 228, 137, 14 },
		{ 228, 138, 14 },
		{ 228, 139, 14 },
		{ 228, 140, 14 },
		{ 228, 374, 14 },
		{ 228, 375, 14 },
		{ 228, 376, 14 },
		{ 228, 377, 14 },
		{ 228, 378, 14 },
		{ 228, 379, 14 },
		{ 228, 380, 14 },
		{ 228, 381, 14 },
		{ 229, 132, 14 },
		{ 229, 133, 14 },
		{ 229, 134, 14 },
		{ 229, 135, 14 },
		{ 229, 136, 14 },
		{ 229, 137, 14 },
		{ 229, 138, 14 },
		{ 229, 139, 14 },
		{ 229, 140, 14 },
		{ 229, 141, 14 },
		{ 229, 142, 14 },
		{ 229, 143, 14 },
		{ 229, 144, 14 },
		{ 229, 370, 14 },
		{ 229, 371, 14 },
		{ 229, 372, 14 },
		{ 229, 373, 14 },
		{ 229, 374, 14 },
		{ 229, 375, 14 },
		{ 229, 376, 14 },
		{ 229, 377, 14 },
		{ 229, 378, 14 },
		{ 229, 379, 14 },
		{ 229, 380, 14 },
		{ 229, 381, 14 },
		{ 229, 382, 14 },
		{ 230, 132, 14 },
		{ 230, 133, 14 },
		{ 230, 134, 14 },
		{ 230, 135, 14 },
		{ 230, 136, 14 },
		{ 230, 137, 14 },
		{ 230, 138, 14 },
		{ 230, 139, 14 },
		{ 230, 140, 14 },
		{ 230, 141, 14 },
		{ 230, 142, 14 },
		{ 230, 143, 14 },
		{ 230, 144, 14 },
		{ 230, 145, 14 },
		{ 230, 146, 14 },
		{ 230, 147, 14 },
		{ 230, 148, 14 },
		{ 230, 366, 14 },
		{ 230, 367, 14 },
		{ 230, 368, 14 },
		{ 230, 369, 14 },
		{ 230, 370, 14 },
		{ 230, 371, 14 },
		{ 230, 372, 14 },
		{ 230, 373, 14 },
		{ 230, 374, 14 },
		{ 230, 375, 14 },
		{ 230, 376, 14 },
		{ 230, 377, 14 },
		{ 230, 378, 14 },
		{ 230, 379, 14 },
		{ 230, 380, 14 },
		{ 230, 381, 14 },
		{ 230, 382, 14 },
		{ 231, 132, 14 },
		{ 231, 133, 14 },
		{ 231, 134, 14 },
		{ 231, 135, 14 },
		{ 231, 136, 14 },
		{ 231, 137, 14 },
		{ 231, 138, 14 },
		{ 231, 139, 14 },
		{ 231, 140, 14 },
		{ 231, 141, 14 },
		{ 231, 142, 14 },
		{ 231, 143, 14 },
		{ 231, 144, 14 },
		{ 231, 145, 14 },
		{ 231, 146, 14 },
		{ 231, 147, 14 },
		{ 231, 148, 14 },
		{ 231, 149, 14 },
		{ 231, 150, 14 },
		{ 231, 151, 14 },
		{ 231, 152, 14 },
		{ 231, 362, 14 },
		{ 231, 363, 14 },
		{ 231, 364, 14 },
		{ 231, 365, 14 },
		{ 231, 366, 14 },
		{ 231, 367, 14 },
		{ 231, 368, 14 },
		{ 231, 369, 14 },
		{ 231, 370, 14 },
		{ 231, 371, 14 },
		{ 231, 372, 14 },
		{ 231, 373, 14 },
		{ 231, 374, 14 },
		{ 231, 375, 14 },
		{ 231, 376, 14 },
		{ 231, 377, 14 },
		{ 231, 378, 14 },
		{ 231, 379, 14 },
		{ 231, 380, 14 },
		{ 231, 381, 14 },
		{ 231, 382, 14 },
		{ 232, 132, 14 },
		{ 232, 133, 14 },
		{ 232, 134, 14 },
		{ 232, 135, 14 },
		{ 232, 136, 14 },
		{ 232, 137, 14 },
		{ 232, 138, 14 },
		{ 232, 139, 14 },
		{ 232, 140, 14 },
		{ 232, 141, 14 },
		{ 232, 142, 14 },
		{ 232, 143, 14 },
		{ 232, 144, 14 },
		{ 232, 145, 14 },
		{ 232, 146, 14 },
		{ 232, 147, 14 },
		{ 232, 148, 14 },
		{ 232, 149, 14 },
		{ 232, 150, 14 },
		{ 232, 151, 14 },
		{ 232, 152, 14 },
		{ 232, 153, 14 },
		{ 232, 154, 14 },
		{ 232, 155, 14 },
		{ 232, 156, 14 },
		{ 232, 358, 14 },
		{ 232, 359, 14 },
		{ 232, 360, 14 },
		{ 232, 361, 14 },
		{ 232, 362, 14 },
		{ 232, 363, 14 },
		{ 232, 364, 14 },
		{ 232, 365, 14 },
		{ 232, 366, 14 },
		{ 232, 367, 14 },
		{ 232, 368, 14 },
		{ 232, 369, 14 },
		{ 232, 370, 14 },
		{ 232, 371, 14 },
		{ 232, 372, 14 },
		{ 232, 373, 14 },
		{ 232, 374, 14 },
		{ 232, 375, 14 },
		{ 232, 376, 14 },
		{ 232, 377, 14 },
		{ 232, 378, 14 },
		{ 232, 379, 14 },
		{ 232, 380, 14 },
		{ 232, 381, 14 },
		{ 232, 382, 14 },
		{ 233, 131, 14 },
		{ 233, 132, 14 },
		{ 233, 133, 14 },
		{ 233, 134, 14 },
		{ 233, 135, 14 },
		{ 233, 136, 14 },
		{ 233, 137, 14 },
		{ 233, 138, 14 },
		{ 233, 139, 14 },
		{ 233, 140, 14 },
		{ 233, 141, 14 },
		{ 233, 142, 14 },
		{ 233, 143, 14 },
		{ 233, 144, 14 },
		{ 233, 145, 14 },
		{ 233, 146, 14 },
		{ 233, 147, 14 },
		{ 233, 148, 14 },
		{ 233, 149, 14 },
		{ 233, 150, 14 },
		{ 233, 151, 14 },
		{ 233, 152, 14 },
		{ 233, 153, 14 },
		{ 233, 154, 14 },
		{ 233, 155, 14 },
		{ 233, 156, 14 },
		{ 233, 157, 14 },
		{ 233, 158, 14 },
		{ 233, 159, 14 },
		{ 233, 160, 14 },
		{ 233, 354, 14 },
		{ 233, 355, 14 },
		{ 233, 356, 14 },
		{ 233, 357, 14 },
		{ 233, 358, 14 },
		{ 233, 359, 14 },
		{ 233, 360, 14 },
		{ 233, 361, 14 },
		{ 233, 362, 14 },
		{ 233, 363, 14 },
		{ 233, 364, 14 },
		{ 233, 365, 14 },
		{ 233, 366, 14 },
		{ 233, 367, 14 },
		{ 233, 368, 14 },
		{ 233, 369, 14 },
		{ 233, 370, 14 },
		{ 233, 371, 14 },
		{ 233, 372, 14 },
		{ 233, 373, 14 },
		{ 233, 374, 14 },
		{ 233, 375, 14 },
		{ 233, 376, 14 },
		{ 233, 377, 14 },
		{ 233, 378, 14 },
		{ 233, 379, 14 },
		{ 233, 380, 14 },
		{ 233, 381, 14 },
		{ 233, 382, 14 },
		{ 233, 383, 14 },
		{ 234, 131, 14 },
		{ 234, 132, 14 },
		{ 234, 133, 14 },
		{ 234, 134, 14 },
		{ 234, 135, 14 },
		{ 234, 136, 14 },
		{ 234, 137, 14 },
		{ 234, 138, 14 },
		{ 234, 139, 14 },
		{ 234, 140, 14 },
		{ 234, 141, 14 },
		{ 234, 142, 14 },
		{ 234, 143, 14 },
		{ 234, 144, 14 },
		{ 234, 145, 14 },
		{ 234, 146, 14 },
		{ 234, 147, 14 },
		{ 234, 148, 14 },
		{ 234, 149, 14 },
		{ 234, 150, 14 },
		{ 234, 151, 14 },
		{ 234, 152, 14 },
		{ 234, 153, 14 },
		{ 234, 154, 14 },
		{ 234, 155, 14 },
		{ 234, 156, 14 },
		{ 234, 157, 14 },
		{ 234, 158, 14 },
		{ 234, 159, 14 },
		{ 234, 160, 14 },
		{ 234, 161, 14 },
		{ 234, 162, 14 },
		{ 234, 163, 14 },
		{ 234, 164, 14 },
		{ 234, 350, 14 },
		{ 234, 351, 14 },
		{ 234, 352, 14 },
		{ 234, 353, 14 },
		{ 234, 354, 14 },
		{ 234, 355, 14 },
		{ 234, 356, 14 },
		{ 234, 357, 14 },
		{ 234, 358, 14 },
		{ 234, 359, 14 },
		{ 234, 360, 14 },
		{ 234, 361, 14 },
		{ 234, 362, 14 },
		{ 234, 363, 14 },
		{ 234, 364, 14 },
		{ 234, 365, 14 },
		{ 234, 366, 14 },
		{ 234, 367, 14 },
		{ 234, 368, 14 },
		{ 234, 369, 14 },
		{ 234, 370, 14 },
		{ 234, 371, 14 },
		{ 234, 372, 14 },
		{ 234, 373, 14 },
		{ 234, 374, 14 },
		{ 234, 375, 14 },
		{ 234, 376, 14 },
		{ 234, 377, 14 },
		{ 234, 378, 14 },
		{ 234, 379, 14 },
		{ 234, 380, 14 },
		{ 234, 381, 14 },
		{ 234, 382, 14 },
		{ 234, 383, 14 },
		{ 235, 131, 14 },
		{ 235, 132, 14 },
		{ 235, 133, 14 },
		{ 235, 134, 14 },
		{ 235, 135, 14 },
		{ 235, 136, 14 },
		{ 235, 137, 14 },
		{ 235, 138, 14 },
		{ 235, 139, 14 },
		{ 235, 140, 14 },
		{ 235, 141, 14 },
		{ 235, 142, 14 },
		{ 235, 143, 14 },
		{ 235, 144, 14 },
		{ 235, 145, 14 },
		{ 235, 146, 14 },
		{ 235, 147, 14 },
		{ 235, 148, 14 },
		{ 235, 149, 14 },
		{ 235, 150, 14 },
		{ 235, 151, 14 },
		{ 235, 152, 14 },
		{ 235, 153, 14 },
		{ 235, 154, 14 },
		{ 235, 155, 14 },
		{ 235, 156, 14 },
		{ 235, 157, 14 },
		{ 235, 158, 14 },
		{ 235, 159, 14 },
		{ 235, 160, 14 },
		{ 235, 161, 14 },
		{ 235, 162, 14 },
		{ 235, 163, 14 },
		{ 235, 164, 14 },
		{ 235, 165, 14 },
		{ 235, 166, 14 },
		{ 235, 167, 14 },
		{ 235, 168, 14 },
		{ 235, 346, 14 },
		{ 235, 347, 14 },
		{ 235, 348, 14 },
		{ 235, 349, 14 },
		{ 235, 350, 14 },
		{ 235, 351, 14 },
		{ 235, 352, 14 },
		{ 235, 353, 14 },
		{ 235, 354, 14 },
		{ 235, 355, 14 },
		{ 235, 356, 14 },
		{ 235, 357, 14 },
		{ 235, 358, 14 },
		{ 235, 359, 14 },
		{ 235, 360, 14 },
		{ 235, 361, 14 },
		{ 235, 362, 14 },
		{ 235, 363, 14 },
		{ 235, 364, 14 },
		{ 235, 365, 14 },
		{ 235, 366, 14 },
		{ 235, 367, 14 },
		{ 235, 368, 14 },
		{ 235, 369, 14 },
		{ 235, 370, 14 },
		{ 235, 371, 14 },
		{ 235, 372, 14 },
		{ 235, 373, 14 },
		{ 235, 374, 14 },
		{ 235, 375, 14 },
		{ 235, 376, 14 },
		{ 235, 377, 14 },
		{ 235, 378, 14 },
		{ 235, 379, 14 },
		{ 235, 380, 14 },
		{ 235, 381, 14 },
		{ 235, 382, 14 },
		{ 235, 383, 14 },
		{ 236, 131, 14 },
		{ 236, 132, 14 },
		{ 236, 133, 14 },
		{ 236, 134, 14 },
		{ 236, 135, 14 },
		{ 236, 136, 14 },
		{ 236, 137, 14 },
		{ 236, 138, 14 },
		{ 236, 139, 14 },
		{ 236, 140, 14 },
		{ 236, 141, 14 },
		{ 236, 142, 14 },
		{ 236, 143, 14 },
		{ 236, 144, 14 },
		{ 236, 145, 14 },
		{ 236, 146, 14 },
		{ 236, 147, 14 },
		{ 236, 148, 14 },
		{ 236, 149, 14 },
		{ 236, 150, 14 },
		{ 236, 151, 14 },
		{ 236, 152, 14 },
		{ 236, 153, 14 },
		{ 236, 154, 14 },
		{ 236, 155, 14 },
		{ 236, 156, 14 },
		{ 236, 157, 14 },
		{ 236, 158, 14 },
		{ 236, 159, 14 },
		{ 236, 160, 14 },
		{ 236, 161, 14 },
		{ 236, 162, 14 },
		{ 236, 163, 14 },
		{ 236, 164, 14 },
		{ 236, 165, 14 },
		{ 236, 166, 14 },
		{ 236, 167, 14 },
		{ 236, 168, 14 },
		{ 236, 169, 14 },
		{ 236, 170, 14 },
		{ 236, 171, 14 },
		{ 236, 172, 14 },
		{ 236, 342, 14 },
		{ 236, 343, 14 },
		{ 236, 344, 14 },
		{ 236, 345, 14 },
		{ 236, 346, 14 },
		{ 236, 347, 14 },
		{ 236, 348, 14 },
		{ 236, 349, 14 },
		{ 236, 350, 14 },
		{ 236, 351, 14 },
		{ 236, 352, 14 },
		{ 236, 353, 14 },
		{ 236, 354, 14 },
		{ 236, 355, 14 },
		{ 236, 356, 14 },
		{ 236, 357, 14 },
		{ 236, 358, 14 },
		{ 236, 359, 14 },
		{ 236, 360, 14 },
		{ 236, 361, 14 },
		{ 236, 362, 14 },
		{ 236, 363, 14 },
		{ 236, 364, 14 },
		{ 236, 365, 14 },
		{ 236, 366, 14 },
		{ 236, 367, 14 },
		{ 236, 368, 14 },
		{ 236, 369, 14 },
		{ 236, 370, 14 },
		{ 236, 371, 14 },
		{ 236, 372, 14 },
		{ 236, 373, 14 },
		{ 236, 374, 14 },
		{ 236, 375, 14 },
		{ 236, 376, 14 },
		{ 236, 377, 14 },
		{ 236, 378, 14 },
		{ 236, 379, 14 },
		{ 236, 380, 14 },
		{ 236, 381, 14 },
		{ 236, 382, 14 },
		{ 236, 383, 14 },
		{ 237, 131, 14 },
		{ 237, 132, 14 },
		{ 237, 133, 14 },
		{ 237, 134, 14 },
		{ 237, 135, 14 },
		{ 237, 136, 14 },
		{ 237, 137, 14 },
		{ 237, 138, 14 },
		{ 237, 139, 14 },
		{ 237, 140, 14 },
		{ 237, 141, 14 },
		{ 237, 142, 14 },
		{ 237, 143, 14 },
		{ 237, 144, 14 },
		{ 237, 145, 14 },
		{ 237, 146, 14 },
		{ 237, 147, 14 },
		{ 237, 148, 14 },
		{ 237, 149, 14 },
		{ 237, 150, 14 },
		{ 237, 151, 14 },
		{ 237, 152, 14 },
		{ 237, 153, 14 },
		{ 237, 154, 14 },
		{ 237, 155, 14 },
		{ 237, 156, 14 },
		{ 237, 157, 14 },
		{ 237, 158, 14 },
		{ 237, 159, 14 },
		{ 237, 160, 14 },
		{ 237, 161, 14 },
		{ 237, 162, 14 },
		{ 237, 163, 14 },
		{ 237, 164, 14 },
		{ 237, 165, 14 },
		{ 237, 166, 14 },
		{ 237, 167, 14 },
		{ 237, 168, 14 },
		{ 237, 169, 14 },
		{ 237, 170, 14 },
		{ 237, 171, 14 },
		{ 237, 172, 14 },
		{ 237, 173, 14 },
		{ 237, 174, 14 },
		{ 237, 175, 14 },
		{ 237, 176, 14 },
		{ 237, 338, 14 },
		{ 237, 339, 14 },
		{ 237, 340, 14 },
		{ 237, 341, 14 },
		{ 237, 342, 14 },
		{ 237, 343, 14 },
		{ 237, 344, 14 },
		{ 237, 345, 14 },
		{ 237, 346, 14 },
		{ 237, 347, 14 },
		{ 237, 348, 14 },
		{ 237, 349, 14 },
		{ 237, 350, 14 },
		{ 237, 351, 14 },
		{ 237, 352, 14 },
		{ 237, 353, 14 },
		{ 237, 354, 14 },
		{ 237, 355, 14 },
		{ 237, 356, 14 },
		{ 237, 357, 14 },
		{ 237, 358, 14 },
		{ 237, 359, 14 },
		{ 237, 360, 14 },
		{ 237, 361, 14 },
		{ 237, 362, 14 },
		{ 237, 363, 14 },
		{ 237, 364, 14 },
		{ 237, 365, 14 },
		{ 237, 366, 14 },
		{ 237, 367, 14 },
		{ 237, 368, 14 },
		{ 237, 369, 14 },
		{ 237, 370, 14 },
		{ 237, 371, 14 },
		{ 237, 372, 14 },
		{ 237, 373, 14 },
		{ 237, 374, 14 },
		{ 237, 375, 14 },
		{ 237, 376, 14 },
		{ 237, 377, 14 },
		{ 237, 378, 14 },
		{ 237, 379, 14 },
		{ 237, 380, 14 },
		{ 237, 381, 14 },
		{ 237, 382, 14 },
		{ 237, 383, 14 },
		{ 238, 131, 14 },
		{ 238, 132, 14 },
		{ 238, 133, 14 },
		{ 238, 134, 14 },
		{ 238, 135, 14 },
		{ 238, 136, 14 },
		{ 238, 137, 14 },
		{ 238, 138, 14 },
		{ 238, 139, 14 },
		{ 238, 140, 14 },
		{ 238, 141, 14 },
		{ 238, 142, 14 },
		{ 238, 143, 14 },
		{ 238, 144, 14 },
		{ 238, 145, 14 },
		{ 238, 146, 14 },
		{ 238, 147, 14 },
		{ 238, 148, 14 },
		{ 238, 149, 14 },
		{ 238, 150, 14 },
		{ 238, 151, 14 },
		{ 238, 152, 14 },
		{ 238, 153, 14 },
		{ 238, 154, 14 },
		{ 238, 155, 14 },
		{ 238, 156, 14 },
		{ 238, 157, 14 },
		{ 238, 158, 14 },
		{ 238, 159, 14 },
		{ 238, 160, 14 },
		{ 238, 161, 14 },
		{ 238, 162, 14 },
		{ 238, 163, 14 },
		{ 238, 164, 14 },
		{ 238, 165, 14 },
		{ 238, 166, 14 },
		{ 238, 167, 14 },
		{ 238, 168, 14 },
		{ 238, 169, 14 },
		{ 238, 170, 14 },
		{ 238, 171, 14 },
		{ 238, 172, 14 },
		{ 238, 173, 14 },
		{ 238, 174, 14 },
		{ 238, 175, 14 },
		{ 238, 176, 14 },
		{ 238, 177, 14 },
		{ 238, 178, 14 },
		{ 238, 179, 14 },
		{ 238, 180, 14 },
		{ 238, 334, 14 },
		{ 238, 335, 14 },
		{ 238, 336, 14 },
		{ 238, 337, 14 },
		{ 238, 338, 14 },
		{ 238, 339, 14 },
		{ 238, 340, 14 },
		{ 238, 341, 14 },
		{ 238, 342, 14 },
		{ 238, 343, 14 },
		{ 238, 344, 14 },
		{ 238, 345, 14 },
		{ 238, 346, 14 },
		{ 238, 347, 14 },
		{ 238, 348, 14 },
		{ 238, 349, 14 },
		{ 238, 350, 14 },
		{ 238, 351, 14 },
		{ 238, 352, 14 },
		{ 238, 353, 14 },
		{ 238, 354, 14 },
		{ 238, 355, 14 },
		{ 238, 356, 14 },
		{ 238, 357, 14 },
		{ 238, 358, 14 },
		{ 238, 359, 14 },
		{ 238, 360, 14 },
		{ 238, 361, 14 },
		{ 238, 362, 14 },
		{ 238, 363, 14 },
		{ 238, 364, 14 },
		{ 238, 365, 14 },
		{ 238, 366, 14 },
		{ 238, 367, 14 },
		{ 238, 368, 14 },
		{ 238, 369, 14 },
		{ 238, 370, 14 },
		{ 238, 371, 14 },
		{ 238, 372, 14 },
		{ 238, 373, 14 },
		{ 238, 374, 14 },
		{ 238, 375, 14 },
		{ 238, 376, 14 },
		{ 238, 377, 14 },
		{ 238, 378, 14 },
		{ 238, 379, 14 },
		{ 238, 380, 14 },
		{ 238, 381, 14 },
		{ 238, 382, 14 },
		{ 238, 383, 14 },
		{ 239, 131, 14 },
		{ 239, 132, 14 },
		{ 239, 133, 14 },
		{ 239, 134, 14 },
		{ 239, 135, 14 },
		{ 239, 136, 14 },
		{ 239, 137, 14 },
		{ 239, 138, 14 },
		{ 239, 139, 14 },
		{ 239, 140, 14 },
		{ 239, 141, 14 },
		{ 239, 142, 14 },
		{ 239, 143, 14 },
		{ 239, 144, 14 },
		{ 239, 145, 14 },
		{ 239, 146, 14 },
		{ 239, 147, 14 },
		{ 239, 148, 14 },
		{ 239, 149, 14 },
		{ 239, 150, 14 },
		{ 239, 151, 14 },
		{ 239, 152, 14 },
		{ 239, 153, 14 },
		{ 239, 154, 14 },
		{ 239, 155, 14 },
		{ 239, 156, 14 },
		{ 239, 157, 14 },
		{ 239, 158, 14 },
		{ 239, 159, 14 },
		{ 239, 160, 14 },
		{ 239, 161, 14 },
		{ 239, 162, 14 },
		{ 239, 163, 14 },
		{ 239, 164, 14 },
		{ 239, 165, 14 },
		{ 239, 166, 14 },
		{ 239, 167, 14 },
		{ 239, 168, 14 },
		{ 239, 169, 14 },
		{ 239, 170, 14 },
		{ 239, 171, 14 },
		{ 239, 172, 14 },
		{ 239, 173, 14 },
		{ 239, 174, 14 },
		{ 239, 175, 14 },
		{ 239, 176, 14 },
		{ 239, 177, 14 },
		{ 239, 178, 14 },
		{ 239, 179, 14 },
		{ 239, 180, 14 },
		{ 239, 181, 14 },
		{ 239, 182, 14 },
		{ 239, 183, 14 },
		{ 239, 184, 14 },
		{ 239, 330, 14 },
		{ 239, 331, 14 },
		{ 239, 332, 14 },
		{ 239, 333, 14 },
		{ 239, 334, 14 },
		{ 239, 335, 14 },
		{ 239, 336, 14 },
		{ 239, 337, 14 },
		{ 239, 338, 14 },
		{ 239, 339, 14 },
		{ 239, 340, 14 },
		{ 239, 341, 14 },
		{ 239, 342, 14 },
		{ 239, 343, 14 },
		{ 239, 344, 14 },
		{ 239, 345, 14 },
		{ 239, 346, 14 },
		{ 239, 347, 14 },
		{ 239, 348, 14 },
		{ 239, 349, 14 },
		{ 239, 350, 14 },
		{ 239, 351, 14 },
		{ 239, 352, 14 },
		{ 239, 353, 14 },
		{ 239, 354, 14 },
		{ 239, 355, 14 },
		{ 239, 356, 14 },
		{ 239, 357, 14 },
		{ 239, 358, 14 },
		{ 239, 359, 14 },
		{ 239, 360, 14 },
		{ 239, 361, 14 },
		{ 239, 362, 14 },
		{ 239, 363, 14 },
		{ 239, 364, 14 },
		{ 239, 365, 14 },
		{ 239, 366, 14 },
		{ 239, 367, 14 },
		{ 239, 368, 14 },
		{ 239, 369, 14 },
		{ 239, 370, 14 },
		{ 239, 371, 14 },
		{ 239, 372, 14 },
		{ 239, 373, 14 },
		{ 239, 374, 14 },
		{ 239, 375, 14 },
		{ 239, 376, 14 },
		{ 239, 377, 14 },
		{ 239, 378, 14 },
		{ 239, 379, 14 },
		{ 239, 380, 14 },
		{ 239, 381, 14 },
		{ 239, 382, 14 },
		{ 239, 383, 14 },
		{ 240, 131, 14 },
		{ 240, 132, 14 },
		{ 240, 133, 14 },
		{ 240, 134, 14 },
		{ 240, 135, 14 },
		{ 240, 136, 14 },
		{ 240, 137, 14 },
		{ 240, 138, 14 },
		{ 240, 139, 14 },
		{ 240, 140, 14 },
		{ 240, 141, 14 },
		{ 240, 142, 14 },
		{ 240, 143, 14 },
		{ 240, 144, 14 },
		{ 240, 145, 14 },
		{ 240, 146, 14 },
		{ 240, 147, 14 },
		{ 240, 148, 14 },
		{ 240, 149, 14 },
		{ 240, 150, 14 },
		{ 240, 151, 14 },
		{ 240, 152, 14 },
		{ 240, 153, 14 },
		{ 240, 154, 14 },
		{ 240, 155, 14 },
		{ 240, 156, 14 },
		{ 240, 157, 14 },
		{ 240, 158, 14 },
		{ 240, 159, 14 },
		{ 240, 160, 14 },
		{ 240, 161, 14 },
		{ 240, 162, 14 },
		{ 240, 163, 14 },
		{ 240, 164, 14 },
		{ 240, 165, 14 },
		{ 240, 166, 14 },
		{ 240, 167, 14 },
		{ 240, 168, 14 },
		{ 240, 169, 14 },
		{ 240, 170, 14 },
		{ 240, 171, 14 },
		{ 240, 172, 14 },
		{ 240, 173, 14 },
		{ 240, 174, 14 },
		{ 240, 175, 14 },
		{ 240, 176, 14 },
		{ 240, 177, 14 },
		{ 240, 178, 14 },
		{ 240, 179, 14 },
		{ 240, 180, 14 },
		{ 240, 181, 14 },
		{ 240, 182, 14 },
		{ 240, 183, 14 },
		{ 240, 184, 14 },
		{ 240, 185, 14 },
		{ 240, 186, 14 },
		{ 240, 187, 14 },
		{ 240, 188, 14 },
		{ 240, 326, 14 },
		{ 240, 327, 14 },
		{ 240, 328, 14 },
		{ 240, 329, 14 },
		{ 240, 330, 14 },
		{ 240, 331, 14 },
		{ 240, 332, 14 },
		{ 240, 333, 14 },
		{ 240, 334, 14 },
		{ 240, 335, 14 },
		{ 240, 336, 14 },
		{ 240, 337, 14 },
		{ 240, 338, 14 },
		{ 240, 339, 14 },
		{ 240, 340, 14 },
		{ 240, 341, 14 },
		{ 240, 342, 14 },
		{ 240, 343, 14 },
		{ 240, 344, 14 },
		{ 240, 345, 14 },
		{ 240, 346, 14 },
		{ 240, 347, 14 },
		{ 240, 348, 14 },
		{ 240, 349, 14 },
		{ 240, 350, 14 },
		{ 240, 351, 14 },
		{ 240, 352, 14 },
		{ 240, 353, 14 },
		{ 240, 354, 14 },
		{ 240, 355, 14 },
		{ 240, 356, 14 },
		{ 240, 357, 14 },
		{ 240, 358, 14 },
		{ 240, 359, 14 },
		{ 240, 360, 14 },
		{ 240, 361, 14 },
		{ 240, 362, 14 },
		{ 240, 363, 14 },
		{ 240, 364, 14 },
		{ 240, 365, 14 },
		{ 240, 366, 14 },
		{ 240, 367, 14 },
		{ 240, 368, 14 },
		{ 240, 369, 14 },
		{ 240, 370, 14 },
		{ 240, 371, 14 },
		{ 240, 372, 14 },
		{ 240, 373, 14 },
		{ 240, 374, 14 },
		{ 240, 375, 14 },
		{ 240, 376, 14 },
		{ 240, 377, 14 },
		{ 240, 378, 14 },
		{ 240, 379, 14 },
		{ 240, 380, 14 },
		{ 240, 381, 14 },
		{ 240, 382, 14 },
		{ 240, 383, 14 },
		{ 241, 131, 14 },
		{ 241, 132, 14 },
		{ 241, 133, 14 },
		{ 241, 134, 14 },
		{ 241, 135, 14 },
		{ 241, 136, 14 },
		{ 241, 137, 14 },
		{ 241, 138, 14 },
		{ 241, 139, 14 },
		{ 241, 140, 14 },
		{ 241, 141, 14 },
		{ 241, 142, 14 },
		{ 241, 143, 14 },
		{ 241, 144, 14 },
		{ 241, 145, 14 },
		{ 241, 146, 14 },
		{ 241, 147, 14 },
		{ 241, 148, 14 },
		{ 241, 149, 14 },
		{ 241, 150, 14 },
		{ 241, 151, 14 },
		{ 241, 152, 14 },
		{ 241, 153, 14 },
		{ 241, 154, 14 },
		{ 241, 155, 14 },
		{ 241, 156, 14 },
		{ 241, 157, 14 },
		{ 241, 158, 14 },
		{ 241, 159, 14 },
		{ 241, 160, 14 },
		{ 241, 161, 14 },
		{ 241, 162, 14 },
		{ 241, 163, 14 },
		{ 241, 164, 14 },
		{ 241, 165, 14 },
		{ 241, 166, 14 },
		{ 241, 167, 14 },
		{ 241, 168, 14 },
		{ 241, 169, 14 },
		{ 241, 170, 14 },
		{ 241, 171, 14 },
		{ 241, 172, 14 },
		{ 241, 173, 14 },
		{ 241, 174, 14 },
		{ 241, 175, 14 },
		{ 241, 176, 14 },
		{ 241, 177, 14 },
		{ 241, 178, 14 },
		{ 241, 179, 14 },
		{ 241, 180, 14 },
		{ 241, 181, 14 },
		{ 241, 182, 14 },
		{ 241, 183, 14 },
		{ 241, 184, 14 },
		{ 241, 185, 14 },
		{ 241, 186, 14 },
		{ 241, 187, 14 },
		{ 241, 188, 14 },
		{ 241, 189, 14 },
		{ 241, 190, 14 },
		{ 241, 191, 14 },
		{ 241, 192, 14 },
		{ 241, 322, 14 },
		{ 241, 323, 14 },
		{ 241, 324, 14 },
		{ 241, 325, 14 },
		{ 241, 326, 14 },
		{ 241, 327, 14 },
		{ 241, 328, 14 },
		{ 241, 329, 14 },
		{ 241, 330, 14 },
		{ 241, 331, 14 },
		{ 241, 332, 14 },
		{ 241, 333, 14 },
		{ 241, 334, 14 },
		{ 241, 335, 14 },
		{ 241, 336, 14 },
		{ 241, 337, 14 },
		{ 241, 338, 14 },
		{ 241, 339, 14 },
		{ 241, 340, 14 },
		{ 241, 341, 14 },
		{ 241, 342, 14 },
		{ 241, 343, 14 },
		{ 241, 344, 14 },
		{ 241, 345, 14 },
		{ 241, 346, 14 },
		{ 241, 347, 14 },
		{ 241, 348, 14 },
		{ 241, 349, 14 },
		{ 241, 350, 14 },
		{ 241, 351, 14 },
		{ 241, 352, 14 },
		{ 241, 353, 14 },
		{ 241, 354, 14 },
		{ 241, 355, 14 },
		{ 241, 356, 14 },
		{ 241, 357, 14 },
		{ 241, 358, 14 },
		{ 241, 359, 14 },
		{ 241, 360, 14 },
		{ 241, 361, 14 },
		{ 241, 362, 14 },
		{ 241, 363, 14 },
		{ 241, 364, 14 },
		{ 241, 365, 14 },
		{ 241, 366, 14 },
		{ 241, 367, 14 },
		{ 241, 368, 14 },
		{ 241, 369, 14 },
		{ 241, 370, 14 },
		{ 241, 371, 14 },
		{ 241, 372, 14 },
		{ 241, 373, 14 },
		{ 241, 374, 14 },
		{ 241, 375, 14 },
		{ 241, 376, 14 },
		{ 241, 377, 14 },
		{ 241, 378, 14 },
		{ 241, 379, 14 },
		{ 241, 380, 14 },
		{ 241, 381, 14 },
		{ 241, 382, 14 },
		{ 241, 383, 14 },
		{ 242, 131, 14 },
		{ 242, 132, 14 },
		{ 242, 133, 14 },
		{ 242, 134, 14 },
		{ 242, 135, 14 },
		{ 242, 136, 14 },
		{ 242, 137, 14 },
		{ 242, 138, 14 },
		{ 242, 139, 14 },
		{ 242, 140, 14 },
		{ 242, 141, 14 },
		{ 242, 142, 14 },
		{ 242, 143, 14 },
		{ 242, 144, 14 },
		{ 242, 145, 14 },
		{ 242, 146, 14 },
		{ 242, 147, 14 },
		{ 242, 148, 14 },
		{ 242, 149, 14 },
		{ 242, 150, 14 },
		{ 242, 151, 14 },
		{ 242, 152, 14 },
		{ 242, 153, 14 },
		{ 242, 154, 14 },
		{ 242, 155, 14 },
		{ 242, 156, 14 },
		{ 242, 157, 14 },
		{ 242, 158, 14 },
		{ 242, 159, 14 },
		{ 242, 160, 14 },
		{ 242, 161, 14 },
		{ 242, 162, 14 },
		{ 242, 163, 14 },
		{ 242, 164, 14 },
		{ 242, 165, 14 },
		{ 242, 166, 14 },
		{ 242, 167, 14 },
		{ 242, 168, 14 },
		{ 242, 169, 14 },
		{ 242, 170, 14 },
		{ 242, 171, 14 },
		{ 242, 172, 14 },
		{ 242, 173, 14 },
		{ 242, 174, 14 },
		{ 242, 175, 14 },
		{ 242, 176, 14 },
		{ 242, 177, 14 },
		{ 242, 178, 14 },
		{ 242, 179, 14 },
		{ 242, 180, 14 },
		{ 242, 181, 14 },
		{ 242, 182, 14 },
		{ 242, 183, 14 },
		{ 242, 184, 14 },
		{ 242, 185, 14 },
		{ 242, 186, 14 },
		{ 242, 187, 14 },
		{ 242, 188, 14 },
		{ 242, 189, 14 },
		{ 242, 190, 14 },
		{ 242, 191, 14 },
		{ 242, 192, 14 },
		{ 242, 193, 14 },
		{ 242, 194, 14 },
		{ 242, 195, 14 },
		{ 242, 196, 14 },
		{ 242, 318, 14 },
		{ 242, 319, 14 },
		{ 242, 320, 14 },
		{ 242, 321, 14 },
		{ 242, 322, 14 },
		{ 242, 323, 14 },
		{ 242, 324, 14 },
		{ 242, 325, 14 },
		{ 242, 326, 14 },
		{ 242, 327, 14 },
		{ 242, 328, 14 },
		{ 242, 329, 14 },
		{ 242, 330, 14 },
		{ 242, 331, 14 },
		{ 242, 332, 14 },
		{ 242, 333, 14 },
		{ 242, 334, 14 },
		{ 242, 335, 14 },
		{ 242, 336, 14 },
		{ 242, 337, 14 },
		{ 242, 338, 14 },
		{ 242, 339, 14 },
		{ 242, 340, 14 },
		{ 242, 341, 14 },
		{ 242, 342, 14 },
		{ 242, 343, 14 },
		{ 242, 344, 14 },
		{ 242, 345, 14 },
		{ 242, 346, 14 },
		{ 242, 347, 14 },
		{ 242, 348, 14 },
		{ 242, 349, 14 },
		{ 242, 350, 14 },
		{ 242, 351, 14 },
		{ 242, 352, 14 },
		{ 242, 353, 14 },
		{ 242, 354, 14 },
		{ 242, 355, 14 },
		{ 242, 356, 14 },
		{ 242, 357, 14 },
		{ 242, 358, 14 },
		{ 242, 359, 14 },
		{ 242, 360, 14 },
		{ 242, 361, 14 },
		{ 242, 362, 14 },
		{ 242, 363, 14 },
		{ 242, 364, 14 },
		{ 242, 365, 14 },
		{ 242, 366, 14 },
		{ 242, 367, 14 },
		{ 242, 368, 14 },
		{ 242, 369, 14 },
		{ 242, 370, 14 },
		{ 242, 371, 14 },
		{ 242, 372, 14 },
		{ 242, 373, 14 },
		{ 242, 374, 14 },
		{ 242, 375, 14 },
		{ 242, 376, 14 },
		{ 242, 377, 14 },
		{ 242, 378, 14 },
		{ 242, 379, 14 },
		{ 242, 380, 14 },
		{ 242, 381, 14 },
		{ 242, 382, 14 },
		{ 242, 383, 14 },
		{ 243, 131, 14 },
		{ 243, 132, 14 },
		{ 243, 133, 14 },
		{ 243, 134, 14 },
		{ 243, 135, 14 },
		{ 243, 136, 14 },
		{ 243, 137, 14 },
		{ 243, 138, 14 },
		{ 243, 139, 14 },
		{ 243, 140, 14 },
		{ 243, 141, 14 },
		{ 243, 142, 14 },
		{ 243, 143, 14 },
		{ 243, 144, 14 },
		{ 243, 145, 14 },
		{ 243, 146, 14 },
		{ 243, 147, 14 },
		{ 243, 148, 14 },
		{ 243, 149, 14 },
		{ 243, 150, 14 },
		{ 243, 151, 14 },
		{ 243, 152, 14 },
		{ 243, 153, 14 },
		{ 243, 154, 14 },
		{ 243, 155, 14 },
		{ 243, 156, 14 },
		{ 243, 157, 14 },
		{ 243, 158, 14 },
		{ 243, 159, 14 },
		{ 243, 160, 14 },
		{ 243, 161, 14 },
		{ 243, 162, 14 },
		{ 243, 163, 14 },
		{ 243, 164, 14 },
		{ 243, 165, 14 },
		{ 243, 166, 14 },
		{ 243, 167, 14 },
		{ 243, 168, 14 },
		{ 243, 169, 14 },
		{ 243, 170, 14 },
		{ 243, 171, 14 },
		{ 243, 172, 14 },
		{ 243, 173, 14 },
		{ 243, 174, 14 },
		{ 243, 175, 14 },
		{ 243, 176, 14 },
		{ 243, 177, 14 },
		{ 243, 178, 14 },
		{ 243, 179, 14 },
		{ 243, 180, 14 },
		{ 243, 181, 14 },
		{ 243, 182, 14 },
		{ 243, 183, 14 },
		{ 243, 184, 14 },
		{ 243, 185, 14 },
		{ 243, 186, 14 },
		{ 243, 187, 14 },
		{ 243, 188, 14 },
		{ 243, 189, 14 },
		{ 243, 190, 14 },
		{ 243, 191, 14 },
		{ 243, 192, 14 },
		{ 243, 193, 14 },
		{ 243, 194, 14 },
		{ 243, 195, 14 },
		{ 243, 196, 14 },
		{ 243, 197, 14 },
		{ 243, 198, 14 },
		{ 243, 199, 14 },
		{ 243, 200, 14 },
		{ 243, 314, 14 },
		{ 243, 315, 14 },
		{ 243, 316, 14 },
		{ 243, 317, 14 },
		{ 243, 318, 14 },
		{ 243, 319, 14 },
		{ 243, 320, 14 },
		{ 243, 321, 14 },
		{ 243, 322, 14 },
		{ 243, 323, 14 },
		{ 243, 324, 14 },
		{ 243, 325, 14 },
		{ 243, 326, 14 },
		{ 243, 327, 14 },
		{ 243, 328, 14 },
		{ 243, 329, 14 },
		{ 243, 330, 14 },
		{ 243, 331, 14 },
		{ 243, 332, 14 },
		{ 243, 333, 14 },
		{ 243, 334, 14 },
		{ 243, 335, 14 },
		{ 243, 336, 14 },
		{ 243, 337, 14 },
		{ 243, 338, 14 },
		{ 243, 339, 14 },
		{ 243, 340, 14 },
		{ 243, 341, 14 },
		{ 243, 342, 14 },
		{ 243, 343, 14 },
		{ 243, 344, 14 },
		{ 243, 345, 14 },
		{ 243, 346, 14 },
		{ 243, 347, 14 },
		{ 243, 348, 14 },
		{ 243, 349, 14 },
		{ 243, 350, 14 },
		{ 243, 351, 14 },
		{ 243, 352, 14 },
		{ 243, 353, 14 },
		{ 243, 354, 14 },
		{ 243, 355, 14 },
		{ 243, 356, 14 },
		{ 243, 357, 14 },
		{ 243, 358, 14 },
		{ 243, 359, 14 },
		{ 243, 360, 14 },
		{ 243, 361, 14 },
		{ 243, 362, 14 },
		{ 243, 363, 14 },
		{ 243, 364, 14 },
		{ 243, 365, 14 },
		{ 243, 366, 14 },
		{ 243, 367, 14 },
		{ 243, 368, 14 },
		{ 243, 369, 14 },
		{ 243, 370, 14 },
		{ 243, 371, 14 },
		{ 243, 372, 14 },
		{ 243, 373, 14 },
		{ 243, 374, 14 },
		{ 243, 375, 14 },
		{ 243, 376, 14 },
		{ 243, 377, 14 },
		{ 243, 378, 14 },
		{ 243, 379, 14 },
		{ 243, 380, 14 },
		{ 243, 381, 14 },
		{ 243, 382, 14 },
		{ 243, 383, 14 },
		{ 244, 131, 14 },
		{ 244, 132, 14 },
		{ 244, 133, 14 },
		{ 244, 134, 14 },
		{ 244, 135, 14 },
		{ 244, 136, 14 },
		{ 244, 137, 14 },
		{ 244, 138, 14 },
		{ 244, 139, 14 },
		{ 244, 140, 14 },
		{ 244, 141, 14 },
		{ 244, 142, 14 },
		{ 244, 143, 14 },
		{ 244, 144, 14 },
		{ 244, 145, 14 },
		{ 244, 146, 14 },
		{ 244, 147, 14 },
		{ 244, 148, 14 },
		{ 244, 149, 14 },
		{ 244, 150, 14 },
		{ 244, 151, 14 },
		{ 244, 152, 14 },
		{ 244, 153, 14 },
		{ 244, 154, 14 },
		{ 244, 155, 14 },
		{ 244, 156, 14 },
		{ 244, 157, 14 },
		{ 244, 158, 14 },
		{ 244, 159, 14 },
		{ 244, 160, 14 },
		{ 244, 161, 14 },
		{ 244, 162, 14 },
		{ 244, 163, 14 },
		{ 244, 164, 14 },
		{ 244, 165, 14 },
		{ 244, 166, 14 },
		{ 244, 167, 14 },
		{ 244, 168, 14 },
		{ 244, 169, 14 },
		{ 244, 170, 14 },
		{ 244, 171, 14 },
		{ 244, 172, 14 },
		{ 244, 173, 14 },
		{ 244, 174, 14 },
		{ 244, 175, 14 },
		{ 244, 176, 14 },
		{ 244, 177, 14 },
		{ 244, 178, 14 },
		{ 244, 179, 14 },
		{ 244, 180, 14 },
		{ 244, 181, 14 },
		{ 244, 182, 14 },
		{ 244, 183, 14 },
		{ 244, 184, 14 },
		{ 244, 185, 14 },
		{ 244, 186, 14 },
		{ 244, 187, 14 },
		{ 244, 188, 14 },
		{ 244, 189, 14 },
		{ 244, 190, 14 },
		{ 244, 191, 14 },
		{ 244, 192, 14 },
		{ 244, 193, 14 },
		{ 244, 194, 14 },
		{ 244, 195, 14 },
		{ 244, 196, 14 },
		{ 244, 197, 14 },
		{ 244, 198, 14 },
		{ 244, 199, 14 },
		{ 244, 200, 14 },
		{ 244, 201, 14 },
		{ 244, 202, 14 },
		{ 244, 203, 14 },
		{ 244, 204, 14 },
		{ 244, 310, 14 },
		{ 244, 311, 14 },
		{ 244, 312, 14 },
		{ 244, 313, 14 },
		{ 244, 314, 14 },
		{ 244, 315, 14 },
		{ 244, 316, 14 },
		{ 244, 317, 14 },
		{ 244, 318, 14 },
		{ 244, 319, 14 },
		{ 244, 320, 14 },
		{ 244, 321, 14 },
		{ 244, 322, 14 },
		{ 244, 323, 14 },
		{ 244, 324, 14 },
		{ 244, 325, 14 },
		{ 244, 326, 14 },
		{ 244, 327, 14 },
		{ 244, 328, 14 },
		{ 244, 329, 14 },
		{ 244, 330, 14 },
		{ 244, 331, 14 },
		{ 244, 332, 14 },
		{ 244, 333, 14 },
		{ 244, 334, 14 },
		{ 244, 335, 14 },
		{ 244, 336, 14 },
		{ 244, 337, 14 },
		{ 244, 338, 14 },
		{ 244, 339, 14 },
		{ 244, 340, 14 },
		{ 244, 341, 14 },
		{ 244, 342, 14 },
		{ 244, 343, 14 },
		{ 244, 344, 14 },
		{ 244, 345, 14 },
		{ 244, 346, 14 },
		{ 244, 347, 14 },
		{ 244, 348, 14 },
		{ 244, 349, 14 },
		{ 244, 350, 14 },
		{ 244, 351, 14 },
		{ 244, 352, 14 },
		{ 244, 353, 14 },
		{ 244, 354, 14 },
		{ 244, 355, 14 },
		{ 244, 356, 14 },
		{ 244, 357, 14 },
		{ 244, 358, 14 },
		{ 244, 359, 14 },
		{ 244, 360, 14 },
		{ 244, 361, 14 },
		{ 244, 362, 14 },
		{ 244, 363, 14 },
		{ 244, 364, 14 },
		{ 244, 365, 14 },
		{ 244, 366, 14 },
		{ 244, 367, 14 },
		{ 244, 368, 14 },
		{ 244, 369, 14 },
		{ 244, 370, 14 },
		{ 244, 371, 14 },
		{ 244, 372, 14 },
		{ 244, 373, 14 },
		{ 244, 374, 14 },
		{ 244, 375, 14 },
		{ 244, 376, 14 },
		{ 244, 377, 14 },
		{ 244, 378, 14 },
		{ 244, 379, 14 },
		{ 244, 380, 14 },
		{ 244, 381, 14 },
		{ 244, 382, 14 },
		{ 244, 383, 14 },
		{ 245, 131, 14 },
		{ 245, 132, 14 },
		{ 245, 133, 14 },
		{ 245, 134, 14 },
		{ 245, 135, 14 },
		{ 245, 136, 14 },
		{ 245, 137, 14 },
		{ 245, 138, 14 },
		{ 245, 139, 14 },
		{ 245, 140, 14 },
		{ 245, 141, 14 },
		{ 245, 142, 14 },
		{ 245, 143, 14 },
		{ 245, 144, 14 },
		{ 245, 145, 14 },
		{ 245, 146, 14 },
		{ 245, 147, 14 },
		{ 245, 148, 14 },
		{ 245, 149, 14 },
		{ 245, 150, 14 },
		{ 245, 151, 14 },
		{ 245, 152, 14 },
		{ 245, 153, 14 },
		{ 245, 154, 14 },
		{ 245, 155, 14 },
		{ 245, 156, 14 },
		{ 245, 157, 14 },
		{ 245, 158, 14 },
		{ 245, 159, 14 },
		{ 245, 160, 14 },
		{ 245, 161, 14 },
		{ 245, 162, 14 },
		{ 245, 163, 14 },
		{ 245, 164, 14 },
		{ 245, 165, 14 },
		{ 245, 166, 14 },
		{ 245, 167, 14 },
		{ 245, 168, 14 },
		{ 245, 169, 14 },
		{ 245, 170, 14 },
		{ 245, 171, 14 },
		{ 245, 172, 14 },
		{ 245, 173, 14 },
		{ 245, 174, 14 },
		{ 245, 175, 14 },
		{ 245, 176, 14 },
		{ 245, 177, 14 },
		{ 245, 178, 14 },
		{ 245, 179, 14 },
		{ 245, 180, 14 },
		{ 245, 181, 14 },
		{ 245, 182, 14 },
		{ 245, 183, 14 },
		{ 245, 184, 14 },
		{ 245, 185, 14 },
		{ 245, 186, 14 },
		{ 245, 187, 14 },
		{ 245, 188, 14 },
		{ 245, 189, 14 },
		{ 245, 190, 14 },
		{ 245, 191, 14 },
		{ 245, 192, 14 },
		{ 245, 193, 14 },
		{ 245, 194, 14 },
		{ 245, 195, 14 },
		{ 245, 196, 14 },
		{ 245, 197, 14 },
		{ 245, 198, 14 },
		{ 245, 199, 14 },
		{ 245, 200, 14 },
		{ 245, 201, 14 },
		{ 245, 202, 14 },
		{ 245, 203, 14 },
		{ 245, 204, 14 },
		{ 245, 205, 14 },
		{ 245, 206, 14 },
		{ 245, 207, 14 },
		{ 245, 208, 14 },
		{ 245, 306, 14 },
		{ 245, 307, 14 },
		{ 245, 308, 14 },
		{ 245, 309, 14 },
		{ 245, 310, 14 },
		{ 245, 311, 14 },
		{ 245, 312, 14 },
		{ 245, 313, 14 },
		{ 245, 314, 14 },
		{ 245, 315, 14 },
		{ 245, 316, 14 },
		{ 245, 317, 14 },
		{ 245, 318, 14 },
		{ 245, 319, 14 },
		{ 245, 320, 14 },
		{ 245, 321, 14 },
		{ 245, 322, 14 },
		{ 245, 323, 14 },
		{ 245, 324, 14 },
		{ 245, 325, 14 },
		{ 245, 326, 14 },
		{ 245, 327, 14 },
		{ 245, 328, 14 },
		{ 245, 329, 14 },
		{ 245, 330, 14 },
		{ 245, 331, 14 },
		{ 245, 332, 14 },
		{ 245, 333, 14 },
		{ 245, 334, 14 },
		{ 245, 335, 14 },
		{ 245, 336, 14 },
		{ 245, 337, 14 },
		{ 245, 338, 14 },
		{ 245, 339, 14 },
		{ 245, 340, 14 },
		{ 245, 341, 14 },
		{ 245, 342, 14 },
		{ 245, 343, 14 },
		{ 245, 344, 14 },
		{ 245, 345, 14 },
		{ 245, 346, 14 },
		{ 245, 347, 14 },
		{ 245, 348, 14 },
		{ 245, 349, 14 },
		{ 245, 350, 14 },
		{ 245, 351, 14 },
		{ 245, 352, 14 },
		{ 245, 353, 14 },
		{ 245, 354, 14 },
		{ 245, 355, 14 },
		{ 245, 356, 14 },
		{ 245, 357, 14 },
		{ 245, 358, 14 },
		{ 245, 359, 14 },
		{ 245, 360, 14 },
		{ 245, 361, 14 },
		{ 245, 362, 14 },
		{ 245, 363, 14 },
		{ 245, 364, 14 },
		{ 245, 365, 14 },
		{ 245, 366, 14 },
		{ 245, 367, 14 },
		{ 245, 368, 14 },
		{ 245, 369, 14 },
		{ 245, 370, 14 },
		{ 245, 371, 14 },
		{ 245, 372, 14 },
		{ 245, 373, 14 },
		{ 245, 374, 14 },
		{ 245, 375, 14 },
		{ 245, 376, 14 },
		{ 245, 377, 14 },
		{ 245, 378, 14 },
		{ 245, 379, 14 },
		{ 245, 380, 14 },
		{ 245, 381, 14 },
		{ 245, 382, 14 },
		{ 245, 383, 14 },
		{ 246, 131, 14 },
		{ 246, 132, 14 },
		{ 246, 133, 14 },
		{ 246, 134, 14 },
		{ 246, 135, 14 },
		{ 246, 136, 14 },
		{ 246, 137, 14 },
		{ 246, 138, 14 },
		{ 246, 139, 14 },
		{ 246, 140, 14 },
		{ 246, 141, 14 },
		{ 246, 142, 14 },
		{ 246, 143, 14 },
		{ 246, 144, 14 },
		{ 246, 145, 14 },
		{ 246, 146, 14 },
		{ 246, 147, 14 },
		{ 246, 148, 14 },
		{ 246, 149, 14 },
		{ 246, 150, 14 },
		{ 246, 151, 14 },
		{ 246, 152, 14 },
		{ 246, 153, 14 },
		{ 246, 154, 14 },
		{ 246, 155, 14 },
		{ 246, 156, 14 },
		{ 246, 157, 14 },
		{ 246, 158, 14 },
		{ 246, 159, 14 },
		{ 246, 160, 14 },
		{ 246, 161, 14 },
		{ 246, 162, 14 },
		{ 246, 163, 14 },
		{ 246, 164, 14 },
		{ 246, 165, 14 },
		{ 246, 166, 14 },
		{ 246, 167, 14 },
		{ 246, 168, 14 },
		{ 246, 169, 14 },
		{ 246, 170, 14 },
		{ 246, 171, 14 },
		{ 246, 172, 14 },
		{ 246, 173, 14 },
		{ 246, 174, 14 },
		{ 246, 175, 14 },
		{ 246, 176, 14 },
		{ 246, 177, 14 },
		{ 246, 178, 14 },
		{ 246, 179, 14 },
		{ 246, 180, 14 },
		{ 246, 181, 14 },
		{ 246, 182, 14 },
		{ 246, 183, 14 },
		{ 246, 184, 14 },
		{ 246, 185, 14 },
		{ 246, 186, 14 },
		{ 246, 187, 14 },
		{ 246, 188, 14 },
		{ 246, 189, 14 },
		{ 246, 190, 14 },
		{ 246, 191, 14 },
		{ 246, 192, 14 },
		{ 246, 193, 14 },
		{ 246, 194, 14 },
		{ 246, 195, 14 },
		{ 246, 196, 14 },
		{ 246, 197, 14 },
		{ 246, 198, 14 },
		{ 246, 199, 14 },
		{ 246, 200, 14 },
		{ 246, 201, 14 },
		{ 246, 202, 14 },
		{ 246, 203, 14 },
		{ 246, 204, 14 },
		{ 246, 205, 14 },
		{ 246, 206, 14 },
		{ 246, 207, 14 },
		{ 246, 208, 14 },
		{ 246, 209, 14 },
		{ 246, 210, 14 },
		{ 246, 211, 14 },
		{ 246, 212, 14 },
		{ 246, 302, 14 },
		{ 246, 303, 14 },
		{ 246, 304, 14 },
		{ 246, 305, 14 },
		{ 246, 306, 14 },
		{ 246, 307, 14 },
		{ 246, 308, 14 },
		{ 246, 309, 14 },
		{ 246, 310, 14 },
		{ 246, 311, 14 },
		{ 246, 312, 14 },
		{ 246, 313, 14 },
		{ 246, 314, 14 },
		{ 246, 315, 14 },
		{ 246, 316, 14 },
		{ 246, 317, 14 },
		{ 246, 318, 14 },
		{ 246, 319, 14 },
		{ 246, 320, 14 },
		{ 246, 321, 14 },
		{ 246, 322, 14 },
		{ 246, 323, 14 },
		{ 246, 324, 14 },
		{ 246, 325, 14 },
		{ 246, 326, 14 },
		{ 246, 327, 14 },
		{ 246, 328, 14 },
		{ 246, 329, 14 },
		{ 246, 330, 14 },
		{ 246, 331, 14 },
		{ 246, 332, 14 },
		{ 246, 333, 14 },
		{ 246, 334, 14 },
		{ 246, 335, 14 },
		{ 246, 336, 14 },
		{ 246, 337, 14 },
		{ 246, 338, 14 },
		{ 246, 339, 14 },
		{ 246, 340, 14 },
		{ 246, 341, 14 },
		{ 246, 342, 14 },
		{ 246, 343, 14 },
		{ 246, 344, 14 },
		{ 246, 345, 14 },
		{ 246, 346, 14 },
		{ 246, 347, 14 },
		{ 246, 348, 14 },
		{ 246, 349, 14 },
		{ 246, 350, 14 },
		{ 246, 351, 14 },
		{ 246, 352, 14 },
		{ 246, 353, 14 },
		{ 246, 354, 14 },
		{ 246, 355, 14 },
		{ 246, 356, 14 },
		{ 246, 357, 14 },
		{ 246, 358, 14 },
		{ 246, 359, 14 },
		{ 246, 360, 14 },
		{ 246, 361, 14 },
		{ 246, 362, 14 },
		{ 246, 363, 14 },
		{ 246, 364, 14 },
		{ 246, 365, 14 },
		{ 246, 366, 14 },
		{ 246, 367, 14 },
		{ 246, 368, 14 },
		{ 246, 369, 14 },
		{ 246, 370, 14 },
		{ 246, 371, 14 },
		{ 246, 372, 14 },
		{ 246, 373, 14 },
		{ 246, 374, 14 },
		{ 246, 375, 14 },
		{ 246, 376, 14 },
		{ 246, 377, 14 },
		{ 246, 378, 14 },
		{ 246, 379, 14 },
		{ 246, 380, 14 },
		{ 246, 381, 14 },
		{ 246, 382, 14 },
		{ 246, 383, 14 },
		{ 247, 131, 14 },
		{ 247, 132, 14 },
		{ 247, 133, 14 },
		{ 247, 134, 14 },
		{ 247, 135, 14 },
		{ 247, 136, 14 },
		{ 247, 137, 14 },
		{ 247, 138, 14 },
		{ 247, 139, 14 },
		{ 247, 140, 14 },
		{ 247, 141, 14 },
		{ 247, 142, 14 },
		{ 247, 143, 14 },
		{ 247, 144, 14 },
		{ 247, 145, 14 },
		{ 247, 146, 14 },
		{ 247, 147, 14 },
		{ 247, 148, 14 },
		{ 247, 149, 14 },
		{ 247, 150, 14 },
		{ 247, 151, 14 },
		{ 247, 152, 14 },
		{ 247, 153, 14 },
		{ 247, 154, 14 },
		{ 247, 155, 14 },
		{ 247, 156, 14 },
		{ 247, 157, 14 },
		{ 247, 158, 14 },
		{ 247, 159, 14 },
		{ 247, 160, 14 },
		{ 247, 161, 14 },
		{ 247, 162, 14 },
		{ 247, 163, 14 },
		{ 247, 164, 14 },
		{ 247, 165, 14 },
		{ 247, 166, 14 },
		{ 247, 167, 14 },
		{ 247, 168, 14 },
		{ 247, 169, 14 },
		{ 247, 170, 14 },
		{ 247, 171, 14 },
		{ 247, 172, 14 },
		{ 247, 173, 14 },
		{ 247, 174, 14 },
		{ 247, 175, 14 },
		{ 247, 176, 14 },
		{ 247, 177, 14 },
		{ 247, 178, 14 },
		{ 247, 179, 14 },
		{ 247, 180, 14 },
		{ 247, 181, 14 },
		{ 247, 182, 14 },
		{ 247, 183, 14 },
		{ 247, 184, 14 },
		{ 247, 185, 14 },
		{ 247, 186, 14 },
		{ 247, 187, 14 },
		{ 247, 188, 14 },
		{ 247, 189, 14 },
		{ 247, 190, 14 },
		{ 247, 191, 14 },
		{ 247, 192, 14 },
		{ 247, 193, 14 },
		{ 247, 194, 14 },
		{ 247, 195, 14 },
		{ 247, 196, 14 },
		{ 247, 197, 14 },
		{ 247, 198, 14 },
		{ 247, 199, 14 },
		{ 247, 200, 14 },
		{ 247, 201, 14 },
		{ 247, 202, 14 },
		{ 247, 203, 14 },
		{ 247, 204, 14 },
		{ 247, 205, 14 },
		{ 247, 206, 14 },
		{ 247, 207, 14 },
		{ 247, 208, 14 },
		{ 247, 209, 14 },
		{ 247, 210, 14 },
		{ 247, 211, 14 },
		{ 247, 212, 14 },
		{ 247, 213, 14 },
		{ 247, 214, 14 },
		{ 247, 215, 14 },
		{ 247, 216, 14 },
		{ 247, 298, 14 },
		{ 247, 299, 14 },
		{ 247, 300, 14 },
		{ 247, 301, 14 },
		{ 247, 302, 14 },
		{ 247, 303, 14 },
		{ 247, 304, 14 },
		{ 247, 305, 14 },
		{ 247, 306, 14 },
		{ 247, 307, 14 },
		{ 247, 308, 14 },
		{ 247, 309, 14 },
		{ 247, 310, 14 },
		{ 247, 311, 14 },
		{ 247, 312, 14 },
		{ 247, 313, 14 },
		{ 247, 314, 14 },
		{ 247, 315, 14 },
		{ 247, 316, 14 },
		{ 247, 317, 14 },
		{ 247, 318, 14 },
		{ 247, 319, 14 },
		{ 247, 320, 14 },
		{ 247, 321, 14 },
		{ 247, 322, 14 },
		{ 247, 323, 14 },
		{ 247, 324, 14 },
		{ 247, 325, 14 },
		{ 247, 326, 14 },
		{ 247, 327, 14 },
		{ 247, 328, 14 },
		{ 247, 329, 14 },
		{ 247, 330, 14 },
		{ 247, 331, 14 },
		{ 247, 332, 14 },
		{ 247, 333, 14 },
		{ 247, 334, 14 },
		{ 247, 335, 14 },
		{ 247, 336, 14 },
		{ 247, 337, 14 },
		{ 247, 338, 14 },
		{ 247, 339, 14 },
		{ 247, 340, 14 },
		{ 247, 341, 14 },
		{ 247, 342, 14 },
		{ 247, 343, 14 },
		{ 247, 344, 14 },
		{ 247, 345, 14 },
		{ 247, 346, 14 },
		{ 247, 347, 14 },
		{ 247, 348, 14 },
		{ 247, 349, 14 },
		{ 247, 350, 14 },
		{ 247, 351, 14 },
		{ 247, 352, 14 },
		{ 247, 353, 14 },
		{ 247, 354, 14 },
		{ 247, 355, 14 },
		{ 247, 356, 14 },
		{ 247, 357, 14 },
		{ 247, 358, 14 },
		{ 247, 359, 14 },
		{ 247, 360, 14 },
		{ 247, 361, 14 },
		{ 247, 362, 14 },
		{ 247, 363, 14 },
		{ 247, 364, 14 },
		{ 247, 365, 14 },
		{ 247, 366, 14 },
		{ 247, 367, 14 },
		{ 247, 368, 14 },
		{ 247, 369, 14 },
		{ 247, 370, 14 },
		{ 247, 371, 14 },
		{ 247, 372, 14 },
		{ 247, 373, 14 },
		{ 247, 374, 14 },
		{ 247, 375, 14 },
		{ 247, 376, 14 },
		{ 247, 377, 14 },
		{ 247, 378, 14 },
		{ 247, 379, 14 },
		{ 247, 380, 14 },
		{ 247, 381, 14 },
		{ 247, 382, 14 },
		{ 247, 383, 14 },
		{ 248, 131, 14 },
		{ 248, 132, 14 },
		{ 248, 133, 14 },
		{ 248, 134, 14 },
		{ 248, 135, 14 },
		{ 248, 136, 14 },
		{ 248, 137, 14 },
		{ 248, 138, 14 },
		{ 248, 139, 14 },
		{ 248, 140, 14 },
		{ 248, 141, 14 },
		{ 248, 142, 14 },
		{ 248, 143, 14 },
		{ 248, 144, 14 },
		{ 248, 145, 14 },
		{ 248, 146, 14 },
		{ 248, 147, 14 },
		{ 248, 148, 14 },
		{ 248, 149, 14 },
		{ 248, 150, 14 },
		{ 248, 151, 14 },
		{ 248, 152, 14 },
		{ 248, 153, 14 },
		{ 248, 154, 14 },
		{ 248, 155, 14 },
		{ 248, 156, 14 },
		{ 248, 157, 14 },
		{ 248, 158, 14 },
		{ 248, 159, 14 },
		{ 248, 160, 14 },
		{ 248, 161, 14 },
		{ 248, 162, 14 },
		{ 248, 163, 14 },
		{ 248, 164, 14 },
		{ 248, 165, 14 },
		{ 248, 166, 14 },
		{ 248, 167, 14 },
		{ 248, 168, 14 },
		{ 248, 169, 14 },
		{ 248, 170, 14 },
		{ 248, 171, 14 },
		{ 248, 172, 14 },
		{ 248, 173, 14 },
		{ 248, 174, 14 },
		{ 248, 175, 14 },
		{ 248, 176, 14 },
		{ 248, 177, 14 },
		{ 248, 178, 14 },
		{ 248, 179, 14 },
		{ 248, 180, 14 },
		{ 248, 181, 14 },
		{ 248, 182, 14 },
		{ 248, 183, 14 },
		{ 248, 184, 14 },
		{ 248, 185, 14 },
		{ 248, 186, 14 },
		{ 248, 187, 14 },
		{ 248, 188, 14 },
		{ 248, 189, 14 },
		{ 248, 190, 14 },
		{ 248, 191, 14 },
		{ 248, 192, 14 },
		{ 248, 193, 14 },
		{ 248, 194, 14 },
		{ 248, 195, 14 },
		{ 248, 196, 14 },
		{ 248, 197, 14 },
		{ 248, 198, 14 },
		{ 248, 199, 14 },
		{ 248, 200, 14 },
		{ 248, 201, 14 },
		{ 248, 202, 14 },
		{ 248, 203, 14 },
		{ 248, 204, 14 },
		{ 248, 205, 14 },
		{ 248, 206, 14 },
		{ 248, 207, 14 },
		{ 248, 208, 14 },
		{ 248, 209, 14 },
		{ 248, 210, 14 },
		{ 248, 211, 14 },
		{ 248, 212, 14 },
		{ 248, 213, 14 },
		{ 248, 214, 14 },
		{ 248, 215, 14 },
		{ 248, 216, 14 },
		{ 248, 217, 14 },
		{ 248, 218, 14 },
		{ 248, 219, 14 },
		{ 248, 220, 14 },
		{ 248, 294, 14 },
		{ 248, 295, 14 },
		{ 248, 296, 14 },
		{ 248, 297, 14 },
		{ 248, 298, 14 },
		{ 248, 299, 14 },
		{ 248, 300, 14 },
		{ 248, 301, 14 },
		{ 248, 302, 14 },
		{ 248, 303, 14 },
		{ 248, 304, 14 },
		{ 248, 305, 14 },
		{ 248, 306, 14 },
		{ 248, 307, 14 },
		{ 248, 308, 14 },
		{ 248, 309, 14 },
		{ 248, 310, 14 },
		{ 248, 311, 14 },
		{ 248, 312, 14 },
		{ 248, 313, 14 },
		{ 248, 314, 14 },
		{ 248, 315, 14 },
		{ 248, 316, 14 },
		{ 248, 317, 14 },
		{ 248, 318, 14 },
		{ 248, 319, 14 },
		{ 248, 320, 14 },
		{ 248, 321, 14 },
		{ 248, 322, 14 },
		{ 248, 323, 14 },
		{ 248, 324, 14 },
		{ 248, 325, 14 },
		{ 248, 326, 14 },
		{ 248, 327, 14 },
		{ 248, 328, 14 },
		{ 248, 329, 14 },
		{ 248, 330, 14 },
		{ 248, 331, 14 },
		{ 248, 332, 14 },
		{ 248, 333, 14 },
		{ 248, 334, 14 },
		{ 248, 335, 14 },
		{ 248, 336, 14 },
		{ 248, 337, 14 },
		{ 248, 338, 14 },
		{ 248, 339, 14 },
		{ 248, 340, 14 },
		{ 248, 341, 14 },
		{ 248, 342, 14 },
		{ 248, 343, 14 },
		{ 248, 344, 14 },
		{ 248, 345, 14 },
		{ 248, 346, 14 },
		{ 248, 347, 14 },
		{ 248, 348, 14 },
		{ 248, 349, 14 },
		{ 248, 350, 14 },
		{ 248, 351, 14 },
		{ 248, 352, 14 },
		{ 248, 353, 14 },
		{ 248, 354, 14 },
		{ 248, 355, 14 },
		{ 248, 356, 14 },
		{ 248, 357, 14 },
		{ 248, 358, 14 },
		{ 248, 359, 14 },
		{ 248, 360, 14 },
		{ 248, 361, 14 },
		{ 248, 362, 14 },
		{ 248, 363, 14 },
		{ 248, 364, 14 },
		{ 248, 365, 14 },
		{ 248, 366, 14 },
		{ 248, 367, 14 },
		{ 248, 368, 14 },
		{ 248, 369, 14 },
		{ 248, 370, 14 },
		{ 248, 371, 14 },
		{ 248, 372, 14 },
		{ 248, 373, 14 },
		{ 248, 374, 14 },
		{ 248, 375, 14 },
		{ 248, 376, 14 },
		{ 248, 377, 14 },
		{ 248, 378, 14 },
		{ 248, 379, 14 },
		{ 248, 380, 14 },
		{ 248, 381, 14 },
		{ 248, 382, 14 },
		{ 248, 383, 14 },
		{ 249, 131, 14 },
		{ 249, 132, 14 },
		{ 249, 133, 14 },
		{ 249, 134, 14 },
		{ 249, 135, 14 },
		{ 249, 136, 14 },
		{ 249, 137, 14 },
		{ 249, 138, 14 },
		{ 249, 139, 14 },
		{ 249, 140, 14 },
		{ 249, 141, 14 },
		{ 249, 142, 14 },
		{ 249, 143, 14 },
		{ 249, 144, 14 },
		{ 249, 145, 14 },
		{ 249, 146, 14 },
		{ 249, 147, 14 },
		{ 249, 148, 14 },
		{ 249, 149, 14 },
		{ 249, 150, 14 },
		{ 249, 151, 14 },
		{ 249, 152, 14 },
		{ 249, 153, 14 },
		{ 249, 154, 14 },
		{ 249, 155, 14 },
		{ 249, 156, 14 },
		{ 249, 157, 14 },
		{ 249, 158, 14 },
		{ 249, 159, 14 },
		{ 249, 160, 14 },
		{ 249, 161, 14 },
		{ 249, 162, 14 },
		{ 249, 163, 14 },
		{ 249, 164, 14 },
		{ 249, 165, 14 },
		{ 249, 166, 14 },
		{ 249, 167, 14 },
		{ 249, 168, 14 },
		{ 249, 169, 14 },
		{ 249, 170, 14 },
		{ 249, 171, 14 },
		{ 249, 172, 14 },
		{ 249, 173, 14 },
		{ 249, 174, 14 },
		{ 249, 175, 14 },
		{ 249, 176, 14 },
		{ 249, 177, 14 },
		{ 249, 178, 14 },
		{ 249, 179, 14 },
		{ 249, 180, 14 },
		{ 249, 181, 14 },
		{ 249, 182, 14 },
		{ 249, 183, 14 },
		{ 249, 184, 14 },
		{ 249, 185, 14 },
		{ 249, 186, 14 },
		{ 249, 187, 14 },
		{ 249, 188, 14 },
		{ 249, 189, 14 },
		{ 249, 190, 14 },
		{ 249, 191, 14 },
		{ 249, 192, 14 },
		{ 249, 193, 14 },
		{ 249, 194, 14 },
		{ 249, 195, 14 },
		{ 249, 196, 14 },
		{ 249, 197, 14 },
		{ 249, 198, 14 },
		{ 249, 199, 14 },
		{ 249, 200, 14 },
		{ 249, 201, 14 },
		{ 249, 202, 14 },
		{ 249, 203, 14 },
		{ 249, 204, 14 },
		{ 249, 205, 14 },
		{ 249, 206, 14 },
		{ 249, 207, 14 },
		{ 249, 208, 14 },
		{ 249, 209, 14 },
		{ 249, 210, 14 },
		{ 249, 211, 14 },
		{ 249, 212, 14 },
		{ 249, 213, 14 },
		{ 249, 214, 14 },
		{ 249, 215, 14 },
		{ 249, 216, 14 },
		{ 249, 217, 14 },
		{ 249, 218, 14 },
		{ 249, 219, 14 },
		{ 249, 220, 14 },
		{ 249, 221, 14 },
		{ 249, 222, 14 },
		{ 249, 223, 14 },
		{ 249, 224, 14 },
		{ 249, 290, 14 },
		{ 249, 291, 14 },
		{ 249, 292, 14 },
		{ 249, 293, 14 },
		{ 249, 294, 14 },
		{ 249, 295, 14 },
		{ 249, 296, 14 },
		{ 249, 297, 14 },
		{ 249, 298, 14 },
		{ 249, 299, 14 },
		{ 249, 300, 14 },
		{ 249, 301, 14 },
		{ 249, 302, 14 },
		{ 249, 303, 14 },
		{ 249, 304, 14 },
		{ 249, 305, 14 },
		{ 249, 306, 14 },
		{ 249, 307, 14 },
		{ 249, 308, 14 },
		{ 249, 309, 14 },
		{ 249, 310, 14 },
		{ 249, 311, 14 },
		{ 249, 312, 14 },
		{ 249, 313, 14 },
		{ 249, 314, 14 },
		{ 249, 315, 14 },
		{ 249, 316, 14 },
		{ 249, 317, 14 },
		{ 249, 318, 14 },
		{ 249, 319, 14 },
		{ 249, 320, 14 },
		{ 249, 321, 14 },
		{ 249, 322, 14 },
		{ 249, 323, 14 },
		{ 249, 324, 14 },
		{ 249, 325, 14 },
		{ 249, 326, 14 },
		{ 249, 327, 14 },
		{ 249, 328, 14 },
		{ 249, 329, 14 },
		{ 249, 330, 14 },
		{ 249, 331, 14 },
		{ 249, 332, 14 },
		{ 249, 333, 14 },
		{ 249, 334, 14 },
		{ 249, 335, 14 },
		{ 249, 336, 14 },
		{ 249, 337, 14 },
		{ 249, 338, 14 },
		{ 249, 339, 14 },
		{ 249, 340, 14 },
		{ 249, 341, 14 },
		{ 249, 342, 14 },
		{ 249, 343, 14 },
		{ 249, 344, 14 },
		{ 249, 345, 14 },
		{ 249, 346, 14 },
		{ 249, 347, 14 },
		{ 249, 348, 14 },
		{ 249, 349, 14 },
		{ 249, 350, 14 },
		{ 249, 351, 14 },
		{ 249, 352, 14 },
		{ 249, 353, 14 },
		{ 249, 354, 14 },
		{ 249, 355, 14 },
		{ 249, 356, 14 },
		{ 249, 357, 14 },
		{ 249, 358, 14 },
		{ 249, 359, 14 },
		{ 249, 360, 14 },
		{ 249, 361, 14 },
		{ 249, 362, 14 },
		{ 249, 363, 14 },
		{ 249, 364, 14 },
		{ 249, 365, 14 },
		{ 249, 366, 14 },
		{ 249, 367, 14 },
		{ 249, 368, 14 },
		{ 249, 369, 14 },
		{ 249, 370, 14 },
		{ 249, 371, 14 },
		{ 249, 372, 14 },
		{ 249, 373, 14 },
		{ 249, 374, 14 },
		{ 249, 375, 14 },
		{ 249, 376, 14 },
		{ 249, 377, 14 },
		{ 249, 378, 14 },
		{ 249, 379, 14 },
		{ 249, 380, 14 },
		{ 249, 381, 14 },
		{ 249, 382, 14 },
		{ 249, 383, 14 },
		{ 250, 131, 14 },
		{ 250, 132, 14 },
		{ 250, 133, 14 },
		{ 250, 134, 14 },
		{ 250, 135, 14 },
		{ 250, 136, 14 },
		{ 250, 137, 14 },
		{ 250, 138, 14 },
		{ 250, 139, 14 },
		{ 250, 140, 14 },
		{ 250, 141, 14 },
		{ 250, 142, 14 },
		{ 250, 143, 14 },
		{ 250, 144, 14 },
		{ 250, 145, 14 },
		{ 250, 146, 14 },
		{ 250, 147, 14 },
		{ 250, 148, 14 },
		{ 250, 149, 14 },
		{ 250, 150, 14 },
		{ 250, 151, 14 },
		{ 250, 152, 14 },
		{ 250, 153, 14 },
		{ 250, 154, 14 },
		{ 250, 155, 14 },
		{ 250, 156, 14 },
		{ 250, 157, 14 },
		{ 250, 158, 14 },
		{ 250, 159, 14 },
		{ 250, 160, 14 },
		{ 250, 161, 14 },
		{ 250, 162, 14 },
		{ 250, 163, 14 },
		{ 250, 164, 14 },
		{ 250, 165, 14 },
		{ 250, 166, 14 },
		{ 250, 167, 14 },
		{ 250, 168, 14 },
		{ 250, 169, 14 },
		{ 250, 170, 14 },
		{ 250, 171, 14 },
		{ 250, 172, 14 },
		{ 250, 173, 14 },
		{ 250, 174, 14 },
		{ 250, 175, 14 },
		{ 250, 176, 14 },
		{ 250, 177, 14 },
		{ 250, 178, 14 },
		{ 250, 179, 14 },
		{ 250, 180, 14 },
		{ 250, 181, 14 },
		{ 250, 182, 14 },
		{ 250, 183, 14 },
		{ 250, 184, 14 },
		{ 250, 185, 14 },
		{ 250, 186, 14 },
		{ 250, 187, 14 },
		{ 250, 188, 14 },
		{ 250, 189, 14 },
		{ 250, 190, 14 },
		{ 250, 191, 14 },
		{ 250, 192, 14 },
		{ 250, 193, 14 },
		{ 250, 194, 14 },
		{ 250, 195, 14 },
		{ 250, 196, 14 },
		{ 250, 197, 14 },
		{ 250, 198, 14 },
		{ 250, 199, 14 },
		{ 250, 200, 14 },
		{ 250, 201, 14 },
		{ 250, 202, 14 },
		{ 250, 203, 14 },
		{ 250, 204, 14 },
		{ 250, 205, 14 },
		{ 250, 206, 14 },
		{ 250, 207, 14 },
		{ 250, 208, 14 },
		{ 250, 209, 14 },
		{ 250, 210, 14 },
		{ 250, 211, 14 },
		{ 250, 212, 14 },
		{ 250, 213, 14 },
		{ 250, 214, 14 },
		{ 250, 215, 14 },
		{ 250, 216, 14 },
		{ 250, 217, 14 },
		{ 250, 218, 14 },
		{ 250, 219, 14 },
		{ 250, 220, 14 },
		{ 250, 221, 14 },
		{ 250, 222, 14 },
		{ 250, 223, 14 },
		{ 250, 224, 14 },
		{ 250, 225, 14 },
		{ 250, 226, 14 },
		{ 250, 227, 14 },
		{ 250, 228, 14 },
		{ 250, 286, 14 },
		{ 250, 287, 14 },
		{ 250, 288, 14 },
		{ 250, 289, 14 },
		{ 250, 290, 14 },
		{ 250, 291, 14 },
		{ 250, 292, 14 },
		{ 250, 293, 14 },
		{ 250, 294, 14 },
		{ 250, 295, 14 },
		{ 250, 296, 14 },
		{ 250, 297, 14 },
		{ 250, 298, 14 },
		{ 250, 299, 14 },
		{ 250, 300, 14 },
		{ 250, 301, 14 },
		{ 250, 302, 14 },
		{ 250, 303, 14 },
		{ 250, 304, 14 },
		{ 250, 305, 14 },
		{ 250, 306, 14 },
		{ 250, 307, 14 },
		{ 250, 308, 14 },
		{ 250, 309, 14 },
		{ 250, 310, 14 },
		{ 250, 311, 14 },
		{ 250, 312, 14 },
		{ 250, 313, 14 },
		{ 250, 314, 14 },
		{ 250, 315, 14 },
		{ 250, 316, 14 },
		{ 250, 317, 14 },
		{ 250, 318, 14 },
		{ 250, 319, 14 },
		{ 250, 320, 14 },
		{ 250, 321, 14 },
		{ 250, 322, 14 },
		{ 250, 323, 14 },
		{ 250, 324, 14 },
		{ 250, 325, 14 },
		{ 250, 326, 14 },
		{ 250, 327, 14 },
		{ 250, 328, 14 },
		{ 250, 329, 14 },
		{ 250, 330, 14 },
		{ 250, 331, 14 },
		{ 250, 332, 14 },
		{ 250, 333, 14 },
		{ 250, 334, 14 },
		{ 250, 335, 14 },
		{ 250, 336, 14 },
		{ 250, 337, 14 },
		{ 250, 338, 14 },
		{ 250, 339, 14 },
		{ 250, 340, 14 },
		{ 250, 341, 14 },
		{ 250, 342, 14 },
		{ 250, 343, 14 },
		{ 250, 344, 14 },
		{ 250, 345, 14 },
		{ 250, 346, 14 },
		{ 250, 347, 14 },
		{ 250, 348, 14 },
		{ 250, 349, 14 },
		{ 250, 350, 14 },
		{ 250, 351, 14 },
		{ 250, 352, 14 },
		{ 250, 353, 14 },
		{ 250, 354, 14 },
		{ 250, 355, 14 },
		{ 250, 356, 14 },
		{ 250, 357, 14 },
		{ 250, 358, 14 },
		{ 250, 359, 14 },
		{ 250, 360, 14 },
		{ 250, 361, 14 },
		{ 250, 362, 14 },
		{ 250, 363, 14 },
		{ 250, 364, 14 },
		{ 250, 365, 14 },
		{ 250, 366, 14 },
		{ 250, 367, 14 },
		{ 250, 368, 14 },
		{ 250, 369, 14 },
		{ 250, 370, 14 },
		{ 250, 371, 14 },
		{ 250, 372, 14 },
		{ 250, 373, 14 },
		{ 250, 374, 14 },
		{ 250, 375, 14 },
		{ 250, 376, 14 },
		{ 250, 377, 14 },
		{ 250, 378, 14 },
		{ 250, 379, 14 },
		{ 250, 380, 14 },
		{ 250, 381, 14 },
		{ 250, 382, 14 },
		{ 250, 383, 14 },
		{ 251, 131, 14 },
		{ 251, 132, 14 },
		{ 251, 133, 14 },
		{ 251, 134, 14 },
		{ 251, 135, 14 },
		{ 251, 136, 14 },
		{ 251, 137, 14 },
		{ 251, 138, 14 },
		{ 251, 139, 14 },
		{ 251, 140, 14 },
		{ 251, 141, 14 },
		{ 251, 142, 14 },
		{ 251, 143, 14 },
		{ 251, 144, 14 },
		{ 251, 145, 14 },
		{ 251, 146, 14 },
		{ 251, 147, 14 },
		{ 251, 148, 14 },
		{ 251, 149, 14 },
		{ 251, 150, 14 },
		{ 251, 151, 14 },
		{ 251, 152, 14 },
		{ 251, 153, 14 },
		{ 251, 154, 14 },
		{ 251, 155, 14 },
		{ 251, 156, 14 },
		{ 251, 157, 14 },
		{ 251, 158, 14 },
		{ 251, 159, 14 },
		{ 251, 160, 14 },
		{ 251, 161, 14 },
		{ 251, 162, 14 },
		{ 251, 163, 14 },
		{ 251, 164, 14 },
		{ 251, 165, 14 },
		{ 251, 166, 14 },
		{ 251, 167, 14 },
		{ 251, 168, 14 },
		{ 251, 169, 14 },
		{ 251, 170, 14 },
		{ 251, 171, 14 },
		{ 251, 172, 14 },
		{ 251, 173, 14 },
		{ 251, 174, 14 },
		{ 251, 175, 14 },
		{ 251, 176, 14 },
		{ 251, 177, 14 },
		{ 251, 178, 14 },
		{ 251, 179, 14 },
		{ 251, 180, 14 },
		{ 251, 181, 14 },
		{ 251, 182, 14 },
		{ 251, 183, 14 },
		{ 251, 184, 14 },
		{ 251, 185, 14 },
		{ 251, 186, 14 },
		{ 251, 187, 14 },
		{ 251, 188, 14 },
		{ 251, 189, 14 },
		{ 251, 190, 14 },
		{ 251, 191, 14 },
		{ 251, 192, 14 },
		{ 251, 193, 14 },
		{ 251, 194, 14 },
		{ 251, 195, 14 },
		{ 251, 196, 14 },
		{ 251, 197, 14 },
		{ 251, 198, 14 },
		{ 251, 199, 14 },
		{ 251, 200, 14 },
		{ 251, 201, 14 },
		{ 251, 202, 14 },
		{ 251, 203, 14 },
		{ 251, 204, 14 },
		{ 251, 205, 14 },
		{ 251, 206, 14 },
		{ 251, 207, 14 },
		{ 251, 208, 14 },
		{ 251, 209, 14 },
		{ 251, 210, 14 },
		{ 251, 211, 14 },
		{ 251, 212, 14 },
		{ 251, 213, 14 },
		{ 251, 214, 14 },
		{ 251, 215, 14 },
		{ 251, 216, 14 },
		{ 251, 217, 14 },
		{ 251, 218, 14 },
		{ 251, 219, 14 },
		{ 251, 220, 14 },
		{ 251, 221, 14 },
		{ 251, 222, 14 },
		{ 251, 223, 14 },
		{ 251, 224, 14 },
		{ 251, 225, 14 },
		{ 251, 226, 14 },
		{ 251, 227, 14 },
		{ 251, 228, 14 },
		{ 251, 229, 14 },
		{ 251, 230, 14 },
		{ 251, 231, 14 },
		{ 251, 232, 14 },
		{ 251, 282, 14 },
		{ 251, 283, 14 },
		{ 251, 284, 14 },
		{ 251, 285, 14 },
		{ 251, 286, 14 },
		{ 251, 287, 14 },
		{ 251, 288, 14 },
		{ 251, 289, 14 },
		{ 251, 290, 14 },
		{ 251, 291, 14 },
		{ 251, 292, 14 },
		{ 251, 293, 14 },
		{ 251, 294, 14 },
		{ 251, 295, 14 },
		{ 251, 296, 14 },
		{ 251, 297, 14 },
		{ 251, 298, 14 },
		{ 251, 299, 14 },
		{ 251, 300, 14 },
		{ 251, 301, 14 },
		{ 251, 302, 14 },
		{ 251, 303, 14 },
		{ 251, 304, 14 },
		{ 251, 305, 14 },
		{ 251, 306, 14 },
		{ 251, 307, 14 },
		{ 251, 308, 14 },
		{ 251, 309, 14 },
		{ 251, 310, 14 },
		{ 251, 311, 14 },
		{ 251, 312, 14 },
		{ 251, 313, 14 },
		{ 251, 314, 14 },
		{ 251, 315, 14 },
		{ 251, 316, 14 },
		{ 251, 317, 14 },
		{ 251, 318, 14 },
		{ 251, 319, 14 },
		{ 251, 320, 14 },
		{ 251, 321, 14 },
		{ 251, 322, 14 },
		{ 251, 323, 14 },
		{ 251, 324, 14 },
		{ 251, 325, 14 },
		{ 251, 326, 14 },
		{ 251, 327, 14 },
		{ 251, 328, 14 },
		{ 251, 329, 14 },
		{ 251, 330, 14 },
		{ 251, 331, 14 },
		{ 251, 332, 14 },
		{ 251, 333, 14 },
		{ 251, 334, 14 },
		{ 251, 335, 14 },
		{ 251, 336, 14 },
		{ 251, 337, 14 },
		{ 251, 338, 14 },
		{ 251, 339, 14 },
		{ 251, 340, 14 },
		{ 251, 341, 14 },
		{ 251, 342, 14 },
		{ 251, 343, 14 },
		{ 251, 344, 14 },
		{ 251, 345, 14 },
		{ 251, 346, 14 },
		{ 251, 347, 14 },
		{ 251, 348, 14 },
		{ 251, 349, 14 },
		{ 251, 350, 14 },
		{ 251, 351, 14 },
		{ 251, 352, 14 },
		{ 251, 353, 14 },
		{ 251, 354, 14 },
		{ 251, 355, 14 },
		{ 251, 356, 14 },
		{ 251, 357, 14 },
		{ 251, 358, 14 },
		{ 251, 359, 14 },
		{ 251, 360, 14 },
		{ 251, 361, 14 },
		{ 251, 362, 14 },
		{ 251, 363, 14 },
		{ 251, 364, 14 },
		{ 251, 365, 14 },
		{ 251, 366, 14 },
		{ 251, 367, 14 },
		{ 251, 368, 14 },
		{ 251, 369, 14 },
		{ 251, 370, 14 },
		{ 251, 371, 14 },
		{ 251, 372, 14 },
		{ 251, 373, 14 },
		{ 251, 374, 14 },
		{ 251, 375, 14 },
		{ 251, 376, 14 },
		{ 251, 377, 14 },
		{ 251, 378, 14 },
		{ 251, 379, 14 },
		{ 251, 380, 14 },
		{ 251, 381, 14 },
		{ 251, 382, 14 },
		{ 251, 383, 14 },
		{ 252, 131, 14 },
		{ 252, 132, 14 },
		{ 252, 133, 14 },
		{ 252, 134, 14 },
		{ 252, 135, 14 },
		{ 252, 136, 14 },
		{ 252, 137, 14 },
		{ 252, 138, 14 },
		{ 252, 139, 14 },
		{ 252, 140, 14 },
		{ 252, 141, 14 },
		{ 252, 142, 14 },
		{ 252, 143, 14 },
		{ 252, 144, 14 },
		{ 252, 145, 14 },
		{ 252, 146, 14 },
		{ 252, 147, 14 },
		{ 252, 148, 14 },
		{ 252, 149, 14 },
		{ 252, 150, 14 },
		{ 252, 151, 14 },
		{ 252, 152, 14 },
		{ 252, 153, 14 },
		{ 252, 154, 14 },
		{ 252, 155, 14 },
		{ 252, 156, 14 },
		{ 252, 157, 14 },
		{ 252, 158, 14 },
		{ 252, 159, 14 },
		{ 252, 160, 14 },
		{ 252, 161, 14 },
		{ 252, 162, 14 },
		{ 252, 163, 14 },
		{ 252, 164, 14 },
		{ 252, 165, 14 },
		{ 252, 166, 14 },
		{ 252, 167, 14 },
		{ 252, 168, 14 },
		{ 252, 169, 14 },
		{ 252, 170, 14 },
		{ 252, 171, 14 },
		{ 252, 172, 14 },
		{ 252, 173, 14 },
		{ 252, 174, 14 },
		{ 252, 175, 14 },
		{ 252, 176, 14 },
		{ 252, 177, 14 },
		{ 252, 178, 14 },
		{ 252, 179, 14 },
		{ 252, 180, 14 },
		{ 252, 181, 14 },
		{ 252, 182, 14 },
		{ 252, 183, 14 },
		{ 252, 184, 14 },
		{ 252, 185, 14 },
		{ 252, 186, 14 },
		{ 252, 187, 14 },
		{ 252, 188, 14 },
		{ 252, 189, 14 },
		{ 252, 190, 14 },
		{ 252, 191, 14 },
		{ 252, 192, 14 },
		{ 252, 193, 14 },
		{ 252, 194, 14 },
		{ 252, 195, 14 },
		{ 252, 196, 14 },
		{ 252, 197, 14 },
		{ 252, 198, 14 },
		{ 252, 199, 14 },
		{ 252, 200, 14 },
		{ 252, 201, 14 },
		{ 252, 202, 14 },
		{ 252, 203, 14 },
		{ 252, 204, 14 },
		{ 252, 205, 14 },
		{ 252, 206, 14 },
		{ 252, 207, 14 },
		{ 252, 208, 14 },
		{ 252, 209, 14 },
		{ 252, 210, 14 },
		{ 252, 211, 14 },
		{ 252, 212, 14 },
		{ 252, 213, 14 },
		{ 252, 214, 14 },
		{ 252, 215, 14 },
		{ 252, 216, 14 },
		{ 252, 217, 14 },
		{ 252, 218, 14 },
		{ 252, 219, 14 },
		{ 252, 220, 14 },
		{ 252, 221, 14 },
		{ 252, 222, 14 },
		{ 252, 223, 14 },
		{ 252, 224, 14 },
		{ 252, 225, 14 },
		{ 252, 226, 14 },
		{ 252, 227, 14 },
		{ 252, 228, 14 },
		{ 252, 229, 14 },
		{ 252, 230, 14 },
		{ 252, 231, 14 },
		{ 252, 232, 14 },
		{ 252, 233, 14 },
		{ 252, 234, 14 },
		{ 252, 235, 14 },
		{ 252, 236, 14 },
		{ 252, 278, 14 },
		{ 252, 279, 14 },
		{ 252, 280, 14 },
		{ 252, 281, 14 },
		{ 252, 282, 14 },
		{ 252, 283, 14 },
		{ 252, 284, 14 },
		{ 252, 285, 14 },
		{ 252, 286, 14 },
		{ 252, 287, 14 },
		{ 252, 288, 14 },
		{ 252, 289, 14 },
		{ 252, 290, 14 },
		{ 252, 291, 14 },
		{ 252, 292, 14 },
		{ 252, 293, 14 },
		{ 252, 294, 14 },
		{ 252, 295, 14 },
		{ 252, 296, 14 },
		{ 252, 297, 14 },
		{ 252, 298, 14 },
		{ 252, 299, 14 },
		{ 252, 300, 14 },
		{ 252, 301, 14 },
		{ 252, 302, 14 },
		{ 252, 303, 14 },
		{ 252, 304, 14 },
		{ 252, 305, 14 },
		{ 252, 306, 14 },
		{ 252, 307, 14 },
		{ 252, 308, 14 },
		{ 252, 309, 14 },
		{ 252, 310, 14 },
		{ 252, 311, 14 },
		{ 252, 312, 14 },
		{ 252, 313, 14 },
		{ 252, 314, 14 },
		{ 252, 315, 14 },
		{ 252, 316, 14 },
		{ 252, 317, 14 },
		{ 252, 318, 14 },
		{ 252, 319, 14 },
		{ 252, 320, 14 },
		{ 252, 321, 14 },
		{ 252, 322, 14 },
		{ 252, 323, 14 },
		{ 252, 324, 14 },
		{ 252, 325, 14 },
		{ 252, 326, 14 },
		{ 252, 327, 14 },
		{ 252, 328, 14 },
		{ 252, 329, 14 },
		{ 252, 330, 14 },
		{ 252, 331, 14 },
		{ 252, 332, 14 },
		{ 252, 333, 14 },
		{ 252, 334, 14 },
		{ 252, 335, 14 },
		{ 252, 336, 14 },
		{ 252, 337, 14 },
		{ 252, 338, 14 },
		{ 252, 339, 14 },
		{ 252, 340, 14 },
		{ 252, 341, 14 },
		{ 252, 342, 14 },
		{ 252, 343, 14 },
		{ 252, 344, 14 },
		{ 252, 345, 14 },
		{ 252, 346, 14 },
		{ 252, 347, 14 },
		{ 252, 348, 14 },
		{ 252, 349, 14 },
		{ 252, 350, 14 },
		{ 252, 351, 14 },
		{ 252, 352, 14 },
		{ 252, 353, 14 },
		{ 252, 354, 14 },
		{ 252, 355, 14 },
		{ 252, 356, 14 },
		{ 252, 357, 14 },
		{ 252, 358, 14 },
		{ 252, 359, 14 },
		{ 252, 360, 14 },
		{ 252, 361, 14 },
		{ 252, 362, 14 },
		{ 252, 363, 14 },
		{ 252, 364, 14 },
		{ 252, 365, 14 },
		{ 252, 366, 14 },
		{ 252, 367, 14 },
		{ 252, 368, 14 },
		{ 252, 369, 14 },
		{ 252, 370, 14 },
		{ 252, 371, 14 },
		{ 252, 372, 14 },
		{ 252, 373, 14 },
		{ 252, 374, 14 },
		{ 252, 375, 14 },
		{ 252, 376, 14 },
		{ 252, 377, 14 },
		{ 252, 378, 14 },
		{ 252, 379, 14 },
		{ 252, 380, 14 },
		{ 252, 381, 14 },
		{ 252, 382, 14 },
		{ 252, 383, 14 },
		{ 253, 131, 14 },
		{ 253, 132, 14 },
		{ 253, 133, 14 },
		{ 253, 134, 14 },
		{ 253, 135, 14 },
		{ 253, 136, 14 },
		{ 253, 137, 14 },
		{ 253, 138, 14 },
		{ 253, 139, 14 },
		{ 253, 140, 14 },
		{ 253, 141, 14 },
		{ 253, 142, 14 },
		{ 253, 143, 14 },
		{ 253, 144, 14 },
		{ 253, 145, 14 },
		{ 253, 146, 14 },
		{ 253, 147, 14 },
		{ 253, 148, 14 },
		{ 253, 149, 14 },
		{ 253, 150, 14 },
		{ 253, 151, 14 },
		{ 253, 152, 14 },
		{ 253, 153, 14 },
		{ 253, 154, 14 },
		{ 253, 155, 14 },
		{ 253, 156, 14 },
		{ 253, 157, 14 },
		{ 253, 158, 14 },
		{ 253, 159, 14 },
		{ 253, 160, 14 },
		{ 253, 161, 14 },
		{ 253, 162, 14 },
		{ 253, 163, 14 },
		{ 253, 164, 14 },
		{ 253, 165, 14 },
		{ 253, 166, 14 },
		{ 253, 167, 14 },
		{ 253, 168, 14 },
		{ 253, 169, 14 },
		{ 253, 170, 14 },
		{ 253, 171, 14 },
		{ 253, 172, 14 },
		{ 253, 173, 14 },
		{ 253, 174, 14 },
		{ 253, 175, 14 },
		{ 253, 176, 14 },
		{ 253, 177, 14 },
		{ 253, 178, 14 },
		{ 253, 179, 14 },
		{ 253, 180, 14 },
		{ 253, 181, 14 },
		{ 253, 182, 14 },
		{ 253, 183, 14 },
		{ 253, 184, 14 },
		{ 253, 185, 14 },
		{ 253, 186, 14 },
		{ 253, 187, 14 },
		{ 253, 188, 14 },
		{ 253, 189, 14 },
		{ 253, 190, 14 },
		{ 253, 191, 14 },
		{ 253, 192, 14 },
		{ 253, 193, 14 },
		{ 253, 194, 14 },
		{ 253, 195, 14 },
		{ 253, 196, 14 },
		{ 253, 197, 14 },
		{ 253, 198, 14 },
		{ 253, 199, 14 },
		{ 253, 200, 14 },
		{ 253, 201, 14 },
		{ 253, 202, 14 },
		{ 253, 203, 14 },
		{ 253, 204, 14 },
		{ 253, 205, 14 },
		{ 253, 206, 14 },
		{ 253, 207, 14 },
		{ 253, 208, 14 },
		{ 253, 209, 14 },
		{ 253, 210, 14 },
		{ 253, 211, 14 },
		{ 253, 212, 14 },
		{ 253, 213, 14 },
		{ 253, 214, 14 },
		{ 253, 215, 14 },
		{ 253, 216, 14 },
		{ 253, 217, 14 },
		{ 253, 218, 14 },
		{ 253, 219, 14 },
		{ 253, 220, 14 },
		{ 253, 221, 14 },
		{ 253, 222, 14 },
		{ 253, 223, 14 },
		{ 253, 224, 14 },
		{ 253, 225, 14 },
		{ 253, 226, 14 },
		{ 253, 227, 14 },
		{ 253, 228, 14 },
		{ 253, 229, 14 },
		{ 253, 230, 14 },
		{ 253, 231, 14 },
		{ 253, 232, 14 },
		{ 253, 233, 14 },
		{ 253, 234, 14 },
		{ 253, 235, 14 },
		{ 253, 236, 14 },
		{ 253, 237, 14 },
		{ 253, 238, 14 },
		{ 253, 239, 14 },
		{ 253, 240, 14 },
		{ 253, 274, 14 },
		{ 253, 275, 14 },
		{ 253, 276, 14 },
		{ 253, 277, 14 },
		{ 253, 278, 14 },
		{ 253, 279, 14 },
		{ 253, 280, 14 },
		{ 253, 281, 14 },
		{ 253, 282, 14 },
		{ 253, 283, 14 },
		{ 253, 284, 14 },
		{ 253, 285, 14 },
		{ 253, 286, 14 },
		{ 253, 287, 14 },
		{ 253, 288, 14 },
		{ 253, 289, 14 },
		{ 253, 290, 14 },
		{ 253, 291, 14 },
		{ 253, 292, 14 },
		{ 253, 293, 14 },
		{ 253, 294, 14 },
		{ 253, 295, 14 },
		{ 253, 296, 14 },
		{ 253, 297, 14 },
		{ 253, 298, 14 },
		{ 253, 299, 14 },
		{ 253, 300, 14 },
		{ 253, 301, 14 },
		{ 253, 302, 14 },
		{ 253, 303, 14 },
		{ 253, 304, 14 },
		{ 253, 305, 14 },
		{ 253, 306, 14 },
		{ 253, 307, 14 },
		{ 253, 308, 14 },
		{ 253, 309, 14 },
		{ 253, 310, 14 },
		{ 253, 311, 14 },
		{ 253, 312, 14 },
		{ 253, 313, 14 },
		{ 253, 314, 14 },
		{ 253, 315, 14 },
		{ 253, 316, 14 },
		{ 253, 317, 14 },
		{ 253, 318, 14 },
		{ 253, 319, 14 },
		{ 253, 320, 14 },
		{ 253, 321, 14 },
		{ 253, 322, 14 },
		{ 253, 323, 14 },
		{ 253, 324, 14 },
		{ 253, 325, 14 },
		{ 253, 326, 14 },
		{ 253, 327, 14 },
		{ 253, 328, 14 },
		{ 253, 329, 14 },
		{ 253, 330, 14 },
		{ 253, 331, 14 },
		{ 253, 332, 14 },
		{ 253, 333, 14 },
		{ 253, 334, 14 },
		{ 253, 335, 14 },
		{ 253, 336, 14 },
		{ 253, 337, 14 },
		{ 253, 338, 14 },
		{ 253, 339, 14 },
		{ 253, 340, 14 },
		{ 253, 341, 14 },
		{ 253, 342, 14 },
		{ 253, 343, 14 },
		{ 253, 344, 14 },
		{ 253, 345, 14 },
		{ 253, 346, 14 },
		{ 253, 347, 14 },
		{ 253, 348, 14 },
		{ 253, 349, 14 },
		{ 253, 350, 14 },
		{ 253, 351, 14 },
		{ 253, 352, 14 },
		{ 253, 353, 14 },
		{ 253, 354, 14 },
		{ 253, 355, 14 },
		{ 253, 356, 14 },
		{ 253, 357, 14 },
		{ 253, 358, 14 },
		{ 253, 359, 14 },
		{ 253, 360, 14 },
		{ 253, 361, 14 },
		{ 253, 362, 14 },
		{ 253, 363, 14 },
		{ 253, 364, 14 },
		{ 253, 365, 14 },
		{ 253, 366, 14 },
		{ 253, 367, 14 },
		{ 253, 368, 14 },
		{ 253, 369, 14 },
		{ 253, 370, 14 },
		{ 253, 371, 14 },
		{ 253, 372, 14 },
		{ 253, 373, 14 },
		{ 253, 374, 14 },
		{ 253, 375, 14 },
		{ 253, 376, 14 },
		{ 253, 377, 14 },
		{ 253, 378, 14 },
		{ 253, 379, 14 },
		{ 253, 380, 14 },
		{ 253, 381, 14 },
		{ 253, 382, 14 },
		{ 253, 383, 14 },
		{ 254, 131, 14 },
		{ 254, 132, 14 },
		{ 254, 133, 14 },
		{ 254, 134, 14 },
		{ 254, 135, 14 },
		{ 254, 136, 14 },
		{ 254, 137, 14 },
		{ 254, 138, 14 },
		{ 254, 139, 14 },
		{ 254, 140, 14 },
		{ 254, 141, 14 },
		{ 254, 142, 14 },
		{ 254, 143, 14 },
		{ 254, 144, 14 },
		{ 254, 145, 14 },
		{ 254, 146, 14 },
		{ 254, 147, 14 },
		{ 254, 148, 14 },
		{ 254, 149, 14 },
		{ 254, 150, 14 },
		{ 254, 151, 14 },
		{ 254, 152, 14 },
		{ 254, 153, 14 },
		{ 254, 154, 14 },
		{ 254, 155, 14 },
		{ 254, 156, 14 },
		{ 254, 157, 14 },
		{ 254, 158, 14 },
		{ 254, 159, 14 },
		{ 254, 160, 14 },
		{ 254, 161, 14 },
		{ 254, 162, 14 },
		{ 254, 163, 14 },
		{ 254, 164, 14 },
		{ 254, 165, 14 },
		{ 254, 166, 14 },
		{ 254, 167, 14 },
		{ 254, 168, 14 },
		{ 254, 169, 14 },
		{ 254, 170, 14 },
		{ 254, 171, 14 },
		{ 254, 172, 14 },
		{ 254, 173, 14 },
		{ 254, 174, 14 },
		{ 254, 175, 14 },
		{ 254, 176, 14 },
		{ 254, 177, 14 },
		{ 254, 178, 14 },
		{ 254, 179, 14 },
		{ 254, 180, 14 },
		{ 254, 181, 14 },
		{ 254, 182, 14 },
		{ 254, 183, 14 },
		{ 254, 184, 14 },
		{ 254, 185, 14 },
		{ 254, 186, 14 },
		{ 254, 187, 14 },
		{ 254, 188, 14 },
		{ 254, 189, 14 },
		{ 254, 190, 14 },
		{ 254, 191, 14 },
		{ 254, 192, 14 },
		{ 254, 193, 14 },
		{ 254, 194, 14 },
		{ 254, 195, 14 },
		{ 254, 196, 14 },
		{ 254, 197, 14 },
		{ 254, 198, 14 },
		{ 254, 199, 14 },
		{ 254, 200, 14 },
		{ 254, 201, 14 },
		{ 254, 202, 14 },
		{ 254, 203, 14 },
		{ 254, 204, 14 },
		{ 254, 205, 14 },
		{ 254, 206, 14 },
		{ 254, 207, 14 },
		{ 254, 208, 14 },
		{ 254, 209, 14 },
		{ 254, 210, 14 },
		{ 254, 211, 14 },
		{ 254, 212, 14 },
		{ 254, 213, 14 },
		{ 254, 214, 14 },
		{ 254, 215, 14 },
		{ 254, 216, 14 },
		{ 254, 217, 14 },
		{ 254, 218, 14 },
		{ 254, 219, 14 },
		{ 254, 220, 14 },
		{ 254, 221, 14 },
		{ 254, 222, 14 },
		{ 254, 223, 14 },
		{ 254, 224, 14 },
		{ 254, 225, 14 },
		{ 254, 226, 14 },
		{ 254, 227, 14 },
		{ 254, 228, 14 },
		{ 254, 229, 14 },
		{ 254, 230, 14 },
		{ 254, 231, 14 },
		{ 254, 232, 14 },
		{ 254, 233, 14 },
		{ 254, 234, 14 },
		{ 254, 235, 14 },
		{ 254, 236, 14 },
		{ 254, 237, 14 },
		{ 254, 238, 14 },
		{ 254, 239, 14 },
		{ 254, 240, 14 },
		{ 254, 274, 14 },
		{ 254, 275, 14 },
		{ 254, 276, 14 },
		{ 254, 277, 14 },
		{ 254, 278, 14 },
		{ 254, 279, 14 },
		{ 254, 280, 14 },
		{ 254, 281, 14 },
		{ 254, 282, 14 },
		{ 254, 283, 14 },
		{ 254, 284, 14 },
		{ 254, 285, 14 },
		{ 254, 286, 14 },
		{ 254, 287, 14 },
		{ 254, 288, 14 },
		{ 254, 289, 14 },
		{ 254, 290, 14 },
		{ 254, 291, 14 },
		{ 254, 292, 14 },
		{ 254, 293, 14 },
		{ 254, 294, 14 },
		{ 254, 295, 14 },
		{ 254, 296, 14 },
		{ 254, 297, 14 },
		{ 254, 298, 14 },
		{ 254, 299, 14 },
		{ 254, 300, 14 },
		{ 254, 301, 14 },
		{ 254, 302, 14 },
		{ 254, 303, 14 },
		{ 254, 304, 14 },
		{ 254, 305, 14 },
		{ 254, 306, 14 },
		{ 254, 307, 14 },
		{ 254, 308, 14 },
		{ 254, 309, 14 },
		{ 254, 310, 14 },
		{ 254, 311, 14 },
		{ 254, 312, 14 },
		{ 254, 313, 14 },
		{ 254, 314, 14 },
		{ 254, 315, 14 },
		{ 254, 316, 14 },
		{ 254, 317, 14 },
		{ 254, 318, 14 },
		{ 254, 319, 14 },
		{ 254, 320, 14 },
		{ 254, 321, 14 },
		{ 254, 322, 14 },
		{ 254, 323, 14 },
		{ 254, 324, 14 },
		{ 254, 325, 14 },
		{ 254, 326, 14 },
		{ 254, 327, 14 },
		{ 254, 328, 14 },
		{ 254, 329, 14 },
		{ 254, 330, 14 },
		{ 254, 331, 14 },
		{ 254, 332, 14 },
		{ 254, 333, 14 },
		{ 254, 334, 14 },
		{ 254, 335, 14 },
		{ 254, 336, 14 },
		{ 254, 337, 14 },
		{ 254, 338, 14 },
		{ 254, 339, 14 },
		{ 254, 340, 14 },
		{ 254, 341, 14 },
		{ 254, 342, 14 },
		{ 254, 343, 14 },
		{ 254, 344, 14 },
		{ 254, 345, 14 },
		{ 254, 346, 14 },
		{ 254, 347, 14 },
		{ 254, 348, 14 },
		{ 254, 349, 14 },
		{ 254, 350, 14 },
		{ 254, 351, 14 },
		{ 254, 352, 14 },
		{ 254, 353, 14 },
		{ 254, 354, 14 },
		{ 254, 355, 14 },
		{ 254, 356, 14 },
		{ 254, 357, 14 },
		{ 254, 358, 14 },
		{ 254, 359, 14 },
		{ 254, 360, 14 },
		{ 254, 361, 14 },
		{ 254, 362, 14 },
		{ 254, 363, 14 },
		{ 254, 364, 14 },
		{ 254, 365, 14 },
		{ 254, 366, 14 },
		{ 254, 367, 14 },
		{ 254, 368, 14 },
		{ 254, 369, 14 },
		{ 254, 370, 14 },
		{ 254, 371, 14 },
		{ 254, 372, 14 },
		{ 254, 373, 14 },
		{ 254, 374, 14 },
		{ 254, 375, 14 },
		{ 254, 376, 14 },
		{ 254, 377, 14 },
		{ 254, 378, 14 },
		{ 254, 379, 14 },
		{ 254, 380, 14 },
		{ 254, 381, 14 },
		{ 254, 382, 14 },
		{ 254, 383, 14 },
		{ 255, 131, 14 },
		{ 255, 132, 14 },
		{ 255, 133, 14 },
		{ 255, 134, 14 },
		{ 255, 135, 14 },
		{ 255, 136, 14 },
		{ 255, 137, 14 },
		{ 255, 138, 14 },
		{ 255, 139, 14 },
		{ 255, 140, 14 },
		{ 255, 141, 14 },
		{ 255, 142, 14 },
		{ 255, 143, 14 },
		{ 255, 144, 14 },
		{ 255, 145, 14 },
		{ 255, 146, 14 },
		{ 255, 147, 14 },
		{ 255, 148, 14 },
		{ 255, 149, 14 },
		{ 255, 150, 14 },
		{ 255, 151, 14 },
		{ 255, 152, 14 },
		{ 255, 153, 14 },
		{ 255, 154, 14 },
		{ 255, 155, 14 },
		{ 255, 156, 14 },
		{ 255, 157, 14 },
		{ 255, 158, 14 },
		{ 255, 159, 14 },
		{ 255, 160, 14 },
		{ 255, 161, 14 },
		{ 255, 162, 14 },
		{ 255, 163, 14 },
		{ 255, 164, 14 },
		{ 255, 165, 14 },
		{ 255, 166, 14 },
		{ 255, 167, 14 },
		{ 255, 168, 14 },
		{ 255, 169, 14 },
		{ 255, 170, 14 },
		{ 255, 171, 14 },
		{ 255, 172, 14 },
		{ 255, 173, 14 },
		{ 255, 174, 14 },
		{ 255, 175, 14 },
		{ 255, 176, 14 },
		{ 255, 177, 14 },
		{ 255, 178, 14 },
		{ 255, 179, 14 },
		{ 255, 180, 14 },
		{ 255, 181, 14 },
		{ 255, 182, 14 },
		{ 255, 183, 14 },
		{ 255, 184, 14 },
		{ 255, 185, 14 },
		{ 255, 186, 14 },
		{ 255, 187, 14 },
		{ 255, 188, 14 },
		{ 255, 189, 14 },
		{ 255, 190, 14 },
		{ 255, 191, 14 },
		{ 255, 192, 14 },
		{ 255, 193, 14 },
		{ 255, 194, 14 },
		{ 255, 195, 14 },
		{ 255, 196, 14 },
		{ 255, 197, 14 },
		{ 255, 198, 14 },
		{ 255, 199, 14 },
		{ 255, 200, 14 },
		{ 255, 201, 14 },
		{ 255, 202, 14 },
		{ 255, 203, 14 },
		{ 255, 204, 14 },
		{ 255, 205, 14 },
		{ 255, 206, 14 },
		{ 255, 207, 14 },
		{ 255, 208, 14 },
		{ 255, 209, 14 },
		{ 255, 210, 14 },
		{ 255, 211, 14 },
		{ 255, 212, 14 },
		{ 255, 213, 14 },
		{ 255, 214, 14 },
		{ 255, 215, 14 },
		{ 255, 216, 14 },
		{ 255, 217, 14 },
		{ 255, 218, 14 },
		{ 255, 219, 14 },
		{ 255, 220, 14 },
		{ 255, 221, 14 },
		{ 255, 222, 14 },
		{ 255, 223, 14 },
		{ 255, 224, 14 },
		{ 255, 225, 14 },
		{ 255, 226, 14 },
		{ 255, 227, 14 },
		{ 255, 228, 14 },
		{ 255, 229, 14 },
		{ 255, 230, 14 },
		{ 255, 231, 14 },
		{ 255, 232, 14 },
		{ 255, 233, 14 },
		{ 255, 234, 14 },
		{ 255, 235, 14 },
		{ 255, 236, 14 },
		{ 255, 237, 14 },
		{ 255, 238, 14 },
		{ 255, 239, 14 },
		{ 255, 240, 14 },
		{ 255, 274, 14 },
		{ 255, 275, 14 },
		{ 255, 276, 14 },
		{ 255, 277, 14 },
		{ 255, 278, 14 },
		{ 255, 279, 14 },
		{ 255, 280, 14 },
		{ 255, 281, 14 },
		{ 255, 282, 14 },
		{ 255, 283, 14 },
		{ 255, 284, 14 },
		{ 255, 285, 14 },
		{ 255, 286, 14 },
		{ 255, 287, 14 },
		{ 255, 288, 14 },
		{ 255, 289, 14 },
		{ 255, 290, 14 },
		{ 255, 291, 14 },
		{ 255, 292, 14 },
		{ 255, 293, 14 },
		{ 255, 294, 14 },
		{ 255, 295, 14 },
		{ 255, 296, 14 },
		{ 255, 297, 14 },
		{ 255, 298, 14 },
		{ 255, 299, 14 },
		{ 255, 300, 14 },
		{ 255, 301, 14 },
		{ 255, 302, 14 },
		{ 255, 303, 14 },
		{ 255, 304, 14 },
		{ 255, 305, 14 },
		{ 255, 306, 14 },
		{ 255, 307, 14 },
		{ 255, 308, 14 },
		{ 255, 309, 14 },
		{ 255, 310, 14 },
		{ 255, 311, 14 },
		{ 255, 312, 14 },
		{ 255, 313, 14 },
		{ 255, 314, 14 },
		{ 255, 315, 14 },
		{ 255, 316, 14 },
		{ 255, 317, 14 },
		{ 255, 318, 14 },
		{ 255, 319, 14 },
		{ 255, 320, 14 },
		{ 255, 321, 14 },
		{ 255, 322, 14 },
		{ 255, 323, 14 },
		{ 255, 324, 14 },
		{ 255, 325, 14 },
		{ 255, 326, 14 },
		{ 255, 327, 14 },
		{ 255, 328, 14 },
		{ 255, 329, 14 },
		{ 255, 330, 14 },
		{ 255, 331, 14 },
		{ 255, 332, 14 },
		{ 255, 333, 14 },
		{ 255, 334, 14 },
		{ 255, 335, 14 },
		{ 255, 336, 14 },
		{ 255, 337, 14 },
		{ 255, 338, 14 },
		{ 255, 339, 14 },
		{ 255, 340, 14 },
		{ 255, 341, 14 },
		{ 255, 342, 14 },
		{ 255, 343, 14 },
		{ 255, 344, 14 },
		{ 255, 345, 14 },
		{ 255, 346, 14 },
		{ 255, 347, 14 },
		{ 255, 348, 14 },
		{ 255, 349, 14 },
		{ 255, 350, 14 },
		{ 255, 351, 14 },
		{ 255, 352, 14 },
		{ 255, 353, 14 },
		{ 255, 354, 14 },
		{ 255, 355, 14 },
		{ 255, 356, 14 },
		{ 255, 357, 14 },
		{ 255, 358, 14 },
		{ 255, 359, 14 },
		{ 255, 360, 14 },
		{ 255, 361, 14 },
		{ 255, 362, 14 },
		{ 255, 363, 14 },
		{ 255, 364, 14 },
		{ 255, 365, 14 },
		{ 255, 366, 14 },
		{ 255, 367, 14 },
		{ 255, 368, 14 },
		{ 255, 369, 14 },
		{ 255, 370, 14 },
		{ 255, 371, 14 },
		{ 255, 372, 14 },
		{ 255, 373, 14 },
		{ 255, 374, 14 },
		{ 255, 375, 14 },
		{ 255, 376, 14 },
		{ 255, 377, 14 },
		{ 255, 378, 14 },
		{ 255, 379, 14 },
		{ 255, 380, 14 },
		{ 255, 381, 14 },
		{ 255, 382, 14 },
		{ 255, 383, 14 },
		{ 256, 131, 14 },
		{ 256, 132, 14 },
		{ 256, 133, 14 },
		{ 256, 134, 14 },
		{ 256, 135, 14 },
		{ 256, 136, 14 },
		{ 256, 137, 14 },
		{ 256, 138, 14 },
		{ 256, 139, 14 },
		{ 256, 140, 14 },
		{ 256, 141, 14 },
		{ 256, 142, 14 },
		{ 256, 143, 14 },
		{ 256, 144, 14 },
		{ 256, 145, 14 },
		{ 256, 146, 14 },
		{ 256, 147, 14 },
		{ 256, 148, 14 },
		{ 256, 149, 14 },
		{ 256, 150, 14 },
		{ 256, 151, 14 },
		{ 256, 152, 14 },
		{ 256, 153, 14 },
		{ 256, 154, 14 },
		{ 256, 155, 14 },
		{ 256, 156, 14 },
		{ 256, 157, 14 },
		{ 256, 158, 14 },
		{ 256, 159, 14 },
		{ 256, 160, 14 },
		{ 256, 161, 14 },
		{ 256, 162, 14 },
		{ 256, 163, 14 },
		{ 256, 164, 14 },
		{ 256, 165, 14 },
		{ 256, 166, 14 },
		{ 256, 167, 14 },
		{ 256, 168, 14 },
		{ 256, 169, 14 },
		{ 256, 170, 14 },
		{ 256, 171, 14 },
		{ 256, 172, 14 },
		{ 256, 173, 14 },
		{ 256, 174, 14 },
		{ 256, 175, 14 },
		{ 256, 176, 14 },
		{ 256, 177, 14 },
		{ 256, 178, 14 },
		{ 256, 179, 14 },
		{ 256, 180, 14 },
		{ 256, 181, 14 },
		{ 256, 182, 14 },
		{ 256, 183, 14 },
		{ 256, 184, 14 },
		{ 256, 185, 14 },
		{ 256, 186, 14 },
		{ 256, 187, 14 },
		{ 256, 188, 14 },
		{ 256, 189, 14 },
		{ 256, 190, 14 },
		{ 256, 191, 14 },
		{ 256, 192, 14 },
		{ 256, 193, 14 },
		{ 256, 194, 14 },
		{ 256, 195, 14 },
		{ 256, 196, 14 },
		{ 256, 197, 14 },
		{ 256, 198, 14 },
		{ 256, 199, 14 },
		{ 256, 200, 14 },
		{ 256, 201, 14 },
		{ 256, 202, 14 },
		{ 256, 203, 14 },
		{ 256, 204, 14 },
		{ 256, 205, 14 },
		{ 256, 206, 14 },
		{ 256, 207, 14 },
		{ 256, 208, 14 },
		{ 256, 209, 14 },
		{ 256, 210, 14 },
		{ 256, 211, 14 },
		{ 256, 212, 14 },
		{ 256, 213, 14 },
		{ 256, 214, 14 },
		{ 256, 215, 14 },
		{ 256, 216, 14 },
		{ 256, 217, 14 },
		{ 256, 218, 14 },
		{ 256, 219, 14 },
		{ 256, 220, 14 },
		{ 256, 221, 14 },
		{ 256, 222, 14 },
		{ 256, 223, 14 },
		{ 256, 224, 14 },
		{ 256, 225, 14 },
		{ 256, 226, 14 },
		{ 256, 227, 14 },
		{ 256, 228, 14 },
		{ 256, 229, 14 },
		{ 256, 230, 14 },
		{ 256, 231, 14 },
		{ 256, 232, 14 },
		{ 256, 233, 14 },
		{ 256, 234, 14 },
		{ 256, 235, 14 },
		{ 256, 236, 14 },
		{ 256, 237, 14 },
		{ 256, 238, 14 },
		{ 256, 239, 14 },
		{ 256, 240, 14 },
		{ 256, 274, 14 },
		{ 256, 275, 14 },
		{ 256, 276, 14 },
		{ 256, 277, 14 },
		{ 256, 278, 14 },
		{ 256, 279, 14 },
		{ 256, 280, 14 },
		{ 256, 281, 14 },
		{ 256, 282, 14 },
		{ 256, 283, 14 },
		{ 256, 284, 14 },
		{ 256, 285, 14 },
		{ 256, 286, 14 },
		{ 256, 287, 14 },
		{ 256, 288, 14 },
		{ 256, 289, 14 },
		{ 256, 290, 14 },
		{ 256, 291, 14 },
		{ 256, 292, 14 },
		{ 256, 293, 14 },
		{ 256, 294, 14 },
		{ 256, 295, 14 },
		{ 256, 296, 14 },
		{ 256, 297, 14 },
		{ 256, 298, 14 },
		{ 256, 299, 14 },
		{ 256, 300, 14 },
		{ 256, 301, 14 },
		{ 256, 302, 14 },
		{ 256, 303, 14 },
		{ 256, 304, 14 },
		{ 256, 305, 14 },
		{ 256, 306, 14 },
		{ 256, 307, 14 },
		{ 256, 308, 14 },
		{ 256, 309, 14 },
		{ 256, 310, 14 },
		{ 256, 311, 14 },
		{ 256, 312, 14 },
		{ 256, 313, 14 },
		{ 256, 314, 14 },
		{ 256, 315, 14 },
		{ 256, 316, 14 },
		{ 256, 317, 14 },
		{ 256, 318, 14 },
		{ 256, 319, 14 },
		{ 256, 320, 14 },
		{ 256, 321, 14 },
		{ 256, 322, 14 },
		{ 256, 323, 14 },
		{ 256, 324, 14 },
		{ 256, 325, 14 },
		{ 256, 326, 14 },
		{ 256, 327, 14 },
		{ 256, 328, 14 },
		{ 256, 329, 14 },
		{ 256, 330, 14 },
		{ 256, 331, 14 },
		{ 256, 332, 14 },
		{ 256, 333, 14 },
		{ 256, 334, 14 },
		{ 256, 335, 14 },
		{ 256, 336, 14 },
		{ 256, 337, 14 },
		{ 256, 338, 14 },
		{ 256, 339, 14 },
		{ 256, 340, 14 },
		{ 256, 341, 14 },
		{ 256, 342, 14 },
		{ 256, 343, 14 },
		{ 256, 344, 14 },
		{ 256, 345, 14 },
		{ 256, 346, 14 },
		{ 256, 347, 14 },
		{ 256, 348, 14 },
		{ 256, 349, 14 },
		{ 256, 350, 14 },
		{ 256, 351, 14 },
		{ 256, 352, 14 },
		{ 256, 353, 14 },
		{ 256, 354, 14 },
		{ 256, 355, 14 },
		{ 256, 356, 14 },
		{ 256, 357, 14 },
		{ 256, 358, 14 },
		{ 256, 359, 14 },
		{ 256, 360, 14 },
		{ 256, 361, 14 },
		{ 256, 362, 14 },
		{ 256, 363, 14 },
		{ 256, 364, 14 },
		{ 256, 365, 14 },
		{ 256, 366, 14 },
		{ 256, 367, 14 },
		{ 256, 368, 14 },
		{ 256, 369, 14 },
		{ 256, 370, 14 },
		{ 256, 371, 14 },
		{ 256, 372, 14 },
		{ 256, 373, 14 },
		{ 256, 374, 14 },
		{ 256, 375, 14 },
		{ 256, 376, 14 },
		{ 256, 377, 14 },
		{ 256, 378, 14 },
		{ 256, 379, 14 },
		{ 256, 380, 14 },
		{ 256, 381, 14 },
		{ 256, 382, 14 },
		{ 256, 383, 14 },
		{ 257, 131, 14 },
		{ 257, 132, 14 },
		{ 257, 133, 14 },
		{ 257, 134, 14 },
		{ 257, 135, 14 },
		{ 257, 136, 14 },
		{ 257, 137, 14 },
		{ 257, 138, 14 },
		{ 257, 139, 14 },
		{ 257, 140, 14 },
		{ 257, 141, 14 },
		{ 257, 142, 14 },
		{ 257, 143, 14 },
		{ 257, 144, 14 },
		{ 257, 145, 14 },
		{ 257, 146, 14 },
		{ 257, 147, 14 },
		{ 257, 148, 14 },
		{ 257, 149, 14 },
		{ 257, 150, 14 },
		{ 257, 151, 14 },
		{ 257, 152, 14 },
		{ 257, 153, 14 },
		{ 257, 154, 14 },
		{ 257, 155, 14 },
		{ 257, 156, 14 },
		{ 257, 157, 14 },
		{ 257, 158, 14 },
		{ 257, 159, 14 },
		{ 257, 160, 14 },
		{ 257, 161, 14 },
		{ 257, 162, 14 },
		{ 257, 163, 14 },
		{ 257, 164, 14 },
		{ 257, 165, 14 },
		{ 257, 166, 14 },
		{ 257, 167, 14 },
		{ 257, 168, 14 },
		{ 257, 169, 14 },
		{ 257, 170, 14 },
		{ 257, 171, 14 },
		{ 257, 172, 14 },
		{ 257, 173, 14 },
		{ 257, 174, 14 },
		{ 257, 175, 14 },
		{ 257, 176, 14 },
		{ 257, 177, 14 },
		{ 257, 178, 14 },
		{ 257, 179, 14 },
		{ 257, 180, 14 },
		{ 257, 181, 14 },
		{ 257, 182, 14 },
		{ 257, 183, 14 },
		{ 257, 184, 14 },
		{ 257, 185, 14 },
		{ 257, 186, 14 },
		{ 257, 187, 14 },
		{ 257, 188, 14 },
		{ 257, 189, 14 },
		{ 257, 190, 14 },
		{ 257, 191, 14 },
		{ 257, 192, 14 },
		{ 257, 193, 14 },
		{ 257, 194, 14 },
		{ 257, 195, 14 },
		{ 257, 196, 14 },
		{ 257, 197, 14 },
		{ 257, 198, 14 },
		{ 257, 199, 14 },
		{ 257, 200, 14 },
		{ 257, 201, 14 },
		{ 257, 202, 14 },
		{ 257, 203, 14 },
		{ 257, 204, 14 },
		{ 257, 205, 14 },
		{ 257, 206, 14 },
		{ 257, 207, 14 },
		{ 257, 208, 14 },
		{ 257, 209, 14 },
		{ 257, 210, 14 },
		{ 257, 211, 14 },
		{ 257, 212, 14 },
		{ 257, 213, 14 },
		{ 257, 214, 14 },
		{ 257, 215, 14 },
		{ 257, 216, 14 },
		{ 257, 217, 14 },
		{ 257, 218, 14 },
		{ 257, 219, 14 },
		{ 257, 220, 14 },
		{ 257, 221, 14 },
		{ 257, 222, 14 },
		{ 257, 223, 14 },
		{ 257, 224, 14 },
		{ 257, 225, 14 },
		{ 257, 226, 14 },
		{ 257, 227, 14 },
		{ 257, 228, 14 },
		{ 257, 229, 14 },
		{ 257, 230, 14 },
		{ 257, 231, 14 },
		{ 257, 232, 14 },
		{ 257, 233, 14 },
		{ 257, 234, 14 },
		{ 257, 235, 14 },
		{ 257, 236, 14 },
		{ 257, 237, 14 },
		{ 257, 238, 14 },
		{ 257, 239, 14 },
		{ 257, 240, 14 },
		{ 257, 274, 14 },
		{ 257, 275, 14 },
		{ 257, 276, 14 },
		{ 257, 277, 14 },
		{ 257, 278, 14 },
		{ 257, 279, 14 },
		{ 257, 280, 14 },
		{ 257, 281, 14 },
		{ 257, 282, 14 },
		{ 257, 283, 14 },
		{ 257, 284, 14 },
		{ 257, 285, 14 },
		{ 257, 286, 14 },
		{ 257, 287, 14 },
		{ 257, 288, 14 },
		{ 257, 289, 14 },
		{ 257, 290, 14 },
		{ 257, 291, 14 },
		{ 257, 292, 14 },
		{ 257, 293, 14 },
		{ 257, 294, 14 },
		{ 257, 295, 14 },
		{ 257, 296, 14 },
		{ 257, 297, 14 },
		{ 257, 298, 14 },
		{ 257, 299, 14 },
		{ 257, 300, 14 },
		{ 257, 301, 14 },
		{ 257, 302, 14 },
		{ 257, 303, 14 },
		{ 257, 304, 14 },
		{ 257, 305, 14 },
		{ 257, 306, 14 },
		{ 257, 307, 14 },
		{ 257, 308, 14 },
		{ 257, 309, 14 },
		{ 257, 310, 14 },
		{ 257, 311, 14 },
		{ 257, 312, 14 },
		{ 257, 313, 14 },
		{ 257, 314, 14 },
		{ 257, 315, 14 },
		{ 257, 316, 14 },
		{ 257, 317, 14 },
		{ 257, 318, 14 },
		{ 257, 319, 14 },
		{ 257, 320, 14 },
		{ 257, 321, 14 },
		{ 257, 322, 14 },
		{ 257, 323, 14 },
		{ 257, 324, 14 },
		{ 257, 325, 14 },
		{ 257, 326, 14 },
		{ 257, 327, 14 },
		{ 257, 328, 14 },
		{ 257, 329, 14 },
		{ 257, 330, 14 },
		{ 257, 331, 14 },
		{ 257, 332, 14 },
		{ 257, 333, 14 },
		{ 257, 334, 14 },
		{ 257, 335, 14 },
		{ 257, 336, 14 },
		{ 257, 337, 14 },
		{ 257, 338, 14 },
		{ 257, 339, 14 },
		{ 257, 340, 14 },
		{ 257, 341, 14 },
		{ 257, 342, 14 },
		{ 257, 343, 14 },
		{ 257, 344, 14 },
		{ 257, 345, 14 },
		{ 257, 346, 14 },
		{ 257, 347, 14 },
		{ 257, 348, 14 },
		{ 257, 349, 14 },
		{ 257, 350, 14 },
		{ 257, 351, 14 },
		{ 257, 352, 14 },
		{ 257, 353, 14 },
		{ 257, 354, 14 },
		{ 257, 355, 14 },
		{ 257, 356, 14 },
		{ 257, 357, 14 },
		{ 257, 358, 14 },
		{ 257, 359, 14 },
		{ 257, 360, 14 },
		{ 257, 361, 14 },
		{ 257, 362, 14 },
		{ 257, 363, 14 },
		{ 257, 364, 14 },
		{ 257, 365, 14 },
		{ 257, 366, 14 },
		{ 257, 367, 14 },
		{ 257, 368, 14 },
		{ 257, 369, 14 },
		{ 257, 370, 14 },
		{ 257, 371, 14 },
		{ 257, 372, 14 },
		{ 257, 373, 14 },
		{ 257, 374, 14 },
		{ 257, 375, 14 },
		{ 257, 376, 14 },
		{ 257, 377, 14 },
		{ 257, 378, 14 },
		{ 257, 379, 14 },
		{ 257, 380, 14 },
		{ 257, 381, 14 },
		{ 257, 382, 14 },
		{ 257, 383, 14 },
		{ 258, 131, 14 },
		{ 258, 132, 14 },
		{ 258, 133, 14 },
		{ 258, 134, 14 },
		{ 258, 135, 14 },
		{ 258, 136, 14 },
		{ 258, 137, 14 },
		{ 258, 138, 14 },
		{ 258, 139, 14 },
		{ 258, 140, 14 },
		{ 258, 141, 14 },
		{ 258, 142, 14 },
		{ 258, 143, 14 },
		{ 258, 144, 14 },
		{ 258, 145, 14 },
		{ 258, 146, 14 },
		{ 258, 147, 14 },
		{ 258, 148, 14 },
		{ 258, 149, 14 },
		{ 258, 150, 14 },
		{ 258, 151, 14 },
		{ 258, 152, 14 },
		{ 258, 153, 14 },
		{ 258, 154, 14 },
		{ 258, 155, 14 },
		{ 258, 156, 14 },
		{ 258, 157, 14 },
		{ 258, 158, 14 },
		{ 258, 159, 14 },
		{ 258, 160, 14 },
		{ 258, 161, 14 },
		{ 258, 162, 14 },
		{ 258, 163, 14 },
		{ 258, 164, 14 },
		{ 258, 165, 14 },
		{ 258, 166, 14 },
		{ 258, 167, 14 },
		{ 258, 168, 14 },
		{ 258, 169, 14 },
		{ 258, 170, 14 },
		{ 258, 171, 14 },
		{ 258, 172, 14 },
		{ 258, 173, 14 },
		{ 258, 174, 14 },
		{ 258, 175, 14 },
		{ 258, 176, 14 },
		{ 258, 177, 14 },
		{ 258, 178, 14 },
		{ 258, 179, 14 },
		{ 258, 180, 14 },
		{ 258, 181, 14 },
		{ 258, 182, 14 },
		{ 258, 183, 14 },
		{ 258, 184, 14 },
		{ 258, 185, 14 },
		{ 258, 186, 14 },
		{ 258, 187, 14 },
		{ 258, 188, 14 },
		{ 258, 189, 14 },
		{ 258, 190, 14 },
		{ 258, 191, 14 },
		{ 258, 192, 14 },
		{ 258, 193, 14 },
		{ 258, 194, 14 },
		{ 258, 195, 14 },
		{ 258, 196, 14 },
		{ 258, 197, 14 },
		{ 258, 198, 14 },
		{ 258, 199, 14 },
		{ 258, 200, 14 },
		{ 258, 201, 14 },
		{ 258, 202, 14 },
		{ 258, 203, 14 },
		{ 258, 204, 14 },
		{ 258, 205, 14 },
		{ 258, 206, 14 },
		{ 258, 207, 14 },
		{ 258, 208, 14 },
		{ 258, 209, 14 },
		{ 258, 210, 14 },
		{ 258, 211, 14 },
		{ 258, 212, 14 },
		{ 258, 213, 14 },
		{ 258, 214, 14 },
		{ 258, 215, 14 },
		{ 258, 216, 14 },
		{ 258, 217, 14 },
		{ 258, 218, 14 },
		{ 258, 219, 14 },
		{ 258, 220, 14 },
		{ 258, 221, 14 },
		{ 258, 222, 14 },
		{ 258, 223, 14 },
		{ 258, 224, 14 },
		{ 258, 225, 14 },
		{ 258, 226, 14 },
		{ 258, 227, 14 },
		{ 258, 228, 14 },
		{ 258, 229, 14 },
		{ 258, 230, 14 },
		{ 258, 231, 14 },
		{ 258, 232, 14 },
		{ 258, 233, 14 },
		{ 258, 234, 14 },
		{ 258, 235, 14 },
		{ 258, 236, 14 },
		{ 258, 237, 14 },
		{ 258, 238, 14 },
		{ 258, 239, 14 },
		{ 258, 240, 14 },
		{ 258, 274, 14 },
		{ 258, 275, 14 },
		{ 258, 276, 14 },
		{ 258, 277, 14 },
		{ 258, 278, 14 },
		{ 258, 279, 14 },
		{ 258, 280, 14 },
		{ 258, 281, 14 },
		{ 258, 282, 14 },
		{ 258, 283, 14 },
		{ 258, 284, 14 },
		{ 258, 285, 14 },
		{ 258, 286, 14 },
		{ 258, 287, 14 },
		{ 258, 288, 14 },
		{ 258, 289, 14 },
		{ 258, 290, 14 },
		{ 258, 291, 14 },
		{ 258, 292, 14 },
		{ 258, 293, 14 },
		{ 258, 294, 14 },
		{ 258, 295, 14 },
		{ 258, 296, 14 },
		{ 258, 297, 14 },
		{ 258, 298, 14 },
		{ 258, 299, 14 },
		{ 258, 300, 14 },
		{ 258, 301, 14 },
		{ 258, 302, 14 },
		{ 258, 303, 14 },
		{ 258, 304, 14 },
		{ 258, 305, 14 },
		{ 258, 306, 14 },
		{ 258, 307, 14 },
		{ 258, 308, 14 },
		{ 258, 309, 14 },
		{ 258, 310, 14 },
		{ 258, 311, 14 },
		{ 258, 312, 14 },
		{ 258, 313, 14 },
		{ 258, 314, 14 },
		{ 258, 315, 14 },
		{ 258, 316, 14 },
		{ 258, 317, 14 },
		{ 258, 318, 14 },
		{ 258, 319, 14 },
		{ 258, 320, 14 },
		{ 258, 321, 14 },
		{ 258, 322, 14 },
		{ 258, 323, 14 },
		{ 258, 324, 14 },
		{ 258, 325, 14 },
		{ 258, 326, 14 },
		{ 258, 327, 14 },
		{ 258, 328, 14 },
		{ 258, 329, 14 },
		{ 258, 330, 14 },
		{ 258, 331, 14 },
		{ 258, 332, 14 },
		{ 258, 333, 14 },
		{ 258, 334, 14 },
		{ 258, 335, 14 },
		{ 258, 336, 14 },
		{ 258, 337, 14 },
		{ 258, 338, 14 },
		{ 258, 339, 14 },
		{ 258, 340, 14 },
		{ 258, 341, 14 },
		{ 258, 342, 14 },
		{ 258, 343, 14 },
		{ 258, 344, 14 },
		{ 258, 345, 14 },
		{ 258, 346, 14 },
		{ 258, 347, 14 },
		{ 258, 348, 14 },
		{ 258, 349, 14 },
		{ 258, 350, 14 },
		{ 258, 351, 14 },
		{ 258, 352, 14 },
		{ 258, 353, 14 },
		{ 258, 354, 14 },
		{ 258, 355, 14 },
		{ 258, 356, 14 },
		{ 258, 357, 14 },
		{ 258, 358, 14 },
		{ 258, 359, 14 },
		{ 258, 360, 14 },
		{ 258, 361, 14 },
		{ 258, 362, 14 },
		{ 258, 363, 14 },
		{ 258, 364, 14 },
		{ 258, 365, 14 },
		{ 258, 366, 14 },
		{ 258, 367, 14 },
		{ 258, 368, 14 },
		{ 258, 369, 14 },
		{ 258, 370, 14 },
		{ 258, 371, 14 },
		{ 258, 372, 14 },
		{ 258, 373, 14 },
		{ 258, 374, 14 },
		{ 258, 375, 14 },
		{ 258, 376, 14 },
		{ 258, 377, 14 },
		{ 258, 378, 14 },
		{ 258, 379, 14 },
		{ 258, 380, 14 },
		{ 258, 381, 14 },
		{ 258, 382, 14 },
		{ 258, 383, 14 },
		{ 259, 131, 14 },
		{ 259, 132, 14 },
		{ 259, 133, 14 },
		{ 259, 134, 14 },
		{ 259, 135, 14 },
		{ 259, 136, 14 },
		{ 259, 137, 14 },
		{ 259, 138, 14 },
		{ 259, 139, 14 },
		{ 259, 140, 14 },
		{ 259, 141, 14 },
		{ 259, 142, 14 },
		{ 259, 143, 14 },
		{ 259, 144, 14 },
		{ 259, 145, 14 },
		{ 259, 146, 14 },
		{ 259, 147, 14 },
		{ 259, 148, 14 },
		{ 259, 149, 14 },
		{ 259, 150, 14 },
		{ 259, 151, 14 },
		{ 259, 152, 14 },
		{ 259, 153, 14 },
		{ 259, 154, 14 },
		{ 259, 155, 14 },
		{ 259, 156, 14 },
		{ 259, 157, 14 },
		{ 259, 158, 14 },
		{ 259, 159, 14 },
		{ 259, 160, 14 },
		{ 259, 161, 14 },
		{ 259, 162, 14 },
		{ 259, 163, 14 },
		{ 259, 164, 14 },
		{ 259, 165, 14 },
		{ 259, 166, 14 },
		{ 259, 167, 14 },
		{ 259, 168, 14 },
		{ 259, 169, 14 },
		{ 259, 170, 14 },
		{ 259, 171, 14 },
		{ 259, 172, 14 },
		{ 259, 173, 14 },
		{ 259, 174, 14 },
		{ 259, 175, 14 },
		{ 259, 176, 14 },
		{ 259, 177, 14 },
		{ 259, 178, 14 },
		{ 259, 179, 14 },
		{ 259, 180, 14 },
		{ 259, 181, 14 },
		{ 259, 182, 14 },
		{ 259, 183, 14 },
		{ 259, 184, 14 },
		{ 259, 185, 14 },
		{ 259, 186, 14 },
		{ 259, 187, 14 },
		{ 259, 188, 14 },
		{ 259, 189, 14 },
		{ 259, 190, 14 },
		{ 259, 191, 14 },
		{ 259, 192, 14 },
		{ 259, 193, 14 },
		{ 259, 194, 14 },
		{ 259, 195, 14 },
		{ 259, 196, 14 },
		{ 259, 197, 14 },
		{ 259, 198, 14 },
		{ 259, 199, 14 },
		{ 259, 200, 14 },
		{ 259, 201, 14 },
		{ 259, 202, 14 },
		{ 259, 203, 14 },
		{ 259, 204, 14 },
		{ 259, 205, 14 },
		{ 259, 206, 14 },
		{ 259, 207, 14 },
		{ 259, 208, 14 },
		{ 259, 209, 14 },
		{ 259, 210, 14 },
		{ 259, 211, 14 },
		{ 259, 212, 14 },
		{ 259, 213, 14 },
		{ 259, 214, 14 },
		{ 259, 215, 14 },
		{ 259, 216, 14 },
		{ 259, 217, 14 },
		{ 259, 218, 14 },
		{ 259, 219, 14 },
		{ 259, 220, 14 },
		{ 259, 221, 14 },
		{ 259, 222, 14 },
		{ 259, 223, 14 },
		{ 259, 224, 14 },
		{ 259, 225, 14 },
		{ 259, 226, 14 },
		{ 259, 227, 14 },
		{ 259, 228, 14 },
		{ 259, 229, 14 },
		{ 259, 230, 14 },
		{ 259, 231, 14 },
		{ 259, 232, 14 },
		{ 259, 233, 14 },
		{ 259, 234, 14 },
		{ 259, 235, 14 },
		{ 259, 236, 14 },
		{ 259, 237, 14 },
		{ 259, 238, 14 },
		{ 259, 239, 14 },
		{ 259, 240, 14 },
		{ 259, 274, 14 },
		{ 259, 275, 14 },
		{ 259, 276, 14 },
		{ 259, 277, 14 },
		{ 259, 278, 14 },
		{ 259, 279, 14 },
		{ 259, 280, 14 },
		{ 259, 281, 14 },
		{ 259, 282, 14 },
		{ 259, 283, 14 },
		{ 259, 284, 14 },
		{ 259, 285, 14 },
		{ 259, 286, 14 },
		{ 259, 287, 14 },
		{ 259, 288, 14 },
		{ 259, 289, 14 },
		{ 259, 290, 14 },
		{ 259, 291, 14 },
		{ 259, 292, 14 },
		{ 259, 293, 14 },
		{ 259, 294, 14 },
		{ 259, 295, 14 },
		{ 259, 296, 14 },
		{ 259, 297, 14 },
		{ 259, 298, 14 },
		{ 259, 299, 14 },
		{ 259, 300, 14 },
		{ 259, 301, 14 },
		{ 259, 302, 14 },
		{ 259, 303, 14 },
		{ 259, 304, 14 },
		{ 259, 305, 14 },
		{ 259, 306, 14 },
		{ 259, 307, 14 },
		{ 259, 308, 14 },
		{ 259, 309, 14 },
		{ 259, 310, 14 },
		{ 259, 311, 14 },
		{ 259, 312, 14 },
		{ 259, 313, 14 },
		{ 259, 314, 14 },
		{ 259, 315, 14 },
		{ 259, 316, 14 },
		{ 259, 317, 14 },
		{ 259, 318, 14 },
		{ 259, 319, 14 },
		{ 259, 320, 14 },
		{ 259, 321, 14 },
		{ 259, 322, 14 },
		{ 259, 323, 14 },
		{ 259, 324, 14 },
		{ 259, 325, 14 },
		{ 259, 326, 14 },
		{ 259, 327, 14 },
		{ 259, 328, 14 },
		{ 259, 329, 14 },
		{ 259, 330, 14 },
		{ 259, 331, 14 },
		{ 259, 332, 14 },
		{ 259, 333, 14 },
		{ 259, 334, 14 },
		{ 259, 335, 14 },
		{ 259, 336, 14 },
		{ 259, 337, 14 },
		{ 259, 338, 14 },
		{ 259, 339, 14 },
		{ 259, 340, 14 },
		{ 259, 341, 14 },
		{ 259, 342, 14 },
		{ 259, 343, 14 },
		{ 259, 344, 14 },
		{ 259, 345, 14 },
		{ 259, 346, 14 },
		{ 259, 347, 14 },
		{ 259, 348, 14 },
		{ 259, 349, 14 },
		{ 259, 350, 14 },
		{ 259, 351, 14 },
		{ 259, 352, 14 },
		{ 259, 353, 14 },
		{ 259, 354, 14 },
		{ 259, 355, 14 },
		{ 259, 356, 14 },
		{ 259, 357, 14 },
		{ 259, 358, 14 },
		{ 259, 359, 14 },
		{ 259, 360, 14 },
		{ 259, 361, 14 },
		{ 259, 362, 14 },
		{ 259, 363, 14 },
		{ 259, 364, 14 },
		{ 259, 365, 14 },
		{ 259, 366, 14 },
		{ 259, 367, 14 },
		{ 259, 368, 14 },
		{ 259, 369, 14 },
		{ 259, 370, 14 },
		{ 259, 371, 14 },
		{ 259, 372, 14 },
		{ 259, 373, 14 },
		{ 259, 374, 14 },
		{ 259, 375, 14 },
		{ 259, 376, 14 },
		{ 259, 377, 14 },
		{ 259, 378, 14 },
		{ 259, 379, 14 },
		{ 259, 380, 14 },
		{ 259, 381, 14 },
		{ 259, 382, 14 },
		{ 259, 383, 14 },
		{ 260, 131, 14 },
		{ 260, 132, 14 },
		{ 260, 133, 14 },
		{ 260, 134, 14 },
		{ 260, 135, 14 },
		{ 260, 136, 14 },
		{ 260, 137, 14 },
		{ 260, 138, 14 },
		{ 260, 139, 14 },
		{ 260, 140, 14 },
		{ 260, 141, 14 },
		{ 260, 142, 14 },
		{ 260, 143, 14 },
		{ 260, 144, 14 },
		{ 260, 145, 14 },
		{ 260, 146, 14 },
		{ 260, 147, 14 },
		{ 260, 148, 14 },
		{ 260, 149, 14 },
		{ 260, 150, 14 },
		{ 260, 151, 14 },
		{ 260, 152, 14 },
		{ 260, 153, 14 },
		{ 260, 154, 14 },
		{ 260, 155, 14 },
		{ 260, 156, 14 },
		{ 260, 157, 14 },
		{ 260, 158, 14 },
		{ 260, 159, 14 },
		{ 260, 160, 14 },
		{ 260, 161, 14 },
		{ 260, 162, 14 },
		{ 260, 163, 14 },
		{ 260, 164, 14 },
		{ 260, 165, 14 },
		{ 260, 166, 14 },
		{ 260, 167, 14 },
		{ 260, 168, 14 },
		{ 260, 169, 14 },
		{ 260, 170, 14 },
		{ 260, 171, 14 },
		{ 260, 172, 14 },
		{ 260, 173, 14 },
		{ 260, 174, 14 },
		{ 260, 175, 14 },
		{ 260, 176, 14 },
		{ 260, 177, 14 },
		{ 260, 178, 14 },
		{ 260, 179, 14 },
		{ 260, 180, 14 },
		{ 260, 181, 14 },
		{ 260, 182, 14 },
		{ 260, 183, 14 },
		{ 260, 184, 14 },
		{ 260, 185, 14 },
		{ 260, 186, 14 },
		{ 260, 187, 14 },
		{ 260, 188, 14 },
		{ 260, 189, 14 },
		{ 260, 190, 14 },
		{ 260, 191, 14 },
		{ 260, 192, 14 },
		{ 260, 193, 14 },
		{ 260, 194, 14 },
		{ 260, 195, 14 },
		{ 260, 196, 14 },
		{ 260, 197, 14 },
		{ 260, 198, 14 },
		{ 260, 199, 14 },
		{ 260, 200, 14 },
		{ 260, 201, 14 },
		{ 260, 202, 14 },
		{ 260, 203, 14 },
		{ 260, 204, 14 },
		{ 260, 205, 14 },
		{ 260, 206, 14 },
		{ 260, 207, 14 },
		{ 260, 208, 14 },
		{ 260, 209, 14 },
		{ 260, 210, 14 },
		{ 260, 211, 14 },
		{ 260, 212, 14 },
		{ 260, 213, 14 },
		{ 260, 214, 14 },
		{ 260, 215, 14 },
		{ 260, 216, 14 },
		{ 260, 217, 14 },
		{ 260, 218, 14 },
		{ 260, 219, 14 },
		{ 260, 220, 14 },
		{ 260, 221, 14 },
		{ 260, 222, 14 },
		{ 260, 223, 14 },
		{ 260, 224, 14 },
		{ 260, 225, 14 },
		{ 260, 226, 14 },
		{ 260, 227, 14 },
		{ 260, 228, 14 },
		{ 260, 229, 14 },
		{ 260, 230, 14 },
		{ 260, 231, 14 },
		{ 260, 232, 14 },
		{ 260, 233, 14 },
		{ 260, 234, 14 },
		{ 260, 235, 14 },
		{ 260, 236, 14 },
		{ 260, 237, 14 },
		{ 260, 238, 14 },
		{ 260, 239, 14 },
		{ 260, 240, 14 },
		{ 260, 274, 14 },
		{ 260, 275, 14 },
		{ 260, 276, 14 },
		{ 260, 277, 14 },
		{ 260, 278, 14 },
		{ 260, 279, 14 },
		{ 260, 280, 14 },
		{ 260, 281, 14 },
		{ 260, 282, 14 },
		{ 260, 283, 14 },
		{ 260, 284, 14 },
		{ 260, 285, 14 },
		{ 260, 286, 14 },
		{ 260, 287, 14 },
		{ 260, 288, 14 },
		{ 260, 289, 14 },
		{ 260, 290, 14 },
		{ 260, 291, 14 },
		{ 260, 292, 14 },
		{ 260, 293, 14 },
		{ 260, 294, 14 },
		{ 260, 295, 14 },
		{ 260, 296, 14 },
		{ 260, 297, 14 },
		{ 260, 298, 14 },
		{ 260, 299, 14 },
		{ 260, 300, 14 },
		{ 260, 301, 14 },
		{ 260, 302, 14 },
		{ 260, 303, 14 },
		{ 260, 304, 14 },
		{ 260, 305, 14 },
		{ 260, 306, 14 },
		{ 260, 307, 14 },
		{ 260, 308, 14 },
		{ 260, 309, 14 },
		{ 260, 310, 14 },
		{ 260, 311, 14 },
		{ 260, 312, 14 },
		{ 260, 313, 14 },
		{ 260, 314, 14 },
		{ 260, 315, 14 },
		{ 260, 316, 14 },
		{ 260, 317, 14 },
		{ 260, 318, 14 },
		{ 260, 319, 14 },
		{ 260, 320, 14 },
		{ 260, 321, 14 },
		{ 260, 322, 14 },
		{ 260, 323, 14 },
		{ 260, 324, 14 },
		{ 260, 325, 14 },
		{ 260, 326, 14 },
		{ 260, 327, 14 },
		{ 260, 328, 14 },
		{ 260, 329, 14 },
		{ 260, 330, 14 },
		{ 260, 331, 14 },
		{ 260, 332, 14 },
		{ 260, 333, 14 },
		{ 260, 334, 14 },
		{ 260, 335, 14 },
		{ 260, 336, 14 },
		{ 260, 337, 14 },
		{ 260, 338, 14 },
		{ 260, 339, 14 },
		{ 260, 340, 14 },
		{ 260, 341, 14 },
		{ 260, 342, 14 },
		{ 260, 343, 14 },
		{ 260, 344, 14 },
		{ 260, 345, 14 },
		{ 260, 346, 14 },
		{ 260, 347, 14 },
		{ 260, 348, 14 },
		{ 260, 349, 14 },
		{ 260, 350, 14 },
		{ 260, 351, 14 },
		{ 260, 352, 14 },
		{ 260, 353, 14 },
		{ 260, 354, 14 },
		{ 260, 355, 14 },
		{ 260, 356, 14 },
		{ 260, 357, 14 },
		{ 260, 358, 14 },
		{ 260, 359, 14 },
		{ 260, 360, 14 },
		{ 260, 361, 14 },
		{ 260, 362, 14 },
		{ 260, 363, 14 },
		{ 260, 364, 14 },
		{ 260, 365, 14 },
		{ 260, 366, 14 },
		{ 260, 367, 14 },
		{ 260, 368, 14 },
		{ 260, 369, 14 },
		{ 260, 370, 14 },
		{ 260, 371, 14 },
		{ 260, 372, 14 },
		{ 260, 373, 14 },
		{ 260, 374, 14 },
		{ 260, 375, 14 },
		{ 260, 376, 14 },
		{ 260, 377, 14 },
		{ 260, 378, 14 },
		{ 260, 379, 14 },
		{ 260, 380, 14 },
		{ 260, 381, 14 },
		{ 260, 382, 14 },
		{ 260, 383, 14 },
		{ 261, 131, 14 },
		{ 261, 132, 14 },
		{ 261, 133, 14 },
		{ 261, 134, 14 },
		{ 261, 135, 14 },
		{ 261, 136, 14 },
		{ 261, 137, 14 },
		{ 261, 138, 14 },
		{ 261, 139, 14 },
		{ 261, 140, 14 },
		{ 261, 141, 14 },
		{ 261, 142, 14 },
		{ 261, 143, 14 },
		{ 261, 144, 14 },
		{ 261, 145, 14 },
		{ 261, 146, 14 },
		{ 261, 147, 14 },
		{ 261, 148, 14 },
		{ 261, 149, 14 },
		{ 261, 150, 14 },
		{ 261, 151, 14 },
		{ 261, 152, 14 },
		{ 261, 153, 14 },
		{ 261, 154, 14 },
		{ 261, 155, 14 },
		{ 261, 156, 14 },
		{ 261, 157, 14 },
		{ 261, 158, 14 },
		{ 261, 159, 14 },
		{ 261, 160, 14 },
		{ 261, 161, 14 },
		{ 261, 162, 14 },
		{ 261, 163, 14 },
		{ 261, 164, 14 },
		{ 261, 165, 14 },
		{ 261, 166, 14 },
		{ 261, 167, 14 },
		{ 261, 168, 14 },
		{ 261, 169, 14 },
		{ 261, 170, 14 },
		{ 261, 171, 14 },
		{ 261, 172, 14 },
		{ 261, 173, 14 },
		{ 261, 174, 14 },
		{ 261, 175, 14 },
		{ 261, 176, 14 },
		{ 261, 177, 14 },
		{ 261, 178, 14 },
		{ 261, 179, 14 },
		{ 261, 180, 14 },
		{ 261, 181, 14 },
		{ 261, 182, 14 },
		{ 261, 183, 14 },
		{ 261, 184, 14 },
		{ 261, 185, 14 },
		{ 261, 186, 14 },
		{ 261, 187, 14 },
		{ 261, 188, 14 },
		{ 261, 189, 14 },
		{ 261, 190, 14 },
		{ 261, 191, 14 },
		{ 261, 192, 14 },
		{ 261, 193, 14 },
		{ 261, 194, 14 },
		{ 261, 195, 14 },
		{ 261, 196, 14 },
		{ 261, 197, 14 },
		{ 261, 198, 14 },
		{ 261, 199, 14 },
		{ 261, 200, 14 },
		{ 261, 201, 14 },
		{ 261, 202, 14 },
		{ 261, 203, 14 },
		{ 261, 204, 14 },
		{ 261, 205, 14 },
		{ 261, 206, 14 },
		{ 261, 207, 14 },
		{ 261, 208, 14 },
		{ 261, 209, 14 },
		{ 261, 210, 14 },
		{ 261, 211, 14 },
		{ 261, 212, 14 },
		{ 261, 213, 14 },
		{ 261, 214, 14 },
		{ 261, 215, 14 },
		{ 261, 216, 14 },
		{ 261, 217, 14 },
		{ 261, 218, 14 },
		{ 261, 219, 14 },
		{ 261, 220, 14 },
		{ 261, 221, 14 },
		{ 261, 222, 14 },
		{ 261, 223, 14 },
		{ 261, 224, 14 },
		{ 261, 225, 14 },
		{ 261, 226, 14 },
		{ 261, 227, 14 },
		{ 261, 228, 14 },
		{ 261, 229, 14 },
		{ 261, 230, 14 },
		{ 261, 231, 14 },
		{ 261, 232, 14 },
		{ 261, 233, 14 },
		{ 261, 234, 14 },
		{ 261, 235, 14 },
		{ 261, 236, 14 },
		{ 261, 237, 14 },
		{ 261, 238, 14 },
		{ 261, 239, 14 },
		{ 261, 240, 14 },
		{ 261, 274, 14 },
		{ 261, 275, 14 },
		{ 261, 276, 14 },
		{ 261, 277, 14 },
		{ 261, 278, 14 },
		{ 261, 279, 14 },
		{ 261, 280, 14 },
		{ 261, 281, 14 },
		{ 261, 282, 14 },
		{ 261, 283, 14 },
		{ 261, 284, 14 },
		{ 261, 285, 14 },
		{ 261, 286, 14 },
		{ 261, 287, 14 },
		{ 261, 288, 14 },
		{ 261, 289, 14 },
		{ 261, 290, 14 },
		{ 261, 291, 14 },
		{ 261, 292, 14 },
		{ 261, 293, 14 },
		{ 261, 294, 14 },
		{ 261, 295, 14 },
		{ 261, 296, 14 },
		{ 261, 297, 14 },
		{ 261, 298, 14 },
		{ 261, 299, 14 },
		{ 261, 300, 14 },
		{ 261, 301, 14 },
		{ 261, 302, 14 },
		{ 261, 303, 14 },
		{ 261, 304, 14 },
		{ 261, 305, 14 },
		{ 261, 306, 14 },
		{ 261, 307, 14 },
		{ 261, 308, 14 },
		{ 261, 309, 14 },
		{ 261, 310, 14 },
		{ 261, 311, 14 },
		{ 261, 312, 14 },
		{ 261, 313, 14 },
		{ 261, 314, 14 },
		{ 261, 315, 14 },
		{ 261, 316, 14 },
		{ 261, 317, 14 },
		{ 261, 318, 14 },
		{ 261, 319, 14 },
		{ 261, 320, 14 },
		{ 261, 321, 14 },
		{ 261, 322, 14 },
		{ 261, 323, 14 },
		{ 261, 324, 14 },
		{ 261, 325, 14 },
		{ 261, 326, 14 },
		{ 261, 327, 14 },
		{ 261, 328, 14 },
		{ 261, 329, 14 },
		{ 261, 330, 14 },
		{ 261, 331, 14 },
		{ 261, 332, 14 },
		{ 261, 333, 14 },
		{ 261, 334, 14 },
		{ 261, 335, 14 },
		{ 261, 336, 14 },
		{ 261, 337, 14 },
		{ 261, 338, 14 },
		{ 261, 339, 14 },
		{ 261, 340, 14 },
		{ 261, 341, 14 },
		{ 261, 342, 14 },
		{ 261, 343, 14 },
		{ 261, 344, 14 },
		{ 261, 345, 14 },
		{ 261, 346, 14 },
		{ 261, 347, 14 },
		{ 261, 348, 14 },
		{ 261, 349, 14 },
		{ 261, 350, 14 },
		{ 261, 351, 14 },
		{ 261, 352, 14 },
		{ 261, 353, 14 },
		{ 261, 354, 14 },
		{ 261, 355, 14 },
		{ 261, 356, 14 },
		{ 261, 357, 14 },
		{ 261, 358, 14 },
		{ 261, 359, 14 },
		{ 261, 360, 14 },
		{ 261, 361, 14 },
		{ 261, 362, 14 },
		{ 261, 363, 14 },
		{ 261, 364, 14 },
		{ 261, 365, 14 },
		{ 261, 366, 14 },
		{ 261, 367, 14 },
		{ 261, 368, 14 },
		{ 261, 369, 14 },
		{ 261, 370, 14 },
		{ 261, 371, 14 },
		{ 261, 372, 14 },
		{ 261, 373, 14 },
		{ 261, 374, 14 },
		{ 261, 375, 14 },
		{ 261, 376, 14 },
		{ 261, 377, 14 },
		{ 261, 378, 14 },
		{ 261, 379, 14 },
		{ 261, 380, 14 },
		{ 261, 381, 14 },
		{ 261, 382, 14 },
		{ 261, 383, 14 },
		{ 262, 131, 14 },
		{ 262, 132, 14 },
		{ 262, 133, 14 },
		{ 262, 134, 14 },
		{ 262, 135, 14 },
		{ 262, 136, 14 },
		{ 262, 137, 14 },
		{ 262, 138, 14 },
		{ 262, 139, 14 },
		{ 262, 140, 14 },
		{ 262, 141, 14 },
		{ 262, 142, 14 },
		{ 262, 143, 14 },
		{ 262, 144, 14 },
		{ 262, 145, 14 },
		{ 262, 146, 14 },
		{ 262, 147, 14 },
		{ 262, 148, 14 },
		{ 262, 149, 14 },
		{ 262, 150, 14 },
		{ 262, 151, 14 },
		{ 262, 152, 14 },
		{ 262, 153, 14 },
		{ 262, 154, 14 },
		{ 262, 155, 14 },
		{ 262, 156, 14 },
		{ 262, 157, 14 },
		{ 262, 158, 14 },
		{ 262, 159, 14 },
		{ 262, 160, 14 },
		{ 262, 161, 14 },
		{ 262, 162, 14 },
		{ 262, 163, 14 },
		{ 262, 164, 14 },
		{ 262, 165, 14 },
		{ 262, 166, 14 },
		{ 262, 167, 14 },
		{ 262, 168, 14 },
		{ 262, 169, 14 },
		{ 262, 170, 14 },
		{ 262, 171, 14 },
		{ 262, 172, 14 },
		{ 262, 173, 14 },
		{ 262, 174, 14 },
		{ 262, 175, 14 },
		{ 262, 176, 14 },
		{ 262, 177, 14 },
		{ 262, 178, 14 },
		{ 262, 179, 14 },
		{ 262, 180, 14 },
		{ 262, 181, 14 },
		{ 262, 182, 14 },
		{ 262, 183, 14 },
		{ 262, 184, 14 },
		{ 262, 185, 14 },
		{ 262, 186, 14 },
		{ 262, 187, 14 },
		{ 262, 188, 14 },
		{ 262, 189, 14 },
		{ 262, 190, 14 },
		{ 262, 191, 14 },
		{ 262, 192, 14 },
		{ 262, 193, 14 },
		{ 262, 194, 14 },
		{ 262, 195, 14 },
		{ 262, 196, 14 },
		{ 262, 197, 14 },
		{ 262, 198, 14 },
		{ 262, 199, 14 },
		{ 262, 200, 14 },
		{ 262, 201, 14 },
		{ 262, 202, 14 },
		{ 262, 203, 14 },
		{ 262, 204, 14 },
		{ 262, 205, 14 },
		{ 262, 206, 14 },
		{ 262, 207, 14 },
		{ 262, 208, 14 },
		{ 262, 209, 14 },
		{ 262, 210, 14 },
		{ 262, 211, 14 },
		{ 262, 212, 14 },
		{ 262, 213, 14 },
		{ 262, 214, 14 },
		{ 262, 215, 14 },
		{ 262, 216, 14 },
		{ 262, 217, 14 },
		{ 262, 218, 14 },
		{ 262, 219, 14 },
		{ 262, 220, 14 },
		{ 262, 221, 14 },
		{ 262, 222, 14 },
		{ 262, 223, 14 },
		{ 262, 224, 14 },
		{ 262, 225, 14 },
		{ 262, 226, 14 },
		{ 262, 227, 14 },
		{ 262, 228, 14 },
		{ 262, 229, 14 },
		{ 262, 230, 14 },
		{ 262, 231, 14 },
		{ 262, 232, 14 },
		{ 262, 233, 14 },
		{ 262, 234, 14 },
		{ 262, 235, 14 },
		{ 262, 236, 14 },
		{ 262, 278, 14 },
		{ 262, 279, 14 },
		{ 262, 280, 14 },
		{ 262, 281, 14 },
		{ 262, 282, 14 },
		{ 262, 283, 14 },
		{ 262, 284, 14 },
		{ 262, 285, 14 },
		{ 262, 286, 14 },
		{ 262, 287, 14 },
		{ 262, 288, 14 },
		{ 262, 289, 14 },
		{ 262, 290, 14 },
		{ 262, 291, 14 },
		{ 262, 292, 14 },
		{ 262, 293, 14 },
		{ 262, 294, 14 },
		{ 262, 295, 14 },
		{ 262, 296, 14 },
		{ 262, 297, 14 },
		{ 262, 298, 14 },
		{ 262, 299, 14 },
		{ 262, 300, 14 },
		{ 262, 301, 14 },
		{ 262, 302, 14 },
		{ 262, 303, 14 },
		{ 262, 304, 14 },
		{ 262, 305, 14 },
		{ 262, 306, 14 },
		{ 262, 307, 14 },
		{ 262, 308, 14 },
		{ 262, 309, 14 },
		{ 262, 310, 14 },
		{ 262, 311, 14 },
		{ 262, 312, 14 },
		{ 262, 313, 14 },
		{ 262, 314, 14 },
		{ 262, 315, 14 },
		{ 262, 316, 14 },
		{ 262, 317, 14 },
		{ 262, 318, 14 },
		{ 262, 319, 14 },
		{ 262, 320, 14 },
		{ 262, 321, 14 },
		{ 262, 322, 14 },
		{ 262, 323, 14 },
		{ 262, 324, 14 },
		{ 262, 325, 14 },
		{ 262, 326, 14 },
		{ 262, 327, 14 },
		{ 262, 328, 14 },
		{ 262, 329, 14 },
		{ 262, 330, 14 },
		{ 262, 331, 14 },
		{ 262, 332, 14 },
		{ 262, 333, 14 },
		{ 262, 334, 14 },
		{ 262, 335, 14 },
		{ 262, 336, 14 },
		{ 262, 337, 14 },
		{ 262, 338, 14 },
		{ 262, 339, 14 },
		{ 262, 340, 14 },
		{ 262, 341, 14 },
		{ 262, 342, 14 },
		{ 262, 343, 14 },
		{ 262, 344, 14 },
		{ 262, 345, 14 },
		{ 262, 346, 14 },
		{ 262, 347, 14 },
		{ 262, 348, 14 },
		{ 262, 349, 14 },
		{ 262, 350, 14 },
		{ 262, 351, 14 },
		{ 262, 352, 14 },
		{ 262, 353, 14 },
		{ 262, 354, 14 },
		{ 262, 355, 14 },
		{ 262, 356, 14 },
		{ 262, 357, 14 },
		{ 262, 358, 14 },
		{ 262, 359, 14 },
		{ 262, 360, 14 },
		{ 262, 361, 14 },
		{ 262, 362, 14 },
		{ 262, 363, 14 },
		{ 262, 364, 14 },
		{ 262, 365, 14 },
		{ 262, 366, 14 },
		{ 262, 367, 14 },
		{ 262, 368, 14 },
		{ 262, 369, 14 },
		{ 262, 370, 14 },
		{ 262, 371, 14 },
		{ 262, 372, 14 },
		{ 262, 373, 14 },
		{ 262, 374, 14 },
		{ 262, 375, 14 },
		{ 262, 376, 14 },
		{ 262, 377, 14 },
		{ 262, 378, 14 },
		{ 262, 379, 14 },
		{ 262, 380, 14 },
		{ 262, 381, 14 },
		{ 262, 382, 14 },
		{ 262, 383, 14 },
		{ 263, 131, 14 },
		{ 263, 132, 14 },
		{ 263, 133, 14 },
		{ 263, 134, 14 },
		{ 263, 135, 14 },
		{ 263, 136, 14 },
		{ 263, 137, 14 },
		{ 263, 138, 14 },
		{ 263, 139, 14 },
		{ 263, 140, 14 },
		{ 263, 141, 14 },
		{ 263, 142, 14 },
		{ 263, 143, 14 },
		{ 263, 144, 14 },
		{ 263, 145, 14 },
		{ 263, 146, 14 },
		{ 263, 147, 14 },
		{ 263, 148, 14 },
		{ 263, 149, 14 },
		{ 263, 150, 14 },
		{ 263, 151, 14 },
		{ 263, 152, 14 },
		{ 263, 153, 14 },
		{ 263, 154, 14 },
		{ 263, 155, 14 },
		{ 263, 156, 14 },
		{ 263, 157, 14 },
		{ 263, 158, 14 },
		{ 263, 159, 14 },
		{ 263, 160, 14 },
		{ 263, 161, 14 },
		{ 263, 162, 14 },
		{ 263, 163, 14 },
		{ 263, 164, 14 },
		{ 263, 165, 14 },
		{ 263, 166, 14 },
		{ 263, 167, 14 },
		{ 263, 168, 14 },
		{ 263, 169, 14 },
		{ 263, 170, 14 },
		{ 263, 171, 14 },
		{ 263, 172, 14 },
		{ 263, 173, 14 },
		{ 263, 174, 14 },
		{ 263, 175, 14 },
		{ 263, 176, 14 },
		{ 263, 177, 14 },
		{ 263, 178, 14 },
		{ 263, 179, 14 },
		{ 263, 180, 14 },
		{ 263, 181, 14 },
		{ 263, 182, 14 },
		{ 263, 183, 14 },
		{ 263, 184, 14 },
		{ 263, 185, 14 },
		{ 263, 186, 14 },
		{ 263, 187, 14 },
		{ 263, 188, 14 },
		{ 263, 189, 14 },
		{ 263, 190, 14 },
		{ 263, 191, 14 },
		{ 263, 192, 14 },
		{ 263, 193, 14 },
		{ 263, 194, 14 },
		{ 263, 195, 14 },
		{ 263, 196, 14 },
		{ 263, 197, 14 },
		{ 263, 198, 14 },
		{ 263, 199, 14 },
		{ 263, 200, 14 },
		{ 263, 201, 14 },
		{ 263, 202, 14 },
		{ 263, 203, 14 },
		{ 263, 204, 14 },
		{ 263, 205, 14 },
		{ 263, 206, 14 },
		{ 263, 207, 14 },
		{ 263, 208, 14 },
		{ 263, 209, 14 },
		{ 263, 210, 14 },
		{ 263, 211, 14 },
		{ 263, 212, 14 },
		{ 263, 213, 14 },
		{ 263, 214, 14 },
		{ 263, 215, 14 },
		{ 263, 216, 14 },
		{ 263, 217, 14 },
		{ 263, 218, 14 },
		{ 263, 219, 14 },
		{ 263, 220, 14 },
		{ 263, 221, 14 },
		{ 263, 222, 14 },
		{ 263, 223, 14 },
		{ 263, 224, 14 },
		{ 263, 225, 14 },
		{ 263, 226, 14 },
		{ 263, 227, 14 },
		{ 263, 228, 14 },
		{ 263, 229, 14 },
		{ 263, 230, 14 },
		{ 263, 231, 14 },
		{ 263, 232, 14 },
		{ 263, 282, 14 },
		{ 263, 283, 14 },
		{ 263, 284, 14 },
		{ 263, 285, 14 },
		{ 263, 286, 14 },
		{ 263, 287, 14 },
		{ 263, 288, 14 },
		{ 263, 289, 14 },
		{ 263, 290, 14 },
		{ 263, 291, 14 },
		{ 263, 292, 14 },
		{ 263, 293, 14 },
		{ 263, 294, 14 },
		{ 263, 295, 14 },
		{ 263, 296, 14 },
		{ 263, 297, 14 },
		{ 263, 298, 14 },
		{ 263, 299, 14 },
		{ 263, 300, 14 },
		{ 263, 301, 14 },
		{ 263, 302, 14 },
		{ 263, 303, 14 },
		{ 263, 304, 14 },
		{ 263, 305, 14 },
		{ 263, 306, 14 },
		{ 263, 307, 14 },
		{ 263, 308, 14 },
		{ 263, 309, 14 },
		{ 263, 310, 14 },
		{ 263, 311, 14 },
		{ 263, 312, 14 },
		{ 263, 313, 14 },
		{ 263, 314, 14 },
		{ 263, 315, 14 },
		{ 263, 316, 14 },
		{ 263, 317, 14 },
		{ 263, 318, 14 },
		{ 263, 319, 14 },
		{ 263, 320, 14 },
		{ 263, 321, 14 },
		{ 263, 322, 14 },
		{ 263, 323, 14 },
		{ 263, 324, 14 },
		{ 263, 325, 14 },
		{ 263, 326, 14 },
		{ 263, 327, 14 },
		{ 263, 328, 14 },
		{ 263, 329, 14 },
		{ 263, 330, 14 },
		{ 263, 331, 14 },
		{ 263, 332, 14 },
		{ 263, 333, 14 },
		{ 263, 334, 14 },
		{ 263, 335, 14 },
		{ 263, 336, 14 },
		{ 263, 337, 14 },
		{ 263, 338, 14 },
		{ 263, 339, 14 },
		{ 263, 340, 14 },
		{ 263, 341, 14 },
		{ 263, 342, 14 },
		{ 263, 343, 14 },
		{ 263, 344, 14 },
		{ 263, 345, 14 },
		{ 263, 346, 14 },
		{ 263, 347, 14 },
		{ 263, 348, 14 },
		{ 263, 349, 14 },
		{ 263, 350, 14 },
		{ 263, 351, 14 },
		{ 263, 352, 14 },
		{ 263, 353, 14 },
		{ 263, 354, 14 },
		{ 263, 355, 14 },
		{ 263, 356, 14 },
		{ 263, 357, 14 },
		{ 263, 358, 14 },
		{ 263, 359, 14 },
		{ 263, 360, 14 },
		{ 263, 361, 14 },
		{ 263, 362, 14 },
		{ 263, 363, 14 },
		{ 263, 364, 14 },
		{ 263, 365, 14 },
		{ 263, 366, 14 },
		{ 263, 367, 14 },
		{ 263, 368, 14 },
		{ 263, 369, 14 },
		{ 263, 370, 14 },
		{ 263, 371, 14 },
		{ 263, 372, 14 },
		{ 263, 373, 14 },
		{ 263, 374, 14 },
		{ 263, 375, 14 },
		{ 263, 376, 14 },
		{ 263, 377, 14 },
		{ 263, 378, 14 },
		{ 263, 379, 14 },
		{ 263, 380, 14 },
		{ 263, 381, 14 },
		{ 263, 382, 14 },
		{ 263, 383, 14 },
		{ 264, 131, 14 },
		{ 264, 132, 14 },
		{ 264, 133, 14 },
		{ 264, 134, 14 },
		{ 264, 135, 14 },
		{ 264, 136, 14 },
		{ 264, 137, 14 },
		{ 264, 138, 14 },
		{ 264, 139, 14 },
		{ 264, 140, 14 },
		{ 264, 141, 14 },
		{ 264, 142, 14 },
		{ 264, 143, 14 },
		{ 264, 144, 14 },
		{ 264, 145, 14 },
		{ 264, 146, 14 },
		{ 264, 147, 14 },
		{ 264, 148, 14 },
		{ 264, 149, 14 },
		{ 264, 150, 14 },
		{ 264, 151, 14 },
		{ 264, 152, 14 },
		{ 264, 153, 14 },
		{ 264, 154, 14 },
		{ 264, 155, 14 },
		{ 264, 156, 14 },
		{ 264, 157, 14 },
		{ 264, 158, 14 },
		{ 264, 159, 14 },
		{ 264, 160, 14 },
		{ 264, 161, 14 },
		{ 264, 162, 14 },
		{ 264, 163, 14 },
		{ 264, 164, 14 },
		{ 264, 165, 14 },
		{ 264, 166, 14 },
		{ 264, 167, 14 },
		{ 264, 168, 14 },
		{ 264, 169, 14 },
		{ 264, 170, 14 },
		{ 264, 171, 14 },
		{ 264, 172, 14 },
		{ 264, 173, 14 },
		{ 264, 174, 14 },
		{ 264, 175, 14 },
		{ 264, 176, 14 },
		{ 264, 177, 14 },
		{ 264, 178, 14 },
		{ 264, 179, 14 },
		{ 264, 180, 14 },
		{ 264, 181, 14 },
		{ 264, 182, 14 },
		{ 264, 183, 14 },
		{ 264, 184, 14 },
		{ 264, 185, 14 },
		{ 264, 186, 14 },
		{ 264, 187, 14 },
		{ 264, 188, 14 },
		{ 264, 189, 14 },
		{ 264, 190, 14 },
		{ 264, 191, 14 },
		{ 264, 192, 14 },
		{ 264, 193, 14 },
		{ 264, 194, 14 },
		{ 264, 195, 14 },
		{ 264, 196, 14 },
		{ 264, 197, 14 },
		{ 264, 198, 14 },
		{ 264, 199, 14 },
		{ 264, 200, 14 },
		{ 264, 201, 14 },
		{ 264, 202, 14 },
		{ 264, 203, 14 },
		{ 264, 204, 14 },
		{ 264, 205, 14 },
		{ 264, 206, 14 },
		{ 264, 207, 14 },
		{ 264, 208, 14 },
		{ 264, 209, 14 },
		{ 264, 210, 14 },
		{ 264, 211, 14 },
		{ 264, 212, 14 },
		{ 264, 213, 14 },
		{ 264, 214, 14 },
		{ 264, 215, 14 },
		{ 264, 216, 14 },
		{ 264, 217, 14 },
		{ 264, 218, 14 },
		{ 264, 219, 14 },
		{ 264, 220, 14 },
		{ 264, 221, 14 },
		{ 264, 222, 14 },
		{ 264, 223, 14 },
		{ 264, 224, 14 },
		{ 264, 225, 14 },
		{ 264, 226, 14 },
		{ 264, 227, 14 },
		{ 264, 228, 14 },
		{ 264, 286, 14 },
		{ 264, 287, 14 },
		{ 264, 288, 14 },
		{ 264, 289, 14 },
		{ 264, 290, 14 },
		{ 264, 291, 14 },
		{ 264, 292, 14 },
		{ 264, 293, 14 },
		{ 264, 294, 14 },
		{ 264, 295, 14 },
		{ 264, 296, 14 },
		{ 264, 297, 14 },
		{ 264, 298, 14 },
		{ 264, 299, 14 },
		{ 264, 300, 14 },
		{ 264, 301, 14 },
		{ 264, 302, 14 },
		{ 264, 303, 14 },
		{ 264, 304, 14 },
		{ 264, 305, 14 },
		{ 264, 306, 14 },
		{ 264, 307, 14 },
		{ 264, 308, 14 },
		{ 264, 309, 14 },
		{ 264, 310, 14 },
		{ 264, 311, 14 },
		{ 264, 312, 14 },
		{ 264, 313, 14 },
		{ 264, 314, 14 },
		{ 264, 315, 14 },
		{ 264, 316, 14 },
		{ 264, 317, 14 },
		{ 264, 318, 14 },
		{ 264, 319, 14 },
		{ 264, 320, 14 },
		{ 264, 321, 14 },
		{ 264, 322, 14 },
		{ 264, 323, 14 },
		{ 264, 324, 14 },
		{ 264, 325, 14 },
		{ 264, 326, 14 },
		{ 264, 327, 14 },
		{ 264, 328, 14 },
		{ 264, 329, 14 },
		{ 264, 330, 14 },
		{ 264, 331, 14 },
		{ 264, 332, 14 },
		{ 264, 333, 14 },
		{ 264, 334, 14 },
		{ 264, 335, 14 },
		{ 264, 336, 14 },
		{ 264, 337, 14 },
		{ 264, 338, 14 },
		{ 264, 339, 14 },
		{ 264, 340, 14 },
		{ 264, 341, 14 },
		{ 264, 342, 14 },
		{ 264, 343, 14 },
		{ 264, 344, 14 },
		{ 264, 345, 14 },
		{ 264, 346, 14 },
		{ 264, 347, 14 },
		{ 264, 348, 14 },
		{ 264, 349, 14 },
		{ 264, 350, 14 },
		{ 264, 351, 14 },
		{ 264, 352, 14 },
		{ 264, 353, 14 },
		{ 264, 354, 14 },
		{ 264, 355, 14 },
		{ 264, 356, 14 },
		{ 264, 357, 14 },
		{ 264, 358, 14 },
		{ 264, 359, 14 },
		{ 264, 360, 14 },
		{ 264, 361, 14 },
		{ 264, 362, 14 },
		{ 264, 363, 14 },
		{ 264, 364, 14 },
		{ 264, 365, 14 },
		{ 264, 366, 14 },
		{ 264, 367, 14 },
		{ 264, 368, 14 },
		{ 264, 369, 14 },
		{ 264, 370, 14 },
		{ 264, 371, 14 },
		{ 264, 372, 14 },
		{ 264, 373, 14 },
		{ 264, 374, 14 },
		{ 264, 375, 14 },
		{ 264, 376, 14 },
		{ 264, 377, 14 },
		{ 264, 378, 14 },
		{ 264, 379, 14 },
		{ 264, 380, 14 },
		{ 264, 381, 14 },
		{ 264, 382, 14 },
		{ 264, 383, 14 },
		{ 265, 131, 14 },
		{ 265, 132, 14 },
		{ 265, 133, 14 },
		{ 265, 134, 14 },
		{ 265, 135, 14 },
		{ 265, 136, 14 },
		{ 265, 137, 14 },
		{ 265, 138, 14 },
		{ 265, 139, 14 },
		{ 265, 140, 14 },
		{ 265, 141, 14 },
		{ 265, 142, 14 },
		{ 265, 143, 14 },
		{ 265, 144, 14 },
		{ 265, 145, 14 },
		{ 265, 146, 14 },
		{ 265, 147, 14 },
		{ 265, 148, 14 },
		{ 265, 149, 14 },
		{ 265, 150, 14 },
		{ 265, 151, 14 },
		{ 265, 152, 14 },
		{ 265, 153, 14 },
		{ 265, 154, 14 },
		{ 265, 155, 14 },
		{ 265, 156, 14 },
		{ 265, 157, 14 },
		{ 265, 158, 14 },
		{ 265, 159, 14 },
		{ 265, 160, 14 },
		{ 265, 161, 14 },
		{ 265, 162, 14 },
		{ 265, 163, 14 },
		{ 265, 164, 14 },
		{ 265, 165, 14 },
		{ 265, 166, 14 },
		{ 265, 167, 14 },
		{ 265, 168, 14 },
		{ 265, 169, 14 },
		{ 265, 170, 14 },
		{ 265, 171, 14 },
		{ 265, 172, 14 },
		{ 265, 173, 14 },
		{ 265, 174, 14 },
		{ 265, 175, 14 },
		{ 265, 176, 14 },
		{ 265, 177, 14 },
		{ 265, 178, 14 },
		{ 265, 179, 14 },
		{ 265, 180, 14 },
		{ 265, 181, 14 },
		{ 265, 182, 14 },
		{ 265, 183, 14 },
		{ 265, 184, 14 },
		{ 265, 185, 14 },
		{ 265, 186, 14 },
		{ 265, 187, 14 },
		{ 265, 188, 14 },
		{ 265, 189, 14 },
		{ 265, 190, 14 },
		{ 265, 191, 14 },
		{ 265, 192, 14 },
		{ 265, 193, 14 },
		{ 265, 194, 14 },
		{ 265, 195, 14 },
		{ 265, 196, 14 },
		{ 265, 197, 14 },
		{ 265, 198, 14 },
		{ 265, 199, 14 },
		{ 265, 200, 14 },
		{ 265, 201, 14 },
		{ 265, 202, 14 },
		{ 265, 203, 14 },
		{ 265, 204, 14 },
		{ 265, 205, 14 },
		{ 265, 206, 14 },
		{ 265, 207, 14 },
		{ 265, 208, 14 },
		{ 265, 209, 14 },
		{ 265, 210, 14 },
		{ 265, 211, 14 },
		{ 265, 212, 14 },
		{ 265, 213, 14 },
		{ 265, 214, 14 },
		{ 265, 215, 14 },
		{ 265, 216, 14 },
		{ 265, 217, 14 },
		{ 265, 218, 14 },
		{ 265, 219, 14 },
		{ 265, 220, 14 },
		{ 265, 221, 14 },
		{ 265, 222, 14 },
		{ 265, 223, 14 },
		{ 265, 224, 14 },
		{ 265, 290, 14 },
		{ 265, 291, 14 },
		{ 265, 292, 14 },
		{ 265, 293, 14 },
		{ 265, 294, 14 },
		{ 265, 295, 14 },
		{ 265, 296, 14 },
		{ 265, 297, 14 },
		{ 265, 298, 14 },
		{ 265, 299, 14 },
		{ 265, 300, 14 },
		{ 265, 301, 14 },
		{ 265, 302, 14 },
		{ 265, 303, 14 },
		{ 265, 304, 14 },
		{ 265, 305, 14 },
		{ 265, 306, 14 },
		{ 265, 307, 14 },
		{ 265, 308, 14 },
		{ 265, 309, 14 },
		{ 265, 310, 14 },
		{ 265, 311, 14 },
		{ 265, 312, 14 },
		{ 265, 313, 14 },
		{ 265, 314, 14 },
		{ 265, 315, 14 },
		{ 265, 316, 14 },
		{ 265, 317, 14 },
		{ 265, 318, 14 },
		{ 265, 319, 14 },
		{ 265, 320, 14 },
		{ 265, 321, 14 },
		{ 265, 322, 14 },
		{ 265, 323, 14 },
		{ 265, 324, 14 },
		{ 265, 325, 14 },
		{ 265, 326, 14 },
		{ 265, 327, 14 },
		{ 265, 328, 14 },
		{ 265, 329, 14 },
		{ 265, 330, 14 },
		{ 265, 331, 14 },
		{ 265, 332, 14 },
		{ 265, 333, 14 },
		{ 265, 334, 14 },
		{ 265, 335, 14 },
		{ 265, 336, 14 },
		{ 265, 337, 14 },
		{ 265, 338, 14 },
		{ 265, 339, 14 },
		{ 265, 340, 14 },
		{ 265, 341, 14 },
		{ 265, 342, 14 },
		{ 265, 343, 14 },
		{ 265, 344, 14 },
		{ 265, 345, 14 },
		{ 265, 346, 14 },
		{ 265, 347, 14 },
		{ 265, 348, 14 },
		{ 265, 349, 14 },
		{ 265, 350, 14 },
		{ 265, 351, 14 },
		{ 265, 352, 14 },
		{ 265, 353, 14 },
		{ 265, 354, 14 },
		{ 265, 355, 14 },
		{ 265, 356, 14 },
		{ 265, 357, 14 },
		{ 265, 358, 14 },
		{ 265, 359, 14 },
		{ 265, 360, 14 },
		{ 265, 361, 14 },
		{ 265, 362, 14 },
		{ 265, 363, 14 },
		{ 265, 364, 14 },
		{ 265, 365, 14 },
		{ 265, 366, 14 },
		{ 265, 367, 14 },
		{ 265, 368, 14 },
		{ 265, 369, 14 },
		{ 265, 370, 14 },
		{ 265, 371, 14 },
		{ 265, 372, 14 },
		{ 265, 373, 14 },
		{ 265, 374, 14 },
		{ 265, 375, 14 },
		{ 265, 376, 14 },
		{ 265, 377, 14 },
		{ 265, 378, 14 },
		{ 265, 379, 14 },
		{ 265, 380, 14 },
		{ 265, 381, 14 },
		{ 265, 382, 14 },
		{ 265, 383, 14 },
		{ 266, 131, 14 },
		{ 266, 132, 14 },
		{ 266, 133, 14 },
		{ 266, 134, 14 },
		{ 266, 135, 14 },
		{ 266, 136, 14 },
		{ 266, 137, 14 },
		{ 266, 138, 14 },
		{ 266, 139, 14 },
		{ 266, 140, 14 },
		{ 266, 141, 14 },
		{ 266, 142, 14 },
		{ 266, 143, 14 },
		{ 266, 144, 14 },
		{ 266, 145, 14 },
		{ 266, 146, 14 },
		{ 266, 147, 14 },
		{ 266, 148, 14 },
		{ 266, 149, 14 },
		{ 266, 150, 14 },
		{ 266, 151, 14 },
		{ 266, 152, 14 },
		{ 266, 153, 14 },
		{ 266, 154, 14 },
		{ 266, 155, 14 },
		{ 266, 156, 14 },
		{ 266, 157, 14 },
		{ 266, 158, 14 },
		{ 266, 159, 14 },
		{ 266, 160, 14 },
		{ 266, 161, 14 },
		{ 266, 162, 14 },
		{ 266, 163, 14 },
		{ 266, 164, 14 },
		{ 266, 165, 14 },
		{ 266, 166, 14 },
		{ 266, 167, 14 },
		{ 266, 168, 14 },
		{ 266, 169, 14 },
		{ 266, 170, 14 },
		{ 266, 171, 14 },
		{ 266, 172, 14 },
		{ 266, 173, 14 },
		{ 266, 174, 14 },
		{ 266, 175, 14 },
		{ 266, 176, 14 },
		{ 266, 177, 14 },
		{ 266, 178, 14 },
		{ 266, 179, 14 },
		{ 266, 180, 14 },
		{ 266, 181, 14 },
		{ 266, 182, 14 },
		{ 266, 183, 14 },
		{ 266, 184, 14 },
		{ 266, 185, 14 },
		{ 266, 186, 14 },
		{ 266, 187, 14 },
		{ 266, 188, 14 },
		{ 266, 189, 14 },
		{ 266, 190, 14 },
		{ 266, 191, 14 },
		{ 266, 192, 14 },
		{ 266, 193, 14 },
		{ 266, 194, 14 },
		{ 266, 195, 14 },
		{ 266, 196, 14 },
		{ 266, 197, 14 },
		{ 266, 198, 14 },
		{ 266, 199, 14 },
		{ 266, 200, 14 },
		{ 266, 201, 14 },
		{ 266, 202, 14 },
		{ 266, 203, 14 },
		{ 266, 204, 14 },
		{ 266, 205, 14 },
		{ 266, 206, 14 },
		{ 266, 207, 14 },
		{ 266, 208, 14 },
		{ 266, 209, 14 },
		{ 266, 210, 14 },
		{ 266, 211, 14 },
		{ 266, 212, 14 },
		{ 266, 213, 14 },
		{ 266, 214, 14 },
		{ 266, 215, 14 },
		{ 266, 216, 14 },
		{ 266, 217, 14 },
		{ 266, 218, 14 },
		{ 266, 219, 14 },
		{ 266, 220, 14 },
		{ 266, 294, 14 },
		{ 266, 295, 14 },
		{ 266, 296, 14 },
		{ 266, 297, 14 },
		{ 266, 298, 14 },
		{ 266, 299, 14 },
		{ 266, 300, 14 },
		{ 266, 301, 14 },
		{ 266, 302, 14 },
		{ 266, 303, 14 },
		{ 266, 304, 14 },
		{ 266, 305, 14 },
		{ 266, 306, 14 },
		{ 266, 307, 14 },
		{ 266, 308, 14 },
		{ 266, 309, 14 },
		{ 266, 310, 14 },
		{ 266, 311, 14 },
		{ 266, 312, 14 },
		{ 266, 313, 14 },
		{ 266, 314, 14 },
		{ 266, 315, 14 },
		{ 266, 316, 14 },
		{ 266, 317, 14 },
		{ 266, 318, 14 },
		{ 266, 319, 14 },
		{ 266, 320, 14 },
		{ 266, 321, 14 },
		{ 266, 322, 14 },
		{ 266, 323, 14 },
		{ 266, 324, 14 },
		{ 266, 325, 14 },
		{ 266, 326, 14 },
		{ 266, 327, 14 },
		{ 266, 328, 14 },
		{ 266, 329, 14 },
		{ 266, 330, 14 },
		{ 266, 331, 14 },
		{ 266, 332, 14 },
		{ 266, 333, 14 },
		{ 266, 334, 14 },
		{ 266, 335, 14 },
		{ 266, 336, 14 },
		{ 266, 337, 14 },
		{ 266, 338, 14 },
		{ 266, 339, 14 },
		{ 266, 340, 14 },
		{ 266, 341, 14 },
		{ 266, 342, 14 },
		{ 266, 343, 14 },
		{ 266, 344, 14 },
		{ 266, 345, 14 },
		{ 266, 346, 14 },
		{ 266, 347, 14 },
		{ 266, 348, 14 },
		{ 266, 349, 14 },
		{ 266, 350, 14 },
		{ 266, 351, 14 },
		{ 266, 352, 14 },
		{ 266, 353, 14 },
		{ 266, 354, 14 },
		{ 266, 355, 14 },
		{ 266, 356, 14 },
		{ 266, 357, 14 },
		{ 266, 358, 14 },
		{ 266, 359, 14 },
		{ 266, 360, 14 },
		{ 266, 361, 14 },
		{ 266, 362, 14 },
		{ 266, 363, 14 },
		{ 266, 364, 14 },
		{ 266, 365, 14 },
		{ 266, 366, 14 },
		{ 266, 367, 14 },
		{ 266, 368, 14 },
		{ 266, 369, 14 },
		{ 266, 370, 14 },
		{ 266, 371, 14 },
		{ 266, 372, 14 },
		{ 266, 373, 14 },
		{ 266, 374, 14 },
		{ 266, 375, 14 },
		{ 266, 376, 14 },
		{ 266, 377, 14 },
		{ 266, 378, 14 },
		{ 266, 379, 14 },
		{ 266, 380, 14 },
		{ 266, 381, 14 },
		{ 266, 382, 14 },
		{ 266, 383, 14 },
		{ 267, 131, 14 },
		{ 267, 132, 14 },
		{ 267, 133, 14 },
		{ 267, 134, 14 },
		{ 267, 135, 14 },
		{ 267, 136, 14 },
		{ 267, 137, 14 },
		{ 267, 138, 14 },
		{ 267, 139, 14 },
		{ 267, 140, 14 },
		{ 267, 141, 14 },
		{ 267, 142, 14 },
		{ 267, 143, 14 },
		{ 267, 144, 14 },
		{ 267, 145, 14 },
		{ 267, 146, 14 },
		{ 267, 147, 14 },
		{ 267, 148, 14 },
		{ 267, 149, 14 },
		{ 267, 150, 14 },
		{ 267, 151, 14 },
		{ 267, 152, 14 },
		{ 267, 153, 14 },
		{ 267, 154, 14 },
		{ 267, 155, 14 },
		{ 267, 156, 14 },
		{ 267, 157, 14 },
		{ 267, 158, 14 },
		{ 267, 159, 14 },
		{ 267, 160, 14 },
		{ 267, 161, 14 },
		{ 267, 162, 14 },
		{ 267, 163, 14 },
		{ 267, 164, 14 },
		{ 267, 165, 14 },
		{ 267, 166, 14 },
		{ 267, 167, 14 },
		{ 267, 168, 14 },
		{ 267, 169, 14 },
		{ 267, 170, 14 },
		{ 267, 171, 14 },
		{ 267, 172, 14 },
		{ 267, 173, 14 },
		{ 267, 174, 14 },
		{ 267, 175, 14 },
		{ 267, 176, 14 },
		{ 267, 177, 14 },
		{ 267, 178, 14 },
		{ 267, 179, 14 },
		{ 267, 180, 14 },
		{ 267, 181, 14 },
		{ 267, 182, 14 },
		{ 267, 183, 14 },
		{ 267, 184, 14 },
		{ 267, 185, 14 },
		{ 267, 186, 14 },
		{ 267, 187, 14 },
		{ 267, 188, 14 },
		{ 267, 189, 14 },
		{ 267, 190, 14 },
		{ 267, 191, 14 },
		{ 267, 192, 14 },
		{ 267, 193, 14 },
		{ 267, 194, 14 },
		{ 267, 195, 14 },
		{ 267, 196, 14 },
		{ 267, 197, 14 },
		{ 267, 198, 14 },
		{ 267, 199, 14 },
		{ 267, 200, 14 },
		{ 267, 201, 14 },
		{ 267, 202, 14 },
		{ 267, 203, 14 },
		{ 267, 204, 14 },
		{ 267, 205, 14 },
		{ 267, 206, 14 },
		{ 267, 207, 14 },
		{ 267, 208, 14 },
		{ 267, 209, 14 },
		{ 267, 210, 14 },
		{ 267, 211, 14 },
		{ 267, 212, 14 },
		{ 267, 213, 14 },
		{ 267, 214, 14 },
		{ 267, 215, 14 },
		{ 267, 216, 14 },
		{ 267, 298, 14 },
		{ 267, 299, 14 },
		{ 267, 300, 14 },
		{ 267, 301, 14 },
		{ 267, 302, 14 },
		{ 267, 303, 14 },
		{ 267, 304, 14 },
		{ 267, 305, 14 },
		{ 267, 306, 14 },
		{ 267, 307, 14 },
		{ 267, 308, 14 },
		{ 267, 309, 14 },
		{ 267, 310, 14 },
		{ 267, 311, 14 },
		{ 267, 312, 14 },
		{ 267, 313, 14 },
		{ 267, 314, 14 },
		{ 267, 315, 14 },
		{ 267, 316, 14 },
		{ 267, 317, 14 },
		{ 267, 318, 14 },
		{ 267, 319, 14 },
		{ 267, 320, 14 },
		{ 267, 321, 14 },
		{ 267, 322, 14 },
		{ 267, 323, 14 },
		{ 267, 324, 14 },
		{ 267, 325, 14 },
		{ 267, 326, 14 },
		{ 267, 327, 14 },
		{ 267, 328, 14 },
		{ 267, 329, 14 },
		{ 267, 330, 14 },
		{ 267, 331, 14 },
		{ 267, 332, 14 },
		{ 267, 333, 14 },
		{ 267, 334, 14 },
		{ 267, 335, 14 },
		{ 267, 336, 14 },
		{ 267, 337, 14 },
		{ 267, 338, 14 },
		{ 267, 339, 14 },
		{ 267, 340, 14 },
		{ 267, 341, 14 },
		{ 267, 342, 14 },
		{ 267, 343, 14 },
		{ 267, 344, 14 },
		{ 267, 345, 14 },
		{ 267, 346, 14 },
		{ 267, 347, 14 },
		{ 267, 348, 14 },
		{ 267, 349, 14 },
		{ 267, 350, 14 },
		{ 267, 351, 14 },
		{ 267, 352, 14 },
		{ 267, 353, 14 },
		{ 267, 354, 14 },
		{ 267, 355, 14 },
		{ 267, 356, 14 },
		{ 267, 357, 14 },
		{ 267, 358, 14 },
		{ 267, 359, 14 },
		{ 267, 360, 14 },
		{ 267, 361, 14 },
		{ 267, 362, 14 },
		{ 267, 363, 14 },
		{ 267, 364, 14 },
		{ 267, 365, 14 },
		{ 267, 366, 14 },
		{ 267, 367, 14 },
		{ 267, 368, 14 },
		{ 267, 369, 14 },
		{ 267, 370, 14 },
		{ 267, 371, 14 },
		{ 267, 372, 14 },
		{ 267, 373, 14 },
		{ 267, 374, 14 },
		{ 267, 375, 14 },
		{ 267, 376, 14 },
		{ 267, 377, 14 },
		{ 267, 378, 14 },
		{ 267, 379, 14 },
		{ 267, 380, 14 },
		{ 267, 381, 14 },
		{ 267, 382, 14 },
		{ 267, 383, 14 },
		{ 268, 131, 14 },
		{ 268, 132, 14 },
		{ 268, 133, 14 },
		{ 268, 134, 14 },
		{ 268, 135, 14 },
		{ 268, 136, 14 },
		{ 268, 137, 14 },
		{ 268, 138, 14 },
		{ 268, 139, 14 },
		{ 268, 140, 14 },
		{ 268, 141, 14 },
		{ 268, 142, 14 },
		{ 268, 143, 14 },
		{ 268, 144, 14 },
		{ 268, 145, 14 },
		{ 268, 146, 14 },
		{ 268, 147, 14 },
		{ 268, 148, 14 },
		{ 268, 149, 14 },
		{ 268, 150, 14 },
		{ 268, 151, 14 },
		{ 268, 152, 14 },
		{ 268, 153, 14 },
		{ 268, 154, 14 },
		{ 268, 155, 14 },
		{ 268, 156, 14 },
		{ 268, 157, 14 },
		{ 268, 158, 14 },
		{ 268, 159, 14 },
		{ 268, 160, 14 },
		{ 268, 161, 14 },
		{ 268, 162, 14 },
		{ 268, 163, 14 },
		{ 268, 164, 14 },
		{ 268, 165, 14 },
		{ 268, 166, 14 },
		{ 268, 167, 14 },
		{ 268, 168, 14 },
		{ 268, 169, 14 },
		{ 268, 170, 14 },
		{ 268, 171, 14 },
		{ 268, 172, 14 },
		{ 268, 173, 14 },
		{ 268, 174, 14 },
		{ 268, 175, 14 },
		{ 268, 176, 14 },
		{ 268, 177, 14 },
		{ 268, 178, 14 },
		{ 268, 179, 14 },
		{ 268, 180, 14 },
		{ 268, 181, 14 },
		{ 268, 182, 14 },
		{ 268, 183, 14 },
		{ 268, 184, 14 },
		{ 268, 185, 14 },
		{ 268, 186, 14 },
		{ 268, 187, 14 },
		{ 268, 188, 14 },
		{ 268, 189, 14 },
		{ 268, 190, 14 },
		{ 268, 191, 14 },
		{ 268, 192, 14 },
		{ 268, 193, 14 },
		{ 268, 194, 14 },
		{ 268, 195, 14 },
		{ 268, 196, 14 },
		{ 268, 197, 14 },
		{ 268, 198, 14 },
		{ 268, 199, 14 },
		{ 268, 200, 14 },
		{ 268, 201, 14 },
		{ 268, 202, 14 },
		{ 268, 203, 14 },
		{ 268, 204, 14 },
		{ 268, 205, 14 },
		{ 268, 206, 14 },
		{ 268, 207, 14 },
		{ 268, 208, 14 },
		{ 268, 209, 14 },
		{ 268, 210, 14 },
		{ 268, 211, 14 },
		{ 268, 212, 14 },
		{ 268, 302, 14 },
		{ 268, 303, 14 },
		{ 268, 304, 14 },
		{ 268, 305, 14 },
		{ 268, 306, 14 },
		{ 268, 307, 14 },
		{ 268, 308, 14 },
		{ 268, 309, 14 },
		{ 268, 310, 14 },
		{ 268, 311, 14 },
		{ 268, 312, 14 },
		{ 268, 313, 14 },
		{ 268, 314, 14 },
		{ 268, 315, 14 },
		{ 268, 316, 14 },
		{ 268, 317, 14 },
		{ 268, 318, 14 },
		{ 268, 319, 14 },
		{ 268, 320, 14 },
		{ 268, 321, 14 },
		{ 268, 322, 14 },
		{ 268, 323, 14 },
		{ 268, 324, 14 },
		{ 268, 325, 14 },
		{ 268, 326, 14 },
		{ 268, 327, 14 },
		{ 268, 328, 14 },
		{ 268, 329, 14 },
		{ 268, 330, 14 },
		{ 268, 331, 14 },
		{ 268, 332, 14 },
		{ 268, 333, 14 },
		{ 268, 334, 14 },
		{ 268, 335, 14 },
		{ 268, 336, 14 },
		{ 268, 337, 14 },
		{ 268, 338, 14 },
		{ 268, 339, 14 },
		{ 268, 340, 14 },
		{ 268, 341, 14 },
		{ 268, 342, 14 },
		{ 268, 343, 14 },
		{ 268, 344, 14 },
		{ 268, 345, 14 },
		{ 268, 346, 14 },
		{ 268, 347, 14 },
		{ 268, 348, 14 },
		{ 268, 349, 14 },
		{ 268, 350, 14 },
		{ 268, 351, 14 },
		{ 268, 352, 14 },
		{ 268, 353, 14 },
		{ 268, 354, 14 },
		{ 268, 355, 14 },
		{ 268, 356, 14 },
		{ 268, 357, 14 },
		{ 268, 358, 14 },
		{ 268, 359, 14 },
		{ 268, 360, 14 },
		{ 268, 361, 14 },
		{ 268, 362, 14 },
		{ 268, 363, 14 },
		{ 268, 364, 14 },
		{ 268, 365, 14 },
		{ 268, 366, 14 },
		{ 268, 367, 14 },
		{ 268, 368, 14 },
		{ 268, 369, 14 },
		{ 268, 370, 14 },
		{ 268, 371, 14 },
		{ 268, 372, 14 },
		{ 268, 373, 14 },
		{ 268, 374, 14 },
		{ 268, 375, 14 },
		{ 268, 376, 14 },
		{ 268, 377, 14 },
		{ 268, 378, 14 },
		{ 268, 379, 14 },
		{ 268, 380, 14 },
		{ 268, 381, 14 },
		{ 268, 382, 14 },
		{ 268, 383, 14 },
		{ 269, 131, 14 },
		{ 269, 132, 14 },
		{ 269, 133, 14 },
		{ 269, 134, 14 },
		{ 269, 135, 14 },
		{ 269, 136, 14 },
		{ 269, 137, 14 },
		{ 269, 138, 14 },
		{ 269, 139, 14 },
		{ 269, 140, 14 },
		{ 269, 141, 14 },
		{ 269, 142, 14 },
		{ 269, 143, 14 },
		{ 269, 144, 14 },
		{ 269, 145, 14 },
		{ 269, 146, 14 },
		{ 269, 147, 14 },
		{ 269, 148, 14 },
		{ 269, 149, 14 },
		{ 269, 150, 14 },
		{ 269, 151, 14 },
		{ 269, 152, 14 },
		{ 269, 153, 14 },
		{ 269, 154, 14 },
		{ 269, 155, 14 },
		{ 269, 156, 14 },
		{ 269, 157, 14 },
		{ 269, 158, 14 },
		{ 269, 159, 14 },
		{ 269, 160, 14 },
		{ 269, 161, 14 },
		{ 269, 162, 14 },
		{ 269, 163, 14 },
		{ 269, 164, 14 },
		{ 269, 165, 14 },
		{ 269, 166, 14 },
		{ 269, 167, 14 },
		{ 269, 168, 14 },
		{ 269, 169, 14 },
		{ 269, 170, 14 },
		{ 269, 171, 14 },
		{ 269, 172, 14 },
		{ 269, 173, 14 },
		{ 269, 174, 14 },
		{ 269, 175, 14 },
		{ 269, 176, 14 },
		{ 269, 177, 14 },
		{ 269, 178, 14 },
		{ 269, 179, 14 },
		{ 269, 180, 14 },
		{ 269, 181, 14 },
		{ 269, 182, 14 },
		{ 269, 183, 14 },
		{ 269, 184, 14 },
		{ 269, 185, 14 },
		{ 269, 186, 14 },
		{ 269, 187, 14 },
		{ 269, 188, 14 },
		{ 269, 189, 14 },
		{ 269, 190, 14 },
		{ 269, 191, 14 },
		{ 269, 192, 14 },
		{ 269, 193, 14 },
		{ 269, 194, 14 },
		{ 269, 195, 14 },
		{ 269, 196, 14 },
		{ 269, 197, 14 },
		{ 269, 198, 14 },
		{ 269, 199, 14 },
		{ 269, 200, 14 },
		{ 269, 201, 14 },
		{ 269, 202, 14 },
		{ 269, 203, 14 },
		{ 269, 204, 14 },
		{ 269, 205, 14 },
		{ 269, 206, 14 },
		{ 269, 207, 14 },
		{ 269, 208, 14 },
		{ 269, 306, 14 },
		{ 269, 307, 14 },
		{ 269, 308, 14 },
		{ 269, 309, 14 },
		{ 269, 310, 14 },
		{ 269, 311, 14 },
		{ 269, 312, 14 },
		{ 269, 313, 14 },
		{ 269, 314, 14 },
		{ 269, 315, 14 },
		{ 269, 316, 14 },
		{ 269, 317, 14 },
		{ 269, 318, 14 },
		{ 269, 319, 14 },
		{ 269, 320, 14 },
		{ 269, 321, 14 },
		{ 269, 322, 14 },
		{ 269, 323, 14 },
		{ 269, 324, 14 },
		{ 269, 325, 14 },
		{ 269, 326, 14 },
		{ 269, 327, 14 },
		{ 269, 328, 14 },
		{ 269, 329, 14 },
		{ 269, 330, 14 },
		{ 269, 331, 14 },
		{ 269, 332, 14 },
		{ 269, 333, 14 },
		{ 269, 334, 14 },
		{ 269, 335, 14 },
		{ 269, 336, 14 },
		{ 269, 337, 14 },
		{ 269, 338, 14 },
		{ 269, 339, 14 },
		{ 269, 340, 14 },
		{ 269, 341, 14 },
		{ 269, 342, 14 },
		{ 269, 343, 14 },
		{ 269, 344, 14 },
		{ 269, 345, 14 },
		{ 269, 346, 14 },
		{ 269, 347, 14 },
		{ 269, 348, 14 },
		{ 269, 349, 14 },
		{ 269, 350, 14 },
		{ 269, 351, 14 },
		{ 269, 352, 14 },
		{ 269, 353, 14 },
		{ 269, 354, 14 },
		{ 269, 355, 14 },
		{ 269, 356, 14 },
		{ 269, 357, 14 },
		{ 269, 358, 14 },
		{ 269, 359, 14 },
		{ 269, 360, 14 },
		{ 269, 361, 14 },
		{ 269, 362, 14 },
		{ 269, 363, 14 },
		{ 269, 364, 14 },
		{ 269, 365, 14 },
		{ 269, 366, 14 },
		{ 269, 367, 14 },
		{ 269, 368, 14 },
		{ 269, 369, 14 },
		{ 269, 370, 14 },
		{ 269, 371, 14 },
		{ 269, 372, 14 },
		{ 269, 373, 14 },
		{ 269, 374, 14 },
		{ 269, 375, 14 },
		{ 269, 376, 14 },
		{ 269, 377, 14 },
		{ 269, 378, 14 },
		{ 269, 379, 14 },
		{ 269, 380, 14 },
		{ 269, 381, 14 },
		{ 269, 382, 14 },
		{ 269, 383, 14 },
		{ 270, 131, 14 },
		{ 270, 132, 14 },
		{ 270, 133, 14 },
		{ 270, 134, 14 },
		{ 270, 135, 14 },
		{ 270, 136, 14 },
		{ 270, 137, 14 },
		{ 270, 138, 14 },
		{ 270, 139, 14 },
		{ 270, 140, 14 },
		{ 270, 141, 14 },
		{ 270, 142, 14 },
		{ 270, 143, 14 },
		{ 270, 144, 14 },
		{ 270, 145, 14 },
		{ 270, 146, 14 },
		{ 270, 147, 14 },
		{ 270, 148, 14 },
		{ 270, 149, 14 },
		{ 270, 150, 14 },
		{ 270, 151, 14 },
		{ 270, 152, 14 },
		{ 270, 153, 14 },
		{ 270, 154, 14 },
		{ 270, 155, 14 },
		{ 270, 156, 14 },
		{ 270, 157, 14 },
		{ 270, 158, 14 },
		{ 270, 159, 14 },
		{ 270, 160, 14 },
		{ 270, 161, 14 },
		{ 270, 162, 14 },
		{ 270, 163, 14 },
		{ 270, 164, 14 },
		{ 270, 165, 14 },
		{ 270, 166, 14 },
		{ 270, 167, 14 },
		{ 270, 168, 14 },
		{ 270, 169, 14 },
		{ 270, 170, 14 },
		{ 270, 171, 14 },
		{ 270, 172, 14 },
		{ 270, 173, 14 },
		{ 270, 174, 14 },
		{ 270, 175, 14 },
		{ 270, 176, 14 },
		{ 270, 177, 14 },
		{ 270, 178, 14 },
		{ 270, 179, 14 },
		{ 270, 180, 14 },
		{ 270, 181, 14 },
		{ 270, 182, 14 },
		{ 270, 183, 14 },
		{ 270, 184, 14 },
		{ 270, 185, 14 },
		{ 270, 186, 14 },
		{ 270, 187, 14 },
		{ 270, 188, 14 },
		{ 270, 189, 14 },
		{ 270, 190, 14 },
		{ 270, 191, 14 },
		{ 270, 192, 14 },
		{ 270, 193, 14 },
		{ 270, 194, 14 },
		{ 270, 195, 14 },
		{ 270, 196, 14 },
		{ 270, 197, 14 },
		{ 270, 198, 14 },
		{ 270, 199, 14 },
		{ 270, 200, 14 },
		{ 270, 201, 14 },
		{ 270, 202, 14 },
		{ 270, 203, 14 },
		{ 270, 204, 14 },
		{ 270, 310, 14 },
		{ 270, 311, 14 },
		{ 270, 312, 14 },
		{ 270, 313, 14 },
		{ 270, 314, 14 },
		{ 270, 315, 14 },
		{ 270, 316, 14 },
		{ 270, 317, 14 },
		{ 270, 318, 14 },
		{ 270, 319, 14 },
		{ 270, 320, 14 },
		{ 270, 321, 14 },
		{ 270, 322, 14 },
		{ 270, 323, 14 },
		{ 270, 324, 14 },
		{ 270, 325, 14 },
		{ 270, 326, 14 },
		{ 270, 327, 14 },
		{ 270, 328, 14 },
		{ 270, 329, 14 },
		{ 270, 330, 14 },
		{ 270, 331, 14 },
		{ 270, 332, 14 },
		{ 270, 333, 14 },
		{ 270, 334, 14 },
		{ 270, 335, 14 },
		{ 270, 336, 14 },
		{ 270, 337, 14 },
		{ 270, 338, 14 },
		{ 270, 339, 14 },
		{ 270, 340, 14 },
		{ 270, 341, 14 },
		{ 270, 342, 14 },
		{ 270, 343, 14 },
		{ 270, 344, 14 },
		{ 270, 345, 14 },
		{ 270, 346, 14 },
		{ 270, 347, 14 },
		{ 270, 348, 14 },
		{ 270, 349, 14 },
		{ 270, 350, 14 },
		{ 270, 351, 14 },
		{ 270, 352, 14 },
		{ 270, 353, 14 },
		{ 270, 354, 14 },
		{ 270, 355, 14 },
		{ 270, 356, 14 },
		{ 270, 357, 14 },
		{ 270, 358, 14 },
		{ 270, 359, 14 },
		{ 270, 360, 14 },
		{ 270, 361, 14 },
		{ 270, 362, 14 },
		{ 270, 363, 14 },
		{ 270, 364, 14 },
		{ 270, 365, 14 },
		{ 270, 366, 14 },
		{ 270, 367, 14 },
		{ 270, 368, 14 },
		{ 270, 369, 14 },
		{ 270, 370, 14 },
		{ 270, 371, 14 },
		{ 270, 372, 14 },
		{ 270, 373, 14 },
		{ 270, 374, 14 },
		{ 270, 375, 14 },
		{ 270, 376, 14 },
		{ 270, 377, 14 },
		{ 270, 378, 14 },
		{ 270, 379, 14 },
		{ 270, 380, 14 },
		{ 270, 381, 14 },
		{ 270, 382, 14 },
		{ 270, 383, 14 },
		{ 271, 131, 14 },
		{ 271, 132, 14 },
		{ 271, 133, 14 },
		{ 271, 134, 14 },
		{ 271, 135, 14 },
		{ 271, 136, 14 },
		{ 271, 137, 14 },
		{ 271, 138, 14 },
		{ 271, 139, 14 },
		{ 271, 140, 14 },
		{ 271, 141, 14 },
		{ 271, 142, 14 },
		{ 271, 143, 14 },
		{ 271, 144, 14 },
		{ 271, 145, 14 },
		{ 271, 146, 14 },
		{ 271, 147, 14 },
		{ 271, 148, 14 },
		{ 271, 149, 14 },
		{ 271, 150, 14 },
		{ 271, 151, 14 },
		{ 271, 152, 14 },
		{ 271, 153, 14 },
		{ 271, 154, 14 },
		{ 271, 155, 14 },
		{ 271, 156, 14 },
		{ 271, 157, 14 },
		{ 271, 158, 14 },
		{ 271, 159, 14 },
		{ 271, 160, 14 },
		{ 271, 161, 14 },
		{ 271, 162, 14 },
		{ 271, 163, 14 },
		{ 271, 164, 14 },
		{ 271, 165, 14 },
		{ 271, 166, 14 },
		{ 271, 167, 14 },
		{ 271, 168, 14 },
		{ 271, 169, 14 },
		{ 271, 170, 14 },
		{ 271, 171, 14 },
		{ 271, 172, 14 },
		{ 271, 173, 14 },
		{ 271, 174, 14 },
		{ 271, 175, 14 },
		{ 271, 176, 14 },
		{ 271, 177, 14 },
		{ 271, 178, 14 },
		{ 271, 179, 14 },
		{ 271, 180, 14 },
		{ 271, 181, 14 },
		{ 271, 182, 14 },
		{ 271, 183, 14 },
		{ 271, 184, 14 },
		{ 271, 185, 14 },
		{ 271, 186, 14 },
		{ 271, 187, 14 },
		{ 271, 188, 14 },
		{ 271, 189, 14 },
		{ 271, 190, 14 },
		{ 271, 191, 14 },
		{ 271, 192, 14 },
		{ 271, 193, 14 },
		{ 271, 194, 14 },
		{ 271, 195, 14 },
		{ 271, 196, 14 },
		{ 271, 197, 14 },
		{ 271, 198, 14 },
		{ 271, 199, 14 },
		{ 271, 200, 14 },
		{ 271, 314, 14 },
		{ 271, 315, 14 },
		{ 271, 316, 14 },
		{ 271, 317, 14 },
		{ 271, 318, 14 },
		{ 271, 319, 14 },
		{ 271, 320, 14 },
		{ 271, 321, 14 },
		{ 271, 322, 14 },
		{ 271, 323, 14 },
		{ 271, 324, 14 },
		{ 271, 325, 14 },
		{ 271, 326, 14 },
		{ 271, 327, 14 },
		{ 271, 328, 14 },
		{ 271, 329, 14 },
		{ 271, 330, 14 },
		{ 271, 331, 14 },
		{ 271, 332, 14 },
		{ 271, 333, 14 },
		{ 271, 334, 14 },
		{ 271, 335, 14 },
		{ 271, 336, 14 },
		{ 271, 337, 14 },
		{ 271, 338, 14 },
		{ 271, 339, 14 },
		{ 271, 340, 14 },
		{ 271, 341, 14 },
		{ 271, 342, 14 },
		{ 271, 343, 14 },
		{ 271, 344, 14 },
		{ 271, 345, 14 },
		{ 271, 346, 14 },
		{ 271, 347, 14 },
		{ 271, 348, 14 },
		{ 271, 349, 14 },
		{ 271, 350, 14 },
		{ 271, 351, 14 },
		{ 271, 352, 14 },
		{ 271, 353, 14 },
		{ 271, 354, 14 },
		{ 271, 355, 14 },
		{ 271, 356, 14 },
		{ 271, 357, 14 },
		{ 271, 358, 14 },
		{ 271, 359, 14 },
		{ 271, 360, 14 },
		{ 271, 361, 14 },
		{ 271, 362, 14 },
		{ 271, 363, 14 },
		{ 271, 364, 14 },
		{ 271, 365, 14 },
		{ 271, 366, 14 },
		{ 271, 367, 14 },
		{ 271, 368, 14 },
		{ 271, 369, 14 },
		{ 271, 370, 14 },
		{ 271, 371, 14 },
		{ 271, 372, 14 },
		{ 271, 373, 14 },
		{ 271, 374, 14 },
		{ 271, 375, 14 },
		{ 271, 376, 14 },
		{ 271, 377, 14 },
		{ 271, 378, 14 },
		{ 271, 379, 14 },
		{ 271, 380, 14 },
		{ 271, 381, 14 },
		{ 271, 382, 14 },
		{ 271, 383, 14 },
		{ 272, 131, 14 },
		{ 272, 132, 14 },
		{ 272, 133, 14 },
		{ 272, 134, 14 },
		{ 272, 135, 14 },
		{ 272, 136, 14 },
		{ 272, 137, 14 },
		{ 272, 138, 14 },
		{ 272, 139, 14 },
		{ 272, 140, 14 },
		{ 272, 141, 14 },
		{ 272, 142, 14 },
		{ 272, 143, 14 },
		{ 272, 144, 14 },
		{ 272, 145, 14 },
		{ 272, 146, 14 },
		{ 272, 147, 14 },
		{ 272, 148, 14 },
		{ 272, 149, 14 },
		{ 272, 150, 14 },
		{ 272, 151, 14 },
		{ 272, 152, 14 },
		{ 272, 153, 14 },
		{ 272, 154, 14 },
		{ 272, 155, 14 },
		{ 272, 156, 14 },
		{ 272, 157, 14 },
		{ 272, 158, 14 },
		{ 272, 159, 14 },
		{ 272, 160, 14 },
		{ 272, 161, 14 },
		{ 272, 162, 14 },
		{ 272, 163, 14 },
		{ 272, 164, 14 },
		{ 272, 165, 14 },
		{ 272, 166, 14 },
		{ 272, 167, 14 },
		{ 272, 168, 14 },
		{ 272, 169, 14 },
		{ 272, 170, 14 },
		{ 272, 171, 14 },
		{ 272, 172, 14 },
		{ 272, 173, 14 },
		{ 272, 174, 14 },
		{ 272, 175, 14 },
		{ 272, 176, 14 },
		{ 272, 177, 14 },
		{ 272, 178, 14 },
		{ 272, 179, 14 },
		{ 272, 180, 14 },
		{ 272, 181, 14 },
		{ 272, 182, 14 },
		{ 272, 183, 14 },
		{ 272, 184, 14 },
		{ 272, 185, 14 },
		{ 272, 186, 14 },
		{ 272, 187, 14 },
		{ 272, 188, 14 },
		{ 272, 189, 14 },
		{ 272, 190, 14 },
		{ 272, 191, 14 },
		{ 272, 192, 14 },
		{ 272, 193, 14 },
		{ 272, 194, 14 },
		{ 272, 195, 14 },
		{ 272, 196, 14 },
		{ 272, 318, 14 },
		{ 272, 319, 14 },
		{ 272, 320, 14 },
		{ 272, 321, 14 },
		{ 272, 322, 14 },
		{ 272, 323, 14 },
		{ 272, 324, 14 },
		{ 272, 325, 14 },
		{ 272, 326, 14 },
		{ 272, 327, 14 },
		{ 272, 328, 14 },
		{ 272, 329, 14 },
		{ 272, 330, 14 },
		{ 272, 331, 14 },
		{ 272, 332, 14 },
		{ 272, 333, 14 },
		{ 272, 334, 14 },
		{ 272, 335, 14 },
		{ 272, 336, 14 },
		{ 272, 337, 14 },
		{ 272, 338, 14 },
		{ 272, 339, 14 },
		{ 272, 340, 14 },
		{ 272, 341, 14 },
		{ 272, 342, 14 },
		{ 272, 343, 14 },
		{ 272, 344, 14 },
		{ 272, 345, 14 },
		{ 272, 346, 14 },
		{ 272, 347, 14 },
		{ 272, 348, 14 },
		{ 272, 349, 14 },
		{ 272, 350, 14 },
		{ 272, 351, 14 },
		{ 272, 352, 14 },
		{ 272, 353, 14 },
		{ 272, 354, 14 },
		{ 272, 355, 14 },
		{ 272, 356, 14 },
		{ 272, 357, 14 },
		{ 272, 358, 14 },
		{ 272, 359, 14 },
		{ 272, 360, 14 },
		{ 272, 361, 14 },
		{ 272, 362, 14 },
		{ 272, 363, 14 },
		{ 272, 364, 14 },
		{ 272, 365, 14 },
		{ 272, 366, 14 },
		{ 272, 367, 14 },
		{ 272, 368, 14 },
		{ 272, 369, 14 },
		{ 272, 370, 14 },
		{ 272, 371, 14 },
		{ 272, 372, 14 },
		{ 272, 373, 14 },
		{ 272, 374, 14 },
		{ 272, 375, 14 },
		{ 272, 376, 14 },
		{ 272, 377, 14 },
		{ 272, 378, 14 },
		{ 272, 379, 14 },
		{ 272, 380, 14 },
		{ 272, 381, 14 },
		{ 272, 382, 14 },
		{ 272, 383, 14 },
		{ 273, 131, 14 },
		{ 273, 132, 14 },
		{ 273, 133, 14 },
		{ 273, 134, 14 },
		{ 273, 135, 14 },
		{ 273, 136, 14 },
		{ 273, 137, 14 },
		{ 273, 138, 14 },
		{ 273, 139, 14 },
		{ 273, 140, 14 },
		{ 273, 141, 14 },
		{ 273, 142, 14 },
		{ 273, 143, 14 },
		{ 273, 144, 14 },
		{ 273, 145, 14 },
		{ 273, 146, 14 },
		{ 273, 147, 14 },
		{ 273, 148, 14 },
		{ 273, 149, 14 },
		{ 273, 150, 14 },
		{ 273, 151, 14 },
		{ 273, 152, 14 },
		{ 273, 153, 14 },
		{ 273, 154, 14 },
		{ 273, 155, 14 },
		{ 273, 156, 14 },
		{ 273, 157, 14 },
		{ 273, 158, 14 },
		{ 273, 159, 14 },
		{ 273, 160, 14 },
		{ 273, 161, 14 },
		{ 273, 162, 14 },
		{ 273, 163, 14 },
		{ 273, 164, 14 },
		{ 273, 165, 14 },
		{ 273, 166, 14 },
		{ 273, 167, 14 },
		{ 273, 168, 14 },
		{ 273, 169, 14 },
		{ 273, 170, 14 },
		{ 273, 171, 14 },
		{ 273, 172, 14 },
		{ 273, 173, 14 },
		{ 273, 174, 14 },
		{ 273, 175, 14 },
		{ 273, 176, 14 },
		{ 273, 177, 14 },
		{ 273, 178, 14 },
		{ 273, 179, 14 },
		{ 273, 180, 14 },
		{ 273, 181, 14 },
		{ 273, 182, 14 },
		{ 273, 183, 14 },
		{ 273, 184, 14 },
		{ 273, 185, 14 },
		{ 273, 186, 14 },
		{ 273, 187, 14 },
		{ 273, 188, 14 },
		{ 273, 189, 14 },
		{ 273, 190, 14 },
		{ 273, 191, 14 },
		{ 273, 192, 14 },
		{ 273, 322, 14 },
		{ 273, 323, 14 },
		{ 273, 324, 14 },
		{ 273, 325, 14 },
		{ 273, 326, 14 },
		{ 273, 327, 14 },
		{ 273, 328, 14 },
		{ 273, 329, 14 },
		{ 273, 330, 14 },
		{ 273, 331, 14 },
		{ 273, 332, 14 },
		{ 273, 333, 14 },
		{ 273, 334, 14 },
		{ 273, 335, 14 },
		{ 273, 336, 14 },
		{ 273, 337, 14 },
		{ 273, 338, 14 },
		{ 273, 339, 14 },
		{ 273, 340, 14 },
		{ 273, 341, 14 },
		{ 273, 342, 14 },
		{ 273, 343, 14 },
		{ 273, 344, 14 },
		{ 273, 345, 14 },
		{ 273, 346, 14 },
		{ 273, 347, 14 },
		{ 273, 348, 14 },
		{ 273, 349, 14 },
		{ 273, 350, 14 },
		{ 273, 351, 14 },
		{ 273, 352, 14 },
		{ 273, 353, 14 },
		{ 273, 354, 14 },
		{ 273, 355, 14 },
		{ 273, 356, 14 },
		{ 273, 357, 14 },
		{ 273, 358, 14 },
		{ 273, 359, 14 },
		{ 273, 360, 14 },
		{ 273, 361, 14 },
		{ 273, 362, 14 },
		{ 273, 363, 14 },
		{ 273, 364, 14 },
		{ 273, 365, 14 },
		{ 273, 366, 14 },
		{ 273, 367, 14 },
		{ 273, 368, 14 },
		{ 273, 369, 14 },
		{ 273, 370, 14 },
		{ 273, 371, 14 },
		{ 273, 372, 14 },
		{ 273, 373, 14 },
		{ 273, 374, 14 },
		{ 273, 375, 14 },
		{ 273, 376, 14 },
		{ 273, 377, 14 },
		{ 273, 378, 14 },
		{ 273, 379, 14 },
		{ 273, 380, 14 },
		{ 273, 381, 14 },
		{ 273, 382, 14 },
		{ 273, 383, 14 },
		{ 274, 131, 14 },
		{ 274, 132, 14 },
		{ 274, 133, 14 },
		{ 274, 134, 14 },
		{ 274, 135, 14 },
		{ 274, 136, 14 },
		{ 274, 137, 14 },
		{ 274, 138, 14 },
		{ 274, 139, 14 },
		{ 274, 140, 14 },
		{ 274, 141, 14 },
		{ 274, 142, 14 },
		{ 274, 143, 14 },
		{ 274, 144, 14 },
		{ 274, 145, 14 },
		{ 274, 146, 14 },
		{ 274, 147, 14 },
		{ 274, 148, 14 },
		{ 274, 149, 14 },
		{ 274, 150, 14 },
		{ 274, 151, 14 },
		{ 274, 152, 14 },
		{ 274, 153, 14 },
		{ 274, 154, 14 },
		{ 274, 155, 14 },
		{ 274, 156, 14 },
		{ 274, 157, 14 },
		{ 274, 158, 14 },
		{ 274, 159, 14 },
		{ 274, 160, 14 },
		{ 274, 161, 14 },
		{ 274, 162, 14 },
		{ 274, 163, 14 },
		{ 274, 164, 14 },
		{ 274, 165, 14 },
		{ 274, 166, 14 },
		{ 274, 167, 14 },
		{ 274, 168, 14 },
		{ 274, 169, 14 },
		{ 274, 170, 14 },
		{ 274, 171, 14 },
		{ 274, 172, 14 },
		{ 274, 173, 14 },
		{ 274, 174, 14 },
		{ 274, 175, 14 },
		{ 274, 176, 14 },
		{ 274, 177, 14 },
		{ 274, 178, 14 },
		{ 274, 179, 14 },
		{ 274, 180, 14 },
		{ 274, 181, 14 },
		{ 274, 182, 14 },
		{ 274, 183, 14 },
		{ 274, 184, 14 },
		{ 274, 185, 14 },
		{ 274, 186, 14 },
		{ 274, 187, 14 },
		{ 274, 188, 14 },
		{ 274, 326, 14 },
		{ 274, 327, 14 },
		{ 274, 328, 14 },
		{ 274, 329, 14 },
		{ 274, 330, 14 },
		{ 274, 331, 14 },
		{ 274, 332, 14 },
		{ 274, 333, 14 },
		{ 274, 334, 14 },
		{ 274, 335, 14 },
		{ 274, 336, 14 },
		{ 274, 337, 14 },
		{ 274, 338, 14 },
		{ 274, 339, 14 },
		{ 274, 340, 14 },
		{ 274, 341, 14 },
		{ 274, 342, 14 },
		{ 274, 343, 14 },
		{ 274, 344, 14 },
		{ 274, 345, 14 },
		{ 274, 346, 14 },
		{ 274, 347, 14 },
		{ 274, 348, 14 },
		{ 274, 349, 14 },
		{ 274, 350, 14 },
		{ 274, 351, 14 },
		{ 274, 352, 14 },
		{ 274, 353, 14 },
		{ 274, 354, 14 },
		{ 274, 355, 14 },
		{ 274, 356, 14 },
		{ 274, 357, 14 },
		{ 274, 358, 14 },
		{ 274, 359, 14 },
		{ 274, 360, 14 },
		{ 274, 361, 14 },
		{ 274, 362, 14 },
		{ 274, 363, 14 },
		{ 274, 364, 14 },
		{ 274, 365, 14 },
		{ 274, 366, 14 },
		{ 274, 367, 14 },
		{ 274, 368, 14 },
		{ 274, 369, 14 },
		{ 274, 370, 14 },
		{ 274, 371, 14 },
		{ 274, 372, 14 },
		{ 274, 373, 14 },
		{ 274, 374, 14 },
		{ 274, 375, 14 },
		{ 274, 376, 14 },
		{ 274, 377, 14 },
		{ 274, 378, 14 },
		{ 274, 379, 14 },
		{ 274, 380, 14 },
		{ 274, 381, 14 },
		{ 274, 382, 14 },
		{ 274, 383, 14 },
		{ 275, 131, 14 },
		{ 275, 132, 14 },
		{ 275, 133, 14 },
		{ 275, 134, 14 },
		{ 275, 135, 14 },
		{ 275, 136, 14 },
		{ 275, 137, 14 },
		{ 275, 138, 14 },
		{ 275, 139, 14 },
		{ 275, 140, 14 },
		{ 275, 141, 14 },
		{ 275, 142, 14 },
		{ 275, 143, 14 },
		{ 275, 144, 14 },
		{ 275, 145, 14 },
		{ 275, 146, 14 },
		{ 275, 147, 14 },
		{ 275, 148, 14 },
		{ 275, 149, 14 },
		{ 275, 150, 14 },
		{ 275, 151, 14 },
		{ 275, 152, 14 },
		{ 275, 153, 14 },
		{ 275, 154, 14 },
		{ 275, 155, 14 },
		{ 275, 156, 14 },
		{ 275, 157, 14 },
		{ 275, 158, 14 },
		{ 275, 159, 14 },
		{ 275, 160, 14 },
		{ 275, 161, 14 },
		{ 275, 162, 14 },
		{ 275, 163, 14 },
		{ 275, 164, 14 },
		{ 275, 165, 14 },
		{ 275, 166, 14 },
		{ 275, 167, 14 },
		{ 275, 168, 14 },
		{ 275, 169, 14 },
		{ 275, 170, 14 },
		{ 275, 171, 14 },
		{ 275, 172, 14 },
		{ 275, 173, 14 },
		{ 275, 174, 14 },
		{ 275, 175, 14 },
		{ 275, 176, 14 },
		{ 275, 177, 14 },
		{ 275, 178, 14 },
		{ 275, 179, 14 },
		{ 275, 180, 14 },
		{ 275, 181, 14 },
		{ 275, 182, 14 },
		{ 275, 183, 14 },
		{ 275, 184, 14 },
		{ 275, 330, 14 },
		{ 275, 331, 14 },
		{ 275, 332, 14 },
		{ 275, 333, 14 },
		{ 275, 334, 14 },
		{ 275, 335, 14 },
		{ 275, 336, 14 },
		{ 275, 337, 14 },
		{ 275, 338, 14 },
		{ 275, 339, 14 },
		{ 275, 340, 14 },
		{ 275, 341, 14 },
		{ 275, 342, 14 },
		{ 275, 343, 14 },
		{ 275, 344, 14 },
		{ 275, 345, 14 },
		{ 275, 346, 14 },
		{ 275, 347, 14 },
		{ 275, 348, 14 },
		{ 275, 349, 14 },
		{ 275, 350, 14 },
		{ 275, 351, 14 },
		{ 275, 352, 14 },
		{ 275, 353, 14 },
		{ 275, 354, 14 },
		{ 275, 355, 14 },
		{ 275, 356, 14 },
		{ 275, 357, 14 },
		{ 275, 358, 14 },
		{ 275, 359, 14 },
		{ 275, 360, 14 },
		{ 275, 361, 14 },
		{ 275, 362, 14 },
		{ 275, 363, 14 },
		{ 275, 364, 14 },
		{ 275, 365, 14 },
		{ 275, 366, 14 },
		{ 275, 367, 14 },
		{ 275, 368, 14 },
		{ 275, 369, 14 },
		{ 275, 370, 14 },
		{ 275, 371, 14 },
		{ 275, 372, 14 },
		{ 275, 373, 14 },
		{ 275, 374, 14 },
		{ 275, 375, 14 },
		{ 275, 376, 14 },
		{ 275, 377, 14 },
		{ 275, 378, 14 },
		{ 275, 379, 14 },
		{ 275, 380, 14 },
		{ 275, 381, 14 },
		{ 275, 382, 14 },
		{ 275, 383, 14 },
		{ 276, 131, 14 },
		{ 276, 132, 14 },
		{ 276, 133, 14 },
		{ 276, 134, 14 },
		{ 276, 135, 14 },
		{ 276, 136, 14 },
		{ 276, 137, 14 },
		{ 276, 138, 14 },
		{ 276, 139, 14 },
		{ 276, 140, 14 },
		{ 276, 141, 14 },
		{ 276, 142, 14 },
		{ 276, 143, 14 },
		{ 276, 144, 14 },
		{ 276, 145, 14 },
		{ 276, 146, 14 },
		{ 276, 147, 14 },
		{ 276, 148, 14 },
		{ 276, 149, 14 },
		{ 276, 150, 14 },
		{ 276, 151, 14 },
		{ 276, 152, 14 },
		{ 276, 153, 14 },
		{ 276, 154, 14 },
		{ 276, 155, 14 },
		{ 276, 156, 14 },
		{ 276, 157, 14 },
		{ 276, 158, 14 },
		{ 276, 159, 14 },
		{ 276, 160, 14 },
		{ 276, 161, 14 },
		{ 276, 162, 14 },
		{ 276, 163, 14 },
		{ 276, 164, 14 },
		{ 276, 165, 14 },
		{ 276, 166, 14 },
		{ 276, 167, 14 },
		{ 276, 168, 14 },
		{ 276, 169, 14 },
		{ 276, 170, 14 },
		{ 276, 171, 14 },
		{ 276, 172, 14 },
		{ 276, 173, 14 },
		{ 276, 174, 14 },
		{ 276, 175, 14 },
		{ 276, 176, 14 },
		{ 276, 177, 14 },
		{ 276, 178, 14 },
		{ 276, 179, 14 },
		{ 276, 180, 14 },
		{ 276, 334, 14 },
		{ 276, 335, 14 },
		{ 276, 336, 14 },
		{ 276, 337, 14 },
		{ 276, 338, 14 },
		{ 276, 339, 14 },
		{ 276, 340, 14 },
		{ 276, 341, 14 },
		{ 276, 342, 14 },
		{ 276, 343, 14 },
		{ 276, 344, 14 },
		{ 276, 345, 14 },
		{ 276, 346, 14 },
		{ 276, 347, 14 },
		{ 276, 348, 14 },
		{ 276, 349, 14 },
		{ 276, 350, 14 },
		{ 276, 351, 14 },
		{ 276, 352, 14 },
		{ 276, 353, 14 },
		{ 276, 354, 14 },
		{ 276, 355, 14 },
		{ 276, 356, 14 },
		{ 276, 357, 14 },
		{ 276, 358, 14 },
		{ 276, 359, 14 },
		{ 276, 360, 14 },
		{ 276, 361, 14 },
		{ 276, 362, 14 },
		{ 276, 363, 14 },
		{ 276, 364, 14 },
		{ 276, 365, 14 },
		{ 276, 366, 14 },
		{ 276, 367, 14 },
		{ 276, 368, 14 },
		{ 276, 369, 14 },
		{ 276, 370, 14 },
		{ 276, 371, 14 },
		{ 276, 372, 14 },
		{ 276, 373, 14 },
		{ 276, 374, 14 },
		{ 276, 375, 14 },
		{ 276, 376, 14 },
		{ 276, 377, 14 },
		{ 276, 378, 14 },
		{ 276, 379, 14 },
		{ 276, 380, 14 },
		{ 276, 381, 14 },
		{ 276, 382, 14 },
		{ 276, 383, 14 },
		{ 277, 131, 14 },
		{ 277, 132, 14 },
		{ 277, 133, 14 },
		{ 277, 134, 14 },
		{ 277, 135, 14 },
		{ 277, 136, 14 },
		{ 277, 137, 14 },
		{ 277, 138, 14 },
		{ 277, 139, 14 },
		{ 277, 140, 14 },
		{ 277, 141, 14 },
		{ 277, 142, 14 },
		{ 277, 143, 14 },
		{ 277, 144, 14 },
		{ 277, 145, 14 },
		{ 277, 146, 14 },
		{ 277, 147, 14 },
		{ 277, 148, 14 },
		{ 277, 149, 14 },
		{ 277, 150, 14 },
		{ 277, 151, 14 },
		{ 277, 152, 14 },
		{ 277, 153, 14 },
		{ 277, 154, 14 },
		{ 277, 155, 14 },
		{ 277, 156, 14 },
		{ 277, 157, 14 },
		{ 277, 158, 14 },
		{ 277, 159, 14 },
		{ 277, 160, 14 },
		{ 277, 161, 14 },
		{ 277, 162, 14 },
		{ 277, 163, 14 },
		{ 277, 164, 14 },
		{ 277, 165, 14 },
		{ 277, 166, 14 },
		{ 277, 167, 14 },
		{ 277, 168, 14 },
		{ 277, 169, 14 },
		{ 277, 170, 14 },
		{ 277, 171, 14 },
		{ 277, 172, 14 },
		{ 277, 173, 14 },
		{ 277, 174, 14 },
		{ 277, 175, 14 },
		{ 277, 176, 14 },
		{ 277, 338, 14 },
		{ 277, 339, 14 },
		{ 277, 340, 14 },
		{ 277, 341, 14 },
		{ 277, 342, 14 },
		{ 277, 343, 14 },
		{ 277, 344, 14 },
		{ 277, 345, 14 },
		{ 277, 346, 14 },
		{ 277, 347, 14 },
		{ 277, 348, 14 },
		{ 277, 349, 14 },
		{ 277, 350, 14 },
		{ 277, 351, 14 },
		{ 277, 352, 14 },
		{ 277, 353, 14 },
		{ 277, 354, 14 },
		{ 277, 355, 14 },
		{ 277, 356, 14 },
		{ 277, 357, 14 },
		{ 277, 358, 14 },
		{ 277, 359, 14 },
		{ 277, 360, 14 },
		{ 277, 361, 14 },
		{ 277, 362, 14 },
		{ 277, 363, 14 },
		{ 277, 364, 14 },
		{ 277, 365, 14 },
		{ 277, 366, 14 },
		{ 277, 367, 14 },
		{ 277, 368, 14 },
		{ 277, 369, 14 },
		{ 277, 370, 14 },
		{ 277, 371, 14 },
		{ 277, 372, 14 },
		{ 277, 373, 14 },
		{ 277, 374, 14 },
		{ 277, 375, 14 },
		{ 277, 376, 14 },
		{ 277, 377, 14 },
		{ 277, 378, 14 },
		{ 277, 379, 14 },
		{ 277, 380, 14 },
		{ 277, 381, 14 },
		{ 277, 382, 14 },
		{ 277, 383, 14 },
		{ 278, 131, 14 },
		{ 278, 132, 14 },
		{ 278, 133, 14 },
		{ 278, 134, 14 },
		{ 278, 135, 14 },
		{ 278, 136, 14 },
		{ 278, 137, 14 },
		{ 278, 138, 14 },
		{ 278, 139, 14 },
		{ 278, 140, 14 },
		{ 278, 141, 14 },
		{ 278, 142, 14 },
		{ 278, 143, 14 },
		{ 278, 144, 14 },
		{ 278, 145, 14 },
		{ 278, 146, 14 },
		{ 278, 147, 14 },
		{ 278, 148, 14 },
		{ 278, 149, 14 },
		{ 278, 150, 14 },
		{ 278, 151, 14 },
		{ 278, 152, 14 },
		{ 278, 153, 14 },
		{ 278, 154, 14 },
		{ 278, 155, 14 },
		{ 278, 156, 14 },
		{ 278, 157, 14 },
		{ 278, 158, 14 },
		{ 278, 159, 14 },
		{ 278, 160, 14 },
		{ 278, 161, 14 },
		{ 278, 162, 14 },
		{ 278, 163, 14 },
		{ 278, 164, 14 },
		{ 278, 165, 14 },
		{ 278, 166, 14 },
		{ 278, 167, 14 },
		{ 278, 168, 14 },
		{ 278, 169, 14 },
		{ 278, 170, 14 },
		{ 278, 171, 14 },
		{ 278, 172, 14 },
		{ 278, 342, 14 },
		{ 278, 343, 14 },
		{ 278, 344, 14 },
		{ 278, 345, 14 },
		{ 278, 346, 14 },
		{ 278, 347, 14 },
		{ 278, 348, 14 },
		{ 278, 349, 14 },
		{ 278, 350, 14 },
		{ 278, 351, 14 },
		{ 278, 352, 14 },
		{ 278, 353, 14 },
		{ 278, 354, 14 },
		{ 278, 355, 14 },
		{ 278, 356, 14 },
		{ 278, 357, 14 },
		{ 278, 358, 14 },
		{ 278, 359, 14 },
		{ 278, 360, 14 },
		{ 278, 361, 14 },
		{ 278, 362, 14 },
		{ 278, 363, 14 },
		{ 278, 364, 14 },
		{ 278, 365, 14 },
		{ 278, 366, 14 },
		{ 278, 367, 14 },
		{ 278, 368, 14 },
		{ 278, 369, 14 },
		{ 278, 370, 14 },
		{ 278, 371, 14 },
		{ 278, 372, 14 },
		{ 278, 373, 14 },
		{ 278, 374, 14 },
		{ 278, 375, 14 },
		{ 278, 376, 14 },
		{ 278, 377, 14 },
		{ 278, 378, 14 },
		{ 278, 379, 14 },
		{ 278, 380, 14 },
		{ 278, 381, 14 },
		{ 278, 382, 14 },
		{ 278, 383, 14 },
		{ 279, 131, 14 },
		{ 279, 132, 14 },
		{ 279, 133, 14 },
		{ 279, 134, 14 },
		{ 279, 135, 14 },
		{ 279, 136, 14 },
		{ 279, 137, 14 },
		{ 279, 138, 14 },
		{ 279, 139, 14 },
		{ 279, 140, 14 },
		{ 279, 141, 14 },
		{ 279, 142, 14 },
		{ 279, 143, 14 },
		{ 279, 144, 14 },
		{ 279, 145, 14 },
		{ 279, 146, 14 },
		{ 279, 147, 14 },
		{ 279, 148, 14 },
		{ 279, 149, 14 },
		{ 279, 150, 14 },
		{ 279, 151, 14 },
		{ 279, 152, 14 },
		{ 279, 153, 14 },
		{ 279, 154, 14 },
		{ 279, 155, 14 },
		{ 279, 156, 14 },
		{ 279, 157, 14 },
		{ 279, 158, 14 },
		{ 279, 159, 14 },
		{ 279, 160, 14 },
		{ 279, 161, 14 },
		{ 279, 162, 14 },
		{ 279, 163, 14 },
		{ 279, 164, 14 },
		{ 279, 165, 14 },
		{ 279, 166, 14 },
		{ 279, 167, 14 },
		{ 279, 168, 14 },
		{ 279, 346, 14 },
		{ 279, 347, 14 },
		{ 279, 348, 14 },
		{ 279, 349, 14 },
		{ 279, 350, 14 },
		{ 279, 351, 14 },
		{ 279, 352, 14 },
		{ 279, 353, 14 },
		{ 279, 354, 14 },
		{ 279, 355, 14 },
		{ 279, 356, 14 },
		{ 279, 357, 14 },
		{ 279, 358, 14 },
		{ 279, 359, 14 },
		{ 279, 360, 14 },
		{ 279, 361, 14 },
		{ 279, 362, 14 },
		{ 279, 363, 14 },
		{ 279, 364, 14 },
		{ 279, 365, 14 },
		{ 279, 366, 14 },
		{ 279, 367, 14 },
		{ 279, 368, 14 },
		{ 279, 369, 14 },
		{ 279, 370, 14 },
		{ 279, 371, 14 },
		{ 279, 372, 14 },
		{ 279, 373, 14 },
		{ 279, 374, 14 },
		{ 279, 375, 14 },
		{ 279, 376, 14 },
		{ 279, 377, 14 },
		{ 279, 378, 14 },
		{ 279, 379, 14 },
		{ 279, 380, 14 },
		{ 279, 381, 14 },
		{ 279, 382, 14 },
		{ 279, 383, 14 },
		{ 280, 131, 14 },
		{ 280, 132, 14 },
		{ 280, 133, 14 },
		{ 280, 134, 14 },
		{ 280, 135, 14 },
		{ 280, 136, 14 },
		{ 280, 137, 14 },
		{ 280, 138, 14 },
		{ 280, 139, 14 },
		{ 280, 140, 14 },
		{ 280, 141, 14 },
		{ 280, 142, 14 },
		{ 280, 143, 14 },
		{ 280, 144, 14 },
		{ 280, 145, 14 },
		{ 280, 146, 14 },
		{ 280, 147, 14 },
		{ 280, 148, 14 },
		{ 280, 149, 14 },
		{ 280, 150, 14 },
		{ 280, 151, 14 },
		{ 280, 152, 14 },
		{ 280, 153, 14 },
		{ 280, 154, 14 },
		{ 280, 155, 14 },
		{ 280, 156, 14 },
		{ 280, 157, 14 },
		{ 280, 158, 14 },
		{ 280, 159, 14 },
		{ 280, 160, 14 },
		{ 280, 161, 14 },
		{ 280, 162, 14 },
		{ 280, 163, 14 },
		{ 280, 164, 14 },
		{ 280, 350, 14 },
		{ 280, 351, 14 },
		{ 280, 352, 14 },
		{ 280, 353, 14 },
		{ 280, 354, 14 },
		{ 280, 355, 14 },
		{ 280, 356, 14 },
		{ 280, 357, 14 },
		{ 280, 358, 14 },
		{ 280, 359, 14 },
		{ 280, 360, 14 },
		{ 280, 361, 14 },
		{ 280, 362, 14 },
		{ 280, 363, 14 },
		{ 280, 364, 14 },
		{ 280, 365, 14 },
		{ 280, 366, 14 },
		{ 280, 367, 14 },
		{ 280, 368, 14 },
		{ 280, 369, 14 },
		{ 280, 370, 14 },
		{ 280, 371, 14 },
		{ 280, 372, 14 },
		{ 280, 373, 14 },
		{ 280, 374, 14 },
		{ 280, 375, 14 },
		{ 280, 376, 14 },
		{ 280, 377, 14 },
		{ 280, 378, 14 },
		{ 280, 379, 14 },
		{ 280, 380, 14 },
		{ 280, 381, 14 },
		{ 280, 382, 14 },
		{ 280, 383, 14 },
		{ 281, 131, 14 },
		{ 281, 132, 14 },
		{ 281, 133, 14 },
		{ 281, 134, 14 },
		{ 281, 135, 14 },
		{ 281, 136, 14 },
		{ 281, 137, 14 },
		{ 281, 138, 14 },
		{ 281, 139, 14 },
		{ 281, 140, 14 },
		{ 281, 141, 14 },
		{ 281, 142, 14 },
		{ 281, 143, 14 },
		{ 281, 144, 14 },
		{ 281, 145, 14 },
		{ 281, 146, 14 },
		{ 281, 147, 14 },
		{ 281, 148, 14 },
		{ 281, 149, 14 },
		{ 281, 150, 14 },
		{ 281, 151, 14 },
		{ 281, 152, 14 },
		{ 281, 153, 14 },
		{ 281, 154, 14 },
		{ 281, 155, 14 },
		{ 281, 156, 14 },
		{ 281, 157, 14 },
		{ 281, 158, 14 },
		{ 281, 159, 14 },
		{ 281, 160, 14 },
		{ 281, 354, 14 },
		{ 281, 355, 14 },
		{ 281, 356, 14 },
		{ 281, 357, 14 },
		{ 281, 358, 14 },
		{ 281, 359, 14 },
		{ 281, 360, 14 },
		{ 281, 361, 14 },
		{ 281, 362, 14 },
		{ 281, 363, 14 },
		{ 281, 364, 14 },
		{ 281, 365, 14 },
		{ 281, 366, 14 },
		{ 281, 367, 14 },
		{ 281, 368, 14 },
		{ 281, 369, 14 },
		{ 281, 370, 14 },
		{ 281, 371, 14 },
		{ 281, 372, 14 },
		{ 281, 373, 14 },
		{ 281, 374, 14 },
		{ 281, 375, 14 },
		{ 281, 376, 14 },
		{ 281, 377, 14 },
		{ 281, 378, 14 },
		{ 281, 379, 14 },
		{ 281, 380, 14 },
		{ 281, 381, 14 },
		{ 281, 382, 14 },
		{ 281, 383, 14 },
		{ 282, 132, 14 },
		{ 282, 133, 14 },
		{ 282, 134, 14 },
		{ 282, 135, 14 },
		{ 282, 136, 14 },
		{ 282, 137, 14 },
		{ 282, 138, 14 },
		{ 282, 139, 14 },
		{ 282, 140, 14 },
		{ 282, 141, 14 },
		{ 282, 142, 14 },
		{ 282, 143, 14 },
		{ 282, 144, 14 },
		{ 282, 145, 14 },
		{ 282, 146, 14 },
		{ 282, 147, 14 },
		{ 282, 148, 14 },
		{ 282, 149, 14 },
		{ 282, 150, 14 },
		{ 282, 151, 14 },
		{ 282, 152, 14 },
		{ 282, 153, 14 },
		{ 282, 154, 14 },
		{ 282, 155, 14 },
		{ 282, 156, 14 },
		{ 282, 358, 14 },
		{ 282, 359, 14 },
		{ 282, 360, 14 },
		{ 282, 361, 14 },
		{ 282, 362, 14 },
		{ 282, 363, 14 },
		{ 282, 364, 14 },
		{ 282, 365, 14 },
		{ 282, 366, 14 },
		{ 282, 367, 14 },
		{ 282, 368, 14 },
		{ 282, 369, 14 },
		{ 282, 370, 14 },
		{ 282, 371, 14 },
		{ 282, 372, 14 },
		{ 282, 373, 14 },
		{ 282, 374, 14 },
		{ 282, 375, 14 },
		{ 282, 376, 14 },
		{ 282, 377, 14 },
		{ 282, 378, 14 },
		{ 282, 379, 14 },
		{ 282, 380, 14 },
		{ 282, 381, 14 },
		{ 282, 382, 14 },
		{ 283, 132, 14 },
		{ 283, 133, 14 },
		{ 283, 134, 14 },
		{ 283, 135, 14 },
		{ 283, 136, 14 },
		{ 283, 137, 14 },
		{ 283, 138, 14 },
		{ 283, 139, 14 },
		{ 283, 140, 14 },
		{ 283, 141, 14 },
		{ 283, 142, 14 },
		{ 283, 143, 14 },
		{ 283, 144, 14 },
		{ 283, 145, 14 },
		{ 283, 146, 14 },
		{ 283, 147, 14 },
		{ 283, 148, 14 },
		{ 283, 149, 14 },
		{ 283, 150, 14 },
		{ 283, 151, 14 },
		{ 283, 152, 14 },
		{ 283, 362, 14 },
		{ 283, 363, 14 },
		{ 283, 364, 14 },
		{ 283, 365, 14 },
		{ 283, 366, 14 },
		{ 283, 367, 14 },
		{ 283, 368, 14 },
		{ 283, 369, 14 },
		{ 283, 370, 14 },
		{ 283, 371, 14 },
		{ 283, 372, 14 },
		{ 283, 373, 14 },
		{ 283, 374, 14 },
		{ 283, 375, 14 },
		{ 283, 376, 14 },
		{ 283, 377, 14 },
		{ 283, 378, 14 },
		{ 283, 379, 14 },
		{ 283, 380, 14 },
		{ 283, 381, 14 },
		{ 283, 382, 14 },
		{ 284, 132, 14 },
		{ 284, 133, 14 },
		{ 284, 134, 14 },
		{ 284, 135, 14 },
		{ 284, 136, 14 },
		{ 284, 137, 14 },
		{ 284, 138, 14 },
		{ 284, 139, 14 },
		{ 284, 140, 14 },
		{ 284, 141, 14 },
		{ 284, 142, 14 },
		{ 284, 143, 14 },
		{ 284, 144, 14 },
		{ 284, 145, 14 },
		{ 284, 146, 14 },
		{ 284, 147, 14 },
		{ 284, 148, 14 },
		{ 284, 366, 14 },
		{ 284, 367, 14 },
		{ 284, 368, 14 },
		{ 284, 369, 14 },
		{ 284, 370, 14 },
		{ 284, 371, 14 },
		{ 284, 372, 14 },
		{ 284, 373, 14 },
		{ 284, 374, 14 },
		{ 284, 375, 14 },
		{ 284, 376, 14 },
		{ 284, 377, 14 },
		{ 284, 378, 14 },
		{ 284, 379, 14 },
		{ 284, 380, 14 },
		{ 284, 381, 14 },
		{ 284, 382, 14 },
		{ 285, 132, 14 },
		{ 285, 133, 14 },
		{ 285, 134, 14 },
		{ 285, 135, 14 },
		{ 285, 136, 14 },
		{ 285, 137, 14 },
		{ 285, 138, 14 },
		{ 285, 139, 14 },
		{ 285, 140, 14 },
		{ 285, 141, 14 },
		{ 285, 142, 14 },
		{ 285, 143, 14 },
		{ 285, 144, 14 },
		{ 285, 370, 14 },
		{ 285, 371, 14 },
		{ 285, 372, 14 },
		{ 285, 373, 14 },
		{ 285, 374, 14 },
		{ 285, 375, 14 },
		{ 285, 376, 14 },
		{ 285, 377, 14 },
		{ 285, 378, 14 },
		{ 285, 379, 14 },
		{ 285, 380, 14 },
		{ 285, 381, 14 },
		{ 285, 382, 14 },
		{ 286, 133, 14 },
		{ 286, 134, 14 },
		{ 286, 135, 14 },
		{ 286, 136, 14 },
		{ 286, 137, 14 },
		{ 286, 138, 14 },
		{ 286, 139, 14 },
		{ 286, 140, 14 },
		{ 286, 374, 14 },
		{ 286, 375, 14 },
		{ 286, 376, 14 },
		{ 286, 377, 14 },
		{ 286, 378, 14 },
		{ 286, 379, 14 },
		{ 286, 380, 14 },
		{ 286, 381, 14 },
		{ 197, 379, 15 },
		{ 197, 380, 15 },
		{ 197, 381, 15 },
		{ 198, 376, 15 },
		{ 198, 377, 15 },
		{ 198, 378, 15 },
		{ 198, 379, 15 },
		{ 198, 380, 15 },
		{ 198, 381, 15 },
		{ 198, 382, 15 },
		{ 199, 374, 15 },
		{ 199, 375, 15 },
		{ 199, 376, 15 },
		{ 199, 377, 15 },
		{ 199, 378, 15 },
		{ 199, 379, 15 },
		{ 199, 380, 15 },
		{ 199, 381, 15 },
		{ 199, 382, 15 },
		{ 200, 372, 15 },
		{ 200, 373, 15 },
		{ 200, 374, 15 },
		{ 200, 375, 15 },
		{ 200, 376, 15 },
		{ 200, 377, 15 },
		{ 200, 378, 15 },
		{ 200, 379, 15 },
		{ 200, 380, 15 },
		{ 200, 381, 15 },
		{ 200, 382, 15 },
		{ 201, 370, 15 },
		{ 201, 371, 15 },
		{ 201, 372, 15 },
		{ 201, 373, 15 },
		{ 201, 374, 15 },
		{ 201, 375, 15 },
		{ 201, 376, 15 },
		{ 201, 377, 15 },
		{ 201, 378, 15 },
		{ 201, 379, 15 },
		{ 201, 380, 15 },
		{ 201, 381, 15 },
		{ 201, 382, 15 },
		{ 201, 383, 15 },
		{ 202, 368, 15 },
		{ 202, 369, 15 },
		{ 202, 370, 15 },
		{ 202, 371, 15 },
		{ 202, 372, 15 },
		{ 202, 373, 15 },
		{ 202, 374, 15 },
		{ 202, 375, 15 },
		{ 202, 376, 15 },
		{ 202, 377, 15 },
		{ 202, 378, 15 },
		{ 202, 379, 15 },
		{ 202, 380, 15 },
		{ 202, 381, 15 },
		{ 202, 382, 15 },
		{ 202, 383, 15 },
		{ 203, 366, 15 },
		{ 203, 367, 15 },
		{ 203, 368, 15 },
		{ 203, 369, 15 },
		{ 203, 370, 15 },
		{ 203, 371, 15 },
		{ 203, 372, 15 },
		{ 203, 373, 15 },
		{ 203, 374, 15 },
		{ 203, 375, 15 },
		{ 203, 376, 15 },
		{ 203, 377, 15 },
		{ 203, 378, 15 },
		{ 203, 379, 15 },
		{ 203, 380, 15 },
		{ 203, 381, 15 },
		{ 203, 382, 15 },
		{ 203, 383, 15 },
		{ 204, 364, 15 },
		{ 204, 365, 15 },
		{ 204, 366, 15 },
		{ 204, 367, 15 },
		{ 204, 368, 15 },
		{ 204, 369, 15 },
		{ 204, 370, 15 },
		{ 204, 371, 15 },
		{ 204, 372, 15 },
		{ 204, 373, 15 },
		{ 204, 374, 15 },
		{ 204, 375, 15 },
		{ 204, 376, 15 },
		{ 204, 377, 15 },
		{ 204, 378, 15 },
		{ 204, 379, 15 },
		{ 204, 380, 15 },
		{ 204, 381, 15 },
		{ 204, 382, 15 },
		{ 204, 383, 15 },
		{ 205, 362, 15 },
		{ 205, 363, 15 },
		{ 205, 364, 15 },
		{ 205, 365, 15 },
		{ 205, 366, 15 },
		{ 205, 367, 15 },
		{ 205, 368, 15 },
		{ 205, 369, 15 },
		{ 205, 370, 15 },
		{ 205, 371, 15 },
		{ 205, 372, 15 },
		{ 205, 373, 15 },
		{ 205, 374, 15 },
		{ 205, 375, 15 },
		{ 205, 376, 15 },
		{ 205, 377, 15 },
		{ 205, 378, 15 },
		{ 205, 379, 15 },
		{ 205, 380, 15 },
		{ 205, 381, 15 },
		{ 205, 382, 15 },
		{ 205, 383, 15 },
		{ 206, 360, 15 },
		{ 206, 361, 15 },
		{ 206, 362, 15 },
		{ 206, 363, 15 },
		{ 206, 364, 15 },
		{ 206, 365, 15 },
		{ 206, 366, 15 },
		{ 206, 367, 15 },
		{ 206, 368, 15 },
		{ 206, 369, 15 },
		{ 206, 370, 15 },
		{ 206, 371, 15 },
		{ 206, 372, 15 },
		{ 206, 373, 15 },
		{ 206, 374, 15 },
		{ 206, 375, 15 },
		{ 206, 376, 15 },
		{ 206, 377, 15 },
		{ 206, 378, 15 },
		{ 206, 379, 15 },
		{ 206, 380, 15 },
		{ 206, 381, 15 },
		{ 206, 382, 15 },
		{ 206, 383, 15 },
		{ 207, 358, 15 },
		{ 207, 359, 15 },
		{ 207, 360, 15 },
		{ 207, 361, 15 },
		{ 207, 362, 15 },
		{ 207, 363, 15 },
		{ 207, 364, 15 },
		{ 207, 365, 15 },
		{ 207, 366, 15 },
		{ 207, 367, 15 },
		{ 207, 368, 15 },
		{ 207, 369, 15 },
		{ 207, 370, 15 },
		{ 207, 371, 15 },
		{ 207, 372, 15 },
		{ 207, 373, 15 },
		{ 207, 374, 15 },
		{ 207, 375, 15 },
		{ 207, 376, 15 },
		{ 207, 377, 15 },
		{ 207, 378, 15 },
		{ 207, 379, 15 },
		{ 207, 380, 15 },
		{ 207, 381, 15 },
		{ 207, 382, 15 },
		{ 207, 383, 15 },
		{ 208, 356, 15 },
		{ 208, 357, 15 },
		{ 208, 358, 15 },
		{ 208, 359, 15 },
		{ 208, 360, 15 },
		{ 208, 361, 15 },
		{ 208, 362, 15 },
		{ 208, 363, 15 },
		{ 208, 364, 15 },
		{ 208, 365, 15 },
		{ 208, 366, 15 },
		{ 208, 367, 15 },
		{ 208, 368, 15 },
		{ 208, 369, 15 },
		{ 208, 370, 15 },
		{ 208, 371, 15 },
		{ 208, 372, 15 },
		{ 208, 373, 15 },
		{ 208, 374, 15 },
		{ 208, 375, 15 },
		{ 208, 376, 15 },
		{ 208, 377, 15 },
		{ 208, 378, 15 },
		{ 208, 379, 15 },
		{ 208, 380, 15 },
		{ 208, 381, 15 },
		{ 208, 382, 15 },
		{ 208, 383, 15 },
		{ 209, 354, 15 },
		{ 209, 355, 15 },
		{ 209, 356, 15 },
		{ 209, 357, 15 },
		{ 209, 358, 15 },
		{ 209, 359, 15 },
		{ 209, 360, 15 },
		{ 209, 361, 15 },
		{ 209, 362, 15 },
		{ 209, 363, 15 },
		{ 209, 364, 15 },
		{ 209, 365, 15 },
		{ 209, 366, 15 },
		{ 209, 367, 15 },
		{ 209, 368, 15 },
		{ 209, 369, 15 },
		{ 209, 370, 15 },
		{ 209, 371, 15 },
		{ 209, 372, 15 },
		{ 209, 373, 15 },
		{ 209, 374, 15 },
		{ 209, 375, 15 },
		{ 209, 376, 15 },
		{ 209, 377, 15 },
		{ 209, 378, 15 },
		{ 209, 379, 15 },
		{ 209, 380, 15 },
		{ 209, 381, 15 },
		{ 209, 382, 15 },
		{ 209, 383, 15 },
		{ 210, 352, 15 },
		{ 210, 353, 15 },
		{ 210, 354, 15 },
		{ 210, 355, 15 },
		{ 210, 356, 15 },
		{ 210, 357, 15 },
		{ 210, 358, 15 },
		{ 210, 359, 15 },
		{ 210, 360, 15 },
		{ 210, 361, 15 },
		{ 210, 362, 15 },
		{ 210, 363, 15 },
		{ 210, 364, 15 },
		{ 210, 365, 15 },
		{ 210, 366, 15 },
		{ 210, 367, 15 },
		{ 210, 368, 15 },
		{ 210, 369, 15 },
		{ 210, 370, 15 },
		{ 210, 371, 15 },
		{ 210, 372, 15 },
		{ 210, 373, 15 },
		{ 210, 374, 15 },
		{ 210, 375, 15 },
		{ 210, 376, 15 },
		{ 210, 377, 15 },
		{ 210, 378, 15 },
		{ 210, 379, 15 },
		{ 210, 380, 15 },
		{ 210, 381, 15 },
		{ 210, 382, 15 },
		{ 210, 383, 15 },
		{ 211, 350, 15 },
		{ 211, 351, 15 },
		{ 211, 352, 15 },
		{ 211, 353, 15 },
		{ 211, 354, 15 },
		{ 211, 355, 15 },
		{ 211, 356, 15 },
		{ 211, 357, 15 },
		{ 211, 358, 15 },
		{ 211, 359, 15 },
		{ 211, 360, 15 },
		{ 211, 361, 15 },
		{ 211, 362, 15 },
		{ 211, 363, 15 },
		{ 211, 364, 15 },
		{ 211, 365, 15 },
		{ 211, 366, 15 },
		{ 211, 367, 15 },
		{ 211, 368, 15 },
		{ 211, 369, 15 },
		{ 211, 370, 15 },
		{ 211, 371, 15 },
		{ 211, 372, 15 },
		{ 211, 373, 15 },
		{ 211, 374, 15 },
		{ 211, 375, 15 },
		{ 211, 376, 15 },
		{ 211, 377, 15 },
		{ 211, 378, 15 },
		{ 211, 379, 15 },
		{ 211, 380, 15 },
		{ 211, 381, 15 },
		{ 211, 382, 15 },
		{ 211, 383, 15 },
		{ 212, 348, 15 },
		{ 212, 349, 15 },
		{ 212, 350, 15 },
		{ 212, 351, 15 },
		{ 212, 352, 15 },
		{ 212, 353, 15 },
		{ 212, 354, 15 },
		{ 212, 355, 15 },
		{ 212, 356, 15 },
		{ 212, 357, 15 },
		{ 212, 358, 15 },
		{ 212, 359, 15 },
		{ 212, 360, 15 },
		{ 212, 361, 15 },
		{ 212, 362, 15 },
		{ 212, 363, 15 },
		{ 212, 364, 15 },
		{ 212, 365, 15 },
		{ 212, 366, 15 },
		{ 212, 367, 15 },
		{ 212, 368, 15 },
		{ 212, 369, 15 },
		{ 212, 370, 15 },
		{ 212, 371, 15 },
		{ 212, 372, 15 },
		{ 212, 373, 15 },
		{ 212, 374, 15 },
		{ 212, 375, 15 },
		{ 212, 376, 15 },
		{ 212, 377, 15 },
		{ 212, 378, 15 },
		{ 212, 379, 15 },
		{ 212, 380, 15 },
		{ 212, 381, 15 },
		{ 212, 382, 15 },
		{ 212, 383, 15 },
		{ 213, 346, 15 },
		{ 213, 347, 15 },
		{ 213, 348, 15 },
		{ 213, 349, 15 },
		{ 213, 350, 15 },
		{ 213, 351, 15 },
		{ 213, 352, 15 },
		{ 213, 353, 15 },
		{ 213, 354, 15 },
		{ 213, 355, 15 },
		{ 213, 356, 15 },
		{ 213, 357, 15 },
		{ 213, 358, 15 },
		{ 213, 359, 15 },
		{ 213, 360, 15 },
		{ 213, 361, 15 },
		{ 213, 362, 15 },
		{ 213, 363, 15 },
		{ 213, 364, 15 },
		{ 213, 365, 15 },
		{ 213, 366, 15 },
		{ 213, 367, 15 },
		{ 213, 368, 15 },
		{ 213, 369, 15 },
		{ 213, 370, 15 },
		{ 213, 371, 15 },
		{ 213, 372, 15 },
		{ 213, 373, 15 },
		{ 213, 374, 15 },
		{ 213, 375, 15 },
		{ 213, 376, 15 },
		{ 213, 377, 15 },
		{ 213, 378, 15 },
		{ 213, 379, 15 },
		{ 213, 380, 15 },
		{ 213, 381, 15 },
		{ 213, 382, 15 },
		{ 213, 383, 15 },
		{ 214, 344, 15 },
		{ 214, 345, 15 },
		{ 214, 346, 15 },
		{ 214, 347, 15 },
		{ 214, 348, 15 },
		{ 214, 349, 15 },
		{ 214, 350, 15 },
		{ 214, 351, 15 },
		{ 214, 352, 15 },
		{ 214, 353, 15 },
		{ 214, 354, 15 },
		{ 214, 355, 15 },
		{ 214, 356, 15 },
		{ 214, 357, 15 },
		{ 214, 358, 15 },
		{ 214, 359, 15 },
		{ 214, 360, 15 },
		{ 214, 361, 15 },
		{ 214, 362, 15 },
		{ 214, 363, 15 },
		{ 214, 364, 15 },
		{ 214, 365, 15 },
		{ 214, 366, 15 },
		{ 214, 367, 15 },
		{ 214, 368, 15 },
		{ 214, 369, 15 },
		{ 214, 370, 15 },
		{ 214, 371, 15 },
		{ 214, 372, 15 },
		{ 214, 373, 15 },
		{ 214, 374, 15 },
		{ 214, 375, 15 },
		{ 214, 376, 15 },
		{ 214, 377, 15 },
		{ 214, 378, 15 },
		{ 214, 379, 15 },
		{ 214, 380, 15 },
		{ 214, 381, 15 },
		{ 214, 382, 15 },
		{ 214, 383, 15 },
		{ 215, 342, 15 },
		{ 215, 343, 15 },
		{ 215, 344, 15 },
		{ 215, 345, 15 },
		{ 215, 346, 15 },
		{ 215, 347, 15 },
		{ 215, 348, 15 },
		{ 215, 349, 15 },
		{ 215, 350, 15 },
		{ 215, 351, 15 },
		{ 215, 352, 15 },
		{ 215, 353, 15 },
		{ 215, 354, 15 },
		{ 215, 355, 15 },
		{ 215, 356, 15 },
		{ 215, 357, 15 },
		{ 215, 358, 15 },
		{ 215, 359, 15 },
		{ 215, 360, 15 },
		{ 215, 361, 15 },
		{ 215, 362, 15 },
		{ 215, 363, 15 },
		{ 215, 364, 15 },
		{ 215, 365, 15 },
		{ 215, 366, 15 },
		{ 215, 367, 15 },
		{ 215, 368, 15 },
		{ 215, 369, 15 },
		{ 215, 370, 15 },
		{ 215, 371, 15 },
		{ 215, 372, 15 },
		{ 215, 373, 15 },
		{ 215, 374, 15 },
		{ 215, 375, 15 },
		{ 215, 376, 15 },
		{ 215, 377, 15 },
		{ 215, 378, 15 },
		{ 215, 379, 15 },
		{ 215, 380, 15 },
		{ 215, 381, 15 },
		{ 215, 382, 15 },
		{ 215, 383, 15 },
		{ 216, 340, 15 },
		{ 216, 341, 15 },
		{ 216, 342, 15 },
		{ 216, 343, 15 },
		{ 216, 344, 15 },
		{ 216, 345, 15 },
		{ 216, 346, 15 },
		{ 216, 347, 15 },
		{ 216, 348, 15 },
		{ 216, 349, 15 },
		{ 216, 350, 15 },
		{ 216, 351, 15 },
		{ 216, 352, 15 },
		{ 216, 353, 15 },
		{ 216, 354, 15 },
		{ 216, 355, 15 },
		{ 216, 356, 15 },
		{ 216, 357, 15 },
		{ 216, 358, 15 },
		{ 216, 359, 15 },
		{ 216, 360, 15 },
		{ 216, 361, 15 },
		{ 216, 362, 15 },
		{ 216, 363, 15 },
		{ 216, 364, 15 },
		{ 216, 365, 15 },
		{ 216, 366, 15 },
		{ 216, 367, 15 },
		{ 216, 368, 15 },
		{ 216, 369, 15 },
		{ 216, 370, 15 },
		{ 216, 371, 15 },
		{ 216, 372, 15 },
		{ 216, 373, 15 },
		{ 216, 374, 15 },
		{ 216, 375, 15 },
		{ 216, 376, 15 },
		{ 216, 377, 15 },
		{ 216, 378, 15 },
		{ 216, 379, 15 },
		{ 216, 380, 15 },
		{ 216, 381, 15 },
		{ 216, 382, 15 },
		{ 216, 383, 15 },
		{ 217, 338, 15 },
		{ 217, 339, 15 },
		{ 217, 340, 15 },
		{ 217, 341, 15 },
		{ 217, 342, 15 },
		{ 217, 343, 15 },
		{ 217, 344, 15 },
		{ 217, 345, 15 },
		{ 217, 346, 15 },
		{ 217, 347, 15 },
		{ 217, 348, 15 },
		{ 217, 349, 15 },
		{ 217, 350, 15 },
		{ 217, 351, 15 },
		{ 217, 352, 15 },
		{ 217, 353, 15 },
		{ 217, 354, 15 },
		{ 217, 355, 15 },
		{ 217, 356, 15 },
		{ 217, 357, 15 },
		{ 217, 358, 15 },
		{ 217, 359, 15 },
		{ 217, 360, 15 },
		{ 217, 361, 15 },
		{ 217, 362, 15 },
		{ 217, 363, 15 },
		{ 217, 364, 15 },
		{ 217, 365, 15 },
		{ 217, 366, 15 },
		{ 217, 367, 15 },
		{ 217, 368, 15 },
		{ 217, 369, 15 },
		{ 217, 370, 15 },
		{ 217, 371, 15 },
		{ 217, 372, 15 },
		{ 217, 373, 15 },
		{ 217, 374, 15 },
		{ 217, 375, 15 },
		{ 217, 376, 15 },
		{ 217, 377, 15 },
		{ 217, 378, 15 },
		{ 217, 379, 15 },
		{ 217, 380, 15 },
		{ 217, 381, 15 },
		{ 217, 382, 15 },
		{ 217, 383, 15 },
		{ 218, 336, 15 },
		{ 218, 337, 15 },
		{ 218, 338, 15 },
		{ 218, 339, 15 },
		{ 218, 340, 15 },
		{ 218, 341, 15 },
		{ 218, 342, 15 },
		{ 218, 343, 15 },
		{ 218, 344, 15 },
		{ 218, 345, 15 },
		{ 218, 346, 15 },
		{ 218, 347, 15 },
		{ 218, 348, 15 },
		{ 218, 349, 15 },
		{ 218, 350, 15 },
		{ 218, 351, 15 },
		{ 218, 352, 15 },
		{ 218, 353, 15 },
		{ 218, 354, 15 },
		{ 218, 355, 15 },
		{ 218, 356, 15 },
		{ 218, 357, 15 },
		{ 218, 358, 15 },
		{ 218, 359, 15 },
		{ 218, 360, 15 },
		{ 218, 361, 15 },
		{ 218, 362, 15 },
		{ 218, 363, 15 },
		{ 218, 364, 15 },
		{ 218, 365, 15 },
		{ 218, 366, 15 },
		{ 218, 367, 15 },
		{ 218, 368, 15 },
		{ 218, 369, 15 },
		{ 218, 370, 15 },
		{ 218, 371, 15 },
		{ 218, 372, 15 },
		{ 218, 373, 15 },
		{ 218, 374, 15 },
		{ 218, 375, 15 },
		{ 218, 376, 15 },
		{ 218, 377, 15 },
		{ 218, 378, 15 },
		{ 218, 379, 15 },
		{ 218, 380, 15 },
		{ 218, 381, 15 },
		{ 218, 382, 15 },
		{ 218, 383, 15 },
		{ 219, 334, 15 },
		{ 219, 335, 15 },
		{ 219, 336, 15 },
		{ 219, 337, 15 },
		{ 219, 338, 15 },
		{ 219, 339, 15 },
		{ 219, 340, 15 },
		{ 219, 341, 15 },
		{ 219, 342, 15 },
		{ 219, 343, 15 },
		{ 219, 344, 15 },
		{ 219, 345, 15 },
		{ 219, 346, 15 },
		{ 219, 347, 15 },
		{ 219, 348, 15 },
		{ 219, 349, 15 },
		{ 219, 350, 15 },
		{ 219, 351, 15 },
		{ 219, 352, 15 },
		{ 219, 353, 15 },
		{ 219, 354, 15 },
		{ 219, 355, 15 },
		{ 219, 356, 15 },
		{ 219, 357, 15 },
		{ 219, 358, 15 },
		{ 219, 359, 15 },
		{ 219, 360, 15 },
		{ 219, 361, 15 },
		{ 219, 362, 15 },
		{ 219, 363, 15 },
		{ 219, 364, 15 },
		{ 219, 365, 15 },
		{ 219, 366, 15 },
		{ 219, 367, 15 },
		{ 219, 368, 15 },
		{ 219, 369, 15 },
		{ 219, 370, 15 },
		{ 219, 371, 15 },
		{ 219, 372, 15 },
		{ 219, 373, 15 },
		{ 219, 374, 15 },
		{ 219, 375, 15 },
		{ 219, 376, 15 },
		{ 219, 377, 15 },
		{ 219, 378, 15 },
		{ 219, 379, 15 },
		{ 219, 380, 15 },
		{ 219, 381, 15 },
		{ 219, 382, 15 },
		{ 219, 383, 15 },
		{ 220, 332, 15 },
		{ 220, 333, 15 },
		{ 220, 334, 15 },
		{ 220, 335, 15 },
		{ 220, 336, 15 },
		{ 220, 337, 15 },
		{ 220, 338, 15 },
		{ 220, 339, 15 },
		{ 220, 340, 15 },
		{ 220, 341, 15 },
		{ 220, 342, 15 },
		{ 220, 343, 15 },
		{ 220, 344, 15 },
		{ 220, 345, 15 },
		{ 220, 346, 15 },
		{ 220, 347, 15 },
		{ 220, 348, 15 },
		{ 220, 349, 15 },
		{ 220, 350, 15 },
		{ 220, 351, 15 },
		{ 220, 352, 15 },
		{ 220, 353, 15 },
		{ 220, 354, 15 },
		{ 220, 355, 15 },
		{ 220, 356, 15 },
		{ 220, 357, 15 },
		{ 220, 358, 15 },
		{ 220, 359, 15 },
		{ 220, 360, 15 },
		{ 220, 361, 15 },
		{ 220, 362, 15 },
		{ 220, 363, 15 },
		{ 220, 364, 15 },
		{ 220, 365, 15 },
		{ 220, 366, 15 },
		{ 220, 367, 15 },
		{ 220, 368, 15 },
		{ 220, 369, 15 },
		{ 220, 370, 15 },
		{ 220, 371, 15 },
		{ 220, 372, 15 },
		{ 220, 373, 15 },
		{ 220, 374, 15 },
		{ 220, 375, 15 },
		{ 220, 376, 15 },
		{ 220, 377, 15 },
		{ 220, 378, 15 },
		{ 220, 379, 15 },
		{ 220, 380, 15 },
		{ 220, 381, 15 },
		{ 220, 382, 15 },
		{ 220, 383, 15 },
		{ 221, 330, 15 },
		{ 221, 331, 15 },
		{ 221, 332, 15 },
		{ 221, 333, 15 },
		{ 221, 334, 15 },
		{ 221, 335, 15 },
		{ 221, 336, 15 },
		{ 221, 337, 15 },
		{ 221, 338, 15 },
		{ 221, 339, 15 },
		{ 221, 340, 15 },
		{ 221, 341, 15 },
		{ 221, 342, 15 },
		{ 221, 343, 15 },
		{ 221, 344, 15 },
		{ 221, 345, 15 },
		{ 221, 346, 15 },
		{ 221, 347, 15 },
		{ 221, 348, 15 },
		{ 221, 349, 15 },
		{ 221, 350, 15 },
		{ 221, 351, 15 },
		{ 221, 352, 15 },
		{ 221, 353, 15 },
		{ 221, 354, 15 },
		{ 221, 355, 15 },
		{ 221, 356, 15 },
		{ 221, 357, 15 },
		{ 221, 358, 15 },
		{ 221, 359, 15 },
		{ 221, 360, 15 },
		{ 221, 361, 15 },
		{ 221, 362, 15 },
		{ 221, 363, 15 },
		{ 221, 364, 15 },
		{ 221, 365, 15 },
		{ 221, 366, 15 },
		{ 221, 367, 15 },
		{ 221, 368, 15 },
		{ 221, 369, 15 },
		{ 221, 370, 15 },
		{ 221, 371, 15 },
		{ 221, 372, 15 },
		{ 221, 373, 15 },
		{ 221, 374, 15 },
		{ 221, 375, 15 },
		{ 221, 376, 15 },
		{ 221, 377, 15 },
		{ 221, 378, 15 },
		{ 221, 379, 15 },
		{ 221, 380, 15 },
		{ 221, 381, 15 },
		{ 221, 382, 15 },
		{ 221, 383, 15 },
		{ 222, 328, 15 },
		{ 222, 329, 15 },
		{ 222, 330, 15 },
		{ 222, 331, 15 },
		{ 222, 332, 15 },
		{ 222, 333, 15 },
		{ 222, 334, 15 },
		{ 222, 335, 15 },
		{ 222, 336, 15 },
		{ 222, 337, 15 },
		{ 222, 338, 15 },
		{ 222, 339, 15 },
		{ 222, 340, 15 },
		{ 222, 341, 15 },
		{ 222, 342, 15 },
		{ 222, 343, 15 },
		{ 222, 344, 15 },
		{ 222, 345, 15 },
		{ 222, 346, 15 },
		{ 222, 347, 15 },
		{ 222, 348, 15 },
		{ 222, 349, 15 },
		{ 222, 350, 15 },
		{ 222, 351, 15 },
		{ 222, 352, 15 },
		{ 222, 353, 15 },
		{ 222, 354, 15 },
		{ 222, 355, 15 },
		{ 222, 356, 15 },
		{ 222, 357, 15 },
		{ 222, 358, 15 },
		{ 222, 359, 15 },
		{ 222, 360, 15 },
		{ 222, 361, 15 },
		{ 222, 362, 15 },
		{ 222, 363, 15 },
		{ 222, 364, 15 },
		{ 222, 365, 15 },
		{ 222, 366, 15 },
		{ 222, 367, 15 },
		{ 222, 368, 15 },
		{ 222, 369, 15 },
		{ 222, 370, 15 },
		{ 222, 371, 15 },
		{ 222, 372, 15 },
		{ 222, 373, 15 },
		{ 222, 374, 15 },
		{ 222, 375, 15 },
		{ 222, 376, 15 },
		{ 222, 377, 15 },
		{ 222, 378, 15 },
		{ 222, 379, 15 },
		{ 222, 380, 15 },
		{ 222, 381, 15 },
		{ 222, 382, 15 },
		{ 222, 383, 15 },
		{ 223, 326, 15 },
		{ 223, 327, 15 },
		{ 223, 328, 15 },
		{ 223, 329, 15 },
		{ 223, 330, 15 },
		{ 223, 331, 15 },
		{ 223, 332, 15 },
		{ 223, 333, 15 },
		{ 223, 334, 15 },
		{ 223, 335, 15 },
		{ 223, 336, 15 },
		{ 223, 337, 15 },
		{ 223, 338, 15 },
		{ 223, 339, 15 },
		{ 223, 340, 15 },
		{ 223, 341, 15 },
		{ 223, 342, 15 },
		{ 223, 343, 15 },
		{ 223, 344, 15 },
		{ 223, 345, 15 },
		{ 223, 346, 15 },
		{ 223, 347, 15 },
		{ 223, 348, 15 },
		{ 223, 349, 15 },
		{ 223, 350, 15 },
		{ 223, 351, 15 },
		{ 223, 352, 15 },
		{ 223, 353, 15 },
		{ 223, 354, 15 },
		{ 223, 355, 15 },
		{ 223, 356, 15 },
		{ 223, 357, 15 },
		{ 223, 358, 15 },
		{ 223, 359, 15 },
		{ 223, 360, 15 },
		{ 223, 361, 15 },
		{ 223, 362, 15 },
		{ 223, 363, 15 },
		{ 223, 364, 15 },
		{ 223, 365, 15 },
		{ 223, 366, 15 },
		{ 223, 367, 15 },
		{ 223, 368, 15 },
		{ 223, 369, 15 },
		{ 223, 370, 15 },
		{ 223, 371, 15 },
		{ 223, 372, 15 },
		{ 223, 373, 15 },
		{ 223, 374, 15 },
		{ 223, 375, 15 },
		{ 223, 376, 15 },
		{ 223, 377, 15 },
		{ 223, 378, 15 },
		{ 223, 379, 15 },
		{ 223, 380, 15 },
		{ 223, 381, 15 },
		{ 223, 382, 15 },
		{ 223, 383, 15 },
		{ 224, 324, 15 },
		{ 224, 325, 15 },
		{ 224, 326, 15 },
		{ 224, 327, 15 },
		{ 224, 328, 15 },
		{ 224, 329, 15 },
		{ 224, 330, 15 },
		{ 224, 331, 15 },
		{ 224, 332, 15 },
		{ 224, 333, 15 },
		{ 224, 334, 15 },
		{ 224, 335, 15 },
		{ 224, 336, 15 },
		{ 224, 337, 15 },
		{ 224, 338, 15 },
		{ 224, 339, 15 },
		{ 224, 340, 15 },
		{ 224, 341, 15 },
		{ 224, 342, 15 },
		{ 224, 343, 15 },
		{ 224, 344, 15 },
		{ 224, 345, 15 },
		{ 224, 346, 15 },
		{ 224, 347, 15 },
		{ 224, 348, 15 },
		{ 224, 349, 15 },
		{ 224, 350, 15 },
		{ 224, 351, 15 },
		{ 224, 352, 15 },
		{ 224, 353, 15 },
		{ 224, 354, 15 },
		{ 224, 355, 15 },
		{ 224, 356, 15 },
		{ 224, 357, 15 },
		{ 224, 358, 15 },
		{ 224, 359, 15 },
		{ 224, 360, 15 },
		{ 224, 361, 15 },
		{ 224, 362, 15 },
		{ 224, 363, 15 },
		{ 224, 364, 15 },
		{ 224, 365, 15 },
		{ 224, 366, 15 },
		{ 224, 367, 15 },
		{ 224, 368, 15 },
		{ 224, 369, 15 },
		{ 224, 370, 15 },
		{ 224, 371, 15 },
		{ 224, 372, 15 },
		{ 224, 373, 15 },
		{ 224, 374, 15 },
		{ 224, 375, 15 },
		{ 224, 376, 15 },
		{ 224, 377, 15 },
		{ 224, 378, 15 },
		{ 224, 379, 15 },
		{ 224, 380, 15 },
		{ 224, 381, 15 },
		{ 224, 382, 15 },
		{ 224, 383, 15 },
		{ 225, 322, 15 },
		{ 225, 323, 15 },
		{ 225, 324, 15 },
		{ 225, 325, 15 },
		{ 225, 326, 15 },
		{ 225, 327, 15 },
		{ 225, 328, 15 },
		{ 225, 329, 15 },
		{ 225, 330, 15 },
		{ 225, 331, 15 },
		{ 225, 332, 15 },
		{ 225, 333, 15 },
		{ 225, 334, 15 },
		{ 225, 335, 15 },
		{ 225, 336, 15 },
		{ 225, 337, 15 },
		{ 225, 338, 15 },
		{ 225, 339, 15 },
		{ 225, 340, 15 },
		{ 225, 341, 15 },
		{ 225, 342, 15 },
		{ 225, 343, 15 },
		{ 225, 344, 15 },
		{ 225, 345, 15 },
		{ 225, 346, 15 },
		{ 225, 347, 15 },
		{ 225, 348, 15 },
		{ 225, 349, 15 },
		{ 225, 350, 15 },
		{ 225, 351, 15 },
		{ 225, 352, 15 },
		{ 225, 353, 15 },
		{ 225, 354, 15 },
		{ 225, 355, 15 },
		{ 225, 356, 15 },
		{ 225, 357, 15 },
		{ 225, 358, 15 },
		{ 225, 359, 15 },
		{ 225, 360, 15 },
		{ 225, 361, 15 },
		{ 225, 362, 15 },
		{ 225, 363, 15 },
		{ 225, 364, 15 },
		{ 225, 365, 15 },
		{ 225, 366, 15 },
		{ 225, 367, 15 },
		{ 225, 368, 15 },
		{ 225, 369, 15 },
		{ 225, 370, 15 },
		{ 225, 371, 15 },
		{ 225, 372, 15 },
		{ 225, 373, 15 },
		{ 225, 374, 15 },
		{ 225, 375, 15 },
		{ 225, 376, 15 },
		{ 225, 377, 15 },
		{ 225, 378, 15 },
		{ 225, 379, 15 },
		{ 225, 380, 15 },
		{ 225, 381, 15 },
		{ 225, 382, 15 },
		{ 225, 383, 15 },
		{ 226, 320, 15 },
		{ 226, 321, 15 },
		{ 226, 322, 15 },
		{ 226, 323, 15 },
		{ 226, 324, 15 },
		{ 226, 325, 15 },
		{ 226, 326, 15 },
		{ 226, 327, 15 },
		{ 226, 328, 15 },
		{ 226, 329, 15 },
		{ 226, 330, 15 },
		{ 226, 331, 15 },
		{ 226, 332, 15 },
		{ 226, 333, 15 },
		{ 226, 334, 15 },
		{ 226, 335, 15 },
		{ 226, 336, 15 },
		{ 226, 337, 15 },
		{ 226, 338, 15 },
		{ 226, 339, 15 },
		{ 226, 340, 15 },
		{ 226, 341, 15 },
		{ 226, 342, 15 },
		{ 226, 343, 15 },
		{ 226, 344, 15 },
		{ 226, 345, 15 },
		{ 226, 346, 15 },
		{ 226, 347, 15 },
		{ 226, 348, 15 },
		{ 226, 349, 15 },
		{ 226, 350, 15 },
		{ 226, 351, 15 },
		{ 226, 352, 15 },
		{ 226, 353, 15 },
		{ 226, 354, 15 },
		{ 226, 355, 15 },
		{ 226, 356, 15 },
		{ 226, 357, 15 },
		{ 226, 358, 15 },
		{ 226, 359, 15 },
		{ 226, 360, 15 },
		{ 226, 361, 15 },
		{ 226, 362, 15 },
		{ 226, 363, 15 },
		{ 226, 364, 15 },
		{ 226, 365, 15 },
		{ 226, 366, 15 },
		{ 226, 367, 15 },
		{ 226, 368, 15 },
		{ 226, 369, 15 },
		{ 226, 370, 15 },
		{ 226, 371, 15 },
		{ 226, 372, 15 },
		{ 226, 373, 15 },
		{ 226, 374, 15 },
		{ 226, 375, 15 },
		{ 226, 376, 15 },
		{ 226, 377, 15 },
		{ 226, 378, 15 },
		{ 226, 379, 15 },
		{ 226, 380, 15 },
		{ 226, 381, 15 },
		{ 226, 382, 15 },
		{ 226, 383, 15 },
		{ 227, 318, 15 },
		{ 227, 319, 15 },
		{ 227, 320, 15 },
		{ 227, 321, 15 },
		{ 227, 322, 15 },
		{ 227, 323, 15 },
		{ 227, 324, 15 },
		{ 227, 325, 15 },
		{ 227, 326, 15 },
		{ 227, 327, 15 },
		{ 227, 328, 15 },
		{ 227, 329, 15 },
		{ 227, 330, 15 },
		{ 227, 331, 15 },
		{ 227, 332, 15 },
		{ 227, 333, 15 },
		{ 227, 334, 15 },
		{ 227, 335, 15 },
		{ 227, 336, 15 },
		{ 227, 337, 15 },
		{ 227, 338, 15 },
		{ 227, 339, 15 },
		{ 227, 340, 15 },
		{ 227, 341, 15 },
		{ 227, 342, 15 },
		{ 227, 343, 15 },
		{ 227, 344, 15 },
		{ 227, 345, 15 },
		{ 227, 346, 15 },
		{ 227, 347, 15 },
		{ 227, 348, 15 },
		{ 227, 349, 15 },
		{ 227, 350, 15 },
		{ 227, 351, 15 },
		{ 227, 352, 15 },
		{ 227, 353, 15 },
		{ 227, 354, 15 },
		{ 227, 355, 15 },
		{ 227, 356, 15 },
		{ 227, 357, 15 },
		{ 227, 358, 15 },
		{ 227, 359, 15 },
		{ 227, 360, 15 },
		{ 227, 361, 15 },
		{ 227, 362, 15 },
		{ 227, 363, 15 },
		{ 227, 364, 15 },
		{ 227, 365, 15 },
		{ 227, 366, 15 },
		{ 227, 367, 15 },
		{ 227, 368, 15 },
		{ 227, 369, 15 },
		{ 227, 370, 15 },
		{ 227, 371, 15 },
		{ 227, 372, 15 },
		{ 227, 373, 15 },
		{ 227, 374, 15 },
		{ 227, 375, 15 },
		{ 227, 376, 15 },
		{ 227, 377, 15 },
		{ 227, 378, 15 },
		{ 227, 379, 15 },
		{ 227, 380, 15 },
		{ 227, 381, 15 },
		{ 227, 382, 15 },
		{ 227, 383, 15 },
		{ 228, 316, 15 },
		{ 228, 317, 15 },
		{ 228, 318, 15 },
		{ 228, 319, 15 },
		{ 228, 320, 15 },
		{ 228, 321, 15 },
		{ 228, 322, 15 },
		{ 228, 323, 15 },
		{ 228, 324, 15 },
		{ 228, 325, 15 },
		{ 228, 326, 15 },
		{ 228, 327, 15 },
		{ 228, 328, 15 },
		{ 228, 329, 15 },
		{ 228, 330, 15 },
		{ 228, 331, 15 },
		{ 228, 332, 15 },
		{ 228, 333, 15 },
		{ 228, 334, 15 },
		{ 228, 335, 15 },
		{ 228, 336, 15 },
		{ 228, 337, 15 },
		{ 228, 338, 15 },
		{ 228, 339, 15 },
		{ 228, 340, 15 },
		{ 228, 341, 15 },
		{ 228, 342, 15 },
		{ 228, 343, 15 },
		{ 228, 344, 15 },
		{ 228, 345, 15 },
		{ 228, 346, 15 },
		{ 228, 347, 15 },
		{ 228, 348, 15 },
		{ 228, 349, 15 },
		{ 228, 350, 15 },
		{ 228, 351, 15 },
		{ 228, 352, 15 },
		{ 228, 353, 15 },
		{ 228, 354, 15 },
		{ 228, 355, 15 },
		{ 228, 356, 15 },
		{ 228, 357, 15 },
		{ 228, 358, 15 },
		{ 228, 359, 15 },
		{ 228, 360, 15 },
		{ 228, 361, 15 },
		{ 228, 362, 15 },
		{ 228, 363, 15 },
		{ 228, 364, 15 },
		{ 228, 365, 15 },
		{ 228, 366, 15 },
		{ 228, 367, 15 },
		{ 228, 368, 15 },
		{ 228, 369, 15 },
		{ 228, 370, 15 },
		{ 228, 371, 15 },
		{ 228, 372, 15 },
		{ 228, 373, 15 },
		{ 228, 374, 15 },
		{ 228, 375, 15 },
		{ 228, 376, 15 },
		{ 228, 377, 15 },
		{ 228, 378, 15 },
		{ 228, 379, 15 },
		{ 228, 380, 15 },
		{ 228, 381, 15 },
		{ 228, 382, 15 },
		{ 228, 383, 15 },
		{ 229, 314, 15 },
		{ 229, 315, 15 },
		{ 229, 316, 15 },
		{ 229, 317, 15 },
		{ 229, 318, 15 },
		{ 229, 319, 15 },
		{ 229, 320, 15 },
		{ 229, 321, 15 },
		{ 229, 322, 15 },
		{ 229, 323, 15 },
		{ 229, 324, 15 },
		{ 229, 325, 15 },
		{ 229, 326, 15 },
		{ 229, 327, 15 },
		{ 229, 328, 15 },
		{ 229, 329, 15 },
		{ 229, 330, 15 },
		{ 229, 331, 15 },
		{ 229, 332, 15 },
		{ 229, 333, 15 },
		{ 229, 334, 15 },
		{ 229, 335, 15 },
		{ 229, 336, 15 },
		{ 229, 337, 15 },
		{ 229, 338, 15 },
		{ 229, 339, 15 },
		{ 229, 340, 15 },
		{ 229, 341, 15 },
		{ 229, 342, 15 },
		{ 229, 343, 15 },
		{ 229, 344, 15 },
		{ 229, 345, 15 },
		{ 229, 346, 15 },
		{ 229, 347, 15 },
		{ 229, 348, 15 },
		{ 229, 349, 15 },
		{ 229, 350, 15 },
		{ 229, 351, 15 },
		{ 229, 352, 15 },
		{ 229, 353, 15 },
		{ 229, 354, 15 },
		{ 229, 355, 15 },
		{ 229, 356, 15 },
		{ 229, 357, 15 },
		{ 229, 358, 15 },
		{ 229, 359, 15 },
		{ 229, 360, 15 },
		{ 229, 361, 15 },
		{ 229, 362, 15 },
		{ 229, 363, 15 },
		{ 229, 364, 15 },
		{ 229, 365, 15 },
		{ 229, 366, 15 },
		{ 229, 367, 15 },
		{ 229, 368, 15 },
		{ 229, 369, 15 },
		{ 229, 370, 15 },
		{ 229, 371, 15 },
		{ 229, 372, 15 },
		{ 229, 373, 15 },
		{ 229, 374, 15 },
		{ 229, 375, 15 },
		{ 229, 376, 15 },
		{ 229, 377, 15 },
		{ 229, 378, 15 },
		{ 229, 379, 15 },
		{ 229, 380, 15 },
		{ 229, 381, 15 },
		{ 229, 382, 15 },
		{ 229, 383, 15 },
		{ 230, 312, 15 },
		{ 230, 313, 15 },
		{ 230, 314, 15 },
		{ 230, 315, 15 },
		{ 230, 316, 15 },
		{ 230, 317, 15 },
		{ 230, 318, 15 },
		{ 230, 319, 15 },
		{ 230, 320, 15 },
		{ 230, 321, 15 },
		{ 230, 322, 15 },
		{ 230, 323, 15 },
		{ 230, 324, 15 },
		{ 230, 325, 15 },
		{ 230, 326, 15 },
		{ 230, 327, 15 },
		{ 230, 328, 15 },
		{ 230, 329, 15 },
		{ 230, 330, 15 },
		{ 230, 331, 15 },
		{ 230, 332, 15 },
		{ 230, 333, 15 },
		{ 230, 334, 15 },
		{ 230, 335, 15 },
		{ 230, 336, 15 },
		{ 230, 337, 15 },
		{ 230, 338, 15 },
		{ 230, 339, 15 },
		{ 230, 340, 15 },
		{ 230, 341, 15 },
		{ 230, 342, 15 },
		{ 230, 343, 15 },
		{ 230, 344, 15 },
		{ 230, 345, 15 },
		{ 230, 346, 15 },
		{ 230, 347, 15 },
		{ 230, 348, 15 },
		{ 230, 349, 15 },
		{ 230, 350, 15 },
		{ 230, 351, 15 },
		{ 230, 352, 15 },
		{ 230, 353, 15 },
		{ 230, 354, 15 },
		{ 230, 355, 15 },
		{ 230, 356, 15 },
		{ 230, 357, 15 },
		{ 230, 358, 15 },
		{ 230, 359, 15 },
		{ 230, 360, 15 },
		{ 230, 361, 15 },
		{ 230, 362, 15 },
		{ 230, 363, 15 },
		{ 230, 364, 15 },
		{ 230, 365, 15 },
		{ 230, 366, 15 },
		{ 230, 367, 15 },
		{ 230, 368, 15 },
		{ 230, 369, 15 },
		{ 230, 370, 15 },
		{ 230, 371, 15 },
		{ 230, 372, 15 },
		{ 230, 373, 15 },
		{ 230, 374, 15 },
		{ 230, 375, 15 },
		{ 230, 376, 15 },
		{ 230, 377, 15 },
		{ 230, 378, 15 },
		{ 230, 379, 15 },
		{ 230, 380, 15 },
		{ 230, 381, 15 },
		{ 230, 382, 15 },
		{ 230, 383, 15 },
		{ 231, 310, 15 },
		{ 231, 311, 15 },
		{ 231, 312, 15 },
		{ 231, 313, 15 },
		{ 231, 314, 15 },
		{ 231, 315, 15 },
		{ 231, 316, 15 },
		{ 231, 317, 15 },
		{ 231, 318, 15 },
		{ 231, 319, 15 },
		{ 231, 320, 15 },
		{ 231, 321, 15 },
		{ 231, 322, 15 },
		{ 231, 323, 15 },
		{ 231, 324, 15 },
		{ 231, 325, 15 },
		{ 231, 326, 15 },
		{ 231, 327, 15 },
		{ 231, 328, 15 },
		{ 231, 329, 15 },
		{ 231, 330, 15 },
		{ 231, 331, 15 },
		{ 231, 332, 15 },
		{ 231, 333, 15 },
		{ 231, 334, 15 },
		{ 231, 335, 15 },
		{ 231, 336, 15 },
		{ 231, 337, 15 },
		{ 231, 338, 15 },
		{ 231, 339, 15 },
		{ 231, 340, 15 },
		{ 231, 341, 15 },
		{ 231, 342, 15 },
		{ 231, 343, 15 },
		{ 231, 344, 15 },
		{ 231, 345, 15 },
		{ 231, 346, 15 },
		{ 231, 347, 15 },
		{ 231, 348, 15 },
		{ 231, 349, 15 },
		{ 231, 350, 15 },
		{ 231, 351, 15 },
		{ 231, 352, 15 },
		{ 231, 353, 15 },
		{ 231, 354, 15 },
		{ 231, 355, 15 },
		{ 231, 356, 15 },
		{ 231, 357, 15 },
		{ 231, 358, 15 },
		{ 231, 359, 15 },
		{ 231, 360, 15 },
		{ 231, 361, 15 },
		{ 231, 362, 15 },
		{ 231, 363, 15 },
		{ 231, 364, 15 },
		{ 231, 365, 15 },
		{ 231, 366, 15 },
		{ 231, 367, 15 },
		{ 231, 368, 15 },
		{ 231, 369, 15 },
		{ 231, 370, 15 },
		{ 231, 371, 15 },
		{ 231, 372, 15 },
		{ 231, 373, 15 },
		{ 231, 374, 15 },
		{ 231, 375, 15 },
		{ 231, 376, 15 },
		{ 231, 377, 15 },
		{ 231, 378, 15 },
		{ 231, 379, 15 },
		{ 231, 380, 15 },
		{ 231, 381, 15 },
		{ 231, 382, 15 },
		{ 231, 383, 15 },
		{ 232, 308, 15 },
		{ 232, 309, 15 },
		{ 232, 310, 15 },
		{ 232, 311, 15 },
		{ 232, 312, 15 },
		{ 232, 313, 15 },
		{ 232, 314, 15 },
		{ 232, 315, 15 },
		{ 232, 316, 15 },
		{ 232, 317, 15 },
		{ 232, 318, 15 },
		{ 232, 319, 15 },
		{ 232, 320, 15 },
		{ 232, 321, 15 },
		{ 232, 322, 15 },
		{ 232, 323, 15 },
		{ 232, 324, 15 },
		{ 232, 325, 15 },
		{ 232, 326, 15 },
		{ 232, 327, 15 },
		{ 232, 328, 15 },
		{ 232, 329, 15 },
		{ 232, 330, 15 },
		{ 232, 331, 15 },
		{ 232, 332, 15 },
		{ 232, 333, 15 },
		{ 232, 334, 15 },
		{ 232, 335, 15 },
		{ 232, 336, 15 },
		{ 232, 337, 15 },
		{ 232, 338, 15 },
		{ 232, 339, 15 },
		{ 232, 340, 15 },
		{ 232, 341, 15 },
		{ 232, 342, 15 },
		{ 232, 343, 15 },
		{ 232, 344, 15 },
		{ 232, 345, 15 },
		{ 232, 346, 15 },
		{ 232, 347, 15 },
		{ 232, 348, 15 },
		{ 232, 349, 15 },
		{ 232, 350, 15 },
		{ 232, 351, 15 },
		{ 232, 352, 15 },
		{ 232, 353, 15 },
		{ 232, 354, 15 },
		{ 232, 355, 15 },
		{ 232, 356, 15 },
		{ 232, 357, 15 },
		{ 232, 358, 15 },
		{ 232, 359, 15 },
		{ 232, 360, 15 },
		{ 232, 361, 15 },
		{ 232, 362, 15 },
		{ 232, 363, 15 },
		{ 232, 364, 15 },
		{ 232, 365, 15 },
		{ 232, 366, 15 },
		{ 232, 367, 15 },
		{ 232, 368, 15 },
		{ 232, 369, 15 },
		{ 232, 370, 15 },
		{ 232, 371, 15 },
		{ 232, 372, 15 },
		{ 232, 373, 15 },
		{ 232, 374, 15 },
		{ 232, 375, 15 },
		{ 232, 376, 15 },
		{ 232, 377, 15 },
		{ 232, 378, 15 },
		{ 232, 379, 15 },
		{ 232, 380, 15 },
		{ 232, 381, 15 },
		{ 232, 382, 15 },
		{ 232, 383, 15 },
		{ 233, 306, 15 },
		{ 233, 307, 15 },
		{ 233, 308, 15 },
		{ 233, 309, 15 },
		{ 233, 310, 15 },
		{ 233, 311, 15 },
		{ 233, 312, 15 },
		{ 233, 313, 15 },
		{ 233, 314, 15 },
		{ 233, 315, 15 },
		{ 233, 316, 15 },
		{ 233, 317, 15 },
		{ 233, 318, 15 },
		{ 233, 319, 15 },
		{ 233, 320, 15 },
		{ 233, 321, 15 },
		{ 233, 322, 15 },
		{ 233, 323, 15 },
		{ 233, 324, 15 },
		{ 233, 325, 15 },
		{ 233, 326, 15 },
		{ 233, 327, 15 },
		{ 233, 328, 15 },
		{ 233, 329, 15 },
		{ 233, 330, 15 },
		{ 233, 331, 15 },
		{ 233, 332, 15 },
		{ 233, 333, 15 },
		{ 233, 334, 15 },
		{ 233, 335, 15 },
		{ 233, 336, 15 },
		{ 233, 337, 15 },
		{ 233, 338, 15 },
		{ 233, 339, 15 },
		{ 233, 340, 15 },
		{ 233, 341, 15 },
		{ 233, 342, 15 },
		{ 233, 343, 15 },
		{ 233, 344, 15 },
		{ 233, 345, 15 },
		{ 233, 346, 15 },
		{ 233, 347, 15 },
		{ 233, 348, 15 },
		{ 233, 349, 15 },
		{ 233, 350, 15 },
		{ 233, 351, 15 },
		{ 233, 352, 15 },
		{ 233, 353, 15 },
		{ 233, 354, 15 },
		{ 233, 355, 15 },
		{ 233, 356, 15 },
		{ 233, 357, 15 },
		{ 233, 358, 15 },
		{ 233, 359, 15 },
		{ 233, 360, 15 },
		{ 233, 361, 15 },
		{ 233, 362, 15 },
		{ 233, 363, 15 },
		{ 233, 364, 15 },
		{ 233, 365, 15 },
		{ 233, 366, 15 },
		{ 233, 367, 15 },
		{ 233, 368, 15 },
		{ 233, 369, 15 },
		{ 233, 370, 15 },
		{ 233, 371, 15 },
		{ 233, 372, 15 },
		{ 233, 373, 15 },
		{ 233, 374, 15 },
		{ 233, 375, 15 },
		{ 233, 376, 15 },
		{ 233, 377, 15 },
		{ 233, 378, 15 },
		{ 233, 379, 15 },
		{ 233, 380, 15 },
		{ 233, 381, 15 },
		{ 233, 382, 15 },
		{ 233, 383, 15 },
		{ 234, 304, 15 },
		{ 234, 305, 15 },
		{ 234, 306, 15 },
		{ 234, 307, 15 },
		{ 234, 308, 15 },
		{ 234, 309, 15 },
		{ 234, 310, 15 },
		{ 234, 311, 15 },
		{ 234, 312, 15 },
		{ 234, 313, 15 },
		{ 234, 314, 15 },
		{ 234, 315, 15 },
		{ 234, 316, 15 },
		{ 234, 317, 15 },
		{ 234, 318, 15 },
		{ 234, 319, 15 },
		{ 234, 320, 15 },
		{ 234, 321, 15 },
		{ 234, 322, 15 },
		{ 234, 323, 15 },
		{ 234, 324, 15 },
		{ 234, 325, 15 },
		{ 234, 326, 15 },
		{ 234, 327, 15 },
		{ 234, 328, 15 },
		{ 234, 329, 15 },
		{ 234, 330, 15 },
		{ 234, 331, 15 },
		{ 234, 332, 15 },
		{ 234, 333, 15 },
		{ 234, 334, 15 },
		{ 234, 335, 15 },
		{ 234, 336, 15 },
		{ 234, 337, 15 },
		{ 234, 338, 15 },
		{ 234, 339, 15 },
		{ 234, 340, 15 },
		{ 234, 341, 15 },
		{ 234, 342, 15 },
		{ 234, 343, 15 },
		{ 234, 344, 15 },
		{ 234, 345, 15 },
		{ 234, 346, 15 },
		{ 234, 347, 15 },
		{ 234, 348, 15 },
		{ 234, 349, 15 },
		{ 234, 350, 15 },
		{ 234, 351, 15 },
		{ 234, 352, 15 },
		{ 234, 353, 15 },
		{ 234, 354, 15 },
		{ 234, 355, 15 },
		{ 234, 356, 15 },
		{ 234, 357, 15 },
		{ 234, 358, 15 },
		{ 234, 359, 15 },
		{ 234, 360, 15 },
		{ 234, 361, 15 },
		{ 234, 362, 15 },
		{ 234, 363, 15 },
		{ 234, 364, 15 },
		{ 234, 365, 15 },
		{ 234, 366, 15 },
		{ 234, 367, 15 },
		{ 234, 368, 15 },
		{ 234, 369, 15 },
		{ 234, 370, 15 },
		{ 234, 371, 15 },
		{ 234, 372, 15 },
		{ 234, 373, 15 },
		{ 234, 374, 15 },
		{ 234, 375, 15 },
		{ 234, 376, 15 },
		{ 234, 377, 15 },
		{ 234, 378, 15 },
		{ 234, 379, 15 },
		{ 234, 380, 15 },
		{ 234, 381, 15 },
		{ 234, 382, 15 },
		{ 234, 383, 15 },
		{ 235, 302, 15 },
		{ 235, 303, 15 },
		{ 235, 304, 15 },
		{ 235, 305, 15 },
		{ 235, 306, 15 },
		{ 235, 307, 15 },
		{ 235, 308, 15 },
		{ 235, 309, 15 },
		{ 235, 310, 15 },
		{ 235, 311, 15 },
		{ 235, 312, 15 },
		{ 235, 313, 15 },
		{ 235, 314, 15 },
		{ 235, 315, 15 },
		{ 235, 316, 15 },
		{ 235, 317, 15 },
		{ 235, 318, 15 },
		{ 235, 319, 15 },
		{ 235, 320, 15 },
		{ 235, 321, 15 },
		{ 235, 322, 15 },
		{ 235, 323, 15 },
		{ 235, 324, 15 },
		{ 235, 325, 15 },
		{ 235, 326, 15 },
		{ 235, 327, 15 },
		{ 235, 328, 15 },
		{ 235, 329, 15 },
		{ 235, 330, 15 },
		{ 235, 331, 15 },
		{ 235, 332, 15 },
		{ 235, 333, 15 },
		{ 235, 334, 15 },
		{ 235, 335, 15 },
		{ 235, 336, 15 },
		{ 235, 337, 15 },
		{ 235, 338, 15 },
		{ 235, 339, 15 },
		{ 235, 340, 15 },
		{ 235, 341, 15 },
		{ 235, 342, 15 },
		{ 235, 343, 15 },
		{ 235, 344, 15 },
		{ 235, 345, 15 },
		{ 235, 346, 15 },
		{ 235, 347, 15 },
		{ 235, 348, 15 },
		{ 235, 349, 15 },
		{ 235, 350, 15 },
		{ 235, 351, 15 },
		{ 235, 352, 15 },
		{ 235, 353, 15 },
		{ 235, 354, 15 },
		{ 235, 355, 15 },
		{ 235, 356, 15 },
		{ 235, 357, 15 },
		{ 235, 358, 15 },
		{ 235, 359, 15 },
		{ 235, 360, 15 },
		{ 235, 361, 15 },
		{ 235, 362, 15 },
		{ 235, 363, 15 },
		{ 235, 364, 15 },
		{ 235, 365, 15 },
		{ 235, 366, 15 },
		{ 235, 367, 15 },
		{ 235, 368, 15 },
		{ 235, 369, 15 },
		{ 235, 370, 15 },
		{ 235, 371, 15 },
		{ 235, 372, 15 },
		{ 235, 373, 15 },
		{ 235, 374, 15 },
		{ 235, 375, 15 },
		{ 235, 376, 15 },
		{ 235, 377, 15 },
		{ 235, 378, 15 },
		{ 235, 379, 15 },
		{ 235, 380, 15 },
		{ 235, 381, 15 },
		{ 235, 382, 15 },
		{ 235, 383, 15 },
		{ 236, 300, 15 },
		{ 236, 301, 15 },
		{ 236, 302, 15 },
		{ 236, 303, 15 },
		{ 236, 304, 15 },
		{ 236, 305, 15 },
		{ 236, 306, 15 },
		{ 236, 307, 15 },
		{ 236, 308, 15 },
		{ 236, 309, 15 },
		{ 236, 310, 15 },
		{ 236, 311, 15 },
		{ 236, 312, 15 },
		{ 236, 313, 15 },
		{ 236, 314, 15 },
		{ 236, 315, 15 },
		{ 236, 316, 15 },
		{ 236, 317, 15 },
		{ 236, 318, 15 },
		{ 236, 319, 15 },
		{ 236, 320, 15 },
		{ 236, 321, 15 },
		{ 236, 322, 15 },
		{ 236, 323, 15 },
		{ 236, 324, 15 },
		{ 236, 325, 15 },
		{ 236, 326, 15 },
		{ 236, 327, 15 },
		{ 236, 328, 15 },
		{ 236, 329, 15 },
		{ 236, 330, 15 },
		{ 236, 331, 15 },
		{ 236, 332, 15 },
		{ 236, 333, 15 },
		{ 236, 334, 15 },
		{ 236, 335, 15 },
		{ 236, 336, 15 },
		{ 236, 337, 15 },
		{ 236, 338, 15 },
		{ 236, 339, 15 },
		{ 236, 340, 15 },
		{ 236, 341, 15 },
		{ 236, 342, 15 },
		{ 236, 343, 15 },
		{ 236, 344, 15 },
		{ 236, 345, 15 },
		{ 236, 346, 15 },
		{ 236, 347, 15 },
		{ 236, 348, 15 },
		{ 236, 349, 15 },
		{ 236, 350, 15 },
		{ 236, 351, 15 },
		{ 236, 352, 15 },
		{ 236, 353, 15 },
		{ 236, 354, 15 },
		{ 236, 355, 15 },
		{ 236, 356, 15 },
		{ 236, 357, 15 },
		{ 236, 358, 15 },
		{ 236, 359, 15 },
		{ 236, 360, 15 },
		{ 236, 361, 15 },
		{ 236, 362, 15 },
		{ 236, 363, 15 },
		{ 236, 364, 15 },
		{ 236, 365, 15 },
		{ 236, 366, 15 },
		{ 236, 367, 15 },
		{ 236, 368, 15 },
		{ 236, 369, 15 },
		{ 236, 370, 15 },
		{ 236, 371, 15 },
		{ 236, 372, 15 },
		{ 236, 373, 15 },
		{ 236, 374, 15 },
		{ 236, 375, 15 },
		{ 236, 376, 15 },
		{ 236, 377, 15 },
		{ 236, 378, 15 },
		{ 236, 379, 15 },
		{ 236, 380, 15 },
		{ 236, 381, 15 },
		{ 236, 382, 15 },
		{ 236, 383, 15 },
		{ 237, 298, 15 },
		{ 237, 299, 15 },
		{ 237, 300, 15 },
		{ 237, 301, 15 },
		{ 237, 302, 15 },
		{ 237, 303, 15 },
		{ 237, 304, 15 },
		{ 237, 305, 15 },
		{ 237, 306, 15 },
		{ 237, 307, 15 },
		{ 237, 308, 15 },
		{ 237, 309, 15 },
		{ 237, 310, 15 },
		{ 237, 311, 15 },
		{ 237, 312, 15 },
		{ 237, 313, 15 },
		{ 237, 314, 15 },
		{ 237, 315, 15 },
		{ 237, 316, 15 },
		{ 237, 317, 15 },
		{ 237, 318, 15 },
		{ 237, 319, 15 },
		{ 237, 320, 15 },
		{ 237, 321, 15 },
		{ 237, 322, 15 },
		{ 237, 323, 15 },
		{ 237, 324, 15 },
		{ 237, 325, 15 },
		{ 237, 326, 15 },
		{ 237, 327, 15 },
		{ 237, 328, 15 },
		{ 237, 329, 15 },
		{ 237, 330, 15 },
		{ 237, 331, 15 },
		{ 237, 332, 15 },
		{ 237, 333, 15 },
		{ 237, 334, 15 },
		{ 237, 335, 15 },
		{ 237, 336, 15 },
		{ 237, 337, 15 },
		{ 237, 338, 15 },
		{ 237, 339, 15 },
		{ 237, 340, 15 },
		{ 237, 341, 15 },
		{ 237, 342, 15 },
		{ 237, 343, 15 },
		{ 237, 344, 15 },
		{ 237, 345, 15 },
		{ 237, 346, 15 },
		{ 237, 347, 15 },
		{ 237, 348, 15 },
		{ 237, 349, 15 },
		{ 237, 350, 15 },
		{ 237, 351, 15 },
		{ 237, 352, 15 },
		{ 237, 353, 15 },
		{ 237, 354, 15 },
		{ 237, 355, 15 },
		{ 237, 356, 15 },
		{ 237, 357, 15 },
		{ 237, 358, 15 },
		{ 237, 359, 15 },
		{ 237, 360, 15 },
		{ 237, 361, 15 },
		{ 237, 362, 15 },
		{ 237, 363, 15 },
		{ 237, 364, 15 },
		{ 237, 365, 15 },
		{ 237, 366, 15 },
		{ 237, 367, 15 },
		{ 237, 368, 15 },
		{ 237, 369, 15 },
		{ 237, 370, 15 },
		{ 237, 371, 15 },
		{ 237, 372, 15 },
		{ 237, 373, 15 },
		{ 237, 374, 15 },
		{ 237, 375, 15 },
		{ 237, 376, 15 },
		{ 237, 377, 15 },
		{ 237, 378, 15 },
		{ 237, 379, 15 },
		{ 237, 380, 15 },
		{ 237, 381, 15 },
		{ 237, 382, 15 },
		{ 237, 383, 15 },
		{ 238, 296, 15 },
		{ 238, 297, 15 },
		{ 238, 298, 15 },
		{ 238, 299, 15 },
		{ 238, 300, 15 },
		{ 238, 301, 15 },
		{ 238, 302, 15 },
		{ 238, 303, 15 },
		{ 238, 304, 15 },
		{ 238, 305, 15 },
		{ 238, 306, 15 },
		{ 238, 307, 15 },
		{ 238, 308, 15 },
		{ 238, 309, 15 },
		{ 238, 310, 15 },
		{ 238, 311, 15 },
		{ 238, 312, 15 },
		{ 238, 313, 15 },
		{ 238, 314, 15 },
		{ 238, 315, 15 },
		{ 238, 316, 15 },
		{ 238, 317, 15 },
		{ 238, 318, 15 },
		{ 238, 319, 15 },
		{ 238, 320, 15 },
		{ 238, 321, 15 },
		{ 238, 322, 15 },
		{ 238, 323, 15 },
		{ 238, 324, 15 },
		{ 238, 325, 15 },
		{ 238, 326, 15 },
		{ 238, 327, 15 },
		{ 238, 328, 15 },
		{ 238, 329, 15 },
		{ 238, 330, 15 },
		{ 238, 331, 15 },
		{ 238, 332, 15 },
		{ 238, 333, 15 },
		{ 238, 334, 15 },
		{ 238, 335, 15 },
		{ 238, 336, 15 },
		{ 238, 337, 15 },
		{ 238, 338, 15 },
		{ 238, 339, 15 },
		{ 238, 340, 15 },
		{ 238, 341, 15 },
		{ 238, 342, 15 },
		{ 238, 343, 15 },
		{ 238, 344, 15 },
		{ 238, 345, 15 },
		{ 238, 346, 15 },
		{ 238, 347, 15 },
		{ 238, 348, 15 },
		{ 238, 349, 15 },
		{ 238, 350, 15 },
		{ 238, 351, 15 },
		{ 238, 352, 15 },
		{ 238, 353, 15 },
		{ 238, 354, 15 },
		{ 238, 355, 15 },
		{ 238, 356, 15 },
		{ 238, 357, 15 },
		{ 238, 358, 15 },
		{ 238, 359, 15 },
		{ 238, 360, 15 },
		{ 238, 361, 15 },
		{ 238, 362, 15 },
		{ 238, 363, 15 },
		{ 238, 364, 15 },
		{ 238, 365, 15 },
		{ 238, 366, 15 },
		{ 238, 367, 15 },
		{ 238, 368, 15 },
		{ 238, 369, 15 },
		{ 238, 370, 15 },
		{ 238, 371, 15 },
		{ 238, 372, 15 },
		{ 238, 373, 15 },
		{ 238, 374, 15 },
		{ 238, 375, 15 },
		{ 238, 376, 15 },
		{ 238, 377, 15 },
		{ 238, 378, 15 },
		{ 238, 379, 15 },
		{ 238, 380, 15 },
		{ 238, 381, 15 },
		{ 238, 382, 15 },
		{ 238, 383, 15 },
		{ 239, 294, 15 },
		{ 239, 295, 15 },
		{ 239, 296, 15 },
		{ 239, 297, 15 },
		{ 239, 298, 15 },
		{ 239, 299, 15 },
		{ 239, 300, 15 },
		{ 239, 301, 15 },
		{ 239, 302, 15 },
		{ 239, 303, 15 },
		{ 239, 304, 15 },
		{ 239, 305, 15 },
		{ 239, 306, 15 },
		{ 239, 307, 15 },
		{ 239, 308, 15 },
		{ 239, 309, 15 },
		{ 239, 310, 15 },
		{ 239, 311, 15 },
		{ 239, 312, 15 },
		{ 239, 313, 15 },
		{ 239, 314, 15 },
		{ 239, 315, 15 },
		{ 239, 316, 15 },
		{ 239, 317, 15 },
		{ 239, 318, 15 },
		{ 239, 319, 15 },
		{ 239, 320, 15 },
		{ 239, 321, 15 },
		{ 239, 322, 15 },
		{ 239, 323, 15 },
		{ 239, 324, 15 },
		{ 239, 325, 15 },
		{ 239, 326, 15 },
		{ 239, 327, 15 },
		{ 239, 328, 15 },
		{ 239, 329, 15 },
		{ 239, 330, 15 },
		{ 239, 331, 15 },
		{ 239, 332, 15 },
		{ 239, 333, 15 },
		{ 239, 334, 15 },
		{ 239, 335, 15 },
		{ 239, 336, 15 },
		{ 239, 337, 15 },
		{ 239, 338, 15 },
		{ 239, 339, 15 },
		{ 239, 340, 15 },
		{ 239, 341, 15 },
		{ 239, 342, 15 },
		{ 239, 343, 15 },
		{ 239, 344, 15 },
		{ 239, 345, 15 },
		{ 239, 346, 15 },
		{ 239, 347, 15 },
		{ 239, 348, 15 },
		{ 239, 349, 15 },
		{ 239, 350, 15 },
		{ 239, 351, 15 },
		{ 239, 352, 15 },
		{ 239, 353, 15 },
		{ 239, 354, 15 },
		{ 239, 355, 15 },
		{ 239, 356, 15 },
		{ 239, 357, 15 },
		{ 239, 358, 15 },
		{ 239, 359, 15 },
		{ 239, 360, 15 },
		{ 239, 361, 15 },
		{ 239, 362, 15 },
		{ 239, 363, 15 },
		{ 239, 364, 15 },
		{ 239, 365, 15 },
		{ 239, 366, 15 },
		{ 239, 367, 15 },
		{ 239, 368, 15 },
		{ 239, 369, 15 },
		{ 239, 370, 15 },
		{ 239, 371, 15 },
		{ 239, 372, 15 },
		{ 239, 373, 15 },
		{ 239, 374, 15 },
		{ 239, 375, 15 },
		{ 239, 376, 15 },
		{ 239, 377, 15 },
		{ 239, 378, 15 },
		{ 239, 379, 15 },
		{ 239, 380, 15 },
		{ 239, 381, 15 },
		{ 239, 382, 15 },
		{ 239, 383, 15 },
		{ 240, 292, 15 },
		{ 240, 293, 15 },
		{ 240, 294, 15 },
		{ 240, 295, 15 },
		{ 240, 296, 15 },
		{ 240, 297, 15 },
		{ 240, 298, 15 },
		{ 240, 299, 15 },
		{ 240, 300, 15 },
		{ 240, 301, 15 },
		{ 240, 302, 15 },
		{ 240, 303, 15 },
		{ 240, 304, 15 },
		{ 240, 305, 15 },
		{ 240, 306, 15 },
		{ 240, 307, 15 },
		{ 240, 308, 15 },
		{ 240, 309, 15 },
		{ 240, 310, 15 },
		{ 240, 311, 15 },
		{ 240, 312, 15 },
		{ 240, 313, 15 },
		{ 240, 314, 15 },
		{ 240, 315, 15 },
		{ 240, 316, 15 },
		{ 240, 317, 15 },
		{ 240, 318, 15 },
		{ 240, 319, 15 },
		{ 240, 320, 15 },
		{ 240, 321, 15 },
		{ 240, 322, 15 },
		{ 240, 323, 15 },
		{ 240, 324, 15 },
		{ 240, 325, 15 },
		{ 240, 326, 15 },
		{ 240, 327, 15 },
		{ 240, 328, 15 },
		{ 240, 329, 15 },
		{ 240, 330, 15 },
		{ 240, 331, 15 },
		{ 240, 332, 15 },
		{ 240, 333, 15 },
		{ 240, 334, 15 },
		{ 240, 335, 15 },
		{ 240, 336, 15 },
		{ 240, 337, 15 },
		{ 240, 338, 15 },
		{ 240, 339, 15 },
		{ 240, 340, 15 },
		{ 240, 341, 15 },
		{ 240, 342, 15 },
		{ 240, 343, 15 },
		{ 240, 344, 15 },
		{ 240, 345, 15 },
		{ 240, 346, 15 },
		{ 240, 347, 15 },
		{ 240, 348, 15 },
		{ 240, 349, 15 },
		{ 240, 350, 15 },
		{ 240, 351, 15 },
		{ 240, 352, 15 },
		{ 240, 353, 15 },
		{ 240, 354, 15 },
		{ 240, 355, 15 },
		{ 240, 356, 15 },
		{ 240, 357, 15 },
		{ 240, 358, 15 },
		{ 240, 359, 15 },
		{ 240, 360, 15 },
		{ 240, 361, 15 },
		{ 240, 362, 15 },
		{ 240, 363, 15 },
		{ 240, 364, 15 },
		{ 240, 365, 15 },
		{ 240, 366, 15 },
		{ 240, 367, 15 },
		{ 240, 368, 15 },
		{ 240, 369, 15 },
		{ 240, 370, 15 },
		{ 240, 371, 15 },
		{ 240, 372, 15 },
		{ 240, 373, 15 },
		{ 240, 374, 15 },
		{ 240, 375, 15 },
		{ 240, 376, 15 },
		{ 240, 377, 15 },
		{ 240, 378, 15 },
		{ 240, 379, 15 },
		{ 240, 380, 15 },
		{ 240, 381, 15 },
		{ 240, 382, 15 },
		{ 240, 383, 15 },
		{ 241, 290, 15 },
		{ 241, 291, 15 },
		{ 241, 292, 15 },
		{ 241, 293, 15 },
		{ 241, 294, 15 },
		{ 241, 295, 15 },
		{ 241, 296, 15 },
		{ 241, 297, 15 },
		{ 241, 298, 15 },
		{ 241, 299, 15 },
		{ 241, 300, 15 },
		{ 241, 301, 15 },
		{ 241, 302, 15 },
		{ 241, 303, 15 },
		{ 241, 304, 15 },
		{ 241, 305, 15 },
		{ 241, 306, 15 },
		{ 241, 307, 15 },
		{ 241, 308, 15 },
		{ 241, 309, 15 },
		{ 241, 310, 15 },
		{ 241, 311, 15 },
		{ 241, 312, 15 },
		{ 241, 313, 15 },
		{ 241, 314, 15 },
		{ 241, 315, 15 },
		{ 241, 316, 15 },
		{ 241, 317, 15 },
		{ 241, 318, 15 },
		{ 241, 319, 15 },
		{ 241, 320, 15 },
		{ 241, 321, 15 },
		{ 241, 322, 15 },
		{ 241, 323, 15 },
		{ 241, 324, 15 },
		{ 241, 325, 15 },
		{ 241, 326, 15 },
		{ 241, 327, 15 },
		{ 241, 328, 15 },
		{ 241, 329, 15 },
		{ 241, 330, 15 },
		{ 241, 331, 15 },
		{ 241, 332, 15 },
		{ 241, 333, 15 },
		{ 241, 334, 15 },
		{ 241, 335, 15 },
		{ 241, 336, 15 },
		{ 241, 337, 15 },
		{ 241, 338, 15 },
		{ 241, 339, 15 },
		{ 241, 340, 15 },
		{ 241, 341, 15 },
		{ 241, 342, 15 },
		{ 241, 343, 15 },
		{ 241, 344, 15 },
		{ 241, 345, 15 },
		{ 241, 346, 15 },
		{ 241, 347, 15 },
		{ 241, 348, 15 },
		{ 241, 349, 15 },
		{ 241, 350, 15 },
		{ 241, 351, 15 },
		{ 241, 352, 15 },
		{ 241, 353, 15 },
		{ 241, 354, 15 },
		{ 241, 355, 15 },
		{ 241, 356, 15 },
		{ 241, 357, 15 },
		{ 241, 358, 15 },
		{ 241, 359, 15 },
		{ 241, 360, 15 },
		{ 241, 361, 15 },
		{ 241, 362, 15 },
		{ 241, 363, 15 },
		{ 241, 364, 15 },
		{ 241, 365, 15 },
		{ 241, 366, 15 },
		{ 241, 367, 15 },
		{ 241, 368, 15 },
		{ 241, 369, 15 },
		{ 241, 370, 15 },
		{ 241, 371, 15 },
		{ 241, 372, 15 },
		{ 241, 373, 15 },
		{ 241, 374, 15 },
		{ 241, 375, 15 },
		{ 241, 376, 15 },
		{ 241, 377, 15 },
		{ 241, 378, 15 },
		{ 241, 379, 15 },
		{ 241, 380, 15 },
		{ 241, 381, 15 },
		{ 241, 382, 15 },
		{ 241, 383, 15 },
		{ 242, 288, 15 },
		{ 242, 289, 15 },
		{ 242, 290, 15 },
		{ 242, 291, 15 },
		{ 242, 292, 15 },
		{ 242, 293, 15 },
		{ 242, 294, 15 },
		{ 242, 295, 15 },
		{ 242, 296, 15 },
		{ 242, 297, 15 },
		{ 242, 298, 15 },
		{ 242, 299, 15 },
		{ 242, 300, 15 },
		{ 242, 301, 15 },
		{ 242, 302, 15 },
		{ 242, 303, 15 },
		{ 242, 304, 15 },
		{ 242, 305, 15 },
		{ 242, 306, 15 },
		{ 242, 307, 15 },
		{ 242, 308, 15 },
		{ 242, 309, 15 },
		{ 242, 310, 15 },
		{ 242, 311, 15 },
		{ 242, 312, 15 },
		{ 242, 313, 15 },
		{ 242, 314, 15 },
		{ 242, 315, 15 },
		{ 242, 316, 15 },
		{ 242, 317, 15 },
		{ 242, 318, 15 },
		{ 242, 319, 15 },
		{ 242, 320, 15 },
		{ 242, 321, 15 },
		{ 242, 322, 15 },
		{ 242, 323, 15 },
		{ 242, 324, 15 },
		{ 242, 325, 15 },
		{ 242, 326, 15 },
		{ 242, 327, 15 },
		{ 242, 328, 15 },
		{ 242, 329, 15 },
		{ 242, 330, 15 },
		{ 242, 331, 15 },
		{ 242, 332, 15 },
		{ 242, 333, 15 },
		{ 242, 334, 15 },
		{ 242, 335, 15 },
		{ 242, 336, 15 },
		{ 242, 337, 15 },
		{ 242, 338, 15 },
		{ 242, 339, 15 },
		{ 242, 340, 15 },
		{ 242, 341, 15 },
		{ 242, 342, 15 },
		{ 242, 343, 15 },
		{ 242, 344, 15 },
		{ 242, 345, 15 },
		{ 242, 346, 15 },
		{ 242, 347, 15 },
		{ 242, 348, 15 },
		{ 242, 349, 15 },
		{ 242, 350, 15 },
		{ 242, 351, 15 },
		{ 242, 352, 15 },
		{ 242, 353, 15 },
		{ 242, 354, 15 },
		{ 242, 355, 15 },
		{ 242, 356, 15 },
		{ 242, 357, 15 },
		{ 242, 358, 15 },
		{ 242, 359, 15 },
		{ 242, 360, 15 },
		{ 242, 361, 15 },
		{ 242, 362, 15 },
		{ 242, 363, 15 },
		{ 242, 364, 15 },
		{ 242, 365, 15 },
		{ 242, 366, 15 },
		{ 242, 367, 15 },
		{ 242, 368, 15 },
		{ 242, 369, 15 },
		{ 242, 370, 15 },
		{ 242, 371, 15 },
		{ 242, 372, 15 },
		{ 242, 373, 15 },
		{ 242, 374, 15 },
		{ 242, 375, 15 },
		{ 242, 376, 15 },
		{ 242, 377, 15 },
		{ 242, 378, 15 },
		{ 242, 379, 15 },
		{ 242, 380, 15 },
		{ 242, 381, 15 },
		{ 242, 382, 15 },
		{ 242, 383, 15 },
		{ 243, 286, 15 },
		{ 243, 287, 15 },
		{ 243, 288, 15 },
		{ 243, 289, 15 },
		{ 243, 290, 15 },
		{ 243, 291, 15 },
		{ 243, 292, 15 },
		{ 243, 293, 15 },
		{ 243, 294, 15 },
		{ 243, 295, 15 },
		{ 243, 296, 15 },
		{ 243, 297, 15 },
		{ 243, 298, 15 },
		{ 243, 299, 15 },
		{ 243, 300, 15 },
		{ 243, 301, 15 },
		{ 243, 302, 15 },
		{ 243, 303, 15 },
		{ 243, 304, 15 },
		{ 243, 305, 15 },
		{ 243, 306, 15 },
		{ 243, 307, 15 },
		{ 243, 308, 15 },
		{ 243, 309, 15 },
		{ 243, 310, 15 },
		{ 243, 311, 15 },
		{ 243, 312, 15 },
		{ 243, 313, 15 },
		{ 243, 314, 15 },
		{ 243, 315, 15 },
		{ 243, 316, 15 },
		{ 243, 317, 15 },
		{ 243, 318, 15 },
		{ 243, 319, 15 },
		{ 243, 320, 15 },
		{ 243, 321, 15 },
		{ 243, 322, 15 },
		{ 243, 323, 15 },
		{ 243, 324, 15 },
		{ 243, 325, 15 },
		{ 243, 326, 15 },
		{ 243, 327, 15 },
		{ 243, 328, 15 },
		{ 243, 329, 15 },
		{ 243, 330, 15 },
		{ 243, 331, 15 },
		{ 243, 332, 15 },
		{ 243, 333, 15 },
		{ 243, 334, 15 },
		{ 243, 335, 15 },
		{ 243, 336, 15 },
		{ 243, 337, 15 },
		{ 243, 338, 15 },
		{ 243, 339, 15 },
		{ 243, 340, 15 },
		{ 243, 341, 15 },
		{ 243, 342, 15 },
		{ 243, 343, 15 },
		{ 243, 344, 15 },
		{ 243, 345, 15 },
		{ 243, 346, 15 },
		{ 243, 347, 15 },
		{ 243, 348, 15 },
		{ 243, 349, 15 },
		{ 243, 350, 15 },
		{ 243, 351, 15 },
		{ 243, 352, 15 },
		{ 243, 353, 15 },
		{ 243, 354, 15 },
		{ 243, 355, 15 },
		{ 243, 356, 15 },
		{ 243, 357, 15 },
		{ 243, 358, 15 },
		{ 243, 359, 15 },
		{ 243, 360, 15 },
		{ 243, 361, 15 },
		{ 243, 362, 15 },
		{ 243, 363, 15 },
		{ 243, 364, 15 },
		{ 243, 365, 15 },
		{ 243, 366, 15 },
		{ 243, 367, 15 },
		{ 243, 368, 15 },
		{ 243, 369, 15 },
		{ 243, 370, 15 },
		{ 243, 371, 15 },
		{ 243, 372, 15 },
		{ 243, 373, 15 },
		{ 243, 374, 15 },
		{ 243, 375, 15 },
		{ 243, 376, 15 },
		{ 243, 377, 15 },
		{ 243, 378, 15 },
		{ 243, 379, 15 },
		{ 243, 380, 15 },
		{ 243, 381, 15 },
		{ 243, 382, 15 },
		{ 243, 383, 15 },
		{ 244, 284, 15 },
		{ 244, 285, 15 },
		{ 244, 286, 15 },
		{ 244, 287, 15 },
		{ 244, 288, 15 },
		{ 244, 289, 15 },
		{ 244, 290, 15 },
		{ 244, 291, 15 },
		{ 244, 292, 15 },
		{ 244, 293, 15 },
		{ 244, 294, 15 },
		{ 244, 295, 15 },
		{ 244, 296, 15 },
		{ 244, 297, 15 },
		{ 244, 298, 15 },
		{ 244, 299, 15 },
		{ 244, 300, 15 },
		{ 244, 301, 15 },
		{ 244, 302, 15 },
		{ 244, 303, 15 },
		{ 244, 304, 15 },
		{ 244, 305, 15 },
		{ 244, 306, 15 },
		{ 244, 307, 15 },
		{ 244, 308, 15 },
		{ 244, 309, 15 },
		{ 244, 310, 15 },
		{ 244, 311, 15 },
		{ 244, 312, 15 },
		{ 244, 313, 15 },
		{ 244, 314, 15 },
		{ 244, 315, 15 },
		{ 244, 316, 15 },
		{ 244, 317, 15 },
		{ 244, 318, 15 },
		{ 244, 319, 15 },
		{ 244, 320, 15 },
		{ 244, 321, 15 },
		{ 244, 322, 15 },
		{ 244, 323, 15 },
		{ 244, 324, 15 },
		{ 244, 325, 15 },
		{ 244, 326, 15 },
		{ 244, 327, 15 },
		{ 244, 328, 15 },
		{ 244, 329, 15 },
		{ 244, 330, 15 },
		{ 244, 331, 15 },
		{ 244, 332, 15 },
		{ 244, 333, 15 },
		{ 244, 334, 15 },
		{ 244, 335, 15 },
		{ 244, 336, 15 },
		{ 244, 337, 15 },
		{ 244, 338, 15 },
		{ 244, 339, 15 },
		{ 244, 340, 15 },
		{ 244, 341, 15 },
		{ 244, 342, 15 },
		{ 244, 343, 15 },
		{ 244, 344, 15 },
		{ 244, 345, 15 },
		{ 244, 346, 15 },
		{ 244, 347, 15 },
		{ 244, 348, 15 },
		{ 244, 349, 15 },
		{ 244, 350, 15 },
		{ 244, 351, 15 },
		{ 244, 352, 15 },
		{ 244, 353, 15 },
		{ 244, 354, 15 },
		{ 244, 355, 15 },
		{ 244, 356, 15 },
		{ 244, 357, 15 },
		{ 244, 358, 15 },
		{ 244, 359, 15 },
		{ 244, 360, 15 },
		{ 244, 361, 15 },
		{ 244, 362, 15 },
		{ 244, 363, 15 },
		{ 244, 364, 15 },
		{ 244, 365, 15 },
		{ 244, 366, 15 },
		{ 244, 367, 15 },
		{ 244, 368, 15 },
		{ 244, 369, 15 },
		{ 244, 370, 15 },
		{ 244, 371, 15 },
		{ 244, 372, 15 },
		{ 244, 373, 15 },
		{ 244, 374, 15 },
		{ 244, 375, 15 },
		{ 244, 376, 15 },
		{ 244, 377, 15 },
		{ 244, 378, 15 },
		{ 244, 379, 15 },
		{ 244, 380, 15 },
		{ 244, 381, 15 },
		{ 244, 382, 15 },
		{ 244, 383, 15 },
		{ 245, 282, 15 },
		{ 245, 283, 15 },
		{ 245, 284, 15 },
		{ 245, 285, 15 },
		{ 245, 286, 15 },
		{ 245, 287, 15 },
		{ 245, 288, 15 },
		{ 245, 289, 15 },
		{ 245, 290, 15 },
		{ 245, 291, 15 },
		{ 245, 292, 15 },
		{ 245, 293, 15 },
		{ 245, 294, 15 },
		{ 245, 295, 15 },
		{ 245, 296, 15 },
		{ 245, 297, 15 },
		{ 245, 298, 15 },
		{ 245, 299, 15 },
		{ 245, 300, 15 },
		{ 245, 301, 15 },
		{ 245, 302, 15 },
		{ 245, 303, 15 },
		{ 245, 304, 15 },
		{ 245, 305, 15 },
		{ 245, 306, 15 },
		{ 245, 307, 15 },
		{ 245, 308, 15 },
		{ 245, 309, 15 },
		{ 245, 310, 15 },
		{ 245, 311, 15 },
		{ 245, 312, 15 },
		{ 245, 313, 15 },
		{ 245, 314, 15 },
		{ 245, 315, 15 },
		{ 245, 316, 15 },
		{ 245, 317, 15 },
		{ 245, 318, 15 },
		{ 245, 319, 15 },
		{ 245, 320, 15 },
		{ 245, 321, 15 },
		{ 245, 322, 15 },
		{ 245, 323, 15 },
		{ 245, 324, 15 },
		{ 245, 325, 15 },
		{ 245, 326, 15 },
		{ 245, 327, 15 },
		{ 245, 328, 15 },
		{ 245, 329, 15 },
		{ 245, 330, 15 },
		{ 245, 331, 15 },
		{ 245, 332, 15 },
		{ 245, 333, 15 },
		{ 245, 334, 15 },
		{ 245, 335, 15 },
		{ 245, 336, 15 },
		{ 245, 337, 15 },
		{ 245, 338, 15 },
		{ 245, 339, 15 },
		{ 245, 340, 15 },
		{ 245, 341, 15 },
		{ 245, 342, 15 },
		{ 245, 343, 15 },
		{ 245, 344, 15 },
		{ 245, 345, 15 },
		{ 245, 346, 15 },
		{ 245, 347, 15 },
		{ 245, 348, 15 },
		{ 245, 349, 15 },
		{ 245, 350, 15 },
		{ 245, 351, 15 },
		{ 245, 352, 15 },
		{ 245, 353, 15 },
		{ 245, 354, 15 },
		{ 245, 355, 15 },
		{ 245, 356, 15 },
		{ 245, 357, 15 },
		{ 245, 358, 15 },
		{ 245, 359, 15 },
		{ 245, 360, 15 },
		{ 245, 361, 15 },
		{ 245, 362, 15 },
		{ 245, 363, 15 },
		{ 245, 364, 15 },
		{ 245, 365, 15 },
		{ 245, 366, 15 },
		{ 245, 367, 15 },
		{ 245, 368, 15 },
		{ 245, 369, 15 },
		{ 245, 370, 15 },
		{ 245, 371, 15 },
		{ 245, 372, 15 },
		{ 245, 373, 15 },
		{ 245, 374, 15 },
		{ 245, 375, 15 },
		{ 245, 376, 15 },
		{ 245, 377, 15 },
		{ 245, 378, 15 },
		{ 245, 379, 15 },
		{ 245, 380, 15 },
		{ 245, 381, 15 },
		{ 245, 382, 15 },
		{ 245, 383, 15 },
		{ 246, 280, 15 },
		{ 246, 281, 15 },
		{ 246, 282, 15 },
		{ 246, 283, 15 },
		{ 246, 284, 15 },
		{ 246, 285, 15 },
		{ 246, 286, 15 },
		{ 246, 287, 15 },
		{ 246, 288, 15 },
		{ 246, 289, 15 },
		{ 246, 290, 15 },
		{ 246, 291, 15 },
		{ 246, 292, 15 },
		{ 246, 293, 15 },
		{ 246, 294, 15 },
		{ 246, 295, 15 },
		{ 246, 296, 15 },
		{ 246, 297, 15 },
		{ 246, 298, 15 },
		{ 246, 299, 15 },
		{ 246, 300, 15 },
		{ 246, 301, 15 },
		{ 246, 302, 15 },
		{ 246, 303, 15 },
		{ 246, 304, 15 },
		{ 246, 305, 15 },
		{ 246, 306, 15 },
		{ 246, 307, 15 },
		{ 246, 308, 15 },
		{ 246, 309, 15 },
		{ 246, 310, 15 },
		{ 246, 311, 15 },
		{ 246, 312, 15 },
		{ 246, 313, 15 },
		{ 246, 314, 15 },
		{ 246, 315, 15 },
		{ 246, 316, 15 },
		{ 246, 317, 15 },
		{ 246, 318, 15 },
		{ 246, 319, 15 },
		{ 246, 320, 15 },
		{ 246, 321, 15 },
		{ 246, 322, 15 },
		{ 246, 323, 15 },
		{ 246, 324, 15 },
		{ 246, 325, 15 },
		{ 246, 326, 15 },
		{ 246, 327, 15 },
		{ 246, 328, 15 },
		{ 246, 329, 15 },
		{ 246, 330, 15 },
		{ 246, 331, 15 },
		{ 246, 332, 15 },
		{ 246, 333, 15 },
		{ 246, 334, 15 },
		{ 246, 335, 15 },
		{ 246, 336, 15 },
		{ 246, 337, 15 },
		{ 246, 338, 15 },
		{ 246, 339, 15 },
		{ 246, 340, 15 },
		{ 246, 341, 15 },
		{ 246, 342, 15 },
		{ 246, 343, 15 },
		{ 246, 344, 15 },
		{ 246, 345, 15 },
		{ 246, 346, 15 },
		{ 246, 347, 15 },
		{ 246, 348, 15 },
		{ 246, 349, 15 },
		{ 246, 350, 15 },
		{ 246, 351, 15 },
		{ 246, 352, 15 },
		{ 246, 353, 15 },
		{ 246, 354, 15 },
		{ 246, 355, 15 },
		{ 246, 356, 15 },
		{ 246, 357, 15 },
		{ 246, 358, 15 },
		{ 246, 359, 15 },
		{ 246, 360, 15 },
		{ 246, 361, 15 },
		{ 246, 362, 15 },
		{ 246, 363, 15 },
		{ 246, 364, 15 },
		{ 246, 365, 15 },
		{ 246, 366, 15 },
		{ 246, 367, 15 },
		{ 246, 368, 15 },
		{ 246, 369, 15 },
		{ 246, 370, 15 },
		{ 246, 371, 15 },
		{ 246, 372, 15 },
		{ 246, 373, 15 },
		{ 246, 374, 15 },
		{ 246, 375, 15 },
		{ 246, 376, 15 },
		{ 246, 377, 15 },
		{ 246, 378, 15 },
		{ 246, 379, 15 },
		{ 246, 380, 15 },
		{ 246, 381, 15 },
		{ 246, 382, 15 },
		{ 246, 383, 15 },
		{ 247, 278, 15 },
		{ 247, 279, 15 },
		{ 247, 280, 15 },
		{ 247, 281, 15 },
		{ 247, 282, 15 },
		{ 247, 283, 15 },
		{ 247, 284, 15 },
		{ 247, 285, 15 },
		{ 247, 286, 15 },
		{ 247, 287, 15 },
		{ 247, 288, 15 },
		{ 247, 289, 15 },
		{ 247, 290, 15 },
		{ 247, 291, 15 },
		{ 247, 292, 15 },
		{ 247, 293, 15 },
		{ 247, 294, 15 },
		{ 247, 295, 15 },
		{ 247, 296, 15 },
		{ 247, 297, 15 },
		{ 247, 298, 15 },
		{ 247, 299, 15 },
		{ 247, 300, 15 },
		{ 247, 301, 15 },
		{ 247, 302, 15 },
		{ 247, 303, 15 },
		{ 247, 304, 15 },
		{ 247, 305, 15 },
		{ 247, 306, 15 },
		{ 247, 307, 15 },
		{ 247, 308, 15 },
		{ 247, 309, 15 },
		{ 247, 310, 15 },
		{ 247, 311, 15 },
		{ 247, 312, 15 },
		{ 247, 313, 15 },
		{ 247, 314, 15 },
		{ 247, 315, 15 },
		{ 247, 316, 15 },
		{ 247, 317, 15 },
		{ 247, 318, 15 },
		{ 247, 319, 15 },
		{ 247, 320, 15 },
		{ 247, 321, 15 },
		{ 247, 322, 15 },
		{ 247, 323, 15 },
		{ 247, 324, 15 },
		{ 247, 325, 15 },
		{ 247, 326, 15 },
		{ 247, 327, 15 },
		{ 247, 328, 15 },
		{ 247, 329, 15 },
		{ 247, 330, 15 },
		{ 247, 331, 15 },
		{ 247, 332, 15 },
		{ 247, 333, 15 },
		{ 247, 334, 15 },
		{ 247, 335, 15 },
		{ 247, 336, 15 },
		{ 247, 337, 15 },
		{ 247, 338, 15 },
		{ 247, 339, 15 },
		{ 247, 340, 15 },
		{ 247, 341, 15 },
		{ 247, 342, 15 },
		{ 247, 343, 15 },
		{ 247, 344, 15 },
		{ 247, 345, 15 },
		{ 247, 346, 15 },
		{ 247, 347, 15 },
		{ 247, 348, 15 },
		{ 247, 349, 15 },
		{ 247, 350, 15 },
		{ 247, 351, 15 },
		{ 247, 352, 15 },
		{ 247, 353, 15 },
		{ 247, 354, 15 },
		{ 247, 355, 15 },
		{ 247, 356, 15 },
		{ 247, 357, 15 },
		{ 247, 358, 15 },
		{ 247, 359, 15 },
		{ 247, 360, 15 },
		{ 247, 361, 15 },
		{ 247, 362, 15 },
		{ 247, 363, 15 },
		{ 247, 364, 15 },
		{ 247, 365, 15 },
		{ 247, 366, 15 },
		{ 247, 367, 15 },
		{ 247, 368, 15 },
		{ 247, 369, 15 },
		{ 247, 370, 15 },
		{ 247, 371, 15 },
		{ 247, 372, 15 },
		{ 247, 373, 15 },
		{ 247, 374, 15 },
		{ 247, 375, 15 },
		{ 247, 376, 15 },
		{ 247, 377, 15 },
		{ 247, 378, 15 },
		{ 247, 379, 15 },
		{ 247, 380, 15 },
		{ 247, 381, 15 },
		{ 247, 382, 15 },
		{ 247, 383, 15 },
		{ 248, 276, 15 },
		{ 248, 277, 15 },
		{ 248, 278, 15 },
		{ 248, 279, 15 },
		{ 248, 280, 15 },
		{ 248, 281, 15 },
		{ 248, 282, 15 },
		{ 248, 283, 15 },
		{ 248, 284, 15 },
		{ 248, 285, 15 },
		{ 248, 286, 15 },
		{ 248, 287, 15 },
		{ 248, 288, 15 },
		{ 248, 289, 15 },
		{ 248, 290, 15 },
		{ 248, 291, 15 },
		{ 248, 292, 15 },
		{ 248, 293, 15 },
		{ 248, 294, 15 },
		{ 248, 295, 15 },
		{ 248, 296, 15 },
		{ 248, 297, 15 },
		{ 248, 298, 15 },
		{ 248, 299, 15 },
		{ 248, 300, 15 },
		{ 248, 301, 15 },
		{ 248, 302, 15 },
		{ 248, 303, 15 },
		{ 248, 304, 15 },
		{ 248, 305, 15 },
		{ 248, 306, 15 },
		{ 248, 307, 15 },
		{ 248, 308, 15 },
		{ 248, 309, 15 },
		{ 248, 310, 15 },
		{ 248, 311, 15 },
		{ 248, 312, 15 },
		{ 248, 313, 15 },
		{ 248, 314, 15 },
		{ 248, 315, 15 },
		{ 248, 316, 15 },
		{ 248, 317, 15 },
		{ 248, 318, 15 },
		{ 248, 319, 15 },
		{ 248, 320, 15 },
		{ 248, 321, 15 },
		{ 248, 322, 15 },
		{ 248, 323, 15 },
		{ 248, 324, 15 },
		{ 248, 325, 15 },
		{ 248, 326, 15 },
		{ 248, 327, 15 },
		{ 248, 328, 15 },
		{ 248, 329, 15 },
		{ 248, 330, 15 },
		{ 248, 331, 15 },
		{ 248, 332, 15 },
		{ 248, 333, 15 },
		{ 248, 334, 15 },
		{ 248, 335, 15 },
		{ 248, 336, 15 },
		{ 248, 337, 15 },
		{ 248, 338, 15 },
		{ 248, 339, 15 },
		{ 248, 340, 15 },
		{ 248, 341, 15 },
		{ 248, 342, 15 },
		{ 248, 343, 15 },
		{ 248, 344, 15 },
		{ 248, 345, 15 },
		{ 248, 346, 15 },
		{ 248, 347, 15 },
		{ 248, 348, 15 },
		{ 248, 349, 15 },
		{ 248, 350, 15 },
		{ 248, 351, 15 },
		{ 248, 352, 15 },
		{ 248, 353, 15 },
		{ 248, 354, 15 },
		{ 248, 355, 15 },
		{ 248, 356, 15 },
		{ 248, 357, 15 },
		{ 248, 358, 15 },
		{ 248, 359, 15 },
		{ 248, 360, 15 },
		{ 248, 361, 15 },
		{ 248, 362, 15 },
		{ 248, 363, 15 },
		{ 248, 364, 15 },
		{ 248, 365, 15 },
		{ 248, 366, 15 },
		{ 248, 367, 15 },
		{ 248, 368, 15 },
		{ 248, 369, 15 },
		{ 248, 370, 15 },
		{ 248, 371, 15 },
		{ 248, 372, 15 },
		{ 248, 373, 15 },
		{ 248, 374, 15 },
		{ 248, 375, 15 },
		{ 248, 376, 15 },
		{ 248, 377, 15 },
		{ 248, 378, 15 },
		{ 248, 379, 15 },
		{ 248, 380, 15 },
		{ 248, 381, 15 },
		{ 248, 382, 15 },
		{ 248, 383, 15 },
		{ 249, 274, 15 },
		{ 249, 275, 15 },
		{ 249, 276, 15 },
		{ 249, 277, 15 },
		{ 249, 278, 15 },
		{ 249, 279, 15 },
		{ 249, 280, 15 },
		{ 249, 281, 15 },
		{ 249, 282, 15 },
		{ 249, 283, 15 },
		{ 249, 284, 15 },
		{ 249, 285, 15 },
		{ 249, 286, 15 },
		{ 249, 287, 15 },
		{ 249, 288, 15 },
		{ 249, 289, 15 },
		{ 249, 290, 15 },
		{ 249, 291, 15 },
		{ 249, 292, 15 },
		{ 249, 293, 15 },
		{ 249, 294, 15 },
		{ 249, 295, 15 },
		{ 249, 296, 15 },
		{ 249, 297, 15 },
		{ 249, 298, 15 },
		{ 249, 299, 15 },
		{ 249, 300, 15 },
		{ 249, 301, 15 },
		{ 249, 302, 15 },
		{ 249, 303, 15 },
		{ 249, 304, 15 },
		{ 249, 305, 15 },
		{ 249, 306, 15 },
		{ 249, 307, 15 },
		{ 249, 308, 15 },
		{ 249, 309, 15 },
		{ 249, 310, 15 },
		{ 249, 311, 15 },
		{ 249, 312, 15 },
		{ 249, 313, 15 },
		{ 249, 314, 15 },
		{ 249, 315, 15 },
		{ 249, 316, 15 },
		{ 249, 317, 15 },
		{ 249, 318, 15 },
		{ 249, 319, 15 },
		{ 249, 320, 15 },
		{ 249, 321, 15 },
		{ 249, 322, 15 },
		{ 249, 323, 15 },
		{ 249, 324, 15 },
		{ 249, 325, 15 },
		{ 249, 326, 15 },
		{ 249, 327, 15 },
		{ 249, 328, 15 },
		{ 249, 329, 15 },
		{ 249, 330, 15 },
		{ 249, 331, 15 },
		{ 249, 332, 15 },
		{ 249, 333, 15 },
		{ 249, 334, 15 },
		{ 249, 335, 15 },
		{ 249, 336, 15 },
		{ 249, 337, 15 },
		{ 249, 338, 15 },
		{ 249, 339, 15 },
		{ 249, 340, 15 },
		{ 249, 341, 15 },
		{ 249, 342, 15 },
		{ 249, 343, 15 },
		{ 249, 344, 15 },
		{ 249, 345, 15 },
		{ 249, 346, 15 },
		{ 249, 347, 15 },
		{ 249, 348, 15 },
		{ 249, 349, 15 },
		{ 249, 350, 15 },
		{ 249, 351, 15 },
		{ 249, 352, 15 },
		{ 249, 353, 15 },
		{ 249, 354, 15 },
		{ 249, 355, 15 },
		{ 249, 356, 15 },
		{ 249, 357, 15 },
		{ 249, 358, 15 },
		{ 249, 359, 15 },
		{ 249, 360, 15 },
		{ 249, 361, 15 },
		{ 249, 362, 15 },
		{ 249, 363, 15 },
		{ 249, 364, 15 },
		{ 249, 365, 15 },
		{ 249, 366, 15 },
		{ 249, 367, 15 },
		{ 249, 368, 15 },
		{ 249, 369, 15 },
		{ 249, 370, 15 },
		{ 249, 371, 15 },
		{ 249, 372, 15 },
		{ 249, 373, 15 },
		{ 249, 374, 15 },
		{ 249, 375, 15 },
		{ 249, 376, 15 },
		{ 249, 377, 15 },
		{ 249, 378, 15 },
		{ 249, 379, 15 },
		{ 249, 380, 15 },
		{ 249, 381, 15 },
		{ 249, 382, 15 },
		{ 249, 383, 15 },
		{ 250, 274, 15 },
		{ 250, 275, 15 },
		{ 250, 276, 15 },
		{ 250, 277, 15 },
		{ 250, 278, 15 },
		{ 250, 279, 15 },
		{ 250, 280, 15 },
		{ 250, 281, 15 },
		{ 250, 282, 15 },
		{ 250, 283, 15 },
		{ 250, 284, 15 },
		{ 250, 285, 15 },
		{ 250, 286, 15 },
		{ 250, 287, 15 },
		{ 250, 288, 15 },
		{ 250, 289, 15 },
		{ 250, 290, 15 },
		{ 250, 291, 15 },
		{ 250, 292, 15 },
		{ 250, 293, 15 },
		{ 250, 294, 15 },
		{ 250, 295, 15 },
		{ 250, 296, 15 },
		{ 250, 297, 15 },
		{ 250, 298, 15 },
		{ 250, 299, 15 },
		{ 250, 300, 15 },
		{ 250, 301, 15 },
		{ 250, 302, 15 },
		{ 250, 303, 15 },
		{ 250, 304, 15 },
		{ 250, 305, 15 },
		{ 250, 306, 15 },
		{ 250, 307, 15 },
		{ 250, 308, 15 },
		{ 250, 309, 15 },
		{ 250, 310, 15 },
		{ 250, 311, 15 },
		{ 250, 312, 15 },
		{ 250, 313, 15 },
		{ 250, 314, 15 },
		{ 250, 315, 15 },
		{ 250, 316, 15 },
		{ 250, 317, 15 },
		{ 250, 318, 15 },
		{ 250, 319, 15 },
		{ 250, 320, 15 },
		{ 250, 321, 15 },
		{ 250, 322, 15 },
		{ 250, 323, 15 },
		{ 250, 324, 15 },
		{ 250, 325, 15 },
		{ 250, 326, 15 },
		{ 250, 327, 15 },
		{ 250, 328, 15 },
		{ 250, 329, 15 },
		{ 250, 330, 15 },
		{ 250, 331, 15 },
		{ 250, 332, 15 },
		{ 250, 333, 15 },
		{ 250, 334, 15 },
		{ 250, 335, 15 },
		{ 250, 336, 15 },
		{ 250, 337, 15 },
		{ 250, 338, 15 },
		{ 250, 339, 15 },
		{ 250, 340, 15 },
		{ 250, 341, 15 },
		{ 250, 342, 15 },
		{ 250, 343, 15 },
		{ 250, 344, 15 },
		{ 250, 345, 15 },
		{ 250, 346, 15 },
		{ 250, 347, 15 },
		{ 250, 348, 15 },
		{ 250, 349, 15 },
		{ 250, 350, 15 },
		{ 250, 351, 15 },
		{ 250, 352, 15 },
		{ 250, 353, 15 },
		{ 250, 354, 15 },
		{ 250, 355, 15 },
		{ 250, 356, 15 },
		{ 250, 357, 15 },
		{ 250, 358, 15 },
		{ 250, 359, 15 },
		{ 250, 360, 15 },
		{ 250, 361, 15 },
		{ 250, 362, 15 },
		{ 250, 363, 15 },
		{ 250, 364, 15 },
		{ 250, 365, 15 },
		{ 250, 366, 15 },
		{ 250, 367, 15 },
		{ 250, 368, 15 },
		{ 250, 369, 15 },
		{ 250, 370, 15 },
		{ 250, 371, 15 },
		{ 250, 372, 15 },
		{ 250, 373, 15 },
		{ 250, 374, 15 },
		{ 250, 375, 15 },
		{ 250, 376, 15 },
		{ 250, 377, 15 },
		{ 250, 378, 15 },
		{ 250, 379, 15 },
		{ 250, 380, 15 },
		{ 250, 381, 15 },
		{ 250, 382, 15 },
		{ 250, 383, 15 },
		{ 251, 274, 15 },
		{ 251, 275, 15 },
		{ 251, 276, 15 },
		{ 251, 277, 15 },
		{ 251, 278, 15 },
		{ 251, 279, 15 },
		{ 251, 280, 15 },
		{ 251, 281, 15 },
		{ 251, 282, 15 },
		{ 251, 283, 15 },
		{ 251, 284, 15 },
		{ 251, 285, 15 },
		{ 251, 286, 15 },
		{ 251, 287, 15 },
		{ 251, 288, 15 },
		{ 251, 289, 15 },
		{ 251, 290, 15 },
		{ 251, 291, 15 },
		{ 251, 292, 15 },
		{ 251, 293, 15 },
		{ 251, 294, 15 },
		{ 251, 295, 15 },
		{ 251, 296, 15 },
		{ 251, 297, 15 },
		{ 251, 298, 15 },
		{ 251, 299, 15 },
		{ 251, 300, 15 },
		{ 251, 301, 15 },
		{ 251, 302, 15 },
		{ 251, 303, 15 },
		{ 251, 304, 15 },
		{ 251, 305, 15 },
		{ 251, 306, 15 },
		{ 251, 307, 15 },
		{ 251, 308, 15 },
		{ 251, 309, 15 },
		{ 251, 310, 15 },
		{ 251, 311, 15 },
		{ 251, 312, 15 },
		{ 251, 313, 15 },
		{ 251, 314, 15 },
		{ 251, 315, 15 },
		{ 251, 316, 15 },
		{ 251, 317, 15 },
		{ 251, 318, 15 },
		{ 251, 319, 15 },
		{ 251, 320, 15 },
		{ 251, 321, 15 },
		{ 251, 322, 15 },
		{ 251, 323, 15 },
		{ 251, 324, 15 },
		{ 251, 325, 15 },
		{ 251, 326, 15 },
		{ 251, 327, 15 },
		{ 251, 328, 15 },
		{ 251, 329, 15 },
		{ 251, 330, 15 },
		{ 251, 331, 15 },
		{ 251, 332, 15 },
		{ 251, 333, 15 },
		{ 251, 334, 15 },
		{ 251, 335, 15 },
		{ 251, 336, 15 },
		{ 251, 337, 15 },
		{ 251, 338, 15 },
		{ 251, 339, 15 },
		{ 251, 340, 15 },
		{ 251, 341, 15 },
		{ 251, 342, 15 },
		{ 251, 343, 15 },
		{ 251, 344, 15 },
		{ 251, 345, 15 },
		{ 251, 346, 15 },
		{ 251, 347, 15 },
		{ 251, 348, 15 },
		{ 251, 349, 15 },
		{ 251, 350, 15 },
		{ 251, 351, 15 },
		{ 251, 352, 15 },
		{ 251, 353, 15 },
		{ 251, 354, 15 },
		{ 251, 355, 15 },
		{ 251, 356, 15 },
		{ 251, 357, 15 },
		{ 251, 358, 15 },
		{ 251, 359, 15 },
		{ 251, 360, 15 },
		{ 251, 361, 15 },
		{ 251, 362, 15 },
		{ 251, 363, 15 },
		{ 251, 364, 15 },
		{ 251, 365, 15 },
		{ 251, 366, 15 },
		{ 251, 367, 15 },
		{ 251, 368, 15 },
		{ 251, 369, 15 },
		{ 251, 370, 15 },
		{ 251, 371, 15 },
		{ 251, 372, 15 },
		{ 251, 373, 15 },
		{ 251, 374, 15 },
		{ 251, 375, 15 },
		{ 251, 376, 15 },
		{ 251, 377, 15 },
		{ 251, 378, 15 },
		{ 251, 379, 15 },
		{ 251, 380, 15 },
		{ 251, 381, 15 },
		{ 251, 382, 15 },
		{ 252, 274, 15 },
		{ 252, 275, 15 },
		{ 252, 276, 15 },
		{ 252, 277, 15 },
		{ 252, 278, 15 },
		{ 252, 279, 15 },
		{ 252, 280, 15 },
		{ 252, 281, 15 },
		{ 252, 282, 15 },
		{ 252, 283, 15 },
		{ 252, 284, 15 },
		{ 252, 285, 15 },
		{ 252, 286, 15 },
		{ 252, 287, 15 },
		{ 252, 288, 15 },
		{ 252, 289, 15 },
		{ 252, 290, 15 },
		{ 252, 291, 15 },
		{ 252, 292, 15 },
		{ 252, 293, 15 },
		{ 252, 294, 15 },
		{ 252, 295, 15 },
		{ 252, 296, 15 },
		{ 252, 297, 15 },
		{ 252, 298, 15 },
		{ 252, 299, 15 },
		{ 252, 300, 15 },
		{ 252, 301, 15 },
		{ 252, 302, 15 },
		{ 252, 303, 15 },
		{ 252, 304, 15 },
		{ 252, 305, 15 },
		{ 252, 306, 15 },
		{ 252, 307, 15 },
		{ 252, 308, 15 },
		{ 252, 309, 15 },
		{ 252, 310, 15 },
		{ 252, 311, 15 },
		{ 252, 312, 15 },
		{ 252, 313, 15 },
		{ 252, 314, 15 },
		{ 252, 315, 15 },
		{ 252, 316, 15 },
		{ 252, 317, 15 },
		{ 252, 318, 15 },
		{ 252, 319, 15 },
		{ 252, 320, 15 },
		{ 252, 321, 15 },
		{ 252, 322, 15 },
		{ 252, 323, 15 },
		{ 252, 324, 15 },
		{ 252, 325, 15 },
		{ 252, 326, 15 },
		{ 252, 327, 15 },
		{ 252, 328, 15 },
		{ 252, 329, 15 },
		{ 252, 330, 15 },
		{ 252, 331, 15 },
		{ 252, 332, 15 },
		{ 252, 333, 15 },
		{ 252, 334, 15 },
		{ 252, 335, 15 },
		{ 252, 336, 15 },
		{ 252, 337, 15 },
		{ 252, 338, 15 },
		{ 252, 339, 15 },
		{ 252, 340, 15 },
		{ 252, 341, 15 },
		{ 252, 342, 15 },
		{ 252, 343, 15 },
		{ 252, 344, 15 },
		{ 252, 345, 15 },
		{ 252, 346, 15 },
		{ 252, 347, 15 },
		{ 252, 348, 15 },
		{ 252, 349, 15 },
		{ 252, 350, 15 },
		{ 252, 351, 15 },
		{ 252, 352, 15 },
		{ 252, 353, 15 },
		{ 252, 354, 15 },
		{ 252, 355, 15 },
		{ 252, 356, 15 },
		{ 252, 357, 15 },
		{ 252, 358, 15 },
		{ 252, 359, 15 },
		{ 252, 360, 15 },
		{ 252, 361, 15 },
		{ 252, 362, 15 },
		{ 252, 363, 15 },
		{ 252, 364, 15 },
		{ 252, 365, 15 },
		{ 252, 366, 15 },
		{ 252, 367, 15 },
		{ 252, 368, 15 },
		{ 252, 369, 15 },
		{ 252, 370, 15 },
		{ 252, 371, 15 },
		{ 252, 372, 15 },
		{ 252, 373, 15 },
		{ 252, 374, 15 },
		{ 252, 375, 15 },
		{ 252, 376, 15 },
		{ 252, 377, 15 },
		{ 252, 378, 15 },
		{ 252, 379, 15 },
		{ 252, 380, 15 },
		{ 252, 381, 15 },
		{ 252, 382, 15 },
		{ 253, 274, 15 },
		{ 253, 275, 15 },
		{ 253, 276, 15 },
		{ 253, 277, 15 },
		{ 253, 278, 15 },
		{ 253, 279, 15 },
		{ 253, 280, 15 },
		{ 253, 281, 15 },
		{ 253, 282, 15 },
		{ 253, 283, 15 },
		{ 253, 284, 15 },
		{ 253, 285, 15 },
		{ 253, 286, 15 },
		{ 253, 287, 15 },
		{ 253, 288, 15 },
		{ 253, 289, 15 },
		{ 253, 290, 15 },
		{ 253, 291, 15 },
		{ 253, 292, 15 },
		{ 253, 293, 15 },
		{ 253, 294, 15 },
		{ 253, 295, 15 },
		{ 253, 296, 15 },
		{ 253, 297, 15 },
		{ 253, 298, 15 },
		{ 253, 299, 15 },
		{ 253, 300, 15 },
		{ 253, 301, 15 },
		{ 253, 302, 15 },
		{ 253, 303, 15 },
		{ 253, 304, 15 },
		{ 253, 305, 15 },
		{ 253, 306, 15 },
		{ 253, 307, 15 },
		{ 253, 308, 15 },
		{ 253, 309, 15 },
		{ 253, 310, 15 },
		{ 253, 311, 15 },
		{ 253, 312, 15 },
		{ 253, 313, 15 },
		{ 253, 314, 15 },
		{ 253, 315, 15 },
		{ 253, 316, 15 },
		{ 253, 317, 15 },
		{ 253, 318, 15 },
		{ 253, 319, 15 },
		{ 253, 320, 15 },
		{ 253, 321, 15 },
		{ 253, 322, 15 },
		{ 253, 323, 15 },
		{ 253, 324, 15 },
		{ 253, 325, 15 },
		{ 253, 326, 15 },
		{ 253, 327, 15 },
		{ 253, 328, 15 },
		{ 253, 329, 15 },
		{ 253, 330, 15 },
		{ 253, 331, 15 },
		{ 253, 332, 15 },
		{ 253, 333, 15 },
		{ 253, 334, 15 },
		{ 253, 335, 15 },
		{ 253, 336, 15 },
		{ 253, 337, 15 },
		{ 253, 338, 15 },
		{ 253, 339, 15 },
		{ 253, 340, 15 },
		{ 253, 341, 15 },
		{ 253, 342, 15 },
		{ 253, 343, 15 },
		{ 253, 344, 15 },
		{ 253, 345, 15 },
		{ 253, 346, 15 },
		{ 253, 347, 15 },
		{ 253, 348, 15 },
		{ 253, 349, 15 },
		{ 253, 350, 15 },
		{ 253, 351, 15 },
		{ 253, 352, 15 },
		{ 253, 353, 15 },
		{ 253, 354, 15 },
		{ 253, 355, 15 },
		{ 253, 356, 15 },
		{ 253, 357, 15 },
		{ 253, 358, 15 },
		{ 253, 359, 15 },
		{ 253, 360, 15 },
		{ 253, 361, 15 },
		{ 253, 362, 15 },
		{ 253, 363, 15 },
		{ 253, 364, 15 },
		{ 253, 365, 15 },
		{ 253, 366, 15 },
		{ 253, 367, 15 },
		{ 253, 368, 15 },
		{ 253, 369, 15 },
		{ 253, 370, 15 },
		{ 253, 371, 15 },
		{ 253, 372, 15 },
		{ 253, 373, 15 },
		{ 253, 374, 15 },
		{ 253, 375, 15 },
		{ 253, 376, 15 },
		{ 253, 377, 15 },
		{ 253, 378, 15 },
		{ 253, 379, 15 },
		{ 253, 380, 15 },
		{ 253, 381, 15 },
		{ 253, 382, 15 },
		{ 254, 274, 15 },
		{ 254, 275, 15 },
		{ 254, 276, 15 },
		{ 254, 277, 15 },
		{ 254, 278, 15 },
		{ 254, 279, 15 },
		{ 254, 280, 15 },
		{ 254, 281, 15 },
		{ 254, 282, 15 },
		{ 254, 283, 15 },
		{ 254, 284, 15 },
		{ 254, 285, 15 },
		{ 254, 286, 15 },
		{ 254, 287, 15 },
		{ 254, 288, 15 },
		{ 254, 289, 15 },
		{ 254, 290, 15 },
		{ 254, 291, 15 },
		{ 254, 292, 15 },
		{ 254, 293, 15 },
		{ 254, 294, 15 },
		{ 254, 295, 15 },
		{ 254, 296, 15 },
		{ 254, 297, 15 },
		{ 254, 298, 15 },
		{ 254, 299, 15 },
		{ 254, 300, 15 },
		{ 254, 301, 15 },
		{ 254, 302, 15 },
		{ 254, 303, 15 },
		{ 254, 304, 15 },
		{ 254, 305, 15 },
		{ 254, 306, 15 },
		{ 254, 307, 15 },
		{ 254, 308, 15 },
		{ 254, 309, 15 },
		{ 254, 310, 15 },
		{ 254, 311, 15 },
		{ 254, 312, 15 },
		{ 254, 313, 15 },
		{ 254, 314, 15 },
		{ 254, 315, 15 },
		{ 254, 316, 15 },
		{ 254, 317, 15 },
		{ 254, 318, 15 },
		{ 254, 319, 15 },
		{ 254, 320, 15 },
		{ 254, 321, 15 },
		{ 254, 322, 15 },
		{ 254, 323, 15 },
		{ 254, 324, 15 },
		{ 254, 325, 15 },
		{ 254, 326, 15 },
		{ 254, 327, 15 },
		{ 254, 328, 15 },
		{ 254, 329, 15 },
		{ 254, 330, 15 },
		{ 254, 331, 15 },
		{ 254, 332, 15 },
		{ 254, 333, 15 },
		{ 254, 334, 15 },
		{ 254, 335, 15 },
		{ 254, 336, 15 },
		{ 254, 337, 15 },
		{ 254, 338, 15 },
		{ 254, 339, 15 },
		{ 254, 340, 15 },
		{ 254, 341, 15 },
		{ 254, 342, 15 },
		{ 254, 343, 15 },
		{ 254, 344, 15 },
		{ 254, 345, 15 },
		{ 254, 346, 15 },
		{ 254, 347, 15 },
		{ 254, 348, 15 },
		{ 254, 349, 15 },
		{ 254, 350, 15 },
		{ 254, 351, 15 },
		{ 254, 352, 15 },
		{ 254, 353, 15 },
		{ 254, 354, 15 },
		{ 254, 355, 15 },
		{ 254, 356, 15 },
		{ 254, 357, 15 },
		{ 254, 358, 15 },
		{ 254, 359, 15 },
		{ 254, 360, 15 },
		{ 254, 361, 15 },
		{ 254, 362, 15 },
		{ 254, 363, 15 },
		{ 254, 364, 15 },
		{ 254, 365, 15 },
		{ 254, 366, 15 },
		{ 254, 367, 15 },
		{ 254, 368, 15 },
		{ 254, 369, 15 },
		{ 254, 370, 15 },
		{ 254, 371, 15 },
		{ 254, 372, 15 },
		{ 254, 373, 15 },
		{ 254, 374, 15 },
		{ 254, 375, 15 },
		{ 254, 376, 15 },
		{ 254, 377, 15 },
		{ 254, 378, 15 },
		{ 254, 379, 15 },
		{ 254, 380, 15 },
		{ 254, 381, 15 },
		{ 254, 382, 15 },
		{ 255, 274, 15 },
		{ 255, 275, 15 },
		{ 255, 276, 15 },
		{ 255, 277, 15 },
		{ 255, 278, 15 },
		{ 255, 279, 15 },
		{ 255, 280, 15 },
		{ 255, 281, 15 },
		{ 255, 282, 15 },
		{ 255, 283, 15 },
		{ 255, 284, 15 },
		{ 255, 285, 15 },
		{ 255, 286, 15 },
		{ 255, 287, 15 },
		{ 255, 288, 15 },
		{ 255, 289, 15 },
		{ 255, 290, 15 },
		{ 255, 291, 15 },
		{ 255, 292, 15 },
		{ 255, 293, 15 },
		{ 255, 294, 15 },
		{ 255, 295, 15 },
		{ 255, 296, 15 },
		{ 255, 297, 15 },
		{ 255, 298, 15 },
		{ 255, 299, 15 },
		{ 255, 300, 15 },
		{ 255, 301, 15 },
		{ 255, 302, 15 },
		{ 255, 303, 15 },
		{ 255, 304, 15 },
		{ 255, 305, 15 },
		{ 255, 306, 15 },
		{ 255, 307, 15 },
		{ 255, 308, 15 },
		{ 255, 309, 15 },
		{ 255, 310, 15 },
		{ 255, 311, 15 },
		{ 255, 312, 15 },
		{ 255, 313, 15 },
		{ 255, 314, 15 },
		{ 255, 315, 15 },
		{ 255, 316, 15 },
		{ 255, 317, 15 },
		{ 255, 318, 15 },
		{ 255, 319, 15 },
		{ 255, 320, 15 },
		{ 255, 321, 15 },
		{ 255, 322, 15 },
		{ 255, 323, 15 },
		{ 255, 324, 15 },
		{ 255, 325, 15 },
		{ 255, 326, 15 },
		{ 255, 327, 15 },
		{ 255, 328, 15 },
		{ 255, 329, 15 },
		{ 255, 330, 15 },
		{ 255, 331, 15 },
		{ 255, 332, 15 },
		{ 255, 333, 15 },
		{ 255, 334, 15 },
		{ 255, 335, 15 },
		{ 255, 336, 15 },
		{ 255, 337, 15 },
		{ 255, 338, 15 },
		{ 255, 339, 15 },
		{ 255, 340, 15 },
		{ 255, 341, 15 },
		{ 255, 342, 15 },
		{ 255, 343, 15 },
		{ 255, 344, 15 },
		{ 255, 345, 15 },
		{ 255, 346, 15 },
		{ 255, 347, 15 },
		{ 255, 348, 15 },
		{ 255, 349, 15 },
		{ 255, 350, 15 },
		{ 255, 351, 15 },
		{ 255, 352, 15 },
		{ 255, 353, 15 },
		{ 255, 354, 15 },
		{ 255, 355, 15 },
		{ 255, 356, 15 },
		{ 255, 357, 15 },
		{ 255, 358, 15 },
		{ 255, 359, 15 },
		{ 255, 360, 15 },
		{ 255, 361, 15 },
		{ 255, 362, 15 },
		{ 255, 363, 15 },
		{ 255, 364, 15 },
		{ 255, 365, 15 },
		{ 255, 366, 15 },
		{ 255, 367, 15 },
		{ 255, 368, 15 },
		{ 255, 369, 15 },
		{ 255, 370, 15 },
		{ 255, 371, 15 },
		{ 255, 372, 15 },
		{ 255, 373, 15 },
		{ 255, 374, 15 },
		{ 255, 375, 15 },
		{ 255, 376, 15 },
		{ 255, 377, 15 },
		{ 255, 378, 15 },
		{ 255, 379, 15 },
		{ 255, 380, 15 },
		{ 255, 381, 15 },
		{ 256, 274, 15 },
		{ 256, 275, 15 },
		{ 256, 276, 15 },
		{ 256, 277, 15 },
		{ 256, 278, 15 },
		{ 256, 279, 15 },
		{ 256, 280, 15 },
		{ 256, 281, 15 },
		{ 256, 282, 15 },
		{ 256, 283, 15 },
		{ 256, 284, 15 },
		{ 256, 285, 15 },
		{ 256, 286, 15 },
		{ 256, 287, 15 },
		{ 256, 288, 15 },
		{ 256, 289, 15 },
		{ 256, 290, 15 },
		{ 256, 291, 15 },
		{ 256, 292, 15 },
		{ 256, 293, 15 },
		{ 256, 294, 15 },
		{ 256, 295, 15 },
		{ 256, 296, 15 },
		{ 256, 297, 15 },
		{ 256, 298, 15 },
		{ 256, 299, 15 },
		{ 256, 300, 15 },
		{ 256, 301, 15 },
		{ 256, 302, 15 },
		{ 256, 303, 15 },
		{ 256, 304, 15 },
		{ 256, 305, 15 },
		{ 256, 306, 15 },
		{ 256, 307, 15 },
		{ 256, 308, 15 },
		{ 256, 309, 15 },
		{ 256, 310, 15 },
		{ 256, 311, 15 },
		{ 256, 312, 15 },
		{ 256, 313, 15 },
		{ 256, 314, 15 },
		{ 256, 315, 15 },
		{ 256, 316, 15 },
		{ 256, 317, 15 },
		{ 256, 318, 15 },
		{ 256, 319, 15 },
		{ 256, 320, 15 },
		{ 256, 321, 15 },
		{ 256, 322, 15 },
		{ 256, 323, 15 },
		{ 256, 324, 15 },
		{ 256, 325, 15 },
		{ 256, 326, 15 },
		{ 256, 327, 15 },
		{ 256, 328, 15 },
		{ 256, 329, 15 },
		{ 256, 330, 15 },
		{ 256, 331, 15 },
		{ 256, 332, 15 },
		{ 256, 333, 15 },
		{ 256, 334, 15 },
		{ 256, 335, 15 },
		{ 256, 336, 15 },
		{ 256, 337, 15 },
		{ 256, 338, 15 },
		{ 256, 339, 15 },
		{ 256, 340, 15 },
		{ 256, 341, 15 },
		{ 256, 342, 15 },
		{ 256, 343, 15 },
		{ 256, 344, 15 },
		{ 256, 345, 15 },
		{ 256, 346, 15 },
		{ 256, 347, 15 },
		{ 256, 348, 15 },
		{ 256, 349, 15 },
		{ 256, 350, 15 },
		{ 256, 351, 15 },
		{ 256, 352, 15 },
		{ 256, 353, 15 },
		{ 256, 354, 15 },
		{ 256, 355, 15 },
		{ 256, 356, 15 },
		{ 256, 357, 15 },
		{ 256, 358, 15 },
		{ 256, 359, 15 },
		{ 256, 360, 15 },
		{ 256, 361, 15 },
		{ 256, 362, 15 },
		{ 256, 363, 15 },
		{ 256, 364, 15 },
		{ 256, 365, 15 },
		{ 256, 366, 15 },
		{ 256, 367, 15 },
		{ 256, 368, 15 },
		{ 256, 369, 15 },
		{ 256, 370, 15 },
		{ 256, 371, 15 },
		{ 256, 372, 15 },
		{ 256, 373, 15 },
		{ 256, 374, 15 },
		{ 256, 375, 15 },
		{ 256, 376, 15 },
		{ 256, 377, 15 },
		{ 256, 378, 15 },
		{ 256, 379, 15 },
		{ 256, 380, 15 },
		{ 258, 134, 15 },
		{ 258, 135, 15 },
		{ 258, 136, 15 },
		{ 258, 137, 15 },
		{ 258, 138, 15 },
		{ 258, 139, 15 },
		{ 258, 140, 15 },
		{ 258, 141, 15 },
		{ 258, 142, 15 },
		{ 258, 143, 15 },
		{ 258, 144, 15 },
		{ 258, 145, 15 },
		{ 258, 146, 15 },
		{ 258, 147, 15 },
		{ 258, 148, 15 },
		{ 258, 149, 15 },
		{ 258, 150, 15 },
		{ 258, 151, 15 },
		{ 258, 152, 15 },
		{ 258, 153, 15 },
		{ 258, 154, 15 },
		{ 258, 155, 15 },
		{ 258, 156, 15 },
		{ 258, 157, 15 },
		{ 258, 158, 15 },
		{ 258, 159, 15 },
		{ 258, 160, 15 },
		{ 258, 161, 15 },
		{ 258, 162, 15 },
		{ 258, 163, 15 },
		{ 258, 164, 15 },
		{ 258, 165, 15 },
		{ 258, 166, 15 },
		{ 258, 167, 15 },
		{ 258, 168, 15 },
		{ 258, 169, 15 },
		{ 258, 170, 15 },
		{ 258, 171, 15 },
		{ 258, 172, 15 },
		{ 258, 173, 15 },
		{ 258, 174, 15 },
		{ 258, 175, 15 },
		{ 258, 176, 15 },
		{ 258, 177, 15 },
		{ 258, 178, 15 },
		{ 258, 179, 15 },
		{ 258, 180, 15 },
		{ 258, 181, 15 },
		{ 258, 182, 15 },
		{ 258, 183, 15 },
		{ 258, 184, 15 },
		{ 258, 185, 15 },
		{ 258, 186, 15 },
		{ 258, 187, 15 },
		{ 258, 188, 15 },
		{ 258, 189, 15 },
		{ 258, 190, 15 },
		{ 258, 191, 15 },
		{ 258, 192, 15 },
		{ 258, 193, 15 },
		{ 258, 194, 15 },
		{ 258, 195, 15 },
		{ 258, 196, 15 },
		{ 258, 197, 15 },
		{ 258, 198, 15 },
		{ 258, 199, 15 },
		{ 258, 200, 15 },
		{ 258, 201, 15 },
		{ 258, 202, 15 },
		{ 258, 203, 15 },
		{ 258, 204, 15 },
		{ 258, 205, 15 },
		{ 258, 206, 15 },
		{ 258, 207, 15 },
		{ 258, 208, 15 },
		{ 258, 209, 15 },
		{ 258, 210, 15 },
		{ 258, 211, 15 },
		{ 258, 212, 15 },
		{ 258, 213, 15 },
		{ 258, 214, 15 },
		{ 258, 215, 15 },
		{ 258, 216, 15 },
		{ 258, 217, 15 },
		{ 258, 218, 15 },
		{ 258, 219, 15 },
		{ 258, 220, 15 },
		{ 258, 221, 15 },
		{ 258, 222, 15 },
		{ 258, 223, 15 },
		{ 258, 224, 15 },
		{ 258, 225, 15 },
		{ 258, 226, 15 },
		{ 258, 227, 15 },
		{ 258, 228, 15 },
		{ 258, 229, 15 },
		{ 258, 230, 15 },
		{ 258, 231, 15 },
		{ 258, 232, 15 },
		{ 258, 233, 15 },
		{ 258, 234, 15 },
		{ 258, 235, 15 },
		{ 258, 236, 15 },
		{ 258, 237, 15 },
		{ 258, 238, 15 },
		{ 258, 239, 15 },
		{ 258, 240, 15 },
		{ 259, 133, 15 },
		{ 259, 134, 15 },
		{ 259, 135, 15 },
		{ 259, 136, 15 },
		{ 259, 137, 15 },
		{ 259, 138, 15 },
		{ 259, 139, 15 },
		{ 259, 140, 15 },
		{ 259, 141, 15 },
		{ 259, 142, 15 },
		{ 259, 143, 15 },
		{ 259, 144, 15 },
		{ 259, 145, 15 },
		{ 259, 146, 15 },
		{ 259, 147, 15 },
		{ 259, 148, 15 },
		{ 259, 149, 15 },
		{ 259, 150, 15 },
		{ 259, 151, 15 },
		{ 259, 152, 15 },
		{ 259, 153, 15 },
		{ 259, 154, 15 },
		{ 259, 155, 15 },
		{ 259, 156, 15 },
		{ 259, 157, 15 },
		{ 259, 158, 15 },
		{ 259, 159, 15 },
		{ 259, 160, 15 },
		{ 259, 161, 15 },
		{ 259, 162, 15 },
		{ 259, 163, 15 },
		{ 259, 164, 15 },
		{ 259, 165, 15 },
		{ 259, 166, 15 },
		{ 259, 167, 15 },
		{ 259, 168, 15 },
		{ 259, 169, 15 },
		{ 259, 170, 15 },
		{ 259, 171, 15 },
		{ 259, 172, 15 },
		{ 259, 173, 15 },
		{ 259, 174, 15 },
		{ 259, 175, 15 },
		{ 259, 176, 15 },
		{ 259, 177, 15 },
		{ 259, 178, 15 },
		{ 259, 179, 15 },
		{ 259, 180, 15 },
		{ 259, 181, 15 },
		{ 259, 182, 15 },
		{ 259, 183, 15 },
		{ 259, 184, 15 },
		{ 259, 185, 15 },
		{ 259, 186, 15 },
		{ 259, 187, 15 },
		{ 259, 188, 15 },
		{ 259, 189, 15 },
		{ 259, 190, 15 },
		{ 259, 191, 15 },
		{ 259, 192, 15 },
		{ 259, 193, 15 },
		{ 259, 194, 15 },
		{ 259, 195, 15 },
		{ 259, 196, 15 },
		{ 259, 197, 15 },
		{ 259, 198, 15 },
		{ 259, 199, 15 },
		{ 259, 200, 15 },
		{ 259, 201, 15 },
		{ 259, 202, 15 },
		{ 259, 203, 15 },
		{ 259, 204, 15 },
		{ 259, 205, 15 },
		{ 259, 206, 15 },
		{ 259, 207, 15 },
		{ 259, 208, 15 },
		{ 259, 209, 15 },
		{ 259, 210, 15 },
		{ 259, 211, 15 },
		{ 259, 212, 15 },
		{ 259, 213, 15 },
		{ 259, 214, 15 },
		{ 259, 215, 15 },
		{ 259, 216, 15 },
		{ 259, 217, 15 },
		{ 259, 218, 15 },
		{ 259, 219, 15 },
		{ 259, 220, 15 },
		{ 259, 221, 15 },
		{ 259, 222, 15 },
		{ 259, 223, 15 },
		{ 259, 224, 15 },
		{ 259, 225, 15 },
		{ 259, 226, 15 },
		{ 259, 227, 15 },
		{ 259, 228, 15 },
		{ 259, 229, 15 },
		{ 259, 230, 15 },
		{ 259, 231, 15 },
		{ 259, 232, 15 },
		{ 259, 233, 15 },
		{ 259, 234, 15 },
		{ 259, 235, 15 },
		{ 259, 236, 15 },
		{ 259, 237, 15 },
		{ 259, 238, 15 },
		{ 259, 239, 15 },
		{ 259, 240, 15 },
		{ 260, 132, 15 },
		{ 260, 133, 15 },
		{ 260, 134, 15 },
		{ 260, 135, 15 },
		{ 260, 136, 15 },
		{ 260, 137, 15 },
		{ 260, 138, 15 },
		{ 260, 139, 15 },
		{ 260, 140, 15 },
		{ 260, 141, 15 },
		{ 260, 142, 15 },
		{ 260, 143, 15 },
		{ 260, 144, 15 },
		{ 260, 145, 15 },
		{ 260, 146, 15 },
		{ 260, 147, 15 },
		{ 260, 148, 15 },
		{ 260, 149, 15 },
		{ 260, 150, 15 },
		{ 260, 151, 15 },
		{ 260, 152, 15 },
		{ 260, 153, 15 },
		{ 260, 154, 15 },
		{ 260, 155, 15 },
		{ 260, 156, 15 },
		{ 260, 157, 15 },
		{ 260, 158, 15 },
		{ 260, 159, 15 },
		{ 260, 160, 15 },
		{ 260, 161, 15 },
		{ 260, 162, 15 },
		{ 260, 163, 15 },
		{ 260, 164, 15 },
		{ 260, 165, 15 },
		{ 260, 166, 15 },
		{ 260, 167, 15 },
		{ 260, 168, 15 },
		{ 260, 169, 15 },
		{ 260, 170, 15 },
		{ 260, 171, 15 },
		{ 260, 172, 15 },
		{ 260, 173, 15 },
		{ 260, 174, 15 },
		{ 260, 175, 15 },
		{ 260, 176, 15 },
		{ 260, 177, 15 },
		{ 260, 178, 15 },
		{ 260, 179, 15 },
		{ 260, 180, 15 },
		{ 260, 181, 15 },
		{ 260, 182, 15 },
		{ 260, 183, 15 },
		{ 260, 184, 15 },
		{ 260, 185, 15 },
		{ 260, 186, 15 },
		{ 260, 187, 15 },
		{ 260, 188, 15 },
		{ 260, 189, 15 },
		{ 260, 190, 15 },
		{ 260, 191, 15 },
		{ 260, 192, 15 },
		{ 260, 193, 15 },
		{ 260, 194, 15 },
		{ 260, 195, 15 },
		{ 260, 196, 15 },
		{ 260, 197, 15 },
		{ 260, 198, 15 },
		{ 260, 199, 15 },
		{ 260, 200, 15 },
		{ 260, 201, 15 },
		{ 260, 202, 15 },
		{ 260, 203, 15 },
		{ 260, 204, 15 },
		{ 260, 205, 15 },
		{ 260, 206, 15 },
		{ 260, 207, 15 },
		{ 260, 208, 15 },
		{ 260, 209, 15 },
		{ 260, 210, 15 },
		{ 260, 211, 15 },
		{ 260, 212, 15 },
		{ 260, 213, 15 },
		{ 260, 214, 15 },
		{ 260, 215, 15 },
		{ 260, 216, 15 },
		{ 260, 217, 15 },
		{ 260, 218, 15 },
		{ 260, 219, 15 },
		{ 260, 220, 15 },
		{ 260, 221, 15 },
		{ 260, 222, 15 },
		{ 260, 223, 15 },
		{ 260, 224, 15 },
		{ 260, 225, 15 },
		{ 260, 226, 15 },
		{ 260, 227, 15 },
		{ 260, 228, 15 },
		{ 260, 229, 15 },
		{ 260, 230, 15 },
		{ 260, 231, 15 },
		{ 260, 232, 15 },
		{ 260, 233, 15 },
		{ 260, 234, 15 },
		{ 260, 235, 15 },
		{ 260, 236, 15 },
		{ 260, 237, 15 },
		{ 260, 238, 15 },
		{ 260, 239, 15 },
		{ 260, 240, 15 },
		{ 261, 132, 15 },
		{ 261, 133, 15 },
		{ 261, 134, 15 },
		{ 261, 135, 15 },
		{ 261, 136, 15 },
		{ 261, 137, 15 },
		{ 261, 138, 15 },
		{ 261, 139, 15 },
		{ 261, 140, 15 },
		{ 261, 141, 15 },
		{ 261, 142, 15 },
		{ 261, 143, 15 },
		{ 261, 144, 15 },
		{ 261, 145, 15 },
		{ 261, 146, 15 },
		{ 261, 147, 15 },
		{ 261, 148, 15 },
		{ 261, 149, 15 },
		{ 261, 150, 15 },
		{ 261, 151, 15 },
		{ 261, 152, 15 },
		{ 261, 153, 15 },
		{ 261, 154, 15 },
		{ 261, 155, 15 },
		{ 261, 156, 15 },
		{ 261, 157, 15 },
		{ 261, 158, 15 },
		{ 261, 159, 15 },
		{ 261, 160, 15 },
		{ 261, 161, 15 },
		{ 261, 162, 15 },
		{ 261, 163, 15 },
		{ 261, 164, 15 },
		{ 261, 165, 15 },
		{ 261, 166, 15 },
		{ 261, 167, 15 },
		{ 261, 168, 15 },
		{ 261, 169, 15 },
		{ 261, 170, 15 },
		{ 261, 171, 15 },
		{ 261, 172, 15 },
		{ 261, 173, 15 },
		{ 261, 174, 15 },
		{ 261, 175, 15 },
		{ 261, 176, 15 },
		{ 261, 177, 15 },
		{ 261, 178, 15 },
		{ 261, 179, 15 },
		{ 261, 180, 15 },
		{ 261, 181, 15 },
		{ 261, 182, 15 },
		{ 261, 183, 15 },
		{ 261, 184, 15 },
		{ 261, 185, 15 },
		{ 261, 186, 15 },
		{ 261, 187, 15 },
		{ 261, 188, 15 },
		{ 261, 189, 15 },
		{ 261, 190, 15 },
		{ 261, 191, 15 },
		{ 261, 192, 15 },
		{ 261, 193, 15 },
		{ 261, 194, 15 },
		{ 261, 195, 15 },
		{ 261, 196, 15 },
		{ 261, 197, 15 },
		{ 261, 198, 15 },
		{ 261, 199, 15 },
		{ 261, 200, 15 },
		{ 261, 201, 15 },
		{ 261, 202, 15 },
		{ 261, 203, 15 },
		{ 261, 204, 15 },
		{ 261, 205, 15 },
		{ 261, 206, 15 },
		{ 261, 207, 15 },
		{ 261, 208, 15 },
		{ 261, 209, 15 },
		{ 261, 210, 15 },
		{ 261, 211, 15 },
		{ 261, 212, 15 },
		{ 261, 213, 15 },
		{ 261, 214, 15 },
		{ 261, 215, 15 },
		{ 261, 216, 15 },
		{ 261, 217, 15 },
		{ 261, 218, 15 },
		{ 261, 219, 15 },
		{ 261, 220, 15 },
		{ 261, 221, 15 },
		{ 261, 222, 15 },
		{ 261, 223, 15 },
		{ 261, 224, 15 },
		{ 261, 225, 15 },
		{ 261, 226, 15 },
		{ 261, 227, 15 },
		{ 261, 228, 15 },
		{ 261, 229, 15 },
		{ 261, 230, 15 },
		{ 261, 231, 15 },
		{ 261, 232, 15 },
		{ 261, 233, 15 },
		{ 261, 234, 15 },
		{ 261, 235, 15 },
		{ 261, 236, 15 },
		{ 261, 237, 15 },
		{ 261, 238, 15 },
		{ 261, 239, 15 },
		{ 261, 240, 15 },
		{ 262, 132, 15 },
		{ 262, 133, 15 },
		{ 262, 134, 15 },
		{ 262, 135, 15 },
		{ 262, 136, 15 },
		{ 262, 137, 15 },
		{ 262, 138, 15 },
		{ 262, 139, 15 },
		{ 262, 140, 15 },
		{ 262, 141, 15 },
		{ 262, 142, 15 },
		{ 262, 143, 15 },
		{ 262, 144, 15 },
		{ 262, 145, 15 },
		{ 262, 146, 15 },
		{ 262, 147, 15 },
		{ 262, 148, 15 },
		{ 262, 149, 15 },
		{ 262, 150, 15 },
		{ 262, 151, 15 },
		{ 262, 152, 15 },
		{ 262, 153, 15 },
		{ 262, 154, 15 },
		{ 262, 155, 15 },
		{ 262, 156, 15 },
		{ 262, 157, 15 },
		{ 262, 158, 15 },
		{ 262, 159, 15 },
		{ 262, 160, 15 },
		{ 262, 161, 15 },
		{ 262, 162, 15 },
		{ 262, 163, 15 },
		{ 262, 164, 15 },
		{ 262, 165, 15 },
		{ 262, 166, 15 },
		{ 262, 167, 15 },
		{ 262, 168, 15 },
		{ 262, 169, 15 },
		{ 262, 170, 15 },
		{ 262, 171, 15 },
		{ 262, 172, 15 },
		{ 262, 173, 15 },
		{ 262, 174, 15 },
		{ 262, 175, 15 },
		{ 262, 176, 15 },
		{ 262, 177, 15 },
		{ 262, 178, 15 },
		{ 262, 179, 15 },
		{ 262, 180, 15 },
		{ 262, 181, 15 },
		{ 262, 182, 15 },
		{ 262, 183, 15 },
		{ 262, 184, 15 },
		{ 262, 185, 15 },
		{ 262, 186, 15 },
		{ 262, 187, 15 },
		{ 262, 188, 15 },
		{ 262, 189, 15 },
		{ 262, 190, 15 },
		{ 262, 191, 15 },
		{ 262, 192, 15 },
		{ 262, 193, 15 },
		{ 262, 194, 15 },
		{ 262, 195, 15 },
		{ 262, 196, 15 },
		{ 262, 197, 15 },
		{ 262, 198, 15 },
		{ 262, 199, 15 },
		{ 262, 200, 15 },
		{ 262, 201, 15 },
		{ 262, 202, 15 },
		{ 262, 203, 15 },
		{ 262, 204, 15 },
		{ 262, 205, 15 },
		{ 262, 206, 15 },
		{ 262, 207, 15 },
		{ 262, 208, 15 },
		{ 262, 209, 15 },
		{ 262, 210, 15 },
		{ 262, 211, 15 },
		{ 262, 212, 15 },
		{ 262, 213, 15 },
		{ 262, 214, 15 },
		{ 262, 215, 15 },
		{ 262, 216, 15 },
		{ 262, 217, 15 },
		{ 262, 218, 15 },
		{ 262, 219, 15 },
		{ 262, 220, 15 },
		{ 262, 221, 15 },
		{ 262, 222, 15 },
		{ 262, 223, 15 },
		{ 262, 224, 15 },
		{ 262, 225, 15 },
		{ 262, 226, 15 },
		{ 262, 227, 15 },
		{ 262, 228, 15 },
		{ 262, 229, 15 },
		{ 262, 230, 15 },
		{ 262, 231, 15 },
		{ 262, 232, 15 },
		{ 262, 233, 15 },
		{ 262, 234, 15 },
		{ 262, 235, 15 },
		{ 262, 236, 15 },
		{ 262, 237, 15 },
		{ 262, 238, 15 },
		{ 262, 239, 15 },
		{ 262, 240, 15 },
		{ 263, 132, 15 },
		{ 263, 133, 15 },
		{ 263, 134, 15 },
		{ 263, 135, 15 },
		{ 263, 136, 15 },
		{ 263, 137, 15 },
		{ 263, 138, 15 },
		{ 263, 139, 15 },
		{ 263, 140, 15 },
		{ 263, 141, 15 },
		{ 263, 142, 15 },
		{ 263, 143, 15 },
		{ 263, 144, 15 },
		{ 263, 145, 15 },
		{ 263, 146, 15 },
		{ 263, 147, 15 },
		{ 263, 148, 15 },
		{ 263, 149, 15 },
		{ 263, 150, 15 },
		{ 263, 151, 15 },
		{ 263, 152, 15 },
		{ 263, 153, 15 },
		{ 263, 154, 15 },
		{ 263, 155, 15 },
		{ 263, 156, 15 },
		{ 263, 157, 15 },
		{ 263, 158, 15 },
		{ 263, 159, 15 },
		{ 263, 160, 15 },
		{ 263, 161, 15 },
		{ 263, 162, 15 },
		{ 263, 163, 15 },
		{ 263, 164, 15 },
		{ 263, 165, 15 },
		{ 263, 166, 15 },
		{ 263, 167, 15 },
		{ 263, 168, 15 },
		{ 263, 169, 15 },
		{ 263, 170, 15 },
		{ 263, 171, 15 },
		{ 263, 172, 15 },
		{ 263, 173, 15 },
		{ 263, 174, 15 },
		{ 263, 175, 15 },
		{ 263, 176, 15 },
		{ 263, 177, 15 },
		{ 263, 178, 15 },
		{ 263, 179, 15 },
		{ 263, 180, 15 },
		{ 263, 181, 15 },
		{ 263, 182, 15 },
		{ 263, 183, 15 },
		{ 263, 184, 15 },
		{ 263, 185, 15 },
		{ 263, 186, 15 },
		{ 263, 187, 15 },
		{ 263, 188, 15 },
		{ 263, 189, 15 },
		{ 263, 190, 15 },
		{ 263, 191, 15 },
		{ 263, 192, 15 },
		{ 263, 193, 15 },
		{ 263, 194, 15 },
		{ 263, 195, 15 },
		{ 263, 196, 15 },
		{ 263, 197, 15 },
		{ 263, 198, 15 },
		{ 263, 199, 15 },
		{ 263, 200, 15 },
		{ 263, 201, 15 },
		{ 263, 202, 15 },
		{ 263, 203, 15 },
		{ 263, 204, 15 },
		{ 263, 205, 15 },
		{ 263, 206, 15 },
		{ 263, 207, 15 },
		{ 263, 208, 15 },
		{ 263, 209, 15 },
		{ 263, 210, 15 },
		{ 263, 211, 15 },
		{ 263, 212, 15 },
		{ 263, 213, 15 },
		{ 263, 214, 15 },
		{ 263, 215, 15 },
		{ 263, 216, 15 },
		{ 263, 217, 15 },
		{ 263, 218, 15 },
		{ 263, 219, 15 },
		{ 263, 220, 15 },
		{ 263, 221, 15 },
		{ 263, 222, 15 },
		{ 263, 223, 15 },
		{ 263, 224, 15 },
		{ 263, 225, 15 },
		{ 263, 226, 15 },
		{ 263, 227, 15 },
		{ 263, 228, 15 },
		{ 263, 229, 15 },
		{ 263, 230, 15 },
		{ 263, 231, 15 },
		{ 263, 232, 15 },
		{ 263, 233, 15 },
		{ 263, 234, 15 },
		{ 263, 235, 15 },
		{ 263, 236, 15 },
		{ 263, 237, 15 },
		{ 263, 238, 15 },
		{ 263, 239, 15 },
		{ 263, 240, 15 },
		{ 264, 131, 15 },
		{ 264, 132, 15 },
		{ 264, 133, 15 },
		{ 264, 134, 15 },
		{ 264, 135, 15 },
		{ 264, 136, 15 },
		{ 264, 137, 15 },
		{ 264, 138, 15 },
		{ 264, 139, 15 },
		{ 264, 140, 15 },
		{ 264, 141, 15 },
		{ 264, 142, 15 },
		{ 264, 143, 15 },
		{ 264, 144, 15 },
		{ 264, 145, 15 },
		{ 264, 146, 15 },
		{ 264, 147, 15 },
		{ 264, 148, 15 },
		{ 264, 149, 15 },
		{ 264, 150, 15 },
		{ 264, 151, 15 },
		{ 264, 152, 15 },
		{ 264, 153, 15 },
		{ 264, 154, 15 },
		{ 264, 155, 15 },
		{ 264, 156, 15 },
		{ 264, 157, 15 },
		{ 264, 158, 15 },
		{ 264, 159, 15 },
		{ 264, 160, 15 },
		{ 264, 161, 15 },
		{ 264, 162, 15 },
		{ 264, 163, 15 },
		{ 264, 164, 15 },
		{ 264, 165, 15 },
		{ 264, 166, 15 },
		{ 264, 167, 15 },
		{ 264, 168, 15 },
		{ 264, 169, 15 },
		{ 264, 170, 15 },
		{ 264, 171, 15 },
		{ 264, 172, 15 },
		{ 264, 173, 15 },
		{ 264, 174, 15 },
		{ 264, 175, 15 },
		{ 264, 176, 15 },
		{ 264, 177, 15 },
		{ 264, 178, 15 },
		{ 264, 179, 15 },
		{ 264, 180, 15 },
		{ 264, 181, 15 },
		{ 264, 182, 15 },
		{ 264, 183, 15 },
		{ 264, 184, 15 },
		{ 264, 185, 15 },
		{ 264, 186, 15 },
		{ 264, 187, 15 },
		{ 264, 188, 15 },
		{ 264, 189, 15 },
		{ 264, 190, 15 },
		{ 264, 191, 15 },
		{ 264, 192, 15 },
		{ 264, 193, 15 },
		{ 264, 194, 15 },
		{ 264, 195, 15 },
		{ 264, 196, 15 },
		{ 264, 197, 15 },
		{ 264, 198, 15 },
		{ 264, 199, 15 },
		{ 264, 200, 15 },
		{ 264, 201, 15 },
		{ 264, 202, 15 },
		{ 264, 203, 15 },
		{ 264, 204, 15 },
		{ 264, 205, 15 },
		{ 264, 206, 15 },
		{ 264, 207, 15 },
		{ 264, 208, 15 },
		{ 264, 209, 15 },
		{ 264, 210, 15 },
		{ 264, 211, 15 },
		{ 264, 212, 15 },
		{ 264, 213, 15 },
		{ 264, 214, 15 },
		{ 264, 215, 15 },
		{ 264, 216, 15 },
		{ 264, 217, 15 },
		{ 264, 218, 15 },
		{ 264, 219, 15 },
		{ 264, 220, 15 },
		{ 264, 221, 15 },
		{ 264, 222, 15 },
		{ 264, 223, 15 },
		{ 264, 224, 15 },
		{ 264, 225, 15 },
		{ 264, 226, 15 },
		{ 264, 227, 15 },
		{ 264, 228, 15 },
		{ 264, 229, 15 },
		{ 264, 230, 15 },
		{ 264, 231, 15 },
		{ 264, 232, 15 },
		{ 264, 233, 15 },
		{ 264, 234, 15 },
		{ 264, 235, 15 },
		{ 264, 236, 15 },
		{ 264, 237, 15 },
		{ 264, 238, 15 },
		{ 264, 239, 15 },
		{ 264, 240, 15 },
		{ 265, 131, 15 },
		{ 265, 132, 15 },
		{ 265, 133, 15 },
		{ 265, 134, 15 },
		{ 265, 135, 15 },
		{ 265, 136, 15 },
		{ 265, 137, 15 },
		{ 265, 138, 15 },
		{ 265, 139, 15 },
		{ 265, 140, 15 },
		{ 265, 141, 15 },
		{ 265, 142, 15 },
		{ 265, 143, 15 },
		{ 265, 144, 15 },
		{ 265, 145, 15 },
		{ 265, 146, 15 },
		{ 265, 147, 15 },
		{ 265, 148, 15 },
		{ 265, 149, 15 },
		{ 265, 150, 15 },
		{ 265, 151, 15 },
		{ 265, 152, 15 },
		{ 265, 153, 15 },
		{ 265, 154, 15 },
		{ 265, 155, 15 },
		{ 265, 156, 15 },
		{ 265, 157, 15 },
		{ 265, 158, 15 },
		{ 265, 159, 15 },
		{ 265, 160, 15 },
		{ 265, 161, 15 },
		{ 265, 162, 15 },
		{ 265, 163, 15 },
		{ 265, 164, 15 },
		{ 265, 165, 15 },
		{ 265, 166, 15 },
		{ 265, 167, 15 },
		{ 265, 168, 15 },
		{ 265, 169, 15 },
		{ 265, 170, 15 },
		{ 265, 171, 15 },
		{ 265, 172, 15 },
		{ 265, 173, 15 },
		{ 265, 174, 15 },
		{ 265, 175, 15 },
		{ 265, 176, 15 },
		{ 265, 177, 15 },
		{ 265, 178, 15 },
		{ 265, 179, 15 },
		{ 265, 180, 15 },
		{ 265, 181, 15 },
		{ 265, 182, 15 },
		{ 265, 183, 15 },
		{ 265, 184, 15 },
		{ 265, 185, 15 },
		{ 265, 186, 15 },
		{ 265, 187, 15 },
		{ 265, 188, 15 },
		{ 265, 189, 15 },
		{ 265, 190, 15 },
		{ 265, 191, 15 },
		{ 265, 192, 15 },
		{ 265, 193, 15 },
		{ 265, 194, 15 },
		{ 265, 195, 15 },
		{ 265, 196, 15 },
		{ 265, 197, 15 },
		{ 265, 198, 15 },
		{ 265, 199, 15 },
		{ 265, 200, 15 },
		{ 265, 201, 15 },
		{ 265, 202, 15 },
		{ 265, 203, 15 },
		{ 265, 204, 15 },
		{ 265, 205, 15 },
		{ 265, 206, 15 },
		{ 265, 207, 15 },
		{ 265, 208, 15 },
		{ 265, 209, 15 },
		{ 265, 210, 15 },
		{ 265, 211, 15 },
		{ 265, 212, 15 },
		{ 265, 213, 15 },
		{ 265, 214, 15 },
		{ 265, 215, 15 },
		{ 265, 216, 15 },
		{ 265, 217, 15 },
		{ 265, 218, 15 },
		{ 265, 219, 15 },
		{ 265, 220, 15 },
		{ 265, 221, 15 },
		{ 265, 222, 15 },
		{ 265, 223, 15 },
		{ 265, 224, 15 },
		{ 265, 225, 15 },
		{ 265, 226, 15 },
		{ 265, 227, 15 },
		{ 265, 228, 15 },
		{ 265, 229, 15 },
		{ 265, 230, 15 },
		{ 265, 231, 15 },
		{ 265, 232, 15 },
		{ 265, 233, 15 },
		{ 265, 234, 15 },
		{ 265, 235, 15 },
		{ 265, 236, 15 },
		{ 265, 237, 15 },
		{ 265, 238, 15 },
		{ 265, 239, 15 },
		{ 265, 240, 15 },
		{ 266, 131, 15 },
		{ 266, 132, 15 },
		{ 266, 133, 15 },
		{ 266, 134, 15 },
		{ 266, 135, 15 },
		{ 266, 136, 15 },
		{ 266, 137, 15 },
		{ 266, 138, 15 },
		{ 266, 139, 15 },
		{ 266, 140, 15 },
		{ 266, 141, 15 },
		{ 266, 142, 15 },
		{ 266, 143, 15 },
		{ 266, 144, 15 },
		{ 266, 145, 15 },
		{ 266, 146, 15 },
		{ 266, 147, 15 },
		{ 266, 148, 15 },
		{ 266, 149, 15 },
		{ 266, 150, 15 },
		{ 266, 151, 15 },
		{ 266, 152, 15 },
		{ 266, 153, 15 },
		{ 266, 154, 15 },
		{ 266, 155, 15 },
		{ 266, 156, 15 },
		{ 266, 157, 15 },
		{ 266, 158, 15 },
		{ 266, 159, 15 },
		{ 266, 160, 15 },
		{ 266, 161, 15 },
		{ 266, 162, 15 },
		{ 266, 163, 15 },
		{ 266, 164, 15 },
		{ 266, 165, 15 },
		{ 266, 166, 15 },
		{ 266, 167, 15 },
		{ 266, 168, 15 },
		{ 266, 169, 15 },
		{ 266, 170, 15 },
		{ 266, 171, 15 },
		{ 266, 172, 15 },
		{ 266, 173, 15 },
		{ 266, 174, 15 },
		{ 266, 175, 15 },
		{ 266, 176, 15 },
		{ 266, 177, 15 },
		{ 266, 178, 15 },
		{ 266, 179, 15 },
		{ 266, 180, 15 },
		{ 266, 181, 15 },
		{ 266, 182, 15 },
		{ 266, 183, 15 },
		{ 266, 184, 15 },
		{ 266, 185, 15 },
		{ 266, 186, 15 },
		{ 266, 187, 15 },
		{ 266, 188, 15 },
		{ 266, 189, 15 },
		{ 266, 190, 15 },
		{ 266, 191, 15 },
		{ 266, 192, 15 },
		{ 266, 193, 15 },
		{ 266, 194, 15 },
		{ 266, 195, 15 },
		{ 266, 196, 15 },
		{ 266, 197, 15 },
		{ 266, 198, 15 },
		{ 266, 199, 15 },
		{ 266, 200, 15 },
		{ 266, 201, 15 },
		{ 266, 202, 15 },
		{ 266, 203, 15 },
		{ 266, 204, 15 },
		{ 266, 205, 15 },
		{ 266, 206, 15 },
		{ 266, 207, 15 },
		{ 266, 208, 15 },
		{ 266, 209, 15 },
		{ 266, 210, 15 },
		{ 266, 211, 15 },
		{ 266, 212, 15 },
		{ 266, 213, 15 },
		{ 266, 214, 15 },
		{ 266, 215, 15 },
		{ 266, 216, 15 },
		{ 266, 217, 15 },
		{ 266, 218, 15 },
		{ 266, 219, 15 },
		{ 266, 220, 15 },
		{ 266, 221, 15 },
		{ 266, 222, 15 },
		{ 266, 223, 15 },
		{ 266, 224, 15 },
		{ 266, 225, 15 },
		{ 266, 226, 15 },
		{ 266, 227, 15 },
		{ 266, 228, 15 },
		{ 266, 229, 15 },
		{ 266, 230, 15 },
		{ 266, 231, 15 },
		{ 266, 232, 15 },
		{ 266, 233, 15 },
		{ 266, 234, 15 },
		{ 266, 235, 15 },
		{ 266, 236, 15 },
		{ 266, 237, 15 },
		{ 266, 238, 15 },
		{ 267, 131, 15 },
		{ 267, 132, 15 },
		{ 267, 133, 15 },
		{ 267, 134, 15 },
		{ 267, 135, 15 },
		{ 267, 136, 15 },
		{ 267, 137, 15 },
		{ 267, 138, 15 },
		{ 267, 139, 15 },
		{ 267, 140, 15 },
		{ 267, 141, 15 },
		{ 267, 142, 15 },
		{ 267, 143, 15 },
		{ 267, 144, 15 },
		{ 267, 145, 15 },
		{ 267, 146, 15 },
		{ 267, 147, 15 },
		{ 267, 148, 15 },
		{ 267, 149, 15 },
		{ 267, 150, 15 },
		{ 267, 151, 15 },
		{ 267, 152, 15 },
		{ 267, 153, 15 },
		{ 267, 154, 15 },
		{ 267, 155, 15 },
		{ 267, 156, 15 },
		{ 267, 157, 15 },
		{ 267, 158, 15 },
		{ 267, 159, 15 },
		{ 267, 160, 15 },
		{ 267, 161, 15 },
		{ 267, 162, 15 },
		{ 267, 163, 15 },
		{ 267, 164, 15 },
		{ 267, 165, 15 },
		{ 267, 166, 15 },
		{ 267, 167, 15 },
		{ 267, 168, 15 },
		{ 267, 169, 15 },
		{ 267, 170, 15 },
		{ 267, 171, 15 },
		{ 267, 172, 15 },
		{ 267, 173, 15 },
		{ 267, 174, 15 },
		{ 267, 175, 15 },
		{ 267, 176, 15 },
		{ 267, 177, 15 },
		{ 267, 178, 15 },
		{ 267, 179, 15 },
		{ 267, 180, 15 },
		{ 267, 181, 15 },
		{ 267, 182, 15 },
		{ 267, 183, 15 },
		{ 267, 184, 15 },
		{ 267, 185, 15 },
		{ 267, 186, 15 },
		{ 267, 187, 15 },
		{ 267, 188, 15 },
		{ 267, 189, 15 },
		{ 267, 190, 15 },
		{ 267, 191, 15 },
		{ 267, 192, 15 },
		{ 267, 193, 15 },
		{ 267, 194, 15 },
		{ 267, 195, 15 },
		{ 267, 196, 15 },
		{ 267, 197, 15 },
		{ 267, 198, 15 },
		{ 267, 199, 15 },
		{ 267, 200, 15 },
		{ 267, 201, 15 },
		{ 267, 202, 15 },
		{ 267, 203, 15 },
		{ 267, 204, 15 },
		{ 267, 205, 15 },
		{ 267, 206, 15 },
		{ 267, 207, 15 },
		{ 267, 208, 15 },
		{ 267, 209, 15 },
		{ 267, 210, 15 },
		{ 267, 211, 15 },
		{ 267, 212, 15 },
		{ 267, 213, 15 },
		{ 267, 214, 15 },
		{ 267, 215, 15 },
		{ 267, 216, 15 },
		{ 267, 217, 15 },
		{ 267, 218, 15 },
		{ 267, 219, 15 },
		{ 267, 220, 15 },
		{ 267, 221, 15 },
		{ 267, 222, 15 },
		{ 267, 223, 15 },
		{ 267, 224, 15 },
		{ 267, 225, 15 },
		{ 267, 226, 15 },
		{ 267, 227, 15 },
		{ 267, 228, 15 },
		{ 267, 229, 15 },
		{ 267, 230, 15 },
		{ 267, 231, 15 },
		{ 267, 232, 15 },
		{ 267, 233, 15 },
		{ 267, 234, 15 },
		{ 267, 235, 15 },
		{ 267, 236, 15 },
		{ 268, 131, 15 },
		{ 268, 132, 15 },
		{ 268, 133, 15 },
		{ 268, 134, 15 },
		{ 268, 135, 15 },
		{ 268, 136, 15 },
		{ 268, 137, 15 },
		{ 268, 138, 15 },
		{ 268, 139, 15 },
		{ 268, 140, 15 },
		{ 268, 141, 15 },
		{ 268, 142, 15 },
		{ 268, 143, 15 },
		{ 268, 144, 15 },
		{ 268, 145, 15 },
		{ 268, 146, 15 },
		{ 268, 147, 15 },
		{ 268, 148, 15 },
		{ 268, 149, 15 },
		{ 268, 150, 15 },
		{ 268, 151, 15 },
		{ 268, 152, 15 },
		{ 268, 153, 15 },
		{ 268, 154, 15 },
		{ 268, 155, 15 },
		{ 268, 156, 15 },
		{ 268, 157, 15 },
		{ 268, 158, 15 },
		{ 268, 159, 15 },
		{ 268, 160, 15 },
		{ 268, 161, 15 },
		{ 268, 162, 15 },
		{ 268, 163, 15 },
		{ 268, 164, 15 },
		{ 268, 165, 15 },
		{ 268, 166, 15 },
		{ 268, 167, 15 },
		{ 268, 168, 15 },
		{ 268, 169, 15 },
		{ 268, 170, 15 },
		{ 268, 171, 15 },
		{ 268, 172, 15 },
		{ 268, 173, 15 },
		{ 268, 174, 15 },
		{ 268, 175, 15 },
		{ 268, 176, 15 },
		{ 268, 177, 15 },
		{ 268, 178, 15 },
		{ 268, 179, 15 },
		{ 268, 180, 15 },
		{ 268, 181, 15 },
		{ 268, 182, 15 },
		{ 268, 183, 15 },
		{ 268, 184, 15 },
		{ 268, 185, 15 },
		{ 268, 186, 15 },
		{ 268, 187, 15 },
		{ 268, 188, 15 },
		{ 268, 189, 15 },
		{ 268, 190, 15 },
		{ 268, 191, 15 },
		{ 268, 192, 15 },
		{ 268, 193, 15 },
		{ 268, 194, 15 },
		{ 268, 195, 15 },
		{ 268, 196, 15 },
		{ 268, 197, 15 },
		{ 268, 198, 15 },
		{ 268, 199, 15 },
		{ 268, 200, 15 },
		{ 268, 201, 15 },
		{ 268, 202, 15 },
		{ 268, 203, 15 },
		{ 268, 204, 15 },
		{ 268, 205, 15 },
		{ 268, 206, 15 },
		{ 268, 207, 15 },
		{ 268, 208, 15 },
		{ 268, 209, 15 },
		{ 268, 210, 15 },
		{ 268, 211, 15 },
		{ 268, 212, 15 },
		{ 268, 213, 15 },
		{ 268, 214, 15 },
		{ 268, 215, 15 },
		{ 268, 216, 15 },
		{ 268, 217, 15 },
		{ 268, 218, 15 },
		{ 268, 219, 15 },
		{ 268, 220, 15 },
		{ 268, 221, 15 },
		{ 268, 222, 15 },
		{ 268, 223, 15 },
		{ 268, 224, 15 },
		{ 268, 225, 15 },
		{ 268, 226, 15 },
		{ 268, 227, 15 },
		{ 268, 228, 15 },
		{ 268, 229, 15 },
		{ 268, 230, 15 },
		{ 268, 231, 15 },
		{ 268, 232, 15 },
		{ 268, 233, 15 },
		{ 268, 234, 15 },
		{ 269, 131, 15 },
		{ 269, 132, 15 },
		{ 269, 133, 15 },
		{ 269, 134, 15 },
		{ 269, 135, 15 },
		{ 269, 136, 15 },
		{ 269, 137, 15 },
		{ 269, 138, 15 },
		{ 269, 139, 15 },
		{ 269, 140, 15 },
		{ 269, 141, 15 },
		{ 269, 142, 15 },
		{ 269, 143, 15 },
		{ 269, 144, 15 },
		{ 269, 145, 15 },
		{ 269, 146, 15 },
		{ 269, 147, 15 },
		{ 269, 148, 15 },
		{ 269, 149, 15 },
		{ 269, 150, 15 },
		{ 269, 151, 15 },
		{ 269, 152, 15 },
		{ 269, 153, 15 },
		{ 269, 154, 15 },
		{ 269, 155, 15 },
		{ 269, 156, 15 },
		{ 269, 157, 15 },
		{ 269, 158, 15 },
		{ 269, 159, 15 },
		{ 269, 160, 15 },
		{ 269, 161, 15 },
		{ 269, 162, 15 },
		{ 269, 163, 15 },
		{ 269, 164, 15 },
		{ 269, 165, 15 },
		{ 269, 166, 15 },
		{ 269, 167, 15 },
		{ 269, 168, 15 },
		{ 269, 169, 15 },
		{ 269, 170, 15 },
		{ 269, 171, 15 },
		{ 269, 172, 15 },
		{ 269, 173, 15 },
		{ 269, 174, 15 },
		{ 269, 175, 15 },
		{ 269, 176, 15 },
		{ 269, 177, 15 },
		{ 269, 178, 15 },
		{ 269, 179, 15 },
		{ 269, 180, 15 },
		{ 269, 181, 15 },
		{ 269, 182, 15 },
		{ 269, 183, 15 },
		{ 269, 184, 15 },
		{ 269, 185, 15 },
		{ 269, 186, 15 },
		{ 269, 187, 15 },
		{ 269, 188, 15 },
		{ 269, 189, 15 },
		{ 269, 190, 15 },
		{ 269, 191, 15 },
		{ 269, 192, 15 },
		{ 269, 193, 15 },
		{ 269, 194, 15 },
		{ 269, 195, 15 },
		{ 269, 196, 15 },
		{ 269, 197, 15 },
		{ 269, 198, 15 },
		{ 269, 199, 15 },
		{ 269, 200, 15 },
		{ 269, 201, 15 },
		{ 269, 202, 15 },
		{ 269, 203, 15 },
		{ 269, 204, 15 },
		{ 269, 205, 15 },
		{ 269, 206, 15 },
		{ 269, 207, 15 },
		{ 269, 208, 15 },
		{ 269, 209, 15 },
		{ 269, 210, 15 },
		{ 269, 211, 15 },
		{ 269, 212, 15 },
		{ 269, 213, 15 },
		{ 269, 214, 15 },
		{ 269, 215, 15 },
		{ 269, 216, 15 },
		{ 269, 217, 15 },
		{ 269, 218, 15 },
		{ 269, 219, 15 },
		{ 269, 220, 15 },
		{ 269, 221, 15 },
		{ 269, 222, 15 },
		{ 269, 223, 15 },
		{ 269, 224, 15 },
		{ 269, 225, 15 },
		{ 269, 226, 15 },
		{ 269, 227, 15 },
		{ 269, 228, 15 },
		{ 269, 229, 15 },
		{ 269, 230, 15 },
		{ 269, 231, 15 },
		{ 269, 232, 15 },
		{ 270, 131, 15 },
		{ 270, 132, 15 },
		{ 270, 133, 15 },
		{ 270, 134, 15 },
		{ 270, 135, 15 },
		{ 270, 136, 15 },
		{ 270, 137, 15 },
		{ 270, 138, 15 },
		{ 270, 139, 15 },
		{ 270, 140, 15 },
		{ 270, 141, 15 },
		{ 270, 142, 15 },
		{ 270, 143, 15 },
		{ 270, 144, 15 },
		{ 270, 145, 15 },
		{ 270, 146, 15 },
		{ 270, 147, 15 },
		{ 270, 148, 15 },
		{ 270, 149, 15 },
		{ 270, 150, 15 },
		{ 270, 151, 15 },
		{ 270, 152, 15 },
		{ 270, 153, 15 },
		{ 270, 154, 15 },
		{ 270, 155, 15 },
		{ 270, 156, 15 },
		{ 270, 157, 15 },
		{ 270, 158, 15 },
		{ 270, 159, 15 },
		{ 270, 160, 15 },
		{ 270, 161, 15 },
		{ 270, 162, 15 },
		{ 270, 163, 15 },
		{ 270, 164, 15 },
		{ 270, 165, 15 },
		{ 270, 166, 15 },
		{ 270, 167, 15 },
		{ 270, 168, 15 },
		{ 270, 169, 15 },
		{ 270, 170, 15 },
		{ 270, 171, 15 },
		{ 270, 172, 15 },
		{ 270, 173, 15 },
		{ 270, 174, 15 },
		{ 270, 175, 15 },
		{ 270, 176, 15 },
		{ 270, 177, 15 },
		{ 270, 178, 15 },
		{ 270, 179, 15 },
		{ 270, 180, 15 },
		{ 270, 181, 15 },
		{ 270, 182, 15 },
		{ 270, 183, 15 },
		{ 270, 184, 15 },
		{ 270, 185, 15 },
		{ 270, 186, 15 },
		{ 270, 187, 15 },
		{ 270, 188, 15 },
		{ 270, 189, 15 },
		{ 270, 190, 15 },
		{ 270, 191, 15 },
		{ 270, 192, 15 },
		{ 270, 193, 15 },
		{ 270, 194, 15 },
		{ 270, 195, 15 },
		{ 270, 196, 15 },
		{ 270, 197, 15 },
		{ 270, 198, 15 },
		{ 270, 199, 15 },
		{ 270, 200, 15 },
		{ 270, 201, 15 },
		{ 270, 202, 15 },
		{ 270, 203, 15 },
		{ 270, 204, 15 },
		{ 270, 205, 15 },
		{ 270, 206, 15 },
		{ 270, 207, 15 },
		{ 270, 208, 15 },
		{ 270, 209, 15 },
		{ 270, 210, 15 },
		{ 270, 211, 15 },
		{ 270, 212, 15 },
		{ 270, 213, 15 },
		{ 270, 214, 15 },
		{ 270, 215, 15 },
		{ 270, 216, 15 },
		{ 270, 217, 15 },
		{ 270, 218, 15 },
		{ 270, 219, 15 },
		{ 270, 220, 15 },
		{ 270, 221, 15 },
		{ 270, 222, 15 },
		{ 270, 223, 15 },
		{ 270, 224, 15 },
		{ 270, 225, 15 },
		{ 270, 226, 15 },
		{ 270, 227, 15 },
		{ 270, 228, 15 },
		{ 270, 229, 15 },
		{ 270, 230, 15 },
		{ 271, 131, 15 },
		{ 271, 132, 15 },
		{ 271, 133, 15 },
		{ 271, 134, 15 },
		{ 271, 135, 15 },
		{ 271, 136, 15 },
		{ 271, 137, 15 },
		{ 271, 138, 15 },
		{ 271, 139, 15 },
		{ 271, 140, 15 },
		{ 271, 141, 15 },
		{ 271, 142, 15 },
		{ 271, 143, 15 },
		{ 271, 144, 15 },
		{ 271, 145, 15 },
		{ 271, 146, 15 },
		{ 271, 147, 15 },
		{ 271, 148, 15 },
		{ 271, 149, 15 },
		{ 271, 150, 15 },
		{ 271, 151, 15 },
		{ 271, 152, 15 },
		{ 271, 153, 15 },
		{ 271, 154, 15 },
		{ 271, 155, 15 },
		{ 271, 156, 15 },
		{ 271, 157, 15 },
		{ 271, 158, 15 },
		{ 271, 159, 15 },
		{ 271, 160, 15 },
		{ 271, 161, 15 },
		{ 271, 162, 15 },
		{ 271, 163, 15 },
		{ 271, 164, 15 },
		{ 271, 165, 15 },
		{ 271, 166, 15 },
		{ 271, 167, 15 },
		{ 271, 168, 15 },
		{ 271, 169, 15 },
		{ 271, 170, 15 },
		{ 271, 171, 15 },
		{ 271, 172, 15 },
		{ 271, 173, 15 },
		{ 271, 174, 15 },
		{ 271, 175, 15 },
		{ 271, 176, 15 },
		{ 271, 177, 15 },
		{ 271, 178, 15 },
		{ 271, 179, 15 },
		{ 271, 180, 15 },
		{ 271, 181, 15 },
		{ 271, 182, 15 },
		{ 271, 183, 15 },
		{ 271, 184, 15 },
		{ 271, 185, 15 },
		{ 271, 186, 15 },
		{ 271, 187, 15 },
		{ 271, 188, 15 },
		{ 271, 189, 15 },
		{ 271, 190, 15 },
		{ 271, 191, 15 },
		{ 271, 192, 15 },
		{ 271, 193, 15 },
		{ 271, 194, 15 },
		{ 271, 195, 15 },
		{ 271, 196, 15 },
		{ 271, 197, 15 },
		{ 271, 198, 15 },
		{ 271, 199, 15 },
		{ 271, 200, 15 },
		{ 271, 201, 15 },
		{ 271, 202, 15 },
		{ 271, 203, 15 },
		{ 271, 204, 15 },
		{ 271, 205, 15 },
		{ 271, 206, 15 },
		{ 271, 207, 15 },
		{ 271, 208, 15 },
		{ 271, 209, 15 },
		{ 271, 210, 15 },
		{ 271, 211, 15 },
		{ 271, 212, 15 },
		{ 271, 213, 15 },
		{ 271, 214, 15 },
		{ 271, 215, 15 },
		{ 271, 216, 15 },
		{ 271, 217, 15 },
		{ 271, 218, 15 },
		{ 271, 219, 15 },
		{ 271, 220, 15 },
		{ 271, 221, 15 },
		{ 271, 222, 15 },
		{ 271, 223, 15 },
		{ 271, 224, 15 },
		{ 271, 225, 15 },
		{ 271, 226, 15 },
		{ 271, 227, 15 },
		{ 271, 228, 15 },
		{ 272, 131, 15 },
		{ 272, 132, 15 },
		{ 272, 133, 15 },
		{ 272, 134, 15 },
		{ 272, 135, 15 },
		{ 272, 136, 15 },
		{ 272, 137, 15 },
		{ 272, 138, 15 },
		{ 272, 139, 15 },
		{ 272, 140, 15 },
		{ 272, 141, 15 },
		{ 272, 142, 15 },
		{ 272, 143, 15 },
		{ 272, 144, 15 },
		{ 272, 145, 15 },
		{ 272, 146, 15 },
		{ 272, 147, 15 },
		{ 272, 148, 15 },
		{ 272, 149, 15 },
		{ 272, 150, 15 },
		{ 272, 151, 15 },
		{ 272, 152, 15 },
		{ 272, 153, 15 },
		{ 272, 154, 15 },
		{ 272, 155, 15 },
		{ 272, 156, 15 },
		{ 272, 157, 15 },
		{ 272, 158, 15 },
		{ 272, 159, 15 },
		{ 272, 160, 15 },
		{ 272, 161, 15 },
		{ 272, 162, 15 },
		{ 272, 163, 15 },
		{ 272, 164, 15 },
		{ 272, 165, 15 },
		{ 272, 166, 15 },
		{ 272, 167, 15 },
		{ 272, 168, 15 },
		{ 272, 169, 15 },
		{ 272, 170, 15 },
		{ 272, 171, 15 },
		{ 272, 172, 15 },
		{ 272, 173, 15 },
		{ 272, 174, 15 },
		{ 272, 175, 15 },
		{ 272, 176, 15 },
		{ 272, 177, 15 },
		{ 272, 178, 15 },
		{ 272, 179, 15 },
		{ 272, 180, 15 },
		{ 272, 181, 15 },
		{ 272, 182, 15 },
		{ 272, 183, 15 },
		{ 272, 184, 15 },
		{ 272, 185, 15 },
		{ 272, 186, 15 },
		{ 272, 187, 15 },
		{ 272, 188, 15 },
		{ 272, 189, 15 },
		{ 272, 190, 15 },
		{ 272, 191, 15 },
		{ 272, 192, 15 },
		{ 272, 193, 15 },
		{ 272, 194, 15 },
		{ 272, 195, 15 },
		{ 272, 196, 15 },
		{ 272, 197, 15 },
		{ 272, 198, 15 },
		{ 272, 199, 15 },
		{ 272, 200, 15 },
		{ 272, 201, 15 },
		{ 272, 202, 15 },
		{ 272, 203, 15 },
		{ 272, 204, 15 },
		{ 272, 205, 15 },
		{ 272, 206, 15 },
		{ 272, 207, 15 },
		{ 272, 208, 15 },
		{ 272, 209, 15 },
		{ 272, 210, 15 },
		{ 272, 211, 15 },
		{ 272, 212, 15 },
		{ 272, 213, 15 },
		{ 272, 214, 15 },
		{ 272, 215, 15 },
		{ 272, 216, 15 },
		{ 272, 217, 15 },
		{ 272, 218, 15 },
		{ 272, 219, 15 },
		{ 272, 220, 15 },
		{ 272, 221, 15 },
		{ 272, 222, 15 },
		{ 272, 223, 15 },
		{ 272, 224, 15 },
		{ 272, 225, 15 },
		{ 272, 226, 15 },
		{ 273, 131, 15 },
		{ 273, 132, 15 },
		{ 273, 133, 15 },
		{ 273, 134, 15 },
		{ 273, 135, 15 },
		{ 273, 136, 15 },
		{ 273, 137, 15 },
		{ 273, 138, 15 },
		{ 273, 139, 15 },
		{ 273, 140, 15 },
		{ 273, 141, 15 },
		{ 273, 142, 15 },
		{ 273, 143, 15 },
		{ 273, 144, 15 },
		{ 273, 145, 15 },
		{ 273, 146, 15 },
		{ 273, 147, 15 },
		{ 273, 148, 15 },
		{ 273, 149, 15 },
		{ 273, 150, 15 },
		{ 273, 151, 15 },
		{ 273, 152, 15 },
		{ 273, 153, 15 },
		{ 273, 154, 15 },
		{ 273, 155, 15 },
		{ 273, 156, 15 },
		{ 273, 157, 15 },
		{ 273, 158, 15 },
		{ 273, 159, 15 },
		{ 273, 160, 15 },
		{ 273, 161, 15 },
		{ 273, 162, 15 },
		{ 273, 163, 15 },
		{ 273, 164, 15 },
		{ 273, 165, 15 },
		{ 273, 166, 15 },
		{ 273, 167, 15 },
		{ 273, 168, 15 },
		{ 273, 169, 15 },
		{ 273, 170, 15 },
		{ 273, 171, 15 },
		{ 273, 172, 15 },
		{ 273, 173, 15 },
		{ 273, 174, 15 },
		{ 273, 175, 15 },
		{ 273, 176, 15 },
		{ 273, 177, 15 },
		{ 273, 178, 15 },
		{ 273, 179, 15 },
		{ 273, 180, 15 },
		{ 273, 181, 15 },
		{ 273, 182, 15 },
		{ 273, 183, 15 },
		{ 273, 184, 15 },
		{ 273, 185, 15 },
		{ 273, 186, 15 },
		{ 273, 187, 15 },
		{ 273, 188, 15 },
		{ 273, 189, 15 },
		{ 273, 190, 15 },
		{ 273, 191, 15 },
		{ 273, 192, 15 },
		{ 273, 193, 15 },
		{ 273, 194, 15 },
		{ 273, 195, 15 },
		{ 273, 196, 15 },
		{ 273, 197, 15 },
		{ 273, 198, 15 },
		{ 273, 199, 15 },
		{ 273, 200, 15 },
		{ 273, 201, 15 },
		{ 273, 202, 15 },
		{ 273, 203, 15 },
		{ 273, 204, 15 },
		{ 273, 205, 15 },
		{ 273, 206, 15 },
		{ 273, 207, 15 },
		{ 273, 208, 15 },
		{ 273, 209, 15 },
		{ 273, 210, 15 },
		{ 273, 211, 15 },
		{ 273, 212, 15 },
		{ 273, 213, 15 },
		{ 273, 214, 15 },
		{ 273, 215, 15 },
		{ 273, 216, 15 },
		{ 273, 217, 15 },
		{ 273, 218, 15 },
		{ 273, 219, 15 },
		{ 273, 220, 15 },
		{ 273, 221, 15 },
		{ 273, 222, 15 },
		{ 273, 223, 15 },
		{ 273, 224, 15 },
		{ 274, 131, 15 },
		{ 274, 132, 15 },
		{ 274, 133, 15 },
		{ 274, 134, 15 },
		{ 274, 135, 15 },
		{ 274, 136, 15 },
		{ 274, 137, 15 },
		{ 274, 138, 15 },
		{ 274, 139, 15 },
		{ 274, 140, 15 },
		{ 274, 141, 15 },
		{ 274, 142, 15 },
		{ 274, 143, 15 },
		{ 274, 144, 15 },
		{ 274, 145, 15 },
		{ 274, 146, 15 },
		{ 274, 147, 15 },
		{ 274, 148, 15 },
		{ 274, 149, 15 },
		{ 274, 150, 15 },
		{ 274, 151, 15 },
		{ 274, 152, 15 },
		{ 274, 153, 15 },
		{ 274, 154, 15 },
		{ 274, 155, 15 },
		{ 274, 156, 15 },
		{ 274, 157, 15 },
		{ 274, 158, 15 },
		{ 274, 159, 15 },
		{ 274, 160, 15 },
		{ 274, 161, 15 },
		{ 274, 162, 15 },
		{ 274, 163, 15 },
		{ 274, 164, 15 },
		{ 274, 165, 15 },
		{ 274, 166, 15 },
		{ 274, 167, 15 },
		{ 274, 168, 15 },
		{ 274, 169, 15 },
		{ 274, 170, 15 },
		{ 274, 171, 15 },
		{ 274, 172, 15 },
		{ 274, 173, 15 },
		{ 274, 174, 15 },
		{ 274, 175, 15 },
		{ 274, 176, 15 },
		{ 274, 177, 15 },
		{ 274, 178, 15 },
		{ 274, 179, 15 },
		{ 274, 180, 15 },
		{ 274, 181, 15 },
		{ 274, 182, 15 },
		{ 274, 183, 15 },
		{ 274, 184, 15 },
		{ 274, 185, 15 },
		{ 274, 186, 15 },
		{ 274, 187, 15 },
		{ 274, 188, 15 },
		{ 274, 189, 15 },
		{ 274, 190, 15 },
		{ 274, 191, 15 },
		{ 274, 192, 15 },
		{ 274, 193, 15 },
		{ 274, 194, 15 },
		{ 274, 195, 15 },
		{ 274, 196, 15 },
		{ 274, 197, 15 },
		{ 274, 198, 15 },
		{ 274, 199, 15 },
		{ 274, 200, 15 },
		{ 274, 201, 15 },
		{ 274, 202, 15 },
		{ 274, 203, 15 },
		{ 274, 204, 15 },
		{ 274, 205, 15 },
		{ 274, 206, 15 },
		{ 274, 207, 15 },
		{ 274, 208, 15 },
		{ 274, 209, 15 },
		{ 274, 210, 15 },
		{ 274, 211, 15 },
		{ 274, 212, 15 },
		{ 274, 213, 15 },
		{ 274, 214, 15 },
		{ 274, 215, 15 },
		{ 274, 216, 15 },
		{ 274, 217, 15 },
		{ 274, 218, 15 },
		{ 274, 219, 15 },
		{ 274, 220, 15 },
		{ 274, 221, 15 },
		{ 274, 222, 15 },
		{ 275, 131, 15 },
		{ 275, 132, 15 },
		{ 275, 133, 15 },
		{ 275, 134, 15 },
		{ 275, 135, 15 },
		{ 275, 136, 15 },
		{ 275, 137, 15 },
		{ 275, 138, 15 },
		{ 275, 139, 15 },
		{ 275, 140, 15 },
		{ 275, 141, 15 },
		{ 275, 142, 15 },
		{ 275, 143, 15 },
		{ 275, 144, 15 },
		{ 275, 145, 15 },
		{ 275, 146, 15 },
		{ 275, 147, 15 },
		{ 275, 148, 15 },
		{ 275, 149, 15 },
		{ 275, 150, 15 },
		{ 275, 151, 15 },
		{ 275, 152, 15 },
		{ 275, 153, 15 },
		{ 275, 154, 15 },
		{ 275, 155, 15 },
		{ 275, 156, 15 },
		{ 275, 157, 15 },
		{ 275, 158, 15 },
		{ 275, 159, 15 },
		{ 275, 160, 15 },
		{ 275, 161, 15 },
		{ 275, 162, 15 },
		{ 275, 163, 15 },
		{ 275, 164, 15 },
		{ 275, 165, 15 },
		{ 275, 166, 15 },
		{ 275, 167, 15 },
		{ 275, 168, 15 },
		{ 275, 169, 15 },
		{ 275, 170, 15 },
		{ 275, 171, 15 },
		{ 275, 172, 15 },
		{ 275, 173, 15 },
		{ 275, 174, 15 },
		{ 275, 175, 15 },
		{ 275, 176, 15 },
		{ 275, 177, 15 },
		{ 275, 178, 15 },
		{ 275, 179, 15 },
		{ 275, 180, 15 },
		{ 275, 181, 15 },
		{ 275, 182, 15 },
		{ 275, 183, 15 },
		{ 275, 184, 15 },
		{ 275, 185, 15 },
		{ 275, 186, 15 },
		{ 275, 187, 15 },
		{ 275, 188, 15 },
		{ 275, 189, 15 },
		{ 275, 190, 15 },
		{ 275, 191, 15 },
		{ 275, 192, 15 },
		{ 275, 193, 15 },
		{ 275, 194, 15 },
		{ 275, 195, 15 },
		{ 275, 196, 15 },
		{ 275, 197, 15 },
		{ 275, 198, 15 },
		{ 275, 199, 15 },
		{ 275, 200, 15 },
		{ 275, 201, 15 },
		{ 275, 202, 15 },
		{ 275, 203, 15 },
		{ 275, 204, 15 },
		{ 275, 205, 15 },
		{ 275, 206, 15 },
		{ 275, 207, 15 },
		{ 275, 208, 15 },
		{ 275, 209, 15 },
		{ 275, 210, 15 },
		{ 275, 211, 15 },
		{ 275, 212, 15 },
		{ 275, 213, 15 },
		{ 275, 214, 15 },
		{ 275, 215, 15 },
		{ 275, 216, 15 },
		{ 275, 217, 15 },
		{ 275, 218, 15 },
		{ 275, 219, 15 },
		{ 275, 220, 15 },
		{ 276, 131, 15 },
		{ 276, 132, 15 },
		{ 276, 133, 15 },
		{ 276, 134, 15 },
		{ 276, 135, 15 },
		{ 276, 136, 15 },
		{ 276, 137, 15 },
		{ 276, 138, 15 },
		{ 276, 139, 15 },
		{ 276, 140, 15 },
		{ 276, 141, 15 },
		{ 276, 142, 15 },
		{ 276, 143, 15 },
		{ 276, 144, 15 },
		{ 276, 145, 15 },
		{ 276, 146, 15 },
		{ 276, 147, 15 },
		{ 276, 148, 15 },
		{ 276, 149, 15 },
		{ 276, 150, 15 },
		{ 276, 151, 15 },
		{ 276, 152, 15 },
		{ 276, 153, 15 },
		{ 276, 154, 15 },
		{ 276, 155, 15 },
		{ 276, 156, 15 },
		{ 276, 157, 15 },
		{ 276, 158, 15 },
		{ 276, 159, 15 },
		{ 276, 160, 15 },
		{ 276, 161, 15 },
		{ 276, 162, 15 },
		{ 276, 163, 15 },
		{ 276, 164, 15 },
		{ 276, 165, 15 },
		{ 276, 166, 15 },
		{ 276, 167, 15 },
		{ 276, 168, 15 },
		{ 276, 169, 15 },
		{ 276, 170, 15 },
		{ 276, 171, 15 },
		{ 276, 172, 15 },
		{ 276, 173, 15 },
		{ 276, 174, 15 },
		{ 276, 175, 15 },
		{ 276, 176, 15 },
		{ 276, 177, 15 },
		{ 276, 178, 15 },
		{ 276, 179, 15 },
		{ 276, 180, 15 },
		{ 276, 181, 15 },
		{ 276, 182, 15 },
		{ 276, 183, 15 },
		{ 276, 184, 15 },
		{ 276, 185, 15 },
		{ 276, 186, 15 },
		{ 276, 187, 15 },
		{ 276, 188, 15 },
		{ 276, 189, 15 },
		{ 276, 190, 15 },
		{ 276, 191, 15 },
		{ 276, 192, 15 },
		{ 276, 193, 15 },
		{ 276, 194, 15 },
		{ 276, 195, 15 },
		{ 276, 196, 15 },
		{ 276, 197, 15 },
		{ 276, 198, 15 },
		{ 276, 199, 15 },
		{ 276, 200, 15 },
		{ 276, 201, 15 },
		{ 276, 202, 15 },
		{ 276, 203, 15 },
		{ 276, 204, 15 },
		{ 276, 205, 15 },
		{ 276, 206, 15 },
		{ 276, 207, 15 },
		{ 276, 208, 15 },
		{ 276, 209, 15 },
		{ 276, 210, 15 },
		{ 276, 211, 15 },
		{ 276, 212, 15 },
		{ 276, 213, 15 },
		{ 276, 214, 15 },
		{ 276, 215, 15 },
		{ 276, 216, 15 },
		{ 276, 217, 15 },
		{ 276, 218, 15 },
		{ 277, 131, 15 },
		{ 277, 132, 15 },
		{ 277, 133, 15 },
		{ 277, 134, 15 },
		{ 277, 135, 15 },
		{ 277, 136, 15 },
		{ 277, 137, 15 },
		{ 277, 138, 15 },
		{ 277, 139, 15 },
		{ 277, 140, 15 },
		{ 277, 141, 15 },
		{ 277, 142, 15 },
		{ 277, 143, 15 },
		{ 277, 144, 15 },
		{ 277, 145, 15 },
		{ 277, 146, 15 },
		{ 277, 147, 15 },
		{ 277, 148, 15 },
		{ 277, 149, 15 },
		{ 277, 150, 15 },
		{ 277, 151, 15 },
		{ 277, 152, 15 },
		{ 277, 153, 15 },
		{ 277, 154, 15 },
		{ 277, 155, 15 },
		{ 277, 156, 15 },
		{ 277, 157, 15 },
		{ 277, 158, 15 },
		{ 277, 159, 15 },
		{ 277, 160, 15 },
		{ 277, 161, 15 },
		{ 277, 162, 15 },
		{ 277, 163, 15 },
		{ 277, 164, 15 },
		{ 277, 165, 15 },
		{ 277, 166, 15 },
		{ 277, 167, 15 },
		{ 277, 168, 15 },
		{ 277, 169, 15 },
		{ 277, 170, 15 },
		{ 277, 171, 15 },
		{ 277, 172, 15 },
		{ 277, 173, 15 },
		{ 277, 174, 15 },
		{ 277, 175, 15 },
		{ 277, 176, 15 },
		{ 277, 177, 15 },
		{ 277, 178, 15 },
		{ 277, 179, 15 },
		{ 277, 180, 15 },
		{ 277, 181, 15 },
		{ 277, 182, 15 },
		{ 277, 183, 15 },
		{ 277, 184, 15 },
		{ 277, 185, 15 },
		{ 277, 186, 15 },
		{ 277, 187, 15 },
		{ 277, 188, 15 },
		{ 277, 189, 15 },
		{ 277, 190, 15 },
		{ 277, 191, 15 },
		{ 277, 192, 15 },
		{ 277, 193, 15 },
		{ 277, 194, 15 },
		{ 277, 195, 15 },
		{ 277, 196, 15 },
		{ 277, 197, 15 },
		{ 277, 198, 15 },
		{ 277, 199, 15 },
		{ 277, 200, 15 },
		{ 277, 201, 15 },
		{ 277, 202, 15 },
		{ 277, 203, 15 },
		{ 277, 204, 15 },
		{ 277, 205, 15 },
		{ 277, 206, 15 },
		{ 277, 207, 15 },
		{ 277, 208, 15 },
		{ 277, 209, 15 },
		{ 277, 210, 15 },
		{ 277, 211, 15 },
		{ 277, 212, 15 },
		{ 277, 213, 15 },
		{ 277, 214, 15 },
		{ 277, 215, 15 },
		{ 277, 216, 15 },
		{ 278, 131, 15 },
		{ 278, 132, 15 },
		{ 278, 133, 15 },
		{ 278, 134, 15 },
		{ 278, 135, 15 },
		{ 278, 136, 15 },
		{ 278, 137, 15 },
		{ 278, 138, 15 },
		{ 278, 139, 15 },
		{ 278, 140, 15 },
		{ 278, 141, 15 },
		{ 278, 142, 15 },
		{ 278, 143, 15 },
		{ 278, 144, 15 },
		{ 278, 145, 15 },
		{ 278, 146, 15 },
		{ 278, 147, 15 },
		{ 278, 148, 15 },
		{ 278, 149, 15 },
		{ 278, 150, 15 },
		{ 278, 151, 15 },
		{ 278, 152, 15 },
		{ 278, 153, 15 },
		{ 278, 154, 15 },
		{ 278, 155, 15 },
		{ 278, 156, 15 },
		{ 278, 157, 15 },
		{ 278, 158, 15 },
		{ 278, 159, 15 },
		{ 278, 160, 15 },
		{ 278, 161, 15 },
		{ 278, 162, 15 },
		{ 278, 163, 15 },
		{ 278, 164, 15 },
		{ 278, 165, 15 },
		{ 278, 166, 15 },
		{ 278, 167, 15 },
		{ 278, 168, 15 },
		{ 278, 169, 15 },
		{ 278, 170, 15 },
		{ 278, 171, 15 },
		{ 278, 172, 15 },
		{ 278, 173, 15 },
		{ 278, 174, 15 },
		{ 278, 175, 15 },
		{ 278, 176, 15 },
		{ 278, 177, 15 },
		{ 278, 178, 15 },
		{ 278, 179, 15 },
		{ 278, 180, 15 },
		{ 278, 181, 15 },
		{ 278, 182, 15 },
		{ 278, 183, 15 },
		{ 278, 184, 15 },
		{ 278, 185, 15 },
		{ 278, 186, 15 },
		{ 278, 187, 15 },
		{ 278, 188, 15 },
		{ 278, 189, 15 },
		{ 278, 190, 15 },
		{ 278, 191, 15 },
		{ 278, 192, 15 },
		{ 278, 193, 15 },
		{ 278, 194, 15 },
		{ 278, 195, 15 },
		{ 278, 196, 15 },
		{ 278, 197, 15 },
		{ 278, 198, 15 },
		{ 278, 199, 15 },
		{ 278, 200, 15 },
		{ 278, 201, 15 },
		{ 278, 202, 15 },
		{ 278, 203, 15 },
		{ 278, 204, 15 },
		{ 278, 205, 15 },
		{ 278, 206, 15 },
		{ 278, 207, 15 },
		{ 278, 208, 15 },
		{ 278, 209, 15 },
		{ 278, 210, 15 },
		{ 278, 211, 15 },
		{ 278, 212, 15 },
		{ 278, 213, 15 },
		{ 278, 214, 15 },
		{ 279, 131, 15 },
		{ 279, 132, 15 },
		{ 279, 133, 15 },
		{ 279, 134, 15 },
		{ 279, 135, 15 },
		{ 279, 136, 15 },
		{ 279, 137, 15 },
		{ 279, 138, 15 },
		{ 279, 139, 15 },
		{ 279, 140, 15 },
		{ 279, 141, 15 },
		{ 279, 142, 15 },
		{ 279, 143, 15 },
		{ 279, 144, 15 },
		{ 279, 145, 15 },
		{ 279, 146, 15 },
		{ 279, 147, 15 },
		{ 279, 148, 15 },
		{ 279, 149, 15 },
		{ 279, 150, 15 },
		{ 279, 151, 15 },
		{ 279, 152, 15 },
		{ 279, 153, 15 },
		{ 279, 154, 15 },
		{ 279, 155, 15 },
		{ 279, 156, 15 },
		{ 279, 157, 15 },
		{ 279, 158, 15 },
		{ 279, 159, 15 },
		{ 279, 160, 15 },
		{ 279, 161, 15 },
		{ 279, 162, 15 },
		{ 279, 163, 15 },
		{ 279, 164, 15 },
		{ 279, 165, 15 },
		{ 279, 166, 15 },
		{ 279, 167, 15 },
		{ 279, 168, 15 },
		{ 279, 169, 15 },
		{ 279, 170, 15 },
		{ 279, 171, 15 },
		{ 279, 172, 15 },
		{ 279, 173, 15 },
		{ 279, 174, 15 },
		{ 279, 175, 15 },
		{ 279, 176, 15 },
		{ 279, 177, 15 },
		{ 279, 178, 15 },
		{ 279, 179, 15 },
		{ 279, 180, 15 },
		{ 279, 181, 15 },
		{ 279, 182, 15 },
		{ 279, 183, 15 },
		{ 279, 184, 15 },
		{ 279, 185, 15 },
		{ 279, 186, 15 },
		{ 279, 187, 15 },
		{ 279, 188, 15 },
		{ 279, 189, 15 },
		{ 279, 190, 15 },
		{ 279, 191, 15 },
		{ 279, 192, 15 },
		{ 279, 193, 15 },
		{ 279, 194, 15 },
		{ 279, 195, 15 },
		{ 279, 196, 15 },
		{ 279, 197, 15 },
		{ 279, 198, 15 },
		{ 279, 199, 15 },
		{ 279, 200, 15 },
		{ 279, 201, 15 },
		{ 279, 202, 15 },
		{ 279, 203, 15 },
		{ 279, 204, 15 },
		{ 279, 205, 15 },
		{ 279, 206, 15 },
		{ 279, 207, 15 },
		{ 279, 208, 15 },
		{ 279, 209, 15 },
		{ 279, 210, 15 },
		{ 279, 211, 15 },
		{ 279, 212, 15 },
		{ 280, 131, 15 },
		{ 280, 132, 15 },
		{ 280, 133, 15 },
		{ 280, 134, 15 },
		{ 280, 135, 15 },
		{ 280, 136, 15 },
		{ 280, 137, 15 },
		{ 280, 138, 15 },
		{ 280, 139, 15 },
		{ 280, 140, 15 },
		{ 280, 141, 15 },
		{ 280, 142, 15 },
		{ 280, 143, 15 },
		{ 280, 144, 15 },
		{ 280, 145, 15 },
		{ 280, 146, 15 },
		{ 280, 147, 15 },
		{ 280, 148, 15 },
		{ 280, 149, 15 },
		{ 280, 150, 15 },
		{ 280, 151, 15 },
		{ 280, 152, 15 },
		{ 280, 153, 15 },
		{ 280, 154, 15 },
		{ 280, 155, 15 },
		{ 280, 156, 15 },
		{ 280, 157, 15 },
		{ 280, 158, 15 },
		{ 280, 159, 15 },
		{ 280, 160, 15 },
		{ 280, 161, 15 },
		{ 280, 162, 15 },
		{ 280, 163, 15 },
		{ 280, 164, 15 },
		{ 280, 165, 15 },
		{ 280, 166, 15 },
		{ 280, 167, 15 },
		{ 280, 168, 15 },
		{ 280, 169, 15 },
		{ 280, 170, 15 },
		{ 280, 171, 15 },
		{ 280, 172, 15 },
		{ 280, 173, 15 },
		{ 280, 174, 15 },
		{ 280, 175, 15 },
		{ 280, 176, 15 },
		{ 280, 177, 15 },
		{ 280, 178, 15 },
		{ 280, 179, 15 },
		{ 280, 180, 15 },
		{ 280, 181, 15 },
		{ 280, 182, 15 },
		{ 280, 183, 15 },
		{ 280, 184, 15 },
		{ 280, 185, 15 },
		{ 280, 186, 15 },
		{ 280, 187, 15 },
		{ 280, 188, 15 },
		{ 280, 189, 15 },
		{ 280, 190, 15 },
		{ 280, 191, 15 },
		{ 280, 192, 15 },
		{ 280, 193, 15 },
		{ 280, 194, 15 },
		{ 280, 195, 15 },
		{ 280, 196, 15 },
		{ 280, 197, 15 },
		{ 280, 198, 15 },
		{ 280, 199, 15 },
		{ 280, 200, 15 },
		{ 280, 201, 15 },
		{ 280, 202, 15 },
		{ 280, 203, 15 },
		{ 280, 204, 15 },
		{ 280, 205, 15 },
		{ 280, 206, 15 },
		{ 280, 207, 15 },
		{ 280, 208, 15 },
		{ 280, 209, 15 },
		{ 280, 210, 15 },
		{ 281, 131, 15 },
		{ 281, 132, 15 },
		{ 281, 133, 15 },
		{ 281, 134, 15 },
		{ 281, 135, 15 },
		{ 281, 136, 15 },
		{ 281, 137, 15 },
		{ 281, 138, 15 },
		{ 281, 139, 15 },
		{ 281, 140, 15 },
		{ 281, 141, 15 },
		{ 281, 142, 15 },
		{ 281, 143, 15 },
		{ 281, 144, 15 },
		{ 281, 145, 15 },
		{ 281, 146, 15 },
		{ 281, 147, 15 },
		{ 281, 148, 15 },
		{ 281, 149, 15 },
		{ 281, 150, 15 },
		{ 281, 151, 15 },
		{ 281, 152, 15 },
		{ 281, 153, 15 },
		{ 281, 154, 15 },
		{ 281, 155, 15 },
		{ 281, 156, 15 },
		{ 281, 157, 15 },
		{ 281, 158, 15 },
		{ 281, 159, 15 },
		{ 281, 160, 15 },
		{ 281, 161, 15 },
		{ 281, 162, 15 },
		{ 281, 163, 15 },
		{ 281, 164, 15 },
		{ 281, 165, 15 },
		{ 281, 166, 15 },
		{ 281, 167, 15 },
		{ 281, 168, 15 },
		{ 281, 169, 15 },
		{ 281, 170, 15 },
		{ 281, 171, 15 },
		{ 281, 172, 15 },
		{ 281, 173, 15 },
		{ 281, 174, 15 },
		{ 281, 175, 15 },
		{ 281, 176, 15 },
		{ 281, 177, 15 },
		{ 281, 178, 15 },
		{ 281, 179, 15 },
		{ 281, 180, 15 },
		{ 281, 181, 15 },
		{ 281, 182, 15 },
		{ 281, 183, 15 },
		{ 281, 184, 15 },
		{ 281, 185, 15 },
		{ 281, 186, 15 },
		{ 281, 187, 15 },
		{ 281, 188, 15 },
		{ 281, 189, 15 },
		{ 281, 190, 15 },
		{ 281, 191, 15 },
		{ 281, 192, 15 },
		{ 281, 193, 15 },
		{ 281, 194, 15 },
		{ 281, 195, 15 },
		{ 281, 196, 15 },
		{ 281, 197, 15 },
		{ 281, 198, 15 },
		{ 281, 199, 15 },
		{ 281, 200, 15 },
		{ 281, 201, 15 },
		{ 281, 202, 15 },
		{ 281, 203, 15 },
		{ 281, 204, 15 },
		{ 281, 205, 15 },
		{ 281, 206, 15 },
		{ 281, 207, 15 },
		{ 281, 208, 15 },
		{ 282, 131, 15 },
		{ 282, 132, 15 },
		{ 282, 133, 15 },
		{ 282, 134, 15 },
		{ 282, 135, 15 },
		{ 282, 136, 15 },
		{ 282, 137, 15 },
		{ 282, 138, 15 },
		{ 282, 139, 15 },
		{ 282, 140, 15 },
		{ 282, 141, 15 },
		{ 282, 142, 15 },
		{ 282, 143, 15 },
		{ 282, 144, 15 },
		{ 282, 145, 15 },
		{ 282, 146, 15 },
		{ 282, 147, 15 },
		{ 282, 148, 15 },
		{ 282, 149, 15 },
		{ 282, 150, 15 },
		{ 282, 151, 15 },
		{ 282, 152, 15 },
		{ 282, 153, 15 },
		{ 282, 154, 15 },
		{ 282, 155, 15 },
		{ 282, 156, 15 },
		{ 282, 157, 15 },
		{ 282, 158, 15 },
		{ 282, 159, 15 },
		{ 282, 160, 15 },
		{ 282, 161, 15 },
		{ 282, 162, 15 },
		{ 282, 163, 15 },
		{ 282, 164, 15 },
		{ 282, 165, 15 },
		{ 282, 166, 15 },
		{ 282, 167, 15 },
		{ 282, 168, 15 },
		{ 282, 169, 15 },
		{ 282, 170, 15 },
		{ 282, 171, 15 },
		{ 282, 172, 15 },
		{ 282, 173, 15 },
		{ 282, 174, 15 },
		{ 282, 175, 15 },
		{ 282, 176, 15 },
		{ 282, 177, 15 },
		{ 282, 178, 15 },
		{ 282, 179, 15 },
		{ 282, 180, 15 },
		{ 282, 181, 15 },
		{ 282, 182, 15 },
		{ 282, 183, 15 },
		{ 282, 184, 15 },
		{ 282, 185, 15 },
		{ 282, 186, 15 },
		{ 282, 187, 15 },
		{ 282, 188, 15 },
		{ 282, 189, 15 },
		{ 282, 190, 15 },
		{ 282, 191, 15 },
		{ 282, 192, 15 },
		{ 282, 193, 15 },
		{ 282, 194, 15 },
		{ 282, 195, 15 },
		{ 282, 196, 15 },
		{ 282, 197, 15 },
		{ 282, 198, 15 },
		{ 282, 199, 15 },
		{ 282, 200, 15 },
		{ 282, 201, 15 },
		{ 282, 202, 15 },
		{ 282, 203, 15 },
		{ 282, 204, 15 },
		{ 282, 205, 15 },
		{ 282, 206, 15 },
		{ 283, 131, 15 },
		{ 283, 132, 15 },
		{ 283, 133, 15 },
		{ 283, 134, 15 },
		{ 283, 135, 15 },
		{ 283, 136, 15 },
		{ 283, 137, 15 },
		{ 283, 138, 15 },
		{ 283, 139, 15 },
		{ 283, 140, 15 },
		{ 283, 141, 15 },
		{ 283, 142, 15 },
		{ 283, 143, 15 },
		{ 283, 144, 15 },
		{ 283, 145, 15 },
		{ 283, 146, 15 },
		{ 283, 147, 15 },
		{ 283, 148, 15 },
		{ 283, 149, 15 },
		{ 283, 150, 15 },
		{ 283, 151, 15 },
		{ 283, 152, 15 },
		{ 283, 153, 15 },
		{ 283, 154, 15 },
		{ 283, 155, 15 },
		{ 283, 156, 15 },
		{ 283, 157, 15 },
		{ 283, 158, 15 },
		{ 283, 159, 15 },
		{ 283, 160, 15 },
		{ 283, 161, 15 },
		{ 283, 162, 15 },
		{ 283, 163, 15 },
		{ 283, 164, 15 },
		{ 283, 165, 15 },
		{ 283, 166, 15 },
		{ 283, 167, 15 },
		{ 283, 168, 15 },
		{ 283, 169, 15 },
		{ 283, 170, 15 },
		{ 283, 171, 15 },
		{ 283, 172, 15 },
		{ 283, 173, 15 },
		{ 283, 174, 15 },
		{ 283, 175, 15 },
		{ 283, 176, 15 },
		{ 283, 177, 15 },
		{ 283, 178, 15 },
		{ 283, 179, 15 },
		{ 283, 180, 15 },
		{ 283, 181, 15 },
		{ 283, 182, 15 },
		{ 283, 183, 15 },
		{ 283, 184, 15 },
		{ 283, 185, 15 },
		{ 283, 186, 15 },
		{ 283, 187, 15 },
		{ 283, 188, 15 },
		{ 283, 189, 15 },
		{ 283, 190, 15 },
		{ 283, 191, 15 },
		{ 283, 192, 15 },
		{ 283, 193, 15 },
		{ 283, 194, 15 },
		{ 283, 195, 15 },
		{ 283, 196, 15 },
		{ 283, 197, 15 },
		{ 283, 198, 15 },
		{ 283, 199, 15 },
		{ 283, 200, 15 },
		{ 283, 201, 15 },
		{ 283, 202, 15 },
		{ 283, 203, 15 },
		{ 283, 204, 15 },
		{ 284, 131, 15 },
		{ 284, 132, 15 },
		{ 284, 133, 15 },
		{ 284, 134, 15 },
		{ 284, 135, 15 },
		{ 284, 136, 15 },
		{ 284, 137, 15 },
		{ 284, 138, 15 },
		{ 284, 139, 15 },
		{ 284, 140, 15 },
		{ 284, 141, 15 },
		{ 284, 142, 15 },
		{ 284, 143, 15 },
		{ 284, 144, 15 },
		{ 284, 145, 15 },
		{ 284, 146, 15 },
		{ 284, 147, 15 },
		{ 284, 148, 15 },
		{ 284, 149, 15 },
		{ 284, 150, 15 },
		{ 284, 151, 15 },
		{ 284, 152, 15 },
		{ 284, 153, 15 },
		{ 284, 154, 15 },
		{ 284, 155, 15 },
		{ 284, 156, 15 },
		{ 284, 157, 15 },
		{ 284, 158, 15 },
		{ 284, 159, 15 },
		{ 284, 160, 15 },
		{ 284, 161, 15 },
		{ 284, 162, 15 },
		{ 284, 163, 15 },
		{ 284, 164, 15 },
		{ 284, 165, 15 },
		{ 284, 166, 15 },
		{ 284, 167, 15 },
		{ 284, 168, 15 },
		{ 284, 169, 15 },
		{ 284, 170, 15 },
		{ 284, 171, 15 },
		{ 284, 172, 15 },
		{ 284, 173, 15 },
		{ 284, 174, 15 },
		{ 284, 175, 15 },
		{ 284, 176, 15 },
		{ 284, 177, 15 },
		{ 284, 178, 15 },
		{ 284, 179, 15 },
		{ 284, 180, 15 },
		{ 284, 181, 15 },
		{ 284, 182, 15 },
		{ 284, 183, 15 },
		{ 284, 184, 15 },
		{ 284, 185, 15 },
		{ 284, 186, 15 },
		{ 284, 187, 15 },
		{ 284, 188, 15 },
		{ 284, 189, 15 },
		{ 284, 190, 15 },
		{ 284, 191, 15 },
		{ 284, 192, 15 },
		{ 284, 193, 15 },
		{ 284, 194, 15 },
		{ 284, 195, 15 },
		{ 284, 196, 15 },
		{ 284, 197, 15 },
		{ 284, 198, 15 },
		{ 284, 199, 15 },
		{ 284, 200, 15 },
		{ 284, 201, 15 },
		{ 284, 202, 15 },
		{ 285, 131, 15 },
		{ 285, 132, 15 },
		{ 285, 133, 15 },
		{ 285, 134, 15 },
		{ 285, 135, 15 },
		{ 285, 136, 15 },
		{ 285, 137, 15 },
		{ 285, 138, 15 },
		{ 285, 139, 15 },
		{ 285, 140, 15 },
		{ 285, 141, 15 },
		{ 285, 142, 15 },
		{ 285, 143, 15 },
		{ 285, 144, 15 },
		{ 285, 145, 15 },
		{ 285, 146, 15 },
		{ 285, 147, 15 },
		{ 285, 148, 15 },
		{ 285, 149, 15 },
		{ 285, 150, 15 },
		{ 285, 151, 15 },
		{ 285, 152, 15 },
		{ 285, 153, 15 },
		{ 285, 154, 15 },
		{ 285, 155, 15 },
		{ 285, 156, 15 },
		{ 285, 157, 15 },
		{ 285, 158, 15 },
		{ 285, 159, 15 },
		{ 285, 160, 15 },
		{ 285, 161, 15 },
		{ 285, 162, 15 },
		{ 285, 163, 15 },
		{ 285, 164, 15 },
		{ 285, 165, 15 },
		{ 285, 166, 15 },
		{ 285, 167, 15 },
		{ 285, 168, 15 },
		{ 285, 169, 15 },
		{ 285, 170, 15 },
		{ 285, 171, 15 },
		{ 285, 172, 15 },
		{ 285, 173, 15 },
		{ 285, 174, 15 },
		{ 285, 175, 15 },
		{ 285, 176, 15 },
		{ 285, 177, 15 },
		{ 285, 178, 15 },
		{ 285, 179, 15 },
		{ 285, 180, 15 },
		{ 285, 181, 15 },
		{ 285, 182, 15 },
		{ 285, 183, 15 },
		{ 285, 184, 15 },
		{ 285, 185, 15 },
		{ 285, 186, 15 },
		{ 285, 187, 15 },
		{ 285, 188, 15 },
		{ 285, 189, 15 },
		{ 285, 190, 15 },
		{ 285, 191, 15 },
		{ 285, 192, 15 },
		{ 285, 193, 15 },
		{ 285, 194, 15 },
		{ 285, 195, 15 },
		{ 285, 196, 15 },
		{ 285, 197, 15 },
		{ 285, 198, 15 },
		{ 285, 199, 15 },
		{ 285, 200, 15 },
		{ 286, 131, 15 },
		{ 286, 132, 15 },
		{ 286, 133, 15 },
		{ 286, 134, 15 },
		{ 286, 135, 15 },
		{ 286, 136, 15 },
		{ 286, 137, 15 },
		{ 286, 138, 15 },
		{ 286, 139, 15 },
		{ 286, 140, 15 },
		{ 286, 141, 15 },
		{ 286, 142, 15 },
		{ 286, 143, 15 },
		{ 286, 144, 15 },
		{ 286, 145, 15 },
		{ 286, 146, 15 },
		{ 286, 147, 15 },
		{ 286, 148, 15 },
		{ 286, 149, 15 },
		{ 286, 150, 15 },
		{ 286, 151, 15 },
		{ 286, 152, 15 },
		{ 286, 153, 15 },
		{ 286, 154, 15 },
		{ 286, 155, 15 },
		{ 286, 156, 15 },
		{ 286, 157, 15 },
		{ 286, 158, 15 },
		{ 286, 159, 15 },
		{ 286, 160, 15 },
		{ 286, 161, 15 },
		{ 286, 162, 15 },
		{ 286, 163, 15 },
		{ 286, 164, 15 },
		{ 286, 165, 15 },
		{ 286, 166, 15 },
		{ 286, 167, 15 },
		{ 286, 168, 15 },
		{ 286, 169, 15 },
		{ 286, 170, 15 },
		{ 286, 171, 15 },
		{ 286, 172, 15 },
		{ 286, 173, 15 },
		{ 286, 174, 15 },
		{ 286, 175, 15 },
		{ 286, 176, 15 },
		{ 286, 177, 15 },
		{ 286, 178, 15 },
		{ 286, 179, 15 },
		{ 286, 180, 15 },
		{ 286, 181, 15 },
		{ 286, 182, 15 },
		{ 286, 183, 15 },
		{ 286, 184, 15 },
		{ 286, 185, 15 },
		{ 286, 186, 15 },
		{ 286, 187, 15 },
		{ 286, 188, 15 },
		{ 286, 189, 15 },
		{ 286, 190, 15 },
		{ 286, 191, 15 },
		{ 286, 192, 15 },
		{ 286, 193, 15 },
		{ 286, 194, 15 },
		{ 286, 195, 15 },
		{ 286, 196, 15 },
		{ 286, 197, 15 },
		{ 286, 198, 15 },
		{ 287, 131, 15 },
		{ 287, 132, 15 },
		{ 287, 133, 15 },
		{ 287, 134, 15 },
		{ 287, 135, 15 },
		{ 287, 136, 15 },
		{ 287, 137, 15 },
		{ 287, 138, 15 },
		{ 287, 139, 15 },
		{ 287, 140, 15 },
		{ 287, 141, 15 },
		{ 287, 142, 15 },
		{ 287, 143, 15 },
		{ 287, 144, 15 },
		{ 287, 145, 15 },
		{ 287, 146, 15 },
		{ 287, 147, 15 },
		{ 287, 148, 15 },
		{ 287, 149, 15 },
		{ 287, 150, 15 },
		{ 287, 151, 15 },
		{ 287, 152, 15 },
		{ 287, 153, 15 },
		{ 287, 154, 15 },
		{ 287, 155, 15 },
		{ 287, 156, 15 },
		{ 287, 157, 15 },
		{ 287, 158, 15 },
		{ 287, 159, 15 },
		{ 287, 160, 15 },
		{ 287, 161, 15 },
		{ 287, 162, 15 },
		{ 287, 163, 15 },
		{ 287, 164, 15 },
		{ 287, 165, 15 },
		{ 287, 166, 15 },
		{ 287, 167, 15 },
		{ 287, 168, 15 },
		{ 287, 169, 15 },
		{ 287, 170, 15 },
		{ 287, 171, 15 },
		{ 287, 172, 15 },
		{ 287, 173, 15 },
		{ 287, 174, 15 },
		{ 287, 175, 15 },
		{ 287, 176, 15 },
		{ 287, 177, 15 },
		{ 287, 178, 15 },
		{ 287, 179, 15 },
		{ 287, 180, 15 },
		{ 287, 181, 15 },
		{ 287, 182, 15 },
		{ 287, 183, 15 },
		{ 287, 184, 15 },
		{ 287, 185, 15 },
		{ 287, 186, 15 },
		{ 287, 187, 15 },
		{ 287, 188, 15 },
		{ 287, 189, 15 },
		{ 287, 190, 15 },
		{ 287, 191, 15 },
		{ 287, 192, 15 },
		{ 287, 193, 15 },
		{ 287, 194, 15 },
		{ 287, 195, 15 },
		{ 287, 196, 15 },
		{ 288, 131, 15 },
		{ 288, 132, 15 },
		{ 288, 133, 15 },
		{ 288, 134, 15 },
		{ 288, 135, 15 },
		{ 288, 136, 15 },
		{ 288, 137, 15 },
		{ 288, 138, 15 },
		{ 288, 139, 15 },
		{ 288, 140, 15 },
		{ 288, 141, 15 },
		{ 288, 142, 15 },
		{ 288, 143, 15 },
		{ 288, 144, 15 },
		{ 288, 145, 15 },
		{ 288, 146, 15 },
		{ 288, 147, 15 },
		{ 288, 148, 15 },
		{ 288, 149, 15 },
		{ 288, 150, 15 },
		{ 288, 151, 15 },
		{ 288, 152, 15 },
		{ 288, 153, 15 },
		{ 288, 154, 15 },
		{ 288, 155, 15 },
		{ 288, 156, 15 },
		{ 288, 157, 15 },
		{ 288, 158, 15 },
		{ 288, 159, 15 },
		{ 288, 160, 15 },
		{ 288, 161, 15 },
		{ 288, 162, 15 },
		{ 288, 163, 15 },
		{ 288, 164, 15 },
		{ 288, 165, 15 },
		{ 288, 166, 15 },
		{ 288, 167, 15 },
		{ 288, 168, 15 },
		{ 288, 169, 15 },
		{ 288, 170, 15 },
		{ 288, 171, 15 },
		{ 288, 172, 15 },
		{ 288, 173, 15 },
		{ 288, 174, 15 },
		{ 288, 175, 15 },
		{ 288, 176, 15 },
		{ 288, 177, 15 },
		{ 288, 178, 15 },
		{ 288, 179, 15 },
		{ 288, 180, 15 },
		{ 288, 181, 15 },
		{ 288, 182, 15 },
		{ 288, 183, 15 },
		{ 288, 184, 15 },
		{ 288, 185, 15 },
		{ 288, 186, 15 },
		{ 288, 187, 15 },
		{ 288, 188, 15 },
		{ 288, 189, 15 },
		{ 288, 190, 15 },
		{ 288, 191, 15 },
		{ 288, 192, 15 },
		{ 288, 193, 15 },
		{ 288, 194, 15 },
		{ 289, 131, 15 },
		{ 289, 132, 15 },
		{ 289, 133, 15 },
		{ 289, 134, 15 },
		{ 289, 135, 15 },
		{ 289, 136, 15 },
		{ 289, 137, 15 },
		{ 289, 138, 15 },
		{ 289, 139, 15 },
		{ 289, 140, 15 },
		{ 289, 141, 15 },
		{ 289, 142, 15 },
		{ 289, 143, 15 },
		{ 289, 144, 15 },
		{ 289, 145, 15 },
		{ 289, 146, 15 },
		{ 289, 147, 15 },
		{ 289, 148, 15 },
		{ 289, 149, 15 },
		{ 289, 150, 15 },
		{ 289, 151, 15 },
		{ 289, 152, 15 },
		{ 289, 153, 15 },
		{ 289, 154, 15 },
		{ 289, 155, 15 },
		{ 289, 156, 15 },
		{ 289, 157, 15 },
		{ 289, 158, 15 },
		{ 289, 159, 15 },
		{ 289, 160, 15 },
		{ 289, 161, 15 },
		{ 289, 162, 15 },
		{ 289, 163, 15 },
		{ 289, 164, 15 },
		{ 289, 165, 15 },
		{ 289, 166, 15 },
		{ 289, 167, 15 },
		{ 289, 168, 15 },
		{ 289, 169, 15 },
		{ 289, 170, 15 },
		{ 289, 171, 15 },
		{ 289, 172, 15 },
		{ 289, 173, 15 },
		{ 289, 174, 15 },
		{ 289, 175, 15 },
		{ 289, 176, 15 },
		{ 289, 177, 15 },
		{ 289, 178, 15 },
		{ 289, 179, 15 },
		{ 289, 180, 15 },
		{ 289, 181, 15 },
		{ 289, 182, 15 },
		{ 289, 183, 15 },
		{ 289, 184, 15 },
		{ 289, 185, 15 },
		{ 289, 186, 15 },
		{ 289, 187, 15 },
		{ 289, 188, 15 },
		{ 289, 189, 15 },
		{ 289, 190, 15 },
		{ 289, 191, 15 },
		{ 289, 192, 15 },
		{ 290, 131, 15 },
		{ 290, 132, 15 },
		{ 290, 133, 15 },
		{ 290, 134, 15 },
		{ 290, 135, 15 },
		{ 290, 136, 15 },
		{ 290, 137, 15 },
		{ 290, 138, 15 },
		{ 290, 139, 15 },
		{ 290, 140, 15 },
		{ 290, 141, 15 },
		{ 290, 142, 15 },
		{ 290, 143, 15 },
		{ 290, 144, 15 },
		{ 290, 145, 15 },
		{ 290, 146, 15 },
		{ 290, 147, 15 },
		{ 290, 148, 15 },
		{ 290, 149, 15 },
		{ 290, 150, 15 },
		{ 290, 151, 15 },
		{ 290, 152, 15 },
		{ 290, 153, 15 },
		{ 290, 154, 15 },
		{ 290, 155, 15 },
		{ 290, 156, 15 },
		{ 290, 157, 15 },
		{ 290, 158, 15 },
		{ 290, 159, 15 },
		{ 290, 160, 15 },
		{ 290, 161, 15 },
		{ 290, 162, 15 },
		{ 290, 163, 15 },
		{ 290, 164, 15 },
		{ 290, 165, 15 },
		{ 290, 166, 15 },
		{ 290, 167, 15 },
		{ 290, 168, 15 },
		{ 290, 169, 15 },
		{ 290, 170, 15 },
		{ 290, 171, 15 },
		{ 290, 172, 15 },
		{ 290, 173, 15 },
		{ 290, 174, 15 },
		{ 290, 175, 15 },
		{ 290, 176, 15 },
		{ 290, 177, 15 },
		{ 290, 178, 15 },
		{ 290, 179, 15 },
		{ 290, 180, 15 },
		{ 290, 181, 15 },
		{ 290, 182, 15 },
		{ 290, 183, 15 },
		{ 290, 184, 15 },
		{ 290, 185, 15 },
		{ 290, 186, 15 },
		{ 290, 187, 15 },
		{ 290, 188, 15 },
		{ 290, 189, 15 },
		{ 290, 190, 15 },
		{ 291, 131, 15 },
		{ 291, 132, 15 },
		{ 291, 133, 15 },
		{ 291, 134, 15 },
		{ 291, 135, 15 },
		{ 291, 136, 15 },
		{ 291, 137, 15 },
		{ 291, 138, 15 },
		{ 291, 139, 15 },
		{ 291, 140, 15 },
		{ 291, 141, 15 },
		{ 291, 142, 15 },
		{ 291, 143, 15 },
		{ 291, 144, 15 },
		{ 291, 145, 15 },
		{ 291, 146, 15 },
		{ 291, 147, 15 },
		{ 291, 148, 15 },
		{ 291, 149, 15 },
		{ 291, 150, 15 },
		{ 291, 151, 15 },
		{ 291, 152, 15 },
		{ 291, 153, 15 },
		{ 291, 154, 15 },
		{ 291, 155, 15 },
		{ 291, 156, 15 },
		{ 291, 157, 15 },
		{ 291, 158, 15 },
		{ 291, 159, 15 },
		{ 291, 160, 15 },
		{ 291, 161, 15 },
		{ 291, 162, 15 },
		{ 291, 163, 15 },
		{ 291, 164, 15 },
		{ 291, 165, 15 },
		{ 291, 166, 15 },
		{ 291, 167, 15 },
		{ 291, 168, 15 },
		{ 291, 169, 15 },
		{ 291, 170, 15 },
		{ 291, 171, 15 },
		{ 291, 172, 15 },
		{ 291, 173, 15 },
		{ 291, 174, 15 },
		{ 291, 175, 15 },
		{ 291, 176, 15 },
		{ 291, 177, 15 },
		{ 291, 178, 15 },
		{ 291, 179, 15 },
		{ 291, 180, 15 },
		{ 291, 181, 15 },
		{ 291, 182, 15 },
		{ 291, 183, 15 },
		{ 291, 184, 15 },
		{ 291, 185, 15 },
		{ 291, 186, 15 },
		{ 291, 187, 15 },
		{ 291, 188, 15 },
		{ 292, 131, 15 },
		{ 292, 132, 15 },
		{ 292, 133, 15 },
		{ 292, 134, 15 },
		{ 292, 135, 15 },
		{ 292, 136, 15 },
		{ 292, 137, 15 },
		{ 292, 138, 15 },
		{ 292, 139, 15 },
		{ 292, 140, 15 },
		{ 292, 141, 15 },
		{ 292, 142, 15 },
		{ 292, 143, 15 },
		{ 292, 144, 15 },
		{ 292, 145, 15 },
		{ 292, 146, 15 },
		{ 292, 147, 15 },
		{ 292, 148, 15 },
		{ 292, 149, 15 },
		{ 292, 150, 15 },
		{ 292, 151, 15 },
		{ 292, 152, 15 },
		{ 292, 153, 15 },
		{ 292, 154, 15 },
		{ 292, 155, 15 },
		{ 292, 156, 15 },
		{ 292, 157, 15 },
		{ 292, 158, 15 },
		{ 292, 159, 15 },
		{ 292, 160, 15 },
		{ 292, 161, 15 },
		{ 292, 162, 15 },
		{ 292, 163, 15 },
		{ 292, 164, 15 },
		{ 292, 165, 15 },
		{ 292, 166, 15 },
		{ 292, 167, 15 },
		{ 292, 168, 15 },
		{ 292, 169, 15 },
		{ 292, 170, 15 },
		{ 292, 171, 15 },
		{ 292, 172, 15 },
		{ 292, 173, 15 },
		{ 292, 174, 15 },
		{ 292, 175, 15 },
		{ 292, 176, 15 },
		{ 292, 177, 15 },
		{ 292, 178, 15 },
		{ 292, 179, 15 },
		{ 292, 180, 15 },
		{ 292, 181, 15 },
		{ 292, 182, 15 },
		{ 292, 183, 15 },
		{ 292, 184, 15 },
		{ 292, 185, 15 },
		{ 292, 186, 15 },
		{ 293, 131, 15 },
		{ 293, 132, 15 },
		{ 293, 133, 15 },
		{ 293, 134, 15 },
		{ 293, 135, 15 },
		{ 293, 136, 15 },
		{ 293, 137, 15 },
		{ 293, 138, 15 },
		{ 293, 139, 15 },
		{ 293, 140, 15 },
		{ 293, 141, 15 },
		{ 293, 142, 15 },
		{ 293, 143, 15 },
		{ 293, 144, 15 },
		{ 293, 145, 15 },
		{ 293, 146, 15 },
		{ 293, 147, 15 },
		{ 293, 148, 15 },
		{ 293, 149, 15 },
		{ 293, 150, 15 },
		{ 293, 151, 15 },
		{ 293, 152, 15 },
		{ 293, 153, 15 },
		{ 293, 154, 15 },
		{ 293, 155, 15 },
		{ 293, 156, 15 },
		{ 293, 157, 15 },
		{ 293, 158, 15 },
		{ 293, 159, 15 },
		{ 293, 160, 15 },
		{ 293, 161, 15 },
		{ 293, 162, 15 },
		{ 293, 163, 15 },
		{ 293, 164, 15 },
		{ 293, 165, 15 },
		{ 293, 166, 15 },
		{ 293, 167, 15 },
		{ 293, 168, 15 },
		{ 293, 169, 15 },
		{ 293, 170, 15 },
		{ 293, 171, 15 },
		{ 293, 172, 15 },
		{ 293, 173, 15 },
		{ 293, 174, 15 },
		{ 293, 175, 15 },
		{ 293, 176, 15 },
		{ 293, 177, 15 },
		{ 293, 178, 15 },
		{ 293, 179, 15 },
		{ 293, 180, 15 },
		{ 293, 181, 15 },
		{ 293, 182, 15 },
		{ 293, 183, 15 },
		{ 293, 184, 15 },
		{ 294, 131, 15 },
		{ 294, 132, 15 },
		{ 294, 133, 15 },
		{ 294, 134, 15 },
		{ 294, 135, 15 },
		{ 294, 136, 15 },
		{ 294, 137, 15 },
		{ 294, 138, 15 },
		{ 294, 139, 15 },
		{ 294, 140, 15 },
		{ 294, 141, 15 },
		{ 294, 142, 15 },
		{ 294, 143, 15 },
		{ 294, 144, 15 },
		{ 294, 145, 15 },
		{ 294, 146, 15 },
		{ 294, 147, 15 },
		{ 294, 148, 15 },
		{ 294, 149, 15 },
		{ 294, 150, 15 },
		{ 294, 151, 15 },
		{ 294, 152, 15 },
		{ 294, 153, 15 },
		{ 294, 154, 15 },
		{ 294, 155, 15 },
		{ 294, 156, 15 },
		{ 294, 157, 15 },
		{ 294, 158, 15 },
		{ 294, 159, 15 },
		{ 294, 160, 15 },
		{ 294, 161, 15 },
		{ 294, 162, 15 },
		{ 294, 163, 15 },
		{ 294, 164, 15 },
		{ 294, 165, 15 },
		{ 294, 166, 15 },
		{ 294, 167, 15 },
		{ 294, 168, 15 },
		{ 294, 169, 15 },
		{ 294, 170, 15 },
		{ 294, 171, 15 },
		{ 294, 172, 15 },
		{ 294, 173, 15 },
		{ 294, 174, 15 },
		{ 294, 175, 15 },
		{ 294, 176, 15 },
		{ 294, 177, 15 },
		{ 294, 178, 15 },
		{ 294, 179, 15 },
		{ 294, 180, 15 },
		{ 294, 181, 15 },
		{ 294, 182, 15 },
		{ 295, 131, 15 },
		{ 295, 132, 15 },
		{ 295, 133, 15 },
		{ 295, 134, 15 },
		{ 295, 135, 15 },
		{ 295, 136, 15 },
		{ 295, 137, 15 },
		{ 295, 138, 15 },
		{ 295, 139, 15 },
		{ 295, 140, 15 },
		{ 295, 141, 15 },
		{ 295, 142, 15 },
		{ 295, 143, 15 },
		{ 295, 144, 15 },
		{ 295, 145, 15 },
		{ 295, 146, 15 },
		{ 295, 147, 15 },
		{ 295, 148, 15 },
		{ 295, 149, 15 },
		{ 295, 150, 15 },
		{ 295, 151, 15 },
		{ 295, 152, 15 },
		{ 295, 153, 15 },
		{ 295, 154, 15 },
		{ 295, 155, 15 },
		{ 295, 156, 15 },
		{ 295, 157, 15 },
		{ 295, 158, 15 },
		{ 295, 159, 15 },
		{ 295, 160, 15 },
		{ 295, 161, 15 },
		{ 295, 162, 15 },
		{ 295, 163, 15 },
		{ 295, 164, 15 },
		{ 295, 165, 15 },
		{ 295, 166, 15 },
		{ 295, 167, 15 },
		{ 295, 168, 15 },
		{ 295, 169, 15 },
		{ 295, 170, 15 },
		{ 295, 171, 15 },
		{ 295, 172, 15 },
		{ 295, 173, 15 },
		{ 295, 174, 15 },
		{ 295, 175, 15 },
		{ 295, 176, 15 },
		{ 295, 177, 15 },
		{ 295, 178, 15 },
		{ 295, 179, 15 },
		{ 295, 180, 15 },
		{ 296, 131, 15 },
		{ 296, 132, 15 },
		{ 296, 133, 15 },
		{ 296, 134, 15 },
		{ 296, 135, 15 },
		{ 296, 136, 15 },
		{ 296, 137, 15 },
		{ 296, 138, 15 },
		{ 296, 139, 15 },
		{ 296, 140, 15 },
		{ 296, 141, 15 },
		{ 296, 142, 15 },
		{ 296, 143, 15 },
		{ 296, 144, 15 },
		{ 296, 145, 15 },
		{ 296, 146, 15 },
		{ 296, 147, 15 },
		{ 296, 148, 15 },
		{ 296, 149, 15 },
		{ 296, 150, 15 },
		{ 296, 151, 15 },
		{ 296, 152, 15 },
		{ 296, 153, 15 },
		{ 296, 154, 15 },
		{ 296, 155, 15 },
		{ 296, 156, 15 },
		{ 296, 157, 15 },
		{ 296, 158, 15 },
		{ 296, 159, 15 },
		{ 296, 160, 15 },
		{ 296, 161, 15 },
		{ 296, 162, 15 },
		{ 296, 163, 15 },
		{ 296, 164, 15 },
		{ 296, 165, 15 },
		{ 296, 166, 15 },
		{ 296, 167, 15 },
		{ 296, 168, 15 },
		{ 296, 169, 15 },
		{ 296, 170, 15 },
		{ 296, 171, 15 },
		{ 296, 172, 15 },
		{ 296, 173, 15 },
		{ 296, 174, 15 },
		{ 296, 175, 15 },
		{ 296, 176, 15 },
		{ 296, 177, 15 },
		{ 296, 178, 15 },
		{ 297, 131, 15 },
		{ 297, 132, 15 },
		{ 297, 133, 15 },
		{ 297, 134, 15 },
		{ 297, 135, 15 },
		{ 297, 136, 15 },
		{ 297, 137, 15 },
		{ 297, 138, 15 },
		{ 297, 139, 15 },
		{ 297, 140, 15 },
		{ 297, 141, 15 },
		{ 297, 142, 15 },
		{ 297, 143, 15 },
		{ 297, 144, 15 },
		{ 297, 145, 15 },
		{ 297, 146, 15 },
		{ 297, 147, 15 },
		{ 297, 148, 15 },
		{ 297, 149, 15 },
		{ 297, 150, 15 },
		{ 297, 151, 15 },
		{ 297, 152, 15 },
		{ 297, 153, 15 },
		{ 297, 154, 15 },
		{ 297, 155, 15 },
		{ 297, 156, 15 },
		{ 297, 157, 15 },
		{ 297, 158, 15 },
		{ 297, 159, 15 },
		{ 297, 160, 15 },
		{ 297, 161, 15 },
		{ 297, 162, 15 },
		{ 297, 163, 15 },
		{ 297, 164, 15 },
		{ 297, 165, 15 },
		{ 297, 166, 15 },
		{ 297, 167, 15 },
		{ 297, 168, 15 },
		{ 297, 169, 15 },
		{ 297, 170, 15 },
		{ 297, 171, 15 },
		{ 297, 172, 15 },
		{ 297, 173, 15 },
		{ 297, 174, 15 },
		{ 297, 175, 15 },
		{ 297, 176, 15 },
		{ 298, 131, 15 },
		{ 298, 132, 15 },
		{ 298, 133, 15 },
		{ 298, 134, 15 },
		{ 298, 135, 15 },
		{ 298, 136, 15 },
		{ 298, 137, 15 },
		{ 298, 138, 15 },
		{ 298, 139, 15 },
		{ 298, 140, 15 },
		{ 298, 141, 15 },
		{ 298, 142, 15 },
		{ 298, 143, 15 },
		{ 298, 144, 15 },
		{ 298, 145, 15 },
		{ 298, 146, 15 },
		{ 298, 147, 15 },
		{ 298, 148, 15 },
		{ 298, 149, 15 },
		{ 298, 150, 15 },
		{ 298, 151, 15 },
		{ 298, 152, 15 },
		{ 298, 153, 15 },
		{ 298, 154, 15 },
		{ 298, 155, 15 },
		{ 298, 156, 15 },
		{ 298, 157, 15 },
		{ 298, 158, 15 },
		{ 298, 159, 15 },
		{ 298, 160, 15 },
		{ 298, 161, 15 },
		{ 298, 162, 15 },
		{ 298, 163, 15 },
		{ 298, 164, 15 },
		{ 298, 165, 15 },
		{ 298, 166, 15 },
		{ 298, 167, 15 },
		{ 298, 168, 15 },
		{ 298, 169, 15 },
		{ 298, 170, 15 },
		{ 298, 171, 15 },
		{ 298, 172, 15 },
		{ 298, 173, 15 },
		{ 298, 174, 15 },
		{ 299, 131, 15 },
		{ 299, 132, 15 },
		{ 299, 133, 15 },
		{ 299, 134, 15 },
		{ 299, 135, 15 },
		{ 299, 136, 15 },
		{ 299, 137, 15 },
		{ 299, 138, 15 },
		{ 299, 139, 15 },
		{ 299, 140, 15 },
		{ 299, 141, 15 },
		{ 299, 142, 15 },
		{ 299, 143, 15 },
		{ 299, 144, 15 },
		{ 299, 145, 15 },
		{ 299, 146, 15 },
		{ 299, 147, 15 },
		{ 299, 148, 15 },
		{ 299, 149, 15 },
		{ 299, 150, 15 },
		{ 299, 151, 15 },
		{ 299, 152, 15 },
		{ 299, 153, 15 },
		{ 299, 154, 15 },
		{ 299, 155, 15 },
		{ 299, 156, 15 },
		{ 299, 157, 15 },
		{ 299, 158, 15 },
		{ 299, 159, 15 },
		{ 299, 160, 15 },
		{ 299, 161, 15 },
		{ 299, 162, 15 },
		{ 299, 163, 15 },
		{ 299, 164, 15 },
		{ 299, 165, 15 },
		{ 299, 166, 15 },
		{ 299, 167, 15 },
		{ 299, 168, 15 },
		{ 299, 169, 15 },
		{ 299, 170, 15 },
		{ 299, 171, 15 },
		{ 299, 172, 15 },
		{ 300, 131, 15 },
		{ 300, 132, 15 },
		{ 300, 133, 15 },
		{ 300, 134, 15 },
		{ 300, 135, 15 },
		{ 300, 136, 15 },
		{ 300, 137, 15 },
		{ 300, 138, 15 },
		{ 300, 139, 15 },
		{ 300, 140, 15 },
		{ 300, 141, 15 },
		{ 300, 142, 15 },
		{ 300, 143, 15 },
		{ 300, 144, 15 },
		{ 300, 145, 15 },
		{ 300, 146, 15 },
		{ 300, 147, 15 },
		{ 300, 148, 15 },
		{ 300, 149, 15 },
		{ 300, 150, 15 },
		{ 300, 151, 15 },
		{ 300, 152, 15 },
		{ 300, 153, 15 },
		{ 300, 154, 15 },
		{ 300, 155, 15 },
		{ 300, 156, 15 },
		{ 300, 157, 15 },
		{ 300, 158, 15 },
		{ 300, 159, 15 },
		{ 300, 160, 15 },
		{ 300, 161, 15 },
		{ 300, 162, 15 },
		{ 300, 163, 15 },
		{ 300, 164, 15 },
		{ 300, 165, 15 },
		{ 300, 166, 15 },
		{ 300, 167, 15 },
		{ 300, 168, 15 },
		{ 300, 169, 15 },
		{ 300, 170, 15 },
		{ 301, 131, 15 },
		{ 301, 132, 15 },
		{ 301, 133, 15 },
		{ 301, 134, 15 },
		{ 301, 135, 15 },
		{ 301, 136, 15 },
		{ 301, 137, 15 },
		{ 301, 138, 15 },
		{ 301, 139, 15 },
		{ 301, 140, 15 },
		{ 301, 141, 15 },
		{ 301, 142, 15 },
		{ 301, 143, 15 },
		{ 301, 144, 15 },
		{ 301, 145, 15 },
		{ 301, 146, 15 },
		{ 301, 147, 15 },
		{ 301, 148, 15 },
		{ 301, 149, 15 },
		{ 301, 150, 15 },
		{ 301, 151, 15 },
		{ 301, 152, 15 },
		{ 301, 153, 15 },
		{ 301, 154, 15 },
		{ 301, 155, 15 },
		{ 301, 156, 15 },
		{ 301, 157, 15 },
		{ 301, 158, 15 },
		{ 301, 159, 15 },
		{ 301, 160, 15 },
		{ 301, 161, 15 },
		{ 301, 162, 15 },
		{ 301, 163, 15 },
		{ 301, 164, 15 },
		{ 301, 165, 15 },
		{ 301, 166, 15 },
		{ 301, 167, 15 },
		{ 301, 168, 15 },
		{ 302, 131, 15 },
		{ 302, 132, 15 },
		{ 302, 133, 15 },
		{ 302, 134, 15 },
		{ 302, 135, 15 },
		{ 302, 136, 15 },
		{ 302, 137, 15 },
		{ 302, 138, 15 },
		{ 302, 139, 15 },
		{ 302, 140, 15 },
		{ 302, 141, 15 },
		{ 302, 142, 15 },
		{ 302, 143, 15 },
		{ 302, 144, 15 },
		{ 302, 145, 15 },
		{ 302, 146, 15 },
		{ 302, 147, 15 },
		{ 302, 148, 15 },
		{ 302, 149, 15 },
		{ 302, 150, 15 },
		{ 302, 151, 15 },
		{ 302, 152, 15 },
		{ 302, 153, 15 },
		{ 302, 154, 15 },
		{ 302, 155, 15 },
		{ 302, 156, 15 },
		{ 302, 157, 15 },
		{ 302, 158, 15 },
		{ 302, 159, 15 },
		{ 302, 160, 15 },
		{ 302, 161, 15 },
		{ 302, 162, 15 },
		{ 302, 163, 15 },
		{ 302, 164, 15 },
		{ 302, 165, 15 },
		{ 302, 166, 15 },
		{ 303, 131, 15 },
		{ 303, 132, 15 },
		{ 303, 133, 15 },
		{ 303, 134, 15 },
		{ 303, 135, 15 },
		{ 303, 136, 15 },
		{ 303, 137, 15 },
		{ 303, 138, 15 },
		{ 303, 139, 15 },
		{ 303, 140, 15 },
		{ 303, 141, 15 },
		{ 303, 142, 15 },
		{ 303, 143, 15 },
		{ 303, 144, 15 },
		{ 303, 145, 15 },
		{ 303, 146, 15 },
		{ 303, 147, 15 },
		{ 303, 148, 15 },
		{ 303, 149, 15 },
		{ 303, 150, 15 },
		{ 303, 151, 15 },
		{ 303, 152, 15 },
		{ 303, 153, 15 },
		{ 303, 154, 15 },
		{ 303, 155, 15 },
		{ 303, 156, 15 },
		{ 303, 157, 15 },
		{ 303, 158, 15 },
		{ 303, 159, 15 },
		{ 303, 160, 15 },
		{ 303, 161, 15 },
		{ 303, 162, 15 },
		{ 303, 163, 15 },
		{ 303, 164, 15 },
		{ 304, 131, 15 },
		{ 304, 132, 15 },
		{ 304, 133, 15 },
		{ 304, 134, 15 },
		{ 304, 135, 15 },
		{ 304, 136, 15 },
		{ 304, 137, 15 },
		{ 304, 138, 15 },
		{ 304, 139, 15 },
		{ 304, 140, 15 },
		{ 304, 141, 15 },
		{ 304, 142, 15 },
		{ 304, 143, 15 },
		{ 304, 144, 15 },
		{ 304, 145, 15 },
		{ 304, 146, 15 },
		{ 304, 147, 15 },
		{ 304, 148, 15 },
		{ 304, 149, 15 },
		{ 304, 150, 15 },
		{ 304, 151, 15 },
		{ 304, 152, 15 },
		{ 304, 153, 15 },
		{ 304, 154, 15 },
		{ 304, 155, 15 },
		{ 304, 156, 15 },
		{ 304, 157, 15 },
		{ 304, 158, 15 },
		{ 304, 159, 15 },
		{ 304, 160, 15 },
		{ 304, 161, 15 },
		{ 304, 162, 15 },
		{ 305, 131, 15 },
		{ 305, 132, 15 },
		{ 305, 133, 15 },
		{ 305, 134, 15 },
		{ 305, 135, 15 },
		{ 305, 136, 15 },
		{ 305, 137, 15 },
		{ 305, 138, 15 },
		{ 305, 139, 15 },
		{ 305, 140, 15 },
		{ 305, 141, 15 },
		{ 305, 142, 15 },
		{ 305, 143, 15 },
		{ 305, 144, 15 },
		{ 305, 145, 15 },
		{ 305, 146, 15 },
		{ 305, 147, 15 },
		{ 305, 148, 15 },
		{ 305, 149, 15 },
		{ 305, 150, 15 },
		{ 305, 151, 15 },
		{ 305, 152, 15 },
		{ 305, 153, 15 },
		{ 305, 154, 15 },
		{ 305, 155, 15 },
		{ 305, 156, 15 },
		{ 305, 157, 15 },
		{ 305, 158, 15 },
		{ 305, 159, 15 },
		{ 305, 160, 15 },
		{ 306, 131, 15 },
		{ 306, 132, 15 },
		{ 306, 133, 15 },
		{ 306, 134, 15 },
		{ 306, 135, 15 },
		{ 306, 136, 15 },
		{ 306, 137, 15 },
		{ 306, 138, 15 },
		{ 306, 139, 15 },
		{ 306, 140, 15 },
		{ 306, 141, 15 },
		{ 306, 142, 15 },
		{ 306, 143, 15 },
		{ 306, 144, 15 },
		{ 306, 145, 15 },
		{ 306, 146, 15 },
		{ 306, 147, 15 },
		{ 306, 148, 15 },
		{ 306, 149, 15 },
		{ 306, 150, 15 },
		{ 306, 151, 15 },
		{ 306, 152, 15 },
		{ 306, 153, 15 },
		{ 306, 154, 15 },
		{ 306, 155, 15 },
		{ 306, 156, 15 },
		{ 306, 157, 15 },
		{ 306, 158, 15 },
		{ 307, 131, 15 },
		{ 307, 132, 15 },
		{ 307, 133, 15 },
		{ 307, 134, 15 },
		{ 307, 135, 15 },
		{ 307, 136, 15 },
		{ 307, 137, 15 },
		{ 307, 138, 15 },
		{ 307, 139, 15 },
		{ 307, 140, 15 },
		{ 307, 141, 15 },
		{ 307, 142, 15 },
		{ 307, 143, 15 },
		{ 307, 144, 15 },
		{ 307, 145, 15 },
		{ 307, 146, 15 },
		{ 307, 147, 15 },
		{ 307, 148, 15 },
		{ 307, 149, 15 },
		{ 307, 150, 15 },
		{ 307, 151, 15 },
		{ 307, 152, 15 },
		{ 307, 153, 15 },
		{ 307, 154, 15 },
		{ 307, 155, 15 },
		{ 307, 156, 15 },
		{ 308, 131, 15 },
		{ 308, 132, 15 },
		{ 308, 133, 15 },
		{ 308, 134, 15 },
		{ 308, 135, 15 },
		{ 308, 136, 15 },
		{ 308, 137, 15 },
		{ 308, 138, 15 },
		{ 308, 139, 15 },
		{ 308, 140, 15 },
		{ 308, 141, 15 },
		{ 308, 142, 15 },
		{ 308, 143, 15 },
		{ 308, 144, 15 },
		{ 308, 145, 15 },
		{ 308, 146, 15 },
		{ 308, 147, 15 },
		{ 308, 148, 15 },
		{ 308, 149, 15 },
		{ 308, 150, 15 },
		{ 308, 151, 15 },
		{ 308, 152, 15 },
		{ 308, 153, 15 },
		{ 308, 154, 15 },
		{ 309, 131, 15 },
		{ 309, 132, 15 },
		{ 309, 133, 15 },
		{ 309, 134, 15 },
		{ 309, 135, 15 },
		{ 309, 136, 15 },
		{ 309, 137, 15 },
		{ 309, 138, 15 },
		{ 309, 139, 15 },
		{ 309, 140, 15 },
		{ 309, 141, 15 },
		{ 309, 142, 15 },
		{ 309, 143, 15 },
		{ 309, 144, 15 },
		{ 309, 145, 15 },
		{ 309, 146, 15 },
		{ 309, 147, 15 },
		{ 309, 148, 15 },
		{ 309, 149, 15 },
		{ 309, 150, 15 },
		{ 309, 151, 15 },
		{ 309, 152, 15 },
		{ 310, 131, 15 },
		{ 310, 132, 15 },
		{ 310, 133, 15 },
		{ 310, 134, 15 },
		{ 310, 135, 15 },
		{ 310, 136, 15 },
		{ 310, 137, 15 },
		{ 310, 138, 15 },
		{ 310, 139, 15 },
		{ 310, 140, 15 },
		{ 310, 141, 15 },
		{ 310, 142, 15 },
		{ 310, 143, 15 },
		{ 310, 144, 15 },
		{ 310, 145, 15 },
		{ 310, 146, 15 },
		{ 310, 147, 15 },
		{ 310, 148, 15 },
		{ 310, 149, 15 },
		{ 310, 150, 15 },
		{ 311, 131, 15 },
		{ 311, 132, 15 },
		{ 311, 133, 15 },
		{ 311, 134, 15 },
		{ 311, 135, 15 },
		{ 311, 136, 15 },
		{ 311, 137, 15 },
		{ 311, 138, 15 },
		{ 311, 139, 15 },
		{ 311, 140, 15 },
		{ 311, 141, 15 },
		{ 311, 142, 15 },
		{ 311, 143, 15 },
		{ 311, 144, 15 },
		{ 311, 145, 15 },
		{ 311, 146, 15 },
		{ 311, 147, 15 },
		{ 311, 148, 15 },
		{ 312, 131, 15 },
		{ 312, 132, 15 },
		{ 312, 133, 15 },
		{ 312, 134, 15 },
		{ 312, 135, 15 },
		{ 312, 136, 15 },
		{ 312, 137, 15 },
		{ 312, 138, 15 },
		{ 312, 139, 15 },
		{ 312, 140, 15 },
		{ 312, 141, 15 },
		{ 312, 142, 15 },
		{ 312, 143, 15 },
		{ 312, 144, 15 },
		{ 312, 145, 15 },
		{ 312, 146, 15 },
		{ 313, 131, 15 },
		{ 313, 132, 15 },
		{ 313, 133, 15 },
		{ 313, 134, 15 },
		{ 313, 135, 15 },
		{ 313, 136, 15 },
		{ 313, 137, 15 },
		{ 313, 138, 15 },
		{ 313, 139, 15 },
		{ 313, 140, 15 },
		{ 313, 141, 15 },
		{ 313, 142, 15 },
		{ 313, 143, 15 },
		{ 313, 144, 15 },
		{ 314, 132, 15 },
		{ 314, 133, 15 },
		{ 314, 134, 15 },
		{ 314, 135, 15 },
		{ 314, 136, 15 },
		{ 314, 137, 15 },
		{ 314, 138, 15 },
		{ 314, 139, 15 },
		{ 314, 140, 15 },
		{ 314, 141, 15 },
		{ 314, 142, 15 },
		{ 315, 132, 15 },
		{ 315, 133, 15 },
		{ 315, 134, 15 },
		{ 315, 135, 15 },
		{ 315, 136, 15 },
		{ 315, 137, 15 },
		{ 315, 138, 15 },
		{ 315, 139, 15 },
		{ 315, 140, 15 },
		{ 316, 132, 15 },
		{ 316, 133, 15 },
		{ 316, 134, 15 },
		{ 316, 135, 15 },
		{ 316, 136, 15 },
		{ 316, 137, 15 },
		{ 316, 138, 15 },
		{ 317, 133, 15 },
		{ 317, 134, 15 },
		{ 317, 135, 15 },
		{ 166, 380, 16 },
		{ 166, 381, 16 },
		{ 167, 378, 16 },
		{ 167, 379, 16 },
		{ 167, 380, 16 },
		{ 167, 381, 16 },
		{ 167, 382, 16 },
		{ 168, 377, 16 },
		{ 168, 378, 16 },
		{ 168, 379, 16 },
		{ 168, 380, 16 },
		{ 168, 381, 16 },
		{ 168, 382, 16 },
		{ 169, 375, 16 },
		{ 169, 376, 16 },
		{ 169, 377, 16 },
		{ 169, 378, 16 },
		{ 169, 379, 16 },
		{ 169, 380, 16 },
		{ 169, 381, 16 },
		{ 169, 382, 16 },
		{ 170, 374, 16 },
		{ 170, 375, 16 },
		{ 170, 376, 16 },
		{ 170, 377, 16 },
		{ 170, 378, 16 },
		{ 170, 379, 16 },
		{ 170, 380, 16 },
		{ 170, 381, 16 },
		{ 170, 382, 16 },
		{ 170, 383, 16 },
		{ 171, 372, 16 },
		{ 171, 373, 16 },
		{ 171, 374, 16 },
		{ 171, 375, 16 },
		{ 171, 376, 16 },
		{ 171, 377, 16 },
		{ 171, 378, 16 },
		{ 171, 379, 16 },
		{ 171, 380, 16 },
		{ 171, 381, 16 },
		{ 171, 382, 16 },
		{ 171, 383, 16 },
		{ 172, 371, 16 },
		{ 172, 372, 16 },
		{ 172, 373, 16 },
		{ 172, 374, 16 },
		{ 172, 375, 16 },
		{ 172, 376, 16 },
		{ 172, 377, 16 },
		{ 172, 378, 16 },
		{ 172, 379, 16 },
		{ 172, 380, 16 },
		{ 172, 381, 16 },
		{ 172, 382, 16 },
		{ 172, 383, 16 },
		{ 173, 370, 16 },
		{ 173, 371, 16 },
		{ 173, 372, 16 },
		{ 173, 373, 16 },
		{ 173, 374, 16 },
		{ 173, 375, 16 },
		{ 173, 376, 16 },
		{ 173, 377, 16 },
		{ 173, 378, 16 },
		{ 173, 379, 16 },
		{ 173, 380, 16 },
		{ 173, 381, 16 },
		{ 173, 382, 16 },
		{ 173, 383, 16 },
		{ 174, 368, 16 },
		{ 174, 369, 16 },
		{ 174, 370, 16 },
		{ 174, 371, 16 },
		{ 174, 372, 16 },
		{ 174, 373, 16 },
		{ 174, 374, 16 },
		{ 174, 375, 16 },
		{ 174, 376, 16 },
		{ 174, 377, 16 },
		{ 174, 378, 16 },
		{ 174, 379, 16 },
		{ 174, 380, 16 },
		{ 174, 381, 16 },
		{ 174, 382, 16 },
		{ 174, 383, 16 },
		{ 175, 367, 16 },
		{ 175, 368, 16 },
		{ 175, 369, 16 },
		{ 175, 370, 16 },
		{ 175, 371, 16 },
		{ 175, 372, 16 },
		{ 175, 373, 16 },
		{ 175, 374, 16 },
		{ 175, 375, 16 },
		{ 175, 376, 16 },
		{ 175, 377, 16 },
		{ 175, 378, 16 },
		{ 175, 379, 16 },
		{ 175, 380, 16 },
		{ 175, 381, 16 },
		{ 175, 382, 16 },
		{ 175, 383, 16 },
		{ 176, 366, 16 },
		{ 176, 367, 16 },
		{ 176, 368, 16 },
		{ 176, 369, 16 },
		{ 176, 370, 16 },
		{ 176, 371, 16 },
		{ 176, 372, 16 },
		{ 176, 373, 16 },
		{ 176, 374, 16 },
		{ 176, 375, 16 },
		{ 176, 376, 16 },
		{ 176, 377, 16 },
		{ 176, 378, 16 },
		{ 176, 379, 16 },
		{ 176, 380, 16 },
		{ 176, 381, 16 },
		{ 176, 382, 16 },
		{ 176, 383, 16 },
		{ 177, 364, 16 },
		{ 177, 365, 16 },
		{ 177, 366, 16 },
		{ 177, 367, 16 },
		{ 177, 368, 16 },
		{ 177, 369, 16 },
		{ 177, 370, 16 },
		{ 177, 371, 16 },
		{ 177, 372, 16 },
		{ 177, 373, 16 },
		{ 177, 374, 16 },
		{ 177, 375, 16 },
		{ 177, 376, 16 },
		{ 177, 377, 16 },
		{ 177, 378, 16 },
		{ 177, 379, 16 },
		{ 177, 380, 16 },
		{ 177, 381, 16 },
		{ 177, 382, 16 },
		{ 177, 383, 16 },
		{ 178, 363, 16 },
		{ 178, 364, 16 },
		{ 178, 365, 16 },
		{ 178, 366, 16 },
		{ 178, 367, 16 },
		{ 178, 368, 16 },
		{ 178, 369, 16 },
		{ 178, 370, 16 },
		{ 178, 371, 16 },
		{ 178, 372, 16 },
		{ 178, 373, 16 },
		{ 178, 374, 16 },
		{ 178, 375, 16 },
		{ 178, 376, 16 },
		{ 178, 377, 16 },
		{ 178, 378, 16 },
		{ 178, 379, 16 },
		{ 178, 380, 16 },
		{ 178, 381, 16 },
		{ 178, 382, 16 },
		{ 178, 383, 16 },
		{ 179, 362, 16 },
		{ 179, 363, 16 },
		{ 179, 364, 16 },
		{ 179, 365, 16 },
		{ 179, 366, 16 },
		{ 179, 367, 16 },
		{ 179, 368, 16 },
		{ 179, 369, 16 },
		{ 179, 370, 16 },
		{ 179, 371, 16 },
		{ 179, 372, 16 },
		{ 179, 373, 16 },
		{ 179, 374, 16 },
		{ 179, 375, 16 },
		{ 179, 376, 16 },
		{ 179, 377, 16 },
		{ 179, 378, 16 },
		{ 179, 379, 16 },
		{ 179, 380, 16 },
		{ 179, 381, 16 },
		{ 179, 382, 16 },
		{ 179, 383, 16 },
		{ 180, 360, 16 },
		{ 180, 361, 16 },
		{ 180, 362, 16 },
		{ 180, 363, 16 },
		{ 180, 364, 16 },
		{ 180, 365, 16 },
		{ 180, 366, 16 },
		{ 180, 367, 16 },
		{ 180, 368, 16 },
		{ 180, 369, 16 },
		{ 180, 370, 16 },
		{ 180, 371, 16 },
		{ 180, 372, 16 },
		{ 180, 373, 16 },
		{ 180, 374, 16 },
		{ 180, 375, 16 },
		{ 180, 376, 16 },
		{ 180, 377, 16 },
		{ 180, 378, 16 },
		{ 180, 379, 16 },
		{ 180, 380, 16 },
		{ 180, 381, 16 },
		{ 180, 382, 16 },
		{ 180, 383, 16 },
		{ 181, 359, 16 },
		{ 181, 360, 16 },
		{ 181, 361, 16 },
		{ 181, 362, 16 },
		{ 181, 363, 16 },
		{ 181, 364, 16 },
		{ 181, 365, 16 },
		{ 181, 366, 16 },
		{ 181, 367, 16 },
		{ 181, 368, 16 },
		{ 181, 369, 16 },
		{ 181, 370, 16 },
		{ 181, 371, 16 },
		{ 181, 372, 16 },
		{ 181, 373, 16 },
		{ 181, 374, 16 },
		{ 181, 375, 16 },
		{ 181, 376, 16 },
		{ 181, 377, 16 },
		{ 181, 378, 16 },
		{ 181, 379, 16 },
		{ 181, 380, 16 },
		{ 181, 381, 16 },
		{ 181, 382, 16 },
		{ 181, 383, 16 },
		{ 182, 358, 16 },
		{ 182, 359, 16 },
		{ 182, 360, 16 },
		{ 182, 361, 16 },
		{ 182, 362, 16 },
		{ 182, 363, 16 },
		{ 182, 364, 16 },
		{ 182, 365, 16 },
		{ 182, 366, 16 },
		{ 182, 367, 16 },
		{ 182, 368, 16 },
		{ 182, 369, 16 },
		{ 182, 370, 16 },
		{ 182, 371, 16 },
		{ 182, 372, 16 },
		{ 182, 373, 16 },
		{ 182, 374, 16 },
		{ 182, 375, 16 },
		{ 182, 376, 16 },
		{ 182, 377, 16 },
		{ 182, 378, 16 },
		{ 182, 379, 16 },
		{ 182, 380, 16 },
		{ 182, 381, 16 },
		{ 182, 382, 16 },
		{ 182, 383, 16 },
		{ 183, 356, 16 },
		{ 183, 357, 16 },
		{ 183, 358, 16 },
		{ 183, 359, 16 },
		{ 183, 360, 16 },
		{ 183, 361, 16 },
		{ 183, 362, 16 },
		{ 183, 363, 16 },
		{ 183, 364, 16 },
		{ 183, 365, 16 },
		{ 183, 366, 16 },
		{ 183, 367, 16 },
		{ 183, 368, 16 },
		{ 183, 369, 16 },
		{ 183, 370, 16 },
		{ 183, 371, 16 },
		{ 183, 372, 16 },
		{ 183, 373, 16 },
		{ 183, 374, 16 },
		{ 183, 375, 16 },
		{ 183, 376, 16 },
		{ 183, 377, 16 },
		{ 183, 378, 16 },
		{ 183, 379, 16 },
		{ 183, 380, 16 },
		{ 183, 381, 16 },
		{ 183, 382, 16 },
		{ 183, 383, 16 },
		{ 184, 355, 16 },
		{ 184, 356, 16 },
		{ 184, 357, 16 },
		{ 184, 358, 16 },
		{ 184, 359, 16 },
		{ 184, 360, 16 },
		{ 184, 361, 16 },
		{ 184, 362, 16 },
		{ 184, 363, 16 },
		{ 184, 364, 16 },
		{ 184, 365, 16 },
		{ 184, 366, 16 },
		{ 184, 367, 16 },
		{ 184, 368, 16 },
		{ 184, 369, 16 },
		{ 184, 370, 16 },
		{ 184, 371, 16 },
		{ 184, 372, 16 },
		{ 184, 373, 16 },
		{ 184, 374, 16 },
		{ 184, 375, 16 },
		{ 184, 376, 16 },
		{ 184, 377, 16 },
		{ 184, 378, 16 },
		{ 184, 379, 16 },
		{ 184, 380, 16 },
		{ 184, 381, 16 },
		{ 184, 382, 16 },
		{ 184, 383, 16 },
		{ 185, 354, 16 },
		{ 185, 355, 16 },
		{ 185, 356, 16 },
		{ 185, 357, 16 },
		{ 185, 358, 16 },
		{ 185, 359, 16 },
		{ 185, 360, 16 },
		{ 185, 361, 16 },
		{ 185, 362, 16 },
		{ 185, 363, 16 },
		{ 185, 364, 16 },
		{ 185, 365, 16 },
		{ 185, 366, 16 },
		{ 185, 367, 16 },
		{ 185, 368, 16 },
		{ 185, 369, 16 },
		{ 185, 370, 16 },
		{ 185, 371, 16 },
		{ 185, 372, 16 },
		{ 185, 373, 16 },
		{ 185, 374, 16 },
		{ 185, 375, 16 },
		{ 185, 376, 16 },
		{ 185, 377, 16 },
		{ 185, 378, 16 },
		{ 185, 379, 16 },
		{ 185, 380, 16 },
		{ 185, 381, 16 },
		{ 185, 382, 16 },
		{ 185, 383, 16 },
		{ 186, 352, 16 },
		{ 186, 353, 16 },
		{ 186, 354, 16 },
		{ 186, 355, 16 },
		{ 186, 356, 16 },
		{ 186, 357, 16 },
		{ 186, 358, 16 },
		{ 186, 359, 16 },
		{ 186, 360, 16 },
		{ 186, 361, 16 },
		{ 186, 362, 16 },
		{ 186, 363, 16 },
		{ 186, 364, 16 },
		{ 186, 365, 16 },
		{ 186, 366, 16 },
		{ 186, 367, 16 },
		{ 186, 368, 16 },
		{ 186, 369, 16 },
		{ 186, 370, 16 },
		{ 186, 371, 16 },
		{ 186, 372, 16 },
		{ 186, 373, 16 },
		{ 186, 374, 16 },
		{ 186, 375, 16 },
		{ 186, 376, 16 },
		{ 186, 377, 16 },
		{ 186, 378, 16 },
		{ 186, 379, 16 },
		{ 186, 380, 16 },
		{ 186, 381, 16 },
		{ 186, 382, 16 },
		{ 186, 383, 16 },
		{ 187, 351, 16 },
		{ 187, 352, 16 },
		{ 187, 353, 16 },
		{ 187, 354, 16 },
		{ 187, 355, 16 },
		{ 187, 356, 16 },
		{ 187, 357, 16 },
		{ 187, 358, 16 },
		{ 187, 359, 16 },
		{ 187, 360, 16 },
		{ 187, 361, 16 },
		{ 187, 362, 16 },
		{ 187, 363, 16 },
		{ 187, 364, 16 },
		{ 187, 365, 16 },
		{ 187, 366, 16 },
		{ 187, 367, 16 },
		{ 187, 368, 16 },
		{ 187, 369, 16 },
		{ 187, 370, 16 },
		{ 187, 371, 16 },
		{ 187, 372, 16 },
		{ 187, 373, 16 },
		{ 187, 374, 16 },
		{ 187, 375, 16 },
		{ 187, 376, 16 },
		{ 187, 377, 16 },
		{ 187, 378, 16 },
		{ 187, 379, 16 },
		{ 187, 380, 16 },
		{ 187, 381, 16 },
		{ 187, 382, 16 },
		{ 187, 383, 16 },
		{ 188, 350, 16 },
		{ 188, 351, 16 },
		{ 188, 352, 16 },
		{ 188, 353, 16 },
		{ 188, 354, 16 },
		{ 188, 355, 16 },
		{ 188, 356, 16 },
		{ 188, 357, 16 },
		{ 188, 358, 16 },
		{ 188, 359, 16 },
		{ 188, 360, 16 },
		{ 188, 361, 16 },
		{ 188, 362, 16 },
		{ 188, 363, 16 },
		{ 188, 364, 16 },
		{ 188, 365, 16 },
		{ 188, 366, 16 },
		{ 188, 367, 16 },
		{ 188, 368, 16 },
		{ 188, 369, 16 },
		{ 188, 370, 16 },
		{ 188, 371, 16 },
		{ 188, 372, 16 },
		{ 188, 373, 16 },
		{ 188, 374, 16 },
		{ 188, 375, 16 },
		{ 188, 376, 16 },
		{ 188, 377, 16 },
		{ 188, 378, 16 },
		{ 188, 379, 16 },
		{ 188, 380, 16 },
		{ 188, 381, 16 },
		{ 188, 382, 16 },
		{ 188, 383, 16 },
		{ 189, 348, 16 },
		{ 189, 349, 16 },
		{ 189, 350, 16 },
		{ 189, 351, 16 },
		{ 189, 352, 16 },
		{ 189, 353, 16 },
		{ 189, 354, 16 },
		{ 189, 355, 16 },
		{ 189, 356, 16 },
		{ 189, 357, 16 },
		{ 189, 358, 16 },
		{ 189, 359, 16 },
		{ 189, 360, 16 },
		{ 189, 361, 16 },
		{ 189, 362, 16 },
		{ 189, 363, 16 },
		{ 189, 364, 16 },
		{ 189, 365, 16 },
		{ 189, 366, 16 },
		{ 189, 367, 16 },
		{ 189, 368, 16 },
		{ 189, 369, 16 },
		{ 189, 370, 16 },
		{ 189, 371, 16 },
		{ 189, 372, 16 },
		{ 189, 373, 16 },
		{ 189, 374, 16 },
		{ 189, 375, 16 },
		{ 189, 376, 16 },
		{ 189, 377, 16 },
		{ 189, 378, 16 },
		{ 189, 379, 16 },
		{ 189, 380, 16 },
		{ 189, 381, 16 },
		{ 189, 382, 16 },
		{ 189, 383, 16 },
		{ 190, 347, 16 },
		{ 190, 348, 16 },
		{ 190, 349, 16 },
		{ 190, 350, 16 },
		{ 190, 351, 16 },
		{ 190, 352, 16 },
		{ 190, 353, 16 },
		{ 190, 354, 16 },
		{ 190, 355, 16 },
		{ 190, 356, 16 },
		{ 190, 357, 16 },
		{ 190, 358, 16 },
		{ 190, 359, 16 },
		{ 190, 360, 16 },
		{ 190, 361, 16 },
		{ 190, 362, 16 },
		{ 190, 363, 16 },
		{ 190, 364, 16 },
		{ 190, 365, 16 },
		{ 190, 366, 16 },
		{ 190, 367, 16 },
		{ 190, 368, 16 },
		{ 190, 369, 16 },
		{ 190, 370, 16 },
		{ 190, 371, 16 },
		{ 190, 372, 16 },
		{ 190, 373, 16 },
		{ 190, 374, 16 },
		{ 190, 375, 16 },
		{ 190, 376, 16 },
		{ 190, 377, 16 },
		{ 190, 378, 16 },
		{ 190, 379, 16 },
		{ 190, 380, 16 },
		{ 190, 381, 16 },
		{ 190, 382, 16 },
		{ 190, 383, 16 },
		{ 191, 346, 16 },
		{ 191, 347, 16 },
		{ 191, 348, 16 },
		{ 191, 349, 16 },
		{ 191, 350, 16 },
		{ 191, 351, 16 },
		{ 191, 352, 16 },
		{ 191, 353, 16 },
		{ 191, 354, 16 },
		{ 191, 355, 16 },
		{ 191, 356, 16 },
		{ 191, 357, 16 },
		{ 191, 358, 16 },
		{ 191, 359, 16 },
		{ 191, 360, 16 },
		{ 191, 361, 16 },
		{ 191, 362, 16 },
		{ 191, 363, 16 },
		{ 191, 364, 16 },
		{ 191, 365, 16 },
		{ 191, 366, 16 },
		{ 191, 367, 16 },
		{ 191, 368, 16 },
		{ 191, 369, 16 },
		{ 191, 370, 16 },
		{ 191, 371, 16 },
		{ 191, 372, 16 },
		{ 191, 373, 16 },
		{ 191, 374, 16 },
		{ 191, 375, 16 },
		{ 191, 376, 16 },
		{ 191, 377, 16 },
		{ 191, 378, 16 },
		{ 191, 379, 16 },
		{ 191, 380, 16 },
		{ 191, 381, 16 },
		{ 191, 382, 16 },
		{ 191, 383, 16 },
		{ 192, 344, 16 },
		{ 192, 345, 16 },
		{ 192, 346, 16 },
		{ 192, 347, 16 },
		{ 192, 348, 16 },
		{ 192, 349, 16 },
		{ 192, 350, 16 },
		{ 192, 351, 16 },
		{ 192, 352, 16 },
		{ 192, 353, 16 },
		{ 192, 354, 16 },
		{ 192, 355, 16 },
		{ 192, 356, 16 },
		{ 192, 357, 16 },
		{ 192, 358, 16 },
		{ 192, 359, 16 },
		{ 192, 360, 16 },
		{ 192, 361, 16 },
		{ 192, 362, 16 },
		{ 192, 363, 16 },
		{ 192, 364, 16 },
		{ 192, 365, 16 },
		{ 192, 366, 16 },
		{ 192, 367, 16 },
		{ 192, 368, 16 },
		{ 192, 369, 16 },
		{ 192, 370, 16 },
		{ 192, 371, 16 },
		{ 192, 372, 16 },
		{ 192, 373, 16 },
		{ 192, 374, 16 },
		{ 192, 375, 16 },
		{ 192, 376, 16 },
		{ 192, 377, 16 },
		{ 192, 378, 16 },
		{ 192, 379, 16 },
		{ 192, 380, 16 },
		{ 192, 381, 16 },
		{ 192, 382, 16 },
		{ 192, 383, 16 },
		{ 193, 343, 16 },
		{ 193, 344, 16 },
		{ 193, 345, 16 },
		{ 193, 346, 16 },
		{ 193, 347, 16 },
		{ 193, 348, 16 },
		{ 193, 349, 16 },
		{ 193, 350, 16 },
		{ 193, 351, 16 },
		{ 193, 352, 16 },
		{ 193, 353, 16 },
		{ 193, 354, 16 },
		{ 193, 355, 16 },
		{ 193, 356, 16 },
		{ 193, 357, 16 },
		{ 193, 358, 16 },
		{ 193, 359, 16 },
		{ 193, 360, 16 },
		{ 193, 361, 16 },
		{ 193, 362, 16 },
		{ 193, 363, 16 },
		{ 193, 364, 16 },
		{ 193, 365, 16 },
		{ 193, 366, 16 },
		{ 193, 367, 16 },
		{ 193, 368, 16 },
		{ 193, 369, 16 },
		{ 193, 370, 16 },
		{ 193, 371, 16 },
		{ 193, 372, 16 },
		{ 193, 373, 16 },
		{ 193, 374, 16 },
		{ 193, 375, 16 },
		{ 193, 376, 16 },
		{ 193, 377, 16 },
		{ 193, 378, 16 },
		{ 193, 379, 16 },
		{ 193, 380, 16 },
		{ 193, 381, 16 },
		{ 193, 382, 16 },
		{ 193, 383, 16 },
		{ 194, 342, 16 },
		{ 194, 343, 16 },
		{ 194, 344, 16 },
		{ 194, 345, 16 },
		{ 194, 346, 16 },
		{ 194, 347, 16 },
		{ 194, 348, 16 },
		{ 194, 349, 16 },
		{ 194, 350, 16 },
		{ 194, 351, 16 },
		{ 194, 352, 16 },
		{ 194, 353, 16 },
		{ 194, 354, 16 },
		{ 194, 355, 16 },
		{ 194, 356, 16 },
		{ 194, 357, 16 },
		{ 194, 358, 16 },
		{ 194, 359, 16 },
		{ 194, 360, 16 },
		{ 194, 361, 16 },
		{ 194, 362, 16 },
		{ 194, 363, 16 },
		{ 194, 364, 16 },
		{ 194, 365, 16 },
		{ 194, 366, 16 },
		{ 194, 367, 16 },
		{ 194, 368, 16 },
		{ 194, 369, 16 },
		{ 194, 370, 16 },
		{ 194, 371, 16 },
		{ 194, 372, 16 },
		{ 194, 373, 16 },
		{ 194, 374, 16 },
		{ 194, 375, 16 },
		{ 194, 376, 16 },
		{ 194, 377, 16 },
		{ 194, 378, 16 },
		{ 194, 379, 16 },
		{ 194, 380, 16 },
		{ 194, 381, 16 },
		{ 194, 382, 16 },
		{ 194, 383, 16 },
		{ 195, 340, 16 },
		{ 195, 341, 16 },
		{ 195, 342, 16 },
		{ 195, 343, 16 },
		{ 195, 344, 16 },
		{ 195, 345, 16 },
		{ 195, 346, 16 },
		{ 195, 347, 16 },
		{ 195, 348, 16 },
		{ 195, 349, 16 },
		{ 195, 350, 16 },
		{ 195, 351, 16 },
		{ 195, 352, 16 },
		{ 195, 353, 16 },
		{ 195, 354, 16 },
		{ 195, 355, 16 },
		{ 195, 356, 16 },
		{ 195, 357, 16 },
		{ 195, 358, 16 },
		{ 195, 359, 16 },
		{ 195, 360, 16 },
		{ 195, 361, 16 },
		{ 195, 362, 16 },
		{ 195, 363, 16 },
		{ 195, 364, 16 },
		{ 195, 365, 16 },
		{ 195, 366, 16 },
		{ 195, 367, 16 },
		{ 195, 368, 16 },
		{ 195, 369, 16 },
		{ 195, 370, 16 },
		{ 195, 371, 16 },
		{ 195, 372, 16 },
		{ 195, 373, 16 },
		{ 195, 374, 16 },
		{ 195, 375, 16 },
		{ 195, 376, 16 },
		{ 195, 377, 16 },
		{ 195, 378, 16 },
		{ 195, 379, 16 },
		{ 195, 380, 16 },
		{ 195, 381, 16 },
		{ 195, 382, 16 },
		{ 195, 383, 16 },
		{ 196, 339, 16 },
		{ 196, 340, 16 },
		{ 196, 341, 16 },
		{ 196, 342, 16 },
		{ 196, 343, 16 },
		{ 196, 344, 16 },
		{ 196, 345, 16 },
		{ 196, 346, 16 },
		{ 196, 347, 16 },
		{ 196, 348, 16 },
		{ 196, 349, 16 },
		{ 196, 350, 16 },
		{ 196, 351, 16 },
		{ 196, 352, 16 },
		{ 196, 353, 16 },
		{ 196, 354, 16 },
		{ 196, 355, 16 },
		{ 196, 356, 16 },
		{ 196, 357, 16 },
		{ 196, 358, 16 },
		{ 196, 359, 16 },
		{ 196, 360, 16 },
		{ 196, 361, 16 },
		{ 196, 362, 16 },
		{ 196, 363, 16 },
		{ 196, 364, 16 },
		{ 196, 365, 16 },
		{ 196, 366, 16 },
		{ 196, 367, 16 },
		{ 196, 368, 16 },
		{ 196, 369, 16 },
		{ 196, 370, 16 },
		{ 196, 371, 16 },
		{ 196, 372, 16 },
		{ 196, 373, 16 },
		{ 196, 374, 16 },
		{ 196, 375, 16 },
		{ 196, 376, 16 },
		{ 196, 377, 16 },
		{ 196, 378, 16 },
		{ 196, 379, 16 },
		{ 196, 380, 16 },
		{ 196, 381, 16 },
		{ 196, 382, 16 },
		{ 196, 383, 16 },
		{ 197, 338, 16 },
		{ 197, 339, 16 },
		{ 197, 340, 16 },
		{ 197, 341, 16 },
		{ 197, 342, 16 },
		{ 197, 343, 16 },
		{ 197, 344, 16 },
		{ 197, 345, 16 },
		{ 197, 346, 16 },
		{ 197, 347, 16 },
		{ 197, 348, 16 },
		{ 197, 349, 16 },
		{ 197, 350, 16 },
		{ 197, 351, 16 },
		{ 197, 352, 16 },
		{ 197, 353, 16 },
		{ 197, 354, 16 },
		{ 197, 355, 16 },
		{ 197, 356, 16 },
		{ 197, 357, 16 },
		{ 197, 358, 16 },
		{ 197, 359, 16 },
		{ 197, 360, 16 },
		{ 197, 361, 16 },
		{ 197, 362, 16 },
		{ 197, 363, 16 },
		{ 197, 364, 16 },
		{ 197, 365, 16 },
		{ 197, 366, 16 },
		{ 197, 367, 16 },
		{ 197, 368, 16 },
		{ 197, 369, 16 },
		{ 197, 370, 16 },
		{ 197, 371, 16 },
		{ 197, 372, 16 },
		{ 197, 373, 16 },
		{ 197, 374, 16 },
		{ 197, 375, 16 },
		{ 197, 376, 16 },
		{ 197, 377, 16 },
		{ 197, 378, 16 },
		{ 197, 379, 16 },
		{ 197, 380, 16 },
		{ 197, 381, 16 },
		{ 197, 382, 16 },
		{ 197, 383, 16 },
		{ 198, 336, 16 },
		{ 198, 337, 16 },
		{ 198, 338, 16 },
		{ 198, 339, 16 },
		{ 198, 340, 16 },
		{ 198, 341, 16 },
		{ 198, 342, 16 },
		{ 198, 343, 16 },
		{ 198, 344, 16 },
		{ 198, 345, 16 },
		{ 198, 346, 16 },
		{ 198, 347, 16 },
		{ 198, 348, 16 },
		{ 198, 349, 16 },
		{ 198, 350, 16 },
		{ 198, 351, 16 },
		{ 198, 352, 16 },
		{ 198, 353, 16 },
		{ 198, 354, 16 },
		{ 198, 355, 16 },
		{ 198, 356, 16 },
		{ 198, 357, 16 },
		{ 198, 358, 16 },
		{ 198, 359, 16 },
		{ 198, 360, 16 },
		{ 198, 361, 16 },
		{ 198, 362, 16 },
		{ 198, 363, 16 },
		{ 198, 364, 16 },
		{ 198, 365, 16 },
		{ 198, 366, 16 },
		{ 198, 367, 16 },
		{ 198, 368, 16 },
		{ 198, 369, 16 },
		{ 198, 370, 16 },
		{ 198, 371, 16 },
		{ 198, 372, 16 },
		{ 198, 373, 16 },
		{ 198, 374, 16 },
		{ 198, 375, 16 },
		{ 198, 376, 16 },
		{ 198, 377, 16 },
		{ 198, 378, 16 },
		{ 198, 379, 16 },
		{ 198, 380, 16 },
		{ 198, 381, 16 },
		{ 198, 382, 16 },
		{ 198, 383, 16 },
		{ 199, 335, 16 },
		{ 199, 336, 16 },
		{ 199, 337, 16 },
		{ 199, 338, 16 },
		{ 199, 339, 16 },
		{ 199, 340, 16 },
		{ 199, 341, 16 },
		{ 199, 342, 16 },
		{ 199, 343, 16 },
		{ 199, 344, 16 },
		{ 199, 345, 16 },
		{ 199, 346, 16 },
		{ 199, 347, 16 },
		{ 199, 348, 16 },
		{ 199, 349, 16 },
		{ 199, 350, 16 },
		{ 199, 351, 16 },
		{ 199, 352, 16 },
		{ 199, 353, 16 },
		{ 199, 354, 16 },
		{ 199, 355, 16 },
		{ 199, 356, 16 },
		{ 199, 357, 16 },
		{ 199, 358, 16 },
		{ 199, 359, 16 },
		{ 199, 360, 16 },
		{ 199, 361, 16 },
		{ 199, 362, 16 },
		{ 199, 363, 16 },
		{ 199, 364, 16 },
		{ 199, 365, 16 },
		{ 199, 366, 16 },
		{ 199, 367, 16 },
		{ 199, 368, 16 },
		{ 199, 369, 16 },
		{ 199, 370, 16 },
		{ 199, 371, 16 },
		{ 199, 372, 16 },
		{ 199, 373, 16 },
		{ 199, 374, 16 },
		{ 199, 375, 16 },
		{ 199, 376, 16 },
		{ 199, 377, 16 },
		{ 199, 378, 16 },
		{ 199, 379, 16 },
		{ 199, 380, 16 },
		{ 199, 381, 16 },
		{ 199, 382, 16 },
		{ 199, 383, 16 },
		{ 200, 334, 16 },
		{ 200, 335, 16 },
		{ 200, 336, 16 },
		{ 200, 337, 16 },
		{ 200, 338, 16 },
		{ 200, 339, 16 },
		{ 200, 340, 16 },
		{ 200, 341, 16 },
		{ 200, 342, 16 },
		{ 200, 343, 16 },
		{ 200, 344, 16 },
		{ 200, 345, 16 },
		{ 200, 346, 16 },
		{ 200, 347, 16 },
		{ 200, 348, 16 },
		{ 200, 349, 16 },
		{ 200, 350, 16 },
		{ 200, 351, 16 },
		{ 200, 352, 16 },
		{ 200, 353, 16 },
		{ 200, 354, 16 },
		{ 200, 355, 16 },
		{ 200, 356, 16 },
		{ 200, 357, 16 },
		{ 200, 358, 16 },
		{ 200, 359, 16 },
		{ 200, 360, 16 },
		{ 200, 361, 16 },
		{ 200, 362, 16 },
		{ 200, 363, 16 },
		{ 200, 364, 16 },
		{ 200, 365, 16 },
		{ 200, 366, 16 },
		{ 200, 367, 16 },
		{ 200, 368, 16 },
		{ 200, 369, 16 },
		{ 200, 370, 16 },
		{ 200, 371, 16 },
		{ 200, 372, 16 },
		{ 200, 373, 16 },
		{ 200, 374, 16 },
		{ 200, 375, 16 },
		{ 200, 376, 16 },
		{ 200, 377, 16 },
		{ 200, 378, 16 },
		{ 200, 379, 16 },
		{ 200, 380, 16 },
		{ 200, 381, 16 },
		{ 200, 382, 16 },
		{ 200, 383, 16 },
		{ 201, 332, 16 },
		{ 201, 333, 16 },
		{ 201, 334, 16 },
		{ 201, 335, 16 },
		{ 201, 336, 16 },
		{ 201, 337, 16 },
		{ 201, 338, 16 },
		{ 201, 339, 16 },
		{ 201, 340, 16 },
		{ 201, 341, 16 },
		{ 201, 342, 16 },
		{ 201, 343, 16 },
		{ 201, 344, 16 },
		{ 201, 345, 16 },
		{ 201, 346, 16 },
		{ 201, 347, 16 },
		{ 201, 348, 16 },
		{ 201, 349, 16 },
		{ 201, 350, 16 },
		{ 201, 351, 16 },
		{ 201, 352, 16 },
		{ 201, 353, 16 },
		{ 201, 354, 16 },
		{ 201, 355, 16 },
		{ 201, 356, 16 },
		{ 201, 357, 16 },
		{ 201, 358, 16 },
		{ 201, 359, 16 },
		{ 201, 360, 16 },
		{ 201, 361, 16 },
		{ 201, 362, 16 },
		{ 201, 363, 16 },
		{ 201, 364, 16 },
		{ 201, 365, 16 },
		{ 201, 366, 16 },
		{ 201, 367, 16 },
		{ 201, 368, 16 },
		{ 201, 369, 16 },
		{ 201, 370, 16 },
		{ 201, 371, 16 },
		{ 201, 372, 16 },
		{ 201, 373, 16 },
		{ 201, 374, 16 },
		{ 201, 375, 16 },
		{ 201, 376, 16 },
		{ 201, 377, 16 },
		{ 201, 378, 16 },
		{ 201, 379, 16 },
		{ 201, 380, 16 },
		{ 201, 381, 16 },
		{ 201, 382, 16 },
		{ 201, 383, 16 },
		{ 202, 331, 16 },
		{ 202, 332, 16 },
		{ 202, 333, 16 },
		{ 202, 334, 16 },
		{ 202, 335, 16 },
		{ 202, 336, 16 },
		{ 202, 337, 16 },
		{ 202, 338, 16 },
		{ 202, 339, 16 },
		{ 202, 340, 16 },
		{ 202, 341, 16 },
		{ 202, 342, 16 },
		{ 202, 343, 16 },
		{ 202, 344, 16 },
		{ 202, 345, 16 },
		{ 202, 346, 16 },
		{ 202, 347, 16 },
		{ 202, 348, 16 },
		{ 202, 349, 16 },
		{ 202, 350, 16 },
		{ 202, 351, 16 },
		{ 202, 352, 16 },
		{ 202, 353, 16 },
		{ 202, 354, 16 },
		{ 202, 355, 16 },
		{ 202, 356, 16 },
		{ 202, 357, 16 },
		{ 202, 358, 16 },
		{ 202, 359, 16 },
		{ 202, 360, 16 },
		{ 202, 361, 16 },
		{ 202, 362, 16 },
		{ 202, 363, 16 },
		{ 202, 364, 16 },
		{ 202, 365, 16 },
		{ 202, 366, 16 },
		{ 202, 367, 16 },
		{ 202, 368, 16 },
		{ 202, 369, 16 },
		{ 202, 370, 16 },
		{ 202, 371, 16 },
		{ 202, 372, 16 },
		{ 202, 373, 16 },
		{ 202, 374, 16 },
		{ 202, 375, 16 },
		{ 202, 376, 16 },
		{ 202, 377, 16 },
		{ 202, 378, 16 },
		{ 202, 379, 16 },
		{ 202, 380, 16 },
		{ 202, 381, 16 },
		{ 202, 382, 16 },
		{ 202, 383, 16 },
		{ 203, 330, 16 },
		{ 203, 331, 16 },
		{ 203, 332, 16 },
		{ 203, 333, 16 },
		{ 203, 334, 16 },
		{ 203, 335, 16 },
		{ 203, 336, 16 },
		{ 203, 337, 16 },
		{ 203, 338, 16 },
		{ 203, 339, 16 },
		{ 203, 340, 16 },
		{ 203, 341, 16 },
		{ 203, 342, 16 },
		{ 203, 343, 16 },
		{ 203, 344, 16 },
		{ 203, 345, 16 },
		{ 203, 346, 16 },
		{ 203, 347, 16 },
		{ 203, 348, 16 },
		{ 203, 349, 16 },
		{ 203, 350, 16 },
		{ 203, 351, 16 },
		{ 203, 352, 16 },
		{ 203, 353, 16 },
		{ 203, 354, 16 },
		{ 203, 355, 16 },
		{ 203, 356, 16 },
		{ 203, 357, 16 },
		{ 203, 358, 16 },
		{ 203, 359, 16 },
		{ 203, 360, 16 },
		{ 203, 361, 16 },
		{ 203, 362, 16 },
		{ 203, 363, 16 },
		{ 203, 364, 16 },
		{ 203, 365, 16 },
		{ 203, 366, 16 },
		{ 203, 367, 16 },
		{ 203, 368, 16 },
		{ 203, 369, 16 },
		{ 203, 370, 16 },
		{ 203, 371, 16 },
		{ 203, 372, 16 },
		{ 203, 373, 16 },
		{ 203, 374, 16 },
		{ 203, 375, 16 },
		{ 203, 376, 16 },
		{ 203, 377, 16 },
		{ 203, 378, 16 },
		{ 203, 379, 16 },
		{ 203, 380, 16 },
		{ 203, 381, 16 },
		{ 203, 382, 16 },
		{ 203, 383, 16 },
		{ 204, 328, 16 },
		{ 204, 329, 16 },
		{ 204, 330, 16 },
		{ 204, 331, 16 },
		{ 204, 332, 16 },
		{ 204, 333, 16 },
		{ 204, 334, 16 },
		{ 204, 335, 16 },
		{ 204, 336, 16 },
		{ 204, 337, 16 },
		{ 204, 338, 16 },
		{ 204, 339, 16 },
		{ 204, 340, 16 },
		{ 204, 341, 16 },
		{ 204, 342, 16 },
		{ 204, 343, 16 },
		{ 204, 344, 16 },
		{ 204, 345, 16 },
		{ 204, 346, 16 },
		{ 204, 347, 16 },
		{ 204, 348, 16 },
		{ 204, 349, 16 },
		{ 204, 350, 16 },
		{ 204, 351, 16 },
		{ 204, 352, 16 },
		{ 204, 353, 16 },
		{ 204, 354, 16 },
		{ 204, 355, 16 },
		{ 204, 356, 16 },
		{ 204, 357, 16 },
		{ 204, 358, 16 },
		{ 204, 359, 16 },
		{ 204, 360, 16 },
		{ 204, 361, 16 },
		{ 204, 362, 16 },
		{ 204, 363, 16 },
		{ 204, 364, 16 },
		{ 204, 365, 16 },
		{ 204, 366, 16 },
		{ 204, 367, 16 },
		{ 204, 368, 16 },
		{ 204, 369, 16 },
		{ 204, 370, 16 },
		{ 204, 371, 16 },
		{ 204, 372, 16 },
		{ 204, 373, 16 },
		{ 204, 374, 16 },
		{ 204, 375, 16 },
		{ 204, 376, 16 },
		{ 204, 377, 16 },
		{ 204, 378, 16 },
		{ 204, 379, 16 },
		{ 204, 380, 16 },
		{ 204, 381, 16 },
		{ 204, 382, 16 },
		{ 204, 383, 16 },
		{ 205, 327, 16 },
		{ 205, 328, 16 },
		{ 205, 329, 16 },
		{ 205, 330, 16 },
		{ 205, 331, 16 },
		{ 205, 332, 16 },
		{ 205, 333, 16 },
		{ 205, 334, 16 },
		{ 205, 335, 16 },
		{ 205, 336, 16 },
		{ 205, 337, 16 },
		{ 205, 338, 16 },
		{ 205, 339, 16 },
		{ 205, 340, 16 },
		{ 205, 341, 16 },
		{ 205, 342, 16 },
		{ 205, 343, 16 },
		{ 205, 344, 16 },
		{ 205, 345, 16 },
		{ 205, 346, 16 },
		{ 205, 347, 16 },
		{ 205, 348, 16 },
		{ 205, 349, 16 },
		{ 205, 350, 16 },
		{ 205, 351, 16 },
		{ 205, 352, 16 },
		{ 205, 353, 16 },
		{ 205, 354, 16 },
		{ 205, 355, 16 },
		{ 205, 356, 16 },
		{ 205, 357, 16 },
		{ 205, 358, 16 },
		{ 205, 359, 16 },
		{ 205, 360, 16 },
		{ 205, 361, 16 },
		{ 205, 362, 16 },
		{ 205, 363, 16 },
		{ 205, 364, 16 },
		{ 205, 365, 16 },
		{ 205, 366, 16 },
		{ 205, 367, 16 },
		{ 205, 368, 16 },
		{ 205, 369, 16 },
		{ 205, 370, 16 },
		{ 205, 371, 16 },
		{ 205, 372, 16 },
		{ 205, 373, 16 },
		{ 205, 374, 16 },
		{ 205, 375, 16 },
		{ 205, 376, 16 },
		{ 205, 377, 16 },
		{ 205, 378, 16 },
		{ 205, 379, 16 },
		{ 205, 380, 16 },
		{ 205, 381, 16 },
		{ 205, 382, 16 },
		{ 205, 383, 16 },
		{ 206, 326, 16 },
		{ 206, 327, 16 },
		{ 206, 328, 16 },
		{ 206, 329, 16 },
		{ 206, 330, 16 },
		{ 206, 331, 16 },
		{ 206, 332, 16 },
		{ 206, 333, 16 },
		{ 206, 334, 16 },
		{ 206, 335, 16 },
		{ 206, 336, 16 },
		{ 206, 337, 16 },
		{ 206, 338, 16 },
		{ 206, 339, 16 },
		{ 206, 340, 16 },
		{ 206, 341, 16 },
		{ 206, 342, 16 },
		{ 206, 343, 16 },
		{ 206, 344, 16 },
		{ 206, 345, 16 },
		{ 206, 346, 16 },
		{ 206, 347, 16 },
		{ 206, 348, 16 },
		{ 206, 349, 16 },
		{ 206, 350, 16 },
		{ 206, 351, 16 },
		{ 206, 352, 16 },
		{ 206, 353, 16 },
		{ 206, 354, 16 },
		{ 206, 355, 16 },
		{ 206, 356, 16 },
		{ 206, 357, 16 },
		{ 206, 358, 16 },
		{ 206, 359, 16 },
		{ 206, 360, 16 },
		{ 206, 361, 16 },
		{ 206, 362, 16 },
		{ 206, 363, 16 },
		{ 206, 364, 16 },
		{ 206, 365, 16 },
		{ 206, 366, 16 },
		{ 206, 367, 16 },
		{ 206, 368, 16 },
		{ 206, 369, 16 },
		{ 206, 370, 16 },
		{ 206, 371, 16 },
		{ 206, 372, 16 },
		{ 206, 373, 16 },
		{ 206, 374, 16 },
		{ 206, 375, 16 },
		{ 206, 376, 16 },
		{ 206, 377, 16 },
		{ 206, 378, 16 },
		{ 206, 379, 16 },
		{ 206, 380, 16 },
		{ 206, 381, 16 },
		{ 206, 382, 16 },
		{ 206, 383, 16 },
		{ 207, 324, 16 },
		{ 207, 325, 16 },
		{ 207, 326, 16 },
		{ 207, 327, 16 },
		{ 207, 328, 16 },
		{ 207, 329, 16 },
		{ 207, 330, 16 },
		{ 207, 331, 16 },
		{ 207, 332, 16 },
		{ 207, 333, 16 },
		{ 207, 334, 16 },
		{ 207, 335, 16 },
		{ 207, 336, 16 },
		{ 207, 337, 16 },
		{ 207, 338, 16 },
		{ 207, 339, 16 },
		{ 207, 340, 16 },
		{ 207, 341, 16 },
		{ 207, 342, 16 },
		{ 207, 343, 16 },
		{ 207, 344, 16 },
		{ 207, 345, 16 },
		{ 207, 346, 16 },
		{ 207, 347, 16 },
		{ 207, 348, 16 },
		{ 207, 349, 16 },
		{ 207, 350, 16 },
		{ 207, 351, 16 },
		{ 207, 352, 16 },
		{ 207, 353, 16 },
		{ 207, 354, 16 },
		{ 207, 355, 16 },
		{ 207, 356, 16 },
		{ 207, 357, 16 },
		{ 207, 358, 16 },
		{ 207, 359, 16 },
		{ 207, 360, 16 },
		{ 207, 361, 16 },
		{ 207, 362, 16 },
		{ 207, 363, 16 },
		{ 207, 364, 16 },
		{ 207, 365, 16 },
		{ 207, 366, 16 },
		{ 207, 367, 16 },
		{ 207, 368, 16 },
		{ 207, 369, 16 },
		{ 207, 370, 16 },
		{ 207, 371, 16 },
		{ 207, 372, 16 },
		{ 207, 373, 16 },
		{ 207, 374, 16 },
		{ 207, 375, 16 },
		{ 207, 376, 16 },
		{ 207, 377, 16 },
		{ 207, 378, 16 },
		{ 207, 379, 16 },
		{ 207, 380, 16 },
		{ 207, 381, 16 },
		{ 207, 382, 16 },
		{ 207, 383, 16 },
		{ 208, 323, 16 },
		{ 208, 324, 16 },
		{ 208, 325, 16 },
		{ 208, 326, 16 },
		{ 208, 327, 16 },
		{ 208, 328, 16 },
		{ 208, 329, 16 },
		{ 208, 330, 16 },
		{ 208, 331, 16 },
		{ 208, 332, 16 },
		{ 208, 333, 16 },
		{ 208, 334, 16 },
		{ 208, 335, 16 },
		{ 208, 336, 16 },
		{ 208, 337, 16 },
		{ 208, 338, 16 },
		{ 208, 339, 16 },
		{ 208, 340, 16 },
		{ 208, 341, 16 },
		{ 208, 342, 16 },
		{ 208, 343, 16 },
		{ 208, 344, 16 },
		{ 208, 345, 16 },
		{ 208, 346, 16 },
		{ 208, 347, 16 },
		{ 208, 348, 16 },
		{ 208, 349, 16 },
		{ 208, 350, 16 },
		{ 208, 351, 16 },
		{ 208, 352, 16 },
		{ 208, 353, 16 },
		{ 208, 354, 16 },
		{ 208, 355, 16 },
		{ 208, 356, 16 },
		{ 208, 357, 16 },
		{ 208, 358, 16 },
		{ 208, 359, 16 },
		{ 208, 360, 16 },
		{ 208, 361, 16 },
		{ 208, 362, 16 },
		{ 208, 363, 16 },
		{ 208, 364, 16 },
		{ 208, 365, 16 },
		{ 208, 366, 16 },
		{ 208, 367, 16 },
		{ 208, 368, 16 },
		{ 208, 369, 16 },
		{ 208, 370, 16 },
		{ 208, 371, 16 },
		{ 208, 372, 16 },
		{ 208, 373, 16 },
		{ 208, 374, 16 },
		{ 208, 375, 16 },
		{ 208, 376, 16 },
		{ 208, 377, 16 },
		{ 208, 378, 16 },
		{ 208, 379, 16 },
		{ 208, 380, 16 },
		{ 208, 381, 16 },
		{ 208, 382, 16 },
		{ 208, 383, 16 },
		{ 209, 322, 16 },
		{ 209, 323, 16 },
		{ 209, 324, 16 },
		{ 209, 325, 16 },
		{ 209, 326, 16 },
		{ 209, 327, 16 },
		{ 209, 328, 16 },
		{ 209, 329, 16 },
		{ 209, 330, 16 },
		{ 209, 331, 16 },
		{ 209, 332, 16 },
		{ 209, 333, 16 },
		{ 209, 334, 16 },
		{ 209, 335, 16 },
		{ 209, 336, 16 },
		{ 209, 337, 16 },
		{ 209, 338, 16 },
		{ 209, 339, 16 },
		{ 209, 340, 16 },
		{ 209, 341, 16 },
		{ 209, 342, 16 },
		{ 209, 343, 16 },
		{ 209, 344, 16 },
		{ 209, 345, 16 },
		{ 209, 346, 16 },
		{ 209, 347, 16 },
		{ 209, 348, 16 },
		{ 209, 349, 16 },
		{ 209, 350, 16 },
		{ 209, 351, 16 },
		{ 209, 352, 16 },
		{ 209, 353, 16 },
		{ 209, 354, 16 },
		{ 209, 355, 16 },
		{ 209, 356, 16 },
		{ 209, 357, 16 },
		{ 209, 358, 16 },
		{ 209, 359, 16 },
		{ 209, 360, 16 },
		{ 209, 361, 16 },
		{ 209, 362, 16 },
		{ 209, 363, 16 },
		{ 209, 364, 16 },
		{ 209, 365, 16 },
		{ 209, 366, 16 },
		{ 209, 367, 16 },
		{ 209, 368, 16 },
		{ 209, 369, 16 },
		{ 209, 370, 16 },
		{ 209, 371, 16 },
		{ 209, 372, 16 },
		{ 209, 373, 16 },
		{ 209, 374, 16 },
		{ 209, 375, 16 },
		{ 209, 376, 16 },
		{ 209, 377, 16 },
		{ 209, 378, 16 },
		{ 209, 379, 16 },
		{ 209, 380, 16 },
		{ 209, 381, 16 },
		{ 209, 382, 16 },
		{ 209, 383, 16 },
		{ 210, 320, 16 },
		{ 210, 321, 16 },
		{ 210, 322, 16 },
		{ 210, 323, 16 },
		{ 210, 324, 16 },
		{ 210, 325, 16 },
		{ 210, 326, 16 },
		{ 210, 327, 16 },
		{ 210, 328, 16 },
		{ 210, 329, 16 },
		{ 210, 330, 16 },
		{ 210, 331, 16 },
		{ 210, 332, 16 },
		{ 210, 333, 16 },
		{ 210, 334, 16 },
		{ 210, 335, 16 },
		{ 210, 336, 16 },
		{ 210, 337, 16 },
		{ 210, 338, 16 },
		{ 210, 339, 16 },
		{ 210, 340, 16 },
		{ 210, 341, 16 },
		{ 210, 342, 16 },
		{ 210, 343, 16 },
		{ 210, 344, 16 },
		{ 210, 345, 16 },
		{ 210, 346, 16 },
		{ 210, 347, 16 },
		{ 210, 348, 16 },
		{ 210, 349, 16 },
		{ 210, 350, 16 },
		{ 210, 351, 16 },
		{ 210, 352, 16 },
		{ 210, 353, 16 },
		{ 210, 354, 16 },
		{ 210, 355, 16 },
		{ 210, 356, 16 },
		{ 210, 357, 16 },
		{ 210, 358, 16 },
		{ 210, 359, 16 },
		{ 210, 360, 16 },
		{ 210, 361, 16 },
		{ 210, 362, 16 },
		{ 210, 363, 16 },
		{ 210, 364, 16 },
		{ 210, 365, 16 },
		{ 210, 366, 16 },
		{ 210, 367, 16 },
		{ 210, 368, 16 },
		{ 210, 369, 16 },
		{ 210, 370, 16 },
		{ 210, 371, 16 },
		{ 210, 372, 16 },
		{ 210, 373, 16 },
		{ 210, 374, 16 },
		{ 210, 375, 16 },
		{ 210, 376, 16 },
		{ 210, 377, 16 },
		{ 210, 378, 16 },
		{ 210, 379, 16 },
		{ 210, 380, 16 },
		{ 210, 381, 16 },
		{ 210, 382, 16 },
		{ 210, 383, 16 },
		{ 211, 319, 16 },
		{ 211, 320, 16 },
		{ 211, 321, 16 },
		{ 211, 322, 16 },
		{ 211, 323, 16 },
		{ 211, 324, 16 },
		{ 211, 325, 16 },
		{ 211, 326, 16 },
		{ 211, 327, 16 },
		{ 211, 328, 16 },
		{ 211, 329, 16 },
		{ 211, 330, 16 },
		{ 211, 331, 16 },
		{ 211, 332, 16 },
		{ 211, 333, 16 },
		{ 211, 334, 16 },
		{ 211, 335, 16 },
		{ 211, 336, 16 },
		{ 211, 337, 16 },
		{ 211, 338, 16 },
		{ 211, 339, 16 },
		{ 211, 340, 16 },
		{ 211, 341, 16 },
		{ 211, 342, 16 },
		{ 211, 343, 16 },
		{ 211, 344, 16 },
		{ 211, 345, 16 },
		{ 211, 346, 16 },
		{ 211, 347, 16 },
		{ 211, 348, 16 },
		{ 211, 349, 16 },
		{ 211, 350, 16 },
		{ 211, 351, 16 },
		{ 211, 352, 16 },
		{ 211, 353, 16 },
		{ 211, 354, 16 },
		{ 211, 355, 16 },
		{ 211, 356, 16 },
		{ 211, 357, 16 },
		{ 211, 358, 16 },
		{ 211, 359, 16 },
		{ 211, 360, 16 },
		{ 211, 361, 16 },
		{ 211, 362, 16 },
		{ 211, 363, 16 },
		{ 211, 364, 16 },
		{ 211, 365, 16 },
		{ 211, 366, 16 },
		{ 211, 367, 16 },
		{ 211, 368, 16 },
		{ 211, 369, 16 },
		{ 211, 370, 16 },
		{ 211, 371, 16 },
		{ 211, 372, 16 },
		{ 211, 373, 16 },
		{ 211, 374, 16 },
		{ 211, 375, 16 },
		{ 211, 376, 16 },
		{ 211, 377, 16 },
		{ 211, 378, 16 },
		{ 211, 379, 16 },
		{ 211, 380, 16 },
		{ 211, 381, 16 },
		{ 211, 382, 16 },
		{ 211, 383, 16 },
		{ 212, 318, 16 },
		{ 212, 319, 16 },
		{ 212, 320, 16 },
		{ 212, 321, 16 },
		{ 212, 322, 16 },
		{ 212, 323, 16 },
		{ 212, 324, 16 },
		{ 212, 325, 16 },
		{ 212, 326, 16 },
		{ 212, 327, 16 },
		{ 212, 328, 16 },
		{ 212, 329, 16 },
		{ 212, 330, 16 },
		{ 212, 331, 16 },
		{ 212, 332, 16 },
		{ 212, 333, 16 },
		{ 212, 334, 16 },
		{ 212, 335, 16 },
		{ 212, 336, 16 },
		{ 212, 337, 16 },
		{ 212, 338, 16 },
		{ 212, 339, 16 },
		{ 212, 340, 16 },
		{ 212, 341, 16 },
		{ 212, 342, 16 },
		{ 212, 343, 16 },
		{ 212, 344, 16 },
		{ 212, 345, 16 },
		{ 212, 346, 16 },
		{ 212, 347, 16 },
		{ 212, 348, 16 },
		{ 212, 349, 16 },
		{ 212, 350, 16 },
		{ 212, 351, 16 },
		{ 212, 352, 16 },
		{ 212, 353, 16 },
		{ 212, 354, 16 },
		{ 212, 355, 16 },
		{ 212, 356, 16 },
		{ 212, 357, 16 },
		{ 212, 358, 16 },
		{ 212, 359, 16 },
		{ 212, 360, 16 },
		{ 212, 361, 16 },
		{ 212, 362, 16 },
		{ 212, 363, 16 },
		{ 212, 364, 16 },
		{ 212, 365, 16 },
		{ 212, 366, 16 },
		{ 212, 367, 16 },
		{ 212, 368, 16 },
		{ 212, 369, 16 },
		{ 212, 370, 16 },
		{ 212, 371, 16 },
		{ 212, 372, 16 },
		{ 212, 373, 16 },
		{ 212, 374, 16 },
		{ 212, 375, 16 },
		{ 212, 376, 16 },
		{ 212, 377, 16 },
		{ 212, 378, 16 },
		{ 212, 379, 16 },
		{ 212, 380, 16 },
		{ 212, 381, 16 },
		{ 212, 382, 16 },
		{ 212, 383, 16 },
		{ 213, 316, 16 },
		{ 213, 317, 16 },
		{ 213, 318, 16 },
		{ 213, 319, 16 },
		{ 213, 320, 16 },
		{ 213, 321, 16 },
		{ 213, 322, 16 },
		{ 213, 323, 16 },
		{ 213, 324, 16 },
		{ 213, 325, 16 },
		{ 213, 326, 16 },
		{ 213, 327, 16 },
		{ 213, 328, 16 },
		{ 213, 329, 16 },
		{ 213, 330, 16 },
		{ 213, 331, 16 },
		{ 213, 332, 16 },
		{ 213, 333, 16 },
		{ 213, 334, 16 },
		{ 213, 335, 16 },
		{ 213, 336, 16 },
		{ 213, 337, 16 },
		{ 213, 338, 16 },
		{ 213, 339, 16 },
		{ 213, 340, 16 },
		{ 213, 341, 16 },
		{ 213, 342, 16 },
		{ 213, 343, 16 },
		{ 213, 344, 16 },
		{ 213, 345, 16 },
		{ 213, 346, 16 },
		{ 213, 347, 16 },
		{ 213, 348, 16 },
		{ 213, 349, 16 },
		{ 213, 350, 16 },
		{ 213, 351, 16 },
		{ 213, 352, 16 },
		{ 213, 353, 16 },
		{ 213, 354, 16 },
		{ 213, 355, 16 },
		{ 213, 356, 16 },
		{ 213, 357, 16 },
		{ 213, 358, 16 },
		{ 213, 359, 16 },
		{ 213, 360, 16 },
		{ 213, 361, 16 },
		{ 213, 362, 16 },
		{ 213, 363, 16 },
		{ 213, 364, 16 },
		{ 213, 365, 16 },
		{ 213, 366, 16 },
		{ 213, 367, 16 },
		{ 213, 368, 16 },
		{ 213, 369, 16 },
		{ 213, 370, 16 },
		{ 213, 371, 16 },
		{ 213, 372, 16 },
		{ 213, 373, 16 },
		{ 213, 374, 16 },
		{ 213, 375, 16 },
		{ 213, 376, 16 },
		{ 213, 377, 16 },
		{ 213, 378, 16 },
		{ 213, 379, 16 },
		{ 213, 380, 16 },
		{ 213, 381, 16 },
		{ 213, 382, 16 },
		{ 213, 383, 16 },
		{ 214, 315, 16 },
		{ 214, 316, 16 },
		{ 214, 317, 16 },
		{ 214, 318, 16 },
		{ 214, 319, 16 },
		{ 214, 320, 16 },
		{ 214, 321, 16 },
		{ 214, 322, 16 },
		{ 214, 323, 16 },
		{ 214, 324, 16 },
		{ 214, 325, 16 },
		{ 214, 326, 16 },
		{ 214, 327, 16 },
		{ 214, 328, 16 },
		{ 214, 329, 16 },
		{ 214, 330, 16 },
		{ 214, 331, 16 },
		{ 214, 332, 16 },
		{ 214, 333, 16 },
		{ 214, 334, 16 },
		{ 214, 335, 16 },
		{ 214, 336, 16 },
		{ 214, 337, 16 },
		{ 214, 338, 16 },
		{ 214, 339, 16 },
		{ 214, 340, 16 },
		{ 214, 341, 16 },
		{ 214, 342, 16 },
		{ 214, 343, 16 },
		{ 214, 344, 16 },
		{ 214, 345, 16 },
		{ 214, 346, 16 },
		{ 214, 347, 16 },
		{ 214, 348, 16 },
		{ 214, 349, 16 },
		{ 214, 350, 16 },
		{ 214, 351, 16 },
		{ 214, 352, 16 },
		{ 214, 353, 16 },
		{ 214, 354, 16 },
		{ 214, 355, 16 },
		{ 214, 356, 16 },
		{ 214, 357, 16 },
		{ 214, 358, 16 },
		{ 214, 359, 16 },
		{ 214, 360, 16 },
		{ 214, 361, 16 },
		{ 214, 362, 16 },
		{ 214, 363, 16 },
		{ 214, 364, 16 },
		{ 214, 365, 16 },
		{ 214, 366, 16 },
		{ 214, 367, 16 },
		{ 214, 368, 16 },
		{ 214, 369, 16 },
		{ 214, 370, 16 },
		{ 214, 371, 16 },
		{ 214, 372, 16 },
		{ 214, 373, 16 },
		{ 214, 374, 16 },
		{ 214, 375, 16 },
		{ 214, 376, 16 },
		{ 214, 377, 16 },
		{ 214, 378, 16 },
		{ 214, 379, 16 },
		{ 214, 380, 16 },
		{ 214, 381, 16 },
		{ 214, 382, 16 },
		{ 214, 383, 16 },
		{ 215, 314, 16 },
		{ 215, 315, 16 },
		{ 215, 316, 16 },
		{ 215, 317, 16 },
		{ 215, 318, 16 },
		{ 215, 319, 16 },
		{ 215, 320, 16 },
		{ 215, 321, 16 },
		{ 215, 322, 16 },
		{ 215, 323, 16 },
		{ 215, 324, 16 },
		{ 215, 325, 16 },
		{ 215, 326, 16 },
		{ 215, 327, 16 },
		{ 215, 328, 16 },
		{ 215, 329, 16 },
		{ 215, 330, 16 },
		{ 215, 331, 16 },
		{ 215, 332, 16 },
		{ 215, 333, 16 },
		{ 215, 334, 16 },
		{ 215, 335, 16 },
		{ 215, 336, 16 },
		{ 215, 337, 16 },
		{ 215, 338, 16 },
		{ 215, 339, 16 },
		{ 215, 340, 16 },
		{ 215, 341, 16 },
		{ 215, 342, 16 },
		{ 215, 343, 16 },
		{ 215, 344, 16 },
		{ 215, 345, 16 },
		{ 215, 346, 16 },
		{ 215, 347, 16 },
		{ 215, 348, 16 },
		{ 215, 349, 16 },
		{ 215, 350, 16 },
		{ 215, 351, 16 },
		{ 215, 352, 16 },
		{ 215, 353, 16 },
		{ 215, 354, 16 },
		{ 215, 355, 16 },
		{ 215, 356, 16 },
		{ 215, 357, 16 },
		{ 215, 358, 16 },
		{ 215, 359, 16 },
		{ 215, 360, 16 },
		{ 215, 361, 16 },
		{ 215, 362, 16 },
		{ 215, 363, 16 },
		{ 215, 364, 16 },
		{ 215, 365, 16 },
		{ 215, 366, 16 },
		{ 215, 367, 16 },
		{ 215, 368, 16 },
		{ 215, 369, 16 },
		{ 215, 370, 16 },
		{ 215, 371, 16 },
		{ 215, 372, 16 },
		{ 215, 373, 16 },
		{ 215, 374, 16 },
		{ 215, 375, 16 },
		{ 215, 376, 16 },
		{ 215, 377, 16 },
		{ 215, 378, 16 },
		{ 215, 379, 16 },
		{ 215, 380, 16 },
		{ 215, 381, 16 },
		{ 215, 382, 16 },
		{ 215, 383, 16 },
		{ 216, 312, 16 },
		{ 216, 313, 16 },
		{ 216, 314, 16 },
		{ 216, 315, 16 },
		{ 216, 316, 16 },
		{ 216, 317, 16 },
		{ 216, 318, 16 },
		{ 216, 319, 16 },
		{ 216, 320, 16 },
		{ 216, 321, 16 },
		{ 216, 322, 16 },
		{ 216, 323, 16 },
		{ 216, 324, 16 },
		{ 216, 325, 16 },
		{ 216, 326, 16 },
		{ 216, 327, 16 },
		{ 216, 328, 16 },
		{ 216, 329, 16 },
		{ 216, 330, 16 },
		{ 216, 331, 16 },
		{ 216, 332, 16 },
		{ 216, 333, 16 },
		{ 216, 334, 16 },
		{ 216, 335, 16 },
		{ 216, 336, 16 },
		{ 216, 337, 16 },
		{ 216, 338, 16 },
		{ 216, 339, 16 },
		{ 216, 340, 16 },
		{ 216, 341, 16 },
		{ 216, 342, 16 },
		{ 216, 343, 16 },
		{ 216, 344, 16 },
		{ 216, 345, 16 },
		{ 216, 346, 16 },
		{ 216, 347, 16 },
		{ 216, 348, 16 },
		{ 216, 349, 16 },
		{ 216, 350, 16 },
		{ 216, 351, 16 },
		{ 216, 352, 16 },
		{ 216, 353, 16 },
		{ 216, 354, 16 },
		{ 216, 355, 16 },
		{ 216, 356, 16 },
		{ 216, 357, 16 },
		{ 216, 358, 16 },
		{ 216, 359, 16 },
		{ 216, 360, 16 },
		{ 216, 361, 16 },
		{ 216, 362, 16 },
		{ 216, 363, 16 },
		{ 216, 364, 16 },
		{ 216, 365, 16 },
		{ 216, 366, 16 },
		{ 216, 367, 16 },
		{ 216, 368, 16 },
		{ 216, 369, 16 },
		{ 216, 370, 16 },
		{ 216, 371, 16 },
		{ 216, 372, 16 },
		{ 216, 373, 16 },
		{ 216, 374, 16 },
		{ 216, 375, 16 },
		{ 216, 376, 16 },
		{ 216, 377, 16 },
		{ 216, 378, 16 },
		{ 216, 379, 16 },
		{ 216, 380, 16 },
		{ 216, 381, 16 },
		{ 216, 382, 16 },
		{ 216, 383, 16 },
		{ 217, 311, 16 },
		{ 217, 312, 16 },
		{ 217, 313, 16 },
		{ 217, 314, 16 },
		{ 217, 315, 16 },
		{ 217, 316, 16 },
		{ 217, 317, 16 },
		{ 217, 318, 16 },
		{ 217, 319, 16 },
		{ 217, 320, 16 },
		{ 217, 321, 16 },
		{ 217, 322, 16 },
		{ 217, 323, 16 },
		{ 217, 324, 16 },
		{ 217, 325, 16 },
		{ 217, 326, 16 },
		{ 217, 327, 16 },
		{ 217, 328, 16 },
		{ 217, 329, 16 },
		{ 217, 330, 16 },
		{ 217, 331, 16 },
		{ 217, 332, 16 },
		{ 217, 333, 16 },
		{ 217, 334, 16 },
		{ 217, 335, 16 },
		{ 217, 336, 16 },
		{ 217, 337, 16 },
		{ 217, 338, 16 },
		{ 217, 339, 16 },
		{ 217, 340, 16 },
		{ 217, 341, 16 },
		{ 217, 342, 16 },
		{ 217, 343, 16 },
		{ 217, 344, 16 },
		{ 217, 345, 16 },
		{ 217, 346, 16 },
		{ 217, 347, 16 },
		{ 217, 348, 16 },
		{ 217, 349, 16 },
		{ 217, 350, 16 },
		{ 217, 351, 16 },
		{ 217, 352, 16 },
		{ 217, 353, 16 },
		{ 217, 354, 16 },
		{ 217, 355, 16 },
		{ 217, 356, 16 },
		{ 217, 357, 16 },
		{ 217, 358, 16 },
		{ 217, 359, 16 },
		{ 217, 360, 16 },
		{ 217, 361, 16 },
		{ 217, 362, 16 },
		{ 217, 363, 16 },
		{ 217, 364, 16 },
		{ 217, 365, 16 },
		{ 217, 366, 16 },
		{ 217, 367, 16 },
		{ 217, 368, 16 },
		{ 217, 369, 16 },
		{ 217, 370, 16 },
		{ 217, 371, 16 },
		{ 217, 372, 16 },
		{ 217, 373, 16 },
		{ 217, 374, 16 },
		{ 217, 375, 16 },
		{ 217, 376, 16 },
		{ 217, 377, 16 },
		{ 217, 378, 16 },
		{ 217, 379, 16 },
		{ 217, 380, 16 },
		{ 217, 381, 16 },
		{ 217, 382, 16 },
		{ 217, 383, 16 },
		{ 218, 310, 16 },
		{ 218, 311, 16 },
		{ 218, 312, 16 },
		{ 218, 313, 16 },
		{ 218, 314, 16 },
		{ 218, 315, 16 },
		{ 218, 316, 16 },
		{ 218, 317, 16 },
		{ 218, 318, 16 },
		{ 218, 319, 16 },
		{ 218, 320, 16 },
		{ 218, 321, 16 },
		{ 218, 322, 16 },
		{ 218, 323, 16 },
		{ 218, 324, 16 },
		{ 218, 325, 16 },
		{ 218, 326, 16 },
		{ 218, 327, 16 },
		{ 218, 328, 16 },
		{ 218, 329, 16 },
		{ 218, 330, 16 },
		{ 218, 331, 16 },
		{ 218, 332, 16 },
		{ 218, 333, 16 },
		{ 218, 334, 16 },
		{ 218, 335, 16 },
		{ 218, 336, 16 },
		{ 218, 337, 16 },
		{ 218, 338, 16 },
		{ 218, 339, 16 },
		{ 218, 340, 16 },
		{ 218, 341, 16 },
		{ 218, 342, 16 },
		{ 218, 343, 16 },
		{ 218, 344, 16 },
		{ 218, 345, 16 },
		{ 218, 346, 16 },
		{ 218, 347, 16 },
		{ 218, 348, 16 },
		{ 218, 349, 16 },
		{ 218, 350, 16 },
		{ 218, 351, 16 },
		{ 218, 352, 16 },
		{ 218, 353, 16 },
		{ 218, 354, 16 },
		{ 218, 355, 16 },
		{ 218, 356, 16 },
		{ 218, 357, 16 },
		{ 218, 358, 16 },
		{ 218, 359, 16 },
		{ 218, 360, 16 },
		{ 218, 361, 16 },
		{ 218, 362, 16 },
		{ 218, 363, 16 },
		{ 218, 364, 16 },
		{ 218, 365, 16 },
		{ 218, 366, 16 },
		{ 218, 367, 16 },
		{ 218, 368, 16 },
		{ 218, 369, 16 },
		{ 218, 370, 16 },
		{ 218, 371, 16 },
		{ 218, 372, 16 },
		{ 218, 373, 16 },
		{ 218, 374, 16 },
		{ 218, 375, 16 },
		{ 218, 376, 16 },
		{ 218, 377, 16 },
		{ 218, 378, 16 },
		{ 218, 379, 16 },
		{ 218, 380, 16 },
		{ 218, 381, 16 },
		{ 218, 382, 16 },
		{ 218, 383, 16 },
		{ 219, 308, 16 },
		{ 219, 309, 16 },
		{ 219, 310, 16 },
		{ 219, 311, 16 },
		{ 219, 312, 16 },
		{ 219, 313, 16 },
		{ 219, 314, 16 },
		{ 219, 315, 16 },
		{ 219, 316, 16 },
		{ 219, 317, 16 },
		{ 219, 318, 16 },
		{ 219, 319, 16 },
		{ 219, 320, 16 },
		{ 219, 321, 16 },
		{ 219, 322, 16 },
		{ 219, 323, 16 },
		{ 219, 324, 16 },
		{ 219, 325, 16 },
		{ 219, 326, 16 },
		{ 219, 327, 16 },
		{ 219, 328, 16 },
		{ 219, 329, 16 },
		{ 219, 330, 16 },
		{ 219, 331, 16 },
		{ 219, 332, 16 },
		{ 219, 333, 16 },
		{ 219, 334, 16 },
		{ 219, 335, 16 },
		{ 219, 336, 16 },
		{ 219, 337, 16 },
		{ 219, 338, 16 },
		{ 219, 339, 16 },
		{ 219, 340, 16 },
		{ 219, 341, 16 },
		{ 219, 342, 16 },
		{ 219, 343, 16 },
		{ 219, 344, 16 },
		{ 219, 345, 16 },
		{ 219, 346, 16 },
		{ 219, 347, 16 },
		{ 219, 348, 16 },
		{ 219, 349, 16 },
		{ 219, 350, 16 },
		{ 219, 351, 16 },
		{ 219, 352, 16 },
		{ 219, 353, 16 },
		{ 219, 354, 16 },
		{ 219, 355, 16 },
		{ 219, 356, 16 },
		{ 219, 357, 16 },
		{ 219, 358, 16 },
		{ 219, 359, 16 },
		{ 219, 360, 16 },
		{ 219, 361, 16 },
		{ 219, 362, 16 },
		{ 219, 363, 16 },
		{ 219, 364, 16 },
		{ 219, 365, 16 },
		{ 219, 366, 16 },
		{ 219, 367, 16 },
		{ 219, 368, 16 },
		{ 219, 369, 16 },
		{ 219, 370, 16 },
		{ 219, 371, 16 },
		{ 219, 372, 16 },
		{ 219, 373, 16 },
		{ 219, 374, 16 },
		{ 219, 375, 16 },
		{ 219, 376, 16 },
		{ 219, 377, 16 },
		{ 219, 378, 16 },
		{ 219, 379, 16 },
		{ 219, 380, 16 },
		{ 219, 381, 16 },
		{ 219, 382, 16 },
		{ 220, 307, 16 },
		{ 220, 308, 16 },
		{ 220, 309, 16 },
		{ 220, 310, 16 },
		{ 220, 311, 16 },
		{ 220, 312, 16 },
		{ 220, 313, 16 },
		{ 220, 314, 16 },
		{ 220, 315, 16 },
		{ 220, 316, 16 },
		{ 220, 317, 16 },
		{ 220, 318, 16 },
		{ 220, 319, 16 },
		{ 220, 320, 16 },
		{ 220, 321, 16 },
		{ 220, 322, 16 },
		{ 220, 323, 16 },
		{ 220, 324, 16 },
		{ 220, 325, 16 },
		{ 220, 326, 16 },
		{ 220, 327, 16 },
		{ 220, 328, 16 },
		{ 220, 329, 16 },
		{ 220, 330, 16 },
		{ 220, 331, 16 },
		{ 220, 332, 16 },
		{ 220, 333, 16 },
		{ 220, 334, 16 },
		{ 220, 335, 16 },
		{ 220, 336, 16 },
		{ 220, 337, 16 },
		{ 220, 338, 16 },
		{ 220, 339, 16 },
		{ 220, 340, 16 },
		{ 220, 341, 16 },
		{ 220, 342, 16 },
		{ 220, 343, 16 },
		{ 220, 344, 16 },
		{ 220, 345, 16 },
		{ 220, 346, 16 },
		{ 220, 347, 16 },
		{ 220, 348, 16 },
		{ 220, 349, 16 },
		{ 220, 350, 16 },
		{ 220, 351, 16 },
		{ 220, 352, 16 },
		{ 220, 353, 16 },
		{ 220, 354, 16 },
		{ 220, 355, 16 },
		{ 220, 356, 16 },
		{ 220, 357, 16 },
		{ 220, 358, 16 },
		{ 220, 359, 16 },
		{ 220, 360, 16 },
		{ 220, 361, 16 },
		{ 220, 362, 16 },
		{ 220, 363, 16 },
		{ 220, 364, 16 },
		{ 220, 365, 16 },
		{ 220, 366, 16 },
		{ 220, 367, 16 },
		{ 220, 368, 16 },
		{ 220, 369, 16 },
		{ 220, 370, 16 },
		{ 220, 371, 16 },
		{ 220, 372, 16 },
		{ 220, 373, 16 },
		{ 220, 374, 16 },
		{ 220, 375, 16 },
		{ 220, 376, 16 },
		{ 220, 377, 16 },
		{ 220, 378, 16 },
		{ 220, 379, 16 },
		{ 220, 380, 16 },
		{ 220, 381, 16 },
		{ 220, 382, 16 },
		{ 221, 306, 16 },
		{ 221, 307, 16 },
		{ 221, 308, 16 },
		{ 221, 309, 16 },
		{ 221, 310, 16 },
		{ 221, 311, 16 },
		{ 221, 312, 16 },
		{ 221, 313, 16 },
		{ 221, 314, 16 },
		{ 221, 315, 16 },
		{ 221, 316, 16 },
		{ 221, 317, 16 },
		{ 221, 318, 16 },
		{ 221, 319, 16 },
		{ 221, 320, 16 },
		{ 221, 321, 16 },
		{ 221, 322, 16 },
		{ 221, 323, 16 },
		{ 221, 324, 16 },
		{ 221, 325, 16 },
		{ 221, 326, 16 },
		{ 221, 327, 16 },
		{ 221, 328, 16 },
		{ 221, 329, 16 },
		{ 221, 330, 16 },
		{ 221, 331, 16 },
		{ 221, 332, 16 },
		{ 221, 333, 16 },
		{ 221, 334, 16 },
		{ 221, 335, 16 },
		{ 221, 336, 16 },
		{ 221, 337, 16 },
		{ 221, 338, 16 },
		{ 221, 339, 16 },
		{ 221, 340, 16 },
		{ 221, 341, 16 },
		{ 221, 342, 16 },
		{ 221, 343, 16 },
		{ 221, 344, 16 },
		{ 221, 345, 16 },
		{ 221, 346, 16 },
		{ 221, 347, 16 },
		{ 221, 348, 16 },
		{ 221, 349, 16 },
		{ 221, 350, 16 },
		{ 221, 351, 16 },
		{ 221, 352, 16 },
		{ 221, 353, 16 },
		{ 221, 354, 16 },
		{ 221, 355, 16 },
		{ 221, 356, 16 },
		{ 221, 357, 16 },
		{ 221, 358, 16 },
		{ 221, 359, 16 },
		{ 221, 360, 16 },
		{ 221, 361, 16 },
		{ 221, 362, 16 },
		{ 221, 363, 16 },
		{ 221, 364, 16 },
		{ 221, 365, 16 },
		{ 221, 366, 16 },
		{ 221, 367, 16 },
		{ 221, 368, 16 },
		{ 221, 369, 16 },
		{ 221, 370, 16 },
		{ 221, 371, 16 },
		{ 221, 372, 16 },
		{ 221, 373, 16 },
		{ 221, 374, 16 },
		{ 221, 375, 16 },
		{ 221, 376, 16 },
		{ 221, 377, 16 },
		{ 221, 378, 16 },
		{ 221, 379, 16 },
		{ 221, 380, 16 },
		{ 221, 381, 16 },
		{ 221, 382, 16 },
		{ 222, 304, 16 },
		{ 222, 305, 16 },
		{ 222, 306, 16 },
		{ 222, 307, 16 },
		{ 222, 308, 16 },
		{ 222, 309, 16 },
		{ 222, 310, 16 },
		{ 222, 311, 16 },
		{ 222, 312, 16 },
		{ 222, 313, 16 },
		{ 222, 314, 16 },
		{ 222, 315, 16 },
		{ 222, 316, 16 },
		{ 222, 317, 16 },
		{ 222, 318, 16 },
		{ 222, 319, 16 },
		{ 222, 320, 16 },
		{ 222, 321, 16 },
		{ 222, 322, 16 },
		{ 222, 323, 16 },
		{ 222, 324, 16 },
		{ 222, 325, 16 },
		{ 222, 326, 16 },
		{ 222, 327, 16 },
		{ 222, 328, 16 },
		{ 222, 329, 16 },
		{ 222, 330, 16 },
		{ 222, 331, 16 },
		{ 222, 332, 16 },
		{ 222, 333, 16 },
		{ 222, 334, 16 },
		{ 222, 335, 16 },
		{ 222, 336, 16 },
		{ 222, 337, 16 },
		{ 222, 338, 16 },
		{ 222, 339, 16 },
		{ 222, 340, 16 },
		{ 222, 341, 16 },
		{ 222, 342, 16 },
		{ 222, 343, 16 },
		{ 222, 344, 16 },
		{ 222, 345, 16 },
		{ 222, 346, 16 },
		{ 222, 347, 16 },
		{ 222, 348, 16 },
		{ 222, 349, 16 },
		{ 222, 350, 16 },
		{ 222, 351, 16 },
		{ 222, 352, 16 },
		{ 222, 353, 16 },
		{ 222, 354, 16 },
		{ 222, 355, 16 },
		{ 222, 356, 16 },
		{ 222, 357, 16 },
		{ 222, 358, 16 },
		{ 222, 359, 16 },
		{ 222, 360, 16 },
		{ 222, 361, 16 },
		{ 222, 362, 16 },
		{ 222, 363, 16 },
		{ 222, 364, 16 },
		{ 222, 365, 16 },
		{ 222, 366, 16 },
		{ 222, 367, 16 },
		{ 222, 368, 16 },
		{ 222, 369, 16 },
		{ 222, 370, 16 },
		{ 222, 371, 16 },
		{ 222, 372, 16 },
		{ 222, 373, 16 },
		{ 222, 374, 16 },
		{ 222, 375, 16 },
		{ 222, 376, 16 },
		{ 222, 377, 16 },
		{ 222, 378, 16 },
		{ 222, 379, 16 },
		{ 222, 380, 16 },
		{ 222, 381, 16 },
		{ 222, 382, 16 },
		{ 223, 303, 16 },
		{ 223, 304, 16 },
		{ 223, 305, 16 },
		{ 223, 306, 16 },
		{ 223, 307, 16 },
		{ 223, 308, 16 },
		{ 223, 309, 16 },
		{ 223, 310, 16 },
		{ 223, 311, 16 },
		{ 223, 312, 16 },
		{ 223, 313, 16 },
		{ 223, 314, 16 },
		{ 223, 315, 16 },
		{ 223, 316, 16 },
		{ 223, 317, 16 },
		{ 223, 318, 16 },
		{ 223, 319, 16 },
		{ 223, 320, 16 },
		{ 223, 321, 16 },
		{ 223, 322, 16 },
		{ 223, 323, 16 },
		{ 223, 324, 16 },
		{ 223, 325, 16 },
		{ 223, 326, 16 },
		{ 223, 327, 16 },
		{ 223, 328, 16 },
		{ 223, 329, 16 },
		{ 223, 330, 16 },
		{ 223, 331, 16 },
		{ 223, 332, 16 },
		{ 223, 333, 16 },
		{ 223, 334, 16 },
		{ 223, 335, 16 },
		{ 223, 336, 16 },
		{ 223, 337, 16 },
		{ 223, 338, 16 },
		{ 223, 339, 16 },
		{ 223, 340, 16 },
		{ 223, 341, 16 },
		{ 223, 342, 16 },
		{ 223, 343, 16 },
		{ 223, 344, 16 },
		{ 223, 345, 16 },
		{ 223, 346, 16 },
		{ 223, 347, 16 },
		{ 223, 348, 16 },
		{ 223, 349, 16 },
		{ 223, 350, 16 },
		{ 223, 351, 16 },
		{ 223, 352, 16 },
		{ 223, 353, 16 },
		{ 223, 354, 16 },
		{ 223, 355, 16 },
		{ 223, 356, 16 },
		{ 223, 357, 16 },
		{ 223, 358, 16 },
		{ 223, 359, 16 },
		{ 223, 360, 16 },
		{ 223, 361, 16 },
		{ 223, 362, 16 },
		{ 223, 363, 16 },
		{ 223, 364, 16 },
		{ 223, 365, 16 },
		{ 223, 366, 16 },
		{ 223, 367, 16 },
		{ 223, 368, 16 },
		{ 223, 369, 16 },
		{ 223, 370, 16 },
		{ 223, 371, 16 },
		{ 223, 372, 16 },
		{ 223, 373, 16 },
		{ 223, 374, 16 },
		{ 223, 375, 16 },
		{ 223, 376, 16 },
		{ 223, 377, 16 },
		{ 223, 378, 16 },
		{ 223, 379, 16 },
		{ 223, 380, 16 },
		{ 223, 381, 16 },
		{ 224, 302, 16 },
		{ 224, 303, 16 },
		{ 224, 304, 16 },
		{ 224, 305, 16 },
		{ 224, 306, 16 },
		{ 224, 307, 16 },
		{ 224, 308, 16 },
		{ 224, 309, 16 },
		{ 224, 310, 16 },
		{ 224, 311, 16 },
		{ 224, 312, 16 },
		{ 224, 313, 16 },
		{ 224, 314, 16 },
		{ 224, 315, 16 },
		{ 224, 316, 16 },
		{ 224, 317, 16 },
		{ 224, 318, 16 },
		{ 224, 319, 16 },
		{ 224, 320, 16 },
		{ 224, 321, 16 },
		{ 224, 322, 16 },
		{ 224, 323, 16 },
		{ 224, 324, 16 },
		{ 224, 325, 16 },
		{ 224, 326, 16 },
		{ 224, 327, 16 },
		{ 224, 328, 16 },
		{ 224, 329, 16 },
		{ 224, 330, 16 },
		{ 224, 331, 16 },
		{ 224, 332, 16 },
		{ 224, 333, 16 },
		{ 224, 334, 16 },
		{ 224, 335, 16 },
		{ 224, 336, 16 },
		{ 224, 337, 16 },
		{ 224, 338, 16 },
		{ 224, 339, 16 },
		{ 224, 340, 16 },
		{ 224, 341, 16 },
		{ 224, 342, 16 },
		{ 224, 343, 16 },
		{ 224, 344, 16 },
		{ 224, 345, 16 },
		{ 224, 346, 16 },
		{ 224, 347, 16 },
		{ 224, 348, 16 },
		{ 224, 349, 16 },
		{ 224, 350, 16 },
		{ 224, 351, 16 },
		{ 224, 352, 16 },
		{ 224, 353, 16 },
		{ 224, 354, 16 },
		{ 224, 355, 16 },
		{ 224, 356, 16 },
		{ 224, 357, 16 },
		{ 224, 358, 16 },
		{ 224, 359, 16 },
		{ 224, 360, 16 },
		{ 224, 361, 16 },
		{ 224, 362, 16 },
		{ 224, 363, 16 },
		{ 224, 364, 16 },
		{ 224, 365, 16 },
		{ 224, 366, 16 },
		{ 224, 367, 16 },
		{ 224, 368, 16 },
		{ 224, 369, 16 },
		{ 224, 370, 16 },
		{ 224, 371, 16 },
		{ 224, 372, 16 },
		{ 224, 373, 16 },
		{ 224, 374, 16 },
		{ 224, 375, 16 },
		{ 224, 376, 16 },
		{ 224, 377, 16 },
		{ 224, 378, 16 },
		{ 224, 379, 16 },
		{ 224, 380, 16 },
		{ 224, 381, 16 },
		{ 225, 300, 16 },
		{ 225, 301, 16 },
		{ 225, 302, 16 },
		{ 225, 303, 16 },
		{ 225, 304, 16 },
		{ 225, 305, 16 },
		{ 225, 306, 16 },
		{ 225, 307, 16 },
		{ 225, 308, 16 },
		{ 225, 309, 16 },
		{ 225, 310, 16 },
		{ 225, 311, 16 },
		{ 225, 312, 16 },
		{ 225, 313, 16 },
		{ 225, 314, 16 },
		{ 225, 315, 16 },
		{ 225, 316, 16 },
		{ 225, 317, 16 },
		{ 225, 318, 16 },
		{ 225, 319, 16 },
		{ 225, 320, 16 },
		{ 225, 321, 16 },
		{ 225, 322, 16 },
		{ 225, 323, 16 },
		{ 225, 324, 16 },
		{ 225, 325, 16 },
		{ 225, 326, 16 },
		{ 225, 327, 16 },
		{ 225, 328, 16 },
		{ 225, 329, 16 },
		{ 225, 330, 16 },
		{ 225, 331, 16 },
		{ 225, 332, 16 },
		{ 225, 333, 16 },
		{ 225, 334, 16 },
		{ 225, 335, 16 },
		{ 225, 336, 16 },
		{ 225, 337, 16 },
		{ 225, 338, 16 },
		{ 225, 339, 16 },
		{ 225, 340, 16 },
		{ 225, 341, 16 },
		{ 225, 342, 16 },
		{ 225, 343, 16 },
		{ 225, 344, 16 },
		{ 225, 345, 16 },
		{ 225, 346, 16 },
		{ 225, 347, 16 },
		{ 225, 348, 16 },
		{ 225, 349, 16 },
		{ 225, 350, 16 },
		{ 225, 351, 16 },
		{ 225, 352, 16 },
		{ 225, 353, 16 },
		{ 225, 354, 16 },
		{ 225, 355, 16 },
		{ 225, 356, 16 },
		{ 225, 357, 16 },
		{ 225, 358, 16 },
		{ 225, 359, 16 },
		{ 225, 360, 16 },
		{ 225, 361, 16 },
		{ 225, 362, 16 },
		{ 225, 363, 16 },
		{ 225, 364, 16 },
		{ 225, 365, 16 },
		{ 225, 366, 16 },
		{ 225, 367, 16 },
		{ 225, 368, 16 },
		{ 225, 369, 16 },
		{ 225, 370, 16 },
		{ 225, 371, 16 },
		{ 225, 372, 16 },
		{ 225, 373, 16 },
		{ 225, 374, 16 },
		{ 225, 375, 16 },
		{ 225, 376, 16 },
		{ 225, 377, 16 },
		{ 225, 378, 16 },
		{ 225, 379, 16 },
		{ 225, 380, 16 },
		{ 226, 299, 16 },
		{ 226, 300, 16 },
		{ 226, 301, 16 },
		{ 226, 302, 16 },
		{ 226, 303, 16 },
		{ 226, 304, 16 },
		{ 226, 305, 16 },
		{ 226, 306, 16 },
		{ 226, 307, 16 },
		{ 226, 308, 16 },
		{ 226, 309, 16 },
		{ 226, 310, 16 },
		{ 226, 311, 16 },
		{ 226, 312, 16 },
		{ 226, 313, 16 },
		{ 226, 314, 16 },
		{ 226, 315, 16 },
		{ 226, 316, 16 },
		{ 226, 317, 16 },
		{ 226, 318, 16 },
		{ 226, 319, 16 },
		{ 226, 320, 16 },
		{ 226, 321, 16 },
		{ 226, 322, 16 },
		{ 226, 323, 16 },
		{ 226, 324, 16 },
		{ 226, 325, 16 },
		{ 226, 326, 16 },
		{ 226, 327, 16 },
		{ 226, 328, 16 },
		{ 226, 329, 16 },
		{ 226, 330, 16 },
		{ 226, 331, 16 },
		{ 226, 332, 16 },
		{ 226, 333, 16 },
		{ 226, 334, 16 },
		{ 226, 335, 16 },
		{ 226, 336, 16 },
		{ 226, 337, 16 },
		{ 226, 338, 16 },
		{ 226, 339, 16 },
		{ 226, 340, 16 },
		{ 226, 341, 16 },
		{ 226, 342, 16 },
		{ 226, 343, 16 },
		{ 226, 344, 16 },
		{ 226, 345, 16 },
		{ 226, 346, 16 },
		{ 226, 347, 16 },
		{ 226, 348, 16 },
		{ 226, 349, 16 },
		{ 226, 350, 16 },
		{ 226, 351, 16 },
		{ 226, 352, 16 },
		{ 226, 353, 16 },
		{ 226, 354, 16 },
		{ 226, 355, 16 },
		{ 226, 356, 16 },
		{ 226, 357, 16 },
		{ 226, 358, 16 },
		{ 226, 359, 16 },
		{ 226, 360, 16 },
		{ 226, 361, 16 },
		{ 226, 362, 16 },
		{ 226, 363, 16 },
		{ 226, 364, 16 },
		{ 226, 365, 16 },
		{ 226, 366, 16 },
		{ 226, 367, 16 },
		{ 226, 368, 16 },
		{ 226, 369, 16 },
		{ 226, 370, 16 },
		{ 226, 371, 16 },
		{ 226, 372, 16 },
		{ 226, 373, 16 },
		{ 226, 374, 16 },
		{ 226, 375, 16 },
		{ 226, 376, 16 },
		{ 226, 377, 16 },
		{ 226, 378, 16 },
		{ 227, 298, 16 },
		{ 227, 299, 16 },
		{ 227, 300, 16 },
		{ 227, 301, 16 },
		{ 227, 302, 16 },
		{ 227, 303, 16 },
		{ 227, 304, 16 },
		{ 227, 305, 16 },
		{ 227, 306, 16 },
		{ 227, 307, 16 },
		{ 227, 308, 16 },
		{ 227, 309, 16 },
		{ 227, 310, 16 },
		{ 227, 311, 16 },
		{ 227, 312, 16 },
		{ 227, 313, 16 },
		{ 227, 314, 16 },
		{ 227, 315, 16 },
		{ 227, 316, 16 },
		{ 227, 317, 16 },
		{ 227, 318, 16 },
		{ 227, 319, 16 },
		{ 227, 320, 16 },
		{ 227, 321, 16 },
		{ 227, 322, 16 },
		{ 227, 323, 16 },
		{ 227, 324, 16 },
		{ 227, 325, 16 },
		{ 227, 326, 16 },
		{ 227, 327, 16 },
		{ 227, 328, 16 },
		{ 227, 329, 16 },
		{ 227, 330, 16 },
		{ 227, 331, 16 },
		{ 227, 332, 16 },
		{ 227, 333, 16 },
		{ 227, 334, 16 },
		{ 227, 335, 16 },
		{ 227, 336, 16 },
		{ 227, 337, 16 },
		{ 227, 338, 16 },
		{ 227, 339, 16 },
		{ 227, 340, 16 },
		{ 227, 341, 16 },
		{ 227, 342, 16 },
		{ 227, 343, 16 },
		{ 227, 344, 16 },
		{ 227, 345, 16 },
		{ 227, 346, 16 },
		{ 227, 347, 16 },
		{ 227, 348, 16 },
		{ 227, 349, 16 },
		{ 227, 350, 16 },
		{ 227, 351, 16 },
		{ 227, 352, 16 },
		{ 227, 353, 16 },
		{ 227, 354, 16 },
		{ 227, 355, 16 },
		{ 227, 356, 16 },
		{ 227, 357, 16 },
		{ 227, 358, 16 },
		{ 227, 359, 16 },
		{ 227, 360, 16 },
		{ 227, 361, 16 },
		{ 227, 362, 16 },
		{ 227, 363, 16 },
		{ 227, 364, 16 },
		{ 227, 365, 16 },
		{ 227, 366, 16 },
		{ 227, 367, 16 },
		{ 227, 368, 16 },
		{ 227, 369, 16 },
		{ 227, 370, 16 },
		{ 227, 371, 16 },
		{ 227, 372, 16 },
		{ 227, 373, 16 },
		{ 227, 374, 16 },
		{ 227, 375, 16 },
		{ 228, 296, 16 },
		{ 228, 297, 16 },
		{ 228, 298, 16 },
		{ 228, 299, 16 },
		{ 228, 300, 16 },
		{ 228, 301, 16 },
		{ 228, 302, 16 },
		{ 228, 303, 16 },
		{ 228, 304, 16 },
		{ 228, 305, 16 },
		{ 228, 306, 16 },
		{ 228, 307, 16 },
		{ 228, 308, 16 },
		{ 228, 309, 16 },
		{ 228, 310, 16 },
		{ 228, 311, 16 },
		{ 228, 312, 16 },
		{ 228, 313, 16 },
		{ 228, 314, 16 },
		{ 228, 315, 16 },
		{ 228, 316, 16 },
		{ 228, 317, 16 },
		{ 228, 318, 16 },
		{ 228, 319, 16 },
		{ 228, 320, 16 },
		{ 228, 321, 16 },
		{ 228, 322, 16 },
		{ 228, 323, 16 },
		{ 228, 324, 16 },
		{ 228, 325, 16 },
		{ 228, 326, 16 },
		{ 228, 327, 16 },
		{ 228, 328, 16 },
		{ 228, 329, 16 },
		{ 228, 330, 16 },
		{ 228, 331, 16 },
		{ 228, 332, 16 },
		{ 228, 333, 16 },
		{ 228, 334, 16 },
		{ 228, 335, 16 },
		{ 228, 336, 16 },
		{ 228, 337, 16 },
		{ 228, 338, 16 },
		{ 228, 339, 16 },
		{ 228, 340, 16 },
		{ 228, 341, 16 },
		{ 228, 342, 16 },
		{ 228, 343, 16 },
		{ 228, 344, 16 },
		{ 228, 345, 16 },
		{ 228, 346, 16 },
		{ 228, 347, 16 },
		{ 228, 348, 16 },
		{ 228, 349, 16 },
		{ 228, 350, 16 },
		{ 228, 351, 16 },
		{ 228, 352, 16 },
		{ 228, 353, 16 },
		{ 228, 354, 16 },
		{ 228, 355, 16 },
		{ 228, 356, 16 },
		{ 228, 357, 16 },
		{ 228, 358, 16 },
		{ 228, 359, 16 },
		{ 228, 360, 16 },
		{ 228, 361, 16 },
		{ 228, 362, 16 },
		{ 228, 363, 16 },
		{ 228, 364, 16 },
		{ 228, 365, 16 },
		{ 228, 366, 16 },
		{ 228, 367, 16 },
		{ 228, 368, 16 },
		{ 228, 369, 16 },
		{ 228, 370, 16 },
		{ 228, 371, 16 },
		{ 228, 372, 16 },
		{ 229, 295, 16 },
		{ 229, 296, 16 },
		{ 229, 297, 16 },
		{ 229, 298, 16 },
		{ 229, 299, 16 },
		{ 229, 300, 16 },
		{ 229, 301, 16 },
		{ 229, 302, 16 },
		{ 229, 303, 16 },
		{ 229, 304, 16 },
		{ 229, 305, 16 },
		{ 229, 306, 16 },
		{ 229, 307, 16 },
		{ 229, 308, 16 },
		{ 229, 309, 16 },
		{ 229, 310, 16 },
		{ 229, 311, 16 },
		{ 229, 312, 16 },
		{ 229, 313, 16 },
		{ 229, 314, 16 },
		{ 229, 315, 16 },
		{ 229, 316, 16 },
		{ 229, 317, 16 },
		{ 229, 318, 16 },
		{ 229, 319, 16 },
		{ 229, 320, 16 },
		{ 229, 321, 16 },
		{ 229, 322, 16 },
		{ 229, 323, 16 },
		{ 229, 324, 16 },
		{ 229, 325, 16 },
		{ 229, 326, 16 },
		{ 229, 327, 16 },
		{ 229, 328, 16 },
		{ 229, 329, 16 },
		{ 229, 330, 16 },
		{ 229, 331, 16 },
		{ 229, 332, 16 },
		{ 229, 333, 16 },
		{ 229, 334, 16 },
		{ 229, 335, 16 },
		{ 229, 336, 16 },
		{ 229, 337, 16 },
		{ 229, 338, 16 },
		{ 229, 339, 16 },
		{ 229, 340, 16 },
		{ 229, 341, 16 },
		{ 229, 342, 16 },
		{ 229, 343, 16 },
		{ 229, 344, 16 },
		{ 229, 345, 16 },
		{ 229, 346, 16 },
		{ 229, 347, 16 },
		{ 229, 348, 16 },
		{ 229, 349, 16 },
		{ 229, 350, 16 },
		{ 229, 351, 16 },
		{ 229, 352, 16 },
		{ 229, 353, 16 },
		{ 229, 354, 16 },
		{ 229, 355, 16 },
		{ 229, 356, 16 },
		{ 229, 357, 16 },
		{ 229, 358, 16 },
		{ 229, 359, 16 },
		{ 229, 360, 16 },
		{ 229, 361, 16 },
		{ 229, 362, 16 },
		{ 229, 363, 16 },
		{ 229, 364, 16 },
		{ 229, 365, 16 },
		{ 229, 366, 16 },
		{ 229, 367, 16 },
		{ 229, 368, 16 },
		{ 230, 294, 16 },
		{ 230, 295, 16 },
		{ 230, 296, 16 },
		{ 230, 297, 16 },
		{ 230, 298, 16 },
		{ 230, 299, 16 },
		{ 230, 300, 16 },
		{ 230, 301, 16 },
		{ 230, 302, 16 },
		{ 230, 303, 16 },
		{ 230, 304, 16 },
		{ 230, 305, 16 },
		{ 230, 306, 16 },
		{ 230, 307, 16 },
		{ 230, 308, 16 },
		{ 230, 309, 16 },
		{ 230, 310, 16 },
		{ 230, 311, 16 },
		{ 230, 312, 16 },
		{ 230, 313, 16 },
		{ 230, 314, 16 },
		{ 230, 315, 16 },
		{ 230, 316, 16 },
		{ 230, 317, 16 },
		{ 230, 318, 16 },
		{ 230, 319, 16 },
		{ 230, 320, 16 },
		{ 230, 321, 16 },
		{ 230, 322, 16 },
		{ 230, 323, 16 },
		{ 230, 324, 16 },
		{ 230, 325, 16 },
		{ 230, 326, 16 },
		{ 230, 327, 16 },
		{ 230, 328, 16 },
		{ 230, 329, 16 },
		{ 230, 330, 16 },
		{ 230, 331, 16 },
		{ 230, 332, 16 },
		{ 230, 333, 16 },
		{ 230, 334, 16 },
		{ 230, 335, 16 },
		{ 230, 336, 16 },
		{ 230, 337, 16 },
		{ 230, 338, 16 },
		{ 230, 339, 16 },
		{ 230, 340, 16 },
		{ 230, 341, 16 },
		{ 230, 342, 16 },
		{ 230, 343, 16 },
		{ 230, 344, 16 },
		{ 230, 345, 16 },
		{ 230, 346, 16 },
		{ 230, 347, 16 },
		{ 230, 348, 16 },
		{ 230, 349, 16 },
		{ 230, 350, 16 },
		{ 230, 351, 16 },
		{ 230, 352, 16 },
		{ 230, 353, 16 },
		{ 230, 354, 16 },
		{ 230, 355, 16 },
		{ 230, 356, 16 },
		{ 230, 357, 16 },
		{ 230, 358, 16 },
		{ 230, 359, 16 },
		{ 230, 360, 16 },
		{ 230, 361, 16 },
		{ 230, 362, 16 },
		{ 230, 363, 16 },
		{ 230, 364, 16 },
		{ 231, 292, 16 },
		{ 231, 293, 16 },
		{ 231, 294, 16 },
		{ 231, 295, 16 },
		{ 231, 296, 16 },
		{ 231, 297, 16 },
		{ 231, 298, 16 },
		{ 231, 299, 16 },
		{ 231, 300, 16 },
		{ 231, 301, 16 },
		{ 231, 302, 16 },
		{ 231, 303, 16 },
		{ 231, 304, 16 },
		{ 231, 305, 16 },
		{ 231, 306, 16 },
		{ 231, 307, 16 },
		{ 231, 308, 16 },
		{ 231, 309, 16 },
		{ 231, 310, 16 },
		{ 231, 311, 16 },
		{ 231, 312, 16 },
		{ 231, 313, 16 },
		{ 231, 314, 16 },
		{ 231, 315, 16 },
		{ 231, 316, 16 },
		{ 231, 317, 16 },
		{ 231, 318, 16 },
		{ 231, 319, 16 },
		{ 231, 320, 16 },
		{ 231, 321, 16 },
		{ 231, 322, 16 },
		{ 231, 323, 16 },
		{ 231, 324, 16 },
		{ 231, 325, 16 },
		{ 231, 326, 16 },
		{ 231, 327, 16 },
		{ 231, 328, 16 },
		{ 231, 329, 16 },
		{ 231, 330, 16 },
		{ 231, 331, 16 },
		{ 231, 332, 16 },
		{ 231, 333, 16 },
		{ 231, 334, 16 },
		{ 231, 335, 16 },
		{ 231, 336, 16 },
		{ 231, 337, 16 },
		{ 231, 338, 16 },
		{ 231, 339, 16 },
		{ 231, 340, 16 },
		{ 231, 341, 16 },
		{ 231, 342, 16 },
		{ 231, 343, 16 },
		{ 231, 344, 16 },
		{ 231, 345, 16 },
		{ 231, 346, 16 },
		{ 231, 347, 16 },
		{ 231, 348, 16 },
		{ 231, 349, 16 },
		{ 231, 350, 16 },
		{ 231, 351, 16 },
		{ 231, 352, 16 },
		{ 231, 353, 16 },
		{ 231, 354, 16 },
		{ 231, 355, 16 },
		{ 231, 356, 16 },
		{ 231, 357, 16 },
		{ 231, 358, 16 },
		{ 231, 359, 16 },
		{ 231, 360, 16 },
		{ 232, 291, 16 },
		{ 232, 292, 16 },
		{ 232, 293, 16 },
		{ 232, 294, 16 },
		{ 232, 295, 16 },
		{ 232, 296, 16 },
		{ 232, 297, 16 },
		{ 232, 298, 16 },
		{ 232, 299, 16 },
		{ 232, 300, 16 },
		{ 232, 301, 16 },
		{ 232, 302, 16 },
		{ 232, 303, 16 },
		{ 232, 304, 16 },
		{ 232, 305, 16 },
		{ 232, 306, 16 },
		{ 232, 307, 16 },
		{ 232, 308, 16 },
		{ 232, 309, 16 },
		{ 232, 310, 16 },
		{ 232, 311, 16 },
		{ 232, 312, 16 },
		{ 232, 313, 16 },
		{ 232, 314, 16 },
		{ 232, 315, 16 },
		{ 232, 316, 16 },
		{ 232, 317, 16 },
		{ 232, 318, 16 },
		{ 232, 319, 16 },
		{ 232, 320, 16 },
		{ 232, 321, 16 },
		{ 232, 322, 16 },
		{ 232, 323, 16 },
		{ 232, 324, 16 },
		{ 232, 325, 16 },
		{ 232, 326, 16 },
		{ 232, 327, 16 },
		{ 232, 328, 16 },
		{ 232, 329, 16 },
		{ 232, 330, 16 },
		{ 232, 331, 16 },
		{ 232, 332, 16 },
		{ 232, 333, 16 },
		{ 232, 334, 16 },
		{ 232, 335, 16 },
		{ 232, 336, 16 },
		{ 232, 337, 16 },
		{ 232, 338, 16 },
		{ 232, 339, 16 },
		{ 232, 340, 16 },
		{ 232, 341, 16 },
		{ 232, 342, 16 },
		{ 232, 343, 16 },
		{ 232, 344, 16 },
		{ 232, 345, 16 },
		{ 232, 346, 16 },
		{ 232, 347, 16 },
		{ 232, 348, 16 },
		{ 232, 349, 16 },
		{ 232, 350, 16 },
		{ 232, 351, 16 },
		{ 232, 352, 16 },
		{ 232, 353, 16 },
		{ 232, 354, 16 },
		{ 232, 355, 16 },
		{ 232, 356, 16 },
		{ 233, 290, 16 },
		{ 233, 291, 16 },
		{ 233, 292, 16 },
		{ 233, 293, 16 },
		{ 233, 294, 16 },
		{ 233, 295, 16 },
		{ 233, 296, 16 },
		{ 233, 297, 16 },
		{ 233, 298, 16 },
		{ 233, 299, 16 },
		{ 233, 300, 16 },
		{ 233, 301, 16 },
		{ 233, 302, 16 },
		{ 233, 303, 16 },
		{ 233, 304, 16 },
		{ 233, 305, 16 },
		{ 233, 306, 16 },
		{ 233, 307, 16 },
		{ 233, 308, 16 },
		{ 233, 309, 16 },
		{ 233, 310, 16 },
		{ 233, 311, 16 },
		{ 233, 312, 16 },
		{ 233, 313, 16 },
		{ 233, 314, 16 },
		{ 233, 315, 16 },
		{ 233, 316, 16 },
		{ 233, 317, 16 },
		{ 233, 318, 16 },
		{ 233, 319, 16 },
		{ 233, 320, 16 },
		{ 233, 321, 16 },
		{ 233, 322, 16 },
		{ 233, 323, 16 },
		{ 233, 324, 16 },
		{ 233, 325, 16 },
		{ 233, 326, 16 },
		{ 233, 327, 16 },
		{ 233, 328, 16 },
		{ 233, 329, 16 },
		{ 233, 330, 16 },
		{ 233, 331, 16 },
		{ 233, 332, 16 },
		{ 233, 333, 16 },
		{ 233, 334, 16 },
		{ 233, 335, 16 },
		{ 233, 336, 16 },
		{ 233, 337, 16 },
		{ 233, 338, 16 },
		{ 233, 339, 16 },
		{ 233, 340, 16 },
		{ 233, 341, 16 },
		{ 233, 342, 16 },
		{ 233, 343, 16 },
		{ 233, 344, 16 },
		{ 233, 345, 16 },
		{ 233, 346, 16 },
		{ 233, 347, 16 },
		{ 233, 348, 16 },
		{ 233, 349, 16 },
		{ 233, 350, 16 },
		{ 233, 351, 16 },
		{ 233, 352, 16 },
		{ 234, 288, 16 },
		{ 234, 289, 16 },
		{ 234, 290, 16 },
		{ 234, 291, 16 },
		{ 234, 292, 16 },
		{ 234, 293, 16 },
		{ 234, 294, 16 },
		{ 234, 295, 16 },
		{ 234, 296, 16 },
		{ 234, 297, 16 },
		{ 234, 298, 16 },
		{ 234, 299, 16 },
		{ 234, 300, 16 },
		{ 234, 301, 16 },
		{ 234, 302, 16 },
		{ 234, 303, 16 },
		{ 234, 304, 16 },
		{ 234, 305, 16 },
		{ 234, 306, 16 },
		{ 234, 307, 16 },
		{ 234, 308, 16 },
		{ 234, 309, 16 },
		{ 234, 310, 16 },
		{ 234, 311, 16 },
		{ 234, 312, 16 },
		{ 234, 313, 16 },
		{ 234, 314, 16 },
		{ 234, 315, 16 },
		{ 234, 316, 16 },
		{ 234, 317, 16 },
		{ 234, 318, 16 },
		{ 234, 319, 16 },
		{ 234, 320, 16 },
		{ 234, 321, 16 },
		{ 234, 322, 16 },
		{ 234, 323, 16 },
		{ 234, 324, 16 },
		{ 234, 325, 16 },
		{ 234, 326, 16 },
		{ 234, 327, 16 },
		{ 234, 328, 16 },
		{ 234, 329, 16 },
		{ 234, 330, 16 },
		{ 234, 331, 16 },
		{ 234, 332, 16 },
		{ 234, 333, 16 },
		{ 234, 334, 16 },
		{ 234, 335, 16 },
		{ 234, 336, 16 },
		{ 234, 337, 16 },
		{ 234, 338, 16 },
		{ 234, 339, 16 },
		{ 234, 340, 16 },
		{ 234, 341, 16 },
		{ 234, 342, 16 },
		{ 234, 343, 16 },
		{ 234, 344, 16 },
		{ 234, 345, 16 },
		{ 234, 346, 16 },
		{ 234, 347, 16 },
		{ 234, 348, 16 },
		{ 235, 287, 16 },
		{ 235, 288, 16 },
		{ 235, 289, 16 },
		{ 235, 290, 16 },
		{ 235, 291, 16 },
		{ 235, 292, 16 },
		{ 235, 293, 16 },
		{ 235, 294, 16 },
		{ 235, 295, 16 },
		{ 235, 296, 16 },
		{ 235, 297, 16 },
		{ 235, 298, 16 },
		{ 235, 299, 16 },
		{ 235, 300, 16 },
		{ 235, 301, 16 },
		{ 235, 302, 16 },
		{ 235, 303, 16 },
		{ 235, 304, 16 },
		{ 235, 305, 16 },
		{ 235, 306, 16 },
		{ 235, 307, 16 },
		{ 235, 308, 16 },
		{ 235, 309, 16 },
		{ 235, 310, 16 },
		{ 235, 311, 16 },
		{ 235, 312, 16 },
		{ 235, 313, 16 },
		{ 235, 314, 16 },
		{ 235, 315, 16 },
		{ 235, 316, 16 },
		{ 235, 317, 16 },
		{ 235, 318, 16 },
		{ 235, 319, 16 },
		{ 235, 320, 16 },
		{ 235, 321, 16 },
		{ 235, 322, 16 },
		{ 235, 323, 16 },
		{ 235, 324, 16 },
		{ 235, 325, 16 },
		{ 235, 326, 16 },
		{ 235, 327, 16 },
		{ 235, 328, 16 },
		{ 235, 329, 16 },
		{ 235, 330, 16 },
		{ 235, 331, 16 },
		{ 235, 332, 16 },
		{ 235, 333, 16 },
		{ 235, 334, 16 },
		{ 235, 335, 16 },
		{ 235, 336, 16 },
		{ 235, 337, 16 },
		{ 235, 338, 16 },
		{ 235, 339, 16 },
		{ 235, 340, 16 },
		{ 235, 341, 16 },
		{ 235, 342, 16 },
		{ 235, 343, 16 },
		{ 235, 344, 16 },
		{ 236, 286, 16 },
		{ 236, 287, 16 },
		{ 236, 288, 16 },
		{ 236, 289, 16 },
		{ 236, 290, 16 },
		{ 236, 291, 16 },
		{ 236, 292, 16 },
		{ 236, 293, 16 },
		{ 236, 294, 16 },
		{ 236, 295, 16 },
		{ 236, 296, 16 },
		{ 236, 297, 16 },
		{ 236, 298, 16 },
		{ 236, 299, 16 },
		{ 236, 300, 16 },
		{ 236, 301, 16 },
		{ 236, 302, 16 },
		{ 236, 303, 16 },
		{ 236, 304, 16 },
		{ 236, 305, 16 },
		{ 236, 306, 16 },
		{ 236, 307, 16 },
		{ 236, 308, 16 },
		{ 236, 309, 16 },
		{ 236, 310, 16 },
		{ 236, 311, 16 },
		{ 236, 312, 16 },
		{ 236, 313, 16 },
		{ 236, 314, 16 },
		{ 236, 315, 16 },
		{ 236, 316, 16 },
		{ 236, 317, 16 },
		{ 236, 318, 16 },
		{ 236, 319, 16 },
		{ 236, 320, 16 },
		{ 236, 321, 16 },
		{ 236, 322, 16 },
		{ 236, 323, 16 },
		{ 236, 324, 16 },
		{ 236, 325, 16 },
		{ 236, 326, 16 },
		{ 236, 327, 16 },
		{ 236, 328, 16 },
		{ 236, 329, 16 },
		{ 236, 330, 16 },
		{ 236, 331, 16 },
		{ 236, 332, 16 },
		{ 236, 333, 16 },
		{ 236, 334, 16 },
		{ 236, 335, 16 },
		{ 236, 336, 16 },
		{ 236, 337, 16 },
		{ 236, 338, 16 },
		{ 236, 339, 16 },
		{ 236, 340, 16 },
		{ 237, 284, 16 },
		{ 237, 285, 16 },
		{ 237, 286, 16 },
		{ 237, 287, 16 },
		{ 237, 288, 16 },
		{ 237, 289, 16 },
		{ 237, 290, 16 },
		{ 237, 291, 16 },
		{ 237, 292, 16 },
		{ 237, 293, 16 },
		{ 237, 294, 16 },
		{ 237, 295, 16 },
		{ 237, 296, 16 },
		{ 237, 297, 16 },
		{ 237, 298, 16 },
		{ 237, 299, 16 },
		{ 237, 300, 16 },
		{ 237, 301, 16 },
		{ 237, 302, 16 },
		{ 237, 303, 16 },
		{ 237, 304, 16 },
		{ 237, 305, 16 },
		{ 237, 306, 16 },
		{ 237, 307, 16 },
		{ 237, 308, 16 },
		{ 237, 309, 16 },
		{ 237, 310, 16 },
		{ 237, 311, 16 },
		{ 237, 312, 16 },
		{ 237, 313, 16 },
		{ 237, 314, 16 },
		{ 237, 315, 16 },
		{ 237, 316, 16 },
		{ 237, 317, 16 },
		{ 237, 318, 16 },
		{ 237, 319, 16 },
		{ 237, 320, 16 },
		{ 237, 321, 16 },
		{ 237, 322, 16 },
		{ 237, 323, 16 },
		{ 237, 324, 16 },
		{ 237, 325, 16 },
		{ 237, 326, 16 },
		{ 237, 327, 16 },
		{ 237, 328, 16 },
		{ 237, 329, 16 },
		{ 237, 330, 16 },
		{ 237, 331, 16 },
		{ 237, 332, 16 },
		{ 237, 333, 16 },
		{ 237, 334, 16 },
		{ 237, 335, 16 },
		{ 237, 336, 16 },
		{ 238, 283, 16 },
		{ 238, 284, 16 },
		{ 238, 285, 16 },
		{ 238, 286, 16 },
		{ 238, 287, 16 },
		{ 238, 288, 16 },
		{ 238, 289, 16 },
		{ 238, 290, 16 },
		{ 238, 291, 16 },
		{ 238, 292, 16 },
		{ 238, 293, 16 },
		{ 238, 294, 16 },
		{ 238, 295, 16 },
		{ 238, 296, 16 },
		{ 238, 297, 16 },
		{ 238, 298, 16 },
		{ 238, 299, 16 },
		{ 238, 300, 16 },
		{ 238, 301, 16 },
		{ 238, 302, 16 },
		{ 238, 303, 16 },
		{ 238, 304, 16 },
		{ 238, 305, 16 },
		{ 238, 306, 16 },
		{ 238, 307, 16 },
		{ 238, 308, 16 },
		{ 238, 309, 16 },
		{ 238, 310, 16 },
		{ 238, 311, 16 },
		{ 238, 312, 16 },
		{ 238, 313, 16 },
		{ 238, 314, 16 },
		{ 238, 315, 16 },
		{ 238, 316, 16 },
		{ 238, 317, 16 },
		{ 238, 318, 16 },
		{ 238, 319, 16 },
		{ 238, 320, 16 },
		{ 238, 321, 16 },
		{ 238, 322, 16 },
		{ 238, 323, 16 },
		{ 238, 324, 16 },
		{ 238, 325, 16 },
		{ 238, 326, 16 },
		{ 238, 327, 16 },
		{ 238, 328, 16 },
		{ 238, 329, 16 },
		{ 238, 330, 16 },
		{ 238, 331, 16 },
		{ 238, 332, 16 },
		{ 239, 282, 16 },
		{ 239, 283, 16 },
		{ 239, 284, 16 },
		{ 239, 285, 16 },
		{ 239, 286, 16 },
		{ 239, 287, 16 },
		{ 239, 288, 16 },
		{ 239, 289, 16 },
		{ 239, 290, 16 },
		{ 239, 291, 16 },
		{ 239, 292, 16 },
		{ 239, 293, 16 },
		{ 239, 294, 16 },
		{ 239, 295, 16 },
		{ 239, 296, 16 },
		{ 239, 297, 16 },
		{ 239, 298, 16 },
		{ 239, 299, 16 },
		{ 239, 300, 16 },
		{ 239, 301, 16 },
		{ 239, 302, 16 },
		{ 239, 303, 16 },
		{ 239, 304, 16 },
		{ 239, 305, 16 },
		{ 239, 306, 16 },
		{ 239, 307, 16 },
		{ 239, 308, 16 },
		{ 239, 309, 16 },
		{ 239, 310, 16 },
		{ 239, 311, 16 },
		{ 239, 312, 16 },
		{ 239, 313, 16 },
		{ 239, 314, 16 },
		{ 239, 315, 16 },
		{ 239, 316, 16 },
		{ 239, 317, 16 },
		{ 239, 318, 16 },
		{ 239, 319, 16 },
		{ 239, 320, 16 },
		{ 239, 321, 16 },
		{ 239, 322, 16 },
		{ 239, 323, 16 },
		{ 239, 324, 16 },
		{ 239, 325, 16 },
		{ 239, 326, 16 },
		{ 239, 327, 16 },
		{ 239, 328, 16 },
		{ 240, 280, 16 },
		{ 240, 281, 16 },
		{ 240, 282, 16 },
		{ 240, 283, 16 },
		{ 240, 284, 16 },
		{ 240, 285, 16 },
		{ 240, 286, 16 },
		{ 240, 287, 16 },
		{ 240, 288, 16 },
		{ 240, 289, 16 },
		{ 240, 290, 16 },
		{ 240, 291, 16 },
		{ 240, 292, 16 },
		{ 240, 293, 16 },
		{ 240, 294, 16 },
		{ 240, 295, 16 },
		{ 240, 296, 16 },
		{ 240, 297, 16 },
		{ 240, 298, 16 },
		{ 240, 299, 16 },
		{ 240, 300, 16 },
		{ 240, 301, 16 },
		{ 240, 302, 16 },
		{ 240, 303, 16 },
		{ 240, 304, 16 },
		{ 240, 305, 16 },
		{ 240, 306, 16 },
		{ 240, 307, 16 },
		{ 240, 308, 16 },
		{ 240, 309, 16 },
		{ 240, 310, 16 },
		{ 240, 311, 16 },
		{ 240, 312, 16 },
		{ 240, 313, 16 },
		{ 240, 314, 16 },
		{ 240, 315, 16 },
		{ 240, 316, 16 },
		{ 240, 317, 16 },
		{ 240, 318, 16 },
		{ 240, 319, 16 },
		{ 240, 320, 16 },
		{ 240, 321, 16 },
		{ 240, 322, 16 },
		{ 240, 323, 16 },
		{ 240, 324, 16 },
		{ 241, 279, 16 },
		{ 241, 280, 16 },
		{ 241, 281, 16 },
		{ 241, 282, 16 },
		{ 241, 283, 16 },
		{ 241, 284, 16 },
		{ 241, 285, 16 },
		{ 241, 286, 16 },
		{ 241, 287, 16 },
		{ 241, 288, 16 },
		{ 241, 289, 16 },
		{ 241, 290, 16 },
		{ 241, 291, 16 },
		{ 241, 292, 16 },
		{ 241, 293, 16 },
		{ 241, 294, 16 },
		{ 241, 295, 16 },
		{ 241, 296, 16 },
		{ 241, 297, 16 },
		{ 241, 298, 16 },
		{ 241, 299, 16 },
		{ 241, 300, 16 },
		{ 241, 301, 16 },
		{ 241, 302, 16 },
		{ 241, 303, 16 },
		{ 241, 304, 16 },
		{ 241, 305, 16 },
		{ 241, 306, 16 },
		{ 241, 307, 16 },
		{ 241, 308, 16 },
		{ 241, 309, 16 },
		{ 241, 310, 16 },
		{ 241, 311, 16 },
		{ 241, 312, 16 },
		{ 241, 313, 16 },
		{ 241, 314, 16 },
		{ 241, 315, 16 },
		{ 241, 316, 16 },
		{ 241, 317, 16 },
		{ 241, 318, 16 },
		{ 241, 319, 16 },
		{ 241, 320, 16 },
		{ 242, 278, 16 },
		{ 242, 279, 16 },
		{ 242, 280, 16 },
		{ 242, 281, 16 },
		{ 242, 282, 16 },
		{ 242, 283, 16 },
		{ 242, 284, 16 },
		{ 242, 285, 16 },
		{ 242, 286, 16 },
		{ 242, 287, 16 },
		{ 242, 288, 16 },
		{ 242, 289, 16 },
		{ 242, 290, 16 },
		{ 242, 291, 16 },
		{ 242, 292, 16 },
		{ 242, 293, 16 },
		{ 242, 294, 16 },
		{ 242, 295, 16 },
		{ 242, 296, 16 },
		{ 242, 297, 16 },
		{ 242, 298, 16 },
		{ 242, 299, 16 },
		{ 242, 300, 16 },
		{ 242, 301, 16 },
		{ 242, 302, 16 },
		{ 242, 303, 16 },
		{ 242, 304, 16 },
		{ 242, 305, 16 },
		{ 242, 306, 16 },
		{ 242, 307, 16 },
		{ 242, 308, 16 },
		{ 242, 309, 16 },
		{ 242, 310, 16 },
		{ 242, 311, 16 },
		{ 242, 312, 16 },
		{ 242, 313, 16 },
		{ 242, 314, 16 },
		{ 242, 315, 16 },
		{ 242, 316, 16 },
		{ 243, 276, 16 },
		{ 243, 277, 16 },
		{ 243, 278, 16 },
		{ 243, 279, 16 },
		{ 243, 280, 16 },
		{ 243, 281, 16 },
		{ 243, 282, 16 },
		{ 243, 283, 16 },
		{ 243, 284, 16 },
		{ 243, 285, 16 },
		{ 243, 286, 16 },
		{ 243, 287, 16 },
		{ 243, 288, 16 },
		{ 243, 289, 16 },
		{ 243, 290, 16 },
		{ 243, 291, 16 },
		{ 243, 292, 16 },
		{ 243, 293, 16 },
		{ 243, 294, 16 },
		{ 243, 295, 16 },
		{ 243, 296, 16 },
		{ 243, 297, 16 },
		{ 243, 298, 16 },
		{ 243, 299, 16 },
		{ 243, 300, 16 },
		{ 243, 301, 16 },
		{ 243, 302, 16 },
		{ 243, 303, 16 },
		{ 243, 304, 16 },
		{ 243, 305, 16 },
		{ 243, 306, 16 },
		{ 243, 307, 16 },
		{ 243, 308, 16 },
		{ 243, 309, 16 },
		{ 243, 310, 16 },
		{ 243, 311, 16 },
		{ 243, 312, 16 },
		{ 244, 275, 16 },
		{ 244, 276, 16 },
		{ 244, 277, 16 },
		{ 244, 278, 16 },
		{ 244, 279, 16 },
		{ 244, 280, 16 },
		{ 244, 281, 16 },
		{ 244, 282, 16 },
		{ 244, 283, 16 },
		{ 244, 284, 16 },
		{ 244, 285, 16 },
		{ 244, 286, 16 },
		{ 244, 287, 16 },
		{ 244, 288, 16 },
		{ 244, 289, 16 },
		{ 244, 290, 16 },
		{ 244, 291, 16 },
		{ 244, 292, 16 },
		{ 244, 293, 16 },
		{ 244, 294, 16 },
		{ 244, 295, 16 },
		{ 244, 296, 16 },
		{ 244, 297, 16 },
		{ 244, 298, 16 },
		{ 244, 299, 16 },
		{ 244, 300, 16 },
		{ 244, 301, 16 },
		{ 244, 302, 16 },
		{ 244, 303, 16 },
		{ 244, 304, 16 },
		{ 244, 305, 16 },
		{ 244, 306, 16 },
		{ 244, 307, 16 },
		{ 244, 308, 16 },
		{ 245, 274, 16 },
		{ 245, 275, 16 },
		{ 245, 276, 16 },
		{ 245, 277, 16 },
		{ 245, 278, 16 },
		{ 245, 279, 16 },
		{ 245, 280, 16 },
		{ 245, 281, 16 },
		{ 245, 282, 16 },
		{ 245, 283, 16 },
		{ 245, 284, 16 },
		{ 245, 285, 16 },
		{ 245, 286, 16 },
		{ 245, 287, 16 },
		{ 245, 288, 16 },
		{ 245, 289, 16 },
		{ 245, 290, 16 },
		{ 245, 291, 16 },
		{ 245, 292, 16 },
		{ 245, 293, 16 },
		{ 245, 294, 16 },
		{ 245, 295, 16 },
		{ 245, 296, 16 },
		{ 245, 297, 16 },
		{ 245, 298, 16 },
		{ 245, 299, 16 },
		{ 245, 300, 16 },
		{ 245, 301, 16 },
		{ 245, 302, 16 },
		{ 245, 303, 16 },
		{ 245, 304, 16 },
		{ 246, 274, 16 },
		{ 246, 275, 16 },
		{ 246, 276, 16 },
		{ 246, 277, 16 },
		{ 246, 278, 16 },
		{ 246, 279, 16 },
		{ 246, 280, 16 },
		{ 246, 281, 16 },
		{ 246, 282, 16 },
		{ 246, 283, 16 },
		{ 246, 284, 16 },
		{ 246, 285, 16 },
		{ 246, 286, 16 },
		{ 246, 287, 16 },
		{ 246, 288, 16 },
		{ 246, 289, 16 },
		{ 246, 290, 16 },
		{ 246, 291, 16 },
		{ 246, 292, 16 },
		{ 246, 293, 16 },
		{ 246, 294, 16 },
		{ 246, 295, 16 },
		{ 246, 296, 16 },
		{ 246, 297, 16 },
		{ 246, 298, 16 },
		{ 246, 299, 16 },
		{ 246, 300, 16 },
		{ 247, 274, 16 },
		{ 247, 275, 16 },
		{ 247, 276, 16 },
		{ 247, 277, 16 },
		{ 247, 278, 16 },
		{ 247, 279, 16 },
		{ 247, 280, 16 },
		{ 247, 281, 16 },
		{ 247, 282, 16 },
		{ 247, 283, 16 },
		{ 247, 284, 16 },
		{ 247, 285, 16 },
		{ 247, 286, 16 },
		{ 247, 287, 16 },
		{ 247, 288, 16 },
		{ 247, 289, 16 },
		{ 247, 290, 16 },
		{ 247, 291, 16 },
		{ 247, 292, 16 },
		{ 247, 293, 16 },
		{ 247, 294, 16 },
		{ 247, 295, 16 },
		{ 247, 296, 16 },
		{ 248, 274, 16 },
		{ 248, 275, 16 },
		{ 248, 276, 16 },
		{ 248, 277, 16 },
		{ 248, 278, 16 },
		{ 248, 279, 16 },
		{ 248, 280, 16 },
		{ 248, 281, 16 },
		{ 248, 282, 16 },
		{ 248, 283, 16 },
		{ 248, 284, 16 },
		{ 248, 285, 16 },
		{ 248, 286, 16 },
		{ 248, 287, 16 },
		{ 248, 288, 16 },
		{ 248, 289, 16 },
		{ 248, 290, 16 },
		{ 248, 291, 16 },
		{ 248, 292, 16 },
		{ 249, 274, 16 },
		{ 249, 275, 16 },
		{ 249, 276, 16 },
		{ 249, 277, 16 },
		{ 249, 278, 16 },
		{ 249, 279, 16 },
		{ 249, 280, 16 },
		{ 249, 281, 16 },
		{ 249, 282, 16 },
		{ 249, 283, 16 },
		{ 249, 284, 16 },
		{ 249, 285, 16 },
		{ 249, 286, 16 },
		{ 249, 287, 16 },
		{ 249, 288, 16 },
		{ 250, 274, 16 },
		{ 250, 275, 16 },
		{ 250, 276, 16 },
		{ 250, 277, 16 },
		{ 250, 278, 16 },
		{ 250, 279, 16 },
		{ 250, 280, 16 },
		{ 250, 281, 16 },
		{ 250, 282, 16 },
		{ 250, 283, 16 },
		{ 250, 284, 16 },
		{ 251, 274, 16 },
		{ 251, 275, 16 },
		{ 251, 276, 16 },
		{ 251, 277, 16 },
		{ 251, 278, 16 },
		{ 251, 279, 16 },
		{ 251, 280, 16 },
		{ 252, 274, 16 },
		{ 252, 275, 16 },
		{ 252, 276, 16 },
		{ 262, 238, 16 },
		{ 262, 239, 16 },
		{ 262, 240, 16 },
		{ 263, 234, 16 },
		{ 263, 235, 16 },
		{ 263, 236, 16 },
		{ 263, 237, 16 },
		{ 263, 238, 16 },
		{ 263, 239, 16 },
		{ 263, 240, 16 },
		{ 264, 230, 16 },
		{ 264, 231, 16 },
		{ 264, 232, 16 },
		{ 264, 233, 16 },
		{ 264, 234, 16 },
		{ 264, 235, 16 },
		{ 264, 236, 16 },
		{ 264, 237, 16 },
		{ 264, 238, 16 },
		{ 264, 239, 16 },
		{ 264, 240, 16 },
		{ 265, 226, 16 },
		{ 265, 227, 16 },
		{ 265, 228, 16 },
		{ 265, 229, 16 },
		{ 265, 230, 16 },
		{ 265, 231, 16 },
		{ 265, 232, 16 },
		{ 265, 233, 16 },
		{ 265, 234, 16 },
		{ 265, 235, 16 },
		{ 265, 236, 16 },
		{ 265, 237, 16 },
		{ 265, 238, 16 },
		{ 265, 239, 16 },
		{ 265, 240, 16 },
		{ 266, 222, 16 },
		{ 266, 223, 16 },
		{ 266, 224, 16 },
		{ 266, 225, 16 },
		{ 266, 226, 16 },
		{ 266, 227, 16 },
		{ 266, 228, 16 },
		{ 266, 229, 16 },
		{ 266, 230, 16 },
		{ 266, 231, 16 },
		{ 266, 232, 16 },
		{ 266, 233, 16 },
		{ 266, 234, 16 },
		{ 266, 235, 16 },
		{ 266, 236, 16 },
		{ 266, 237, 16 },
		{ 266, 238, 16 },
		{ 266, 239, 16 },
		{ 266, 240, 16 },
		{ 267, 218, 16 },
		{ 267, 219, 16 },
		{ 267, 220, 16 },
		{ 267, 221, 16 },
		{ 267, 222, 16 },
		{ 267, 223, 16 },
		{ 267, 224, 16 },
		{ 267, 225, 16 },
		{ 267, 226, 16 },
		{ 267, 227, 16 },
		{ 267, 228, 16 },
		{ 267, 229, 16 },
		{ 267, 230, 16 },
		{ 267, 231, 16 },
		{ 267, 232, 16 },
		{ 267, 233, 16 },
		{ 267, 234, 16 },
		{ 267, 235, 16 },
		{ 267, 236, 16 },
		{ 267, 237, 16 },
		{ 267, 238, 16 },
		{ 267, 239, 16 },
		{ 267, 240, 16 },
		{ 268, 214, 16 },
		{ 268, 215, 16 },
		{ 268, 216, 16 },
		{ 268, 217, 16 },
		{ 268, 218, 16 },
		{ 268, 219, 16 },
		{ 268, 220, 16 },
		{ 268, 221, 16 },
		{ 268, 222, 16 },
		{ 268, 223, 16 },
		{ 268, 224, 16 },
		{ 268, 225, 16 },
		{ 268, 226, 16 },
		{ 268, 227, 16 },
		{ 268, 228, 16 },
		{ 268, 229, 16 },
		{ 268, 230, 16 },
		{ 268, 231, 16 },
		{ 268, 232, 16 },
		{ 268, 233, 16 },
		{ 268, 234, 16 },
		{ 268, 235, 16 },
		{ 268, 236, 16 },
		{ 268, 237, 16 },
		{ 268, 238, 16 },
		{ 268, 239, 16 },
		{ 268, 240, 16 },
		{ 269, 210, 16 },
		{ 269, 211, 16 },
		{ 269, 212, 16 },
		{ 269, 213, 16 },
		{ 269, 214, 16 },
		{ 269, 215, 16 },
		{ 269, 216, 16 },
		{ 269, 217, 16 },
		{ 269, 218, 16 },
		{ 269, 219, 16 },
		{ 269, 220, 16 },
		{ 269, 221, 16 },
		{ 269, 222, 16 },
		{ 269, 223, 16 },
		{ 269, 224, 16 },
		{ 269, 225, 16 },
		{ 269, 226, 16 },
		{ 269, 227, 16 },
		{ 269, 228, 16 },
		{ 269, 229, 16 },
		{ 269, 230, 16 },
		{ 269, 231, 16 },
		{ 269, 232, 16 },
		{ 269, 233, 16 },
		{ 269, 234, 16 },
		{ 269, 235, 16 },
		{ 269, 236, 16 },
		{ 269, 237, 16 },
		{ 269, 238, 16 },
		{ 269, 239, 16 },
		{ 269, 240, 16 },
		{ 270, 206, 16 },
		{ 270, 207, 16 },
		{ 270, 208, 16 },
		{ 270, 209, 16 },
		{ 270, 210, 16 },
		{ 270, 211, 16 },
		{ 270, 212, 16 },
		{ 270, 213, 16 },
		{ 270, 214, 16 },
		{ 270, 215, 16 },
		{ 270, 216, 16 },
		{ 270, 217, 16 },
		{ 270, 218, 16 },
		{ 270, 219, 16 },
		{ 270, 220, 16 },
		{ 270, 221, 16 },
		{ 270, 222, 16 },
		{ 270, 223, 16 },
		{ 270, 224, 16 },
		{ 270, 225, 16 },
		{ 270, 226, 16 },
		{ 270, 227, 16 },
		{ 270, 228, 16 },
		{ 270, 229, 16 },
		{ 270, 230, 16 },
		{ 270, 231, 16 },
		{ 270, 232, 16 },
		{ 270, 233, 16 },
		{ 270, 234, 16 },
		{ 270, 235, 16 },
		{ 270, 236, 16 },
		{ 270, 237, 16 },
		{ 270, 238, 16 },
		{ 270, 239, 16 },
		{ 271, 202, 16 },
		{ 271, 203, 16 },
		{ 271, 204, 16 },
		{ 271, 205, 16 },
		{ 271, 206, 16 },
		{ 271, 207, 16 },
		{ 271, 208, 16 },
		{ 271, 209, 16 },
		{ 271, 210, 16 },
		{ 271, 211, 16 },
		{ 271, 212, 16 },
		{ 271, 213, 16 },
		{ 271, 214, 16 },
		{ 271, 215, 16 },
		{ 271, 216, 16 },
		{ 271, 217, 16 },
		{ 271, 218, 16 },
		{ 271, 219, 16 },
		{ 271, 220, 16 },
		{ 271, 221, 16 },
		{ 271, 222, 16 },
		{ 271, 223, 16 },
		{ 271, 224, 16 },
		{ 271, 225, 16 },
		{ 271, 226, 16 },
		{ 271, 227, 16 },
		{ 271, 228, 16 },
		{ 271, 229, 16 },
		{ 271, 230, 16 },
		{ 271, 231, 16 },
		{ 271, 232, 16 },
		{ 271, 233, 16 },
		{ 271, 234, 16 },
		{ 271, 235, 16 },
		{ 271, 236, 16 },
		{ 271, 237, 16 },
		{ 271, 238, 16 },
		{ 272, 198, 16 },
		{ 272, 199, 16 },
		{ 272, 200, 16 },
		{ 272, 201, 16 },
		{ 272, 202, 16 },
		{ 272, 203, 16 },
		{ 272, 204, 16 },
		{ 272, 205, 16 },
		{ 272, 206, 16 },
		{ 272, 207, 16 },
		{ 272, 208, 16 },
		{ 272, 209, 16 },
		{ 272, 210, 16 },
		{ 272, 211, 16 },
		{ 272, 212, 16 },
		{ 272, 213, 16 },
		{ 272, 214, 16 },
		{ 272, 215, 16 },
		{ 272, 216, 16 },
		{ 272, 217, 16 },
		{ 272, 218, 16 },
		{ 272, 219, 16 },
		{ 272, 220, 16 },
		{ 272, 221, 16 },
		{ 272, 222, 16 },
		{ 272, 223, 16 },
		{ 272, 224, 16 },
		{ 272, 225, 16 },
		{ 272, 226, 16 },
		{ 272, 227, 16 },
		{ 272, 228, 16 },
		{ 272, 229, 16 },
		{ 272, 230, 16 },
		{ 272, 231, 16 },
		{ 272, 232, 16 },
		{ 272, 233, 16 },
		{ 272, 234, 16 },
		{ 272, 235, 16 },
		{ 272, 236, 16 },
		{ 273, 194, 16 },
		{ 273, 195, 16 },
		{ 273, 196, 16 },
		{ 273, 197, 16 },
		{ 273, 198, 16 },
		{ 273, 199, 16 },
		{ 273, 200, 16 },
		{ 273, 201, 16 },
		{ 273, 202, 16 },
		{ 273, 203, 16 },
		{ 273, 204, 16 },
		{ 273, 205, 16 },
		{ 273, 206, 16 },
		{ 273, 207, 16 },
		{ 273, 208, 16 },
		{ 273, 209, 16 },
		{ 273, 210, 16 },
		{ 273, 211, 16 },
		{ 273, 212, 16 },
		{ 273, 213, 16 },
		{ 273, 214, 16 },
		{ 273, 215, 16 },
		{ 273, 216, 16 },
		{ 273, 217, 16 },
		{ 273, 218, 16 },
		{ 273, 219, 16 },
		{ 273, 220, 16 },
		{ 273, 221, 16 },
		{ 273, 222, 16 },
		{ 273, 223, 16 },
		{ 273, 224, 16 },
		{ 273, 225, 16 },
		{ 273, 226, 16 },
		{ 273, 227, 16 },
		{ 273, 228, 16 },
		{ 273, 229, 16 },
		{ 273, 230, 16 },
		{ 273, 231, 16 },
		{ 273, 232, 16 },
		{ 273, 233, 16 },
		{ 273, 234, 16 },
		{ 273, 235, 16 },
		{ 274, 190, 16 },
		{ 274, 191, 16 },
		{ 274, 192, 16 },
		{ 274, 193, 16 },
		{ 274, 194, 16 },
		{ 274, 195, 16 },
		{ 274, 196, 16 },
		{ 274, 197, 16 },
		{ 274, 198, 16 },
		{ 274, 199, 16 },
		{ 274, 200, 16 },
		{ 274, 201, 16 },
		{ 274, 202, 16 },
		{ 274, 203, 16 },
		{ 274, 204, 16 },
		{ 274, 205, 16 },
		{ 274, 206, 16 },
		{ 274, 207, 16 },
		{ 274, 208, 16 },
		{ 274, 209, 16 },
		{ 274, 210, 16 },
		{ 274, 211, 16 },
		{ 274, 212, 16 },
		{ 274, 213, 16 },
		{ 274, 214, 16 },
		{ 274, 215, 16 },
		{ 274, 216, 16 },
		{ 274, 217, 16 },
		{ 274, 218, 16 },
		{ 274, 219, 16 },
		{ 274, 220, 16 },
		{ 274, 221, 16 },
		{ 274, 222, 16 },
		{ 274, 223, 16 },
		{ 274, 224, 16 },
		{ 274, 225, 16 },
		{ 274, 226, 16 },
		{ 274, 227, 16 },
		{ 274, 228, 16 },
		{ 274, 229, 16 },
		{ 274, 230, 16 },
		{ 274, 231, 16 },
		{ 274, 232, 16 },
		{ 274, 233, 16 },
		{ 274, 234, 16 },
		{ 275, 186, 16 },
		{ 275, 187, 16 },
		{ 275, 188, 16 },
		{ 275, 189, 16 },
		{ 275, 190, 16 },
		{ 275, 191, 16 },
		{ 275, 192, 16 },
		{ 275, 193, 16 },
		{ 275, 194, 16 },
		{ 275, 195, 16 },
		{ 275, 196, 16 },
		{ 275, 197, 16 },
		{ 275, 198, 16 },
		{ 275, 199, 16 },
		{ 275, 200, 16 },
		{ 275, 201, 16 },
		{ 275, 202, 16 },
		{ 275, 203, 16 },
		{ 275, 204, 16 },
		{ 275, 205, 16 },
		{ 275, 206, 16 },
		{ 275, 207, 16 },
		{ 275, 208, 16 },
		{ 275, 209, 16 },
		{ 275, 210, 16 },
		{ 275, 211, 16 },
		{ 275, 212, 16 },
		{ 275, 213, 16 },
		{ 275, 214, 16 },
		{ 275, 215, 16 },
		{ 275, 216, 16 },
		{ 275, 217, 16 },
		{ 275, 218, 16 },
		{ 275, 219, 16 },
		{ 275, 220, 16 },
		{ 275, 221, 16 },
		{ 275, 222, 16 },
		{ 275, 223, 16 },
		{ 275, 224, 16 },
		{ 275, 225, 16 },
		{ 275, 226, 16 },
		{ 275, 227, 16 },
		{ 275, 228, 16 },
		{ 275, 229, 16 },
		{ 275, 230, 16 },
		{ 275, 231, 16 },
		{ 275, 232, 16 },
		{ 276, 182, 16 },
		{ 276, 183, 16 },
		{ 276, 184, 16 },
		{ 276, 185, 16 },
		{ 276, 186, 16 },
		{ 276, 187, 16 },
		{ 276, 188, 16 },
		{ 276, 189, 16 },
		{ 276, 190, 16 },
		{ 276, 191, 16 },
		{ 276, 192, 16 },
		{ 276, 193, 16 },
		{ 276, 194, 16 },
		{ 276, 195, 16 },
		{ 276, 196, 16 },
		{ 276, 197, 16 },
		{ 276, 198, 16 },
		{ 276, 199, 16 },
		{ 276, 200, 16 },
		{ 276, 201, 16 },
		{ 276, 202, 16 },
		{ 276, 203, 16 },
		{ 276, 204, 16 },
		{ 276, 205, 16 },
		{ 276, 206, 16 },
		{ 276, 207, 16 },
		{ 276, 208, 16 },
		{ 276, 209, 16 },
		{ 276, 210, 16 },
		{ 276, 211, 16 },
		{ 276, 212, 16 },
		{ 276, 213, 16 },
		{ 276, 214, 16 },
		{ 276, 215, 16 },
		{ 276, 216, 16 },
		{ 276, 217, 16 },
		{ 276, 218, 16 },
		{ 276, 219, 16 },
		{ 276, 220, 16 },
		{ 276, 221, 16 },
		{ 276, 222, 16 },
		{ 276, 223, 16 },
		{ 276, 224, 16 },
		{ 276, 225, 16 },
		{ 276, 226, 16 },
		{ 276, 227, 16 },
		{ 276, 228, 16 },
		{ 276, 229, 16 },
		{ 276, 230, 16 },
		{ 276, 231, 16 },
		{ 277, 178, 16 },
		{ 277, 179, 16 },
		{ 277, 180, 16 },
		{ 277, 181, 16 },
		{ 277, 182, 16 },
		{ 277, 183, 16 },
		{ 277, 184, 16 },
		{ 277, 185, 16 },
		{ 277, 186, 16 },
		{ 277, 187, 16 },
		{ 277, 188, 16 },
		{ 277, 189, 16 },
		{ 277, 190, 16 },
		{ 277, 191, 16 },
		{ 277, 192, 16 },
		{ 277, 193, 16 },
		{ 277, 194, 16 },
		{ 277, 195, 16 },
		{ 277, 196, 16 },
		{ 277, 197, 16 },
		{ 277, 198, 16 },
		{ 277, 199, 16 },
		{ 277, 200, 16 },
		{ 277, 201, 16 },
		{ 277, 202, 16 },
		{ 277, 203, 16 },
		{ 277, 204, 16 },
		{ 277, 205, 16 },
		{ 277, 206, 16 },
		{ 277, 207, 16 },
		{ 277, 208, 16 },
		{ 277, 209, 16 },
		{ 277, 210, 16 },
		{ 277, 211, 16 },
		{ 277, 212, 16 },
		{ 277, 213, 16 },
		{ 277, 214, 16 },
		{ 277, 215, 16 },
		{ 277, 216, 16 },
		{ 277, 217, 16 },
		{ 277, 218, 16 },
		{ 277, 219, 16 },
		{ 277, 220, 16 },
		{ 277, 221, 16 },
		{ 277, 222, 16 },
		{ 277, 223, 16 },
		{ 277, 224, 16 },
		{ 277, 225, 16 },
		{ 277, 226, 16 },
		{ 277, 227, 16 },
		{ 277, 228, 16 },
		{ 277, 229, 16 },
		{ 277, 230, 16 },
		{ 278, 174, 16 },
		{ 278, 175, 16 },
		{ 278, 176, 16 },
		{ 278, 177, 16 },
		{ 278, 178, 16 },
		{ 278, 179, 16 },
		{ 278, 180, 16 },
		{ 278, 181, 16 },
		{ 278, 182, 16 },
		{ 278, 183, 16 },
		{ 278, 184, 16 },
		{ 278, 185, 16 },
		{ 278, 186, 16 },
		{ 278, 187, 16 },
		{ 278, 188, 16 },
		{ 278, 189, 16 },
		{ 278, 190, 16 },
		{ 278, 191, 16 },
		{ 278, 192, 16 },
		{ 278, 193, 16 },
		{ 278, 194, 16 },
		{ 278, 195, 16 },
		{ 278, 196, 16 },
		{ 278, 197, 16 },
		{ 278, 198, 16 },
		{ 278, 199, 16 },
		{ 278, 200, 16 },
		{ 278, 201, 16 },
		{ 278, 202, 16 },
		{ 278, 203, 16 },
		{ 278, 204, 16 },
		{ 278, 205, 16 },
		{ 278, 206, 16 },
		{ 278, 207, 16 },
		{ 278, 208, 16 },
		{ 278, 209, 16 },
		{ 278, 210, 16 },
		{ 278, 211, 16 },
		{ 278, 212, 16 },
		{ 278, 213, 16 },
		{ 278, 214, 16 },
		{ 278, 215, 16 },
		{ 278, 216, 16 },
		{ 278, 217, 16 },
		{ 278, 218, 16 },
		{ 278, 219, 16 },
		{ 278, 220, 16 },
		{ 278, 221, 16 },
		{ 278, 222, 16 },
		{ 278, 223, 16 },
		{ 278, 224, 16 },
		{ 278, 225, 16 },
		{ 278, 226, 16 },
		{ 278, 227, 16 },
		{ 278, 228, 16 },
		{ 279, 170, 16 },
		{ 279, 171, 16 },
		{ 279, 172, 16 },
		{ 279, 173, 16 },
		{ 279, 174, 16 },
		{ 279, 175, 16 },
		{ 279, 176, 16 },
		{ 279, 177, 16 },
		{ 279, 178, 16 },
		{ 279, 179, 16 },
		{ 279, 180, 16 },
		{ 279, 181, 16 },
		{ 279, 182, 16 },
		{ 279, 183, 16 },
		{ 279, 184, 16 },
		{ 279, 185, 16 },
		{ 279, 186, 16 },
		{ 279, 187, 16 },
		{ 279, 188, 16 },
		{ 279, 189, 16 },
		{ 279, 190, 16 },
		{ 279, 191, 16 },
		{ 279, 192, 16 },
		{ 279, 193, 16 },
		{ 279, 194, 16 },
		{ 279, 195, 16 },
		{ 279, 196, 16 },
		{ 279, 197, 16 },
		{ 279, 198, 16 },
		{ 279, 199, 16 },
		{ 279, 200, 16 },
		{ 279, 201, 16 },
		{ 279, 202, 16 },
		{ 279, 203, 16 },
		{ 279, 204, 16 },
		{ 279, 205, 16 },
		{ 279, 206, 16 },
		{ 279, 207, 16 },
		{ 279, 208, 16 },
		{ 279, 209, 16 },
		{ 279, 210, 16 },
		{ 279, 211, 16 },
		{ 279, 212, 16 },
		{ 279, 213, 16 },
		{ 279, 214, 16 },
		{ 279, 215, 16 },
		{ 279, 216, 16 },
		{ 279, 217, 16 },
		{ 279, 218, 16 },
		{ 279, 219, 16 },
		{ 279, 220, 16 },
		{ 279, 221, 16 },
		{ 279, 222, 16 },
		{ 279, 223, 16 },
		{ 279, 224, 16 },
		{ 279, 225, 16 },
		{ 279, 226, 16 },
		{ 279, 227, 16 },
		{ 280, 166, 16 },
		{ 280, 167, 16 },
		{ 280, 168, 16 },
		{ 280, 169, 16 },
		{ 280, 170, 16 },
		{ 280, 171, 16 },
		{ 280, 172, 16 },
		{ 280, 173, 16 },
		{ 280, 174, 16 },
		{ 280, 175, 16 },
		{ 280, 176, 16 },
		{ 280, 177, 16 },
		{ 280, 178, 16 },
		{ 280, 179, 16 },
		{ 280, 180, 16 },
		{ 280, 181, 16 },
		{ 280, 182, 16 },
		{ 280, 183, 16 },
		{ 280, 184, 16 },
		{ 280, 185, 16 },
		{ 280, 186, 16 },
		{ 280, 187, 16 },
		{ 280, 188, 16 },
		{ 280, 189, 16 },
		{ 280, 190, 16 },
		{ 280, 191, 16 },
		{ 280, 192, 16 },
		{ 280, 193, 16 },
		{ 280, 194, 16 },
		{ 280, 195, 16 },
		{ 280, 196, 16 },
		{ 280, 197, 16 },
		{ 280, 198, 16 },
		{ 280, 199, 16 },
		{ 280, 200, 16 },
		{ 280, 201, 16 },
		{ 280, 202, 16 },
		{ 280, 203, 16 },
		{ 280, 204, 16 },
		{ 280, 205, 16 },
		{ 280, 206, 16 },
		{ 280, 207, 16 },
		{ 280, 208, 16 },
		{ 280, 209, 16 },
		{ 280, 210, 16 },
		{ 280, 211, 16 },
		{ 280, 212, 16 },
		{ 280, 213, 16 },
		{ 280, 214, 16 },
		{ 280, 215, 16 },
		{ 280, 216, 16 },
		{ 280, 217, 16 },
		{ 280, 218, 16 },
		{ 280, 219, 16 },
		{ 280, 220, 16 },
		{ 280, 221, 16 },
		{ 280, 222, 16 },
		{ 280, 223, 16 },
		{ 280, 224, 16 },
		{ 280, 225, 16 },
		{ 280, 226, 16 },
		{ 281, 162, 16 },
		{ 281, 163, 16 },
		{ 281, 164, 16 },
		{ 281, 165, 16 },
		{ 281, 166, 16 },
		{ 281, 167, 16 },
		{ 281, 168, 16 },
		{ 281, 169, 16 },
		{ 281, 170, 16 },
		{ 281, 171, 16 },
		{ 281, 172, 16 },
		{ 281, 173, 16 },
		{ 281, 174, 16 },
		{ 281, 175, 16 },
		{ 281, 176, 16 },
		{ 281, 177, 16 },
		{ 281, 178, 16 },
		{ 281, 179, 16 },
		{ 281, 180, 16 },
		{ 281, 181, 16 },
		{ 281, 182, 16 },
		{ 281, 183, 16 },
		{ 281, 184, 16 },
		{ 281, 185, 16 },
		{ 281, 186, 16 },
		{ 281, 187, 16 },
		{ 281, 188, 16 },
		{ 281, 189, 16 },
		{ 281, 190, 16 },
		{ 281, 191, 16 },
		{ 281, 192, 16 },
		{ 281, 193, 16 },
		{ 281, 194, 16 },
		{ 281, 195, 16 },
		{ 281, 196, 16 },
		{ 281, 197, 16 },
		{ 281, 198, 16 },
		{ 281, 199, 16 },
		{ 281, 200, 16 },
		{ 281, 201, 16 },
		{ 281, 202, 16 },
		{ 281, 203, 16 },
		{ 281, 204, 16 },
		{ 281, 205, 16 },
		{ 281, 206, 16 },
		{ 281, 207, 16 },
		{ 281, 208, 16 },
		{ 281, 209, 16 },
		{ 281, 210, 16 },
		{ 281, 211, 16 },
		{ 281, 212, 16 },
		{ 281, 213, 16 },
		{ 281, 214, 16 },
		{ 281, 215, 16 },
		{ 281, 216, 16 },
		{ 281, 217, 16 },
		{ 281, 218, 16 },
		{ 281, 219, 16 },
		{ 281, 220, 16 },
		{ 281, 221, 16 },
		{ 281, 222, 16 },
		{ 281, 223, 16 },
		{ 281, 224, 16 },
		{ 282, 158, 16 },
		{ 282, 159, 16 },
		{ 282, 160, 16 },
		{ 282, 161, 16 },
		{ 282, 162, 16 },
		{ 282, 163, 16 },
		{ 282, 164, 16 },
		{ 282, 165, 16 },
		{ 282, 166, 16 },
		{ 282, 167, 16 },
		{ 282, 168, 16 },
		{ 282, 169, 16 },
		{ 282, 170, 16 },
		{ 282, 171, 16 },
		{ 282, 172, 16 },
		{ 282, 173, 16 },
		{ 282, 174, 16 },
		{ 282, 175, 16 },
		{ 282, 176, 16 },
		{ 282, 177, 16 },
		{ 282, 178, 16 },
		{ 282, 179, 16 },
		{ 282, 180, 16 },
		{ 282, 181, 16 },
		{ 282, 182, 16 },
		{ 282, 183, 16 },
		{ 282, 184, 16 },
		{ 282, 185, 16 },
		{ 282, 186, 16 },
		{ 282, 187, 16 },
		{ 282, 188, 16 },
		{ 282, 189, 16 },
		{ 282, 190, 16 },
		{ 282, 191, 16 },
		{ 282, 192, 16 },
		{ 282, 193, 16 },
		{ 282, 194, 16 },
		{ 282, 195, 16 },
		{ 282, 196, 16 },
		{ 282, 197, 16 },
		{ 282, 198, 16 },
		{ 282, 199, 16 },
		{ 282, 200, 16 },
		{ 282, 201, 16 },
		{ 282, 202, 16 },
		{ 282, 203, 16 },
		{ 282, 204, 16 },
		{ 282, 205, 16 },
		{ 282, 206, 16 },
		{ 282, 207, 16 },
		{ 282, 208, 16 },
		{ 282, 209, 16 },
		{ 282, 210, 16 },
		{ 282, 211, 16 },
		{ 282, 212, 16 },
		{ 282, 213, 16 },
		{ 282, 214, 16 },
		{ 282, 215, 16 },
		{ 282, 216, 16 },
		{ 282, 217, 16 },
		{ 282, 218, 16 },
		{ 282, 219, 16 },
		{ 282, 220, 16 },
		{ 282, 221, 16 },
		{ 282, 222, 16 },
		{ 282, 223, 16 },
		{ 283, 154, 16 },
		{ 283, 155, 16 },
		{ 283, 156, 16 },
		{ 283, 157, 16 },
		{ 283, 158, 16 },
		{ 283, 159, 16 },
		{ 283, 160, 16 },
		{ 283, 161, 16 },
		{ 283, 162, 16 },
		{ 283, 163, 16 },
		{ 283, 164, 16 },
		{ 283, 165, 16 },
		{ 283, 166, 16 },
		{ 283, 167, 16 },
		{ 283, 168, 16 },
		{ 283, 169, 16 },
		{ 283, 170, 16 },
		{ 283, 171, 16 },
		{ 283, 172, 16 },
		{ 283, 173, 16 },
		{ 283, 174, 16 },
		{ 283, 175, 16 },
		{ 283, 176, 16 },
		{ 283, 177, 16 },
		{ 283, 178, 16 },
		{ 283, 179, 16 },
		{ 283, 180, 16 },
		{ 283, 181, 16 },
		{ 283, 182, 16 },
		{ 283, 183, 16 },
		{ 283, 184, 16 },
		{ 283, 185, 16 },
		{ 283, 186, 16 },
		{ 283, 187, 16 },
		{ 283, 188, 16 },
		{ 283, 189, 16 },
		{ 283, 190, 16 },
		{ 283, 191, 16 },
		{ 283, 192, 16 },
		{ 283, 193, 16 },
		{ 283, 194, 16 },
		{ 283, 195, 16 },
		{ 283, 196, 16 },
		{ 283, 197, 16 },
		{ 283, 198, 16 },
		{ 283, 199, 16 },
		{ 283, 200, 16 },
		{ 283, 201, 16 },
		{ 283, 202, 16 },
		{ 283, 203, 16 },
		{ 283, 204, 16 },
		{ 283, 205, 16 },
		{ 283, 206, 16 },
		{ 283, 207, 16 },
		{ 283, 208, 16 },
		{ 283, 209, 16 },
		{ 283, 210, 16 },
		{ 283, 211, 16 },
		{ 283, 212, 16 },
		{ 283, 213, 16 },
		{ 283, 214, 16 },
		{ 283, 215, 16 },
		{ 283, 216, 16 },
		{ 283, 217, 16 },
		{ 283, 218, 16 },
		{ 283, 219, 16 },
		{ 283, 220, 16 },
		{ 283, 221, 16 },
		{ 283, 222, 16 },
		{ 284, 150, 16 },
		{ 284, 151, 16 },
		{ 284, 152, 16 },
		{ 284, 153, 16 },
		{ 284, 154, 16 },
		{ 284, 155, 16 },
		{ 284, 156, 16 },
		{ 284, 157, 16 },
		{ 284, 158, 16 },
		{ 284, 159, 16 },
		{ 284, 160, 16 },
		{ 284, 161, 16 },
		{ 284, 162, 16 },
		{ 284, 163, 16 },
		{ 284, 164, 16 },
		{ 284, 165, 16 },
		{ 284, 166, 16 },
		{ 284, 167, 16 },
		{ 284, 168, 16 },
		{ 284, 169, 16 },
		{ 284, 170, 16 },
		{ 284, 171, 16 },
		{ 284, 172, 16 },
		{ 284, 173, 16 },
		{ 284, 174, 16 },
		{ 284, 175, 16 },
		{ 284, 176, 16 },
		{ 284, 177, 16 },
		{ 284, 178, 16 },
		{ 284, 179, 16 },
		{ 284, 180, 16 },
		{ 284, 181, 16 },
		{ 284, 182, 16 },
		{ 284, 183, 16 },
		{ 284, 184, 16 },
		{ 284, 185, 16 },
		{ 284, 186, 16 },
		{ 284, 187, 16 },
		{ 284, 188, 16 },
		{ 284, 189, 16 },
		{ 284, 190, 16 },
		{ 284, 191, 16 },
		{ 284, 192, 16 },
		{ 284, 193, 16 },
		{ 284, 194, 16 },
		{ 284, 195, 16 },
		{ 284, 196, 16 },
		{ 284, 197, 16 },
		{ 284, 198, 16 },
		{ 284, 199, 16 },
		{ 284, 200, 16 },
		{ 284, 201, 16 },
		{ 284, 202, 16 },
		{ 284, 203, 16 },
		{ 284, 204, 16 },
		{ 284, 205, 16 },
		{ 284, 206, 16 },
		{ 284, 207, 16 },
		{ 284, 208, 16 },
		{ 284, 209, 16 },
		{ 284, 210, 16 },
		{ 284, 211, 16 },
		{ 284, 212, 16 },
		{ 284, 213, 16 },
		{ 284, 214, 16 },
		{ 284, 215, 16 },
		{ 284, 216, 16 },
		{ 284, 217, 16 },
		{ 284, 218, 16 },
		{ 284, 219, 16 },
		{ 284, 220, 16 },
		{ 285, 146, 16 },
		{ 285, 147, 16 },
		{ 285, 148, 16 },
		{ 285, 149, 16 },
		{ 285, 150, 16 },
		{ 285, 151, 16 },
		{ 285, 152, 16 },
		{ 285, 153, 16 },
		{ 285, 154, 16 },
		{ 285, 155, 16 },
		{ 285, 156, 16 },
		{ 285, 157, 16 },
		{ 285, 158, 16 },
		{ 285, 159, 16 },
		{ 285, 160, 16 },
		{ 285, 161, 16 },
		{ 285, 162, 16 },
		{ 285, 163, 16 },
		{ 285, 164, 16 },
		{ 285, 165, 16 },
		{ 285, 166, 16 },
		{ 285, 167, 16 },
		{ 285, 168, 16 },
		{ 285, 169, 16 },
		{ 285, 170, 16 },
		{ 285, 171, 16 },
		{ 285, 172, 16 },
		{ 285, 173, 16 },
		{ 285, 174, 16 },
		{ 285, 175, 16 },
		{ 285, 176, 16 },
		{ 285, 177, 16 },
		{ 285, 178, 16 },
		{ 285, 179, 16 },
		{ 285, 180, 16 },
		{ 285, 181, 16 },
		{ 285, 182, 16 },
		{ 285, 183, 16 },
		{ 285, 184, 16 },
		{ 285, 185, 16 },
		{ 285, 186, 16 },
		{ 285, 187, 16 },
		{ 285, 188, 16 },
		{ 285, 189, 16 },
		{ 285, 190, 16 },
		{ 285, 191, 16 },
		{ 285, 192, 16 },
		{ 285, 193, 16 },
		{ 285, 194, 16 },
		{ 285, 195, 16 },
		{ 285, 196, 16 },
		{ 285, 197, 16 },
		{ 285, 198, 16 },
		{ 285, 199, 16 },
		{ 285, 200, 16 },
		{ 285, 201, 16 },
		{ 285, 202, 16 },
		{ 285, 203, 16 },
		{ 285, 204, 16 },
		{ 285, 205, 16 },
		{ 285, 206, 16 },
		{ 285, 207, 16 },
		{ 285, 208, 16 },
		{ 285, 209, 16 },
		{ 285, 210, 16 },
		{ 285, 211, 16 },
		{ 285, 212, 16 },
		{ 285, 213, 16 },
		{ 285, 214, 16 },
		{ 285, 215, 16 },
		{ 285, 216, 16 },
		{ 285, 217, 16 },
		{ 285, 218, 16 },
		{ 285, 219, 16 },
		{ 286, 142, 16 },
		{ 286, 143, 16 },
		{ 286, 144, 16 },
		{ 286, 145, 16 },
		{ 286, 146, 16 },
		{ 286, 147, 16 },
		{ 286, 148, 16 },
		{ 286, 149, 16 },
		{ 286, 150, 16 },
		{ 286, 151, 16 },
		{ 286, 152, 16 },
		{ 286, 153, 16 },
		{ 286, 154, 16 },
		{ 286, 155, 16 },
		{ 286, 156, 16 },
		{ 286, 157, 16 },
		{ 286, 158, 16 },
		{ 286, 159, 16 },
		{ 286, 160, 16 },
		{ 286, 161, 16 },
		{ 286, 162, 16 },
		{ 286, 163, 16 },
		{ 286, 164, 16 },
		{ 286, 165, 16 },
		{ 286, 166, 16 },
		{ 286, 167, 16 },
		{ 286, 168, 16 },
		{ 286, 169, 16 },
		{ 286, 170, 16 },
		{ 286, 171, 16 },
		{ 286, 172, 16 },
		{ 286, 173, 16 },
		{ 286, 174, 16 },
		{ 286, 175, 16 },
		{ 286, 176, 16 },
		{ 286, 177, 16 },
		{ 286, 178, 16 },
		{ 286, 179, 16 },
		{ 286, 180, 16 },
		{ 286, 181, 16 },
		{ 286, 182, 16 },
		{ 286, 183, 16 },
		{ 286, 184, 16 },
		{ 286, 185, 16 },
		{ 286, 186, 16 },
		{ 286, 187, 16 },
		{ 286, 188, 16 },
		{ 286, 189, 16 },
		{ 286, 190, 16 },
		{ 286, 191, 16 },
		{ 286, 192, 16 },
		{ 286, 193, 16 },
		{ 286, 194, 16 },
		{ 286, 195, 16 },
		{ 286, 196, 16 },
		{ 286, 197, 16 },
		{ 286, 198, 16 },
		{ 286, 199, 16 },
		{ 286, 200, 16 },
		{ 286, 201, 16 },
		{ 286, 202, 16 },
		{ 286, 203, 16 },
		{ 286, 204, 16 },
		{ 286, 205, 16 },
		{ 286, 206, 16 },
		{ 286, 207, 16 },
		{ 286, 208, 16 },
		{ 286, 209, 16 },
		{ 286, 210, 16 },
		{ 286, 211, 16 },
		{ 286, 212, 16 },
		{ 286, 213, 16 },
		{ 286, 214, 16 },
		{ 286, 215, 16 },
		{ 286, 216, 16 },
		{ 286, 217, 16 },
		{ 286, 218, 16 },
		{ 287, 139, 16 },
		{ 287, 140, 16 },
		{ 287, 141, 16 },
		{ 287, 142, 16 },
		{ 287, 143, 16 },
		{ 287, 144, 16 },
		{ 287, 145, 16 },
		{ 287, 146, 16 },
		{ 287, 147, 16 },
		{ 287, 148, 16 },
		{ 287, 149, 16 },
		{ 287, 150, 16 },
		{ 287, 151, 16 },
		{ 287, 152, 16 },
		{ 287, 153, 16 },
		{ 287, 154, 16 },
		{ 287, 155, 16 },
		{ 287, 156, 16 },
		{ 287, 157, 16 },
		{ 287, 158, 16 },
		{ 287, 159, 16 },
		{ 287, 160, 16 },
		{ 287, 161, 16 },
		{ 287, 162, 16 },
		{ 287, 163, 16 },
		{ 287, 164, 16 },
		{ 287, 165, 16 },
		{ 287, 166, 16 },
		{ 287, 167, 16 },
		{ 287, 168, 16 },
		{ 287, 169, 16 },
		{ 287, 170, 16 },
		{ 287, 171, 16 },
		{ 287, 172, 16 },
		{ 287, 173, 16 },
		{ 287, 174, 16 },
		{ 287, 175, 16 },
		{ 287, 176, 16 },
		{ 287, 177, 16 },
		{ 287, 178, 16 },
		{ 287, 179, 16 },
		{ 287, 180, 16 },
		{ 287, 181, 16 },
		{ 287, 182, 16 },
		{ 287, 183, 16 },
		{ 287, 184, 16 },
		{ 287, 185, 16 },
		{ 287, 186, 16 },
		{ 287, 187, 16 },
		{ 287, 188, 16 },
		{ 287, 189, 16 },
		{ 287, 190, 16 },
		{ 287, 191, 16 },
		{ 287, 192, 16 },
		{ 287, 193, 16 },
		{ 287, 194, 16 },
		{ 287, 195, 16 },
		{ 287, 196, 16 },
		{ 287, 197, 16 },
		{ 287, 198, 16 },
		{ 287, 199, 16 },
		{ 287, 200, 16 },
		{ 287, 201, 16 },
		{ 287, 202, 16 },
		{ 287, 203, 16 },
		{ 287, 204, 16 },
		{ 287, 205, 16 },
		{ 287, 206, 16 },
		{ 287, 207, 16 },
		{ 287, 208, 16 },
		{ 287, 209, 16 },
		{ 287, 210, 16 },
		{ 287, 211, 16 },
		{ 287, 212, 16 },
		{ 287, 213, 16 },
		{ 287, 214, 16 },
		{ 287, 215, 16 },
		{ 287, 216, 16 },
		{ 288, 136, 16 },
		{ 288, 137, 16 },
		{ 288, 138, 16 },
		{ 288, 139, 16 },
		{ 288, 140, 16 },
		{ 288, 141, 16 },
		{ 288, 142, 16 },
		{ 288, 143, 16 },
		{ 288, 144, 16 },
		{ 288, 145, 16 },
		{ 288, 146, 16 },
		{ 288, 147, 16 },
		{ 288, 148, 16 },
		{ 288, 149, 16 },
		{ 288, 150, 16 },
		{ 288, 151, 16 },
		{ 288, 152, 16 },
		{ 288, 153, 16 },
		{ 288, 154, 16 },
		{ 288, 155, 16 },
		{ 288, 156, 16 },
		{ 288, 157, 16 },
		{ 288, 158, 16 },
		{ 288, 159, 16 },
		{ 288, 160, 16 },
		{ 288, 161, 16 },
		{ 288, 162, 16 },
		{ 288, 163, 16 },
		{ 288, 164, 16 },
		{ 288, 165, 16 },
		{ 288, 166, 16 },
		{ 288, 167, 16 },
		{ 288, 168, 16 },
		{ 288, 169, 16 },
		{ 288, 170, 16 },
		{ 288, 171, 16 },
		{ 288, 172, 16 },
		{ 288, 173, 16 },
		{ 288, 174, 16 },
		{ 288, 175, 16 },
		{ 288, 176, 16 },
		{ 288, 177, 16 },
		{ 288, 178, 16 },
		{ 288, 179, 16 },
		{ 288, 180, 16 },
		{ 288, 181, 16 },
		{ 288, 182, 16 },
		{ 288, 183, 16 },
		{ 288, 184, 16 },
		{ 288, 185, 16 },
		{ 288, 186, 16 },
		{ 288, 187, 16 },
		{ 288, 188, 16 },
		{ 288, 189, 16 },
		{ 288, 190, 16 },
		{ 288, 191, 16 },
		{ 288, 192, 16 },
		{ 288, 193, 16 },
		{ 288, 194, 16 },
		{ 288, 195, 16 },
		{ 288, 196, 16 },
		{ 288, 197, 16 },
		{ 288, 198, 16 },
		{ 288, 199, 16 },
		{ 288, 200, 16 },
		{ 288, 201, 16 },
		{ 288, 202, 16 },
		{ 288, 203, 16 },
		{ 288, 204, 16 },
		{ 288, 205, 16 },
		{ 288, 206, 16 },
		{ 288, 207, 16 },
		{ 288, 208, 16 },
		{ 288, 209, 16 },
		{ 288, 210, 16 },
		{ 288, 211, 16 },
		{ 288, 212, 16 },
		{ 288, 213, 16 },
		{ 288, 214, 16 },
		{ 288, 215, 16 },
		{ 289, 134, 16 },
		{ 289, 135, 16 },
		{ 289, 136, 16 },
		{ 289, 137, 16 },
		{ 289, 138, 16 },
		{ 289, 139, 16 },
		{ 289, 140, 16 },
		{ 289, 141, 16 },
		{ 289, 142, 16 },
		{ 289, 143, 16 },
		{ 289, 144, 16 },
		{ 289, 145, 16 },
		{ 289, 146, 16 },
		{ 289, 147, 16 },
		{ 289, 148, 16 },
		{ 289, 149, 16 },
		{ 289, 150, 16 },
		{ 289, 151, 16 },
		{ 289, 152, 16 },
		{ 289, 153, 16 },
		{ 289, 154, 16 },
		{ 289, 155, 16 },
		{ 289, 156, 16 },
		{ 289, 157, 16 },
		{ 289, 158, 16 },
		{ 289, 159, 16 },
		{ 289, 160, 16 },
		{ 289, 161, 16 },
		{ 289, 162, 16 },
		{ 289, 163, 16 },
		{ 289, 164, 16 },
		{ 289, 165, 16 },
		{ 289, 166, 16 },
		{ 289, 167, 16 },
		{ 289, 168, 16 },
		{ 289, 169, 16 },
		{ 289, 170, 16 },
		{ 289, 171, 16 },
		{ 289, 172, 16 },
		{ 289, 173, 16 },
		{ 289, 174, 16 },
		{ 289, 175, 16 },
		{ 289, 176, 16 },
		{ 289, 177, 16 },
		{ 289, 178, 16 },
		{ 289, 179, 16 },
		{ 289, 180, 16 },
		{ 289, 181, 16 },
		{ 289, 182, 16 },
		{ 289, 183, 16 },
		{ 289, 184, 16 },
		{ 289, 185, 16 },
		{ 289, 186, 16 },
		{ 289, 187, 16 },
		{ 289, 188, 16 },
		{ 289, 189, 16 },
		{ 289, 190, 16 },
		{ 289, 191, 16 },
		{ 289, 192, 16 },
		{ 289, 193, 16 },
		{ 289, 194, 16 },
		{ 289, 195, 16 },
		{ 289, 196, 16 },
		{ 289, 197, 16 },
		{ 289, 198, 16 },
		{ 289, 199, 16 },
		{ 289, 200, 16 },
		{ 289, 201, 16 },
		{ 289, 202, 16 },
		{ 289, 203, 16 },
		{ 289, 204, 16 },
		{ 289, 205, 16 },
		{ 289, 206, 16 },
		{ 289, 207, 16 },
		{ 289, 208, 16 },
		{ 289, 209, 16 },
		{ 289, 210, 16 },
		{ 289, 211, 16 },
		{ 289, 212, 16 },
		{ 289, 213, 16 },
		{ 289, 214, 16 },
		{ 290, 133, 16 },
		{ 290, 134, 16 },
		{ 290, 135, 16 },
		{ 290, 136, 16 },
		{ 290, 137, 16 },
		{ 290, 138, 16 },
		{ 290, 139, 16 },
		{ 290, 140, 16 },
		{ 290, 141, 16 },
		{ 290, 142, 16 },
		{ 290, 143, 16 },
		{ 290, 144, 16 },
		{ 290, 145, 16 },
		{ 290, 146, 16 },
		{ 290, 147, 16 },
		{ 290, 148, 16 },
		{ 290, 149, 16 },
		{ 290, 150, 16 },
		{ 290, 151, 16 },
		{ 290, 152, 16 },
		{ 290, 153, 16 },
		{ 290, 154, 16 },
		{ 290, 155, 16 },
		{ 290, 156, 16 },
		{ 290, 157, 16 },
		{ 290, 158, 16 },
		{ 290, 159, 16 },
		{ 290, 160, 16 },
		{ 290, 161, 16 },
		{ 290, 162, 16 },
		{ 290, 163, 16 },
		{ 290, 164, 16 },
		{ 290, 165, 16 },
		{ 290, 166, 16 },
		{ 290, 167, 16 },
		{ 290, 168, 16 },
		{ 290, 169, 16 },
		{ 290, 170, 16 },
		{ 290, 171, 16 },
		{ 290, 172, 16 },
		{ 290, 173, 16 },
		{ 290, 174, 16 },
		{ 290, 175, 16 },
		{ 290, 176, 16 },
		{ 290, 177, 16 },
		{ 290, 178, 16 },
		{ 290, 179, 16 },
		{ 290, 180, 16 },
		{ 290, 181, 16 },
		{ 290, 182, 16 },
		{ 290, 183, 16 },
		{ 290, 184, 16 },
		{ 290, 185, 16 },
		{ 290, 186, 16 },
		{ 290, 187, 16 },
		{ 290, 188, 16 },
		{ 290, 189, 16 },
		{ 290, 190, 16 },
		{ 290, 191, 16 },
		{ 290, 192, 16 },
		{ 290, 193, 16 },
		{ 290, 194, 16 },
		{ 290, 195, 16 },
		{ 290, 196, 16 },
		{ 290, 197, 16 },
		{ 290, 198, 16 },
		{ 290, 199, 16 },
		{ 290, 200, 16 },
		{ 290, 201, 16 },
		{ 290, 202, 16 },
		{ 290, 203, 16 },
		{ 290, 204, 16 },
		{ 290, 205, 16 },
		{ 290, 206, 16 },
		{ 290, 207, 16 },
		{ 290, 208, 16 },
		{ 290, 209, 16 },
		{ 290, 210, 16 },
		{ 290, 211, 16 },
		{ 290, 212, 16 },
		{ 291, 133, 16 },
		{ 291, 134, 16 },
		{ 291, 135, 16 },
		{ 291, 136, 16 },
		{ 291, 137, 16 },
		{ 291, 138, 16 },
		{ 291, 139, 16 },
		{ 291, 140, 16 },
		{ 291, 141, 16 },
		{ 291, 142, 16 },
		{ 291, 143, 16 },
		{ 291, 144, 16 },
		{ 291, 145, 16 },
		{ 291, 146, 16 },
		{ 291, 147, 16 },
		{ 291, 148, 16 },
		{ 291, 149, 16 },
		{ 291, 150, 16 },
		{ 291, 151, 16 },
		{ 291, 152, 16 },
		{ 291, 153, 16 },
		{ 291, 154, 16 },
		{ 291, 155, 16 },
		{ 291, 156, 16 },
		{ 291, 157, 16 },
		{ 291, 158, 16 },
		{ 291, 159, 16 },
		{ 291, 160, 16 },
		{ 291, 161, 16 },
		{ 291, 162, 16 },
		{ 291, 163, 16 },
		{ 291, 164, 16 },
		{ 291, 165, 16 },
		{ 291, 166, 16 },
		{ 291, 167, 16 },
		{ 291, 168, 16 },
		{ 291, 169, 16 },
		{ 291, 170, 16 },
		{ 291, 171, 16 },
		{ 291, 172, 16 },
		{ 291, 173, 16 },
		{ 291, 174, 16 },
		{ 291, 175, 16 },
		{ 291, 176, 16 },
		{ 291, 177, 16 },
		{ 291, 178, 16 },
		{ 291, 179, 16 },
		{ 291, 180, 16 },
		{ 291, 181, 16 },
		{ 291, 182, 16 },
		{ 291, 183, 16 },
		{ 291, 184, 16 },
		{ 291, 185, 16 },
		{ 291, 186, 16 },
		{ 291, 187, 16 },
		{ 291, 188, 16 },
		{ 291, 189, 16 },
		{ 291, 190, 16 },
		{ 291, 191, 16 },
		{ 291, 192, 16 },
		{ 291, 193, 16 },
		{ 291, 194, 16 },
		{ 291, 195, 16 },
		{ 291, 196, 16 },
		{ 291, 197, 16 },
		{ 291, 198, 16 },
		{ 291, 199, 16 },
		{ 291, 200, 16 },
		{ 291, 201, 16 },
		{ 291, 202, 16 },
		{ 291, 203, 16 },
		{ 291, 204, 16 },
		{ 291, 205, 16 },
		{ 291, 206, 16 },
		{ 291, 207, 16 },
		{ 291, 208, 16 },
		{ 291, 209, 16 },
		{ 291, 210, 16 },
		{ 291, 211, 16 },
		{ 292, 132, 16 },
		{ 292, 133, 16 },
		{ 292, 134, 16 },
		{ 292, 135, 16 },
		{ 292, 136, 16 },
		{ 292, 137, 16 },
		{ 292, 138, 16 },
		{ 292, 139, 16 },
		{ 292, 140, 16 },
		{ 292, 141, 16 },
		{ 292, 142, 16 },
		{ 292, 143, 16 },
		{ 292, 144, 16 },
		{ 292, 145, 16 },
		{ 292, 146, 16 },
		{ 292, 147, 16 },
		{ 292, 148, 16 },
		{ 292, 149, 16 },
		{ 292, 150, 16 },
		{ 292, 151, 16 },
		{ 292, 152, 16 },
		{ 292, 153, 16 },
		{ 292, 154, 16 },
		{ 292, 155, 16 },
		{ 292, 156, 16 },
		{ 292, 157, 16 },
		{ 292, 158, 16 },
		{ 292, 159, 16 },
		{ 292, 160, 16 },
		{ 292, 161, 16 },
		{ 292, 162, 16 },
		{ 292, 163, 16 },
		{ 292, 164, 16 },
		{ 292, 165, 16 },
		{ 292, 166, 16 },
		{ 292, 167, 16 },
		{ 292, 168, 16 },
		{ 292, 169, 16 },
		{ 292, 170, 16 },
		{ 292, 171, 16 },
		{ 292, 172, 16 },
		{ 292, 173, 16 },
		{ 292, 174, 16 },
		{ 292, 175, 16 },
		{ 292, 176, 16 },
		{ 292, 177, 16 },
		{ 292, 178, 16 },
		{ 292, 179, 16 },
		{ 292, 180, 16 },
		{ 292, 181, 16 },
		{ 292, 182, 16 },
		{ 292, 183, 16 },
		{ 292, 184, 16 },
		{ 292, 185, 16 },
		{ 292, 186, 16 },
		{ 292, 187, 16 },
		{ 292, 188, 16 },
		{ 292, 189, 16 },
		{ 292, 190, 16 },
		{ 292, 191, 16 },
		{ 292, 192, 16 },
		{ 292, 193, 16 },
		{ 292, 194, 16 },
		{ 292, 195, 16 },
		{ 292, 196, 16 },
		{ 292, 197, 16 },
		{ 292, 198, 16 },
		{ 292, 199, 16 },
		{ 292, 200, 16 },
		{ 292, 201, 16 },
		{ 292, 202, 16 },
		{ 292, 203, 16 },
		{ 292, 204, 16 },
		{ 292, 205, 16 },
		{ 292, 206, 16 },
		{ 292, 207, 16 },
		{ 292, 208, 16 },
		{ 292, 209, 16 },
		{ 292, 210, 16 },
		{ 293, 132, 16 },
		{ 293, 133, 16 },
		{ 293, 134, 16 },
		{ 293, 135, 16 },
		{ 293, 136, 16 },
		{ 293, 137, 16 },
		{ 293, 138, 16 },
		{ 293, 139, 16 },
		{ 293, 140, 16 },
		{ 293, 141, 16 },
		{ 293, 142, 16 },
		{ 293, 143, 16 },
		{ 293, 144, 16 },
		{ 293, 145, 16 },
		{ 293, 146, 16 },
		{ 293, 147, 16 },
		{ 293, 148, 16 },
		{ 293, 149, 16 },
		{ 293, 150, 16 },
		{ 293, 151, 16 },
		{ 293, 152, 16 },
		{ 293, 153, 16 },
		{ 293, 154, 16 },
		{ 293, 155, 16 },
		{ 293, 156, 16 },
		{ 293, 157, 16 },
		{ 293, 158, 16 },
		{ 293, 159, 16 },
		{ 293, 160, 16 },
		{ 293, 161, 16 },
		{ 293, 162, 16 },
		{ 293, 163, 16 },
		{ 293, 164, 16 },
		{ 293, 165, 16 },
		{ 293, 166, 16 },
		{ 293, 167, 16 },
		{ 293, 168, 16 },
		{ 293, 169, 16 },
		{ 293, 170, 16 },
		{ 293, 171, 16 },
		{ 293, 172, 16 },
		{ 293, 173, 16 },
		{ 293, 174, 16 },
		{ 293, 175, 16 },
		{ 293, 176, 16 },
		{ 293, 177, 16 },
		{ 293, 178, 16 },
		{ 293, 179, 16 },
		{ 293, 180, 16 },
		{ 293, 181, 16 },
		{ 293, 182, 16 },
		{ 293, 183, 16 },
		{ 293, 184, 16 },
		{ 293, 185, 16 },
		{ 293, 186, 16 },
		{ 293, 187, 16 },
		{ 293, 188, 16 },
		{ 293, 189, 16 },
		{ 293, 190, 16 },
		{ 293, 191, 16 },
		{ 293, 192, 16 },
		{ 293, 193, 16 },
		{ 293, 194, 16 },
		{ 293, 195, 16 },
		{ 293, 196, 16 },
		{ 293, 197, 16 },
		{ 293, 198, 16 },
		{ 293, 199, 16 },
		{ 293, 200, 16 },
		{ 293, 201, 16 },
		{ 293, 202, 16 },
		{ 293, 203, 16 },
		{ 293, 204, 16 },
		{ 293, 205, 16 },
		{ 293, 206, 16 },
		{ 293, 207, 16 },
		{ 293, 208, 16 },
		{ 294, 132, 16 },
		{ 294, 133, 16 },
		{ 294, 134, 16 },
		{ 294, 135, 16 },
		{ 294, 136, 16 },
		{ 294, 137, 16 },
		{ 294, 138, 16 },
		{ 294, 139, 16 },
		{ 294, 140, 16 },
		{ 294, 141, 16 },
		{ 294, 142, 16 },
		{ 294, 143, 16 },
		{ 294, 144, 16 },
		{ 294, 145, 16 },
		{ 294, 146, 16 },
		{ 294, 147, 16 },
		{ 294, 148, 16 },
		{ 294, 149, 16 },
		{ 294, 150, 16 },
		{ 294, 151, 16 },
		{ 294, 152, 16 },
		{ 294, 153, 16 },
		{ 294, 154, 16 },
		{ 294, 155, 16 },
		{ 294, 156, 16 },
		{ 294, 157, 16 },
		{ 294, 158, 16 },
		{ 294, 159, 16 },
		{ 294, 160, 16 },
		{ 294, 161, 16 },
		{ 294, 162, 16 },
		{ 294, 163, 16 },
		{ 294, 164, 16 },
		{ 294, 165, 16 },
		{ 294, 166, 16 },
		{ 294, 167, 16 },
		{ 294, 168, 16 },
		{ 294, 169, 16 },
		{ 294, 170, 16 },
		{ 294, 171, 16 },
		{ 294, 172, 16 },
		{ 294, 173, 16 },
		{ 294, 174, 16 },
		{ 294, 175, 16 },
		{ 294, 176, 16 },
		{ 294, 177, 16 },
		{ 294, 178, 16 },
		{ 294, 179, 16 },
		{ 294, 180, 16 },
		{ 294, 181, 16 },
		{ 294, 182, 16 },
		{ 294, 183, 16 },
		{ 294, 184, 16 },
		{ 294, 185, 16 },
		{ 294, 186, 16 },
		{ 294, 187, 16 },
		{ 294, 188, 16 },
		{ 294, 189, 16 },
		{ 294, 190, 16 },
		{ 294, 191, 16 },
		{ 294, 192, 16 },
		{ 294, 193, 16 },
		{ 294, 194, 16 },
		{ 294, 195, 16 },
		{ 294, 196, 16 },
		{ 294, 197, 16 },
		{ 294, 198, 16 },
		{ 294, 199, 16 },
		{ 294, 200, 16 },
		{ 294, 201, 16 },
		{ 294, 202, 16 },
		{ 294, 203, 16 },
		{ 294, 204, 16 },
		{ 294, 205, 16 },
		{ 294, 206, 16 },
		{ 294, 207, 16 },
		{ 295, 132, 16 },
		{ 295, 133, 16 },
		{ 295, 134, 16 },
		{ 295, 135, 16 },
		{ 295, 136, 16 },
		{ 295, 137, 16 },
		{ 295, 138, 16 },
		{ 295, 139, 16 },
		{ 295, 140, 16 },
		{ 295, 141, 16 },
		{ 295, 142, 16 },
		{ 295, 143, 16 },
		{ 295, 144, 16 },
		{ 295, 145, 16 },
		{ 295, 146, 16 },
		{ 295, 147, 16 },
		{ 295, 148, 16 },
		{ 295, 149, 16 },
		{ 295, 150, 16 },
		{ 295, 151, 16 },
		{ 295, 152, 16 },
		{ 295, 153, 16 },
		{ 295, 154, 16 },
		{ 295, 155, 16 },
		{ 295, 156, 16 },
		{ 295, 157, 16 },
		{ 295, 158, 16 },
		{ 295, 159, 16 },
		{ 295, 160, 16 },
		{ 295, 161, 16 },
		{ 295, 162, 16 },
		{ 295, 163, 16 },
		{ 295, 164, 16 },
		{ 295, 165, 16 },
		{ 295, 166, 16 },
		{ 295, 167, 16 },
		{ 295, 168, 16 },
		{ 295, 169, 16 },
		{ 295, 170, 16 },
		{ 295, 171, 16 },
		{ 295, 172, 16 },
		{ 295, 173, 16 },
		{ 295, 174, 16 },
		{ 295, 175, 16 },
		{ 295, 176, 16 },
		{ 295, 177, 16 },
		{ 295, 178, 16 },
		{ 295, 179, 16 },
		{ 295, 180, 16 },
		{ 295, 181, 16 },
		{ 295, 182, 16 },
		{ 295, 183, 16 },
		{ 295, 184, 16 },
		{ 295, 185, 16 },
		{ 295, 186, 16 },
		{ 295, 187, 16 },
		{ 295, 188, 16 },
		{ 295, 189, 16 },
		{ 295, 190, 16 },
		{ 295, 191, 16 },
		{ 295, 192, 16 },
		{ 295, 193, 16 },
		{ 295, 194, 16 },
		{ 295, 195, 16 },
		{ 295, 196, 16 },
		{ 295, 197, 16 },
		{ 295, 198, 16 },
		{ 295, 199, 16 },
		{ 295, 200, 16 },
		{ 295, 201, 16 },
		{ 295, 202, 16 },
		{ 295, 203, 16 },
		{ 295, 204, 16 },
		{ 295, 205, 16 },
		{ 295, 206, 16 },
		{ 296, 131, 16 },
		{ 296, 132, 16 },
		{ 296, 133, 16 },
		{ 296, 134, 16 },
		{ 296, 135, 16 },
		{ 296, 136, 16 },
		{ 296, 137, 16 },
		{ 296, 138, 16 },
		{ 296, 139, 16 },
		{ 296, 140, 16 },
		{ 296, 141, 16 },
		{ 296, 142, 16 },
		{ 296, 143, 16 },
		{ 296, 144, 16 },
		{ 296, 145, 16 },
		{ 296, 146, 16 },
		{ 296, 147, 16 },
		{ 296, 148, 16 },
		{ 296, 149, 16 },
		{ 296, 150, 16 },
		{ 296, 151, 16 },
		{ 296, 152, 16 },
		{ 296, 153, 16 },
		{ 296, 154, 16 },
		{ 296, 155, 16 },
		{ 296, 156, 16 },
		{ 296, 157, 16 },
		{ 296, 158, 16 },
		{ 296, 159, 16 },
		{ 296, 160, 16 },
		{ 296, 161, 16 },
		{ 296, 162, 16 },
		{ 296, 163, 16 },
		{ 296, 164, 16 },
		{ 296, 165, 16 },
		{ 296, 166, 16 },
		{ 296, 167, 16 },
		{ 296, 168, 16 },
		{ 296, 169, 16 },
		{ 296, 170, 16 },
		{ 296, 171, 16 },
		{ 296, 172, 16 },
		{ 296, 173, 16 },
		{ 296, 174, 16 },
		{ 296, 175, 16 },
		{ 296, 176, 16 },
		{ 296, 177, 16 },
		{ 296, 178, 16 },
		{ 296, 179, 16 },
		{ 296, 180, 16 },
		{ 296, 181, 16 },
		{ 296, 182, 16 },
		{ 296, 183, 16 },
		{ 296, 184, 16 },
		{ 296, 185, 16 },
		{ 296, 186, 16 },
		{ 296, 187, 16 },
		{ 296, 188, 16 },
		{ 296, 189, 16 },
		{ 296, 190, 16 },
		{ 296, 191, 16 },
		{ 296, 192, 16 },
		{ 296, 193, 16 },
		{ 296, 194, 16 },
		{ 296, 195, 16 },
		{ 296, 196, 16 },
		{ 296, 197, 16 },
		{ 296, 198, 16 },
		{ 296, 199, 16 },
		{ 296, 200, 16 },
		{ 296, 201, 16 },
		{ 296, 202, 16 },
		{ 296, 203, 16 },
		{ 296, 204, 16 },
		{ 297, 131, 16 },
		{ 297, 132, 16 },
		{ 297, 133, 16 },
		{ 297, 134, 16 },
		{ 297, 135, 16 },
		{ 297, 136, 16 },
		{ 297, 137, 16 },
		{ 297, 138, 16 },
		{ 297, 139, 16 },
		{ 297, 140, 16 },
		{ 297, 141, 16 },
		{ 297, 142, 16 },
		{ 297, 143, 16 },
		{ 297, 144, 16 },
		{ 297, 145, 16 },
		{ 297, 146, 16 },
		{ 297, 147, 16 },
		{ 297, 148, 16 },
		{ 297, 149, 16 },
		{ 297, 150, 16 },
		{ 297, 151, 16 },
		{ 297, 152, 16 },
		{ 297, 153, 16 },
		{ 297, 154, 16 },
		{ 297, 155, 16 },
		{ 297, 156, 16 },
		{ 297, 157, 16 },
		{ 297, 158, 16 },
		{ 297, 159, 16 },
		{ 297, 160, 16 },
		{ 297, 161, 16 },
		{ 297, 162, 16 },
		{ 297, 163, 16 },
		{ 297, 164, 16 },
		{ 297, 165, 16 },
		{ 297, 166, 16 },
		{ 297, 167, 16 },
		{ 297, 168, 16 },
		{ 297, 169, 16 },
		{ 297, 170, 16 },
		{ 297, 171, 16 },
		{ 297, 172, 16 },
		{ 297, 173, 16 },
		{ 297, 174, 16 },
		{ 297, 175, 16 },
		{ 297, 176, 16 },
		{ 297, 177, 16 },
		{ 297, 178, 16 },
		{ 297, 179, 16 },
		{ 297, 180, 16 },
		{ 297, 181, 16 },
		{ 297, 182, 16 },
		{ 297, 183, 16 },
		{ 297, 184, 16 },
		{ 297, 185, 16 },
		{ 297, 186, 16 },
		{ 297, 187, 16 },
		{ 297, 188, 16 },
		{ 297, 189, 16 },
		{ 297, 190, 16 },
		{ 297, 191, 16 },
		{ 297, 192, 16 },
		{ 297, 193, 16 },
		{ 297, 194, 16 },
		{ 297, 195, 16 },
		{ 297, 196, 16 },
		{ 297, 197, 16 },
		{ 297, 198, 16 },
		{ 297, 199, 16 },
		{ 297, 200, 16 },
		{ 297, 201, 16 },
		{ 297, 202, 16 },
		{ 297, 203, 16 },
		{ 298, 131, 16 },
		{ 298, 132, 16 },
		{ 298, 133, 16 },
		{ 298, 134, 16 },
		{ 298, 135, 16 },
		{ 298, 136, 16 },
		{ 298, 137, 16 },
		{ 298, 138, 16 },
		{ 298, 139, 16 },
		{ 298, 140, 16 },
		{ 298, 141, 16 },
		{ 298, 142, 16 },
		{ 298, 143, 16 },
		{ 298, 144, 16 },
		{ 298, 145, 16 },
		{ 298, 146, 16 },
		{ 298, 147, 16 },
		{ 298, 148, 16 },
		{ 298, 149, 16 },
		{ 298, 150, 16 },
		{ 298, 151, 16 },
		{ 298, 152, 16 },
		{ 298, 153, 16 },
		{ 298, 154, 16 },
		{ 298, 155, 16 },
		{ 298, 156, 16 },
		{ 298, 157, 16 },
		{ 298, 158, 16 },
		{ 298, 159, 16 },
		{ 298, 160, 16 },
		{ 298, 161, 16 },
		{ 298, 162, 16 },
		{ 298, 163, 16 },
		{ 298, 164, 16 },
		{ 298, 165, 16 },
		{ 298, 166, 16 },
		{ 298, 167, 16 },
		{ 298, 168, 16 },
		{ 298, 169, 16 },
		{ 298, 170, 16 },
		{ 298, 171, 16 },
		{ 298, 172, 16 },
		{ 298, 173, 16 },
		{ 298, 174, 16 },
		{ 298, 175, 16 },
		{ 298, 176, 16 },
		{ 298, 177, 16 },
		{ 298, 178, 16 },
		{ 298, 179, 16 },
		{ 298, 180, 16 },
		{ 298, 181, 16 },
		{ 298, 182, 16 },
		{ 298, 183, 16 },
		{ 298, 184, 16 },
		{ 298, 185, 16 },
		{ 298, 186, 16 },
		{ 298, 187, 16 },
		{ 298, 188, 16 },
		{ 298, 189, 16 },
		{ 298, 190, 16 },
		{ 298, 191, 16 },
		{ 298, 192, 16 },
		{ 298, 193, 16 },
		{ 298, 194, 16 },
		{ 298, 195, 16 },
		{ 298, 196, 16 },
		{ 298, 197, 16 },
		{ 298, 198, 16 },
		{ 298, 199, 16 },
		{ 298, 200, 16 },
		{ 298, 201, 16 },
		{ 298, 202, 16 },
		{ 299, 131, 16 },
		{ 299, 132, 16 },
		{ 299, 133, 16 },
		{ 299, 134, 16 },
		{ 299, 135, 16 },
		{ 299, 136, 16 },
		{ 299, 137, 16 },
		{ 299, 138, 16 },
		{ 299, 139, 16 },
		{ 299, 140, 16 },
		{ 299, 141, 16 },
		{ 299, 142, 16 },
		{ 299, 143, 16 },
		{ 299, 144, 16 },
		{ 299, 145, 16 },
		{ 299, 146, 16 },
		{ 299, 147, 16 },
		{ 299, 148, 16 },
		{ 299, 149, 16 },
		{ 299, 150, 16 },
		{ 299, 151, 16 },
		{ 299, 152, 16 },
		{ 299, 153, 16 },
		{ 299, 154, 16 },
		{ 299, 155, 16 },
		{ 299, 156, 16 },
		{ 299, 157, 16 },
		{ 299, 158, 16 },
		{ 299, 159, 16 },
		{ 299, 160, 16 },
		{ 299, 161, 16 },
		{ 299, 162, 16 },
		{ 299, 163, 16 },
		{ 299, 164, 16 },
		{ 299, 165, 16 },
		{ 299, 166, 16 },
		{ 299, 167, 16 },
		{ 299, 168, 16 },
		{ 299, 169, 16 },
		{ 299, 170, 16 },
		{ 299, 171, 16 },
		{ 299, 172, 16 },
		{ 299, 173, 16 },
		{ 299, 174, 16 },
		{ 299, 175, 16 },
		{ 299, 176, 16 },
		{ 299, 177, 16 },
		{ 299, 178, 16 },
		{ 299, 179, 16 },
		{ 299, 180, 16 },
		{ 299, 181, 16 },
		{ 299, 182, 16 },
		{ 299, 183, 16 },
		{ 299, 184, 16 },
		{ 299, 185, 16 },
		{ 299, 186, 16 },
		{ 299, 187, 16 },
		{ 299, 188, 16 },
		{ 299, 189, 16 },
		{ 299, 190, 16 },
		{ 299, 191, 16 },
		{ 299, 192, 16 },
		{ 299, 193, 16 },
		{ 299, 194, 16 },
		{ 299, 195, 16 },
		{ 299, 196, 16 },
		{ 299, 197, 16 },
		{ 299, 198, 16 },
		{ 299, 199, 16 },
		{ 299, 200, 16 },
		{ 300, 131, 16 },
		{ 300, 132, 16 },
		{ 300, 133, 16 },
		{ 300, 134, 16 },
		{ 300, 135, 16 },
		{ 300, 136, 16 },
		{ 300, 137, 16 },
		{ 300, 138, 16 },
		{ 300, 139, 16 },
		{ 300, 140, 16 },
		{ 300, 141, 16 },
		{ 300, 142, 16 },
		{ 300, 143, 16 },
		{ 300, 144, 16 },
		{ 300, 145, 16 },
		{ 300, 146, 16 },
		{ 300, 147, 16 },
		{ 300, 148, 16 },
		{ 300, 149, 16 },
		{ 300, 150, 16 },
		{ 300, 151, 16 },
		{ 300, 152, 16 },
		{ 300, 153, 16 },
		{ 300, 154, 16 },
		{ 300, 155, 16 },
		{ 300, 156, 16 },
		{ 300, 157, 16 },
		{ 300, 158, 16 },
		{ 300, 159, 16 },
		{ 300, 160, 16 },
		{ 300, 161, 16 },
		{ 300, 162, 16 },
		{ 300, 163, 16 },
		{ 300, 164, 16 },
		{ 300, 165, 16 },
		{ 300, 166, 16 },
		{ 300, 167, 16 },
		{ 300, 168, 16 },
		{ 300, 169, 16 },
		{ 300, 170, 16 },
		{ 300, 171, 16 },
		{ 300, 172, 16 },
		{ 300, 173, 16 },
		{ 300, 174, 16 },
		{ 300, 175, 16 },
		{ 300, 176, 16 },
		{ 300, 177, 16 },
		{ 300, 178, 16 },
		{ 300, 179, 16 },
		{ 300, 180, 16 },
		{ 300, 181, 16 },
		{ 300, 182, 16 },
		{ 300, 183, 16 },
		{ 300, 184, 16 },
		{ 300, 185, 16 },
		{ 300, 186, 16 },
		{ 300, 187, 16 },
		{ 300, 188, 16 },
		{ 300, 189, 16 },
		{ 300, 190, 16 },
		{ 300, 191, 16 },
		{ 300, 192, 16 },
		{ 300, 193, 16 },
		{ 300, 194, 16 },
		{ 300, 195, 16 },
		{ 300, 196, 16 },
		{ 300, 197, 16 },
		{ 300, 198, 16 },
		{ 300, 199, 16 },
		{ 301, 131, 16 },
		{ 301, 132, 16 },
		{ 301, 133, 16 },
		{ 301, 134, 16 },
		{ 301, 135, 16 },
		{ 301, 136, 16 },
		{ 301, 137, 16 },
		{ 301, 138, 16 },
		{ 301, 139, 16 },
		{ 301, 140, 16 },
		{ 301, 141, 16 },
		{ 301, 142, 16 },
		{ 301, 143, 16 },
		{ 301, 144, 16 },
		{ 301, 145, 16 },
		{ 301, 146, 16 },
		{ 301, 147, 16 },
		{ 301, 148, 16 },
		{ 301, 149, 16 },
		{ 301, 150, 16 },
		{ 301, 151, 16 },
		{ 301, 152, 16 },
		{ 301, 153, 16 },
		{ 301, 154, 16 },
		{ 301, 155, 16 },
		{ 301, 156, 16 },
		{ 301, 157, 16 },
		{ 301, 158, 16 },
		{ 301, 159, 16 },
		{ 301, 160, 16 },
		{ 301, 161, 16 },
		{ 301, 162, 16 },
		{ 301, 163, 16 },
		{ 301, 164, 16 },
		{ 301, 165, 16 },
		{ 301, 166, 16 },
		{ 301, 167, 16 },
		{ 301, 168, 16 },
		{ 301, 169, 16 },
		{ 301, 170, 16 },
		{ 301, 171, 16 },
		{ 301, 172, 16 },
		{ 301, 173, 16 },
		{ 301, 174, 16 },
		{ 301, 175, 16 },
		{ 301, 176, 16 },
		{ 301, 177, 16 },
		{ 301, 178, 16 },
		{ 301, 179, 16 },
		{ 301, 180, 16 },
		{ 301, 181, 16 },
		{ 301, 182, 16 },
		{ 301, 183, 16 },
		{ 301, 184, 16 },
		{ 301, 185, 16 },
		{ 301, 186, 16 },
		{ 301, 187, 16 },
		{ 301, 188, 16 },
		{ 301, 189, 16 },
		{ 301, 190, 16 },
		{ 301, 191, 16 },
		{ 301, 192, 16 },
		{ 301, 193, 16 },
		{ 301, 194, 16 },
		{ 301, 195, 16 },
		{ 301, 196, 16 },
		{ 301, 197, 16 },
		{ 301, 198, 16 },
		{ 302, 131, 16 },
		{ 302, 132, 16 },
		{ 302, 133, 16 },
		{ 302, 134, 16 },
		{ 302, 135, 16 },
		{ 302, 136, 16 },
		{ 302, 137, 16 },
		{ 302, 138, 16 },
		{ 302, 139, 16 },
		{ 302, 140, 16 },
		{ 302, 141, 16 },
		{ 302, 142, 16 },
		{ 302, 143, 16 },
		{ 302, 144, 16 },
		{ 302, 145, 16 },
		{ 302, 146, 16 },
		{ 302, 147, 16 },
		{ 302, 148, 16 },
		{ 302, 149, 16 },
		{ 302, 150, 16 },
		{ 302, 151, 16 },
		{ 302, 152, 16 },
		{ 302, 153, 16 },
		{ 302, 154, 16 },
		{ 302, 155, 16 },
		{ 302, 156, 16 },
		{ 302, 157, 16 },
		{ 302, 158, 16 },
		{ 302, 159, 16 },
		{ 302, 160, 16 },
		{ 302, 161, 16 },
		{ 302, 162, 16 },
		{ 302, 163, 16 },
		{ 302, 164, 16 },
		{ 302, 165, 16 },
		{ 302, 166, 16 },
		{ 302, 167, 16 },
		{ 302, 168, 16 },
		{ 302, 169, 16 },
		{ 302, 170, 16 },
		{ 302, 171, 16 },
		{ 302, 172, 16 },
		{ 302, 173, 16 },
		{ 302, 174, 16 },
		{ 302, 175, 16 },
		{ 302, 176, 16 },
		{ 302, 177, 16 },
		{ 302, 178, 16 },
		{ 302, 179, 16 },
		{ 302, 180, 16 },
		{ 302, 181, 16 },
		{ 302, 182, 16 },
		{ 302, 183, 16 },
		{ 302, 184, 16 },
		{ 302, 185, 16 },
		{ 302, 186, 16 },
		{ 302, 187, 16 },
		{ 302, 188, 16 },
		{ 302, 189, 16 },
		{ 302, 190, 16 },
		{ 302, 191, 16 },
		{ 302, 192, 16 },
		{ 302, 193, 16 },
		{ 302, 194, 16 },
		{ 302, 195, 16 },
		{ 302, 196, 16 },
		{ 303, 131, 16 },
		{ 303, 132, 16 },
		{ 303, 133, 16 },
		{ 303, 134, 16 },
		{ 303, 135, 16 },
		{ 303, 136, 16 },
		{ 303, 137, 16 },
		{ 303, 138, 16 },
		{ 303, 139, 16 },
		{ 303, 140, 16 },
		{ 303, 141, 16 },
		{ 303, 142, 16 },
		{ 303, 143, 16 },
		{ 303, 144, 16 },
		{ 303, 145, 16 },
		{ 303, 146, 16 },
		{ 303, 147, 16 },
		{ 303, 148, 16 },
		{ 303, 149, 16 },
		{ 303, 150, 16 },
		{ 303, 151, 16 },
		{ 303, 152, 16 },
		{ 303, 153, 16 },
		{ 303, 154, 16 },
		{ 303, 155, 16 },
		{ 303, 156, 16 },
		{ 303, 157, 16 },
		{ 303, 158, 16 },
		{ 303, 159, 16 },
		{ 303, 160, 16 },
		{ 303, 161, 16 },
		{ 303, 162, 16 },
		{ 303, 163, 16 },
		{ 303, 164, 16 },
		{ 303, 165, 16 },
		{ 303, 166, 16 },
		{ 303, 167, 16 },
		{ 303, 168, 16 },
		{ 303, 169, 16 },
		{ 303, 170, 16 },
		{ 303, 171, 16 },
		{ 303, 172, 16 },
		{ 303, 173, 16 },
		{ 303, 174, 16 },
		{ 303, 175, 16 },
		{ 303, 176, 16 },
		{ 303, 177, 16 },
		{ 303, 178, 16 },
		{ 303, 179, 16 },
		{ 303, 180, 16 },
		{ 303, 181, 16 },
		{ 303, 182, 16 },
		{ 303, 183, 16 },
		{ 303, 184, 16 },
		{ 303, 185, 16 },
		{ 303, 186, 16 },
		{ 303, 187, 16 },
		{ 303, 188, 16 },
		{ 303, 189, 16 },
		{ 303, 190, 16 },
		{ 303, 191, 16 },
		{ 303, 192, 16 },
		{ 303, 193, 16 },
		{ 303, 194, 16 },
		{ 303, 195, 16 },
		{ 304, 131, 16 },
		{ 304, 132, 16 },
		{ 304, 133, 16 },
		{ 304, 134, 16 },
		{ 304, 135, 16 },
		{ 304, 136, 16 },
		{ 304, 137, 16 },
		{ 304, 138, 16 },
		{ 304, 139, 16 },
		{ 304, 140, 16 },
		{ 304, 141, 16 },
		{ 304, 142, 16 },
		{ 304, 143, 16 },
		{ 304, 144, 16 },
		{ 304, 145, 16 },
		{ 304, 146, 16 },
		{ 304, 147, 16 },
		{ 304, 148, 16 },
		{ 304, 149, 16 },
		{ 304, 150, 16 },
		{ 304, 151, 16 },
		{ 304, 152, 16 },
		{ 304, 153, 16 },
		{ 304, 154, 16 },
		{ 304, 155, 16 },
		{ 304, 156, 16 },
		{ 304, 157, 16 },
		{ 304, 158, 16 },
		{ 304, 159, 16 },
		{ 304, 160, 16 },
		{ 304, 161, 16 },
		{ 304, 162, 16 },
		{ 304, 163, 16 },
		{ 304, 164, 16 },
		{ 304, 165, 16 },
		{ 304, 166, 16 },
		{ 304, 167, 16 },
		{ 304, 168, 16 },
		{ 304, 169, 16 },
		{ 304, 170, 16 },
		{ 304, 171, 16 },
		{ 304, 172, 16 },
		{ 304, 173, 16 },
		{ 304, 174, 16 },
		{ 304, 175, 16 },
		{ 304, 176, 16 },
		{ 304, 177, 16 },
		{ 304, 178, 16 },
		{ 304, 179, 16 },
		{ 304, 180, 16 },
		{ 304, 181, 16 },
		{ 304, 182, 16 },
		{ 304, 183, 16 },
		{ 304, 184, 16 },
		{ 304, 185, 16 },
		{ 304, 186, 16 },
		{ 304, 187, 16 },
		{ 304, 188, 16 },
		{ 304, 189, 16 },
		{ 304, 190, 16 },
		{ 304, 191, 16 },
		{ 304, 192, 16 },
		{ 304, 193, 16 },
		{ 304, 194, 16 },
		{ 305, 131, 16 },
		{ 305, 132, 16 },
		{ 305, 133, 16 },
		{ 305, 134, 16 },
		{ 305, 135, 16 },
		{ 305, 136, 16 },
		{ 305, 137, 16 },
		{ 305, 138, 16 },
		{ 305, 139, 16 },
		{ 305, 140, 16 },
		{ 305, 141, 16 },
		{ 305, 142, 16 },
		{ 305, 143, 16 },
		{ 305, 144, 16 },
		{ 305, 145, 16 },
		{ 305, 146, 16 },
		{ 305, 147, 16 },
		{ 305, 148, 16 },
		{ 305, 149, 16 },
		{ 305, 150, 16 },
		{ 305, 151, 16 },
		{ 305, 152, 16 },
		{ 305, 153, 16 },
		{ 305, 154, 16 },
		{ 305, 155, 16 },
		{ 305, 156, 16 },
		{ 305, 157, 16 },
		{ 305, 158, 16 },
		{ 305, 159, 16 },
		{ 305, 160, 16 },
		{ 305, 161, 16 },
		{ 305, 162, 16 },
		{ 305, 163, 16 },
		{ 305, 164, 16 },
		{ 305, 165, 16 },
		{ 305, 166, 16 },
		{ 305, 167, 16 },
		{ 305, 168, 16 },
		{ 305, 169, 16 },
		{ 305, 170, 16 },
		{ 305, 171, 16 },
		{ 305, 172, 16 },
		{ 305, 173, 16 },
		{ 305, 174, 16 },
		{ 305, 175, 16 },
		{ 305, 176, 16 },
		{ 305, 177, 16 },
		{ 305, 178, 16 },
		{ 305, 179, 16 },
		{ 305, 180, 16 },
		{ 305, 181, 16 },
		{ 305, 182, 16 },
		{ 305, 183, 16 },
		{ 305, 184, 16 },
		{ 305, 185, 16 },
		{ 305, 186, 16 },
		{ 305, 187, 16 },
		{ 305, 188, 16 },
		{ 305, 189, 16 },
		{ 305, 190, 16 },
		{ 305, 191, 16 },
		{ 305, 192, 16 },
		{ 306, 131, 16 },
		{ 306, 132, 16 },
		{ 306, 133, 16 },
		{ 306, 134, 16 },
		{ 306, 135, 16 },
		{ 306, 136, 16 },
		{ 306, 137, 16 },
		{ 306, 138, 16 },
		{ 306, 139, 16 },
		{ 306, 140, 16 },
		{ 306, 141, 16 },
		{ 306, 142, 16 },
		{ 306, 143, 16 },
		{ 306, 144, 16 },
		{ 306, 145, 16 },
		{ 306, 146, 16 },
		{ 306, 147, 16 },
		{ 306, 148, 16 },
		{ 306, 149, 16 },
		{ 306, 150, 16 },
		{ 306, 151, 16 },
		{ 306, 152, 16 },
		{ 306, 153, 16 },
		{ 306, 154, 16 },
		{ 306, 155, 16 },
		{ 306, 156, 16 },
		{ 306, 157, 16 },
		{ 306, 158, 16 },
		{ 306, 159, 16 },
		{ 306, 160, 16 },
		{ 306, 161, 16 },
		{ 306, 162, 16 },
		{ 306, 163, 16 },
		{ 306, 164, 16 },
		{ 306, 165, 16 },
		{ 306, 166, 16 },
		{ 306, 167, 16 },
		{ 306, 168, 16 },
		{ 306, 169, 16 },
		{ 306, 170, 16 },
		{ 306, 171, 16 },
		{ 306, 172, 16 },
		{ 306, 173, 16 },
		{ 306, 174, 16 },
		{ 306, 175, 16 },
		{ 306, 176, 16 },
		{ 306, 177, 16 },
		{ 306, 178, 16 },
		{ 306, 179, 16 },
		{ 306, 180, 16 },
		{ 306, 181, 16 },
		{ 306, 182, 16 },
		{ 306, 183, 16 },
		{ 306, 184, 16 },
		{ 306, 185, 16 },
		{ 306, 186, 16 },
		{ 306, 187, 16 },
		{ 306, 188, 16 },
		{ 306, 189, 16 },
		{ 306, 190, 16 },
		{ 306, 191, 16 },
		{ 307, 131, 16 },
		{ 307, 132, 16 },
		{ 307, 133, 16 },
		{ 307, 134, 16 },
		{ 307, 135, 16 },
		{ 307, 136, 16 },
		{ 307, 137, 16 },
		{ 307, 138, 16 },
		{ 307, 139, 16 },
		{ 307, 140, 16 },
		{ 307, 141, 16 },
		{ 307, 142, 16 },
		{ 307, 143, 16 },
		{ 307, 144, 16 },
		{ 307, 145, 16 },
		{ 307, 146, 16 },
		{ 307, 147, 16 },
		{ 307, 148, 16 },
		{ 307, 149, 16 },
		{ 307, 150, 16 },
		{ 307, 151, 16 },
		{ 307, 152, 16 },
		{ 307, 153, 16 },
		{ 307, 154, 16 },
		{ 307, 155, 16 },
		{ 307, 156, 16 },
		{ 307, 157, 16 },
		{ 307, 158, 16 },
		{ 307, 159, 16 },
		{ 307, 160, 16 },
		{ 307, 161, 16 },
		{ 307, 162, 16 },
		{ 307, 163, 16 },
		{ 307, 164, 16 },
		{ 307, 165, 16 },
		{ 307, 166, 16 },
		{ 307, 167, 16 },
		{ 307, 168, 16 },
		{ 307, 169, 16 },
		{ 307, 170, 16 },
		{ 307, 171, 16 },
		{ 307, 172, 16 },
		{ 307, 173, 16 },
		{ 307, 174, 16 },
		{ 307, 175, 16 },
		{ 307, 176, 16 },
		{ 307, 177, 16 },
		{ 307, 178, 16 },
		{ 307, 179, 16 },
		{ 307, 180, 16 },
		{ 307, 181, 16 },
		{ 307, 182, 16 },
		{ 307, 183, 16 },
		{ 307, 184, 16 },
		{ 307, 185, 16 },
		{ 307, 186, 16 },
		{ 307, 187, 16 },
		{ 307, 188, 16 },
		{ 307, 189, 16 },
		{ 307, 190, 16 },
		{ 308, 131, 16 },
		{ 308, 132, 16 },
		{ 308, 133, 16 },
		{ 308, 134, 16 },
		{ 308, 135, 16 },
		{ 308, 136, 16 },
		{ 308, 137, 16 },
		{ 308, 138, 16 },
		{ 308, 139, 16 },
		{ 308, 140, 16 },
		{ 308, 141, 16 },
		{ 308, 142, 16 },
		{ 308, 143, 16 },
		{ 308, 144, 16 },
		{ 308, 145, 16 },
		{ 308, 146, 16 },
		{ 308, 147, 16 },
		{ 308, 148, 16 },
		{ 308, 149, 16 },
		{ 308, 150, 16 },
		{ 308, 151, 16 },
		{ 308, 152, 16 },
		{ 308, 153, 16 },
		{ 308, 154, 16 },
		{ 308, 155, 16 },
		{ 308, 156, 16 },
		{ 308, 157, 16 },
		{ 308, 158, 16 },
		{ 308, 159, 16 },
		{ 308, 160, 16 },
		{ 308, 161, 16 },
		{ 308, 162, 16 },
		{ 308, 163, 16 },
		{ 308, 164, 16 },
		{ 308, 165, 16 },
		{ 308, 166, 16 },
		{ 308, 167, 16 },
		{ 308, 168, 16 },
		{ 308, 169, 16 },
		{ 308, 170, 16 },
		{ 308, 171, 16 },
		{ 308, 172, 16 },
		{ 308, 173, 16 },
		{ 308, 174, 16 },
		{ 308, 175, 16 },
		{ 308, 176, 16 },
		{ 308, 177, 16 },
		{ 308, 178, 16 },
		{ 308, 179, 16 },
		{ 308, 180, 16 },
		{ 308, 181, 16 },
		{ 308, 182, 16 },
		{ 308, 183, 16 },
		{ 308, 184, 16 },
		{ 308, 185, 16 },
		{ 308, 186, 16 },
		{ 308, 187, 16 },
		{ 308, 188, 16 },
		{ 309, 131, 16 },
		{ 309, 132, 16 },
		{ 309, 133, 16 },
		{ 309, 134, 16 },
		{ 309, 135, 16 },
		{ 309, 136, 16 },
		{ 309, 137, 16 },
		{ 309, 138, 16 },
		{ 309, 139, 16 },
		{ 309, 140, 16 },
		{ 309, 141, 16 },
		{ 309, 142, 16 },
		{ 309, 143, 16 },
		{ 309, 144, 16 },
		{ 309, 145, 16 },
		{ 309, 146, 16 },
		{ 309, 147, 16 },
		{ 309, 148, 16 },
		{ 309, 149, 16 },
		{ 309, 150, 16 },
		{ 309, 151, 16 },
		{ 309, 152, 16 },
		{ 309, 153, 16 },
		{ 309, 154, 16 },
		{ 309, 155, 16 },
		{ 309, 156, 16 },
		{ 309, 157, 16 },
		{ 309, 158, 16 },
		{ 309, 159, 16 },
		{ 309, 160, 16 },
		{ 309, 161, 16 },
		{ 309, 162, 16 },
		{ 309, 163, 16 },
		{ 309, 164, 16 },
		{ 309, 165, 16 },
		{ 309, 166, 16 },
		{ 309, 167, 16 },
		{ 309, 168, 16 },
		{ 309, 169, 16 },
		{ 309, 170, 16 },
		{ 309, 171, 16 },
		{ 309, 172, 16 },
		{ 309, 173, 16 },
		{ 309, 174, 16 },
		{ 309, 175, 16 },
		{ 309, 176, 16 },
		{ 309, 177, 16 },
		{ 309, 178, 16 },
		{ 309, 179, 16 },
		{ 309, 180, 16 },
		{ 309, 181, 16 },
		{ 309, 182, 16 },
		{ 309, 183, 16 },
		{ 309, 184, 16 },
		{ 309, 185, 16 },
		{ 309, 186, 16 },
		{ 309, 187, 16 },
		{ 310, 131, 16 },
		{ 310, 132, 16 },
		{ 310, 133, 16 },
		{ 310, 134, 16 },
		{ 310, 135, 16 },
		{ 310, 136, 16 },
		{ 310, 137, 16 },
		{ 310, 138, 16 },
		{ 310, 139, 16 },
		{ 310, 140, 16 },
		{ 310, 141, 16 },
		{ 310, 142, 16 },
		{ 310, 143, 16 },
		{ 310, 144, 16 },
		{ 310, 145, 16 },
		{ 310, 146, 16 },
		{ 310, 147, 16 },
		{ 310, 148, 16 },
		{ 310, 149, 16 },
		{ 310, 150, 16 },
		{ 310, 151, 16 },
		{ 310, 152, 16 },
		{ 310, 153, 16 },
		{ 310, 154, 16 },
		{ 310, 155, 16 },
		{ 310, 156, 16 },
		{ 310, 157, 16 },
		{ 310, 158, 16 },
		{ 310, 159, 16 },
		{ 310, 160, 16 },
		{ 310, 161, 16 },
		{ 310, 162, 16 },
		{ 310, 163, 16 },
		{ 310, 164, 16 },
		{ 310, 165, 16 },
		{ 310, 166, 16 },
		{ 310, 167, 16 },
		{ 310, 168, 16 },
		{ 310, 169, 16 },
		{ 310, 170, 16 },
		{ 310, 171, 16 },
		{ 310, 172, 16 },
		{ 310, 173, 16 },
		{ 310, 174, 16 },
		{ 310, 175, 16 },
		{ 310, 176, 16 },
		{ 310, 177, 16 },
		{ 310, 178, 16 },
		{ 310, 179, 16 },
		{ 310, 180, 16 },
		{ 310, 181, 16 },
		{ 310, 182, 16 },
		{ 310, 183, 16 },
		{ 310, 184, 16 },
		{ 310, 185, 16 },
		{ 310, 186, 16 },
		{ 311, 131, 16 },
		{ 311, 132, 16 },
		{ 311, 133, 16 },
		{ 311, 134, 16 },
		{ 311, 135, 16 },
		{ 311, 136, 16 },
		{ 311, 137, 16 },
		{ 311, 138, 16 },
		{ 311, 139, 16 },
		{ 311, 140, 16 },
		{ 311, 141, 16 },
		{ 311, 142, 16 },
		{ 311, 143, 16 },
		{ 311, 144, 16 },
		{ 311, 145, 16 },
		{ 311, 146, 16 },
		{ 311, 147, 16 },
		{ 311, 148, 16 },
		{ 311, 149, 16 },
		{ 311, 150, 16 },
		{ 311, 151, 16 },
		{ 311, 152, 16 },
		{ 311, 153, 16 },
		{ 311, 154, 16 },
		{ 311, 155, 16 },
		{ 311, 156, 16 },
		{ 311, 157, 16 },
		{ 311, 158, 16 },
		{ 311, 159, 16 },
		{ 311, 160, 16 },
		{ 311, 161, 16 },
		{ 311, 162, 16 },
		{ 311, 163, 16 },
		{ 311, 164, 16 },
		{ 311, 165, 16 },
		{ 311, 166, 16 },
		{ 311, 167, 16 },
		{ 311, 168, 16 },
		{ 311, 169, 16 },
		{ 311, 170, 16 },
		{ 311, 171, 16 },
		{ 311, 172, 16 },
		{ 311, 173, 16 },
		{ 311, 174, 16 },
		{ 311, 175, 16 },
		{ 311, 176, 16 },
		{ 311, 177, 16 },
		{ 311, 178, 16 },
		{ 311, 179, 16 },
		{ 311, 180, 16 },
		{ 311, 181, 16 },
		{ 311, 182, 16 },
		{ 311, 183, 16 },
		{ 311, 184, 16 },
		{ 312, 131, 16 },
		{ 312, 132, 16 },
		{ 312, 133, 16 },
		{ 312, 134, 16 },
		{ 312, 135, 16 },
		{ 312, 136, 16 },
		{ 312, 137, 16 },
		{ 312, 138, 16 },
		{ 312, 139, 16 },
		{ 312, 140, 16 },
		{ 312, 141, 16 },
		{ 312, 142, 16 },
		{ 312, 143, 16 },
		{ 312, 144, 16 },
		{ 312, 145, 16 },
		{ 312, 146, 16 },
		{ 312, 147, 16 },
		{ 312, 148, 16 },
		{ 312, 149, 16 },
		{ 312, 150, 16 },
		{ 312, 151, 16 },
		{ 312, 152, 16 },
		{ 312, 153, 16 },
		{ 312, 154, 16 },
		{ 312, 155, 16 },
		{ 312, 156, 16 },
		{ 312, 157, 16 },
		{ 312, 158, 16 },
		{ 312, 159, 16 },
		{ 312, 160, 16 },
		{ 312, 161, 16 },
		{ 312, 162, 16 },
		{ 312, 163, 16 },
		{ 312, 164, 16 },
		{ 312, 165, 16 },
		{ 312, 166, 16 },
		{ 312, 167, 16 },
		{ 312, 168, 16 },
		{ 312, 169, 16 },
		{ 312, 170, 16 },
		{ 312, 171, 16 },
		{ 312, 172, 16 },
		{ 312, 173, 16 },
		{ 312, 174, 16 },
		{ 312, 175, 16 },
		{ 312, 176, 16 },
		{ 312, 177, 16 },
		{ 312, 178, 16 },
		{ 312, 179, 16 },
		{ 312, 180, 16 },
		{ 312, 181, 16 },
		{ 312, 182, 16 },
		{ 312, 183, 16 },
		{ 313, 131, 16 },
		{ 313, 132, 16 },
		{ 313, 133, 16 },
		{ 313, 134, 16 },
		{ 313, 135, 16 },
		{ 313, 136, 16 },
		{ 313, 137, 16 },
		{ 313, 138, 16 },
		{ 313, 139, 16 },
		{ 313, 140, 16 },
		{ 313, 141, 16 },
		{ 313, 142, 16 },
		{ 313, 143, 16 },
		{ 313, 144, 16 },
		{ 313, 145, 16 },
		{ 313, 146, 16 },
		{ 313, 147, 16 },
		{ 313, 148, 16 },
		{ 313, 149, 16 },
		{ 313, 150, 16 },
		{ 313, 151, 16 },
		{ 313, 152, 16 },
		{ 313, 153, 16 },
		{ 313, 154, 16 },
		{ 313, 155, 16 },
		{ 313, 156, 16 },
		{ 313, 157, 16 },
		{ 313, 158, 16 },
		{ 313, 159, 16 },
		{ 313, 160, 16 },
		{ 313, 161, 16 },
		{ 313, 162, 16 },
		{ 313, 163, 16 },
		{ 313, 164, 16 },
		{ 313, 165, 16 },
		{ 313, 166, 16 },
		{ 313, 167, 16 },
		{ 313, 168, 16 },
		{ 313, 169, 16 },
		{ 313, 170, 16 },
		{ 313, 171, 16 },
		{ 313, 172, 16 },
		{ 313, 173, 16 },
		{ 313, 174, 16 },
		{ 313, 175, 16 },
		{ 313, 176, 16 },
		{ 313, 177, 16 },
		{ 313, 178, 16 },
		{ 313, 179, 16 },
		{ 313, 180, 16 },
		{ 313, 181, 16 },
		{ 313, 182, 16 },
		{ 314, 131, 16 },
		{ 314, 132, 16 },
		{ 314, 133, 16 },
		{ 314, 134, 16 },
		{ 314, 135, 16 },
		{ 314, 136, 16 },
		{ 314, 137, 16 },
		{ 314, 138, 16 },
		{ 314, 139, 16 },
		{ 314, 140, 16 },
		{ 314, 141, 16 },
		{ 314, 142, 16 },
		{ 314, 143, 16 },
		{ 314, 144, 16 },
		{ 314, 145, 16 },
		{ 314, 146, 16 },
		{ 314, 147, 16 },
		{ 314, 148, 16 },
		{ 314, 149, 16 },
		{ 314, 150, 16 },
		{ 314, 151, 16 },
		{ 314, 152, 16 },
		{ 314, 153, 16 },
		{ 314, 154, 16 },
		{ 314, 155, 16 },
		{ 314, 156, 16 },
		{ 314, 157, 16 },
		{ 314, 158, 16 },
		{ 314, 159, 16 },
		{ 314, 160, 16 },
		{ 314, 161, 16 },
		{ 314, 162, 16 },
		{ 314, 163, 16 },
		{ 314, 164, 16 },
		{ 314, 165, 16 },
		{ 314, 166, 16 },
		{ 314, 167, 16 },
		{ 314, 168, 16 },
		{ 314, 169, 16 },
		{ 314, 170, 16 },
		{ 314, 171, 16 },
		{ 314, 172, 16 },
		{ 314, 173, 16 },
		{ 314, 174, 16 },
		{ 314, 175, 16 },
		{ 314, 176, 16 },
		{ 314, 177, 16 },
		{ 314, 178, 16 },
		{ 314, 179, 16 },
		{ 314, 180, 16 },
		{ 315, 131, 16 },
		{ 315, 132, 16 },
		{ 315, 133, 16 },
		{ 315, 134, 16 },
		{ 315, 135, 16 },
		{ 315, 136, 16 },
		{ 315, 137, 16 },
		{ 315, 138, 16 },
		{ 315, 139, 16 },
		{ 315, 140, 16 },
		{ 315, 141, 16 },
		{ 315, 142, 16 },
		{ 315, 143, 16 },
		{ 315, 144, 16 },
		{ 315, 145, 16 },
		{ 315, 146, 16 },
		{ 315, 147, 16 },
		{ 315, 148, 16 },
		{ 315, 149, 16 },
		{ 315, 150, 16 },
		{ 315, 151, 16 },
		{ 315, 152, 16 },
		{ 315, 153, 16 },
		{ 315, 154, 16 },
		{ 315, 155, 16 },
		{ 315, 156, 16 },
		{ 315, 157, 16 },
		{ 315, 158, 16 },
		{ 315, 159, 16 },
		{ 315, 160, 16 },
		{ 315, 161, 16 },
		{ 315, 162, 16 },
		{ 315, 163, 16 },
		{ 315, 164, 16 },
		{ 315, 165, 16 },
		{ 315, 166, 16 },
		{ 315, 167, 16 },
		{ 315, 168, 16 },
		{ 315, 169, 16 },
		{ 315, 170, 16 },
		{ 315, 171, 16 },
		{ 315, 172, 16 },
		{ 315, 173, 16 },
		{ 315, 174, 16 },
		{ 315, 175, 16 },
		{ 315, 176, 16 },
		{ 315, 177, 16 },
		{ 315, 178, 16 },
		{ 315, 179, 16 },
		{ 316, 131, 16 },
		{ 316, 132, 16 },
		{ 316, 133, 16 },
		{ 316, 134, 16 },
		{ 316, 135, 16 },
		{ 316, 136, 16 },
		{ 316, 137, 16 },
		{ 316, 138, 16 },
		{ 316, 139, 16 },
		{ 316, 140, 16 },
		{ 316, 141, 16 },
		{ 316, 142, 16 },
		{ 316, 143, 16 },
		{ 316, 144, 16 },
		{ 316, 145, 16 },
		{ 316, 146, 16 },
		{ 316, 147, 16 },
		{ 316, 148, 16 },
		{ 316, 149, 16 },
		{ 316, 150, 16 },
		{ 316, 151, 16 },
		{ 316, 152, 16 },
		{ 316, 153, 16 },
		{ 316, 154, 16 },
		{ 316, 155, 16 },
		{ 316, 156, 16 },
		{ 316, 157, 16 },
		{ 316, 158, 16 },
		{ 316, 159, 16 },
		{ 316, 160, 16 },
		{ 316, 161, 16 },
		{ 316, 162, 16 },
		{ 316, 163, 16 },
		{ 316, 164, 16 },
		{ 316, 165, 16 },
		{ 316, 166, 16 },
		{ 316, 167, 16 },
		{ 316, 168, 16 },
		{ 316, 169, 16 },
		{ 316, 170, 16 },
		{ 316, 171, 16 },
		{ 316, 172, 16 },
		{ 316, 173, 16 },
		{ 316, 174, 16 },
		{ 316, 175, 16 },
		{ 316, 176, 16 },
		{ 316, 177, 16 },
		{ 316, 178, 16 },
		{ 317, 131, 16 },
		{ 317, 132, 16 },
		{ 317, 133, 16 },
		{ 317, 134, 16 },
		{ 317, 135, 16 },
		{ 317, 136, 16 },
		{ 317, 137, 16 },
		{ 317, 138, 16 },
		{ 317, 139, 16 },
		{ 317, 140, 16 },
		{ 317, 141, 16 },
		{ 317, 142, 16 },
		{ 317, 143, 16 },
		{ 317, 144, 16 },
		{ 317, 145, 16 },
		{ 317, 146, 16 },
		{ 317, 147, 16 },
		{ 317, 148, 16 },
		{ 317, 149, 16 },
		{ 317, 150, 16 },
		{ 317, 151, 16 },
		{ 317, 152, 16 },
		{ 317, 153, 16 },
		{ 317, 154, 16 },
		{ 317, 155, 16 },
		{ 317, 156, 16 },
		{ 317, 157, 16 },
		{ 317, 158, 16 },
		{ 317, 159, 16 },
		{ 317, 160, 16 },
		{ 317, 161, 16 },
		{ 317, 162, 16 },
		{ 317, 163, 16 },
		{ 317, 164, 16 },
		{ 317, 165, 16 },
		{ 317, 166, 16 },
		{ 317, 167, 16 },
		{ 317, 168, 16 },
		{ 317, 169, 16 },
		{ 317, 170, 16 },
		{ 317, 171, 16 },
		{ 317, 172, 16 },
		{ 317, 173, 16 },
		{ 317, 174, 16 },
		{ 317, 175, 16 },
		{ 317, 176, 16 },
		{ 318, 131, 16 },
		{ 318, 132, 16 },
		{ 318, 133, 16 },
		{ 318, 134, 16 },
		{ 318, 135, 16 },
		{ 318, 136, 16 },
		{ 318, 137, 16 },
		{ 318, 138, 16 },
		{ 318, 139, 16 },
		{ 318, 140, 16 },
		{ 318, 141, 16 },
		{ 318, 142, 16 },
		{ 318, 143, 16 },
		{ 318, 144, 16 },
		{ 318, 145, 16 },
		{ 318, 146, 16 },
		{ 318, 147, 16 },
		{ 318, 148, 16 },
		{ 318, 149, 16 },
		{ 318, 150, 16 },
		{ 318, 151, 16 },
		{ 318, 152, 16 },
		{ 318, 153, 16 },
		{ 318, 154, 16 },
		{ 318, 155, 16 },
		{ 318, 156, 16 },
		{ 318, 157, 16 },
		{ 318, 158, 16 },
		{ 318, 159, 16 },
		{ 318, 160, 16 },
		{ 318, 161, 16 },
		{ 318, 162, 16 },
		{ 318, 163, 16 },
		{ 318, 164, 16 },
		{ 318, 165, 16 },
		{ 318, 166, 16 },
		{ 318, 167, 16 },
		{ 318, 168, 16 },
		{ 318, 169, 16 },
		{ 318, 170, 16 },
		{ 318, 171, 16 },
		{ 318, 172, 16 },
		{ 318, 173, 16 },
		{ 318, 174, 16 },
		{ 318, 175, 16 },
		{ 319, 131, 16 },
		{ 319, 132, 16 },
		{ 319, 133, 16 },
		{ 319, 134, 16 },
		{ 319, 135, 16 },
		{ 319, 136, 16 },
		{ 319, 137, 16 },
		{ 319, 138, 16 },
		{ 319, 139, 16 },
		{ 319, 140, 16 },
		{ 319, 141, 16 },
		{ 319, 142, 16 },
		{ 319, 143, 16 },
		{ 319, 144, 16 },
		{ 319, 145, 16 },
		{ 319, 146, 16 },
		{ 319, 147, 16 },
		{ 319, 148, 16 },
		{ 319, 149, 16 },
		{ 319, 150, 16 },
		{ 319, 151, 16 },
		{ 319, 152, 16 },
		{ 319, 153, 16 },
		{ 319, 154, 16 },
		{ 319, 155, 16 },
		{ 319, 156, 16 },
		{ 319, 157, 16 },
		{ 319, 158, 16 },
		{ 319, 159, 16 },
		{ 319, 160, 16 },
		{ 319, 161, 16 },
		{ 319, 162, 16 },
		{ 319, 163, 16 },
		{ 319, 164, 16 },
		{ 319, 165, 16 },
		{ 319, 166, 16 },
		{ 319, 167, 16 },
		{ 319, 168, 16 },
		{ 319, 169, 16 },
		{ 319, 170, 16 },
		{ 319, 171, 16 },
		{ 319, 172, 16 },
		{ 319, 173, 16 },
		{ 319, 174, 16 },
		{ 320, 131, 16 },
		{ 320, 132, 16 },
		{ 320, 133, 16 },
		{ 320, 134, 16 },
		{ 320, 135, 16 },
		{ 320, 136, 16 },
		{ 320, 137, 16 },
		{ 320, 138, 16 },
		{ 320, 139, 16 },
		{ 320, 140, 16 },
		{ 320, 141, 16 },
		{ 320, 142, 16 },
		{ 320, 143, 16 },
		{ 320, 144, 16 },
		{ 320, 145, 16 },
		{ 320, 146, 16 },
		{ 320, 147, 16 },
		{ 320, 148, 16 },
		{ 320, 149, 16 },
		{ 320, 150, 16 },
		{ 320, 151, 16 },
		{ 320, 152, 16 },
		{ 320, 153, 16 },
		{ 320, 154, 16 },
		{ 320, 155, 16 },
		{ 320, 156, 16 },
		{ 320, 157, 16 },
		{ 320, 158, 16 },
		{ 320, 159, 16 },
		{ 320, 160, 16 },
		{ 320, 161, 16 },
		{ 320, 162, 16 },
		{ 320, 163, 16 },
		{ 320, 164, 16 },
		{ 320, 165, 16 },
		{ 320, 166, 16 },
		{ 320, 167, 16 },
		{ 320, 168, 16 },
		{ 320, 169, 16 },
		{ 320, 170, 16 },
		{ 320, 171, 16 },
		{ 320, 172, 16 },
		{ 321, 131, 16 },
		{ 321, 132, 16 },
		{ 321, 133, 16 },
		{ 321, 134, 16 },
		{ 321, 135, 16 },
		{ 321, 136, 16 },
		{ 321, 137, 16 },
		{ 321, 138, 16 },
		{ 321, 139, 16 },
		{ 321, 140, 16 },
		{ 321, 141, 16 },
		{ 321, 142, 16 },
		{ 321, 143, 16 },
		{ 321, 144, 16 },
		{ 321, 145, 16 },
		{ 321, 146, 16 },
		{ 321, 147, 16 },
		{ 321, 148, 16 },
		{ 321, 149, 16 },
		{ 321, 150, 16 },
		{ 321, 151, 16 },
		{ 321, 152, 16 },
		{ 321, 153, 16 },
		{ 321, 154, 16 },
		{ 321, 155, 16 },
		{ 321, 156, 16 },
		{ 321, 157, 16 },
		{ 321, 158, 16 },
		{ 321, 159, 16 },
		{ 321, 160, 16 },
		{ 321, 161, 16 },
		{ 321, 162, 16 },
		{ 321, 163, 16 },
		{ 321, 164, 16 },
		{ 321, 165, 16 },
		{ 321, 166, 16 },
		{ 321, 167, 16 },
		{ 321, 168, 16 },
		{ 321, 169, 16 },
		{ 321, 170, 16 },
		{ 321, 171, 16 },
		{ 322, 131, 16 },
		{ 322, 132, 16 },
		{ 322, 133, 16 },
		{ 322, 134, 16 },
		{ 322, 135, 16 },
		{ 322, 136, 16 },
		{ 322, 137, 16 },
		{ 322, 138, 16 },
		{ 322, 139, 16 },
		{ 322, 140, 16 },
		{ 322, 141, 16 },
		{ 322, 142, 16 },
		{ 322, 143, 16 },
		{ 322, 144, 16 },
		{ 322, 145, 16 },
		{ 322, 146, 16 },
		{ 322, 147, 16 },
		{ 322, 148, 16 },
		{ 322, 149, 16 },
		{ 322, 150, 16 },
		{ 322, 151, 16 },
		{ 322, 152, 16 },
		{ 322, 153, 16 },
		{ 322, 154, 16 },
		{ 322, 155, 16 },
		{ 322, 156, 16 },
		{ 322, 157, 16 },
		{ 322, 158, 16 },
		{ 322, 159, 16 },
		{ 322, 160, 16 },
		{ 322, 161, 16 },
		{ 322, 162, 16 },
		{ 322, 163, 16 },
		{ 322, 164, 16 },
		{ 322, 165, 16 },
		{ 322, 166, 16 },
		{ 322, 167, 16 },
		{ 322, 168, 16 },
		{ 322, 169, 16 },
		{ 322, 170, 16 },
		{ 323, 131, 16 },
		{ 323, 132, 16 },
		{ 323, 133, 16 },
		{ 323, 134, 16 },
		{ 323, 135, 16 },
		{ 323, 136, 16 },
		{ 323, 137, 16 },
		{ 323, 138, 16 },
		{ 323, 139, 16 },
		{ 323, 140, 16 },
		{ 323, 141, 16 },
		{ 323, 142, 16 },
		{ 323, 143, 16 },
		{ 323, 144, 16 },
		{ 323, 145, 16 },
		{ 323, 146, 16 },
		{ 323, 147, 16 },
		{ 323, 148, 16 },
		{ 323, 149, 16 },
		{ 323, 150, 16 },
		{ 323, 151, 16 },
		{ 323, 152, 16 },
		{ 323, 153, 16 },
		{ 323, 154, 16 },
		{ 323, 155, 16 },
		{ 323, 156, 16 },
		{ 323, 157, 16 },
		{ 323, 158, 16 },
		{ 323, 159, 16 },
		{ 323, 160, 16 },
		{ 323, 161, 16 },
		{ 323, 162, 16 },
		{ 323, 163, 16 },
		{ 323, 164, 16 },
		{ 323, 165, 16 },
		{ 323, 166, 16 },
		{ 323, 167, 16 },
		{ 323, 168, 16 },
		{ 324, 131, 16 },
		{ 324, 132, 16 },
		{ 324, 133, 16 },
		{ 324, 134, 16 },
		{ 324, 135, 16 },
		{ 324, 136, 16 },
		{ 324, 137, 16 },
		{ 324, 138, 16 },
		{ 324, 139, 16 },
		{ 324, 140, 16 },
		{ 324, 141, 16 },
		{ 324, 142, 16 },
		{ 324, 143, 16 },
		{ 324, 144, 16 },
		{ 324, 145, 16 },
		{ 324, 146, 16 },
		{ 324, 147, 16 },
		{ 324, 148, 16 },
		{ 324, 149, 16 },
		{ 324, 150, 16 },
		{ 324, 151, 16 },
		{ 324, 152, 16 },
		{ 324, 153, 16 },
		{ 324, 154, 16 },
		{ 324, 155, 16 },
		{ 324, 156, 16 },
		{ 324, 157, 16 },
		{ 324, 158, 16 },
		{ 324, 159, 16 },
		{ 324, 160, 16 },
		{ 324, 161, 16 },
		{ 324, 162, 16 },
		{ 324, 163, 16 },
		{ 324, 164, 16 },
		{ 324, 165, 16 },
		{ 324, 166, 16 },
		{ 324, 167, 16 },
		{ 325, 131, 16 },
		{ 325, 132, 16 },
		{ 325, 133, 16 },
		{ 325, 134, 16 },
		{ 325, 135, 16 },
		{ 325, 136, 16 },
		{ 325, 137, 16 },
		{ 325, 138, 16 },
		{ 325, 139, 16 },
		{ 325, 140, 16 },
		{ 325, 141, 16 },
		{ 325, 142, 16 },
		{ 325, 143, 16 },
		{ 325, 144, 16 },
		{ 325, 145, 16 },
		{ 325, 146, 16 },
		{ 325, 147, 16 },
		{ 325, 148, 16 },
		{ 325, 149, 16 },
		{ 325, 150, 16 },
		{ 325, 151, 16 },
		{ 325, 152, 16 },
		{ 325, 153, 16 },
		{ 325, 154, 16 },
		{ 325, 155, 16 },
		{ 325, 156, 16 },
		{ 325, 157, 16 },
		{ 325, 158, 16 },
		{ 325, 159, 16 },
		{ 325, 160, 16 },
		{ 325, 161, 16 },
		{ 325, 162, 16 },
		{ 325, 163, 16 },
		{ 325, 164, 16 },
		{ 325, 165, 16 },
		{ 325, 166, 16 },
		{ 326, 131, 16 },
		{ 326, 132, 16 },
		{ 326, 133, 16 },
		{ 326, 134, 16 },
		{ 326, 135, 16 },
		{ 326, 136, 16 },
		{ 326, 137, 16 },
		{ 326, 138, 16 },
		{ 326, 139, 16 },
		{ 326, 140, 16 },
		{ 326, 141, 16 },
		{ 326, 142, 16 },
		{ 326, 143, 16 },
		{ 326, 144, 16 },
		{ 326, 145, 16 },
		{ 326, 146, 16 },
		{ 326, 147, 16 },
		{ 326, 148, 16 },
		{ 326, 149, 16 },
		{ 326, 150, 16 },
		{ 326, 151, 16 },
		{ 326, 152, 16 },
		{ 326, 153, 16 },
		{ 326, 154, 16 },
		{ 326, 155, 16 },
		{ 326, 156, 16 },
		{ 326, 157, 16 },
		{ 326, 158, 16 },
		{ 326, 159, 16 },
		{ 326, 160, 16 },
		{ 326, 161, 16 },
		{ 326, 162, 16 },
		{ 326, 163, 16 },
		{ 326, 164, 16 },
		{ 327, 131, 16 },
		{ 327, 132, 16 },
		{ 327, 133, 16 },
		{ 327, 134, 16 },
		{ 327, 135, 16 },
		{ 327, 136, 16 },
		{ 327, 137, 16 },
		{ 327, 138, 16 },
		{ 327, 139, 16 },
		{ 327, 140, 16 },
		{ 327, 141, 16 },
		{ 327, 142, 16 },
		{ 327, 143, 16 },
		{ 327, 144, 16 },
		{ 327, 145, 16 },
		{ 327, 146, 16 },
		{ 327, 147, 16 },
		{ 327, 148, 16 },
		{ 327, 149, 16 },
		{ 327, 150, 16 },
		{ 327, 151, 16 },
		{ 327, 152, 16 },
		{ 327, 153, 16 },
		{ 327, 154, 16 },
		{ 327, 155, 16 },
		{ 327, 156, 16 },
		{ 327, 157, 16 },
		{ 327, 158, 16 },
		{ 327, 159, 16 },
		{ 327, 160, 16 },
		{ 327, 161, 16 },
		{ 327, 162, 16 },
		{ 327, 163, 16 },
		{ 328, 131, 16 },
		{ 328, 132, 16 },
		{ 328, 133, 16 },
		{ 328, 134, 16 },
		{ 328, 135, 16 },
		{ 328, 136, 16 },
		{ 328, 137, 16 },
		{ 328, 138, 16 },
		{ 328, 139, 16 },
		{ 328, 140, 16 },
		{ 328, 141, 16 },
		{ 328, 142, 16 },
		{ 328, 143, 16 },
		{ 328, 144, 16 },
		{ 328, 145, 16 },
		{ 328, 146, 16 },
		{ 328, 147, 16 },
		{ 328, 148, 16 },
		{ 328, 149, 16 },
		{ 328, 150, 16 },
		{ 328, 151, 16 },
		{ 328, 152, 16 },
		{ 328, 153, 16 },
		{ 328, 154, 16 },
		{ 328, 155, 16 },
		{ 328, 156, 16 },
		{ 328, 157, 16 },
		{ 328, 158, 16 },
		{ 328, 159, 16 },
		{ 328, 160, 16 },
		{ 328, 161, 16 },
		{ 328, 162, 16 },
		{ 329, 131, 16 },
		{ 329, 132, 16 },
		{ 329, 133, 16 },
		{ 329, 134, 16 },
		{ 329, 135, 16 },
		{ 329, 136, 16 },
		{ 329, 137, 16 },
		{ 329, 138, 16 },
		{ 329, 139, 16 },
		{ 329, 140, 16 },
		{ 329, 141, 16 },
		{ 329, 142, 16 },
		{ 329, 143, 16 },
		{ 329, 144, 16 },
		{ 329, 145, 16 },
		{ 329, 146, 16 },
		{ 329, 147, 16 },
		{ 329, 148, 16 },
		{ 329, 149, 16 },
		{ 329, 150, 16 },
		{ 329, 151, 16 },
		{ 329, 152, 16 },
		{ 329, 153, 16 },
		{ 329, 154, 16 },
		{ 329, 155, 16 },
		{ 329, 156, 16 },
		{ 329, 157, 16 },
		{ 329, 158, 16 },
		{ 329, 159, 16 },
		{ 329, 160, 16 },
		{ 330, 131, 16 },
		{ 330, 132, 16 },
		{ 330, 133, 16 },
		{ 330, 134, 16 },
		{ 330, 135, 16 },
		{ 330, 136, 16 },
		{ 330, 137, 16 },
		{ 330, 138, 16 },
		{ 330, 139, 16 },
		{ 330, 140, 16 },
		{ 330, 141, 16 },
		{ 330, 142, 16 },
		{ 330, 143, 16 },
		{ 330, 144, 16 },
		{ 330, 145, 16 },
		{ 330, 146, 16 },
		{ 330, 147, 16 },
		{ 330, 148, 16 },
		{ 330, 149, 16 },
		{ 330, 150, 16 },
		{ 330, 151, 16 },
		{ 330, 152, 16 },
		{ 330, 153, 16 },
		{ 330, 154, 16 },
		{ 330, 155, 16 },
		{ 330, 156, 16 },
		{ 330, 157, 16 },
		{ 330, 158, 16 },
		{ 330, 159, 16 },
		{ 331, 131, 16 },
		{ 331, 132, 16 },
		{ 331, 133, 16 },
		{ 331, 134, 16 },
		{ 331, 135, 16 },
		{ 331, 136, 16 },
		{ 331, 137, 16 },
		{ 331, 138, 16 },
		{ 331, 139, 16 },
		{ 331, 140, 16 },
		{ 331, 141, 16 },
		{ 331, 142, 16 },
		{ 331, 143, 16 },
		{ 331, 144, 16 },
		{ 331, 145, 16 },
		{ 331, 146, 16 },
		{ 331, 147, 16 },
		{ 331, 148, 16 },
		{ 331, 149, 16 },
		{ 331, 150, 16 },
		{ 331, 151, 16 },
		{ 331, 152, 16 },
		{ 331, 153, 16 },
		{ 331, 154, 16 },
		{ 331, 155, 16 },
		{ 331, 156, 16 },
		{ 331, 157, 16 },
		{ 331, 158, 16 },
		{ 332, 131, 16 },
		{ 332, 132, 16 },
		{ 332, 133, 16 },
		{ 332, 134, 16 },
		{ 332, 135, 16 },
		{ 332, 136, 16 },
		{ 332, 137, 16 },
		{ 332, 138, 16 },
		{ 332, 139, 16 },
		{ 332, 140, 16 },
		{ 332, 141, 16 },
		{ 332, 142, 16 },
		{ 332, 143, 16 },
		{ 332, 144, 16 },
		{ 332, 145, 16 },
		{ 332, 146, 16 },
		{ 332, 147, 16 },
		{ 332, 148, 16 },
		{ 332, 149, 16 },
		{ 332, 150, 16 },
		{ 332, 151, 16 },
		{ 332, 152, 16 },
		{ 332, 153, 16 },
		{ 332, 154, 16 },
		{ 332, 155, 16 },
		{ 332, 156, 16 },
		{ 333, 131, 16 },
		{ 333, 132, 16 },
		{ 333, 133, 16 },
		{ 333, 134, 16 },
		{ 333, 135, 16 },
		{ 333, 136, 16 },
		{ 333, 137, 16 },
		{ 333, 138, 16 },
		{ 333, 139, 16 },
		{ 333, 140, 16 },
		{ 333, 141, 16 },
		{ 333, 142, 16 },
		{ 333, 143, 16 },
		{ 333, 144, 16 },
		{ 333, 145, 16 },
		{ 333, 146, 16 },
		{ 333, 147, 16 },
		{ 333, 148, 16 },
		{ 333, 149, 16 },
		{ 333, 150, 16 },
		{ 333, 151, 16 },
		{ 333, 152, 16 },
		{ 333, 153, 16 },
		{ 333, 154, 16 },
		{ 333, 155, 16 },
		{ 334, 131, 16 },
		{ 334, 132, 16 },
		{ 334, 133, 16 },
		{ 334, 134, 16 },
		{ 334, 135, 16 },
		{ 334, 136, 16 },
		{ 334, 137, 16 },
		{ 334, 138, 16 },
		{ 334, 139, 16 },
		{ 334, 140, 16 },
		{ 334, 141, 16 },
		{ 334, 142, 16 },
		{ 334, 143, 16 },
		{ 334, 144, 16 },
		{ 334, 145, 16 },
		{ 334, 146, 16 },
		{ 334, 147, 16 },
		{ 334, 148, 16 },
		{ 334, 149, 16 },
		{ 334, 150, 16 },
		{ 334, 151, 16 },
		{ 334, 152, 16 },
		{ 334, 153, 16 },
		{ 334, 154, 16 },
		{ 335, 131, 16 },
		{ 335, 132, 16 },
		{ 335, 133, 16 },
		{ 335, 134, 16 },
		{ 335, 135, 16 },
		{ 335, 136, 16 },
		{ 335, 137, 16 },
		{ 335, 138, 16 },
		{ 335, 139, 16 },
		{ 335, 140, 16 },
		{ 335, 141, 16 },
		{ 335, 142, 16 },
		{ 335, 143, 16 },
		{ 335, 144, 16 },
		{ 335, 145, 16 },
		{ 335, 146, 16 },
		{ 335, 147, 16 },
		{ 335, 148, 16 },
		{ 335, 149, 16 },
		{ 335, 150, 16 },
		{ 335, 151, 16 },
		{ 335, 152, 16 },
		{ 336, 131, 16 },
		{ 336, 132, 16 },
		{ 336, 133, 16 },
		{ 336, 134, 16 },
		{ 336, 135, 16 },
		{ 336, 136, 16 },
		{ 336, 137, 16 },
		{ 336, 138, 16 },
		{ 336, 139, 16 },
		{ 336, 140, 16 },
		{ 336, 141, 16 },
		{ 336, 142, 16 },
		{ 336, 143, 16 },
		{ 336, 144, 16 },
		{ 336, 145, 16 },
		{ 336, 146, 16 },
		{ 336, 147, 16 },
		{ 336, 148, 16 },
		{ 336, 149, 16 },
		{ 336, 150, 16 },
		{ 336, 151, 16 },
		{ 337, 131, 16 },
		{ 337, 132, 16 },
		{ 337, 133, 16 },
		{ 337, 134, 16 },
		{ 337, 135, 16 },
		{ 337, 136, 16 },
		{ 337, 137, 16 },
		{ 337, 138, 16 },
		{ 337, 139, 16 },
		{ 337, 140, 16 },
		{ 337, 141, 16 },
		{ 337, 142, 16 },
		{ 337, 143, 16 },
		{ 337, 144, 16 },
		{ 337, 145, 16 },
		{ 337, 146, 16 },
		{ 337, 147, 16 },
		{ 337, 148, 16 },
		{ 337, 149, 16 },
		{ 337, 150, 16 },
		{ 338, 131, 16 },
		{ 338, 132, 16 },
		{ 338, 133, 16 },
		{ 338, 134, 16 },
		{ 338, 135, 16 },
		{ 338, 136, 16 },
		{ 338, 137, 16 },
		{ 338, 138, 16 },
		{ 338, 139, 16 },
		{ 338, 140, 16 },
		{ 338, 141, 16 },
		{ 338, 142, 16 },
		{ 338, 143, 16 },
		{ 338, 144, 16 },
		{ 338, 145, 16 },
		{ 338, 146, 16 },
		{ 338, 147, 16 },
		{ 338, 148, 16 },
		{ 339, 131, 16 },
		{ 339, 132, 16 },
		{ 339, 133, 16 },
		{ 339, 134, 16 },
		{ 339, 135, 16 },
		{ 339, 136, 16 },
		{ 339, 137, 16 },
		{ 339, 138, 16 },
		{ 339, 139, 16 },
		{ 339, 140, 16 },
		{ 339, 141, 16 },
		{ 339, 142, 16 },
		{ 339, 143, 16 },
		{ 339, 144, 16 },
		{ 339, 145, 16 },
		{ 339, 146, 16 },
		{ 339, 147, 16 },
		{ 340, 131, 16 },
		{ 340, 132, 16 },
		{ 340, 133, 16 },
		{ 340, 134, 16 },
		{ 340, 135, 16 },
		{ 340, 136, 16 },
		{ 340, 137, 16 },
		{ 340, 138, 16 },
		{ 340, 139, 16 },
		{ 340, 140, 16 },
		{ 340, 141, 16 },
		{ 340, 142, 16 },
		{ 340, 143, 16 },
		{ 340, 144, 16 },
		{ 340, 145, 16 },
		{ 340, 146, 16 },
		{ 341, 131, 16 },
		{ 341, 132, 16 },
		{ 341, 133, 16 },
		{ 341, 134, 16 },
		{ 341, 135, 16 },
		{ 341, 136, 16 },
		{ 341, 137, 16 },
		{ 341, 138, 16 },
		{ 341, 139, 16 },
		{ 341, 140, 16 },
		{ 341, 141, 16 },
		{ 341, 142, 16 },
		{ 341, 143, 16 },
		{ 341, 144, 16 },
		{ 342, 131, 16 },
		{ 342, 132, 16 },
		{ 342, 133, 16 },
		{ 342, 134, 16 },
		{ 342, 135, 16 },
		{ 342, 136, 16 },
		{ 342, 137, 16 },
		{ 342, 138, 16 },
		{ 342, 139, 16 },
		{ 342, 140, 16 },
		{ 342, 141, 16 },
		{ 342, 142, 16 },
		{ 342, 143, 16 },
		{ 343, 131, 16 },
		{ 343, 132, 16 },
		{ 343, 133, 16 },
		{ 343, 134, 16 },
		{ 343, 135, 16 },
		{ 343, 136, 16 },
		{ 343, 137, 16 },
		{ 343, 138, 16 },
		{ 343, 139, 16 },
		{ 343, 140, 16 },
		{ 343, 141, 16 },
		{ 343, 142, 16 },
		{ 344, 131, 16 },
		{ 344, 132, 16 },
		{ 344, 133, 16 },
		{ 344, 134, 16 },
		{ 344, 135, 16 },
		{ 344, 136, 16 },
		{ 344, 137, 16 },
		{ 344, 138, 16 },
		{ 344, 139, 16 },
		{ 344, 140, 16 },
		{ 345, 132, 16 },
		{ 345, 133, 16 },
		{ 345, 134, 16 },
		{ 345, 135, 16 },
		{ 345, 136, 16 },
		{ 345, 137, 16 },
		{ 345, 138, 16 },
		{ 345, 139, 16 },
		{ 346, 132, 16 },
		{ 346, 133, 16 },
		{ 346, 134, 16 },
		{ 346, 135, 16 },
		{ 346, 136, 16 },
		{ 346, 137, 16 },
		{ 347, 132, 16 },
		{ 347, 133, 16 },
		{ 347, 134, 16 },
		{ 347, 135, 16 },
		{ 347, 136, 16 },
		{ 348, 133, 16 },
		{ 348, 134, 16 },
		{ 135, 380, 17 },
		{ 135, 381, 17 },
		{ 135, 382, 17 },
		{ 136, 379, 17 },
		{ 136, 380, 17 },
		{ 136, 381, 17 },
		{ 136, 382, 17 },
		{ 137, 378, 17 },
		{ 137, 379, 17 },
		{ 137, 380, 17 },
		{ 137, 381, 17 },
		{ 137, 382, 17 },
		{ 138, 377, 17 },
		{ 138, 378, 17 },
		{ 138, 379, 17 },
		{ 138, 380, 17 },
		{ 138, 381, 17 },
		{ 138, 382, 17 },
		{ 138, 383, 17 },
		{ 139, 376, 17 },
		{ 139, 377, 17 },
		{ 139, 378, 17 },
		{ 139, 379, 17 },
		{ 139, 380, 17 },
		{ 139, 381, 17 },
		{ 139, 382, 17 },
		{ 139, 383, 17 },
		{ 140, 375, 17 },
		{ 140, 376, 17 },
		{ 140, 377, 17 },
		{ 140, 378, 17 },
		{ 140, 379, 17 },
		{ 140, 380, 17 },
		{ 140, 381, 17 },
		{ 140, 382, 17 },
		{ 140, 383, 17 },
		{ 141, 374, 17 },
		{ 141, 375, 17 },
		{ 141, 376, 17 },
		{ 141, 377, 17 },
		{ 141, 378, 17 },
		{ 141, 379, 17 },
		{ 141, 380, 17 },
		{ 141, 381, 17 },
		{ 141, 382, 17 },
		{ 141, 383, 17 },
		{ 142, 373, 17 },
		{ 142, 374, 17 },
		{ 142, 375, 17 },
		{ 142, 376, 17 },
		{ 142, 377, 17 },
		{ 142, 378, 17 },
		{ 142, 379, 17 },
		{ 142, 380, 17 },
		{ 142, 381, 17 },
		{ 142, 382, 17 },
		{ 142, 383, 17 },
		{ 143, 372, 17 },
		{ 143, 373, 17 },
		{ 143, 374, 17 },
		{ 143, 375, 17 },
		{ 143, 376, 17 },
		{ 143, 377, 17 },
		{ 143, 378, 17 },
		{ 143, 379, 17 },
		{ 143, 380, 17 },
		{ 143, 381, 17 },
		{ 143, 382, 17 },
		{ 143, 383, 17 },
		{ 144, 371, 17 },
		{ 144, 372, 17 },
		{ 144, 373, 17 },
		{ 144, 374, 17 },
		{ 144, 375, 17 },
		{ 144, 376, 17 },
		{ 144, 377, 17 },
		{ 144, 378, 17 },
		{ 144, 379, 17 },
		{ 144, 380, 17 },
		{ 144, 381, 17 },
		{ 144, 382, 17 },
		{ 144, 383, 17 },
		{ 145, 370, 17 },
		{ 145, 371, 17 },
		{ 145, 372, 17 },
		{ 145, 373, 17 },
		{ 145, 374, 17 },
		{ 145, 375, 17 },
		{ 145, 376, 17 },
		{ 145, 377, 17 },
		{ 145, 378, 17 },
		{ 145, 379, 17 },
		{ 145, 380, 17 },
		{ 145, 381, 17 },
		{ 145, 382, 17 },
		{ 145, 383, 17 },
		{ 146, 369, 17 },
		{ 146, 370, 17 },
		{ 146, 371, 17 },
		{ 146, 372, 17 },
		{ 146, 373, 17 },
		{ 146, 374, 17 },
		{ 146, 375, 17 },
		{ 146, 376, 17 },
		{ 146, 377, 17 },
		{ 146, 378, 17 },
		{ 146, 379, 17 },
		{ 146, 380, 17 },
		{ 146, 381, 17 },
		{ 146, 382, 17 },
		{ 146, 383, 17 },
		{ 147, 368, 17 },
		{ 147, 369, 17 },
		{ 147, 370, 17 },
		{ 147, 371, 17 },
		{ 147, 372, 17 },
		{ 147, 373, 17 },
		{ 147, 374, 17 },
		{ 147, 375, 17 },
		{ 147, 376, 17 },
		{ 147, 377, 17 },
		{ 147, 378, 17 },
		{ 147, 379, 17 },
		{ 147, 380, 17 },
		{ 147, 381, 17 },
		{ 147, 382, 17 },
		{ 147, 383, 17 },
		{ 148, 367, 17 },
		{ 148, 368, 17 },
		{ 148, 369, 17 },
		{ 148, 370, 17 },
		{ 148, 371, 17 },
		{ 148, 372, 17 },
		{ 148, 373, 17 },
		{ 148, 374, 17 },
		{ 148, 375, 17 },
		{ 148, 376, 17 },
		{ 148, 377, 17 },
		{ 148, 378, 17 },
		{ 148, 379, 17 },
		{ 148, 380, 17 },
		{ 148, 381, 17 },
		{ 148, 382, 17 },
		{ 148, 383, 17 },
		{ 149, 366, 17 },
		{ 149, 367, 17 },
		{ 149, 368, 17 },
		{ 149, 369, 17 },
		{ 149, 370, 17 },
		{ 149, 371, 17 },
		{ 149, 372, 17 },
		{ 149, 373, 17 },
		{ 149, 374, 17 },
		{ 149, 375, 17 },
		{ 149, 376, 17 },
		{ 149, 377, 17 },
		{ 149, 378, 17 },
		{ 149, 379, 17 },
		{ 149, 380, 17 },
		{ 149, 381, 17 },
		{ 149, 382, 17 },
		{ 149, 383, 17 },
		{ 150, 365, 17 },
		{ 150, 366, 17 },
		{ 150, 367, 17 },
		{ 150, 368, 17 },
		{ 150, 369, 17 },
		{ 150, 370, 17 },
		{ 150, 371, 17 },
		{ 150, 372, 17 },
		{ 150, 373, 17 },
		{ 150, 374, 17 },
		{ 150, 375, 17 },
		{ 150, 376, 17 },
		{ 150, 377, 17 },
		{ 150, 378, 17 },
		{ 150, 379, 17 },
		{ 150, 380, 17 },
		{ 150, 381, 17 },
		{ 150, 382, 17 },
		{ 150, 383, 17 },
		{ 151, 364, 17 },
		{ 151, 365, 17 },
		{ 151, 366, 17 },
		{ 151, 367, 17 },
		{ 151, 368, 17 },
		{ 151, 369, 17 },
		{ 151, 370, 17 },
		{ 151, 371, 17 },
		{ 151, 372, 17 },
		{ 151, 373, 17 },
		{ 151, 374, 17 },
		{ 151, 375, 17 },
		{ 151, 376, 17 },
		{ 151, 377, 17 },
		{ 151, 378, 17 },
		{ 151, 379, 17 },
		{ 151, 380, 17 },
		{ 151, 381, 17 },
		{ 151, 382, 17 },
		{ 151, 383, 17 },
		{ 152, 363, 17 },
		{ 152, 364, 17 },
		{ 152, 365, 17 },
		{ 152, 366, 17 },
		{ 152, 367, 17 },
		{ 152, 368, 17 },
		{ 152, 369, 17 },
		{ 152, 370, 17 },
		{ 152, 371, 17 },
		{ 152, 372, 17 },
		{ 152, 373, 17 },
		{ 152, 374, 17 },
		{ 152, 375, 17 },
		{ 152, 376, 17 },
		{ 152, 377, 17 },
		{ 152, 378, 17 },
		{ 152, 379, 17 },
		{ 152, 380, 17 },
		{ 152, 381, 17 },
		{ 152, 382, 17 },
		{ 152, 383, 17 },
		{ 153, 362, 17 },
		{ 153, 363, 17 },
		{ 153, 364, 17 },
		{ 153, 365, 17 },
		{ 153, 366, 17 },
		{ 153, 367, 17 },
		{ 153, 368, 17 },
		{ 153, 369, 17 },
		{ 153, 370, 17 },
		{ 153, 371, 17 },
		{ 153, 372, 17 },
		{ 153, 373, 17 },
		{ 153, 374, 17 },
		{ 153, 375, 17 },
		{ 153, 376, 17 },
		{ 153, 377, 17 },
		{ 153, 378, 17 },
		{ 153, 379, 17 },
		{ 153, 380, 17 },
		{ 153, 381, 17 },
		{ 153, 382, 17 },
		{ 153, 383, 17 },
		{ 154, 361, 17 },
		{ 154, 362, 17 },
		{ 154, 363, 17 },
		{ 154, 364, 17 },
		{ 154, 365, 17 },
		{ 154, 366, 17 },
		{ 154, 367, 17 },
		{ 154, 368, 17 },
		{ 154, 369, 17 },
		{ 154, 370, 17 },
		{ 154, 371, 17 },
		{ 154, 372, 17 },
		{ 154, 373, 17 },
		{ 154, 374, 17 },
		{ 154, 375, 17 },
		{ 154, 376, 17 },
		{ 154, 377, 17 },
		{ 154, 378, 17 },
		{ 154, 379, 17 },
		{ 154, 380, 17 },
		{ 154, 381, 17 },
		{ 154, 382, 17 },
		{ 154, 383, 17 },
		{ 155, 360, 17 },
		{ 155, 361, 17 },
		{ 155, 362, 17 },
		{ 155, 363, 17 },
		{ 155, 364, 17 },
		{ 155, 365, 17 },
		{ 155, 366, 17 },
		{ 155, 367, 17 },
		{ 155, 368, 17 },
		{ 155, 369, 17 },
		{ 155, 370, 17 },
		{ 155, 371, 17 },
		{ 155, 372, 17 },
		{ 155, 373, 17 },
		{ 155, 374, 17 },
		{ 155, 375, 17 },
		{ 155, 376, 17 },
		{ 155, 377, 17 },
		{ 155, 378, 17 },
		{ 155, 379, 17 },
		{ 155, 380, 17 },
		{ 155, 381, 17 },
		{ 155, 382, 17 },
		{ 155, 383, 17 },
		{ 156, 359, 17 },
		{ 156, 360, 17 },
		{ 156, 361, 17 },
		{ 156, 362, 17 },
		{ 156, 363, 17 },
		{ 156, 364, 17 },
		{ 156, 365, 17 },
		{ 156, 366, 17 },
		{ 156, 367, 17 },
		{ 156, 368, 17 },
		{ 156, 369, 17 },
		{ 156, 370, 17 },
		{ 156, 371, 17 },
		{ 156, 372, 17 },
		{ 156, 373, 17 },
		{ 156, 374, 17 },
		{ 156, 375, 17 },
		{ 156, 376, 17 },
		{ 156, 377, 17 },
		{ 156, 378, 17 },
		{ 156, 379, 17 },
		{ 156, 380, 17 },
		{ 156, 381, 17 },
		{ 156, 382, 17 },
		{ 156, 383, 17 },
		{ 157, 358, 17 },
		{ 157, 359, 17 },
		{ 157, 360, 17 },
		{ 157, 361, 17 },
		{ 157, 362, 17 },
		{ 157, 363, 17 },
		{ 157, 364, 17 },
		{ 157, 365, 17 },
		{ 157, 366, 17 },
		{ 157, 367, 17 },
		{ 157, 368, 17 },
		{ 157, 369, 17 },
		{ 157, 370, 17 },
		{ 157, 371, 17 },
		{ 157, 372, 17 },
		{ 157, 373, 17 },
		{ 157, 374, 17 },
		{ 157, 375, 17 },
		{ 157, 376, 17 },
		{ 157, 377, 17 },
		{ 157, 378, 17 },
		{ 157, 379, 17 },
		{ 157, 380, 17 },
		{ 157, 381, 17 },
		{ 157, 382, 17 },
		{ 157, 383, 17 },
		{ 158, 357, 17 },
		{ 158, 358, 17 },
		{ 158, 359, 17 },
		{ 158, 360, 17 },
		{ 158, 361, 17 },
		{ 158, 362, 17 },
		{ 158, 363, 17 },
		{ 158, 364, 17 },
		{ 158, 365, 17 },
		{ 158, 366, 17 },
		{ 158, 367, 17 },
		{ 158, 368, 17 },
		{ 158, 369, 17 },
		{ 158, 370, 17 },
		{ 158, 371, 17 },
		{ 158, 372, 17 },
		{ 158, 373, 17 },
		{ 158, 374, 17 },
		{ 158, 375, 17 },
		{ 158, 376, 17 },
		{ 158, 377, 17 },
		{ 158, 378, 17 },
		{ 158, 379, 17 },
		{ 158, 380, 17 },
		{ 158, 381, 17 },
		{ 158, 382, 17 },
		{ 158, 383, 17 },
		{ 159, 356, 17 },
		{ 159, 357, 17 },
		{ 159, 358, 17 },
		{ 159, 359, 17 },
		{ 159, 360, 17 },
		{ 159, 361, 17 },
		{ 159, 362, 17 },
		{ 159, 363, 17 },
		{ 159, 364, 17 },
		{ 159, 365, 17 },
		{ 159, 366, 17 },
		{ 159, 367, 17 },
		{ 159, 368, 17 },
		{ 159, 369, 17 },
		{ 159, 370, 17 },
		{ 159, 371, 17 },
		{ 159, 372, 17 },
		{ 159, 373, 17 },
		{ 159, 374, 17 },
		{ 159, 375, 17 },
		{ 159, 376, 17 },
		{ 159, 377, 17 },
		{ 159, 378, 17 },
		{ 159, 379, 17 },
		{ 159, 380, 17 },
		{ 159, 381, 17 },
		{ 159, 382, 17 },
		{ 159, 383, 17 },
		{ 160, 355, 17 },
		{ 160, 356, 17 },
		{ 160, 357, 17 },
		{ 160, 358, 17 },
		{ 160, 359, 17 },
		{ 160, 360, 17 },
		{ 160, 361, 17 },
		{ 160, 362, 17 },
		{ 160, 363, 17 },
		{ 160, 364, 17 },
		{ 160, 365, 17 },
		{ 160, 366, 17 },
		{ 160, 367, 17 },
		{ 160, 368, 17 },
		{ 160, 369, 17 },
		{ 160, 370, 17 },
		{ 160, 371, 17 },
		{ 160, 372, 17 },
		{ 160, 373, 17 },
		{ 160, 374, 17 },
		{ 160, 375, 17 },
		{ 160, 376, 17 },
		{ 160, 377, 17 },
		{ 160, 378, 17 },
		{ 160, 379, 17 },
		{ 160, 380, 17 },
		{ 160, 381, 17 },
		{ 160, 382, 17 },
		{ 160, 383, 17 },
		{ 161, 354, 17 },
		{ 161, 355, 17 },
		{ 161, 356, 17 },
		{ 161, 357, 17 },
		{ 161, 358, 17 },
		{ 161, 359, 17 },
		{ 161, 360, 17 },
		{ 161, 361, 17 },
		{ 161, 362, 17 },
		{ 161, 363, 17 },
		{ 161, 364, 17 },
		{ 161, 365, 17 },
		{ 161, 366, 17 },
		{ 161, 367, 17 },
		{ 161, 368, 17 },
		{ 161, 369, 17 },
		{ 161, 370, 17 },
		{ 161, 371, 17 },
		{ 161, 372, 17 },
		{ 161, 373, 17 },
		{ 161, 374, 17 },
		{ 161, 375, 17 },
		{ 161, 376, 17 },
		{ 161, 377, 17 },
		{ 161, 378, 17 },
		{ 161, 379, 17 },
		{ 161, 380, 17 },
		{ 161, 381, 17 },
		{ 161, 382, 17 },
		{ 161, 383, 17 },
		{ 162, 353, 17 },
		{ 162, 354, 17 },
		{ 162, 355, 17 },
		{ 162, 356, 17 },
		{ 162, 357, 17 },
		{ 162, 358, 17 },
		{ 162, 359, 17 },
		{ 162, 360, 17 },
		{ 162, 361, 17 },
		{ 162, 362, 17 },
		{ 162, 363, 17 },
		{ 162, 364, 17 },
		{ 162, 365, 17 },
		{ 162, 366, 17 },
		{ 162, 367, 17 },
		{ 162, 368, 17 },
		{ 162, 369, 17 },
		{ 162, 370, 17 },
		{ 162, 371, 17 },
		{ 162, 372, 17 },
		{ 162, 373, 17 },
		{ 162, 374, 17 },
		{ 162, 375, 17 },
		{ 162, 376, 17 },
		{ 162, 377, 17 },
		{ 162, 378, 17 },
		{ 162, 379, 17 },
		{ 162, 380, 17 },
		{ 162, 381, 17 },
		{ 162, 382, 17 },
		{ 162, 383, 17 },
		{ 163, 352, 17 },
		{ 163, 353, 17 },
		{ 163, 354, 17 },
		{ 163, 355, 17 },
		{ 163, 356, 17 },
		{ 163, 357, 17 },
		{ 163, 358, 17 },
		{ 163, 359, 17 },
		{ 163, 360, 17 },
		{ 163, 361, 17 },
		{ 163, 362, 17 },
		{ 163, 363, 17 },
		{ 163, 364, 17 },
		{ 163, 365, 17 },
		{ 163, 366, 17 },
		{ 163, 367, 17 },
		{ 163, 368, 17 },
		{ 163, 369, 17 },
		{ 163, 370, 17 },
		{ 163, 371, 17 },
		{ 163, 372, 17 },
		{ 163, 373, 17 },
		{ 163, 374, 17 },
		{ 163, 375, 17 },
		{ 163, 376, 17 },
		{ 163, 377, 17 },
		{ 163, 378, 17 },
		{ 163, 379, 17 },
		{ 163, 380, 17 },
		{ 163, 381, 17 },
		{ 163, 382, 17 },
		{ 163, 383, 17 },
		{ 164, 351, 17 },
		{ 164, 352, 17 },
		{ 164, 353, 17 },
		{ 164, 354, 17 },
		{ 164, 355, 17 },
		{ 164, 356, 17 },
		{ 164, 357, 17 },
		{ 164, 358, 17 },
		{ 164, 359, 17 },
		{ 164, 360, 17 },
		{ 164, 361, 17 },
		{ 164, 362, 17 },
		{ 164, 363, 17 },
		{ 164, 364, 17 },
		{ 164, 365, 17 },
		{ 164, 366, 17 },
		{ 164, 367, 17 },
		{ 164, 368, 17 },
		{ 164, 369, 17 },
		{ 164, 370, 17 },
		{ 164, 371, 17 },
		{ 164, 372, 17 },
		{ 164, 373, 17 },
		{ 164, 374, 17 },
		{ 164, 375, 17 },
		{ 164, 376, 17 },
		{ 164, 377, 17 },
		{ 164, 378, 17 },
		{ 164, 379, 17 },
		{ 164, 380, 17 },
		{ 164, 381, 17 },
		{ 164, 382, 17 },
		{ 164, 383, 17 },
		{ 165, 350, 17 },
		{ 165, 351, 17 },
		{ 165, 352, 17 },
		{ 165, 353, 17 },
		{ 165, 354, 17 },
		{ 165, 355, 17 },
		{ 165, 356, 17 },
		{ 165, 357, 17 },
		{ 165, 358, 17 },
		{ 165, 359, 17 },
		{ 165, 360, 17 },
		{ 165, 361, 17 },
		{ 165, 362, 17 },
		{ 165, 363, 17 },
		{ 165, 364, 17 },
		{ 165, 365, 17 },
		{ 165, 366, 17 },
		{ 165, 367, 17 },
		{ 165, 368, 17 },
		{ 165, 369, 17 },
		{ 165, 370, 17 },
		{ 165, 371, 17 },
		{ 165, 372, 17 },
		{ 165, 373, 17 },
		{ 165, 374, 17 },
		{ 165, 375, 17 },
		{ 165, 376, 17 },
		{ 165, 377, 17 },
		{ 165, 378, 17 },
		{ 165, 379, 17 },
		{ 165, 380, 17 },
		{ 165, 381, 17 },
		{ 165, 382, 17 },
		{ 165, 383, 17 },
		{ 166, 349, 17 },
		{ 166, 350, 17 },
		{ 166, 351, 17 },
		{ 166, 352, 17 },
		{ 166, 353, 17 },
		{ 166, 354, 17 },
		{ 166, 355, 17 },
		{ 166, 356, 17 },
		{ 166, 357, 17 },
		{ 166, 358, 17 },
		{ 166, 359, 17 },
		{ 166, 360, 17 },
		{ 166, 361, 17 },
		{ 166, 362, 17 },
		{ 166, 363, 17 },
		{ 166, 364, 17 },
		{ 166, 365, 17 },
		{ 166, 366, 17 },
		{ 166, 367, 17 },
		{ 166, 368, 17 },
		{ 166, 369, 17 },
		{ 166, 370, 17 },
		{ 166, 371, 17 },
		{ 166, 372, 17 },
		{ 166, 373, 17 },
		{ 166, 374, 17 },
		{ 166, 375, 17 },
		{ 166, 376, 17 },
		{ 166, 377, 17 },
		{ 166, 378, 17 },
		{ 166, 379, 17 },
		{ 166, 380, 17 },
		{ 166, 381, 17 },
		{ 166, 382, 17 },
		{ 166, 383, 17 },
		{ 167, 348, 17 },
		{ 167, 349, 17 },
		{ 167, 350, 17 },
		{ 167, 351, 17 },
		{ 167, 352, 17 },
		{ 167, 353, 17 },
		{ 167, 354, 17 },
		{ 167, 355, 17 },
		{ 167, 356, 17 },
		{ 167, 357, 17 },
		{ 167, 358, 17 },
		{ 167, 359, 17 },
		{ 167, 360, 17 },
		{ 167, 361, 17 },
		{ 167, 362, 17 },
		{ 167, 363, 17 },
		{ 167, 364, 17 },
		{ 167, 365, 17 },
		{ 167, 366, 17 },
		{ 167, 367, 17 },
		{ 167, 368, 17 },
		{ 167, 369, 17 },
		{ 167, 370, 17 },
		{ 167, 371, 17 },
		{ 167, 372, 17 },
		{ 167, 373, 17 },
		{ 167, 374, 17 },
		{ 167, 375, 17 },
		{ 167, 376, 17 },
		{ 167, 377, 17 },
		{ 167, 378, 17 },
		{ 167, 379, 17 },
		{ 167, 380, 17 },
		{ 167, 381, 17 },
		{ 167, 382, 17 },
		{ 167, 383, 17 },
		{ 168, 347, 17 },
		{ 168, 348, 17 },
		{ 168, 349, 17 },
		{ 168, 350, 17 },
		{ 168, 351, 17 },
		{ 168, 352, 17 },
		{ 168, 353, 17 },
		{ 168, 354, 17 },
		{ 168, 355, 17 },
		{ 168, 356, 17 },
		{ 168, 357, 17 },
		{ 168, 358, 17 },
		{ 168, 359, 17 },
		{ 168, 360, 17 },
		{ 168, 361, 17 },
		{ 168, 362, 17 },
		{ 168, 363, 17 },
		{ 168, 364, 17 },
		{ 168, 365, 17 },
		{ 168, 366, 17 },
		{ 168, 367, 17 },
		{ 168, 368, 17 },
		{ 168, 369, 17 },
		{ 168, 370, 17 },
		{ 168, 371, 17 },
		{ 168, 372, 17 },
		{ 168, 373, 17 },
		{ 168, 374, 17 },
		{ 168, 375, 17 },
		{ 168, 376, 17 },
		{ 168, 377, 17 },
		{ 168, 378, 17 },
		{ 168, 379, 17 },
		{ 168, 380, 17 },
		{ 168, 381, 17 },
		{ 168, 382, 17 },
		{ 168, 383, 17 },
		{ 169, 346, 17 },
		{ 169, 347, 17 },
		{ 169, 348, 17 },
		{ 169, 349, 17 },
		{ 169, 350, 17 },
		{ 169, 351, 17 },
		{ 169, 352, 17 },
		{ 169, 353, 17 },
		{ 169, 354, 17 },
		{ 169, 355, 17 },
		{ 169, 356, 17 },
		{ 169, 357, 17 },
		{ 169, 358, 17 },
		{ 169, 359, 17 },
		{ 169, 360, 17 },
		{ 169, 361, 17 },
		{ 169, 362, 17 },
		{ 169, 363, 17 },
		{ 169, 364, 17 },
		{ 169, 365, 17 },
		{ 169, 366, 17 },
		{ 169, 367, 17 },
		{ 169, 368, 17 },
		{ 169, 369, 17 },
		{ 169, 370, 17 },
		{ 169, 371, 17 },
		{ 169, 372, 17 },
		{ 169, 373, 17 },
		{ 169, 374, 17 },
		{ 169, 375, 17 },
		{ 169, 376, 17 },
		{ 169, 377, 17 },
		{ 169, 378, 17 },
		{ 169, 379, 17 },
		{ 169, 380, 17 },
		{ 169, 381, 17 },
		{ 169, 382, 17 },
		{ 169, 383, 17 },
		{ 170, 345, 17 },
		{ 170, 346, 17 },
		{ 170, 347, 17 },
		{ 170, 348, 17 },
		{ 170, 349, 17 },
		{ 170, 350, 17 },
		{ 170, 351, 17 },
		{ 170, 352, 17 },
		{ 170, 353, 17 },
		{ 170, 354, 17 },
		{ 170, 355, 17 },
		{ 170, 356, 17 },
		{ 170, 357, 17 },
		{ 170, 358, 17 },
		{ 170, 359, 17 },
		{ 170, 360, 17 },
		{ 170, 361, 17 },
		{ 170, 362, 17 },
		{ 170, 363, 17 },
		{ 170, 364, 17 },
		{ 170, 365, 17 },
		{ 170, 366, 17 },
		{ 170, 367, 17 },
		{ 170, 368, 17 },
		{ 170, 369, 17 },
		{ 170, 370, 17 },
		{ 170, 371, 17 },
		{ 170, 372, 17 },
		{ 170, 373, 17 },
		{ 170, 374, 17 },
		{ 170, 375, 17 },
		{ 170, 376, 17 },
		{ 170, 377, 17 },
		{ 170, 378, 17 },
		{ 170, 379, 17 },
		{ 170, 380, 17 },
		{ 170, 381, 17 },
		{ 170, 382, 17 },
		{ 170, 383, 17 },
		{ 171, 344, 17 },
		{ 171, 345, 17 },
		{ 171, 346, 17 },
		{ 171, 347, 17 },
		{ 171, 348, 17 },
		{ 171, 349, 17 },
		{ 171, 350, 17 },
		{ 171, 351, 17 },
		{ 171, 352, 17 },
		{ 171, 353, 17 },
		{ 171, 354, 17 },
		{ 171, 355, 17 },
		{ 171, 356, 17 },
		{ 171, 357, 17 },
		{ 171, 358, 17 },
		{ 171, 359, 17 },
		{ 171, 360, 17 },
		{ 171, 361, 17 },
		{ 171, 362, 17 },
		{ 171, 363, 17 },
		{ 171, 364, 17 },
		{ 171, 365, 17 },
		{ 171, 366, 17 },
		{ 171, 367, 17 },
		{ 171, 368, 17 },
		{ 171, 369, 17 },
		{ 171, 370, 17 },
		{ 171, 371, 17 },
		{ 171, 372, 17 },
		{ 171, 373, 17 },
		{ 171, 374, 17 },
		{ 171, 375, 17 },
		{ 171, 376, 17 },
		{ 171, 377, 17 },
		{ 171, 378, 17 },
		{ 171, 379, 17 },
		{ 171, 380, 17 },
		{ 171, 381, 17 },
		{ 171, 382, 17 },
		{ 171, 383, 17 },
		{ 172, 343, 17 },
		{ 172, 344, 17 },
		{ 172, 345, 17 },
		{ 172, 346, 17 },
		{ 172, 347, 17 },
		{ 172, 348, 17 },
		{ 172, 349, 17 },
		{ 172, 350, 17 },
		{ 172, 351, 17 },
		{ 172, 352, 17 },
		{ 172, 353, 17 },
		{ 172, 354, 17 },
		{ 172, 355, 17 },
		{ 172, 356, 17 },
		{ 172, 357, 17 },
		{ 172, 358, 17 },
		{ 172, 359, 17 },
		{ 172, 360, 17 },
		{ 172, 361, 17 },
		{ 172, 362, 17 },
		{ 172, 363, 17 },
		{ 172, 364, 17 },
		{ 172, 365, 17 },
		{ 172, 366, 17 },
		{ 172, 367, 17 },
		{ 172, 368, 17 },
		{ 172, 369, 17 },
		{ 172, 370, 17 },
		{ 172, 371, 17 },
		{ 172, 372, 17 },
		{ 172, 373, 17 },
		{ 172, 374, 17 },
		{ 172, 375, 17 },
		{ 172, 376, 17 },
		{ 172, 377, 17 },
		{ 172, 378, 17 },
		{ 172, 379, 17 },
		{ 172, 380, 17 },
		{ 172, 381, 17 },
		{ 172, 382, 17 },
		{ 172, 383, 17 },
		{ 173, 342, 17 },
		{ 173, 343, 17 },
		{ 173, 344, 17 },
		{ 173, 345, 17 },
		{ 173, 346, 17 },
		{ 173, 347, 17 },
		{ 173, 348, 17 },
		{ 173, 349, 17 },
		{ 173, 350, 17 },
		{ 173, 351, 17 },
		{ 173, 352, 17 },
		{ 173, 353, 17 },
		{ 173, 354, 17 },
		{ 173, 355, 17 },
		{ 173, 356, 17 },
		{ 173, 357, 17 },
		{ 173, 358, 17 },
		{ 173, 359, 17 },
		{ 173, 360, 17 },
		{ 173, 361, 17 },
		{ 173, 362, 17 },
		{ 173, 363, 17 },
		{ 173, 364, 17 },
		{ 173, 365, 17 },
		{ 173, 366, 17 },
		{ 173, 367, 17 },
		{ 173, 368, 17 },
		{ 173, 369, 17 },
		{ 173, 370, 17 },
		{ 173, 371, 17 },
		{ 173, 372, 17 },
		{ 173, 373, 17 },
		{ 173, 374, 17 },
		{ 173, 375, 17 },
		{ 173, 376, 17 },
		{ 173, 377, 17 },
		{ 173, 378, 17 },
		{ 173, 379, 17 },
		{ 173, 380, 17 },
		{ 173, 381, 17 },
		{ 173, 382, 17 },
		{ 173, 383, 17 },
		{ 174, 341, 17 },
		{ 174, 342, 17 },
		{ 174, 343, 17 },
		{ 174, 344, 17 },
		{ 174, 345, 17 },
		{ 174, 346, 17 },
		{ 174, 347, 17 },
		{ 174, 348, 17 },
		{ 174, 349, 17 },
		{ 174, 350, 17 },
		{ 174, 351, 17 },
		{ 174, 352, 17 },
		{ 174, 353, 17 },
		{ 174, 354, 17 },
		{ 174, 355, 17 },
		{ 174, 356, 17 },
		{ 174, 357, 17 },
		{ 174, 358, 17 },
		{ 174, 359, 17 },
		{ 174, 360, 17 },
		{ 174, 361, 17 },
		{ 174, 362, 17 },
		{ 174, 363, 17 },
		{ 174, 364, 17 },
		{ 174, 365, 17 },
		{ 174, 366, 17 },
		{ 174, 367, 17 },
		{ 174, 368, 17 },
		{ 174, 369, 17 },
		{ 174, 370, 17 },
		{ 174, 371, 17 },
		{ 174, 372, 17 },
		{ 174, 373, 17 },
		{ 174, 374, 17 },
		{ 174, 375, 17 },
		{ 174, 376, 17 },
		{ 174, 377, 17 },
		{ 174, 378, 17 },
		{ 174, 379, 17 },
		{ 174, 380, 17 },
		{ 174, 381, 17 },
		{ 174, 382, 17 },
		{ 174, 383, 17 },
		{ 175, 340, 17 },
		{ 175, 341, 17 },
		{ 175, 342, 17 },
		{ 175, 343, 17 },
		{ 175, 344, 17 },
		{ 175, 345, 17 },
		{ 175, 346, 17 },
		{ 175, 347, 17 },
		{ 175, 348, 17 },
		{ 175, 349, 17 },
		{ 175, 350, 17 },
		{ 175, 351, 17 },
		{ 175, 352, 17 },
		{ 175, 353, 17 },
		{ 175, 354, 17 },
		{ 175, 355, 17 },
		{ 175, 356, 17 },
		{ 175, 357, 17 },
		{ 175, 358, 17 },
		{ 175, 359, 17 },
		{ 175, 360, 17 },
		{ 175, 361, 17 },
		{ 175, 362, 17 },
		{ 175, 363, 17 },
		{ 175, 364, 17 },
		{ 175, 365, 17 },
		{ 175, 366, 17 },
		{ 175, 367, 17 },
		{ 175, 368, 17 },
		{ 175, 369, 17 },
		{ 175, 370, 17 },
		{ 175, 371, 17 },
		{ 175, 372, 17 },
		{ 175, 373, 17 },
		{ 175, 374, 17 },
		{ 175, 375, 17 },
		{ 175, 376, 17 },
		{ 175, 377, 17 },
		{ 175, 378, 17 },
		{ 175, 379, 17 },
		{ 175, 380, 17 },
		{ 175, 381, 17 },
		{ 175, 382, 17 },
		{ 175, 383, 17 },
		{ 176, 339, 17 },
		{ 176, 340, 17 },
		{ 176, 341, 17 },
		{ 176, 342, 17 },
		{ 176, 343, 17 },
		{ 176, 344, 17 },
		{ 176, 345, 17 },
		{ 176, 346, 17 },
		{ 176, 347, 17 },
		{ 176, 348, 17 },
		{ 176, 349, 17 },
		{ 176, 350, 17 },
		{ 176, 351, 17 },
		{ 176, 352, 17 },
		{ 176, 353, 17 },
		{ 176, 354, 17 },
		{ 176, 355, 17 },
		{ 176, 356, 17 },
		{ 176, 357, 17 },
		{ 176, 358, 17 },
		{ 176, 359, 17 },
		{ 176, 360, 17 },
		{ 176, 361, 17 },
		{ 176, 362, 17 },
		{ 176, 363, 17 },
		{ 176, 364, 17 },
		{ 176, 365, 17 },
		{ 176, 366, 17 },
		{ 176, 367, 17 },
		{ 176, 368, 17 },
		{ 176, 369, 17 },
		{ 176, 370, 17 },
		{ 176, 371, 17 },
		{ 176, 372, 17 },
		{ 176, 373, 17 },
		{ 176, 374, 17 },
		{ 176, 375, 17 },
		{ 176, 376, 17 },
		{ 176, 377, 17 },
		{ 176, 378, 17 },
		{ 176, 379, 17 },
		{ 176, 380, 17 },
		{ 176, 381, 17 },
		{ 176, 382, 17 },
		{ 176, 383, 17 },
		{ 177, 338, 17 },
		{ 177, 339, 17 },
		{ 177, 340, 17 },
		{ 177, 341, 17 },
		{ 177, 342, 17 },
		{ 177, 343, 17 },
		{ 177, 344, 17 },
		{ 177, 345, 17 },
		{ 177, 346, 17 },
		{ 177, 347, 17 },
		{ 177, 348, 17 },
		{ 177, 349, 17 },
		{ 177, 350, 17 },
		{ 177, 351, 17 },
		{ 177, 352, 17 },
		{ 177, 353, 17 },
		{ 177, 354, 17 },
		{ 177, 355, 17 },
		{ 177, 356, 17 },
		{ 177, 357, 17 },
		{ 177, 358, 17 },
		{ 177, 359, 17 },
		{ 177, 360, 17 },
		{ 177, 361, 17 },
		{ 177, 362, 17 },
		{ 177, 363, 17 },
		{ 177, 364, 17 },
		{ 177, 365, 17 },
		{ 177, 366, 17 },
		{ 177, 367, 17 },
		{ 177, 368, 17 },
		{ 177, 369, 17 },
		{ 177, 370, 17 },
		{ 177, 371, 17 },
		{ 177, 372, 17 },
		{ 177, 373, 17 },
		{ 177, 374, 17 },
		{ 177, 375, 17 },
		{ 177, 376, 17 },
		{ 177, 377, 17 },
		{ 177, 378, 17 },
		{ 177, 379, 17 },
		{ 177, 380, 17 },
		{ 177, 381, 17 },
		{ 177, 382, 17 },
		{ 177, 383, 17 },
		{ 178, 337, 17 },
		{ 178, 338, 17 },
		{ 178, 339, 17 },
		{ 178, 340, 17 },
		{ 178, 341, 17 },
		{ 178, 342, 17 },
		{ 178, 343, 17 },
		{ 178, 344, 17 },
		{ 178, 345, 17 },
		{ 178, 346, 17 },
		{ 178, 347, 17 },
		{ 178, 348, 17 },
		{ 178, 349, 17 },
		{ 178, 350, 17 },
		{ 178, 351, 17 },
		{ 178, 352, 17 },
		{ 178, 353, 17 },
		{ 178, 354, 17 },
		{ 178, 355, 17 },
		{ 178, 356, 17 },
		{ 178, 357, 17 },
		{ 178, 358, 17 },
		{ 178, 359, 17 },
		{ 178, 360, 17 },
		{ 178, 361, 17 },
		{ 178, 362, 17 },
		{ 178, 363, 17 },
		{ 178, 364, 17 },
		{ 178, 365, 17 },
		{ 178, 366, 17 },
		{ 178, 367, 17 },
		{ 178, 368, 17 },
		{ 178, 369, 17 },
		{ 178, 370, 17 },
		{ 178, 371, 17 },
		{ 178, 372, 17 },
		{ 178, 373, 17 },
		{ 178, 374, 17 },
		{ 178, 375, 17 },
		{ 178, 376, 17 },
		{ 178, 377, 17 },
		{ 178, 378, 17 },
		{ 178, 379, 17 },
		{ 178, 380, 17 },
		{ 178, 381, 17 },
		{ 178, 382, 17 },
		{ 178, 383, 17 },
		{ 179, 336, 17 },
		{ 179, 337, 17 },
		{ 179, 338, 17 },
		{ 179, 339, 17 },
		{ 179, 340, 17 },
		{ 179, 341, 17 },
		{ 179, 342, 17 },
		{ 179, 343, 17 },
		{ 179, 344, 17 },
		{ 179, 345, 17 },
		{ 179, 346, 17 },
		{ 179, 347, 17 },
		{ 179, 348, 17 },
		{ 179, 349, 17 },
		{ 179, 350, 17 },
		{ 179, 351, 17 },
		{ 179, 352, 17 },
		{ 179, 353, 17 },
		{ 179, 354, 17 },
		{ 179, 355, 17 },
		{ 179, 356, 17 },
		{ 179, 357, 17 },
		{ 179, 358, 17 },
		{ 179, 359, 17 },
		{ 179, 360, 17 },
		{ 179, 361, 17 },
		{ 179, 362, 17 },
		{ 179, 363, 17 },
		{ 179, 364, 17 },
		{ 179, 365, 17 },
		{ 179, 366, 17 },
		{ 179, 367, 17 },
		{ 179, 368, 17 },
		{ 179, 369, 17 },
		{ 179, 370, 17 },
		{ 179, 371, 17 },
		{ 179, 372, 17 },
		{ 179, 373, 17 },
		{ 179, 374, 17 },
		{ 179, 375, 17 },
		{ 179, 376, 17 },
		{ 179, 377, 17 },
		{ 179, 378, 17 },
		{ 179, 379, 17 },
		{ 179, 380, 17 },
		{ 179, 381, 17 },
		{ 179, 382, 17 },
		{ 179, 383, 17 },
		{ 180, 335, 17 },
		{ 180, 336, 17 },
		{ 180, 337, 17 },
		{ 180, 338, 17 },
		{ 180, 339, 17 },
		{ 180, 340, 17 },
		{ 180, 341, 17 },
		{ 180, 342, 17 },
		{ 180, 343, 17 },
		{ 180, 344, 17 },
		{ 180, 345, 17 },
		{ 180, 346, 17 },
		{ 180, 347, 17 },
		{ 180, 348, 17 },
		{ 180, 349, 17 },
		{ 180, 350, 17 },
		{ 180, 351, 17 },
		{ 180, 352, 17 },
		{ 180, 353, 17 },
		{ 180, 354, 17 },
		{ 180, 355, 17 },
		{ 180, 356, 17 },
		{ 180, 357, 17 },
		{ 180, 358, 17 },
		{ 180, 359, 17 },
		{ 180, 360, 17 },
		{ 180, 361, 17 },
		{ 180, 362, 17 },
		{ 180, 363, 17 },
		{ 180, 364, 17 },
		{ 180, 365, 17 },
		{ 180, 366, 17 },
		{ 180, 367, 17 },
		{ 180, 368, 17 },
		{ 180, 369, 17 },
		{ 180, 370, 17 },
		{ 180, 371, 17 },
		{ 180, 372, 17 },
		{ 180, 373, 17 },
		{ 180, 374, 17 },
		{ 180, 375, 17 },
		{ 180, 376, 17 },
		{ 180, 377, 17 },
		{ 180, 378, 17 },
		{ 180, 379, 17 },
		{ 180, 380, 17 },
		{ 180, 381, 17 },
		{ 180, 382, 17 },
		{ 180, 383, 17 },
		{ 181, 334, 17 },
		{ 181, 335, 17 },
		{ 181, 336, 17 },
		{ 181, 337, 17 },
		{ 181, 338, 17 },
		{ 181, 339, 17 },
		{ 181, 340, 17 },
		{ 181, 341, 17 },
		{ 181, 342, 17 },
		{ 181, 343, 17 },
		{ 181, 344, 17 },
		{ 181, 345, 17 },
		{ 181, 346, 17 },
		{ 181, 347, 17 },
		{ 181, 348, 17 },
		{ 181, 349, 17 },
		{ 181, 350, 17 },
		{ 181, 351, 17 },
		{ 181, 352, 17 },
		{ 181, 353, 17 },
		{ 181, 354, 17 },
		{ 181, 355, 17 },
		{ 181, 356, 17 },
		{ 181, 357, 17 },
		{ 181, 358, 17 },
		{ 181, 359, 17 },
		{ 181, 360, 17 },
		{ 181, 361, 17 },
		{ 181, 362, 17 },
		{ 181, 363, 17 },
		{ 181, 364, 17 },
		{ 181, 365, 17 },
		{ 181, 366, 17 },
		{ 181, 367, 17 },
		{ 181, 368, 17 },
		{ 181, 369, 17 },
		{ 181, 370, 17 },
		{ 181, 371, 17 },
		{ 181, 372, 17 },
		{ 181, 373, 17 },
		{ 181, 374, 17 },
		{ 181, 375, 17 },
		{ 181, 376, 17 },
		{ 181, 377, 17 },
		{ 181, 378, 17 },
		{ 181, 379, 17 },
		{ 181, 380, 17 },
		{ 181, 381, 17 },
		{ 181, 382, 17 },
		{ 181, 383, 17 },
		{ 182, 333, 17 },
		{ 182, 334, 17 },
		{ 182, 335, 17 },
		{ 182, 336, 17 },
		{ 182, 337, 17 },
		{ 182, 338, 17 },
		{ 182, 339, 17 },
		{ 182, 340, 17 },
		{ 182, 341, 17 },
		{ 182, 342, 17 },
		{ 182, 343, 17 },
		{ 182, 344, 17 },
		{ 182, 345, 17 },
		{ 182, 346, 17 },
		{ 182, 347, 17 },
		{ 182, 348, 17 },
		{ 182, 349, 17 },
		{ 182, 350, 17 },
		{ 182, 351, 17 },
		{ 182, 352, 17 },
		{ 182, 353, 17 },
		{ 182, 354, 17 },
		{ 182, 355, 17 },
		{ 182, 356, 17 },
		{ 182, 357, 17 },
		{ 182, 358, 17 },
		{ 182, 359, 17 },
		{ 182, 360, 17 },
		{ 182, 361, 17 },
		{ 182, 362, 17 },
		{ 182, 363, 17 },
		{ 182, 364, 17 },
		{ 182, 365, 17 },
		{ 182, 366, 17 },
		{ 182, 367, 17 },
		{ 182, 368, 17 },
		{ 182, 369, 17 },
		{ 182, 370, 17 },
		{ 182, 371, 17 },
		{ 182, 372, 17 },
		{ 182, 373, 17 },
		{ 182, 374, 17 },
		{ 182, 375, 17 },
		{ 182, 376, 17 },
		{ 182, 377, 17 },
		{ 182, 378, 17 },
		{ 182, 379, 17 },
		{ 182, 380, 17 },
		{ 182, 381, 17 },
		{ 182, 382, 17 },
		{ 182, 383, 17 },
		{ 183, 332, 17 },
		{ 183, 333, 17 },
		{ 183, 334, 17 },
		{ 183, 335, 17 },
		{ 183, 336, 17 },
		{ 183, 337, 17 },
		{ 183, 338, 17 },
		{ 183, 339, 17 },
		{ 183, 340, 17 },
		{ 183, 341, 17 },
		{ 183, 342, 17 },
		{ 183, 343, 17 },
		{ 183, 344, 17 },
		{ 183, 345, 17 },
		{ 183, 346, 17 },
		{ 183, 347, 17 },
		{ 183, 348, 17 },
		{ 183, 349, 17 },
		{ 183, 350, 17 },
		{ 183, 351, 17 },
		{ 183, 352, 17 },
		{ 183, 353, 17 },
		{ 183, 354, 17 },
		{ 183, 355, 17 },
		{ 183, 356, 17 },
		{ 183, 357, 17 },
		{ 183, 358, 17 },
		{ 183, 359, 17 },
		{ 183, 360, 17 },
		{ 183, 361, 17 },
		{ 183, 362, 17 },
		{ 183, 363, 17 },
		{ 183, 364, 17 },
		{ 183, 365, 17 },
		{ 183, 366, 17 },
		{ 183, 367, 17 },
		{ 183, 368, 17 },
		{ 183, 369, 17 },
		{ 183, 370, 17 },
		{ 183, 371, 17 },
		{ 183, 372, 17 },
		{ 183, 373, 17 },
		{ 183, 374, 17 },
		{ 183, 375, 17 },
		{ 183, 376, 17 },
		{ 183, 377, 17 },
		{ 183, 378, 17 },
		{ 183, 379, 17 },
		{ 183, 380, 17 },
		{ 183, 381, 17 },
		{ 183, 382, 17 },
		{ 183, 383, 17 },
		{ 184, 331, 17 },
		{ 184, 332, 17 },
		{ 184, 333, 17 },
		{ 184, 334, 17 },
		{ 184, 335, 17 },
		{ 184, 336, 17 },
		{ 184, 337, 17 },
		{ 184, 338, 17 },
		{ 184, 339, 17 },
		{ 184, 340, 17 },
		{ 184, 341, 17 },
		{ 184, 342, 17 },
		{ 184, 343, 17 },
		{ 184, 344, 17 },
		{ 184, 345, 17 },
		{ 184, 346, 17 },
		{ 184, 347, 17 },
		{ 184, 348, 17 },
		{ 184, 349, 17 },
		{ 184, 350, 17 },
		{ 184, 351, 17 },
		{ 184, 352, 17 },
		{ 184, 353, 17 },
		{ 184, 354, 17 },
		{ 184, 355, 17 },
		{ 184, 356, 17 },
		{ 184, 357, 17 },
		{ 184, 358, 17 },
		{ 184, 359, 17 },
		{ 184, 360, 17 },
		{ 184, 361, 17 },
		{ 184, 362, 17 },
		{ 184, 363, 17 },
		{ 184, 364, 17 },
		{ 184, 365, 17 },
		{ 184, 366, 17 },
		{ 184, 367, 17 },
		{ 184, 368, 17 },
		{ 184, 369, 17 },
		{ 184, 370, 17 },
		{ 184, 371, 17 },
		{ 184, 372, 17 },
		{ 184, 373, 17 },
		{ 184, 374, 17 },
		{ 184, 375, 17 },
		{ 184, 376, 17 },
		{ 184, 377, 17 },
		{ 184, 378, 17 },
		{ 184, 379, 17 },
		{ 184, 380, 17 },
		{ 184, 381, 17 },
		{ 184, 382, 17 },
		{ 184, 383, 17 },
		{ 185, 330, 17 },
		{ 185, 331, 17 },
		{ 185, 332, 17 },
		{ 185, 333, 17 },
		{ 185, 334, 17 },
		{ 185, 335, 17 },
		{ 185, 336, 17 },
		{ 185, 337, 17 },
		{ 185, 338, 17 },
		{ 185, 339, 17 },
		{ 185, 340, 17 },
		{ 185, 341, 17 },
		{ 185, 342, 17 },
		{ 185, 343, 17 },
		{ 185, 344, 17 },
		{ 185, 345, 17 },
		{ 185, 346, 17 },
		{ 185, 347, 17 },
		{ 185, 348, 17 },
		{ 185, 349, 17 },
		{ 185, 350, 17 },
		{ 185, 351, 17 },
		{ 185, 352, 17 },
		{ 185, 353, 17 },
		{ 185, 354, 17 },
		{ 185, 355, 17 },
		{ 185, 356, 17 },
		{ 185, 357, 17 },
		{ 185, 358, 17 },
		{ 185, 359, 17 },
		{ 185, 360, 17 },
		{ 185, 361, 17 },
		{ 185, 362, 17 },
		{ 185, 363, 17 },
		{ 185, 364, 17 },
		{ 185, 365, 17 },
		{ 185, 366, 17 },
		{ 185, 367, 17 },
		{ 185, 368, 17 },
		{ 185, 369, 17 },
		{ 185, 370, 17 },
		{ 185, 371, 17 },
		{ 185, 372, 17 },
		{ 185, 373, 17 },
		{ 185, 374, 17 },
		{ 185, 375, 17 },
		{ 185, 376, 17 },
		{ 185, 377, 17 },
		{ 185, 378, 17 },
		{ 185, 379, 17 },
		{ 185, 380, 17 },
		{ 185, 381, 17 },
		{ 185, 382, 17 },
		{ 185, 383, 17 },
		{ 186, 329, 17 },
		{ 186, 330, 17 },
		{ 186, 331, 17 },
		{ 186, 332, 17 },
		{ 186, 333, 17 },
		{ 186, 334, 17 },
		{ 186, 335, 17 },
		{ 186, 336, 17 },
		{ 186, 337, 17 },
		{ 186, 338, 17 },
		{ 186, 339, 17 },
		{ 186, 340, 17 },
		{ 186, 341, 17 },
		{ 186, 342, 17 },
		{ 186, 343, 17 },
		{ 186, 344, 17 },
		{ 186, 345, 17 },
		{ 186, 346, 17 },
		{ 186, 347, 17 },
		{ 186, 348, 17 },
		{ 186, 349, 17 },
		{ 186, 350, 17 },
		{ 186, 351, 17 },
		{ 186, 352, 17 },
		{ 186, 353, 17 },
		{ 186, 354, 17 },
		{ 186, 355, 17 },
		{ 186, 356, 17 },
		{ 186, 357, 17 },
		{ 186, 358, 17 },
		{ 186, 359, 17 },
		{ 186, 360, 17 },
		{ 186, 361, 17 },
		{ 186, 362, 17 },
		{ 186, 363, 17 },
		{ 186, 364, 17 },
		{ 186, 365, 17 },
		{ 186, 366, 17 },
		{ 186, 367, 17 },
		{ 186, 368, 17 },
		{ 186, 369, 17 },
		{ 186, 370, 17 },
		{ 186, 371, 17 },
		{ 186, 372, 17 },
		{ 186, 373, 17 },
		{ 186, 374, 17 },
		{ 186, 375, 17 },
		{ 186, 376, 17 },
		{ 186, 377, 17 },
		{ 186, 378, 17 },
		{ 186, 379, 17 },
		{ 186, 380, 17 },
		{ 186, 381, 17 },
		{ 186, 382, 17 },
		{ 186, 383, 17 },
		{ 187, 328, 17 },
		{ 187, 329, 17 },
		{ 187, 330, 17 },
		{ 187, 331, 17 },
		{ 187, 332, 17 },
		{ 187, 333, 17 },
		{ 187, 334, 17 },
		{ 187, 335, 17 },
		{ 187, 336, 17 },
		{ 187, 337, 17 },
		{ 187, 338, 17 },
		{ 187, 339, 17 },
		{ 187, 340, 17 },
		{ 187, 341, 17 },
		{ 187, 342, 17 },
		{ 187, 343, 17 },
		{ 187, 344, 17 },
		{ 187, 345, 17 },
		{ 187, 346, 17 },
		{ 187, 347, 17 },
		{ 187, 348, 17 },
		{ 187, 349, 17 },
		{ 187, 350, 17 },
		{ 187, 351, 17 },
		{ 187, 352, 17 },
		{ 187, 353, 17 },
		{ 187, 354, 17 },
		{ 187, 355, 17 },
		{ 187, 356, 17 },
		{ 187, 357, 17 },
		{ 187, 358, 17 },
		{ 187, 359, 17 },
		{ 187, 360, 17 },
		{ 187, 361, 17 },
		{ 187, 362, 17 },
		{ 187, 363, 17 },
		{ 187, 364, 17 },
		{ 187, 365, 17 },
		{ 187, 366, 17 },
		{ 187, 367, 17 },
		{ 187, 368, 17 },
		{ 187, 369, 17 },
		{ 187, 370, 17 },
		{ 187, 371, 17 },
		{ 187, 372, 17 },
		{ 187, 373, 17 },
		{ 187, 374, 17 },
		{ 187, 375, 17 },
		{ 187, 376, 17 },
		{ 187, 377, 17 },
		{ 187, 378, 17 },
		{ 187, 379, 17 },
		{ 187, 380, 17 },
		{ 187, 381, 17 },
		{ 187, 382, 17 },
		{ 187, 383, 17 },
		{ 188, 327, 17 },
		{ 188, 328, 17 },
		{ 188, 329, 17 },
		{ 188, 330, 17 },
		{ 188, 331, 17 },
		{ 188, 332, 17 },
		{ 188, 333, 17 },
		{ 188, 334, 17 },
		{ 188, 335, 17 },
		{ 188, 336, 17 },
		{ 188, 337, 17 },
		{ 188, 338, 17 },
		{ 188, 339, 17 },
		{ 188, 340, 17 },
		{ 188, 341, 17 },
		{ 188, 342, 17 },
		{ 188, 343, 17 },
		{ 188, 344, 17 },
		{ 188, 345, 17 },
		{ 188, 346, 17 },
		{ 188, 347, 17 },
		{ 188, 348, 17 },
		{ 188, 349, 17 },
		{ 188, 350, 17 },
		{ 188, 351, 17 },
		{ 188, 352, 17 },
		{ 188, 353, 17 },
		{ 188, 354, 17 },
		{ 188, 355, 17 },
		{ 188, 356, 17 },
		{ 188, 357, 17 },
		{ 188, 358, 17 },
		{ 188, 359, 17 },
		{ 188, 360, 17 },
		{ 188, 361, 17 },
		{ 188, 362, 17 },
		{ 188, 363, 17 },
		{ 188, 364, 17 },
		{ 188, 365, 17 },
		{ 188, 366, 17 },
		{ 188, 367, 17 },
		{ 188, 368, 17 },
		{ 188, 369, 17 },
		{ 188, 370, 17 },
		{ 188, 371, 17 },
		{ 188, 372, 17 },
		{ 188, 373, 17 },
		{ 188, 374, 17 },
		{ 188, 375, 17 },
		{ 188, 376, 17 },
		{ 188, 377, 17 },
		{ 188, 378, 17 },
		{ 188, 379, 17 },
		{ 188, 380, 17 },
		{ 188, 381, 17 },
		{ 188, 382, 17 },
		{ 189, 326, 17 },
		{ 189, 327, 17 },
		{ 189, 328, 17 },
		{ 189, 329, 17 },
		{ 189, 330, 17 },
		{ 189, 331, 17 },
		{ 189, 332, 17 },
		{ 189, 333, 17 },
		{ 189, 334, 17 },
		{ 189, 335, 17 },
		{ 189, 336, 17 },
		{ 189, 337, 17 },
		{ 189, 338, 17 },
		{ 189, 339, 17 },
		{ 189, 340, 17 },
		{ 189, 341, 17 },
		{ 189, 342, 17 },
		{ 189, 343, 17 },
		{ 189, 344, 17 },
		{ 189, 345, 17 },
		{ 189, 346, 17 },
		{ 189, 347, 17 },
		{ 189, 348, 17 },
		{ 189, 349, 17 },
		{ 189, 350, 17 },
		{ 189, 351, 17 },
		{ 189, 352, 17 },
		{ 189, 353, 17 },
		{ 189, 354, 17 },
		{ 189, 355, 17 },
		{ 189, 356, 17 },
		{ 189, 357, 17 },
		{ 189, 358, 17 },
		{ 189, 359, 17 },
		{ 189, 360, 17 },
		{ 189, 361, 17 },
		{ 189, 362, 17 },
		{ 189, 363, 17 },
		{ 189, 364, 17 },
		{ 189, 365, 17 },
		{ 189, 366, 17 },
		{ 189, 367, 17 },
		{ 189, 368, 17 },
		{ 189, 369, 17 },
		{ 189, 370, 17 },
		{ 189, 371, 17 },
		{ 189, 372, 17 },
		{ 189, 373, 17 },
		{ 189, 374, 17 },
		{ 189, 375, 17 },
		{ 189, 376, 17 },
		{ 189, 377, 17 },
		{ 189, 378, 17 },
		{ 189, 379, 17 },
		{ 189, 380, 17 },
		{ 189, 381, 17 },
		{ 189, 382, 17 },
		{ 190, 325, 17 },
		{ 190, 326, 17 },
		{ 190, 327, 17 },
		{ 190, 328, 17 },
		{ 190, 329, 17 },
		{ 190, 330, 17 },
		{ 190, 331, 17 },
		{ 190, 332, 17 },
		{ 190, 333, 17 },
		{ 190, 334, 17 },
		{ 190, 335, 17 },
		{ 190, 336, 17 },
		{ 190, 337, 17 },
		{ 190, 338, 17 },
		{ 190, 339, 17 },
		{ 190, 340, 17 },
		{ 190, 341, 17 },
		{ 190, 342, 17 },
		{ 190, 343, 17 },
		{ 190, 344, 17 },
		{ 190, 345, 17 },
		{ 190, 346, 17 },
		{ 190, 347, 17 },
		{ 190, 348, 17 },
		{ 190, 349, 17 },
		{ 190, 350, 17 },
		{ 190, 351, 17 },
		{ 190, 352, 17 },
		{ 190, 353, 17 },
		{ 190, 354, 17 },
		{ 190, 355, 17 },
		{ 190, 356, 17 },
		{ 190, 357, 17 },
		{ 190, 358, 17 },
		{ 190, 359, 17 },
		{ 190, 360, 17 },
		{ 190, 361, 17 },
		{ 190, 362, 17 },
		{ 190, 363, 17 },
		{ 190, 364, 17 },
		{ 190, 365, 17 },
		{ 190, 366, 17 },
		{ 190, 367, 17 },
		{ 190, 368, 17 },
		{ 190, 369, 17 },
		{ 190, 370, 17 },
		{ 190, 371, 17 },
		{ 190, 372, 17 },
		{ 190, 373, 17 },
		{ 190, 374, 17 },
		{ 190, 375, 17 },
		{ 190, 376, 17 },
		{ 190, 377, 17 },
		{ 190, 378, 17 },
		{ 190, 379, 17 },
		{ 190, 380, 17 },
		{ 190, 381, 17 },
		{ 190, 382, 17 },
		{ 191, 324, 17 },
		{ 191, 325, 17 },
		{ 191, 326, 17 },
		{ 191, 327, 17 },
		{ 191, 328, 17 },
		{ 191, 329, 17 },
		{ 191, 330, 17 },
		{ 191, 331, 17 },
		{ 191, 332, 17 },
		{ 191, 333, 17 },
		{ 191, 334, 17 },
		{ 191, 335, 17 },
		{ 191, 336, 17 },
		{ 191, 337, 17 },
		{ 191, 338, 17 },
		{ 191, 339, 17 },
		{ 191, 340, 17 },
		{ 191, 341, 17 },
		{ 191, 342, 17 },
		{ 191, 343, 17 },
		{ 191, 344, 17 },
		{ 191, 345, 17 },
		{ 191, 346, 17 },
		{ 191, 347, 17 },
		{ 191, 348, 17 },
		{ 191, 349, 17 },
		{ 191, 350, 17 },
		{ 191, 351, 17 },
		{ 191, 352, 17 },
		{ 191, 353, 17 },
		{ 191, 354, 17 },
		{ 191, 355, 17 },
		{ 191, 356, 17 },
		{ 191, 357, 17 },
		{ 191, 358, 17 },
		{ 191, 359, 17 },
		{ 191, 360, 17 },
		{ 191, 361, 17 },
		{ 191, 362, 17 },
		{ 191, 363, 17 },
		{ 191, 364, 17 },
		{ 191, 365, 17 },
		{ 191, 366, 17 },
		{ 191, 367, 17 },
		{ 191, 368, 17 },
		{ 191, 369, 17 },
		{ 191, 370, 17 },
		{ 191, 371, 17 },
		{ 191, 372, 17 },
		{ 191, 373, 17 },
		{ 191, 374, 17 },
		{ 191, 375, 17 },
		{ 191, 376, 17 },
		{ 191, 377, 17 },
		{ 191, 378, 17 },
		{ 191, 379, 17 },
		{ 191, 380, 17 },
		{ 191, 381, 17 },
		{ 191, 382, 17 },
		{ 192, 323, 17 },
		{ 192, 324, 17 },
		{ 192, 325, 17 },
		{ 192, 326, 17 },
		{ 192, 327, 17 },
		{ 192, 328, 17 },
		{ 192, 329, 17 },
		{ 192, 330, 17 },
		{ 192, 331, 17 },
		{ 192, 332, 17 },
		{ 192, 333, 17 },
		{ 192, 334, 17 },
		{ 192, 335, 17 },
		{ 192, 336, 17 },
		{ 192, 337, 17 },
		{ 192, 338, 17 },
		{ 192, 339, 17 },
		{ 192, 340, 17 },
		{ 192, 341, 17 },
		{ 192, 342, 17 },
		{ 192, 343, 17 },
		{ 192, 344, 17 },
		{ 192, 345, 17 },
		{ 192, 346, 17 },
		{ 192, 347, 17 },
		{ 192, 348, 17 },
		{ 192, 349, 17 },
		{ 192, 350, 17 },
		{ 192, 351, 17 },
		{ 192, 352, 17 },
		{ 192, 353, 17 },
		{ 192, 354, 17 },
		{ 192, 355, 17 },
		{ 192, 356, 17 },
		{ 192, 357, 17 },
		{ 192, 358, 17 },
		{ 192, 359, 17 },
		{ 192, 360, 17 },
		{ 192, 361, 17 },
		{ 192, 362, 17 },
		{ 192, 363, 17 },
		{ 192, 364, 17 },
		{ 192, 365, 17 },
		{ 192, 366, 17 },
		{ 192, 367, 17 },
		{ 192, 368, 17 },
		{ 192, 369, 17 },
		{ 192, 370, 17 },
		{ 192, 371, 17 },
		{ 192, 372, 17 },
		{ 192, 373, 17 },
		{ 192, 374, 17 },
		{ 192, 375, 17 },
		{ 192, 376, 17 },
		{ 192, 377, 17 },
		{ 192, 378, 17 },
		{ 192, 379, 17 },
		{ 192, 380, 17 },
		{ 192, 381, 17 },
		{ 193, 322, 17 },
		{ 193, 323, 17 },
		{ 193, 324, 17 },
		{ 193, 325, 17 },
		{ 193, 326, 17 },
		{ 193, 327, 17 },
		{ 193, 328, 17 },
		{ 193, 329, 17 },
		{ 193, 330, 17 },
		{ 193, 331, 17 },
		{ 193, 332, 17 },
		{ 193, 333, 17 },
		{ 193, 334, 17 },
		{ 193, 335, 17 },
		{ 193, 336, 17 },
		{ 193, 337, 17 },
		{ 193, 338, 17 },
		{ 193, 339, 17 },
		{ 193, 340, 17 },
		{ 193, 341, 17 },
		{ 193, 342, 17 },
		{ 193, 343, 17 },
		{ 193, 344, 17 },
		{ 193, 345, 17 },
		{ 193, 346, 17 },
		{ 193, 347, 17 },
		{ 193, 348, 17 },
		{ 193, 349, 17 },
		{ 193, 350, 17 },
		{ 193, 351, 17 },
		{ 193, 352, 17 },
		{ 193, 353, 17 },
		{ 193, 354, 17 },
		{ 193, 355, 17 },
		{ 193, 356, 17 },
		{ 193, 357, 17 },
		{ 193, 358, 17 },
		{ 193, 359, 17 },
		{ 193, 360, 17 },
		{ 193, 361, 17 },
		{ 193, 362, 17 },
		{ 193, 363, 17 },
		{ 193, 364, 17 },
		{ 193, 365, 17 },
		{ 193, 366, 17 },
		{ 193, 367, 17 },
		{ 193, 368, 17 },
		{ 193, 369, 17 },
		{ 193, 370, 17 },
		{ 193, 371, 17 },
		{ 193, 372, 17 },
		{ 193, 373, 17 },
		{ 193, 374, 17 },
		{ 193, 375, 17 },
		{ 193, 376, 17 },
		{ 193, 377, 17 },
		{ 193, 378, 17 },
		{ 193, 379, 17 },
		{ 193, 380, 17 },
		{ 193, 381, 17 },
		{ 194, 321, 17 },
		{ 194, 322, 17 },
		{ 194, 323, 17 },
		{ 194, 324, 17 },
		{ 194, 325, 17 },
		{ 194, 326, 17 },
		{ 194, 327, 17 },
		{ 194, 328, 17 },
		{ 194, 329, 17 },
		{ 194, 330, 17 },
		{ 194, 331, 17 },
		{ 194, 332, 17 },
		{ 194, 333, 17 },
		{ 194, 334, 17 },
		{ 194, 335, 17 },
		{ 194, 336, 17 },
		{ 194, 337, 17 },
		{ 194, 338, 17 },
		{ 194, 339, 17 },
		{ 194, 340, 17 },
		{ 194, 341, 17 },
		{ 194, 342, 17 },
		{ 194, 343, 17 },
		{ 194, 344, 17 },
		{ 194, 345, 17 },
		{ 194, 346, 17 },
		{ 194, 347, 17 },
		{ 194, 348, 17 },
		{ 194, 349, 17 },
		{ 194, 350, 17 },
		{ 194, 351, 17 },
		{ 194, 352, 17 },
		{ 194, 353, 17 },
		{ 194, 354, 17 },
		{ 194, 355, 17 },
		{ 194, 356, 17 },
		{ 194, 357, 17 },
		{ 194, 358, 17 },
		{ 194, 359, 17 },
		{ 194, 360, 17 },
		{ 194, 361, 17 },
		{ 194, 362, 17 },
		{ 194, 363, 17 },
		{ 194, 364, 17 },
		{ 194, 365, 17 },
		{ 194, 366, 17 },
		{ 194, 367, 17 },
		{ 194, 368, 17 },
		{ 194, 369, 17 },
		{ 194, 370, 17 },
		{ 194, 371, 17 },
		{ 194, 372, 17 },
		{ 194, 373, 17 },
		{ 194, 374, 17 },
		{ 194, 375, 17 },
		{ 194, 376, 17 },
		{ 194, 377, 17 },
		{ 194, 378, 17 },
		{ 194, 379, 17 },
		{ 194, 380, 17 },
		{ 195, 320, 17 },
		{ 195, 321, 17 },
		{ 195, 322, 17 },
		{ 195, 323, 17 },
		{ 195, 324, 17 },
		{ 195, 325, 17 },
		{ 195, 326, 17 },
		{ 195, 327, 17 },
		{ 195, 328, 17 },
		{ 195, 329, 17 },
		{ 195, 330, 17 },
		{ 195, 331, 17 },
		{ 195, 332, 17 },
		{ 195, 333, 17 },
		{ 195, 334, 17 },
		{ 195, 335, 17 },
		{ 195, 336, 17 },
		{ 195, 337, 17 },
		{ 195, 338, 17 },
		{ 195, 339, 17 },
		{ 195, 340, 17 },
		{ 195, 341, 17 },
		{ 195, 342, 17 },
		{ 195, 343, 17 },
		{ 195, 344, 17 },
		{ 195, 345, 17 },
		{ 195, 346, 17 },
		{ 195, 347, 17 },
		{ 195, 348, 17 },
		{ 195, 349, 17 },
		{ 195, 350, 17 },
		{ 195, 351, 17 },
		{ 195, 352, 17 },
		{ 195, 353, 17 },
		{ 195, 354, 17 },
		{ 195, 355, 17 },
		{ 195, 356, 17 },
		{ 195, 357, 17 },
		{ 195, 358, 17 },
		{ 195, 359, 17 },
		{ 195, 360, 17 },
		{ 195, 361, 17 },
		{ 195, 362, 17 },
		{ 195, 363, 17 },
		{ 195, 364, 17 },
		{ 195, 365, 17 },
		{ 195, 366, 17 },
		{ 195, 367, 17 },
		{ 195, 368, 17 },
		{ 195, 369, 17 },
		{ 195, 370, 17 },
		{ 195, 371, 17 },
		{ 195, 372, 17 },
		{ 195, 373, 17 },
		{ 195, 374, 17 },
		{ 195, 375, 17 },
		{ 195, 376, 17 },
		{ 195, 377, 17 },
		{ 195, 378, 17 },
		{ 195, 379, 17 },
		{ 196, 319, 17 },
		{ 196, 320, 17 },
		{ 196, 321, 17 },
		{ 196, 322, 17 },
		{ 196, 323, 17 },
		{ 196, 324, 17 },
		{ 196, 325, 17 },
		{ 196, 326, 17 },
		{ 196, 327, 17 },
		{ 196, 328, 17 },
		{ 196, 329, 17 },
		{ 196, 330, 17 },
		{ 196, 331, 17 },
		{ 196, 332, 17 },
		{ 196, 333, 17 },
		{ 196, 334, 17 },
		{ 196, 335, 17 },
		{ 196, 336, 17 },
		{ 196, 337, 17 },
		{ 196, 338, 17 },
		{ 196, 339, 17 },
		{ 196, 340, 17 },
		{ 196, 341, 17 },
		{ 196, 342, 17 },
		{ 196, 343, 17 },
		{ 196, 344, 17 },
		{ 196, 345, 17 },
		{ 196, 346, 17 },
		{ 196, 347, 17 },
		{ 196, 348, 17 },
		{ 196, 349, 17 },
		{ 196, 350, 17 },
		{ 196, 351, 17 },
		{ 196, 352, 17 },
		{ 196, 353, 17 },
		{ 196, 354, 17 },
		{ 196, 355, 17 },
		{ 196, 356, 17 },
		{ 196, 357, 17 },
		{ 196, 358, 17 },
		{ 196, 359, 17 },
		{ 196, 360, 17 },
		{ 196, 361, 17 },
		{ 196, 362, 17 },
		{ 196, 363, 17 },
		{ 196, 364, 17 },
		{ 196, 365, 17 },
		{ 196, 366, 17 },
		{ 196, 367, 17 },
		{ 196, 368, 17 },
		{ 196, 369, 17 },
		{ 196, 370, 17 },
		{ 196, 371, 17 },
		{ 196, 372, 17 },
		{ 196, 373, 17 },
		{ 196, 374, 17 },
		{ 196, 375, 17 },
		{ 196, 376, 17 },
		{ 196, 377, 17 },
		{ 197, 318, 17 },
		{ 197, 319, 17 },
		{ 197, 320, 17 },
		{ 197, 321, 17 },
		{ 197, 322, 17 },
		{ 197, 323, 17 },
		{ 197, 324, 17 },
		{ 197, 325, 17 },
		{ 197, 326, 17 },
		{ 197, 327, 17 },
		{ 197, 328, 17 },
		{ 197, 329, 17 },
		{ 197, 330, 17 },
		{ 197, 331, 17 },
		{ 197, 332, 17 },
		{ 197, 333, 17 },
		{ 197, 334, 17 },
		{ 197, 335, 17 },
		{ 197, 336, 17 },
		{ 197, 337, 17 },
		{ 197, 338, 17 },
		{ 197, 339, 17 },
		{ 197, 340, 17 },
		{ 197, 341, 17 },
		{ 197, 342, 17 },
		{ 197, 343, 17 },
		{ 197, 344, 17 },
		{ 197, 345, 17 },
		{ 197, 346, 17 },
		{ 197, 347, 17 },
		{ 197, 348, 17 },
		{ 197, 349, 17 },
		{ 197, 350, 17 },
		{ 197, 351, 17 },
		{ 197, 352, 17 },
		{ 197, 353, 17 },
		{ 197, 354, 17 },
		{ 197, 355, 17 },
		{ 197, 356, 17 },
		{ 197, 357, 17 },
		{ 197, 358, 17 },
		{ 197, 359, 17 },
		{ 197, 360, 17 },
		{ 197, 361, 17 },
		{ 197, 362, 17 },
		{ 197, 363, 17 },
		{ 197, 364, 17 },
		{ 197, 365, 17 },
		{ 197, 366, 17 },
		{ 197, 367, 17 },
		{ 197, 368, 17 },
		{ 197, 369, 17 },
		{ 197, 370, 17 },
		{ 197, 371, 17 },
		{ 197, 372, 17 },
		{ 197, 373, 17 },
		{ 197, 374, 17 },
		{ 197, 375, 17 },
		{ 197, 376, 17 },
		{ 198, 317, 17 },
		{ 198, 318, 17 },
		{ 198, 319, 17 },
		{ 198, 320, 17 },
		{ 198, 321, 17 },
		{ 198, 322, 17 },
		{ 198, 323, 17 },
		{ 198, 324, 17 },
		{ 198, 325, 17 },
		{ 198, 326, 17 },
		{ 198, 327, 17 },
		{ 198, 328, 17 },
		{ 198, 329, 17 },
		{ 198, 330, 17 },
		{ 198, 331, 17 },
		{ 198, 332, 17 },
		{ 198, 333, 17 },
		{ 198, 334, 17 },
		{ 198, 335, 17 },
		{ 198, 336, 17 },
		{ 198, 337, 17 },
		{ 198, 338, 17 },
		{ 198, 339, 17 },
		{ 198, 340, 17 },
		{ 198, 341, 17 },
		{ 198, 342, 17 },
		{ 198, 343, 17 },
		{ 198, 344, 17 },
		{ 198, 345, 17 },
		{ 198, 346, 17 },
		{ 198, 347, 17 },
		{ 198, 348, 17 },
		{ 198, 349, 17 },
		{ 198, 350, 17 },
		{ 198, 351, 17 },
		{ 198, 352, 17 },
		{ 198, 353, 17 },
		{ 198, 354, 17 },
		{ 198, 355, 17 },
		{ 198, 356, 17 },
		{ 198, 357, 17 },
		{ 198, 358, 17 },
		{ 198, 359, 17 },
		{ 198, 360, 17 },
		{ 198, 361, 17 },
		{ 198, 362, 17 },
		{ 198, 363, 17 },
		{ 198, 364, 17 },
		{ 198, 365, 17 },
		{ 198, 366, 17 },
		{ 198, 367, 17 },
		{ 198, 368, 17 },
		{ 198, 369, 17 },
		{ 198, 370, 17 },
		{ 198, 371, 17 },
		{ 198, 372, 17 },
		{ 198, 373, 17 },
		{ 198, 374, 17 },
		{ 199, 316, 17 },
		{ 199, 317, 17 },
		{ 199, 318, 17 },
		{ 199, 319, 17 },
		{ 199, 320, 17 },
		{ 199, 321, 17 },
		{ 199, 322, 17 },
		{ 199, 323, 17 },
		{ 199, 324, 17 },
		{ 199, 325, 17 },
		{ 199, 326, 17 },
		{ 199, 327, 17 },
		{ 199, 328, 17 },
		{ 199, 329, 17 },
		{ 199, 330, 17 },
		{ 199, 331, 17 },
		{ 199, 332, 17 },
		{ 199, 333, 17 },
		{ 199, 334, 17 },
		{ 199, 335, 17 },
		{ 199, 336, 17 },
		{ 199, 337, 17 },
		{ 199, 338, 17 },
		{ 199, 339, 17 },
		{ 199, 340, 17 },
		{ 199, 341, 17 },
		{ 199, 342, 17 },
		{ 199, 343, 17 },
		{ 199, 344, 17 },
		{ 199, 345, 17 },
		{ 199, 346, 17 },
		{ 199, 347, 17 },
		{ 199, 348, 17 },
		{ 199, 349, 17 },
		{ 199, 350, 17 },
		{ 199, 351, 17 },
		{ 199, 352, 17 },
		{ 199, 353, 17 },
		{ 199, 354, 17 },
		{ 199, 355, 17 },
		{ 199, 356, 17 },
		{ 199, 357, 17 },
		{ 199, 358, 17 },
		{ 199, 359, 17 },
		{ 199, 360, 17 },
		{ 199, 361, 17 },
		{ 199, 362, 17 },
		{ 199, 363, 17 },
		{ 199, 364, 17 },
		{ 199, 365, 17 },
		{ 199, 366, 17 },
		{ 199, 367, 17 },
		{ 199, 368, 17 },
		{ 199, 369, 17 },
		{ 199, 370, 17 },
		{ 199, 371, 17 },
		{ 199, 372, 17 },
		{ 200, 315, 17 },
		{ 200, 316, 17 },
		{ 200, 317, 17 },
		{ 200, 318, 17 },
		{ 200, 319, 17 },
		{ 200, 320, 17 },
		{ 200, 321, 17 },
		{ 200, 322, 17 },
		{ 200, 323, 17 },
		{ 200, 324, 17 },
		{ 200, 325, 17 },
		{ 200, 326, 17 },
		{ 200, 327, 17 },
		{ 200, 328, 17 },
		{ 200, 329, 17 },
		{ 200, 330, 17 },
		{ 200, 331, 17 },
		{ 200, 332, 17 },
		{ 200, 333, 17 },
		{ 200, 334, 17 },
		{ 200, 335, 17 },
		{ 200, 336, 17 },
		{ 200, 337, 17 },
		{ 200, 338, 17 },
		{ 200, 339, 17 },
		{ 200, 340, 17 },
		{ 200, 341, 17 },
		{ 200, 342, 17 },
		{ 200, 343, 17 },
		{ 200, 344, 17 },
		{ 200, 345, 17 },
		{ 200, 346, 17 },
		{ 200, 347, 17 },
		{ 200, 348, 17 },
		{ 200, 349, 17 },
		{ 200, 350, 17 },
		{ 200, 351, 17 },
		{ 200, 352, 17 },
		{ 200, 353, 17 },
		{ 200, 354, 17 },
		{ 200, 355, 17 },
		{ 200, 356, 17 },
		{ 200, 357, 17 },
		{ 200, 358, 17 },
		{ 200, 359, 17 },
		{ 200, 360, 17 },
		{ 200, 361, 17 },
		{ 200, 362, 17 },
		{ 200, 363, 17 },
		{ 200, 364, 17 },
		{ 200, 365, 17 },
		{ 200, 366, 17 },
		{ 200, 367, 17 },
		{ 200, 368, 17 },
		{ 200, 369, 17 },
		{ 200, 370, 17 },
		{ 201, 314, 17 },
		{ 201, 315, 17 },
		{ 201, 316, 17 },
		{ 201, 317, 17 },
		{ 201, 318, 17 },
		{ 201, 319, 17 },
		{ 201, 320, 17 },
		{ 201, 321, 17 },
		{ 201, 322, 17 },
		{ 201, 323, 17 },
		{ 201, 324, 17 },
		{ 201, 325, 17 },
		{ 201, 326, 17 },
		{ 201, 327, 17 },
		{ 201, 328, 17 },
		{ 201, 329, 17 },
		{ 201, 330, 17 },
		{ 201, 331, 17 },
		{ 201, 332, 17 },
		{ 201, 333, 17 },
		{ 201, 334, 17 },
		{ 201, 335, 17 },
		{ 201, 336, 17 },
		{ 201, 337, 17 },
		{ 201, 338, 17 },
		{ 201, 339, 17 },
		{ 201, 340, 17 },
		{ 201, 341, 17 },
		{ 201, 342, 17 },
		{ 201, 343, 17 },
		{ 201, 344, 17 },
		{ 201, 345, 17 },
		{ 201, 346, 17 },
		{ 201, 347, 17 },
		{ 201, 348, 17 },
		{ 201, 349, 17 },
		{ 201, 350, 17 },
		{ 201, 351, 17 },
		{ 201, 352, 17 },
		{ 201, 353, 17 },
		{ 201, 354, 17 },
		{ 201, 355, 17 },
		{ 201, 356, 17 },
		{ 201, 357, 17 },
		{ 201, 358, 17 },
		{ 201, 359, 17 },
		{ 201, 360, 17 },
		{ 201, 361, 17 },
		{ 201, 362, 17 },
		{ 201, 363, 17 },
		{ 201, 364, 17 },
		{ 201, 365, 17 },
		{ 201, 366, 17 },
		{ 201, 367, 17 },
		{ 201, 368, 17 },
		{ 202, 313, 17 },
		{ 202, 314, 17 },
		{ 202, 315, 17 },
		{ 202, 316, 17 },
		{ 202, 317, 17 },
		{ 202, 318, 17 },
		{ 202, 319, 17 },
		{ 202, 320, 17 },
		{ 202, 321, 17 },
		{ 202, 322, 17 },
		{ 202, 323, 17 },
		{ 202, 324, 17 },
		{ 202, 325, 17 },
		{ 202, 326, 17 },
		{ 202, 327, 17 },
		{ 202, 328, 17 },
		{ 202, 329, 17 },
		{ 202, 330, 17 },
		{ 202, 331, 17 },
		{ 202, 332, 17 },
		{ 202, 333, 17 },
		{ 202, 334, 17 },
		{ 202, 335, 17 },
		{ 202, 336, 17 },
		{ 202, 337, 17 },
		{ 202, 338, 17 },
		{ 202, 339, 17 },
		{ 202, 340, 17 },
		{ 202, 341, 17 },
		{ 202, 342, 17 },
		{ 202, 343, 17 },
		{ 202, 344, 17 },
		{ 202, 345, 17 },
		{ 202, 346, 17 },
		{ 202, 347, 17 },
		{ 202, 348, 17 },
		{ 202, 349, 17 },
		{ 202, 350, 17 },
		{ 202, 351, 17 },
		{ 202, 352, 17 },
		{ 202, 353, 17 },
		{ 202, 354, 17 },
		{ 202, 355, 17 },
		{ 202, 356, 17 },
		{ 202, 357, 17 },
		{ 202, 358, 17 },
		{ 202, 359, 17 },
		{ 202, 360, 17 },
		{ 202, 361, 17 },
		{ 202, 362, 17 },
		{ 202, 363, 17 },
		{ 202, 364, 17 },
		{ 202, 365, 17 },
		{ 202, 366, 17 },
		{ 203, 312, 17 },
		{ 203, 313, 17 },
		{ 203, 314, 17 },
		{ 203, 315, 17 },
		{ 203, 316, 17 },
		{ 203, 317, 17 },
		{ 203, 318, 17 },
		{ 203, 319, 17 },
		{ 203, 320, 17 },
		{ 203, 321, 17 },
		{ 203, 322, 17 },
		{ 203, 323, 17 },
		{ 203, 324, 17 },
		{ 203, 325, 17 },
		{ 203, 326, 17 },
		{ 203, 327, 17 },
		{ 203, 328, 17 },
		{ 203, 329, 17 },
		{ 203, 330, 17 },
		{ 203, 331, 17 },
		{ 203, 332, 17 },
		{ 203, 333, 17 },
		{ 203, 334, 17 },
		{ 203, 335, 17 },
		{ 203, 336, 17 },
		{ 203, 337, 17 },
		{ 203, 338, 17 },
		{ 203, 339, 17 },
		{ 203, 340, 17 },
		{ 203, 341, 17 },
		{ 203, 342, 17 },
		{ 203, 343, 17 },
		{ 203, 344, 17 },
		{ 203, 345, 17 },
		{ 203, 346, 17 },
		{ 203, 347, 17 },
		{ 203, 348, 17 },
		{ 203, 349, 17 },
		{ 203, 350, 17 },
		{ 203, 351, 17 },
		{ 203, 352, 17 },
		{ 203, 353, 17 },
		{ 203, 354, 17 },
		{ 203, 355, 17 },
		{ 203, 356, 17 },
		{ 203, 357, 17 },
		{ 203, 358, 17 },
		{ 203, 359, 17 },
		{ 203, 360, 17 },
		{ 203, 361, 17 },
		{ 203, 362, 17 },
		{ 203, 363, 17 },
		{ 203, 364, 17 },
		{ 204, 311, 17 },
		{ 204, 312, 17 },
		{ 204, 313, 17 },
		{ 204, 314, 17 },
		{ 204, 315, 17 },
		{ 204, 316, 17 },
		{ 204, 317, 17 },
		{ 204, 318, 17 },
		{ 204, 319, 17 },
		{ 204, 320, 17 },
		{ 204, 321, 17 },
		{ 204, 322, 17 },
		{ 204, 323, 17 },
		{ 204, 324, 17 },
		{ 204, 325, 17 },
		{ 204, 326, 17 },
		{ 204, 327, 17 },
		{ 204, 328, 17 },
		{ 204, 329, 17 },
		{ 204, 330, 17 },
		{ 204, 331, 17 },
		{ 204, 332, 17 },
		{ 204, 333, 17 },
		{ 204, 334, 17 },
		{ 204, 335, 17 },
		{ 204, 336, 17 },
		{ 204, 337, 17 },
		{ 204, 338, 17 },
		{ 204, 339, 17 },
		{ 204, 340, 17 },
		{ 204, 341, 17 },
		{ 204, 342, 17 },
		{ 204, 343, 17 },
		{ 204, 344, 17 },
		{ 204, 345, 17 },
		{ 204, 346, 17 },
		{ 204, 347, 17 },
		{ 204, 348, 17 },
		{ 204, 349, 17 },
		{ 204, 350, 17 },
		{ 204, 351, 17 },
		{ 204, 352, 17 },
		{ 204, 353, 17 },
		{ 204, 354, 17 },
		{ 204, 355, 17 },
		{ 204, 356, 17 },
		{ 204, 357, 17 },
		{ 204, 358, 17 },
		{ 204, 359, 17 },
		{ 204, 360, 17 },
		{ 204, 361, 17 },
		{ 204, 362, 17 },
		{ 205, 310, 17 },
		{ 205, 311, 17 },
		{ 205, 312, 17 },
		{ 205, 313, 17 },
		{ 205, 314, 17 },
		{ 205, 315, 17 },
		{ 205, 316, 17 },
		{ 205, 317, 17 },
		{ 205, 318, 17 },
		{ 205, 319, 17 },
		{ 205, 320, 17 },
		{ 205, 321, 17 },
		{ 205, 322, 17 },
		{ 205, 323, 17 },
		{ 205, 324, 17 },
		{ 205, 325, 17 },
		{ 205, 326, 17 },
		{ 205, 327, 17 },
		{ 205, 328, 17 },
		{ 205, 329, 17 },
		{ 205, 330, 17 },
		{ 205, 331, 17 },
		{ 205, 332, 17 },
		{ 205, 333, 17 },
		{ 205, 334, 17 },
		{ 205, 335, 17 },
		{ 205, 336, 17 },
		{ 205, 337, 17 },
		{ 205, 338, 17 },
		{ 205, 339, 17 },
		{ 205, 340, 17 },
		{ 205, 341, 17 },
		{ 205, 342, 17 },
		{ 205, 343, 17 },
		{ 205, 344, 17 },
		{ 205, 345, 17 },
		{ 205, 346, 17 },
		{ 205, 347, 17 },
		{ 205, 348, 17 },
		{ 205, 349, 17 },
		{ 205, 350, 17 },
		{ 205, 351, 17 },
		{ 205, 352, 17 },
		{ 205, 353, 17 },
		{ 205, 354, 17 },
		{ 205, 355, 17 },
		{ 205, 356, 17 },
		{ 205, 357, 17 },
		{ 205, 358, 17 },
		{ 205, 359, 17 },
		{ 205, 360, 17 },
		{ 206, 309, 17 },
		{ 206, 310, 17 },
		{ 206, 311, 17 },
		{ 206, 312, 17 },
		{ 206, 313, 17 },
		{ 206, 314, 17 },
		{ 206, 315, 17 },
		{ 206, 316, 17 },
		{ 206, 317, 17 },
		{ 206, 318, 17 },
		{ 206, 319, 17 },
		{ 206, 320, 17 },
		{ 206, 321, 17 },
		{ 206, 322, 17 },
		{ 206, 323, 17 },
		{ 206, 324, 17 },
		{ 206, 325, 17 },
		{ 206, 326, 17 },
		{ 206, 327, 17 },
		{ 206, 328, 17 },
		{ 206, 329, 17 },
		{ 206, 330, 17 },
		{ 206, 331, 17 },
		{ 206, 332, 17 },
		{ 206, 333, 17 },
		{ 206, 334, 17 },
		{ 206, 335, 17 },
		{ 206, 336, 17 },
		{ 206, 337, 17 },
		{ 206, 338, 17 },
		{ 206, 339, 17 },
		{ 206, 340, 17 },
		{ 206, 341, 17 },
		{ 206, 342, 17 },
		{ 206, 343, 17 },
		{ 206, 344, 17 },
		{ 206, 345, 17 },
		{ 206, 346, 17 },
		{ 206, 347, 17 },
		{ 206, 348, 17 },
		{ 206, 349, 17 },
		{ 206, 350, 17 },
		{ 206, 351, 17 },
		{ 206, 352, 17 },
		{ 206, 353, 17 },
		{ 206, 354, 17 },
		{ 206, 355, 17 },
		{ 206, 356, 17 },
		{ 206, 357, 17 },
		{ 206, 358, 17 },
		{ 207, 308, 17 },
		{ 207, 309, 17 },
		{ 207, 310, 17 },
		{ 207, 311, 17 },
		{ 207, 312, 17 },
		{ 207, 313, 17 },
		{ 207, 314, 17 },
		{ 207, 315, 17 },
		{ 207, 316, 17 },
		{ 207, 317, 17 },
		{ 207, 318, 17 },
		{ 207, 319, 17 },
		{ 207, 320, 17 },
		{ 207, 321, 17 },
		{ 207, 322, 17 },
		{ 207, 323, 17 },
		{ 207, 324, 17 },
		{ 207, 325, 17 },
		{ 207, 326, 17 },
		{ 207, 327, 17 },
		{ 207, 328, 17 },
		{ 207, 329, 17 },
		{ 207, 330, 17 },
		{ 207, 331, 17 },
		{ 207, 332, 17 },
		{ 207, 333, 17 },
		{ 207, 334, 17 },
		{ 207, 335, 17 },
		{ 207, 336, 17 },
		{ 207, 337, 17 },
		{ 207, 338, 17 },
		{ 207, 339, 17 },
		{ 207, 340, 17 },
		{ 207, 341, 17 },
		{ 207, 342, 17 },
		{ 207, 343, 17 },
		{ 207, 344, 17 },
		{ 207, 345, 17 },
		{ 207, 346, 17 },
		{ 207, 347, 17 },
		{ 207, 348, 17 },
		{ 207, 349, 17 },
		{ 207, 350, 17 },
		{ 207, 351, 17 },
		{ 207, 352, 17 },
		{ 207, 353, 17 },
		{ 207, 354, 17 },
		{ 207, 355, 17 },
		{ 207, 356, 17 },
		{ 208, 307, 17 },
		{ 208, 308, 17 },
		{ 208, 309, 17 },
		{ 208, 310, 17 },
		{ 208, 311, 17 },
		{ 208, 312, 17 },
		{ 208, 313, 17 },
		{ 208, 314, 17 },
		{ 208, 315, 17 },
		{ 208, 316, 17 },
		{ 208, 317, 17 },
		{ 208, 318, 17 },
		{ 208, 319, 17 },
		{ 208, 320, 17 },
		{ 208, 321, 17 },
		{ 208, 322, 17 },
		{ 208, 323, 17 },
		{ 208, 324, 17 },
		{ 208, 325, 17 },
		{ 208, 326, 17 },
		{ 208, 327, 17 },
		{ 208, 328, 17 },
		{ 208, 329, 17 },
		{ 208, 330, 17 },
		{ 208, 331, 17 },
		{ 208, 332, 17 },
		{ 208, 333, 17 },
		{ 208, 334, 17 },
		{ 208, 335, 17 },
		{ 208, 336, 17 },
		{ 208, 337, 17 },
		{ 208, 338, 17 },
		{ 208, 339, 17 },
		{ 208, 340, 17 },
		{ 208, 341, 17 },
		{ 208, 342, 17 },
		{ 208, 343, 17 },
		{ 208, 344, 17 },
		{ 208, 345, 17 },
		{ 208, 346, 17 },
		{ 208, 347, 17 },
		{ 208, 348, 17 },
		{ 208, 349, 17 },
		{ 208, 350, 17 },
		{ 208, 351, 17 },
		{ 208, 352, 17 },
		{ 208, 353, 17 },
		{ 208, 354, 17 },
		{ 209, 306, 17 },
		{ 209, 307, 17 },
		{ 209, 308, 17 },
		{ 209, 309, 17 },
		{ 209, 310, 17 },
		{ 209, 311, 17 },
		{ 209, 312, 17 },
		{ 209, 313, 17 },
		{ 209, 314, 17 },
		{ 209, 315, 17 },
		{ 209, 316, 17 },
		{ 209, 317, 17 },
		{ 209, 318, 17 },
		{ 209, 319, 17 },
		{ 209, 320, 17 },
		{ 209, 321, 17 },
		{ 209, 322, 17 },
		{ 209, 323, 17 },
		{ 209, 324, 17 },
		{ 209, 325, 17 },
		{ 209, 326, 17 },
		{ 209, 327, 17 },
		{ 209, 328, 17 },
		{ 209, 329, 17 },
		{ 209, 330, 17 },
		{ 209, 331, 17 },
		{ 209, 332, 17 },
		{ 209, 333, 17 },
		{ 209, 334, 17 },
		{ 209, 335, 17 },
		{ 209, 336, 17 },
		{ 209, 337, 17 },
		{ 209, 338, 17 },
		{ 209, 339, 17 },
		{ 209, 340, 17 },
		{ 209, 341, 17 },
		{ 209, 342, 17 },
		{ 209, 343, 17 },
		{ 209, 344, 17 },
		{ 209, 345, 17 },
		{ 209, 346, 17 },
		{ 209, 347, 17 },
		{ 209, 348, 17 },
		{ 209, 349, 17 },
		{ 209, 350, 17 },
		{ 209, 351, 17 },
		{ 209, 352, 17 },
		{ 210, 305, 17 },
		{ 210, 306, 17 },
		{ 210, 307, 17 },
		{ 210, 308, 17 },
		{ 210, 309, 17 },
		{ 210, 310, 17 },
		{ 210, 311, 17 },
		{ 210, 312, 17 },
		{ 210, 313, 17 },
		{ 210, 314, 17 },
		{ 210, 315, 17 },
		{ 210, 316, 17 },
		{ 210, 317, 17 },
		{ 210, 318, 17 },
		{ 210, 319, 17 },
		{ 210, 320, 17 },
		{ 210, 321, 17 },
		{ 210, 322, 17 },
		{ 210, 323, 17 },
		{ 210, 324, 17 },
		{ 210, 325, 17 },
		{ 210, 326, 17 },
		{ 210, 327, 17 },
		{ 210, 328, 17 },
		{ 210, 329, 17 },
		{ 210, 330, 17 },
		{ 210, 331, 17 },
		{ 210, 332, 17 },
		{ 210, 333, 17 },
		{ 210, 334, 17 },
		{ 210, 335, 17 },
		{ 210, 336, 17 },
		{ 210, 337, 17 },
		{ 210, 338, 17 },
		{ 210, 339, 17 },
		{ 210, 340, 17 },
		{ 210, 341, 17 },
		{ 210, 342, 17 },
		{ 210, 343, 17 },
		{ 210, 344, 17 },
		{ 210, 345, 17 },
		{ 210, 346, 17 },
		{ 210, 347, 17 },
		{ 210, 348, 17 },
		{ 210, 349, 17 },
		{ 210, 350, 17 },
		{ 211, 304, 17 },
		{ 211, 305, 17 },
		{ 211, 306, 17 },
		{ 211, 307, 17 },
		{ 211, 308, 17 },
		{ 211, 309, 17 },
		{ 211, 310, 17 },
		{ 211, 311, 17 },
		{ 211, 312, 17 },
		{ 211, 313, 17 },
		{ 211, 314, 17 },
		{ 211, 315, 17 },
		{ 211, 316, 17 },
		{ 211, 317, 17 },
		{ 211, 318, 17 },
		{ 211, 319, 17 },
		{ 211, 320, 17 },
		{ 211, 321, 17 },
		{ 211, 322, 17 },
		{ 211, 323, 17 },
		{ 211, 324, 17 },
		{ 211, 325, 17 },
		{ 211, 326, 17 },
		{ 211, 327, 17 },
		{ 211, 328, 17 },
		{ 211, 329, 17 },
		{ 211, 330, 17 },
		{ 211, 331, 17 },
		{ 211, 332, 17 },
		{ 211, 333, 17 },
		{ 211, 334, 17 },
		{ 211, 335, 17 },
		{ 211, 336, 17 },
		{ 211, 337, 17 },
		{ 211, 338, 17 },
		{ 211, 339, 17 },
		{ 211, 340, 17 },
		{ 211, 341, 17 },
		{ 211, 342, 17 },
		{ 211, 343, 17 },
		{ 211, 344, 17 },
		{ 211, 345, 17 },
		{ 211, 346, 17 },
		{ 211, 347, 17 },
		{ 211, 348, 17 },
		{ 212, 303, 17 },
		{ 212, 304, 17 },
		{ 212, 305, 17 },
		{ 212, 306, 17 },
		{ 212, 307, 17 },
		{ 212, 308, 17 },
		{ 212, 309, 17 },
		{ 212, 310, 17 },
		{ 212, 311, 17 },
		{ 212, 312, 17 },
		{ 212, 313, 17 },
		{ 212, 314, 17 },
		{ 212, 315, 17 },
		{ 212, 316, 17 },
		{ 212, 317, 17 },
		{ 212, 318, 17 },
		{ 212, 319, 17 },
		{ 212, 320, 17 },
		{ 212, 321, 17 },
		{ 212, 322, 17 },
		{ 212, 323, 17 },
		{ 212, 324, 17 },
		{ 212, 325, 17 },
		{ 212, 326, 17 },
		{ 212, 327, 17 },
		{ 212, 328, 17 },
		{ 212, 329, 17 },
		{ 212, 330, 17 },
		{ 212, 331, 17 },
		{ 212, 332, 17 },
		{ 212, 333, 17 },
		{ 212, 334, 17 },
		{ 212, 335, 17 },
		{ 212, 336, 17 },
		{ 212, 337, 17 },
		{ 212, 338, 17 },
		{ 212, 339, 17 },
		{ 212, 340, 17 },
		{ 212, 341, 17 },
		{ 212, 342, 17 },
		{ 212, 343, 17 },
		{ 212, 344, 17 },
		{ 212, 345, 17 },
		{ 212, 346, 17 },
		{ 213, 302, 17 },
		{ 213, 303, 17 },
		{ 213, 304, 17 },
		{ 213, 305, 17 },
		{ 213, 306, 17 },
		{ 213, 307, 17 },
		{ 213, 308, 17 },
		{ 213, 309, 17 },
		{ 213, 310, 17 },
		{ 213, 311, 17 },
		{ 213, 312, 17 },
		{ 213, 313, 17 },
		{ 213, 314, 17 },
		{ 213, 315, 17 },
		{ 213, 316, 17 },
		{ 213, 317, 17 },
		{ 213, 318, 17 },
		{ 213, 319, 17 },
		{ 213, 320, 17 },
		{ 213, 321, 17 },
		{ 213, 322, 17 },
		{ 213, 323, 17 },
		{ 213, 324, 17 },
		{ 213, 325, 17 },
		{ 213, 326, 17 },
		{ 213, 327, 17 },
		{ 213, 328, 17 },
		{ 213, 329, 17 },
		{ 213, 330, 17 },
		{ 213, 331, 17 },
		{ 213, 332, 17 },
		{ 213, 333, 17 },
		{ 213, 334, 17 },
		{ 213, 335, 17 },
		{ 213, 336, 17 },
		{ 213, 337, 17 },
		{ 213, 338, 17 },
		{ 213, 339, 17 },
		{ 213, 340, 17 },
		{ 213, 341, 17 },
		{ 213, 342, 17 },
		{ 213, 343, 17 },
		{ 213, 344, 17 },
		{ 214, 301, 17 },
		{ 214, 302, 17 },
		{ 214, 303, 17 },
		{ 214, 304, 17 },
		{ 214, 305, 17 },
		{ 214, 306, 17 },
		{ 214, 307, 17 },
		{ 214, 308, 17 },
		{ 214, 309, 17 },
		{ 214, 310, 17 },
		{ 214, 311, 17 },
		{ 214, 312, 17 },
		{ 214, 313, 17 },
		{ 214, 314, 17 },
		{ 214, 315, 17 },
		{ 214, 316, 17 },
		{ 214, 317, 17 },
		{ 214, 318, 17 },
		{ 214, 319, 17 },
		{ 214, 320, 17 },
		{ 214, 321, 17 },
		{ 214, 322, 17 },
		{ 214, 323, 17 },
		{ 214, 324, 17 },
		{ 214, 325, 17 },
		{ 214, 326, 17 },
		{ 214, 327, 17 },
		{ 214, 328, 17 },
		{ 214, 329, 17 },
		{ 214, 330, 17 },
		{ 214, 331, 17 },
		{ 214, 332, 17 },
		{ 214, 333, 17 },
		{ 214, 334, 17 },
		{ 214, 335, 17 },
		{ 214, 336, 17 },
		{ 214, 337, 17 },
		{ 214, 338, 17 },
		{ 214, 339, 17 },
		{ 214, 340, 17 },
		{ 214, 341, 17 },
		{ 214, 342, 17 },
		{ 215, 300, 17 },
		{ 215, 301, 17 },
		{ 215, 302, 17 },
		{ 215, 303, 17 },
		{ 215, 304, 17 },
		{ 215, 305, 17 },
		{ 215, 306, 17 },
		{ 215, 307, 17 },
		{ 215, 308, 17 },
		{ 215, 309, 17 },
		{ 215, 310, 17 },
		{ 215, 311, 17 },
		{ 215, 312, 17 },
		{ 215, 313, 17 },
		{ 215, 314, 17 },
		{ 215, 315, 17 },
		{ 215, 316, 17 },
		{ 215, 317, 17 },
		{ 215, 318, 17 },
		{ 215, 319, 17 },
		{ 215, 320, 17 },
		{ 215, 321, 17 },
		{ 215, 322, 17 },
		{ 215, 323, 17 },
		{ 215, 324, 17 },
		{ 215, 325, 17 },
		{ 215, 326, 17 },
		{ 215, 327, 17 },
		{ 215, 328, 17 },
		{ 215, 329, 17 },
		{ 215, 330, 17 },
		{ 215, 331, 17 },
		{ 215, 332, 17 },
		{ 215, 333, 17 },
		{ 215, 334, 17 },
		{ 215, 335, 17 },
		{ 215, 336, 17 },
		{ 215, 337, 17 },
		{ 215, 338, 17 },
		{ 215, 339, 17 },
		{ 215, 340, 17 },
		{ 216, 299, 17 },
		{ 216, 300, 17 },
		{ 216, 301, 17 },
		{ 216, 302, 17 },
		{ 216, 303, 17 },
		{ 216, 304, 17 },
		{ 216, 305, 17 },
		{ 216, 306, 17 },
		{ 216, 307, 17 },
		{ 216, 308, 17 },
		{ 216, 309, 17 },
		{ 216, 310, 17 },
		{ 216, 311, 17 },
		{ 216, 312, 17 },
		{ 216, 313, 17 },
		{ 216, 314, 17 },
		{ 216, 315, 17 },
		{ 216, 316, 17 },
		{ 216, 317, 17 },
		{ 216, 318, 17 },
		{ 216, 319, 17 },
		{ 216, 320, 17 },
		{ 216, 321, 17 },
		{ 216, 322, 17 },
		{ 216, 323, 17 },
		{ 216, 324, 17 },
		{ 216, 325, 17 },
		{ 216, 326, 17 },
		{ 216, 327, 17 },
		{ 216, 328, 17 },
		{ 216, 329, 17 },
		{ 216, 330, 17 },
		{ 216, 331, 17 },
		{ 216, 332, 17 },
		{ 216, 333, 17 },
		{ 216, 334, 17 },
		{ 216, 335, 17 },
		{ 216, 336, 17 },
		{ 216, 337, 17 },
		{ 216, 338, 17 },
		{ 217, 298, 17 },
		{ 217, 299, 17 },
		{ 217, 300, 17 },
		{ 217, 301, 17 },
		{ 217, 302, 17 },
		{ 217, 303, 17 },
		{ 217, 304, 17 },
		{ 217, 305, 17 },
		{ 217, 306, 17 },
		{ 217, 307, 17 },
		{ 217, 308, 17 },
		{ 217, 309, 17 },
		{ 217, 310, 17 },
		{ 217, 311, 17 },
		{ 217, 312, 17 },
		{ 217, 313, 17 },
		{ 217, 314, 17 },
		{ 217, 315, 17 },
		{ 217, 316, 17 },
		{ 217, 317, 17 },
		{ 217, 318, 17 },
		{ 217, 319, 17 },
		{ 217, 320, 17 },
		{ 217, 321, 17 },
		{ 217, 322, 17 },
		{ 217, 323, 17 },
		{ 217, 324, 17 },
		{ 217, 325, 17 },
		{ 217, 326, 17 },
		{ 217, 327, 17 },
		{ 217, 328, 17 },
		{ 217, 329, 17 },
		{ 217, 330, 17 },
		{ 217, 331, 17 },
		{ 217, 332, 17 },
		{ 217, 333, 17 },
		{ 217, 334, 17 },
		{ 217, 335, 17 },
		{ 217, 336, 17 },
		{ 218, 297, 17 },
		{ 218, 298, 17 },
		{ 218, 299, 17 },
		{ 218, 300, 17 },
		{ 218, 301, 17 },
		{ 218, 302, 17 },
		{ 218, 303, 17 },
		{ 218, 304, 17 },
		{ 218, 305, 17 },
		{ 218, 306, 17 },
		{ 218, 307, 17 },
		{ 218, 308, 17 },
		{ 218, 309, 17 },
		{ 218, 310, 17 },
		{ 218, 311, 17 },
		{ 218, 312, 17 },
		{ 218, 313, 17 },
		{ 218, 314, 17 },
		{ 218, 315, 17 },
		{ 218, 316, 17 },
		{ 218, 317, 17 },
		{ 218, 318, 17 },
		{ 218, 319, 17 },
		{ 218, 320, 17 },
		{ 218, 321, 17 },
		{ 218, 322, 17 },
		{ 218, 323, 17 },
		{ 218, 324, 17 },
		{ 218, 325, 17 },
		{ 218, 326, 17 },
		{ 218, 327, 17 },
		{ 218, 328, 17 },
		{ 218, 329, 17 },
		{ 218, 330, 17 },
		{ 218, 331, 17 },
		{ 218, 332, 17 },
		{ 218, 333, 17 },
		{ 218, 334, 17 },
		{ 219, 296, 17 },
		{ 219, 297, 17 },
		{ 219, 298, 17 },
		{ 219, 299, 17 },
		{ 219, 300, 17 },
		{ 219, 301, 17 },
		{ 219, 302, 17 },
		{ 219, 303, 17 },
		{ 219, 304, 17 },
		{ 219, 305, 17 },
		{ 219, 306, 17 },
		{ 219, 307, 17 },
		{ 219, 308, 17 },
		{ 219, 309, 17 },
		{ 219, 310, 17 },
		{ 219, 311, 17 },
		{ 219, 312, 17 },
		{ 219, 313, 17 },
		{ 219, 314, 17 },
		{ 219, 315, 17 },
		{ 219, 316, 17 },
		{ 219, 317, 17 },
		{ 219, 318, 17 },
		{ 219, 319, 17 },
		{ 219, 320, 17 },
		{ 219, 321, 17 },
		{ 219, 322, 17 },
		{ 219, 323, 17 },
		{ 219, 324, 17 },
		{ 219, 325, 17 },
		{ 219, 326, 17 },
		{ 219, 327, 17 },
		{ 219, 328, 17 },
		{ 219, 329, 17 },
		{ 219, 330, 17 },
		{ 219, 331, 17 },
		{ 219, 332, 17 },
		{ 220, 295, 17 },
		{ 220, 296, 17 },
		{ 220, 297, 17 },
		{ 220, 298, 17 },
		{ 220, 299, 17 },
		{ 220, 300, 17 },
		{ 220, 301, 17 },
		{ 220, 302, 17 },
		{ 220, 303, 17 },
		{ 220, 304, 17 },
		{ 220, 305, 17 },
		{ 220, 306, 17 },
		{ 220, 307, 17 },
		{ 220, 308, 17 },
		{ 220, 309, 17 },
		{ 220, 310, 17 },
		{ 220, 311, 17 },
		{ 220, 312, 17 },
		{ 220, 313, 17 },
		{ 220, 314, 17 },
		{ 220, 315, 17 },
		{ 220, 316, 17 },
		{ 220, 317, 17 },
		{ 220, 318, 17 },
		{ 220, 319, 17 },
		{ 220, 320, 17 },
		{ 220, 321, 17 },
		{ 220, 322, 17 },
		{ 220, 323, 17 },
		{ 220, 324, 17 },
		{ 220, 325, 17 },
		{ 220, 326, 17 },
		{ 220, 327, 17 },
		{ 220, 328, 17 },
		{ 220, 329, 17 },
		{ 220, 330, 17 },
		{ 221, 294, 17 },
		{ 221, 295, 17 },
		{ 221, 296, 17 },
		{ 221, 297, 17 },
		{ 221, 298, 17 },
		{ 221, 299, 17 },
		{ 221, 300, 17 },
		{ 221, 301, 17 },
		{ 221, 302, 17 },
		{ 221, 303, 17 },
		{ 221, 304, 17 },
		{ 221, 305, 17 },
		{ 221, 306, 17 },
		{ 221, 307, 17 },
		{ 221, 308, 17 },
		{ 221, 309, 17 },
		{ 221, 310, 17 },
		{ 221, 311, 17 },
		{ 221, 312, 17 },
		{ 221, 313, 17 },
		{ 221, 314, 17 },
		{ 221, 315, 17 },
		{ 221, 316, 17 },
		{ 221, 317, 17 },
		{ 221, 318, 17 },
		{ 221, 319, 17 },
		{ 221, 320, 17 },
		{ 221, 321, 17 },
		{ 221, 322, 17 },
		{ 221, 323, 17 },
		{ 221, 324, 17 },
		{ 221, 325, 17 },
		{ 221, 326, 17 },
		{ 221, 327, 17 },
		{ 221, 328, 17 },
		{ 222, 293, 17 },
		{ 222, 294, 17 },
		{ 222, 295, 17 },
		{ 222, 296, 17 },
		{ 222, 297, 17 },
		{ 222, 298, 17 },
		{ 222, 299, 17 },
		{ 222, 300, 17 },
		{ 222, 301, 17 },
		{ 222, 302, 17 },
		{ 222, 303, 17 },
		{ 222, 304, 17 },
		{ 222, 305, 17 },
		{ 222, 306, 17 },
		{ 222, 307, 17 },
		{ 222, 308, 17 },
		{ 222, 309, 17 },
		{ 222, 310, 17 },
		{ 222, 311, 17 },
		{ 222, 312, 17 },
		{ 222, 313, 17 },
		{ 222, 314, 17 },
		{ 222, 315, 17 },
		{ 222, 316, 17 },
		{ 222, 317, 17 },
		{ 222, 318, 17 },
		{ 222, 319, 17 },
		{ 222, 320, 17 },
		{ 222, 321, 17 },
		{ 222, 322, 17 },
		{ 222, 323, 17 },
		{ 222, 324, 17 },
		{ 222, 325, 17 },
		{ 222, 326, 17 },
		{ 223, 292, 17 },
		{ 223, 293, 17 },
		{ 223, 294, 17 },
		{ 223, 295, 17 },
		{ 223, 296, 17 },
		{ 223, 297, 17 },
		{ 223, 298, 17 },
		{ 223, 299, 17 },
		{ 223, 300, 17 },
		{ 223, 301, 17 },
		{ 223, 302, 17 },
		{ 223, 303, 17 },
		{ 223, 304, 17 },
		{ 223, 305, 17 },
		{ 223, 306, 17 },
		{ 223, 307, 17 },
		{ 223, 308, 17 },
		{ 223, 309, 17 },
		{ 223, 310, 17 },
		{ 223, 311, 17 },
		{ 223, 312, 17 },
		{ 223, 313, 17 },
		{ 223, 314, 17 },
		{ 223, 315, 17 },
		{ 223, 316, 17 },
		{ 223, 317, 17 },
		{ 223, 318, 17 },
		{ 223, 319, 17 },
		{ 223, 320, 17 },
		{ 223, 321, 17 },
		{ 223, 322, 17 },
		{ 223, 323, 17 },
		{ 223, 324, 17 },
		{ 224, 291, 17 },
		{ 224, 292, 17 },
		{ 224, 293, 17 },
		{ 224, 294, 17 },
		{ 224, 295, 17 },
		{ 224, 296, 17 },
		{ 224, 297, 17 },
		{ 224, 298, 17 },
		{ 224, 299, 17 },
		{ 224, 300, 17 },
		{ 224, 301, 17 },
		{ 224, 302, 17 },
		{ 224, 303, 17 },
		{ 224, 304, 17 },
		{ 224, 305, 17 },
		{ 224, 306, 17 },
		{ 224, 307, 17 },
		{ 224, 308, 17 },
		{ 224, 309, 17 },
		{ 224, 310, 17 },
		{ 224, 311, 17 },
		{ 224, 312, 17 },
		{ 224, 313, 17 },
		{ 224, 314, 17 },
		{ 224, 315, 17 },
		{ 224, 316, 17 },
		{ 224, 317, 17 },
		{ 224, 318, 17 },
		{ 224, 319, 17 },
		{ 224, 320, 17 },
		{ 224, 321, 17 },
		{ 224, 322, 17 },
		{ 225, 290, 17 },
		{ 225, 291, 17 },
		{ 225, 292, 17 },
		{ 225, 293, 17 },
		{ 225, 294, 17 },
		{ 225, 295, 17 },
		{ 225, 296, 17 },
		{ 225, 297, 17 },
		{ 225, 298, 17 },
		{ 225, 299, 17 },
		{ 225, 300, 17 },
		{ 225, 301, 17 },
		{ 225, 302, 17 },
		{ 225, 303, 17 },
		{ 225, 304, 17 },
		{ 225, 305, 17 },
		{ 225, 306, 17 },
		{ 225, 307, 17 },
		{ 225, 308, 17 },
		{ 225, 309, 17 },
		{ 225, 310, 17 },
		{ 225, 311, 17 },
		{ 225, 312, 17 },
		{ 225, 313, 17 },
		{ 225, 314, 17 },
		{ 225, 315, 17 },
		{ 225, 316, 17 },
		{ 225, 317, 17 },
		{ 225, 318, 17 },
		{ 225, 319, 17 },
		{ 225, 320, 17 },
		{ 226, 289, 17 },
		{ 226, 290, 17 },
		{ 226, 291, 17 },
		{ 226, 292, 17 },
		{ 226, 293, 17 },
		{ 226, 294, 17 },
		{ 226, 295, 17 },
		{ 226, 296, 17 },
		{ 226, 297, 17 },
		{ 226, 298, 17 },
		{ 226, 299, 17 },
		{ 226, 300, 17 },
		{ 226, 301, 17 },
		{ 226, 302, 17 },
		{ 226, 303, 17 },
		{ 226, 304, 17 },
		{ 226, 305, 17 },
		{ 226, 306, 17 },
		{ 226, 307, 17 },
		{ 226, 308, 17 },
		{ 226, 309, 17 },
		{ 226, 310, 17 },
		{ 226, 311, 17 },
		{ 226, 312, 17 },
		{ 226, 313, 17 },
		{ 226, 314, 17 },
		{ 226, 315, 17 },
		{ 226, 316, 17 },
		{ 226, 317, 17 },
		{ 226, 318, 17 },
		{ 227, 288, 17 },
		{ 227, 289, 17 },
		{ 227, 290, 17 },
		{ 227, 291, 17 },
		{ 227, 292, 17 },
		{ 227, 293, 17 },
		{ 227, 294, 17 },
		{ 227, 295, 17 },
		{ 227, 296, 17 },
		{ 227, 297, 17 },
		{ 227, 298, 17 },
		{ 227, 299, 17 },
		{ 227, 300, 17 },
		{ 227, 301, 17 },
		{ 227, 302, 17 },
		{ 227, 303, 17 },
		{ 227, 304, 17 },
		{ 227, 305, 17 },
		{ 227, 306, 17 },
		{ 227, 307, 17 },
		{ 227, 308, 17 },
		{ 227, 309, 17 },
		{ 227, 310, 17 },
		{ 227, 311, 17 },
		{ 227, 312, 17 },
		{ 227, 313, 17 },
		{ 227, 314, 17 },
		{ 227, 315, 17 },
		{ 227, 316, 17 },
		{ 228, 287, 17 },
		{ 228, 288, 17 },
		{ 228, 289, 17 },
		{ 228, 290, 17 },
		{ 228, 291, 17 },
		{ 228, 292, 17 },
		{ 228, 293, 17 },
		{ 228, 294, 17 },
		{ 228, 295, 17 },
		{ 228, 296, 17 },
		{ 228, 297, 17 },
		{ 228, 298, 17 },
		{ 228, 299, 17 },
		{ 228, 300, 17 },
		{ 228, 301, 17 },
		{ 228, 302, 17 },
		{ 228, 303, 17 },
		{ 228, 304, 17 },
		{ 228, 305, 17 },
		{ 228, 306, 17 },
		{ 228, 307, 17 },
		{ 228, 308, 17 },
		{ 228, 309, 17 },
		{ 228, 310, 17 },
		{ 228, 311, 17 },
		{ 228, 312, 17 },
		{ 228, 313, 17 },
		{ 228, 314, 17 },
		{ 229, 286, 17 },
		{ 229, 287, 17 },
		{ 229, 288, 17 },
		{ 229, 289, 17 },
		{ 229, 290, 17 },
		{ 229, 291, 17 },
		{ 229, 292, 17 },
		{ 229, 293, 17 },
		{ 229, 294, 17 },
		{ 229, 295, 17 },
		{ 229, 296, 17 },
		{ 229, 297, 17 },
		{ 229, 298, 17 },
		{ 229, 299, 17 },
		{ 229, 300, 17 },
		{ 229, 301, 17 },
		{ 229, 302, 17 },
		{ 229, 303, 17 },
		{ 229, 304, 17 },
		{ 229, 305, 17 },
		{ 229, 306, 17 },
		{ 229, 307, 17 },
		{ 229, 308, 17 },
		{ 229, 309, 17 },
		{ 229, 310, 17 },
		{ 229, 311, 17 },
		{ 229, 312, 17 },
		{ 230, 285, 17 },
		{ 230, 286, 17 },
		{ 230, 287, 17 },
		{ 230, 288, 17 },
		{ 230, 289, 17 },
		{ 230, 290, 17 },
		{ 230, 291, 17 },
		{ 230, 292, 17 },
		{ 230, 293, 17 },
		{ 230, 294, 17 },
		{ 230, 295, 17 },
		{ 230, 296, 17 },
		{ 230, 297, 17 },
		{ 230, 298, 17 },
		{ 230, 299, 17 },
		{ 230, 300, 17 },
		{ 230, 301, 17 },
		{ 230, 302, 17 },
		{ 230, 303, 17 },
		{ 230, 304, 17 },
		{ 230, 305, 17 },
		{ 230, 306, 17 },
		{ 230, 307, 17 },
		{ 230, 308, 17 },
		{ 230, 309, 17 },
		{ 230, 310, 17 },
		{ 231, 284, 17 },
		{ 231, 285, 17 },
		{ 231, 286, 17 },
		{ 231, 287, 17 },
		{ 231, 288, 17 },
		{ 231, 289, 17 },
		{ 231, 290, 17 },
		{ 231, 291, 17 },
		{ 231, 292, 17 },
		{ 231, 293, 17 },
		{ 231, 294, 17 },
		{ 231, 295, 17 },
		{ 231, 296, 17 },
		{ 231, 297, 17 },
		{ 231, 298, 17 },
		{ 231, 299, 17 },
		{ 231, 300, 17 },
		{ 231, 301, 17 },
		{ 231, 302, 17 },
		{ 231, 303, 17 },
		{ 231, 304, 17 },
		{ 231, 305, 17 },
		{ 231, 306, 17 },
		{ 231, 307, 17 },
		{ 231, 308, 17 },
		{ 232, 283, 17 },
		{ 232, 284, 17 },
		{ 232, 285, 17 },
		{ 232, 286, 17 },
		{ 232, 287, 17 },
		{ 232, 288, 17 },
		{ 232, 289, 17 },
		{ 232, 290, 17 },
		{ 232, 291, 17 },
		{ 232, 292, 17 },
		{ 232, 293, 17 },
		{ 232, 294, 17 },
		{ 232, 295, 17 },
		{ 232, 296, 17 },
		{ 232, 297, 17 },
		{ 232, 298, 17 },
		{ 232, 299, 17 },
		{ 232, 300, 17 },
		{ 232, 301, 17 },
		{ 232, 302, 17 },
		{ 232, 303, 17 },
		{ 232, 304, 17 },
		{ 232, 305, 17 },
		{ 232, 306, 17 },
		{ 233, 282, 17 },
		{ 233, 283, 17 },
		{ 233, 284, 17 },
		{ 233, 285, 17 },
		{ 233, 286, 17 },
		{ 233, 287, 17 },
		{ 233, 288, 17 },
		{ 233, 289, 17 },
		{ 233, 290, 17 },
		{ 233, 291, 17 },
		{ 233, 292, 17 },
		{ 233, 293, 17 },
		{ 233, 294, 17 },
		{ 233, 295, 17 },
		{ 233, 296, 17 },
		{ 233, 297, 17 },
		{ 233, 298, 17 },
		{ 233, 299, 17 },
		{ 233, 300, 17 },
		{ 233, 301, 17 },
		{ 233, 302, 17 },
		{ 233, 303, 17 },
		{ 233, 304, 17 },
		{ 234, 281, 17 },
		{ 234, 282, 17 },
		{ 234, 283, 17 },
		{ 234, 284, 17 },
		{ 234, 285, 17 },
		{ 234, 286, 17 },
		{ 234, 287, 17 },
		{ 234, 288, 17 },
		{ 234, 289, 17 },
		{ 234, 290, 17 },
		{ 234, 291, 17 },
		{ 234, 292, 17 },
		{ 234, 293, 17 },
		{ 234, 294, 17 },
		{ 234, 295, 17 },
		{ 234, 296, 17 },
		{ 234, 297, 17 },
		{ 234, 298, 17 },
		{ 234, 299, 17 },
		{ 234, 300, 17 },
		{ 234, 301, 17 },
		{ 234, 302, 17 },
		{ 235, 280, 17 },
		{ 235, 281, 17 },
		{ 235, 282, 17 },
		{ 235, 283, 17 },
		{ 235, 284, 17 },
		{ 235, 285, 17 },
		{ 235, 286, 17 },
		{ 235, 287, 17 },
		{ 235, 288, 17 },
		{ 235, 289, 17 },
		{ 235, 290, 17 },
		{ 235, 291, 17 },
		{ 235, 292, 17 },
		{ 235, 293, 17 },
		{ 235, 294, 17 },
		{ 235, 295, 17 },
		{ 235, 296, 17 },
		{ 235, 297, 17 },
		{ 235, 298, 17 },
		{ 235, 299, 17 },
		{ 235, 300, 17 },
		{ 236, 279, 17 },
		{ 236, 280, 17 },
		{ 236, 281, 17 },
		{ 236, 282, 17 },
		{ 236, 283, 17 },
		{ 236, 284, 17 },
		{ 236, 285, 17 },
		{ 236, 286, 17 },
		{ 236, 287, 17 },
		{ 236, 288, 17 },
		{ 236, 289, 17 },
		{ 236, 290, 17 },
		{ 236, 291, 17 },
		{ 236, 292, 17 },
		{ 236, 293, 17 },
		{ 236, 294, 17 },
		{ 236, 295, 17 },
		{ 236, 296, 17 },
		{ 236, 297, 17 },
		{ 236, 298, 17 },
		{ 237, 278, 17 },
		{ 237, 279, 17 },
		{ 237, 280, 17 },
		{ 237, 281, 17 },
		{ 237, 282, 17 },
		{ 237, 283, 17 },
		{ 237, 284, 17 },
		{ 237, 285, 17 },
		{ 237, 286, 17 },
		{ 237, 287, 17 },
		{ 237, 288, 17 },
		{ 237, 289, 17 },
		{ 237, 290, 17 },
		{ 237, 291, 17 },
		{ 237, 292, 17 },
		{ 237, 293, 17 },
		{ 237, 294, 17 },
		{ 237, 295, 17 },
		{ 237, 296, 17 },
		{ 238, 277, 17 },
		{ 238, 278, 17 },
		{ 238, 279, 17 },
		{ 238, 280, 17 },
		{ 238, 281, 17 },
		{ 238, 282, 17 },
		{ 238, 283, 17 },
		{ 238, 284, 17 },
		{ 238, 285, 17 },
		{ 238, 286, 17 },
		{ 238, 287, 17 },
		{ 238, 288, 17 },
		{ 238, 289, 17 },
		{ 238, 290, 17 },
		{ 238, 291, 17 },
		{ 238, 292, 17 },
		{ 238, 293, 17 },
		{ 238, 294, 17 },
		{ 239, 276, 17 },
		{ 239, 277, 17 },
		{ 239, 278, 17 },
		{ 239, 279, 17 },
		{ 239, 280, 17 },
		{ 239, 281, 17 },
		{ 239, 282, 17 },
		{ 239, 283, 17 },
		{ 239, 284, 17 },
		{ 239, 285, 17 },
		{ 239, 286, 17 },
		{ 239, 287, 17 },
		{ 239, 288, 17 },
		{ 239, 289, 17 },
		{ 239, 290, 17 },
		{ 239, 291, 17 },
		{ 239, 292, 17 },
		{ 240, 275, 17 },
		{ 240, 276, 17 },
		{ 240, 277, 17 },
		{ 240, 278, 17 },
		{ 240, 279, 17 },
		{ 240, 280, 17 },
		{ 240, 281, 17 },
		{ 240, 282, 17 },
		{ 240, 283, 17 },
		{ 240, 284, 17 },
		{ 240, 285, 17 },
		{ 240, 286, 17 },
		{ 240, 287, 17 },
		{ 240, 288, 17 },
		{ 240, 289, 17 },
		{ 240, 290, 17 },
		{ 241, 274, 17 },
		{ 241, 275, 17 },
		{ 241, 276, 17 },
		{ 241, 277, 17 },
		{ 241, 278, 17 },
		{ 241, 279, 17 },
		{ 241, 280, 17 },
		{ 241, 281, 17 },
		{ 241, 282, 17 },
		{ 241, 283, 17 },
		{ 241, 284, 17 },
		{ 241, 285, 17 },
		{ 241, 286, 17 },
		{ 241, 287, 17 },
		{ 241, 288, 17 },
		{ 242, 274, 17 },
		{ 242, 275, 17 },
		{ 242, 276, 17 },
		{ 242, 277, 17 },
		{ 242, 278, 17 },
		{ 242, 279, 17 },
		{ 242, 280, 17 },
		{ 242, 281, 17 },
		{ 242, 282, 17 },
		{ 242, 283, 17 },
		{ 242, 284, 17 },
		{ 242, 285, 17 },
		{ 242, 286, 17 },
		{ 243, 274, 17 },
		{ 243, 275, 17 },
		{ 243, 276, 17 },
		{ 243, 277, 17 },
		{ 243, 278, 17 },
		{ 243, 279, 17 },
		{ 243, 280, 17 },
		{ 243, 281, 17 },
		{ 243, 282, 17 },
		{ 243, 283, 17 },
		{ 243, 284, 17 },
		{ 244, 274, 17 },
		{ 244, 275, 17 },
		{ 244, 276, 17 },
		{ 244, 277, 17 },
		{ 244, 278, 17 },
		{ 244, 279, 17 },
		{ 244, 280, 17 },
		{ 244, 281, 17 },
		{ 244, 282, 17 },
		{ 245, 274, 17 },
		{ 245, 275, 17 },
		{ 245, 276, 17 },
		{ 245, 277, 17 },
		{ 245, 278, 17 },
		{ 245, 279, 17 },
		{ 245, 280, 17 },
		{ 246, 274, 17 },
		{ 246, 275, 17 },
		{ 246, 276, 17 },
		{ 246, 277, 17 },
		{ 246, 278, 17 },
		{ 247, 274, 17 },
		{ 247, 275, 17 },
		{ 247, 276, 17 },
		{ 248, 274, 17 },
		{ 266, 240, 17 },
		{ 267, 238, 17 },
		{ 267, 239, 17 },
		{ 267, 240, 17 },
		{ 268, 236, 17 },
		{ 268, 237, 17 },
		{ 268, 238, 17 },
		{ 268, 239, 17 },
		{ 268, 240, 17 },
		{ 269, 234, 17 },
		{ 269, 235, 17 },
		{ 269, 236, 17 },
		{ 269, 237, 17 },
		{ 269, 238, 17 },
		{ 269, 239, 17 },
		{ 269, 240, 17 },
		{ 270, 232, 17 },
		{ 270, 233, 17 },
		{ 270, 234, 17 },
		{ 270, 235, 17 },
		{ 270, 236, 17 },
		{ 270, 237, 17 },
		{ 270, 238, 17 },
		{ 270, 239, 17 },
		{ 270, 240, 17 },
		{ 271, 230, 17 },
		{ 271, 231, 17 },
		{ 271, 232, 17 },
		{ 271, 233, 17 },
		{ 271, 234, 17 },
		{ 271, 235, 17 },
		{ 271, 236, 17 },
		{ 271, 237, 17 },
		{ 271, 238, 17 },
		{ 271, 239, 17 },
		{ 271, 240, 17 },
		{ 272, 228, 17 },
		{ 272, 229, 17 },
		{ 272, 230, 17 },
		{ 272, 231, 17 },
		{ 272, 232, 17 },
		{ 272, 233, 17 },
		{ 272, 234, 17 },
		{ 272, 235, 17 },
		{ 272, 236, 17 },
		{ 272, 237, 17 },
		{ 272, 238, 17 },
		{ 272, 239, 17 },
		{ 272, 240, 17 },
		{ 273, 226, 17 },
		{ 273, 227, 17 },
		{ 273, 228, 17 },
		{ 273, 229, 17 },
		{ 273, 230, 17 },
		{ 273, 231, 17 },
		{ 273, 232, 17 },
		{ 273, 233, 17 },
		{ 273, 234, 17 },
		{ 273, 235, 17 },
		{ 273, 236, 17 },
		{ 273, 237, 17 },
		{ 273, 238, 17 },
		{ 273, 239, 17 },
		{ 273, 240, 17 },
		{ 274, 224, 17 },
		{ 274, 225, 17 },
		{ 274, 226, 17 },
		{ 274, 227, 17 },
		{ 274, 228, 17 },
		{ 274, 229, 17 },
		{ 274, 230, 17 },
		{ 274, 231, 17 },
		{ 274, 232, 17 },
		{ 274, 233, 17 },
		{ 274, 234, 17 },
		{ 274, 235, 17 },
		{ 274, 236, 17 },
		{ 274, 237, 17 },
		{ 274, 238, 17 },
		{ 274, 239, 17 },
		{ 275, 222, 17 },
		{ 275, 223, 17 },
		{ 275, 224, 17 },
		{ 275, 225, 17 },
		{ 275, 226, 17 },
		{ 275, 227, 17 },
		{ 275, 228, 17 },
		{ 275, 229, 17 },
		{ 275, 230, 17 },
		{ 275, 231, 17 },
		{ 275, 232, 17 },
		{ 275, 233, 17 },
		{ 275, 234, 17 },
		{ 275, 235, 17 },
		{ 275, 236, 17 },
		{ 275, 237, 17 },
		{ 275, 238, 17 },
		{ 276, 220, 17 },
		{ 276, 221, 17 },
		{ 276, 222, 17 },
		{ 276, 223, 17 },
		{ 276, 224, 17 },
		{ 276, 225, 17 },
		{ 276, 226, 17 },
		{ 276, 227, 17 },
		{ 276, 228, 17 },
		{ 276, 229, 17 },
		{ 276, 230, 17 },
		{ 276, 231, 17 },
		{ 276, 232, 17 },
		{ 276, 233, 17 },
		{ 276, 234, 17 },
		{ 276, 235, 17 },
		{ 276, 236, 17 },
		{ 276, 237, 17 },
		{ 277, 218, 17 },
		{ 277, 219, 17 },
		{ 277, 220, 17 },
		{ 277, 221, 17 },
		{ 277, 222, 17 },
		{ 277, 223, 17 },
		{ 277, 224, 17 },
		{ 277, 225, 17 },
		{ 277, 226, 17 },
		{ 277, 227, 17 },
		{ 277, 228, 17 },
		{ 277, 229, 17 },
		{ 277, 230, 17 },
		{ 277, 231, 17 },
		{ 277, 232, 17 },
		{ 277, 233, 17 },
		{ 277, 234, 17 },
		{ 277, 235, 17 },
		{ 277, 236, 17 },
		{ 278, 216, 17 },
		{ 278, 217, 17 },
		{ 278, 218, 17 },
		{ 278, 219, 17 },
		{ 278, 220, 17 },
		{ 278, 221, 17 },
		{ 278, 222, 17 },
		{ 278, 223, 17 },
		{ 278, 224, 17 },
		{ 278, 225, 17 },
		{ 278, 226, 17 },
		{ 278, 227, 17 },
		{ 278, 228, 17 },
		{ 278, 229, 17 },
		{ 278, 230, 17 },
		{ 278, 231, 17 },
		{ 278, 232, 17 },
		{ 278, 233, 17 },
		{ 278, 234, 17 },
		{ 278, 235, 17 },
		{ 279, 214, 17 },
		{ 279, 215, 17 },
		{ 279, 216, 17 },
		{ 279, 217, 17 },
		{ 279, 218, 17 },
		{ 279, 219, 17 },
		{ 279, 220, 17 },
		{ 279, 221, 17 },
		{ 279, 222, 17 },
		{ 279, 223, 17 },
		{ 279, 224, 17 },
		{ 279, 225, 17 },
		{ 279, 226, 17 },
		{ 279, 227, 17 },
		{ 279, 228, 17 },
		{ 279, 229, 17 },
		{ 279, 230, 17 },
		{ 279, 231, 17 },
		{ 279, 232, 17 },
		{ 279, 233, 17 },
		{ 279, 234, 17 },
		{ 280, 212, 17 },
		{ 280, 213, 17 },
		{ 280, 214, 17 },
		{ 280, 215, 17 },
		{ 280, 216, 17 },
		{ 280, 217, 17 },
		{ 280, 218, 17 },
		{ 280, 219, 17 },
		{ 280, 220, 17 },
		{ 280, 221, 17 },
		{ 280, 222, 17 },
		{ 280, 223, 17 },
		{ 280, 224, 17 },
		{ 280, 225, 17 },
		{ 280, 226, 17 },
		{ 280, 227, 17 },
		{ 280, 228, 17 },
		{ 280, 229, 17 },
		{ 280, 230, 17 },
		{ 280, 231, 17 },
		{ 280, 232, 17 },
		{ 280, 233, 17 },
		{ 281, 210, 17 },
		{ 281, 211, 17 },
		{ 281, 212, 17 },
		{ 281, 213, 17 },
		{ 281, 214, 17 },
		{ 281, 215, 17 },
		{ 281, 216, 17 },
		{ 281, 217, 17 },
		{ 281, 218, 17 },
		{ 281, 219, 17 },
		{ 281, 220, 17 },
		{ 281, 221, 17 },
		{ 281, 222, 17 },
		{ 281, 223, 17 },
		{ 281, 224, 17 },
		{ 281, 225, 17 },
		{ 281, 226, 17 },
		{ 281, 227, 17 },
		{ 281, 228, 17 },
		{ 281, 229, 17 },
		{ 281, 230, 17 },
		{ 281, 231, 17 },
		{ 281, 232, 17 },
		{ 282, 208, 17 },
		{ 282, 209, 17 },
		{ 282, 210, 17 },
		{ 282, 211, 17 },
		{ 282, 212, 17 },
		{ 282, 213, 17 },
		{ 282, 214, 17 },
		{ 282, 215, 17 },
		{ 282, 216, 17 },
		{ 282, 217, 17 },
		{ 282, 218, 17 },
		{ 282, 219, 17 },
		{ 282, 220, 17 },
		{ 282, 221, 17 },
		{ 282, 222, 17 },
		{ 282, 223, 17 },
		{ 282, 224, 17 },
		{ 282, 225, 17 },
		{ 282, 226, 17 },
		{ 282, 227, 17 },
		{ 282, 228, 17 },
		{ 282, 229, 17 },
		{ 282, 230, 17 },
		{ 282, 231, 17 },
		{ 283, 206, 17 },
		{ 283, 207, 17 },
		{ 283, 208, 17 },
		{ 283, 209, 17 },
		{ 283, 210, 17 },
		{ 283, 211, 17 },
		{ 283, 212, 17 },
		{ 283, 213, 17 },
		{ 283, 214, 17 },
		{ 283, 215, 17 },
		{ 283, 216, 17 },
		{ 283, 217, 17 },
		{ 283, 218, 17 },
		{ 283, 219, 17 },
		{ 283, 220, 17 },
		{ 283, 221, 17 },
		{ 283, 222, 17 },
		{ 283, 223, 17 },
		{ 283, 224, 17 },
		{ 283, 225, 17 },
		{ 283, 226, 17 },
		{ 283, 227, 17 },
		{ 283, 228, 17 },
		{ 283, 229, 17 },
		{ 283, 230, 17 },
		{ 284, 204, 17 },
		{ 284, 205, 17 },
		{ 284, 206, 17 },
		{ 284, 207, 17 },
		{ 284, 208, 17 },
		{ 284, 209, 17 },
		{ 284, 210, 17 },
		{ 284, 211, 17 },
		{ 284, 212, 17 },
		{ 284, 213, 17 },
		{ 284, 214, 17 },
		{ 284, 215, 17 },
		{ 284, 216, 17 },
		{ 284, 217, 17 },
		{ 284, 218, 17 },
		{ 284, 219, 17 },
		{ 284, 220, 17 },
		{ 284, 221, 17 },
		{ 284, 222, 17 },
		{ 284, 223, 17 },
		{ 284, 224, 17 },
		{ 284, 225, 17 },
		{ 284, 226, 17 },
		{ 284, 227, 17 },
		{ 284, 228, 17 },
		{ 284, 229, 17 },
		{ 285, 202, 17 },
		{ 285, 203, 17 },
		{ 285, 204, 17 },
		{ 285, 205, 17 },
		{ 285, 206, 17 },
		{ 285, 207, 17 },
		{ 285, 208, 17 },
		{ 285, 209, 17 },
		{ 285, 210, 17 },
		{ 285, 211, 17 },
		{ 285, 212, 17 },
		{ 285, 213, 17 },
		{ 285, 214, 17 },
		{ 285, 215, 17 },
		{ 285, 216, 17 },
		{ 285, 217, 17 },
		{ 285, 218, 17 },
		{ 285, 219, 17 },
		{ 285, 220, 17 },
		{ 285, 221, 17 },
		{ 285, 222, 17 },
		{ 285, 223, 17 },
		{ 285, 224, 17 },
		{ 285, 225, 17 },
		{ 285, 226, 17 },
		{ 285, 227, 17 },
		{ 285, 228, 17 },
		{ 286, 200, 17 },
		{ 286, 201, 17 },
		{ 286, 202, 17 },
		{ 286, 203, 17 },
		{ 286, 204, 17 },
		{ 286, 205, 17 },
		{ 286, 206, 17 },
		{ 286, 207, 17 },
		{ 286, 208, 17 },
		{ 286, 209, 17 },
		{ 286, 210, 17 },
		{ 286, 211, 17 },
		{ 286, 212, 17 },
		{ 286, 213, 17 },
		{ 286, 214, 17 },
		{ 286, 215, 17 },
		{ 286, 216, 17 },
		{ 286, 217, 17 },
		{ 286, 218, 17 },
		{ 286, 219, 17 },
		{ 286, 220, 17 },
		{ 286, 221, 17 },
		{ 286, 222, 17 },
		{ 286, 223, 17 },
		{ 286, 224, 17 },
		{ 286, 225, 17 },
		{ 286, 226, 17 },
		{ 286, 227, 17 },
		{ 287, 198, 17 },
		{ 287, 199, 17 },
		{ 287, 200, 17 },
		{ 287, 201, 17 },
		{ 287, 202, 17 },
		{ 287, 203, 17 },
		{ 287, 204, 17 },
		{ 287, 205, 17 },
		{ 287, 206, 17 },
		{ 287, 207, 17 },
		{ 287, 208, 17 },
		{ 287, 209, 17 },
		{ 287, 210, 17 },
		{ 287, 211, 17 },
		{ 287, 212, 17 },
		{ 287, 213, 17 },
		{ 287, 214, 17 },
		{ 287, 215, 17 },
		{ 287, 216, 17 },
		{ 287, 217, 17 },
		{ 287, 218, 17 },
		{ 287, 219, 17 },
		{ 287, 220, 17 },
		{ 287, 221, 17 },
		{ 287, 222, 17 },
		{ 287, 223, 17 },
		{ 287, 224, 17 },
		{ 287, 225, 17 },
		{ 287, 226, 17 },
		{ 288, 196, 17 },
		{ 288, 197, 17 },
		{ 288, 198, 17 },
		{ 288, 199, 17 },
		{ 288, 200, 17 },
		{ 288, 201, 17 },
		{ 288, 202, 17 },
		{ 288, 203, 17 },
		{ 288, 204, 17 },
		{ 288, 205, 17 },
		{ 288, 206, 17 },
		{ 288, 207, 17 },
		{ 288, 208, 17 },
		{ 288, 209, 17 },
		{ 288, 210, 17 },
		{ 288, 211, 17 },
		{ 288, 212, 17 },
		{ 288, 213, 17 },
		{ 288, 214, 17 },
		{ 288, 215, 17 },
		{ 288, 216, 17 },
		{ 288, 217, 17 },
		{ 288, 218, 17 },
		{ 288, 219, 17 },
		{ 288, 220, 17 },
		{ 288, 221, 17 },
		{ 288, 222, 17 },
		{ 288, 223, 17 },
		{ 288, 224, 17 },
		{ 288, 225, 17 },
		{ 289, 194, 17 },
		{ 289, 195, 17 },
		{ 289, 196, 17 },
		{ 289, 197, 17 },
		{ 289, 198, 17 },
		{ 289, 199, 17 },
		{ 289, 200, 17 },
		{ 289, 201, 17 },
		{ 289, 202, 17 },
		{ 289, 203, 17 },
		{ 289, 204, 17 },
		{ 289, 205, 17 },
		{ 289, 206, 17 },
		{ 289, 207, 17 },
		{ 289, 208, 17 },
		{ 289, 209, 17 },
		{ 289, 210, 17 },
		{ 289, 211, 17 },
		{ 289, 212, 17 },
		{ 289, 213, 17 },
		{ 289, 214, 17 },
		{ 289, 215, 17 },
		{ 289, 216, 17 },
		{ 289, 217, 17 },
		{ 289, 218, 17 },
		{ 289, 219, 17 },
		{ 289, 220, 17 },
		{ 289, 221, 17 },
		{ 289, 222, 17 },
		{ 289, 223, 17 },
		{ 289, 224, 17 },
		{ 290, 192, 17 },
		{ 290, 193, 17 },
		{ 290, 194, 17 },
		{ 290, 195, 17 },
		{ 290, 196, 17 },
		{ 290, 197, 17 },
		{ 290, 198, 17 },
		{ 290, 199, 17 },
		{ 290, 200, 17 },
		{ 290, 201, 17 },
		{ 290, 202, 17 },
		{ 290, 203, 17 },
		{ 290, 204, 17 },
		{ 290, 205, 17 },
		{ 290, 206, 17 },
		{ 290, 207, 17 },
		{ 290, 208, 17 },
		{ 290, 209, 17 },
		{ 290, 210, 17 },
		{ 290, 211, 17 },
		{ 290, 212, 17 },
		{ 290, 213, 17 },
		{ 290, 214, 17 },
		{ 290, 215, 17 },
		{ 290, 216, 17 },
		{ 290, 217, 17 },
		{ 290, 218, 17 },
		{ 290, 219, 17 },
		{ 290, 220, 17 },
		{ 290, 221, 17 },
		{ 290, 222, 17 },
		{ 290, 223, 17 },
		{ 291, 190, 17 },
		{ 291, 191, 17 },
		{ 291, 192, 17 },
		{ 291, 193, 17 },
		{ 291, 194, 17 },
		{ 291, 195, 17 },
		{ 291, 196, 17 },
		{ 291, 197, 17 },
		{ 291, 198, 17 },
		{ 291, 199, 17 },
		{ 291, 200, 17 },
		{ 291, 201, 17 },
		{ 291, 202, 17 },
		{ 291, 203, 17 },
		{ 291, 204, 17 },
		{ 291, 205, 17 },
		{ 291, 206, 17 },
		{ 291, 207, 17 },
		{ 291, 208, 17 },
		{ 291, 209, 17 },
		{ 291, 210, 17 },
		{ 291, 211, 17 },
		{ 291, 212, 17 },
		{ 291, 213, 17 },
		{ 291, 214, 17 },
		{ 291, 215, 17 },
		{ 291, 216, 17 },
		{ 291, 217, 17 },
		{ 291, 218, 17 },
		{ 291, 219, 17 },
		{ 291, 220, 17 },
		{ 291, 221, 17 },
		{ 291, 222, 17 },
		{ 292, 188, 17 },
		{ 292, 189, 17 },
		{ 292, 190, 17 },
		{ 292, 191, 17 },
		{ 292, 192, 17 },
		{ 292, 193, 17 },
		{ 292, 194, 17 },
		{ 292, 195, 17 },
		{ 292, 196, 17 },
		{ 292, 197, 17 },
		{ 292, 198, 17 },
		{ 292, 199, 17 },
		{ 292, 200, 17 },
		{ 292, 201, 17 },
		{ 292, 202, 17 },
		{ 292, 203, 17 },
		{ 292, 204, 17 },
		{ 292, 205, 17 },
		{ 292, 206, 17 },
		{ 292, 207, 17 },
		{ 292, 208, 17 },
		{ 292, 209, 17 },
		{ 292, 210, 17 },
		{ 292, 211, 17 },
		{ 292, 212, 17 },
		{ 292, 213, 17 },
		{ 292, 214, 17 },
		{ 292, 215, 17 },
		{ 292, 216, 17 },
		{ 292, 217, 17 },
		{ 292, 218, 17 },
		{ 292, 219, 17 },
		{ 292, 220, 17 },
		{ 292, 221, 17 },
		{ 293, 186, 17 },
		{ 293, 187, 17 },
		{ 293, 188, 17 },
		{ 293, 189, 17 },
		{ 293, 190, 17 },
		{ 293, 191, 17 },
		{ 293, 192, 17 },
		{ 293, 193, 17 },
		{ 293, 194, 17 },
		{ 293, 195, 17 },
		{ 293, 196, 17 },
		{ 293, 197, 17 },
		{ 293, 198, 17 },
		{ 293, 199, 17 },
		{ 293, 200, 17 },
		{ 293, 201, 17 },
		{ 293, 202, 17 },
		{ 293, 203, 17 },
		{ 293, 204, 17 },
		{ 293, 205, 17 },
		{ 293, 206, 17 },
		{ 293, 207, 17 },
		{ 293, 208, 17 },
		{ 293, 209, 17 },
		{ 293, 210, 17 },
		{ 293, 211, 17 },
		{ 293, 212, 17 },
		{ 293, 213, 17 },
		{ 293, 214, 17 },
		{ 293, 215, 17 },
		{ 293, 216, 17 },
		{ 293, 217, 17 },
		{ 293, 218, 17 },
		{ 293, 219, 17 },
		{ 293, 220, 17 },
		{ 294, 184, 17 },
		{ 294, 185, 17 },
		{ 294, 186, 17 },
		{ 294, 187, 17 },
		{ 294, 188, 17 },
		{ 294, 189, 17 },
		{ 294, 190, 17 },
		{ 294, 191, 17 },
		{ 294, 192, 17 },
		{ 294, 193, 17 },
		{ 294, 194, 17 },
		{ 294, 195, 17 },
		{ 294, 196, 17 },
		{ 294, 197, 17 },
		{ 294, 198, 17 },
		{ 294, 199, 17 },
		{ 294, 200, 17 },
		{ 294, 201, 17 },
		{ 294, 202, 17 },
		{ 294, 203, 17 },
		{ 294, 204, 17 },
		{ 294, 205, 17 },
		{ 294, 206, 17 },
		{ 294, 207, 17 },
		{ 294, 208, 17 },
		{ 294, 209, 17 },
		{ 294, 210, 17 },
		{ 294, 211, 17 },
		{ 294, 212, 17 },
		{ 294, 213, 17 },
		{ 294, 214, 17 },
		{ 294, 215, 17 },
		{ 294, 216, 17 },
		{ 294, 217, 17 },
		{ 294, 218, 17 },
		{ 294, 219, 17 },
		{ 295, 182, 17 },
		{ 295, 183, 17 },
		{ 295, 184, 17 },
		{ 295, 185, 17 },
		{ 295, 186, 17 },
		{ 295, 187, 17 },
		{ 295, 188, 17 },
		{ 295, 189, 17 },
		{ 295, 190, 17 },
		{ 295, 191, 17 },
		{ 295, 192, 17 },
		{ 295, 193, 17 },
		{ 295, 194, 17 },
		{ 295, 195, 17 },
		{ 295, 196, 17 },
		{ 295, 197, 17 },
		{ 295, 198, 17 },
		{ 295, 199, 17 },
		{ 295, 200, 17 },
		{ 295, 201, 17 },
		{ 295, 202, 17 },
		{ 295, 203, 17 },
		{ 295, 204, 17 },
		{ 295, 205, 17 },
		{ 295, 206, 17 },
		{ 295, 207, 17 },
		{ 295, 208, 17 },
		{ 295, 209, 17 },
		{ 295, 210, 17 },
		{ 295, 211, 17 },
		{ 295, 212, 17 },
		{ 295, 213, 17 },
		{ 295, 214, 17 },
		{ 295, 215, 17 },
		{ 295, 216, 17 },
		{ 295, 217, 17 },
		{ 295, 218, 17 },
		{ 296, 180, 17 },
		{ 296, 181, 17 },
		{ 296, 182, 17 },
		{ 296, 183, 17 },
		{ 296, 184, 17 },
		{ 296, 185, 17 },
		{ 296, 186, 17 },
		{ 296, 187, 17 },
		{ 296, 188, 17 },
		{ 296, 189, 17 },
		{ 296, 190, 17 },
		{ 296, 191, 17 },
		{ 296, 192, 17 },
		{ 296, 193, 17 },
		{ 296, 194, 17 },
		{ 296, 195, 17 },
		{ 296, 196, 17 },
		{ 296, 197, 17 },
		{ 296, 198, 17 },
		{ 296, 199, 17 },
		{ 296, 200, 17 },
		{ 296, 201, 17 },
		{ 296, 202, 17 },
		{ 296, 203, 17 },
		{ 296, 204, 17 },
		{ 296, 205, 17 },
		{ 296, 206, 17 },
		{ 296, 207, 17 },
		{ 296, 208, 17 },
		{ 296, 209, 17 },
		{ 296, 210, 17 },
		{ 296, 211, 17 },
		{ 296, 212, 17 },
		{ 296, 213, 17 },
		{ 296, 214, 17 },
		{ 296, 215, 17 },
		{ 296, 216, 17 },
		{ 296, 217, 17 },
		{ 297, 178, 17 },
		{ 297, 179, 17 },
		{ 297, 180, 17 },
		{ 297, 181, 17 },
		{ 297, 182, 17 },
		{ 297, 183, 17 },
		{ 297, 184, 17 },
		{ 297, 185, 17 },
		{ 297, 186, 17 },
		{ 297, 187, 17 },
		{ 297, 188, 17 },
		{ 297, 189, 17 },
		{ 297, 190, 17 },
		{ 297, 191, 17 },
		{ 297, 192, 17 },
		{ 297, 193, 17 },
		{ 297, 194, 17 },
		{ 297, 195, 17 },
		{ 297, 196, 17 },
		{ 297, 197, 17 },
		{ 297, 198, 17 },
		{ 297, 199, 17 },
		{ 297, 200, 17 },
		{ 297, 201, 17 },
		{ 297, 202, 17 },
		{ 297, 203, 17 },
		{ 297, 204, 17 },
		{ 297, 205, 17 },
		{ 297, 206, 17 },
		{ 297, 207, 17 },
		{ 297, 208, 17 },
		{ 297, 209, 17 },
		{ 297, 210, 17 },
		{ 297, 211, 17 },
		{ 297, 212, 17 },
		{ 297, 213, 17 },
		{ 297, 214, 17 },
		{ 297, 215, 17 },
		{ 297, 216, 17 },
		{ 298, 176, 17 },
		{ 298, 177, 17 },
		{ 298, 178, 17 },
		{ 298, 179, 17 },
		{ 298, 180, 17 },
		{ 298, 181, 17 },
		{ 298, 182, 17 },
		{ 298, 183, 17 },
		{ 298, 184, 17 },
		{ 298, 185, 17 },
		{ 298, 186, 17 },
		{ 298, 187, 17 },
		{ 298, 188, 17 },
		{ 298, 189, 17 },
		{ 298, 190, 17 },
		{ 298, 191, 17 },
		{ 298, 192, 17 },
		{ 298, 193, 17 },
		{ 298, 194, 17 },
		{ 298, 195, 17 },
		{ 298, 196, 17 },
		{ 298, 197, 17 },
		{ 298, 198, 17 },
		{ 298, 199, 17 },
		{ 298, 200, 17 },
		{ 298, 201, 17 },
		{ 298, 202, 17 },
		{ 298, 203, 17 },
		{ 298, 204, 17 },
		{ 298, 205, 17 },
		{ 298, 206, 17 },
		{ 298, 207, 17 },
		{ 298, 208, 17 },
		{ 298, 209, 17 },
		{ 298, 210, 17 },
		{ 298, 211, 17 },
		{ 298, 212, 17 },
		{ 298, 213, 17 },
		{ 298, 214, 17 },
		{ 298, 215, 17 },
		{ 299, 174, 17 },
		{ 299, 175, 17 },
		{ 299, 176, 17 },
		{ 299, 177, 17 },
		{ 299, 178, 17 },
		{ 299, 179, 17 },
		{ 299, 180, 17 },
		{ 299, 181, 17 },
		{ 299, 182, 17 },
		{ 299, 183, 17 },
		{ 299, 184, 17 },
		{ 299, 185, 17 },
		{ 299, 186, 17 },
		{ 299, 187, 17 },
		{ 299, 188, 17 },
		{ 299, 189, 17 },
		{ 299, 190, 17 },
		{ 299, 191, 17 },
		{ 299, 192, 17 },
		{ 299, 193, 17 },
		{ 299, 194, 17 },
		{ 299, 195, 17 },
		{ 299, 196, 17 },
		{ 299, 197, 17 },
		{ 299, 198, 17 },
		{ 299, 199, 17 },
		{ 299, 200, 17 },
		{ 299, 201, 17 },
		{ 299, 202, 17 },
		{ 299, 203, 17 },
		{ 299, 204, 17 },
		{ 299, 205, 17 },
		{ 299, 206, 17 },
		{ 299, 207, 17 },
		{ 299, 208, 17 },
		{ 299, 209, 17 },
		{ 299, 210, 17 },
		{ 299, 211, 17 },
		{ 299, 212, 17 },
		{ 299, 213, 17 },
		{ 299, 214, 17 },
		{ 300, 172, 17 },
		{ 300, 173, 17 },
		{ 300, 174, 17 },
		{ 300, 175, 17 },
		{ 300, 176, 17 },
		{ 300, 177, 17 },
		{ 300, 178, 17 },
		{ 300, 179, 17 },
		{ 300, 180, 17 },
		{ 300, 181, 17 },
		{ 300, 182, 17 },
		{ 300, 183, 17 },
		{ 300, 184, 17 },
		{ 300, 185, 17 },
		{ 300, 186, 17 },
		{ 300, 187, 17 },
		{ 300, 188, 17 },
		{ 300, 189, 17 },
		{ 300, 190, 17 },
		{ 300, 191, 17 },
		{ 300, 192, 17 },
		{ 300, 193, 17 },
		{ 300, 194, 17 },
		{ 300, 195, 17 },
		{ 300, 196, 17 },
		{ 300, 197, 17 },
		{ 300, 198, 17 },
		{ 300, 199, 17 },
		{ 300, 200, 17 },
		{ 300, 201, 17 },
		{ 300, 202, 17 },
		{ 300, 203, 17 },
		{ 300, 204, 17 },
		{ 300, 205, 17 },
		{ 300, 206, 17 },
		{ 300, 207, 17 },
		{ 300, 208, 17 },
		{ 300, 209, 17 },
		{ 300, 210, 17 },
		{ 300, 211, 17 },
		{ 300, 212, 17 },
		{ 300, 213, 17 },
		{ 301, 170, 17 },
		{ 301, 171, 17 },
		{ 301, 172, 17 },
		{ 301, 173, 17 },
		{ 301, 174, 17 },
		{ 301, 175, 17 },
		{ 301, 176, 17 },
		{ 301, 177, 17 },
		{ 301, 178, 17 },
		{ 301, 179, 17 },
		{ 301, 180, 17 },
		{ 301, 181, 17 },
		{ 301, 182, 17 },
		{ 301, 183, 17 },
		{ 301, 184, 17 },
		{ 301, 185, 17 },
		{ 301, 186, 17 },
		{ 301, 187, 17 },
		{ 301, 188, 17 },
		{ 301, 189, 17 },
		{ 301, 190, 17 },
		{ 301, 191, 17 },
		{ 301, 192, 17 },
		{ 301, 193, 17 },
		{ 301, 194, 17 },
		{ 301, 195, 17 },
		{ 301, 196, 17 },
		{ 301, 197, 17 },
		{ 301, 198, 17 },
		{ 301, 199, 17 },
		{ 301, 200, 17 },
		{ 301, 201, 17 },
		{ 301, 202, 17 },
		{ 301, 203, 17 },
		{ 301, 204, 17 },
		{ 301, 205, 17 },
		{ 301, 206, 17 },
		{ 301, 207, 17 },
		{ 301, 208, 17 },
		{ 301, 209, 17 },
		{ 301, 210, 17 },
		{ 301, 211, 17 },
		{ 301, 212, 17 },
		{ 302, 168, 17 },
		{ 302, 169, 17 },
		{ 302, 170, 17 },
		{ 302, 171, 17 },
		{ 302, 172, 17 },
		{ 302, 173, 17 },
		{ 302, 174, 17 },
		{ 302, 175, 17 },
		{ 302, 176, 17 },
		{ 302, 177, 17 },
		{ 302, 178, 17 },
		{ 302, 179, 17 },
		{ 302, 180, 17 },
		{ 302, 181, 17 },
		{ 302, 182, 17 },
		{ 302, 183, 17 },
		{ 302, 184, 17 },
		{ 302, 185, 17 },
		{ 302, 186, 17 },
		{ 302, 187, 17 },
		{ 302, 188, 17 },
		{ 302, 189, 17 },
		{ 302, 190, 17 },
		{ 302, 191, 17 },
		{ 302, 192, 17 },
		{ 302, 193, 17 },
		{ 302, 194, 17 },
		{ 302, 195, 17 },
		{ 302, 196, 17 },
		{ 302, 197, 17 },
		{ 302, 198, 17 },
		{ 302, 199, 17 },
		{ 302, 200, 17 },
		{ 302, 201, 17 },
		{ 302, 202, 17 },
		{ 302, 203, 17 },
		{ 302, 204, 17 },
		{ 302, 205, 17 },
		{ 302, 206, 17 },
		{ 302, 207, 17 },
		{ 302, 208, 17 },
		{ 302, 209, 17 },
		{ 302, 210, 17 },
		{ 302, 211, 17 },
		{ 303, 166, 17 },
		{ 303, 167, 17 },
		{ 303, 168, 17 },
		{ 303, 169, 17 },
		{ 303, 170, 17 },
		{ 303, 171, 17 },
		{ 303, 172, 17 },
		{ 303, 173, 17 },
		{ 303, 174, 17 },
		{ 303, 175, 17 },
		{ 303, 176, 17 },
		{ 303, 177, 17 },
		{ 303, 178, 17 },
		{ 303, 179, 17 },
		{ 303, 180, 17 },
		{ 303, 181, 17 },
		{ 303, 182, 17 },
		{ 303, 183, 17 },
		{ 303, 184, 17 },
		{ 303, 185, 17 },
		{ 303, 186, 17 },
		{ 303, 187, 17 },
		{ 303, 188, 17 },
		{ 303, 189, 17 },
		{ 303, 190, 17 },
		{ 303, 191, 17 },
		{ 303, 192, 17 },
		{ 303, 193, 17 },
		{ 303, 194, 17 },
		{ 303, 195, 17 },
		{ 303, 196, 17 },
		{ 303, 197, 17 },
		{ 303, 198, 17 },
		{ 303, 199, 17 },
		{ 303, 200, 17 },
		{ 303, 201, 17 },
		{ 303, 202, 17 },
		{ 303, 203, 17 },
		{ 303, 204, 17 },
		{ 303, 205, 17 },
		{ 303, 206, 17 },
		{ 303, 207, 17 },
		{ 303, 208, 17 },
		{ 303, 209, 17 },
		{ 303, 210, 17 },
		{ 304, 164, 17 },
		{ 304, 165, 17 },
		{ 304, 166, 17 },
		{ 304, 167, 17 },
		{ 304, 168, 17 },
		{ 304, 169, 17 },
		{ 304, 170, 17 },
		{ 304, 171, 17 },
		{ 304, 172, 17 },
		{ 304, 173, 17 },
		{ 304, 174, 17 },
		{ 304, 175, 17 },
		{ 304, 176, 17 },
		{ 304, 177, 17 },
		{ 304, 178, 17 },
		{ 304, 179, 17 },
		{ 304, 180, 17 },
		{ 304, 181, 17 },
		{ 304, 182, 17 },
		{ 304, 183, 17 },
		{ 304, 184, 17 },
		{ 304, 185, 17 },
		{ 304, 186, 17 },
		{ 304, 187, 17 },
		{ 304, 188, 17 },
		{ 304, 189, 17 },
		{ 304, 190, 17 },
		{ 304, 191, 17 },
		{ 304, 192, 17 },
		{ 304, 193, 17 },
		{ 304, 194, 17 },
		{ 304, 195, 17 },
		{ 304, 196, 17 },
		{ 304, 197, 17 },
		{ 304, 198, 17 },
		{ 304, 199, 17 },
		{ 304, 200, 17 },
		{ 304, 201, 17 },
		{ 304, 202, 17 },
		{ 304, 203, 17 },
		{ 304, 204, 17 },
		{ 304, 205, 17 },
		{ 304, 206, 17 },
		{ 304, 207, 17 },
		{ 304, 208, 17 },
		{ 304, 209, 17 },
		{ 305, 162, 17 },
		{ 305, 163, 17 },
		{ 305, 164, 17 },
		{ 305, 165, 17 },
		{ 305, 166, 17 },
		{ 305, 167, 17 },
		{ 305, 168, 17 },
		{ 305, 169, 17 },
		{ 305, 170, 17 },
		{ 305, 171, 17 },
		{ 305, 172, 17 },
		{ 305, 173, 17 },
		{ 305, 174, 17 },
		{ 305, 175, 17 },
		{ 305, 176, 17 },
		{ 305, 177, 17 },
		{ 305, 178, 17 },
		{ 305, 179, 17 },
		{ 305, 180, 17 },
		{ 305, 181, 17 },
		{ 305, 182, 17 },
		{ 305, 183, 17 },
		{ 305, 184, 17 },
		{ 305, 185, 17 },
		{ 305, 186, 17 },
		{ 305, 187, 17 },
		{ 305, 188, 17 },
		{ 305, 189, 17 },
		{ 305, 190, 17 },
		{ 305, 191, 17 },
		{ 305, 192, 17 },
		{ 305, 193, 17 },
		{ 305, 194, 17 },
		{ 305, 195, 17 },
		{ 305, 196, 17 },
		{ 305, 197, 17 },
		{ 305, 198, 17 },
		{ 305, 199, 17 },
		{ 305, 200, 17 },
		{ 305, 201, 17 },
		{ 305, 202, 17 },
		{ 305, 203, 17 },
		{ 305, 204, 17 },
		{ 305, 205, 17 },
		{ 305, 206, 17 },
		{ 305, 207, 17 },
		{ 305, 208, 17 },
		{ 306, 160, 17 },
		{ 306, 161, 17 },
		{ 306, 162, 17 },
		{ 306, 163, 17 },
		{ 306, 164, 17 },
		{ 306, 165, 17 },
		{ 306, 166, 17 },
		{ 306, 167, 17 },
		{ 306, 168, 17 },
		{ 306, 169, 17 },
		{ 306, 170, 17 },
		{ 306, 171, 17 },
		{ 306, 172, 17 },
		{ 306, 173, 17 },
		{ 306, 174, 17 },
		{ 306, 175, 17 },
		{ 306, 176, 17 },
		{ 306, 177, 17 },
		{ 306, 178, 17 },
		{ 306, 179, 17 },
		{ 306, 180, 17 },
		{ 306, 181, 17 },
		{ 306, 182, 17 },
		{ 306, 183, 17 },
		{ 306, 184, 17 },
		{ 306, 185, 17 },
		{ 306, 186, 17 },
		{ 306, 187, 17 },
		{ 306, 188, 17 },
		{ 306, 189, 17 },
		{ 306, 190, 17 },
		{ 306, 191, 17 },
		{ 306, 192, 17 },
		{ 306, 193, 17 },
		{ 306, 194, 17 },
		{ 306, 195, 17 },
		{ 306, 196, 17 },
		{ 306, 197, 17 },
		{ 306, 198, 17 },
		{ 306, 199, 17 },
		{ 306, 200, 17 },
		{ 306, 201, 17 },
		{ 306, 202, 17 },
		{ 306, 203, 17 },
		{ 306, 204, 17 },
		{ 306, 205, 17 },
		{ 306, 206, 17 },
		{ 306, 207, 17 },
		{ 307, 158, 17 },
		{ 307, 159, 17 },
		{ 307, 160, 17 },
		{ 307, 161, 17 },
		{ 307, 162, 17 },
		{ 307, 163, 17 },
		{ 307, 164, 17 },
		{ 307, 165, 17 },
		{ 307, 166, 17 },
		{ 307, 167, 17 },
		{ 307, 168, 17 },
		{ 307, 169, 17 },
		{ 307, 170, 17 },
		{ 307, 171, 17 },
		{ 307, 172, 17 },
		{ 307, 173, 17 },
		{ 307, 174, 17 },
		{ 307, 175, 17 },
		{ 307, 176, 17 },
		{ 307, 177, 17 },
		{ 307, 178, 17 },
		{ 307, 179, 17 },
		{ 307, 180, 17 },
		{ 307, 181, 17 },
		{ 307, 182, 17 },
		{ 307, 183, 17 },
		{ 307, 184, 17 },
		{ 307, 185, 17 },
		{ 307, 186, 17 },
		{ 307, 187, 17 },
		{ 307, 188, 17 },
		{ 307, 189, 17 },
		{ 307, 190, 17 },
		{ 307, 191, 17 },
		{ 307, 192, 17 },
		{ 307, 193, 17 },
		{ 307, 194, 17 },
		{ 307, 195, 17 },
		{ 307, 196, 17 },
		{ 307, 197, 17 },
		{ 307, 198, 17 },
		{ 307, 199, 17 },
		{ 307, 200, 17 },
		{ 307, 201, 17 },
		{ 307, 202, 17 },
		{ 307, 203, 17 },
		{ 307, 204, 17 },
		{ 307, 205, 17 },
		{ 307, 206, 17 },
		{ 308, 156, 17 },
		{ 308, 157, 17 },
		{ 308, 158, 17 },
		{ 308, 159, 17 },
		{ 308, 160, 17 },
		{ 308, 161, 17 },
		{ 308, 162, 17 },
		{ 308, 163, 17 },
		{ 308, 164, 17 },
		{ 308, 165, 17 },
		{ 308, 166, 17 },
		{ 308, 167, 17 },
		{ 308, 168, 17 },
		{ 308, 169, 17 },
		{ 308, 170, 17 },
		{ 308, 171, 17 },
		{ 308, 172, 17 },
		{ 308, 173, 17 },
		{ 308, 174, 17 },
		{ 308, 175, 17 },
		{ 308, 176, 17 },
		{ 308, 177, 17 },
		{ 308, 178, 17 },
		{ 308, 179, 17 },
		{ 308, 180, 17 },
		{ 308, 181, 17 },
		{ 308, 182, 17 },
		{ 308, 183, 17 },
		{ 308, 184, 17 },
		{ 308, 185, 17 },
		{ 308, 186, 17 },
		{ 308, 187, 17 },
		{ 308, 188, 17 },
		{ 308, 189, 17 },
		{ 308, 190, 17 },
		{ 308, 191, 17 },
		{ 308, 192, 17 },
		{ 308, 193, 17 },
		{ 308, 194, 17 },
		{ 308, 195, 17 },
		{ 308, 196, 17 },
		{ 308, 197, 17 },
		{ 308, 198, 17 },
		{ 308, 199, 17 },
		{ 308, 200, 17 },
		{ 308, 201, 17 },
		{ 308, 202, 17 },
		{ 308, 203, 17 },
		{ 308, 204, 17 },
		{ 308, 205, 17 },
		{ 309, 154, 17 },
		{ 309, 155, 17 },
		{ 309, 156, 17 },
		{ 309, 157, 17 },
		{ 309, 158, 17 },
		{ 309, 159, 17 },
		{ 309, 160, 17 },
		{ 309, 161, 17 },
		{ 309, 162, 17 },
		{ 309, 163, 17 },
		{ 309, 164, 17 },
		{ 309, 165, 17 },
		{ 309, 166, 17 },
		{ 309, 167, 17 },
		{ 309, 168, 17 },
		{ 309, 169, 17 },
		{ 309, 170, 17 },
		{ 309, 171, 17 },
		{ 309, 172, 17 },
		{ 309, 173, 17 },
		{ 309, 174, 17 },
		{ 309, 175, 17 },
		{ 309, 176, 17 },
		{ 309, 177, 17 },
		{ 309, 178, 17 },
		{ 309, 179, 17 },
		{ 309, 180, 17 },
		{ 309, 181, 17 },
		{ 309, 182, 17 },
		{ 309, 183, 17 },
		{ 309, 184, 17 },
		{ 309, 185, 17 },
		{ 309, 186, 17 },
		{ 309, 187, 17 },
		{ 309, 188, 17 },
		{ 309, 189, 17 },
		{ 309, 190, 17 },
		{ 309, 191, 17 },
		{ 309, 192, 17 },
		{ 309, 193, 17 },
		{ 309, 194, 17 },
		{ 309, 195, 17 },
		{ 309, 196, 17 },
		{ 309, 197, 17 },
		{ 309, 198, 17 },
		{ 309, 199, 17 },
		{ 309, 200, 17 },
		{ 309, 201, 17 },
		{ 309, 202, 17 },
		{ 309, 203, 17 },
		{ 309, 204, 17 },
		{ 310, 152, 17 },
		{ 310, 153, 17 },
		{ 310, 154, 17 },
		{ 310, 155, 17 },
		{ 310, 156, 17 },
		{ 310, 157, 17 },
		{ 310, 158, 17 },
		{ 310, 159, 17 },
		{ 310, 160, 17 },
		{ 310, 161, 17 },
		{ 310, 162, 17 },
		{ 310, 163, 17 },
		{ 310, 164, 17 },
		{ 310, 165, 17 },
		{ 310, 166, 17 },
		{ 310, 167, 17 },
		{ 310, 168, 17 },
		{ 310, 169, 17 },
		{ 310, 170, 17 },
		{ 310, 171, 17 },
		{ 310, 172, 17 },
		{ 310, 173, 17 },
		{ 310, 174, 17 },
		{ 310, 175, 17 },
		{ 310, 176, 17 },
		{ 310, 177, 17 },
		{ 310, 178, 17 },
		{ 310, 179, 17 },
		{ 310, 180, 17 },
		{ 310, 181, 17 },
		{ 310, 182, 17 },
		{ 310, 183, 17 },
		{ 310, 184, 17 },
		{ 310, 185, 17 },
		{ 310, 186, 17 },
		{ 310, 187, 17 },
		{ 310, 188, 17 },
		{ 310, 189, 17 },
		{ 310, 190, 17 },
		{ 310, 191, 17 },
		{ 310, 192, 17 },
		{ 310, 193, 17 },
		{ 310, 194, 17 },
		{ 310, 195, 17 },
		{ 310, 196, 17 },
		{ 310, 197, 17 },
		{ 310, 198, 17 },
		{ 310, 199, 17 },
		{ 310, 200, 17 },
		{ 310, 201, 17 },
		{ 310, 202, 17 },
		{ 310, 203, 17 },
		{ 311, 150, 17 },
		{ 311, 151, 17 },
		{ 311, 152, 17 },
		{ 311, 153, 17 },
		{ 311, 154, 17 },
		{ 311, 155, 17 },
		{ 311, 156, 17 },
		{ 311, 157, 17 },
		{ 311, 158, 17 },
		{ 311, 159, 17 },
		{ 311, 160, 17 },
		{ 311, 161, 17 },
		{ 311, 162, 17 },
		{ 311, 163, 17 },
		{ 311, 164, 17 },
		{ 311, 165, 17 },
		{ 311, 166, 17 },
		{ 311, 167, 17 },
		{ 311, 168, 17 },
		{ 311, 169, 17 },
		{ 311, 170, 17 },
		{ 311, 171, 17 },
		{ 311, 172, 17 },
		{ 311, 173, 17 },
		{ 311, 174, 17 },
		{ 311, 175, 17 },
		{ 311, 176, 17 },
		{ 311, 177, 17 },
		{ 311, 178, 17 },
		{ 311, 179, 17 },
		{ 311, 180, 17 },
		{ 311, 181, 17 },
		{ 311, 182, 17 },
		{ 311, 183, 17 },
		{ 311, 184, 17 },
		{ 311, 185, 17 },
		{ 311, 186, 17 },
		{ 311, 187, 17 },
		{ 311, 188, 17 },
		{ 311, 189, 17 },
		{ 311, 190, 17 },
		{ 311, 191, 17 },
		{ 311, 192, 17 },
		{ 311, 193, 17 },
		{ 311, 194, 17 },
		{ 311, 195, 17 },
		{ 311, 196, 17 },
		{ 311, 197, 17 },
		{ 311, 198, 17 },
		{ 311, 199, 17 },
		{ 311, 200, 17 },
		{ 311, 201, 17 },
		{ 311, 202, 17 },
		{ 312, 148, 17 },
		{ 312, 149, 17 },
		{ 312, 150, 17 },
		{ 312, 151, 17 },
		{ 312, 152, 17 },
		{ 312, 153, 17 },
		{ 312, 154, 17 },
		{ 312, 155, 17 },
		{ 312, 156, 17 },
		{ 312, 157, 17 },
		{ 312, 158, 17 },
		{ 312, 159, 17 },
		{ 312, 160, 17 },
		{ 312, 161, 17 },
		{ 312, 162, 17 },
		{ 312, 163, 17 },
		{ 312, 164, 17 },
		{ 312, 165, 17 },
		{ 312, 166, 17 },
		{ 312, 167, 17 },
		{ 312, 168, 17 },
		{ 312, 169, 17 },
		{ 312, 170, 17 },
		{ 312, 171, 17 },
		{ 312, 172, 17 },
		{ 312, 173, 17 },
		{ 312, 174, 17 },
		{ 312, 175, 17 },
		{ 312, 176, 17 },
		{ 312, 177, 17 },
		{ 312, 178, 17 },
		{ 312, 179, 17 },
		{ 312, 180, 17 },
		{ 312, 181, 17 },
		{ 312, 182, 17 },
		{ 312, 183, 17 },
		{ 312, 184, 17 },
		{ 312, 185, 17 },
		{ 312, 186, 17 },
		{ 312, 187, 17 },
		{ 312, 188, 17 },
		{ 312, 189, 17 },
		{ 312, 190, 17 },
		{ 312, 191, 17 },
		{ 312, 192, 17 },
		{ 312, 193, 17 },
		{ 312, 194, 17 },
		{ 312, 195, 17 },
		{ 312, 196, 17 },
		{ 312, 197, 17 },
		{ 312, 198, 17 },
		{ 312, 199, 17 },
		{ 312, 200, 17 },
		{ 312, 201, 17 },
		{ 313, 146, 17 },
		{ 313, 147, 17 },
		{ 313, 148, 17 },
		{ 313, 149, 17 },
		{ 313, 150, 17 },
		{ 313, 151, 17 },
		{ 313, 152, 17 },
		{ 313, 153, 17 },
		{ 313, 154, 17 },
		{ 313, 155, 17 },
		{ 313, 156, 17 },
		{ 313, 157, 17 },
		{ 313, 158, 17 },
		{ 313, 159, 17 },
		{ 313, 160, 17 },
		{ 313, 161, 17 },
		{ 313, 162, 17 },
		{ 313, 163, 17 },
		{ 313, 164, 17 },
		{ 313, 165, 17 },
		{ 313, 166, 17 },
		{ 313, 167, 17 },
		{ 313, 168, 17 },
		{ 313, 169, 17 },
		{ 313, 170, 17 },
		{ 313, 171, 17 },
		{ 313, 172, 17 },
		{ 313, 173, 17 },
		{ 313, 174, 17 },
		{ 313, 175, 17 },
		{ 313, 176, 17 },
		{ 313, 177, 17 },
		{ 313, 178, 17 },
		{ 313, 179, 17 },
		{ 313, 180, 17 },
		{ 313, 181, 17 },
		{ 313, 182, 17 },
		{ 313, 183, 17 },
		{ 313, 184, 17 },
		{ 313, 185, 17 },
		{ 313, 186, 17 },
		{ 313, 187, 17 },
		{ 313, 188, 17 },
		{ 313, 189, 17 },
		{ 313, 190, 17 },
		{ 313, 191, 17 },
		{ 313, 192, 17 },
		{ 313, 193, 17 },
		{ 313, 194, 17 },
		{ 313, 195, 17 },
		{ 313, 196, 17 },
		{ 313, 197, 17 },
		{ 313, 198, 17 },
		{ 313, 199, 17 },
		{ 313, 200, 17 },
		{ 314, 144, 17 },
		{ 314, 145, 17 },
		{ 314, 146, 17 },
		{ 314, 147, 17 },
		{ 314, 148, 17 },
		{ 314, 149, 17 },
		{ 314, 150, 17 },
		{ 314, 151, 17 },
		{ 314, 152, 17 },
		{ 314, 153, 17 },
		{ 314, 154, 17 },
		{ 314, 155, 17 },
		{ 314, 156, 17 },
		{ 314, 157, 17 },
		{ 314, 158, 17 },
		{ 314, 159, 17 },
		{ 314, 160, 17 },
		{ 314, 161, 17 },
		{ 314, 162, 17 },
		{ 314, 163, 17 },
		{ 314, 164, 17 },
		{ 314, 165, 17 },
		{ 314, 166, 17 },
		{ 314, 167, 17 },
		{ 314, 168, 17 },
		{ 314, 169, 17 },
		{ 314, 170, 17 },
		{ 314, 171, 17 },
		{ 314, 172, 17 },
		{ 314, 173, 17 },
		{ 314, 174, 17 },
		{ 314, 175, 17 },
		{ 314, 176, 17 },
		{ 314, 177, 17 },
		{ 314, 178, 17 },
		{ 314, 179, 17 },
		{ 314, 180, 17 },
		{ 314, 181, 17 },
		{ 314, 182, 17 },
		{ 314, 183, 17 },
		{ 314, 184, 17 },
		{ 314, 185, 17 },
		{ 314, 186, 17 },
		{ 314, 187, 17 },
		{ 314, 188, 17 },
		{ 314, 189, 17 },
		{ 314, 190, 17 },
		{ 314, 191, 17 },
		{ 314, 192, 17 },
		{ 314, 193, 17 },
		{ 314, 194, 17 },
		{ 314, 195, 17 },
		{ 314, 196, 17 },
		{ 314, 197, 17 },
		{ 314, 198, 17 },
		{ 314, 199, 17 },
		{ 315, 142, 17 },
		{ 315, 143, 17 },
		{ 315, 144, 17 },
		{ 315, 145, 17 },
		{ 315, 146, 17 },
		{ 315, 147, 17 },
		{ 315, 148, 17 },
		{ 315, 149, 17 },
		{ 315, 150, 17 },
		{ 315, 151, 17 },
		{ 315, 152, 17 },
		{ 315, 153, 17 },
		{ 315, 154, 17 },
		{ 315, 155, 17 },
		{ 315, 156, 17 },
		{ 315, 157, 17 },
		{ 315, 158, 17 },
		{ 315, 159, 17 },
		{ 315, 160, 17 },
		{ 315, 161, 17 },
		{ 315, 162, 17 },
		{ 315, 163, 17 },
		{ 315, 164, 17 },
		{ 315, 165, 17 },
		{ 315, 166, 17 },
		{ 315, 167, 17 },
		{ 315, 168, 17 },
		{ 315, 169, 17 },
		{ 315, 170, 17 },
		{ 315, 171, 17 },
		{ 315, 172, 17 },
		{ 315, 173, 17 },
		{ 315, 174, 17 },
		{ 315, 175, 17 },
		{ 315, 176, 17 },
		{ 315, 177, 17 },
		{ 315, 178, 17 },
		{ 315, 179, 17 },
		{ 315, 180, 17 },
		{ 315, 181, 17 },
		{ 315, 182, 17 },
		{ 315, 183, 17 },
		{ 315, 184, 17 },
		{ 315, 185, 17 },
		{ 315, 186, 17 },
		{ 315, 187, 17 },
		{ 315, 188, 17 },
		{ 315, 189, 17 },
		{ 315, 190, 17 },
		{ 315, 191, 17 },
		{ 315, 192, 17 },
		{ 315, 193, 17 },
		{ 315, 194, 17 },
		{ 315, 195, 17 },
		{ 315, 196, 17 },
		{ 315, 197, 17 },
		{ 315, 198, 17 },
		{ 316, 140, 17 },
		{ 316, 141, 17 },
		{ 316, 142, 17 },
		{ 316, 143, 17 },
		{ 316, 144, 17 },
		{ 316, 145, 17 },
		{ 316, 146, 17 },
		{ 316, 147, 17 },
		{ 316, 148, 17 },
		{ 316, 149, 17 },
		{ 316, 150, 17 },
		{ 316, 151, 17 },
		{ 316, 152, 17 },
		{ 316, 153, 17 },
		{ 316, 154, 17 },
		{ 316, 155, 17 },
		{ 316, 156, 17 },
		{ 316, 157, 17 },
		{ 316, 158, 17 },
		{ 316, 159, 17 },
		{ 316, 160, 17 },
		{ 316, 161, 17 },
		{ 316, 162, 17 },
		{ 316, 163, 17 },
		{ 316, 164, 17 },
		{ 316, 165, 17 },
		{ 316, 166, 17 },
		{ 316, 167, 17 },
		{ 316, 168, 17 },
		{ 316, 169, 17 },
		{ 316, 170, 17 },
		{ 316, 171, 17 },
		{ 316, 172, 17 },
		{ 316, 173, 17 },
		{ 316, 174, 17 },
		{ 316, 175, 17 },
		{ 316, 176, 17 },
		{ 316, 177, 17 },
		{ 316, 178, 17 },
		{ 316, 179, 17 },
		{ 316, 180, 17 },
		{ 316, 181, 17 },
		{ 316, 182, 17 },
		{ 316, 183, 17 },
		{ 316, 184, 17 },
		{ 316, 185, 17 },
		{ 316, 186, 17 },
		{ 316, 187, 17 },
		{ 316, 188, 17 },
		{ 316, 189, 17 },
		{ 316, 190, 17 },
		{ 316, 191, 17 },
		{ 316, 192, 17 },
		{ 316, 193, 17 },
		{ 316, 194, 17 },
		{ 316, 195, 17 },
		{ 316, 196, 17 },
		{ 316, 197, 17 },
		{ 317, 138, 17 },
		{ 317, 139, 17 },
		{ 317, 140, 17 },
		{ 317, 141, 17 },
		{ 317, 142, 17 },
		{ 317, 143, 17 },
		{ 317, 144, 17 },
		{ 317, 145, 17 },
		{ 317, 146, 17 },
		{ 317, 147, 17 },
		{ 317, 148, 17 },
		{ 317, 149, 17 },
		{ 317, 150, 17 },
		{ 317, 151, 17 },
		{ 317, 152, 17 },
		{ 317, 153, 17 },
		{ 317, 154, 17 },
		{ 317, 155, 17 },
		{ 317, 156, 17 },
		{ 317, 157, 17 },
		{ 317, 158, 17 },
		{ 317, 159, 17 },
		{ 317, 160, 17 },
		{ 317, 161, 17 },
		{ 317, 162, 17 },
		{ 317, 163, 17 },
		{ 317, 164, 17 },
		{ 317, 165, 17 },
		{ 317, 166, 17 },
		{ 317, 167, 17 },
		{ 317, 168, 17 },
		{ 317, 169, 17 },
		{ 317, 170, 17 },
		{ 317, 171, 17 },
		{ 317, 172, 17 },
		{ 317, 173, 17 },
		{ 317, 174, 17 },
		{ 317, 175, 17 },
		{ 317, 176, 17 },
		{ 317, 177, 17 },
		{ 317, 178, 17 },
		{ 317, 179, 17 },
		{ 317, 180, 17 },
		{ 317, 181, 17 },
		{ 317, 182, 17 },
		{ 317, 183, 17 },
		{ 317, 184, 17 },
		{ 317, 185, 17 },
		{ 317, 186, 17 },
		{ 317, 187, 17 },
		{ 317, 188, 17 },
		{ 317, 189, 17 },
		{ 317, 190, 17 },
		{ 317, 191, 17 },
		{ 317, 192, 17 },
		{ 317, 193, 17 },
		{ 317, 194, 17 },
		{ 317, 195, 17 },
		{ 317, 196, 17 },
		{ 318, 137, 17 },
		{ 318, 138, 17 },
		{ 318, 139, 17 },
		{ 318, 140, 17 },
		{ 318, 141, 17 },
		{ 318, 142, 17 },
		{ 318, 143, 17 },
		{ 318, 144, 17 },
		{ 318, 145, 17 },
		{ 318, 146, 17 },
		{ 318, 147, 17 },
		{ 318, 148, 17 },
		{ 318, 149, 17 },
		{ 318, 150, 17 },
		{ 318, 151, 17 },
		{ 318, 152, 17 },
		{ 318, 153, 17 },
		{ 318, 154, 17 },
		{ 318, 155, 17 },
		{ 318, 156, 17 },
		{ 318, 157, 17 },
		{ 318, 158, 17 },
		{ 318, 159, 17 },
		{ 318, 160, 17 },
		{ 318, 161, 17 },
		{ 318, 162, 17 },
		{ 318, 163, 17 },
		{ 318, 164, 17 },
		{ 318, 165, 17 },
		{ 318, 166, 17 },
		{ 318, 167, 17 },
		{ 318, 168, 17 },
		{ 318, 169, 17 },
		{ 318, 170, 17 },
		{ 318, 171, 17 },
		{ 318, 172, 17 },
		{ 318, 173, 17 },
		{ 318, 174, 17 },
		{ 318, 175, 17 },
		{ 318, 176, 17 },
		{ 318, 177, 17 },
		{ 318, 178, 17 },
		{ 318, 179, 17 },
		{ 318, 180, 17 },
		{ 318, 181, 17 },
		{ 318, 182, 17 },
		{ 318, 183, 17 },
		{ 318, 184, 17 },
		{ 318, 185, 17 },
		{ 318, 186, 17 },
		{ 318, 187, 17 },
		{ 318, 188, 17 },
		{ 318, 189, 17 },
		{ 318, 190, 17 },
		{ 318, 191, 17 },
		{ 318, 192, 17 },
		{ 318, 193, 17 },
		{ 318, 194, 17 },
		{ 318, 195, 17 },
		{ 319, 135, 17 },
		{ 319, 136, 17 },
		{ 319, 137, 17 },
		{ 319, 138, 17 },
		{ 319, 139, 17 },
		{ 319, 140, 17 },
		{ 319, 141, 17 },
		{ 319, 142, 17 },
		{ 319, 143, 17 },
		{ 319, 144, 17 },
		{ 319, 145, 17 },
		{ 319, 146, 17 },
		{ 319, 147, 17 },
		{ 319, 148, 17 },
		{ 319, 149, 17 },
		{ 319, 150, 17 },
		{ 319, 151, 17 },
		{ 319, 152, 17 },
		{ 319, 153, 17 },
		{ 319, 154, 17 },
		{ 319, 155, 17 },
		{ 319, 156, 17 },
		{ 319, 157, 17 },
		{ 319, 158, 17 },
		{ 319, 159, 17 },
		{ 319, 160, 17 },
		{ 319, 161, 17 },
		{ 319, 162, 17 },
		{ 319, 163, 17 },
		{ 319, 164, 17 },
		{ 319, 165, 17 },
		{ 319, 166, 17 },
		{ 319, 167, 17 },
		{ 319, 168, 17 },
		{ 319, 169, 17 },
		{ 319, 170, 17 },
		{ 319, 171, 17 },
		{ 319, 172, 17 },
		{ 319, 173, 17 },
		{ 319, 174, 17 },
		{ 319, 175, 17 },
		{ 319, 176, 17 },
		{ 319, 177, 17 },
		{ 319, 178, 17 },
		{ 319, 179, 17 },
		{ 319, 180, 17 },
		{ 319, 181, 17 },
		{ 319, 182, 17 },
		{ 319, 183, 17 },
		{ 319, 184, 17 },
		{ 319, 185, 17 },
		{ 319, 186, 17 },
		{ 319, 187, 17 },
		{ 319, 188, 17 },
		{ 319, 189, 17 },
		{ 319, 190, 17 },
		{ 319, 191, 17 },
		{ 319, 192, 17 },
		{ 319, 193, 17 },
		{ 319, 194, 17 },
		{ 320, 134, 17 },
		{ 320, 135, 17 },
		{ 320, 136, 17 },
		{ 320, 137, 17 },
		{ 320, 138, 17 },
		{ 320, 139, 17 },
		{ 320, 140, 17 },
		{ 320, 141, 17 },
		{ 320, 142, 17 },
		{ 320, 143, 17 },
		{ 320, 144, 17 },
		{ 320, 145, 17 },
		{ 320, 146, 17 },
		{ 320, 147, 17 },
		{ 320, 148, 17 },
		{ 320, 149, 17 },
		{ 320, 150, 17 },
		{ 320, 151, 17 },
		{ 320, 152, 17 },
		{ 320, 153, 17 },
		{ 320, 154, 17 },
		{ 320, 155, 17 },
		{ 320, 156, 17 },
		{ 320, 157, 17 },
		{ 320, 158, 17 },
		{ 320, 159, 17 },
		{ 320, 160, 17 },
		{ 320, 161, 17 },
		{ 320, 162, 17 },
		{ 320, 163, 17 },
		{ 320, 164, 17 },
		{ 320, 165, 17 },
		{ 320, 166, 17 },
		{ 320, 167, 17 },
		{ 320, 168, 17 },
		{ 320, 169, 17 },
		{ 320, 170, 17 },
		{ 320, 171, 17 },
		{ 320, 172, 17 },
		{ 320, 173, 17 },
		{ 320, 174, 17 },
		{ 320, 175, 17 },
		{ 320, 176, 17 },
		{ 320, 177, 17 },
		{ 320, 178, 17 },
		{ 320, 179, 17 },
		{ 320, 180, 17 },
		{ 320, 181, 17 },
		{ 320, 182, 17 },
		{ 320, 183, 17 },
		{ 320, 184, 17 },
		{ 320, 185, 17 },
		{ 320, 186, 17 },
		{ 320, 187, 17 },
		{ 320, 188, 17 },
		{ 320, 189, 17 },
		{ 320, 190, 17 },
		{ 320, 191, 17 },
		{ 320, 192, 17 },
		{ 320, 193, 17 },
		{ 321, 133, 17 },
		{ 321, 134, 17 },
		{ 321, 135, 17 },
		{ 321, 136, 17 },
		{ 321, 137, 17 },
		{ 321, 138, 17 },
		{ 321, 139, 17 },
		{ 321, 140, 17 },
		{ 321, 141, 17 },
		{ 321, 142, 17 },
		{ 321, 143, 17 },
		{ 321, 144, 17 },
		{ 321, 145, 17 },
		{ 321, 146, 17 },
		{ 321, 147, 17 },
		{ 321, 148, 17 },
		{ 321, 149, 17 },
		{ 321, 150, 17 },
		{ 321, 151, 17 },
		{ 321, 152, 17 },
		{ 321, 153, 17 },
		{ 321, 154, 17 },
		{ 321, 155, 17 },
		{ 321, 156, 17 },
		{ 321, 157, 17 },
		{ 321, 158, 17 },
		{ 321, 159, 17 },
		{ 321, 160, 17 },
		{ 321, 161, 17 },
		{ 321, 162, 17 },
		{ 321, 163, 17 },
		{ 321, 164, 17 },
		{ 321, 165, 17 },
		{ 321, 166, 17 },
		{ 321, 167, 17 },
		{ 321, 168, 17 },
		{ 321, 169, 17 },
		{ 321, 170, 17 },
		{ 321, 171, 17 },
		{ 321, 172, 17 },
		{ 321, 173, 17 },
		{ 321, 174, 17 },
		{ 321, 175, 17 },
		{ 321, 176, 17 },
		{ 321, 177, 17 },
		{ 321, 178, 17 },
		{ 321, 179, 17 },
		{ 321, 180, 17 },
		{ 321, 181, 17 },
		{ 321, 182, 17 },
		{ 321, 183, 17 },
		{ 321, 184, 17 },
		{ 321, 185, 17 },
		{ 321, 186, 17 },
		{ 321, 187, 17 },
		{ 321, 188, 17 },
		{ 321, 189, 17 },
		{ 321, 190, 17 },
		{ 321, 191, 17 },
		{ 321, 192, 17 },
		{ 322, 133, 17 },
		{ 322, 134, 17 },
		{ 322, 135, 17 },
		{ 322, 136, 17 },
		{ 322, 137, 17 },
		{ 322, 138, 17 },
		{ 322, 139, 17 },
		{ 322, 140, 17 },
		{ 322, 141, 17 },
		{ 322, 142, 17 },
		{ 322, 143, 17 },
		{ 322, 144, 17 },
		{ 322, 145, 17 },
		{ 322, 146, 17 },
		{ 322, 147, 17 },
		{ 322, 148, 17 },
		{ 322, 149, 17 },
		{ 322, 150, 17 },
		{ 322, 151, 17 },
		{ 322, 152, 17 },
		{ 322, 153, 17 },
		{ 322, 154, 17 },
		{ 322, 155, 17 },
		{ 322, 156, 17 },
		{ 322, 157, 17 },
		{ 322, 158, 17 },
		{ 322, 159, 17 },
		{ 322, 160, 17 },
		{ 322, 161, 17 },
		{ 322, 162, 17 },
		{ 322, 163, 17 },
		{ 322, 164, 17 },
		{ 322, 165, 17 },
		{ 322, 166, 17 },
		{ 322, 167, 17 },
		{ 322, 168, 17 },
		{ 322, 169, 17 },
		{ 322, 170, 17 },
		{ 322, 171, 17 },
		{ 322, 172, 17 },
		{ 322, 173, 17 },
		{ 322, 174, 17 },
		{ 322, 175, 17 },
		{ 322, 176, 17 },
		{ 322, 177, 17 },
		{ 322, 178, 17 },
		{ 322, 179, 17 },
		{ 322, 180, 17 },
		{ 322, 181, 17 },
		{ 322, 182, 17 },
		{ 322, 183, 17 },
		{ 322, 184, 17 },
		{ 322, 185, 17 },
		{ 322, 186, 17 },
		{ 322, 187, 17 },
		{ 322, 188, 17 },
		{ 322, 189, 17 },
		{ 322, 190, 17 },
		{ 322, 191, 17 },
		{ 323, 132, 17 },
		{ 323, 133, 17 },
		{ 323, 134, 17 },
		{ 323, 135, 17 },
		{ 323, 136, 17 },
		{ 323, 137, 17 },
		{ 323, 138, 17 },
		{ 323, 139, 17 },
		{ 323, 140, 17 },
		{ 323, 141, 17 },
		{ 323, 142, 17 },
		{ 323, 143, 17 },
		{ 323, 144, 17 },
		{ 323, 145, 17 },
		{ 323, 146, 17 },
		{ 323, 147, 17 },
		{ 323, 148, 17 },
		{ 323, 149, 17 },
		{ 323, 150, 17 },
		{ 323, 151, 17 },
		{ 323, 152, 17 },
		{ 323, 153, 17 },
		{ 323, 154, 17 },
		{ 323, 155, 17 },
		{ 323, 156, 17 },
		{ 323, 157, 17 },
		{ 323, 158, 17 },
		{ 323, 159, 17 },
		{ 323, 160, 17 },
		{ 323, 161, 17 },
		{ 323, 162, 17 },
		{ 323, 163, 17 },
		{ 323, 164, 17 },
		{ 323, 165, 17 },
		{ 323, 166, 17 },
		{ 323, 167, 17 },
		{ 323, 168, 17 },
		{ 323, 169, 17 },
		{ 323, 170, 17 },
		{ 323, 171, 17 },
		{ 323, 172, 17 },
		{ 323, 173, 17 },
		{ 323, 174, 17 },
		{ 323, 175, 17 },
		{ 323, 176, 17 },
		{ 323, 177, 17 },
		{ 323, 178, 17 },
		{ 323, 179, 17 },
		{ 323, 180, 17 },
		{ 323, 181, 17 },
		{ 323, 182, 17 },
		{ 323, 183, 17 },
		{ 323, 184, 17 },
		{ 323, 185, 17 },
		{ 323, 186, 17 },
		{ 323, 187, 17 },
		{ 323, 188, 17 },
		{ 323, 189, 17 },
		{ 323, 190, 17 },
		{ 324, 132, 17 },
		{ 324, 133, 17 },
		{ 324, 134, 17 },
		{ 324, 135, 17 },
		{ 324, 136, 17 },
		{ 324, 137, 17 },
		{ 324, 138, 17 },
		{ 324, 139, 17 },
		{ 324, 140, 17 },
		{ 324, 141, 17 },
		{ 324, 142, 17 },
		{ 324, 143, 17 },
		{ 324, 144, 17 },
		{ 324, 145, 17 },
		{ 324, 146, 17 },
		{ 324, 147, 17 },
		{ 324, 148, 17 },
		{ 324, 149, 17 },
		{ 324, 150, 17 },
		{ 324, 151, 17 },
		{ 324, 152, 17 },
		{ 324, 153, 17 },
		{ 324, 154, 17 },
		{ 324, 155, 17 },
		{ 324, 156, 17 },
		{ 324, 157, 17 },
		{ 324, 158, 17 },
		{ 324, 159, 17 },
		{ 324, 160, 17 },
		{ 324, 161, 17 },
		{ 324, 162, 17 },
		{ 324, 163, 17 },
		{ 324, 164, 17 },
		{ 324, 165, 17 },
		{ 324, 166, 17 },
		{ 324, 167, 17 },
		{ 324, 168, 17 },
		{ 324, 169, 17 },
		{ 324, 170, 17 },
		{ 324, 171, 17 },
		{ 324, 172, 17 },
		{ 324, 173, 17 },
		{ 324, 174, 17 },
		{ 324, 175, 17 },
		{ 324, 176, 17 },
		{ 324, 177, 17 },
		{ 324, 178, 17 },
		{ 324, 179, 17 },
		{ 324, 180, 17 },
		{ 324, 181, 17 },
		{ 324, 182, 17 },
		{ 324, 183, 17 },
		{ 324, 184, 17 },
		{ 324, 185, 17 },
		{ 324, 186, 17 },
		{ 324, 187, 17 },
		{ 324, 188, 17 },
		{ 324, 189, 17 },
		{ 325, 132, 17 },
		{ 325, 133, 17 },
		{ 325, 134, 17 },
		{ 325, 135, 17 },
		{ 325, 136, 17 },
		{ 325, 137, 17 },
		{ 325, 138, 17 },
		{ 325, 139, 17 },
		{ 325, 140, 17 },
		{ 325, 141, 17 },
		{ 325, 142, 17 },
		{ 325, 143, 17 },
		{ 325, 144, 17 },
		{ 325, 145, 17 },
		{ 325, 146, 17 },
		{ 325, 147, 17 },
		{ 325, 148, 17 },
		{ 325, 149, 17 },
		{ 325, 150, 17 },
		{ 325, 151, 17 },
		{ 325, 152, 17 },
		{ 325, 153, 17 },
		{ 325, 154, 17 },
		{ 325, 155, 17 },
		{ 325, 156, 17 },
		{ 325, 157, 17 },
		{ 325, 158, 17 },
		{ 325, 159, 17 },
		{ 325, 160, 17 },
		{ 325, 161, 17 },
		{ 325, 162, 17 },
		{ 325, 163, 17 },
		{ 325, 164, 17 },
		{ 325, 165, 17 },
		{ 325, 166, 17 },
		{ 325, 167, 17 },
		{ 325, 168, 17 },
		{ 325, 169, 17 },
		{ 325, 170, 17 },
		{ 325, 171, 17 },
		{ 325, 172, 17 },
		{ 325, 173, 17 },
		{ 325, 174, 17 },
		{ 325, 175, 17 },
		{ 325, 176, 17 },
		{ 325, 177, 17 },
		{ 325, 178, 17 },
		{ 325, 179, 17 },
		{ 325, 180, 17 },
		{ 325, 181, 17 },
		{ 325, 182, 17 },
		{ 325, 183, 17 },
		{ 325, 184, 17 },
		{ 325, 185, 17 },
		{ 325, 186, 17 },
		{ 325, 187, 17 },
		{ 325, 188, 17 },
		{ 326, 132, 17 },
		{ 326, 133, 17 },
		{ 326, 134, 17 },
		{ 326, 135, 17 },
		{ 326, 136, 17 },
		{ 326, 137, 17 },
		{ 326, 138, 17 },
		{ 326, 139, 17 },
		{ 326, 140, 17 },
		{ 326, 141, 17 },
		{ 326, 142, 17 },
		{ 326, 143, 17 },
		{ 326, 144, 17 },
		{ 326, 145, 17 },
		{ 326, 146, 17 },
		{ 326, 147, 17 },
		{ 326, 148, 17 },
		{ 326, 149, 17 },
		{ 326, 150, 17 },
		{ 326, 151, 17 },
		{ 326, 152, 17 },
		{ 326, 153, 17 },
		{ 326, 154, 17 },
		{ 326, 155, 17 },
		{ 326, 156, 17 },
		{ 326, 157, 17 },
		{ 326, 158, 17 },
		{ 326, 159, 17 },
		{ 326, 160, 17 },
		{ 326, 161, 17 },
		{ 326, 162, 17 },
		{ 326, 163, 17 },
		{ 326, 164, 17 },
		{ 326, 165, 17 },
		{ 326, 166, 17 },
		{ 326, 167, 17 },
		{ 326, 168, 17 },
		{ 326, 169, 17 },
		{ 326, 170, 17 },
		{ 326, 171, 17 },
		{ 326, 172, 17 },
		{ 326, 173, 17 },
		{ 326, 174, 17 },
		{ 326, 175, 17 },
		{ 326, 176, 17 },
		{ 326, 177, 17 },
		{ 326, 178, 17 },
		{ 326, 179, 17 },
		{ 326, 180, 17 },
		{ 326, 181, 17 },
		{ 326, 182, 17 },
		{ 326, 183, 17 },
		{ 326, 184, 17 },
		{ 326, 185, 17 },
		{ 326, 186, 17 },
		{ 326, 187, 17 },
		{ 327, 131, 17 },
		{ 327, 132, 17 },
		{ 327, 133, 17 },
		{ 327, 134, 17 },
		{ 327, 135, 17 },
		{ 327, 136, 17 },
		{ 327, 137, 17 },
		{ 327, 138, 17 },
		{ 327, 139, 17 },
		{ 327, 140, 17 },
		{ 327, 141, 17 },
		{ 327, 142, 17 },
		{ 327, 143, 17 },
		{ 327, 144, 17 },
		{ 327, 145, 17 },
		{ 327, 146, 17 },
		{ 327, 147, 17 },
		{ 327, 148, 17 },
		{ 327, 149, 17 },
		{ 327, 150, 17 },
		{ 327, 151, 17 },
		{ 327, 152, 17 },
		{ 327, 153, 17 },
		{ 327, 154, 17 },
		{ 327, 155, 17 },
		{ 327, 156, 17 },
		{ 327, 157, 17 },
		{ 327, 158, 17 },
		{ 327, 159, 17 },
		{ 327, 160, 17 },
		{ 327, 161, 17 },
		{ 327, 162, 17 },
		{ 327, 163, 17 },
		{ 327, 164, 17 },
		{ 327, 165, 17 },
		{ 327, 166, 17 },
		{ 327, 167, 17 },
		{ 327, 168, 17 },
		{ 327, 169, 17 },
		{ 327, 170, 17 },
		{ 327, 171, 17 },
		{ 327, 172, 17 },
		{ 327, 173, 17 },
		{ 327, 174, 17 },
		{ 327, 175, 17 },
		{ 327, 176, 17 },
		{ 327, 177, 17 },
		{ 327, 178, 17 },
		{ 327, 179, 17 },
		{ 327, 180, 17 },
		{ 327, 181, 17 },
		{ 327, 182, 17 },
		{ 327, 183, 17 },
		{ 327, 184, 17 },
		{ 327, 185, 17 },
		{ 327, 186, 17 },
		{ 328, 131, 17 },
		{ 328, 132, 17 },
		{ 328, 133, 17 },
		{ 328, 134, 17 },
		{ 328, 135, 17 },
		{ 328, 136, 17 },
		{ 328, 137, 17 },
		{ 328, 138, 17 },
		{ 328, 139, 17 },
		{ 328, 140, 17 },
		{ 328, 141, 17 },
		{ 328, 142, 17 },
		{ 328, 143, 17 },
		{ 328, 144, 17 },
		{ 328, 145, 17 },
		{ 328, 146, 17 },
		{ 328, 147, 17 },
		{ 328, 148, 17 },
		{ 328, 149, 17 },
		{ 328, 150, 17 },
		{ 328, 151, 17 },
		{ 328, 152, 17 },
		{ 328, 153, 17 },
		{ 328, 154, 17 },
		{ 328, 155, 17 },
		{ 328, 156, 17 },
		{ 328, 157, 17 },
		{ 328, 158, 17 },
		{ 328, 159, 17 },
		{ 328, 160, 17 },
		{ 328, 161, 17 },
		{ 328, 162, 17 },
		{ 328, 163, 17 },
		{ 328, 164, 17 },
		{ 328, 165, 17 },
		{ 328, 166, 17 },
		{ 328, 167, 17 },
		{ 328, 168, 17 },
		{ 328, 169, 17 },
		{ 328, 170, 17 },
		{ 328, 171, 17 },
		{ 328, 172, 17 },
		{ 328, 173, 17 },
		{ 328, 174, 17 },
		{ 328, 175, 17 },
		{ 328, 176, 17 },
		{ 328, 177, 17 },
		{ 328, 178, 17 },
		{ 328, 179, 17 },
		{ 328, 180, 17 },
		{ 328, 181, 17 },
		{ 328, 182, 17 },
		{ 328, 183, 17 },
		{ 328, 184, 17 },
		{ 328, 185, 17 },
		{ 329, 131, 17 },
		{ 329, 132, 17 },
		{ 329, 133, 17 },
		{ 329, 134, 17 },
		{ 329, 135, 17 },
		{ 329, 136, 17 },
		{ 329, 137, 17 },
		{ 329, 138, 17 },
		{ 329, 139, 17 },
		{ 329, 140, 17 },
		{ 329, 141, 17 },
		{ 329, 142, 17 },
		{ 329, 143, 17 },
		{ 329, 144, 17 },
		{ 329, 145, 17 },
		{ 329, 146, 17 },
		{ 329, 147, 17 },
		{ 329, 148, 17 },
		{ 329, 149, 17 },
		{ 329, 150, 17 },
		{ 329, 151, 17 },
		{ 329, 152, 17 },
		{ 329, 153, 17 },
		{ 329, 154, 17 },
		{ 329, 155, 17 },
		{ 329, 156, 17 },
		{ 329, 157, 17 },
		{ 329, 158, 17 },
		{ 329, 159, 17 },
		{ 329, 160, 17 },
		{ 329, 161, 17 },
		{ 329, 162, 17 },
		{ 329, 163, 17 },
		{ 329, 164, 17 },
		{ 329, 165, 17 },
		{ 329, 166, 17 },
		{ 329, 167, 17 },
		{ 329, 168, 17 },
		{ 329, 169, 17 },
		{ 329, 170, 17 },
		{ 329, 171, 17 },
		{ 329, 172, 17 },
		{ 329, 173, 17 },
		{ 329, 174, 17 },
		{ 329, 175, 17 },
		{ 329, 176, 17 },
		{ 329, 177, 17 },
		{ 329, 178, 17 },
		{ 329, 179, 17 },
		{ 329, 180, 17 },
		{ 329, 181, 17 },
		{ 329, 182, 17 },
		{ 329, 183, 17 },
		{ 329, 184, 17 },
		{ 330, 131, 17 },
		{ 330, 132, 17 },
		{ 330, 133, 17 },
		{ 330, 134, 17 },
		{ 330, 135, 17 },
		{ 330, 136, 17 },
		{ 330, 137, 17 },
		{ 330, 138, 17 },
		{ 330, 139, 17 },
		{ 330, 140, 17 },
		{ 330, 141, 17 },
		{ 330, 142, 17 },
		{ 330, 143, 17 },
		{ 330, 144, 17 },
		{ 330, 145, 17 },
		{ 330, 146, 17 },
		{ 330, 147, 17 },
		{ 330, 148, 17 },
		{ 330, 149, 17 },
		{ 330, 150, 17 },
		{ 330, 151, 17 },
		{ 330, 152, 17 },
		{ 330, 153, 17 },
		{ 330, 154, 17 },
		{ 330, 155, 17 },
		{ 330, 156, 17 },
		{ 330, 157, 17 },
		{ 330, 158, 17 },
		{ 330, 159, 17 },
		{ 330, 160, 17 },
		{ 330, 161, 17 },
		{ 330, 162, 17 },
		{ 330, 163, 17 },
		{ 330, 164, 17 },
		{ 330, 165, 17 },
		{ 330, 166, 17 },
		{ 330, 167, 17 },
		{ 330, 168, 17 },
		{ 330, 169, 17 },
		{ 330, 170, 17 },
		{ 330, 171, 17 },
		{ 330, 172, 17 },
		{ 330, 173, 17 },
		{ 330, 174, 17 },
		{ 330, 175, 17 },
		{ 330, 176, 17 },
		{ 330, 177, 17 },
		{ 330, 178, 17 },
		{ 330, 179, 17 },
		{ 330, 180, 17 },
		{ 330, 181, 17 },
		{ 330, 182, 17 },
		{ 330, 183, 17 },
		{ 331, 131, 17 },
		{ 331, 132, 17 },
		{ 331, 133, 17 },
		{ 331, 134, 17 },
		{ 331, 135, 17 },
		{ 331, 136, 17 },
		{ 331, 137, 17 },
		{ 331, 138, 17 },
		{ 331, 139, 17 },
		{ 331, 140, 17 },
		{ 331, 141, 17 },
		{ 331, 142, 17 },
		{ 331, 143, 17 },
		{ 331, 144, 17 },
		{ 331, 145, 17 },
		{ 331, 146, 17 },
		{ 331, 147, 17 },
		{ 331, 148, 17 },
		{ 331, 149, 17 },
		{ 331, 150, 17 },
		{ 331, 151, 17 },
		{ 331, 152, 17 },
		{ 331, 153, 17 },
		{ 331, 154, 17 },
		{ 331, 155, 17 },
		{ 331, 156, 17 },
		{ 331, 157, 17 },
		{ 331, 158, 17 },
		{ 331, 159, 17 },
		{ 331, 160, 17 },
		{ 331, 161, 17 },
		{ 331, 162, 17 },
		{ 331, 163, 17 },
		{ 331, 164, 17 },
		{ 331, 165, 17 },
		{ 331, 166, 17 },
		{ 331, 167, 17 },
		{ 331, 168, 17 },
		{ 331, 169, 17 },
		{ 331, 170, 17 },
		{ 331, 171, 17 },
		{ 331, 172, 17 },
		{ 331, 173, 17 },
		{ 331, 174, 17 },
		{ 331, 175, 17 },
		{ 331, 176, 17 },
		{ 331, 177, 17 },
		{ 331, 178, 17 },
		{ 331, 179, 17 },
		{ 331, 180, 17 },
		{ 331, 181, 17 },
		{ 331, 182, 17 },
		{ 332, 131, 17 },
		{ 332, 132, 17 },
		{ 332, 133, 17 },
		{ 332, 134, 17 },
		{ 332, 135, 17 },
		{ 332, 136, 17 },
		{ 332, 137, 17 },
		{ 332, 138, 17 },
		{ 332, 139, 17 },
		{ 332, 140, 17 },
		{ 332, 141, 17 },
		{ 332, 142, 17 },
		{ 332, 143, 17 },
		{ 332, 144, 17 },
		{ 332, 145, 17 },
		{ 332, 146, 17 },
		{ 332, 147, 17 },
		{ 332, 148, 17 },
		{ 332, 149, 17 },
		{ 332, 150, 17 },
		{ 332, 151, 17 },
		{ 332, 152, 17 },
		{ 332, 153, 17 },
		{ 332, 154, 17 },
		{ 332, 155, 17 },
		{ 332, 156, 17 },
		{ 332, 157, 17 },
		{ 332, 158, 17 },
		{ 332, 159, 17 },
		{ 332, 160, 17 },
		{ 332, 161, 17 },
		{ 332, 162, 17 },
		{ 332, 163, 17 },
		{ 332, 164, 17 },
		{ 332, 165, 17 },
		{ 332, 166, 17 },
		{ 332, 167, 17 },
		{ 332, 168, 17 },
		{ 332, 169, 17 },
		{ 332, 170, 17 },
		{ 332, 171, 17 },
		{ 332, 172, 17 },
		{ 332, 173, 17 },
		{ 332, 174, 17 },
		{ 332, 175, 17 },
		{ 332, 176, 17 },
		{ 332, 177, 17 },
		{ 332, 178, 17 },
		{ 332, 179, 17 },
		{ 332, 180, 17 },
		{ 332, 181, 17 },
		{ 333, 131, 17 },
		{ 333, 132, 17 },
		{ 333, 133, 17 },
		{ 333, 134, 17 },
		{ 333, 135, 17 },
		{ 333, 136, 17 },
		{ 333, 137, 17 },
		{ 333, 138, 17 },
		{ 333, 139, 17 },
		{ 333, 140, 17 },
		{ 333, 141, 17 },
		{ 333, 142, 17 },
		{ 333, 143, 17 },
		{ 333, 144, 17 },
		{ 333, 145, 17 },
		{ 333, 146, 17 },
		{ 333, 147, 17 },
		{ 333, 148, 17 },
		{ 333, 149, 17 },
		{ 333, 150, 17 },
		{ 333, 151, 17 },
		{ 333, 152, 17 },
		{ 333, 153, 17 },
		{ 333, 154, 17 },
		{ 333, 155, 17 },
		{ 333, 156, 17 },
		{ 333, 157, 17 },
		{ 333, 158, 17 },
		{ 333, 159, 17 },
		{ 333, 160, 17 },
		{ 333, 161, 17 },
		{ 333, 162, 17 },
		{ 333, 163, 17 },
		{ 333, 164, 17 },
		{ 333, 165, 17 },
		{ 333, 166, 17 },
		{ 333, 167, 17 },
		{ 333, 168, 17 },
		{ 333, 169, 17 },
		{ 333, 170, 17 },
		{ 333, 171, 17 },
		{ 333, 172, 17 },
		{ 333, 173, 17 },
		{ 333, 174, 17 },
		{ 333, 175, 17 },
		{ 333, 176, 17 },
		{ 333, 177, 17 },
		{ 333, 178, 17 },
		{ 333, 179, 17 },
		{ 333, 180, 17 },
		{ 334, 131, 17 },
		{ 334, 132, 17 },
		{ 334, 133, 17 },
		{ 334, 134, 17 },
		{ 334, 135, 17 },
		{ 334, 136, 17 },
		{ 334, 137, 17 },
		{ 334, 138, 17 },
		{ 334, 139, 17 },
		{ 334, 140, 17 },
		{ 334, 141, 17 },
		{ 334, 142, 17 },
		{ 334, 143, 17 },
		{ 334, 144, 17 },
		{ 334, 145, 17 },
		{ 334, 146, 17 },
		{ 334, 147, 17 },
		{ 334, 148, 17 },
		{ 334, 149, 17 },
		{ 334, 150, 17 },
		{ 334, 151, 17 },
		{ 334, 152, 17 },
		{ 334, 153, 17 },
		{ 334, 154, 17 },
		{ 334, 155, 17 },
		{ 334, 156, 17 },
		{ 334, 157, 17 },
		{ 334, 158, 17 },
		{ 334, 159, 17 },
		{ 334, 160, 17 },
		{ 334, 161, 17 },
		{ 334, 162, 17 },
		{ 334, 163, 17 },
		{ 334, 164, 17 },
		{ 334, 165, 17 },
		{ 334, 166, 17 },
		{ 334, 167, 17 },
		{ 334, 168, 17 },
		{ 334, 169, 17 },
		{ 334, 170, 17 },
		{ 334, 171, 17 },
		{ 334, 172, 17 },
		{ 334, 173, 17 },
		{ 334, 174, 17 },
		{ 334, 175, 17 },
		{ 334, 176, 17 },
		{ 334, 177, 17 },
		{ 334, 178, 17 },
		{ 334, 179, 17 },
		{ 335, 131, 17 },
		{ 335, 132, 17 },
		{ 335, 133, 17 },
		{ 335, 134, 17 },
		{ 335, 135, 17 },
		{ 335, 136, 17 },
		{ 335, 137, 17 },
		{ 335, 138, 17 },
		{ 335, 139, 17 },
		{ 335, 140, 17 },
		{ 335, 141, 17 },
		{ 335, 142, 17 },
		{ 335, 143, 17 },
		{ 335, 144, 17 },
		{ 335, 145, 17 },
		{ 335, 146, 17 },
		{ 335, 147, 17 },
		{ 335, 148, 17 },
		{ 335, 149, 17 },
		{ 335, 150, 17 },
		{ 335, 151, 17 },
		{ 335, 152, 17 },
		{ 335, 153, 17 },
		{ 335, 154, 17 },
		{ 335, 155, 17 },
		{ 335, 156, 17 },
		{ 335, 157, 17 },
		{ 335, 158, 17 },
		{ 335, 159, 17 },
		{ 335, 160, 17 },
		{ 335, 161, 17 },
		{ 335, 162, 17 },
		{ 335, 163, 17 },
		{ 335, 164, 17 },
		{ 335, 165, 17 },
		{ 335, 166, 17 },
		{ 335, 167, 17 },
		{ 335, 168, 17 },
		{ 335, 169, 17 },
		{ 335, 170, 17 },
		{ 335, 171, 17 },
		{ 335, 172, 17 },
		{ 335, 173, 17 },
		{ 335, 174, 17 },
		{ 335, 175, 17 },
		{ 335, 176, 17 },
		{ 335, 177, 17 },
		{ 335, 178, 17 },
		{ 336, 131, 17 },
		{ 336, 132, 17 },
		{ 336, 133, 17 },
		{ 336, 134, 17 },
		{ 336, 135, 17 },
		{ 336, 136, 17 },
		{ 336, 137, 17 },
		{ 336, 138, 17 },
		{ 336, 139, 17 },
		{ 336, 140, 17 },
		{ 336, 141, 17 },
		{ 336, 142, 17 },
		{ 336, 143, 17 },
		{ 336, 144, 17 },
		{ 336, 145, 17 },
		{ 336, 146, 17 },
		{ 336, 147, 17 },
		{ 336, 148, 17 },
		{ 336, 149, 17 },
		{ 336, 150, 17 },
		{ 336, 151, 17 },
		{ 336, 152, 17 },
		{ 336, 153, 17 },
		{ 336, 154, 17 },
		{ 336, 155, 17 },
		{ 336, 156, 17 },
		{ 336, 157, 17 },
		{ 336, 158, 17 },
		{ 336, 159, 17 },
		{ 336, 160, 17 },
		{ 336, 161, 17 },
		{ 336, 162, 17 },
		{ 336, 163, 17 },
		{ 336, 164, 17 },
		{ 336, 165, 17 },
		{ 336, 166, 17 },
		{ 336, 167, 17 },
		{ 336, 168, 17 },
		{ 336, 169, 17 },
		{ 336, 170, 17 },
		{ 336, 171, 17 },
		{ 336, 172, 17 },
		{ 336, 173, 17 },
		{ 336, 174, 17 },
		{ 336, 175, 17 },
		{ 336, 176, 17 },
		{ 336, 177, 17 },
		{ 337, 131, 17 },
		{ 337, 132, 17 },
		{ 337, 133, 17 },
		{ 337, 134, 17 },
		{ 337, 135, 17 },
		{ 337, 136, 17 },
		{ 337, 137, 17 },
		{ 337, 138, 17 },
		{ 337, 139, 17 },
		{ 337, 140, 17 },
		{ 337, 141, 17 },
		{ 337, 142, 17 },
		{ 337, 143, 17 },
		{ 337, 144, 17 },
		{ 337, 145, 17 },
		{ 337, 146, 17 },
		{ 337, 147, 17 },
		{ 337, 148, 17 },
		{ 337, 149, 17 },
		{ 337, 150, 17 },
		{ 337, 151, 17 },
		{ 337, 152, 17 },
		{ 337, 153, 17 },
		{ 337, 154, 17 },
		{ 337, 155, 17 },
		{ 337, 156, 17 },
		{ 337, 157, 17 },
		{ 337, 158, 17 },
		{ 337, 159, 17 },
		{ 337, 160, 17 },
		{ 337, 161, 17 },
		{ 337, 162, 17 },
		{ 337, 163, 17 },
		{ 337, 164, 17 },
		{ 337, 165, 17 },
		{ 337, 166, 17 },
		{ 337, 167, 17 },
		{ 337, 168, 17 },
		{ 337, 169, 17 },
		{ 337, 170, 17 },
		{ 337, 171, 17 },
		{ 337, 172, 17 },
		{ 337, 173, 17 },
		{ 337, 174, 17 },
		{ 337, 175, 17 },
		{ 337, 176, 17 },
		{ 338, 131, 17 },
		{ 338, 132, 17 },
		{ 338, 133, 17 },
		{ 338, 134, 17 },
		{ 338, 135, 17 },
		{ 338, 136, 17 },
		{ 338, 137, 17 },
		{ 338, 138, 17 },
		{ 338, 139, 17 },
		{ 338, 140, 17 },
		{ 338, 141, 17 },
		{ 338, 142, 17 },
		{ 338, 143, 17 },
		{ 338, 144, 17 },
		{ 338, 145, 17 },
		{ 338, 146, 17 },
		{ 338, 147, 17 },
		{ 338, 148, 17 },
		{ 338, 149, 17 },
		{ 338, 150, 17 },
		{ 338, 151, 17 },
		{ 338, 152, 17 },
		{ 338, 153, 17 },
		{ 338, 154, 17 },
		{ 338, 155, 17 },
		{ 338, 156, 17 },
		{ 338, 157, 17 },
		{ 338, 158, 17 },
		{ 338, 159, 17 },
		{ 338, 160, 17 },
		{ 338, 161, 17 },
		{ 338, 162, 17 },
		{ 338, 163, 17 },
		{ 338, 164, 17 },
		{ 338, 165, 17 },
		{ 338, 166, 17 },
		{ 338, 167, 17 },
		{ 338, 168, 17 },
		{ 338, 169, 17 },
		{ 338, 170, 17 },
		{ 338, 171, 17 },
		{ 338, 172, 17 },
		{ 338, 173, 17 },
		{ 338, 174, 17 },
		{ 338, 175, 17 },
		{ 339, 131, 17 },
		{ 339, 132, 17 },
		{ 339, 133, 17 },
		{ 339, 134, 17 },
		{ 339, 135, 17 },
		{ 339, 136, 17 },
		{ 339, 137, 17 },
		{ 339, 138, 17 },
		{ 339, 139, 17 },
		{ 339, 140, 17 },
		{ 339, 141, 17 },
		{ 339, 142, 17 },
		{ 339, 143, 17 },
		{ 339, 144, 17 },
		{ 339, 145, 17 },
		{ 339, 146, 17 },
		{ 339, 147, 17 },
		{ 339, 148, 17 },
		{ 339, 149, 17 },
		{ 339, 150, 17 },
		{ 339, 151, 17 },
		{ 339, 152, 17 },
		{ 339, 153, 17 },
		{ 339, 154, 17 },
		{ 339, 155, 17 },
		{ 339, 156, 17 },
		{ 339, 157, 17 },
		{ 339, 158, 17 },
		{ 339, 159, 17 },
		{ 339, 160, 17 },
		{ 339, 161, 17 },
		{ 339, 162, 17 },
		{ 339, 163, 17 },
		{ 339, 164, 17 },
		{ 339, 165, 17 },
		{ 339, 166, 17 },
		{ 339, 167, 17 },
		{ 339, 168, 17 },
		{ 339, 169, 17 },
		{ 339, 170, 17 },
		{ 339, 171, 17 },
		{ 339, 172, 17 },
		{ 339, 173, 17 },
		{ 339, 174, 17 },
		{ 340, 131, 17 },
		{ 340, 132, 17 },
		{ 340, 133, 17 },
		{ 340, 134, 17 },
		{ 340, 135, 17 },
		{ 340, 136, 17 },
		{ 340, 137, 17 },
		{ 340, 138, 17 },
		{ 340, 139, 17 },
		{ 340, 140, 17 },
		{ 340, 141, 17 },
		{ 340, 142, 17 },
		{ 340, 143, 17 },
		{ 340, 144, 17 },
		{ 340, 145, 17 },
		{ 340, 146, 17 },
		{ 340, 147, 17 },
		{ 340, 148, 17 },
		{ 340, 149, 17 },
		{ 340, 150, 17 },
		{ 340, 151, 17 },
		{ 340, 152, 17 },
		{ 340, 153, 17 },
		{ 340, 154, 17 },
		{ 340, 155, 17 },
		{ 340, 156, 17 },
		{ 340, 157, 17 },
		{ 340, 158, 17 },
		{ 340, 159, 17 },
		{ 340, 160, 17 },
		{ 340, 161, 17 },
		{ 340, 162, 17 },
		{ 340, 163, 17 },
		{ 340, 164, 17 },
		{ 340, 165, 17 },
		{ 340, 166, 17 },
		{ 340, 167, 17 },
		{ 340, 168, 17 },
		{ 340, 169, 17 },
		{ 340, 170, 17 },
		{ 340, 171, 17 },
		{ 340, 172, 17 },
		{ 340, 173, 17 },
		{ 341, 131, 17 },
		{ 341, 132, 17 },
		{ 341, 133, 17 },
		{ 341, 134, 17 },
		{ 341, 135, 17 },
		{ 341, 136, 17 },
		{ 341, 137, 17 },
		{ 341, 138, 17 },
		{ 341, 139, 17 },
		{ 341, 140, 17 },
		{ 341, 141, 17 },
		{ 341, 142, 17 },
		{ 341, 143, 17 },
		{ 341, 144, 17 },
		{ 341, 145, 17 },
		{ 341, 146, 17 },
		{ 341, 147, 17 },
		{ 341, 148, 17 },
		{ 341, 149, 17 },
		{ 341, 150, 17 },
		{ 341, 151, 17 },
		{ 341, 152, 17 },
		{ 341, 153, 17 },
		{ 341, 154, 17 },
		{ 341, 155, 17 },
		{ 341, 156, 17 },
		{ 341, 157, 17 },
		{ 341, 158, 17 },
		{ 341, 159, 17 },
		{ 341, 160, 17 },
		{ 341, 161, 17 },
		{ 341, 162, 17 },
		{ 341, 163, 17 },
		{ 341, 164, 17 },
		{ 341, 165, 17 },
		{ 341, 166, 17 },
		{ 341, 167, 17 },
		{ 341, 168, 17 },
		{ 341, 169, 17 },
		{ 341, 170, 17 },
		{ 341, 171, 17 },
		{ 341, 172, 17 },
		{ 342, 131, 17 },
		{ 342, 132, 17 },
		{ 342, 133, 17 },
		{ 342, 134, 17 },
		{ 342, 135, 17 },
		{ 342, 136, 17 },
		{ 342, 137, 17 },
		{ 342, 138, 17 },
		{ 342, 139, 17 },
		{ 342, 140, 17 },
		{ 342, 141, 17 },
		{ 342, 142, 17 },
		{ 342, 143, 17 },
		{ 342, 144, 17 },
		{ 342, 145, 17 },
		{ 342, 146, 17 },
		{ 342, 147, 17 },
		{ 342, 148, 17 },
		{ 342, 149, 17 },
		{ 342, 150, 17 },
		{ 342, 151, 17 },
		{ 342, 152, 17 },
		{ 342, 153, 17 },
		{ 342, 154, 17 },
		{ 342, 155, 17 },
		{ 342, 156, 17 },
		{ 342, 157, 17 },
		{ 342, 158, 17 },
		{ 342, 159, 17 },
		{ 342, 160, 17 },
		{ 342, 161, 17 },
		{ 342, 162, 17 },
		{ 342, 163, 17 },
		{ 342, 164, 17 },
		{ 342, 165, 17 },
		{ 342, 166, 17 },
		{ 342, 167, 17 },
		{ 342, 168, 17 },
		{ 342, 169, 17 },
		{ 342, 170, 17 },
		{ 342, 171, 17 },
		{ 343, 131, 17 },
		{ 343, 132, 17 },
		{ 343, 133, 17 },
		{ 343, 134, 17 },
		{ 343, 135, 17 },
		{ 343, 136, 17 },
		{ 343, 137, 17 },
		{ 343, 138, 17 },
		{ 343, 139, 17 },
		{ 343, 140, 17 },
		{ 343, 141, 17 },
		{ 343, 142, 17 },
		{ 343, 143, 17 },
		{ 343, 144, 17 },
		{ 343, 145, 17 },
		{ 343, 146, 17 },
		{ 343, 147, 17 },
		{ 343, 148, 17 },
		{ 343, 149, 17 },
		{ 343, 150, 17 },
		{ 343, 151, 17 },
		{ 343, 152, 17 },
		{ 343, 153, 17 },
		{ 343, 154, 17 },
		{ 343, 155, 17 },
		{ 343, 156, 17 },
		{ 343, 157, 17 },
		{ 343, 158, 17 },
		{ 343, 159, 17 },
		{ 343, 160, 17 },
		{ 343, 161, 17 },
		{ 343, 162, 17 },
		{ 343, 163, 17 },
		{ 343, 164, 17 },
		{ 343, 165, 17 },
		{ 343, 166, 17 },
		{ 343, 167, 17 },
		{ 343, 168, 17 },
		{ 343, 169, 17 },
		{ 343, 170, 17 },
		{ 344, 131, 17 },
		{ 344, 132, 17 },
		{ 344, 133, 17 },
		{ 344, 134, 17 },
		{ 344, 135, 17 },
		{ 344, 136, 17 },
		{ 344, 137, 17 },
		{ 344, 138, 17 },
		{ 344, 139, 17 },
		{ 344, 140, 17 },
		{ 344, 141, 17 },
		{ 344, 142, 17 },
		{ 344, 143, 17 },
		{ 344, 144, 17 },
		{ 344, 145, 17 },
		{ 344, 146, 17 },
		{ 344, 147, 17 },
		{ 344, 148, 17 },
		{ 344, 149, 17 },
		{ 344, 150, 17 },
		{ 344, 151, 17 },
		{ 344, 152, 17 },
		{ 344, 153, 17 },
		{ 344, 154, 17 },
		{ 344, 155, 17 },
		{ 344, 156, 17 },
		{ 344, 157, 17 },
		{ 344, 158, 17 },
		{ 344, 159, 17 },
		{ 344, 160, 17 },
		{ 344, 161, 17 },
		{ 344, 162, 17 },
		{ 344, 163, 17 },
		{ 344, 164, 17 },
		{ 344, 165, 17 },
		{ 344, 166, 17 },
		{ 344, 167, 17 },
		{ 344, 168, 17 },
		{ 344, 169, 17 },
		{ 345, 131, 17 },
		{ 345, 132, 17 },
		{ 345, 133, 17 },
		{ 345, 134, 17 },
		{ 345, 135, 17 },
		{ 345, 136, 17 },
		{ 345, 137, 17 },
		{ 345, 138, 17 },
		{ 345, 139, 17 },
		{ 345, 140, 17 },
		{ 345, 141, 17 },
		{ 345, 142, 17 },
		{ 345, 143, 17 },
		{ 345, 144, 17 },
		{ 345, 145, 17 },
		{ 345, 146, 17 },
		{ 345, 147, 17 },
		{ 345, 148, 17 },
		{ 345, 149, 17 },
		{ 345, 150, 17 },
		{ 345, 151, 17 },
		{ 345, 152, 17 },
		{ 345, 153, 17 },
		{ 345, 154, 17 },
		{ 345, 155, 17 },
		{ 345, 156, 17 },
		{ 345, 157, 17 },
		{ 345, 158, 17 },
		{ 345, 159, 17 },
		{ 345, 160, 17 },
		{ 345, 161, 17 },
		{ 345, 162, 17 },
		{ 345, 163, 17 },
		{ 345, 164, 17 },
		{ 345, 165, 17 },
		{ 345, 166, 17 },
		{ 345, 167, 17 },
		{ 345, 168, 17 },
		{ 346, 131, 17 },
		{ 346, 132, 17 },
		{ 346, 133, 17 },
		{ 346, 134, 17 },
		{ 346, 135, 17 },
		{ 346, 136, 17 },
		{ 346, 137, 17 },
		{ 346, 138, 17 },
		{ 346, 139, 17 },
		{ 346, 140, 17 },
		{ 346, 141, 17 },
		{ 346, 142, 17 },
		{ 346, 143, 17 },
		{ 346, 144, 17 },
		{ 346, 145, 17 },
		{ 346, 146, 17 },
		{ 346, 147, 17 },
		{ 346, 148, 17 },
		{ 346, 149, 17 },
		{ 346, 150, 17 },
		{ 346, 151, 17 },
		{ 346, 152, 17 },
		{ 346, 153, 17 },
		{ 346, 154, 17 },
		{ 346, 155, 17 },
		{ 346, 156, 17 },
		{ 346, 157, 17 },
		{ 346, 158, 17 },
		{ 346, 159, 17 },
		{ 346, 160, 17 },
		{ 346, 161, 17 },
		{ 346, 162, 17 },
		{ 346, 163, 17 },
		{ 346, 164, 17 },
		{ 346, 165, 17 },
		{ 346, 166, 17 },
		{ 346, 167, 17 },
		{ 347, 131, 17 },
		{ 347, 132, 17 },
		{ 347, 133, 17 },
		{ 347, 134, 17 },
		{ 347, 135, 17 },
		{ 347, 136, 17 },
		{ 347, 137, 17 },
		{ 347, 138, 17 },
		{ 347, 139, 17 },
		{ 347, 140, 17 },
		{ 347, 141, 17 },
		{ 347, 142, 17 },
		{ 347, 143, 17 },
		{ 347, 144, 17 },
		{ 347, 145, 17 },
		{ 347, 146, 17 },
		{ 347, 147, 17 },
		{ 347, 148, 17 },
		{ 347, 149, 17 },
		{ 347, 150, 17 },
		{ 347, 151, 17 },
		{ 347, 152, 17 },
		{ 347, 153, 17 },
		{ 347, 154, 17 },
		{ 347, 155, 17 },
		{ 347, 156, 17 },
		{ 347, 157, 17 },
		{ 347, 158, 17 },
		{ 347, 159, 17 },
		{ 347, 160, 17 },
		{ 347, 161, 17 },
		{ 347, 162, 17 },
		{ 347, 163, 17 },
		{ 347, 164, 17 },
		{ 347, 165, 17 },
		{ 347, 166, 17 },
		{ 348, 131, 17 },
		{ 348, 132, 17 },
		{ 348, 133, 17 },
		{ 348, 134, 17 },
		{ 348, 135, 17 },
		{ 348, 136, 17 },
		{ 348, 137, 17 },
		{ 348, 138, 17 },
		{ 348, 139, 17 },
		{ 348, 140, 17 },
		{ 348, 141, 17 },
		{ 348, 142, 17 },
		{ 348, 143, 17 },
		{ 348, 144, 17 },
		{ 348, 145, 17 },
		{ 348, 146, 17 },
		{ 348, 147, 17 },
		{ 348, 148, 17 },
		{ 348, 149, 17 },
		{ 348, 150, 17 },
		{ 348, 151, 17 },
		{ 348, 152, 17 },
		{ 348, 153, 17 },
		{ 348, 154, 17 },
		{ 348, 155, 17 },
		{ 348, 156, 17 },
		{ 348, 157, 17 },
		{ 348, 158, 17 },
		{ 348, 159, 17 },
		{ 348, 160, 17 },
		{ 348, 161, 17 },
		{ 348, 162, 17 },
		{ 348, 163, 17 },
		{ 348, 164, 17 },
		{ 348, 165, 17 },
		{ 349, 131, 17 },
		{ 349, 132, 17 },
		{ 349, 133, 17 },
		{ 349, 134, 17 },
		{ 349, 135, 17 },
		{ 349, 136, 17 },
		{ 349, 137, 17 },
		{ 349, 138, 17 },
		{ 349, 139, 17 },
		{ 349, 140, 17 },
		{ 349, 141, 17 },
		{ 349, 142, 17 },
		{ 349, 143, 17 },
		{ 349, 144, 17 },
		{ 349, 145, 17 },
		{ 349, 146, 17 },
		{ 349, 147, 17 },
		{ 349, 148, 17 },
		{ 349, 149, 17 },
		{ 349, 150, 17 },
		{ 349, 151, 17 },
		{ 349, 152, 17 },
		{ 349, 153, 17 },
		{ 349, 154, 17 },
		{ 349, 155, 17 },
		{ 349, 156, 17 },
		{ 349, 157, 17 },
		{ 349, 158, 17 },
		{ 349, 159, 17 },
		{ 349, 160, 17 },
		{ 349, 161, 17 },
		{ 349, 162, 17 },
		{ 349, 163, 17 },
		{ 349, 164, 17 },
		{ 350, 131, 17 },
		{ 350, 132, 17 },
		{ 350, 133, 17 },
		{ 350, 134, 17 },
		{ 350, 135, 17 },
		{ 350, 136, 17 },
		{ 350, 137, 17 },
		{ 350, 138, 17 },
		{ 350, 139, 17 },
		{ 350, 140, 17 },
		{ 350, 141, 17 },
		{ 350, 142, 17 },
		{ 350, 143, 17 },
		{ 350, 144, 17 },
		{ 350, 145, 17 },
		{ 350, 146, 17 },
		{ 350, 147, 17 },
		{ 350, 148, 17 },
		{ 350, 149, 17 },
		{ 350, 150, 17 },
		{ 350, 151, 17 },
		{ 350, 152, 17 },
		{ 350, 153, 17 },
		{ 350, 154, 17 },
		{ 350, 155, 17 },
		{ 350, 156, 17 },
		{ 350, 157, 17 },
		{ 350, 158, 17 },
		{ 350, 159, 17 },
		{ 350, 160, 17 },
		{ 350, 161, 17 },
		{ 350, 162, 17 },
		{ 350, 163, 17 },
		{ 351, 131, 17 },
		{ 351, 132, 17 },
		{ 351, 133, 17 },
		{ 351, 134, 17 },
		{ 351, 135, 17 },
		{ 351, 136, 17 },
		{ 351, 137, 17 },
		{ 351, 138, 17 },
		{ 351, 139, 17 },
		{ 351, 140, 17 },
		{ 351, 141, 17 },
		{ 351, 142, 17 },
		{ 351, 143, 17 },
		{ 351, 144, 17 },
		{ 351, 145, 17 },
		{ 351, 146, 17 },
		{ 351, 147, 17 },
		{ 351, 148, 17 },
		{ 351, 149, 17 },
		{ 351, 150, 17 },
		{ 351, 151, 17 },
		{ 351, 152, 17 },
		{ 351, 153, 17 },
		{ 351, 154, 17 },
		{ 351, 155, 17 },
		{ 351, 156, 17 },
		{ 351, 157, 17 },
		{ 351, 158, 17 },
		{ 351, 159, 17 },
		{ 351, 160, 17 },
		{ 351, 161, 17 },
		{ 351, 162, 17 },
		{ 352, 131, 17 },
		{ 352, 132, 17 },
		{ 352, 133, 17 },
		{ 352, 134, 17 },
		{ 352, 135, 17 },
		{ 352, 136, 17 },
		{ 352, 137, 17 },
		{ 352, 138, 17 },
		{ 352, 139, 17 },
		{ 352, 140, 17 },
		{ 352, 141, 17 },
		{ 352, 142, 17 },
		{ 352, 143, 17 },
		{ 352, 144, 17 },
		{ 352, 145, 17 },
		{ 352, 146, 17 },
		{ 352, 147, 17 },
		{ 352, 148, 17 },
		{ 352, 149, 17 },
		{ 352, 150, 17 },
		{ 352, 151, 17 },
		{ 352, 152, 17 },
		{ 352, 153, 17 },
		{ 352, 154, 17 },
		{ 352, 155, 17 },
		{ 352, 156, 17 },
		{ 352, 157, 17 },
		{ 352, 158, 17 },
		{ 352, 159, 17 },
		{ 352, 160, 17 },
		{ 352, 161, 17 },
		{ 353, 131, 17 },
		{ 353, 132, 17 },
		{ 353, 133, 17 },
		{ 353, 134, 17 },
		{ 353, 135, 17 },
		{ 353, 136, 17 },
		{ 353, 137, 17 },
		{ 353, 138, 17 },
		{ 353, 139, 17 },
		{ 353, 140, 17 },
		{ 353, 141, 17 },
		{ 353, 142, 17 },
		{ 353, 143, 17 },
		{ 353, 144, 17 },
		{ 353, 145, 17 },
		{ 353, 146, 17 },
		{ 353, 147, 17 },
		{ 353, 148, 17 },
		{ 353, 149, 17 },
		{ 353, 150, 17 },
		{ 353, 151, 17 },
		{ 353, 152, 17 },
		{ 353, 153, 17 },
		{ 353, 154, 17 },
		{ 353, 155, 17 },
		{ 353, 156, 17 },
		{ 353, 157, 17 },
		{ 353, 158, 17 },
		{ 353, 159, 17 },
		{ 353, 160, 17 },
		{ 354, 131, 17 },
		{ 354, 132, 17 },
		{ 354, 133, 17 },
		{ 354, 134, 17 },
		{ 354, 135, 17 },
		{ 354, 136, 17 },
		{ 354, 137, 17 },
		{ 354, 138, 17 },
		{ 354, 139, 17 },
		{ 354, 140, 17 },
		{ 354, 141, 17 },
		{ 354, 142, 17 },
		{ 354, 143, 17 },
		{ 354, 144, 17 },
		{ 354, 145, 17 },
		{ 354, 146, 17 },
		{ 354, 147, 17 },
		{ 354, 148, 17 },
		{ 354, 149, 17 },
		{ 354, 150, 17 },
		{ 354, 151, 17 },
		{ 354, 152, 17 },
		{ 354, 153, 17 },
		{ 354, 154, 17 },
		{ 354, 155, 17 },
		{ 354, 156, 17 },
		{ 354, 157, 17 },
		{ 354, 158, 17 },
		{ 354, 159, 17 },
		{ 355, 131, 17 },
		{ 355, 132, 17 },
		{ 355, 133, 17 },
		{ 355, 134, 17 },
		{ 355, 135, 17 },
		{ 355, 136, 17 },
		{ 355, 137, 17 },
		{ 355, 138, 17 },
		{ 355, 139, 17 },
		{ 355, 140, 17 },
		{ 355, 141, 17 },
		{ 355, 142, 17 },
		{ 355, 143, 17 },
		{ 355, 144, 17 },
		{ 355, 145, 17 },
		{ 355, 146, 17 },
		{ 355, 147, 17 },
		{ 355, 148, 17 },
		{ 355, 149, 17 },
		{ 355, 150, 17 },
		{ 355, 151, 17 },
		{ 355, 152, 17 },
		{ 355, 153, 17 },
		{ 355, 154, 17 },
		{ 355, 155, 17 },
		{ 355, 156, 17 },
		{ 355, 157, 17 },
		{ 355, 158, 17 },
		{ 356, 131, 17 },
		{ 356, 132, 17 },
		{ 356, 133, 17 },
		{ 356, 134, 17 },
		{ 356, 135, 17 },
		{ 356, 136, 17 },
		{ 356, 137, 17 },
		{ 356, 138, 17 },
		{ 356, 139, 17 },
		{ 356, 140, 17 },
		{ 356, 141, 17 },
		{ 356, 142, 17 },
		{ 356, 143, 17 },
		{ 356, 144, 17 },
		{ 356, 145, 17 },
		{ 356, 146, 17 },
		{ 356, 147, 17 },
		{ 356, 148, 17 },
		{ 356, 149, 17 },
		{ 356, 150, 17 },
		{ 356, 151, 17 },
		{ 356, 152, 17 },
		{ 356, 153, 17 },
		{ 356, 154, 17 },
		{ 356, 155, 17 },
		{ 356, 156, 17 },
		{ 356, 157, 17 },
		{ 357, 131, 17 },
		{ 357, 132, 17 },
		{ 357, 133, 17 },
		{ 357, 134, 17 },
		{ 357, 135, 17 },
		{ 357, 136, 17 },
		{ 357, 137, 17 },
		{ 357, 138, 17 },
		{ 357, 139, 17 },
		{ 357, 140, 17 },
		{ 357, 141, 17 },
		{ 357, 142, 17 },
		{ 357, 143, 17 },
		{ 357, 144, 17 },
		{ 357, 145, 17 },
		{ 357, 146, 17 },
		{ 357, 147, 17 },
		{ 357, 148, 17 },
		{ 357, 149, 17 },
		{ 357, 150, 17 },
		{ 357, 151, 17 },
		{ 357, 152, 17 },
		{ 357, 153, 17 },
		{ 357, 154, 17 },
		{ 357, 155, 17 },
		{ 357, 156, 17 },
		{ 358, 131, 17 },
		{ 358, 132, 17 },
		{ 358, 133, 17 },
		{ 358, 134, 17 },
		{ 358, 135, 17 },
		{ 358, 136, 17 },
		{ 358, 137, 17 },
		{ 358, 138, 17 },
		{ 358, 139, 17 },
		{ 358, 140, 17 },
		{ 358, 141, 17 },
		{ 358, 142, 17 },
		{ 358, 143, 17 },
		{ 358, 144, 17 },
		{ 358, 145, 17 },
		{ 358, 146, 17 },
		{ 358, 147, 17 },
		{ 358, 148, 17 },
		{ 358, 149, 17 },
		{ 358, 150, 17 },
		{ 358, 151, 17 },
		{ 358, 152, 17 },
		{ 358, 153, 17 },
		{ 358, 154, 17 },
		{ 358, 155, 17 },
		{ 359, 131, 17 },
		{ 359, 132, 17 },
		{ 359, 133, 17 },
		{ 359, 134, 17 },
		{ 359, 135, 17 },
		{ 359, 136, 17 },
		{ 359, 137, 17 },
		{ 359, 138, 17 },
		{ 359, 139, 17 },
		{ 359, 140, 17 },
		{ 359, 141, 17 },
		{ 359, 142, 17 },
		{ 359, 143, 17 },
		{ 359, 144, 17 },
		{ 359, 145, 17 },
		{ 359, 146, 17 },
		{ 359, 147, 17 },
		{ 359, 148, 17 },
		{ 359, 149, 17 },
		{ 359, 150, 17 },
		{ 359, 151, 17 },
		{ 359, 152, 17 },
		{ 359, 153, 17 },
		{ 359, 154, 17 },
		{ 360, 131, 17 },
		{ 360, 132, 17 },
		{ 360, 133, 17 },
		{ 360, 134, 17 },
		{ 360, 135, 17 },
		{ 360, 136, 17 },
		{ 360, 137, 17 },
		{ 360, 138, 17 },
		{ 360, 139, 17 },
		{ 360, 140, 17 },
		{ 360, 141, 17 },
		{ 360, 142, 17 },
		{ 360, 143, 17 },
		{ 360, 144, 17 },
		{ 360, 145, 17 },
		{ 360, 146, 17 },
		{ 360, 147, 17 },
		{ 360, 148, 17 },
		{ 360, 149, 17 },
		{ 360, 150, 17 },
		{ 360, 151, 17 },
		{ 360, 152, 17 },
		{ 360, 153, 17 },
		{ 361, 131, 17 },
		{ 361, 132, 17 },
		{ 361, 133, 17 },
		{ 361, 134, 17 },
		{ 361, 135, 17 },
		{ 361, 136, 17 },
		{ 361, 137, 17 },
		{ 361, 138, 17 },
		{ 361, 139, 17 },
		{ 361, 140, 17 },
		{ 361, 141, 17 },
		{ 361, 142, 17 },
		{ 361, 143, 17 },
		{ 361, 144, 17 },
		{ 361, 145, 17 },
		{ 361, 146, 17 },
		{ 361, 147, 17 },
		{ 361, 148, 17 },
		{ 361, 149, 17 },
		{ 361, 150, 17 },
		{ 361, 151, 17 },
		{ 361, 152, 17 },
		{ 362, 131, 17 },
		{ 362, 132, 17 },
		{ 362, 133, 17 },
		{ 362, 134, 17 },
		{ 362, 135, 17 },
		{ 362, 136, 17 },
		{ 362, 137, 17 },
		{ 362, 138, 17 },
		{ 362, 139, 17 },
		{ 362, 140, 17 },
		{ 362, 141, 17 },
		{ 362, 142, 17 },
		{ 362, 143, 17 },
		{ 362, 144, 17 },
		{ 362, 145, 17 },
		{ 362, 146, 17 },
		{ 362, 147, 17 },
		{ 362, 148, 17 },
		{ 362, 149, 17 },
		{ 362, 150, 17 },
		{ 362, 151, 17 },
		{ 363, 131, 17 },
		{ 363, 132, 17 },
		{ 363, 133, 17 },
		{ 363, 134, 17 },
		{ 363, 135, 17 },
		{ 363, 136, 17 },
		{ 363, 137, 17 },
		{ 363, 138, 17 },
		{ 363, 139, 17 },
		{ 363, 140, 17 },
		{ 363, 141, 17 },
		{ 363, 142, 17 },
		{ 363, 143, 17 },
		{ 363, 144, 17 },
		{ 363, 145, 17 },
		{ 363, 146, 17 },
		{ 363, 147, 17 },
		{ 363, 148, 17 },
		{ 363, 149, 17 },
		{ 363, 150, 17 },
		{ 364, 131, 17 },
		{ 364, 132, 17 },
		{ 364, 133, 17 },
		{ 364, 134, 17 },
		{ 364, 135, 17 },
		{ 364, 136, 17 },
		{ 364, 137, 17 },
		{ 364, 138, 17 },
		{ 364, 139, 17 },
		{ 364, 140, 17 },
		{ 364, 141, 17 },
		{ 364, 142, 17 },
		{ 364, 143, 17 },
		{ 364, 144, 17 },
		{ 364, 145, 17 },
		{ 364, 146, 17 },
		{ 364, 147, 17 },
		{ 364, 148, 17 },
		{ 364, 149, 17 },
		{ 365, 131, 17 },
		{ 365, 132, 17 },
		{ 365, 133, 17 },
		{ 365, 134, 17 },
		{ 365, 135, 17 },
		{ 365, 136, 17 },
		{ 365, 137, 17 },
		{ 365, 138, 17 },
		{ 365, 139, 17 },
		{ 365, 140, 17 },
		{ 365, 141, 17 },
		{ 365, 142, 17 },
		{ 365, 143, 17 },
		{ 365, 144, 17 },
		{ 365, 145, 17 },
		{ 365, 146, 17 },
		{ 365, 147, 17 },
		{ 365, 148, 17 },
		{ 366, 131, 17 },
		{ 366, 132, 17 },
		{ 366, 133, 17 },
		{ 366, 134, 17 },
		{ 366, 135, 17 },
		{ 366, 136, 17 },
		{ 366, 137, 17 },
		{ 366, 138, 17 },
		{ 366, 139, 17 },
		{ 366, 140, 17 },
		{ 366, 141, 17 },
		{ 366, 142, 17 },
		{ 366, 143, 17 },
		{ 366, 144, 17 },
		{ 366, 145, 17 },
		{ 366, 146, 17 },
		{ 366, 147, 17 },
		{ 367, 131, 17 },
		{ 367, 132, 17 },
		{ 367, 133, 17 },
		{ 367, 134, 17 },
		{ 367, 135, 17 },
		{ 367, 136, 17 },
		{ 367, 137, 17 },
		{ 367, 138, 17 },
		{ 367, 139, 17 },
		{ 367, 140, 17 },
		{ 367, 141, 17 },
		{ 367, 142, 17 },
		{ 367, 143, 17 },
		{ 367, 144, 17 },
		{ 367, 145, 17 },
		{ 367, 146, 17 },
		{ 368, 131, 17 },
		{ 368, 132, 17 },
		{ 368, 133, 17 },
		{ 368, 134, 17 },
		{ 368, 135, 17 },
		{ 368, 136, 17 },
		{ 368, 137, 17 },
		{ 368, 138, 17 },
		{ 368, 139, 17 },
		{ 368, 140, 17 },
		{ 368, 141, 17 },
		{ 368, 142, 17 },
		{ 368, 143, 17 },
		{ 368, 144, 17 },
		{ 368, 145, 17 },
		{ 369, 131, 17 },
		{ 369, 132, 17 },
		{ 369, 133, 17 },
		{ 369, 134, 17 },
		{ 369, 135, 17 },
		{ 369, 136, 17 },
		{ 369, 137, 17 },
		{ 369, 138, 17 },
		{ 369, 139, 17 },
		{ 369, 140, 17 },
		{ 369, 141, 17 },
		{ 369, 142, 17 },
		{ 369, 143, 17 },
		{ 369, 144, 17 },
		{ 370, 131, 17 },
		{ 370, 132, 17 },
		{ 370, 133, 17 },
		{ 370, 134, 17 },
		{ 370, 135, 17 },
		{ 370, 136, 17 },
		{ 370, 137, 17 },
		{ 370, 138, 17 },
		{ 370, 139, 17 },
		{ 370, 140, 17 },
		{ 370, 141, 17 },
		{ 370, 142, 17 },
		{ 370, 143, 17 },
		{ 371, 131, 17 },
		{ 371, 132, 17 },
		{ 371, 133, 17 },
		{ 371, 134, 17 },
		{ 371, 135, 17 },
		{ 371, 136, 17 },
		{ 371, 137, 17 },
		{ 371, 138, 17 },
		{ 371, 139, 17 },
		{ 371, 140, 17 },
		{ 371, 141, 17 },
		{ 371, 142, 17 },
		{ 372, 131, 17 },
		{ 372, 132, 17 },
		{ 372, 133, 17 },
		{ 372, 134, 17 },
		{ 372, 135, 17 },
		{ 372, 136, 17 },
		{ 372, 137, 17 },
		{ 372, 138, 17 },
		{ 372, 139, 17 },
		{ 372, 140, 17 },
		{ 372, 141, 17 },
		{ 373, 131, 17 },
		{ 373, 132, 17 },
		{ 373, 133, 17 },
		{ 373, 134, 17 },
		{ 373, 135, 17 },
		{ 373, 136, 17 },
		{ 373, 137, 17 },
		{ 373, 138, 17 },
		{ 373, 139, 17 },
		{ 373, 140, 17 },
		{ 374, 131, 17 },
		{ 374, 132, 17 },
		{ 374, 133, 17 },
		{ 374, 134, 17 },
		{ 374, 135, 17 },
		{ 374, 136, 17 },
		{ 374, 137, 17 },
		{ 374, 138, 17 },
		{ 374, 139, 17 },
		{ 375, 131, 17 },
		{ 375, 132, 17 },
		{ 375, 133, 17 },
		{ 375, 134, 17 },
		{ 375, 135, 17 },
		{ 375, 136, 17 },
		{ 375, 137, 17 },
		{ 375, 138, 17 },
		{ 376, 131, 17 },
		{ 376, 132, 17 },
		{ 376, 133, 17 },
		{ 376, 134, 17 },
		{ 376, 135, 17 },
		{ 376, 136, 17 },
		{ 376, 137, 17 },
		{ 377, 132, 17 },
		{ 377, 133, 17 },
		{ 377, 134, 17 },
		{ 377, 135, 17 },
		{ 377, 136, 17 },
		{ 378, 132, 17 },
		{ 378, 133, 17 },
		{ 378, 134, 17 },
		{ 378, 135, 17 },
		{ 379, 132, 17 },
		{ 379, 133, 17 },
		{ 379, 134, 17 },
		{ 131, 359, 18 },
		{ 131, 360, 18 },
		{ 131, 361, 18 },
		{ 131, 362, 18 },
		{ 131, 363, 18 },
		{ 131, 364, 18 },
		{ 131, 365, 18 },
		{ 131, 366, 18 },
		{ 131, 367, 18 },
		{ 131, 368, 18 },
		{ 131, 369, 18 },
		{ 131, 370, 18 },
		{ 131, 371, 18 },
		{ 131, 372, 18 },
		{ 131, 373, 18 },
		{ 131, 374, 18 },
		{ 131, 375, 18 },
		{ 131, 376, 18 },
		{ 131, 377, 18 },
		{ 131, 378, 18 },
		{ 131, 379, 18 },
		{ 131, 380, 18 },
		{ 131, 381, 18 },
		{ 131, 382, 18 },
		{ 131, 383, 18 },
		{ 132, 354, 18 },
		{ 132, 355, 18 },
		{ 132, 356, 18 },
		{ 132, 357, 18 },
		{ 132, 358, 18 },
		{ 132, 359, 18 },
		{ 132, 360, 18 },
		{ 132, 361, 18 },
		{ 132, 362, 18 },
		{ 132, 363, 18 },
		{ 132, 364, 18 },
		{ 132, 365, 18 },
		{ 132, 366, 18 },
		{ 132, 367, 18 },
		{ 132, 368, 18 },
		{ 132, 369, 18 },
		{ 132, 370, 18 },
		{ 132, 371, 18 },
		{ 132, 372, 18 },
		{ 132, 373, 18 },
		{ 132, 374, 18 },
		{ 132, 375, 18 },
		{ 132, 376, 18 },
		{ 132, 377, 18 },
		{ 132, 378, 18 },
		{ 132, 379, 18 },
		{ 132, 380, 18 },
		{ 132, 381, 18 },
		{ 132, 382, 18 },
		{ 132, 383, 18 },
		{ 133, 352, 18 },
		{ 133, 353, 18 },
		{ 133, 354, 18 },
		{ 133, 355, 18 },
		{ 133, 356, 18 },
		{ 133, 357, 18 },
		{ 133, 358, 18 },
		{ 133, 359, 18 },
		{ 133, 360, 18 },
		{ 133, 361, 18 },
		{ 133, 362, 18 },
		{ 133, 363, 18 },
		{ 133, 364, 18 },
		{ 133, 365, 18 },
		{ 133, 366, 18 },
		{ 133, 367, 18 },
		{ 133, 368, 18 },
		{ 133, 369, 18 },
		{ 133, 370, 18 },
		{ 133, 371, 18 },
		{ 133, 372, 18 },
		{ 133, 373, 18 },
		{ 133, 374, 18 },
		{ 133, 375, 18 },
		{ 133, 376, 18 },
		{ 133, 377, 18 },
		{ 133, 378, 18 },
		{ 133, 379, 18 },
		{ 133, 380, 18 },
		{ 133, 381, 18 },
		{ 133, 382, 18 },
		{ 133, 383, 18 },
		{ 134, 351, 18 },
		{ 134, 352, 18 },
		{ 134, 353, 18 },
		{ 134, 354, 18 },
		{ 134, 355, 18 },
		{ 134, 356, 18 },
		{ 134, 357, 18 },
		{ 134, 358, 18 },
		{ 134, 359, 18 },
		{ 134, 360, 18 },
		{ 134, 361, 18 },
		{ 134, 362, 18 },
		{ 134, 363, 18 },
		{ 134, 364, 18 },
		{ 134, 365, 18 },
		{ 134, 366, 18 },
		{ 134, 367, 18 },
		{ 134, 368, 18 },
		{ 134, 369, 18 },
		{ 134, 370, 18 },
		{ 134, 371, 18 },
		{ 134, 372, 18 },
		{ 134, 373, 18 },
		{ 134, 374, 18 },
		{ 134, 375, 18 },
		{ 134, 376, 18 },
		{ 134, 377, 18 },
		{ 134, 378, 18 },
		{ 134, 379, 18 },
		{ 134, 380, 18 },
		{ 134, 381, 18 },
		{ 134, 382, 18 },
		{ 134, 383, 18 },
		{ 135, 350, 18 },
		{ 135, 351, 18 },
		{ 135, 352, 18 },
		{ 135, 353, 18 },
		{ 135, 354, 18 },
		{ 135, 355, 18 },
		{ 135, 356, 18 },
		{ 135, 357, 18 },
		{ 135, 358, 18 },
		{ 135, 359, 18 },
		{ 135, 360, 18 },
		{ 135, 361, 18 },
		{ 135, 362, 18 },
		{ 135, 363, 18 },
		{ 135, 364, 18 },
		{ 135, 365, 18 },
		{ 135, 366, 18 },
		{ 135, 367, 18 },
		{ 135, 368, 18 },
		{ 135, 369, 18 },
		{ 135, 370, 18 },
		{ 135, 371, 18 },
		{ 135, 372, 18 },
		{ 135, 373, 18 },
		{ 135, 374, 18 },
		{ 135, 375, 18 },
		{ 135, 376, 18 },
		{ 135, 377, 18 },
		{ 135, 378, 18 },
		{ 135, 379, 18 },
		{ 135, 380, 18 },
		{ 135, 381, 18 },
		{ 135, 382, 18 },
		{ 135, 383, 18 },
		{ 136, 349, 18 },
		{ 136, 350, 18 },
		{ 136, 351, 18 },
		{ 136, 352, 18 },
		{ 136, 353, 18 },
		{ 136, 354, 18 },
		{ 136, 355, 18 },
		{ 136, 356, 18 },
		{ 136, 357, 18 },
		{ 136, 358, 18 },
		{ 136, 359, 18 },
		{ 136, 360, 18 },
		{ 136, 361, 18 },
		{ 136, 362, 18 },
		{ 136, 363, 18 },
		{ 136, 364, 18 },
		{ 136, 365, 18 },
		{ 136, 366, 18 },
		{ 136, 367, 18 },
		{ 136, 368, 18 },
		{ 136, 369, 18 },
		{ 136, 370, 18 },
		{ 136, 371, 18 },
		{ 136, 372, 18 },
		{ 136, 373, 18 },
		{ 136, 374, 18 },
		{ 136, 375, 18 },
		{ 136, 376, 18 },
		{ 136, 377, 18 },
		{ 136, 378, 18 },
		{ 136, 379, 18 },
		{ 136, 380, 18 },
		{ 136, 381, 18 },
		{ 136, 382, 18 },
		{ 136, 383, 18 },
		{ 137, 348, 18 },
		{ 137, 349, 18 },
		{ 137, 350, 18 },
		{ 137, 351, 18 },
		{ 137, 352, 18 },
		{ 137, 353, 18 },
		{ 137, 354, 18 },
		{ 137, 355, 18 },
		{ 137, 356, 18 },
		{ 137, 357, 18 },
		{ 137, 358, 18 },
		{ 137, 359, 18 },
		{ 137, 360, 18 },
		{ 137, 361, 18 },
		{ 137, 362, 18 },
		{ 137, 363, 18 },
		{ 137, 364, 18 },
		{ 137, 365, 18 },
		{ 137, 366, 18 },
		{ 137, 367, 18 },
		{ 137, 368, 18 },
		{ 137, 369, 18 },
		{ 137, 370, 18 },
		{ 137, 371, 18 },
		{ 137, 372, 18 },
		{ 137, 373, 18 },
		{ 137, 374, 18 },
		{ 137, 375, 18 },
		{ 137, 376, 18 },
		{ 137, 377, 18 },
		{ 137, 378, 18 },
		{ 137, 379, 18 },
		{ 137, 380, 18 },
		{ 137, 381, 18 },
		{ 137, 382, 18 },
		{ 137, 383, 18 },
		{ 138, 347, 18 },
		{ 138, 348, 18 },
		{ 138, 349, 18 },
		{ 138, 350, 18 },
		{ 138, 351, 18 },
		{ 138, 352, 18 },
		{ 138, 353, 18 },
		{ 138, 354, 18 },
		{ 138, 355, 18 },
		{ 138, 356, 18 },
		{ 138, 357, 18 },
		{ 138, 358, 18 },
		{ 138, 359, 18 },
		{ 138, 360, 18 },
		{ 138, 361, 18 },
		{ 138, 362, 18 },
		{ 138, 363, 18 },
		{ 138, 364, 18 },
		{ 138, 365, 18 },
		{ 138, 366, 18 },
		{ 138, 367, 18 },
		{ 138, 368, 18 },
		{ 138, 369, 18 },
		{ 138, 370, 18 },
		{ 138, 371, 18 },
		{ 138, 372, 18 },
		{ 138, 373, 18 },
		{ 138, 374, 18 },
		{ 138, 375, 18 },
		{ 138, 376, 18 },
		{ 138, 377, 18 },
		{ 138, 378, 18 },
		{ 138, 379, 18 },
		{ 138, 380, 18 },
		{ 138, 381, 18 },
		{ 138, 382, 18 },
		{ 138, 383, 18 },
		{ 139, 346, 18 },
		{ 139, 347, 18 },
		{ 139, 348, 18 },
		{ 139, 349, 18 },
		{ 139, 350, 18 },
		{ 139, 351, 18 },
		{ 139, 352, 18 },
		{ 139, 353, 18 },
		{ 139, 354, 18 },
		{ 139, 355, 18 },
		{ 139, 356, 18 },
		{ 139, 357, 18 },
		{ 139, 358, 18 },
		{ 139, 359, 18 },
		{ 139, 360, 18 },
		{ 139, 361, 18 },
		{ 139, 362, 18 },
		{ 139, 363, 18 },
		{ 139, 364, 18 },
		{ 139, 365, 18 },
		{ 139, 366, 18 },
		{ 139, 367, 18 },
		{ 139, 368, 18 },
		{ 139, 369, 18 },
		{ 139, 370, 18 },
		{ 139, 371, 18 },
		{ 139, 372, 18 },
		{ 139, 373, 18 },
		{ 139, 374, 18 },
		{ 139, 375, 18 },
		{ 139, 376, 18 },
		{ 139, 377, 18 },
		{ 139, 378, 18 },
		{ 139, 379, 18 },
		{ 139, 380, 18 },
		{ 139, 381, 18 },
		{ 139, 382, 18 },
		{ 139, 383, 18 },
		{ 140, 345, 18 },
		{ 140, 346, 18 },
		{ 140, 347, 18 },
		{ 140, 348, 18 },
		{ 140, 349, 18 },
		{ 140, 350, 18 },
		{ 140, 351, 18 },
		{ 140, 352, 18 },
		{ 140, 353, 18 },
		{ 140, 354, 18 },
		{ 140, 355, 18 },
		{ 140, 356, 18 },
		{ 140, 357, 18 },
		{ 140, 358, 18 },
		{ 140, 359, 18 },
		{ 140, 360, 18 },
		{ 140, 361, 18 },
		{ 140, 362, 18 },
		{ 140, 363, 18 },
		{ 140, 364, 18 },
		{ 140, 365, 18 },
		{ 140, 366, 18 },
		{ 140, 367, 18 },
		{ 140, 368, 18 },
		{ 140, 369, 18 },
		{ 140, 370, 18 },
		{ 140, 371, 18 },
		{ 140, 372, 18 },
		{ 140, 373, 18 },
		{ 140, 374, 18 },
		{ 140, 375, 18 },
		{ 140, 376, 18 },
		{ 140, 377, 18 },
		{ 140, 378, 18 },
		{ 140, 379, 18 },
		{ 140, 380, 18 },
		{ 140, 381, 18 },
		{ 140, 382, 18 },
		{ 140, 383, 18 },
		{ 141, 345, 18 },
		{ 141, 346, 18 },
		{ 141, 347, 18 },
		{ 141, 348, 18 },
		{ 141, 349, 18 },
		{ 141, 350, 18 },
		{ 141, 351, 18 },
		{ 141, 352, 18 },
		{ 141, 353, 18 },
		{ 141, 354, 18 },
		{ 141, 355, 18 },
		{ 141, 356, 18 },
		{ 141, 357, 18 },
		{ 141, 358, 18 },
		{ 141, 359, 18 },
		{ 141, 360, 18 },
		{ 141, 361, 18 },
		{ 141, 362, 18 },
		{ 141, 363, 18 },
		{ 141, 364, 18 },
		{ 141, 365, 18 },
		{ 141, 366, 18 },
		{ 141, 367, 18 },
		{ 141, 368, 18 },
		{ 141, 369, 18 },
		{ 141, 370, 18 },
		{ 141, 371, 18 },
		{ 141, 372, 18 },
		{ 141, 373, 18 },
		{ 141, 374, 18 },
		{ 141, 375, 18 },
		{ 141, 376, 18 },
		{ 141, 377, 18 },
		{ 141, 378, 18 },
		{ 141, 379, 18 },
		{ 141, 380, 18 },
		{ 141, 381, 18 },
		{ 141, 382, 18 },
		{ 141, 383, 18 },
		{ 142, 344, 18 },
		{ 142, 345, 18 },
		{ 142, 346, 18 },
		{ 142, 347, 18 },
		{ 142, 348, 18 },
		{ 142, 349, 18 },
		{ 142, 350, 18 },
		{ 142, 351, 18 },
		{ 142, 352, 18 },
		{ 142, 353, 18 },
		{ 142, 354, 18 },
		{ 142, 355, 18 },
		{ 142, 356, 18 },
		{ 142, 357, 18 },
		{ 142, 358, 18 },
		{ 142, 359, 18 },
		{ 142, 360, 18 },
		{ 142, 361, 18 },
		{ 142, 362, 18 },
		{ 142, 363, 18 },
		{ 142, 364, 18 },
		{ 142, 365, 18 },
		{ 142, 366, 18 },
		{ 142, 367, 18 },
		{ 142, 368, 18 },
		{ 142, 369, 18 },
		{ 142, 370, 18 },
		{ 142, 371, 18 },
		{ 142, 372, 18 },
		{ 142, 373, 18 },
		{ 142, 374, 18 },
		{ 142, 375, 18 },
		{ 142, 376, 18 },
		{ 142, 377, 18 },
		{ 142, 378, 18 },
		{ 142, 379, 18 },
		{ 142, 380, 18 },
		{ 142, 381, 18 },
		{ 142, 382, 18 },
		{ 142, 383, 18 },
		{ 143, 343, 18 },
		{ 143, 344, 18 },
		{ 143, 345, 18 },
		{ 143, 346, 18 },
		{ 143, 347, 18 },
		{ 143, 348, 18 },
		{ 143, 349, 18 },
		{ 143, 350, 18 },
		{ 143, 351, 18 },
		{ 143, 352, 18 },
		{ 143, 353, 18 },
		{ 143, 354, 18 },
		{ 143, 355, 18 },
		{ 143, 356, 18 },
		{ 143, 357, 18 },
		{ 143, 358, 18 },
		{ 143, 359, 18 },
		{ 143, 360, 18 },
		{ 143, 361, 18 },
		{ 143, 362, 18 },
		{ 143, 363, 18 },
		{ 143, 364, 18 },
		{ 143, 365, 18 },
		{ 143, 366, 18 },
		{ 143, 367, 18 },
		{ 143, 368, 18 },
		{ 143, 369, 18 },
		{ 143, 370, 18 },
		{ 143, 371, 18 },
		{ 143, 372, 18 },
		{ 143, 373, 18 },
		{ 143, 374, 18 },
		{ 143, 375, 18 },
		{ 143, 376, 18 },
		{ 143, 377, 18 },
		{ 143, 378, 18 },
		{ 143, 379, 18 },
		{ 143, 380, 18 },
		{ 143, 381, 18 },
		{ 143, 382, 18 },
		{ 143, 383, 18 },
		{ 144, 342, 18 },
		{ 144, 343, 18 },
		{ 144, 344, 18 },
		{ 144, 345, 18 },
		{ 144, 346, 18 },
		{ 144, 347, 18 },
		{ 144, 348, 18 },
		{ 144, 349, 18 },
		{ 144, 350, 18 },
		{ 144, 351, 18 },
		{ 144, 352, 18 },
		{ 144, 353, 18 },
		{ 144, 354, 18 },
		{ 144, 355, 18 },
		{ 144, 356, 18 },
		{ 144, 357, 18 },
		{ 144, 358, 18 },
		{ 144, 359, 18 },
		{ 144, 360, 18 },
		{ 144, 361, 18 },
		{ 144, 362, 18 },
		{ 144, 363, 18 },
		{ 144, 364, 18 },
		{ 144, 365, 18 },
		{ 144, 366, 18 },
		{ 144, 367, 18 },
		{ 144, 368, 18 },
		{ 144, 369, 18 },
		{ 144, 370, 18 },
		{ 144, 371, 18 },
		{ 144, 372, 18 },
		{ 144, 373, 18 },
		{ 144, 374, 18 },
		{ 144, 375, 18 },
		{ 144, 376, 18 },
		{ 144, 377, 18 },
		{ 144, 378, 18 },
		{ 144, 379, 18 },
		{ 144, 380, 18 },
		{ 144, 381, 18 },
		{ 144, 382, 18 },
		{ 144, 383, 18 },
		{ 145, 342, 18 },
		{ 145, 343, 18 },
		{ 145, 344, 18 },
		{ 145, 345, 18 },
		{ 145, 346, 18 },
		{ 145, 347, 18 },
		{ 145, 348, 18 },
		{ 145, 349, 18 },
		{ 145, 350, 18 },
		{ 145, 351, 18 },
		{ 145, 352, 18 },
		{ 145, 353, 18 },
		{ 145, 354, 18 },
		{ 145, 355, 18 },
		{ 145, 356, 18 },
		{ 145, 357, 18 },
		{ 145, 358, 18 },
		{ 145, 359, 18 },
		{ 145, 360, 18 },
		{ 145, 361, 18 },
		{ 145, 362, 18 },
		{ 145, 363, 18 },
		{ 145, 364, 18 },
		{ 145, 365, 18 },
		{ 145, 366, 18 },
		{ 145, 367, 18 },
		{ 145, 368, 18 },
		{ 145, 369, 18 },
		{ 145, 370, 18 },
		{ 145, 371, 18 },
		{ 145, 372, 18 },
		{ 145, 373, 18 },
		{ 145, 374, 18 },
		{ 145, 375, 18 },
		{ 145, 376, 18 },
		{ 145, 377, 18 },
		{ 145, 378, 18 },
		{ 145, 379, 18 },
		{ 145, 380, 18 },
		{ 145, 381, 18 },
		{ 145, 382, 18 },
		{ 145, 383, 18 },
		{ 146, 341, 18 },
		{ 146, 342, 18 },
		{ 146, 343, 18 },
		{ 146, 344, 18 },
		{ 146, 345, 18 },
		{ 146, 346, 18 },
		{ 146, 347, 18 },
		{ 146, 348, 18 },
		{ 146, 349, 18 },
		{ 146, 350, 18 },
		{ 146, 351, 18 },
		{ 146, 352, 18 },
		{ 146, 353, 18 },
		{ 146, 354, 18 },
		{ 146, 355, 18 },
		{ 146, 356, 18 },
		{ 146, 357, 18 },
		{ 146, 358, 18 },
		{ 146, 359, 18 },
		{ 146, 360, 18 },
		{ 146, 361, 18 },
		{ 146, 362, 18 },
		{ 146, 363, 18 },
		{ 146, 364, 18 },
		{ 146, 365, 18 },
		{ 146, 366, 18 },
		{ 146, 367, 18 },
		{ 146, 368, 18 },
		{ 146, 369, 18 },
		{ 146, 370, 18 },
		{ 146, 371, 18 },
		{ 146, 372, 18 },
		{ 146, 373, 18 },
		{ 146, 374, 18 },
		{ 146, 375, 18 },
		{ 146, 376, 18 },
		{ 146, 377, 18 },
		{ 146, 378, 18 },
		{ 146, 379, 18 },
		{ 146, 380, 18 },
		{ 146, 381, 18 },
		{ 146, 382, 18 },
		{ 146, 383, 18 },
		{ 147, 340, 18 },
		{ 147, 341, 18 },
		{ 147, 342, 18 },
		{ 147, 343, 18 },
		{ 147, 344, 18 },
		{ 147, 345, 18 },
		{ 147, 346, 18 },
		{ 147, 347, 18 },
		{ 147, 348, 18 },
		{ 147, 349, 18 },
		{ 147, 350, 18 },
		{ 147, 351, 18 },
		{ 147, 352, 18 },
		{ 147, 353, 18 },
		{ 147, 354, 18 },
		{ 147, 355, 18 },
		{ 147, 356, 18 },
		{ 147, 357, 18 },
		{ 147, 358, 18 },
		{ 147, 359, 18 },
		{ 147, 360, 18 },
		{ 147, 361, 18 },
		{ 147, 362, 18 },
		{ 147, 363, 18 },
		{ 147, 364, 18 },
		{ 147, 365, 18 },
		{ 147, 366, 18 },
		{ 147, 367, 18 },
		{ 147, 368, 18 },
		{ 147, 369, 18 },
		{ 147, 370, 18 },
		{ 147, 371, 18 },
		{ 147, 372, 18 },
		{ 147, 373, 18 },
		{ 147, 374, 18 },
		{ 147, 375, 18 },
		{ 147, 376, 18 },
		{ 147, 377, 18 },
		{ 147, 378, 18 },
		{ 147, 379, 18 },
		{ 147, 380, 18 },
		{ 147, 381, 18 },
		{ 147, 382, 18 },
		{ 147, 383, 18 },
		{ 148, 339, 18 },
		{ 148, 340, 18 },
		{ 148, 341, 18 },
		{ 148, 342, 18 },
		{ 148, 343, 18 },
		{ 148, 344, 18 },
		{ 148, 345, 18 },
		{ 148, 346, 18 },
		{ 148, 347, 18 },
		{ 148, 348, 18 },
		{ 148, 349, 18 },
		{ 148, 350, 18 },
		{ 148, 351, 18 },
		{ 148, 352, 18 },
		{ 148, 353, 18 },
		{ 148, 354, 18 },
		{ 148, 355, 18 },
		{ 148, 356, 18 },
		{ 148, 357, 18 },
		{ 148, 358, 18 },
		{ 148, 359, 18 },
		{ 148, 360, 18 },
		{ 148, 361, 18 },
		{ 148, 362, 18 },
		{ 148, 363, 18 },
		{ 148, 364, 18 },
		{ 148, 365, 18 },
		{ 148, 366, 18 },
		{ 148, 367, 18 },
		{ 148, 368, 18 },
		{ 148, 369, 18 },
		{ 148, 370, 18 },
		{ 148, 371, 18 },
		{ 148, 372, 18 },
		{ 148, 373, 18 },
		{ 148, 374, 18 },
		{ 148, 375, 18 },
		{ 148, 376, 18 },
		{ 148, 377, 18 },
		{ 148, 378, 18 },
		{ 148, 379, 18 },
		{ 148, 380, 18 },
		{ 148, 381, 18 },
		{ 148, 382, 18 },
		{ 148, 383, 18 },
		{ 149, 339, 18 },
		{ 149, 340, 18 },
		{ 149, 341, 18 },
		{ 149, 342, 18 },
		{ 149, 343, 18 },
		{ 149, 344, 18 },
		{ 149, 345, 18 },
		{ 149, 346, 18 },
		{ 149, 347, 18 },
		{ 149, 348, 18 },
		{ 149, 349, 18 },
		{ 149, 350, 18 },
		{ 149, 351, 18 },
		{ 149, 352, 18 },
		{ 149, 353, 18 },
		{ 149, 354, 18 },
		{ 149, 355, 18 },
		{ 149, 356, 18 },
		{ 149, 357, 18 },
		{ 149, 358, 18 },
		{ 149, 359, 18 },
		{ 149, 360, 18 },
		{ 149, 361, 18 },
		{ 149, 362, 18 },
		{ 149, 363, 18 },
		{ 149, 364, 18 },
		{ 149, 365, 18 },
		{ 149, 366, 18 },
		{ 149, 367, 18 },
		{ 149, 368, 18 },
		{ 149, 369, 18 },
		{ 149, 370, 18 },
		{ 149, 371, 18 },
		{ 149, 372, 18 },
		{ 149, 373, 18 },
		{ 149, 374, 18 },
		{ 149, 375, 18 },
		{ 149, 376, 18 },
		{ 149, 377, 18 },
		{ 149, 378, 18 },
		{ 149, 379, 18 },
		{ 149, 380, 18 },
		{ 149, 381, 18 },
		{ 149, 382, 18 },
		{ 149, 383, 18 },
		{ 150, 338, 18 },
		{ 150, 339, 18 },
		{ 150, 340, 18 },
		{ 150, 341, 18 },
		{ 150, 342, 18 },
		{ 150, 343, 18 },
		{ 150, 344, 18 },
		{ 150, 345, 18 },
		{ 150, 346, 18 },
		{ 150, 347, 18 },
		{ 150, 348, 18 },
		{ 150, 349, 18 },
		{ 150, 350, 18 },
		{ 150, 351, 18 },
		{ 150, 352, 18 },
		{ 150, 353, 18 },
		{ 150, 354, 18 },
		{ 150, 355, 18 },
		{ 150, 356, 18 },
		{ 150, 357, 18 },
		{ 150, 358, 18 },
		{ 150, 359, 18 },
		{ 150, 360, 18 },
		{ 150, 361, 18 },
		{ 150, 362, 18 },
		{ 150, 363, 18 },
		{ 150, 364, 18 },
		{ 150, 365, 18 },
		{ 150, 366, 18 },
		{ 150, 367, 18 },
		{ 150, 368, 18 },
		{ 150, 369, 18 },
		{ 150, 370, 18 },
		{ 150, 371, 18 },
		{ 150, 372, 18 },
		{ 150, 373, 18 },
		{ 150, 374, 18 },
		{ 150, 375, 18 },
		{ 150, 376, 18 },
		{ 150, 377, 18 },
		{ 150, 378, 18 },
		{ 150, 379, 18 },
		{ 150, 380, 18 },
		{ 150, 381, 18 },
		{ 150, 382, 18 },
		{ 150, 383, 18 },
		{ 151, 337, 18 },
		{ 151, 338, 18 },
		{ 151, 339, 18 },
		{ 151, 340, 18 },
		{ 151, 341, 18 },
		{ 151, 342, 18 },
		{ 151, 343, 18 },
		{ 151, 344, 18 },
		{ 151, 345, 18 },
		{ 151, 346, 18 },
		{ 151, 347, 18 },
		{ 151, 348, 18 },
		{ 151, 349, 18 },
		{ 151, 350, 18 },
		{ 151, 351, 18 },
		{ 151, 352, 18 },
		{ 151, 353, 18 },
		{ 151, 354, 18 },
		{ 151, 355, 18 },
		{ 151, 356, 18 },
		{ 151, 357, 18 },
		{ 151, 358, 18 },
		{ 151, 359, 18 },
		{ 151, 360, 18 },
		{ 151, 361, 18 },
		{ 151, 362, 18 },
		{ 151, 363, 18 },
		{ 151, 364, 18 },
		{ 151, 365, 18 },
		{ 151, 366, 18 },
		{ 151, 367, 18 },
		{ 151, 368, 18 },
		{ 151, 369, 18 },
		{ 151, 370, 18 },
		{ 151, 371, 18 },
		{ 151, 372, 18 },
		{ 151, 373, 18 },
		{ 151, 374, 18 },
		{ 151, 375, 18 },
		{ 151, 376, 18 },
		{ 151, 377, 18 },
		{ 151, 378, 18 },
		{ 151, 379, 18 },
		{ 151, 380, 18 },
		{ 151, 381, 18 },
		{ 151, 382, 18 },
		{ 151, 383, 18 },
		{ 152, 336, 18 },
		{ 152, 337, 18 },
		{ 152, 338, 18 },
		{ 152, 339, 18 },
		{ 152, 340, 18 },
		{ 152, 341, 18 },
		{ 152, 342, 18 },
		{ 152, 343, 18 },
		{ 152, 344, 18 },
		{ 152, 345, 18 },
		{ 152, 346, 18 },
		{ 152, 347, 18 },
		{ 152, 348, 18 },
		{ 152, 349, 18 },
		{ 152, 350, 18 },
		{ 152, 351, 18 },
		{ 152, 352, 18 },
		{ 152, 353, 18 },
		{ 152, 354, 18 },
		{ 152, 355, 18 },
		{ 152, 356, 18 },
		{ 152, 357, 18 },
		{ 152, 358, 18 },
		{ 152, 359, 18 },
		{ 152, 360, 18 },
		{ 152, 361, 18 },
		{ 152, 362, 18 },
		{ 152, 363, 18 },
		{ 152, 364, 18 },
		{ 152, 365, 18 },
		{ 152, 366, 18 },
		{ 152, 367, 18 },
		{ 152, 368, 18 },
		{ 152, 369, 18 },
		{ 152, 370, 18 },
		{ 152, 371, 18 },
		{ 152, 372, 18 },
		{ 152, 373, 18 },
		{ 152, 374, 18 },
		{ 152, 375, 18 },
		{ 152, 376, 18 },
		{ 152, 377, 18 },
		{ 152, 378, 18 },
		{ 152, 379, 18 },
		{ 152, 380, 18 },
		{ 152, 381, 18 },
		{ 152, 382, 18 },
		{ 152, 383, 18 },
		{ 153, 336, 18 },
		{ 153, 337, 18 },
		{ 153, 338, 18 },
		{ 153, 339, 18 },
		{ 153, 340, 18 },
		{ 153, 341, 18 },
		{ 153, 342, 18 },
		{ 153, 343, 18 },
		{ 153, 344, 18 },
		{ 153, 345, 18 },
		{ 153, 346, 18 },
		{ 153, 347, 18 },
		{ 153, 348, 18 },
		{ 153, 349, 18 },
		{ 153, 350, 18 },
		{ 153, 351, 18 },
		{ 153, 352, 18 },
		{ 153, 353, 18 },
		{ 153, 354, 18 },
		{ 153, 355, 18 },
		{ 153, 356, 18 },
		{ 153, 357, 18 },
		{ 153, 358, 18 },
		{ 153, 359, 18 },
		{ 153, 360, 18 },
		{ 153, 361, 18 },
		{ 153, 362, 18 },
		{ 153, 363, 18 },
		{ 153, 364, 18 },
		{ 153, 365, 18 },
		{ 153, 366, 18 },
		{ 153, 367, 18 },
		{ 153, 368, 18 },
		{ 153, 369, 18 },
		{ 153, 370, 18 },
		{ 153, 371, 18 },
		{ 153, 372, 18 },
		{ 153, 373, 18 },
		{ 153, 374, 18 },
		{ 153, 375, 18 },
		{ 153, 376, 18 },
		{ 153, 377, 18 },
		{ 153, 378, 18 },
		{ 153, 379, 18 },
		{ 153, 380, 18 },
		{ 153, 381, 18 },
		{ 153, 382, 18 },
		{ 153, 383, 18 },
		{ 154, 335, 18 },
		{ 154, 336, 18 },
		{ 154, 337, 18 },
		{ 154, 338, 18 },
		{ 154, 339, 18 },
		{ 154, 340, 18 },
		{ 154, 341, 18 },
		{ 154, 342, 18 },
		{ 154, 343, 18 },
		{ 154, 344, 18 },
		{ 154, 345, 18 },
		{ 154, 346, 18 },
		{ 154, 347, 18 },
		{ 154, 348, 18 },
		{ 154, 349, 18 },
		{ 154, 350, 18 },
		{ 154, 351, 18 },
		{ 154, 352, 18 },
		{ 154, 353, 18 },
		{ 154, 354, 18 },
		{ 154, 355, 18 },
		{ 154, 356, 18 },
		{ 154, 357, 18 },
		{ 154, 358, 18 },
		{ 154, 359, 18 },
		{ 154, 360, 18 },
		{ 154, 361, 18 },
		{ 154, 362, 18 },
		{ 154, 363, 18 },
		{ 154, 364, 18 },
		{ 154, 365, 18 },
		{ 154, 366, 18 },
		{ 154, 367, 18 },
		{ 154, 368, 18 },
		{ 154, 369, 18 },
		{ 154, 370, 18 },
		{ 154, 371, 18 },
		{ 154, 372, 18 },
		{ 154, 373, 18 },
		{ 154, 374, 18 },
		{ 154, 375, 18 },
		{ 154, 376, 18 },
		{ 154, 377, 18 },
		{ 154, 378, 18 },
		{ 154, 379, 18 },
		{ 154, 380, 18 },
		{ 154, 381, 18 },
		{ 154, 382, 18 },
		{ 154, 383, 18 },
		{ 155, 334, 18 },
		{ 155, 335, 18 },
		{ 155, 336, 18 },
		{ 155, 337, 18 },
		{ 155, 338, 18 },
		{ 155, 339, 18 },
		{ 155, 340, 18 },
		{ 155, 341, 18 },
		{ 155, 342, 18 },
		{ 155, 343, 18 },
		{ 155, 344, 18 },
		{ 155, 345, 18 },
		{ 155, 346, 18 },
		{ 155, 347, 18 },
		{ 155, 348, 18 },
		{ 155, 349, 18 },
		{ 155, 350, 18 },
		{ 155, 351, 18 },
		{ 155, 352, 18 },
		{ 155, 353, 18 },
		{ 155, 354, 18 },
		{ 155, 355, 18 },
		{ 155, 356, 18 },
		{ 155, 357, 18 },
		{ 155, 358, 18 },
		{ 155, 359, 18 },
		{ 155, 360, 18 },
		{ 155, 361, 18 },
		{ 155, 362, 18 },
		{ 155, 363, 18 },
		{ 155, 364, 18 },
		{ 155, 365, 18 },
		{ 155, 366, 18 },
		{ 155, 367, 18 },
		{ 155, 368, 18 },
		{ 155, 369, 18 },
		{ 155, 370, 18 },
		{ 155, 371, 18 },
		{ 155, 372, 18 },
		{ 155, 373, 18 },
		{ 155, 374, 18 },
		{ 155, 375, 18 },
		{ 155, 376, 18 },
		{ 155, 377, 18 },
		{ 155, 378, 18 },
		{ 155, 379, 18 },
		{ 155, 380, 18 },
		{ 155, 381, 18 },
		{ 155, 382, 18 },
		{ 155, 383, 18 },
		{ 156, 333, 18 },
		{ 156, 334, 18 },
		{ 156, 335, 18 },
		{ 156, 336, 18 },
		{ 156, 337, 18 },
		{ 156, 338, 18 },
		{ 156, 339, 18 },
		{ 156, 340, 18 },
		{ 156, 341, 18 },
		{ 156, 342, 18 },
		{ 156, 343, 18 },
		{ 156, 344, 18 },
		{ 156, 345, 18 },
		{ 156, 346, 18 },
		{ 156, 347, 18 },
		{ 156, 348, 18 },
		{ 156, 349, 18 },
		{ 156, 350, 18 },
		{ 156, 351, 18 },
		{ 156, 352, 18 },
		{ 156, 353, 18 },
		{ 156, 354, 18 },
		{ 156, 355, 18 },
		{ 156, 356, 18 },
		{ 156, 357, 18 },
		{ 156, 358, 18 },
		{ 156, 359, 18 },
		{ 156, 360, 18 },
		{ 156, 361, 18 },
		{ 156, 362, 18 },
		{ 156, 363, 18 },
		{ 156, 364, 18 },
		{ 156, 365, 18 },
		{ 156, 366, 18 },
		{ 156, 367, 18 },
		{ 156, 368, 18 },
		{ 156, 369, 18 },
		{ 156, 370, 18 },
		{ 156, 371, 18 },
		{ 156, 372, 18 },
		{ 156, 373, 18 },
		{ 156, 374, 18 },
		{ 156, 375, 18 },
		{ 156, 376, 18 },
		{ 156, 377, 18 },
		{ 156, 378, 18 },
		{ 156, 379, 18 },
		{ 156, 380, 18 },
		{ 156, 381, 18 },
		{ 156, 382, 18 },
		{ 157, 333, 18 },
		{ 157, 334, 18 },
		{ 157, 335, 18 },
		{ 157, 336, 18 },
		{ 157, 337, 18 },
		{ 157, 338, 18 },
		{ 157, 339, 18 },
		{ 157, 340, 18 },
		{ 157, 341, 18 },
		{ 157, 342, 18 },
		{ 157, 343, 18 },
		{ 157, 344, 18 },
		{ 157, 345, 18 },
		{ 157, 346, 18 },
		{ 157, 347, 18 },
		{ 157, 348, 18 },
		{ 157, 349, 18 },
		{ 157, 350, 18 },
		{ 157, 351, 18 },
		{ 157, 352, 18 },
		{ 157, 353, 18 },
		{ 157, 354, 18 },
		{ 157, 355, 18 },
		{ 157, 356, 18 },
		{ 157, 357, 18 },
		{ 157, 358, 18 },
		{ 157, 359, 18 },
		{ 157, 360, 18 },
		{ 157, 361, 18 },
		{ 157, 362, 18 },
		{ 157, 363, 18 },
		{ 157, 364, 18 },
		{ 157, 365, 18 },
		{ 157, 366, 18 },
		{ 157, 367, 18 },
		{ 157, 368, 18 },
		{ 157, 369, 18 },
		{ 157, 370, 18 },
		{ 157, 371, 18 },
		{ 157, 372, 18 },
		{ 157, 373, 18 },
		{ 157, 374, 18 },
		{ 157, 375, 18 },
		{ 157, 376, 18 },
		{ 157, 377, 18 },
		{ 157, 378, 18 },
		{ 157, 379, 18 },
		{ 157, 380, 18 },
		{ 157, 381, 18 },
		{ 157, 382, 18 },
		{ 158, 332, 18 },
		{ 158, 333, 18 },
		{ 158, 334, 18 },
		{ 158, 335, 18 },
		{ 158, 336, 18 },
		{ 158, 337, 18 },
		{ 158, 338, 18 },
		{ 158, 339, 18 },
		{ 158, 340, 18 },
		{ 158, 341, 18 },
		{ 158, 342, 18 },
		{ 158, 343, 18 },
		{ 158, 344, 18 },
		{ 158, 345, 18 },
		{ 158, 346, 18 },
		{ 158, 347, 18 },
		{ 158, 348, 18 },
		{ 158, 349, 18 },
		{ 158, 350, 18 },
		{ 158, 351, 18 },
		{ 158, 352, 18 },
		{ 158, 353, 18 },
		{ 158, 354, 18 },
		{ 158, 355, 18 },
		{ 158, 356, 18 },
		{ 158, 357, 18 },
		{ 158, 358, 18 },
		{ 158, 359, 18 },
		{ 158, 360, 18 },
		{ 158, 361, 18 },
		{ 158, 362, 18 },
		{ 158, 363, 18 },
		{ 158, 364, 18 },
		{ 158, 365, 18 },
		{ 158, 366, 18 },
		{ 158, 367, 18 },
		{ 158, 368, 18 },
		{ 158, 369, 18 },
		{ 158, 370, 18 },
		{ 158, 371, 18 },
		{ 158, 372, 18 },
		{ 158, 373, 18 },
		{ 158, 374, 18 },
		{ 158, 375, 18 },
		{ 158, 376, 18 },
		{ 158, 377, 18 },
		{ 158, 378, 18 },
		{ 158, 379, 18 },
		{ 158, 380, 18 },
		{ 158, 381, 18 },
		{ 158, 382, 18 },
		{ 159, 331, 18 },
		{ 159, 332, 18 },
		{ 159, 333, 18 },
		{ 159, 334, 18 },
		{ 159, 335, 18 },
		{ 159, 336, 18 },
		{ 159, 337, 18 },
		{ 159, 338, 18 },
		{ 159, 339, 18 },
		{ 159, 340, 18 },
		{ 159, 341, 18 },
		{ 159, 342, 18 },
		{ 159, 343, 18 },
		{ 159, 344, 18 },
		{ 159, 345, 18 },
		{ 159, 346, 18 },
		{ 159, 347, 18 },
		{ 159, 348, 18 },
		{ 159, 349, 18 },
		{ 159, 350, 18 },
		{ 159, 351, 18 },
		{ 159, 352, 18 },
		{ 159, 353, 18 },
		{ 159, 354, 18 },
		{ 159, 355, 18 },
		{ 159, 356, 18 },
		{ 159, 357, 18 },
		{ 159, 358, 18 },
		{ 159, 359, 18 },
		{ 159, 360, 18 },
		{ 159, 361, 18 },
		{ 159, 362, 18 },
		{ 159, 363, 18 },
		{ 159, 364, 18 },
		{ 159, 365, 18 },
		{ 159, 366, 18 },
		{ 159, 367, 18 },
		{ 159, 368, 18 },
		{ 159, 369, 18 },
		{ 159, 370, 18 },
		{ 159, 371, 18 },
		{ 159, 372, 18 },
		{ 159, 373, 18 },
		{ 159, 374, 18 },
		{ 159, 375, 18 },
		{ 159, 376, 18 },
		{ 159, 377, 18 },
		{ 159, 378, 18 },
		{ 159, 379, 18 },
		{ 159, 380, 18 },
		{ 159, 381, 18 },
		{ 159, 382, 18 },
		{ 160, 330, 18 },
		{ 160, 331, 18 },
		{ 160, 332, 18 },
		{ 160, 333, 18 },
		{ 160, 334, 18 },
		{ 160, 335, 18 },
		{ 160, 336, 18 },
		{ 160, 337, 18 },
		{ 160, 338, 18 },
		{ 160, 339, 18 },
		{ 160, 340, 18 },
		{ 160, 341, 18 },
		{ 160, 342, 18 },
		{ 160, 343, 18 },
		{ 160, 344, 18 },
		{ 160, 345, 18 },
		{ 160, 346, 18 },
		{ 160, 347, 18 },
		{ 160, 348, 18 },
		{ 160, 349, 18 },
		{ 160, 350, 18 },
		{ 160, 351, 18 },
		{ 160, 352, 18 },
		{ 160, 353, 18 },
		{ 160, 354, 18 },
		{ 160, 355, 18 },
		{ 160, 356, 18 },
		{ 160, 357, 18 },
		{ 160, 358, 18 },
		{ 160, 359, 18 },
		{ 160, 360, 18 },
		{ 160, 361, 18 },
		{ 160, 362, 18 },
		{ 160, 363, 18 },
		{ 160, 364, 18 },
		{ 160, 365, 18 },
		{ 160, 366, 18 },
		{ 160, 367, 18 },
		{ 160, 368, 18 },
		{ 160, 369, 18 },
		{ 160, 370, 18 },
		{ 160, 371, 18 },
		{ 160, 372, 18 },
		{ 160, 373, 18 },
		{ 160, 374, 18 },
		{ 160, 375, 18 },
		{ 160, 376, 18 },
		{ 160, 377, 18 },
		{ 160, 378, 18 },
		{ 160, 379, 18 },
		{ 160, 380, 18 },
		{ 160, 381, 18 },
		{ 160, 382, 18 },
		{ 161, 330, 18 },
		{ 161, 331, 18 },
		{ 161, 332, 18 },
		{ 161, 333, 18 },
		{ 161, 334, 18 },
		{ 161, 335, 18 },
		{ 161, 336, 18 },
		{ 161, 337, 18 },
		{ 161, 338, 18 },
		{ 161, 339, 18 },
		{ 161, 340, 18 },
		{ 161, 341, 18 },
		{ 161, 342, 18 },
		{ 161, 343, 18 },
		{ 161, 344, 18 },
		{ 161, 345, 18 },
		{ 161, 346, 18 },
		{ 161, 347, 18 },
		{ 161, 348, 18 },
		{ 161, 349, 18 },
		{ 161, 350, 18 },
		{ 161, 351, 18 },
		{ 161, 352, 18 },
		{ 161, 353, 18 },
		{ 161, 354, 18 },
		{ 161, 355, 18 },
		{ 161, 356, 18 },
		{ 161, 357, 18 },
		{ 161, 358, 18 },
		{ 161, 359, 18 },
		{ 161, 360, 18 },
		{ 161, 361, 18 },
		{ 161, 362, 18 },
		{ 161, 363, 18 },
		{ 161, 364, 18 },
		{ 161, 365, 18 },
		{ 161, 366, 18 },
		{ 161, 367, 18 },
		{ 161, 368, 18 },
		{ 161, 369, 18 },
		{ 161, 370, 18 },
		{ 161, 371, 18 },
		{ 161, 372, 18 },
		{ 161, 373, 18 },
		{ 161, 374, 18 },
		{ 161, 375, 18 },
		{ 161, 376, 18 },
		{ 161, 377, 18 },
		{ 161, 378, 18 },
		{ 161, 379, 18 },
		{ 161, 380, 18 },
		{ 161, 381, 18 },
		{ 162, 329, 18 },
		{ 162, 330, 18 },
		{ 162, 331, 18 },
		{ 162, 332, 18 },
		{ 162, 333, 18 },
		{ 162, 334, 18 },
		{ 162, 335, 18 },
		{ 162, 336, 18 },
		{ 162, 337, 18 },
		{ 162, 338, 18 },
		{ 162, 339, 18 },
		{ 162, 340, 18 },
		{ 162, 341, 18 },
		{ 162, 342, 18 },
		{ 162, 343, 18 },
		{ 162, 344, 18 },
		{ 162, 345, 18 },
		{ 162, 346, 18 },
		{ 162, 347, 18 },
		{ 162, 348, 18 },
		{ 162, 349, 18 },
		{ 162, 350, 18 },
		{ 162, 351, 18 },
		{ 162, 352, 18 },
		{ 162, 353, 18 },
		{ 162, 354, 18 },
		{ 162, 355, 18 },
		{ 162, 356, 18 },
		{ 162, 357, 18 },
		{ 162, 358, 18 },
		{ 162, 359, 18 },
		{ 162, 360, 18 },
		{ 162, 361, 18 },
		{ 162, 362, 18 },
		{ 162, 363, 18 },
		{ 162, 364, 18 },
		{ 162, 365, 18 },
		{ 162, 366, 18 },
		{ 162, 367, 18 },
		{ 162, 368, 18 },
		{ 162, 369, 18 },
		{ 162, 370, 18 },
		{ 162, 371, 18 },
		{ 162, 372, 18 },
		{ 162, 373, 18 },
		{ 162, 374, 18 },
		{ 162, 375, 18 },
		{ 162, 376, 18 },
		{ 162, 377, 18 },
		{ 162, 378, 18 },
		{ 162, 379, 18 },
		{ 162, 380, 18 },
		{ 162, 381, 18 },
		{ 163, 328, 18 },
		{ 163, 329, 18 },
		{ 163, 330, 18 },
		{ 163, 331, 18 },
		{ 163, 332, 18 },
		{ 163, 333, 18 },
		{ 163, 334, 18 },
		{ 163, 335, 18 },
		{ 163, 336, 18 },
		{ 163, 337, 18 },
		{ 163, 338, 18 },
		{ 163, 339, 18 },
		{ 163, 340, 18 },
		{ 163, 341, 18 },
		{ 163, 342, 18 },
		{ 163, 343, 18 },
		{ 163, 344, 18 },
		{ 163, 345, 18 },
		{ 163, 346, 18 },
		{ 163, 347, 18 },
		{ 163, 348, 18 },
		{ 163, 349, 18 },
		{ 163, 350, 18 },
		{ 163, 351, 18 },
		{ 163, 352, 18 },
		{ 163, 353, 18 },
		{ 163, 354, 18 },
		{ 163, 355, 18 },
		{ 163, 356, 18 },
		{ 163, 357, 18 },
		{ 163, 358, 18 },
		{ 163, 359, 18 },
		{ 163, 360, 18 },
		{ 163, 361, 18 },
		{ 163, 362, 18 },
		{ 163, 363, 18 },
		{ 163, 364, 18 },
		{ 163, 365, 18 },
		{ 163, 366, 18 },
		{ 163, 367, 18 },
		{ 163, 368, 18 },
		{ 163, 369, 18 },
		{ 163, 370, 18 },
		{ 163, 371, 18 },
		{ 163, 372, 18 },
		{ 163, 373, 18 },
		{ 163, 374, 18 },
		{ 163, 375, 18 },
		{ 163, 376, 18 },
		{ 163, 377, 18 },
		{ 163, 378, 18 },
		{ 163, 379, 18 },
		{ 163, 380, 18 },
		{ 164, 327, 18 },
		{ 164, 328, 18 },
		{ 164, 329, 18 },
		{ 164, 330, 18 },
		{ 164, 331, 18 },
		{ 164, 332, 18 },
		{ 164, 333, 18 },
		{ 164, 334, 18 },
		{ 164, 335, 18 },
		{ 164, 336, 18 },
		{ 164, 337, 18 },
		{ 164, 338, 18 },
		{ 164, 339, 18 },
		{ 164, 340, 18 },
		{ 164, 341, 18 },
		{ 164, 342, 18 },
		{ 164, 343, 18 },
		{ 164, 344, 18 },
		{ 164, 345, 18 },
		{ 164, 346, 18 },
		{ 164, 347, 18 },
		{ 164, 348, 18 },
		{ 164, 349, 18 },
		{ 164, 350, 18 },
		{ 164, 351, 18 },
		{ 164, 352, 18 },
		{ 164, 353, 18 },
		{ 164, 354, 18 },
		{ 164, 355, 18 },
		{ 164, 356, 18 },
		{ 164, 357, 18 },
		{ 164, 358, 18 },
		{ 164, 359, 18 },
		{ 164, 360, 18 },
		{ 164, 361, 18 },
		{ 164, 362, 18 },
		{ 164, 363, 18 },
		{ 164, 364, 18 },
		{ 164, 365, 18 },
		{ 164, 366, 18 },
		{ 164, 367, 18 },
		{ 164, 368, 18 },
		{ 164, 369, 18 },
		{ 164, 370, 18 },
		{ 164, 371, 18 },
		{ 164, 372, 18 },
		{ 164, 373, 18 },
		{ 164, 374, 18 },
		{ 164, 375, 18 },
		{ 164, 376, 18 },
		{ 164, 377, 18 },
		{ 164, 378, 18 },
		{ 164, 379, 18 },
		{ 165, 327, 18 },
		{ 165, 328, 18 },
		{ 165, 329, 18 },
		{ 165, 330, 18 },
		{ 165, 331, 18 },
		{ 165, 332, 18 },
		{ 165, 333, 18 },
		{ 165, 334, 18 },
		{ 165, 335, 18 },
		{ 165, 336, 18 },
		{ 165, 337, 18 },
		{ 165, 338, 18 },
		{ 165, 339, 18 },
		{ 165, 340, 18 },
		{ 165, 341, 18 },
		{ 165, 342, 18 },
		{ 165, 343, 18 },
		{ 165, 344, 18 },
		{ 165, 345, 18 },
		{ 165, 346, 18 },
		{ 165, 347, 18 },
		{ 165, 348, 18 },
		{ 165, 349, 18 },
		{ 165, 350, 18 },
		{ 165, 351, 18 },
		{ 165, 352, 18 },
		{ 165, 353, 18 },
		{ 165, 354, 18 },
		{ 165, 355, 18 },
		{ 165, 356, 18 },
		{ 165, 357, 18 },
		{ 165, 358, 18 },
		{ 165, 359, 18 },
		{ 165, 360, 18 },
		{ 165, 361, 18 },
		{ 165, 362, 18 },
		{ 165, 363, 18 },
		{ 165, 364, 18 },
		{ 165, 365, 18 },
		{ 165, 366, 18 },
		{ 165, 367, 18 },
		{ 165, 368, 18 },
		{ 165, 369, 18 },
		{ 165, 370, 18 },
		{ 165, 371, 18 },
		{ 165, 372, 18 },
		{ 165, 373, 18 },
		{ 165, 374, 18 },
		{ 165, 375, 18 },
		{ 165, 376, 18 },
		{ 165, 377, 18 },
		{ 165, 378, 18 },
		{ 166, 326, 18 },
		{ 166, 327, 18 },
		{ 166, 328, 18 },
		{ 166, 329, 18 },
		{ 166, 330, 18 },
		{ 166, 331, 18 },
		{ 166, 332, 18 },
		{ 166, 333, 18 },
		{ 166, 334, 18 },
		{ 166, 335, 18 },
		{ 166, 336, 18 },
		{ 166, 337, 18 },
		{ 166, 338, 18 },
		{ 166, 339, 18 },
		{ 166, 340, 18 },
		{ 166, 341, 18 },
		{ 166, 342, 18 },
		{ 166, 343, 18 },
		{ 166, 344, 18 },
		{ 166, 345, 18 },
		{ 166, 346, 18 },
		{ 166, 347, 18 },
		{ 166, 348, 18 },
		{ 166, 349, 18 },
		{ 166, 350, 18 },
		{ 166, 351, 18 },
		{ 166, 352, 18 },
		{ 166, 353, 18 },
		{ 166, 354, 18 },
		{ 166, 355, 18 },
		{ 166, 356, 18 },
		{ 166, 357, 18 },
		{ 166, 358, 18 },
		{ 166, 359, 18 },
		{ 166, 360, 18 },
		{ 166, 361, 18 },
		{ 166, 362, 18 },
		{ 166, 363, 18 },
		{ 166, 364, 18 },
		{ 166, 365, 18 },
		{ 166, 366, 18 },
		{ 166, 367, 18 },
		{ 166, 368, 18 },
		{ 166, 369, 18 },
		{ 166, 370, 18 },
		{ 166, 371, 18 },
		{ 166, 372, 18 },
		{ 166, 373, 18 },
		{ 166, 374, 18 },
		{ 166, 375, 18 },
		{ 166, 376, 18 },
		{ 166, 377, 18 },
		{ 167, 325, 18 },
		{ 167, 326, 18 },
		{ 167, 327, 18 },
		{ 167, 328, 18 },
		{ 167, 329, 18 },
		{ 167, 330, 18 },
		{ 167, 331, 18 },
		{ 167, 332, 18 },
		{ 167, 333, 18 },
		{ 167, 334, 18 },
		{ 167, 335, 18 },
		{ 167, 336, 18 },
		{ 167, 337, 18 },
		{ 167, 338, 18 },
		{ 167, 339, 18 },
		{ 167, 340, 18 },
		{ 167, 341, 18 },
		{ 167, 342, 18 },
		{ 167, 343, 18 },
		{ 167, 344, 18 },
		{ 167, 345, 18 },
		{ 167, 346, 18 },
		{ 167, 347, 18 },
		{ 167, 348, 18 },
		{ 167, 349, 18 },
		{ 167, 350, 18 },
		{ 167, 351, 18 },
		{ 167, 352, 18 },
		{ 167, 353, 18 },
		{ 167, 354, 18 },
		{ 167, 355, 18 },
		{ 167, 356, 18 },
		{ 167, 357, 18 },
		{ 167, 358, 18 },
		{ 167, 359, 18 },
		{ 167, 360, 18 },
		{ 167, 361, 18 },
		{ 167, 362, 18 },
		{ 167, 363, 18 },
		{ 167, 364, 18 },
		{ 167, 365, 18 },
		{ 167, 366, 18 },
		{ 167, 367, 18 },
		{ 167, 368, 18 },
		{ 167, 369, 18 },
		{ 167, 370, 18 },
		{ 167, 371, 18 },
		{ 167, 372, 18 },
		{ 167, 373, 18 },
		{ 167, 374, 18 },
		{ 167, 375, 18 },
		{ 167, 376, 18 },
		{ 168, 324, 18 },
		{ 168, 325, 18 },
		{ 168, 326, 18 },
		{ 168, 327, 18 },
		{ 168, 328, 18 },
		{ 168, 329, 18 },
		{ 168, 330, 18 },
		{ 168, 331, 18 },
		{ 168, 332, 18 },
		{ 168, 333, 18 },
		{ 168, 334, 18 },
		{ 168, 335, 18 },
		{ 168, 336, 18 },
		{ 168, 337, 18 },
		{ 168, 338, 18 },
		{ 168, 339, 18 },
		{ 168, 340, 18 },
		{ 168, 341, 18 },
		{ 168, 342, 18 },
		{ 168, 343, 18 },
		{ 168, 344, 18 },
		{ 168, 345, 18 },
		{ 168, 346, 18 },
		{ 168, 347, 18 },
		{ 168, 348, 18 },
		{ 168, 349, 18 },
		{ 168, 350, 18 },
		{ 168, 351, 18 },
		{ 168, 352, 18 },
		{ 168, 353, 18 },
		{ 168, 354, 18 },
		{ 168, 355, 18 },
		{ 168, 356, 18 },
		{ 168, 357, 18 },
		{ 168, 358, 18 },
		{ 168, 359, 18 },
		{ 168, 360, 18 },
		{ 168, 361, 18 },
		{ 168, 362, 18 },
		{ 168, 363, 18 },
		{ 168, 364, 18 },
		{ 168, 365, 18 },
		{ 168, 366, 18 },
		{ 168, 367, 18 },
		{ 168, 368, 18 },
		{ 168, 369, 18 },
		{ 168, 370, 18 },
		{ 168, 371, 18 },
		{ 168, 372, 18 },
		{ 168, 373, 18 },
		{ 168, 374, 18 },
		{ 168, 375, 18 },
		{ 169, 324, 18 },
		{ 169, 325, 18 },
		{ 169, 326, 18 },
		{ 169, 327, 18 },
		{ 169, 328, 18 },
		{ 169, 329, 18 },
		{ 169, 330, 18 },
		{ 169, 331, 18 },
		{ 169, 332, 18 },
		{ 169, 333, 18 },
		{ 169, 334, 18 },
		{ 169, 335, 18 },
		{ 169, 336, 18 },
		{ 169, 337, 18 },
		{ 169, 338, 18 },
		{ 169, 339, 18 },
		{ 169, 340, 18 },
		{ 169, 341, 18 },
		{ 169, 342, 18 },
		{ 169, 343, 18 },
		{ 169, 344, 18 },
		{ 169, 345, 18 },
		{ 169, 346, 18 },
		{ 169, 347, 18 },
		{ 169, 348, 18 },
		{ 169, 349, 18 },
		{ 169, 350, 18 },
		{ 169, 351, 18 },
		{ 169, 352, 18 },
		{ 169, 353, 18 },
		{ 169, 354, 18 },
		{ 169, 355, 18 },
		{ 169, 356, 18 },
		{ 169, 357, 18 },
		{ 169, 358, 18 },
		{ 169, 359, 18 },
		{ 169, 360, 18 },
		{ 169, 361, 18 },
		{ 169, 362, 18 },
		{ 169, 363, 18 },
		{ 169, 364, 18 },
		{ 169, 365, 18 },
		{ 169, 366, 18 },
		{ 169, 367, 18 },
		{ 169, 368, 18 },
		{ 169, 369, 18 },
		{ 169, 370, 18 },
		{ 169, 371, 18 },
		{ 169, 372, 18 },
		{ 169, 373, 18 },
		{ 169, 374, 18 },
		{ 170, 323, 18 },
		{ 170, 324, 18 },
		{ 170, 325, 18 },
		{ 170, 326, 18 },
		{ 170, 327, 18 },
		{ 170, 328, 18 },
		{ 170, 329, 18 },
		{ 170, 330, 18 },
		{ 170, 331, 18 },
		{ 170, 332, 18 },
		{ 170, 333, 18 },
		{ 170, 334, 18 },
		{ 170, 335, 18 },
		{ 170, 336, 18 },
		{ 170, 337, 18 },
		{ 170, 338, 18 },
		{ 170, 339, 18 },
		{ 170, 340, 18 },
		{ 170, 341, 18 },
		{ 170, 342, 18 },
		{ 170, 343, 18 },
		{ 170, 344, 18 },
		{ 170, 345, 18 },
		{ 170, 346, 18 },
		{ 170, 347, 18 },
		{ 170, 348, 18 },
		{ 170, 349, 18 },
		{ 170, 350, 18 },
		{ 170, 351, 18 },
		{ 170, 352, 18 },
		{ 170, 353, 18 },
		{ 170, 354, 18 },
		{ 170, 355, 18 },
		{ 170, 356, 18 },
		{ 170, 357, 18 },
		{ 170, 358, 18 },
		{ 170, 359, 18 },
		{ 170, 360, 18 },
		{ 170, 361, 18 },
		{ 170, 362, 18 },
		{ 170, 363, 18 },
		{ 170, 364, 18 },
		{ 170, 365, 18 },
		{ 170, 366, 18 },
		{ 170, 367, 18 },
		{ 170, 368, 18 },
		{ 170, 369, 18 },
		{ 170, 370, 18 },
		{ 170, 371, 18 },
		{ 170, 372, 18 },
		{ 171, 322, 18 },
		{ 171, 323, 18 },
		{ 171, 324, 18 },
		{ 171, 325, 18 },
		{ 171, 326, 18 },
		{ 171, 327, 18 },
		{ 171, 328, 18 },
		{ 171, 329, 18 },
		{ 171, 330, 18 },
		{ 171, 331, 18 },
		{ 171, 332, 18 },
		{ 171, 333, 18 },
		{ 171, 334, 18 },
		{ 171, 335, 18 },
		{ 171, 336, 18 },
		{ 171, 337, 18 },
		{ 171, 338, 18 },
		{ 171, 339, 18 },
		{ 171, 340, 18 },
		{ 171, 341, 18 },
		{ 171, 342, 18 },
		{ 171, 343, 18 },
		{ 171, 344, 18 },
		{ 171, 345, 18 },
		{ 171, 346, 18 },
		{ 171, 347, 18 },
		{ 171, 348, 18 },
		{ 171, 349, 18 },
		{ 171, 350, 18 },
		{ 171, 351, 18 },
		{ 171, 352, 18 },
		{ 171, 353, 18 },
		{ 171, 354, 18 },
		{ 171, 355, 18 },
		{ 171, 356, 18 },
		{ 171, 357, 18 },
		{ 171, 358, 18 },
		{ 171, 359, 18 },
		{ 171, 360, 18 },
		{ 171, 361, 18 },
		{ 171, 362, 18 },
		{ 171, 363, 18 },
		{ 171, 364, 18 },
		{ 171, 365, 18 },
		{ 171, 366, 18 },
		{ 171, 367, 18 },
		{ 171, 368, 18 },
		{ 171, 369, 18 },
		{ 171, 370, 18 },
		{ 171, 371, 18 },
		{ 172, 321, 18 },
		{ 172, 322, 18 },
		{ 172, 323, 18 },
		{ 172, 324, 18 },
		{ 172, 325, 18 },
		{ 172, 326, 18 },
		{ 172, 327, 18 },
		{ 172, 328, 18 },
		{ 172, 329, 18 },
		{ 172, 330, 18 },
		{ 172, 331, 18 },
		{ 172, 332, 18 },
		{ 172, 333, 18 },
		{ 172, 334, 18 },
		{ 172, 335, 18 },
		{ 172, 336, 18 },
		{ 172, 337, 18 },
		{ 172, 338, 18 },
		{ 172, 339, 18 },
		{ 172, 340, 18 },
		{ 172, 341, 18 },
		{ 172, 342, 18 },
		{ 172, 343, 18 },
		{ 172, 344, 18 },
		{ 172, 345, 18 },
		{ 172, 346, 18 },
		{ 172, 347, 18 },
		{ 172, 348, 18 },
		{ 172, 349, 18 },
		{ 172, 350, 18 },
		{ 172, 351, 18 },
		{ 172, 352, 18 },
		{ 172, 353, 18 },
		{ 172, 354, 18 },
		{ 172, 355, 18 },
		{ 172, 356, 18 },
		{ 172, 357, 18 },
		{ 172, 358, 18 },
		{ 172, 359, 18 },
		{ 172, 360, 18 },
		{ 172, 361, 18 },
		{ 172, 362, 18 },
		{ 172, 363, 18 },
		{ 172, 364, 18 },
		{ 172, 365, 18 },
		{ 172, 366, 18 },
		{ 172, 367, 18 },
		{ 172, 368, 18 },
		{ 172, 369, 18 },
		{ 172, 370, 18 },
		{ 173, 321, 18 },
		{ 173, 322, 18 },
		{ 173, 323, 18 },
		{ 173, 324, 18 },
		{ 173, 325, 18 },
		{ 173, 326, 18 },
		{ 173, 327, 18 },
		{ 173, 328, 18 },
		{ 173, 329, 18 },
		{ 173, 330, 18 },
		{ 173, 331, 18 },
		{ 173, 332, 18 },
		{ 173, 333, 18 },
		{ 173, 334, 18 },
		{ 173, 335, 18 },
		{ 173, 336, 18 },
		{ 173, 337, 18 },
		{ 173, 338, 18 },
		{ 173, 339, 18 },
		{ 173, 340, 18 },
		{ 173, 341, 18 },
		{ 173, 342, 18 },
		{ 173, 343, 18 },
		{ 173, 344, 18 },
		{ 173, 345, 18 },
		{ 173, 346, 18 },
		{ 173, 347, 18 },
		{ 173, 348, 18 },
		{ 173, 349, 18 },
		{ 173, 350, 18 },
		{ 173, 351, 18 },
		{ 173, 352, 18 },
		{ 173, 353, 18 },
		{ 173, 354, 18 },
		{ 173, 355, 18 },
		{ 173, 356, 18 },
		{ 173, 357, 18 },
		{ 173, 358, 18 },
		{ 173, 359, 18 },
		{ 173, 360, 18 },
		{ 173, 361, 18 },
		{ 173, 362, 18 },
		{ 173, 363, 18 },
		{ 173, 364, 18 },
		{ 173, 365, 18 },
		{ 173, 366, 18 },
		{ 173, 367, 18 },
		{ 173, 368, 18 },
		{ 174, 320, 18 },
		{ 174, 321, 18 },
		{ 174, 322, 18 },
		{ 174, 323, 18 },
		{ 174, 324, 18 },
		{ 174, 325, 18 },
		{ 174, 326, 18 },
		{ 174, 327, 18 },
		{ 174, 328, 18 },
		{ 174, 329, 18 },
		{ 174, 330, 18 },
		{ 174, 331, 18 },
		{ 174, 332, 18 },
		{ 174, 333, 18 },
		{ 174, 334, 18 },
		{ 174, 335, 18 },
		{ 174, 336, 18 },
		{ 174, 337, 18 },
		{ 174, 338, 18 },
		{ 174, 339, 18 },
		{ 174, 340, 18 },
		{ 174, 341, 18 },
		{ 174, 342, 18 },
		{ 174, 343, 18 },
		{ 174, 344, 18 },
		{ 174, 345, 18 },
		{ 174, 346, 18 },
		{ 174, 347, 18 },
		{ 174, 348, 18 },
		{ 174, 349, 18 },
		{ 174, 350, 18 },
		{ 174, 351, 18 },
		{ 174, 352, 18 },
		{ 174, 353, 18 },
		{ 174, 354, 18 },
		{ 174, 355, 18 },
		{ 174, 356, 18 },
		{ 174, 357, 18 },
		{ 174, 358, 18 },
		{ 174, 359, 18 },
		{ 174, 360, 18 },
		{ 174, 361, 18 },
		{ 174, 362, 18 },
		{ 174, 363, 18 },
		{ 174, 364, 18 },
		{ 174, 365, 18 },
		{ 174, 366, 18 },
		{ 174, 367, 18 },
		{ 175, 319, 18 },
		{ 175, 320, 18 },
		{ 175, 321, 18 },
		{ 175, 322, 18 },
		{ 175, 323, 18 },
		{ 175, 324, 18 },
		{ 175, 325, 18 },
		{ 175, 326, 18 },
		{ 175, 327, 18 },
		{ 175, 328, 18 },
		{ 175, 329, 18 },
		{ 175, 330, 18 },
		{ 175, 331, 18 },
		{ 175, 332, 18 },
		{ 175, 333, 18 },
		{ 175, 334, 18 },
		{ 175, 335, 18 },
		{ 175, 336, 18 },
		{ 175, 337, 18 },
		{ 175, 338, 18 },
		{ 175, 339, 18 },
		{ 175, 340, 18 },
		{ 175, 341, 18 },
		{ 175, 342, 18 },
		{ 175, 343, 18 },
		{ 175, 344, 18 },
		{ 175, 345, 18 },
		{ 175, 346, 18 },
		{ 175, 347, 18 },
		{ 175, 348, 18 },
		{ 175, 349, 18 },
		{ 175, 350, 18 },
		{ 175, 351, 18 },
		{ 175, 352, 18 },
		{ 175, 353, 18 },
		{ 175, 354, 18 },
		{ 175, 355, 18 },
		{ 175, 356, 18 },
		{ 175, 357, 18 },
		{ 175, 358, 18 },
		{ 175, 359, 18 },
		{ 175, 360, 18 },
		{ 175, 361, 18 },
		{ 175, 362, 18 },
		{ 175, 363, 18 },
		{ 175, 364, 18 },
		{ 175, 365, 18 },
		{ 175, 366, 18 },
		{ 176, 318, 18 },
		{ 176, 319, 18 },
		{ 176, 320, 18 },
		{ 176, 321, 18 },
		{ 176, 322, 18 },
		{ 176, 323, 18 },
		{ 176, 324, 18 },
		{ 176, 325, 18 },
		{ 176, 326, 18 },
		{ 176, 327, 18 },
		{ 176, 328, 18 },
		{ 176, 329, 18 },
		{ 176, 330, 18 },
		{ 176, 331, 18 },
		{ 176, 332, 18 },
		{ 176, 333, 18 },
		{ 176, 334, 18 },
		{ 176, 335, 18 },
		{ 176, 336, 18 },
		{ 176, 337, 18 },
		{ 176, 338, 18 },
		{ 176, 339, 18 },
		{ 176, 340, 18 },
		{ 176, 341, 18 },
		{ 176, 342, 18 },
		{ 176, 343, 18 },
		{ 176, 344, 18 },
		{ 176, 345, 18 },
		{ 176, 346, 18 },
		{ 176, 347, 18 },
		{ 176, 348, 18 },
		{ 176, 349, 18 },
		{ 176, 350, 18 },
		{ 176, 351, 18 },
		{ 176, 352, 18 },
		{ 176, 353, 18 },
		{ 176, 354, 18 },
		{ 176, 355, 18 },
		{ 176, 356, 18 },
		{ 176, 357, 18 },
		{ 176, 358, 18 },
		{ 176, 359, 18 },
		{ 176, 360, 18 },
		{ 176, 361, 18 },
		{ 176, 362, 18 },
		{ 176, 363, 18 },
		{ 176, 364, 18 },
		{ 177, 318, 18 },
		{ 177, 319, 18 },
		{ 177, 320, 18 },
		{ 177, 321, 18 },
		{ 177, 322, 18 },
		{ 177, 323, 18 },
		{ 177, 324, 18 },
		{ 177, 325, 18 },
		{ 177, 326, 18 },
		{ 177, 327, 18 },
		{ 177, 328, 18 },
		{ 177, 329, 18 },
		{ 177, 330, 18 },
		{ 177, 331, 18 },
		{ 177, 332, 18 },
		{ 177, 333, 18 },
		{ 177, 334, 18 },
		{ 177, 335, 18 },
		{ 177, 336, 18 },
		{ 177, 337, 18 },
		{ 177, 338, 18 },
		{ 177, 339, 18 },
		{ 177, 340, 18 },
		{ 177, 341, 18 },
		{ 177, 342, 18 },
		{ 177, 343, 18 },
		{ 177, 344, 18 },
		{ 177, 345, 18 },
		{ 177, 346, 18 },
		{ 177, 347, 18 },
		{ 177, 348, 18 },
		{ 177, 349, 18 },
		{ 177, 350, 18 },
		{ 177, 351, 18 },
		{ 177, 352, 18 },
		{ 177, 353, 18 },
		{ 177, 354, 18 },
		{ 177, 355, 18 },
		{ 177, 356, 18 },
		{ 177, 357, 18 },
		{ 177, 358, 18 },
		{ 177, 359, 18 },
		{ 177, 360, 18 },
		{ 177, 361, 18 },
		{ 177, 362, 18 },
		{ 177, 363, 18 },
		{ 178, 317, 18 },
		{ 178, 318, 18 },
		{ 178, 319, 18 },
		{ 178, 320, 18 },
		{ 178, 321, 18 },
		{ 178, 322, 18 },
		{ 178, 323, 18 },
		{ 178, 324, 18 },
		{ 178, 325, 18 },
		{ 178, 326, 18 },
		{ 178, 327, 18 },
		{ 178, 328, 18 },
		{ 178, 329, 18 },
		{ 178, 330, 18 },
		{ 178, 331, 18 },
		{ 178, 332, 18 },
		{ 178, 333, 18 },
		{ 178, 334, 18 },
		{ 178, 335, 18 },
		{ 178, 336, 18 },
		{ 178, 337, 18 },
		{ 178, 338, 18 },
		{ 178, 339, 18 },
		{ 178, 340, 18 },
		{ 178, 341, 18 },
		{ 178, 342, 18 },
		{ 178, 343, 18 },
		{ 178, 344, 18 },
		{ 178, 345, 18 },
		{ 178, 346, 18 },
		{ 178, 347, 18 },
		{ 178, 348, 18 },
		{ 178, 349, 18 },
		{ 178, 350, 18 },
		{ 178, 351, 18 },
		{ 178, 352, 18 },
		{ 178, 353, 18 },
		{ 178, 354, 18 },
		{ 178, 355, 18 },
		{ 178, 356, 18 },
		{ 178, 357, 18 },
		{ 178, 358, 18 },
		{ 178, 359, 18 },
		{ 178, 360, 18 },
		{ 178, 361, 18 },
		{ 178, 362, 18 },
		{ 179, 316, 18 },
		{ 179, 317, 18 },
		{ 179, 318, 18 },
		{ 179, 319, 18 },
		{ 179, 320, 18 },
		{ 179, 321, 18 },
		{ 179, 322, 18 },
		{ 179, 323, 18 },
		{ 179, 324, 18 },
		{ 179, 325, 18 },
		{ 179, 326, 18 },
		{ 179, 327, 18 },
		{ 179, 328, 18 },
		{ 179, 329, 18 },
		{ 179, 330, 18 },
		{ 179, 331, 18 },
		{ 179, 332, 18 },
		{ 179, 333, 18 },
		{ 179, 334, 18 },
		{ 179, 335, 18 },
		{ 179, 336, 18 },
		{ 179, 337, 18 },
		{ 179, 338, 18 },
		{ 179, 339, 18 },
		{ 179, 340, 18 },
		{ 179, 341, 18 },
		{ 179, 342, 18 },
		{ 179, 343, 18 },
		{ 179, 344, 18 },
		{ 179, 345, 18 },
		{ 179, 346, 18 },
		{ 179, 347, 18 },
		{ 179, 348, 18 },
		{ 179, 349, 18 },
		{ 179, 350, 18 },
		{ 179, 351, 18 },
		{ 179, 352, 18 },
		{ 179, 353, 18 },
		{ 179, 354, 18 },
		{ 179, 355, 18 },
		{ 179, 356, 18 },
		{ 179, 357, 18 },
		{ 179, 358, 18 },
		{ 179, 359, 18 },
		{ 179, 360, 18 },
		{ 180, 315, 18 },
		{ 180, 316, 18 },
		{ 180, 317, 18 },
		{ 180, 318, 18 },
		{ 180, 319, 18 },
		{ 180, 320, 18 },
		{ 180, 321, 18 },
		{ 180, 322, 18 },
		{ 180, 323, 18 },
		{ 180, 324, 18 },
		{ 180, 325, 18 },
		{ 180, 326, 18 },
		{ 180, 327, 18 },
		{ 180, 328, 18 },
		{ 180, 329, 18 },
		{ 180, 330, 18 },
		{ 180, 331, 18 },
		{ 180, 332, 18 },
		{ 180, 333, 18 },
		{ 180, 334, 18 },
		{ 180, 335, 18 },
		{ 180, 336, 18 },
		{ 180, 337, 18 },
		{ 180, 338, 18 },
		{ 180, 339, 18 },
		{ 180, 340, 18 },
		{ 180, 341, 18 },
		{ 180, 342, 18 },
		{ 180, 343, 18 },
		{ 180, 344, 18 },
		{ 180, 345, 18 },
		{ 180, 346, 18 },
		{ 180, 347, 18 },
		{ 180, 348, 18 },
		{ 180, 349, 18 },
		{ 180, 350, 18 },
		{ 180, 351, 18 },
		{ 180, 352, 18 },
		{ 180, 353, 18 },
		{ 180, 354, 18 },
		{ 180, 355, 18 },
		{ 180, 356, 18 },
		{ 180, 357, 18 },
		{ 180, 358, 18 },
		{ 180, 359, 18 },
		{ 181, 315, 18 },
		{ 181, 316, 18 },
		{ 181, 317, 18 },
		{ 181, 318, 18 },
		{ 181, 319, 18 },
		{ 181, 320, 18 },
		{ 181, 321, 18 },
		{ 181, 322, 18 },
		{ 181, 323, 18 },
		{ 181, 324, 18 },
		{ 181, 325, 18 },
		{ 181, 326, 18 },
		{ 181, 327, 18 },
		{ 181, 328, 18 },
		{ 181, 329, 18 },
		{ 181, 330, 18 },
		{ 181, 331, 18 },
		{ 181, 332, 18 },
		{ 181, 333, 18 },
		{ 181, 334, 18 },
		{ 181, 335, 18 },
		{ 181, 336, 18 },
		{ 181, 337, 18 },
		{ 181, 338, 18 },
		{ 181, 339, 18 },
		{ 181, 340, 18 },
		{ 181, 341, 18 },
		{ 181, 342, 18 },
		{ 181, 343, 18 },
		{ 181, 344, 18 },
		{ 181, 345, 18 },
		{ 181, 346, 18 },
		{ 181, 347, 18 },
		{ 181, 348, 18 },
		{ 181, 349, 18 },
		{ 181, 350, 18 },
		{ 181, 351, 18 },
		{ 181, 352, 18 },
		{ 181, 353, 18 },
		{ 181, 354, 18 },
		{ 181, 355, 18 },
		{ 181, 356, 18 },
		{ 181, 357, 18 },
		{ 181, 358, 18 },
		{ 182, 314, 18 },
		{ 182, 315, 18 },
		{ 182, 316, 18 },
		{ 182, 317, 18 },
		{ 182, 318, 18 },
		{ 182, 319, 18 },
		{ 182, 320, 18 },
		{ 182, 321, 18 },
		{ 182, 322, 18 },
		{ 182, 323, 18 },
		{ 182, 324, 18 },
		{ 182, 325, 18 },
		{ 182, 326, 18 },
		{ 182, 327, 18 },
		{ 182, 328, 18 },
		{ 182, 329, 18 },
		{ 182, 330, 18 },
		{ 182, 331, 18 },
		{ 182, 332, 18 },
		{ 182, 333, 18 },
		{ 182, 334, 18 },
		{ 182, 335, 18 },
		{ 182, 336, 18 },
		{ 182, 337, 18 },
		{ 182, 338, 18 },
		{ 182, 339, 18 },
		{ 182, 340, 18 },
		{ 182, 341, 18 },
		{ 182, 342, 18 },
		{ 182, 343, 18 },
		{ 182, 344, 18 },
		{ 182, 345, 18 },
		{ 182, 346, 18 },
		{ 182, 347, 18 },
		{ 182, 348, 18 },
		{ 182, 349, 18 },
		{ 182, 350, 18 },
		{ 182, 351, 18 },
		{ 182, 352, 18 },
		{ 182, 353, 18 },
		{ 182, 354, 18 },
		{ 182, 355, 18 },
		{ 182, 356, 18 },
		{ 183, 313, 18 },
		{ 183, 314, 18 },
		{ 183, 315, 18 },
		{ 183, 316, 18 },
		{ 183, 317, 18 },
		{ 183, 318, 18 },
		{ 183, 319, 18 },
		{ 183, 320, 18 },
		{ 183, 321, 18 },
		{ 183, 322, 18 },
		{ 183, 323, 18 },
		{ 183, 324, 18 },
		{ 183, 325, 18 },
		{ 183, 326, 18 },
		{ 183, 327, 18 },
		{ 183, 328, 18 },
		{ 183, 329, 18 },
		{ 183, 330, 18 },
		{ 183, 331, 18 },
		{ 183, 332, 18 },
		{ 183, 333, 18 },
		{ 183, 334, 18 },
		{ 183, 335, 18 },
		{ 183, 336, 18 },
		{ 183, 337, 18 },
		{ 183, 338, 18 },
		{ 183, 339, 18 },
		{ 183, 340, 18 },
		{ 183, 341, 18 },
		{ 183, 342, 18 },
		{ 183, 343, 18 },
		{ 183, 344, 18 },
		{ 183, 345, 18 },
		{ 183, 346, 18 },
		{ 183, 347, 18 },
		{ 183, 348, 18 },
		{ 183, 349, 18 },
		{ 183, 350, 18 },
		{ 183, 351, 18 },
		{ 183, 352, 18 },
		{ 183, 353, 18 },
		{ 183, 354, 18 },
		{ 183, 355, 18 },
		{ 184, 312, 18 },
		{ 184, 313, 18 },
		{ 184, 314, 18 },
		{ 184, 315, 18 },
		{ 184, 316, 18 },
		{ 184, 317, 18 },
		{ 184, 318, 18 },
		{ 184, 319, 18 },
		{ 184, 320, 18 },
		{ 184, 321, 18 },
		{ 184, 322, 18 },
		{ 184, 323, 18 },
		{ 184, 324, 18 },
		{ 184, 325, 18 },
		{ 184, 326, 18 },
		{ 184, 327, 18 },
		{ 184, 328, 18 },
		{ 184, 329, 18 },
		{ 184, 330, 18 },
		{ 184, 331, 18 },
		{ 184, 332, 18 },
		{ 184, 333, 18 },
		{ 184, 334, 18 },
		{ 184, 335, 18 },
		{ 184, 336, 18 },
		{ 184, 337, 18 },
		{ 184, 338, 18 },
		{ 184, 339, 18 },
		{ 184, 340, 18 },
		{ 184, 341, 18 },
		{ 184, 342, 18 },
		{ 184, 343, 18 },
		{ 184, 344, 18 },
		{ 184, 345, 18 },
		{ 184, 346, 18 },
		{ 184, 347, 18 },
		{ 184, 348, 18 },
		{ 184, 349, 18 },
		{ 184, 350, 18 },
		{ 184, 351, 18 },
		{ 184, 352, 18 },
		{ 184, 353, 18 },
		{ 184, 354, 18 },
		{ 185, 312, 18 },
		{ 185, 313, 18 },
		{ 185, 314, 18 },
		{ 185, 315, 18 },
		{ 185, 316, 18 },
		{ 185, 317, 18 },
		{ 185, 318, 18 },
		{ 185, 319, 18 },
		{ 185, 320, 18 },
		{ 185, 321, 18 },
		{ 185, 322, 18 },
		{ 185, 323, 18 },
		{ 185, 324, 18 },
		{ 185, 325, 18 },
		{ 185, 326, 18 },
		{ 185, 327, 18 },
		{ 185, 328, 18 },
		{ 185, 329, 18 },
		{ 185, 330, 18 },
		{ 185, 331, 18 },
		{ 185, 332, 18 },
		{ 185, 333, 18 },
		{ 185, 334, 18 },
		{ 185, 335, 18 },
		{ 185, 336, 18 },
		{ 185, 337, 18 },
		{ 185, 338, 18 },
		{ 185, 339, 18 },
		{ 185, 340, 18 },
		{ 185, 341, 18 },
		{ 185, 342, 18 },
		{ 185, 343, 18 },
		{ 185, 344, 18 },
		{ 185, 345, 18 },
		{ 185, 346, 18 },
		{ 185, 347, 18 },
		{ 185, 348, 18 },
		{ 185, 349, 18 },
		{ 185, 350, 18 },
		{ 185, 351, 18 },
		{ 185, 352, 18 },
		{ 186, 311, 18 },
		{ 186, 312, 18 },
		{ 186, 313, 18 },
		{ 186, 314, 18 },
		{ 186, 315, 18 },
		{ 186, 316, 18 },
		{ 186, 317, 18 },
		{ 186, 318, 18 },
		{ 186, 319, 18 },
		{ 186, 320, 18 },
		{ 186, 321, 18 },
		{ 186, 322, 18 },
		{ 186, 323, 18 },
		{ 186, 324, 18 },
		{ 186, 325, 18 },
		{ 186, 326, 18 },
		{ 186, 327, 18 },
		{ 186, 328, 18 },
		{ 186, 329, 18 },
		{ 186, 330, 18 },
		{ 186, 331, 18 },
		{ 186, 332, 18 },
		{ 186, 333, 18 },
		{ 186, 334, 18 },
		{ 186, 335, 18 },
		{ 186, 336, 18 },
		{ 186, 337, 18 },
		{ 186, 338, 18 },
		{ 186, 339, 18 },
		{ 186, 340, 18 },
		{ 186, 341, 18 },
		{ 186, 342, 18 },
		{ 186, 343, 18 },
		{ 186, 344, 18 },
		{ 186, 345, 18 },
		{ 186, 346, 18 },
		{ 186, 347, 18 },
		{ 186, 348, 18 },
		{ 186, 349, 18 },
		{ 186, 350, 18 },
		{ 186, 351, 18 },
		{ 187, 310, 18 },
		{ 187, 311, 18 },
		{ 187, 312, 18 },
		{ 187, 313, 18 },
		{ 187, 314, 18 },
		{ 187, 315, 18 },
		{ 187, 316, 18 },
		{ 187, 317, 18 },
		{ 187, 318, 18 },
		{ 187, 319, 18 },
		{ 187, 320, 18 },
		{ 187, 321, 18 },
		{ 187, 322, 18 },
		{ 187, 323, 18 },
		{ 187, 324, 18 },
		{ 187, 325, 18 },
		{ 187, 326, 18 },
		{ 187, 327, 18 },
		{ 187, 328, 18 },
		{ 187, 329, 18 },
		{ 187, 330, 18 },
		{ 187, 331, 18 },
		{ 187, 332, 18 },
		{ 187, 333, 18 },
		{ 187, 334, 18 },
		{ 187, 335, 18 },
		{ 187, 336, 18 },
		{ 187, 337, 18 },
		{ 187, 338, 18 },
		{ 187, 339, 18 },
		{ 187, 340, 18 },
		{ 187, 341, 18 },
		{ 187, 342, 18 },
		{ 187, 343, 18 },
		{ 187, 344, 18 },
		{ 187, 345, 18 },
		{ 187, 346, 18 },
		{ 187, 347, 18 },
		{ 187, 348, 18 },
		{ 187, 349, 18 },
		{ 187, 350, 18 },
		{ 188, 309, 18 },
		{ 188, 310, 18 },
		{ 188, 311, 18 },
		{ 188, 312, 18 },
		{ 188, 313, 18 },
		{ 188, 314, 18 },
		{ 188, 315, 18 },
		{ 188, 316, 18 },
		{ 188, 317, 18 },
		{ 188, 318, 18 },
		{ 188, 319, 18 },
		{ 188, 320, 18 },
		{ 188, 321, 18 },
		{ 188, 322, 18 },
		{ 188, 323, 18 },
		{ 188, 324, 18 },
		{ 188, 325, 18 },
		{ 188, 326, 18 },
		{ 188, 327, 18 },
		{ 188, 328, 18 },
		{ 188, 329, 18 },
		{ 188, 330, 18 },
		{ 188, 331, 18 },
		{ 188, 332, 18 },
		{ 188, 333, 18 },
		{ 188, 334, 18 },
		{ 188, 335, 18 },
		{ 188, 336, 18 },
		{ 188, 337, 18 },
		{ 188, 338, 18 },
		{ 188, 339, 18 },
		{ 188, 340, 18 },
		{ 188, 341, 18 },
		{ 188, 342, 18 },
		{ 188, 343, 18 },
		{ 188, 344, 18 },
		{ 188, 345, 18 },
		{ 188, 346, 18 },
		{ 188, 347, 18 },
		{ 188, 348, 18 },
		{ 189, 309, 18 },
		{ 189, 310, 18 },
		{ 189, 311, 18 },
		{ 189, 312, 18 },
		{ 189, 313, 18 },
		{ 189, 314, 18 },
		{ 189, 315, 18 },
		{ 189, 316, 18 },
		{ 189, 317, 18 },
		{ 189, 318, 18 },
		{ 189, 319, 18 },
		{ 189, 320, 18 },
		{ 189, 321, 18 },
		{ 189, 322, 18 },
		{ 189, 323, 18 },
		{ 189, 324, 18 },
		{ 189, 325, 18 },
		{ 189, 326, 18 },
		{ 189, 327, 18 },
		{ 189, 328, 18 },
		{ 189, 329, 18 },
		{ 189, 330, 18 },
		{ 189, 331, 18 },
		{ 189, 332, 18 },
		{ 189, 333, 18 },
		{ 189, 334, 18 },
		{ 189, 335, 18 },
		{ 189, 336, 18 },
		{ 189, 337, 18 },
		{ 189, 338, 18 },
		{ 189, 339, 18 },
		{ 189, 340, 18 },
		{ 189, 341, 18 },
		{ 189, 342, 18 },
		{ 189, 343, 18 },
		{ 189, 344, 18 },
		{ 189, 345, 18 },
		{ 189, 346, 18 },
		{ 189, 347, 18 },
		{ 190, 308, 18 },
		{ 190, 309, 18 },
		{ 190, 310, 18 },
		{ 190, 311, 18 },
		{ 190, 312, 18 },
		{ 190, 313, 18 },
		{ 190, 314, 18 },
		{ 190, 315, 18 },
		{ 190, 316, 18 },
		{ 190, 317, 18 },
		{ 190, 318, 18 },
		{ 190, 319, 18 },
		{ 190, 320, 18 },
		{ 190, 321, 18 },
		{ 190, 322, 18 },
		{ 190, 323, 18 },
		{ 190, 324, 18 },
		{ 190, 325, 18 },
		{ 190, 326, 18 },
		{ 190, 327, 18 },
		{ 190, 328, 18 },
		{ 190, 329, 18 },
		{ 190, 330, 18 },
		{ 190, 331, 18 },
		{ 190, 332, 18 },
		{ 190, 333, 18 },
		{ 190, 334, 18 },
		{ 190, 335, 18 },
		{ 190, 336, 18 },
		{ 190, 337, 18 },
		{ 190, 338, 18 },
		{ 190, 339, 18 },
		{ 190, 340, 18 },
		{ 190, 341, 18 },
		{ 190, 342, 18 },
		{ 190, 343, 18 },
		{ 190, 344, 18 },
		{ 190, 345, 18 },
		{ 190, 346, 18 },
		{ 191, 307, 18 },
		{ 191, 308, 18 },
		{ 191, 309, 18 },
		{ 191, 310, 18 },
		{ 191, 311, 18 },
		{ 191, 312, 18 },
		{ 191, 313, 18 },
		{ 191, 314, 18 },
		{ 191, 315, 18 },
		{ 191, 316, 18 },
		{ 191, 317, 18 },
		{ 191, 318, 18 },
		{ 191, 319, 18 },
		{ 191, 320, 18 },
		{ 191, 321, 18 },
		{ 191, 322, 18 },
		{ 191, 323, 18 },
		{ 191, 324, 18 },
		{ 191, 325, 18 },
		{ 191, 326, 18 },
		{ 191, 327, 18 },
		{ 191, 328, 18 },
		{ 191, 329, 18 },
		{ 191, 330, 18 },
		{ 191, 331, 18 },
		{ 191, 332, 18 },
		{ 191, 333, 18 },
		{ 191, 334, 18 },
		{ 191, 335, 18 },
		{ 191, 336, 18 },
		{ 191, 337, 18 },
		{ 191, 338, 18 },
		{ 191, 339, 18 },
		{ 191, 340, 18 },
		{ 191, 341, 18 },
		{ 191, 342, 18 },
		{ 191, 343, 18 },
		{ 191, 344, 18 },
		{ 192, 306, 18 },
		{ 192, 307, 18 },
		{ 192, 308, 18 },
		{ 192, 309, 18 },
		{ 192, 310, 18 },
		{ 192, 311, 18 },
		{ 192, 312, 18 },
		{ 192, 313, 18 },
		{ 192, 314, 18 },
		{ 192, 315, 18 },
		{ 192, 316, 18 },
		{ 192, 317, 18 },
		{ 192, 318, 18 },
		{ 192, 319, 18 },
		{ 192, 320, 18 },
		{ 192, 321, 18 },
		{ 192, 322, 18 },
		{ 192, 323, 18 },
		{ 192, 324, 18 },
		{ 192, 325, 18 },
		{ 192, 326, 18 },
		{ 192, 327, 18 },
		{ 192, 328, 18 },
		{ 192, 329, 18 },
		{ 192, 330, 18 },
		{ 192, 331, 18 },
		{ 192, 332, 18 },
		{ 192, 333, 18 },
		{ 192, 334, 18 },
		{ 192, 335, 18 },
		{ 192, 336, 18 },
		{ 192, 337, 18 },
		{ 192, 338, 18 },
		{ 192, 339, 18 },
		{ 192, 340, 18 },
		{ 192, 341, 18 },
		{ 192, 342, 18 },
		{ 192, 343, 18 },
		{ 193, 306, 18 },
		{ 193, 307, 18 },
		{ 193, 308, 18 },
		{ 193, 309, 18 },
		{ 193, 310, 18 },
		{ 193, 311, 18 },
		{ 193, 312, 18 },
		{ 193, 313, 18 },
		{ 193, 314, 18 },
		{ 193, 315, 18 },
		{ 193, 316, 18 },
		{ 193, 317, 18 },
		{ 193, 318, 18 },
		{ 193, 319, 18 },
		{ 193, 320, 18 },
		{ 193, 321, 18 },
		{ 193, 322, 18 },
		{ 193, 323, 18 },
		{ 193, 324, 18 },
		{ 193, 325, 18 },
		{ 193, 326, 18 },
		{ 193, 327, 18 },
		{ 193, 328, 18 },
		{ 193, 329, 18 },
		{ 193, 330, 18 },
		{ 193, 331, 18 },
		{ 193, 332, 18 },
		{ 193, 333, 18 },
		{ 193, 334, 18 },
		{ 193, 335, 18 },
		{ 193, 336, 18 },
		{ 193, 337, 18 },
		{ 193, 338, 18 },
		{ 193, 339, 18 },
		{ 193, 340, 18 },
		{ 193, 341, 18 },
		{ 193, 342, 18 },
		{ 194, 305, 18 },
		{ 194, 306, 18 },
		{ 194, 307, 18 },
		{ 194, 308, 18 },
		{ 194, 309, 18 },
		{ 194, 310, 18 },
		{ 194, 311, 18 },
		{ 194, 312, 18 },
		{ 194, 313, 18 },
		{ 194, 314, 18 },
		{ 194, 315, 18 },
		{ 194, 316, 18 },
		{ 194, 317, 18 },
		{ 194, 318, 18 },
		{ 194, 319, 18 },
		{ 194, 320, 18 },
		{ 194, 321, 18 },
		{ 194, 322, 18 },
		{ 194, 323, 18 },
		{ 194, 324, 18 },
		{ 194, 325, 18 },
		{ 194, 326, 18 },
		{ 194, 327, 18 },
		{ 194, 328, 18 },
		{ 194, 329, 18 },
		{ 194, 330, 18 },
		{ 194, 331, 18 },
		{ 194, 332, 18 },
		{ 194, 333, 18 },
		{ 194, 334, 18 },
		{ 194, 335, 18 },
		{ 194, 336, 18 },
		{ 194, 337, 18 },
		{ 194, 338, 18 },
		{ 194, 339, 18 },
		{ 194, 340, 18 },
		{ 195, 304, 18 },
		{ 195, 305, 18 },
		{ 195, 306, 18 },
		{ 195, 307, 18 },
		{ 195, 308, 18 },
		{ 195, 309, 18 },
		{ 195, 310, 18 },
		{ 195, 311, 18 },
		{ 195, 312, 18 },
		{ 195, 313, 18 },
		{ 195, 314, 18 },
		{ 195, 315, 18 },
		{ 195, 316, 18 },
		{ 195, 317, 18 },
		{ 195, 318, 18 },
		{ 195, 319, 18 },
		{ 195, 320, 18 },
		{ 195, 321, 18 },
		{ 195, 322, 18 },
		{ 195, 323, 18 },
		{ 195, 324, 18 },
		{ 195, 325, 18 },
		{ 195, 326, 18 },
		{ 195, 327, 18 },
		{ 195, 328, 18 },
		{ 195, 329, 18 },
		{ 195, 330, 18 },
		{ 195, 331, 18 },
		{ 195, 332, 18 },
		{ 195, 333, 18 },
		{ 195, 334, 18 },
		{ 195, 335, 18 },
		{ 195, 336, 18 },
		{ 195, 337, 18 },
		{ 195, 338, 18 },
		{ 195, 339, 18 },
		{ 196, 303, 18 },
		{ 196, 304, 18 },
		{ 196, 305, 18 },
		{ 196, 306, 18 },
		{ 196, 307, 18 },
		{ 196, 308, 18 },
		{ 196, 309, 18 },
		{ 196, 310, 18 },
		{ 196, 311, 18 },
		{ 196, 312, 18 },
		{ 196, 313, 18 },
		{ 196, 314, 18 },
		{ 196, 315, 18 },
		{ 196, 316, 18 },
		{ 196, 317, 18 },
		{ 196, 318, 18 },
		{ 196, 319, 18 },
		{ 196, 320, 18 },
		{ 196, 321, 18 },
		{ 196, 322, 18 },
		{ 196, 323, 18 },
		{ 196, 324, 18 },
		{ 196, 325, 18 },
		{ 196, 326, 18 },
		{ 196, 327, 18 },
		{ 196, 328, 18 },
		{ 196, 329, 18 },
		{ 196, 330, 18 },
		{ 196, 331, 18 },
		{ 196, 332, 18 },
		{ 196, 333, 18 },
		{ 196, 334, 18 },
		{ 196, 335, 18 },
		{ 196, 336, 18 },
		{ 196, 337, 18 },
		{ 196, 338, 18 },
		{ 197, 303, 18 },
		{ 197, 304, 18 },
		{ 197, 305, 18 },
		{ 197, 306, 18 },
		{ 197, 307, 18 },
		{ 197, 308, 18 },
		{ 197, 309, 18 },
		{ 197, 310, 18 },
		{ 197, 311, 18 },
		{ 197, 312, 18 },
		{ 197, 313, 18 },
		{ 197, 314, 18 },
		{ 197, 315, 18 },
		{ 197, 316, 18 },
		{ 197, 317, 18 },
		{ 197, 318, 18 },
		{ 197, 319, 18 },
		{ 197, 320, 18 },
		{ 197, 321, 18 },
		{ 197, 322, 18 },
		{ 197, 323, 18 },
		{ 197, 324, 18 },
		{ 197, 325, 18 },
		{ 197, 326, 18 },
		{ 197, 327, 18 },
		{ 197, 328, 18 },
		{ 197, 329, 18 },
		{ 197, 330, 18 },
		{ 197, 331, 18 },
		{ 197, 332, 18 },
		{ 197, 333, 18 },
		{ 197, 334, 18 },
		{ 197, 335, 18 },
		{ 197, 336, 18 },
		{ 198, 302, 18 },
		{ 198, 303, 18 },
		{ 198, 304, 18 },
		{ 198, 305, 18 },
		{ 198, 306, 18 },
		{ 198, 307, 18 },
		{ 198, 308, 18 },
		{ 198, 309, 18 },
		{ 198, 310, 18 },
		{ 198, 311, 18 },
		{ 198, 312, 18 },
		{ 198, 313, 18 },
		{ 198, 314, 18 },
		{ 198, 315, 18 },
		{ 198, 316, 18 },
		{ 198, 317, 18 },
		{ 198, 318, 18 },
		{ 198, 319, 18 },
		{ 198, 320, 18 },
		{ 198, 321, 18 },
		{ 198, 322, 18 },
		{ 198, 323, 18 },
		{ 198, 324, 18 },
		{ 198, 325, 18 },
		{ 198, 326, 18 },
		{ 198, 327, 18 },
		{ 198, 328, 18 },
		{ 198, 329, 18 },
		{ 198, 330, 18 },
		{ 198, 331, 18 },
		{ 198, 332, 18 },
		{ 198, 333, 18 },
		{ 198, 334, 18 },
		{ 198, 335, 18 },
		{ 199, 301, 18 },
		{ 199, 302, 18 },
		{ 199, 303, 18 },
		{ 199, 304, 18 },
		{ 199, 305, 18 },
		{ 199, 306, 18 },
		{ 199, 307, 18 },
		{ 199, 308, 18 },
		{ 199, 309, 18 },
		{ 199, 310, 18 },
		{ 199, 311, 18 },
		{ 199, 312, 18 },
		{ 199, 313, 18 },
		{ 199, 314, 18 },
		{ 199, 315, 18 },
		{ 199, 316, 18 },
		{ 199, 317, 18 },
		{ 199, 318, 18 },
		{ 199, 319, 18 },
		{ 199, 320, 18 },
		{ 199, 321, 18 },
		{ 199, 322, 18 },
		{ 199, 323, 18 },
		{ 199, 324, 18 },
		{ 199, 325, 18 },
		{ 199, 326, 18 },
		{ 199, 327, 18 },
		{ 199, 328, 18 },
		{ 199, 329, 18 },
		{ 199, 330, 18 },
		{ 199, 331, 18 },
		{ 199, 332, 18 },
		{ 199, 333, 18 },
		{ 199, 334, 18 },
		{ 200, 300, 18 },
		{ 200, 301, 18 },
		{ 200, 302, 18 },
		{ 200, 303, 18 },
		{ 200, 304, 18 },
		{ 200, 305, 18 },
		{ 200, 306, 18 },
		{ 200, 307, 18 },
		{ 200, 308, 18 },
		{ 200, 309, 18 },
		{ 200, 310, 18 },
		{ 200, 311, 18 },
		{ 200, 312, 18 },
		{ 200, 313, 18 },
		{ 200, 314, 18 },
		{ 200, 315, 18 },
		{ 200, 316, 18 },
		{ 200, 317, 18 },
		{ 200, 318, 18 },
		{ 200, 319, 18 },
		{ 200, 320, 18 },
		{ 200, 321, 18 },
		{ 200, 322, 18 },
		{ 200, 323, 18 },
		{ 200, 324, 18 },
		{ 200, 325, 18 },
		{ 200, 326, 18 },
		{ 200, 327, 18 },
		{ 200, 328, 18 },
		{ 200, 329, 18 },
		{ 200, 330, 18 },
		{ 200, 331, 18 },
		{ 200, 332, 18 },
		{ 201, 300, 18 },
		{ 201, 301, 18 },
		{ 201, 302, 18 },
		{ 201, 303, 18 },
		{ 201, 304, 18 },
		{ 201, 305, 18 },
		{ 201, 306, 18 },
		{ 201, 307, 18 },
		{ 201, 308, 18 },
		{ 201, 309, 18 },
		{ 201, 310, 18 },
		{ 201, 311, 18 },
		{ 201, 312, 18 },
		{ 201, 313, 18 },
		{ 201, 314, 18 },
		{ 201, 315, 18 },
		{ 201, 316, 18 },
		{ 201, 317, 18 },
		{ 201, 318, 18 },
		{ 201, 319, 18 },
		{ 201, 320, 18 },
		{ 201, 321, 18 },
		{ 201, 322, 18 },
		{ 201, 323, 18 },
		{ 201, 324, 18 },
		{ 201, 325, 18 },
		{ 201, 326, 18 },
		{ 201, 327, 18 },
		{ 201, 328, 18 },
		{ 201, 329, 18 },
		{ 201, 330, 18 },
		{ 201, 331, 18 },
		{ 202, 299, 18 },
		{ 202, 300, 18 },
		{ 202, 301, 18 },
		{ 202, 302, 18 },
		{ 202, 303, 18 },
		{ 202, 304, 18 },
		{ 202, 305, 18 },
		{ 202, 306, 18 },
		{ 202, 307, 18 },
		{ 202, 308, 18 },
		{ 202, 309, 18 },
		{ 202, 310, 18 },
		{ 202, 311, 18 },
		{ 202, 312, 18 },
		{ 202, 313, 18 },
		{ 202, 314, 18 },
		{ 202, 315, 18 },
		{ 202, 316, 18 },
		{ 202, 317, 18 },
		{ 202, 318, 18 },
		{ 202, 319, 18 },
		{ 202, 320, 18 },
		{ 202, 321, 18 },
		{ 202, 322, 18 },
		{ 202, 323, 18 },
		{ 202, 324, 18 },
		{ 202, 325, 18 },
		{ 202, 326, 18 },
		{ 202, 327, 18 },
		{ 202, 328, 18 },
		{ 202, 329, 18 },
		{ 202, 330, 18 },
		{ 203, 298, 18 },
		{ 203, 299, 18 },
		{ 203, 300, 18 },
		{ 203, 301, 18 },
		{ 203, 302, 18 },
		{ 203, 303, 18 },
		{ 203, 304, 18 },
		{ 203, 305, 18 },
		{ 203, 306, 18 },
		{ 203, 307, 18 },
		{ 203, 308, 18 },
		{ 203, 309, 18 },
		{ 203, 310, 18 },
		{ 203, 311, 18 },
		{ 203, 312, 18 },
		{ 203, 313, 18 },
		{ 203, 314, 18 },
		{ 203, 315, 18 },
		{ 203, 316, 18 },
		{ 203, 317, 18 },
		{ 203, 318, 18 },
		{ 203, 319, 18 },
		{ 203, 320, 18 },
		{ 203, 321, 18 },
		{ 203, 322, 18 },
		{ 203, 323, 18 },
		{ 203, 324, 18 },
		{ 203, 325, 18 },
		{ 203, 326, 18 },
		{ 203, 327, 18 },
		{ 203, 328, 18 },
		{ 204, 297, 18 },
		{ 204, 298, 18 },
		{ 204, 299, 18 },
		{ 204, 300, 18 },
		{ 204, 301, 18 },
		{ 204, 302, 18 },
		{ 204, 303, 18 },
		{ 204, 304, 18 },
		{ 204, 305, 18 },
		{ 204, 306, 18 },
		{ 204, 307, 18 },
		{ 204, 308, 18 },
		{ 204, 309, 18 },
		{ 204, 310, 18 },
		{ 204, 311, 18 },
		{ 204, 312, 18 },
		{ 204, 313, 18 },
		{ 204, 314, 18 },
		{ 204, 315, 18 },
		{ 204, 316, 18 },
		{ 204, 317, 18 },
		{ 204, 318, 18 },
		{ 204, 319, 18 },
		{ 204, 320, 18 },
		{ 204, 321, 18 },
		{ 204, 322, 18 },
		{ 204, 323, 18 },
		{ 204, 324, 18 },
		{ 204, 325, 18 },
		{ 204, 326, 18 },
		{ 204, 327, 18 },
		{ 205, 297, 18 },
		{ 205, 298, 18 },
		{ 205, 299, 18 },
		{ 205, 300, 18 },
		{ 205, 301, 18 },
		{ 205, 302, 18 },
		{ 205, 303, 18 },
		{ 205, 304, 18 },
		{ 205, 305, 18 },
		{ 205, 306, 18 },
		{ 205, 307, 18 },
		{ 205, 308, 18 },
		{ 205, 309, 18 },
		{ 205, 310, 18 },
		{ 205, 311, 18 },
		{ 205, 312, 18 },
		{ 205, 313, 18 },
		{ 205, 314, 18 },
		{ 205, 315, 18 },
		{ 205, 316, 18 },
		{ 205, 317, 18 },
		{ 205, 318, 18 },
		{ 205, 319, 18 },
		{ 205, 320, 18 },
		{ 205, 321, 18 },
		{ 205, 322, 18 },
		{ 205, 323, 18 },
		{ 205, 324, 18 },
		{ 205, 325, 18 },
		{ 205, 326, 18 },
		{ 206, 296, 18 },
		{ 206, 297, 18 },
		{ 206, 298, 18 },
		{ 206, 299, 18 },
		{ 206, 300, 18 },
		{ 206, 301, 18 },
		{ 206, 302, 18 },
		{ 206, 303, 18 },
		{ 206, 304, 18 },
		{ 206, 305, 18 },
		{ 206, 306, 18 },
		{ 206, 307, 18 },
		{ 206, 308, 18 },
		{ 206, 309, 18 },
		{ 206, 310, 18 },
		{ 206, 311, 18 },
		{ 206, 312, 18 },
		{ 206, 313, 18 },
		{ 206, 314, 18 },
		{ 206, 315, 18 },
		{ 206, 316, 18 },
		{ 206, 317, 18 },
		{ 206, 318, 18 },
		{ 206, 319, 18 },
		{ 206, 320, 18 },
		{ 206, 321, 18 },
		{ 206, 322, 18 },
		{ 206, 323, 18 },
		{ 206, 324, 18 },
		{ 207, 295, 18 },
		{ 207, 296, 18 },
		{ 207, 297, 18 },
		{ 207, 298, 18 },
		{ 207, 299, 18 },
		{ 207, 300, 18 },
		{ 207, 301, 18 },
		{ 207, 302, 18 },
		{ 207, 303, 18 },
		{ 207, 304, 18 },
		{ 207, 305, 18 },
		{ 207, 306, 18 },
		{ 207, 307, 18 },
		{ 207, 308, 18 },
		{ 207, 309, 18 },
		{ 207, 310, 18 },
		{ 207, 311, 18 },
		{ 207, 312, 18 },
		{ 207, 313, 18 },
		{ 207, 314, 18 },
		{ 207, 315, 18 },
		{ 207, 316, 18 },
		{ 207, 317, 18 },
		{ 207, 318, 18 },
		{ 207, 319, 18 },
		{ 207, 320, 18 },
		{ 207, 321, 18 },
		{ 207, 322, 18 },
		{ 207, 323, 18 },
		{ 208, 294, 18 },
		{ 208, 295, 18 },
		{ 208, 296, 18 },
		{ 208, 297, 18 },
		{ 208, 298, 18 },
		{ 208, 299, 18 },
		{ 208, 300, 18 },
		{ 208, 301, 18 },
		{ 208, 302, 18 },
		{ 208, 303, 18 },
		{ 208, 304, 18 },
		{ 208, 305, 18 },
		{ 208, 306, 18 },
		{ 208, 307, 18 },
		{ 208, 308, 18 },
		{ 208, 309, 18 },
		{ 208, 310, 18 },
		{ 208, 311, 18 },
		{ 208, 312, 18 },
		{ 208, 313, 18 },
		{ 208, 314, 18 },
		{ 208, 315, 18 },
		{ 208, 316, 18 },
		{ 208, 317, 18 },
		{ 208, 318, 18 },
		{ 208, 319, 18 },
		{ 208, 320, 18 },
		{ 208, 321, 18 },
		{ 208, 322, 18 },
		{ 209, 294, 18 },
		{ 209, 295, 18 },
		{ 209, 296, 18 },
		{ 209, 297, 18 },
		{ 209, 298, 18 },
		{ 209, 299, 18 },
		{ 209, 300, 18 },
		{ 209, 301, 18 },
		{ 209, 302, 18 },
		{ 209, 303, 18 },
		{ 209, 304, 18 },
		{ 209, 305, 18 },
		{ 209, 306, 18 },
		{ 209, 307, 18 },
		{ 209, 308, 18 },
		{ 209, 309, 18 },
		{ 209, 310, 18 },
		{ 209, 311, 18 },
		{ 209, 312, 18 },
		{ 209, 313, 18 },
		{ 209, 314, 18 },
		{ 209, 315, 18 },
		{ 209, 316, 18 },
		{ 209, 317, 18 },
		{ 209, 318, 18 },
		{ 209, 319, 18 },
		{ 209, 320, 18 },
		{ 210, 293, 18 },
		{ 210, 294, 18 },
		{ 210, 295, 18 },
		{ 210, 296, 18 },
		{ 210, 297, 18 },
		{ 210, 298, 18 },
		{ 210, 299, 18 },
		{ 210, 300, 18 },
		{ 210, 301, 18 },
		{ 210, 302, 18 },
		{ 210, 303, 18 },
		{ 210, 304, 18 },
		{ 210, 305, 18 },
		{ 210, 306, 18 },
		{ 210, 307, 18 },
		{ 210, 308, 18 },
		{ 210, 309, 18 },
		{ 210, 310, 18 },
		{ 210, 311, 18 },
		{ 210, 312, 18 },
		{ 210, 313, 18 },
		{ 210, 314, 18 },
		{ 210, 315, 18 },
		{ 210, 316, 18 },
		{ 210, 317, 18 },
		{ 210, 318, 18 },
		{ 210, 319, 18 },
		{ 211, 292, 18 },
		{ 211, 293, 18 },
		{ 211, 294, 18 },
		{ 211, 295, 18 },
		{ 211, 296, 18 },
		{ 211, 297, 18 },
		{ 211, 298, 18 },
		{ 211, 299, 18 },
		{ 211, 300, 18 },
		{ 211, 301, 18 },
		{ 211, 302, 18 },
		{ 211, 303, 18 },
		{ 211, 304, 18 },
		{ 211, 305, 18 },
		{ 211, 306, 18 },
		{ 211, 307, 18 },
		{ 211, 308, 18 },
		{ 211, 309, 18 },
		{ 211, 310, 18 },
		{ 211, 311, 18 },
		{ 211, 312, 18 },
		{ 211, 313, 18 },
		{ 211, 314, 18 },
		{ 211, 315, 18 },
		{ 211, 316, 18 },
		{ 211, 317, 18 },
		{ 211, 318, 18 },
		{ 212, 291, 18 },
		{ 212, 292, 18 },
		{ 212, 293, 18 },
		{ 212, 294, 18 },
		{ 212, 295, 18 },
		{ 212, 296, 18 },
		{ 212, 297, 18 },
		{ 212, 298, 18 },
		{ 212, 299, 18 },
		{ 212, 300, 18 },
		{ 212, 301, 18 },
		{ 212, 302, 18 },
		{ 212, 303, 18 },
		{ 212, 304, 18 },
		{ 212, 305, 18 },
		{ 212, 306, 18 },
		{ 212, 307, 18 },
		{ 212, 308, 18 },
		{ 212, 309, 18 },
		{ 212, 310, 18 },
		{ 212, 311, 18 },
		{ 212, 312, 18 },
		{ 212, 313, 18 },
		{ 212, 314, 18 },
		{ 212, 315, 18 },
		{ 212, 316, 18 },
		{ 213, 291, 18 },
		{ 213, 292, 18 },
		{ 213, 293, 18 },
		{ 213, 294, 18 },
		{ 213, 295, 18 },
		{ 213, 296, 18 },
		{ 213, 297, 18 },
		{ 213, 298, 18 },
		{ 213, 299, 18 },
		{ 213, 300, 18 },
		{ 213, 301, 18 },
		{ 213, 302, 18 },
		{ 213, 303, 18 },
		{ 213, 304, 18 },
		{ 213, 305, 18 },
		{ 213, 306, 18 },
		{ 213, 307, 18 },
		{ 213, 308, 18 },
		{ 213, 309, 18 },
		{ 213, 310, 18 },
		{ 213, 311, 18 },
		{ 213, 312, 18 },
		{ 213, 313, 18 },
		{ 213, 314, 18 },
		{ 213, 315, 18 },
		{ 214, 290, 18 },
		{ 214, 291, 18 },
		{ 214, 292, 18 },
		{ 214, 293, 18 },
		{ 214, 294, 18 },
		{ 214, 295, 18 },
		{ 214, 296, 18 },
		{ 214, 297, 18 },
		{ 214, 298, 18 },
		{ 214, 299, 18 },
		{ 214, 300, 18 },
		{ 214, 301, 18 },
		{ 214, 302, 18 },
		{ 214, 303, 18 },
		{ 214, 304, 18 },
		{ 214, 305, 18 },
		{ 214, 306, 18 },
		{ 214, 307, 18 },
		{ 214, 308, 18 },
		{ 214, 309, 18 },
		{ 214, 310, 18 },
		{ 214, 311, 18 },
		{ 214, 312, 18 },
		{ 214, 313, 18 },
		{ 214, 314, 18 },
		{ 215, 289, 18 },
		{ 215, 290, 18 },
		{ 215, 291, 18 },
		{ 215, 292, 18 },
		{ 215, 293, 18 },
		{ 215, 294, 18 },
		{ 215, 295, 18 },
		{ 215, 296, 18 },
		{ 215, 297, 18 },
		{ 215, 298, 18 },
		{ 215, 299, 18 },
		{ 215, 300, 18 },
		{ 215, 301, 18 },
		{ 215, 302, 18 },
		{ 215, 303, 18 },
		{ 215, 304, 18 },
		{ 215, 305, 18 },
		{ 215, 306, 18 },
		{ 215, 307, 18 },
		{ 215, 308, 18 },
		{ 215, 309, 18 },
		{ 215, 310, 18 },
		{ 215, 311, 18 },
		{ 215, 312, 18 },
		{ 216, 288, 18 },
		{ 216, 289, 18 },
		{ 216, 290, 18 },
		{ 216, 291, 18 },
		{ 216, 292, 18 },
		{ 216, 293, 18 },
		{ 216, 294, 18 },
		{ 216, 295, 18 },
		{ 216, 296, 18 },
		{ 216, 297, 18 },
		{ 216, 298, 18 },
		{ 216, 299, 18 },
		{ 216, 300, 18 },
		{ 216, 301, 18 },
		{ 216, 302, 18 },
		{ 216, 303, 18 },
		{ 216, 304, 18 },
		{ 216, 305, 18 },
		{ 216, 306, 18 },
		{ 216, 307, 18 },
		{ 216, 308, 18 },
		{ 216, 309, 18 },
		{ 216, 310, 18 },
		{ 216, 311, 18 },
		{ 217, 288, 18 },
		{ 217, 289, 18 },
		{ 217, 290, 18 },
		{ 217, 291, 18 },
		{ 217, 292, 18 },
		{ 217, 293, 18 },
		{ 217, 294, 18 },
		{ 217, 295, 18 },
		{ 217, 296, 18 },
		{ 217, 297, 18 },
		{ 217, 298, 18 },
		{ 217, 299, 18 },
		{ 217, 300, 18 },
		{ 217, 301, 18 },
		{ 217, 302, 18 },
		{ 217, 303, 18 },
		{ 217, 304, 18 },
		{ 217, 305, 18 },
		{ 217, 306, 18 },
		{ 217, 307, 18 },
		{ 217, 308, 18 },
		{ 217, 309, 18 },
		{ 217, 310, 18 },
		{ 218, 287, 18 },
		{ 218, 288, 18 },
		{ 218, 289, 18 },
		{ 218, 290, 18 },
		{ 218, 291, 18 },
		{ 218, 292, 18 },
		{ 218, 293, 18 },
		{ 218, 294, 18 },
		{ 218, 295, 18 },
		{ 218, 296, 18 },
		{ 218, 297, 18 },
		{ 218, 298, 18 },
		{ 218, 299, 18 },
		{ 218, 300, 18 },
		{ 218, 301, 18 },
		{ 218, 302, 18 },
		{ 218, 303, 18 },
		{ 218, 304, 18 },
		{ 218, 305, 18 },
		{ 218, 306, 18 },
		{ 218, 307, 18 },
		{ 218, 308, 18 },
		{ 219, 286, 18 },
		{ 219, 287, 18 },
		{ 219, 288, 18 },
		{ 219, 289, 18 },
		{ 219, 290, 18 },
		{ 219, 291, 18 },
		{ 219, 292, 18 },
		{ 219, 293, 18 },
		{ 219, 294, 18 },
		{ 219, 295, 18 },
		{ 219, 296, 18 },
		{ 219, 297, 18 },
		{ 219, 298, 18 },
		{ 219, 299, 18 },
		{ 219, 300, 18 },
		{ 219, 301, 18 },
		{ 219, 302, 18 },
		{ 219, 303, 18 },
		{ 219, 304, 18 },
		{ 219, 305, 18 },
		{ 219, 306, 18 },
		{ 219, 307, 18 },
		{ 220, 285, 18 },
		{ 220, 286, 18 },
		{ 220, 287, 18 },
		{ 220, 288, 18 },
		{ 220, 289, 18 },
		{ 220, 290, 18 },
		{ 220, 291, 18 },
		{ 220, 292, 18 },
		{ 220, 293, 18 },
		{ 220, 294, 18 },
		{ 220, 295, 18 },
		{ 220, 296, 18 },
		{ 220, 297, 18 },
		{ 220, 298, 18 },
		{ 220, 299, 18 },
		{ 220, 300, 18 },
		{ 220, 301, 18 },
		{ 220, 302, 18 },
		{ 220, 303, 18 },
		{ 220, 304, 18 },
		{ 220, 305, 18 },
		{ 220, 306, 18 },
		{ 221, 285, 18 },
		{ 221, 286, 18 },
		{ 221, 287, 18 },
		{ 221, 288, 18 },
		{ 221, 289, 18 },
		{ 221, 290, 18 },
		{ 221, 291, 18 },
		{ 221, 292, 18 },
		{ 221, 293, 18 },
		{ 221, 294, 18 },
		{ 221, 295, 18 },
		{ 221, 296, 18 },
		{ 221, 297, 18 },
		{ 221, 298, 18 },
		{ 221, 299, 18 },
		{ 221, 300, 18 },
		{ 221, 301, 18 },
		{ 221, 302, 18 },
		{ 221, 303, 18 },
		{ 221, 304, 18 },
		{ 222, 284, 18 },
		{ 222, 285, 18 },
		{ 222, 286, 18 },
		{ 222, 287, 18 },
		{ 222, 288, 18 },
		{ 222, 289, 18 },
		{ 222, 290, 18 },
		{ 222, 291, 18 },
		{ 222, 292, 18 },
		{ 222, 293, 18 },
		{ 222, 294, 18 },
		{ 222, 295, 18 },
		{ 222, 296, 18 },
		{ 222, 297, 18 },
		{ 222, 298, 18 },
		{ 222, 299, 18 },
		{ 222, 300, 18 },
		{ 222, 301, 18 },
		{ 222, 302, 18 },
		{ 222, 303, 18 },
		{ 223, 283, 18 },
		{ 223, 284, 18 },
		{ 223, 285, 18 },
		{ 223, 286, 18 },
		{ 223, 287, 18 },
		{ 223, 288, 18 },
		{ 223, 289, 18 },
		{ 223, 290, 18 },
		{ 223, 291, 18 },
		{ 223, 292, 18 },
		{ 223, 293, 18 },
		{ 223, 294, 18 },
		{ 223, 295, 18 },
		{ 223, 296, 18 },
		{ 223, 297, 18 },
		{ 223, 298, 18 },
		{ 223, 299, 18 },
		{ 223, 300, 18 },
		{ 223, 301, 18 },
		{ 223, 302, 18 },
		{ 224, 282, 18 },
		{ 224, 283, 18 },
		{ 224, 284, 18 },
		{ 224, 285, 18 },
		{ 224, 286, 18 },
		{ 224, 287, 18 },
		{ 224, 288, 18 },
		{ 224, 289, 18 },
		{ 224, 290, 18 },
		{ 224, 291, 18 },
		{ 224, 292, 18 },
		{ 224, 293, 18 },
		{ 224, 294, 18 },
		{ 224, 295, 18 },
		{ 224, 296, 18 },
		{ 224, 297, 18 },
		{ 224, 298, 18 },
		{ 224, 299, 18 },
		{ 224, 300, 18 },
		{ 225, 282, 18 },
		{ 225, 283, 18 },
		{ 225, 284, 18 },
		{ 225, 285, 18 },
		{ 225, 286, 18 },
		{ 225, 287, 18 },
		{ 225, 288, 18 },
		{ 225, 289, 18 },
		{ 225, 290, 18 },
		{ 225, 291, 18 },
		{ 225, 292, 18 },
		{ 225, 293, 18 },
		{ 225, 294, 18 },
		{ 225, 295, 18 },
		{ 225, 296, 18 },
		{ 225, 297, 18 },
		{ 225, 298, 18 },
		{ 225, 299, 18 },
		{ 226, 281, 18 },
		{ 226, 282, 18 },
		{ 226, 283, 18 },
		{ 226, 284, 18 },
		{ 226, 285, 18 },
		{ 226, 286, 18 },
		{ 226, 287, 18 },
		{ 226, 288, 18 },
		{ 226, 289, 18 },
		{ 226, 290, 18 },
		{ 226, 291, 18 },
		{ 226, 292, 18 },
		{ 226, 293, 18 },
		{ 226, 294, 18 },
		{ 226, 295, 18 },
		{ 226, 296, 18 },
		{ 226, 297, 18 },
		{ 226, 298, 18 },
		{ 227, 280, 18 },
		{ 227, 281, 18 },
		{ 227, 282, 18 },
		{ 227, 283, 18 },
		{ 227, 284, 18 },
		{ 227, 285, 18 },
		{ 227, 286, 18 },
		{ 227, 287, 18 },
		{ 227, 288, 18 },
		{ 227, 289, 18 },
		{ 227, 290, 18 },
		{ 227, 291, 18 },
		{ 227, 292, 18 },
		{ 227, 293, 18 },
		{ 227, 294, 18 },
		{ 227, 295, 18 },
		{ 227, 296, 18 },
		{ 228, 279, 18 },
		{ 228, 280, 18 },
		{ 228, 281, 18 },
		{ 228, 282, 18 },
		{ 228, 283, 18 },
		{ 228, 284, 18 },
		{ 228, 285, 18 },
		{ 228, 286, 18 },
		{ 228, 287, 18 },
		{ 228, 288, 18 },
		{ 228, 289, 18 },
		{ 228, 290, 18 },
		{ 228, 291, 18 },
		{ 228, 292, 18 },
		{ 228, 293, 18 },
		{ 228, 294, 18 },
		{ 228, 295, 18 },
		{ 229, 279, 18 },
		{ 229, 280, 18 },
		{ 229, 281, 18 },
		{ 229, 282, 18 },
		{ 229, 283, 18 },
		{ 229, 284, 18 },
		{ 229, 285, 18 },
		{ 229, 286, 18 },
		{ 229, 287, 18 },
		{ 229, 288, 18 },
		{ 229, 289, 18 },
		{ 229, 290, 18 },
		{ 229, 291, 18 },
		{ 229, 292, 18 },
		{ 229, 293, 18 },
		{ 229, 294, 18 },
		{ 230, 278, 18 },
		{ 230, 279, 18 },
		{ 230, 280, 18 },
		{ 230, 281, 18 },
		{ 230, 282, 18 },
		{ 230, 283, 18 },
		{ 230, 284, 18 },
		{ 230, 285, 18 },
		{ 230, 286, 18 },
		{ 230, 287, 18 },
		{ 230, 288, 18 },
		{ 230, 289, 18 },
		{ 230, 290, 18 },
		{ 230, 291, 18 },
		{ 230, 292, 18 },
		{ 231, 277, 18 },
		{ 231, 278, 18 },
		{ 231, 279, 18 },
		{ 231, 280, 18 },
		{ 231, 281, 18 },
		{ 231, 282, 18 },
		{ 231, 283, 18 },
		{ 231, 284, 18 },
		{ 231, 285, 18 },
		{ 231, 286, 18 },
		{ 231, 287, 18 },
		{ 231, 288, 18 },
		{ 231, 289, 18 },
		{ 231, 290, 18 },
		{ 231, 291, 18 },
		{ 232, 276, 18 },
		{ 232, 277, 18 },
		{ 232, 278, 18 },
		{ 232, 279, 18 },
		{ 232, 280, 18 },
		{ 232, 281, 18 },
		{ 232, 282, 18 },
		{ 232, 283, 18 },
		{ 232, 284, 18 },
		{ 232, 285, 18 },
		{ 232, 286, 18 },
		{ 232, 287, 18 },
		{ 232, 288, 18 },
		{ 232, 289, 18 },
		{ 232, 290, 18 },
		{ 233, 276, 18 },
		{ 233, 277, 18 },
		{ 233, 278, 18 },
		{ 233, 279, 18 },
		{ 233, 280, 18 },
		{ 233, 281, 18 },
		{ 233, 282, 18 },
		{ 233, 283, 18 },
		{ 233, 284, 18 },
		{ 233, 285, 18 },
		{ 233, 286, 18 },
		{ 233, 287, 18 },
		{ 233, 288, 18 },
		{ 234, 275, 18 },
		{ 234, 276, 18 },
		{ 234, 277, 18 },
		{ 234, 278, 18 },
		{ 234, 279, 18 },
		{ 234, 280, 18 },
		{ 234, 281, 18 },
		{ 234, 282, 18 },
		{ 234, 283, 18 },
		{ 234, 284, 18 },
		{ 234, 285, 18 },
		{ 234, 286, 18 },
		{ 234, 287, 18 },
		{ 235, 274, 18 },
		{ 235, 275, 18 },
		{ 235, 276, 18 },
		{ 235, 277, 18 },
		{ 235, 278, 18 },
		{ 235, 279, 18 },
		{ 235, 280, 18 },
		{ 235, 281, 18 },
		{ 235, 282, 18 },
		{ 235, 283, 18 },
		{ 235, 284, 18 },
		{ 235, 285, 18 },
		{ 235, 286, 18 },
		{ 236, 273, 18 },
		{ 236, 274, 18 },
		{ 236, 275, 18 },
		{ 236, 276, 18 },
		{ 236, 277, 18 },
		{ 236, 278, 18 },
		{ 236, 279, 18 },
		{ 236, 280, 18 },
		{ 236, 281, 18 },
		{ 236, 282, 18 },
		{ 236, 283, 18 },
		{ 236, 284, 18 },
		{ 237, 273, 18 },
		{ 237, 274, 18 },
		{ 237, 275, 18 },
		{ 237, 276, 18 },
		{ 237, 277, 18 },
		{ 237, 278, 18 },
		{ 237, 279, 18 },
		{ 237, 280, 18 },
		{ 237, 281, 18 },
		{ 237, 282, 18 },
		{ 237, 283, 18 },
		{ 238, 272, 18 },
		{ 238, 273, 18 },
		{ 238, 274, 18 },
		{ 238, 275, 18 },
		{ 238, 276, 18 },
		{ 238, 277, 18 },
		{ 238, 278, 18 },
		{ 238, 279, 18 },
		{ 238, 280, 18 },
		{ 238, 281, 18 },
		{ 238, 282, 18 },
		{ 239, 271, 18 },
		{ 239, 272, 18 },
		{ 239, 273, 18 },
		{ 239, 274, 18 },
		{ 239, 275, 18 },
		{ 239, 276, 18 },
		{ 239, 277, 18 },
		{ 239, 278, 18 },
		{ 239, 279, 18 },
		{ 239, 280, 18 },
		{ 240, 270, 18 },
		{ 240, 271, 18 },
		{ 240, 272, 18 },
		{ 240, 273, 18 },
		{ 240, 274, 18 },
		{ 240, 275, 18 },
		{ 240, 276, 18 },
		{ 240, 277, 18 },
		{ 240, 278, 18 },
		{ 240, 279, 18 },
		{ 241, 274, 18 },
		{ 241, 275, 18 },
		{ 241, 276, 18 },
		{ 241, 277, 18 },
		{ 241, 278, 18 },
		{ 242, 274, 18 },
		{ 242, 275, 18 },
		{ 242, 276, 18 },
		{ 243, 274, 18 },
		{ 243, 275, 18 },
		{ 244, 274, 18 },
		{ 270, 240, 18 },
		{ 271, 239, 18 },
		{ 271, 240, 18 },
		{ 272, 238, 18 },
		{ 272, 239, 18 },
		{ 272, 240, 18 },
		{ 273, 236, 18 },
		{ 273, 237, 18 },
		{ 273, 238, 18 },
		{ 273, 239, 18 },
		{ 273, 240, 18 },
		{ 274, 235, 18 },
		{ 274, 236, 18 },
		{ 274, 237, 18 },
		{ 274, 238, 18 },
		{ 274, 239, 18 },
		{ 274, 240, 18 },
		{ 274, 241, 18 },
		{ 274, 242, 18 },
		{ 274, 243, 18 },
		{ 274, 244, 18 },
		{ 275, 234, 18 },
		{ 275, 235, 18 },
		{ 275, 236, 18 },
		{ 275, 237, 18 },
		{ 275, 238, 18 },
		{ 275, 239, 18 },
		{ 275, 240, 18 },
		{ 275, 241, 18 },
		{ 275, 242, 18 },
		{ 275, 243, 18 },
		{ 276, 232, 18 },
		{ 276, 233, 18 },
		{ 276, 234, 18 },
		{ 276, 235, 18 },
		{ 276, 236, 18 },
		{ 276, 237, 18 },
		{ 276, 238, 18 },
		{ 276, 239, 18 },
		{ 276, 240, 18 },
		{ 276, 241, 18 },
		{ 276, 242, 18 },
		{ 277, 231, 18 },
		{ 277, 232, 18 },
		{ 277, 233, 18 },
		{ 277, 234, 18 },
		{ 277, 235, 18 },
		{ 277, 236, 18 },
		{ 277, 237, 18 },
		{ 277, 238, 18 },
		{ 277, 239, 18 },
		{ 277, 240, 18 },
		{ 277, 241, 18 },
		{ 278, 230, 18 },
		{ 278, 231, 18 },
		{ 278, 232, 18 },
		{ 278, 233, 18 },
		{ 278, 234, 18 },
		{ 278, 235, 18 },
		{ 278, 236, 18 },
		{ 278, 237, 18 },
		{ 278, 238, 18 },
		{ 278, 239, 18 },
		{ 278, 240, 18 },
		{ 278, 241, 18 },
		{ 279, 228, 18 },
		{ 279, 229, 18 },
		{ 279, 230, 18 },
		{ 279, 231, 18 },
		{ 279, 232, 18 },
		{ 279, 233, 18 },
		{ 279, 234, 18 },
		{ 279, 235, 18 },
		{ 279, 236, 18 },
		{ 279, 237, 18 },
		{ 279, 238, 18 },
		{ 279, 239, 18 },
		{ 279, 240, 18 },
		{ 280, 227, 18 },
		{ 280, 228, 18 },
		{ 280, 229, 18 },
		{ 280, 230, 18 },
		{ 280, 231, 18 },
		{ 280, 232, 18 },
		{ 280, 233, 18 },
		{ 280, 234, 18 },
		{ 280, 235, 18 },
		{ 280, 236, 18 },
		{ 280, 237, 18 },
		{ 280, 238, 18 },
		{ 280, 239, 18 },
		{ 281, 226, 18 },
		{ 281, 227, 18 },
		{ 281, 228, 18 },
		{ 281, 229, 18 },
		{ 281, 230, 18 },
		{ 281, 231, 18 },
		{ 281, 232, 18 },
		{ 281, 233, 18 },
		{ 281, 234, 18 },
		{ 281, 235, 18 },
		{ 281, 236, 18 },
		{ 281, 237, 18 },
		{ 281, 238, 18 },
		{ 282, 224, 18 },
		{ 282, 225, 18 },
		{ 282, 226, 18 },
		{ 282, 227, 18 },
		{ 282, 228, 18 },
		{ 282, 229, 18 },
		{ 282, 230, 18 },
		{ 282, 231, 18 },
		{ 282, 232, 18 },
		{ 282, 233, 18 },
		{ 282, 234, 18 },
		{ 282, 235, 18 },
		{ 282, 236, 18 },
		{ 282, 237, 18 },
		{ 282, 238, 18 },
		{ 283, 223, 18 },
		{ 283, 224, 18 },
		{ 283, 225, 18 },
		{ 283, 226, 18 },
		{ 283, 227, 18 },
		{ 283, 228, 18 },
		{ 283, 229, 18 },
		{ 283, 230, 18 },
		{ 283, 231, 18 },
		{ 283, 232, 18 },
		{ 283, 233, 18 },
		{ 283, 234, 18 },
		{ 283, 235, 18 },
		{ 283, 236, 18 },
		{ 283, 237, 18 },
		{ 284, 222, 18 },
		{ 284, 223, 18 },
		{ 284, 224, 18 },
		{ 284, 225, 18 },
		{ 284, 226, 18 },
		{ 284, 227, 18 },
		{ 284, 228, 18 },
		{ 284, 229, 18 },
		{ 284, 230, 18 },
		{ 284, 231, 18 },
		{ 284, 232, 18 },
		{ 284, 233, 18 },
		{ 284, 234, 18 },
		{ 284, 235, 18 },
		{ 284, 236, 18 },
		{ 285, 220, 18 },
		{ 285, 221, 18 },
		{ 285, 222, 18 },
		{ 285, 223, 18 },
		{ 285, 224, 18 },
		{ 285, 225, 18 },
		{ 285, 226, 18 },
		{ 285, 227, 18 },
		{ 285, 228, 18 },
		{ 285, 229, 18 },
		{ 285, 230, 18 },
		{ 285, 231, 18 },
		{ 285, 232, 18 },
		{ 285, 233, 18 },
		{ 285, 234, 18 },
		{ 285, 235, 18 },
		{ 286, 219, 18 },
		{ 286, 220, 18 },
		{ 286, 221, 18 },
		{ 286, 222, 18 },
		{ 286, 223, 18 },
		{ 286, 224, 18 },
		{ 286, 225, 18 },
		{ 286, 226, 18 },
		{ 286, 227, 18 },
		{ 286, 228, 18 },
		{ 286, 229, 18 },
		{ 286, 230, 18 },
		{ 286, 231, 18 },
		{ 286, 232, 18 },
		{ 286, 233, 18 },
		{ 286, 234, 18 },
		{ 286, 235, 18 },
		{ 287, 218, 18 },
		{ 287, 219, 18 },
		{ 287, 220, 18 },
		{ 287, 221, 18 },
		{ 287, 222, 18 },
		{ 287, 223, 18 },
		{ 287, 224, 18 },
		{ 287, 225, 18 },
		{ 287, 226, 18 },
		{ 287, 227, 18 },
		{ 287, 228, 18 },
		{ 287, 229, 18 },
		{ 287, 230, 18 },
		{ 287, 231, 18 },
		{ 287, 232, 18 },
		{ 287, 233, 18 },
		{ 287, 234, 18 },
		{ 288, 216, 18 },
		{ 288, 217, 18 },
		{ 288, 218, 18 },
		{ 288, 219, 18 },
		{ 288, 220, 18 },
		{ 288, 221, 18 },
		{ 288, 222, 18 },
		{ 288, 223, 18 },
		{ 288, 224, 18 },
		{ 288, 225, 18 },
		{ 288, 226, 18 },
		{ 288, 227, 18 },
		{ 288, 228, 18 },
		{ 288, 229, 18 },
		{ 288, 230, 18 },
		{ 288, 231, 18 },
		{ 288, 232, 18 },
		{ 288, 233, 18 },
		{ 289, 215, 18 },
		{ 289, 216, 18 },
		{ 289, 217, 18 },
		{ 289, 218, 18 },
		{ 289, 219, 18 },
		{ 289, 220, 18 },
		{ 289, 221, 18 },
		{ 289, 222, 18 },
		{ 289, 223, 18 },
		{ 289, 224, 18 },
		{ 289, 225, 18 },
		{ 289, 226, 18 },
		{ 289, 227, 18 },
		{ 289, 228, 18 },
		{ 289, 229, 18 },
		{ 289, 230, 18 },
		{ 289, 231, 18 },
		{ 289, 232, 18 },
		{ 290, 214, 18 },
		{ 290, 215, 18 },
		{ 290, 216, 18 },
		{ 290, 217, 18 },
		{ 290, 218, 18 },
		{ 290, 219, 18 },
		{ 290, 220, 18 },
		{ 290, 221, 18 },
		{ 290, 222, 18 },
		{ 290, 223, 18 },
		{ 290, 224, 18 },
		{ 290, 225, 18 },
		{ 290, 226, 18 },
		{ 290, 227, 18 },
		{ 290, 228, 18 },
		{ 290, 229, 18 },
		{ 290, 230, 18 },
		{ 290, 231, 18 },
		{ 290, 232, 18 },
		{ 291, 212, 18 },
		{ 291, 213, 18 },
		{ 291, 214, 18 },
		{ 291, 215, 18 },
		{ 291, 216, 18 },
		{ 291, 217, 18 },
		{ 291, 218, 18 },
		{ 291, 219, 18 },
		{ 291, 220, 18 },
		{ 291, 221, 18 },
		{ 291, 222, 18 },
		{ 291, 223, 18 },
		{ 291, 224, 18 },
		{ 291, 225, 18 },
		{ 291, 226, 18 },
		{ 291, 227, 18 },
		{ 291, 228, 18 },
		{ 291, 229, 18 },
		{ 291, 230, 18 },
		{ 291, 231, 18 },
		{ 292, 211, 18 },
		{ 292, 212, 18 },
		{ 292, 213, 18 },
		{ 292, 214, 18 },
		{ 292, 215, 18 },
		{ 292, 216, 18 },
		{ 292, 217, 18 },
		{ 292, 218, 18 },
		{ 292, 219, 18 },
		{ 292, 220, 18 },
		{ 292, 221, 18 },
		{ 292, 222, 18 },
		{ 292, 223, 18 },
		{ 292, 224, 18 },
		{ 292, 225, 18 },
		{ 292, 226, 18 },
		{ 292, 227, 18 },
		{ 292, 228, 18 },
		{ 292, 229, 18 },
		{ 292, 230, 18 },
		{ 293, 210, 18 },
		{ 293, 211, 18 },
		{ 293, 212, 18 },
		{ 293, 213, 18 },
		{ 293, 214, 18 },
		{ 293, 215, 18 },
		{ 293, 216, 18 },
		{ 293, 217, 18 },
		{ 293, 218, 18 },
		{ 293, 219, 18 },
		{ 293, 220, 18 },
		{ 293, 221, 18 },
		{ 293, 222, 18 },
		{ 293, 223, 18 },
		{ 293, 224, 18 },
		{ 293, 225, 18 },
		{ 293, 226, 18 },
		{ 293, 227, 18 },
		{ 293, 228, 18 },
		{ 293, 229, 18 },
		{ 294, 208, 18 },
		{ 294, 209, 18 },
		{ 294, 210, 18 },
		{ 294, 211, 18 },
		{ 294, 212, 18 },
		{ 294, 213, 18 },
		{ 294, 214, 18 },
		{ 294, 215, 18 },
		{ 294, 216, 18 },
		{ 294, 217, 18 },
		{ 294, 218, 18 },
		{ 294, 219, 18 },
		{ 294, 220, 18 },
		{ 294, 221, 18 },
		{ 294, 222, 18 },
		{ 294, 223, 18 },
		{ 294, 224, 18 },
		{ 294, 225, 18 },
		{ 294, 226, 18 },
		{ 294, 227, 18 },
		{ 294, 228, 18 },
		{ 294, 229, 18 },
		{ 295, 207, 18 },
		{ 295, 208, 18 },
		{ 295, 209, 18 },
		{ 295, 210, 18 },
		{ 295, 211, 18 },
		{ 295, 212, 18 },
		{ 295, 213, 18 },
		{ 295, 214, 18 },
		{ 295, 215, 18 },
		{ 295, 216, 18 },
		{ 295, 217, 18 },
		{ 295, 218, 18 },
		{ 295, 219, 18 },
		{ 295, 220, 18 },
		{ 295, 221, 18 },
		{ 295, 222, 18 },
		{ 295, 223, 18 },
		{ 295, 224, 18 },
		{ 295, 225, 18 },
		{ 295, 226, 18 },
		{ 295, 227, 18 },
		{ 295, 228, 18 },
		{ 296, 206, 18 },
		{ 296, 207, 18 },
		{ 296, 208, 18 },
		{ 296, 209, 18 },
		{ 296, 210, 18 },
		{ 296, 211, 18 },
		{ 296, 212, 18 },
		{ 296, 213, 18 },
		{ 296, 214, 18 },
		{ 296, 215, 18 },
		{ 296, 216, 18 },
		{ 296, 217, 18 },
		{ 296, 218, 18 },
		{ 296, 219, 18 },
		{ 296, 220, 18 },
		{ 296, 221, 18 },
		{ 296, 222, 18 },
		{ 296, 223, 18 },
		{ 296, 224, 18 },
		{ 296, 225, 18 },
		{ 296, 226, 18 },
		{ 296, 227, 18 },
		{ 297, 204, 18 },
		{ 297, 205, 18 },
		{ 297, 206, 18 },
		{ 297, 207, 18 },
		{ 297, 208, 18 },
		{ 297, 209, 18 },
		{ 297, 210, 18 },
		{ 297, 211, 18 },
		{ 297, 212, 18 },
		{ 297, 213, 18 },
		{ 297, 214, 18 },
		{ 297, 215, 18 },
		{ 297, 216, 18 },
		{ 297, 217, 18 },
		{ 297, 218, 18 },
		{ 297, 219, 18 },
		{ 297, 220, 18 },
		{ 297, 221, 18 },
		{ 297, 222, 18 },
		{ 297, 223, 18 },
		{ 297, 224, 18 },
		{ 297, 225, 18 },
		{ 297, 226, 18 },
		{ 298, 203, 18 },
		{ 298, 204, 18 },
		{ 298, 205, 18 },
		{ 298, 206, 18 },
		{ 298, 207, 18 },
		{ 298, 208, 18 },
		{ 298, 209, 18 },
		{ 298, 210, 18 },
		{ 298, 211, 18 },
		{ 298, 212, 18 },
		{ 298, 213, 18 },
		{ 298, 214, 18 },
		{ 298, 215, 18 },
		{ 298, 216, 18 },
		{ 298, 217, 18 },
		{ 298, 218, 18 },
		{ 298, 219, 18 },
		{ 298, 220, 18 },
		{ 298, 221, 18 },
		{ 298, 222, 18 },
		{ 298, 223, 18 },
		{ 298, 224, 18 },
		{ 298, 225, 18 },
		{ 298, 226, 18 },
		{ 299, 202, 18 },
		{ 299, 203, 18 },
		{ 299, 204, 18 },
		{ 299, 205, 18 },
		{ 299, 206, 18 },
		{ 299, 207, 18 },
		{ 299, 208, 18 },
		{ 299, 209, 18 },
		{ 299, 210, 18 },
		{ 299, 211, 18 },
		{ 299, 212, 18 },
		{ 299, 213, 18 },
		{ 299, 214, 18 },
		{ 299, 215, 18 },
		{ 299, 216, 18 },
		{ 299, 217, 18 },
		{ 299, 218, 18 },
		{ 299, 219, 18 },
		{ 299, 220, 18 },
		{ 299, 221, 18 },
		{ 299, 222, 18 },
		{ 299, 223, 18 },
		{ 299, 224, 18 },
		{ 299, 225, 18 },
		{ 300, 200, 18 },
		{ 300, 201, 18 },
		{ 300, 202, 18 },
		{ 300, 203, 18 },
		{ 300, 204, 18 },
		{ 300, 205, 18 },
		{ 300, 206, 18 },
		{ 300, 207, 18 },
		{ 300, 208, 18 },
		{ 300, 209, 18 },
		{ 300, 210, 18 },
		{ 300, 211, 18 },
		{ 300, 212, 18 },
		{ 300, 213, 18 },
		{ 300, 214, 18 },
		{ 300, 215, 18 },
		{ 300, 216, 18 },
		{ 300, 217, 18 },
		{ 300, 218, 18 },
		{ 300, 219, 18 },
		{ 300, 220, 18 },
		{ 300, 221, 18 },
		{ 300, 222, 18 },
		{ 300, 223, 18 },
		{ 300, 224, 18 },
		{ 301, 199, 18 },
		{ 301, 200, 18 },
		{ 301, 201, 18 },
		{ 301, 202, 18 },
		{ 301, 203, 18 },
		{ 301, 204, 18 },
		{ 301, 205, 18 },
		{ 301, 206, 18 },
		{ 301, 207, 18 },
		{ 301, 208, 18 },
		{ 301, 209, 18 },
		{ 301, 210, 18 },
		{ 301, 211, 18 },
		{ 301, 212, 18 },
		{ 301, 213, 18 },
		{ 301, 214, 18 },
		{ 301, 215, 18 },
		{ 301, 216, 18 },
		{ 301, 217, 18 },
		{ 301, 218, 18 },
		{ 301, 219, 18 },
		{ 301, 220, 18 },
		{ 301, 221, 18 },
		{ 301, 222, 18 },
		{ 301, 223, 18 },
		{ 302, 198, 18 },
		{ 302, 199, 18 },
		{ 302, 200, 18 },
		{ 302, 201, 18 },
		{ 302, 202, 18 },
		{ 302, 203, 18 },
		{ 302, 204, 18 },
		{ 302, 205, 18 },
		{ 302, 206, 18 },
		{ 302, 207, 18 },
		{ 302, 208, 18 },
		{ 302, 209, 18 },
		{ 302, 210, 18 },
		{ 302, 211, 18 },
		{ 302, 212, 18 },
		{ 302, 213, 18 },
		{ 302, 214, 18 },
		{ 302, 215, 18 },
		{ 302, 216, 18 },
		{ 302, 217, 18 },
		{ 302, 218, 18 },
		{ 302, 219, 18 },
		{ 302, 220, 18 },
		{ 302, 221, 18 },
		{ 302, 222, 18 },
		{ 302, 223, 18 },
		{ 303, 196, 18 },
		{ 303, 197, 18 },
		{ 303, 198, 18 },
		{ 303, 199, 18 },
		{ 303, 200, 18 },
		{ 303, 201, 18 },
		{ 303, 202, 18 },
		{ 303, 203, 18 },
		{ 303, 204, 18 },
		{ 303, 205, 18 },
		{ 303, 206, 18 },
		{ 303, 207, 18 },
		{ 303, 208, 18 },
		{ 303, 209, 18 },
		{ 303, 210, 18 },
		{ 303, 211, 18 },
		{ 303, 212, 18 },
		{ 303, 213, 18 },
		{ 303, 214, 18 },
		{ 303, 215, 18 },
		{ 303, 216, 18 },
		{ 303, 217, 18 },
		{ 303, 218, 18 },
		{ 303, 219, 18 },
		{ 303, 220, 18 },
		{ 303, 221, 18 },
		{ 303, 222, 18 },
		{ 304, 195, 18 },
		{ 304, 196, 18 },
		{ 304, 197, 18 },
		{ 304, 198, 18 },
		{ 304, 199, 18 },
		{ 304, 200, 18 },
		{ 304, 201, 18 },
		{ 304, 202, 18 },
		{ 304, 203, 18 },
		{ 304, 204, 18 },
		{ 304, 205, 18 },
		{ 304, 206, 18 },
		{ 304, 207, 18 },
		{ 304, 208, 18 },
		{ 304, 209, 18 },
		{ 304, 210, 18 },
		{ 304, 211, 18 },
		{ 304, 212, 18 },
		{ 304, 213, 18 },
		{ 304, 214, 18 },
		{ 304, 215, 18 },
		{ 304, 216, 18 },
		{ 304, 217, 18 },
		{ 304, 218, 18 },
		{ 304, 219, 18 },
		{ 304, 220, 18 },
		{ 304, 221, 18 },
		{ 305, 194, 18 },
		{ 305, 195, 18 },
		{ 305, 196, 18 },
		{ 305, 197, 18 },
		{ 305, 198, 18 },
		{ 305, 199, 18 },
		{ 305, 200, 18 },
		{ 305, 201, 18 },
		{ 305, 202, 18 },
		{ 305, 203, 18 },
		{ 305, 204, 18 },
		{ 305, 205, 18 },
		{ 305, 206, 18 },
		{ 305, 207, 18 },
		{ 305, 208, 18 },
		{ 305, 209, 18 },
		{ 305, 210, 18 },
		{ 305, 211, 18 },
		{ 305, 212, 18 },
		{ 305, 213, 18 },
		{ 305, 214, 18 },
		{ 305, 215, 18 },
		{ 305, 216, 18 },
		{ 305, 217, 18 },
		{ 305, 218, 18 },
		{ 305, 219, 18 },
		{ 305, 220, 18 },
		{ 306, 192, 18 },
		{ 306, 193, 18 },
		{ 306, 194, 18 },
		{ 306, 195, 18 },
		{ 306, 196, 18 },
		{ 306, 197, 18 },
		{ 306, 198, 18 },
		{ 306, 199, 18 },
		{ 306, 200, 18 },
		{ 306, 201, 18 },
		{ 306, 202, 18 },
		{ 306, 203, 18 },
		{ 306, 204, 18 },
		{ 306, 205, 18 },
		{ 306, 206, 18 },
		{ 306, 207, 18 },
		{ 306, 208, 18 },
		{ 306, 209, 18 },
		{ 306, 210, 18 },
		{ 306, 211, 18 },
		{ 306, 212, 18 },
		{ 306, 213, 18 },
		{ 306, 214, 18 },
		{ 306, 215, 18 },
		{ 306, 216, 18 },
		{ 306, 217, 18 },
		{ 306, 218, 18 },
		{ 306, 219, 18 },
		{ 306, 220, 18 },
		{ 307, 191, 18 },
		{ 307, 192, 18 },
		{ 307, 193, 18 },
		{ 307, 194, 18 },
		{ 307, 195, 18 },
		{ 307, 196, 18 },
		{ 307, 197, 18 },
		{ 307, 198, 18 },
		{ 307, 199, 18 },
		{ 307, 200, 18 },
		{ 307, 201, 18 },
		{ 307, 202, 18 },
		{ 307, 203, 18 },
		{ 307, 204, 18 },
		{ 307, 205, 18 },
		{ 307, 206, 18 },
		{ 307, 207, 18 },
		{ 307, 208, 18 },
		{ 307, 209, 18 },
		{ 307, 210, 18 },
		{ 307, 211, 18 },
		{ 307, 212, 18 },
		{ 307, 213, 18 },
		{ 307, 214, 18 },
		{ 307, 215, 18 },
		{ 307, 216, 18 },
		{ 307, 217, 18 },
		{ 307, 218, 18 },
		{ 307, 219, 18 },
		{ 308, 190, 18 },
		{ 308, 191, 18 },
		{ 308, 192, 18 },
		{ 308, 193, 18 },
		{ 308, 194, 18 },
		{ 308, 195, 18 },
		{ 308, 196, 18 },
		{ 308, 197, 18 },
		{ 308, 198, 18 },
		{ 308, 199, 18 },
		{ 308, 200, 18 },
		{ 308, 201, 18 },
		{ 308, 202, 18 },
		{ 308, 203, 18 },
		{ 308, 204, 18 },
		{ 308, 205, 18 },
		{ 308, 206, 18 },
		{ 308, 207, 18 },
		{ 308, 208, 18 },
		{ 308, 209, 18 },
		{ 308, 210, 18 },
		{ 308, 211, 18 },
		{ 308, 212, 18 },
		{ 308, 213, 18 },
		{ 308, 214, 18 },
		{ 308, 215, 18 },
		{ 308, 216, 18 },
		{ 308, 217, 18 },
		{ 308, 218, 18 },
		{ 309, 188, 18 },
		{ 309, 189, 18 },
		{ 309, 190, 18 },
		{ 309, 191, 18 },
		{ 309, 192, 18 },
		{ 309, 193, 18 },
		{ 309, 194, 18 },
		{ 309, 195, 18 },
		{ 309, 196, 18 },
		{ 309, 197, 18 },
		{ 309, 198, 18 },
		{ 309, 199, 18 },
		{ 309, 200, 18 },
		{ 309, 201, 18 },
		{ 309, 202, 18 },
		{ 309, 203, 18 },
		{ 309, 204, 18 },
		{ 309, 205, 18 },
		{ 309, 206, 18 },
		{ 309, 207, 18 },
		{ 309, 208, 18 },
		{ 309, 209, 18 },
		{ 309, 210, 18 },
		{ 309, 211, 18 },
		{ 309, 212, 18 },
		{ 309, 213, 18 },
		{ 309, 214, 18 },
		{ 309, 215, 18 },
		{ 309, 216, 18 },
		{ 309, 217, 18 },
		{ 310, 187, 18 },
		{ 310, 188, 18 },
		{ 310, 189, 18 },
		{ 310, 190, 18 },
		{ 310, 191, 18 },
		{ 310, 192, 18 },
		{ 310, 193, 18 },
		{ 310, 194, 18 },
		{ 310, 195, 18 },
		{ 310, 196, 18 },
		{ 310, 197, 18 },
		{ 310, 198, 18 },
		{ 310, 199, 18 },
		{ 310, 200, 18 },
		{ 310, 201, 18 },
		{ 310, 202, 18 },
		{ 310, 203, 18 },
		{ 310, 204, 18 },
		{ 310, 205, 18 },
		{ 310, 206, 18 },
		{ 310, 207, 18 },
		{ 310, 208, 18 },
		{ 310, 209, 18 },
		{ 310, 210, 18 },
		{ 310, 211, 18 },
		{ 310, 212, 18 },
		{ 310, 213, 18 },
		{ 310, 214, 18 },
		{ 310, 215, 18 },
		{ 310, 216, 18 },
		{ 310, 217, 18 },
		{ 311, 186, 18 },
		{ 311, 187, 18 },
		{ 311, 188, 18 },
		{ 311, 189, 18 },
		{ 311, 190, 18 },
		{ 311, 191, 18 },
		{ 311, 192, 18 },
		{ 311, 193, 18 },
		{ 311, 194, 18 },
		{ 311, 195, 18 },
		{ 311, 196, 18 },
		{ 311, 197, 18 },
		{ 311, 198, 18 },
		{ 311, 199, 18 },
		{ 311, 200, 18 },
		{ 311, 201, 18 },
		{ 311, 202, 18 },
		{ 311, 203, 18 },
		{ 311, 204, 18 },
		{ 311, 205, 18 },
		{ 311, 206, 18 },
		{ 311, 207, 18 },
		{ 311, 208, 18 },
		{ 311, 209, 18 },
		{ 311, 210, 18 },
		{ 311, 211, 18 },
		{ 311, 212, 18 },
		{ 311, 213, 18 },
		{ 311, 214, 18 },
		{ 311, 215, 18 },
		{ 311, 216, 18 },
		{ 312, 184, 18 },
		{ 312, 185, 18 },
		{ 312, 186, 18 },
		{ 312, 187, 18 },
		{ 312, 188, 18 },
		{ 312, 189, 18 },
		{ 312, 190, 18 },
		{ 312, 191, 18 },
		{ 312, 192, 18 },
		{ 312, 193, 18 },
		{ 312, 194, 18 },
		{ 312, 195, 18 },
		{ 312, 196, 18 },
		{ 312, 197, 18 },
		{ 312, 198, 18 },
		{ 312, 199, 18 },
		{ 312, 200, 18 },
		{ 312, 201, 18 },
		{ 312, 202, 18 },
		{ 312, 203, 18 },
		{ 312, 204, 18 },
		{ 312, 205, 18 },
		{ 312, 206, 18 },
		{ 312, 207, 18 },
		{ 312, 208, 18 },
		{ 312, 209, 18 },
		{ 312, 210, 18 },
		{ 312, 211, 18 },
		{ 312, 212, 18 },
		{ 312, 213, 18 },
		{ 312, 214, 18 },
		{ 312, 215, 18 },
		{ 313, 183, 18 },
		{ 313, 184, 18 },
		{ 313, 185, 18 },
		{ 313, 186, 18 },
		{ 313, 187, 18 },
		{ 313, 188, 18 },
		{ 313, 189, 18 },
		{ 313, 190, 18 },
		{ 313, 191, 18 },
		{ 313, 192, 18 },
		{ 313, 193, 18 },
		{ 313, 194, 18 },
		{ 313, 195, 18 },
		{ 313, 196, 18 },
		{ 313, 197, 18 },
		{ 313, 198, 18 },
		{ 313, 199, 18 },
		{ 313, 200, 18 },
		{ 313, 201, 18 },
		{ 313, 202, 18 },
		{ 313, 203, 18 },
		{ 313, 204, 18 },
		{ 313, 205, 18 },
		{ 313, 206, 18 },
		{ 313, 207, 18 },
		{ 313, 208, 18 },
		{ 313, 209, 18 },
		{ 313, 210, 18 },
		{ 313, 211, 18 },
		{ 313, 212, 18 },
		{ 313, 213, 18 },
		{ 313, 214, 18 },
		{ 314, 182, 18 },
		{ 314, 183, 18 },
		{ 314, 184, 18 },
		{ 314, 185, 18 },
		{ 314, 186, 18 },
		{ 314, 187, 18 },
		{ 314, 188, 18 },
		{ 314, 189, 18 },
		{ 314, 190, 18 },
		{ 314, 191, 18 },
		{ 314, 192, 18 },
		{ 314, 193, 18 },
		{ 314, 194, 18 },
		{ 314, 195, 18 },
		{ 314, 196, 18 },
		{ 314, 197, 18 },
		{ 314, 198, 18 },
		{ 314, 199, 18 },
		{ 314, 200, 18 },
		{ 314, 201, 18 },
		{ 314, 202, 18 },
		{ 314, 203, 18 },
		{ 314, 204, 18 },
		{ 314, 205, 18 },
		{ 314, 206, 18 },
		{ 314, 207, 18 },
		{ 314, 208, 18 },
		{ 314, 209, 18 },
		{ 314, 210, 18 },
		{ 314, 211, 18 },
		{ 314, 212, 18 },
		{ 314, 213, 18 },
		{ 314, 214, 18 },
		{ 315, 180, 18 },
		{ 315, 181, 18 },
		{ 315, 182, 18 },
		{ 315, 183, 18 },
		{ 315, 184, 18 },
		{ 315, 185, 18 },
		{ 315, 186, 18 },
		{ 315, 187, 18 },
		{ 315, 188, 18 },
		{ 315, 189, 18 },
		{ 315, 190, 18 },
		{ 315, 191, 18 },
		{ 315, 192, 18 },
		{ 315, 193, 18 },
		{ 315, 194, 18 },
		{ 315, 195, 18 },
		{ 315, 196, 18 },
		{ 315, 197, 18 },
		{ 315, 198, 18 },
		{ 315, 199, 18 },
		{ 315, 200, 18 },
		{ 315, 201, 18 },
		{ 315, 202, 18 },
		{ 315, 203, 18 },
		{ 315, 204, 18 },
		{ 315, 205, 18 },
		{ 315, 206, 18 },
		{ 315, 207, 18 },
		{ 315, 208, 18 },
		{ 315, 209, 18 },
		{ 315, 210, 18 },
		{ 315, 211, 18 },
		{ 315, 212, 18 },
		{ 315, 213, 18 },
		{ 316, 179, 18 },
		{ 316, 180, 18 },
		{ 316, 181, 18 },
		{ 316, 182, 18 },
		{ 316, 183, 18 },
		{ 316, 184, 18 },
		{ 316, 185, 18 },
		{ 316, 186, 18 },
		{ 316, 187, 18 },
		{ 316, 188, 18 },
		{ 316, 189, 18 },
		{ 316, 190, 18 },
		{ 316, 191, 18 },
		{ 316, 192, 18 },
		{ 316, 193, 18 },
		{ 316, 194, 18 },
		{ 316, 195, 18 },
		{ 316, 196, 18 },
		{ 316, 197, 18 },
		{ 316, 198, 18 },
		{ 316, 199, 18 },
		{ 316, 200, 18 },
		{ 316, 201, 18 },
		{ 316, 202, 18 },
		{ 316, 203, 18 },
		{ 316, 204, 18 },
		{ 316, 205, 18 },
		{ 316, 206, 18 },
		{ 316, 207, 18 },
		{ 316, 208, 18 },
		{ 316, 209, 18 },
		{ 316, 210, 18 },
		{ 316, 211, 18 },
		{ 316, 212, 18 },
		{ 317, 178, 18 },
		{ 317, 179, 18 },
		{ 317, 180, 18 },
		{ 317, 181, 18 },
		{ 317, 182, 18 },
		{ 317, 183, 18 },
		{ 317, 184, 18 },
		{ 317, 185, 18 },
		{ 317, 186, 18 },
		{ 317, 187, 18 },
		{ 317, 188, 18 },
		{ 317, 189, 18 },
		{ 317, 190, 18 },
		{ 317, 191, 18 },
		{ 317, 192, 18 },
		{ 317, 193, 18 },
		{ 317, 194, 18 },
		{ 317, 195, 18 },
		{ 317, 196, 18 },
		{ 317, 197, 18 },
		{ 317, 198, 18 },
		{ 317, 199, 18 },
		{ 317, 200, 18 },
		{ 317, 201, 18 },
		{ 317, 202, 18 },
		{ 317, 203, 18 },
		{ 317, 204, 18 },
		{ 317, 205, 18 },
		{ 317, 206, 18 },
		{ 317, 207, 18 },
		{ 317, 208, 18 },
		{ 317, 209, 18 },
		{ 317, 210, 18 },
		{ 317, 211, 18 },
		{ 318, 176, 18 },
		{ 318, 177, 18 },
		{ 318, 178, 18 },
		{ 318, 179, 18 },
		{ 318, 180, 18 },
		{ 318, 181, 18 },
		{ 318, 182, 18 },
		{ 318, 183, 18 },
		{ 318, 184, 18 },
		{ 318, 185, 18 },
		{ 318, 186, 18 },
		{ 318, 187, 18 },
		{ 318, 188, 18 },
		{ 318, 189, 18 },
		{ 318, 190, 18 },
		{ 318, 191, 18 },
		{ 318, 192, 18 },
		{ 318, 193, 18 },
		{ 318, 194, 18 },
		{ 318, 195, 18 },
		{ 318, 196, 18 },
		{ 318, 197, 18 },
		{ 318, 198, 18 },
		{ 318, 199, 18 },
		{ 318, 200, 18 },
		{ 318, 201, 18 },
		{ 318, 202, 18 },
		{ 318, 203, 18 },
		{ 318, 204, 18 },
		{ 318, 205, 18 },
		{ 318, 206, 18 },
		{ 318, 207, 18 },
		{ 318, 208, 18 },
		{ 318, 209, 18 },
		{ 318, 210, 18 },
		{ 318, 211, 18 },
		{ 319, 175, 18 },
		{ 319, 176, 18 },
		{ 319, 177, 18 },
		{ 319, 178, 18 },
		{ 319, 179, 18 },
		{ 319, 180, 18 },
		{ 319, 181, 18 },
		{ 319, 182, 18 },
		{ 319, 183, 18 },
		{ 319, 184, 18 },
		{ 319, 185, 18 },
		{ 319, 186, 18 },
		{ 319, 187, 18 },
		{ 319, 188, 18 },
		{ 319, 189, 18 },
		{ 319, 190, 18 },
		{ 319, 191, 18 },
		{ 319, 192, 18 },
		{ 319, 193, 18 },
		{ 319, 194, 18 },
		{ 319, 195, 18 },
		{ 319, 196, 18 },
		{ 319, 197, 18 },
		{ 319, 198, 18 },
		{ 319, 199, 18 },
		{ 319, 200, 18 },
		{ 319, 201, 18 },
		{ 319, 202, 18 },
		{ 319, 203, 18 },
		{ 319, 204, 18 },
		{ 319, 205, 18 },
		{ 319, 206, 18 },
		{ 319, 207, 18 },
		{ 319, 208, 18 },
		{ 319, 209, 18 },
		{ 319, 210, 18 },
		{ 320, 174, 18 },
		{ 320, 175, 18 },
		{ 320, 176, 18 },
		{ 320, 177, 18 },
		{ 320, 178, 18 },
		{ 320, 179, 18 },
		{ 320, 180, 18 },
		{ 320, 181, 18 },
		{ 320, 182, 18 },
		{ 320, 183, 18 },
		{ 320, 184, 18 },
		{ 320, 185, 18 },
		{ 320, 186, 18 },
		{ 320, 187, 18 },
		{ 320, 188, 18 },
		{ 320, 189, 18 },
		{ 320, 190, 18 },
		{ 320, 191, 18 },
		{ 320, 192, 18 },
		{ 320, 193, 18 },
		{ 320, 194, 18 },
		{ 320, 195, 18 },
		{ 320, 196, 18 },
		{ 320, 197, 18 },
		{ 320, 198, 18 },
		{ 320, 199, 18 },
		{ 320, 200, 18 },
		{ 320, 201, 18 },
		{ 320, 202, 18 },
		{ 320, 203, 18 },
		{ 320, 204, 18 },
		{ 320, 205, 18 },
		{ 320, 206, 18 },
		{ 320, 207, 18 },
		{ 320, 208, 18 },
		{ 320, 209, 18 },
		{ 321, 172, 18 },
		{ 321, 173, 18 },
		{ 321, 174, 18 },
		{ 321, 175, 18 },
		{ 321, 176, 18 },
		{ 321, 177, 18 },
		{ 321, 178, 18 },
		{ 321, 179, 18 },
		{ 321, 180, 18 },
		{ 321, 181, 18 },
		{ 321, 182, 18 },
		{ 321, 183, 18 },
		{ 321, 184, 18 },
		{ 321, 185, 18 },
		{ 321, 186, 18 },
		{ 321, 187, 18 },
		{ 321, 188, 18 },
		{ 321, 189, 18 },
		{ 321, 190, 18 },
		{ 321, 191, 18 },
		{ 321, 192, 18 },
		{ 321, 193, 18 },
		{ 321, 194, 18 },
		{ 321, 195, 18 },
		{ 321, 196, 18 },
		{ 321, 197, 18 },
		{ 321, 198, 18 },
		{ 321, 199, 18 },
		{ 321, 200, 18 },
		{ 321, 201, 18 },
		{ 321, 202, 18 },
		{ 321, 203, 18 },
		{ 321, 204, 18 },
		{ 321, 205, 18 },
		{ 321, 206, 18 },
		{ 321, 207, 18 },
		{ 321, 208, 18 },
		{ 322, 171, 18 },
		{ 322, 172, 18 },
		{ 322, 173, 18 },
		{ 322, 174, 18 },
		{ 322, 175, 18 },
		{ 322, 176, 18 },
		{ 322, 177, 18 },
		{ 322, 178, 18 },
		{ 322, 179, 18 },
		{ 322, 180, 18 },
		{ 322, 181, 18 },
		{ 322, 182, 18 },
		{ 322, 183, 18 },
		{ 322, 184, 18 },
		{ 322, 185, 18 },
		{ 322, 186, 18 },
		{ 322, 187, 18 },
		{ 322, 188, 18 },
		{ 322, 189, 18 },
		{ 322, 190, 18 },
		{ 322, 191, 18 },
		{ 322, 192, 18 },
		{ 322, 193, 18 },
		{ 322, 194, 18 },
		{ 322, 195, 18 },
		{ 322, 196, 18 },
		{ 322, 197, 18 },
		{ 322, 198, 18 },
		{ 322, 199, 18 },
		{ 322, 200, 18 },
		{ 322, 201, 18 },
		{ 322, 202, 18 },
		{ 322, 203, 18 },
		{ 322, 204, 18 },
		{ 322, 205, 18 },
		{ 322, 206, 18 },
		{ 322, 207, 18 },
		{ 322, 208, 18 },
		{ 323, 170, 18 },
		{ 323, 171, 18 },
		{ 323, 172, 18 },
		{ 323, 173, 18 },
		{ 323, 174, 18 },
		{ 323, 175, 18 },
		{ 323, 176, 18 },
		{ 323, 177, 18 },
		{ 323, 178, 18 },
		{ 323, 179, 18 },
		{ 323, 180, 18 },
		{ 323, 181, 18 },
		{ 323, 182, 18 },
		{ 323, 183, 18 },
		{ 323, 184, 18 },
		{ 323, 185, 18 },
		{ 323, 186, 18 },
		{ 323, 187, 18 },
		{ 323, 188, 18 },
		{ 323, 189, 18 },
		{ 323, 190, 18 },
		{ 323, 191, 18 },
		{ 323, 192, 18 },
		{ 323, 193, 18 },
		{ 323, 194, 18 },
		{ 323, 195, 18 },
		{ 323, 196, 18 },
		{ 323, 197, 18 },
		{ 323, 198, 18 },
		{ 323, 199, 18 },
		{ 323, 200, 18 },
		{ 323, 201, 18 },
		{ 323, 202, 18 },
		{ 323, 203, 18 },
		{ 323, 204, 18 },
		{ 323, 205, 18 },
		{ 323, 206, 18 },
		{ 323, 207, 18 },
		{ 324, 168, 18 },
		{ 324, 169, 18 },
		{ 324, 170, 18 },
		{ 324, 171, 18 },
		{ 324, 172, 18 },
		{ 324, 173, 18 },
		{ 324, 174, 18 },
		{ 324, 175, 18 },
		{ 324, 176, 18 },
		{ 324, 177, 18 },
		{ 324, 178, 18 },
		{ 324, 179, 18 },
		{ 324, 180, 18 },
		{ 324, 181, 18 },
		{ 324, 182, 18 },
		{ 324, 183, 18 },
		{ 324, 184, 18 },
		{ 324, 185, 18 },
		{ 324, 186, 18 },
		{ 324, 187, 18 },
		{ 324, 188, 18 },
		{ 324, 189, 18 },
		{ 324, 190, 18 },
		{ 324, 191, 18 },
		{ 324, 192, 18 },
		{ 324, 193, 18 },
		{ 324, 194, 18 },
		{ 324, 195, 18 },
		{ 324, 196, 18 },
		{ 324, 197, 18 },
		{ 324, 198, 18 },
		{ 324, 199, 18 },
		{ 324, 200, 18 },
		{ 324, 201, 18 },
		{ 324, 202, 18 },
		{ 324, 203, 18 },
		{ 324, 204, 18 },
		{ 324, 205, 18 },
		{ 324, 206, 18 },
		{ 325, 167, 18 },
		{ 325, 168, 18 },
		{ 325, 169, 18 },
		{ 325, 170, 18 },
		{ 325, 171, 18 },
		{ 325, 172, 18 },
		{ 325, 173, 18 },
		{ 325, 174, 18 },
		{ 325, 175, 18 },
		{ 325, 176, 18 },
		{ 325, 177, 18 },
		{ 325, 178, 18 },
		{ 325, 179, 18 },
		{ 325, 180, 18 },
		{ 325, 181, 18 },
		{ 325, 182, 18 },
		{ 325, 183, 18 },
		{ 325, 184, 18 },
		{ 325, 185, 18 },
		{ 325, 186, 18 },
		{ 325, 187, 18 },
		{ 325, 188, 18 },
		{ 325, 189, 18 },
		{ 325, 190, 18 },
		{ 325, 191, 18 },
		{ 325, 192, 18 },
		{ 325, 193, 18 },
		{ 325, 194, 18 },
		{ 325, 195, 18 },
		{ 325, 196, 18 },
		{ 325, 197, 18 },
		{ 325, 198, 18 },
		{ 325, 199, 18 },
		{ 325, 200, 18 },
		{ 325, 201, 18 },
		{ 325, 202, 18 },
		{ 325, 203, 18 },
		{ 325, 204, 18 },
		{ 325, 205, 18 },
		{ 326, 166, 18 },
		{ 326, 167, 18 },
		{ 326, 168, 18 },
		{ 326, 169, 18 },
		{ 326, 170, 18 },
		{ 326, 171, 18 },
		{ 326, 172, 18 },
		{ 326, 173, 18 },
		{ 326, 174, 18 },
		{ 326, 175, 18 },
		{ 326, 176, 18 },
		{ 326, 177, 18 },
		{ 326, 178, 18 },
		{ 326, 179, 18 },
		{ 326, 180, 18 },
		{ 326, 181, 18 },
		{ 326, 182, 18 },
		{ 326, 183, 18 },
		{ 326, 184, 18 },
		{ 326, 185, 18 },
		{ 326, 186, 18 },
		{ 326, 187, 18 },
		{ 326, 188, 18 },
		{ 326, 189, 18 },
		{ 326, 190, 18 },
		{ 326, 191, 18 },
		{ 326, 192, 18 },
		{ 326, 193, 18 },
		{ 326, 194, 18 },
		{ 326, 195, 18 },
		{ 326, 196, 18 },
		{ 326, 197, 18 },
		{ 326, 198, 18 },
		{ 326, 199, 18 },
		{ 326, 200, 18 },
		{ 326, 201, 18 },
		{ 326, 202, 18 },
		{ 326, 203, 18 },
		{ 326, 204, 18 },
		{ 326, 205, 18 },
		{ 327, 164, 18 },
		{ 327, 165, 18 },
		{ 327, 166, 18 },
		{ 327, 167, 18 },
		{ 327, 168, 18 },
		{ 327, 169, 18 },
		{ 327, 170, 18 },
		{ 327, 171, 18 },
		{ 327, 172, 18 },
		{ 327, 173, 18 },
		{ 327, 174, 18 },
		{ 327, 175, 18 },
		{ 327, 176, 18 },
		{ 327, 177, 18 },
		{ 327, 178, 18 },
		{ 327, 179, 18 },
		{ 327, 180, 18 },
		{ 327, 181, 18 },
		{ 327, 182, 18 },
		{ 327, 183, 18 },
		{ 327, 184, 18 },
		{ 327, 185, 18 },
		{ 327, 186, 18 },
		{ 327, 187, 18 },
		{ 327, 188, 18 },
		{ 327, 189, 18 },
		{ 327, 190, 18 },
		{ 327, 191, 18 },
		{ 327, 192, 18 },
		{ 327, 193, 18 },
		{ 327, 194, 18 },
		{ 327, 195, 18 },
		{ 327, 196, 18 },
		{ 327, 197, 18 },
		{ 327, 198, 18 },
		{ 327, 199, 18 },
		{ 327, 200, 18 },
		{ 327, 201, 18 },
		{ 327, 202, 18 },
		{ 327, 203, 18 },
		{ 327, 204, 18 },
		{ 328, 163, 18 },
		{ 328, 164, 18 },
		{ 328, 165, 18 },
		{ 328, 166, 18 },
		{ 328, 167, 18 },
		{ 328, 168, 18 },
		{ 328, 169, 18 },
		{ 328, 170, 18 },
		{ 328, 171, 18 },
		{ 328, 172, 18 },
		{ 328, 173, 18 },
		{ 328, 174, 18 },
		{ 328, 175, 18 },
		{ 328, 176, 18 },
		{ 328, 177, 18 },
		{ 328, 178, 18 },
		{ 328, 179, 18 },
		{ 328, 180, 18 },
		{ 328, 181, 18 },
		{ 328, 182, 18 },
		{ 328, 183, 18 },
		{ 328, 184, 18 },
		{ 328, 185, 18 },
		{ 328, 186, 18 },
		{ 328, 187, 18 },
		{ 328, 188, 18 },
		{ 328, 189, 18 },
		{ 328, 190, 18 },
		{ 328, 191, 18 },
		{ 328, 192, 18 },
		{ 328, 193, 18 },
		{ 328, 194, 18 },
		{ 328, 195, 18 },
		{ 328, 196, 18 },
		{ 328, 197, 18 },
		{ 328, 198, 18 },
		{ 328, 199, 18 },
		{ 328, 200, 18 },
		{ 328, 201, 18 },
		{ 328, 202, 18 },
		{ 328, 203, 18 },
		{ 329, 162, 18 },
		{ 329, 163, 18 },
		{ 329, 164, 18 },
		{ 329, 165, 18 },
		{ 329, 166, 18 },
		{ 329, 167, 18 },
		{ 329, 168, 18 },
		{ 329, 169, 18 },
		{ 329, 170, 18 },
		{ 329, 171, 18 },
		{ 329, 172, 18 },
		{ 329, 173, 18 },
		{ 329, 174, 18 },
		{ 329, 175, 18 },
		{ 329, 176, 18 },
		{ 329, 177, 18 },
		{ 329, 178, 18 },
		{ 329, 179, 18 },
		{ 329, 180, 18 },
		{ 329, 181, 18 },
		{ 329, 182, 18 },
		{ 329, 183, 18 },
		{ 329, 184, 18 },
		{ 329, 185, 18 },
		{ 329, 186, 18 },
		{ 329, 187, 18 },
		{ 329, 188, 18 },
		{ 329, 189, 18 },
		{ 329, 190, 18 },
		{ 329, 191, 18 },
		{ 329, 192, 18 },
		{ 329, 193, 18 },
		{ 329, 194, 18 },
		{ 329, 195, 18 },
		{ 329, 196, 18 },
		{ 329, 197, 18 },
		{ 329, 198, 18 },
		{ 329, 199, 18 },
		{ 329, 200, 18 },
		{ 329, 201, 18 },
		{ 329, 202, 18 },
		{ 330, 160, 18 },
		{ 330, 161, 18 },
		{ 330, 162, 18 },
		{ 330, 163, 18 },
		{ 330, 164, 18 },
		{ 330, 165, 18 },
		{ 330, 166, 18 },
		{ 330, 167, 18 },
		{ 330, 168, 18 },
		{ 330, 169, 18 },
		{ 330, 170, 18 },
		{ 330, 171, 18 },
		{ 330, 172, 18 },
		{ 330, 173, 18 },
		{ 330, 174, 18 },
		{ 330, 175, 18 },
		{ 330, 176, 18 },
		{ 330, 177, 18 },
		{ 330, 178, 18 },
		{ 330, 179, 18 },
		{ 330, 180, 18 },
		{ 330, 181, 18 },
		{ 330, 182, 18 },
		{ 330, 183, 18 },
		{ 330, 184, 18 },
		{ 330, 185, 18 },
		{ 330, 186, 18 },
		{ 330, 187, 18 },
		{ 330, 188, 18 },
		{ 330, 189, 18 },
		{ 330, 190, 18 },
		{ 330, 191, 18 },
		{ 330, 192, 18 },
		{ 330, 193, 18 },
		{ 330, 194, 18 },
		{ 330, 195, 18 },
		{ 330, 196, 18 },
		{ 330, 197, 18 },
		{ 330, 198, 18 },
		{ 330, 199, 18 },
		{ 330, 200, 18 },
		{ 330, 201, 18 },
		{ 330, 202, 18 },
		{ 331, 159, 18 },
		{ 331, 160, 18 },
		{ 331, 161, 18 },
		{ 331, 162, 18 },
		{ 331, 163, 18 },
		{ 331, 164, 18 },
		{ 331, 165, 18 },
		{ 331, 166, 18 },
		{ 331, 167, 18 },
		{ 331, 168, 18 },
		{ 331, 169, 18 },
		{ 331, 170, 18 },
		{ 331, 171, 18 },
		{ 331, 172, 18 },
		{ 331, 173, 18 },
		{ 331, 174, 18 },
		{ 331, 175, 18 },
		{ 331, 176, 18 },
		{ 331, 177, 18 },
		{ 331, 178, 18 },
		{ 331, 179, 18 },
		{ 331, 180, 18 },
		{ 331, 181, 18 },
		{ 331, 182, 18 },
		{ 331, 183, 18 },
		{ 331, 184, 18 },
		{ 331, 185, 18 },
		{ 331, 186, 18 },
		{ 331, 187, 18 },
		{ 331, 188, 18 },
		{ 331, 189, 18 },
		{ 331, 190, 18 },
		{ 331, 191, 18 },
		{ 331, 192, 18 },
		{ 331, 193, 18 },
		{ 331, 194, 18 },
		{ 331, 195, 18 },
		{ 331, 196, 18 },
		{ 331, 197, 18 },
		{ 331, 198, 18 },
		{ 331, 199, 18 },
		{ 331, 200, 18 },
		{ 331, 201, 18 },
		{ 332, 158, 18 },
		{ 332, 159, 18 },
		{ 332, 160, 18 },
		{ 332, 161, 18 },
		{ 332, 162, 18 },
		{ 332, 163, 18 },
		{ 332, 164, 18 },
		{ 332, 165, 18 },
		{ 332, 166, 18 },
		{ 332, 167, 18 },
		{ 332, 168, 18 },
		{ 332, 169, 18 },
		{ 332, 170, 18 },
		{ 332, 171, 18 },
		{ 332, 172, 18 },
		{ 332, 173, 18 },
		{ 332, 174, 18 },
		{ 332, 175, 18 },
		{ 332, 176, 18 },
		{ 332, 177, 18 },
		{ 332, 178, 18 },
		{ 332, 179, 18 },
		{ 332, 180, 18 },
		{ 332, 181, 18 },
		{ 332, 182, 18 },
		{ 332, 183, 18 },
		{ 332, 184, 18 },
		{ 332, 185, 18 },
		{ 332, 186, 18 },
		{ 332, 187, 18 },
		{ 332, 188, 18 },
		{ 332, 189, 18 },
		{ 332, 190, 18 },
		{ 332, 191, 18 },
		{ 332, 192, 18 },
		{ 332, 193, 18 },
		{ 332, 194, 18 },
		{ 332, 195, 18 },
		{ 332, 196, 18 },
		{ 332, 197, 18 },
		{ 332, 198, 18 },
		{ 332, 199, 18 },
		{ 332, 200, 18 },
		{ 333, 156, 18 },
		{ 333, 157, 18 },
		{ 333, 158, 18 },
		{ 333, 159, 18 },
		{ 333, 160, 18 },
		{ 333, 161, 18 },
		{ 333, 162, 18 },
		{ 333, 163, 18 },
		{ 333, 164, 18 },
		{ 333, 165, 18 },
		{ 333, 166, 18 },
		{ 333, 167, 18 },
		{ 333, 168, 18 },
		{ 333, 169, 18 },
		{ 333, 170, 18 },
		{ 333, 171, 18 },
		{ 333, 172, 18 },
		{ 333, 173, 18 },
		{ 333, 174, 18 },
		{ 333, 175, 18 },
		{ 333, 176, 18 },
		{ 333, 177, 18 },
		{ 333, 178, 18 },
		{ 333, 179, 18 },
		{ 333, 180, 18 },
		{ 333, 181, 18 },
		{ 333, 182, 18 },
		{ 333, 183, 18 },
		{ 333, 184, 18 },
		{ 333, 185, 18 },
		{ 333, 186, 18 },
		{ 333, 187, 18 },
		{ 333, 188, 18 },
		{ 333, 189, 18 },
		{ 333, 190, 18 },
		{ 333, 191, 18 },
		{ 333, 192, 18 },
		{ 333, 193, 18 },
		{ 333, 194, 18 },
		{ 333, 195, 18 },
		{ 333, 196, 18 },
		{ 333, 197, 18 },
		{ 333, 198, 18 },
		{ 333, 199, 18 },
		{ 334, 155, 18 },
		{ 334, 156, 18 },
		{ 334, 157, 18 },
		{ 334, 158, 18 },
		{ 334, 159, 18 },
		{ 334, 160, 18 },
		{ 334, 161, 18 },
		{ 334, 162, 18 },
		{ 334, 163, 18 },
		{ 334, 164, 18 },
		{ 334, 165, 18 },
		{ 334, 166, 18 },
		{ 334, 167, 18 },
		{ 334, 168, 18 },
		{ 334, 169, 18 },
		{ 334, 170, 18 },
		{ 334, 171, 18 },
		{ 334, 172, 18 },
		{ 334, 173, 18 },
		{ 334, 174, 18 },
		{ 334, 175, 18 },
		{ 334, 176, 18 },
		{ 334, 177, 18 },
		{ 334, 178, 18 },
		{ 334, 179, 18 },
		{ 334, 180, 18 },
		{ 334, 181, 18 },
		{ 334, 182, 18 },
		{ 334, 183, 18 },
		{ 334, 184, 18 },
		{ 334, 185, 18 },
		{ 334, 186, 18 },
		{ 334, 187, 18 },
		{ 334, 188, 18 },
		{ 334, 189, 18 },
		{ 334, 190, 18 },
		{ 334, 191, 18 },
		{ 334, 192, 18 },
		{ 334, 193, 18 },
		{ 334, 194, 18 },
		{ 334, 195, 18 },
		{ 334, 196, 18 },
		{ 334, 197, 18 },
		{ 334, 198, 18 },
		{ 334, 199, 18 },
		{ 335, 154, 18 },
		{ 335, 155, 18 },
		{ 335, 156, 18 },
		{ 335, 157, 18 },
		{ 335, 158, 18 },
		{ 335, 159, 18 },
		{ 335, 160, 18 },
		{ 335, 161, 18 },
		{ 335, 162, 18 },
		{ 335, 163, 18 },
		{ 335, 164, 18 },
		{ 335, 165, 18 },
		{ 335, 166, 18 },
		{ 335, 167, 18 },
		{ 335, 168, 18 },
		{ 335, 169, 18 },
		{ 335, 170, 18 },
		{ 335, 171, 18 },
		{ 335, 172, 18 },
		{ 335, 173, 18 },
		{ 335, 174, 18 },
		{ 335, 175, 18 },
		{ 335, 176, 18 },
		{ 335, 177, 18 },
		{ 335, 178, 18 },
		{ 335, 179, 18 },
		{ 335, 180, 18 },
		{ 335, 181, 18 },
		{ 335, 182, 18 },
		{ 335, 183, 18 },
		{ 335, 184, 18 },
		{ 335, 185, 18 },
		{ 335, 186, 18 },
		{ 335, 187, 18 },
		{ 335, 188, 18 },
		{ 335, 189, 18 },
		{ 335, 190, 18 },
		{ 335, 191, 18 },
		{ 335, 192, 18 },
		{ 335, 193, 18 },
		{ 335, 194, 18 },
		{ 335, 195, 18 },
		{ 335, 196, 18 },
		{ 335, 197, 18 },
		{ 335, 198, 18 },
		{ 336, 152, 18 },
		{ 336, 153, 18 },
		{ 336, 154, 18 },
		{ 336, 155, 18 },
		{ 336, 156, 18 },
		{ 336, 157, 18 },
		{ 336, 158, 18 },
		{ 336, 159, 18 },
		{ 336, 160, 18 },
		{ 336, 161, 18 },
		{ 336, 162, 18 },
		{ 336, 163, 18 },
		{ 336, 164, 18 },
		{ 336, 165, 18 },
		{ 336, 166, 18 },
		{ 336, 167, 18 },
		{ 336, 168, 18 },
		{ 336, 169, 18 },
		{ 336, 170, 18 },
		{ 336, 171, 18 },
		{ 336, 172, 18 },
		{ 336, 173, 18 },
		{ 336, 174, 18 },
		{ 336, 175, 18 },
		{ 336, 176, 18 },
		{ 336, 177, 18 },
		{ 336, 178, 18 },
		{ 336, 179, 18 },
		{ 336, 180, 18 },
		{ 336, 181, 18 },
		{ 336, 182, 18 },
		{ 336, 183, 18 },
		{ 336, 184, 18 },
		{ 336, 185, 18 },
		{ 336, 186, 18 },
		{ 336, 187, 18 },
		{ 336, 188, 18 },
		{ 336, 189, 18 },
		{ 336, 190, 18 },
		{ 336, 191, 18 },
		{ 336, 192, 18 },
		{ 336, 193, 18 },
		{ 336, 194, 18 },
		{ 336, 195, 18 },
		{ 336, 196, 18 },
		{ 336, 197, 18 },
		{ 337, 151, 18 },
		{ 337, 152, 18 },
		{ 337, 153, 18 },
		{ 337, 154, 18 },
		{ 337, 155, 18 },
		{ 337, 156, 18 },
		{ 337, 157, 18 },
		{ 337, 158, 18 },
		{ 337, 159, 18 },
		{ 337, 160, 18 },
		{ 337, 161, 18 },
		{ 337, 162, 18 },
		{ 337, 163, 18 },
		{ 337, 164, 18 },
		{ 337, 165, 18 },
		{ 337, 166, 18 },
		{ 337, 167, 18 },
		{ 337, 168, 18 },
		{ 337, 169, 18 },
		{ 337, 170, 18 },
		{ 337, 171, 18 },
		{ 337, 172, 18 },
		{ 337, 173, 18 },
		{ 337, 174, 18 },
		{ 337, 175, 18 },
		{ 337, 176, 18 },
		{ 337, 177, 18 },
		{ 337, 178, 18 },
		{ 337, 179, 18 },
		{ 337, 180, 18 },
		{ 337, 181, 18 },
		{ 337, 182, 18 },
		{ 337, 183, 18 },
		{ 337, 184, 18 },
		{ 337, 185, 18 },
		{ 337, 186, 18 },
		{ 337, 187, 18 },
		{ 337, 188, 18 },
		{ 337, 189, 18 },
		{ 337, 190, 18 },
		{ 337, 191, 18 },
		{ 337, 192, 18 },
		{ 337, 193, 18 },
		{ 337, 194, 18 },
		{ 337, 195, 18 },
		{ 337, 196, 18 },
		{ 338, 150, 18 },
		{ 338, 151, 18 },
		{ 338, 152, 18 },
		{ 338, 153, 18 },
		{ 338, 154, 18 },
		{ 338, 155, 18 },
		{ 338, 156, 18 },
		{ 338, 157, 18 },
		{ 338, 158, 18 },
		{ 338, 159, 18 },
		{ 338, 160, 18 },
		{ 338, 161, 18 },
		{ 338, 162, 18 },
		{ 338, 163, 18 },
		{ 338, 164, 18 },
		{ 338, 165, 18 },
		{ 338, 166, 18 },
		{ 338, 167, 18 },
		{ 338, 168, 18 },
		{ 338, 169, 18 },
		{ 338, 170, 18 },
		{ 338, 171, 18 },
		{ 338, 172, 18 },
		{ 338, 173, 18 },
		{ 338, 174, 18 },
		{ 338, 175, 18 },
		{ 338, 176, 18 },
		{ 338, 177, 18 },
		{ 338, 178, 18 },
		{ 338, 179, 18 },
		{ 338, 180, 18 },
		{ 338, 181, 18 },
		{ 338, 182, 18 },
		{ 338, 183, 18 },
		{ 338, 184, 18 },
		{ 338, 185, 18 },
		{ 338, 186, 18 },
		{ 338, 187, 18 },
		{ 338, 188, 18 },
		{ 338, 189, 18 },
		{ 338, 190, 18 },
		{ 338, 191, 18 },
		{ 338, 192, 18 },
		{ 338, 193, 18 },
		{ 338, 194, 18 },
		{ 338, 195, 18 },
		{ 338, 196, 18 },
		{ 339, 148, 18 },
		{ 339, 149, 18 },
		{ 339, 150, 18 },
		{ 339, 151, 18 },
		{ 339, 152, 18 },
		{ 339, 153, 18 },
		{ 339, 154, 18 },
		{ 339, 155, 18 },
		{ 339, 156, 18 },
		{ 339, 157, 18 },
		{ 339, 158, 18 },
		{ 339, 159, 18 },
		{ 339, 160, 18 },
		{ 339, 161, 18 },
		{ 339, 162, 18 },
		{ 339, 163, 18 },
		{ 339, 164, 18 },
		{ 339, 165, 18 },
		{ 339, 166, 18 },
		{ 339, 167, 18 },
		{ 339, 168, 18 },
		{ 339, 169, 18 },
		{ 339, 170, 18 },
		{ 339, 171, 18 },
		{ 339, 172, 18 },
		{ 339, 173, 18 },
		{ 339, 174, 18 },
		{ 339, 175, 18 },
		{ 339, 176, 18 },
		{ 339, 177, 18 },
		{ 339, 178, 18 },
		{ 339, 179, 18 },
		{ 339, 180, 18 },
		{ 339, 181, 18 },
		{ 339, 182, 18 },
		{ 339, 183, 18 },
		{ 339, 184, 18 },
		{ 339, 185, 18 },
		{ 339, 186, 18 },
		{ 339, 187, 18 },
		{ 339, 188, 18 },
		{ 339, 189, 18 },
		{ 339, 190, 18 },
		{ 339, 191, 18 },
		{ 339, 192, 18 },
		{ 339, 193, 18 },
		{ 339, 194, 18 },
		{ 339, 195, 18 },
		{ 340, 147, 18 },
		{ 340, 148, 18 },
		{ 340, 149, 18 },
		{ 340, 150, 18 },
		{ 340, 151, 18 },
		{ 340, 152, 18 },
		{ 340, 153, 18 },
		{ 340, 154, 18 },
		{ 340, 155, 18 },
		{ 340, 156, 18 },
		{ 340, 157, 18 },
		{ 340, 158, 18 },
		{ 340, 159, 18 },
		{ 340, 160, 18 },
		{ 340, 161, 18 },
		{ 340, 162, 18 },
		{ 340, 163, 18 },
		{ 340, 164, 18 },
		{ 340, 165, 18 },
		{ 340, 166, 18 },
		{ 340, 167, 18 },
		{ 340, 168, 18 },
		{ 340, 169, 18 },
		{ 340, 170, 18 },
		{ 340, 171, 18 },
		{ 340, 172, 18 },
		{ 340, 173, 18 },
		{ 340, 174, 18 },
		{ 340, 175, 18 },
		{ 340, 176, 18 },
		{ 340, 177, 18 },
		{ 340, 178, 18 },
		{ 340, 179, 18 },
		{ 340, 180, 18 },
		{ 340, 181, 18 },
		{ 340, 182, 18 },
		{ 340, 183, 18 },
		{ 340, 184, 18 },
		{ 340, 185, 18 },
		{ 340, 186, 18 },
		{ 340, 187, 18 },
		{ 340, 188, 18 },
		{ 340, 189, 18 },
		{ 340, 190, 18 },
		{ 340, 191, 18 },
		{ 340, 192, 18 },
		{ 340, 193, 18 },
		{ 340, 194, 18 },
		{ 341, 146, 18 },
		{ 341, 147, 18 },
		{ 341, 148, 18 },
		{ 341, 149, 18 },
		{ 341, 150, 18 },
		{ 341, 151, 18 },
		{ 341, 152, 18 },
		{ 341, 153, 18 },
		{ 341, 154, 18 },
		{ 341, 155, 18 },
		{ 341, 156, 18 },
		{ 341, 157, 18 },
		{ 341, 158, 18 },
		{ 341, 159, 18 },
		{ 341, 160, 18 },
		{ 341, 161, 18 },
		{ 341, 162, 18 },
		{ 341, 163, 18 },
		{ 341, 164, 18 },
		{ 341, 165, 18 },
		{ 341, 166, 18 },
		{ 341, 167, 18 },
		{ 341, 168, 18 },
		{ 341, 169, 18 },
		{ 341, 170, 18 },
		{ 341, 171, 18 },
		{ 341, 172, 18 },
		{ 341, 173, 18 },
		{ 341, 174, 18 },
		{ 341, 175, 18 },
		{ 341, 176, 18 },
		{ 341, 177, 18 },
		{ 341, 178, 18 },
		{ 341, 179, 18 },
		{ 341, 180, 18 },
		{ 341, 181, 18 },
		{ 341, 182, 18 },
		{ 341, 183, 18 },
		{ 341, 184, 18 },
		{ 341, 185, 18 },
		{ 341, 186, 18 },
		{ 341, 187, 18 },
		{ 341, 188, 18 },
		{ 341, 189, 18 },
		{ 341, 190, 18 },
		{ 341, 191, 18 },
		{ 341, 192, 18 },
		{ 341, 193, 18 },
		{ 342, 144, 18 },
		{ 342, 145, 18 },
		{ 342, 146, 18 },
		{ 342, 147, 18 },
		{ 342, 148, 18 },
		{ 342, 149, 18 },
		{ 342, 150, 18 },
		{ 342, 151, 18 },
		{ 342, 152, 18 },
		{ 342, 153, 18 },
		{ 342, 154, 18 },
		{ 342, 155, 18 },
		{ 342, 156, 18 },
		{ 342, 157, 18 },
		{ 342, 158, 18 },
		{ 342, 159, 18 },
		{ 342, 160, 18 },
		{ 342, 161, 18 },
		{ 342, 162, 18 },
		{ 342, 163, 18 },
		{ 342, 164, 18 },
		{ 342, 165, 18 },
		{ 342, 166, 18 },
		{ 342, 167, 18 },
		{ 342, 168, 18 },
		{ 342, 169, 18 },
		{ 342, 170, 18 },
		{ 342, 171, 18 },
		{ 342, 172, 18 },
		{ 342, 173, 18 },
		{ 342, 174, 18 },
		{ 342, 175, 18 },
		{ 342, 176, 18 },
		{ 342, 177, 18 },
		{ 342, 178, 18 },
		{ 342, 179, 18 },
		{ 342, 180, 18 },
		{ 342, 181, 18 },
		{ 342, 182, 18 },
		{ 342, 183, 18 },
		{ 342, 184, 18 },
		{ 342, 185, 18 },
		{ 342, 186, 18 },
		{ 342, 187, 18 },
		{ 342, 188, 18 },
		{ 342, 189, 18 },
		{ 342, 190, 18 },
		{ 342, 191, 18 },
		{ 342, 192, 18 },
		{ 342, 193, 18 },
		{ 343, 143, 18 },
		{ 343, 144, 18 },
		{ 343, 145, 18 },
		{ 343, 146, 18 },
		{ 343, 147, 18 },
		{ 343, 148, 18 },
		{ 343, 149, 18 },
		{ 343, 150, 18 },
		{ 343, 151, 18 },
		{ 343, 152, 18 },
		{ 343, 153, 18 },
		{ 343, 154, 18 },
		{ 343, 155, 18 },
		{ 343, 156, 18 },
		{ 343, 157, 18 },
		{ 343, 158, 18 },
		{ 343, 159, 18 },
		{ 343, 160, 18 },
		{ 343, 161, 18 },
		{ 343, 162, 18 },
		{ 343, 163, 18 },
		{ 343, 164, 18 },
		{ 343, 165, 18 },
		{ 343, 166, 18 },
		{ 343, 167, 18 },
		{ 343, 168, 18 },
		{ 343, 169, 18 },
		{ 343, 170, 18 },
		{ 343, 171, 18 },
		{ 343, 172, 18 },
		{ 343, 173, 18 },
		{ 343, 174, 18 },
		{ 343, 175, 18 },
		{ 343, 176, 18 },
		{ 343, 177, 18 },
		{ 343, 178, 18 },
		{ 343, 179, 18 },
		{ 343, 180, 18 },
		{ 343, 181, 18 },
		{ 343, 182, 18 },
		{ 343, 183, 18 },
		{ 343, 184, 18 },
		{ 343, 185, 18 },
		{ 343, 186, 18 },
		{ 343, 187, 18 },
		{ 343, 188, 18 },
		{ 343, 189, 18 },
		{ 343, 190, 18 },
		{ 343, 191, 18 },
		{ 343, 192, 18 },
		{ 344, 142, 18 },
		{ 344, 143, 18 },
		{ 344, 144, 18 },
		{ 344, 145, 18 },
		{ 344, 146, 18 },
		{ 344, 147, 18 },
		{ 344, 148, 18 },
		{ 344, 149, 18 },
		{ 344, 150, 18 },
		{ 344, 151, 18 },
		{ 344, 152, 18 },
		{ 344, 153, 18 },
		{ 344, 154, 18 },
		{ 344, 155, 18 },
		{ 344, 156, 18 },
		{ 344, 157, 18 },
		{ 344, 158, 18 },
		{ 344, 159, 18 },
		{ 344, 160, 18 },
		{ 344, 161, 18 },
		{ 344, 162, 18 },
		{ 344, 163, 18 },
		{ 344, 164, 18 },
		{ 344, 165, 18 },
		{ 344, 166, 18 },
		{ 344, 167, 18 },
		{ 344, 168, 18 },
		{ 344, 169, 18 },
		{ 344, 170, 18 },
		{ 344, 171, 18 },
		{ 344, 172, 18 },
		{ 344, 173, 18 },
		{ 344, 174, 18 },
		{ 344, 175, 18 },
		{ 344, 176, 18 },
		{ 344, 177, 18 },
		{ 344, 178, 18 },
		{ 344, 179, 18 },
		{ 344, 180, 18 },
		{ 344, 181, 18 },
		{ 344, 182, 18 },
		{ 344, 183, 18 },
		{ 344, 184, 18 },
		{ 344, 185, 18 },
		{ 344, 186, 18 },
		{ 344, 187, 18 },
		{ 344, 188, 18 },
		{ 344, 189, 18 },
		{ 344, 190, 18 },
		{ 344, 191, 18 },
		{ 345, 140, 18 },
		{ 345, 141, 18 },
		{ 345, 142, 18 },
		{ 345, 143, 18 },
		{ 345, 144, 18 },
		{ 345, 145, 18 },
		{ 345, 146, 18 },
		{ 345, 147, 18 },
		{ 345, 148, 18 },
		{ 345, 149, 18 },
		{ 345, 150, 18 },
		{ 345, 151, 18 },
		{ 345, 152, 18 },
		{ 345, 153, 18 },
		{ 345, 154, 18 },
		{ 345, 155, 18 },
		{ 345, 156, 18 },
		{ 345, 157, 18 },
		{ 345, 158, 18 },
		{ 345, 159, 18 },
		{ 345, 160, 18 },
		{ 345, 161, 18 },
		{ 345, 162, 18 },
		{ 345, 163, 18 },
		{ 345, 164, 18 },
		{ 345, 165, 18 },
		{ 345, 166, 18 },
		{ 345, 167, 18 },
		{ 345, 168, 18 },
		{ 345, 169, 18 },
		{ 345, 170, 18 },
		{ 345, 171, 18 },
		{ 345, 172, 18 },
		{ 345, 173, 18 },
		{ 345, 174, 18 },
		{ 345, 175, 18 },
		{ 345, 176, 18 },
		{ 345, 177, 18 },
		{ 345, 178, 18 },
		{ 345, 179, 18 },
		{ 345, 180, 18 },
		{ 345, 181, 18 },
		{ 345, 182, 18 },
		{ 345, 183, 18 },
		{ 345, 184, 18 },
		{ 345, 185, 18 },
		{ 345, 186, 18 },
		{ 345, 187, 18 },
		{ 345, 188, 18 },
		{ 345, 189, 18 },
		{ 345, 190, 18 },
		{ 346, 139, 18 },
		{ 346, 140, 18 },
		{ 346, 141, 18 },
		{ 346, 142, 18 },
		{ 346, 143, 18 },
		{ 346, 144, 18 },
		{ 346, 145, 18 },
		{ 346, 146, 18 },
		{ 346, 147, 18 },
		{ 346, 148, 18 },
		{ 346, 149, 18 },
		{ 346, 150, 18 },
		{ 346, 151, 18 },
		{ 346, 152, 18 },
		{ 346, 153, 18 },
		{ 346, 154, 18 },
		{ 346, 155, 18 },
		{ 346, 156, 18 },
		{ 346, 157, 18 },
		{ 346, 158, 18 },
		{ 346, 159, 18 },
		{ 346, 160, 18 },
		{ 346, 161, 18 },
		{ 346, 162, 18 },
		{ 346, 163, 18 },
		{ 346, 164, 18 },
		{ 346, 165, 18 },
		{ 346, 166, 18 },
		{ 346, 167, 18 },
		{ 346, 168, 18 },
		{ 346, 169, 18 },
		{ 346, 170, 18 },
		{ 346, 171, 18 },
		{ 346, 172, 18 },
		{ 346, 173, 18 },
		{ 346, 174, 18 },
		{ 346, 175, 18 },
		{ 346, 176, 18 },
		{ 346, 177, 18 },
		{ 346, 178, 18 },
		{ 346, 179, 18 },
		{ 346, 180, 18 },
		{ 346, 181, 18 },
		{ 346, 182, 18 },
		{ 346, 183, 18 },
		{ 346, 184, 18 },
		{ 346, 185, 18 },
		{ 346, 186, 18 },
		{ 346, 187, 18 },
		{ 346, 188, 18 },
		{ 346, 189, 18 },
		{ 346, 190, 18 },
		{ 347, 138, 18 },
		{ 347, 139, 18 },
		{ 347, 140, 18 },
		{ 347, 141, 18 },
		{ 347, 142, 18 },
		{ 347, 143, 18 },
		{ 347, 144, 18 },
		{ 347, 145, 18 },
		{ 347, 146, 18 },
		{ 347, 147, 18 },
		{ 347, 148, 18 },
		{ 347, 149, 18 },
		{ 347, 150, 18 },
		{ 347, 151, 18 },
		{ 347, 152, 18 },
		{ 347, 153, 18 },
		{ 347, 154, 18 },
		{ 347, 155, 18 },
		{ 347, 156, 18 },
		{ 347, 157, 18 },
		{ 347, 158, 18 },
		{ 347, 159, 18 },
		{ 347, 160, 18 },
		{ 347, 161, 18 },
		{ 347, 162, 18 },
		{ 347, 163, 18 },
		{ 347, 164, 18 },
		{ 347, 165, 18 },
		{ 347, 166, 18 },
		{ 347, 167, 18 },
		{ 347, 168, 18 },
		{ 347, 169, 18 },
		{ 347, 170, 18 },
		{ 347, 171, 18 },
		{ 347, 172, 18 },
		{ 347, 173, 18 },
		{ 347, 174, 18 },
		{ 347, 175, 18 },
		{ 347, 176, 18 },
		{ 347, 177, 18 },
		{ 347, 178, 18 },
		{ 347, 179, 18 },
		{ 347, 180, 18 },
		{ 347, 181, 18 },
		{ 347, 182, 18 },
		{ 347, 183, 18 },
		{ 347, 184, 18 },
		{ 347, 185, 18 },
		{ 347, 186, 18 },
		{ 347, 187, 18 },
		{ 347, 188, 18 },
		{ 347, 189, 18 },
		{ 348, 137, 18 },
		{ 348, 138, 18 },
		{ 348, 139, 18 },
		{ 348, 140, 18 },
		{ 348, 141, 18 },
		{ 348, 142, 18 },
		{ 348, 143, 18 },
		{ 348, 144, 18 },
		{ 348, 145, 18 },
		{ 348, 146, 18 },
		{ 348, 147, 18 },
		{ 348, 148, 18 },
		{ 348, 149, 18 },
		{ 348, 150, 18 },
		{ 348, 151, 18 },
		{ 348, 152, 18 },
		{ 348, 153, 18 },
		{ 348, 154, 18 },
		{ 348, 155, 18 },
		{ 348, 156, 18 },
		{ 348, 157, 18 },
		{ 348, 158, 18 },
		{ 348, 159, 18 },
		{ 348, 160, 18 },
		{ 348, 161, 18 },
		{ 348, 162, 18 },
		{ 348, 163, 18 },
		{ 348, 164, 18 },
		{ 348, 165, 18 },
		{ 348, 166, 18 },
		{ 348, 167, 18 },
		{ 348, 168, 18 },
		{ 348, 169, 18 },
		{ 348, 170, 18 },
		{ 348, 171, 18 },
		{ 348, 172, 18 },
		{ 348, 173, 18 },
		{ 348, 174, 18 },
		{ 348, 175, 18 },
		{ 348, 176, 18 },
		{ 348, 177, 18 },
		{ 348, 178, 18 },
		{ 348, 179, 18 },
		{ 348, 180, 18 },
		{ 348, 181, 18 },
		{ 348, 182, 18 },
		{ 348, 183, 18 },
		{ 348, 184, 18 },
		{ 348, 185, 18 },
		{ 348, 186, 18 },
		{ 348, 187, 18 },
		{ 348, 188, 18 },
		{ 349, 136, 18 },
		{ 349, 137, 18 },
		{ 349, 138, 18 },
		{ 349, 139, 18 },
		{ 349, 140, 18 },
		{ 349, 141, 18 },
		{ 349, 142, 18 },
		{ 349, 143, 18 },
		{ 349, 144, 18 },
		{ 349, 145, 18 },
		{ 349, 146, 18 },
		{ 349, 147, 18 },
		{ 349, 148, 18 },
		{ 349, 149, 18 },
		{ 349, 150, 18 },
		{ 349, 151, 18 },
		{ 349, 152, 18 },
		{ 349, 153, 18 },
		{ 349, 154, 18 },
		{ 349, 155, 18 },
		{ 349, 156, 18 },
		{ 349, 157, 18 },
		{ 349, 158, 18 },
		{ 349, 159, 18 },
		{ 349, 160, 18 },
		{ 349, 161, 18 },
		{ 349, 162, 18 },
		{ 349, 163, 18 },
		{ 349, 164, 18 },
		{ 349, 165, 18 },
		{ 349, 166, 18 },
		{ 349, 167, 18 },
		{ 349, 168, 18 },
		{ 349, 169, 18 },
		{ 349, 170, 18 },
		{ 349, 171, 18 },
		{ 349, 172, 18 },
		{ 349, 173, 18 },
		{ 349, 174, 18 },
		{ 349, 175, 18 },
		{ 349, 176, 18 },
		{ 349, 177, 18 },
		{ 349, 178, 18 },
		{ 349, 179, 18 },
		{ 349, 180, 18 },
		{ 349, 181, 18 },
		{ 349, 182, 18 },
		{ 349, 183, 18 },
		{ 349, 184, 18 },
		{ 349, 185, 18 },
		{ 349, 186, 18 },
		{ 349, 187, 18 },
		{ 350, 135, 18 },
		{ 350, 136, 18 },
		{ 350, 137, 18 },
		{ 350, 138, 18 },
		{ 350, 139, 18 },
		{ 350, 140, 18 },
		{ 350, 141, 18 },
		{ 350, 142, 18 },
		{ 350, 143, 18 },
		{ 350, 144, 18 },
		{ 350, 145, 18 },
		{ 350, 146, 18 },
		{ 350, 147, 18 },
		{ 350, 148, 18 },
		{ 350, 149, 18 },
		{ 350, 150, 18 },
		{ 350, 151, 18 },
		{ 350, 152, 18 },
		{ 350, 153, 18 },
		{ 350, 154, 18 },
		{ 350, 155, 18 },
		{ 350, 156, 18 },
		{ 350, 157, 18 },
		{ 350, 158, 18 },
		{ 350, 159, 18 },
		{ 350, 160, 18 },
		{ 350, 161, 18 },
		{ 350, 162, 18 },
		{ 350, 163, 18 },
		{ 350, 164, 18 },
		{ 350, 165, 18 },
		{ 350, 166, 18 },
		{ 350, 167, 18 },
		{ 350, 168, 18 },
		{ 350, 169, 18 },
		{ 350, 170, 18 },
		{ 350, 171, 18 },
		{ 350, 172, 18 },
		{ 350, 173, 18 },
		{ 350, 174, 18 },
		{ 350, 175, 18 },
		{ 350, 176, 18 },
		{ 350, 177, 18 },
		{ 350, 178, 18 },
		{ 350, 179, 18 },
		{ 350, 180, 18 },
		{ 350, 181, 18 },
		{ 350, 182, 18 },
		{ 350, 183, 18 },
		{ 350, 184, 18 },
		{ 350, 185, 18 },
		{ 350, 186, 18 },
		{ 350, 187, 18 },
		{ 351, 134, 18 },
		{ 351, 135, 18 },
		{ 351, 136, 18 },
		{ 351, 137, 18 },
		{ 351, 138, 18 },
		{ 351, 139, 18 },
		{ 351, 140, 18 },
		{ 351, 141, 18 },
		{ 351, 142, 18 },
		{ 351, 143, 18 },
		{ 351, 144, 18 },
		{ 351, 145, 18 },
		{ 351, 146, 18 },
		{ 351, 147, 18 },
		{ 351, 148, 18 },
		{ 351, 149, 18 },
		{ 351, 150, 18 },
		{ 351, 151, 18 },
		{ 351, 152, 18 },
		{ 351, 153, 18 },
		{ 351, 154, 18 },
		{ 351, 155, 18 },
		{ 351, 156, 18 },
		{ 351, 157, 18 },
		{ 351, 158, 18 },
		{ 351, 159, 18 },
		{ 351, 160, 18 },
		{ 351, 161, 18 },
		{ 351, 162, 18 },
		{ 351, 163, 18 },
		{ 351, 164, 18 },
		{ 351, 165, 18 },
		{ 351, 166, 18 },
		{ 351, 167, 18 },
		{ 351, 168, 18 },
		{ 351, 169, 18 },
		{ 351, 170, 18 },
		{ 351, 171, 18 },
		{ 351, 172, 18 },
		{ 351, 173, 18 },
		{ 351, 174, 18 },
		{ 351, 175, 18 },
		{ 351, 176, 18 },
		{ 351, 177, 18 },
		{ 351, 178, 18 },
		{ 351, 179, 18 },
		{ 351, 180, 18 },
		{ 351, 181, 18 },
		{ 351, 182, 18 },
		{ 351, 183, 18 },
		{ 351, 184, 18 },
		{ 351, 185, 18 },
		{ 351, 186, 18 },
		{ 352, 133, 18 },
		{ 352, 134, 18 },
		{ 352, 135, 18 },
		{ 352, 136, 18 },
		{ 352, 137, 18 },
		{ 352, 138, 18 },
		{ 352, 139, 18 },
		{ 352, 140, 18 },
		{ 352, 141, 18 },
		{ 352, 142, 18 },
		{ 352, 143, 18 },
		{ 352, 144, 18 },
		{ 352, 145, 18 },
		{ 352, 146, 18 },
		{ 352, 147, 18 },
		{ 352, 148, 18 },
		{ 352, 149, 18 },
		{ 352, 150, 18 },
		{ 352, 151, 18 },
		{ 352, 152, 18 },
		{ 352, 153, 18 },
		{ 352, 154, 18 },
		{ 352, 155, 18 },
		{ 352, 156, 18 },
		{ 352, 157, 18 },
		{ 352, 158, 18 },
		{ 352, 159, 18 },
		{ 352, 160, 18 },
		{ 352, 161, 18 },
		{ 352, 162, 18 },
		{ 352, 163, 18 },
		{ 352, 164, 18 },
		{ 352, 165, 18 },
		{ 352, 166, 18 },
		{ 352, 167, 18 },
		{ 352, 168, 18 },
		{ 352, 169, 18 },
		{ 352, 170, 18 },
		{ 352, 171, 18 },
		{ 352, 172, 18 },
		{ 352, 173, 18 },
		{ 352, 174, 18 },
		{ 352, 175, 18 },
		{ 352, 176, 18 },
		{ 352, 177, 18 },
		{ 352, 178, 18 },
		{ 352, 179, 18 },
		{ 352, 180, 18 },
		{ 352, 181, 18 },
		{ 352, 182, 18 },
		{ 352, 183, 18 },
		{ 352, 184, 18 },
		{ 352, 185, 18 },
		{ 353, 133, 18 },
		{ 353, 134, 18 },
		{ 353, 135, 18 },
		{ 353, 136, 18 },
		{ 353, 137, 18 },
		{ 353, 138, 18 },
		{ 353, 139, 18 },
		{ 353, 140, 18 },
		{ 353, 141, 18 },
		{ 353, 142, 18 },
		{ 353, 143, 18 },
		{ 353, 144, 18 },
		{ 353, 145, 18 },
		{ 353, 146, 18 },
		{ 353, 147, 18 },
		{ 353, 148, 18 },
		{ 353, 149, 18 },
		{ 353, 150, 18 },
		{ 353, 151, 18 },
		{ 353, 152, 18 },
		{ 353, 153, 18 },
		{ 353, 154, 18 },
		{ 353, 155, 18 },
		{ 353, 156, 18 },
		{ 353, 157, 18 },
		{ 353, 158, 18 },
		{ 353, 159, 18 },
		{ 353, 160, 18 },
		{ 353, 161, 18 },
		{ 353, 162, 18 },
		{ 353, 163, 18 },
		{ 353, 164, 18 },
		{ 353, 165, 18 },
		{ 353, 166, 18 },
		{ 353, 167, 18 },
		{ 353, 168, 18 },
		{ 353, 169, 18 },
		{ 353, 170, 18 },
		{ 353, 171, 18 },
		{ 353, 172, 18 },
		{ 353, 173, 18 },
		{ 353, 174, 18 },
		{ 353, 175, 18 },
		{ 353, 176, 18 },
		{ 353, 177, 18 },
		{ 353, 178, 18 },
		{ 353, 179, 18 },
		{ 353, 180, 18 },
		{ 353, 181, 18 },
		{ 353, 182, 18 },
		{ 353, 183, 18 },
		{ 353, 184, 18 },
		{ 354, 132, 18 },
		{ 354, 133, 18 },
		{ 354, 134, 18 },
		{ 354, 135, 18 },
		{ 354, 136, 18 },
		{ 354, 137, 18 },
		{ 354, 138, 18 },
		{ 354, 139, 18 },
		{ 354, 140, 18 },
		{ 354, 141, 18 },
		{ 354, 142, 18 },
		{ 354, 143, 18 },
		{ 354, 144, 18 },
		{ 354, 145, 18 },
		{ 354, 146, 18 },
		{ 354, 147, 18 },
		{ 354, 148, 18 },
		{ 354, 149, 18 },
		{ 354, 150, 18 },
		{ 354, 151, 18 },
		{ 354, 152, 18 },
		{ 354, 153, 18 },
		{ 354, 154, 18 },
		{ 354, 155, 18 },
		{ 354, 156, 18 },
		{ 354, 157, 18 },
		{ 354, 158, 18 },
		{ 354, 159, 18 },
		{ 354, 160, 18 },
		{ 354, 161, 18 },
		{ 354, 162, 18 },
		{ 354, 163, 18 },
		{ 354, 164, 18 },
		{ 354, 165, 18 },
		{ 354, 166, 18 },
		{ 354, 167, 18 },
		{ 354, 168, 18 },
		{ 354, 169, 18 },
		{ 354, 170, 18 },
		{ 354, 171, 18 },
		{ 354, 172, 18 },
		{ 354, 173, 18 },
		{ 354, 174, 18 },
		{ 354, 175, 18 },
		{ 354, 176, 18 },
		{ 354, 177, 18 },
		{ 354, 178, 18 },
		{ 354, 179, 18 },
		{ 354, 180, 18 },
		{ 354, 181, 18 },
		{ 354, 182, 18 },
		{ 354, 183, 18 },
		{ 354, 184, 18 },
		{ 355, 132, 18 },
		{ 355, 133, 18 },
		{ 355, 134, 18 },
		{ 355, 135, 18 },
		{ 355, 136, 18 },
		{ 355, 137, 18 },
		{ 355, 138, 18 },
		{ 355, 139, 18 },
		{ 355, 140, 18 },
		{ 355, 141, 18 },
		{ 355, 142, 18 },
		{ 355, 143, 18 },
		{ 355, 144, 18 },
		{ 355, 145, 18 },
		{ 355, 146, 18 },
		{ 355, 147, 18 },
		{ 355, 148, 18 },
		{ 355, 149, 18 },
		{ 355, 150, 18 },
		{ 355, 151, 18 },
		{ 355, 152, 18 },
		{ 355, 153, 18 },
		{ 355, 154, 18 },
		{ 355, 155, 18 },
		{ 355, 156, 18 },
		{ 355, 157, 18 },
		{ 355, 158, 18 },
		{ 355, 159, 18 },
		{ 355, 160, 18 },
		{ 355, 161, 18 },
		{ 355, 162, 18 },
		{ 355, 163, 18 },
		{ 355, 164, 18 },
		{ 355, 165, 18 },
		{ 355, 166, 18 },
		{ 355, 167, 18 },
		{ 355, 168, 18 },
		{ 355, 169, 18 },
		{ 355, 170, 18 },
		{ 355, 171, 18 },
		{ 355, 172, 18 },
		{ 355, 173, 18 },
		{ 355, 174, 18 },
		{ 355, 175, 18 },
		{ 355, 176, 18 },
		{ 355, 177, 18 },
		{ 355, 178, 18 },
		{ 355, 179, 18 },
		{ 355, 180, 18 },
		{ 355, 181, 18 },
		{ 355, 182, 18 },
		{ 355, 183, 18 },
		{ 356, 132, 18 },
		{ 356, 133, 18 },
		{ 356, 134, 18 },
		{ 356, 135, 18 },
		{ 356, 136, 18 },
		{ 356, 137, 18 },
		{ 356, 138, 18 },
		{ 356, 139, 18 },
		{ 356, 140, 18 },
		{ 356, 141, 18 },
		{ 356, 142, 18 },
		{ 356, 143, 18 },
		{ 356, 144, 18 },
		{ 356, 145, 18 },
		{ 356, 146, 18 },
		{ 356, 147, 18 },
		{ 356, 148, 18 },
		{ 356, 149, 18 },
		{ 356, 150, 18 },
		{ 356, 151, 18 },
		{ 356, 152, 18 },
		{ 356, 153, 18 },
		{ 356, 154, 18 },
		{ 356, 155, 18 },
		{ 356, 156, 18 },
		{ 356, 157, 18 },
		{ 356, 158, 18 },
		{ 356, 159, 18 },
		{ 356, 160, 18 },
		{ 356, 161, 18 },
		{ 356, 162, 18 },
		{ 356, 163, 18 },
		{ 356, 164, 18 },
		{ 356, 165, 18 },
		{ 356, 166, 18 },
		{ 356, 167, 18 },
		{ 356, 168, 18 },
		{ 356, 169, 18 },
		{ 356, 170, 18 },
		{ 356, 171, 18 },
		{ 356, 172, 18 },
		{ 356, 173, 18 },
		{ 356, 174, 18 },
		{ 356, 175, 18 },
		{ 356, 176, 18 },
		{ 356, 177, 18 },
		{ 356, 178, 18 },
		{ 356, 179, 18 },
		{ 356, 180, 18 },
		{ 356, 181, 18 },
		{ 356, 182, 18 },
		{ 357, 132, 18 },
		{ 357, 133, 18 },
		{ 357, 134, 18 },
		{ 357, 135, 18 },
		{ 357, 136, 18 },
		{ 357, 137, 18 },
		{ 357, 138, 18 },
		{ 357, 139, 18 },
		{ 357, 140, 18 },
		{ 357, 141, 18 },
		{ 357, 142, 18 },
		{ 357, 143, 18 },
		{ 357, 144, 18 },
		{ 357, 145, 18 },
		{ 357, 146, 18 },
		{ 357, 147, 18 },
		{ 357, 148, 18 },
		{ 357, 149, 18 },
		{ 357, 150, 18 },
		{ 357, 151, 18 },
		{ 357, 152, 18 },
		{ 357, 153, 18 },
		{ 357, 154, 18 },
		{ 357, 155, 18 },
		{ 357, 156, 18 },
		{ 357, 157, 18 },
		{ 357, 158, 18 },
		{ 357, 159, 18 },
		{ 357, 160, 18 },
		{ 357, 161, 18 },
		{ 357, 162, 18 },
		{ 357, 163, 18 },
		{ 357, 164, 18 },
		{ 357, 165, 18 },
		{ 357, 166, 18 },
		{ 357, 167, 18 },
		{ 357, 168, 18 },
		{ 357, 169, 18 },
		{ 357, 170, 18 },
		{ 357, 171, 18 },
		{ 357, 172, 18 },
		{ 357, 173, 18 },
		{ 357, 174, 18 },
		{ 357, 175, 18 },
		{ 357, 176, 18 },
		{ 357, 177, 18 },
		{ 357, 178, 18 },
		{ 357, 179, 18 },
		{ 357, 180, 18 },
		{ 357, 181, 18 },
		{ 358, 132, 18 },
		{ 358, 133, 18 },
		{ 358, 134, 18 },
		{ 358, 135, 18 },
		{ 358, 136, 18 },
		{ 358, 137, 18 },
		{ 358, 138, 18 },
		{ 358, 139, 18 },
		{ 358, 140, 18 },
		{ 358, 141, 18 },
		{ 358, 142, 18 },
		{ 358, 143, 18 },
		{ 358, 144, 18 },
		{ 358, 145, 18 },
		{ 358, 146, 18 },
		{ 358, 147, 18 },
		{ 358, 148, 18 },
		{ 358, 149, 18 },
		{ 358, 150, 18 },
		{ 358, 151, 18 },
		{ 358, 152, 18 },
		{ 358, 153, 18 },
		{ 358, 154, 18 },
		{ 358, 155, 18 },
		{ 358, 156, 18 },
		{ 358, 157, 18 },
		{ 358, 158, 18 },
		{ 358, 159, 18 },
		{ 358, 160, 18 },
		{ 358, 161, 18 },
		{ 358, 162, 18 },
		{ 358, 163, 18 },
		{ 358, 164, 18 },
		{ 358, 165, 18 },
		{ 358, 166, 18 },
		{ 358, 167, 18 },
		{ 358, 168, 18 },
		{ 358, 169, 18 },
		{ 358, 170, 18 },
		{ 358, 171, 18 },
		{ 358, 172, 18 },
		{ 358, 173, 18 },
		{ 358, 174, 18 },
		{ 358, 175, 18 },
		{ 358, 176, 18 },
		{ 358, 177, 18 },
		{ 358, 178, 18 },
		{ 358, 179, 18 },
		{ 358, 180, 18 },
		{ 358, 181, 18 },
		{ 359, 131, 18 },
		{ 359, 132, 18 },
		{ 359, 133, 18 },
		{ 359, 134, 18 },
		{ 359, 135, 18 },
		{ 359, 136, 18 },
		{ 359, 137, 18 },
		{ 359, 138, 18 },
		{ 359, 139, 18 },
		{ 359, 140, 18 },
		{ 359, 141, 18 },
		{ 359, 142, 18 },
		{ 359, 143, 18 },
		{ 359, 144, 18 },
		{ 359, 145, 18 },
		{ 359, 146, 18 },
		{ 359, 147, 18 },
		{ 359, 148, 18 },
		{ 359, 149, 18 },
		{ 359, 150, 18 },
		{ 359, 151, 18 },
		{ 359, 152, 18 },
		{ 359, 153, 18 },
		{ 359, 154, 18 },
		{ 359, 155, 18 },
		{ 359, 156, 18 },
		{ 359, 157, 18 },
		{ 359, 158, 18 },
		{ 359, 159, 18 },
		{ 359, 160, 18 },
		{ 359, 161, 18 },
		{ 359, 162, 18 },
		{ 359, 163, 18 },
		{ 359, 164, 18 },
		{ 359, 165, 18 },
		{ 359, 166, 18 },
		{ 359, 167, 18 },
		{ 359, 168, 18 },
		{ 359, 169, 18 },
		{ 359, 170, 18 },
		{ 359, 171, 18 },
		{ 359, 172, 18 },
		{ 359, 173, 18 },
		{ 359, 174, 18 },
		{ 359, 175, 18 },
		{ 359, 176, 18 },
		{ 359, 177, 18 },
		{ 359, 178, 18 },
		{ 359, 179, 18 },
		{ 359, 180, 18 },
		{ 360, 131, 18 },
		{ 360, 132, 18 },
		{ 360, 133, 18 },
		{ 360, 134, 18 },
		{ 360, 135, 18 },
		{ 360, 136, 18 },
		{ 360, 137, 18 },
		{ 360, 138, 18 },
		{ 360, 139, 18 },
		{ 360, 140, 18 },
		{ 360, 141, 18 },
		{ 360, 142, 18 },
		{ 360, 143, 18 },
		{ 360, 144, 18 },
		{ 360, 145, 18 },
		{ 360, 146, 18 },
		{ 360, 147, 18 },
		{ 360, 148, 18 },
		{ 360, 149, 18 },
		{ 360, 150, 18 },
		{ 360, 151, 18 },
		{ 360, 152, 18 },
		{ 360, 153, 18 },
		{ 360, 154, 18 },
		{ 360, 155, 18 },
		{ 360, 156, 18 },
		{ 360, 157, 18 },
		{ 360, 158, 18 },
		{ 360, 159, 18 },
		{ 360, 160, 18 },
		{ 360, 161, 18 },
		{ 360, 162, 18 },
		{ 360, 163, 18 },
		{ 360, 164, 18 },
		{ 360, 165, 18 },
		{ 360, 166, 18 },
		{ 360, 167, 18 },
		{ 360, 168, 18 },
		{ 360, 169, 18 },
		{ 360, 170, 18 },
		{ 360, 171, 18 },
		{ 360, 172, 18 },
		{ 360, 173, 18 },
		{ 360, 174, 18 },
		{ 360, 175, 18 },
		{ 360, 176, 18 },
		{ 360, 177, 18 },
		{ 360, 178, 18 },
		{ 360, 179, 18 },
		{ 361, 131, 18 },
		{ 361, 132, 18 },
		{ 361, 133, 18 },
		{ 361, 134, 18 },
		{ 361, 135, 18 },
		{ 361, 136, 18 },
		{ 361, 137, 18 },
		{ 361, 138, 18 },
		{ 361, 139, 18 },
		{ 361, 140, 18 },
		{ 361, 141, 18 },
		{ 361, 142, 18 },
		{ 361, 143, 18 },
		{ 361, 144, 18 },
		{ 361, 145, 18 },
		{ 361, 146, 18 },
		{ 361, 147, 18 },
		{ 361, 148, 18 },
		{ 361, 149, 18 },
		{ 361, 150, 18 },
		{ 361, 151, 18 },
		{ 361, 152, 18 },
		{ 361, 153, 18 },
		{ 361, 154, 18 },
		{ 361, 155, 18 },
		{ 361, 156, 18 },
		{ 361, 157, 18 },
		{ 361, 158, 18 },
		{ 361, 159, 18 },
		{ 361, 160, 18 },
		{ 361, 161, 18 },
		{ 361, 162, 18 },
		{ 361, 163, 18 },
		{ 361, 164, 18 },
		{ 361, 165, 18 },
		{ 361, 166, 18 },
		{ 361, 167, 18 },
		{ 361, 168, 18 },
		{ 361, 169, 18 },
		{ 361, 170, 18 },
		{ 361, 171, 18 },
		{ 361, 172, 18 },
		{ 361, 173, 18 },
		{ 361, 174, 18 },
		{ 361, 175, 18 },
		{ 361, 176, 18 },
		{ 361, 177, 18 },
		{ 361, 178, 18 },
		{ 362, 131, 18 },
		{ 362, 132, 18 },
		{ 362, 133, 18 },
		{ 362, 134, 18 },
		{ 362, 135, 18 },
		{ 362, 136, 18 },
		{ 362, 137, 18 },
		{ 362, 138, 18 },
		{ 362, 139, 18 },
		{ 362, 140, 18 },
		{ 362, 141, 18 },
		{ 362, 142, 18 },
		{ 362, 143, 18 },
		{ 362, 144, 18 },
		{ 362, 145, 18 },
		{ 362, 146, 18 },
		{ 362, 147, 18 },
		{ 362, 148, 18 },
		{ 362, 149, 18 },
		{ 362, 150, 18 },
		{ 362, 151, 18 },
		{ 362, 152, 18 },
		{ 362, 153, 18 },
		{ 362, 154, 18 },
		{ 362, 155, 18 },
		{ 362, 156, 18 },
		{ 362, 157, 18 },
		{ 362, 158, 18 },
		{ 362, 159, 18 },
		{ 362, 160, 18 },
		{ 362, 161, 18 },
		{ 362, 162, 18 },
		{ 362, 163, 18 },
		{ 362, 164, 18 },
		{ 362, 165, 18 },
		{ 362, 166, 18 },
		{ 362, 167, 18 },
		{ 362, 168, 18 },
		{ 362, 169, 18 },
		{ 362, 170, 18 },
		{ 362, 171, 18 },
		{ 362, 172, 18 },
		{ 362, 173, 18 },
		{ 362, 174, 18 },
		{ 362, 175, 18 },
		{ 362, 176, 18 },
		{ 362, 177, 18 },
		{ 362, 178, 18 },
		{ 363, 131, 18 },
		{ 363, 132, 18 },
		{ 363, 133, 18 },
		{ 363, 134, 18 },
		{ 363, 135, 18 },
		{ 363, 136, 18 },
		{ 363, 137, 18 },
		{ 363, 138, 18 },
		{ 363, 139, 18 },
		{ 363, 140, 18 },
		{ 363, 141, 18 },
		{ 363, 142, 18 },
		{ 363, 143, 18 },
		{ 363, 144, 18 },
		{ 363, 145, 18 },
		{ 363, 146, 18 },
		{ 363, 147, 18 },
		{ 363, 148, 18 },
		{ 363, 149, 18 },
		{ 363, 150, 18 },
		{ 363, 151, 18 },
		{ 363, 152, 18 },
		{ 363, 153, 18 },
		{ 363, 154, 18 },
		{ 363, 155, 18 },
		{ 363, 156, 18 },
		{ 363, 157, 18 },
		{ 363, 158, 18 },
		{ 363, 159, 18 },
		{ 363, 160, 18 },
		{ 363, 161, 18 },
		{ 363, 162, 18 },
		{ 363, 163, 18 },
		{ 363, 164, 18 },
		{ 363, 165, 18 },
		{ 363, 166, 18 },
		{ 363, 167, 18 },
		{ 363, 168, 18 },
		{ 363, 169, 18 },
		{ 363, 170, 18 },
		{ 363, 171, 18 },
		{ 363, 172, 18 },
		{ 363, 173, 18 },
		{ 363, 174, 18 },
		{ 363, 175, 18 },
		{ 363, 176, 18 },
		{ 363, 177, 18 },
		{ 364, 131, 18 },
		{ 364, 132, 18 },
		{ 364, 133, 18 },
		{ 364, 134, 18 },
		{ 364, 135, 18 },
		{ 364, 136, 18 },
		{ 364, 137, 18 },
		{ 364, 138, 18 },
		{ 364, 139, 18 },
		{ 364, 140, 18 },
		{ 364, 141, 18 },
		{ 364, 142, 18 },
		{ 364, 143, 18 },
		{ 364, 144, 18 },
		{ 364, 145, 18 },
		{ 364, 146, 18 },
		{ 364, 147, 18 },
		{ 364, 148, 18 },
		{ 364, 149, 18 },
		{ 364, 150, 18 },
		{ 364, 151, 18 },
		{ 364, 152, 18 },
		{ 364, 153, 18 },
		{ 364, 154, 18 },
		{ 364, 155, 18 },
		{ 364, 156, 18 },
		{ 364, 157, 18 },
		{ 364, 158, 18 },
		{ 364, 159, 18 },
		{ 364, 160, 18 },
		{ 364, 161, 18 },
		{ 364, 162, 18 },
		{ 364, 163, 18 },
		{ 364, 164, 18 },
		{ 364, 165, 18 },
		{ 364, 166, 18 },
		{ 364, 167, 18 },
		{ 364, 168, 18 },
		{ 364, 169, 18 },
		{ 364, 170, 18 },
		{ 364, 171, 18 },
		{ 364, 172, 18 },
		{ 364, 173, 18 },
		{ 364, 174, 18 },
		{ 364, 175, 18 },
		{ 364, 176, 18 },
		{ 365, 131, 18 },
		{ 365, 132, 18 },
		{ 365, 133, 18 },
		{ 365, 134, 18 },
		{ 365, 135, 18 },
		{ 365, 136, 18 },
		{ 365, 137, 18 },
		{ 365, 138, 18 },
		{ 365, 139, 18 },
		{ 365, 140, 18 },
		{ 365, 141, 18 },
		{ 365, 142, 18 },
		{ 365, 143, 18 },
		{ 365, 144, 18 },
		{ 365, 145, 18 },
		{ 365, 146, 18 },
		{ 365, 147, 18 },
		{ 365, 148, 18 },
		{ 365, 149, 18 },
		{ 365, 150, 18 },
		{ 365, 151, 18 },
		{ 365, 152, 18 },
		{ 365, 153, 18 },
		{ 365, 154, 18 },
		{ 365, 155, 18 },
		{ 365, 156, 18 },
		{ 365, 157, 18 },
		{ 365, 158, 18 },
		{ 365, 159, 18 },
		{ 365, 160, 18 },
		{ 365, 161, 18 },
		{ 365, 162, 18 },
		{ 365, 163, 18 },
		{ 365, 164, 18 },
		{ 365, 165, 18 },
		{ 365, 166, 18 },
		{ 365, 167, 18 },
		{ 365, 168, 18 },
		{ 365, 169, 18 },
		{ 365, 170, 18 },
		{ 365, 171, 18 },
		{ 365, 172, 18 },
		{ 365, 173, 18 },
		{ 365, 174, 18 },
		{ 365, 175, 18 },
		{ 366, 131, 18 },
		{ 366, 132, 18 },
		{ 366, 133, 18 },
		{ 366, 134, 18 },
		{ 366, 135, 18 },
		{ 366, 136, 18 },
		{ 366, 137, 18 },
		{ 366, 138, 18 },
		{ 366, 139, 18 },
		{ 366, 140, 18 },
		{ 366, 141, 18 },
		{ 366, 142, 18 },
		{ 366, 143, 18 },
		{ 366, 144, 18 },
		{ 366, 145, 18 },
		{ 366, 146, 18 },
		{ 366, 147, 18 },
		{ 366, 148, 18 },
		{ 366, 149, 18 },
		{ 366, 150, 18 },
		{ 366, 151, 18 },
		{ 366, 152, 18 },
		{ 366, 153, 18 },
		{ 366, 154, 18 },
		{ 366, 155, 18 },
		{ 366, 156, 18 },
		{ 366, 157, 18 },
		{ 366, 158, 18 },
		{ 366, 159, 18 },
		{ 366, 160, 18 },
		{ 366, 161, 18 },
		{ 366, 162, 18 },
		{ 366, 163, 18 },
		{ 366, 164, 18 },
		{ 366, 165, 18 },
		{ 366, 166, 18 },
		{ 366, 167, 18 },
		{ 366, 168, 18 },
		{ 366, 169, 18 },
		{ 366, 170, 18 },
		{ 366, 171, 18 },
		{ 366, 172, 18 },
		{ 366, 173, 18 },
		{ 366, 174, 18 },
		{ 366, 175, 18 },
		{ 367, 131, 18 },
		{ 367, 132, 18 },
		{ 367, 133, 18 },
		{ 367, 134, 18 },
		{ 367, 135, 18 },
		{ 367, 136, 18 },
		{ 367, 137, 18 },
		{ 367, 138, 18 },
		{ 367, 139, 18 },
		{ 367, 140, 18 },
		{ 367, 141, 18 },
		{ 367, 142, 18 },
		{ 367, 143, 18 },
		{ 367, 144, 18 },
		{ 367, 145, 18 },
		{ 367, 146, 18 },
		{ 367, 147, 18 },
		{ 367, 148, 18 },
		{ 367, 149, 18 },
		{ 367, 150, 18 },
		{ 367, 151, 18 },
		{ 367, 152, 18 },
		{ 367, 153, 18 },
		{ 367, 154, 18 },
		{ 367, 155, 18 },
		{ 367, 156, 18 },
		{ 367, 157, 18 },
		{ 367, 158, 18 },
		{ 367, 159, 18 },
		{ 367, 160, 18 },
		{ 367, 161, 18 },
		{ 367, 162, 18 },
		{ 367, 163, 18 },
		{ 367, 164, 18 },
		{ 367, 165, 18 },
		{ 367, 166, 18 },
		{ 367, 167, 18 },
		{ 367, 168, 18 },
		{ 367, 169, 18 },
		{ 367, 170, 18 },
		{ 367, 171, 18 },
		{ 367, 172, 18 },
		{ 367, 173, 18 },
		{ 367, 174, 18 },
		{ 368, 131, 18 },
		{ 368, 132, 18 },
		{ 368, 133, 18 },
		{ 368, 134, 18 },
		{ 368, 135, 18 },
		{ 368, 136, 18 },
		{ 368, 137, 18 },
		{ 368, 138, 18 },
		{ 368, 139, 18 },
		{ 368, 140, 18 },
		{ 368, 141, 18 },
		{ 368, 142, 18 },
		{ 368, 143, 18 },
		{ 368, 144, 18 },
		{ 368, 145, 18 },
		{ 368, 146, 18 },
		{ 368, 147, 18 },
		{ 368, 148, 18 },
		{ 368, 149, 18 },
		{ 368, 150, 18 },
		{ 368, 151, 18 },
		{ 368, 152, 18 },
		{ 368, 153, 18 },
		{ 368, 154, 18 },
		{ 368, 155, 18 },
		{ 368, 156, 18 },
		{ 368, 157, 18 },
		{ 368, 158, 18 },
		{ 368, 159, 18 },
		{ 368, 160, 18 },
		{ 368, 161, 18 },
		{ 368, 162, 18 },
		{ 368, 163, 18 },
		{ 368, 164, 18 },
		{ 368, 165, 18 },
		{ 368, 166, 18 },
		{ 368, 167, 18 },
		{ 368, 168, 18 },
		{ 368, 169, 18 },
		{ 368, 170, 18 },
		{ 368, 171, 18 },
		{ 368, 172, 18 },
		{ 368, 173, 18 },
		{ 369, 131, 18 },
		{ 369, 132, 18 },
		{ 369, 133, 18 },
		{ 369, 134, 18 },
		{ 369, 135, 18 },
		{ 369, 136, 18 },
		{ 369, 137, 18 },
		{ 369, 138, 18 },
		{ 369, 139, 18 },
		{ 369, 140, 18 },
		{ 369, 141, 18 },
		{ 369, 142, 18 },
		{ 369, 143, 18 },
		{ 369, 144, 18 },
		{ 369, 145, 18 },
		{ 369, 146, 18 },
		{ 369, 147, 18 },
		{ 369, 148, 18 },
		{ 369, 149, 18 },
		{ 369, 150, 18 },
		{ 369, 151, 18 },
		{ 369, 152, 18 },
		{ 369, 153, 18 },
		{ 369, 154, 18 },
		{ 369, 155, 18 },
		{ 369, 156, 18 },
		{ 369, 157, 18 },
		{ 369, 158, 18 },
		{ 369, 159, 18 },
		{ 369, 160, 18 },
		{ 369, 161, 18 },
		{ 369, 162, 18 },
		{ 369, 163, 18 },
		{ 369, 164, 18 },
		{ 369, 165, 18 },
		{ 369, 166, 18 },
		{ 369, 167, 18 },
		{ 369, 168, 18 },
		{ 369, 169, 18 },
		{ 369, 170, 18 },
		{ 369, 171, 18 },
		{ 369, 172, 18 },
		{ 370, 131, 18 },
		{ 370, 132, 18 },
		{ 370, 133, 18 },
		{ 370, 134, 18 },
		{ 370, 135, 18 },
		{ 370, 136, 18 },
		{ 370, 137, 18 },
		{ 370, 138, 18 },
		{ 370, 139, 18 },
		{ 370, 140, 18 },
		{ 370, 141, 18 },
		{ 370, 142, 18 },
		{ 370, 143, 18 },
		{ 370, 144, 18 },
		{ 370, 145, 18 },
		{ 370, 146, 18 },
		{ 370, 147, 18 },
		{ 370, 148, 18 },
		{ 370, 149, 18 },
		{ 370, 150, 18 },
		{ 370, 151, 18 },
		{ 370, 152, 18 },
		{ 370, 153, 18 },
		{ 370, 154, 18 },
		{ 370, 155, 18 },
		{ 370, 156, 18 },
		{ 370, 157, 18 },
		{ 370, 158, 18 },
		{ 370, 159, 18 },
		{ 370, 160, 18 },
		{ 370, 161, 18 },
		{ 370, 162, 18 },
		{ 370, 163, 18 },
		{ 370, 164, 18 },
		{ 370, 165, 18 },
		{ 370, 166, 18 },
		{ 370, 167, 18 },
		{ 370, 168, 18 },
		{ 370, 169, 18 },
		{ 370, 170, 18 },
		{ 370, 171, 18 },
		{ 370, 172, 18 },
		{ 371, 131, 18 },
		{ 371, 132, 18 },
		{ 371, 133, 18 },
		{ 371, 134, 18 },
		{ 371, 135, 18 },
		{ 371, 136, 18 },
		{ 371, 137, 18 },
		{ 371, 138, 18 },
		{ 371, 139, 18 },
		{ 371, 140, 18 },
		{ 371, 141, 18 },
		{ 371, 142, 18 },
		{ 371, 143, 18 },
		{ 371, 144, 18 },
		{ 371, 145, 18 },
		{ 371, 146, 18 },
		{ 371, 147, 18 },
		{ 371, 148, 18 },
		{ 371, 149, 18 },
		{ 371, 150, 18 },
		{ 371, 151, 18 },
		{ 371, 152, 18 },
		{ 371, 153, 18 },
		{ 371, 154, 18 },
		{ 371, 155, 18 },
		{ 371, 156, 18 },
		{ 371, 157, 18 },
		{ 371, 158, 18 },
		{ 371, 159, 18 },
		{ 371, 160, 18 },
		{ 371, 161, 18 },
		{ 371, 162, 18 },
		{ 371, 163, 18 },
		{ 371, 164, 18 },
		{ 371, 165, 18 },
		{ 371, 166, 18 },
		{ 371, 167, 18 },
		{ 371, 168, 18 },
		{ 371, 169, 18 },
		{ 371, 170, 18 },
		{ 371, 171, 18 },
		{ 372, 131, 18 },
		{ 372, 132, 18 },
		{ 372, 133, 18 },
		{ 372, 134, 18 },
		{ 372, 135, 18 },
		{ 372, 136, 18 },
		{ 372, 137, 18 },
		{ 372, 138, 18 },
		{ 372, 139, 18 },
		{ 372, 140, 18 },
		{ 372, 141, 18 },
		{ 372, 142, 18 },
		{ 372, 143, 18 },
		{ 372, 144, 18 },
		{ 372, 145, 18 },
		{ 372, 146, 18 },
		{ 372, 147, 18 },
		{ 372, 148, 18 },
		{ 372, 149, 18 },
		{ 372, 150, 18 },
		{ 372, 151, 18 },
		{ 372, 152, 18 },
		{ 372, 153, 18 },
		{ 372, 154, 18 },
		{ 372, 155, 18 },
		{ 372, 156, 18 },
		{ 372, 157, 18 },
		{ 372, 158, 18 },
		{ 372, 159, 18 },
		{ 372, 160, 18 },
		{ 372, 161, 18 },
		{ 372, 162, 18 },
		{ 372, 163, 18 },
		{ 372, 164, 18 },
		{ 372, 165, 18 },
		{ 372, 166, 18 },
		{ 372, 167, 18 },
		{ 372, 168, 18 },
		{ 372, 169, 18 },
		{ 372, 170, 18 },
		{ 373, 131, 18 },
		{ 373, 132, 18 },
		{ 373, 133, 18 },
		{ 373, 134, 18 },
		{ 373, 135, 18 },
		{ 373, 136, 18 },
		{ 373, 137, 18 },
		{ 373, 138, 18 },
		{ 373, 139, 18 },
		{ 373, 140, 18 },
		{ 373, 141, 18 },
		{ 373, 142, 18 },
		{ 373, 143, 18 },
		{ 373, 144, 18 },
		{ 373, 145, 18 },
		{ 373, 146, 18 },
		{ 373, 147, 18 },
		{ 373, 148, 18 },
		{ 373, 149, 18 },
		{ 373, 150, 18 },
		{ 373, 151, 18 },
		{ 373, 152, 18 },
		{ 373, 153, 18 },
		{ 373, 154, 18 },
		{ 373, 155, 18 },
		{ 373, 156, 18 },
		{ 373, 157, 18 },
		{ 373, 158, 18 },
		{ 373, 159, 18 },
		{ 373, 160, 18 },
		{ 373, 161, 18 },
		{ 373, 162, 18 },
		{ 373, 163, 18 },
		{ 373, 164, 18 },
		{ 373, 165, 18 },
		{ 373, 166, 18 },
		{ 373, 167, 18 },
		{ 373, 168, 18 },
		{ 373, 169, 18 },
		{ 374, 131, 18 },
		{ 374, 132, 18 },
		{ 374, 133, 18 },
		{ 374, 134, 18 },
		{ 374, 135, 18 },
		{ 374, 136, 18 },
		{ 374, 137, 18 },
		{ 374, 138, 18 },
		{ 374, 139, 18 },
		{ 374, 140, 18 },
		{ 374, 141, 18 },
		{ 374, 142, 18 },
		{ 374, 143, 18 },
		{ 374, 144, 18 },
		{ 374, 145, 18 },
		{ 374, 146, 18 },
		{ 374, 147, 18 },
		{ 374, 148, 18 },
		{ 374, 149, 18 },
		{ 374, 150, 18 },
		{ 374, 151, 18 },
		{ 374, 152, 18 },
		{ 374, 153, 18 },
		{ 374, 154, 18 },
		{ 374, 155, 18 },
		{ 374, 156, 18 },
		{ 374, 157, 18 },
		{ 374, 158, 18 },
		{ 374, 159, 18 },
		{ 374, 160, 18 },
		{ 374, 161, 18 },
		{ 374, 162, 18 },
		{ 374, 163, 18 },
		{ 374, 164, 18 },
		{ 374, 165, 18 },
		{ 374, 166, 18 },
		{ 374, 167, 18 },
		{ 374, 168, 18 },
		{ 374, 169, 18 },
		{ 375, 131, 18 },
		{ 375, 132, 18 },
		{ 375, 133, 18 },
		{ 375, 134, 18 },
		{ 375, 135, 18 },
		{ 375, 136, 18 },
		{ 375, 137, 18 },
		{ 375, 138, 18 },
		{ 375, 139, 18 },
		{ 375, 140, 18 },
		{ 375, 141, 18 },
		{ 375, 142, 18 },
		{ 375, 143, 18 },
		{ 375, 144, 18 },
		{ 375, 145, 18 },
		{ 375, 146, 18 },
		{ 375, 147, 18 },
		{ 375, 148, 18 },
		{ 375, 149, 18 },
		{ 375, 150, 18 },
		{ 375, 151, 18 },
		{ 375, 152, 18 },
		{ 375, 153, 18 },
		{ 375, 154, 18 },
		{ 375, 155, 18 },
		{ 375, 156, 18 },
		{ 375, 157, 18 },
		{ 375, 158, 18 },
		{ 375, 159, 18 },
		{ 375, 160, 18 },
		{ 375, 161, 18 },
		{ 375, 162, 18 },
		{ 375, 163, 18 },
		{ 375, 164, 18 },
		{ 375, 165, 18 },
		{ 375, 166, 18 },
		{ 375, 167, 18 },
		{ 375, 168, 18 },
		{ 376, 131, 18 },
		{ 376, 132, 18 },
		{ 376, 133, 18 },
		{ 376, 134, 18 },
		{ 376, 135, 18 },
		{ 376, 136, 18 },
		{ 376, 137, 18 },
		{ 376, 138, 18 },
		{ 376, 139, 18 },
		{ 376, 140, 18 },
		{ 376, 141, 18 },
		{ 376, 142, 18 },
		{ 376, 143, 18 },
		{ 376, 144, 18 },
		{ 376, 145, 18 },
		{ 376, 146, 18 },
		{ 376, 147, 18 },
		{ 376, 148, 18 },
		{ 376, 149, 18 },
		{ 376, 150, 18 },
		{ 376, 151, 18 },
		{ 376, 152, 18 },
		{ 376, 153, 18 },
		{ 376, 154, 18 },
		{ 376, 155, 18 },
		{ 376, 156, 18 },
		{ 376, 157, 18 },
		{ 376, 158, 18 },
		{ 376, 159, 18 },
		{ 376, 160, 18 },
		{ 376, 161, 18 },
		{ 376, 162, 18 },
		{ 376, 163, 18 },
		{ 376, 164, 18 },
		{ 376, 165, 18 },
		{ 376, 166, 18 },
		{ 376, 167, 18 },
		{ 377, 131, 18 },
		{ 377, 132, 18 },
		{ 377, 133, 18 },
		{ 377, 134, 18 },
		{ 377, 135, 18 },
		{ 377, 136, 18 },
		{ 377, 137, 18 },
		{ 377, 138, 18 },
		{ 377, 139, 18 },
		{ 377, 140, 18 },
		{ 377, 141, 18 },
		{ 377, 142, 18 },
		{ 377, 143, 18 },
		{ 377, 144, 18 },
		{ 377, 145, 18 },
		{ 377, 146, 18 },
		{ 377, 147, 18 },
		{ 377, 148, 18 },
		{ 377, 149, 18 },
		{ 377, 150, 18 },
		{ 377, 151, 18 },
		{ 377, 152, 18 },
		{ 377, 153, 18 },
		{ 377, 154, 18 },
		{ 377, 155, 18 },
		{ 377, 156, 18 },
		{ 377, 157, 18 },
		{ 377, 158, 18 },
		{ 377, 159, 18 },
		{ 377, 160, 18 },
		{ 377, 161, 18 },
		{ 377, 162, 18 },
		{ 377, 163, 18 },
		{ 377, 164, 18 },
		{ 377, 165, 18 },
		{ 377, 166, 18 },
		{ 378, 131, 18 },
		{ 378, 132, 18 },
		{ 378, 133, 18 },
		{ 378, 134, 18 },
		{ 378, 135, 18 },
		{ 378, 136, 18 },
		{ 378, 137, 18 },
		{ 378, 138, 18 },
		{ 378, 139, 18 },
		{ 378, 140, 18 },
		{ 378, 141, 18 },
		{ 378, 142, 18 },
		{ 378, 143, 18 },
		{ 378, 144, 18 },
		{ 378, 145, 18 },
		{ 378, 146, 18 },
		{ 378, 147, 18 },
		{ 378, 148, 18 },
		{ 378, 149, 18 },
		{ 378, 150, 18 },
		{ 378, 151, 18 },
		{ 378, 152, 18 },
		{ 378, 153, 18 },
		{ 378, 154, 18 },
		{ 378, 155, 18 },
		{ 378, 156, 18 },
		{ 378, 157, 18 },
		{ 378, 158, 18 },
		{ 378, 159, 18 },
		{ 378, 160, 18 },
		{ 378, 161, 18 },
		{ 378, 162, 18 },
		{ 378, 163, 18 },
		{ 378, 164, 18 },
		{ 378, 165, 18 },
		{ 379, 131, 18 },
		{ 379, 132, 18 },
		{ 379, 133, 18 },
		{ 379, 134, 18 },
		{ 379, 135, 18 },
		{ 379, 136, 18 },
		{ 379, 137, 18 },
		{ 379, 138, 18 },
		{ 379, 139, 18 },
		{ 379, 140, 18 },
		{ 379, 141, 18 },
		{ 379, 142, 18 },
		{ 379, 143, 18 },
		{ 379, 144, 18 },
		{ 379, 145, 18 },
		{ 379, 146, 18 },
		{ 379, 147, 18 },
		{ 379, 148, 18 },
		{ 379, 149, 18 },
		{ 379, 150, 18 },
		{ 379, 151, 18 },
		{ 379, 152, 18 },
		{ 379, 153, 18 },
		{ 379, 154, 18 },
		{ 379, 155, 18 },
		{ 379, 156, 18 },
		{ 379, 157, 18 },
		{ 379, 158, 18 },
		{ 379, 159, 18 },
		{ 379, 160, 18 },
		{ 379, 161, 18 },
		{ 379, 162, 18 },
		{ 379, 163, 18 },
		{ 379, 164, 18 },
		{ 380, 131, 18 },
		{ 380, 132, 18 },
		{ 380, 133, 18 },
		{ 380, 134, 18 },
		{ 380, 135, 18 },
		{ 380, 136, 18 },
		{ 380, 137, 18 },
		{ 380, 138, 18 },
		{ 380, 139, 18 },
		{ 380, 140, 18 },
		{ 380, 141, 18 },
		{ 380, 142, 18 },
		{ 380, 143, 18 },
		{ 380, 144, 18 },
		{ 380, 145, 18 },
		{ 380, 146, 18 },
		{ 380, 147, 18 },
		{ 380, 148, 18 },
		{ 380, 149, 18 },
		{ 380, 150, 18 },
		{ 380, 151, 18 },
		{ 380, 152, 18 },
		{ 380, 153, 18 },
		{ 380, 154, 18 },
		{ 380, 155, 18 },
		{ 380, 156, 18 },
		{ 380, 157, 18 },
		{ 380, 158, 18 },
		{ 380, 159, 18 },
		{ 380, 160, 18 },
		{ 380, 161, 18 },
		{ 380, 162, 18 },
		{ 380, 163, 18 },
		{ 381, 131, 18 },
		{ 381, 132, 18 },
		{ 381, 133, 18 },
		{ 381, 134, 18 },
		{ 381, 135, 18 },
		{ 381, 136, 18 },
		{ 381, 137, 18 },
		{ 381, 138, 18 },
		{ 381, 139, 18 },
		{ 381, 140, 18 },
		{ 381, 141, 18 },
		{ 381, 142, 18 },
		{ 381, 143, 18 },
		{ 381, 144, 18 },
		{ 381, 145, 18 },
		{ 381, 146, 18 },
		{ 381, 147, 18 },
		{ 381, 148, 18 },
		{ 381, 149, 18 },
		{ 381, 150, 18 },
		{ 381, 151, 18 },
		{ 381, 152, 18 },
		{ 381, 153, 18 },
		{ 381, 154, 18 },
		{ 381, 155, 18 },
		{ 381, 156, 18 },
		{ 381, 157, 18 },
		{ 381, 158, 18 },
		{ 381, 159, 18 },
		{ 381, 160, 18 },
		{ 381, 161, 18 },
		{ 381, 162, 18 },
		{ 382, 131, 18 },
		{ 382, 132, 18 },
		{ 382, 133, 18 },
		{ 382, 134, 18 },
		{ 382, 135, 18 },
		{ 382, 136, 18 },
		{ 382, 137, 18 },
		{ 382, 138, 18 },
		{ 382, 139, 18 },
		{ 382, 140, 18 },
		{ 382, 141, 18 },
		{ 382, 142, 18 },
		{ 382, 143, 18 },
		{ 382, 144, 18 },
		{ 382, 145, 18 },
		{ 382, 146, 18 },
		{ 382, 147, 18 },
		{ 382, 148, 18 },
		{ 382, 149, 18 },
		{ 382, 150, 18 },
		{ 382, 151, 18 },
		{ 382, 152, 18 },
		{ 382, 153, 18 },
		{ 382, 154, 18 },
		{ 382, 155, 18 },
		{ 382, 156, 18 },
		{ 382, 157, 18 },
		{ 382, 158, 18 },
		{ 382, 159, 18 },
		{ 382, 160, 18 },
		{ 383, 131, 18 },
		{ 383, 132, 18 },
		{ 383, 133, 18 },
		{ 383, 134, 18 },
		{ 383, 135, 18 },
		{ 383, 136, 18 },
		{ 383, 137, 18 },
		{ 383, 138, 18 },
		{ 383, 139, 18 },
		{ 383, 140, 18 },
		{ 383, 141, 18 },
		{ 383, 142, 18 },
		{ 383, 143, 18 },
		{ 383, 144, 18 },
		{ 383, 145, 18 },
		{ 383, 146, 18 },
		{ 383, 147, 18 },
		{ 383, 148, 18 },
		{ 383, 149, 18 },
		{ 383, 150, 18 },
		{ 383, 151, 18 },
		{ 383, 152, 18 },
		{ 383, 153, 18 },
		{ 383, 154, 18 },
		{ 383, 155, 18 },
		{ 131, 327, 19 },
		{ 131, 328, 19 },
		{ 131, 329, 19 },
		{ 131, 330, 19 },
		{ 131, 331, 19 },
		{ 131, 332, 19 },
		{ 131, 333, 19 },
		{ 131, 334, 19 },
		{ 131, 335, 19 },
		{ 131, 336, 19 },
		{ 131, 337, 19 },
		{ 131, 338, 19 },
		{ 131, 339, 19 },
		{ 131, 340, 19 },
		{ 131, 341, 19 },
		{ 131, 342, 19 },
		{ 131, 343, 19 },
		{ 131, 344, 19 },
		{ 131, 345, 19 },
		{ 131, 346, 19 },
		{ 131, 347, 19 },
		{ 131, 348, 19 },
		{ 131, 349, 19 },
		{ 131, 350, 19 },
		{ 131, 351, 19 },
		{ 131, 352, 19 },
		{ 131, 353, 19 },
		{ 131, 354, 19 },
		{ 131, 355, 19 },
		{ 131, 356, 19 },
		{ 131, 357, 19 },
		{ 131, 358, 19 },
		{ 131, 359, 19 },
		{ 131, 360, 19 },
		{ 131, 361, 19 },
		{ 131, 362, 19 },
		{ 131, 363, 19 },
		{ 131, 364, 19 },
		{ 131, 365, 19 },
		{ 131, 366, 19 },
		{ 131, 367, 19 },
		{ 131, 368, 19 },
		{ 131, 369, 19 },
		{ 131, 370, 19 },
		{ 131, 371, 19 },
		{ 131, 372, 19 },
		{ 131, 373, 19 },
		{ 131, 374, 19 },
		{ 131, 375, 19 },
		{ 131, 376, 19 },
		{ 132, 323, 19 },
		{ 132, 324, 19 },
		{ 132, 325, 19 },
		{ 132, 326, 19 },
		{ 132, 327, 19 },
		{ 132, 328, 19 },
		{ 132, 329, 19 },
		{ 132, 330, 19 },
		{ 132, 331, 19 },
		{ 132, 332, 19 },
		{ 132, 333, 19 },
		{ 132, 334, 19 },
		{ 132, 335, 19 },
		{ 132, 336, 19 },
		{ 132, 337, 19 },
		{ 132, 338, 19 },
		{ 132, 339, 19 },
		{ 132, 340, 19 },
		{ 132, 341, 19 },
		{ 132, 342, 19 },
		{ 132, 343, 19 },
		{ 132, 344, 19 },
		{ 132, 345, 19 },
		{ 132, 346, 19 },
		{ 132, 347, 19 },
		{ 132, 348, 19 },
		{ 132, 349, 19 },
		{ 132, 350, 19 },
		{ 132, 351, 19 },
		{ 132, 352, 19 },
		{ 132, 353, 19 },
		{ 132, 354, 19 },
		{ 132, 355, 19 },
		{ 132, 356, 19 },
		{ 132, 357, 19 },
		{ 132, 358, 19 },
		{ 132, 359, 19 },
		{ 132, 360, 19 },
		{ 132, 361, 19 },
		{ 132, 362, 19 },
		{ 132, 363, 19 },
		{ 132, 364, 19 },
		{ 132, 365, 19 },
		{ 132, 366, 19 },
		{ 132, 367, 19 },
		{ 132, 368, 19 },
		{ 132, 369, 19 },
		{ 132, 370, 19 },
		{ 132, 371, 19 },
		{ 132, 372, 19 },
		{ 132, 373, 19 },
		{ 132, 374, 19 },
		{ 132, 375, 19 },
		{ 132, 376, 19 },
		{ 132, 377, 19 },
		{ 132, 378, 19 },
		{ 132, 379, 19 },
		{ 133, 321, 19 },
		{ 133, 322, 19 },
		{ 133, 323, 19 },
		{ 133, 324, 19 },
		{ 133, 325, 19 },
		{ 133, 326, 19 },
		{ 133, 327, 19 },
		{ 133, 328, 19 },
		{ 133, 329, 19 },
		{ 133, 330, 19 },
		{ 133, 331, 19 },
		{ 133, 332, 19 },
		{ 133, 333, 19 },
		{ 133, 334, 19 },
		{ 133, 335, 19 },
		{ 133, 336, 19 },
		{ 133, 337, 19 },
		{ 133, 338, 19 },
		{ 133, 339, 19 },
		{ 133, 340, 19 },
		{ 133, 341, 19 },
		{ 133, 342, 19 },
		{ 133, 343, 19 },
		{ 133, 344, 19 },
		{ 133, 345, 19 },
		{ 133, 346, 19 },
		{ 133, 347, 19 },
		{ 133, 348, 19 },
		{ 133, 349, 19 },
		{ 133, 350, 19 },
		{ 133, 351, 19 },
		{ 133, 352, 19 },
		{ 133, 353, 19 },
		{ 133, 354, 19 },
		{ 133, 355, 19 },
		{ 133, 356, 19 },
		{ 133, 357, 19 },
		{ 133, 358, 19 },
		{ 133, 359, 19 },
		{ 133, 360, 19 },
		{ 133, 361, 19 },
		{ 133, 362, 19 },
		{ 133, 363, 19 },
		{ 133, 364, 19 },
		{ 133, 365, 19 },
		{ 133, 366, 19 },
		{ 133, 367, 19 },
		{ 133, 368, 19 },
		{ 133, 369, 19 },
		{ 133, 370, 19 },
		{ 133, 371, 19 },
		{ 133, 372, 19 },
		{ 133, 373, 19 },
		{ 133, 374, 19 },
		{ 133, 375, 19 },
		{ 133, 376, 19 },
		{ 133, 377, 19 },
		{ 133, 378, 19 },
		{ 133, 379, 19 },
		{ 134, 320, 19 },
		{ 134, 321, 19 },
		{ 134, 322, 19 },
		{ 134, 323, 19 },
		{ 134, 324, 19 },
		{ 134, 325, 19 },
		{ 134, 326, 19 },
		{ 134, 327, 19 },
		{ 134, 328, 19 },
		{ 134, 329, 19 },
		{ 134, 330, 19 },
		{ 134, 331, 19 },
		{ 134, 332, 19 },
		{ 134, 333, 19 },
		{ 134, 334, 19 },
		{ 134, 335, 19 },
		{ 134, 336, 19 },
		{ 134, 337, 19 },
		{ 134, 338, 19 },
		{ 134, 339, 19 },
		{ 134, 340, 19 },
		{ 134, 341, 19 },
		{ 134, 342, 19 },
		{ 134, 343, 19 },
		{ 134, 344, 19 },
		{ 134, 345, 19 },
		{ 134, 346, 19 },
		{ 134, 347, 19 },
		{ 134, 348, 19 },
		{ 134, 349, 19 },
		{ 134, 350, 19 },
		{ 134, 351, 19 },
		{ 134, 352, 19 },
		{ 134, 353, 19 },
		{ 134, 354, 19 },
		{ 134, 355, 19 },
		{ 134, 356, 19 },
		{ 134, 357, 19 },
		{ 134, 358, 19 },
		{ 134, 359, 19 },
		{ 134, 360, 19 },
		{ 134, 361, 19 },
		{ 134, 362, 19 },
		{ 134, 363, 19 },
		{ 134, 364, 19 },
		{ 134, 365, 19 },
		{ 134, 366, 19 },
		{ 134, 367, 19 },
		{ 134, 368, 19 },
		{ 134, 369, 19 },
		{ 134, 370, 19 },
		{ 134, 371, 19 },
		{ 134, 372, 19 },
		{ 134, 373, 19 },
		{ 134, 374, 19 },
		{ 134, 375, 19 },
		{ 134, 376, 19 },
		{ 134, 377, 19 },
		{ 134, 378, 19 },
		{ 134, 379, 19 },
		{ 135, 319, 19 },
		{ 135, 320, 19 },
		{ 135, 321, 19 },
		{ 135, 322, 19 },
		{ 135, 323, 19 },
		{ 135, 324, 19 },
		{ 135, 325, 19 },
		{ 135, 326, 19 },
		{ 135, 327, 19 },
		{ 135, 328, 19 },
		{ 135, 329, 19 },
		{ 135, 330, 19 },
		{ 135, 331, 19 },
		{ 135, 332, 19 },
		{ 135, 333, 19 },
		{ 135, 334, 19 },
		{ 135, 335, 19 },
		{ 135, 336, 19 },
		{ 135, 337, 19 },
		{ 135, 338, 19 },
		{ 135, 339, 19 },
		{ 135, 340, 19 },
		{ 135, 341, 19 },
		{ 135, 342, 19 },
		{ 135, 343, 19 },
		{ 135, 344, 19 },
		{ 135, 345, 19 },
		{ 135, 346, 19 },
		{ 135, 347, 19 },
		{ 135, 348, 19 },
		{ 135, 349, 19 },
		{ 135, 350, 19 },
		{ 135, 351, 19 },
		{ 135, 352, 19 },
		{ 135, 353, 19 },
		{ 135, 354, 19 },
		{ 135, 355, 19 },
		{ 135, 356, 19 },
		{ 135, 357, 19 },
		{ 135, 358, 19 },
		{ 135, 359, 19 },
		{ 135, 360, 19 },
		{ 135, 361, 19 },
		{ 135, 362, 19 },
		{ 135, 363, 19 },
		{ 135, 364, 19 },
		{ 135, 365, 19 },
		{ 135, 366, 19 },
		{ 135, 367, 19 },
		{ 135, 368, 19 },
		{ 135, 369, 19 },
		{ 135, 370, 19 },
		{ 135, 371, 19 },
		{ 135, 372, 19 },
		{ 135, 373, 19 },
		{ 135, 374, 19 },
		{ 135, 375, 19 },
		{ 135, 376, 19 },
		{ 135, 377, 19 },
		{ 135, 378, 19 },
		{ 136, 319, 19 },
		{ 136, 320, 19 },
		{ 136, 321, 19 },
		{ 136, 322, 19 },
		{ 136, 323, 19 },
		{ 136, 324, 19 },
		{ 136, 325, 19 },
		{ 136, 326, 19 },
		{ 136, 327, 19 },
		{ 136, 328, 19 },
		{ 136, 329, 19 },
		{ 136, 330, 19 },
		{ 136, 331, 19 },
		{ 136, 332, 19 },
		{ 136, 333, 19 },
		{ 136, 334, 19 },
		{ 136, 335, 19 },
		{ 136, 336, 19 },
		{ 136, 337, 19 },
		{ 136, 338, 19 },
		{ 136, 339, 19 },
		{ 136, 340, 19 },
		{ 136, 341, 19 },
		{ 136, 342, 19 },
		{ 136, 343, 19 },
		{ 136, 344, 19 },
		{ 136, 345, 19 },
		{ 136, 346, 19 },
		{ 136, 347, 19 },
		{ 136, 348, 19 },
		{ 136, 349, 19 },
		{ 136, 350, 19 },
		{ 136, 351, 19 },
		{ 136, 352, 19 },
		{ 136, 353, 19 },
		{ 136, 354, 19 },
		{ 136, 355, 19 },
		{ 136, 356, 19 },
		{ 136, 357, 19 },
		{ 136, 358, 19 },
		{ 136, 359, 19 },
		{ 136, 360, 19 },
		{ 136, 361, 19 },
		{ 136, 362, 19 },
		{ 136, 363, 19 },
		{ 136, 364, 19 },
		{ 136, 365, 19 },
		{ 136, 366, 19 },
		{ 136, 367, 19 },
		{ 136, 368, 19 },
		{ 136, 369, 19 },
		{ 136, 370, 19 },
		{ 136, 371, 19 },
		{ 136, 372, 19 },
		{ 136, 373, 19 },
		{ 136, 374, 19 },
		{ 136, 375, 19 },
		{ 136, 376, 19 },
		{ 136, 377, 19 },
		{ 137, 318, 19 },
		{ 137, 319, 19 },
		{ 137, 320, 19 },
		{ 137, 321, 19 },
		{ 137, 322, 19 },
		{ 137, 323, 19 },
		{ 137, 324, 19 },
		{ 137, 325, 19 },
		{ 137, 326, 19 },
		{ 137, 327, 19 },
		{ 137, 328, 19 },
		{ 137, 329, 19 },
		{ 137, 330, 19 },
		{ 137, 331, 19 },
		{ 137, 332, 19 },
		{ 137, 333, 19 },
		{ 137, 334, 19 },
		{ 137, 335, 19 },
		{ 137, 336, 19 },
		{ 137, 337, 19 },
		{ 137, 338, 19 },
		{ 137, 339, 19 },
		{ 137, 340, 19 },
		{ 137, 341, 19 },
		{ 137, 342, 19 },
		{ 137, 343, 19 },
		{ 137, 344, 19 },
		{ 137, 345, 19 },
		{ 137, 346, 19 },
		{ 137, 347, 19 },
		{ 137, 348, 19 },
		{ 137, 349, 19 },
		{ 137, 350, 19 },
		{ 137, 351, 19 },
		{ 137, 352, 19 },
		{ 137, 353, 19 },
		{ 137, 354, 19 },
		{ 137, 355, 19 },
		{ 137, 356, 19 },
		{ 137, 357, 19 },
		{ 137, 358, 19 },
		{ 137, 359, 19 },
		{ 137, 360, 19 },
		{ 137, 361, 19 },
		{ 137, 362, 19 },
		{ 137, 363, 19 },
		{ 137, 364, 19 },
		{ 137, 365, 19 },
		{ 137, 366, 19 },
		{ 137, 367, 19 },
		{ 137, 368, 19 },
		{ 137, 369, 19 },
		{ 137, 370, 19 },
		{ 137, 371, 19 },
		{ 137, 372, 19 },
		{ 137, 373, 19 },
		{ 137, 374, 19 },
		{ 137, 375, 19 },
		{ 137, 376, 19 },
		{ 138, 317, 19 },
		{ 138, 318, 19 },
		{ 138, 319, 19 },
		{ 138, 320, 19 },
		{ 138, 321, 19 },
		{ 138, 322, 19 },
		{ 138, 323, 19 },
		{ 138, 324, 19 },
		{ 138, 325, 19 },
		{ 138, 326, 19 },
		{ 138, 327, 19 },
		{ 138, 328, 19 },
		{ 138, 329, 19 },
		{ 138, 330, 19 },
		{ 138, 331, 19 },
		{ 138, 332, 19 },
		{ 138, 333, 19 },
		{ 138, 334, 19 },
		{ 138, 335, 19 },
		{ 138, 336, 19 },
		{ 138, 337, 19 },
		{ 138, 338, 19 },
		{ 138, 339, 19 },
		{ 138, 340, 19 },
		{ 138, 341, 19 },
		{ 138, 342, 19 },
		{ 138, 343, 19 },
		{ 138, 344, 19 },
		{ 138, 345, 19 },
		{ 138, 346, 19 },
		{ 138, 347, 19 },
		{ 138, 348, 19 },
		{ 138, 349, 19 },
		{ 138, 350, 19 },
		{ 138, 351, 19 },
		{ 138, 352, 19 },
		{ 138, 353, 19 },
		{ 138, 354, 19 },
		{ 138, 355, 19 },
		{ 138, 356, 19 },
		{ 138, 357, 19 },
		{ 138, 358, 19 },
		{ 138, 359, 19 },
		{ 138, 360, 19 },
		{ 138, 361, 19 },
		{ 138, 362, 19 },
		{ 138, 363, 19 },
		{ 138, 364, 19 },
		{ 138, 365, 19 },
		{ 138, 366, 19 },
		{ 138, 367, 19 },
		{ 138, 368, 19 },
		{ 138, 369, 19 },
		{ 138, 370, 19 },
		{ 138, 371, 19 },
		{ 138, 372, 19 },
		{ 138, 373, 19 },
		{ 138, 374, 19 },
		{ 138, 375, 19 },
		{ 139, 317, 19 },
		{ 139, 318, 19 },
		{ 139, 319, 19 },
		{ 139, 320, 19 },
		{ 139, 321, 19 },
		{ 139, 322, 19 },
		{ 139, 323, 19 },
		{ 139, 324, 19 },
		{ 139, 325, 19 },
		{ 139, 326, 19 },
		{ 139, 327, 19 },
		{ 139, 328, 19 },
		{ 139, 329, 19 },
		{ 139, 330, 19 },
		{ 139, 331, 19 },
		{ 139, 332, 19 },
		{ 139, 333, 19 },
		{ 139, 334, 19 },
		{ 139, 335, 19 },
		{ 139, 336, 19 },
		{ 139, 337, 19 },
		{ 139, 338, 19 },
		{ 139, 339, 19 },
		{ 139, 340, 19 },
		{ 139, 341, 19 },
		{ 139, 342, 19 },
		{ 139, 343, 19 },
		{ 139, 344, 19 },
		{ 139, 345, 19 },
		{ 139, 346, 19 },
		{ 139, 347, 19 },
		{ 139, 348, 19 },
		{ 139, 349, 19 },
		{ 139, 350, 19 },
		{ 139, 351, 19 },
		{ 139, 352, 19 },
		{ 139, 353, 19 },
		{ 139, 354, 19 },
		{ 139, 355, 19 },
		{ 139, 356, 19 },
		{ 139, 357, 19 },
		{ 139, 358, 19 },
		{ 139, 359, 19 },
		{ 139, 360, 19 },
		{ 139, 361, 19 },
		{ 139, 362, 19 },
		{ 139, 363, 19 },
		{ 139, 364, 19 },
		{ 139, 365, 19 },
		{ 139, 366, 19 },
		{ 139, 367, 19 },
		{ 139, 368, 19 },
		{ 139, 369, 19 },
		{ 139, 370, 19 },
		{ 139, 371, 19 },
		{ 139, 372, 19 },
		{ 139, 373, 19 },
		{ 139, 374, 19 },
		{ 140, 316, 19 },
		{ 140, 317, 19 },
		{ 140, 318, 19 },
		{ 140, 319, 19 },
		{ 140, 320, 19 },
		{ 140, 321, 19 },
		{ 140, 322, 19 },
		{ 140, 323, 19 },
		{ 140, 324, 19 },
		{ 140, 325, 19 },
		{ 140, 326, 19 },
		{ 140, 327, 19 },
		{ 140, 328, 19 },
		{ 140, 329, 19 },
		{ 140, 330, 19 },
		{ 140, 331, 19 },
		{ 140, 332, 19 },
		{ 140, 333, 19 },
		{ 140, 334, 19 },
		{ 140, 335, 19 },
		{ 140, 336, 19 },
		{ 140, 337, 19 },
		{ 140, 338, 19 },
		{ 140, 339, 19 },
		{ 140, 340, 19 },
		{ 140, 341, 19 },
		{ 140, 342, 19 },
		{ 140, 343, 19 },
		{ 140, 344, 19 },
		{ 140, 345, 19 },
		{ 140, 346, 19 },
		{ 140, 347, 19 },
		{ 140, 348, 19 },
		{ 140, 349, 19 },
		{ 140, 350, 19 },
		{ 140, 351, 19 },
		{ 140, 352, 19 },
		{ 140, 353, 19 },
		{ 140, 354, 19 },
		{ 140, 355, 19 },
		{ 140, 356, 19 },
		{ 140, 357, 19 },
		{ 140, 358, 19 },
		{ 140, 359, 19 },
		{ 140, 360, 19 },
		{ 140, 361, 19 },
		{ 140, 362, 19 },
		{ 140, 363, 19 },
		{ 140, 364, 19 },
		{ 140, 365, 19 },
		{ 140, 366, 19 },
		{ 140, 367, 19 },
		{ 140, 368, 19 },
		{ 140, 369, 19 },
		{ 140, 370, 19 },
		{ 140, 371, 19 },
		{ 140, 372, 19 },
		{ 140, 373, 19 },
		{ 141, 316, 19 },
		{ 141, 317, 19 },
		{ 141, 318, 19 },
		{ 141, 319, 19 },
		{ 141, 320, 19 },
		{ 141, 321, 19 },
		{ 141, 322, 19 },
		{ 141, 323, 19 },
		{ 141, 324, 19 },
		{ 141, 325, 19 },
		{ 141, 326, 19 },
		{ 141, 327, 19 },
		{ 141, 328, 19 },
		{ 141, 329, 19 },
		{ 141, 330, 19 },
		{ 141, 331, 19 },
		{ 141, 332, 19 },
		{ 141, 333, 19 },
		{ 141, 334, 19 },
		{ 141, 335, 19 },
		{ 141, 336, 19 },
		{ 141, 337, 19 },
		{ 141, 338, 19 },
		{ 141, 339, 19 },
		{ 141, 340, 19 },
		{ 141, 341, 19 },
		{ 141, 342, 19 },
		{ 141, 343, 19 },
		{ 141, 344, 19 },
		{ 141, 345, 19 },
		{ 141, 346, 19 },
		{ 141, 347, 19 },
		{ 141, 348, 19 },
		{ 141, 349, 19 },
		{ 141, 350, 19 },
		{ 141, 351, 19 },
		{ 141, 352, 19 },
		{ 141, 353, 19 },
		{ 141, 354, 19 },
		{ 141, 355, 19 },
		{ 141, 356, 19 },
		{ 141, 357, 19 },
		{ 141, 358, 19 },
		{ 141, 359, 19 },
		{ 141, 360, 19 },
		{ 141, 361, 19 },
		{ 141, 362, 19 },
		{ 141, 363, 19 },
		{ 141, 364, 19 },
		{ 141, 365, 19 },
		{ 141, 366, 19 },
		{ 141, 367, 19 },
		{ 141, 368, 19 },
		{ 141, 369, 19 },
		{ 141, 370, 19 },
		{ 141, 371, 19 },
		{ 141, 372, 19 },
		{ 142, 315, 19 },
		{ 142, 316, 19 },
		{ 142, 317, 19 },
		{ 142, 318, 19 },
		{ 142, 319, 19 },
		{ 142, 320, 19 },
		{ 142, 321, 19 },
		{ 142, 322, 19 },
		{ 142, 323, 19 },
		{ 142, 324, 19 },
		{ 142, 325, 19 },
		{ 142, 326, 19 },
		{ 142, 327, 19 },
		{ 142, 328, 19 },
		{ 142, 329, 19 },
		{ 142, 330, 19 },
		{ 142, 331, 19 },
		{ 142, 332, 19 },
		{ 142, 333, 19 },
		{ 142, 334, 19 },
		{ 142, 335, 19 },
		{ 142, 336, 19 },
		{ 142, 337, 19 },
		{ 142, 338, 19 },
		{ 142, 339, 19 },
		{ 142, 340, 19 },
		{ 142, 341, 19 },
		{ 142, 342, 19 },
		{ 142, 343, 19 },
		{ 142, 344, 19 },
		{ 142, 345, 19 },
		{ 142, 346, 19 },
		{ 142, 347, 19 },
		{ 142, 348, 19 },
		{ 142, 349, 19 },
		{ 142, 350, 19 },
		{ 142, 351, 19 },
		{ 142, 352, 19 },
		{ 142, 353, 19 },
		{ 142, 354, 19 },
		{ 142, 355, 19 },
		{ 142, 356, 19 },
		{ 142, 357, 19 },
		{ 142, 358, 19 },
		{ 142, 359, 19 },
		{ 142, 360, 19 },
		{ 142, 361, 19 },
		{ 142, 362, 19 },
		{ 142, 363, 19 },
		{ 142, 364, 19 },
		{ 142, 365, 19 },
		{ 142, 366, 19 },
		{ 142, 367, 19 },
		{ 142, 368, 19 },
		{ 142, 369, 19 },
		{ 142, 370, 19 },
		{ 142, 371, 19 },
		{ 143, 315, 19 },
		{ 143, 316, 19 },
		{ 143, 317, 19 },
		{ 143, 318, 19 },
		{ 143, 319, 19 },
		{ 143, 320, 19 },
		{ 143, 321, 19 },
		{ 143, 322, 19 },
		{ 143, 323, 19 },
		{ 143, 324, 19 },
		{ 143, 325, 19 },
		{ 143, 326, 19 },
		{ 143, 327, 19 },
		{ 143, 328, 19 },
		{ 143, 329, 19 },
		{ 143, 330, 19 },
		{ 143, 331, 19 },
		{ 143, 332, 19 },
		{ 143, 333, 19 },
		{ 143, 334, 19 },
		{ 143, 335, 19 },
		{ 143, 336, 19 },
		{ 143, 337, 19 },
		{ 143, 338, 19 },
		{ 143, 339, 19 },
		{ 143, 340, 19 },
		{ 143, 341, 19 },
		{ 143, 342, 19 },
		{ 143, 343, 19 },
		{ 143, 344, 19 },
		{ 143, 345, 19 },
		{ 143, 346, 19 },
		{ 143, 347, 19 },
		{ 143, 348, 19 },
		{ 143, 349, 19 },
		{ 143, 350, 19 },
		{ 143, 351, 19 },
		{ 143, 352, 19 },
		{ 143, 353, 19 },
		{ 143, 354, 19 },
		{ 143, 355, 19 },
		{ 143, 356, 19 },
		{ 143, 357, 19 },
		{ 143, 358, 19 },
		{ 143, 359, 19 },
		{ 143, 360, 19 },
		{ 143, 361, 19 },
		{ 143, 362, 19 },
		{ 143, 363, 19 },
		{ 143, 364, 19 },
		{ 143, 365, 19 },
		{ 143, 366, 19 },
		{ 143, 367, 19 },
		{ 143, 368, 19 },
		{ 143, 369, 19 },
		{ 143, 370, 19 },
		{ 144, 314, 19 },
		{ 144, 315, 19 },
		{ 144, 316, 19 },
		{ 144, 317, 19 },
		{ 144, 318, 19 },
		{ 144, 319, 19 },
		{ 144, 320, 19 },
		{ 144, 321, 19 },
		{ 144, 322, 19 },
		{ 144, 323, 19 },
		{ 144, 324, 19 },
		{ 144, 325, 19 },
		{ 144, 326, 19 },
		{ 144, 327, 19 },
		{ 144, 328, 19 },
		{ 144, 329, 19 },
		{ 144, 330, 19 },
		{ 144, 331, 19 },
		{ 144, 332, 19 },
		{ 144, 333, 19 },
		{ 144, 334, 19 },
		{ 144, 335, 19 },
		{ 144, 336, 19 },
		{ 144, 337, 19 },
		{ 144, 338, 19 },
		{ 144, 339, 19 },
		{ 144, 340, 19 },
		{ 144, 341, 19 },
		{ 144, 342, 19 },
		{ 144, 343, 19 },
		{ 144, 344, 19 },
		{ 144, 345, 19 },
		{ 144, 346, 19 },
		{ 144, 347, 19 },
		{ 144, 348, 19 },
		{ 144, 349, 19 },
		{ 144, 350, 19 },
		{ 144, 351, 19 },
		{ 144, 352, 19 },
		{ 144, 353, 19 },
		{ 144, 354, 19 },
		{ 144, 355, 19 },
		{ 144, 356, 19 },
		{ 144, 357, 19 },
		{ 144, 358, 19 },
		{ 144, 359, 19 },
		{ 144, 360, 19 },
		{ 144, 361, 19 },
		{ 144, 362, 19 },
		{ 144, 363, 19 },
		{ 144, 364, 19 },
		{ 144, 365, 19 },
		{ 144, 366, 19 },
		{ 144, 367, 19 },
		{ 144, 368, 19 },
		{ 144, 369, 19 },
		{ 145, 314, 19 },
		{ 145, 315, 19 },
		{ 145, 316, 19 },
		{ 145, 317, 19 },
		{ 145, 318, 19 },
		{ 145, 319, 19 },
		{ 145, 320, 19 },
		{ 145, 321, 19 },
		{ 145, 322, 19 },
		{ 145, 323, 19 },
		{ 145, 324, 19 },
		{ 145, 325, 19 },
		{ 145, 326, 19 },
		{ 145, 327, 19 },
		{ 145, 328, 19 },
		{ 145, 329, 19 },
		{ 145, 330, 19 },
		{ 145, 331, 19 },
		{ 145, 332, 19 },
		{ 145, 333, 19 },
		{ 145, 334, 19 },
		{ 145, 335, 19 },
		{ 145, 336, 19 },
		{ 145, 337, 19 },
		{ 145, 338, 19 },
		{ 145, 339, 19 },
		{ 145, 340, 19 },
		{ 145, 341, 19 },
		{ 145, 342, 19 },
		{ 145, 343, 19 },
		{ 145, 344, 19 },
		{ 145, 345, 19 },
		{ 145, 346, 19 },
		{ 145, 347, 19 },
		{ 145, 348, 19 },
		{ 145, 349, 19 },
		{ 145, 350, 19 },
		{ 145, 351, 19 },
		{ 145, 352, 19 },
		{ 145, 353, 19 },
		{ 145, 354, 19 },
		{ 145, 355, 19 },
		{ 145, 356, 19 },
		{ 145, 357, 19 },
		{ 145, 358, 19 },
		{ 145, 359, 19 },
		{ 145, 360, 19 },
		{ 145, 361, 19 },
		{ 145, 362, 19 },
		{ 145, 363, 19 },
		{ 145, 364, 19 },
		{ 145, 365, 19 },
		{ 145, 366, 19 },
		{ 145, 367, 19 },
		{ 145, 368, 19 },
		{ 146, 313, 19 },
		{ 146, 314, 19 },
		{ 146, 315, 19 },
		{ 146, 316, 19 },
		{ 146, 317, 19 },
		{ 146, 318, 19 },
		{ 146, 319, 19 },
		{ 146, 320, 19 },
		{ 146, 321, 19 },
		{ 146, 322, 19 },
		{ 146, 323, 19 },
		{ 146, 324, 19 },
		{ 146, 325, 19 },
		{ 146, 326, 19 },
		{ 146, 327, 19 },
		{ 146, 328, 19 },
		{ 146, 329, 19 },
		{ 146, 330, 19 },
		{ 146, 331, 19 },
		{ 146, 332, 19 },
		{ 146, 333, 19 },
		{ 146, 334, 19 },
		{ 146, 335, 19 },
		{ 146, 336, 19 },
		{ 146, 337, 19 },
		{ 146, 338, 19 },
		{ 146, 339, 19 },
		{ 146, 340, 19 },
		{ 146, 341, 19 },
		{ 146, 342, 19 },
		{ 146, 343, 19 },
		{ 146, 344, 19 },
		{ 146, 345, 19 },
		{ 146, 346, 19 },
		{ 146, 347, 19 },
		{ 146, 348, 19 },
		{ 146, 349, 19 },
		{ 146, 350, 19 },
		{ 146, 351, 19 },
		{ 146, 352, 19 },
		{ 146, 353, 19 },
		{ 146, 354, 19 },
		{ 146, 355, 19 },
		{ 146, 356, 19 },
		{ 146, 357, 19 },
		{ 146, 358, 19 },
		{ 146, 359, 19 },
		{ 146, 360, 19 },
		{ 146, 361, 19 },
		{ 146, 362, 19 },
		{ 146, 363, 19 },
		{ 146, 364, 19 },
		{ 146, 365, 19 },
		{ 146, 366, 19 },
		{ 146, 367, 19 },
		{ 147, 313, 19 },
		{ 147, 314, 19 },
		{ 147, 315, 19 },
		{ 147, 316, 19 },
		{ 147, 317, 19 },
		{ 147, 318, 19 },
		{ 147, 319, 19 },
		{ 147, 320, 19 },
		{ 147, 321, 19 },
		{ 147, 322, 19 },
		{ 147, 323, 19 },
		{ 147, 324, 19 },
		{ 147, 325, 19 },
		{ 147, 326, 19 },
		{ 147, 327, 19 },
		{ 147, 328, 19 },
		{ 147, 329, 19 },
		{ 147, 330, 19 },
		{ 147, 331, 19 },
		{ 147, 332, 19 },
		{ 147, 333, 19 },
		{ 147, 334, 19 },
		{ 147, 335, 19 },
		{ 147, 336, 19 },
		{ 147, 337, 19 },
		{ 147, 338, 19 },
		{ 147, 339, 19 },
		{ 147, 340, 19 },
		{ 147, 341, 19 },
		{ 147, 342, 19 },
		{ 147, 343, 19 },
		{ 147, 344, 19 },
		{ 147, 345, 19 },
		{ 147, 346, 19 },
		{ 147, 347, 19 },
		{ 147, 348, 19 },
		{ 147, 349, 19 },
		{ 147, 350, 19 },
		{ 147, 351, 19 },
		{ 147, 352, 19 },
		{ 147, 353, 19 },
		{ 147, 354, 19 },
		{ 147, 355, 19 },
		{ 147, 356, 19 },
		{ 147, 357, 19 },
		{ 147, 358, 19 },
		{ 147, 359, 19 },
		{ 147, 360, 19 },
		{ 147, 361, 19 },
		{ 147, 362, 19 },
		{ 147, 363, 19 },
		{ 147, 364, 19 },
		{ 147, 365, 19 },
		{ 147, 366, 19 },
		{ 148, 312, 19 },
		{ 148, 313, 19 },
		{ 148, 314, 19 },
		{ 148, 315, 19 },
		{ 148, 316, 19 },
		{ 148, 317, 19 },
		{ 148, 318, 19 },
		{ 148, 319, 19 },
		{ 148, 320, 19 },
		{ 148, 321, 19 },
		{ 148, 322, 19 },
		{ 148, 323, 19 },
		{ 148, 324, 19 },
		{ 148, 325, 19 },
		{ 148, 326, 19 },
		{ 148, 327, 19 },
		{ 148, 328, 19 },
		{ 148, 329, 19 },
		{ 148, 330, 19 },
		{ 148, 331, 19 },
		{ 148, 332, 19 },
		{ 148, 333, 19 },
		{ 148, 334, 19 },
		{ 148, 335, 19 },
		{ 148, 336, 19 },
		{ 148, 337, 19 },
		{ 148, 338, 19 },
		{ 148, 339, 19 },
		{ 148, 340, 19 },
		{ 148, 341, 19 },
		{ 148, 342, 19 },
		{ 148, 343, 19 },
		{ 148, 344, 19 },
		{ 148, 345, 19 },
		{ 148, 346, 19 },
		{ 148, 347, 19 },
		{ 148, 348, 19 },
		{ 148, 349, 19 },
		{ 148, 350, 19 },
		{ 148, 351, 19 },
		{ 148, 352, 19 },
		{ 148, 353, 19 },
		{ 148, 354, 19 },
		{ 148, 355, 19 },
		{ 148, 356, 19 },
		{ 148, 357, 19 },
		{ 148, 358, 19 },
		{ 148, 359, 19 },
		{ 148, 360, 19 },
		{ 148, 361, 19 },
		{ 148, 362, 19 },
		{ 148, 363, 19 },
		{ 148, 364, 19 },
		{ 148, 365, 19 },
		{ 149, 312, 19 },
		{ 149, 313, 19 },
		{ 149, 314, 19 },
		{ 149, 315, 19 },
		{ 149, 316, 19 },
		{ 149, 317, 19 },
		{ 149, 318, 19 },
		{ 149, 319, 19 },
		{ 149, 320, 19 },
		{ 149, 321, 19 },
		{ 149, 322, 19 },
		{ 149, 323, 19 },
		{ 149, 324, 19 },
		{ 149, 325, 19 },
		{ 149, 326, 19 },
		{ 149, 327, 19 },
		{ 149, 328, 19 },
		{ 149, 329, 19 },
		{ 149, 330, 19 },
		{ 149, 331, 19 },
		{ 149, 332, 19 },
		{ 149, 333, 19 },
		{ 149, 334, 19 },
		{ 149, 335, 19 },
		{ 149, 336, 19 },
		{ 149, 337, 19 },
		{ 149, 338, 19 },
		{ 149, 339, 19 },
		{ 149, 340, 19 },
		{ 149, 341, 19 },
		{ 149, 342, 19 },
		{ 149, 343, 19 },
		{ 149, 344, 19 },
		{ 149, 345, 19 },
		{ 149, 346, 19 },
		{ 149, 347, 19 },
		{ 149, 348, 19 },
		{ 149, 349, 19 },
		{ 149, 350, 19 },
		{ 149, 351, 19 },
		{ 149, 352, 19 },
		{ 149, 353, 19 },
		{ 149, 354, 19 },
		{ 149, 355, 19 },
		{ 149, 356, 19 },
		{ 149, 357, 19 },
		{ 149, 358, 19 },
		{ 149, 359, 19 },
		{ 149, 360, 19 },
		{ 149, 361, 19 },
		{ 149, 362, 19 },
		{ 149, 363, 19 },
		{ 149, 364, 19 },
		{ 150, 311, 19 },
		{ 150, 312, 19 },
		{ 150, 313, 19 },
		{ 150, 314, 19 },
		{ 150, 315, 19 },
		{ 150, 316, 19 },
		{ 150, 317, 19 },
		{ 150, 318, 19 },
		{ 150, 319, 19 },
		{ 150, 320, 19 },
		{ 150, 321, 19 },
		{ 150, 322, 19 },
		{ 150, 323, 19 },
		{ 150, 324, 19 },
		{ 150, 325, 19 },
		{ 150, 326, 19 },
		{ 150, 327, 19 },
		{ 150, 328, 19 },
		{ 150, 329, 19 },
		{ 150, 330, 19 },
		{ 150, 331, 19 },
		{ 150, 332, 19 },
		{ 150, 333, 19 },
		{ 150, 334, 19 },
		{ 150, 335, 19 },
		{ 150, 336, 19 },
		{ 150, 337, 19 },
		{ 150, 338, 19 },
		{ 150, 339, 19 },
		{ 150, 340, 19 },
		{ 150, 341, 19 },
		{ 150, 342, 19 },
		{ 150, 343, 19 },
		{ 150, 344, 19 },
		{ 150, 345, 19 },
		{ 150, 346, 19 },
		{ 150, 347, 19 },
		{ 150, 348, 19 },
		{ 150, 349, 19 },
		{ 150, 350, 19 },
		{ 150, 351, 19 },
		{ 150, 352, 19 },
		{ 150, 353, 19 },
		{ 150, 354, 19 },
		{ 150, 355, 19 },
		{ 150, 356, 19 },
		{ 150, 357, 19 },
		{ 150, 358, 19 },
		{ 150, 359, 19 },
		{ 150, 360, 19 },
		{ 150, 361, 19 },
		{ 150, 362, 19 },
		{ 150, 363, 19 },
		{ 151, 311, 19 },
		{ 151, 312, 19 },
		{ 151, 313, 19 },
		{ 151, 314, 19 },
		{ 151, 315, 19 },
		{ 151, 316, 19 },
		{ 151, 317, 19 },
		{ 151, 318, 19 },
		{ 151, 319, 19 },
		{ 151, 320, 19 },
		{ 151, 321, 19 },
		{ 151, 322, 19 },
		{ 151, 323, 19 },
		{ 151, 324, 19 },
		{ 151, 325, 19 },
		{ 151, 326, 19 },
		{ 151, 327, 19 },
		{ 151, 328, 19 },
		{ 151, 329, 19 },
		{ 151, 330, 19 },
		{ 151, 331, 19 },
		{ 151, 332, 19 },
		{ 151, 333, 19 },
		{ 151, 334, 19 },
		{ 151, 335, 19 },
		{ 151, 336, 19 },
		{ 151, 337, 19 },
		{ 151, 338, 19 },
		{ 151, 339, 19 },
		{ 151, 340, 19 },
		{ 151, 341, 19 },
		{ 151, 342, 19 },
		{ 151, 343, 19 },
		{ 151, 344, 19 },
		{ 151, 345, 19 },
		{ 151, 346, 19 },
		{ 151, 347, 19 },
		{ 151, 348, 19 },
		{ 151, 349, 19 },
		{ 151, 350, 19 },
		{ 151, 351, 19 },
		{ 151, 352, 19 },
		{ 151, 353, 19 },
		{ 151, 354, 19 },
		{ 151, 355, 19 },
		{ 151, 356, 19 },
		{ 151, 357, 19 },
		{ 151, 358, 19 },
		{ 151, 359, 19 },
		{ 151, 360, 19 },
		{ 151, 361, 19 },
		{ 151, 362, 19 },
		{ 152, 310, 19 },
		{ 152, 311, 19 },
		{ 152, 312, 19 },
		{ 152, 313, 19 },
		{ 152, 314, 19 },
		{ 152, 315, 19 },
		{ 152, 316, 19 },
		{ 152, 317, 19 },
		{ 152, 318, 19 },
		{ 152, 319, 19 },
		{ 152, 320, 19 },
		{ 152, 321, 19 },
		{ 152, 322, 19 },
		{ 152, 323, 19 },
		{ 152, 324, 19 },
		{ 152, 325, 19 },
		{ 152, 326, 19 },
		{ 152, 327, 19 },
		{ 152, 328, 19 },
		{ 152, 329, 19 },
		{ 152, 330, 19 },
		{ 152, 331, 19 },
		{ 152, 332, 19 },
		{ 152, 333, 19 },
		{ 152, 334, 19 },
		{ 152, 335, 19 },
		{ 152, 336, 19 },
		{ 152, 337, 19 },
		{ 152, 338, 19 },
		{ 152, 339, 19 },
		{ 152, 340, 19 },
		{ 152, 341, 19 },
		{ 152, 342, 19 },
		{ 152, 343, 19 },
		{ 152, 344, 19 },
		{ 152, 345, 19 },
		{ 152, 346, 19 },
		{ 152, 347, 19 },
		{ 152, 348, 19 },
		{ 152, 349, 19 },
		{ 152, 350, 19 },
		{ 152, 351, 19 },
		{ 152, 352, 19 },
		{ 152, 353, 19 },
		{ 152, 354, 19 },
		{ 152, 355, 19 },
		{ 152, 356, 19 },
		{ 152, 357, 19 },
		{ 152, 358, 19 },
		{ 152, 359, 19 },
		{ 152, 360, 19 },
		{ 152, 361, 19 },
		{ 153, 310, 19 },
		{ 153, 311, 19 },
		{ 153, 312, 19 },
		{ 153, 313, 19 },
		{ 153, 314, 19 },
		{ 153, 315, 19 },
		{ 153, 316, 19 },
		{ 153, 317, 19 },
		{ 153, 318, 19 },
		{ 153, 319, 19 },
		{ 153, 320, 19 },
		{ 153, 321, 19 },
		{ 153, 322, 19 },
		{ 153, 323, 19 },
		{ 153, 324, 19 },
		{ 153, 325, 19 },
		{ 153, 326, 19 },
		{ 153, 327, 19 },
		{ 153, 328, 19 },
		{ 153, 329, 19 },
		{ 153, 330, 19 },
		{ 153, 331, 19 },
		{ 153, 332, 19 },
		{ 153, 333, 19 },
		{ 153, 334, 19 },
		{ 153, 335, 19 },
		{ 153, 336, 19 },
		{ 153, 337, 19 },
		{ 153, 338, 19 },
		{ 153, 339, 19 },
		{ 153, 340, 19 },
		{ 153, 341, 19 },
		{ 153, 342, 19 },
		{ 153, 343, 19 },
		{ 153, 344, 19 },
		{ 153, 345, 19 },
		{ 153, 346, 19 },
		{ 153, 347, 19 },
		{ 153, 348, 19 },
		{ 153, 349, 19 },
		{ 153, 350, 19 },
		{ 153, 351, 19 },
		{ 153, 352, 19 },
		{ 153, 353, 19 },
		{ 153, 354, 19 },
		{ 153, 355, 19 },
		{ 153, 356, 19 },
		{ 153, 357, 19 },
		{ 153, 358, 19 },
		{ 153, 359, 19 },
		{ 153, 360, 19 },
		{ 154, 309, 19 },
		{ 154, 310, 19 },
		{ 154, 311, 19 },
		{ 154, 312, 19 },
		{ 154, 313, 19 },
		{ 154, 314, 19 },
		{ 154, 315, 19 },
		{ 154, 316, 19 },
		{ 154, 317, 19 },
		{ 154, 318, 19 },
		{ 154, 319, 19 },
		{ 154, 320, 19 },
		{ 154, 321, 19 },
		{ 154, 322, 19 },
		{ 154, 323, 19 },
		{ 154, 324, 19 },
		{ 154, 325, 19 },
		{ 154, 326, 19 },
		{ 154, 327, 19 },
		{ 154, 328, 19 },
		{ 154, 329, 19 },
		{ 154, 330, 19 },
		{ 154, 331, 19 },
		{ 154, 332, 19 },
		{ 154, 333, 19 },
		{ 154, 334, 19 },
		{ 154, 335, 19 },
		{ 154, 336, 19 },
		{ 154, 337, 19 },
		{ 154, 338, 19 },
		{ 154, 339, 19 },
		{ 154, 340, 19 },
		{ 154, 341, 19 },
		{ 154, 342, 19 },
		{ 154, 343, 19 },
		{ 154, 344, 19 },
		{ 154, 345, 19 },
		{ 154, 346, 19 },
		{ 154, 347, 19 },
		{ 154, 348, 19 },
		{ 154, 349, 19 },
		{ 154, 350, 19 },
		{ 154, 351, 19 },
		{ 154, 352, 19 },
		{ 154, 353, 19 },
		{ 154, 354, 19 },
		{ 154, 355, 19 },
		{ 154, 356, 19 },
		{ 154, 357, 19 },
		{ 154, 358, 19 },
		{ 154, 359, 19 },
		{ 155, 309, 19 },
		{ 155, 310, 19 },
		{ 155, 311, 19 },
		{ 155, 312, 19 },
		{ 155, 313, 19 },
		{ 155, 314, 19 },
		{ 155, 315, 19 },
		{ 155, 316, 19 },
		{ 155, 317, 19 },
		{ 155, 318, 19 },
		{ 155, 319, 19 },
		{ 155, 320, 19 },
		{ 155, 321, 19 },
		{ 155, 322, 19 },
		{ 155, 323, 19 },
		{ 155, 324, 19 },
		{ 155, 325, 19 },
		{ 155, 326, 19 },
		{ 155, 327, 19 },
		{ 155, 328, 19 },
		{ 155, 329, 19 },
		{ 155, 330, 19 },
		{ 155, 331, 19 },
		{ 155, 332, 19 },
		{ 155, 333, 19 },
		{ 155, 334, 19 },
		{ 155, 335, 19 },
		{ 155, 336, 19 },
		{ 155, 337, 19 },
		{ 155, 338, 19 },
		{ 155, 339, 19 },
		{ 155, 340, 19 },
		{ 155, 341, 19 },
		{ 155, 342, 19 },
		{ 155, 343, 19 },
		{ 155, 344, 19 },
		{ 155, 345, 19 },
		{ 155, 346, 19 },
		{ 155, 347, 19 },
		{ 155, 348, 19 },
		{ 155, 349, 19 },
		{ 155, 350, 19 },
		{ 155, 351, 19 },
		{ 155, 352, 19 },
		{ 155, 353, 19 },
		{ 155, 354, 19 },
		{ 155, 355, 19 },
		{ 155, 356, 19 },
		{ 155, 357, 19 },
		{ 155, 358, 19 },
		{ 156, 308, 19 },
		{ 156, 309, 19 },
		{ 156, 310, 19 },
		{ 156, 311, 19 },
		{ 156, 312, 19 },
		{ 156, 313, 19 },
		{ 156, 314, 19 },
		{ 156, 315, 19 },
		{ 156, 316, 19 },
		{ 156, 317, 19 },
		{ 156, 318, 19 },
		{ 156, 319, 19 },
		{ 156, 320, 19 },
		{ 156, 321, 19 },
		{ 156, 322, 19 },
		{ 156, 323, 19 },
		{ 156, 324, 19 },
		{ 156, 325, 19 },
		{ 156, 326, 19 },
		{ 156, 327, 19 },
		{ 156, 328, 19 },
		{ 156, 329, 19 },
		{ 156, 330, 19 },
		{ 156, 331, 19 },
		{ 156, 332, 19 },
		{ 156, 333, 19 },
		{ 156, 334, 19 },
		{ 156, 335, 19 },
		{ 156, 336, 19 },
		{ 156, 337, 19 },
		{ 156, 338, 19 },
		{ 156, 339, 19 },
		{ 156, 340, 19 },
		{ 156, 341, 19 },
		{ 156, 342, 19 },
		{ 156, 343, 19 },
		{ 156, 344, 19 },
		{ 156, 345, 19 },
		{ 156, 346, 19 },
		{ 156, 347, 19 },
		{ 156, 348, 19 },
		{ 156, 349, 19 },
		{ 156, 350, 19 },
		{ 156, 351, 19 },
		{ 156, 352, 19 },
		{ 156, 353, 19 },
		{ 156, 354, 19 },
		{ 156, 355, 19 },
		{ 156, 356, 19 },
		{ 156, 357, 19 },
		{ 157, 308, 19 },
		{ 157, 309, 19 },
		{ 157, 310, 19 },
		{ 157, 311, 19 },
		{ 157, 312, 19 },
		{ 157, 313, 19 },
		{ 157, 314, 19 },
		{ 157, 315, 19 },
		{ 157, 316, 19 },
		{ 157, 317, 19 },
		{ 157, 318, 19 },
		{ 157, 319, 19 },
		{ 157, 320, 19 },
		{ 157, 321, 19 },
		{ 157, 322, 19 },
		{ 157, 323, 19 },
		{ 157, 324, 19 },
		{ 157, 325, 19 },
		{ 157, 326, 19 },
		{ 157, 327, 19 },
		{ 157, 328, 19 },
		{ 157, 329, 19 },
		{ 157, 330, 19 },
		{ 157, 331, 19 },
		{ 157, 332, 19 },
		{ 157, 333, 19 },
		{ 157, 334, 19 },
		{ 157, 335, 19 },
		{ 157, 336, 19 },
		{ 157, 337, 19 },
		{ 157, 338, 19 },
		{ 157, 339, 19 },
		{ 157, 340, 19 },
		{ 157, 341, 19 },
		{ 157, 342, 19 },
		{ 157, 343, 19 },
		{ 157, 344, 19 },
		{ 157, 345, 19 },
		{ 157, 346, 19 },
		{ 157, 347, 19 },
		{ 157, 348, 19 },
		{ 157, 349, 19 },
		{ 157, 350, 19 },
		{ 157, 351, 19 },
		{ 157, 352, 19 },
		{ 157, 353, 19 },
		{ 157, 354, 19 },
		{ 157, 355, 19 },
		{ 157, 356, 19 },
		{ 158, 307, 19 },
		{ 158, 308, 19 },
		{ 158, 309, 19 },
		{ 158, 310, 19 },
		{ 158, 311, 19 },
		{ 158, 312, 19 },
		{ 158, 313, 19 },
		{ 158, 314, 19 },
		{ 158, 315, 19 },
		{ 158, 316, 19 },
		{ 158, 317, 19 },
		{ 158, 318, 19 },
		{ 158, 319, 19 },
		{ 158, 320, 19 },
		{ 158, 321, 19 },
		{ 158, 322, 19 },
		{ 158, 323, 19 },
		{ 158, 324, 19 },
		{ 158, 325, 19 },
		{ 158, 326, 19 },
		{ 158, 327, 19 },
		{ 158, 328, 19 },
		{ 158, 329, 19 },
		{ 158, 330, 19 },
		{ 158, 331, 19 },
		{ 158, 332, 19 },
		{ 158, 333, 19 },
		{ 158, 334, 19 },
		{ 158, 335, 19 },
		{ 158, 336, 19 },
		{ 158, 337, 19 },
		{ 158, 338, 19 },
		{ 158, 339, 19 },
		{ 158, 340, 19 },
		{ 158, 341, 19 },
		{ 158, 342, 19 },
		{ 158, 343, 19 },
		{ 158, 344, 19 },
		{ 158, 345, 19 },
		{ 158, 346, 19 },
		{ 158, 347, 19 },
		{ 158, 348, 19 },
		{ 158, 349, 19 },
		{ 158, 350, 19 },
		{ 158, 351, 19 },
		{ 158, 352, 19 },
		{ 158, 353, 19 },
		{ 158, 354, 19 },
		{ 158, 355, 19 },
		{ 159, 307, 19 },
		{ 159, 308, 19 },
		{ 159, 309, 19 },
		{ 159, 310, 19 },
		{ 159, 311, 19 },
		{ 159, 312, 19 },
		{ 159, 313, 19 },
		{ 159, 314, 19 },
		{ 159, 315, 19 },
		{ 159, 316, 19 },
		{ 159, 317, 19 },
		{ 159, 318, 19 },
		{ 159, 319, 19 },
		{ 159, 320, 19 },
		{ 159, 321, 19 },
		{ 159, 322, 19 },
		{ 159, 323, 19 },
		{ 159, 324, 19 },
		{ 159, 325, 19 },
		{ 159, 326, 19 },
		{ 159, 327, 19 },
		{ 159, 328, 19 },
		{ 159, 329, 19 },
		{ 159, 330, 19 },
		{ 159, 331, 19 },
		{ 159, 332, 19 },
		{ 159, 333, 19 },
		{ 159, 334, 19 },
		{ 159, 335, 19 },
		{ 159, 336, 19 },
		{ 159, 337, 19 },
		{ 159, 338, 19 },
		{ 159, 339, 19 },
		{ 159, 340, 19 },
		{ 159, 341, 19 },
		{ 159, 342, 19 },
		{ 159, 343, 19 },
		{ 159, 344, 19 },
		{ 159, 345, 19 },
		{ 159, 346, 19 },
		{ 159, 347, 19 },
		{ 159, 348, 19 },
		{ 159, 349, 19 },
		{ 159, 350, 19 },
		{ 159, 351, 19 },
		{ 159, 352, 19 },
		{ 159, 353, 19 },
		{ 159, 354, 19 },
		{ 160, 306, 19 },
		{ 160, 307, 19 },
		{ 160, 308, 19 },
		{ 160, 309, 19 },
		{ 160, 310, 19 },
		{ 160, 311, 19 },
		{ 160, 312, 19 },
		{ 160, 313, 19 },
		{ 160, 314, 19 },
		{ 160, 315, 19 },
		{ 160, 316, 19 },
		{ 160, 317, 19 },
		{ 160, 318, 19 },
		{ 160, 319, 19 },
		{ 160, 320, 19 },
		{ 160, 321, 19 },
		{ 160, 322, 19 },
		{ 160, 323, 19 },
		{ 160, 324, 19 },
		{ 160, 325, 19 },
		{ 160, 326, 19 },
		{ 160, 327, 19 },
		{ 160, 328, 19 },
		{ 160, 329, 19 },
		{ 160, 330, 19 },
		{ 160, 331, 19 },
		{ 160, 332, 19 },
		{ 160, 333, 19 },
		{ 160, 334, 19 },
		{ 160, 335, 19 },
		{ 160, 336, 19 },
		{ 160, 337, 19 },
		{ 160, 338, 19 },
		{ 160, 339, 19 },
		{ 160, 340, 19 },
		{ 160, 341, 19 },
		{ 160, 342, 19 },
		{ 160, 343, 19 },
		{ 160, 344, 19 },
		{ 160, 345, 19 },
		{ 160, 346, 19 },
		{ 160, 347, 19 },
		{ 160, 348, 19 },
		{ 160, 349, 19 },
		{ 160, 350, 19 },
		{ 160, 351, 19 },
		{ 160, 352, 19 },
		{ 160, 353, 19 },
		{ 161, 306, 19 },
		{ 161, 307, 19 },
		{ 161, 308, 19 },
		{ 161, 309, 19 },
		{ 161, 310, 19 },
		{ 161, 311, 19 },
		{ 161, 312, 19 },
		{ 161, 313, 19 },
		{ 161, 314, 19 },
		{ 161, 315, 19 },
		{ 161, 316, 19 },
		{ 161, 317, 19 },
		{ 161, 318, 19 },
		{ 161, 319, 19 },
		{ 161, 320, 19 },
		{ 161, 321, 19 },
		{ 161, 322, 19 },
		{ 161, 323, 19 },
		{ 161, 324, 19 },
		{ 161, 325, 19 },
		{ 161, 326, 19 },
		{ 161, 327, 19 },
		{ 161, 328, 19 },
		{ 161, 329, 19 },
		{ 161, 330, 19 },
		{ 161, 331, 19 },
		{ 161, 332, 19 },
		{ 161, 333, 19 },
		{ 161, 334, 19 },
		{ 161, 335, 19 },
		{ 161, 336, 19 },
		{ 161, 337, 19 },
		{ 161, 338, 19 },
		{ 161, 339, 19 },
		{ 161, 340, 19 },
		{ 161, 341, 19 },
		{ 161, 342, 19 },
		{ 161, 343, 19 },
		{ 161, 344, 19 },
		{ 161, 345, 19 },
		{ 161, 346, 19 },
		{ 161, 347, 19 },
		{ 161, 348, 19 },
		{ 161, 349, 19 },
		{ 161, 350, 19 },
		{ 161, 351, 19 },
		{ 161, 352, 19 },
		{ 162, 305, 19 },
		{ 162, 306, 19 },
		{ 162, 307, 19 },
		{ 162, 308, 19 },
		{ 162, 309, 19 },
		{ 162, 310, 19 },
		{ 162, 311, 19 },
		{ 162, 312, 19 },
		{ 162, 313, 19 },
		{ 162, 314, 19 },
		{ 162, 315, 19 },
		{ 162, 316, 19 },
		{ 162, 317, 19 },
		{ 162, 318, 19 },
		{ 162, 319, 19 },
		{ 162, 320, 19 },
		{ 162, 321, 19 },
		{ 162, 322, 19 },
		{ 162, 323, 19 },
		{ 162, 324, 19 },
		{ 162, 325, 19 },
		{ 162, 326, 19 },
		{ 162, 327, 19 },
		{ 162, 328, 19 },
		{ 162, 329, 19 },
		{ 162, 330, 19 },
		{ 162, 331, 19 },
		{ 162, 332, 19 },
		{ 162, 333, 19 },
		{ 162, 334, 19 },
		{ 162, 335, 19 },
		{ 162, 336, 19 },
		{ 162, 337, 19 },
		{ 162, 338, 19 },
		{ 162, 339, 19 },
		{ 162, 340, 19 },
		{ 162, 341, 19 },
		{ 162, 342, 19 },
		{ 162, 343, 19 },
		{ 162, 344, 19 },
		{ 162, 345, 19 },
		{ 162, 346, 19 },
		{ 162, 347, 19 },
		{ 162, 348, 19 },
		{ 162, 349, 19 },
		{ 162, 350, 19 },
		{ 162, 351, 19 },
		{ 163, 305, 19 },
		{ 163, 306, 19 },
		{ 163, 307, 19 },
		{ 163, 308, 19 },
		{ 163, 309, 19 },
		{ 163, 310, 19 },
		{ 163, 311, 19 },
		{ 163, 312, 19 },
		{ 163, 313, 19 },
		{ 163, 314, 19 },
		{ 163, 315, 19 },
		{ 163, 316, 19 },
		{ 163, 317, 19 },
		{ 163, 318, 19 },
		{ 163, 319, 19 },
		{ 163, 320, 19 },
		{ 163, 321, 19 },
		{ 163, 322, 19 },
		{ 163, 323, 19 },
		{ 163, 324, 19 },
		{ 163, 325, 19 },
		{ 163, 326, 19 },
		{ 163, 327, 19 },
		{ 163, 328, 19 },
		{ 163, 329, 19 },
		{ 163, 330, 19 },
		{ 163, 331, 19 },
		{ 163, 332, 19 },
		{ 163, 333, 19 },
		{ 163, 334, 19 },
		{ 163, 335, 19 },
		{ 163, 336, 19 },
		{ 163, 337, 19 },
		{ 163, 338, 19 },
		{ 163, 339, 19 },
		{ 163, 340, 19 },
		{ 163, 341, 19 },
		{ 163, 342, 19 },
		{ 163, 343, 19 },
		{ 163, 344, 19 },
		{ 163, 345, 19 },
		{ 163, 346, 19 },
		{ 163, 347, 19 },
		{ 163, 348, 19 },
		{ 163, 349, 19 },
		{ 163, 350, 19 },
		{ 164, 304, 19 },
		{ 164, 305, 19 },
		{ 164, 306, 19 },
		{ 164, 307, 19 },
		{ 164, 308, 19 },
		{ 164, 309, 19 },
		{ 164, 310, 19 },
		{ 164, 311, 19 },
		{ 164, 312, 19 },
		{ 164, 313, 19 },
		{ 164, 314, 19 },
		{ 164, 315, 19 },
		{ 164, 316, 19 },
		{ 164, 317, 19 },
		{ 164, 318, 19 },
		{ 164, 319, 19 },
		{ 164, 320, 19 },
		{ 164, 321, 19 },
		{ 164, 322, 19 },
		{ 164, 323, 19 },
		{ 164, 324, 19 },
		{ 164, 325, 19 },
		{ 164, 326, 19 },
		{ 164, 327, 19 },
		{ 164, 328, 19 },
		{ 164, 329, 19 },
		{ 164, 330, 19 },
		{ 164, 331, 19 },
		{ 164, 332, 19 },
		{ 164, 333, 19 },
		{ 164, 334, 19 },
		{ 164, 335, 19 },
		{ 164, 336, 19 },
		{ 164, 337, 19 },
		{ 164, 338, 19 },
		{ 164, 339, 19 },
		{ 164, 340, 19 },
		{ 164, 341, 19 },
		{ 164, 342, 19 },
		{ 164, 343, 19 },
		{ 164, 344, 19 },
		{ 164, 345, 19 },
		{ 164, 346, 19 },
		{ 164, 347, 19 },
		{ 164, 348, 19 },
		{ 164, 349, 19 },
		{ 165, 304, 19 },
		{ 165, 305, 19 },
		{ 165, 306, 19 },
		{ 165, 307, 19 },
		{ 165, 308, 19 },
		{ 165, 309, 19 },
		{ 165, 310, 19 },
		{ 165, 311, 19 },
		{ 165, 312, 19 },
		{ 165, 313, 19 },
		{ 165, 314, 19 },
		{ 165, 315, 19 },
		{ 165, 316, 19 },
		{ 165, 317, 19 },
		{ 165, 318, 19 },
		{ 165, 319, 19 },
		{ 165, 320, 19 },
		{ 165, 321, 19 },
		{ 165, 322, 19 },
		{ 165, 323, 19 },
		{ 165, 324, 19 },
		{ 165, 325, 19 },
		{ 165, 326, 19 },
		{ 165, 327, 19 },
		{ 165, 328, 19 },
		{ 165, 329, 19 },
		{ 165, 330, 19 },
		{ 165, 331, 19 },
		{ 165, 332, 19 },
		{ 165, 333, 19 },
		{ 165, 334, 19 },
		{ 165, 335, 19 },
		{ 165, 336, 19 },
		{ 165, 337, 19 },
		{ 165, 338, 19 },
		{ 165, 339, 19 },
		{ 165, 340, 19 },
		{ 165, 341, 19 },
		{ 165, 342, 19 },
		{ 165, 343, 19 },
		{ 165, 344, 19 },
		{ 165, 345, 19 },
		{ 165, 346, 19 },
		{ 165, 347, 19 },
		{ 165, 348, 19 },
		{ 166, 303, 19 },
		{ 166, 304, 19 },
		{ 166, 305, 19 },
		{ 166, 306, 19 },
		{ 166, 307, 19 },
		{ 166, 308, 19 },
		{ 166, 309, 19 },
		{ 166, 310, 19 },
		{ 166, 311, 19 },
		{ 166, 312, 19 },
		{ 166, 313, 19 },
		{ 166, 314, 19 },
		{ 166, 315, 19 },
		{ 166, 316, 19 },
		{ 166, 317, 19 },
		{ 166, 318, 19 },
		{ 166, 319, 19 },
		{ 166, 320, 19 },
		{ 166, 321, 19 },
		{ 166, 322, 19 },
		{ 166, 323, 19 },
		{ 166, 324, 19 },
		{ 166, 325, 19 },
		{ 166, 326, 19 },
		{ 166, 327, 19 },
		{ 166, 328, 19 },
		{ 166, 329, 19 },
		{ 166, 330, 19 },
		{ 166, 331, 19 },
		{ 166, 332, 19 },
		{ 166, 333, 19 },
		{ 166, 334, 19 },
		{ 166, 335, 19 },
		{ 166, 336, 19 },
		{ 166, 337, 19 },
		{ 166, 338, 19 },
		{ 166, 339, 19 },
		{ 166, 340, 19 },
		{ 166, 341, 19 },
		{ 166, 342, 19 },
		{ 166, 343, 19 },
		{ 166, 344, 19 },
		{ 166, 345, 19 },
		{ 166, 346, 19 },
		{ 166, 347, 19 },
		{ 167, 303, 19 },
		{ 167, 304, 19 },
		{ 167, 305, 19 },
		{ 167, 306, 19 },
		{ 167, 307, 19 },
		{ 167, 308, 19 },
		{ 167, 309, 19 },
		{ 167, 310, 19 },
		{ 167, 311, 19 },
		{ 167, 312, 19 },
		{ 167, 313, 19 },
		{ 167, 314, 19 },
		{ 167, 315, 19 },
		{ 167, 316, 19 },
		{ 167, 317, 19 },
		{ 167, 318, 19 },
		{ 167, 319, 19 },
		{ 167, 320, 19 },
		{ 167, 321, 19 },
		{ 167, 322, 19 },
		{ 167, 323, 19 },
		{ 167, 324, 19 },
		{ 167, 325, 19 },
		{ 167, 326, 19 },
		{ 167, 327, 19 },
		{ 167, 328, 19 },
		{ 167, 329, 19 },
		{ 167, 330, 19 },
		{ 167, 331, 19 },
		{ 167, 332, 19 },
		{ 167, 333, 19 },
		{ 167, 334, 19 },
		{ 167, 335, 19 },
		{ 167, 336, 19 },
		{ 167, 337, 19 },
		{ 167, 338, 19 },
		{ 167, 339, 19 },
		{ 167, 340, 19 },
		{ 167, 341, 19 },
		{ 167, 342, 19 },
		{ 167, 343, 19 },
		{ 167, 344, 19 },
		{ 167, 345, 19 },
		{ 167, 346, 19 },
		{ 168, 302, 19 },
		{ 168, 303, 19 },
		{ 168, 304, 19 },
		{ 168, 305, 19 },
		{ 168, 306, 19 },
		{ 168, 307, 19 },
		{ 168, 308, 19 },
		{ 168, 309, 19 },
		{ 168, 310, 19 },
		{ 168, 311, 19 },
		{ 168, 312, 19 },
		{ 168, 313, 19 },
		{ 168, 314, 19 },
		{ 168, 315, 19 },
		{ 168, 316, 19 },
		{ 168, 317, 19 },
		{ 168, 318, 19 },
		{ 168, 319, 19 },
		{ 168, 320, 19 },
		{ 168, 321, 19 },
		{ 168, 322, 19 },
		{ 168, 323, 19 },
		{ 168, 324, 19 },
		{ 168, 325, 19 },
		{ 168, 326, 19 },
		{ 168, 327, 19 },
		{ 168, 328, 19 },
		{ 168, 329, 19 },
		{ 168, 330, 19 },
		{ 168, 331, 19 },
		{ 168, 332, 19 },
		{ 168, 333, 19 },
		{ 168, 334, 19 },
		{ 168, 335, 19 },
		{ 168, 336, 19 },
		{ 168, 337, 19 },
		{ 168, 338, 19 },
		{ 168, 339, 19 },
		{ 168, 340, 19 },
		{ 168, 341, 19 },
		{ 168, 342, 19 },
		{ 168, 343, 19 },
		{ 168, 344, 19 },
		{ 168, 345, 19 },
		{ 169, 302, 19 },
		{ 169, 303, 19 },
		{ 169, 304, 19 },
		{ 169, 305, 19 },
		{ 169, 306, 19 },
		{ 169, 307, 19 },
		{ 169, 308, 19 },
		{ 169, 309, 19 },
		{ 169, 310, 19 },
		{ 169, 311, 19 },
		{ 169, 312, 19 },
		{ 169, 313, 19 },
		{ 169, 314, 19 },
		{ 169, 315, 19 },
		{ 169, 316, 19 },
		{ 169, 317, 19 },
		{ 169, 318, 19 },
		{ 169, 319, 19 },
		{ 169, 320, 19 },
		{ 169, 321, 19 },
		{ 169, 322, 19 },
		{ 169, 323, 19 },
		{ 169, 324, 19 },
		{ 169, 325, 19 },
		{ 169, 326, 19 },
		{ 169, 327, 19 },
		{ 169, 328, 19 },
		{ 169, 329, 19 },
		{ 169, 330, 19 },
		{ 169, 331, 19 },
		{ 169, 332, 19 },
		{ 169, 333, 19 },
		{ 169, 334, 19 },
		{ 169, 335, 19 },
		{ 169, 336, 19 },
		{ 169, 337, 19 },
		{ 169, 338, 19 },
		{ 169, 339, 19 },
		{ 169, 340, 19 },
		{ 169, 341, 19 },
		{ 169, 342, 19 },
		{ 169, 343, 19 },
		{ 169, 344, 19 },
		{ 170, 301, 19 },
		{ 170, 302, 19 },
		{ 170, 303, 19 },
		{ 170, 304, 19 },
		{ 170, 305, 19 },
		{ 170, 306, 19 },
		{ 170, 307, 19 },
		{ 170, 308, 19 },
		{ 170, 309, 19 },
		{ 170, 310, 19 },
		{ 170, 311, 19 },
		{ 170, 312, 19 },
		{ 170, 313, 19 },
		{ 170, 314, 19 },
		{ 170, 315, 19 },
		{ 170, 316, 19 },
		{ 170, 317, 19 },
		{ 170, 318, 19 },
		{ 170, 319, 19 },
		{ 170, 320, 19 },
		{ 170, 321, 19 },
		{ 170, 322, 19 },
		{ 170, 323, 19 },
		{ 170, 324, 19 },
		{ 170, 325, 19 },
		{ 170, 326, 19 },
		{ 170, 327, 19 },
		{ 170, 328, 19 },
		{ 170, 329, 19 },
		{ 170, 330, 19 },
		{ 170, 331, 19 },
		{ 170, 332, 19 },
		{ 170, 333, 19 },
		{ 170, 334, 19 },
		{ 170, 335, 19 },
		{ 170, 336, 19 },
		{ 170, 337, 19 },
		{ 170, 338, 19 },
		{ 170, 339, 19 },
		{ 170, 340, 19 },
		{ 170, 341, 19 },
		{ 170, 342, 19 },
		{ 170, 343, 19 },
		{ 171, 301, 19 },
		{ 171, 302, 19 },
		{ 171, 303, 19 },
		{ 171, 304, 19 },
		{ 171, 305, 19 },
		{ 171, 306, 19 },
		{ 171, 307, 19 },
		{ 171, 308, 19 },
		{ 171, 309, 19 },
		{ 171, 310, 19 },
		{ 171, 311, 19 },
		{ 171, 312, 19 },
		{ 171, 313, 19 },
		{ 171, 314, 19 },
		{ 171, 315, 19 },
		{ 171, 316, 19 },
		{ 171, 317, 19 },
		{ 171, 318, 19 },
		{ 171, 319, 19 },
		{ 171, 320, 19 },
		{ 171, 321, 19 },
		{ 171, 322, 19 },
		{ 171, 323, 19 },
		{ 171, 324, 19 },
		{ 171, 325, 19 },
		{ 171, 326, 19 },
		{ 171, 327, 19 },
		{ 171, 328, 19 },
		{ 171, 329, 19 },
		{ 171, 330, 19 },
		{ 171, 331, 19 },
		{ 171, 332, 19 },
		{ 171, 333, 19 },
		{ 171, 334, 19 },
		{ 171, 335, 19 },
		{ 171, 336, 19 },
		{ 171, 337, 19 },
		{ 171, 338, 19 },
		{ 171, 339, 19 },
		{ 171, 340, 19 },
		{ 171, 341, 19 },
		{ 171, 342, 19 },
		{ 172, 300, 19 },
		{ 172, 301, 19 },
		{ 172, 302, 19 },
		{ 172, 303, 19 },
		{ 172, 304, 19 },
		{ 172, 305, 19 },
		{ 172, 306, 19 },
		{ 172, 307, 19 },
		{ 172, 308, 19 },
		{ 172, 309, 19 },
		{ 172, 310, 19 },
		{ 172, 311, 19 },
		{ 172, 312, 19 },
		{ 172, 313, 19 },
		{ 172, 314, 19 },
		{ 172, 315, 19 },
		{ 172, 316, 19 },
		{ 172, 317, 19 },
		{ 172, 318, 19 },
		{ 172, 319, 19 },
		{ 172, 320, 19 },
		{ 172, 321, 19 },
		{ 172, 322, 19 },
		{ 172, 323, 19 },
		{ 172, 324, 19 },
		{ 172, 325, 19 },
		{ 172, 326, 19 },
		{ 172, 327, 19 },
		{ 172, 328, 19 },
		{ 172, 329, 19 },
		{ 172, 330, 19 },
		{ 172, 331, 19 },
		{ 172, 332, 19 },
		{ 172, 333, 19 },
		{ 172, 334, 19 },
		{ 172, 335, 19 },
		{ 172, 336, 19 },
		{ 172, 337, 19 },
		{ 172, 338, 19 },
		{ 172, 339, 19 },
		{ 172, 340, 19 },
		{ 172, 341, 19 },
		{ 173, 300, 19 },
		{ 173, 301, 19 },
		{ 173, 302, 19 },
		{ 173, 303, 19 },
		{ 173, 304, 19 },
		{ 173, 305, 19 },
		{ 173, 306, 19 },
		{ 173, 307, 19 },
		{ 173, 308, 19 },
		{ 173, 309, 19 },
		{ 173, 310, 19 },
		{ 173, 311, 19 },
		{ 173, 312, 19 },
		{ 173, 313, 19 },
		{ 173, 314, 19 },
		{ 173, 315, 19 },
		{ 173, 316, 19 },
		{ 173, 317, 19 },
		{ 173, 318, 19 },
		{ 173, 319, 19 },
		{ 173, 320, 19 },
		{ 173, 321, 19 },
		{ 173, 322, 19 },
		{ 173, 323, 19 },
		{ 173, 324, 19 },
		{ 173, 325, 19 },
		{ 173, 326, 19 },
		{ 173, 327, 19 },
		{ 173, 328, 19 },
		{ 173, 329, 19 },
		{ 173, 330, 19 },
		{ 173, 331, 19 },
		{ 173, 332, 19 },
		{ 173, 333, 19 },
		{ 173, 334, 19 },
		{ 173, 335, 19 },
		{ 173, 336, 19 },
		{ 173, 337, 19 },
		{ 173, 338, 19 },
		{ 173, 339, 19 },
		{ 173, 340, 19 },
		{ 174, 299, 19 },
		{ 174, 300, 19 },
		{ 174, 301, 19 },
		{ 174, 302, 19 },
		{ 174, 303, 19 },
		{ 174, 304, 19 },
		{ 174, 305, 19 },
		{ 174, 306, 19 },
		{ 174, 307, 19 },
		{ 174, 308, 19 },
		{ 174, 309, 19 },
		{ 174, 310, 19 },
		{ 174, 311, 19 },
		{ 174, 312, 19 },
		{ 174, 313, 19 },
		{ 174, 314, 19 },
		{ 174, 315, 19 },
		{ 174, 316, 19 },
		{ 174, 317, 19 },
		{ 174, 318, 19 },
		{ 174, 319, 19 },
		{ 174, 320, 19 },
		{ 174, 321, 19 },
		{ 174, 322, 19 },
		{ 174, 323, 19 },
		{ 174, 324, 19 },
		{ 174, 325, 19 },
		{ 174, 326, 19 },
		{ 174, 327, 19 },
		{ 174, 328, 19 },
		{ 174, 329, 19 },
		{ 174, 330, 19 },
		{ 174, 331, 19 },
		{ 174, 332, 19 },
		{ 174, 333, 19 },
		{ 174, 334, 19 },
		{ 174, 335, 19 },
		{ 174, 336, 19 },
		{ 174, 337, 19 },
		{ 174, 338, 19 },
		{ 174, 339, 19 },
		{ 175, 299, 19 },
		{ 175, 300, 19 },
		{ 175, 301, 19 },
		{ 175, 302, 19 },
		{ 175, 303, 19 },
		{ 175, 304, 19 },
		{ 175, 305, 19 },
		{ 175, 306, 19 },
		{ 175, 307, 19 },
		{ 175, 308, 19 },
		{ 175, 309, 19 },
		{ 175, 310, 19 },
		{ 175, 311, 19 },
		{ 175, 312, 19 },
		{ 175, 313, 19 },
		{ 175, 314, 19 },
		{ 175, 315, 19 },
		{ 175, 316, 19 },
		{ 175, 317, 19 },
		{ 175, 318, 19 },
		{ 175, 319, 19 },
		{ 175, 320, 19 },
		{ 175, 321, 19 },
		{ 175, 322, 19 },
		{ 175, 323, 19 },
		{ 175, 324, 19 },
		{ 175, 325, 19 },
		{ 175, 326, 19 },
		{ 175, 327, 19 },
		{ 175, 328, 19 },
		{ 175, 329, 19 },
		{ 175, 330, 19 },
		{ 175, 331, 19 },
		{ 175, 332, 19 },
		{ 175, 333, 19 },
		{ 175, 334, 19 },
		{ 175, 335, 19 },
		{ 175, 336, 19 },
		{ 175, 337, 19 },
		{ 175, 338, 19 },
		{ 176, 298, 19 },
		{ 176, 299, 19 },
		{ 176, 300, 19 },
		{ 176, 301, 19 },
		{ 176, 302, 19 },
		{ 176, 303, 19 },
		{ 176, 304, 19 },
		{ 176, 305, 19 },
		{ 176, 306, 19 },
		{ 176, 307, 19 },
		{ 176, 308, 19 },
		{ 176, 309, 19 },
		{ 176, 310, 19 },
		{ 176, 311, 19 },
		{ 176, 312, 19 },
		{ 176, 313, 19 },
		{ 176, 314, 19 },
		{ 176, 315, 19 },
		{ 176, 316, 19 },
		{ 176, 317, 19 },
		{ 176, 318, 19 },
		{ 176, 319, 19 },
		{ 176, 320, 19 },
		{ 176, 321, 19 },
		{ 176, 322, 19 },
		{ 176, 323, 19 },
		{ 176, 324, 19 },
		{ 176, 325, 19 },
		{ 176, 326, 19 },
		{ 176, 327, 19 },
		{ 176, 328, 19 },
		{ 176, 329, 19 },
		{ 176, 330, 19 },
		{ 176, 331, 19 },
		{ 176, 332, 19 },
		{ 176, 333, 19 },
		{ 176, 334, 19 },
		{ 176, 335, 19 },
		{ 176, 336, 19 },
		{ 176, 337, 19 },
		{ 177, 298, 19 },
		{ 177, 299, 19 },
		{ 177, 300, 19 },
		{ 177, 301, 19 },
		{ 177, 302, 19 },
		{ 177, 303, 19 },
		{ 177, 304, 19 },
		{ 177, 305, 19 },
		{ 177, 306, 19 },
		{ 177, 307, 19 },
		{ 177, 308, 19 },
		{ 177, 309, 19 },
		{ 177, 310, 19 },
		{ 177, 311, 19 },
		{ 177, 312, 19 },
		{ 177, 313, 19 },
		{ 177, 314, 19 },
		{ 177, 315, 19 },
		{ 177, 316, 19 },
		{ 177, 317, 19 },
		{ 177, 318, 19 },
		{ 177, 319, 19 },
		{ 177, 320, 19 },
		{ 177, 321, 19 },
		{ 177, 322, 19 },
		{ 177, 323, 19 },
		{ 177, 324, 19 },
		{ 177, 325, 19 },
		{ 177, 326, 19 },
		{ 177, 327, 19 },
		{ 177, 328, 19 },
		{ 177, 329, 19 },
		{ 177, 330, 19 },
		{ 177, 331, 19 },
		{ 177, 332, 19 },
		{ 177, 333, 19 },
		{ 177, 334, 19 },
		{ 177, 335, 19 },
		{ 177, 336, 19 },
		{ 178, 297, 19 },
		{ 178, 298, 19 },
		{ 178, 299, 19 },
		{ 178, 300, 19 },
		{ 178, 301, 19 },
		{ 178, 302, 19 },
		{ 178, 303, 19 },
		{ 178, 304, 19 },
		{ 178, 305, 19 },
		{ 178, 306, 19 },
		{ 178, 307, 19 },
		{ 178, 308, 19 },
		{ 178, 309, 19 },
		{ 178, 310, 19 },
		{ 178, 311, 19 },
		{ 178, 312, 19 },
		{ 178, 313, 19 },
		{ 178, 314, 19 },
		{ 178, 315, 19 },
		{ 178, 316, 19 },
		{ 178, 317, 19 },
		{ 178, 318, 19 },
		{ 178, 319, 19 },
		{ 178, 320, 19 },
		{ 178, 321, 19 },
		{ 178, 322, 19 },
		{ 178, 323, 19 },
		{ 178, 324, 19 },
		{ 178, 325, 19 },
		{ 178, 326, 19 },
		{ 178, 327, 19 },
		{ 178, 328, 19 },
		{ 178, 329, 19 },
		{ 178, 330, 19 },
		{ 178, 331, 19 },
		{ 178, 332, 19 },
		{ 178, 333, 19 },
		{ 178, 334, 19 },
		{ 178, 335, 19 },
		{ 179, 297, 19 },
		{ 179, 298, 19 },
		{ 179, 299, 19 },
		{ 179, 300, 19 },
		{ 179, 301, 19 },
		{ 179, 302, 19 },
		{ 179, 303, 19 },
		{ 179, 304, 19 },
		{ 179, 305, 19 },
		{ 179, 306, 19 },
		{ 179, 307, 19 },
		{ 179, 308, 19 },
		{ 179, 309, 19 },
		{ 179, 310, 19 },
		{ 179, 311, 19 },
		{ 179, 312, 19 },
		{ 179, 313, 19 },
		{ 179, 314, 19 },
		{ 179, 315, 19 },
		{ 179, 316, 19 },
		{ 179, 317, 19 },
		{ 179, 318, 19 },
		{ 179, 319, 19 },
		{ 179, 320, 19 },
		{ 179, 321, 19 },
		{ 179, 322, 19 },
		{ 179, 323, 19 },
		{ 179, 324, 19 },
		{ 179, 325, 19 },
		{ 179, 326, 19 },
		{ 179, 327, 19 },
		{ 179, 328, 19 },
		{ 179, 329, 19 },
		{ 179, 330, 19 },
		{ 179, 331, 19 },
		{ 179, 332, 19 },
		{ 179, 333, 19 },
		{ 179, 334, 19 },
		{ 180, 296, 19 },
		{ 180, 297, 19 },
		{ 180, 298, 19 },
		{ 180, 299, 19 },
		{ 180, 300, 19 },
		{ 180, 301, 19 },
		{ 180, 302, 19 },
		{ 180, 303, 19 },
		{ 180, 304, 19 },
		{ 180, 305, 19 },
		{ 180, 306, 19 },
		{ 180, 307, 19 },
		{ 180, 308, 19 },
		{ 180, 309, 19 },
		{ 180, 310, 19 },
		{ 180, 311, 19 },
		{ 180, 312, 19 },
		{ 180, 313, 19 },
		{ 180, 314, 19 },
		{ 180, 315, 19 },
		{ 180, 316, 19 },
		{ 180, 317, 19 },
		{ 180, 318, 19 },
		{ 180, 319, 19 },
		{ 180, 320, 19 },
		{ 180, 321, 19 },
		{ 180, 322, 19 },
		{ 180, 323, 19 },
		{ 180, 324, 19 },
		{ 180, 325, 19 },
		{ 180, 326, 19 },
		{ 180, 327, 19 },
		{ 180, 328, 19 },
		{ 180, 329, 19 },
		{ 180, 330, 19 },
		{ 180, 331, 19 },
		{ 180, 332, 19 },
		{ 180, 333, 19 },
		{ 181, 296, 19 },
		{ 181, 297, 19 },
		{ 181, 298, 19 },
		{ 181, 299, 19 },
		{ 181, 300, 19 },
		{ 181, 301, 19 },
		{ 181, 302, 19 },
		{ 181, 303, 19 },
		{ 181, 304, 19 },
		{ 181, 305, 19 },
		{ 181, 306, 19 },
		{ 181, 307, 19 },
		{ 181, 308, 19 },
		{ 181, 309, 19 },
		{ 181, 310, 19 },
		{ 181, 311, 19 },
		{ 181, 312, 19 },
		{ 181, 313, 19 },
		{ 181, 314, 19 },
		{ 181, 315, 19 },
		{ 181, 316, 19 },
		{ 181, 317, 19 },
		{ 181, 318, 19 },
		{ 181, 319, 19 },
		{ 181, 320, 19 },
		{ 181, 321, 19 },
		{ 181, 322, 19 },
		{ 181, 323, 19 },
		{ 181, 324, 19 },
		{ 181, 325, 19 },
		{ 181, 326, 19 },
		{ 181, 327, 19 },
		{ 181, 328, 19 },
		{ 181, 329, 19 },
		{ 181, 330, 19 },
		{ 181, 331, 19 },
		{ 181, 332, 19 },
		{ 182, 295, 19 },
		{ 182, 296, 19 },
		{ 182, 297, 19 },
		{ 182, 298, 19 },
		{ 182, 299, 19 },
		{ 182, 300, 19 },
		{ 182, 301, 19 },
		{ 182, 302, 19 },
		{ 182, 303, 19 },
		{ 182, 304, 19 },
		{ 182, 305, 19 },
		{ 182, 306, 19 },
		{ 182, 307, 19 },
		{ 182, 308, 19 },
		{ 182, 309, 19 },
		{ 182, 310, 19 },
		{ 182, 311, 19 },
		{ 182, 312, 19 },
		{ 182, 313, 19 },
		{ 182, 314, 19 },
		{ 182, 315, 19 },
		{ 182, 316, 19 },
		{ 182, 317, 19 },
		{ 182, 318, 19 },
		{ 182, 319, 19 },
		{ 182, 320, 19 },
		{ 182, 321, 19 },
		{ 182, 322, 19 },
		{ 182, 323, 19 },
		{ 182, 324, 19 },
		{ 182, 325, 19 },
		{ 182, 326, 19 },
		{ 182, 327, 19 },
		{ 182, 328, 19 },
		{ 182, 329, 19 },
		{ 182, 330, 19 },
		{ 182, 331, 19 },
		{ 183, 295, 19 },
		{ 183, 296, 19 },
		{ 183, 297, 19 },
		{ 183, 298, 19 },
		{ 183, 299, 19 },
		{ 183, 300, 19 },
		{ 183, 301, 19 },
		{ 183, 302, 19 },
		{ 183, 303, 19 },
		{ 183, 304, 19 },
		{ 183, 305, 19 },
		{ 183, 306, 19 },
		{ 183, 307, 19 },
		{ 183, 308, 19 },
		{ 183, 309, 19 },
		{ 183, 310, 19 },
		{ 183, 311, 19 },
		{ 183, 312, 19 },
		{ 183, 313, 19 },
		{ 183, 314, 19 },
		{ 183, 315, 19 },
		{ 183, 316, 19 },
		{ 183, 317, 19 },
		{ 183, 318, 19 },
		{ 183, 319, 19 },
		{ 183, 320, 19 },
		{ 183, 321, 19 },
		{ 183, 322, 19 },
		{ 183, 323, 19 },
		{ 183, 324, 19 },
		{ 183, 325, 19 },
		{ 183, 326, 19 },
		{ 183, 327, 19 },
		{ 183, 328, 19 },
		{ 183, 329, 19 },
		{ 183, 330, 19 },
		{ 184, 294, 19 },
		{ 184, 295, 19 },
		{ 184, 296, 19 },
		{ 184, 297, 19 },
		{ 184, 298, 19 },
		{ 184, 299, 19 },
		{ 184, 300, 19 },
		{ 184, 301, 19 },
		{ 184, 302, 19 },
		{ 184, 303, 19 },
		{ 184, 304, 19 },
		{ 184, 305, 19 },
		{ 184, 306, 19 },
		{ 184, 307, 19 },
		{ 184, 308, 19 },
		{ 184, 309, 19 },
		{ 184, 310, 19 },
		{ 184, 311, 19 },
		{ 184, 312, 19 },
		{ 184, 313, 19 },
		{ 184, 314, 19 },
		{ 184, 315, 19 },
		{ 184, 316, 19 },
		{ 184, 317, 19 },
		{ 184, 318, 19 },
		{ 184, 319, 19 },
		{ 184, 320, 19 },
		{ 184, 321, 19 },
		{ 184, 322, 19 },
		{ 184, 323, 19 },
		{ 184, 324, 19 },
		{ 184, 325, 19 },
		{ 184, 326, 19 },
		{ 184, 327, 19 },
		{ 184, 328, 19 },
		{ 184, 329, 19 },
		{ 185, 294, 19 },
		{ 185, 295, 19 },
		{ 185, 296, 19 },
		{ 185, 297, 19 },
		{ 185, 298, 19 },
		{ 185, 299, 19 },
		{ 185, 300, 19 },
		{ 185, 301, 19 },
		{ 185, 302, 19 },
		{ 185, 303, 19 },
		{ 185, 304, 19 },
		{ 185, 305, 19 },
		{ 185, 306, 19 },
		{ 185, 307, 19 },
		{ 185, 308, 19 },
		{ 185, 309, 19 },
		{ 185, 310, 19 },
		{ 185, 311, 19 },
		{ 185, 312, 19 },
		{ 185, 313, 19 },
		{ 185, 314, 19 },
		{ 185, 315, 19 },
		{ 185, 316, 19 },
		{ 185, 317, 19 },
		{ 185, 318, 19 },
		{ 185, 319, 19 },
		{ 185, 320, 19 },
		{ 185, 321, 19 },
		{ 185, 322, 19 },
		{ 185, 323, 19 },
		{ 185, 324, 19 },
		{ 185, 325, 19 },
		{ 185, 326, 19 },
		{ 185, 327, 19 },
		{ 185, 328, 19 },
		{ 186, 293, 19 },
		{ 186, 294, 19 },
		{ 186, 295, 19 },
		{ 186, 296, 19 },
		{ 186, 297, 19 },
		{ 186, 298, 19 },
		{ 186, 299, 19 },
		{ 186, 300, 19 },
		{ 186, 301, 19 },
		{ 186, 302, 19 },
		{ 186, 303, 19 },
		{ 186, 304, 19 },
		{ 186, 305, 19 },
		{ 186, 306, 19 },
		{ 186, 307, 19 },
		{ 186, 308, 19 },
		{ 186, 309, 19 },
		{ 186, 310, 19 },
		{ 186, 311, 19 },
		{ 186, 312, 19 },
		{ 186, 313, 19 },
		{ 186, 314, 19 },
		{ 186, 315, 19 },
		{ 186, 316, 19 },
		{ 186, 317, 19 },
		{ 186, 318, 19 },
		{ 186, 319, 19 },
		{ 186, 320, 19 },
		{ 186, 321, 19 },
		{ 186, 322, 19 },
		{ 186, 323, 19 },
		{ 186, 324, 19 },
		{ 186, 325, 19 },
		{ 186, 326, 19 },
		{ 186, 327, 19 },
		{ 187, 293, 19 },
		{ 187, 294, 19 },
		{ 187, 295, 19 },
		{ 187, 296, 19 },
		{ 187, 297, 19 },
		{ 187, 298, 19 },
		{ 187, 299, 19 },
		{ 187, 300, 19 },
		{ 187, 301, 19 },
		{ 187, 302, 19 },
		{ 187, 303, 19 },
		{ 187, 304, 19 },
		{ 187, 305, 19 },
		{ 187, 306, 19 },
		{ 187, 307, 19 },
		{ 187, 308, 19 },
		{ 187, 309, 19 },
		{ 187, 310, 19 },
		{ 187, 311, 19 },
		{ 187, 312, 19 },
		{ 187, 313, 19 },
		{ 187, 314, 19 },
		{ 187, 315, 19 },
		{ 187, 316, 19 },
		{ 187, 317, 19 },
		{ 187, 318, 19 },
		{ 187, 319, 19 },
		{ 187, 320, 19 },
		{ 187, 321, 19 },
		{ 187, 322, 19 },
		{ 187, 323, 19 },
		{ 187, 324, 19 },
		{ 187, 325, 19 },
		{ 187, 326, 19 },
		{ 188, 292, 19 },
		{ 188, 293, 19 },
		{ 188, 294, 19 },
		{ 188, 295, 19 },
		{ 188, 296, 19 },
		{ 188, 297, 19 },
		{ 188, 298, 19 },
		{ 188, 299, 19 },
		{ 188, 300, 19 },
		{ 188, 301, 19 },
		{ 188, 302, 19 },
		{ 188, 303, 19 },
		{ 188, 304, 19 },
		{ 188, 305, 19 },
		{ 188, 306, 19 },
		{ 188, 307, 19 },
		{ 188, 308, 19 },
		{ 188, 309, 19 },
		{ 188, 310, 19 },
		{ 188, 311, 19 },
		{ 188, 312, 19 },
		{ 188, 313, 19 },
		{ 188, 314, 19 },
		{ 188, 315, 19 },
		{ 188, 316, 19 },
		{ 188, 317, 19 },
		{ 188, 318, 19 },
		{ 188, 319, 19 },
		{ 188, 320, 19 },
		{ 188, 321, 19 },
		{ 188, 322, 19 },
		{ 188, 323, 19 },
		{ 188, 324, 19 },
		{ 188, 325, 19 },
		{ 189, 292, 19 },
		{ 189, 293, 19 },
		{ 189, 294, 19 },
		{ 189, 295, 19 },
		{ 189, 296, 19 },
		{ 189, 297, 19 },
		{ 189, 298, 19 },
		{ 189, 299, 19 },
		{ 189, 300, 19 },
		{ 189, 301, 19 },
		{ 189, 302, 19 },
		{ 189, 303, 19 },
		{ 189, 304, 19 },
		{ 189, 305, 19 },
		{ 189, 306, 19 },
		{ 189, 307, 19 },
		{ 189, 308, 19 },
		{ 189, 309, 19 },
		{ 189, 310, 19 },
		{ 189, 311, 19 },
		{ 189, 312, 19 },
		{ 189, 313, 19 },
		{ 189, 314, 19 },
		{ 189, 315, 19 },
		{ 189, 316, 19 },
		{ 189, 317, 19 },
		{ 189, 318, 19 },
		{ 189, 319, 19 },
		{ 189, 320, 19 },
		{ 189, 321, 19 },
		{ 189, 322, 19 },
		{ 189, 323, 19 },
		{ 189, 324, 19 },
		{ 190, 291, 19 },
		{ 190, 292, 19 },
		{ 190, 293, 19 },
		{ 190, 294, 19 },
		{ 190, 295, 19 },
		{ 190, 296, 19 },
		{ 190, 297, 19 },
		{ 190, 298, 19 },
		{ 190, 299, 19 },
		{ 190, 300, 19 },
		{ 190, 301, 19 },
		{ 190, 302, 19 },
		{ 190, 303, 19 },
		{ 190, 304, 19 },
		{ 190, 305, 19 },
		{ 190, 306, 19 },
		{ 190, 307, 19 },
		{ 190, 308, 19 },
		{ 190, 309, 19 },
		{ 190, 310, 19 },
		{ 190, 311, 19 },
		{ 190, 312, 19 },
		{ 190, 313, 19 },
		{ 190, 314, 19 },
		{ 190, 315, 19 },
		{ 190, 316, 19 },
		{ 190, 317, 19 },
		{ 190, 318, 19 },
		{ 190, 319, 19 },
		{ 190, 320, 19 },
		{ 190, 321, 19 },
		{ 190, 322, 19 },
		{ 190, 323, 19 },
		{ 191, 291, 19 },
		{ 191, 292, 19 },
		{ 191, 293, 19 },
		{ 191, 294, 19 },
		{ 191, 295, 19 },
		{ 191, 296, 19 },
		{ 191, 297, 19 },
		{ 191, 298, 19 },
		{ 191, 299, 19 },
		{ 191, 300, 19 },
		{ 191, 301, 19 },
		{ 191, 302, 19 },
		{ 191, 303, 19 },
		{ 191, 304, 19 },
		{ 191, 305, 19 },
		{ 191, 306, 19 },
		{ 191, 307, 19 },
		{ 191, 308, 19 },
		{ 191, 309, 19 },
		{ 191, 310, 19 },
		{ 191, 311, 19 },
		{ 191, 312, 19 },
		{ 191, 313, 19 },
		{ 191, 314, 19 },
		{ 191, 315, 19 },
		{ 191, 316, 19 },
		{ 191, 317, 19 },
		{ 191, 318, 19 },
		{ 191, 319, 19 },
		{ 191, 320, 19 },
		{ 191, 321, 19 },
		{ 191, 322, 19 },
		{ 192, 290, 19 },
		{ 192, 291, 19 },
		{ 192, 292, 19 },
		{ 192, 293, 19 },
		{ 192, 294, 19 },
		{ 192, 295, 19 },
		{ 192, 296, 19 },
		{ 192, 297, 19 },
		{ 192, 298, 19 },
		{ 192, 299, 19 },
		{ 192, 300, 19 },
		{ 192, 301, 19 },
		{ 192, 302, 19 },
		{ 192, 303, 19 },
		{ 192, 304, 19 },
		{ 192, 305, 19 },
		{ 192, 306, 19 },
		{ 192, 307, 19 },
		{ 192, 308, 19 },
		{ 192, 309, 19 },
		{ 192, 310, 19 },
		{ 192, 311, 19 },
		{ 192, 312, 19 },
		{ 192, 313, 19 },
		{ 192, 314, 19 },
		{ 192, 315, 19 },
		{ 192, 316, 19 },
		{ 192, 317, 19 },
		{ 192, 318, 19 },
		{ 192, 319, 19 },
		{ 192, 320, 19 },
		{ 192, 321, 19 },
		{ 193, 290, 19 },
		{ 193, 291, 19 },
		{ 193, 292, 19 },
		{ 193, 293, 19 },
		{ 193, 294, 19 },
		{ 193, 295, 19 },
		{ 193, 296, 19 },
		{ 193, 297, 19 },
		{ 193, 298, 19 },
		{ 193, 299, 19 },
		{ 193, 300, 19 },
		{ 193, 301, 19 },
		{ 193, 302, 19 },
		{ 193, 303, 19 },
		{ 193, 304, 19 },
		{ 193, 305, 19 },
		{ 193, 306, 19 },
		{ 193, 307, 19 },
		{ 193, 308, 19 },
		{ 193, 309, 19 },
		{ 193, 310, 19 },
		{ 193, 311, 19 },
		{ 193, 312, 19 },
		{ 193, 313, 19 },
		{ 193, 314, 19 },
		{ 193, 315, 19 },
		{ 193, 316, 19 },
		{ 193, 317, 19 },
		{ 193, 318, 19 },
		{ 193, 319, 19 },
		{ 193, 320, 19 },
		{ 194, 289, 19 },
		{ 194, 290, 19 },
		{ 194, 291, 19 },
		{ 194, 292, 19 },
		{ 194, 293, 19 },
		{ 194, 294, 19 },
		{ 194, 295, 19 },
		{ 194, 296, 19 },
		{ 194, 297, 19 },
		{ 194, 298, 19 },
		{ 194, 299, 19 },
		{ 194, 300, 19 },
		{ 194, 301, 19 },
		{ 194, 302, 19 },
		{ 194, 303, 19 },
		{ 194, 304, 19 },
		{ 194, 305, 19 },
		{ 194, 306, 19 },
		{ 194, 307, 19 },
		{ 194, 308, 19 },
		{ 194, 309, 19 },
		{ 194, 310, 19 },
		{ 194, 311, 19 },
		{ 194, 312, 19 },
		{ 194, 313, 19 },
		{ 194, 314, 19 },
		{ 194, 315, 19 },
		{ 194, 316, 19 },
		{ 194, 317, 19 },
		{ 194, 318, 19 },
		{ 194, 319, 19 },
		{ 195, 289, 19 },
		{ 195, 290, 19 },
		{ 195, 291, 19 },
		{ 195, 292, 19 },
		{ 195, 293, 19 },
		{ 195, 294, 19 },
		{ 195, 295, 19 },
		{ 195, 296, 19 },
		{ 195, 297, 19 },
		{ 195, 298, 19 },
		{ 195, 299, 19 },
		{ 195, 300, 19 },
		{ 195, 301, 19 },
		{ 195, 302, 19 },
		{ 195, 303, 19 },
		{ 195, 304, 19 },
		{ 195, 305, 19 },
		{ 195, 306, 19 },
		{ 195, 307, 19 },
		{ 195, 308, 19 },
		{ 195, 309, 19 },
		{ 195, 310, 19 },
		{ 195, 311, 19 },
		{ 195, 312, 19 },
		{ 195, 313, 19 },
		{ 195, 314, 19 },
		{ 195, 315, 19 },
		{ 195, 316, 19 },
		{ 195, 317, 19 },
		{ 195, 318, 19 },
		{ 196, 288, 19 },
		{ 196, 289, 19 },
		{ 196, 290, 19 },
		{ 196, 291, 19 },
		{ 196, 292, 19 },
		{ 196, 293, 19 },
		{ 196, 294, 19 },
		{ 196, 295, 19 },
		{ 196, 296, 19 },
		{ 196, 297, 19 },
		{ 196, 298, 19 },
		{ 196, 299, 19 },
		{ 196, 300, 19 },
		{ 196, 301, 19 },
		{ 196, 302, 19 },
		{ 196, 303, 19 },
		{ 196, 304, 19 },
		{ 196, 305, 19 },
		{ 196, 306, 19 },
		{ 196, 307, 19 },
		{ 196, 308, 19 },
		{ 196, 309, 19 },
		{ 196, 310, 19 },
		{ 196, 311, 19 },
		{ 196, 312, 19 },
		{ 196, 313, 19 },
		{ 196, 314, 19 },
		{ 196, 315, 19 },
		{ 196, 316, 19 },
		{ 196, 317, 19 },
		{ 197, 288, 19 },
		{ 197, 289, 19 },
		{ 197, 290, 19 },
		{ 197, 291, 19 },
		{ 197, 292, 19 },
		{ 197, 293, 19 },
		{ 197, 294, 19 },
		{ 197, 295, 19 },
		{ 197, 296, 19 },
		{ 197, 297, 19 },
		{ 197, 298, 19 },
		{ 197, 299, 19 },
		{ 197, 300, 19 },
		{ 197, 301, 19 },
		{ 197, 302, 19 },
		{ 197, 303, 19 },
		{ 197, 304, 19 },
		{ 197, 305, 19 },
		{ 197, 306, 19 },
		{ 197, 307, 19 },
		{ 197, 308, 19 },
		{ 197, 309, 19 },
		{ 197, 310, 19 },
		{ 197, 311, 19 },
		{ 197, 312, 19 },
		{ 197, 313, 19 },
		{ 197, 314, 19 },
		{ 197, 315, 19 },
		{ 197, 316, 19 },
		{ 198, 287, 19 },
		{ 198, 288, 19 },
		{ 198, 289, 19 },
		{ 198, 290, 19 },
		{ 198, 291, 19 },
		{ 198, 292, 19 },
		{ 198, 293, 19 },
		{ 198, 294, 19 },
		{ 198, 295, 19 },
		{ 198, 296, 19 },
		{ 198, 297, 19 },
		{ 198, 298, 19 },
		{ 198, 299, 19 },
		{ 198, 300, 19 },
		{ 198, 301, 19 },
		{ 198, 302, 19 },
		{ 198, 303, 19 },
		{ 198, 304, 19 },
		{ 198, 305, 19 },
		{ 198, 306, 19 },
		{ 198, 307, 19 },
		{ 198, 308, 19 },
		{ 198, 309, 19 },
		{ 198, 310, 19 },
		{ 198, 311, 19 },
		{ 198, 312, 19 },
		{ 198, 313, 19 },
		{ 198, 314, 19 },
		{ 198, 315, 19 },
		{ 199, 287, 19 },
		{ 199, 288, 19 },
		{ 199, 289, 19 },
		{ 199, 290, 19 },
		{ 199, 291, 19 },
		{ 199, 292, 19 },
		{ 199, 293, 19 },
		{ 199, 294, 19 },
		{ 199, 295, 19 },
		{ 199, 296, 19 },
		{ 199, 297, 19 },
		{ 199, 298, 19 },
		{ 199, 299, 19 },
		{ 199, 300, 19 },
		{ 199, 301, 19 },
		{ 199, 302, 19 },
		{ 199, 303, 19 },
		{ 199, 304, 19 },
		{ 199, 305, 19 },
		{ 199, 306, 19 },
		{ 199, 307, 19 },
		{ 199, 308, 19 },
		{ 199, 309, 19 },
		{ 199, 310, 19 },
		{ 199, 311, 19 },
		{ 199, 312, 19 },
		{ 199, 313, 19 },
		{ 199, 314, 19 },
		{ 200, 286, 19 },
		{ 200, 287, 19 },
		{ 200, 288, 19 },
		{ 200, 289, 19 },
		{ 200, 290, 19 },
		{ 200, 291, 19 },
		{ 200, 292, 19 },
		{ 200, 293, 19 },
		{ 200, 294, 19 },
		{ 200, 295, 19 },
		{ 200, 296, 19 },
		{ 200, 297, 19 },
		{ 200, 298, 19 },
		{ 200, 299, 19 },
		{ 200, 300, 19 },
		{ 200, 301, 19 },
		{ 200, 302, 19 },
		{ 200, 303, 19 },
		{ 200, 304, 19 },
		{ 200, 305, 19 },
		{ 200, 306, 19 },
		{ 200, 307, 19 },
		{ 200, 308, 19 },
		{ 200, 309, 19 },
		{ 200, 310, 19 },
		{ 200, 311, 19 },
		{ 200, 312, 19 },
		{ 200, 313, 19 },
		{ 201, 286, 19 },
		{ 201, 287, 19 },
		{ 201, 288, 19 },
		{ 201, 289, 19 },
		{ 201, 290, 19 },
		{ 201, 291, 19 },
		{ 201, 292, 19 },
		{ 201, 293, 19 },
		{ 201, 294, 19 },
		{ 201, 295, 19 },
		{ 201, 296, 19 },
		{ 201, 297, 19 },
		{ 201, 298, 19 },
		{ 201, 299, 19 },
		{ 201, 300, 19 },
		{ 201, 301, 19 },
		{ 201, 302, 19 },
		{ 201, 303, 19 },
		{ 201, 304, 19 },
		{ 201, 305, 19 },
		{ 201, 306, 19 },
		{ 201, 307, 19 },
		{ 201, 308, 19 },
		{ 201, 309, 19 },
		{ 201, 310, 19 },
		{ 201, 311, 19 },
		{ 201, 312, 19 },
		{ 202, 285, 19 },
		{ 202, 286, 19 },
		{ 202, 287, 19 },
		{ 202, 288, 19 },
		{ 202, 289, 19 },
		{ 202, 290, 19 },
		{ 202, 291, 19 },
		{ 202, 292, 19 },
		{ 202, 293, 19 },
		{ 202, 294, 19 },
		{ 202, 295, 19 },
		{ 202, 296, 19 },
		{ 202, 297, 19 },
		{ 202, 298, 19 },
		{ 202, 299, 19 },
		{ 202, 300, 19 },
		{ 202, 301, 19 },
		{ 202, 302, 19 },
		{ 202, 303, 19 },
		{ 202, 304, 19 },
		{ 202, 305, 19 },
		{ 202, 306, 19 },
		{ 202, 307, 19 },
		{ 202, 308, 19 },
		{ 202, 309, 19 },
		{ 202, 310, 19 },
		{ 202, 311, 19 },
		{ 203, 285, 19 },
		{ 203, 286, 19 },
		{ 203, 287, 19 },
		{ 203, 288, 19 },
		{ 203, 289, 19 },
		{ 203, 290, 19 },
		{ 203, 291, 19 },
		{ 203, 292, 19 },
		{ 203, 293, 19 },
		{ 203, 294, 19 },
		{ 203, 295, 19 },
		{ 203, 296, 19 },
		{ 203, 297, 19 },
		{ 203, 298, 19 },
		{ 203, 299, 19 },
		{ 203, 300, 19 },
		{ 203, 301, 19 },
		{ 203, 302, 19 },
		{ 203, 303, 19 },
		{ 203, 304, 19 },
		{ 203, 305, 19 },
		{ 203, 306, 19 },
		{ 203, 307, 19 },
		{ 203, 308, 19 },
		{ 203, 309, 19 },
		{ 203, 310, 19 },
		{ 204, 284, 19 },
		{ 204, 285, 19 },
		{ 204, 286, 19 },
		{ 204, 287, 19 },
		{ 204, 288, 19 },
		{ 204, 289, 19 },
		{ 204, 290, 19 },
		{ 204, 291, 19 },
		{ 204, 292, 19 },
		{ 204, 293, 19 },
		{ 204, 294, 19 },
		{ 204, 295, 19 },
		{ 204, 296, 19 },
		{ 204, 297, 19 },
		{ 204, 298, 19 },
		{ 204, 299, 19 },
		{ 204, 300, 19 },
		{ 204, 301, 19 },
		{ 204, 302, 19 },
		{ 204, 303, 19 },
		{ 204, 304, 19 },
		{ 204, 305, 19 },
		{ 204, 306, 19 },
		{ 204, 307, 19 },
		{ 204, 308, 19 },
		{ 204, 309, 19 },
		{ 205, 284, 19 },
		{ 205, 285, 19 },
		{ 205, 286, 19 },
		{ 205, 287, 19 },
		{ 205, 288, 19 },
		{ 205, 289, 19 },
		{ 205, 290, 19 },
		{ 205, 291, 19 },
		{ 205, 292, 19 },
		{ 205, 293, 19 },
		{ 205, 294, 19 },
		{ 205, 295, 19 },
		{ 205, 296, 19 },
		{ 205, 297, 19 },
		{ 205, 298, 19 },
		{ 205, 299, 19 },
		{ 205, 300, 19 },
		{ 205, 301, 19 },
		{ 205, 302, 19 },
		{ 205, 303, 19 },
		{ 205, 304, 19 },
		{ 205, 305, 19 },
		{ 205, 306, 19 },
		{ 205, 307, 19 },
		{ 205, 308, 19 },
		{ 206, 283, 19 },
		{ 206, 284, 19 },
		{ 206, 285, 19 },
		{ 206, 286, 19 },
		{ 206, 287, 19 },
		{ 206, 288, 19 },
		{ 206, 289, 19 },
		{ 206, 290, 19 },
		{ 206, 291, 19 },
		{ 206, 292, 19 },
		{ 206, 293, 19 },
		{ 206, 294, 19 },
		{ 206, 295, 19 },
		{ 206, 296, 19 },
		{ 206, 297, 19 },
		{ 206, 298, 19 },
		{ 206, 299, 19 },
		{ 206, 300, 19 },
		{ 206, 301, 19 },
		{ 206, 302, 19 },
		{ 206, 303, 19 },
		{ 206, 304, 19 },
		{ 206, 305, 19 },
		{ 206, 306, 19 },
		{ 206, 307, 19 },
		{ 207, 283, 19 },
		{ 207, 284, 19 },
		{ 207, 285, 19 },
		{ 207, 286, 19 },
		{ 207, 287, 19 },
		{ 207, 288, 19 },
		{ 207, 289, 19 },
		{ 207, 290, 19 },
		{ 207, 291, 19 },
		{ 207, 292, 19 },
		{ 207, 293, 19 },
		{ 207, 294, 19 },
		{ 207, 295, 19 },
		{ 207, 296, 19 },
		{ 207, 297, 19 },
		{ 207, 298, 19 },
		{ 207, 299, 19 },
		{ 207, 300, 19 },
		{ 207, 301, 19 },
		{ 207, 302, 19 },
		{ 207, 303, 19 },
		{ 207, 304, 19 },
		{ 207, 305, 19 },
		{ 207, 306, 19 },
		{ 208, 282, 19 },
		{ 208, 283, 19 },
		{ 208, 284, 19 },
		{ 208, 285, 19 },
		{ 208, 286, 19 },
		{ 208, 287, 19 },
		{ 208, 288, 19 },
		{ 208, 289, 19 },
		{ 208, 290, 19 },
		{ 208, 291, 19 },
		{ 208, 292, 19 },
		{ 208, 293, 19 },
		{ 208, 294, 19 },
		{ 208, 295, 19 },
		{ 208, 296, 19 },
		{ 208, 297, 19 },
		{ 208, 298, 19 },
		{ 208, 299, 19 },
		{ 208, 300, 19 },
		{ 208, 301, 19 },
		{ 208, 302, 19 },
		{ 208, 303, 19 },
		{ 208, 304, 19 },
		{ 208, 305, 19 },
		{ 209, 282, 19 },
		{ 209, 283, 19 },
		{ 209, 284, 19 },
		{ 209, 285, 19 },
		{ 209, 286, 19 },
		{ 209, 287, 19 },
		{ 209, 288, 19 },
		{ 209, 289, 19 },
		{ 209, 290, 19 },
		{ 209, 291, 19 },
		{ 209, 292, 19 },
		{ 209, 293, 19 },
		{ 209, 294, 19 },
		{ 209, 295, 19 },
		{ 209, 296, 19 },
		{ 209, 297, 19 },
		{ 209, 298, 19 },
		{ 209, 299, 19 },
		{ 209, 300, 19 },
		{ 209, 301, 19 },
		{ 209, 302, 19 },
		{ 209, 303, 19 },
		{ 209, 304, 19 },
		{ 210, 281, 19 },
		{ 210, 282, 19 },
		{ 210, 283, 19 },
		{ 210, 284, 19 },
		{ 210, 285, 19 },
		{ 210, 286, 19 },
		{ 210, 287, 19 },
		{ 210, 288, 19 },
		{ 210, 289, 19 },
		{ 210, 290, 19 },
		{ 210, 291, 19 },
		{ 210, 292, 19 },
		{ 210, 293, 19 },
		{ 210, 294, 19 },
		{ 210, 295, 19 },
		{ 210, 296, 19 },
		{ 210, 297, 19 },
		{ 210, 298, 19 },
		{ 210, 299, 19 },
		{ 210, 300, 19 },
		{ 210, 301, 19 },
		{ 210, 302, 19 },
		{ 210, 303, 19 },
		{ 211, 281, 19 },
		{ 211, 282, 19 },
		{ 211, 283, 19 },
		{ 211, 284, 19 },
		{ 211, 285, 19 },
		{ 211, 286, 19 },
		{ 211, 287, 19 },
		{ 211, 288, 19 },
		{ 211, 289, 19 },
		{ 211, 290, 19 },
		{ 211, 291, 19 },
		{ 211, 292, 19 },
		{ 211, 293, 19 },
		{ 211, 294, 19 },
		{ 211, 295, 19 },
		{ 211, 296, 19 },
		{ 211, 297, 19 },
		{ 211, 298, 19 },
		{ 211, 299, 19 },
		{ 211, 300, 19 },
		{ 211, 301, 19 },
		{ 211, 302, 19 },
		{ 212, 280, 19 },
		{ 212, 281, 19 },
		{ 212, 282, 19 },
		{ 212, 283, 19 },
		{ 212, 284, 19 },
		{ 212, 285, 19 },
		{ 212, 286, 19 },
		{ 212, 287, 19 },
		{ 212, 288, 19 },
		{ 212, 289, 19 },
		{ 212, 290, 19 },
		{ 212, 291, 19 },
		{ 212, 292, 19 },
		{ 212, 293, 19 },
		{ 212, 294, 19 },
		{ 212, 295, 19 },
		{ 212, 296, 19 },
		{ 212, 297, 19 },
		{ 212, 298, 19 },
		{ 212, 299, 19 },
		{ 212, 300, 19 },
		{ 212, 301, 19 },
		{ 213, 280, 19 },
		{ 213, 281, 19 },
		{ 213, 282, 19 },
		{ 213, 283, 19 },
		{ 213, 284, 19 },
		{ 213, 285, 19 },
		{ 213, 286, 19 },
		{ 213, 287, 19 },
		{ 213, 288, 19 },
		{ 213, 289, 19 },
		{ 213, 290, 19 },
		{ 213, 291, 19 },
		{ 213, 292, 19 },
		{ 213, 293, 19 },
		{ 213, 294, 19 },
		{ 213, 295, 19 },
		{ 213, 296, 19 },
		{ 213, 297, 19 },
		{ 213, 298, 19 },
		{ 213, 299, 19 },
		{ 213, 300, 19 },
		{ 214, 279, 19 },
		{ 214, 280, 19 },
		{ 214, 281, 19 },
		{ 214, 282, 19 },
		{ 214, 283, 19 },
		{ 214, 284, 19 },
		{ 214, 285, 19 },
		{ 214, 286, 19 },
		{ 214, 287, 19 },
		{ 214, 288, 19 },
		{ 214, 289, 19 },
		{ 214, 290, 19 },
		{ 214, 291, 19 },
		{ 214, 292, 19 },
		{ 214, 293, 19 },
		{ 214, 294, 19 },
		{ 214, 295, 19 },
		{ 214, 296, 19 },
		{ 214, 297, 19 },
		{ 214, 298, 19 },
		{ 214, 299, 19 },
		{ 215, 279, 19 },
		{ 215, 280, 19 },
		{ 215, 281, 19 },
		{ 215, 282, 19 },
		{ 215, 283, 19 },
		{ 215, 284, 19 },
		{ 215, 285, 19 },
		{ 215, 286, 19 },
		{ 215, 287, 19 },
		{ 215, 288, 19 },
		{ 215, 289, 19 },
		{ 215, 290, 19 },
		{ 215, 291, 19 },
		{ 215, 292, 19 },
		{ 215, 293, 19 },
		{ 215, 294, 19 },
		{ 215, 295, 19 },
		{ 215, 296, 19 },
		{ 215, 297, 19 },
		{ 215, 298, 19 },
		{ 216, 278, 19 },
		{ 216, 279, 19 },
		{ 216, 280, 19 },
		{ 216, 281, 19 },
		{ 216, 282, 19 },
		{ 216, 283, 19 },
		{ 216, 284, 19 },
		{ 216, 285, 19 },
		{ 216, 286, 19 },
		{ 216, 287, 19 },
		{ 216, 288, 19 },
		{ 216, 289, 19 },
		{ 216, 290, 19 },
		{ 216, 291, 19 },
		{ 216, 292, 19 },
		{ 216, 293, 19 },
		{ 216, 294, 19 },
		{ 216, 295, 19 },
		{ 216, 296, 19 },
		{ 216, 297, 19 },
		{ 217, 278, 19 },
		{ 217, 279, 19 },
		{ 217, 280, 19 },
		{ 217, 281, 19 },
		{ 217, 282, 19 },
		{ 217, 283, 19 },
		{ 217, 284, 19 },
		{ 217, 285, 19 },
		{ 217, 286, 19 },
		{ 217, 287, 19 },
		{ 217, 288, 19 },
		{ 217, 289, 19 },
		{ 217, 290, 19 },
		{ 217, 291, 19 },
		{ 217, 292, 19 },
		{ 217, 293, 19 },
		{ 217, 294, 19 },
		{ 217, 295, 19 },
		{ 217, 296, 19 },
		{ 218, 277, 19 },
		{ 218, 278, 19 },
		{ 218, 279, 19 },
		{ 218, 280, 19 },
		{ 218, 281, 19 },
		{ 218, 282, 19 },
		{ 218, 283, 19 },
		{ 218, 284, 19 },
		{ 218, 285, 19 },
		{ 218, 286, 19 },
		{ 218, 287, 19 },
		{ 218, 288, 19 },
		{ 218, 289, 19 },
		{ 218, 290, 19 },
		{ 218, 291, 19 },
		{ 218, 292, 19 },
		{ 218, 293, 19 },
		{ 218, 294, 19 },
		{ 218, 295, 19 },
		{ 219, 277, 19 },
		{ 219, 278, 19 },
		{ 219, 279, 19 },
		{ 219, 280, 19 },
		{ 219, 281, 19 },
		{ 219, 282, 19 },
		{ 219, 283, 19 },
		{ 219, 284, 19 },
		{ 219, 285, 19 },
		{ 219, 286, 19 },
		{ 219, 287, 19 },
		{ 219, 288, 19 },
		{ 219, 289, 19 },
		{ 219, 290, 19 },
		{ 219, 291, 19 },
		{ 219, 292, 19 },
		{ 219, 293, 19 },
		{ 219, 294, 19 },
		{ 220, 276, 19 },
		{ 220, 277, 19 },
		{ 220, 278, 19 },
		{ 220, 279, 19 },
		{ 220, 280, 19 },
		{ 220, 281, 19 },
		{ 220, 282, 19 },
		{ 220, 283, 19 },
		{ 220, 284, 19 },
		{ 220, 285, 19 },
		{ 220, 286, 19 },
		{ 220, 287, 19 },
		{ 220, 288, 19 },
		{ 220, 289, 19 },
		{ 220, 290, 19 },
		{ 220, 291, 19 },
		{ 220, 292, 19 },
		{ 220, 293, 19 },
		{ 221, 276, 19 },
		{ 221, 277, 19 },
		{ 221, 278, 19 },
		{ 221, 279, 19 },
		{ 221, 280, 19 },
		{ 221, 281, 19 },
		{ 221, 282, 19 },
		{ 221, 283, 19 },
		{ 221, 284, 19 },
		{ 221, 285, 19 },
		{ 221, 286, 19 },
		{ 221, 287, 19 },
		{ 221, 288, 19 },
		{ 221, 289, 19 },
		{ 221, 290, 19 },
		{ 221, 291, 19 },
		{ 221, 292, 19 },
		{ 222, 275, 19 },
		{ 222, 276, 19 },
		{ 222, 277, 19 },
		{ 222, 278, 19 },
		{ 222, 279, 19 },
		{ 222, 280, 19 },
		{ 222, 281, 19 },
		{ 222, 282, 19 },
		{ 222, 283, 19 },
		{ 222, 284, 19 },
		{ 222, 285, 19 },
		{ 222, 286, 19 },
		{ 222, 287, 19 },
		{ 222, 288, 19 },
		{ 222, 289, 19 },
		{ 222, 290, 19 },
		{ 222, 291, 19 },
		{ 223, 275, 19 },
		{ 223, 276, 19 },
		{ 223, 277, 19 },
		{ 223, 278, 19 },
		{ 223, 279, 19 },
		{ 223, 280, 19 },
		{ 223, 281, 19 },
		{ 223, 282, 19 },
		{ 223, 283, 19 },
		{ 223, 284, 19 },
		{ 223, 285, 19 },
		{ 223, 286, 19 },
		{ 223, 287, 19 },
		{ 223, 288, 19 },
		{ 223, 289, 19 },
		{ 223, 290, 19 },
		{ 224, 274, 19 },
		{ 224, 275, 19 },
		{ 224, 276, 19 },
		{ 224, 277, 19 },
		{ 224, 278, 19 },
		{ 224, 279, 19 },
		{ 224, 280, 19 },
		{ 224, 281, 19 },
		{ 224, 282, 19 },
		{ 224, 283, 19 },
		{ 224, 284, 19 },
		{ 224, 285, 19 },
		{ 224, 286, 19 },
		{ 224, 287, 19 },
		{ 224, 288, 19 },
		{ 224, 289, 19 },
		{ 225, 274, 19 },
		{ 225, 275, 19 },
		{ 225, 276, 19 },
		{ 225, 277, 19 },
		{ 225, 278, 19 },
		{ 225, 279, 19 },
		{ 225, 280, 19 },
		{ 225, 281, 19 },
		{ 225, 282, 19 },
		{ 225, 283, 19 },
		{ 225, 284, 19 },
		{ 225, 285, 19 },
		{ 225, 286, 19 },
		{ 225, 287, 19 },
		{ 225, 288, 19 },
		{ 226, 273, 19 },
		{ 226, 274, 19 },
		{ 226, 275, 19 },
		{ 226, 276, 19 },
		{ 226, 277, 19 },
		{ 226, 278, 19 },
		{ 226, 279, 19 },
		{ 226, 280, 19 },
		{ 226, 281, 19 },
		{ 226, 282, 19 },
		{ 226, 283, 19 },
		{ 226, 284, 19 },
		{ 226, 285, 19 },
		{ 226, 286, 19 },
		{ 226, 287, 19 },
		{ 227, 273, 19 },
		{ 227, 274, 19 },
		{ 227, 275, 19 },
		{ 227, 276, 19 },
		{ 227, 277, 19 },
		{ 227, 278, 19 },
		{ 227, 279, 19 },
		{ 227, 280, 19 },
		{ 227, 281, 19 },
		{ 227, 282, 19 },
		{ 227, 283, 19 },
		{ 227, 284, 19 },
		{ 227, 285, 19 },
		{ 227, 286, 19 },
		{ 228, 272, 19 },
		{ 228, 273, 19 },
		{ 228, 274, 19 },
		{ 228, 275, 19 },
		{ 228, 276, 19 },
		{ 228, 277, 19 },
		{ 228, 278, 19 },
		{ 228, 279, 19 },
		{ 228, 280, 19 },
		{ 228, 281, 19 },
		{ 228, 282, 19 },
		{ 228, 283, 19 },
		{ 228, 284, 19 },
		{ 228, 285, 19 },
		{ 229, 272, 19 },
		{ 229, 273, 19 },
		{ 229, 274, 19 },
		{ 229, 275, 19 },
		{ 229, 276, 19 },
		{ 229, 277, 19 },
		{ 229, 278, 19 },
		{ 229, 279, 19 },
		{ 229, 280, 19 },
		{ 229, 281, 19 },
		{ 229, 282, 19 },
		{ 229, 283, 19 },
		{ 229, 284, 19 },
		{ 230, 271, 19 },
		{ 230, 272, 19 },
		{ 230, 273, 19 },
		{ 230, 274, 19 },
		{ 230, 275, 19 },
		{ 230, 276, 19 },
		{ 230, 277, 19 },
		{ 230, 278, 19 },
		{ 230, 279, 19 },
		{ 230, 280, 19 },
		{ 230, 281, 19 },
		{ 230, 282, 19 },
		{ 230, 283, 19 },
		{ 231, 271, 19 },
		{ 231, 272, 19 },
		{ 231, 273, 19 },
		{ 231, 274, 19 },
		{ 231, 275, 19 },
		{ 231, 276, 19 },
		{ 231, 277, 19 },
		{ 231, 278, 19 },
		{ 231, 279, 19 },
		{ 231, 280, 19 },
		{ 231, 281, 19 },
		{ 231, 282, 19 },
		{ 232, 270, 19 },
		{ 232, 271, 19 },
		{ 232, 272, 19 },
		{ 232, 273, 19 },
		{ 232, 274, 19 },
		{ 232, 275, 19 },
		{ 232, 276, 19 },
		{ 232, 277, 19 },
		{ 232, 278, 19 },
		{ 232, 279, 19 },
		{ 232, 280, 19 },
		{ 232, 281, 19 },
		{ 233, 270, 19 },
		{ 233, 271, 19 },
		{ 233, 272, 19 },
		{ 233, 273, 19 },
		{ 233, 274, 19 },
		{ 233, 275, 19 },
		{ 233, 276, 19 },
		{ 233, 277, 19 },
		{ 233, 278, 19 },
		{ 233, 279, 19 },
		{ 233, 280, 19 },
		{ 234, 269, 19 },
		{ 234, 270, 19 },
		{ 234, 271, 19 },
		{ 234, 272, 19 },
		{ 234, 273, 19 },
		{ 234, 274, 19 },
		{ 234, 275, 19 },
		{ 234, 276, 19 },
		{ 234, 277, 19 },
		{ 234, 278, 19 },
		{ 234, 279, 19 },
		{ 235, 269, 19 },
		{ 235, 270, 19 },
		{ 235, 271, 19 },
		{ 235, 272, 19 },
		{ 235, 273, 19 },
		{ 235, 274, 19 },
		{ 235, 275, 19 },
		{ 235, 276, 19 },
		{ 235, 277, 19 },
		{ 235, 278, 19 },
		{ 236, 268, 19 },
		{ 236, 269, 19 },
		{ 236, 270, 19 },
		{ 236, 271, 19 },
		{ 236, 272, 19 },
		{ 236, 273, 19 },
		{ 236, 274, 19 },
		{ 236, 275, 19 },
		{ 236, 276, 19 },
		{ 236, 277, 19 },
		{ 237, 268, 19 },
		{ 237, 269, 19 },
		{ 237, 270, 19 },
		{ 237, 271, 19 },
		{ 237, 272, 19 },
		{ 237, 273, 19 },
		{ 237, 274, 19 },
		{ 237, 275, 19 },
		{ 237, 276, 19 },
		{ 238, 267, 19 },
		{ 238, 268, 19 },
		{ 238, 269, 19 },
		{ 238, 270, 19 },
		{ 238, 271, 19 },
		{ 238, 272, 19 },
		{ 238, 273, 19 },
		{ 238, 274, 19 },
		{ 238, 275, 19 },
		{ 239, 267, 19 },
		{ 239, 268, 19 },
		{ 239, 269, 19 },
		{ 239, 270, 19 },
		{ 239, 271, 19 },
		{ 239, 272, 19 },
		{ 239, 273, 19 },
		{ 239, 274, 19 },
		{ 240, 266, 19 },
		{ 240, 267, 19 },
		{ 240, 268, 19 },
		{ 240, 269, 19 },
		{ 240, 270, 19 },
		{ 240, 271, 19 },
		{ 240, 272, 19 },
		{ 240, 273, 19 },
		{ 274, 241, 19 },
		{ 274, 242, 19 },
		{ 274, 243, 19 },
		{ 274, 244, 19 },
		{ 274, 245, 19 },
		{ 274, 246, 19 },
		{ 274, 247, 19 },
		{ 274, 248, 19 },
		{ 275, 240, 19 },
		{ 275, 241, 19 },
		{ 275, 242, 19 },
		{ 275, 243, 19 },
		{ 275, 244, 19 },
		{ 275, 245, 19 },
		{ 275, 246, 19 },
		{ 275, 247, 19 },
		{ 276, 239, 19 },
		{ 276, 240, 19 },
		{ 276, 241, 19 },
		{ 276, 242, 19 },
		{ 276, 243, 19 },
		{ 276, 244, 19 },
		{ 276, 245, 19 },
		{ 276, 246, 19 },
		{ 276, 247, 19 },
		{ 277, 238, 19 },
		{ 277, 239, 19 },
		{ 277, 240, 19 },
		{ 277, 241, 19 },
		{ 277, 242, 19 },
		{ 277, 243, 19 },
		{ 277, 244, 19 },
		{ 277, 245, 19 },
		{ 277, 246, 19 },
		{ 278, 237, 19 },
		{ 278, 238, 19 },
		{ 278, 239, 19 },
		{ 278, 240, 19 },
		{ 278, 241, 19 },
		{ 278, 242, 19 },
		{ 278, 243, 19 },
		{ 278, 244, 19 },
		{ 278, 245, 19 },
		{ 278, 246, 19 },
		{ 279, 236, 19 },
		{ 279, 237, 19 },
		{ 279, 238, 19 },
		{ 279, 239, 19 },
		{ 279, 240, 19 },
		{ 279, 241, 19 },
		{ 279, 242, 19 },
		{ 279, 243, 19 },
		{ 279, 244, 19 },
		{ 279, 245, 19 },
		{ 280, 235, 19 },
		{ 280, 236, 19 },
		{ 280, 237, 19 },
		{ 280, 238, 19 },
		{ 280, 239, 19 },
		{ 280, 240, 19 },
		{ 280, 241, 19 },
		{ 280, 242, 19 },
		{ 280, 243, 19 },
		{ 280, 244, 19 },
		{ 280, 245, 19 },
		{ 281, 234, 19 },
		{ 281, 235, 19 },
		{ 281, 236, 19 },
		{ 281, 237, 19 },
		{ 281, 238, 19 },
		{ 281, 239, 19 },
		{ 281, 240, 19 },
		{ 281, 241, 19 },
		{ 281, 242, 19 },
		{ 281, 243, 19 },
		{ 281, 244, 19 },
		{ 282, 233, 19 },
		{ 282, 234, 19 },
		{ 282, 235, 19 },
		{ 282, 236, 19 },
		{ 282, 237, 19 },
		{ 282, 238, 19 },
		{ 282, 239, 19 },
		{ 282, 240, 19 },
		{ 282, 241, 19 },
		{ 282, 242, 19 },
		{ 282, 243, 19 },
		{ 282, 244, 19 },
		{ 283, 232, 19 },
		{ 283, 233, 19 },
		{ 283, 234, 19 },
		{ 283, 235, 19 },
		{ 283, 236, 19 },
		{ 283, 237, 19 },
		{ 283, 238, 19 },
		{ 283, 239, 19 },
		{ 283, 240, 19 },
		{ 283, 241, 19 },
		{ 283, 242, 19 },
		{ 283, 243, 19 },
		{ 284, 231, 19 },
		{ 284, 232, 19 },
		{ 284, 233, 19 },
		{ 284, 234, 19 },
		{ 284, 235, 19 },
		{ 284, 236, 19 },
		{ 284, 237, 19 },
		{ 284, 238, 19 },
		{ 284, 239, 19 },
		{ 284, 240, 19 },
		{ 284, 241, 19 },
		{ 284, 242, 19 },
		{ 284, 243, 19 },
		{ 285, 230, 19 },
		{ 285, 231, 19 },
		{ 285, 232, 19 },
		{ 285, 233, 19 },
		{ 285, 234, 19 },
		{ 285, 235, 19 },
		{ 285, 236, 19 },
		{ 285, 237, 19 },
		{ 285, 238, 19 },
		{ 285, 239, 19 },
		{ 285, 240, 19 },
		{ 285, 241, 19 },
		{ 285, 242, 19 },
		{ 286, 229, 19 },
		{ 286, 230, 19 },
		{ 286, 231, 19 },
		{ 286, 232, 19 },
		{ 286, 233, 19 },
		{ 286, 234, 19 },
		{ 286, 235, 19 },
		{ 286, 236, 19 },
		{ 286, 237, 19 },
		{ 286, 238, 19 },
		{ 286, 239, 19 },
		{ 286, 240, 19 },
		{ 286, 241, 19 },
		{ 286, 242, 19 },
		{ 287, 228, 19 },
		{ 287, 229, 19 },
		{ 287, 230, 19 },
		{ 287, 231, 19 },
		{ 287, 232, 19 },
		{ 287, 233, 19 },
		{ 287, 234, 19 },
		{ 287, 235, 19 },
		{ 287, 236, 19 },
		{ 287, 237, 19 },
		{ 287, 238, 19 },
		{ 287, 239, 19 },
		{ 287, 240, 19 },
		{ 287, 241, 19 },
		{ 288, 227, 19 },
		{ 288, 228, 19 },
		{ 288, 229, 19 },
		{ 288, 230, 19 },
		{ 288, 231, 19 },
		{ 288, 232, 19 },
		{ 288, 233, 19 },
		{ 288, 234, 19 },
		{ 288, 235, 19 },
		{ 288, 236, 19 },
		{ 288, 237, 19 },
		{ 288, 238, 19 },
		{ 288, 239, 19 },
		{ 288, 240, 19 },
		{ 288, 241, 19 },
		{ 289, 226, 19 },
		{ 289, 227, 19 },
		{ 289, 228, 19 },
		{ 289, 229, 19 },
		{ 289, 230, 19 },
		{ 289, 231, 19 },
		{ 289, 232, 19 },
		{ 289, 233, 19 },
		{ 289, 234, 19 },
		{ 289, 235, 19 },
		{ 289, 236, 19 },
		{ 289, 237, 19 },
		{ 289, 238, 19 },
		{ 289, 239, 19 },
		{ 289, 240, 19 },
		{ 290, 225, 19 },
		{ 290, 226, 19 },
		{ 290, 227, 19 },
		{ 290, 228, 19 },
		{ 290, 229, 19 },
		{ 290, 230, 19 },
		{ 290, 231, 19 },
		{ 290, 232, 19 },
		{ 290, 233, 19 },
		{ 290, 234, 19 },
		{ 290, 235, 19 },
		{ 290, 236, 19 },
		{ 290, 237, 19 },
		{ 290, 238, 19 },
		{ 290, 239, 19 },
		{ 290, 240, 19 },
		{ 291, 224, 19 },
		{ 291, 225, 19 },
		{ 291, 226, 19 },
		{ 291, 227, 19 },
		{ 291, 228, 19 },
		{ 291, 229, 19 },
		{ 291, 230, 19 },
		{ 291, 231, 19 },
		{ 291, 232, 19 },
		{ 291, 233, 19 },
		{ 291, 234, 19 },
		{ 291, 235, 19 },
		{ 291, 236, 19 },
		{ 291, 237, 19 },
		{ 291, 238, 19 },
		{ 291, 239, 19 },
		{ 292, 223, 19 },
		{ 292, 224, 19 },
		{ 292, 225, 19 },
		{ 292, 226, 19 },
		{ 292, 227, 19 },
		{ 292, 228, 19 },
		{ 292, 229, 19 },
		{ 292, 230, 19 },
		{ 292, 231, 19 },
		{ 292, 232, 19 },
		{ 292, 233, 19 },
		{ 292, 234, 19 },
		{ 292, 235, 19 },
		{ 292, 236, 19 },
		{ 292, 237, 19 },
		{ 292, 238, 19 },
		{ 292, 239, 19 },
		{ 293, 222, 19 },
		{ 293, 223, 19 },
		{ 293, 224, 19 },
		{ 293, 225, 19 },
		{ 293, 226, 19 },
		{ 293, 227, 19 },
		{ 293, 228, 19 },
		{ 293, 229, 19 },
		{ 293, 230, 19 },
		{ 293, 231, 19 },
		{ 293, 232, 19 },
		{ 293, 233, 19 },
		{ 293, 234, 19 },
		{ 293, 235, 19 },
		{ 293, 236, 19 },
		{ 293, 237, 19 },
		{ 293, 238, 19 },
		{ 294, 221, 19 },
		{ 294, 222, 19 },
		{ 294, 223, 19 },
		{ 294, 224, 19 },
		{ 294, 225, 19 },
		{ 294, 226, 19 },
		{ 294, 227, 19 },
		{ 294, 228, 19 },
		{ 294, 229, 19 },
		{ 294, 230, 19 },
		{ 294, 231, 19 },
		{ 294, 232, 19 },
		{ 294, 233, 19 },
		{ 294, 234, 19 },
		{ 294, 235, 19 },
		{ 294, 236, 19 },
		{ 294, 237, 19 },
		{ 294, 238, 19 },
		{ 295, 220, 19 },
		{ 295, 221, 19 },
		{ 295, 222, 19 },
		{ 295, 223, 19 },
		{ 295, 224, 19 },
		{ 295, 225, 19 },
		{ 295, 226, 19 },
		{ 295, 227, 19 },
		{ 295, 228, 19 },
		{ 295, 229, 19 },
		{ 295, 230, 19 },
		{ 295, 231, 19 },
		{ 295, 232, 19 },
		{ 295, 233, 19 },
		{ 295, 234, 19 },
		{ 295, 235, 19 },
		{ 295, 236, 19 },
		{ 295, 237, 19 },
		{ 296, 219, 19 },
		{ 296, 220, 19 },
		{ 296, 221, 19 },
		{ 296, 222, 19 },
		{ 296, 223, 19 },
		{ 296, 224, 19 },
		{ 296, 225, 19 },
		{ 296, 226, 19 },
		{ 296, 227, 19 },
		{ 296, 228, 19 },
		{ 296, 229, 19 },
		{ 296, 230, 19 },
		{ 296, 231, 19 },
		{ 296, 232, 19 },
		{ 296, 233, 19 },
		{ 296, 234, 19 },
		{ 296, 235, 19 },
		{ 296, 236, 19 },
		{ 296, 237, 19 },
		{ 297, 218, 19 },
		{ 297, 219, 19 },
		{ 297, 220, 19 },
		{ 297, 221, 19 },
		{ 297, 222, 19 },
		{ 297, 223, 19 },
		{ 297, 224, 19 },
		{ 297, 225, 19 },
		{ 297, 226, 19 },
		{ 297, 227, 19 },
		{ 297, 228, 19 },
		{ 297, 229, 19 },
		{ 297, 230, 19 },
		{ 297, 231, 19 },
		{ 297, 232, 19 },
		{ 297, 233, 19 },
		{ 297, 234, 19 },
		{ 297, 235, 19 },
		{ 297, 236, 19 },
		{ 298, 217, 19 },
		{ 298, 218, 19 },
		{ 298, 219, 19 },
		{ 298, 220, 19 },
		{ 298, 221, 19 },
		{ 298, 222, 19 },
		{ 298, 223, 19 },
		{ 298, 224, 19 },
		{ 298, 225, 19 },
		{ 298, 226, 19 },
		{ 298, 227, 19 },
		{ 298, 228, 19 },
		{ 298, 229, 19 },
		{ 298, 230, 19 },
		{ 298, 231, 19 },
		{ 298, 232, 19 },
		{ 298, 233, 19 },
		{ 298, 234, 19 },
		{ 298, 235, 19 },
		{ 298, 236, 19 },
		{ 299, 216, 19 },
		{ 299, 217, 19 },
		{ 299, 218, 19 },
		{ 299, 219, 19 },
		{ 299, 220, 19 },
		{ 299, 221, 19 },
		{ 299, 222, 19 },
		{ 299, 223, 19 },
		{ 299, 224, 19 },
		{ 299, 225, 19 },
		{ 299, 226, 19 },
		{ 299, 227, 19 },
		{ 299, 228, 19 },
		{ 299, 229, 19 },
		{ 299, 230, 19 },
		{ 299, 231, 19 },
		{ 299, 232, 19 },
		{ 299, 233, 19 },
		{ 299, 234, 19 },
		{ 299, 235, 19 },
		{ 300, 215, 19 },
		{ 300, 216, 19 },
		{ 300, 217, 19 },
		{ 300, 218, 19 },
		{ 300, 219, 19 },
		{ 300, 220, 19 },
		{ 300, 221, 19 },
		{ 300, 222, 19 },
		{ 300, 223, 19 },
		{ 300, 224, 19 },
		{ 300, 225, 19 },
		{ 300, 226, 19 },
		{ 300, 227, 19 },
		{ 300, 228, 19 },
		{ 300, 229, 19 },
		{ 300, 230, 19 },
		{ 300, 231, 19 },
		{ 300, 232, 19 },
		{ 300, 233, 19 },
		{ 300, 234, 19 },
		{ 300, 235, 19 },
		{ 301, 214, 19 },
		{ 301, 215, 19 },
		{ 301, 216, 19 },
		{ 301, 217, 19 },
		{ 301, 218, 19 },
		{ 301, 219, 19 },
		{ 301, 220, 19 },
		{ 301, 221, 19 },
		{ 301, 222, 19 },
		{ 301, 223, 19 },
		{ 301, 224, 19 },
		{ 301, 225, 19 },
		{ 301, 226, 19 },
		{ 301, 227, 19 },
		{ 301, 228, 19 },
		{ 301, 229, 19 },
		{ 301, 230, 19 },
		{ 301, 231, 19 },
		{ 301, 232, 19 },
		{ 301, 233, 19 },
		{ 301, 234, 19 },
		{ 302, 213, 19 },
		{ 302, 214, 19 },
		{ 302, 215, 19 },
		{ 302, 216, 19 },
		{ 302, 217, 19 },
		{ 302, 218, 19 },
		{ 302, 219, 19 },
		{ 302, 220, 19 },
		{ 302, 221, 19 },
		{ 302, 222, 19 },
		{ 302, 223, 19 },
		{ 302, 224, 19 },
		{ 302, 225, 19 },
		{ 302, 226, 19 },
		{ 302, 227, 19 },
		{ 302, 228, 19 },
		{ 302, 229, 19 },
		{ 302, 230, 19 },
		{ 302, 231, 19 },
		{ 302, 232, 19 },
		{ 302, 233, 19 },
		{ 302, 234, 19 },
		{ 303, 212, 19 },
		{ 303, 213, 19 },
		{ 303, 214, 19 },
		{ 303, 215, 19 },
		{ 303, 216, 19 },
		{ 303, 217, 19 },
		{ 303, 218, 19 },
		{ 303, 219, 19 },
		{ 303, 220, 19 },
		{ 303, 221, 19 },
		{ 303, 222, 19 },
		{ 303, 223, 19 },
		{ 303, 224, 19 },
		{ 303, 225, 19 },
		{ 303, 226, 19 },
		{ 303, 227, 19 },
		{ 303, 228, 19 },
		{ 303, 229, 19 },
		{ 303, 230, 19 },
		{ 303, 231, 19 },
		{ 303, 232, 19 },
		{ 303, 233, 19 },
		{ 304, 211, 19 },
		{ 304, 212, 19 },
		{ 304, 213, 19 },
		{ 304, 214, 19 },
		{ 304, 215, 19 },
		{ 304, 216, 19 },
		{ 304, 217, 19 },
		{ 304, 218, 19 },
		{ 304, 219, 19 },
		{ 304, 220, 19 },
		{ 304, 221, 19 },
		{ 304, 222, 19 },
		{ 304, 223, 19 },
		{ 304, 224, 19 },
		{ 304, 225, 19 },
		{ 304, 226, 19 },
		{ 304, 227, 19 },
		{ 304, 228, 19 },
		{ 304, 229, 19 },
		{ 304, 230, 19 },
		{ 304, 231, 19 },
		{ 304, 232, 19 },
		{ 304, 233, 19 },
		{ 305, 210, 19 },
		{ 305, 211, 19 },
		{ 305, 212, 19 },
		{ 305, 213, 19 },
		{ 305, 214, 19 },
		{ 305, 215, 19 },
		{ 305, 216, 19 },
		{ 305, 217, 19 },
		{ 305, 218, 19 },
		{ 305, 219, 19 },
		{ 305, 220, 19 },
		{ 305, 221, 19 },
		{ 305, 222, 19 },
		{ 305, 223, 19 },
		{ 305, 224, 19 },
		{ 305, 225, 19 },
		{ 305, 226, 19 },
		{ 305, 227, 19 },
		{ 305, 228, 19 },
		{ 305, 229, 19 },
		{ 305, 230, 19 },
		{ 305, 231, 19 },
		{ 305, 232, 19 },
		{ 306, 209, 19 },
		{ 306, 210, 19 },
		{ 306, 211, 19 },
		{ 306, 212, 19 },
		{ 306, 213, 19 },
		{ 306, 214, 19 },
		{ 306, 215, 19 },
		{ 306, 216, 19 },
		{ 306, 217, 19 },
		{ 306, 218, 19 },
		{ 306, 219, 19 },
		{ 306, 220, 19 },
		{ 306, 221, 19 },
		{ 306, 222, 19 },
		{ 306, 223, 19 },
		{ 306, 224, 19 },
		{ 306, 225, 19 },
		{ 306, 226, 19 },
		{ 306, 227, 19 },
		{ 306, 228, 19 },
		{ 306, 229, 19 },
		{ 306, 230, 19 },
		{ 306, 231, 19 },
		{ 306, 232, 19 },
		{ 307, 208, 19 },
		{ 307, 209, 19 },
		{ 307, 210, 19 },
		{ 307, 211, 19 },
		{ 307, 212, 19 },
		{ 307, 213, 19 },
		{ 307, 214, 19 },
		{ 307, 215, 19 },
		{ 307, 216, 19 },
		{ 307, 217, 19 },
		{ 307, 218, 19 },
		{ 307, 219, 19 },
		{ 307, 220, 19 },
		{ 307, 221, 19 },
		{ 307, 222, 19 },
		{ 307, 223, 19 },
		{ 307, 224, 19 },
		{ 307, 225, 19 },
		{ 307, 226, 19 },
		{ 307, 227, 19 },
		{ 307, 228, 19 },
		{ 307, 229, 19 },
		{ 307, 230, 19 },
		{ 307, 231, 19 },
		{ 308, 207, 19 },
		{ 308, 208, 19 },
		{ 308, 209, 19 },
		{ 308, 210, 19 },
		{ 308, 211, 19 },
		{ 308, 212, 19 },
		{ 308, 213, 19 },
		{ 308, 214, 19 },
		{ 308, 215, 19 },
		{ 308, 216, 19 },
		{ 308, 217, 19 },
		{ 308, 218, 19 },
		{ 308, 219, 19 },
		{ 308, 220, 19 },
		{ 308, 221, 19 },
		{ 308, 222, 19 },
		{ 308, 223, 19 },
		{ 308, 224, 19 },
		{ 308, 225, 19 },
		{ 308, 226, 19 },
		{ 308, 227, 19 },
		{ 308, 228, 19 },
		{ 308, 229, 19 },
		{ 308, 230, 19 },
		{ 308, 231, 19 },
		{ 309, 206, 19 },
		{ 309, 207, 19 },
		{ 309, 208, 19 },
		{ 309, 209, 19 },
		{ 309, 210, 19 },
		{ 309, 211, 19 },
		{ 309, 212, 19 },
		{ 309, 213, 19 },
		{ 309, 214, 19 },
		{ 309, 215, 19 },
		{ 309, 216, 19 },
		{ 309, 217, 19 },
		{ 309, 218, 19 },
		{ 309, 219, 19 },
		{ 309, 220, 19 },
		{ 309, 221, 19 },
		{ 309, 222, 19 },
		{ 309, 223, 19 },
		{ 309, 224, 19 },
		{ 309, 225, 19 },
		{ 309, 226, 19 },
		{ 309, 227, 19 },
		{ 309, 228, 19 },
		{ 309, 229, 19 },
		{ 309, 230, 19 },
		{ 310, 205, 19 },
		{ 310, 206, 19 },
		{ 310, 207, 19 },
		{ 310, 208, 19 },
		{ 310, 209, 19 },
		{ 310, 210, 19 },
		{ 310, 211, 19 },
		{ 310, 212, 19 },
		{ 310, 213, 19 },
		{ 310, 214, 19 },
		{ 310, 215, 19 },
		{ 310, 216, 19 },
		{ 310, 217, 19 },
		{ 310, 218, 19 },
		{ 310, 219, 19 },
		{ 310, 220, 19 },
		{ 310, 221, 19 },
		{ 310, 222, 19 },
		{ 310, 223, 19 },
		{ 310, 224, 19 },
		{ 310, 225, 19 },
		{ 310, 226, 19 },
		{ 310, 227, 19 },
		{ 310, 228, 19 },
		{ 310, 229, 19 },
		{ 310, 230, 19 },
		{ 311, 204, 19 },
		{ 311, 205, 19 },
		{ 311, 206, 19 },
		{ 311, 207, 19 },
		{ 311, 208, 19 },
		{ 311, 209, 19 },
		{ 311, 210, 19 },
		{ 311, 211, 19 },
		{ 311, 212, 19 },
		{ 311, 213, 19 },
		{ 311, 214, 19 },
		{ 311, 215, 19 },
		{ 311, 216, 19 },
		{ 311, 217, 19 },
		{ 311, 218, 19 },
		{ 311, 219, 19 },
		{ 311, 220, 19 },
		{ 311, 221, 19 },
		{ 311, 222, 19 },
		{ 311, 223, 19 },
		{ 311, 224, 19 },
		{ 311, 225, 19 },
		{ 311, 226, 19 },
		{ 311, 227, 19 },
		{ 311, 228, 19 },
		{ 311, 229, 19 },
		{ 312, 203, 19 },
		{ 312, 204, 19 },
		{ 312, 205, 19 },
		{ 312, 206, 19 },
		{ 312, 207, 19 },
		{ 312, 208, 19 },
		{ 312, 209, 19 },
		{ 312, 210, 19 },
		{ 312, 211, 19 },
		{ 312, 212, 19 },
		{ 312, 213, 19 },
		{ 312, 214, 19 },
		{ 312, 215, 19 },
		{ 312, 216, 19 },
		{ 312, 217, 19 },
		{ 312, 218, 19 },
		{ 312, 219, 19 },
		{ 312, 220, 19 },
		{ 312, 221, 19 },
		{ 312, 222, 19 },
		{ 312, 223, 19 },
		{ 312, 224, 19 },
		{ 312, 225, 19 },
		{ 312, 226, 19 },
		{ 312, 227, 19 },
		{ 312, 228, 19 },
		{ 312, 229, 19 },
		{ 313, 202, 19 },
		{ 313, 203, 19 },
		{ 313, 204, 19 },
		{ 313, 205, 19 },
		{ 313, 206, 19 },
		{ 313, 207, 19 },
		{ 313, 208, 19 },
		{ 313, 209, 19 },
		{ 313, 210, 19 },
		{ 313, 211, 19 },
		{ 313, 212, 19 },
		{ 313, 213, 19 },
		{ 313, 214, 19 },
		{ 313, 215, 19 },
		{ 313, 216, 19 },
		{ 313, 217, 19 },
		{ 313, 218, 19 },
		{ 313, 219, 19 },
		{ 313, 220, 19 },
		{ 313, 221, 19 },
		{ 313, 222, 19 },
		{ 313, 223, 19 },
		{ 313, 224, 19 },
		{ 313, 225, 19 },
		{ 313, 226, 19 },
		{ 313, 227, 19 },
		{ 313, 228, 19 },
		{ 314, 201, 19 },
		{ 314, 202, 19 },
		{ 314, 203, 19 },
		{ 314, 204, 19 },
		{ 314, 205, 19 },
		{ 314, 206, 19 },
		{ 314, 207, 19 },
		{ 314, 208, 19 },
		{ 314, 209, 19 },
		{ 314, 210, 19 },
		{ 314, 211, 19 },
		{ 314, 212, 19 },
		{ 314, 213, 19 },
		{ 314, 214, 19 },
		{ 314, 215, 19 },
		{ 314, 216, 19 },
		{ 314, 217, 19 },
		{ 314, 218, 19 },
		{ 314, 219, 19 },
		{ 314, 220, 19 },
		{ 314, 221, 19 },
		{ 314, 222, 19 },
		{ 314, 223, 19 },
		{ 314, 224, 19 },
		{ 314, 225, 19 },
		{ 314, 226, 19 },
		{ 314, 227, 19 },
		{ 314, 228, 19 },
		{ 315, 200, 19 },
		{ 315, 201, 19 },
		{ 315, 202, 19 },
		{ 315, 203, 19 },
		{ 315, 204, 19 },
		{ 315, 205, 19 },
		{ 315, 206, 19 },
		{ 315, 207, 19 },
		{ 315, 208, 19 },
		{ 315, 209, 19 },
		{ 315, 210, 19 },
		{ 315, 211, 19 },
		{ 315, 212, 19 },
		{ 315, 213, 19 },
		{ 315, 214, 19 },
		{ 315, 215, 19 },
		{ 315, 216, 19 },
		{ 315, 217, 19 },
		{ 315, 218, 19 },
		{ 315, 219, 19 },
		{ 315, 220, 19 },
		{ 315, 221, 19 },
		{ 315, 222, 19 },
		{ 315, 223, 19 },
		{ 315, 224, 19 },
		{ 315, 225, 19 },
		{ 315, 226, 19 },
		{ 315, 227, 19 },
		{ 316, 199, 19 },
		{ 316, 200, 19 },
		{ 316, 201, 19 },
		{ 316, 202, 19 },
		{ 316, 203, 19 },
		{ 316, 204, 19 },
		{ 316, 205, 19 },
		{ 316, 206, 19 },
		{ 316, 207, 19 },
		{ 316, 208, 19 },
		{ 316, 209, 19 },
		{ 316, 210, 19 },
		{ 316, 211, 19 },
		{ 316, 212, 19 },
		{ 316, 213, 19 },
		{ 316, 214, 19 },
		{ 316, 215, 19 },
		{ 316, 216, 19 },
		{ 316, 217, 19 },
		{ 316, 218, 19 },
		{ 316, 219, 19 },
		{ 316, 220, 19 },
		{ 316, 221, 19 },
		{ 316, 222, 19 },
		{ 316, 223, 19 },
		{ 316, 224, 19 },
		{ 316, 225, 19 },
		{ 316, 226, 19 },
		{ 316, 227, 19 },
		{ 317, 198, 19 },
		{ 317, 199, 19 },
		{ 317, 200, 19 },
		{ 317, 201, 19 },
		{ 317, 202, 19 },
		{ 317, 203, 19 },
		{ 317, 204, 19 },
		{ 317, 205, 19 },
		{ 317, 206, 19 },
		{ 317, 207, 19 },
		{ 317, 208, 19 },
		{ 317, 209, 19 },
		{ 317, 210, 19 },
		{ 317, 211, 19 },
		{ 317, 212, 19 },
		{ 317, 213, 19 },
		{ 317, 214, 19 },
		{ 317, 215, 19 },
		{ 317, 216, 19 },
		{ 317, 217, 19 },
		{ 317, 218, 19 },
		{ 317, 219, 19 },
		{ 317, 220, 19 },
		{ 317, 221, 19 },
		{ 317, 222, 19 },
		{ 317, 223, 19 },
		{ 317, 224, 19 },
		{ 317, 225, 19 },
		{ 317, 226, 19 },
		{ 318, 197, 19 },
		{ 318, 198, 19 },
		{ 318, 199, 19 },
		{ 318, 200, 19 },
		{ 318, 201, 19 },
		{ 318, 202, 19 },
		{ 318, 203, 19 },
		{ 318, 204, 19 },
		{ 318, 205, 19 },
		{ 318, 206, 19 },
		{ 318, 207, 19 },
		{ 318, 208, 19 },
		{ 318, 209, 19 },
		{ 318, 210, 19 },
		{ 318, 211, 19 },
		{ 318, 212, 19 },
		{ 318, 213, 19 },
		{ 318, 214, 19 },
		{ 318, 215, 19 },
		{ 318, 216, 19 },
		{ 318, 217, 19 },
		{ 318, 218, 19 },
		{ 318, 219, 19 },
		{ 318, 220, 19 },
		{ 318, 221, 19 },
		{ 318, 222, 19 },
		{ 318, 223, 19 },
		{ 318, 224, 19 },
		{ 318, 225, 19 },
		{ 318, 226, 19 },
		{ 319, 196, 19 },
		{ 319, 197, 19 },
		{ 319, 198, 19 },
		{ 319, 199, 19 },
		{ 319, 200, 19 },
		{ 319, 201, 19 },
		{ 319, 202, 19 },
		{ 319, 203, 19 },
		{ 319, 204, 19 },
		{ 319, 205, 19 },
		{ 319, 206, 19 },
		{ 319, 207, 19 },
		{ 319, 208, 19 },
		{ 319, 209, 19 },
		{ 319, 210, 19 },
		{ 319, 211, 19 },
		{ 319, 212, 19 },
		{ 319, 213, 19 },
		{ 319, 214, 19 },
		{ 319, 215, 19 },
		{ 319, 216, 19 },
		{ 319, 217, 19 },
		{ 319, 218, 19 },
		{ 319, 219, 19 },
		{ 319, 220, 19 },
		{ 319, 221, 19 },
		{ 319, 222, 19 },
		{ 319, 223, 19 },
		{ 319, 224, 19 },
		{ 319, 225, 19 },
		{ 320, 195, 19 },
		{ 320, 196, 19 },
		{ 320, 197, 19 },
		{ 320, 198, 19 },
		{ 320, 199, 19 },
		{ 320, 200, 19 },
		{ 320, 201, 19 },
		{ 320, 202, 19 },
		{ 320, 203, 19 },
		{ 320, 204, 19 },
		{ 320, 205, 19 },
		{ 320, 206, 19 },
		{ 320, 207, 19 },
		{ 320, 208, 19 },
		{ 320, 209, 19 },
		{ 320, 210, 19 },
		{ 320, 211, 19 },
		{ 320, 212, 19 },
		{ 320, 213, 19 },
		{ 320, 214, 19 },
		{ 320, 215, 19 },
		{ 320, 216, 19 },
		{ 320, 217, 19 },
		{ 320, 218, 19 },
		{ 320, 219, 19 },
		{ 320, 220, 19 },
		{ 320, 221, 19 },
		{ 320, 222, 19 },
		{ 320, 223, 19 },
		{ 320, 224, 19 },
		{ 320, 225, 19 },
		{ 321, 194, 19 },
		{ 321, 195, 19 },
		{ 321, 196, 19 },
		{ 321, 197, 19 },
		{ 321, 198, 19 },
		{ 321, 199, 19 },
		{ 321, 200, 19 },
		{ 321, 201, 19 },
		{ 321, 202, 19 },
		{ 321, 203, 19 },
		{ 321, 204, 19 },
		{ 321, 205, 19 },
		{ 321, 206, 19 },
		{ 321, 207, 19 },
		{ 321, 208, 19 },
		{ 321, 209, 19 },
		{ 321, 210, 19 },
		{ 321, 211, 19 },
		{ 321, 212, 19 },
		{ 321, 213, 19 },
		{ 321, 214, 19 },
		{ 321, 215, 19 },
		{ 321, 216, 19 },
		{ 321, 217, 19 },
		{ 321, 218, 19 },
		{ 321, 219, 19 },
		{ 321, 220, 19 },
		{ 321, 221, 19 },
		{ 321, 222, 19 },
		{ 321, 223, 19 },
		{ 321, 224, 19 },
		{ 322, 193, 19 },
		{ 322, 194, 19 },
		{ 322, 195, 19 },
		{ 322, 196, 19 },
		{ 322, 197, 19 },
		{ 322, 198, 19 },
		{ 322, 199, 19 },
		{ 322, 200, 19 },
		{ 322, 201, 19 },
		{ 322, 202, 19 },
		{ 322, 203, 19 },
		{ 322, 204, 19 },
		{ 322, 205, 19 },
		{ 322, 206, 19 },
		{ 322, 207, 19 },
		{ 322, 208, 19 },
		{ 322, 209, 19 },
		{ 322, 210, 19 },
		{ 322, 211, 19 },
		{ 322, 212, 19 },
		{ 322, 213, 19 },
		{ 322, 214, 19 },
		{ 322, 215, 19 },
		{ 322, 216, 19 },
		{ 322, 217, 19 },
		{ 322, 218, 19 },
		{ 322, 219, 19 },
		{ 322, 220, 19 },
		{ 322, 221, 19 },
		{ 322, 222, 19 },
		{ 322, 223, 19 },
		{ 322, 224, 19 },
		{ 323, 192, 19 },
		{ 323, 193, 19 },
		{ 323, 194, 19 },
		{ 323, 195, 19 },
		{ 323, 196, 19 },
		{ 323, 197, 19 },
		{ 323, 198, 19 },
		{ 323, 199, 19 },
		{ 323, 200, 19 },
		{ 323, 201, 19 },
		{ 323, 202, 19 },
		{ 323, 203, 19 },
		{ 323, 204, 19 },
		{ 323, 205, 19 },
		{ 323, 206, 19 },
		{ 323, 207, 19 },
		{ 323, 208, 19 },
		{ 323, 209, 19 },
		{ 323, 210, 19 },
		{ 323, 211, 19 },
		{ 323, 212, 19 },
		{ 323, 213, 19 },
		{ 323, 214, 19 },
		{ 323, 215, 19 },
		{ 323, 216, 19 },
		{ 323, 217, 19 },
		{ 323, 218, 19 },
		{ 323, 219, 19 },
		{ 323, 220, 19 },
		{ 323, 221, 19 },
		{ 323, 222, 19 },
		{ 323, 223, 19 },
		{ 324, 191, 19 },
		{ 324, 192, 19 },
		{ 324, 193, 19 },
		{ 324, 194, 19 },
		{ 324, 195, 19 },
		{ 324, 196, 19 },
		{ 324, 197, 19 },
		{ 324, 198, 19 },
		{ 324, 199, 19 },
		{ 324, 200, 19 },
		{ 324, 201, 19 },
		{ 324, 202, 19 },
		{ 324, 203, 19 },
		{ 324, 204, 19 },
		{ 324, 205, 19 },
		{ 324, 206, 19 },
		{ 324, 207, 19 },
		{ 324, 208, 19 },
		{ 324, 209, 19 },
		{ 324, 210, 19 },
		{ 324, 211, 19 },
		{ 324, 212, 19 },
		{ 324, 213, 19 },
		{ 324, 214, 19 },
		{ 324, 215, 19 },
		{ 324, 216, 19 },
		{ 324, 217, 19 },
		{ 324, 218, 19 },
		{ 324, 219, 19 },
		{ 324, 220, 19 },
		{ 324, 221, 19 },
		{ 324, 222, 19 },
		{ 324, 223, 19 },
		{ 325, 190, 19 },
		{ 325, 191, 19 },
		{ 325, 192, 19 },
		{ 325, 193, 19 },
		{ 325, 194, 19 },
		{ 325, 195, 19 },
		{ 325, 196, 19 },
		{ 325, 197, 19 },
		{ 325, 198, 19 },
		{ 325, 199, 19 },
		{ 325, 200, 19 },
		{ 325, 201, 19 },
		{ 325, 202, 19 },
		{ 325, 203, 19 },
		{ 325, 204, 19 },
		{ 325, 205, 19 },
		{ 325, 206, 19 },
		{ 325, 207, 19 },
		{ 325, 208, 19 },
		{ 325, 209, 19 },
		{ 325, 210, 19 },
		{ 325, 211, 19 },
		{ 325, 212, 19 },
		{ 325, 213, 19 },
		{ 325, 214, 19 },
		{ 325, 215, 19 },
		{ 325, 216, 19 },
		{ 325, 217, 19 },
		{ 325, 218, 19 },
		{ 325, 219, 19 },
		{ 325, 220, 19 },
		{ 325, 221, 19 },
		{ 325, 222, 19 },
		{ 326, 189, 19 },
		{ 326, 190, 19 },
		{ 326, 191, 19 },
		{ 326, 192, 19 },
		{ 326, 193, 19 },
		{ 326, 194, 19 },
		{ 326, 195, 19 },
		{ 326, 196, 19 },
		{ 326, 197, 19 },
		{ 326, 198, 19 },
		{ 326, 199, 19 },
		{ 326, 200, 19 },
		{ 326, 201, 19 },
		{ 326, 202, 19 },
		{ 326, 203, 19 },
		{ 326, 204, 19 },
		{ 326, 205, 19 },
		{ 326, 206, 19 },
		{ 326, 207, 19 },
		{ 326, 208, 19 },
		{ 326, 209, 19 },
		{ 326, 210, 19 },
		{ 326, 211, 19 },
		{ 326, 212, 19 },
		{ 326, 213, 19 },
		{ 326, 214, 19 },
		{ 326, 215, 19 },
		{ 326, 216, 19 },
		{ 326, 217, 19 },
		{ 326, 218, 19 },
		{ 326, 219, 19 },
		{ 326, 220, 19 },
		{ 326, 221, 19 },
		{ 326, 222, 19 },
		{ 327, 188, 19 },
		{ 327, 189, 19 },
		{ 327, 190, 19 },
		{ 327, 191, 19 },
		{ 327, 192, 19 },
		{ 327, 193, 19 },
		{ 327, 194, 19 },
		{ 327, 195, 19 },
		{ 327, 196, 19 },
		{ 327, 197, 19 },
		{ 327, 198, 19 },
		{ 327, 199, 19 },
		{ 327, 200, 19 },
		{ 327, 201, 19 },
		{ 327, 202, 19 },
		{ 327, 203, 19 },
		{ 327, 204, 19 },
		{ 327, 205, 19 },
		{ 327, 206, 19 },
		{ 327, 207, 19 },
		{ 327, 208, 19 },
		{ 327, 209, 19 },
		{ 327, 210, 19 },
		{ 327, 211, 19 },
		{ 327, 212, 19 },
		{ 327, 213, 19 },
		{ 327, 214, 19 },
		{ 327, 215, 19 },
		{ 327, 216, 19 },
		{ 327, 217, 19 },
		{ 327, 218, 19 },
		{ 327, 219, 19 },
		{ 327, 220, 19 },
		{ 327, 221, 19 },
		{ 328, 187, 19 },
		{ 328, 188, 19 },
		{ 328, 189, 19 },
		{ 328, 190, 19 },
		{ 328, 191, 19 },
		{ 328, 192, 19 },
		{ 328, 193, 19 },
		{ 328, 194, 19 },
		{ 328, 195, 19 },
		{ 328, 196, 19 },
		{ 328, 197, 19 },
		{ 328, 198, 19 },
		{ 328, 199, 19 },
		{ 328, 200, 19 },
		{ 328, 201, 19 },
		{ 328, 202, 19 },
		{ 328, 203, 19 },
		{ 328, 204, 19 },
		{ 328, 205, 19 },
		{ 328, 206, 19 },
		{ 328, 207, 19 },
		{ 328, 208, 19 },
		{ 328, 209, 19 },
		{ 328, 210, 19 },
		{ 328, 211, 19 },
		{ 328, 212, 19 },
		{ 328, 213, 19 },
		{ 328, 214, 19 },
		{ 328, 215, 19 },
		{ 328, 216, 19 },
		{ 328, 217, 19 },
		{ 328, 218, 19 },
		{ 328, 219, 19 },
		{ 328, 220, 19 },
		{ 328, 221, 19 },
		{ 329, 186, 19 },
		{ 329, 187, 19 },
		{ 329, 188, 19 },
		{ 329, 189, 19 },
		{ 329, 190, 19 },
		{ 329, 191, 19 },
		{ 329, 192, 19 },
		{ 329, 193, 19 },
		{ 329, 194, 19 },
		{ 329, 195, 19 },
		{ 329, 196, 19 },
		{ 329, 197, 19 },
		{ 329, 198, 19 },
		{ 329, 199, 19 },
		{ 329, 200, 19 },
		{ 329, 201, 19 },
		{ 329, 202, 19 },
		{ 329, 203, 19 },
		{ 329, 204, 19 },
		{ 329, 205, 19 },
		{ 329, 206, 19 },
		{ 329, 207, 19 },
		{ 329, 208, 19 },
		{ 329, 209, 19 },
		{ 329, 210, 19 },
		{ 329, 211, 19 },
		{ 329, 212, 19 },
		{ 329, 213, 19 },
		{ 329, 214, 19 },
		{ 329, 215, 19 },
		{ 329, 216, 19 },
		{ 329, 217, 19 },
		{ 329, 218, 19 },
		{ 329, 219, 19 },
		{ 329, 220, 19 },
		{ 330, 185, 19 },
		{ 330, 186, 19 },
		{ 330, 187, 19 },
		{ 330, 188, 19 },
		{ 330, 189, 19 },
		{ 330, 190, 19 },
		{ 330, 191, 19 },
		{ 330, 192, 19 },
		{ 330, 193, 19 },
		{ 330, 194, 19 },
		{ 330, 195, 19 },
		{ 330, 196, 19 },
		{ 330, 197, 19 },
		{ 330, 198, 19 },
		{ 330, 199, 19 },
		{ 330, 200, 19 },
		{ 330, 201, 19 },
		{ 330, 202, 19 },
		{ 330, 203, 19 },
		{ 330, 204, 19 },
		{ 330, 205, 19 },
		{ 330, 206, 19 },
		{ 330, 207, 19 },
		{ 330, 208, 19 },
		{ 330, 209, 19 },
		{ 330, 210, 19 },
		{ 330, 211, 19 },
		{ 330, 212, 19 },
		{ 330, 213, 19 },
		{ 330, 214, 19 },
		{ 330, 215, 19 },
		{ 330, 216, 19 },
		{ 330, 217, 19 },
		{ 330, 218, 19 },
		{ 330, 219, 19 },
		{ 330, 220, 19 },
		{ 331, 184, 19 },
		{ 331, 185, 19 },
		{ 331, 186, 19 },
		{ 331, 187, 19 },
		{ 331, 188, 19 },
		{ 331, 189, 19 },
		{ 331, 190, 19 },
		{ 331, 191, 19 },
		{ 331, 192, 19 },
		{ 331, 193, 19 },
		{ 331, 194, 19 },
		{ 331, 195, 19 },
		{ 331, 196, 19 },
		{ 331, 197, 19 },
		{ 331, 198, 19 },
		{ 331, 199, 19 },
		{ 331, 200, 19 },
		{ 331, 201, 19 },
		{ 331, 202, 19 },
		{ 331, 203, 19 },
		{ 331, 204, 19 },
		{ 331, 205, 19 },
		{ 331, 206, 19 },
		{ 331, 207, 19 },
		{ 331, 208, 19 },
		{ 331, 209, 19 },
		{ 331, 210, 19 },
		{ 331, 211, 19 },
		{ 331, 212, 19 },
		{ 331, 213, 19 },
		{ 331, 214, 19 },
		{ 331, 215, 19 },
		{ 331, 216, 19 },
		{ 331, 217, 19 },
		{ 331, 218, 19 },
		{ 331, 219, 19 },
		{ 332, 183, 19 },
		{ 332, 184, 19 },
		{ 332, 185, 19 },
		{ 332, 186, 19 },
		{ 332, 187, 19 },
		{ 332, 188, 19 },
		{ 332, 189, 19 },
		{ 332, 190, 19 },
		{ 332, 191, 19 },
		{ 332, 192, 19 },
		{ 332, 193, 19 },
		{ 332, 194, 19 },
		{ 332, 195, 19 },
		{ 332, 196, 19 },
		{ 332, 197, 19 },
		{ 332, 198, 19 },
		{ 332, 199, 19 },
		{ 332, 200, 19 },
		{ 332, 201, 19 },
		{ 332, 202, 19 },
		{ 332, 203, 19 },
		{ 332, 204, 19 },
		{ 332, 205, 19 },
		{ 332, 206, 19 },
		{ 332, 207, 19 },
		{ 332, 208, 19 },
		{ 332, 209, 19 },
		{ 332, 210, 19 },
		{ 332, 211, 19 },
		{ 332, 212, 19 },
		{ 332, 213, 19 },
		{ 332, 214, 19 },
		{ 332, 215, 19 },
		{ 332, 216, 19 },
		{ 332, 217, 19 },
		{ 332, 218, 19 },
		{ 332, 219, 19 },
		{ 333, 182, 19 },
		{ 333, 183, 19 },
		{ 333, 184, 19 },
		{ 333, 185, 19 },
		{ 333, 186, 19 },
		{ 333, 187, 19 },
		{ 333, 188, 19 },
		{ 333, 189, 19 },
		{ 333, 190, 19 },
		{ 333, 191, 19 },
		{ 333, 192, 19 },
		{ 333, 193, 19 },
		{ 333, 194, 19 },
		{ 333, 195, 19 },
		{ 333, 196, 19 },
		{ 333, 197, 19 },
		{ 333, 198, 19 },
		{ 333, 199, 19 },
		{ 333, 200, 19 },
		{ 333, 201, 19 },
		{ 333, 202, 19 },
		{ 333, 203, 19 },
		{ 333, 204, 19 },
		{ 333, 205, 19 },
		{ 333, 206, 19 },
		{ 333, 207, 19 },
		{ 333, 208, 19 },
		{ 333, 209, 19 },
		{ 333, 210, 19 },
		{ 333, 211, 19 },
		{ 333, 212, 19 },
		{ 333, 213, 19 },
		{ 333, 214, 19 },
		{ 333, 215, 19 },
		{ 333, 216, 19 },
		{ 333, 217, 19 },
		{ 333, 218, 19 },
		{ 334, 181, 19 },
		{ 334, 182, 19 },
		{ 334, 183, 19 },
		{ 334, 184, 19 },
		{ 334, 185, 19 },
		{ 334, 186, 19 },
		{ 334, 187, 19 },
		{ 334, 188, 19 },
		{ 334, 189, 19 },
		{ 334, 190, 19 },
		{ 334, 191, 19 },
		{ 334, 192, 19 },
		{ 334, 193, 19 },
		{ 334, 194, 19 },
		{ 334, 195, 19 },
		{ 334, 196, 19 },
		{ 334, 197, 19 },
		{ 334, 198, 19 },
		{ 334, 199, 19 },
		{ 334, 200, 19 },
		{ 334, 201, 19 },
		{ 334, 202, 19 },
		{ 334, 203, 19 },
		{ 334, 204, 19 },
		{ 334, 205, 19 },
		{ 334, 206, 19 },
		{ 334, 207, 19 },
		{ 334, 208, 19 },
		{ 334, 209, 19 },
		{ 334, 210, 19 },
		{ 334, 211, 19 },
		{ 334, 212, 19 },
		{ 334, 213, 19 },
		{ 334, 214, 19 },
		{ 334, 215, 19 },
		{ 334, 216, 19 },
		{ 334, 217, 19 },
		{ 334, 218, 19 },
		{ 335, 180, 19 },
		{ 335, 181, 19 },
		{ 335, 182, 19 },
		{ 335, 183, 19 },
		{ 335, 184, 19 },
		{ 335, 185, 19 },
		{ 335, 186, 19 },
		{ 335, 187, 19 },
		{ 335, 188, 19 },
		{ 335, 189, 19 },
		{ 335, 190, 19 },
		{ 335, 191, 19 },
		{ 335, 192, 19 },
		{ 335, 193, 19 },
		{ 335, 194, 19 },
		{ 335, 195, 19 },
		{ 335, 196, 19 },
		{ 335, 197, 19 },
		{ 335, 198, 19 },
		{ 335, 199, 19 },
		{ 335, 200, 19 },
		{ 335, 201, 19 },
		{ 335, 202, 19 },
		{ 335, 203, 19 },
		{ 335, 204, 19 },
		{ 335, 205, 19 },
		{ 335, 206, 19 },
		{ 335, 207, 19 },
		{ 335, 208, 19 },
		{ 335, 209, 19 },
		{ 335, 210, 19 },
		{ 335, 211, 19 },
		{ 335, 212, 19 },
		{ 335, 213, 19 },
		{ 335, 214, 19 },
		{ 335, 215, 19 },
		{ 335, 216, 19 },
		{ 335, 217, 19 },
		{ 336, 179, 19 },
		{ 336, 180, 19 },
		{ 336, 181, 19 },
		{ 336, 182, 19 },
		{ 336, 183, 19 },
		{ 336, 184, 19 },
		{ 336, 185, 19 },
		{ 336, 186, 19 },
		{ 336, 187, 19 },
		{ 336, 188, 19 },
		{ 336, 189, 19 },
		{ 336, 190, 19 },
		{ 336, 191, 19 },
		{ 336, 192, 19 },
		{ 336, 193, 19 },
		{ 336, 194, 19 },
		{ 336, 195, 19 },
		{ 336, 196, 19 },
		{ 336, 197, 19 },
		{ 336, 198, 19 },
		{ 336, 199, 19 },
		{ 336, 200, 19 },
		{ 336, 201, 19 },
		{ 336, 202, 19 },
		{ 336, 203, 19 },
		{ 336, 204, 19 },
		{ 336, 205, 19 },
		{ 336, 206, 19 },
		{ 336, 207, 19 },
		{ 336, 208, 19 },
		{ 336, 209, 19 },
		{ 336, 210, 19 },
		{ 336, 211, 19 },
		{ 336, 212, 19 },
		{ 336, 213, 19 },
		{ 336, 214, 19 },
		{ 336, 215, 19 },
		{ 336, 216, 19 },
		{ 336, 217, 19 },
		{ 337, 178, 19 },
		{ 337, 179, 19 },
		{ 337, 180, 19 },
		{ 337, 181, 19 },
		{ 337, 182, 19 },
		{ 337, 183, 19 },
		{ 337, 184, 19 },
		{ 337, 185, 19 },
		{ 337, 186, 19 },
		{ 337, 187, 19 },
		{ 337, 188, 19 },
		{ 337, 189, 19 },
		{ 337, 190, 19 },
		{ 337, 191, 19 },
		{ 337, 192, 19 },
		{ 337, 193, 19 },
		{ 337, 194, 19 },
		{ 337, 195, 19 },
		{ 337, 196, 19 },
		{ 337, 197, 19 },
		{ 337, 198, 19 },
		{ 337, 199, 19 },
		{ 337, 200, 19 },
		{ 337, 201, 19 },
		{ 337, 202, 19 },
		{ 337, 203, 19 },
		{ 337, 204, 19 },
		{ 337, 205, 19 },
		{ 337, 206, 19 },
		{ 337, 207, 19 },
		{ 337, 208, 19 },
		{ 337, 209, 19 },
		{ 337, 210, 19 },
		{ 337, 211, 19 },
		{ 337, 212, 19 },
		{ 337, 213, 19 },
		{ 337, 214, 19 },
		{ 337, 215, 19 },
		{ 337, 216, 19 },
		{ 338, 177, 19 },
		{ 338, 178, 19 },
		{ 338, 179, 19 },
		{ 338, 180, 19 },
		{ 338, 181, 19 },
		{ 338, 182, 19 },
		{ 338, 183, 19 },
		{ 338, 184, 19 },
		{ 338, 185, 19 },
		{ 338, 186, 19 },
		{ 338, 187, 19 },
		{ 338, 188, 19 },
		{ 338, 189, 19 },
		{ 338, 190, 19 },
		{ 338, 191, 19 },
		{ 338, 192, 19 },
		{ 338, 193, 19 },
		{ 338, 194, 19 },
		{ 338, 195, 19 },
		{ 338, 196, 19 },
		{ 338, 197, 19 },
		{ 338, 198, 19 },
		{ 338, 199, 19 },
		{ 338, 200, 19 },
		{ 338, 201, 19 },
		{ 338, 202, 19 },
		{ 338, 203, 19 },
		{ 338, 204, 19 },
		{ 338, 205, 19 },
		{ 338, 206, 19 },
		{ 338, 207, 19 },
		{ 338, 208, 19 },
		{ 338, 209, 19 },
		{ 338, 210, 19 },
		{ 338, 211, 19 },
		{ 338, 212, 19 },
		{ 338, 213, 19 },
		{ 338, 214, 19 },
		{ 338, 215, 19 },
		{ 338, 216, 19 },
		{ 339, 176, 19 },
		{ 339, 177, 19 },
		{ 339, 178, 19 },
		{ 339, 179, 19 },
		{ 339, 180, 19 },
		{ 339, 181, 19 },
		{ 339, 182, 19 },
		{ 339, 183, 19 },
		{ 339, 184, 19 },
		{ 339, 185, 19 },
		{ 339, 186, 19 },
		{ 339, 187, 19 },
		{ 339, 188, 19 },
		{ 339, 189, 19 },
		{ 339, 190, 19 },
		{ 339, 191, 19 },
		{ 339, 192, 19 },
		{ 339, 193, 19 },
		{ 339, 194, 19 },
		{ 339, 195, 19 },
		{ 339, 196, 19 },
		{ 339, 197, 19 },
		{ 339, 198, 19 },
		{ 339, 199, 19 },
		{ 339, 200, 19 },
		{ 339, 201, 19 },
		{ 339, 202, 19 },
		{ 339, 203, 19 },
		{ 339, 204, 19 },
		{ 339, 205, 19 },
		{ 339, 206, 19 },
		{ 339, 207, 19 },
		{ 339, 208, 19 },
		{ 339, 209, 19 },
		{ 339, 210, 19 },
		{ 339, 211, 19 },
		{ 339, 212, 19 },
		{ 339, 213, 19 },
		{ 339, 214, 19 },
		{ 339, 215, 19 },
		{ 340, 175, 19 },
		{ 340, 176, 19 },
		{ 340, 177, 19 },
		{ 340, 178, 19 },
		{ 340, 179, 19 },
		{ 340, 180, 19 },
		{ 340, 181, 19 },
		{ 340, 182, 19 },
		{ 340, 183, 19 },
		{ 340, 184, 19 },
		{ 340, 185, 19 },
		{ 340, 186, 19 },
		{ 340, 187, 19 },
		{ 340, 188, 19 },
		{ 340, 189, 19 },
		{ 340, 190, 19 },
		{ 340, 191, 19 },
		{ 340, 192, 19 },
		{ 340, 193, 19 },
		{ 340, 194, 19 },
		{ 340, 195, 19 },
		{ 340, 196, 19 },
		{ 340, 197, 19 },
		{ 340, 198, 19 },
		{ 340, 199, 19 },
		{ 340, 200, 19 },
		{ 340, 201, 19 },
		{ 340, 202, 19 },
		{ 340, 203, 19 },
		{ 340, 204, 19 },
		{ 340, 205, 19 },
		{ 340, 206, 19 },
		{ 340, 207, 19 },
		{ 340, 208, 19 },
		{ 340, 209, 19 },
		{ 340, 210, 19 },
		{ 340, 211, 19 },
		{ 340, 212, 19 },
		{ 340, 213, 19 },
		{ 340, 214, 19 },
		{ 340, 215, 19 },
		{ 341, 174, 19 },
		{ 341, 175, 19 },
		{ 341, 176, 19 },
		{ 341, 177, 19 },
		{ 341, 178, 19 },
		{ 341, 179, 19 },
		{ 341, 180, 19 },
		{ 341, 181, 19 },
		{ 341, 182, 19 },
		{ 341, 183, 19 },
		{ 341, 184, 19 },
		{ 341, 185, 19 },
		{ 341, 186, 19 },
		{ 341, 187, 19 },
		{ 341, 188, 19 },
		{ 341, 189, 19 },
		{ 341, 190, 19 },
		{ 341, 191, 19 },
		{ 341, 192, 19 },
		{ 341, 193, 19 },
		{ 341, 194, 19 },
		{ 341, 195, 19 },
		{ 341, 196, 19 },
		{ 341, 197, 19 },
		{ 341, 198, 19 },
		{ 341, 199, 19 },
		{ 341, 200, 19 },
		{ 341, 201, 19 },
		{ 341, 202, 19 },
		{ 341, 203, 19 },
		{ 341, 204, 19 },
		{ 341, 205, 19 },
		{ 341, 206, 19 },
		{ 341, 207, 19 },
		{ 341, 208, 19 },
		{ 341, 209, 19 },
		{ 341, 210, 19 },
		{ 341, 211, 19 },
		{ 341, 212, 19 },
		{ 341, 213, 19 },
		{ 341, 214, 19 },
		{ 342, 173, 19 },
		{ 342, 174, 19 },
		{ 342, 175, 19 },
		{ 342, 176, 19 },
		{ 342, 177, 19 },
		{ 342, 178, 19 },
		{ 342, 179, 19 },
		{ 342, 180, 19 },
		{ 342, 181, 19 },
		{ 342, 182, 19 },
		{ 342, 183, 19 },
		{ 342, 184, 19 },
		{ 342, 185, 19 },
		{ 342, 186, 19 },
		{ 342, 187, 19 },
		{ 342, 188, 19 },
		{ 342, 189, 19 },
		{ 342, 190, 19 },
		{ 342, 191, 19 },
		{ 342, 192, 19 },
		{ 342, 193, 19 },
		{ 342, 194, 19 },
		{ 342, 195, 19 },
		{ 342, 196, 19 },
		{ 342, 197, 19 },
		{ 342, 198, 19 },
		{ 342, 199, 19 },
		{ 342, 200, 19 },
		{ 342, 201, 19 },
		{ 342, 202, 19 },
		{ 342, 203, 19 },
		{ 342, 204, 19 },
		{ 342, 205, 19 },
		{ 342, 206, 19 },
		{ 342, 207, 19 },
		{ 342, 208, 19 },
		{ 342, 209, 19 },
		{ 342, 210, 19 },
		{ 342, 211, 19 },
		{ 342, 212, 19 },
		{ 342, 213, 19 },
		{ 342, 214, 19 },
		{ 343, 172, 19 },
		{ 343, 173, 19 },
		{ 343, 174, 19 },
		{ 343, 175, 19 },
		{ 343, 176, 19 },
		{ 343, 177, 19 },
		{ 343, 178, 19 },
		{ 343, 179, 19 },
		{ 343, 180, 19 },
		{ 343, 181, 19 },
		{ 343, 182, 19 },
		{ 343, 183, 19 },
		{ 343, 184, 19 },
		{ 343, 185, 19 },
		{ 343, 186, 19 },
		{ 343, 187, 19 },
		{ 343, 188, 19 },
		{ 343, 189, 19 },
		{ 343, 190, 19 },
		{ 343, 191, 19 },
		{ 343, 192, 19 },
		{ 343, 193, 19 },
		{ 343, 194, 19 },
		{ 343, 195, 19 },
		{ 343, 196, 19 },
		{ 343, 197, 19 },
		{ 343, 198, 19 },
		{ 343, 199, 19 },
		{ 343, 200, 19 },
		{ 343, 201, 19 },
		{ 343, 202, 19 },
		{ 343, 203, 19 },
		{ 343, 204, 19 },
		{ 343, 205, 19 },
		{ 343, 206, 19 },
		{ 343, 207, 19 },
		{ 343, 208, 19 },
		{ 343, 209, 19 },
		{ 343, 210, 19 },
		{ 343, 211, 19 },
		{ 343, 212, 19 },
		{ 343, 213, 19 },
		{ 344, 171, 19 },
		{ 344, 172, 19 },
		{ 344, 173, 19 },
		{ 344, 174, 19 },
		{ 344, 175, 19 },
		{ 344, 176, 19 },
		{ 344, 177, 19 },
		{ 344, 178, 19 },
		{ 344, 179, 19 },
		{ 344, 180, 19 },
		{ 344, 181, 19 },
		{ 344, 182, 19 },
		{ 344, 183, 19 },
		{ 344, 184, 19 },
		{ 344, 185, 19 },
		{ 344, 186, 19 },
		{ 344, 187, 19 },
		{ 344, 188, 19 },
		{ 344, 189, 19 },
		{ 344, 190, 19 },
		{ 344, 191, 19 },
		{ 344, 192, 19 },
		{ 344, 193, 19 },
		{ 344, 194, 19 },
		{ 344, 195, 19 },
		{ 344, 196, 19 },
		{ 344, 197, 19 },
		{ 344, 198, 19 },
		{ 344, 199, 19 },
		{ 344, 200, 19 },
		{ 344, 201, 19 },
		{ 344, 202, 19 },
		{ 344, 203, 19 },
		{ 344, 204, 19 },
		{ 344, 205, 19 },
		{ 344, 206, 19 },
		{ 344, 207, 19 },
		{ 344, 208, 19 },
		{ 344, 209, 19 },
		{ 344, 210, 19 },
		{ 344, 211, 19 },
		{ 344, 212, 19 },
		{ 344, 213, 19 },
		{ 345, 170, 19 },
		{ 345, 171, 19 },
		{ 345, 172, 19 },
		{ 345, 173, 19 },
		{ 345, 174, 19 },
		{ 345, 175, 19 },
		{ 345, 176, 19 },
		{ 345, 177, 19 },
		{ 345, 178, 19 },
		{ 345, 179, 19 },
		{ 345, 180, 19 },
		{ 345, 181, 19 },
		{ 345, 182, 19 },
		{ 345, 183, 19 },
		{ 345, 184, 19 },
		{ 345, 185, 19 },
		{ 345, 186, 19 },
		{ 345, 187, 19 },
		{ 345, 188, 19 },
		{ 345, 189, 19 },
		{ 345, 190, 19 },
		{ 345, 191, 19 },
		{ 345, 192, 19 },
		{ 345, 193, 19 },
		{ 345, 194, 19 },
		{ 345, 195, 19 },
		{ 345, 196, 19 },
		{ 345, 197, 19 },
		{ 345, 198, 19 },
		{ 345, 199, 19 },
		{ 345, 200, 19 },
		{ 345, 201, 19 },
		{ 345, 202, 19 },
		{ 345, 203, 19 },
		{ 345, 204, 19 },
		{ 345, 205, 19 },
		{ 345, 206, 19 },
		{ 345, 207, 19 },
		{ 345, 208, 19 },
		{ 345, 209, 19 },
		{ 345, 210, 19 },
		{ 345, 211, 19 },
		{ 345, 212, 19 },
		{ 346, 169, 19 },
		{ 346, 170, 19 },
		{ 346, 171, 19 },
		{ 346, 172, 19 },
		{ 346, 173, 19 },
		{ 346, 174, 19 },
		{ 346, 175, 19 },
		{ 346, 176, 19 },
		{ 346, 177, 19 },
		{ 346, 178, 19 },
		{ 346, 179, 19 },
		{ 346, 180, 19 },
		{ 346, 181, 19 },
		{ 346, 182, 19 },
		{ 346, 183, 19 },
		{ 346, 184, 19 },
		{ 346, 185, 19 },
		{ 346, 186, 19 },
		{ 346, 187, 19 },
		{ 346, 188, 19 },
		{ 346, 189, 19 },
		{ 346, 190, 19 },
		{ 346, 191, 19 },
		{ 346, 192, 19 },
		{ 346, 193, 19 },
		{ 346, 194, 19 },
		{ 346, 195, 19 },
		{ 346, 196, 19 },
		{ 346, 197, 19 },
		{ 346, 198, 19 },
		{ 346, 199, 19 },
		{ 346, 200, 19 },
		{ 346, 201, 19 },
		{ 346, 202, 19 },
		{ 346, 203, 19 },
		{ 346, 204, 19 },
		{ 346, 205, 19 },
		{ 346, 206, 19 },
		{ 346, 207, 19 },
		{ 346, 208, 19 },
		{ 346, 209, 19 },
		{ 346, 210, 19 },
		{ 346, 211, 19 },
		{ 346, 212, 19 },
		{ 347, 168, 19 },
		{ 347, 169, 19 },
		{ 347, 170, 19 },
		{ 347, 171, 19 },
		{ 347, 172, 19 },
		{ 347, 173, 19 },
		{ 347, 174, 19 },
		{ 347, 175, 19 },
		{ 347, 176, 19 },
		{ 347, 177, 19 },
		{ 347, 178, 19 },
		{ 347, 179, 19 },
		{ 347, 180, 19 },
		{ 347, 181, 19 },
		{ 347, 182, 19 },
		{ 347, 183, 19 },
		{ 347, 184, 19 },
		{ 347, 185, 19 },
		{ 347, 186, 19 },
		{ 347, 187, 19 },
		{ 347, 188, 19 },
		{ 347, 189, 19 },
		{ 347, 190, 19 },
		{ 347, 191, 19 },
		{ 347, 192, 19 },
		{ 347, 193, 19 },
		{ 347, 194, 19 },
		{ 347, 195, 19 },
		{ 347, 196, 19 },
		{ 347, 197, 19 },
		{ 347, 198, 19 },
		{ 347, 199, 19 },
		{ 347, 200, 19 },
		{ 347, 201, 19 },
		{ 347, 202, 19 },
		{ 347, 203, 19 },
		{ 347, 204, 19 },
		{ 347, 205, 19 },
		{ 347, 206, 19 },
		{ 347, 207, 19 },
		{ 347, 208, 19 },
		{ 347, 209, 19 },
		{ 347, 210, 19 },
		{ 347, 211, 19 },
		{ 348, 167, 19 },
		{ 348, 168, 19 },
		{ 348, 169, 19 },
		{ 348, 170, 19 },
		{ 348, 171, 19 },
		{ 348, 172, 19 },
		{ 348, 173, 19 },
		{ 348, 174, 19 },
		{ 348, 175, 19 },
		{ 348, 176, 19 },
		{ 348, 177, 19 },
		{ 348, 178, 19 },
		{ 348, 179, 19 },
		{ 348, 180, 19 },
		{ 348, 181, 19 },
		{ 348, 182, 19 },
		{ 348, 183, 19 },
		{ 348, 184, 19 },
		{ 348, 185, 19 },
		{ 348, 186, 19 },
		{ 348, 187, 19 },
		{ 348, 188, 19 },
		{ 348, 189, 19 },
		{ 348, 190, 19 },
		{ 348, 191, 19 },
		{ 348, 192, 19 },
		{ 348, 193, 19 },
		{ 348, 194, 19 },
		{ 348, 195, 19 },
		{ 348, 196, 19 },
		{ 348, 197, 19 },
		{ 348, 198, 19 },
		{ 348, 199, 19 },
		{ 348, 200, 19 },
		{ 348, 201, 19 },
		{ 348, 202, 19 },
		{ 348, 203, 19 },
		{ 348, 204, 19 },
		{ 348, 205, 19 },
		{ 348, 206, 19 },
		{ 348, 207, 19 },
		{ 348, 208, 19 },
		{ 348, 209, 19 },
		{ 348, 210, 19 },
		{ 348, 211, 19 },
		{ 349, 166, 19 },
		{ 349, 167, 19 },
		{ 349, 168, 19 },
		{ 349, 169, 19 },
		{ 349, 170, 19 },
		{ 349, 171, 19 },
		{ 349, 172, 19 },
		{ 349, 173, 19 },
		{ 349, 174, 19 },
		{ 349, 175, 19 },
		{ 349, 176, 19 },
		{ 349, 177, 19 },
		{ 349, 178, 19 },
		{ 349, 179, 19 },
		{ 349, 180, 19 },
		{ 349, 181, 19 },
		{ 349, 182, 19 },
		{ 349, 183, 19 },
		{ 349, 184, 19 },
		{ 349, 185, 19 },
		{ 349, 186, 19 },
		{ 349, 187, 19 },
		{ 349, 188, 19 },
		{ 349, 189, 19 },
		{ 349, 190, 19 },
		{ 349, 191, 19 },
		{ 349, 192, 19 },
		{ 349, 193, 19 },
		{ 349, 194, 19 },
		{ 349, 195, 19 },
		{ 349, 196, 19 },
		{ 349, 197, 19 },
		{ 349, 198, 19 },
		{ 349, 199, 19 },
		{ 349, 200, 19 },
		{ 349, 201, 19 },
		{ 349, 202, 19 },
		{ 349, 203, 19 },
		{ 349, 204, 19 },
		{ 349, 205, 19 },
		{ 349, 206, 19 },
		{ 349, 207, 19 },
		{ 349, 208, 19 },
		{ 349, 209, 19 },
		{ 349, 210, 19 },
		{ 350, 165, 19 },
		{ 350, 166, 19 },
		{ 350, 167, 19 },
		{ 350, 168, 19 },
		{ 350, 169, 19 },
		{ 350, 170, 19 },
		{ 350, 171, 19 },
		{ 350, 172, 19 },
		{ 350, 173, 19 },
		{ 350, 174, 19 },
		{ 350, 175, 19 },
		{ 350, 176, 19 },
		{ 350, 177, 19 },
		{ 350, 178, 19 },
		{ 350, 179, 19 },
		{ 350, 180, 19 },
		{ 350, 181, 19 },
		{ 350, 182, 19 },
		{ 350, 183, 19 },
		{ 350, 184, 19 },
		{ 350, 185, 19 },
		{ 350, 186, 19 },
		{ 350, 187, 19 },
		{ 350, 188, 19 },
		{ 350, 189, 19 },
		{ 350, 190, 19 },
		{ 350, 191, 19 },
		{ 350, 192, 19 },
		{ 350, 193, 19 },
		{ 350, 194, 19 },
		{ 350, 195, 19 },
		{ 350, 196, 19 },
		{ 350, 197, 19 },
		{ 350, 198, 19 },
		{ 350, 199, 19 },
		{ 350, 200, 19 },
		{ 350, 201, 19 },
		{ 350, 202, 19 },
		{ 350, 203, 19 },
		{ 350, 204, 19 },
		{ 350, 205, 19 },
		{ 350, 206, 19 },
		{ 350, 207, 19 },
		{ 350, 208, 19 },
		{ 350, 209, 19 },
		{ 350, 210, 19 },
		{ 351, 164, 19 },
		{ 351, 165, 19 },
		{ 351, 166, 19 },
		{ 351, 167, 19 },
		{ 351, 168, 19 },
		{ 351, 169, 19 },
		{ 351, 170, 19 },
		{ 351, 171, 19 },
		{ 351, 172, 19 },
		{ 351, 173, 19 },
		{ 351, 174, 19 },
		{ 351, 175, 19 },
		{ 351, 176, 19 },
		{ 351, 177, 19 },
		{ 351, 178, 19 },
		{ 351, 179, 19 },
		{ 351, 180, 19 },
		{ 351, 181, 19 },
		{ 351, 182, 19 },
		{ 351, 183, 19 },
		{ 351, 184, 19 },
		{ 351, 185, 19 },
		{ 351, 186, 19 },
		{ 351, 187, 19 },
		{ 351, 188, 19 },
		{ 351, 189, 19 },
		{ 351, 190, 19 },
		{ 351, 191, 19 },
		{ 351, 192, 19 },
		{ 351, 193, 19 },
		{ 351, 194, 19 },
		{ 351, 195, 19 },
		{ 351, 196, 19 },
		{ 351, 197, 19 },
		{ 351, 198, 19 },
		{ 351, 199, 19 },
		{ 351, 200, 19 },
		{ 351, 201, 19 },
		{ 351, 202, 19 },
		{ 351, 203, 19 },
		{ 351, 204, 19 },
		{ 351, 205, 19 },
		{ 351, 206, 19 },
		{ 351, 207, 19 },
		{ 351, 208, 19 },
		{ 351, 209, 19 },
		{ 352, 163, 19 },
		{ 352, 164, 19 },
		{ 352, 165, 19 },
		{ 352, 166, 19 },
		{ 352, 167, 19 },
		{ 352, 168, 19 },
		{ 352, 169, 19 },
		{ 352, 170, 19 },
		{ 352, 171, 19 },
		{ 352, 172, 19 },
		{ 352, 173, 19 },
		{ 352, 174, 19 },
		{ 352, 175, 19 },
		{ 352, 176, 19 },
		{ 352, 177, 19 },
		{ 352, 178, 19 },
		{ 352, 179, 19 },
		{ 352, 180, 19 },
		{ 352, 181, 19 },
		{ 352, 182, 19 },
		{ 352, 183, 19 },
		{ 352, 184, 19 },
		{ 352, 185, 19 },
		{ 352, 186, 19 },
		{ 352, 187, 19 },
		{ 352, 188, 19 },
		{ 352, 189, 19 },
		{ 352, 190, 19 },
		{ 352, 191, 19 },
		{ 352, 192, 19 },
		{ 352, 193, 19 },
		{ 352, 194, 19 },
		{ 352, 195, 19 },
		{ 352, 196, 19 },
		{ 352, 197, 19 },
		{ 352, 198, 19 },
		{ 352, 199, 19 },
		{ 352, 200, 19 },
		{ 352, 201, 19 },
		{ 352, 202, 19 },
		{ 352, 203, 19 },
		{ 352, 204, 19 },
		{ 352, 205, 19 },
		{ 352, 206, 19 },
		{ 352, 207, 19 },
		{ 352, 208, 19 },
		{ 352, 209, 19 },
		{ 353, 162, 19 },
		{ 353, 163, 19 },
		{ 353, 164, 19 },
		{ 353, 165, 19 },
		{ 353, 166, 19 },
		{ 353, 167, 19 },
		{ 353, 168, 19 },
		{ 353, 169, 19 },
		{ 353, 170, 19 },
		{ 353, 171, 19 },
		{ 353, 172, 19 },
		{ 353, 173, 19 },
		{ 353, 174, 19 },
		{ 353, 175, 19 },
		{ 353, 176, 19 },
		{ 353, 177, 19 },
		{ 353, 178, 19 },
		{ 353, 179, 19 },
		{ 353, 180, 19 },
		{ 353, 181, 19 },
		{ 353, 182, 19 },
		{ 353, 183, 19 },
		{ 353, 184, 19 },
		{ 353, 185, 19 },
		{ 353, 186, 19 },
		{ 353, 187, 19 },
		{ 353, 188, 19 },
		{ 353, 189, 19 },
		{ 353, 190, 19 },
		{ 353, 191, 19 },
		{ 353, 192, 19 },
		{ 353, 193, 19 },
		{ 353, 194, 19 },
		{ 353, 195, 19 },
		{ 353, 196, 19 },
		{ 353, 197, 19 },
		{ 353, 198, 19 },
		{ 353, 199, 19 },
		{ 353, 200, 19 },
		{ 353, 201, 19 },
		{ 353, 202, 19 },
		{ 353, 203, 19 },
		{ 353, 204, 19 },
		{ 353, 205, 19 },
		{ 353, 206, 19 },
		{ 353, 207, 19 },
		{ 353, 208, 19 },
		{ 354, 161, 19 },
		{ 354, 162, 19 },
		{ 354, 163, 19 },
		{ 354, 164, 19 },
		{ 354, 165, 19 },
		{ 354, 166, 19 },
		{ 354, 167, 19 },
		{ 354, 168, 19 },
		{ 354, 169, 19 },
		{ 354, 170, 19 },
		{ 354, 171, 19 },
		{ 354, 172, 19 },
		{ 354, 173, 19 },
		{ 354, 174, 19 },
		{ 354, 175, 19 },
		{ 354, 176, 19 },
		{ 354, 177, 19 },
		{ 354, 178, 19 },
		{ 354, 179, 19 },
		{ 354, 180, 19 },
		{ 354, 181, 19 },
		{ 354, 182, 19 },
		{ 354, 183, 19 },
		{ 354, 184, 19 },
		{ 354, 185, 19 },
		{ 354, 186, 19 },
		{ 354, 187, 19 },
		{ 354, 188, 19 },
		{ 354, 189, 19 },
		{ 354, 190, 19 },
		{ 354, 191, 19 },
		{ 354, 192, 19 },
		{ 354, 193, 19 },
		{ 354, 194, 19 },
		{ 354, 195, 19 },
		{ 354, 196, 19 },
		{ 354, 197, 19 },
		{ 354, 198, 19 },
		{ 354, 199, 19 },
		{ 354, 200, 19 },
		{ 354, 201, 19 },
		{ 354, 202, 19 },
		{ 354, 203, 19 },
		{ 354, 204, 19 },
		{ 354, 205, 19 },
		{ 354, 206, 19 },
		{ 354, 207, 19 },
		{ 354, 208, 19 },
		{ 355, 160, 19 },
		{ 355, 161, 19 },
		{ 355, 162, 19 },
		{ 355, 163, 19 },
		{ 355, 164, 19 },
		{ 355, 165, 19 },
		{ 355, 166, 19 },
		{ 355, 167, 19 },
		{ 355, 168, 19 },
		{ 355, 169, 19 },
		{ 355, 170, 19 },
		{ 355, 171, 19 },
		{ 355, 172, 19 },
		{ 355, 173, 19 },
		{ 355, 174, 19 },
		{ 355, 175, 19 },
		{ 355, 176, 19 },
		{ 355, 177, 19 },
		{ 355, 178, 19 },
		{ 355, 179, 19 },
		{ 355, 180, 19 },
		{ 355, 181, 19 },
		{ 355, 182, 19 },
		{ 355, 183, 19 },
		{ 355, 184, 19 },
		{ 355, 185, 19 },
		{ 355, 186, 19 },
		{ 355, 187, 19 },
		{ 355, 188, 19 },
		{ 355, 189, 19 },
		{ 355, 190, 19 },
		{ 355, 191, 19 },
		{ 355, 192, 19 },
		{ 355, 193, 19 },
		{ 355, 194, 19 },
		{ 355, 195, 19 },
		{ 355, 196, 19 },
		{ 355, 197, 19 },
		{ 355, 198, 19 },
		{ 355, 199, 19 },
		{ 355, 200, 19 },
		{ 355, 201, 19 },
		{ 355, 202, 19 },
		{ 355, 203, 19 },
		{ 355, 204, 19 },
		{ 355, 205, 19 },
		{ 355, 206, 19 },
		{ 355, 207, 19 },
		{ 356, 159, 19 },
		{ 356, 160, 19 },
		{ 356, 161, 19 },
		{ 356, 162, 19 },
		{ 356, 163, 19 },
		{ 356, 164, 19 },
		{ 356, 165, 19 },
		{ 356, 166, 19 },
		{ 356, 167, 19 },
		{ 356, 168, 19 },
		{ 356, 169, 19 },
		{ 356, 170, 19 },
		{ 356, 171, 19 },
		{ 356, 172, 19 },
		{ 356, 173, 19 },
		{ 356, 174, 19 },
		{ 356, 175, 19 },
		{ 356, 176, 19 },
		{ 356, 177, 19 },
		{ 356, 178, 19 },
		{ 356, 179, 19 },
		{ 356, 180, 19 },
		{ 356, 181, 19 },
		{ 356, 182, 19 },
		{ 356, 183, 19 },
		{ 356, 184, 19 },
		{ 356, 185, 19 },
		{ 356, 186, 19 },
		{ 356, 187, 19 },
		{ 356, 188, 19 },
		{ 356, 189, 19 },
		{ 356, 190, 19 },
		{ 356, 191, 19 },
		{ 356, 192, 19 },
		{ 356, 193, 19 },
		{ 356, 194, 19 },
		{ 356, 195, 19 },
		{ 356, 196, 19 },
		{ 356, 197, 19 },
		{ 356, 198, 19 },
		{ 356, 199, 19 },
		{ 356, 200, 19 },
		{ 356, 201, 19 },
		{ 356, 202, 19 },
		{ 356, 203, 19 },
		{ 356, 204, 19 },
		{ 356, 205, 19 },
		{ 356, 206, 19 },
		{ 356, 207, 19 },
		{ 357, 158, 19 },
		{ 357, 159, 19 },
		{ 357, 160, 19 },
		{ 357, 161, 19 },
		{ 357, 162, 19 },
		{ 357, 163, 19 },
		{ 357, 164, 19 },
		{ 357, 165, 19 },
		{ 357, 166, 19 },
		{ 357, 167, 19 },
		{ 357, 168, 19 },
		{ 357, 169, 19 },
		{ 357, 170, 19 },
		{ 357, 171, 19 },
		{ 357, 172, 19 },
		{ 357, 173, 19 },
		{ 357, 174, 19 },
		{ 357, 175, 19 },
		{ 357, 176, 19 },
		{ 357, 177, 19 },
		{ 357, 178, 19 },
		{ 357, 179, 19 },
		{ 357, 180, 19 },
		{ 357, 181, 19 },
		{ 357, 182, 19 },
		{ 357, 183, 19 },
		{ 357, 184, 19 },
		{ 357, 185, 19 },
		{ 357, 186, 19 },
		{ 357, 187, 19 },
		{ 357, 188, 19 },
		{ 357, 189, 19 },
		{ 357, 190, 19 },
		{ 357, 191, 19 },
		{ 357, 192, 19 },
		{ 357, 193, 19 },
		{ 357, 194, 19 },
		{ 357, 195, 19 },
		{ 357, 196, 19 },
		{ 357, 197, 19 },
		{ 357, 198, 19 },
		{ 357, 199, 19 },
		{ 357, 200, 19 },
		{ 357, 201, 19 },
		{ 357, 202, 19 },
		{ 357, 203, 19 },
		{ 357, 204, 19 },
		{ 357, 205, 19 },
		{ 357, 206, 19 },
		{ 358, 157, 19 },
		{ 358, 158, 19 },
		{ 358, 159, 19 },
		{ 358, 160, 19 },
		{ 358, 161, 19 },
		{ 358, 162, 19 },
		{ 358, 163, 19 },
		{ 358, 164, 19 },
		{ 358, 165, 19 },
		{ 358, 166, 19 },
		{ 358, 167, 19 },
		{ 358, 168, 19 },
		{ 358, 169, 19 },
		{ 358, 170, 19 },
		{ 358, 171, 19 },
		{ 358, 172, 19 },
		{ 358, 173, 19 },
		{ 358, 174, 19 },
		{ 358, 175, 19 },
		{ 358, 176, 19 },
		{ 358, 177, 19 },
		{ 358, 178, 19 },
		{ 358, 179, 19 },
		{ 358, 180, 19 },
		{ 358, 181, 19 },
		{ 358, 182, 19 },
		{ 358, 183, 19 },
		{ 358, 184, 19 },
		{ 358, 185, 19 },
		{ 358, 186, 19 },
		{ 358, 187, 19 },
		{ 358, 188, 19 },
		{ 358, 189, 19 },
		{ 358, 190, 19 },
		{ 358, 191, 19 },
		{ 358, 192, 19 },
		{ 358, 193, 19 },
		{ 358, 194, 19 },
		{ 358, 195, 19 },
		{ 358, 196, 19 },
		{ 358, 197, 19 },
		{ 358, 198, 19 },
		{ 358, 199, 19 },
		{ 358, 200, 19 },
		{ 358, 201, 19 },
		{ 358, 202, 19 },
		{ 358, 203, 19 },
		{ 358, 204, 19 },
		{ 358, 205, 19 },
		{ 358, 206, 19 },
		{ 359, 156, 19 },
		{ 359, 157, 19 },
		{ 359, 158, 19 },
		{ 359, 159, 19 },
		{ 359, 160, 19 },
		{ 359, 161, 19 },
		{ 359, 162, 19 },
		{ 359, 163, 19 },
		{ 359, 164, 19 },
		{ 359, 165, 19 },
		{ 359, 166, 19 },
		{ 359, 167, 19 },
		{ 359, 168, 19 },
		{ 359, 169, 19 },
		{ 359, 170, 19 },
		{ 359, 171, 19 },
		{ 359, 172, 19 },
		{ 359, 173, 19 },
		{ 359, 174, 19 },
		{ 359, 175, 19 },
		{ 359, 176, 19 },
		{ 359, 177, 19 },
		{ 359, 178, 19 },
		{ 359, 179, 19 },
		{ 359, 180, 19 },
		{ 359, 181, 19 },
		{ 359, 182, 19 },
		{ 359, 183, 19 },
		{ 359, 184, 19 },
		{ 359, 185, 19 },
		{ 359, 186, 19 },
		{ 359, 187, 19 },
		{ 359, 188, 19 },
		{ 359, 189, 19 },
		{ 359, 190, 19 },
		{ 359, 191, 19 },
		{ 359, 192, 19 },
		{ 359, 193, 19 },
		{ 359, 194, 19 },
		{ 359, 195, 19 },
		{ 359, 196, 19 },
		{ 359, 197, 19 },
		{ 359, 198, 19 },
		{ 359, 199, 19 },
		{ 359, 200, 19 },
		{ 359, 201, 19 },
		{ 359, 202, 19 },
		{ 359, 203, 19 },
		{ 359, 204, 19 },
		{ 359, 205, 19 },
		{ 360, 155, 19 },
		{ 360, 156, 19 },
		{ 360, 157, 19 },
		{ 360, 158, 19 },
		{ 360, 159, 19 },
		{ 360, 160, 19 },
		{ 360, 161, 19 },
		{ 360, 162, 19 },
		{ 360, 163, 19 },
		{ 360, 164, 19 },
		{ 360, 165, 19 },
		{ 360, 166, 19 },
		{ 360, 167, 19 },
		{ 360, 168, 19 },
		{ 360, 169, 19 },
		{ 360, 170, 19 },
		{ 360, 171, 19 },
		{ 360, 172, 19 },
		{ 360, 173, 19 },
		{ 360, 174, 19 },
		{ 360, 175, 19 },
		{ 360, 176, 19 },
		{ 360, 177, 19 },
		{ 360, 178, 19 },
		{ 360, 179, 19 },
		{ 360, 180, 19 },
		{ 360, 181, 19 },
		{ 360, 182, 19 },
		{ 360, 183, 19 },
		{ 360, 184, 19 },
		{ 360, 185, 19 },
		{ 360, 186, 19 },
		{ 360, 187, 19 },
		{ 360, 188, 19 },
		{ 360, 189, 19 },
		{ 360, 190, 19 },
		{ 360, 191, 19 },
		{ 360, 192, 19 },
		{ 360, 193, 19 },
		{ 360, 194, 19 },
		{ 360, 195, 19 },
		{ 360, 196, 19 },
		{ 360, 197, 19 },
		{ 360, 198, 19 },
		{ 360, 199, 19 },
		{ 360, 200, 19 },
		{ 360, 201, 19 },
		{ 360, 202, 19 },
		{ 360, 203, 19 },
		{ 360, 204, 19 },
		{ 360, 205, 19 },
		{ 361, 154, 19 },
		{ 361, 155, 19 },
		{ 361, 156, 19 },
		{ 361, 157, 19 },
		{ 361, 158, 19 },
		{ 361, 159, 19 },
		{ 361, 160, 19 },
		{ 361, 161, 19 },
		{ 361, 162, 19 },
		{ 361, 163, 19 },
		{ 361, 164, 19 },
		{ 361, 165, 19 },
		{ 361, 166, 19 },
		{ 361, 167, 19 },
		{ 361, 168, 19 },
		{ 361, 169, 19 },
		{ 361, 170, 19 },
		{ 361, 171, 19 },
		{ 361, 172, 19 },
		{ 361, 173, 19 },
		{ 361, 174, 19 },
		{ 361, 175, 19 },
		{ 361, 176, 19 },
		{ 361, 177, 19 },
		{ 361, 178, 19 },
		{ 361, 179, 19 },
		{ 361, 180, 19 },
		{ 361, 181, 19 },
		{ 361, 182, 19 },
		{ 361, 183, 19 },
		{ 361, 184, 19 },
		{ 361, 185, 19 },
		{ 361, 186, 19 },
		{ 361, 187, 19 },
		{ 361, 188, 19 },
		{ 361, 189, 19 },
		{ 361, 190, 19 },
		{ 361, 191, 19 },
		{ 361, 192, 19 },
		{ 361, 193, 19 },
		{ 361, 194, 19 },
		{ 361, 195, 19 },
		{ 361, 196, 19 },
		{ 361, 197, 19 },
		{ 361, 198, 19 },
		{ 361, 199, 19 },
		{ 361, 200, 19 },
		{ 361, 201, 19 },
		{ 361, 202, 19 },
		{ 361, 203, 19 },
		{ 361, 204, 19 },
		{ 362, 153, 19 },
		{ 362, 154, 19 },
		{ 362, 155, 19 },
		{ 362, 156, 19 },
		{ 362, 157, 19 },
		{ 362, 158, 19 },
		{ 362, 159, 19 },
		{ 362, 160, 19 },
		{ 362, 161, 19 },
		{ 362, 162, 19 },
		{ 362, 163, 19 },
		{ 362, 164, 19 },
		{ 362, 165, 19 },
		{ 362, 166, 19 },
		{ 362, 167, 19 },
		{ 362, 168, 19 },
		{ 362, 169, 19 },
		{ 362, 170, 19 },
		{ 362, 171, 19 },
		{ 362, 172, 19 },
		{ 362, 173, 19 },
		{ 362, 174, 19 },
		{ 362, 175, 19 },
		{ 362, 176, 19 },
		{ 362, 177, 19 },
		{ 362, 178, 19 },
		{ 362, 179, 19 },
		{ 362, 180, 19 },
		{ 362, 181, 19 },
		{ 362, 182, 19 },
		{ 362, 183, 19 },
		{ 362, 184, 19 },
		{ 362, 185, 19 },
		{ 362, 186, 19 },
		{ 362, 187, 19 },
		{ 362, 188, 19 },
		{ 362, 189, 19 },
		{ 362, 190, 19 },
		{ 362, 191, 19 },
		{ 362, 192, 19 },
		{ 362, 193, 19 },
		{ 362, 194, 19 },
		{ 362, 195, 19 },
		{ 362, 196, 19 },
		{ 362, 197, 19 },
		{ 362, 198, 19 },
		{ 362, 199, 19 },
		{ 362, 200, 19 },
		{ 362, 201, 19 },
		{ 362, 202, 19 },
		{ 362, 203, 19 },
		{ 362, 204, 19 },
		{ 363, 152, 19 },
		{ 363, 153, 19 },
		{ 363, 154, 19 },
		{ 363, 155, 19 },
		{ 363, 156, 19 },
		{ 363, 157, 19 },
		{ 363, 158, 19 },
		{ 363, 159, 19 },
		{ 363, 160, 19 },
		{ 363, 161, 19 },
		{ 363, 162, 19 },
		{ 363, 163, 19 },
		{ 363, 164, 19 },
		{ 363, 165, 19 },
		{ 363, 166, 19 },
		{ 363, 167, 19 },
		{ 363, 168, 19 },
		{ 363, 169, 19 },
		{ 363, 170, 19 },
		{ 363, 171, 19 },
		{ 363, 172, 19 },
		{ 363, 173, 19 },
		{ 363, 174, 19 },
		{ 363, 175, 19 },
		{ 363, 176, 19 },
		{ 363, 177, 19 },
		{ 363, 178, 19 },
		{ 363, 179, 19 },
		{ 363, 180, 19 },
		{ 363, 181, 19 },
		{ 363, 182, 19 },
		{ 363, 183, 19 },
		{ 363, 184, 19 },
		{ 363, 185, 19 },
		{ 363, 186, 19 },
		{ 363, 187, 19 },
		{ 363, 188, 19 },
		{ 363, 189, 19 },
		{ 363, 190, 19 },
		{ 363, 191, 19 },
		{ 363, 192, 19 },
		{ 363, 193, 19 },
		{ 363, 194, 19 },
		{ 363, 195, 19 },
		{ 363, 196, 19 },
		{ 363, 197, 19 },
		{ 363, 198, 19 },
		{ 363, 199, 19 },
		{ 363, 200, 19 },
		{ 363, 201, 19 },
		{ 363, 202, 19 },
		{ 363, 203, 19 },
		{ 364, 151, 19 },
		{ 364, 152, 19 },
		{ 364, 153, 19 },
		{ 364, 154, 19 },
		{ 364, 155, 19 },
		{ 364, 156, 19 },
		{ 364, 157, 19 },
		{ 364, 158, 19 },
		{ 364, 159, 19 },
		{ 364, 160, 19 },
		{ 364, 161, 19 },
		{ 364, 162, 19 },
		{ 364, 163, 19 },
		{ 364, 164, 19 },
		{ 364, 165, 19 },
		{ 364, 166, 19 },
		{ 364, 167, 19 },
		{ 364, 168, 19 },
		{ 364, 169, 19 },
		{ 364, 170, 19 },
		{ 364, 171, 19 },
		{ 364, 172, 19 },
		{ 364, 173, 19 },
		{ 364, 174, 19 },
		{ 364, 175, 19 },
		{ 364, 176, 19 },
		{ 364, 177, 19 },
		{ 364, 178, 19 },
		{ 364, 179, 19 },
		{ 364, 180, 19 },
		{ 364, 181, 19 },
		{ 364, 182, 19 },
		{ 364, 183, 19 },
		{ 364, 184, 19 },
		{ 364, 185, 19 },
		{ 364, 186, 19 },
		{ 364, 187, 19 },
		{ 364, 188, 19 },
		{ 364, 189, 19 },
		{ 364, 190, 19 },
		{ 364, 191, 19 },
		{ 364, 192, 19 },
		{ 364, 193, 19 },
		{ 364, 194, 19 },
		{ 364, 195, 19 },
		{ 364, 196, 19 },
		{ 364, 197, 19 },
		{ 364, 198, 19 },
		{ 364, 199, 19 },
		{ 364, 200, 19 },
		{ 364, 201, 19 },
		{ 364, 202, 19 },
		{ 364, 203, 19 },
		{ 365, 150, 19 },
		{ 365, 151, 19 },
		{ 365, 152, 19 },
		{ 365, 153, 19 },
		{ 365, 154, 19 },
		{ 365, 155, 19 },
		{ 365, 156, 19 },
		{ 365, 157, 19 },
		{ 365, 158, 19 },
		{ 365, 159, 19 },
		{ 365, 160, 19 },
		{ 365, 161, 19 },
		{ 365, 162, 19 },
		{ 365, 163, 19 },
		{ 365, 164, 19 },
		{ 365, 165, 19 },
		{ 365, 166, 19 },
		{ 365, 167, 19 },
		{ 365, 168, 19 },
		{ 365, 169, 19 },
		{ 365, 170, 19 },
		{ 365, 171, 19 },
		{ 365, 172, 19 },
		{ 365, 173, 19 },
		{ 365, 174, 19 },
		{ 365, 175, 19 },
		{ 365, 176, 19 },
		{ 365, 177, 19 },
		{ 365, 178, 19 },
		{ 365, 179, 19 },
		{ 365, 180, 19 },
		{ 365, 181, 19 },
		{ 365, 182, 19 },
		{ 365, 183, 19 },
		{ 365, 184, 19 },
		{ 365, 185, 19 },
		{ 365, 186, 19 },
		{ 365, 187, 19 },
		{ 365, 188, 19 },
		{ 365, 189, 19 },
		{ 365, 190, 19 },
		{ 365, 191, 19 },
		{ 365, 192, 19 },
		{ 365, 193, 19 },
		{ 365, 194, 19 },
		{ 365, 195, 19 },
		{ 365, 196, 19 },
		{ 365, 197, 19 },
		{ 365, 198, 19 },
		{ 365, 199, 19 },
		{ 365, 200, 19 },
		{ 365, 201, 19 },
		{ 365, 202, 19 },
		{ 366, 149, 19 },
		{ 366, 150, 19 },
		{ 366, 151, 19 },
		{ 366, 152, 19 },
		{ 366, 153, 19 },
		{ 366, 154, 19 },
		{ 366, 155, 19 },
		{ 366, 156, 19 },
		{ 366, 157, 19 },
		{ 366, 158, 19 },
		{ 366, 159, 19 },
		{ 366, 160, 19 },
		{ 366, 161, 19 },
		{ 366, 162, 19 },
		{ 366, 163, 19 },
		{ 366, 164, 19 },
		{ 366, 165, 19 },
		{ 366, 166, 19 },
		{ 366, 167, 19 },
		{ 366, 168, 19 },
		{ 366, 169, 19 },
		{ 366, 170, 19 },
		{ 366, 171, 19 },
		{ 366, 172, 19 },
		{ 366, 173, 19 },
		{ 366, 174, 19 },
		{ 366, 175, 19 },
		{ 366, 176, 19 },
		{ 366, 177, 19 },
		{ 366, 178, 19 },
		{ 366, 179, 19 },
		{ 366, 180, 19 },
		{ 366, 181, 19 },
		{ 366, 182, 19 },
		{ 366, 183, 19 },
		{ 366, 184, 19 },
		{ 366, 185, 19 },
		{ 366, 186, 19 },
		{ 366, 187, 19 },
		{ 366, 188, 19 },
		{ 366, 189, 19 },
		{ 366, 190, 19 },
		{ 366, 191, 19 },
		{ 366, 192, 19 },
		{ 366, 193, 19 },
		{ 366, 194, 19 },
		{ 366, 195, 19 },
		{ 366, 196, 19 },
		{ 366, 197, 19 },
		{ 366, 198, 19 },
		{ 366, 199, 19 },
		{ 366, 200, 19 },
		{ 366, 201, 19 },
		{ 366, 202, 19 },
		{ 367, 148, 19 },
		{ 367, 149, 19 },
		{ 367, 150, 19 },
		{ 367, 151, 19 },
		{ 367, 152, 19 },
		{ 367, 153, 19 },
		{ 367, 154, 19 },
		{ 367, 155, 19 },
		{ 367, 156, 19 },
		{ 367, 157, 19 },
		{ 367, 158, 19 },
		{ 367, 159, 19 },
		{ 367, 160, 19 },
		{ 367, 161, 19 },
		{ 367, 162, 19 },
		{ 367, 163, 19 },
		{ 367, 164, 19 },
		{ 367, 165, 19 },
		{ 367, 166, 19 },
		{ 367, 167, 19 },
		{ 367, 168, 19 },
		{ 367, 169, 19 },
		{ 367, 170, 19 },
		{ 367, 171, 19 },
		{ 367, 172, 19 },
		{ 367, 173, 19 },
		{ 367, 174, 19 },
		{ 367, 175, 19 },
		{ 367, 176, 19 },
		{ 367, 177, 19 },
		{ 367, 178, 19 },
		{ 367, 179, 19 },
		{ 367, 180, 19 },
		{ 367, 181, 19 },
		{ 367, 182, 19 },
		{ 367, 183, 19 },
		{ 367, 184, 19 },
		{ 367, 185, 19 },
		{ 367, 186, 19 },
		{ 367, 187, 19 },
		{ 367, 188, 19 },
		{ 367, 189, 19 },
		{ 367, 190, 19 },
		{ 367, 191, 19 },
		{ 367, 192, 19 },
		{ 367, 193, 19 },
		{ 367, 194, 19 },
		{ 367, 195, 19 },
		{ 367, 196, 19 },
		{ 367, 197, 19 },
		{ 367, 198, 19 },
		{ 367, 199, 19 },
		{ 367, 200, 19 },
		{ 367, 201, 19 },
		{ 368, 147, 19 },
		{ 368, 148, 19 },
		{ 368, 149, 19 },
		{ 368, 150, 19 },
		{ 368, 151, 19 },
		{ 368, 152, 19 },
		{ 368, 153, 19 },
		{ 368, 154, 19 },
		{ 368, 155, 19 },
		{ 368, 156, 19 },
		{ 368, 157, 19 },
		{ 368, 158, 19 },
		{ 368, 159, 19 },
		{ 368, 160, 19 },
		{ 368, 161, 19 },
		{ 368, 162, 19 },
		{ 368, 163, 19 },
		{ 368, 164, 19 },
		{ 368, 165, 19 },
		{ 368, 166, 19 },
		{ 368, 167, 19 },
		{ 368, 168, 19 },
		{ 368, 169, 19 },
		{ 368, 170, 19 },
		{ 368, 171, 19 },
		{ 368, 172, 19 },
		{ 368, 173, 19 },
		{ 368, 174, 19 },
		{ 368, 175, 19 },
		{ 368, 176, 19 },
		{ 368, 177, 19 },
		{ 368, 178, 19 },
		{ 368, 179, 19 },
		{ 368, 180, 19 },
		{ 368, 181, 19 },
		{ 368, 182, 19 },
		{ 368, 183, 19 },
		{ 368, 184, 19 },
		{ 368, 185, 19 },
		{ 368, 186, 19 },
		{ 368, 187, 19 },
		{ 368, 188, 19 },
		{ 368, 189, 19 },
		{ 368, 190, 19 },
		{ 368, 191, 19 },
		{ 368, 192, 19 },
		{ 368, 193, 19 },
		{ 368, 194, 19 },
		{ 368, 195, 19 },
		{ 368, 196, 19 },
		{ 368, 197, 19 },
		{ 368, 198, 19 },
		{ 368, 199, 19 },
		{ 368, 200, 19 },
		{ 368, 201, 19 },
		{ 369, 146, 19 },
		{ 369, 147, 19 },
		{ 369, 148, 19 },
		{ 369, 149, 19 },
		{ 369, 150, 19 },
		{ 369, 151, 19 },
		{ 369, 152, 19 },
		{ 369, 153, 19 },
		{ 369, 154, 19 },
		{ 369, 155, 19 },
		{ 369, 156, 19 },
		{ 369, 157, 19 },
		{ 369, 158, 19 },
		{ 369, 159, 19 },
		{ 369, 160, 19 },
		{ 369, 161, 19 },
		{ 369, 162, 19 },
		{ 369, 163, 19 },
		{ 369, 164, 19 },
		{ 369, 165, 19 },
		{ 369, 166, 19 },
		{ 369, 167, 19 },
		{ 369, 168, 19 },
		{ 369, 169, 19 },
		{ 369, 170, 19 },
		{ 369, 171, 19 },
		{ 369, 172, 19 },
		{ 369, 173, 19 },
		{ 369, 174, 19 },
		{ 369, 175, 19 },
		{ 369, 176, 19 },
		{ 369, 177, 19 },
		{ 369, 178, 19 },
		{ 369, 179, 19 },
		{ 369, 180, 19 },
		{ 369, 181, 19 },
		{ 369, 182, 19 },
		{ 369, 183, 19 },
		{ 369, 184, 19 },
		{ 369, 185, 19 },
		{ 369, 186, 19 },
		{ 369, 187, 19 },
		{ 369, 188, 19 },
		{ 369, 189, 19 },
		{ 369, 190, 19 },
		{ 369, 191, 19 },
		{ 369, 192, 19 },
		{ 369, 193, 19 },
		{ 369, 194, 19 },
		{ 369, 195, 19 },
		{ 369, 196, 19 },
		{ 369, 197, 19 },
		{ 369, 198, 19 },
		{ 369, 199, 19 },
		{ 369, 200, 19 },
		{ 370, 145, 19 },
		{ 370, 146, 19 },
		{ 370, 147, 19 },
		{ 370, 148, 19 },
		{ 370, 149, 19 },
		{ 370, 150, 19 },
		{ 370, 151, 19 },
		{ 370, 152, 19 },
		{ 370, 153, 19 },
		{ 370, 154, 19 },
		{ 370, 155, 19 },
		{ 370, 156, 19 },
		{ 370, 157, 19 },
		{ 370, 158, 19 },
		{ 370, 159, 19 },
		{ 370, 160, 19 },
		{ 370, 161, 19 },
		{ 370, 162, 19 },
		{ 370, 163, 19 },
		{ 370, 164, 19 },
		{ 370, 165, 19 },
		{ 370, 166, 19 },
		{ 370, 167, 19 },
		{ 370, 168, 19 },
		{ 370, 169, 19 },
		{ 370, 170, 19 },
		{ 370, 171, 19 },
		{ 370, 172, 19 },
		{ 370, 173, 19 },
		{ 370, 174, 19 },
		{ 370, 175, 19 },
		{ 370, 176, 19 },
		{ 370, 177, 19 },
		{ 370, 178, 19 },
		{ 370, 179, 19 },
		{ 370, 180, 19 },
		{ 370, 181, 19 },
		{ 370, 182, 19 },
		{ 370, 183, 19 },
		{ 370, 184, 19 },
		{ 370, 185, 19 },
		{ 370, 186, 19 },
		{ 370, 187, 19 },
		{ 370, 188, 19 },
		{ 370, 189, 19 },
		{ 370, 190, 19 },
		{ 370, 191, 19 },
		{ 370, 192, 19 },
		{ 370, 193, 19 },
		{ 370, 194, 19 },
		{ 370, 195, 19 },
		{ 370, 196, 19 },
		{ 370, 197, 19 },
		{ 370, 198, 19 },
		{ 370, 199, 19 },
		{ 370, 200, 19 },
		{ 371, 144, 19 },
		{ 371, 145, 19 },
		{ 371, 146, 19 },
		{ 371, 147, 19 },
		{ 371, 148, 19 },
		{ 371, 149, 19 },
		{ 371, 150, 19 },
		{ 371, 151, 19 },
		{ 371, 152, 19 },
		{ 371, 153, 19 },
		{ 371, 154, 19 },
		{ 371, 155, 19 },
		{ 371, 156, 19 },
		{ 371, 157, 19 },
		{ 371, 158, 19 },
		{ 371, 159, 19 },
		{ 371, 160, 19 },
		{ 371, 161, 19 },
		{ 371, 162, 19 },
		{ 371, 163, 19 },
		{ 371, 164, 19 },
		{ 371, 165, 19 },
		{ 371, 166, 19 },
		{ 371, 167, 19 },
		{ 371, 168, 19 },
		{ 371, 169, 19 },
		{ 371, 170, 19 },
		{ 371, 171, 19 },
		{ 371, 172, 19 },
		{ 371, 173, 19 },
		{ 371, 174, 19 },
		{ 371, 175, 19 },
		{ 371, 176, 19 },
		{ 371, 177, 19 },
		{ 371, 178, 19 },
		{ 371, 179, 19 },
		{ 371, 180, 19 },
		{ 371, 181, 19 },
		{ 371, 182, 19 },
		{ 371, 183, 19 },
		{ 371, 184, 19 },
		{ 371, 185, 19 },
		{ 371, 186, 19 },
		{ 371, 187, 19 },
		{ 371, 188, 19 },
		{ 371, 189, 19 },
		{ 371, 190, 19 },
		{ 371, 191, 19 },
		{ 371, 192, 19 },
		{ 371, 193, 19 },
		{ 371, 194, 19 },
		{ 371, 195, 19 },
		{ 371, 196, 19 },
		{ 371, 197, 19 },
		{ 371, 198, 19 },
		{ 371, 199, 19 },
		{ 372, 143, 19 },
		{ 372, 144, 19 },
		{ 372, 145, 19 },
		{ 372, 146, 19 },
		{ 372, 147, 19 },
		{ 372, 148, 19 },
		{ 372, 149, 19 },
		{ 372, 150, 19 },
		{ 372, 151, 19 },
		{ 372, 152, 19 },
		{ 372, 153, 19 },
		{ 372, 154, 19 },
		{ 372, 155, 19 },
		{ 372, 156, 19 },
		{ 372, 157, 19 },
		{ 372, 158, 19 },
		{ 372, 159, 19 },
		{ 372, 160, 19 },
		{ 372, 161, 19 },
		{ 372, 162, 19 },
		{ 372, 163, 19 },
		{ 372, 164, 19 },
		{ 372, 165, 19 },
		{ 372, 166, 19 },
		{ 372, 167, 19 },
		{ 372, 168, 19 },
		{ 372, 169, 19 },
		{ 372, 170, 19 },
		{ 372, 171, 19 },
		{ 372, 172, 19 },
		{ 372, 173, 19 },
		{ 372, 174, 19 },
		{ 372, 175, 19 },
		{ 372, 176, 19 },
		{ 372, 177, 19 },
		{ 372, 178, 19 },
		{ 372, 179, 19 },
		{ 372, 180, 19 },
		{ 372, 181, 19 },
		{ 372, 182, 19 },
		{ 372, 183, 19 },
		{ 372, 184, 19 },
		{ 372, 185, 19 },
		{ 372, 186, 19 },
		{ 372, 187, 19 },
		{ 372, 188, 19 },
		{ 372, 189, 19 },
		{ 372, 190, 19 },
		{ 372, 191, 19 },
		{ 372, 192, 19 },
		{ 372, 193, 19 },
		{ 372, 194, 19 },
		{ 372, 195, 19 },
		{ 372, 196, 19 },
		{ 372, 197, 19 },
		{ 372, 198, 19 },
		{ 372, 199, 19 },
		{ 373, 142, 19 },
		{ 373, 143, 19 },
		{ 373, 144, 19 },
		{ 373, 145, 19 },
		{ 373, 146, 19 },
		{ 373, 147, 19 },
		{ 373, 148, 19 },
		{ 373, 149, 19 },
		{ 373, 150, 19 },
		{ 373, 151, 19 },
		{ 373, 152, 19 },
		{ 373, 153, 19 },
		{ 373, 154, 19 },
		{ 373, 155, 19 },
		{ 373, 156, 19 },
		{ 373, 157, 19 },
		{ 373, 158, 19 },
		{ 373, 159, 19 },
		{ 373, 160, 19 },
		{ 373, 161, 19 },
		{ 373, 162, 19 },
		{ 373, 163, 19 },
		{ 373, 164, 19 },
		{ 373, 165, 19 },
		{ 373, 166, 19 },
		{ 373, 167, 19 },
		{ 373, 168, 19 },
		{ 373, 169, 19 },
		{ 373, 170, 19 },
		{ 373, 171, 19 },
		{ 373, 172, 19 },
		{ 373, 173, 19 },
		{ 373, 174, 19 },
		{ 373, 175, 19 },
		{ 373, 176, 19 },
		{ 373, 177, 19 },
		{ 373, 178, 19 },
		{ 373, 179, 19 },
		{ 373, 180, 19 },
		{ 373, 181, 19 },
		{ 373, 182, 19 },
		{ 373, 183, 19 },
		{ 373, 184, 19 },
		{ 373, 185, 19 },
		{ 373, 186, 19 },
		{ 373, 187, 19 },
		{ 373, 188, 19 },
		{ 373, 189, 19 },
		{ 373, 190, 19 },
		{ 373, 191, 19 },
		{ 373, 192, 19 },
		{ 373, 193, 19 },
		{ 373, 194, 19 },
		{ 373, 195, 19 },
		{ 373, 196, 19 },
		{ 373, 197, 19 },
		{ 373, 198, 19 },
		{ 374, 141, 19 },
		{ 374, 142, 19 },
		{ 374, 143, 19 },
		{ 374, 144, 19 },
		{ 374, 145, 19 },
		{ 374, 146, 19 },
		{ 374, 147, 19 },
		{ 374, 148, 19 },
		{ 374, 149, 19 },
		{ 374, 150, 19 },
		{ 374, 151, 19 },
		{ 374, 152, 19 },
		{ 374, 153, 19 },
		{ 374, 154, 19 },
		{ 374, 155, 19 },
		{ 374, 156, 19 },
		{ 374, 157, 19 },
		{ 374, 158, 19 },
		{ 374, 159, 19 },
		{ 374, 160, 19 },
		{ 374, 161, 19 },
		{ 374, 162, 19 },
		{ 374, 163, 19 },
		{ 374, 164, 19 },
		{ 374, 165, 19 },
		{ 374, 166, 19 },
		{ 374, 167, 19 },
		{ 374, 168, 19 },
		{ 374, 169, 19 },
		{ 374, 170, 19 },
		{ 374, 171, 19 },
		{ 374, 172, 19 },
		{ 374, 173, 19 },
		{ 374, 174, 19 },
		{ 374, 175, 19 },
		{ 374, 176, 19 },
		{ 374, 177, 19 },
		{ 374, 178, 19 },
		{ 374, 179, 19 },
		{ 374, 180, 19 },
		{ 374, 181, 19 },
		{ 374, 182, 19 },
		{ 374, 183, 19 },
		{ 374, 184, 19 },
		{ 374, 185, 19 },
		{ 374, 186, 19 },
		{ 374, 187, 19 },
		{ 374, 188, 19 },
		{ 374, 189, 19 },
		{ 374, 190, 19 },
		{ 374, 191, 19 },
		{ 374, 192, 19 },
		{ 374, 193, 19 },
		{ 374, 194, 19 },
		{ 374, 195, 19 },
		{ 374, 196, 19 },
		{ 374, 197, 19 },
		{ 374, 198, 19 },
		{ 375, 140, 19 },
		{ 375, 141, 19 },
		{ 375, 142, 19 },
		{ 375, 143, 19 },
		{ 375, 144, 19 },
		{ 375, 145, 19 },
		{ 375, 146, 19 },
		{ 375, 147, 19 },
		{ 375, 148, 19 },
		{ 375, 149, 19 },
		{ 375, 150, 19 },
		{ 375, 151, 19 },
		{ 375, 152, 19 },
		{ 375, 153, 19 },
		{ 375, 154, 19 },
		{ 375, 155, 19 },
		{ 375, 156, 19 },
		{ 375, 157, 19 },
		{ 375, 158, 19 },
		{ 375, 159, 19 },
		{ 375, 160, 19 },
		{ 375, 161, 19 },
		{ 375, 162, 19 },
		{ 375, 163, 19 },
		{ 375, 164, 19 },
		{ 375, 165, 19 },
		{ 375, 166, 19 },
		{ 375, 167, 19 },
		{ 375, 168, 19 },
		{ 375, 169, 19 },
		{ 375, 170, 19 },
		{ 375, 171, 19 },
		{ 375, 172, 19 },
		{ 375, 173, 19 },
		{ 375, 174, 19 },
		{ 375, 175, 19 },
		{ 375, 176, 19 },
		{ 375, 177, 19 },
		{ 375, 178, 19 },
		{ 375, 179, 19 },
		{ 375, 180, 19 },
		{ 375, 181, 19 },
		{ 375, 182, 19 },
		{ 375, 183, 19 },
		{ 375, 184, 19 },
		{ 375, 185, 19 },
		{ 375, 186, 19 },
		{ 375, 187, 19 },
		{ 375, 188, 19 },
		{ 375, 189, 19 },
		{ 375, 190, 19 },
		{ 375, 191, 19 },
		{ 375, 192, 19 },
		{ 375, 193, 19 },
		{ 375, 194, 19 },
		{ 375, 195, 19 },
		{ 375, 196, 19 },
		{ 375, 197, 19 },
		{ 376, 139, 19 },
		{ 376, 140, 19 },
		{ 376, 141, 19 },
		{ 376, 142, 19 },
		{ 376, 143, 19 },
		{ 376, 144, 19 },
		{ 376, 145, 19 },
		{ 376, 146, 19 },
		{ 376, 147, 19 },
		{ 376, 148, 19 },
		{ 376, 149, 19 },
		{ 376, 150, 19 },
		{ 376, 151, 19 },
		{ 376, 152, 19 },
		{ 376, 153, 19 },
		{ 376, 154, 19 },
		{ 376, 155, 19 },
		{ 376, 156, 19 },
		{ 376, 157, 19 },
		{ 376, 158, 19 },
		{ 376, 159, 19 },
		{ 376, 160, 19 },
		{ 376, 161, 19 },
		{ 376, 162, 19 },
		{ 376, 163, 19 },
		{ 376, 164, 19 },
		{ 376, 165, 19 },
		{ 376, 166, 19 },
		{ 376, 167, 19 },
		{ 376, 168, 19 },
		{ 376, 169, 19 },
		{ 376, 170, 19 },
		{ 376, 171, 19 },
		{ 376, 172, 19 },
		{ 376, 173, 19 },
		{ 376, 174, 19 },
		{ 376, 175, 19 },
		{ 376, 176, 19 },
		{ 376, 177, 19 },
		{ 376, 178, 19 },
		{ 376, 179, 19 },
		{ 376, 180, 19 },
		{ 376, 181, 19 },
		{ 376, 182, 19 },
		{ 376, 183, 19 },
		{ 376, 184, 19 },
		{ 376, 185, 19 },
		{ 376, 186, 19 },
		{ 376, 187, 19 },
		{ 376, 188, 19 },
		{ 376, 189, 19 },
		{ 376, 190, 19 },
		{ 376, 191, 19 },
		{ 376, 192, 19 },
		{ 376, 193, 19 },
		{ 376, 194, 19 },
		{ 376, 195, 19 },
		{ 376, 196, 19 },
		{ 376, 197, 19 },
		{ 377, 138, 19 },
		{ 377, 139, 19 },
		{ 377, 140, 19 },
		{ 377, 141, 19 },
		{ 377, 142, 19 },
		{ 377, 143, 19 },
		{ 377, 144, 19 },
		{ 377, 145, 19 },
		{ 377, 146, 19 },
		{ 377, 147, 19 },
		{ 377, 148, 19 },
		{ 377, 149, 19 },
		{ 377, 150, 19 },
		{ 377, 151, 19 },
		{ 377, 152, 19 },
		{ 377, 153, 19 },
		{ 377, 154, 19 },
		{ 377, 155, 19 },
		{ 377, 156, 19 },
		{ 377, 157, 19 },
		{ 377, 158, 19 },
		{ 377, 159, 19 },
		{ 377, 160, 19 },
		{ 377, 161, 19 },
		{ 377, 162, 19 },
		{ 377, 163, 19 },
		{ 377, 164, 19 },
		{ 377, 165, 19 },
		{ 377, 166, 19 },
		{ 377, 167, 19 },
		{ 377, 168, 19 },
		{ 377, 169, 19 },
		{ 377, 170, 19 },
		{ 377, 171, 19 },
		{ 377, 172, 19 },
		{ 377, 173, 19 },
		{ 377, 174, 19 },
		{ 377, 175, 19 },
		{ 377, 176, 19 },
		{ 377, 177, 19 },
		{ 377, 178, 19 },
		{ 377, 179, 19 },
		{ 377, 180, 19 },
		{ 377, 181, 19 },
		{ 377, 182, 19 },
		{ 377, 183, 19 },
		{ 377, 184, 19 },
		{ 377, 185, 19 },
		{ 377, 186, 19 },
		{ 377, 187, 19 },
		{ 377, 188, 19 },
		{ 377, 189, 19 },
		{ 377, 190, 19 },
		{ 377, 191, 19 },
		{ 377, 192, 19 },
		{ 377, 193, 19 },
		{ 377, 194, 19 },
		{ 377, 195, 19 },
		{ 377, 196, 19 },
		{ 378, 137, 19 },
		{ 378, 138, 19 },
		{ 378, 139, 19 },
		{ 378, 140, 19 },
		{ 378, 141, 19 },
		{ 378, 142, 19 },
		{ 378, 143, 19 },
		{ 378, 144, 19 },
		{ 378, 145, 19 },
		{ 378, 146, 19 },
		{ 378, 147, 19 },
		{ 378, 148, 19 },
		{ 378, 149, 19 },
		{ 378, 150, 19 },
		{ 378, 151, 19 },
		{ 378, 152, 19 },
		{ 378, 153, 19 },
		{ 378, 154, 19 },
		{ 378, 155, 19 },
		{ 378, 156, 19 },
		{ 378, 157, 19 },
		{ 378, 158, 19 },
		{ 378, 159, 19 },
		{ 378, 160, 19 },
		{ 378, 161, 19 },
		{ 378, 162, 19 },
		{ 378, 163, 19 },
		{ 378, 164, 19 },
		{ 378, 165, 19 },
		{ 378, 166, 19 },
		{ 378, 167, 19 },
		{ 378, 168, 19 },
		{ 378, 169, 19 },
		{ 378, 170, 19 },
		{ 378, 171, 19 },
		{ 378, 172, 19 },
		{ 378, 173, 19 },
		{ 378, 174, 19 },
		{ 378, 175, 19 },
		{ 378, 176, 19 },
		{ 378, 177, 19 },
		{ 378, 178, 19 },
		{ 378, 179, 19 },
		{ 378, 180, 19 },
		{ 378, 181, 19 },
		{ 378, 182, 19 },
		{ 378, 183, 19 },
		{ 378, 184, 19 },
		{ 378, 185, 19 },
		{ 378, 186, 19 },
		{ 378, 187, 19 },
		{ 378, 188, 19 },
		{ 378, 189, 19 },
		{ 378, 190, 19 },
		{ 378, 191, 19 },
		{ 378, 192, 19 },
		{ 378, 193, 19 },
		{ 378, 194, 19 },
		{ 378, 195, 19 },
		{ 379, 136, 19 },
		{ 379, 137, 19 },
		{ 379, 138, 19 },
		{ 379, 139, 19 },
		{ 379, 140, 19 },
		{ 379, 141, 19 },
		{ 379, 142, 19 },
		{ 379, 143, 19 },
		{ 379, 144, 19 },
		{ 379, 145, 19 },
		{ 379, 146, 19 },
		{ 379, 147, 19 },
		{ 379, 148, 19 },
		{ 379, 149, 19 },
		{ 379, 150, 19 },
		{ 379, 151, 19 },
		{ 379, 152, 19 },
		{ 379, 153, 19 },
		{ 379, 154, 19 },
		{ 379, 155, 19 },
		{ 379, 156, 19 },
		{ 379, 157, 19 },
		{ 379, 158, 19 },
		{ 379, 159, 19 },
		{ 379, 160, 19 },
		{ 379, 161, 19 },
		{ 379, 162, 19 },
		{ 379, 163, 19 },
		{ 379, 164, 19 },
		{ 379, 165, 19 },
		{ 379, 166, 19 },
		{ 379, 167, 19 },
		{ 379, 168, 19 },
		{ 379, 169, 19 },
		{ 379, 170, 19 },
		{ 379, 171, 19 },
		{ 379, 172, 19 },
		{ 379, 173, 19 },
		{ 379, 174, 19 },
		{ 379, 175, 19 },
		{ 379, 176, 19 },
		{ 379, 177, 19 },
		{ 379, 178, 19 },
		{ 379, 179, 19 },
		{ 379, 180, 19 },
		{ 379, 181, 19 },
		{ 379, 182, 19 },
		{ 379, 183, 19 },
		{ 379, 184, 19 },
		{ 379, 185, 19 },
		{ 379, 186, 19 },
		{ 379, 187, 19 },
		{ 379, 188, 19 },
		{ 379, 189, 19 },
		{ 379, 190, 19 },
		{ 379, 191, 19 },
		{ 379, 192, 19 },
		{ 379, 193, 19 },
		{ 379, 194, 19 },
		{ 379, 195, 19 },
		{ 380, 135, 19 },
		{ 380, 136, 19 },
		{ 380, 137, 19 },
		{ 380, 138, 19 },
		{ 380, 139, 19 },
		{ 380, 140, 19 },
		{ 380, 141, 19 },
		{ 380, 142, 19 },
		{ 380, 143, 19 },
		{ 380, 144, 19 },
		{ 380, 145, 19 },
		{ 380, 146, 19 },
		{ 380, 147, 19 },
		{ 380, 148, 19 },
		{ 380, 149, 19 },
		{ 380, 150, 19 },
		{ 380, 151, 19 },
		{ 380, 152, 19 },
		{ 380, 153, 19 },
		{ 380, 154, 19 },
		{ 380, 155, 19 },
		{ 380, 156, 19 },
		{ 380, 157, 19 },
		{ 380, 158, 19 },
		{ 380, 159, 19 },
		{ 380, 160, 19 },
		{ 380, 161, 19 },
		{ 380, 162, 19 },
		{ 380, 163, 19 },
		{ 380, 164, 19 },
		{ 380, 165, 19 },
		{ 380, 166, 19 },
		{ 380, 167, 19 },
		{ 380, 168, 19 },
		{ 380, 169, 19 },
		{ 380, 170, 19 },
		{ 380, 171, 19 },
		{ 380, 172, 19 },
		{ 380, 173, 19 },
		{ 380, 174, 19 },
		{ 380, 175, 19 },
		{ 380, 176, 19 },
		{ 380, 177, 19 },
		{ 380, 178, 19 },
		{ 380, 179, 19 },
		{ 380, 180, 19 },
		{ 380, 181, 19 },
		{ 380, 182, 19 },
		{ 380, 183, 19 },
		{ 380, 184, 19 },
		{ 380, 185, 19 },
		{ 380, 186, 19 },
		{ 380, 187, 19 },
		{ 380, 188, 19 },
		{ 380, 189, 19 },
		{ 380, 190, 19 },
		{ 380, 191, 19 },
		{ 380, 192, 19 },
		{ 380, 193, 19 },
		{ 380, 194, 19 },
		{ 381, 135, 19 },
		{ 381, 136, 19 },
		{ 381, 137, 19 },
		{ 381, 138, 19 },
		{ 381, 139, 19 },
		{ 381, 140, 19 },
		{ 381, 141, 19 },
		{ 381, 142, 19 },
		{ 381, 143, 19 },
		{ 381, 144, 19 },
		{ 381, 145, 19 },
		{ 381, 146, 19 },
		{ 381, 147, 19 },
		{ 381, 148, 19 },
		{ 381, 149, 19 },
		{ 381, 150, 19 },
		{ 381, 151, 19 },
		{ 381, 152, 19 },
		{ 381, 153, 19 },
		{ 381, 154, 19 },
		{ 381, 155, 19 },
		{ 381, 156, 19 },
		{ 381, 157, 19 },
		{ 381, 158, 19 },
		{ 381, 159, 19 },
		{ 381, 160, 19 },
		{ 381, 161, 19 },
		{ 381, 162, 19 },
		{ 381, 163, 19 },
		{ 381, 164, 19 },
		{ 381, 165, 19 },
		{ 381, 166, 19 },
		{ 381, 167, 19 },
		{ 381, 168, 19 },
		{ 381, 169, 19 },
		{ 381, 170, 19 },
		{ 381, 171, 19 },
		{ 381, 172, 19 },
		{ 381, 173, 19 },
		{ 381, 174, 19 },
		{ 381, 175, 19 },
		{ 381, 176, 19 },
		{ 381, 177, 19 },
		{ 381, 178, 19 },
		{ 381, 179, 19 },
		{ 381, 180, 19 },
		{ 381, 181, 19 },
		{ 381, 182, 19 },
		{ 381, 183, 19 },
		{ 381, 184, 19 },
		{ 381, 185, 19 },
		{ 381, 186, 19 },
		{ 381, 187, 19 },
		{ 381, 188, 19 },
		{ 381, 189, 19 },
		{ 381, 190, 19 },
		{ 381, 191, 19 },
		{ 381, 192, 19 },
		{ 381, 193, 19 },
		{ 382, 135, 19 },
		{ 382, 136, 19 },
		{ 382, 137, 19 },
		{ 382, 138, 19 },
		{ 382, 139, 19 },
		{ 382, 140, 19 },
		{ 382, 141, 19 },
		{ 382, 142, 19 },
		{ 382, 143, 19 },
		{ 382, 144, 19 },
		{ 382, 145, 19 },
		{ 382, 146, 19 },
		{ 382, 147, 19 },
		{ 382, 148, 19 },
		{ 382, 149, 19 },
		{ 382, 150, 19 },
		{ 382, 151, 19 },
		{ 382, 152, 19 },
		{ 382, 153, 19 },
		{ 382, 154, 19 },
		{ 382, 155, 19 },
		{ 382, 156, 19 },
		{ 382, 157, 19 },
		{ 382, 158, 19 },
		{ 382, 159, 19 },
		{ 382, 160, 19 },
		{ 382, 161, 19 },
		{ 382, 162, 19 },
		{ 382, 163, 19 },
		{ 382, 164, 19 },
		{ 382, 165, 19 },
		{ 382, 166, 19 },
		{ 382, 167, 19 },
		{ 382, 168, 19 },
		{ 382, 169, 19 },
		{ 382, 170, 19 },
		{ 382, 171, 19 },
		{ 382, 172, 19 },
		{ 382, 173, 19 },
		{ 382, 174, 19 },
		{ 382, 175, 19 },
		{ 382, 176, 19 },
		{ 382, 177, 19 },
		{ 382, 178, 19 },
		{ 382, 179, 19 },
		{ 382, 180, 19 },
		{ 382, 181, 19 },
		{ 382, 182, 19 },
		{ 382, 183, 19 },
		{ 382, 184, 19 },
		{ 382, 185, 19 },
		{ 382, 186, 19 },
		{ 382, 187, 19 },
		{ 382, 188, 19 },
		{ 382, 189, 19 },
		{ 382, 190, 19 },
		{ 382, 191, 19 },
		{ 383, 138, 19 },
		{ 383, 139, 19 },
		{ 383, 140, 19 },
		{ 383, 141, 19 },
		{ 383, 142, 19 },
		{ 383, 143, 19 },
		{ 383, 144, 19 },
		{ 383, 145, 19 },
		{ 383, 146, 19 },
		{ 383, 147, 19 },
		{ 383, 148, 19 },
		{ 383, 149, 19 },
		{ 383, 150, 19 },
		{ 383, 151, 19 },
		{ 383, 152, 19 },
		{ 383, 153, 19 },
		{ 383, 154, 19 },
		{ 383, 155, 19 },
		{ 383, 156, 19 },
		{ 383, 157, 19 },
		{ 383, 158, 19 },
		{ 383, 159, 19 },
		{ 383, 160, 19 },
		{ 383, 161, 19 },
		{ 383, 162, 19 },
		{ 383, 163, 19 },
		{ 383, 164, 19 },
		{ 383, 165, 19 },
		{ 383, 166, 19 },
		{ 383, 167, 19 },
		{ 383, 168, 19 },
		{ 383, 169, 19 },
		{ 383, 170, 19 },
		{ 383, 171, 19 },
		{ 383, 172, 19 },
		{ 383, 173, 19 },
		{ 383, 174, 19 },
		{ 383, 175, 19 },
		{ 383, 176, 19 },
		{ 383, 177, 19 },
		{ 383, 178, 19 },
		{ 383, 179, 19 },
		{ 383, 180, 19 },
		{ 383, 181, 19 },
		{ 383, 182, 19 },
		{ 383, 183, 19 },
		{ 383, 184, 19 },
		{ 383, 185, 19 },
		{ 383, 186, 19 },
		{ 383, 187, 19 },
		{ 131, 296, 20 },
		{ 131, 297, 20 },
		{ 131, 298, 20 },
		{ 131, 299, 20 },
		{ 131, 300, 20 },
		{ 131, 301, 20 },
		{ 131, 302, 20 },
		{ 131, 303, 20 },
		{ 131, 304, 20 },
		{ 131, 305, 20 },
		{ 131, 306, 20 },
		{ 131, 307, 20 },
		{ 131, 308, 20 },
		{ 131, 309, 20 },
		{ 131, 310, 20 },
		{ 131, 311, 20 },
		{ 131, 312, 20 },
		{ 131, 313, 20 },
		{ 131, 314, 20 },
		{ 131, 315, 20 },
		{ 131, 316, 20 },
		{ 131, 317, 20 },
		{ 131, 318, 20 },
		{ 131, 319, 20 },
		{ 131, 320, 20 },
		{ 131, 321, 20 },
		{ 131, 322, 20 },
		{ 131, 323, 20 },
		{ 131, 324, 20 },
		{ 131, 325, 20 },
		{ 131, 326, 20 },
		{ 131, 327, 20 },
		{ 131, 328, 20 },
		{ 131, 329, 20 },
		{ 131, 330, 20 },
		{ 131, 331, 20 },
		{ 131, 332, 20 },
		{ 131, 333, 20 },
		{ 131, 334, 20 },
		{ 131, 335, 20 },
		{ 131, 336, 20 },
		{ 131, 337, 20 },
		{ 131, 338, 20 },
		{ 131, 339, 20 },
		{ 131, 340, 20 },
		{ 131, 341, 20 },
		{ 131, 342, 20 },
		{ 131, 343, 20 },
		{ 131, 344, 20 },
		{ 132, 292, 20 },
		{ 132, 293, 20 },
		{ 132, 294, 20 },
		{ 132, 295, 20 },
		{ 132, 296, 20 },
		{ 132, 297, 20 },
		{ 132, 298, 20 },
		{ 132, 299, 20 },
		{ 132, 300, 20 },
		{ 132, 301, 20 },
		{ 132, 302, 20 },
		{ 132, 303, 20 },
		{ 132, 304, 20 },
		{ 132, 305, 20 },
		{ 132, 306, 20 },
		{ 132, 307, 20 },
		{ 132, 308, 20 },
		{ 132, 309, 20 },
		{ 132, 310, 20 },
		{ 132, 311, 20 },
		{ 132, 312, 20 },
		{ 132, 313, 20 },
		{ 132, 314, 20 },
		{ 132, 315, 20 },
		{ 132, 316, 20 },
		{ 132, 317, 20 },
		{ 132, 318, 20 },
		{ 132, 319, 20 },
		{ 132, 320, 20 },
		{ 132, 321, 20 },
		{ 132, 322, 20 },
		{ 132, 323, 20 },
		{ 132, 324, 20 },
		{ 132, 325, 20 },
		{ 132, 326, 20 },
		{ 132, 327, 20 },
		{ 132, 328, 20 },
		{ 132, 329, 20 },
		{ 132, 330, 20 },
		{ 132, 331, 20 },
		{ 132, 332, 20 },
		{ 132, 333, 20 },
		{ 132, 334, 20 },
		{ 132, 335, 20 },
		{ 132, 336, 20 },
		{ 132, 337, 20 },
		{ 132, 338, 20 },
		{ 132, 339, 20 },
		{ 132, 340, 20 },
		{ 132, 341, 20 },
		{ 132, 342, 20 },
		{ 132, 343, 20 },
		{ 132, 344, 20 },
		{ 132, 345, 20 },
		{ 132, 346, 20 },
		{ 132, 347, 20 },
		{ 133, 290, 20 },
		{ 133, 291, 20 },
		{ 133, 292, 20 },
		{ 133, 293, 20 },
		{ 133, 294, 20 },
		{ 133, 295, 20 },
		{ 133, 296, 20 },
		{ 133, 297, 20 },
		{ 133, 298, 20 },
		{ 133, 299, 20 },
		{ 133, 300, 20 },
		{ 133, 301, 20 },
		{ 133, 302, 20 },
		{ 133, 303, 20 },
		{ 133, 304, 20 },
		{ 133, 305, 20 },
		{ 133, 306, 20 },
		{ 133, 307, 20 },
		{ 133, 308, 20 },
		{ 133, 309, 20 },
		{ 133, 310, 20 },
		{ 133, 311, 20 },
		{ 133, 312, 20 },
		{ 133, 313, 20 },
		{ 133, 314, 20 },
		{ 133, 315, 20 },
		{ 133, 316, 20 },
		{ 133, 317, 20 },
		{ 133, 318, 20 },
		{ 133, 319, 20 },
		{ 133, 320, 20 },
		{ 133, 321, 20 },
		{ 133, 322, 20 },
		{ 133, 323, 20 },
		{ 133, 324, 20 },
		{ 133, 325, 20 },
		{ 133, 326, 20 },
		{ 133, 327, 20 },
		{ 133, 328, 20 },
		{ 133, 329, 20 },
		{ 133, 330, 20 },
		{ 133, 331, 20 },
		{ 133, 332, 20 },
		{ 133, 333, 20 },
		{ 133, 334, 20 },
		{ 133, 335, 20 },
		{ 133, 336, 20 },
		{ 133, 337, 20 },
		{ 133, 338, 20 },
		{ 133, 339, 20 },
		{ 133, 340, 20 },
		{ 133, 341, 20 },
		{ 133, 342, 20 },
		{ 133, 343, 20 },
		{ 133, 344, 20 },
		{ 133, 345, 20 },
		{ 133, 346, 20 },
		{ 133, 347, 20 },
		{ 133, 348, 20 },
		{ 134, 289, 20 },
		{ 134, 290, 20 },
		{ 134, 291, 20 },
		{ 134, 292, 20 },
		{ 134, 293, 20 },
		{ 134, 294, 20 },
		{ 134, 295, 20 },
		{ 134, 296, 20 },
		{ 134, 297, 20 },
		{ 134, 298, 20 },
		{ 134, 299, 20 },
		{ 134, 300, 20 },
		{ 134, 301, 20 },
		{ 134, 302, 20 },
		{ 134, 303, 20 },
		{ 134, 304, 20 },
		{ 134, 305, 20 },
		{ 134, 306, 20 },
		{ 134, 307, 20 },
		{ 134, 308, 20 },
		{ 134, 309, 20 },
		{ 134, 310, 20 },
		{ 134, 311, 20 },
		{ 134, 312, 20 },
		{ 134, 313, 20 },
		{ 134, 314, 20 },
		{ 134, 315, 20 },
		{ 134, 316, 20 },
		{ 134, 317, 20 },
		{ 134, 318, 20 },
		{ 134, 319, 20 },
		{ 134, 320, 20 },
		{ 134, 321, 20 },
		{ 134, 322, 20 },
		{ 134, 323, 20 },
		{ 134, 324, 20 },
		{ 134, 325, 20 },
		{ 134, 326, 20 },
		{ 134, 327, 20 },
		{ 134, 328, 20 },
		{ 134, 329, 20 },
		{ 134, 330, 20 },
		{ 134, 331, 20 },
		{ 134, 332, 20 },
		{ 134, 333, 20 },
		{ 134, 334, 20 },
		{ 134, 335, 20 },
		{ 134, 336, 20 },
		{ 134, 337, 20 },
		{ 134, 338, 20 },
		{ 134, 339, 20 },
		{ 134, 340, 20 },
		{ 134, 341, 20 },
		{ 134, 342, 20 },
		{ 134, 343, 20 },
		{ 134, 344, 20 },
		{ 134, 345, 20 },
		{ 134, 346, 20 },
		{ 134, 347, 20 },
		{ 134, 348, 20 },
		{ 135, 289, 20 },
		{ 135, 290, 20 },
		{ 135, 291, 20 },
		{ 135, 292, 20 },
		{ 135, 293, 20 },
		{ 135, 294, 20 },
		{ 135, 295, 20 },
		{ 135, 296, 20 },
		{ 135, 297, 20 },
		{ 135, 298, 20 },
		{ 135, 299, 20 },
		{ 135, 300, 20 },
		{ 135, 301, 20 },
		{ 135, 302, 20 },
		{ 135, 303, 20 },
		{ 135, 304, 20 },
		{ 135, 305, 20 },
		{ 135, 306, 20 },
		{ 135, 307, 20 },
		{ 135, 308, 20 },
		{ 135, 309, 20 },
		{ 135, 310, 20 },
		{ 135, 311, 20 },
		{ 135, 312, 20 },
		{ 135, 313, 20 },
		{ 135, 314, 20 },
		{ 135, 315, 20 },
		{ 135, 316, 20 },
		{ 135, 317, 20 },
		{ 135, 318, 20 },
		{ 135, 319, 20 },
		{ 135, 320, 20 },
		{ 135, 321, 20 },
		{ 135, 322, 20 },
		{ 135, 323, 20 },
		{ 135, 324, 20 },
		{ 135, 325, 20 },
		{ 135, 326, 20 },
		{ 135, 327, 20 },
		{ 135, 328, 20 },
		{ 135, 329, 20 },
		{ 135, 330, 20 },
		{ 135, 331, 20 },
		{ 135, 332, 20 },
		{ 135, 333, 20 },
		{ 135, 334, 20 },
		{ 135, 335, 20 },
		{ 135, 336, 20 },
		{ 135, 337, 20 },
		{ 135, 338, 20 },
		{ 135, 339, 20 },
		{ 135, 340, 20 },
		{ 135, 341, 20 },
		{ 135, 342, 20 },
		{ 135, 343, 20 },
		{ 135, 344, 20 },
		{ 135, 345, 20 },
		{ 135, 346, 20 },
		{ 135, 347, 20 },
		{ 136, 288, 20 },
		{ 136, 289, 20 },
		{ 136, 290, 20 },
		{ 136, 291, 20 },
		{ 136, 292, 20 },
		{ 136, 293, 20 },
		{ 136, 294, 20 },
		{ 136, 295, 20 },
		{ 136, 296, 20 },
		{ 136, 297, 20 },
		{ 136, 298, 20 },
		{ 136, 299, 20 },
		{ 136, 300, 20 },
		{ 136, 301, 20 },
		{ 136, 302, 20 },
		{ 136, 303, 20 },
		{ 136, 304, 20 },
		{ 136, 305, 20 },
		{ 136, 306, 20 },
		{ 136, 307, 20 },
		{ 136, 308, 20 },
		{ 136, 309, 20 },
		{ 136, 310, 20 },
		{ 136, 311, 20 },
		{ 136, 312, 20 },
		{ 136, 313, 20 },
		{ 136, 314, 20 },
		{ 136, 315, 20 },
		{ 136, 316, 20 },
		{ 136, 317, 20 },
		{ 136, 318, 20 },
		{ 136, 319, 20 },
		{ 136, 320, 20 },
		{ 136, 321, 20 },
		{ 136, 322, 20 },
		{ 136, 323, 20 },
		{ 136, 324, 20 },
		{ 136, 325, 20 },
		{ 136, 326, 20 },
		{ 136, 327, 20 },
		{ 136, 328, 20 },
		{ 136, 329, 20 },
		{ 136, 330, 20 },
		{ 136, 331, 20 },
		{ 136, 332, 20 },
		{ 136, 333, 20 },
		{ 136, 334, 20 },
		{ 136, 335, 20 },
		{ 136, 336, 20 },
		{ 136, 337, 20 },
		{ 136, 338, 20 },
		{ 136, 339, 20 },
		{ 136, 340, 20 },
		{ 136, 341, 20 },
		{ 136, 342, 20 },
		{ 136, 343, 20 },
		{ 136, 344, 20 },
		{ 136, 345, 20 },
		{ 136, 346, 20 },
		{ 136, 347, 20 },
		{ 137, 288, 20 },
		{ 137, 289, 20 },
		{ 137, 290, 20 },
		{ 137, 291, 20 },
		{ 137, 292, 20 },
		{ 137, 293, 20 },
		{ 137, 294, 20 },
		{ 137, 295, 20 },
		{ 137, 296, 20 },
		{ 137, 297, 20 },
		{ 137, 298, 20 },
		{ 137, 299, 20 },
		{ 137, 300, 20 },
		{ 137, 301, 20 },
		{ 137, 302, 20 },
		{ 137, 303, 20 },
		{ 137, 304, 20 },
		{ 137, 305, 20 },
		{ 137, 306, 20 },
		{ 137, 307, 20 },
		{ 137, 308, 20 },
		{ 137, 309, 20 },
		{ 137, 310, 20 },
		{ 137, 311, 20 },
		{ 137, 312, 20 },
		{ 137, 313, 20 },
		{ 137, 314, 20 },
		{ 137, 315, 20 },
		{ 137, 316, 20 },
		{ 137, 317, 20 },
		{ 137, 318, 20 },
		{ 137, 319, 20 },
		{ 137, 320, 20 },
		{ 137, 321, 20 },
		{ 137, 322, 20 },
		{ 137, 323, 20 },
		{ 137, 324, 20 },
		{ 137, 325, 20 },
		{ 137, 326, 20 },
		{ 137, 327, 20 },
		{ 137, 328, 20 },
		{ 137, 329, 20 },
		{ 137, 330, 20 },
		{ 137, 331, 20 },
		{ 137, 332, 20 },
		{ 137, 333, 20 },
		{ 137, 334, 20 },
		{ 137, 335, 20 },
		{ 137, 336, 20 },
		{ 137, 337, 20 },
		{ 137, 338, 20 },
		{ 137, 339, 20 },
		{ 137, 340, 20 },
		{ 137, 341, 20 },
		{ 137, 342, 20 },
		{ 137, 343, 20 },
		{ 137, 344, 20 },
		{ 137, 345, 20 },
		{ 137, 346, 20 },
		{ 138, 288, 20 },
		{ 138, 289, 20 },
		{ 138, 290, 20 },
		{ 138, 291, 20 },
		{ 138, 292, 20 },
		{ 138, 293, 20 },
		{ 138, 294, 20 },
		{ 138, 295, 20 },
		{ 138, 296, 20 },
		{ 138, 297, 20 },
		{ 138, 298, 20 },
		{ 138, 299, 20 },
		{ 138, 300, 20 },
		{ 138, 301, 20 },
		{ 138, 302, 20 },
		{ 138, 303, 20 },
		{ 138, 304, 20 },
		{ 138, 305, 20 },
		{ 138, 306, 20 },
		{ 138, 307, 20 },
		{ 138, 308, 20 },
		{ 138, 309, 20 },
		{ 138, 310, 20 },
		{ 138, 311, 20 },
		{ 138, 312, 20 },
		{ 138, 313, 20 },
		{ 138, 314, 20 },
		{ 138, 315, 20 },
		{ 138, 316, 20 },
		{ 138, 317, 20 },
		{ 138, 318, 20 },
		{ 138, 319, 20 },
		{ 138, 320, 20 },
		{ 138, 321, 20 },
		{ 138, 322, 20 },
		{ 138, 323, 20 },
		{ 138, 324, 20 },
		{ 138, 325, 20 },
		{ 138, 326, 20 },
		{ 138, 327, 20 },
		{ 138, 328, 20 },
		{ 138, 329, 20 },
		{ 138, 330, 20 },
		{ 138, 331, 20 },
		{ 138, 332, 20 },
		{ 138, 333, 20 },
		{ 138, 334, 20 },
		{ 138, 335, 20 },
		{ 138, 336, 20 },
		{ 138, 337, 20 },
		{ 138, 338, 20 },
		{ 138, 339, 20 },
		{ 138, 340, 20 },
		{ 138, 341, 20 },
		{ 138, 342, 20 },
		{ 138, 343, 20 },
		{ 138, 344, 20 },
		{ 138, 345, 20 },
		{ 139, 287, 20 },
		{ 139, 288, 20 },
		{ 139, 289, 20 },
		{ 139, 290, 20 },
		{ 139, 291, 20 },
		{ 139, 292, 20 },
		{ 139, 293, 20 },
		{ 139, 294, 20 },
		{ 139, 295, 20 },
		{ 139, 296, 20 },
		{ 139, 297, 20 },
		{ 139, 298, 20 },
		{ 139, 299, 20 },
		{ 139, 300, 20 },
		{ 139, 301, 20 },
		{ 139, 302, 20 },
		{ 139, 303, 20 },
		{ 139, 304, 20 },
		{ 139, 305, 20 },
		{ 139, 306, 20 },
		{ 139, 307, 20 },
		{ 139, 308, 20 },
		{ 139, 309, 20 },
		{ 139, 310, 20 },
		{ 139, 311, 20 },
		{ 139, 312, 20 },
		{ 139, 313, 20 },
		{ 139, 314, 20 },
		{ 139, 315, 20 },
		{ 139, 316, 20 },
		{ 139, 317, 20 },
		{ 139, 318, 20 },
		{ 139, 319, 20 },
		{ 139, 320, 20 },
		{ 139, 321, 20 },
		{ 139, 322, 20 },
		{ 139, 323, 20 },
		{ 139, 324, 20 },
		{ 139, 325, 20 },
		{ 139, 326, 20 },
		{ 139, 327, 20 },
		{ 139, 328, 20 },
		{ 139, 329, 20 },
		{ 139, 330, 20 },
		{ 139, 331, 20 },
		{ 139, 332, 20 },
		{ 139, 333, 20 },
		{ 139, 334, 20 },
		{ 139, 335, 20 },
		{ 139, 336, 20 },
		{ 139, 337, 20 },
		{ 139, 338, 20 },
		{ 139, 339, 20 },
		{ 139, 340, 20 },
		{ 139, 341, 20 },
		{ 139, 342, 20 },
		{ 139, 343, 20 },
		{ 139, 344, 20 },
		{ 139, 345, 20 },
		{ 140, 287, 20 },
		{ 140, 288, 20 },
		{ 140, 289, 20 },
		{ 140, 290, 20 },
		{ 140, 291, 20 },
		{ 140, 292, 20 },
		{ 140, 293, 20 },
		{ 140, 294, 20 },
		{ 140, 295, 20 },
		{ 140, 296, 20 },
		{ 140, 297, 20 },
		{ 140, 298, 20 },
		{ 140, 299, 20 },
		{ 140, 300, 20 },
		{ 140, 301, 20 },
		{ 140, 302, 20 },
		{ 140, 303, 20 },
		{ 140, 304, 20 },
		{ 140, 305, 20 },
		{ 140, 306, 20 },
		{ 140, 307, 20 },
		{ 140, 308, 20 },
		{ 140, 309, 20 },
		{ 140, 310, 20 },
		{ 140, 311, 20 },
		{ 140, 312, 20 },
		{ 140, 313, 20 },
		{ 140, 314, 20 },
		{ 140, 315, 20 },
		{ 140, 316, 20 },
		{ 140, 317, 20 },
		{ 140, 318, 20 },
		{ 140, 319, 20 },
		{ 140, 320, 20 },
		{ 140, 321, 20 },
		{ 140, 322, 20 },
		{ 140, 323, 20 },
		{ 140, 324, 20 },
		{ 140, 325, 20 },
		{ 140, 326, 20 },
		{ 140, 327, 20 },
		{ 140, 328, 20 },
		{ 140, 329, 20 },
		{ 140, 330, 20 },
		{ 140, 331, 20 },
		{ 140, 332, 20 },
		{ 140, 333, 20 },
		{ 140, 334, 20 },
		{ 140, 335, 20 },
		{ 140, 336, 20 },
		{ 140, 337, 20 },
		{ 140, 338, 20 },
		{ 140, 339, 20 },
		{ 140, 340, 20 },
		{ 140, 341, 20 },
		{ 140, 342, 20 },
		{ 140, 343, 20 },
		{ 140, 344, 20 },
		{ 141, 287, 20 },
		{ 141, 288, 20 },
		{ 141, 289, 20 },
		{ 141, 290, 20 },
		{ 141, 291, 20 },
		{ 141, 292, 20 },
		{ 141, 293, 20 },
		{ 141, 294, 20 },
		{ 141, 295, 20 },
		{ 141, 296, 20 },
		{ 141, 297, 20 },
		{ 141, 298, 20 },
		{ 141, 299, 20 },
		{ 141, 300, 20 },
		{ 141, 301, 20 },
		{ 141, 302, 20 },
		{ 141, 303, 20 },
		{ 141, 304, 20 },
		{ 141, 305, 20 },
		{ 141, 306, 20 },
		{ 141, 307, 20 },
		{ 141, 308, 20 },
		{ 141, 309, 20 },
		{ 141, 310, 20 },
		{ 141, 311, 20 },
		{ 141, 312, 20 },
		{ 141, 313, 20 },
		{ 141, 314, 20 },
		{ 141, 315, 20 },
		{ 141, 316, 20 },
		{ 141, 317, 20 },
		{ 141, 318, 20 },
		{ 141, 319, 20 },
		{ 141, 320, 20 },
		{ 141, 321, 20 },
		{ 141, 322, 20 },
		{ 141, 323, 20 },
		{ 141, 324, 20 },
		{ 141, 325, 20 },
		{ 141, 326, 20 },
		{ 141, 327, 20 },
		{ 141, 328, 20 },
		{ 141, 329, 20 },
		{ 141, 330, 20 },
		{ 141, 331, 20 },
		{ 141, 332, 20 },
		{ 141, 333, 20 },
		{ 141, 334, 20 },
		{ 141, 335, 20 },
		{ 141, 336, 20 },
		{ 141, 337, 20 },
		{ 141, 338, 20 },
		{ 141, 339, 20 },
		{ 141, 340, 20 },
		{ 141, 341, 20 },
		{ 141, 342, 20 },
		{ 141, 343, 20 },
		{ 142, 286, 20 },
		{ 142, 287, 20 },
		{ 142, 288, 20 },
		{ 142, 289, 20 },
		{ 142, 290, 20 },
		{ 142, 291, 20 },
		{ 142, 292, 20 },
		{ 142, 293, 20 },
		{ 142, 294, 20 },
		{ 142, 295, 20 },
		{ 142, 296, 20 },
		{ 142, 297, 20 },
		{ 142, 298, 20 },
		{ 142, 299, 20 },
		{ 142, 300, 20 },
		{ 142, 301, 20 },
		{ 142, 302, 20 },
		{ 142, 303, 20 },
		{ 142, 304, 20 },
		{ 142, 305, 20 },
		{ 142, 306, 20 },
		{ 142, 307, 20 },
		{ 142, 308, 20 },
		{ 142, 309, 20 },
		{ 142, 310, 20 },
		{ 142, 311, 20 },
		{ 142, 312, 20 },
		{ 142, 313, 20 },
		{ 142, 314, 20 },
		{ 142, 315, 20 },
		{ 142, 316, 20 },
		{ 142, 317, 20 },
		{ 142, 318, 20 },
		{ 142, 319, 20 },
		{ 142, 320, 20 },
		{ 142, 321, 20 },
		{ 142, 322, 20 },
		{ 142, 323, 20 },
		{ 142, 324, 20 },
		{ 142, 325, 20 },
		{ 142, 326, 20 },
		{ 142, 327, 20 },
		{ 142, 328, 20 },
		{ 142, 329, 20 },
		{ 142, 330, 20 },
		{ 142, 331, 20 },
		{ 142, 332, 20 },
		{ 142, 333, 20 },
		{ 142, 334, 20 },
		{ 142, 335, 20 },
		{ 142, 336, 20 },
		{ 142, 337, 20 },
		{ 142, 338, 20 },
		{ 142, 339, 20 },
		{ 142, 340, 20 },
		{ 142, 341, 20 },
		{ 142, 342, 20 },
		{ 142, 343, 20 },
		{ 143, 286, 20 },
		{ 143, 287, 20 },
		{ 143, 288, 20 },
		{ 143, 289, 20 },
		{ 143, 290, 20 },
		{ 143, 291, 20 },
		{ 143, 292, 20 },
		{ 143, 293, 20 },
		{ 143, 294, 20 },
		{ 143, 295, 20 },
		{ 143, 296, 20 },
		{ 143, 297, 20 },
		{ 143, 298, 20 },
		{ 143, 299, 20 },
		{ 143, 300, 20 },
		{ 143, 301, 20 },
		{ 143, 302, 20 },
		{ 143, 303, 20 },
		{ 143, 304, 20 },
		{ 143, 305, 20 },
		{ 143, 306, 20 },
		{ 143, 307, 20 },
		{ 143, 308, 20 },
		{ 143, 309, 20 },
		{ 143, 310, 20 },
		{ 143, 311, 20 },
		{ 143, 312, 20 },
		{ 143, 313, 20 },
		{ 143, 314, 20 },
		{ 143, 315, 20 },
		{ 143, 316, 20 },
		{ 143, 317, 20 },
		{ 143, 318, 20 },
		{ 143, 319, 20 },
		{ 143, 320, 20 },
		{ 143, 321, 20 },
		{ 143, 322, 20 },
		{ 143, 323, 20 },
		{ 143, 324, 20 },
		{ 143, 325, 20 },
		{ 143, 326, 20 },
		{ 143, 327, 20 },
		{ 143, 328, 20 },
		{ 143, 329, 20 },
		{ 143, 330, 20 },
		{ 143, 331, 20 },
		{ 143, 332, 20 },
		{ 143, 333, 20 },
		{ 143, 334, 20 },
		{ 143, 335, 20 },
		{ 143, 336, 20 },
		{ 143, 337, 20 },
		{ 143, 338, 20 },
		{ 143, 339, 20 },
		{ 143, 340, 20 },
		{ 143, 341, 20 },
		{ 143, 342, 20 },
		{ 144, 286, 20 },
		{ 144, 287, 20 },
		{ 144, 288, 20 },
		{ 144, 289, 20 },
		{ 144, 290, 20 },
		{ 144, 291, 20 },
		{ 144, 292, 20 },
		{ 144, 293, 20 },
		{ 144, 294, 20 },
		{ 144, 295, 20 },
		{ 144, 296, 20 },
		{ 144, 297, 20 },
		{ 144, 298, 20 },
		{ 144, 299, 20 },
		{ 144, 300, 20 },
		{ 144, 301, 20 },
		{ 144, 302, 20 },
		{ 144, 303, 20 },
		{ 144, 304, 20 },
		{ 144, 305, 20 },
		{ 144, 306, 20 },
		{ 144, 307, 20 },
		{ 144, 308, 20 },
		{ 144, 309, 20 },
		{ 144, 310, 20 },
		{ 144, 311, 20 },
		{ 144, 312, 20 },
		{ 144, 313, 20 },
		{ 144, 314, 20 },
		{ 144, 315, 20 },
		{ 144, 316, 20 },
		{ 144, 317, 20 },
		{ 144, 318, 20 },
		{ 144, 319, 20 },
		{ 144, 320, 20 },
		{ 144, 321, 20 },
		{ 144, 322, 20 },
		{ 144, 323, 20 },
		{ 144, 324, 20 },
		{ 144, 325, 20 },
		{ 144, 326, 20 },
		{ 144, 327, 20 },
		{ 144, 328, 20 },
		{ 144, 329, 20 },
		{ 144, 330, 20 },
		{ 144, 331, 20 },
		{ 144, 332, 20 },
		{ 144, 333, 20 },
		{ 144, 334, 20 },
		{ 144, 335, 20 },
		{ 144, 336, 20 },
		{ 144, 337, 20 },
		{ 144, 338, 20 },
		{ 144, 339, 20 },
		{ 144, 340, 20 },
		{ 144, 341, 20 },
		{ 145, 286, 20 },
		{ 145, 287, 20 },
		{ 145, 288, 20 },
		{ 145, 289, 20 },
		{ 145, 290, 20 },
		{ 145, 291, 20 },
		{ 145, 292, 20 },
		{ 145, 293, 20 },
		{ 145, 294, 20 },
		{ 145, 295, 20 },
		{ 145, 296, 20 },
		{ 145, 297, 20 },
		{ 145, 298, 20 },
		{ 145, 299, 20 },
		{ 145, 300, 20 },
		{ 145, 301, 20 },
		{ 145, 302, 20 },
		{ 145, 303, 20 },
		{ 145, 304, 20 },
		{ 145, 305, 20 },
		{ 145, 306, 20 },
		{ 145, 307, 20 },
		{ 145, 308, 20 },
		{ 145, 309, 20 },
		{ 145, 310, 20 },
		{ 145, 311, 20 },
		{ 145, 312, 20 },
		{ 145, 313, 20 },
		{ 145, 314, 20 },
		{ 145, 315, 20 },
		{ 145, 316, 20 },
		{ 145, 317, 20 },
		{ 145, 318, 20 },
		{ 145, 319, 20 },
		{ 145, 320, 20 },
		{ 145, 321, 20 },
		{ 145, 322, 20 },
		{ 145, 323, 20 },
		{ 145, 324, 20 },
		{ 145, 325, 20 },
		{ 145, 326, 20 },
		{ 145, 327, 20 },
		{ 145, 328, 20 },
		{ 145, 329, 20 },
		{ 145, 330, 20 },
		{ 145, 331, 20 },
		{ 145, 332, 20 },
		{ 145, 333, 20 },
		{ 145, 334, 20 },
		{ 145, 335, 20 },
		{ 145, 336, 20 },
		{ 145, 337, 20 },
		{ 145, 338, 20 },
		{ 145, 339, 20 },
		{ 145, 340, 20 },
		{ 146, 285, 20 },
		{ 146, 286, 20 },
		{ 146, 287, 20 },
		{ 146, 288, 20 },
		{ 146, 289, 20 },
		{ 146, 290, 20 },
		{ 146, 291, 20 },
		{ 146, 292, 20 },
		{ 146, 293, 20 },
		{ 146, 294, 20 },
		{ 146, 295, 20 },
		{ 146, 296, 20 },
		{ 146, 297, 20 },
		{ 146, 298, 20 },
		{ 146, 299, 20 },
		{ 146, 300, 20 },
		{ 146, 301, 20 },
		{ 146, 302, 20 },
		{ 146, 303, 20 },
		{ 146, 304, 20 },
		{ 146, 305, 20 },
		{ 146, 306, 20 },
		{ 146, 307, 20 },
		{ 146, 308, 20 },
		{ 146, 309, 20 },
		{ 146, 310, 20 },
		{ 146, 311, 20 },
		{ 146, 312, 20 },
		{ 146, 313, 20 },
		{ 146, 314, 20 },
		{ 146, 315, 20 },
		{ 146, 316, 20 },
		{ 146, 317, 20 },
		{ 146, 318, 20 },
		{ 146, 319, 20 },
		{ 146, 320, 20 },
		{ 146, 321, 20 },
		{ 146, 322, 20 },
		{ 146, 323, 20 },
		{ 146, 324, 20 },
		{ 146, 325, 20 },
		{ 146, 326, 20 },
		{ 146, 327, 20 },
		{ 146, 328, 20 },
		{ 146, 329, 20 },
		{ 146, 330, 20 },
		{ 146, 331, 20 },
		{ 146, 332, 20 },
		{ 146, 333, 20 },
		{ 146, 334, 20 },
		{ 146, 335, 20 },
		{ 146, 336, 20 },
		{ 146, 337, 20 },
		{ 146, 338, 20 },
		{ 146, 339, 20 },
		{ 146, 340, 20 },
		{ 147, 285, 20 },
		{ 147, 286, 20 },
		{ 147, 287, 20 },
		{ 147, 288, 20 },
		{ 147, 289, 20 },
		{ 147, 290, 20 },
		{ 147, 291, 20 },
		{ 147, 292, 20 },
		{ 147, 293, 20 },
		{ 147, 294, 20 },
		{ 147, 295, 20 },
		{ 147, 296, 20 },
		{ 147, 297, 20 },
		{ 147, 298, 20 },
		{ 147, 299, 20 },
		{ 147, 300, 20 },
		{ 147, 301, 20 },
		{ 147, 302, 20 },
		{ 147, 303, 20 },
		{ 147, 304, 20 },
		{ 147, 305, 20 },
		{ 147, 306, 20 },
		{ 147, 307, 20 },
		{ 147, 308, 20 },
		{ 147, 309, 20 },
		{ 147, 310, 20 },
		{ 147, 311, 20 },
		{ 147, 312, 20 },
		{ 147, 313, 20 },
		{ 147, 314, 20 },
		{ 147, 315, 20 },
		{ 147, 316, 20 },
		{ 147, 317, 20 },
		{ 147, 318, 20 },
		{ 147, 319, 20 },
		{ 147, 320, 20 },
		{ 147, 321, 20 },
		{ 147, 322, 20 },
		{ 147, 323, 20 },
		{ 147, 324, 20 },
		{ 147, 325, 20 },
		{ 147, 326, 20 },
		{ 147, 327, 20 },
		{ 147, 328, 20 },
		{ 147, 329, 20 },
		{ 147, 330, 20 },
		{ 147, 331, 20 },
		{ 147, 332, 20 },
		{ 147, 333, 20 },
		{ 147, 334, 20 },
		{ 147, 335, 20 },
		{ 147, 336, 20 },
		{ 147, 337, 20 },
		{ 147, 338, 20 },
		{ 147, 339, 20 },
		{ 148, 285, 20 },
		{ 148, 286, 20 },
		{ 148, 287, 20 },
		{ 148, 288, 20 },
		{ 148, 289, 20 },
		{ 148, 290, 20 },
		{ 148, 291, 20 },
		{ 148, 292, 20 },
		{ 148, 293, 20 },
		{ 148, 294, 20 },
		{ 148, 295, 20 },
		{ 148, 296, 20 },
		{ 148, 297, 20 },
		{ 148, 298, 20 },
		{ 148, 299, 20 },
		{ 148, 300, 20 },
		{ 148, 301, 20 },
		{ 148, 302, 20 },
		{ 148, 303, 20 },
		{ 148, 304, 20 },
		{ 148, 305, 20 },
		{ 148, 306, 20 },
		{ 148, 307, 20 },
		{ 148, 308, 20 },
		{ 148, 309, 20 },
		{ 148, 310, 20 },
		{ 148, 311, 20 },
		{ 148, 312, 20 },
		{ 148, 313, 20 },
		{ 148, 314, 20 },
		{ 148, 315, 20 },
		{ 148, 316, 20 },
		{ 148, 317, 20 },
		{ 148, 318, 20 },
		{ 148, 319, 20 },
		{ 148, 320, 20 },
		{ 148, 321, 20 },
		{ 148, 322, 20 },
		{ 148, 323, 20 },
		{ 148, 324, 20 },
		{ 148, 325, 20 },
		{ 148, 326, 20 },
		{ 148, 327, 20 },
		{ 148, 328, 20 },
		{ 148, 329, 20 },
		{ 148, 330, 20 },
		{ 148, 331, 20 },
		{ 148, 332, 20 },
		{ 148, 333, 20 },
		{ 148, 334, 20 },
		{ 148, 335, 20 },
		{ 148, 336, 20 },
		{ 148, 337, 20 },
		{ 148, 338, 20 },
		{ 149, 285, 20 },
		{ 149, 286, 20 },
		{ 149, 287, 20 },
		{ 149, 288, 20 },
		{ 149, 289, 20 },
		{ 149, 290, 20 },
		{ 149, 291, 20 },
		{ 149, 292, 20 },
		{ 149, 293, 20 },
		{ 149, 294, 20 },
		{ 149, 295, 20 },
		{ 149, 296, 20 },
		{ 149, 297, 20 },
		{ 149, 298, 20 },
		{ 149, 299, 20 },
		{ 149, 300, 20 },
		{ 149, 301, 20 },
		{ 149, 302, 20 },
		{ 149, 303, 20 },
		{ 149, 304, 20 },
		{ 149, 305, 20 },
		{ 149, 306, 20 },
		{ 149, 307, 20 },
		{ 149, 308, 20 },
		{ 149, 309, 20 },
		{ 149, 310, 20 },
		{ 149, 311, 20 },
		{ 149, 312, 20 },
		{ 149, 313, 20 },
		{ 149, 314, 20 },
		{ 149, 315, 20 },
		{ 149, 316, 20 },
		{ 149, 317, 20 },
		{ 149, 318, 20 },
		{ 149, 319, 20 },
		{ 149, 320, 20 },
		{ 149, 321, 20 },
		{ 149, 322, 20 },
		{ 149, 323, 20 },
		{ 149, 324, 20 },
		{ 149, 325, 20 },
		{ 149, 326, 20 },
		{ 149, 327, 20 },
		{ 149, 328, 20 },
		{ 149, 329, 20 },
		{ 149, 330, 20 },
		{ 149, 331, 20 },
		{ 149, 332, 20 },
		{ 149, 333, 20 },
		{ 149, 334, 20 },
		{ 149, 335, 20 },
		{ 149, 336, 20 },
		{ 149, 337, 20 },
		{ 150, 284, 20 },
		{ 150, 285, 20 },
		{ 150, 286, 20 },
		{ 150, 287, 20 },
		{ 150, 288, 20 },
		{ 150, 289, 20 },
		{ 150, 290, 20 },
		{ 150, 291, 20 },
		{ 150, 292, 20 },
		{ 150, 293, 20 },
		{ 150, 294, 20 },
		{ 150, 295, 20 },
		{ 150, 296, 20 },
		{ 150, 297, 20 },
		{ 150, 298, 20 },
		{ 150, 299, 20 },
		{ 150, 300, 20 },
		{ 150, 301, 20 },
		{ 150, 302, 20 },
		{ 150, 303, 20 },
		{ 150, 304, 20 },
		{ 150, 305, 20 },
		{ 150, 306, 20 },
		{ 150, 307, 20 },
		{ 150, 308, 20 },
		{ 150, 309, 20 },
		{ 150, 310, 20 },
		{ 150, 311, 20 },
		{ 150, 312, 20 },
		{ 150, 313, 20 },
		{ 150, 314, 20 },
		{ 150, 315, 20 },
		{ 150, 316, 20 },
		{ 150, 317, 20 },
		{ 150, 318, 20 },
		{ 150, 319, 20 },
		{ 150, 320, 20 },
		{ 150, 321, 20 },
		{ 150, 322, 20 },
		{ 150, 323, 20 },
		{ 150, 324, 20 },
		{ 150, 325, 20 },
		{ 150, 326, 20 },
		{ 150, 327, 20 },
		{ 150, 328, 20 },
		{ 150, 329, 20 },
		{ 150, 330, 20 },
		{ 150, 331, 20 },
		{ 150, 332, 20 },
		{ 150, 333, 20 },
		{ 150, 334, 20 },
		{ 150, 335, 20 },
		{ 150, 336, 20 },
		{ 150, 337, 20 },
		{ 151, 284, 20 },
		{ 151, 285, 20 },
		{ 151, 286, 20 },
		{ 151, 287, 20 },
		{ 151, 288, 20 },
		{ 151, 289, 20 },
		{ 151, 290, 20 },
		{ 151, 291, 20 },
		{ 151, 292, 20 },
		{ 151, 293, 20 },
		{ 151, 294, 20 },
		{ 151, 295, 20 },
		{ 151, 296, 20 },
		{ 151, 297, 20 },
		{ 151, 298, 20 },
		{ 151, 299, 20 },
		{ 151, 300, 20 },
		{ 151, 301, 20 },
		{ 151, 302, 20 },
		{ 151, 303, 20 },
		{ 151, 304, 20 },
		{ 151, 305, 20 },
		{ 151, 306, 20 },
		{ 151, 307, 20 },
		{ 151, 308, 20 },
		{ 151, 309, 20 },
		{ 151, 310, 20 },
		{ 151, 311, 20 },
		{ 151, 312, 20 },
		{ 151, 313, 20 },
		{ 151, 314, 20 },
		{ 151, 315, 20 },
		{ 151, 316, 20 },
		{ 151, 317, 20 },
		{ 151, 318, 20 },
		{ 151, 319, 20 },
		{ 151, 320, 20 },
		{ 151, 321, 20 },
		{ 151, 322, 20 },
		{ 151, 323, 20 },
		{ 151, 324, 20 },
		{ 151, 325, 20 },
		{ 151, 326, 20 },
		{ 151, 327, 20 },
		{ 151, 328, 20 },
		{ 151, 329, 20 },
		{ 151, 330, 20 },
		{ 151, 331, 20 },
		{ 151, 332, 20 },
		{ 151, 333, 20 },
		{ 151, 334, 20 },
		{ 151, 335, 20 },
		{ 151, 336, 20 },
		{ 152, 284, 20 },
		{ 152, 285, 20 },
		{ 152, 286, 20 },
		{ 152, 287, 20 },
		{ 152, 288, 20 },
		{ 152, 289, 20 },
		{ 152, 290, 20 },
		{ 152, 291, 20 },
		{ 152, 292, 20 },
		{ 152, 293, 20 },
		{ 152, 294, 20 },
		{ 152, 295, 20 },
		{ 152, 296, 20 },
		{ 152, 297, 20 },
		{ 152, 298, 20 },
		{ 152, 299, 20 },
		{ 152, 300, 20 },
		{ 152, 301, 20 },
		{ 152, 302, 20 },
		{ 152, 303, 20 },
		{ 152, 304, 20 },
		{ 152, 305, 20 },
		{ 152, 306, 20 },
		{ 152, 307, 20 },
		{ 152, 308, 20 },
		{ 152, 309, 20 },
		{ 152, 310, 20 },
		{ 152, 311, 20 },
		{ 152, 312, 20 },
		{ 152, 313, 20 },
		{ 152, 314, 20 },
		{ 152, 315, 20 },
		{ 152, 316, 20 },
		{ 152, 317, 20 },
		{ 152, 318, 20 },
		{ 152, 319, 20 },
		{ 152, 320, 20 },
		{ 152, 321, 20 },
		{ 152, 322, 20 },
		{ 152, 323, 20 },
		{ 152, 324, 20 },
		{ 152, 325, 20 },
		{ 152, 326, 20 },
		{ 152, 327, 20 },
		{ 152, 328, 20 },
		{ 152, 329, 20 },
		{ 152, 330, 20 },
		{ 152, 331, 20 },
		{ 152, 332, 20 },
		{ 152, 333, 20 },
		{ 152, 334, 20 },
		{ 152, 335, 20 },
		{ 153, 284, 20 },
		{ 153, 285, 20 },
		{ 153, 286, 20 },
		{ 153, 287, 20 },
		{ 153, 288, 20 },
		{ 153, 289, 20 },
		{ 153, 290, 20 },
		{ 153, 291, 20 },
		{ 153, 292, 20 },
		{ 153, 293, 20 },
		{ 153, 294, 20 },
		{ 153, 295, 20 },
		{ 153, 296, 20 },
		{ 153, 297, 20 },
		{ 153, 298, 20 },
		{ 153, 299, 20 },
		{ 153, 300, 20 },
		{ 153, 301, 20 },
		{ 153, 302, 20 },
		{ 153, 303, 20 },
		{ 153, 304, 20 },
		{ 153, 305, 20 },
		{ 153, 306, 20 },
		{ 153, 307, 20 },
		{ 153, 308, 20 },
		{ 153, 309, 20 },
		{ 153, 310, 20 },
		{ 153, 311, 20 },
		{ 153, 312, 20 },
		{ 153, 313, 20 },
		{ 153, 314, 20 },
		{ 153, 315, 20 },
		{ 153, 316, 20 },
		{ 153, 317, 20 },
		{ 153, 318, 20 },
		{ 153, 319, 20 },
		{ 153, 320, 20 },
		{ 153, 321, 20 },
		{ 153, 322, 20 },
		{ 153, 323, 20 },
		{ 153, 324, 20 },
		{ 153, 325, 20 },
		{ 153, 326, 20 },
		{ 153, 327, 20 },
		{ 153, 328, 20 },
		{ 153, 329, 20 },
		{ 153, 330, 20 },
		{ 153, 331, 20 },
		{ 153, 332, 20 },
		{ 153, 333, 20 },
		{ 153, 334, 20 },
		{ 154, 283, 20 },
		{ 154, 284, 20 },
		{ 154, 285, 20 },
		{ 154, 286, 20 },
		{ 154, 287, 20 },
		{ 154, 288, 20 },
		{ 154, 289, 20 },
		{ 154, 290, 20 },
		{ 154, 291, 20 },
		{ 154, 292, 20 },
		{ 154, 293, 20 },
		{ 154, 294, 20 },
		{ 154, 295, 20 },
		{ 154, 296, 20 },
		{ 154, 297, 20 },
		{ 154, 298, 20 },
		{ 154, 299, 20 },
		{ 154, 300, 20 },
		{ 154, 301, 20 },
		{ 154, 302, 20 },
		{ 154, 303, 20 },
		{ 154, 304, 20 },
		{ 154, 305, 20 },
		{ 154, 306, 20 },
		{ 154, 307, 20 },
		{ 154, 308, 20 },
		{ 154, 309, 20 },
		{ 154, 310, 20 },
		{ 154, 311, 20 },
		{ 154, 312, 20 },
		{ 154, 313, 20 },
		{ 154, 314, 20 },
		{ 154, 315, 20 },
		{ 154, 316, 20 },
		{ 154, 317, 20 },
		{ 154, 318, 20 },
		{ 154, 319, 20 },
		{ 154, 320, 20 },
		{ 154, 321, 20 },
		{ 154, 322, 20 },
		{ 154, 323, 20 },
		{ 154, 324, 20 },
		{ 154, 325, 20 },
		{ 154, 326, 20 },
		{ 154, 327, 20 },
		{ 154, 328, 20 },
		{ 154, 329, 20 },
		{ 154, 330, 20 },
		{ 154, 331, 20 },
		{ 154, 332, 20 },
		{ 154, 333, 20 },
		{ 154, 334, 20 },
		{ 155, 283, 20 },
		{ 155, 284, 20 },
		{ 155, 285, 20 },
		{ 155, 286, 20 },
		{ 155, 287, 20 },
		{ 155, 288, 20 },
		{ 155, 289, 20 },
		{ 155, 290, 20 },
		{ 155, 291, 20 },
		{ 155, 292, 20 },
		{ 155, 293, 20 },
		{ 155, 294, 20 },
		{ 155, 295, 20 },
		{ 155, 296, 20 },
		{ 155, 297, 20 },
		{ 155, 298, 20 },
		{ 155, 299, 20 },
		{ 155, 300, 20 },
		{ 155, 301, 20 },
		{ 155, 302, 20 },
		{ 155, 303, 20 },
		{ 155, 304, 20 },
		{ 155, 305, 20 },
		{ 155, 306, 20 },
		{ 155, 307, 20 },
		{ 155, 308, 20 },
		{ 155, 309, 20 },
		{ 155, 310, 20 },
		{ 155, 311, 20 },
		{ 155, 312, 20 },
		{ 155, 313, 20 },
		{ 155, 314, 20 },
		{ 155, 315, 20 },
		{ 155, 316, 20 },
		{ 155, 317, 20 },
		{ 155, 318, 20 },
		{ 155, 319, 20 },
		{ 155, 320, 20 },
		{ 155, 321, 20 },
		{ 155, 322, 20 },
		{ 155, 323, 20 },
		{ 155, 324, 20 },
		{ 155, 325, 20 },
		{ 155, 326, 20 },
		{ 155, 327, 20 },
		{ 155, 328, 20 },
		{ 155, 329, 20 },
		{ 155, 330, 20 },
		{ 155, 331, 20 },
		{ 155, 332, 20 },
		{ 155, 333, 20 },
		{ 156, 283, 20 },
		{ 156, 284, 20 },
		{ 156, 285, 20 },
		{ 156, 286, 20 },
		{ 156, 287, 20 },
		{ 156, 288, 20 },
		{ 156, 289, 20 },
		{ 156, 290, 20 },
		{ 156, 291, 20 },
		{ 156, 292, 20 },
		{ 156, 293, 20 },
		{ 156, 294, 20 },
		{ 156, 295, 20 },
		{ 156, 296, 20 },
		{ 156, 297, 20 },
		{ 156, 298, 20 },
		{ 156, 299, 20 },
		{ 156, 300, 20 },
		{ 156, 301, 20 },
		{ 156, 302, 20 },
		{ 156, 303, 20 },
		{ 156, 304, 20 },
		{ 156, 305, 20 },
		{ 156, 306, 20 },
		{ 156, 307, 20 },
		{ 156, 308, 20 },
		{ 156, 309, 20 },
		{ 156, 310, 20 },
		{ 156, 311, 20 },
		{ 156, 312, 20 },
		{ 156, 313, 20 },
		{ 156, 314, 20 },
		{ 156, 315, 20 },
		{ 156, 316, 20 },
		{ 156, 317, 20 },
		{ 156, 318, 20 },
		{ 156, 319, 20 },
		{ 156, 320, 20 },
		{ 156, 321, 20 },
		{ 156, 322, 20 },
		{ 156, 323, 20 },
		{ 156, 324, 20 },
		{ 156, 325, 20 },
		{ 156, 326, 20 },
		{ 156, 327, 20 },
		{ 156, 328, 20 },
		{ 156, 329, 20 },
		{ 156, 330, 20 },
		{ 156, 331, 20 },
		{ 156, 332, 20 },
		{ 157, 283, 20 },
		{ 157, 284, 20 },
		{ 157, 285, 20 },
		{ 157, 286, 20 },
		{ 157, 287, 20 },
		{ 157, 288, 20 },
		{ 157, 289, 20 },
		{ 157, 290, 20 },
		{ 157, 291, 20 },
		{ 157, 292, 20 },
		{ 157, 293, 20 },
		{ 157, 294, 20 },
		{ 157, 295, 20 },
		{ 157, 296, 20 },
		{ 157, 297, 20 },
		{ 157, 298, 20 },
		{ 157, 299, 20 },
		{ 157, 300, 20 },
		{ 157, 301, 20 },
		{ 157, 302, 20 },
		{ 157, 303, 20 },
		{ 157, 304, 20 },
		{ 157, 305, 20 },
		{ 157, 306, 20 },
		{ 157, 307, 20 },
		{ 157, 308, 20 },
		{ 157, 309, 20 },
		{ 157, 310, 20 },
		{ 157, 311, 20 },
		{ 157, 312, 20 },
		{ 157, 313, 20 },
		{ 157, 314, 20 },
		{ 157, 315, 20 },
		{ 157, 316, 20 },
		{ 157, 317, 20 },
		{ 157, 318, 20 },
		{ 157, 319, 20 },
		{ 157, 320, 20 },
		{ 157, 321, 20 },
		{ 157, 322, 20 },
		{ 157, 323, 20 },
		{ 157, 324, 20 },
		{ 157, 325, 20 },
		{ 157, 326, 20 },
		{ 157, 327, 20 },
		{ 157, 328, 20 },
		{ 157, 329, 20 },
		{ 157, 330, 20 },
		{ 157, 331, 20 },
		{ 158, 282, 20 },
		{ 158, 283, 20 },
		{ 158, 284, 20 },
		{ 158, 285, 20 },
		{ 158, 286, 20 },
		{ 158, 287, 20 },
		{ 158, 288, 20 },
		{ 158, 289, 20 },
		{ 158, 290, 20 },
		{ 158, 291, 20 },
		{ 158, 292, 20 },
		{ 158, 293, 20 },
		{ 158, 294, 20 },
		{ 158, 295, 20 },
		{ 158, 296, 20 },
		{ 158, 297, 20 },
		{ 158, 298, 20 },
		{ 158, 299, 20 },
		{ 158, 300, 20 },
		{ 158, 301, 20 },
		{ 158, 302, 20 },
		{ 158, 303, 20 },
		{ 158, 304, 20 },
		{ 158, 305, 20 },
		{ 158, 306, 20 },
		{ 158, 307, 20 },
		{ 158, 308, 20 },
		{ 158, 309, 20 },
		{ 158, 310, 20 },
		{ 158, 311, 20 },
		{ 158, 312, 20 },
		{ 158, 313, 20 },
		{ 158, 314, 20 },
		{ 158, 315, 20 },
		{ 158, 316, 20 },
		{ 158, 317, 20 },
		{ 158, 318, 20 },
		{ 158, 319, 20 },
		{ 158, 320, 20 },
		{ 158, 321, 20 },
		{ 158, 322, 20 },
		{ 158, 323, 20 },
		{ 158, 324, 20 },
		{ 158, 325, 20 },
		{ 158, 326, 20 },
		{ 158, 327, 20 },
		{ 158, 328, 20 },
		{ 158, 329, 20 },
		{ 158, 330, 20 },
		{ 158, 331, 20 },
		{ 159, 282, 20 },
		{ 159, 283, 20 },
		{ 159, 284, 20 },
		{ 159, 285, 20 },
		{ 159, 286, 20 },
		{ 159, 287, 20 },
		{ 159, 288, 20 },
		{ 159, 289, 20 },
		{ 159, 290, 20 },
		{ 159, 291, 20 },
		{ 159, 292, 20 },
		{ 159, 293, 20 },
		{ 159, 294, 20 },
		{ 159, 295, 20 },
		{ 159, 296, 20 },
		{ 159, 297, 20 },
		{ 159, 298, 20 },
		{ 159, 299, 20 },
		{ 159, 300, 20 },
		{ 159, 301, 20 },
		{ 159, 302, 20 },
		{ 159, 303, 20 },
		{ 159, 304, 20 },
		{ 159, 305, 20 },
		{ 159, 306, 20 },
		{ 159, 307, 20 },
		{ 159, 308, 20 },
		{ 159, 309, 20 },
		{ 159, 310, 20 },
		{ 159, 311, 20 },
		{ 159, 312, 20 },
		{ 159, 313, 20 },
		{ 159, 314, 20 },
		{ 159, 315, 20 },
		{ 159, 316, 20 },
		{ 159, 317, 20 },
		{ 159, 318, 20 },
		{ 159, 319, 20 },
		{ 159, 320, 20 },
		{ 159, 321, 20 },
		{ 159, 322, 20 },
		{ 159, 323, 20 },
		{ 159, 324, 20 },
		{ 159, 325, 20 },
		{ 159, 326, 20 },
		{ 159, 327, 20 },
		{ 159, 328, 20 },
		{ 159, 329, 20 },
		{ 159, 330, 20 },
		{ 160, 282, 20 },
		{ 160, 283, 20 },
		{ 160, 284, 20 },
		{ 160, 285, 20 },
		{ 160, 286, 20 },
		{ 160, 287, 20 },
		{ 160, 288, 20 },
		{ 160, 289, 20 },
		{ 160, 290, 20 },
		{ 160, 291, 20 },
		{ 160, 292, 20 },
		{ 160, 293, 20 },
		{ 160, 294, 20 },
		{ 160, 295, 20 },
		{ 160, 296, 20 },
		{ 160, 297, 20 },
		{ 160, 298, 20 },
		{ 160, 299, 20 },
		{ 160, 300, 20 },
		{ 160, 301, 20 },
		{ 160, 302, 20 },
		{ 160, 303, 20 },
		{ 160, 304, 20 },
		{ 160, 305, 20 },
		{ 160, 306, 20 },
		{ 160, 307, 20 },
		{ 160, 308, 20 },
		{ 160, 309, 20 },
		{ 160, 310, 20 },
		{ 160, 311, 20 },
		{ 160, 312, 20 },
		{ 160, 313, 20 },
		{ 160, 314, 20 },
		{ 160, 315, 20 },
		{ 160, 316, 20 },
		{ 160, 317, 20 },
		{ 160, 318, 20 },
		{ 160, 319, 20 },
		{ 160, 320, 20 },
		{ 160, 321, 20 },
		{ 160, 322, 20 },
		{ 160, 323, 20 },
		{ 160, 324, 20 },
		{ 160, 325, 20 },
		{ 160, 326, 20 },
		{ 160, 327, 20 },
		{ 160, 328, 20 },
		{ 160, 329, 20 },
		{ 161, 282, 20 },
		{ 161, 283, 20 },
		{ 161, 284, 20 },
		{ 161, 285, 20 },
		{ 161, 286, 20 },
		{ 161, 287, 20 },
		{ 161, 288, 20 },
		{ 161, 289, 20 },
		{ 161, 290, 20 },
		{ 161, 291, 20 },
		{ 161, 292, 20 },
		{ 161, 293, 20 },
		{ 161, 294, 20 },
		{ 161, 295, 20 },
		{ 161, 296, 20 },
		{ 161, 297, 20 },
		{ 161, 298, 20 },
		{ 161, 299, 20 },
		{ 161, 300, 20 },
		{ 161, 301, 20 },
		{ 161, 302, 20 },
		{ 161, 303, 20 },
		{ 161, 304, 20 },
		{ 161, 305, 20 },
		{ 161, 306, 20 },
		{ 161, 307, 20 },
		{ 161, 308, 20 },
		{ 161, 309, 20 },
		{ 161, 310, 20 },
		{ 161, 311, 20 },
		{ 161, 312, 20 },
		{ 161, 313, 20 },
		{ 161, 314, 20 },
		{ 161, 315, 20 },
		{ 161, 316, 20 },
		{ 161, 317, 20 },
		{ 161, 318, 20 },
		{ 161, 319, 20 },
		{ 161, 320, 20 },
		{ 161, 321, 20 },
		{ 161, 322, 20 },
		{ 161, 323, 20 },
		{ 161, 324, 20 },
		{ 161, 325, 20 },
		{ 161, 326, 20 },
		{ 161, 327, 20 },
		{ 161, 328, 20 },
		{ 162, 281, 20 },
		{ 162, 282, 20 },
		{ 162, 283, 20 },
		{ 162, 284, 20 },
		{ 162, 285, 20 },
		{ 162, 286, 20 },
		{ 162, 287, 20 },
		{ 162, 288, 20 },
		{ 162, 289, 20 },
		{ 162, 290, 20 },
		{ 162, 291, 20 },
		{ 162, 292, 20 },
		{ 162, 293, 20 },
		{ 162, 294, 20 },
		{ 162, 295, 20 },
		{ 162, 296, 20 },
		{ 162, 297, 20 },
		{ 162, 298, 20 },
		{ 162, 299, 20 },
		{ 162, 300, 20 },
		{ 162, 301, 20 },
		{ 162, 302, 20 },
		{ 162, 303, 20 },
		{ 162, 304, 20 },
		{ 162, 305, 20 },
		{ 162, 306, 20 },
		{ 162, 307, 20 },
		{ 162, 308, 20 },
		{ 162, 309, 20 },
		{ 162, 310, 20 },
		{ 162, 311, 20 },
		{ 162, 312, 20 },
		{ 162, 313, 20 },
		{ 162, 314, 20 },
		{ 162, 315, 20 },
		{ 162, 316, 20 },
		{ 162, 317, 20 },
		{ 162, 318, 20 },
		{ 162, 319, 20 },
		{ 162, 320, 20 },
		{ 162, 321, 20 },
		{ 162, 322, 20 },
		{ 162, 323, 20 },
		{ 162, 324, 20 },
		{ 162, 325, 20 },
		{ 162, 326, 20 },
		{ 162, 327, 20 },
		{ 162, 328, 20 },
		{ 163, 281, 20 },
		{ 163, 282, 20 },
		{ 163, 283, 20 },
		{ 163, 284, 20 },
		{ 163, 285, 20 },
		{ 163, 286, 20 },
		{ 163, 287, 20 },
		{ 163, 288, 20 },
		{ 163, 289, 20 },
		{ 163, 290, 20 },
		{ 163, 291, 20 },
		{ 163, 292, 20 },
		{ 163, 293, 20 },
		{ 163, 294, 20 },
		{ 163, 295, 20 },
		{ 163, 296, 20 },
		{ 163, 297, 20 },
		{ 163, 298, 20 },
		{ 163, 299, 20 },
		{ 163, 300, 20 },
		{ 163, 301, 20 },
		{ 163, 302, 20 },
		{ 163, 303, 20 },
		{ 163, 304, 20 },
		{ 163, 305, 20 },
		{ 163, 306, 20 },
		{ 163, 307, 20 },
		{ 163, 308, 20 },
		{ 163, 309, 20 },
		{ 163, 310, 20 },
		{ 163, 311, 20 },
		{ 163, 312, 20 },
		{ 163, 313, 20 },
		{ 163, 314, 20 },
		{ 163, 315, 20 },
		{ 163, 316, 20 },
		{ 163, 317, 20 },
		{ 163, 318, 20 },
		{ 163, 319, 20 },
		{ 163, 320, 20 },
		{ 163, 321, 20 },
		{ 163, 322, 20 },
		{ 163, 323, 20 },
		{ 163, 324, 20 },
		{ 163, 325, 20 },
		{ 163, 326, 20 },
		{ 163, 327, 20 },
		{ 164, 281, 20 },
		{ 164, 282, 20 },
		{ 164, 283, 20 },
		{ 164, 284, 20 },
		{ 164, 285, 20 },
		{ 164, 286, 20 },
		{ 164, 287, 20 },
		{ 164, 288, 20 },
		{ 164, 289, 20 },
		{ 164, 290, 20 },
		{ 164, 291, 20 },
		{ 164, 292, 20 },
		{ 164, 293, 20 },
		{ 164, 294, 20 },
		{ 164, 295, 20 },
		{ 164, 296, 20 },
		{ 164, 297, 20 },
		{ 164, 298, 20 },
		{ 164, 299, 20 },
		{ 164, 300, 20 },
		{ 164, 301, 20 },
		{ 164, 302, 20 },
		{ 164, 303, 20 },
		{ 164, 304, 20 },
		{ 164, 305, 20 },
		{ 164, 306, 20 },
		{ 164, 307, 20 },
		{ 164, 308, 20 },
		{ 164, 309, 20 },
		{ 164, 310, 20 },
		{ 164, 311, 20 },
		{ 164, 312, 20 },
		{ 164, 313, 20 },
		{ 164, 314, 20 },
		{ 164, 315, 20 },
		{ 164, 316, 20 },
		{ 164, 317, 20 },
		{ 164, 318, 20 },
		{ 164, 319, 20 },
		{ 164, 320, 20 },
		{ 164, 321, 20 },
		{ 164, 322, 20 },
		{ 164, 323, 20 },
		{ 164, 324, 20 },
		{ 164, 325, 20 },
		{ 164, 326, 20 },
		{ 165, 281, 20 },
		{ 165, 282, 20 },
		{ 165, 283, 20 },
		{ 165, 284, 20 },
		{ 165, 285, 20 },
		{ 165, 286, 20 },
		{ 165, 287, 20 },
		{ 165, 288, 20 },
		{ 165, 289, 20 },
		{ 165, 290, 20 },
		{ 165, 291, 20 },
		{ 165, 292, 20 },
		{ 165, 293, 20 },
		{ 165, 294, 20 },
		{ 165, 295, 20 },
		{ 165, 296, 20 },
		{ 165, 297, 20 },
		{ 165, 298, 20 },
		{ 165, 299, 20 },
		{ 165, 300, 20 },
		{ 165, 301, 20 },
		{ 165, 302, 20 },
		{ 165, 303, 20 },
		{ 165, 304, 20 },
		{ 165, 305, 20 },
		{ 165, 306, 20 },
		{ 165, 307, 20 },
		{ 165, 308, 20 },
		{ 165, 309, 20 },
		{ 165, 310, 20 },
		{ 165, 311, 20 },
		{ 165, 312, 20 },
		{ 165, 313, 20 },
		{ 165, 314, 20 },
		{ 165, 315, 20 },
		{ 165, 316, 20 },
		{ 165, 317, 20 },
		{ 165, 318, 20 },
		{ 165, 319, 20 },
		{ 165, 320, 20 },
		{ 165, 321, 20 },
		{ 165, 322, 20 },
		{ 165, 323, 20 },
		{ 165, 324, 20 },
		{ 165, 325, 20 },
		{ 166, 280, 20 },
		{ 166, 281, 20 },
		{ 166, 282, 20 },
		{ 166, 283, 20 },
		{ 166, 284, 20 },
		{ 166, 285, 20 },
		{ 166, 286, 20 },
		{ 166, 287, 20 },
		{ 166, 288, 20 },
		{ 166, 289, 20 },
		{ 166, 290, 20 },
		{ 166, 291, 20 },
		{ 166, 292, 20 },
		{ 166, 293, 20 },
		{ 166, 294, 20 },
		{ 166, 295, 20 },
		{ 166, 296, 20 },
		{ 166, 297, 20 },
		{ 166, 298, 20 },
		{ 166, 299, 20 },
		{ 166, 300, 20 },
		{ 166, 301, 20 },
		{ 166, 302, 20 },
		{ 166, 303, 20 },
		{ 166, 304, 20 },
		{ 166, 305, 20 },
		{ 166, 306, 20 },
		{ 166, 307, 20 },
		{ 166, 308, 20 },
		{ 166, 309, 20 },
		{ 166, 310, 20 },
		{ 166, 311, 20 },
		{ 166, 312, 20 },
		{ 166, 313, 20 },
		{ 166, 314, 20 },
		{ 166, 315, 20 },
		{ 166, 316, 20 },
		{ 166, 317, 20 },
		{ 166, 318, 20 },
		{ 166, 319, 20 },
		{ 166, 320, 20 },
		{ 166, 321, 20 },
		{ 166, 322, 20 },
		{ 166, 323, 20 },
		{ 166, 324, 20 },
		{ 166, 325, 20 },
		{ 167, 280, 20 },
		{ 167, 281, 20 },
		{ 167, 282, 20 },
		{ 167, 283, 20 },
		{ 167, 284, 20 },
		{ 167, 285, 20 },
		{ 167, 286, 20 },
		{ 167, 287, 20 },
		{ 167, 288, 20 },
		{ 167, 289, 20 },
		{ 167, 290, 20 },
		{ 167, 291, 20 },
		{ 167, 292, 20 },
		{ 167, 293, 20 },
		{ 167, 294, 20 },
		{ 167, 295, 20 },
		{ 167, 296, 20 },
		{ 167, 297, 20 },
		{ 167, 298, 20 },
		{ 167, 299, 20 },
		{ 167, 300, 20 },
		{ 167, 301, 20 },
		{ 167, 302, 20 },
		{ 167, 303, 20 },
		{ 167, 304, 20 },
		{ 167, 305, 20 },
		{ 167, 306, 20 },
		{ 167, 307, 20 },
		{ 167, 308, 20 },
		{ 167, 309, 20 },
		{ 167, 310, 20 },
		{ 167, 311, 20 },
		{ 167, 312, 20 },
		{ 167, 313, 20 },
		{ 167, 314, 20 },
		{ 167, 315, 20 },
		{ 167, 316, 20 },
		{ 167, 317, 20 },
		{ 167, 318, 20 },
		{ 167, 319, 20 },
		{ 167, 320, 20 },
		{ 167, 321, 20 },
		{ 167, 322, 20 },
		{ 167, 323, 20 },
		{ 167, 324, 20 },
		{ 168, 280, 20 },
		{ 168, 281, 20 },
		{ 168, 282, 20 },
		{ 168, 283, 20 },
		{ 168, 284, 20 },
		{ 168, 285, 20 },
		{ 168, 286, 20 },
		{ 168, 287, 20 },
		{ 168, 288, 20 },
		{ 168, 289, 20 },
		{ 168, 290, 20 },
		{ 168, 291, 20 },
		{ 168, 292, 20 },
		{ 168, 293, 20 },
		{ 168, 294, 20 },
		{ 168, 295, 20 },
		{ 168, 296, 20 },
		{ 168, 297, 20 },
		{ 168, 298, 20 },
		{ 168, 299, 20 },
		{ 168, 300, 20 },
		{ 168, 301, 20 },
		{ 168, 302, 20 },
		{ 168, 303, 20 },
		{ 168, 304, 20 },
		{ 168, 305, 20 },
		{ 168, 306, 20 },
		{ 168, 307, 20 },
		{ 168, 308, 20 },
		{ 168, 309, 20 },
		{ 168, 310, 20 },
		{ 168, 311, 20 },
		{ 168, 312, 20 },
		{ 168, 313, 20 },
		{ 168, 314, 20 },
		{ 168, 315, 20 },
		{ 168, 316, 20 },
		{ 168, 317, 20 },
		{ 168, 318, 20 },
		{ 168, 319, 20 },
		{ 168, 320, 20 },
		{ 168, 321, 20 },
		{ 168, 322, 20 },
		{ 168, 323, 20 },
		{ 169, 280, 20 },
		{ 169, 281, 20 },
		{ 169, 282, 20 },
		{ 169, 283, 20 },
		{ 169, 284, 20 },
		{ 169, 285, 20 },
		{ 169, 286, 20 },
		{ 169, 287, 20 },
		{ 169, 288, 20 },
		{ 169, 289, 20 },
		{ 169, 290, 20 },
		{ 169, 291, 20 },
		{ 169, 292, 20 },
		{ 169, 293, 20 },
		{ 169, 294, 20 },
		{ 169, 295, 20 },
		{ 169, 296, 20 },
		{ 169, 297, 20 },
		{ 169, 298, 20 },
		{ 169, 299, 20 },
		{ 169, 300, 20 },
		{ 169, 301, 20 },
		{ 169, 302, 20 },
		{ 169, 303, 20 },
		{ 169, 304, 20 },
		{ 169, 305, 20 },
		{ 169, 306, 20 },
		{ 169, 307, 20 },
		{ 169, 308, 20 },
		{ 169, 309, 20 },
		{ 169, 310, 20 },
		{ 169, 311, 20 },
		{ 169, 312, 20 },
		{ 169, 313, 20 },
		{ 169, 314, 20 },
		{ 169, 315, 20 },
		{ 169, 316, 20 },
		{ 169, 317, 20 },
		{ 169, 318, 20 },
		{ 169, 319, 20 },
		{ 169, 320, 20 },
		{ 169, 321, 20 },
		{ 169, 322, 20 },
		{ 170, 279, 20 },
		{ 170, 280, 20 },
		{ 170, 281, 20 },
		{ 170, 282, 20 },
		{ 170, 283, 20 },
		{ 170, 284, 20 },
		{ 170, 285, 20 },
		{ 170, 286, 20 },
		{ 170, 287, 20 },
		{ 170, 288, 20 },
		{ 170, 289, 20 },
		{ 170, 290, 20 },
		{ 170, 291, 20 },
		{ 170, 292, 20 },
		{ 170, 293, 20 },
		{ 170, 294, 20 },
		{ 170, 295, 20 },
		{ 170, 296, 20 },
		{ 170, 297, 20 },
		{ 170, 298, 20 },
		{ 170, 299, 20 },
		{ 170, 300, 20 },
		{ 170, 301, 20 },
		{ 170, 302, 20 },
		{ 170, 303, 20 },
		{ 170, 304, 20 },
		{ 170, 305, 20 },
		{ 170, 306, 20 },
		{ 170, 307, 20 },
		{ 170, 308, 20 },
		{ 170, 309, 20 },
		{ 170, 310, 20 },
		{ 170, 311, 20 },
		{ 170, 312, 20 },
		{ 170, 313, 20 },
		{ 170, 314, 20 },
		{ 170, 315, 20 },
		{ 170, 316, 20 },
		{ 170, 317, 20 },
		{ 170, 318, 20 },
		{ 170, 319, 20 },
		{ 170, 320, 20 },
		{ 170, 321, 20 },
		{ 170, 322, 20 },
		{ 171, 279, 20 },
		{ 171, 280, 20 },
		{ 171, 281, 20 },
		{ 171, 282, 20 },
		{ 171, 283, 20 },
		{ 171, 284, 20 },
		{ 171, 285, 20 },
		{ 171, 286, 20 },
		{ 171, 287, 20 },
		{ 171, 288, 20 },
		{ 171, 289, 20 },
		{ 171, 290, 20 },
		{ 171, 291, 20 },
		{ 171, 292, 20 },
		{ 171, 293, 20 },
		{ 171, 294, 20 },
		{ 171, 295, 20 },
		{ 171, 296, 20 },
		{ 171, 297, 20 },
		{ 171, 298, 20 },
		{ 171, 299, 20 },
		{ 171, 300, 20 },
		{ 171, 301, 20 },
		{ 171, 302, 20 },
		{ 171, 303, 20 },
		{ 171, 304, 20 },
		{ 171, 305, 20 },
		{ 171, 306, 20 },
		{ 171, 307, 20 },
		{ 171, 308, 20 },
		{ 171, 309, 20 },
		{ 171, 310, 20 },
		{ 171, 311, 20 },
		{ 171, 312, 20 },
		{ 171, 313, 20 },
		{ 171, 314, 20 },
		{ 171, 315, 20 },
		{ 171, 316, 20 },
		{ 171, 317, 20 },
		{ 171, 318, 20 },
		{ 171, 319, 20 },
		{ 171, 320, 20 },
		{ 171, 321, 20 },
		{ 172, 279, 20 },
		{ 172, 280, 20 },
		{ 172, 281, 20 },
		{ 172, 282, 20 },
		{ 172, 283, 20 },
		{ 172, 284, 20 },
		{ 172, 285, 20 },
		{ 172, 286, 20 },
		{ 172, 287, 20 },
		{ 172, 288, 20 },
		{ 172, 289, 20 },
		{ 172, 290, 20 },
		{ 172, 291, 20 },
		{ 172, 292, 20 },
		{ 172, 293, 20 },
		{ 172, 294, 20 },
		{ 172, 295, 20 },
		{ 172, 296, 20 },
		{ 172, 297, 20 },
		{ 172, 298, 20 },
		{ 172, 299, 20 },
		{ 172, 300, 20 },
		{ 172, 301, 20 },
		{ 172, 302, 20 },
		{ 172, 303, 20 },
		{ 172, 304, 20 },
		{ 172, 305, 20 },
		{ 172, 306, 20 },
		{ 172, 307, 20 },
		{ 172, 308, 20 },
		{ 172, 309, 20 },
		{ 172, 310, 20 },
		{ 172, 311, 20 },
		{ 172, 312, 20 },
		{ 172, 313, 20 },
		{ 172, 314, 20 },
		{ 172, 315, 20 },
		{ 172, 316, 20 },
		{ 172, 317, 20 },
		{ 172, 318, 20 },
		{ 172, 319, 20 },
		{ 172, 320, 20 },
		{ 173, 279, 20 },
		{ 173, 280, 20 },
		{ 173, 281, 20 },
		{ 173, 282, 20 },
		{ 173, 283, 20 },
		{ 173, 284, 20 },
		{ 173, 285, 20 },
		{ 173, 286, 20 },
		{ 173, 287, 20 },
		{ 173, 288, 20 },
		{ 173, 289, 20 },
		{ 173, 290, 20 },
		{ 173, 291, 20 },
		{ 173, 292, 20 },
		{ 173, 293, 20 },
		{ 173, 294, 20 },
		{ 173, 295, 20 },
		{ 173, 296, 20 },
		{ 173, 297, 20 },
		{ 173, 298, 20 },
		{ 173, 299, 20 },
		{ 173, 300, 20 },
		{ 173, 301, 20 },
		{ 173, 302, 20 },
		{ 173, 303, 20 },
		{ 173, 304, 20 },
		{ 173, 305, 20 },
		{ 173, 306, 20 },
		{ 173, 307, 20 },
		{ 173, 308, 20 },
		{ 173, 309, 20 },
		{ 173, 310, 20 },
		{ 173, 311, 20 },
		{ 173, 312, 20 },
		{ 173, 313, 20 },
		{ 173, 314, 20 },
		{ 173, 315, 20 },
		{ 173, 316, 20 },
		{ 173, 317, 20 },
		{ 173, 318, 20 },
		{ 173, 319, 20 },
		{ 174, 278, 20 },
		{ 174, 279, 20 },
		{ 174, 280, 20 },
		{ 174, 281, 20 },
		{ 174, 282, 20 },
		{ 174, 283, 20 },
		{ 174, 284, 20 },
		{ 174, 285, 20 },
		{ 174, 286, 20 },
		{ 174, 287, 20 },
		{ 174, 288, 20 },
		{ 174, 289, 20 },
		{ 174, 290, 20 },
		{ 174, 291, 20 },
		{ 174, 292, 20 },
		{ 174, 293, 20 },
		{ 174, 294, 20 },
		{ 174, 295, 20 },
		{ 174, 296, 20 },
		{ 174, 297, 20 },
		{ 174, 298, 20 },
		{ 174, 299, 20 },
		{ 174, 300, 20 },
		{ 174, 301, 20 },
		{ 174, 302, 20 },
		{ 174, 303, 20 },
		{ 174, 304, 20 },
		{ 174, 305, 20 },
		{ 174, 306, 20 },
		{ 174, 307, 20 },
		{ 174, 308, 20 },
		{ 174, 309, 20 },
		{ 174, 310, 20 },
		{ 174, 311, 20 },
		{ 174, 312, 20 },
		{ 174, 313, 20 },
		{ 174, 314, 20 },
		{ 174, 315, 20 },
		{ 174, 316, 20 },
		{ 174, 317, 20 },
		{ 174, 318, 20 },
		{ 174, 319, 20 },
		{ 175, 278, 20 },
		{ 175, 279, 20 },
		{ 175, 280, 20 },
		{ 175, 281, 20 },
		{ 175, 282, 20 },
		{ 175, 283, 20 },
		{ 175, 284, 20 },
		{ 175, 285, 20 },
		{ 175, 286, 20 },
		{ 175, 287, 20 },
		{ 175, 288, 20 },
		{ 175, 289, 20 },
		{ 175, 290, 20 },
		{ 175, 291, 20 },
		{ 175, 292, 20 },
		{ 175, 293, 20 },
		{ 175, 294, 20 },
		{ 175, 295, 20 },
		{ 175, 296, 20 },
		{ 175, 297, 20 },
		{ 175, 298, 20 },
		{ 175, 299, 20 },
		{ 175, 300, 20 },
		{ 175, 301, 20 },
		{ 175, 302, 20 },
		{ 175, 303, 20 },
		{ 175, 304, 20 },
		{ 175, 305, 20 },
		{ 175, 306, 20 },
		{ 175, 307, 20 },
		{ 175, 308, 20 },
		{ 175, 309, 20 },
		{ 175, 310, 20 },
		{ 175, 311, 20 },
		{ 175, 312, 20 },
		{ 175, 313, 20 },
		{ 175, 314, 20 },
		{ 175, 315, 20 },
		{ 175, 316, 20 },
		{ 175, 317, 20 },
		{ 175, 318, 20 },
		{ 176, 278, 20 },
		{ 176, 279, 20 },
		{ 176, 280, 20 },
		{ 176, 281, 20 },
		{ 176, 282, 20 },
		{ 176, 283, 20 },
		{ 176, 284, 20 },
		{ 176, 285, 20 },
		{ 176, 286, 20 },
		{ 176, 287, 20 },
		{ 176, 288, 20 },
		{ 176, 289, 20 },
		{ 176, 290, 20 },
		{ 176, 291, 20 },
		{ 176, 292, 20 },
		{ 176, 293, 20 },
		{ 176, 294, 20 },
		{ 176, 295, 20 },
		{ 176, 296, 20 },
		{ 176, 297, 20 },
		{ 176, 298, 20 },
		{ 176, 299, 20 },
		{ 176, 300, 20 },
		{ 176, 301, 20 },
		{ 176, 302, 20 },
		{ 176, 303, 20 },
		{ 176, 304, 20 },
		{ 176, 305, 20 },
		{ 176, 306, 20 },
		{ 176, 307, 20 },
		{ 176, 308, 20 },
		{ 176, 309, 20 },
		{ 176, 310, 20 },
		{ 176, 311, 20 },
		{ 176, 312, 20 },
		{ 176, 313, 20 },
		{ 176, 314, 20 },
		{ 176, 315, 20 },
		{ 176, 316, 20 },
		{ 176, 317, 20 },
		{ 177, 278, 20 },
		{ 177, 279, 20 },
		{ 177, 280, 20 },
		{ 177, 281, 20 },
		{ 177, 282, 20 },
		{ 177, 283, 20 },
		{ 177, 284, 20 },
		{ 177, 285, 20 },
		{ 177, 286, 20 },
		{ 177, 287, 20 },
		{ 177, 288, 20 },
		{ 177, 289, 20 },
		{ 177, 290, 20 },
		{ 177, 291, 20 },
		{ 177, 292, 20 },
		{ 177, 293, 20 },
		{ 177, 294, 20 },
		{ 177, 295, 20 },
		{ 177, 296, 20 },
		{ 177, 297, 20 },
		{ 177, 298, 20 },
		{ 177, 299, 20 },
		{ 177, 300, 20 },
		{ 177, 301, 20 },
		{ 177, 302, 20 },
		{ 177, 303, 20 },
		{ 177, 304, 20 },
		{ 177, 305, 20 },
		{ 177, 306, 20 },
		{ 177, 307, 20 },
		{ 177, 308, 20 },
		{ 177, 309, 20 },
		{ 177, 310, 20 },
		{ 177, 311, 20 },
		{ 177, 312, 20 },
		{ 177, 313, 20 },
		{ 177, 314, 20 },
		{ 177, 315, 20 },
		{ 177, 316, 20 },
		{ 178, 277, 20 },
		{ 178, 278, 20 },
		{ 178, 279, 20 },
		{ 178, 280, 20 },
		{ 178, 281, 20 },
		{ 178, 282, 20 },
		{ 178, 283, 20 },
		{ 178, 284, 20 },
		{ 178, 285, 20 },
		{ 178, 286, 20 },
		{ 178, 287, 20 },
		{ 178, 288, 20 },
		{ 178, 289, 20 },
		{ 178, 290, 20 },
		{ 178, 291, 20 },
		{ 178, 292, 20 },
		{ 178, 293, 20 },
		{ 178, 294, 20 },
		{ 178, 295, 20 },
		{ 178, 296, 20 },
		{ 178, 297, 20 },
		{ 178, 298, 20 },
		{ 178, 299, 20 },
		{ 178, 300, 20 },
		{ 178, 301, 20 },
		{ 178, 302, 20 },
		{ 178, 303, 20 },
		{ 178, 304, 20 },
		{ 178, 305, 20 },
		{ 178, 306, 20 },
		{ 178, 307, 20 },
		{ 178, 308, 20 },
		{ 178, 309, 20 },
		{ 178, 310, 20 },
		{ 178, 311, 20 },
		{ 178, 312, 20 },
		{ 178, 313, 20 },
		{ 178, 314, 20 },
		{ 178, 315, 20 },
		{ 178, 316, 20 },
		{ 179, 277, 20 },
		{ 179, 278, 20 },
		{ 179, 279, 20 },
		{ 179, 280, 20 },
		{ 179, 281, 20 },
		{ 179, 282, 20 },
		{ 179, 283, 20 },
		{ 179, 284, 20 },
		{ 179, 285, 20 },
		{ 179, 286, 20 },
		{ 179, 287, 20 },
		{ 179, 288, 20 },
		{ 179, 289, 20 },
		{ 179, 290, 20 },
		{ 179, 291, 20 },
		{ 179, 292, 20 },
		{ 179, 293, 20 },
		{ 179, 294, 20 },
		{ 179, 295, 20 },
		{ 179, 296, 20 },
		{ 179, 297, 20 },
		{ 179, 298, 20 },
		{ 179, 299, 20 },
		{ 179, 300, 20 },
		{ 179, 301, 20 },
		{ 179, 302, 20 },
		{ 179, 303, 20 },
		{ 179, 304, 20 },
		{ 179, 305, 20 },
		{ 179, 306, 20 },
		{ 179, 307, 20 },
		{ 179, 308, 20 },
		{ 179, 309, 20 },
		{ 179, 310, 20 },
		{ 179, 311, 20 },
		{ 179, 312, 20 },
		{ 179, 313, 20 },
		{ 179, 314, 20 },
		{ 179, 315, 20 },
		{ 180, 277, 20 },
		{ 180, 278, 20 },
		{ 180, 279, 20 },
		{ 180, 280, 20 },
		{ 180, 281, 20 },
		{ 180, 282, 20 },
		{ 180, 283, 20 },
		{ 180, 284, 20 },
		{ 180, 285, 20 },
		{ 180, 286, 20 },
		{ 180, 287, 20 },
		{ 180, 288, 20 },
		{ 180, 289, 20 },
		{ 180, 290, 20 },
		{ 180, 291, 20 },
		{ 180, 292, 20 },
		{ 180, 293, 20 },
		{ 180, 294, 20 },
		{ 180, 295, 20 },
		{ 180, 296, 20 },
		{ 180, 297, 20 },
		{ 180, 298, 20 },
		{ 180, 299, 20 },
		{ 180, 300, 20 },
		{ 180, 301, 20 },
		{ 180, 302, 20 },
		{ 180, 303, 20 },
		{ 180, 304, 20 },
		{ 180, 305, 20 },
		{ 180, 306, 20 },
		{ 180, 307, 20 },
		{ 180, 308, 20 },
		{ 180, 309, 20 },
		{ 180, 310, 20 },
		{ 180, 311, 20 },
		{ 180, 312, 20 },
		{ 180, 313, 20 },
		{ 180, 314, 20 },
		{ 181, 277, 20 },
		{ 181, 278, 20 },
		{ 181, 279, 20 },
		{ 181, 280, 20 },
		{ 181, 281, 20 },
		{ 181, 282, 20 },
		{ 181, 283, 20 },
		{ 181, 284, 20 },
		{ 181, 285, 20 },
		{ 181, 286, 20 },
		{ 181, 287, 20 },
		{ 181, 288, 20 },
		{ 181, 289, 20 },
		{ 181, 290, 20 },
		{ 181, 291, 20 },
		{ 181, 292, 20 },
		{ 181, 293, 20 },
		{ 181, 294, 20 },
		{ 181, 295, 20 },
		{ 181, 296, 20 },
		{ 181, 297, 20 },
		{ 181, 298, 20 },
		{ 181, 299, 20 },
		{ 181, 300, 20 },
		{ 181, 301, 20 },
		{ 181, 302, 20 },
		{ 181, 303, 20 },
		{ 181, 304, 20 },
		{ 181, 305, 20 },
		{ 181, 306, 20 },
		{ 181, 307, 20 },
		{ 181, 308, 20 },
		{ 181, 309, 20 },
		{ 181, 310, 20 },
		{ 181, 311, 20 },
		{ 181, 312, 20 },
		{ 181, 313, 20 },
		{ 182, 276, 20 },
		{ 182, 277, 20 },
		{ 182, 278, 20 },
		{ 182, 279, 20 },
		{ 182, 280, 20 },
		{ 182, 281, 20 },
		{ 182, 282, 20 },
		{ 182, 283, 20 },
		{ 182, 284, 20 },
		{ 182, 285, 20 },
		{ 182, 286, 20 },
		{ 182, 287, 20 },
		{ 182, 288, 20 },
		{ 182, 289, 20 },
		{ 182, 290, 20 },
		{ 182, 291, 20 },
		{ 182, 292, 20 },
		{ 182, 293, 20 },
		{ 182, 294, 20 },
		{ 182, 295, 20 },
		{ 182, 296, 20 },
		{ 182, 297, 20 },
		{ 182, 298, 20 },
		{ 182, 299, 20 },
		{ 182, 300, 20 },
		{ 182, 301, 20 },
		{ 182, 302, 20 },
		{ 182, 303, 20 },
		{ 182, 304, 20 },
		{ 182, 305, 20 },
		{ 182, 306, 20 },
		{ 182, 307, 20 },
		{ 182, 308, 20 },
		{ 182, 309, 20 },
		{ 182, 310, 20 },
		{ 182, 311, 20 },
		{ 182, 312, 20 },
		{ 182, 313, 20 },
		{ 183, 276, 20 },
		{ 183, 277, 20 },
		{ 183, 278, 20 },
		{ 183, 279, 20 },
		{ 183, 280, 20 },
		{ 183, 281, 20 },
		{ 183, 282, 20 },
		{ 183, 283, 20 },
		{ 183, 284, 20 },
		{ 183, 285, 20 },
		{ 183, 286, 20 },
		{ 183, 287, 20 },
		{ 183, 288, 20 },
		{ 183, 289, 20 },
		{ 183, 290, 20 },
		{ 183, 291, 20 },
		{ 183, 292, 20 },
		{ 183, 293, 20 },
		{ 183, 294, 20 },
		{ 183, 295, 20 },
		{ 183, 296, 20 },
		{ 183, 297, 20 },
		{ 183, 298, 20 },
		{ 183, 299, 20 },
		{ 183, 300, 20 },
		{ 183, 301, 20 },
		{ 183, 302, 20 },
		{ 183, 303, 20 },
		{ 183, 304, 20 },
		{ 183, 305, 20 },
		{ 183, 306, 20 },
		{ 183, 307, 20 },
		{ 183, 308, 20 },
		{ 183, 309, 20 },
		{ 183, 310, 20 },
		{ 183, 311, 20 },
		{ 183, 312, 20 },
		{ 184, 276, 20 },
		{ 184, 277, 20 },
		{ 184, 278, 20 },
		{ 184, 279, 20 },
		{ 184, 280, 20 },
		{ 184, 281, 20 },
		{ 184, 282, 20 },
		{ 184, 283, 20 },
		{ 184, 284, 20 },
		{ 184, 285, 20 },
		{ 184, 286, 20 },
		{ 184, 287, 20 },
		{ 184, 288, 20 },
		{ 184, 289, 20 },
		{ 184, 290, 20 },
		{ 184, 291, 20 },
		{ 184, 292, 20 },
		{ 184, 293, 20 },
		{ 184, 294, 20 },
		{ 184, 295, 20 },
		{ 184, 296, 20 },
		{ 184, 297, 20 },
		{ 184, 298, 20 },
		{ 184, 299, 20 },
		{ 184, 300, 20 },
		{ 184, 301, 20 },
		{ 184, 302, 20 },
		{ 184, 303, 20 },
		{ 184, 304, 20 },
		{ 184, 305, 20 },
		{ 184, 306, 20 },
		{ 184, 307, 20 },
		{ 184, 308, 20 },
		{ 184, 309, 20 },
		{ 184, 310, 20 },
		{ 184, 311, 20 },
		{ 185, 276, 20 },
		{ 185, 277, 20 },
		{ 185, 278, 20 },
		{ 185, 279, 20 },
		{ 185, 280, 20 },
		{ 185, 281, 20 },
		{ 185, 282, 20 },
		{ 185, 283, 20 },
		{ 185, 284, 20 },
		{ 185, 285, 20 },
		{ 185, 286, 20 },
		{ 185, 287, 20 },
		{ 185, 288, 20 },
		{ 185, 289, 20 },
		{ 185, 290, 20 },
		{ 185, 291, 20 },
		{ 185, 292, 20 },
		{ 185, 293, 20 },
		{ 185, 294, 20 },
		{ 185, 295, 20 },
		{ 185, 296, 20 },
		{ 185, 297, 20 },
		{ 185, 298, 20 },
		{ 185, 299, 20 },
		{ 185, 300, 20 },
		{ 185, 301, 20 },
		{ 185, 302, 20 },
		{ 185, 303, 20 },
		{ 185, 304, 20 },
		{ 185, 305, 20 },
		{ 185, 306, 20 },
		{ 185, 307, 20 },
		{ 185, 308, 20 },
		{ 185, 309, 20 },
		{ 185, 310, 20 },
		{ 186, 275, 20 },
		{ 186, 276, 20 },
		{ 186, 277, 20 },
		{ 186, 278, 20 },
		{ 186, 279, 20 },
		{ 186, 280, 20 },
		{ 186, 281, 20 },
		{ 186, 282, 20 },
		{ 186, 283, 20 },
		{ 186, 284, 20 },
		{ 186, 285, 20 },
		{ 186, 286, 20 },
		{ 186, 287, 20 },
		{ 186, 288, 20 },
		{ 186, 289, 20 },
		{ 186, 290, 20 },
		{ 186, 291, 20 },
		{ 186, 292, 20 },
		{ 186, 293, 20 },
		{ 186, 294, 20 },
		{ 186, 295, 20 },
		{ 186, 296, 20 },
		{ 186, 297, 20 },
		{ 186, 298, 20 },
		{ 186, 299, 20 },
		{ 186, 300, 20 },
		{ 186, 301, 20 },
		{ 186, 302, 20 },
		{ 186, 303, 20 },
		{ 186, 304, 20 },
		{ 186, 305, 20 },
		{ 186, 306, 20 },
		{ 186, 307, 20 },
		{ 186, 308, 20 },
		{ 186, 309, 20 },
		{ 186, 310, 20 },
		{ 187, 275, 20 },
		{ 187, 276, 20 },
		{ 187, 277, 20 },
		{ 187, 278, 20 },
		{ 187, 279, 20 },
		{ 187, 280, 20 },
		{ 187, 281, 20 },
		{ 187, 282, 20 },
		{ 187, 283, 20 },
		{ 187, 284, 20 },
		{ 187, 285, 20 },
		{ 187, 286, 20 },
		{ 187, 287, 20 },
		{ 187, 288, 20 },
		{ 187, 289, 20 },
		{ 187, 290, 20 },
		{ 187, 291, 20 },
		{ 187, 292, 20 },
		{ 187, 293, 20 },
		{ 187, 294, 20 },
		{ 187, 295, 20 },
		{ 187, 296, 20 },
		{ 187, 297, 20 },
		{ 187, 298, 20 },
		{ 187, 299, 20 },
		{ 187, 300, 20 },
		{ 187, 301, 20 },
		{ 187, 302, 20 },
		{ 187, 303, 20 },
		{ 187, 304, 20 },
		{ 187, 305, 20 },
		{ 187, 306, 20 },
		{ 187, 307, 20 },
		{ 187, 308, 20 },
		{ 187, 309, 20 },
		{ 188, 275, 20 },
		{ 188, 276, 20 },
		{ 188, 277, 20 },
		{ 188, 278, 20 },
		{ 188, 279, 20 },
		{ 188, 280, 20 },
		{ 188, 281, 20 },
		{ 188, 282, 20 },
		{ 188, 283, 20 },
		{ 188, 284, 20 },
		{ 188, 285, 20 },
		{ 188, 286, 20 },
		{ 188, 287, 20 },
		{ 188, 288, 20 },
		{ 188, 289, 20 },
		{ 188, 290, 20 },
		{ 188, 291, 20 },
		{ 188, 292, 20 },
		{ 188, 293, 20 },
		{ 188, 294, 20 },
		{ 188, 295, 20 },
		{ 188, 296, 20 },
		{ 188, 297, 20 },
		{ 188, 298, 20 },
		{ 188, 299, 20 },
		{ 188, 300, 20 },
		{ 188, 301, 20 },
		{ 188, 302, 20 },
		{ 188, 303, 20 },
		{ 188, 304, 20 },
		{ 188, 305, 20 },
		{ 188, 306, 20 },
		{ 188, 307, 20 },
		{ 188, 308, 20 },
		{ 189, 275, 20 },
		{ 189, 276, 20 },
		{ 189, 277, 20 },
		{ 189, 278, 20 },
		{ 189, 279, 20 },
		{ 189, 280, 20 },
		{ 189, 281, 20 },
		{ 189, 282, 20 },
		{ 189, 283, 20 },
		{ 189, 284, 20 },
		{ 189, 285, 20 },
		{ 189, 286, 20 },
		{ 189, 287, 20 },
		{ 189, 288, 20 },
		{ 189, 289, 20 },
		{ 189, 290, 20 },
		{ 189, 291, 20 },
		{ 189, 292, 20 },
		{ 189, 293, 20 },
		{ 189, 294, 20 },
		{ 189, 295, 20 },
		{ 189, 296, 20 },
		{ 189, 297, 20 },
		{ 189, 298, 20 },
		{ 189, 299, 20 },
		{ 189, 300, 20 },
		{ 189, 301, 20 },
		{ 189, 302, 20 },
		{ 189, 303, 20 },
		{ 189, 304, 20 },
		{ 189, 305, 20 },
		{ 189, 306, 20 },
		{ 189, 307, 20 },
		{ 190, 274, 20 },
		{ 190, 275, 20 },
		{ 190, 276, 20 },
		{ 190, 277, 20 },
		{ 190, 278, 20 },
		{ 190, 279, 20 },
		{ 190, 280, 20 },
		{ 190, 281, 20 },
		{ 190, 282, 20 },
		{ 190, 283, 20 },
		{ 190, 284, 20 },
		{ 190, 285, 20 },
		{ 190, 286, 20 },
		{ 190, 287, 20 },
		{ 190, 288, 20 },
		{ 190, 289, 20 },
		{ 190, 290, 20 },
		{ 190, 291, 20 },
		{ 190, 292, 20 },
		{ 190, 293, 20 },
		{ 190, 294, 20 },
		{ 190, 295, 20 },
		{ 190, 296, 20 },
		{ 190, 297, 20 },
		{ 190, 298, 20 },
		{ 190, 299, 20 },
		{ 190, 300, 20 },
		{ 190, 301, 20 },
		{ 190, 302, 20 },
		{ 190, 303, 20 },
		{ 190, 304, 20 },
		{ 190, 305, 20 },
		{ 190, 306, 20 },
		{ 190, 307, 20 },
		{ 191, 274, 20 },
		{ 191, 275, 20 },
		{ 191, 276, 20 },
		{ 191, 277, 20 },
		{ 191, 278, 20 },
		{ 191, 279, 20 },
		{ 191, 280, 20 },
		{ 191, 281, 20 },
		{ 191, 282, 20 },
		{ 191, 283, 20 },
		{ 191, 284, 20 },
		{ 191, 285, 20 },
		{ 191, 286, 20 },
		{ 191, 287, 20 },
		{ 191, 288, 20 },
		{ 191, 289, 20 },
		{ 191, 290, 20 },
		{ 191, 291, 20 },
		{ 191, 292, 20 },
		{ 191, 293, 20 },
		{ 191, 294, 20 },
		{ 191, 295, 20 },
		{ 191, 296, 20 },
		{ 191, 297, 20 },
		{ 191, 298, 20 },
		{ 191, 299, 20 },
		{ 191, 300, 20 },
		{ 191, 301, 20 },
		{ 191, 302, 20 },
		{ 191, 303, 20 },
		{ 191, 304, 20 },
		{ 191, 305, 20 },
		{ 191, 306, 20 },
		{ 192, 274, 20 },
		{ 192, 275, 20 },
		{ 192, 276, 20 },
		{ 192, 277, 20 },
		{ 192, 278, 20 },
		{ 192, 279, 20 },
		{ 192, 280, 20 },
		{ 192, 281, 20 },
		{ 192, 282, 20 },
		{ 192, 283, 20 },
		{ 192, 284, 20 },
		{ 192, 285, 20 },
		{ 192, 286, 20 },
		{ 192, 287, 20 },
		{ 192, 288, 20 },
		{ 192, 289, 20 },
		{ 192, 290, 20 },
		{ 192, 291, 20 },
		{ 192, 292, 20 },
		{ 192, 293, 20 },
		{ 192, 294, 20 },
		{ 192, 295, 20 },
		{ 192, 296, 20 },
		{ 192, 297, 20 },
		{ 192, 298, 20 },
		{ 192, 299, 20 },
		{ 192, 300, 20 },
		{ 192, 301, 20 },
		{ 192, 302, 20 },
		{ 192, 303, 20 },
		{ 192, 304, 20 },
		{ 192, 305, 20 },
		{ 193, 274, 20 },
		{ 193, 275, 20 },
		{ 193, 276, 20 },
		{ 193, 277, 20 },
		{ 193, 278, 20 },
		{ 193, 279, 20 },
		{ 193, 280, 20 },
		{ 193, 281, 20 },
		{ 193, 282, 20 },
		{ 193, 283, 20 },
		{ 193, 284, 20 },
		{ 193, 285, 20 },
		{ 193, 286, 20 },
		{ 193, 287, 20 },
		{ 193, 288, 20 },
		{ 193, 289, 20 },
		{ 193, 290, 20 },
		{ 193, 291, 20 },
		{ 193, 292, 20 },
		{ 193, 293, 20 },
		{ 193, 294, 20 },
		{ 193, 295, 20 },
		{ 193, 296, 20 },
		{ 193, 297, 20 },
		{ 193, 298, 20 },
		{ 193, 299, 20 },
		{ 193, 300, 20 },
		{ 193, 301, 20 },
		{ 193, 302, 20 },
		{ 193, 303, 20 },
		{ 193, 304, 20 },
		{ 194, 273, 20 },
		{ 194, 274, 20 },
		{ 194, 275, 20 },
		{ 194, 276, 20 },
		{ 194, 277, 20 },
		{ 194, 278, 20 },
		{ 194, 279, 20 },
		{ 194, 280, 20 },
		{ 194, 281, 20 },
		{ 194, 282, 20 },
		{ 194, 283, 20 },
		{ 194, 284, 20 },
		{ 194, 285, 20 },
		{ 194, 286, 20 },
		{ 194, 287, 20 },
		{ 194, 288, 20 },
		{ 194, 289, 20 },
		{ 194, 290, 20 },
		{ 194, 291, 20 },
		{ 194, 292, 20 },
		{ 194, 293, 20 },
		{ 194, 294, 20 },
		{ 194, 295, 20 },
		{ 194, 296, 20 },
		{ 194, 297, 20 },
		{ 194, 298, 20 },
		{ 194, 299, 20 },
		{ 194, 300, 20 },
		{ 194, 301, 20 },
		{ 194, 302, 20 },
		{ 194, 303, 20 },
		{ 194, 304, 20 },
		{ 195, 273, 20 },
		{ 195, 274, 20 },
		{ 195, 275, 20 },
		{ 195, 276, 20 },
		{ 195, 277, 20 },
		{ 195, 278, 20 },
		{ 195, 279, 20 },
		{ 195, 280, 20 },
		{ 195, 281, 20 },
		{ 195, 282, 20 },
		{ 195, 283, 20 },
		{ 195, 284, 20 },
		{ 195, 285, 20 },
		{ 195, 286, 20 },
		{ 195, 287, 20 },
		{ 195, 288, 20 },
		{ 195, 289, 20 },
		{ 195, 290, 20 },
		{ 195, 291, 20 },
		{ 195, 292, 20 },
		{ 195, 293, 20 },
		{ 195, 294, 20 },
		{ 195, 295, 20 },
		{ 195, 296, 20 },
		{ 195, 297, 20 },
		{ 195, 298, 20 },
		{ 195, 299, 20 },
		{ 195, 300, 20 },
		{ 195, 301, 20 },
		{ 195, 302, 20 },
		{ 195, 303, 20 },
		{ 196, 273, 20 },
		{ 196, 274, 20 },
		{ 196, 275, 20 },
		{ 196, 276, 20 },
		{ 196, 277, 20 },
		{ 196, 278, 20 },
		{ 196, 279, 20 },
		{ 196, 280, 20 },
		{ 196, 281, 20 },
		{ 196, 282, 20 },
		{ 196, 283, 20 },
		{ 196, 284, 20 },
		{ 196, 285, 20 },
		{ 196, 286, 20 },
		{ 196, 287, 20 },
		{ 196, 288, 20 },
		{ 196, 289, 20 },
		{ 196, 290, 20 },
		{ 196, 291, 20 },
		{ 196, 292, 20 },
		{ 196, 293, 20 },
		{ 196, 294, 20 },
		{ 196, 295, 20 },
		{ 196, 296, 20 },
		{ 196, 297, 20 },
		{ 196, 298, 20 },
		{ 196, 299, 20 },
		{ 196, 300, 20 },
		{ 196, 301, 20 },
		{ 196, 302, 20 },
		{ 197, 273, 20 },
		{ 197, 274, 20 },
		{ 197, 275, 20 },
		{ 197, 276, 20 },
		{ 197, 277, 20 },
		{ 197, 278, 20 },
		{ 197, 279, 20 },
		{ 197, 280, 20 },
		{ 197, 281, 20 },
		{ 197, 282, 20 },
		{ 197, 283, 20 },
		{ 197, 284, 20 },
		{ 197, 285, 20 },
		{ 197, 286, 20 },
		{ 197, 287, 20 },
		{ 197, 288, 20 },
		{ 197, 289, 20 },
		{ 197, 290, 20 },
		{ 197, 291, 20 },
		{ 197, 292, 20 },
		{ 197, 293, 20 },
		{ 197, 294, 20 },
		{ 197, 295, 20 },
		{ 197, 296, 20 },
		{ 197, 297, 20 },
		{ 197, 298, 20 },
		{ 197, 299, 20 },
		{ 197, 300, 20 },
		{ 197, 301, 20 },
		{ 198, 272, 20 },
		{ 198, 273, 20 },
		{ 198, 274, 20 },
		{ 198, 275, 20 },
		{ 198, 276, 20 },
		{ 198, 277, 20 },
		{ 198, 278, 20 },
		{ 198, 279, 20 },
		{ 198, 280, 20 },
		{ 198, 281, 20 },
		{ 198, 282, 20 },
		{ 198, 283, 20 },
		{ 198, 284, 20 },
		{ 198, 285, 20 },
		{ 198, 286, 20 },
		{ 198, 287, 20 },
		{ 198, 288, 20 },
		{ 198, 289, 20 },
		{ 198, 290, 20 },
		{ 198, 291, 20 },
		{ 198, 292, 20 },
		{ 198, 293, 20 },
		{ 198, 294, 20 },
		{ 198, 295, 20 },
		{ 198, 296, 20 },
		{ 198, 297, 20 },
		{ 198, 298, 20 },
		{ 198, 299, 20 },
		{ 198, 300, 20 },
		{ 198, 301, 20 },
		{ 199, 272, 20 },
		{ 199, 273, 20 },
		{ 199, 274, 20 },
		{ 199, 275, 20 },
		{ 199, 276, 20 },
		{ 199, 277, 20 },
		{ 199, 278, 20 },
		{ 199, 279, 20 },
		{ 199, 280, 20 },
		{ 199, 281, 20 },
		{ 199, 282, 20 },
		{ 199, 283, 20 },
		{ 199, 284, 20 },
		{ 199, 285, 20 },
		{ 199, 286, 20 },
		{ 199, 287, 20 },
		{ 199, 288, 20 },
		{ 199, 289, 20 },
		{ 199, 290, 20 },
		{ 199, 291, 20 },
		{ 199, 292, 20 },
		{ 199, 293, 20 },
		{ 199, 294, 20 },
		{ 199, 295, 20 },
		{ 199, 296, 20 },
		{ 199, 297, 20 },
		{ 199, 298, 20 },
		{ 199, 299, 20 },
		{ 199, 300, 20 },
		{ 200, 272, 20 },
		{ 200, 273, 20 },
		{ 200, 274, 20 },
		{ 200, 275, 20 },
		{ 200, 276, 20 },
		{ 200, 277, 20 },
		{ 200, 278, 20 },
		{ 200, 279, 20 },
		{ 200, 280, 20 },
		{ 200, 281, 20 },
		{ 200, 282, 20 },
		{ 200, 283, 20 },
		{ 200, 284, 20 },
		{ 200, 285, 20 },
		{ 200, 286, 20 },
		{ 200, 287, 20 },
		{ 200, 288, 20 },
		{ 200, 289, 20 },
		{ 200, 290, 20 },
		{ 200, 291, 20 },
		{ 200, 292, 20 },
		{ 200, 293, 20 },
		{ 200, 294, 20 },
		{ 200, 295, 20 },
		{ 200, 296, 20 },
		{ 200, 297, 20 },
		{ 200, 298, 20 },
		{ 200, 299, 20 },
		{ 201, 272, 20 },
		{ 201, 273, 20 },
		{ 201, 274, 20 },
		{ 201, 275, 20 },
		{ 201, 276, 20 },
		{ 201, 277, 20 },
		{ 201, 278, 20 },
		{ 201, 279, 20 },
		{ 201, 280, 20 },
		{ 201, 281, 20 },
		{ 201, 282, 20 },
		{ 201, 283, 20 },
		{ 201, 284, 20 },
		{ 201, 285, 20 },
		{ 201, 286, 20 },
		{ 201, 287, 20 },
		{ 201, 288, 20 },
		{ 201, 289, 20 },
		{ 201, 290, 20 },
		{ 201, 291, 20 },
		{ 201, 292, 20 },
		{ 201, 293, 20 },
		{ 201, 294, 20 },
		{ 201, 295, 20 },
		{ 201, 296, 20 },
		{ 201, 297, 20 },
		{ 201, 298, 20 },
		{ 202, 271, 20 },
		{ 202, 272, 20 },
		{ 202, 273, 20 },
		{ 202, 274, 20 },
		{ 202, 275, 20 },
		{ 202, 276, 20 },
		{ 202, 277, 20 },
		{ 202, 278, 20 },
		{ 202, 279, 20 },
		{ 202, 280, 20 },
		{ 202, 281, 20 },
		{ 202, 282, 20 },
		{ 202, 283, 20 },
		{ 202, 284, 20 },
		{ 202, 285, 20 },
		{ 202, 286, 20 },
		{ 202, 287, 20 },
		{ 202, 288, 20 },
		{ 202, 289, 20 },
		{ 202, 290, 20 },
		{ 202, 291, 20 },
		{ 202, 292, 20 },
		{ 202, 293, 20 },
		{ 202, 294, 20 },
		{ 202, 295, 20 },
		{ 202, 296, 20 },
		{ 202, 297, 20 },
		{ 202, 298, 20 },
		{ 203, 271, 20 },
		{ 203, 272, 20 },
		{ 203, 273, 20 },
		{ 203, 274, 20 },
		{ 203, 275, 20 },
		{ 203, 276, 20 },
		{ 203, 277, 20 },
		{ 203, 278, 20 },
		{ 203, 279, 20 },
		{ 203, 280, 20 },
		{ 203, 281, 20 },
		{ 203, 282, 20 },
		{ 203, 283, 20 },
		{ 203, 284, 20 },
		{ 203, 285, 20 },
		{ 203, 286, 20 },
		{ 203, 287, 20 },
		{ 203, 288, 20 },
		{ 203, 289, 20 },
		{ 203, 290, 20 },
		{ 203, 291, 20 },
		{ 203, 292, 20 },
		{ 203, 293, 20 },
		{ 203, 294, 20 },
		{ 203, 295, 20 },
		{ 203, 296, 20 },
		{ 203, 297, 20 },
		{ 204, 271, 20 },
		{ 204, 272, 20 },
		{ 204, 273, 20 },
		{ 204, 274, 20 },
		{ 204, 275, 20 },
		{ 204, 276, 20 },
		{ 204, 277, 20 },
		{ 204, 278, 20 },
		{ 204, 279, 20 },
		{ 204, 280, 20 },
		{ 204, 281, 20 },
		{ 204, 282, 20 },
		{ 204, 283, 20 },
		{ 204, 284, 20 },
		{ 204, 285, 20 },
		{ 204, 286, 20 },
		{ 204, 287, 20 },
		{ 204, 288, 20 },
		{ 204, 289, 20 },
		{ 204, 290, 20 },
		{ 204, 291, 20 },
		{ 204, 292, 20 },
		{ 204, 293, 20 },
		{ 204, 294, 20 },
		{ 204, 295, 20 },
		{ 204, 296, 20 },
		{ 205, 271, 20 },
		{ 205, 272, 20 },
		{ 205, 273, 20 },
		{ 205, 274, 20 },
		{ 205, 275, 20 },
		{ 205, 276, 20 },
		{ 205, 277, 20 },
		{ 205, 278, 20 },
		{ 205, 279, 20 },
		{ 205, 280, 20 },
		{ 205, 281, 20 },
		{ 205, 282, 20 },
		{ 205, 283, 20 },
		{ 205, 284, 20 },
		{ 205, 285, 20 },
		{ 205, 286, 20 },
		{ 205, 287, 20 },
		{ 205, 288, 20 },
		{ 205, 289, 20 },
		{ 205, 290, 20 },
		{ 205, 291, 20 },
		{ 205, 292, 20 },
		{ 205, 293, 20 },
		{ 205, 294, 20 },
		{ 205, 295, 20 },
		{ 206, 270, 20 },
		{ 206, 271, 20 },
		{ 206, 272, 20 },
		{ 206, 273, 20 },
		{ 206, 274, 20 },
		{ 206, 275, 20 },
		{ 206, 276, 20 },
		{ 206, 277, 20 },
		{ 206, 278, 20 },
		{ 206, 279, 20 },
		{ 206, 280, 20 },
		{ 206, 281, 20 },
		{ 206, 282, 20 },
		{ 206, 283, 20 },
		{ 206, 284, 20 },
		{ 206, 285, 20 },
		{ 206, 286, 20 },
		{ 206, 287, 20 },
		{ 206, 288, 20 },
		{ 206, 289, 20 },
		{ 206, 290, 20 },
		{ 206, 291, 20 },
		{ 206, 292, 20 },
		{ 206, 293, 20 },
		{ 206, 294, 20 },
		{ 206, 295, 20 },
		{ 207, 270, 20 },
		{ 207, 271, 20 },
		{ 207, 272, 20 },
		{ 207, 273, 20 },
		{ 207, 274, 20 },
		{ 207, 275, 20 },
		{ 207, 276, 20 },
		{ 207, 277, 20 },
		{ 207, 278, 20 },
		{ 207, 279, 20 },
		{ 207, 280, 20 },
		{ 207, 281, 20 },
		{ 207, 282, 20 },
		{ 207, 283, 20 },
		{ 207, 284, 20 },
		{ 207, 285, 20 },
		{ 207, 286, 20 },
		{ 207, 287, 20 },
		{ 207, 288, 20 },
		{ 207, 289, 20 },
		{ 207, 290, 20 },
		{ 207, 291, 20 },
		{ 207, 292, 20 },
		{ 207, 293, 20 },
		{ 207, 294, 20 },
		{ 208, 270, 20 },
		{ 208, 271, 20 },
		{ 208, 272, 20 },
		{ 208, 273, 20 },
		{ 208, 274, 20 },
		{ 208, 275, 20 },
		{ 208, 276, 20 },
		{ 208, 277, 20 },
		{ 208, 278, 20 },
		{ 208, 279, 20 },
		{ 208, 280, 20 },
		{ 208, 281, 20 },
		{ 208, 282, 20 },
		{ 208, 283, 20 },
		{ 208, 284, 20 },
		{ 208, 285, 20 },
		{ 208, 286, 20 },
		{ 208, 287, 20 },
		{ 208, 288, 20 },
		{ 208, 289, 20 },
		{ 208, 290, 20 },
		{ 208, 291, 20 },
		{ 208, 292, 20 },
		{ 208, 293, 20 },
		{ 209, 270, 20 },
		{ 209, 271, 20 },
		{ 209, 272, 20 },
		{ 209, 273, 20 },
		{ 209, 274, 20 },
		{ 209, 275, 20 },
		{ 209, 276, 20 },
		{ 209, 277, 20 },
		{ 209, 278, 20 },
		{ 209, 279, 20 },
		{ 209, 280, 20 },
		{ 209, 281, 20 },
		{ 209, 282, 20 },
		{ 209, 283, 20 },
		{ 209, 284, 20 },
		{ 209, 285, 20 },
		{ 209, 286, 20 },
		{ 209, 287, 20 },
		{ 209, 288, 20 },
		{ 209, 289, 20 },
		{ 209, 290, 20 },
		{ 209, 291, 20 },
		{ 209, 292, 20 },
		{ 210, 269, 20 },
		{ 210, 270, 20 },
		{ 210, 271, 20 },
		{ 210, 272, 20 },
		{ 210, 273, 20 },
		{ 210, 274, 20 },
		{ 210, 275, 20 },
		{ 210, 276, 20 },
		{ 210, 277, 20 },
		{ 210, 278, 20 },
		{ 210, 279, 20 },
		{ 210, 280, 20 },
		{ 210, 281, 20 },
		{ 210, 282, 20 },
		{ 210, 283, 20 },
		{ 210, 284, 20 },
		{ 210, 285, 20 },
		{ 210, 286, 20 },
		{ 210, 287, 20 },
		{ 210, 288, 20 },
		{ 210, 289, 20 },
		{ 210, 290, 20 },
		{ 210, 291, 20 },
		{ 210, 292, 20 },
		{ 211, 269, 20 },
		{ 211, 270, 20 },
		{ 211, 271, 20 },
		{ 211, 272, 20 },
		{ 211, 273, 20 },
		{ 211, 274, 20 },
		{ 211, 275, 20 },
		{ 211, 276, 20 },
		{ 211, 277, 20 },
		{ 211, 278, 20 },
		{ 211, 279, 20 },
		{ 211, 280, 20 },
		{ 211, 281, 20 },
		{ 211, 282, 20 },
		{ 211, 283, 20 },
		{ 211, 284, 20 },
		{ 211, 285, 20 },
		{ 211, 286, 20 },
		{ 211, 287, 20 },
		{ 211, 288, 20 },
		{ 211, 289, 20 },
		{ 211, 290, 20 },
		{ 211, 291, 20 },
		{ 212, 269, 20 },
		{ 212, 270, 20 },
		{ 212, 271, 20 },
		{ 212, 272, 20 },
		{ 212, 273, 20 },
		{ 212, 274, 20 },
		{ 212, 275, 20 },
		{ 212, 276, 20 },
		{ 212, 277, 20 },
		{ 212, 278, 20 },
		{ 212, 279, 20 },
		{ 212, 280, 20 },
		{ 212, 281, 20 },
		{ 212, 282, 20 },
		{ 212, 283, 20 },
		{ 212, 284, 20 },
		{ 212, 285, 20 },
		{ 212, 286, 20 },
		{ 212, 287, 20 },
		{ 212, 288, 20 },
		{ 212, 289, 20 },
		{ 212, 290, 20 },
		{ 213, 269, 20 },
		{ 213, 270, 20 },
		{ 213, 271, 20 },
		{ 213, 272, 20 },
		{ 213, 273, 20 },
		{ 213, 274, 20 },
		{ 213, 275, 20 },
		{ 213, 276, 20 },
		{ 213, 277, 20 },
		{ 213, 278, 20 },
		{ 213, 279, 20 },
		{ 213, 280, 20 },
		{ 213, 281, 20 },
		{ 213, 282, 20 },
		{ 213, 283, 20 },
		{ 213, 284, 20 },
		{ 213, 285, 20 },
		{ 213, 286, 20 },
		{ 213, 287, 20 },
		{ 213, 288, 20 },
		{ 213, 289, 20 },
		{ 214, 268, 20 },
		{ 214, 269, 20 },
		{ 214, 270, 20 },
		{ 214, 271, 20 },
		{ 214, 272, 20 },
		{ 214, 273, 20 },
		{ 214, 274, 20 },
		{ 214, 275, 20 },
		{ 214, 276, 20 },
		{ 214, 277, 20 },
		{ 214, 278, 20 },
		{ 214, 279, 20 },
		{ 214, 280, 20 },
		{ 214, 281, 20 },
		{ 214, 282, 20 },
		{ 214, 283, 20 },
		{ 214, 284, 20 },
		{ 214, 285, 20 },
		{ 214, 286, 20 },
		{ 214, 287, 20 },
		{ 214, 288, 20 },
		{ 214, 289, 20 },
		{ 215, 268, 20 },
		{ 215, 269, 20 },
		{ 215, 270, 20 },
		{ 215, 271, 20 },
		{ 215, 272, 20 },
		{ 215, 273, 20 },
		{ 215, 274, 20 },
		{ 215, 275, 20 },
		{ 215, 276, 20 },
		{ 215, 277, 20 },
		{ 215, 278, 20 },
		{ 215, 279, 20 },
		{ 215, 280, 20 },
		{ 215, 281, 20 },
		{ 215, 282, 20 },
		{ 215, 283, 20 },
		{ 215, 284, 20 },
		{ 215, 285, 20 },
		{ 215, 286, 20 },
		{ 215, 287, 20 },
		{ 215, 288, 20 },
		{ 216, 268, 20 },
		{ 216, 269, 20 },
		{ 216, 270, 20 },
		{ 216, 271, 20 },
		{ 216, 272, 20 },
		{ 216, 273, 20 },
		{ 216, 274, 20 },
		{ 216, 275, 20 },
		{ 216, 276, 20 },
		{ 216, 277, 20 },
		{ 216, 278, 20 },
		{ 216, 279, 20 },
		{ 216, 280, 20 },
		{ 216, 281, 20 },
		{ 216, 282, 20 },
		{ 216, 283, 20 },
		{ 216, 284, 20 },
		{ 216, 285, 20 },
		{ 216, 286, 20 },
		{ 216, 287, 20 },
		{ 217, 268, 20 },
		{ 217, 269, 20 },
		{ 217, 270, 20 },
		{ 217, 271, 20 },
		{ 217, 272, 20 },
		{ 217, 273, 20 },
		{ 217, 274, 20 },
		{ 217, 275, 20 },
		{ 217, 276, 20 },
		{ 217, 277, 20 },
		{ 217, 278, 20 },
		{ 217, 279, 20 },
		{ 217, 280, 20 },
		{ 217, 281, 20 },
		{ 217, 282, 20 },
		{ 217, 283, 20 },
		{ 217, 284, 20 },
		{ 217, 285, 20 },
		{ 217, 286, 20 },
		{ 218, 267, 20 },
		{ 218, 268, 20 },
		{ 218, 269, 20 },
		{ 218, 270, 20 },
		{ 218, 271, 20 },
		{ 218, 272, 20 },
		{ 218, 273, 20 },
		{ 218, 274, 20 },
		{ 218, 275, 20 },
		{ 218, 276, 20 },
		{ 218, 277, 20 },
		{ 218, 278, 20 },
		{ 218, 279, 20 },
		{ 218, 280, 20 },
		{ 218, 281, 20 },
		{ 218, 282, 20 },
		{ 218, 283, 20 },
		{ 218, 284, 20 },
		{ 218, 285, 20 },
		{ 218, 286, 20 },
		{ 219, 267, 20 },
		{ 219, 268, 20 },
		{ 219, 269, 20 },
		{ 219, 270, 20 },
		{ 219, 271, 20 },
		{ 219, 272, 20 },
		{ 219, 273, 20 },
		{ 219, 274, 20 },
		{ 219, 275, 20 },
		{ 219, 276, 20 },
		{ 219, 277, 20 },
		{ 219, 278, 20 },
		{ 219, 279, 20 },
		{ 219, 280, 20 },
		{ 219, 281, 20 },
		{ 219, 282, 20 },
		{ 219, 283, 20 },
		{ 219, 284, 20 },
		{ 219, 285, 20 },
		{ 220, 267, 20 },
		{ 220, 268, 20 },
		{ 220, 269, 20 },
		{ 220, 270, 20 },
		{ 220, 271, 20 },
		{ 220, 272, 20 },
		{ 220, 273, 20 },
		{ 220, 274, 20 },
		{ 220, 275, 20 },
		{ 220, 276, 20 },
		{ 220, 277, 20 },
		{ 220, 278, 20 },
		{ 220, 279, 20 },
		{ 220, 280, 20 },
		{ 220, 281, 20 },
		{ 220, 282, 20 },
		{ 220, 283, 20 },
		{ 220, 284, 20 },
		{ 221, 267, 20 },
		{ 221, 268, 20 },
		{ 221, 269, 20 },
		{ 221, 270, 20 },
		{ 221, 271, 20 },
		{ 221, 272, 20 },
		{ 221, 273, 20 },
		{ 221, 274, 20 },
		{ 221, 275, 20 },
		{ 221, 276, 20 },
		{ 221, 277, 20 },
		{ 221, 278, 20 },
		{ 221, 279, 20 },
		{ 221, 280, 20 },
		{ 221, 281, 20 },
		{ 221, 282, 20 },
		{ 221, 283, 20 },
		{ 222, 266, 20 },
		{ 222, 267, 20 },
		{ 222, 268, 20 },
		{ 222, 269, 20 },
		{ 222, 270, 20 },
		{ 222, 271, 20 },
		{ 222, 272, 20 },
		{ 222, 273, 20 },
		{ 222, 274, 20 },
		{ 222, 275, 20 },
		{ 222, 276, 20 },
		{ 222, 277, 20 },
		{ 222, 278, 20 },
		{ 222, 279, 20 },
		{ 222, 280, 20 },
		{ 222, 281, 20 },
		{ 222, 282, 20 },
		{ 222, 283, 20 },
		{ 223, 266, 20 },
		{ 223, 267, 20 },
		{ 223, 268, 20 },
		{ 223, 269, 20 },
		{ 223, 270, 20 },
		{ 223, 271, 20 },
		{ 223, 272, 20 },
		{ 223, 273, 20 },
		{ 223, 274, 20 },
		{ 223, 275, 20 },
		{ 223, 276, 20 },
		{ 223, 277, 20 },
		{ 223, 278, 20 },
		{ 223, 279, 20 },
		{ 223, 280, 20 },
		{ 223, 281, 20 },
		{ 223, 282, 20 },
		{ 224, 266, 20 },
		{ 224, 267, 20 },
		{ 224, 268, 20 },
		{ 224, 269, 20 },
		{ 224, 270, 20 },
		{ 224, 271, 20 },
		{ 224, 272, 20 },
		{ 224, 273, 20 },
		{ 224, 274, 20 },
		{ 224, 275, 20 },
		{ 224, 276, 20 },
		{ 224, 277, 20 },
		{ 224, 278, 20 },
		{ 224, 279, 20 },
		{ 224, 280, 20 },
		{ 224, 281, 20 },
		{ 225, 266, 20 },
		{ 225, 267, 20 },
		{ 225, 268, 20 },
		{ 225, 269, 20 },
		{ 225, 270, 20 },
		{ 225, 271, 20 },
		{ 225, 272, 20 },
		{ 225, 273, 20 },
		{ 225, 274, 20 },
		{ 225, 275, 20 },
		{ 225, 276, 20 },
		{ 225, 277, 20 },
		{ 225, 278, 20 },
		{ 225, 279, 20 },
		{ 225, 280, 20 },
		{ 226, 265, 20 },
		{ 226, 266, 20 },
		{ 226, 267, 20 },
		{ 226, 268, 20 },
		{ 226, 269, 20 },
		{ 226, 270, 20 },
		{ 226, 271, 20 },
		{ 226, 272, 20 },
		{ 226, 273, 20 },
		{ 226, 274, 20 },
		{ 226, 275, 20 },
		{ 226, 276, 20 },
		{ 226, 277, 20 },
		{ 226, 278, 20 },
		{ 226, 279, 20 },
		{ 226, 280, 20 },
		{ 227, 265, 20 },
		{ 227, 266, 20 },
		{ 227, 267, 20 },
		{ 227, 268, 20 },
		{ 227, 269, 20 },
		{ 227, 270, 20 },
		{ 227, 271, 20 },
		{ 227, 272, 20 },
		{ 227, 273, 20 },
		{ 227, 274, 20 },
		{ 227, 275, 20 },
		{ 227, 276, 20 },
		{ 227, 277, 20 },
		{ 227, 278, 20 },
		{ 227, 279, 20 },
		{ 228, 265, 20 },
		{ 228, 266, 20 },
		{ 228, 267, 20 },
		{ 228, 268, 20 },
		{ 228, 269, 20 },
		{ 228, 270, 20 },
		{ 228, 271, 20 },
		{ 228, 272, 20 },
		{ 228, 273, 20 },
		{ 228, 274, 20 },
		{ 228, 275, 20 },
		{ 228, 276, 20 },
		{ 228, 277, 20 },
		{ 228, 278, 20 },
		{ 229, 265, 20 },
		{ 229, 266, 20 },
		{ 229, 267, 20 },
		{ 229, 268, 20 },
		{ 229, 269, 20 },
		{ 229, 270, 20 },
		{ 229, 271, 20 },
		{ 229, 272, 20 },
		{ 229, 273, 20 },
		{ 229, 274, 20 },
		{ 229, 275, 20 },
		{ 229, 276, 20 },
		{ 229, 277, 20 },
		{ 230, 264, 20 },
		{ 230, 265, 20 },
		{ 230, 266, 20 },
		{ 230, 267, 20 },
		{ 230, 268, 20 },
		{ 230, 269, 20 },
		{ 230, 270, 20 },
		{ 230, 271, 20 },
		{ 230, 272, 20 },
		{ 230, 273, 20 },
		{ 230, 274, 20 },
		{ 230, 275, 20 },
		{ 230, 276, 20 },
		{ 230, 277, 20 },
		{ 231, 264, 20 },
		{ 231, 265, 20 },
		{ 231, 266, 20 },
		{ 231, 267, 20 },
		{ 231, 268, 20 },
		{ 231, 269, 20 },
		{ 231, 270, 20 },
		{ 231, 271, 20 },
		{ 231, 272, 20 },
		{ 231, 273, 20 },
		{ 231, 274, 20 },
		{ 231, 275, 20 },
		{ 231, 276, 20 },
		{ 232, 264, 20 },
		{ 232, 265, 20 },
		{ 232, 266, 20 },
		{ 232, 267, 20 },
		{ 232, 268, 20 },
		{ 232, 269, 20 },
		{ 232, 270, 20 },
		{ 232, 271, 20 },
		{ 232, 272, 20 },
		{ 232, 273, 20 },
		{ 232, 274, 20 },
		{ 232, 275, 20 },
		{ 233, 264, 20 },
		{ 233, 265, 20 },
		{ 233, 266, 20 },
		{ 233, 267, 20 },
		{ 233, 268, 20 },
		{ 233, 269, 20 },
		{ 233, 270, 20 },
		{ 233, 271, 20 },
		{ 233, 272, 20 },
		{ 233, 273, 20 },
		{ 233, 274, 20 },
		{ 234, 263, 20 },
		{ 234, 264, 20 },
		{ 234, 265, 20 },
		{ 234, 266, 20 },
		{ 234, 267, 20 },
		{ 234, 268, 20 },
		{ 234, 269, 20 },
		{ 234, 270, 20 },
		{ 234, 271, 20 },
		{ 234, 272, 20 },
		{ 234, 273, 20 },
		{ 234, 274, 20 },
		{ 235, 263, 20 },
		{ 235, 264, 20 },
		{ 235, 265, 20 },
		{ 235, 266, 20 },
		{ 235, 267, 20 },
		{ 235, 268, 20 },
		{ 235, 269, 20 },
		{ 235, 270, 20 },
		{ 235, 271, 20 },
		{ 235, 272, 20 },
		{ 235, 273, 20 },
		{ 236, 263, 20 },
		{ 236, 264, 20 },
		{ 236, 265, 20 },
		{ 236, 266, 20 },
		{ 236, 267, 20 },
		{ 236, 268, 20 },
		{ 236, 269, 20 },
		{ 236, 270, 20 },
		{ 236, 271, 20 },
		{ 236, 272, 20 },
		{ 237, 263, 20 },
		{ 237, 264, 20 },
		{ 237, 265, 20 },
		{ 237, 266, 20 },
		{ 237, 267, 20 },
		{ 237, 268, 20 },
		{ 237, 269, 20 },
		{ 237, 270, 20 },
		{ 237, 271, 20 },
		{ 238, 262, 20 },
		{ 238, 263, 20 },
		{ 238, 264, 20 },
		{ 238, 265, 20 },
		{ 238, 266, 20 },
		{ 238, 267, 20 },
		{ 238, 268, 20 },
		{ 238, 269, 20 },
		{ 238, 270, 20 },
		{ 238, 271, 20 },
		{ 239, 262, 20 },
		{ 239, 263, 20 },
		{ 239, 264, 20 },
		{ 239, 265, 20 },
		{ 239, 266, 20 },
		{ 239, 267, 20 },
		{ 239, 268, 20 },
		{ 239, 269, 20 },
		{ 239, 270, 20 },
		{ 240, 262, 20 },
		{ 240, 263, 20 },
		{ 240, 264, 20 },
		{ 240, 265, 20 },
		{ 240, 266, 20 },
		{ 240, 267, 20 },
		{ 240, 268, 20 },
		{ 240, 269, 20 },
		{ 274, 245, 20 },
		{ 274, 246, 20 },
		{ 274, 247, 20 },
		{ 274, 248, 20 },
		{ 274, 249, 20 },
		{ 274, 250, 20 },
		{ 274, 251, 20 },
		{ 274, 252, 20 },
		{ 275, 244, 20 },
		{ 275, 245, 20 },
		{ 275, 246, 20 },
		{ 275, 247, 20 },
		{ 275, 248, 20 },
		{ 275, 249, 20 },
		{ 275, 250, 20 },
		{ 275, 251, 20 },
		{ 275, 252, 20 },
		{ 276, 243, 20 },
		{ 276, 244, 20 },
		{ 276, 245, 20 },
		{ 276, 246, 20 },
		{ 276, 247, 20 },
		{ 276, 248, 20 },
		{ 276, 249, 20 },
		{ 276, 250, 20 },
		{ 276, 251, 20 },
		{ 276, 252, 20 },
		{ 277, 243, 20 },
		{ 277, 244, 20 },
		{ 277, 245, 20 },
		{ 277, 246, 20 },
		{ 277, 247, 20 },
		{ 277, 248, 20 },
		{ 277, 249, 20 },
		{ 277, 250, 20 },
		{ 277, 251, 20 },
		{ 278, 242, 20 },
		{ 278, 243, 20 },
		{ 278, 244, 20 },
		{ 278, 245, 20 },
		{ 278, 246, 20 },
		{ 278, 247, 20 },
		{ 278, 248, 20 },
		{ 278, 249, 20 },
		{ 278, 250, 20 },
		{ 278, 251, 20 },
		{ 279, 241, 20 },
		{ 279, 242, 20 },
		{ 279, 243, 20 },
		{ 279, 244, 20 },
		{ 279, 245, 20 },
		{ 279, 246, 20 },
		{ 279, 247, 20 },
		{ 279, 248, 20 },
		{ 279, 249, 20 },
		{ 279, 250, 20 },
		{ 279, 251, 20 },
		{ 280, 240, 20 },
		{ 280, 241, 20 },
		{ 280, 242, 20 },
		{ 280, 243, 20 },
		{ 280, 244, 20 },
		{ 280, 245, 20 },
		{ 280, 246, 20 },
		{ 280, 247, 20 },
		{ 280, 248, 20 },
		{ 280, 249, 20 },
		{ 280, 250, 20 },
		{ 280, 251, 20 },
		{ 281, 240, 20 },
		{ 281, 241, 20 },
		{ 281, 242, 20 },
		{ 281, 243, 20 },
		{ 281, 244, 20 },
		{ 281, 245, 20 },
		{ 281, 246, 20 },
		{ 281, 247, 20 },
		{ 281, 248, 20 },
		{ 281, 249, 20 },
		{ 281, 250, 20 },
		{ 282, 239, 20 },
		{ 282, 240, 20 },
		{ 282, 241, 20 },
		{ 282, 242, 20 },
		{ 282, 243, 20 },
		{ 282, 244, 20 },
		{ 282, 245, 20 },
		{ 282, 246, 20 },
		{ 282, 247, 20 },
		{ 282, 248, 20 },
		{ 282, 249, 20 },
		{ 282, 250, 20 },
		{ 283, 238, 20 },
		{ 283, 239, 20 },
		{ 283, 240, 20 },
		{ 283, 241, 20 },
		{ 283, 242, 20 },
		{ 283, 243, 20 },
		{ 283, 244, 20 },
		{ 283, 245, 20 },
		{ 283, 246, 20 },
		{ 283, 247, 20 },
		{ 283, 248, 20 },
		{ 283, 249, 20 },
		{ 283, 250, 20 },
		{ 284, 237, 20 },
		{ 284, 238, 20 },
		{ 284, 239, 20 },
		{ 284, 240, 20 },
		{ 284, 241, 20 },
		{ 284, 242, 20 },
		{ 284, 243, 20 },
		{ 284, 244, 20 },
		{ 284, 245, 20 },
		{ 284, 246, 20 },
		{ 284, 247, 20 },
		{ 284, 248, 20 },
		{ 284, 249, 20 },
		{ 284, 250, 20 },
		{ 285, 237, 20 },
		{ 285, 238, 20 },
		{ 285, 239, 20 },
		{ 285, 240, 20 },
		{ 285, 241, 20 },
		{ 285, 242, 20 },
		{ 285, 243, 20 },
		{ 285, 244, 20 },
		{ 285, 245, 20 },
		{ 285, 246, 20 },
		{ 285, 247, 20 },
		{ 285, 248, 20 },
		{ 285, 249, 20 },
		{ 286, 236, 20 },
		{ 286, 237, 20 },
		{ 286, 238, 20 },
		{ 286, 239, 20 },
		{ 286, 240, 20 },
		{ 286, 241, 20 },
		{ 286, 242, 20 },
		{ 286, 243, 20 },
		{ 286, 244, 20 },
		{ 286, 245, 20 },
		{ 286, 246, 20 },
		{ 286, 247, 20 },
		{ 286, 248, 20 },
		{ 286, 249, 20 },
		{ 287, 235, 20 },
		{ 287, 236, 20 },
		{ 287, 237, 20 },
		{ 287, 238, 20 },
		{ 287, 239, 20 },
		{ 287, 240, 20 },
		{ 287, 241, 20 },
		{ 287, 242, 20 },
		{ 287, 243, 20 },
		{ 287, 244, 20 },
		{ 287, 245, 20 },
		{ 287, 246, 20 },
		{ 287, 247, 20 },
		{ 287, 248, 20 },
		{ 287, 249, 20 },
		{ 288, 234, 20 },
		{ 288, 235, 20 },
		{ 288, 236, 20 },
		{ 288, 237, 20 },
		{ 288, 238, 20 },
		{ 288, 239, 20 },
		{ 288, 240, 20 },
		{ 288, 241, 20 },
		{ 288, 242, 20 },
		{ 288, 243, 20 },
		{ 288, 244, 20 },
		{ 288, 245, 20 },
		{ 288, 246, 20 },
		{ 288, 247, 20 },
		{ 288, 248, 20 },
		{ 288, 249, 20 },
		{ 289, 234, 20 },
		{ 289, 235, 20 },
		{ 289, 236, 20 },
		{ 289, 237, 20 },
		{ 289, 238, 20 },
		{ 289, 239, 20 },
		{ 289, 240, 20 },
		{ 289, 241, 20 },
		{ 289, 242, 20 },
		{ 289, 243, 20 },
		{ 289, 244, 20 },
		{ 289, 245, 20 },
		{ 289, 246, 20 },
		{ 289, 247, 20 },
		{ 289, 248, 20 },
		{ 290, 233, 20 },
		{ 290, 234, 20 },
		{ 290, 235, 20 },
		{ 290, 236, 20 },
		{ 290, 237, 20 },
		{ 290, 238, 20 },
		{ 290, 239, 20 },
		{ 290, 240, 20 },
		{ 290, 241, 20 },
		{ 290, 242, 20 },
		{ 290, 243, 20 },
		{ 290, 244, 20 },
		{ 290, 245, 20 },
		{ 290, 246, 20 },
		{ 290, 247, 20 },
		{ 290, 248, 20 },
		{ 291, 232, 20 },
		{ 291, 233, 20 },
		{ 291, 234, 20 },
		{ 291, 235, 20 },
		{ 291, 236, 20 },
		{ 291, 237, 20 },
		{ 291, 238, 20 },
		{ 291, 239, 20 },
		{ 291, 240, 20 },
		{ 291, 241, 20 },
		{ 291, 242, 20 },
		{ 291, 243, 20 },
		{ 291, 244, 20 },
		{ 291, 245, 20 },
		{ 291, 246, 20 },
		{ 291, 247, 20 },
		{ 291, 248, 20 },
		{ 292, 231, 20 },
		{ 292, 232, 20 },
		{ 292, 233, 20 },
		{ 292, 234, 20 },
		{ 292, 235, 20 },
		{ 292, 236, 20 },
		{ 292, 237, 20 },
		{ 292, 238, 20 },
		{ 292, 239, 20 },
		{ 292, 240, 20 },
		{ 292, 241, 20 },
		{ 292, 242, 20 },
		{ 292, 243, 20 },
		{ 292, 244, 20 },
		{ 292, 245, 20 },
		{ 292, 246, 20 },
		{ 292, 247, 20 },
		{ 292, 248, 20 },
		{ 293, 231, 20 },
		{ 293, 232, 20 },
		{ 293, 233, 20 },
		{ 293, 234, 20 },
		{ 293, 235, 20 },
		{ 293, 236, 20 },
		{ 293, 237, 20 },
		{ 293, 238, 20 },
		{ 293, 239, 20 },
		{ 293, 240, 20 },
		{ 293, 241, 20 },
		{ 293, 242, 20 },
		{ 293, 243, 20 },
		{ 293, 244, 20 },
		{ 293, 245, 20 },
		{ 293, 246, 20 },
		{ 293, 247, 20 },
		{ 294, 230, 20 },
		{ 294, 231, 20 },
		{ 294, 232, 20 },
		{ 294, 233, 20 },
		{ 294, 234, 20 },
		{ 294, 235, 20 },
		{ 294, 236, 20 },
		{ 294, 237, 20 },
		{ 294, 238, 20 },
		{ 294, 239, 20 },
		{ 294, 240, 20 },
		{ 294, 241, 20 },
		{ 294, 242, 20 },
		{ 294, 243, 20 },
		{ 294, 244, 20 },
		{ 294, 245, 20 },
		{ 294, 246, 20 },
		{ 294, 247, 20 },
		{ 295, 229, 20 },
		{ 295, 230, 20 },
		{ 295, 231, 20 },
		{ 295, 232, 20 },
		{ 295, 233, 20 },
		{ 295, 234, 20 },
		{ 295, 235, 20 },
		{ 295, 236, 20 },
		{ 295, 237, 20 },
		{ 295, 238, 20 },
		{ 295, 239, 20 },
		{ 295, 240, 20 },
		{ 295, 241, 20 },
		{ 295, 242, 20 },
		{ 295, 243, 20 },
		{ 295, 244, 20 },
		{ 295, 245, 20 },
		{ 295, 246, 20 },
		{ 295, 247, 20 },
		{ 296, 228, 20 },
		{ 296, 229, 20 },
		{ 296, 230, 20 },
		{ 296, 231, 20 },
		{ 296, 232, 20 },
		{ 296, 233, 20 },
		{ 296, 234, 20 },
		{ 296, 235, 20 },
		{ 296, 236, 20 },
		{ 296, 237, 20 },
		{ 296, 238, 20 },
		{ 296, 239, 20 },
		{ 296, 240, 20 },
		{ 296, 241, 20 },
		{ 296, 242, 20 },
		{ 296, 243, 20 },
		{ 296, 244, 20 },
		{ 296, 245, 20 },
		{ 296, 246, 20 },
		{ 296, 247, 20 },
		{ 297, 228, 20 },
		{ 297, 229, 20 },
		{ 297, 230, 20 },
		{ 297, 231, 20 },
		{ 297, 232, 20 },
		{ 297, 233, 20 },
		{ 297, 234, 20 },
		{ 297, 235, 20 },
		{ 297, 236, 20 },
		{ 297, 237, 20 },
		{ 297, 238, 20 },
		{ 297, 239, 20 },
		{ 297, 240, 20 },
		{ 297, 241, 20 },
		{ 297, 242, 20 },
		{ 297, 243, 20 },
		{ 297, 244, 20 },
		{ 297, 245, 20 },
		{ 297, 246, 20 },
		{ 298, 227, 20 },
		{ 298, 228, 20 },
		{ 298, 229, 20 },
		{ 298, 230, 20 },
		{ 298, 231, 20 },
		{ 298, 232, 20 },
		{ 298, 233, 20 },
		{ 298, 234, 20 },
		{ 298, 235, 20 },
		{ 298, 236, 20 },
		{ 298, 237, 20 },
		{ 298, 238, 20 },
		{ 298, 239, 20 },
		{ 298, 240, 20 },
		{ 298, 241, 20 },
		{ 298, 242, 20 },
		{ 298, 243, 20 },
		{ 298, 244, 20 },
		{ 298, 245, 20 },
		{ 298, 246, 20 },
		{ 299, 226, 20 },
		{ 299, 227, 20 },
		{ 299, 228, 20 },
		{ 299, 229, 20 },
		{ 299, 230, 20 },
		{ 299, 231, 20 },
		{ 299, 232, 20 },
		{ 299, 233, 20 },
		{ 299, 234, 20 },
		{ 299, 235, 20 },
		{ 299, 236, 20 },
		{ 299, 237, 20 },
		{ 299, 238, 20 },
		{ 299, 239, 20 },
		{ 299, 240, 20 },
		{ 299, 241, 20 },
		{ 299, 242, 20 },
		{ 299, 243, 20 },
		{ 299, 244, 20 },
		{ 299, 245, 20 },
		{ 299, 246, 20 },
		{ 300, 225, 20 },
		{ 300, 226, 20 },
		{ 300, 227, 20 },
		{ 300, 228, 20 },
		{ 300, 229, 20 },
		{ 300, 230, 20 },
		{ 300, 231, 20 },
		{ 300, 232, 20 },
		{ 300, 233, 20 },
		{ 300, 234, 20 },
		{ 300, 235, 20 },
		{ 300, 236, 20 },
		{ 300, 237, 20 },
		{ 300, 238, 20 },
		{ 300, 239, 20 },
		{ 300, 240, 20 },
		{ 300, 241, 20 },
		{ 300, 242, 20 },
		{ 300, 243, 20 },
		{ 300, 244, 20 },
		{ 300, 245, 20 },
		{ 300, 246, 20 },
		{ 301, 225, 20 },
		{ 301, 226, 20 },
		{ 301, 227, 20 },
		{ 301, 228, 20 },
		{ 301, 229, 20 },
		{ 301, 230, 20 },
		{ 301, 231, 20 },
		{ 301, 232, 20 },
		{ 301, 233, 20 },
		{ 301, 234, 20 },
		{ 301, 235, 20 },
		{ 301, 236, 20 },
		{ 301, 237, 20 },
		{ 301, 238, 20 },
		{ 301, 239, 20 },
		{ 301, 240, 20 },
		{ 301, 241, 20 },
		{ 301, 242, 20 },
		{ 301, 243, 20 },
		{ 301, 244, 20 },
		{ 301, 245, 20 },
		{ 302, 224, 20 },
		{ 302, 225, 20 },
		{ 302, 226, 20 },
		{ 302, 227, 20 },
		{ 302, 228, 20 },
		{ 302, 229, 20 },
		{ 302, 230, 20 },
		{ 302, 231, 20 },
		{ 302, 232, 20 },
		{ 302, 233, 20 },
		{ 302, 234, 20 },
		{ 302, 235, 20 },
		{ 302, 236, 20 },
		{ 302, 237, 20 },
		{ 302, 238, 20 },
		{ 302, 239, 20 },
		{ 302, 240, 20 },
		{ 302, 241, 20 },
		{ 302, 242, 20 },
		{ 302, 243, 20 },
		{ 302, 244, 20 },
		{ 302, 245, 20 },
		{ 303, 223, 20 },
		{ 303, 224, 20 },
		{ 303, 225, 20 },
		{ 303, 226, 20 },
		{ 303, 227, 20 },
		{ 303, 228, 20 },
		{ 303, 229, 20 },
		{ 303, 230, 20 },
		{ 303, 231, 20 },
		{ 303, 232, 20 },
		{ 303, 233, 20 },
		{ 303, 234, 20 },
		{ 303, 235, 20 },
		{ 303, 236, 20 },
		{ 303, 237, 20 },
		{ 303, 238, 20 },
		{ 303, 239, 20 },
		{ 303, 240, 20 },
		{ 303, 241, 20 },
		{ 303, 242, 20 },
		{ 303, 243, 20 },
		{ 303, 244, 20 },
		{ 303, 245, 20 },
		{ 304, 222, 20 },
		{ 304, 223, 20 },
		{ 304, 224, 20 },
		{ 304, 225, 20 },
		{ 304, 226, 20 },
		{ 304, 227, 20 },
		{ 304, 228, 20 },
		{ 304, 229, 20 },
		{ 304, 230, 20 },
		{ 304, 231, 20 },
		{ 304, 232, 20 },
		{ 304, 233, 20 },
		{ 304, 234, 20 },
		{ 304, 235, 20 },
		{ 304, 236, 20 },
		{ 304, 237, 20 },
		{ 304, 238, 20 },
		{ 304, 239, 20 },
		{ 304, 240, 20 },
		{ 304, 241, 20 },
		{ 304, 242, 20 },
		{ 304, 243, 20 },
		{ 304, 244, 20 },
		{ 304, 245, 20 },
		{ 305, 222, 20 },
		{ 305, 223, 20 },
		{ 305, 224, 20 },
		{ 305, 225, 20 },
		{ 305, 226, 20 },
		{ 305, 227, 20 },
		{ 305, 228, 20 },
		{ 305, 229, 20 },
		{ 305, 230, 20 },
		{ 305, 231, 20 },
		{ 305, 232, 20 },
		{ 305, 233, 20 },
		{ 305, 234, 20 },
		{ 305, 235, 20 },
		{ 305, 236, 20 },
		{ 305, 237, 20 },
		{ 305, 238, 20 },
		{ 305, 239, 20 },
		{ 305, 240, 20 },
		{ 305, 241, 20 },
		{ 305, 242, 20 },
		{ 305, 243, 20 },
		{ 305, 244, 20 },
		{ 306, 221, 20 },
		{ 306, 222, 20 },
		{ 306, 223, 20 },
		{ 306, 224, 20 },
		{ 306, 225, 20 },
		{ 306, 226, 20 },
		{ 306, 227, 20 },
		{ 306, 228, 20 },
		{ 306, 229, 20 },
		{ 306, 230, 20 },
		{ 306, 231, 20 },
		{ 306, 232, 20 },
		{ 306, 233, 20 },
		{ 306, 234, 20 },
		{ 306, 235, 20 },
		{ 306, 236, 20 },
		{ 306, 237, 20 },
		{ 306, 238, 20 },
		{ 306, 239, 20 },
		{ 306, 240, 20 },
		{ 306, 241, 20 },
		{ 306, 242, 20 },
		{ 306, 243, 20 },
		{ 306, 244, 20 },
		{ 307, 220, 20 },
		{ 307, 221, 20 },
		{ 307, 222, 20 },
		{ 307, 223, 20 },
		{ 307, 224, 20 },
		{ 307, 225, 20 },
		{ 307, 226, 20 },
		{ 307, 227, 20 },
		{ 307, 228, 20 },
		{ 307, 229, 20 },
		{ 307, 230, 20 },
		{ 307, 231, 20 },
		{ 307, 232, 20 },
		{ 307, 233, 20 },
		{ 307, 234, 20 },
		{ 307, 235, 20 },
		{ 307, 236, 20 },
		{ 307, 237, 20 },
		{ 307, 238, 20 },
		{ 307, 239, 20 },
		{ 307, 240, 20 },
		{ 307, 241, 20 },
		{ 307, 242, 20 },
		{ 307, 243, 20 },
		{ 307, 244, 20 },
		{ 308, 219, 20 },
		{ 308, 220, 20 },
		{ 308, 221, 20 },
		{ 308, 222, 20 },
		{ 308, 223, 20 },
		{ 308, 224, 20 },
		{ 308, 225, 20 },
		{ 308, 226, 20 },
		{ 308, 227, 20 },
		{ 308, 228, 20 },
		{ 308, 229, 20 },
		{ 308, 230, 20 },
		{ 308, 231, 20 },
		{ 308, 232, 20 },
		{ 308, 233, 20 },
		{ 308, 234, 20 },
		{ 308, 235, 20 },
		{ 308, 236, 20 },
		{ 308, 237, 20 },
		{ 308, 238, 20 },
		{ 308, 239, 20 },
		{ 308, 240, 20 },
		{ 308, 241, 20 },
		{ 308, 242, 20 },
		{ 308, 243, 20 },
		{ 308, 244, 20 },
		{ 309, 219, 20 },
		{ 309, 220, 20 },
		{ 309, 221, 20 },
		{ 309, 222, 20 },
		{ 309, 223, 20 },
		{ 309, 224, 20 },
		{ 309, 225, 20 },
		{ 309, 226, 20 },
		{ 309, 227, 20 },
		{ 309, 228, 20 },
		{ 309, 229, 20 },
		{ 309, 230, 20 },
		{ 309, 231, 20 },
		{ 309, 232, 20 },
		{ 309, 233, 20 },
		{ 309, 234, 20 },
		{ 309, 235, 20 },
		{ 309, 236, 20 },
		{ 309, 237, 20 },
		{ 309, 238, 20 },
		{ 309, 239, 20 },
		{ 309, 240, 20 },
		{ 309, 241, 20 },
		{ 309, 242, 20 },
		{ 309, 243, 20 },
		{ 310, 218, 20 },
		{ 310, 219, 20 },
		{ 310, 220, 20 },
		{ 310, 221, 20 },
		{ 310, 222, 20 },
		{ 310, 223, 20 },
		{ 310, 224, 20 },
		{ 310, 225, 20 },
		{ 310, 226, 20 },
		{ 310, 227, 20 },
		{ 310, 228, 20 },
		{ 310, 229, 20 },
		{ 310, 230, 20 },
		{ 310, 231, 20 },
		{ 310, 232, 20 },
		{ 310, 233, 20 },
		{ 310, 234, 20 },
		{ 310, 235, 20 },
		{ 310, 236, 20 },
		{ 310, 237, 20 },
		{ 310, 238, 20 },
		{ 310, 239, 20 },
		{ 310, 240, 20 },
		{ 310, 241, 20 },
		{ 310, 242, 20 },
		{ 310, 243, 20 },
		{ 311, 217, 20 },
		{ 311, 218, 20 },
		{ 311, 219, 20 },
		{ 311, 220, 20 },
		{ 311, 221, 20 },
		{ 311, 222, 20 },
		{ 311, 223, 20 },
		{ 311, 224, 20 },
		{ 311, 225, 20 },
		{ 311, 226, 20 },
		{ 311, 227, 20 },
		{ 311, 228, 20 },
		{ 311, 229, 20 },
		{ 311, 230, 20 },
		{ 311, 231, 20 },
		{ 311, 232, 20 },
		{ 311, 233, 20 },
		{ 311, 234, 20 },
		{ 311, 235, 20 },
		{ 311, 236, 20 },
		{ 311, 237, 20 },
		{ 311, 238, 20 },
		{ 311, 239, 20 },
		{ 311, 240, 20 },
		{ 311, 241, 20 },
		{ 311, 242, 20 },
		{ 311, 243, 20 },
		{ 312, 216, 20 },
		{ 312, 217, 20 },
		{ 312, 218, 20 },
		{ 312, 219, 20 },
		{ 312, 220, 20 },
		{ 312, 221, 20 },
		{ 312, 222, 20 },
		{ 312, 223, 20 },
		{ 312, 224, 20 },
		{ 312, 225, 20 },
		{ 312, 226, 20 },
		{ 312, 227, 20 },
		{ 312, 228, 20 },
		{ 312, 229, 20 },
		{ 312, 230, 20 },
		{ 312, 231, 20 },
		{ 312, 232, 20 },
		{ 312, 233, 20 },
		{ 312, 234, 20 },
		{ 312, 235, 20 },
		{ 312, 236, 20 },
		{ 312, 237, 20 },
		{ 312, 238, 20 },
		{ 312, 239, 20 },
		{ 312, 240, 20 },
		{ 312, 241, 20 },
		{ 312, 242, 20 },
		{ 312, 243, 20 },
		{ 313, 216, 20 },
		{ 313, 217, 20 },
		{ 313, 218, 20 },
		{ 313, 219, 20 },
		{ 313, 220, 20 },
		{ 313, 221, 20 },
		{ 313, 222, 20 },
		{ 313, 223, 20 },
		{ 313, 224, 20 },
		{ 313, 225, 20 },
		{ 313, 226, 20 },
		{ 313, 227, 20 },
		{ 313, 228, 20 },
		{ 313, 229, 20 },
		{ 313, 230, 20 },
		{ 313, 231, 20 },
		{ 313, 232, 20 },
		{ 313, 233, 20 },
		{ 313, 234, 20 },
		{ 313, 235, 20 },
		{ 313, 236, 20 },
		{ 313, 237, 20 },
		{ 313, 238, 20 },
		{ 313, 239, 20 },
		{ 313, 240, 20 },
		{ 313, 241, 20 },
		{ 313, 242, 20 },
		{ 314, 215, 20 },
		{ 314, 216, 20 },
		{ 314, 217, 20 },
		{ 314, 218, 20 },
		{ 314, 219, 20 },
		{ 314, 220, 20 },
		{ 314, 221, 20 },
		{ 314, 222, 20 },
		{ 314, 223, 20 },
		{ 314, 224, 20 },
		{ 314, 225, 20 },
		{ 314, 226, 20 },
		{ 314, 227, 20 },
		{ 314, 228, 20 },
		{ 314, 229, 20 },
		{ 314, 230, 20 },
		{ 314, 231, 20 },
		{ 314, 232, 20 },
		{ 314, 233, 20 },
		{ 314, 234, 20 },
		{ 314, 235, 20 },
		{ 314, 236, 20 },
		{ 314, 237, 20 },
		{ 314, 238, 20 },
		{ 314, 239, 20 },
		{ 314, 240, 20 },
		{ 314, 241, 20 },
		{ 314, 242, 20 },
		{ 315, 214, 20 },
		{ 315, 215, 20 },
		{ 315, 216, 20 },
		{ 315, 217, 20 },
		{ 315, 218, 20 },
		{ 315, 219, 20 },
		{ 315, 220, 20 },
		{ 315, 221, 20 },
		{ 315, 222, 20 },
		{ 315, 223, 20 },
		{ 315, 224, 20 },
		{ 315, 225, 20 },
		{ 315, 226, 20 },
		{ 315, 227, 20 },
		{ 315, 228, 20 },
		{ 315, 229, 20 },
		{ 315, 230, 20 },
		{ 315, 231, 20 },
		{ 315, 232, 20 },
		{ 315, 233, 20 },
		{ 315, 234, 20 },
		{ 315, 235, 20 },
		{ 315, 236, 20 },
		{ 315, 237, 20 },
		{ 315, 238, 20 },
		{ 315, 239, 20 },
		{ 315, 240, 20 },
		{ 315, 241, 20 },
		{ 315, 242, 20 },
		{ 316, 213, 20 },
		{ 316, 214, 20 },
		{ 316, 215, 20 },
		{ 316, 216, 20 },
		{ 316, 217, 20 },
		{ 316, 218, 20 },
		{ 316, 219, 20 },
		{ 316, 220, 20 },
		{ 316, 221, 20 },
		{ 316, 222, 20 },
		{ 316, 223, 20 },
		{ 316, 224, 20 },
		{ 316, 225, 20 },
		{ 316, 226, 20 },
		{ 316, 227, 20 },
		{ 316, 228, 20 },
		{ 316, 229, 20 },
		{ 316, 230, 20 },
		{ 316, 231, 20 },
		{ 316, 232, 20 },
		{ 316, 233, 20 },
		{ 316, 234, 20 },
		{ 316, 235, 20 },
		{ 316, 236, 20 },
		{ 316, 237, 20 },
		{ 316, 238, 20 },
		{ 316, 239, 20 },
		{ 316, 240, 20 },
		{ 316, 241, 20 },
		{ 316, 242, 20 },
		{ 317, 213, 20 },
		{ 317, 214, 20 },
		{ 317, 215, 20 },
		{ 317, 216, 20 },
		{ 317, 217, 20 },
		{ 317, 218, 20 },
		{ 317, 219, 20 },
		{ 317, 220, 20 },
		{ 317, 221, 20 },
		{ 317, 222, 20 },
		{ 317, 223, 20 },
		{ 317, 224, 20 },
		{ 317, 225, 20 },
		{ 317, 226, 20 },
		{ 317, 227, 20 },
		{ 317, 228, 20 },
		{ 317, 229, 20 },
		{ 317, 230, 20 },
		{ 317, 231, 20 },
		{ 317, 232, 20 },
		{ 317, 233, 20 },
		{ 317, 234, 20 },
		{ 317, 235, 20 },
		{ 317, 236, 20 },
		{ 317, 237, 20 },
		{ 317, 238, 20 },
		{ 317, 239, 20 },
		{ 317, 240, 20 },
		{ 317, 241, 20 },
		{ 318, 212, 20 },
		{ 318, 213, 20 },
		{ 318, 214, 20 },
		{ 318, 215, 20 },
		{ 318, 216, 20 },
		{ 318, 217, 20 },
		{ 318, 218, 20 },
		{ 318, 219, 20 },
		{ 318, 220, 20 },
		{ 318, 221, 20 },
		{ 318, 222, 20 },
		{ 318, 223, 20 },
		{ 318, 224, 20 },
		{ 318, 225, 20 },
		{ 318, 226, 20 },
		{ 318, 227, 20 },
		{ 318, 228, 20 },
		{ 318, 229, 20 },
		{ 318, 230, 20 },
		{ 318, 231, 20 },
		{ 318, 232, 20 },
		{ 318, 233, 20 },
		{ 318, 234, 20 },
		{ 318, 235, 20 },
		{ 318, 236, 20 },
		{ 318, 237, 20 },
		{ 318, 238, 20 },
		{ 318, 239, 20 },
		{ 318, 240, 20 },
		{ 318, 241, 20 },
		{ 319, 211, 20 },
		{ 319, 212, 20 },
		{ 319, 213, 20 },
		{ 319, 214, 20 },
		{ 319, 215, 20 },
		{ 319, 216, 20 },
		{ 319, 217, 20 },
		{ 319, 218, 20 },
		{ 319, 219, 20 },
		{ 319, 220, 20 },
		{ 319, 221, 20 },
		{ 319, 222, 20 },
		{ 319, 223, 20 },
		{ 319, 224, 20 },
		{ 319, 225, 20 },
		{ 319, 226, 20 },
		{ 319, 227, 20 },
		{ 319, 228, 20 },
		{ 319, 229, 20 },
		{ 319, 230, 20 },
		{ 319, 231, 20 },
		{ 319, 232, 20 },
		{ 319, 233, 20 },
		{ 319, 234, 20 },
		{ 319, 235, 20 },
		{ 319, 236, 20 },
		{ 319, 237, 20 },
		{ 319, 238, 20 },
		{ 319, 239, 20 },
		{ 319, 240, 20 },
		{ 319, 241, 20 },
		{ 320, 210, 20 },
		{ 320, 211, 20 },
		{ 320, 212, 20 },
		{ 320, 213, 20 },
		{ 320, 214, 20 },
		{ 320, 215, 20 },
		{ 320, 216, 20 },
		{ 320, 217, 20 },
		{ 320, 218, 20 },
		{ 320, 219, 20 },
		{ 320, 220, 20 },
		{ 320, 221, 20 },
		{ 320, 222, 20 },
		{ 320, 223, 20 },
		{ 320, 224, 20 },
		{ 320, 225, 20 },
		{ 320, 226, 20 },
		{ 320, 227, 20 },
		{ 320, 228, 20 },
		{ 320, 229, 20 },
		{ 320, 230, 20 },
		{ 320, 231, 20 },
		{ 320, 232, 20 },
		{ 320, 233, 20 },
		{ 320, 234, 20 },
		{ 320, 235, 20 },
		{ 320, 236, 20 },
		{ 320, 237, 20 },
		{ 320, 238, 20 },
		{ 320, 239, 20 },
		{ 320, 240, 20 },
		{ 320, 241, 20 },
		{ 321, 210, 20 },
		{ 321, 211, 20 },
		{ 321, 212, 20 },
		{ 321, 213, 20 },
		{ 321, 214, 20 },
		{ 321, 215, 20 },
		{ 321, 216, 20 },
		{ 321, 217, 20 },
		{ 321, 218, 20 },
		{ 321, 219, 20 },
		{ 321, 220, 20 },
		{ 321, 221, 20 },
		{ 321, 222, 20 },
		{ 321, 223, 20 },
		{ 321, 224, 20 },
		{ 321, 225, 20 },
		{ 321, 226, 20 },
		{ 321, 227, 20 },
		{ 321, 228, 20 },
		{ 321, 229, 20 },
		{ 321, 230, 20 },
		{ 321, 231, 20 },
		{ 321, 232, 20 },
		{ 321, 233, 20 },
		{ 321, 234, 20 },
		{ 321, 235, 20 },
		{ 321, 236, 20 },
		{ 321, 237, 20 },
		{ 321, 238, 20 },
		{ 321, 239, 20 },
		{ 321, 240, 20 },
		{ 322, 209, 20 },
		{ 322, 210, 20 },
		{ 322, 211, 20 },
		{ 322, 212, 20 },
		{ 322, 213, 20 },
		{ 322, 214, 20 },
		{ 322, 215, 20 },
		{ 322, 216, 20 },
		{ 322, 217, 20 },
		{ 322, 218, 20 },
		{ 322, 219, 20 },
		{ 322, 220, 20 },
		{ 322, 221, 20 },
		{ 322, 222, 20 },
		{ 322, 223, 20 },
		{ 322, 224, 20 },
		{ 322, 225, 20 },
		{ 322, 226, 20 },
		{ 322, 227, 20 },
		{ 322, 228, 20 },
		{ 322, 229, 20 },
		{ 322, 230, 20 },
		{ 322, 231, 20 },
		{ 322, 232, 20 },
		{ 322, 233, 20 },
		{ 322, 234, 20 },
		{ 322, 235, 20 },
		{ 322, 236, 20 },
		{ 322, 237, 20 },
		{ 322, 238, 20 },
		{ 322, 239, 20 },
		{ 322, 240, 20 },
		{ 323, 208, 20 },
		{ 323, 209, 20 },
		{ 323, 210, 20 },
		{ 323, 211, 20 },
		{ 323, 212, 20 },
		{ 323, 213, 20 },
		{ 323, 214, 20 },
		{ 323, 215, 20 },
		{ 323, 216, 20 },
		{ 323, 217, 20 },
		{ 323, 218, 20 },
		{ 323, 219, 20 },
		{ 323, 220, 20 },
		{ 323, 221, 20 },
		{ 323, 222, 20 },
		{ 323, 223, 20 },
		{ 323, 224, 20 },
		{ 323, 225, 20 },
		{ 323, 226, 20 },
		{ 323, 227, 20 },
		{ 323, 228, 20 },
		{ 323, 229, 20 },
		{ 323, 230, 20 },
		{ 323, 231, 20 },
		{ 323, 232, 20 },
		{ 323, 233, 20 },
		{ 323, 234, 20 },
		{ 323, 235, 20 },
		{ 323, 236, 20 },
		{ 323, 237, 20 },
		{ 323, 238, 20 },
		{ 323, 239, 20 },
		{ 323, 240, 20 },
		{ 324, 207, 20 },
		{ 324, 208, 20 },
		{ 324, 209, 20 },
		{ 324, 210, 20 },
		{ 324, 211, 20 },
		{ 324, 212, 20 },
		{ 324, 213, 20 },
		{ 324, 214, 20 },
		{ 324, 215, 20 },
		{ 324, 216, 20 },
		{ 324, 217, 20 },
		{ 324, 218, 20 },
		{ 324, 219, 20 },
		{ 324, 220, 20 },
		{ 324, 221, 20 },
		{ 324, 222, 20 },
		{ 324, 223, 20 },
		{ 324, 224, 20 },
		{ 324, 225, 20 },
		{ 324, 226, 20 },
		{ 324, 227, 20 },
		{ 324, 228, 20 },
		{ 324, 229, 20 },
		{ 324, 230, 20 },
		{ 324, 231, 20 },
		{ 324, 232, 20 },
		{ 324, 233, 20 },
		{ 324, 234, 20 },
		{ 324, 235, 20 },
		{ 324, 236, 20 },
		{ 324, 237, 20 },
		{ 324, 238, 20 },
		{ 324, 239, 20 },
		{ 324, 240, 20 },
		{ 325, 207, 20 },
		{ 325, 208, 20 },
		{ 325, 209, 20 },
		{ 325, 210, 20 },
		{ 325, 211, 20 },
		{ 325, 212, 20 },
		{ 325, 213, 20 },
		{ 325, 214, 20 },
		{ 325, 215, 20 },
		{ 325, 216, 20 },
		{ 325, 217, 20 },
		{ 325, 218, 20 },
		{ 325, 219, 20 },
		{ 325, 220, 20 },
		{ 325, 221, 20 },
		{ 325, 222, 20 },
		{ 325, 223, 20 },
		{ 325, 224, 20 },
		{ 325, 225, 20 },
		{ 325, 226, 20 },
		{ 325, 227, 20 },
		{ 325, 228, 20 },
		{ 325, 229, 20 },
		{ 325, 230, 20 },
		{ 325, 231, 20 },
		{ 325, 232, 20 },
		{ 325, 233, 20 },
		{ 325, 234, 20 },
		{ 325, 235, 20 },
		{ 325, 236, 20 },
		{ 325, 237, 20 },
		{ 325, 238, 20 },
		{ 325, 239, 20 },
		{ 326, 206, 20 },
		{ 326, 207, 20 },
		{ 326, 208, 20 },
		{ 326, 209, 20 },
		{ 326, 210, 20 },
		{ 326, 211, 20 },
		{ 326, 212, 20 },
		{ 326, 213, 20 },
		{ 326, 214, 20 },
		{ 326, 215, 20 },
		{ 326, 216, 20 },
		{ 326, 217, 20 },
		{ 326, 218, 20 },
		{ 326, 219, 20 },
		{ 326, 220, 20 },
		{ 326, 221, 20 },
		{ 326, 222, 20 },
		{ 326, 223, 20 },
		{ 326, 224, 20 },
		{ 326, 225, 20 },
		{ 326, 226, 20 },
		{ 326, 227, 20 },
		{ 326, 228, 20 },
		{ 326, 229, 20 },
		{ 326, 230, 20 },
		{ 326, 231, 20 },
		{ 326, 232, 20 },
		{ 326, 233, 20 },
		{ 326, 234, 20 },
		{ 326, 235, 20 },
		{ 326, 236, 20 },
		{ 326, 237, 20 },
		{ 326, 238, 20 },
		{ 326, 239, 20 },
		{ 327, 205, 20 },
		{ 327, 206, 20 },
		{ 327, 207, 20 },
		{ 327, 208, 20 },
		{ 327, 209, 20 },
		{ 327, 210, 20 },
		{ 327, 211, 20 },
		{ 327, 212, 20 },
		{ 327, 213, 20 },
		{ 327, 214, 20 },
		{ 327, 215, 20 },
		{ 327, 216, 20 },
		{ 327, 217, 20 },
		{ 327, 218, 20 },
		{ 327, 219, 20 },
		{ 327, 220, 20 },
		{ 327, 221, 20 },
		{ 327, 222, 20 },
		{ 327, 223, 20 },
		{ 327, 224, 20 },
		{ 327, 225, 20 },
		{ 327, 226, 20 },
		{ 327, 227, 20 },
		{ 327, 228, 20 },
		{ 327, 229, 20 },
		{ 327, 230, 20 },
		{ 327, 231, 20 },
		{ 327, 232, 20 },
		{ 327, 233, 20 },
		{ 327, 234, 20 },
		{ 327, 235, 20 },
		{ 327, 236, 20 },
		{ 327, 237, 20 },
		{ 327, 238, 20 },
		{ 327, 239, 20 },
		{ 328, 204, 20 },
		{ 328, 205, 20 },
		{ 328, 206, 20 },
		{ 328, 207, 20 },
		{ 328, 208, 20 },
		{ 328, 209, 20 },
		{ 328, 210, 20 },
		{ 328, 211, 20 },
		{ 328, 212, 20 },
		{ 328, 213, 20 },
		{ 328, 214, 20 },
		{ 328, 215, 20 },
		{ 328, 216, 20 },
		{ 328, 217, 20 },
		{ 328, 218, 20 },
		{ 328, 219, 20 },
		{ 328, 220, 20 },
		{ 328, 221, 20 },
		{ 328, 222, 20 },
		{ 328, 223, 20 },
		{ 328, 224, 20 },
		{ 328, 225, 20 },
		{ 328, 226, 20 },
		{ 328, 227, 20 },
		{ 328, 228, 20 },
		{ 328, 229, 20 },
		{ 328, 230, 20 },
		{ 328, 231, 20 },
		{ 328, 232, 20 },
		{ 328, 233, 20 },
		{ 328, 234, 20 },
		{ 328, 235, 20 },
		{ 328, 236, 20 },
		{ 328, 237, 20 },
		{ 328, 238, 20 },
		{ 328, 239, 20 },
		{ 329, 204, 20 },
		{ 329, 205, 20 },
		{ 329, 206, 20 },
		{ 329, 207, 20 },
		{ 329, 208, 20 },
		{ 329, 209, 20 },
		{ 329, 210, 20 },
		{ 329, 211, 20 },
		{ 329, 212, 20 },
		{ 329, 213, 20 },
		{ 329, 214, 20 },
		{ 329, 215, 20 },
		{ 329, 216, 20 },
		{ 329, 217, 20 },
		{ 329, 218, 20 },
		{ 329, 219, 20 },
		{ 329, 220, 20 },
		{ 329, 221, 20 },
		{ 329, 222, 20 },
		{ 329, 223, 20 },
		{ 329, 224, 20 },
		{ 329, 225, 20 },
		{ 329, 226, 20 },
		{ 329, 227, 20 },
		{ 329, 228, 20 },
		{ 329, 229, 20 },
		{ 329, 230, 20 },
		{ 329, 231, 20 },
		{ 329, 232, 20 },
		{ 329, 233, 20 },
		{ 329, 234, 20 },
		{ 329, 235, 20 },
		{ 329, 236, 20 },
		{ 329, 237, 20 },
		{ 329, 238, 20 },
		{ 330, 203, 20 },
		{ 330, 204, 20 },
		{ 330, 205, 20 },
		{ 330, 206, 20 },
		{ 330, 207, 20 },
		{ 330, 208, 20 },
		{ 330, 209, 20 },
		{ 330, 210, 20 },
		{ 330, 211, 20 },
		{ 330, 212, 20 },
		{ 330, 213, 20 },
		{ 330, 214, 20 },
		{ 330, 215, 20 },
		{ 330, 216, 20 },
		{ 330, 217, 20 },
		{ 330, 218, 20 },
		{ 330, 219, 20 },
		{ 330, 220, 20 },
		{ 330, 221, 20 },
		{ 330, 222, 20 },
		{ 330, 223, 20 },
		{ 330, 224, 20 },
		{ 330, 225, 20 },
		{ 330, 226, 20 },
		{ 330, 227, 20 },
		{ 330, 228, 20 },
		{ 330, 229, 20 },
		{ 330, 230, 20 },
		{ 330, 231, 20 },
		{ 330, 232, 20 },
		{ 330, 233, 20 },
		{ 330, 234, 20 },
		{ 330, 235, 20 },
		{ 330, 236, 20 },
		{ 330, 237, 20 },
		{ 330, 238, 20 },
		{ 331, 202, 20 },
		{ 331, 203, 20 },
		{ 331, 204, 20 },
		{ 331, 205, 20 },
		{ 331, 206, 20 },
		{ 331, 207, 20 },
		{ 331, 208, 20 },
		{ 331, 209, 20 },
		{ 331, 210, 20 },
		{ 331, 211, 20 },
		{ 331, 212, 20 },
		{ 331, 213, 20 },
		{ 331, 214, 20 },
		{ 331, 215, 20 },
		{ 331, 216, 20 },
		{ 331, 217, 20 },
		{ 331, 218, 20 },
		{ 331, 219, 20 },
		{ 331, 220, 20 },
		{ 331, 221, 20 },
		{ 331, 222, 20 },
		{ 331, 223, 20 },
		{ 331, 224, 20 },
		{ 331, 225, 20 },
		{ 331, 226, 20 },
		{ 331, 227, 20 },
		{ 331, 228, 20 },
		{ 331, 229, 20 },
		{ 331, 230, 20 },
		{ 331, 231, 20 },
		{ 331, 232, 20 },
		{ 331, 233, 20 },
		{ 331, 234, 20 },
		{ 331, 235, 20 },
		{ 331, 236, 20 },
		{ 331, 237, 20 },
		{ 331, 238, 20 },
		{ 332, 201, 20 },
		{ 332, 202, 20 },
		{ 332, 203, 20 },
		{ 332, 204, 20 },
		{ 332, 205, 20 },
		{ 332, 206, 20 },
		{ 332, 207, 20 },
		{ 332, 208, 20 },
		{ 332, 209, 20 },
		{ 332, 210, 20 },
		{ 332, 211, 20 },
		{ 332, 212, 20 },
		{ 332, 213, 20 },
		{ 332, 214, 20 },
		{ 332, 215, 20 },
		{ 332, 216, 20 },
		{ 332, 217, 20 },
		{ 332, 218, 20 },
		{ 332, 219, 20 },
		{ 332, 220, 20 },
		{ 332, 221, 20 },
		{ 332, 222, 20 },
		{ 332, 223, 20 },
		{ 332, 224, 20 },
		{ 332, 225, 20 },
		{ 332, 226, 20 },
		{ 332, 227, 20 },
		{ 332, 228, 20 },
		{ 332, 229, 20 },
		{ 332, 230, 20 },
		{ 332, 231, 20 },
		{ 332, 232, 20 },
		{ 332, 233, 20 },
		{ 332, 234, 20 },
		{ 332, 235, 20 },
		{ 332, 236, 20 },
		{ 332, 237, 20 },
		{ 332, 238, 20 },
		{ 333, 201, 20 },
		{ 333, 202, 20 },
		{ 333, 203, 20 },
		{ 333, 204, 20 },
		{ 333, 205, 20 },
		{ 333, 206, 20 },
		{ 333, 207, 20 },
		{ 333, 208, 20 },
		{ 333, 209, 20 },
		{ 333, 210, 20 },
		{ 333, 211, 20 },
		{ 333, 212, 20 },
		{ 333, 213, 20 },
		{ 333, 214, 20 },
		{ 333, 215, 20 },
		{ 333, 216, 20 },
		{ 333, 217, 20 },
		{ 333, 218, 20 },
		{ 333, 219, 20 },
		{ 333, 220, 20 },
		{ 333, 221, 20 },
		{ 333, 222, 20 },
		{ 333, 223, 20 },
		{ 333, 224, 20 },
		{ 333, 225, 20 },
		{ 333, 226, 20 },
		{ 333, 227, 20 },
		{ 333, 228, 20 },
		{ 333, 229, 20 },
		{ 333, 230, 20 },
		{ 333, 231, 20 },
		{ 333, 232, 20 },
		{ 333, 233, 20 },
		{ 333, 234, 20 },
		{ 333, 235, 20 },
		{ 333, 236, 20 },
		{ 333, 237, 20 },
		{ 334, 200, 20 },
		{ 334, 201, 20 },
		{ 334, 202, 20 },
		{ 334, 203, 20 },
		{ 334, 204, 20 },
		{ 334, 205, 20 },
		{ 334, 206, 20 },
		{ 334, 207, 20 },
		{ 334, 208, 20 },
		{ 334, 209, 20 },
		{ 334, 210, 20 },
		{ 334, 211, 20 },
		{ 334, 212, 20 },
		{ 334, 213, 20 },
		{ 334, 214, 20 },
		{ 334, 215, 20 },
		{ 334, 216, 20 },
		{ 334, 217, 20 },
		{ 334, 218, 20 },
		{ 334, 219, 20 },
		{ 334, 220, 20 },
		{ 334, 221, 20 },
		{ 334, 222, 20 },
		{ 334, 223, 20 },
		{ 334, 224, 20 },
		{ 334, 225, 20 },
		{ 334, 226, 20 },
		{ 334, 227, 20 },
		{ 334, 228, 20 },
		{ 334, 229, 20 },
		{ 334, 230, 20 },
		{ 334, 231, 20 },
		{ 334, 232, 20 },
		{ 334, 233, 20 },
		{ 334, 234, 20 },
		{ 334, 235, 20 },
		{ 334, 236, 20 },
		{ 334, 237, 20 },
		{ 335, 199, 20 },
		{ 335, 200, 20 },
		{ 335, 201, 20 },
		{ 335, 202, 20 },
		{ 335, 203, 20 },
		{ 335, 204, 20 },
		{ 335, 205, 20 },
		{ 335, 206, 20 },
		{ 335, 207, 20 },
		{ 335, 208, 20 },
		{ 335, 209, 20 },
		{ 335, 210, 20 },
		{ 335, 211, 20 },
		{ 335, 212, 20 },
		{ 335, 213, 20 },
		{ 335, 214, 20 },
		{ 335, 215, 20 },
		{ 335, 216, 20 },
		{ 335, 217, 20 },
		{ 335, 218, 20 },
		{ 335, 219, 20 },
		{ 335, 220, 20 },
		{ 335, 221, 20 },
		{ 335, 222, 20 },
		{ 335, 223, 20 },
		{ 335, 224, 20 },
		{ 335, 225, 20 },
		{ 335, 226, 20 },
		{ 335, 227, 20 },
		{ 335, 228, 20 },
		{ 335, 229, 20 },
		{ 335, 230, 20 },
		{ 335, 231, 20 },
		{ 335, 232, 20 },
		{ 335, 233, 20 },
		{ 335, 234, 20 },
		{ 335, 235, 20 },
		{ 335, 236, 20 },
		{ 335, 237, 20 },
		{ 336, 198, 20 },
		{ 336, 199, 20 },
		{ 336, 200, 20 },
		{ 336, 201, 20 },
		{ 336, 202, 20 },
		{ 336, 203, 20 },
		{ 336, 204, 20 },
		{ 336, 205, 20 },
		{ 336, 206, 20 },
		{ 336, 207, 20 },
		{ 336, 208, 20 },
		{ 336, 209, 20 },
		{ 336, 210, 20 },
		{ 336, 211, 20 },
		{ 336, 212, 20 },
		{ 336, 213, 20 },
		{ 336, 214, 20 },
		{ 336, 215, 20 },
		{ 336, 216, 20 },
		{ 336, 217, 20 },
		{ 336, 218, 20 },
		{ 336, 219, 20 },
		{ 336, 220, 20 },
		{ 336, 221, 20 },
		{ 336, 222, 20 },
		{ 336, 223, 20 },
		{ 336, 224, 20 },
		{ 336, 225, 20 },
		{ 336, 226, 20 },
		{ 336, 227, 20 },
		{ 336, 228, 20 },
		{ 336, 229, 20 },
		{ 336, 230, 20 },
		{ 336, 231, 20 },
		{ 336, 232, 20 },
		{ 336, 233, 20 },
		{ 336, 234, 20 },
		{ 336, 235, 20 },
		{ 336, 236, 20 },
		{ 336, 237, 20 },
		{ 337, 198, 20 },
		{ 337, 199, 20 },
		{ 337, 200, 20 },
		{ 337, 201, 20 },
		{ 337, 202, 20 },
		{ 337, 203, 20 },
		{ 337, 204, 20 },
		{ 337, 205, 20 },
		{ 337, 206, 20 },
		{ 337, 207, 20 },
		{ 337, 208, 20 },
		{ 337, 209, 20 },
		{ 337, 210, 20 },
		{ 337, 211, 20 },
		{ 337, 212, 20 },
		{ 337, 213, 20 },
		{ 337, 214, 20 },
		{ 337, 215, 20 },
		{ 337, 216, 20 },
		{ 337, 217, 20 },
		{ 337, 218, 20 },
		{ 337, 219, 20 },
		{ 337, 220, 20 },
		{ 337, 221, 20 },
		{ 337, 222, 20 },
		{ 337, 223, 20 },
		{ 337, 224, 20 },
		{ 337, 225, 20 },
		{ 337, 226, 20 },
		{ 337, 227, 20 },
		{ 337, 228, 20 },
		{ 337, 229, 20 },
		{ 337, 230, 20 },
		{ 337, 231, 20 },
		{ 337, 232, 20 },
		{ 337, 233, 20 },
		{ 337, 234, 20 },
		{ 337, 235, 20 },
		{ 337, 236, 20 },
		{ 338, 197, 20 },
		{ 338, 198, 20 },
		{ 338, 199, 20 },
		{ 338, 200, 20 },
		{ 338, 201, 20 },
		{ 338, 202, 20 },
		{ 338, 203, 20 },
		{ 338, 204, 20 },
		{ 338, 205, 20 },
		{ 338, 206, 20 },
		{ 338, 207, 20 },
		{ 338, 208, 20 },
		{ 338, 209, 20 },
		{ 338, 210, 20 },
		{ 338, 211, 20 },
		{ 338, 212, 20 },
		{ 338, 213, 20 },
		{ 338, 214, 20 },
		{ 338, 215, 20 },
		{ 338, 216, 20 },
		{ 338, 217, 20 },
		{ 338, 218, 20 },
		{ 338, 219, 20 },
		{ 338, 220, 20 },
		{ 338, 221, 20 },
		{ 338, 222, 20 },
		{ 338, 223, 20 },
		{ 338, 224, 20 },
		{ 338, 225, 20 },
		{ 338, 226, 20 },
		{ 338, 227, 20 },
		{ 338, 228, 20 },
		{ 338, 229, 20 },
		{ 338, 230, 20 },
		{ 338, 231, 20 },
		{ 338, 232, 20 },
		{ 338, 233, 20 },
		{ 338, 234, 20 },
		{ 338, 235, 20 },
		{ 338, 236, 20 },
		{ 339, 196, 20 },
		{ 339, 197, 20 },
		{ 339, 198, 20 },
		{ 339, 199, 20 },
		{ 339, 200, 20 },
		{ 339, 201, 20 },
		{ 339, 202, 20 },
		{ 339, 203, 20 },
		{ 339, 204, 20 },
		{ 339, 205, 20 },
		{ 339, 206, 20 },
		{ 339, 207, 20 },
		{ 339, 208, 20 },
		{ 339, 209, 20 },
		{ 339, 210, 20 },
		{ 339, 211, 20 },
		{ 339, 212, 20 },
		{ 339, 213, 20 },
		{ 339, 214, 20 },
		{ 339, 215, 20 },
		{ 339, 216, 20 },
		{ 339, 217, 20 },
		{ 339, 218, 20 },
		{ 339, 219, 20 },
		{ 339, 220, 20 },
		{ 339, 221, 20 },
		{ 339, 222, 20 },
		{ 339, 223, 20 },
		{ 339, 224, 20 },
		{ 339, 225, 20 },
		{ 339, 226, 20 },
		{ 339, 227, 20 },
		{ 339, 228, 20 },
		{ 339, 229, 20 },
		{ 339, 230, 20 },
		{ 339, 231, 20 },
		{ 339, 232, 20 },
		{ 339, 233, 20 },
		{ 339, 234, 20 },
		{ 339, 235, 20 },
		{ 339, 236, 20 },
		{ 340, 195, 20 },
		{ 340, 196, 20 },
		{ 340, 197, 20 },
		{ 340, 198, 20 },
		{ 340, 199, 20 },
		{ 340, 200, 20 },
		{ 340, 201, 20 },
		{ 340, 202, 20 },
		{ 340, 203, 20 },
		{ 340, 204, 20 },
		{ 340, 205, 20 },
		{ 340, 206, 20 },
		{ 340, 207, 20 },
		{ 340, 208, 20 },
		{ 340, 209, 20 },
		{ 340, 210, 20 },
		{ 340, 211, 20 },
		{ 340, 212, 20 },
		{ 340, 213, 20 },
		{ 340, 214, 20 },
		{ 340, 215, 20 },
		{ 340, 216, 20 },
		{ 340, 217, 20 },
		{ 340, 218, 20 },
		{ 340, 219, 20 },
		{ 340, 220, 20 },
		{ 340, 221, 20 },
		{ 340, 222, 20 },
		{ 340, 223, 20 },
		{ 340, 224, 20 },
		{ 340, 225, 20 },
		{ 340, 226, 20 },
		{ 340, 227, 20 },
		{ 340, 228, 20 },
		{ 340, 229, 20 },
		{ 340, 230, 20 },
		{ 340, 231, 20 },
		{ 340, 232, 20 },
		{ 340, 233, 20 },
		{ 340, 234, 20 },
		{ 340, 235, 20 },
		{ 340, 236, 20 },
		{ 341, 195, 20 },
		{ 341, 196, 20 },
		{ 341, 197, 20 },
		{ 341, 198, 20 },
		{ 341, 199, 20 },
		{ 341, 200, 20 },
		{ 341, 201, 20 },
		{ 341, 202, 20 },
		{ 341, 203, 20 },
		{ 341, 204, 20 },
		{ 341, 205, 20 },
		{ 341, 206, 20 },
		{ 341, 207, 20 },
		{ 341, 208, 20 },
		{ 341, 209, 20 },
		{ 341, 210, 20 },
		{ 341, 211, 20 },
		{ 341, 212, 20 },
		{ 341, 213, 20 },
		{ 341, 214, 20 },
		{ 341, 215, 20 },
		{ 341, 216, 20 },
		{ 341, 217, 20 },
		{ 341, 218, 20 },
		{ 341, 219, 20 },
		{ 341, 220, 20 },
		{ 341, 221, 20 },
		{ 341, 222, 20 },
		{ 341, 223, 20 },
		{ 341, 224, 20 },
		{ 341, 225, 20 },
		{ 341, 226, 20 },
		{ 341, 227, 20 },
		{ 341, 228, 20 },
		{ 341, 229, 20 },
		{ 341, 230, 20 },
		{ 341, 231, 20 },
		{ 341, 232, 20 },
		{ 341, 233, 20 },
		{ 341, 234, 20 },
		{ 341, 235, 20 },
		{ 342, 194, 20 },
		{ 342, 195, 20 },
		{ 342, 196, 20 },
		{ 342, 197, 20 },
		{ 342, 198, 20 },
		{ 342, 199, 20 },
		{ 342, 200, 20 },
		{ 342, 201, 20 },
		{ 342, 202, 20 },
		{ 342, 203, 20 },
		{ 342, 204, 20 },
		{ 342, 205, 20 },
		{ 342, 206, 20 },
		{ 342, 207, 20 },
		{ 342, 208, 20 },
		{ 342, 209, 20 },
		{ 342, 210, 20 },
		{ 342, 211, 20 },
		{ 342, 212, 20 },
		{ 342, 213, 20 },
		{ 342, 214, 20 },
		{ 342, 215, 20 },
		{ 342, 216, 20 },
		{ 342, 217, 20 },
		{ 342, 218, 20 },
		{ 342, 219, 20 },
		{ 342, 220, 20 },
		{ 342, 221, 20 },
		{ 342, 222, 20 },
		{ 342, 223, 20 },
		{ 342, 224, 20 },
		{ 342, 225, 20 },
		{ 342, 226, 20 },
		{ 342, 227, 20 },
		{ 342, 228, 20 },
		{ 342, 229, 20 },
		{ 342, 230, 20 },
		{ 342, 231, 20 },
		{ 342, 232, 20 },
		{ 342, 233, 20 },
		{ 342, 234, 20 },
		{ 342, 235, 20 },
		{ 343, 193, 20 },
		{ 343, 194, 20 },
		{ 343, 195, 20 },
		{ 343, 196, 20 },
		{ 343, 197, 20 },
		{ 343, 198, 20 },
		{ 343, 199, 20 },
		{ 343, 200, 20 },
		{ 343, 201, 20 },
		{ 343, 202, 20 },
		{ 343, 203, 20 },
		{ 343, 204, 20 },
		{ 343, 205, 20 },
		{ 343, 206, 20 },
		{ 343, 207, 20 },
		{ 343, 208, 20 },
		{ 343, 209, 20 },
		{ 343, 210, 20 },
		{ 343, 211, 20 },
		{ 343, 212, 20 },
		{ 343, 213, 20 },
		{ 343, 214, 20 },
		{ 343, 215, 20 },
		{ 343, 216, 20 },
		{ 343, 217, 20 },
		{ 343, 218, 20 },
		{ 343, 219, 20 },
		{ 343, 220, 20 },
		{ 343, 221, 20 },
		{ 343, 222, 20 },
		{ 343, 223, 20 },
		{ 343, 224, 20 },
		{ 343, 225, 20 },
		{ 343, 226, 20 },
		{ 343, 227, 20 },
		{ 343, 228, 20 },
		{ 343, 229, 20 },
		{ 343, 230, 20 },
		{ 343, 231, 20 },
		{ 343, 232, 20 },
		{ 343, 233, 20 },
		{ 343, 234, 20 },
		{ 343, 235, 20 },
		{ 344, 192, 20 },
		{ 344, 193, 20 },
		{ 344, 194, 20 },
		{ 344, 195, 20 },
		{ 344, 196, 20 },
		{ 344, 197, 20 },
		{ 344, 198, 20 },
		{ 344, 199, 20 },
		{ 344, 200, 20 },
		{ 344, 201, 20 },
		{ 344, 202, 20 },
		{ 344, 203, 20 },
		{ 344, 204, 20 },
		{ 344, 205, 20 },
		{ 344, 206, 20 },
		{ 344, 207, 20 },
		{ 344, 208, 20 },
		{ 344, 209, 20 },
		{ 344, 210, 20 },
		{ 344, 211, 20 },
		{ 344, 212, 20 },
		{ 344, 213, 20 },
		{ 344, 214, 20 },
		{ 344, 215, 20 },
		{ 344, 216, 20 },
		{ 344, 217, 20 },
		{ 344, 218, 20 },
		{ 344, 219, 20 },
		{ 344, 220, 20 },
		{ 344, 221, 20 },
		{ 344, 222, 20 },
		{ 344, 223, 20 },
		{ 344, 224, 20 },
		{ 344, 225, 20 },
		{ 344, 226, 20 },
		{ 344, 227, 20 },
		{ 344, 228, 20 },
		{ 344, 229, 20 },
		{ 344, 230, 20 },
		{ 344, 231, 20 },
		{ 344, 232, 20 },
		{ 344, 233, 20 },
		{ 344, 234, 20 },
		{ 344, 235, 20 },
		{ 345, 192, 20 },
		{ 345, 193, 20 },
		{ 345, 194, 20 },
		{ 345, 195, 20 },
		{ 345, 196, 20 },
		{ 345, 197, 20 },
		{ 345, 198, 20 },
		{ 345, 199, 20 },
		{ 345, 200, 20 },
		{ 345, 201, 20 },
		{ 345, 202, 20 },
		{ 345, 203, 20 },
		{ 345, 204, 20 },
		{ 345, 205, 20 },
		{ 345, 206, 20 },
		{ 345, 207, 20 },
		{ 345, 208, 20 },
		{ 345, 209, 20 },
		{ 345, 210, 20 },
		{ 345, 211, 20 },
		{ 345, 212, 20 },
		{ 345, 213, 20 },
		{ 345, 214, 20 },
		{ 345, 215, 20 },
		{ 345, 216, 20 },
		{ 345, 217, 20 },
		{ 345, 218, 20 },
		{ 345, 219, 20 },
		{ 345, 220, 20 },
		{ 345, 221, 20 },
		{ 345, 222, 20 },
		{ 345, 223, 20 },
		{ 345, 224, 20 },
		{ 345, 225, 20 },
		{ 345, 226, 20 },
		{ 345, 227, 20 },
		{ 345, 228, 20 },
		{ 345, 229, 20 },
		{ 345, 230, 20 },
		{ 345, 231, 20 },
		{ 345, 232, 20 },
		{ 345, 233, 20 },
		{ 345, 234, 20 },
		{ 346, 191, 20 },
		{ 346, 192, 20 },
		{ 346, 193, 20 },
		{ 346, 194, 20 },
		{ 346, 195, 20 },
		{ 346, 196, 20 },
		{ 346, 197, 20 },
		{ 346, 198, 20 },
		{ 346, 199, 20 },
		{ 346, 200, 20 },
		{ 346, 201, 20 },
		{ 346, 202, 20 },
		{ 346, 203, 20 },
		{ 346, 204, 20 },
		{ 346, 205, 20 },
		{ 346, 206, 20 },
		{ 346, 207, 20 },
		{ 346, 208, 20 },
		{ 346, 209, 20 },
		{ 346, 210, 20 },
		{ 346, 211, 20 },
		{ 346, 212, 20 },
		{ 346, 213, 20 },
		{ 346, 214, 20 },
		{ 346, 215, 20 },
		{ 346, 216, 20 },
		{ 346, 217, 20 },
		{ 346, 218, 20 },
		{ 346, 219, 20 },
		{ 346, 220, 20 },
		{ 346, 221, 20 },
		{ 346, 222, 20 },
		{ 346, 223, 20 },
		{ 346, 224, 20 },
		{ 346, 225, 20 },
		{ 346, 226, 20 },
		{ 346, 227, 20 },
		{ 346, 228, 20 },
		{ 346, 229, 20 },
		{ 346, 230, 20 },
		{ 346, 231, 20 },
		{ 346, 232, 20 },
		{ 346, 233, 20 },
		{ 346, 234, 20 },
		{ 347, 190, 20 },
		{ 347, 191, 20 },
		{ 347, 192, 20 },
		{ 347, 193, 20 },
		{ 347, 194, 20 },
		{ 347, 195, 20 },
		{ 347, 196, 20 },
		{ 347, 197, 20 },
		{ 347, 198, 20 },
		{ 347, 199, 20 },
		{ 347, 200, 20 },
		{ 347, 201, 20 },
		{ 347, 202, 20 },
		{ 347, 203, 20 },
		{ 347, 204, 20 },
		{ 347, 205, 20 },
		{ 347, 206, 20 },
		{ 347, 207, 20 },
		{ 347, 208, 20 },
		{ 347, 209, 20 },
		{ 347, 210, 20 },
		{ 347, 211, 20 },
		{ 347, 212, 20 },
		{ 347, 213, 20 },
		{ 347, 214, 20 },
		{ 347, 215, 20 },
		{ 347, 216, 20 },
		{ 347, 217, 20 },
		{ 347, 218, 20 },
		{ 347, 219, 20 },
		{ 347, 220, 20 },
		{ 347, 221, 20 },
		{ 347, 222, 20 },
		{ 347, 223, 20 },
		{ 347, 224, 20 },
		{ 347, 225, 20 },
		{ 347, 226, 20 },
		{ 347, 227, 20 },
		{ 347, 228, 20 },
		{ 347, 229, 20 },
		{ 347, 230, 20 },
		{ 347, 231, 20 },
		{ 347, 232, 20 },
		{ 347, 233, 20 },
		{ 347, 234, 20 },
		{ 348, 189, 20 },
		{ 348, 190, 20 },
		{ 348, 191, 20 },
		{ 348, 192, 20 },
		{ 348, 193, 20 },
		{ 348, 194, 20 },
		{ 348, 195, 20 },
		{ 348, 196, 20 },
		{ 348, 197, 20 },
		{ 348, 198, 20 },
		{ 348, 199, 20 },
		{ 348, 200, 20 },
		{ 348, 201, 20 },
		{ 348, 202, 20 },
		{ 348, 203, 20 },
		{ 348, 204, 20 },
		{ 348, 205, 20 },
		{ 348, 206, 20 },
		{ 348, 207, 20 },
		{ 348, 208, 20 },
		{ 348, 209, 20 },
		{ 348, 210, 20 },
		{ 348, 211, 20 },
		{ 348, 212, 20 },
		{ 348, 213, 20 },
		{ 348, 214, 20 },
		{ 348, 215, 20 },
		{ 348, 216, 20 },
		{ 348, 217, 20 },
		{ 348, 218, 20 },
		{ 348, 219, 20 },
		{ 348, 220, 20 },
		{ 348, 221, 20 },
		{ 348, 222, 20 },
		{ 348, 223, 20 },
		{ 348, 224, 20 },
		{ 348, 225, 20 },
		{ 348, 226, 20 },
		{ 348, 227, 20 },
		{ 348, 228, 20 },
		{ 348, 229, 20 },
		{ 348, 230, 20 },
		{ 348, 231, 20 },
		{ 348, 232, 20 },
		{ 348, 233, 20 },
		{ 348, 234, 20 },
		{ 349, 189, 20 },
		{ 349, 190, 20 },
		{ 349, 191, 20 },
		{ 349, 192, 20 },
		{ 349, 193, 20 },
		{ 349, 194, 20 },
		{ 349, 195, 20 },
		{ 349, 196, 20 },
		{ 349, 197, 20 },
		{ 349, 198, 20 },
		{ 349, 199, 20 },
		{ 349, 200, 20 },
		{ 349, 201, 20 },
		{ 349, 202, 20 },
		{ 349, 203, 20 },
		{ 349, 204, 20 },
		{ 349, 205, 20 },
		{ 349, 206, 20 },
		{ 349, 207, 20 },
		{ 349, 208, 20 },
		{ 349, 209, 20 },
		{ 349, 210, 20 },
		{ 349, 211, 20 },
		{ 349, 212, 20 },
		{ 349, 213, 20 },
		{ 349, 214, 20 },
		{ 349, 215, 20 },
		{ 349, 216, 20 },
		{ 349, 217, 20 },
		{ 349, 218, 20 },
		{ 349, 219, 20 },
		{ 349, 220, 20 },
		{ 349, 221, 20 },
		{ 349, 222, 20 },
		{ 349, 223, 20 },
		{ 349, 224, 20 },
		{ 349, 225, 20 },
		{ 349, 226, 20 },
		{ 349, 227, 20 },
		{ 349, 228, 20 },
		{ 349, 229, 20 },
		{ 349, 230, 20 },
		{ 349, 231, 20 },
		{ 349, 232, 20 },
		{ 349, 233, 20 },
		{ 350, 188, 20 },
		{ 350, 189, 20 },
		{ 350, 190, 20 },
		{ 350, 191, 20 },
		{ 350, 192, 20 },
		{ 350, 193, 20 },
		{ 350, 194, 20 },
		{ 350, 195, 20 },
		{ 350, 196, 20 },
		{ 350, 197, 20 },
		{ 350, 198, 20 },
		{ 350, 199, 20 },
		{ 350, 200, 20 },
		{ 350, 201, 20 },
		{ 350, 202, 20 },
		{ 350, 203, 20 },
		{ 350, 204, 20 },
		{ 350, 205, 20 },
		{ 350, 206, 20 },
		{ 350, 207, 20 },
		{ 350, 208, 20 },
		{ 350, 209, 20 },
		{ 350, 210, 20 },
		{ 350, 211, 20 },
		{ 350, 212, 20 },
		{ 350, 213, 20 },
		{ 350, 214, 20 },
		{ 350, 215, 20 },
		{ 350, 216, 20 },
		{ 350, 217, 20 },
		{ 350, 218, 20 },
		{ 350, 219, 20 },
		{ 350, 220, 20 },
		{ 350, 221, 20 },
		{ 350, 222, 20 },
		{ 350, 223, 20 },
		{ 350, 224, 20 },
		{ 350, 225, 20 },
		{ 350, 226, 20 },
		{ 350, 227, 20 },
		{ 350, 228, 20 },
		{ 350, 229, 20 },
		{ 350, 230, 20 },
		{ 350, 231, 20 },
		{ 350, 232, 20 },
		{ 350, 233, 20 },
		{ 351, 187, 20 },
		{ 351, 188, 20 },
		{ 351, 189, 20 },
		{ 351, 190, 20 },
		{ 351, 191, 20 },
		{ 351, 192, 20 },
		{ 351, 193, 20 },
		{ 351, 194, 20 },
		{ 351, 195, 20 },
		{ 351, 196, 20 },
		{ 351, 197, 20 },
		{ 351, 198, 20 },
		{ 351, 199, 20 },
		{ 351, 200, 20 },
		{ 351, 201, 20 },
		{ 351, 202, 20 },
		{ 351, 203, 20 },
		{ 351, 204, 20 },
		{ 351, 205, 20 },
		{ 351, 206, 20 },
		{ 351, 207, 20 },
		{ 351, 208, 20 },
		{ 351, 209, 20 },
		{ 351, 210, 20 },
		{ 351, 211, 20 },
		{ 351, 212, 20 },
		{ 351, 213, 20 },
		{ 351, 214, 20 },
		{ 351, 215, 20 },
		{ 351, 216, 20 },
		{ 351, 217, 20 },
		{ 351, 218, 20 },
		{ 351, 219, 20 },
		{ 351, 220, 20 },
		{ 351, 221, 20 },
		{ 351, 222, 20 },
		{ 351, 223, 20 },
		{ 351, 224, 20 },
		{ 351, 225, 20 },
		{ 351, 226, 20 },
		{ 351, 227, 20 },
		{ 351, 228, 20 },
		{ 351, 229, 20 },
		{ 351, 230, 20 },
		{ 351, 231, 20 },
		{ 351, 232, 20 },
		{ 351, 233, 20 },
		{ 352, 186, 20 },
		{ 352, 187, 20 },
		{ 352, 188, 20 },
		{ 352, 189, 20 },
		{ 352, 190, 20 },
		{ 352, 191, 20 },
		{ 352, 192, 20 },
		{ 352, 193, 20 },
		{ 352, 194, 20 },
		{ 352, 195, 20 },
		{ 352, 196, 20 },
		{ 352, 197, 20 },
		{ 352, 198, 20 },
		{ 352, 199, 20 },
		{ 352, 200, 20 },
		{ 352, 201, 20 },
		{ 352, 202, 20 },
		{ 352, 203, 20 },
		{ 352, 204, 20 },
		{ 352, 205, 20 },
		{ 352, 206, 20 },
		{ 352, 207, 20 },
		{ 352, 208, 20 },
		{ 352, 209, 20 },
		{ 352, 210, 20 },
		{ 352, 211, 20 },
		{ 352, 212, 20 },
		{ 352, 213, 20 },
		{ 352, 214, 20 },
		{ 352, 215, 20 },
		{ 352, 216, 20 },
		{ 352, 217, 20 },
		{ 352, 218, 20 },
		{ 352, 219, 20 },
		{ 352, 220, 20 },
		{ 352, 221, 20 },
		{ 352, 222, 20 },
		{ 352, 223, 20 },
		{ 352, 224, 20 },
		{ 352, 225, 20 },
		{ 352, 226, 20 },
		{ 352, 227, 20 },
		{ 352, 228, 20 },
		{ 352, 229, 20 },
		{ 352, 230, 20 },
		{ 352, 231, 20 },
		{ 352, 232, 20 },
		{ 352, 233, 20 },
		{ 353, 186, 20 },
		{ 353, 187, 20 },
		{ 353, 188, 20 },
		{ 353, 189, 20 },
		{ 353, 190, 20 },
		{ 353, 191, 20 },
		{ 353, 192, 20 },
		{ 353, 193, 20 },
		{ 353, 194, 20 },
		{ 353, 195, 20 },
		{ 353, 196, 20 },
		{ 353, 197, 20 },
		{ 353, 198, 20 },
		{ 353, 199, 20 },
		{ 353, 200, 20 },
		{ 353, 201, 20 },
		{ 353, 202, 20 },
		{ 353, 203, 20 },
		{ 353, 204, 20 },
		{ 353, 205, 20 },
		{ 353, 206, 20 },
		{ 353, 207, 20 },
		{ 353, 208, 20 },
		{ 353, 209, 20 },
		{ 353, 210, 20 },
		{ 353, 211, 20 },
		{ 353, 212, 20 },
		{ 353, 213, 20 },
		{ 353, 214, 20 },
		{ 353, 215, 20 },
		{ 353, 216, 20 },
		{ 353, 217, 20 },
		{ 353, 218, 20 },
		{ 353, 219, 20 },
		{ 353, 220, 20 },
		{ 353, 221, 20 },
		{ 353, 222, 20 },
		{ 353, 223, 20 },
		{ 353, 224, 20 },
		{ 353, 225, 20 },
		{ 353, 226, 20 },
		{ 353, 227, 20 },
		{ 353, 228, 20 },
		{ 353, 229, 20 },
		{ 353, 230, 20 },
		{ 353, 231, 20 },
		{ 353, 232, 20 },
		{ 354, 185, 20 },
		{ 354, 186, 20 },
		{ 354, 187, 20 },
		{ 354, 188, 20 },
		{ 354, 189, 20 },
		{ 354, 190, 20 },
		{ 354, 191, 20 },
		{ 354, 192, 20 },
		{ 354, 193, 20 },
		{ 354, 194, 20 },
		{ 354, 195, 20 },
		{ 354, 196, 20 },
		{ 354, 197, 20 },
		{ 354, 198, 20 },
		{ 354, 199, 20 },
		{ 354, 200, 20 },
		{ 354, 201, 20 },
		{ 354, 202, 20 },
		{ 354, 203, 20 },
		{ 354, 204, 20 },
		{ 354, 205, 20 },
		{ 354, 206, 20 },
		{ 354, 207, 20 },
		{ 354, 208, 20 },
		{ 354, 209, 20 },
		{ 354, 210, 20 },
		{ 354, 211, 20 },
		{ 354, 212, 20 },
		{ 354, 213, 20 },
		{ 354, 214, 20 },
		{ 354, 215, 20 },
		{ 354, 216, 20 },
		{ 354, 217, 20 },
		{ 354, 218, 20 },
		{ 354, 219, 20 },
		{ 354, 220, 20 },
		{ 354, 221, 20 },
		{ 354, 222, 20 },
		{ 354, 223, 20 },
		{ 354, 224, 20 },
		{ 354, 225, 20 },
		{ 354, 226, 20 },
		{ 354, 227, 20 },
		{ 354, 228, 20 },
		{ 354, 229, 20 },
		{ 354, 230, 20 },
		{ 354, 231, 20 },
		{ 354, 232, 20 },
		{ 355, 184, 20 },
		{ 355, 185, 20 },
		{ 355, 186, 20 },
		{ 355, 187, 20 },
		{ 355, 188, 20 },
		{ 355, 189, 20 },
		{ 355, 190, 20 },
		{ 355, 191, 20 },
		{ 355, 192, 20 },
		{ 355, 193, 20 },
		{ 355, 194, 20 },
		{ 355, 195, 20 },
		{ 355, 196, 20 },
		{ 355, 197, 20 },
		{ 355, 198, 20 },
		{ 355, 199, 20 },
		{ 355, 200, 20 },
		{ 355, 201, 20 },
		{ 355, 202, 20 },
		{ 355, 203, 20 },
		{ 355, 204, 20 },
		{ 355, 205, 20 },
		{ 355, 206, 20 },
		{ 355, 207, 20 },
		{ 355, 208, 20 },
		{ 355, 209, 20 },
		{ 355, 210, 20 },
		{ 355, 211, 20 },
		{ 355, 212, 20 },
		{ 355, 213, 20 },
		{ 355, 214, 20 },
		{ 355, 215, 20 },
		{ 355, 216, 20 },
		{ 355, 217, 20 },
		{ 355, 218, 20 },
		{ 355, 219, 20 },
		{ 355, 220, 20 },
		{ 355, 221, 20 },
		{ 355, 222, 20 },
		{ 355, 223, 20 },
		{ 355, 224, 20 },
		{ 355, 225, 20 },
		{ 355, 226, 20 },
		{ 355, 227, 20 },
		{ 355, 228, 20 },
		{ 355, 229, 20 },
		{ 355, 230, 20 },
		{ 355, 231, 20 },
		{ 355, 232, 20 },
		{ 356, 183, 20 },
		{ 356, 184, 20 },
		{ 356, 185, 20 },
		{ 356, 186, 20 },
		{ 356, 187, 20 },
		{ 356, 188, 20 },
		{ 356, 189, 20 },
		{ 356, 190, 20 },
		{ 356, 191, 20 },
		{ 356, 192, 20 },
		{ 356, 193, 20 },
		{ 356, 194, 20 },
		{ 356, 195, 20 },
		{ 356, 196, 20 },
		{ 356, 197, 20 },
		{ 356, 198, 20 },
		{ 356, 199, 20 },
		{ 356, 200, 20 },
		{ 356, 201, 20 },
		{ 356, 202, 20 },
		{ 356, 203, 20 },
		{ 356, 204, 20 },
		{ 356, 205, 20 },
		{ 356, 206, 20 },
		{ 356, 207, 20 },
		{ 356, 208, 20 },
		{ 356, 209, 20 },
		{ 356, 210, 20 },
		{ 356, 211, 20 },
		{ 356, 212, 20 },
		{ 356, 213, 20 },
		{ 356, 214, 20 },
		{ 356, 215, 20 },
		{ 356, 216, 20 },
		{ 356, 217, 20 },
		{ 356, 218, 20 },
		{ 356, 219, 20 },
		{ 356, 220, 20 },
		{ 356, 221, 20 },
		{ 356, 222, 20 },
		{ 356, 223, 20 },
		{ 356, 224, 20 },
		{ 356, 225, 20 },
		{ 356, 226, 20 },
		{ 356, 227, 20 },
		{ 356, 228, 20 },
		{ 356, 229, 20 },
		{ 356, 230, 20 },
		{ 356, 231, 20 },
		{ 356, 232, 20 },
		{ 357, 183, 20 },
		{ 357, 184, 20 },
		{ 357, 185, 20 },
		{ 357, 186, 20 },
		{ 357, 187, 20 },
		{ 357, 188, 20 },
		{ 357, 189, 20 },
		{ 357, 190, 20 },
		{ 357, 191, 20 },
		{ 357, 192, 20 },
		{ 357, 193, 20 },
		{ 357, 194, 20 },
		{ 357, 195, 20 },
		{ 357, 196, 20 },
		{ 357, 197, 20 },
		{ 357, 198, 20 },
		{ 357, 199, 20 },
		{ 357, 200, 20 },
		{ 357, 201, 20 },
		{ 357, 202, 20 },
		{ 357, 203, 20 },
		{ 357, 204, 20 },
		{ 357, 205, 20 },
		{ 357, 206, 20 },
		{ 357, 207, 20 },
		{ 357, 208, 20 },
		{ 357, 209, 20 },
		{ 357, 210, 20 },
		{ 357, 211, 20 },
		{ 357, 212, 20 },
		{ 357, 213, 20 },
		{ 357, 214, 20 },
		{ 357, 215, 20 },
		{ 357, 216, 20 },
		{ 357, 217, 20 },
		{ 357, 218, 20 },
		{ 357, 219, 20 },
		{ 357, 220, 20 },
		{ 357, 221, 20 },
		{ 357, 222, 20 },
		{ 357, 223, 20 },
		{ 357, 224, 20 },
		{ 357, 225, 20 },
		{ 357, 226, 20 },
		{ 357, 227, 20 },
		{ 357, 228, 20 },
		{ 357, 229, 20 },
		{ 357, 230, 20 },
		{ 357, 231, 20 },
		{ 358, 182, 20 },
		{ 358, 183, 20 },
		{ 358, 184, 20 },
		{ 358, 185, 20 },
		{ 358, 186, 20 },
		{ 358, 187, 20 },
		{ 358, 188, 20 },
		{ 358, 189, 20 },
		{ 358, 190, 20 },
		{ 358, 191, 20 },
		{ 358, 192, 20 },
		{ 358, 193, 20 },
		{ 358, 194, 20 },
		{ 358, 195, 20 },
		{ 358, 196, 20 },
		{ 358, 197, 20 },
		{ 358, 198, 20 },
		{ 358, 199, 20 },
		{ 358, 200, 20 },
		{ 358, 201, 20 },
		{ 358, 202, 20 },
		{ 358, 203, 20 },
		{ 358, 204, 20 },
		{ 358, 205, 20 },
		{ 358, 206, 20 },
		{ 358, 207, 20 },
		{ 358, 208, 20 },
		{ 358, 209, 20 },
		{ 358, 210, 20 },
		{ 358, 211, 20 },
		{ 358, 212, 20 },
		{ 358, 213, 20 },
		{ 358, 214, 20 },
		{ 358, 215, 20 },
		{ 358, 216, 20 },
		{ 358, 217, 20 },
		{ 358, 218, 20 },
		{ 358, 219, 20 },
		{ 358, 220, 20 },
		{ 358, 221, 20 },
		{ 358, 222, 20 },
		{ 358, 223, 20 },
		{ 358, 224, 20 },
		{ 358, 225, 20 },
		{ 358, 226, 20 },
		{ 358, 227, 20 },
		{ 358, 228, 20 },
		{ 358, 229, 20 },
		{ 358, 230, 20 },
		{ 358, 231, 20 },
		{ 359, 181, 20 },
		{ 359, 182, 20 },
		{ 359, 183, 20 },
		{ 359, 184, 20 },
		{ 359, 185, 20 },
		{ 359, 186, 20 },
		{ 359, 187, 20 },
		{ 359, 188, 20 },
		{ 359, 189, 20 },
		{ 359, 190, 20 },
		{ 359, 191, 20 },
		{ 359, 192, 20 },
		{ 359, 193, 20 },
		{ 359, 194, 20 },
		{ 359, 195, 20 },
		{ 359, 196, 20 },
		{ 359, 197, 20 },
		{ 359, 198, 20 },
		{ 359, 199, 20 },
		{ 359, 200, 20 },
		{ 359, 201, 20 },
		{ 359, 202, 20 },
		{ 359, 203, 20 },
		{ 359, 204, 20 },
		{ 359, 205, 20 },
		{ 359, 206, 20 },
		{ 359, 207, 20 },
		{ 359, 208, 20 },
		{ 359, 209, 20 },
		{ 359, 210, 20 },
		{ 359, 211, 20 },
		{ 359, 212, 20 },
		{ 359, 213, 20 },
		{ 359, 214, 20 },
		{ 359, 215, 20 },
		{ 359, 216, 20 },
		{ 359, 217, 20 },
		{ 359, 218, 20 },
		{ 359, 219, 20 },
		{ 359, 220, 20 },
		{ 359, 221, 20 },
		{ 359, 222, 20 },
		{ 359, 223, 20 },
		{ 359, 224, 20 },
		{ 359, 225, 20 },
		{ 359, 226, 20 },
		{ 359, 227, 20 },
		{ 359, 228, 20 },
		{ 359, 229, 20 },
		{ 359, 230, 20 },
		{ 359, 231, 20 },
		{ 360, 180, 20 },
		{ 360, 181, 20 },
		{ 360, 182, 20 },
		{ 360, 183, 20 },
		{ 360, 184, 20 },
		{ 360, 185, 20 },
		{ 360, 186, 20 },
		{ 360, 187, 20 },
		{ 360, 188, 20 },
		{ 360, 189, 20 },
		{ 360, 190, 20 },
		{ 360, 191, 20 },
		{ 360, 192, 20 },
		{ 360, 193, 20 },
		{ 360, 194, 20 },
		{ 360, 195, 20 },
		{ 360, 196, 20 },
		{ 360, 197, 20 },
		{ 360, 198, 20 },
		{ 360, 199, 20 },
		{ 360, 200, 20 },
		{ 360, 201, 20 },
		{ 360, 202, 20 },
		{ 360, 203, 20 },
		{ 360, 204, 20 },
		{ 360, 205, 20 },
		{ 360, 206, 20 },
		{ 360, 207, 20 },
		{ 360, 208, 20 },
		{ 360, 209, 20 },
		{ 360, 210, 20 },
		{ 360, 211, 20 },
		{ 360, 212, 20 },
		{ 360, 213, 20 },
		{ 360, 214, 20 },
		{ 360, 215, 20 },
		{ 360, 216, 20 },
		{ 360, 217, 20 },
		{ 360, 218, 20 },
		{ 360, 219, 20 },
		{ 360, 220, 20 },
		{ 360, 221, 20 },
		{ 360, 222, 20 },
		{ 360, 223, 20 },
		{ 360, 224, 20 },
		{ 360, 225, 20 },
		{ 360, 226, 20 },
		{ 360, 227, 20 },
		{ 360, 228, 20 },
		{ 360, 229, 20 },
		{ 360, 230, 20 },
		{ 360, 231, 20 },
		{ 361, 180, 20 },
		{ 361, 181, 20 },
		{ 361, 182, 20 },
		{ 361, 183, 20 },
		{ 361, 184, 20 },
		{ 361, 185, 20 },
		{ 361, 186, 20 },
		{ 361, 187, 20 },
		{ 361, 188, 20 },
		{ 361, 189, 20 },
		{ 361, 190, 20 },
		{ 361, 191, 20 },
		{ 361, 192, 20 },
		{ 361, 193, 20 },
		{ 361, 194, 20 },
		{ 361, 195, 20 },
		{ 361, 196, 20 },
		{ 361, 197, 20 },
		{ 361, 198, 20 },
		{ 361, 199, 20 },
		{ 361, 200, 20 },
		{ 361, 201, 20 },
		{ 361, 202, 20 },
		{ 361, 203, 20 },
		{ 361, 204, 20 },
		{ 361, 205, 20 },
		{ 361, 206, 20 },
		{ 361, 207, 20 },
		{ 361, 208, 20 },
		{ 361, 209, 20 },
		{ 361, 210, 20 },
		{ 361, 211, 20 },
		{ 361, 212, 20 },
		{ 361, 213, 20 },
		{ 361, 214, 20 },
		{ 361, 215, 20 },
		{ 361, 216, 20 },
		{ 361, 217, 20 },
		{ 361, 218, 20 },
		{ 361, 219, 20 },
		{ 361, 220, 20 },
		{ 361, 221, 20 },
		{ 361, 222, 20 },
		{ 361, 223, 20 },
		{ 361, 224, 20 },
		{ 361, 225, 20 },
		{ 361, 226, 20 },
		{ 361, 227, 20 },
		{ 361, 228, 20 },
		{ 361, 229, 20 },
		{ 361, 230, 20 },
		{ 362, 179, 20 },
		{ 362, 180, 20 },
		{ 362, 181, 20 },
		{ 362, 182, 20 },
		{ 362, 183, 20 },
		{ 362, 184, 20 },
		{ 362, 185, 20 },
		{ 362, 186, 20 },
		{ 362, 187, 20 },
		{ 362, 188, 20 },
		{ 362, 189, 20 },
		{ 362, 190, 20 },
		{ 362, 191, 20 },
		{ 362, 192, 20 },
		{ 362, 193, 20 },
		{ 362, 194, 20 },
		{ 362, 195, 20 },
		{ 362, 196, 20 },
		{ 362, 197, 20 },
		{ 362, 198, 20 },
		{ 362, 199, 20 },
		{ 362, 200, 20 },
		{ 362, 201, 20 },
		{ 362, 202, 20 },
		{ 362, 203, 20 },
		{ 362, 204, 20 },
		{ 362, 205, 20 },
		{ 362, 206, 20 },
		{ 362, 207, 20 },
		{ 362, 208, 20 },
		{ 362, 209, 20 },
		{ 362, 210, 20 },
		{ 362, 211, 20 },
		{ 362, 212, 20 },
		{ 362, 213, 20 },
		{ 362, 214, 20 },
		{ 362, 215, 20 },
		{ 362, 216, 20 },
		{ 362, 217, 20 },
		{ 362, 218, 20 },
		{ 362, 219, 20 },
		{ 362, 220, 20 },
		{ 362, 221, 20 },
		{ 362, 222, 20 },
		{ 362, 223, 20 },
		{ 362, 224, 20 },
		{ 362, 225, 20 },
		{ 362, 226, 20 },
		{ 362, 227, 20 },
		{ 362, 228, 20 },
		{ 362, 229, 20 },
		{ 362, 230, 20 },
		{ 363, 178, 20 },
		{ 363, 179, 20 },
		{ 363, 180, 20 },
		{ 363, 181, 20 },
		{ 363, 182, 20 },
		{ 363, 183, 20 },
		{ 363, 184, 20 },
		{ 363, 185, 20 },
		{ 363, 186, 20 },
		{ 363, 187, 20 },
		{ 363, 188, 20 },
		{ 363, 189, 20 },
		{ 363, 190, 20 },
		{ 363, 191, 20 },
		{ 363, 192, 20 },
		{ 363, 193, 20 },
		{ 363, 194, 20 },
		{ 363, 195, 20 },
		{ 363, 196, 20 },
		{ 363, 197, 20 },
		{ 363, 198, 20 },
		{ 363, 199, 20 },
		{ 363, 200, 20 },
		{ 363, 201, 20 },
		{ 363, 202, 20 },
		{ 363, 203, 20 },
		{ 363, 204, 20 },
		{ 363, 205, 20 },
		{ 363, 206, 20 },
		{ 363, 207, 20 },
		{ 363, 208, 20 },
		{ 363, 209, 20 },
		{ 363, 210, 20 },
		{ 363, 211, 20 },
		{ 363, 212, 20 },
		{ 363, 213, 20 },
		{ 363, 214, 20 },
		{ 363, 215, 20 },
		{ 363, 216, 20 },
		{ 363, 217, 20 },
		{ 363, 218, 20 },
		{ 363, 219, 20 },
		{ 363, 220, 20 },
		{ 363, 221, 20 },
		{ 363, 222, 20 },
		{ 363, 223, 20 },
		{ 363, 224, 20 },
		{ 363, 225, 20 },
		{ 363, 226, 20 },
		{ 363, 227, 20 },
		{ 363, 228, 20 },
		{ 363, 229, 20 },
		{ 363, 230, 20 },
		{ 364, 177, 20 },
		{ 364, 178, 20 },
		{ 364, 179, 20 },
		{ 364, 180, 20 },
		{ 364, 181, 20 },
		{ 364, 182, 20 },
		{ 364, 183, 20 },
		{ 364, 184, 20 },
		{ 364, 185, 20 },
		{ 364, 186, 20 },
		{ 364, 187, 20 },
		{ 364, 188, 20 },
		{ 364, 189, 20 },
		{ 364, 190, 20 },
		{ 364, 191, 20 },
		{ 364, 192, 20 },
		{ 364, 193, 20 },
		{ 364, 194, 20 },
		{ 364, 195, 20 },
		{ 364, 196, 20 },
		{ 364, 197, 20 },
		{ 364, 198, 20 },
		{ 364, 199, 20 },
		{ 364, 200, 20 },
		{ 364, 201, 20 },
		{ 364, 202, 20 },
		{ 364, 203, 20 },
		{ 364, 204, 20 },
		{ 364, 205, 20 },
		{ 364, 206, 20 },
		{ 364, 207, 20 },
		{ 364, 208, 20 },
		{ 364, 209, 20 },
		{ 364, 210, 20 },
		{ 364, 211, 20 },
		{ 364, 212, 20 },
		{ 364, 213, 20 },
		{ 364, 214, 20 },
		{ 364, 215, 20 },
		{ 364, 216, 20 },
		{ 364, 217, 20 },
		{ 364, 218, 20 },
		{ 364, 219, 20 },
		{ 364, 220, 20 },
		{ 364, 221, 20 },
		{ 364, 222, 20 },
		{ 364, 223, 20 },
		{ 364, 224, 20 },
		{ 364, 225, 20 },
		{ 364, 226, 20 },
		{ 364, 227, 20 },
		{ 364, 228, 20 },
		{ 364, 229, 20 },
		{ 364, 230, 20 },
		{ 365, 177, 20 },
		{ 365, 178, 20 },
		{ 365, 179, 20 },
		{ 365, 180, 20 },
		{ 365, 181, 20 },
		{ 365, 182, 20 },
		{ 365, 183, 20 },
		{ 365, 184, 20 },
		{ 365, 185, 20 },
		{ 365, 186, 20 },
		{ 365, 187, 20 },
		{ 365, 188, 20 },
		{ 365, 189, 20 },
		{ 365, 190, 20 },
		{ 365, 191, 20 },
		{ 365, 192, 20 },
		{ 365, 193, 20 },
		{ 365, 194, 20 },
		{ 365, 195, 20 },
		{ 365, 196, 20 },
		{ 365, 197, 20 },
		{ 365, 198, 20 },
		{ 365, 199, 20 },
		{ 365, 200, 20 },
		{ 365, 201, 20 },
		{ 365, 202, 20 },
		{ 365, 203, 20 },
		{ 365, 204, 20 },
		{ 365, 205, 20 },
		{ 365, 206, 20 },
		{ 365, 207, 20 },
		{ 365, 208, 20 },
		{ 365, 209, 20 },
		{ 365, 210, 20 },
		{ 365, 211, 20 },
		{ 365, 212, 20 },
		{ 365, 213, 20 },
		{ 365, 214, 20 },
		{ 365, 215, 20 },
		{ 365, 216, 20 },
		{ 365, 217, 20 },
		{ 365, 218, 20 },
		{ 365, 219, 20 },
		{ 365, 220, 20 },
		{ 365, 221, 20 },
		{ 365, 222, 20 },
		{ 365, 223, 20 },
		{ 365, 224, 20 },
		{ 365, 225, 20 },
		{ 365, 226, 20 },
		{ 365, 227, 20 },
		{ 365, 228, 20 },
		{ 365, 229, 20 },
		{ 366, 176, 20 },
		{ 366, 177, 20 },
		{ 366, 178, 20 },
		{ 366, 179, 20 },
		{ 366, 180, 20 },
		{ 366, 181, 20 },
		{ 366, 182, 20 },
		{ 366, 183, 20 },
		{ 366, 184, 20 },
		{ 366, 185, 20 },
		{ 366, 186, 20 },
		{ 366, 187, 20 },
		{ 366, 188, 20 },
		{ 366, 189, 20 },
		{ 366, 190, 20 },
		{ 366, 191, 20 },
		{ 366, 192, 20 },
		{ 366, 193, 20 },
		{ 366, 194, 20 },
		{ 366, 195, 20 },
		{ 366, 196, 20 },
		{ 366, 197, 20 },
		{ 366, 198, 20 },
		{ 366, 199, 20 },
		{ 366, 200, 20 },
		{ 366, 201, 20 },
		{ 366, 202, 20 },
		{ 366, 203, 20 },
		{ 366, 204, 20 },
		{ 366, 205, 20 },
		{ 366, 206, 20 },
		{ 366, 207, 20 },
		{ 366, 208, 20 },
		{ 366, 209, 20 },
		{ 366, 210, 20 },
		{ 366, 211, 20 },
		{ 366, 212, 20 },
		{ 366, 213, 20 },
		{ 366, 214, 20 },
		{ 366, 215, 20 },
		{ 366, 216, 20 },
		{ 366, 217, 20 },
		{ 366, 218, 20 },
		{ 366, 219, 20 },
		{ 366, 220, 20 },
		{ 366, 221, 20 },
		{ 366, 222, 20 },
		{ 366, 223, 20 },
		{ 366, 224, 20 },
		{ 366, 225, 20 },
		{ 366, 226, 20 },
		{ 366, 227, 20 },
		{ 366, 228, 20 },
		{ 366, 229, 20 },
		{ 367, 175, 20 },
		{ 367, 176, 20 },
		{ 367, 177, 20 },
		{ 367, 178, 20 },
		{ 367, 179, 20 },
		{ 367, 180, 20 },
		{ 367, 181, 20 },
		{ 367, 182, 20 },
		{ 367, 183, 20 },
		{ 367, 184, 20 },
		{ 367, 185, 20 },
		{ 367, 186, 20 },
		{ 367, 187, 20 },
		{ 367, 188, 20 },
		{ 367, 189, 20 },
		{ 367, 190, 20 },
		{ 367, 191, 20 },
		{ 367, 192, 20 },
		{ 367, 193, 20 },
		{ 367, 194, 20 },
		{ 367, 195, 20 },
		{ 367, 196, 20 },
		{ 367, 197, 20 },
		{ 367, 198, 20 },
		{ 367, 199, 20 },
		{ 367, 200, 20 },
		{ 367, 201, 20 },
		{ 367, 202, 20 },
		{ 367, 203, 20 },
		{ 367, 204, 20 },
		{ 367, 205, 20 },
		{ 367, 206, 20 },
		{ 367, 207, 20 },
		{ 367, 208, 20 },
		{ 367, 209, 20 },
		{ 367, 210, 20 },
		{ 367, 211, 20 },
		{ 367, 212, 20 },
		{ 367, 213, 20 },
		{ 367, 214, 20 },
		{ 367, 215, 20 },
		{ 367, 216, 20 },
		{ 367, 217, 20 },
		{ 367, 218, 20 },
		{ 367, 219, 20 },
		{ 367, 220, 20 },
		{ 367, 221, 20 },
		{ 367, 222, 20 },
		{ 367, 223, 20 },
		{ 367, 224, 20 },
		{ 367, 225, 20 },
		{ 367, 226, 20 },
		{ 367, 227, 20 },
		{ 367, 228, 20 },
		{ 367, 229, 20 },
		{ 368, 174, 20 },
		{ 368, 175, 20 },
		{ 368, 176, 20 },
		{ 368, 177, 20 },
		{ 368, 178, 20 },
		{ 368, 179, 20 },
		{ 368, 180, 20 },
		{ 368, 181, 20 },
		{ 368, 182, 20 },
		{ 368, 183, 20 },
		{ 368, 184, 20 },
		{ 368, 185, 20 },
		{ 368, 186, 20 },
		{ 368, 187, 20 },
		{ 368, 188, 20 },
		{ 368, 189, 20 },
		{ 368, 190, 20 },
		{ 368, 191, 20 },
		{ 368, 192, 20 },
		{ 368, 193, 20 },
		{ 368, 194, 20 },
		{ 368, 195, 20 },
		{ 368, 196, 20 },
		{ 368, 197, 20 },
		{ 368, 198, 20 },
		{ 368, 199, 20 },
		{ 368, 200, 20 },
		{ 368, 201, 20 },
		{ 368, 202, 20 },
		{ 368, 203, 20 },
		{ 368, 204, 20 },
		{ 368, 205, 20 },
		{ 368, 206, 20 },
		{ 368, 207, 20 },
		{ 368, 208, 20 },
		{ 368, 209, 20 },
		{ 368, 210, 20 },
		{ 368, 211, 20 },
		{ 368, 212, 20 },
		{ 368, 213, 20 },
		{ 368, 214, 20 },
		{ 368, 215, 20 },
		{ 368, 216, 20 },
		{ 368, 217, 20 },
		{ 368, 218, 20 },
		{ 368, 219, 20 },
		{ 368, 220, 20 },
		{ 368, 221, 20 },
		{ 368, 222, 20 },
		{ 368, 223, 20 },
		{ 368, 224, 20 },
		{ 368, 225, 20 },
		{ 368, 226, 20 },
		{ 368, 227, 20 },
		{ 368, 228, 20 },
		{ 368, 229, 20 },
		{ 369, 174, 20 },
		{ 369, 175, 20 },
		{ 369, 176, 20 },
		{ 369, 177, 20 },
		{ 369, 178, 20 },
		{ 369, 179, 20 },
		{ 369, 180, 20 },
		{ 369, 181, 20 },
		{ 369, 182, 20 },
		{ 369, 183, 20 },
		{ 369, 184, 20 },
		{ 369, 185, 20 },
		{ 369, 186, 20 },
		{ 369, 187, 20 },
		{ 369, 188, 20 },
		{ 369, 189, 20 },
		{ 369, 190, 20 },
		{ 369, 191, 20 },
		{ 369, 192, 20 },
		{ 369, 193, 20 },
		{ 369, 194, 20 },
		{ 369, 195, 20 },
		{ 369, 196, 20 },
		{ 369, 197, 20 },
		{ 369, 198, 20 },
		{ 369, 199, 20 },
		{ 369, 200, 20 },
		{ 369, 201, 20 },
		{ 369, 202, 20 },
		{ 369, 203, 20 },
		{ 369, 204, 20 },
		{ 369, 205, 20 },
		{ 369, 206, 20 },
		{ 369, 207, 20 },
		{ 369, 208, 20 },
		{ 369, 209, 20 },
		{ 369, 210, 20 },
		{ 369, 211, 20 },
		{ 369, 212, 20 },
		{ 369, 213, 20 },
		{ 369, 214, 20 },
		{ 369, 215, 20 },
		{ 369, 216, 20 },
		{ 369, 217, 20 },
		{ 369, 218, 20 },
		{ 369, 219, 20 },
		{ 369, 220, 20 },
		{ 369, 221, 20 },
		{ 369, 222, 20 },
		{ 369, 223, 20 },
		{ 369, 224, 20 },
		{ 369, 225, 20 },
		{ 369, 226, 20 },
		{ 369, 227, 20 },
		{ 369, 228, 20 },
		{ 370, 173, 20 },
		{ 370, 174, 20 },
		{ 370, 175, 20 },
		{ 370, 176, 20 },
		{ 370, 177, 20 },
		{ 370, 178, 20 },
		{ 370, 179, 20 },
		{ 370, 180, 20 },
		{ 370, 181, 20 },
		{ 370, 182, 20 },
		{ 370, 183, 20 },
		{ 370, 184, 20 },
		{ 370, 185, 20 },
		{ 370, 186, 20 },
		{ 370, 187, 20 },
		{ 370, 188, 20 },
		{ 370, 189, 20 },
		{ 370, 190, 20 },
		{ 370, 191, 20 },
		{ 370, 192, 20 },
		{ 370, 193, 20 },
		{ 370, 194, 20 },
		{ 370, 195, 20 },
		{ 370, 196, 20 },
		{ 370, 197, 20 },
		{ 370, 198, 20 },
		{ 370, 199, 20 },
		{ 370, 200, 20 },
		{ 370, 201, 20 },
		{ 370, 202, 20 },
		{ 370, 203, 20 },
		{ 370, 204, 20 },
		{ 370, 205, 20 },
		{ 370, 206, 20 },
		{ 370, 207, 20 },
		{ 370, 208, 20 },
		{ 370, 209, 20 },
		{ 370, 210, 20 },
		{ 370, 211, 20 },
		{ 370, 212, 20 },
		{ 370, 213, 20 },
		{ 370, 214, 20 },
		{ 370, 215, 20 },
		{ 370, 216, 20 },
		{ 370, 217, 20 },
		{ 370, 218, 20 },
		{ 370, 219, 20 },
		{ 370, 220, 20 },
		{ 370, 221, 20 },
		{ 370, 222, 20 },
		{ 370, 223, 20 },
		{ 370, 224, 20 },
		{ 370, 225, 20 },
		{ 370, 226, 20 },
		{ 370, 227, 20 },
		{ 370, 228, 20 },
		{ 371, 172, 20 },
		{ 371, 173, 20 },
		{ 371, 174, 20 },
		{ 371, 175, 20 },
		{ 371, 176, 20 },
		{ 371, 177, 20 },
		{ 371, 178, 20 },
		{ 371, 179, 20 },
		{ 371, 180, 20 },
		{ 371, 181, 20 },
		{ 371, 182, 20 },
		{ 371, 183, 20 },
		{ 371, 184, 20 },
		{ 371, 185, 20 },
		{ 371, 186, 20 },
		{ 371, 187, 20 },
		{ 371, 188, 20 },
		{ 371, 189, 20 },
		{ 371, 190, 20 },
		{ 371, 191, 20 },
		{ 371, 192, 20 },
		{ 371, 193, 20 },
		{ 371, 194, 20 },
		{ 371, 195, 20 },
		{ 371, 196, 20 },
		{ 371, 197, 20 },
		{ 371, 198, 20 },
		{ 371, 199, 20 },
		{ 371, 200, 20 },
		{ 371, 201, 20 },
		{ 371, 202, 20 },
		{ 371, 203, 20 },
		{ 371, 204, 20 },
		{ 371, 205, 20 },
		{ 371, 206, 20 },
		{ 371, 207, 20 },
		{ 371, 208, 20 },
		{ 371, 209, 20 },
		{ 371, 210, 20 },
		{ 371, 211, 20 },
		{ 371, 212, 20 },
		{ 371, 213, 20 },
		{ 371, 214, 20 },
		{ 371, 215, 20 },
		{ 371, 216, 20 },
		{ 371, 217, 20 },
		{ 371, 218, 20 },
		{ 371, 219, 20 },
		{ 371, 220, 20 },
		{ 371, 221, 20 },
		{ 371, 222, 20 },
		{ 371, 223, 20 },
		{ 371, 224, 20 },
		{ 371, 225, 20 },
		{ 371, 226, 20 },
		{ 371, 227, 20 },
		{ 371, 228, 20 },
		{ 372, 171, 20 },
		{ 372, 172, 20 },
		{ 372, 173, 20 },
		{ 372, 174, 20 },
		{ 372, 175, 20 },
		{ 372, 176, 20 },
		{ 372, 177, 20 },
		{ 372, 178, 20 },
		{ 372, 179, 20 },
		{ 372, 180, 20 },
		{ 372, 181, 20 },
		{ 372, 182, 20 },
		{ 372, 183, 20 },
		{ 372, 184, 20 },
		{ 372, 185, 20 },
		{ 372, 186, 20 },
		{ 372, 187, 20 },
		{ 372, 188, 20 },
		{ 372, 189, 20 },
		{ 372, 190, 20 },
		{ 372, 191, 20 },
		{ 372, 192, 20 },
		{ 372, 193, 20 },
		{ 372, 194, 20 },
		{ 372, 195, 20 },
		{ 372, 196, 20 },
		{ 372, 197, 20 },
		{ 372, 198, 20 },
		{ 372, 199, 20 },
		{ 372, 200, 20 },
		{ 372, 201, 20 },
		{ 372, 202, 20 },
		{ 372, 203, 20 },
		{ 372, 204, 20 },
		{ 372, 205, 20 },
		{ 372, 206, 20 },
		{ 372, 207, 20 },
		{ 372, 208, 20 },
		{ 372, 209, 20 },
		{ 372, 210, 20 },
		{ 372, 211, 20 },
		{ 372, 212, 20 },
		{ 372, 213, 20 },
		{ 372, 214, 20 },
		{ 372, 215, 20 },
		{ 372, 216, 20 },
		{ 372, 217, 20 },
		{ 372, 218, 20 },
		{ 372, 219, 20 },
		{ 372, 220, 20 },
		{ 372, 221, 20 },
		{ 372, 222, 20 },
		{ 372, 223, 20 },
		{ 372, 224, 20 },
		{ 372, 225, 20 },
		{ 372, 226, 20 },
		{ 372, 227, 20 },
		{ 372, 228, 20 },
		{ 373, 171, 20 },
		{ 373, 172, 20 },
		{ 373, 173, 20 },
		{ 373, 174, 20 },
		{ 373, 175, 20 },
		{ 373, 176, 20 },
		{ 373, 177, 20 },
		{ 373, 178, 20 },
		{ 373, 179, 20 },
		{ 373, 180, 20 },
		{ 373, 181, 20 },
		{ 373, 182, 20 },
		{ 373, 183, 20 },
		{ 373, 184, 20 },
		{ 373, 185, 20 },
		{ 373, 186, 20 },
		{ 373, 187, 20 },
		{ 373, 188, 20 },
		{ 373, 189, 20 },
		{ 373, 190, 20 },
		{ 373, 191, 20 },
		{ 373, 192, 20 },
		{ 373, 193, 20 },
		{ 373, 194, 20 },
		{ 373, 195, 20 },
		{ 373, 196, 20 },
		{ 373, 197, 20 },
		{ 373, 198, 20 },
		{ 373, 199, 20 },
		{ 373, 200, 20 },
		{ 373, 201, 20 },
		{ 373, 202, 20 },
		{ 373, 203, 20 },
		{ 373, 204, 20 },
		{ 373, 205, 20 },
		{ 373, 206, 20 },
		{ 373, 207, 20 },
		{ 373, 208, 20 },
		{ 373, 209, 20 },
		{ 373, 210, 20 },
		{ 373, 211, 20 },
		{ 373, 212, 20 },
		{ 373, 213, 20 },
		{ 373, 214, 20 },
		{ 373, 215, 20 },
		{ 373, 216, 20 },
		{ 373, 217, 20 },
		{ 373, 218, 20 },
		{ 373, 219, 20 },
		{ 373, 220, 20 },
		{ 373, 221, 20 },
		{ 373, 222, 20 },
		{ 373, 223, 20 },
		{ 373, 224, 20 },
		{ 373, 225, 20 },
		{ 373, 226, 20 },
		{ 373, 227, 20 },
		{ 374, 170, 20 },
		{ 374, 171, 20 },
		{ 374, 172, 20 },
		{ 374, 173, 20 },
		{ 374, 174, 20 },
		{ 374, 175, 20 },
		{ 374, 176, 20 },
		{ 374, 177, 20 },
		{ 374, 178, 20 },
		{ 374, 179, 20 },
		{ 374, 180, 20 },
		{ 374, 181, 20 },
		{ 374, 182, 20 },
		{ 374, 183, 20 },
		{ 374, 184, 20 },
		{ 374, 185, 20 },
		{ 374, 186, 20 },
		{ 374, 187, 20 },
		{ 374, 188, 20 },
		{ 374, 189, 20 },
		{ 374, 190, 20 },
		{ 374, 191, 20 },
		{ 374, 192, 20 },
		{ 374, 193, 20 },
		{ 374, 194, 20 },
		{ 374, 195, 20 },
		{ 374, 196, 20 },
		{ 374, 197, 20 },
		{ 374, 198, 20 },
		{ 374, 199, 20 },
		{ 374, 200, 20 },
		{ 374, 201, 20 },
		{ 374, 202, 20 },
		{ 374, 203, 20 },
		{ 374, 204, 20 },
		{ 374, 205, 20 },
		{ 374, 206, 20 },
		{ 374, 207, 20 },
		{ 374, 208, 20 },
		{ 374, 209, 20 },
		{ 374, 210, 20 },
		{ 374, 211, 20 },
		{ 374, 212, 20 },
		{ 374, 213, 20 },
		{ 374, 214, 20 },
		{ 374, 215, 20 },
		{ 374, 216, 20 },
		{ 374, 217, 20 },
		{ 374, 218, 20 },
		{ 374, 219, 20 },
		{ 374, 220, 20 },
		{ 374, 221, 20 },
		{ 374, 222, 20 },
		{ 374, 223, 20 },
		{ 374, 224, 20 },
		{ 374, 225, 20 },
		{ 374, 226, 20 },
		{ 374, 227, 20 },
		{ 375, 169, 20 },
		{ 375, 170, 20 },
		{ 375, 171, 20 },
		{ 375, 172, 20 },
		{ 375, 173, 20 },
		{ 375, 174, 20 },
		{ 375, 175, 20 },
		{ 375, 176, 20 },
		{ 375, 177, 20 },
		{ 375, 178, 20 },
		{ 375, 179, 20 },
		{ 375, 180, 20 },
		{ 375, 181, 20 },
		{ 375, 182, 20 },
		{ 375, 183, 20 },
		{ 375, 184, 20 },
		{ 375, 185, 20 },
		{ 375, 186, 20 },
		{ 375, 187, 20 },
		{ 375, 188, 20 },
		{ 375, 189, 20 },
		{ 375, 190, 20 },
		{ 375, 191, 20 },
		{ 375, 192, 20 },
		{ 375, 193, 20 },
		{ 375, 194, 20 },
		{ 375, 195, 20 },
		{ 375, 196, 20 },
		{ 375, 197, 20 },
		{ 375, 198, 20 },
		{ 375, 199, 20 },
		{ 375, 200, 20 },
		{ 375, 201, 20 },
		{ 375, 202, 20 },
		{ 375, 203, 20 },
		{ 375, 204, 20 },
		{ 375, 205, 20 },
		{ 375, 206, 20 },
		{ 375, 207, 20 },
		{ 375, 208, 20 },
		{ 375, 209, 20 },
		{ 375, 210, 20 },
		{ 375, 211, 20 },
		{ 375, 212, 20 },
		{ 375, 213, 20 },
		{ 375, 214, 20 },
		{ 375, 215, 20 },
		{ 375, 216, 20 },
		{ 375, 217, 20 },
		{ 375, 218, 20 },
		{ 375, 219, 20 },
		{ 375, 220, 20 },
		{ 375, 221, 20 },
		{ 375, 222, 20 },
		{ 375, 223, 20 },
		{ 375, 224, 20 },
		{ 375, 225, 20 },
		{ 375, 226, 20 },
		{ 375, 227, 20 },
		{ 376, 169, 20 },
		{ 376, 170, 20 },
		{ 376, 171, 20 },
		{ 376, 172, 20 },
		{ 376, 173, 20 },
		{ 376, 174, 20 },
		{ 376, 175, 20 },
		{ 376, 176, 20 },
		{ 376, 177, 20 },
		{ 376, 178, 20 },
		{ 376, 179, 20 },
		{ 376, 180, 20 },
		{ 376, 181, 20 },
		{ 376, 182, 20 },
		{ 376, 183, 20 },
		{ 376, 184, 20 },
		{ 376, 185, 20 },
		{ 376, 186, 20 },
		{ 376, 187, 20 },
		{ 376, 188, 20 },
		{ 376, 189, 20 },
		{ 376, 190, 20 },
		{ 376, 191, 20 },
		{ 376, 192, 20 },
		{ 376, 193, 20 },
		{ 376, 194, 20 },
		{ 376, 195, 20 },
		{ 376, 196, 20 },
		{ 376, 197, 20 },
		{ 376, 198, 20 },
		{ 376, 199, 20 },
		{ 376, 200, 20 },
		{ 376, 201, 20 },
		{ 376, 202, 20 },
		{ 376, 203, 20 },
		{ 376, 204, 20 },
		{ 376, 205, 20 },
		{ 376, 206, 20 },
		{ 376, 207, 20 },
		{ 376, 208, 20 },
		{ 376, 209, 20 },
		{ 376, 210, 20 },
		{ 376, 211, 20 },
		{ 376, 212, 20 },
		{ 376, 213, 20 },
		{ 376, 214, 20 },
		{ 376, 215, 20 },
		{ 376, 216, 20 },
		{ 376, 217, 20 },
		{ 376, 218, 20 },
		{ 376, 219, 20 },
		{ 376, 220, 20 },
		{ 376, 221, 20 },
		{ 376, 222, 20 },
		{ 376, 223, 20 },
		{ 376, 224, 20 },
		{ 376, 225, 20 },
		{ 376, 226, 20 },
		{ 377, 168, 20 },
		{ 377, 169, 20 },
		{ 377, 170, 20 },
		{ 377, 171, 20 },
		{ 377, 172, 20 },
		{ 377, 173, 20 },
		{ 377, 174, 20 },
		{ 377, 175, 20 },
		{ 377, 176, 20 },
		{ 377, 177, 20 },
		{ 377, 178, 20 },
		{ 377, 179, 20 },
		{ 377, 180, 20 },
		{ 377, 181, 20 },
		{ 377, 182, 20 },
		{ 377, 183, 20 },
		{ 377, 184, 20 },
		{ 377, 185, 20 },
		{ 377, 186, 20 },
		{ 377, 187, 20 },
		{ 377, 188, 20 },
		{ 377, 189, 20 },
		{ 377, 190, 20 },
		{ 377, 191, 20 },
		{ 377, 192, 20 },
		{ 377, 193, 20 },
		{ 377, 194, 20 },
		{ 377, 195, 20 },
		{ 377, 196, 20 },
		{ 377, 197, 20 },
		{ 377, 198, 20 },
		{ 377, 199, 20 },
		{ 377, 200, 20 },
		{ 377, 201, 20 },
		{ 377, 202, 20 },
		{ 377, 203, 20 },
		{ 377, 204, 20 },
		{ 377, 205, 20 },
		{ 377, 206, 20 },
		{ 377, 207, 20 },
		{ 377, 208, 20 },
		{ 377, 209, 20 },
		{ 377, 210, 20 },
		{ 377, 211, 20 },
		{ 377, 212, 20 },
		{ 377, 213, 20 },
		{ 377, 214, 20 },
		{ 377, 215, 20 },
		{ 377, 216, 20 },
		{ 377, 217, 20 },
		{ 377, 218, 20 },
		{ 377, 219, 20 },
		{ 377, 220, 20 },
		{ 377, 221, 20 },
		{ 377, 222, 20 },
		{ 377, 223, 20 },
		{ 377, 224, 20 },
		{ 377, 225, 20 },
		{ 377, 226, 20 },
		{ 378, 167, 20 },
		{ 378, 168, 20 },
		{ 378, 169, 20 },
		{ 378, 170, 20 },
		{ 378, 171, 20 },
		{ 378, 172, 20 },
		{ 378, 173, 20 },
		{ 378, 174, 20 },
		{ 378, 175, 20 },
		{ 378, 176, 20 },
		{ 378, 177, 20 },
		{ 378, 178, 20 },
		{ 378, 179, 20 },
		{ 378, 180, 20 },
		{ 378, 181, 20 },
		{ 378, 182, 20 },
		{ 378, 183, 20 },
		{ 378, 184, 20 },
		{ 378, 185, 20 },
		{ 378, 186, 20 },
		{ 378, 187, 20 },
		{ 378, 188, 20 },
		{ 378, 189, 20 },
		{ 378, 190, 20 },
		{ 378, 191, 20 },
		{ 378, 192, 20 },
		{ 378, 193, 20 },
		{ 378, 194, 20 },
		{ 378, 195, 20 },
		{ 378, 196, 20 },
		{ 378, 197, 20 },
		{ 378, 198, 20 },
		{ 378, 199, 20 },
		{ 378, 200, 20 },
		{ 378, 201, 20 },
		{ 378, 202, 20 },
		{ 378, 203, 20 },
		{ 378, 204, 20 },
		{ 378, 205, 20 },
		{ 378, 206, 20 },
		{ 378, 207, 20 },
		{ 378, 208, 20 },
		{ 378, 209, 20 },
		{ 378, 210, 20 },
		{ 378, 211, 20 },
		{ 378, 212, 20 },
		{ 378, 213, 20 },
		{ 378, 214, 20 },
		{ 378, 215, 20 },
		{ 378, 216, 20 },
		{ 378, 217, 20 },
		{ 378, 218, 20 },
		{ 378, 219, 20 },
		{ 378, 220, 20 },
		{ 378, 221, 20 },
		{ 378, 222, 20 },
		{ 378, 223, 20 },
		{ 378, 224, 20 },
		{ 378, 225, 20 },
		{ 378, 226, 20 },
		{ 379, 167, 20 },
		{ 379, 168, 20 },
		{ 379, 169, 20 },
		{ 379, 170, 20 },
		{ 379, 171, 20 },
		{ 379, 172, 20 },
		{ 379, 173, 20 },
		{ 379, 174, 20 },
		{ 379, 175, 20 },
		{ 379, 176, 20 },
		{ 379, 177, 20 },
		{ 379, 178, 20 },
		{ 379, 179, 20 },
		{ 379, 180, 20 },
		{ 379, 181, 20 },
		{ 379, 182, 20 },
		{ 379, 183, 20 },
		{ 379, 184, 20 },
		{ 379, 185, 20 },
		{ 379, 186, 20 },
		{ 379, 187, 20 },
		{ 379, 188, 20 },
		{ 379, 189, 20 },
		{ 379, 190, 20 },
		{ 379, 191, 20 },
		{ 379, 192, 20 },
		{ 379, 193, 20 },
		{ 379, 194, 20 },
		{ 379, 195, 20 },
		{ 379, 196, 20 },
		{ 379, 197, 20 },
		{ 379, 198, 20 },
		{ 379, 199, 20 },
		{ 379, 200, 20 },
		{ 379, 201, 20 },
		{ 379, 202, 20 },
		{ 379, 203, 20 },
		{ 379, 204, 20 },
		{ 379, 205, 20 },
		{ 379, 206, 20 },
		{ 379, 207, 20 },
		{ 379, 208, 20 },
		{ 379, 209, 20 },
		{ 379, 210, 20 },
		{ 379, 211, 20 },
		{ 379, 212, 20 },
		{ 379, 213, 20 },
		{ 379, 214, 20 },
		{ 379, 215, 20 },
		{ 379, 216, 20 },
		{ 379, 217, 20 },
		{ 379, 218, 20 },
		{ 379, 219, 20 },
		{ 379, 220, 20 },
		{ 379, 221, 20 },
		{ 379, 222, 20 },
		{ 379, 223, 20 },
		{ 379, 224, 20 },
		{ 379, 225, 20 },
		{ 380, 166, 20 },
		{ 380, 167, 20 },
		{ 380, 168, 20 },
		{ 380, 169, 20 },
		{ 380, 170, 20 },
		{ 380, 171, 20 },
		{ 380, 172, 20 },
		{ 380, 173, 20 },
		{ 380, 174, 20 },
		{ 380, 175, 20 },
		{ 380, 176, 20 },
		{ 380, 177, 20 },
		{ 380, 178, 20 },
		{ 380, 179, 20 },
		{ 380, 180, 20 },
		{ 380, 181, 20 },
		{ 380, 182, 20 },
		{ 380, 183, 20 },
		{ 380, 184, 20 },
		{ 380, 185, 20 },
		{ 380, 186, 20 },
		{ 380, 187, 20 },
		{ 380, 188, 20 },
		{ 380, 189, 20 },
		{ 380, 190, 20 },
		{ 380, 191, 20 },
		{ 380, 192, 20 },
		{ 380, 193, 20 },
		{ 380, 194, 20 },
		{ 380, 195, 20 },
		{ 380, 196, 20 },
		{ 380, 197, 20 },
		{ 380, 198, 20 },
		{ 380, 199, 20 },
		{ 380, 200, 20 },
		{ 380, 201, 20 },
		{ 380, 202, 20 },
		{ 380, 203, 20 },
		{ 380, 204, 20 },
		{ 380, 205, 20 },
		{ 380, 206, 20 },
		{ 380, 207, 20 },
		{ 380, 208, 20 },
		{ 380, 209, 20 },
		{ 380, 210, 20 },
		{ 380, 211, 20 },
		{ 380, 212, 20 },
		{ 380, 213, 20 },
		{ 380, 214, 20 },
		{ 380, 215, 20 },
		{ 380, 216, 20 },
		{ 380, 217, 20 },
		{ 380, 218, 20 },
		{ 380, 219, 20 },
		{ 380, 220, 20 },
		{ 380, 221, 20 },
		{ 380, 222, 20 },
		{ 380, 223, 20 },
		{ 380, 224, 20 },
		{ 380, 225, 20 },
		{ 381, 166, 20 },
		{ 381, 167, 20 },
		{ 381, 168, 20 },
		{ 381, 169, 20 },
		{ 381, 170, 20 },
		{ 381, 171, 20 },
		{ 381, 172, 20 },
		{ 381, 173, 20 },
		{ 381, 174, 20 },
		{ 381, 175, 20 },
		{ 381, 176, 20 },
		{ 381, 177, 20 },
		{ 381, 178, 20 },
		{ 381, 179, 20 },
		{ 381, 180, 20 },
		{ 381, 181, 20 },
		{ 381, 182, 20 },
		{ 381, 183, 20 },
		{ 381, 184, 20 },
		{ 381, 185, 20 },
		{ 381, 186, 20 },
		{ 381, 187, 20 },
		{ 381, 188, 20 },
		{ 381, 189, 20 },
		{ 381, 190, 20 },
		{ 381, 191, 20 },
		{ 381, 192, 20 },
		{ 381, 193, 20 },
		{ 381, 194, 20 },
		{ 381, 195, 20 },
		{ 381, 196, 20 },
		{ 381, 197, 20 },
		{ 381, 198, 20 },
		{ 381, 199, 20 },
		{ 381, 200, 20 },
		{ 381, 201, 20 },
		{ 381, 202, 20 },
		{ 381, 203, 20 },
		{ 381, 204, 20 },
		{ 381, 205, 20 },
		{ 381, 206, 20 },
		{ 381, 207, 20 },
		{ 381, 208, 20 },
		{ 381, 209, 20 },
		{ 381, 210, 20 },
		{ 381, 211, 20 },
		{ 381, 212, 20 },
		{ 381, 213, 20 },
		{ 381, 214, 20 },
		{ 381, 215, 20 },
		{ 381, 216, 20 },
		{ 381, 217, 20 },
		{ 381, 218, 20 },
		{ 381, 219, 20 },
		{ 381, 220, 20 },
		{ 381, 221, 20 },
		{ 381, 222, 20 },
		{ 381, 223, 20 },
		{ 381, 224, 20 },
		{ 382, 167, 20 },
		{ 382, 168, 20 },
		{ 382, 169, 20 },
		{ 382, 170, 20 },
		{ 382, 171, 20 },
		{ 382, 172, 20 },
		{ 382, 173, 20 },
		{ 382, 174, 20 },
		{ 382, 175, 20 },
		{ 382, 176, 20 },
		{ 382, 177, 20 },
		{ 382, 178, 20 },
		{ 382, 179, 20 },
		{ 382, 180, 20 },
		{ 382, 181, 20 },
		{ 382, 182, 20 },
		{ 382, 183, 20 },
		{ 382, 184, 20 },
		{ 382, 185, 20 },
		{ 382, 186, 20 },
		{ 382, 187, 20 },
		{ 382, 188, 20 },
		{ 382, 189, 20 },
		{ 382, 190, 20 },
		{ 382, 191, 20 },
		{ 382, 192, 20 },
		{ 382, 193, 20 },
		{ 382, 194, 20 },
		{ 382, 195, 20 },
		{ 382, 196, 20 },
		{ 382, 197, 20 },
		{ 382, 198, 20 },
		{ 382, 199, 20 },
		{ 382, 200, 20 },
		{ 382, 201, 20 },
		{ 382, 202, 20 },
		{ 382, 203, 20 },
		{ 382, 204, 20 },
		{ 382, 205, 20 },
		{ 382, 206, 20 },
		{ 382, 207, 20 },
		{ 382, 208, 20 },
		{ 382, 209, 20 },
		{ 382, 210, 20 },
		{ 382, 211, 20 },
		{ 382, 212, 20 },
		{ 382, 213, 20 },
		{ 382, 214, 20 },
		{ 382, 215, 20 },
		{ 382, 216, 20 },
		{ 382, 217, 20 },
		{ 382, 218, 20 },
		{ 382, 219, 20 },
		{ 382, 220, 20 },
		{ 382, 221, 20 },
		{ 382, 222, 20 },
		{ 383, 170, 20 },
		{ 383, 171, 20 },
		{ 383, 172, 20 },
		{ 383, 173, 20 },
		{ 383, 174, 20 },
		{ 383, 175, 20 },
		{ 383, 176, 20 },
		{ 383, 177, 20 },
		{ 383, 178, 20 },
		{ 383, 179, 20 },
		{ 383, 180, 20 },
		{ 383, 181, 20 },
		{ 383, 182, 20 },
		{ 383, 183, 20 },
		{ 383, 184, 20 },
		{ 383, 185, 20 },
		{ 383, 186, 20 },
		{ 383, 187, 20 },
		{ 383, 188, 20 },
		{ 383, 189, 20 },
		{ 383, 190, 20 },
		{ 383, 191, 20 },
		{ 383, 192, 20 },
		{ 383, 193, 20 },
		{ 383, 194, 20 },
		{ 383, 195, 20 },
		{ 383, 196, 20 },
		{ 383, 197, 20 },
		{ 383, 198, 20 },
		{ 383, 199, 20 },
		{ 383, 200, 20 },
		{ 383, 201, 20 },
		{ 383, 202, 20 },
		{ 383, 203, 20 },
		{ 383, 204, 20 },
		{ 383, 205, 20 },
		{ 383, 206, 20 },
		{ 383, 207, 20 },
		{ 383, 208, 20 },
		{ 383, 209, 20 },
		{ 383, 210, 20 },
		{ 383, 211, 20 },
		{ 383, 212, 20 },
		{ 383, 213, 20 },
		{ 383, 214, 20 },
		{ 383, 215, 20 },
		{ 383, 216, 20 },
		{ 383, 217, 20 },
		{ 383, 218, 20 },
		{ 131, 264, 21 },
		{ 131, 265, 21 },
		{ 131, 266, 21 },
		{ 131, 267, 21 },
		{ 131, 268, 21 },
		{ 131, 269, 21 },
		{ 131, 270, 21 },
		{ 131, 271, 21 },
		{ 131, 272, 21 },
		{ 131, 273, 21 },
		{ 131, 274, 21 },
		{ 131, 275, 21 },
		{ 131, 276, 21 },
		{ 131, 277, 21 },
		{ 131, 278, 21 },
		{ 131, 279, 21 },
		{ 131, 280, 21 },
		{ 131, 281, 21 },
		{ 131, 282, 21 },
		{ 131, 283, 21 },
		{ 131, 284, 21 },
		{ 131, 285, 21 },
		{ 131, 286, 21 },
		{ 131, 287, 21 },
		{ 131, 288, 21 },
		{ 131, 289, 21 },
		{ 131, 290, 21 },
		{ 131, 291, 21 },
		{ 131, 292, 21 },
		{ 131, 293, 21 },
		{ 131, 294, 21 },
		{ 131, 295, 21 },
		{ 131, 296, 21 },
		{ 131, 297, 21 },
		{ 131, 298, 21 },
		{ 131, 299, 21 },
		{ 131, 300, 21 },
		{ 131, 301, 21 },
		{ 131, 302, 21 },
		{ 131, 303, 21 },
		{ 131, 304, 21 },
		{ 131, 305, 21 },
		{ 131, 306, 21 },
		{ 131, 307, 21 },
		{ 131, 308, 21 },
		{ 131, 309, 21 },
		{ 131, 310, 21 },
		{ 131, 311, 21 },
		{ 131, 312, 21 },
		{ 131, 313, 21 },
		{ 132, 260, 21 },
		{ 132, 261, 21 },
		{ 132, 262, 21 },
		{ 132, 263, 21 },
		{ 132, 264, 21 },
		{ 132, 265, 21 },
		{ 132, 266, 21 },
		{ 132, 267, 21 },
		{ 132, 268, 21 },
		{ 132, 269, 21 },
		{ 132, 270, 21 },
		{ 132, 271, 21 },
		{ 132, 272, 21 },
		{ 132, 273, 21 },
		{ 132, 274, 21 },
		{ 132, 275, 21 },
		{ 132, 276, 21 },
		{ 132, 277, 21 },
		{ 132, 278, 21 },
		{ 132, 279, 21 },
		{ 132, 280, 21 },
		{ 132, 281, 21 },
		{ 132, 282, 21 },
		{ 132, 283, 21 },
		{ 132, 284, 21 },
		{ 132, 285, 21 },
		{ 132, 286, 21 },
		{ 132, 287, 21 },
		{ 132, 288, 21 },
		{ 132, 289, 21 },
		{ 132, 290, 21 },
		{ 132, 291, 21 },
		{ 132, 292, 21 },
		{ 132, 293, 21 },
		{ 132, 294, 21 },
		{ 132, 295, 21 },
		{ 132, 296, 21 },
		{ 132, 297, 21 },
		{ 132, 298, 21 },
		{ 132, 299, 21 },
		{ 132, 300, 21 },
		{ 132, 301, 21 },
		{ 132, 302, 21 },
		{ 132, 303, 21 },
		{ 132, 304, 21 },
		{ 132, 305, 21 },
		{ 132, 306, 21 },
		{ 132, 307, 21 },
		{ 132, 308, 21 },
		{ 132, 309, 21 },
		{ 132, 310, 21 },
		{ 132, 311, 21 },
		{ 132, 312, 21 },
		{ 132, 313, 21 },
		{ 132, 314, 21 },
		{ 132, 315, 21 },
		{ 132, 316, 21 },
		{ 133, 259, 21 },
		{ 133, 260, 21 },
		{ 133, 261, 21 },
		{ 133, 262, 21 },
		{ 133, 263, 21 },
		{ 133, 264, 21 },
		{ 133, 265, 21 },
		{ 133, 266, 21 },
		{ 133, 267, 21 },
		{ 133, 268, 21 },
		{ 133, 269, 21 },
		{ 133, 270, 21 },
		{ 133, 271, 21 },
		{ 133, 272, 21 },
		{ 133, 273, 21 },
		{ 133, 274, 21 },
		{ 133, 275, 21 },
		{ 133, 276, 21 },
		{ 133, 277, 21 },
		{ 133, 278, 21 },
		{ 133, 279, 21 },
		{ 133, 280, 21 },
		{ 133, 281, 21 },
		{ 133, 282, 21 },
		{ 133, 283, 21 },
		{ 133, 284, 21 },
		{ 133, 285, 21 },
		{ 133, 286, 21 },
		{ 133, 287, 21 },
		{ 133, 288, 21 },
		{ 133, 289, 21 },
		{ 133, 290, 21 },
		{ 133, 291, 21 },
		{ 133, 292, 21 },
		{ 133, 293, 21 },
		{ 133, 294, 21 },
		{ 133, 295, 21 },
		{ 133, 296, 21 },
		{ 133, 297, 21 },
		{ 133, 298, 21 },
		{ 133, 299, 21 },
		{ 133, 300, 21 },
		{ 133, 301, 21 },
		{ 133, 302, 21 },
		{ 133, 303, 21 },
		{ 133, 304, 21 },
		{ 133, 305, 21 },
		{ 133, 306, 21 },
		{ 133, 307, 21 },
		{ 133, 308, 21 },
		{ 133, 309, 21 },
		{ 133, 310, 21 },
		{ 133, 311, 21 },
		{ 133, 312, 21 },
		{ 133, 313, 21 },
		{ 133, 314, 21 },
		{ 133, 315, 21 },
		{ 133, 316, 21 },
		{ 133, 317, 21 },
		{ 134, 258, 21 },
		{ 134, 259, 21 },
		{ 134, 260, 21 },
		{ 134, 261, 21 },
		{ 134, 262, 21 },
		{ 134, 263, 21 },
		{ 134, 264, 21 },
		{ 134, 265, 21 },
		{ 134, 266, 21 },
		{ 134, 267, 21 },
		{ 134, 268, 21 },
		{ 134, 269, 21 },
		{ 134, 270, 21 },
		{ 134, 271, 21 },
		{ 134, 272, 21 },
		{ 134, 273, 21 },
		{ 134, 274, 21 },
		{ 134, 275, 21 },
		{ 134, 276, 21 },
		{ 134, 277, 21 },
		{ 134, 278, 21 },
		{ 134, 279, 21 },
		{ 134, 280, 21 },
		{ 134, 281, 21 },
		{ 134, 282, 21 },
		{ 134, 283, 21 },
		{ 134, 284, 21 },
		{ 134, 285, 21 },
		{ 134, 286, 21 },
		{ 134, 287, 21 },
		{ 134, 288, 21 },
		{ 134, 289, 21 },
		{ 134, 290, 21 },
		{ 134, 291, 21 },
		{ 134, 292, 21 },
		{ 134, 293, 21 },
		{ 134, 294, 21 },
		{ 134, 295, 21 },
		{ 134, 296, 21 },
		{ 134, 297, 21 },
		{ 134, 298, 21 },
		{ 134, 299, 21 },
		{ 134, 300, 21 },
		{ 134, 301, 21 },
		{ 134, 302, 21 },
		{ 134, 303, 21 },
		{ 134, 304, 21 },
		{ 134, 305, 21 },
		{ 134, 306, 21 },
		{ 134, 307, 21 },
		{ 134, 308, 21 },
		{ 134, 309, 21 },
		{ 134, 310, 21 },
		{ 134, 311, 21 },
		{ 134, 312, 21 },
		{ 134, 313, 21 },
		{ 134, 314, 21 },
		{ 134, 315, 21 },
		{ 134, 316, 21 },
		{ 134, 317, 21 },
		{ 135, 258, 21 },
		{ 135, 259, 21 },
		{ 135, 260, 21 },
		{ 135, 261, 21 },
		{ 135, 262, 21 },
		{ 135, 263, 21 },
		{ 135, 264, 21 },
		{ 135, 265, 21 },
		{ 135, 266, 21 },
		{ 135, 267, 21 },
		{ 135, 268, 21 },
		{ 135, 269, 21 },
		{ 135, 270, 21 },
		{ 135, 271, 21 },
		{ 135, 272, 21 },
		{ 135, 273, 21 },
		{ 135, 274, 21 },
		{ 135, 275, 21 },
		{ 135, 276, 21 },
		{ 135, 277, 21 },
		{ 135, 278, 21 },
		{ 135, 279, 21 },
		{ 135, 280, 21 },
		{ 135, 281, 21 },
		{ 135, 282, 21 },
		{ 135, 283, 21 },
		{ 135, 284, 21 },
		{ 135, 285, 21 },
		{ 135, 286, 21 },
		{ 135, 287, 21 },
		{ 135, 288, 21 },
		{ 135, 289, 21 },
		{ 135, 290, 21 },
		{ 135, 291, 21 },
		{ 135, 292, 21 },
		{ 135, 293, 21 },
		{ 135, 294, 21 },
		{ 135, 295, 21 },
		{ 135, 296, 21 },
		{ 135, 297, 21 },
		{ 135, 298, 21 },
		{ 135, 299, 21 },
		{ 135, 300, 21 },
		{ 135, 301, 21 },
		{ 135, 302, 21 },
		{ 135, 303, 21 },
		{ 135, 304, 21 },
		{ 135, 305, 21 },
		{ 135, 306, 21 },
		{ 135, 307, 21 },
		{ 135, 308, 21 },
		{ 135, 309, 21 },
		{ 135, 310, 21 },
		{ 135, 311, 21 },
		{ 135, 312, 21 },
		{ 135, 313, 21 },
		{ 135, 314, 21 },
		{ 135, 315, 21 },
		{ 135, 316, 21 },
		{ 135, 317, 21 },
		{ 136, 258, 21 },
		{ 136, 259, 21 },
		{ 136, 260, 21 },
		{ 136, 261, 21 },
		{ 136, 262, 21 },
		{ 136, 263, 21 },
		{ 136, 264, 21 },
		{ 136, 265, 21 },
		{ 136, 266, 21 },
		{ 136, 267, 21 },
		{ 136, 268, 21 },
		{ 136, 269, 21 },
		{ 136, 270, 21 },
		{ 136, 271, 21 },
		{ 136, 272, 21 },
		{ 136, 273, 21 },
		{ 136, 274, 21 },
		{ 136, 275, 21 },
		{ 136, 276, 21 },
		{ 136, 277, 21 },
		{ 136, 278, 21 },
		{ 136, 279, 21 },
		{ 136, 280, 21 },
		{ 136, 281, 21 },
		{ 136, 282, 21 },
		{ 136, 283, 21 },
		{ 136, 284, 21 },
		{ 136, 285, 21 },
		{ 136, 286, 21 },
		{ 136, 287, 21 },
		{ 136, 288, 21 },
		{ 136, 289, 21 },
		{ 136, 290, 21 },
		{ 136, 291, 21 },
		{ 136, 292, 21 },
		{ 136, 293, 21 },
		{ 136, 294, 21 },
		{ 136, 295, 21 },
		{ 136, 296, 21 },
		{ 136, 297, 21 },
		{ 136, 298, 21 },
		{ 136, 299, 21 },
		{ 136, 300, 21 },
		{ 136, 301, 21 },
		{ 136, 302, 21 },
		{ 136, 303, 21 },
		{ 136, 304, 21 },
		{ 136, 305, 21 },
		{ 136, 306, 21 },
		{ 136, 307, 21 },
		{ 136, 308, 21 },
		{ 136, 309, 21 },
		{ 136, 310, 21 },
		{ 136, 311, 21 },
		{ 136, 312, 21 },
		{ 136, 313, 21 },
		{ 136, 314, 21 },
		{ 136, 315, 21 },
		{ 136, 316, 21 },
		{ 137, 258, 21 },
		{ 137, 259, 21 },
		{ 137, 260, 21 },
		{ 137, 261, 21 },
		{ 137, 262, 21 },
		{ 137, 263, 21 },
		{ 137, 264, 21 },
		{ 137, 265, 21 },
		{ 137, 266, 21 },
		{ 137, 267, 21 },
		{ 137, 268, 21 },
		{ 137, 269, 21 },
		{ 137, 270, 21 },
		{ 137, 271, 21 },
		{ 137, 272, 21 },
		{ 137, 273, 21 },
		{ 137, 274, 21 },
		{ 137, 275, 21 },
		{ 137, 276, 21 },
		{ 137, 277, 21 },
		{ 137, 278, 21 },
		{ 137, 279, 21 },
		{ 137, 280, 21 },
		{ 137, 281, 21 },
		{ 137, 282, 21 },
		{ 137, 283, 21 },
		{ 137, 284, 21 },
		{ 137, 285, 21 },
		{ 137, 286, 21 },
		{ 137, 287, 21 },
		{ 137, 288, 21 },
		{ 137, 289, 21 },
		{ 137, 290, 21 },
		{ 137, 291, 21 },
		{ 137, 292, 21 },
		{ 137, 293, 21 },
		{ 137, 294, 21 },
		{ 137, 295, 21 },
		{ 137, 296, 21 },
		{ 137, 297, 21 },
		{ 137, 298, 21 },
		{ 137, 299, 21 },
		{ 137, 300, 21 },
		{ 137, 301, 21 },
		{ 137, 302, 21 },
		{ 137, 303, 21 },
		{ 137, 304, 21 },
		{ 137, 305, 21 },
		{ 137, 306, 21 },
		{ 137, 307, 21 },
		{ 137, 308, 21 },
		{ 137, 309, 21 },
		{ 137, 310, 21 },
		{ 137, 311, 21 },
		{ 137, 312, 21 },
		{ 137, 313, 21 },
		{ 137, 314, 21 },
		{ 137, 315, 21 },
		{ 137, 316, 21 },
		{ 138, 258, 21 },
		{ 138, 259, 21 },
		{ 138, 260, 21 },
		{ 138, 261, 21 },
		{ 138, 262, 21 },
		{ 138, 263, 21 },
		{ 138, 264, 21 },
		{ 138, 265, 21 },
		{ 138, 266, 21 },
		{ 138, 267, 21 },
		{ 138, 268, 21 },
		{ 138, 269, 21 },
		{ 138, 270, 21 },
		{ 138, 271, 21 },
		{ 138, 272, 21 },
		{ 138, 273, 21 },
		{ 138, 274, 21 },
		{ 138, 275, 21 },
		{ 138, 276, 21 },
		{ 138, 277, 21 },
		{ 138, 278, 21 },
		{ 138, 279, 21 },
		{ 138, 280, 21 },
		{ 138, 281, 21 },
		{ 138, 282, 21 },
		{ 138, 283, 21 },
		{ 138, 284, 21 },
		{ 138, 285, 21 },
		{ 138, 286, 21 },
		{ 138, 287, 21 },
		{ 138, 288, 21 },
		{ 138, 289, 21 },
		{ 138, 290, 21 },
		{ 138, 291, 21 },
		{ 138, 292, 21 },
		{ 138, 293, 21 },
		{ 138, 294, 21 },
		{ 138, 295, 21 },
		{ 138, 296, 21 },
		{ 138, 297, 21 },
		{ 138, 298, 21 },
		{ 138, 299, 21 },
		{ 138, 300, 21 },
		{ 138, 301, 21 },
		{ 138, 302, 21 },
		{ 138, 303, 21 },
		{ 138, 304, 21 },
		{ 138, 305, 21 },
		{ 138, 306, 21 },
		{ 138, 307, 21 },
		{ 138, 308, 21 },
		{ 138, 309, 21 },
		{ 138, 310, 21 },
		{ 138, 311, 21 },
		{ 138, 312, 21 },
		{ 138, 313, 21 },
		{ 138, 314, 21 },
		{ 138, 315, 21 },
		{ 138, 316, 21 },
		{ 139, 258, 21 },
		{ 139, 259, 21 },
		{ 139, 260, 21 },
		{ 139, 261, 21 },
		{ 139, 262, 21 },
		{ 139, 263, 21 },
		{ 139, 264, 21 },
		{ 139, 265, 21 },
		{ 139, 266, 21 },
		{ 139, 267, 21 },
		{ 139, 268, 21 },
		{ 139, 269, 21 },
		{ 139, 270, 21 },
		{ 139, 271, 21 },
		{ 139, 272, 21 },
		{ 139, 273, 21 },
		{ 139, 274, 21 },
		{ 139, 275, 21 },
		{ 139, 276, 21 },
		{ 139, 277, 21 },
		{ 139, 278, 21 },
		{ 139, 279, 21 },
		{ 139, 280, 21 },
		{ 139, 281, 21 },
		{ 139, 282, 21 },
		{ 139, 283, 21 },
		{ 139, 284, 21 },
		{ 139, 285, 21 },
		{ 139, 286, 21 },
		{ 139, 287, 21 },
		{ 139, 288, 21 },
		{ 139, 289, 21 },
		{ 139, 290, 21 },
		{ 139, 291, 21 },
		{ 139, 292, 21 },
		{ 139, 293, 21 },
		{ 139, 294, 21 },
		{ 139, 295, 21 },
		{ 139, 296, 21 },
		{ 139, 297, 21 },
		{ 139, 298, 21 },
		{ 139, 299, 21 },
		{ 139, 300, 21 },
		{ 139, 301, 21 },
		{ 139, 302, 21 },
		{ 139, 303, 21 },
		{ 139, 304, 21 },
		{ 139, 305, 21 },
		{ 139, 306, 21 },
		{ 139, 307, 21 },
		{ 139, 308, 21 },
		{ 139, 309, 21 },
		{ 139, 310, 21 },
		{ 139, 311, 21 },
		{ 139, 312, 21 },
		{ 139, 313, 21 },
		{ 139, 314, 21 },
		{ 139, 315, 21 },
		{ 140, 258, 21 },
		{ 140, 259, 21 },
		{ 140, 260, 21 },
		{ 140, 261, 21 },
		{ 140, 262, 21 },
		{ 140, 263, 21 },
		{ 140, 264, 21 },
		{ 140, 265, 21 },
		{ 140, 266, 21 },
		{ 140, 267, 21 },
		{ 140, 268, 21 },
		{ 140, 269, 21 },
		{ 140, 270, 21 },
		{ 140, 271, 21 },
		{ 140, 272, 21 },
		{ 140, 273, 21 },
		{ 140, 274, 21 },
		{ 140, 275, 21 },
		{ 140, 276, 21 },
		{ 140, 277, 21 },
		{ 140, 278, 21 },
		{ 140, 279, 21 },
		{ 140, 280, 21 },
		{ 140, 281, 21 },
		{ 140, 282, 21 },
		{ 140, 283, 21 },
		{ 140, 284, 21 },
		{ 140, 285, 21 },
		{ 140, 286, 21 },
		{ 140, 287, 21 },
		{ 140, 288, 21 },
		{ 140, 289, 21 },
		{ 140, 290, 21 },
		{ 140, 291, 21 },
		{ 140, 292, 21 },
		{ 140, 293, 21 },
		{ 140, 294, 21 },
		{ 140, 295, 21 },
		{ 140, 296, 21 },
		{ 140, 297, 21 },
		{ 140, 298, 21 },
		{ 140, 299, 21 },
		{ 140, 300, 21 },
		{ 140, 301, 21 },
		{ 140, 302, 21 },
		{ 140, 303, 21 },
		{ 140, 304, 21 },
		{ 140, 305, 21 },
		{ 140, 306, 21 },
		{ 140, 307, 21 },
		{ 140, 308, 21 },
		{ 140, 309, 21 },
		{ 140, 310, 21 },
		{ 140, 311, 21 },
		{ 140, 312, 21 },
		{ 140, 313, 21 },
		{ 140, 314, 21 },
		{ 140, 315, 21 },
		{ 141, 258, 21 },
		{ 141, 259, 21 },
		{ 141, 260, 21 },
		{ 141, 261, 21 },
		{ 141, 262, 21 },
		{ 141, 263, 21 },
		{ 141, 264, 21 },
		{ 141, 265, 21 },
		{ 141, 266, 21 },
		{ 141, 267, 21 },
		{ 141, 268, 21 },
		{ 141, 269, 21 },
		{ 141, 270, 21 },
		{ 141, 271, 21 },
		{ 141, 272, 21 },
		{ 141, 273, 21 },
		{ 141, 274, 21 },
		{ 141, 275, 21 },
		{ 141, 276, 21 },
		{ 141, 277, 21 },
		{ 141, 278, 21 },
		{ 141, 279, 21 },
		{ 141, 280, 21 },
		{ 141, 281, 21 },
		{ 141, 282, 21 },
		{ 141, 283, 21 },
		{ 141, 284, 21 },
		{ 141, 285, 21 },
		{ 141, 286, 21 },
		{ 141, 287, 21 },
		{ 141, 288, 21 },
		{ 141, 289, 21 },
		{ 141, 290, 21 },
		{ 141, 291, 21 },
		{ 141, 292, 21 },
		{ 141, 293, 21 },
		{ 141, 294, 21 },
		{ 141, 295, 21 },
		{ 141, 296, 21 },
		{ 141, 297, 21 },
		{ 141, 298, 21 },
		{ 141, 299, 21 },
		{ 141, 300, 21 },
		{ 141, 301, 21 },
		{ 141, 302, 21 },
		{ 141, 303, 21 },
		{ 141, 304, 21 },
		{ 141, 305, 21 },
		{ 141, 306, 21 },
		{ 141, 307, 21 },
		{ 141, 308, 21 },
		{ 141, 309, 21 },
		{ 141, 310, 21 },
		{ 141, 311, 21 },
		{ 141, 312, 21 },
		{ 141, 313, 21 },
		{ 141, 314, 21 },
		{ 142, 258, 21 },
		{ 142, 259, 21 },
		{ 142, 260, 21 },
		{ 142, 261, 21 },
		{ 142, 262, 21 },
		{ 142, 263, 21 },
		{ 142, 264, 21 },
		{ 142, 265, 21 },
		{ 142, 266, 21 },
		{ 142, 267, 21 },
		{ 142, 268, 21 },
		{ 142, 269, 21 },
		{ 142, 270, 21 },
		{ 142, 271, 21 },
		{ 142, 272, 21 },
		{ 142, 273, 21 },
		{ 142, 274, 21 },
		{ 142, 275, 21 },
		{ 142, 276, 21 },
		{ 142, 277, 21 },
		{ 142, 278, 21 },
		{ 142, 279, 21 },
		{ 142, 280, 21 },
		{ 142, 281, 21 },
		{ 142, 282, 21 },
		{ 142, 283, 21 },
		{ 142, 284, 21 },
		{ 142, 285, 21 },
		{ 142, 286, 21 },
		{ 142, 287, 21 },
		{ 142, 288, 21 },
		{ 142, 289, 21 },
		{ 142, 290, 21 },
		{ 142, 291, 21 },
		{ 142, 292, 21 },
		{ 142, 293, 21 },
		{ 142, 294, 21 },
		{ 142, 295, 21 },
		{ 142, 296, 21 },
		{ 142, 297, 21 },
		{ 142, 298, 21 },
		{ 142, 299, 21 },
		{ 142, 300, 21 },
		{ 142, 301, 21 },
		{ 142, 302, 21 },
		{ 142, 303, 21 },
		{ 142, 304, 21 },
		{ 142, 305, 21 },
		{ 142, 306, 21 },
		{ 142, 307, 21 },
		{ 142, 308, 21 },
		{ 142, 309, 21 },
		{ 142, 310, 21 },
		{ 142, 311, 21 },
		{ 142, 312, 21 },
		{ 142, 313, 21 },
		{ 142, 314, 21 },
		{ 143, 258, 21 },
		{ 143, 259, 21 },
		{ 143, 260, 21 },
		{ 143, 261, 21 },
		{ 143, 262, 21 },
		{ 143, 263, 21 },
		{ 143, 264, 21 },
		{ 143, 265, 21 },
		{ 143, 266, 21 },
		{ 143, 267, 21 },
		{ 143, 268, 21 },
		{ 143, 269, 21 },
		{ 143, 270, 21 },
		{ 143, 271, 21 },
		{ 143, 272, 21 },
		{ 143, 273, 21 },
		{ 143, 274, 21 },
		{ 143, 275, 21 },
		{ 143, 276, 21 },
		{ 143, 277, 21 },
		{ 143, 278, 21 },
		{ 143, 279, 21 },
		{ 143, 280, 21 },
		{ 143, 281, 21 },
		{ 143, 282, 21 },
		{ 143, 283, 21 },
		{ 143, 284, 21 },
		{ 143, 285, 21 },
		{ 143, 286, 21 },
		{ 143, 287, 21 },
		{ 143, 288, 21 },
		{ 143, 289, 21 },
		{ 143, 290, 21 },
		{ 143, 291, 21 },
		{ 143, 292, 21 },
		{ 143, 293, 21 },
		{ 143, 294, 21 },
		{ 143, 295, 21 },
		{ 143, 296, 21 },
		{ 143, 297, 21 },
		{ 143, 298, 21 },
		{ 143, 299, 21 },
		{ 143, 300, 21 },
		{ 143, 301, 21 },
		{ 143, 302, 21 },
		{ 143, 303, 21 },
		{ 143, 304, 21 },
		{ 143, 305, 21 },
		{ 143, 306, 21 },
		{ 143, 307, 21 },
		{ 143, 308, 21 },
		{ 143, 309, 21 },
		{ 143, 310, 21 },
		{ 143, 311, 21 },
		{ 143, 312, 21 },
		{ 143, 313, 21 },
		{ 144, 258, 21 },
		{ 144, 259, 21 },
		{ 144, 260, 21 },
		{ 144, 261, 21 },
		{ 144, 262, 21 },
		{ 144, 263, 21 },
		{ 144, 264, 21 },
		{ 144, 265, 21 },
		{ 144, 266, 21 },
		{ 144, 267, 21 },
		{ 144, 268, 21 },
		{ 144, 269, 21 },
		{ 144, 270, 21 },
		{ 144, 271, 21 },
		{ 144, 272, 21 },
		{ 144, 273, 21 },
		{ 144, 274, 21 },
		{ 144, 275, 21 },
		{ 144, 276, 21 },
		{ 144, 277, 21 },
		{ 144, 278, 21 },
		{ 144, 279, 21 },
		{ 144, 280, 21 },
		{ 144, 281, 21 },
		{ 144, 282, 21 },
		{ 144, 283, 21 },
		{ 144, 284, 21 },
		{ 144, 285, 21 },
		{ 144, 286, 21 },
		{ 144, 287, 21 },
		{ 144, 288, 21 },
		{ 144, 289, 21 },
		{ 144, 290, 21 },
		{ 144, 291, 21 },
		{ 144, 292, 21 },
		{ 144, 293, 21 },
		{ 144, 294, 21 },
		{ 144, 295, 21 },
		{ 144, 296, 21 },
		{ 144, 297, 21 },
		{ 144, 298, 21 },
		{ 144, 299, 21 },
		{ 144, 300, 21 },
		{ 144, 301, 21 },
		{ 144, 302, 21 },
		{ 144, 303, 21 },
		{ 144, 304, 21 },
		{ 144, 305, 21 },
		{ 144, 306, 21 },
		{ 144, 307, 21 },
		{ 144, 308, 21 },
		{ 144, 309, 21 },
		{ 144, 310, 21 },
		{ 144, 311, 21 },
		{ 144, 312, 21 },
		{ 144, 313, 21 },
		{ 145, 258, 21 },
		{ 145, 259, 21 },
		{ 145, 260, 21 },
		{ 145, 261, 21 },
		{ 145, 262, 21 },
		{ 145, 263, 21 },
		{ 145, 264, 21 },
		{ 145, 265, 21 },
		{ 145, 266, 21 },
		{ 145, 267, 21 },
		{ 145, 268, 21 },
		{ 145, 269, 21 },
		{ 145, 270, 21 },
		{ 145, 271, 21 },
		{ 145, 272, 21 },
		{ 145, 273, 21 },
		{ 145, 274, 21 },
		{ 145, 275, 21 },
		{ 145, 276, 21 },
		{ 145, 277, 21 },
		{ 145, 278, 21 },
		{ 145, 279, 21 },
		{ 145, 280, 21 },
		{ 145, 281, 21 },
		{ 145, 282, 21 },
		{ 145, 283, 21 },
		{ 145, 284, 21 },
		{ 145, 285, 21 },
		{ 145, 286, 21 },
		{ 145, 287, 21 },
		{ 145, 288, 21 },
		{ 145, 289, 21 },
		{ 145, 290, 21 },
		{ 145, 291, 21 },
		{ 145, 292, 21 },
		{ 145, 293, 21 },
		{ 145, 294, 21 },
		{ 145, 295, 21 },
		{ 145, 296, 21 },
		{ 145, 297, 21 },
		{ 145, 298, 21 },
		{ 145, 299, 21 },
		{ 145, 300, 21 },
		{ 145, 301, 21 },
		{ 145, 302, 21 },
		{ 145, 303, 21 },
		{ 145, 304, 21 },
		{ 145, 305, 21 },
		{ 145, 306, 21 },
		{ 145, 307, 21 },
		{ 145, 308, 21 },
		{ 145, 309, 21 },
		{ 145, 310, 21 },
		{ 145, 311, 21 },
		{ 145, 312, 21 },
		{ 146, 258, 21 },
		{ 146, 259, 21 },
		{ 146, 260, 21 },
		{ 146, 261, 21 },
		{ 146, 262, 21 },
		{ 146, 263, 21 },
		{ 146, 264, 21 },
		{ 146, 265, 21 },
		{ 146, 266, 21 },
		{ 146, 267, 21 },
		{ 146, 268, 21 },
		{ 146, 269, 21 },
		{ 146, 270, 21 },
		{ 146, 271, 21 },
		{ 146, 272, 21 },
		{ 146, 273, 21 },
		{ 146, 274, 21 },
		{ 146, 275, 21 },
		{ 146, 276, 21 },
		{ 146, 277, 21 },
		{ 146, 278, 21 },
		{ 146, 279, 21 },
		{ 146, 280, 21 },
		{ 146, 281, 21 },
		{ 146, 282, 21 },
		{ 146, 283, 21 },
		{ 146, 284, 21 },
		{ 146, 285, 21 },
		{ 146, 286, 21 },
		{ 146, 287, 21 },
		{ 146, 288, 21 },
		{ 146, 289, 21 },
		{ 146, 290, 21 },
		{ 146, 291, 21 },
		{ 146, 292, 21 },
		{ 146, 293, 21 },
		{ 146, 294, 21 },
		{ 146, 295, 21 },
		{ 146, 296, 21 },
		{ 146, 297, 21 },
		{ 146, 298, 21 },
		{ 146, 299, 21 },
		{ 146, 300, 21 },
		{ 146, 301, 21 },
		{ 146, 302, 21 },
		{ 146, 303, 21 },
		{ 146, 304, 21 },
		{ 146, 305, 21 },
		{ 146, 306, 21 },
		{ 146, 307, 21 },
		{ 146, 308, 21 },
		{ 146, 309, 21 },
		{ 146, 310, 21 },
		{ 146, 311, 21 },
		{ 146, 312, 21 },
		{ 147, 258, 21 },
		{ 147, 259, 21 },
		{ 147, 260, 21 },
		{ 147, 261, 21 },
		{ 147, 262, 21 },
		{ 147, 263, 21 },
		{ 147, 264, 21 },
		{ 147, 265, 21 },
		{ 147, 266, 21 },
		{ 147, 267, 21 },
		{ 147, 268, 21 },
		{ 147, 269, 21 },
		{ 147, 270, 21 },
		{ 147, 271, 21 },
		{ 147, 272, 21 },
		{ 147, 273, 21 },
		{ 147, 274, 21 },
		{ 147, 275, 21 },
		{ 147, 276, 21 },
		{ 147, 277, 21 },
		{ 147, 278, 21 },
		{ 147, 279, 21 },
		{ 147, 280, 21 },
		{ 147, 281, 21 },
		{ 147, 282, 21 },
		{ 147, 283, 21 },
		{ 147, 284, 21 },
		{ 147, 285, 21 },
		{ 147, 286, 21 },
		{ 147, 287, 21 },
		{ 147, 288, 21 },
		{ 147, 289, 21 },
		{ 147, 290, 21 },
		{ 147, 291, 21 },
		{ 147, 292, 21 },
		{ 147, 293, 21 },
		{ 147, 294, 21 },
		{ 147, 295, 21 },
		{ 147, 296, 21 },
		{ 147, 297, 21 },
		{ 147, 298, 21 },
		{ 147, 299, 21 },
		{ 147, 300, 21 },
		{ 147, 301, 21 },
		{ 147, 302, 21 },
		{ 147, 303, 21 },
		{ 147, 304, 21 },
		{ 147, 305, 21 },
		{ 147, 306, 21 },
		{ 147, 307, 21 },
		{ 147, 308, 21 },
		{ 147, 309, 21 },
		{ 147, 310, 21 },
		{ 147, 311, 21 },
		{ 148, 258, 21 },
		{ 148, 259, 21 },
		{ 148, 260, 21 },
		{ 148, 261, 21 },
		{ 148, 262, 21 },
		{ 148, 263, 21 },
		{ 148, 264, 21 },
		{ 148, 265, 21 },
		{ 148, 266, 21 },
		{ 148, 267, 21 },
		{ 148, 268, 21 },
		{ 148, 269, 21 },
		{ 148, 270, 21 },
		{ 148, 271, 21 },
		{ 148, 272, 21 },
		{ 148, 273, 21 },
		{ 148, 274, 21 },
		{ 148, 275, 21 },
		{ 148, 276, 21 },
		{ 148, 277, 21 },
		{ 148, 278, 21 },
		{ 148, 279, 21 },
		{ 148, 280, 21 },
		{ 148, 281, 21 },
		{ 148, 282, 21 },
		{ 148, 283, 21 },
		{ 148, 284, 21 },
		{ 148, 285, 21 },
		{ 148, 286, 21 },
		{ 148, 287, 21 },
		{ 148, 288, 21 },
		{ 148, 289, 21 },
		{ 148, 290, 21 },
		{ 148, 291, 21 },
		{ 148, 292, 21 },
		{ 148, 293, 21 },
		{ 148, 294, 21 },
		{ 148, 295, 21 },
		{ 148, 296, 21 },
		{ 148, 297, 21 },
		{ 148, 298, 21 },
		{ 148, 299, 21 },
		{ 148, 300, 21 },
		{ 148, 301, 21 },
		{ 148, 302, 21 },
		{ 148, 303, 21 },
		{ 148, 304, 21 },
		{ 148, 305, 21 },
		{ 148, 306, 21 },
		{ 148, 307, 21 },
		{ 148, 308, 21 },
		{ 148, 309, 21 },
		{ 148, 310, 21 },
		{ 148, 311, 21 },
		{ 149, 258, 21 },
		{ 149, 259, 21 },
		{ 149, 260, 21 },
		{ 149, 261, 21 },
		{ 149, 262, 21 },
		{ 149, 263, 21 },
		{ 149, 264, 21 },
		{ 149, 265, 21 },
		{ 149, 266, 21 },
		{ 149, 267, 21 },
		{ 149, 268, 21 },
		{ 149, 269, 21 },
		{ 149, 270, 21 },
		{ 149, 271, 21 },
		{ 149, 272, 21 },
		{ 149, 273, 21 },
		{ 149, 274, 21 },
		{ 149, 275, 21 },
		{ 149, 276, 21 },
		{ 149, 277, 21 },
		{ 149, 278, 21 },
		{ 149, 279, 21 },
		{ 149, 280, 21 },
		{ 149, 281, 21 },
		{ 149, 282, 21 },
		{ 149, 283, 21 },
		{ 149, 284, 21 },
		{ 149, 285, 21 },
		{ 149, 286, 21 },
		{ 149, 287, 21 },
		{ 149, 288, 21 },
		{ 149, 289, 21 },
		{ 149, 290, 21 },
		{ 149, 291, 21 },
		{ 149, 292, 21 },
		{ 149, 293, 21 },
		{ 149, 294, 21 },
		{ 149, 295, 21 },
		{ 149, 296, 21 },
		{ 149, 297, 21 },
		{ 149, 298, 21 },
		{ 149, 299, 21 },
		{ 149, 300, 21 },
		{ 149, 301, 21 },
		{ 149, 302, 21 },
		{ 149, 303, 21 },
		{ 149, 304, 21 },
		{ 149, 305, 21 },
		{ 149, 306, 21 },
		{ 149, 307, 21 },
		{ 149, 308, 21 },
		{ 149, 309, 21 },
		{ 149, 310, 21 },
		{ 150, 258, 21 },
		{ 150, 259, 21 },
		{ 150, 260, 21 },
		{ 150, 261, 21 },
		{ 150, 262, 21 },
		{ 150, 263, 21 },
		{ 150, 264, 21 },
		{ 150, 265, 21 },
		{ 150, 266, 21 },
		{ 150, 267, 21 },
		{ 150, 268, 21 },
		{ 150, 269, 21 },
		{ 150, 270, 21 },
		{ 150, 271, 21 },
		{ 150, 272, 21 },
		{ 150, 273, 21 },
		{ 150, 274, 21 },
		{ 150, 275, 21 },
		{ 150, 276, 21 },
		{ 150, 277, 21 },
		{ 150, 278, 21 },
		{ 150, 279, 21 },
		{ 150, 280, 21 },
		{ 150, 281, 21 },
		{ 150, 282, 21 },
		{ 150, 283, 21 },
		{ 150, 284, 21 },
		{ 150, 285, 21 },
		{ 150, 286, 21 },
		{ 150, 287, 21 },
		{ 150, 288, 21 },
		{ 150, 289, 21 },
		{ 150, 290, 21 },
		{ 150, 291, 21 },
		{ 150, 292, 21 },
		{ 150, 293, 21 },
		{ 150, 294, 21 },
		{ 150, 295, 21 },
		{ 150, 296, 21 },
		{ 150, 297, 21 },
		{ 150, 298, 21 },
		{ 150, 299, 21 },
		{ 150, 300, 21 },
		{ 150, 301, 21 },
		{ 150, 302, 21 },
		{ 150, 303, 21 },
		{ 150, 304, 21 },
		{ 150, 305, 21 },
		{ 150, 306, 21 },
		{ 150, 307, 21 },
		{ 150, 308, 21 },
		{ 150, 309, 21 },
		{ 150, 310, 21 },
		{ 151, 258, 21 },
		{ 151, 259, 21 },
		{ 151, 260, 21 },
		{ 151, 261, 21 },
		{ 151, 262, 21 },
		{ 151, 263, 21 },
		{ 151, 264, 21 },
		{ 151, 265, 21 },
		{ 151, 266, 21 },
		{ 151, 267, 21 },
		{ 151, 268, 21 },
		{ 151, 269, 21 },
		{ 151, 270, 21 },
		{ 151, 271, 21 },
		{ 151, 272, 21 },
		{ 151, 273, 21 },
		{ 151, 274, 21 },
		{ 151, 275, 21 },
		{ 151, 276, 21 },
		{ 151, 277, 21 },
		{ 151, 278, 21 },
		{ 151, 279, 21 },
		{ 151, 280, 21 },
		{ 151, 281, 21 },
		{ 151, 282, 21 },
		{ 151, 283, 21 },
		{ 151, 284, 21 },
		{ 151, 285, 21 },
		{ 151, 286, 21 },
		{ 151, 287, 21 },
		{ 151, 288, 21 },
		{ 151, 289, 21 },
		{ 151, 290, 21 },
		{ 151, 291, 21 },
		{ 151, 292, 21 },
		{ 151, 293, 21 },
		{ 151, 294, 21 },
		{ 151, 295, 21 },
		{ 151, 296, 21 },
		{ 151, 297, 21 },
		{ 151, 298, 21 },
		{ 151, 299, 21 },
		{ 151, 300, 21 },
		{ 151, 301, 21 },
		{ 151, 302, 21 },
		{ 151, 303, 21 },
		{ 151, 304, 21 },
		{ 151, 305, 21 },
		{ 151, 306, 21 },
		{ 151, 307, 21 },
		{ 151, 308, 21 },
		{ 151, 309, 21 },
		{ 152, 258, 21 },
		{ 152, 259, 21 },
		{ 152, 260, 21 },
		{ 152, 261, 21 },
		{ 152, 262, 21 },
		{ 152, 263, 21 },
		{ 152, 264, 21 },
		{ 152, 265, 21 },
		{ 152, 266, 21 },
		{ 152, 267, 21 },
		{ 152, 268, 21 },
		{ 152, 269, 21 },
		{ 152, 270, 21 },
		{ 152, 271, 21 },
		{ 152, 272, 21 },
		{ 152, 273, 21 },
		{ 152, 274, 21 },
		{ 152, 275, 21 },
		{ 152, 276, 21 },
		{ 152, 277, 21 },
		{ 152, 278, 21 },
		{ 152, 279, 21 },
		{ 152, 280, 21 },
		{ 152, 281, 21 },
		{ 152, 282, 21 },
		{ 152, 283, 21 },
		{ 152, 284, 21 },
		{ 152, 285, 21 },
		{ 152, 286, 21 },
		{ 152, 287, 21 },
		{ 152, 288, 21 },
		{ 152, 289, 21 },
		{ 152, 290, 21 },
		{ 152, 291, 21 },
		{ 152, 292, 21 },
		{ 152, 293, 21 },
		{ 152, 294, 21 },
		{ 152, 295, 21 },
		{ 152, 296, 21 },
		{ 152, 297, 21 },
		{ 152, 298, 21 },
		{ 152, 299, 21 },
		{ 152, 300, 21 },
		{ 152, 301, 21 },
		{ 152, 302, 21 },
		{ 152, 303, 21 },
		{ 152, 304, 21 },
		{ 152, 305, 21 },
		{ 152, 306, 21 },
		{ 152, 307, 21 },
		{ 152, 308, 21 },
		{ 152, 309, 21 },
		{ 153, 258, 21 },
		{ 153, 259, 21 },
		{ 153, 260, 21 },
		{ 153, 261, 21 },
		{ 153, 262, 21 },
		{ 153, 263, 21 },
		{ 153, 264, 21 },
		{ 153, 265, 21 },
		{ 153, 266, 21 },
		{ 153, 267, 21 },
		{ 153, 268, 21 },
		{ 153, 269, 21 },
		{ 153, 270, 21 },
		{ 153, 271, 21 },
		{ 153, 272, 21 },
		{ 153, 273, 21 },
		{ 153, 274, 21 },
		{ 153, 275, 21 },
		{ 153, 276, 21 },
		{ 153, 277, 21 },
		{ 153, 278, 21 },
		{ 153, 279, 21 },
		{ 153, 280, 21 },
		{ 153, 281, 21 },
		{ 153, 282, 21 },
		{ 153, 283, 21 },
		{ 153, 284, 21 },
		{ 153, 285, 21 },
		{ 153, 286, 21 },
		{ 153, 287, 21 },
		{ 153, 288, 21 },
		{ 153, 289, 21 },
		{ 153, 290, 21 },
		{ 153, 291, 21 },
		{ 153, 292, 21 },
		{ 153, 293, 21 },
		{ 153, 294, 21 },
		{ 153, 295, 21 },
		{ 153, 296, 21 },
		{ 153, 297, 21 },
		{ 153, 298, 21 },
		{ 153, 299, 21 },
		{ 153, 300, 21 },
		{ 153, 301, 21 },
		{ 153, 302, 21 },
		{ 153, 303, 21 },
		{ 153, 304, 21 },
		{ 153, 305, 21 },
		{ 153, 306, 21 },
		{ 153, 307, 21 },
		{ 153, 308, 21 },
		{ 154, 258, 21 },
		{ 154, 259, 21 },
		{ 154, 260, 21 },
		{ 154, 261, 21 },
		{ 154, 262, 21 },
		{ 154, 263, 21 },
		{ 154, 264, 21 },
		{ 154, 265, 21 },
		{ 154, 266, 21 },
		{ 154, 267, 21 },
		{ 154, 268, 21 },
		{ 154, 269, 21 },
		{ 154, 270, 21 },
		{ 154, 271, 21 },
		{ 154, 272, 21 },
		{ 154, 273, 21 },
		{ 154, 274, 21 },
		{ 154, 275, 21 },
		{ 154, 276, 21 },
		{ 154, 277, 21 },
		{ 154, 278, 21 },
		{ 154, 279, 21 },
		{ 154, 280, 21 },
		{ 154, 281, 21 },
		{ 154, 282, 21 },
		{ 154, 283, 21 },
		{ 154, 284, 21 },
		{ 154, 285, 21 },
		{ 154, 286, 21 },
		{ 154, 287, 21 },
		{ 154, 288, 21 },
		{ 154, 289, 21 },
		{ 154, 290, 21 },
		{ 154, 291, 21 },
		{ 154, 292, 21 },
		{ 154, 293, 21 },
		{ 154, 294, 21 },
		{ 154, 295, 21 },
		{ 154, 296, 21 },
		{ 154, 297, 21 },
		{ 154, 298, 21 },
		{ 154, 299, 21 },
		{ 154, 300, 21 },
		{ 154, 301, 21 },
		{ 154, 302, 21 },
		{ 154, 303, 21 },
		{ 154, 304, 21 },
		{ 154, 305, 21 },
		{ 154, 306, 21 },
		{ 154, 307, 21 },
		{ 154, 308, 21 },
		{ 155, 258, 21 },
		{ 155, 259, 21 },
		{ 155, 260, 21 },
		{ 155, 261, 21 },
		{ 155, 262, 21 },
		{ 155, 263, 21 },
		{ 155, 264, 21 },
		{ 155, 265, 21 },
		{ 155, 266, 21 },
		{ 155, 267, 21 },
		{ 155, 268, 21 },
		{ 155, 269, 21 },
		{ 155, 270, 21 },
		{ 155, 271, 21 },
		{ 155, 272, 21 },
		{ 155, 273, 21 },
		{ 155, 274, 21 },
		{ 155, 275, 21 },
		{ 155, 276, 21 },
		{ 155, 277, 21 },
		{ 155, 278, 21 },
		{ 155, 279, 21 },
		{ 155, 280, 21 },
		{ 155, 281, 21 },
		{ 155, 282, 21 },
		{ 155, 283, 21 },
		{ 155, 284, 21 },
		{ 155, 285, 21 },
		{ 155, 286, 21 },
		{ 155, 287, 21 },
		{ 155, 288, 21 },
		{ 155, 289, 21 },
		{ 155, 290, 21 },
		{ 155, 291, 21 },
		{ 155, 292, 21 },
		{ 155, 293, 21 },
		{ 155, 294, 21 },
		{ 155, 295, 21 },
		{ 155, 296, 21 },
		{ 155, 297, 21 },
		{ 155, 298, 21 },
		{ 155, 299, 21 },
		{ 155, 300, 21 },
		{ 155, 301, 21 },
		{ 155, 302, 21 },
		{ 155, 303, 21 },
		{ 155, 304, 21 },
		{ 155, 305, 21 },
		{ 155, 306, 21 },
		{ 155, 307, 21 },
		{ 156, 258, 21 },
		{ 156, 259, 21 },
		{ 156, 260, 21 },
		{ 156, 261, 21 },
		{ 156, 262, 21 },
		{ 156, 263, 21 },
		{ 156, 264, 21 },
		{ 156, 265, 21 },
		{ 156, 266, 21 },
		{ 156, 267, 21 },
		{ 156, 268, 21 },
		{ 156, 269, 21 },
		{ 156, 270, 21 },
		{ 156, 271, 21 },
		{ 156, 272, 21 },
		{ 156, 273, 21 },
		{ 156, 274, 21 },
		{ 156, 275, 21 },
		{ 156, 276, 21 },
		{ 156, 277, 21 },
		{ 156, 278, 21 },
		{ 156, 279, 21 },
		{ 156, 280, 21 },
		{ 156, 281, 21 },
		{ 156, 282, 21 },
		{ 156, 283, 21 },
		{ 156, 284, 21 },
		{ 156, 285, 21 },
		{ 156, 286, 21 },
		{ 156, 287, 21 },
		{ 156, 288, 21 },
		{ 156, 289, 21 },
		{ 156, 290, 21 },
		{ 156, 291, 21 },
		{ 156, 292, 21 },
		{ 156, 293, 21 },
		{ 156, 294, 21 },
		{ 156, 295, 21 },
		{ 156, 296, 21 },
		{ 156, 297, 21 },
		{ 156, 298, 21 },
		{ 156, 299, 21 },
		{ 156, 300, 21 },
		{ 156, 301, 21 },
		{ 156, 302, 21 },
		{ 156, 303, 21 },
		{ 156, 304, 21 },
		{ 156, 305, 21 },
		{ 156, 306, 21 },
		{ 156, 307, 21 },
		{ 157, 258, 21 },
		{ 157, 259, 21 },
		{ 157, 260, 21 },
		{ 157, 261, 21 },
		{ 157, 262, 21 },
		{ 157, 263, 21 },
		{ 157, 264, 21 },
		{ 157, 265, 21 },
		{ 157, 266, 21 },
		{ 157, 267, 21 },
		{ 157, 268, 21 },
		{ 157, 269, 21 },
		{ 157, 270, 21 },
		{ 157, 271, 21 },
		{ 157, 272, 21 },
		{ 157, 273, 21 },
		{ 157, 274, 21 },
		{ 157, 275, 21 },
		{ 157, 276, 21 },
		{ 157, 277, 21 },
		{ 157, 278, 21 },
		{ 157, 279, 21 },
		{ 157, 280, 21 },
		{ 157, 281, 21 },
		{ 157, 282, 21 },
		{ 157, 283, 21 },
		{ 157, 284, 21 },
		{ 157, 285, 21 },
		{ 157, 286, 21 },
		{ 157, 287, 21 },
		{ 157, 288, 21 },
		{ 157, 289, 21 },
		{ 157, 290, 21 },
		{ 157, 291, 21 },
		{ 157, 292, 21 },
		{ 157, 293, 21 },
		{ 157, 294, 21 },
		{ 157, 295, 21 },
		{ 157, 296, 21 },
		{ 157, 297, 21 },
		{ 157, 298, 21 },
		{ 157, 299, 21 },
		{ 157, 300, 21 },
		{ 157, 301, 21 },
		{ 157, 302, 21 },
		{ 157, 303, 21 },
		{ 157, 304, 21 },
		{ 157, 305, 21 },
		{ 157, 306, 21 },
		{ 158, 258, 21 },
		{ 158, 259, 21 },
		{ 158, 260, 21 },
		{ 158, 261, 21 },
		{ 158, 262, 21 },
		{ 158, 263, 21 },
		{ 158, 264, 21 },
		{ 158, 265, 21 },
		{ 158, 266, 21 },
		{ 158, 267, 21 },
		{ 158, 268, 21 },
		{ 158, 269, 21 },
		{ 158, 270, 21 },
		{ 158, 271, 21 },
		{ 158, 272, 21 },
		{ 158, 273, 21 },
		{ 158, 274, 21 },
		{ 158, 275, 21 },
		{ 158, 276, 21 },
		{ 158, 277, 21 },
		{ 158, 278, 21 },
		{ 158, 279, 21 },
		{ 158, 280, 21 },
		{ 158, 281, 21 },
		{ 158, 282, 21 },
		{ 158, 283, 21 },
		{ 158, 284, 21 },
		{ 158, 285, 21 },
		{ 158, 286, 21 },
		{ 158, 287, 21 },
		{ 158, 288, 21 },
		{ 158, 289, 21 },
		{ 158, 290, 21 },
		{ 158, 291, 21 },
		{ 158, 292, 21 },
		{ 158, 293, 21 },
		{ 158, 294, 21 },
		{ 158, 295, 21 },
		{ 158, 296, 21 },
		{ 158, 297, 21 },
		{ 158, 298, 21 },
		{ 158, 299, 21 },
		{ 158, 300, 21 },
		{ 158, 301, 21 },
		{ 158, 302, 21 },
		{ 158, 303, 21 },
		{ 158, 304, 21 },
		{ 158, 305, 21 },
		{ 158, 306, 21 },
		{ 159, 258, 21 },
		{ 159, 259, 21 },
		{ 159, 260, 21 },
		{ 159, 261, 21 },
		{ 159, 262, 21 },
		{ 159, 263, 21 },
		{ 159, 264, 21 },
		{ 159, 265, 21 },
		{ 159, 266, 21 },
		{ 159, 267, 21 },
		{ 159, 268, 21 },
		{ 159, 269, 21 },
		{ 159, 270, 21 },
		{ 159, 271, 21 },
		{ 159, 272, 21 },
		{ 159, 273, 21 },
		{ 159, 274, 21 },
		{ 159, 275, 21 },
		{ 159, 276, 21 },
		{ 159, 277, 21 },
		{ 159, 278, 21 },
		{ 159, 279, 21 },
		{ 159, 280, 21 },
		{ 159, 281, 21 },
		{ 159, 282, 21 },
		{ 159, 283, 21 },
		{ 159, 284, 21 },
		{ 159, 285, 21 },
		{ 159, 286, 21 },
		{ 159, 287, 21 },
		{ 159, 288, 21 },
		{ 159, 289, 21 },
		{ 159, 290, 21 },
		{ 159, 291, 21 },
		{ 159, 292, 21 },
		{ 159, 293, 21 },
		{ 159, 294, 21 },
		{ 159, 295, 21 },
		{ 159, 296, 21 },
		{ 159, 297, 21 },
		{ 159, 298, 21 },
		{ 159, 299, 21 },
		{ 159, 300, 21 },
		{ 159, 301, 21 },
		{ 159, 302, 21 },
		{ 159, 303, 21 },
		{ 159, 304, 21 },
		{ 159, 305, 21 },
		{ 160, 258, 21 },
		{ 160, 259, 21 },
		{ 160, 260, 21 },
		{ 160, 261, 21 },
		{ 160, 262, 21 },
		{ 160, 263, 21 },
		{ 160, 264, 21 },
		{ 160, 265, 21 },
		{ 160, 266, 21 },
		{ 160, 267, 21 },
		{ 160, 268, 21 },
		{ 160, 269, 21 },
		{ 160, 270, 21 },
		{ 160, 271, 21 },
		{ 160, 272, 21 },
		{ 160, 273, 21 },
		{ 160, 274, 21 },
		{ 160, 275, 21 },
		{ 160, 276, 21 },
		{ 160, 277, 21 },
		{ 160, 278, 21 },
		{ 160, 279, 21 },
		{ 160, 280, 21 },
		{ 160, 281, 21 },
		{ 160, 282, 21 },
		{ 160, 283, 21 },
		{ 160, 284, 21 },
		{ 160, 285, 21 },
		{ 160, 286, 21 },
		{ 160, 287, 21 },
		{ 160, 288, 21 },
		{ 160, 289, 21 },
		{ 160, 290, 21 },
		{ 160, 291, 21 },
		{ 160, 292, 21 },
		{ 160, 293, 21 },
		{ 160, 294, 21 },
		{ 160, 295, 21 },
		{ 160, 296, 21 },
		{ 160, 297, 21 },
		{ 160, 298, 21 },
		{ 160, 299, 21 },
		{ 160, 300, 21 },
		{ 160, 301, 21 },
		{ 160, 302, 21 },
		{ 160, 303, 21 },
		{ 160, 304, 21 },
		{ 160, 305, 21 },
		{ 161, 258, 21 },
		{ 161, 259, 21 },
		{ 161, 260, 21 },
		{ 161, 261, 21 },
		{ 161, 262, 21 },
		{ 161, 263, 21 },
		{ 161, 264, 21 },
		{ 161, 265, 21 },
		{ 161, 266, 21 },
		{ 161, 267, 21 },
		{ 161, 268, 21 },
		{ 161, 269, 21 },
		{ 161, 270, 21 },
		{ 161, 271, 21 },
		{ 161, 272, 21 },
		{ 161, 273, 21 },
		{ 161, 274, 21 },
		{ 161, 275, 21 },
		{ 161, 276, 21 },
		{ 161, 277, 21 },
		{ 161, 278, 21 },
		{ 161, 279, 21 },
		{ 161, 280, 21 },
		{ 161, 281, 21 },
		{ 161, 282, 21 },
		{ 161, 283, 21 },
		{ 161, 284, 21 },
		{ 161, 285, 21 },
		{ 161, 286, 21 },
		{ 161, 287, 21 },
		{ 161, 288, 21 },
		{ 161, 289, 21 },
		{ 161, 290, 21 },
		{ 161, 291, 21 },
		{ 161, 292, 21 },
		{ 161, 293, 21 },
		{ 161, 294, 21 },
		{ 161, 295, 21 },
		{ 161, 296, 21 },
		{ 161, 297, 21 },
		{ 161, 298, 21 },
		{ 161, 299, 21 },
		{ 161, 300, 21 },
		{ 161, 301, 21 },
		{ 161, 302, 21 },
		{ 161, 303, 21 },
		{ 161, 304, 21 },
		{ 162, 258, 21 },
		{ 162, 259, 21 },
		{ 162, 260, 21 },
		{ 162, 261, 21 },
		{ 162, 262, 21 },
		{ 162, 263, 21 },
		{ 162, 264, 21 },
		{ 162, 265, 21 },
		{ 162, 266, 21 },
		{ 162, 267, 21 },
		{ 162, 268, 21 },
		{ 162, 269, 21 },
		{ 162, 270, 21 },
		{ 162, 271, 21 },
		{ 162, 272, 21 },
		{ 162, 273, 21 },
		{ 162, 274, 21 },
		{ 162, 275, 21 },
		{ 162, 276, 21 },
		{ 162, 277, 21 },
		{ 162, 278, 21 },
		{ 162, 279, 21 },
		{ 162, 280, 21 },
		{ 162, 281, 21 },
		{ 162, 282, 21 },
		{ 162, 283, 21 },
		{ 162, 284, 21 },
		{ 162, 285, 21 },
		{ 162, 286, 21 },
		{ 162, 287, 21 },
		{ 162, 288, 21 },
		{ 162, 289, 21 },
		{ 162, 290, 21 },
		{ 162, 291, 21 },
		{ 162, 292, 21 },
		{ 162, 293, 21 },
		{ 162, 294, 21 },
		{ 162, 295, 21 },
		{ 162, 296, 21 },
		{ 162, 297, 21 },
		{ 162, 298, 21 },
		{ 162, 299, 21 },
		{ 162, 300, 21 },
		{ 162, 301, 21 },
		{ 162, 302, 21 },
		{ 162, 303, 21 },
		{ 162, 304, 21 },
		{ 163, 258, 21 },
		{ 163, 259, 21 },
		{ 163, 260, 21 },
		{ 163, 261, 21 },
		{ 163, 262, 21 },
		{ 163, 263, 21 },
		{ 163, 264, 21 },
		{ 163, 265, 21 },
		{ 163, 266, 21 },
		{ 163, 267, 21 },
		{ 163, 268, 21 },
		{ 163, 269, 21 },
		{ 163, 270, 21 },
		{ 163, 271, 21 },
		{ 163, 272, 21 },
		{ 163, 273, 21 },
		{ 163, 274, 21 },
		{ 163, 275, 21 },
		{ 163, 276, 21 },
		{ 163, 277, 21 },
		{ 163, 278, 21 },
		{ 163, 279, 21 },
		{ 163, 280, 21 },
		{ 163, 281, 21 },
		{ 163, 282, 21 },
		{ 163, 283, 21 },
		{ 163, 284, 21 },
		{ 163, 285, 21 },
		{ 163, 286, 21 },
		{ 163, 287, 21 },
		{ 163, 288, 21 },
		{ 163, 289, 21 },
		{ 163, 290, 21 },
		{ 163, 291, 21 },
		{ 163, 292, 21 },
		{ 163, 293, 21 },
		{ 163, 294, 21 },
		{ 163, 295, 21 },
		{ 163, 296, 21 },
		{ 163, 297, 21 },
		{ 163, 298, 21 },
		{ 163, 299, 21 },
		{ 163, 300, 21 },
		{ 163, 301, 21 },
		{ 163, 302, 21 },
		{ 163, 303, 21 },
		{ 164, 258, 21 },
		{ 164, 259, 21 },
		{ 164, 260, 21 },
		{ 164, 261, 21 },
		{ 164, 262, 21 },
		{ 164, 263, 21 },
		{ 164, 264, 21 },
		{ 164, 265, 21 },
		{ 164, 266, 21 },
		{ 164, 267, 21 },
		{ 164, 268, 21 },
		{ 164, 269, 21 },
		{ 164, 270, 21 },
		{ 164, 271, 21 },
		{ 164, 272, 21 },
		{ 164, 273, 21 },
		{ 164, 274, 21 },
		{ 164, 275, 21 },
		{ 164, 276, 21 },
		{ 164, 277, 21 },
		{ 164, 278, 21 },
		{ 164, 279, 21 },
		{ 164, 280, 21 },
		{ 164, 281, 21 },
		{ 164, 282, 21 },
		{ 164, 283, 21 },
		{ 164, 284, 21 },
		{ 164, 285, 21 },
		{ 164, 286, 21 },
		{ 164, 287, 21 },
		{ 164, 288, 21 },
		{ 164, 289, 21 },
		{ 164, 290, 21 },
		{ 164, 291, 21 },
		{ 164, 292, 21 },
		{ 164, 293, 21 },
		{ 164, 294, 21 },
		{ 164, 295, 21 },
		{ 164, 296, 21 },
		{ 164, 297, 21 },
		{ 164, 298, 21 },
		{ 164, 299, 21 },
		{ 164, 300, 21 },
		{ 164, 301, 21 },
		{ 164, 302, 21 },
		{ 164, 303, 21 },
		{ 165, 258, 21 },
		{ 165, 259, 21 },
		{ 165, 260, 21 },
		{ 165, 261, 21 },
		{ 165, 262, 21 },
		{ 165, 263, 21 },
		{ 165, 264, 21 },
		{ 165, 265, 21 },
		{ 165, 266, 21 },
		{ 165, 267, 21 },
		{ 165, 268, 21 },
		{ 165, 269, 21 },
		{ 165, 270, 21 },
		{ 165, 271, 21 },
		{ 165, 272, 21 },
		{ 165, 273, 21 },
		{ 165, 274, 21 },
		{ 165, 275, 21 },
		{ 165, 276, 21 },
		{ 165, 277, 21 },
		{ 165, 278, 21 },
		{ 165, 279, 21 },
		{ 165, 280, 21 },
		{ 165, 281, 21 },
		{ 165, 282, 21 },
		{ 165, 283, 21 },
		{ 165, 284, 21 },
		{ 165, 285, 21 },
		{ 165, 286, 21 },
		{ 165, 287, 21 },
		{ 165, 288, 21 },
		{ 165, 289, 21 },
		{ 165, 290, 21 },
		{ 165, 291, 21 },
		{ 165, 292, 21 },
		{ 165, 293, 21 },
		{ 165, 294, 21 },
		{ 165, 295, 21 },
		{ 165, 296, 21 },
		{ 165, 297, 21 },
		{ 165, 298, 21 },
		{ 165, 299, 21 },
		{ 165, 300, 21 },
		{ 165, 301, 21 },
		{ 165, 302, 21 },
		{ 166, 258, 21 },
		{ 166, 259, 21 },
		{ 166, 260, 21 },
		{ 166, 261, 21 },
		{ 166, 262, 21 },
		{ 166, 263, 21 },
		{ 166, 264, 21 },
		{ 166, 265, 21 },
		{ 166, 266, 21 },
		{ 166, 267, 21 },
		{ 166, 268, 21 },
		{ 166, 269, 21 },
		{ 166, 270, 21 },
		{ 166, 271, 21 },
		{ 166, 272, 21 },
		{ 166, 273, 21 },
		{ 166, 274, 21 },
		{ 166, 275, 21 },
		{ 166, 276, 21 },
		{ 166, 277, 21 },
		{ 166, 278, 21 },
		{ 166, 279, 21 },
		{ 166, 280, 21 },
		{ 166, 281, 21 },
		{ 166, 282, 21 },
		{ 166, 283, 21 },
		{ 166, 284, 21 },
		{ 166, 285, 21 },
		{ 166, 286, 21 },
		{ 166, 287, 21 },
		{ 166, 288, 21 },
		{ 166, 289, 21 },
		{ 166, 290, 21 },
		{ 166, 291, 21 },
		{ 166, 292, 21 },
		{ 166, 293, 21 },
		{ 166, 294, 21 },
		{ 166, 295, 21 },
		{ 166, 296, 21 },
		{ 166, 297, 21 },
		{ 166, 298, 21 },
		{ 166, 299, 21 },
		{ 166, 300, 21 },
		{ 166, 301, 21 },
		{ 166, 302, 21 },
		{ 167, 258, 21 },
		{ 167, 259, 21 },
		{ 167, 260, 21 },
		{ 167, 261, 21 },
		{ 167, 262, 21 },
		{ 167, 263, 21 },
		{ 167, 264, 21 },
		{ 167, 265, 21 },
		{ 167, 266, 21 },
		{ 167, 267, 21 },
		{ 167, 268, 21 },
		{ 167, 269, 21 },
		{ 167, 270, 21 },
		{ 167, 271, 21 },
		{ 167, 272, 21 },
		{ 167, 273, 21 },
		{ 167, 274, 21 },
		{ 167, 275, 21 },
		{ 167, 276, 21 },
		{ 167, 277, 21 },
		{ 167, 278, 21 },
		{ 167, 279, 21 },
		{ 167, 280, 21 },
		{ 167, 281, 21 },
		{ 167, 282, 21 },
		{ 167, 283, 21 },
		{ 167, 284, 21 },
		{ 167, 285, 21 },
		{ 167, 286, 21 },
		{ 167, 287, 21 },
		{ 167, 288, 21 },
		{ 167, 289, 21 },
		{ 167, 290, 21 },
		{ 167, 291, 21 },
		{ 167, 292, 21 },
		{ 167, 293, 21 },
		{ 167, 294, 21 },
		{ 167, 295, 21 },
		{ 167, 296, 21 },
		{ 167, 297, 21 },
		{ 167, 298, 21 },
		{ 167, 299, 21 },
		{ 167, 300, 21 },
		{ 167, 301, 21 },
		{ 168, 258, 21 },
		{ 168, 259, 21 },
		{ 168, 260, 21 },
		{ 168, 261, 21 },
		{ 168, 262, 21 },
		{ 168, 263, 21 },
		{ 168, 264, 21 },
		{ 168, 265, 21 },
		{ 168, 266, 21 },
		{ 168, 267, 21 },
		{ 168, 268, 21 },
		{ 168, 269, 21 },
		{ 168, 270, 21 },
		{ 168, 271, 21 },
		{ 168, 272, 21 },
		{ 168, 273, 21 },
		{ 168, 274, 21 },
		{ 168, 275, 21 },
		{ 168, 276, 21 },
		{ 168, 277, 21 },
		{ 168, 278, 21 },
		{ 168, 279, 21 },
		{ 168, 280, 21 },
		{ 168, 281, 21 },
		{ 168, 282, 21 },
		{ 168, 283, 21 },
		{ 168, 284, 21 },
		{ 168, 285, 21 },
		{ 168, 286, 21 },
		{ 168, 287, 21 },
		{ 168, 288, 21 },
		{ 168, 289, 21 },
		{ 168, 290, 21 },
		{ 168, 291, 21 },
		{ 168, 292, 21 },
		{ 168, 293, 21 },
		{ 168, 294, 21 },
		{ 168, 295, 21 },
		{ 168, 296, 21 },
		{ 168, 297, 21 },
		{ 168, 298, 21 },
		{ 168, 299, 21 },
		{ 168, 300, 21 },
		{ 168, 301, 21 },
		{ 169, 258, 21 },
		{ 169, 259, 21 },
		{ 169, 260, 21 },
		{ 169, 261, 21 },
		{ 169, 262, 21 },
		{ 169, 263, 21 },
		{ 169, 264, 21 },
		{ 169, 265, 21 },
		{ 169, 266, 21 },
		{ 169, 267, 21 },
		{ 169, 268, 21 },
		{ 169, 269, 21 },
		{ 169, 270, 21 },
		{ 169, 271, 21 },
		{ 169, 272, 21 },
		{ 169, 273, 21 },
		{ 169, 274, 21 },
		{ 169, 275, 21 },
		{ 169, 276, 21 },
		{ 169, 277, 21 },
		{ 169, 278, 21 },
		{ 169, 279, 21 },
		{ 169, 280, 21 },
		{ 169, 281, 21 },
		{ 169, 282, 21 },
		{ 169, 283, 21 },
		{ 169, 284, 21 },
		{ 169, 285, 21 },
		{ 169, 286, 21 },
		{ 169, 287, 21 },
		{ 169, 288, 21 },
		{ 169, 289, 21 },
		{ 169, 290, 21 },
		{ 169, 291, 21 },
		{ 169, 292, 21 },
		{ 169, 293, 21 },
		{ 169, 294, 21 },
		{ 169, 295, 21 },
		{ 169, 296, 21 },
		{ 169, 297, 21 },
		{ 169, 298, 21 },
		{ 169, 299, 21 },
		{ 169, 300, 21 },
		{ 170, 258, 21 },
		{ 170, 259, 21 },
		{ 170, 260, 21 },
		{ 170, 261, 21 },
		{ 170, 262, 21 },
		{ 170, 263, 21 },
		{ 170, 264, 21 },
		{ 170, 265, 21 },
		{ 170, 266, 21 },
		{ 170, 267, 21 },
		{ 170, 268, 21 },
		{ 170, 269, 21 },
		{ 170, 270, 21 },
		{ 170, 271, 21 },
		{ 170, 272, 21 },
		{ 170, 273, 21 },
		{ 170, 274, 21 },
		{ 170, 275, 21 },
		{ 170, 276, 21 },
		{ 170, 277, 21 },
		{ 170, 278, 21 },
		{ 170, 279, 21 },
		{ 170, 280, 21 },
		{ 170, 281, 21 },
		{ 170, 282, 21 },
		{ 170, 283, 21 },
		{ 170, 284, 21 },
		{ 170, 285, 21 },
		{ 170, 286, 21 },
		{ 170, 287, 21 },
		{ 170, 288, 21 },
		{ 170, 289, 21 },
		{ 170, 290, 21 },
		{ 170, 291, 21 },
		{ 170, 292, 21 },
		{ 170, 293, 21 },
		{ 170, 294, 21 },
		{ 170, 295, 21 },
		{ 170, 296, 21 },
		{ 170, 297, 21 },
		{ 170, 298, 21 },
		{ 170, 299, 21 },
		{ 170, 300, 21 },
		{ 171, 258, 21 },
		{ 171, 259, 21 },
		{ 171, 260, 21 },
		{ 171, 261, 21 },
		{ 171, 262, 21 },
		{ 171, 263, 21 },
		{ 171, 264, 21 },
		{ 171, 265, 21 },
		{ 171, 266, 21 },
		{ 171, 267, 21 },
		{ 171, 268, 21 },
		{ 171, 269, 21 },
		{ 171, 270, 21 },
		{ 171, 271, 21 },
		{ 171, 272, 21 },
		{ 171, 273, 21 },
		{ 171, 274, 21 },
		{ 171, 275, 21 },
		{ 171, 276, 21 },
		{ 171, 277, 21 },
		{ 171, 278, 21 },
		{ 171, 279, 21 },
		{ 171, 280, 21 },
		{ 171, 281, 21 },
		{ 171, 282, 21 },
		{ 171, 283, 21 },
		{ 171, 284, 21 },
		{ 171, 285, 21 },
		{ 171, 286, 21 },
		{ 171, 287, 21 },
		{ 171, 288, 21 },
		{ 171, 289, 21 },
		{ 171, 290, 21 },
		{ 171, 291, 21 },
		{ 171, 292, 21 },
		{ 171, 293, 21 },
		{ 171, 294, 21 },
		{ 171, 295, 21 },
		{ 171, 296, 21 },
		{ 171, 297, 21 },
		{ 171, 298, 21 },
		{ 171, 299, 21 },
		{ 172, 258, 21 },
		{ 172, 259, 21 },
		{ 172, 260, 21 },
		{ 172, 261, 21 },
		{ 172, 262, 21 },
		{ 172, 263, 21 },
		{ 172, 264, 21 },
		{ 172, 265, 21 },
		{ 172, 266, 21 },
		{ 172, 267, 21 },
		{ 172, 268, 21 },
		{ 172, 269, 21 },
		{ 172, 270, 21 },
		{ 172, 271, 21 },
		{ 172, 272, 21 },
		{ 172, 273, 21 },
		{ 172, 274, 21 },
		{ 172, 275, 21 },
		{ 172, 276, 21 },
		{ 172, 277, 21 },
		{ 172, 278, 21 },
		{ 172, 279, 21 },
		{ 172, 280, 21 },
		{ 172, 281, 21 },
		{ 172, 282, 21 },
		{ 172, 283, 21 },
		{ 172, 284, 21 },
		{ 172, 285, 21 },
		{ 172, 286, 21 },
		{ 172, 287, 21 },
		{ 172, 288, 21 },
		{ 172, 289, 21 },
		{ 172, 290, 21 },
		{ 172, 291, 21 },
		{ 172, 292, 21 },
		{ 172, 293, 21 },
		{ 172, 294, 21 },
		{ 172, 295, 21 },
		{ 172, 296, 21 },
		{ 172, 297, 21 },
		{ 172, 298, 21 },
		{ 172, 299, 21 },
		{ 173, 258, 21 },
		{ 173, 259, 21 },
		{ 173, 260, 21 },
		{ 173, 261, 21 },
		{ 173, 262, 21 },
		{ 173, 263, 21 },
		{ 173, 264, 21 },
		{ 173, 265, 21 },
		{ 173, 266, 21 },
		{ 173, 267, 21 },
		{ 173, 268, 21 },
		{ 173, 269, 21 },
		{ 173, 270, 21 },
		{ 173, 271, 21 },
		{ 173, 272, 21 },
		{ 173, 273, 21 },
		{ 173, 274, 21 },
		{ 173, 275, 21 },
		{ 173, 276, 21 },
		{ 173, 277, 21 },
		{ 173, 278, 21 },
		{ 173, 279, 21 },
		{ 173, 280, 21 },
		{ 173, 281, 21 },
		{ 173, 282, 21 },
		{ 173, 283, 21 },
		{ 173, 284, 21 },
		{ 173, 285, 21 },
		{ 173, 286, 21 },
		{ 173, 287, 21 },
		{ 173, 288, 21 },
		{ 173, 289, 21 },
		{ 173, 290, 21 },
		{ 173, 291, 21 },
		{ 173, 292, 21 },
		{ 173, 293, 21 },
		{ 173, 294, 21 },
		{ 173, 295, 21 },
		{ 173, 296, 21 },
		{ 173, 297, 21 },
		{ 173, 298, 21 },
		{ 174, 258, 21 },
		{ 174, 259, 21 },
		{ 174, 260, 21 },
		{ 174, 261, 21 },
		{ 174, 262, 21 },
		{ 174, 263, 21 },
		{ 174, 264, 21 },
		{ 174, 265, 21 },
		{ 174, 266, 21 },
		{ 174, 267, 21 },
		{ 174, 268, 21 },
		{ 174, 269, 21 },
		{ 174, 270, 21 },
		{ 174, 271, 21 },
		{ 174, 272, 21 },
		{ 174, 273, 21 },
		{ 174, 274, 21 },
		{ 174, 275, 21 },
		{ 174, 276, 21 },
		{ 174, 277, 21 },
		{ 174, 278, 21 },
		{ 174, 279, 21 },
		{ 174, 280, 21 },
		{ 174, 281, 21 },
		{ 174, 282, 21 },
		{ 174, 283, 21 },
		{ 174, 284, 21 },
		{ 174, 285, 21 },
		{ 174, 286, 21 },
		{ 174, 287, 21 },
		{ 174, 288, 21 },
		{ 174, 289, 21 },
		{ 174, 290, 21 },
		{ 174, 291, 21 },
		{ 174, 292, 21 },
		{ 174, 293, 21 },
		{ 174, 294, 21 },
		{ 174, 295, 21 },
		{ 174, 296, 21 },
		{ 174, 297, 21 },
		{ 174, 298, 21 },
		{ 175, 258, 21 },
		{ 175, 259, 21 },
		{ 175, 260, 21 },
		{ 175, 261, 21 },
		{ 175, 262, 21 },
		{ 175, 263, 21 },
		{ 175, 264, 21 },
		{ 175, 265, 21 },
		{ 175, 266, 21 },
		{ 175, 267, 21 },
		{ 175, 268, 21 },
		{ 175, 269, 21 },
		{ 175, 270, 21 },
		{ 175, 271, 21 },
		{ 175, 272, 21 },
		{ 175, 273, 21 },
		{ 175, 274, 21 },
		{ 175, 275, 21 },
		{ 175, 276, 21 },
		{ 175, 277, 21 },
		{ 175, 278, 21 },
		{ 175, 279, 21 },
		{ 175, 280, 21 },
		{ 175, 281, 21 },
		{ 175, 282, 21 },
		{ 175, 283, 21 },
		{ 175, 284, 21 },
		{ 175, 285, 21 },
		{ 175, 286, 21 },
		{ 175, 287, 21 },
		{ 175, 288, 21 },
		{ 175, 289, 21 },
		{ 175, 290, 21 },
		{ 175, 291, 21 },
		{ 175, 292, 21 },
		{ 175, 293, 21 },
		{ 175, 294, 21 },
		{ 175, 295, 21 },
		{ 175, 296, 21 },
		{ 175, 297, 21 },
		{ 176, 258, 21 },
		{ 176, 259, 21 },
		{ 176, 260, 21 },
		{ 176, 261, 21 },
		{ 176, 262, 21 },
		{ 176, 263, 21 },
		{ 176, 264, 21 },
		{ 176, 265, 21 },
		{ 176, 266, 21 },
		{ 176, 267, 21 },
		{ 176, 268, 21 },
		{ 176, 269, 21 },
		{ 176, 270, 21 },
		{ 176, 271, 21 },
		{ 176, 272, 21 },
		{ 176, 273, 21 },
		{ 176, 274, 21 },
		{ 176, 275, 21 },
		{ 176, 276, 21 },
		{ 176, 277, 21 },
		{ 176, 278, 21 },
		{ 176, 279, 21 },
		{ 176, 280, 21 },
		{ 176, 281, 21 },
		{ 176, 282, 21 },
		{ 176, 283, 21 },
		{ 176, 284, 21 },
		{ 176, 285, 21 },
		{ 176, 286, 21 },
		{ 176, 287, 21 },
		{ 176, 288, 21 },
		{ 176, 289, 21 },
		{ 176, 290, 21 },
		{ 176, 291, 21 },
		{ 176, 292, 21 },
		{ 176, 293, 21 },
		{ 176, 294, 21 },
		{ 176, 295, 21 },
		{ 176, 296, 21 },
		{ 176, 297, 21 },
		{ 177, 258, 21 },
		{ 177, 259, 21 },
		{ 177, 260, 21 },
		{ 177, 261, 21 },
		{ 177, 262, 21 },
		{ 177, 263, 21 },
		{ 177, 264, 21 },
		{ 177, 265, 21 },
		{ 177, 266, 21 },
		{ 177, 267, 21 },
		{ 177, 268, 21 },
		{ 177, 269, 21 },
		{ 177, 270, 21 },
		{ 177, 271, 21 },
		{ 177, 272, 21 },
		{ 177, 273, 21 },
		{ 177, 274, 21 },
		{ 177, 275, 21 },
		{ 177, 276, 21 },
		{ 177, 277, 21 },
		{ 177, 278, 21 },
		{ 177, 279, 21 },
		{ 177, 280, 21 },
		{ 177, 281, 21 },
		{ 177, 282, 21 },
		{ 177, 283, 21 },
		{ 177, 284, 21 },
		{ 177, 285, 21 },
		{ 177, 286, 21 },
		{ 177, 287, 21 },
		{ 177, 288, 21 },
		{ 177, 289, 21 },
		{ 177, 290, 21 },
		{ 177, 291, 21 },
		{ 177, 292, 21 },
		{ 177, 293, 21 },
		{ 177, 294, 21 },
		{ 177, 295, 21 },
		{ 177, 296, 21 },
		{ 178, 258, 21 },
		{ 178, 259, 21 },
		{ 178, 260, 21 },
		{ 178, 261, 21 },
		{ 178, 262, 21 },
		{ 178, 263, 21 },
		{ 178, 264, 21 },
		{ 178, 265, 21 },
		{ 178, 266, 21 },
		{ 178, 267, 21 },
		{ 178, 268, 21 },
		{ 178, 269, 21 },
		{ 178, 270, 21 },
		{ 178, 271, 21 },
		{ 178, 272, 21 },
		{ 178, 273, 21 },
		{ 178, 274, 21 },
		{ 178, 275, 21 },
		{ 178, 276, 21 },
		{ 178, 277, 21 },
		{ 178, 278, 21 },
		{ 178, 279, 21 },
		{ 178, 280, 21 },
		{ 178, 281, 21 },
		{ 178, 282, 21 },
		{ 178, 283, 21 },
		{ 178, 284, 21 },
		{ 178, 285, 21 },
		{ 178, 286, 21 },
		{ 178, 287, 21 },
		{ 178, 288, 21 },
		{ 178, 289, 21 },
		{ 178, 290, 21 },
		{ 178, 291, 21 },
		{ 178, 292, 21 },
		{ 178, 293, 21 },
		{ 178, 294, 21 },
		{ 178, 295, 21 },
		{ 178, 296, 21 },
		{ 179, 258, 21 },
		{ 179, 259, 21 },
		{ 179, 260, 21 },
		{ 179, 261, 21 },
		{ 179, 262, 21 },
		{ 179, 263, 21 },
		{ 179, 264, 21 },
		{ 179, 265, 21 },
		{ 179, 266, 21 },
		{ 179, 267, 21 },
		{ 179, 268, 21 },
		{ 179, 269, 21 },
		{ 179, 270, 21 },
		{ 179, 271, 21 },
		{ 179, 272, 21 },
		{ 179, 273, 21 },
		{ 179, 274, 21 },
		{ 179, 275, 21 },
		{ 179, 276, 21 },
		{ 179, 277, 21 },
		{ 179, 278, 21 },
		{ 179, 279, 21 },
		{ 179, 280, 21 },
		{ 179, 281, 21 },
		{ 179, 282, 21 },
		{ 179, 283, 21 },
		{ 179, 284, 21 },
		{ 179, 285, 21 },
		{ 179, 286, 21 },
		{ 179, 287, 21 },
		{ 179, 288, 21 },
		{ 179, 289, 21 },
		{ 179, 290, 21 },
		{ 179, 291, 21 },
		{ 179, 292, 21 },
		{ 179, 293, 21 },
		{ 179, 294, 21 },
		{ 179, 295, 21 },
		{ 180, 258, 21 },
		{ 180, 259, 21 },
		{ 180, 260, 21 },
		{ 180, 261, 21 },
		{ 180, 262, 21 },
		{ 180, 263, 21 },
		{ 180, 264, 21 },
		{ 180, 265, 21 },
		{ 180, 266, 21 },
		{ 180, 267, 21 },
		{ 180, 268, 21 },
		{ 180, 269, 21 },
		{ 180, 270, 21 },
		{ 180, 271, 21 },
		{ 180, 272, 21 },
		{ 180, 273, 21 },
		{ 180, 274, 21 },
		{ 180, 275, 21 },
		{ 180, 276, 21 },
		{ 180, 277, 21 },
		{ 180, 278, 21 },
		{ 180, 279, 21 },
		{ 180, 280, 21 },
		{ 180, 281, 21 },
		{ 180, 282, 21 },
		{ 180, 283, 21 },
		{ 180, 284, 21 },
		{ 180, 285, 21 },
		{ 180, 286, 21 },
		{ 180, 287, 21 },
		{ 180, 288, 21 },
		{ 180, 289, 21 },
		{ 180, 290, 21 },
		{ 180, 291, 21 },
		{ 180, 292, 21 },
		{ 180, 293, 21 },
		{ 180, 294, 21 },
		{ 180, 295, 21 },
		{ 181, 258, 21 },
		{ 181, 259, 21 },
		{ 181, 260, 21 },
		{ 181, 261, 21 },
		{ 181, 262, 21 },
		{ 181, 263, 21 },
		{ 181, 264, 21 },
		{ 181, 265, 21 },
		{ 181, 266, 21 },
		{ 181, 267, 21 },
		{ 181, 268, 21 },
		{ 181, 269, 21 },
		{ 181, 270, 21 },
		{ 181, 271, 21 },
		{ 181, 272, 21 },
		{ 181, 273, 21 },
		{ 181, 274, 21 },
		{ 181, 275, 21 },
		{ 181, 276, 21 },
		{ 181, 277, 21 },
		{ 181, 278, 21 },
		{ 181, 279, 21 },
		{ 181, 280, 21 },
		{ 181, 281, 21 },
		{ 181, 282, 21 },
		{ 181, 283, 21 },
		{ 181, 284, 21 },
		{ 181, 285, 21 },
		{ 181, 286, 21 },
		{ 181, 287, 21 },
		{ 181, 288, 21 },
		{ 181, 289, 21 },
		{ 181, 290, 21 },
		{ 181, 291, 21 },
		{ 181, 292, 21 },
		{ 181, 293, 21 },
		{ 181, 294, 21 },
		{ 182, 258, 21 },
		{ 182, 259, 21 },
		{ 182, 260, 21 },
		{ 182, 261, 21 },
		{ 182, 262, 21 },
		{ 182, 263, 21 },
		{ 182, 264, 21 },
		{ 182, 265, 21 },
		{ 182, 266, 21 },
		{ 182, 267, 21 },
		{ 182, 268, 21 },
		{ 182, 269, 21 },
		{ 182, 270, 21 },
		{ 182, 271, 21 },
		{ 182, 272, 21 },
		{ 182, 273, 21 },
		{ 182, 274, 21 },
		{ 182, 275, 21 },
		{ 182, 276, 21 },
		{ 182, 277, 21 },
		{ 182, 278, 21 },
		{ 182, 279, 21 },
		{ 182, 280, 21 },
		{ 182, 281, 21 },
		{ 182, 282, 21 },
		{ 182, 283, 21 },
		{ 182, 284, 21 },
		{ 182, 285, 21 },
		{ 182, 286, 21 },
		{ 182, 287, 21 },
		{ 182, 288, 21 },
		{ 182, 289, 21 },
		{ 182, 290, 21 },
		{ 182, 291, 21 },
		{ 182, 292, 21 },
		{ 182, 293, 21 },
		{ 182, 294, 21 },
		{ 183, 258, 21 },
		{ 183, 259, 21 },
		{ 183, 260, 21 },
		{ 183, 261, 21 },
		{ 183, 262, 21 },
		{ 183, 263, 21 },
		{ 183, 264, 21 },
		{ 183, 265, 21 },
		{ 183, 266, 21 },
		{ 183, 267, 21 },
		{ 183, 268, 21 },
		{ 183, 269, 21 },
		{ 183, 270, 21 },
		{ 183, 271, 21 },
		{ 183, 272, 21 },
		{ 183, 273, 21 },
		{ 183, 274, 21 },
		{ 183, 275, 21 },
		{ 183, 276, 21 },
		{ 183, 277, 21 },
		{ 183, 278, 21 },
		{ 183, 279, 21 },
		{ 183, 280, 21 },
		{ 183, 281, 21 },
		{ 183, 282, 21 },
		{ 183, 283, 21 },
		{ 183, 284, 21 },
		{ 183, 285, 21 },
		{ 183, 286, 21 },
		{ 183, 287, 21 },
		{ 183, 288, 21 },
		{ 183, 289, 21 },
		{ 183, 290, 21 },
		{ 183, 291, 21 },
		{ 183, 292, 21 },
		{ 183, 293, 21 },
		{ 184, 258, 21 },
		{ 184, 259, 21 },
		{ 184, 260, 21 },
		{ 184, 261, 21 },
		{ 184, 262, 21 },
		{ 184, 263, 21 },
		{ 184, 264, 21 },
		{ 184, 265, 21 },
		{ 184, 266, 21 },
		{ 184, 267, 21 },
		{ 184, 268, 21 },
		{ 184, 269, 21 },
		{ 184, 270, 21 },
		{ 184, 271, 21 },
		{ 184, 272, 21 },
		{ 184, 273, 21 },
		{ 184, 274, 21 },
		{ 184, 275, 21 },
		{ 184, 276, 21 },
		{ 184, 277, 21 },
		{ 184, 278, 21 },
		{ 184, 279, 21 },
		{ 184, 280, 21 },
		{ 184, 281, 21 },
		{ 184, 282, 21 },
		{ 184, 283, 21 },
		{ 184, 284, 21 },
		{ 184, 285, 21 },
		{ 184, 286, 21 },
		{ 184, 287, 21 },
		{ 184, 288, 21 },
		{ 184, 289, 21 },
		{ 184, 290, 21 },
		{ 184, 291, 21 },
		{ 184, 292, 21 },
		{ 184, 293, 21 },
		{ 185, 258, 21 },
		{ 185, 259, 21 },
		{ 185, 260, 21 },
		{ 185, 261, 21 },
		{ 185, 262, 21 },
		{ 185, 263, 21 },
		{ 185, 264, 21 },
		{ 185, 265, 21 },
		{ 185, 266, 21 },
		{ 185, 267, 21 },
		{ 185, 268, 21 },
		{ 185, 269, 21 },
		{ 185, 270, 21 },
		{ 185, 271, 21 },
		{ 185, 272, 21 },
		{ 185, 273, 21 },
		{ 185, 274, 21 },
		{ 185, 275, 21 },
		{ 185, 276, 21 },
		{ 185, 277, 21 },
		{ 185, 278, 21 },
		{ 185, 279, 21 },
		{ 185, 280, 21 },
		{ 185, 281, 21 },
		{ 185, 282, 21 },
		{ 185, 283, 21 },
		{ 185, 284, 21 },
		{ 185, 285, 21 },
		{ 185, 286, 21 },
		{ 185, 287, 21 },
		{ 185, 288, 21 },
		{ 185, 289, 21 },
		{ 185, 290, 21 },
		{ 185, 291, 21 },
		{ 185, 292, 21 },
		{ 186, 258, 21 },
		{ 186, 259, 21 },
		{ 186, 260, 21 },
		{ 186, 261, 21 },
		{ 186, 262, 21 },
		{ 186, 263, 21 },
		{ 186, 264, 21 },
		{ 186, 265, 21 },
		{ 186, 266, 21 },
		{ 186, 267, 21 },
		{ 186, 268, 21 },
		{ 186, 269, 21 },
		{ 186, 270, 21 },
		{ 186, 271, 21 },
		{ 186, 272, 21 },
		{ 186, 273, 21 },
		{ 186, 274, 21 },
		{ 186, 275, 21 },
		{ 186, 276, 21 },
		{ 186, 277, 21 },
		{ 186, 278, 21 },
		{ 186, 279, 21 },
		{ 186, 280, 21 },
		{ 186, 281, 21 },
		{ 186, 282, 21 },
		{ 186, 283, 21 },
		{ 186, 284, 21 },
		{ 186, 285, 21 },
		{ 186, 286, 21 },
		{ 186, 287, 21 },
		{ 186, 288, 21 },
		{ 186, 289, 21 },
		{ 186, 290, 21 },
		{ 186, 291, 21 },
		{ 186, 292, 21 },
		{ 187, 258, 21 },
		{ 187, 259, 21 },
		{ 187, 260, 21 },
		{ 187, 261, 21 },
		{ 187, 262, 21 },
		{ 187, 263, 21 },
		{ 187, 264, 21 },
		{ 187, 265, 21 },
		{ 187, 266, 21 },
		{ 187, 267, 21 },
		{ 187, 268, 21 },
		{ 187, 269, 21 },
		{ 187, 270, 21 },
		{ 187, 271, 21 },
		{ 187, 272, 21 },
		{ 187, 273, 21 },
		{ 187, 274, 21 },
		{ 187, 275, 21 },
		{ 187, 276, 21 },
		{ 187, 277, 21 },
		{ 187, 278, 21 },
		{ 187, 279, 21 },
		{ 187, 280, 21 },
		{ 187, 281, 21 },
		{ 187, 282, 21 },
		{ 187, 283, 21 },
		{ 187, 284, 21 },
		{ 187, 285, 21 },
		{ 187, 286, 21 },
		{ 187, 287, 21 },
		{ 187, 288, 21 },
		{ 187, 289, 21 },
		{ 187, 290, 21 },
		{ 187, 291, 21 },
		{ 188, 258, 21 },
		{ 188, 259, 21 },
		{ 188, 260, 21 },
		{ 188, 261, 21 },
		{ 188, 262, 21 },
		{ 188, 263, 21 },
		{ 188, 264, 21 },
		{ 188, 265, 21 },
		{ 188, 266, 21 },
		{ 188, 267, 21 },
		{ 188, 268, 21 },
		{ 188, 269, 21 },
		{ 188, 270, 21 },
		{ 188, 271, 21 },
		{ 188, 272, 21 },
		{ 188, 273, 21 },
		{ 188, 274, 21 },
		{ 188, 275, 21 },
		{ 188, 276, 21 },
		{ 188, 277, 21 },
		{ 188, 278, 21 },
		{ 188, 279, 21 },
		{ 188, 280, 21 },
		{ 188, 281, 21 },
		{ 188, 282, 21 },
		{ 188, 283, 21 },
		{ 188, 284, 21 },
		{ 188, 285, 21 },
		{ 188, 286, 21 },
		{ 188, 287, 21 },
		{ 188, 288, 21 },
		{ 188, 289, 21 },
		{ 188, 290, 21 },
		{ 188, 291, 21 },
		{ 189, 258, 21 },
		{ 189, 259, 21 },
		{ 189, 260, 21 },
		{ 189, 261, 21 },
		{ 189, 262, 21 },
		{ 189, 263, 21 },
		{ 189, 264, 21 },
		{ 189, 265, 21 },
		{ 189, 266, 21 },
		{ 189, 267, 21 },
		{ 189, 268, 21 },
		{ 189, 269, 21 },
		{ 189, 270, 21 },
		{ 189, 271, 21 },
		{ 189, 272, 21 },
		{ 189, 273, 21 },
		{ 189, 274, 21 },
		{ 189, 275, 21 },
		{ 189, 276, 21 },
		{ 189, 277, 21 },
		{ 189, 278, 21 },
		{ 189, 279, 21 },
		{ 189, 280, 21 },
		{ 189, 281, 21 },
		{ 189, 282, 21 },
		{ 189, 283, 21 },
		{ 189, 284, 21 },
		{ 189, 285, 21 },
		{ 189, 286, 21 },
		{ 189, 287, 21 },
		{ 189, 288, 21 },
		{ 189, 289, 21 },
		{ 189, 290, 21 },
		{ 190, 258, 21 },
		{ 190, 259, 21 },
		{ 190, 260, 21 },
		{ 190, 261, 21 },
		{ 190, 262, 21 },
		{ 190, 263, 21 },
		{ 190, 264, 21 },
		{ 190, 265, 21 },
		{ 190, 266, 21 },
		{ 190, 267, 21 },
		{ 190, 268, 21 },
		{ 190, 269, 21 },
		{ 190, 270, 21 },
		{ 190, 271, 21 },
		{ 190, 272, 21 },
		{ 190, 273, 21 },
		{ 190, 274, 21 },
		{ 190, 275, 21 },
		{ 190, 276, 21 },
		{ 190, 277, 21 },
		{ 190, 278, 21 },
		{ 190, 279, 21 },
		{ 190, 280, 21 },
		{ 190, 281, 21 },
		{ 190, 282, 21 },
		{ 190, 283, 21 },
		{ 190, 284, 21 },
		{ 190, 285, 21 },
		{ 190, 286, 21 },
		{ 190, 287, 21 },
		{ 190, 288, 21 },
		{ 190, 289, 21 },
		{ 190, 290, 21 },
		{ 191, 258, 21 },
		{ 191, 259, 21 },
		{ 191, 260, 21 },
		{ 191, 261, 21 },
		{ 191, 262, 21 },
		{ 191, 263, 21 },
		{ 191, 264, 21 },
		{ 191, 265, 21 },
		{ 191, 266, 21 },
		{ 191, 267, 21 },
		{ 191, 268, 21 },
		{ 191, 269, 21 },
		{ 191, 270, 21 },
		{ 191, 271, 21 },
		{ 191, 272, 21 },
		{ 191, 273, 21 },
		{ 191, 274, 21 },
		{ 191, 275, 21 },
		{ 191, 276, 21 },
		{ 191, 277, 21 },
		{ 191, 278, 21 },
		{ 191, 279, 21 },
		{ 191, 280, 21 },
		{ 191, 281, 21 },
		{ 191, 282, 21 },
		{ 191, 283, 21 },
		{ 191, 284, 21 },
		{ 191, 285, 21 },
		{ 191, 286, 21 },
		{ 191, 287, 21 },
		{ 191, 288, 21 },
		{ 191, 289, 21 },
		{ 192, 258, 21 },
		{ 192, 259, 21 },
		{ 192, 260, 21 },
		{ 192, 261, 21 },
		{ 192, 262, 21 },
		{ 192, 263, 21 },
		{ 192, 264, 21 },
		{ 192, 265, 21 },
		{ 192, 266, 21 },
		{ 192, 267, 21 },
		{ 192, 268, 21 },
		{ 192, 269, 21 },
		{ 192, 270, 21 },
		{ 192, 271, 21 },
		{ 192, 272, 21 },
		{ 192, 273, 21 },
		{ 192, 274, 21 },
		{ 192, 275, 21 },
		{ 192, 276, 21 },
		{ 192, 277, 21 },
		{ 192, 278, 21 },
		{ 192, 279, 21 },
		{ 192, 280, 21 },
		{ 192, 281, 21 },
		{ 192, 282, 21 },
		{ 192, 283, 21 },
		{ 192, 284, 21 },
		{ 192, 285, 21 },
		{ 192, 286, 21 },
		{ 192, 287, 21 },
		{ 192, 288, 21 },
		{ 192, 289, 21 },
		{ 193, 258, 21 },
		{ 193, 259, 21 },
		{ 193, 260, 21 },
		{ 193, 261, 21 },
		{ 193, 262, 21 },
		{ 193, 263, 21 },
		{ 193, 264, 21 },
		{ 193, 265, 21 },
		{ 193, 266, 21 },
		{ 193, 267, 21 },
		{ 193, 268, 21 },
		{ 193, 269, 21 },
		{ 193, 270, 21 },
		{ 193, 271, 21 },
		{ 193, 272, 21 },
		{ 193, 273, 21 },
		{ 193, 274, 21 },
		{ 193, 275, 21 },
		{ 193, 276, 21 },
		{ 193, 277, 21 },
		{ 193, 278, 21 },
		{ 193, 279, 21 },
		{ 193, 280, 21 },
		{ 193, 281, 21 },
		{ 193, 282, 21 },
		{ 193, 283, 21 },
		{ 193, 284, 21 },
		{ 193, 285, 21 },
		{ 193, 286, 21 },
		{ 193, 287, 21 },
		{ 193, 288, 21 },
		{ 194, 258, 21 },
		{ 194, 259, 21 },
		{ 194, 260, 21 },
		{ 194, 261, 21 },
		{ 194, 262, 21 },
		{ 194, 263, 21 },
		{ 194, 264, 21 },
		{ 194, 265, 21 },
		{ 194, 266, 21 },
		{ 194, 267, 21 },
		{ 194, 268, 21 },
		{ 194, 269, 21 },
		{ 194, 270, 21 },
		{ 194, 271, 21 },
		{ 194, 272, 21 },
		{ 194, 273, 21 },
		{ 194, 274, 21 },
		{ 194, 275, 21 },
		{ 194, 276, 21 },
		{ 194, 277, 21 },
		{ 194, 278, 21 },
		{ 194, 279, 21 },
		{ 194, 280, 21 },
		{ 194, 281, 21 },
		{ 194, 282, 21 },
		{ 194, 283, 21 },
		{ 194, 284, 21 },
		{ 194, 285, 21 },
		{ 194, 286, 21 },
		{ 194, 287, 21 },
		{ 194, 288, 21 },
		{ 195, 258, 21 },
		{ 195, 259, 21 },
		{ 195, 260, 21 },
		{ 195, 261, 21 },
		{ 195, 262, 21 },
		{ 195, 263, 21 },
		{ 195, 264, 21 },
		{ 195, 265, 21 },
		{ 195, 266, 21 },
		{ 195, 267, 21 },
		{ 195, 268, 21 },
		{ 195, 269, 21 },
		{ 195, 270, 21 },
		{ 195, 271, 21 },
		{ 195, 272, 21 },
		{ 195, 273, 21 },
		{ 195, 274, 21 },
		{ 195, 275, 21 },
		{ 195, 276, 21 },
		{ 195, 277, 21 },
		{ 195, 278, 21 },
		{ 195, 279, 21 },
		{ 195, 280, 21 },
		{ 195, 281, 21 },
		{ 195, 282, 21 },
		{ 195, 283, 21 },
		{ 195, 284, 21 },
		{ 195, 285, 21 },
		{ 195, 286, 21 },
		{ 195, 287, 21 },
		{ 196, 258, 21 },
		{ 196, 259, 21 },
		{ 196, 260, 21 },
		{ 196, 261, 21 },
		{ 196, 262, 21 },
		{ 196, 263, 21 },
		{ 196, 264, 21 },
		{ 196, 265, 21 },
		{ 196, 266, 21 },
		{ 196, 267, 21 },
		{ 196, 268, 21 },
		{ 196, 269, 21 },
		{ 196, 270, 21 },
		{ 196, 271, 21 },
		{ 196, 272, 21 },
		{ 196, 273, 21 },
		{ 196, 274, 21 },
		{ 196, 275, 21 },
		{ 196, 276, 21 },
		{ 196, 277, 21 },
		{ 196, 278, 21 },
		{ 196, 279, 21 },
		{ 196, 280, 21 },
		{ 196, 281, 21 },
		{ 196, 282, 21 },
		{ 196, 283, 21 },
		{ 196, 284, 21 },
		{ 196, 285, 21 },
		{ 196, 286, 21 },
		{ 196, 287, 21 },
		{ 197, 258, 21 },
		{ 197, 259, 21 },
		{ 197, 260, 21 },
		{ 197, 261, 21 },
		{ 197, 262, 21 },
		{ 197, 263, 21 },
		{ 197, 264, 21 },
		{ 197, 265, 21 },
		{ 197, 266, 21 },
		{ 197, 267, 21 },
		{ 197, 268, 21 },
		{ 197, 269, 21 },
		{ 197, 270, 21 },
		{ 197, 271, 21 },
		{ 197, 272, 21 },
		{ 197, 273, 21 },
		{ 197, 274, 21 },
		{ 197, 275, 21 },
		{ 197, 276, 21 },
		{ 197, 277, 21 },
		{ 197, 278, 21 },
		{ 197, 279, 21 },
		{ 197, 280, 21 },
		{ 197, 281, 21 },
		{ 197, 282, 21 },
		{ 197, 283, 21 },
		{ 197, 284, 21 },
		{ 197, 285, 21 },
		{ 197, 286, 21 },
		{ 198, 258, 21 },
		{ 198, 259, 21 },
		{ 198, 260, 21 },
		{ 198, 261, 21 },
		{ 198, 262, 21 },
		{ 198, 263, 21 },
		{ 198, 264, 21 },
		{ 198, 265, 21 },
		{ 198, 266, 21 },
		{ 198, 267, 21 },
		{ 198, 268, 21 },
		{ 198, 269, 21 },
		{ 198, 270, 21 },
		{ 198, 271, 21 },
		{ 198, 272, 21 },
		{ 198, 273, 21 },
		{ 198, 274, 21 },
		{ 198, 275, 21 },
		{ 198, 276, 21 },
		{ 198, 277, 21 },
		{ 198, 278, 21 },
		{ 198, 279, 21 },
		{ 198, 280, 21 },
		{ 198, 281, 21 },
		{ 198, 282, 21 },
		{ 198, 283, 21 },
		{ 198, 284, 21 },
		{ 198, 285, 21 },
		{ 198, 286, 21 },
		{ 199, 258, 21 },
		{ 199, 259, 21 },
		{ 199, 260, 21 },
		{ 199, 261, 21 },
		{ 199, 262, 21 },
		{ 199, 263, 21 },
		{ 199, 264, 21 },
		{ 199, 265, 21 },
		{ 199, 266, 21 },
		{ 199, 267, 21 },
		{ 199, 268, 21 },
		{ 199, 269, 21 },
		{ 199, 270, 21 },
		{ 199, 271, 21 },
		{ 199, 272, 21 },
		{ 199, 273, 21 },
		{ 199, 274, 21 },
		{ 199, 275, 21 },
		{ 199, 276, 21 },
		{ 199, 277, 21 },
		{ 199, 278, 21 },
		{ 199, 279, 21 },
		{ 199, 280, 21 },
		{ 199, 281, 21 },
		{ 199, 282, 21 },
		{ 199, 283, 21 },
		{ 199, 284, 21 },
		{ 199, 285, 21 },
		{ 200, 258, 21 },
		{ 200, 259, 21 },
		{ 200, 260, 21 },
		{ 200, 261, 21 },
		{ 200, 262, 21 },
		{ 200, 263, 21 },
		{ 200, 264, 21 },
		{ 200, 265, 21 },
		{ 200, 266, 21 },
		{ 200, 267, 21 },
		{ 200, 268, 21 },
		{ 200, 269, 21 },
		{ 200, 270, 21 },
		{ 200, 271, 21 },
		{ 200, 272, 21 },
		{ 200, 273, 21 },
		{ 200, 274, 21 },
		{ 200, 275, 21 },
		{ 200, 276, 21 },
		{ 200, 277, 21 },
		{ 200, 278, 21 },
		{ 200, 279, 21 },
		{ 200, 280, 21 },
		{ 200, 281, 21 },
		{ 200, 282, 21 },
		{ 200, 283, 21 },
		{ 200, 284, 21 },
		{ 200, 285, 21 },
		{ 201, 258, 21 },
		{ 201, 259, 21 },
		{ 201, 260, 21 },
		{ 201, 261, 21 },
		{ 201, 262, 21 },
		{ 201, 263, 21 },
		{ 201, 264, 21 },
		{ 201, 265, 21 },
		{ 201, 266, 21 },
		{ 201, 267, 21 },
		{ 201, 268, 21 },
		{ 201, 269, 21 },
		{ 201, 270, 21 },
		{ 201, 271, 21 },
		{ 201, 272, 21 },
		{ 201, 273, 21 },
		{ 201, 274, 21 },
		{ 201, 275, 21 },
		{ 201, 276, 21 },
		{ 201, 277, 21 },
		{ 201, 278, 21 },
		{ 201, 279, 21 },
		{ 201, 280, 21 },
		{ 201, 281, 21 },
		{ 201, 282, 21 },
		{ 201, 283, 21 },
		{ 201, 284, 21 },
		{ 202, 258, 21 },
		{ 202, 259, 21 },
		{ 202, 260, 21 },
		{ 202, 261, 21 },
		{ 202, 262, 21 },
		{ 202, 263, 21 },
		{ 202, 264, 21 },
		{ 202, 265, 21 },
		{ 202, 266, 21 },
		{ 202, 267, 21 },
		{ 202, 268, 21 },
		{ 202, 269, 21 },
		{ 202, 270, 21 },
		{ 202, 271, 21 },
		{ 202, 272, 21 },
		{ 202, 273, 21 },
		{ 202, 274, 21 },
		{ 202, 275, 21 },
		{ 202, 276, 21 },
		{ 202, 277, 21 },
		{ 202, 278, 21 },
		{ 202, 279, 21 },
		{ 202, 280, 21 },
		{ 202, 281, 21 },
		{ 202, 282, 21 },
		{ 202, 283, 21 },
		{ 202, 284, 21 },
		{ 203, 258, 21 },
		{ 203, 259, 21 },
		{ 203, 260, 21 },
		{ 203, 261, 21 },
		{ 203, 262, 21 },
		{ 203, 263, 21 },
		{ 203, 264, 21 },
		{ 203, 265, 21 },
		{ 203, 266, 21 },
		{ 203, 267, 21 },
		{ 203, 268, 21 },
		{ 203, 269, 21 },
		{ 203, 270, 21 },
		{ 203, 271, 21 },
		{ 203, 272, 21 },
		{ 203, 273, 21 },
		{ 203, 274, 21 },
		{ 203, 275, 21 },
		{ 203, 276, 21 },
		{ 203, 277, 21 },
		{ 203, 278, 21 },
		{ 203, 279, 21 },
		{ 203, 280, 21 },
		{ 203, 281, 21 },
		{ 203, 282, 21 },
		{ 203, 283, 21 },
		{ 204, 258, 21 },
		{ 204, 259, 21 },
		{ 204, 260, 21 },
		{ 204, 261, 21 },
		{ 204, 262, 21 },
		{ 204, 263, 21 },
		{ 204, 264, 21 },
		{ 204, 265, 21 },
		{ 204, 266, 21 },
		{ 204, 267, 21 },
		{ 204, 268, 21 },
		{ 204, 269, 21 },
		{ 204, 270, 21 },
		{ 204, 271, 21 },
		{ 204, 272, 21 },
		{ 204, 273, 21 },
		{ 204, 274, 21 },
		{ 204, 275, 21 },
		{ 204, 276, 21 },
		{ 204, 277, 21 },
		{ 204, 278, 21 },
		{ 204, 279, 21 },
		{ 204, 280, 21 },
		{ 204, 281, 21 },
		{ 204, 282, 21 },
		{ 204, 283, 21 },
		{ 205, 258, 21 },
		{ 205, 259, 21 },
		{ 205, 260, 21 },
		{ 205, 261, 21 },
		{ 205, 262, 21 },
		{ 205, 263, 21 },
		{ 205, 264, 21 },
		{ 205, 265, 21 },
		{ 205, 266, 21 },
		{ 205, 267, 21 },
		{ 205, 268, 21 },
		{ 205, 269, 21 },
		{ 205, 270, 21 },
		{ 205, 271, 21 },
		{ 205, 272, 21 },
		{ 205, 273, 21 },
		{ 205, 274, 21 },
		{ 205, 275, 21 },
		{ 205, 276, 21 },
		{ 205, 277, 21 },
		{ 205, 278, 21 },
		{ 205, 279, 21 },
		{ 205, 280, 21 },
		{ 205, 281, 21 },
		{ 205, 282, 21 },
		{ 206, 258, 21 },
		{ 206, 259, 21 },
		{ 206, 260, 21 },
		{ 206, 261, 21 },
		{ 206, 262, 21 },
		{ 206, 263, 21 },
		{ 206, 264, 21 },
		{ 206, 265, 21 },
		{ 206, 266, 21 },
		{ 206, 267, 21 },
		{ 206, 268, 21 },
		{ 206, 269, 21 },
		{ 206, 270, 21 },
		{ 206, 271, 21 },
		{ 206, 272, 21 },
		{ 206, 273, 21 },
		{ 206, 274, 21 },
		{ 206, 275, 21 },
		{ 206, 276, 21 },
		{ 206, 277, 21 },
		{ 206, 278, 21 },
		{ 206, 279, 21 },
		{ 206, 280, 21 },
		{ 206, 281, 21 },
		{ 206, 282, 21 },
		{ 207, 258, 21 },
		{ 207, 259, 21 },
		{ 207, 260, 21 },
		{ 207, 261, 21 },
		{ 207, 262, 21 },
		{ 207, 263, 21 },
		{ 207, 264, 21 },
		{ 207, 265, 21 },
		{ 207, 266, 21 },
		{ 207, 267, 21 },
		{ 207, 268, 21 },
		{ 207, 269, 21 },
		{ 207, 270, 21 },
		{ 207, 271, 21 },
		{ 207, 272, 21 },
		{ 207, 273, 21 },
		{ 207, 274, 21 },
		{ 207, 275, 21 },
		{ 207, 276, 21 },
		{ 207, 277, 21 },
		{ 207, 278, 21 },
		{ 207, 279, 21 },
		{ 207, 280, 21 },
		{ 207, 281, 21 },
		{ 208, 258, 21 },
		{ 208, 259, 21 },
		{ 208, 260, 21 },
		{ 208, 261, 21 },
		{ 208, 262, 21 },
		{ 208, 263, 21 },
		{ 208, 264, 21 },
		{ 208, 265, 21 },
		{ 208, 266, 21 },
		{ 208, 267, 21 },
		{ 208, 268, 21 },
		{ 208, 269, 21 },
		{ 208, 270, 21 },
		{ 208, 271, 21 },
		{ 208, 272, 21 },
		{ 208, 273, 21 },
		{ 208, 274, 21 },
		{ 208, 275, 21 },
		{ 208, 276, 21 },
		{ 208, 277, 21 },
		{ 208, 278, 21 },
		{ 208, 279, 21 },
		{ 208, 280, 21 },
		{ 208, 281, 21 },
		{ 209, 258, 21 },
		{ 209, 259, 21 },
		{ 209, 260, 21 },
		{ 209, 261, 21 },
		{ 209, 262, 21 },
		{ 209, 263, 21 },
		{ 209, 264, 21 },
		{ 209, 265, 21 },
		{ 209, 266, 21 },
		{ 209, 267, 21 },
		{ 209, 268, 21 },
		{ 209, 269, 21 },
		{ 209, 270, 21 },
		{ 209, 271, 21 },
		{ 209, 272, 21 },
		{ 209, 273, 21 },
		{ 209, 274, 21 },
		{ 209, 275, 21 },
		{ 209, 276, 21 },
		{ 209, 277, 21 },
		{ 209, 278, 21 },
		{ 209, 279, 21 },
		{ 209, 280, 21 },
		{ 210, 258, 21 },
		{ 210, 259, 21 },
		{ 210, 260, 21 },
		{ 210, 261, 21 },
		{ 210, 262, 21 },
		{ 210, 263, 21 },
		{ 210, 264, 21 },
		{ 210, 265, 21 },
		{ 210, 266, 21 },
		{ 210, 267, 21 },
		{ 210, 268, 21 },
		{ 210, 269, 21 },
		{ 210, 270, 21 },
		{ 210, 271, 21 },
		{ 210, 272, 21 },
		{ 210, 273, 21 },
		{ 210, 274, 21 },
		{ 210, 275, 21 },
		{ 210, 276, 21 },
		{ 210, 277, 21 },
		{ 210, 278, 21 },
		{ 210, 279, 21 },
		{ 210, 280, 21 },
		{ 211, 258, 21 },
		{ 211, 259, 21 },
		{ 211, 260, 21 },
		{ 211, 261, 21 },
		{ 211, 262, 21 },
		{ 211, 263, 21 },
		{ 211, 264, 21 },
		{ 211, 265, 21 },
		{ 211, 266, 21 },
		{ 211, 267, 21 },
		{ 211, 268, 21 },
		{ 211, 269, 21 },
		{ 211, 270, 21 },
		{ 211, 271, 21 },
		{ 211, 272, 21 },
		{ 211, 273, 21 },
		{ 211, 274, 21 },
		{ 211, 275, 21 },
		{ 211, 276, 21 },
		{ 211, 277, 21 },
		{ 211, 278, 21 },
		{ 211, 279, 21 },
		{ 212, 258, 21 },
		{ 212, 259, 21 },
		{ 212, 260, 21 },
		{ 212, 261, 21 },
		{ 212, 262, 21 },
		{ 212, 263, 21 },
		{ 212, 264, 21 },
		{ 212, 265, 21 },
		{ 212, 266, 21 },
		{ 212, 267, 21 },
		{ 212, 268, 21 },
		{ 212, 269, 21 },
		{ 212, 270, 21 },
		{ 212, 271, 21 },
		{ 212, 272, 21 },
		{ 212, 273, 21 },
		{ 212, 274, 21 },
		{ 212, 275, 21 },
		{ 212, 276, 21 },
		{ 212, 277, 21 },
		{ 212, 278, 21 },
		{ 212, 279, 21 },
		{ 213, 258, 21 },
		{ 213, 259, 21 },
		{ 213, 260, 21 },
		{ 213, 261, 21 },
		{ 213, 262, 21 },
		{ 213, 263, 21 },
		{ 213, 264, 21 },
		{ 213, 265, 21 },
		{ 213, 266, 21 },
		{ 213, 267, 21 },
		{ 213, 268, 21 },
		{ 213, 269, 21 },
		{ 213, 270, 21 },
		{ 213, 271, 21 },
		{ 213, 272, 21 },
		{ 213, 273, 21 },
		{ 213, 274, 21 },
		{ 213, 275, 21 },
		{ 213, 276, 21 },
		{ 213, 277, 21 },
		{ 213, 278, 21 },
		{ 214, 258, 21 },
		{ 214, 259, 21 },
		{ 214, 260, 21 },
		{ 214, 261, 21 },
		{ 214, 262, 21 },
		{ 214, 263, 21 },
		{ 214, 264, 21 },
		{ 214, 265, 21 },
		{ 214, 266, 21 },
		{ 214, 267, 21 },
		{ 214, 268, 21 },
		{ 214, 269, 21 },
		{ 214, 270, 21 },
		{ 214, 271, 21 },
		{ 214, 272, 21 },
		{ 214, 273, 21 },
		{ 214, 274, 21 },
		{ 214, 275, 21 },
		{ 214, 276, 21 },
		{ 214, 277, 21 },
		{ 214, 278, 21 },
		{ 215, 258, 21 },
		{ 215, 259, 21 },
		{ 215, 260, 21 },
		{ 215, 261, 21 },
		{ 215, 262, 21 },
		{ 215, 263, 21 },
		{ 215, 264, 21 },
		{ 215, 265, 21 },
		{ 215, 266, 21 },
		{ 215, 267, 21 },
		{ 215, 268, 21 },
		{ 215, 269, 21 },
		{ 215, 270, 21 },
		{ 215, 271, 21 },
		{ 215, 272, 21 },
		{ 215, 273, 21 },
		{ 215, 274, 21 },
		{ 215, 275, 21 },
		{ 215, 276, 21 },
		{ 215, 277, 21 },
		{ 216, 258, 21 },
		{ 216, 259, 21 },
		{ 216, 260, 21 },
		{ 216, 261, 21 },
		{ 216, 262, 21 },
		{ 216, 263, 21 },
		{ 216, 264, 21 },
		{ 216, 265, 21 },
		{ 216, 266, 21 },
		{ 216, 267, 21 },
		{ 216, 268, 21 },
		{ 216, 269, 21 },
		{ 216, 270, 21 },
		{ 216, 271, 21 },
		{ 216, 272, 21 },
		{ 216, 273, 21 },
		{ 216, 274, 21 },
		{ 216, 275, 21 },
		{ 216, 276, 21 },
		{ 216, 277, 21 },
		{ 217, 258, 21 },
		{ 217, 259, 21 },
		{ 217, 260, 21 },
		{ 217, 261, 21 },
		{ 217, 262, 21 },
		{ 217, 263, 21 },
		{ 217, 264, 21 },
		{ 217, 265, 21 },
		{ 217, 266, 21 },
		{ 217, 267, 21 },
		{ 217, 268, 21 },
		{ 217, 269, 21 },
		{ 217, 270, 21 },
		{ 217, 271, 21 },
		{ 217, 272, 21 },
		{ 217, 273, 21 },
		{ 217, 274, 21 },
		{ 217, 275, 21 },
		{ 217, 276, 21 },
		{ 218, 258, 21 },
		{ 218, 259, 21 },
		{ 218, 260, 21 },
		{ 218, 261, 21 },
		{ 218, 262, 21 },
		{ 218, 263, 21 },
		{ 218, 264, 21 },
		{ 218, 265, 21 },
		{ 218, 266, 21 },
		{ 218, 267, 21 },
		{ 218, 268, 21 },
		{ 218, 269, 21 },
		{ 218, 270, 21 },
		{ 218, 271, 21 },
		{ 218, 272, 21 },
		{ 218, 273, 21 },
		{ 218, 274, 21 },
		{ 218, 275, 21 },
		{ 218, 276, 21 },
		{ 219, 258, 21 },
		{ 219, 259, 21 },
		{ 219, 260, 21 },
		{ 219, 261, 21 },
		{ 219, 262, 21 },
		{ 219, 263, 21 },
		{ 219, 264, 21 },
		{ 219, 265, 21 },
		{ 219, 266, 21 },
		{ 219, 267, 21 },
		{ 219, 268, 21 },
		{ 219, 269, 21 },
		{ 219, 270, 21 },
		{ 219, 271, 21 },
		{ 219, 272, 21 },
		{ 219, 273, 21 },
		{ 219, 274, 21 },
		{ 219, 275, 21 },
		{ 220, 258, 21 },
		{ 220, 259, 21 },
		{ 220, 260, 21 },
		{ 220, 261, 21 },
		{ 220, 262, 21 },
		{ 220, 263, 21 },
		{ 220, 264, 21 },
		{ 220, 265, 21 },
		{ 220, 266, 21 },
		{ 220, 267, 21 },
		{ 220, 268, 21 },
		{ 220, 269, 21 },
		{ 220, 270, 21 },
		{ 220, 271, 21 },
		{ 220, 272, 21 },
		{ 220, 273, 21 },
		{ 220, 274, 21 },
		{ 220, 275, 21 },
		{ 221, 258, 21 },
		{ 221, 259, 21 },
		{ 221, 260, 21 },
		{ 221, 261, 21 },
		{ 221, 262, 21 },
		{ 221, 263, 21 },
		{ 221, 264, 21 },
		{ 221, 265, 21 },
		{ 221, 266, 21 },
		{ 221, 267, 21 },
		{ 221, 268, 21 },
		{ 221, 269, 21 },
		{ 221, 270, 21 },
		{ 221, 271, 21 },
		{ 221, 272, 21 },
		{ 221, 273, 21 },
		{ 221, 274, 21 },
		{ 222, 258, 21 },
		{ 222, 259, 21 },
		{ 222, 260, 21 },
		{ 222, 261, 21 },
		{ 222, 262, 21 },
		{ 222, 263, 21 },
		{ 222, 264, 21 },
		{ 222, 265, 21 },
		{ 222, 266, 21 },
		{ 222, 267, 21 },
		{ 222, 268, 21 },
		{ 222, 269, 21 },
		{ 222, 270, 21 },
		{ 222, 271, 21 },
		{ 222, 272, 21 },
		{ 222, 273, 21 },
		{ 222, 274, 21 },
		{ 223, 258, 21 },
		{ 223, 259, 21 },
		{ 223, 260, 21 },
		{ 223, 261, 21 },
		{ 223, 262, 21 },
		{ 223, 263, 21 },
		{ 223, 264, 21 },
		{ 223, 265, 21 },
		{ 223, 266, 21 },
		{ 223, 267, 21 },
		{ 223, 268, 21 },
		{ 223, 269, 21 },
		{ 223, 270, 21 },
		{ 223, 271, 21 },
		{ 223, 272, 21 },
		{ 223, 273, 21 },
		{ 224, 258, 21 },
		{ 224, 259, 21 },
		{ 224, 260, 21 },
		{ 224, 261, 21 },
		{ 224, 262, 21 },
		{ 224, 263, 21 },
		{ 224, 264, 21 },
		{ 224, 265, 21 },
		{ 224, 266, 21 },
		{ 224, 267, 21 },
		{ 224, 268, 21 },
		{ 224, 269, 21 },
		{ 224, 270, 21 },
		{ 224, 271, 21 },
		{ 224, 272, 21 },
		{ 224, 273, 21 },
		{ 225, 258, 21 },
		{ 225, 259, 21 },
		{ 225, 260, 21 },
		{ 225, 261, 21 },
		{ 225, 262, 21 },
		{ 225, 263, 21 },
		{ 225, 264, 21 },
		{ 225, 265, 21 },
		{ 225, 266, 21 },
		{ 225, 267, 21 },
		{ 225, 268, 21 },
		{ 225, 269, 21 },
		{ 225, 270, 21 },
		{ 225, 271, 21 },
		{ 225, 272, 21 },
		{ 226, 258, 21 },
		{ 226, 259, 21 },
		{ 226, 260, 21 },
		{ 226, 261, 21 },
		{ 226, 262, 21 },
		{ 226, 263, 21 },
		{ 226, 264, 21 },
		{ 226, 265, 21 },
		{ 226, 266, 21 },
		{ 226, 267, 21 },
		{ 226, 268, 21 },
		{ 226, 269, 21 },
		{ 226, 270, 21 },
		{ 226, 271, 21 },
		{ 226, 272, 21 },
		{ 227, 258, 21 },
		{ 227, 259, 21 },
		{ 227, 260, 21 },
		{ 227, 261, 21 },
		{ 227, 262, 21 },
		{ 227, 263, 21 },
		{ 227, 264, 21 },
		{ 227, 265, 21 },
		{ 227, 266, 21 },
		{ 227, 267, 21 },
		{ 227, 268, 21 },
		{ 227, 269, 21 },
		{ 227, 270, 21 },
		{ 227, 271, 21 },
		{ 228, 258, 21 },
		{ 228, 259, 21 },
		{ 228, 260, 21 },
		{ 228, 261, 21 },
		{ 228, 262, 21 },
		{ 228, 263, 21 },
		{ 228, 264, 21 },
		{ 228, 265, 21 },
		{ 228, 266, 21 },
		{ 228, 267, 21 },
		{ 228, 268, 21 },
		{ 228, 269, 21 },
		{ 228, 270, 21 },
		{ 228, 271, 21 },
		{ 229, 258, 21 },
		{ 229, 259, 21 },
		{ 229, 260, 21 },
		{ 229, 261, 21 },
		{ 229, 262, 21 },
		{ 229, 263, 21 },
		{ 229, 264, 21 },
		{ 229, 265, 21 },
		{ 229, 266, 21 },
		{ 229, 267, 21 },
		{ 229, 268, 21 },
		{ 229, 269, 21 },
		{ 229, 270, 21 },
		{ 230, 258, 21 },
		{ 230, 259, 21 },
		{ 230, 260, 21 },
		{ 230, 261, 21 },
		{ 230, 262, 21 },
		{ 230, 263, 21 },
		{ 230, 264, 21 },
		{ 230, 265, 21 },
		{ 230, 266, 21 },
		{ 230, 267, 21 },
		{ 230, 268, 21 },
		{ 230, 269, 21 },
		{ 230, 270, 21 },
		{ 231, 258, 21 },
		{ 231, 259, 21 },
		{ 231, 260, 21 },
		{ 231, 261, 21 },
		{ 231, 262, 21 },
		{ 231, 263, 21 },
		{ 231, 264, 21 },
		{ 231, 265, 21 },
		{ 231, 266, 21 },
		{ 231, 267, 21 },
		{ 231, 268, 21 },
		{ 231, 269, 21 },
		{ 232, 258, 21 },
		{ 232, 259, 21 },
		{ 232, 260, 21 },
		{ 232, 261, 21 },
		{ 232, 262, 21 },
		{ 232, 263, 21 },
		{ 232, 264, 21 },
		{ 232, 265, 21 },
		{ 232, 266, 21 },
		{ 232, 267, 21 },
		{ 232, 268, 21 },
		{ 232, 269, 21 },
		{ 233, 258, 21 },
		{ 233, 259, 21 },
		{ 233, 260, 21 },
		{ 233, 261, 21 },
		{ 233, 262, 21 },
		{ 233, 263, 21 },
		{ 233, 264, 21 },
		{ 233, 265, 21 },
		{ 233, 266, 21 },
		{ 233, 267, 21 },
		{ 233, 268, 21 },
		{ 234, 258, 21 },
		{ 234, 259, 21 },
		{ 234, 260, 21 },
		{ 234, 261, 21 },
		{ 234, 262, 21 },
		{ 234, 263, 21 },
		{ 234, 264, 21 },
		{ 234, 265, 21 },
		{ 234, 266, 21 },
		{ 234, 267, 21 },
		{ 234, 268, 21 },
		{ 235, 258, 21 },
		{ 235, 259, 21 },
		{ 235, 260, 21 },
		{ 235, 261, 21 },
		{ 235, 262, 21 },
		{ 235, 263, 21 },
		{ 235, 264, 21 },
		{ 235, 265, 21 },
		{ 235, 266, 21 },
		{ 235, 267, 21 },
		{ 236, 258, 21 },
		{ 236, 259, 21 },
		{ 236, 260, 21 },
		{ 236, 261, 21 },
		{ 236, 262, 21 },
		{ 236, 263, 21 },
		{ 236, 264, 21 },
		{ 236, 265, 21 },
		{ 236, 266, 21 },
		{ 236, 267, 21 },
		{ 237, 258, 21 },
		{ 237, 259, 21 },
		{ 237, 260, 21 },
		{ 237, 261, 21 },
		{ 237, 262, 21 },
		{ 237, 263, 21 },
		{ 237, 264, 21 },
		{ 237, 265, 21 },
		{ 237, 266, 21 },
		{ 238, 258, 21 },
		{ 238, 259, 21 },
		{ 238, 260, 21 },
		{ 238, 261, 21 },
		{ 238, 262, 21 },
		{ 238, 263, 21 },
		{ 238, 264, 21 },
		{ 238, 265, 21 },
		{ 238, 266, 21 },
		{ 239, 258, 21 },
		{ 239, 259, 21 },
		{ 239, 260, 21 },
		{ 239, 261, 21 },
		{ 239, 262, 21 },
		{ 239, 263, 21 },
		{ 239, 264, 21 },
		{ 239, 265, 21 },
		{ 240, 258, 21 },
		{ 240, 259, 21 },
		{ 240, 260, 21 },
		{ 240, 261, 21 },
		{ 240, 262, 21 },
		{ 240, 263, 21 },
		{ 240, 264, 21 },
		{ 240, 265, 21 },
		{ 274, 249, 21 },
		{ 274, 250, 21 },
		{ 274, 251, 21 },
		{ 274, 252, 21 },
		{ 274, 253, 21 },
		{ 274, 254, 21 },
		{ 274, 255, 21 },
		{ 274, 256, 21 },
		{ 275, 249, 21 },
		{ 275, 250, 21 },
		{ 275, 251, 21 },
		{ 275, 252, 21 },
		{ 275, 253, 21 },
		{ 275, 254, 21 },
		{ 275, 255, 21 },
		{ 275, 256, 21 },
		{ 276, 248, 21 },
		{ 276, 249, 21 },
		{ 276, 250, 21 },
		{ 276, 251, 21 },
		{ 276, 252, 21 },
		{ 276, 253, 21 },
		{ 276, 254, 21 },
		{ 276, 255, 21 },
		{ 276, 256, 21 },
		{ 277, 248, 21 },
		{ 277, 249, 21 },
		{ 277, 250, 21 },
		{ 277, 251, 21 },
		{ 277, 252, 21 },
		{ 277, 253, 21 },
		{ 277, 254, 21 },
		{ 277, 255, 21 },
		{ 277, 256, 21 },
		{ 278, 247, 21 },
		{ 278, 248, 21 },
		{ 278, 249, 21 },
		{ 278, 250, 21 },
		{ 278, 251, 21 },
		{ 278, 252, 21 },
		{ 278, 253, 21 },
		{ 278, 254, 21 },
		{ 278, 255, 21 },
		{ 278, 256, 21 },
		{ 279, 247, 21 },
		{ 279, 248, 21 },
		{ 279, 249, 21 },
		{ 279, 250, 21 },
		{ 279, 251, 21 },
		{ 279, 252, 21 },
		{ 279, 253, 21 },
		{ 279, 254, 21 },
		{ 279, 255, 21 },
		{ 279, 256, 21 },
		{ 280, 246, 21 },
		{ 280, 247, 21 },
		{ 280, 248, 21 },
		{ 280, 249, 21 },
		{ 280, 250, 21 },
		{ 280, 251, 21 },
		{ 280, 252, 21 },
		{ 280, 253, 21 },
		{ 280, 254, 21 },
		{ 280, 255, 21 },
		{ 280, 256, 21 },
		{ 281, 246, 21 },
		{ 281, 247, 21 },
		{ 281, 248, 21 },
		{ 281, 249, 21 },
		{ 281, 250, 21 },
		{ 281, 251, 21 },
		{ 281, 252, 21 },
		{ 281, 253, 21 },
		{ 281, 254, 21 },
		{ 281, 255, 21 },
		{ 281, 256, 21 },
		{ 282, 245, 21 },
		{ 282, 246, 21 },
		{ 282, 247, 21 },
		{ 282, 248, 21 },
		{ 282, 249, 21 },
		{ 282, 250, 21 },
		{ 282, 251, 21 },
		{ 282, 252, 21 },
		{ 282, 253, 21 },
		{ 282, 254, 21 },
		{ 282, 255, 21 },
		{ 282, 256, 21 },
		{ 283, 245, 21 },
		{ 283, 246, 21 },
		{ 283, 247, 21 },
		{ 283, 248, 21 },
		{ 283, 249, 21 },
		{ 283, 250, 21 },
		{ 283, 251, 21 },
		{ 283, 252, 21 },
		{ 283, 253, 21 },
		{ 283, 254, 21 },
		{ 283, 255, 21 },
		{ 283, 256, 21 },
		{ 284, 244, 21 },
		{ 284, 245, 21 },
		{ 284, 246, 21 },
		{ 284, 247, 21 },
		{ 284, 248, 21 },
		{ 284, 249, 21 },
		{ 284, 250, 21 },
		{ 284, 251, 21 },
		{ 284, 252, 21 },
		{ 284, 253, 21 },
		{ 284, 254, 21 },
		{ 284, 255, 21 },
		{ 284, 256, 21 },
		{ 285, 244, 21 },
		{ 285, 245, 21 },
		{ 285, 246, 21 },
		{ 285, 247, 21 },
		{ 285, 248, 21 },
		{ 285, 249, 21 },
		{ 285, 250, 21 },
		{ 285, 251, 21 },
		{ 285, 252, 21 },
		{ 285, 253, 21 },
		{ 285, 254, 21 },
		{ 285, 255, 21 },
		{ 285, 256, 21 },
		{ 286, 243, 21 },
		{ 286, 244, 21 },
		{ 286, 245, 21 },
		{ 286, 246, 21 },
		{ 286, 247, 21 },
		{ 286, 248, 21 },
		{ 286, 249, 21 },
		{ 286, 250, 21 },
		{ 286, 251, 21 },
		{ 286, 252, 21 },
		{ 286, 253, 21 },
		{ 286, 254, 21 },
		{ 286, 255, 21 },
		{ 286, 256, 21 },
		{ 287, 243, 21 },
		{ 287, 244, 21 },
		{ 287, 245, 21 },
		{ 287, 246, 21 },
		{ 287, 247, 21 },
		{ 287, 248, 21 },
		{ 287, 249, 21 },
		{ 287, 250, 21 },
		{ 287, 251, 21 },
		{ 287, 252, 21 },
		{ 287, 253, 21 },
		{ 287, 254, 21 },
		{ 287, 255, 21 },
		{ 287, 256, 21 },
		{ 288, 242, 21 },
		{ 288, 243, 21 },
		{ 288, 244, 21 },
		{ 288, 245, 21 },
		{ 288, 246, 21 },
		{ 288, 247, 21 },
		{ 288, 248, 21 },
		{ 288, 249, 21 },
		{ 288, 250, 21 },
		{ 288, 251, 21 },
		{ 288, 252, 21 },
		{ 288, 253, 21 },
		{ 288, 254, 21 },
		{ 288, 255, 21 },
		{ 288, 256, 21 },
		{ 289, 242, 21 },
		{ 289, 243, 21 },
		{ 289, 244, 21 },
		{ 289, 245, 21 },
		{ 289, 246, 21 },
		{ 289, 247, 21 },
		{ 289, 248, 21 },
		{ 289, 249, 21 },
		{ 289, 250, 21 },
		{ 289, 251, 21 },
		{ 289, 252, 21 },
		{ 289, 253, 21 },
		{ 289, 254, 21 },
		{ 289, 255, 21 },
		{ 289, 256, 21 },
		{ 290, 241, 21 },
		{ 290, 242, 21 },
		{ 290, 243, 21 },
		{ 290, 244, 21 },
		{ 290, 245, 21 },
		{ 290, 246, 21 },
		{ 290, 247, 21 },
		{ 290, 248, 21 },
		{ 290, 249, 21 },
		{ 290, 250, 21 },
		{ 290, 251, 21 },
		{ 290, 252, 21 },
		{ 290, 253, 21 },
		{ 290, 254, 21 },
		{ 290, 255, 21 },
		{ 290, 256, 21 },
		{ 291, 241, 21 },
		{ 291, 242, 21 },
		{ 291, 243, 21 },
		{ 291, 244, 21 },
		{ 291, 245, 21 },
		{ 291, 246, 21 },
		{ 291, 247, 21 },
		{ 291, 248, 21 },
		{ 291, 249, 21 },
		{ 291, 250, 21 },
		{ 291, 251, 21 },
		{ 291, 252, 21 },
		{ 291, 253, 21 },
		{ 291, 254, 21 },
		{ 291, 255, 21 },
		{ 291, 256, 21 },
		{ 292, 240, 21 },
		{ 292, 241, 21 },
		{ 292, 242, 21 },
		{ 292, 243, 21 },
		{ 292, 244, 21 },
		{ 292, 245, 21 },
		{ 292, 246, 21 },
		{ 292, 247, 21 },
		{ 292, 248, 21 },
		{ 292, 249, 21 },
		{ 292, 250, 21 },
		{ 292, 251, 21 },
		{ 292, 252, 21 },
		{ 292, 253, 21 },
		{ 292, 254, 21 },
		{ 292, 255, 21 },
		{ 292, 256, 21 },
		{ 293, 240, 21 },
		{ 293, 241, 21 },
		{ 293, 242, 21 },
		{ 293, 243, 21 },
		{ 293, 244, 21 },
		{ 293, 245, 21 },
		{ 293, 246, 21 },
		{ 293, 247, 21 },
		{ 293, 248, 21 },
		{ 293, 249, 21 },
		{ 293, 250, 21 },
		{ 293, 251, 21 },
		{ 293, 252, 21 },
		{ 293, 253, 21 },
		{ 293, 254, 21 },
		{ 293, 255, 21 },
		{ 293, 256, 21 },
		{ 294, 239, 21 },
		{ 294, 240, 21 },
		{ 294, 241, 21 },
		{ 294, 242, 21 },
		{ 294, 243, 21 },
		{ 294, 244, 21 },
		{ 294, 245, 21 },
		{ 294, 246, 21 },
		{ 294, 247, 21 },
		{ 294, 248, 21 },
		{ 294, 249, 21 },
		{ 294, 250, 21 },
		{ 294, 251, 21 },
		{ 294, 252, 21 },
		{ 294, 253, 21 },
		{ 294, 254, 21 },
		{ 294, 255, 21 },
		{ 294, 256, 21 },
		{ 295, 239, 21 },
		{ 295, 240, 21 },
		{ 295, 241, 21 },
		{ 295, 242, 21 },
		{ 295, 243, 21 },
		{ 295, 244, 21 },
		{ 295, 245, 21 },
		{ 295, 246, 21 },
		{ 295, 247, 21 },
		{ 295, 248, 21 },
		{ 295, 249, 21 },
		{ 295, 250, 21 },
		{ 295, 251, 21 },
		{ 295, 252, 21 },
		{ 295, 253, 21 },
		{ 295, 254, 21 },
		{ 295, 255, 21 },
		{ 295, 256, 21 },
		{ 296, 238, 21 },
		{ 296, 239, 21 },
		{ 296, 240, 21 },
		{ 296, 241, 21 },
		{ 296, 242, 21 },
		{ 296, 243, 21 },
		{ 296, 244, 21 },
		{ 296, 245, 21 },
		{ 296, 246, 21 },
		{ 296, 247, 21 },
		{ 296, 248, 21 },
		{ 296, 249, 21 },
		{ 296, 250, 21 },
		{ 296, 251, 21 },
		{ 296, 252, 21 },
		{ 296, 253, 21 },
		{ 296, 254, 21 },
		{ 296, 255, 21 },
		{ 296, 256, 21 },
		{ 297, 238, 21 },
		{ 297, 239, 21 },
		{ 297, 240, 21 },
		{ 297, 241, 21 },
		{ 297, 242, 21 },
		{ 297, 243, 21 },
		{ 297, 244, 21 },
		{ 297, 245, 21 },
		{ 297, 246, 21 },
		{ 297, 247, 21 },
		{ 297, 248, 21 },
		{ 297, 249, 21 },
		{ 297, 250, 21 },
		{ 297, 251, 21 },
		{ 297, 252, 21 },
		{ 297, 253, 21 },
		{ 297, 254, 21 },
		{ 297, 255, 21 },
		{ 297, 256, 21 },
		{ 298, 237, 21 },
		{ 298, 238, 21 },
		{ 298, 239, 21 },
		{ 298, 240, 21 },
		{ 298, 241, 21 },
		{ 298, 242, 21 },
		{ 298, 243, 21 },
		{ 298, 244, 21 },
		{ 298, 245, 21 },
		{ 298, 246, 21 },
		{ 298, 247, 21 },
		{ 298, 248, 21 },
		{ 298, 249, 21 },
		{ 298, 250, 21 },
		{ 298, 251, 21 },
		{ 298, 252, 21 },
		{ 298, 253, 21 },
		{ 298, 254, 21 },
		{ 298, 255, 21 },
		{ 298, 256, 21 },
		{ 299, 237, 21 },
		{ 299, 238, 21 },
		{ 299, 239, 21 },
		{ 299, 240, 21 },
		{ 299, 241, 21 },
		{ 299, 242, 21 },
		{ 299, 243, 21 },
		{ 299, 244, 21 },
		{ 299, 245, 21 },
		{ 299, 246, 21 },
		{ 299, 247, 21 },
		{ 299, 248, 21 },
		{ 299, 249, 21 },
		{ 299, 250, 21 },
		{ 299, 251, 21 },
		{ 299, 252, 21 },
		{ 299, 253, 21 },
		{ 299, 254, 21 },
		{ 299, 255, 21 },
		{ 299, 256, 21 },
		{ 300, 236, 21 },
		{ 300, 237, 21 },
		{ 300, 238, 21 },
		{ 300, 239, 21 },
		{ 300, 240, 21 },
		{ 300, 241, 21 },
		{ 300, 242, 21 },
		{ 300, 243, 21 },
		{ 300, 244, 21 },
		{ 300, 245, 21 },
		{ 300, 246, 21 },
		{ 300, 247, 21 },
		{ 300, 248, 21 },
		{ 300, 249, 21 },
		{ 300, 250, 21 },
		{ 300, 251, 21 },
		{ 300, 252, 21 },
		{ 300, 253, 21 },
		{ 300, 254, 21 },
		{ 300, 255, 21 },
		{ 300, 256, 21 },
		{ 301, 236, 21 },
		{ 301, 237, 21 },
		{ 301, 238, 21 },
		{ 301, 239, 21 },
		{ 301, 240, 21 },
		{ 301, 241, 21 },
		{ 301, 242, 21 },
		{ 301, 243, 21 },
		{ 301, 244, 21 },
		{ 301, 245, 21 },
		{ 301, 246, 21 },
		{ 301, 247, 21 },
		{ 301, 248, 21 },
		{ 301, 249, 21 },
		{ 301, 250, 21 },
		{ 301, 251, 21 },
		{ 301, 252, 21 },
		{ 301, 253, 21 },
		{ 301, 254, 21 },
		{ 301, 255, 21 },
		{ 301, 256, 21 },
		{ 302, 235, 21 },
		{ 302, 236, 21 },
		{ 302, 237, 21 },
		{ 302, 238, 21 },
		{ 302, 239, 21 },
		{ 302, 240, 21 },
		{ 302, 241, 21 },
		{ 302, 242, 21 },
		{ 302, 243, 21 },
		{ 302, 244, 21 },
		{ 302, 245, 21 },
		{ 302, 246, 21 },
		{ 302, 247, 21 },
		{ 302, 248, 21 },
		{ 302, 249, 21 },
		{ 302, 250, 21 },
		{ 302, 251, 21 },
		{ 302, 252, 21 },
		{ 302, 253, 21 },
		{ 302, 254, 21 },
		{ 302, 255, 21 },
		{ 302, 256, 21 },
		{ 303, 235, 21 },
		{ 303, 236, 21 },
		{ 303, 237, 21 },
		{ 303, 238, 21 },
		{ 303, 239, 21 },
		{ 303, 240, 21 },
		{ 303, 241, 21 },
		{ 303, 242, 21 },
		{ 303, 243, 21 },
		{ 303, 244, 21 },
		{ 303, 245, 21 },
		{ 303, 246, 21 },
		{ 303, 247, 21 },
		{ 303, 248, 21 },
		{ 303, 249, 21 },
		{ 303, 250, 21 },
		{ 303, 251, 21 },
		{ 303, 252, 21 },
		{ 303, 253, 21 },
		{ 303, 254, 21 },
		{ 303, 255, 21 },
		{ 303, 256, 21 },
		{ 304, 234, 21 },
		{ 304, 235, 21 },
		{ 304, 236, 21 },
		{ 304, 237, 21 },
		{ 304, 238, 21 },
		{ 304, 239, 21 },
		{ 304, 240, 21 },
		{ 304, 241, 21 },
		{ 304, 242, 21 },
		{ 304, 243, 21 },
		{ 304, 244, 21 },
		{ 304, 245, 21 },
		{ 304, 246, 21 },
		{ 304, 247, 21 },
		{ 304, 248, 21 },
		{ 304, 249, 21 },
		{ 304, 250, 21 },
		{ 304, 251, 21 },
		{ 304, 252, 21 },
		{ 304, 253, 21 },
		{ 304, 254, 21 },
		{ 304, 255, 21 },
		{ 304, 256, 21 },
		{ 305, 234, 21 },
		{ 305, 235, 21 },
		{ 305, 236, 21 },
		{ 305, 237, 21 },
		{ 305, 238, 21 },
		{ 305, 239, 21 },
		{ 305, 240, 21 },
		{ 305, 241, 21 },
		{ 305, 242, 21 },
		{ 305, 243, 21 },
		{ 305, 244, 21 },
		{ 305, 245, 21 },
		{ 305, 246, 21 },
		{ 305, 247, 21 },
		{ 305, 248, 21 },
		{ 305, 249, 21 },
		{ 305, 250, 21 },
		{ 305, 251, 21 },
		{ 305, 252, 21 },
		{ 305, 253, 21 },
		{ 305, 254, 21 },
		{ 305, 255, 21 },
		{ 305, 256, 21 },
		{ 306, 233, 21 },
		{ 306, 234, 21 },
		{ 306, 235, 21 },
		{ 306, 236, 21 },
		{ 306, 237, 21 },
		{ 306, 238, 21 },
		{ 306, 239, 21 },
		{ 306, 240, 21 },
		{ 306, 241, 21 },
		{ 306, 242, 21 },
		{ 306, 243, 21 },
		{ 306, 244, 21 },
		{ 306, 245, 21 },
		{ 306, 246, 21 },
		{ 306, 247, 21 },
		{ 306, 248, 21 },
		{ 306, 249, 21 },
		{ 306, 250, 21 },
		{ 306, 251, 21 },
		{ 306, 252, 21 },
		{ 306, 253, 21 },
		{ 306, 254, 21 },
		{ 306, 255, 21 },
		{ 306, 256, 21 },
		{ 307, 233, 21 },
		{ 307, 234, 21 },
		{ 307, 235, 21 },
		{ 307, 236, 21 },
		{ 307, 237, 21 },
		{ 307, 238, 21 },
		{ 307, 239, 21 },
		{ 307, 240, 21 },
		{ 307, 241, 21 },
		{ 307, 242, 21 },
		{ 307, 243, 21 },
		{ 307, 244, 21 },
		{ 307, 245, 21 },
		{ 307, 246, 21 },
		{ 307, 247, 21 },
		{ 307, 248, 21 },
		{ 307, 249, 21 },
		{ 307, 250, 21 },
		{ 307, 251, 21 },
		{ 307, 252, 21 },
		{ 307, 253, 21 },
		{ 307, 254, 21 },
		{ 307, 255, 21 },
		{ 307, 256, 21 },
		{ 308, 232, 21 },
		{ 308, 233, 21 },
		{ 308, 234, 21 },
		{ 308, 235, 21 },
		{ 308, 236, 21 },
		{ 308, 237, 21 },
		{ 308, 238, 21 },
		{ 308, 239, 21 },
		{ 308, 240, 21 },
		{ 308, 241, 21 },
		{ 308, 242, 21 },
		{ 308, 243, 21 },
		{ 308, 244, 21 },
		{ 308, 245, 21 },
		{ 308, 246, 21 },
		{ 308, 247, 21 },
		{ 308, 248, 21 },
		{ 308, 249, 21 },
		{ 308, 250, 21 },
		{ 308, 251, 21 },
		{ 308, 252, 21 },
		{ 308, 253, 21 },
		{ 308, 254, 21 },
		{ 308, 255, 21 },
		{ 308, 256, 21 },
		{ 309, 232, 21 },
		{ 309, 233, 21 },
		{ 309, 234, 21 },
		{ 309, 235, 21 },
		{ 309, 236, 21 },
		{ 309, 237, 21 },
		{ 309, 238, 21 },
		{ 309, 239, 21 },
		{ 309, 240, 21 },
		{ 309, 241, 21 },
		{ 309, 242, 21 },
		{ 309, 243, 21 },
		{ 309, 244, 21 },
		{ 309, 245, 21 },
		{ 309, 246, 21 },
		{ 309, 247, 21 },
		{ 309, 248, 21 },
		{ 309, 249, 21 },
		{ 309, 250, 21 },
		{ 309, 251, 21 },
		{ 309, 252, 21 },
		{ 309, 253, 21 },
		{ 309, 254, 21 },
		{ 309, 255, 21 },
		{ 309, 256, 21 },
		{ 310, 231, 21 },
		{ 310, 232, 21 },
		{ 310, 233, 21 },
		{ 310, 234, 21 },
		{ 310, 235, 21 },
		{ 310, 236, 21 },
		{ 310, 237, 21 },
		{ 310, 238, 21 },
		{ 310, 239, 21 },
		{ 310, 240, 21 },
		{ 310, 241, 21 },
		{ 310, 242, 21 },
		{ 310, 243, 21 },
		{ 310, 244, 21 },
		{ 310, 245, 21 },
		{ 310, 246, 21 },
		{ 310, 247, 21 },
		{ 310, 248, 21 },
		{ 310, 249, 21 },
		{ 310, 250, 21 },
		{ 310, 251, 21 },
		{ 310, 252, 21 },
		{ 310, 253, 21 },
		{ 310, 254, 21 },
		{ 310, 255, 21 },
		{ 310, 256, 21 },
		{ 311, 231, 21 },
		{ 311, 232, 21 },
		{ 311, 233, 21 },
		{ 311, 234, 21 },
		{ 311, 235, 21 },
		{ 311, 236, 21 },
		{ 311, 237, 21 },
		{ 311, 238, 21 },
		{ 311, 239, 21 },
		{ 311, 240, 21 },
		{ 311, 241, 21 },
		{ 311, 242, 21 },
		{ 311, 243, 21 },
		{ 311, 244, 21 },
		{ 311, 245, 21 },
		{ 311, 246, 21 },
		{ 311, 247, 21 },
		{ 311, 248, 21 },
		{ 311, 249, 21 },
		{ 311, 250, 21 },
		{ 311, 251, 21 },
		{ 311, 252, 21 },
		{ 311, 253, 21 },
		{ 311, 254, 21 },
		{ 311, 255, 21 },
		{ 311, 256, 21 },
		{ 312, 230, 21 },
		{ 312, 231, 21 },
		{ 312, 232, 21 },
		{ 312, 233, 21 },
		{ 312, 234, 21 },
		{ 312, 235, 21 },
		{ 312, 236, 21 },
		{ 312, 237, 21 },
		{ 312, 238, 21 },
		{ 312, 239, 21 },
		{ 312, 240, 21 },
		{ 312, 241, 21 },
		{ 312, 242, 21 },
		{ 312, 243, 21 },
		{ 312, 244, 21 },
		{ 312, 245, 21 },
		{ 312, 246, 21 },
		{ 312, 247, 21 },
		{ 312, 248, 21 },
		{ 312, 249, 21 },
		{ 312, 250, 21 },
		{ 312, 251, 21 },
		{ 312, 252, 21 },
		{ 312, 253, 21 },
		{ 312, 254, 21 },
		{ 312, 255, 21 },
		{ 312, 256, 21 },
		{ 313, 230, 21 },
		{ 313, 231, 21 },
		{ 313, 232, 21 },
		{ 313, 233, 21 },
		{ 313, 234, 21 },
		{ 313, 235, 21 },
		{ 313, 236, 21 },
		{ 313, 237, 21 },
		{ 313, 238, 21 },
		{ 313, 239, 21 },
		{ 313, 240, 21 },
		{ 313, 241, 21 },
		{ 313, 242, 21 },
		{ 313, 243, 21 },
		{ 313, 244, 21 },
		{ 313, 245, 21 },
		{ 313, 246, 21 },
		{ 313, 247, 21 },
		{ 313, 248, 21 },
		{ 313, 249, 21 },
		{ 313, 250, 21 },
		{ 313, 251, 21 },
		{ 313, 252, 21 },
		{ 313, 253, 21 },
		{ 313, 254, 21 },
		{ 313, 255, 21 },
		{ 313, 256, 21 },
		{ 314, 229, 21 },
		{ 314, 230, 21 },
		{ 314, 231, 21 },
		{ 314, 232, 21 },
		{ 314, 233, 21 },
		{ 314, 234, 21 },
		{ 314, 235, 21 },
		{ 314, 236, 21 },
		{ 314, 237, 21 },
		{ 314, 238, 21 },
		{ 314, 239, 21 },
		{ 314, 240, 21 },
		{ 314, 241, 21 },
		{ 314, 242, 21 },
		{ 314, 243, 21 },
		{ 314, 244, 21 },
		{ 314, 245, 21 },
		{ 314, 246, 21 },
		{ 314, 247, 21 },
		{ 314, 248, 21 },
		{ 314, 249, 21 },
		{ 314, 250, 21 },
		{ 314, 251, 21 },
		{ 314, 252, 21 },
		{ 314, 253, 21 },
		{ 314, 254, 21 },
		{ 314, 255, 21 },
		{ 314, 256, 21 },
		{ 315, 229, 21 },
		{ 315, 230, 21 },
		{ 315, 231, 21 },
		{ 315, 232, 21 },
		{ 315, 233, 21 },
		{ 315, 234, 21 },
		{ 315, 235, 21 },
		{ 315, 236, 21 },
		{ 315, 237, 21 },
		{ 315, 238, 21 },
		{ 315, 239, 21 },
		{ 315, 240, 21 },
		{ 315, 241, 21 },
		{ 315, 242, 21 },
		{ 315, 243, 21 },
		{ 315, 244, 21 },
		{ 315, 245, 21 },
		{ 315, 246, 21 },
		{ 315, 247, 21 },
		{ 315, 248, 21 },
		{ 315, 249, 21 },
		{ 315, 250, 21 },
		{ 315, 251, 21 },
		{ 315, 252, 21 },
		{ 315, 253, 21 },
		{ 315, 254, 21 },
		{ 315, 255, 21 },
		{ 315, 256, 21 },
		{ 316, 228, 21 },
		{ 316, 229, 21 },
		{ 316, 230, 21 },
		{ 316, 231, 21 },
		{ 316, 232, 21 },
		{ 316, 233, 21 },
		{ 316, 234, 21 },
		{ 316, 235, 21 },
		{ 316, 236, 21 },
		{ 316, 237, 21 },
		{ 316, 238, 21 },
		{ 316, 239, 21 },
		{ 316, 240, 21 },
		{ 316, 241, 21 },
		{ 316, 242, 21 },
		{ 316, 243, 21 },
		{ 316, 244, 21 },
		{ 316, 245, 21 },
		{ 316, 246, 21 },
		{ 316, 247, 21 },
		{ 316, 248, 21 },
		{ 316, 249, 21 },
		{ 316, 250, 21 },
		{ 316, 251, 21 },
		{ 316, 252, 21 },
		{ 316, 253, 21 },
		{ 316, 254, 21 },
		{ 316, 255, 21 },
		{ 316, 256, 21 },
		{ 317, 228, 21 },
		{ 317, 229, 21 },
		{ 317, 230, 21 },
		{ 317, 231, 21 },
		{ 317, 232, 21 },
		{ 317, 233, 21 },
		{ 317, 234, 21 },
		{ 317, 235, 21 },
		{ 317, 236, 21 },
		{ 317, 237, 21 },
		{ 317, 238, 21 },
		{ 317, 239, 21 },
		{ 317, 240, 21 },
		{ 317, 241, 21 },
		{ 317, 242, 21 },
		{ 317, 243, 21 },
		{ 317, 244, 21 },
		{ 317, 245, 21 },
		{ 317, 246, 21 },
		{ 317, 247, 21 },
		{ 317, 248, 21 },
		{ 317, 249, 21 },
		{ 317, 250, 21 },
		{ 317, 251, 21 },
		{ 317, 252, 21 },
		{ 317, 253, 21 },
		{ 317, 254, 21 },
		{ 317, 255, 21 },
		{ 317, 256, 21 },
		{ 318, 227, 21 },
		{ 318, 228, 21 },
		{ 318, 229, 21 },
		{ 318, 230, 21 },
		{ 318, 231, 21 },
		{ 318, 232, 21 },
		{ 318, 233, 21 },
		{ 318, 234, 21 },
		{ 318, 235, 21 },
		{ 318, 236, 21 },
		{ 318, 237, 21 },
		{ 318, 238, 21 },
		{ 318, 239, 21 },
		{ 318, 240, 21 },
		{ 318, 241, 21 },
		{ 318, 242, 21 },
		{ 318, 243, 21 },
		{ 318, 244, 21 },
		{ 318, 245, 21 },
		{ 318, 246, 21 },
		{ 318, 247, 21 },
		{ 318, 248, 21 },
		{ 318, 249, 21 },
		{ 318, 250, 21 },
		{ 318, 251, 21 },
		{ 318, 252, 21 },
		{ 318, 253, 21 },
		{ 318, 254, 21 },
		{ 318, 255, 21 },
		{ 318, 256, 21 },
		{ 319, 227, 21 },
		{ 319, 228, 21 },
		{ 319, 229, 21 },
		{ 319, 230, 21 },
		{ 319, 231, 21 },
		{ 319, 232, 21 },
		{ 319, 233, 21 },
		{ 319, 234, 21 },
		{ 319, 235, 21 },
		{ 319, 236, 21 },
		{ 319, 237, 21 },
		{ 319, 238, 21 },
		{ 319, 239, 21 },
		{ 319, 240, 21 },
		{ 319, 241, 21 },
		{ 319, 242, 21 },
		{ 319, 243, 21 },
		{ 319, 244, 21 },
		{ 319, 245, 21 },
		{ 319, 246, 21 },
		{ 319, 247, 21 },
		{ 319, 248, 21 },
		{ 319, 249, 21 },
		{ 319, 250, 21 },
		{ 319, 251, 21 },
		{ 319, 252, 21 },
		{ 319, 253, 21 },
		{ 319, 254, 21 },
		{ 319, 255, 21 },
		{ 319, 256, 21 },
		{ 320, 226, 21 },
		{ 320, 227, 21 },
		{ 320, 228, 21 },
		{ 320, 229, 21 },
		{ 320, 230, 21 },
		{ 320, 231, 21 },
		{ 320, 232, 21 },
		{ 320, 233, 21 },
		{ 320, 234, 21 },
		{ 320, 235, 21 },
		{ 320, 236, 21 },
		{ 320, 237, 21 },
		{ 320, 238, 21 },
		{ 320, 239, 21 },
		{ 320, 240, 21 },
		{ 320, 241, 21 },
		{ 320, 242, 21 },
		{ 320, 243, 21 },
		{ 320, 244, 21 },
		{ 320, 245, 21 },
		{ 320, 246, 21 },
		{ 320, 247, 21 },
		{ 320, 248, 21 },
		{ 320, 249, 21 },
		{ 320, 250, 21 },
		{ 320, 251, 21 },
		{ 320, 252, 21 },
		{ 320, 253, 21 },
		{ 320, 254, 21 },
		{ 320, 255, 21 },
		{ 320, 256, 21 },
		{ 321, 226, 21 },
		{ 321, 227, 21 },
		{ 321, 228, 21 },
		{ 321, 229, 21 },
		{ 321, 230, 21 },
		{ 321, 231, 21 },
		{ 321, 232, 21 },
		{ 321, 233, 21 },
		{ 321, 234, 21 },
		{ 321, 235, 21 },
		{ 321, 236, 21 },
		{ 321, 237, 21 },
		{ 321, 238, 21 },
		{ 321, 239, 21 },
		{ 321, 240, 21 },
		{ 321, 241, 21 },
		{ 321, 242, 21 },
		{ 321, 243, 21 },
		{ 321, 244, 21 },
		{ 321, 245, 21 },
		{ 321, 246, 21 },
		{ 321, 247, 21 },
		{ 321, 248, 21 },
		{ 321, 249, 21 },
		{ 321, 250, 21 },
		{ 321, 251, 21 },
		{ 321, 252, 21 },
		{ 321, 253, 21 },
		{ 321, 254, 21 },
		{ 321, 255, 21 },
		{ 321, 256, 21 },
		{ 322, 225, 21 },
		{ 322, 226, 21 },
		{ 322, 227, 21 },
		{ 322, 228, 21 },
		{ 322, 229, 21 },
		{ 322, 230, 21 },
		{ 322, 231, 21 },
		{ 322, 232, 21 },
		{ 322, 233, 21 },
		{ 322, 234, 21 },
		{ 322, 235, 21 },
		{ 322, 236, 21 },
		{ 322, 237, 21 },
		{ 322, 238, 21 },
		{ 322, 239, 21 },
		{ 322, 240, 21 },
		{ 322, 241, 21 },
		{ 322, 242, 21 },
		{ 322, 243, 21 },
		{ 322, 244, 21 },
		{ 322, 245, 21 },
		{ 322, 246, 21 },
		{ 322, 247, 21 },
		{ 322, 248, 21 },
		{ 322, 249, 21 },
		{ 322, 250, 21 },
		{ 322, 251, 21 },
		{ 322, 252, 21 },
		{ 322, 253, 21 },
		{ 322, 254, 21 },
		{ 322, 255, 21 },
		{ 322, 256, 21 },
		{ 323, 225, 21 },
		{ 323, 226, 21 },
		{ 323, 227, 21 },
		{ 323, 228, 21 },
		{ 323, 229, 21 },
		{ 323, 230, 21 },
		{ 323, 231, 21 },
		{ 323, 232, 21 },
		{ 323, 233, 21 },
		{ 323, 234, 21 },
		{ 323, 235, 21 },
		{ 323, 236, 21 },
		{ 323, 237, 21 },
		{ 323, 238, 21 },
		{ 323, 239, 21 },
		{ 323, 240, 21 },
		{ 323, 241, 21 },
		{ 323, 242, 21 },
		{ 323, 243, 21 },
		{ 323, 244, 21 },
		{ 323, 245, 21 },
		{ 323, 246, 21 },
		{ 323, 247, 21 },
		{ 323, 248, 21 },
		{ 323, 249, 21 },
		{ 323, 250, 21 },
		{ 323, 251, 21 },
		{ 323, 252, 21 },
		{ 323, 253, 21 },
		{ 323, 254, 21 },
		{ 323, 255, 21 },
		{ 323, 256, 21 },
		{ 324, 224, 21 },
		{ 324, 225, 21 },
		{ 324, 226, 21 },
		{ 324, 227, 21 },
		{ 324, 228, 21 },
		{ 324, 229, 21 },
		{ 324, 230, 21 },
		{ 324, 231, 21 },
		{ 324, 232, 21 },
		{ 324, 233, 21 },
		{ 324, 234, 21 },
		{ 324, 235, 21 },
		{ 324, 236, 21 },
		{ 324, 237, 21 },
		{ 324, 238, 21 },
		{ 324, 239, 21 },
		{ 324, 240, 21 },
		{ 324, 241, 21 },
		{ 324, 242, 21 },
		{ 324, 243, 21 },
		{ 324, 244, 21 },
		{ 324, 245, 21 },
		{ 324, 246, 21 },
		{ 324, 247, 21 },
		{ 324, 248, 21 },
		{ 324, 249, 21 },
		{ 324, 250, 21 },
		{ 324, 251, 21 },
		{ 324, 252, 21 },
		{ 324, 253, 21 },
		{ 324, 254, 21 },
		{ 324, 255, 21 },
		{ 324, 256, 21 },
		{ 325, 224, 21 },
		{ 325, 225, 21 },
		{ 325, 226, 21 },
		{ 325, 227, 21 },
		{ 325, 228, 21 },
		{ 325, 229, 21 },
		{ 325, 230, 21 },
		{ 325, 231, 21 },
		{ 325, 232, 21 },
		{ 325, 233, 21 },
		{ 325, 234, 21 },
		{ 325, 235, 21 },
		{ 325, 236, 21 },
		{ 325, 237, 21 },
		{ 325, 238, 21 },
		{ 325, 239, 21 },
		{ 325, 240, 21 },
		{ 325, 241, 21 },
		{ 325, 242, 21 },
		{ 325, 243, 21 },
		{ 325, 244, 21 },
		{ 325, 245, 21 },
		{ 325, 246, 21 },
		{ 325, 247, 21 },
		{ 325, 248, 21 },
		{ 325, 249, 21 },
		{ 325, 250, 21 },
		{ 325, 251, 21 },
		{ 325, 252, 21 },
		{ 325, 253, 21 },
		{ 325, 254, 21 },
		{ 325, 255, 21 },
		{ 325, 256, 21 },
		{ 326, 223, 21 },
		{ 326, 224, 21 },
		{ 326, 225, 21 },
		{ 326, 226, 21 },
		{ 326, 227, 21 },
		{ 326, 228, 21 },
		{ 326, 229, 21 },
		{ 326, 230, 21 },
		{ 326, 231, 21 },
		{ 326, 232, 21 },
		{ 326, 233, 21 },
		{ 326, 234, 21 },
		{ 326, 235, 21 },
		{ 326, 236, 21 },
		{ 326, 237, 21 },
		{ 326, 238, 21 },
		{ 326, 239, 21 },
		{ 326, 240, 21 },
		{ 326, 241, 21 },
		{ 326, 242, 21 },
		{ 326, 243, 21 },
		{ 326, 244, 21 },
		{ 326, 245, 21 },
		{ 326, 246, 21 },
		{ 326, 247, 21 },
		{ 326, 248, 21 },
		{ 326, 249, 21 },
		{ 326, 250, 21 },
		{ 326, 251, 21 },
		{ 326, 252, 21 },
		{ 326, 253, 21 },
		{ 326, 254, 21 },
		{ 326, 255, 21 },
		{ 326, 256, 21 },
		{ 327, 223, 21 },
		{ 327, 224, 21 },
		{ 327, 225, 21 },
		{ 327, 226, 21 },
		{ 327, 227, 21 },
		{ 327, 228, 21 },
		{ 327, 229, 21 },
		{ 327, 230, 21 },
		{ 327, 231, 21 },
		{ 327, 232, 21 },
		{ 327, 233, 21 },
		{ 327, 234, 21 },
		{ 327, 235, 21 },
		{ 327, 236, 21 },
		{ 327, 237, 21 },
		{ 327, 238, 21 },
		{ 327, 239, 21 },
		{ 327, 240, 21 },
		{ 327, 241, 21 },
		{ 327, 242, 21 },
		{ 327, 243, 21 },
		{ 327, 244, 21 },
		{ 327, 245, 21 },
		{ 327, 246, 21 },
		{ 327, 247, 21 },
		{ 327, 248, 21 },
		{ 327, 249, 21 },
		{ 327, 250, 21 },
		{ 327, 251, 21 },
		{ 327, 252, 21 },
		{ 327, 253, 21 },
		{ 327, 254, 21 },
		{ 327, 255, 21 },
		{ 327, 256, 21 },
		{ 328, 222, 21 },
		{ 328, 223, 21 },
		{ 328, 224, 21 },
		{ 328, 225, 21 },
		{ 328, 226, 21 },
		{ 328, 227, 21 },
		{ 328, 228, 21 },
		{ 328, 229, 21 },
		{ 328, 230, 21 },
		{ 328, 231, 21 },
		{ 328, 232, 21 },
		{ 328, 233, 21 },
		{ 328, 234, 21 },
		{ 328, 235, 21 },
		{ 328, 236, 21 },
		{ 328, 237, 21 },
		{ 328, 238, 21 },
		{ 328, 239, 21 },
		{ 328, 240, 21 },
		{ 328, 241, 21 },
		{ 328, 242, 21 },
		{ 328, 243, 21 },
		{ 328, 244, 21 },
		{ 328, 245, 21 },
		{ 328, 246, 21 },
		{ 328, 247, 21 },
		{ 328, 248, 21 },
		{ 328, 249, 21 },
		{ 328, 250, 21 },
		{ 328, 251, 21 },
		{ 328, 252, 21 },
		{ 328, 253, 21 },
		{ 328, 254, 21 },
		{ 328, 255, 21 },
		{ 328, 256, 21 },
		{ 329, 222, 21 },
		{ 329, 223, 21 },
		{ 329, 224, 21 },
		{ 329, 225, 21 },
		{ 329, 226, 21 },
		{ 329, 227, 21 },
		{ 329, 228, 21 },
		{ 329, 229, 21 },
		{ 329, 230, 21 },
		{ 329, 231, 21 },
		{ 329, 232, 21 },
		{ 329, 233, 21 },
		{ 329, 234, 21 },
		{ 329, 235, 21 },
		{ 329, 236, 21 },
		{ 329, 237, 21 },
		{ 329, 238, 21 },
		{ 329, 239, 21 },
		{ 329, 240, 21 },
		{ 329, 241, 21 },
		{ 329, 242, 21 },
		{ 329, 243, 21 },
		{ 329, 244, 21 },
		{ 329, 245, 21 },
		{ 329, 246, 21 },
		{ 329, 247, 21 },
		{ 329, 248, 21 },
		{ 329, 249, 21 },
		{ 329, 250, 21 },
		{ 329, 251, 21 },
		{ 329, 252, 21 },
		{ 329, 253, 21 },
		{ 329, 254, 21 },
		{ 329, 255, 21 },
		{ 329, 256, 21 },
		{ 330, 221, 21 },
		{ 330, 222, 21 },
		{ 330, 223, 21 },
		{ 330, 224, 21 },
		{ 330, 225, 21 },
		{ 330, 226, 21 },
		{ 330, 227, 21 },
		{ 330, 228, 21 },
		{ 330, 229, 21 },
		{ 330, 230, 21 },
		{ 330, 231, 21 },
		{ 330, 232, 21 },
		{ 330, 233, 21 },
		{ 330, 234, 21 },
		{ 330, 235, 21 },
		{ 330, 236, 21 },
		{ 330, 237, 21 },
		{ 330, 238, 21 },
		{ 330, 239, 21 },
		{ 330, 240, 21 },
		{ 330, 241, 21 },
		{ 330, 242, 21 },
		{ 330, 243, 21 },
		{ 330, 244, 21 },
		{ 330, 245, 21 },
		{ 330, 246, 21 },
		{ 330, 247, 21 },
		{ 330, 248, 21 },
		{ 330, 249, 21 },
		{ 330, 250, 21 },
		{ 330, 251, 21 },
		{ 330, 252, 21 },
		{ 330, 253, 21 },
		{ 330, 254, 21 },
		{ 330, 255, 21 },
		{ 330, 256, 21 },
		{ 331, 221, 21 },
		{ 331, 222, 21 },
		{ 331, 223, 21 },
		{ 331, 224, 21 },
		{ 331, 225, 21 },
		{ 331, 226, 21 },
		{ 331, 227, 21 },
		{ 331, 228, 21 },
		{ 331, 229, 21 },
		{ 331, 230, 21 },
		{ 331, 231, 21 },
		{ 331, 232, 21 },
		{ 331, 233, 21 },
		{ 331, 234, 21 },
		{ 331, 235, 21 },
		{ 331, 236, 21 },
		{ 331, 237, 21 },
		{ 331, 238, 21 },
		{ 331, 239, 21 },
		{ 331, 240, 21 },
		{ 331, 241, 21 },
		{ 331, 242, 21 },
		{ 331, 243, 21 },
		{ 331, 244, 21 },
		{ 331, 245, 21 },
		{ 331, 246, 21 },
		{ 331, 247, 21 },
		{ 331, 248, 21 },
		{ 331, 249, 21 },
		{ 331, 250, 21 },
		{ 331, 251, 21 },
		{ 331, 252, 21 },
		{ 331, 253, 21 },
		{ 331, 254, 21 },
		{ 331, 255, 21 },
		{ 331, 256, 21 },
		{ 332, 220, 21 },
		{ 332, 221, 21 },
		{ 332, 222, 21 },
		{ 332, 223, 21 },
		{ 332, 224, 21 },
		{ 332, 225, 21 },
		{ 332, 226, 21 },
		{ 332, 227, 21 },
		{ 332, 228, 21 },
		{ 332, 229, 21 },
		{ 332, 230, 21 },
		{ 332, 231, 21 },
		{ 332, 232, 21 },
		{ 332, 233, 21 },
		{ 332, 234, 21 },
		{ 332, 235, 21 },
		{ 332, 236, 21 },
		{ 332, 237, 21 },
		{ 332, 238, 21 },
		{ 332, 239, 21 },
		{ 332, 240, 21 },
		{ 332, 241, 21 },
		{ 332, 242, 21 },
		{ 332, 243, 21 },
		{ 332, 244, 21 },
		{ 332, 245, 21 },
		{ 332, 246, 21 },
		{ 332, 247, 21 },
		{ 332, 248, 21 },
		{ 332, 249, 21 },
		{ 332, 250, 21 },
		{ 332, 251, 21 },
		{ 332, 252, 21 },
		{ 332, 253, 21 },
		{ 332, 254, 21 },
		{ 332, 255, 21 },
		{ 332, 256, 21 },
		{ 333, 220, 21 },
		{ 333, 221, 21 },
		{ 333, 222, 21 },
		{ 333, 223, 21 },
		{ 333, 224, 21 },
		{ 333, 225, 21 },
		{ 333, 226, 21 },
		{ 333, 227, 21 },
		{ 333, 228, 21 },
		{ 333, 229, 21 },
		{ 333, 230, 21 },
		{ 333, 231, 21 },
		{ 333, 232, 21 },
		{ 333, 233, 21 },
		{ 333, 234, 21 },
		{ 333, 235, 21 },
		{ 333, 236, 21 },
		{ 333, 237, 21 },
		{ 333, 238, 21 },
		{ 333, 239, 21 },
		{ 333, 240, 21 },
		{ 333, 241, 21 },
		{ 333, 242, 21 },
		{ 333, 243, 21 },
		{ 333, 244, 21 },
		{ 333, 245, 21 },
		{ 333, 246, 21 },
		{ 333, 247, 21 },
		{ 333, 248, 21 },
		{ 333, 249, 21 },
		{ 333, 250, 21 },
		{ 333, 251, 21 },
		{ 333, 252, 21 },
		{ 333, 253, 21 },
		{ 333, 254, 21 },
		{ 333, 255, 21 },
		{ 333, 256, 21 },
		{ 334, 219, 21 },
		{ 334, 220, 21 },
		{ 334, 221, 21 },
		{ 334, 222, 21 },
		{ 334, 223, 21 },
		{ 334, 224, 21 },
		{ 334, 225, 21 },
		{ 334, 226, 21 },
		{ 334, 227, 21 },
		{ 334, 228, 21 },
		{ 334, 229, 21 },
		{ 334, 230, 21 },
		{ 334, 231, 21 },
		{ 334, 232, 21 },
		{ 334, 233, 21 },
		{ 334, 234, 21 },
		{ 334, 235, 21 },
		{ 334, 236, 21 },
		{ 334, 237, 21 },
		{ 334, 238, 21 },
		{ 334, 239, 21 },
		{ 334, 240, 21 },
		{ 334, 241, 21 },
		{ 334, 242, 21 },
		{ 334, 243, 21 },
		{ 334, 244, 21 },
		{ 334, 245, 21 },
		{ 334, 246, 21 },
		{ 334, 247, 21 },
		{ 334, 248, 21 },
		{ 334, 249, 21 },
		{ 334, 250, 21 },
		{ 334, 251, 21 },
		{ 334, 252, 21 },
		{ 334, 253, 21 },
		{ 334, 254, 21 },
		{ 334, 255, 21 },
		{ 334, 256, 21 },
		{ 335, 219, 21 },
		{ 335, 220, 21 },
		{ 335, 221, 21 },
		{ 335, 222, 21 },
		{ 335, 223, 21 },
		{ 335, 224, 21 },
		{ 335, 225, 21 },
		{ 335, 226, 21 },
		{ 335, 227, 21 },
		{ 335, 228, 21 },
		{ 335, 229, 21 },
		{ 335, 230, 21 },
		{ 335, 231, 21 },
		{ 335, 232, 21 },
		{ 335, 233, 21 },
		{ 335, 234, 21 },
		{ 335, 235, 21 },
		{ 335, 236, 21 },
		{ 335, 237, 21 },
		{ 335, 238, 21 },
		{ 335, 239, 21 },
		{ 335, 240, 21 },
		{ 335, 241, 21 },
		{ 335, 242, 21 },
		{ 335, 243, 21 },
		{ 335, 244, 21 },
		{ 335, 245, 21 },
		{ 335, 246, 21 },
		{ 335, 247, 21 },
		{ 335, 248, 21 },
		{ 335, 249, 21 },
		{ 335, 250, 21 },
		{ 335, 251, 21 },
		{ 335, 252, 21 },
		{ 335, 253, 21 },
		{ 335, 254, 21 },
		{ 335, 255, 21 },
		{ 335, 256, 21 },
		{ 336, 218, 21 },
		{ 336, 219, 21 },
		{ 336, 220, 21 },
		{ 336, 221, 21 },
		{ 336, 222, 21 },
		{ 336, 223, 21 },
		{ 336, 224, 21 },
		{ 336, 225, 21 },
		{ 336, 226, 21 },
		{ 336, 227, 21 },
		{ 336, 228, 21 },
		{ 336, 229, 21 },
		{ 336, 230, 21 },
		{ 336, 231, 21 },
		{ 336, 232, 21 },
		{ 336, 233, 21 },
		{ 336, 234, 21 },
		{ 336, 235, 21 },
		{ 336, 236, 21 },
		{ 336, 237, 21 },
		{ 336, 238, 21 },
		{ 336, 239, 21 },
		{ 336, 240, 21 },
		{ 336, 241, 21 },
		{ 336, 242, 21 },
		{ 336, 243, 21 },
		{ 336, 244, 21 },
		{ 336, 245, 21 },
		{ 336, 246, 21 },
		{ 336, 247, 21 },
		{ 336, 248, 21 },
		{ 336, 249, 21 },
		{ 336, 250, 21 },
		{ 336, 251, 21 },
		{ 336, 252, 21 },
		{ 336, 253, 21 },
		{ 336, 254, 21 },
		{ 336, 255, 21 },
		{ 336, 256, 21 },
		{ 337, 218, 21 },
		{ 337, 219, 21 },
		{ 337, 220, 21 },
		{ 337, 221, 21 },
		{ 337, 222, 21 },
		{ 337, 223, 21 },
		{ 337, 224, 21 },
		{ 337, 225, 21 },
		{ 337, 226, 21 },
		{ 337, 227, 21 },
		{ 337, 228, 21 },
		{ 337, 229, 21 },
		{ 337, 230, 21 },
		{ 337, 231, 21 },
		{ 337, 232, 21 },
		{ 337, 233, 21 },
		{ 337, 234, 21 },
		{ 337, 235, 21 },
		{ 337, 236, 21 },
		{ 337, 237, 21 },
		{ 337, 238, 21 },
		{ 337, 239, 21 },
		{ 337, 240, 21 },
		{ 337, 241, 21 },
		{ 337, 242, 21 },
		{ 337, 243, 21 },
		{ 337, 244, 21 },
		{ 337, 245, 21 },
		{ 337, 246, 21 },
		{ 337, 247, 21 },
		{ 337, 248, 21 },
		{ 337, 249, 21 },
		{ 337, 250, 21 },
		{ 337, 251, 21 },
		{ 337, 252, 21 },
		{ 337, 253, 21 },
		{ 337, 254, 21 },
		{ 337, 255, 21 },
		{ 337, 256, 21 },
		{ 338, 217, 21 },
		{ 338, 218, 21 },
		{ 338, 219, 21 },
		{ 338, 220, 21 },
		{ 338, 221, 21 },
		{ 338, 222, 21 },
		{ 338, 223, 21 },
		{ 338, 224, 21 },
		{ 338, 225, 21 },
		{ 338, 226, 21 },
		{ 338, 227, 21 },
		{ 338, 228, 21 },
		{ 338, 229, 21 },
		{ 338, 230, 21 },
		{ 338, 231, 21 },
		{ 338, 232, 21 },
		{ 338, 233, 21 },
		{ 338, 234, 21 },
		{ 338, 235, 21 },
		{ 338, 236, 21 },
		{ 338, 237, 21 },
		{ 338, 238, 21 },
		{ 338, 239, 21 },
		{ 338, 240, 21 },
		{ 338, 241, 21 },
		{ 338, 242, 21 },
		{ 338, 243, 21 },
		{ 338, 244, 21 },
		{ 338, 245, 21 },
		{ 338, 246, 21 },
		{ 338, 247, 21 },
		{ 338, 248, 21 },
		{ 338, 249, 21 },
		{ 338, 250, 21 },
		{ 338, 251, 21 },
		{ 338, 252, 21 },
		{ 338, 253, 21 },
		{ 338, 254, 21 },
		{ 338, 255, 21 },
		{ 338, 256, 21 },
		{ 339, 217, 21 },
		{ 339, 218, 21 },
		{ 339, 219, 21 },
		{ 339, 220, 21 },
		{ 339, 221, 21 },
		{ 339, 222, 21 },
		{ 339, 223, 21 },
		{ 339, 224, 21 },
		{ 339, 225, 21 },
		{ 339, 226, 21 },
		{ 339, 227, 21 },
		{ 339, 228, 21 },
		{ 339, 229, 21 },
		{ 339, 230, 21 },
		{ 339, 231, 21 },
		{ 339, 232, 21 },
		{ 339, 233, 21 },
		{ 339, 234, 21 },
		{ 339, 235, 21 },
		{ 339, 236, 21 },
		{ 339, 237, 21 },
		{ 339, 238, 21 },
		{ 339, 239, 21 },
		{ 339, 240, 21 },
		{ 339, 241, 21 },
		{ 339, 242, 21 },
		{ 339, 243, 21 },
		{ 339, 244, 21 },
		{ 339, 245, 21 },
		{ 339, 246, 21 },
		{ 339, 247, 21 },
		{ 339, 248, 21 },
		{ 339, 249, 21 },
		{ 339, 250, 21 },
		{ 339, 251, 21 },
		{ 339, 252, 21 },
		{ 339, 253, 21 },
		{ 339, 254, 21 },
		{ 339, 255, 21 },
		{ 339, 256, 21 },
		{ 340, 216, 21 },
		{ 340, 217, 21 },
		{ 340, 218, 21 },
		{ 340, 219, 21 },
		{ 340, 220, 21 },
		{ 340, 221, 21 },
		{ 340, 222, 21 },
		{ 340, 223, 21 },
		{ 340, 224, 21 },
		{ 340, 225, 21 },
		{ 340, 226, 21 },
		{ 340, 227, 21 },
		{ 340, 228, 21 },
		{ 340, 229, 21 },
		{ 340, 230, 21 },
		{ 340, 231, 21 },
		{ 340, 232, 21 },
		{ 340, 233, 21 },
		{ 340, 234, 21 },
		{ 340, 235, 21 },
		{ 340, 236, 21 },
		{ 340, 237, 21 },
		{ 340, 238, 21 },
		{ 340, 239, 21 },
		{ 340, 240, 21 },
		{ 340, 241, 21 },
		{ 340, 242, 21 },
		{ 340, 243, 21 },
		{ 340, 244, 21 },
		{ 340, 245, 21 },
		{ 340, 246, 21 },
		{ 340, 247, 21 },
		{ 340, 248, 21 },
		{ 340, 249, 21 },
		{ 340, 250, 21 },
		{ 340, 251, 21 },
		{ 340, 252, 21 },
		{ 340, 253, 21 },
		{ 340, 254, 21 },
		{ 340, 255, 21 },
		{ 340, 256, 21 },
		{ 341, 216, 21 },
		{ 341, 217, 21 },
		{ 341, 218, 21 },
		{ 341, 219, 21 },
		{ 341, 220, 21 },
		{ 341, 221, 21 },
		{ 341, 222, 21 },
		{ 341, 223, 21 },
		{ 341, 224, 21 },
		{ 341, 225, 21 },
		{ 341, 226, 21 },
		{ 341, 227, 21 },
		{ 341, 228, 21 },
		{ 341, 229, 21 },
		{ 341, 230, 21 },
		{ 341, 231, 21 },
		{ 341, 232, 21 },
		{ 341, 233, 21 },
		{ 341, 234, 21 },
		{ 341, 235, 21 },
		{ 341, 236, 21 },
		{ 341, 237, 21 },
		{ 341, 238, 21 },
		{ 341, 239, 21 },
		{ 341, 240, 21 },
		{ 341, 241, 21 },
		{ 341, 242, 21 },
		{ 341, 243, 21 },
		{ 341, 244, 21 },
		{ 341, 245, 21 },
		{ 341, 246, 21 },
		{ 341, 247, 21 },
		{ 341, 248, 21 },
		{ 341, 249, 21 },
		{ 341, 250, 21 },
		{ 341, 251, 21 },
		{ 341, 252, 21 },
		{ 341, 253, 21 },
		{ 341, 254, 21 },
		{ 341, 255, 21 },
		{ 341, 256, 21 },
		{ 342, 215, 21 },
		{ 342, 216, 21 },
		{ 342, 217, 21 },
		{ 342, 218, 21 },
		{ 342, 219, 21 },
		{ 342, 220, 21 },
		{ 342, 221, 21 },
		{ 342, 222, 21 },
		{ 342, 223, 21 },
		{ 342, 224, 21 },
		{ 342, 225, 21 },
		{ 342, 226, 21 },
		{ 342, 227, 21 },
		{ 342, 228, 21 },
		{ 342, 229, 21 },
		{ 342, 230, 21 },
		{ 342, 231, 21 },
		{ 342, 232, 21 },
		{ 342, 233, 21 },
		{ 342, 234, 21 },
		{ 342, 235, 21 },
		{ 342, 236, 21 },
		{ 342, 237, 21 },
		{ 342, 238, 21 },
		{ 342, 239, 21 },
		{ 342, 240, 21 },
		{ 342, 241, 21 },
		{ 342, 242, 21 },
		{ 342, 243, 21 },
		{ 342, 244, 21 },
		{ 342, 245, 21 },
		{ 342, 246, 21 },
		{ 342, 247, 21 },
		{ 342, 248, 21 },
		{ 342, 249, 21 },
		{ 342, 250, 21 },
		{ 342, 251, 21 },
		{ 342, 252, 21 },
		{ 342, 253, 21 },
		{ 342, 254, 21 },
		{ 342, 255, 21 },
		{ 342, 256, 21 },
		{ 343, 215, 21 },
		{ 343, 216, 21 },
		{ 343, 217, 21 },
		{ 343, 218, 21 },
		{ 343, 219, 21 },
		{ 343, 220, 21 },
		{ 343, 221, 21 },
		{ 343, 222, 21 },
		{ 343, 223, 21 },
		{ 343, 224, 21 },
		{ 343, 225, 21 },
		{ 343, 226, 21 },
		{ 343, 227, 21 },
		{ 343, 228, 21 },
		{ 343, 229, 21 },
		{ 343, 230, 21 },
		{ 343, 231, 21 },
		{ 343, 232, 21 },
		{ 343, 233, 21 },
		{ 343, 234, 21 },
		{ 343, 235, 21 },
		{ 343, 236, 21 },
		{ 343, 237, 21 },
		{ 343, 238, 21 },
		{ 343, 239, 21 },
		{ 343, 240, 21 },
		{ 343, 241, 21 },
		{ 343, 242, 21 },
		{ 343, 243, 21 },
		{ 343, 244, 21 },
		{ 343, 245, 21 },
		{ 343, 246, 21 },
		{ 343, 247, 21 },
		{ 343, 248, 21 },
		{ 343, 249, 21 },
		{ 343, 250, 21 },
		{ 343, 251, 21 },
		{ 343, 252, 21 },
		{ 343, 253, 21 },
		{ 343, 254, 21 },
		{ 343, 255, 21 },
		{ 343, 256, 21 },
		{ 344, 214, 21 },
		{ 344, 215, 21 },
		{ 344, 216, 21 },
		{ 344, 217, 21 },
		{ 344, 218, 21 },
		{ 344, 219, 21 },
		{ 344, 220, 21 },
		{ 344, 221, 21 },
		{ 344, 222, 21 },
		{ 344, 223, 21 },
		{ 344, 224, 21 },
		{ 344, 225, 21 },
		{ 344, 226, 21 },
		{ 344, 227, 21 },
		{ 344, 228, 21 },
		{ 344, 229, 21 },
		{ 344, 230, 21 },
		{ 344, 231, 21 },
		{ 344, 232, 21 },
		{ 344, 233, 21 },
		{ 344, 234, 21 },
		{ 344, 235, 21 },
		{ 344, 236, 21 },
		{ 344, 237, 21 },
		{ 344, 238, 21 },
		{ 344, 239, 21 },
		{ 344, 240, 21 },
		{ 344, 241, 21 },
		{ 344, 242, 21 },
		{ 344, 243, 21 },
		{ 344, 244, 21 },
		{ 344, 245, 21 },
		{ 344, 246, 21 },
		{ 344, 247, 21 },
		{ 344, 248, 21 },
		{ 344, 249, 21 },
		{ 344, 250, 21 },
		{ 344, 251, 21 },
		{ 344, 252, 21 },
		{ 344, 253, 21 },
		{ 344, 254, 21 },
		{ 344, 255, 21 },
		{ 344, 256, 21 },
		{ 345, 214, 21 },
		{ 345, 215, 21 },
		{ 345, 216, 21 },
		{ 345, 217, 21 },
		{ 345, 218, 21 },
		{ 345, 219, 21 },
		{ 345, 220, 21 },
		{ 345, 221, 21 },
		{ 345, 222, 21 },
		{ 345, 223, 21 },
		{ 345, 224, 21 },
		{ 345, 225, 21 },
		{ 345, 226, 21 },
		{ 345, 227, 21 },
		{ 345, 228, 21 },
		{ 345, 229, 21 },
		{ 345, 230, 21 },
		{ 345, 231, 21 },
		{ 345, 232, 21 },
		{ 345, 233, 21 },
		{ 345, 234, 21 },
		{ 345, 235, 21 },
		{ 345, 236, 21 },
		{ 345, 237, 21 },
		{ 345, 238, 21 },
		{ 345, 239, 21 },
		{ 345, 240, 21 },
		{ 345, 241, 21 },
		{ 345, 242, 21 },
		{ 345, 243, 21 },
		{ 345, 244, 21 },
		{ 345, 245, 21 },
		{ 345, 246, 21 },
		{ 345, 247, 21 },
		{ 345, 248, 21 },
		{ 345, 249, 21 },
		{ 345, 250, 21 },
		{ 345, 251, 21 },
		{ 345, 252, 21 },
		{ 345, 253, 21 },
		{ 345, 254, 21 },
		{ 345, 255, 21 },
		{ 345, 256, 21 },
		{ 346, 213, 21 },
		{ 346, 214, 21 },
		{ 346, 215, 21 },
		{ 346, 216, 21 },
		{ 346, 217, 21 },
		{ 346, 218, 21 },
		{ 346, 219, 21 },
		{ 346, 220, 21 },
		{ 346, 221, 21 },
		{ 346, 222, 21 },
		{ 346, 223, 21 },
		{ 346, 224, 21 },
		{ 346, 225, 21 },
		{ 346, 226, 21 },
		{ 346, 227, 21 },
		{ 346, 228, 21 },
		{ 346, 229, 21 },
		{ 346, 230, 21 },
		{ 346, 231, 21 },
		{ 346, 232, 21 },
		{ 346, 233, 21 },
		{ 346, 234, 21 },
		{ 346, 235, 21 },
		{ 346, 236, 21 },
		{ 346, 237, 21 },
		{ 346, 238, 21 },
		{ 346, 239, 21 },
		{ 346, 240, 21 },
		{ 346, 241, 21 },
		{ 346, 242, 21 },
		{ 346, 243, 21 },
		{ 346, 244, 21 },
		{ 346, 245, 21 },
		{ 346, 246, 21 },
		{ 346, 247, 21 },
		{ 346, 248, 21 },
		{ 346, 249, 21 },
		{ 346, 250, 21 },
		{ 346, 251, 21 },
		{ 346, 252, 21 },
		{ 346, 253, 21 },
		{ 346, 254, 21 },
		{ 346, 255, 21 },
		{ 346, 256, 21 },
		{ 347, 213, 21 },
		{ 347, 214, 21 },
		{ 347, 215, 21 },
		{ 347, 216, 21 },
		{ 347, 217, 21 },
		{ 347, 218, 21 },
		{ 347, 219, 21 },
		{ 347, 220, 21 },
		{ 347, 221, 21 },
		{ 347, 222, 21 },
		{ 347, 223, 21 },
		{ 347, 224, 21 },
		{ 347, 225, 21 },
		{ 347, 226, 21 },
		{ 347, 227, 21 },
		{ 347, 228, 21 },
		{ 347, 229, 21 },
		{ 347, 230, 21 },
		{ 347, 231, 21 },
		{ 347, 232, 21 },
		{ 347, 233, 21 },
		{ 347, 234, 21 },
		{ 347, 235, 21 },
		{ 347, 236, 21 },
		{ 347, 237, 21 },
		{ 347, 238, 21 },
		{ 347, 239, 21 },
		{ 347, 240, 21 },
		{ 347, 241, 21 },
		{ 347, 242, 21 },
		{ 347, 243, 21 },
		{ 347, 244, 21 },
		{ 347, 245, 21 },
		{ 347, 246, 21 },
		{ 347, 247, 21 },
		{ 347, 248, 21 },
		{ 347, 249, 21 },
		{ 347, 250, 21 },
		{ 347, 251, 21 },
		{ 347, 252, 21 },
		{ 347, 253, 21 },
		{ 347, 254, 21 },
		{ 347, 255, 21 },
		{ 347, 256, 21 },
		{ 348, 212, 21 },
		{ 348, 213, 21 },
		{ 348, 214, 21 },
		{ 348, 215, 21 },
		{ 348, 216, 21 },
		{ 348, 217, 21 },
		{ 348, 218, 21 },
		{ 348, 219, 21 },
		{ 348, 220, 21 },
		{ 348, 221, 21 },
		{ 348, 222, 21 },
		{ 348, 223, 21 },
		{ 348, 224, 21 },
		{ 348, 225, 21 },
		{ 348, 226, 21 },
		{ 348, 227, 21 },
		{ 348, 228, 21 },
		{ 348, 229, 21 },
		{ 348, 230, 21 },
		{ 348, 231, 21 },
		{ 348, 232, 21 },
		{ 348, 233, 21 },
		{ 348, 234, 21 },
		{ 348, 235, 21 },
		{ 348, 236, 21 },
		{ 348, 237, 21 },
		{ 348, 238, 21 },
		{ 348, 239, 21 },
		{ 348, 240, 21 },
		{ 348, 241, 21 },
		{ 348, 242, 21 },
		{ 348, 243, 21 },
		{ 348, 244, 21 },
		{ 348, 245, 21 },
		{ 348, 246, 21 },
		{ 348, 247, 21 },
		{ 348, 248, 21 },
		{ 348, 249, 21 },
		{ 348, 250, 21 },
		{ 348, 251, 21 },
		{ 348, 252, 21 },
		{ 348, 253, 21 },
		{ 348, 254, 21 },
		{ 348, 255, 21 },
		{ 348, 256, 21 },
		{ 349, 212, 21 },
		{ 349, 213, 21 },
		{ 349, 214, 21 },
		{ 349, 215, 21 },
		{ 349, 216, 21 },
		{ 349, 217, 21 },
		{ 349, 218, 21 },
		{ 349, 219, 21 },
		{ 349, 220, 21 },
		{ 349, 221, 21 },
		{ 349, 222, 21 },
		{ 349, 223, 21 },
		{ 349, 224, 21 },
		{ 349, 225, 21 },
		{ 349, 226, 21 },
		{ 349, 227, 21 },
		{ 349, 228, 21 },
		{ 349, 229, 21 },
		{ 349, 230, 21 },
		{ 349, 231, 21 },
		{ 349, 232, 21 },
		{ 349, 233, 21 },
		{ 349, 234, 21 },
		{ 349, 235, 21 },
		{ 349, 236, 21 },
		{ 349, 237, 21 },
		{ 349, 238, 21 },
		{ 349, 239, 21 },
		{ 349, 240, 21 },
		{ 349, 241, 21 },
		{ 349, 242, 21 },
		{ 349, 243, 21 },
		{ 349, 244, 21 },
		{ 349, 245, 21 },
		{ 349, 246, 21 },
		{ 349, 247, 21 },
		{ 349, 248, 21 },
		{ 349, 249, 21 },
		{ 349, 250, 21 },
		{ 349, 251, 21 },
		{ 349, 252, 21 },
		{ 349, 253, 21 },
		{ 349, 254, 21 },
		{ 349, 255, 21 },
		{ 349, 256, 21 },
		{ 350, 211, 21 },
		{ 350, 212, 21 },
		{ 350, 213, 21 },
		{ 350, 214, 21 },
		{ 350, 215, 21 },
		{ 350, 216, 21 },
		{ 350, 217, 21 },
		{ 350, 218, 21 },
		{ 350, 219, 21 },
		{ 350, 220, 21 },
		{ 350, 221, 21 },
		{ 350, 222, 21 },
		{ 350, 223, 21 },
		{ 350, 224, 21 },
		{ 350, 225, 21 },
		{ 350, 226, 21 },
		{ 350, 227, 21 },
		{ 350, 228, 21 },
		{ 350, 229, 21 },
		{ 350, 230, 21 },
		{ 350, 231, 21 },
		{ 350, 232, 21 },
		{ 350, 233, 21 },
		{ 350, 234, 21 },
		{ 350, 235, 21 },
		{ 350, 236, 21 },
		{ 350, 237, 21 },
		{ 350, 238, 21 },
		{ 350, 239, 21 },
		{ 350, 240, 21 },
		{ 350, 241, 21 },
		{ 350, 242, 21 },
		{ 350, 243, 21 },
		{ 350, 244, 21 },
		{ 350, 245, 21 },
		{ 350, 246, 21 },
		{ 350, 247, 21 },
		{ 350, 248, 21 },
		{ 350, 249, 21 },
		{ 350, 250, 21 },
		{ 350, 251, 21 },
		{ 350, 252, 21 },
		{ 350, 253, 21 },
		{ 350, 254, 21 },
		{ 350, 255, 21 },
		{ 350, 256, 21 },
		{ 351, 211, 21 },
		{ 351, 212, 21 },
		{ 351, 213, 21 },
		{ 351, 214, 21 },
		{ 351, 215, 21 },
		{ 351, 216, 21 },
		{ 351, 217, 21 },
		{ 351, 218, 21 },
		{ 351, 219, 21 },
		{ 351, 220, 21 },
		{ 351, 221, 21 },
		{ 351, 222, 21 },
		{ 351, 223, 21 },
		{ 351, 224, 21 },
		{ 351, 225, 21 },
		{ 351, 226, 21 },
		{ 351, 227, 21 },
		{ 351, 228, 21 },
		{ 351, 229, 21 },
		{ 351, 230, 21 },
		{ 351, 231, 21 },
		{ 351, 232, 21 },
		{ 351, 233, 21 },
		{ 351, 234, 21 },
		{ 351, 235, 21 },
		{ 351, 236, 21 },
		{ 351, 237, 21 },
		{ 351, 238, 21 },
		{ 351, 239, 21 },
		{ 351, 240, 21 },
		{ 351, 241, 21 },
		{ 351, 242, 21 },
		{ 351, 243, 21 },
		{ 351, 244, 21 },
		{ 351, 245, 21 },
		{ 351, 246, 21 },
		{ 351, 247, 21 },
		{ 351, 248, 21 },
		{ 351, 249, 21 },
		{ 351, 250, 21 },
		{ 351, 251, 21 },
		{ 351, 252, 21 },
		{ 351, 253, 21 },
		{ 351, 254, 21 },
		{ 351, 255, 21 },
		{ 351, 256, 21 },
		{ 352, 210, 21 },
		{ 352, 211, 21 },
		{ 352, 212, 21 },
		{ 352, 213, 21 },
		{ 352, 214, 21 },
		{ 352, 215, 21 },
		{ 352, 216, 21 },
		{ 352, 217, 21 },
		{ 352, 218, 21 },
		{ 352, 219, 21 },
		{ 352, 220, 21 },
		{ 352, 221, 21 },
		{ 352, 222, 21 },
		{ 352, 223, 21 },
		{ 352, 224, 21 },
		{ 352, 225, 21 },
		{ 352, 226, 21 },
		{ 352, 227, 21 },
		{ 352, 228, 21 },
		{ 352, 229, 21 },
		{ 352, 230, 21 },
		{ 352, 231, 21 },
		{ 352, 232, 21 },
		{ 352, 233, 21 },
		{ 352, 234, 21 },
		{ 352, 235, 21 },
		{ 352, 236, 21 },
		{ 352, 237, 21 },
		{ 352, 238, 21 },
		{ 352, 239, 21 },
		{ 352, 240, 21 },
		{ 352, 241, 21 },
		{ 352, 242, 21 },
		{ 352, 243, 21 },
		{ 352, 244, 21 },
		{ 352, 245, 21 },
		{ 352, 246, 21 },
		{ 352, 247, 21 },
		{ 352, 248, 21 },
		{ 352, 249, 21 },
		{ 352, 250, 21 },
		{ 352, 251, 21 },
		{ 352, 252, 21 },
		{ 352, 253, 21 },
		{ 352, 254, 21 },
		{ 352, 255, 21 },
		{ 352, 256, 21 },
		{ 353, 210, 21 },
		{ 353, 211, 21 },
		{ 353, 212, 21 },
		{ 353, 213, 21 },
		{ 353, 214, 21 },
		{ 353, 215, 21 },
		{ 353, 216, 21 },
		{ 353, 217, 21 },
		{ 353, 218, 21 },
		{ 353, 219, 21 },
		{ 353, 220, 21 },
		{ 353, 221, 21 },
		{ 353, 222, 21 },
		{ 353, 223, 21 },
		{ 353, 224, 21 },
		{ 353, 225, 21 },
		{ 353, 226, 21 },
		{ 353, 227, 21 },
		{ 353, 228, 21 },
		{ 353, 229, 21 },
		{ 353, 230, 21 },
		{ 353, 231, 21 },
		{ 353, 232, 21 },
		{ 353, 233, 21 },
		{ 353, 234, 21 },
		{ 353, 235, 21 },
		{ 353, 236, 21 },
		{ 353, 237, 21 },
		{ 353, 238, 21 },
		{ 353, 239, 21 },
		{ 353, 240, 21 },
		{ 353, 241, 21 },
		{ 353, 242, 21 },
		{ 353, 243, 21 },
		{ 353, 244, 21 },
		{ 353, 245, 21 },
		{ 353, 246, 21 },
		{ 353, 247, 21 },
		{ 353, 248, 21 },
		{ 353, 249, 21 },
		{ 353, 250, 21 },
		{ 353, 251, 21 },
		{ 353, 252, 21 },
		{ 353, 253, 21 },
		{ 353, 254, 21 },
		{ 353, 255, 21 },
		{ 353, 256, 21 },
		{ 354, 209, 21 },
		{ 354, 210, 21 },
		{ 354, 211, 21 },
		{ 354, 212, 21 },
		{ 354, 213, 21 },
		{ 354, 214, 21 },
		{ 354, 215, 21 },
		{ 354, 216, 21 },
		{ 354, 217, 21 },
		{ 354, 218, 21 },
		{ 354, 219, 21 },
		{ 354, 220, 21 },
		{ 354, 221, 21 },
		{ 354, 222, 21 },
		{ 354, 223, 21 },
		{ 354, 224, 21 },
		{ 354, 225, 21 },
		{ 354, 226, 21 },
		{ 354, 227, 21 },
		{ 354, 228, 21 },
		{ 354, 229, 21 },
		{ 354, 230, 21 },
		{ 354, 231, 21 },
		{ 354, 232, 21 },
		{ 354, 233, 21 },
		{ 354, 234, 21 },
		{ 354, 235, 21 },
		{ 354, 236, 21 },
		{ 354, 237, 21 },
		{ 354, 238, 21 },
		{ 354, 239, 21 },
		{ 354, 240, 21 },
		{ 354, 241, 21 },
		{ 354, 242, 21 },
		{ 354, 243, 21 },
		{ 354, 244, 21 },
		{ 354, 245, 21 },
		{ 354, 246, 21 },
		{ 354, 247, 21 },
		{ 354, 248, 21 },
		{ 354, 249, 21 },
		{ 354, 250, 21 },
		{ 354, 251, 21 },
		{ 354, 252, 21 },
		{ 354, 253, 21 },
		{ 354, 254, 21 },
		{ 354, 255, 21 },
		{ 354, 256, 21 },
		{ 355, 209, 21 },
		{ 355, 210, 21 },
		{ 355, 211, 21 },
		{ 355, 212, 21 },
		{ 355, 213, 21 },
		{ 355, 214, 21 },
		{ 355, 215, 21 },
		{ 355, 216, 21 },
		{ 355, 217, 21 },
		{ 355, 218, 21 },
		{ 355, 219, 21 },
		{ 355, 220, 21 },
		{ 355, 221, 21 },
		{ 355, 222, 21 },
		{ 355, 223, 21 },
		{ 355, 224, 21 },
		{ 355, 225, 21 },
		{ 355, 226, 21 },
		{ 355, 227, 21 },
		{ 355, 228, 21 },
		{ 355, 229, 21 },
		{ 355, 230, 21 },
		{ 355, 231, 21 },
		{ 355, 232, 21 },
		{ 355, 233, 21 },
		{ 355, 234, 21 },
		{ 355, 235, 21 },
		{ 355, 236, 21 },
		{ 355, 237, 21 },
		{ 355, 238, 21 },
		{ 355, 239, 21 },
		{ 355, 240, 21 },
		{ 355, 241, 21 },
		{ 355, 242, 21 },
		{ 355, 243, 21 },
		{ 355, 244, 21 },
		{ 355, 245, 21 },
		{ 355, 246, 21 },
		{ 355, 247, 21 },
		{ 355, 248, 21 },
		{ 355, 249, 21 },
		{ 355, 250, 21 },
		{ 355, 251, 21 },
		{ 355, 252, 21 },
		{ 355, 253, 21 },
		{ 355, 254, 21 },
		{ 355, 255, 21 },
		{ 355, 256, 21 },
		{ 356, 208, 21 },
		{ 356, 209, 21 },
		{ 356, 210, 21 },
		{ 356, 211, 21 },
		{ 356, 212, 21 },
		{ 356, 213, 21 },
		{ 356, 214, 21 },
		{ 356, 215, 21 },
		{ 356, 216, 21 },
		{ 356, 217, 21 },
		{ 356, 218, 21 },
		{ 356, 219, 21 },
		{ 356, 220, 21 },
		{ 356, 221, 21 },
		{ 356, 222, 21 },
		{ 356, 223, 21 },
		{ 356, 224, 21 },
		{ 356, 225, 21 },
		{ 356, 226, 21 },
		{ 356, 227, 21 },
		{ 356, 228, 21 },
		{ 356, 229, 21 },
		{ 356, 230, 21 },
		{ 356, 231, 21 },
		{ 356, 232, 21 },
		{ 356, 233, 21 },
		{ 356, 234, 21 },
		{ 356, 235, 21 },
		{ 356, 236, 21 },
		{ 356, 237, 21 },
		{ 356, 238, 21 },
		{ 356, 239, 21 },
		{ 356, 240, 21 },
		{ 356, 241, 21 },
		{ 356, 242, 21 },
		{ 356, 243, 21 },
		{ 356, 244, 21 },
		{ 356, 245, 21 },
		{ 356, 246, 21 },
		{ 356, 247, 21 },
		{ 356, 248, 21 },
		{ 356, 249, 21 },
		{ 356, 250, 21 },
		{ 356, 251, 21 },
		{ 356, 252, 21 },
		{ 356, 253, 21 },
		{ 356, 254, 21 },
		{ 356, 255, 21 },
		{ 356, 256, 21 },
		{ 357, 208, 21 },
		{ 357, 209, 21 },
		{ 357, 210, 21 },
		{ 357, 211, 21 },
		{ 357, 212, 21 },
		{ 357, 213, 21 },
		{ 357, 214, 21 },
		{ 357, 215, 21 },
		{ 357, 216, 21 },
		{ 357, 217, 21 },
		{ 357, 218, 21 },
		{ 357, 219, 21 },
		{ 357, 220, 21 },
		{ 357, 221, 21 },
		{ 357, 222, 21 },
		{ 357, 223, 21 },
		{ 357, 224, 21 },
		{ 357, 225, 21 },
		{ 357, 226, 21 },
		{ 357, 227, 21 },
		{ 357, 228, 21 },
		{ 357, 229, 21 },
		{ 357, 230, 21 },
		{ 357, 231, 21 },
		{ 357, 232, 21 },
		{ 357, 233, 21 },
		{ 357, 234, 21 },
		{ 357, 235, 21 },
		{ 357, 236, 21 },
		{ 357, 237, 21 },
		{ 357, 238, 21 },
		{ 357, 239, 21 },
		{ 357, 240, 21 },
		{ 357, 241, 21 },
		{ 357, 242, 21 },
		{ 357, 243, 21 },
		{ 357, 244, 21 },
		{ 357, 245, 21 },
		{ 357, 246, 21 },
		{ 357, 247, 21 },
		{ 357, 248, 21 },
		{ 357, 249, 21 },
		{ 357, 250, 21 },
		{ 357, 251, 21 },
		{ 357, 252, 21 },
		{ 357, 253, 21 },
		{ 357, 254, 21 },
		{ 357, 255, 21 },
		{ 357, 256, 21 },
		{ 358, 207, 21 },
		{ 358, 208, 21 },
		{ 358, 209, 21 },
		{ 358, 210, 21 },
		{ 358, 211, 21 },
		{ 358, 212, 21 },
		{ 358, 213, 21 },
		{ 358, 214, 21 },
		{ 358, 215, 21 },
		{ 358, 216, 21 },
		{ 358, 217, 21 },
		{ 358, 218, 21 },
		{ 358, 219, 21 },
		{ 358, 220, 21 },
		{ 358, 221, 21 },
		{ 358, 222, 21 },
		{ 358, 223, 21 },
		{ 358, 224, 21 },
		{ 358, 225, 21 },
		{ 358, 226, 21 },
		{ 358, 227, 21 },
		{ 358, 228, 21 },
		{ 358, 229, 21 },
		{ 358, 230, 21 },
		{ 358, 231, 21 },
		{ 358, 232, 21 },
		{ 358, 233, 21 },
		{ 358, 234, 21 },
		{ 358, 235, 21 },
		{ 358, 236, 21 },
		{ 358, 237, 21 },
		{ 358, 238, 21 },
		{ 358, 239, 21 },
		{ 358, 240, 21 },
		{ 358, 241, 21 },
		{ 358, 242, 21 },
		{ 358, 243, 21 },
		{ 358, 244, 21 },
		{ 358, 245, 21 },
		{ 358, 246, 21 },
		{ 358, 247, 21 },
		{ 358, 248, 21 },
		{ 358, 249, 21 },
		{ 358, 250, 21 },
		{ 358, 251, 21 },
		{ 358, 252, 21 },
		{ 358, 253, 21 },
		{ 358, 254, 21 },
		{ 358, 255, 21 },
		{ 358, 256, 21 },
		{ 359, 207, 21 },
		{ 359, 208, 21 },
		{ 359, 209, 21 },
		{ 359, 210, 21 },
		{ 359, 211, 21 },
		{ 359, 212, 21 },
		{ 359, 213, 21 },
		{ 359, 214, 21 },
		{ 359, 215, 21 },
		{ 359, 216, 21 },
		{ 359, 217, 21 },
		{ 359, 218, 21 },
		{ 359, 219, 21 },
		{ 359, 220, 21 },
		{ 359, 221, 21 },
		{ 359, 222, 21 },
		{ 359, 223, 21 },
		{ 359, 224, 21 },
		{ 359, 225, 21 },
		{ 359, 226, 21 },
		{ 359, 227, 21 },
		{ 359, 228, 21 },
		{ 359, 229, 21 },
		{ 359, 230, 21 },
		{ 359, 231, 21 },
		{ 359, 232, 21 },
		{ 359, 233, 21 },
		{ 359, 234, 21 },
		{ 359, 235, 21 },
		{ 359, 236, 21 },
		{ 359, 237, 21 },
		{ 359, 238, 21 },
		{ 359, 239, 21 },
		{ 359, 240, 21 },
		{ 359, 241, 21 },
		{ 359, 242, 21 },
		{ 359, 243, 21 },
		{ 359, 244, 21 },
		{ 359, 245, 21 },
		{ 359, 246, 21 },
		{ 359, 247, 21 },
		{ 359, 248, 21 },
		{ 359, 249, 21 },
		{ 359, 250, 21 },
		{ 359, 251, 21 },
		{ 359, 252, 21 },
		{ 359, 253, 21 },
		{ 359, 254, 21 },
		{ 359, 255, 21 },
		{ 359, 256, 21 },
		{ 360, 206, 21 },
		{ 360, 207, 21 },
		{ 360, 208, 21 },
		{ 360, 209, 21 },
		{ 360, 210, 21 },
		{ 360, 211, 21 },
		{ 360, 212, 21 },
		{ 360, 213, 21 },
		{ 360, 214, 21 },
		{ 360, 215, 21 },
		{ 360, 216, 21 },
		{ 360, 217, 21 },
		{ 360, 218, 21 },
		{ 360, 219, 21 },
		{ 360, 220, 21 },
		{ 360, 221, 21 },
		{ 360, 222, 21 },
		{ 360, 223, 21 },
		{ 360, 224, 21 },
		{ 360, 225, 21 },
		{ 360, 226, 21 },
		{ 360, 227, 21 },
		{ 360, 228, 21 },
		{ 360, 229, 21 },
		{ 360, 230, 21 },
		{ 360, 231, 21 },
		{ 360, 232, 21 },
		{ 360, 233, 21 },
		{ 360, 234, 21 },
		{ 360, 235, 21 },
		{ 360, 236, 21 },
		{ 360, 237, 21 },
		{ 360, 238, 21 },
		{ 360, 239, 21 },
		{ 360, 240, 21 },
		{ 360, 241, 21 },
		{ 360, 242, 21 },
		{ 360, 243, 21 },
		{ 360, 244, 21 },
		{ 360, 245, 21 },
		{ 360, 246, 21 },
		{ 360, 247, 21 },
		{ 360, 248, 21 },
		{ 360, 249, 21 },
		{ 360, 250, 21 },
		{ 360, 251, 21 },
		{ 360, 252, 21 },
		{ 360, 253, 21 },
		{ 360, 254, 21 },
		{ 360, 255, 21 },
		{ 360, 256, 21 },
		{ 361, 206, 21 },
		{ 361, 207, 21 },
		{ 361, 208, 21 },
		{ 361, 209, 21 },
		{ 361, 210, 21 },
		{ 361, 211, 21 },
		{ 361, 212, 21 },
		{ 361, 213, 21 },
		{ 361, 214, 21 },
		{ 361, 215, 21 },
		{ 361, 216, 21 },
		{ 361, 217, 21 },
		{ 361, 218, 21 },
		{ 361, 219, 21 },
		{ 361, 220, 21 },
		{ 361, 221, 21 },
		{ 361, 222, 21 },
		{ 361, 223, 21 },
		{ 361, 224, 21 },
		{ 361, 225, 21 },
		{ 361, 226, 21 },
		{ 361, 227, 21 },
		{ 361, 228, 21 },
		{ 361, 229, 21 },
		{ 361, 230, 21 },
		{ 361, 231, 21 },
		{ 361, 232, 21 },
		{ 361, 233, 21 },
		{ 361, 234, 21 },
		{ 361, 235, 21 },
		{ 361, 236, 21 },
		{ 361, 237, 21 },
		{ 361, 238, 21 },
		{ 361, 239, 21 },
		{ 361, 240, 21 },
		{ 361, 241, 21 },
		{ 361, 242, 21 },
		{ 361, 243, 21 },
		{ 361, 244, 21 },
		{ 361, 245, 21 },
		{ 361, 246, 21 },
		{ 361, 247, 21 },
		{ 361, 248, 21 },
		{ 361, 249, 21 },
		{ 361, 250, 21 },
		{ 361, 251, 21 },
		{ 361, 252, 21 },
		{ 361, 253, 21 },
		{ 361, 254, 21 },
		{ 361, 255, 21 },
		{ 361, 256, 21 },
		{ 362, 205, 21 },
		{ 362, 206, 21 },
		{ 362, 207, 21 },
		{ 362, 208, 21 },
		{ 362, 209, 21 },
		{ 362, 210, 21 },
		{ 362, 211, 21 },
		{ 362, 212, 21 },
		{ 362, 213, 21 },
		{ 362, 214, 21 },
		{ 362, 215, 21 },
		{ 362, 216, 21 },
		{ 362, 217, 21 },
		{ 362, 218, 21 },
		{ 362, 219, 21 },
		{ 362, 220, 21 },
		{ 362, 221, 21 },
		{ 362, 222, 21 },
		{ 362, 223, 21 },
		{ 362, 224, 21 },
		{ 362, 225, 21 },
		{ 362, 226, 21 },
		{ 362, 227, 21 },
		{ 362, 228, 21 },
		{ 362, 229, 21 },
		{ 362, 230, 21 },
		{ 362, 231, 21 },
		{ 362, 232, 21 },
		{ 362, 233, 21 },
		{ 362, 234, 21 },
		{ 362, 235, 21 },
		{ 362, 236, 21 },
		{ 362, 237, 21 },
		{ 362, 238, 21 },
		{ 362, 239, 21 },
		{ 362, 240, 21 },
		{ 362, 241, 21 },
		{ 362, 242, 21 },
		{ 362, 243, 21 },
		{ 362, 244, 21 },
		{ 362, 245, 21 },
		{ 362, 246, 21 },
		{ 362, 247, 21 },
		{ 362, 248, 21 },
		{ 362, 249, 21 },
		{ 362, 250, 21 },
		{ 362, 251, 21 },
		{ 362, 252, 21 },
		{ 362, 253, 21 },
		{ 362, 254, 21 },
		{ 362, 255, 21 },
		{ 362, 256, 21 },
		{ 363, 205, 21 },
		{ 363, 206, 21 },
		{ 363, 207, 21 },
		{ 363, 208, 21 },
		{ 363, 209, 21 },
		{ 363, 210, 21 },
		{ 363, 211, 21 },
		{ 363, 212, 21 },
		{ 363, 213, 21 },
		{ 363, 214, 21 },
		{ 363, 215, 21 },
		{ 363, 216, 21 },
		{ 363, 217, 21 },
		{ 363, 218, 21 },
		{ 363, 219, 21 },
		{ 363, 220, 21 },
		{ 363, 221, 21 },
		{ 363, 222, 21 },
		{ 363, 223, 21 },
		{ 363, 224, 21 },
		{ 363, 225, 21 },
		{ 363, 226, 21 },
		{ 363, 227, 21 },
		{ 363, 228, 21 },
		{ 363, 229, 21 },
		{ 363, 230, 21 },
		{ 363, 231, 21 },
		{ 363, 232, 21 },
		{ 363, 233, 21 },
		{ 363, 234, 21 },
		{ 363, 235, 21 },
		{ 363, 236, 21 },
		{ 363, 237, 21 },
		{ 363, 238, 21 },
		{ 363, 239, 21 },
		{ 363, 240, 21 },
		{ 363, 241, 21 },
		{ 363, 242, 21 },
		{ 363, 243, 21 },
		{ 363, 244, 21 },
		{ 363, 245, 21 },
		{ 363, 246, 21 },
		{ 363, 247, 21 },
		{ 363, 248, 21 },
		{ 363, 249, 21 },
		{ 363, 250, 21 },
		{ 363, 251, 21 },
		{ 363, 252, 21 },
		{ 363, 253, 21 },
		{ 363, 254, 21 },
		{ 363, 255, 21 },
		{ 363, 256, 21 },
		{ 364, 204, 21 },
		{ 364, 205, 21 },
		{ 364, 206, 21 },
		{ 364, 207, 21 },
		{ 364, 208, 21 },
		{ 364, 209, 21 },
		{ 364, 210, 21 },
		{ 364, 211, 21 },
		{ 364, 212, 21 },
		{ 364, 213, 21 },
		{ 364, 214, 21 },
		{ 364, 215, 21 },
		{ 364, 216, 21 },
		{ 364, 217, 21 },
		{ 364, 218, 21 },
		{ 364, 219, 21 },
		{ 364, 220, 21 },
		{ 364, 221, 21 },
		{ 364, 222, 21 },
		{ 364, 223, 21 },
		{ 364, 224, 21 },
		{ 364, 225, 21 },
		{ 364, 226, 21 },
		{ 364, 227, 21 },
		{ 364, 228, 21 },
		{ 364, 229, 21 },
		{ 364, 230, 21 },
		{ 364, 231, 21 },
		{ 364, 232, 21 },
		{ 364, 233, 21 },
		{ 364, 234, 21 },
		{ 364, 235, 21 },
		{ 364, 236, 21 },
		{ 364, 237, 21 },
		{ 364, 238, 21 },
		{ 364, 239, 21 },
		{ 364, 240, 21 },
		{ 364, 241, 21 },
		{ 364, 242, 21 },
		{ 364, 243, 21 },
		{ 364, 244, 21 },
		{ 364, 245, 21 },
		{ 364, 246, 21 },
		{ 364, 247, 21 },
		{ 364, 248, 21 },
		{ 364, 249, 21 },
		{ 364, 250, 21 },
		{ 364, 251, 21 },
		{ 364, 252, 21 },
		{ 364, 253, 21 },
		{ 364, 254, 21 },
		{ 364, 255, 21 },
		{ 364, 256, 21 },
		{ 365, 204, 21 },
		{ 365, 205, 21 },
		{ 365, 206, 21 },
		{ 365, 207, 21 },
		{ 365, 208, 21 },
		{ 365, 209, 21 },
		{ 365, 210, 21 },
		{ 365, 211, 21 },
		{ 365, 212, 21 },
		{ 365, 213, 21 },
		{ 365, 214, 21 },
		{ 365, 215, 21 },
		{ 365, 216, 21 },
		{ 365, 217, 21 },
		{ 365, 218, 21 },
		{ 365, 219, 21 },
		{ 365, 220, 21 },
		{ 365, 221, 21 },
		{ 365, 222, 21 },
		{ 365, 223, 21 },
		{ 365, 224, 21 },
		{ 365, 225, 21 },
		{ 365, 226, 21 },
		{ 365, 227, 21 },
		{ 365, 228, 21 },
		{ 365, 229, 21 },
		{ 365, 230, 21 },
		{ 365, 231, 21 },
		{ 365, 232, 21 },
		{ 365, 233, 21 },
		{ 365, 234, 21 },
		{ 365, 235, 21 },
		{ 365, 236, 21 },
		{ 365, 237, 21 },
		{ 365, 238, 21 },
		{ 365, 239, 21 },
		{ 365, 240, 21 },
		{ 365, 241, 21 },
		{ 365, 242, 21 },
		{ 365, 243, 21 },
		{ 365, 244, 21 },
		{ 365, 245, 21 },
		{ 365, 246, 21 },
		{ 365, 247, 21 },
		{ 365, 248, 21 },
		{ 365, 249, 21 },
		{ 365, 250, 21 },
		{ 365, 251, 21 },
		{ 365, 252, 21 },
		{ 365, 253, 21 },
		{ 365, 254, 21 },
		{ 365, 255, 21 },
		{ 365, 256, 21 },
		{ 366, 203, 21 },
		{ 366, 204, 21 },
		{ 366, 205, 21 },
		{ 366, 206, 21 },
		{ 366, 207, 21 },
		{ 366, 208, 21 },
		{ 366, 209, 21 },
		{ 366, 210, 21 },
		{ 366, 211, 21 },
		{ 366, 212, 21 },
		{ 366, 213, 21 },
		{ 366, 214, 21 },
		{ 366, 215, 21 },
		{ 366, 216, 21 },
		{ 366, 217, 21 },
		{ 366, 218, 21 },
		{ 366, 219, 21 },
		{ 366, 220, 21 },
		{ 366, 221, 21 },
		{ 366, 222, 21 },
		{ 366, 223, 21 },
		{ 366, 224, 21 },
		{ 366, 225, 21 },
		{ 366, 226, 21 },
		{ 366, 227, 21 },
		{ 366, 228, 21 },
		{ 366, 229, 21 },
		{ 366, 230, 21 },
		{ 366, 231, 21 },
		{ 366, 232, 21 },
		{ 366, 233, 21 },
		{ 366, 234, 21 },
		{ 366, 235, 21 },
		{ 366, 236, 21 },
		{ 366, 237, 21 },
		{ 366, 238, 21 },
		{ 366, 239, 21 },
		{ 366, 240, 21 },
		{ 366, 241, 21 },
		{ 366, 242, 21 },
		{ 366, 243, 21 },
		{ 366, 244, 21 },
		{ 366, 245, 21 },
		{ 366, 246, 21 },
		{ 366, 247, 21 },
		{ 366, 248, 21 },
		{ 366, 249, 21 },
		{ 366, 250, 21 },
		{ 366, 251, 21 },
		{ 366, 252, 21 },
		{ 366, 253, 21 },
		{ 366, 254, 21 },
		{ 366, 255, 21 },
		{ 366, 256, 21 },
		{ 367, 203, 21 },
		{ 367, 204, 21 },
		{ 367, 205, 21 },
		{ 367, 206, 21 },
		{ 367, 207, 21 },
		{ 367, 208, 21 },
		{ 367, 209, 21 },
		{ 367, 210, 21 },
		{ 367, 211, 21 },
		{ 367, 212, 21 },
		{ 367, 213, 21 },
		{ 367, 214, 21 },
		{ 367, 215, 21 },
		{ 367, 216, 21 },
		{ 367, 217, 21 },
		{ 367, 218, 21 },
		{ 367, 219, 21 },
		{ 367, 220, 21 },
		{ 367, 221, 21 },
		{ 367, 222, 21 },
		{ 367, 223, 21 },
		{ 367, 224, 21 },
		{ 367, 225, 21 },
		{ 367, 226, 21 },
		{ 367, 227, 21 },
		{ 367, 228, 21 },
		{ 367, 229, 21 },
		{ 367, 230, 21 },
		{ 367, 231, 21 },
		{ 367, 232, 21 },
		{ 367, 233, 21 },
		{ 367, 234, 21 },
		{ 367, 235, 21 },
		{ 367, 236, 21 },
		{ 367, 237, 21 },
		{ 367, 238, 21 },
		{ 367, 239, 21 },
		{ 367, 240, 21 },
		{ 367, 241, 21 },
		{ 367, 242, 21 },
		{ 367, 243, 21 },
		{ 367, 244, 21 },
		{ 367, 245, 21 },
		{ 367, 246, 21 },
		{ 367, 247, 21 },
		{ 367, 248, 21 },
		{ 367, 249, 21 },
		{ 367, 250, 21 },
		{ 367, 251, 21 },
		{ 367, 252, 21 },
		{ 367, 253, 21 },
		{ 367, 254, 21 },
		{ 367, 255, 21 },
		{ 367, 256, 21 },
		{ 368, 202, 21 },
		{ 368, 203, 21 },
		{ 368, 204, 21 },
		{ 368, 205, 21 },
		{ 368, 206, 21 },
		{ 368, 207, 21 },
		{ 368, 208, 21 },
		{ 368, 209, 21 },
		{ 368, 210, 21 },
		{ 368, 211, 21 },
		{ 368, 212, 21 },
		{ 368, 213, 21 },
		{ 368, 214, 21 },
		{ 368, 215, 21 },
		{ 368, 216, 21 },
		{ 368, 217, 21 },
		{ 368, 218, 21 },
		{ 368, 219, 21 },
		{ 368, 220, 21 },
		{ 368, 221, 21 },
		{ 368, 222, 21 },
		{ 368, 223, 21 },
		{ 368, 224, 21 },
		{ 368, 225, 21 },
		{ 368, 226, 21 },
		{ 368, 227, 21 },
		{ 368, 228, 21 },
		{ 368, 229, 21 },
		{ 368, 230, 21 },
		{ 368, 231, 21 },
		{ 368, 232, 21 },
		{ 368, 233, 21 },
		{ 368, 234, 21 },
		{ 368, 235, 21 },
		{ 368, 236, 21 },
		{ 368, 237, 21 },
		{ 368, 238, 21 },
		{ 368, 239, 21 },
		{ 368, 240, 21 },
		{ 368, 241, 21 },
		{ 368, 242, 21 },
		{ 368, 243, 21 },
		{ 368, 244, 21 },
		{ 368, 245, 21 },
		{ 368, 246, 21 },
		{ 368, 247, 21 },
		{ 368, 248, 21 },
		{ 368, 249, 21 },
		{ 368, 250, 21 },
		{ 368, 251, 21 },
		{ 368, 252, 21 },
		{ 368, 253, 21 },
		{ 368, 254, 21 },
		{ 368, 255, 21 },
		{ 368, 256, 21 },
		{ 369, 202, 21 },
		{ 369, 203, 21 },
		{ 369, 204, 21 },
		{ 369, 205, 21 },
		{ 369, 206, 21 },
		{ 369, 207, 21 },
		{ 369, 208, 21 },
		{ 369, 209, 21 },
		{ 369, 210, 21 },
		{ 369, 211, 21 },
		{ 369, 212, 21 },
		{ 369, 213, 21 },
		{ 369, 214, 21 },
		{ 369, 215, 21 },
		{ 369, 216, 21 },
		{ 369, 217, 21 },
		{ 369, 218, 21 },
		{ 369, 219, 21 },
		{ 369, 220, 21 },
		{ 369, 221, 21 },
		{ 369, 222, 21 },
		{ 369, 223, 21 },
		{ 369, 224, 21 },
		{ 369, 225, 21 },
		{ 369, 226, 21 },
		{ 369, 227, 21 },
		{ 369, 228, 21 },
		{ 369, 229, 21 },
		{ 369, 230, 21 },
		{ 369, 231, 21 },
		{ 369, 232, 21 },
		{ 369, 233, 21 },
		{ 369, 234, 21 },
		{ 369, 235, 21 },
		{ 369, 236, 21 },
		{ 369, 237, 21 },
		{ 369, 238, 21 },
		{ 369, 239, 21 },
		{ 369, 240, 21 },
		{ 369, 241, 21 },
		{ 369, 242, 21 },
		{ 369, 243, 21 },
		{ 369, 244, 21 },
		{ 369, 245, 21 },
		{ 369, 246, 21 },
		{ 369, 247, 21 },
		{ 369, 248, 21 },
		{ 369, 249, 21 },
		{ 369, 250, 21 },
		{ 369, 251, 21 },
		{ 369, 252, 21 },
		{ 369, 253, 21 },
		{ 369, 254, 21 },
		{ 369, 255, 21 },
		{ 369, 256, 21 },
		{ 370, 201, 21 },
		{ 370, 202, 21 },
		{ 370, 203, 21 },
		{ 370, 204, 21 },
		{ 370, 205, 21 },
		{ 370, 206, 21 },
		{ 370, 207, 21 },
		{ 370, 208, 21 },
		{ 370, 209, 21 },
		{ 370, 210, 21 },
		{ 370, 211, 21 },
		{ 370, 212, 21 },
		{ 370, 213, 21 },
		{ 370, 214, 21 },
		{ 370, 215, 21 },
		{ 370, 216, 21 },
		{ 370, 217, 21 },
		{ 370, 218, 21 },
		{ 370, 219, 21 },
		{ 370, 220, 21 },
		{ 370, 221, 21 },
		{ 370, 222, 21 },
		{ 370, 223, 21 },
		{ 370, 224, 21 },
		{ 370, 225, 21 },
		{ 370, 226, 21 },
		{ 370, 227, 21 },
		{ 370, 228, 21 },
		{ 370, 229, 21 },
		{ 370, 230, 21 },
		{ 370, 231, 21 },
		{ 370, 232, 21 },
		{ 370, 233, 21 },
		{ 370, 234, 21 },
		{ 370, 235, 21 },
		{ 370, 236, 21 },
		{ 370, 237, 21 },
		{ 370, 238, 21 },
		{ 370, 239, 21 },
		{ 370, 240, 21 },
		{ 370, 241, 21 },
		{ 370, 242, 21 },
		{ 370, 243, 21 },
		{ 370, 244, 21 },
		{ 370, 245, 21 },
		{ 370, 246, 21 },
		{ 370, 247, 21 },
		{ 370, 248, 21 },
		{ 370, 249, 21 },
		{ 370, 250, 21 },
		{ 370, 251, 21 },
		{ 370, 252, 21 },
		{ 370, 253, 21 },
		{ 370, 254, 21 },
		{ 370, 255, 21 },
		{ 370, 256, 21 },
		{ 371, 201, 21 },
		{ 371, 202, 21 },
		{ 371, 203, 21 },
		{ 371, 204, 21 },
		{ 371, 205, 21 },
		{ 371, 206, 21 },
		{ 371, 207, 21 },
		{ 371, 208, 21 },
		{ 371, 209, 21 },
		{ 371, 210, 21 },
		{ 371, 211, 21 },
		{ 371, 212, 21 },
		{ 371, 213, 21 },
		{ 371, 214, 21 },
		{ 371, 215, 21 },
		{ 371, 216, 21 },
		{ 371, 217, 21 },
		{ 371, 218, 21 },
		{ 371, 219, 21 },
		{ 371, 220, 21 },
		{ 371, 221, 21 },
		{ 371, 222, 21 },
		{ 371, 223, 21 },
		{ 371, 224, 21 },
		{ 371, 225, 21 },
		{ 371, 226, 21 },
		{ 371, 227, 21 },
		{ 371, 228, 21 },
		{ 371, 229, 21 },
		{ 371, 230, 21 },
		{ 371, 231, 21 },
		{ 371, 232, 21 },
		{ 371, 233, 21 },
		{ 371, 234, 21 },
		{ 371, 235, 21 },
		{ 371, 236, 21 },
		{ 371, 237, 21 },
		{ 371, 238, 21 },
		{ 371, 239, 21 },
		{ 371, 240, 21 },
		{ 371, 241, 21 },
		{ 371, 242, 21 },
		{ 371, 243, 21 },
		{ 371, 244, 21 },
		{ 371, 245, 21 },
		{ 371, 246, 21 },
		{ 371, 247, 21 },
		{ 371, 248, 21 },
		{ 371, 249, 21 },
		{ 371, 250, 21 },
		{ 371, 251, 21 },
		{ 371, 252, 21 },
		{ 371, 253, 21 },
		{ 371, 254, 21 },
		{ 371, 255, 21 },
		{ 371, 256, 21 },
		{ 372, 200, 21 },
		{ 372, 201, 21 },
		{ 372, 202, 21 },
		{ 372, 203, 21 },
		{ 372, 204, 21 },
		{ 372, 205, 21 },
		{ 372, 206, 21 },
		{ 372, 207, 21 },
		{ 372, 208, 21 },
		{ 372, 209, 21 },
		{ 372, 210, 21 },
		{ 372, 211, 21 },
		{ 372, 212, 21 },
		{ 372, 213, 21 },
		{ 372, 214, 21 },
		{ 372, 215, 21 },
		{ 372, 216, 21 },
		{ 372, 217, 21 },
		{ 372, 218, 21 },
		{ 372, 219, 21 },
		{ 372, 220, 21 },
		{ 372, 221, 21 },
		{ 372, 222, 21 },
		{ 372, 223, 21 },
		{ 372, 224, 21 },
		{ 372, 225, 21 },
		{ 372, 226, 21 },
		{ 372, 227, 21 },
		{ 372, 228, 21 },
		{ 372, 229, 21 },
		{ 372, 230, 21 },
		{ 372, 231, 21 },
		{ 372, 232, 21 },
		{ 372, 233, 21 },
		{ 372, 234, 21 },
		{ 372, 235, 21 },
		{ 372, 236, 21 },
		{ 372, 237, 21 },
		{ 372, 238, 21 },
		{ 372, 239, 21 },
		{ 372, 240, 21 },
		{ 372, 241, 21 },
		{ 372, 242, 21 },
		{ 372, 243, 21 },
		{ 372, 244, 21 },
		{ 372, 245, 21 },
		{ 372, 246, 21 },
		{ 372, 247, 21 },
		{ 372, 248, 21 },
		{ 372, 249, 21 },
		{ 372, 250, 21 },
		{ 372, 251, 21 },
		{ 372, 252, 21 },
		{ 372, 253, 21 },
		{ 372, 254, 21 },
		{ 372, 255, 21 },
		{ 372, 256, 21 },
		{ 373, 200, 21 },
		{ 373, 201, 21 },
		{ 373, 202, 21 },
		{ 373, 203, 21 },
		{ 373, 204, 21 },
		{ 373, 205, 21 },
		{ 373, 206, 21 },
		{ 373, 207, 21 },
		{ 373, 208, 21 },
		{ 373, 209, 21 },
		{ 373, 210, 21 },
		{ 373, 211, 21 },
		{ 373, 212, 21 },
		{ 373, 213, 21 },
		{ 373, 214, 21 },
		{ 373, 215, 21 },
		{ 373, 216, 21 },
		{ 373, 217, 21 },
		{ 373, 218, 21 },
		{ 373, 219, 21 },
		{ 373, 220, 21 },
		{ 373, 221, 21 },
		{ 373, 222, 21 },
		{ 373, 223, 21 },
		{ 373, 224, 21 },
		{ 373, 225, 21 },
		{ 373, 226, 21 },
		{ 373, 227, 21 },
		{ 373, 228, 21 },
		{ 373, 229, 21 },
		{ 373, 230, 21 },
		{ 373, 231, 21 },
		{ 373, 232, 21 },
		{ 373, 233, 21 },
		{ 373, 234, 21 },
		{ 373, 235, 21 },
		{ 373, 236, 21 },
		{ 373, 237, 21 },
		{ 373, 238, 21 },
		{ 373, 239, 21 },
		{ 373, 240, 21 },
		{ 373, 241, 21 },
		{ 373, 242, 21 },
		{ 373, 243, 21 },
		{ 373, 244, 21 },
		{ 373, 245, 21 },
		{ 373, 246, 21 },
		{ 373, 247, 21 },
		{ 373, 248, 21 },
		{ 373, 249, 21 },
		{ 373, 250, 21 },
		{ 373, 251, 21 },
		{ 373, 252, 21 },
		{ 373, 253, 21 },
		{ 373, 254, 21 },
		{ 373, 255, 21 },
		{ 373, 256, 21 },
		{ 374, 199, 21 },
		{ 374, 200, 21 },
		{ 374, 201, 21 },
		{ 374, 202, 21 },
		{ 374, 203, 21 },
		{ 374, 204, 21 },
		{ 374, 205, 21 },
		{ 374, 206, 21 },
		{ 374, 207, 21 },
		{ 374, 208, 21 },
		{ 374, 209, 21 },
		{ 374, 210, 21 },
		{ 374, 211, 21 },
		{ 374, 212, 21 },
		{ 374, 213, 21 },
		{ 374, 214, 21 },
		{ 374, 215, 21 },
		{ 374, 216, 21 },
		{ 374, 217, 21 },
		{ 374, 218, 21 },
		{ 374, 219, 21 },
		{ 374, 220, 21 },
		{ 374, 221, 21 },
		{ 374, 222, 21 },
		{ 374, 223, 21 },
		{ 374, 224, 21 },
		{ 374, 225, 21 },
		{ 374, 226, 21 },
		{ 374, 227, 21 },
		{ 374, 228, 21 },
		{ 374, 229, 21 },
		{ 374, 230, 21 },
		{ 374, 231, 21 },
		{ 374, 232, 21 },
		{ 374, 233, 21 },
		{ 374, 234, 21 },
		{ 374, 235, 21 },
		{ 374, 236, 21 },
		{ 374, 237, 21 },
		{ 374, 238, 21 },
		{ 374, 239, 21 },
		{ 374, 240, 21 },
		{ 374, 241, 21 },
		{ 374, 242, 21 },
		{ 374, 243, 21 },
		{ 374, 244, 21 },
		{ 374, 245, 21 },
		{ 374, 246, 21 },
		{ 374, 247, 21 },
		{ 374, 248, 21 },
		{ 374, 249, 21 },
		{ 374, 250, 21 },
		{ 374, 251, 21 },
		{ 374, 252, 21 },
		{ 374, 253, 21 },
		{ 374, 254, 21 },
		{ 374, 255, 21 },
		{ 374, 256, 21 },
		{ 375, 199, 21 },
		{ 375, 200, 21 },
		{ 375, 201, 21 },
		{ 375, 202, 21 },
		{ 375, 203, 21 },
		{ 375, 204, 21 },
		{ 375, 205, 21 },
		{ 375, 206, 21 },
		{ 375, 207, 21 },
		{ 375, 208, 21 },
		{ 375, 209, 21 },
		{ 375, 210, 21 },
		{ 375, 211, 21 },
		{ 375, 212, 21 },
		{ 375, 213, 21 },
		{ 375, 214, 21 },
		{ 375, 215, 21 },
		{ 375, 216, 21 },
		{ 375, 217, 21 },
		{ 375, 218, 21 },
		{ 375, 219, 21 },
		{ 375, 220, 21 },
		{ 375, 221, 21 },
		{ 375, 222, 21 },
		{ 375, 223, 21 },
		{ 375, 224, 21 },
		{ 375, 225, 21 },
		{ 375, 226, 21 },
		{ 375, 227, 21 },
		{ 375, 228, 21 },
		{ 375, 229, 21 },
		{ 375, 230, 21 },
		{ 375, 231, 21 },
		{ 375, 232, 21 },
		{ 375, 233, 21 },
		{ 375, 234, 21 },
		{ 375, 235, 21 },
		{ 375, 236, 21 },
		{ 375, 237, 21 },
		{ 375, 238, 21 },
		{ 375, 239, 21 },
		{ 375, 240, 21 },
		{ 375, 241, 21 },
		{ 375, 242, 21 },
		{ 375, 243, 21 },
		{ 375, 244, 21 },
		{ 375, 245, 21 },
		{ 375, 246, 21 },
		{ 375, 247, 21 },
		{ 375, 248, 21 },
		{ 375, 249, 21 },
		{ 375, 250, 21 },
		{ 375, 251, 21 },
		{ 375, 252, 21 },
		{ 375, 253, 21 },
		{ 375, 254, 21 },
		{ 375, 255, 21 },
		{ 375, 256, 21 },
		{ 376, 198, 21 },
		{ 376, 199, 21 },
		{ 376, 200, 21 },
		{ 376, 201, 21 },
		{ 376, 202, 21 },
		{ 376, 203, 21 },
		{ 376, 204, 21 },
		{ 376, 205, 21 },
		{ 376, 206, 21 },
		{ 376, 207, 21 },
		{ 376, 208, 21 },
		{ 376, 209, 21 },
		{ 376, 210, 21 },
		{ 376, 211, 21 },
		{ 376, 212, 21 },
		{ 376, 213, 21 },
		{ 376, 214, 21 },
		{ 376, 215, 21 },
		{ 376, 216, 21 },
		{ 376, 217, 21 },
		{ 376, 218, 21 },
		{ 376, 219, 21 },
		{ 376, 220, 21 },
		{ 376, 221, 21 },
		{ 376, 222, 21 },
		{ 376, 223, 21 },
		{ 376, 224, 21 },
		{ 376, 225, 21 },
		{ 376, 226, 21 },
		{ 376, 227, 21 },
		{ 376, 228, 21 },
		{ 376, 229, 21 },
		{ 376, 230, 21 },
		{ 376, 231, 21 },
		{ 376, 232, 21 },
		{ 376, 233, 21 },
		{ 376, 234, 21 },
		{ 376, 235, 21 },
		{ 376, 236, 21 },
		{ 376, 237, 21 },
		{ 376, 238, 21 },
		{ 376, 239, 21 },
		{ 376, 240, 21 },
		{ 376, 241, 21 },
		{ 376, 242, 21 },
		{ 376, 243, 21 },
		{ 376, 244, 21 },
		{ 376, 245, 21 },
		{ 376, 246, 21 },
		{ 376, 247, 21 },
		{ 376, 248, 21 },
		{ 376, 249, 21 },
		{ 376, 250, 21 },
		{ 376, 251, 21 },
		{ 376, 252, 21 },
		{ 376, 253, 21 },
		{ 376, 254, 21 },
		{ 376, 255, 21 },
		{ 376, 256, 21 },
		{ 377, 198, 21 },
		{ 377, 199, 21 },
		{ 377, 200, 21 },
		{ 377, 201, 21 },
		{ 377, 202, 21 },
		{ 377, 203, 21 },
		{ 377, 204, 21 },
		{ 377, 205, 21 },
		{ 377, 206, 21 },
		{ 377, 207, 21 },
		{ 377, 208, 21 },
		{ 377, 209, 21 },
		{ 377, 210, 21 },
		{ 377, 211, 21 },
		{ 377, 212, 21 },
		{ 377, 213, 21 },
		{ 377, 214, 21 },
		{ 377, 215, 21 },
		{ 377, 216, 21 },
		{ 377, 217, 21 },
		{ 377, 218, 21 },
		{ 377, 219, 21 },
		{ 377, 220, 21 },
		{ 377, 221, 21 },
		{ 377, 222, 21 },
		{ 377, 223, 21 },
		{ 377, 224, 21 },
		{ 377, 225, 21 },
		{ 377, 226, 21 },
		{ 377, 227, 21 },
		{ 377, 228, 21 },
		{ 377, 229, 21 },
		{ 377, 230, 21 },
		{ 377, 231, 21 },
		{ 377, 232, 21 },
		{ 377, 233, 21 },
		{ 377, 234, 21 },
		{ 377, 235, 21 },
		{ 377, 236, 21 },
		{ 377, 237, 21 },
		{ 377, 238, 21 },
		{ 377, 239, 21 },
		{ 377, 240, 21 },
		{ 377, 241, 21 },
		{ 377, 242, 21 },
		{ 377, 243, 21 },
		{ 377, 244, 21 },
		{ 377, 245, 21 },
		{ 377, 246, 21 },
		{ 377, 247, 21 },
		{ 377, 248, 21 },
		{ 377, 249, 21 },
		{ 377, 250, 21 },
		{ 377, 251, 21 },
		{ 377, 252, 21 },
		{ 377, 253, 21 },
		{ 377, 254, 21 },
		{ 377, 255, 21 },
		{ 377, 256, 21 },
		{ 378, 198, 21 },
		{ 378, 199, 21 },
		{ 378, 200, 21 },
		{ 378, 201, 21 },
		{ 378, 202, 21 },
		{ 378, 203, 21 },
		{ 378, 204, 21 },
		{ 378, 205, 21 },
		{ 378, 206, 21 },
		{ 378, 207, 21 },
		{ 378, 208, 21 },
		{ 378, 209, 21 },
		{ 378, 210, 21 },
		{ 378, 211, 21 },
		{ 378, 212, 21 },
		{ 378, 213, 21 },
		{ 378, 214, 21 },
		{ 378, 215, 21 },
		{ 378, 216, 21 },
		{ 378, 217, 21 },
		{ 378, 218, 21 },
		{ 378, 219, 21 },
		{ 378, 220, 21 },
		{ 378, 221, 21 },
		{ 378, 222, 21 },
		{ 378, 223, 21 },
		{ 378, 224, 21 },
		{ 378, 225, 21 },
		{ 378, 226, 21 },
		{ 378, 227, 21 },
		{ 378, 228, 21 },
		{ 378, 229, 21 },
		{ 378, 230, 21 },
		{ 378, 231, 21 },
		{ 378, 232, 21 },
		{ 378, 233, 21 },
		{ 378, 234, 21 },
		{ 378, 235, 21 },
		{ 378, 236, 21 },
		{ 378, 237, 21 },
		{ 378, 238, 21 },
		{ 378, 239, 21 },
		{ 378, 240, 21 },
		{ 378, 241, 21 },
		{ 378, 242, 21 },
		{ 378, 243, 21 },
		{ 378, 244, 21 },
		{ 378, 245, 21 },
		{ 378, 246, 21 },
		{ 378, 247, 21 },
		{ 378, 248, 21 },
		{ 378, 249, 21 },
		{ 378, 250, 21 },
		{ 378, 251, 21 },
		{ 378, 252, 21 },
		{ 378, 253, 21 },
		{ 378, 254, 21 },
		{ 378, 255, 21 },
		{ 378, 256, 21 },
		{ 379, 197, 21 },
		{ 379, 198, 21 },
		{ 379, 199, 21 },
		{ 379, 200, 21 },
		{ 379, 201, 21 },
		{ 379, 202, 21 },
		{ 379, 203, 21 },
		{ 379, 204, 21 },
		{ 379, 205, 21 },
		{ 379, 206, 21 },
		{ 379, 207, 21 },
		{ 379, 208, 21 },
		{ 379, 209, 21 },
		{ 379, 210, 21 },
		{ 379, 211, 21 },
		{ 379, 212, 21 },
		{ 379, 213, 21 },
		{ 379, 214, 21 },
		{ 379, 215, 21 },
		{ 379, 216, 21 },
		{ 379, 217, 21 },
		{ 379, 218, 21 },
		{ 379, 219, 21 },
		{ 379, 220, 21 },
		{ 379, 221, 21 },
		{ 379, 222, 21 },
		{ 379, 223, 21 },
		{ 379, 224, 21 },
		{ 379, 225, 21 },
		{ 379, 226, 21 },
		{ 379, 227, 21 },
		{ 379, 228, 21 },
		{ 379, 229, 21 },
		{ 379, 230, 21 },
		{ 379, 231, 21 },
		{ 379, 232, 21 },
		{ 379, 233, 21 },
		{ 379, 234, 21 },
		{ 379, 235, 21 },
		{ 379, 236, 21 },
		{ 379, 237, 21 },
		{ 379, 238, 21 },
		{ 379, 239, 21 },
		{ 379, 240, 21 },
		{ 379, 241, 21 },
		{ 379, 242, 21 },
		{ 379, 243, 21 },
		{ 379, 244, 21 },
		{ 379, 245, 21 },
		{ 379, 246, 21 },
		{ 379, 247, 21 },
		{ 379, 248, 21 },
		{ 379, 249, 21 },
		{ 379, 250, 21 },
		{ 379, 251, 21 },
		{ 379, 252, 21 },
		{ 379, 253, 21 },
		{ 379, 254, 21 },
		{ 379, 255, 21 },
		{ 379, 256, 21 },
		{ 380, 197, 21 },
		{ 380, 198, 21 },
		{ 380, 199, 21 },
		{ 380, 200, 21 },
		{ 380, 201, 21 },
		{ 380, 202, 21 },
		{ 380, 203, 21 },
		{ 380, 204, 21 },
		{ 380, 205, 21 },
		{ 380, 206, 21 },
		{ 380, 207, 21 },
		{ 380, 208, 21 },
		{ 380, 209, 21 },
		{ 380, 210, 21 },
		{ 380, 211, 21 },
		{ 380, 212, 21 },
		{ 380, 213, 21 },
		{ 380, 214, 21 },
		{ 380, 215, 21 },
		{ 380, 216, 21 },
		{ 380, 217, 21 },
		{ 380, 218, 21 },
		{ 380, 219, 21 },
		{ 380, 220, 21 },
		{ 380, 221, 21 },
		{ 380, 222, 21 },
		{ 380, 223, 21 },
		{ 380, 224, 21 },
		{ 380, 225, 21 },
		{ 380, 226, 21 },
		{ 380, 227, 21 },
		{ 380, 228, 21 },
		{ 380, 229, 21 },
		{ 380, 230, 21 },
		{ 380, 231, 21 },
		{ 380, 232, 21 },
		{ 380, 233, 21 },
		{ 380, 234, 21 },
		{ 380, 235, 21 },
		{ 380, 236, 21 },
		{ 380, 237, 21 },
		{ 380, 238, 21 },
		{ 380, 239, 21 },
		{ 380, 240, 21 },
		{ 380, 241, 21 },
		{ 380, 242, 21 },
		{ 380, 243, 21 },
		{ 380, 244, 21 },
		{ 380, 245, 21 },
		{ 380, 246, 21 },
		{ 380, 247, 21 },
		{ 380, 248, 21 },
		{ 380, 249, 21 },
		{ 380, 250, 21 },
		{ 380, 251, 21 },
		{ 380, 252, 21 },
		{ 380, 253, 21 },
		{ 380, 254, 21 },
		{ 380, 255, 21 },
		{ 380, 256, 21 },
		{ 381, 197, 21 },
		{ 381, 198, 21 },
		{ 381, 199, 21 },
		{ 381, 200, 21 },
		{ 381, 201, 21 },
		{ 381, 202, 21 },
		{ 381, 203, 21 },
		{ 381, 204, 21 },
		{ 381, 205, 21 },
		{ 381, 206, 21 },
		{ 381, 207, 21 },
		{ 381, 208, 21 },
		{ 381, 209, 21 },
		{ 381, 210, 21 },
		{ 381, 211, 21 },
		{ 381, 212, 21 },
		{ 381, 213, 21 },
		{ 381, 214, 21 },
		{ 381, 215, 21 },
		{ 381, 216, 21 },
		{ 381, 217, 21 },
		{ 381, 218, 21 },
		{ 381, 219, 21 },
		{ 381, 220, 21 },
		{ 381, 221, 21 },
		{ 381, 222, 21 },
		{ 381, 223, 21 },
		{ 381, 224, 21 },
		{ 381, 225, 21 },
		{ 381, 226, 21 },
		{ 381, 227, 21 },
		{ 381, 228, 21 },
		{ 381, 229, 21 },
		{ 381, 230, 21 },
		{ 381, 231, 21 },
		{ 381, 232, 21 },
		{ 381, 233, 21 },
		{ 381, 234, 21 },
		{ 381, 235, 21 },
		{ 381, 236, 21 },
		{ 381, 237, 21 },
		{ 381, 238, 21 },
		{ 381, 239, 21 },
		{ 381, 240, 21 },
		{ 381, 241, 21 },
		{ 381, 242, 21 },
		{ 381, 243, 21 },
		{ 381, 244, 21 },
		{ 381, 245, 21 },
		{ 381, 246, 21 },
		{ 381, 247, 21 },
		{ 381, 248, 21 },
		{ 381, 249, 21 },
		{ 381, 250, 21 },
		{ 381, 251, 21 },
		{ 381, 252, 21 },
		{ 381, 253, 21 },
		{ 381, 254, 21 },
		{ 381, 255, 21 },
		{ 382, 198, 21 },
		{ 382, 199, 21 },
		{ 382, 200, 21 },
		{ 382, 201, 21 },
		{ 382, 202, 21 },
		{ 382, 203, 21 },
		{ 382, 204, 21 },
		{ 382, 205, 21 },
		{ 382, 206, 21 },
		{ 382, 207, 21 },
		{ 382, 208, 21 },
		{ 382, 209, 21 },
		{ 382, 210, 21 },
		{ 382, 211, 21 },
		{ 382, 212, 21 },
		{ 382, 213, 21 },
		{ 382, 214, 21 },
		{ 382, 215, 21 },
		{ 382, 216, 21 },
		{ 382, 217, 21 },
		{ 382, 218, 21 },
		{ 382, 219, 21 },
		{ 382, 220, 21 },
		{ 382, 221, 21 },
		{ 382, 222, 21 },
		{ 382, 223, 21 },
		{ 382, 224, 21 },
		{ 382, 225, 21 },
		{ 382, 226, 21 },
		{ 382, 227, 21 },
		{ 382, 228, 21 },
		{ 382, 229, 21 },
		{ 382, 230, 21 },
		{ 382, 231, 21 },
		{ 382, 232, 21 },
		{ 382, 233, 21 },
		{ 382, 234, 21 },
		{ 382, 235, 21 },
		{ 382, 236, 21 },
		{ 382, 237, 21 },
		{ 382, 238, 21 },
		{ 382, 239, 21 },
		{ 382, 240, 21 },
		{ 382, 241, 21 },
		{ 382, 242, 21 },
		{ 382, 243, 21 },
		{ 382, 244, 21 },
		{ 382, 245, 21 },
		{ 382, 246, 21 },
		{ 382, 247, 21 },
		{ 382, 248, 21 },
		{ 382, 249, 21 },
		{ 382, 250, 21 },
		{ 382, 251, 21 },
		{ 382, 252, 21 },
		{ 382, 253, 21 },
		{ 382, 254, 21 },
		{ 383, 201, 21 },
		{ 383, 202, 21 },
		{ 383, 203, 21 },
		{ 383, 204, 21 },
		{ 383, 205, 21 },
		{ 383, 206, 21 },
		{ 383, 207, 21 },
		{ 383, 208, 21 },
		{ 383, 209, 21 },
		{ 383, 210, 21 },
		{ 383, 211, 21 },
		{ 383, 212, 21 },
		{ 383, 213, 21 },
		{ 383, 214, 21 },
		{ 383, 215, 21 },
		{ 383, 216, 21 },
		{ 383, 217, 21 },
		{ 383, 218, 21 },
		{ 383, 219, 21 },
		{ 383, 220, 21 },
		{ 383, 221, 21 },
		{ 383, 222, 21 },
		{ 383, 223, 21 },
		{ 383, 224, 21 },
		{ 383, 225, 21 },
		{ 383, 226, 21 },
		{ 383, 227, 21 },
		{ 383, 228, 21 },
		{ 383, 229, 21 },
		{ 383, 230, 21 },
		{ 383, 231, 21 },
		{ 383, 232, 21 },
		{ 383, 233, 21 },
		{ 383, 234, 21 },
		{ 383, 235, 21 },
		{ 383, 236, 21 },
		{ 383, 237, 21 },
		{ 383, 238, 21 },
		{ 383, 239, 21 },
		{ 383, 240, 21 },
		{ 383, 241, 21 },
		{ 383, 242, 21 },
		{ 383, 243, 21 },
		{ 383, 244, 21 },
		{ 383, 245, 21 },
		{ 383, 246, 21 },
		{ 383, 247, 21 },
		{ 383, 248, 21 },
		{ 383, 249, 21 },
		{ 383, 250, 21 },
		{ 131, 233, 22 },
		{ 131, 234, 22 },
		{ 131, 235, 22 },
		{ 131, 236, 22 },
		{ 131, 237, 22 },
		{ 131, 238, 22 },
		{ 131, 239, 22 },
		{ 131, 240, 22 },
		{ 131, 241, 22 },
		{ 131, 242, 22 },
		{ 131, 243, 22 },
		{ 131, 244, 22 },
		{ 131, 245, 22 },
		{ 131, 246, 22 },
		{ 131, 247, 22 },
		{ 131, 248, 22 },
		{ 131, 249, 22 },
		{ 131, 250, 22 },
		{ 131, 251, 22 },
		{ 131, 252, 22 },
		{ 131, 253, 22 },
		{ 131, 254, 22 },
		{ 131, 255, 22 },
		{ 131, 256, 22 },
		{ 131, 257, 22 },
		{ 131, 258, 22 },
		{ 131, 259, 22 },
		{ 131, 260, 22 },
		{ 131, 261, 22 },
		{ 131, 262, 22 },
		{ 131, 263, 22 },
		{ 131, 264, 22 },
		{ 131, 265, 22 },
		{ 131, 266, 22 },
		{ 131, 267, 22 },
		{ 131, 268, 22 },
		{ 131, 269, 22 },
		{ 131, 270, 22 },
		{ 131, 271, 22 },
		{ 131, 272, 22 },
		{ 131, 273, 22 },
		{ 131, 274, 22 },
		{ 131, 275, 22 },
		{ 131, 276, 22 },
		{ 131, 277, 22 },
		{ 131, 278, 22 },
		{ 131, 279, 22 },
		{ 131, 280, 22 },
		{ 131, 281, 22 },
		{ 132, 229, 22 },
		{ 132, 230, 22 },
		{ 132, 231, 22 },
		{ 132, 232, 22 },
		{ 132, 233, 22 },
		{ 132, 234, 22 },
		{ 132, 235, 22 },
		{ 132, 236, 22 },
		{ 132, 237, 22 },
		{ 132, 238, 22 },
		{ 132, 239, 22 },
		{ 132, 240, 22 },
		{ 132, 241, 22 },
		{ 132, 242, 22 },
		{ 132, 243, 22 },
		{ 132, 244, 22 },
		{ 132, 245, 22 },
		{ 132, 246, 22 },
		{ 132, 247, 22 },
		{ 132, 248, 22 },
		{ 132, 249, 22 },
		{ 132, 250, 22 },
		{ 132, 251, 22 },
		{ 132, 252, 22 },
		{ 132, 253, 22 },
		{ 132, 254, 22 },
		{ 132, 255, 22 },
		{ 132, 256, 22 },
		{ 132, 257, 22 },
		{ 132, 258, 22 },
		{ 132, 259, 22 },
		{ 132, 260, 22 },
		{ 132, 261, 22 },
		{ 132, 262, 22 },
		{ 132, 263, 22 },
		{ 132, 264, 22 },
		{ 132, 265, 22 },
		{ 132, 266, 22 },
		{ 132, 267, 22 },
		{ 132, 268, 22 },
		{ 132, 269, 22 },
		{ 132, 270, 22 },
		{ 132, 271, 22 },
		{ 132, 272, 22 },
		{ 132, 273, 22 },
		{ 132, 274, 22 },
		{ 132, 275, 22 },
		{ 132, 276, 22 },
		{ 132, 277, 22 },
		{ 132, 278, 22 },
		{ 132, 279, 22 },
		{ 132, 280, 22 },
		{ 132, 281, 22 },
		{ 132, 282, 22 },
		{ 132, 283, 22 },
		{ 132, 284, 22 },
		{ 132, 285, 22 },
		{ 133, 228, 22 },
		{ 133, 229, 22 },
		{ 133, 230, 22 },
		{ 133, 231, 22 },
		{ 133, 232, 22 },
		{ 133, 233, 22 },
		{ 133, 234, 22 },
		{ 133, 235, 22 },
		{ 133, 236, 22 },
		{ 133, 237, 22 },
		{ 133, 238, 22 },
		{ 133, 239, 22 },
		{ 133, 240, 22 },
		{ 133, 241, 22 },
		{ 133, 242, 22 },
		{ 133, 243, 22 },
		{ 133, 244, 22 },
		{ 133, 245, 22 },
		{ 133, 246, 22 },
		{ 133, 247, 22 },
		{ 133, 248, 22 },
		{ 133, 249, 22 },
		{ 133, 250, 22 },
		{ 133, 251, 22 },
		{ 133, 252, 22 },
		{ 133, 253, 22 },
		{ 133, 254, 22 },
		{ 133, 255, 22 },
		{ 133, 256, 22 },
		{ 133, 257, 22 },
		{ 133, 258, 22 },
		{ 133, 259, 22 },
		{ 133, 260, 22 },
		{ 133, 261, 22 },
		{ 133, 262, 22 },
		{ 133, 263, 22 },
		{ 133, 264, 22 },
		{ 133, 265, 22 },
		{ 133, 266, 22 },
		{ 133, 267, 22 },
		{ 133, 268, 22 },
		{ 133, 269, 22 },
		{ 133, 270, 22 },
		{ 133, 271, 22 },
		{ 133, 272, 22 },
		{ 133, 273, 22 },
		{ 133, 274, 22 },
		{ 133, 275, 22 },
		{ 133, 276, 22 },
		{ 133, 277, 22 },
		{ 133, 278, 22 },
		{ 133, 279, 22 },
		{ 133, 280, 22 },
		{ 133, 281, 22 },
		{ 133, 282, 22 },
		{ 133, 283, 22 },
		{ 133, 284, 22 },
		{ 133, 285, 22 },
		{ 133, 286, 22 },
		{ 134, 228, 22 },
		{ 134, 229, 22 },
		{ 134, 230, 22 },
		{ 134, 231, 22 },
		{ 134, 232, 22 },
		{ 134, 233, 22 },
		{ 134, 234, 22 },
		{ 134, 235, 22 },
		{ 134, 236, 22 },
		{ 134, 237, 22 },
		{ 134, 238, 22 },
		{ 134, 239, 22 },
		{ 134, 240, 22 },
		{ 134, 241, 22 },
		{ 134, 242, 22 },
		{ 134, 243, 22 },
		{ 134, 244, 22 },
		{ 134, 245, 22 },
		{ 134, 246, 22 },
		{ 134, 247, 22 },
		{ 134, 248, 22 },
		{ 134, 249, 22 },
		{ 134, 250, 22 },
		{ 134, 251, 22 },
		{ 134, 252, 22 },
		{ 134, 253, 22 },
		{ 134, 254, 22 },
		{ 134, 255, 22 },
		{ 134, 256, 22 },
		{ 134, 257, 22 },
		{ 134, 258, 22 },
		{ 134, 259, 22 },
		{ 134, 260, 22 },
		{ 134, 261, 22 },
		{ 134, 262, 22 },
		{ 134, 263, 22 },
		{ 134, 264, 22 },
		{ 134, 265, 22 },
		{ 134, 266, 22 },
		{ 134, 267, 22 },
		{ 134, 268, 22 },
		{ 134, 269, 22 },
		{ 134, 270, 22 },
		{ 134, 271, 22 },
		{ 134, 272, 22 },
		{ 134, 273, 22 },
		{ 134, 274, 22 },
		{ 134, 275, 22 },
		{ 134, 276, 22 },
		{ 134, 277, 22 },
		{ 134, 278, 22 },
		{ 134, 279, 22 },
		{ 134, 280, 22 },
		{ 134, 281, 22 },
		{ 134, 282, 22 },
		{ 134, 283, 22 },
		{ 134, 284, 22 },
		{ 134, 285, 22 },
		{ 134, 286, 22 },
		{ 135, 228, 22 },
		{ 135, 229, 22 },
		{ 135, 230, 22 },
		{ 135, 231, 22 },
		{ 135, 232, 22 },
		{ 135, 233, 22 },
		{ 135, 234, 22 },
		{ 135, 235, 22 },
		{ 135, 236, 22 },
		{ 135, 237, 22 },
		{ 135, 238, 22 },
		{ 135, 239, 22 },
		{ 135, 240, 22 },
		{ 135, 241, 22 },
		{ 135, 242, 22 },
		{ 135, 243, 22 },
		{ 135, 244, 22 },
		{ 135, 245, 22 },
		{ 135, 246, 22 },
		{ 135, 247, 22 },
		{ 135, 248, 22 },
		{ 135, 249, 22 },
		{ 135, 250, 22 },
		{ 135, 251, 22 },
		{ 135, 252, 22 },
		{ 135, 253, 22 },
		{ 135, 254, 22 },
		{ 135, 255, 22 },
		{ 135, 256, 22 },
		{ 135, 257, 22 },
		{ 135, 258, 22 },
		{ 135, 259, 22 },
		{ 135, 260, 22 },
		{ 135, 261, 22 },
		{ 135, 262, 22 },
		{ 135, 263, 22 },
		{ 135, 264, 22 },
		{ 135, 265, 22 },
		{ 135, 266, 22 },
		{ 135, 267, 22 },
		{ 135, 268, 22 },
		{ 135, 269, 22 },
		{ 135, 270, 22 },
		{ 135, 271, 22 },
		{ 135, 272, 22 },
		{ 135, 273, 22 },
		{ 135, 274, 22 },
		{ 135, 275, 22 },
		{ 135, 276, 22 },
		{ 135, 277, 22 },
		{ 135, 278, 22 },
		{ 135, 279, 22 },
		{ 135, 280, 22 },
		{ 135, 281, 22 },
		{ 135, 282, 22 },
		{ 135, 283, 22 },
		{ 135, 284, 22 },
		{ 135, 285, 22 },
		{ 135, 286, 22 },
		{ 136, 228, 22 },
		{ 136, 229, 22 },
		{ 136, 230, 22 },
		{ 136, 231, 22 },
		{ 136, 232, 22 },
		{ 136, 233, 22 },
		{ 136, 234, 22 },
		{ 136, 235, 22 },
		{ 136, 236, 22 },
		{ 136, 237, 22 },
		{ 136, 238, 22 },
		{ 136, 239, 22 },
		{ 136, 240, 22 },
		{ 136, 241, 22 },
		{ 136, 242, 22 },
		{ 136, 243, 22 },
		{ 136, 244, 22 },
		{ 136, 245, 22 },
		{ 136, 246, 22 },
		{ 136, 247, 22 },
		{ 136, 248, 22 },
		{ 136, 249, 22 },
		{ 136, 250, 22 },
		{ 136, 251, 22 },
		{ 136, 252, 22 },
		{ 136, 253, 22 },
		{ 136, 254, 22 },
		{ 136, 255, 22 },
		{ 136, 256, 22 },
		{ 136, 257, 22 },
		{ 136, 258, 22 },
		{ 136, 259, 22 },
		{ 136, 260, 22 },
		{ 136, 261, 22 },
		{ 136, 262, 22 },
		{ 136, 263, 22 },
		{ 136, 264, 22 },
		{ 136, 265, 22 },
		{ 136, 266, 22 },
		{ 136, 267, 22 },
		{ 136, 268, 22 },
		{ 136, 269, 22 },
		{ 136, 270, 22 },
		{ 136, 271, 22 },
		{ 136, 272, 22 },
		{ 136, 273, 22 },
		{ 136, 274, 22 },
		{ 136, 275, 22 },
		{ 136, 276, 22 },
		{ 136, 277, 22 },
		{ 136, 278, 22 },
		{ 136, 279, 22 },
		{ 136, 280, 22 },
		{ 136, 281, 22 },
		{ 136, 282, 22 },
		{ 136, 283, 22 },
		{ 136, 284, 22 },
		{ 136, 285, 22 },
		{ 136, 286, 22 },
		{ 137, 228, 22 },
		{ 137, 229, 22 },
		{ 137, 230, 22 },
		{ 137, 231, 22 },
		{ 137, 232, 22 },
		{ 137, 233, 22 },
		{ 137, 234, 22 },
		{ 137, 235, 22 },
		{ 137, 236, 22 },
		{ 137, 237, 22 },
		{ 137, 238, 22 },
		{ 137, 239, 22 },
		{ 137, 240, 22 },
		{ 137, 241, 22 },
		{ 137, 242, 22 },
		{ 137, 243, 22 },
		{ 137, 244, 22 },
		{ 137, 245, 22 },
		{ 137, 246, 22 },
		{ 137, 247, 22 },
		{ 137, 248, 22 },
		{ 137, 249, 22 },
		{ 137, 250, 22 },
		{ 137, 251, 22 },
		{ 137, 252, 22 },
		{ 137, 253, 22 },
		{ 137, 254, 22 },
		{ 137, 255, 22 },
		{ 137, 256, 22 },
		{ 137, 257, 22 },
		{ 137, 258, 22 },
		{ 137, 259, 22 },
		{ 137, 260, 22 },
		{ 137, 261, 22 },
		{ 137, 262, 22 },
		{ 137, 263, 22 },
		{ 137, 264, 22 },
		{ 137, 265, 22 },
		{ 137, 266, 22 },
		{ 137, 267, 22 },
		{ 137, 268, 22 },
		{ 137, 269, 22 },
		{ 137, 270, 22 },
		{ 137, 271, 22 },
		{ 137, 272, 22 },
		{ 137, 273, 22 },
		{ 137, 274, 22 },
		{ 137, 275, 22 },
		{ 137, 276, 22 },
		{ 137, 277, 22 },
		{ 137, 278, 22 },
		{ 137, 279, 22 },
		{ 137, 280, 22 },
		{ 137, 281, 22 },
		{ 137, 282, 22 },
		{ 137, 283, 22 },
		{ 137, 284, 22 },
		{ 137, 285, 22 },
		{ 137, 286, 22 },
		{ 138, 228, 22 },
		{ 138, 229, 22 },
		{ 138, 230, 22 },
		{ 138, 231, 22 },
		{ 138, 232, 22 },
		{ 138, 233, 22 },
		{ 138, 234, 22 },
		{ 138, 235, 22 },
		{ 138, 236, 22 },
		{ 138, 237, 22 },
		{ 138, 238, 22 },
		{ 138, 239, 22 },
		{ 138, 240, 22 },
		{ 138, 241, 22 },
		{ 138, 242, 22 },
		{ 138, 243, 22 },
		{ 138, 244, 22 },
		{ 138, 245, 22 },
		{ 138, 246, 22 },
		{ 138, 247, 22 },
		{ 138, 248, 22 },
		{ 138, 249, 22 },
		{ 138, 250, 22 },
		{ 138, 251, 22 },
		{ 138, 252, 22 },
		{ 138, 253, 22 },
		{ 138, 254, 22 },
		{ 138, 255, 22 },
		{ 138, 256, 22 },
		{ 138, 257, 22 },
		{ 138, 258, 22 },
		{ 138, 259, 22 },
		{ 138, 260, 22 },
		{ 138, 261, 22 },
		{ 138, 262, 22 },
		{ 138, 263, 22 },
		{ 138, 264, 22 },
		{ 138, 265, 22 },
		{ 138, 266, 22 },
		{ 138, 267, 22 },
		{ 138, 268, 22 },
		{ 138, 269, 22 },
		{ 138, 270, 22 },
		{ 138, 271, 22 },
		{ 138, 272, 22 },
		{ 138, 273, 22 },
		{ 138, 274, 22 },
		{ 138, 275, 22 },
		{ 138, 276, 22 },
		{ 138, 277, 22 },
		{ 138, 278, 22 },
		{ 138, 279, 22 },
		{ 138, 280, 22 },
		{ 138, 281, 22 },
		{ 138, 282, 22 },
		{ 138, 283, 22 },
		{ 138, 284, 22 },
		{ 138, 285, 22 },
		{ 138, 286, 22 },
		{ 139, 228, 22 },
		{ 139, 229, 22 },
		{ 139, 230, 22 },
		{ 139, 231, 22 },
		{ 139, 232, 22 },
		{ 139, 233, 22 },
		{ 139, 234, 22 },
		{ 139, 235, 22 },
		{ 139, 236, 22 },
		{ 139, 237, 22 },
		{ 139, 238, 22 },
		{ 139, 239, 22 },
		{ 139, 240, 22 },
		{ 139, 241, 22 },
		{ 139, 242, 22 },
		{ 139, 243, 22 },
		{ 139, 244, 22 },
		{ 139, 245, 22 },
		{ 139, 246, 22 },
		{ 139, 247, 22 },
		{ 139, 248, 22 },
		{ 139, 249, 22 },
		{ 139, 250, 22 },
		{ 139, 251, 22 },
		{ 139, 252, 22 },
		{ 139, 253, 22 },
		{ 139, 254, 22 },
		{ 139, 255, 22 },
		{ 139, 256, 22 },
		{ 139, 257, 22 },
		{ 139, 258, 22 },
		{ 139, 259, 22 },
		{ 139, 260, 22 },
		{ 139, 261, 22 },
		{ 139, 262, 22 },
		{ 139, 263, 22 },
		{ 139, 264, 22 },
		{ 139, 265, 22 },
		{ 139, 266, 22 },
		{ 139, 267, 22 },
		{ 139, 268, 22 },
		{ 139, 269, 22 },
		{ 139, 270, 22 },
		{ 139, 271, 22 },
		{ 139, 272, 22 },
		{ 139, 273, 22 },
		{ 139, 274, 22 },
		{ 139, 275, 22 },
		{ 139, 276, 22 },
		{ 139, 277, 22 },
		{ 139, 278, 22 },
		{ 139, 279, 22 },
		{ 139, 280, 22 },
		{ 139, 281, 22 },
		{ 139, 282, 22 },
		{ 139, 283, 22 },
		{ 139, 284, 22 },
		{ 139, 285, 22 },
		{ 139, 286, 22 },
		{ 140, 228, 22 },
		{ 140, 229, 22 },
		{ 140, 230, 22 },
		{ 140, 231, 22 },
		{ 140, 232, 22 },
		{ 140, 233, 22 },
		{ 140, 234, 22 },
		{ 140, 235, 22 },
		{ 140, 236, 22 },
		{ 140, 237, 22 },
		{ 140, 238, 22 },
		{ 140, 239, 22 },
		{ 140, 240, 22 },
		{ 140, 241, 22 },
		{ 140, 242, 22 },
		{ 140, 243, 22 },
		{ 140, 244, 22 },
		{ 140, 245, 22 },
		{ 140, 246, 22 },
		{ 140, 247, 22 },
		{ 140, 248, 22 },
		{ 140, 249, 22 },
		{ 140, 250, 22 },
		{ 140, 251, 22 },
		{ 140, 252, 22 },
		{ 140, 253, 22 },
		{ 140, 254, 22 },
		{ 140, 255, 22 },
		{ 140, 256, 22 },
		{ 140, 257, 22 },
		{ 140, 258, 22 },
		{ 140, 259, 22 },
		{ 140, 260, 22 },
		{ 140, 261, 22 },
		{ 140, 262, 22 },
		{ 140, 263, 22 },
		{ 140, 264, 22 },
		{ 140, 265, 22 },
		{ 140, 266, 22 },
		{ 140, 267, 22 },
		{ 140, 268, 22 },
		{ 140, 269, 22 },
		{ 140, 270, 22 },
		{ 140, 271, 22 },
		{ 140, 272, 22 },
		{ 140, 273, 22 },
		{ 140, 274, 22 },
		{ 140, 275, 22 },
		{ 140, 276, 22 },
		{ 140, 277, 22 },
		{ 140, 278, 22 },
		{ 140, 279, 22 },
		{ 140, 280, 22 },
		{ 140, 281, 22 },
		{ 140, 282, 22 },
		{ 140, 283, 22 },
		{ 140, 284, 22 },
		{ 140, 285, 22 },
		{ 140, 286, 22 },
		{ 141, 229, 22 },
		{ 141, 230, 22 },
		{ 141, 231, 22 },
		{ 141, 232, 22 },
		{ 141, 233, 22 },
		{ 141, 234, 22 },
		{ 141, 235, 22 },
		{ 141, 236, 22 },
		{ 141, 237, 22 },
		{ 141, 238, 22 },
		{ 141, 239, 22 },
		{ 141, 240, 22 },
		{ 141, 241, 22 },
		{ 141, 242, 22 },
		{ 141, 243, 22 },
		{ 141, 244, 22 },
		{ 141, 245, 22 },
		{ 141, 246, 22 },
		{ 141, 247, 22 },
		{ 141, 248, 22 },
		{ 141, 249, 22 },
		{ 141, 250, 22 },
		{ 141, 251, 22 },
		{ 141, 252, 22 },
		{ 141, 253, 22 },
		{ 141, 254, 22 },
		{ 141, 255, 22 },
		{ 141, 256, 22 },
		{ 141, 257, 22 },
		{ 141, 258, 22 },
		{ 141, 259, 22 },
		{ 141, 260, 22 },
		{ 141, 261, 22 },
		{ 141, 262, 22 },
		{ 141, 263, 22 },
		{ 141, 264, 22 },
		{ 141, 265, 22 },
		{ 141, 266, 22 },
		{ 141, 267, 22 },
		{ 141, 268, 22 },
		{ 141, 269, 22 },
		{ 141, 270, 22 },
		{ 141, 271, 22 },
		{ 141, 272, 22 },
		{ 141, 273, 22 },
		{ 141, 274, 22 },
		{ 141, 275, 22 },
		{ 141, 276, 22 },
		{ 141, 277, 22 },
		{ 141, 278, 22 },
		{ 141, 279, 22 },
		{ 141, 280, 22 },
		{ 141, 281, 22 },
		{ 141, 282, 22 },
		{ 141, 283, 22 },
		{ 141, 284, 22 },
		{ 141, 285, 22 },
		{ 142, 229, 22 },
		{ 142, 230, 22 },
		{ 142, 231, 22 },
		{ 142, 232, 22 },
		{ 142, 233, 22 },
		{ 142, 234, 22 },
		{ 142, 235, 22 },
		{ 142, 236, 22 },
		{ 142, 237, 22 },
		{ 142, 238, 22 },
		{ 142, 239, 22 },
		{ 142, 240, 22 },
		{ 142, 241, 22 },
		{ 142, 242, 22 },
		{ 142, 243, 22 },
		{ 142, 244, 22 },
		{ 142, 245, 22 },
		{ 142, 246, 22 },
		{ 142, 247, 22 },
		{ 142, 248, 22 },
		{ 142, 249, 22 },
		{ 142, 250, 22 },
		{ 142, 251, 22 },
		{ 142, 252, 22 },
		{ 142, 253, 22 },
		{ 142, 254, 22 },
		{ 142, 255, 22 },
		{ 142, 256, 22 },
		{ 142, 257, 22 },
		{ 142, 258, 22 },
		{ 142, 259, 22 },
		{ 142, 260, 22 },
		{ 142, 261, 22 },
		{ 142, 262, 22 },
		{ 142, 263, 22 },
		{ 142, 264, 22 },
		{ 142, 265, 22 },
		{ 142, 266, 22 },
		{ 142, 267, 22 },
		{ 142, 268, 22 },
		{ 142, 269, 22 },
		{ 142, 270, 22 },
		{ 142, 271, 22 },
		{ 142, 272, 22 },
		{ 142, 273, 22 },
		{ 142, 274, 22 },
		{ 142, 275, 22 },
		{ 142, 276, 22 },
		{ 142, 277, 22 },
		{ 142, 278, 22 },
		{ 142, 279, 22 },
		{ 142, 280, 22 },
		{ 142, 281, 22 },
		{ 142, 282, 22 },
		{ 142, 283, 22 },
		{ 142, 284, 22 },
		{ 142, 285, 22 },
		{ 143, 229, 22 },
		{ 143, 230, 22 },
		{ 143, 231, 22 },
		{ 143, 232, 22 },
		{ 143, 233, 22 },
		{ 143, 234, 22 },
		{ 143, 235, 22 },
		{ 143, 236, 22 },
		{ 143, 237, 22 },
		{ 143, 238, 22 },
		{ 143, 239, 22 },
		{ 143, 240, 22 },
		{ 143, 241, 22 },
		{ 143, 242, 22 },
		{ 143, 243, 22 },
		{ 143, 244, 22 },
		{ 143, 245, 22 },
		{ 143, 246, 22 },
		{ 143, 247, 22 },
		{ 143, 248, 22 },
		{ 143, 249, 22 },
		{ 143, 250, 22 },
		{ 143, 251, 22 },
		{ 143, 252, 22 },
		{ 143, 253, 22 },
		{ 143, 254, 22 },
		{ 143, 255, 22 },
		{ 143, 256, 22 },
		{ 143, 257, 22 },
		{ 143, 258, 22 },
		{ 143, 259, 22 },
		{ 143, 260, 22 },
		{ 143, 261, 22 },
		{ 143, 262, 22 },
		{ 143, 263, 22 },
		{ 143, 264, 22 },
		{ 143, 265, 22 },
		{ 143, 266, 22 },
		{ 143, 267, 22 },
		{ 143, 268, 22 },
		{ 143, 269, 22 },
		{ 143, 270, 22 },
		{ 143, 271, 22 },
		{ 143, 272, 22 },
		{ 143, 273, 22 },
		{ 143, 274, 22 },
		{ 143, 275, 22 },
		{ 143, 276, 22 },
		{ 143, 277, 22 },
		{ 143, 278, 22 },
		{ 143, 279, 22 },
		{ 143, 280, 22 },
		{ 143, 281, 22 },
		{ 143, 282, 22 },
		{ 143, 283, 22 },
		{ 143, 284, 22 },
		{ 143, 285, 22 },
		{ 144, 229, 22 },
		{ 144, 230, 22 },
		{ 144, 231, 22 },
		{ 144, 232, 22 },
		{ 144, 233, 22 },
		{ 144, 234, 22 },
		{ 144, 235, 22 },
		{ 144, 236, 22 },
		{ 144, 237, 22 },
		{ 144, 238, 22 },
		{ 144, 239, 22 },
		{ 144, 240, 22 },
		{ 144, 241, 22 },
		{ 144, 242, 22 },
		{ 144, 243, 22 },
		{ 144, 244, 22 },
		{ 144, 245, 22 },
		{ 144, 246, 22 },
		{ 144, 247, 22 },
		{ 144, 248, 22 },
		{ 144, 249, 22 },
		{ 144, 250, 22 },
		{ 144, 251, 22 },
		{ 144, 252, 22 },
		{ 144, 253, 22 },
		{ 144, 254, 22 },
		{ 144, 255, 22 },
		{ 144, 256, 22 },
		{ 144, 257, 22 },
		{ 144, 258, 22 },
		{ 144, 259, 22 },
		{ 144, 260, 22 },
		{ 144, 261, 22 },
		{ 144, 262, 22 },
		{ 144, 263, 22 },
		{ 144, 264, 22 },
		{ 144, 265, 22 },
		{ 144, 266, 22 },
		{ 144, 267, 22 },
		{ 144, 268, 22 },
		{ 144, 269, 22 },
		{ 144, 270, 22 },
		{ 144, 271, 22 },
		{ 144, 272, 22 },
		{ 144, 273, 22 },
		{ 144, 274, 22 },
		{ 144, 275, 22 },
		{ 144, 276, 22 },
		{ 144, 277, 22 },
		{ 144, 278, 22 },
		{ 144, 279, 22 },
		{ 144, 280, 22 },
		{ 144, 281, 22 },
		{ 144, 282, 22 },
		{ 144, 283, 22 },
		{ 144, 284, 22 },
		{ 144, 285, 22 },
		{ 145, 230, 22 },
		{ 145, 231, 22 },
		{ 145, 232, 22 },
		{ 145, 233, 22 },
		{ 145, 234, 22 },
		{ 145, 235, 22 },
		{ 145, 236, 22 },
		{ 145, 237, 22 },
		{ 145, 238, 22 },
		{ 145, 239, 22 },
		{ 145, 240, 22 },
		{ 145, 241, 22 },
		{ 145, 242, 22 },
		{ 145, 243, 22 },
		{ 145, 244, 22 },
		{ 145, 245, 22 },
		{ 145, 246, 22 },
		{ 145, 247, 22 },
		{ 145, 248, 22 },
		{ 145, 249, 22 },
		{ 145, 250, 22 },
		{ 145, 251, 22 },
		{ 145, 252, 22 },
		{ 145, 253, 22 },
		{ 145, 254, 22 },
		{ 145, 255, 22 },
		{ 145, 256, 22 },
		{ 145, 257, 22 },
		{ 145, 258, 22 },
		{ 145, 259, 22 },
		{ 145, 260, 22 },
		{ 145, 261, 22 },
		{ 145, 262, 22 },
		{ 145, 263, 22 },
		{ 145, 264, 22 },
		{ 145, 265, 22 },
		{ 145, 266, 22 },
		{ 145, 267, 22 },
		{ 145, 268, 22 },
		{ 145, 269, 22 },
		{ 145, 270, 22 },
		{ 145, 271, 22 },
		{ 145, 272, 22 },
		{ 145, 273, 22 },
		{ 145, 274, 22 },
		{ 145, 275, 22 },
		{ 145, 276, 22 },
		{ 145, 277, 22 },
		{ 145, 278, 22 },
		{ 145, 279, 22 },
		{ 145, 280, 22 },
		{ 145, 281, 22 },
		{ 145, 282, 22 },
		{ 145, 283, 22 },
		{ 145, 284, 22 },
		{ 146, 230, 22 },
		{ 146, 231, 22 },
		{ 146, 232, 22 },
		{ 146, 233, 22 },
		{ 146, 234, 22 },
		{ 146, 235, 22 },
		{ 146, 236, 22 },
		{ 146, 237, 22 },
		{ 146, 238, 22 },
		{ 146, 239, 22 },
		{ 146, 240, 22 },
		{ 146, 241, 22 },
		{ 146, 242, 22 },
		{ 146, 243, 22 },
		{ 146, 244, 22 },
		{ 146, 245, 22 },
		{ 146, 246, 22 },
		{ 146, 247, 22 },
		{ 146, 248, 22 },
		{ 146, 249, 22 },
		{ 146, 250, 22 },
		{ 146, 251, 22 },
		{ 146, 252, 22 },
		{ 146, 253, 22 },
		{ 146, 254, 22 },
		{ 146, 255, 22 },
		{ 146, 256, 22 },
		{ 146, 257, 22 },
		{ 146, 258, 22 },
		{ 146, 259, 22 },
		{ 146, 260, 22 },
		{ 146, 261, 22 },
		{ 146, 262, 22 },
		{ 146, 263, 22 },
		{ 146, 264, 22 },
		{ 146, 265, 22 },
		{ 146, 266, 22 },
		{ 146, 267, 22 },
		{ 146, 268, 22 },
		{ 146, 269, 22 },
		{ 146, 270, 22 },
		{ 146, 271, 22 },
		{ 146, 272, 22 },
		{ 146, 273, 22 },
		{ 146, 274, 22 },
		{ 146, 275, 22 },
		{ 146, 276, 22 },
		{ 146, 277, 22 },
		{ 146, 278, 22 },
		{ 146, 279, 22 },
		{ 146, 280, 22 },
		{ 146, 281, 22 },
		{ 146, 282, 22 },
		{ 146, 283, 22 },
		{ 146, 284, 22 },
		{ 147, 230, 22 },
		{ 147, 231, 22 },
		{ 147, 232, 22 },
		{ 147, 233, 22 },
		{ 147, 234, 22 },
		{ 147, 235, 22 },
		{ 147, 236, 22 },
		{ 147, 237, 22 },
		{ 147, 238, 22 },
		{ 147, 239, 22 },
		{ 147, 240, 22 },
		{ 147, 241, 22 },
		{ 147, 242, 22 },
		{ 147, 243, 22 },
		{ 147, 244, 22 },
		{ 147, 245, 22 },
		{ 147, 246, 22 },
		{ 147, 247, 22 },
		{ 147, 248, 22 },
		{ 147, 249, 22 },
		{ 147, 250, 22 },
		{ 147, 251, 22 },
		{ 147, 252, 22 },
		{ 147, 253, 22 },
		{ 147, 254, 22 },
		{ 147, 255, 22 },
		{ 147, 256, 22 },
		{ 147, 257, 22 },
		{ 147, 258, 22 },
		{ 147, 259, 22 },
		{ 147, 260, 22 },
		{ 147, 261, 22 },
		{ 147, 262, 22 },
		{ 147, 263, 22 },
		{ 147, 264, 22 },
		{ 147, 265, 22 },
		{ 147, 266, 22 },
		{ 147, 267, 22 },
		{ 147, 268, 22 },
		{ 147, 269, 22 },
		{ 147, 270, 22 },
		{ 147, 271, 22 },
		{ 147, 272, 22 },
		{ 147, 273, 22 },
		{ 147, 274, 22 },
		{ 147, 275, 22 },
		{ 147, 276, 22 },
		{ 147, 277, 22 },
		{ 147, 278, 22 },
		{ 147, 279, 22 },
		{ 147, 280, 22 },
		{ 147, 281, 22 },
		{ 147, 282, 22 },
		{ 147, 283, 22 },
		{ 147, 284, 22 },
		{ 148, 230, 22 },
		{ 148, 231, 22 },
		{ 148, 232, 22 },
		{ 148, 233, 22 },
		{ 148, 234, 22 },
		{ 148, 235, 22 },
		{ 148, 236, 22 },
		{ 148, 237, 22 },
		{ 148, 238, 22 },
		{ 148, 239, 22 },
		{ 148, 240, 22 },
		{ 148, 241, 22 },
		{ 148, 242, 22 },
		{ 148, 243, 22 },
		{ 148, 244, 22 },
		{ 148, 245, 22 },
		{ 148, 246, 22 },
		{ 148, 247, 22 },
		{ 148, 248, 22 },
		{ 148, 249, 22 },
		{ 148, 250, 22 },
		{ 148, 251, 22 },
		{ 148, 252, 22 },
		{ 148, 253, 22 },
		{ 148, 254, 22 },
		{ 148, 255, 22 },
		{ 148, 256, 22 },
		{ 148, 257, 22 },
		{ 148, 258, 22 },
		{ 148, 259, 22 },
		{ 148, 260, 22 },
		{ 148, 261, 22 },
		{ 148, 262, 22 },
		{ 148, 263, 22 },
		{ 148, 264, 22 },
		{ 148, 265, 22 },
		{ 148, 266, 22 },
		{ 148, 267, 22 },
		{ 148, 268, 22 },
		{ 148, 269, 22 },
		{ 148, 270, 22 },
		{ 148, 271, 22 },
		{ 148, 272, 22 },
		{ 148, 273, 22 },
		{ 148, 274, 22 },
		{ 148, 275, 22 },
		{ 148, 276, 22 },
		{ 148, 277, 22 },
		{ 148, 278, 22 },
		{ 148, 279, 22 },
		{ 148, 280, 22 },
		{ 148, 281, 22 },
		{ 148, 282, 22 },
		{ 148, 283, 22 },
		{ 148, 284, 22 },
		{ 149, 231, 22 },
		{ 149, 232, 22 },
		{ 149, 233, 22 },
		{ 149, 234, 22 },
		{ 149, 235, 22 },
		{ 149, 236, 22 },
		{ 149, 237, 22 },
		{ 149, 238, 22 },
		{ 149, 239, 22 },
		{ 149, 240, 22 },
		{ 149, 241, 22 },
		{ 149, 242, 22 },
		{ 149, 243, 22 },
		{ 149, 244, 22 },
		{ 149, 245, 22 },
		{ 149, 246, 22 },
		{ 149, 247, 22 },
		{ 149, 248, 22 },
		{ 149, 249, 22 },
		{ 149, 250, 22 },
		{ 149, 251, 22 },
		{ 149, 252, 22 },
		{ 149, 253, 22 },
		{ 149, 254, 22 },
		{ 149, 255, 22 },
		{ 149, 256, 22 },
		{ 149, 257, 22 },
		{ 149, 258, 22 },
		{ 149, 259, 22 },
		{ 149, 260, 22 },
		{ 149, 261, 22 },
		{ 149, 262, 22 },
		{ 149, 263, 22 },
		{ 149, 264, 22 },
		{ 149, 265, 22 },
		{ 149, 266, 22 },
		{ 149, 267, 22 },
		{ 149, 268, 22 },
		{ 149, 269, 22 },
		{ 149, 270, 22 },
		{ 149, 271, 22 },
		{ 149, 272, 22 },
		{ 149, 273, 22 },
		{ 149, 274, 22 },
		{ 149, 275, 22 },
		{ 149, 276, 22 },
		{ 149, 277, 22 },
		{ 149, 278, 22 },
		{ 149, 279, 22 },
		{ 149, 280, 22 },
		{ 149, 281, 22 },
		{ 149, 282, 22 },
		{ 149, 283, 22 },
		{ 150, 231, 22 },
		{ 150, 232, 22 },
		{ 150, 233, 22 },
		{ 150, 234, 22 },
		{ 150, 235, 22 },
		{ 150, 236, 22 },
		{ 150, 237, 22 },
		{ 150, 238, 22 },
		{ 150, 239, 22 },
		{ 150, 240, 22 },
		{ 150, 241, 22 },
		{ 150, 242, 22 },
		{ 150, 243, 22 },
		{ 150, 244, 22 },
		{ 150, 245, 22 },
		{ 150, 246, 22 },
		{ 150, 247, 22 },
		{ 150, 248, 22 },
		{ 150, 249, 22 },
		{ 150, 250, 22 },
		{ 150, 251, 22 },
		{ 150, 252, 22 },
		{ 150, 253, 22 },
		{ 150, 254, 22 },
		{ 150, 255, 22 },
		{ 150, 256, 22 },
		{ 150, 257, 22 },
		{ 150, 258, 22 },
		{ 150, 259, 22 },
		{ 150, 260, 22 },
		{ 150, 261, 22 },
		{ 150, 262, 22 },
		{ 150, 263, 22 },
		{ 150, 264, 22 },
		{ 150, 265, 22 },
		{ 150, 266, 22 },
		{ 150, 267, 22 },
		{ 150, 268, 22 },
		{ 150, 269, 22 },
		{ 150, 270, 22 },
		{ 150, 271, 22 },
		{ 150, 272, 22 },
		{ 150, 273, 22 },
		{ 150, 274, 22 },
		{ 150, 275, 22 },
		{ 150, 276, 22 },
		{ 150, 277, 22 },
		{ 150, 278, 22 },
		{ 150, 279, 22 },
		{ 150, 280, 22 },
		{ 150, 281, 22 },
		{ 150, 282, 22 },
		{ 150, 283, 22 },
		{ 151, 231, 22 },
		{ 151, 232, 22 },
		{ 151, 233, 22 },
		{ 151, 234, 22 },
		{ 151, 235, 22 },
		{ 151, 236, 22 },
		{ 151, 237, 22 },
		{ 151, 238, 22 },
		{ 151, 239, 22 },
		{ 151, 240, 22 },
		{ 151, 241, 22 },
		{ 151, 242, 22 },
		{ 151, 243, 22 },
		{ 151, 244, 22 },
		{ 151, 245, 22 },
		{ 151, 246, 22 },
		{ 151, 247, 22 },
		{ 151, 248, 22 },
		{ 151, 249, 22 },
		{ 151, 250, 22 },
		{ 151, 251, 22 },
		{ 151, 252, 22 },
		{ 151, 253, 22 },
		{ 151, 254, 22 },
		{ 151, 255, 22 },
		{ 151, 256, 22 },
		{ 151, 257, 22 },
		{ 151, 258, 22 },
		{ 151, 259, 22 },
		{ 151, 260, 22 },
		{ 151, 261, 22 },
		{ 151, 262, 22 },
		{ 151, 263, 22 },
		{ 151, 264, 22 },
		{ 151, 265, 22 },
		{ 151, 266, 22 },
		{ 151, 267, 22 },
		{ 151, 268, 22 },
		{ 151, 269, 22 },
		{ 151, 270, 22 },
		{ 151, 271, 22 },
		{ 151, 272, 22 },
		{ 151, 273, 22 },
		{ 151, 274, 22 },
		{ 151, 275, 22 },
		{ 151, 276, 22 },
		{ 151, 277, 22 },
		{ 151, 278, 22 },
		{ 151, 279, 22 },
		{ 151, 280, 22 },
		{ 151, 281, 22 },
		{ 151, 282, 22 },
		{ 151, 283, 22 },
		{ 152, 231, 22 },
		{ 152, 232, 22 },
		{ 152, 233, 22 },
		{ 152, 234, 22 },
		{ 152, 235, 22 },
		{ 152, 236, 22 },
		{ 152, 237, 22 },
		{ 152, 238, 22 },
		{ 152, 239, 22 },
		{ 152, 240, 22 },
		{ 152, 241, 22 },
		{ 152, 242, 22 },
		{ 152, 243, 22 },
		{ 152, 244, 22 },
		{ 152, 245, 22 },
		{ 152, 246, 22 },
		{ 152, 247, 22 },
		{ 152, 248, 22 },
		{ 152, 249, 22 },
		{ 152, 250, 22 },
		{ 152, 251, 22 },
		{ 152, 252, 22 },
		{ 152, 253, 22 },
		{ 152, 254, 22 },
		{ 152, 255, 22 },
		{ 152, 256, 22 },
		{ 152, 257, 22 },
		{ 152, 258, 22 },
		{ 152, 259, 22 },
		{ 152, 260, 22 },
		{ 152, 261, 22 },
		{ 152, 262, 22 },
		{ 152, 263, 22 },
		{ 152, 264, 22 },
		{ 152, 265, 22 },
		{ 152, 266, 22 },
		{ 152, 267, 22 },
		{ 152, 268, 22 },
		{ 152, 269, 22 },
		{ 152, 270, 22 },
		{ 152, 271, 22 },
		{ 152, 272, 22 },
		{ 152, 273, 22 },
		{ 152, 274, 22 },
		{ 152, 275, 22 },
		{ 152, 276, 22 },
		{ 152, 277, 22 },
		{ 152, 278, 22 },
		{ 152, 279, 22 },
		{ 152, 280, 22 },
		{ 152, 281, 22 },
		{ 152, 282, 22 },
		{ 152, 283, 22 },
		{ 153, 232, 22 },
		{ 153, 233, 22 },
		{ 153, 234, 22 },
		{ 153, 235, 22 },
		{ 153, 236, 22 },
		{ 153, 237, 22 },
		{ 153, 238, 22 },
		{ 153, 239, 22 },
		{ 153, 240, 22 },
		{ 153, 241, 22 },
		{ 153, 242, 22 },
		{ 153, 243, 22 },
		{ 153, 244, 22 },
		{ 153, 245, 22 },
		{ 153, 246, 22 },
		{ 153, 247, 22 },
		{ 153, 248, 22 },
		{ 153, 249, 22 },
		{ 153, 250, 22 },
		{ 153, 251, 22 },
		{ 153, 252, 22 },
		{ 153, 253, 22 },
		{ 153, 254, 22 },
		{ 153, 255, 22 },
		{ 153, 256, 22 },
		{ 153, 257, 22 },
		{ 153, 258, 22 },
		{ 153, 259, 22 },
		{ 153, 260, 22 },
		{ 153, 261, 22 },
		{ 153, 262, 22 },
		{ 153, 263, 22 },
		{ 153, 264, 22 },
		{ 153, 265, 22 },
		{ 153, 266, 22 },
		{ 153, 267, 22 },
		{ 153, 268, 22 },
		{ 153, 269, 22 },
		{ 153, 270, 22 },
		{ 153, 271, 22 },
		{ 153, 272, 22 },
		{ 153, 273, 22 },
		{ 153, 274, 22 },
		{ 153, 275, 22 },
		{ 153, 276, 22 },
		{ 153, 277, 22 },
		{ 153, 278, 22 },
		{ 153, 279, 22 },
		{ 153, 280, 22 },
		{ 153, 281, 22 },
		{ 153, 282, 22 },
		{ 154, 232, 22 },
		{ 154, 233, 22 },
		{ 154, 234, 22 },
		{ 154, 235, 22 },
		{ 154, 236, 22 },
		{ 154, 237, 22 },
		{ 154, 238, 22 },
		{ 154, 239, 22 },
		{ 154, 240, 22 },
		{ 154, 241, 22 },
		{ 154, 242, 22 },
		{ 154, 243, 22 },
		{ 154, 244, 22 },
		{ 154, 245, 22 },
		{ 154, 246, 22 },
		{ 154, 247, 22 },
		{ 154, 248, 22 },
		{ 154, 249, 22 },
		{ 154, 250, 22 },
		{ 154, 251, 22 },
		{ 154, 252, 22 },
		{ 154, 253, 22 },
		{ 154, 254, 22 },
		{ 154, 255, 22 },
		{ 154, 256, 22 },
		{ 154, 257, 22 },
		{ 154, 258, 22 },
		{ 154, 259, 22 },
		{ 154, 260, 22 },
		{ 154, 261, 22 },
		{ 154, 262, 22 },
		{ 154, 263, 22 },
		{ 154, 264, 22 },
		{ 154, 265, 22 },
		{ 154, 266, 22 },
		{ 154, 267, 22 },
		{ 154, 268, 22 },
		{ 154, 269, 22 },
		{ 154, 270, 22 },
		{ 154, 271, 22 },
		{ 154, 272, 22 },
		{ 154, 273, 22 },
		{ 154, 274, 22 },
		{ 154, 275, 22 },
		{ 154, 276, 22 },
		{ 154, 277, 22 },
		{ 154, 278, 22 },
		{ 154, 279, 22 },
		{ 154, 280, 22 },
		{ 154, 281, 22 },
		{ 154, 282, 22 },
		{ 155, 232, 22 },
		{ 155, 233, 22 },
		{ 155, 234, 22 },
		{ 155, 235, 22 },
		{ 155, 236, 22 },
		{ 155, 237, 22 },
		{ 155, 238, 22 },
		{ 155, 239, 22 },
		{ 155, 240, 22 },
		{ 155, 241, 22 },
		{ 155, 242, 22 },
		{ 155, 243, 22 },
		{ 155, 244, 22 },
		{ 155, 245, 22 },
		{ 155, 246, 22 },
		{ 155, 247, 22 },
		{ 155, 248, 22 },
		{ 155, 249, 22 },
		{ 155, 250, 22 },
		{ 155, 251, 22 },
		{ 155, 252, 22 },
		{ 155, 253, 22 },
		{ 155, 254, 22 },
		{ 155, 255, 22 },
		{ 155, 256, 22 },
		{ 155, 257, 22 },
		{ 155, 258, 22 },
		{ 155, 259, 22 },
		{ 155, 260, 22 },
		{ 155, 261, 22 },
		{ 155, 262, 22 },
		{ 155, 263, 22 },
		{ 155, 264, 22 },
		{ 155, 265, 22 },
		{ 155, 266, 22 },
		{ 155, 267, 22 },
		{ 155, 268, 22 },
		{ 155, 269, 22 },
		{ 155, 270, 22 },
		{ 155, 271, 22 },
		{ 155, 272, 22 },
		{ 155, 273, 22 },
		{ 155, 274, 22 },
		{ 155, 275, 22 },
		{ 155, 276, 22 },
		{ 155, 277, 22 },
		{ 155, 278, 22 },
		{ 155, 279, 22 },
		{ 155, 280, 22 },
		{ 155, 281, 22 },
		{ 155, 282, 22 },
		{ 156, 232, 22 },
		{ 156, 233, 22 },
		{ 156, 234, 22 },
		{ 156, 235, 22 },
		{ 156, 236, 22 },
		{ 156, 237, 22 },
		{ 156, 238, 22 },
		{ 156, 239, 22 },
		{ 156, 240, 22 },
		{ 156, 241, 22 },
		{ 156, 242, 22 },
		{ 156, 243, 22 },
		{ 156, 244, 22 },
		{ 156, 245, 22 },
		{ 156, 246, 22 },
		{ 156, 247, 22 },
		{ 156, 248, 22 },
		{ 156, 249, 22 },
		{ 156, 250, 22 },
		{ 156, 251, 22 },
		{ 156, 252, 22 },
		{ 156, 253, 22 },
		{ 156, 254, 22 },
		{ 156, 255, 22 },
		{ 156, 256, 22 },
		{ 156, 257, 22 },
		{ 156, 258, 22 },
		{ 156, 259, 22 },
		{ 156, 260, 22 },
		{ 156, 261, 22 },
		{ 156, 262, 22 },
		{ 156, 263, 22 },
		{ 156, 264, 22 },
		{ 156, 265, 22 },
		{ 156, 266, 22 },
		{ 156, 267, 22 },
		{ 156, 268, 22 },
		{ 156, 269, 22 },
		{ 156, 270, 22 },
		{ 156, 271, 22 },
		{ 156, 272, 22 },
		{ 156, 273, 22 },
		{ 156, 274, 22 },
		{ 156, 275, 22 },
		{ 156, 276, 22 },
		{ 156, 277, 22 },
		{ 156, 278, 22 },
		{ 156, 279, 22 },
		{ 156, 280, 22 },
		{ 156, 281, 22 },
		{ 156, 282, 22 },
		{ 157, 233, 22 },
		{ 157, 234, 22 },
		{ 157, 235, 22 },
		{ 157, 236, 22 },
		{ 157, 237, 22 },
		{ 157, 238, 22 },
		{ 157, 239, 22 },
		{ 157, 240, 22 },
		{ 157, 241, 22 },
		{ 157, 242, 22 },
		{ 157, 243, 22 },
		{ 157, 244, 22 },
		{ 157, 245, 22 },
		{ 157, 246, 22 },
		{ 157, 247, 22 },
		{ 157, 248, 22 },
		{ 157, 249, 22 },
		{ 157, 250, 22 },
		{ 157, 251, 22 },
		{ 157, 252, 22 },
		{ 157, 253, 22 },
		{ 157, 254, 22 },
		{ 157, 255, 22 },
		{ 157, 256, 22 },
		{ 157, 257, 22 },
		{ 157, 258, 22 },
		{ 157, 259, 22 },
		{ 157, 260, 22 },
		{ 157, 261, 22 },
		{ 157, 262, 22 },
		{ 157, 263, 22 },
		{ 157, 264, 22 },
		{ 157, 265, 22 },
		{ 157, 266, 22 },
		{ 157, 267, 22 },
		{ 157, 268, 22 },
		{ 157, 269, 22 },
		{ 157, 270, 22 },
		{ 157, 271, 22 },
		{ 157, 272, 22 },
		{ 157, 273, 22 },
		{ 157, 274, 22 },
		{ 157, 275, 22 },
		{ 157, 276, 22 },
		{ 157, 277, 22 },
		{ 157, 278, 22 },
		{ 157, 279, 22 },
		{ 157, 280, 22 },
		{ 157, 281, 22 },
		{ 158, 233, 22 },
		{ 158, 234, 22 },
		{ 158, 235, 22 },
		{ 158, 236, 22 },
		{ 158, 237, 22 },
		{ 158, 238, 22 },
		{ 158, 239, 22 },
		{ 158, 240, 22 },
		{ 158, 241, 22 },
		{ 158, 242, 22 },
		{ 158, 243, 22 },
		{ 158, 244, 22 },
		{ 158, 245, 22 },
		{ 158, 246, 22 },
		{ 158, 247, 22 },
		{ 158, 248, 22 },
		{ 158, 249, 22 },
		{ 158, 250, 22 },
		{ 158, 251, 22 },
		{ 158, 252, 22 },
		{ 158, 253, 22 },
		{ 158, 254, 22 },
		{ 158, 255, 22 },
		{ 158, 256, 22 },
		{ 158, 257, 22 },
		{ 158, 258, 22 },
		{ 158, 259, 22 },
		{ 158, 260, 22 },
		{ 158, 261, 22 },
		{ 158, 262, 22 },
		{ 158, 263, 22 },
		{ 158, 264, 22 },
		{ 158, 265, 22 },
		{ 158, 266, 22 },
		{ 158, 267, 22 },
		{ 158, 268, 22 },
		{ 158, 269, 22 },
		{ 158, 270, 22 },
		{ 158, 271, 22 },
		{ 158, 272, 22 },
		{ 158, 273, 22 },
		{ 158, 274, 22 },
		{ 158, 275, 22 },
		{ 158, 276, 22 },
		{ 158, 277, 22 },
		{ 158, 278, 22 },
		{ 158, 279, 22 },
		{ 158, 280, 22 },
		{ 158, 281, 22 },
		{ 159, 233, 22 },
		{ 159, 234, 22 },
		{ 159, 235, 22 },
		{ 159, 236, 22 },
		{ 159, 237, 22 },
		{ 159, 238, 22 },
		{ 159, 239, 22 },
		{ 159, 240, 22 },
		{ 159, 241, 22 },
		{ 159, 242, 22 },
		{ 159, 243, 22 },
		{ 159, 244, 22 },
		{ 159, 245, 22 },
		{ 159, 246, 22 },
		{ 159, 247, 22 },
		{ 159, 248, 22 },
		{ 159, 249, 22 },
		{ 159, 250, 22 },
		{ 159, 251, 22 },
		{ 159, 252, 22 },
		{ 159, 253, 22 },
		{ 159, 254, 22 },
		{ 159, 255, 22 },
		{ 159, 256, 22 },
		{ 159, 257, 22 },
		{ 159, 258, 22 },
		{ 159, 259, 22 },
		{ 159, 260, 22 },
		{ 159, 261, 22 },
		{ 159, 262, 22 },
		{ 159, 263, 22 },
		{ 159, 264, 22 },
		{ 159, 265, 22 },
		{ 159, 266, 22 },
		{ 159, 267, 22 },
		{ 159, 268, 22 },
		{ 159, 269, 22 },
		{ 159, 270, 22 },
		{ 159, 271, 22 },
		{ 159, 272, 22 },
		{ 159, 273, 22 },
		{ 159, 274, 22 },
		{ 159, 275, 22 },
		{ 159, 276, 22 },
		{ 159, 277, 22 },
		{ 159, 278, 22 },
		{ 159, 279, 22 },
		{ 159, 280, 22 },
		{ 159, 281, 22 },
		{ 160, 233, 22 },
		{ 160, 234, 22 },
		{ 160, 235, 22 },
		{ 160, 236, 22 },
		{ 160, 237, 22 },
		{ 160, 238, 22 },
		{ 160, 239, 22 },
		{ 160, 240, 22 },
		{ 160, 241, 22 },
		{ 160, 242, 22 },
		{ 160, 243, 22 },
		{ 160, 244, 22 },
		{ 160, 245, 22 },
		{ 160, 246, 22 },
		{ 160, 247, 22 },
		{ 160, 248, 22 },
		{ 160, 249, 22 },
		{ 160, 250, 22 },
		{ 160, 251, 22 },
		{ 160, 252, 22 },
		{ 160, 253, 22 },
		{ 160, 254, 22 },
		{ 160, 255, 22 },
		{ 160, 256, 22 },
		{ 160, 257, 22 },
		{ 160, 258, 22 },
		{ 160, 259, 22 },
		{ 160, 260, 22 },
		{ 160, 261, 22 },
		{ 160, 262, 22 },
		{ 160, 263, 22 },
		{ 160, 264, 22 },
		{ 160, 265, 22 },
		{ 160, 266, 22 },
		{ 160, 267, 22 },
		{ 160, 268, 22 },
		{ 160, 269, 22 },
		{ 160, 270, 22 },
		{ 160, 271, 22 },
		{ 160, 272, 22 },
		{ 160, 273, 22 },
		{ 160, 274, 22 },
		{ 160, 275, 22 },
		{ 160, 276, 22 },
		{ 160, 277, 22 },
		{ 160, 278, 22 },
		{ 160, 279, 22 },
		{ 160, 280, 22 },
		{ 160, 281, 22 },
		{ 161, 234, 22 },
		{ 161, 235, 22 },
		{ 161, 236, 22 },
		{ 161, 237, 22 },
		{ 161, 238, 22 },
		{ 161, 239, 22 },
		{ 161, 240, 22 },
		{ 161, 241, 22 },
		{ 161, 242, 22 },
		{ 161, 243, 22 },
		{ 161, 244, 22 },
		{ 161, 245, 22 },
		{ 161, 246, 22 },
		{ 161, 247, 22 },
		{ 161, 248, 22 },
		{ 161, 249, 22 },
		{ 161, 250, 22 },
		{ 161, 251, 22 },
		{ 161, 252, 22 },
		{ 161, 253, 22 },
		{ 161, 254, 22 },
		{ 161, 255, 22 },
		{ 161, 256, 22 },
		{ 161, 257, 22 },
		{ 161, 258, 22 },
		{ 161, 259, 22 },
		{ 161, 260, 22 },
		{ 161, 261, 22 },
		{ 161, 262, 22 },
		{ 161, 263, 22 },
		{ 161, 264, 22 },
		{ 161, 265, 22 },
		{ 161, 266, 22 },
		{ 161, 267, 22 },
		{ 161, 268, 22 },
		{ 161, 269, 22 },
		{ 161, 270, 22 },
		{ 161, 271, 22 },
		{ 161, 272, 22 },
		{ 161, 273, 22 },
		{ 161, 274, 22 },
		{ 161, 275, 22 },
		{ 161, 276, 22 },
		{ 161, 277, 22 },
		{ 161, 278, 22 },
		{ 161, 279, 22 },
		{ 161, 280, 22 },
		{ 162, 234, 22 },
		{ 162, 235, 22 },
		{ 162, 236, 22 },
		{ 162, 237, 22 },
		{ 162, 238, 22 },
		{ 162, 239, 22 },
		{ 162, 240, 22 },
		{ 162, 241, 22 },
		{ 162, 242, 22 },
		{ 162, 243, 22 },
		{ 162, 244, 22 },
		{ 162, 245, 22 },
		{ 162, 246, 22 },
		{ 162, 247, 22 },
		{ 162, 248, 22 },
		{ 162, 249, 22 },
		{ 162, 250, 22 },
		{ 162, 251, 22 },
		{ 162, 252, 22 },
		{ 162, 253, 22 },
		{ 162, 254, 22 },
		{ 162, 255, 22 },
		{ 162, 256, 22 },
		{ 162, 257, 22 },
		{ 162, 258, 22 },
		{ 162, 259, 22 },
		{ 162, 260, 22 },
		{ 162, 261, 22 },
		{ 162, 262, 22 },
		{ 162, 263, 22 },
		{ 162, 264, 22 },
		{ 162, 265, 22 },
		{ 162, 266, 22 },
		{ 162, 267, 22 },
		{ 162, 268, 22 },
		{ 162, 269, 22 },
		{ 162, 270, 22 },
		{ 162, 271, 22 },
		{ 162, 272, 22 },
		{ 162, 273, 22 },
		{ 162, 274, 22 },
		{ 162, 275, 22 },
		{ 162, 276, 22 },
		{ 162, 277, 22 },
		{ 162, 278, 22 },
		{ 162, 279, 22 },
		{ 162, 280, 22 },
		{ 163, 234, 22 },
		{ 163, 235, 22 },
		{ 163, 236, 22 },
		{ 163, 237, 22 },
		{ 163, 238, 22 },
		{ 163, 239, 22 },
		{ 163, 240, 22 },
		{ 163, 241, 22 },
		{ 163, 242, 22 },
		{ 163, 243, 22 },
		{ 163, 244, 22 },
		{ 163, 245, 22 },
		{ 163, 246, 22 },
		{ 163, 247, 22 },
		{ 163, 248, 22 },
		{ 163, 249, 22 },
		{ 163, 250, 22 },
		{ 163, 251, 22 },
		{ 163, 252, 22 },
		{ 163, 253, 22 },
		{ 163, 254, 22 },
		{ 163, 255, 22 },
		{ 163, 256, 22 },
		{ 163, 257, 22 },
		{ 163, 258, 22 },
		{ 163, 259, 22 },
		{ 163, 260, 22 },
		{ 163, 261, 22 },
		{ 163, 262, 22 },
		{ 163, 263, 22 },
		{ 163, 264, 22 },
		{ 163, 265, 22 },
		{ 163, 266, 22 },
		{ 163, 267, 22 },
		{ 163, 268, 22 },
		{ 163, 269, 22 },
		{ 163, 270, 22 },
		{ 163, 271, 22 },
		{ 163, 272, 22 },
		{ 163, 273, 22 },
		{ 163, 274, 22 },
		{ 163, 275, 22 },
		{ 163, 276, 22 },
		{ 163, 277, 22 },
		{ 163, 278, 22 },
		{ 163, 279, 22 },
		{ 163, 280, 22 },
		{ 164, 234, 22 },
		{ 164, 235, 22 },
		{ 164, 236, 22 },
		{ 164, 237, 22 },
		{ 164, 238, 22 },
		{ 164, 239, 22 },
		{ 164, 240, 22 },
		{ 164, 241, 22 },
		{ 164, 242, 22 },
		{ 164, 243, 22 },
		{ 164, 244, 22 },
		{ 164, 245, 22 },
		{ 164, 246, 22 },
		{ 164, 247, 22 },
		{ 164, 248, 22 },
		{ 164, 249, 22 },
		{ 164, 250, 22 },
		{ 164, 251, 22 },
		{ 164, 252, 22 },
		{ 164, 253, 22 },
		{ 164, 254, 22 },
		{ 164, 255, 22 },
		{ 164, 256, 22 },
		{ 164, 257, 22 },
		{ 164, 258, 22 },
		{ 164, 259, 22 },
		{ 164, 260, 22 },
		{ 164, 261, 22 },
		{ 164, 262, 22 },
		{ 164, 263, 22 },
		{ 164, 264, 22 },
		{ 164, 265, 22 },
		{ 164, 266, 22 },
		{ 164, 267, 22 },
		{ 164, 268, 22 },
		{ 164, 269, 22 },
		{ 164, 270, 22 },
		{ 164, 271, 22 },
		{ 164, 272, 22 },
		{ 164, 273, 22 },
		{ 164, 274, 22 },
		{ 164, 275, 22 },
		{ 164, 276, 22 },
		{ 164, 277, 22 },
		{ 164, 278, 22 },
		{ 164, 279, 22 },
		{ 164, 280, 22 },
		{ 165, 235, 22 },
		{ 165, 236, 22 },
		{ 165, 237, 22 },
		{ 165, 238, 22 },
		{ 165, 239, 22 },
		{ 165, 240, 22 },
		{ 165, 241, 22 },
		{ 165, 242, 22 },
		{ 165, 243, 22 },
		{ 165, 244, 22 },
		{ 165, 245, 22 },
		{ 165, 246, 22 },
		{ 165, 247, 22 },
		{ 165, 248, 22 },
		{ 165, 249, 22 },
		{ 165, 250, 22 },
		{ 165, 251, 22 },
		{ 165, 252, 22 },
		{ 165, 253, 22 },
		{ 165, 254, 22 },
		{ 165, 255, 22 },
		{ 165, 256, 22 },
		{ 165, 257, 22 },
		{ 165, 258, 22 },
		{ 165, 259, 22 },
		{ 165, 260, 22 },
		{ 165, 261, 22 },
		{ 165, 262, 22 },
		{ 165, 263, 22 },
		{ 165, 264, 22 },
		{ 165, 265, 22 },
		{ 165, 266, 22 },
		{ 165, 267, 22 },
		{ 165, 268, 22 },
		{ 165, 269, 22 },
		{ 165, 270, 22 },
		{ 165, 271, 22 },
		{ 165, 272, 22 },
		{ 165, 273, 22 },
		{ 165, 274, 22 },
		{ 165, 275, 22 },
		{ 165, 276, 22 },
		{ 165, 277, 22 },
		{ 165, 278, 22 },
		{ 165, 279, 22 },
		{ 166, 235, 22 },
		{ 166, 236, 22 },
		{ 166, 237, 22 },
		{ 166, 238, 22 },
		{ 166, 239, 22 },
		{ 166, 240, 22 },
		{ 166, 241, 22 },
		{ 166, 242, 22 },
		{ 166, 243, 22 },
		{ 166, 244, 22 },
		{ 166, 245, 22 },
		{ 166, 246, 22 },
		{ 166, 247, 22 },
		{ 166, 248, 22 },
		{ 166, 249, 22 },
		{ 166, 250, 22 },
		{ 166, 251, 22 },
		{ 166, 252, 22 },
		{ 166, 253, 22 },
		{ 166, 254, 22 },
		{ 166, 255, 22 },
		{ 166, 256, 22 },
		{ 166, 257, 22 },
		{ 166, 258, 22 },
		{ 166, 259, 22 },
		{ 166, 260, 22 },
		{ 166, 261, 22 },
		{ 166, 262, 22 },
		{ 166, 263, 22 },
		{ 166, 264, 22 },
		{ 166, 265, 22 },
		{ 166, 266, 22 },
		{ 166, 267, 22 },
		{ 166, 268, 22 },
		{ 166, 269, 22 },
		{ 166, 270, 22 },
		{ 166, 271, 22 },
		{ 166, 272, 22 },
		{ 166, 273, 22 },
		{ 166, 274, 22 },
		{ 166, 275, 22 },
		{ 166, 276, 22 },
		{ 166, 277, 22 },
		{ 166, 278, 22 },
		{ 166, 279, 22 },
		{ 167, 235, 22 },
		{ 167, 236, 22 },
		{ 167, 237, 22 },
		{ 167, 238, 22 },
		{ 167, 239, 22 },
		{ 167, 240, 22 },
		{ 167, 241, 22 },
		{ 167, 242, 22 },
		{ 167, 243, 22 },
		{ 167, 244, 22 },
		{ 167, 245, 22 },
		{ 167, 246, 22 },
		{ 167, 247, 22 },
		{ 167, 248, 22 },
		{ 167, 249, 22 },
		{ 167, 250, 22 },
		{ 167, 251, 22 },
		{ 167, 252, 22 },
		{ 167, 253, 22 },
		{ 167, 254, 22 },
		{ 167, 255, 22 },
		{ 167, 256, 22 },
		{ 167, 257, 22 },
		{ 167, 258, 22 },
		{ 167, 259, 22 },
		{ 167, 260, 22 },
		{ 167, 261, 22 },
		{ 167, 262, 22 },
		{ 167, 263, 22 },
		{ 167, 264, 22 },
		{ 167, 265, 22 },
		{ 167, 266, 22 },
		{ 167, 267, 22 },
		{ 167, 268, 22 },
		{ 167, 269, 22 },
		{ 167, 270, 22 },
		{ 167, 271, 22 },
		{ 167, 272, 22 },
		{ 167, 273, 22 },
		{ 167, 274, 22 },
		{ 167, 275, 22 },
		{ 167, 276, 22 },
		{ 167, 277, 22 },
		{ 167, 278, 22 },
		{ 167, 279, 22 },
		{ 168, 235, 22 },
		{ 168, 236, 22 },
		{ 168, 237, 22 },
		{ 168, 238, 22 },
		{ 168, 239, 22 },
		{ 168, 240, 22 },
		{ 168, 241, 22 },
		{ 168, 242, 22 },
		{ 168, 243, 22 },
		{ 168, 244, 22 },
		{ 168, 245, 22 },
		{ 168, 246, 22 },
		{ 168, 247, 22 },
		{ 168, 248, 22 },
		{ 168, 249, 22 },
		{ 168, 250, 22 },
		{ 168, 251, 22 },
		{ 168, 252, 22 },
		{ 168, 253, 22 },
		{ 168, 254, 22 },
		{ 168, 255, 22 },
		{ 168, 256, 22 },
		{ 168, 257, 22 },
		{ 168, 258, 22 },
		{ 168, 259, 22 },
		{ 168, 260, 22 },
		{ 168, 261, 22 },
		{ 168, 262, 22 },
		{ 168, 263, 22 },
		{ 168, 264, 22 },
		{ 168, 265, 22 },
		{ 168, 266, 22 },
		{ 168, 267, 22 },
		{ 168, 268, 22 },
		{ 168, 269, 22 },
		{ 168, 270, 22 },
		{ 168, 271, 22 },
		{ 168, 272, 22 },
		{ 168, 273, 22 },
		{ 168, 274, 22 },
		{ 168, 275, 22 },
		{ 168, 276, 22 },
		{ 168, 277, 22 },
		{ 168, 278, 22 },
		{ 168, 279, 22 },
		{ 169, 236, 22 },
		{ 169, 237, 22 },
		{ 169, 238, 22 },
		{ 169, 239, 22 },
		{ 169, 240, 22 },
		{ 169, 241, 22 },
		{ 169, 242, 22 },
		{ 169, 243, 22 },
		{ 169, 244, 22 },
		{ 169, 245, 22 },
		{ 169, 246, 22 },
		{ 169, 247, 22 },
		{ 169, 248, 22 },
		{ 169, 249, 22 },
		{ 169, 250, 22 },
		{ 169, 251, 22 },
		{ 169, 252, 22 },
		{ 169, 253, 22 },
		{ 169, 254, 22 },
		{ 169, 255, 22 },
		{ 169, 256, 22 },
		{ 169, 257, 22 },
		{ 169, 258, 22 },
		{ 169, 259, 22 },
		{ 169, 260, 22 },
		{ 169, 261, 22 },
		{ 169, 262, 22 },
		{ 169, 263, 22 },
		{ 169, 264, 22 },
		{ 169, 265, 22 },
		{ 169, 266, 22 },
		{ 169, 267, 22 },
		{ 169, 268, 22 },
		{ 169, 269, 22 },
		{ 169, 270, 22 },
		{ 169, 271, 22 },
		{ 169, 272, 22 },
		{ 169, 273, 22 },
		{ 169, 274, 22 },
		{ 169, 275, 22 },
		{ 169, 276, 22 },
		{ 169, 277, 22 },
		{ 169, 278, 22 },
		{ 170, 236, 22 },
		{ 170, 237, 22 },
		{ 170, 238, 22 },
		{ 170, 239, 22 },
		{ 170, 240, 22 },
		{ 170, 241, 22 },
		{ 170, 242, 22 },
		{ 170, 243, 22 },
		{ 170, 244, 22 },
		{ 170, 245, 22 },
		{ 170, 246, 22 },
		{ 170, 247, 22 },
		{ 170, 248, 22 },
		{ 170, 249, 22 },
		{ 170, 250, 22 },
		{ 170, 251, 22 },
		{ 170, 252, 22 },
		{ 170, 253, 22 },
		{ 170, 254, 22 },
		{ 170, 255, 22 },
		{ 170, 256, 22 },
		{ 170, 257, 22 },
		{ 170, 258, 22 },
		{ 170, 259, 22 },
		{ 170, 260, 22 },
		{ 170, 261, 22 },
		{ 170, 262, 22 },
		{ 170, 263, 22 },
		{ 170, 264, 22 },
		{ 170, 265, 22 },
		{ 170, 266, 22 },
		{ 170, 267, 22 },
		{ 170, 268, 22 },
		{ 170, 269, 22 },
		{ 170, 270, 22 },
		{ 170, 271, 22 },
		{ 170, 272, 22 },
		{ 170, 273, 22 },
		{ 170, 274, 22 },
		{ 170, 275, 22 },
		{ 170, 276, 22 },
		{ 170, 277, 22 },
		{ 170, 278, 22 },
		{ 171, 236, 22 },
		{ 171, 237, 22 },
		{ 171, 238, 22 },
		{ 171, 239, 22 },
		{ 171, 240, 22 },
		{ 171, 241, 22 },
		{ 171, 242, 22 },
		{ 171, 243, 22 },
		{ 171, 244, 22 },
		{ 171, 245, 22 },
		{ 171, 246, 22 },
		{ 171, 247, 22 },
		{ 171, 248, 22 },
		{ 171, 249, 22 },
		{ 171, 250, 22 },
		{ 171, 251, 22 },
		{ 171, 252, 22 },
		{ 171, 253, 22 },
		{ 171, 254, 22 },
		{ 171, 255, 22 },
		{ 171, 256, 22 },
		{ 171, 257, 22 },
		{ 171, 258, 22 },
		{ 171, 259, 22 },
		{ 171, 260, 22 },
		{ 171, 261, 22 },
		{ 171, 262, 22 },
		{ 171, 263, 22 },
		{ 171, 264, 22 },
		{ 171, 265, 22 },
		{ 171, 266, 22 },
		{ 171, 267, 22 },
		{ 171, 268, 22 },
		{ 171, 269, 22 },
		{ 171, 270, 22 },
		{ 171, 271, 22 },
		{ 171, 272, 22 },
		{ 171, 273, 22 },
		{ 171, 274, 22 },
		{ 171, 275, 22 },
		{ 171, 276, 22 },
		{ 171, 277, 22 },
		{ 171, 278, 22 },
		{ 172, 236, 22 },
		{ 172, 237, 22 },
		{ 172, 238, 22 },
		{ 172, 239, 22 },
		{ 172, 240, 22 },
		{ 172, 241, 22 },
		{ 172, 242, 22 },
		{ 172, 243, 22 },
		{ 172, 244, 22 },
		{ 172, 245, 22 },
		{ 172, 246, 22 },
		{ 172, 247, 22 },
		{ 172, 248, 22 },
		{ 172, 249, 22 },
		{ 172, 250, 22 },
		{ 172, 251, 22 },
		{ 172, 252, 22 },
		{ 172, 253, 22 },
		{ 172, 254, 22 },
		{ 172, 255, 22 },
		{ 172, 256, 22 },
		{ 172, 257, 22 },
		{ 172, 258, 22 },
		{ 172, 259, 22 },
		{ 172, 260, 22 },
		{ 172, 261, 22 },
		{ 172, 262, 22 },
		{ 172, 263, 22 },
		{ 172, 264, 22 },
		{ 172, 265, 22 },
		{ 172, 266, 22 },
		{ 172, 267, 22 },
		{ 172, 268, 22 },
		{ 172, 269, 22 },
		{ 172, 270, 22 },
		{ 172, 271, 22 },
		{ 172, 272, 22 },
		{ 172, 273, 22 },
		{ 172, 274, 22 },
		{ 172, 275, 22 },
		{ 172, 276, 22 },
		{ 172, 277, 22 },
		{ 172, 278, 22 },
		{ 173, 237, 22 },
		{ 173, 238, 22 },
		{ 173, 239, 22 },
		{ 173, 240, 22 },
		{ 173, 241, 22 },
		{ 173, 242, 22 },
		{ 173, 243, 22 },
		{ 173, 244, 22 },
		{ 173, 245, 22 },
		{ 173, 246, 22 },
		{ 173, 247, 22 },
		{ 173, 248, 22 },
		{ 173, 249, 22 },
		{ 173, 250, 22 },
		{ 173, 251, 22 },
		{ 173, 252, 22 },
		{ 173, 253, 22 },
		{ 173, 254, 22 },
		{ 173, 255, 22 },
		{ 173, 256, 22 },
		{ 173, 257, 22 },
		{ 173, 258, 22 },
		{ 173, 259, 22 },
		{ 173, 260, 22 },
		{ 173, 261, 22 },
		{ 173, 262, 22 },
		{ 173, 263, 22 },
		{ 173, 264, 22 },
		{ 173, 265, 22 },
		{ 173, 266, 22 },
		{ 173, 267, 22 },
		{ 173, 268, 22 },
		{ 173, 269, 22 },
		{ 173, 270, 22 },
		{ 173, 271, 22 },
		{ 173, 272, 22 },
		{ 173, 273, 22 },
		{ 173, 274, 22 },
		{ 173, 275, 22 },
		{ 173, 276, 22 },
		{ 173, 277, 22 },
		{ 174, 237, 22 },
		{ 174, 238, 22 },
		{ 174, 239, 22 },
		{ 174, 240, 22 },
		{ 174, 241, 22 },
		{ 174, 242, 22 },
		{ 174, 243, 22 },
		{ 174, 244, 22 },
		{ 174, 245, 22 },
		{ 174, 246, 22 },
		{ 174, 247, 22 },
		{ 174, 248, 22 },
		{ 174, 249, 22 },
		{ 174, 250, 22 },
		{ 174, 251, 22 },
		{ 174, 252, 22 },
		{ 174, 253, 22 },
		{ 174, 254, 22 },
		{ 174, 255, 22 },
		{ 174, 256, 22 },
		{ 174, 257, 22 },
		{ 174, 258, 22 },
		{ 174, 259, 22 },
		{ 174, 260, 22 },
		{ 174, 261, 22 },
		{ 174, 262, 22 },
		{ 174, 263, 22 },
		{ 174, 264, 22 },
		{ 174, 265, 22 },
		{ 174, 266, 22 },
		{ 174, 267, 22 },
		{ 174, 268, 22 },
		{ 174, 269, 22 },
		{ 174, 270, 22 },
		{ 174, 271, 22 },
		{ 174, 272, 22 },
		{ 174, 273, 22 },
		{ 174, 274, 22 },
		{ 174, 275, 22 },
		{ 174, 276, 22 },
		{ 174, 277, 22 },
		{ 175, 237, 22 },
		{ 175, 238, 22 },
		{ 175, 239, 22 },
		{ 175, 240, 22 },
		{ 175, 241, 22 },
		{ 175, 242, 22 },
		{ 175, 243, 22 },
		{ 175, 244, 22 },
		{ 175, 245, 22 },
		{ 175, 246, 22 },
		{ 175, 247, 22 },
		{ 175, 248, 22 },
		{ 175, 249, 22 },
		{ 175, 250, 22 },
		{ 175, 251, 22 },
		{ 175, 252, 22 },
		{ 175, 253, 22 },
		{ 175, 254, 22 },
		{ 175, 255, 22 },
		{ 175, 256, 22 },
		{ 175, 257, 22 },
		{ 175, 258, 22 },
		{ 175, 259, 22 },
		{ 175, 260, 22 },
		{ 175, 261, 22 },
		{ 175, 262, 22 },
		{ 175, 263, 22 },
		{ 175, 264, 22 },
		{ 175, 265, 22 },
		{ 175, 266, 22 },
		{ 175, 267, 22 },
		{ 175, 268, 22 },
		{ 175, 269, 22 },
		{ 175, 270, 22 },
		{ 175, 271, 22 },
		{ 175, 272, 22 },
		{ 175, 273, 22 },
		{ 175, 274, 22 },
		{ 175, 275, 22 },
		{ 175, 276, 22 },
		{ 175, 277, 22 },
		{ 176, 237, 22 },
		{ 176, 238, 22 },
		{ 176, 239, 22 },
		{ 176, 240, 22 },
		{ 176, 241, 22 },
		{ 176, 242, 22 },
		{ 176, 243, 22 },
		{ 176, 244, 22 },
		{ 176, 245, 22 },
		{ 176, 246, 22 },
		{ 176, 247, 22 },
		{ 176, 248, 22 },
		{ 176, 249, 22 },
		{ 176, 250, 22 },
		{ 176, 251, 22 },
		{ 176, 252, 22 },
		{ 176, 253, 22 },
		{ 176, 254, 22 },
		{ 176, 255, 22 },
		{ 176, 256, 22 },
		{ 176, 257, 22 },
		{ 176, 258, 22 },
		{ 176, 259, 22 },
		{ 176, 260, 22 },
		{ 176, 261, 22 },
		{ 176, 262, 22 },
		{ 176, 263, 22 },
		{ 176, 264, 22 },
		{ 176, 265, 22 },
		{ 176, 266, 22 },
		{ 176, 267, 22 },
		{ 176, 268, 22 },
		{ 176, 269, 22 },
		{ 176, 270, 22 },
		{ 176, 271, 22 },
		{ 176, 272, 22 },
		{ 176, 273, 22 },
		{ 176, 274, 22 },
		{ 176, 275, 22 },
		{ 176, 276, 22 },
		{ 176, 277, 22 },
		{ 177, 238, 22 },
		{ 177, 239, 22 },
		{ 177, 240, 22 },
		{ 177, 241, 22 },
		{ 177, 242, 22 },
		{ 177, 243, 22 },
		{ 177, 244, 22 },
		{ 177, 245, 22 },
		{ 177, 246, 22 },
		{ 177, 247, 22 },
		{ 177, 248, 22 },
		{ 177, 249, 22 },
		{ 177, 250, 22 },
		{ 177, 251, 22 },
		{ 177, 252, 22 },
		{ 177, 253, 22 },
		{ 177, 254, 22 },
		{ 177, 255, 22 },
		{ 177, 256, 22 },
		{ 177, 257, 22 },
		{ 177, 258, 22 },
		{ 177, 259, 22 },
		{ 177, 260, 22 },
		{ 177, 261, 22 },
		{ 177, 262, 22 },
		{ 177, 263, 22 },
		{ 177, 264, 22 },
		{ 177, 265, 22 },
		{ 177, 266, 22 },
		{ 177, 267, 22 },
		{ 177, 268, 22 },
		{ 177, 269, 22 },
		{ 177, 270, 22 },
		{ 177, 271, 22 },
		{ 177, 272, 22 },
		{ 177, 273, 22 },
		{ 177, 274, 22 },
		{ 177, 275, 22 },
		{ 177, 276, 22 },
		{ 178, 238, 22 },
		{ 178, 239, 22 },
		{ 178, 240, 22 },
		{ 178, 241, 22 },
		{ 178, 242, 22 },
		{ 178, 243, 22 },
		{ 178, 244, 22 },
		{ 178, 245, 22 },
		{ 178, 246, 22 },
		{ 178, 247, 22 },
		{ 178, 248, 22 },
		{ 178, 249, 22 },
		{ 178, 250, 22 },
		{ 178, 251, 22 },
		{ 178, 252, 22 },
		{ 178, 253, 22 },
		{ 178, 254, 22 },
		{ 178, 255, 22 },
		{ 178, 256, 22 },
		{ 178, 257, 22 },
		{ 178, 258, 22 },
		{ 178, 259, 22 },
		{ 178, 260, 22 },
		{ 178, 261, 22 },
		{ 178, 262, 22 },
		{ 178, 263, 22 },
		{ 178, 264, 22 },
		{ 178, 265, 22 },
		{ 178, 266, 22 },
		{ 178, 267, 22 },
		{ 178, 268, 22 },
		{ 178, 269, 22 },
		{ 178, 270, 22 },
		{ 178, 271, 22 },
		{ 178, 272, 22 },
		{ 178, 273, 22 },
		{ 178, 274, 22 },
		{ 178, 275, 22 },
		{ 178, 276, 22 },
		{ 179, 238, 22 },
		{ 179, 239, 22 },
		{ 179, 240, 22 },
		{ 179, 241, 22 },
		{ 179, 242, 22 },
		{ 179, 243, 22 },
		{ 179, 244, 22 },
		{ 179, 245, 22 },
		{ 179, 246, 22 },
		{ 179, 247, 22 },
		{ 179, 248, 22 },
		{ 179, 249, 22 },
		{ 179, 250, 22 },
		{ 179, 251, 22 },
		{ 179, 252, 22 },
		{ 179, 253, 22 },
		{ 179, 254, 22 },
		{ 179, 255, 22 },
		{ 179, 256, 22 },
		{ 179, 257, 22 },
		{ 179, 258, 22 },
		{ 179, 259, 22 },
		{ 179, 260, 22 },
		{ 179, 261, 22 },
		{ 179, 262, 22 },
		{ 179, 263, 22 },
		{ 179, 264, 22 },
		{ 179, 265, 22 },
		{ 179, 266, 22 },
		{ 179, 267, 22 },
		{ 179, 268, 22 },
		{ 179, 269, 22 },
		{ 179, 270, 22 },
		{ 179, 271, 22 },
		{ 179, 272, 22 },
		{ 179, 273, 22 },
		{ 179, 274, 22 },
		{ 179, 275, 22 },
		{ 179, 276, 22 },
		{ 180, 238, 22 },
		{ 180, 239, 22 },
		{ 180, 240, 22 },
		{ 180, 241, 22 },
		{ 180, 242, 22 },
		{ 180, 243, 22 },
		{ 180, 244, 22 },
		{ 180, 245, 22 },
		{ 180, 246, 22 },
		{ 180, 247, 22 },
		{ 180, 248, 22 },
		{ 180, 249, 22 },
		{ 180, 250, 22 },
		{ 180, 251, 22 },
		{ 180, 252, 22 },
		{ 180, 253, 22 },
		{ 180, 254, 22 },
		{ 180, 255, 22 },
		{ 180, 256, 22 },
		{ 180, 257, 22 },
		{ 180, 258, 22 },
		{ 180, 259, 22 },
		{ 180, 260, 22 },
		{ 180, 261, 22 },
		{ 180, 262, 22 },
		{ 180, 263, 22 },
		{ 180, 264, 22 },
		{ 180, 265, 22 },
		{ 180, 266, 22 },
		{ 180, 267, 22 },
		{ 180, 268, 22 },
		{ 180, 269, 22 },
		{ 180, 270, 22 },
		{ 180, 271, 22 },
		{ 180, 272, 22 },
		{ 180, 273, 22 },
		{ 180, 274, 22 },
		{ 180, 275, 22 },
		{ 180, 276, 22 },
		{ 181, 239, 22 },
		{ 181, 240, 22 },
		{ 181, 241, 22 },
		{ 181, 242, 22 },
		{ 181, 243, 22 },
		{ 181, 244, 22 },
		{ 181, 245, 22 },
		{ 181, 246, 22 },
		{ 181, 247, 22 },
		{ 181, 248, 22 },
		{ 181, 249, 22 },
		{ 181, 250, 22 },
		{ 181, 251, 22 },
		{ 181, 252, 22 },
		{ 181, 253, 22 },
		{ 181, 254, 22 },
		{ 181, 255, 22 },
		{ 181, 256, 22 },
		{ 181, 257, 22 },
		{ 181, 258, 22 },
		{ 181, 259, 22 },
		{ 181, 260, 22 },
		{ 181, 261, 22 },
		{ 181, 262, 22 },
		{ 181, 263, 22 },
		{ 181, 264, 22 },
		{ 181, 265, 22 },
		{ 181, 266, 22 },
		{ 181, 267, 22 },
		{ 181, 268, 22 },
		{ 181, 269, 22 },
		{ 181, 270, 22 },
		{ 181, 271, 22 },
		{ 181, 272, 22 },
		{ 181, 273, 22 },
		{ 181, 274, 22 },
		{ 181, 275, 22 },
		{ 182, 239, 22 },
		{ 182, 240, 22 },
		{ 182, 241, 22 },
		{ 182, 242, 22 },
		{ 182, 243, 22 },
		{ 182, 244, 22 },
		{ 182, 245, 22 },
		{ 182, 246, 22 },
		{ 182, 247, 22 },
		{ 182, 248, 22 },
		{ 182, 249, 22 },
		{ 182, 250, 22 },
		{ 182, 251, 22 },
		{ 182, 252, 22 },
		{ 182, 253, 22 },
		{ 182, 254, 22 },
		{ 182, 255, 22 },
		{ 182, 256, 22 },
		{ 182, 257, 22 },
		{ 182, 258, 22 },
		{ 182, 259, 22 },
		{ 182, 260, 22 },
		{ 182, 261, 22 },
		{ 182, 262, 22 },
		{ 182, 263, 22 },
		{ 182, 264, 22 },
		{ 182, 265, 22 },
		{ 182, 266, 22 },
		{ 182, 267, 22 },
		{ 182, 268, 22 },
		{ 182, 269, 22 },
		{ 182, 270, 22 },
		{ 182, 271, 22 },
		{ 182, 272, 22 },
		{ 182, 273, 22 },
		{ 182, 274, 22 },
		{ 182, 275, 22 },
		{ 183, 239, 22 },
		{ 183, 240, 22 },
		{ 183, 241, 22 },
		{ 183, 242, 22 },
		{ 183, 243, 22 },
		{ 183, 244, 22 },
		{ 183, 245, 22 },
		{ 183, 246, 22 },
		{ 183, 247, 22 },
		{ 183, 248, 22 },
		{ 183, 249, 22 },
		{ 183, 250, 22 },
		{ 183, 251, 22 },
		{ 183, 252, 22 },
		{ 183, 253, 22 },
		{ 183, 254, 22 },
		{ 183, 255, 22 },
		{ 183, 256, 22 },
		{ 183, 257, 22 },
		{ 183, 258, 22 },
		{ 183, 259, 22 },
		{ 183, 260, 22 },
		{ 183, 261, 22 },
		{ 183, 262, 22 },
		{ 183, 263, 22 },
		{ 183, 264, 22 },
		{ 183, 265, 22 },
		{ 183, 266, 22 },
		{ 183, 267, 22 },
		{ 183, 268, 22 },
		{ 183, 269, 22 },
		{ 183, 270, 22 },
		{ 183, 271, 22 },
		{ 183, 272, 22 },
		{ 183, 273, 22 },
		{ 183, 274, 22 },
		{ 183, 275, 22 },
		{ 184, 239, 22 },
		{ 184, 240, 22 },
		{ 184, 241, 22 },
		{ 184, 242, 22 },
		{ 184, 243, 22 },
		{ 184, 244, 22 },
		{ 184, 245, 22 },
		{ 184, 246, 22 },
		{ 184, 247, 22 },
		{ 184, 248, 22 },
		{ 184, 249, 22 },
		{ 184, 250, 22 },
		{ 184, 251, 22 },
		{ 184, 252, 22 },
		{ 184, 253, 22 },
		{ 184, 254, 22 },
		{ 184, 255, 22 },
		{ 184, 256, 22 },
		{ 184, 257, 22 },
		{ 184, 258, 22 },
		{ 184, 259, 22 },
		{ 184, 260, 22 },
		{ 184, 261, 22 },
		{ 184, 262, 22 },
		{ 184, 263, 22 },
		{ 184, 264, 22 },
		{ 184, 265, 22 },
		{ 184, 266, 22 },
		{ 184, 267, 22 },
		{ 184, 268, 22 },
		{ 184, 269, 22 },
		{ 184, 270, 22 },
		{ 184, 271, 22 },
		{ 184, 272, 22 },
		{ 184, 273, 22 },
		{ 184, 274, 22 },
		{ 184, 275, 22 },
		{ 185, 240, 22 },
		{ 185, 241, 22 },
		{ 185, 242, 22 },
		{ 185, 243, 22 },
		{ 185, 244, 22 },
		{ 185, 245, 22 },
		{ 185, 246, 22 },
		{ 185, 247, 22 },
		{ 185, 248, 22 },
		{ 185, 249, 22 },
		{ 185, 250, 22 },
		{ 185, 251, 22 },
		{ 185, 252, 22 },
		{ 185, 253, 22 },
		{ 185, 254, 22 },
		{ 185, 255, 22 },
		{ 185, 256, 22 },
		{ 185, 257, 22 },
		{ 185, 258, 22 },
		{ 185, 259, 22 },
		{ 185, 260, 22 },
		{ 185, 261, 22 },
		{ 185, 262, 22 },
		{ 185, 263, 22 },
		{ 185, 264, 22 },
		{ 185, 265, 22 },
		{ 185, 266, 22 },
		{ 185, 267, 22 },
		{ 185, 268, 22 },
		{ 185, 269, 22 },
		{ 185, 270, 22 },
		{ 185, 271, 22 },
		{ 185, 272, 22 },
		{ 185, 273, 22 },
		{ 185, 274, 22 },
		{ 186, 240, 22 },
		{ 186, 241, 22 },
		{ 186, 242, 22 },
		{ 186, 243, 22 },
		{ 186, 244, 22 },
		{ 186, 245, 22 },
		{ 186, 246, 22 },
		{ 186, 247, 22 },
		{ 186, 248, 22 },
		{ 186, 249, 22 },
		{ 186, 250, 22 },
		{ 186, 251, 22 },
		{ 186, 252, 22 },
		{ 186, 253, 22 },
		{ 186, 254, 22 },
		{ 186, 255, 22 },
		{ 186, 256, 22 },
		{ 186, 257, 22 },
		{ 186, 258, 22 },
		{ 186, 259, 22 },
		{ 186, 260, 22 },
		{ 186, 261, 22 },
		{ 186, 262, 22 },
		{ 186, 263, 22 },
		{ 186, 264, 22 },
		{ 186, 265, 22 },
		{ 186, 266, 22 },
		{ 186, 267, 22 },
		{ 186, 268, 22 },
		{ 186, 269, 22 },
		{ 186, 270, 22 },
		{ 186, 271, 22 },
		{ 186, 272, 22 },
		{ 186, 273, 22 },
		{ 186, 274, 22 },
		{ 187, 240, 22 },
		{ 187, 241, 22 },
		{ 187, 242, 22 },
		{ 187, 243, 22 },
		{ 187, 244, 22 },
		{ 187, 245, 22 },
		{ 187, 246, 22 },
		{ 187, 247, 22 },
		{ 187, 248, 22 },
		{ 187, 249, 22 },
		{ 187, 250, 22 },
		{ 187, 251, 22 },
		{ 187, 252, 22 },
		{ 187, 253, 22 },
		{ 187, 254, 22 },
		{ 187, 255, 22 },
		{ 187, 256, 22 },
		{ 187, 257, 22 },
		{ 187, 258, 22 },
		{ 187, 259, 22 },
		{ 187, 260, 22 },
		{ 187, 261, 22 },
		{ 187, 262, 22 },
		{ 187, 263, 22 },
		{ 187, 264, 22 },
		{ 187, 265, 22 },
		{ 187, 266, 22 },
		{ 187, 267, 22 },
		{ 187, 268, 22 },
		{ 187, 269, 22 },
		{ 187, 270, 22 },
		{ 187, 271, 22 },
		{ 187, 272, 22 },
		{ 187, 273, 22 },
		{ 187, 274, 22 },
		{ 188, 240, 22 },
		{ 188, 241, 22 },
		{ 188, 242, 22 },
		{ 188, 243, 22 },
		{ 188, 244, 22 },
		{ 188, 245, 22 },
		{ 188, 246, 22 },
		{ 188, 247, 22 },
		{ 188, 248, 22 },
		{ 188, 249, 22 },
		{ 188, 250, 22 },
		{ 188, 251, 22 },
		{ 188, 252, 22 },
		{ 188, 253, 22 },
		{ 188, 254, 22 },
		{ 188, 255, 22 },
		{ 188, 256, 22 },
		{ 188, 257, 22 },
		{ 188, 258, 22 },
		{ 188, 259, 22 },
		{ 188, 260, 22 },
		{ 188, 261, 22 },
		{ 188, 262, 22 },
		{ 188, 263, 22 },
		{ 188, 264, 22 },
		{ 188, 265, 22 },
		{ 188, 266, 22 },
		{ 188, 267, 22 },
		{ 188, 268, 22 },
		{ 188, 269, 22 },
		{ 188, 270, 22 },
		{ 188, 271, 22 },
		{ 188, 272, 22 },
		{ 188, 273, 22 },
		{ 188, 274, 22 },
		{ 189, 241, 22 },
		{ 189, 242, 22 },
		{ 189, 243, 22 },
		{ 189, 244, 22 },
		{ 189, 245, 22 },
		{ 189, 246, 22 },
		{ 189, 247, 22 },
		{ 189, 248, 22 },
		{ 189, 249, 22 },
		{ 189, 250, 22 },
		{ 189, 251, 22 },
		{ 189, 252, 22 },
		{ 189, 253, 22 },
		{ 189, 254, 22 },
		{ 189, 255, 22 },
		{ 189, 256, 22 },
		{ 189, 257, 22 },
		{ 189, 258, 22 },
		{ 189, 259, 22 },
		{ 189, 260, 22 },
		{ 189, 261, 22 },
		{ 189, 262, 22 },
		{ 189, 263, 22 },
		{ 189, 264, 22 },
		{ 189, 265, 22 },
		{ 189, 266, 22 },
		{ 189, 267, 22 },
		{ 189, 268, 22 },
		{ 189, 269, 22 },
		{ 189, 270, 22 },
		{ 189, 271, 22 },
		{ 189, 272, 22 },
		{ 189, 273, 22 },
		{ 190, 241, 22 },
		{ 190, 242, 22 },
		{ 190, 243, 22 },
		{ 190, 244, 22 },
		{ 190, 245, 22 },
		{ 190, 246, 22 },
		{ 190, 247, 22 },
		{ 190, 248, 22 },
		{ 190, 249, 22 },
		{ 190, 250, 22 },
		{ 190, 251, 22 },
		{ 190, 252, 22 },
		{ 190, 253, 22 },
		{ 190, 254, 22 },
		{ 190, 255, 22 },
		{ 190, 256, 22 },
		{ 190, 257, 22 },
		{ 190, 258, 22 },
		{ 190, 259, 22 },
		{ 190, 260, 22 },
		{ 190, 261, 22 },
		{ 190, 262, 22 },
		{ 190, 263, 22 },
		{ 190, 264, 22 },
		{ 190, 265, 22 },
		{ 190, 266, 22 },
		{ 190, 267, 22 },
		{ 190, 268, 22 },
		{ 190, 269, 22 },
		{ 190, 270, 22 },
		{ 190, 271, 22 },
		{ 190, 272, 22 },
		{ 190, 273, 22 },
		{ 191, 241, 22 },
		{ 191, 242, 22 },
		{ 191, 243, 22 },
		{ 191, 244, 22 },
		{ 191, 245, 22 },
		{ 191, 246, 22 },
		{ 191, 247, 22 },
		{ 191, 248, 22 },
		{ 191, 249, 22 },
		{ 191, 250, 22 },
		{ 191, 251, 22 },
		{ 191, 252, 22 },
		{ 191, 253, 22 },
		{ 191, 254, 22 },
		{ 191, 255, 22 },
		{ 191, 256, 22 },
		{ 191, 257, 22 },
		{ 191, 258, 22 },
		{ 191, 259, 22 },
		{ 191, 260, 22 },
		{ 191, 261, 22 },
		{ 191, 262, 22 },
		{ 191, 263, 22 },
		{ 191, 264, 22 },
		{ 191, 265, 22 },
		{ 191, 266, 22 },
		{ 191, 267, 22 },
		{ 191, 268, 22 },
		{ 191, 269, 22 },
		{ 191, 270, 22 },
		{ 191, 271, 22 },
		{ 191, 272, 22 },
		{ 191, 273, 22 },
		{ 192, 241, 22 },
		{ 192, 242, 22 },
		{ 192, 243, 22 },
		{ 192, 244, 22 },
		{ 192, 245, 22 },
		{ 192, 246, 22 },
		{ 192, 247, 22 },
		{ 192, 248, 22 },
		{ 192, 249, 22 },
		{ 192, 250, 22 },
		{ 192, 251, 22 },
		{ 192, 252, 22 },
		{ 192, 253, 22 },
		{ 192, 254, 22 },
		{ 192, 255, 22 },
		{ 192, 256, 22 },
		{ 192, 257, 22 },
		{ 192, 258, 22 },
		{ 192, 259, 22 },
		{ 192, 260, 22 },
		{ 192, 261, 22 },
		{ 192, 262, 22 },
		{ 192, 263, 22 },
		{ 192, 264, 22 },
		{ 192, 265, 22 },
		{ 192, 266, 22 },
		{ 192, 267, 22 },
		{ 192, 268, 22 },
		{ 192, 269, 22 },
		{ 192, 270, 22 },
		{ 192, 271, 22 },
		{ 192, 272, 22 },
		{ 192, 273, 22 },
		{ 193, 242, 22 },
		{ 193, 243, 22 },
		{ 193, 244, 22 },
		{ 193, 245, 22 },
		{ 193, 246, 22 },
		{ 193, 247, 22 },
		{ 193, 248, 22 },
		{ 193, 249, 22 },
		{ 193, 250, 22 },
		{ 193, 251, 22 },
		{ 193, 252, 22 },
		{ 193, 253, 22 },
		{ 193, 254, 22 },
		{ 193, 255, 22 },
		{ 193, 256, 22 },
		{ 193, 257, 22 },
		{ 193, 258, 22 },
		{ 193, 259, 22 },
		{ 193, 260, 22 },
		{ 193, 261, 22 },
		{ 193, 262, 22 },
		{ 193, 263, 22 },
		{ 193, 264, 22 },
		{ 193, 265, 22 },
		{ 193, 266, 22 },
		{ 193, 267, 22 },
		{ 193, 268, 22 },
		{ 193, 269, 22 },
		{ 193, 270, 22 },
		{ 193, 271, 22 },
		{ 193, 272, 22 },
		{ 194, 242, 22 },
		{ 194, 243, 22 },
		{ 194, 244, 22 },
		{ 194, 245, 22 },
		{ 194, 246, 22 },
		{ 194, 247, 22 },
		{ 194, 248, 22 },
		{ 194, 249, 22 },
		{ 194, 250, 22 },
		{ 194, 251, 22 },
		{ 194, 252, 22 },
		{ 194, 253, 22 },
		{ 194, 254, 22 },
		{ 194, 255, 22 },
		{ 194, 256, 22 },
		{ 194, 257, 22 },
		{ 194, 258, 22 },
		{ 194, 259, 22 },
		{ 194, 260, 22 },
		{ 194, 261, 22 },
		{ 194, 262, 22 },
		{ 194, 263, 22 },
		{ 194, 264, 22 },
		{ 194, 265, 22 },
		{ 194, 266, 22 },
		{ 194, 267, 22 },
		{ 194, 268, 22 },
		{ 194, 269, 22 },
		{ 194, 270, 22 },
		{ 194, 271, 22 },
		{ 194, 272, 22 },
		{ 195, 242, 22 },
		{ 195, 243, 22 },
		{ 195, 244, 22 },
		{ 195, 245, 22 },
		{ 195, 246, 22 },
		{ 195, 247, 22 },
		{ 195, 248, 22 },
		{ 195, 249, 22 },
		{ 195, 250, 22 },
		{ 195, 251, 22 },
		{ 195, 252, 22 },
		{ 195, 253, 22 },
		{ 195, 254, 22 },
		{ 195, 255, 22 },
		{ 195, 256, 22 },
		{ 195, 257, 22 },
		{ 195, 258, 22 },
		{ 195, 259, 22 },
		{ 195, 260, 22 },
		{ 195, 261, 22 },
		{ 195, 262, 22 },
		{ 195, 263, 22 },
		{ 195, 264, 22 },
		{ 195, 265, 22 },
		{ 195, 266, 22 },
		{ 195, 267, 22 },
		{ 195, 268, 22 },
		{ 195, 269, 22 },
		{ 195, 270, 22 },
		{ 195, 271, 22 },
		{ 195, 272, 22 },
		{ 196, 242, 22 },
		{ 196, 243, 22 },
		{ 196, 244, 22 },
		{ 196, 245, 22 },
		{ 196, 246, 22 },
		{ 196, 247, 22 },
		{ 196, 248, 22 },
		{ 196, 249, 22 },
		{ 196, 250, 22 },
		{ 196, 251, 22 },
		{ 196, 252, 22 },
		{ 196, 253, 22 },
		{ 196, 254, 22 },
		{ 196, 255, 22 },
		{ 196, 256, 22 },
		{ 196, 257, 22 },
		{ 196, 258, 22 },
		{ 196, 259, 22 },
		{ 196, 260, 22 },
		{ 196, 261, 22 },
		{ 196, 262, 22 },
		{ 196, 263, 22 },
		{ 196, 264, 22 },
		{ 196, 265, 22 },
		{ 196, 266, 22 },
		{ 196, 267, 22 },
		{ 196, 268, 22 },
		{ 196, 269, 22 },
		{ 196, 270, 22 },
		{ 196, 271, 22 },
		{ 196, 272, 22 },
		{ 197, 243, 22 },
		{ 197, 244, 22 },
		{ 197, 245, 22 },
		{ 197, 246, 22 },
		{ 197, 247, 22 },
		{ 197, 248, 22 },
		{ 197, 249, 22 },
		{ 197, 250, 22 },
		{ 197, 251, 22 },
		{ 197, 252, 22 },
		{ 197, 253, 22 },
		{ 197, 254, 22 },
		{ 197, 255, 22 },
		{ 197, 256, 22 },
		{ 197, 257, 22 },
		{ 197, 258, 22 },
		{ 197, 259, 22 },
		{ 197, 260, 22 },
		{ 197, 261, 22 },
		{ 197, 262, 22 },
		{ 197, 263, 22 },
		{ 197, 264, 22 },
		{ 197, 265, 22 },
		{ 197, 266, 22 },
		{ 197, 267, 22 },
		{ 197, 268, 22 },
		{ 197, 269, 22 },
		{ 197, 270, 22 },
		{ 197, 271, 22 },
		{ 198, 243, 22 },
		{ 198, 244, 22 },
		{ 198, 245, 22 },
		{ 198, 246, 22 },
		{ 198, 247, 22 },
		{ 198, 248, 22 },
		{ 198, 249, 22 },
		{ 198, 250, 22 },
		{ 198, 251, 22 },
		{ 198, 252, 22 },
		{ 198, 253, 22 },
		{ 198, 254, 22 },
		{ 198, 255, 22 },
		{ 198, 256, 22 },
		{ 198, 257, 22 },
		{ 198, 258, 22 },
		{ 198, 259, 22 },
		{ 198, 260, 22 },
		{ 198, 261, 22 },
		{ 198, 262, 22 },
		{ 198, 263, 22 },
		{ 198, 264, 22 },
		{ 198, 265, 22 },
		{ 198, 266, 22 },
		{ 198, 267, 22 },
		{ 198, 268, 22 },
		{ 198, 269, 22 },
		{ 198, 270, 22 },
		{ 198, 271, 22 },
		{ 199, 243, 22 },
		{ 199, 244, 22 },
		{ 199, 245, 22 },
		{ 199, 246, 22 },
		{ 199, 247, 22 },
		{ 199, 248, 22 },
		{ 199, 249, 22 },
		{ 199, 250, 22 },
		{ 199, 251, 22 },
		{ 199, 252, 22 },
		{ 199, 253, 22 },
		{ 199, 254, 22 },
		{ 199, 255, 22 },
		{ 199, 256, 22 },
		{ 199, 257, 22 },
		{ 199, 258, 22 },
		{ 199, 259, 22 },
		{ 199, 260, 22 },
		{ 199, 261, 22 },
		{ 199, 262, 22 },
		{ 199, 263, 22 },
		{ 199, 264, 22 },
		{ 199, 265, 22 },
		{ 199, 266, 22 },
		{ 199, 267, 22 },
		{ 199, 268, 22 },
		{ 199, 269, 22 },
		{ 199, 270, 22 },
		{ 199, 271, 22 },
		{ 200, 243, 22 },
		{ 200, 244, 22 },
		{ 200, 245, 22 },
		{ 200, 246, 22 },
		{ 200, 247, 22 },
		{ 200, 248, 22 },
		{ 200, 249, 22 },
		{ 200, 250, 22 },
		{ 200, 251, 22 },
		{ 200, 252, 22 },
		{ 200, 253, 22 },
		{ 200, 254, 22 },
		{ 200, 255, 22 },
		{ 200, 256, 22 },
		{ 200, 257, 22 },
		{ 200, 258, 22 },
		{ 200, 259, 22 },
		{ 200, 260, 22 },
		{ 200, 261, 22 },
		{ 200, 262, 22 },
		{ 200, 263, 22 },
		{ 200, 264, 22 },
		{ 200, 265, 22 },
		{ 200, 266, 22 },
		{ 200, 267, 22 },
		{ 200, 268, 22 },
		{ 200, 269, 22 },
		{ 200, 270, 22 },
		{ 200, 271, 22 },
		{ 201, 244, 22 },
		{ 201, 245, 22 },
		{ 201, 246, 22 },
		{ 201, 247, 22 },
		{ 201, 248, 22 },
		{ 201, 249, 22 },
		{ 201, 250, 22 },
		{ 201, 251, 22 },
		{ 201, 252, 22 },
		{ 201, 253, 22 },
		{ 201, 254, 22 },
		{ 201, 255, 22 },
		{ 201, 256, 22 },
		{ 201, 257, 22 },
		{ 201, 258, 22 },
		{ 201, 259, 22 },
		{ 201, 260, 22 },
		{ 201, 261, 22 },
		{ 201, 262, 22 },
		{ 201, 263, 22 },
		{ 201, 264, 22 },
		{ 201, 265, 22 },
		{ 201, 266, 22 },
		{ 201, 267, 22 },
		{ 201, 268, 22 },
		{ 201, 269, 22 },
		{ 201, 270, 22 },
		{ 202, 244, 22 },
		{ 202, 245, 22 },
		{ 202, 246, 22 },
		{ 202, 247, 22 },
		{ 202, 248, 22 },
		{ 202, 249, 22 },
		{ 202, 250, 22 },
		{ 202, 251, 22 },
		{ 202, 252, 22 },
		{ 202, 253, 22 },
		{ 202, 254, 22 },
		{ 202, 255, 22 },
		{ 202, 256, 22 },
		{ 202, 257, 22 },
		{ 202, 258, 22 },
		{ 202, 259, 22 },
		{ 202, 260, 22 },
		{ 202, 261, 22 },
		{ 202, 262, 22 },
		{ 202, 263, 22 },
		{ 202, 264, 22 },
		{ 202, 265, 22 },
		{ 202, 266, 22 },
		{ 202, 267, 22 },
		{ 202, 268, 22 },
		{ 202, 269, 22 },
		{ 202, 270, 22 },
		{ 203, 244, 22 },
		{ 203, 245, 22 },
		{ 203, 246, 22 },
		{ 203, 247, 22 },
		{ 203, 248, 22 },
		{ 203, 249, 22 },
		{ 203, 250, 22 },
		{ 203, 251, 22 },
		{ 203, 252, 22 },
		{ 203, 253, 22 },
		{ 203, 254, 22 },
		{ 203, 255, 22 },
		{ 203, 256, 22 },
		{ 203, 257, 22 },
		{ 203, 258, 22 },
		{ 203, 259, 22 },
		{ 203, 260, 22 },
		{ 203, 261, 22 },
		{ 203, 262, 22 },
		{ 203, 263, 22 },
		{ 203, 264, 22 },
		{ 203, 265, 22 },
		{ 203, 266, 22 },
		{ 203, 267, 22 },
		{ 203, 268, 22 },
		{ 203, 269, 22 },
		{ 203, 270, 22 },
		{ 204, 244, 22 },
		{ 204, 245, 22 },
		{ 204, 246, 22 },
		{ 204, 247, 22 },
		{ 204, 248, 22 },
		{ 204, 249, 22 },
		{ 204, 250, 22 },
		{ 204, 251, 22 },
		{ 204, 252, 22 },
		{ 204, 253, 22 },
		{ 204, 254, 22 },
		{ 204, 255, 22 },
		{ 204, 256, 22 },
		{ 204, 257, 22 },
		{ 204, 258, 22 },
		{ 204, 259, 22 },
		{ 204, 260, 22 },
		{ 204, 261, 22 },
		{ 204, 262, 22 },
		{ 204, 263, 22 },
		{ 204, 264, 22 },
		{ 204, 265, 22 },
		{ 204, 266, 22 },
		{ 204, 267, 22 },
		{ 204, 268, 22 },
		{ 204, 269, 22 },
		{ 204, 270, 22 },
		{ 205, 245, 22 },
		{ 205, 246, 22 },
		{ 205, 247, 22 },
		{ 205, 248, 22 },
		{ 205, 249, 22 },
		{ 205, 250, 22 },
		{ 205, 251, 22 },
		{ 205, 252, 22 },
		{ 205, 253, 22 },
		{ 205, 254, 22 },
		{ 205, 255, 22 },
		{ 205, 256, 22 },
		{ 205, 257, 22 },
		{ 205, 258, 22 },
		{ 205, 259, 22 },
		{ 205, 260, 22 },
		{ 205, 261, 22 },
		{ 205, 262, 22 },
		{ 205, 263, 22 },
		{ 205, 264, 22 },
		{ 205, 265, 22 },
		{ 205, 266, 22 },
		{ 205, 267, 22 },
		{ 205, 268, 22 },
		{ 205, 269, 22 },
		{ 206, 245, 22 },
		{ 206, 246, 22 },
		{ 206, 247, 22 },
		{ 206, 248, 22 },
		{ 206, 249, 22 },
		{ 206, 250, 22 },
		{ 206, 251, 22 },
		{ 206, 252, 22 },
		{ 206, 253, 22 },
		{ 206, 254, 22 },
		{ 206, 255, 22 },
		{ 206, 256, 22 },
		{ 206, 257, 22 },
		{ 206, 258, 22 },
		{ 206, 259, 22 },
		{ 206, 260, 22 },
		{ 206, 261, 22 },
		{ 206, 262, 22 },
		{ 206, 263, 22 },
		{ 206, 264, 22 },
		{ 206, 265, 22 },
		{ 206, 266, 22 },
		{ 206, 267, 22 },
		{ 206, 268, 22 },
		{ 206, 269, 22 },
		{ 207, 245, 22 },
		{ 207, 246, 22 },
		{ 207, 247, 22 },
		{ 207, 248, 22 },
		{ 207, 249, 22 },
		{ 207, 250, 22 },
		{ 207, 251, 22 },
		{ 207, 252, 22 },
		{ 207, 253, 22 },
		{ 207, 254, 22 },
		{ 207, 255, 22 },
		{ 207, 256, 22 },
		{ 207, 257, 22 },
		{ 207, 258, 22 },
		{ 207, 259, 22 },
		{ 207, 260, 22 },
		{ 207, 261, 22 },
		{ 207, 262, 22 },
		{ 207, 263, 22 },
		{ 207, 264, 22 },
		{ 207, 265, 22 },
		{ 207, 266, 22 },
		{ 207, 267, 22 },
		{ 207, 268, 22 },
		{ 207, 269, 22 },
		{ 208, 245, 22 },
		{ 208, 246, 22 },
		{ 208, 247, 22 },
		{ 208, 248, 22 },
		{ 208, 249, 22 },
		{ 208, 250, 22 },
		{ 208, 251, 22 },
		{ 208, 252, 22 },
		{ 208, 253, 22 },
		{ 208, 254, 22 },
		{ 208, 255, 22 },
		{ 208, 256, 22 },
		{ 208, 257, 22 },
		{ 208, 258, 22 },
		{ 208, 259, 22 },
		{ 208, 260, 22 },
		{ 208, 261, 22 },
		{ 208, 262, 22 },
		{ 208, 263, 22 },
		{ 208, 264, 22 },
		{ 208, 265, 22 },
		{ 208, 266, 22 },
		{ 208, 267, 22 },
		{ 208, 268, 22 },
		{ 208, 269, 22 },
		{ 209, 246, 22 },
		{ 209, 247, 22 },
		{ 209, 248, 22 },
		{ 209, 249, 22 },
		{ 209, 250, 22 },
		{ 209, 251, 22 },
		{ 209, 252, 22 },
		{ 209, 253, 22 },
		{ 209, 254, 22 },
		{ 209, 255, 22 },
		{ 209, 256, 22 },
		{ 209, 257, 22 },
		{ 209, 258, 22 },
		{ 209, 259, 22 },
		{ 209, 260, 22 },
		{ 209, 261, 22 },
		{ 209, 262, 22 },
		{ 209, 263, 22 },
		{ 209, 264, 22 },
		{ 209, 265, 22 },
		{ 209, 266, 22 },
		{ 209, 267, 22 },
		{ 209, 268, 22 },
		{ 210, 246, 22 },
		{ 210, 247, 22 },
		{ 210, 248, 22 },
		{ 210, 249, 22 },
		{ 210, 250, 22 },
		{ 210, 251, 22 },
		{ 210, 252, 22 },
		{ 210, 253, 22 },
		{ 210, 254, 22 },
		{ 210, 255, 22 },
		{ 210, 256, 22 },
		{ 210, 257, 22 },
		{ 210, 258, 22 },
		{ 210, 259, 22 },
		{ 210, 260, 22 },
		{ 210, 261, 22 },
		{ 210, 262, 22 },
		{ 210, 263, 22 },
		{ 210, 264, 22 },
		{ 210, 265, 22 },
		{ 210, 266, 22 },
		{ 210, 267, 22 },
		{ 210, 268, 22 },
		{ 211, 246, 22 },
		{ 211, 247, 22 },
		{ 211, 248, 22 },
		{ 211, 249, 22 },
		{ 211, 250, 22 },
		{ 211, 251, 22 },
		{ 211, 252, 22 },
		{ 211, 253, 22 },
		{ 211, 254, 22 },
		{ 211, 255, 22 },
		{ 211, 256, 22 },
		{ 211, 257, 22 },
		{ 211, 258, 22 },
		{ 211, 259, 22 },
		{ 211, 260, 22 },
		{ 211, 261, 22 },
		{ 211, 262, 22 },
		{ 211, 263, 22 },
		{ 211, 264, 22 },
		{ 211, 265, 22 },
		{ 211, 266, 22 },
		{ 211, 267, 22 },
		{ 211, 268, 22 },
		{ 212, 246, 22 },
		{ 212, 247, 22 },
		{ 212, 248, 22 },
		{ 212, 249, 22 },
		{ 212, 250, 22 },
		{ 212, 251, 22 },
		{ 212, 252, 22 },
		{ 212, 253, 22 },
		{ 212, 254, 22 },
		{ 212, 255, 22 },
		{ 212, 256, 22 },
		{ 212, 257, 22 },
		{ 212, 258, 22 },
		{ 212, 259, 22 },
		{ 212, 260, 22 },
		{ 212, 261, 22 },
		{ 212, 262, 22 },
		{ 212, 263, 22 },
		{ 212, 264, 22 },
		{ 212, 265, 22 },
		{ 212, 266, 22 },
		{ 212, 267, 22 },
		{ 212, 268, 22 },
		{ 213, 247, 22 },
		{ 213, 248, 22 },
		{ 213, 249, 22 },
		{ 213, 250, 22 },
		{ 213, 251, 22 },
		{ 213, 252, 22 },
		{ 213, 253, 22 },
		{ 213, 254, 22 },
		{ 213, 255, 22 },
		{ 213, 256, 22 },
		{ 213, 257, 22 },
		{ 213, 258, 22 },
		{ 213, 259, 22 },
		{ 213, 260, 22 },
		{ 213, 261, 22 },
		{ 213, 262, 22 },
		{ 213, 263, 22 },
		{ 213, 264, 22 },
		{ 213, 265, 22 },
		{ 213, 266, 22 },
		{ 213, 267, 22 },
		{ 214, 247, 22 },
		{ 214, 248, 22 },
		{ 214, 249, 22 },
		{ 214, 250, 22 },
		{ 214, 251, 22 },
		{ 214, 252, 22 },
		{ 214, 253, 22 },
		{ 214, 254, 22 },
		{ 214, 255, 22 },
		{ 214, 256, 22 },
		{ 214, 257, 22 },
		{ 214, 258, 22 },
		{ 214, 259, 22 },
		{ 214, 260, 22 },
		{ 214, 261, 22 },
		{ 214, 262, 22 },
		{ 214, 263, 22 },
		{ 214, 264, 22 },
		{ 214, 265, 22 },
		{ 214, 266, 22 },
		{ 214, 267, 22 },
		{ 215, 247, 22 },
		{ 215, 248, 22 },
		{ 215, 249, 22 },
		{ 215, 250, 22 },
		{ 215, 251, 22 },
		{ 215, 252, 22 },
		{ 215, 253, 22 },
		{ 215, 254, 22 },
		{ 215, 255, 22 },
		{ 215, 256, 22 },
		{ 215, 257, 22 },
		{ 215, 258, 22 },
		{ 215, 259, 22 },
		{ 215, 260, 22 },
		{ 215, 261, 22 },
		{ 215, 262, 22 },
		{ 215, 263, 22 },
		{ 215, 264, 22 },
		{ 215, 265, 22 },
		{ 215, 266, 22 },
		{ 215, 267, 22 },
		{ 216, 247, 22 },
		{ 216, 248, 22 },
		{ 216, 249, 22 },
		{ 216, 250, 22 },
		{ 216, 251, 22 },
		{ 216, 252, 22 },
		{ 216, 253, 22 },
		{ 216, 254, 22 },
		{ 216, 255, 22 },
		{ 216, 256, 22 },
		{ 216, 257, 22 },
		{ 216, 258, 22 },
		{ 216, 259, 22 },
		{ 216, 260, 22 },
		{ 216, 261, 22 },
		{ 216, 262, 22 },
		{ 216, 263, 22 },
		{ 216, 264, 22 },
		{ 216, 265, 22 },
		{ 216, 266, 22 },
		{ 216, 267, 22 },
		{ 217, 248, 22 },
		{ 217, 249, 22 },
		{ 217, 250, 22 },
		{ 217, 251, 22 },
		{ 217, 252, 22 },
		{ 217, 253, 22 },
		{ 217, 254, 22 },
		{ 217, 255, 22 },
		{ 217, 256, 22 },
		{ 217, 257, 22 },
		{ 217, 258, 22 },
		{ 217, 259, 22 },
		{ 217, 260, 22 },
		{ 217, 261, 22 },
		{ 217, 262, 22 },
		{ 217, 263, 22 },
		{ 217, 264, 22 },
		{ 217, 265, 22 },
		{ 217, 266, 22 },
		{ 218, 248, 22 },
		{ 218, 249, 22 },
		{ 218, 250, 22 },
		{ 218, 251, 22 },
		{ 218, 252, 22 },
		{ 218, 253, 22 },
		{ 218, 254, 22 },
		{ 218, 255, 22 },
		{ 218, 256, 22 },
		{ 218, 257, 22 },
		{ 218, 258, 22 },
		{ 218, 259, 22 },
		{ 218, 260, 22 },
		{ 218, 261, 22 },
		{ 218, 262, 22 },
		{ 218, 263, 22 },
		{ 218, 264, 22 },
		{ 218, 265, 22 },
		{ 218, 266, 22 },
		{ 219, 248, 22 },
		{ 219, 249, 22 },
		{ 219, 250, 22 },
		{ 219, 251, 22 },
		{ 219, 252, 22 },
		{ 219, 253, 22 },
		{ 219, 254, 22 },
		{ 219, 255, 22 },
		{ 219, 256, 22 },
		{ 219, 257, 22 },
		{ 219, 258, 22 },
		{ 219, 259, 22 },
		{ 219, 260, 22 },
		{ 219, 261, 22 },
		{ 219, 262, 22 },
		{ 219, 263, 22 },
		{ 219, 264, 22 },
		{ 219, 265, 22 },
		{ 219, 266, 22 },
		{ 220, 248, 22 },
		{ 220, 249, 22 },
		{ 220, 250, 22 },
		{ 220, 251, 22 },
		{ 220, 252, 22 },
		{ 220, 253, 22 },
		{ 220, 254, 22 },
		{ 220, 255, 22 },
		{ 220, 256, 22 },
		{ 220, 257, 22 },
		{ 220, 258, 22 },
		{ 220, 259, 22 },
		{ 220, 260, 22 },
		{ 220, 261, 22 },
		{ 220, 262, 22 },
		{ 220, 263, 22 },
		{ 220, 264, 22 },
		{ 220, 265, 22 },
		{ 220, 266, 22 },
		{ 221, 249, 22 },
		{ 221, 250, 22 },
		{ 221, 251, 22 },
		{ 221, 252, 22 },
		{ 221, 253, 22 },
		{ 221, 254, 22 },
		{ 221, 255, 22 },
		{ 221, 256, 22 },
		{ 221, 257, 22 },
		{ 221, 258, 22 },
		{ 221, 259, 22 },
		{ 221, 260, 22 },
		{ 221, 261, 22 },
		{ 221, 262, 22 },
		{ 221, 263, 22 },
		{ 221, 264, 22 },
		{ 221, 265, 22 },
		{ 222, 249, 22 },
		{ 222, 250, 22 },
		{ 222, 251, 22 },
		{ 222, 252, 22 },
		{ 222, 253, 22 },
		{ 222, 254, 22 },
		{ 222, 255, 22 },
		{ 222, 256, 22 },
		{ 222, 257, 22 },
		{ 222, 258, 22 },
		{ 222, 259, 22 },
		{ 222, 260, 22 },
		{ 222, 261, 22 },
		{ 222, 262, 22 },
		{ 222, 263, 22 },
		{ 222, 264, 22 },
		{ 222, 265, 22 },
		{ 223, 249, 22 },
		{ 223, 250, 22 },
		{ 223, 251, 22 },
		{ 223, 252, 22 },
		{ 223, 253, 22 },
		{ 223, 254, 22 },
		{ 223, 255, 22 },
		{ 223, 256, 22 },
		{ 223, 257, 22 },
		{ 223, 258, 22 },
		{ 223, 259, 22 },
		{ 223, 260, 22 },
		{ 223, 261, 22 },
		{ 223, 262, 22 },
		{ 223, 263, 22 },
		{ 223, 264, 22 },
		{ 223, 265, 22 },
		{ 224, 249, 22 },
		{ 224, 250, 22 },
		{ 224, 251, 22 },
		{ 224, 252, 22 },
		{ 224, 253, 22 },
		{ 224, 254, 22 },
		{ 224, 255, 22 },
		{ 224, 256, 22 },
		{ 224, 257, 22 },
		{ 224, 258, 22 },
		{ 224, 259, 22 },
		{ 224, 260, 22 },
		{ 224, 261, 22 },
		{ 224, 262, 22 },
		{ 224, 263, 22 },
		{ 224, 264, 22 },
		{ 224, 265, 22 },
		{ 225, 250, 22 },
		{ 225, 251, 22 },
		{ 225, 252, 22 },
		{ 225, 253, 22 },
		{ 225, 254, 22 },
		{ 225, 255, 22 },
		{ 225, 256, 22 },
		{ 225, 257, 22 },
		{ 225, 258, 22 },
		{ 225, 259, 22 },
		{ 225, 260, 22 },
		{ 225, 261, 22 },
		{ 225, 262, 22 },
		{ 225, 263, 22 },
		{ 225, 264, 22 },
		{ 226, 250, 22 },
		{ 226, 251, 22 },
		{ 226, 252, 22 },
		{ 226, 253, 22 },
		{ 226, 254, 22 },
		{ 226, 255, 22 },
		{ 226, 256, 22 },
		{ 226, 257, 22 },
		{ 226, 258, 22 },
		{ 226, 259, 22 },
		{ 226, 260, 22 },
		{ 226, 261, 22 },
		{ 226, 262, 22 },
		{ 226, 263, 22 },
		{ 226, 264, 22 },
		{ 227, 250, 22 },
		{ 227, 251, 22 },
		{ 227, 252, 22 },
		{ 227, 253, 22 },
		{ 227, 254, 22 },
		{ 227, 255, 22 },
		{ 227, 256, 22 },
		{ 227, 257, 22 },
		{ 227, 258, 22 },
		{ 227, 259, 22 },
		{ 227, 260, 22 },
		{ 227, 261, 22 },
		{ 227, 262, 22 },
		{ 227, 263, 22 },
		{ 227, 264, 22 },
		{ 228, 250, 22 },
		{ 228, 251, 22 },
		{ 228, 252, 22 },
		{ 228, 253, 22 },
		{ 228, 254, 22 },
		{ 228, 255, 22 },
		{ 228, 256, 22 },
		{ 228, 257, 22 },
		{ 228, 258, 22 },
		{ 228, 259, 22 },
		{ 228, 260, 22 },
		{ 228, 261, 22 },
		{ 228, 262, 22 },
		{ 228, 263, 22 },
		{ 228, 264, 22 },
		{ 229, 251, 22 },
		{ 229, 252, 22 },
		{ 229, 253, 22 },
		{ 229, 254, 22 },
		{ 229, 255, 22 },
		{ 229, 256, 22 },
		{ 229, 257, 22 },
		{ 229, 258, 22 },
		{ 229, 259, 22 },
		{ 229, 260, 22 },
		{ 229, 261, 22 },
		{ 229, 262, 22 },
		{ 229, 263, 22 },
		{ 230, 251, 22 },
		{ 230, 252, 22 },
		{ 230, 253, 22 },
		{ 230, 254, 22 },
		{ 230, 255, 22 },
		{ 230, 256, 22 },
		{ 230, 257, 22 },
		{ 230, 258, 22 },
		{ 230, 259, 22 },
		{ 230, 260, 22 },
		{ 230, 261, 22 },
		{ 230, 262, 22 },
		{ 230, 263, 22 },
		{ 231, 251, 22 },
		{ 231, 252, 22 },
		{ 231, 253, 22 },
		{ 231, 254, 22 },
		{ 231, 255, 22 },
		{ 231, 256, 22 },
		{ 231, 257, 22 },
		{ 231, 258, 22 },
		{ 231, 259, 22 },
		{ 231, 260, 22 },
		{ 231, 261, 22 },
		{ 231, 262, 22 },
		{ 231, 263, 22 },
		{ 232, 251, 22 },
		{ 232, 252, 22 },
		{ 232, 253, 22 },
		{ 232, 254, 22 },
		{ 232, 255, 22 },
		{ 232, 256, 22 },
		{ 232, 257, 22 },
		{ 232, 258, 22 },
		{ 232, 259, 22 },
		{ 232, 260, 22 },
		{ 232, 261, 22 },
		{ 232, 262, 22 },
		{ 232, 263, 22 },
		{ 233, 252, 22 },
		{ 233, 253, 22 },
		{ 233, 254, 22 },
		{ 233, 255, 22 },
		{ 233, 256, 22 },
		{ 233, 257, 22 },
		{ 233, 258, 22 },
		{ 233, 259, 22 },
		{ 233, 260, 22 },
		{ 233, 261, 22 },
		{ 233, 262, 22 },
		{ 234, 252, 22 },
		{ 234, 253, 22 },
		{ 234, 254, 22 },
		{ 234, 255, 22 },
		{ 234, 256, 22 },
		{ 234, 257, 22 },
		{ 234, 258, 22 },
		{ 234, 259, 22 },
		{ 234, 260, 22 },
		{ 234, 261, 22 },
		{ 234, 262, 22 },
		{ 235, 252, 22 },
		{ 235, 253, 22 },
		{ 235, 254, 22 },
		{ 235, 255, 22 },
		{ 235, 256, 22 },
		{ 235, 257, 22 },
		{ 235, 258, 22 },
		{ 235, 259, 22 },
		{ 235, 260, 22 },
		{ 235, 261, 22 },
		{ 235, 262, 22 },
		{ 236, 252, 22 },
		{ 236, 253, 22 },
		{ 236, 254, 22 },
		{ 236, 255, 22 },
		{ 236, 256, 22 },
		{ 236, 257, 22 },
		{ 236, 258, 22 },
		{ 236, 259, 22 },
		{ 236, 260, 22 },
		{ 236, 261, 22 },
		{ 236, 262, 22 },
		{ 237, 253, 22 },
		{ 237, 254, 22 },
		{ 237, 255, 22 },
		{ 237, 256, 22 },
		{ 237, 257, 22 },
		{ 237, 258, 22 },
		{ 237, 259, 22 },
		{ 237, 260, 22 },
		{ 237, 261, 22 },
		{ 238, 253, 22 },
		{ 238, 254, 22 },
		{ 238, 255, 22 },
		{ 238, 256, 22 },
		{ 238, 257, 22 },
		{ 238, 258, 22 },
		{ 238, 259, 22 },
		{ 238, 260, 22 },
		{ 238, 261, 22 },
		{ 239, 253, 22 },
		{ 239, 254, 22 },
		{ 239, 255, 22 },
		{ 239, 256, 22 },
		{ 239, 257, 22 },
		{ 239, 258, 22 },
		{ 239, 259, 22 },
		{ 239, 260, 22 },
		{ 239, 261, 22 },
		{ 240, 253, 22 },
		{ 240, 254, 22 },
		{ 240, 255, 22 },
		{ 240, 256, 22 },
		{ 240, 257, 22 },
		{ 240, 258, 22 },
		{ 240, 259, 22 },
		{ 240, 260, 22 },
		{ 240, 261, 22 },
		{ 274, 253, 22 },
		{ 274, 254, 22 },
		{ 274, 255, 22 },
		{ 274, 256, 22 },
		{ 274, 257, 22 },
		{ 274, 258, 22 },
		{ 274, 259, 22 },
		{ 274, 260, 22 },
		{ 274, 261, 22 },
		{ 275, 253, 22 },
		{ 275, 254, 22 },
		{ 275, 255, 22 },
		{ 275, 256, 22 },
		{ 275, 257, 22 },
		{ 275, 258, 22 },
		{ 275, 259, 22 },
		{ 275, 260, 22 },
		{ 275, 261, 22 },
		{ 276, 253, 22 },
		{ 276, 254, 22 },
		{ 276, 255, 22 },
		{ 276, 256, 22 },
		{ 276, 257, 22 },
		{ 276, 258, 22 },
		{ 276, 259, 22 },
		{ 276, 260, 22 },
		{ 276, 261, 22 },
		{ 277, 253, 22 },
		{ 277, 254, 22 },
		{ 277, 255, 22 },
		{ 277, 256, 22 },
		{ 277, 257, 22 },
		{ 277, 258, 22 },
		{ 277, 259, 22 },
		{ 277, 260, 22 },
		{ 277, 261, 22 },
		{ 278, 252, 22 },
		{ 278, 253, 22 },
		{ 278, 254, 22 },
		{ 278, 255, 22 },
		{ 278, 256, 22 },
		{ 278, 257, 22 },
		{ 278, 258, 22 },
		{ 278, 259, 22 },
		{ 278, 260, 22 },
		{ 278, 261, 22 },
		{ 278, 262, 22 },
		{ 279, 252, 22 },
		{ 279, 253, 22 },
		{ 279, 254, 22 },
		{ 279, 255, 22 },
		{ 279, 256, 22 },
		{ 279, 257, 22 },
		{ 279, 258, 22 },
		{ 279, 259, 22 },
		{ 279, 260, 22 },
		{ 279, 261, 22 },
		{ 279, 262, 22 },
		{ 280, 252, 22 },
		{ 280, 253, 22 },
		{ 280, 254, 22 },
		{ 280, 255, 22 },
		{ 280, 256, 22 },
		{ 280, 257, 22 },
		{ 280, 258, 22 },
		{ 280, 259, 22 },
		{ 280, 260, 22 },
		{ 280, 261, 22 },
		{ 280, 262, 22 },
		{ 281, 252, 22 },
		{ 281, 253, 22 },
		{ 281, 254, 22 },
		{ 281, 255, 22 },
		{ 281, 256, 22 },
		{ 281, 257, 22 },
		{ 281, 258, 22 },
		{ 281, 259, 22 },
		{ 281, 260, 22 },
		{ 281, 261, 22 },
		{ 281, 262, 22 },
		{ 282, 251, 22 },
		{ 282, 252, 22 },
		{ 282, 253, 22 },
		{ 282, 254, 22 },
		{ 282, 255, 22 },
		{ 282, 256, 22 },
		{ 282, 257, 22 },
		{ 282, 258, 22 },
		{ 282, 259, 22 },
		{ 282, 260, 22 },
		{ 282, 261, 22 },
		{ 282, 262, 22 },
		{ 282, 263, 22 },
		{ 283, 251, 22 },
		{ 283, 252, 22 },
		{ 283, 253, 22 },
		{ 283, 254, 22 },
		{ 283, 255, 22 },
		{ 283, 256, 22 },
		{ 283, 257, 22 },
		{ 283, 258, 22 },
		{ 283, 259, 22 },
		{ 283, 260, 22 },
		{ 283, 261, 22 },
		{ 283, 262, 22 },
		{ 283, 263, 22 },
		{ 284, 251, 22 },
		{ 284, 252, 22 },
		{ 284, 253, 22 },
		{ 284, 254, 22 },
		{ 284, 255, 22 },
		{ 284, 256, 22 },
		{ 284, 257, 22 },
		{ 284, 258, 22 },
		{ 284, 259, 22 },
		{ 284, 260, 22 },
		{ 284, 261, 22 },
		{ 284, 262, 22 },
		{ 284, 263, 22 },
		{ 285, 251, 22 },
		{ 285, 252, 22 },
		{ 285, 253, 22 },
		{ 285, 254, 22 },
		{ 285, 255, 22 },
		{ 285, 256, 22 },
		{ 285, 257, 22 },
		{ 285, 258, 22 },
		{ 285, 259, 22 },
		{ 285, 260, 22 },
		{ 285, 261, 22 },
		{ 285, 262, 22 },
		{ 285, 263, 22 },
		{ 286, 250, 22 },
		{ 286, 251, 22 },
		{ 286, 252, 22 },
		{ 286, 253, 22 },
		{ 286, 254, 22 },
		{ 286, 255, 22 },
		{ 286, 256, 22 },
		{ 286, 257, 22 },
		{ 286, 258, 22 },
		{ 286, 259, 22 },
		{ 286, 260, 22 },
		{ 286, 261, 22 },
		{ 286, 262, 22 },
		{ 286, 263, 22 },
		{ 286, 264, 22 },
		{ 287, 250, 22 },
		{ 287, 251, 22 },
		{ 287, 252, 22 },
		{ 287, 253, 22 },
		{ 287, 254, 22 },
		{ 287, 255, 22 },
		{ 287, 256, 22 },
		{ 287, 257, 22 },
		{ 287, 258, 22 },
		{ 287, 259, 22 },
		{ 287, 260, 22 },
		{ 287, 261, 22 },
		{ 287, 262, 22 },
		{ 287, 263, 22 },
		{ 287, 264, 22 },
		{ 288, 250, 22 },
		{ 288, 251, 22 },
		{ 288, 252, 22 },
		{ 288, 253, 22 },
		{ 288, 254, 22 },
		{ 288, 255, 22 },
		{ 288, 256, 22 },
		{ 288, 257, 22 },
		{ 288, 258, 22 },
		{ 288, 259, 22 },
		{ 288, 260, 22 },
		{ 288, 261, 22 },
		{ 288, 262, 22 },
		{ 288, 263, 22 },
		{ 288, 264, 22 },
		{ 289, 250, 22 },
		{ 289, 251, 22 },
		{ 289, 252, 22 },
		{ 289, 253, 22 },
		{ 289, 254, 22 },
		{ 289, 255, 22 },
		{ 289, 256, 22 },
		{ 289, 257, 22 },
		{ 289, 258, 22 },
		{ 289, 259, 22 },
		{ 289, 260, 22 },
		{ 289, 261, 22 },
		{ 289, 262, 22 },
		{ 289, 263, 22 },
		{ 289, 264, 22 },
		{ 290, 249, 22 },
		{ 290, 250, 22 },
		{ 290, 251, 22 },
		{ 290, 252, 22 },
		{ 290, 253, 22 },
		{ 290, 254, 22 },
		{ 290, 255, 22 },
		{ 290, 256, 22 },
		{ 290, 257, 22 },
		{ 290, 258, 22 },
		{ 290, 259, 22 },
		{ 290, 260, 22 },
		{ 290, 261, 22 },
		{ 290, 262, 22 },
		{ 290, 263, 22 },
		{ 290, 264, 22 },
		{ 290, 265, 22 },
		{ 291, 249, 22 },
		{ 291, 250, 22 },
		{ 291, 251, 22 },
		{ 291, 252, 22 },
		{ 291, 253, 22 },
		{ 291, 254, 22 },
		{ 291, 255, 22 },
		{ 291, 256, 22 },
		{ 291, 257, 22 },
		{ 291, 258, 22 },
		{ 291, 259, 22 },
		{ 291, 260, 22 },
		{ 291, 261, 22 },
		{ 291, 262, 22 },
		{ 291, 263, 22 },
		{ 291, 264, 22 },
		{ 291, 265, 22 },
		{ 292, 249, 22 },
		{ 292, 250, 22 },
		{ 292, 251, 22 },
		{ 292, 252, 22 },
		{ 292, 253, 22 },
		{ 292, 254, 22 },
		{ 292, 255, 22 },
		{ 292, 256, 22 },
		{ 292, 257, 22 },
		{ 292, 258, 22 },
		{ 292, 259, 22 },
		{ 292, 260, 22 },
		{ 292, 261, 22 },
		{ 292, 262, 22 },
		{ 292, 263, 22 },
		{ 292, 264, 22 },
		{ 292, 265, 22 },
		{ 293, 249, 22 },
		{ 293, 250, 22 },
		{ 293, 251, 22 },
		{ 293, 252, 22 },
		{ 293, 253, 22 },
		{ 293, 254, 22 },
		{ 293, 255, 22 },
		{ 293, 256, 22 },
		{ 293, 257, 22 },
		{ 293, 258, 22 },
		{ 293, 259, 22 },
		{ 293, 260, 22 },
		{ 293, 261, 22 },
		{ 293, 262, 22 },
		{ 293, 263, 22 },
		{ 293, 264, 22 },
		{ 293, 265, 22 },
		{ 294, 248, 22 },
		{ 294, 249, 22 },
		{ 294, 250, 22 },
		{ 294, 251, 22 },
		{ 294, 252, 22 },
		{ 294, 253, 22 },
		{ 294, 254, 22 },
		{ 294, 255, 22 },
		{ 294, 256, 22 },
		{ 294, 257, 22 },
		{ 294, 258, 22 },
		{ 294, 259, 22 },
		{ 294, 260, 22 },
		{ 294, 261, 22 },
		{ 294, 262, 22 },
		{ 294, 263, 22 },
		{ 294, 264, 22 },
		{ 294, 265, 22 },
		{ 294, 266, 22 },
		{ 295, 248, 22 },
		{ 295, 249, 22 },
		{ 295, 250, 22 },
		{ 295, 251, 22 },
		{ 295, 252, 22 },
		{ 295, 253, 22 },
		{ 295, 254, 22 },
		{ 295, 255, 22 },
		{ 295, 256, 22 },
		{ 295, 257, 22 },
		{ 295, 258, 22 },
		{ 295, 259, 22 },
		{ 295, 260, 22 },
		{ 295, 261, 22 },
		{ 295, 262, 22 },
		{ 295, 263, 22 },
		{ 295, 264, 22 },
		{ 295, 265, 22 },
		{ 295, 266, 22 },
		{ 296, 248, 22 },
		{ 296, 249, 22 },
		{ 296, 250, 22 },
		{ 296, 251, 22 },
		{ 296, 252, 22 },
		{ 296, 253, 22 },
		{ 296, 254, 22 },
		{ 296, 255, 22 },
		{ 296, 256, 22 },
		{ 296, 257, 22 },
		{ 296, 258, 22 },
		{ 296, 259, 22 },
		{ 296, 260, 22 },
		{ 296, 261, 22 },
		{ 296, 262, 22 },
		{ 296, 263, 22 },
		{ 296, 264, 22 },
		{ 296, 265, 22 },
		{ 296, 266, 22 },
		{ 297, 248, 22 },
		{ 297, 249, 22 },
		{ 297, 250, 22 },
		{ 297, 251, 22 },
		{ 297, 252, 22 },
		{ 297, 253, 22 },
		{ 297, 254, 22 },
		{ 297, 255, 22 },
		{ 297, 256, 22 },
		{ 297, 257, 22 },
		{ 297, 258, 22 },
		{ 297, 259, 22 },
		{ 297, 260, 22 },
		{ 297, 261, 22 },
		{ 297, 262, 22 },
		{ 297, 263, 22 },
		{ 297, 264, 22 },
		{ 297, 265, 22 },
		{ 297, 266, 22 },
		{ 298, 247, 22 },
		{ 298, 248, 22 },
		{ 298, 249, 22 },
		{ 298, 250, 22 },
		{ 298, 251, 22 },
		{ 298, 252, 22 },
		{ 298, 253, 22 },
		{ 298, 254, 22 },
		{ 298, 255, 22 },
		{ 298, 256, 22 },
		{ 298, 257, 22 },
		{ 298, 258, 22 },
		{ 298, 259, 22 },
		{ 298, 260, 22 },
		{ 298, 261, 22 },
		{ 298, 262, 22 },
		{ 298, 263, 22 },
		{ 298, 264, 22 },
		{ 298, 265, 22 },
		{ 298, 266, 22 },
		{ 298, 267, 22 },
		{ 299, 247, 22 },
		{ 299, 248, 22 },
		{ 299, 249, 22 },
		{ 299, 250, 22 },
		{ 299, 251, 22 },
		{ 299, 252, 22 },
		{ 299, 253, 22 },
		{ 299, 254, 22 },
		{ 299, 255, 22 },
		{ 299, 256, 22 },
		{ 299, 257, 22 },
		{ 299, 258, 22 },
		{ 299, 259, 22 },
		{ 299, 260, 22 },
		{ 299, 261, 22 },
		{ 299, 262, 22 },
		{ 299, 263, 22 },
		{ 299, 264, 22 },
		{ 299, 265, 22 },
		{ 299, 266, 22 },
		{ 299, 267, 22 },
		{ 300, 247, 22 },
		{ 300, 248, 22 },
		{ 300, 249, 22 },
		{ 300, 250, 22 },
		{ 300, 251, 22 },
		{ 300, 252, 22 },
		{ 300, 253, 22 },
		{ 300, 254, 22 },
		{ 300, 255, 22 },
		{ 300, 256, 22 },
		{ 300, 257, 22 },
		{ 300, 258, 22 },
		{ 300, 259, 22 },
		{ 300, 260, 22 },
		{ 300, 261, 22 },
		{ 300, 262, 22 },
		{ 300, 263, 22 },
		{ 300, 264, 22 },
		{ 300, 265, 22 },
		{ 300, 266, 22 },
		{ 300, 267, 22 },
		{ 301, 247, 22 },
		{ 301, 248, 22 },
		{ 301, 249, 22 },
		{ 301, 250, 22 },
		{ 301, 251, 22 },
		{ 301, 252, 22 },
		{ 301, 253, 22 },
		{ 301, 254, 22 },
		{ 301, 255, 22 },
		{ 301, 256, 22 },
		{ 301, 257, 22 },
		{ 301, 258, 22 },
		{ 301, 259, 22 },
		{ 301, 260, 22 },
		{ 301, 261, 22 },
		{ 301, 262, 22 },
		{ 301, 263, 22 },
		{ 301, 264, 22 },
		{ 301, 265, 22 },
		{ 301, 266, 22 },
		{ 301, 267, 22 },
		{ 302, 246, 22 },
		{ 302, 247, 22 },
		{ 302, 248, 22 },
		{ 302, 249, 22 },
		{ 302, 250, 22 },
		{ 302, 251, 22 },
		{ 302, 252, 22 },
		{ 302, 253, 22 },
		{ 302, 254, 22 },
		{ 302, 255, 22 },
		{ 302, 256, 22 },
		{ 302, 257, 22 },
		{ 302, 258, 22 },
		{ 302, 259, 22 },
		{ 302, 260, 22 },
		{ 302, 261, 22 },
		{ 302, 262, 22 },
		{ 302, 263, 22 },
		{ 302, 264, 22 },
		{ 302, 265, 22 },
		{ 302, 266, 22 },
		{ 302, 267, 22 },
		{ 302, 268, 22 },
		{ 303, 246, 22 },
		{ 303, 247, 22 },
		{ 303, 248, 22 },
		{ 303, 249, 22 },
		{ 303, 250, 22 },
		{ 303, 251, 22 },
		{ 303, 252, 22 },
		{ 303, 253, 22 },
		{ 303, 254, 22 },
		{ 303, 255, 22 },
		{ 303, 256, 22 },
		{ 303, 257, 22 },
		{ 303, 258, 22 },
		{ 303, 259, 22 },
		{ 303, 260, 22 },
		{ 303, 261, 22 },
		{ 303, 262, 22 },
		{ 303, 263, 22 },
		{ 303, 264, 22 },
		{ 303, 265, 22 },
		{ 303, 266, 22 },
		{ 303, 267, 22 },
		{ 303, 268, 22 },
		{ 304, 246, 22 },
		{ 304, 247, 22 },
		{ 304, 248, 22 },
		{ 304, 249, 22 },
		{ 304, 250, 22 },
		{ 304, 251, 22 },
		{ 304, 252, 22 },
		{ 304, 253, 22 },
		{ 304, 254, 22 },
		{ 304, 255, 22 },
		{ 304, 256, 22 },
		{ 304, 257, 22 },
		{ 304, 258, 22 },
		{ 304, 259, 22 },
		{ 304, 260, 22 },
		{ 304, 261, 22 },
		{ 304, 262, 22 },
		{ 304, 263, 22 },
		{ 304, 264, 22 },
		{ 304, 265, 22 },
		{ 304, 266, 22 },
		{ 304, 267, 22 },
		{ 304, 268, 22 },
		{ 305, 246, 22 },
		{ 305, 247, 22 },
		{ 305, 248, 22 },
		{ 305, 249, 22 },
		{ 305, 250, 22 },
		{ 305, 251, 22 },
		{ 305, 252, 22 },
		{ 305, 253, 22 },
		{ 305, 254, 22 },
		{ 305, 255, 22 },
		{ 305, 256, 22 },
		{ 305, 257, 22 },
		{ 305, 258, 22 },
		{ 305, 259, 22 },
		{ 305, 260, 22 },
		{ 305, 261, 22 },
		{ 305, 262, 22 },
		{ 305, 263, 22 },
		{ 305, 264, 22 },
		{ 305, 265, 22 },
		{ 305, 266, 22 },
		{ 305, 267, 22 },
		{ 305, 268, 22 },
		{ 306, 245, 22 },
		{ 306, 246, 22 },
		{ 306, 247, 22 },
		{ 306, 248, 22 },
		{ 306, 249, 22 },
		{ 306, 250, 22 },
		{ 306, 251, 22 },
		{ 306, 252, 22 },
		{ 306, 253, 22 },
		{ 306, 254, 22 },
		{ 306, 255, 22 },
		{ 306, 256, 22 },
		{ 306, 257, 22 },
		{ 306, 258, 22 },
		{ 306, 259, 22 },
		{ 306, 260, 22 },
		{ 306, 261, 22 },
		{ 306, 262, 22 },
		{ 306, 263, 22 },
		{ 306, 264, 22 },
		{ 306, 265, 22 },
		{ 306, 266, 22 },
		{ 306, 267, 22 },
		{ 306, 268, 22 },
		{ 306, 269, 22 },
		{ 307, 245, 22 },
		{ 307, 246, 22 },
		{ 307, 247, 22 },
		{ 307, 248, 22 },
		{ 307, 249, 22 },
		{ 307, 250, 22 },
		{ 307, 251, 22 },
		{ 307, 252, 22 },
		{ 307, 253, 22 },
		{ 307, 254, 22 },
		{ 307, 255, 22 },
		{ 307, 256, 22 },
		{ 307, 257, 22 },
		{ 307, 258, 22 },
		{ 307, 259, 22 },
		{ 307, 260, 22 },
		{ 307, 261, 22 },
		{ 307, 262, 22 },
		{ 307, 263, 22 },
		{ 307, 264, 22 },
		{ 307, 265, 22 },
		{ 307, 266, 22 },
		{ 307, 267, 22 },
		{ 307, 268, 22 },
		{ 307, 269, 22 },
		{ 308, 245, 22 },
		{ 308, 246, 22 },
		{ 308, 247, 22 },
		{ 308, 248, 22 },
		{ 308, 249, 22 },
		{ 308, 250, 22 },
		{ 308, 251, 22 },
		{ 308, 252, 22 },
		{ 308, 253, 22 },
		{ 308, 254, 22 },
		{ 308, 255, 22 },
		{ 308, 256, 22 },
		{ 308, 257, 22 },
		{ 308, 258, 22 },
		{ 308, 259, 22 },
		{ 308, 260, 22 },
		{ 308, 261, 22 },
		{ 308, 262, 22 },
		{ 308, 263, 22 },
		{ 308, 264, 22 },
		{ 308, 265, 22 },
		{ 308, 266, 22 },
		{ 308, 267, 22 },
		{ 308, 268, 22 },
		{ 308, 269, 22 },
		{ 309, 245, 22 },
		{ 309, 246, 22 },
		{ 309, 247, 22 },
		{ 309, 248, 22 },
		{ 309, 249, 22 },
		{ 309, 250, 22 },
		{ 309, 251, 22 },
		{ 309, 252, 22 },
		{ 309, 253, 22 },
		{ 309, 254, 22 },
		{ 309, 255, 22 },
		{ 309, 256, 22 },
		{ 309, 257, 22 },
		{ 309, 258, 22 },
		{ 309, 259, 22 },
		{ 309, 260, 22 },
		{ 309, 261, 22 },
		{ 309, 262, 22 },
		{ 309, 263, 22 },
		{ 309, 264, 22 },
		{ 309, 265, 22 },
		{ 309, 266, 22 },
		{ 309, 267, 22 },
		{ 309, 268, 22 },
		{ 309, 269, 22 },
		{ 310, 244, 22 },
		{ 310, 245, 22 },
		{ 310, 246, 22 },
		{ 310, 247, 22 },
		{ 310, 248, 22 },
		{ 310, 249, 22 },
		{ 310, 250, 22 },
		{ 310, 251, 22 },
		{ 310, 252, 22 },
		{ 310, 253, 22 },
		{ 310, 254, 22 },
		{ 310, 255, 22 },
		{ 310, 256, 22 },
		{ 310, 257, 22 },
		{ 310, 258, 22 },
		{ 310, 259, 22 },
		{ 310, 260, 22 },
		{ 310, 261, 22 },
		{ 310, 262, 22 },
		{ 310, 263, 22 },
		{ 310, 264, 22 },
		{ 310, 265, 22 },
		{ 310, 266, 22 },
		{ 310, 267, 22 },
		{ 310, 268, 22 },
		{ 310, 269, 22 },
		{ 310, 270, 22 },
		{ 311, 244, 22 },
		{ 311, 245, 22 },
		{ 311, 246, 22 },
		{ 311, 247, 22 },
		{ 311, 248, 22 },
		{ 311, 249, 22 },
		{ 311, 250, 22 },
		{ 311, 251, 22 },
		{ 311, 252, 22 },
		{ 311, 253, 22 },
		{ 311, 254, 22 },
		{ 311, 255, 22 },
		{ 311, 256, 22 },
		{ 311, 257, 22 },
		{ 311, 258, 22 },
		{ 311, 259, 22 },
		{ 311, 260, 22 },
		{ 311, 261, 22 },
		{ 311, 262, 22 },
		{ 311, 263, 22 },
		{ 311, 264, 22 },
		{ 311, 265, 22 },
		{ 311, 266, 22 },
		{ 311, 267, 22 },
		{ 311, 268, 22 },
		{ 311, 269, 22 },
		{ 311, 270, 22 },
		{ 312, 244, 22 },
		{ 312, 245, 22 },
		{ 312, 246, 22 },
		{ 312, 247, 22 },
		{ 312, 248, 22 },
		{ 312, 249, 22 },
		{ 312, 250, 22 },
		{ 312, 251, 22 },
		{ 312, 252, 22 },
		{ 312, 253, 22 },
		{ 312, 254, 22 },
		{ 312, 255, 22 },
		{ 312, 256, 22 },
		{ 312, 257, 22 },
		{ 312, 258, 22 },
		{ 312, 259, 22 },
		{ 312, 260, 22 },
		{ 312, 261, 22 },
		{ 312, 262, 22 },
		{ 312, 263, 22 },
		{ 312, 264, 22 },
		{ 312, 265, 22 },
		{ 312, 266, 22 },
		{ 312, 267, 22 },
		{ 312, 268, 22 },
		{ 312, 269, 22 },
		{ 312, 270, 22 },
		{ 313, 244, 22 },
		{ 313, 245, 22 },
		{ 313, 246, 22 },
		{ 313, 247, 22 },
		{ 313, 248, 22 },
		{ 313, 249, 22 },
		{ 313, 250, 22 },
		{ 313, 251, 22 },
		{ 313, 252, 22 },
		{ 313, 253, 22 },
		{ 313, 254, 22 },
		{ 313, 255, 22 },
		{ 313, 256, 22 },
		{ 313, 257, 22 },
		{ 313, 258, 22 },
		{ 313, 259, 22 },
		{ 313, 260, 22 },
		{ 313, 261, 22 },
		{ 313, 262, 22 },
		{ 313, 263, 22 },
		{ 313, 264, 22 },
		{ 313, 265, 22 },
		{ 313, 266, 22 },
		{ 313, 267, 22 },
		{ 313, 268, 22 },
		{ 313, 269, 22 },
		{ 313, 270, 22 },
		{ 314, 243, 22 },
		{ 314, 244, 22 },
		{ 314, 245, 22 },
		{ 314, 246, 22 },
		{ 314, 247, 22 },
		{ 314, 248, 22 },
		{ 314, 249, 22 },
		{ 314, 250, 22 },
		{ 314, 251, 22 },
		{ 314, 252, 22 },
		{ 314, 253, 22 },
		{ 314, 254, 22 },
		{ 314, 255, 22 },
		{ 314, 256, 22 },
		{ 314, 257, 22 },
		{ 314, 258, 22 },
		{ 314, 259, 22 },
		{ 314, 260, 22 },
		{ 314, 261, 22 },
		{ 314, 262, 22 },
		{ 314, 263, 22 },
		{ 314, 264, 22 },
		{ 314, 265, 22 },
		{ 314, 266, 22 },
		{ 314, 267, 22 },
		{ 314, 268, 22 },
		{ 314, 269, 22 },
		{ 314, 270, 22 },
		{ 314, 271, 22 },
		{ 315, 243, 22 },
		{ 315, 244, 22 },
		{ 315, 245, 22 },
		{ 315, 246, 22 },
		{ 315, 247, 22 },
		{ 315, 248, 22 },
		{ 315, 249, 22 },
		{ 315, 250, 22 },
		{ 315, 251, 22 },
		{ 315, 252, 22 },
		{ 315, 253, 22 },
		{ 315, 254, 22 },
		{ 315, 255, 22 },
		{ 315, 256, 22 },
		{ 315, 257, 22 },
		{ 315, 258, 22 },
		{ 315, 259, 22 },
		{ 315, 260, 22 },
		{ 315, 261, 22 },
		{ 315, 262, 22 },
		{ 315, 263, 22 },
		{ 315, 264, 22 },
		{ 315, 265, 22 },
		{ 315, 266, 22 },
		{ 315, 267, 22 },
		{ 315, 268, 22 },
		{ 315, 269, 22 },
		{ 315, 270, 22 },
		{ 315, 271, 22 },
		{ 316, 243, 22 },
		{ 316, 244, 22 },
		{ 316, 245, 22 },
		{ 316, 246, 22 },
		{ 316, 247, 22 },
		{ 316, 248, 22 },
		{ 316, 249, 22 },
		{ 316, 250, 22 },
		{ 316, 251, 22 },
		{ 316, 252, 22 },
		{ 316, 253, 22 },
		{ 316, 254, 22 },
		{ 316, 255, 22 },
		{ 316, 256, 22 },
		{ 316, 257, 22 },
		{ 316, 258, 22 },
		{ 316, 259, 22 },
		{ 316, 260, 22 },
		{ 316, 261, 22 },
		{ 316, 262, 22 },
		{ 316, 263, 22 },
		{ 316, 264, 22 },
		{ 316, 265, 22 },
		{ 316, 266, 22 },
		{ 316, 267, 22 },
		{ 316, 268, 22 },
		{ 316, 269, 22 },
		{ 316, 270, 22 },
		{ 316, 271, 22 },
		{ 317, 243, 22 },
		{ 317, 244, 22 },
		{ 317, 245, 22 },
		{ 317, 246, 22 },
		{ 317, 247, 22 },
		{ 317, 248, 22 },
		{ 317, 249, 22 },
		{ 317, 250, 22 },
		{ 317, 251, 22 },
		{ 317, 252, 22 },
		{ 317, 253, 22 },
		{ 317, 254, 22 },
		{ 317, 255, 22 },
		{ 317, 256, 22 },
		{ 317, 257, 22 },
		{ 317, 258, 22 },
		{ 317, 259, 22 },
		{ 317, 260, 22 },
		{ 317, 261, 22 },
		{ 317, 262, 22 },
		{ 317, 263, 22 },
		{ 317, 264, 22 },
		{ 317, 265, 22 },
		{ 317, 266, 22 },
		{ 317, 267, 22 },
		{ 317, 268, 22 },
		{ 317, 269, 22 },
		{ 317, 270, 22 },
		{ 317, 271, 22 },
		{ 318, 242, 22 },
		{ 318, 243, 22 },
		{ 318, 244, 22 },
		{ 318, 245, 22 },
		{ 318, 246, 22 },
		{ 318, 247, 22 },
		{ 318, 248, 22 },
		{ 318, 249, 22 },
		{ 318, 250, 22 },
		{ 318, 251, 22 },
		{ 318, 252, 22 },
		{ 318, 253, 22 },
		{ 318, 254, 22 },
		{ 318, 255, 22 },
		{ 318, 256, 22 },
		{ 318, 257, 22 },
		{ 318, 258, 22 },
		{ 318, 259, 22 },
		{ 318, 260, 22 },
		{ 318, 261, 22 },
		{ 318, 262, 22 },
		{ 318, 263, 22 },
		{ 318, 264, 22 },
		{ 318, 265, 22 },
		{ 318, 266, 22 },
		{ 318, 267, 22 },
		{ 318, 268, 22 },
		{ 318, 269, 22 },
		{ 318, 270, 22 },
		{ 318, 271, 22 },
		{ 318, 272, 22 },
		{ 319, 242, 22 },
		{ 319, 243, 22 },
		{ 319, 244, 22 },
		{ 319, 245, 22 },
		{ 319, 246, 22 },
		{ 319, 247, 22 },
		{ 319, 248, 22 },
		{ 319, 249, 22 },
		{ 319, 250, 22 },
		{ 319, 251, 22 },
		{ 319, 252, 22 },
		{ 319, 253, 22 },
		{ 319, 254, 22 },
		{ 319, 255, 22 },
		{ 319, 256, 22 },
		{ 319, 257, 22 },
		{ 319, 258, 22 },
		{ 319, 259, 22 },
		{ 319, 260, 22 },
		{ 319, 261, 22 },
		{ 319, 262, 22 },
		{ 319, 263, 22 },
		{ 319, 264, 22 },
		{ 319, 265, 22 },
		{ 319, 266, 22 },
		{ 319, 267, 22 },
		{ 319, 268, 22 },
		{ 319, 269, 22 },
		{ 319, 270, 22 },
		{ 319, 271, 22 },
		{ 319, 272, 22 },
		{ 320, 242, 22 },
		{ 320, 243, 22 },
		{ 320, 244, 22 },
		{ 320, 245, 22 },
		{ 320, 246, 22 },
		{ 320, 247, 22 },
		{ 320, 248, 22 },
		{ 320, 249, 22 },
		{ 320, 250, 22 },
		{ 320, 251, 22 },
		{ 320, 252, 22 },
		{ 320, 253, 22 },
		{ 320, 254, 22 },
		{ 320, 255, 22 },
		{ 320, 256, 22 },
		{ 320, 257, 22 },
		{ 320, 258, 22 },
		{ 320, 259, 22 },
		{ 320, 260, 22 },
		{ 320, 261, 22 },
		{ 320, 262, 22 },
		{ 320, 263, 22 },
		{ 320, 264, 22 },
		{ 320, 265, 22 },
		{ 320, 266, 22 },
		{ 320, 267, 22 },
		{ 320, 268, 22 },
		{ 320, 269, 22 },
		{ 320, 270, 22 },
		{ 320, 271, 22 },
		{ 320, 272, 22 },
		{ 321, 242, 22 },
		{ 321, 243, 22 },
		{ 321, 244, 22 },
		{ 321, 245, 22 },
		{ 321, 246, 22 },
		{ 321, 247, 22 },
		{ 321, 248, 22 },
		{ 321, 249, 22 },
		{ 321, 250, 22 },
		{ 321, 251, 22 },
		{ 321, 252, 22 },
		{ 321, 253, 22 },
		{ 321, 254, 22 },
		{ 321, 255, 22 },
		{ 321, 256, 22 },
		{ 321, 257, 22 },
		{ 321, 258, 22 },
		{ 321, 259, 22 },
		{ 321, 260, 22 },
		{ 321, 261, 22 },
		{ 321, 262, 22 },
		{ 321, 263, 22 },
		{ 321, 264, 22 },
		{ 321, 265, 22 },
		{ 321, 266, 22 },
		{ 321, 267, 22 },
		{ 321, 268, 22 },
		{ 321, 269, 22 },
		{ 321, 270, 22 },
		{ 321, 271, 22 },
		{ 321, 272, 22 },
		{ 322, 241, 22 },
		{ 322, 242, 22 },
		{ 322, 243, 22 },
		{ 322, 244, 22 },
		{ 322, 245, 22 },
		{ 322, 246, 22 },
		{ 322, 247, 22 },
		{ 322, 248, 22 },
		{ 322, 249, 22 },
		{ 322, 250, 22 },
		{ 322, 251, 22 },
		{ 322, 252, 22 },
		{ 322, 253, 22 },
		{ 322, 254, 22 },
		{ 322, 255, 22 },
		{ 322, 256, 22 },
		{ 322, 257, 22 },
		{ 322, 258, 22 },
		{ 322, 259, 22 },
		{ 322, 260, 22 },
		{ 322, 261, 22 },
		{ 322, 262, 22 },
		{ 322, 263, 22 },
		{ 322, 264, 22 },
		{ 322, 265, 22 },
		{ 322, 266, 22 },
		{ 322, 267, 22 },
		{ 322, 268, 22 },
		{ 322, 269, 22 },
		{ 322, 270, 22 },
		{ 322, 271, 22 },
		{ 322, 272, 22 },
		{ 322, 273, 22 },
		{ 323, 241, 22 },
		{ 323, 242, 22 },
		{ 323, 243, 22 },
		{ 323, 244, 22 },
		{ 323, 245, 22 },
		{ 323, 246, 22 },
		{ 323, 247, 22 },
		{ 323, 248, 22 },
		{ 323, 249, 22 },
		{ 323, 250, 22 },
		{ 323, 251, 22 },
		{ 323, 252, 22 },
		{ 323, 253, 22 },
		{ 323, 254, 22 },
		{ 323, 255, 22 },
		{ 323, 256, 22 },
		{ 323, 257, 22 },
		{ 323, 258, 22 },
		{ 323, 259, 22 },
		{ 323, 260, 22 },
		{ 323, 261, 22 },
		{ 323, 262, 22 },
		{ 323, 263, 22 },
		{ 323, 264, 22 },
		{ 323, 265, 22 },
		{ 323, 266, 22 },
		{ 323, 267, 22 },
		{ 323, 268, 22 },
		{ 323, 269, 22 },
		{ 323, 270, 22 },
		{ 323, 271, 22 },
		{ 323, 272, 22 },
		{ 323, 273, 22 },
		{ 324, 241, 22 },
		{ 324, 242, 22 },
		{ 324, 243, 22 },
		{ 324, 244, 22 },
		{ 324, 245, 22 },
		{ 324, 246, 22 },
		{ 324, 247, 22 },
		{ 324, 248, 22 },
		{ 324, 249, 22 },
		{ 324, 250, 22 },
		{ 324, 251, 22 },
		{ 324, 252, 22 },
		{ 324, 253, 22 },
		{ 324, 254, 22 },
		{ 324, 255, 22 },
		{ 324, 256, 22 },
		{ 324, 257, 22 },
		{ 324, 258, 22 },
		{ 324, 259, 22 },
		{ 324, 260, 22 },
		{ 324, 261, 22 },
		{ 324, 262, 22 },
		{ 324, 263, 22 },
		{ 324, 264, 22 },
		{ 324, 265, 22 },
		{ 324, 266, 22 },
		{ 324, 267, 22 },
		{ 324, 268, 22 },
		{ 324, 269, 22 },
		{ 324, 270, 22 },
		{ 324, 271, 22 },
		{ 324, 272, 22 },
		{ 324, 273, 22 },
		{ 325, 241, 22 },
		{ 325, 242, 22 },
		{ 325, 243, 22 },
		{ 325, 244, 22 },
		{ 325, 245, 22 },
		{ 325, 246, 22 },
		{ 325, 247, 22 },
		{ 325, 248, 22 },
		{ 325, 249, 22 },
		{ 325, 250, 22 },
		{ 325, 251, 22 },
		{ 325, 252, 22 },
		{ 325, 253, 22 },
		{ 325, 254, 22 },
		{ 325, 255, 22 },
		{ 325, 256, 22 },
		{ 325, 257, 22 },
		{ 325, 258, 22 },
		{ 325, 259, 22 },
		{ 325, 260, 22 },
		{ 325, 261, 22 },
		{ 325, 262, 22 },
		{ 325, 263, 22 },
		{ 325, 264, 22 },
		{ 325, 265, 22 },
		{ 325, 266, 22 },
		{ 325, 267, 22 },
		{ 325, 268, 22 },
		{ 325, 269, 22 },
		{ 325, 270, 22 },
		{ 325, 271, 22 },
		{ 325, 272, 22 },
		{ 325, 273, 22 },
		{ 326, 240, 22 },
		{ 326, 241, 22 },
		{ 326, 242, 22 },
		{ 326, 243, 22 },
		{ 326, 244, 22 },
		{ 326, 245, 22 },
		{ 326, 246, 22 },
		{ 326, 247, 22 },
		{ 326, 248, 22 },
		{ 326, 249, 22 },
		{ 326, 250, 22 },
		{ 326, 251, 22 },
		{ 326, 252, 22 },
		{ 326, 253, 22 },
		{ 326, 254, 22 },
		{ 326, 255, 22 },
		{ 326, 256, 22 },
		{ 326, 257, 22 },
		{ 326, 258, 22 },
		{ 326, 259, 22 },
		{ 326, 260, 22 },
		{ 326, 261, 22 },
		{ 326, 262, 22 },
		{ 326, 263, 22 },
		{ 326, 264, 22 },
		{ 326, 265, 22 },
		{ 326, 266, 22 },
		{ 326, 267, 22 },
		{ 326, 268, 22 },
		{ 326, 269, 22 },
		{ 326, 270, 22 },
		{ 326, 271, 22 },
		{ 326, 272, 22 },
		{ 326, 273, 22 },
		{ 326, 274, 22 },
		{ 327, 240, 22 },
		{ 327, 241, 22 },
		{ 327, 242, 22 },
		{ 327, 243, 22 },
		{ 327, 244, 22 },
		{ 327, 245, 22 },
		{ 327, 246, 22 },
		{ 327, 247, 22 },
		{ 327, 248, 22 },
		{ 327, 249, 22 },
		{ 327, 250, 22 },
		{ 327, 251, 22 },
		{ 327, 252, 22 },
		{ 327, 253, 22 },
		{ 327, 254, 22 },
		{ 327, 255, 22 },
		{ 327, 256, 22 },
		{ 327, 257, 22 },
		{ 327, 258, 22 },
		{ 327, 259, 22 },
		{ 327, 260, 22 },
		{ 327, 261, 22 },
		{ 327, 262, 22 },
		{ 327, 263, 22 },
		{ 327, 264, 22 },
		{ 327, 265, 22 },
		{ 327, 266, 22 },
		{ 327, 267, 22 },
		{ 327, 268, 22 },
		{ 327, 269, 22 },
		{ 327, 270, 22 },
		{ 327, 271, 22 },
		{ 327, 272, 22 },
		{ 327, 273, 22 },
		{ 327, 274, 22 },
		{ 328, 240, 22 },
		{ 328, 241, 22 },
		{ 328, 242, 22 },
		{ 328, 243, 22 },
		{ 328, 244, 22 },
		{ 328, 245, 22 },
		{ 328, 246, 22 },
		{ 328, 247, 22 },
		{ 328, 248, 22 },
		{ 328, 249, 22 },
		{ 328, 250, 22 },
		{ 328, 251, 22 },
		{ 328, 252, 22 },
		{ 328, 253, 22 },
		{ 328, 254, 22 },
		{ 328, 255, 22 },
		{ 328, 256, 22 },
		{ 328, 257, 22 },
		{ 328, 258, 22 },
		{ 328, 259, 22 },
		{ 328, 260, 22 },
		{ 328, 261, 22 },
		{ 328, 262, 22 },
		{ 328, 263, 22 },
		{ 328, 264, 22 },
		{ 328, 265, 22 },
		{ 328, 266, 22 },
		{ 328, 267, 22 },
		{ 328, 268, 22 },
		{ 328, 269, 22 },
		{ 328, 270, 22 },
		{ 328, 271, 22 },
		{ 328, 272, 22 },
		{ 328, 273, 22 },
		{ 328, 274, 22 },
		{ 329, 240, 22 },
		{ 329, 241, 22 },
		{ 329, 242, 22 },
		{ 329, 243, 22 },
		{ 329, 244, 22 },
		{ 329, 245, 22 },
		{ 329, 246, 22 },
		{ 329, 247, 22 },
		{ 329, 248, 22 },
		{ 329, 249, 22 },
		{ 329, 250, 22 },
		{ 329, 251, 22 },
		{ 329, 252, 22 },
		{ 329, 253, 22 },
		{ 329, 254, 22 },
		{ 329, 255, 22 },
		{ 329, 256, 22 },
		{ 329, 257, 22 },
		{ 329, 258, 22 },
		{ 329, 259, 22 },
		{ 329, 260, 22 },
		{ 329, 261, 22 },
		{ 329, 262, 22 },
		{ 329, 263, 22 },
		{ 329, 264, 22 },
		{ 329, 265, 22 },
		{ 329, 266, 22 },
		{ 329, 267, 22 },
		{ 329, 268, 22 },
		{ 329, 269, 22 },
		{ 329, 270, 22 },
		{ 329, 271, 22 },
		{ 329, 272, 22 },
		{ 329, 273, 22 },
		{ 329, 274, 22 },
		{ 330, 239, 22 },
		{ 330, 240, 22 },
		{ 330, 241, 22 },
		{ 330, 242, 22 },
		{ 330, 243, 22 },
		{ 330, 244, 22 },
		{ 330, 245, 22 },
		{ 330, 246, 22 },
		{ 330, 247, 22 },
		{ 330, 248, 22 },
		{ 330, 249, 22 },
		{ 330, 250, 22 },
		{ 330, 251, 22 },
		{ 330, 252, 22 },
		{ 330, 253, 22 },
		{ 330, 254, 22 },
		{ 330, 255, 22 },
		{ 330, 256, 22 },
		{ 330, 257, 22 },
		{ 330, 258, 22 },
		{ 330, 259, 22 },
		{ 330, 260, 22 },
		{ 330, 261, 22 },
		{ 330, 262, 22 },
		{ 330, 263, 22 },
		{ 330, 264, 22 },
		{ 330, 265, 22 },
		{ 330, 266, 22 },
		{ 330, 267, 22 },
		{ 330, 268, 22 },
		{ 330, 269, 22 },
		{ 330, 270, 22 },
		{ 330, 271, 22 },
		{ 330, 272, 22 },
		{ 330, 273, 22 },
		{ 330, 274, 22 },
		{ 330, 275, 22 },
		{ 331, 239, 22 },
		{ 331, 240, 22 },
		{ 331, 241, 22 },
		{ 331, 242, 22 },
		{ 331, 243, 22 },
		{ 331, 244, 22 },
		{ 331, 245, 22 },
		{ 331, 246, 22 },
		{ 331, 247, 22 },
		{ 331, 248, 22 },
		{ 331, 249, 22 },
		{ 331, 250, 22 },
		{ 331, 251, 22 },
		{ 331, 252, 22 },
		{ 331, 253, 22 },
		{ 331, 254, 22 },
		{ 331, 255, 22 },
		{ 331, 256, 22 },
		{ 331, 257, 22 },
		{ 331, 258, 22 },
		{ 331, 259, 22 },
		{ 331, 260, 22 },
		{ 331, 261, 22 },
		{ 331, 262, 22 },
		{ 331, 263, 22 },
		{ 331, 264, 22 },
		{ 331, 265, 22 },
		{ 331, 266, 22 },
		{ 331, 267, 22 },
		{ 331, 268, 22 },
		{ 331, 269, 22 },
		{ 331, 270, 22 },
		{ 331, 271, 22 },
		{ 331, 272, 22 },
		{ 331, 273, 22 },
		{ 331, 274, 22 },
		{ 331, 275, 22 },
		{ 332, 239, 22 },
		{ 332, 240, 22 },
		{ 332, 241, 22 },
		{ 332, 242, 22 },
		{ 332, 243, 22 },
		{ 332, 244, 22 },
		{ 332, 245, 22 },
		{ 332, 246, 22 },
		{ 332, 247, 22 },
		{ 332, 248, 22 },
		{ 332, 249, 22 },
		{ 332, 250, 22 },
		{ 332, 251, 22 },
		{ 332, 252, 22 },
		{ 332, 253, 22 },
		{ 332, 254, 22 },
		{ 332, 255, 22 },
		{ 332, 256, 22 },
		{ 332, 257, 22 },
		{ 332, 258, 22 },
		{ 332, 259, 22 },
		{ 332, 260, 22 },
		{ 332, 261, 22 },
		{ 332, 262, 22 },
		{ 332, 263, 22 },
		{ 332, 264, 22 },
		{ 332, 265, 22 },
		{ 332, 266, 22 },
		{ 332, 267, 22 },
		{ 332, 268, 22 },
		{ 332, 269, 22 },
		{ 332, 270, 22 },
		{ 332, 271, 22 },
		{ 332, 272, 22 },
		{ 332, 273, 22 },
		{ 332, 274, 22 },
		{ 332, 275, 22 },
		{ 333, 239, 22 },
		{ 333, 240, 22 },
		{ 333, 241, 22 },
		{ 333, 242, 22 },
		{ 333, 243, 22 },
		{ 333, 244, 22 },
		{ 333, 245, 22 },
		{ 333, 246, 22 },
		{ 333, 247, 22 },
		{ 333, 248, 22 },
		{ 333, 249, 22 },
		{ 333, 250, 22 },
		{ 333, 251, 22 },
		{ 333, 252, 22 },
		{ 333, 253, 22 },
		{ 333, 254, 22 },
		{ 333, 255, 22 },
		{ 333, 256, 22 },
		{ 333, 257, 22 },
		{ 333, 258, 22 },
		{ 333, 259, 22 },
		{ 333, 260, 22 },
		{ 333, 261, 22 },
		{ 333, 262, 22 },
		{ 333, 263, 22 },
		{ 333, 264, 22 },
		{ 333, 265, 22 },
		{ 333, 266, 22 },
		{ 333, 267, 22 },
		{ 333, 268, 22 },
		{ 333, 269, 22 },
		{ 333, 270, 22 },
		{ 333, 271, 22 },
		{ 333, 272, 22 },
		{ 333, 273, 22 },
		{ 333, 274, 22 },
		{ 333, 275, 22 },
		{ 334, 238, 22 },
		{ 334, 239, 22 },
		{ 334, 240, 22 },
		{ 334, 241, 22 },
		{ 334, 242, 22 },
		{ 334, 243, 22 },
		{ 334, 244, 22 },
		{ 334, 245, 22 },
		{ 334, 246, 22 },
		{ 334, 247, 22 },
		{ 334, 248, 22 },
		{ 334, 249, 22 },
		{ 334, 250, 22 },
		{ 334, 251, 22 },
		{ 334, 252, 22 },
		{ 334, 253, 22 },
		{ 334, 254, 22 },
		{ 334, 255, 22 },
		{ 334, 256, 22 },
		{ 334, 257, 22 },
		{ 334, 258, 22 },
		{ 334, 259, 22 },
		{ 334, 260, 22 },
		{ 334, 261, 22 },
		{ 334, 262, 22 },
		{ 334, 263, 22 },
		{ 334, 264, 22 },
		{ 334, 265, 22 },
		{ 334, 266, 22 },
		{ 334, 267, 22 },
		{ 334, 268, 22 },
		{ 334, 269, 22 },
		{ 334, 270, 22 },
		{ 334, 271, 22 },
		{ 334, 272, 22 },
		{ 334, 273, 22 },
		{ 334, 274, 22 },
		{ 334, 275, 22 },
		{ 334, 276, 22 },
		{ 335, 238, 22 },
		{ 335, 239, 22 },
		{ 335, 240, 22 },
		{ 335, 241, 22 },
		{ 335, 242, 22 },
		{ 335, 243, 22 },
		{ 335, 244, 22 },
		{ 335, 245, 22 },
		{ 335, 246, 22 },
		{ 335, 247, 22 },
		{ 335, 248, 22 },
		{ 335, 249, 22 },
		{ 335, 250, 22 },
		{ 335, 251, 22 },
		{ 335, 252, 22 },
		{ 335, 253, 22 },
		{ 335, 254, 22 },
		{ 335, 255, 22 },
		{ 335, 256, 22 },
		{ 335, 257, 22 },
		{ 335, 258, 22 },
		{ 335, 259, 22 },
		{ 335, 260, 22 },
		{ 335, 261, 22 },
		{ 335, 262, 22 },
		{ 335, 263, 22 },
		{ 335, 264, 22 },
		{ 335, 265, 22 },
		{ 335, 266, 22 },
		{ 335, 267, 22 },
		{ 335, 268, 22 },
		{ 335, 269, 22 },
		{ 335, 270, 22 },
		{ 335, 271, 22 },
		{ 335, 272, 22 },
		{ 335, 273, 22 },
		{ 335, 274, 22 },
		{ 335, 275, 22 },
		{ 335, 276, 22 },
		{ 336, 238, 22 },
		{ 336, 239, 22 },
		{ 336, 240, 22 },
		{ 336, 241, 22 },
		{ 336, 242, 22 },
		{ 336, 243, 22 },
		{ 336, 244, 22 },
		{ 336, 245, 22 },
		{ 336, 246, 22 },
		{ 336, 247, 22 },
		{ 336, 248, 22 },
		{ 336, 249, 22 },
		{ 336, 250, 22 },
		{ 336, 251, 22 },
		{ 336, 252, 22 },
		{ 336, 253, 22 },
		{ 336, 254, 22 },
		{ 336, 255, 22 },
		{ 336, 256, 22 },
		{ 336, 257, 22 },
		{ 336, 258, 22 },
		{ 336, 259, 22 },
		{ 336, 260, 22 },
		{ 336, 261, 22 },
		{ 336, 262, 22 },
		{ 336, 263, 22 },
		{ 336, 264, 22 },
		{ 336, 265, 22 },
		{ 336, 266, 22 },
		{ 336, 267, 22 },
		{ 336, 268, 22 },
		{ 336, 269, 22 },
		{ 336, 270, 22 },
		{ 336, 271, 22 },
		{ 336, 272, 22 },
		{ 336, 273, 22 },
		{ 336, 274, 22 },
		{ 336, 275, 22 },
		{ 336, 276, 22 },
		{ 337, 238, 22 },
		{ 337, 239, 22 },
		{ 337, 240, 22 },
		{ 337, 241, 22 },
		{ 337, 242, 22 },
		{ 337, 243, 22 },
		{ 337, 244, 22 },
		{ 337, 245, 22 },
		{ 337, 246, 22 },
		{ 337, 247, 22 },
		{ 337, 248, 22 },
		{ 337, 249, 22 },
		{ 337, 250, 22 },
		{ 337, 251, 22 },
		{ 337, 252, 22 },
		{ 337, 253, 22 },
		{ 337, 254, 22 },
		{ 337, 255, 22 },
		{ 337, 256, 22 },
		{ 337, 257, 22 },
		{ 337, 258, 22 },
		{ 337, 259, 22 },
		{ 337, 260, 22 },
		{ 337, 261, 22 },
		{ 337, 262, 22 },
		{ 337, 263, 22 },
		{ 337, 264, 22 },
		{ 337, 265, 22 },
		{ 337, 266, 22 },
		{ 337, 267, 22 },
		{ 337, 268, 22 },
		{ 337, 269, 22 },
		{ 337, 270, 22 },
		{ 337, 271, 22 },
		{ 337, 272, 22 },
		{ 337, 273, 22 },
		{ 337, 274, 22 },
		{ 337, 275, 22 },
		{ 337, 276, 22 },
		{ 338, 237, 22 },
		{ 338, 238, 22 },
		{ 338, 239, 22 },
		{ 338, 240, 22 },
		{ 338, 241, 22 },
		{ 338, 242, 22 },
		{ 338, 243, 22 },
		{ 338, 244, 22 },
		{ 338, 245, 22 },
		{ 338, 246, 22 },
		{ 338, 247, 22 },
		{ 338, 248, 22 },
		{ 338, 249, 22 },
		{ 338, 250, 22 },
		{ 338, 251, 22 },
		{ 338, 252, 22 },
		{ 338, 253, 22 },
		{ 338, 254, 22 },
		{ 338, 255, 22 },
		{ 338, 256, 22 },
		{ 338, 257, 22 },
		{ 338, 258, 22 },
		{ 338, 259, 22 },
		{ 338, 260, 22 },
		{ 338, 261, 22 },
		{ 338, 262, 22 },
		{ 338, 263, 22 },
		{ 338, 264, 22 },
		{ 338, 265, 22 },
		{ 338, 266, 22 },
		{ 338, 267, 22 },
		{ 338, 268, 22 },
		{ 338, 269, 22 },
		{ 338, 270, 22 },
		{ 338, 271, 22 },
		{ 338, 272, 22 },
		{ 338, 273, 22 },
		{ 338, 274, 22 },
		{ 338, 275, 22 },
		{ 338, 276, 22 },
		{ 338, 277, 22 },
		{ 339, 237, 22 },
		{ 339, 238, 22 },
		{ 339, 239, 22 },
		{ 339, 240, 22 },
		{ 339, 241, 22 },
		{ 339, 242, 22 },
		{ 339, 243, 22 },
		{ 339, 244, 22 },
		{ 339, 245, 22 },
		{ 339, 246, 22 },
		{ 339, 247, 22 },
		{ 339, 248, 22 },
		{ 339, 249, 22 },
		{ 339, 250, 22 },
		{ 339, 251, 22 },
		{ 339, 252, 22 },
		{ 339, 253, 22 },
		{ 339, 254, 22 },
		{ 339, 255, 22 },
		{ 339, 256, 22 },
		{ 339, 257, 22 },
		{ 339, 258, 22 },
		{ 339, 259, 22 },
		{ 339, 260, 22 },
		{ 339, 261, 22 },
		{ 339, 262, 22 },
		{ 339, 263, 22 },
		{ 339, 264, 22 },
		{ 339, 265, 22 },
		{ 339, 266, 22 },
		{ 339, 267, 22 },
		{ 339, 268, 22 },
		{ 339, 269, 22 },
		{ 339, 270, 22 },
		{ 339, 271, 22 },
		{ 339, 272, 22 },
		{ 339, 273, 22 },
		{ 339, 274, 22 },
		{ 339, 275, 22 },
		{ 339, 276, 22 },
		{ 339, 277, 22 },
		{ 340, 237, 22 },
		{ 340, 238, 22 },
		{ 340, 239, 22 },
		{ 340, 240, 22 },
		{ 340, 241, 22 },
		{ 340, 242, 22 },
		{ 340, 243, 22 },
		{ 340, 244, 22 },
		{ 340, 245, 22 },
		{ 340, 246, 22 },
		{ 340, 247, 22 },
		{ 340, 248, 22 },
		{ 340, 249, 22 },
		{ 340, 250, 22 },
		{ 340, 251, 22 },
		{ 340, 252, 22 },
		{ 340, 253, 22 },
		{ 340, 254, 22 },
		{ 340, 255, 22 },
		{ 340, 256, 22 },
		{ 340, 257, 22 },
		{ 340, 258, 22 },
		{ 340, 259, 22 },
		{ 340, 260, 22 },
		{ 340, 261, 22 },
		{ 340, 262, 22 },
		{ 340, 263, 22 },
		{ 340, 264, 22 },
		{ 340, 265, 22 },
		{ 340, 266, 22 },
		{ 340, 267, 22 },
		{ 340, 268, 22 },
		{ 340, 269, 22 },
		{ 340, 270, 22 },
		{ 340, 271, 22 },
		{ 340, 272, 22 },
		{ 340, 273, 22 },
		{ 340, 274, 22 },
		{ 340, 275, 22 },
		{ 340, 276, 22 },
		{ 340, 277, 22 },
		{ 341, 237, 22 },
		{ 341, 238, 22 },
		{ 341, 239, 22 },
		{ 341, 240, 22 },
		{ 341, 241, 22 },
		{ 341, 242, 22 },
		{ 341, 243, 22 },
		{ 341, 244, 22 },
		{ 341, 245, 22 },
		{ 341, 246, 22 },
		{ 341, 247, 22 },
		{ 341, 248, 22 },
		{ 341, 249, 22 },
		{ 341, 250, 22 },
		{ 341, 251, 22 },
		{ 341, 252, 22 },
		{ 341, 253, 22 },
		{ 341, 254, 22 },
		{ 341, 255, 22 },
		{ 341, 256, 22 },
		{ 341, 257, 22 },
		{ 341, 258, 22 },
		{ 341, 259, 22 },
		{ 341, 260, 22 },
		{ 341, 261, 22 },
		{ 341, 262, 22 },
		{ 341, 263, 22 },
		{ 341, 264, 22 },
		{ 341, 265, 22 },
		{ 341, 266, 22 },
		{ 341, 267, 22 },
		{ 341, 268, 22 },
		{ 341, 269, 22 },
		{ 341, 270, 22 },
		{ 341, 271, 22 },
		{ 341, 272, 22 },
		{ 341, 273, 22 },
		{ 341, 274, 22 },
		{ 341, 275, 22 },
		{ 341, 276, 22 },
		{ 341, 277, 22 },
		{ 342, 236, 22 },
		{ 342, 237, 22 },
		{ 342, 238, 22 },
		{ 342, 239, 22 },
		{ 342, 240, 22 },
		{ 342, 241, 22 },
		{ 342, 242, 22 },
		{ 342, 243, 22 },
		{ 342, 244, 22 },
		{ 342, 245, 22 },
		{ 342, 246, 22 },
		{ 342, 247, 22 },
		{ 342, 248, 22 },
		{ 342, 249, 22 },
		{ 342, 250, 22 },
		{ 342, 251, 22 },
		{ 342, 252, 22 },
		{ 342, 253, 22 },
		{ 342, 254, 22 },
		{ 342, 255, 22 },
		{ 342, 256, 22 },
		{ 342, 257, 22 },
		{ 342, 258, 22 },
		{ 342, 259, 22 },
		{ 342, 260, 22 },
		{ 342, 261, 22 },
		{ 342, 262, 22 },
		{ 342, 263, 22 },
		{ 342, 264, 22 },
		{ 342, 265, 22 },
		{ 342, 266, 22 },
		{ 342, 267, 22 },
		{ 342, 268, 22 },
		{ 342, 269, 22 },
		{ 342, 270, 22 },
		{ 342, 271, 22 },
		{ 342, 272, 22 },
		{ 342, 273, 22 },
		{ 342, 274, 22 },
		{ 342, 275, 22 },
		{ 342, 276, 22 },
		{ 342, 277, 22 },
		{ 342, 278, 22 },
		{ 343, 236, 22 },
		{ 343, 237, 22 },
		{ 343, 238, 22 },
		{ 343, 239, 22 },
		{ 343, 240, 22 },
		{ 343, 241, 22 },
		{ 343, 242, 22 },
		{ 343, 243, 22 },
		{ 343, 244, 22 },
		{ 343, 245, 22 },
		{ 343, 246, 22 },
		{ 343, 247, 22 },
		{ 343, 248, 22 },
		{ 343, 249, 22 },
		{ 343, 250, 22 },
		{ 343, 251, 22 },
		{ 343, 252, 22 },
		{ 343, 253, 22 },
		{ 343, 254, 22 },
		{ 343, 255, 22 },
		{ 343, 256, 22 },
		{ 343, 257, 22 },
		{ 343, 258, 22 },
		{ 343, 259, 22 },
		{ 343, 260, 22 },
		{ 343, 261, 22 },
		{ 343, 262, 22 },
		{ 343, 263, 22 },
		{ 343, 264, 22 },
		{ 343, 265, 22 },
		{ 343, 266, 22 },
		{ 343, 267, 22 },
		{ 343, 268, 22 },
		{ 343, 269, 22 },
		{ 343, 270, 22 },
		{ 343, 271, 22 },
		{ 343, 272, 22 },
		{ 343, 273, 22 },
		{ 343, 274, 22 },
		{ 343, 275, 22 },
		{ 343, 276, 22 },
		{ 343, 277, 22 },
		{ 343, 278, 22 },
		{ 344, 236, 22 },
		{ 344, 237, 22 },
		{ 344, 238, 22 },
		{ 344, 239, 22 },
		{ 344, 240, 22 },
		{ 344, 241, 22 },
		{ 344, 242, 22 },
		{ 344, 243, 22 },
		{ 344, 244, 22 },
		{ 344, 245, 22 },
		{ 344, 246, 22 },
		{ 344, 247, 22 },
		{ 344, 248, 22 },
		{ 344, 249, 22 },
		{ 344, 250, 22 },
		{ 344, 251, 22 },
		{ 344, 252, 22 },
		{ 344, 253, 22 },
		{ 344, 254, 22 },
		{ 344, 255, 22 },
		{ 344, 256, 22 },
		{ 344, 257, 22 },
		{ 344, 258, 22 },
		{ 344, 259, 22 },
		{ 344, 260, 22 },
		{ 344, 261, 22 },
		{ 344, 262, 22 },
		{ 344, 263, 22 },
		{ 344, 264, 22 },
		{ 344, 265, 22 },
		{ 344, 266, 22 },
		{ 344, 267, 22 },
		{ 344, 268, 22 },
		{ 344, 269, 22 },
		{ 344, 270, 22 },
		{ 344, 271, 22 },
		{ 344, 272, 22 },
		{ 344, 273, 22 },
		{ 344, 274, 22 },
		{ 344, 275, 22 },
		{ 344, 276, 22 },
		{ 344, 277, 22 },
		{ 344, 278, 22 },
		{ 345, 236, 22 },
		{ 345, 237, 22 },
		{ 345, 238, 22 },
		{ 345, 239, 22 },
		{ 345, 240, 22 },
		{ 345, 241, 22 },
		{ 345, 242, 22 },
		{ 345, 243, 22 },
		{ 345, 244, 22 },
		{ 345, 245, 22 },
		{ 345, 246, 22 },
		{ 345, 247, 22 },
		{ 345, 248, 22 },
		{ 345, 249, 22 },
		{ 345, 250, 22 },
		{ 345, 251, 22 },
		{ 345, 252, 22 },
		{ 345, 253, 22 },
		{ 345, 254, 22 },
		{ 345, 255, 22 },
		{ 345, 256, 22 },
		{ 345, 257, 22 },
		{ 345, 258, 22 },
		{ 345, 259, 22 },
		{ 345, 260, 22 },
		{ 345, 261, 22 },
		{ 345, 262, 22 },
		{ 345, 263, 22 },
		{ 345, 264, 22 },
		{ 345, 265, 22 },
		{ 345, 266, 22 },
		{ 345, 267, 22 },
		{ 345, 268, 22 },
		{ 345, 269, 22 },
		{ 345, 270, 22 },
		{ 345, 271, 22 },
		{ 345, 272, 22 },
		{ 345, 273, 22 },
		{ 345, 274, 22 },
		{ 345, 275, 22 },
		{ 345, 276, 22 },
		{ 345, 277, 22 },
		{ 345, 278, 22 },
		{ 346, 235, 22 },
		{ 346, 236, 22 },
		{ 346, 237, 22 },
		{ 346, 238, 22 },
		{ 346, 239, 22 },
		{ 346, 240, 22 },
		{ 346, 241, 22 },
		{ 346, 242, 22 },
		{ 346, 243, 22 },
		{ 346, 244, 22 },
		{ 346, 245, 22 },
		{ 346, 246, 22 },
		{ 346, 247, 22 },
		{ 346, 248, 22 },
		{ 346, 249, 22 },
		{ 346, 250, 22 },
		{ 346, 251, 22 },
		{ 346, 252, 22 },
		{ 346, 253, 22 },
		{ 346, 254, 22 },
		{ 346, 255, 22 },
		{ 346, 256, 22 },
		{ 346, 257, 22 },
		{ 346, 258, 22 },
		{ 346, 259, 22 },
		{ 346, 260, 22 },
		{ 346, 261, 22 },
		{ 346, 262, 22 },
		{ 346, 263, 22 },
		{ 346, 264, 22 },
		{ 346, 265, 22 },
		{ 346, 266, 22 },
		{ 346, 267, 22 },
		{ 346, 268, 22 },
		{ 346, 269, 22 },
		{ 346, 270, 22 },
		{ 346, 271, 22 },
		{ 346, 272, 22 },
		{ 346, 273, 22 },
		{ 346, 274, 22 },
		{ 346, 275, 22 },
		{ 346, 276, 22 },
		{ 346, 277, 22 },
		{ 346, 278, 22 },
		{ 346, 279, 22 },
		{ 347, 235, 22 },
		{ 347, 236, 22 },
		{ 347, 237, 22 },
		{ 347, 238, 22 },
		{ 347, 239, 22 },
		{ 347, 240, 22 },
		{ 347, 241, 22 },
		{ 347, 242, 22 },
		{ 347, 243, 22 },
		{ 347, 244, 22 },
		{ 347, 245, 22 },
		{ 347, 246, 22 },
		{ 347, 247, 22 },
		{ 347, 248, 22 },
		{ 347, 249, 22 },
		{ 347, 250, 22 },
		{ 347, 251, 22 },
		{ 347, 252, 22 },
		{ 347, 253, 22 },
		{ 347, 254, 22 },
		{ 347, 255, 22 },
		{ 347, 256, 22 },
		{ 347, 257, 22 },
		{ 347, 258, 22 },
		{ 347, 259, 22 },
		{ 347, 260, 22 },
		{ 347, 261, 22 },
		{ 347, 262, 22 },
		{ 347, 263, 22 },
		{ 347, 264, 22 },
		{ 347, 265, 22 },
		{ 347, 266, 22 },
		{ 347, 267, 22 },
		{ 347, 268, 22 },
		{ 347, 269, 22 },
		{ 347, 270, 22 },
		{ 347, 271, 22 },
		{ 347, 272, 22 },
		{ 347, 273, 22 },
		{ 347, 274, 22 },
		{ 347, 275, 22 },
		{ 347, 276, 22 },
		{ 347, 277, 22 },
		{ 347, 278, 22 },
		{ 347, 279, 22 },
		{ 348, 235, 22 },
		{ 348, 236, 22 },
		{ 348, 237, 22 },
		{ 348, 238, 22 },
		{ 348, 239, 22 },
		{ 348, 240, 22 },
		{ 348, 241, 22 },
		{ 348, 242, 22 },
		{ 348, 243, 22 },
		{ 348, 244, 22 },
		{ 348, 245, 22 },
		{ 348, 246, 22 },
		{ 348, 247, 22 },
		{ 348, 248, 22 },
		{ 348, 249, 22 },
		{ 348, 250, 22 },
		{ 348, 251, 22 },
		{ 348, 252, 22 },
		{ 348, 253, 22 },
		{ 348, 254, 22 },
		{ 348, 255, 22 },
		{ 348, 256, 22 },
		{ 348, 257, 22 },
		{ 348, 258, 22 },
		{ 348, 259, 22 },
		{ 348, 260, 22 },
		{ 348, 261, 22 },
		{ 348, 262, 22 },
		{ 348, 263, 22 },
		{ 348, 264, 22 },
		{ 348, 265, 22 },
		{ 348, 266, 22 },
		{ 348, 267, 22 },
		{ 348, 268, 22 },
		{ 348, 269, 22 },
		{ 348, 270, 22 },
		{ 348, 271, 22 },
		{ 348, 272, 22 },
		{ 348, 273, 22 },
		{ 348, 274, 22 },
		{ 348, 275, 22 },
		{ 348, 276, 22 },
		{ 348, 277, 22 },
		{ 348, 278, 22 },
		{ 348, 279, 22 },
		{ 349, 235, 22 },
		{ 349, 236, 22 },
		{ 349, 237, 22 },
		{ 349, 238, 22 },
		{ 349, 239, 22 },
		{ 349, 240, 22 },
		{ 349, 241, 22 },
		{ 349, 242, 22 },
		{ 349, 243, 22 },
		{ 349, 244, 22 },
		{ 349, 245, 22 },
		{ 349, 246, 22 },
		{ 349, 247, 22 },
		{ 349, 248, 22 },
		{ 349, 249, 22 },
		{ 349, 250, 22 },
		{ 349, 251, 22 },
		{ 349, 252, 22 },
		{ 349, 253, 22 },
		{ 349, 254, 22 },
		{ 349, 255, 22 },
		{ 349, 256, 22 },
		{ 349, 257, 22 },
		{ 349, 258, 22 },
		{ 349, 259, 22 },
		{ 349, 260, 22 },
		{ 349, 261, 22 },
		{ 349, 262, 22 },
		{ 349, 263, 22 },
		{ 349, 264, 22 },
		{ 349, 265, 22 },
		{ 349, 266, 22 },
		{ 349, 267, 22 },
		{ 349, 268, 22 },
		{ 349, 269, 22 },
		{ 349, 270, 22 },
		{ 349, 271, 22 },
		{ 349, 272, 22 },
		{ 349, 273, 22 },
		{ 349, 274, 22 },
		{ 349, 275, 22 },
		{ 349, 276, 22 },
		{ 349, 277, 22 },
		{ 349, 278, 22 },
		{ 349, 279, 22 },
		{ 350, 234, 22 },
		{ 350, 235, 22 },
		{ 350, 236, 22 },
		{ 350, 237, 22 },
		{ 350, 238, 22 },
		{ 350, 239, 22 },
		{ 350, 240, 22 },
		{ 350, 241, 22 },
		{ 350, 242, 22 },
		{ 350, 243, 22 },
		{ 350, 244, 22 },
		{ 350, 245, 22 },
		{ 350, 246, 22 },
		{ 350, 247, 22 },
		{ 350, 248, 22 },
		{ 350, 249, 22 },
		{ 350, 250, 22 },
		{ 350, 251, 22 },
		{ 350, 252, 22 },
		{ 350, 253, 22 },
		{ 350, 254, 22 },
		{ 350, 255, 22 },
		{ 350, 256, 22 },
		{ 350, 257, 22 },
		{ 350, 258, 22 },
		{ 350, 259, 22 },
		{ 350, 260, 22 },
		{ 350, 261, 22 },
		{ 350, 262, 22 },
		{ 350, 263, 22 },
		{ 350, 264, 22 },
		{ 350, 265, 22 },
		{ 350, 266, 22 },
		{ 350, 267, 22 },
		{ 350, 268, 22 },
		{ 350, 269, 22 },
		{ 350, 270, 22 },
		{ 350, 271, 22 },
		{ 350, 272, 22 },
		{ 350, 273, 22 },
		{ 350, 274, 22 },
		{ 350, 275, 22 },
		{ 350, 276, 22 },
		{ 350, 277, 22 },
		{ 350, 278, 22 },
		{ 350, 279, 22 },
		{ 350, 280, 22 },
		{ 351, 234, 22 },
		{ 351, 235, 22 },
		{ 351, 236, 22 },
		{ 351, 237, 22 },
		{ 351, 238, 22 },
		{ 351, 239, 22 },
		{ 351, 240, 22 },
		{ 351, 241, 22 },
		{ 351, 242, 22 },
		{ 351, 243, 22 },
		{ 351, 244, 22 },
		{ 351, 245, 22 },
		{ 351, 246, 22 },
		{ 351, 247, 22 },
		{ 351, 248, 22 },
		{ 351, 249, 22 },
		{ 351, 250, 22 },
		{ 351, 251, 22 },
		{ 351, 252, 22 },
		{ 351, 253, 22 },
		{ 351, 254, 22 },
		{ 351, 255, 22 },
		{ 351, 256, 22 },
		{ 351, 257, 22 },
		{ 351, 258, 22 },
		{ 351, 259, 22 },
		{ 351, 260, 22 },
		{ 351, 261, 22 },
		{ 351, 262, 22 },
		{ 351, 263, 22 },
		{ 351, 264, 22 },
		{ 351, 265, 22 },
		{ 351, 266, 22 },
		{ 351, 267, 22 },
		{ 351, 268, 22 },
		{ 351, 269, 22 },
		{ 351, 270, 22 },
		{ 351, 271, 22 },
		{ 351, 272, 22 },
		{ 351, 273, 22 },
		{ 351, 274, 22 },
		{ 351, 275, 22 },
		{ 351, 276, 22 },
		{ 351, 277, 22 },
		{ 351, 278, 22 },
		{ 351, 279, 22 },
		{ 351, 280, 22 },
		{ 352, 234, 22 },
		{ 352, 235, 22 },
		{ 352, 236, 22 },
		{ 352, 237, 22 },
		{ 352, 238, 22 },
		{ 352, 239, 22 },
		{ 352, 240, 22 },
		{ 352, 241, 22 },
		{ 352, 242, 22 },
		{ 352, 243, 22 },
		{ 352, 244, 22 },
		{ 352, 245, 22 },
		{ 352, 246, 22 },
		{ 352, 247, 22 },
		{ 352, 248, 22 },
		{ 352, 249, 22 },
		{ 352, 250, 22 },
		{ 352, 251, 22 },
		{ 352, 252, 22 },
		{ 352, 253, 22 },
		{ 352, 254, 22 },
		{ 352, 255, 22 },
		{ 352, 256, 22 },
		{ 352, 257, 22 },
		{ 352, 258, 22 },
		{ 352, 259, 22 },
		{ 352, 260, 22 },
		{ 352, 261, 22 },
		{ 352, 262, 22 },
		{ 352, 263, 22 },
		{ 352, 264, 22 },
		{ 352, 265, 22 },
		{ 352, 266, 22 },
		{ 352, 267, 22 },
		{ 352, 268, 22 },
		{ 352, 269, 22 },
		{ 352, 270, 22 },
		{ 352, 271, 22 },
		{ 352, 272, 22 },
		{ 352, 273, 22 },
		{ 352, 274, 22 },
		{ 352, 275, 22 },
		{ 352, 276, 22 },
		{ 352, 277, 22 },
		{ 352, 278, 22 },
		{ 352, 279, 22 },
		{ 352, 280, 22 },
		{ 353, 234, 22 },
		{ 353, 235, 22 },
		{ 353, 236, 22 },
		{ 353, 237, 22 },
		{ 353, 238, 22 },
		{ 353, 239, 22 },
		{ 353, 240, 22 },
		{ 353, 241, 22 },
		{ 353, 242, 22 },
		{ 353, 243, 22 },
		{ 353, 244, 22 },
		{ 353, 245, 22 },
		{ 353, 246, 22 },
		{ 353, 247, 22 },
		{ 353, 248, 22 },
		{ 353, 249, 22 },
		{ 353, 250, 22 },
		{ 353, 251, 22 },
		{ 353, 252, 22 },
		{ 353, 253, 22 },
		{ 353, 254, 22 },
		{ 353, 255, 22 },
		{ 353, 256, 22 },
		{ 353, 257, 22 },
		{ 353, 258, 22 },
		{ 353, 259, 22 },
		{ 353, 260, 22 },
		{ 353, 261, 22 },
		{ 353, 262, 22 },
		{ 353, 263, 22 },
		{ 353, 264, 22 },
		{ 353, 265, 22 },
		{ 353, 266, 22 },
		{ 353, 267, 22 },
		{ 353, 268, 22 },
		{ 353, 269, 22 },
		{ 353, 270, 22 },
		{ 353, 271, 22 },
		{ 353, 272, 22 },
		{ 353, 273, 22 },
		{ 353, 274, 22 },
		{ 353, 275, 22 },
		{ 353, 276, 22 },
		{ 353, 277, 22 },
		{ 353, 278, 22 },
		{ 353, 279, 22 },
		{ 353, 280, 22 },
		{ 354, 233, 22 },
		{ 354, 234, 22 },
		{ 354, 235, 22 },
		{ 354, 236, 22 },
		{ 354, 237, 22 },
		{ 354, 238, 22 },
		{ 354, 239, 22 },
		{ 354, 240, 22 },
		{ 354, 241, 22 },
		{ 354, 242, 22 },
		{ 354, 243, 22 },
		{ 354, 244, 22 },
		{ 354, 245, 22 },
		{ 354, 246, 22 },
		{ 354, 247, 22 },
		{ 354, 248, 22 },
		{ 354, 249, 22 },
		{ 354, 250, 22 },
		{ 354, 251, 22 },
		{ 354, 252, 22 },
		{ 354, 253, 22 },
		{ 354, 254, 22 },
		{ 354, 255, 22 },
		{ 354, 256, 22 },
		{ 354, 257, 22 },
		{ 354, 258, 22 },
		{ 354, 259, 22 },
		{ 354, 260, 22 },
		{ 354, 261, 22 },
		{ 354, 262, 22 },
		{ 354, 263, 22 },
		{ 354, 264, 22 },
		{ 354, 265, 22 },
		{ 354, 266, 22 },
		{ 354, 267, 22 },
		{ 354, 268, 22 },
		{ 354, 269, 22 },
		{ 354, 270, 22 },
		{ 354, 271, 22 },
		{ 354, 272, 22 },
		{ 354, 273, 22 },
		{ 354, 274, 22 },
		{ 354, 275, 22 },
		{ 354, 276, 22 },
		{ 354, 277, 22 },
		{ 354, 278, 22 },
		{ 354, 279, 22 },
		{ 354, 280, 22 },
		{ 354, 281, 22 },
		{ 355, 233, 22 },
		{ 355, 234, 22 },
		{ 355, 235, 22 },
		{ 355, 236, 22 },
		{ 355, 237, 22 },
		{ 355, 238, 22 },
		{ 355, 239, 22 },
		{ 355, 240, 22 },
		{ 355, 241, 22 },
		{ 355, 242, 22 },
		{ 355, 243, 22 },
		{ 355, 244, 22 },
		{ 355, 245, 22 },
		{ 355, 246, 22 },
		{ 355, 247, 22 },
		{ 355, 248, 22 },
		{ 355, 249, 22 },
		{ 355, 250, 22 },
		{ 355, 251, 22 },
		{ 355, 252, 22 },
		{ 355, 253, 22 },
		{ 355, 254, 22 },
		{ 355, 255, 22 },
		{ 355, 256, 22 },
		{ 355, 257, 22 },
		{ 355, 258, 22 },
		{ 355, 259, 22 },
		{ 355, 260, 22 },
		{ 355, 261, 22 },
		{ 355, 262, 22 },
		{ 355, 263, 22 },
		{ 355, 264, 22 },
		{ 355, 265, 22 },
		{ 355, 266, 22 },
		{ 355, 267, 22 },
		{ 355, 268, 22 },
		{ 355, 269, 22 },
		{ 355, 270, 22 },
		{ 355, 271, 22 },
		{ 355, 272, 22 },
		{ 355, 273, 22 },
		{ 355, 274, 22 },
		{ 355, 275, 22 },
		{ 355, 276, 22 },
		{ 355, 277, 22 },
		{ 355, 278, 22 },
		{ 355, 279, 22 },
		{ 355, 280, 22 },
		{ 355, 281, 22 },
		{ 356, 233, 22 },
		{ 356, 234, 22 },
		{ 356, 235, 22 },
		{ 356, 236, 22 },
		{ 356, 237, 22 },
		{ 356, 238, 22 },
		{ 356, 239, 22 },
		{ 356, 240, 22 },
		{ 356, 241, 22 },
		{ 356, 242, 22 },
		{ 356, 243, 22 },
		{ 356, 244, 22 },
		{ 356, 245, 22 },
		{ 356, 246, 22 },
		{ 356, 247, 22 },
		{ 356, 248, 22 },
		{ 356, 249, 22 },
		{ 356, 250, 22 },
		{ 356, 251, 22 },
		{ 356, 252, 22 },
		{ 356, 253, 22 },
		{ 356, 254, 22 },
		{ 356, 255, 22 },
		{ 356, 256, 22 },
		{ 356, 257, 22 },
		{ 356, 258, 22 },
		{ 356, 259, 22 },
		{ 356, 260, 22 },
		{ 356, 261, 22 },
		{ 356, 262, 22 },
		{ 356, 263, 22 },
		{ 356, 264, 22 },
		{ 356, 265, 22 },
		{ 356, 266, 22 },
		{ 356, 267, 22 },
		{ 356, 268, 22 },
		{ 356, 269, 22 },
		{ 356, 270, 22 },
		{ 356, 271, 22 },
		{ 356, 272, 22 },
		{ 356, 273, 22 },
		{ 356, 274, 22 },
		{ 356, 275, 22 },
		{ 356, 276, 22 },
		{ 356, 277, 22 },
		{ 356, 278, 22 },
		{ 356, 279, 22 },
		{ 356, 280, 22 },
		{ 356, 281, 22 },
		{ 357, 233, 22 },
		{ 357, 234, 22 },
		{ 357, 235, 22 },
		{ 357, 236, 22 },
		{ 357, 237, 22 },
		{ 357, 238, 22 },
		{ 357, 239, 22 },
		{ 357, 240, 22 },
		{ 357, 241, 22 },
		{ 357, 242, 22 },
		{ 357, 243, 22 },
		{ 357, 244, 22 },
		{ 357, 245, 22 },
		{ 357, 246, 22 },
		{ 357, 247, 22 },
		{ 357, 248, 22 },
		{ 357, 249, 22 },
		{ 357, 250, 22 },
		{ 357, 251, 22 },
		{ 357, 252, 22 },
		{ 357, 253, 22 },
		{ 357, 254, 22 },
		{ 357, 255, 22 },
		{ 357, 256, 22 },
		{ 357, 257, 22 },
		{ 357, 258, 22 },
		{ 357, 259, 22 },
		{ 357, 260, 22 },
		{ 357, 261, 22 },
		{ 357, 262, 22 },
		{ 357, 263, 22 },
		{ 357, 264, 22 },
		{ 357, 265, 22 },
		{ 357, 266, 22 },
		{ 357, 267, 22 },
		{ 357, 268, 22 },
		{ 357, 269, 22 },
		{ 357, 270, 22 },
		{ 357, 271, 22 },
		{ 357, 272, 22 },
		{ 357, 273, 22 },
		{ 357, 274, 22 },
		{ 357, 275, 22 },
		{ 357, 276, 22 },
		{ 357, 277, 22 },
		{ 357, 278, 22 },
		{ 357, 279, 22 },
		{ 357, 280, 22 },
		{ 357, 281, 22 },
		{ 358, 232, 22 },
		{ 358, 233, 22 },
		{ 358, 234, 22 },
		{ 358, 235, 22 },
		{ 358, 236, 22 },
		{ 358, 237, 22 },
		{ 358, 238, 22 },
		{ 358, 239, 22 },
		{ 358, 240, 22 },
		{ 358, 241, 22 },
		{ 358, 242, 22 },
		{ 358, 243, 22 },
		{ 358, 244, 22 },
		{ 358, 245, 22 },
		{ 358, 246, 22 },
		{ 358, 247, 22 },
		{ 358, 248, 22 },
		{ 358, 249, 22 },
		{ 358, 250, 22 },
		{ 358, 251, 22 },
		{ 358, 252, 22 },
		{ 358, 253, 22 },
		{ 358, 254, 22 },
		{ 358, 255, 22 },
		{ 358, 256, 22 },
		{ 358, 257, 22 },
		{ 358, 258, 22 },
		{ 358, 259, 22 },
		{ 358, 260, 22 },
		{ 358, 261, 22 },
		{ 358, 262, 22 },
		{ 358, 263, 22 },
		{ 358, 264, 22 },
		{ 358, 265, 22 },
		{ 358, 266, 22 },
		{ 358, 267, 22 },
		{ 358, 268, 22 },
		{ 358, 269, 22 },
		{ 358, 270, 22 },
		{ 358, 271, 22 },
		{ 358, 272, 22 },
		{ 358, 273, 22 },
		{ 358, 274, 22 },
		{ 358, 275, 22 },
		{ 358, 276, 22 },
		{ 358, 277, 22 },
		{ 358, 278, 22 },
		{ 358, 279, 22 },
		{ 358, 280, 22 },
		{ 358, 281, 22 },
		{ 358, 282, 22 },
		{ 359, 232, 22 },
		{ 359, 233, 22 },
		{ 359, 234, 22 },
		{ 359, 235, 22 },
		{ 359, 236, 22 },
		{ 359, 237, 22 },
		{ 359, 238, 22 },
		{ 359, 239, 22 },
		{ 359, 240, 22 },
		{ 359, 241, 22 },
		{ 359, 242, 22 },
		{ 359, 243, 22 },
		{ 359, 244, 22 },
		{ 359, 245, 22 },
		{ 359, 246, 22 },
		{ 359, 247, 22 },
		{ 359, 248, 22 },
		{ 359, 249, 22 },
		{ 359, 250, 22 },
		{ 359, 251, 22 },
		{ 359, 252, 22 },
		{ 359, 253, 22 },
		{ 359, 254, 22 },
		{ 359, 255, 22 },
		{ 359, 256, 22 },
		{ 359, 257, 22 },
		{ 359, 258, 22 },
		{ 359, 259, 22 },
		{ 359, 260, 22 },
		{ 359, 261, 22 },
		{ 359, 262, 22 },
		{ 359, 263, 22 },
		{ 359, 264, 22 },
		{ 359, 265, 22 },
		{ 359, 266, 22 },
		{ 359, 267, 22 },
		{ 359, 268, 22 },
		{ 359, 269, 22 },
		{ 359, 270, 22 },
		{ 359, 271, 22 },
		{ 359, 272, 22 },
		{ 359, 273, 22 },
		{ 359, 274, 22 },
		{ 359, 275, 22 },
		{ 359, 276, 22 },
		{ 359, 277, 22 },
		{ 359, 278, 22 },
		{ 359, 279, 22 },
		{ 359, 280, 22 },
		{ 359, 281, 22 },
		{ 359, 282, 22 },
		{ 360, 232, 22 },
		{ 360, 233, 22 },
		{ 360, 234, 22 },
		{ 360, 235, 22 },
		{ 360, 236, 22 },
		{ 360, 237, 22 },
		{ 360, 238, 22 },
		{ 360, 239, 22 },
		{ 360, 240, 22 },
		{ 360, 241, 22 },
		{ 360, 242, 22 },
		{ 360, 243, 22 },
		{ 360, 244, 22 },
		{ 360, 245, 22 },
		{ 360, 246, 22 },
		{ 360, 247, 22 },
		{ 360, 248, 22 },
		{ 360, 249, 22 },
		{ 360, 250, 22 },
		{ 360, 251, 22 },
		{ 360, 252, 22 },
		{ 360, 253, 22 },
		{ 360, 254, 22 },
		{ 360, 255, 22 },
		{ 360, 256, 22 },
		{ 360, 257, 22 },
		{ 360, 258, 22 },
		{ 360, 259, 22 },
		{ 360, 260, 22 },
		{ 360, 261, 22 },
		{ 360, 262, 22 },
		{ 360, 263, 22 },
		{ 360, 264, 22 },
		{ 360, 265, 22 },
		{ 360, 266, 22 },
		{ 360, 267, 22 },
		{ 360, 268, 22 },
		{ 360, 269, 22 },
		{ 360, 270, 22 },
		{ 360, 271, 22 },
		{ 360, 272, 22 },
		{ 360, 273, 22 },
		{ 360, 274, 22 },
		{ 360, 275, 22 },
		{ 360, 276, 22 },
		{ 360, 277, 22 },
		{ 360, 278, 22 },
		{ 360, 279, 22 },
		{ 360, 280, 22 },
		{ 360, 281, 22 },
		{ 360, 282, 22 },
		{ 361, 232, 22 },
		{ 361, 233, 22 },
		{ 361, 234, 22 },
		{ 361, 235, 22 },
		{ 361, 236, 22 },
		{ 361, 237, 22 },
		{ 361, 238, 22 },
		{ 361, 239, 22 },
		{ 361, 240, 22 },
		{ 361, 241, 22 },
		{ 361, 242, 22 },
		{ 361, 243, 22 },
		{ 361, 244, 22 },
		{ 361, 245, 22 },
		{ 361, 246, 22 },
		{ 361, 247, 22 },
		{ 361, 248, 22 },
		{ 361, 249, 22 },
		{ 361, 250, 22 },
		{ 361, 251, 22 },
		{ 361, 252, 22 },
		{ 361, 253, 22 },
		{ 361, 254, 22 },
		{ 361, 255, 22 },
		{ 361, 256, 22 },
		{ 361, 257, 22 },
		{ 361, 258, 22 },
		{ 361, 259, 22 },
		{ 361, 260, 22 },
		{ 361, 261, 22 },
		{ 361, 262, 22 },
		{ 361, 263, 22 },
		{ 361, 264, 22 },
		{ 361, 265, 22 },
		{ 361, 266, 22 },
		{ 361, 267, 22 },
		{ 361, 268, 22 },
		{ 361, 269, 22 },
		{ 361, 270, 22 },
		{ 361, 271, 22 },
		{ 361, 272, 22 },
		{ 361, 273, 22 },
		{ 361, 274, 22 },
		{ 361, 275, 22 },
		{ 361, 276, 22 },
		{ 361, 277, 22 },
		{ 361, 278, 22 },
		{ 361, 279, 22 },
		{ 361, 280, 22 },
		{ 361, 281, 22 },
		{ 361, 282, 22 },
		{ 362, 231, 22 },
		{ 362, 232, 22 },
		{ 362, 233, 22 },
		{ 362, 234, 22 },
		{ 362, 235, 22 },
		{ 362, 236, 22 },
		{ 362, 237, 22 },
		{ 362, 238, 22 },
		{ 362, 239, 22 },
		{ 362, 240, 22 },
		{ 362, 241, 22 },
		{ 362, 242, 22 },
		{ 362, 243, 22 },
		{ 362, 244, 22 },
		{ 362, 245, 22 },
		{ 362, 246, 22 },
		{ 362, 247, 22 },
		{ 362, 248, 22 },
		{ 362, 249, 22 },
		{ 362, 250, 22 },
		{ 362, 251, 22 },
		{ 362, 252, 22 },
		{ 362, 253, 22 },
		{ 362, 254, 22 },
		{ 362, 255, 22 },
		{ 362, 256, 22 },
		{ 362, 257, 22 },
		{ 362, 258, 22 },
		{ 362, 259, 22 },
		{ 362, 260, 22 },
		{ 362, 261, 22 },
		{ 362, 262, 22 },
		{ 362, 263, 22 },
		{ 362, 264, 22 },
		{ 362, 265, 22 },
		{ 362, 266, 22 },
		{ 362, 267, 22 },
		{ 362, 268, 22 },
		{ 362, 269, 22 },
		{ 362, 270, 22 },
		{ 362, 271, 22 },
		{ 362, 272, 22 },
		{ 362, 273, 22 },
		{ 362, 274, 22 },
		{ 362, 275, 22 },
		{ 362, 276, 22 },
		{ 362, 277, 22 },
		{ 362, 278, 22 },
		{ 362, 279, 22 },
		{ 362, 280, 22 },
		{ 362, 281, 22 },
		{ 362, 282, 22 },
		{ 362, 283, 22 },
		{ 363, 231, 22 },
		{ 363, 232, 22 },
		{ 363, 233, 22 },
		{ 363, 234, 22 },
		{ 363, 235, 22 },
		{ 363, 236, 22 },
		{ 363, 237, 22 },
		{ 363, 238, 22 },
		{ 363, 239, 22 },
		{ 363, 240, 22 },
		{ 363, 241, 22 },
		{ 363, 242, 22 },
		{ 363, 243, 22 },
		{ 363, 244, 22 },
		{ 363, 245, 22 },
		{ 363, 246, 22 },
		{ 363, 247, 22 },
		{ 363, 248, 22 },
		{ 363, 249, 22 },
		{ 363, 250, 22 },
		{ 363, 251, 22 },
		{ 363, 252, 22 },
		{ 363, 253, 22 },
		{ 363, 254, 22 },
		{ 363, 255, 22 },
		{ 363, 256, 22 },
		{ 363, 257, 22 },
		{ 363, 258, 22 },
		{ 363, 259, 22 },
		{ 363, 260, 22 },
		{ 363, 261, 22 },
		{ 363, 262, 22 },
		{ 363, 263, 22 },
		{ 363, 264, 22 },
		{ 363, 265, 22 },
		{ 363, 266, 22 },
		{ 363, 267, 22 },
		{ 363, 268, 22 },
		{ 363, 269, 22 },
		{ 363, 270, 22 },
		{ 363, 271, 22 },
		{ 363, 272, 22 },
		{ 363, 273, 22 },
		{ 363, 274, 22 },
		{ 363, 275, 22 },
		{ 363, 276, 22 },
		{ 363, 277, 22 },
		{ 363, 278, 22 },
		{ 363, 279, 22 },
		{ 363, 280, 22 },
		{ 363, 281, 22 },
		{ 363, 282, 22 },
		{ 363, 283, 22 },
		{ 364, 231, 22 },
		{ 364, 232, 22 },
		{ 364, 233, 22 },
		{ 364, 234, 22 },
		{ 364, 235, 22 },
		{ 364, 236, 22 },
		{ 364, 237, 22 },
		{ 364, 238, 22 },
		{ 364, 239, 22 },
		{ 364, 240, 22 },
		{ 364, 241, 22 },
		{ 364, 242, 22 },
		{ 364, 243, 22 },
		{ 364, 244, 22 },
		{ 364, 245, 22 },
		{ 364, 246, 22 },
		{ 364, 247, 22 },
		{ 364, 248, 22 },
		{ 364, 249, 22 },
		{ 364, 250, 22 },
		{ 364, 251, 22 },
		{ 364, 252, 22 },
		{ 364, 253, 22 },
		{ 364, 254, 22 },
		{ 364, 255, 22 },
		{ 364, 256, 22 },
		{ 364, 257, 22 },
		{ 364, 258, 22 },
		{ 364, 259, 22 },
		{ 364, 260, 22 },
		{ 364, 261, 22 },
		{ 364, 262, 22 },
		{ 364, 263, 22 },
		{ 364, 264, 22 },
		{ 364, 265, 22 },
		{ 364, 266, 22 },
		{ 364, 267, 22 },
		{ 364, 268, 22 },
		{ 364, 269, 22 },
		{ 364, 270, 22 },
		{ 364, 271, 22 },
		{ 364, 272, 22 },
		{ 364, 273, 22 },
		{ 364, 274, 22 },
		{ 364, 275, 22 },
		{ 364, 276, 22 },
		{ 364, 277, 22 },
		{ 364, 278, 22 },
		{ 364, 279, 22 },
		{ 364, 280, 22 },
		{ 364, 281, 22 },
		{ 364, 282, 22 },
		{ 364, 283, 22 },
		{ 365, 231, 22 },
		{ 365, 232, 22 },
		{ 365, 233, 22 },
		{ 365, 234, 22 },
		{ 365, 235, 22 },
		{ 365, 236, 22 },
		{ 365, 237, 22 },
		{ 365, 238, 22 },
		{ 365, 239, 22 },
		{ 365, 240, 22 },
		{ 365, 241, 22 },
		{ 365, 242, 22 },
		{ 365, 243, 22 },
		{ 365, 244, 22 },
		{ 365, 245, 22 },
		{ 365, 246, 22 },
		{ 365, 247, 22 },
		{ 365, 248, 22 },
		{ 365, 249, 22 },
		{ 365, 250, 22 },
		{ 365, 251, 22 },
		{ 365, 252, 22 },
		{ 365, 253, 22 },
		{ 365, 254, 22 },
		{ 365, 255, 22 },
		{ 365, 256, 22 },
		{ 365, 257, 22 },
		{ 365, 258, 22 },
		{ 365, 259, 22 },
		{ 365, 260, 22 },
		{ 365, 261, 22 },
		{ 365, 262, 22 },
		{ 365, 263, 22 },
		{ 365, 264, 22 },
		{ 365, 265, 22 },
		{ 365, 266, 22 },
		{ 365, 267, 22 },
		{ 365, 268, 22 },
		{ 365, 269, 22 },
		{ 365, 270, 22 },
		{ 365, 271, 22 },
		{ 365, 272, 22 },
		{ 365, 273, 22 },
		{ 365, 274, 22 },
		{ 365, 275, 22 },
		{ 365, 276, 22 },
		{ 365, 277, 22 },
		{ 365, 278, 22 },
		{ 365, 279, 22 },
		{ 365, 280, 22 },
		{ 365, 281, 22 },
		{ 365, 282, 22 },
		{ 365, 283, 22 },
		{ 366, 230, 22 },
		{ 366, 231, 22 },
		{ 366, 232, 22 },
		{ 366, 233, 22 },
		{ 366, 234, 22 },
		{ 366, 235, 22 },
		{ 366, 236, 22 },
		{ 366, 237, 22 },
		{ 366, 238, 22 },
		{ 366, 239, 22 },
		{ 366, 240, 22 },
		{ 366, 241, 22 },
		{ 366, 242, 22 },
		{ 366, 243, 22 },
		{ 366, 244, 22 },
		{ 366, 245, 22 },
		{ 366, 246, 22 },
		{ 366, 247, 22 },
		{ 366, 248, 22 },
		{ 366, 249, 22 },
		{ 366, 250, 22 },
		{ 366, 251, 22 },
		{ 366, 252, 22 },
		{ 366, 253, 22 },
		{ 366, 254, 22 },
		{ 366, 255, 22 },
		{ 366, 256, 22 },
		{ 366, 257, 22 },
		{ 366, 258, 22 },
		{ 366, 259, 22 },
		{ 366, 260, 22 },
		{ 366, 261, 22 },
		{ 366, 262, 22 },
		{ 366, 263, 22 },
		{ 366, 264, 22 },
		{ 366, 265, 22 },
		{ 366, 266, 22 },
		{ 366, 267, 22 },
		{ 366, 268, 22 },
		{ 366, 269, 22 },
		{ 366, 270, 22 },
		{ 366, 271, 22 },
		{ 366, 272, 22 },
		{ 366, 273, 22 },
		{ 366, 274, 22 },
		{ 366, 275, 22 },
		{ 366, 276, 22 },
		{ 366, 277, 22 },
		{ 366, 278, 22 },
		{ 366, 279, 22 },
		{ 366, 280, 22 },
		{ 366, 281, 22 },
		{ 366, 282, 22 },
		{ 366, 283, 22 },
		{ 366, 284, 22 },
		{ 367, 230, 22 },
		{ 367, 231, 22 },
		{ 367, 232, 22 },
		{ 367, 233, 22 },
		{ 367, 234, 22 },
		{ 367, 235, 22 },
		{ 367, 236, 22 },
		{ 367, 237, 22 },
		{ 367, 238, 22 },
		{ 367, 239, 22 },
		{ 367, 240, 22 },
		{ 367, 241, 22 },
		{ 367, 242, 22 },
		{ 367, 243, 22 },
		{ 367, 244, 22 },
		{ 367, 245, 22 },
		{ 367, 246, 22 },
		{ 367, 247, 22 },
		{ 367, 248, 22 },
		{ 367, 249, 22 },
		{ 367, 250, 22 },
		{ 367, 251, 22 },
		{ 367, 252, 22 },
		{ 367, 253, 22 },
		{ 367, 254, 22 },
		{ 367, 255, 22 },
		{ 367, 256, 22 },
		{ 367, 257, 22 },
		{ 367, 258, 22 },
		{ 367, 259, 22 },
		{ 367, 260, 22 },
		{ 367, 261, 22 },
		{ 367, 262, 22 },
		{ 367, 263, 22 },
		{ 367, 264, 22 },
		{ 367, 265, 22 },
		{ 367, 266, 22 },
		{ 367, 267, 22 },
		{ 367, 268, 22 },
		{ 367, 269, 22 },
		{ 367, 270, 22 },
		{ 367, 271, 22 },
		{ 367, 272, 22 },
		{ 367, 273, 22 },
		{ 367, 274, 22 },
		{ 367, 275, 22 },
		{ 367, 276, 22 },
		{ 367, 277, 22 },
		{ 367, 278, 22 },
		{ 367, 279, 22 },
		{ 367, 280, 22 },
		{ 367, 281, 22 },
		{ 367, 282, 22 },
		{ 367, 283, 22 },
		{ 367, 284, 22 },
		{ 368, 230, 22 },
		{ 368, 231, 22 },
		{ 368, 232, 22 },
		{ 368, 233, 22 },
		{ 368, 234, 22 },
		{ 368, 235, 22 },
		{ 368, 236, 22 },
		{ 368, 237, 22 },
		{ 368, 238, 22 },
		{ 368, 239, 22 },
		{ 368, 240, 22 },
		{ 368, 241, 22 },
		{ 368, 242, 22 },
		{ 368, 243, 22 },
		{ 368, 244, 22 },
		{ 368, 245, 22 },
		{ 368, 246, 22 },
		{ 368, 247, 22 },
		{ 368, 248, 22 },
		{ 368, 249, 22 },
		{ 368, 250, 22 },
		{ 368, 251, 22 },
		{ 368, 252, 22 },
		{ 368, 253, 22 },
		{ 368, 254, 22 },
		{ 368, 255, 22 },
		{ 368, 256, 22 },
		{ 368, 257, 22 },
		{ 368, 258, 22 },
		{ 368, 259, 22 },
		{ 368, 260, 22 },
		{ 368, 261, 22 },
		{ 368, 262, 22 },
		{ 368, 263, 22 },
		{ 368, 264, 22 },
		{ 368, 265, 22 },
		{ 368, 266, 22 },
		{ 368, 267, 22 },
		{ 368, 268, 22 },
		{ 368, 269, 22 },
		{ 368, 270, 22 },
		{ 368, 271, 22 },
		{ 368, 272, 22 },
		{ 368, 273, 22 },
		{ 368, 274, 22 },
		{ 368, 275, 22 },
		{ 368, 276, 22 },
		{ 368, 277, 22 },
		{ 368, 278, 22 },
		{ 368, 279, 22 },
		{ 368, 280, 22 },
		{ 368, 281, 22 },
		{ 368, 282, 22 },
		{ 368, 283, 22 },
		{ 368, 284, 22 },
		{ 369, 230, 22 },
		{ 369, 231, 22 },
		{ 369, 232, 22 },
		{ 369, 233, 22 },
		{ 369, 234, 22 },
		{ 369, 235, 22 },
		{ 369, 236, 22 },
		{ 369, 237, 22 },
		{ 369, 238, 22 },
		{ 369, 239, 22 },
		{ 369, 240, 22 },
		{ 369, 241, 22 },
		{ 369, 242, 22 },
		{ 369, 243, 22 },
		{ 369, 244, 22 },
		{ 369, 245, 22 },
		{ 369, 246, 22 },
		{ 369, 247, 22 },
		{ 369, 248, 22 },
		{ 369, 249, 22 },
		{ 369, 250, 22 },
		{ 369, 251, 22 },
		{ 369, 252, 22 },
		{ 369, 253, 22 },
		{ 369, 254, 22 },
		{ 369, 255, 22 },
		{ 369, 256, 22 },
		{ 369, 257, 22 },
		{ 369, 258, 22 },
		{ 369, 259, 22 },
		{ 369, 260, 22 },
		{ 369, 261, 22 },
		{ 369, 262, 22 },
		{ 369, 263, 22 },
		{ 369, 264, 22 },
		{ 369, 265, 22 },
		{ 369, 266, 22 },
		{ 369, 267, 22 },
		{ 369, 268, 22 },
		{ 369, 269, 22 },
		{ 369, 270, 22 },
		{ 369, 271, 22 },
		{ 369, 272, 22 },
		{ 369, 273, 22 },
		{ 369, 274, 22 },
		{ 369, 275, 22 },
		{ 369, 276, 22 },
		{ 369, 277, 22 },
		{ 369, 278, 22 },
		{ 369, 279, 22 },
		{ 369, 280, 22 },
		{ 369, 281, 22 },
		{ 369, 282, 22 },
		{ 369, 283, 22 },
		{ 369, 284, 22 },
		{ 370, 229, 22 },
		{ 370, 230, 22 },
		{ 370, 231, 22 },
		{ 370, 232, 22 },
		{ 370, 233, 22 },
		{ 370, 234, 22 },
		{ 370, 235, 22 },
		{ 370, 236, 22 },
		{ 370, 237, 22 },
		{ 370, 238, 22 },
		{ 370, 239, 22 },
		{ 370, 240, 22 },
		{ 370, 241, 22 },
		{ 370, 242, 22 },
		{ 370, 243, 22 },
		{ 370, 244, 22 },
		{ 370, 245, 22 },
		{ 370, 246, 22 },
		{ 370, 247, 22 },
		{ 370, 248, 22 },
		{ 370, 249, 22 },
		{ 370, 250, 22 },
		{ 370, 251, 22 },
		{ 370, 252, 22 },
		{ 370, 253, 22 },
		{ 370, 254, 22 },
		{ 370, 255, 22 },
		{ 370, 256, 22 },
		{ 370, 257, 22 },
		{ 370, 258, 22 },
		{ 370, 259, 22 },
		{ 370, 260, 22 },
		{ 370, 261, 22 },
		{ 370, 262, 22 },
		{ 370, 263, 22 },
		{ 370, 264, 22 },
		{ 370, 265, 22 },
		{ 370, 266, 22 },
		{ 370, 267, 22 },
		{ 370, 268, 22 },
		{ 370, 269, 22 },
		{ 370, 270, 22 },
		{ 370, 271, 22 },
		{ 370, 272, 22 },
		{ 370, 273, 22 },
		{ 370, 274, 22 },
		{ 370, 275, 22 },
		{ 370, 276, 22 },
		{ 370, 277, 22 },
		{ 370, 278, 22 },
		{ 370, 279, 22 },
		{ 370, 280, 22 },
		{ 370, 281, 22 },
		{ 370, 282, 22 },
		{ 370, 283, 22 },
		{ 370, 284, 22 },
		{ 370, 285, 22 },
		{ 371, 229, 22 },
		{ 371, 230, 22 },
		{ 371, 231, 22 },
		{ 371, 232, 22 },
		{ 371, 233, 22 },
		{ 371, 234, 22 },
		{ 371, 235, 22 },
		{ 371, 236, 22 },
		{ 371, 237, 22 },
		{ 371, 238, 22 },
		{ 371, 239, 22 },
		{ 371, 240, 22 },
		{ 371, 241, 22 },
		{ 371, 242, 22 },
		{ 371, 243, 22 },
		{ 371, 244, 22 },
		{ 371, 245, 22 },
		{ 371, 246, 22 },
		{ 371, 247, 22 },
		{ 371, 248, 22 },
		{ 371, 249, 22 },
		{ 371, 250, 22 },
		{ 371, 251, 22 },
		{ 371, 252, 22 },
		{ 371, 253, 22 },
		{ 371, 254, 22 },
		{ 371, 255, 22 },
		{ 371, 256, 22 },
		{ 371, 257, 22 },
		{ 371, 258, 22 },
		{ 371, 259, 22 },
		{ 371, 260, 22 },
		{ 371, 261, 22 },
		{ 371, 262, 22 },
		{ 371, 263, 22 },
		{ 371, 264, 22 },
		{ 371, 265, 22 },
		{ 371, 266, 22 },
		{ 371, 267, 22 },
		{ 371, 268, 22 },
		{ 371, 269, 22 },
		{ 371, 270, 22 },
		{ 371, 271, 22 },
		{ 371, 272, 22 },
		{ 371, 273, 22 },
		{ 371, 274, 22 },
		{ 371, 275, 22 },
		{ 371, 276, 22 },
		{ 371, 277, 22 },
		{ 371, 278, 22 },
		{ 371, 279, 22 },
		{ 371, 280, 22 },
		{ 371, 281, 22 },
		{ 371, 282, 22 },
		{ 371, 283, 22 },
		{ 371, 284, 22 },
		{ 371, 285, 22 },
		{ 372, 229, 22 },
		{ 372, 230, 22 },
		{ 372, 231, 22 },
		{ 372, 232, 22 },
		{ 372, 233, 22 },
		{ 372, 234, 22 },
		{ 372, 235, 22 },
		{ 372, 236, 22 },
		{ 372, 237, 22 },
		{ 372, 238, 22 },
		{ 372, 239, 22 },
		{ 372, 240, 22 },
		{ 372, 241, 22 },
		{ 372, 242, 22 },
		{ 372, 243, 22 },
		{ 372, 244, 22 },
		{ 372, 245, 22 },
		{ 372, 246, 22 },
		{ 372, 247, 22 },
		{ 372, 248, 22 },
		{ 372, 249, 22 },
		{ 372, 250, 22 },
		{ 372, 251, 22 },
		{ 372, 252, 22 },
		{ 372, 253, 22 },
		{ 372, 254, 22 },
		{ 372, 255, 22 },
		{ 372, 256, 22 },
		{ 372, 257, 22 },
		{ 372, 258, 22 },
		{ 372, 259, 22 },
		{ 372, 260, 22 },
		{ 372, 261, 22 },
		{ 372, 262, 22 },
		{ 372, 263, 22 },
		{ 372, 264, 22 },
		{ 372, 265, 22 },
		{ 372, 266, 22 },
		{ 372, 267, 22 },
		{ 372, 268, 22 },
		{ 372, 269, 22 },
		{ 372, 270, 22 },
		{ 372, 271, 22 },
		{ 372, 272, 22 },
		{ 372, 273, 22 },
		{ 372, 274, 22 },
		{ 372, 275, 22 },
		{ 372, 276, 22 },
		{ 372, 277, 22 },
		{ 372, 278, 22 },
		{ 372, 279, 22 },
		{ 372, 280, 22 },
		{ 372, 281, 22 },
		{ 372, 282, 22 },
		{ 372, 283, 22 },
		{ 372, 284, 22 },
		{ 372, 285, 22 },
		{ 373, 229, 22 },
		{ 373, 230, 22 },
		{ 373, 231, 22 },
		{ 373, 232, 22 },
		{ 373, 233, 22 },
		{ 373, 234, 22 },
		{ 373, 235, 22 },
		{ 373, 236, 22 },
		{ 373, 237, 22 },
		{ 373, 238, 22 },
		{ 373, 239, 22 },
		{ 373, 240, 22 },
		{ 373, 241, 22 },
		{ 373, 242, 22 },
		{ 373, 243, 22 },
		{ 373, 244, 22 },
		{ 373, 245, 22 },
		{ 373, 246, 22 },
		{ 373, 247, 22 },
		{ 373, 248, 22 },
		{ 373, 249, 22 },
		{ 373, 250, 22 },
		{ 373, 251, 22 },
		{ 373, 252, 22 },
		{ 373, 253, 22 },
		{ 373, 254, 22 },
		{ 373, 255, 22 },
		{ 373, 256, 22 },
		{ 373, 257, 22 },
		{ 373, 258, 22 },
		{ 373, 259, 22 },
		{ 373, 260, 22 },
		{ 373, 261, 22 },
		{ 373, 262, 22 },
		{ 373, 263, 22 },
		{ 373, 264, 22 },
		{ 373, 265, 22 },
		{ 373, 266, 22 },
		{ 373, 267, 22 },
		{ 373, 268, 22 },
		{ 373, 269, 22 },
		{ 373, 270, 22 },
		{ 373, 271, 22 },
		{ 373, 272, 22 },
		{ 373, 273, 22 },
		{ 373, 274, 22 },
		{ 373, 275, 22 },
		{ 373, 276, 22 },
		{ 373, 277, 22 },
		{ 373, 278, 22 },
		{ 373, 279, 22 },
		{ 373, 280, 22 },
		{ 373, 281, 22 },
		{ 373, 282, 22 },
		{ 373, 283, 22 },
		{ 373, 284, 22 },
		{ 373, 285, 22 },
		{ 374, 228, 22 },
		{ 374, 229, 22 },
		{ 374, 230, 22 },
		{ 374, 231, 22 },
		{ 374, 232, 22 },
		{ 374, 233, 22 },
		{ 374, 234, 22 },
		{ 374, 235, 22 },
		{ 374, 236, 22 },
		{ 374, 237, 22 },
		{ 374, 238, 22 },
		{ 374, 239, 22 },
		{ 374, 240, 22 },
		{ 374, 241, 22 },
		{ 374, 242, 22 },
		{ 374, 243, 22 },
		{ 374, 244, 22 },
		{ 374, 245, 22 },
		{ 374, 246, 22 },
		{ 374, 247, 22 },
		{ 374, 248, 22 },
		{ 374, 249, 22 },
		{ 374, 250, 22 },
		{ 374, 251, 22 },
		{ 374, 252, 22 },
		{ 374, 253, 22 },
		{ 374, 254, 22 },
		{ 374, 255, 22 },
		{ 374, 256, 22 },
		{ 374, 257, 22 },
		{ 374, 258, 22 },
		{ 374, 259, 22 },
		{ 374, 260, 22 },
		{ 374, 261, 22 },
		{ 374, 262, 22 },
		{ 374, 263, 22 },
		{ 374, 264, 22 },
		{ 374, 265, 22 },
		{ 374, 266, 22 },
		{ 374, 267, 22 },
		{ 374, 268, 22 },
		{ 374, 269, 22 },
		{ 374, 270, 22 },
		{ 374, 271, 22 },
		{ 374, 272, 22 },
		{ 374, 273, 22 },
		{ 374, 274, 22 },
		{ 374, 275, 22 },
		{ 374, 276, 22 },
		{ 374, 277, 22 },
		{ 374, 278, 22 },
		{ 374, 279, 22 },
		{ 374, 280, 22 },
		{ 374, 281, 22 },
		{ 374, 282, 22 },
		{ 374, 283, 22 },
		{ 374, 284, 22 },
		{ 374, 285, 22 },
		{ 374, 286, 22 },
		{ 375, 228, 22 },
		{ 375, 229, 22 },
		{ 375, 230, 22 },
		{ 375, 231, 22 },
		{ 375, 232, 22 },
		{ 375, 233, 22 },
		{ 375, 234, 22 },
		{ 375, 235, 22 },
		{ 375, 236, 22 },
		{ 375, 237, 22 },
		{ 375, 238, 22 },
		{ 375, 239, 22 },
		{ 375, 240, 22 },
		{ 375, 241, 22 },
		{ 375, 242, 22 },
		{ 375, 243, 22 },
		{ 375, 244, 22 },
		{ 375, 245, 22 },
		{ 375, 246, 22 },
		{ 375, 247, 22 },
		{ 375, 248, 22 },
		{ 375, 249, 22 },
		{ 375, 250, 22 },
		{ 375, 251, 22 },
		{ 375, 252, 22 },
		{ 375, 253, 22 },
		{ 375, 254, 22 },
		{ 375, 255, 22 },
		{ 375, 256, 22 },
		{ 375, 257, 22 },
		{ 375, 258, 22 },
		{ 375, 259, 22 },
		{ 375, 260, 22 },
		{ 375, 261, 22 },
		{ 375, 262, 22 },
		{ 375, 263, 22 },
		{ 375, 264, 22 },
		{ 375, 265, 22 },
		{ 375, 266, 22 },
		{ 375, 267, 22 },
		{ 375, 268, 22 },
		{ 375, 269, 22 },
		{ 375, 270, 22 },
		{ 375, 271, 22 },
		{ 375, 272, 22 },
		{ 375, 273, 22 },
		{ 375, 274, 22 },
		{ 375, 275, 22 },
		{ 375, 276, 22 },
		{ 375, 277, 22 },
		{ 375, 278, 22 },
		{ 375, 279, 22 },
		{ 375, 280, 22 },
		{ 375, 281, 22 },
		{ 375, 282, 22 },
		{ 375, 283, 22 },
		{ 375, 284, 22 },
		{ 375, 285, 22 },
		{ 375, 286, 22 },
		{ 376, 228, 22 },
		{ 376, 229, 22 },
		{ 376, 230, 22 },
		{ 376, 231, 22 },
		{ 376, 232, 22 },
		{ 376, 233, 22 },
		{ 376, 234, 22 },
		{ 376, 235, 22 },
		{ 376, 236, 22 },
		{ 376, 237, 22 },
		{ 376, 238, 22 },
		{ 376, 239, 22 },
		{ 376, 240, 22 },
		{ 376, 241, 22 },
		{ 376, 242, 22 },
		{ 376, 243, 22 },
		{ 376, 244, 22 },
		{ 376, 245, 22 },
		{ 376, 246, 22 },
		{ 376, 247, 22 },
		{ 376, 248, 22 },
		{ 376, 249, 22 },
		{ 376, 250, 22 },
		{ 376, 251, 22 },
		{ 376, 252, 22 },
		{ 376, 253, 22 },
		{ 376, 254, 22 },
		{ 376, 255, 22 },
		{ 376, 256, 22 },
		{ 376, 257, 22 },
		{ 376, 258, 22 },
		{ 376, 259, 22 },
		{ 376, 260, 22 },
		{ 376, 261, 22 },
		{ 376, 262, 22 },
		{ 376, 263, 22 },
		{ 376, 264, 22 },
		{ 376, 265, 22 },
		{ 376, 266, 22 },
		{ 376, 267, 22 },
		{ 376, 268, 22 },
		{ 376, 269, 22 },
		{ 376, 270, 22 },
		{ 376, 271, 22 },
		{ 376, 272, 22 },
		{ 376, 273, 22 },
		{ 376, 274, 22 },
		{ 376, 275, 22 },
		{ 376, 276, 22 },
		{ 376, 277, 22 },
		{ 376, 278, 22 },
		{ 376, 279, 22 },
		{ 376, 280, 22 },
		{ 376, 281, 22 },
		{ 376, 282, 22 },
		{ 376, 283, 22 },
		{ 376, 284, 22 },
		{ 376, 285, 22 },
		{ 376, 286, 22 },
		{ 377, 228, 22 },
		{ 377, 229, 22 },
		{ 377, 230, 22 },
		{ 377, 231, 22 },
		{ 377, 232, 22 },
		{ 377, 233, 22 },
		{ 377, 234, 22 },
		{ 377, 235, 22 },
		{ 377, 236, 22 },
		{ 377, 237, 22 },
		{ 377, 238, 22 },
		{ 377, 239, 22 },
		{ 377, 240, 22 },
		{ 377, 241, 22 },
		{ 377, 242, 22 },
		{ 377, 243, 22 },
		{ 377, 244, 22 },
		{ 377, 245, 22 },
		{ 377, 246, 22 },
		{ 377, 247, 22 },
		{ 377, 248, 22 },
		{ 377, 249, 22 },
		{ 377, 250, 22 },
		{ 377, 251, 22 },
		{ 377, 252, 22 },
		{ 377, 253, 22 },
		{ 377, 254, 22 },
		{ 377, 255, 22 },
		{ 377, 256, 22 },
		{ 377, 257, 22 },
		{ 377, 258, 22 },
		{ 377, 259, 22 },
		{ 377, 260, 22 },
		{ 377, 261, 22 },
		{ 377, 262, 22 },
		{ 377, 263, 22 },
		{ 377, 264, 22 },
		{ 377, 265, 22 },
		{ 377, 266, 22 },
		{ 377, 267, 22 },
		{ 377, 268, 22 },
		{ 377, 269, 22 },
		{ 377, 270, 22 },
		{ 377, 271, 22 },
		{ 377, 272, 22 },
		{ 377, 273, 22 },
		{ 377, 274, 22 },
		{ 377, 275, 22 },
		{ 377, 276, 22 },
		{ 377, 277, 22 },
		{ 377, 278, 22 },
		{ 377, 279, 22 },
		{ 377, 280, 22 },
		{ 377, 281, 22 },
		{ 377, 282, 22 },
		{ 377, 283, 22 },
		{ 377, 284, 22 },
		{ 377, 285, 22 },
		{ 377, 286, 22 },
		{ 378, 228, 22 },
		{ 378, 229, 22 },
		{ 378, 230, 22 },
		{ 378, 231, 22 },
		{ 378, 232, 22 },
		{ 378, 233, 22 },
		{ 378, 234, 22 },
		{ 378, 235, 22 },
		{ 378, 236, 22 },
		{ 378, 237, 22 },
		{ 378, 238, 22 },
		{ 378, 239, 22 },
		{ 378, 240, 22 },
		{ 378, 241, 22 },
		{ 378, 242, 22 },
		{ 378, 243, 22 },
		{ 378, 244, 22 },
		{ 378, 245, 22 },
		{ 378, 246, 22 },
		{ 378, 247, 22 },
		{ 378, 248, 22 },
		{ 378, 249, 22 },
		{ 378, 250, 22 },
		{ 378, 251, 22 },
		{ 378, 252, 22 },
		{ 378, 253, 22 },
		{ 378, 254, 22 },
		{ 378, 255, 22 },
		{ 378, 256, 22 },
		{ 378, 257, 22 },
		{ 378, 258, 22 },
		{ 378, 259, 22 },
		{ 378, 260, 22 },
		{ 378, 261, 22 },
		{ 378, 262, 22 },
		{ 378, 263, 22 },
		{ 378, 264, 22 },
		{ 378, 265, 22 },
		{ 378, 266, 22 },
		{ 378, 267, 22 },
		{ 378, 268, 22 },
		{ 378, 269, 22 },
		{ 378, 270, 22 },
		{ 378, 271, 22 },
		{ 378, 272, 22 },
		{ 378, 273, 22 },
		{ 378, 274, 22 },
		{ 378, 275, 22 },
		{ 378, 276, 22 },
		{ 378, 277, 22 },
		{ 378, 278, 22 },
		{ 378, 279, 22 },
		{ 378, 280, 22 },
		{ 378, 281, 22 },
		{ 378, 282, 22 },
		{ 378, 283, 22 },
		{ 378, 284, 22 },
		{ 378, 285, 22 },
		{ 378, 286, 22 },
		{ 379, 228, 22 },
		{ 379, 229, 22 },
		{ 379, 230, 22 },
		{ 379, 231, 22 },
		{ 379, 232, 22 },
		{ 379, 233, 22 },
		{ 379, 234, 22 },
		{ 379, 235, 22 },
		{ 379, 236, 22 },
		{ 379, 237, 22 },
		{ 379, 238, 22 },
		{ 379, 239, 22 },
		{ 379, 240, 22 },
		{ 379, 241, 22 },
		{ 379, 242, 22 },
		{ 379, 243, 22 },
		{ 379, 244, 22 },
		{ 379, 245, 22 },
		{ 379, 246, 22 },
		{ 379, 247, 22 },
		{ 379, 248, 22 },
		{ 379, 249, 22 },
		{ 379, 250, 22 },
		{ 379, 251, 22 },
		{ 379, 252, 22 },
		{ 379, 253, 22 },
		{ 379, 254, 22 },
		{ 379, 255, 22 },
		{ 379, 256, 22 },
		{ 379, 257, 22 },
		{ 379, 258, 22 },
		{ 379, 259, 22 },
		{ 379, 260, 22 },
		{ 379, 261, 22 },
		{ 379, 262, 22 },
		{ 379, 263, 22 },
		{ 379, 264, 22 },
		{ 379, 265, 22 },
		{ 379, 266, 22 },
		{ 379, 267, 22 },
		{ 379, 268, 22 },
		{ 379, 269, 22 },
		{ 379, 270, 22 },
		{ 379, 271, 22 },
		{ 379, 272, 22 },
		{ 379, 273, 22 },
		{ 379, 274, 22 },
		{ 379, 275, 22 },
		{ 379, 276, 22 },
		{ 379, 277, 22 },
		{ 379, 278, 22 },
		{ 379, 279, 22 },
		{ 379, 280, 22 },
		{ 379, 281, 22 },
		{ 379, 282, 22 },
		{ 379, 283, 22 },
		{ 379, 284, 22 },
		{ 379, 285, 22 },
		{ 379, 286, 22 },
		{ 380, 228, 22 },
		{ 380, 229, 22 },
		{ 380, 230, 22 },
		{ 380, 231, 22 },
		{ 380, 232, 22 },
		{ 380, 233, 22 },
		{ 380, 234, 22 },
		{ 380, 235, 22 },
		{ 380, 236, 22 },
		{ 380, 237, 22 },
		{ 380, 238, 22 },
		{ 380, 239, 22 },
		{ 380, 240, 22 },
		{ 380, 241, 22 },
		{ 380, 242, 22 },
		{ 380, 243, 22 },
		{ 380, 244, 22 },
		{ 380, 245, 22 },
		{ 380, 246, 22 },
		{ 380, 247, 22 },
		{ 380, 248, 22 },
		{ 380, 249, 22 },
		{ 380, 250, 22 },
		{ 380, 251, 22 },
		{ 380, 252, 22 },
		{ 380, 253, 22 },
		{ 380, 254, 22 },
		{ 380, 255, 22 },
		{ 380, 256, 22 },
		{ 380, 257, 22 },
		{ 380, 258, 22 },
		{ 380, 259, 22 },
		{ 380, 260, 22 },
		{ 380, 261, 22 },
		{ 380, 262, 22 },
		{ 380, 263, 22 },
		{ 380, 264, 22 },
		{ 380, 265, 22 },
		{ 380, 266, 22 },
		{ 380, 267, 22 },
		{ 380, 268, 22 },
		{ 380, 269, 22 },
		{ 380, 270, 22 },
		{ 380, 271, 22 },
		{ 380, 272, 22 },
		{ 380, 273, 22 },
		{ 380, 274, 22 },
		{ 380, 275, 22 },
		{ 380, 276, 22 },
		{ 380, 277, 22 },
		{ 380, 278, 22 },
		{ 380, 279, 22 },
		{ 380, 280, 22 },
		{ 380, 281, 22 },
		{ 380, 282, 22 },
		{ 380, 283, 22 },
		{ 380, 284, 22 },
		{ 380, 285, 22 },
		{ 380, 286, 22 },
		{ 381, 228, 22 },
		{ 381, 229, 22 },
		{ 381, 230, 22 },
		{ 381, 231, 22 },
		{ 381, 232, 22 },
		{ 381, 233, 22 },
		{ 381, 234, 22 },
		{ 381, 235, 22 },
		{ 381, 236, 22 },
		{ 381, 237, 22 },
		{ 381, 238, 22 },
		{ 381, 239, 22 },
		{ 381, 240, 22 },
		{ 381, 241, 22 },
		{ 381, 242, 22 },
		{ 381, 243, 22 },
		{ 381, 244, 22 },
		{ 381, 245, 22 },
		{ 381, 246, 22 },
		{ 381, 247, 22 },
		{ 381, 248, 22 },
		{ 381, 249, 22 },
		{ 381, 250, 22 },
		{ 381, 251, 22 },
		{ 381, 252, 22 },
		{ 381, 253, 22 },
		{ 381, 254, 22 },
		{ 381, 255, 22 },
		{ 381, 256, 22 },
		{ 381, 257, 22 },
		{ 381, 258, 22 },
		{ 381, 259, 22 },
		{ 381, 260, 22 },
		{ 381, 261, 22 },
		{ 381, 262, 22 },
		{ 381, 263, 22 },
		{ 381, 264, 22 },
		{ 381, 265, 22 },
		{ 381, 266, 22 },
		{ 381, 267, 22 },
		{ 381, 268, 22 },
		{ 381, 269, 22 },
		{ 381, 270, 22 },
		{ 381, 271, 22 },
		{ 381, 272, 22 },
		{ 381, 273, 22 },
		{ 381, 274, 22 },
		{ 381, 275, 22 },
		{ 381, 276, 22 },
		{ 381, 277, 22 },
		{ 381, 278, 22 },
		{ 381, 279, 22 },
		{ 381, 280, 22 },
		{ 381, 281, 22 },
		{ 381, 282, 22 },
		{ 381, 283, 22 },
		{ 381, 284, 22 },
		{ 381, 285, 22 },
		{ 381, 286, 22 },
		{ 382, 229, 22 },
		{ 382, 230, 22 },
		{ 382, 231, 22 },
		{ 382, 232, 22 },
		{ 382, 233, 22 },
		{ 382, 234, 22 },
		{ 382, 235, 22 },
		{ 382, 236, 22 },
		{ 382, 237, 22 },
		{ 382, 238, 22 },
		{ 382, 239, 22 },
		{ 382, 240, 22 },
		{ 382, 241, 22 },
		{ 382, 242, 22 },
		{ 382, 243, 22 },
		{ 382, 244, 22 },
		{ 382, 245, 22 },
		{ 382, 246, 22 },
		{ 382, 247, 22 },
		{ 382, 248, 22 },
		{ 382, 249, 22 },
		{ 382, 250, 22 },
		{ 382, 251, 22 },
		{ 382, 252, 22 },
		{ 382, 253, 22 },
		{ 382, 254, 22 },
		{ 382, 255, 22 },
		{ 382, 256, 22 },
		{ 382, 257, 22 },
		{ 382, 258, 22 },
		{ 382, 259, 22 },
		{ 382, 260, 22 },
		{ 382, 261, 22 },
		{ 382, 262, 22 },
		{ 382, 263, 22 },
		{ 382, 264, 22 },
		{ 382, 265, 22 },
		{ 382, 266, 22 },
		{ 382, 267, 22 },
		{ 382, 268, 22 },
		{ 382, 269, 22 },
		{ 382, 270, 22 },
		{ 382, 271, 22 },
		{ 382, 272, 22 },
		{ 382, 273, 22 },
		{ 382, 274, 22 },
		{ 382, 275, 22 },
		{ 382, 276, 22 },
		{ 382, 277, 22 },
		{ 382, 278, 22 },
		{ 382, 279, 22 },
		{ 382, 280, 22 },
		{ 382, 281, 22 },
		{ 382, 282, 22 },
		{ 382, 283, 22 },
		{ 382, 284, 22 },
		{ 382, 285, 22 },
		{ 383, 233, 22 },
		{ 383, 234, 22 },
		{ 383, 235, 22 },
		{ 383, 236, 22 },
		{ 383, 237, 22 },
		{ 383, 238, 22 },
		{ 383, 239, 22 },
		{ 383, 240, 22 },
		{ 383, 241, 22 },
		{ 383, 242, 22 },
		{ 383, 243, 22 },
		{ 383, 244, 22 },
		{ 383, 245, 22 },
		{ 383, 246, 22 },
		{ 383, 247, 22 },
		{ 383, 248, 22 },
		{ 383, 249, 22 },
		{ 383, 250, 22 },
		{ 383, 251, 22 },
		{ 383, 252, 22 },
		{ 383, 253, 22 },
		{ 383, 254, 22 },
		{ 383, 255, 22 },
		{ 383, 256, 22 },
		{ 383, 257, 22 },
		{ 383, 258, 22 },
		{ 383, 259, 22 },
		{ 383, 260, 22 },
		{ 383, 261, 22 },
		{ 383, 262, 22 },
		{ 383, 263, 22 },
		{ 383, 264, 22 },
		{ 383, 265, 22 },
		{ 383, 266, 22 },
		{ 383, 267, 22 },
		{ 383, 268, 22 },
		{ 383, 269, 22 },
		{ 383, 270, 22 },
		{ 383, 271, 22 },
		{ 383, 272, 22 },
		{ 383, 273, 22 },
		{ 383, 274, 22 },
		{ 383, 275, 22 },
		{ 383, 276, 22 },
		{ 383, 277, 22 },
		{ 383, 278, 22 },
		{ 383, 279, 22 },
		{ 383, 280, 22 },
		{ 383, 281, 22 },
		{ 131, 201, 23 },
		{ 131, 202, 23 },
		{ 131, 203, 23 },
		{ 131, 204, 23 },
		{ 131, 205, 23 },
		{ 131, 206, 23 },
		{ 131, 207, 23 },
		{ 131, 208, 23 },
		{ 131, 209, 23 },
		{ 131, 210, 23 },
		{ 131, 211, 23 },
		{ 131, 212, 23 },
		{ 131, 213, 23 },
		{ 131, 214, 23 },
		{ 131, 215, 23 },
		{ 131, 216, 23 },
		{ 131, 217, 23 },
		{ 131, 218, 23 },
		{ 131, 219, 23 },
		{ 131, 220, 23 },
		{ 131, 221, 23 },
		{ 131, 222, 23 },
		{ 131, 223, 23 },
		{ 131, 224, 23 },
		{ 131, 225, 23 },
		{ 131, 226, 23 },
		{ 131, 227, 23 },
		{ 131, 228, 23 },
		{ 131, 229, 23 },
		{ 131, 230, 23 },
		{ 131, 231, 23 },
		{ 131, 232, 23 },
		{ 131, 233, 23 },
		{ 131, 234, 23 },
		{ 131, 235, 23 },
		{ 131, 236, 23 },
		{ 131, 237, 23 },
		{ 131, 238, 23 },
		{ 131, 239, 23 },
		{ 131, 240, 23 },
		{ 131, 241, 23 },
		{ 131, 242, 23 },
		{ 131, 243, 23 },
		{ 131, 244, 23 },
		{ 131, 245, 23 },
		{ 131, 246, 23 },
		{ 131, 247, 23 },
		{ 131, 248, 23 },
		{ 131, 249, 23 },
		{ 131, 250, 23 },
		{ 132, 198, 23 },
		{ 132, 199, 23 },
		{ 132, 200, 23 },
		{ 132, 201, 23 },
		{ 132, 202, 23 },
		{ 132, 203, 23 },
		{ 132, 204, 23 },
		{ 132, 205, 23 },
		{ 132, 206, 23 },
		{ 132, 207, 23 },
		{ 132, 208, 23 },
		{ 132, 209, 23 },
		{ 132, 210, 23 },
		{ 132, 211, 23 },
		{ 132, 212, 23 },
		{ 132, 213, 23 },
		{ 132, 214, 23 },
		{ 132, 215, 23 },
		{ 132, 216, 23 },
		{ 132, 217, 23 },
		{ 132, 218, 23 },
		{ 132, 219, 23 },
		{ 132, 220, 23 },
		{ 132, 221, 23 },
		{ 132, 222, 23 },
		{ 132, 223, 23 },
		{ 132, 224, 23 },
		{ 132, 225, 23 },
		{ 132, 226, 23 },
		{ 132, 227, 23 },
		{ 132, 228, 23 },
		{ 132, 229, 23 },
		{ 132, 230, 23 },
		{ 132, 231, 23 },
		{ 132, 232, 23 },
		{ 132, 233, 23 },
		{ 132, 234, 23 },
		{ 132, 235, 23 },
		{ 132, 236, 23 },
		{ 132, 237, 23 },
		{ 132, 238, 23 },
		{ 132, 239, 23 },
		{ 132, 240, 23 },
		{ 132, 241, 23 },
		{ 132, 242, 23 },
		{ 132, 243, 23 },
		{ 132, 244, 23 },
		{ 132, 245, 23 },
		{ 132, 246, 23 },
		{ 132, 247, 23 },
		{ 132, 248, 23 },
		{ 132, 249, 23 },
		{ 132, 250, 23 },
		{ 132, 251, 23 },
		{ 132, 252, 23 },
		{ 132, 253, 23 },
		{ 132, 254, 23 },
		{ 133, 197, 23 },
		{ 133, 198, 23 },
		{ 133, 199, 23 },
		{ 133, 200, 23 },
		{ 133, 201, 23 },
		{ 133, 202, 23 },
		{ 133, 203, 23 },
		{ 133, 204, 23 },
		{ 133, 205, 23 },
		{ 133, 206, 23 },
		{ 133, 207, 23 },
		{ 133, 208, 23 },
		{ 133, 209, 23 },
		{ 133, 210, 23 },
		{ 133, 211, 23 },
		{ 133, 212, 23 },
		{ 133, 213, 23 },
		{ 133, 214, 23 },
		{ 133, 215, 23 },
		{ 133, 216, 23 },
		{ 133, 217, 23 },
		{ 133, 218, 23 },
		{ 133, 219, 23 },
		{ 133, 220, 23 },
		{ 133, 221, 23 },
		{ 133, 222, 23 },
		{ 133, 223, 23 },
		{ 133, 224, 23 },
		{ 133, 225, 23 },
		{ 133, 226, 23 },
		{ 133, 227, 23 },
		{ 133, 228, 23 },
		{ 133, 229, 23 },
		{ 133, 230, 23 },
		{ 133, 231, 23 },
		{ 133, 232, 23 },
		{ 133, 233, 23 },
		{ 133, 234, 23 },
		{ 133, 235, 23 },
		{ 133, 236, 23 },
		{ 133, 237, 23 },
		{ 133, 238, 23 },
		{ 133, 239, 23 },
		{ 133, 240, 23 },
		{ 133, 241, 23 },
		{ 133, 242, 23 },
		{ 133, 243, 23 },
		{ 133, 244, 23 },
		{ 133, 245, 23 },
		{ 133, 246, 23 },
		{ 133, 247, 23 },
		{ 133, 248, 23 },
		{ 133, 249, 23 },
		{ 133, 250, 23 },
		{ 133, 251, 23 },
		{ 133, 252, 23 },
		{ 133, 253, 23 },
		{ 133, 254, 23 },
		{ 133, 255, 23 },
		{ 134, 197, 23 },
		{ 134, 198, 23 },
		{ 134, 199, 23 },
		{ 134, 200, 23 },
		{ 134, 201, 23 },
		{ 134, 202, 23 },
		{ 134, 203, 23 },
		{ 134, 204, 23 },
		{ 134, 205, 23 },
		{ 134, 206, 23 },
		{ 134, 207, 23 },
		{ 134, 208, 23 },
		{ 134, 209, 23 },
		{ 134, 210, 23 },
		{ 134, 211, 23 },
		{ 134, 212, 23 },
		{ 134, 213, 23 },
		{ 134, 214, 23 },
		{ 134, 215, 23 },
		{ 134, 216, 23 },
		{ 134, 217, 23 },
		{ 134, 218, 23 },
		{ 134, 219, 23 },
		{ 134, 220, 23 },
		{ 134, 221, 23 },
		{ 134, 222, 23 },
		{ 134, 223, 23 },
		{ 134, 224, 23 },
		{ 134, 225, 23 },
		{ 134, 226, 23 },
		{ 134, 227, 23 },
		{ 134, 228, 23 },
		{ 134, 229, 23 },
		{ 134, 230, 23 },
		{ 134, 231, 23 },
		{ 134, 232, 23 },
		{ 134, 233, 23 },
		{ 134, 234, 23 },
		{ 134, 235, 23 },
		{ 134, 236, 23 },
		{ 134, 237, 23 },
		{ 134, 238, 23 },
		{ 134, 239, 23 },
		{ 134, 240, 23 },
		{ 134, 241, 23 },
		{ 134, 242, 23 },
		{ 134, 243, 23 },
		{ 134, 244, 23 },
		{ 134, 245, 23 },
		{ 134, 246, 23 },
		{ 134, 247, 23 },
		{ 134, 248, 23 },
		{ 134, 249, 23 },
		{ 134, 250, 23 },
		{ 134, 251, 23 },
		{ 134, 252, 23 },
		{ 134, 253, 23 },
		{ 134, 254, 23 },
		{ 134, 255, 23 },
		{ 134, 256, 23 },
		{ 135, 197, 23 },
		{ 135, 198, 23 },
		{ 135, 199, 23 },
		{ 135, 200, 23 },
		{ 135, 201, 23 },
		{ 135, 202, 23 },
		{ 135, 203, 23 },
		{ 135, 204, 23 },
		{ 135, 205, 23 },
		{ 135, 206, 23 },
		{ 135, 207, 23 },
		{ 135, 208, 23 },
		{ 135, 209, 23 },
		{ 135, 210, 23 },
		{ 135, 211, 23 },
		{ 135, 212, 23 },
		{ 135, 213, 23 },
		{ 135, 214, 23 },
		{ 135, 215, 23 },
		{ 135, 216, 23 },
		{ 135, 217, 23 },
		{ 135, 218, 23 },
		{ 135, 219, 23 },
		{ 135, 220, 23 },
		{ 135, 221, 23 },
		{ 135, 222, 23 },
		{ 135, 223, 23 },
		{ 135, 224, 23 },
		{ 135, 225, 23 },
		{ 135, 226, 23 },
		{ 135, 227, 23 },
		{ 135, 228, 23 },
		{ 135, 229, 23 },
		{ 135, 230, 23 },
		{ 135, 231, 23 },
		{ 135, 232, 23 },
		{ 135, 233, 23 },
		{ 135, 234, 23 },
		{ 135, 235, 23 },
		{ 135, 236, 23 },
		{ 135, 237, 23 },
		{ 135, 238, 23 },
		{ 135, 239, 23 },
		{ 135, 240, 23 },
		{ 135, 241, 23 },
		{ 135, 242, 23 },
		{ 135, 243, 23 },
		{ 135, 244, 23 },
		{ 135, 245, 23 },
		{ 135, 246, 23 },
		{ 135, 247, 23 },
		{ 135, 248, 23 },
		{ 135, 249, 23 },
		{ 135, 250, 23 },
		{ 135, 251, 23 },
		{ 135, 252, 23 },
		{ 135, 253, 23 },
		{ 135, 254, 23 },
		{ 135, 255, 23 },
		{ 135, 256, 23 },
		{ 136, 198, 23 },
		{ 136, 199, 23 },
		{ 136, 200, 23 },
		{ 136, 201, 23 },
		{ 136, 202, 23 },
		{ 136, 203, 23 },
		{ 136, 204, 23 },
		{ 136, 205, 23 },
		{ 136, 206, 23 },
		{ 136, 207, 23 },
		{ 136, 208, 23 },
		{ 136, 209, 23 },
		{ 136, 210, 23 },
		{ 136, 211, 23 },
		{ 136, 212, 23 },
		{ 136, 213, 23 },
		{ 136, 214, 23 },
		{ 136, 215, 23 },
		{ 136, 216, 23 },
		{ 136, 217, 23 },
		{ 136, 218, 23 },
		{ 136, 219, 23 },
		{ 136, 220, 23 },
		{ 136, 221, 23 },
		{ 136, 222, 23 },
		{ 136, 223, 23 },
		{ 136, 224, 23 },
		{ 136, 225, 23 },
		{ 136, 226, 23 },
		{ 136, 227, 23 },
		{ 136, 228, 23 },
		{ 136, 229, 23 },
		{ 136, 230, 23 },
		{ 136, 231, 23 },
		{ 136, 232, 23 },
		{ 136, 233, 23 },
		{ 136, 234, 23 },
		{ 136, 235, 23 },
		{ 136, 236, 23 },
		{ 136, 237, 23 },
		{ 136, 238, 23 },
		{ 136, 239, 23 },
		{ 136, 240, 23 },
		{ 136, 241, 23 },
		{ 136, 242, 23 },
		{ 136, 243, 23 },
		{ 136, 244, 23 },
		{ 136, 245, 23 },
		{ 136, 246, 23 },
		{ 136, 247, 23 },
		{ 136, 248, 23 },
		{ 136, 249, 23 },
		{ 136, 250, 23 },
		{ 136, 251, 23 },
		{ 136, 252, 23 },
		{ 136, 253, 23 },
		{ 136, 254, 23 },
		{ 136, 255, 23 },
		{ 136, 256, 23 },
		{ 137, 198, 23 },
		{ 137, 199, 23 },
		{ 137, 200, 23 },
		{ 137, 201, 23 },
		{ 137, 202, 23 },
		{ 137, 203, 23 },
		{ 137, 204, 23 },
		{ 137, 205, 23 },
		{ 137, 206, 23 },
		{ 137, 207, 23 },
		{ 137, 208, 23 },
		{ 137, 209, 23 },
		{ 137, 210, 23 },
		{ 137, 211, 23 },
		{ 137, 212, 23 },
		{ 137, 213, 23 },
		{ 137, 214, 23 },
		{ 137, 215, 23 },
		{ 137, 216, 23 },
		{ 137, 217, 23 },
		{ 137, 218, 23 },
		{ 137, 219, 23 },
		{ 137, 220, 23 },
		{ 137, 221, 23 },
		{ 137, 222, 23 },
		{ 137, 223, 23 },
		{ 137, 224, 23 },
		{ 137, 225, 23 },
		{ 137, 226, 23 },
		{ 137, 227, 23 },
		{ 137, 228, 23 },
		{ 137, 229, 23 },
		{ 137, 230, 23 },
		{ 137, 231, 23 },
		{ 137, 232, 23 },
		{ 137, 233, 23 },
		{ 137, 234, 23 },
		{ 137, 235, 23 },
		{ 137, 236, 23 },
		{ 137, 237, 23 },
		{ 137, 238, 23 },
		{ 137, 239, 23 },
		{ 137, 240, 23 },
		{ 137, 241, 23 },
		{ 137, 242, 23 },
		{ 137, 243, 23 },
		{ 137, 244, 23 },
		{ 137, 245, 23 },
		{ 137, 246, 23 },
		{ 137, 247, 23 },
		{ 137, 248, 23 },
		{ 137, 249, 23 },
		{ 137, 250, 23 },
		{ 137, 251, 23 },
		{ 137, 252, 23 },
		{ 137, 253, 23 },
		{ 137, 254, 23 },
		{ 137, 255, 23 },
		{ 137, 256, 23 },
		{ 138, 198, 23 },
		{ 138, 199, 23 },
		{ 138, 200, 23 },
		{ 138, 201, 23 },
		{ 138, 202, 23 },
		{ 138, 203, 23 },
		{ 138, 204, 23 },
		{ 138, 205, 23 },
		{ 138, 206, 23 },
		{ 138, 207, 23 },
		{ 138, 208, 23 },
		{ 138, 209, 23 },
		{ 138, 210, 23 },
		{ 138, 211, 23 },
		{ 138, 212, 23 },
		{ 138, 213, 23 },
		{ 138, 214, 23 },
		{ 138, 215, 23 },
		{ 138, 216, 23 },
		{ 138, 217, 23 },
		{ 138, 218, 23 },
		{ 138, 219, 23 },
		{ 138, 220, 23 },
		{ 138, 221, 23 },
		{ 138, 222, 23 },
		{ 138, 223, 23 },
		{ 138, 224, 23 },
		{ 138, 225, 23 },
		{ 138, 226, 23 },
		{ 138, 227, 23 },
		{ 138, 228, 23 },
		{ 138, 229, 23 },
		{ 138, 230, 23 },
		{ 138, 231, 23 },
		{ 138, 232, 23 },
		{ 138, 233, 23 },
		{ 138, 234, 23 },
		{ 138, 235, 23 },
		{ 138, 236, 23 },
		{ 138, 237, 23 },
		{ 138, 238, 23 },
		{ 138, 239, 23 },
		{ 138, 240, 23 },
		{ 138, 241, 23 },
		{ 138, 242, 23 },
		{ 138, 243, 23 },
		{ 138, 244, 23 },
		{ 138, 245, 23 },
		{ 138, 246, 23 },
		{ 138, 247, 23 },
		{ 138, 248, 23 },
		{ 138, 249, 23 },
		{ 138, 250, 23 },
		{ 138, 251, 23 },
		{ 138, 252, 23 },
		{ 138, 253, 23 },
		{ 138, 254, 23 },
		{ 138, 255, 23 },
		{ 138, 256, 23 },
		{ 139, 199, 23 },
		{ 139, 200, 23 },
		{ 139, 201, 23 },
		{ 139, 202, 23 },
		{ 139, 203, 23 },
		{ 139, 204, 23 },
		{ 139, 205, 23 },
		{ 139, 206, 23 },
		{ 139, 207, 23 },
		{ 139, 208, 23 },
		{ 139, 209, 23 },
		{ 139, 210, 23 },
		{ 139, 211, 23 },
		{ 139, 212, 23 },
		{ 139, 213, 23 },
		{ 139, 214, 23 },
		{ 139, 215, 23 },
		{ 139, 216, 23 },
		{ 139, 217, 23 },
		{ 139, 218, 23 },
		{ 139, 219, 23 },
		{ 139, 220, 23 },
		{ 139, 221, 23 },
		{ 139, 222, 23 },
		{ 139, 223, 23 },
		{ 139, 224, 23 },
		{ 139, 225, 23 },
		{ 139, 226, 23 },
		{ 139, 227, 23 },
		{ 139, 228, 23 },
		{ 139, 229, 23 },
		{ 139, 230, 23 },
		{ 139, 231, 23 },
		{ 139, 232, 23 },
		{ 139, 233, 23 },
		{ 139, 234, 23 },
		{ 139, 235, 23 },
		{ 139, 236, 23 },
		{ 139, 237, 23 },
		{ 139, 238, 23 },
		{ 139, 239, 23 },
		{ 139, 240, 23 },
		{ 139, 241, 23 },
		{ 139, 242, 23 },
		{ 139, 243, 23 },
		{ 139, 244, 23 },
		{ 139, 245, 23 },
		{ 139, 246, 23 },
		{ 139, 247, 23 },
		{ 139, 248, 23 },
		{ 139, 249, 23 },
		{ 139, 250, 23 },
		{ 139, 251, 23 },
		{ 139, 252, 23 },
		{ 139, 253, 23 },
		{ 139, 254, 23 },
		{ 139, 255, 23 },
		{ 139, 256, 23 },
		{ 140, 199, 23 },
		{ 140, 200, 23 },
		{ 140, 201, 23 },
		{ 140, 202, 23 },
		{ 140, 203, 23 },
		{ 140, 204, 23 },
		{ 140, 205, 23 },
		{ 140, 206, 23 },
		{ 140, 207, 23 },
		{ 140, 208, 23 },
		{ 140, 209, 23 },
		{ 140, 210, 23 },
		{ 140, 211, 23 },
		{ 140, 212, 23 },
		{ 140, 213, 23 },
		{ 140, 214, 23 },
		{ 140, 215, 23 },
		{ 140, 216, 23 },
		{ 140, 217, 23 },
		{ 140, 218, 23 },
		{ 140, 219, 23 },
		{ 140, 220, 23 },
		{ 140, 221, 23 },
		{ 140, 222, 23 },
		{ 140, 223, 23 },
		{ 140, 224, 23 },
		{ 140, 225, 23 },
		{ 140, 226, 23 },
		{ 140, 227, 23 },
		{ 140, 228, 23 },
		{ 140, 229, 23 },
		{ 140, 230, 23 },
		{ 140, 231, 23 },
		{ 140, 232, 23 },
		{ 140, 233, 23 },
		{ 140, 234, 23 },
		{ 140, 235, 23 },
		{ 140, 236, 23 },
		{ 140, 237, 23 },
		{ 140, 238, 23 },
		{ 140, 239, 23 },
		{ 140, 240, 23 },
		{ 140, 241, 23 },
		{ 140, 242, 23 },
		{ 140, 243, 23 },
		{ 140, 244, 23 },
		{ 140, 245, 23 },
		{ 140, 246, 23 },
		{ 140, 247, 23 },
		{ 140, 248, 23 },
		{ 140, 249, 23 },
		{ 140, 250, 23 },
		{ 140, 251, 23 },
		{ 140, 252, 23 },
		{ 140, 253, 23 },
		{ 140, 254, 23 },
		{ 140, 255, 23 },
		{ 140, 256, 23 },
		{ 141, 200, 23 },
		{ 141, 201, 23 },
		{ 141, 202, 23 },
		{ 141, 203, 23 },
		{ 141, 204, 23 },
		{ 141, 205, 23 },
		{ 141, 206, 23 },
		{ 141, 207, 23 },
		{ 141, 208, 23 },
		{ 141, 209, 23 },
		{ 141, 210, 23 },
		{ 141, 211, 23 },
		{ 141, 212, 23 },
		{ 141, 213, 23 },
		{ 141, 214, 23 },
		{ 141, 215, 23 },
		{ 141, 216, 23 },
		{ 141, 217, 23 },
		{ 141, 218, 23 },
		{ 141, 219, 23 },
		{ 141, 220, 23 },
		{ 141, 221, 23 },
		{ 141, 222, 23 },
		{ 141, 223, 23 },
		{ 141, 224, 23 },
		{ 141, 225, 23 },
		{ 141, 226, 23 },
		{ 141, 227, 23 },
		{ 141, 228, 23 },
		{ 141, 229, 23 },
		{ 141, 230, 23 },
		{ 141, 231, 23 },
		{ 141, 232, 23 },
		{ 141, 233, 23 },
		{ 141, 234, 23 },
		{ 141, 235, 23 },
		{ 141, 236, 23 },
		{ 141, 237, 23 },
		{ 141, 238, 23 },
		{ 141, 239, 23 },
		{ 141, 240, 23 },
		{ 141, 241, 23 },
		{ 141, 242, 23 },
		{ 141, 243, 23 },
		{ 141, 244, 23 },
		{ 141, 245, 23 },
		{ 141, 246, 23 },
		{ 141, 247, 23 },
		{ 141, 248, 23 },
		{ 141, 249, 23 },
		{ 141, 250, 23 },
		{ 141, 251, 23 },
		{ 141, 252, 23 },
		{ 141, 253, 23 },
		{ 141, 254, 23 },
		{ 141, 255, 23 },
		{ 141, 256, 23 },
		{ 142, 200, 23 },
		{ 142, 201, 23 },
		{ 142, 202, 23 },
		{ 142, 203, 23 },
		{ 142, 204, 23 },
		{ 142, 205, 23 },
		{ 142, 206, 23 },
		{ 142, 207, 23 },
		{ 142, 208, 23 },
		{ 142, 209, 23 },
		{ 142, 210, 23 },
		{ 142, 211, 23 },
		{ 142, 212, 23 },
		{ 142, 213, 23 },
		{ 142, 214, 23 },
		{ 142, 215, 23 },
		{ 142, 216, 23 },
		{ 142, 217, 23 },
		{ 142, 218, 23 },
		{ 142, 219, 23 },
		{ 142, 220, 23 },
		{ 142, 221, 23 },
		{ 142, 222, 23 },
		{ 142, 223, 23 },
		{ 142, 224, 23 },
		{ 142, 225, 23 },
		{ 142, 226, 23 },
		{ 142, 227, 23 },
		{ 142, 228, 23 },
		{ 142, 229, 23 },
		{ 142, 230, 23 },
		{ 142, 231, 23 },
		{ 142, 232, 23 },
		{ 142, 233, 23 },
		{ 142, 234, 23 },
		{ 142, 235, 23 },
		{ 142, 236, 23 },
		{ 142, 237, 23 },
		{ 142, 238, 23 },
		{ 142, 239, 23 },
		{ 142, 240, 23 },
		{ 142, 241, 23 },
		{ 142, 242, 23 },
		{ 142, 243, 23 },
		{ 142, 244, 23 },
		{ 142, 245, 23 },
		{ 142, 246, 23 },
		{ 142, 247, 23 },
		{ 142, 248, 23 },
		{ 142, 249, 23 },
		{ 142, 250, 23 },
		{ 142, 251, 23 },
		{ 142, 252, 23 },
		{ 142, 253, 23 },
		{ 142, 254, 23 },
		{ 142, 255, 23 },
		{ 142, 256, 23 },
		{ 143, 201, 23 },
		{ 143, 202, 23 },
		{ 143, 203, 23 },
		{ 143, 204, 23 },
		{ 143, 205, 23 },
		{ 143, 206, 23 },
		{ 143, 207, 23 },
		{ 143, 208, 23 },
		{ 143, 209, 23 },
		{ 143, 210, 23 },
		{ 143, 211, 23 },
		{ 143, 212, 23 },
		{ 143, 213, 23 },
		{ 143, 214, 23 },
		{ 143, 215, 23 },
		{ 143, 216, 23 },
		{ 143, 217, 23 },
		{ 143, 218, 23 },
		{ 143, 219, 23 },
		{ 143, 220, 23 },
		{ 143, 221, 23 },
		{ 143, 222, 23 },
		{ 143, 223, 23 },
		{ 143, 224, 23 },
		{ 143, 225, 23 },
		{ 143, 226, 23 },
		{ 143, 227, 23 },
		{ 143, 228, 23 },
		{ 143, 229, 23 },
		{ 143, 230, 23 },
		{ 143, 231, 23 },
		{ 143, 232, 23 },
		{ 143, 233, 23 },
		{ 143, 234, 23 },
		{ 143, 235, 23 },
		{ 143, 236, 23 },
		{ 143, 237, 23 },
		{ 143, 238, 23 },
		{ 143, 239, 23 },
		{ 143, 240, 23 },
		{ 143, 241, 23 },
		{ 143, 242, 23 },
		{ 143, 243, 23 },
		{ 143, 244, 23 },
		{ 143, 245, 23 },
		{ 143, 246, 23 },
		{ 143, 247, 23 },
		{ 143, 248, 23 },
		{ 143, 249, 23 },
		{ 143, 250, 23 },
		{ 143, 251, 23 },
		{ 143, 252, 23 },
		{ 143, 253, 23 },
		{ 143, 254, 23 },
		{ 143, 255, 23 },
		{ 143, 256, 23 },
		{ 144, 201, 23 },
		{ 144, 202, 23 },
		{ 144, 203, 23 },
		{ 144, 204, 23 },
		{ 144, 205, 23 },
		{ 144, 206, 23 },
		{ 144, 207, 23 },
		{ 144, 208, 23 },
		{ 144, 209, 23 },
		{ 144, 210, 23 },
		{ 144, 211, 23 },
		{ 144, 212, 23 },
		{ 144, 213, 23 },
		{ 144, 214, 23 },
		{ 144, 215, 23 },
		{ 144, 216, 23 },
		{ 144, 217, 23 },
		{ 144, 218, 23 },
		{ 144, 219, 23 },
		{ 144, 220, 23 },
		{ 144, 221, 23 },
		{ 144, 222, 23 },
		{ 144, 223, 23 },
		{ 144, 224, 23 },
		{ 144, 225, 23 },
		{ 144, 226, 23 },
		{ 144, 227, 23 },
		{ 144, 228, 23 },
		{ 144, 229, 23 },
		{ 144, 230, 23 },
		{ 144, 231, 23 },
		{ 144, 232, 23 },
		{ 144, 233, 23 },
		{ 144, 234, 23 },
		{ 144, 235, 23 },
		{ 144, 236, 23 },
		{ 144, 237, 23 },
		{ 144, 238, 23 },
		{ 144, 239, 23 },
		{ 144, 240, 23 },
		{ 144, 241, 23 },
		{ 144, 242, 23 },
		{ 144, 243, 23 },
		{ 144, 244, 23 },
		{ 144, 245, 23 },
		{ 144, 246, 23 },
		{ 144, 247, 23 },
		{ 144, 248, 23 },
		{ 144, 249, 23 },
		{ 144, 250, 23 },
		{ 144, 251, 23 },
		{ 144, 252, 23 },
		{ 144, 253, 23 },
		{ 144, 254, 23 },
		{ 144, 255, 23 },
		{ 144, 256, 23 },
		{ 145, 202, 23 },
		{ 145, 203, 23 },
		{ 145, 204, 23 },
		{ 145, 205, 23 },
		{ 145, 206, 23 },
		{ 145, 207, 23 },
		{ 145, 208, 23 },
		{ 145, 209, 23 },
		{ 145, 210, 23 },
		{ 145, 211, 23 },
		{ 145, 212, 23 },
		{ 145, 213, 23 },
		{ 145, 214, 23 },
		{ 145, 215, 23 },
		{ 145, 216, 23 },
		{ 145, 217, 23 },
		{ 145, 218, 23 },
		{ 145, 219, 23 },
		{ 145, 220, 23 },
		{ 145, 221, 23 },
		{ 145, 222, 23 },
		{ 145, 223, 23 },
		{ 145, 224, 23 },
		{ 145, 225, 23 },
		{ 145, 226, 23 },
		{ 145, 227, 23 },
		{ 145, 228, 23 },
		{ 145, 229, 23 },
		{ 145, 230, 23 },
		{ 145, 231, 23 },
		{ 145, 232, 23 },
		{ 145, 233, 23 },
		{ 145, 234, 23 },
		{ 145, 235, 23 },
		{ 145, 236, 23 },
		{ 145, 237, 23 },
		{ 145, 238, 23 },
		{ 145, 239, 23 },
		{ 145, 240, 23 },
		{ 145, 241, 23 },
		{ 145, 242, 23 },
		{ 145, 243, 23 },
		{ 145, 244, 23 },
		{ 145, 245, 23 },
		{ 145, 246, 23 },
		{ 145, 247, 23 },
		{ 145, 248, 23 },
		{ 145, 249, 23 },
		{ 145, 250, 23 },
		{ 145, 251, 23 },
		{ 145, 252, 23 },
		{ 145, 253, 23 },
		{ 145, 254, 23 },
		{ 145, 255, 23 },
		{ 145, 256, 23 },
		{ 146, 202, 23 },
		{ 146, 203, 23 },
		{ 146, 204, 23 },
		{ 146, 205, 23 },
		{ 146, 206, 23 },
		{ 146, 207, 23 },
		{ 146, 208, 23 },
		{ 146, 209, 23 },
		{ 146, 210, 23 },
		{ 146, 211, 23 },
		{ 146, 212, 23 },
		{ 146, 213, 23 },
		{ 146, 214, 23 },
		{ 146, 215, 23 },
		{ 146, 216, 23 },
		{ 146, 217, 23 },
		{ 146, 218, 23 },
		{ 146, 219, 23 },
		{ 146, 220, 23 },
		{ 146, 221, 23 },
		{ 146, 222, 23 },
		{ 146, 223, 23 },
		{ 146, 224, 23 },
		{ 146, 225, 23 },
		{ 146, 226, 23 },
		{ 146, 227, 23 },
		{ 146, 228, 23 },
		{ 146, 229, 23 },
		{ 146, 230, 23 },
		{ 146, 231, 23 },
		{ 146, 232, 23 },
		{ 146, 233, 23 },
		{ 146, 234, 23 },
		{ 146, 235, 23 },
		{ 146, 236, 23 },
		{ 146, 237, 23 },
		{ 146, 238, 23 },
		{ 146, 239, 23 },
		{ 146, 240, 23 },
		{ 146, 241, 23 },
		{ 146, 242, 23 },
		{ 146, 243, 23 },
		{ 146, 244, 23 },
		{ 146, 245, 23 },
		{ 146, 246, 23 },
		{ 146, 247, 23 },
		{ 146, 248, 23 },
		{ 146, 249, 23 },
		{ 146, 250, 23 },
		{ 146, 251, 23 },
		{ 146, 252, 23 },
		{ 146, 253, 23 },
		{ 146, 254, 23 },
		{ 146, 255, 23 },
		{ 146, 256, 23 },
		{ 147, 203, 23 },
		{ 147, 204, 23 },
		{ 147, 205, 23 },
		{ 147, 206, 23 },
		{ 147, 207, 23 },
		{ 147, 208, 23 },
		{ 147, 209, 23 },
		{ 147, 210, 23 },
		{ 147, 211, 23 },
		{ 147, 212, 23 },
		{ 147, 213, 23 },
		{ 147, 214, 23 },
		{ 147, 215, 23 },
		{ 147, 216, 23 },
		{ 147, 217, 23 },
		{ 147, 218, 23 },
		{ 147, 219, 23 },
		{ 147, 220, 23 },
		{ 147, 221, 23 },
		{ 147, 222, 23 },
		{ 147, 223, 23 },
		{ 147, 224, 23 },
		{ 147, 225, 23 },
		{ 147, 226, 23 },
		{ 147, 227, 23 },
		{ 147, 228, 23 },
		{ 147, 229, 23 },
		{ 147, 230, 23 },
		{ 147, 231, 23 },
		{ 147, 232, 23 },
		{ 147, 233, 23 },
		{ 147, 234, 23 },
		{ 147, 235, 23 },
		{ 147, 236, 23 },
		{ 147, 237, 23 },
		{ 147, 238, 23 },
		{ 147, 239, 23 },
		{ 147, 240, 23 },
		{ 147, 241, 23 },
		{ 147, 242, 23 },
		{ 147, 243, 23 },
		{ 147, 244, 23 },
		{ 147, 245, 23 },
		{ 147, 246, 23 },
		{ 147, 247, 23 },
		{ 147, 248, 23 },
		{ 147, 249, 23 },
		{ 147, 250, 23 },
		{ 147, 251, 23 },
		{ 147, 252, 23 },
		{ 147, 253, 23 },
		{ 147, 254, 23 },
		{ 147, 255, 23 },
		{ 147, 256, 23 },
		{ 148, 203, 23 },
		{ 148, 204, 23 },
		{ 148, 205, 23 },
		{ 148, 206, 23 },
		{ 148, 207, 23 },
		{ 148, 208, 23 },
		{ 148, 209, 23 },
		{ 148, 210, 23 },
		{ 148, 211, 23 },
		{ 148, 212, 23 },
		{ 148, 213, 23 },
		{ 148, 214, 23 },
		{ 148, 215, 23 },
		{ 148, 216, 23 },
		{ 148, 217, 23 },
		{ 148, 218, 23 },
		{ 148, 219, 23 },
		{ 148, 220, 23 },
		{ 148, 221, 23 },
		{ 148, 222, 23 },
		{ 148, 223, 23 },
		{ 148, 224, 23 },
		{ 148, 225, 23 },
		{ 148, 226, 23 },
		{ 148, 227, 23 },
		{ 148, 228, 23 },
		{ 148, 229, 23 },
		{ 148, 230, 23 },
		{ 148, 231, 23 },
		{ 148, 232, 23 },
		{ 148, 233, 23 },
		{ 148, 234, 23 },
		{ 148, 235, 23 },
		{ 148, 236, 23 },
		{ 148, 237, 23 },
		{ 148, 238, 23 },
		{ 148, 239, 23 },
		{ 148, 240, 23 },
		{ 148, 241, 23 },
		{ 148, 242, 23 },
		{ 148, 243, 23 },
		{ 148, 244, 23 },
		{ 148, 245, 23 },
		{ 148, 246, 23 },
		{ 148, 247, 23 },
		{ 148, 248, 23 },
		{ 148, 249, 23 },
		{ 148, 250, 23 },
		{ 148, 251, 23 },
		{ 148, 252, 23 },
		{ 148, 253, 23 },
		{ 148, 254, 23 },
		{ 148, 255, 23 },
		{ 148, 256, 23 },
		{ 149, 204, 23 },
		{ 149, 205, 23 },
		{ 149, 206, 23 },
		{ 149, 207, 23 },
		{ 149, 208, 23 },
		{ 149, 209, 23 },
		{ 149, 210, 23 },
		{ 149, 211, 23 },
		{ 149, 212, 23 },
		{ 149, 213, 23 },
		{ 149, 214, 23 },
		{ 149, 215, 23 },
		{ 149, 216, 23 },
		{ 149, 217, 23 },
		{ 149, 218, 23 },
		{ 149, 219, 23 },
		{ 149, 220, 23 },
		{ 149, 221, 23 },
		{ 149, 222, 23 },
		{ 149, 223, 23 },
		{ 149, 224, 23 },
		{ 149, 225, 23 },
		{ 149, 226, 23 },
		{ 149, 227, 23 },
		{ 149, 228, 23 },
		{ 149, 229, 23 },
		{ 149, 230, 23 },
		{ 149, 231, 23 },
		{ 149, 232, 23 },
		{ 149, 233, 23 },
		{ 149, 234, 23 },
		{ 149, 235, 23 },
		{ 149, 236, 23 },
		{ 149, 237, 23 },
		{ 149, 238, 23 },
		{ 149, 239, 23 },
		{ 149, 240, 23 },
		{ 149, 241, 23 },
		{ 149, 242, 23 },
		{ 149, 243, 23 },
		{ 149, 244, 23 },
		{ 149, 245, 23 },
		{ 149, 246, 23 },
		{ 149, 247, 23 },
		{ 149, 248, 23 },
		{ 149, 249, 23 },
		{ 149, 250, 23 },
		{ 149, 251, 23 },
		{ 149, 252, 23 },
		{ 149, 253, 23 },
		{ 149, 254, 23 },
		{ 149, 255, 23 },
		{ 149, 256, 23 },
		{ 150, 204, 23 },
		{ 150, 205, 23 },
		{ 150, 206, 23 },
		{ 150, 207, 23 },
		{ 150, 208, 23 },
		{ 150, 209, 23 },
		{ 150, 210, 23 },
		{ 150, 211, 23 },
		{ 150, 212, 23 },
		{ 150, 213, 23 },
		{ 150, 214, 23 },
		{ 150, 215, 23 },
		{ 150, 216, 23 },
		{ 150, 217, 23 },
		{ 150, 218, 23 },
		{ 150, 219, 23 },
		{ 150, 220, 23 },
		{ 150, 221, 23 },
		{ 150, 222, 23 },
		{ 150, 223, 23 },
		{ 150, 224, 23 },
		{ 150, 225, 23 },
		{ 150, 226, 23 },
		{ 150, 227, 23 },
		{ 150, 228, 23 },
		{ 150, 229, 23 },
		{ 150, 230, 23 },
		{ 150, 231, 23 },
		{ 150, 232, 23 },
		{ 150, 233, 23 },
		{ 150, 234, 23 },
		{ 150, 235, 23 },
		{ 150, 236, 23 },
		{ 150, 237, 23 },
		{ 150, 238, 23 },
		{ 150, 239, 23 },
		{ 150, 240, 23 },
		{ 150, 241, 23 },
		{ 150, 242, 23 },
		{ 150, 243, 23 },
		{ 150, 244, 23 },
		{ 150, 245, 23 },
		{ 150, 246, 23 },
		{ 150, 247, 23 },
		{ 150, 248, 23 },
		{ 150, 249, 23 },
		{ 150, 250, 23 },
		{ 150, 251, 23 },
		{ 150, 252, 23 },
		{ 150, 253, 23 },
		{ 150, 254, 23 },
		{ 150, 255, 23 },
		{ 150, 256, 23 },
		{ 151, 205, 23 },
		{ 151, 206, 23 },
		{ 151, 207, 23 },
		{ 151, 208, 23 },
		{ 151, 209, 23 },
		{ 151, 210, 23 },
		{ 151, 211, 23 },
		{ 151, 212, 23 },
		{ 151, 213, 23 },
		{ 151, 214, 23 },
		{ 151, 215, 23 },
		{ 151, 216, 23 },
		{ 151, 217, 23 },
		{ 151, 218, 23 },
		{ 151, 219, 23 },
		{ 151, 220, 23 },
		{ 151, 221, 23 },
		{ 151, 222, 23 },
		{ 151, 223, 23 },
		{ 151, 224, 23 },
		{ 151, 225, 23 },
		{ 151, 226, 23 },
		{ 151, 227, 23 },
		{ 151, 228, 23 },
		{ 151, 229, 23 },
		{ 151, 230, 23 },
		{ 151, 231, 23 },
		{ 151, 232, 23 },
		{ 151, 233, 23 },
		{ 151, 234, 23 },
		{ 151, 235, 23 },
		{ 151, 236, 23 },
		{ 151, 237, 23 },
		{ 151, 238, 23 },
		{ 151, 239, 23 },
		{ 151, 240, 23 },
		{ 151, 241, 23 },
		{ 151, 242, 23 },
		{ 151, 243, 23 },
		{ 151, 244, 23 },
		{ 151, 245, 23 },
		{ 151, 246, 23 },
		{ 151, 247, 23 },
		{ 151, 248, 23 },
		{ 151, 249, 23 },
		{ 151, 250, 23 },
		{ 151, 251, 23 },
		{ 151, 252, 23 },
		{ 151, 253, 23 },
		{ 151, 254, 23 },
		{ 151, 255, 23 },
		{ 151, 256, 23 },
		{ 152, 205, 23 },
		{ 152, 206, 23 },
		{ 152, 207, 23 },
		{ 152, 208, 23 },
		{ 152, 209, 23 },
		{ 152, 210, 23 },
		{ 152, 211, 23 },
		{ 152, 212, 23 },
		{ 152, 213, 23 },
		{ 152, 214, 23 },
		{ 152, 215, 23 },
		{ 152, 216, 23 },
		{ 152, 217, 23 },
		{ 152, 218, 23 },
		{ 152, 219, 23 },
		{ 152, 220, 23 },
		{ 152, 221, 23 },
		{ 152, 222, 23 },
		{ 152, 223, 23 },
		{ 152, 224, 23 },
		{ 152, 225, 23 },
		{ 152, 226, 23 },
		{ 152, 227, 23 },
		{ 152, 228, 23 },
		{ 152, 229, 23 },
		{ 152, 230, 23 },
		{ 152, 231, 23 },
		{ 152, 232, 23 },
		{ 152, 233, 23 },
		{ 152, 234, 23 },
		{ 152, 235, 23 },
		{ 152, 236, 23 },
		{ 152, 237, 23 },
		{ 152, 238, 23 },
		{ 152, 239, 23 },
		{ 152, 240, 23 },
		{ 152, 241, 23 },
		{ 152, 242, 23 },
		{ 152, 243, 23 },
		{ 152, 244, 23 },
		{ 152, 245, 23 },
		{ 152, 246, 23 },
		{ 152, 247, 23 },
		{ 152, 248, 23 },
		{ 152, 249, 23 },
		{ 152, 250, 23 },
		{ 152, 251, 23 },
		{ 152, 252, 23 },
		{ 152, 253, 23 },
		{ 152, 254, 23 },
		{ 152, 255, 23 },
		{ 152, 256, 23 },
		{ 153, 206, 23 },
		{ 153, 207, 23 },
		{ 153, 208, 23 },
		{ 153, 209, 23 },
		{ 153, 210, 23 },
		{ 153, 211, 23 },
		{ 153, 212, 23 },
		{ 153, 213, 23 },
		{ 153, 214, 23 },
		{ 153, 215, 23 },
		{ 153, 216, 23 },
		{ 153, 217, 23 },
		{ 153, 218, 23 },
		{ 153, 219, 23 },
		{ 153, 220, 23 },
		{ 153, 221, 23 },
		{ 153, 222, 23 },
		{ 153, 223, 23 },
		{ 153, 224, 23 },
		{ 153, 225, 23 },
		{ 153, 226, 23 },
		{ 153, 227, 23 },
		{ 153, 228, 23 },
		{ 153, 229, 23 },
		{ 153, 230, 23 },
		{ 153, 231, 23 },
		{ 153, 232, 23 },
		{ 153, 233, 23 },
		{ 153, 234, 23 },
		{ 153, 235, 23 },
		{ 153, 236, 23 },
		{ 153, 237, 23 },
		{ 153, 238, 23 },
		{ 153, 239, 23 },
		{ 153, 240, 23 },
		{ 153, 241, 23 },
		{ 153, 242, 23 },
		{ 153, 243, 23 },
		{ 153, 244, 23 },
		{ 153, 245, 23 },
		{ 153, 246, 23 },
		{ 153, 247, 23 },
		{ 153, 248, 23 },
		{ 153, 249, 23 },
		{ 153, 250, 23 },
		{ 153, 251, 23 },
		{ 153, 252, 23 },
		{ 153, 253, 23 },
		{ 153, 254, 23 },
		{ 153, 255, 23 },
		{ 153, 256, 23 },
		{ 154, 206, 23 },
		{ 154, 207, 23 },
		{ 154, 208, 23 },
		{ 154, 209, 23 },
		{ 154, 210, 23 },
		{ 154, 211, 23 },
		{ 154, 212, 23 },
		{ 154, 213, 23 },
		{ 154, 214, 23 },
		{ 154, 215, 23 },
		{ 154, 216, 23 },
		{ 154, 217, 23 },
		{ 154, 218, 23 },
		{ 154, 219, 23 },
		{ 154, 220, 23 },
		{ 154, 221, 23 },
		{ 154, 222, 23 },
		{ 154, 223, 23 },
		{ 154, 224, 23 },
		{ 154, 225, 23 },
		{ 154, 226, 23 },
		{ 154, 227, 23 },
		{ 154, 228, 23 },
		{ 154, 229, 23 },
		{ 154, 230, 23 },
		{ 154, 231, 23 },
		{ 154, 232, 23 },
		{ 154, 233, 23 },
		{ 154, 234, 23 },
		{ 154, 235, 23 },
		{ 154, 236, 23 },
		{ 154, 237, 23 },
		{ 154, 238, 23 },
		{ 154, 239, 23 },
		{ 154, 240, 23 },
		{ 154, 241, 23 },
		{ 154, 242, 23 },
		{ 154, 243, 23 },
		{ 154, 244, 23 },
		{ 154, 245, 23 },
		{ 154, 246, 23 },
		{ 154, 247, 23 },
		{ 154, 248, 23 },
		{ 154, 249, 23 },
		{ 154, 250, 23 },
		{ 154, 251, 23 },
		{ 154, 252, 23 },
		{ 154, 253, 23 },
		{ 154, 254, 23 },
		{ 154, 255, 23 },
		{ 154, 256, 23 },
		{ 155, 207, 23 },
		{ 155, 208, 23 },
		{ 155, 209, 23 },
		{ 155, 210, 23 },
		{ 155, 211, 23 },
		{ 155, 212, 23 },
		{ 155, 213, 23 },
		{ 155, 214, 23 },
		{ 155, 215, 23 },
		{ 155, 216, 23 },
		{ 155, 217, 23 },
		{ 155, 218, 23 },
		{ 155, 219, 23 },
		{ 155, 220, 23 },
		{ 155, 221, 23 },
		{ 155, 222, 23 },
		{ 155, 223, 23 },
		{ 155, 224, 23 },
		{ 155, 225, 23 },
		{ 155, 226, 23 },
		{ 155, 227, 23 },
		{ 155, 228, 23 },
		{ 155, 229, 23 },
		{ 155, 230, 23 },
		{ 155, 231, 23 },
		{ 155, 232, 23 },
		{ 155, 233, 23 },
		{ 155, 234, 23 },
		{ 155, 235, 23 },
		{ 155, 236, 23 },
		{ 155, 237, 23 },
		{ 155, 238, 23 },
		{ 155, 239, 23 },
		{ 155, 240, 23 },
		{ 155, 241, 23 },
		{ 155, 242, 23 },
		{ 155, 243, 23 },
		{ 155, 244, 23 },
		{ 155, 245, 23 },
		{ 155, 246, 23 },
		{ 155, 247, 23 },
		{ 155, 248, 23 },
		{ 155, 249, 23 },
		{ 155, 250, 23 },
		{ 155, 251, 23 },
		{ 155, 252, 23 },
		{ 155, 253, 23 },
		{ 155, 254, 23 },
		{ 155, 255, 23 },
		{ 155, 256, 23 },
		{ 156, 207, 23 },
		{ 156, 208, 23 },
		{ 156, 209, 23 },
		{ 156, 210, 23 },
		{ 156, 211, 23 },
		{ 156, 212, 23 },
		{ 156, 213, 23 },
		{ 156, 214, 23 },
		{ 156, 215, 23 },
		{ 156, 216, 23 },
		{ 156, 217, 23 },
		{ 156, 218, 23 },
		{ 156, 219, 23 },
		{ 156, 220, 23 },
		{ 156, 221, 23 },
		{ 156, 222, 23 },
		{ 156, 223, 23 },
		{ 156, 224, 23 },
		{ 156, 225, 23 },
		{ 156, 226, 23 },
		{ 156, 227, 23 },
		{ 156, 228, 23 },
		{ 156, 229, 23 },
		{ 156, 230, 23 },
		{ 156, 231, 23 },
		{ 156, 232, 23 },
		{ 156, 233, 23 },
		{ 156, 234, 23 },
		{ 156, 235, 23 },
		{ 156, 236, 23 },
		{ 156, 237, 23 },
		{ 156, 238, 23 },
		{ 156, 239, 23 },
		{ 156, 240, 23 },
		{ 156, 241, 23 },
		{ 156, 242, 23 },
		{ 156, 243, 23 },
		{ 156, 244, 23 },
		{ 156, 245, 23 },
		{ 156, 246, 23 },
		{ 156, 247, 23 },
		{ 156, 248, 23 },
		{ 156, 249, 23 },
		{ 156, 250, 23 },
		{ 156, 251, 23 },
		{ 156, 252, 23 },
		{ 156, 253, 23 },
		{ 156, 254, 23 },
		{ 156, 255, 23 },
		{ 156, 256, 23 },
		{ 157, 208, 23 },
		{ 157, 209, 23 },
		{ 157, 210, 23 },
		{ 157, 211, 23 },
		{ 157, 212, 23 },
		{ 157, 213, 23 },
		{ 157, 214, 23 },
		{ 157, 215, 23 },
		{ 157, 216, 23 },
		{ 157, 217, 23 },
		{ 157, 218, 23 },
		{ 157, 219, 23 },
		{ 157, 220, 23 },
		{ 157, 221, 23 },
		{ 157, 222, 23 },
		{ 157, 223, 23 },
		{ 157, 224, 23 },
		{ 157, 225, 23 },
		{ 157, 226, 23 },
		{ 157, 227, 23 },
		{ 157, 228, 23 },
		{ 157, 229, 23 },
		{ 157, 230, 23 },
		{ 157, 231, 23 },
		{ 157, 232, 23 },
		{ 157, 233, 23 },
		{ 157, 234, 23 },
		{ 157, 235, 23 },
		{ 157, 236, 23 },
		{ 157, 237, 23 },
		{ 157, 238, 23 },
		{ 157, 239, 23 },
		{ 157, 240, 23 },
		{ 157, 241, 23 },
		{ 157, 242, 23 },
		{ 157, 243, 23 },
		{ 157, 244, 23 },
		{ 157, 245, 23 },
		{ 157, 246, 23 },
		{ 157, 247, 23 },
		{ 157, 248, 23 },
		{ 157, 249, 23 },
		{ 157, 250, 23 },
		{ 157, 251, 23 },
		{ 157, 252, 23 },
		{ 157, 253, 23 },
		{ 157, 254, 23 },
		{ 157, 255, 23 },
		{ 157, 256, 23 },
		{ 158, 208, 23 },
		{ 158, 209, 23 },
		{ 158, 210, 23 },
		{ 158, 211, 23 },
		{ 158, 212, 23 },
		{ 158, 213, 23 },
		{ 158, 214, 23 },
		{ 158, 215, 23 },
		{ 158, 216, 23 },
		{ 158, 217, 23 },
		{ 158, 218, 23 },
		{ 158, 219, 23 },
		{ 158, 220, 23 },
		{ 158, 221, 23 },
		{ 158, 222, 23 },
		{ 158, 223, 23 },
		{ 158, 224, 23 },
		{ 158, 225, 23 },
		{ 158, 226, 23 },
		{ 158, 227, 23 },
		{ 158, 228, 23 },
		{ 158, 229, 23 },
		{ 158, 230, 23 },
		{ 158, 231, 23 },
		{ 158, 232, 23 },
		{ 158, 233, 23 },
		{ 158, 234, 23 },
		{ 158, 235, 23 },
		{ 158, 236, 23 },
		{ 158, 237, 23 },
		{ 158, 238, 23 },
		{ 158, 239, 23 },
		{ 158, 240, 23 },
		{ 158, 241, 23 },
		{ 158, 242, 23 },
		{ 158, 243, 23 },
		{ 158, 244, 23 },
		{ 158, 245, 23 },
		{ 158, 246, 23 },
		{ 158, 247, 23 },
		{ 158, 248, 23 },
		{ 158, 249, 23 },
		{ 158, 250, 23 },
		{ 158, 251, 23 },
		{ 158, 252, 23 },
		{ 158, 253, 23 },
		{ 158, 254, 23 },
		{ 158, 255, 23 },
		{ 158, 256, 23 },
		{ 159, 209, 23 },
		{ 159, 210, 23 },
		{ 159, 211, 23 },
		{ 159, 212, 23 },
		{ 159, 213, 23 },
		{ 159, 214, 23 },
		{ 159, 215, 23 },
		{ 159, 216, 23 },
		{ 159, 217, 23 },
		{ 159, 218, 23 },
		{ 159, 219, 23 },
		{ 159, 220, 23 },
		{ 159, 221, 23 },
		{ 159, 222, 23 },
		{ 159, 223, 23 },
		{ 159, 224, 23 },
		{ 159, 225, 23 },
		{ 159, 226, 23 },
		{ 159, 227, 23 },
		{ 159, 228, 23 },
		{ 159, 229, 23 },
		{ 159, 230, 23 },
		{ 159, 231, 23 },
		{ 159, 232, 23 },
		{ 159, 233, 23 },
		{ 159, 234, 23 },
		{ 159, 235, 23 },
		{ 159, 236, 23 },
		{ 159, 237, 23 },
		{ 159, 238, 23 },
		{ 159, 239, 23 },
		{ 159, 240, 23 },
		{ 159, 241, 23 },
		{ 159, 242, 23 },
		{ 159, 243, 23 },
		{ 159, 244, 23 },
		{ 159, 245, 23 },
		{ 159, 246, 23 },
		{ 159, 247, 23 },
		{ 159, 248, 23 },
		{ 159, 249, 23 },
		{ 159, 250, 23 },
		{ 159, 251, 23 },
		{ 159, 252, 23 },
		{ 159, 253, 23 },
		{ 159, 254, 23 },
		{ 159, 255, 23 },
		{ 159, 256, 23 },
		{ 160, 209, 23 },
		{ 160, 210, 23 },
		{ 160, 211, 23 },
		{ 160, 212, 23 },
		{ 160, 213, 23 },
		{ 160, 214, 23 },
		{ 160, 215, 23 },
		{ 160, 216, 23 },
		{ 160, 217, 23 },
		{ 160, 218, 23 },
		{ 160, 219, 23 },
		{ 160, 220, 23 },
		{ 160, 221, 23 },
		{ 160, 222, 23 },
		{ 160, 223, 23 },
		{ 160, 224, 23 },
		{ 160, 225, 23 },
		{ 160, 226, 23 },
		{ 160, 227, 23 },
		{ 160, 228, 23 },
		{ 160, 229, 23 },
		{ 160, 230, 23 },
		{ 160, 231, 23 },
		{ 160, 232, 23 },
		{ 160, 233, 23 },
		{ 160, 234, 23 },
		{ 160, 235, 23 },
		{ 160, 236, 23 },
		{ 160, 237, 23 },
		{ 160, 238, 23 },
		{ 160, 239, 23 },
		{ 160, 240, 23 },
		{ 160, 241, 23 },
		{ 160, 242, 23 },
		{ 160, 243, 23 },
		{ 160, 244, 23 },
		{ 160, 245, 23 },
		{ 160, 246, 23 },
		{ 160, 247, 23 },
		{ 160, 248, 23 },
		{ 160, 249, 23 },
		{ 160, 250, 23 },
		{ 160, 251, 23 },
		{ 160, 252, 23 },
		{ 160, 253, 23 },
		{ 160, 254, 23 },
		{ 160, 255, 23 },
		{ 160, 256, 23 },
		{ 161, 210, 23 },
		{ 161, 211, 23 },
		{ 161, 212, 23 },
		{ 161, 213, 23 },
		{ 161, 214, 23 },
		{ 161, 215, 23 },
		{ 161, 216, 23 },
		{ 161, 217, 23 },
		{ 161, 218, 23 },
		{ 161, 219, 23 },
		{ 161, 220, 23 },
		{ 161, 221, 23 },
		{ 161, 222, 23 },
		{ 161, 223, 23 },
		{ 161, 224, 23 },
		{ 161, 225, 23 },
		{ 161, 226, 23 },
		{ 161, 227, 23 },
		{ 161, 228, 23 },
		{ 161, 229, 23 },
		{ 161, 230, 23 },
		{ 161, 231, 23 },
		{ 161, 232, 23 },
		{ 161, 233, 23 },
		{ 161, 234, 23 },
		{ 161, 235, 23 },
		{ 161, 236, 23 },
		{ 161, 237, 23 },
		{ 161, 238, 23 },
		{ 161, 239, 23 },
		{ 161, 240, 23 },
		{ 161, 241, 23 },
		{ 161, 242, 23 },
		{ 161, 243, 23 },
		{ 161, 244, 23 },
		{ 161, 245, 23 },
		{ 161, 246, 23 },
		{ 161, 247, 23 },
		{ 161, 248, 23 },
		{ 161, 249, 23 },
		{ 161, 250, 23 },
		{ 161, 251, 23 },
		{ 161, 252, 23 },
		{ 161, 253, 23 },
		{ 161, 254, 23 },
		{ 161, 255, 23 },
		{ 161, 256, 23 },
		{ 162, 210, 23 },
		{ 162, 211, 23 },
		{ 162, 212, 23 },
		{ 162, 213, 23 },
		{ 162, 214, 23 },
		{ 162, 215, 23 },
		{ 162, 216, 23 },
		{ 162, 217, 23 },
		{ 162, 218, 23 },
		{ 162, 219, 23 },
		{ 162, 220, 23 },
		{ 162, 221, 23 },
		{ 162, 222, 23 },
		{ 162, 223, 23 },
		{ 162, 224, 23 },
		{ 162, 225, 23 },
		{ 162, 226, 23 },
		{ 162, 227, 23 },
		{ 162, 228, 23 },
		{ 162, 229, 23 },
		{ 162, 230, 23 },
		{ 162, 231, 23 },
		{ 162, 232, 23 },
		{ 162, 233, 23 },
		{ 162, 234, 23 },
		{ 162, 235, 23 },
		{ 162, 236, 23 },
		{ 162, 237, 23 },
		{ 162, 238, 23 },
		{ 162, 239, 23 },
		{ 162, 240, 23 },
		{ 162, 241, 23 },
		{ 162, 242, 23 },
		{ 162, 243, 23 },
		{ 162, 244, 23 },
		{ 162, 245, 23 },
		{ 162, 246, 23 },
		{ 162, 247, 23 },
		{ 162, 248, 23 },
		{ 162, 249, 23 },
		{ 162, 250, 23 },
		{ 162, 251, 23 },
		{ 162, 252, 23 },
		{ 162, 253, 23 },
		{ 162, 254, 23 },
		{ 162, 255, 23 },
		{ 162, 256, 23 },
		{ 163, 211, 23 },
		{ 163, 212, 23 },
		{ 163, 213, 23 },
		{ 163, 214, 23 },
		{ 163, 215, 23 },
		{ 163, 216, 23 },
		{ 163, 217, 23 },
		{ 163, 218, 23 },
		{ 163, 219, 23 },
		{ 163, 220, 23 },
		{ 163, 221, 23 },
		{ 163, 222, 23 },
		{ 163, 223, 23 },
		{ 163, 224, 23 },
		{ 163, 225, 23 },
		{ 163, 226, 23 },
		{ 163, 227, 23 },
		{ 163, 228, 23 },
		{ 163, 229, 23 },
		{ 163, 230, 23 },
		{ 163, 231, 23 },
		{ 163, 232, 23 },
		{ 163, 233, 23 },
		{ 163, 234, 23 },
		{ 163, 235, 23 },
		{ 163, 236, 23 },
		{ 163, 237, 23 },
		{ 163, 238, 23 },
		{ 163, 239, 23 },
		{ 163, 240, 23 },
		{ 163, 241, 23 },
		{ 163, 242, 23 },
		{ 163, 243, 23 },
		{ 163, 244, 23 },
		{ 163, 245, 23 },
		{ 163, 246, 23 },
		{ 163, 247, 23 },
		{ 163, 248, 23 },
		{ 163, 249, 23 },
		{ 163, 250, 23 },
		{ 163, 251, 23 },
		{ 163, 252, 23 },
		{ 163, 253, 23 },
		{ 163, 254, 23 },
		{ 163, 255, 23 },
		{ 163, 256, 23 },
		{ 164, 211, 23 },
		{ 164, 212, 23 },
		{ 164, 213, 23 },
		{ 164, 214, 23 },
		{ 164, 215, 23 },
		{ 164, 216, 23 },
		{ 164, 217, 23 },
		{ 164, 218, 23 },
		{ 164, 219, 23 },
		{ 164, 220, 23 },
		{ 164, 221, 23 },
		{ 164, 222, 23 },
		{ 164, 223, 23 },
		{ 164, 224, 23 },
		{ 164, 225, 23 },
		{ 164, 226, 23 },
		{ 164, 227, 23 },
		{ 164, 228, 23 },
		{ 164, 229, 23 },
		{ 164, 230, 23 },
		{ 164, 231, 23 },
		{ 164, 232, 23 },
		{ 164, 233, 23 },
		{ 164, 234, 23 },
		{ 164, 235, 23 },
		{ 164, 236, 23 },
		{ 164, 237, 23 },
		{ 164, 238, 23 },
		{ 164, 239, 23 },
		{ 164, 240, 23 },
		{ 164, 241, 23 },
		{ 164, 242, 23 },
		{ 164, 243, 23 },
		{ 164, 244, 23 },
		{ 164, 245, 23 },
		{ 164, 246, 23 },
		{ 164, 247, 23 },
		{ 164, 248, 23 },
		{ 164, 249, 23 },
		{ 164, 250, 23 },
		{ 164, 251, 23 },
		{ 164, 252, 23 },
		{ 164, 253, 23 },
		{ 164, 254, 23 },
		{ 164, 255, 23 },
		{ 164, 256, 23 },
		{ 165, 212, 23 },
		{ 165, 213, 23 },
		{ 165, 214, 23 },
		{ 165, 215, 23 },
		{ 165, 216, 23 },
		{ 165, 217, 23 },
		{ 165, 218, 23 },
		{ 165, 219, 23 },
		{ 165, 220, 23 },
		{ 165, 221, 23 },
		{ 165, 222, 23 },
		{ 165, 223, 23 },
		{ 165, 224, 23 },
		{ 165, 225, 23 },
		{ 165, 226, 23 },
		{ 165, 227, 23 },
		{ 165, 228, 23 },
		{ 165, 229, 23 },
		{ 165, 230, 23 },
		{ 165, 231, 23 },
		{ 165, 232, 23 },
		{ 165, 233, 23 },
		{ 165, 234, 23 },
		{ 165, 235, 23 },
		{ 165, 236, 23 },
		{ 165, 237, 23 },
		{ 165, 238, 23 },
		{ 165, 239, 23 },
		{ 165, 240, 23 },
		{ 165, 241, 23 },
		{ 165, 242, 23 },
		{ 165, 243, 23 },
		{ 165, 244, 23 },
		{ 165, 245, 23 },
		{ 165, 246, 23 },
		{ 165, 247, 23 },
		{ 165, 248, 23 },
		{ 165, 249, 23 },
		{ 165, 250, 23 },
		{ 165, 251, 23 },
		{ 165, 252, 23 },
		{ 165, 253, 23 },
		{ 165, 254, 23 },
		{ 165, 255, 23 },
		{ 165, 256, 23 },
		{ 166, 212, 23 },
		{ 166, 213, 23 },
		{ 166, 214, 23 },
		{ 166, 215, 23 },
		{ 166, 216, 23 },
		{ 166, 217, 23 },
		{ 166, 218, 23 },
		{ 166, 219, 23 },
		{ 166, 220, 23 },
		{ 166, 221, 23 },
		{ 166, 222, 23 },
		{ 166, 223, 23 },
		{ 166, 224, 23 },
		{ 166, 225, 23 },
		{ 166, 226, 23 },
		{ 166, 227, 23 },
		{ 166, 228, 23 },
		{ 166, 229, 23 },
		{ 166, 230, 23 },
		{ 166, 231, 23 },
		{ 166, 232, 23 },
		{ 166, 233, 23 },
		{ 166, 234, 23 },
		{ 166, 235, 23 },
		{ 166, 236, 23 },
		{ 166, 237, 23 },
		{ 166, 238, 23 },
		{ 166, 239, 23 },
		{ 166, 240, 23 },
		{ 166, 241, 23 },
		{ 166, 242, 23 },
		{ 166, 243, 23 },
		{ 166, 244, 23 },
		{ 166, 245, 23 },
		{ 166, 246, 23 },
		{ 166, 247, 23 },
		{ 166, 248, 23 },
		{ 166, 249, 23 },
		{ 166, 250, 23 },
		{ 166, 251, 23 },
		{ 166, 252, 23 },
		{ 166, 253, 23 },
		{ 166, 254, 23 },
		{ 166, 255, 23 },
		{ 166, 256, 23 },
		{ 167, 213, 23 },
		{ 167, 214, 23 },
		{ 167, 215, 23 },
		{ 167, 216, 23 },
		{ 167, 217, 23 },
		{ 167, 218, 23 },
		{ 167, 219, 23 },
		{ 167, 220, 23 },
		{ 167, 221, 23 },
		{ 167, 222, 23 },
		{ 167, 223, 23 },
		{ 167, 224, 23 },
		{ 167, 225, 23 },
		{ 167, 226, 23 },
		{ 167, 227, 23 },
		{ 167, 228, 23 },
		{ 167, 229, 23 },
		{ 167, 230, 23 },
		{ 167, 231, 23 },
		{ 167, 232, 23 },
		{ 167, 233, 23 },
		{ 167, 234, 23 },
		{ 167, 235, 23 },
		{ 167, 236, 23 },
		{ 167, 237, 23 },
		{ 167, 238, 23 },
		{ 167, 239, 23 },
		{ 167, 240, 23 },
		{ 167, 241, 23 },
		{ 167, 242, 23 },
		{ 167, 243, 23 },
		{ 167, 244, 23 },
		{ 167, 245, 23 },
		{ 167, 246, 23 },
		{ 167, 247, 23 },
		{ 167, 248, 23 },
		{ 167, 249, 23 },
		{ 167, 250, 23 },
		{ 167, 251, 23 },
		{ 167, 252, 23 },
		{ 167, 253, 23 },
		{ 167, 254, 23 },
		{ 167, 255, 23 },
		{ 167, 256, 23 },
		{ 168, 213, 23 },
		{ 168, 214, 23 },
		{ 168, 215, 23 },
		{ 168, 216, 23 },
		{ 168, 217, 23 },
		{ 168, 218, 23 },
		{ 168, 219, 23 },
		{ 168, 220, 23 },
		{ 168, 221, 23 },
		{ 168, 222, 23 },
		{ 168, 223, 23 },
		{ 168, 224, 23 },
		{ 168, 225, 23 },
		{ 168, 226, 23 },
		{ 168, 227, 23 },
		{ 168, 228, 23 },
		{ 168, 229, 23 },
		{ 168, 230, 23 },
		{ 168, 231, 23 },
		{ 168, 232, 23 },
		{ 168, 233, 23 },
		{ 168, 234, 23 },
		{ 168, 235, 23 },
		{ 168, 236, 23 },
		{ 168, 237, 23 },
		{ 168, 238, 23 },
		{ 168, 239, 23 },
		{ 168, 240, 23 },
		{ 168, 241, 23 },
		{ 168, 242, 23 },
		{ 168, 243, 23 },
		{ 168, 244, 23 },
		{ 168, 245, 23 },
		{ 168, 246, 23 },
		{ 168, 247, 23 },
		{ 168, 248, 23 },
		{ 168, 249, 23 },
		{ 168, 250, 23 },
		{ 168, 251, 23 },
		{ 168, 252, 23 },
		{ 168, 253, 23 },
		{ 168, 254, 23 },
		{ 168, 255, 23 },
		{ 168, 256, 23 },
		{ 169, 214, 23 },
		{ 169, 215, 23 },
		{ 169, 216, 23 },
		{ 169, 217, 23 },
		{ 169, 218, 23 },
		{ 169, 219, 23 },
		{ 169, 220, 23 },
		{ 169, 221, 23 },
		{ 169, 222, 23 },
		{ 169, 223, 23 },
		{ 169, 224, 23 },
		{ 169, 225, 23 },
		{ 169, 226, 23 },
		{ 169, 227, 23 },
		{ 169, 228, 23 },
		{ 169, 229, 23 },
		{ 169, 230, 23 },
		{ 169, 231, 23 },
		{ 169, 232, 23 },
		{ 169, 233, 23 },
		{ 169, 234, 23 },
		{ 169, 235, 23 },
		{ 169, 236, 23 },
		{ 169, 237, 23 },
		{ 169, 238, 23 },
		{ 169, 239, 23 },
		{ 169, 240, 23 },
		{ 169, 241, 23 },
		{ 169, 242, 23 },
		{ 169, 243, 23 },
		{ 169, 244, 23 },
		{ 169, 245, 23 },
		{ 169, 246, 23 },
		{ 169, 247, 23 },
		{ 169, 248, 23 },
		{ 169, 249, 23 },
		{ 169, 250, 23 },
		{ 169, 251, 23 },
		{ 169, 252, 23 },
		{ 169, 253, 23 },
		{ 169, 254, 23 },
		{ 169, 255, 23 },
		{ 169, 256, 23 },
		{ 170, 214, 23 },
		{ 170, 215, 23 },
		{ 170, 216, 23 },
		{ 170, 217, 23 },
		{ 170, 218, 23 },
		{ 170, 219, 23 },
		{ 170, 220, 23 },
		{ 170, 221, 23 },
		{ 170, 222, 23 },
		{ 170, 223, 23 },
		{ 170, 224, 23 },
		{ 170, 225, 23 },
		{ 170, 226, 23 },
		{ 170, 227, 23 },
		{ 170, 228, 23 },
		{ 170, 229, 23 },
		{ 170, 230, 23 },
		{ 170, 231, 23 },
		{ 170, 232, 23 },
		{ 170, 233, 23 },
		{ 170, 234, 23 },
		{ 170, 235, 23 },
		{ 170, 236, 23 },
		{ 170, 237, 23 },
		{ 170, 238, 23 },
		{ 170, 239, 23 },
		{ 170, 240, 23 },
		{ 170, 241, 23 },
		{ 170, 242, 23 },
		{ 170, 243, 23 },
		{ 170, 244, 23 },
		{ 170, 245, 23 },
		{ 170, 246, 23 },
		{ 170, 247, 23 },
		{ 170, 248, 23 },
		{ 170, 249, 23 },
		{ 170, 250, 23 },
		{ 170, 251, 23 },
		{ 170, 252, 23 },
		{ 170, 253, 23 },
		{ 170, 254, 23 },
		{ 170, 255, 23 },
		{ 170, 256, 23 },
		{ 171, 215, 23 },
		{ 171, 216, 23 },
		{ 171, 217, 23 },
		{ 171, 218, 23 },
		{ 171, 219, 23 },
		{ 171, 220, 23 },
		{ 171, 221, 23 },
		{ 171, 222, 23 },
		{ 171, 223, 23 },
		{ 171, 224, 23 },
		{ 171, 225, 23 },
		{ 171, 226, 23 },
		{ 171, 227, 23 },
		{ 171, 228, 23 },
		{ 171, 229, 23 },
		{ 171, 230, 23 },
		{ 171, 231, 23 },
		{ 171, 232, 23 },
		{ 171, 233, 23 },
		{ 171, 234, 23 },
		{ 171, 235, 23 },
		{ 171, 236, 23 },
		{ 171, 237, 23 },
		{ 171, 238, 23 },
		{ 171, 239, 23 },
		{ 171, 240, 23 },
		{ 171, 241, 23 },
		{ 171, 242, 23 },
		{ 171, 243, 23 },
		{ 171, 244, 23 },
		{ 171, 245, 23 },
		{ 171, 246, 23 },
		{ 171, 247, 23 },
		{ 171, 248, 23 },
		{ 171, 249, 23 },
		{ 171, 250, 23 },
		{ 171, 251, 23 },
		{ 171, 252, 23 },
		{ 171, 253, 23 },
		{ 171, 254, 23 },
		{ 171, 255, 23 },
		{ 171, 256, 23 },
		{ 172, 215, 23 },
		{ 172, 216, 23 },
		{ 172, 217, 23 },
		{ 172, 218, 23 },
		{ 172, 219, 23 },
		{ 172, 220, 23 },
		{ 172, 221, 23 },
		{ 172, 222, 23 },
		{ 172, 223, 23 },
		{ 172, 224, 23 },
		{ 172, 225, 23 },
		{ 172, 226, 23 },
		{ 172, 227, 23 },
		{ 172, 228, 23 },
		{ 172, 229, 23 },
		{ 172, 230, 23 },
		{ 172, 231, 23 },
		{ 172, 232, 23 },
		{ 172, 233, 23 },
		{ 172, 234, 23 },
		{ 172, 235, 23 },
		{ 172, 236, 23 },
		{ 172, 237, 23 },
		{ 172, 238, 23 },
		{ 172, 239, 23 },
		{ 172, 240, 23 },
		{ 172, 241, 23 },
		{ 172, 242, 23 },
		{ 172, 243, 23 },
		{ 172, 244, 23 },
		{ 172, 245, 23 },
		{ 172, 246, 23 },
		{ 172, 247, 23 },
		{ 172, 248, 23 },
		{ 172, 249, 23 },
		{ 172, 250, 23 },
		{ 172, 251, 23 },
		{ 172, 252, 23 },
		{ 172, 253, 23 },
		{ 172, 254, 23 },
		{ 172, 255, 23 },
		{ 172, 256, 23 },
		{ 173, 216, 23 },
		{ 173, 217, 23 },
		{ 173, 218, 23 },
		{ 173, 219, 23 },
		{ 173, 220, 23 },
		{ 173, 221, 23 },
		{ 173, 222, 23 },
		{ 173, 223, 23 },
		{ 173, 224, 23 },
		{ 173, 225, 23 },
		{ 173, 226, 23 },
		{ 173, 227, 23 },
		{ 173, 228, 23 },
		{ 173, 229, 23 },
		{ 173, 230, 23 },
		{ 173, 231, 23 },
		{ 173, 232, 23 },
		{ 173, 233, 23 },
		{ 173, 234, 23 },
		{ 173, 235, 23 },
		{ 173, 236, 23 },
		{ 173, 237, 23 },
		{ 173, 238, 23 },
		{ 173, 239, 23 },
		{ 173, 240, 23 },
		{ 173, 241, 23 },
		{ 173, 242, 23 },
		{ 173, 243, 23 },
		{ 173, 244, 23 },
		{ 173, 245, 23 },
		{ 173, 246, 23 },
		{ 173, 247, 23 },
		{ 173, 248, 23 },
		{ 173, 249, 23 },
		{ 173, 250, 23 },
		{ 173, 251, 23 },
		{ 173, 252, 23 },
		{ 173, 253, 23 },
		{ 173, 254, 23 },
		{ 173, 255, 23 },
		{ 173, 256, 23 },
		{ 174, 216, 23 },
		{ 174, 217, 23 },
		{ 174, 218, 23 },
		{ 174, 219, 23 },
		{ 174, 220, 23 },
		{ 174, 221, 23 },
		{ 174, 222, 23 },
		{ 174, 223, 23 },
		{ 174, 224, 23 },
		{ 174, 225, 23 },
		{ 174, 226, 23 },
		{ 174, 227, 23 },
		{ 174, 228, 23 },
		{ 174, 229, 23 },
		{ 174, 230, 23 },
		{ 174, 231, 23 },
		{ 174, 232, 23 },
		{ 174, 233, 23 },
		{ 174, 234, 23 },
		{ 174, 235, 23 },
		{ 174, 236, 23 },
		{ 174, 237, 23 },
		{ 174, 238, 23 },
		{ 174, 239, 23 },
		{ 174, 240, 23 },
		{ 174, 241, 23 },
		{ 174, 242, 23 },
		{ 174, 243, 23 },
		{ 174, 244, 23 },
		{ 174, 245, 23 },
		{ 174, 246, 23 },
		{ 174, 247, 23 },
		{ 174, 248, 23 },
		{ 174, 249, 23 },
		{ 174, 250, 23 },
		{ 174, 251, 23 },
		{ 174, 252, 23 },
		{ 174, 253, 23 },
		{ 174, 254, 23 },
		{ 174, 255, 23 },
		{ 174, 256, 23 },
		{ 175, 217, 23 },
		{ 175, 218, 23 },
		{ 175, 219, 23 },
		{ 175, 220, 23 },
		{ 175, 221, 23 },
		{ 175, 222, 23 },
		{ 175, 223, 23 },
		{ 175, 224, 23 },
		{ 175, 225, 23 },
		{ 175, 226, 23 },
		{ 175, 227, 23 },
		{ 175, 228, 23 },
		{ 175, 229, 23 },
		{ 175, 230, 23 },
		{ 175, 231, 23 },
		{ 175, 232, 23 },
		{ 175, 233, 23 },
		{ 175, 234, 23 },
		{ 175, 235, 23 },
		{ 175, 236, 23 },
		{ 175, 237, 23 },
		{ 175, 238, 23 },
		{ 175, 239, 23 },
		{ 175, 240, 23 },
		{ 175, 241, 23 },
		{ 175, 242, 23 },
		{ 175, 243, 23 },
		{ 175, 244, 23 },
		{ 175, 245, 23 },
		{ 175, 246, 23 },
		{ 175, 247, 23 },
		{ 175, 248, 23 },
		{ 175, 249, 23 },
		{ 175, 250, 23 },
		{ 175, 251, 23 },
		{ 175, 252, 23 },
		{ 175, 253, 23 },
		{ 175, 254, 23 },
		{ 175, 255, 23 },
		{ 175, 256, 23 },
		{ 176, 217, 23 },
		{ 176, 218, 23 },
		{ 176, 219, 23 },
		{ 176, 220, 23 },
		{ 176, 221, 23 },
		{ 176, 222, 23 },
		{ 176, 223, 23 },
		{ 176, 224, 23 },
		{ 176, 225, 23 },
		{ 176, 226, 23 },
		{ 176, 227, 23 },
		{ 176, 228, 23 },
		{ 176, 229, 23 },
		{ 176, 230, 23 },
		{ 176, 231, 23 },
		{ 176, 232, 23 },
		{ 176, 233, 23 },
		{ 176, 234, 23 },
		{ 176, 235, 23 },
		{ 176, 236, 23 },
		{ 176, 237, 23 },
		{ 176, 238, 23 },
		{ 176, 239, 23 },
		{ 176, 240, 23 },
		{ 176, 241, 23 },
		{ 176, 242, 23 },
		{ 176, 243, 23 },
		{ 176, 244, 23 },
		{ 176, 245, 23 },
		{ 176, 246, 23 },
		{ 176, 247, 23 },
		{ 176, 248, 23 },
		{ 176, 249, 23 },
		{ 176, 250, 23 },
		{ 176, 251, 23 },
		{ 176, 252, 23 },
		{ 176, 253, 23 },
		{ 176, 254, 23 },
		{ 176, 255, 23 },
		{ 176, 256, 23 },
		{ 177, 218, 23 },
		{ 177, 219, 23 },
		{ 177, 220, 23 },
		{ 177, 221, 23 },
		{ 177, 222, 23 },
		{ 177, 223, 23 },
		{ 177, 224, 23 },
		{ 177, 225, 23 },
		{ 177, 226, 23 },
		{ 177, 227, 23 },
		{ 177, 228, 23 },
		{ 177, 229, 23 },
		{ 177, 230, 23 },
		{ 177, 231, 23 },
		{ 177, 232, 23 },
		{ 177, 233, 23 },
		{ 177, 234, 23 },
		{ 177, 235, 23 },
		{ 177, 236, 23 },
		{ 177, 237, 23 },
		{ 177, 238, 23 },
		{ 177, 239, 23 },
		{ 177, 240, 23 },
		{ 177, 241, 23 },
		{ 177, 242, 23 },
		{ 177, 243, 23 },
		{ 177, 244, 23 },
		{ 177, 245, 23 },
		{ 177, 246, 23 },
		{ 177, 247, 23 },
		{ 177, 248, 23 },
		{ 177, 249, 23 },
		{ 177, 250, 23 },
		{ 177, 251, 23 },
		{ 177, 252, 23 },
		{ 177, 253, 23 },
		{ 177, 254, 23 },
		{ 177, 255, 23 },
		{ 177, 256, 23 },
		{ 178, 218, 23 },
		{ 178, 219, 23 },
		{ 178, 220, 23 },
		{ 178, 221, 23 },
		{ 178, 222, 23 },
		{ 178, 223, 23 },
		{ 178, 224, 23 },
		{ 178, 225, 23 },
		{ 178, 226, 23 },
		{ 178, 227, 23 },
		{ 178, 228, 23 },
		{ 178, 229, 23 },
		{ 178, 230, 23 },
		{ 178, 231, 23 },
		{ 178, 232, 23 },
		{ 178, 233, 23 },
		{ 178, 234, 23 },
		{ 178, 235, 23 },
		{ 178, 236, 23 },
		{ 178, 237, 23 },
		{ 178, 238, 23 },
		{ 178, 239, 23 },
		{ 178, 240, 23 },
		{ 178, 241, 23 },
		{ 178, 242, 23 },
		{ 178, 243, 23 },
		{ 178, 244, 23 },
		{ 178, 245, 23 },
		{ 178, 246, 23 },
		{ 178, 247, 23 },
		{ 178, 248, 23 },
		{ 178, 249, 23 },
		{ 178, 250, 23 },
		{ 178, 251, 23 },
		{ 178, 252, 23 },
		{ 178, 253, 23 },
		{ 178, 254, 23 },
		{ 178, 255, 23 },
		{ 178, 256, 23 },
		{ 179, 219, 23 },
		{ 179, 220, 23 },
		{ 179, 221, 23 },
		{ 179, 222, 23 },
		{ 179, 223, 23 },
		{ 179, 224, 23 },
		{ 179, 225, 23 },
		{ 179, 226, 23 },
		{ 179, 227, 23 },
		{ 179, 228, 23 },
		{ 179, 229, 23 },
		{ 179, 230, 23 },
		{ 179, 231, 23 },
		{ 179, 232, 23 },
		{ 179, 233, 23 },
		{ 179, 234, 23 },
		{ 179, 235, 23 },
		{ 179, 236, 23 },
		{ 179, 237, 23 },
		{ 179, 238, 23 },
		{ 179, 239, 23 },
		{ 179, 240, 23 },
		{ 179, 241, 23 },
		{ 179, 242, 23 },
		{ 179, 243, 23 },
		{ 179, 244, 23 },
		{ 179, 245, 23 },
		{ 179, 246, 23 },
		{ 179, 247, 23 },
		{ 179, 248, 23 },
		{ 179, 249, 23 },
		{ 179, 250, 23 },
		{ 179, 251, 23 },
		{ 179, 252, 23 },
		{ 179, 253, 23 },
		{ 179, 254, 23 },
		{ 179, 255, 23 },
		{ 179, 256, 23 },
		{ 180, 219, 23 },
		{ 180, 220, 23 },
		{ 180, 221, 23 },
		{ 180, 222, 23 },
		{ 180, 223, 23 },
		{ 180, 224, 23 },
		{ 180, 225, 23 },
		{ 180, 226, 23 },
		{ 180, 227, 23 },
		{ 180, 228, 23 },
		{ 180, 229, 23 },
		{ 180, 230, 23 },
		{ 180, 231, 23 },
		{ 180, 232, 23 },
		{ 180, 233, 23 },
		{ 180, 234, 23 },
		{ 180, 235, 23 },
		{ 180, 236, 23 },
		{ 180, 237, 23 },
		{ 180, 238, 23 },
		{ 180, 239, 23 },
		{ 180, 240, 23 },
		{ 180, 241, 23 },
		{ 180, 242, 23 },
		{ 180, 243, 23 },
		{ 180, 244, 23 },
		{ 180, 245, 23 },
		{ 180, 246, 23 },
		{ 180, 247, 23 },
		{ 180, 248, 23 },
		{ 180, 249, 23 },
		{ 180, 250, 23 },
		{ 180, 251, 23 },
		{ 180, 252, 23 },
		{ 180, 253, 23 },
		{ 180, 254, 23 },
		{ 180, 255, 23 },
		{ 180, 256, 23 },
		{ 181, 220, 23 },
		{ 181, 221, 23 },
		{ 181, 222, 23 },
		{ 181, 223, 23 },
		{ 181, 224, 23 },
		{ 181, 225, 23 },
		{ 181, 226, 23 },
		{ 181, 227, 23 },
		{ 181, 228, 23 },
		{ 181, 229, 23 },
		{ 181, 230, 23 },
		{ 181, 231, 23 },
		{ 181, 232, 23 },
		{ 181, 233, 23 },
		{ 181, 234, 23 },
		{ 181, 235, 23 },
		{ 181, 236, 23 },
		{ 181, 237, 23 },
		{ 181, 238, 23 },
		{ 181, 239, 23 },
		{ 181, 240, 23 },
		{ 181, 241, 23 },
		{ 181, 242, 23 },
		{ 181, 243, 23 },
		{ 181, 244, 23 },
		{ 181, 245, 23 },
		{ 181, 246, 23 },
		{ 181, 247, 23 },
		{ 181, 248, 23 },
		{ 181, 249, 23 },
		{ 181, 250, 23 },
		{ 181, 251, 23 },
		{ 181, 252, 23 },
		{ 181, 253, 23 },
		{ 181, 254, 23 },
		{ 181, 255, 23 },
		{ 181, 256, 23 },
		{ 182, 220, 23 },
		{ 182, 221, 23 },
		{ 182, 222, 23 },
		{ 182, 223, 23 },
		{ 182, 224, 23 },
		{ 182, 225, 23 },
		{ 182, 226, 23 },
		{ 182, 227, 23 },
		{ 182, 228, 23 },
		{ 182, 229, 23 },
		{ 182, 230, 23 },
		{ 182, 231, 23 },
		{ 182, 232, 23 },
		{ 182, 233, 23 },
		{ 182, 234, 23 },
		{ 182, 235, 23 },
		{ 182, 236, 23 },
		{ 182, 237, 23 },
		{ 182, 238, 23 },
		{ 182, 239, 23 },
		{ 182, 240, 23 },
		{ 182, 241, 23 },
		{ 182, 242, 23 },
		{ 182, 243, 23 },
		{ 182, 244, 23 },
		{ 182, 245, 23 },
		{ 182, 246, 23 },
		{ 182, 247, 23 },
		{ 182, 248, 23 },
		{ 182, 249, 23 },
		{ 182, 250, 23 },
		{ 182, 251, 23 },
		{ 182, 252, 23 },
		{ 182, 253, 23 },
		{ 182, 254, 23 },
		{ 182, 255, 23 },
		{ 182, 256, 23 },
		{ 183, 221, 23 },
		{ 183, 222, 23 },
		{ 183, 223, 23 },
		{ 183, 224, 23 },
		{ 183, 225, 23 },
		{ 183, 226, 23 },
		{ 183, 227, 23 },
		{ 183, 228, 23 },
		{ 183, 229, 23 },
		{ 183, 230, 23 },
		{ 183, 231, 23 },
		{ 183, 232, 23 },
		{ 183, 233, 23 },
		{ 183, 234, 23 },
		{ 183, 235, 23 },
		{ 183, 236, 23 },
		{ 183, 237, 23 },
		{ 183, 238, 23 },
		{ 183, 239, 23 },
		{ 183, 240, 23 },
		{ 183, 241, 23 },
		{ 183, 242, 23 },
		{ 183, 243, 23 },
		{ 183, 244, 23 },
		{ 183, 245, 23 },
		{ 183, 246, 23 },
		{ 183, 247, 23 },
		{ 183, 248, 23 },
		{ 183, 249, 23 },
		{ 183, 250, 23 },
		{ 183, 251, 23 },
		{ 183, 252, 23 },
		{ 183, 253, 23 },
		{ 183, 254, 23 },
		{ 183, 255, 23 },
		{ 183, 256, 23 },
		{ 184, 221, 23 },
		{ 184, 222, 23 },
		{ 184, 223, 23 },
		{ 184, 224, 23 },
		{ 184, 225, 23 },
		{ 184, 226, 23 },
		{ 184, 227, 23 },
		{ 184, 228, 23 },
		{ 184, 229, 23 },
		{ 184, 230, 23 },
		{ 184, 231, 23 },
		{ 184, 232, 23 },
		{ 184, 233, 23 },
		{ 184, 234, 23 },
		{ 184, 235, 23 },
		{ 184, 236, 23 },
		{ 184, 237, 23 },
		{ 184, 238, 23 },
		{ 184, 239, 23 },
		{ 184, 240, 23 },
		{ 184, 241, 23 },
		{ 184, 242, 23 },
		{ 184, 243, 23 },
		{ 184, 244, 23 },
		{ 184, 245, 23 },
		{ 184, 246, 23 },
		{ 184, 247, 23 },
		{ 184, 248, 23 },
		{ 184, 249, 23 },
		{ 184, 250, 23 },
		{ 184, 251, 23 },
		{ 184, 252, 23 },
		{ 184, 253, 23 },
		{ 184, 254, 23 },
		{ 184, 255, 23 },
		{ 184, 256, 23 },
		{ 185, 222, 23 },
		{ 185, 223, 23 },
		{ 185, 224, 23 },
		{ 185, 225, 23 },
		{ 185, 226, 23 },
		{ 185, 227, 23 },
		{ 185, 228, 23 },
		{ 185, 229, 23 },
		{ 185, 230, 23 },
		{ 185, 231, 23 },
		{ 185, 232, 23 },
		{ 185, 233, 23 },
		{ 185, 234, 23 },
		{ 185, 235, 23 },
		{ 185, 236, 23 },
		{ 185, 237, 23 },
		{ 185, 238, 23 },
		{ 185, 239, 23 },
		{ 185, 240, 23 },
		{ 185, 241, 23 },
		{ 185, 242, 23 },
		{ 185, 243, 23 },
		{ 185, 244, 23 },
		{ 185, 245, 23 },
		{ 185, 246, 23 },
		{ 185, 247, 23 },
		{ 185, 248, 23 },
		{ 185, 249, 23 },
		{ 185, 250, 23 },
		{ 185, 251, 23 },
		{ 185, 252, 23 },
		{ 185, 253, 23 },
		{ 185, 254, 23 },
		{ 185, 255, 23 },
		{ 185, 256, 23 },
		{ 186, 222, 23 },
		{ 186, 223, 23 },
		{ 186, 224, 23 },
		{ 186, 225, 23 },
		{ 186, 226, 23 },
		{ 186, 227, 23 },
		{ 186, 228, 23 },
		{ 186, 229, 23 },
		{ 186, 230, 23 },
		{ 186, 231, 23 },
		{ 186, 232, 23 },
		{ 186, 233, 23 },
		{ 186, 234, 23 },
		{ 186, 235, 23 },
		{ 186, 236, 23 },
		{ 186, 237, 23 },
		{ 186, 238, 23 },
		{ 186, 239, 23 },
		{ 186, 240, 23 },
		{ 186, 241, 23 },
		{ 186, 242, 23 },
		{ 186, 243, 23 },
		{ 186, 244, 23 },
		{ 186, 245, 23 },
		{ 186, 246, 23 },
		{ 186, 247, 23 },
		{ 186, 248, 23 },
		{ 186, 249, 23 },
		{ 186, 250, 23 },
		{ 186, 251, 23 },
		{ 186, 252, 23 },
		{ 186, 253, 23 },
		{ 186, 254, 23 },
		{ 186, 255, 23 },
		{ 186, 256, 23 },
		{ 187, 223, 23 },
		{ 187, 224, 23 },
		{ 187, 225, 23 },
		{ 187, 226, 23 },
		{ 187, 227, 23 },
		{ 187, 228, 23 },
		{ 187, 229, 23 },
		{ 187, 230, 23 },
		{ 187, 231, 23 },
		{ 187, 232, 23 },
		{ 187, 233, 23 },
		{ 187, 234, 23 },
		{ 187, 235, 23 },
		{ 187, 236, 23 },
		{ 187, 237, 23 },
		{ 187, 238, 23 },
		{ 187, 239, 23 },
		{ 187, 240, 23 },
		{ 187, 241, 23 },
		{ 187, 242, 23 },
		{ 187, 243, 23 },
		{ 187, 244, 23 },
		{ 187, 245, 23 },
		{ 187, 246, 23 },
		{ 187, 247, 23 },
		{ 187, 248, 23 },
		{ 187, 249, 23 },
		{ 187, 250, 23 },
		{ 187, 251, 23 },
		{ 187, 252, 23 },
		{ 187, 253, 23 },
		{ 187, 254, 23 },
		{ 187, 255, 23 },
		{ 187, 256, 23 },
		{ 188, 223, 23 },
		{ 188, 224, 23 },
		{ 188, 225, 23 },
		{ 188, 226, 23 },
		{ 188, 227, 23 },
		{ 188, 228, 23 },
		{ 188, 229, 23 },
		{ 188, 230, 23 },
		{ 188, 231, 23 },
		{ 188, 232, 23 },
		{ 188, 233, 23 },
		{ 188, 234, 23 },
		{ 188, 235, 23 },
		{ 188, 236, 23 },
		{ 188, 237, 23 },
		{ 188, 238, 23 },
		{ 188, 239, 23 },
		{ 188, 240, 23 },
		{ 188, 241, 23 },
		{ 188, 242, 23 },
		{ 188, 243, 23 },
		{ 188, 244, 23 },
		{ 188, 245, 23 },
		{ 188, 246, 23 },
		{ 188, 247, 23 },
		{ 188, 248, 23 },
		{ 188, 249, 23 },
		{ 188, 250, 23 },
		{ 188, 251, 23 },
		{ 188, 252, 23 },
		{ 188, 253, 23 },
		{ 188, 254, 23 },
		{ 188, 255, 23 },
		{ 188, 256, 23 },
		{ 189, 224, 23 },
		{ 189, 225, 23 },
		{ 189, 226, 23 },
		{ 189, 227, 23 },
		{ 189, 228, 23 },
		{ 189, 229, 23 },
		{ 189, 230, 23 },
		{ 189, 231, 23 },
		{ 189, 232, 23 },
		{ 189, 233, 23 },
		{ 189, 234, 23 },
		{ 189, 235, 23 },
		{ 189, 236, 23 },
		{ 189, 237, 23 },
		{ 189, 238, 23 },
		{ 189, 239, 23 },
		{ 189, 240, 23 },
		{ 189, 241, 23 },
		{ 189, 242, 23 },
		{ 189, 243, 23 },
		{ 189, 244, 23 },
		{ 189, 245, 23 },
		{ 189, 246, 23 },
		{ 189, 247, 23 },
		{ 189, 248, 23 },
		{ 189, 249, 23 },
		{ 189, 250, 23 },
		{ 189, 251, 23 },
		{ 189, 252, 23 },
		{ 189, 253, 23 },
		{ 189, 254, 23 },
		{ 189, 255, 23 },
		{ 189, 256, 23 },
		{ 190, 224, 23 },
		{ 190, 225, 23 },
		{ 190, 226, 23 },
		{ 190, 227, 23 },
		{ 190, 228, 23 },
		{ 190, 229, 23 },
		{ 190, 230, 23 },
		{ 190, 231, 23 },
		{ 190, 232, 23 },
		{ 190, 233, 23 },
		{ 190, 234, 23 },
		{ 190, 235, 23 },
		{ 190, 236, 23 },
		{ 190, 237, 23 },
		{ 190, 238, 23 },
		{ 190, 239, 23 },
		{ 190, 240, 23 },
		{ 190, 241, 23 },
		{ 190, 242, 23 },
		{ 190, 243, 23 },
		{ 190, 244, 23 },
		{ 190, 245, 23 },
		{ 190, 246, 23 },
		{ 190, 247, 23 },
		{ 190, 248, 23 },
		{ 190, 249, 23 },
		{ 190, 250, 23 },
		{ 190, 251, 23 },
		{ 190, 252, 23 },
		{ 190, 253, 23 },
		{ 190, 254, 23 },
		{ 190, 255, 23 },
		{ 190, 256, 23 },
		{ 191, 225, 23 },
		{ 191, 226, 23 },
		{ 191, 227, 23 },
		{ 191, 228, 23 },
		{ 191, 229, 23 },
		{ 191, 230, 23 },
		{ 191, 231, 23 },
		{ 191, 232, 23 },
		{ 191, 233, 23 },
		{ 191, 234, 23 },
		{ 191, 235, 23 },
		{ 191, 236, 23 },
		{ 191, 237, 23 },
		{ 191, 238, 23 },
		{ 191, 239, 23 },
		{ 191, 240, 23 },
		{ 191, 241, 23 },
		{ 191, 242, 23 },
		{ 191, 243, 23 },
		{ 191, 244, 23 },
		{ 191, 245, 23 },
		{ 191, 246, 23 },
		{ 191, 247, 23 },
		{ 191, 248, 23 },
		{ 191, 249, 23 },
		{ 191, 250, 23 },
		{ 191, 251, 23 },
		{ 191, 252, 23 },
		{ 191, 253, 23 },
		{ 191, 254, 23 },
		{ 191, 255, 23 },
		{ 191, 256, 23 },
		{ 192, 225, 23 },
		{ 192, 226, 23 },
		{ 192, 227, 23 },
		{ 192, 228, 23 },
		{ 192, 229, 23 },
		{ 192, 230, 23 },
		{ 192, 231, 23 },
		{ 192, 232, 23 },
		{ 192, 233, 23 },
		{ 192, 234, 23 },
		{ 192, 235, 23 },
		{ 192, 236, 23 },
		{ 192, 237, 23 },
		{ 192, 238, 23 },
		{ 192, 239, 23 },
		{ 192, 240, 23 },
		{ 192, 241, 23 },
		{ 192, 242, 23 },
		{ 192, 243, 23 },
		{ 192, 244, 23 },
		{ 192, 245, 23 },
		{ 192, 246, 23 },
		{ 192, 247, 23 },
		{ 192, 248, 23 },
		{ 192, 249, 23 },
		{ 192, 250, 23 },
		{ 192, 251, 23 },
		{ 192, 252, 23 },
		{ 192, 253, 23 },
		{ 192, 254, 23 },
		{ 192, 255, 23 },
		{ 192, 256, 23 },
		{ 193, 226, 23 },
		{ 193, 227, 23 },
		{ 193, 228, 23 },
		{ 193, 229, 23 },
		{ 193, 230, 23 },
		{ 193, 231, 23 },
		{ 193, 232, 23 },
		{ 193, 233, 23 },
		{ 193, 234, 23 },
		{ 193, 235, 23 },
		{ 193, 236, 23 },
		{ 193, 237, 23 },
		{ 193, 238, 23 },
		{ 193, 239, 23 },
		{ 193, 240, 23 },
		{ 193, 241, 23 },
		{ 193, 242, 23 },
		{ 193, 243, 23 },
		{ 193, 244, 23 },
		{ 193, 245, 23 },
		{ 193, 246, 23 },
		{ 193, 247, 23 },
		{ 193, 248, 23 },
		{ 193, 249, 23 },
		{ 193, 250, 23 },
		{ 193, 251, 23 },
		{ 193, 252, 23 },
		{ 193, 253, 23 },
		{ 193, 254, 23 },
		{ 193, 255, 23 },
		{ 193, 256, 23 },
		{ 194, 226, 23 },
		{ 194, 227, 23 },
		{ 194, 228, 23 },
		{ 194, 229, 23 },
		{ 194, 230, 23 },
		{ 194, 231, 23 },
		{ 194, 232, 23 },
		{ 194, 233, 23 },
		{ 194, 234, 23 },
		{ 194, 235, 23 },
		{ 194, 236, 23 },
		{ 194, 237, 23 },
		{ 194, 238, 23 },
		{ 194, 239, 23 },
		{ 194, 240, 23 },
		{ 194, 241, 23 },
		{ 194, 242, 23 },
		{ 194, 243, 23 },
		{ 194, 244, 23 },
		{ 194, 245, 23 },
		{ 194, 246, 23 },
		{ 194, 247, 23 },
		{ 194, 248, 23 },
		{ 194, 249, 23 },
		{ 194, 250, 23 },
		{ 194, 251, 23 },
		{ 194, 252, 23 },
		{ 194, 253, 23 },
		{ 194, 254, 23 },
		{ 194, 255, 23 },
		{ 194, 256, 23 },
		{ 195, 227, 23 },
		{ 195, 228, 23 },
		{ 195, 229, 23 },
		{ 195, 230, 23 },
		{ 195, 231, 23 },
		{ 195, 232, 23 },
		{ 195, 233, 23 },
		{ 195, 234, 23 },
		{ 195, 235, 23 },
		{ 195, 236, 23 },
		{ 195, 237, 23 },
		{ 195, 238, 23 },
		{ 195, 239, 23 },
		{ 195, 240, 23 },
		{ 195, 241, 23 },
		{ 195, 242, 23 },
		{ 195, 243, 23 },
		{ 195, 244, 23 },
		{ 195, 245, 23 },
		{ 195, 246, 23 },
		{ 195, 247, 23 },
		{ 195, 248, 23 },
		{ 195, 249, 23 },
		{ 195, 250, 23 },
		{ 195, 251, 23 },
		{ 195, 252, 23 },
		{ 195, 253, 23 },
		{ 195, 254, 23 },
		{ 195, 255, 23 },
		{ 195, 256, 23 },
		{ 196, 227, 23 },
		{ 196, 228, 23 },
		{ 196, 229, 23 },
		{ 196, 230, 23 },
		{ 196, 231, 23 },
		{ 196, 232, 23 },
		{ 196, 233, 23 },
		{ 196, 234, 23 },
		{ 196, 235, 23 },
		{ 196, 236, 23 },
		{ 196, 237, 23 },
		{ 196, 238, 23 },
		{ 196, 239, 23 },
		{ 196, 240, 23 },
		{ 196, 241, 23 },
		{ 196, 242, 23 },
		{ 196, 243, 23 },
		{ 196, 244, 23 },
		{ 196, 245, 23 },
		{ 196, 246, 23 },
		{ 196, 247, 23 },
		{ 196, 248, 23 },
		{ 196, 249, 23 },
		{ 196, 250, 23 },
		{ 196, 251, 23 },
		{ 196, 252, 23 },
		{ 196, 253, 23 },
		{ 196, 254, 23 },
		{ 196, 255, 23 },
		{ 196, 256, 23 },
		{ 197, 228, 23 },
		{ 197, 229, 23 },
		{ 197, 230, 23 },
		{ 197, 231, 23 },
		{ 197, 232, 23 },
		{ 197, 233, 23 },
		{ 197, 234, 23 },
		{ 197, 235, 23 },
		{ 197, 236, 23 },
		{ 197, 237, 23 },
		{ 197, 238, 23 },
		{ 197, 239, 23 },
		{ 197, 240, 23 },
		{ 197, 241, 23 },
		{ 197, 242, 23 },
		{ 197, 243, 23 },
		{ 197, 244, 23 },
		{ 197, 245, 23 },
		{ 197, 246, 23 },
		{ 197, 247, 23 },
		{ 197, 248, 23 },
		{ 197, 249, 23 },
		{ 197, 250, 23 },
		{ 197, 251, 23 },
		{ 197, 252, 23 },
		{ 197, 253, 23 },
		{ 197, 254, 23 },
		{ 197, 255, 23 },
		{ 197, 256, 23 },
		{ 198, 228, 23 },
		{ 198, 229, 23 },
		{ 198, 230, 23 },
		{ 198, 231, 23 },
		{ 198, 232, 23 },
		{ 198, 233, 23 },
		{ 198, 234, 23 },
		{ 198, 235, 23 },
		{ 198, 236, 23 },
		{ 198, 237, 23 },
		{ 198, 238, 23 },
		{ 198, 239, 23 },
		{ 198, 240, 23 },
		{ 198, 241, 23 },
		{ 198, 242, 23 },
		{ 198, 243, 23 },
		{ 198, 244, 23 },
		{ 198, 245, 23 },
		{ 198, 246, 23 },
		{ 198, 247, 23 },
		{ 198, 248, 23 },
		{ 198, 249, 23 },
		{ 198, 250, 23 },
		{ 198, 251, 23 },
		{ 198, 252, 23 },
		{ 198, 253, 23 },
		{ 198, 254, 23 },
		{ 198, 255, 23 },
		{ 198, 256, 23 },
		{ 199, 229, 23 },
		{ 199, 230, 23 },
		{ 199, 231, 23 },
		{ 199, 232, 23 },
		{ 199, 233, 23 },
		{ 199, 234, 23 },
		{ 199, 235, 23 },
		{ 199, 236, 23 },
		{ 199, 237, 23 },
		{ 199, 238, 23 },
		{ 199, 239, 23 },
		{ 199, 240, 23 },
		{ 199, 241, 23 },
		{ 199, 242, 23 },
		{ 199, 243, 23 },
		{ 199, 244, 23 },
		{ 199, 245, 23 },
		{ 199, 246, 23 },
		{ 199, 247, 23 },
		{ 199, 248, 23 },
		{ 199, 249, 23 },
		{ 199, 250, 23 },
		{ 199, 251, 23 },
		{ 199, 252, 23 },
		{ 199, 253, 23 },
		{ 199, 254, 23 },
		{ 199, 255, 23 },
		{ 199, 256, 23 },
		{ 200, 229, 23 },
		{ 200, 230, 23 },
		{ 200, 231, 23 },
		{ 200, 232, 23 },
		{ 200, 233, 23 },
		{ 200, 234, 23 },
		{ 200, 235, 23 },
		{ 200, 236, 23 },
		{ 200, 237, 23 },
		{ 200, 238, 23 },
		{ 200, 239, 23 },
		{ 200, 240, 23 },
		{ 200, 241, 23 },
		{ 200, 242, 23 },
		{ 200, 243, 23 },
		{ 200, 244, 23 },
		{ 200, 245, 23 },
		{ 200, 246, 23 },
		{ 200, 247, 23 },
		{ 200, 248, 23 },
		{ 200, 249, 23 },
		{ 200, 250, 23 },
		{ 200, 251, 23 },
		{ 200, 252, 23 },
		{ 200, 253, 23 },
		{ 200, 254, 23 },
		{ 200, 255, 23 },
		{ 200, 256, 23 },
		{ 201, 230, 23 },
		{ 201, 231, 23 },
		{ 201, 232, 23 },
		{ 201, 233, 23 },
		{ 201, 234, 23 },
		{ 201, 235, 23 },
		{ 201, 236, 23 },
		{ 201, 237, 23 },
		{ 201, 238, 23 },
		{ 201, 239, 23 },
		{ 201, 240, 23 },
		{ 201, 241, 23 },
		{ 201, 242, 23 },
		{ 201, 243, 23 },
		{ 201, 244, 23 },
		{ 201, 245, 23 },
		{ 201, 246, 23 },
		{ 201, 247, 23 },
		{ 201, 248, 23 },
		{ 201, 249, 23 },
		{ 201, 250, 23 },
		{ 201, 251, 23 },
		{ 201, 252, 23 },
		{ 201, 253, 23 },
		{ 201, 254, 23 },
		{ 201, 255, 23 },
		{ 201, 256, 23 },
		{ 202, 230, 23 },
		{ 202, 231, 23 },
		{ 202, 232, 23 },
		{ 202, 233, 23 },
		{ 202, 234, 23 },
		{ 202, 235, 23 },
		{ 202, 236, 23 },
		{ 202, 237, 23 },
		{ 202, 238, 23 },
		{ 202, 239, 23 },
		{ 202, 240, 23 },
		{ 202, 241, 23 },
		{ 202, 242, 23 },
		{ 202, 243, 23 },
		{ 202, 244, 23 },
		{ 202, 245, 23 },
		{ 202, 246, 23 },
		{ 202, 247, 23 },
		{ 202, 248, 23 },
		{ 202, 249, 23 },
		{ 202, 250, 23 },
		{ 202, 251, 23 },
		{ 202, 252, 23 },
		{ 202, 253, 23 },
		{ 202, 254, 23 },
		{ 202, 255, 23 },
		{ 202, 256, 23 },
		{ 203, 231, 23 },
		{ 203, 232, 23 },
		{ 203, 233, 23 },
		{ 203, 234, 23 },
		{ 203, 235, 23 },
		{ 203, 236, 23 },
		{ 203, 237, 23 },
		{ 203, 238, 23 },
		{ 203, 239, 23 },
		{ 203, 240, 23 },
		{ 203, 241, 23 },
		{ 203, 242, 23 },
		{ 203, 243, 23 },
		{ 203, 244, 23 },
		{ 203, 245, 23 },
		{ 203, 246, 23 },
		{ 203, 247, 23 },
		{ 203, 248, 23 },
		{ 203, 249, 23 },
		{ 203, 250, 23 },
		{ 203, 251, 23 },
		{ 203, 252, 23 },
		{ 203, 253, 23 },
		{ 203, 254, 23 },
		{ 203, 255, 23 },
		{ 203, 256, 23 },
		{ 204, 231, 23 },
		{ 204, 232, 23 },
		{ 204, 233, 23 },
		{ 204, 234, 23 },
		{ 204, 235, 23 },
		{ 204, 236, 23 },
		{ 204, 237, 23 },
		{ 204, 238, 23 },
		{ 204, 239, 23 },
		{ 204, 240, 23 },
		{ 204, 241, 23 },
		{ 204, 242, 23 },
		{ 204, 243, 23 },
		{ 204, 244, 23 },
		{ 204, 245, 23 },
		{ 204, 246, 23 },
		{ 204, 247, 23 },
		{ 204, 248, 23 },
		{ 204, 249, 23 },
		{ 204, 250, 23 },
		{ 204, 251, 23 },
		{ 204, 252, 23 },
		{ 204, 253, 23 },
		{ 204, 254, 23 },
		{ 204, 255, 23 },
		{ 204, 256, 23 },
		{ 205, 232, 23 },
		{ 205, 233, 23 },
		{ 205, 234, 23 },
		{ 205, 235, 23 },
		{ 205, 236, 23 },
		{ 205, 237, 23 },
		{ 205, 238, 23 },
		{ 205, 239, 23 },
		{ 205, 240, 23 },
		{ 205, 241, 23 },
		{ 205, 242, 23 },
		{ 205, 243, 23 },
		{ 205, 244, 23 },
		{ 205, 245, 23 },
		{ 205, 246, 23 },
		{ 205, 247, 23 },
		{ 205, 248, 23 },
		{ 205, 249, 23 },
		{ 205, 250, 23 },
		{ 205, 251, 23 },
		{ 205, 252, 23 },
		{ 205, 253, 23 },
		{ 205, 254, 23 },
		{ 205, 255, 23 },
		{ 205, 256, 23 },
		{ 206, 232, 23 },
		{ 206, 233, 23 },
		{ 206, 234, 23 },
		{ 206, 235, 23 },
		{ 206, 236, 23 },
		{ 206, 237, 23 },
		{ 206, 238, 23 },
		{ 206, 239, 23 },
		{ 206, 240, 23 },
		{ 206, 241, 23 },
		{ 206, 242, 23 },
		{ 206, 243, 23 },
		{ 206, 244, 23 },
		{ 206, 245, 23 },
		{ 206, 246, 23 },
		{ 206, 247, 23 },
		{ 206, 248, 23 },
		{ 206, 249, 23 },
		{ 206, 250, 23 },
		{ 206, 251, 23 },
		{ 206, 252, 23 },
		{ 206, 253, 23 },
		{ 206, 254, 23 },
		{ 206, 255, 23 },
		{ 206, 256, 23 },
		{ 207, 233, 23 },
		{ 207, 234, 23 },
		{ 207, 235, 23 },
		{ 207, 236, 23 },
		{ 207, 237, 23 },
		{ 207, 238, 23 },
		{ 207, 239, 23 },
		{ 207, 240, 23 },
		{ 207, 241, 23 },
		{ 207, 242, 23 },
		{ 207, 243, 23 },
		{ 207, 244, 23 },
		{ 207, 245, 23 },
		{ 207, 246, 23 },
		{ 207, 247, 23 },
		{ 207, 248, 23 },
		{ 207, 249, 23 },
		{ 207, 250, 23 },
		{ 207, 251, 23 },
		{ 207, 252, 23 },
		{ 207, 253, 23 },
		{ 207, 254, 23 },
		{ 207, 255, 23 },
		{ 207, 256, 23 },
		{ 208, 233, 23 },
		{ 208, 234, 23 },
		{ 208, 235, 23 },
		{ 208, 236, 23 },
		{ 208, 237, 23 },
		{ 208, 238, 23 },
		{ 208, 239, 23 },
		{ 208, 240, 23 },
		{ 208, 241, 23 },
		{ 208, 242, 23 },
		{ 208, 243, 23 },
		{ 208, 244, 23 },
		{ 208, 245, 23 },
		{ 208, 246, 23 },
		{ 208, 247, 23 },
		{ 208, 248, 23 },
		{ 208, 249, 23 },
		{ 208, 250, 23 },
		{ 208, 251, 23 },
		{ 208, 252, 23 },
		{ 208, 253, 23 },
		{ 208, 254, 23 },
		{ 208, 255, 23 },
		{ 208, 256, 23 },
		{ 209, 234, 23 },
		{ 209, 235, 23 },
		{ 209, 236, 23 },
		{ 209, 237, 23 },
		{ 209, 238, 23 },
		{ 209, 239, 23 },
		{ 209, 240, 23 },
		{ 209, 241, 23 },
		{ 209, 242, 23 },
		{ 209, 243, 23 },
		{ 209, 244, 23 },
		{ 209, 245, 23 },
		{ 209, 246, 23 },
		{ 209, 247, 23 },
		{ 209, 248, 23 },
		{ 209, 249, 23 },
		{ 209, 250, 23 },
		{ 209, 251, 23 },
		{ 209, 252, 23 },
		{ 209, 253, 23 },
		{ 209, 254, 23 },
		{ 209, 255, 23 },
		{ 209, 256, 23 },
		{ 210, 234, 23 },
		{ 210, 235, 23 },
		{ 210, 236, 23 },
		{ 210, 237, 23 },
		{ 210, 238, 23 },
		{ 210, 239, 23 },
		{ 210, 240, 23 },
		{ 210, 241, 23 },
		{ 210, 242, 23 },
		{ 210, 243, 23 },
		{ 210, 244, 23 },
		{ 210, 245, 23 },
		{ 210, 246, 23 },
		{ 210, 247, 23 },
		{ 210, 248, 23 },
		{ 210, 249, 23 },
		{ 210, 250, 23 },
		{ 210, 251, 23 },
		{ 210, 252, 23 },
		{ 210, 253, 23 },
		{ 210, 254, 23 },
		{ 210, 255, 23 },
		{ 210, 256, 23 },
		{ 211, 235, 23 },
		{ 211, 236, 23 },
		{ 211, 237, 23 },
		{ 211, 238, 23 },
		{ 211, 239, 23 },
		{ 211, 240, 23 },
		{ 211, 241, 23 },
		{ 211, 242, 23 },
		{ 211, 243, 23 },
		{ 211, 244, 23 },
		{ 211, 245, 23 },
		{ 211, 246, 23 },
		{ 211, 247, 23 },
		{ 211, 248, 23 },
		{ 211, 249, 23 },
		{ 211, 250, 23 },
		{ 211, 251, 23 },
		{ 211, 252, 23 },
		{ 211, 253, 23 },
		{ 211, 254, 23 },
		{ 211, 255, 23 },
		{ 211, 256, 23 },
		{ 212, 235, 23 },
		{ 212, 236, 23 },
		{ 212, 237, 23 },
		{ 212, 238, 23 },
		{ 212, 239, 23 },
		{ 212, 240, 23 },
		{ 212, 241, 23 },
		{ 212, 242, 23 },
		{ 212, 243, 23 },
		{ 212, 244, 23 },
		{ 212, 245, 23 },
		{ 212, 246, 23 },
		{ 212, 247, 23 },
		{ 212, 248, 23 },
		{ 212, 249, 23 },
		{ 212, 250, 23 },
		{ 212, 251, 23 },
		{ 212, 252, 23 },
		{ 212, 253, 23 },
		{ 212, 254, 23 },
		{ 212, 255, 23 },
		{ 212, 256, 23 },
		{ 213, 236, 23 },
		{ 213, 237, 23 },
		{ 213, 238, 23 },
		{ 213, 239, 23 },
		{ 213, 240, 23 },
		{ 213, 241, 23 },
		{ 213, 242, 23 },
		{ 213, 243, 23 },
		{ 213, 244, 23 },
		{ 213, 245, 23 },
		{ 213, 246, 23 },
		{ 213, 247, 23 },
		{ 213, 248, 23 },
		{ 213, 249, 23 },
		{ 213, 250, 23 },
		{ 213, 251, 23 },
		{ 213, 252, 23 },
		{ 213, 253, 23 },
		{ 213, 254, 23 },
		{ 213, 255, 23 },
		{ 213, 256, 23 },
		{ 214, 236, 23 },
		{ 214, 237, 23 },
		{ 214, 238, 23 },
		{ 214, 239, 23 },
		{ 214, 240, 23 },
		{ 214, 241, 23 },
		{ 214, 242, 23 },
		{ 214, 243, 23 },
		{ 214, 244, 23 },
		{ 214, 245, 23 },
		{ 214, 246, 23 },
		{ 214, 247, 23 },
		{ 214, 248, 23 },
		{ 214, 249, 23 },
		{ 214, 250, 23 },
		{ 214, 251, 23 },
		{ 214, 252, 23 },
		{ 214, 253, 23 },
		{ 214, 254, 23 },
		{ 214, 255, 23 },
		{ 214, 256, 23 },
		{ 215, 237, 23 },
		{ 215, 238, 23 },
		{ 215, 239, 23 },
		{ 215, 240, 23 },
		{ 215, 241, 23 },
		{ 215, 242, 23 },
		{ 215, 243, 23 },
		{ 215, 244, 23 },
		{ 215, 245, 23 },
		{ 215, 246, 23 },
		{ 215, 247, 23 },
		{ 215, 248, 23 },
		{ 215, 249, 23 },
		{ 215, 250, 23 },
		{ 215, 251, 23 },
		{ 215, 252, 23 },
		{ 215, 253, 23 },
		{ 215, 254, 23 },
		{ 215, 255, 23 },
		{ 215, 256, 23 },
		{ 216, 237, 23 },
		{ 216, 238, 23 },
		{ 216, 239, 23 },
		{ 216, 240, 23 },
		{ 216, 241, 23 },
		{ 216, 242, 23 },
		{ 216, 243, 23 },
		{ 216, 244, 23 },
		{ 216, 245, 23 },
		{ 216, 246, 23 },
		{ 216, 247, 23 },
		{ 216, 248, 23 },
		{ 216, 249, 23 },
		{ 216, 250, 23 },
		{ 216, 251, 23 },
		{ 216, 252, 23 },
		{ 216, 253, 23 },
		{ 216, 254, 23 },
		{ 216, 255, 23 },
		{ 216, 256, 23 },
		{ 217, 238, 23 },
		{ 217, 239, 23 },
		{ 217, 240, 23 },
		{ 217, 241, 23 },
		{ 217, 242, 23 },
		{ 217, 243, 23 },
		{ 217, 244, 23 },
		{ 217, 245, 23 },
		{ 217, 246, 23 },
		{ 217, 247, 23 },
		{ 217, 248, 23 },
		{ 217, 249, 23 },
		{ 217, 250, 23 },
		{ 217, 251, 23 },
		{ 217, 252, 23 },
		{ 217, 253, 23 },
		{ 217, 254, 23 },
		{ 217, 255, 23 },
		{ 217, 256, 23 },
		{ 218, 238, 23 },
		{ 218, 239, 23 },
		{ 218, 240, 23 },
		{ 218, 241, 23 },
		{ 218, 242, 23 },
		{ 218, 243, 23 },
		{ 218, 244, 23 },
		{ 218, 245, 23 },
		{ 218, 246, 23 },
		{ 218, 247, 23 },
		{ 218, 248, 23 },
		{ 218, 249, 23 },
		{ 218, 250, 23 },
		{ 218, 251, 23 },
		{ 218, 252, 23 },
		{ 218, 253, 23 },
		{ 218, 254, 23 },
		{ 218, 255, 23 },
		{ 218, 256, 23 },
		{ 219, 239, 23 },
		{ 219, 240, 23 },
		{ 219, 241, 23 },
		{ 219, 242, 23 },
		{ 219, 243, 23 },
		{ 219, 244, 23 },
		{ 219, 245, 23 },
		{ 219, 246, 23 },
		{ 219, 247, 23 },
		{ 219, 248, 23 },
		{ 219, 249, 23 },
		{ 219, 250, 23 },
		{ 219, 251, 23 },
		{ 219, 252, 23 },
		{ 219, 253, 23 },
		{ 219, 254, 23 },
		{ 219, 255, 23 },
		{ 219, 256, 23 },
		{ 220, 239, 23 },
		{ 220, 240, 23 },
		{ 220, 241, 23 },
		{ 220, 242, 23 },
		{ 220, 243, 23 },
		{ 220, 244, 23 },
		{ 220, 245, 23 },
		{ 220, 246, 23 },
		{ 220, 247, 23 },
		{ 220, 248, 23 },
		{ 220, 249, 23 },
		{ 220, 250, 23 },
		{ 220, 251, 23 },
		{ 220, 252, 23 },
		{ 220, 253, 23 },
		{ 220, 254, 23 },
		{ 220, 255, 23 },
		{ 220, 256, 23 },
		{ 221, 240, 23 },
		{ 221, 241, 23 },
		{ 221, 242, 23 },
		{ 221, 243, 23 },
		{ 221, 244, 23 },
		{ 221, 245, 23 },
		{ 221, 246, 23 },
		{ 221, 247, 23 },
		{ 221, 248, 23 },
		{ 221, 249, 23 },
		{ 221, 250, 23 },
		{ 221, 251, 23 },
		{ 221, 252, 23 },
		{ 221, 253, 23 },
		{ 221, 254, 23 },
		{ 221, 255, 23 },
		{ 221, 256, 23 },
		{ 222, 240, 23 },
		{ 222, 241, 23 },
		{ 222, 242, 23 },
		{ 222, 243, 23 },
		{ 222, 244, 23 },
		{ 222, 245, 23 },
		{ 222, 246, 23 },
		{ 222, 247, 23 },
		{ 222, 248, 23 },
		{ 222, 249, 23 },
		{ 222, 250, 23 },
		{ 222, 251, 23 },
		{ 222, 252, 23 },
		{ 222, 253, 23 },
		{ 222, 254, 23 },
		{ 222, 255, 23 },
		{ 222, 256, 23 },
		{ 223, 241, 23 },
		{ 223, 242, 23 },
		{ 223, 243, 23 },
		{ 223, 244, 23 },
		{ 223, 245, 23 },
		{ 223, 246, 23 },
		{ 223, 247, 23 },
		{ 223, 248, 23 },
		{ 223, 249, 23 },
		{ 223, 250, 23 },
		{ 223, 251, 23 },
		{ 223, 252, 23 },
		{ 223, 253, 23 },
		{ 223, 254, 23 },
		{ 223, 255, 23 },
		{ 223, 256, 23 },
		{ 224, 241, 23 },
		{ 224, 242, 23 },
		{ 224, 243, 23 },
		{ 224, 244, 23 },
		{ 224, 245, 23 },
		{ 224, 246, 23 },
		{ 224, 247, 23 },
		{ 224, 248, 23 },
		{ 224, 249, 23 },
		{ 224, 250, 23 },
		{ 224, 251, 23 },
		{ 224, 252, 23 },
		{ 224, 253, 23 },
		{ 224, 254, 23 },
		{ 224, 255, 23 },
		{ 224, 256, 23 },
		{ 225, 242, 23 },
		{ 225, 243, 23 },
		{ 225, 244, 23 },
		{ 225, 245, 23 },
		{ 225, 246, 23 },
		{ 225, 247, 23 },
		{ 225, 248, 23 },
		{ 225, 249, 23 },
		{ 225, 250, 23 },
		{ 225, 251, 23 },
		{ 225, 252, 23 },
		{ 225, 253, 23 },
		{ 225, 254, 23 },
		{ 225, 255, 23 },
		{ 225, 256, 23 },
		{ 226, 242, 23 },
		{ 226, 243, 23 },
		{ 226, 244, 23 },
		{ 226, 245, 23 },
		{ 226, 246, 23 },
		{ 226, 247, 23 },
		{ 226, 248, 23 },
		{ 226, 249, 23 },
		{ 226, 250, 23 },
		{ 226, 251, 23 },
		{ 226, 252, 23 },
		{ 226, 253, 23 },
		{ 226, 254, 23 },
		{ 226, 255, 23 },
		{ 226, 256, 23 },
		{ 227, 243, 23 },
		{ 227, 244, 23 },
		{ 227, 245, 23 },
		{ 227, 246, 23 },
		{ 227, 247, 23 },
		{ 227, 248, 23 },
		{ 227, 249, 23 },
		{ 227, 250, 23 },
		{ 227, 251, 23 },
		{ 227, 252, 23 },
		{ 227, 253, 23 },
		{ 227, 254, 23 },
		{ 227, 255, 23 },
		{ 227, 256, 23 },
		{ 228, 243, 23 },
		{ 228, 244, 23 },
		{ 228, 245, 23 },
		{ 228, 246, 23 },
		{ 228, 247, 23 },
		{ 228, 248, 23 },
		{ 228, 249, 23 },
		{ 228, 250, 23 },
		{ 228, 251, 23 },
		{ 228, 252, 23 },
		{ 228, 253, 23 },
		{ 228, 254, 23 },
		{ 228, 255, 23 },
		{ 228, 256, 23 },
		{ 229, 244, 23 },
		{ 229, 245, 23 },
		{ 229, 246, 23 },
		{ 229, 247, 23 },
		{ 229, 248, 23 },
		{ 229, 249, 23 },
		{ 229, 250, 23 },
		{ 229, 251, 23 },
		{ 229, 252, 23 },
		{ 229, 253, 23 },
		{ 229, 254, 23 },
		{ 229, 255, 23 },
		{ 229, 256, 23 },
		{ 230, 244, 23 },
		{ 230, 245, 23 },
		{ 230, 246, 23 },
		{ 230, 247, 23 },
		{ 230, 248, 23 },
		{ 230, 249, 23 },
		{ 230, 250, 23 },
		{ 230, 251, 23 },
		{ 230, 252, 23 },
		{ 230, 253, 23 },
		{ 230, 254, 23 },
		{ 230, 255, 23 },
		{ 230, 256, 23 },
		{ 231, 245, 23 },
		{ 231, 246, 23 },
		{ 231, 247, 23 },
		{ 231, 248, 23 },
		{ 231, 249, 23 },
		{ 231, 250, 23 },
		{ 231, 251, 23 },
		{ 231, 252, 23 },
		{ 231, 253, 23 },
		{ 231, 254, 23 },
		{ 231, 255, 23 },
		{ 231, 256, 23 },
		{ 232, 245, 23 },
		{ 232, 246, 23 },
		{ 232, 247, 23 },
		{ 232, 248, 23 },
		{ 232, 249, 23 },
		{ 232, 250, 23 },
		{ 232, 251, 23 },
		{ 232, 252, 23 },
		{ 232, 253, 23 },
		{ 232, 254, 23 },
		{ 232, 255, 23 },
		{ 232, 256, 23 },
		{ 233, 246, 23 },
		{ 233, 247, 23 },
		{ 233, 248, 23 },
		{ 233, 249, 23 },
		{ 233, 250, 23 },
		{ 233, 251, 23 },
		{ 233, 252, 23 },
		{ 233, 253, 23 },
		{ 233, 254, 23 },
		{ 233, 255, 23 },
		{ 233, 256, 23 },
		{ 234, 246, 23 },
		{ 234, 247, 23 },
		{ 234, 248, 23 },
		{ 234, 249, 23 },
		{ 234, 250, 23 },
		{ 234, 251, 23 },
		{ 234, 252, 23 },
		{ 234, 253, 23 },
		{ 234, 254, 23 },
		{ 234, 255, 23 },
		{ 234, 256, 23 },
		{ 235, 247, 23 },
		{ 235, 248, 23 },
		{ 235, 249, 23 },
		{ 235, 250, 23 },
		{ 235, 251, 23 },
		{ 235, 252, 23 },
		{ 235, 253, 23 },
		{ 235, 254, 23 },
		{ 235, 255, 23 },
		{ 235, 256, 23 },
		{ 236, 247, 23 },
		{ 236, 248, 23 },
		{ 236, 249, 23 },
		{ 236, 250, 23 },
		{ 236, 251, 23 },
		{ 236, 252, 23 },
		{ 236, 253, 23 },
		{ 236, 254, 23 },
		{ 236, 255, 23 },
		{ 236, 256, 23 },
		{ 237, 248, 23 },
		{ 237, 249, 23 },
		{ 237, 250, 23 },
		{ 237, 251, 23 },
		{ 237, 252, 23 },
		{ 237, 253, 23 },
		{ 237, 254, 23 },
		{ 237, 255, 23 },
		{ 237, 256, 23 },
		{ 238, 248, 23 },
		{ 238, 249, 23 },
		{ 238, 250, 23 },
		{ 238, 251, 23 },
		{ 238, 252, 23 },
		{ 238, 253, 23 },
		{ 238, 254, 23 },
		{ 238, 255, 23 },
		{ 238, 256, 23 },
		{ 239, 249, 23 },
		{ 239, 250, 23 },
		{ 239, 251, 23 },
		{ 239, 252, 23 },
		{ 239, 253, 23 },
		{ 239, 254, 23 },
		{ 239, 255, 23 },
		{ 239, 256, 23 },
		{ 240, 249, 23 },
		{ 240, 250, 23 },
		{ 240, 251, 23 },
		{ 240, 252, 23 },
		{ 240, 253, 23 },
		{ 240, 254, 23 },
		{ 240, 255, 23 },
		{ 240, 256, 23 },
		{ 274, 258, 23 },
		{ 274, 259, 23 },
		{ 274, 260, 23 },
		{ 274, 261, 23 },
		{ 274, 262, 23 },
		{ 274, 263, 23 },
		{ 274, 264, 23 },
		{ 274, 265, 23 },
		{ 275, 258, 23 },
		{ 275, 259, 23 },
		{ 275, 260, 23 },
		{ 275, 261, 23 },
		{ 275, 262, 23 },
		{ 275, 263, 23 },
		{ 275, 264, 23 },
		{ 275, 265, 23 },
		{ 276, 258, 23 },
		{ 276, 259, 23 },
		{ 276, 260, 23 },
		{ 276, 261, 23 },
		{ 276, 262, 23 },
		{ 276, 263, 23 },
		{ 276, 264, 23 },
		{ 276, 265, 23 },
		{ 276, 266, 23 },
		{ 277, 258, 23 },
		{ 277, 259, 23 },
		{ 277, 260, 23 },
		{ 277, 261, 23 },
		{ 277, 262, 23 },
		{ 277, 263, 23 },
		{ 277, 264, 23 },
		{ 277, 265, 23 },
		{ 277, 266, 23 },
		{ 278, 258, 23 },
		{ 278, 259, 23 },
		{ 278, 260, 23 },
		{ 278, 261, 23 },
		{ 278, 262, 23 },
		{ 278, 263, 23 },
		{ 278, 264, 23 },
		{ 278, 265, 23 },
		{ 278, 266, 23 },
		{ 278, 267, 23 },
		{ 279, 258, 23 },
		{ 279, 259, 23 },
		{ 279, 260, 23 },
		{ 279, 261, 23 },
		{ 279, 262, 23 },
		{ 279, 263, 23 },
		{ 279, 264, 23 },
		{ 279, 265, 23 },
		{ 279, 266, 23 },
		{ 279, 267, 23 },
		{ 280, 258, 23 },
		{ 280, 259, 23 },
		{ 280, 260, 23 },
		{ 280, 261, 23 },
		{ 280, 262, 23 },
		{ 280, 263, 23 },
		{ 280, 264, 23 },
		{ 280, 265, 23 },
		{ 280, 266, 23 },
		{ 280, 267, 23 },
		{ 280, 268, 23 },
		{ 281, 258, 23 },
		{ 281, 259, 23 },
		{ 281, 260, 23 },
		{ 281, 261, 23 },
		{ 281, 262, 23 },
		{ 281, 263, 23 },
		{ 281, 264, 23 },
		{ 281, 265, 23 },
		{ 281, 266, 23 },
		{ 281, 267, 23 },
		{ 281, 268, 23 },
		{ 282, 258, 23 },
		{ 282, 259, 23 },
		{ 282, 260, 23 },
		{ 282, 261, 23 },
		{ 282, 262, 23 },
		{ 282, 263, 23 },
		{ 282, 264, 23 },
		{ 282, 265, 23 },
		{ 282, 266, 23 },
		{ 282, 267, 23 },
		{ 282, 268, 23 },
		{ 282, 269, 23 },
		{ 283, 258, 23 },
		{ 283, 259, 23 },
		{ 283, 260, 23 },
		{ 283, 261, 23 },
		{ 283, 262, 23 },
		{ 283, 263, 23 },
		{ 283, 264, 23 },
		{ 283, 265, 23 },
		{ 283, 266, 23 },
		{ 283, 267, 23 },
		{ 283, 268, 23 },
		{ 283, 269, 23 },
		{ 284, 258, 23 },
		{ 284, 259, 23 },
		{ 284, 260, 23 },
		{ 284, 261, 23 },
		{ 284, 262, 23 },
		{ 284, 263, 23 },
		{ 284, 264, 23 },
		{ 284, 265, 23 },
		{ 284, 266, 23 },
		{ 284, 267, 23 },
		{ 284, 268, 23 },
		{ 284, 269, 23 },
		{ 284, 270, 23 },
		{ 285, 258, 23 },
		{ 285, 259, 23 },
		{ 285, 260, 23 },
		{ 285, 261, 23 },
		{ 285, 262, 23 },
		{ 285, 263, 23 },
		{ 285, 264, 23 },
		{ 285, 265, 23 },
		{ 285, 266, 23 },
		{ 285, 267, 23 },
		{ 285, 268, 23 },
		{ 285, 269, 23 },
		{ 285, 270, 23 },
		{ 286, 258, 23 },
		{ 286, 259, 23 },
		{ 286, 260, 23 },
		{ 286, 261, 23 },
		{ 286, 262, 23 },
		{ 286, 263, 23 },
		{ 286, 264, 23 },
		{ 286, 265, 23 },
		{ 286, 266, 23 },
		{ 286, 267, 23 },
		{ 286, 268, 23 },
		{ 286, 269, 23 },
		{ 286, 270, 23 },
		{ 286, 271, 23 },
		{ 287, 258, 23 },
		{ 287, 259, 23 },
		{ 287, 260, 23 },
		{ 287, 261, 23 },
		{ 287, 262, 23 },
		{ 287, 263, 23 },
		{ 287, 264, 23 },
		{ 287, 265, 23 },
		{ 287, 266, 23 },
		{ 287, 267, 23 },
		{ 287, 268, 23 },
		{ 287, 269, 23 },
		{ 287, 270, 23 },
		{ 287, 271, 23 },
		{ 288, 258, 23 },
		{ 288, 259, 23 },
		{ 288, 260, 23 },
		{ 288, 261, 23 },
		{ 288, 262, 23 },
		{ 288, 263, 23 },
		{ 288, 264, 23 },
		{ 288, 265, 23 },
		{ 288, 266, 23 },
		{ 288, 267, 23 },
		{ 288, 268, 23 },
		{ 288, 269, 23 },
		{ 288, 270, 23 },
		{ 288, 271, 23 },
		{ 288, 272, 23 },
		{ 289, 258, 23 },
		{ 289, 259, 23 },
		{ 289, 260, 23 },
		{ 289, 261, 23 },
		{ 289, 262, 23 },
		{ 289, 263, 23 },
		{ 289, 264, 23 },
		{ 289, 265, 23 },
		{ 289, 266, 23 },
		{ 289, 267, 23 },
		{ 289, 268, 23 },
		{ 289, 269, 23 },
		{ 289, 270, 23 },
		{ 289, 271, 23 },
		{ 289, 272, 23 },
		{ 290, 258, 23 },
		{ 290, 259, 23 },
		{ 290, 260, 23 },
		{ 290, 261, 23 },
		{ 290, 262, 23 },
		{ 290, 263, 23 },
		{ 290, 264, 23 },
		{ 290, 265, 23 },
		{ 290, 266, 23 },
		{ 290, 267, 23 },
		{ 290, 268, 23 },
		{ 290, 269, 23 },
		{ 290, 270, 23 },
		{ 290, 271, 23 },
		{ 290, 272, 23 },
		{ 290, 273, 23 },
		{ 291, 258, 23 },
		{ 291, 259, 23 },
		{ 291, 260, 23 },
		{ 291, 261, 23 },
		{ 291, 262, 23 },
		{ 291, 263, 23 },
		{ 291, 264, 23 },
		{ 291, 265, 23 },
		{ 291, 266, 23 },
		{ 291, 267, 23 },
		{ 291, 268, 23 },
		{ 291, 269, 23 },
		{ 291, 270, 23 },
		{ 291, 271, 23 },
		{ 291, 272, 23 },
		{ 291, 273, 23 },
		{ 292, 258, 23 },
		{ 292, 259, 23 },
		{ 292, 260, 23 },
		{ 292, 261, 23 },
		{ 292, 262, 23 },
		{ 292, 263, 23 },
		{ 292, 264, 23 },
		{ 292, 265, 23 },
		{ 292, 266, 23 },
		{ 292, 267, 23 },
		{ 292, 268, 23 },
		{ 292, 269, 23 },
		{ 292, 270, 23 },
		{ 292, 271, 23 },
		{ 292, 272, 23 },
		{ 292, 273, 23 },
		{ 292, 274, 23 },
		{ 293, 258, 23 },
		{ 293, 259, 23 },
		{ 293, 260, 23 },
		{ 293, 261, 23 },
		{ 293, 262, 23 },
		{ 293, 263, 23 },
		{ 293, 264, 23 },
		{ 293, 265, 23 },
		{ 293, 266, 23 },
		{ 293, 267, 23 },
		{ 293, 268, 23 },
		{ 293, 269, 23 },
		{ 293, 270, 23 },
		{ 293, 271, 23 },
		{ 293, 272, 23 },
		{ 293, 273, 23 },
		{ 293, 274, 23 },
		{ 294, 258, 23 },
		{ 294, 259, 23 },
		{ 294, 260, 23 },
		{ 294, 261, 23 },
		{ 294, 262, 23 },
		{ 294, 263, 23 },
		{ 294, 264, 23 },
		{ 294, 265, 23 },
		{ 294, 266, 23 },
		{ 294, 267, 23 },
		{ 294, 268, 23 },
		{ 294, 269, 23 },
		{ 294, 270, 23 },
		{ 294, 271, 23 },
		{ 294, 272, 23 },
		{ 294, 273, 23 },
		{ 294, 274, 23 },
		{ 294, 275, 23 },
		{ 295, 258, 23 },
		{ 295, 259, 23 },
		{ 295, 260, 23 },
		{ 295, 261, 23 },
		{ 295, 262, 23 },
		{ 295, 263, 23 },
		{ 295, 264, 23 },
		{ 295, 265, 23 },
		{ 295, 266, 23 },
		{ 295, 267, 23 },
		{ 295, 268, 23 },
		{ 295, 269, 23 },
		{ 295, 270, 23 },
		{ 295, 271, 23 },
		{ 295, 272, 23 },
		{ 295, 273, 23 },
		{ 295, 274, 23 },
		{ 295, 275, 23 },
		{ 296, 258, 23 },
		{ 296, 259, 23 },
		{ 296, 260, 23 },
		{ 296, 261, 23 },
		{ 296, 262, 23 },
		{ 296, 263, 23 },
		{ 296, 264, 23 },
		{ 296, 265, 23 },
		{ 296, 266, 23 },
		{ 296, 267, 23 },
		{ 296, 268, 23 },
		{ 296, 269, 23 },
		{ 296, 270, 23 },
		{ 296, 271, 23 },
		{ 296, 272, 23 },
		{ 296, 273, 23 },
		{ 296, 274, 23 },
		{ 296, 275, 23 },
		{ 296, 276, 23 },
		{ 297, 258, 23 },
		{ 297, 259, 23 },
		{ 297, 260, 23 },
		{ 297, 261, 23 },
		{ 297, 262, 23 },
		{ 297, 263, 23 },
		{ 297, 264, 23 },
		{ 297, 265, 23 },
		{ 297, 266, 23 },
		{ 297, 267, 23 },
		{ 297, 268, 23 },
		{ 297, 269, 23 },
		{ 297, 270, 23 },
		{ 297, 271, 23 },
		{ 297, 272, 23 },
		{ 297, 273, 23 },
		{ 297, 274, 23 },
		{ 297, 275, 23 },
		{ 297, 276, 23 },
		{ 298, 258, 23 },
		{ 298, 259, 23 },
		{ 298, 260, 23 },
		{ 298, 261, 23 },
		{ 298, 262, 23 },
		{ 298, 263, 23 },
		{ 298, 264, 23 },
		{ 298, 265, 23 },
		{ 298, 266, 23 },
		{ 298, 267, 23 },
		{ 298, 268, 23 },
		{ 298, 269, 23 },
		{ 298, 270, 23 },
		{ 298, 271, 23 },
		{ 298, 272, 23 },
		{ 298, 273, 23 },
		{ 298, 274, 23 },
		{ 298, 275, 23 },
		{ 298, 276, 23 },
		{ 298, 277, 23 },
		{ 299, 258, 23 },
		{ 299, 259, 23 },
		{ 299, 260, 23 },
		{ 299, 261, 23 },
		{ 299, 262, 23 },
		{ 299, 263, 23 },
		{ 299, 264, 23 },
		{ 299, 265, 23 },
		{ 299, 266, 23 },
		{ 299, 267, 23 },
		{ 299, 268, 23 },
		{ 299, 269, 23 },
		{ 299, 270, 23 },
		{ 299, 271, 23 },
		{ 299, 272, 23 },
		{ 299, 273, 23 },
		{ 299, 274, 23 },
		{ 299, 275, 23 },
		{ 299, 276, 23 },
		{ 299, 277, 23 },
		{ 300, 258, 23 },
		{ 300, 259, 23 },
		{ 300, 260, 23 },
		{ 300, 261, 23 },
		{ 300, 262, 23 },
		{ 300, 263, 23 },
		{ 300, 264, 23 },
		{ 300, 265, 23 },
		{ 300, 266, 23 },
		{ 300, 267, 23 },
		{ 300, 268, 23 },
		{ 300, 269, 23 },
		{ 300, 270, 23 },
		{ 300, 271, 23 },
		{ 300, 272, 23 },
		{ 300, 273, 23 },
		{ 300, 274, 23 },
		{ 300, 275, 23 },
		{ 300, 276, 23 },
		{ 300, 277, 23 },
		{ 300, 278, 23 },
		{ 301, 258, 23 },
		{ 301, 259, 23 },
		{ 301, 260, 23 },
		{ 301, 261, 23 },
		{ 301, 262, 23 },
		{ 301, 263, 23 },
		{ 301, 264, 23 },
		{ 301, 265, 23 },
		{ 301, 266, 23 },
		{ 301, 267, 23 },
		{ 301, 268, 23 },
		{ 301, 269, 23 },
		{ 301, 270, 23 },
		{ 301, 271, 23 },
		{ 301, 272, 23 },
		{ 301, 273, 23 },
		{ 301, 274, 23 },
		{ 301, 275, 23 },
		{ 301, 276, 23 },
		{ 301, 277, 23 },
		{ 301, 278, 23 },
		{ 302, 258, 23 },
		{ 302, 259, 23 },
		{ 302, 260, 23 },
		{ 302, 261, 23 },
		{ 302, 262, 23 },
		{ 302, 263, 23 },
		{ 302, 264, 23 },
		{ 302, 265, 23 },
		{ 302, 266, 23 },
		{ 302, 267, 23 },
		{ 302, 268, 23 },
		{ 302, 269, 23 },
		{ 302, 270, 23 },
		{ 302, 271, 23 },
		{ 302, 272, 23 },
		{ 302, 273, 23 },
		{ 302, 274, 23 },
		{ 302, 275, 23 },
		{ 302, 276, 23 },
		{ 302, 277, 23 },
		{ 302, 278, 23 },
		{ 302, 279, 23 },
		{ 303, 258, 23 },
		{ 303, 259, 23 },
		{ 303, 260, 23 },
		{ 303, 261, 23 },
		{ 303, 262, 23 },
		{ 303, 263, 23 },
		{ 303, 264, 23 },
		{ 303, 265, 23 },
		{ 303, 266, 23 },
		{ 303, 267, 23 },
		{ 303, 268, 23 },
		{ 303, 269, 23 },
		{ 303, 270, 23 },
		{ 303, 271, 23 },
		{ 303, 272, 23 },
		{ 303, 273, 23 },
		{ 303, 274, 23 },
		{ 303, 275, 23 },
		{ 303, 276, 23 },
		{ 303, 277, 23 },
		{ 303, 278, 23 },
		{ 303, 279, 23 },
		{ 304, 258, 23 },
		{ 304, 259, 23 },
		{ 304, 260, 23 },
		{ 304, 261, 23 },
		{ 304, 262, 23 },
		{ 304, 263, 23 },
		{ 304, 264, 23 },
		{ 304, 265, 23 },
		{ 304, 266, 23 },
		{ 304, 267, 23 },
		{ 304, 268, 23 },
		{ 304, 269, 23 },
		{ 304, 270, 23 },
		{ 304, 271, 23 },
		{ 304, 272, 23 },
		{ 304, 273, 23 },
		{ 304, 274, 23 },
		{ 304, 275, 23 },
		{ 304, 276, 23 },
		{ 304, 277, 23 },
		{ 304, 278, 23 },
		{ 304, 279, 23 },
		{ 304, 280, 23 },
		{ 305, 258, 23 },
		{ 305, 259, 23 },
		{ 305, 260, 23 },
		{ 305, 261, 23 },
		{ 305, 262, 23 },
		{ 305, 263, 23 },
		{ 305, 264, 23 },
		{ 305, 265, 23 },
		{ 305, 266, 23 },
		{ 305, 267, 23 },
		{ 305, 268, 23 },
		{ 305, 269, 23 },
		{ 305, 270, 23 },
		{ 305, 271, 23 },
		{ 305, 272, 23 },
		{ 305, 273, 23 },
		{ 305, 274, 23 },
		{ 305, 275, 23 },
		{ 305, 276, 23 },
		{ 305, 277, 23 },
		{ 305, 278, 23 },
		{ 305, 279, 23 },
		{ 305, 280, 23 },
		{ 306, 258, 23 },
		{ 306, 259, 23 },
		{ 306, 260, 23 },
		{ 306, 261, 23 },
		{ 306, 262, 23 },
		{ 306, 263, 23 },
		{ 306, 264, 23 },
		{ 306, 265, 23 },
		{ 306, 266, 23 },
		{ 306, 267, 23 },
		{ 306, 268, 23 },
		{ 306, 269, 23 },
		{ 306, 270, 23 },
		{ 306, 271, 23 },
		{ 306, 272, 23 },
		{ 306, 273, 23 },
		{ 306, 274, 23 },
		{ 306, 275, 23 },
		{ 306, 276, 23 },
		{ 306, 277, 23 },
		{ 306, 278, 23 },
		{ 306, 279, 23 },
		{ 306, 280, 23 },
		{ 306, 281, 23 },
		{ 307, 258, 23 },
		{ 307, 259, 23 },
		{ 307, 260, 23 },
		{ 307, 261, 23 },
		{ 307, 262, 23 },
		{ 307, 263, 23 },
		{ 307, 264, 23 },
		{ 307, 265, 23 },
		{ 307, 266, 23 },
		{ 307, 267, 23 },
		{ 307, 268, 23 },
		{ 307, 269, 23 },
		{ 307, 270, 23 },
		{ 307, 271, 23 },
		{ 307, 272, 23 },
		{ 307, 273, 23 },
		{ 307, 274, 23 },
		{ 307, 275, 23 },
		{ 307, 276, 23 },
		{ 307, 277, 23 },
		{ 307, 278, 23 },
		{ 307, 279, 23 },
		{ 307, 280, 23 },
		{ 307, 281, 23 },
		{ 308, 258, 23 },
		{ 308, 259, 23 },
		{ 308, 260, 23 },
		{ 308, 261, 23 },
		{ 308, 262, 23 },
		{ 308, 263, 23 },
		{ 308, 264, 23 },
		{ 308, 265, 23 },
		{ 308, 266, 23 },
		{ 308, 267, 23 },
		{ 308, 268, 23 },
		{ 308, 269, 23 },
		{ 308, 270, 23 },
		{ 308, 271, 23 },
		{ 308, 272, 23 },
		{ 308, 273, 23 },
		{ 308, 274, 23 },
		{ 308, 275, 23 },
		{ 308, 276, 23 },
		{ 308, 277, 23 },
		{ 308, 278, 23 },
		{ 308, 279, 23 },
		{ 308, 280, 23 },
		{ 308, 281, 23 },
		{ 308, 282, 23 },
		{ 309, 258, 23 },
		{ 309, 259, 23 },
		{ 309, 260, 23 },
		{ 309, 261, 23 },
		{ 309, 262, 23 },
		{ 309, 263, 23 },
		{ 309, 264, 23 },
		{ 309, 265, 23 },
		{ 309, 266, 23 },
		{ 309, 267, 23 },
		{ 309, 268, 23 },
		{ 309, 269, 23 },
		{ 309, 270, 23 },
		{ 309, 271, 23 },
		{ 309, 272, 23 },
		{ 309, 273, 23 },
		{ 309, 274, 23 },
		{ 309, 275, 23 },
		{ 309, 276, 23 },
		{ 309, 277, 23 },
		{ 309, 278, 23 },
		{ 309, 279, 23 },
		{ 309, 280, 23 },
		{ 309, 281, 23 },
		{ 309, 282, 23 },
		{ 310, 258, 23 },
		{ 310, 259, 23 },
		{ 310, 260, 23 },
		{ 310, 261, 23 },
		{ 310, 262, 23 },
		{ 310, 263, 23 },
		{ 310, 264, 23 },
		{ 310, 265, 23 },
		{ 310, 266, 23 },
		{ 310, 267, 23 },
		{ 310, 268, 23 },
		{ 310, 269, 23 },
		{ 310, 270, 23 },
		{ 310, 271, 23 },
		{ 310, 272, 23 },
		{ 310, 273, 23 },
		{ 310, 274, 23 },
		{ 310, 275, 23 },
		{ 310, 276, 23 },
		{ 310, 277, 23 },
		{ 310, 278, 23 },
		{ 310, 279, 23 },
		{ 310, 280, 23 },
		{ 310, 281, 23 },
		{ 310, 282, 23 },
		{ 310, 283, 23 },
		{ 311, 258, 23 },
		{ 311, 259, 23 },
		{ 311, 260, 23 },
		{ 311, 261, 23 },
		{ 311, 262, 23 },
		{ 311, 263, 23 },
		{ 311, 264, 23 },
		{ 311, 265, 23 },
		{ 311, 266, 23 },
		{ 311, 267, 23 },
		{ 311, 268, 23 },
		{ 311, 269, 23 },
		{ 311, 270, 23 },
		{ 311, 271, 23 },
		{ 311, 272, 23 },
		{ 311, 273, 23 },
		{ 311, 274, 23 },
		{ 311, 275, 23 },
		{ 311, 276, 23 },
		{ 311, 277, 23 },
		{ 311, 278, 23 },
		{ 311, 279, 23 },
		{ 311, 280, 23 },
		{ 311, 281, 23 },
		{ 311, 282, 23 },
		{ 311, 283, 23 },
		{ 312, 258, 23 },
		{ 312, 259, 23 },
		{ 312, 260, 23 },
		{ 312, 261, 23 },
		{ 312, 262, 23 },
		{ 312, 263, 23 },
		{ 312, 264, 23 },
		{ 312, 265, 23 },
		{ 312, 266, 23 },
		{ 312, 267, 23 },
		{ 312, 268, 23 },
		{ 312, 269, 23 },
		{ 312, 270, 23 },
		{ 312, 271, 23 },
		{ 312, 272, 23 },
		{ 312, 273, 23 },
		{ 312, 274, 23 },
		{ 312, 275, 23 },
		{ 312, 276, 23 },
		{ 312, 277, 23 },
		{ 312, 278, 23 },
		{ 312, 279, 23 },
		{ 312, 280, 23 },
		{ 312, 281, 23 },
		{ 312, 282, 23 },
		{ 312, 283, 23 },
		{ 312, 284, 23 },
		{ 313, 258, 23 },
		{ 313, 259, 23 },
		{ 313, 260, 23 },
		{ 313, 261, 23 },
		{ 313, 262, 23 },
		{ 313, 263, 23 },
		{ 313, 264, 23 },
		{ 313, 265, 23 },
		{ 313, 266, 23 },
		{ 313, 267, 23 },
		{ 313, 268, 23 },
		{ 313, 269, 23 },
		{ 313, 270, 23 },
		{ 313, 271, 23 },
		{ 313, 272, 23 },
		{ 313, 273, 23 },
		{ 313, 274, 23 },
		{ 313, 275, 23 },
		{ 313, 276, 23 },
		{ 313, 277, 23 },
		{ 313, 278, 23 },
		{ 313, 279, 23 },
		{ 313, 280, 23 },
		{ 313, 281, 23 },
		{ 313, 282, 23 },
		{ 313, 283, 23 },
		{ 313, 284, 23 },
		{ 314, 258, 23 },
		{ 314, 259, 23 },
		{ 314, 260, 23 },
		{ 314, 261, 23 },
		{ 314, 262, 23 },
		{ 314, 263, 23 },
		{ 314, 264, 23 },
		{ 314, 265, 23 },
		{ 314, 266, 23 },
		{ 314, 267, 23 },
		{ 314, 268, 23 },
		{ 314, 269, 23 },
		{ 314, 270, 23 },
		{ 314, 271, 23 },
		{ 314, 272, 23 },
		{ 314, 273, 23 },
		{ 314, 274, 23 },
		{ 314, 275, 23 },
		{ 314, 276, 23 },
		{ 314, 277, 23 },
		{ 314, 278, 23 },
		{ 314, 279, 23 },
		{ 314, 280, 23 },
		{ 314, 281, 23 },
		{ 314, 282, 23 },
		{ 314, 283, 23 },
		{ 314, 284, 23 },
		{ 314, 285, 23 },
		{ 315, 258, 23 },
		{ 315, 259, 23 },
		{ 315, 260, 23 },
		{ 315, 261, 23 },
		{ 315, 262, 23 },
		{ 315, 263, 23 },
		{ 315, 264, 23 },
		{ 315, 265, 23 },
		{ 315, 266, 23 },
		{ 315, 267, 23 },
		{ 315, 268, 23 },
		{ 315, 269, 23 },
		{ 315, 270, 23 },
		{ 315, 271, 23 },
		{ 315, 272, 23 },
		{ 315, 273, 23 },
		{ 315, 274, 23 },
		{ 315, 275, 23 },
		{ 315, 276, 23 },
		{ 315, 277, 23 },
		{ 315, 278, 23 },
		{ 315, 279, 23 },
		{ 315, 280, 23 },
		{ 315, 281, 23 },
		{ 315, 282, 23 },
		{ 315, 283, 23 },
		{ 315, 284, 23 },
		{ 315, 285, 23 },
		{ 316, 258, 23 },
		{ 316, 259, 23 },
		{ 316, 260, 23 },
		{ 316, 261, 23 },
		{ 316, 262, 23 },
		{ 316, 263, 23 },
		{ 316, 264, 23 },
		{ 316, 265, 23 },
		{ 316, 266, 23 },
		{ 316, 267, 23 },
		{ 316, 268, 23 },
		{ 316, 269, 23 },
		{ 316, 270, 23 },
		{ 316, 271, 23 },
		{ 316, 272, 23 },
		{ 316, 273, 23 },
		{ 316, 274, 23 },
		{ 316, 275, 23 },
		{ 316, 276, 23 },
		{ 316, 277, 23 },
		{ 316, 278, 23 },
		{ 316, 279, 23 },
		{ 316, 280, 23 },
		{ 316, 281, 23 },
		{ 316, 282, 23 },
		{ 316, 283, 23 },
		{ 316, 284, 23 },
		{ 316, 285, 23 },
		{ 316, 286, 23 },
		{ 317, 258, 23 },
		{ 317, 259, 23 },
		{ 317, 260, 23 },
		{ 317, 261, 23 },
		{ 317, 262, 23 },
		{ 317, 263, 23 },
		{ 317, 264, 23 },
		{ 317, 265, 23 },
		{ 317, 266, 23 },
		{ 317, 267, 23 },
		{ 317, 268, 23 },
		{ 317, 269, 23 },
		{ 317, 270, 23 },
		{ 317, 271, 23 },
		{ 317, 272, 23 },
		{ 317, 273, 23 },
		{ 317, 274, 23 },
		{ 317, 275, 23 },
		{ 317, 276, 23 },
		{ 317, 277, 23 },
		{ 317, 278, 23 },
		{ 317, 279, 23 },
		{ 317, 280, 23 },
		{ 317, 281, 23 },
		{ 317, 282, 23 },
		{ 317, 283, 23 },
		{ 317, 284, 23 },
		{ 317, 285, 23 },
		{ 317, 286, 23 },
		{ 318, 258, 23 },
		{ 318, 259, 23 },
		{ 318, 260, 23 },
		{ 318, 261, 23 },
		{ 318, 262, 23 },
		{ 318, 263, 23 },
		{ 318, 264, 23 },
		{ 318, 265, 23 },
		{ 318, 266, 23 },
		{ 318, 267, 23 },
		{ 318, 268, 23 },
		{ 318, 269, 23 },
		{ 318, 270, 23 },
		{ 318, 271, 23 },
		{ 318, 272, 23 },
		{ 318, 273, 23 },
		{ 318, 274, 23 },
		{ 318, 275, 23 },
		{ 318, 276, 23 },
		{ 318, 277, 23 },
		{ 318, 278, 23 },
		{ 318, 279, 23 },
		{ 318, 280, 23 },
		{ 318, 281, 23 },
		{ 318, 282, 23 },
		{ 318, 283, 23 },
		{ 318, 284, 23 },
		{ 318, 285, 23 },
		{ 318, 286, 23 },
		{ 318, 287, 23 },
		{ 319, 258, 23 },
		{ 319, 259, 23 },
		{ 319, 260, 23 },
		{ 319, 261, 23 },
		{ 319, 262, 23 },
		{ 319, 263, 23 },
		{ 319, 264, 23 },
		{ 319, 265, 23 },
		{ 319, 266, 23 },
		{ 319, 267, 23 },
		{ 319, 268, 23 },
		{ 319, 269, 23 },
		{ 319, 270, 23 },
		{ 319, 271, 23 },
		{ 319, 272, 23 },
		{ 319, 273, 23 },
		{ 319, 274, 23 },
		{ 319, 275, 23 },
		{ 319, 276, 23 },
		{ 319, 277, 23 },
		{ 319, 278, 23 },
		{ 319, 279, 23 },
		{ 319, 280, 23 },
		{ 319, 281, 23 },
		{ 319, 282, 23 },
		{ 319, 283, 23 },
		{ 319, 284, 23 },
		{ 319, 285, 23 },
		{ 319, 286, 23 },
		{ 319, 287, 23 },
		{ 320, 258, 23 },
		{ 320, 259, 23 },
		{ 320, 260, 23 },
		{ 320, 261, 23 },
		{ 320, 262, 23 },
		{ 320, 263, 23 },
		{ 320, 264, 23 },
		{ 320, 265, 23 },
		{ 320, 266, 23 },
		{ 320, 267, 23 },
		{ 320, 268, 23 },
		{ 320, 269, 23 },
		{ 320, 270, 23 },
		{ 320, 271, 23 },
		{ 320, 272, 23 },
		{ 320, 273, 23 },
		{ 320, 274, 23 },
		{ 320, 275, 23 },
		{ 320, 276, 23 },
		{ 320, 277, 23 },
		{ 320, 278, 23 },
		{ 320, 279, 23 },
		{ 320, 280, 23 },
		{ 320, 281, 23 },
		{ 320, 282, 23 },
		{ 320, 283, 23 },
		{ 320, 284, 23 },
		{ 320, 285, 23 },
		{ 320, 286, 23 },
		{ 320, 287, 23 },
		{ 320, 288, 23 },
		{ 321, 258, 23 },
		{ 321, 259, 23 },
		{ 321, 260, 23 },
		{ 321, 261, 23 },
		{ 321, 262, 23 },
		{ 321, 263, 23 },
		{ 321, 264, 23 },
		{ 321, 265, 23 },
		{ 321, 266, 23 },
		{ 321, 267, 23 },
		{ 321, 268, 23 },
		{ 321, 269, 23 },
		{ 321, 270, 23 },
		{ 321, 271, 23 },
		{ 321, 272, 23 },
		{ 321, 273, 23 },
		{ 321, 274, 23 },
		{ 321, 275, 23 },
		{ 321, 276, 23 },
		{ 321, 277, 23 },
		{ 321, 278, 23 },
		{ 321, 279, 23 },
		{ 321, 280, 23 },
		{ 321, 281, 23 },
		{ 321, 282, 23 },
		{ 321, 283, 23 },
		{ 321, 284, 23 },
		{ 321, 285, 23 },
		{ 321, 286, 23 },
		{ 321, 287, 23 },
		{ 321, 288, 23 },
		{ 322, 258, 23 },
		{ 322, 259, 23 },
		{ 322, 260, 23 },
		{ 322, 261, 23 },
		{ 322, 262, 23 },
		{ 322, 263, 23 },
		{ 322, 264, 23 },
		{ 322, 265, 23 },
		{ 322, 266, 23 },
		{ 322, 267, 23 },
		{ 322, 268, 23 },
		{ 322, 269, 23 },
		{ 322, 270, 23 },
		{ 322, 271, 23 },
		{ 322, 272, 23 },
		{ 322, 273, 23 },
		{ 322, 274, 23 },
		{ 322, 275, 23 },
		{ 322, 276, 23 },
		{ 322, 277, 23 },
		{ 322, 278, 23 },
		{ 322, 279, 23 },
		{ 322, 280, 23 },
		{ 322, 281, 23 },
		{ 322, 282, 23 },
		{ 322, 283, 23 },
		{ 322, 284, 23 },
		{ 322, 285, 23 },
		{ 322, 286, 23 },
		{ 322, 287, 23 },
		{ 322, 288, 23 },
		{ 322, 289, 23 },
		{ 323, 258, 23 },
		{ 323, 259, 23 },
		{ 323, 260, 23 },
		{ 323, 261, 23 },
		{ 323, 262, 23 },
		{ 323, 263, 23 },
		{ 323, 264, 23 },
		{ 323, 265, 23 },
		{ 323, 266, 23 },
		{ 323, 267, 23 },
		{ 323, 268, 23 },
		{ 323, 269, 23 },
		{ 323, 270, 23 },
		{ 323, 271, 23 },
		{ 323, 272, 23 },
		{ 323, 273, 23 },
		{ 323, 274, 23 },
		{ 323, 275, 23 },
		{ 323, 276, 23 },
		{ 323, 277, 23 },
		{ 323, 278, 23 },
		{ 323, 279, 23 },
		{ 323, 280, 23 },
		{ 323, 281, 23 },
		{ 323, 282, 23 },
		{ 323, 283, 23 },
		{ 323, 284, 23 },
		{ 323, 285, 23 },
		{ 323, 286, 23 },
		{ 323, 287, 23 },
		{ 323, 288, 23 },
		{ 323, 289, 23 },
		{ 324, 258, 23 },
		{ 324, 259, 23 },
		{ 324, 260, 23 },
		{ 324, 261, 23 },
		{ 324, 262, 23 },
		{ 324, 263, 23 },
		{ 324, 264, 23 },
		{ 324, 265, 23 },
		{ 324, 266, 23 },
		{ 324, 267, 23 },
		{ 324, 268, 23 },
		{ 324, 269, 23 },
		{ 324, 270, 23 },
		{ 324, 271, 23 },
		{ 324, 272, 23 },
		{ 324, 273, 23 },
		{ 324, 274, 23 },
		{ 324, 275, 23 },
		{ 324, 276, 23 },
		{ 324, 277, 23 },
		{ 324, 278, 23 },
		{ 324, 279, 23 },
		{ 324, 280, 23 },
		{ 324, 281, 23 },
		{ 324, 282, 23 },
		{ 324, 283, 23 },
		{ 324, 284, 23 },
		{ 324, 285, 23 },
		{ 324, 286, 23 },
		{ 324, 287, 23 },
		{ 324, 288, 23 },
		{ 324, 289, 23 },
		{ 324, 290, 23 },
		{ 325, 258, 23 },
		{ 325, 259, 23 },
		{ 325, 260, 23 },
		{ 325, 261, 23 },
		{ 325, 262, 23 },
		{ 325, 263, 23 },
		{ 325, 264, 23 },
		{ 325, 265, 23 },
		{ 325, 266, 23 },
		{ 325, 267, 23 },
		{ 325, 268, 23 },
		{ 325, 269, 23 },
		{ 325, 270, 23 },
		{ 325, 271, 23 },
		{ 325, 272, 23 },
		{ 325, 273, 23 },
		{ 325, 274, 23 },
		{ 325, 275, 23 },
		{ 325, 276, 23 },
		{ 325, 277, 23 },
		{ 325, 278, 23 },
		{ 325, 279, 23 },
		{ 325, 280, 23 },
		{ 325, 281, 23 },
		{ 325, 282, 23 },
		{ 325, 283, 23 },
		{ 325, 284, 23 },
		{ 325, 285, 23 },
		{ 325, 286, 23 },
		{ 325, 287, 23 },
		{ 325, 288, 23 },
		{ 325, 289, 23 },
		{ 325, 290, 23 },
		{ 326, 258, 23 },
		{ 326, 259, 23 },
		{ 326, 260, 23 },
		{ 326, 261, 23 },
		{ 326, 262, 23 },
		{ 326, 263, 23 },
		{ 326, 264, 23 },
		{ 326, 265, 23 },
		{ 326, 266, 23 },
		{ 326, 267, 23 },
		{ 326, 268, 23 },
		{ 326, 269, 23 },
		{ 326, 270, 23 },
		{ 326, 271, 23 },
		{ 326, 272, 23 },
		{ 326, 273, 23 },
		{ 326, 274, 23 },
		{ 326, 275, 23 },
		{ 326, 276, 23 },
		{ 326, 277, 23 },
		{ 326, 278, 23 },
		{ 326, 279, 23 },
		{ 326, 280, 23 },
		{ 326, 281, 23 },
		{ 326, 282, 23 },
		{ 326, 283, 23 },
		{ 326, 284, 23 },
		{ 326, 285, 23 },
		{ 326, 286, 23 },
		{ 326, 287, 23 },
		{ 326, 288, 23 },
		{ 326, 289, 23 },
		{ 326, 290, 23 },
		{ 326, 291, 23 },
		{ 327, 258, 23 },
		{ 327, 259, 23 },
		{ 327, 260, 23 },
		{ 327, 261, 23 },
		{ 327, 262, 23 },
		{ 327, 263, 23 },
		{ 327, 264, 23 },
		{ 327, 265, 23 },
		{ 327, 266, 23 },
		{ 327, 267, 23 },
		{ 327, 268, 23 },
		{ 327, 269, 23 },
		{ 327, 270, 23 },
		{ 327, 271, 23 },
		{ 327, 272, 23 },
		{ 327, 273, 23 },
		{ 327, 274, 23 },
		{ 327, 275, 23 },
		{ 327, 276, 23 },
		{ 327, 277, 23 },
		{ 327, 278, 23 },
		{ 327, 279, 23 },
		{ 327, 280, 23 },
		{ 327, 281, 23 },
		{ 327, 282, 23 },
		{ 327, 283, 23 },
		{ 327, 284, 23 },
		{ 327, 285, 23 },
		{ 327, 286, 23 },
		{ 327, 287, 23 },
		{ 327, 288, 23 },
		{ 327, 289, 23 },
		{ 327, 290, 23 },
		{ 327, 291, 23 },
		{ 328, 258, 23 },
		{ 328, 259, 23 },
		{ 328, 260, 23 },
		{ 328, 261, 23 },
		{ 328, 262, 23 },
		{ 328, 263, 23 },
		{ 328, 264, 23 },
		{ 328, 265, 23 },
		{ 328, 266, 23 },
		{ 328, 267, 23 },
		{ 328, 268, 23 },
		{ 328, 269, 23 },
		{ 328, 270, 23 },
		{ 328, 271, 23 },
		{ 328, 272, 23 },
		{ 328, 273, 23 },
		{ 328, 274, 23 },
		{ 328, 275, 23 },
		{ 328, 276, 23 },
		{ 328, 277, 23 },
		{ 328, 278, 23 },
		{ 328, 279, 23 },
		{ 328, 280, 23 },
		{ 328, 281, 23 },
		{ 328, 282, 23 },
		{ 328, 283, 23 },
		{ 328, 284, 23 },
		{ 328, 285, 23 },
		{ 328, 286, 23 },
		{ 328, 287, 23 },
		{ 328, 288, 23 },
		{ 328, 289, 23 },
		{ 328, 290, 23 },
		{ 328, 291, 23 },
		{ 328, 292, 23 },
		{ 329, 258, 23 },
		{ 329, 259, 23 },
		{ 329, 260, 23 },
		{ 329, 261, 23 },
		{ 329, 262, 23 },
		{ 329, 263, 23 },
		{ 329, 264, 23 },
		{ 329, 265, 23 },
		{ 329, 266, 23 },
		{ 329, 267, 23 },
		{ 329, 268, 23 },
		{ 329, 269, 23 },
		{ 329, 270, 23 },
		{ 329, 271, 23 },
		{ 329, 272, 23 },
		{ 329, 273, 23 },
		{ 329, 274, 23 },
		{ 329, 275, 23 },
		{ 329, 276, 23 },
		{ 329, 277, 23 },
		{ 329, 278, 23 },
		{ 329, 279, 23 },
		{ 329, 280, 23 },
		{ 329, 281, 23 },
		{ 329, 282, 23 },
		{ 329, 283, 23 },
		{ 329, 284, 23 },
		{ 329, 285, 23 },
		{ 329, 286, 23 },
		{ 329, 287, 23 },
		{ 329, 288, 23 },
		{ 329, 289, 23 },
		{ 329, 290, 23 },
		{ 329, 291, 23 },
		{ 329, 292, 23 },
		{ 330, 258, 23 },
		{ 330, 259, 23 },
		{ 330, 260, 23 },
		{ 330, 261, 23 },
		{ 330, 262, 23 },
		{ 330, 263, 23 },
		{ 330, 264, 23 },
		{ 330, 265, 23 },
		{ 330, 266, 23 },
		{ 330, 267, 23 },
		{ 330, 268, 23 },
		{ 330, 269, 23 },
		{ 330, 270, 23 },
		{ 330, 271, 23 },
		{ 330, 272, 23 },
		{ 330, 273, 23 },
		{ 330, 274, 23 },
		{ 330, 275, 23 },
		{ 330, 276, 23 },
		{ 330, 277, 23 },
		{ 330, 278, 23 },
		{ 330, 279, 23 },
		{ 330, 280, 23 },
		{ 330, 281, 23 },
		{ 330, 282, 23 },
		{ 330, 283, 23 },
		{ 330, 284, 23 },
		{ 330, 285, 23 },
		{ 330, 286, 23 },
		{ 330, 287, 23 },
		{ 330, 288, 23 },
		{ 330, 289, 23 },
		{ 330, 290, 23 },
		{ 330, 291, 23 },
		{ 330, 292, 23 },
		{ 330, 293, 23 },
		{ 331, 258, 23 },
		{ 331, 259, 23 },
		{ 331, 260, 23 },
		{ 331, 261, 23 },
		{ 331, 262, 23 },
		{ 331, 263, 23 },
		{ 331, 264, 23 },
		{ 331, 265, 23 },
		{ 331, 266, 23 },
		{ 331, 267, 23 },
		{ 331, 268, 23 },
		{ 331, 269, 23 },
		{ 331, 270, 23 },
		{ 331, 271, 23 },
		{ 331, 272, 23 },
		{ 331, 273, 23 },
		{ 331, 274, 23 },
		{ 331, 275, 23 },
		{ 331, 276, 23 },
		{ 331, 277, 23 },
		{ 331, 278, 23 },
		{ 331, 279, 23 },
		{ 331, 280, 23 },
		{ 331, 281, 23 },
		{ 331, 282, 23 },
		{ 331, 283, 23 },
		{ 331, 284, 23 },
		{ 331, 285, 23 },
		{ 331, 286, 23 },
		{ 331, 287, 23 },
		{ 331, 288, 23 },
		{ 331, 289, 23 },
		{ 331, 290, 23 },
		{ 331, 291, 23 },
		{ 331, 292, 23 },
		{ 331, 293, 23 },
		{ 332, 258, 23 },
		{ 332, 259, 23 },
		{ 332, 260, 23 },
		{ 332, 261, 23 },
		{ 332, 262, 23 },
		{ 332, 263, 23 },
		{ 332, 264, 23 },
		{ 332, 265, 23 },
		{ 332, 266, 23 },
		{ 332, 267, 23 },
		{ 332, 268, 23 },
		{ 332, 269, 23 },
		{ 332, 270, 23 },
		{ 332, 271, 23 },
		{ 332, 272, 23 },
		{ 332, 273, 23 },
		{ 332, 274, 23 },
		{ 332, 275, 23 },
		{ 332, 276, 23 },
		{ 332, 277, 23 },
		{ 332, 278, 23 },
		{ 332, 279, 23 },
		{ 332, 280, 23 },
		{ 332, 281, 23 },
		{ 332, 282, 23 },
		{ 332, 283, 23 },
		{ 332, 284, 23 },
		{ 332, 285, 23 },
		{ 332, 286, 23 },
		{ 332, 287, 23 },
		{ 332, 288, 23 },
		{ 332, 289, 23 },
		{ 332, 290, 23 },
		{ 332, 291, 23 },
		{ 332, 292, 23 },
		{ 332, 293, 23 },
		{ 332, 294, 23 },
		{ 333, 258, 23 },
		{ 333, 259, 23 },
		{ 333, 260, 23 },
		{ 333, 261, 23 },
		{ 333, 262, 23 },
		{ 333, 263, 23 },
		{ 333, 264, 23 },
		{ 333, 265, 23 },
		{ 333, 266, 23 },
		{ 333, 267, 23 },
		{ 333, 268, 23 },
		{ 333, 269, 23 },
		{ 333, 270, 23 },
		{ 333, 271, 23 },
		{ 333, 272, 23 },
		{ 333, 273, 23 },
		{ 333, 274, 23 },
		{ 333, 275, 23 },
		{ 333, 276, 23 },
		{ 333, 277, 23 },
		{ 333, 278, 23 },
		{ 333, 279, 23 },
		{ 333, 280, 23 },
		{ 333, 281, 23 },
		{ 333, 282, 23 },
		{ 333, 283, 23 },
		{ 333, 284, 23 },
		{ 333, 285, 23 },
		{ 333, 286, 23 },
		{ 333, 287, 23 },
		{ 333, 288, 23 },
		{ 333, 289, 23 },
		{ 333, 290, 23 },
		{ 333, 291, 23 },
		{ 333, 292, 23 },
		{ 333, 293, 23 },
		{ 333, 294, 23 },
		{ 334, 258, 23 },
		{ 334, 259, 23 },
		{ 334, 260, 23 },
		{ 334, 261, 23 },
		{ 334, 262, 23 },
		{ 334, 263, 23 },
		{ 334, 264, 23 },
		{ 334, 265, 23 },
		{ 334, 266, 23 },
		{ 334, 267, 23 },
		{ 334, 268, 23 },
		{ 334, 269, 23 },
		{ 334, 270, 23 },
		{ 334, 271, 23 },
		{ 334, 272, 23 },
		{ 334, 273, 23 },
		{ 334, 274, 23 },
		{ 334, 275, 23 },
		{ 334, 276, 23 },
		{ 334, 277, 23 },
		{ 334, 278, 23 },
		{ 334, 279, 23 },
		{ 334, 280, 23 },
		{ 334, 281, 23 },
		{ 334, 282, 23 },
		{ 334, 283, 23 },
		{ 334, 284, 23 },
		{ 334, 285, 23 },
		{ 334, 286, 23 },
		{ 334, 287, 23 },
		{ 334, 288, 23 },
		{ 334, 289, 23 },
		{ 334, 290, 23 },
		{ 334, 291, 23 },
		{ 334, 292, 23 },
		{ 334, 293, 23 },
		{ 334, 294, 23 },
		{ 334, 295, 23 },
		{ 335, 258, 23 },
		{ 335, 259, 23 },
		{ 335, 260, 23 },
		{ 335, 261, 23 },
		{ 335, 262, 23 },
		{ 335, 263, 23 },
		{ 335, 264, 23 },
		{ 335, 265, 23 },
		{ 335, 266, 23 },
		{ 335, 267, 23 },
		{ 335, 268, 23 },
		{ 335, 269, 23 },
		{ 335, 270, 23 },
		{ 335, 271, 23 },
		{ 335, 272, 23 },
		{ 335, 273, 23 },
		{ 335, 274, 23 },
		{ 335, 275, 23 },
		{ 335, 276, 23 },
		{ 335, 277, 23 },
		{ 335, 278, 23 },
		{ 335, 279, 23 },
		{ 335, 280, 23 },
		{ 335, 281, 23 },
		{ 335, 282, 23 },
		{ 335, 283, 23 },
		{ 335, 284, 23 },
		{ 335, 285, 23 },
		{ 335, 286, 23 },
		{ 335, 287, 23 },
		{ 335, 288, 23 },
		{ 335, 289, 23 },
		{ 335, 290, 23 },
		{ 335, 291, 23 },
		{ 335, 292, 23 },
		{ 335, 293, 23 },
		{ 335, 294, 23 },
		{ 335, 295, 23 },
		{ 336, 258, 23 },
		{ 336, 259, 23 },
		{ 336, 260, 23 },
		{ 336, 261, 23 },
		{ 336, 262, 23 },
		{ 336, 263, 23 },
		{ 336, 264, 23 },
		{ 336, 265, 23 },
		{ 336, 266, 23 },
		{ 336, 267, 23 },
		{ 336, 268, 23 },
		{ 336, 269, 23 },
		{ 336, 270, 23 },
		{ 336, 271, 23 },
		{ 336, 272, 23 },
		{ 336, 273, 23 },
		{ 336, 274, 23 },
		{ 336, 275, 23 },
		{ 336, 276, 23 },
		{ 336, 277, 23 },
		{ 336, 278, 23 },
		{ 336, 279, 23 },
		{ 336, 280, 23 },
		{ 336, 281, 23 },
		{ 336, 282, 23 },
		{ 336, 283, 23 },
		{ 336, 284, 23 },
		{ 336, 285, 23 },
		{ 336, 286, 23 },
		{ 336, 287, 23 },
		{ 336, 288, 23 },
		{ 336, 289, 23 },
		{ 336, 290, 23 },
		{ 336, 291, 23 },
		{ 336, 292, 23 },
		{ 336, 293, 23 },
		{ 336, 294, 23 },
		{ 336, 295, 23 },
		{ 336, 296, 23 },
		{ 337, 258, 23 },
		{ 337, 259, 23 },
		{ 337, 260, 23 },
		{ 337, 261, 23 },
		{ 337, 262, 23 },
		{ 337, 263, 23 },
		{ 337, 264, 23 },
		{ 337, 265, 23 },
		{ 337, 266, 23 },
		{ 337, 267, 23 },
		{ 337, 268, 23 },
		{ 337, 269, 23 },
		{ 337, 270, 23 },
		{ 337, 271, 23 },
		{ 337, 272, 23 },
		{ 337, 273, 23 },
		{ 337, 274, 23 },
		{ 337, 275, 23 },
		{ 337, 276, 23 },
		{ 337, 277, 23 },
		{ 337, 278, 23 },
		{ 337, 279, 23 },
		{ 337, 280, 23 },
		{ 337, 281, 23 },
		{ 337, 282, 23 },
		{ 337, 283, 23 },
		{ 337, 284, 23 },
		{ 337, 285, 23 },
		{ 337, 286, 23 },
		{ 337, 287, 23 },
		{ 337, 288, 23 },
		{ 337, 289, 23 },
		{ 337, 290, 23 },
		{ 337, 291, 23 },
		{ 337, 292, 23 },
		{ 337, 293, 23 },
		{ 337, 294, 23 },
		{ 337, 295, 23 },
		{ 337, 296, 23 },
		{ 338, 258, 23 },
		{ 338, 259, 23 },
		{ 338, 260, 23 },
		{ 338, 261, 23 },
		{ 338, 262, 23 },
		{ 338, 263, 23 },
		{ 338, 264, 23 },
		{ 338, 265, 23 },
		{ 338, 266, 23 },
		{ 338, 267, 23 },
		{ 338, 268, 23 },
		{ 338, 269, 23 },
		{ 338, 270, 23 },
		{ 338, 271, 23 },
		{ 338, 272, 23 },
		{ 338, 273, 23 },
		{ 338, 274, 23 },
		{ 338, 275, 23 },
		{ 338, 276, 23 },
		{ 338, 277, 23 },
		{ 338, 278, 23 },
		{ 338, 279, 23 },
		{ 338, 280, 23 },
		{ 338, 281, 23 },
		{ 338, 282, 23 },
		{ 338, 283, 23 },
		{ 338, 284, 23 },
		{ 338, 285, 23 },
		{ 338, 286, 23 },
		{ 338, 287, 23 },
		{ 338, 288, 23 },
		{ 338, 289, 23 },
		{ 338, 290, 23 },
		{ 338, 291, 23 },
		{ 338, 292, 23 },
		{ 338, 293, 23 },
		{ 338, 294, 23 },
		{ 338, 295, 23 },
		{ 338, 296, 23 },
		{ 338, 297, 23 },
		{ 339, 258, 23 },
		{ 339, 259, 23 },
		{ 339, 260, 23 },
		{ 339, 261, 23 },
		{ 339, 262, 23 },
		{ 339, 263, 23 },
		{ 339, 264, 23 },
		{ 339, 265, 23 },
		{ 339, 266, 23 },
		{ 339, 267, 23 },
		{ 339, 268, 23 },
		{ 339, 269, 23 },
		{ 339, 270, 23 },
		{ 339, 271, 23 },
		{ 339, 272, 23 },
		{ 339, 273, 23 },
		{ 339, 274, 23 },
		{ 339, 275, 23 },
		{ 339, 276, 23 },
		{ 339, 277, 23 },
		{ 339, 278, 23 },
		{ 339, 279, 23 },
		{ 339, 280, 23 },
		{ 339, 281, 23 },
		{ 339, 282, 23 },
		{ 339, 283, 23 },
		{ 339, 284, 23 },
		{ 339, 285, 23 },
		{ 339, 286, 23 },
		{ 339, 287, 23 },
		{ 339, 288, 23 },
		{ 339, 289, 23 },
		{ 339, 290, 23 },
		{ 339, 291, 23 },
		{ 339, 292, 23 },
		{ 339, 293, 23 },
		{ 339, 294, 23 },
		{ 339, 295, 23 },
		{ 339, 296, 23 },
		{ 339, 297, 23 },
		{ 340, 258, 23 },
		{ 340, 259, 23 },
		{ 340, 260, 23 },
		{ 340, 261, 23 },
		{ 340, 262, 23 },
		{ 340, 263, 23 },
		{ 340, 264, 23 },
		{ 340, 265, 23 },
		{ 340, 266, 23 },
		{ 340, 267, 23 },
		{ 340, 268, 23 },
		{ 340, 269, 23 },
		{ 340, 270, 23 },
		{ 340, 271, 23 },
		{ 340, 272, 23 },
		{ 340, 273, 23 },
		{ 340, 274, 23 },
		{ 340, 275, 23 },
		{ 340, 276, 23 },
		{ 340, 277, 23 },
		{ 340, 278, 23 },
		{ 340, 279, 23 },
		{ 340, 280, 23 },
		{ 340, 281, 23 },
		{ 340, 282, 23 },
		{ 340, 283, 23 },
		{ 340, 284, 23 },
		{ 340, 285, 23 },
		{ 340, 286, 23 },
		{ 340, 287, 23 },
		{ 340, 288, 23 },
		{ 340, 289, 23 },
		{ 340, 290, 23 },
		{ 340, 291, 23 },
		{ 340, 292, 23 },
		{ 340, 293, 23 },
		{ 340, 294, 23 },
		{ 340, 295, 23 },
		{ 340, 296, 23 },
		{ 340, 297, 23 },
		{ 340, 298, 23 },
		{ 341, 258, 23 },
		{ 341, 259, 23 },
		{ 341, 260, 23 },
		{ 341, 261, 23 },
		{ 341, 262, 23 },
		{ 341, 263, 23 },
		{ 341, 264, 23 },
		{ 341, 265, 23 },
		{ 341, 266, 23 },
		{ 341, 267, 23 },
		{ 341, 268, 23 },
		{ 341, 269, 23 },
		{ 341, 270, 23 },
		{ 341, 271, 23 },
		{ 341, 272, 23 },
		{ 341, 273, 23 },
		{ 341, 274, 23 },
		{ 341, 275, 23 },
		{ 341, 276, 23 },
		{ 341, 277, 23 },
		{ 341, 278, 23 },
		{ 341, 279, 23 },
		{ 341, 280, 23 },
		{ 341, 281, 23 },
		{ 341, 282, 23 },
		{ 341, 283, 23 },
		{ 341, 284, 23 },
		{ 341, 285, 23 },
		{ 341, 286, 23 },
		{ 341, 287, 23 },
		{ 341, 288, 23 },
		{ 341, 289, 23 },
		{ 341, 290, 23 },
		{ 341, 291, 23 },
		{ 341, 292, 23 },
		{ 341, 293, 23 },
		{ 341, 294, 23 },
		{ 341, 295, 23 },
		{ 341, 296, 23 },
		{ 341, 297, 23 },
		{ 341, 298, 23 },
		{ 342, 258, 23 },
		{ 342, 259, 23 },
		{ 342, 260, 23 },
		{ 342, 261, 23 },
		{ 342, 262, 23 },
		{ 342, 263, 23 },
		{ 342, 264, 23 },
		{ 342, 265, 23 },
		{ 342, 266, 23 },
		{ 342, 267, 23 },
		{ 342, 268, 23 },
		{ 342, 269, 23 },
		{ 342, 270, 23 },
		{ 342, 271, 23 },
		{ 342, 272, 23 },
		{ 342, 273, 23 },
		{ 342, 274, 23 },
		{ 342, 275, 23 },
		{ 342, 276, 23 },
		{ 342, 277, 23 },
		{ 342, 278, 23 },
		{ 342, 279, 23 },
		{ 342, 280, 23 },
		{ 342, 281, 23 },
		{ 342, 282, 23 },
		{ 342, 283, 23 },
		{ 342, 284, 23 },
		{ 342, 285, 23 },
		{ 342, 286, 23 },
		{ 342, 287, 23 },
		{ 342, 288, 23 },
		{ 342, 289, 23 },
		{ 342, 290, 23 },
		{ 342, 291, 23 },
		{ 342, 292, 23 },
		{ 342, 293, 23 },
		{ 342, 294, 23 },
		{ 342, 295, 23 },
		{ 342, 296, 23 },
		{ 342, 297, 23 },
		{ 342, 298, 23 },
		{ 342, 299, 23 },
		{ 343, 258, 23 },
		{ 343, 259, 23 },
		{ 343, 260, 23 },
		{ 343, 261, 23 },
		{ 343, 262, 23 },
		{ 343, 263, 23 },
		{ 343, 264, 23 },
		{ 343, 265, 23 },
		{ 343, 266, 23 },
		{ 343, 267, 23 },
		{ 343, 268, 23 },
		{ 343, 269, 23 },
		{ 343, 270, 23 },
		{ 343, 271, 23 },
		{ 343, 272, 23 },
		{ 343, 273, 23 },
		{ 343, 274, 23 },
		{ 343, 275, 23 },
		{ 343, 276, 23 },
		{ 343, 277, 23 },
		{ 343, 278, 23 },
		{ 343, 279, 23 },
		{ 343, 280, 23 },
		{ 343, 281, 23 },
		{ 343, 282, 23 },
		{ 343, 283, 23 },
		{ 343, 284, 23 },
		{ 343, 285, 23 },
		{ 343, 286, 23 },
		{ 343, 287, 23 },
		{ 343, 288, 23 },
		{ 343, 289, 23 },
		{ 343, 290, 23 },
		{ 343, 291, 23 },
		{ 343, 292, 23 },
		{ 343, 293, 23 },
		{ 343, 294, 23 },
		{ 343, 295, 23 },
		{ 343, 296, 23 },
		{ 343, 297, 23 },
		{ 343, 298, 23 },
		{ 343, 299, 23 },
		{ 344, 258, 23 },
		{ 344, 259, 23 },
		{ 344, 260, 23 },
		{ 344, 261, 23 },
		{ 344, 262, 23 },
		{ 344, 263, 23 },
		{ 344, 264, 23 },
		{ 344, 265, 23 },
		{ 344, 266, 23 },
		{ 344, 267, 23 },
		{ 344, 268, 23 },
		{ 344, 269, 23 },
		{ 344, 270, 23 },
		{ 344, 271, 23 },
		{ 344, 272, 23 },
		{ 344, 273, 23 },
		{ 344, 274, 23 },
		{ 344, 275, 23 },
		{ 344, 276, 23 },
		{ 344, 277, 23 },
		{ 344, 278, 23 },
		{ 344, 279, 23 },
		{ 344, 280, 23 },
		{ 344, 281, 23 },
		{ 344, 282, 23 },
		{ 344, 283, 23 },
		{ 344, 284, 23 },
		{ 344, 285, 23 },
		{ 344, 286, 23 },
		{ 344, 287, 23 },
		{ 344, 288, 23 },
		{ 344, 289, 23 },
		{ 344, 290, 23 },
		{ 344, 291, 23 },
		{ 344, 292, 23 },
		{ 344, 293, 23 },
		{ 344, 294, 23 },
		{ 344, 295, 23 },
		{ 344, 296, 23 },
		{ 344, 297, 23 },
		{ 344, 298, 23 },
		{ 344, 299, 23 },
		{ 344, 300, 23 },
		{ 345, 258, 23 },
		{ 345, 259, 23 },
		{ 345, 260, 23 },
		{ 345, 261, 23 },
		{ 345, 262, 23 },
		{ 345, 263, 23 },
		{ 345, 264, 23 },
		{ 345, 265, 23 },
		{ 345, 266, 23 },
		{ 345, 267, 23 },
		{ 345, 268, 23 },
		{ 345, 269, 23 },
		{ 345, 270, 23 },
		{ 345, 271, 23 },
		{ 345, 272, 23 },
		{ 345, 273, 23 },
		{ 345, 274, 23 },
		{ 345, 275, 23 },
		{ 345, 276, 23 },
		{ 345, 277, 23 },
		{ 345, 278, 23 },
		{ 345, 279, 23 },
		{ 345, 280, 23 },
		{ 345, 281, 23 },
		{ 345, 282, 23 },
		{ 345, 283, 23 },
		{ 345, 284, 23 },
		{ 345, 285, 23 },
		{ 345, 286, 23 },
		{ 345, 287, 23 },
		{ 345, 288, 23 },
		{ 345, 289, 23 },
		{ 345, 290, 23 },
		{ 345, 291, 23 },
		{ 345, 292, 23 },
		{ 345, 293, 23 },
		{ 345, 294, 23 },
		{ 345, 295, 23 },
		{ 345, 296, 23 },
		{ 345, 297, 23 },
		{ 345, 298, 23 },
		{ 345, 299, 23 },
		{ 345, 300, 23 },
		{ 346, 258, 23 },
		{ 346, 259, 23 },
		{ 346, 260, 23 },
		{ 346, 261, 23 },
		{ 346, 262, 23 },
		{ 346, 263, 23 },
		{ 346, 264, 23 },
		{ 346, 265, 23 },
		{ 346, 266, 23 },
		{ 346, 267, 23 },
		{ 346, 268, 23 },
		{ 346, 269, 23 },
		{ 346, 270, 23 },
		{ 346, 271, 23 },
		{ 346, 272, 23 },
		{ 346, 273, 23 },
		{ 346, 274, 23 },
		{ 346, 275, 23 },
		{ 346, 276, 23 },
		{ 346, 277, 23 },
		{ 346, 278, 23 },
		{ 346, 279, 23 },
		{ 346, 280, 23 },
		{ 346, 281, 23 },
		{ 346, 282, 23 },
		{ 346, 283, 23 },
		{ 346, 284, 23 },
		{ 346, 285, 23 },
		{ 346, 286, 23 },
		{ 346, 287, 23 },
		{ 346, 288, 23 },
		{ 346, 289, 23 },
		{ 346, 290, 23 },
		{ 346, 291, 23 },
		{ 346, 292, 23 },
		{ 346, 293, 23 },
		{ 346, 294, 23 },
		{ 346, 295, 23 },
		{ 346, 296, 23 },
		{ 346, 297, 23 },
		{ 346, 298, 23 },
		{ 346, 299, 23 },
		{ 346, 300, 23 },
		{ 346, 301, 23 },
		{ 347, 258, 23 },
		{ 347, 259, 23 },
		{ 347, 260, 23 },
		{ 347, 261, 23 },
		{ 347, 262, 23 },
		{ 347, 263, 23 },
		{ 347, 264, 23 },
		{ 347, 265, 23 },
		{ 347, 266, 23 },
		{ 347, 267, 23 },
		{ 347, 268, 23 },
		{ 347, 269, 23 },
		{ 347, 270, 23 },
		{ 347, 271, 23 },
		{ 347, 272, 23 },
		{ 347, 273, 23 },
		{ 347, 274, 23 },
		{ 347, 275, 23 },
		{ 347, 276, 23 },
		{ 347, 277, 23 },
		{ 347, 278, 23 },
		{ 347, 279, 23 },
		{ 347, 280, 23 },
		{ 347, 281, 23 },
		{ 347, 282, 23 },
		{ 347, 283, 23 },
		{ 347, 284, 23 },
		{ 347, 285, 23 },
		{ 347, 286, 23 },
		{ 347, 287, 23 },
		{ 347, 288, 23 },
		{ 347, 289, 23 },
		{ 347, 290, 23 },
		{ 347, 291, 23 },
		{ 347, 292, 23 },
		{ 347, 293, 23 },
		{ 347, 294, 23 },
		{ 347, 295, 23 },
		{ 347, 296, 23 },
		{ 347, 297, 23 },
		{ 347, 298, 23 },
		{ 347, 299, 23 },
		{ 347, 300, 23 },
		{ 347, 301, 23 },
		{ 348, 258, 23 },
		{ 348, 259, 23 },
		{ 348, 260, 23 },
		{ 348, 261, 23 },
		{ 348, 262, 23 },
		{ 348, 263, 23 },
		{ 348, 264, 23 },
		{ 348, 265, 23 },
		{ 348, 266, 23 },
		{ 348, 267, 23 },
		{ 348, 268, 23 },
		{ 348, 269, 23 },
		{ 348, 270, 23 },
		{ 348, 271, 23 },
		{ 348, 272, 23 },
		{ 348, 273, 23 },
		{ 348, 274, 23 },
		{ 348, 275, 23 },
		{ 348, 276, 23 },
		{ 348, 277, 23 },
		{ 348, 278, 23 },
		{ 348, 279, 23 },
		{ 348, 280, 23 },
		{ 348, 281, 23 },
		{ 348, 282, 23 },
		{ 348, 283, 23 },
		{ 348, 284, 23 },
		{ 348, 285, 23 },
		{ 348, 286, 23 },
		{ 348, 287, 23 },
		{ 348, 288, 23 },
		{ 348, 289, 23 },
		{ 348, 290, 23 },
		{ 348, 291, 23 },
		{ 348, 292, 23 },
		{ 348, 293, 23 },
		{ 348, 294, 23 },
		{ 348, 295, 23 },
		{ 348, 296, 23 },
		{ 348, 297, 23 },
		{ 348, 298, 23 },
		{ 348, 299, 23 },
		{ 348, 300, 23 },
		{ 348, 301, 23 },
		{ 348, 302, 23 },
		{ 349, 258, 23 },
		{ 349, 259, 23 },
		{ 349, 260, 23 },
		{ 349, 261, 23 },
		{ 349, 262, 23 },
		{ 349, 263, 23 },
		{ 349, 264, 23 },
		{ 349, 265, 23 },
		{ 349, 266, 23 },
		{ 349, 267, 23 },
		{ 349, 268, 23 },
		{ 349, 269, 23 },
		{ 349, 270, 23 },
		{ 349, 271, 23 },
		{ 349, 272, 23 },
		{ 349, 273, 23 },
		{ 349, 274, 23 },
		{ 349, 275, 23 },
		{ 349, 276, 23 },
		{ 349, 277, 23 },
		{ 349, 278, 23 },
		{ 349, 279, 23 },
		{ 349, 280, 23 },
		{ 349, 281, 23 },
		{ 349, 282, 23 },
		{ 349, 283, 23 },
		{ 349, 284, 23 },
		{ 349, 285, 23 },
		{ 349, 286, 23 },
		{ 349, 287, 23 },
		{ 349, 288, 23 },
		{ 349, 289, 23 },
		{ 349, 290, 23 },
		{ 349, 291, 23 },
		{ 349, 292, 23 },
		{ 349, 293, 23 },
		{ 349, 294, 23 },
		{ 349, 295, 23 },
		{ 349, 296, 23 },
		{ 349, 297, 23 },
		{ 349, 298, 23 },
		{ 349, 299, 23 },
		{ 349, 300, 23 },
		{ 349, 301, 23 },
		{ 349, 302, 23 },
		{ 350, 258, 23 },
		{ 350, 259, 23 },
		{ 350, 260, 23 },
		{ 350, 261, 23 },
		{ 350, 262, 23 },
		{ 350, 263, 23 },
		{ 350, 264, 23 },
		{ 350, 265, 23 },
		{ 350, 266, 23 },
		{ 350, 267, 23 },
		{ 350, 268, 23 },
		{ 350, 269, 23 },
		{ 350, 270, 23 },
		{ 350, 271, 23 },
		{ 350, 272, 23 },
		{ 350, 273, 23 },
		{ 350, 274, 23 },
		{ 350, 275, 23 },
		{ 350, 276, 23 },
		{ 350, 277, 23 },
		{ 350, 278, 23 },
		{ 350, 279, 23 },
		{ 350, 280, 23 },
		{ 350, 281, 23 },
		{ 350, 282, 23 },
		{ 350, 283, 23 },
		{ 350, 284, 23 },
		{ 350, 285, 23 },
		{ 350, 286, 23 },
		{ 350, 287, 23 },
		{ 350, 288, 23 },
		{ 350, 289, 23 },
		{ 350, 290, 23 },
		{ 350, 291, 23 },
		{ 350, 292, 23 },
		{ 350, 293, 23 },
		{ 350, 294, 23 },
		{ 350, 295, 23 },
		{ 350, 296, 23 },
		{ 350, 297, 23 },
		{ 350, 298, 23 },
		{ 350, 299, 23 },
		{ 350, 300, 23 },
		{ 350, 301, 23 },
		{ 350, 302, 23 },
		{ 350, 303, 23 },
		{ 351, 258, 23 },
		{ 351, 259, 23 },
		{ 351, 260, 23 },
		{ 351, 261, 23 },
		{ 351, 262, 23 },
		{ 351, 263, 23 },
		{ 351, 264, 23 },
		{ 351, 265, 23 },
		{ 351, 266, 23 },
		{ 351, 267, 23 },
		{ 351, 268, 23 },
		{ 351, 269, 23 },
		{ 351, 270, 23 },
		{ 351, 271, 23 },
		{ 351, 272, 23 },
		{ 351, 273, 23 },
		{ 351, 274, 23 },
		{ 351, 275, 23 },
		{ 351, 276, 23 },
		{ 351, 277, 23 },
		{ 351, 278, 23 },
		{ 351, 279, 23 },
		{ 351, 280, 23 },
		{ 351, 281, 23 },
		{ 351, 282, 23 },
		{ 351, 283, 23 },
		{ 351, 284, 23 },
		{ 351, 285, 23 },
		{ 351, 286, 23 },
		{ 351, 287, 23 },
		{ 351, 288, 23 },
		{ 351, 289, 23 },
		{ 351, 290, 23 },
		{ 351, 291, 23 },
		{ 351, 292, 23 },
		{ 351, 293, 23 },
		{ 351, 294, 23 },
		{ 351, 295, 23 },
		{ 351, 296, 23 },
		{ 351, 297, 23 },
		{ 351, 298, 23 },
		{ 351, 299, 23 },
		{ 351, 300, 23 },
		{ 351, 301, 23 },
		{ 351, 302, 23 },
		{ 351, 303, 23 },
		{ 352, 258, 23 },
		{ 352, 259, 23 },
		{ 352, 260, 23 },
		{ 352, 261, 23 },
		{ 352, 262, 23 },
		{ 352, 263, 23 },
		{ 352, 264, 23 },
		{ 352, 265, 23 },
		{ 352, 266, 23 },
		{ 352, 267, 23 },
		{ 352, 268, 23 },
		{ 352, 269, 23 },
		{ 352, 270, 23 },
		{ 352, 271, 23 },
		{ 352, 272, 23 },
		{ 352, 273, 23 },
		{ 352, 274, 23 },
		{ 352, 275, 23 },
		{ 352, 276, 23 },
		{ 352, 277, 23 },
		{ 352, 278, 23 },
		{ 352, 279, 23 },
		{ 352, 280, 23 },
		{ 352, 281, 23 },
		{ 352, 282, 23 },
		{ 352, 283, 23 },
		{ 352, 284, 23 },
		{ 352, 285, 23 },
		{ 352, 286, 23 },
		{ 352, 287, 23 },
		{ 352, 288, 23 },
		{ 352, 289, 23 },
		{ 352, 290, 23 },
		{ 352, 291, 23 },
		{ 352, 292, 23 },
		{ 352, 293, 23 },
		{ 352, 294, 23 },
		{ 352, 295, 23 },
		{ 352, 296, 23 },
		{ 352, 297, 23 },
		{ 352, 298, 23 },
		{ 352, 299, 23 },
		{ 352, 300, 23 },
		{ 352, 301, 23 },
		{ 352, 302, 23 },
		{ 352, 303, 23 },
		{ 352, 304, 23 },
		{ 353, 258, 23 },
		{ 353, 259, 23 },
		{ 353, 260, 23 },
		{ 353, 261, 23 },
		{ 353, 262, 23 },
		{ 353, 263, 23 },
		{ 353, 264, 23 },
		{ 353, 265, 23 },
		{ 353, 266, 23 },
		{ 353, 267, 23 },
		{ 353, 268, 23 },
		{ 353, 269, 23 },
		{ 353, 270, 23 },
		{ 353, 271, 23 },
		{ 353, 272, 23 },
		{ 353, 273, 23 },
		{ 353, 274, 23 },
		{ 353, 275, 23 },
		{ 353, 276, 23 },
		{ 353, 277, 23 },
		{ 353, 278, 23 },
		{ 353, 279, 23 },
		{ 353, 280, 23 },
		{ 353, 281, 23 },
		{ 353, 282, 23 },
		{ 353, 283, 23 },
		{ 353, 284, 23 },
		{ 353, 285, 23 },
		{ 353, 286, 23 },
		{ 353, 287, 23 },
		{ 353, 288, 23 },
		{ 353, 289, 23 },
		{ 353, 290, 23 },
		{ 353, 291, 23 },
		{ 353, 292, 23 },
		{ 353, 293, 23 },
		{ 353, 294, 23 },
		{ 353, 295, 23 },
		{ 353, 296, 23 },
		{ 353, 297, 23 },
		{ 353, 298, 23 },
		{ 353, 299, 23 },
		{ 353, 300, 23 },
		{ 353, 301, 23 },
		{ 353, 302, 23 },
		{ 353, 303, 23 },
		{ 353, 304, 23 },
		{ 354, 258, 23 },
		{ 354, 259, 23 },
		{ 354, 260, 23 },
		{ 354, 261, 23 },
		{ 354, 262, 23 },
		{ 354, 263, 23 },
		{ 354, 264, 23 },
		{ 354, 265, 23 },
		{ 354, 266, 23 },
		{ 354, 267, 23 },
		{ 354, 268, 23 },
		{ 354, 269, 23 },
		{ 354, 270, 23 },
		{ 354, 271, 23 },
		{ 354, 272, 23 },
		{ 354, 273, 23 },
		{ 354, 274, 23 },
		{ 354, 275, 23 },
		{ 354, 276, 23 },
		{ 354, 277, 23 },
		{ 354, 278, 23 },
		{ 354, 279, 23 },
		{ 354, 280, 23 },
		{ 354, 281, 23 },
		{ 354, 282, 23 },
		{ 354, 283, 23 },
		{ 354, 284, 23 },
		{ 354, 285, 23 },
		{ 354, 286, 23 },
		{ 354, 287, 23 },
		{ 354, 288, 23 },
		{ 354, 289, 23 },
		{ 354, 290, 23 },
		{ 354, 291, 23 },
		{ 354, 292, 23 },
		{ 354, 293, 23 },
		{ 354, 294, 23 },
		{ 354, 295, 23 },
		{ 354, 296, 23 },
		{ 354, 297, 23 },
		{ 354, 298, 23 },
		{ 354, 299, 23 },
		{ 354, 300, 23 },
		{ 354, 301, 23 },
		{ 354, 302, 23 },
		{ 354, 303, 23 },
		{ 354, 304, 23 },
		{ 354, 305, 23 },
		{ 355, 258, 23 },
		{ 355, 259, 23 },
		{ 355, 260, 23 },
		{ 355, 261, 23 },
		{ 355, 262, 23 },
		{ 355, 263, 23 },
		{ 355, 264, 23 },
		{ 355, 265, 23 },
		{ 355, 266, 23 },
		{ 355, 267, 23 },
		{ 355, 268, 23 },
		{ 355, 269, 23 },
		{ 355, 270, 23 },
		{ 355, 271, 23 },
		{ 355, 272, 23 },
		{ 355, 273, 23 },
		{ 355, 274, 23 },
		{ 355, 275, 23 },
		{ 355, 276, 23 },
		{ 355, 277, 23 },
		{ 355, 278, 23 },
		{ 355, 279, 23 },
		{ 355, 280, 23 },
		{ 355, 281, 23 },
		{ 355, 282, 23 },
		{ 355, 283, 23 },
		{ 355, 284, 23 },
		{ 355, 285, 23 },
		{ 355, 286, 23 },
		{ 355, 287, 23 },
		{ 355, 288, 23 },
		{ 355, 289, 23 },
		{ 355, 290, 23 },
		{ 355, 291, 23 },
		{ 355, 292, 23 },
		{ 355, 293, 23 },
		{ 355, 294, 23 },
		{ 355, 295, 23 },
		{ 355, 296, 23 },
		{ 355, 297, 23 },
		{ 355, 298, 23 },
		{ 355, 299, 23 },
		{ 355, 300, 23 },
		{ 355, 301, 23 },
		{ 355, 302, 23 },
		{ 355, 303, 23 },
		{ 355, 304, 23 },
		{ 355, 305, 23 },
		{ 356, 258, 23 },
		{ 356, 259, 23 },
		{ 356, 260, 23 },
		{ 356, 261, 23 },
		{ 356, 262, 23 },
		{ 356, 263, 23 },
		{ 356, 264, 23 },
		{ 356, 265, 23 },
		{ 356, 266, 23 },
		{ 356, 267, 23 },
		{ 356, 268, 23 },
		{ 356, 269, 23 },
		{ 356, 270, 23 },
		{ 356, 271, 23 },
		{ 356, 272, 23 },
		{ 356, 273, 23 },
		{ 356, 274, 23 },
		{ 356, 275, 23 },
		{ 356, 276, 23 },
		{ 356, 277, 23 },
		{ 356, 278, 23 },
		{ 356, 279, 23 },
		{ 356, 280, 23 },
		{ 356, 281, 23 },
		{ 356, 282, 23 },
		{ 356, 283, 23 },
		{ 356, 284, 23 },
		{ 356, 285, 23 },
		{ 356, 286, 23 },
		{ 356, 287, 23 },
		{ 356, 288, 23 },
		{ 356, 289, 23 },
		{ 356, 290, 23 },
		{ 356, 291, 23 },
		{ 356, 292, 23 },
		{ 356, 293, 23 },
		{ 356, 294, 23 },
		{ 356, 295, 23 },
		{ 356, 296, 23 },
		{ 356, 297, 23 },
		{ 356, 298, 23 },
		{ 356, 299, 23 },
		{ 356, 300, 23 },
		{ 356, 301, 23 },
		{ 356, 302, 23 },
		{ 356, 303, 23 },
		{ 356, 304, 23 },
		{ 356, 305, 23 },
		{ 356, 306, 23 },
		{ 357, 258, 23 },
		{ 357, 259, 23 },
		{ 357, 260, 23 },
		{ 357, 261, 23 },
		{ 357, 262, 23 },
		{ 357, 263, 23 },
		{ 357, 264, 23 },
		{ 357, 265, 23 },
		{ 357, 266, 23 },
		{ 357, 267, 23 },
		{ 357, 268, 23 },
		{ 357, 269, 23 },
		{ 357, 270, 23 },
		{ 357, 271, 23 },
		{ 357, 272, 23 },
		{ 357, 273, 23 },
		{ 357, 274, 23 },
		{ 357, 275, 23 },
		{ 357, 276, 23 },
		{ 357, 277, 23 },
		{ 357, 278, 23 },
		{ 357, 279, 23 },
		{ 357, 280, 23 },
		{ 357, 281, 23 },
		{ 357, 282, 23 },
		{ 357, 283, 23 },
		{ 357, 284, 23 },
		{ 357, 285, 23 },
		{ 357, 286, 23 },
		{ 357, 287, 23 },
		{ 357, 288, 23 },
		{ 357, 289, 23 },
		{ 357, 290, 23 },
		{ 357, 291, 23 },
		{ 357, 292, 23 },
		{ 357, 293, 23 },
		{ 357, 294, 23 },
		{ 357, 295, 23 },
		{ 357, 296, 23 },
		{ 357, 297, 23 },
		{ 357, 298, 23 },
		{ 357, 299, 23 },
		{ 357, 300, 23 },
		{ 357, 301, 23 },
		{ 357, 302, 23 },
		{ 357, 303, 23 },
		{ 357, 304, 23 },
		{ 357, 305, 23 },
		{ 357, 306, 23 },
		{ 358, 258, 23 },
		{ 358, 259, 23 },
		{ 358, 260, 23 },
		{ 358, 261, 23 },
		{ 358, 262, 23 },
		{ 358, 263, 23 },
		{ 358, 264, 23 },
		{ 358, 265, 23 },
		{ 358, 266, 23 },
		{ 358, 267, 23 },
		{ 358, 268, 23 },
		{ 358, 269, 23 },
		{ 358, 270, 23 },
		{ 358, 271, 23 },
		{ 358, 272, 23 },
		{ 358, 273, 23 },
		{ 358, 274, 23 },
		{ 358, 275, 23 },
		{ 358, 276, 23 },
		{ 358, 277, 23 },
		{ 358, 278, 23 },
		{ 358, 279, 23 },
		{ 358, 280, 23 },
		{ 358, 281, 23 },
		{ 358, 282, 23 },
		{ 358, 283, 23 },
		{ 358, 284, 23 },
		{ 358, 285, 23 },
		{ 358, 286, 23 },
		{ 358, 287, 23 },
		{ 358, 288, 23 },
		{ 358, 289, 23 },
		{ 358, 290, 23 },
		{ 358, 291, 23 },
		{ 358, 292, 23 },
		{ 358, 293, 23 },
		{ 358, 294, 23 },
		{ 358, 295, 23 },
		{ 358, 296, 23 },
		{ 358, 297, 23 },
		{ 358, 298, 23 },
		{ 358, 299, 23 },
		{ 358, 300, 23 },
		{ 358, 301, 23 },
		{ 358, 302, 23 },
		{ 358, 303, 23 },
		{ 358, 304, 23 },
		{ 358, 305, 23 },
		{ 358, 306, 23 },
		{ 358, 307, 23 },
		{ 359, 258, 23 },
		{ 359, 259, 23 },
		{ 359, 260, 23 },
		{ 359, 261, 23 },
		{ 359, 262, 23 },
		{ 359, 263, 23 },
		{ 359, 264, 23 },
		{ 359, 265, 23 },
		{ 359, 266, 23 },
		{ 359, 267, 23 },
		{ 359, 268, 23 },
		{ 359, 269, 23 },
		{ 359, 270, 23 },
		{ 359, 271, 23 },
		{ 359, 272, 23 },
		{ 359, 273, 23 },
		{ 359, 274, 23 },
		{ 359, 275, 23 },
		{ 359, 276, 23 },
		{ 359, 277, 23 },
		{ 359, 278, 23 },
		{ 359, 279, 23 },
		{ 359, 280, 23 },
		{ 359, 281, 23 },
		{ 359, 282, 23 },
		{ 359, 283, 23 },
		{ 359, 284, 23 },
		{ 359, 285, 23 },
		{ 359, 286, 23 },
		{ 359, 287, 23 },
		{ 359, 288, 23 },
		{ 359, 289, 23 },
		{ 359, 290, 23 },
		{ 359, 291, 23 },
		{ 359, 292, 23 },
		{ 359, 293, 23 },
		{ 359, 294, 23 },
		{ 359, 295, 23 },
		{ 359, 296, 23 },
		{ 359, 297, 23 },
		{ 359, 298, 23 },
		{ 359, 299, 23 },
		{ 359, 300, 23 },
		{ 359, 301, 23 },
		{ 359, 302, 23 },
		{ 359, 303, 23 },
		{ 359, 304, 23 },
		{ 359, 305, 23 },
		{ 359, 306, 23 },
		{ 359, 307, 23 },
		{ 360, 258, 23 },
		{ 360, 259, 23 },
		{ 360, 260, 23 },
		{ 360, 261, 23 },
		{ 360, 262, 23 },
		{ 360, 263, 23 },
		{ 360, 264, 23 },
		{ 360, 265, 23 },
		{ 360, 266, 23 },
		{ 360, 267, 23 },
		{ 360, 268, 23 },
		{ 360, 269, 23 },
		{ 360, 270, 23 },
		{ 360, 271, 23 },
		{ 360, 272, 23 },
		{ 360, 273, 23 },
		{ 360, 274, 23 },
		{ 360, 275, 23 },
		{ 360, 276, 23 },
		{ 360, 277, 23 },
		{ 360, 278, 23 },
		{ 360, 279, 23 },
		{ 360, 280, 23 },
		{ 360, 281, 23 },
		{ 360, 282, 23 },
		{ 360, 283, 23 },
		{ 360, 284, 23 },
		{ 360, 285, 23 },
		{ 360, 286, 23 },
		{ 360, 287, 23 },
		{ 360, 288, 23 },
		{ 360, 289, 23 },
		{ 360, 290, 23 },
		{ 360, 291, 23 },
		{ 360, 292, 23 },
		{ 360, 293, 23 },
		{ 360, 294, 23 },
		{ 360, 295, 23 },
		{ 360, 296, 23 },
		{ 360, 297, 23 },
		{ 360, 298, 23 },
		{ 360, 299, 23 },
		{ 360, 300, 23 },
		{ 360, 301, 23 },
		{ 360, 302, 23 },
		{ 360, 303, 23 },
		{ 360, 304, 23 },
		{ 360, 305, 23 },
		{ 360, 306, 23 },
		{ 360, 307, 23 },
		{ 360, 308, 23 },
		{ 361, 258, 23 },
		{ 361, 259, 23 },
		{ 361, 260, 23 },
		{ 361, 261, 23 },
		{ 361, 262, 23 },
		{ 361, 263, 23 },
		{ 361, 264, 23 },
		{ 361, 265, 23 },
		{ 361, 266, 23 },
		{ 361, 267, 23 },
		{ 361, 268, 23 },
		{ 361, 269, 23 },
		{ 361, 270, 23 },
		{ 361, 271, 23 },
		{ 361, 272, 23 },
		{ 361, 273, 23 },
		{ 361, 274, 23 },
		{ 361, 275, 23 },
		{ 361, 276, 23 },
		{ 361, 277, 23 },
		{ 361, 278, 23 },
		{ 361, 279, 23 },
		{ 361, 280, 23 },
		{ 361, 281, 23 },
		{ 361, 282, 23 },
		{ 361, 283, 23 },
		{ 361, 284, 23 },
		{ 361, 285, 23 },
		{ 361, 286, 23 },
		{ 361, 287, 23 },
		{ 361, 288, 23 },
		{ 361, 289, 23 },
		{ 361, 290, 23 },
		{ 361, 291, 23 },
		{ 361, 292, 23 },
		{ 361, 293, 23 },
		{ 361, 294, 23 },
		{ 361, 295, 23 },
		{ 361, 296, 23 },
		{ 361, 297, 23 },
		{ 361, 298, 23 },
		{ 361, 299, 23 },
		{ 361, 300, 23 },
		{ 361, 301, 23 },
		{ 361, 302, 23 },
		{ 361, 303, 23 },
		{ 361, 304, 23 },
		{ 361, 305, 23 },
		{ 361, 306, 23 },
		{ 361, 307, 23 },
		{ 361, 308, 23 },
		{ 362, 258, 23 },
		{ 362, 259, 23 },
		{ 362, 260, 23 },
		{ 362, 261, 23 },
		{ 362, 262, 23 },
		{ 362, 263, 23 },
		{ 362, 264, 23 },
		{ 362, 265, 23 },
		{ 362, 266, 23 },
		{ 362, 267, 23 },
		{ 362, 268, 23 },
		{ 362, 269, 23 },
		{ 362, 270, 23 },
		{ 362, 271, 23 },
		{ 362, 272, 23 },
		{ 362, 273, 23 },
		{ 362, 274, 23 },
		{ 362, 275, 23 },
		{ 362, 276, 23 },
		{ 362, 277, 23 },
		{ 362, 278, 23 },
		{ 362, 279, 23 },
		{ 362, 280, 23 },
		{ 362, 281, 23 },
		{ 362, 282, 23 },
		{ 362, 283, 23 },
		{ 362, 284, 23 },
		{ 362, 285, 23 },
		{ 362, 286, 23 },
		{ 362, 287, 23 },
		{ 362, 288, 23 },
		{ 362, 289, 23 },
		{ 362, 290, 23 },
		{ 362, 291, 23 },
		{ 362, 292, 23 },
		{ 362, 293, 23 },
		{ 362, 294, 23 },
		{ 362, 295, 23 },
		{ 362, 296, 23 },
		{ 362, 297, 23 },
		{ 362, 298, 23 },
		{ 362, 299, 23 },
		{ 362, 300, 23 },
		{ 362, 301, 23 },
		{ 362, 302, 23 },
		{ 362, 303, 23 },
		{ 362, 304, 23 },
		{ 362, 305, 23 },
		{ 362, 306, 23 },
		{ 362, 307, 23 },
		{ 362, 308, 23 },
		{ 362, 309, 23 },
		{ 363, 258, 23 },
		{ 363, 259, 23 },
		{ 363, 260, 23 },
		{ 363, 261, 23 },
		{ 363, 262, 23 },
		{ 363, 263, 23 },
		{ 363, 264, 23 },
		{ 363, 265, 23 },
		{ 363, 266, 23 },
		{ 363, 267, 23 },
		{ 363, 268, 23 },
		{ 363, 269, 23 },
		{ 363, 270, 23 },
		{ 363, 271, 23 },
		{ 363, 272, 23 },
		{ 363, 273, 23 },
		{ 363, 274, 23 },
		{ 363, 275, 23 },
		{ 363, 276, 23 },
		{ 363, 277, 23 },
		{ 363, 278, 23 },
		{ 363, 279, 23 },
		{ 363, 280, 23 },
		{ 363, 281, 23 },
		{ 363, 282, 23 },
		{ 363, 283, 23 },
		{ 363, 284, 23 },
		{ 363, 285, 23 },
		{ 363, 286, 23 },
		{ 363, 287, 23 },
		{ 363, 288, 23 },
		{ 363, 289, 23 },
		{ 363, 290, 23 },
		{ 363, 291, 23 },
		{ 363, 292, 23 },
		{ 363, 293, 23 },
		{ 363, 294, 23 },
		{ 363, 295, 23 },
		{ 363, 296, 23 },
		{ 363, 297, 23 },
		{ 363, 298, 23 },
		{ 363, 299, 23 },
		{ 363, 300, 23 },
		{ 363, 301, 23 },
		{ 363, 302, 23 },
		{ 363, 303, 23 },
		{ 363, 304, 23 },
		{ 363, 305, 23 },
		{ 363, 306, 23 },
		{ 363, 307, 23 },
		{ 363, 308, 23 },
		{ 363, 309, 23 },
		{ 364, 258, 23 },
		{ 364, 259, 23 },
		{ 364, 260, 23 },
		{ 364, 261, 23 },
		{ 364, 262, 23 },
		{ 364, 263, 23 },
		{ 364, 264, 23 },
		{ 364, 265, 23 },
		{ 364, 266, 23 },
		{ 364, 267, 23 },
		{ 364, 268, 23 },
		{ 364, 269, 23 },
		{ 364, 270, 23 },
		{ 364, 271, 23 },
		{ 364, 272, 23 },
		{ 364, 273, 23 },
		{ 364, 274, 23 },
		{ 364, 275, 23 },
		{ 364, 276, 23 },
		{ 364, 277, 23 },
		{ 364, 278, 23 },
		{ 364, 279, 23 },
		{ 364, 280, 23 },
		{ 364, 281, 23 },
		{ 364, 282, 23 },
		{ 364, 283, 23 },
		{ 364, 284, 23 },
		{ 364, 285, 23 },
		{ 364, 286, 23 },
		{ 364, 287, 23 },
		{ 364, 288, 23 },
		{ 364, 289, 23 },
		{ 364, 290, 23 },
		{ 364, 291, 23 },
		{ 364, 292, 23 },
		{ 364, 293, 23 },
		{ 364, 294, 23 },
		{ 364, 295, 23 },
		{ 364, 296, 23 },
		{ 364, 297, 23 },
		{ 364, 298, 23 },
		{ 364, 299, 23 },
		{ 364, 300, 23 },
		{ 364, 301, 23 },
		{ 364, 302, 23 },
		{ 364, 303, 23 },
		{ 364, 304, 23 },
		{ 364, 305, 23 },
		{ 364, 306, 23 },
		{ 364, 307, 23 },
		{ 364, 308, 23 },
		{ 364, 309, 23 },
		{ 364, 310, 23 },
		{ 365, 258, 23 },
		{ 365, 259, 23 },
		{ 365, 260, 23 },
		{ 365, 261, 23 },
		{ 365, 262, 23 },
		{ 365, 263, 23 },
		{ 365, 264, 23 },
		{ 365, 265, 23 },
		{ 365, 266, 23 },
		{ 365, 267, 23 },
		{ 365, 268, 23 },
		{ 365, 269, 23 },
		{ 365, 270, 23 },
		{ 365, 271, 23 },
		{ 365, 272, 23 },
		{ 365, 273, 23 },
		{ 365, 274, 23 },
		{ 365, 275, 23 },
		{ 365, 276, 23 },
		{ 365, 277, 23 },
		{ 365, 278, 23 },
		{ 365, 279, 23 },
		{ 365, 280, 23 },
		{ 365, 281, 23 },
		{ 365, 282, 23 },
		{ 365, 283, 23 },
		{ 365, 284, 23 },
		{ 365, 285, 23 },
		{ 365, 286, 23 },
		{ 365, 287, 23 },
		{ 365, 288, 23 },
		{ 365, 289, 23 },
		{ 365, 290, 23 },
		{ 365, 291, 23 },
		{ 365, 292, 23 },
		{ 365, 293, 23 },
		{ 365, 294, 23 },
		{ 365, 295, 23 },
		{ 365, 296, 23 },
		{ 365, 297, 23 },
		{ 365, 298, 23 },
		{ 365, 299, 23 },
		{ 365, 300, 23 },
		{ 365, 301, 23 },
		{ 365, 302, 23 },
		{ 365, 303, 23 },
		{ 365, 304, 23 },
		{ 365, 305, 23 },
		{ 365, 306, 23 },
		{ 365, 307, 23 },
		{ 365, 308, 23 },
		{ 365, 309, 23 },
		{ 365, 310, 23 },
		{ 366, 258, 23 },
		{ 366, 259, 23 },
		{ 366, 260, 23 },
		{ 366, 261, 23 },
		{ 366, 262, 23 },
		{ 366, 263, 23 },
		{ 366, 264, 23 },
		{ 366, 265, 23 },
		{ 366, 266, 23 },
		{ 366, 267, 23 },
		{ 366, 268, 23 },
		{ 366, 269, 23 },
		{ 366, 270, 23 },
		{ 366, 271, 23 },
		{ 366, 272, 23 },
		{ 366, 273, 23 },
		{ 366, 274, 23 },
		{ 366, 275, 23 },
		{ 366, 276, 23 },
		{ 366, 277, 23 },
		{ 366, 278, 23 },
		{ 366, 279, 23 },
		{ 366, 280, 23 },
		{ 366, 281, 23 },
		{ 366, 282, 23 },
		{ 366, 283, 23 },
		{ 366, 284, 23 },
		{ 366, 285, 23 },
		{ 366, 286, 23 },
		{ 366, 287, 23 },
		{ 366, 288, 23 },
		{ 366, 289, 23 },
		{ 366, 290, 23 },
		{ 366, 291, 23 },
		{ 366, 292, 23 },
		{ 366, 293, 23 },
		{ 366, 294, 23 },
		{ 366, 295, 23 },
		{ 366, 296, 23 },
		{ 366, 297, 23 },
		{ 366, 298, 23 },
		{ 366, 299, 23 },
		{ 366, 300, 23 },
		{ 366, 301, 23 },
		{ 366, 302, 23 },
		{ 366, 303, 23 },
		{ 366, 304, 23 },
		{ 366, 305, 23 },
		{ 366, 306, 23 },
		{ 366, 307, 23 },
		{ 366, 308, 23 },
		{ 366, 309, 23 },
		{ 366, 310, 23 },
		{ 366, 311, 23 },
		{ 367, 258, 23 },
		{ 367, 259, 23 },
		{ 367, 260, 23 },
		{ 367, 261, 23 },
		{ 367, 262, 23 },
		{ 367, 263, 23 },
		{ 367, 264, 23 },
		{ 367, 265, 23 },
		{ 367, 266, 23 },
		{ 367, 267, 23 },
		{ 367, 268, 23 },
		{ 367, 269, 23 },
		{ 367, 270, 23 },
		{ 367, 271, 23 },
		{ 367, 272, 23 },
		{ 367, 273, 23 },
		{ 367, 274, 23 },
		{ 367, 275, 23 },
		{ 367, 276, 23 },
		{ 367, 277, 23 },
		{ 367, 278, 23 },
		{ 367, 279, 23 },
		{ 367, 280, 23 },
		{ 367, 281, 23 },
		{ 367, 282, 23 },
		{ 367, 283, 23 },
		{ 367, 284, 23 },
		{ 367, 285, 23 },
		{ 367, 286, 23 },
		{ 367, 287, 23 },
		{ 367, 288, 23 },
		{ 367, 289, 23 },
		{ 367, 290, 23 },
		{ 367, 291, 23 },
		{ 367, 292, 23 },
		{ 367, 293, 23 },
		{ 367, 294, 23 },
		{ 367, 295, 23 },
		{ 367, 296, 23 },
		{ 367, 297, 23 },
		{ 367, 298, 23 },
		{ 367, 299, 23 },
		{ 367, 300, 23 },
		{ 367, 301, 23 },
		{ 367, 302, 23 },
		{ 367, 303, 23 },
		{ 367, 304, 23 },
		{ 367, 305, 23 },
		{ 367, 306, 23 },
		{ 367, 307, 23 },
		{ 367, 308, 23 },
		{ 367, 309, 23 },
		{ 367, 310, 23 },
		{ 367, 311, 23 },
		{ 368, 258, 23 },
		{ 368, 259, 23 },
		{ 368, 260, 23 },
		{ 368, 261, 23 },
		{ 368, 262, 23 },
		{ 368, 263, 23 },
		{ 368, 264, 23 },
		{ 368, 265, 23 },
		{ 368, 266, 23 },
		{ 368, 267, 23 },
		{ 368, 268, 23 },
		{ 368, 269, 23 },
		{ 368, 270, 23 },
		{ 368, 271, 23 },
		{ 368, 272, 23 },
		{ 368, 273, 23 },
		{ 368, 274, 23 },
		{ 368, 275, 23 },
		{ 368, 276, 23 },
		{ 368, 277, 23 },
		{ 368, 278, 23 },
		{ 368, 279, 23 },
		{ 368, 280, 23 },
		{ 368, 281, 23 },
		{ 368, 282, 23 },
		{ 368, 283, 23 },
		{ 368, 284, 23 },
		{ 368, 285, 23 },
		{ 368, 286, 23 },
		{ 368, 287, 23 },
		{ 368, 288, 23 },
		{ 368, 289, 23 },
		{ 368, 290, 23 },
		{ 368, 291, 23 },
		{ 368, 292, 23 },
		{ 368, 293, 23 },
		{ 368, 294, 23 },
		{ 368, 295, 23 },
		{ 368, 296, 23 },
		{ 368, 297, 23 },
		{ 368, 298, 23 },
		{ 368, 299, 23 },
		{ 368, 300, 23 },
		{ 368, 301, 23 },
		{ 368, 302, 23 },
		{ 368, 303, 23 },
		{ 368, 304, 23 },
		{ 368, 305, 23 },
		{ 368, 306, 23 },
		{ 368, 307, 23 },
		{ 368, 308, 23 },
		{ 368, 309, 23 },
		{ 368, 310, 23 },
		{ 368, 311, 23 },
		{ 368, 312, 23 },
		{ 369, 258, 23 },
		{ 369, 259, 23 },
		{ 369, 260, 23 },
		{ 369, 261, 23 },
		{ 369, 262, 23 },
		{ 369, 263, 23 },
		{ 369, 264, 23 },
		{ 369, 265, 23 },
		{ 369, 266, 23 },
		{ 369, 267, 23 },
		{ 369, 268, 23 },
		{ 369, 269, 23 },
		{ 369, 270, 23 },
		{ 369, 271, 23 },
		{ 369, 272, 23 },
		{ 369, 273, 23 },
		{ 369, 274, 23 },
		{ 369, 275, 23 },
		{ 369, 276, 23 },
		{ 369, 277, 23 },
		{ 369, 278, 23 },
		{ 369, 279, 23 },
		{ 369, 280, 23 },
		{ 369, 281, 23 },
		{ 369, 282, 23 },
		{ 369, 283, 23 },
		{ 369, 284, 23 },
		{ 369, 285, 23 },
		{ 369, 286, 23 },
		{ 369, 287, 23 },
		{ 369, 288, 23 },
		{ 369, 289, 23 },
		{ 369, 290, 23 },
		{ 369, 291, 23 },
		{ 369, 292, 23 },
		{ 369, 293, 23 },
		{ 369, 294, 23 },
		{ 369, 295, 23 },
		{ 369, 296, 23 },
		{ 369, 297, 23 },
		{ 369, 298, 23 },
		{ 369, 299, 23 },
		{ 369, 300, 23 },
		{ 369, 301, 23 },
		{ 369, 302, 23 },
		{ 369, 303, 23 },
		{ 369, 304, 23 },
		{ 369, 305, 23 },
		{ 369, 306, 23 },
		{ 369, 307, 23 },
		{ 369, 308, 23 },
		{ 369, 309, 23 },
		{ 369, 310, 23 },
		{ 369, 311, 23 },
		{ 369, 312, 23 },
		{ 370, 258, 23 },
		{ 370, 259, 23 },
		{ 370, 260, 23 },
		{ 370, 261, 23 },
		{ 370, 262, 23 },
		{ 370, 263, 23 },
		{ 370, 264, 23 },
		{ 370, 265, 23 },
		{ 370, 266, 23 },
		{ 370, 267, 23 },
		{ 370, 268, 23 },
		{ 370, 269, 23 },
		{ 370, 270, 23 },
		{ 370, 271, 23 },
		{ 370, 272, 23 },
		{ 370, 273, 23 },
		{ 370, 274, 23 },
		{ 370, 275, 23 },
		{ 370, 276, 23 },
		{ 370, 277, 23 },
		{ 370, 278, 23 },
		{ 370, 279, 23 },
		{ 370, 280, 23 },
		{ 370, 281, 23 },
		{ 370, 282, 23 },
		{ 370, 283, 23 },
		{ 370, 284, 23 },
		{ 370, 285, 23 },
		{ 370, 286, 23 },
		{ 370, 287, 23 },
		{ 370, 288, 23 },
		{ 370, 289, 23 },
		{ 370, 290, 23 },
		{ 370, 291, 23 },
		{ 370, 292, 23 },
		{ 370, 293, 23 },
		{ 370, 294, 23 },
		{ 370, 295, 23 },
		{ 370, 296, 23 },
		{ 370, 297, 23 },
		{ 370, 298, 23 },
		{ 370, 299, 23 },
		{ 370, 300, 23 },
		{ 370, 301, 23 },
		{ 370, 302, 23 },
		{ 370, 303, 23 },
		{ 370, 304, 23 },
		{ 370, 305, 23 },
		{ 370, 306, 23 },
		{ 370, 307, 23 },
		{ 370, 308, 23 },
		{ 370, 309, 23 },
		{ 370, 310, 23 },
		{ 370, 311, 23 },
		{ 370, 312, 23 },
		{ 370, 313, 23 },
		{ 371, 258, 23 },
		{ 371, 259, 23 },
		{ 371, 260, 23 },
		{ 371, 261, 23 },
		{ 371, 262, 23 },
		{ 371, 263, 23 },
		{ 371, 264, 23 },
		{ 371, 265, 23 },
		{ 371, 266, 23 },
		{ 371, 267, 23 },
		{ 371, 268, 23 },
		{ 371, 269, 23 },
		{ 371, 270, 23 },
		{ 371, 271, 23 },
		{ 371, 272, 23 },
		{ 371, 273, 23 },
		{ 371, 274, 23 },
		{ 371, 275, 23 },
		{ 371, 276, 23 },
		{ 371, 277, 23 },
		{ 371, 278, 23 },
		{ 371, 279, 23 },
		{ 371, 280, 23 },
		{ 371, 281, 23 },
		{ 371, 282, 23 },
		{ 371, 283, 23 },
		{ 371, 284, 23 },
		{ 371, 285, 23 },
		{ 371, 286, 23 },
		{ 371, 287, 23 },
		{ 371, 288, 23 },
		{ 371, 289, 23 },
		{ 371, 290, 23 },
		{ 371, 291, 23 },
		{ 371, 292, 23 },
		{ 371, 293, 23 },
		{ 371, 294, 23 },
		{ 371, 295, 23 },
		{ 371, 296, 23 },
		{ 371, 297, 23 },
		{ 371, 298, 23 },
		{ 371, 299, 23 },
		{ 371, 300, 23 },
		{ 371, 301, 23 },
		{ 371, 302, 23 },
		{ 371, 303, 23 },
		{ 371, 304, 23 },
		{ 371, 305, 23 },
		{ 371, 306, 23 },
		{ 371, 307, 23 },
		{ 371, 308, 23 },
		{ 371, 309, 23 },
		{ 371, 310, 23 },
		{ 371, 311, 23 },
		{ 371, 312, 23 },
		{ 371, 313, 23 },
		{ 372, 258, 23 },
		{ 372, 259, 23 },
		{ 372, 260, 23 },
		{ 372, 261, 23 },
		{ 372, 262, 23 },
		{ 372, 263, 23 },
		{ 372, 264, 23 },
		{ 372, 265, 23 },
		{ 372, 266, 23 },
		{ 372, 267, 23 },
		{ 372, 268, 23 },
		{ 372, 269, 23 },
		{ 372, 270, 23 },
		{ 372, 271, 23 },
		{ 372, 272, 23 },
		{ 372, 273, 23 },
		{ 372, 274, 23 },
		{ 372, 275, 23 },
		{ 372, 276, 23 },
		{ 372, 277, 23 },
		{ 372, 278, 23 },
		{ 372, 279, 23 },
		{ 372, 280, 23 },
		{ 372, 281, 23 },
		{ 372, 282, 23 },
		{ 372, 283, 23 },
		{ 372, 284, 23 },
		{ 372, 285, 23 },
		{ 372, 286, 23 },
		{ 372, 287, 23 },
		{ 372, 288, 23 },
		{ 372, 289, 23 },
		{ 372, 290, 23 },
		{ 372, 291, 23 },
		{ 372, 292, 23 },
		{ 372, 293, 23 },
		{ 372, 294, 23 },
		{ 372, 295, 23 },
		{ 372, 296, 23 },
		{ 372, 297, 23 },
		{ 372, 298, 23 },
		{ 372, 299, 23 },
		{ 372, 300, 23 },
		{ 372, 301, 23 },
		{ 372, 302, 23 },
		{ 372, 303, 23 },
		{ 372, 304, 23 },
		{ 372, 305, 23 },
		{ 372, 306, 23 },
		{ 372, 307, 23 },
		{ 372, 308, 23 },
		{ 372, 309, 23 },
		{ 372, 310, 23 },
		{ 372, 311, 23 },
		{ 372, 312, 23 },
		{ 372, 313, 23 },
		{ 372, 314, 23 },
		{ 373, 258, 23 },
		{ 373, 259, 23 },
		{ 373, 260, 23 },
		{ 373, 261, 23 },
		{ 373, 262, 23 },
		{ 373, 263, 23 },
		{ 373, 264, 23 },
		{ 373, 265, 23 },
		{ 373, 266, 23 },
		{ 373, 267, 23 },
		{ 373, 268, 23 },
		{ 373, 269, 23 },
		{ 373, 270, 23 },
		{ 373, 271, 23 },
		{ 373, 272, 23 },
		{ 373, 273, 23 },
		{ 373, 274, 23 },
		{ 373, 275, 23 },
		{ 373, 276, 23 },
		{ 373, 277, 23 },
		{ 373, 278, 23 },
		{ 373, 279, 23 },
		{ 373, 280, 23 },
		{ 373, 281, 23 },
		{ 373, 282, 23 },
		{ 373, 283, 23 },
		{ 373, 284, 23 },
		{ 373, 285, 23 },
		{ 373, 286, 23 },
		{ 373, 287, 23 },
		{ 373, 288, 23 },
		{ 373, 289, 23 },
		{ 373, 290, 23 },
		{ 373, 291, 23 },
		{ 373, 292, 23 },
		{ 373, 293, 23 },
		{ 373, 294, 23 },
		{ 373, 295, 23 },
		{ 373, 296, 23 },
		{ 373, 297, 23 },
		{ 373, 298, 23 },
		{ 373, 299, 23 },
		{ 373, 300, 23 },
		{ 373, 301, 23 },
		{ 373, 302, 23 },
		{ 373, 303, 23 },
		{ 373, 304, 23 },
		{ 373, 305, 23 },
		{ 373, 306, 23 },
		{ 373, 307, 23 },
		{ 373, 308, 23 },
		{ 373, 309, 23 },
		{ 373, 310, 23 },
		{ 373, 311, 23 },
		{ 373, 312, 23 },
		{ 373, 313, 23 },
		{ 373, 314, 23 },
		{ 374, 258, 23 },
		{ 374, 259, 23 },
		{ 374, 260, 23 },
		{ 374, 261, 23 },
		{ 374, 262, 23 },
		{ 374, 263, 23 },
		{ 374, 264, 23 },
		{ 374, 265, 23 },
		{ 374, 266, 23 },
		{ 374, 267, 23 },
		{ 374, 268, 23 },
		{ 374, 269, 23 },
		{ 374, 270, 23 },
		{ 374, 271, 23 },
		{ 374, 272, 23 },
		{ 374, 273, 23 },
		{ 374, 274, 23 },
		{ 374, 275, 23 },
		{ 374, 276, 23 },
		{ 374, 277, 23 },
		{ 374, 278, 23 },
		{ 374, 279, 23 },
		{ 374, 280, 23 },
		{ 374, 281, 23 },
		{ 374, 282, 23 },
		{ 374, 283, 23 },
		{ 374, 284, 23 },
		{ 374, 285, 23 },
		{ 374, 286, 23 },
		{ 374, 287, 23 },
		{ 374, 288, 23 },
		{ 374, 289, 23 },
		{ 374, 290, 23 },
		{ 374, 291, 23 },
		{ 374, 292, 23 },
		{ 374, 293, 23 },
		{ 374, 294, 23 },
		{ 374, 295, 23 },
		{ 374, 296, 23 },
		{ 374, 297, 23 },
		{ 374, 298, 23 },
		{ 374, 299, 23 },
		{ 374, 300, 23 },
		{ 374, 301, 23 },
		{ 374, 302, 23 },
		{ 374, 303, 23 },
		{ 374, 304, 23 },
		{ 374, 305, 23 },
		{ 374, 306, 23 },
		{ 374, 307, 23 },
		{ 374, 308, 23 },
		{ 374, 309, 23 },
		{ 374, 310, 23 },
		{ 374, 311, 23 },
		{ 374, 312, 23 },
		{ 374, 313, 23 },
		{ 374, 314, 23 },
		{ 374, 315, 23 },
		{ 375, 258, 23 },
		{ 375, 259, 23 },
		{ 375, 260, 23 },
		{ 375, 261, 23 },
		{ 375, 262, 23 },
		{ 375, 263, 23 },
		{ 375, 264, 23 },
		{ 375, 265, 23 },
		{ 375, 266, 23 },
		{ 375, 267, 23 },
		{ 375, 268, 23 },
		{ 375, 269, 23 },
		{ 375, 270, 23 },
		{ 375, 271, 23 },
		{ 375, 272, 23 },
		{ 375, 273, 23 },
		{ 375, 274, 23 },
		{ 375, 275, 23 },
		{ 375, 276, 23 },
		{ 375, 277, 23 },
		{ 375, 278, 23 },
		{ 375, 279, 23 },
		{ 375, 280, 23 },
		{ 375, 281, 23 },
		{ 375, 282, 23 },
		{ 375, 283, 23 },
		{ 375, 284, 23 },
		{ 375, 285, 23 },
		{ 375, 286, 23 },
		{ 375, 287, 23 },
		{ 375, 288, 23 },
		{ 375, 289, 23 },
		{ 375, 290, 23 },
		{ 375, 291, 23 },
		{ 375, 292, 23 },
		{ 375, 293, 23 },
		{ 375, 294, 23 },
		{ 375, 295, 23 },
		{ 375, 296, 23 },
		{ 375, 297, 23 },
		{ 375, 298, 23 },
		{ 375, 299, 23 },
		{ 375, 300, 23 },
		{ 375, 301, 23 },
		{ 375, 302, 23 },
		{ 375, 303, 23 },
		{ 375, 304, 23 },
		{ 375, 305, 23 },
		{ 375, 306, 23 },
		{ 375, 307, 23 },
		{ 375, 308, 23 },
		{ 375, 309, 23 },
		{ 375, 310, 23 },
		{ 375, 311, 23 },
		{ 375, 312, 23 },
		{ 375, 313, 23 },
		{ 375, 314, 23 },
		{ 375, 315, 23 },
		{ 376, 258, 23 },
		{ 376, 259, 23 },
		{ 376, 260, 23 },
		{ 376, 261, 23 },
		{ 376, 262, 23 },
		{ 376, 263, 23 },
		{ 376, 264, 23 },
		{ 376, 265, 23 },
		{ 376, 266, 23 },
		{ 376, 267, 23 },
		{ 376, 268, 23 },
		{ 376, 269, 23 },
		{ 376, 270, 23 },
		{ 376, 271, 23 },
		{ 376, 272, 23 },
		{ 376, 273, 23 },
		{ 376, 274, 23 },
		{ 376, 275, 23 },
		{ 376, 276, 23 },
		{ 376, 277, 23 },
		{ 376, 278, 23 },
		{ 376, 279, 23 },
		{ 376, 280, 23 },
		{ 376, 281, 23 },
		{ 376, 282, 23 },
		{ 376, 283, 23 },
		{ 376, 284, 23 },
		{ 376, 285, 23 },
		{ 376, 286, 23 },
		{ 376, 287, 23 },
		{ 376, 288, 23 },
		{ 376, 289, 23 },
		{ 376, 290, 23 },
		{ 376, 291, 23 },
		{ 376, 292, 23 },
		{ 376, 293, 23 },
		{ 376, 294, 23 },
		{ 376, 295, 23 },
		{ 376, 296, 23 },
		{ 376, 297, 23 },
		{ 376, 298, 23 },
		{ 376, 299, 23 },
		{ 376, 300, 23 },
		{ 376, 301, 23 },
		{ 376, 302, 23 },
		{ 376, 303, 23 },
		{ 376, 304, 23 },
		{ 376, 305, 23 },
		{ 376, 306, 23 },
		{ 376, 307, 23 },
		{ 376, 308, 23 },
		{ 376, 309, 23 },
		{ 376, 310, 23 },
		{ 376, 311, 23 },
		{ 376, 312, 23 },
		{ 376, 313, 23 },
		{ 376, 314, 23 },
		{ 376, 315, 23 },
		{ 376, 316, 23 },
		{ 377, 258, 23 },
		{ 377, 259, 23 },
		{ 377, 260, 23 },
		{ 377, 261, 23 },
		{ 377, 262, 23 },
		{ 377, 263, 23 },
		{ 377, 264, 23 },
		{ 377, 265, 23 },
		{ 377, 266, 23 },
		{ 377, 267, 23 },
		{ 377, 268, 23 },
		{ 377, 269, 23 },
		{ 377, 270, 23 },
		{ 377, 271, 23 },
		{ 377, 272, 23 },
		{ 377, 273, 23 },
		{ 377, 274, 23 },
		{ 377, 275, 23 },
		{ 377, 276, 23 },
		{ 377, 277, 23 },
		{ 377, 278, 23 },
		{ 377, 279, 23 },
		{ 377, 280, 23 },
		{ 377, 281, 23 },
		{ 377, 282, 23 },
		{ 377, 283, 23 },
		{ 377, 284, 23 },
		{ 377, 285, 23 },
		{ 377, 286, 23 },
		{ 377, 287, 23 },
		{ 377, 288, 23 },
		{ 377, 289, 23 },
		{ 377, 290, 23 },
		{ 377, 291, 23 },
		{ 377, 292, 23 },
		{ 377, 293, 23 },
		{ 377, 294, 23 },
		{ 377, 295, 23 },
		{ 377, 296, 23 },
		{ 377, 297, 23 },
		{ 377, 298, 23 },
		{ 377, 299, 23 },
		{ 377, 300, 23 },
		{ 377, 301, 23 },
		{ 377, 302, 23 },
		{ 377, 303, 23 },
		{ 377, 304, 23 },
		{ 377, 305, 23 },
		{ 377, 306, 23 },
		{ 377, 307, 23 },
		{ 377, 308, 23 },
		{ 377, 309, 23 },
		{ 377, 310, 23 },
		{ 377, 311, 23 },
		{ 377, 312, 23 },
		{ 377, 313, 23 },
		{ 377, 314, 23 },
		{ 377, 315, 23 },
		{ 377, 316, 23 },
		{ 378, 258, 23 },
		{ 378, 259, 23 },
		{ 378, 260, 23 },
		{ 378, 261, 23 },
		{ 378, 262, 23 },
		{ 378, 263, 23 },
		{ 378, 264, 23 },
		{ 378, 265, 23 },
		{ 378, 266, 23 },
		{ 378, 267, 23 },
		{ 378, 268, 23 },
		{ 378, 269, 23 },
		{ 378, 270, 23 },
		{ 378, 271, 23 },
		{ 378, 272, 23 },
		{ 378, 273, 23 },
		{ 378, 274, 23 },
		{ 378, 275, 23 },
		{ 378, 276, 23 },
		{ 378, 277, 23 },
		{ 378, 278, 23 },
		{ 378, 279, 23 },
		{ 378, 280, 23 },
		{ 378, 281, 23 },
		{ 378, 282, 23 },
		{ 378, 283, 23 },
		{ 378, 284, 23 },
		{ 378, 285, 23 },
		{ 378, 286, 23 },
		{ 378, 287, 23 },
		{ 378, 288, 23 },
		{ 378, 289, 23 },
		{ 378, 290, 23 },
		{ 378, 291, 23 },
		{ 378, 292, 23 },
		{ 378, 293, 23 },
		{ 378, 294, 23 },
		{ 378, 295, 23 },
		{ 378, 296, 23 },
		{ 378, 297, 23 },
		{ 378, 298, 23 },
		{ 378, 299, 23 },
		{ 378, 300, 23 },
		{ 378, 301, 23 },
		{ 378, 302, 23 },
		{ 378, 303, 23 },
		{ 378, 304, 23 },
		{ 378, 305, 23 },
		{ 378, 306, 23 },
		{ 378, 307, 23 },
		{ 378, 308, 23 },
		{ 378, 309, 23 },
		{ 378, 310, 23 },
		{ 378, 311, 23 },
		{ 378, 312, 23 },
		{ 378, 313, 23 },
		{ 378, 314, 23 },
		{ 378, 315, 23 },
		{ 378, 316, 23 },
		{ 379, 258, 23 },
		{ 379, 259, 23 },
		{ 379, 260, 23 },
		{ 379, 261, 23 },
		{ 379, 262, 23 },
		{ 379, 263, 23 },
		{ 379, 264, 23 },
		{ 379, 265, 23 },
		{ 379, 266, 23 },
		{ 379, 267, 23 },
		{ 379, 268, 23 },
		{ 379, 269, 23 },
		{ 379, 270, 23 },
		{ 379, 271, 23 },
		{ 379, 272, 23 },
		{ 379, 273, 23 },
		{ 379, 274, 23 },
		{ 379, 275, 23 },
		{ 379, 276, 23 },
		{ 379, 277, 23 },
		{ 379, 278, 23 },
		{ 379, 279, 23 },
		{ 379, 280, 23 },
		{ 379, 281, 23 },
		{ 379, 282, 23 },
		{ 379, 283, 23 },
		{ 379, 284, 23 },
		{ 379, 285, 23 },
		{ 379, 286, 23 },
		{ 379, 287, 23 },
		{ 379, 288, 23 },
		{ 379, 289, 23 },
		{ 379, 290, 23 },
		{ 379, 291, 23 },
		{ 379, 292, 23 },
		{ 379, 293, 23 },
		{ 379, 294, 23 },
		{ 379, 295, 23 },
		{ 379, 296, 23 },
		{ 379, 297, 23 },
		{ 379, 298, 23 },
		{ 379, 299, 23 },
		{ 379, 300, 23 },
		{ 379, 301, 23 },
		{ 379, 302, 23 },
		{ 379, 303, 23 },
		{ 379, 304, 23 },
		{ 379, 305, 23 },
		{ 379, 306, 23 },
		{ 379, 307, 23 },
		{ 379, 308, 23 },
		{ 379, 309, 23 },
		{ 379, 310, 23 },
		{ 379, 311, 23 },
		{ 379, 312, 23 },
		{ 379, 313, 23 },
		{ 379, 314, 23 },
		{ 379, 315, 23 },
		{ 379, 316, 23 },
		{ 379, 317, 23 },
		{ 380, 258, 23 },
		{ 380, 259, 23 },
		{ 380, 260, 23 },
		{ 380, 261, 23 },
		{ 380, 262, 23 },
		{ 380, 263, 23 },
		{ 380, 264, 23 },
		{ 380, 265, 23 },
		{ 380, 266, 23 },
		{ 380, 267, 23 },
		{ 380, 268, 23 },
		{ 380, 269, 23 },
		{ 380, 270, 23 },
		{ 380, 271, 23 },
		{ 380, 272, 23 },
		{ 380, 273, 23 },
		{ 380, 274, 23 },
		{ 380, 275, 23 },
		{ 380, 276, 23 },
		{ 380, 277, 23 },
		{ 380, 278, 23 },
		{ 380, 279, 23 },
		{ 380, 280, 23 },
		{ 380, 281, 23 },
		{ 380, 282, 23 },
		{ 380, 283, 23 },
		{ 380, 284, 23 },
		{ 380, 285, 23 },
		{ 380, 286, 23 },
		{ 380, 287, 23 },
		{ 380, 288, 23 },
		{ 380, 289, 23 },
		{ 380, 290, 23 },
		{ 380, 291, 23 },
		{ 380, 292, 23 },
		{ 380, 293, 23 },
		{ 380, 294, 23 },
		{ 380, 295, 23 },
		{ 380, 296, 23 },
		{ 380, 297, 23 },
		{ 380, 298, 23 },
		{ 380, 299, 23 },
		{ 380, 300, 23 },
		{ 380, 301, 23 },
		{ 380, 302, 23 },
		{ 380, 303, 23 },
		{ 380, 304, 23 },
		{ 380, 305, 23 },
		{ 380, 306, 23 },
		{ 380, 307, 23 },
		{ 380, 308, 23 },
		{ 380, 309, 23 },
		{ 380, 310, 23 },
		{ 380, 311, 23 },
		{ 380, 312, 23 },
		{ 380, 313, 23 },
		{ 380, 314, 23 },
		{ 380, 315, 23 },
		{ 380, 316, 23 },
		{ 380, 317, 23 },
		{ 381, 259, 23 },
		{ 381, 260, 23 },
		{ 381, 261, 23 },
		{ 381, 262, 23 },
		{ 381, 263, 23 },
		{ 381, 264, 23 },
		{ 381, 265, 23 },
		{ 381, 266, 23 },
		{ 381, 267, 23 },
		{ 381, 268, 23 },
		{ 381, 269, 23 },
		{ 381, 270, 23 },
		{ 381, 271, 23 },
		{ 381, 272, 23 },
		{ 381, 273, 23 },
		{ 381, 274, 23 },
		{ 381, 275, 23 },
		{ 381, 276, 23 },
		{ 381, 277, 23 },
		{ 381, 278, 23 },
		{ 381, 279, 23 },
		{ 381, 280, 23 },
		{ 381, 281, 23 },
		{ 381, 282, 23 },
		{ 381, 283, 23 },
		{ 381, 284, 23 },
		{ 381, 285, 23 },
		{ 381, 286, 23 },
		{ 381, 287, 23 },
		{ 381, 288, 23 },
		{ 381, 289, 23 },
		{ 381, 290, 23 },
		{ 381, 291, 23 },
		{ 381, 292, 23 },
		{ 381, 293, 23 },
		{ 381, 294, 23 },
		{ 381, 295, 23 },
		{ 381, 296, 23 },
		{ 381, 297, 23 },
		{ 381, 298, 23 },
		{ 381, 299, 23 },
		{ 381, 300, 23 },
		{ 381, 301, 23 },
		{ 381, 302, 23 },
		{ 381, 303, 23 },
		{ 381, 304, 23 },
		{ 381, 305, 23 },
		{ 381, 306, 23 },
		{ 381, 307, 23 },
		{ 381, 308, 23 },
		{ 381, 309, 23 },
		{ 381, 310, 23 },
		{ 381, 311, 23 },
		{ 381, 312, 23 },
		{ 381, 313, 23 },
		{ 381, 314, 23 },
		{ 381, 315, 23 },
		{ 381, 316, 23 },
		{ 381, 317, 23 },
		{ 382, 260, 23 },
		{ 382, 261, 23 },
		{ 382, 262, 23 },
		{ 382, 263, 23 },
		{ 382, 264, 23 },
		{ 382, 265, 23 },
		{ 382, 266, 23 },
		{ 382, 267, 23 },
		{ 382, 268, 23 },
		{ 382, 269, 23 },
		{ 382, 270, 23 },
		{ 382, 271, 23 },
		{ 382, 272, 23 },
		{ 382, 273, 23 },
		{ 382, 274, 23 },
		{ 382, 275, 23 },
		{ 382, 276, 23 },
		{ 382, 277, 23 },
		{ 382, 278, 23 },
		{ 382, 279, 23 },
		{ 382, 280, 23 },
		{ 382, 281, 23 },
		{ 382, 282, 23 },
		{ 382, 283, 23 },
		{ 382, 284, 23 },
		{ 382, 285, 23 },
		{ 382, 286, 23 },
		{ 382, 287, 23 },
		{ 382, 288, 23 },
		{ 382, 289, 23 },
		{ 382, 290, 23 },
		{ 382, 291, 23 },
		{ 382, 292, 23 },
		{ 382, 293, 23 },
		{ 382, 294, 23 },
		{ 382, 295, 23 },
		{ 382, 296, 23 },
		{ 382, 297, 23 },
		{ 382, 298, 23 },
		{ 382, 299, 23 },
		{ 382, 300, 23 },
		{ 382, 301, 23 },
		{ 382, 302, 23 },
		{ 382, 303, 23 },
		{ 382, 304, 23 },
		{ 382, 305, 23 },
		{ 382, 306, 23 },
		{ 382, 307, 23 },
		{ 382, 308, 23 },
		{ 382, 309, 23 },
		{ 382, 310, 23 },
		{ 382, 311, 23 },
		{ 382, 312, 23 },
		{ 382, 313, 23 },
		{ 382, 314, 23 },
		{ 382, 315, 23 },
		{ 382, 316, 23 },
		{ 383, 264, 23 },
		{ 383, 265, 23 },
		{ 383, 266, 23 },
		{ 383, 267, 23 },
		{ 383, 268, 23 },
		{ 383, 269, 23 },
		{ 383, 270, 23 },
		{ 383, 271, 23 },
		{ 383, 272, 23 },
		{ 383, 273, 23 },
		{ 383, 274, 23 },
		{ 383, 275, 23 },
		{ 383, 276, 23 },
		{ 383, 277, 23 },
		{ 383, 278, 23 },
		{ 383, 279, 23 },
		{ 383, 280, 23 },
		{ 383, 281, 23 },
		{ 383, 282, 23 },
		{ 383, 283, 23 },
		{ 383, 284, 23 },
		{ 383, 285, 23 },
		{ 383, 286, 23 },
		{ 383, 287, 23 },
		{ 383, 288, 23 },
		{ 383, 289, 23 },
		{ 383, 290, 23 },
		{ 383, 291, 23 },
		{ 383, 292, 23 },
		{ 383, 293, 23 },
		{ 383, 294, 23 },
		{ 383, 295, 23 },
		{ 383, 296, 23 },
		{ 383, 297, 23 },
		{ 383, 298, 23 },
		{ 383, 299, 23 },
		{ 383, 300, 23 },
		{ 383, 301, 23 },
		{ 383, 302, 23 },
		{ 383, 303, 23 },
		{ 383, 304, 23 },
		{ 383, 305, 23 },
		{ 383, 306, 23 },
		{ 383, 307, 23 },
		{ 383, 308, 23 },
		{ 383, 309, 23 },
		{ 383, 310, 23 },
		{ 383, 311, 23 },
		{ 383, 312, 23 },
		{ 383, 313, 23 },
		{ 131, 170, 24 },
		{ 131, 171, 24 },
		{ 131, 172, 24 },
		{ 131, 173, 24 },
		{ 131, 174, 24 },
		{ 131, 175, 24 },
		{ 131, 176, 24 },
		{ 131, 177, 24 },
		{ 131, 178, 24 },
		{ 131, 179, 24 },
		{ 131, 180, 24 },
		{ 131, 181, 24 },
		{ 131, 182, 24 },
		{ 131, 183, 24 },
		{ 131, 184, 24 },
		{ 131, 185, 24 },
		{ 131, 186, 24 },
		{ 131, 187, 24 },
		{ 131, 188, 24 },
		{ 131, 189, 24 },
		{ 131, 190, 24 },
		{ 131, 191, 24 },
		{ 131, 192, 24 },
		{ 131, 193, 24 },
		{ 131, 194, 24 },
		{ 131, 195, 24 },
		{ 131, 196, 24 },
		{ 131, 197, 24 },
		{ 131, 198, 24 },
		{ 131, 199, 24 },
		{ 131, 200, 24 },
		{ 131, 201, 24 },
		{ 131, 202, 24 },
		{ 131, 203, 24 },
		{ 131, 204, 24 },
		{ 131, 205, 24 },
		{ 131, 206, 24 },
		{ 131, 207, 24 },
		{ 131, 208, 24 },
		{ 131, 209, 24 },
		{ 131, 210, 24 },
		{ 131, 211, 24 },
		{ 131, 212, 24 },
		{ 131, 213, 24 },
		{ 131, 214, 24 },
		{ 131, 215, 24 },
		{ 131, 216, 24 },
		{ 131, 217, 24 },
		{ 131, 218, 24 },
		{ 132, 167, 24 },
		{ 132, 168, 24 },
		{ 132, 169, 24 },
		{ 132, 170, 24 },
		{ 132, 171, 24 },
		{ 132, 172, 24 },
		{ 132, 173, 24 },
		{ 132, 174, 24 },
		{ 132, 175, 24 },
		{ 132, 176, 24 },
		{ 132, 177, 24 },
		{ 132, 178, 24 },
		{ 132, 179, 24 },
		{ 132, 180, 24 },
		{ 132, 181, 24 },
		{ 132, 182, 24 },
		{ 132, 183, 24 },
		{ 132, 184, 24 },
		{ 132, 185, 24 },
		{ 132, 186, 24 },
		{ 132, 187, 24 },
		{ 132, 188, 24 },
		{ 132, 189, 24 },
		{ 132, 190, 24 },
		{ 132, 191, 24 },
		{ 132, 192, 24 },
		{ 132, 193, 24 },
		{ 132, 194, 24 },
		{ 132, 195, 24 },
		{ 132, 196, 24 },
		{ 132, 197, 24 },
		{ 132, 198, 24 },
		{ 132, 199, 24 },
		{ 132, 200, 24 },
		{ 132, 201, 24 },
		{ 132, 202, 24 },
		{ 132, 203, 24 },
		{ 132, 204, 24 },
		{ 132, 205, 24 },
		{ 132, 206, 24 },
		{ 132, 207, 24 },
		{ 132, 208, 24 },
		{ 132, 209, 24 },
		{ 132, 210, 24 },
		{ 132, 211, 24 },
		{ 132, 212, 24 },
		{ 132, 213, 24 },
		{ 132, 214, 24 },
		{ 132, 215, 24 },
		{ 132, 216, 24 },
		{ 132, 217, 24 },
		{ 132, 218, 24 },
		{ 132, 219, 24 },
		{ 132, 220, 24 },
		{ 132, 221, 24 },
		{ 132, 222, 24 },
		{ 133, 166, 24 },
		{ 133, 167, 24 },
		{ 133, 168, 24 },
		{ 133, 169, 24 },
		{ 133, 170, 24 },
		{ 133, 171, 24 },
		{ 133, 172, 24 },
		{ 133, 173, 24 },
		{ 133, 174, 24 },
		{ 133, 175, 24 },
		{ 133, 176, 24 },
		{ 133, 177, 24 },
		{ 133, 178, 24 },
		{ 133, 179, 24 },
		{ 133, 180, 24 },
		{ 133, 181, 24 },
		{ 133, 182, 24 },
		{ 133, 183, 24 },
		{ 133, 184, 24 },
		{ 133, 185, 24 },
		{ 133, 186, 24 },
		{ 133, 187, 24 },
		{ 133, 188, 24 },
		{ 133, 189, 24 },
		{ 133, 190, 24 },
		{ 133, 191, 24 },
		{ 133, 192, 24 },
		{ 133, 193, 24 },
		{ 133, 194, 24 },
		{ 133, 195, 24 },
		{ 133, 196, 24 },
		{ 133, 197, 24 },
		{ 133, 198, 24 },
		{ 133, 199, 24 },
		{ 133, 200, 24 },
		{ 133, 201, 24 },
		{ 133, 202, 24 },
		{ 133, 203, 24 },
		{ 133, 204, 24 },
		{ 133, 205, 24 },
		{ 133, 206, 24 },
		{ 133, 207, 24 },
		{ 133, 208, 24 },
		{ 133, 209, 24 },
		{ 133, 210, 24 },
		{ 133, 211, 24 },
		{ 133, 212, 24 },
		{ 133, 213, 24 },
		{ 133, 214, 24 },
		{ 133, 215, 24 },
		{ 133, 216, 24 },
		{ 133, 217, 24 },
		{ 133, 218, 24 },
		{ 133, 219, 24 },
		{ 133, 220, 24 },
		{ 133, 221, 24 },
		{ 133, 222, 24 },
		{ 133, 223, 24 },
		{ 133, 224, 24 },
		{ 134, 166, 24 },
		{ 134, 167, 24 },
		{ 134, 168, 24 },
		{ 134, 169, 24 },
		{ 134, 170, 24 },
		{ 134, 171, 24 },
		{ 134, 172, 24 },
		{ 134, 173, 24 },
		{ 134, 174, 24 },
		{ 134, 175, 24 },
		{ 134, 176, 24 },
		{ 134, 177, 24 },
		{ 134, 178, 24 },
		{ 134, 179, 24 },
		{ 134, 180, 24 },
		{ 134, 181, 24 },
		{ 134, 182, 24 },
		{ 134, 183, 24 },
		{ 134, 184, 24 },
		{ 134, 185, 24 },
		{ 134, 186, 24 },
		{ 134, 187, 24 },
		{ 134, 188, 24 },
		{ 134, 189, 24 },
		{ 134, 190, 24 },
		{ 134, 191, 24 },
		{ 134, 192, 24 },
		{ 134, 193, 24 },
		{ 134, 194, 24 },
		{ 134, 195, 24 },
		{ 134, 196, 24 },
		{ 134, 197, 24 },
		{ 134, 198, 24 },
		{ 134, 199, 24 },
		{ 134, 200, 24 },
		{ 134, 201, 24 },
		{ 134, 202, 24 },
		{ 134, 203, 24 },
		{ 134, 204, 24 },
		{ 134, 205, 24 },
		{ 134, 206, 24 },
		{ 134, 207, 24 },
		{ 134, 208, 24 },
		{ 134, 209, 24 },
		{ 134, 210, 24 },
		{ 134, 211, 24 },
		{ 134, 212, 24 },
		{ 134, 213, 24 },
		{ 134, 214, 24 },
		{ 134, 215, 24 },
		{ 134, 216, 24 },
		{ 134, 217, 24 },
		{ 134, 218, 24 },
		{ 134, 219, 24 },
		{ 134, 220, 24 },
		{ 134, 221, 24 },
		{ 134, 222, 24 },
		{ 134, 223, 24 },
		{ 134, 224, 24 },
		{ 134, 225, 24 },
		{ 135, 167, 24 },
		{ 135, 168, 24 },
		{ 135, 169, 24 },
		{ 135, 170, 24 },
		{ 135, 171, 24 },
		{ 135, 172, 24 },
		{ 135, 173, 24 },
		{ 135, 174, 24 },
		{ 135, 175, 24 },
		{ 135, 176, 24 },
		{ 135, 177, 24 },
		{ 135, 178, 24 },
		{ 135, 179, 24 },
		{ 135, 180, 24 },
		{ 135, 181, 24 },
		{ 135, 182, 24 },
		{ 135, 183, 24 },
		{ 135, 184, 24 },
		{ 135, 185, 24 },
		{ 135, 186, 24 },
		{ 135, 187, 24 },
		{ 135, 188, 24 },
		{ 135, 189, 24 },
		{ 135, 190, 24 },
		{ 135, 191, 24 },
		{ 135, 192, 24 },
		{ 135, 193, 24 },
		{ 135, 194, 24 },
		{ 135, 195, 24 },
		{ 135, 196, 24 },
		{ 135, 197, 24 },
		{ 135, 198, 24 },
		{ 135, 199, 24 },
		{ 135, 200, 24 },
		{ 135, 201, 24 },
		{ 135, 202, 24 },
		{ 135, 203, 24 },
		{ 135, 204, 24 },
		{ 135, 205, 24 },
		{ 135, 206, 24 },
		{ 135, 207, 24 },
		{ 135, 208, 24 },
		{ 135, 209, 24 },
		{ 135, 210, 24 },
		{ 135, 211, 24 },
		{ 135, 212, 24 },
		{ 135, 213, 24 },
		{ 135, 214, 24 },
		{ 135, 215, 24 },
		{ 135, 216, 24 },
		{ 135, 217, 24 },
		{ 135, 218, 24 },
		{ 135, 219, 24 },
		{ 135, 220, 24 },
		{ 135, 221, 24 },
		{ 135, 222, 24 },
		{ 135, 223, 24 },
		{ 135, 224, 24 },
		{ 135, 225, 24 },
		{ 136, 167, 24 },
		{ 136, 168, 24 },
		{ 136, 169, 24 },
		{ 136, 170, 24 },
		{ 136, 171, 24 },
		{ 136, 172, 24 },
		{ 136, 173, 24 },
		{ 136, 174, 24 },
		{ 136, 175, 24 },
		{ 136, 176, 24 },
		{ 136, 177, 24 },
		{ 136, 178, 24 },
		{ 136, 179, 24 },
		{ 136, 180, 24 },
		{ 136, 181, 24 },
		{ 136, 182, 24 },
		{ 136, 183, 24 },
		{ 136, 184, 24 },
		{ 136, 185, 24 },
		{ 136, 186, 24 },
		{ 136, 187, 24 },
		{ 136, 188, 24 },
		{ 136, 189, 24 },
		{ 136, 190, 24 },
		{ 136, 191, 24 },
		{ 136, 192, 24 },
		{ 136, 193, 24 },
		{ 136, 194, 24 },
		{ 136, 195, 24 },
		{ 136, 196, 24 },
		{ 136, 197, 24 },
		{ 136, 198, 24 },
		{ 136, 199, 24 },
		{ 136, 200, 24 },
		{ 136, 201, 24 },
		{ 136, 202, 24 },
		{ 136, 203, 24 },
		{ 136, 204, 24 },
		{ 136, 205, 24 },
		{ 136, 206, 24 },
		{ 136, 207, 24 },
		{ 136, 208, 24 },
		{ 136, 209, 24 },
		{ 136, 210, 24 },
		{ 136, 211, 24 },
		{ 136, 212, 24 },
		{ 136, 213, 24 },
		{ 136, 214, 24 },
		{ 136, 215, 24 },
		{ 136, 216, 24 },
		{ 136, 217, 24 },
		{ 136, 218, 24 },
		{ 136, 219, 24 },
		{ 136, 220, 24 },
		{ 136, 221, 24 },
		{ 136, 222, 24 },
		{ 136, 223, 24 },
		{ 136, 224, 24 },
		{ 136, 225, 24 },
		{ 136, 226, 24 },
		{ 137, 168, 24 },
		{ 137, 169, 24 },
		{ 137, 170, 24 },
		{ 137, 171, 24 },
		{ 137, 172, 24 },
		{ 137, 173, 24 },
		{ 137, 174, 24 },
		{ 137, 175, 24 },
		{ 137, 176, 24 },
		{ 137, 177, 24 },
		{ 137, 178, 24 },
		{ 137, 179, 24 },
		{ 137, 180, 24 },
		{ 137, 181, 24 },
		{ 137, 182, 24 },
		{ 137, 183, 24 },
		{ 137, 184, 24 },
		{ 137, 185, 24 },
		{ 137, 186, 24 },
		{ 137, 187, 24 },
		{ 137, 188, 24 },
		{ 137, 189, 24 },
		{ 137, 190, 24 },
		{ 137, 191, 24 },
		{ 137, 192, 24 },
		{ 137, 193, 24 },
		{ 137, 194, 24 },
		{ 137, 195, 24 },
		{ 137, 196, 24 },
		{ 137, 197, 24 },
		{ 137, 198, 24 },
		{ 137, 199, 24 },
		{ 137, 200, 24 },
		{ 137, 201, 24 },
		{ 137, 202, 24 },
		{ 137, 203, 24 },
		{ 137, 204, 24 },
		{ 137, 205, 24 },
		{ 137, 206, 24 },
		{ 137, 207, 24 },
		{ 137, 208, 24 },
		{ 137, 209, 24 },
		{ 137, 210, 24 },
		{ 137, 211, 24 },
		{ 137, 212, 24 },
		{ 137, 213, 24 },
		{ 137, 214, 24 },
		{ 137, 215, 24 },
		{ 137, 216, 24 },
		{ 137, 217, 24 },
		{ 137, 218, 24 },
		{ 137, 219, 24 },
		{ 137, 220, 24 },
		{ 137, 221, 24 },
		{ 137, 222, 24 },
		{ 137, 223, 24 },
		{ 137, 224, 24 },
		{ 137, 225, 24 },
		{ 137, 226, 24 },
		{ 138, 169, 24 },
		{ 138, 170, 24 },
		{ 138, 171, 24 },
		{ 138, 172, 24 },
		{ 138, 173, 24 },
		{ 138, 174, 24 },
		{ 138, 175, 24 },
		{ 138, 176, 24 },
		{ 138, 177, 24 },
		{ 138, 178, 24 },
		{ 138, 179, 24 },
		{ 138, 180, 24 },
		{ 138, 181, 24 },
		{ 138, 182, 24 },
		{ 138, 183, 24 },
		{ 138, 184, 24 },
		{ 138, 185, 24 },
		{ 138, 186, 24 },
		{ 138, 187, 24 },
		{ 138, 188, 24 },
		{ 138, 189, 24 },
		{ 138, 190, 24 },
		{ 138, 191, 24 },
		{ 138, 192, 24 },
		{ 138, 193, 24 },
		{ 138, 194, 24 },
		{ 138, 195, 24 },
		{ 138, 196, 24 },
		{ 138, 197, 24 },
		{ 138, 198, 24 },
		{ 138, 199, 24 },
		{ 138, 200, 24 },
		{ 138, 201, 24 },
		{ 138, 202, 24 },
		{ 138, 203, 24 },
		{ 138, 204, 24 },
		{ 138, 205, 24 },
		{ 138, 206, 24 },
		{ 138, 207, 24 },
		{ 138, 208, 24 },
		{ 138, 209, 24 },
		{ 138, 210, 24 },
		{ 138, 211, 24 },
		{ 138, 212, 24 },
		{ 138, 213, 24 },
		{ 138, 214, 24 },
		{ 138, 215, 24 },
		{ 138, 216, 24 },
		{ 138, 217, 24 },
		{ 138, 218, 24 },
		{ 138, 219, 24 },
		{ 138, 220, 24 },
		{ 138, 221, 24 },
		{ 138, 222, 24 },
		{ 138, 223, 24 },
		{ 138, 224, 24 },
		{ 138, 225, 24 },
		{ 138, 226, 24 },
		{ 139, 169, 24 },
		{ 139, 170, 24 },
		{ 139, 171, 24 },
		{ 139, 172, 24 },
		{ 139, 173, 24 },
		{ 139, 174, 24 },
		{ 139, 175, 24 },
		{ 139, 176, 24 },
		{ 139, 177, 24 },
		{ 139, 178, 24 },
		{ 139, 179, 24 },
		{ 139, 180, 24 },
		{ 139, 181, 24 },
		{ 139, 182, 24 },
		{ 139, 183, 24 },
		{ 139, 184, 24 },
		{ 139, 185, 24 },
		{ 139, 186, 24 },
		{ 139, 187, 24 },
		{ 139, 188, 24 },
		{ 139, 189, 24 },
		{ 139, 190, 24 },
		{ 139, 191, 24 },
		{ 139, 192, 24 },
		{ 139, 193, 24 },
		{ 139, 194, 24 },
		{ 139, 195, 24 },
		{ 139, 196, 24 },
		{ 139, 197, 24 },
		{ 139, 198, 24 },
		{ 139, 199, 24 },
		{ 139, 200, 24 },
		{ 139, 201, 24 },
		{ 139, 202, 24 },
		{ 139, 203, 24 },
		{ 139, 204, 24 },
		{ 139, 205, 24 },
		{ 139, 206, 24 },
		{ 139, 207, 24 },
		{ 139, 208, 24 },
		{ 139, 209, 24 },
		{ 139, 210, 24 },
		{ 139, 211, 24 },
		{ 139, 212, 24 },
		{ 139, 213, 24 },
		{ 139, 214, 24 },
		{ 139, 215, 24 },
		{ 139, 216, 24 },
		{ 139, 217, 24 },
		{ 139, 218, 24 },
		{ 139, 219, 24 },
		{ 139, 220, 24 },
		{ 139, 221, 24 },
		{ 139, 222, 24 },
		{ 139, 223, 24 },
		{ 139, 224, 24 },
		{ 139, 225, 24 },
		{ 139, 226, 24 },
		{ 139, 227, 24 },
		{ 140, 170, 24 },
		{ 140, 171, 24 },
		{ 140, 172, 24 },
		{ 140, 173, 24 },
		{ 140, 174, 24 },
		{ 140, 175, 24 },
		{ 140, 176, 24 },
		{ 140, 177, 24 },
		{ 140, 178, 24 },
		{ 140, 179, 24 },
		{ 140, 180, 24 },
		{ 140, 181, 24 },
		{ 140, 182, 24 },
		{ 140, 183, 24 },
		{ 140, 184, 24 },
		{ 140, 185, 24 },
		{ 140, 186, 24 },
		{ 140, 187, 24 },
		{ 140, 188, 24 },
		{ 140, 189, 24 },
		{ 140, 190, 24 },
		{ 140, 191, 24 },
		{ 140, 192, 24 },
		{ 140, 193, 24 },
		{ 140, 194, 24 },
		{ 140, 195, 24 },
		{ 140, 196, 24 },
		{ 140, 197, 24 },
		{ 140, 198, 24 },
		{ 140, 199, 24 },
		{ 140, 200, 24 },
		{ 140, 201, 24 },
		{ 140, 202, 24 },
		{ 140, 203, 24 },
		{ 140, 204, 24 },
		{ 140, 205, 24 },
		{ 140, 206, 24 },
		{ 140, 207, 24 },
		{ 140, 208, 24 },
		{ 140, 209, 24 },
		{ 140, 210, 24 },
		{ 140, 211, 24 },
		{ 140, 212, 24 },
		{ 140, 213, 24 },
		{ 140, 214, 24 },
		{ 140, 215, 24 },
		{ 140, 216, 24 },
		{ 140, 217, 24 },
		{ 140, 218, 24 },
		{ 140, 219, 24 },
		{ 140, 220, 24 },
		{ 140, 221, 24 },
		{ 140, 222, 24 },
		{ 140, 223, 24 },
		{ 140, 224, 24 },
		{ 140, 225, 24 },
		{ 140, 226, 24 },
		{ 140, 227, 24 },
		{ 141, 171, 24 },
		{ 141, 172, 24 },
		{ 141, 173, 24 },
		{ 141, 174, 24 },
		{ 141, 175, 24 },
		{ 141, 176, 24 },
		{ 141, 177, 24 },
		{ 141, 178, 24 },
		{ 141, 179, 24 },
		{ 141, 180, 24 },
		{ 141, 181, 24 },
		{ 141, 182, 24 },
		{ 141, 183, 24 },
		{ 141, 184, 24 },
		{ 141, 185, 24 },
		{ 141, 186, 24 },
		{ 141, 187, 24 },
		{ 141, 188, 24 },
		{ 141, 189, 24 },
		{ 141, 190, 24 },
		{ 141, 191, 24 },
		{ 141, 192, 24 },
		{ 141, 193, 24 },
		{ 141, 194, 24 },
		{ 141, 195, 24 },
		{ 141, 196, 24 },
		{ 141, 197, 24 },
		{ 141, 198, 24 },
		{ 141, 199, 24 },
		{ 141, 200, 24 },
		{ 141, 201, 24 },
		{ 141, 202, 24 },
		{ 141, 203, 24 },
		{ 141, 204, 24 },
		{ 141, 205, 24 },
		{ 141, 206, 24 },
		{ 141, 207, 24 },
		{ 141, 208, 24 },
		{ 141, 209, 24 },
		{ 141, 210, 24 },
		{ 141, 211, 24 },
		{ 141, 212, 24 },
		{ 141, 213, 24 },
		{ 141, 214, 24 },
		{ 141, 215, 24 },
		{ 141, 216, 24 },
		{ 141, 217, 24 },
		{ 141, 218, 24 },
		{ 141, 219, 24 },
		{ 141, 220, 24 },
		{ 141, 221, 24 },
		{ 141, 222, 24 },
		{ 141, 223, 24 },
		{ 141, 224, 24 },
		{ 141, 225, 24 },
		{ 141, 226, 24 },
		{ 141, 227, 24 },
		{ 142, 171, 24 },
		{ 142, 172, 24 },
		{ 142, 173, 24 },
		{ 142, 174, 24 },
		{ 142, 175, 24 },
		{ 142, 176, 24 },
		{ 142, 177, 24 },
		{ 142, 178, 24 },
		{ 142, 179, 24 },
		{ 142, 180, 24 },
		{ 142, 181, 24 },
		{ 142, 182, 24 },
		{ 142, 183, 24 },
		{ 142, 184, 24 },
		{ 142, 185, 24 },
		{ 142, 186, 24 },
		{ 142, 187, 24 },
		{ 142, 188, 24 },
		{ 142, 189, 24 },
		{ 142, 190, 24 },
		{ 142, 191, 24 },
		{ 142, 192, 24 },
		{ 142, 193, 24 },
		{ 142, 194, 24 },
		{ 142, 195, 24 },
		{ 142, 196, 24 },
		{ 142, 197, 24 },
		{ 142, 198, 24 },
		{ 142, 199, 24 },
		{ 142, 200, 24 },
		{ 142, 201, 24 },
		{ 142, 202, 24 },
		{ 142, 203, 24 },
		{ 142, 204, 24 },
		{ 142, 205, 24 },
		{ 142, 206, 24 },
		{ 142, 207, 24 },
		{ 142, 208, 24 },
		{ 142, 209, 24 },
		{ 142, 210, 24 },
		{ 142, 211, 24 },
		{ 142, 212, 24 },
		{ 142, 213, 24 },
		{ 142, 214, 24 },
		{ 142, 215, 24 },
		{ 142, 216, 24 },
		{ 142, 217, 24 },
		{ 142, 218, 24 },
		{ 142, 219, 24 },
		{ 142, 220, 24 },
		{ 142, 221, 24 },
		{ 142, 222, 24 },
		{ 142, 223, 24 },
		{ 142, 224, 24 },
		{ 142, 225, 24 },
		{ 142, 226, 24 },
		{ 142, 227, 24 },
		{ 142, 228, 24 },
		{ 143, 172, 24 },
		{ 143, 173, 24 },
		{ 143, 174, 24 },
		{ 143, 175, 24 },
		{ 143, 176, 24 },
		{ 143, 177, 24 },
		{ 143, 178, 24 },
		{ 143, 179, 24 },
		{ 143, 180, 24 },
		{ 143, 181, 24 },
		{ 143, 182, 24 },
		{ 143, 183, 24 },
		{ 143, 184, 24 },
		{ 143, 185, 24 },
		{ 143, 186, 24 },
		{ 143, 187, 24 },
		{ 143, 188, 24 },
		{ 143, 189, 24 },
		{ 143, 190, 24 },
		{ 143, 191, 24 },
		{ 143, 192, 24 },
		{ 143, 193, 24 },
		{ 143, 194, 24 },
		{ 143, 195, 24 },
		{ 143, 196, 24 },
		{ 143, 197, 24 },
		{ 143, 198, 24 },
		{ 143, 199, 24 },
		{ 143, 200, 24 },
		{ 143, 201, 24 },
		{ 143, 202, 24 },
		{ 143, 203, 24 },
		{ 143, 204, 24 },
		{ 143, 205, 24 },
		{ 143, 206, 24 },
		{ 143, 207, 24 },
		{ 143, 208, 24 },
		{ 143, 209, 24 },
		{ 143, 210, 24 },
		{ 143, 211, 24 },
		{ 143, 212, 24 },
		{ 143, 213, 24 },
		{ 143, 214, 24 },
		{ 143, 215, 24 },
		{ 143, 216, 24 },
		{ 143, 217, 24 },
		{ 143, 218, 24 },
		{ 143, 219, 24 },
		{ 143, 220, 24 },
		{ 143, 221, 24 },
		{ 143, 222, 24 },
		{ 143, 223, 24 },
		{ 143, 224, 24 },
		{ 143, 225, 24 },
		{ 143, 226, 24 },
		{ 143, 227, 24 },
		{ 143, 228, 24 },
		{ 144, 173, 24 },
		{ 144, 174, 24 },
		{ 144, 175, 24 },
		{ 144, 176, 24 },
		{ 144, 177, 24 },
		{ 144, 178, 24 },
		{ 144, 179, 24 },
		{ 144, 180, 24 },
		{ 144, 181, 24 },
		{ 144, 182, 24 },
		{ 144, 183, 24 },
		{ 144, 184, 24 },
		{ 144, 185, 24 },
		{ 144, 186, 24 },
		{ 144, 187, 24 },
		{ 144, 188, 24 },
		{ 144, 189, 24 },
		{ 144, 190, 24 },
		{ 144, 191, 24 },
		{ 144, 192, 24 },
		{ 144, 193, 24 },
		{ 144, 194, 24 },
		{ 144, 195, 24 },
		{ 144, 196, 24 },
		{ 144, 197, 24 },
		{ 144, 198, 24 },
		{ 144, 199, 24 },
		{ 144, 200, 24 },
		{ 144, 201, 24 },
		{ 144, 202, 24 },
		{ 144, 203, 24 },
		{ 144, 204, 24 },
		{ 144, 205, 24 },
		{ 144, 206, 24 },
		{ 144, 207, 24 },
		{ 144, 208, 24 },
		{ 144, 209, 24 },
		{ 144, 210, 24 },
		{ 144, 211, 24 },
		{ 144, 212, 24 },
		{ 144, 213, 24 },
		{ 144, 214, 24 },
		{ 144, 215, 24 },
		{ 144, 216, 24 },
		{ 144, 217, 24 },
		{ 144, 218, 24 },
		{ 144, 219, 24 },
		{ 144, 220, 24 },
		{ 144, 221, 24 },
		{ 144, 222, 24 },
		{ 144, 223, 24 },
		{ 144, 224, 24 },
		{ 144, 225, 24 },
		{ 144, 226, 24 },
		{ 144, 227, 24 },
		{ 144, 228, 24 },
		{ 145, 174, 24 },
		{ 145, 175, 24 },
		{ 145, 176, 24 },
		{ 145, 177, 24 },
		{ 145, 178, 24 },
		{ 145, 179, 24 },
		{ 145, 180, 24 },
		{ 145, 181, 24 },
		{ 145, 182, 24 },
		{ 145, 183, 24 },
		{ 145, 184, 24 },
		{ 145, 185, 24 },
		{ 145, 186, 24 },
		{ 145, 187, 24 },
		{ 145, 188, 24 },
		{ 145, 189, 24 },
		{ 145, 190, 24 },
		{ 145, 191, 24 },
		{ 145, 192, 24 },
		{ 145, 193, 24 },
		{ 145, 194, 24 },
		{ 145, 195, 24 },
		{ 145, 196, 24 },
		{ 145, 197, 24 },
		{ 145, 198, 24 },
		{ 145, 199, 24 },
		{ 145, 200, 24 },
		{ 145, 201, 24 },
		{ 145, 202, 24 },
		{ 145, 203, 24 },
		{ 145, 204, 24 },
		{ 145, 205, 24 },
		{ 145, 206, 24 },
		{ 145, 207, 24 },
		{ 145, 208, 24 },
		{ 145, 209, 24 },
		{ 145, 210, 24 },
		{ 145, 211, 24 },
		{ 145, 212, 24 },
		{ 145, 213, 24 },
		{ 145, 214, 24 },
		{ 145, 215, 24 },
		{ 145, 216, 24 },
		{ 145, 217, 24 },
		{ 145, 218, 24 },
		{ 145, 219, 24 },
		{ 145, 220, 24 },
		{ 145, 221, 24 },
		{ 145, 222, 24 },
		{ 145, 223, 24 },
		{ 145, 224, 24 },
		{ 145, 225, 24 },
		{ 145, 226, 24 },
		{ 145, 227, 24 },
		{ 145, 228, 24 },
		{ 146, 174, 24 },
		{ 146, 175, 24 },
		{ 146, 176, 24 },
		{ 146, 177, 24 },
		{ 146, 178, 24 },
		{ 146, 179, 24 },
		{ 146, 180, 24 },
		{ 146, 181, 24 },
		{ 146, 182, 24 },
		{ 146, 183, 24 },
		{ 146, 184, 24 },
		{ 146, 185, 24 },
		{ 146, 186, 24 },
		{ 146, 187, 24 },
		{ 146, 188, 24 },
		{ 146, 189, 24 },
		{ 146, 190, 24 },
		{ 146, 191, 24 },
		{ 146, 192, 24 },
		{ 146, 193, 24 },
		{ 146, 194, 24 },
		{ 146, 195, 24 },
		{ 146, 196, 24 },
		{ 146, 197, 24 },
		{ 146, 198, 24 },
		{ 146, 199, 24 },
		{ 146, 200, 24 },
		{ 146, 201, 24 },
		{ 146, 202, 24 },
		{ 146, 203, 24 },
		{ 146, 204, 24 },
		{ 146, 205, 24 },
		{ 146, 206, 24 },
		{ 146, 207, 24 },
		{ 146, 208, 24 },
		{ 146, 209, 24 },
		{ 146, 210, 24 },
		{ 146, 211, 24 },
		{ 146, 212, 24 },
		{ 146, 213, 24 },
		{ 146, 214, 24 },
		{ 146, 215, 24 },
		{ 146, 216, 24 },
		{ 146, 217, 24 },
		{ 146, 218, 24 },
		{ 146, 219, 24 },
		{ 146, 220, 24 },
		{ 146, 221, 24 },
		{ 146, 222, 24 },
		{ 146, 223, 24 },
		{ 146, 224, 24 },
		{ 146, 225, 24 },
		{ 146, 226, 24 },
		{ 146, 227, 24 },
		{ 146, 228, 24 },
		{ 146, 229, 24 },
		{ 147, 175, 24 },
		{ 147, 176, 24 },
		{ 147, 177, 24 },
		{ 147, 178, 24 },
		{ 147, 179, 24 },
		{ 147, 180, 24 },
		{ 147, 181, 24 },
		{ 147, 182, 24 },
		{ 147, 183, 24 },
		{ 147, 184, 24 },
		{ 147, 185, 24 },
		{ 147, 186, 24 },
		{ 147, 187, 24 },
		{ 147, 188, 24 },
		{ 147, 189, 24 },
		{ 147, 190, 24 },
		{ 147, 191, 24 },
		{ 147, 192, 24 },
		{ 147, 193, 24 },
		{ 147, 194, 24 },
		{ 147, 195, 24 },
		{ 147, 196, 24 },
		{ 147, 197, 24 },
		{ 147, 198, 24 },
		{ 147, 199, 24 },
		{ 147, 200, 24 },
		{ 147, 201, 24 },
		{ 147, 202, 24 },
		{ 147, 203, 24 },
		{ 147, 204, 24 },
		{ 147, 205, 24 },
		{ 147, 206, 24 },
		{ 147, 207, 24 },
		{ 147, 208, 24 },
		{ 147, 209, 24 },
		{ 147, 210, 24 },
		{ 147, 211, 24 },
		{ 147, 212, 24 },
		{ 147, 213, 24 },
		{ 147, 214, 24 },
		{ 147, 215, 24 },
		{ 147, 216, 24 },
		{ 147, 217, 24 },
		{ 147, 218, 24 },
		{ 147, 219, 24 },
		{ 147, 220, 24 },
		{ 147, 221, 24 },
		{ 147, 222, 24 },
		{ 147, 223, 24 },
		{ 147, 224, 24 },
		{ 147, 225, 24 },
		{ 147, 226, 24 },
		{ 147, 227, 24 },
		{ 147, 228, 24 },
		{ 147, 229, 24 },
		{ 148, 176, 24 },
		{ 148, 177, 24 },
		{ 148, 178, 24 },
		{ 148, 179, 24 },
		{ 148, 180, 24 },
		{ 148, 181, 24 },
		{ 148, 182, 24 },
		{ 148, 183, 24 },
		{ 148, 184, 24 },
		{ 148, 185, 24 },
		{ 148, 186, 24 },
		{ 148, 187, 24 },
		{ 148, 188, 24 },
		{ 148, 189, 24 },
		{ 148, 190, 24 },
		{ 148, 191, 24 },
		{ 148, 192, 24 },
		{ 148, 193, 24 },
		{ 148, 194, 24 },
		{ 148, 195, 24 },
		{ 148, 196, 24 },
		{ 148, 197, 24 },
		{ 148, 198, 24 },
		{ 148, 199, 24 },
		{ 148, 200, 24 },
		{ 148, 201, 24 },
		{ 148, 202, 24 },
		{ 148, 203, 24 },
		{ 148, 204, 24 },
		{ 148, 205, 24 },
		{ 148, 206, 24 },
		{ 148, 207, 24 },
		{ 148, 208, 24 },
		{ 148, 209, 24 },
		{ 148, 210, 24 },
		{ 148, 211, 24 },
		{ 148, 212, 24 },
		{ 148, 213, 24 },
		{ 148, 214, 24 },
		{ 148, 215, 24 },
		{ 148, 216, 24 },
		{ 148, 217, 24 },
		{ 148, 218, 24 },
		{ 148, 219, 24 },
		{ 148, 220, 24 },
		{ 148, 221, 24 },
		{ 148, 222, 24 },
		{ 148, 223, 24 },
		{ 148, 224, 24 },
		{ 148, 225, 24 },
		{ 148, 226, 24 },
		{ 148, 227, 24 },
		{ 148, 228, 24 },
		{ 148, 229, 24 },
		{ 149, 177, 24 },
		{ 149, 178, 24 },
		{ 149, 179, 24 },
		{ 149, 180, 24 },
		{ 149, 181, 24 },
		{ 149, 182, 24 },
		{ 149, 183, 24 },
		{ 149, 184, 24 },
		{ 149, 185, 24 },
		{ 149, 186, 24 },
		{ 149, 187, 24 },
		{ 149, 188, 24 },
		{ 149, 189, 24 },
		{ 149, 190, 24 },
		{ 149, 191, 24 },
		{ 149, 192, 24 },
		{ 149, 193, 24 },
		{ 149, 194, 24 },
		{ 149, 195, 24 },
		{ 149, 196, 24 },
		{ 149, 197, 24 },
		{ 149, 198, 24 },
		{ 149, 199, 24 },
		{ 149, 200, 24 },
		{ 149, 201, 24 },
		{ 149, 202, 24 },
		{ 149, 203, 24 },
		{ 149, 204, 24 },
		{ 149, 205, 24 },
		{ 149, 206, 24 },
		{ 149, 207, 24 },
		{ 149, 208, 24 },
		{ 149, 209, 24 },
		{ 149, 210, 24 },
		{ 149, 211, 24 },
		{ 149, 212, 24 },
		{ 149, 213, 24 },
		{ 149, 214, 24 },
		{ 149, 215, 24 },
		{ 149, 216, 24 },
		{ 149, 217, 24 },
		{ 149, 218, 24 },
		{ 149, 219, 24 },
		{ 149, 220, 24 },
		{ 149, 221, 24 },
		{ 149, 222, 24 },
		{ 149, 223, 24 },
		{ 149, 224, 24 },
		{ 149, 225, 24 },
		{ 149, 226, 24 },
		{ 149, 227, 24 },
		{ 149, 228, 24 },
		{ 149, 229, 24 },
		{ 150, 177, 24 },
		{ 150, 178, 24 },
		{ 150, 179, 24 },
		{ 150, 180, 24 },
		{ 150, 181, 24 },
		{ 150, 182, 24 },
		{ 150, 183, 24 },
		{ 150, 184, 24 },
		{ 150, 185, 24 },
		{ 150, 186, 24 },
		{ 150, 187, 24 },
		{ 150, 188, 24 },
		{ 150, 189, 24 },
		{ 150, 190, 24 },
		{ 150, 191, 24 },
		{ 150, 192, 24 },
		{ 150, 193, 24 },
		{ 150, 194, 24 },
		{ 150, 195, 24 },
		{ 150, 196, 24 },
		{ 150, 197, 24 },
		{ 150, 198, 24 },
		{ 150, 199, 24 },
		{ 150, 200, 24 },
		{ 150, 201, 24 },
		{ 150, 202, 24 },
		{ 150, 203, 24 },
		{ 150, 204, 24 },
		{ 150, 205, 24 },
		{ 150, 206, 24 },
		{ 150, 207, 24 },
		{ 150, 208, 24 },
		{ 150, 209, 24 },
		{ 150, 210, 24 },
		{ 150, 211, 24 },
		{ 150, 212, 24 },
		{ 150, 213, 24 },
		{ 150, 214, 24 },
		{ 150, 215, 24 },
		{ 150, 216, 24 },
		{ 150, 217, 24 },
		{ 150, 218, 24 },
		{ 150, 219, 24 },
		{ 150, 220, 24 },
		{ 150, 221, 24 },
		{ 150, 222, 24 },
		{ 150, 223, 24 },
		{ 150, 224, 24 },
		{ 150, 225, 24 },
		{ 150, 226, 24 },
		{ 150, 227, 24 },
		{ 150, 228, 24 },
		{ 150, 229, 24 },
		{ 150, 230, 24 },
		{ 151, 178, 24 },
		{ 151, 179, 24 },
		{ 151, 180, 24 },
		{ 151, 181, 24 },
		{ 151, 182, 24 },
		{ 151, 183, 24 },
		{ 151, 184, 24 },
		{ 151, 185, 24 },
		{ 151, 186, 24 },
		{ 151, 187, 24 },
		{ 151, 188, 24 },
		{ 151, 189, 24 },
		{ 151, 190, 24 },
		{ 151, 191, 24 },
		{ 151, 192, 24 },
		{ 151, 193, 24 },
		{ 151, 194, 24 },
		{ 151, 195, 24 },
		{ 151, 196, 24 },
		{ 151, 197, 24 },
		{ 151, 198, 24 },
		{ 151, 199, 24 },
		{ 151, 200, 24 },
		{ 151, 201, 24 },
		{ 151, 202, 24 },
		{ 151, 203, 24 },
		{ 151, 204, 24 },
		{ 151, 205, 24 },
		{ 151, 206, 24 },
		{ 151, 207, 24 },
		{ 151, 208, 24 },
		{ 151, 209, 24 },
		{ 151, 210, 24 },
		{ 151, 211, 24 },
		{ 151, 212, 24 },
		{ 151, 213, 24 },
		{ 151, 214, 24 },
		{ 151, 215, 24 },
		{ 151, 216, 24 },
		{ 151, 217, 24 },
		{ 151, 218, 24 },
		{ 151, 219, 24 },
		{ 151, 220, 24 },
		{ 151, 221, 24 },
		{ 151, 222, 24 },
		{ 151, 223, 24 },
		{ 151, 224, 24 },
		{ 151, 225, 24 },
		{ 151, 226, 24 },
		{ 151, 227, 24 },
		{ 151, 228, 24 },
		{ 151, 229, 24 },
		{ 151, 230, 24 },
		{ 152, 179, 24 },
		{ 152, 180, 24 },
		{ 152, 181, 24 },
		{ 152, 182, 24 },
		{ 152, 183, 24 },
		{ 152, 184, 24 },
		{ 152, 185, 24 },
		{ 152, 186, 24 },
		{ 152, 187, 24 },
		{ 152, 188, 24 },
		{ 152, 189, 24 },
		{ 152, 190, 24 },
		{ 152, 191, 24 },
		{ 152, 192, 24 },
		{ 152, 193, 24 },
		{ 152, 194, 24 },
		{ 152, 195, 24 },
		{ 152, 196, 24 },
		{ 152, 197, 24 },
		{ 152, 198, 24 },
		{ 152, 199, 24 },
		{ 152, 200, 24 },
		{ 152, 201, 24 },
		{ 152, 202, 24 },
		{ 152, 203, 24 },
		{ 152, 204, 24 },
		{ 152, 205, 24 },
		{ 152, 206, 24 },
		{ 152, 207, 24 },
		{ 152, 208, 24 },
		{ 152, 209, 24 },
		{ 152, 210, 24 },
		{ 152, 211, 24 },
		{ 152, 212, 24 },
		{ 152, 213, 24 },
		{ 152, 214, 24 },
		{ 152, 215, 24 },
		{ 152, 216, 24 },
		{ 152, 217, 24 },
		{ 152, 218, 24 },
		{ 152, 219, 24 },
		{ 152, 220, 24 },
		{ 152, 221, 24 },
		{ 152, 222, 24 },
		{ 152, 223, 24 },
		{ 152, 224, 24 },
		{ 152, 225, 24 },
		{ 152, 226, 24 },
		{ 152, 227, 24 },
		{ 152, 228, 24 },
		{ 152, 229, 24 },
		{ 152, 230, 24 },
		{ 153, 180, 24 },
		{ 153, 181, 24 },
		{ 153, 182, 24 },
		{ 153, 183, 24 },
		{ 153, 184, 24 },
		{ 153, 185, 24 },
		{ 153, 186, 24 },
		{ 153, 187, 24 },
		{ 153, 188, 24 },
		{ 153, 189, 24 },
		{ 153, 190, 24 },
		{ 153, 191, 24 },
		{ 153, 192, 24 },
		{ 153, 193, 24 },
		{ 153, 194, 24 },
		{ 153, 195, 24 },
		{ 153, 196, 24 },
		{ 153, 197, 24 },
		{ 153, 198, 24 },
		{ 153, 199, 24 },
		{ 153, 200, 24 },
		{ 153, 201, 24 },
		{ 153, 202, 24 },
		{ 153, 203, 24 },
		{ 153, 204, 24 },
		{ 153, 205, 24 },
		{ 153, 206, 24 },
		{ 153, 207, 24 },
		{ 153, 208, 24 },
		{ 153, 209, 24 },
		{ 153, 210, 24 },
		{ 153, 211, 24 },
		{ 153, 212, 24 },
		{ 153, 213, 24 },
		{ 153, 214, 24 },
		{ 153, 215, 24 },
		{ 153, 216, 24 },
		{ 153, 217, 24 },
		{ 153, 218, 24 },
		{ 153, 219, 24 },
		{ 153, 220, 24 },
		{ 153, 221, 24 },
		{ 153, 222, 24 },
		{ 153, 223, 24 },
		{ 153, 224, 24 },
		{ 153, 225, 24 },
		{ 153, 226, 24 },
		{ 153, 227, 24 },
		{ 153, 228, 24 },
		{ 153, 229, 24 },
		{ 153, 230, 24 },
		{ 154, 180, 24 },
		{ 154, 181, 24 },
		{ 154, 182, 24 },
		{ 154, 183, 24 },
		{ 154, 184, 24 },
		{ 154, 185, 24 },
		{ 154, 186, 24 },
		{ 154, 187, 24 },
		{ 154, 188, 24 },
		{ 154, 189, 24 },
		{ 154, 190, 24 },
		{ 154, 191, 24 },
		{ 154, 192, 24 },
		{ 154, 193, 24 },
		{ 154, 194, 24 },
		{ 154, 195, 24 },
		{ 154, 196, 24 },
		{ 154, 197, 24 },
		{ 154, 198, 24 },
		{ 154, 199, 24 },
		{ 154, 200, 24 },
		{ 154, 201, 24 },
		{ 154, 202, 24 },
		{ 154, 203, 24 },
		{ 154, 204, 24 },
		{ 154, 205, 24 },
		{ 154, 206, 24 },
		{ 154, 207, 24 },
		{ 154, 208, 24 },
		{ 154, 209, 24 },
		{ 154, 210, 24 },
		{ 154, 211, 24 },
		{ 154, 212, 24 },
		{ 154, 213, 24 },
		{ 154, 214, 24 },
		{ 154, 215, 24 },
		{ 154, 216, 24 },
		{ 154, 217, 24 },
		{ 154, 218, 24 },
		{ 154, 219, 24 },
		{ 154, 220, 24 },
		{ 154, 221, 24 },
		{ 154, 222, 24 },
		{ 154, 223, 24 },
		{ 154, 224, 24 },
		{ 154, 225, 24 },
		{ 154, 226, 24 },
		{ 154, 227, 24 },
		{ 154, 228, 24 },
		{ 154, 229, 24 },
		{ 154, 230, 24 },
		{ 154, 231, 24 },
		{ 155, 181, 24 },
		{ 155, 182, 24 },
		{ 155, 183, 24 },
		{ 155, 184, 24 },
		{ 155, 185, 24 },
		{ 155, 186, 24 },
		{ 155, 187, 24 },
		{ 155, 188, 24 },
		{ 155, 189, 24 },
		{ 155, 190, 24 },
		{ 155, 191, 24 },
		{ 155, 192, 24 },
		{ 155, 193, 24 },
		{ 155, 194, 24 },
		{ 155, 195, 24 },
		{ 155, 196, 24 },
		{ 155, 197, 24 },
		{ 155, 198, 24 },
		{ 155, 199, 24 },
		{ 155, 200, 24 },
		{ 155, 201, 24 },
		{ 155, 202, 24 },
		{ 155, 203, 24 },
		{ 155, 204, 24 },
		{ 155, 205, 24 },
		{ 155, 206, 24 },
		{ 155, 207, 24 },
		{ 155, 208, 24 },
		{ 155, 209, 24 },
		{ 155, 210, 24 },
		{ 155, 211, 24 },
		{ 155, 212, 24 },
		{ 155, 213, 24 },
		{ 155, 214, 24 },
		{ 155, 215, 24 },
		{ 155, 216, 24 },
		{ 155, 217, 24 },
		{ 155, 218, 24 },
		{ 155, 219, 24 },
		{ 155, 220, 24 },
		{ 155, 221, 24 },
		{ 155, 222, 24 },
		{ 155, 223, 24 },
		{ 155, 224, 24 },
		{ 155, 225, 24 },
		{ 155, 226, 24 },
		{ 155, 227, 24 },
		{ 155, 228, 24 },
		{ 155, 229, 24 },
		{ 155, 230, 24 },
		{ 155, 231, 24 },
		{ 156, 182, 24 },
		{ 156, 183, 24 },
		{ 156, 184, 24 },
		{ 156, 185, 24 },
		{ 156, 186, 24 },
		{ 156, 187, 24 },
		{ 156, 188, 24 },
		{ 156, 189, 24 },
		{ 156, 190, 24 },
		{ 156, 191, 24 },
		{ 156, 192, 24 },
		{ 156, 193, 24 },
		{ 156, 194, 24 },
		{ 156, 195, 24 },
		{ 156, 196, 24 },
		{ 156, 197, 24 },
		{ 156, 198, 24 },
		{ 156, 199, 24 },
		{ 156, 200, 24 },
		{ 156, 201, 24 },
		{ 156, 202, 24 },
		{ 156, 203, 24 },
		{ 156, 204, 24 },
		{ 156, 205, 24 },
		{ 156, 206, 24 },
		{ 156, 207, 24 },
		{ 156, 208, 24 },
		{ 156, 209, 24 },
		{ 156, 210, 24 },
		{ 156, 211, 24 },
		{ 156, 212, 24 },
		{ 156, 213, 24 },
		{ 156, 214, 24 },
		{ 156, 215, 24 },
		{ 156, 216, 24 },
		{ 156, 217, 24 },
		{ 156, 218, 24 },
		{ 156, 219, 24 },
		{ 156, 220, 24 },
		{ 156, 221, 24 },
		{ 156, 222, 24 },
		{ 156, 223, 24 },
		{ 156, 224, 24 },
		{ 156, 225, 24 },
		{ 156, 226, 24 },
		{ 156, 227, 24 },
		{ 156, 228, 24 },
		{ 156, 229, 24 },
		{ 156, 230, 24 },
		{ 156, 231, 24 },
		{ 157, 183, 24 },
		{ 157, 184, 24 },
		{ 157, 185, 24 },
		{ 157, 186, 24 },
		{ 157, 187, 24 },
		{ 157, 188, 24 },
		{ 157, 189, 24 },
		{ 157, 190, 24 },
		{ 157, 191, 24 },
		{ 157, 192, 24 },
		{ 157, 193, 24 },
		{ 157, 194, 24 },
		{ 157, 195, 24 },
		{ 157, 196, 24 },
		{ 157, 197, 24 },
		{ 157, 198, 24 },
		{ 157, 199, 24 },
		{ 157, 200, 24 },
		{ 157, 201, 24 },
		{ 157, 202, 24 },
		{ 157, 203, 24 },
		{ 157, 204, 24 },
		{ 157, 205, 24 },
		{ 157, 206, 24 },
		{ 157, 207, 24 },
		{ 157, 208, 24 },
		{ 157, 209, 24 },
		{ 157, 210, 24 },
		{ 157, 211, 24 },
		{ 157, 212, 24 },
		{ 157, 213, 24 },
		{ 157, 214, 24 },
		{ 157, 215, 24 },
		{ 157, 216, 24 },
		{ 157, 217, 24 },
		{ 157, 218, 24 },
		{ 157, 219, 24 },
		{ 157, 220, 24 },
		{ 157, 221, 24 },
		{ 157, 222, 24 },
		{ 157, 223, 24 },
		{ 157, 224, 24 },
		{ 157, 225, 24 },
		{ 157, 226, 24 },
		{ 157, 227, 24 },
		{ 157, 228, 24 },
		{ 157, 229, 24 },
		{ 157, 230, 24 },
		{ 157, 231, 24 },
		{ 158, 183, 24 },
		{ 158, 184, 24 },
		{ 158, 185, 24 },
		{ 158, 186, 24 },
		{ 158, 187, 24 },
		{ 158, 188, 24 },
		{ 158, 189, 24 },
		{ 158, 190, 24 },
		{ 158, 191, 24 },
		{ 158, 192, 24 },
		{ 158, 193, 24 },
		{ 158, 194, 24 },
		{ 158, 195, 24 },
		{ 158, 196, 24 },
		{ 158, 197, 24 },
		{ 158, 198, 24 },
		{ 158, 199, 24 },
		{ 158, 200, 24 },
		{ 158, 201, 24 },
		{ 158, 202, 24 },
		{ 158, 203, 24 },
		{ 158, 204, 24 },
		{ 158, 205, 24 },
		{ 158, 206, 24 },
		{ 158, 207, 24 },
		{ 158, 208, 24 },
		{ 158, 209, 24 },
		{ 158, 210, 24 },
		{ 158, 211, 24 },
		{ 158, 212, 24 },
		{ 158, 213, 24 },
		{ 158, 214, 24 },
		{ 158, 215, 24 },
		{ 158, 216, 24 },
		{ 158, 217, 24 },
		{ 158, 218, 24 },
		{ 158, 219, 24 },
		{ 158, 220, 24 },
		{ 158, 221, 24 },
		{ 158, 222, 24 },
		{ 158, 223, 24 },
		{ 158, 224, 24 },
		{ 158, 225, 24 },
		{ 158, 226, 24 },
		{ 158, 227, 24 },
		{ 158, 228, 24 },
		{ 158, 229, 24 },
		{ 158, 230, 24 },
		{ 158, 231, 24 },
		{ 158, 232, 24 },
		{ 159, 184, 24 },
		{ 159, 185, 24 },
		{ 159, 186, 24 },
		{ 159, 187, 24 },
		{ 159, 188, 24 },
		{ 159, 189, 24 },
		{ 159, 190, 24 },
		{ 159, 191, 24 },
		{ 159, 192, 24 },
		{ 159, 193, 24 },
		{ 159, 194, 24 },
		{ 159, 195, 24 },
		{ 159, 196, 24 },
		{ 159, 197, 24 },
		{ 159, 198, 24 },
		{ 159, 199, 24 },
		{ 159, 200, 24 },
		{ 159, 201, 24 },
		{ 159, 202, 24 },
		{ 159, 203, 24 },
		{ 159, 204, 24 },
		{ 159, 205, 24 },
		{ 159, 206, 24 },
		{ 159, 207, 24 },
		{ 159, 208, 24 },
		{ 159, 209, 24 },
		{ 159, 210, 24 },
		{ 159, 211, 24 },
		{ 159, 212, 24 },
		{ 159, 213, 24 },
		{ 159, 214, 24 },
		{ 159, 215, 24 },
		{ 159, 216, 24 },
		{ 159, 217, 24 },
		{ 159, 218, 24 },
		{ 159, 219, 24 },
		{ 159, 220, 24 },
		{ 159, 221, 24 },
		{ 159, 222, 24 },
		{ 159, 223, 24 },
		{ 159, 224, 24 },
		{ 159, 225, 24 },
		{ 159, 226, 24 },
		{ 159, 227, 24 },
		{ 159, 228, 24 },
		{ 159, 229, 24 },
		{ 159, 230, 24 },
		{ 159, 231, 24 },
		{ 159, 232, 24 },
		{ 160, 185, 24 },
		{ 160, 186, 24 },
		{ 160, 187, 24 },
		{ 160, 188, 24 },
		{ 160, 189, 24 },
		{ 160, 190, 24 },
		{ 160, 191, 24 },
		{ 160, 192, 24 },
		{ 160, 193, 24 },
		{ 160, 194, 24 },
		{ 160, 195, 24 },
		{ 160, 196, 24 },
		{ 160, 197, 24 },
		{ 160, 198, 24 },
		{ 160, 199, 24 },
		{ 160, 200, 24 },
		{ 160, 201, 24 },
		{ 160, 202, 24 },
		{ 160, 203, 24 },
		{ 160, 204, 24 },
		{ 160, 205, 24 },
		{ 160, 206, 24 },
		{ 160, 207, 24 },
		{ 160, 208, 24 },
		{ 160, 209, 24 },
		{ 160, 210, 24 },
		{ 160, 211, 24 },
		{ 160, 212, 24 },
		{ 160, 213, 24 },
		{ 160, 214, 24 },
		{ 160, 215, 24 },
		{ 160, 216, 24 },
		{ 160, 217, 24 },
		{ 160, 218, 24 },
		{ 160, 219, 24 },
		{ 160, 220, 24 },
		{ 160, 221, 24 },
		{ 160, 222, 24 },
		{ 160, 223, 24 },
		{ 160, 224, 24 },
		{ 160, 225, 24 },
		{ 160, 226, 24 },
		{ 160, 227, 24 },
		{ 160, 228, 24 },
		{ 160, 229, 24 },
		{ 160, 230, 24 },
		{ 160, 231, 24 },
		{ 160, 232, 24 },
		{ 161, 186, 24 },
		{ 161, 187, 24 },
		{ 161, 188, 24 },
		{ 161, 189, 24 },
		{ 161, 190, 24 },
		{ 161, 191, 24 },
		{ 161, 192, 24 },
		{ 161, 193, 24 },
		{ 161, 194, 24 },
		{ 161, 195, 24 },
		{ 161, 196, 24 },
		{ 161, 197, 24 },
		{ 161, 198, 24 },
		{ 161, 199, 24 },
		{ 161, 200, 24 },
		{ 161, 201, 24 },
		{ 161, 202, 24 },
		{ 161, 203, 24 },
		{ 161, 204, 24 },
		{ 161, 205, 24 },
		{ 161, 206, 24 },
		{ 161, 207, 24 },
		{ 161, 208, 24 },
		{ 161, 209, 24 },
		{ 161, 210, 24 },
		{ 161, 211, 24 },
		{ 161, 212, 24 },
		{ 161, 213, 24 },
		{ 161, 214, 24 },
		{ 161, 215, 24 },
		{ 161, 216, 24 },
		{ 161, 217, 24 },
		{ 161, 218, 24 },
		{ 161, 219, 24 },
		{ 161, 220, 24 },
		{ 161, 221, 24 },
		{ 161, 222, 24 },
		{ 161, 223, 24 },
		{ 161, 224, 24 },
		{ 161, 225, 24 },
		{ 161, 226, 24 },
		{ 161, 227, 24 },
		{ 161, 228, 24 },
		{ 161, 229, 24 },
		{ 161, 230, 24 },
		{ 161, 231, 24 },
		{ 161, 232, 24 },
		{ 162, 186, 24 },
		{ 162, 187, 24 },
		{ 162, 188, 24 },
		{ 162, 189, 24 },
		{ 162, 190, 24 },
		{ 162, 191, 24 },
		{ 162, 192, 24 },
		{ 162, 193, 24 },
		{ 162, 194, 24 },
		{ 162, 195, 24 },
		{ 162, 196, 24 },
		{ 162, 197, 24 },
		{ 162, 198, 24 },
		{ 162, 199, 24 },
		{ 162, 200, 24 },
		{ 162, 201, 24 },
		{ 162, 202, 24 },
		{ 162, 203, 24 },
		{ 162, 204, 24 },
		{ 162, 205, 24 },
		{ 162, 206, 24 },
		{ 162, 207, 24 },
		{ 162, 208, 24 },
		{ 162, 209, 24 },
		{ 162, 210, 24 },
		{ 162, 211, 24 },
		{ 162, 212, 24 },
		{ 162, 213, 24 },
		{ 162, 214, 24 },
		{ 162, 215, 24 },
		{ 162, 216, 24 },
		{ 162, 217, 24 },
		{ 162, 218, 24 },
		{ 162, 219, 24 },
		{ 162, 220, 24 },
		{ 162, 221, 24 },
		{ 162, 222, 24 },
		{ 162, 223, 24 },
		{ 162, 224, 24 },
		{ 162, 225, 24 },
		{ 162, 226, 24 },
		{ 162, 227, 24 },
		{ 162, 228, 24 },
		{ 162, 229, 24 },
		{ 162, 230, 24 },
		{ 162, 231, 24 },
		{ 162, 232, 24 },
		{ 162, 233, 24 },
		{ 163, 187, 24 },
		{ 163, 188, 24 },
		{ 163, 189, 24 },
		{ 163, 190, 24 },
		{ 163, 191, 24 },
		{ 163, 192, 24 },
		{ 163, 193, 24 },
		{ 163, 194, 24 },
		{ 163, 195, 24 },
		{ 163, 196, 24 },
		{ 163, 197, 24 },
		{ 163, 198, 24 },
		{ 163, 199, 24 },
		{ 163, 200, 24 },
		{ 163, 201, 24 },
		{ 163, 202, 24 },
		{ 163, 203, 24 },
		{ 163, 204, 24 },
		{ 163, 205, 24 },
		{ 163, 206, 24 },
		{ 163, 207, 24 },
		{ 163, 208, 24 },
		{ 163, 209, 24 },
		{ 163, 210, 24 },
		{ 163, 211, 24 },
		{ 163, 212, 24 },
		{ 163, 213, 24 },
		{ 163, 214, 24 },
		{ 163, 215, 24 },
		{ 163, 216, 24 },
		{ 163, 217, 24 },
		{ 163, 218, 24 },
		{ 163, 219, 24 },
		{ 163, 220, 24 },
		{ 163, 221, 24 },
		{ 163, 222, 24 },
		{ 163, 223, 24 },
		{ 163, 224, 24 },
		{ 163, 225, 24 },
		{ 163, 226, 24 },
		{ 163, 227, 24 },
		{ 163, 228, 24 },
		{ 163, 229, 24 },
		{ 163, 230, 24 },
		{ 163, 231, 24 },
		{ 163, 232, 24 },
		{ 163, 233, 24 },
		{ 164, 188, 24 },
		{ 164, 189, 24 },
		{ 164, 190, 24 },
		{ 164, 191, 24 },
		{ 164, 192, 24 },
		{ 164, 193, 24 },
		{ 164, 194, 24 },
		{ 164, 195, 24 },
		{ 164, 196, 24 },
		{ 164, 197, 24 },
		{ 164, 198, 24 },
		{ 164, 199, 24 },
		{ 164, 200, 24 },
		{ 164, 201, 24 },
		{ 164, 202, 24 },
		{ 164, 203, 24 },
		{ 164, 204, 24 },
		{ 164, 205, 24 },
		{ 164, 206, 24 },
		{ 164, 207, 24 },
		{ 164, 208, 24 },
		{ 164, 209, 24 },
		{ 164, 210, 24 },
		{ 164, 211, 24 },
		{ 164, 212, 24 },
		{ 164, 213, 24 },
		{ 164, 214, 24 },
		{ 164, 215, 24 },
		{ 164, 216, 24 },
		{ 164, 217, 24 },
		{ 164, 218, 24 },
		{ 164, 219, 24 },
		{ 164, 220, 24 },
		{ 164, 221, 24 },
		{ 164, 222, 24 },
		{ 164, 223, 24 },
		{ 164, 224, 24 },
		{ 164, 225, 24 },
		{ 164, 226, 24 },
		{ 164, 227, 24 },
		{ 164, 228, 24 },
		{ 164, 229, 24 },
		{ 164, 230, 24 },
		{ 164, 231, 24 },
		{ 164, 232, 24 },
		{ 164, 233, 24 },
		{ 165, 189, 24 },
		{ 165, 190, 24 },
		{ 165, 191, 24 },
		{ 165, 192, 24 },
		{ 165, 193, 24 },
		{ 165, 194, 24 },
		{ 165, 195, 24 },
		{ 165, 196, 24 },
		{ 165, 197, 24 },
		{ 165, 198, 24 },
		{ 165, 199, 24 },
		{ 165, 200, 24 },
		{ 165, 201, 24 },
		{ 165, 202, 24 },
		{ 165, 203, 24 },
		{ 165, 204, 24 },
		{ 165, 205, 24 },
		{ 165, 206, 24 },
		{ 165, 207, 24 },
		{ 165, 208, 24 },
		{ 165, 209, 24 },
		{ 165, 210, 24 },
		{ 165, 211, 24 },
		{ 165, 212, 24 },
		{ 165, 213, 24 },
		{ 165, 214, 24 },
		{ 165, 215, 24 },
		{ 165, 216, 24 },
		{ 165, 217, 24 },
		{ 165, 218, 24 },
		{ 165, 219, 24 },
		{ 165, 220, 24 },
		{ 165, 221, 24 },
		{ 165, 222, 24 },
		{ 165, 223, 24 },
		{ 165, 224, 24 },
		{ 165, 225, 24 },
		{ 165, 226, 24 },
		{ 165, 227, 24 },
		{ 165, 228, 24 },
		{ 165, 229, 24 },
		{ 165, 230, 24 },
		{ 165, 231, 24 },
		{ 165, 232, 24 },
		{ 165, 233, 24 },
		{ 166, 189, 24 },
		{ 166, 190, 24 },
		{ 166, 191, 24 },
		{ 166, 192, 24 },
		{ 166, 193, 24 },
		{ 166, 194, 24 },
		{ 166, 195, 24 },
		{ 166, 196, 24 },
		{ 166, 197, 24 },
		{ 166, 198, 24 },
		{ 166, 199, 24 },
		{ 166, 200, 24 },
		{ 166, 201, 24 },
		{ 166, 202, 24 },
		{ 166, 203, 24 },
		{ 166, 204, 24 },
		{ 166, 205, 24 },
		{ 166, 206, 24 },
		{ 166, 207, 24 },
		{ 166, 208, 24 },
		{ 166, 209, 24 },
		{ 166, 210, 24 },
		{ 166, 211, 24 },
		{ 166, 212, 24 },
		{ 166, 213, 24 },
		{ 166, 214, 24 },
		{ 166, 215, 24 },
		{ 166, 216, 24 },
		{ 166, 217, 24 },
		{ 166, 218, 24 },
		{ 166, 219, 24 },
		{ 166, 220, 24 },
		{ 166, 221, 24 },
		{ 166, 222, 24 },
		{ 166, 223, 24 },
		{ 166, 224, 24 },
		{ 166, 225, 24 },
		{ 166, 226, 24 },
		{ 166, 227, 24 },
		{ 166, 228, 24 },
		{ 166, 229, 24 },
		{ 166, 230, 24 },
		{ 166, 231, 24 },
		{ 166, 232, 24 },
		{ 166, 233, 24 },
		{ 166, 234, 24 },
		{ 167, 190, 24 },
		{ 167, 191, 24 },
		{ 167, 192, 24 },
		{ 167, 193, 24 },
		{ 167, 194, 24 },
		{ 167, 195, 24 },
		{ 167, 196, 24 },
		{ 167, 197, 24 },
		{ 167, 198, 24 },
		{ 167, 199, 24 },
		{ 167, 200, 24 },
		{ 167, 201, 24 },
		{ 167, 202, 24 },
		{ 167, 203, 24 },
		{ 167, 204, 24 },
		{ 167, 205, 24 },
		{ 167, 206, 24 },
		{ 167, 207, 24 },
		{ 167, 208, 24 },
		{ 167, 209, 24 },
		{ 167, 210, 24 },
		{ 167, 211, 24 },
		{ 167, 212, 24 },
		{ 167, 213, 24 },
		{ 167, 214, 24 },
		{ 167, 215, 24 },
		{ 167, 216, 24 },
		{ 167, 217, 24 },
		{ 167, 218, 24 },
		{ 167, 219, 24 },
		{ 167, 220, 24 },
		{ 167, 221, 24 },
		{ 167, 222, 24 },
		{ 167, 223, 24 },
		{ 167, 224, 24 },
		{ 167, 225, 24 },
		{ 167, 226, 24 },
		{ 167, 227, 24 },
		{ 167, 228, 24 },
		{ 167, 229, 24 },
		{ 167, 230, 24 },
		{ 167, 231, 24 },
		{ 167, 232, 24 },
		{ 167, 233, 24 },
		{ 167, 234, 24 },
		{ 168, 191, 24 },
		{ 168, 192, 24 },
		{ 168, 193, 24 },
		{ 168, 194, 24 },
		{ 168, 195, 24 },
		{ 168, 196, 24 },
		{ 168, 197, 24 },
		{ 168, 198, 24 },
		{ 168, 199, 24 },
		{ 168, 200, 24 },
		{ 168, 201, 24 },
		{ 168, 202, 24 },
		{ 168, 203, 24 },
		{ 168, 204, 24 },
		{ 168, 205, 24 },
		{ 168, 206, 24 },
		{ 168, 207, 24 },
		{ 168, 208, 24 },
		{ 168, 209, 24 },
		{ 168, 210, 24 },
		{ 168, 211, 24 },
		{ 168, 212, 24 },
		{ 168, 213, 24 },
		{ 168, 214, 24 },
		{ 168, 215, 24 },
		{ 168, 216, 24 },
		{ 168, 217, 24 },
		{ 168, 218, 24 },
		{ 168, 219, 24 },
		{ 168, 220, 24 },
		{ 168, 221, 24 },
		{ 168, 222, 24 },
		{ 168, 223, 24 },
		{ 168, 224, 24 },
		{ 168, 225, 24 },
		{ 168, 226, 24 },
		{ 168, 227, 24 },
		{ 168, 228, 24 },
		{ 168, 229, 24 },
		{ 168, 230, 24 },
		{ 168, 231, 24 },
		{ 168, 232, 24 },
		{ 168, 233, 24 },
		{ 168, 234, 24 },
		{ 169, 192, 24 },
		{ 169, 193, 24 },
		{ 169, 194, 24 },
		{ 169, 195, 24 },
		{ 169, 196, 24 },
		{ 169, 197, 24 },
		{ 169, 198, 24 },
		{ 169, 199, 24 },
		{ 169, 200, 24 },
		{ 169, 201, 24 },
		{ 169, 202, 24 },
		{ 169, 203, 24 },
		{ 169, 204, 24 },
		{ 169, 205, 24 },
		{ 169, 206, 24 },
		{ 169, 207, 24 },
		{ 169, 208, 24 },
		{ 169, 209, 24 },
		{ 169, 210, 24 },
		{ 169, 211, 24 },
		{ 169, 212, 24 },
		{ 169, 213, 24 },
		{ 169, 214, 24 },
		{ 169, 215, 24 },
		{ 169, 216, 24 },
		{ 169, 217, 24 },
		{ 169, 218, 24 },
		{ 169, 219, 24 },
		{ 169, 220, 24 },
		{ 169, 221, 24 },
		{ 169, 222, 24 },
		{ 169, 223, 24 },
		{ 169, 224, 24 },
		{ 169, 225, 24 },
		{ 169, 226, 24 },
		{ 169, 227, 24 },
		{ 169, 228, 24 },
		{ 169, 229, 24 },
		{ 169, 230, 24 },
		{ 169, 231, 24 },
		{ 169, 232, 24 },
		{ 169, 233, 24 },
		{ 169, 234, 24 },
		{ 170, 192, 24 },
		{ 170, 193, 24 },
		{ 170, 194, 24 },
		{ 170, 195, 24 },
		{ 170, 196, 24 },
		{ 170, 197, 24 },
		{ 170, 198, 24 },
		{ 170, 199, 24 },
		{ 170, 200, 24 },
		{ 170, 201, 24 },
		{ 170, 202, 24 },
		{ 170, 203, 24 },
		{ 170, 204, 24 },
		{ 170, 205, 24 },
		{ 170, 206, 24 },
		{ 170, 207, 24 },
		{ 170, 208, 24 },
		{ 170, 209, 24 },
		{ 170, 210, 24 },
		{ 170, 211, 24 },
		{ 170, 212, 24 },
		{ 170, 213, 24 },
		{ 170, 214, 24 },
		{ 170, 215, 24 },
		{ 170, 216, 24 },
		{ 170, 217, 24 },
		{ 170, 218, 24 },
		{ 170, 219, 24 },
		{ 170, 220, 24 },
		{ 170, 221, 24 },
		{ 170, 222, 24 },
		{ 170, 223, 24 },
		{ 170, 224, 24 },
		{ 170, 225, 24 },
		{ 170, 226, 24 },
		{ 170, 227, 24 },
		{ 170, 228, 24 },
		{ 170, 229, 24 },
		{ 170, 230, 24 },
		{ 170, 231, 24 },
		{ 170, 232, 24 },
		{ 170, 233, 24 },
		{ 170, 234, 24 },
		{ 170, 235, 24 },
		{ 171, 193, 24 },
		{ 171, 194, 24 },
		{ 171, 195, 24 },
		{ 171, 196, 24 },
		{ 171, 197, 24 },
		{ 171, 198, 24 },
		{ 171, 199, 24 },
		{ 171, 200, 24 },
		{ 171, 201, 24 },
		{ 171, 202, 24 },
		{ 171, 203, 24 },
		{ 171, 204, 24 },
		{ 171, 205, 24 },
		{ 171, 206, 24 },
		{ 171, 207, 24 },
		{ 171, 208, 24 },
		{ 171, 209, 24 },
		{ 171, 210, 24 },
		{ 171, 211, 24 },
		{ 171, 212, 24 },
		{ 171, 213, 24 },
		{ 171, 214, 24 },
		{ 171, 215, 24 },
		{ 171, 216, 24 },
		{ 171, 217, 24 },
		{ 171, 218, 24 },
		{ 171, 219, 24 },
		{ 171, 220, 24 },
		{ 171, 221, 24 },
		{ 171, 222, 24 },
		{ 171, 223, 24 },
		{ 171, 224, 24 },
		{ 171, 225, 24 },
		{ 171, 226, 24 },
		{ 171, 227, 24 },
		{ 171, 228, 24 },
		{ 171, 229, 24 },
		{ 171, 230, 24 },
		{ 171, 231, 24 },
		{ 171, 232, 24 },
		{ 171, 233, 24 },
		{ 171, 234, 24 },
		{ 171, 235, 24 },
		{ 172, 194, 24 },
		{ 172, 195, 24 },
		{ 172, 196, 24 },
		{ 172, 197, 24 },
		{ 172, 198, 24 },
		{ 172, 199, 24 },
		{ 172, 200, 24 },
		{ 172, 201, 24 },
		{ 172, 202, 24 },
		{ 172, 203, 24 },
		{ 172, 204, 24 },
		{ 172, 205, 24 },
		{ 172, 206, 24 },
		{ 172, 207, 24 },
		{ 172, 208, 24 },
		{ 172, 209, 24 },
		{ 172, 210, 24 },
		{ 172, 211, 24 },
		{ 172, 212, 24 },
		{ 172, 213, 24 },
		{ 172, 214, 24 },
		{ 172, 215, 24 },
		{ 172, 216, 24 },
		{ 172, 217, 24 },
		{ 172, 218, 24 },
		{ 172, 219, 24 },
		{ 172, 220, 24 },
		{ 172, 221, 24 },
		{ 172, 222, 24 },
		{ 172, 223, 24 },
		{ 172, 224, 24 },
		{ 172, 225, 24 },
		{ 172, 226, 24 },
		{ 172, 227, 24 },
		{ 172, 228, 24 },
		{ 172, 229, 24 },
		{ 172, 230, 24 },
		{ 172, 231, 24 },
		{ 172, 232, 24 },
		{ 172, 233, 24 },
		{ 172, 234, 24 },
		{ 172, 235, 24 },
		{ 173, 195, 24 },
		{ 173, 196, 24 },
		{ 173, 197, 24 },
		{ 173, 198, 24 },
		{ 173, 199, 24 },
		{ 173, 200, 24 },
		{ 173, 201, 24 },
		{ 173, 202, 24 },
		{ 173, 203, 24 },
		{ 173, 204, 24 },
		{ 173, 205, 24 },
		{ 173, 206, 24 },
		{ 173, 207, 24 },
		{ 173, 208, 24 },
		{ 173, 209, 24 },
		{ 173, 210, 24 },
		{ 173, 211, 24 },
		{ 173, 212, 24 },
		{ 173, 213, 24 },
		{ 173, 214, 24 },
		{ 173, 215, 24 },
		{ 173, 216, 24 },
		{ 173, 217, 24 },
		{ 173, 218, 24 },
		{ 173, 219, 24 },
		{ 173, 220, 24 },
		{ 173, 221, 24 },
		{ 173, 222, 24 },
		{ 173, 223, 24 },
		{ 173, 224, 24 },
		{ 173, 225, 24 },
		{ 173, 226, 24 },
		{ 173, 227, 24 },
		{ 173, 228, 24 },
		{ 173, 229, 24 },
		{ 173, 230, 24 },
		{ 173, 231, 24 },
		{ 173, 232, 24 },
		{ 173, 233, 24 },
		{ 173, 234, 24 },
		{ 173, 235, 24 },
		{ 174, 195, 24 },
		{ 174, 196, 24 },
		{ 174, 197, 24 },
		{ 174, 198, 24 },
		{ 174, 199, 24 },
		{ 174, 200, 24 },
		{ 174, 201, 24 },
		{ 174, 202, 24 },
		{ 174, 203, 24 },
		{ 174, 204, 24 },
		{ 174, 205, 24 },
		{ 174, 206, 24 },
		{ 174, 207, 24 },
		{ 174, 208, 24 },
		{ 174, 209, 24 },
		{ 174, 210, 24 },
		{ 174, 211, 24 },
		{ 174, 212, 24 },
		{ 174, 213, 24 },
		{ 174, 214, 24 },
		{ 174, 215, 24 },
		{ 174, 216, 24 },
		{ 174, 217, 24 },
		{ 174, 218, 24 },
		{ 174, 219, 24 },
		{ 174, 220, 24 },
		{ 174, 221, 24 },
		{ 174, 222, 24 },
		{ 174, 223, 24 },
		{ 174, 224, 24 },
		{ 174, 225, 24 },
		{ 174, 226, 24 },
		{ 174, 227, 24 },
		{ 174, 228, 24 },
		{ 174, 229, 24 },
		{ 174, 230, 24 },
		{ 174, 231, 24 },
		{ 174, 232, 24 },
		{ 174, 233, 24 },
		{ 174, 234, 24 },
		{ 174, 235, 24 },
		{ 174, 236, 24 },
		{ 175, 196, 24 },
		{ 175, 197, 24 },
		{ 175, 198, 24 },
		{ 175, 199, 24 },
		{ 175, 200, 24 },
		{ 175, 201, 24 },
		{ 175, 202, 24 },
		{ 175, 203, 24 },
		{ 175, 204, 24 },
		{ 175, 205, 24 },
		{ 175, 206, 24 },
		{ 175, 207, 24 },
		{ 175, 208, 24 },
		{ 175, 209, 24 },
		{ 175, 210, 24 },
		{ 175, 211, 24 },
		{ 175, 212, 24 },
		{ 175, 213, 24 },
		{ 175, 214, 24 },
		{ 175, 215, 24 },
		{ 175, 216, 24 },
		{ 175, 217, 24 },
		{ 175, 218, 24 },
		{ 175, 219, 24 },
		{ 175, 220, 24 },
		{ 175, 221, 24 },
		{ 175, 222, 24 },
		{ 175, 223, 24 },
		{ 175, 224, 24 },
		{ 175, 225, 24 },
		{ 175, 226, 24 },
		{ 175, 227, 24 },
		{ 175, 228, 24 },
		{ 175, 229, 24 },
		{ 175, 230, 24 },
		{ 175, 231, 24 },
		{ 175, 232, 24 },
		{ 175, 233, 24 },
		{ 175, 234, 24 },
		{ 175, 235, 24 },
		{ 175, 236, 24 },
		{ 176, 197, 24 },
		{ 176, 198, 24 },
		{ 176, 199, 24 },
		{ 176, 200, 24 },
		{ 176, 201, 24 },
		{ 176, 202, 24 },
		{ 176, 203, 24 },
		{ 176, 204, 24 },
		{ 176, 205, 24 },
		{ 176, 206, 24 },
		{ 176, 207, 24 },
		{ 176, 208, 24 },
		{ 176, 209, 24 },
		{ 176, 210, 24 },
		{ 176, 211, 24 },
		{ 176, 212, 24 },
		{ 176, 213, 24 },
		{ 176, 214, 24 },
		{ 176, 215, 24 },
		{ 176, 216, 24 },
		{ 176, 217, 24 },
		{ 176, 218, 24 },
		{ 176, 219, 24 },
		{ 176, 220, 24 },
		{ 176, 221, 24 },
		{ 176, 222, 24 },
		{ 176, 223, 24 },
		{ 176, 224, 24 },
		{ 176, 225, 24 },
		{ 176, 226, 24 },
		{ 176, 227, 24 },
		{ 176, 228, 24 },
		{ 176, 229, 24 },
		{ 176, 230, 24 },
		{ 176, 231, 24 },
		{ 176, 232, 24 },
		{ 176, 233, 24 },
		{ 176, 234, 24 },
		{ 176, 235, 24 },
		{ 176, 236, 24 },
		{ 177, 198, 24 },
		{ 177, 199, 24 },
		{ 177, 200, 24 },
		{ 177, 201, 24 },
		{ 177, 202, 24 },
		{ 177, 203, 24 },
		{ 177, 204, 24 },
		{ 177, 205, 24 },
		{ 177, 206, 24 },
		{ 177, 207, 24 },
		{ 177, 208, 24 },
		{ 177, 209, 24 },
		{ 177, 210, 24 },
		{ 177, 211, 24 },
		{ 177, 212, 24 },
		{ 177, 213, 24 },
		{ 177, 214, 24 },
		{ 177, 215, 24 },
		{ 177, 216, 24 },
		{ 177, 217, 24 },
		{ 177, 218, 24 },
		{ 177, 219, 24 },
		{ 177, 220, 24 },
		{ 177, 221, 24 },
		{ 177, 222, 24 },
		{ 177, 223, 24 },
		{ 177, 224, 24 },
		{ 177, 225, 24 },
		{ 177, 226, 24 },
		{ 177, 227, 24 },
		{ 177, 228, 24 },
		{ 177, 229, 24 },
		{ 177, 230, 24 },
		{ 177, 231, 24 },
		{ 177, 232, 24 },
		{ 177, 233, 24 },
		{ 177, 234, 24 },
		{ 177, 235, 24 },
		{ 177, 236, 24 },
		{ 178, 198, 24 },
		{ 178, 199, 24 },
		{ 178, 200, 24 },
		{ 178, 201, 24 },
		{ 178, 202, 24 },
		{ 178, 203, 24 },
		{ 178, 204, 24 },
		{ 178, 205, 24 },
		{ 178, 206, 24 },
		{ 178, 207, 24 },
		{ 178, 208, 24 },
		{ 178, 209, 24 },
		{ 178, 210, 24 },
		{ 178, 211, 24 },
		{ 178, 212, 24 },
		{ 178, 213, 24 },
		{ 178, 214, 24 },
		{ 178, 215, 24 },
		{ 178, 216, 24 },
		{ 178, 217, 24 },
		{ 178, 218, 24 },
		{ 178, 219, 24 },
		{ 178, 220, 24 },
		{ 178, 221, 24 },
		{ 178, 222, 24 },
		{ 178, 223, 24 },
		{ 178, 224, 24 },
		{ 178, 225, 24 },
		{ 178, 226, 24 },
		{ 178, 227, 24 },
		{ 178, 228, 24 },
		{ 178, 229, 24 },
		{ 178, 230, 24 },
		{ 178, 231, 24 },
		{ 178, 232, 24 },
		{ 178, 233, 24 },
		{ 178, 234, 24 },
		{ 178, 235, 24 },
		{ 178, 236, 24 },
		{ 178, 237, 24 },
		{ 179, 199, 24 },
		{ 179, 200, 24 },
		{ 179, 201, 24 },
		{ 179, 202, 24 },
		{ 179, 203, 24 },
		{ 179, 204, 24 },
		{ 179, 205, 24 },
		{ 179, 206, 24 },
		{ 179, 207, 24 },
		{ 179, 208, 24 },
		{ 179, 209, 24 },
		{ 179, 210, 24 },
		{ 179, 211, 24 },
		{ 179, 212, 24 },
		{ 179, 213, 24 },
		{ 179, 214, 24 },
		{ 179, 215, 24 },
		{ 179, 216, 24 },
		{ 179, 217, 24 },
		{ 179, 218, 24 },
		{ 179, 219, 24 },
		{ 179, 220, 24 },
		{ 179, 221, 24 },
		{ 179, 222, 24 },
		{ 179, 223, 24 },
		{ 179, 224, 24 },
		{ 179, 225, 24 },
		{ 179, 226, 24 },
		{ 179, 227, 24 },
		{ 179, 228, 24 },
		{ 179, 229, 24 },
		{ 179, 230, 24 },
		{ 179, 231, 24 },
		{ 179, 232, 24 },
		{ 179, 233, 24 },
		{ 179, 234, 24 },
		{ 179, 235, 24 },
		{ 179, 236, 24 },
		{ 179, 237, 24 },
		{ 180, 200, 24 },
		{ 180, 201, 24 },
		{ 180, 202, 24 },
		{ 180, 203, 24 },
		{ 180, 204, 24 },
		{ 180, 205, 24 },
		{ 180, 206, 24 },
		{ 180, 207, 24 },
		{ 180, 208, 24 },
		{ 180, 209, 24 },
		{ 180, 210, 24 },
		{ 180, 211, 24 },
		{ 180, 212, 24 },
		{ 180, 213, 24 },
		{ 180, 214, 24 },
		{ 180, 215, 24 },
		{ 180, 216, 24 },
		{ 180, 217, 24 },
		{ 180, 218, 24 },
		{ 180, 219, 24 },
		{ 180, 220, 24 },
		{ 180, 221, 24 },
		{ 180, 222, 24 },
		{ 180, 223, 24 },
		{ 180, 224, 24 },
		{ 180, 225, 24 },
		{ 180, 226, 24 },
		{ 180, 227, 24 },
		{ 180, 228, 24 },
		{ 180, 229, 24 },
		{ 180, 230, 24 },
		{ 180, 231, 24 },
		{ 180, 232, 24 },
		{ 180, 233, 24 },
		{ 180, 234, 24 },
		{ 180, 235, 24 },
		{ 180, 236, 24 },
		{ 180, 237, 24 },
		{ 181, 201, 24 },
		{ 181, 202, 24 },
		{ 181, 203, 24 },
		{ 181, 204, 24 },
		{ 181, 205, 24 },
		{ 181, 206, 24 },
		{ 181, 207, 24 },
		{ 181, 208, 24 },
		{ 181, 209, 24 },
		{ 181, 210, 24 },
		{ 181, 211, 24 },
		{ 181, 212, 24 },
		{ 181, 213, 24 },
		{ 181, 214, 24 },
		{ 181, 215, 24 },
		{ 181, 216, 24 },
		{ 181, 217, 24 },
		{ 181, 218, 24 },
		{ 181, 219, 24 },
		{ 181, 220, 24 },
		{ 181, 221, 24 },
		{ 181, 222, 24 },
		{ 181, 223, 24 },
		{ 181, 224, 24 },
		{ 181, 225, 24 },
		{ 181, 226, 24 },
		{ 181, 227, 24 },
		{ 181, 228, 24 },
		{ 181, 229, 24 },
		{ 181, 230, 24 },
		{ 181, 231, 24 },
		{ 181, 232, 24 },
		{ 181, 233, 24 },
		{ 181, 234, 24 },
		{ 181, 235, 24 },
		{ 181, 236, 24 },
		{ 181, 237, 24 },
		{ 182, 201, 24 },
		{ 182, 202, 24 },
		{ 182, 203, 24 },
		{ 182, 204, 24 },
		{ 182, 205, 24 },
		{ 182, 206, 24 },
		{ 182, 207, 24 },
		{ 182, 208, 24 },
		{ 182, 209, 24 },
		{ 182, 210, 24 },
		{ 182, 211, 24 },
		{ 182, 212, 24 },
		{ 182, 213, 24 },
		{ 182, 214, 24 },
		{ 182, 215, 24 },
		{ 182, 216, 24 },
		{ 182, 217, 24 },
		{ 182, 218, 24 },
		{ 182, 219, 24 },
		{ 182, 220, 24 },
		{ 182, 221, 24 },
		{ 182, 222, 24 },
		{ 182, 223, 24 },
		{ 182, 224, 24 },
		{ 182, 225, 24 },
		{ 182, 226, 24 },
		{ 182, 227, 24 },
		{ 182, 228, 24 },
		{ 182, 229, 24 },
		{ 182, 230, 24 },
		{ 182, 231, 24 },
		{ 182, 232, 24 },
		{ 182, 233, 24 },
		{ 182, 234, 24 },
		{ 182, 235, 24 },
		{ 182, 236, 24 },
		{ 182, 237, 24 },
		{ 182, 238, 24 },
		{ 183, 202, 24 },
		{ 183, 203, 24 },
		{ 183, 204, 24 },
		{ 183, 205, 24 },
		{ 183, 206, 24 },
		{ 183, 207, 24 },
		{ 183, 208, 24 },
		{ 183, 209, 24 },
		{ 183, 210, 24 },
		{ 183, 211, 24 },
		{ 183, 212, 24 },
		{ 183, 213, 24 },
		{ 183, 214, 24 },
		{ 183, 215, 24 },
		{ 183, 216, 24 },
		{ 183, 217, 24 },
		{ 183, 218, 24 },
		{ 183, 219, 24 },
		{ 183, 220, 24 },
		{ 183, 221, 24 },
		{ 183, 222, 24 },
		{ 183, 223, 24 },
		{ 183, 224, 24 },
		{ 183, 225, 24 },
		{ 183, 226, 24 },
		{ 183, 227, 24 },
		{ 183, 228, 24 },
		{ 183, 229, 24 },
		{ 183, 230, 24 },
		{ 183, 231, 24 },
		{ 183, 232, 24 },
		{ 183, 233, 24 },
		{ 183, 234, 24 },
		{ 183, 235, 24 },
		{ 183, 236, 24 },
		{ 183, 237, 24 },
		{ 183, 238, 24 },
		{ 184, 203, 24 },
		{ 184, 204, 24 },
		{ 184, 205, 24 },
		{ 184, 206, 24 },
		{ 184, 207, 24 },
		{ 184, 208, 24 },
		{ 184, 209, 24 },
		{ 184, 210, 24 },
		{ 184, 211, 24 },
		{ 184, 212, 24 },
		{ 184, 213, 24 },
		{ 184, 214, 24 },
		{ 184, 215, 24 },
		{ 184, 216, 24 },
		{ 184, 217, 24 },
		{ 184, 218, 24 },
		{ 184, 219, 24 },
		{ 184, 220, 24 },
		{ 184, 221, 24 },
		{ 184, 222, 24 },
		{ 184, 223, 24 },
		{ 184, 224, 24 },
		{ 184, 225, 24 },
		{ 184, 226, 24 },
		{ 184, 227, 24 },
		{ 184, 228, 24 },
		{ 184, 229, 24 },
		{ 184, 230, 24 },
		{ 184, 231, 24 },
		{ 184, 232, 24 },
		{ 184, 233, 24 },
		{ 184, 234, 24 },
		{ 184, 235, 24 },
		{ 184, 236, 24 },
		{ 184, 237, 24 },
		{ 184, 238, 24 },
		{ 185, 204, 24 },
		{ 185, 205, 24 },
		{ 185, 206, 24 },
		{ 185, 207, 24 },
		{ 185, 208, 24 },
		{ 185, 209, 24 },
		{ 185, 210, 24 },
		{ 185, 211, 24 },
		{ 185, 212, 24 },
		{ 185, 213, 24 },
		{ 185, 214, 24 },
		{ 185, 215, 24 },
		{ 185, 216, 24 },
		{ 185, 217, 24 },
		{ 185, 218, 24 },
		{ 185, 219, 24 },
		{ 185, 220, 24 },
		{ 185, 221, 24 },
		{ 185, 222, 24 },
		{ 185, 223, 24 },
		{ 185, 224, 24 },
		{ 185, 225, 24 },
		{ 185, 226, 24 },
		{ 185, 227, 24 },
		{ 185, 228, 24 },
		{ 185, 229, 24 },
		{ 185, 230, 24 },
		{ 185, 231, 24 },
		{ 185, 232, 24 },
		{ 185, 233, 24 },
		{ 185, 234, 24 },
		{ 185, 235, 24 },
		{ 185, 236, 24 },
		{ 185, 237, 24 },
		{ 185, 238, 24 },
		{ 186, 204, 24 },
		{ 186, 205, 24 },
		{ 186, 206, 24 },
		{ 186, 207, 24 },
		{ 186, 208, 24 },
		{ 186, 209, 24 },
		{ 186, 210, 24 },
		{ 186, 211, 24 },
		{ 186, 212, 24 },
		{ 186, 213, 24 },
		{ 186, 214, 24 },
		{ 186, 215, 24 },
		{ 186, 216, 24 },
		{ 186, 217, 24 },
		{ 186, 218, 24 },
		{ 186, 219, 24 },
		{ 186, 220, 24 },
		{ 186, 221, 24 },
		{ 186, 222, 24 },
		{ 186, 223, 24 },
		{ 186, 224, 24 },
		{ 186, 225, 24 },
		{ 186, 226, 24 },
		{ 186, 227, 24 },
		{ 186, 228, 24 },
		{ 186, 229, 24 },
		{ 186, 230, 24 },
		{ 186, 231, 24 },
		{ 186, 232, 24 },
		{ 186, 233, 24 },
		{ 186, 234, 24 },
		{ 186, 235, 24 },
		{ 186, 236, 24 },
		{ 186, 237, 24 },
		{ 186, 238, 24 },
		{ 186, 239, 24 },
		{ 187, 205, 24 },
		{ 187, 206, 24 },
		{ 187, 207, 24 },
		{ 187, 208, 24 },
		{ 187, 209, 24 },
		{ 187, 210, 24 },
		{ 187, 211, 24 },
		{ 187, 212, 24 },
		{ 187, 213, 24 },
		{ 187, 214, 24 },
		{ 187, 215, 24 },
		{ 187, 216, 24 },
		{ 187, 217, 24 },
		{ 187, 218, 24 },
		{ 187, 219, 24 },
		{ 187, 220, 24 },
		{ 187, 221, 24 },
		{ 187, 222, 24 },
		{ 187, 223, 24 },
		{ 187, 224, 24 },
		{ 187, 225, 24 },
		{ 187, 226, 24 },
		{ 187, 227, 24 },
		{ 187, 228, 24 },
		{ 187, 229, 24 },
		{ 187, 230, 24 },
		{ 187, 231, 24 },
		{ 187, 232, 24 },
		{ 187, 233, 24 },
		{ 187, 234, 24 },
		{ 187, 235, 24 },
		{ 187, 236, 24 },
		{ 187, 237, 24 },
		{ 187, 238, 24 },
		{ 187, 239, 24 },
		{ 188, 206, 24 },
		{ 188, 207, 24 },
		{ 188, 208, 24 },
		{ 188, 209, 24 },
		{ 188, 210, 24 },
		{ 188, 211, 24 },
		{ 188, 212, 24 },
		{ 188, 213, 24 },
		{ 188, 214, 24 },
		{ 188, 215, 24 },
		{ 188, 216, 24 },
		{ 188, 217, 24 },
		{ 188, 218, 24 },
		{ 188, 219, 24 },
		{ 188, 220, 24 },
		{ 188, 221, 24 },
		{ 188, 222, 24 },
		{ 188, 223, 24 },
		{ 188, 224, 24 },
		{ 188, 225, 24 },
		{ 188, 226, 24 },
		{ 188, 227, 24 },
		{ 188, 228, 24 },
		{ 188, 229, 24 },
		{ 188, 230, 24 },
		{ 188, 231, 24 },
		{ 188, 232, 24 },
		{ 188, 233, 24 },
		{ 188, 234, 24 },
		{ 188, 235, 24 },
		{ 188, 236, 24 },
		{ 188, 237, 24 },
		{ 188, 238, 24 },
		{ 188, 239, 24 },
		{ 189, 207, 24 },
		{ 189, 208, 24 },
		{ 189, 209, 24 },
		{ 189, 210, 24 },
		{ 189, 211, 24 },
		{ 189, 212, 24 },
		{ 189, 213, 24 },
		{ 189, 214, 24 },
		{ 189, 215, 24 },
		{ 189, 216, 24 },
		{ 189, 217, 24 },
		{ 189, 218, 24 },
		{ 189, 219, 24 },
		{ 189, 220, 24 },
		{ 189, 221, 24 },
		{ 189, 222, 24 },
		{ 189, 223, 24 },
		{ 189, 224, 24 },
		{ 189, 225, 24 },
		{ 189, 226, 24 },
		{ 189, 227, 24 },
		{ 189, 228, 24 },
		{ 189, 229, 24 },
		{ 189, 230, 24 },
		{ 189, 231, 24 },
		{ 189, 232, 24 },
		{ 189, 233, 24 },
		{ 189, 234, 24 },
		{ 189, 235, 24 },
		{ 189, 236, 24 },
		{ 189, 237, 24 },
		{ 189, 238, 24 },
		{ 189, 239, 24 },
		{ 190, 207, 24 },
		{ 190, 208, 24 },
		{ 190, 209, 24 },
		{ 190, 210, 24 },
		{ 190, 211, 24 },
		{ 190, 212, 24 },
		{ 190, 213, 24 },
		{ 190, 214, 24 },
		{ 190, 215, 24 },
		{ 190, 216, 24 },
		{ 190, 217, 24 },
		{ 190, 218, 24 },
		{ 190, 219, 24 },
		{ 190, 220, 24 },
		{ 190, 221, 24 },
		{ 190, 222, 24 },
		{ 190, 223, 24 },
		{ 190, 224, 24 },
		{ 190, 225, 24 },
		{ 190, 226, 24 },
		{ 190, 227, 24 },
		{ 190, 228, 24 },
		{ 190, 229, 24 },
		{ 190, 230, 24 },
		{ 190, 231, 24 },
		{ 190, 232, 24 },
		{ 190, 233, 24 },
		{ 190, 234, 24 },
		{ 190, 235, 24 },
		{ 190, 236, 24 },
		{ 190, 237, 24 },
		{ 190, 238, 24 },
		{ 190, 239, 24 },
		{ 190, 240, 24 },
		{ 191, 208, 24 },
		{ 191, 209, 24 },
		{ 191, 210, 24 },
		{ 191, 211, 24 },
		{ 191, 212, 24 },
		{ 191, 213, 24 },
		{ 191, 214, 24 },
		{ 191, 215, 24 },
		{ 191, 216, 24 },
		{ 191, 217, 24 },
		{ 191, 218, 24 },
		{ 191, 219, 24 },
		{ 191, 220, 24 },
		{ 191, 221, 24 },
		{ 191, 222, 24 },
		{ 191, 223, 24 },
		{ 191, 224, 24 },
		{ 191, 225, 24 },
		{ 191, 226, 24 },
		{ 191, 227, 24 },
		{ 191, 228, 24 },
		{ 191, 229, 24 },
		{ 191, 230, 24 },
		{ 191, 231, 24 },
		{ 191, 232, 24 },
		{ 191, 233, 24 },
		{ 191, 234, 24 },
		{ 191, 235, 24 },
		{ 191, 236, 24 },
		{ 191, 237, 24 },
		{ 191, 238, 24 },
		{ 191, 239, 24 },
		{ 191, 240, 24 },
		{ 192, 209, 24 },
		{ 192, 210, 24 },
		{ 192, 211, 24 },
		{ 192, 212, 24 },
		{ 192, 213, 24 },
		{ 192, 214, 24 },
		{ 192, 215, 24 },
		{ 192, 216, 24 },
		{ 192, 217, 24 },
		{ 192, 218, 24 },
		{ 192, 219, 24 },
		{ 192, 220, 24 },
		{ 192, 221, 24 },
		{ 192, 222, 24 },
		{ 192, 223, 24 },
		{ 192, 224, 24 },
		{ 192, 225, 24 },
		{ 192, 226, 24 },
		{ 192, 227, 24 },
		{ 192, 228, 24 },
		{ 192, 229, 24 },
		{ 192, 230, 24 },
		{ 192, 231, 24 },
		{ 192, 232, 24 },
		{ 192, 233, 24 },
		{ 192, 234, 24 },
		{ 192, 235, 24 },
		{ 192, 236, 24 },
		{ 192, 237, 24 },
		{ 192, 238, 24 },
		{ 192, 239, 24 },
		{ 192, 240, 24 },
		{ 193, 210, 24 },
		{ 193, 211, 24 },
		{ 193, 212, 24 },
		{ 193, 213, 24 },
		{ 193, 214, 24 },
		{ 193, 215, 24 },
		{ 193, 216, 24 },
		{ 193, 217, 24 },
		{ 193, 218, 24 },
		{ 193, 219, 24 },
		{ 193, 220, 24 },
		{ 193, 221, 24 },
		{ 193, 222, 24 },
		{ 193, 223, 24 },
		{ 193, 224, 24 },
		{ 193, 225, 24 },
		{ 193, 226, 24 },
		{ 193, 227, 24 },
		{ 193, 228, 24 },
		{ 193, 229, 24 },
		{ 193, 230, 24 },
		{ 193, 231, 24 },
		{ 193, 232, 24 },
		{ 193, 233, 24 },
		{ 193, 234, 24 },
		{ 193, 235, 24 },
		{ 193, 236, 24 },
		{ 193, 237, 24 },
		{ 193, 238, 24 },
		{ 193, 239, 24 },
		{ 193, 240, 24 },
		{ 194, 210, 24 },
		{ 194, 211, 24 },
		{ 194, 212, 24 },
		{ 194, 213, 24 },
		{ 194, 214, 24 },
		{ 194, 215, 24 },
		{ 194, 216, 24 },
		{ 194, 217, 24 },
		{ 194, 218, 24 },
		{ 194, 219, 24 },
		{ 194, 220, 24 },
		{ 194, 221, 24 },
		{ 194, 222, 24 },
		{ 194, 223, 24 },
		{ 194, 224, 24 },
		{ 194, 225, 24 },
		{ 194, 226, 24 },
		{ 194, 227, 24 },
		{ 194, 228, 24 },
		{ 194, 229, 24 },
		{ 194, 230, 24 },
		{ 194, 231, 24 },
		{ 194, 232, 24 },
		{ 194, 233, 24 },
		{ 194, 234, 24 },
		{ 194, 235, 24 },
		{ 194, 236, 24 },
		{ 194, 237, 24 },
		{ 194, 238, 24 },
		{ 194, 239, 24 },
		{ 194, 240, 24 },
		{ 194, 241, 24 },
		{ 195, 211, 24 },
		{ 195, 212, 24 },
		{ 195, 213, 24 },
		{ 195, 214, 24 },
		{ 195, 215, 24 },
		{ 195, 216, 24 },
		{ 195, 217, 24 },
		{ 195, 218, 24 },
		{ 195, 219, 24 },
		{ 195, 220, 24 },
		{ 195, 221, 24 },
		{ 195, 222, 24 },
		{ 195, 223, 24 },
		{ 195, 224, 24 },
		{ 195, 225, 24 },
		{ 195, 226, 24 },
		{ 195, 227, 24 },
		{ 195, 228, 24 },
		{ 195, 229, 24 },
		{ 195, 230, 24 },
		{ 195, 231, 24 },
		{ 195, 232, 24 },
		{ 195, 233, 24 },
		{ 195, 234, 24 },
		{ 195, 235, 24 },
		{ 195, 236, 24 },
		{ 195, 237, 24 },
		{ 195, 238, 24 },
		{ 195, 239, 24 },
		{ 195, 240, 24 },
		{ 195, 241, 24 },
		{ 196, 212, 24 },
		{ 196, 213, 24 },
		{ 196, 214, 24 },
		{ 196, 215, 24 },
		{ 196, 216, 24 },
		{ 196, 217, 24 },
		{ 196, 218, 24 },
		{ 196, 219, 24 },
		{ 196, 220, 24 },
		{ 196, 221, 24 },
		{ 196, 222, 24 },
		{ 196, 223, 24 },
		{ 196, 224, 24 },
		{ 196, 225, 24 },
		{ 196, 226, 24 },
		{ 196, 227, 24 },
		{ 196, 228, 24 },
		{ 196, 229, 24 },
		{ 196, 230, 24 },
		{ 196, 231, 24 },
		{ 196, 232, 24 },
		{ 196, 233, 24 },
		{ 196, 234, 24 },
		{ 196, 235, 24 },
		{ 196, 236, 24 },
		{ 196, 237, 24 },
		{ 196, 238, 24 },
		{ 196, 239, 24 },
		{ 196, 240, 24 },
		{ 196, 241, 24 },
		{ 197, 213, 24 },
		{ 197, 214, 24 },
		{ 197, 215, 24 },
		{ 197, 216, 24 },
		{ 197, 217, 24 },
		{ 197, 218, 24 },
		{ 197, 219, 24 },
		{ 197, 220, 24 },
		{ 197, 221, 24 },
		{ 197, 222, 24 },
		{ 197, 223, 24 },
		{ 197, 224, 24 },
		{ 197, 225, 24 },
		{ 197, 226, 24 },
		{ 197, 227, 24 },
		{ 197, 228, 24 },
		{ 197, 229, 24 },
		{ 197, 230, 24 },
		{ 197, 231, 24 },
		{ 197, 232, 24 },
		{ 197, 233, 24 },
		{ 197, 234, 24 },
		{ 197, 235, 24 },
		{ 197, 236, 24 },
		{ 197, 237, 24 },
		{ 197, 238, 24 },
		{ 197, 239, 24 },
		{ 197, 240, 24 },
		{ 197, 241, 24 },
		{ 198, 213, 24 },
		{ 198, 214, 24 },
		{ 198, 215, 24 },
		{ 198, 216, 24 },
		{ 198, 217, 24 },
		{ 198, 218, 24 },
		{ 198, 219, 24 },
		{ 198, 220, 24 },
		{ 198, 221, 24 },
		{ 198, 222, 24 },
		{ 198, 223, 24 },
		{ 198, 224, 24 },
		{ 198, 225, 24 },
		{ 198, 226, 24 },
		{ 198, 227, 24 },
		{ 198, 228, 24 },
		{ 198, 229, 24 },
		{ 198, 230, 24 },
		{ 198, 231, 24 },
		{ 198, 232, 24 },
		{ 198, 233, 24 },
		{ 198, 234, 24 },
		{ 198, 235, 24 },
		{ 198, 236, 24 },
		{ 198, 237, 24 },
		{ 198, 238, 24 },
		{ 198, 239, 24 },
		{ 198, 240, 24 },
		{ 198, 241, 24 },
		{ 198, 242, 24 },
		{ 199, 214, 24 },
		{ 199, 215, 24 },
		{ 199, 216, 24 },
		{ 199, 217, 24 },
		{ 199, 218, 24 },
		{ 199, 219, 24 },
		{ 199, 220, 24 },
		{ 199, 221, 24 },
		{ 199, 222, 24 },
		{ 199, 223, 24 },
		{ 199, 224, 24 },
		{ 199, 225, 24 },
		{ 199, 226, 24 },
		{ 199, 227, 24 },
		{ 199, 228, 24 },
		{ 199, 229, 24 },
		{ 199, 230, 24 },
		{ 199, 231, 24 },
		{ 199, 232, 24 },
		{ 199, 233, 24 },
		{ 199, 234, 24 },
		{ 199, 235, 24 },
		{ 199, 236, 24 },
		{ 199, 237, 24 },
		{ 199, 238, 24 },
		{ 199, 239, 24 },
		{ 199, 240, 24 },
		{ 199, 241, 24 },
		{ 199, 242, 24 },
		{ 200, 215, 24 },
		{ 200, 216, 24 },
		{ 200, 217, 24 },
		{ 200, 218, 24 },
		{ 200, 219, 24 },
		{ 200, 220, 24 },
		{ 200, 221, 24 },
		{ 200, 222, 24 },
		{ 200, 223, 24 },
		{ 200, 224, 24 },
		{ 200, 225, 24 },
		{ 200, 226, 24 },
		{ 200, 227, 24 },
		{ 200, 228, 24 },
		{ 200, 229, 24 },
		{ 200, 230, 24 },
		{ 200, 231, 24 },
		{ 200, 232, 24 },
		{ 200, 233, 24 },
		{ 200, 234, 24 },
		{ 200, 235, 24 },
		{ 200, 236, 24 },
		{ 200, 237, 24 },
		{ 200, 238, 24 },
		{ 200, 239, 24 },
		{ 200, 240, 24 },
		{ 200, 241, 24 },
		{ 200, 242, 24 },
		{ 201, 216, 24 },
		{ 201, 217, 24 },
		{ 201, 218, 24 },
		{ 201, 219, 24 },
		{ 201, 220, 24 },
		{ 201, 221, 24 },
		{ 201, 222, 24 },
		{ 201, 223, 24 },
		{ 201, 224, 24 },
		{ 201, 225, 24 },
		{ 201, 226, 24 },
		{ 201, 227, 24 },
		{ 201, 228, 24 },
		{ 201, 229, 24 },
		{ 201, 230, 24 },
		{ 201, 231, 24 },
		{ 201, 232, 24 },
		{ 201, 233, 24 },
		{ 201, 234, 24 },
		{ 201, 235, 24 },
		{ 201, 236, 24 },
		{ 201, 237, 24 },
		{ 201, 238, 24 },
		{ 201, 239, 24 },
		{ 201, 240, 24 },
		{ 201, 241, 24 },
		{ 201, 242, 24 },
		{ 202, 216, 24 },
		{ 202, 217, 24 },
		{ 202, 218, 24 },
		{ 202, 219, 24 },
		{ 202, 220, 24 },
		{ 202, 221, 24 },
		{ 202, 222, 24 },
		{ 202, 223, 24 },
		{ 202, 224, 24 },
		{ 202, 225, 24 },
		{ 202, 226, 24 },
		{ 202, 227, 24 },
		{ 202, 228, 24 },
		{ 202, 229, 24 },
		{ 202, 230, 24 },
		{ 202, 231, 24 },
		{ 202, 232, 24 },
		{ 202, 233, 24 },
		{ 202, 234, 24 },
		{ 202, 235, 24 },
		{ 202, 236, 24 },
		{ 202, 237, 24 },
		{ 202, 238, 24 },
		{ 202, 239, 24 },
		{ 202, 240, 24 },
		{ 202, 241, 24 },
		{ 202, 242, 24 },
		{ 202, 243, 24 },
		{ 203, 217, 24 },
		{ 203, 218, 24 },
		{ 203, 219, 24 },
		{ 203, 220, 24 },
		{ 203, 221, 24 },
		{ 203, 222, 24 },
		{ 203, 223, 24 },
		{ 203, 224, 24 },
		{ 203, 225, 24 },
		{ 203, 226, 24 },
		{ 203, 227, 24 },
		{ 203, 228, 24 },
		{ 203, 229, 24 },
		{ 203, 230, 24 },
		{ 203, 231, 24 },
		{ 203, 232, 24 },
		{ 203, 233, 24 },
		{ 203, 234, 24 },
		{ 203, 235, 24 },
		{ 203, 236, 24 },
		{ 203, 237, 24 },
		{ 203, 238, 24 },
		{ 203, 239, 24 },
		{ 203, 240, 24 },
		{ 203, 241, 24 },
		{ 203, 242, 24 },
		{ 203, 243, 24 },
		{ 204, 218, 24 },
		{ 204, 219, 24 },
		{ 204, 220, 24 },
		{ 204, 221, 24 },
		{ 204, 222, 24 },
		{ 204, 223, 24 },
		{ 204, 224, 24 },
		{ 204, 225, 24 },
		{ 204, 226, 24 },
		{ 204, 227, 24 },
		{ 204, 228, 24 },
		{ 204, 229, 24 },
		{ 204, 230, 24 },
		{ 204, 231, 24 },
		{ 204, 232, 24 },
		{ 204, 233, 24 },
		{ 204, 234, 24 },
		{ 204, 235, 24 },
		{ 204, 236, 24 },
		{ 204, 237, 24 },
		{ 204, 238, 24 },
		{ 204, 239, 24 },
		{ 204, 240, 24 },
		{ 204, 241, 24 },
		{ 204, 242, 24 },
		{ 204, 243, 24 },
		{ 205, 219, 24 },
		{ 205, 220, 24 },
		{ 205, 221, 24 },
		{ 205, 222, 24 },
		{ 205, 223, 24 },
		{ 205, 224, 24 },
		{ 205, 225, 24 },
		{ 205, 226, 24 },
		{ 205, 227, 24 },
		{ 205, 228, 24 },
		{ 205, 229, 24 },
		{ 205, 230, 24 },
		{ 205, 231, 24 },
		{ 205, 232, 24 },
		{ 205, 233, 24 },
		{ 205, 234, 24 },
		{ 205, 235, 24 },
		{ 205, 236, 24 },
		{ 205, 237, 24 },
		{ 205, 238, 24 },
		{ 205, 239, 24 },
		{ 205, 240, 24 },
		{ 205, 241, 24 },
		{ 205, 242, 24 },
		{ 205, 243, 24 },
		{ 206, 219, 24 },
		{ 206, 220, 24 },
		{ 206, 221, 24 },
		{ 206, 222, 24 },
		{ 206, 223, 24 },
		{ 206, 224, 24 },
		{ 206, 225, 24 },
		{ 206, 226, 24 },
		{ 206, 227, 24 },
		{ 206, 228, 24 },
		{ 206, 229, 24 },
		{ 206, 230, 24 },
		{ 206, 231, 24 },
		{ 206, 232, 24 },
		{ 206, 233, 24 },
		{ 206, 234, 24 },
		{ 206, 235, 24 },
		{ 206, 236, 24 },
		{ 206, 237, 24 },
		{ 206, 238, 24 },
		{ 206, 239, 24 },
		{ 206, 240, 24 },
		{ 206, 241, 24 },
		{ 206, 242, 24 },
		{ 206, 243, 24 },
		{ 206, 244, 24 },
		{ 207, 220, 24 },
		{ 207, 221, 24 },
		{ 207, 222, 24 },
		{ 207, 223, 24 },
		{ 207, 224, 24 },
		{ 207, 225, 24 },
		{ 207, 226, 24 },
		{ 207, 227, 24 },
		{ 207, 228, 24 },
		{ 207, 229, 24 },
		{ 207, 230, 24 },
		{ 207, 231, 24 },
		{ 207, 232, 24 },
		{ 207, 233, 24 },
		{ 207, 234, 24 },
		{ 207, 235, 24 },
		{ 207, 236, 24 },
		{ 207, 237, 24 },
		{ 207, 238, 24 },
		{ 207, 239, 24 },
		{ 207, 240, 24 },
		{ 207, 241, 24 },
		{ 207, 242, 24 },
		{ 207, 243, 24 },
		{ 207, 244, 24 },
		{ 208, 221, 24 },
		{ 208, 222, 24 },
		{ 208, 223, 24 },
		{ 208, 224, 24 },
		{ 208, 225, 24 },
		{ 208, 226, 24 },
		{ 208, 227, 24 },
		{ 208, 228, 24 },
		{ 208, 229, 24 },
		{ 208, 230, 24 },
		{ 208, 231, 24 },
		{ 208, 232, 24 },
		{ 208, 233, 24 },
		{ 208, 234, 24 },
		{ 208, 235, 24 },
		{ 208, 236, 24 },
		{ 208, 237, 24 },
		{ 208, 238, 24 },
		{ 208, 239, 24 },
		{ 208, 240, 24 },
		{ 208, 241, 24 },
		{ 208, 242, 24 },
		{ 208, 243, 24 },
		{ 208, 244, 24 },
		{ 209, 222, 24 },
		{ 209, 223, 24 },
		{ 209, 224, 24 },
		{ 209, 225, 24 },
		{ 209, 226, 24 },
		{ 209, 227, 24 },
		{ 209, 228, 24 },
		{ 209, 229, 24 },
		{ 209, 230, 24 },
		{ 209, 231, 24 },
		{ 209, 232, 24 },
		{ 209, 233, 24 },
		{ 209, 234, 24 },
		{ 209, 235, 24 },
		{ 209, 236, 24 },
		{ 209, 237, 24 },
		{ 209, 238, 24 },
		{ 209, 239, 24 },
		{ 209, 240, 24 },
		{ 209, 241, 24 },
		{ 209, 242, 24 },
		{ 209, 243, 24 },
		{ 209, 244, 24 },
		{ 210, 222, 24 },
		{ 210, 223, 24 },
		{ 210, 224, 24 },
		{ 210, 225, 24 },
		{ 210, 226, 24 },
		{ 210, 227, 24 },
		{ 210, 228, 24 },
		{ 210, 229, 24 },
		{ 210, 230, 24 },
		{ 210, 231, 24 },
		{ 210, 232, 24 },
		{ 210, 233, 24 },
		{ 210, 234, 24 },
		{ 210, 235, 24 },
		{ 210, 236, 24 },
		{ 210, 237, 24 },
		{ 210, 238, 24 },
		{ 210, 239, 24 },
		{ 210, 240, 24 },
		{ 210, 241, 24 },
		{ 210, 242, 24 },
		{ 210, 243, 24 },
		{ 210, 244, 24 },
		{ 210, 245, 24 },
		{ 211, 223, 24 },
		{ 211, 224, 24 },
		{ 211, 225, 24 },
		{ 211, 226, 24 },
		{ 211, 227, 24 },
		{ 211, 228, 24 },
		{ 211, 229, 24 },
		{ 211, 230, 24 },
		{ 211, 231, 24 },
		{ 211, 232, 24 },
		{ 211, 233, 24 },
		{ 211, 234, 24 },
		{ 211, 235, 24 },
		{ 211, 236, 24 },
		{ 211, 237, 24 },
		{ 211, 238, 24 },
		{ 211, 239, 24 },
		{ 211, 240, 24 },
		{ 211, 241, 24 },
		{ 211, 242, 24 },
		{ 211, 243, 24 },
		{ 211, 244, 24 },
		{ 211, 245, 24 },
		{ 212, 224, 24 },
		{ 212, 225, 24 },
		{ 212, 226, 24 },
		{ 212, 227, 24 },
		{ 212, 228, 24 },
		{ 212, 229, 24 },
		{ 212, 230, 24 },
		{ 212, 231, 24 },
		{ 212, 232, 24 },
		{ 212, 233, 24 },
		{ 212, 234, 24 },
		{ 212, 235, 24 },
		{ 212, 236, 24 },
		{ 212, 237, 24 },
		{ 212, 238, 24 },
		{ 212, 239, 24 },
		{ 212, 240, 24 },
		{ 212, 241, 24 },
		{ 212, 242, 24 },
		{ 212, 243, 24 },
		{ 212, 244, 24 },
		{ 212, 245, 24 },
		{ 213, 225, 24 },
		{ 213, 226, 24 },
		{ 213, 227, 24 },
		{ 213, 228, 24 },
		{ 213, 229, 24 },
		{ 213, 230, 24 },
		{ 213, 231, 24 },
		{ 213, 232, 24 },
		{ 213, 233, 24 },
		{ 213, 234, 24 },
		{ 213, 235, 24 },
		{ 213, 236, 24 },
		{ 213, 237, 24 },
		{ 213, 238, 24 },
		{ 213, 239, 24 },
		{ 213, 240, 24 },
		{ 213, 241, 24 },
		{ 213, 242, 24 },
		{ 213, 243, 24 },
		{ 213, 244, 24 },
		{ 213, 245, 24 },
		{ 214, 225, 24 },
		{ 214, 226, 24 },
		{ 214, 227, 24 },
		{ 214, 228, 24 },
		{ 214, 229, 24 },
		{ 214, 230, 24 },
		{ 214, 231, 24 },
		{ 214, 232, 24 },
		{ 214, 233, 24 },
		{ 214, 234, 24 },
		{ 214, 235, 24 },
		{ 214, 236, 24 },
		{ 214, 237, 24 },
		{ 214, 238, 24 },
		{ 214, 239, 24 },
		{ 214, 240, 24 },
		{ 214, 241, 24 },
		{ 214, 242, 24 },
		{ 214, 243, 24 },
		{ 214, 244, 24 },
		{ 214, 245, 24 },
		{ 214, 246, 24 },
		{ 215, 226, 24 },
		{ 215, 227, 24 },
		{ 215, 228, 24 },
		{ 215, 229, 24 },
		{ 215, 230, 24 },
		{ 215, 231, 24 },
		{ 215, 232, 24 },
		{ 215, 233, 24 },
		{ 215, 234, 24 },
		{ 215, 235, 24 },
		{ 215, 236, 24 },
		{ 215, 237, 24 },
		{ 215, 238, 24 },
		{ 215, 239, 24 },
		{ 215, 240, 24 },
		{ 215, 241, 24 },
		{ 215, 242, 24 },
		{ 215, 243, 24 },
		{ 215, 244, 24 },
		{ 215, 245, 24 },
		{ 215, 246, 24 },
		{ 216, 227, 24 },
		{ 216, 228, 24 },
		{ 216, 229, 24 },
		{ 216, 230, 24 },
		{ 216, 231, 24 },
		{ 216, 232, 24 },
		{ 216, 233, 24 },
		{ 216, 234, 24 },
		{ 216, 235, 24 },
		{ 216, 236, 24 },
		{ 216, 237, 24 },
		{ 216, 238, 24 },
		{ 216, 239, 24 },
		{ 216, 240, 24 },
		{ 216, 241, 24 },
		{ 216, 242, 24 },
		{ 216, 243, 24 },
		{ 216, 244, 24 },
		{ 216, 245, 24 },
		{ 216, 246, 24 },
		{ 217, 228, 24 },
		{ 217, 229, 24 },
		{ 217, 230, 24 },
		{ 217, 231, 24 },
		{ 217, 232, 24 },
		{ 217, 233, 24 },
		{ 217, 234, 24 },
		{ 217, 235, 24 },
		{ 217, 236, 24 },
		{ 217, 237, 24 },
		{ 217, 238, 24 },
		{ 217, 239, 24 },
		{ 217, 240, 24 },
		{ 217, 241, 24 },
		{ 217, 242, 24 },
		{ 217, 243, 24 },
		{ 217, 244, 24 },
		{ 217, 245, 24 },
		{ 217, 246, 24 },
		{ 218, 228, 24 },
		{ 218, 229, 24 },
		{ 218, 230, 24 },
		{ 218, 231, 24 },
		{ 218, 232, 24 },
		{ 218, 233, 24 },
		{ 218, 234, 24 },
		{ 218, 235, 24 },
		{ 218, 236, 24 },
		{ 218, 237, 24 },
		{ 218, 238, 24 },
		{ 218, 239, 24 },
		{ 218, 240, 24 },
		{ 218, 241, 24 },
		{ 218, 242, 24 },
		{ 218, 243, 24 },
		{ 218, 244, 24 },
		{ 218, 245, 24 },
		{ 218, 246, 24 },
		{ 218, 247, 24 },
		{ 219, 229, 24 },
		{ 219, 230, 24 },
		{ 219, 231, 24 },
		{ 219, 232, 24 },
		{ 219, 233, 24 },
		{ 219, 234, 24 },
		{ 219, 235, 24 },
		{ 219, 236, 24 },
		{ 219, 237, 24 },
		{ 219, 238, 24 },
		{ 219, 239, 24 },
		{ 219, 240, 24 },
		{ 219, 241, 24 },
		{ 219, 242, 24 },
		{ 219, 243, 24 },
		{ 219, 244, 24 },
		{ 219, 245, 24 },
		{ 219, 246, 24 },
		{ 219, 247, 24 },
		{ 220, 230, 24 },
		{ 220, 231, 24 },
		{ 220, 232, 24 },
		{ 220, 233, 24 },
		{ 220, 234, 24 },
		{ 220, 235, 24 },
		{ 220, 236, 24 },
		{ 220, 237, 24 },
		{ 220, 238, 24 },
		{ 220, 239, 24 },
		{ 220, 240, 24 },
		{ 220, 241, 24 },
		{ 220, 242, 24 },
		{ 220, 243, 24 },
		{ 220, 244, 24 },
		{ 220, 245, 24 },
		{ 220, 246, 24 },
		{ 220, 247, 24 },
		{ 221, 231, 24 },
		{ 221, 232, 24 },
		{ 221, 233, 24 },
		{ 221, 234, 24 },
		{ 221, 235, 24 },
		{ 221, 236, 24 },
		{ 221, 237, 24 },
		{ 221, 238, 24 },
		{ 221, 239, 24 },
		{ 221, 240, 24 },
		{ 221, 241, 24 },
		{ 221, 242, 24 },
		{ 221, 243, 24 },
		{ 221, 244, 24 },
		{ 221, 245, 24 },
		{ 221, 246, 24 },
		{ 221, 247, 24 },
		{ 222, 231, 24 },
		{ 222, 232, 24 },
		{ 222, 233, 24 },
		{ 222, 234, 24 },
		{ 222, 235, 24 },
		{ 222, 236, 24 },
		{ 222, 237, 24 },
		{ 222, 238, 24 },
		{ 222, 239, 24 },
		{ 222, 240, 24 },
		{ 222, 241, 24 },
		{ 222, 242, 24 },
		{ 222, 243, 24 },
		{ 222, 244, 24 },
		{ 222, 245, 24 },
		{ 222, 246, 24 },
		{ 222, 247, 24 },
		{ 222, 248, 24 },
		{ 223, 232, 24 },
		{ 223, 233, 24 },
		{ 223, 234, 24 },
		{ 223, 235, 24 },
		{ 223, 236, 24 },
		{ 223, 237, 24 },
		{ 223, 238, 24 },
		{ 223, 239, 24 },
		{ 223, 240, 24 },
		{ 223, 241, 24 },
		{ 223, 242, 24 },
		{ 223, 243, 24 },
		{ 223, 244, 24 },
		{ 223, 245, 24 },
		{ 223, 246, 24 },
		{ 223, 247, 24 },
		{ 223, 248, 24 },
		{ 224, 233, 24 },
		{ 224, 234, 24 },
		{ 224, 235, 24 },
		{ 224, 236, 24 },
		{ 224, 237, 24 },
		{ 224, 238, 24 },
		{ 224, 239, 24 },
		{ 224, 240, 24 },
		{ 224, 241, 24 },
		{ 224, 242, 24 },
		{ 224, 243, 24 },
		{ 224, 244, 24 },
		{ 224, 245, 24 },
		{ 224, 246, 24 },
		{ 224, 247, 24 },
		{ 224, 248, 24 },
		{ 225, 234, 24 },
		{ 225, 235, 24 },
		{ 225, 236, 24 },
		{ 225, 237, 24 },
		{ 225, 238, 24 },
		{ 225, 239, 24 },
		{ 225, 240, 24 },
		{ 225, 241, 24 },
		{ 225, 242, 24 },
		{ 225, 243, 24 },
		{ 225, 244, 24 },
		{ 225, 245, 24 },
		{ 225, 246, 24 },
		{ 225, 247, 24 },
		{ 225, 248, 24 },
		{ 226, 234, 24 },
		{ 226, 235, 24 },
		{ 226, 236, 24 },
		{ 226, 237, 24 },
		{ 226, 238, 24 },
		{ 226, 239, 24 },
		{ 226, 240, 24 },
		{ 226, 241, 24 },
		{ 226, 242, 24 },
		{ 226, 243, 24 },
		{ 226, 244, 24 },
		{ 226, 245, 24 },
		{ 226, 246, 24 },
		{ 226, 247, 24 },
		{ 226, 248, 24 },
		{ 226, 249, 24 },
		{ 227, 235, 24 },
		{ 227, 236, 24 },
		{ 227, 237, 24 },
		{ 227, 238, 24 },
		{ 227, 239, 24 },
		{ 227, 240, 24 },
		{ 227, 241, 24 },
		{ 227, 242, 24 },
		{ 227, 243, 24 },
		{ 227, 244, 24 },
		{ 227, 245, 24 },
		{ 227, 246, 24 },
		{ 227, 247, 24 },
		{ 227, 248, 24 },
		{ 227, 249, 24 },
		{ 228, 236, 24 },
		{ 228, 237, 24 },
		{ 228, 238, 24 },
		{ 228, 239, 24 },
		{ 228, 240, 24 },
		{ 228, 241, 24 },
		{ 228, 242, 24 },
		{ 228, 243, 24 },
		{ 228, 244, 24 },
		{ 228, 245, 24 },
		{ 228, 246, 24 },
		{ 228, 247, 24 },
		{ 228, 248, 24 },
		{ 228, 249, 24 },
		{ 229, 237, 24 },
		{ 229, 238, 24 },
		{ 229, 239, 24 },
		{ 229, 240, 24 },
		{ 229, 241, 24 },
		{ 229, 242, 24 },
		{ 229, 243, 24 },
		{ 229, 244, 24 },
		{ 229, 245, 24 },
		{ 229, 246, 24 },
		{ 229, 247, 24 },
		{ 229, 248, 24 },
		{ 229, 249, 24 },
		{ 230, 237, 24 },
		{ 230, 238, 24 },
		{ 230, 239, 24 },
		{ 230, 240, 24 },
		{ 230, 241, 24 },
		{ 230, 242, 24 },
		{ 230, 243, 24 },
		{ 230, 244, 24 },
		{ 230, 245, 24 },
		{ 230, 246, 24 },
		{ 230, 247, 24 },
		{ 230, 248, 24 },
		{ 230, 249, 24 },
		{ 230, 250, 24 },
		{ 231, 238, 24 },
		{ 231, 239, 24 },
		{ 231, 240, 24 },
		{ 231, 241, 24 },
		{ 231, 242, 24 },
		{ 231, 243, 24 },
		{ 231, 244, 24 },
		{ 231, 245, 24 },
		{ 231, 246, 24 },
		{ 231, 247, 24 },
		{ 231, 248, 24 },
		{ 231, 249, 24 },
		{ 231, 250, 24 },
		{ 232, 239, 24 },
		{ 232, 240, 24 },
		{ 232, 241, 24 },
		{ 232, 242, 24 },
		{ 232, 243, 24 },
		{ 232, 244, 24 },
		{ 232, 245, 24 },
		{ 232, 246, 24 },
		{ 232, 247, 24 },
		{ 232, 248, 24 },
		{ 232, 249, 24 },
		{ 232, 250, 24 },
		{ 233, 240, 24 },
		{ 233, 241, 24 },
		{ 233, 242, 24 },
		{ 233, 243, 24 },
		{ 233, 244, 24 },
		{ 233, 245, 24 },
		{ 233, 246, 24 },
		{ 233, 247, 24 },
		{ 233, 248, 24 },
		{ 233, 249, 24 },
		{ 233, 250, 24 },
		{ 234, 240, 24 },
		{ 234, 241, 24 },
		{ 234, 242, 24 },
		{ 234, 243, 24 },
		{ 234, 244, 24 },
		{ 234, 245, 24 },
		{ 234, 246, 24 },
		{ 234, 247, 24 },
		{ 234, 248, 24 },
		{ 234, 249, 24 },
		{ 234, 250, 24 },
		{ 234, 251, 24 },
		{ 235, 241, 24 },
		{ 235, 242, 24 },
		{ 235, 243, 24 },
		{ 235, 244, 24 },
		{ 235, 245, 24 },
		{ 235, 246, 24 },
		{ 235, 247, 24 },
		{ 235, 248, 24 },
		{ 235, 249, 24 },
		{ 235, 250, 24 },
		{ 235, 251, 24 },
		{ 236, 242, 24 },
		{ 236, 243, 24 },
		{ 236, 244, 24 },
		{ 236, 245, 24 },
		{ 236, 246, 24 },
		{ 236, 247, 24 },
		{ 236, 248, 24 },
		{ 236, 249, 24 },
		{ 236, 250, 24 },
		{ 236, 251, 24 },
		{ 237, 243, 24 },
		{ 237, 244, 24 },
		{ 237, 245, 24 },
		{ 237, 246, 24 },
		{ 237, 247, 24 },
		{ 237, 248, 24 },
		{ 237, 249, 24 },
		{ 237, 250, 24 },
		{ 237, 251, 24 },
		{ 238, 243, 24 },
		{ 238, 244, 24 },
		{ 238, 245, 24 },
		{ 238, 246, 24 },
		{ 238, 247, 24 },
		{ 238, 248, 24 },
		{ 238, 249, 24 },
		{ 238, 250, 24 },
		{ 238, 251, 24 },
		{ 238, 252, 24 },
		{ 239, 244, 24 },
		{ 239, 245, 24 },
		{ 239, 246, 24 },
		{ 239, 247, 24 },
		{ 239, 248, 24 },
		{ 239, 249, 24 },
		{ 239, 250, 24 },
		{ 239, 251, 24 },
		{ 239, 252, 24 },
		{ 240, 245, 24 },
		{ 240, 246, 24 },
		{ 240, 247, 24 },
		{ 240, 248, 24 },
		{ 240, 249, 24 },
		{ 240, 250, 24 },
		{ 240, 251, 24 },
		{ 240, 252, 24 },
		{ 274, 262, 24 },
		{ 274, 263, 24 },
		{ 274, 264, 24 },
		{ 274, 265, 24 },
		{ 274, 266, 24 },
		{ 274, 267, 24 },
		{ 274, 268, 24 },
		{ 274, 269, 24 },
		{ 275, 262, 24 },
		{ 275, 263, 24 },
		{ 275, 264, 24 },
		{ 275, 265, 24 },
		{ 275, 266, 24 },
		{ 275, 267, 24 },
		{ 275, 268, 24 },
		{ 275, 269, 24 },
		{ 275, 270, 24 },
		{ 276, 262, 24 },
		{ 276, 263, 24 },
		{ 276, 264, 24 },
		{ 276, 265, 24 },
		{ 276, 266, 24 },
		{ 276, 267, 24 },
		{ 276, 268, 24 },
		{ 276, 269, 24 },
		{ 276, 270, 24 },
		{ 276, 271, 24 },
		{ 277, 263, 24 },
		{ 277, 264, 24 },
		{ 277, 265, 24 },
		{ 277, 266, 24 },
		{ 277, 267, 24 },
		{ 277, 268, 24 },
		{ 277, 269, 24 },
		{ 277, 270, 24 },
		{ 277, 271, 24 },
		{ 278, 263, 24 },
		{ 278, 264, 24 },
		{ 278, 265, 24 },
		{ 278, 266, 24 },
		{ 278, 267, 24 },
		{ 278, 268, 24 },
		{ 278, 269, 24 },
		{ 278, 270, 24 },
		{ 278, 271, 24 },
		{ 278, 272, 24 },
		{ 279, 263, 24 },
		{ 279, 264, 24 },
		{ 279, 265, 24 },
		{ 279, 266, 24 },
		{ 279, 267, 24 },
		{ 279, 268, 24 },
		{ 279, 269, 24 },
		{ 279, 270, 24 },
		{ 279, 271, 24 },
		{ 279, 272, 24 },
		{ 279, 273, 24 },
		{ 280, 263, 24 },
		{ 280, 264, 24 },
		{ 280, 265, 24 },
		{ 280, 266, 24 },
		{ 280, 267, 24 },
		{ 280, 268, 24 },
		{ 280, 269, 24 },
		{ 280, 270, 24 },
		{ 280, 271, 24 },
		{ 280, 272, 24 },
		{ 280, 273, 24 },
		{ 280, 274, 24 },
		{ 281, 264, 24 },
		{ 281, 265, 24 },
		{ 281, 266, 24 },
		{ 281, 267, 24 },
		{ 281, 268, 24 },
		{ 281, 269, 24 },
		{ 281, 270, 24 },
		{ 281, 271, 24 },
		{ 281, 272, 24 },
		{ 281, 273, 24 },
		{ 281, 274, 24 },
		{ 282, 264, 24 },
		{ 282, 265, 24 },
		{ 282, 266, 24 },
		{ 282, 267, 24 },
		{ 282, 268, 24 },
		{ 282, 269, 24 },
		{ 282, 270, 24 },
		{ 282, 271, 24 },
		{ 282, 272, 24 },
		{ 282, 273, 24 },
		{ 282, 274, 24 },
		{ 282, 275, 24 },
		{ 283, 264, 24 },
		{ 283, 265, 24 },
		{ 283, 266, 24 },
		{ 283, 267, 24 },
		{ 283, 268, 24 },
		{ 283, 269, 24 },
		{ 283, 270, 24 },
		{ 283, 271, 24 },
		{ 283, 272, 24 },
		{ 283, 273, 24 },
		{ 283, 274, 24 },
		{ 283, 275, 24 },
		{ 283, 276, 24 },
		{ 284, 264, 24 },
		{ 284, 265, 24 },
		{ 284, 266, 24 },
		{ 284, 267, 24 },
		{ 284, 268, 24 },
		{ 284, 269, 24 },
		{ 284, 270, 24 },
		{ 284, 271, 24 },
		{ 284, 272, 24 },
		{ 284, 273, 24 },
		{ 284, 274, 24 },
		{ 284, 275, 24 },
		{ 284, 276, 24 },
		{ 284, 277, 24 },
		{ 285, 265, 24 },
		{ 285, 266, 24 },
		{ 285, 267, 24 },
		{ 285, 268, 24 },
		{ 285, 269, 24 },
		{ 285, 270, 24 },
		{ 285, 271, 24 },
		{ 285, 272, 24 },
		{ 285, 273, 24 },
		{ 285, 274, 24 },
		{ 285, 275, 24 },
		{ 285, 276, 24 },
		{ 285, 277, 24 },
		{ 286, 265, 24 },
		{ 286, 266, 24 },
		{ 286, 267, 24 },
		{ 286, 268, 24 },
		{ 286, 269, 24 },
		{ 286, 270, 24 },
		{ 286, 271, 24 },
		{ 286, 272, 24 },
		{ 286, 273, 24 },
		{ 286, 274, 24 },
		{ 286, 275, 24 },
		{ 286, 276, 24 },
		{ 286, 277, 24 },
		{ 286, 278, 24 },
		{ 287, 265, 24 },
		{ 287, 266, 24 },
		{ 287, 267, 24 },
		{ 287, 268, 24 },
		{ 287, 269, 24 },
		{ 287, 270, 24 },
		{ 287, 271, 24 },
		{ 287, 272, 24 },
		{ 287, 273, 24 },
		{ 287, 274, 24 },
		{ 287, 275, 24 },
		{ 287, 276, 24 },
		{ 287, 277, 24 },
		{ 287, 278, 24 },
		{ 287, 279, 24 },
		{ 288, 265, 24 },
		{ 288, 266, 24 },
		{ 288, 267, 24 },
		{ 288, 268, 24 },
		{ 288, 269, 24 },
		{ 288, 270, 24 },
		{ 288, 271, 24 },
		{ 288, 272, 24 },
		{ 288, 273, 24 },
		{ 288, 274, 24 },
		{ 288, 275, 24 },
		{ 288, 276, 24 },
		{ 288, 277, 24 },
		{ 288, 278, 24 },
		{ 288, 279, 24 },
		{ 288, 280, 24 },
		{ 289, 266, 24 },
		{ 289, 267, 24 },
		{ 289, 268, 24 },
		{ 289, 269, 24 },
		{ 289, 270, 24 },
		{ 289, 271, 24 },
		{ 289, 272, 24 },
		{ 289, 273, 24 },
		{ 289, 274, 24 },
		{ 289, 275, 24 },
		{ 289, 276, 24 },
		{ 289, 277, 24 },
		{ 289, 278, 24 },
		{ 289, 279, 24 },
		{ 289, 280, 24 },
		{ 290, 266, 24 },
		{ 290, 267, 24 },
		{ 290, 268, 24 },
		{ 290, 269, 24 },
		{ 290, 270, 24 },
		{ 290, 271, 24 },
		{ 290, 272, 24 },
		{ 290, 273, 24 },
		{ 290, 274, 24 },
		{ 290, 275, 24 },
		{ 290, 276, 24 },
		{ 290, 277, 24 },
		{ 290, 278, 24 },
		{ 290, 279, 24 },
		{ 290, 280, 24 },
		{ 290, 281, 24 },
		{ 291, 266, 24 },
		{ 291, 267, 24 },
		{ 291, 268, 24 },
		{ 291, 269, 24 },
		{ 291, 270, 24 },
		{ 291, 271, 24 },
		{ 291, 272, 24 },
		{ 291, 273, 24 },
		{ 291, 274, 24 },
		{ 291, 275, 24 },
		{ 291, 276, 24 },
		{ 291, 277, 24 },
		{ 291, 278, 24 },
		{ 291, 279, 24 },
		{ 291, 280, 24 },
		{ 291, 281, 24 },
		{ 291, 282, 24 },
		{ 292, 266, 24 },
		{ 292, 267, 24 },
		{ 292, 268, 24 },
		{ 292, 269, 24 },
		{ 292, 270, 24 },
		{ 292, 271, 24 },
		{ 292, 272, 24 },
		{ 292, 273, 24 },
		{ 292, 274, 24 },
		{ 292, 275, 24 },
		{ 292, 276, 24 },
		{ 292, 277, 24 },
		{ 292, 278, 24 },
		{ 292, 279, 24 },
		{ 292, 280, 24 },
		{ 292, 281, 24 },
		{ 292, 282, 24 },
		{ 292, 283, 24 },
		{ 293, 267, 24 },
		{ 293, 268, 24 },
		{ 293, 269, 24 },
		{ 293, 270, 24 },
		{ 293, 271, 24 },
		{ 293, 272, 24 },
		{ 293, 273, 24 },
		{ 293, 274, 24 },
		{ 293, 275, 24 },
		{ 293, 276, 24 },
		{ 293, 277, 24 },
		{ 293, 278, 24 },
		{ 293, 279, 24 },
		{ 293, 280, 24 },
		{ 293, 281, 24 },
		{ 293, 282, 24 },
		{ 293, 283, 24 },
		{ 294, 267, 24 },
		{ 294, 268, 24 },
		{ 294, 269, 24 },
		{ 294, 270, 24 },
		{ 294, 271, 24 },
		{ 294, 272, 24 },
		{ 294, 273, 24 },
		{ 294, 274, 24 },
		{ 294, 275, 24 },
		{ 294, 276, 24 },
		{ 294, 277, 24 },
		{ 294, 278, 24 },
		{ 294, 279, 24 },
		{ 294, 280, 24 },
		{ 294, 281, 24 },
		{ 294, 282, 24 },
		{ 294, 283, 24 },
		{ 294, 284, 24 },
		{ 295, 267, 24 },
		{ 295, 268, 24 },
		{ 295, 269, 24 },
		{ 295, 270, 24 },
		{ 295, 271, 24 },
		{ 295, 272, 24 },
		{ 295, 273, 24 },
		{ 295, 274, 24 },
		{ 295, 275, 24 },
		{ 295, 276, 24 },
		{ 295, 277, 24 },
		{ 295, 278, 24 },
		{ 295, 279, 24 },
		{ 295, 280, 24 },
		{ 295, 281, 24 },
		{ 295, 282, 24 },
		{ 295, 283, 24 },
		{ 295, 284, 24 },
		{ 295, 285, 24 },
		{ 296, 267, 24 },
		{ 296, 268, 24 },
		{ 296, 269, 24 },
		{ 296, 270, 24 },
		{ 296, 271, 24 },
		{ 296, 272, 24 },
		{ 296, 273, 24 },
		{ 296, 274, 24 },
		{ 296, 275, 24 },
		{ 296, 276, 24 },
		{ 296, 277, 24 },
		{ 296, 278, 24 },
		{ 296, 279, 24 },
		{ 296, 280, 24 },
		{ 296, 281, 24 },
		{ 296, 282, 24 },
		{ 296, 283, 24 },
		{ 296, 284, 24 },
		{ 296, 285, 24 },
		{ 296, 286, 24 },
		{ 297, 268, 24 },
		{ 297, 269, 24 },
		{ 297, 270, 24 },
		{ 297, 271, 24 },
		{ 297, 272, 24 },
		{ 297, 273, 24 },
		{ 297, 274, 24 },
		{ 297, 275, 24 },
		{ 297, 276, 24 },
		{ 297, 277, 24 },
		{ 297, 278, 24 },
		{ 297, 279, 24 },
		{ 297, 280, 24 },
		{ 297, 281, 24 },
		{ 297, 282, 24 },
		{ 297, 283, 24 },
		{ 297, 284, 24 },
		{ 297, 285, 24 },
		{ 297, 286, 24 },
		{ 298, 268, 24 },
		{ 298, 269, 24 },
		{ 298, 270, 24 },
		{ 298, 271, 24 },
		{ 298, 272, 24 },
		{ 298, 273, 24 },
		{ 298, 274, 24 },
		{ 298, 275, 24 },
		{ 298, 276, 24 },
		{ 298, 277, 24 },
		{ 298, 278, 24 },
		{ 298, 279, 24 },
		{ 298, 280, 24 },
		{ 298, 281, 24 },
		{ 298, 282, 24 },
		{ 298, 283, 24 },
		{ 298, 284, 24 },
		{ 298, 285, 24 },
		{ 298, 286, 24 },
		{ 298, 287, 24 },
		{ 299, 268, 24 },
		{ 299, 269, 24 },
		{ 299, 270, 24 },
		{ 299, 271, 24 },
		{ 299, 272, 24 },
		{ 299, 273, 24 },
		{ 299, 274, 24 },
		{ 299, 275, 24 },
		{ 299, 276, 24 },
		{ 299, 277, 24 },
		{ 299, 278, 24 },
		{ 299, 279, 24 },
		{ 299, 280, 24 },
		{ 299, 281, 24 },
		{ 299, 282, 24 },
		{ 299, 283, 24 },
		{ 299, 284, 24 },
		{ 299, 285, 24 },
		{ 299, 286, 24 },
		{ 299, 287, 24 },
		{ 299, 288, 24 },
		{ 300, 268, 24 },
		{ 300, 269, 24 },
		{ 300, 270, 24 },
		{ 300, 271, 24 },
		{ 300, 272, 24 },
		{ 300, 273, 24 },
		{ 300, 274, 24 },
		{ 300, 275, 24 },
		{ 300, 276, 24 },
		{ 300, 277, 24 },
		{ 300, 278, 24 },
		{ 300, 279, 24 },
		{ 300, 280, 24 },
		{ 300, 281, 24 },
		{ 300, 282, 24 },
		{ 300, 283, 24 },
		{ 300, 284, 24 },
		{ 300, 285, 24 },
		{ 300, 286, 24 },
		{ 300, 287, 24 },
		{ 300, 288, 24 },
		{ 300, 289, 24 },
		{ 301, 269, 24 },
		{ 301, 270, 24 },
		{ 301, 271, 24 },
		{ 301, 272, 24 },
		{ 301, 273, 24 },
		{ 301, 274, 24 },
		{ 301, 275, 24 },
		{ 301, 276, 24 },
		{ 301, 277, 24 },
		{ 301, 278, 24 },
		{ 301, 279, 24 },
		{ 301, 280, 24 },
		{ 301, 281, 24 },
		{ 301, 282, 24 },
		{ 301, 283, 24 },
		{ 301, 284, 24 },
		{ 301, 285, 24 },
		{ 301, 286, 24 },
		{ 301, 287, 24 },
		{ 301, 288, 24 },
		{ 301, 289, 24 },
		{ 302, 269, 24 },
		{ 302, 270, 24 },
		{ 302, 271, 24 },
		{ 302, 272, 24 },
		{ 302, 273, 24 },
		{ 302, 274, 24 },
		{ 302, 275, 24 },
		{ 302, 276, 24 },
		{ 302, 277, 24 },
		{ 302, 278, 24 },
		{ 302, 279, 24 },
		{ 302, 280, 24 },
		{ 302, 281, 24 },
		{ 302, 282, 24 },
		{ 302, 283, 24 },
		{ 302, 284, 24 },
		{ 302, 285, 24 },
		{ 302, 286, 24 },
		{ 302, 287, 24 },
		{ 302, 288, 24 },
		{ 302, 289, 24 },
		{ 302, 290, 24 },
		{ 303, 269, 24 },
		{ 303, 270, 24 },
		{ 303, 271, 24 },
		{ 303, 272, 24 },
		{ 303, 273, 24 },
		{ 303, 274, 24 },
		{ 303, 275, 24 },
		{ 303, 276, 24 },
		{ 303, 277, 24 },
		{ 303, 278, 24 },
		{ 303, 279, 24 },
		{ 303, 280, 24 },
		{ 303, 281, 24 },
		{ 303, 282, 24 },
		{ 303, 283, 24 },
		{ 303, 284, 24 },
		{ 303, 285, 24 },
		{ 303, 286, 24 },
		{ 303, 287, 24 },
		{ 303, 288, 24 },
		{ 303, 289, 24 },
		{ 303, 290, 24 },
		{ 303, 291, 24 },
		{ 304, 269, 24 },
		{ 304, 270, 24 },
		{ 304, 271, 24 },
		{ 304, 272, 24 },
		{ 304, 273, 24 },
		{ 304, 274, 24 },
		{ 304, 275, 24 },
		{ 304, 276, 24 },
		{ 304, 277, 24 },
		{ 304, 278, 24 },
		{ 304, 279, 24 },
		{ 304, 280, 24 },
		{ 304, 281, 24 },
		{ 304, 282, 24 },
		{ 304, 283, 24 },
		{ 304, 284, 24 },
		{ 304, 285, 24 },
		{ 304, 286, 24 },
		{ 304, 287, 24 },
		{ 304, 288, 24 },
		{ 304, 289, 24 },
		{ 304, 290, 24 },
		{ 304, 291, 24 },
		{ 304, 292, 24 },
		{ 305, 270, 24 },
		{ 305, 271, 24 },
		{ 305, 272, 24 },
		{ 305, 273, 24 },
		{ 305, 274, 24 },
		{ 305, 275, 24 },
		{ 305, 276, 24 },
		{ 305, 277, 24 },
		{ 305, 278, 24 },
		{ 305, 279, 24 },
		{ 305, 280, 24 },
		{ 305, 281, 24 },
		{ 305, 282, 24 },
		{ 305, 283, 24 },
		{ 305, 284, 24 },
		{ 305, 285, 24 },
		{ 305, 286, 24 },
		{ 305, 287, 24 },
		{ 305, 288, 24 },
		{ 305, 289, 24 },
		{ 305, 290, 24 },
		{ 305, 291, 24 },
		{ 305, 292, 24 },
		{ 306, 270, 24 },
		{ 306, 271, 24 },
		{ 306, 272, 24 },
		{ 306, 273, 24 },
		{ 306, 274, 24 },
		{ 306, 275, 24 },
		{ 306, 276, 24 },
		{ 306, 277, 24 },
		{ 306, 278, 24 },
		{ 306, 279, 24 },
		{ 306, 280, 24 },
		{ 306, 281, 24 },
		{ 306, 282, 24 },
		{ 306, 283, 24 },
		{ 306, 284, 24 },
		{ 306, 285, 24 },
		{ 306, 286, 24 },
		{ 306, 287, 24 },
		{ 306, 288, 24 },
		{ 306, 289, 24 },
		{ 306, 290, 24 },
		{ 306, 291, 24 },
		{ 306, 292, 24 },
		{ 306, 293, 24 },
		{ 307, 270, 24 },
		{ 307, 271, 24 },
		{ 307, 272, 24 },
		{ 307, 273, 24 },
		{ 307, 274, 24 },
		{ 307, 275, 24 },
		{ 307, 276, 24 },
		{ 307, 277, 24 },
		{ 307, 278, 24 },
		{ 307, 279, 24 },
		{ 307, 280, 24 },
		{ 307, 281, 24 },
		{ 307, 282, 24 },
		{ 307, 283, 24 },
		{ 307, 284, 24 },
		{ 307, 285, 24 },
		{ 307, 286, 24 },
		{ 307, 287, 24 },
		{ 307, 288, 24 },
		{ 307, 289, 24 },
		{ 307, 290, 24 },
		{ 307, 291, 24 },
		{ 307, 292, 24 },
		{ 307, 293, 24 },
		{ 307, 294, 24 },
		{ 308, 270, 24 },
		{ 308, 271, 24 },
		{ 308, 272, 24 },
		{ 308, 273, 24 },
		{ 308, 274, 24 },
		{ 308, 275, 24 },
		{ 308, 276, 24 },
		{ 308, 277, 24 },
		{ 308, 278, 24 },
		{ 308, 279, 24 },
		{ 308, 280, 24 },
		{ 308, 281, 24 },
		{ 308, 282, 24 },
		{ 308, 283, 24 },
		{ 308, 284, 24 },
		{ 308, 285, 24 },
		{ 308, 286, 24 },
		{ 308, 287, 24 },
		{ 308, 288, 24 },
		{ 308, 289, 24 },
		{ 308, 290, 24 },
		{ 308, 291, 24 },
		{ 308, 292, 24 },
		{ 308, 293, 24 },
		{ 308, 294, 24 },
		{ 308, 295, 24 },
		{ 309, 271, 24 },
		{ 309, 272, 24 },
		{ 309, 273, 24 },
		{ 309, 274, 24 },
		{ 309, 275, 24 },
		{ 309, 276, 24 },
		{ 309, 277, 24 },
		{ 309, 278, 24 },
		{ 309, 279, 24 },
		{ 309, 280, 24 },
		{ 309, 281, 24 },
		{ 309, 282, 24 },
		{ 309, 283, 24 },
		{ 309, 284, 24 },
		{ 309, 285, 24 },
		{ 309, 286, 24 },
		{ 309, 287, 24 },
		{ 309, 288, 24 },
		{ 309, 289, 24 },
		{ 309, 290, 24 },
		{ 309, 291, 24 },
		{ 309, 292, 24 },
		{ 309, 293, 24 },
		{ 309, 294, 24 },
		{ 309, 295, 24 },
		{ 310, 271, 24 },
		{ 310, 272, 24 },
		{ 310, 273, 24 },
		{ 310, 274, 24 },
		{ 310, 275, 24 },
		{ 310, 276, 24 },
		{ 310, 277, 24 },
		{ 310, 278, 24 },
		{ 310, 279, 24 },
		{ 310, 280, 24 },
		{ 310, 281, 24 },
		{ 310, 282, 24 },
		{ 310, 283, 24 },
		{ 310, 284, 24 },
		{ 310, 285, 24 },
		{ 310, 286, 24 },
		{ 310, 287, 24 },
		{ 310, 288, 24 },
		{ 310, 289, 24 },
		{ 310, 290, 24 },
		{ 310, 291, 24 },
		{ 310, 292, 24 },
		{ 310, 293, 24 },
		{ 310, 294, 24 },
		{ 310, 295, 24 },
		{ 310, 296, 24 },
		{ 311, 271, 24 },
		{ 311, 272, 24 },
		{ 311, 273, 24 },
		{ 311, 274, 24 },
		{ 311, 275, 24 },
		{ 311, 276, 24 },
		{ 311, 277, 24 },
		{ 311, 278, 24 },
		{ 311, 279, 24 },
		{ 311, 280, 24 },
		{ 311, 281, 24 },
		{ 311, 282, 24 },
		{ 311, 283, 24 },
		{ 311, 284, 24 },
		{ 311, 285, 24 },
		{ 311, 286, 24 },
		{ 311, 287, 24 },
		{ 311, 288, 24 },
		{ 311, 289, 24 },
		{ 311, 290, 24 },
		{ 311, 291, 24 },
		{ 311, 292, 24 },
		{ 311, 293, 24 },
		{ 311, 294, 24 },
		{ 311, 295, 24 },
		{ 311, 296, 24 },
		{ 311, 297, 24 },
		{ 312, 271, 24 },
		{ 312, 272, 24 },
		{ 312, 273, 24 },
		{ 312, 274, 24 },
		{ 312, 275, 24 },
		{ 312, 276, 24 },
		{ 312, 277, 24 },
		{ 312, 278, 24 },
		{ 312, 279, 24 },
		{ 312, 280, 24 },
		{ 312, 281, 24 },
		{ 312, 282, 24 },
		{ 312, 283, 24 },
		{ 312, 284, 24 },
		{ 312, 285, 24 },
		{ 312, 286, 24 },
		{ 312, 287, 24 },
		{ 312, 288, 24 },
		{ 312, 289, 24 },
		{ 312, 290, 24 },
		{ 312, 291, 24 },
		{ 312, 292, 24 },
		{ 312, 293, 24 },
		{ 312, 294, 24 },
		{ 312, 295, 24 },
		{ 312, 296, 24 },
		{ 312, 297, 24 },
		{ 312, 298, 24 },
		{ 313, 272, 24 },
		{ 313, 273, 24 },
		{ 313, 274, 24 },
		{ 313, 275, 24 },
		{ 313, 276, 24 },
		{ 313, 277, 24 },
		{ 313, 278, 24 },
		{ 313, 279, 24 },
		{ 313, 280, 24 },
		{ 313, 281, 24 },
		{ 313, 282, 24 },
		{ 313, 283, 24 },
		{ 313, 284, 24 },
		{ 313, 285, 24 },
		{ 313, 286, 24 },
		{ 313, 287, 24 },
		{ 313, 288, 24 },
		{ 313, 289, 24 },
		{ 313, 290, 24 },
		{ 313, 291, 24 },
		{ 313, 292, 24 },
		{ 313, 293, 24 },
		{ 313, 294, 24 },
		{ 313, 295, 24 },
		{ 313, 296, 24 },
		{ 313, 297, 24 },
		{ 313, 298, 24 },
		{ 314, 272, 24 },
		{ 314, 273, 24 },
		{ 314, 274, 24 },
		{ 314, 275, 24 },
		{ 314, 276, 24 },
		{ 314, 277, 24 },
		{ 314, 278, 24 },
		{ 314, 279, 24 },
		{ 314, 280, 24 },
		{ 314, 281, 24 },
		{ 314, 282, 24 },
		{ 314, 283, 24 },
		{ 314, 284, 24 },
		{ 314, 285, 24 },
		{ 314, 286, 24 },
		{ 314, 287, 24 },
		{ 314, 288, 24 },
		{ 314, 289, 24 },
		{ 314, 290, 24 },
		{ 314, 291, 24 },
		{ 314, 292, 24 },
		{ 314, 293, 24 },
		{ 314, 294, 24 },
		{ 314, 295, 24 },
		{ 314, 296, 24 },
		{ 314, 297, 24 },
		{ 314, 298, 24 },
		{ 314, 299, 24 },
		{ 315, 272, 24 },
		{ 315, 273, 24 },
		{ 315, 274, 24 },
		{ 315, 275, 24 },
		{ 315, 276, 24 },
		{ 315, 277, 24 },
		{ 315, 278, 24 },
		{ 315, 279, 24 },
		{ 315, 280, 24 },
		{ 315, 281, 24 },
		{ 315, 282, 24 },
		{ 315, 283, 24 },
		{ 315, 284, 24 },
		{ 315, 285, 24 },
		{ 315, 286, 24 },
		{ 315, 287, 24 },
		{ 315, 288, 24 },
		{ 315, 289, 24 },
		{ 315, 290, 24 },
		{ 315, 291, 24 },
		{ 315, 292, 24 },
		{ 315, 293, 24 },
		{ 315, 294, 24 },
		{ 315, 295, 24 },
		{ 315, 296, 24 },
		{ 315, 297, 24 },
		{ 315, 298, 24 },
		{ 315, 299, 24 },
		{ 315, 300, 24 },
		{ 316, 272, 24 },
		{ 316, 273, 24 },
		{ 316, 274, 24 },
		{ 316, 275, 24 },
		{ 316, 276, 24 },
		{ 316, 277, 24 },
		{ 316, 278, 24 },
		{ 316, 279, 24 },
		{ 316, 280, 24 },
		{ 316, 281, 24 },
		{ 316, 282, 24 },
		{ 316, 283, 24 },
		{ 316, 284, 24 },
		{ 316, 285, 24 },
		{ 316, 286, 24 },
		{ 316, 287, 24 },
		{ 316, 288, 24 },
		{ 316, 289, 24 },
		{ 316, 290, 24 },
		{ 316, 291, 24 },
		{ 316, 292, 24 },
		{ 316, 293, 24 },
		{ 316, 294, 24 },
		{ 316, 295, 24 },
		{ 316, 296, 24 },
		{ 316, 297, 24 },
		{ 316, 298, 24 },
		{ 316, 299, 24 },
		{ 316, 300, 24 },
		{ 316, 301, 24 },
		{ 317, 273, 24 },
		{ 317, 274, 24 },
		{ 317, 275, 24 },
		{ 317, 276, 24 },
		{ 317, 277, 24 },
		{ 317, 278, 24 },
		{ 317, 279, 24 },
		{ 317, 280, 24 },
		{ 317, 281, 24 },
		{ 317, 282, 24 },
		{ 317, 283, 24 },
		{ 317, 284, 24 },
		{ 317, 285, 24 },
		{ 317, 286, 24 },
		{ 317, 287, 24 },
		{ 317, 288, 24 },
		{ 317, 289, 24 },
		{ 317, 290, 24 },
		{ 317, 291, 24 },
		{ 317, 292, 24 },
		{ 317, 293, 24 },
		{ 317, 294, 24 },
		{ 317, 295, 24 },
		{ 317, 296, 24 },
		{ 317, 297, 24 },
		{ 317, 298, 24 },
		{ 317, 299, 24 },
		{ 317, 300, 24 },
		{ 317, 301, 24 },
		{ 318, 273, 24 },
		{ 318, 274, 24 },
		{ 318, 275, 24 },
		{ 318, 276, 24 },
		{ 318, 277, 24 },
		{ 318, 278, 24 },
		{ 318, 279, 24 },
		{ 318, 280, 24 },
		{ 318, 281, 24 },
		{ 318, 282, 24 },
		{ 318, 283, 24 },
		{ 318, 284, 24 },
		{ 318, 285, 24 },
		{ 318, 286, 24 },
		{ 318, 287, 24 },
		{ 318, 288, 24 },
		{ 318, 289, 24 },
		{ 318, 290, 24 },
		{ 318, 291, 24 },
		{ 318, 292, 24 },
		{ 318, 293, 24 },
		{ 318, 294, 24 },
		{ 318, 295, 24 },
		{ 318, 296, 24 },
		{ 318, 297, 24 },
		{ 318, 298, 24 },
		{ 318, 299, 24 },
		{ 318, 300, 24 },
		{ 318, 301, 24 },
		{ 318, 302, 24 },
		{ 319, 273, 24 },
		{ 319, 274, 24 },
		{ 319, 275, 24 },
		{ 319, 276, 24 },
		{ 319, 277, 24 },
		{ 319, 278, 24 },
		{ 319, 279, 24 },
		{ 319, 280, 24 },
		{ 319, 281, 24 },
		{ 319, 282, 24 },
		{ 319, 283, 24 },
		{ 319, 284, 24 },
		{ 319, 285, 24 },
		{ 319, 286, 24 },
		{ 319, 287, 24 },
		{ 319, 288, 24 },
		{ 319, 289, 24 },
		{ 319, 290, 24 },
		{ 319, 291, 24 },
		{ 319, 292, 24 },
		{ 319, 293, 24 },
		{ 319, 294, 24 },
		{ 319, 295, 24 },
		{ 319, 296, 24 },
		{ 319, 297, 24 },
		{ 319, 298, 24 },
		{ 319, 299, 24 },
		{ 319, 300, 24 },
		{ 319, 301, 24 },
		{ 319, 302, 24 },
		{ 319, 303, 24 },
		{ 320, 273, 24 },
		{ 320, 274, 24 },
		{ 320, 275, 24 },
		{ 320, 276, 24 },
		{ 320, 277, 24 },
		{ 320, 278, 24 },
		{ 320, 279, 24 },
		{ 320, 280, 24 },
		{ 320, 281, 24 },
		{ 320, 282, 24 },
		{ 320, 283, 24 },
		{ 320, 284, 24 },
		{ 320, 285, 24 },
		{ 320, 286, 24 },
		{ 320, 287, 24 },
		{ 320, 288, 24 },
		{ 320, 289, 24 },
		{ 320, 290, 24 },
		{ 320, 291, 24 },
		{ 320, 292, 24 },
		{ 320, 293, 24 },
		{ 320, 294, 24 },
		{ 320, 295, 24 },
		{ 320, 296, 24 },
		{ 320, 297, 24 },
		{ 320, 298, 24 },
		{ 320, 299, 24 },
		{ 320, 300, 24 },
		{ 320, 301, 24 },
		{ 320, 302, 24 },
		{ 320, 303, 24 },
		{ 320, 304, 24 },
		{ 321, 274, 24 },
		{ 321, 275, 24 },
		{ 321, 276, 24 },
		{ 321, 277, 24 },
		{ 321, 278, 24 },
		{ 321, 279, 24 },
		{ 321, 280, 24 },
		{ 321, 281, 24 },
		{ 321, 282, 24 },
		{ 321, 283, 24 },
		{ 321, 284, 24 },
		{ 321, 285, 24 },
		{ 321, 286, 24 },
		{ 321, 287, 24 },
		{ 321, 288, 24 },
		{ 321, 289, 24 },
		{ 321, 290, 24 },
		{ 321, 291, 24 },
		{ 321, 292, 24 },
		{ 321, 293, 24 },
		{ 321, 294, 24 },
		{ 321, 295, 24 },
		{ 321, 296, 24 },
		{ 321, 297, 24 },
		{ 321, 298, 24 },
		{ 321, 299, 24 },
		{ 321, 300, 24 },
		{ 321, 301, 24 },
		{ 321, 302, 24 },
		{ 321, 303, 24 },
		{ 321, 304, 24 },
		{ 322, 274, 24 },
		{ 322, 275, 24 },
		{ 322, 276, 24 },
		{ 322, 277, 24 },
		{ 322, 278, 24 },
		{ 322, 279, 24 },
		{ 322, 280, 24 },
		{ 322, 281, 24 },
		{ 322, 282, 24 },
		{ 322, 283, 24 },
		{ 322, 284, 24 },
		{ 322, 285, 24 },
		{ 322, 286, 24 },
		{ 322, 287, 24 },
		{ 322, 288, 24 },
		{ 322, 289, 24 },
		{ 322, 290, 24 },
		{ 322, 291, 24 },
		{ 322, 292, 24 },
		{ 322, 293, 24 },
		{ 322, 294, 24 },
		{ 322, 295, 24 },
		{ 322, 296, 24 },
		{ 322, 297, 24 },
		{ 322, 298, 24 },
		{ 322, 299, 24 },
		{ 322, 300, 24 },
		{ 322, 301, 24 },
		{ 322, 302, 24 },
		{ 322, 303, 24 },
		{ 322, 304, 24 },
		{ 322, 305, 24 },
		{ 323, 274, 24 },
		{ 323, 275, 24 },
		{ 323, 276, 24 },
		{ 323, 277, 24 },
		{ 323, 278, 24 },
		{ 323, 279, 24 },
		{ 323, 280, 24 },
		{ 323, 281, 24 },
		{ 323, 282, 24 },
		{ 323, 283, 24 },
		{ 323, 284, 24 },
		{ 323, 285, 24 },
		{ 323, 286, 24 },
		{ 323, 287, 24 },
		{ 323, 288, 24 },
		{ 323, 289, 24 },
		{ 323, 290, 24 },
		{ 323, 291, 24 },
		{ 323, 292, 24 },
		{ 323, 293, 24 },
		{ 323, 294, 24 },
		{ 323, 295, 24 },
		{ 323, 296, 24 },
		{ 323, 297, 24 },
		{ 323, 298, 24 },
		{ 323, 299, 24 },
		{ 323, 300, 24 },
		{ 323, 301, 24 },
		{ 323, 302, 24 },
		{ 323, 303, 24 },
		{ 323, 304, 24 },
		{ 323, 305, 24 },
		{ 323, 306, 24 },
		{ 324, 274, 24 },
		{ 324, 275, 24 },
		{ 324, 276, 24 },
		{ 324, 277, 24 },
		{ 324, 278, 24 },
		{ 324, 279, 24 },
		{ 324, 280, 24 },
		{ 324, 281, 24 },
		{ 324, 282, 24 },
		{ 324, 283, 24 },
		{ 324, 284, 24 },
		{ 324, 285, 24 },
		{ 324, 286, 24 },
		{ 324, 287, 24 },
		{ 324, 288, 24 },
		{ 324, 289, 24 },
		{ 324, 290, 24 },
		{ 324, 291, 24 },
		{ 324, 292, 24 },
		{ 324, 293, 24 },
		{ 324, 294, 24 },
		{ 324, 295, 24 },
		{ 324, 296, 24 },
		{ 324, 297, 24 },
		{ 324, 298, 24 },
		{ 324, 299, 24 },
		{ 324, 300, 24 },
		{ 324, 301, 24 },
		{ 324, 302, 24 },
		{ 324, 303, 24 },
		{ 324, 304, 24 },
		{ 324, 305, 24 },
		{ 324, 306, 24 },
		{ 324, 307, 24 },
		{ 325, 275, 24 },
		{ 325, 276, 24 },
		{ 325, 277, 24 },
		{ 325, 278, 24 },
		{ 325, 279, 24 },
		{ 325, 280, 24 },
		{ 325, 281, 24 },
		{ 325, 282, 24 },
		{ 325, 283, 24 },
		{ 325, 284, 24 },
		{ 325, 285, 24 },
		{ 325, 286, 24 },
		{ 325, 287, 24 },
		{ 325, 288, 24 },
		{ 325, 289, 24 },
		{ 325, 290, 24 },
		{ 325, 291, 24 },
		{ 325, 292, 24 },
		{ 325, 293, 24 },
		{ 325, 294, 24 },
		{ 325, 295, 24 },
		{ 325, 296, 24 },
		{ 325, 297, 24 },
		{ 325, 298, 24 },
		{ 325, 299, 24 },
		{ 325, 300, 24 },
		{ 325, 301, 24 },
		{ 325, 302, 24 },
		{ 325, 303, 24 },
		{ 325, 304, 24 },
		{ 325, 305, 24 },
		{ 325, 306, 24 },
		{ 325, 307, 24 },
		{ 326, 275, 24 },
		{ 326, 276, 24 },
		{ 326, 277, 24 },
		{ 326, 278, 24 },
		{ 326, 279, 24 },
		{ 326, 280, 24 },
		{ 326, 281, 24 },
		{ 326, 282, 24 },
		{ 326, 283, 24 },
		{ 326, 284, 24 },
		{ 326, 285, 24 },
		{ 326, 286, 24 },
		{ 326, 287, 24 },
		{ 326, 288, 24 },
		{ 326, 289, 24 },
		{ 326, 290, 24 },
		{ 326, 291, 24 },
		{ 326, 292, 24 },
		{ 326, 293, 24 },
		{ 326, 294, 24 },
		{ 326, 295, 24 },
		{ 326, 296, 24 },
		{ 326, 297, 24 },
		{ 326, 298, 24 },
		{ 326, 299, 24 },
		{ 326, 300, 24 },
		{ 326, 301, 24 },
		{ 326, 302, 24 },
		{ 326, 303, 24 },
		{ 326, 304, 24 },
		{ 326, 305, 24 },
		{ 326, 306, 24 },
		{ 326, 307, 24 },
		{ 326, 308, 24 },
		{ 327, 275, 24 },
		{ 327, 276, 24 },
		{ 327, 277, 24 },
		{ 327, 278, 24 },
		{ 327, 279, 24 },
		{ 327, 280, 24 },
		{ 327, 281, 24 },
		{ 327, 282, 24 },
		{ 327, 283, 24 },
		{ 327, 284, 24 },
		{ 327, 285, 24 },
		{ 327, 286, 24 },
		{ 327, 287, 24 },
		{ 327, 288, 24 },
		{ 327, 289, 24 },
		{ 327, 290, 24 },
		{ 327, 291, 24 },
		{ 327, 292, 24 },
		{ 327, 293, 24 },
		{ 327, 294, 24 },
		{ 327, 295, 24 },
		{ 327, 296, 24 },
		{ 327, 297, 24 },
		{ 327, 298, 24 },
		{ 327, 299, 24 },
		{ 327, 300, 24 },
		{ 327, 301, 24 },
		{ 327, 302, 24 },
		{ 327, 303, 24 },
		{ 327, 304, 24 },
		{ 327, 305, 24 },
		{ 327, 306, 24 },
		{ 327, 307, 24 },
		{ 327, 308, 24 },
		{ 327, 309, 24 },
		{ 328, 275, 24 },
		{ 328, 276, 24 },
		{ 328, 277, 24 },
		{ 328, 278, 24 },
		{ 328, 279, 24 },
		{ 328, 280, 24 },
		{ 328, 281, 24 },
		{ 328, 282, 24 },
		{ 328, 283, 24 },
		{ 328, 284, 24 },
		{ 328, 285, 24 },
		{ 328, 286, 24 },
		{ 328, 287, 24 },
		{ 328, 288, 24 },
		{ 328, 289, 24 },
		{ 328, 290, 24 },
		{ 328, 291, 24 },
		{ 328, 292, 24 },
		{ 328, 293, 24 },
		{ 328, 294, 24 },
		{ 328, 295, 24 },
		{ 328, 296, 24 },
		{ 328, 297, 24 },
		{ 328, 298, 24 },
		{ 328, 299, 24 },
		{ 328, 300, 24 },
		{ 328, 301, 24 },
		{ 328, 302, 24 },
		{ 328, 303, 24 },
		{ 328, 304, 24 },
		{ 328, 305, 24 },
		{ 328, 306, 24 },
		{ 328, 307, 24 },
		{ 328, 308, 24 },
		{ 328, 309, 24 },
		{ 328, 310, 24 },
		{ 329, 276, 24 },
		{ 329, 277, 24 },
		{ 329, 278, 24 },
		{ 329, 279, 24 },
		{ 329, 280, 24 },
		{ 329, 281, 24 },
		{ 329, 282, 24 },
		{ 329, 283, 24 },
		{ 329, 284, 24 },
		{ 329, 285, 24 },
		{ 329, 286, 24 },
		{ 329, 287, 24 },
		{ 329, 288, 24 },
		{ 329, 289, 24 },
		{ 329, 290, 24 },
		{ 329, 291, 24 },
		{ 329, 292, 24 },
		{ 329, 293, 24 },
		{ 329, 294, 24 },
		{ 329, 295, 24 },
		{ 329, 296, 24 },
		{ 329, 297, 24 },
		{ 329, 298, 24 },
		{ 329, 299, 24 },
		{ 329, 300, 24 },
		{ 329, 301, 24 },
		{ 329, 302, 24 },
		{ 329, 303, 24 },
		{ 329, 304, 24 },
		{ 329, 305, 24 },
		{ 329, 306, 24 },
		{ 329, 307, 24 },
		{ 329, 308, 24 },
		{ 329, 309, 24 },
		{ 329, 310, 24 },
		{ 330, 276, 24 },
		{ 330, 277, 24 },
		{ 330, 278, 24 },
		{ 330, 279, 24 },
		{ 330, 280, 24 },
		{ 330, 281, 24 },
		{ 330, 282, 24 },
		{ 330, 283, 24 },
		{ 330, 284, 24 },
		{ 330, 285, 24 },
		{ 330, 286, 24 },
		{ 330, 287, 24 },
		{ 330, 288, 24 },
		{ 330, 289, 24 },
		{ 330, 290, 24 },
		{ 330, 291, 24 },
		{ 330, 292, 24 },
		{ 330, 293, 24 },
		{ 330, 294, 24 },
		{ 330, 295, 24 },
		{ 330, 296, 24 },
		{ 330, 297, 24 },
		{ 330, 298, 24 },
		{ 330, 299, 24 },
		{ 330, 300, 24 },
		{ 330, 301, 24 },
		{ 330, 302, 24 },
		{ 330, 303, 24 },
		{ 330, 304, 24 },
		{ 330, 305, 24 },
		{ 330, 306, 24 },
		{ 330, 307, 24 },
		{ 330, 308, 24 },
		{ 330, 309, 24 },
		{ 330, 310, 24 },
		{ 330, 311, 24 },
		{ 331, 276, 24 },
		{ 331, 277, 24 },
		{ 331, 278, 24 },
		{ 331, 279, 24 },
		{ 331, 280, 24 },
		{ 331, 281, 24 },
		{ 331, 282, 24 },
		{ 331, 283, 24 },
		{ 331, 284, 24 },
		{ 331, 285, 24 },
		{ 331, 286, 24 },
		{ 331, 287, 24 },
		{ 331, 288, 24 },
		{ 331, 289, 24 },
		{ 331, 290, 24 },
		{ 331, 291, 24 },
		{ 331, 292, 24 },
		{ 331, 293, 24 },
		{ 331, 294, 24 },
		{ 331, 295, 24 },
		{ 331, 296, 24 },
		{ 331, 297, 24 },
		{ 331, 298, 24 },
		{ 331, 299, 24 },
		{ 331, 300, 24 },
		{ 331, 301, 24 },
		{ 331, 302, 24 },
		{ 331, 303, 24 },
		{ 331, 304, 24 },
		{ 331, 305, 24 },
		{ 331, 306, 24 },
		{ 331, 307, 24 },
		{ 331, 308, 24 },
		{ 331, 309, 24 },
		{ 331, 310, 24 },
		{ 331, 311, 24 },
		{ 331, 312, 24 },
		{ 332, 276, 24 },
		{ 332, 277, 24 },
		{ 332, 278, 24 },
		{ 332, 279, 24 },
		{ 332, 280, 24 },
		{ 332, 281, 24 },
		{ 332, 282, 24 },
		{ 332, 283, 24 },
		{ 332, 284, 24 },
		{ 332, 285, 24 },
		{ 332, 286, 24 },
		{ 332, 287, 24 },
		{ 332, 288, 24 },
		{ 332, 289, 24 },
		{ 332, 290, 24 },
		{ 332, 291, 24 },
		{ 332, 292, 24 },
		{ 332, 293, 24 },
		{ 332, 294, 24 },
		{ 332, 295, 24 },
		{ 332, 296, 24 },
		{ 332, 297, 24 },
		{ 332, 298, 24 },
		{ 332, 299, 24 },
		{ 332, 300, 24 },
		{ 332, 301, 24 },
		{ 332, 302, 24 },
		{ 332, 303, 24 },
		{ 332, 304, 24 },
		{ 332, 305, 24 },
		{ 332, 306, 24 },
		{ 332, 307, 24 },
		{ 332, 308, 24 },
		{ 332, 309, 24 },
		{ 332, 310, 24 },
		{ 332, 311, 24 },
		{ 332, 312, 24 },
		{ 332, 313, 24 },
		{ 333, 277, 24 },
		{ 333, 278, 24 },
		{ 333, 279, 24 },
		{ 333, 280, 24 },
		{ 333, 281, 24 },
		{ 333, 282, 24 },
		{ 333, 283, 24 },
		{ 333, 284, 24 },
		{ 333, 285, 24 },
		{ 333, 286, 24 },
		{ 333, 287, 24 },
		{ 333, 288, 24 },
		{ 333, 289, 24 },
		{ 333, 290, 24 },
		{ 333, 291, 24 },
		{ 333, 292, 24 },
		{ 333, 293, 24 },
		{ 333, 294, 24 },
		{ 333, 295, 24 },
		{ 333, 296, 24 },
		{ 333, 297, 24 },
		{ 333, 298, 24 },
		{ 333, 299, 24 },
		{ 333, 300, 24 },
		{ 333, 301, 24 },
		{ 333, 302, 24 },
		{ 333, 303, 24 },
		{ 333, 304, 24 },
		{ 333, 305, 24 },
		{ 333, 306, 24 },
		{ 333, 307, 24 },
		{ 333, 308, 24 },
		{ 333, 309, 24 },
		{ 333, 310, 24 },
		{ 333, 311, 24 },
		{ 333, 312, 24 },
		{ 333, 313, 24 },
		{ 334, 277, 24 },
		{ 334, 278, 24 },
		{ 334, 279, 24 },
		{ 334, 280, 24 },
		{ 334, 281, 24 },
		{ 334, 282, 24 },
		{ 334, 283, 24 },
		{ 334, 284, 24 },
		{ 334, 285, 24 },
		{ 334, 286, 24 },
		{ 334, 287, 24 },
		{ 334, 288, 24 },
		{ 334, 289, 24 },
		{ 334, 290, 24 },
		{ 334, 291, 24 },
		{ 334, 292, 24 },
		{ 334, 293, 24 },
		{ 334, 294, 24 },
		{ 334, 295, 24 },
		{ 334, 296, 24 },
		{ 334, 297, 24 },
		{ 334, 298, 24 },
		{ 334, 299, 24 },
		{ 334, 300, 24 },
		{ 334, 301, 24 },
		{ 334, 302, 24 },
		{ 334, 303, 24 },
		{ 334, 304, 24 },
		{ 334, 305, 24 },
		{ 334, 306, 24 },
		{ 334, 307, 24 },
		{ 334, 308, 24 },
		{ 334, 309, 24 },
		{ 334, 310, 24 },
		{ 334, 311, 24 },
		{ 334, 312, 24 },
		{ 334, 313, 24 },
		{ 334, 314, 24 },
		{ 335, 277, 24 },
		{ 335, 278, 24 },
		{ 335, 279, 24 },
		{ 335, 280, 24 },
		{ 335, 281, 24 },
		{ 335, 282, 24 },
		{ 335, 283, 24 },
		{ 335, 284, 24 },
		{ 335, 285, 24 },
		{ 335, 286, 24 },
		{ 335, 287, 24 },
		{ 335, 288, 24 },
		{ 335, 289, 24 },
		{ 335, 290, 24 },
		{ 335, 291, 24 },
		{ 335, 292, 24 },
		{ 335, 293, 24 },
		{ 335, 294, 24 },
		{ 335, 295, 24 },
		{ 335, 296, 24 },
		{ 335, 297, 24 },
		{ 335, 298, 24 },
		{ 335, 299, 24 },
		{ 335, 300, 24 },
		{ 335, 301, 24 },
		{ 335, 302, 24 },
		{ 335, 303, 24 },
		{ 335, 304, 24 },
		{ 335, 305, 24 },
		{ 335, 306, 24 },
		{ 335, 307, 24 },
		{ 335, 308, 24 },
		{ 335, 309, 24 },
		{ 335, 310, 24 },
		{ 335, 311, 24 },
		{ 335, 312, 24 },
		{ 335, 313, 24 },
		{ 335, 314, 24 },
		{ 335, 315, 24 },
		{ 336, 277, 24 },
		{ 336, 278, 24 },
		{ 336, 279, 24 },
		{ 336, 280, 24 },
		{ 336, 281, 24 },
		{ 336, 282, 24 },
		{ 336, 283, 24 },
		{ 336, 284, 24 },
		{ 336, 285, 24 },
		{ 336, 286, 24 },
		{ 336, 287, 24 },
		{ 336, 288, 24 },
		{ 336, 289, 24 },
		{ 336, 290, 24 },
		{ 336, 291, 24 },
		{ 336, 292, 24 },
		{ 336, 293, 24 },
		{ 336, 294, 24 },
		{ 336, 295, 24 },
		{ 336, 296, 24 },
		{ 336, 297, 24 },
		{ 336, 298, 24 },
		{ 336, 299, 24 },
		{ 336, 300, 24 },
		{ 336, 301, 24 },
		{ 336, 302, 24 },
		{ 336, 303, 24 },
		{ 336, 304, 24 },
		{ 336, 305, 24 },
		{ 336, 306, 24 },
		{ 336, 307, 24 },
		{ 336, 308, 24 },
		{ 336, 309, 24 },
		{ 336, 310, 24 },
		{ 336, 311, 24 },
		{ 336, 312, 24 },
		{ 336, 313, 24 },
		{ 336, 314, 24 },
		{ 336, 315, 24 },
		{ 336, 316, 24 },
		{ 337, 278, 24 },
		{ 337, 279, 24 },
		{ 337, 280, 24 },
		{ 337, 281, 24 },
		{ 337, 282, 24 },
		{ 337, 283, 24 },
		{ 337, 284, 24 },
		{ 337, 285, 24 },
		{ 337, 286, 24 },
		{ 337, 287, 24 },
		{ 337, 288, 24 },
		{ 337, 289, 24 },
		{ 337, 290, 24 },
		{ 337, 291, 24 },
		{ 337, 292, 24 },
		{ 337, 293, 24 },
		{ 337, 294, 24 },
		{ 337, 295, 24 },
		{ 337, 296, 24 },
		{ 337, 297, 24 },
		{ 337, 298, 24 },
		{ 337, 299, 24 },
		{ 337, 300, 24 },
		{ 337, 301, 24 },
		{ 337, 302, 24 },
		{ 337, 303, 24 },
		{ 337, 304, 24 },
		{ 337, 305, 24 },
		{ 337, 306, 24 },
		{ 337, 307, 24 },
		{ 337, 308, 24 },
		{ 337, 309, 24 },
		{ 337, 310, 24 },
		{ 337, 311, 24 },
		{ 337, 312, 24 },
		{ 337, 313, 24 },
		{ 337, 314, 24 },
		{ 337, 315, 24 },
		{ 337, 316, 24 },
		{ 338, 278, 24 },
		{ 338, 279, 24 },
		{ 338, 280, 24 },
		{ 338, 281, 24 },
		{ 338, 282, 24 },
		{ 338, 283, 24 },
		{ 338, 284, 24 },
		{ 338, 285, 24 },
		{ 338, 286, 24 },
		{ 338, 287, 24 },
		{ 338, 288, 24 },
		{ 338, 289, 24 },
		{ 338, 290, 24 },
		{ 338, 291, 24 },
		{ 338, 292, 24 },
		{ 338, 293, 24 },
		{ 338, 294, 24 },
		{ 338, 295, 24 },
		{ 338, 296, 24 },
		{ 338, 297, 24 },
		{ 338, 298, 24 },
		{ 338, 299, 24 },
		{ 338, 300, 24 },
		{ 338, 301, 24 },
		{ 338, 302, 24 },
		{ 338, 303, 24 },
		{ 338, 304, 24 },
		{ 338, 305, 24 },
		{ 338, 306, 24 },
		{ 338, 307, 24 },
		{ 338, 308, 24 },
		{ 338, 309, 24 },
		{ 338, 310, 24 },
		{ 338, 311, 24 },
		{ 338, 312, 24 },
		{ 338, 313, 24 },
		{ 338, 314, 24 },
		{ 338, 315, 24 },
		{ 338, 316, 24 },
		{ 338, 317, 24 },
		{ 339, 278, 24 },
		{ 339, 279, 24 },
		{ 339, 280, 24 },
		{ 339, 281, 24 },
		{ 339, 282, 24 },
		{ 339, 283, 24 },
		{ 339, 284, 24 },
		{ 339, 285, 24 },
		{ 339, 286, 24 },
		{ 339, 287, 24 },
		{ 339, 288, 24 },
		{ 339, 289, 24 },
		{ 339, 290, 24 },
		{ 339, 291, 24 },
		{ 339, 292, 24 },
		{ 339, 293, 24 },
		{ 339, 294, 24 },
		{ 339, 295, 24 },
		{ 339, 296, 24 },
		{ 339, 297, 24 },
		{ 339, 298, 24 },
		{ 339, 299, 24 },
		{ 339, 300, 24 },
		{ 339, 301, 24 },
		{ 339, 302, 24 },
		{ 339, 303, 24 },
		{ 339, 304, 24 },
		{ 339, 305, 24 },
		{ 339, 306, 24 },
		{ 339, 307, 24 },
		{ 339, 308, 24 },
		{ 339, 309, 24 },
		{ 339, 310, 24 },
		{ 339, 311, 24 },
		{ 339, 312, 24 },
		{ 339, 313, 24 },
		{ 339, 314, 24 },
		{ 339, 315, 24 },
		{ 339, 316, 24 },
		{ 339, 317, 24 },
		{ 339, 318, 24 },
		{ 340, 278, 24 },
		{ 340, 279, 24 },
		{ 340, 280, 24 },
		{ 340, 281, 24 },
		{ 340, 282, 24 },
		{ 340, 283, 24 },
		{ 340, 284, 24 },
		{ 340, 285, 24 },
		{ 340, 286, 24 },
		{ 340, 287, 24 },
		{ 340, 288, 24 },
		{ 340, 289, 24 },
		{ 340, 290, 24 },
		{ 340, 291, 24 },
		{ 340, 292, 24 },
		{ 340, 293, 24 },
		{ 340, 294, 24 },
		{ 340, 295, 24 },
		{ 340, 296, 24 },
		{ 340, 297, 24 },
		{ 340, 298, 24 },
		{ 340, 299, 24 },
		{ 340, 300, 24 },
		{ 340, 301, 24 },
		{ 340, 302, 24 },
		{ 340, 303, 24 },
		{ 340, 304, 24 },
		{ 340, 305, 24 },
		{ 340, 306, 24 },
		{ 340, 307, 24 },
		{ 340, 308, 24 },
		{ 340, 309, 24 },
		{ 340, 310, 24 },
		{ 340, 311, 24 },
		{ 340, 312, 24 },
		{ 340, 313, 24 },
		{ 340, 314, 24 },
		{ 340, 315, 24 },
		{ 340, 316, 24 },
		{ 340, 317, 24 },
		{ 340, 318, 24 },
		{ 340, 319, 24 },
		{ 341, 279, 24 },
		{ 341, 280, 24 },
		{ 341, 281, 24 },
		{ 341, 282, 24 },
		{ 341, 283, 24 },
		{ 341, 284, 24 },
		{ 341, 285, 24 },
		{ 341, 286, 24 },
		{ 341, 287, 24 },
		{ 341, 288, 24 },
		{ 341, 289, 24 },
		{ 341, 290, 24 },
		{ 341, 291, 24 },
		{ 341, 292, 24 },
		{ 341, 293, 24 },
		{ 341, 294, 24 },
		{ 341, 295, 24 },
		{ 341, 296, 24 },
		{ 341, 297, 24 },
		{ 341, 298, 24 },
		{ 341, 299, 24 },
		{ 341, 300, 24 },
		{ 341, 301, 24 },
		{ 341, 302, 24 },
		{ 341, 303, 24 },
		{ 341, 304, 24 },
		{ 341, 305, 24 },
		{ 341, 306, 24 },
		{ 341, 307, 24 },
		{ 341, 308, 24 },
		{ 341, 309, 24 },
		{ 341, 310, 24 },
		{ 341, 311, 24 },
		{ 341, 312, 24 },
		{ 341, 313, 24 },
		{ 341, 314, 24 },
		{ 341, 315, 24 },
		{ 341, 316, 24 },
		{ 341, 317, 24 },
		{ 341, 318, 24 },
		{ 341, 319, 24 },
		{ 342, 279, 24 },
		{ 342, 280, 24 },
		{ 342, 281, 24 },
		{ 342, 282, 24 },
		{ 342, 283, 24 },
		{ 342, 284, 24 },
		{ 342, 285, 24 },
		{ 342, 286, 24 },
		{ 342, 287, 24 },
		{ 342, 288, 24 },
		{ 342, 289, 24 },
		{ 342, 290, 24 },
		{ 342, 291, 24 },
		{ 342, 292, 24 },
		{ 342, 293, 24 },
		{ 342, 294, 24 },
		{ 342, 295, 24 },
		{ 342, 296, 24 },
		{ 342, 297, 24 },
		{ 342, 298, 24 },
		{ 342, 299, 24 },
		{ 342, 300, 24 },
		{ 342, 301, 24 },
		{ 342, 302, 24 },
		{ 342, 303, 24 },
		{ 342, 304, 24 },
		{ 342, 305, 24 },
		{ 342, 306, 24 },
		{ 342, 307, 24 },
		{ 342, 308, 24 },
		{ 342, 309, 24 },
		{ 342, 310, 24 },
		{ 342, 311, 24 },
		{ 342, 312, 24 },
		{ 342, 313, 24 },
		{ 342, 314, 24 },
		{ 342, 315, 24 },
		{ 342, 316, 24 },
		{ 342, 317, 24 },
		{ 342, 318, 24 },
		{ 342, 319, 24 },
		{ 342, 320, 24 },
		{ 343, 279, 24 },
		{ 343, 280, 24 },
		{ 343, 281, 24 },
		{ 343, 282, 24 },
		{ 343, 283, 24 },
		{ 343, 284, 24 },
		{ 343, 285, 24 },
		{ 343, 286, 24 },
		{ 343, 287, 24 },
		{ 343, 288, 24 },
		{ 343, 289, 24 },
		{ 343, 290, 24 },
		{ 343, 291, 24 },
		{ 343, 292, 24 },
		{ 343, 293, 24 },
		{ 343, 294, 24 },
		{ 343, 295, 24 },
		{ 343, 296, 24 },
		{ 343, 297, 24 },
		{ 343, 298, 24 },
		{ 343, 299, 24 },
		{ 343, 300, 24 },
		{ 343, 301, 24 },
		{ 343, 302, 24 },
		{ 343, 303, 24 },
		{ 343, 304, 24 },
		{ 343, 305, 24 },
		{ 343, 306, 24 },
		{ 343, 307, 24 },
		{ 343, 308, 24 },
		{ 343, 309, 24 },
		{ 343, 310, 24 },
		{ 343, 311, 24 },
		{ 343, 312, 24 },
		{ 343, 313, 24 },
		{ 343, 314, 24 },
		{ 343, 315, 24 },
		{ 343, 316, 24 },
		{ 343, 317, 24 },
		{ 343, 318, 24 },
		{ 343, 319, 24 },
		{ 343, 320, 24 },
		{ 343, 321, 24 },
		{ 344, 279, 24 },
		{ 344, 280, 24 },
		{ 344, 281, 24 },
		{ 344, 282, 24 },
		{ 344, 283, 24 },
		{ 344, 284, 24 },
		{ 344, 285, 24 },
		{ 344, 286, 24 },
		{ 344, 287, 24 },
		{ 344, 288, 24 },
		{ 344, 289, 24 },
		{ 344, 290, 24 },
		{ 344, 291, 24 },
		{ 344, 292, 24 },
		{ 344, 293, 24 },
		{ 344, 294, 24 },
		{ 344, 295, 24 },
		{ 344, 296, 24 },
		{ 344, 297, 24 },
		{ 344, 298, 24 },
		{ 344, 299, 24 },
		{ 344, 300, 24 },
		{ 344, 301, 24 },
		{ 344, 302, 24 },
		{ 344, 303, 24 },
		{ 344, 304, 24 },
		{ 344, 305, 24 },
		{ 344, 306, 24 },
		{ 344, 307, 24 },
		{ 344, 308, 24 },
		{ 344, 309, 24 },
		{ 344, 310, 24 },
		{ 344, 311, 24 },
		{ 344, 312, 24 },
		{ 344, 313, 24 },
		{ 344, 314, 24 },
		{ 344, 315, 24 },
		{ 344, 316, 24 },
		{ 344, 317, 24 },
		{ 344, 318, 24 },
		{ 344, 319, 24 },
		{ 344, 320, 24 },
		{ 344, 321, 24 },
		{ 344, 322, 24 },
		{ 345, 280, 24 },
		{ 345, 281, 24 },
		{ 345, 282, 24 },
		{ 345, 283, 24 },
		{ 345, 284, 24 },
		{ 345, 285, 24 },
		{ 345, 286, 24 },
		{ 345, 287, 24 },
		{ 345, 288, 24 },
		{ 345, 289, 24 },
		{ 345, 290, 24 },
		{ 345, 291, 24 },
		{ 345, 292, 24 },
		{ 345, 293, 24 },
		{ 345, 294, 24 },
		{ 345, 295, 24 },
		{ 345, 296, 24 },
		{ 345, 297, 24 },
		{ 345, 298, 24 },
		{ 345, 299, 24 },
		{ 345, 300, 24 },
		{ 345, 301, 24 },
		{ 345, 302, 24 },
		{ 345, 303, 24 },
		{ 345, 304, 24 },
		{ 345, 305, 24 },
		{ 345, 306, 24 },
		{ 345, 307, 24 },
		{ 345, 308, 24 },
		{ 345, 309, 24 },
		{ 345, 310, 24 },
		{ 345, 311, 24 },
		{ 345, 312, 24 },
		{ 345, 313, 24 },
		{ 345, 314, 24 },
		{ 345, 315, 24 },
		{ 345, 316, 24 },
		{ 345, 317, 24 },
		{ 345, 318, 24 },
		{ 345, 319, 24 },
		{ 345, 320, 24 },
		{ 345, 321, 24 },
		{ 345, 322, 24 },
		{ 346, 280, 24 },
		{ 346, 281, 24 },
		{ 346, 282, 24 },
		{ 346, 283, 24 },
		{ 346, 284, 24 },
		{ 346, 285, 24 },
		{ 346, 286, 24 },
		{ 346, 287, 24 },
		{ 346, 288, 24 },
		{ 346, 289, 24 },
		{ 346, 290, 24 },
		{ 346, 291, 24 },
		{ 346, 292, 24 },
		{ 346, 293, 24 },
		{ 346, 294, 24 },
		{ 346, 295, 24 },
		{ 346, 296, 24 },
		{ 346, 297, 24 },
		{ 346, 298, 24 },
		{ 346, 299, 24 },
		{ 346, 300, 24 },
		{ 346, 301, 24 },
		{ 346, 302, 24 },
		{ 346, 303, 24 },
		{ 346, 304, 24 },
		{ 346, 305, 24 },
		{ 346, 306, 24 },
		{ 346, 307, 24 },
		{ 346, 308, 24 },
		{ 346, 309, 24 },
		{ 346, 310, 24 },
		{ 346, 311, 24 },
		{ 346, 312, 24 },
		{ 346, 313, 24 },
		{ 346, 314, 24 },
		{ 346, 315, 24 },
		{ 346, 316, 24 },
		{ 346, 317, 24 },
		{ 346, 318, 24 },
		{ 346, 319, 24 },
		{ 346, 320, 24 },
		{ 346, 321, 24 },
		{ 346, 322, 24 },
		{ 346, 323, 24 },
		{ 347, 280, 24 },
		{ 347, 281, 24 },
		{ 347, 282, 24 },
		{ 347, 283, 24 },
		{ 347, 284, 24 },
		{ 347, 285, 24 },
		{ 347, 286, 24 },
		{ 347, 287, 24 },
		{ 347, 288, 24 },
		{ 347, 289, 24 },
		{ 347, 290, 24 },
		{ 347, 291, 24 },
		{ 347, 292, 24 },
		{ 347, 293, 24 },
		{ 347, 294, 24 },
		{ 347, 295, 24 },
		{ 347, 296, 24 },
		{ 347, 297, 24 },
		{ 347, 298, 24 },
		{ 347, 299, 24 },
		{ 347, 300, 24 },
		{ 347, 301, 24 },
		{ 347, 302, 24 },
		{ 347, 303, 24 },
		{ 347, 304, 24 },
		{ 347, 305, 24 },
		{ 347, 306, 24 },
		{ 347, 307, 24 },
		{ 347, 308, 24 },
		{ 347, 309, 24 },
		{ 347, 310, 24 },
		{ 347, 311, 24 },
		{ 347, 312, 24 },
		{ 347, 313, 24 },
		{ 347, 314, 24 },
		{ 347, 315, 24 },
		{ 347, 316, 24 },
		{ 347, 317, 24 },
		{ 347, 318, 24 },
		{ 347, 319, 24 },
		{ 347, 320, 24 },
		{ 347, 321, 24 },
		{ 347, 322, 24 },
		{ 347, 323, 24 },
		{ 347, 324, 24 },
		{ 348, 280, 24 },
		{ 348, 281, 24 },
		{ 348, 282, 24 },
		{ 348, 283, 24 },
		{ 348, 284, 24 },
		{ 348, 285, 24 },
		{ 348, 286, 24 },
		{ 348, 287, 24 },
		{ 348, 288, 24 },
		{ 348, 289, 24 },
		{ 348, 290, 24 },
		{ 348, 291, 24 },
		{ 348, 292, 24 },
		{ 348, 293, 24 },
		{ 348, 294, 24 },
		{ 348, 295, 24 },
		{ 348, 296, 24 },
		{ 348, 297, 24 },
		{ 348, 298, 24 },
		{ 348, 299, 24 },
		{ 348, 300, 24 },
		{ 348, 301, 24 },
		{ 348, 302, 24 },
		{ 348, 303, 24 },
		{ 348, 304, 24 },
		{ 348, 305, 24 },
		{ 348, 306, 24 },
		{ 348, 307, 24 },
		{ 348, 308, 24 },
		{ 348, 309, 24 },
		{ 348, 310, 24 },
		{ 348, 311, 24 },
		{ 348, 312, 24 },
		{ 348, 313, 24 },
		{ 348, 314, 24 },
		{ 348, 315, 24 },
		{ 348, 316, 24 },
		{ 348, 317, 24 },
		{ 348, 318, 24 },
		{ 348, 319, 24 },
		{ 348, 320, 24 },
		{ 348, 321, 24 },
		{ 348, 322, 24 },
		{ 348, 323, 24 },
		{ 348, 324, 24 },
		{ 348, 325, 24 },
		{ 349, 281, 24 },
		{ 349, 282, 24 },
		{ 349, 283, 24 },
		{ 349, 284, 24 },
		{ 349, 285, 24 },
		{ 349, 286, 24 },
		{ 349, 287, 24 },
		{ 349, 288, 24 },
		{ 349, 289, 24 },
		{ 349, 290, 24 },
		{ 349, 291, 24 },
		{ 349, 292, 24 },
		{ 349, 293, 24 },
		{ 349, 294, 24 },
		{ 349, 295, 24 },
		{ 349, 296, 24 },
		{ 349, 297, 24 },
		{ 349, 298, 24 },
		{ 349, 299, 24 },
		{ 349, 300, 24 },
		{ 349, 301, 24 },
		{ 349, 302, 24 },
		{ 349, 303, 24 },
		{ 349, 304, 24 },
		{ 349, 305, 24 },
		{ 349, 306, 24 },
		{ 349, 307, 24 },
		{ 349, 308, 24 },
		{ 349, 309, 24 },
		{ 349, 310, 24 },
		{ 349, 311, 24 },
		{ 349, 312, 24 },
		{ 349, 313, 24 },
		{ 349, 314, 24 },
		{ 349, 315, 24 },
		{ 349, 316, 24 },
		{ 349, 317, 24 },
		{ 349, 318, 24 },
		{ 349, 319, 24 },
		{ 349, 320, 24 },
		{ 349, 321, 24 },
		{ 349, 322, 24 },
		{ 349, 323, 24 },
		{ 349, 324, 24 },
		{ 349, 325, 24 },
		{ 350, 281, 24 },
		{ 350, 282, 24 },
		{ 350, 283, 24 },
		{ 350, 284, 24 },
		{ 350, 285, 24 },
		{ 350, 286, 24 },
		{ 350, 287, 24 },
		{ 350, 288, 24 },
		{ 350, 289, 24 },
		{ 350, 290, 24 },
		{ 350, 291, 24 },
		{ 350, 292, 24 },
		{ 350, 293, 24 },
		{ 350, 294, 24 },
		{ 350, 295, 24 },
		{ 350, 296, 24 },
		{ 350, 297, 24 },
		{ 350, 298, 24 },
		{ 350, 299, 24 },
		{ 350, 300, 24 },
		{ 350, 301, 24 },
		{ 350, 302, 24 },
		{ 350, 303, 24 },
		{ 350, 304, 24 },
		{ 350, 305, 24 },
		{ 350, 306, 24 },
		{ 350, 307, 24 },
		{ 350, 308, 24 },
		{ 350, 309, 24 },
		{ 350, 310, 24 },
		{ 350, 311, 24 },
		{ 350, 312, 24 },
		{ 350, 313, 24 },
		{ 350, 314, 24 },
		{ 350, 315, 24 },
		{ 350, 316, 24 },
		{ 350, 317, 24 },
		{ 350, 318, 24 },
		{ 350, 319, 24 },
		{ 350, 320, 24 },
		{ 350, 321, 24 },
		{ 350, 322, 24 },
		{ 350, 323, 24 },
		{ 350, 324, 24 },
		{ 350, 325, 24 },
		{ 350, 326, 24 },
		{ 351, 281, 24 },
		{ 351, 282, 24 },
		{ 351, 283, 24 },
		{ 351, 284, 24 },
		{ 351, 285, 24 },
		{ 351, 286, 24 },
		{ 351, 287, 24 },
		{ 351, 288, 24 },
		{ 351, 289, 24 },
		{ 351, 290, 24 },
		{ 351, 291, 24 },
		{ 351, 292, 24 },
		{ 351, 293, 24 },
		{ 351, 294, 24 },
		{ 351, 295, 24 },
		{ 351, 296, 24 },
		{ 351, 297, 24 },
		{ 351, 298, 24 },
		{ 351, 299, 24 },
		{ 351, 300, 24 },
		{ 351, 301, 24 },
		{ 351, 302, 24 },
		{ 351, 303, 24 },
		{ 351, 304, 24 },
		{ 351, 305, 24 },
		{ 351, 306, 24 },
		{ 351, 307, 24 },
		{ 351, 308, 24 },
		{ 351, 309, 24 },
		{ 351, 310, 24 },
		{ 351, 311, 24 },
		{ 351, 312, 24 },
		{ 351, 313, 24 },
		{ 351, 314, 24 },
		{ 351, 315, 24 },
		{ 351, 316, 24 },
		{ 351, 317, 24 },
		{ 351, 318, 24 },
		{ 351, 319, 24 },
		{ 351, 320, 24 },
		{ 351, 321, 24 },
		{ 351, 322, 24 },
		{ 351, 323, 24 },
		{ 351, 324, 24 },
		{ 351, 325, 24 },
		{ 351, 326, 24 },
		{ 351, 327, 24 },
		{ 352, 281, 24 },
		{ 352, 282, 24 },
		{ 352, 283, 24 },
		{ 352, 284, 24 },
		{ 352, 285, 24 },
		{ 352, 286, 24 },
		{ 352, 287, 24 },
		{ 352, 288, 24 },
		{ 352, 289, 24 },
		{ 352, 290, 24 },
		{ 352, 291, 24 },
		{ 352, 292, 24 },
		{ 352, 293, 24 },
		{ 352, 294, 24 },
		{ 352, 295, 24 },
		{ 352, 296, 24 },
		{ 352, 297, 24 },
		{ 352, 298, 24 },
		{ 352, 299, 24 },
		{ 352, 300, 24 },
		{ 352, 301, 24 },
		{ 352, 302, 24 },
		{ 352, 303, 24 },
		{ 352, 304, 24 },
		{ 352, 305, 24 },
		{ 352, 306, 24 },
		{ 352, 307, 24 },
		{ 352, 308, 24 },
		{ 352, 309, 24 },
		{ 352, 310, 24 },
		{ 352, 311, 24 },
		{ 352, 312, 24 },
		{ 352, 313, 24 },
		{ 352, 314, 24 },
		{ 352, 315, 24 },
		{ 352, 316, 24 },
		{ 352, 317, 24 },
		{ 352, 318, 24 },
		{ 352, 319, 24 },
		{ 352, 320, 24 },
		{ 352, 321, 24 },
		{ 352, 322, 24 },
		{ 352, 323, 24 },
		{ 352, 324, 24 },
		{ 352, 325, 24 },
		{ 352, 326, 24 },
		{ 352, 327, 24 },
		{ 352, 328, 24 },
		{ 353, 282, 24 },
		{ 353, 283, 24 },
		{ 353, 284, 24 },
		{ 353, 285, 24 },
		{ 353, 286, 24 },
		{ 353, 287, 24 },
		{ 353, 288, 24 },
		{ 353, 289, 24 },
		{ 353, 290, 24 },
		{ 353, 291, 24 },
		{ 353, 292, 24 },
		{ 353, 293, 24 },
		{ 353, 294, 24 },
		{ 353, 295, 24 },
		{ 353, 296, 24 },
		{ 353, 297, 24 },
		{ 353, 298, 24 },
		{ 353, 299, 24 },
		{ 353, 300, 24 },
		{ 353, 301, 24 },
		{ 353, 302, 24 },
		{ 353, 303, 24 },
		{ 353, 304, 24 },
		{ 353, 305, 24 },
		{ 353, 306, 24 },
		{ 353, 307, 24 },
		{ 353, 308, 24 },
		{ 353, 309, 24 },
		{ 353, 310, 24 },
		{ 353, 311, 24 },
		{ 353, 312, 24 },
		{ 353, 313, 24 },
		{ 353, 314, 24 },
		{ 353, 315, 24 },
		{ 353, 316, 24 },
		{ 353, 317, 24 },
		{ 353, 318, 24 },
		{ 353, 319, 24 },
		{ 353, 320, 24 },
		{ 353, 321, 24 },
		{ 353, 322, 24 },
		{ 353, 323, 24 },
		{ 353, 324, 24 },
		{ 353, 325, 24 },
		{ 353, 326, 24 },
		{ 353, 327, 24 },
		{ 353, 328, 24 },
		{ 354, 282, 24 },
		{ 354, 283, 24 },
		{ 354, 284, 24 },
		{ 354, 285, 24 },
		{ 354, 286, 24 },
		{ 354, 287, 24 },
		{ 354, 288, 24 },
		{ 354, 289, 24 },
		{ 354, 290, 24 },
		{ 354, 291, 24 },
		{ 354, 292, 24 },
		{ 354, 293, 24 },
		{ 354, 294, 24 },
		{ 354, 295, 24 },
		{ 354, 296, 24 },
		{ 354, 297, 24 },
		{ 354, 298, 24 },
		{ 354, 299, 24 },
		{ 354, 300, 24 },
		{ 354, 301, 24 },
		{ 354, 302, 24 },
		{ 354, 303, 24 },
		{ 354, 304, 24 },
		{ 354, 305, 24 },
		{ 354, 306, 24 },
		{ 354, 307, 24 },
		{ 354, 308, 24 },
		{ 354, 309, 24 },
		{ 354, 310, 24 },
		{ 354, 311, 24 },
		{ 354, 312, 24 },
		{ 354, 313, 24 },
		{ 354, 314, 24 },
		{ 354, 315, 24 },
		{ 354, 316, 24 },
		{ 354, 317, 24 },
		{ 354, 318, 24 },
		{ 354, 319, 24 },
		{ 354, 320, 24 },
		{ 354, 321, 24 },
		{ 354, 322, 24 },
		{ 354, 323, 24 },
		{ 354, 324, 24 },
		{ 354, 325, 24 },
		{ 354, 326, 24 },
		{ 354, 327, 24 },
		{ 354, 328, 24 },
		{ 354, 329, 24 },
		{ 355, 282, 24 },
		{ 355, 283, 24 },
		{ 355, 284, 24 },
		{ 355, 285, 24 },
		{ 355, 286, 24 },
		{ 355, 287, 24 },
		{ 355, 288, 24 },
		{ 355, 289, 24 },
		{ 355, 290, 24 },
		{ 355, 291, 24 },
		{ 355, 292, 24 },
		{ 355, 293, 24 },
		{ 355, 294, 24 },
		{ 355, 295, 24 },
		{ 355, 296, 24 },
		{ 355, 297, 24 },
		{ 355, 298, 24 },
		{ 355, 299, 24 },
		{ 355, 300, 24 },
		{ 355, 301, 24 },
		{ 355, 302, 24 },
		{ 355, 303, 24 },
		{ 355, 304, 24 },
		{ 355, 305, 24 },
		{ 355, 306, 24 },
		{ 355, 307, 24 },
		{ 355, 308, 24 },
		{ 355, 309, 24 },
		{ 355, 310, 24 },
		{ 355, 311, 24 },
		{ 355, 312, 24 },
		{ 355, 313, 24 },
		{ 355, 314, 24 },
		{ 355, 315, 24 },
		{ 355, 316, 24 },
		{ 355, 317, 24 },
		{ 355, 318, 24 },
		{ 355, 319, 24 },
		{ 355, 320, 24 },
		{ 355, 321, 24 },
		{ 355, 322, 24 },
		{ 355, 323, 24 },
		{ 355, 324, 24 },
		{ 355, 325, 24 },
		{ 355, 326, 24 },
		{ 355, 327, 24 },
		{ 355, 328, 24 },
		{ 355, 329, 24 },
		{ 355, 330, 24 },
		{ 356, 282, 24 },
		{ 356, 283, 24 },
		{ 356, 284, 24 },
		{ 356, 285, 24 },
		{ 356, 286, 24 },
		{ 356, 287, 24 },
		{ 356, 288, 24 },
		{ 356, 289, 24 },
		{ 356, 290, 24 },
		{ 356, 291, 24 },
		{ 356, 292, 24 },
		{ 356, 293, 24 },
		{ 356, 294, 24 },
		{ 356, 295, 24 },
		{ 356, 296, 24 },
		{ 356, 297, 24 },
		{ 356, 298, 24 },
		{ 356, 299, 24 },
		{ 356, 300, 24 },
		{ 356, 301, 24 },
		{ 356, 302, 24 },
		{ 356, 303, 24 },
		{ 356, 304, 24 },
		{ 356, 305, 24 },
		{ 356, 306, 24 },
		{ 356, 307, 24 },
		{ 356, 308, 24 },
		{ 356, 309, 24 },
		{ 356, 310, 24 },
		{ 356, 311, 24 },
		{ 356, 312, 24 },
		{ 356, 313, 24 },
		{ 356, 314, 24 },
		{ 356, 315, 24 },
		{ 356, 316, 24 },
		{ 356, 317, 24 },
		{ 356, 318, 24 },
		{ 356, 319, 24 },
		{ 356, 320, 24 },
		{ 356, 321, 24 },
		{ 356, 322, 24 },
		{ 356, 323, 24 },
		{ 356, 324, 24 },
		{ 356, 325, 24 },
		{ 356, 326, 24 },
		{ 356, 327, 24 },
		{ 356, 328, 24 },
		{ 356, 329, 24 },
		{ 356, 330, 24 },
		{ 356, 331, 24 },
		{ 357, 283, 24 },
		{ 357, 284, 24 },
		{ 357, 285, 24 },
		{ 357, 286, 24 },
		{ 357, 287, 24 },
		{ 357, 288, 24 },
		{ 357, 289, 24 },
		{ 357, 290, 24 },
		{ 357, 291, 24 },
		{ 357, 292, 24 },
		{ 357, 293, 24 },
		{ 357, 294, 24 },
		{ 357, 295, 24 },
		{ 357, 296, 24 },
		{ 357, 297, 24 },
		{ 357, 298, 24 },
		{ 357, 299, 24 },
		{ 357, 300, 24 },
		{ 357, 301, 24 },
		{ 357, 302, 24 },
		{ 357, 303, 24 },
		{ 357, 304, 24 },
		{ 357, 305, 24 },
		{ 357, 306, 24 },
		{ 357, 307, 24 },
		{ 357, 308, 24 },
		{ 357, 309, 24 },
		{ 357, 310, 24 },
		{ 357, 311, 24 },
		{ 357, 312, 24 },
		{ 357, 313, 24 },
		{ 357, 314, 24 },
		{ 357, 315, 24 },
		{ 357, 316, 24 },
		{ 357, 317, 24 },
		{ 357, 318, 24 },
		{ 357, 319, 24 },
		{ 357, 320, 24 },
		{ 357, 321, 24 },
		{ 357, 322, 24 },
		{ 357, 323, 24 },
		{ 357, 324, 24 },
		{ 357, 325, 24 },
		{ 357, 326, 24 },
		{ 357, 327, 24 },
		{ 357, 328, 24 },
		{ 357, 329, 24 },
		{ 357, 330, 24 },
		{ 357, 331, 24 },
		{ 358, 283, 24 },
		{ 358, 284, 24 },
		{ 358, 285, 24 },
		{ 358, 286, 24 },
		{ 358, 287, 24 },
		{ 358, 288, 24 },
		{ 358, 289, 24 },
		{ 358, 290, 24 },
		{ 358, 291, 24 },
		{ 358, 292, 24 },
		{ 358, 293, 24 },
		{ 358, 294, 24 },
		{ 358, 295, 24 },
		{ 358, 296, 24 },
		{ 358, 297, 24 },
		{ 358, 298, 24 },
		{ 358, 299, 24 },
		{ 358, 300, 24 },
		{ 358, 301, 24 },
		{ 358, 302, 24 },
		{ 358, 303, 24 },
		{ 358, 304, 24 },
		{ 358, 305, 24 },
		{ 358, 306, 24 },
		{ 358, 307, 24 },
		{ 358, 308, 24 },
		{ 358, 309, 24 },
		{ 358, 310, 24 },
		{ 358, 311, 24 },
		{ 358, 312, 24 },
		{ 358, 313, 24 },
		{ 358, 314, 24 },
		{ 358, 315, 24 },
		{ 358, 316, 24 },
		{ 358, 317, 24 },
		{ 358, 318, 24 },
		{ 358, 319, 24 },
		{ 358, 320, 24 },
		{ 358, 321, 24 },
		{ 358, 322, 24 },
		{ 358, 323, 24 },
		{ 358, 324, 24 },
		{ 358, 325, 24 },
		{ 358, 326, 24 },
		{ 358, 327, 24 },
		{ 358, 328, 24 },
		{ 358, 329, 24 },
		{ 358, 330, 24 },
		{ 358, 331, 24 },
		{ 358, 332, 24 },
		{ 359, 283, 24 },
		{ 359, 284, 24 },
		{ 359, 285, 24 },
		{ 359, 286, 24 },
		{ 359, 287, 24 },
		{ 359, 288, 24 },
		{ 359, 289, 24 },
		{ 359, 290, 24 },
		{ 359, 291, 24 },
		{ 359, 292, 24 },
		{ 359, 293, 24 },
		{ 359, 294, 24 },
		{ 359, 295, 24 },
		{ 359, 296, 24 },
		{ 359, 297, 24 },
		{ 359, 298, 24 },
		{ 359, 299, 24 },
		{ 359, 300, 24 },
		{ 359, 301, 24 },
		{ 359, 302, 24 },
		{ 359, 303, 24 },
		{ 359, 304, 24 },
		{ 359, 305, 24 },
		{ 359, 306, 24 },
		{ 359, 307, 24 },
		{ 359, 308, 24 },
		{ 359, 309, 24 },
		{ 359, 310, 24 },
		{ 359, 311, 24 },
		{ 359, 312, 24 },
		{ 359, 313, 24 },
		{ 359, 314, 24 },
		{ 359, 315, 24 },
		{ 359, 316, 24 },
		{ 359, 317, 24 },
		{ 359, 318, 24 },
		{ 359, 319, 24 },
		{ 359, 320, 24 },
		{ 359, 321, 24 },
		{ 359, 322, 24 },
		{ 359, 323, 24 },
		{ 359, 324, 24 },
		{ 359, 325, 24 },
		{ 359, 326, 24 },
		{ 359, 327, 24 },
		{ 359, 328, 24 },
		{ 359, 329, 24 },
		{ 359, 330, 24 },
		{ 359, 331, 24 },
		{ 359, 332, 24 },
		{ 359, 333, 24 },
		{ 360, 283, 24 },
		{ 360, 284, 24 },
		{ 360, 285, 24 },
		{ 360, 286, 24 },
		{ 360, 287, 24 },
		{ 360, 288, 24 },
		{ 360, 289, 24 },
		{ 360, 290, 24 },
		{ 360, 291, 24 },
		{ 360, 292, 24 },
		{ 360, 293, 24 },
		{ 360, 294, 24 },
		{ 360, 295, 24 },
		{ 360, 296, 24 },
		{ 360, 297, 24 },
		{ 360, 298, 24 },
		{ 360, 299, 24 },
		{ 360, 300, 24 },
		{ 360, 301, 24 },
		{ 360, 302, 24 },
		{ 360, 303, 24 },
		{ 360, 304, 24 },
		{ 360, 305, 24 },
		{ 360, 306, 24 },
		{ 360, 307, 24 },
		{ 360, 308, 24 },
		{ 360, 309, 24 },
		{ 360, 310, 24 },
		{ 360, 311, 24 },
		{ 360, 312, 24 },
		{ 360, 313, 24 },
		{ 360, 314, 24 },
		{ 360, 315, 24 },
		{ 360, 316, 24 },
		{ 360, 317, 24 },
		{ 360, 318, 24 },
		{ 360, 319, 24 },
		{ 360, 320, 24 },
		{ 360, 321, 24 },
		{ 360, 322, 24 },
		{ 360, 323, 24 },
		{ 360, 324, 24 },
		{ 360, 325, 24 },
		{ 360, 326, 24 },
		{ 360, 327, 24 },
		{ 360, 328, 24 },
		{ 360, 329, 24 },
		{ 360, 330, 24 },
		{ 360, 331, 24 },
		{ 360, 332, 24 },
		{ 360, 333, 24 },
		{ 360, 334, 24 },
		{ 361, 284, 24 },
		{ 361, 285, 24 },
		{ 361, 286, 24 },
		{ 361, 287, 24 },
		{ 361, 288, 24 },
		{ 361, 289, 24 },
		{ 361, 290, 24 },
		{ 361, 291, 24 },
		{ 361, 292, 24 },
		{ 361, 293, 24 },
		{ 361, 294, 24 },
		{ 361, 295, 24 },
		{ 361, 296, 24 },
		{ 361, 297, 24 },
		{ 361, 298, 24 },
		{ 361, 299, 24 },
		{ 361, 300, 24 },
		{ 361, 301, 24 },
		{ 361, 302, 24 },
		{ 361, 303, 24 },
		{ 361, 304, 24 },
		{ 361, 305, 24 },
		{ 361, 306, 24 },
		{ 361, 307, 24 },
		{ 361, 308, 24 },
		{ 361, 309, 24 },
		{ 361, 310, 24 },
		{ 361, 311, 24 },
		{ 361, 312, 24 },
		{ 361, 313, 24 },
		{ 361, 314, 24 },
		{ 361, 315, 24 },
		{ 361, 316, 24 },
		{ 361, 317, 24 },
		{ 361, 318, 24 },
		{ 361, 319, 24 },
		{ 361, 320, 24 },
		{ 361, 321, 24 },
		{ 361, 322, 24 },
		{ 361, 323, 24 },
		{ 361, 324, 24 },
		{ 361, 325, 24 },
		{ 361, 326, 24 },
		{ 361, 327, 24 },
		{ 361, 328, 24 },
		{ 361, 329, 24 },
		{ 361, 330, 24 },
		{ 361, 331, 24 },
		{ 361, 332, 24 },
		{ 361, 333, 24 },
		{ 361, 334, 24 },
		{ 362, 284, 24 },
		{ 362, 285, 24 },
		{ 362, 286, 24 },
		{ 362, 287, 24 },
		{ 362, 288, 24 },
		{ 362, 289, 24 },
		{ 362, 290, 24 },
		{ 362, 291, 24 },
		{ 362, 292, 24 },
		{ 362, 293, 24 },
		{ 362, 294, 24 },
		{ 362, 295, 24 },
		{ 362, 296, 24 },
		{ 362, 297, 24 },
		{ 362, 298, 24 },
		{ 362, 299, 24 },
		{ 362, 300, 24 },
		{ 362, 301, 24 },
		{ 362, 302, 24 },
		{ 362, 303, 24 },
		{ 362, 304, 24 },
		{ 362, 305, 24 },
		{ 362, 306, 24 },
		{ 362, 307, 24 },
		{ 362, 308, 24 },
		{ 362, 309, 24 },
		{ 362, 310, 24 },
		{ 362, 311, 24 },
		{ 362, 312, 24 },
		{ 362, 313, 24 },
		{ 362, 314, 24 },
		{ 362, 315, 24 },
		{ 362, 316, 24 },
		{ 362, 317, 24 },
		{ 362, 318, 24 },
		{ 362, 319, 24 },
		{ 362, 320, 24 },
		{ 362, 321, 24 },
		{ 362, 322, 24 },
		{ 362, 323, 24 },
		{ 362, 324, 24 },
		{ 362, 325, 24 },
		{ 362, 326, 24 },
		{ 362, 327, 24 },
		{ 362, 328, 24 },
		{ 362, 329, 24 },
		{ 362, 330, 24 },
		{ 362, 331, 24 },
		{ 362, 332, 24 },
		{ 362, 333, 24 },
		{ 362, 334, 24 },
		{ 362, 335, 24 },
		{ 363, 284, 24 },
		{ 363, 285, 24 },
		{ 363, 286, 24 },
		{ 363, 287, 24 },
		{ 363, 288, 24 },
		{ 363, 289, 24 },
		{ 363, 290, 24 },
		{ 363, 291, 24 },
		{ 363, 292, 24 },
		{ 363, 293, 24 },
		{ 363, 294, 24 },
		{ 363, 295, 24 },
		{ 363, 296, 24 },
		{ 363, 297, 24 },
		{ 363, 298, 24 },
		{ 363, 299, 24 },
		{ 363, 300, 24 },
		{ 363, 301, 24 },
		{ 363, 302, 24 },
		{ 363, 303, 24 },
		{ 363, 304, 24 },
		{ 363, 305, 24 },
		{ 363, 306, 24 },
		{ 363, 307, 24 },
		{ 363, 308, 24 },
		{ 363, 309, 24 },
		{ 363, 310, 24 },
		{ 363, 311, 24 },
		{ 363, 312, 24 },
		{ 363, 313, 24 },
		{ 363, 314, 24 },
		{ 363, 315, 24 },
		{ 363, 316, 24 },
		{ 363, 317, 24 },
		{ 363, 318, 24 },
		{ 363, 319, 24 },
		{ 363, 320, 24 },
		{ 363, 321, 24 },
		{ 363, 322, 24 },
		{ 363, 323, 24 },
		{ 363, 324, 24 },
		{ 363, 325, 24 },
		{ 363, 326, 24 },
		{ 363, 327, 24 },
		{ 363, 328, 24 },
		{ 363, 329, 24 },
		{ 363, 330, 24 },
		{ 363, 331, 24 },
		{ 363, 332, 24 },
		{ 363, 333, 24 },
		{ 363, 334, 24 },
		{ 363, 335, 24 },
		{ 363, 336, 24 },
		{ 364, 284, 24 },
		{ 364, 285, 24 },
		{ 364, 286, 24 },
		{ 364, 287, 24 },
		{ 364, 288, 24 },
		{ 364, 289, 24 },
		{ 364, 290, 24 },
		{ 364, 291, 24 },
		{ 364, 292, 24 },
		{ 364, 293, 24 },
		{ 364, 294, 24 },
		{ 364, 295, 24 },
		{ 364, 296, 24 },
		{ 364, 297, 24 },
		{ 364, 298, 24 },
		{ 364, 299, 24 },
		{ 364, 300, 24 },
		{ 364, 301, 24 },
		{ 364, 302, 24 },
		{ 364, 303, 24 },
		{ 364, 304, 24 },
		{ 364, 305, 24 },
		{ 364, 306, 24 },
		{ 364, 307, 24 },
		{ 364, 308, 24 },
		{ 364, 309, 24 },
		{ 364, 310, 24 },
		{ 364, 311, 24 },
		{ 364, 312, 24 },
		{ 364, 313, 24 },
		{ 364, 314, 24 },
		{ 364, 315, 24 },
		{ 364, 316, 24 },
		{ 364, 317, 24 },
		{ 364, 318, 24 },
		{ 364, 319, 24 },
		{ 364, 320, 24 },
		{ 364, 321, 24 },
		{ 364, 322, 24 },
		{ 364, 323, 24 },
		{ 364, 324, 24 },
		{ 364, 325, 24 },
		{ 364, 326, 24 },
		{ 364, 327, 24 },
		{ 364, 328, 24 },
		{ 364, 329, 24 },
		{ 364, 330, 24 },
		{ 364, 331, 24 },
		{ 364, 332, 24 },
		{ 364, 333, 24 },
		{ 364, 334, 24 },
		{ 364, 335, 24 },
		{ 364, 336, 24 },
		{ 364, 337, 24 },
		{ 365, 285, 24 },
		{ 365, 286, 24 },
		{ 365, 287, 24 },
		{ 365, 288, 24 },
		{ 365, 289, 24 },
		{ 365, 290, 24 },
		{ 365, 291, 24 },
		{ 365, 292, 24 },
		{ 365, 293, 24 },
		{ 365, 294, 24 },
		{ 365, 295, 24 },
		{ 365, 296, 24 },
		{ 365, 297, 24 },
		{ 365, 298, 24 },
		{ 365, 299, 24 },
		{ 365, 300, 24 },
		{ 365, 301, 24 },
		{ 365, 302, 24 },
		{ 365, 303, 24 },
		{ 365, 304, 24 },
		{ 365, 305, 24 },
		{ 365, 306, 24 },
		{ 365, 307, 24 },
		{ 365, 308, 24 },
		{ 365, 309, 24 },
		{ 365, 310, 24 },
		{ 365, 311, 24 },
		{ 365, 312, 24 },
		{ 365, 313, 24 },
		{ 365, 314, 24 },
		{ 365, 315, 24 },
		{ 365, 316, 24 },
		{ 365, 317, 24 },
		{ 365, 318, 24 },
		{ 365, 319, 24 },
		{ 365, 320, 24 },
		{ 365, 321, 24 },
		{ 365, 322, 24 },
		{ 365, 323, 24 },
		{ 365, 324, 24 },
		{ 365, 325, 24 },
		{ 365, 326, 24 },
		{ 365, 327, 24 },
		{ 365, 328, 24 },
		{ 365, 329, 24 },
		{ 365, 330, 24 },
		{ 365, 331, 24 },
		{ 365, 332, 24 },
		{ 365, 333, 24 },
		{ 365, 334, 24 },
		{ 365, 335, 24 },
		{ 365, 336, 24 },
		{ 365, 337, 24 },
		{ 366, 285, 24 },
		{ 366, 286, 24 },
		{ 366, 287, 24 },
		{ 366, 288, 24 },
		{ 366, 289, 24 },
		{ 366, 290, 24 },
		{ 366, 291, 24 },
		{ 366, 292, 24 },
		{ 366, 293, 24 },
		{ 366, 294, 24 },
		{ 366, 295, 24 },
		{ 366, 296, 24 },
		{ 366, 297, 24 },
		{ 366, 298, 24 },
		{ 366, 299, 24 },
		{ 366, 300, 24 },
		{ 366, 301, 24 },
		{ 366, 302, 24 },
		{ 366, 303, 24 },
		{ 366, 304, 24 },
		{ 366, 305, 24 },
		{ 366, 306, 24 },
		{ 366, 307, 24 },
		{ 366, 308, 24 },
		{ 366, 309, 24 },
		{ 366, 310, 24 },
		{ 366, 311, 24 },
		{ 366, 312, 24 },
		{ 366, 313, 24 },
		{ 366, 314, 24 },
		{ 366, 315, 24 },
		{ 366, 316, 24 },
		{ 366, 317, 24 },
		{ 366, 318, 24 },
		{ 366, 319, 24 },
		{ 366, 320, 24 },
		{ 366, 321, 24 },
		{ 366, 322, 24 },
		{ 366, 323, 24 },
		{ 366, 324, 24 },
		{ 366, 325, 24 },
		{ 366, 326, 24 },
		{ 366, 327, 24 },
		{ 366, 328, 24 },
		{ 366, 329, 24 },
		{ 366, 330, 24 },
		{ 366, 331, 24 },
		{ 366, 332, 24 },
		{ 366, 333, 24 },
		{ 366, 334, 24 },
		{ 366, 335, 24 },
		{ 366, 336, 24 },
		{ 366, 337, 24 },
		{ 366, 338, 24 },
		{ 367, 285, 24 },
		{ 367, 286, 24 },
		{ 367, 287, 24 },
		{ 367, 288, 24 },
		{ 367, 289, 24 },
		{ 367, 290, 24 },
		{ 367, 291, 24 },
		{ 367, 292, 24 },
		{ 367, 293, 24 },
		{ 367, 294, 24 },
		{ 367, 295, 24 },
		{ 367, 296, 24 },
		{ 367, 297, 24 },
		{ 367, 298, 24 },
		{ 367, 299, 24 },
		{ 367, 300, 24 },
		{ 367, 301, 24 },
		{ 367, 302, 24 },
		{ 367, 303, 24 },
		{ 367, 304, 24 },
		{ 367, 305, 24 },
		{ 367, 306, 24 },
		{ 367, 307, 24 },
		{ 367, 308, 24 },
		{ 367, 309, 24 },
		{ 367, 310, 24 },
		{ 367, 311, 24 },
		{ 367, 312, 24 },
		{ 367, 313, 24 },
		{ 367, 314, 24 },
		{ 367, 315, 24 },
		{ 367, 316, 24 },
		{ 367, 317, 24 },
		{ 367, 318, 24 },
		{ 367, 319, 24 },
		{ 367, 320, 24 },
		{ 367, 321, 24 },
		{ 367, 322, 24 },
		{ 367, 323, 24 },
		{ 367, 324, 24 },
		{ 367, 325, 24 },
		{ 367, 326, 24 },
		{ 367, 327, 24 },
		{ 367, 328, 24 },
		{ 367, 329, 24 },
		{ 367, 330, 24 },
		{ 367, 331, 24 },
		{ 367, 332, 24 },
		{ 367, 333, 24 },
		{ 367, 334, 24 },
		{ 367, 335, 24 },
		{ 367, 336, 24 },
		{ 367, 337, 24 },
		{ 367, 338, 24 },
		{ 367, 339, 24 },
		{ 368, 285, 24 },
		{ 368, 286, 24 },
		{ 368, 287, 24 },
		{ 368, 288, 24 },
		{ 368, 289, 24 },
		{ 368, 290, 24 },
		{ 368, 291, 24 },
		{ 368, 292, 24 },
		{ 368, 293, 24 },
		{ 368, 294, 24 },
		{ 368, 295, 24 },
		{ 368, 296, 24 },
		{ 368, 297, 24 },
		{ 368, 298, 24 },
		{ 368, 299, 24 },
		{ 368, 300, 24 },
		{ 368, 301, 24 },
		{ 368, 302, 24 },
		{ 368, 303, 24 },
		{ 368, 304, 24 },
		{ 368, 305, 24 },
		{ 368, 306, 24 },
		{ 368, 307, 24 },
		{ 368, 308, 24 },
		{ 368, 309, 24 },
		{ 368, 310, 24 },
		{ 368, 311, 24 },
		{ 368, 312, 24 },
		{ 368, 313, 24 },
		{ 368, 314, 24 },
		{ 368, 315, 24 },
		{ 368, 316, 24 },
		{ 368, 317, 24 },
		{ 368, 318, 24 },
		{ 368, 319, 24 },
		{ 368, 320, 24 },
		{ 368, 321, 24 },
		{ 368, 322, 24 },
		{ 368, 323, 24 },
		{ 368, 324, 24 },
		{ 368, 325, 24 },
		{ 368, 326, 24 },
		{ 368, 327, 24 },
		{ 368, 328, 24 },
		{ 368, 329, 24 },
		{ 368, 330, 24 },
		{ 368, 331, 24 },
		{ 368, 332, 24 },
		{ 368, 333, 24 },
		{ 368, 334, 24 },
		{ 368, 335, 24 },
		{ 368, 336, 24 },
		{ 368, 337, 24 },
		{ 368, 338, 24 },
		{ 368, 339, 24 },
		{ 368, 340, 24 },
		{ 369, 286, 24 },
		{ 369, 287, 24 },
		{ 369, 288, 24 },
		{ 369, 289, 24 },
		{ 369, 290, 24 },
		{ 369, 291, 24 },
		{ 369, 292, 24 },
		{ 369, 293, 24 },
		{ 369, 294, 24 },
		{ 369, 295, 24 },
		{ 369, 296, 24 },
		{ 369, 297, 24 },
		{ 369, 298, 24 },
		{ 369, 299, 24 },
		{ 369, 300, 24 },
		{ 369, 301, 24 },
		{ 369, 302, 24 },
		{ 369, 303, 24 },
		{ 369, 304, 24 },
		{ 369, 305, 24 },
		{ 369, 306, 24 },
		{ 369, 307, 24 },
		{ 369, 308, 24 },
		{ 369, 309, 24 },
		{ 369, 310, 24 },
		{ 369, 311, 24 },
		{ 369, 312, 24 },
		{ 369, 313, 24 },
		{ 369, 314, 24 },
		{ 369, 315, 24 },
		{ 369, 316, 24 },
		{ 369, 317, 24 },
		{ 369, 318, 24 },
		{ 369, 319, 24 },
		{ 369, 320, 24 },
		{ 369, 321, 24 },
		{ 369, 322, 24 },
		{ 369, 323, 24 },
		{ 369, 324, 24 },
		{ 369, 325, 24 },
		{ 369, 326, 24 },
		{ 369, 327, 24 },
		{ 369, 328, 24 },
		{ 369, 329, 24 },
		{ 369, 330, 24 },
		{ 369, 331, 24 },
		{ 369, 332, 24 },
		{ 369, 333, 24 },
		{ 369, 334, 24 },
		{ 369, 335, 24 },
		{ 369, 336, 24 },
		{ 369, 337, 24 },
		{ 369, 338, 24 },
		{ 369, 339, 24 },
		{ 369, 340, 24 },
		{ 370, 286, 24 },
		{ 370, 287, 24 },
		{ 370, 288, 24 },
		{ 370, 289, 24 },
		{ 370, 290, 24 },
		{ 370, 291, 24 },
		{ 370, 292, 24 },
		{ 370, 293, 24 },
		{ 370, 294, 24 },
		{ 370, 295, 24 },
		{ 370, 296, 24 },
		{ 370, 297, 24 },
		{ 370, 298, 24 },
		{ 370, 299, 24 },
		{ 370, 300, 24 },
		{ 370, 301, 24 },
		{ 370, 302, 24 },
		{ 370, 303, 24 },
		{ 370, 304, 24 },
		{ 370, 305, 24 },
		{ 370, 306, 24 },
		{ 370, 307, 24 },
		{ 370, 308, 24 },
		{ 370, 309, 24 },
		{ 370, 310, 24 },
		{ 370, 311, 24 },
		{ 370, 312, 24 },
		{ 370, 313, 24 },
		{ 370, 314, 24 },
		{ 370, 315, 24 },
		{ 370, 316, 24 },
		{ 370, 317, 24 },
		{ 370, 318, 24 },
		{ 370, 319, 24 },
		{ 370, 320, 24 },
		{ 370, 321, 24 },
		{ 370, 322, 24 },
		{ 370, 323, 24 },
		{ 370, 324, 24 },
		{ 370, 325, 24 },
		{ 370, 326, 24 },
		{ 370, 327, 24 },
		{ 370, 328, 24 },
		{ 370, 329, 24 },
		{ 370, 330, 24 },
		{ 370, 331, 24 },
		{ 370, 332, 24 },
		{ 370, 333, 24 },
		{ 370, 334, 24 },
		{ 370, 335, 24 },
		{ 370, 336, 24 },
		{ 370, 337, 24 },
		{ 370, 338, 24 },
		{ 370, 339, 24 },
		{ 370, 340, 24 },
		{ 370, 341, 24 },
		{ 371, 286, 24 },
		{ 371, 287, 24 },
		{ 371, 288, 24 },
		{ 371, 289, 24 },
		{ 371, 290, 24 },
		{ 371, 291, 24 },
		{ 371, 292, 24 },
		{ 371, 293, 24 },
		{ 371, 294, 24 },
		{ 371, 295, 24 },
		{ 371, 296, 24 },
		{ 371, 297, 24 },
		{ 371, 298, 24 },
		{ 371, 299, 24 },
		{ 371, 300, 24 },
		{ 371, 301, 24 },
		{ 371, 302, 24 },
		{ 371, 303, 24 },
		{ 371, 304, 24 },
		{ 371, 305, 24 },
		{ 371, 306, 24 },
		{ 371, 307, 24 },
		{ 371, 308, 24 },
		{ 371, 309, 24 },
		{ 371, 310, 24 },
		{ 371, 311, 24 },
		{ 371, 312, 24 },
		{ 371, 313, 24 },
		{ 371, 314, 24 },
		{ 371, 315, 24 },
		{ 371, 316, 24 },
		{ 371, 317, 24 },
		{ 371, 318, 24 },
		{ 371, 319, 24 },
		{ 371, 320, 24 },
		{ 371, 321, 24 },
		{ 371, 322, 24 },
		{ 371, 323, 24 },
		{ 371, 324, 24 },
		{ 371, 325, 24 },
		{ 371, 326, 24 },
		{ 371, 327, 24 },
		{ 371, 328, 24 },
		{ 371, 329, 24 },
		{ 371, 330, 24 },
		{ 371, 331, 24 },
		{ 371, 332, 24 },
		{ 371, 333, 24 },
		{ 371, 334, 24 },
		{ 371, 335, 24 },
		{ 371, 336, 24 },
		{ 371, 337, 24 },
		{ 371, 338, 24 },
		{ 371, 339, 24 },
		{ 371, 340, 24 },
		{ 371, 341, 24 },
		{ 371, 342, 24 },
		{ 372, 286, 24 },
		{ 372, 287, 24 },
		{ 372, 288, 24 },
		{ 372, 289, 24 },
		{ 372, 290, 24 },
		{ 372, 291, 24 },
		{ 372, 292, 24 },
		{ 372, 293, 24 },
		{ 372, 294, 24 },
		{ 372, 295, 24 },
		{ 372, 296, 24 },
		{ 372, 297, 24 },
		{ 372, 298, 24 },
		{ 372, 299, 24 },
		{ 372, 300, 24 },
		{ 372, 301, 24 },
		{ 372, 302, 24 },
		{ 372, 303, 24 },
		{ 372, 304, 24 },
		{ 372, 305, 24 },
		{ 372, 306, 24 },
		{ 372, 307, 24 },
		{ 372, 308, 24 },
		{ 372, 309, 24 },
		{ 372, 310, 24 },
		{ 372, 311, 24 },
		{ 372, 312, 24 },
		{ 372, 313, 24 },
		{ 372, 314, 24 },
		{ 372, 315, 24 },
		{ 372, 316, 24 },
		{ 372, 317, 24 },
		{ 372, 318, 24 },
		{ 372, 319, 24 },
		{ 372, 320, 24 },
		{ 372, 321, 24 },
		{ 372, 322, 24 },
		{ 372, 323, 24 },
		{ 372, 324, 24 },
		{ 372, 325, 24 },
		{ 372, 326, 24 },
		{ 372, 327, 24 },
		{ 372, 328, 24 },
		{ 372, 329, 24 },
		{ 372, 330, 24 },
		{ 372, 331, 24 },
		{ 372, 332, 24 },
		{ 372, 333, 24 },
		{ 372, 334, 24 },
		{ 372, 335, 24 },
		{ 372, 336, 24 },
		{ 372, 337, 24 },
		{ 372, 338, 24 },
		{ 372, 339, 24 },
		{ 372, 340, 24 },
		{ 372, 341, 24 },
		{ 372, 342, 24 },
		{ 372, 343, 24 },
		{ 373, 287, 24 },
		{ 373, 288, 24 },
		{ 373, 289, 24 },
		{ 373, 290, 24 },
		{ 373, 291, 24 },
		{ 373, 292, 24 },
		{ 373, 293, 24 },
		{ 373, 294, 24 },
		{ 373, 295, 24 },
		{ 373, 296, 24 },
		{ 373, 297, 24 },
		{ 373, 298, 24 },
		{ 373, 299, 24 },
		{ 373, 300, 24 },
		{ 373, 301, 24 },
		{ 373, 302, 24 },
		{ 373, 303, 24 },
		{ 373, 304, 24 },
		{ 373, 305, 24 },
		{ 373, 306, 24 },
		{ 373, 307, 24 },
		{ 373, 308, 24 },
		{ 373, 309, 24 },
		{ 373, 310, 24 },
		{ 373, 311, 24 },
		{ 373, 312, 24 },
		{ 373, 313, 24 },
		{ 373, 314, 24 },
		{ 373, 315, 24 },
		{ 373, 316, 24 },
		{ 373, 317, 24 },
		{ 373, 318, 24 },
		{ 373, 319, 24 },
		{ 373, 320, 24 },
		{ 373, 321, 24 },
		{ 373, 322, 24 },
		{ 373, 323, 24 },
		{ 373, 324, 24 },
		{ 373, 325, 24 },
		{ 373, 326, 24 },
		{ 373, 327, 24 },
		{ 373, 328, 24 },
		{ 373, 329, 24 },
		{ 373, 330, 24 },
		{ 373, 331, 24 },
		{ 373, 332, 24 },
		{ 373, 333, 24 },
		{ 373, 334, 24 },
		{ 373, 335, 24 },
		{ 373, 336, 24 },
		{ 373, 337, 24 },
		{ 373, 338, 24 },
		{ 373, 339, 24 },
		{ 373, 340, 24 },
		{ 373, 341, 24 },
		{ 373, 342, 24 },
		{ 373, 343, 24 },
		{ 374, 287, 24 },
		{ 374, 288, 24 },
		{ 374, 289, 24 },
		{ 374, 290, 24 },
		{ 374, 291, 24 },
		{ 374, 292, 24 },
		{ 374, 293, 24 },
		{ 374, 294, 24 },
		{ 374, 295, 24 },
		{ 374, 296, 24 },
		{ 374, 297, 24 },
		{ 374, 298, 24 },
		{ 374, 299, 24 },
		{ 374, 300, 24 },
		{ 374, 301, 24 },
		{ 374, 302, 24 },
		{ 374, 303, 24 },
		{ 374, 304, 24 },
		{ 374, 305, 24 },
		{ 374, 306, 24 },
		{ 374, 307, 24 },
		{ 374, 308, 24 },
		{ 374, 309, 24 },
		{ 374, 310, 24 },
		{ 374, 311, 24 },
		{ 374, 312, 24 },
		{ 374, 313, 24 },
		{ 374, 314, 24 },
		{ 374, 315, 24 },
		{ 374, 316, 24 },
		{ 374, 317, 24 },
		{ 374, 318, 24 },
		{ 374, 319, 24 },
		{ 374, 320, 24 },
		{ 374, 321, 24 },
		{ 374, 322, 24 },
		{ 374, 323, 24 },
		{ 374, 324, 24 },
		{ 374, 325, 24 },
		{ 374, 326, 24 },
		{ 374, 327, 24 },
		{ 374, 328, 24 },
		{ 374, 329, 24 },
		{ 374, 330, 24 },
		{ 374, 331, 24 },
		{ 374, 332, 24 },
		{ 374, 333, 24 },
		{ 374, 334, 24 },
		{ 374, 335, 24 },
		{ 374, 336, 24 },
		{ 374, 337, 24 },
		{ 374, 338, 24 },
		{ 374, 339, 24 },
		{ 374, 340, 24 },
		{ 374, 341, 24 },
		{ 374, 342, 24 },
		{ 374, 343, 24 },
		{ 374, 344, 24 },
		{ 375, 287, 24 },
		{ 375, 288, 24 },
		{ 375, 289, 24 },
		{ 375, 290, 24 },
		{ 375, 291, 24 },
		{ 375, 292, 24 },
		{ 375, 293, 24 },
		{ 375, 294, 24 },
		{ 375, 295, 24 },
		{ 375, 296, 24 },
		{ 375, 297, 24 },
		{ 375, 298, 24 },
		{ 375, 299, 24 },
		{ 375, 300, 24 },
		{ 375, 301, 24 },
		{ 375, 302, 24 },
		{ 375, 303, 24 },
		{ 375, 304, 24 },
		{ 375, 305, 24 },
		{ 375, 306, 24 },
		{ 375, 307, 24 },
		{ 375, 308, 24 },
		{ 375, 309, 24 },
		{ 375, 310, 24 },
		{ 375, 311, 24 },
		{ 375, 312, 24 },
		{ 375, 313, 24 },
		{ 375, 314, 24 },
		{ 375, 315, 24 },
		{ 375, 316, 24 },
		{ 375, 317, 24 },
		{ 375, 318, 24 },
		{ 375, 319, 24 },
		{ 375, 320, 24 },
		{ 375, 321, 24 },
		{ 375, 322, 24 },
		{ 375, 323, 24 },
		{ 375, 324, 24 },
		{ 375, 325, 24 },
		{ 375, 326, 24 },
		{ 375, 327, 24 },
		{ 375, 328, 24 },
		{ 375, 329, 24 },
		{ 375, 330, 24 },
		{ 375, 331, 24 },
		{ 375, 332, 24 },
		{ 375, 333, 24 },
		{ 375, 334, 24 },
		{ 375, 335, 24 },
		{ 375, 336, 24 },
		{ 375, 337, 24 },
		{ 375, 338, 24 },
		{ 375, 339, 24 },
		{ 375, 340, 24 },
		{ 375, 341, 24 },
		{ 375, 342, 24 },
		{ 375, 343, 24 },
		{ 375, 344, 24 },
		{ 375, 345, 24 },
		{ 376, 288, 24 },
		{ 376, 289, 24 },
		{ 376, 290, 24 },
		{ 376, 291, 24 },
		{ 376, 292, 24 },
		{ 376, 293, 24 },
		{ 376, 294, 24 },
		{ 376, 295, 24 },
		{ 376, 296, 24 },
		{ 376, 297, 24 },
		{ 376, 298, 24 },
		{ 376, 299, 24 },
		{ 376, 300, 24 },
		{ 376, 301, 24 },
		{ 376, 302, 24 },
		{ 376, 303, 24 },
		{ 376, 304, 24 },
		{ 376, 305, 24 },
		{ 376, 306, 24 },
		{ 376, 307, 24 },
		{ 376, 308, 24 },
		{ 376, 309, 24 },
		{ 376, 310, 24 },
		{ 376, 311, 24 },
		{ 376, 312, 24 },
		{ 376, 313, 24 },
		{ 376, 314, 24 },
		{ 376, 315, 24 },
		{ 376, 316, 24 },
		{ 376, 317, 24 },
		{ 376, 318, 24 },
		{ 376, 319, 24 },
		{ 376, 320, 24 },
		{ 376, 321, 24 },
		{ 376, 322, 24 },
		{ 376, 323, 24 },
		{ 376, 324, 24 },
		{ 376, 325, 24 },
		{ 376, 326, 24 },
		{ 376, 327, 24 },
		{ 376, 328, 24 },
		{ 376, 329, 24 },
		{ 376, 330, 24 },
		{ 376, 331, 24 },
		{ 376, 332, 24 },
		{ 376, 333, 24 },
		{ 376, 334, 24 },
		{ 376, 335, 24 },
		{ 376, 336, 24 },
		{ 376, 337, 24 },
		{ 376, 338, 24 },
		{ 376, 339, 24 },
		{ 376, 340, 24 },
		{ 376, 341, 24 },
		{ 376, 342, 24 },
		{ 376, 343, 24 },
		{ 376, 344, 24 },
		{ 376, 345, 24 },
		{ 377, 288, 24 },
		{ 377, 289, 24 },
		{ 377, 290, 24 },
		{ 377, 291, 24 },
		{ 377, 292, 24 },
		{ 377, 293, 24 },
		{ 377, 294, 24 },
		{ 377, 295, 24 },
		{ 377, 296, 24 },
		{ 377, 297, 24 },
		{ 377, 298, 24 },
		{ 377, 299, 24 },
		{ 377, 300, 24 },
		{ 377, 301, 24 },
		{ 377, 302, 24 },
		{ 377, 303, 24 },
		{ 377, 304, 24 },
		{ 377, 305, 24 },
		{ 377, 306, 24 },
		{ 377, 307, 24 },
		{ 377, 308, 24 },
		{ 377, 309, 24 },
		{ 377, 310, 24 },
		{ 377, 311, 24 },
		{ 377, 312, 24 },
		{ 377, 313, 24 },
		{ 377, 314, 24 },
		{ 377, 315, 24 },
		{ 377, 316, 24 },
		{ 377, 317, 24 },
		{ 377, 318, 24 },
		{ 377, 319, 24 },
		{ 377, 320, 24 },
		{ 377, 321, 24 },
		{ 377, 322, 24 },
		{ 377, 323, 24 },
		{ 377, 324, 24 },
		{ 377, 325, 24 },
		{ 377, 326, 24 },
		{ 377, 327, 24 },
		{ 377, 328, 24 },
		{ 377, 329, 24 },
		{ 377, 330, 24 },
		{ 377, 331, 24 },
		{ 377, 332, 24 },
		{ 377, 333, 24 },
		{ 377, 334, 24 },
		{ 377, 335, 24 },
		{ 377, 336, 24 },
		{ 377, 337, 24 },
		{ 377, 338, 24 },
		{ 377, 339, 24 },
		{ 377, 340, 24 },
		{ 377, 341, 24 },
		{ 377, 342, 24 },
		{ 377, 343, 24 },
		{ 377, 344, 24 },
		{ 377, 345, 24 },
		{ 377, 346, 24 },
		{ 378, 288, 24 },
		{ 378, 289, 24 },
		{ 378, 290, 24 },
		{ 378, 291, 24 },
		{ 378, 292, 24 },
		{ 378, 293, 24 },
		{ 378, 294, 24 },
		{ 378, 295, 24 },
		{ 378, 296, 24 },
		{ 378, 297, 24 },
		{ 378, 298, 24 },
		{ 378, 299, 24 },
		{ 378, 300, 24 },
		{ 378, 301, 24 },
		{ 378, 302, 24 },
		{ 378, 303, 24 },
		{ 378, 304, 24 },
		{ 378, 305, 24 },
		{ 378, 306, 24 },
		{ 378, 307, 24 },
		{ 378, 308, 24 },
		{ 378, 309, 24 },
		{ 378, 310, 24 },
		{ 378, 311, 24 },
		{ 378, 312, 24 },
		{ 378, 313, 24 },
		{ 378, 314, 24 },
		{ 378, 315, 24 },
		{ 378, 316, 24 },
		{ 378, 317, 24 },
		{ 378, 318, 24 },
		{ 378, 319, 24 },
		{ 378, 320, 24 },
		{ 378, 321, 24 },
		{ 378, 322, 24 },
		{ 378, 323, 24 },
		{ 378, 324, 24 },
		{ 378, 325, 24 },
		{ 378, 326, 24 },
		{ 378, 327, 24 },
		{ 378, 328, 24 },
		{ 378, 329, 24 },
		{ 378, 330, 24 },
		{ 378, 331, 24 },
		{ 378, 332, 24 },
		{ 378, 333, 24 },
		{ 378, 334, 24 },
		{ 378, 335, 24 },
		{ 378, 336, 24 },
		{ 378, 337, 24 },
		{ 378, 338, 24 },
		{ 378, 339, 24 },
		{ 378, 340, 24 },
		{ 378, 341, 24 },
		{ 378, 342, 24 },
		{ 378, 343, 24 },
		{ 378, 344, 24 },
		{ 378, 345, 24 },
		{ 378, 346, 24 },
		{ 378, 347, 24 },
		{ 379, 289, 24 },
		{ 379, 290, 24 },
		{ 379, 291, 24 },
		{ 379, 292, 24 },
		{ 379, 293, 24 },
		{ 379, 294, 24 },
		{ 379, 295, 24 },
		{ 379, 296, 24 },
		{ 379, 297, 24 },
		{ 379, 298, 24 },
		{ 379, 299, 24 },
		{ 379, 300, 24 },
		{ 379, 301, 24 },
		{ 379, 302, 24 },
		{ 379, 303, 24 },
		{ 379, 304, 24 },
		{ 379, 305, 24 },
		{ 379, 306, 24 },
		{ 379, 307, 24 },
		{ 379, 308, 24 },
		{ 379, 309, 24 },
		{ 379, 310, 24 },
		{ 379, 311, 24 },
		{ 379, 312, 24 },
		{ 379, 313, 24 },
		{ 379, 314, 24 },
		{ 379, 315, 24 },
		{ 379, 316, 24 },
		{ 379, 317, 24 },
		{ 379, 318, 24 },
		{ 379, 319, 24 },
		{ 379, 320, 24 },
		{ 379, 321, 24 },
		{ 379, 322, 24 },
		{ 379, 323, 24 },
		{ 379, 324, 24 },
		{ 379, 325, 24 },
		{ 379, 326, 24 },
		{ 379, 327, 24 },
		{ 379, 328, 24 },
		{ 379, 329, 24 },
		{ 379, 330, 24 },
		{ 379, 331, 24 },
		{ 379, 332, 24 },
		{ 379, 333, 24 },
		{ 379, 334, 24 },
		{ 379, 335, 24 },
		{ 379, 336, 24 },
		{ 379, 337, 24 },
		{ 379, 338, 24 },
		{ 379, 339, 24 },
		{ 379, 340, 24 },
		{ 379, 341, 24 },
		{ 379, 342, 24 },
		{ 379, 343, 24 },
		{ 379, 344, 24 },
		{ 379, 345, 24 },
		{ 379, 346, 24 },
		{ 379, 347, 24 },
		{ 380, 289, 24 },
		{ 380, 290, 24 },
		{ 380, 291, 24 },
		{ 380, 292, 24 },
		{ 380, 293, 24 },
		{ 380, 294, 24 },
		{ 380, 295, 24 },
		{ 380, 296, 24 },
		{ 380, 297, 24 },
		{ 380, 298, 24 },
		{ 380, 299, 24 },
		{ 380, 300, 24 },
		{ 380, 301, 24 },
		{ 380, 302, 24 },
		{ 380, 303, 24 },
		{ 380, 304, 24 },
		{ 380, 305, 24 },
		{ 380, 306, 24 },
		{ 380, 307, 24 },
		{ 380, 308, 24 },
		{ 380, 309, 24 },
		{ 380, 310, 24 },
		{ 380, 311, 24 },
		{ 380, 312, 24 },
		{ 380, 313, 24 },
		{ 380, 314, 24 },
		{ 380, 315, 24 },
		{ 380, 316, 24 },
		{ 380, 317, 24 },
		{ 380, 318, 24 },
		{ 380, 319, 24 },
		{ 380, 320, 24 },
		{ 380, 321, 24 },
		{ 380, 322, 24 },
		{ 380, 323, 24 },
		{ 380, 324, 24 },
		{ 380, 325, 24 },
		{ 380, 326, 24 },
		{ 380, 327, 24 },
		{ 380, 328, 24 },
		{ 380, 329, 24 },
		{ 380, 330, 24 },
		{ 380, 331, 24 },
		{ 380, 332, 24 },
		{ 380, 333, 24 },
		{ 380, 334, 24 },
		{ 380, 335, 24 },
		{ 380, 336, 24 },
		{ 380, 337, 24 },
		{ 380, 338, 24 },
		{ 380, 339, 24 },
		{ 380, 340, 24 },
		{ 380, 341, 24 },
		{ 380, 342, 24 },
		{ 380, 343, 24 },
		{ 380, 344, 24 },
		{ 380, 345, 24 },
		{ 380, 346, 24 },
		{ 380, 347, 24 },
		{ 380, 348, 24 },
		{ 381, 290, 24 },
		{ 381, 291, 24 },
		{ 381, 292, 24 },
		{ 381, 293, 24 },
		{ 381, 294, 24 },
		{ 381, 295, 24 },
		{ 381, 296, 24 },
		{ 381, 297, 24 },
		{ 381, 298, 24 },
		{ 381, 299, 24 },
		{ 381, 300, 24 },
		{ 381, 301, 24 },
		{ 381, 302, 24 },
		{ 381, 303, 24 },
		{ 381, 304, 24 },
		{ 381, 305, 24 },
		{ 381, 306, 24 },
		{ 381, 307, 24 },
		{ 381, 308, 24 },
		{ 381, 309, 24 },
		{ 381, 310, 24 },
		{ 381, 311, 24 },
		{ 381, 312, 24 },
		{ 381, 313, 24 },
		{ 381, 314, 24 },
		{ 381, 315, 24 },
		{ 381, 316, 24 },
		{ 381, 317, 24 },
		{ 381, 318, 24 },
		{ 381, 319, 24 },
		{ 381, 320, 24 },
		{ 381, 321, 24 },
		{ 381, 322, 24 },
		{ 381, 323, 24 },
		{ 381, 324, 24 },
		{ 381, 325, 24 },
		{ 381, 326, 24 },
		{ 381, 327, 24 },
		{ 381, 328, 24 },
		{ 381, 329, 24 },
		{ 381, 330, 24 },
		{ 381, 331, 24 },
		{ 381, 332, 24 },
		{ 381, 333, 24 },
		{ 381, 334, 24 },
		{ 381, 335, 24 },
		{ 381, 336, 24 },
		{ 381, 337, 24 },
		{ 381, 338, 24 },
		{ 381, 339, 24 },
		{ 381, 340, 24 },
		{ 381, 341, 24 },
		{ 381, 342, 24 },
		{ 381, 343, 24 },
		{ 381, 344, 24 },
		{ 381, 345, 24 },
		{ 381, 346, 24 },
		{ 381, 347, 24 },
		{ 381, 348, 24 },
		{ 382, 292, 24 },
		{ 382, 293, 24 },
		{ 382, 294, 24 },
		{ 382, 295, 24 },
		{ 382, 296, 24 },
		{ 382, 297, 24 },
		{ 382, 298, 24 },
		{ 382, 299, 24 },
		{ 382, 300, 24 },
		{ 382, 301, 24 },
		{ 382, 302, 24 },
		{ 382, 303, 24 },
		{ 382, 304, 24 },
		{ 382, 305, 24 },
		{ 382, 306, 24 },
		{ 382, 307, 24 },
		{ 382, 308, 24 },
		{ 382, 309, 24 },
		{ 382, 310, 24 },
		{ 382, 311, 24 },
		{ 382, 312, 24 },
		{ 382, 313, 24 },
		{ 382, 314, 24 },
		{ 382, 315, 24 },
		{ 382, 316, 24 },
		{ 382, 317, 24 },
		{ 382, 318, 24 },
		{ 382, 319, 24 },
		{ 382, 320, 24 },
		{ 382, 321, 24 },
		{ 382, 322, 24 },
		{ 382, 323, 24 },
		{ 382, 324, 24 },
		{ 382, 325, 24 },
		{ 382, 326, 24 },
		{ 382, 327, 24 },
		{ 382, 328, 24 },
		{ 382, 329, 24 },
		{ 382, 330, 24 },
		{ 382, 331, 24 },
		{ 382, 332, 24 },
		{ 382, 333, 24 },
		{ 382, 334, 24 },
		{ 382, 335, 24 },
		{ 382, 336, 24 },
		{ 382, 337, 24 },
		{ 382, 338, 24 },
		{ 382, 339, 24 },
		{ 382, 340, 24 },
		{ 382, 341, 24 },
		{ 382, 342, 24 },
		{ 382, 343, 24 },
		{ 382, 344, 24 },
		{ 382, 345, 24 },
		{ 382, 346, 24 },
		{ 382, 347, 24 },
		{ 383, 296, 24 },
		{ 383, 297, 24 },
		{ 383, 298, 24 },
		{ 383, 299, 24 },
		{ 383, 300, 24 },
		{ 383, 301, 24 },
		{ 383, 302, 24 },
		{ 383, 303, 24 },
		{ 383, 304, 24 },
		{ 383, 305, 24 },
		{ 383, 306, 24 },
		{ 383, 307, 24 },
		{ 383, 308, 24 },
		{ 383, 309, 24 },
		{ 383, 310, 24 },
		{ 383, 311, 24 },
		{ 383, 312, 24 },
		{ 383, 313, 24 },
		{ 383, 314, 24 },
		{ 383, 315, 24 },
		{ 383, 316, 24 },
		{ 383, 317, 24 },
		{ 383, 318, 24 },
		{ 383, 319, 24 },
		{ 383, 320, 24 },
		{ 383, 321, 24 },
		{ 383, 322, 24 },
		{ 383, 323, 24 },
		{ 383, 324, 24 },
		{ 383, 325, 24 },
		{ 383, 326, 24 },
		{ 383, 327, 24 },
		{ 383, 328, 24 },
		{ 383, 329, 24 },
		{ 383, 330, 24 },
		{ 383, 331, 24 },
		{ 383, 332, 24 },
		{ 383, 333, 24 },
		{ 383, 334, 24 },
		{ 383, 335, 24 },
		{ 383, 336, 24 },
		{ 383, 337, 24 },
		{ 383, 338, 24 },
		{ 383, 339, 24 },
		{ 383, 340, 24 },
		{ 383, 341, 24 },
		{ 383, 342, 24 },
		{ 383, 343, 24 },
		{ 383, 344, 24 },
		{ 131, 138, 25 },
		{ 131, 139, 25 },
		{ 131, 140, 25 },
		{ 131, 141, 25 },
		{ 131, 142, 25 },
		{ 131, 143, 25 },
		{ 131, 144, 25 },
		{ 131, 145, 25 },
		{ 131, 146, 25 },
		{ 131, 147, 25 },
		{ 131, 148, 25 },
		{ 131, 149, 25 },
		{ 131, 150, 25 },
		{ 131, 151, 25 },
		{ 131, 152, 25 },
		{ 131, 153, 25 },
		{ 131, 154, 25 },
		{ 131, 155, 25 },
		{ 131, 156, 25 },
		{ 131, 157, 25 },
		{ 131, 158, 25 },
		{ 131, 159, 25 },
		{ 131, 160, 25 },
		{ 131, 161, 25 },
		{ 131, 162, 25 },
		{ 131, 163, 25 },
		{ 131, 164, 25 },
		{ 131, 165, 25 },
		{ 131, 166, 25 },
		{ 131, 167, 25 },
		{ 131, 168, 25 },
		{ 131, 169, 25 },
		{ 131, 170, 25 },
		{ 131, 171, 25 },
		{ 131, 172, 25 },
		{ 131, 173, 25 },
		{ 131, 174, 25 },
		{ 131, 175, 25 },
		{ 131, 176, 25 },
		{ 131, 177, 25 },
		{ 131, 178, 25 },
		{ 131, 179, 25 },
		{ 131, 180, 25 },
		{ 131, 181, 25 },
		{ 131, 182, 25 },
		{ 131, 183, 25 },
		{ 131, 184, 25 },
		{ 131, 185, 25 },
		{ 131, 186, 25 },
		{ 131, 187, 25 },
		{ 132, 135, 25 },
		{ 132, 136, 25 },
		{ 132, 137, 25 },
		{ 132, 138, 25 },
		{ 132, 139, 25 },
		{ 132, 140, 25 },
		{ 132, 141, 25 },
		{ 132, 142, 25 },
		{ 132, 143, 25 },
		{ 132, 144, 25 },
		{ 132, 145, 25 },
		{ 132, 146, 25 },
		{ 132, 147, 25 },
		{ 132, 148, 25 },
		{ 132, 149, 25 },
		{ 132, 150, 25 },
		{ 132, 151, 25 },
		{ 132, 152, 25 },
		{ 132, 153, 25 },
		{ 132, 154, 25 },
		{ 132, 155, 25 },
		{ 132, 156, 25 },
		{ 132, 157, 25 },
		{ 132, 158, 25 },
		{ 132, 159, 25 },
		{ 132, 160, 25 },
		{ 132, 161, 25 },
		{ 132, 162, 25 },
		{ 132, 163, 25 },
		{ 132, 164, 25 },
		{ 132, 165, 25 },
		{ 132, 166, 25 },
		{ 132, 167, 25 },
		{ 132, 168, 25 },
		{ 132, 169, 25 },
		{ 132, 170, 25 },
		{ 132, 171, 25 },
		{ 132, 172, 25 },
		{ 132, 173, 25 },
		{ 132, 174, 25 },
		{ 132, 175, 25 },
		{ 132, 176, 25 },
		{ 132, 177, 25 },
		{ 132, 178, 25 },
		{ 132, 179, 25 },
		{ 132, 180, 25 },
		{ 132, 181, 25 },
		{ 132, 182, 25 },
		{ 132, 183, 25 },
		{ 132, 184, 25 },
		{ 132, 185, 25 },
		{ 132, 186, 25 },
		{ 132, 187, 25 },
		{ 132, 188, 25 },
		{ 132, 189, 25 },
		{ 132, 190, 25 },
		{ 132, 191, 25 },
		{ 133, 135, 25 },
		{ 133, 136, 25 },
		{ 133, 137, 25 },
		{ 133, 138, 25 },
		{ 133, 139, 25 },
		{ 133, 140, 25 },
		{ 133, 141, 25 },
		{ 133, 142, 25 },
		{ 133, 143, 25 },
		{ 133, 144, 25 },
		{ 133, 145, 25 },
		{ 133, 146, 25 },
		{ 133, 147, 25 },
		{ 133, 148, 25 },
		{ 133, 149, 25 },
		{ 133, 150, 25 },
		{ 133, 151, 25 },
		{ 133, 152, 25 },
		{ 133, 153, 25 },
		{ 133, 154, 25 },
		{ 133, 155, 25 },
		{ 133, 156, 25 },
		{ 133, 157, 25 },
		{ 133, 158, 25 },
		{ 133, 159, 25 },
		{ 133, 160, 25 },
		{ 133, 161, 25 },
		{ 133, 162, 25 },
		{ 133, 163, 25 },
		{ 133, 164, 25 },
		{ 133, 165, 25 },
		{ 133, 166, 25 },
		{ 133, 167, 25 },
		{ 133, 168, 25 },
		{ 133, 169, 25 },
		{ 133, 170, 25 },
		{ 133, 171, 25 },
		{ 133, 172, 25 },
		{ 133, 173, 25 },
		{ 133, 174, 25 },
		{ 133, 175, 25 },
		{ 133, 176, 25 },
		{ 133, 177, 25 },
		{ 133, 178, 25 },
		{ 133, 179, 25 },
		{ 133, 180, 25 },
		{ 133, 181, 25 },
		{ 133, 182, 25 },
		{ 133, 183, 25 },
		{ 133, 184, 25 },
		{ 133, 185, 25 },
		{ 133, 186, 25 },
		{ 133, 187, 25 },
		{ 133, 188, 25 },
		{ 133, 189, 25 },
		{ 133, 190, 25 },
		{ 133, 191, 25 },
		{ 133, 192, 25 },
		{ 133, 193, 25 },
		{ 134, 135, 25 },
		{ 134, 136, 25 },
		{ 134, 137, 25 },
		{ 134, 138, 25 },
		{ 134, 139, 25 },
		{ 134, 140, 25 },
		{ 134, 141, 25 },
		{ 134, 142, 25 },
		{ 134, 143, 25 },
		{ 134, 144, 25 },
		{ 134, 145, 25 },
		{ 134, 146, 25 },
		{ 134, 147, 25 },
		{ 134, 148, 25 },
		{ 134, 149, 25 },
		{ 134, 150, 25 },
		{ 134, 151, 25 },
		{ 134, 152, 25 },
		{ 134, 153, 25 },
		{ 134, 154, 25 },
		{ 134, 155, 25 },
		{ 134, 156, 25 },
		{ 134, 157, 25 },
		{ 134, 158, 25 },
		{ 134, 159, 25 },
		{ 134, 160, 25 },
		{ 134, 161, 25 },
		{ 134, 162, 25 },
		{ 134, 163, 25 },
		{ 134, 164, 25 },
		{ 134, 165, 25 },
		{ 134, 166, 25 },
		{ 134, 167, 25 },
		{ 134, 168, 25 },
		{ 134, 169, 25 },
		{ 134, 170, 25 },
		{ 134, 171, 25 },
		{ 134, 172, 25 },
		{ 134, 173, 25 },
		{ 134, 174, 25 },
		{ 134, 175, 25 },
		{ 134, 176, 25 },
		{ 134, 177, 25 },
		{ 134, 178, 25 },
		{ 134, 179, 25 },
		{ 134, 180, 25 },
		{ 134, 181, 25 },
		{ 134, 182, 25 },
		{ 134, 183, 25 },
		{ 134, 184, 25 },
		{ 134, 185, 25 },
		{ 134, 186, 25 },
		{ 134, 187, 25 },
		{ 134, 188, 25 },
		{ 134, 189, 25 },
		{ 134, 190, 25 },
		{ 134, 191, 25 },
		{ 134, 192, 25 },
		{ 134, 193, 25 },
		{ 134, 194, 25 },
		{ 135, 136, 25 },
		{ 135, 137, 25 },
		{ 135, 138, 25 },
		{ 135, 139, 25 },
		{ 135, 140, 25 },
		{ 135, 141, 25 },
		{ 135, 142, 25 },
		{ 135, 143, 25 },
		{ 135, 144, 25 },
		{ 135, 145, 25 },
		{ 135, 146, 25 },
		{ 135, 147, 25 },
		{ 135, 148, 25 },
		{ 135, 149, 25 },
		{ 135, 150, 25 },
		{ 135, 151, 25 },
		{ 135, 152, 25 },
		{ 135, 153, 25 },
		{ 135, 154, 25 },
		{ 135, 155, 25 },
		{ 135, 156, 25 },
		{ 135, 157, 25 },
		{ 135, 158, 25 },
		{ 135, 159, 25 },
		{ 135, 160, 25 },
		{ 135, 161, 25 },
		{ 135, 162, 25 },
		{ 135, 163, 25 },
		{ 135, 164, 25 },
		{ 135, 165, 25 },
		{ 135, 166, 25 },
		{ 135, 167, 25 },
		{ 135, 168, 25 },
		{ 135, 169, 25 },
		{ 135, 170, 25 },
		{ 135, 171, 25 },
		{ 135, 172, 25 },
		{ 135, 173, 25 },
		{ 135, 174, 25 },
		{ 135, 175, 25 },
		{ 135, 176, 25 },
		{ 135, 177, 25 },
		{ 135, 178, 25 },
		{ 135, 179, 25 },
		{ 135, 180, 25 },
		{ 135, 181, 25 },
		{ 135, 182, 25 },
		{ 135, 183, 25 },
		{ 135, 184, 25 },
		{ 135, 185, 25 },
		{ 135, 186, 25 },
		{ 135, 187, 25 },
		{ 135, 188, 25 },
		{ 135, 189, 25 },
		{ 135, 190, 25 },
		{ 135, 191, 25 },
		{ 135, 192, 25 },
		{ 135, 193, 25 },
		{ 135, 194, 25 },
		{ 135, 195, 25 },
		{ 136, 137, 25 },
		{ 136, 138, 25 },
		{ 136, 139, 25 },
		{ 136, 140, 25 },
		{ 136, 141, 25 },
		{ 136, 142, 25 },
		{ 136, 143, 25 },
		{ 136, 144, 25 },
		{ 136, 145, 25 },
		{ 136, 146, 25 },
		{ 136, 147, 25 },
		{ 136, 148, 25 },
		{ 136, 149, 25 },
		{ 136, 150, 25 },
		{ 136, 151, 25 },
		{ 136, 152, 25 },
		{ 136, 153, 25 },
		{ 136, 154, 25 },
		{ 136, 155, 25 },
		{ 136, 156, 25 },
		{ 136, 157, 25 },
		{ 136, 158, 25 },
		{ 136, 159, 25 },
		{ 136, 160, 25 },
		{ 136, 161, 25 },
		{ 136, 162, 25 },
		{ 136, 163, 25 },
		{ 136, 164, 25 },
		{ 136, 165, 25 },
		{ 136, 166, 25 },
		{ 136, 167, 25 },
		{ 136, 168, 25 },
		{ 136, 169, 25 },
		{ 136, 170, 25 },
		{ 136, 171, 25 },
		{ 136, 172, 25 },
		{ 136, 173, 25 },
		{ 136, 174, 25 },
		{ 136, 175, 25 },
		{ 136, 176, 25 },
		{ 136, 177, 25 },
		{ 136, 178, 25 },
		{ 136, 179, 25 },
		{ 136, 180, 25 },
		{ 136, 181, 25 },
		{ 136, 182, 25 },
		{ 136, 183, 25 },
		{ 136, 184, 25 },
		{ 136, 185, 25 },
		{ 136, 186, 25 },
		{ 136, 187, 25 },
		{ 136, 188, 25 },
		{ 136, 189, 25 },
		{ 136, 190, 25 },
		{ 136, 191, 25 },
		{ 136, 192, 25 },
		{ 136, 193, 25 },
		{ 136, 194, 25 },
		{ 136, 195, 25 },
		{ 137, 138, 25 },
		{ 137, 139, 25 },
		{ 137, 140, 25 },
		{ 137, 141, 25 },
		{ 137, 142, 25 },
		{ 137, 143, 25 },
		{ 137, 144, 25 },
		{ 137, 145, 25 },
		{ 137, 146, 25 },
		{ 137, 147, 25 },
		{ 137, 148, 25 },
		{ 137, 149, 25 },
		{ 137, 150, 25 },
		{ 137, 151, 25 },
		{ 137, 152, 25 },
		{ 137, 153, 25 },
		{ 137, 154, 25 },
		{ 137, 155, 25 },
		{ 137, 156, 25 },
		{ 137, 157, 25 },
		{ 137, 158, 25 },
		{ 137, 159, 25 },
		{ 137, 160, 25 },
		{ 137, 161, 25 },
		{ 137, 162, 25 },
		{ 137, 163, 25 },
		{ 137, 164, 25 },
		{ 137, 165, 25 },
		{ 137, 166, 25 },
		{ 137, 167, 25 },
		{ 137, 168, 25 },
		{ 137, 169, 25 },
		{ 137, 170, 25 },
		{ 137, 171, 25 },
		{ 137, 172, 25 },
		{ 137, 173, 25 },
		{ 137, 174, 25 },
		{ 137, 175, 25 },
		{ 137, 176, 25 },
		{ 137, 177, 25 },
		{ 137, 178, 25 },
		{ 137, 179, 25 },
		{ 137, 180, 25 },
		{ 137, 181, 25 },
		{ 137, 182, 25 },
		{ 137, 183, 25 },
		{ 137, 184, 25 },
		{ 137, 185, 25 },
		{ 137, 186, 25 },
		{ 137, 187, 25 },
		{ 137, 188, 25 },
		{ 137, 189, 25 },
		{ 137, 190, 25 },
		{ 137, 191, 25 },
		{ 137, 192, 25 },
		{ 137, 193, 25 },
		{ 137, 194, 25 },
		{ 137, 195, 25 },
		{ 137, 196, 25 },
		{ 138, 139, 25 },
		{ 138, 140, 25 },
		{ 138, 141, 25 },
		{ 138, 142, 25 },
		{ 138, 143, 25 },
		{ 138, 144, 25 },
		{ 138, 145, 25 },
		{ 138, 146, 25 },
		{ 138, 147, 25 },
		{ 138, 148, 25 },
		{ 138, 149, 25 },
		{ 138, 150, 25 },
		{ 138, 151, 25 },
		{ 138, 152, 25 },
		{ 138, 153, 25 },
		{ 138, 154, 25 },
		{ 138, 155, 25 },
		{ 138, 156, 25 },
		{ 138, 157, 25 },
		{ 138, 158, 25 },
		{ 138, 159, 25 },
		{ 138, 160, 25 },
		{ 138, 161, 25 },
		{ 138, 162, 25 },
		{ 138, 163, 25 },
		{ 138, 164, 25 },
		{ 138, 165, 25 },
		{ 138, 166, 25 },
		{ 138, 167, 25 },
		{ 138, 168, 25 },
		{ 138, 169, 25 },
		{ 138, 170, 25 },
		{ 138, 171, 25 },
		{ 138, 172, 25 },
		{ 138, 173, 25 },
		{ 138, 174, 25 },
		{ 138, 175, 25 },
		{ 138, 176, 25 },
		{ 138, 177, 25 },
		{ 138, 178, 25 },
		{ 138, 179, 25 },
		{ 138, 180, 25 },
		{ 138, 181, 25 },
		{ 138, 182, 25 },
		{ 138, 183, 25 },
		{ 138, 184, 25 },
		{ 138, 185, 25 },
		{ 138, 186, 25 },
		{ 138, 187, 25 },
		{ 138, 188, 25 },
		{ 138, 189, 25 },
		{ 138, 190, 25 },
		{ 138, 191, 25 },
		{ 138, 192, 25 },
		{ 138, 193, 25 },
		{ 138, 194, 25 },
		{ 138, 195, 25 },
		{ 138, 196, 25 },
		{ 138, 197, 25 },
		{ 139, 140, 25 },
		{ 139, 141, 25 },
		{ 139, 142, 25 },
		{ 139, 143, 25 },
		{ 139, 144, 25 },
		{ 139, 145, 25 },
		{ 139, 146, 25 },
		{ 139, 147, 25 },
		{ 139, 148, 25 },
		{ 139, 149, 25 },
		{ 139, 150, 25 },
		{ 139, 151, 25 },
		{ 139, 152, 25 },
		{ 139, 153, 25 },
		{ 139, 154, 25 },
		{ 139, 155, 25 },
		{ 139, 156, 25 },
		{ 139, 157, 25 },
		{ 139, 158, 25 },
		{ 139, 159, 25 },
		{ 139, 160, 25 },
		{ 139, 161, 25 },
		{ 139, 162, 25 },
		{ 139, 163, 25 },
		{ 139, 164, 25 },
		{ 139, 165, 25 },
		{ 139, 166, 25 },
		{ 139, 167, 25 },
		{ 139, 168, 25 },
		{ 139, 169, 25 },
		{ 139, 170, 25 },
		{ 139, 171, 25 },
		{ 139, 172, 25 },
		{ 139, 173, 25 },
		{ 139, 174, 25 },
		{ 139, 175, 25 },
		{ 139, 176, 25 },
		{ 139, 177, 25 },
		{ 139, 178, 25 },
		{ 139, 179, 25 },
		{ 139, 180, 25 },
		{ 139, 181, 25 },
		{ 139, 182, 25 },
		{ 139, 183, 25 },
		{ 139, 184, 25 },
		{ 139, 185, 25 },
		{ 139, 186, 25 },
		{ 139, 187, 25 },
		{ 139, 188, 25 },
		{ 139, 189, 25 },
		{ 139, 190, 25 },
		{ 139, 191, 25 },
		{ 139, 192, 25 },
		{ 139, 193, 25 },
		{ 139, 194, 25 },
		{ 139, 195, 25 },
		{ 139, 196, 25 },
		{ 139, 197, 25 },
		{ 140, 141, 25 },
		{ 140, 142, 25 },
		{ 140, 143, 25 },
		{ 140, 144, 25 },
		{ 140, 145, 25 },
		{ 140, 146, 25 },
		{ 140, 147, 25 },
		{ 140, 148, 25 },
		{ 140, 149, 25 },
		{ 140, 150, 25 },
		{ 140, 151, 25 },
		{ 140, 152, 25 },
		{ 140, 153, 25 },
		{ 140, 154, 25 },
		{ 140, 155, 25 },
		{ 140, 156, 25 },
		{ 140, 157, 25 },
		{ 140, 158, 25 },
		{ 140, 159, 25 },
		{ 140, 160, 25 },
		{ 140, 161, 25 },
		{ 140, 162, 25 },
		{ 140, 163, 25 },
		{ 140, 164, 25 },
		{ 140, 165, 25 },
		{ 140, 166, 25 },
		{ 140, 167, 25 },
		{ 140, 168, 25 },
		{ 140, 169, 25 },
		{ 140, 170, 25 },
		{ 140, 171, 25 },
		{ 140, 172, 25 },
		{ 140, 173, 25 },
		{ 140, 174, 25 },
		{ 140, 175, 25 },
		{ 140, 176, 25 },
		{ 140, 177, 25 },
		{ 140, 178, 25 },
		{ 140, 179, 25 },
		{ 140, 180, 25 },
		{ 140, 181, 25 },
		{ 140, 182, 25 },
		{ 140, 183, 25 },
		{ 140, 184, 25 },
		{ 140, 185, 25 },
		{ 140, 186, 25 },
		{ 140, 187, 25 },
		{ 140, 188, 25 },
		{ 140, 189, 25 },
		{ 140, 190, 25 },
		{ 140, 191, 25 },
		{ 140, 192, 25 },
		{ 140, 193, 25 },
		{ 140, 194, 25 },
		{ 140, 195, 25 },
		{ 140, 196, 25 },
		{ 140, 197, 25 },
		{ 140, 198, 25 },
		{ 141, 142, 25 },
		{ 141, 143, 25 },
		{ 141, 144, 25 },
		{ 141, 145, 25 },
		{ 141, 146, 25 },
		{ 141, 147, 25 },
		{ 141, 148, 25 },
		{ 141, 149, 25 },
		{ 141, 150, 25 },
		{ 141, 151, 25 },
		{ 141, 152, 25 },
		{ 141, 153, 25 },
		{ 141, 154, 25 },
		{ 141, 155, 25 },
		{ 141, 156, 25 },
		{ 141, 157, 25 },
		{ 141, 158, 25 },
		{ 141, 159, 25 },
		{ 141, 160, 25 },
		{ 141, 161, 25 },
		{ 141, 162, 25 },
		{ 141, 163, 25 },
		{ 141, 164, 25 },
		{ 141, 165, 25 },
		{ 141, 166, 25 },
		{ 141, 167, 25 },
		{ 141, 168, 25 },
		{ 141, 169, 25 },
		{ 141, 170, 25 },
		{ 141, 171, 25 },
		{ 141, 172, 25 },
		{ 141, 173, 25 },
		{ 141, 174, 25 },
		{ 141, 175, 25 },
		{ 141, 176, 25 },
		{ 141, 177, 25 },
		{ 141, 178, 25 },
		{ 141, 179, 25 },
		{ 141, 180, 25 },
		{ 141, 181, 25 },
		{ 141, 182, 25 },
		{ 141, 183, 25 },
		{ 141, 184, 25 },
		{ 141, 185, 25 },
		{ 141, 186, 25 },
		{ 141, 187, 25 },
		{ 141, 188, 25 },
		{ 141, 189, 25 },
		{ 141, 190, 25 },
		{ 141, 191, 25 },
		{ 141, 192, 25 },
		{ 141, 193, 25 },
		{ 141, 194, 25 },
		{ 141, 195, 25 },
		{ 141, 196, 25 },
		{ 141, 197, 25 },
		{ 141, 198, 25 },
		{ 142, 143, 25 },
		{ 142, 144, 25 },
		{ 142, 145, 25 },
		{ 142, 146, 25 },
		{ 142, 147, 25 },
		{ 142, 148, 25 },
		{ 142, 149, 25 },
		{ 142, 150, 25 },
		{ 142, 151, 25 },
		{ 142, 152, 25 },
		{ 142, 153, 25 },
		{ 142, 154, 25 },
		{ 142, 155, 25 },
		{ 142, 156, 25 },
		{ 142, 157, 25 },
		{ 142, 158, 25 },
		{ 142, 159, 25 },
		{ 142, 160, 25 },
		{ 142, 161, 25 },
		{ 142, 162, 25 },
		{ 142, 163, 25 },
		{ 142, 164, 25 },
		{ 142, 165, 25 },
		{ 142, 166, 25 },
		{ 142, 167, 25 },
		{ 142, 168, 25 },
		{ 142, 169, 25 },
		{ 142, 170, 25 },
		{ 142, 171, 25 },
		{ 142, 172, 25 },
		{ 142, 173, 25 },
		{ 142, 174, 25 },
		{ 142, 175, 25 },
		{ 142, 176, 25 },
		{ 142, 177, 25 },
		{ 142, 178, 25 },
		{ 142, 179, 25 },
		{ 142, 180, 25 },
		{ 142, 181, 25 },
		{ 142, 182, 25 },
		{ 142, 183, 25 },
		{ 142, 184, 25 },
		{ 142, 185, 25 },
		{ 142, 186, 25 },
		{ 142, 187, 25 },
		{ 142, 188, 25 },
		{ 142, 189, 25 },
		{ 142, 190, 25 },
		{ 142, 191, 25 },
		{ 142, 192, 25 },
		{ 142, 193, 25 },
		{ 142, 194, 25 },
		{ 142, 195, 25 },
		{ 142, 196, 25 },
		{ 142, 197, 25 },
		{ 142, 198, 25 },
		{ 142, 199, 25 },
		{ 143, 144, 25 },
		{ 143, 145, 25 },
		{ 143, 146, 25 },
		{ 143, 147, 25 },
		{ 143, 148, 25 },
		{ 143, 149, 25 },
		{ 143, 150, 25 },
		{ 143, 151, 25 },
		{ 143, 152, 25 },
		{ 143, 153, 25 },
		{ 143, 154, 25 },
		{ 143, 155, 25 },
		{ 143, 156, 25 },
		{ 143, 157, 25 },
		{ 143, 158, 25 },
		{ 143, 159, 25 },
		{ 143, 160, 25 },
		{ 143, 161, 25 },
		{ 143, 162, 25 },
		{ 143, 163, 25 },
		{ 143, 164, 25 },
		{ 143, 165, 25 },
		{ 143, 166, 25 },
		{ 143, 167, 25 },
		{ 143, 168, 25 },
		{ 143, 169, 25 },
		{ 143, 170, 25 },
		{ 143, 171, 25 },
		{ 143, 172, 25 },
		{ 143, 173, 25 },
		{ 143, 174, 25 },
		{ 143, 175, 25 },
		{ 143, 176, 25 },
		{ 143, 177, 25 },
		{ 143, 178, 25 },
		{ 143, 179, 25 },
		{ 143, 180, 25 },
		{ 143, 181, 25 },
		{ 143, 182, 25 },
		{ 143, 183, 25 },
		{ 143, 184, 25 },
		{ 143, 185, 25 },
		{ 143, 186, 25 },
		{ 143, 187, 25 },
		{ 143, 188, 25 },
		{ 143, 189, 25 },
		{ 143, 190, 25 },
		{ 143, 191, 25 },
		{ 143, 192, 25 },
		{ 143, 193, 25 },
		{ 143, 194, 25 },
		{ 143, 195, 25 },
		{ 143, 196, 25 },
		{ 143, 197, 25 },
		{ 143, 198, 25 },
		{ 143, 199, 25 },
		{ 144, 145, 25 },
		{ 144, 146, 25 },
		{ 144, 147, 25 },
		{ 144, 148, 25 },
		{ 144, 149, 25 },
		{ 144, 150, 25 },
		{ 144, 151, 25 },
		{ 144, 152, 25 },
		{ 144, 153, 25 },
		{ 144, 154, 25 },
		{ 144, 155, 25 },
		{ 144, 156, 25 },
		{ 144, 157, 25 },
		{ 144, 158, 25 },
		{ 144, 159, 25 },
		{ 144, 160, 25 },
		{ 144, 161, 25 },
		{ 144, 162, 25 },
		{ 144, 163, 25 },
		{ 144, 164, 25 },
		{ 144, 165, 25 },
		{ 144, 166, 25 },
		{ 144, 167, 25 },
		{ 144, 168, 25 },
		{ 144, 169, 25 },
		{ 144, 170, 25 },
		{ 144, 171, 25 },
		{ 144, 172, 25 },
		{ 144, 173, 25 },
		{ 144, 174, 25 },
		{ 144, 175, 25 },
		{ 144, 176, 25 },
		{ 144, 177, 25 },
		{ 144, 178, 25 },
		{ 144, 179, 25 },
		{ 144, 180, 25 },
		{ 144, 181, 25 },
		{ 144, 182, 25 },
		{ 144, 183, 25 },
		{ 144, 184, 25 },
		{ 144, 185, 25 },
		{ 144, 186, 25 },
		{ 144, 187, 25 },
		{ 144, 188, 25 },
		{ 144, 189, 25 },
		{ 144, 190, 25 },
		{ 144, 191, 25 },
		{ 144, 192, 25 },
		{ 144, 193, 25 },
		{ 144, 194, 25 },
		{ 144, 195, 25 },
		{ 144, 196, 25 },
		{ 144, 197, 25 },
		{ 144, 198, 25 },
		{ 144, 199, 25 },
		{ 144, 200, 25 },
		{ 145, 146, 25 },
		{ 145, 147, 25 },
		{ 145, 148, 25 },
		{ 145, 149, 25 },
		{ 145, 150, 25 },
		{ 145, 151, 25 },
		{ 145, 152, 25 },
		{ 145, 153, 25 },
		{ 145, 154, 25 },
		{ 145, 155, 25 },
		{ 145, 156, 25 },
		{ 145, 157, 25 },
		{ 145, 158, 25 },
		{ 145, 159, 25 },
		{ 145, 160, 25 },
		{ 145, 161, 25 },
		{ 145, 162, 25 },
		{ 145, 163, 25 },
		{ 145, 164, 25 },
		{ 145, 165, 25 },
		{ 145, 166, 25 },
		{ 145, 167, 25 },
		{ 145, 168, 25 },
		{ 145, 169, 25 },
		{ 145, 170, 25 },
		{ 145, 171, 25 },
		{ 145, 172, 25 },
		{ 145, 173, 25 },
		{ 145, 174, 25 },
		{ 145, 175, 25 },
		{ 145, 176, 25 },
		{ 145, 177, 25 },
		{ 145, 178, 25 },
		{ 145, 179, 25 },
		{ 145, 180, 25 },
		{ 145, 181, 25 },
		{ 145, 182, 25 },
		{ 145, 183, 25 },
		{ 145, 184, 25 },
		{ 145, 185, 25 },
		{ 145, 186, 25 },
		{ 145, 187, 25 },
		{ 145, 188, 25 },
		{ 145, 189, 25 },
		{ 145, 190, 25 },
		{ 145, 191, 25 },
		{ 145, 192, 25 },
		{ 145, 193, 25 },
		{ 145, 194, 25 },
		{ 145, 195, 25 },
		{ 145, 196, 25 },
		{ 145, 197, 25 },
		{ 145, 198, 25 },
		{ 145, 199, 25 },
		{ 145, 200, 25 },
		{ 146, 147, 25 },
		{ 146, 148, 25 },
		{ 146, 149, 25 },
		{ 146, 150, 25 },
		{ 146, 151, 25 },
		{ 146, 152, 25 },
		{ 146, 153, 25 },
		{ 146, 154, 25 },
		{ 146, 155, 25 },
		{ 146, 156, 25 },
		{ 146, 157, 25 },
		{ 146, 158, 25 },
		{ 146, 159, 25 },
		{ 146, 160, 25 },
		{ 146, 161, 25 },
		{ 146, 162, 25 },
		{ 146, 163, 25 },
		{ 146, 164, 25 },
		{ 146, 165, 25 },
		{ 146, 166, 25 },
		{ 146, 167, 25 },
		{ 146, 168, 25 },
		{ 146, 169, 25 },
		{ 146, 170, 25 },
		{ 146, 171, 25 },
		{ 146, 172, 25 },
		{ 146, 173, 25 },
		{ 146, 174, 25 },
		{ 146, 175, 25 },
		{ 146, 176, 25 },
		{ 146, 177, 25 },
		{ 146, 178, 25 },
		{ 146, 179, 25 },
		{ 146, 180, 25 },
		{ 146, 181, 25 },
		{ 146, 182, 25 },
		{ 146, 183, 25 },
		{ 146, 184, 25 },
		{ 146, 185, 25 },
		{ 146, 186, 25 },
		{ 146, 187, 25 },
		{ 146, 188, 25 },
		{ 146, 189, 25 },
		{ 146, 190, 25 },
		{ 146, 191, 25 },
		{ 146, 192, 25 },
		{ 146, 193, 25 },
		{ 146, 194, 25 },
		{ 146, 195, 25 },
		{ 146, 196, 25 },
		{ 146, 197, 25 },
		{ 146, 198, 25 },
		{ 146, 199, 25 },
		{ 146, 200, 25 },
		{ 146, 201, 25 },
		{ 147, 148, 25 },
		{ 147, 149, 25 },
		{ 147, 150, 25 },
		{ 147, 151, 25 },
		{ 147, 152, 25 },
		{ 147, 153, 25 },
		{ 147, 154, 25 },
		{ 147, 155, 25 },
		{ 147, 156, 25 },
		{ 147, 157, 25 },
		{ 147, 158, 25 },
		{ 147, 159, 25 },
		{ 147, 160, 25 },
		{ 147, 161, 25 },
		{ 147, 162, 25 },
		{ 147, 163, 25 },
		{ 147, 164, 25 },
		{ 147, 165, 25 },
		{ 147, 166, 25 },
		{ 147, 167, 25 },
		{ 147, 168, 25 },
		{ 147, 169, 25 },
		{ 147, 170, 25 },
		{ 147, 171, 25 },
		{ 147, 172, 25 },
		{ 147, 173, 25 },
		{ 147, 174, 25 },
		{ 147, 175, 25 },
		{ 147, 176, 25 },
		{ 147, 177, 25 },
		{ 147, 178, 25 },
		{ 147, 179, 25 },
		{ 147, 180, 25 },
		{ 147, 181, 25 },
		{ 147, 182, 25 },
		{ 147, 183, 25 },
		{ 147, 184, 25 },
		{ 147, 185, 25 },
		{ 147, 186, 25 },
		{ 147, 187, 25 },
		{ 147, 188, 25 },
		{ 147, 189, 25 },
		{ 147, 190, 25 },
		{ 147, 191, 25 },
		{ 147, 192, 25 },
		{ 147, 193, 25 },
		{ 147, 194, 25 },
		{ 147, 195, 25 },
		{ 147, 196, 25 },
		{ 147, 197, 25 },
		{ 147, 198, 25 },
		{ 147, 199, 25 },
		{ 147, 200, 25 },
		{ 147, 201, 25 },
		{ 148, 149, 25 },
		{ 148, 150, 25 },
		{ 148, 151, 25 },
		{ 148, 152, 25 },
		{ 148, 153, 25 },
		{ 148, 154, 25 },
		{ 148, 155, 25 },
		{ 148, 156, 25 },
		{ 148, 157, 25 },
		{ 148, 158, 25 },
		{ 148, 159, 25 },
		{ 148, 160, 25 },
		{ 148, 161, 25 },
		{ 148, 162, 25 },
		{ 148, 163, 25 },
		{ 148, 164, 25 },
		{ 148, 165, 25 },
		{ 148, 166, 25 },
		{ 148, 167, 25 },
		{ 148, 168, 25 },
		{ 148, 169, 25 },
		{ 148, 170, 25 },
		{ 148, 171, 25 },
		{ 148, 172, 25 },
		{ 148, 173, 25 },
		{ 148, 174, 25 },
		{ 148, 175, 25 },
		{ 148, 176, 25 },
		{ 148, 177, 25 },
		{ 148, 178, 25 },
		{ 148, 179, 25 },
		{ 148, 180, 25 },
		{ 148, 181, 25 },
		{ 148, 182, 25 },
		{ 148, 183, 25 },
		{ 148, 184, 25 },
		{ 148, 185, 25 },
		{ 148, 186, 25 },
		{ 148, 187, 25 },
		{ 148, 188, 25 },
		{ 148, 189, 25 },
		{ 148, 190, 25 },
		{ 148, 191, 25 },
		{ 148, 192, 25 },
		{ 148, 193, 25 },
		{ 148, 194, 25 },
		{ 148, 195, 25 },
		{ 148, 196, 25 },
		{ 148, 197, 25 },
		{ 148, 198, 25 },
		{ 148, 199, 25 },
		{ 148, 200, 25 },
		{ 148, 201, 25 },
		{ 148, 202, 25 },
		{ 149, 150, 25 },
		{ 149, 151, 25 },
		{ 149, 152, 25 },
		{ 149, 153, 25 },
		{ 149, 154, 25 },
		{ 149, 155, 25 },
		{ 149, 156, 25 },
		{ 149, 157, 25 },
		{ 149, 158, 25 },
		{ 149, 159, 25 },
		{ 149, 160, 25 },
		{ 149, 161, 25 },
		{ 149, 162, 25 },
		{ 149, 163, 25 },
		{ 149, 164, 25 },
		{ 149, 165, 25 },
		{ 149, 166, 25 },
		{ 149, 167, 25 },
		{ 149, 168, 25 },
		{ 149, 169, 25 },
		{ 149, 170, 25 },
		{ 149, 171, 25 },
		{ 149, 172, 25 },
		{ 149, 173, 25 },
		{ 149, 174, 25 },
		{ 149, 175, 25 },
		{ 149, 176, 25 },
		{ 149, 177, 25 },
		{ 149, 178, 25 },
		{ 149, 179, 25 },
		{ 149, 180, 25 },
		{ 149, 181, 25 },
		{ 149, 182, 25 },
		{ 149, 183, 25 },
		{ 149, 184, 25 },
		{ 149, 185, 25 },
		{ 149, 186, 25 },
		{ 149, 187, 25 },
		{ 149, 188, 25 },
		{ 149, 189, 25 },
		{ 149, 190, 25 },
		{ 149, 191, 25 },
		{ 149, 192, 25 },
		{ 149, 193, 25 },
		{ 149, 194, 25 },
		{ 149, 195, 25 },
		{ 149, 196, 25 },
		{ 149, 197, 25 },
		{ 149, 198, 25 },
		{ 149, 199, 25 },
		{ 149, 200, 25 },
		{ 149, 201, 25 },
		{ 149, 202, 25 },
		{ 150, 151, 25 },
		{ 150, 152, 25 },
		{ 150, 153, 25 },
		{ 150, 154, 25 },
		{ 150, 155, 25 },
		{ 150, 156, 25 },
		{ 150, 157, 25 },
		{ 150, 158, 25 },
		{ 150, 159, 25 },
		{ 150, 160, 25 },
		{ 150, 161, 25 },
		{ 150, 162, 25 },
		{ 150, 163, 25 },
		{ 150, 164, 25 },
		{ 150, 165, 25 },
		{ 150, 166, 25 },
		{ 150, 167, 25 },
		{ 150, 168, 25 },
		{ 150, 169, 25 },
		{ 150, 170, 25 },
		{ 150, 171, 25 },
		{ 150, 172, 25 },
		{ 150, 173, 25 },
		{ 150, 174, 25 },
		{ 150, 175, 25 },
		{ 150, 176, 25 },
		{ 150, 177, 25 },
		{ 150, 178, 25 },
		{ 150, 179, 25 },
		{ 150, 180, 25 },
		{ 150, 181, 25 },
		{ 150, 182, 25 },
		{ 150, 183, 25 },
		{ 150, 184, 25 },
		{ 150, 185, 25 },
		{ 150, 186, 25 },
		{ 150, 187, 25 },
		{ 150, 188, 25 },
		{ 150, 189, 25 },
		{ 150, 190, 25 },
		{ 150, 191, 25 },
		{ 150, 192, 25 },
		{ 150, 193, 25 },
		{ 150, 194, 25 },
		{ 150, 195, 25 },
		{ 150, 196, 25 },
		{ 150, 197, 25 },
		{ 150, 198, 25 },
		{ 150, 199, 25 },
		{ 150, 200, 25 },
		{ 150, 201, 25 },
		{ 150, 202, 25 },
		{ 150, 203, 25 },
		{ 151, 152, 25 },
		{ 151, 153, 25 },
		{ 151, 154, 25 },
		{ 151, 155, 25 },
		{ 151, 156, 25 },
		{ 151, 157, 25 },
		{ 151, 158, 25 },
		{ 151, 159, 25 },
		{ 151, 160, 25 },
		{ 151, 161, 25 },
		{ 151, 162, 25 },
		{ 151, 163, 25 },
		{ 151, 164, 25 },
		{ 151, 165, 25 },
		{ 151, 166, 25 },
		{ 151, 167, 25 },
		{ 151, 168, 25 },
		{ 151, 169, 25 },
		{ 151, 170, 25 },
		{ 151, 171, 25 },
		{ 151, 172, 25 },
		{ 151, 173, 25 },
		{ 151, 174, 25 },
		{ 151, 175, 25 },
		{ 151, 176, 25 },
		{ 151, 177, 25 },
		{ 151, 178, 25 },
		{ 151, 179, 25 },
		{ 151, 180, 25 },
		{ 151, 181, 25 },
		{ 151, 182, 25 },
		{ 151, 183, 25 },
		{ 151, 184, 25 },
		{ 151, 185, 25 },
		{ 151, 186, 25 },
		{ 151, 187, 25 },
		{ 151, 188, 25 },
		{ 151, 189, 25 },
		{ 151, 190, 25 },
		{ 151, 191, 25 },
		{ 151, 192, 25 },
		{ 151, 193, 25 },
		{ 151, 194, 25 },
		{ 151, 195, 25 },
		{ 151, 196, 25 },
		{ 151, 197, 25 },
		{ 151, 198, 25 },
		{ 151, 199, 25 },
		{ 151, 200, 25 },
		{ 151, 201, 25 },
		{ 151, 202, 25 },
		{ 151, 203, 25 },
		{ 152, 153, 25 },
		{ 152, 154, 25 },
		{ 152, 155, 25 },
		{ 152, 156, 25 },
		{ 152, 157, 25 },
		{ 152, 158, 25 },
		{ 152, 159, 25 },
		{ 152, 160, 25 },
		{ 152, 161, 25 },
		{ 152, 162, 25 },
		{ 152, 163, 25 },
		{ 152, 164, 25 },
		{ 152, 165, 25 },
		{ 152, 166, 25 },
		{ 152, 167, 25 },
		{ 152, 168, 25 },
		{ 152, 169, 25 },
		{ 152, 170, 25 },
		{ 152, 171, 25 },
		{ 152, 172, 25 },
		{ 152, 173, 25 },
		{ 152, 174, 25 },
		{ 152, 175, 25 },
		{ 152, 176, 25 },
		{ 152, 177, 25 },
		{ 152, 178, 25 },
		{ 152, 179, 25 },
		{ 152, 180, 25 },
		{ 152, 181, 25 },
		{ 152, 182, 25 },
		{ 152, 183, 25 },
		{ 152, 184, 25 },
		{ 152, 185, 25 },
		{ 152, 186, 25 },
		{ 152, 187, 25 },
		{ 152, 188, 25 },
		{ 152, 189, 25 },
		{ 152, 190, 25 },
		{ 152, 191, 25 },
		{ 152, 192, 25 },
		{ 152, 193, 25 },
		{ 152, 194, 25 },
		{ 152, 195, 25 },
		{ 152, 196, 25 },
		{ 152, 197, 25 },
		{ 152, 198, 25 },
		{ 152, 199, 25 },
		{ 152, 200, 25 },
		{ 152, 201, 25 },
		{ 152, 202, 25 },
		{ 152, 203, 25 },
		{ 152, 204, 25 },
		{ 153, 154, 25 },
		{ 153, 155, 25 },
		{ 153, 156, 25 },
		{ 153, 157, 25 },
		{ 153, 158, 25 },
		{ 153, 159, 25 },
		{ 153, 160, 25 },
		{ 153, 161, 25 },
		{ 153, 162, 25 },
		{ 153, 163, 25 },
		{ 153, 164, 25 },
		{ 153, 165, 25 },
		{ 153, 166, 25 },
		{ 153, 167, 25 },
		{ 153, 168, 25 },
		{ 153, 169, 25 },
		{ 153, 170, 25 },
		{ 153, 171, 25 },
		{ 153, 172, 25 },
		{ 153, 173, 25 },
		{ 153, 174, 25 },
		{ 153, 175, 25 },
		{ 153, 176, 25 },
		{ 153, 177, 25 },
		{ 153, 178, 25 },
		{ 153, 179, 25 },
		{ 153, 180, 25 },
		{ 153, 181, 25 },
		{ 153, 182, 25 },
		{ 153, 183, 25 },
		{ 153, 184, 25 },
		{ 153, 185, 25 },
		{ 153, 186, 25 },
		{ 153, 187, 25 },
		{ 153, 188, 25 },
		{ 153, 189, 25 },
		{ 153, 190, 25 },
		{ 153, 191, 25 },
		{ 153, 192, 25 },
		{ 153, 193, 25 },
		{ 153, 194, 25 },
		{ 153, 195, 25 },
		{ 153, 196, 25 },
		{ 153, 197, 25 },
		{ 153, 198, 25 },
		{ 153, 199, 25 },
		{ 153, 200, 25 },
		{ 153, 201, 25 },
		{ 153, 202, 25 },
		{ 153, 203, 25 },
		{ 153, 204, 25 },
		{ 154, 155, 25 },
		{ 154, 156, 25 },
		{ 154, 157, 25 },
		{ 154, 158, 25 },
		{ 154, 159, 25 },
		{ 154, 160, 25 },
		{ 154, 161, 25 },
		{ 154, 162, 25 },
		{ 154, 163, 25 },
		{ 154, 164, 25 },
		{ 154, 165, 25 },
		{ 154, 166, 25 },
		{ 154, 167, 25 },
		{ 154, 168, 25 },
		{ 154, 169, 25 },
		{ 154, 170, 25 },
		{ 154, 171, 25 },
		{ 154, 172, 25 },
		{ 154, 173, 25 },
		{ 154, 174, 25 },
		{ 154, 175, 25 },
		{ 154, 176, 25 },
		{ 154, 177, 25 },
		{ 154, 178, 25 },
		{ 154, 179, 25 },
		{ 154, 180, 25 },
		{ 154, 181, 25 },
		{ 154, 182, 25 },
		{ 154, 183, 25 },
		{ 154, 184, 25 },
		{ 154, 185, 25 },
		{ 154, 186, 25 },
		{ 154, 187, 25 },
		{ 154, 188, 25 },
		{ 154, 189, 25 },
		{ 154, 190, 25 },
		{ 154, 191, 25 },
		{ 154, 192, 25 },
		{ 154, 193, 25 },
		{ 154, 194, 25 },
		{ 154, 195, 25 },
		{ 154, 196, 25 },
		{ 154, 197, 25 },
		{ 154, 198, 25 },
		{ 154, 199, 25 },
		{ 154, 200, 25 },
		{ 154, 201, 25 },
		{ 154, 202, 25 },
		{ 154, 203, 25 },
		{ 154, 204, 25 },
		{ 154, 205, 25 },
		{ 155, 156, 25 },
		{ 155, 157, 25 },
		{ 155, 158, 25 },
		{ 155, 159, 25 },
		{ 155, 160, 25 },
		{ 155, 161, 25 },
		{ 155, 162, 25 },
		{ 155, 163, 25 },
		{ 155, 164, 25 },
		{ 155, 165, 25 },
		{ 155, 166, 25 },
		{ 155, 167, 25 },
		{ 155, 168, 25 },
		{ 155, 169, 25 },
		{ 155, 170, 25 },
		{ 155, 171, 25 },
		{ 155, 172, 25 },
		{ 155, 173, 25 },
		{ 155, 174, 25 },
		{ 155, 175, 25 },
		{ 155, 176, 25 },
		{ 155, 177, 25 },
		{ 155, 178, 25 },
		{ 155, 179, 25 },
		{ 155, 180, 25 },
		{ 155, 181, 25 },
		{ 155, 182, 25 },
		{ 155, 183, 25 },
		{ 155, 184, 25 },
		{ 155, 185, 25 },
		{ 155, 186, 25 },
		{ 155, 187, 25 },
		{ 155, 188, 25 },
		{ 155, 189, 25 },
		{ 155, 190, 25 },
		{ 155, 191, 25 },
		{ 155, 192, 25 },
		{ 155, 193, 25 },
		{ 155, 194, 25 },
		{ 155, 195, 25 },
		{ 155, 196, 25 },
		{ 155, 197, 25 },
		{ 155, 198, 25 },
		{ 155, 199, 25 },
		{ 155, 200, 25 },
		{ 155, 201, 25 },
		{ 155, 202, 25 },
		{ 155, 203, 25 },
		{ 155, 204, 25 },
		{ 155, 205, 25 },
		{ 156, 157, 25 },
		{ 156, 158, 25 },
		{ 156, 159, 25 },
		{ 156, 160, 25 },
		{ 156, 161, 25 },
		{ 156, 162, 25 },
		{ 156, 163, 25 },
		{ 156, 164, 25 },
		{ 156, 165, 25 },
		{ 156, 166, 25 },
		{ 156, 167, 25 },
		{ 156, 168, 25 },
		{ 156, 169, 25 },
		{ 156, 170, 25 },
		{ 156, 171, 25 },
		{ 156, 172, 25 },
		{ 156, 173, 25 },
		{ 156, 174, 25 },
		{ 156, 175, 25 },
		{ 156, 176, 25 },
		{ 156, 177, 25 },
		{ 156, 178, 25 },
		{ 156, 179, 25 },
		{ 156, 180, 25 },
		{ 156, 181, 25 },
		{ 156, 182, 25 },
		{ 156, 183, 25 },
		{ 156, 184, 25 },
		{ 156, 185, 25 },
		{ 156, 186, 25 },
		{ 156, 187, 25 },
		{ 156, 188, 25 },
		{ 156, 189, 25 },
		{ 156, 190, 25 },
		{ 156, 191, 25 },
		{ 156, 192, 25 },
		{ 156, 193, 25 },
		{ 156, 194, 25 },
		{ 156, 195, 25 },
		{ 156, 196, 25 },
		{ 156, 197, 25 },
		{ 156, 198, 25 },
		{ 156, 199, 25 },
		{ 156, 200, 25 },
		{ 156, 201, 25 },
		{ 156, 202, 25 },
		{ 156, 203, 25 },
		{ 156, 204, 25 },
		{ 156, 205, 25 },
		{ 156, 206, 25 },
		{ 157, 158, 25 },
		{ 157, 159, 25 },
		{ 157, 160, 25 },
		{ 157, 161, 25 },
		{ 157, 162, 25 },
		{ 157, 163, 25 },
		{ 157, 164, 25 },
		{ 157, 165, 25 },
		{ 157, 166, 25 },
		{ 157, 167, 25 },
		{ 157, 168, 25 },
		{ 157, 169, 25 },
		{ 157, 170, 25 },
		{ 157, 171, 25 },
		{ 157, 172, 25 },
		{ 157, 173, 25 },
		{ 157, 174, 25 },
		{ 157, 175, 25 },
		{ 157, 176, 25 },
		{ 157, 177, 25 },
		{ 157, 178, 25 },
		{ 157, 179, 25 },
		{ 157, 180, 25 },
		{ 157, 181, 25 },
		{ 157, 182, 25 },
		{ 157, 183, 25 },
		{ 157, 184, 25 },
		{ 157, 185, 25 },
		{ 157, 186, 25 },
		{ 157, 187, 25 },
		{ 157, 188, 25 },
		{ 157, 189, 25 },
		{ 157, 190, 25 },
		{ 157, 191, 25 },
		{ 157, 192, 25 },
		{ 157, 193, 25 },
		{ 157, 194, 25 },
		{ 157, 195, 25 },
		{ 157, 196, 25 },
		{ 157, 197, 25 },
		{ 157, 198, 25 },
		{ 157, 199, 25 },
		{ 157, 200, 25 },
		{ 157, 201, 25 },
		{ 157, 202, 25 },
		{ 157, 203, 25 },
		{ 157, 204, 25 },
		{ 157, 205, 25 },
		{ 157, 206, 25 },
		{ 158, 159, 25 },
		{ 158, 160, 25 },
		{ 158, 161, 25 },
		{ 158, 162, 25 },
		{ 158, 163, 25 },
		{ 158, 164, 25 },
		{ 158, 165, 25 },
		{ 158, 166, 25 },
		{ 158, 167, 25 },
		{ 158, 168, 25 },
		{ 158, 169, 25 },
		{ 158, 170, 25 },
		{ 158, 171, 25 },
		{ 158, 172, 25 },
		{ 158, 173, 25 },
		{ 158, 174, 25 },
		{ 158, 175, 25 },
		{ 158, 176, 25 },
		{ 158, 177, 25 },
		{ 158, 178, 25 },
		{ 158, 179, 25 },
		{ 158, 180, 25 },
		{ 158, 181, 25 },
		{ 158, 182, 25 },
		{ 158, 183, 25 },
		{ 158, 184, 25 },
		{ 158, 185, 25 },
		{ 158, 186, 25 },
		{ 158, 187, 25 },
		{ 158, 188, 25 },
		{ 158, 189, 25 },
		{ 158, 190, 25 },
		{ 158, 191, 25 },
		{ 158, 192, 25 },
		{ 158, 193, 25 },
		{ 158, 194, 25 },
		{ 158, 195, 25 },
		{ 158, 196, 25 },
		{ 158, 197, 25 },
		{ 158, 198, 25 },
		{ 158, 199, 25 },
		{ 158, 200, 25 },
		{ 158, 201, 25 },
		{ 158, 202, 25 },
		{ 158, 203, 25 },
		{ 158, 204, 25 },
		{ 158, 205, 25 },
		{ 158, 206, 25 },
		{ 158, 207, 25 },
		{ 159, 160, 25 },
		{ 159, 161, 25 },
		{ 159, 162, 25 },
		{ 159, 163, 25 },
		{ 159, 164, 25 },
		{ 159, 165, 25 },
		{ 159, 166, 25 },
		{ 159, 167, 25 },
		{ 159, 168, 25 },
		{ 159, 169, 25 },
		{ 159, 170, 25 },
		{ 159, 171, 25 },
		{ 159, 172, 25 },
		{ 159, 173, 25 },
		{ 159, 174, 25 },
		{ 159, 175, 25 },
		{ 159, 176, 25 },
		{ 159, 177, 25 },
		{ 159, 178, 25 },
		{ 159, 179, 25 },
		{ 159, 180, 25 },
		{ 159, 181, 25 },
		{ 159, 182, 25 },
		{ 159, 183, 25 },
		{ 159, 184, 25 },
		{ 159, 185, 25 },
		{ 159, 186, 25 },
		{ 159, 187, 25 },
		{ 159, 188, 25 },
		{ 159, 189, 25 },
		{ 159, 190, 25 },
		{ 159, 191, 25 },
		{ 159, 192, 25 },
		{ 159, 193, 25 },
		{ 159, 194, 25 },
		{ 159, 195, 25 },
		{ 159, 196, 25 },
		{ 159, 197, 25 },
		{ 159, 198, 25 },
		{ 159, 199, 25 },
		{ 159, 200, 25 },
		{ 159, 201, 25 },
		{ 159, 202, 25 },
		{ 159, 203, 25 },
		{ 159, 204, 25 },
		{ 159, 205, 25 },
		{ 159, 206, 25 },
		{ 159, 207, 25 },
		{ 160, 161, 25 },
		{ 160, 162, 25 },
		{ 160, 163, 25 },
		{ 160, 164, 25 },
		{ 160, 165, 25 },
		{ 160, 166, 25 },
		{ 160, 167, 25 },
		{ 160, 168, 25 },
		{ 160, 169, 25 },
		{ 160, 170, 25 },
		{ 160, 171, 25 },
		{ 160, 172, 25 },
		{ 160, 173, 25 },
		{ 160, 174, 25 },
		{ 160, 175, 25 },
		{ 160, 176, 25 },
		{ 160, 177, 25 },
		{ 160, 178, 25 },
		{ 160, 179, 25 },
		{ 160, 180, 25 },
		{ 160, 181, 25 },
		{ 160, 182, 25 },
		{ 160, 183, 25 },
		{ 160, 184, 25 },
		{ 160, 185, 25 },
		{ 160, 186, 25 },
		{ 160, 187, 25 },
		{ 160, 188, 25 },
		{ 160, 189, 25 },
		{ 160, 190, 25 },
		{ 160, 191, 25 },
		{ 160, 192, 25 },
		{ 160, 193, 25 },
		{ 160, 194, 25 },
		{ 160, 195, 25 },
		{ 160, 196, 25 },
		{ 160, 197, 25 },
		{ 160, 198, 25 },
		{ 160, 199, 25 },
		{ 160, 200, 25 },
		{ 160, 201, 25 },
		{ 160, 202, 25 },
		{ 160, 203, 25 },
		{ 160, 204, 25 },
		{ 160, 205, 25 },
		{ 160, 206, 25 },
		{ 160, 207, 25 },
		{ 160, 208, 25 },
		{ 161, 162, 25 },
		{ 161, 163, 25 },
		{ 161, 164, 25 },
		{ 161, 165, 25 },
		{ 161, 166, 25 },
		{ 161, 167, 25 },
		{ 161, 168, 25 },
		{ 161, 169, 25 },
		{ 161, 170, 25 },
		{ 161, 171, 25 },
		{ 161, 172, 25 },
		{ 161, 173, 25 },
		{ 161, 174, 25 },
		{ 161, 175, 25 },
		{ 161, 176, 25 },
		{ 161, 177, 25 },
		{ 161, 178, 25 },
		{ 161, 179, 25 },
		{ 161, 180, 25 },
		{ 161, 181, 25 },
		{ 161, 182, 25 },
		{ 161, 183, 25 },
		{ 161, 184, 25 },
		{ 161, 185, 25 },
		{ 161, 186, 25 },
		{ 161, 187, 25 },
		{ 161, 188, 25 },
		{ 161, 189, 25 },
		{ 161, 190, 25 },
		{ 161, 191, 25 },
		{ 161, 192, 25 },
		{ 161, 193, 25 },
		{ 161, 194, 25 },
		{ 161, 195, 25 },
		{ 161, 196, 25 },
		{ 161, 197, 25 },
		{ 161, 198, 25 },
		{ 161, 199, 25 },
		{ 161, 200, 25 },
		{ 161, 201, 25 },
		{ 161, 202, 25 },
		{ 161, 203, 25 },
		{ 161, 204, 25 },
		{ 161, 205, 25 },
		{ 161, 206, 25 },
		{ 161, 207, 25 },
		{ 161, 208, 25 },
		{ 162, 163, 25 },
		{ 162, 164, 25 },
		{ 162, 165, 25 },
		{ 162, 166, 25 },
		{ 162, 167, 25 },
		{ 162, 168, 25 },
		{ 162, 169, 25 },
		{ 162, 170, 25 },
		{ 162, 171, 25 },
		{ 162, 172, 25 },
		{ 162, 173, 25 },
		{ 162, 174, 25 },
		{ 162, 175, 25 },
		{ 162, 176, 25 },
		{ 162, 177, 25 },
		{ 162, 178, 25 },
		{ 162, 179, 25 },
		{ 162, 180, 25 },
		{ 162, 181, 25 },
		{ 162, 182, 25 },
		{ 162, 183, 25 },
		{ 162, 184, 25 },
		{ 162, 185, 25 },
		{ 162, 186, 25 },
		{ 162, 187, 25 },
		{ 162, 188, 25 },
		{ 162, 189, 25 },
		{ 162, 190, 25 },
		{ 162, 191, 25 },
		{ 162, 192, 25 },
		{ 162, 193, 25 },
		{ 162, 194, 25 },
		{ 162, 195, 25 },
		{ 162, 196, 25 },
		{ 162, 197, 25 },
		{ 162, 198, 25 },
		{ 162, 199, 25 },
		{ 162, 200, 25 },
		{ 162, 201, 25 },
		{ 162, 202, 25 },
		{ 162, 203, 25 },
		{ 162, 204, 25 },
		{ 162, 205, 25 },
		{ 162, 206, 25 },
		{ 162, 207, 25 },
		{ 162, 208, 25 },
		{ 162, 209, 25 },
		{ 163, 164, 25 },
		{ 163, 165, 25 },
		{ 163, 166, 25 },
		{ 163, 167, 25 },
		{ 163, 168, 25 },
		{ 163, 169, 25 },
		{ 163, 170, 25 },
		{ 163, 171, 25 },
		{ 163, 172, 25 },
		{ 163, 173, 25 },
		{ 163, 174, 25 },
		{ 163, 175, 25 },
		{ 163, 176, 25 },
		{ 163, 177, 25 },
		{ 163, 178, 25 },
		{ 163, 179, 25 },
		{ 163, 180, 25 },
		{ 163, 181, 25 },
		{ 163, 182, 25 },
		{ 163, 183, 25 },
		{ 163, 184, 25 },
		{ 163, 185, 25 },
		{ 163, 186, 25 },
		{ 163, 187, 25 },
		{ 163, 188, 25 },
		{ 163, 189, 25 },
		{ 163, 190, 25 },
		{ 163, 191, 25 },
		{ 163, 192, 25 },
		{ 163, 193, 25 },
		{ 163, 194, 25 },
		{ 163, 195, 25 },
		{ 163, 196, 25 },
		{ 163, 197, 25 },
		{ 163, 198, 25 },
		{ 163, 199, 25 },
		{ 163, 200, 25 },
		{ 163, 201, 25 },
		{ 163, 202, 25 },
		{ 163, 203, 25 },
		{ 163, 204, 25 },
		{ 163, 205, 25 },
		{ 163, 206, 25 },
		{ 163, 207, 25 },
		{ 163, 208, 25 },
		{ 163, 209, 25 },
		{ 164, 165, 25 },
		{ 164, 166, 25 },
		{ 164, 167, 25 },
		{ 164, 168, 25 },
		{ 164, 169, 25 },
		{ 164, 170, 25 },
		{ 164, 171, 25 },
		{ 164, 172, 25 },
		{ 164, 173, 25 },
		{ 164, 174, 25 },
		{ 164, 175, 25 },
		{ 164, 176, 25 },
		{ 164, 177, 25 },
		{ 164, 178, 25 },
		{ 164, 179, 25 },
		{ 164, 180, 25 },
		{ 164, 181, 25 },
		{ 164, 182, 25 },
		{ 164, 183, 25 },
		{ 164, 184, 25 },
		{ 164, 185, 25 },
		{ 164, 186, 25 },
		{ 164, 187, 25 },
		{ 164, 188, 25 },
		{ 164, 189, 25 },
		{ 164, 190, 25 },
		{ 164, 191, 25 },
		{ 164, 192, 25 },
		{ 164, 193, 25 },
		{ 164, 194, 25 },
		{ 164, 195, 25 },
		{ 164, 196, 25 },
		{ 164, 197, 25 },
		{ 164, 198, 25 },
		{ 164, 199, 25 },
		{ 164, 200, 25 },
		{ 164, 201, 25 },
		{ 164, 202, 25 },
		{ 164, 203, 25 },
		{ 164, 204, 25 },
		{ 164, 205, 25 },
		{ 164, 206, 25 },
		{ 164, 207, 25 },
		{ 164, 208, 25 },
		{ 164, 209, 25 },
		{ 164, 210, 25 },
		{ 165, 166, 25 },
		{ 165, 167, 25 },
		{ 165, 168, 25 },
		{ 165, 169, 25 },
		{ 165, 170, 25 },
		{ 165, 171, 25 },
		{ 165, 172, 25 },
		{ 165, 173, 25 },
		{ 165, 174, 25 },
		{ 165, 175, 25 },
		{ 165, 176, 25 },
		{ 165, 177, 25 },
		{ 165, 178, 25 },
		{ 165, 179, 25 },
		{ 165, 180, 25 },
		{ 165, 181, 25 },
		{ 165, 182, 25 },
		{ 165, 183, 25 },
		{ 165, 184, 25 },
		{ 165, 185, 25 },
		{ 165, 186, 25 },
		{ 165, 187, 25 },
		{ 165, 188, 25 },
		{ 165, 189, 25 },
		{ 165, 190, 25 },
		{ 165, 191, 25 },
		{ 165, 192, 25 },
		{ 165, 193, 25 },
		{ 165, 194, 25 },
		{ 165, 195, 25 },
		{ 165, 196, 25 },
		{ 165, 197, 25 },
		{ 165, 198, 25 },
		{ 165, 199, 25 },
		{ 165, 200, 25 },
		{ 165, 201, 25 },
		{ 165, 202, 25 },
		{ 165, 203, 25 },
		{ 165, 204, 25 },
		{ 165, 205, 25 },
		{ 165, 206, 25 },
		{ 165, 207, 25 },
		{ 165, 208, 25 },
		{ 165, 209, 25 },
		{ 165, 210, 25 },
		{ 166, 167, 25 },
		{ 166, 168, 25 },
		{ 166, 169, 25 },
		{ 166, 170, 25 },
		{ 166, 171, 25 },
		{ 166, 172, 25 },
		{ 166, 173, 25 },
		{ 166, 174, 25 },
		{ 166, 175, 25 },
		{ 166, 176, 25 },
		{ 166, 177, 25 },
		{ 166, 178, 25 },
		{ 166, 179, 25 },
		{ 166, 180, 25 },
		{ 166, 181, 25 },
		{ 166, 182, 25 },
		{ 166, 183, 25 },
		{ 166, 184, 25 },
		{ 166, 185, 25 },
		{ 166, 186, 25 },
		{ 166, 187, 25 },
		{ 166, 188, 25 },
		{ 166, 189, 25 },
		{ 166, 190, 25 },
		{ 166, 191, 25 },
		{ 166, 192, 25 },
		{ 166, 193, 25 },
		{ 166, 194, 25 },
		{ 166, 195, 25 },
		{ 166, 196, 25 },
		{ 166, 197, 25 },
		{ 166, 198, 25 },
		{ 166, 199, 25 },
		{ 166, 200, 25 },
		{ 166, 201, 25 },
		{ 166, 202, 25 },
		{ 166, 203, 25 },
		{ 166, 204, 25 },
		{ 166, 205, 25 },
		{ 166, 206, 25 },
		{ 166, 207, 25 },
		{ 166, 208, 25 },
		{ 166, 209, 25 },
		{ 166, 210, 25 },
		{ 166, 211, 25 },
		{ 167, 168, 25 },
		{ 167, 169, 25 },
		{ 167, 170, 25 },
		{ 167, 171, 25 },
		{ 167, 172, 25 },
		{ 167, 173, 25 },
		{ 167, 174, 25 },
		{ 167, 175, 25 },
		{ 167, 176, 25 },
		{ 167, 177, 25 },
		{ 167, 178, 25 },
		{ 167, 179, 25 },
		{ 167, 180, 25 },
		{ 167, 181, 25 },
		{ 167, 182, 25 },
		{ 167, 183, 25 },
		{ 167, 184, 25 },
		{ 167, 185, 25 },
		{ 167, 186, 25 },
		{ 167, 187, 25 },
		{ 167, 188, 25 },
		{ 167, 189, 25 },
		{ 167, 190, 25 },
		{ 167, 191, 25 },
		{ 167, 192, 25 },
		{ 167, 193, 25 },
		{ 167, 194, 25 },
		{ 167, 195, 25 },
		{ 167, 196, 25 },
		{ 167, 197, 25 },
		{ 167, 198, 25 },
		{ 167, 199, 25 },
		{ 167, 200, 25 },
		{ 167, 201, 25 },
		{ 167, 202, 25 },
		{ 167, 203, 25 },
		{ 167, 204, 25 },
		{ 167, 205, 25 },
		{ 167, 206, 25 },
		{ 167, 207, 25 },
		{ 167, 208, 25 },
		{ 167, 209, 25 },
		{ 167, 210, 25 },
		{ 167, 211, 25 },
		{ 168, 169, 25 },
		{ 168, 170, 25 },
		{ 168, 171, 25 },
		{ 168, 172, 25 },
		{ 168, 173, 25 },
		{ 168, 174, 25 },
		{ 168, 175, 25 },
		{ 168, 176, 25 },
		{ 168, 177, 25 },
		{ 168, 178, 25 },
		{ 168, 179, 25 },
		{ 168, 180, 25 },
		{ 168, 181, 25 },
		{ 168, 182, 25 },
		{ 168, 183, 25 },
		{ 168, 184, 25 },
		{ 168, 185, 25 },
		{ 168, 186, 25 },
		{ 168, 187, 25 },
		{ 168, 188, 25 },
		{ 168, 189, 25 },
		{ 168, 190, 25 },
		{ 168, 191, 25 },
		{ 168, 192, 25 },
		{ 168, 193, 25 },
		{ 168, 194, 25 },
		{ 168, 195, 25 },
		{ 168, 196, 25 },
		{ 168, 197, 25 },
		{ 168, 198, 25 },
		{ 168, 199, 25 },
		{ 168, 200, 25 },
		{ 168, 201, 25 },
		{ 168, 202, 25 },
		{ 168, 203, 25 },
		{ 168, 204, 25 },
		{ 168, 205, 25 },
		{ 168, 206, 25 },
		{ 168, 207, 25 },
		{ 168, 208, 25 },
		{ 168, 209, 25 },
		{ 168, 210, 25 },
		{ 168, 211, 25 },
		{ 168, 212, 25 },
		{ 169, 170, 25 },
		{ 169, 171, 25 },
		{ 169, 172, 25 },
		{ 169, 173, 25 },
		{ 169, 174, 25 },
		{ 169, 175, 25 },
		{ 169, 176, 25 },
		{ 169, 177, 25 },
		{ 169, 178, 25 },
		{ 169, 179, 25 },
		{ 169, 180, 25 },
		{ 169, 181, 25 },
		{ 169, 182, 25 },
		{ 169, 183, 25 },
		{ 169, 184, 25 },
		{ 169, 185, 25 },
		{ 169, 186, 25 },
		{ 169, 187, 25 },
		{ 169, 188, 25 },
		{ 169, 189, 25 },
		{ 169, 190, 25 },
		{ 169, 191, 25 },
		{ 169, 192, 25 },
		{ 169, 193, 25 },
		{ 169, 194, 25 },
		{ 169, 195, 25 },
		{ 169, 196, 25 },
		{ 169, 197, 25 },
		{ 169, 198, 25 },
		{ 169, 199, 25 },
		{ 169, 200, 25 },
		{ 169, 201, 25 },
		{ 169, 202, 25 },
		{ 169, 203, 25 },
		{ 169, 204, 25 },
		{ 169, 205, 25 },
		{ 169, 206, 25 },
		{ 169, 207, 25 },
		{ 169, 208, 25 },
		{ 169, 209, 25 },
		{ 169, 210, 25 },
		{ 169, 211, 25 },
		{ 169, 212, 25 },
		{ 170, 171, 25 },
		{ 170, 172, 25 },
		{ 170, 173, 25 },
		{ 170, 174, 25 },
		{ 170, 175, 25 },
		{ 170, 176, 25 },
		{ 170, 177, 25 },
		{ 170, 178, 25 },
		{ 170, 179, 25 },
		{ 170, 180, 25 },
		{ 170, 181, 25 },
		{ 170, 182, 25 },
		{ 170, 183, 25 },
		{ 170, 184, 25 },
		{ 170, 185, 25 },
		{ 170, 186, 25 },
		{ 170, 187, 25 },
		{ 170, 188, 25 },
		{ 170, 189, 25 },
		{ 170, 190, 25 },
		{ 170, 191, 25 },
		{ 170, 192, 25 },
		{ 170, 193, 25 },
		{ 170, 194, 25 },
		{ 170, 195, 25 },
		{ 170, 196, 25 },
		{ 170, 197, 25 },
		{ 170, 198, 25 },
		{ 170, 199, 25 },
		{ 170, 200, 25 },
		{ 170, 201, 25 },
		{ 170, 202, 25 },
		{ 170, 203, 25 },
		{ 170, 204, 25 },
		{ 170, 205, 25 },
		{ 170, 206, 25 },
		{ 170, 207, 25 },
		{ 170, 208, 25 },
		{ 170, 209, 25 },
		{ 170, 210, 25 },
		{ 170, 211, 25 },
		{ 170, 212, 25 },
		{ 170, 213, 25 },
		{ 171, 172, 25 },
		{ 171, 173, 25 },
		{ 171, 174, 25 },
		{ 171, 175, 25 },
		{ 171, 176, 25 },
		{ 171, 177, 25 },
		{ 171, 178, 25 },
		{ 171, 179, 25 },
		{ 171, 180, 25 },
		{ 171, 181, 25 },
		{ 171, 182, 25 },
		{ 171, 183, 25 },
		{ 171, 184, 25 },
		{ 171, 185, 25 },
		{ 171, 186, 25 },
		{ 171, 187, 25 },
		{ 171, 188, 25 },
		{ 171, 189, 25 },
		{ 171, 190, 25 },
		{ 171, 191, 25 },
		{ 171, 192, 25 },
		{ 171, 193, 25 },
		{ 171, 194, 25 },
		{ 171, 195, 25 },
		{ 171, 196, 25 },
		{ 171, 197, 25 },
		{ 171, 198, 25 },
		{ 171, 199, 25 },
		{ 171, 200, 25 },
		{ 171, 201, 25 },
		{ 171, 202, 25 },
		{ 171, 203, 25 },
		{ 171, 204, 25 },
		{ 171, 205, 25 },
		{ 171, 206, 25 },
		{ 171, 207, 25 },
		{ 171, 208, 25 },
		{ 171, 209, 25 },
		{ 171, 210, 25 },
		{ 171, 211, 25 },
		{ 171, 212, 25 },
		{ 171, 213, 25 },
		{ 172, 173, 25 },
		{ 172, 174, 25 },
		{ 172, 175, 25 },
		{ 172, 176, 25 },
		{ 172, 177, 25 },
		{ 172, 178, 25 },
		{ 172, 179, 25 },
		{ 172, 180, 25 },
		{ 172, 181, 25 },
		{ 172, 182, 25 },
		{ 172, 183, 25 },
		{ 172, 184, 25 },
		{ 172, 185, 25 },
		{ 172, 186, 25 },
		{ 172, 187, 25 },
		{ 172, 188, 25 },
		{ 172, 189, 25 },
		{ 172, 190, 25 },
		{ 172, 191, 25 },
		{ 172, 192, 25 },
		{ 172, 193, 25 },
		{ 172, 194, 25 },
		{ 172, 195, 25 },
		{ 172, 196, 25 },
		{ 172, 197, 25 },
		{ 172, 198, 25 },
		{ 172, 199, 25 },
		{ 172, 200, 25 },
		{ 172, 201, 25 },
		{ 172, 202, 25 },
		{ 172, 203, 25 },
		{ 172, 204, 25 },
		{ 172, 205, 25 },
		{ 172, 206, 25 },
		{ 172, 207, 25 },
		{ 172, 208, 25 },
		{ 172, 209, 25 },
		{ 172, 210, 25 },
		{ 172, 211, 25 },
		{ 172, 212, 25 },
		{ 172, 213, 25 },
		{ 172, 214, 25 },
		{ 173, 174, 25 },
		{ 173, 175, 25 },
		{ 173, 176, 25 },
		{ 173, 177, 25 },
		{ 173, 178, 25 },
		{ 173, 179, 25 },
		{ 173, 180, 25 },
		{ 173, 181, 25 },
		{ 173, 182, 25 },
		{ 173, 183, 25 },
		{ 173, 184, 25 },
		{ 173, 185, 25 },
		{ 173, 186, 25 },
		{ 173, 187, 25 },
		{ 173, 188, 25 },
		{ 173, 189, 25 },
		{ 173, 190, 25 },
		{ 173, 191, 25 },
		{ 173, 192, 25 },
		{ 173, 193, 25 },
		{ 173, 194, 25 },
		{ 173, 195, 25 },
		{ 173, 196, 25 },
		{ 173, 197, 25 },
		{ 173, 198, 25 },
		{ 173, 199, 25 },
		{ 173, 200, 25 },
		{ 173, 201, 25 },
		{ 173, 202, 25 },
		{ 173, 203, 25 },
		{ 173, 204, 25 },
		{ 173, 205, 25 },
		{ 173, 206, 25 },
		{ 173, 207, 25 },
		{ 173, 208, 25 },
		{ 173, 209, 25 },
		{ 173, 210, 25 },
		{ 173, 211, 25 },
		{ 173, 212, 25 },
		{ 173, 213, 25 },
		{ 173, 214, 25 },
		{ 174, 175, 25 },
		{ 174, 176, 25 },
		{ 174, 177, 25 },
		{ 174, 178, 25 },
		{ 174, 179, 25 },
		{ 174, 180, 25 },
		{ 174, 181, 25 },
		{ 174, 182, 25 },
		{ 174, 183, 25 },
		{ 174, 184, 25 },
		{ 174, 185, 25 },
		{ 174, 186, 25 },
		{ 174, 187, 25 },
		{ 174, 188, 25 },
		{ 174, 189, 25 },
		{ 174, 190, 25 },
		{ 174, 191, 25 },
		{ 174, 192, 25 },
		{ 174, 193, 25 },
		{ 174, 194, 25 },
		{ 174, 195, 25 },
		{ 174, 196, 25 },
		{ 174, 197, 25 },
		{ 174, 198, 25 },
		{ 174, 199, 25 },
		{ 174, 200, 25 },
		{ 174, 201, 25 },
		{ 174, 202, 25 },
		{ 174, 203, 25 },
		{ 174, 204, 25 },
		{ 174, 205, 25 },
		{ 174, 206, 25 },
		{ 174, 207, 25 },
		{ 174, 208, 25 },
		{ 174, 209, 25 },
		{ 174, 210, 25 },
		{ 174, 211, 25 },
		{ 174, 212, 25 },
		{ 174, 213, 25 },
		{ 174, 214, 25 },
		{ 174, 215, 25 },
		{ 175, 176, 25 },
		{ 175, 177, 25 },
		{ 175, 178, 25 },
		{ 175, 179, 25 },
		{ 175, 180, 25 },
		{ 175, 181, 25 },
		{ 175, 182, 25 },
		{ 175, 183, 25 },
		{ 175, 184, 25 },
		{ 175, 185, 25 },
		{ 175, 186, 25 },
		{ 175, 187, 25 },
		{ 175, 188, 25 },
		{ 175, 189, 25 },
		{ 175, 190, 25 },
		{ 175, 191, 25 },
		{ 175, 192, 25 },
		{ 175, 193, 25 },
		{ 175, 194, 25 },
		{ 175, 195, 25 },
		{ 175, 196, 25 },
		{ 175, 197, 25 },
		{ 175, 198, 25 },
		{ 175, 199, 25 },
		{ 175, 200, 25 },
		{ 175, 201, 25 },
		{ 175, 202, 25 },
		{ 175, 203, 25 },
		{ 175, 204, 25 },
		{ 175, 205, 25 },
		{ 175, 206, 25 },
		{ 175, 207, 25 },
		{ 175, 208, 25 },
		{ 175, 209, 25 },
		{ 175, 210, 25 },
		{ 175, 211, 25 },
		{ 175, 212, 25 },
		{ 175, 213, 25 },
		{ 175, 214, 25 },
		{ 175, 215, 25 },
		{ 176, 177, 25 },
		{ 176, 178, 25 },
		{ 176, 179, 25 },
		{ 176, 180, 25 },
		{ 176, 181, 25 },
		{ 176, 182, 25 },
		{ 176, 183, 25 },
		{ 176, 184, 25 },
		{ 176, 185, 25 },
		{ 176, 186, 25 },
		{ 176, 187, 25 },
		{ 176, 188, 25 },
		{ 176, 189, 25 },
		{ 176, 190, 25 },
		{ 176, 191, 25 },
		{ 176, 192, 25 },
		{ 176, 193, 25 },
		{ 176, 194, 25 },
		{ 176, 195, 25 },
		{ 176, 196, 25 },
		{ 176, 197, 25 },
		{ 176, 198, 25 },
		{ 176, 199, 25 },
		{ 176, 200, 25 },
		{ 176, 201, 25 },
		{ 176, 202, 25 },
		{ 176, 203, 25 },
		{ 176, 204, 25 },
		{ 176, 205, 25 },
		{ 176, 206, 25 },
		{ 176, 207, 25 },
		{ 176, 208, 25 },
		{ 176, 209, 25 },
		{ 176, 210, 25 },
		{ 176, 211, 25 },
		{ 176, 212, 25 },
		{ 176, 213, 25 },
		{ 176, 214, 25 },
		{ 176, 215, 25 },
		{ 176, 216, 25 },
		{ 177, 178, 25 },
		{ 177, 179, 25 },
		{ 177, 180, 25 },
		{ 177, 181, 25 },
		{ 177, 182, 25 },
		{ 177, 183, 25 },
		{ 177, 184, 25 },
		{ 177, 185, 25 },
		{ 177, 186, 25 },
		{ 177, 187, 25 },
		{ 177, 188, 25 },
		{ 177, 189, 25 },
		{ 177, 190, 25 },
		{ 177, 191, 25 },
		{ 177, 192, 25 },
		{ 177, 193, 25 },
		{ 177, 194, 25 },
		{ 177, 195, 25 },
		{ 177, 196, 25 },
		{ 177, 197, 25 },
		{ 177, 198, 25 },
		{ 177, 199, 25 },
		{ 177, 200, 25 },
		{ 177, 201, 25 },
		{ 177, 202, 25 },
		{ 177, 203, 25 },
		{ 177, 204, 25 },
		{ 177, 205, 25 },
		{ 177, 206, 25 },
		{ 177, 207, 25 },
		{ 177, 208, 25 },
		{ 177, 209, 25 },
		{ 177, 210, 25 },
		{ 177, 211, 25 },
		{ 177, 212, 25 },
		{ 177, 213, 25 },
		{ 177, 214, 25 },
		{ 177, 215, 25 },
		{ 177, 216, 25 },
		{ 178, 179, 25 },
		{ 178, 180, 25 },
		{ 178, 181, 25 },
		{ 178, 182, 25 },
		{ 178, 183, 25 },
		{ 178, 184, 25 },
		{ 178, 185, 25 },
		{ 178, 186, 25 },
		{ 178, 187, 25 },
		{ 178, 188, 25 },
		{ 178, 189, 25 },
		{ 178, 190, 25 },
		{ 178, 191, 25 },
		{ 178, 192, 25 },
		{ 178, 193, 25 },
		{ 178, 194, 25 },
		{ 178, 195, 25 },
		{ 178, 196, 25 },
		{ 178, 197, 25 },
		{ 178, 198, 25 },
		{ 178, 199, 25 },
		{ 178, 200, 25 },
		{ 178, 201, 25 },
		{ 178, 202, 25 },
		{ 178, 203, 25 },
		{ 178, 204, 25 },
		{ 178, 205, 25 },
		{ 178, 206, 25 },
		{ 178, 207, 25 },
		{ 178, 208, 25 },
		{ 178, 209, 25 },
		{ 178, 210, 25 },
		{ 178, 211, 25 },
		{ 178, 212, 25 },
		{ 178, 213, 25 },
		{ 178, 214, 25 },
		{ 178, 215, 25 },
		{ 178, 216, 25 },
		{ 178, 217, 25 },
		{ 179, 180, 25 },
		{ 179, 181, 25 },
		{ 179, 182, 25 },
		{ 179, 183, 25 },
		{ 179, 184, 25 },
		{ 179, 185, 25 },
		{ 179, 186, 25 },
		{ 179, 187, 25 },
		{ 179, 188, 25 },
		{ 179, 189, 25 },
		{ 179, 190, 25 },
		{ 179, 191, 25 },
		{ 179, 192, 25 },
		{ 179, 193, 25 },
		{ 179, 194, 25 },
		{ 179, 195, 25 },
		{ 179, 196, 25 },
		{ 179, 197, 25 },
		{ 179, 198, 25 },
		{ 179, 199, 25 },
		{ 179, 200, 25 },
		{ 179, 201, 25 },
		{ 179, 202, 25 },
		{ 179, 203, 25 },
		{ 179, 204, 25 },
		{ 179, 205, 25 },
		{ 179, 206, 25 },
		{ 179, 207, 25 },
		{ 179, 208, 25 },
		{ 179, 209, 25 },
		{ 179, 210, 25 },
		{ 179, 211, 25 },
		{ 179, 212, 25 },
		{ 179, 213, 25 },
		{ 179, 214, 25 },
		{ 179, 215, 25 },
		{ 179, 216, 25 },
		{ 179, 217, 25 },
		{ 180, 181, 25 },
		{ 180, 182, 25 },
		{ 180, 183, 25 },
		{ 180, 184, 25 },
		{ 180, 185, 25 },
		{ 180, 186, 25 },
		{ 180, 187, 25 },
		{ 180, 188, 25 },
		{ 180, 189, 25 },
		{ 180, 190, 25 },
		{ 180, 191, 25 },
		{ 180, 192, 25 },
		{ 180, 193, 25 },
		{ 180, 194, 25 },
		{ 180, 195, 25 },
		{ 180, 196, 25 },
		{ 180, 197, 25 },
		{ 180, 198, 25 },
		{ 180, 199, 25 },
		{ 180, 200, 25 },
		{ 180, 201, 25 },
		{ 180, 202, 25 },
		{ 180, 203, 25 },
		{ 180, 204, 25 },
		{ 180, 205, 25 },
		{ 180, 206, 25 },
		{ 180, 207, 25 },
		{ 180, 208, 25 },
		{ 180, 209, 25 },
		{ 180, 210, 25 },
		{ 180, 211, 25 },
		{ 180, 212, 25 },
		{ 180, 213, 25 },
		{ 180, 214, 25 },
		{ 180, 215, 25 },
		{ 180, 216, 25 },
		{ 180, 217, 25 },
		{ 180, 218, 25 },
		{ 181, 182, 25 },
		{ 181, 183, 25 },
		{ 181, 184, 25 },
		{ 181, 185, 25 },
		{ 181, 186, 25 },
		{ 181, 187, 25 },
		{ 181, 188, 25 },
		{ 181, 189, 25 },
		{ 181, 190, 25 },
		{ 181, 191, 25 },
		{ 181, 192, 25 },
		{ 181, 193, 25 },
		{ 181, 194, 25 },
		{ 181, 195, 25 },
		{ 181, 196, 25 },
		{ 181, 197, 25 },
		{ 181, 198, 25 },
		{ 181, 199, 25 },
		{ 181, 200, 25 },
		{ 181, 201, 25 },
		{ 181, 202, 25 },
		{ 181, 203, 25 },
		{ 181, 204, 25 },
		{ 181, 205, 25 },
		{ 181, 206, 25 },
		{ 181, 207, 25 },
		{ 181, 208, 25 },
		{ 181, 209, 25 },
		{ 181, 210, 25 },
		{ 181, 211, 25 },
		{ 181, 212, 25 },
		{ 181, 213, 25 },
		{ 181, 214, 25 },
		{ 181, 215, 25 },
		{ 181, 216, 25 },
		{ 181, 217, 25 },
		{ 181, 218, 25 },
		{ 182, 183, 25 },
		{ 182, 184, 25 },
		{ 182, 185, 25 },
		{ 182, 186, 25 },
		{ 182, 187, 25 },
		{ 182, 188, 25 },
		{ 182, 189, 25 },
		{ 182, 190, 25 },
		{ 182, 191, 25 },
		{ 182, 192, 25 },
		{ 182, 193, 25 },
		{ 182, 194, 25 },
		{ 182, 195, 25 },
		{ 182, 196, 25 },
		{ 182, 197, 25 },
		{ 182, 198, 25 },
		{ 182, 199, 25 },
		{ 182, 200, 25 },
		{ 182, 201, 25 },
		{ 182, 202, 25 },
		{ 182, 203, 25 },
		{ 182, 204, 25 },
		{ 182, 205, 25 },
		{ 182, 206, 25 },
		{ 182, 207, 25 },
		{ 182, 208, 25 },
		{ 182, 209, 25 },
		{ 182, 210, 25 },
		{ 182, 211, 25 },
		{ 182, 212, 25 },
		{ 182, 213, 25 },
		{ 182, 214, 25 },
		{ 182, 215, 25 },
		{ 182, 216, 25 },
		{ 182, 217, 25 },
		{ 182, 218, 25 },
		{ 182, 219, 25 },
		{ 183, 184, 25 },
		{ 183, 185, 25 },
		{ 183, 186, 25 },
		{ 183, 187, 25 },
		{ 183, 188, 25 },
		{ 183, 189, 25 },
		{ 183, 190, 25 },
		{ 183, 191, 25 },
		{ 183, 192, 25 },
		{ 183, 193, 25 },
		{ 183, 194, 25 },
		{ 183, 195, 25 },
		{ 183, 196, 25 },
		{ 183, 197, 25 },
		{ 183, 198, 25 },
		{ 183, 199, 25 },
		{ 183, 200, 25 },
		{ 183, 201, 25 },
		{ 183, 202, 25 },
		{ 183, 203, 25 },
		{ 183, 204, 25 },
		{ 183, 205, 25 },
		{ 183, 206, 25 },
		{ 183, 207, 25 },
		{ 183, 208, 25 },
		{ 183, 209, 25 },
		{ 183, 210, 25 },
		{ 183, 211, 25 },
		{ 183, 212, 25 },
		{ 183, 213, 25 },
		{ 183, 214, 25 },
		{ 183, 215, 25 },
		{ 183, 216, 25 },
		{ 183, 217, 25 },
		{ 183, 218, 25 },
		{ 183, 219, 25 },
		{ 184, 185, 25 },
		{ 184, 186, 25 },
		{ 184, 187, 25 },
		{ 184, 188, 25 },
		{ 184, 189, 25 },
		{ 184, 190, 25 },
		{ 184, 191, 25 },
		{ 184, 192, 25 },
		{ 184, 193, 25 },
		{ 184, 194, 25 },
		{ 184, 195, 25 },
		{ 184, 196, 25 },
		{ 184, 197, 25 },
		{ 184, 198, 25 },
		{ 184, 199, 25 },
		{ 184, 200, 25 },
		{ 184, 201, 25 },
		{ 184, 202, 25 },
		{ 184, 203, 25 },
		{ 184, 204, 25 },
		{ 184, 205, 25 },
		{ 184, 206, 25 },
		{ 184, 207, 25 },
		{ 184, 208, 25 },
		{ 184, 209, 25 },
		{ 184, 210, 25 },
		{ 184, 211, 25 },
		{ 184, 212, 25 },
		{ 184, 213, 25 },
		{ 184, 214, 25 },
		{ 184, 215, 25 },
		{ 184, 216, 25 },
		{ 184, 217, 25 },
		{ 184, 218, 25 },
		{ 184, 219, 25 },
		{ 184, 220, 25 },
		{ 185, 186, 25 },
		{ 185, 187, 25 },
		{ 185, 188, 25 },
		{ 185, 189, 25 },
		{ 185, 190, 25 },
		{ 185, 191, 25 },
		{ 185, 192, 25 },
		{ 185, 193, 25 },
		{ 185, 194, 25 },
		{ 185, 195, 25 },
		{ 185, 196, 25 },
		{ 185, 197, 25 },
		{ 185, 198, 25 },
		{ 185, 199, 25 },
		{ 185, 200, 25 },
		{ 185, 201, 25 },
		{ 185, 202, 25 },
		{ 185, 203, 25 },
		{ 185, 204, 25 },
		{ 185, 205, 25 },
		{ 185, 206, 25 },
		{ 185, 207, 25 },
		{ 185, 208, 25 },
		{ 185, 209, 25 },
		{ 185, 210, 25 },
		{ 185, 211, 25 },
		{ 185, 212, 25 },
		{ 185, 213, 25 },
		{ 185, 214, 25 },
		{ 185, 215, 25 },
		{ 185, 216, 25 },
		{ 185, 217, 25 },
		{ 185, 218, 25 },
		{ 185, 219, 25 },
		{ 185, 220, 25 },
		{ 186, 187, 25 },
		{ 186, 188, 25 },
		{ 186, 189, 25 },
		{ 186, 190, 25 },
		{ 186, 191, 25 },
		{ 186, 192, 25 },
		{ 186, 193, 25 },
		{ 186, 194, 25 },
		{ 186, 195, 25 },
		{ 186, 196, 25 },
		{ 186, 197, 25 },
		{ 186, 198, 25 },
		{ 186, 199, 25 },
		{ 186, 200, 25 },
		{ 186, 201, 25 },
		{ 186, 202, 25 },
		{ 186, 203, 25 },
		{ 186, 204, 25 },
		{ 186, 205, 25 },
		{ 186, 206, 25 },
		{ 186, 207, 25 },
		{ 186, 208, 25 },
		{ 186, 209, 25 },
		{ 186, 210, 25 },
		{ 186, 211, 25 },
		{ 186, 212, 25 },
		{ 186, 213, 25 },
		{ 186, 214, 25 },
		{ 186, 215, 25 },
		{ 186, 216, 25 },
		{ 186, 217, 25 },
		{ 186, 218, 25 },
		{ 186, 219, 25 },
		{ 186, 220, 25 },
		{ 186, 221, 25 },
		{ 187, 188, 25 },
		{ 187, 189, 25 },
		{ 187, 190, 25 },
		{ 187, 191, 25 },
		{ 187, 192, 25 },
		{ 187, 193, 25 },
		{ 187, 194, 25 },
		{ 187, 195, 25 },
		{ 187, 196, 25 },
		{ 187, 197, 25 },
		{ 187, 198, 25 },
		{ 187, 199, 25 },
		{ 187, 200, 25 },
		{ 187, 201, 25 },
		{ 187, 202, 25 },
		{ 187, 203, 25 },
		{ 187, 204, 25 },
		{ 187, 205, 25 },
		{ 187, 206, 25 },
		{ 187, 207, 25 },
		{ 187, 208, 25 },
		{ 187, 209, 25 },
		{ 187, 210, 25 },
		{ 187, 211, 25 },
		{ 187, 212, 25 },
		{ 187, 213, 25 },
		{ 187, 214, 25 },
		{ 187, 215, 25 },
		{ 187, 216, 25 },
		{ 187, 217, 25 },
		{ 187, 218, 25 },
		{ 187, 219, 25 },
		{ 187, 220, 25 },
		{ 187, 221, 25 },
		{ 188, 189, 25 },
		{ 188, 190, 25 },
		{ 188, 191, 25 },
		{ 188, 192, 25 },
		{ 188, 193, 25 },
		{ 188, 194, 25 },
		{ 188, 195, 25 },
		{ 188, 196, 25 },
		{ 188, 197, 25 },
		{ 188, 198, 25 },
		{ 188, 199, 25 },
		{ 188, 200, 25 },
		{ 188, 201, 25 },
		{ 188, 202, 25 },
		{ 188, 203, 25 },
		{ 188, 204, 25 },
		{ 188, 205, 25 },
		{ 188, 206, 25 },
		{ 188, 207, 25 },
		{ 188, 208, 25 },
		{ 188, 209, 25 },
		{ 188, 210, 25 },
		{ 188, 211, 25 },
		{ 188, 212, 25 },
		{ 188, 213, 25 },
		{ 188, 214, 25 },
		{ 188, 215, 25 },
		{ 188, 216, 25 },
		{ 188, 217, 25 },
		{ 188, 218, 25 },
		{ 188, 219, 25 },
		{ 188, 220, 25 },
		{ 188, 221, 25 },
		{ 188, 222, 25 },
		{ 189, 190, 25 },
		{ 189, 191, 25 },
		{ 189, 192, 25 },
		{ 189, 193, 25 },
		{ 189, 194, 25 },
		{ 189, 195, 25 },
		{ 189, 196, 25 },
		{ 189, 197, 25 },
		{ 189, 198, 25 },
		{ 189, 199, 25 },
		{ 189, 200, 25 },
		{ 189, 201, 25 },
		{ 189, 202, 25 },
		{ 189, 203, 25 },
		{ 189, 204, 25 },
		{ 189, 205, 25 },
		{ 189, 206, 25 },
		{ 189, 207, 25 },
		{ 189, 208, 25 },
		{ 189, 209, 25 },
		{ 189, 210, 25 },
		{ 189, 211, 25 },
		{ 189, 212, 25 },
		{ 189, 213, 25 },
		{ 189, 214, 25 },
		{ 189, 215, 25 },
		{ 189, 216, 25 },
		{ 189, 217, 25 },
		{ 189, 218, 25 },
		{ 189, 219, 25 },
		{ 189, 220, 25 },
		{ 189, 221, 25 },
		{ 189, 222, 25 },
		{ 190, 191, 25 },
		{ 190, 192, 25 },
		{ 190, 193, 25 },
		{ 190, 194, 25 },
		{ 190, 195, 25 },
		{ 190, 196, 25 },
		{ 190, 197, 25 },
		{ 190, 198, 25 },
		{ 190, 199, 25 },
		{ 190, 200, 25 },
		{ 190, 201, 25 },
		{ 190, 202, 25 },
		{ 190, 203, 25 },
		{ 190, 204, 25 },
		{ 190, 205, 25 },
		{ 190, 206, 25 },
		{ 190, 207, 25 },
		{ 190, 208, 25 },
		{ 190, 209, 25 },
		{ 190, 210, 25 },
		{ 190, 211, 25 },
		{ 190, 212, 25 },
		{ 190, 213, 25 },
		{ 190, 214, 25 },
		{ 190, 215, 25 },
		{ 190, 216, 25 },
		{ 190, 217, 25 },
		{ 190, 218, 25 },
		{ 190, 219, 25 },
		{ 190, 220, 25 },
		{ 190, 221, 25 },
		{ 190, 222, 25 },
		{ 190, 223, 25 },
		{ 191, 192, 25 },
		{ 191, 193, 25 },
		{ 191, 194, 25 },
		{ 191, 195, 25 },
		{ 191, 196, 25 },
		{ 191, 197, 25 },
		{ 191, 198, 25 },
		{ 191, 199, 25 },
		{ 191, 200, 25 },
		{ 191, 201, 25 },
		{ 191, 202, 25 },
		{ 191, 203, 25 },
		{ 191, 204, 25 },
		{ 191, 205, 25 },
		{ 191, 206, 25 },
		{ 191, 207, 25 },
		{ 191, 208, 25 },
		{ 191, 209, 25 },
		{ 191, 210, 25 },
		{ 191, 211, 25 },
		{ 191, 212, 25 },
		{ 191, 213, 25 },
		{ 191, 214, 25 },
		{ 191, 215, 25 },
		{ 191, 216, 25 },
		{ 191, 217, 25 },
		{ 191, 218, 25 },
		{ 191, 219, 25 },
		{ 191, 220, 25 },
		{ 191, 221, 25 },
		{ 191, 222, 25 },
		{ 191, 223, 25 },
		{ 192, 193, 25 },
		{ 192, 194, 25 },
		{ 192, 195, 25 },
		{ 192, 196, 25 },
		{ 192, 197, 25 },
		{ 192, 198, 25 },
		{ 192, 199, 25 },
		{ 192, 200, 25 },
		{ 192, 201, 25 },
		{ 192, 202, 25 },
		{ 192, 203, 25 },
		{ 192, 204, 25 },
		{ 192, 205, 25 },
		{ 192, 206, 25 },
		{ 192, 207, 25 },
		{ 192, 208, 25 },
		{ 192, 209, 25 },
		{ 192, 210, 25 },
		{ 192, 211, 25 },
		{ 192, 212, 25 },
		{ 192, 213, 25 },
		{ 192, 214, 25 },
		{ 192, 215, 25 },
		{ 192, 216, 25 },
		{ 192, 217, 25 },
		{ 192, 218, 25 },
		{ 192, 219, 25 },
		{ 192, 220, 25 },
		{ 192, 221, 25 },
		{ 192, 222, 25 },
		{ 192, 223, 25 },
		{ 192, 224, 25 },
		{ 193, 194, 25 },
		{ 193, 195, 25 },
		{ 193, 196, 25 },
		{ 193, 197, 25 },
		{ 193, 198, 25 },
		{ 193, 199, 25 },
		{ 193, 200, 25 },
		{ 193, 201, 25 },
		{ 193, 202, 25 },
		{ 193, 203, 25 },
		{ 193, 204, 25 },
		{ 193, 205, 25 },
		{ 193, 206, 25 },
		{ 193, 207, 25 },
		{ 193, 208, 25 },
		{ 193, 209, 25 },
		{ 193, 210, 25 },
		{ 193, 211, 25 },
		{ 193, 212, 25 },
		{ 193, 213, 25 },
		{ 193, 214, 25 },
		{ 193, 215, 25 },
		{ 193, 216, 25 },
		{ 193, 217, 25 },
		{ 193, 218, 25 },
		{ 193, 219, 25 },
		{ 193, 220, 25 },
		{ 193, 221, 25 },
		{ 193, 222, 25 },
		{ 193, 223, 25 },
		{ 193, 224, 25 },
		{ 194, 195, 25 },
		{ 194, 196, 25 },
		{ 194, 197, 25 },
		{ 194, 198, 25 },
		{ 194, 199, 25 },
		{ 194, 200, 25 },
		{ 194, 201, 25 },
		{ 194, 202, 25 },
		{ 194, 203, 25 },
		{ 194, 204, 25 },
		{ 194, 205, 25 },
		{ 194, 206, 25 },
		{ 194, 207, 25 },
		{ 194, 208, 25 },
		{ 194, 209, 25 },
		{ 194, 210, 25 },
		{ 194, 211, 25 },
		{ 194, 212, 25 },
		{ 194, 213, 25 },
		{ 194, 214, 25 },
		{ 194, 215, 25 },
		{ 194, 216, 25 },
		{ 194, 217, 25 },
		{ 194, 218, 25 },
		{ 194, 219, 25 },
		{ 194, 220, 25 },
		{ 194, 221, 25 },
		{ 194, 222, 25 },
		{ 194, 223, 25 },
		{ 194, 224, 25 },
		{ 194, 225, 25 },
		{ 195, 196, 25 },
		{ 195, 197, 25 },
		{ 195, 198, 25 },
		{ 195, 199, 25 },
		{ 195, 200, 25 },
		{ 195, 201, 25 },
		{ 195, 202, 25 },
		{ 195, 203, 25 },
		{ 195, 204, 25 },
		{ 195, 205, 25 },
		{ 195, 206, 25 },
		{ 195, 207, 25 },
		{ 195, 208, 25 },
		{ 195, 209, 25 },
		{ 195, 210, 25 },
		{ 195, 211, 25 },
		{ 195, 212, 25 },
		{ 195, 213, 25 },
		{ 195, 214, 25 },
		{ 195, 215, 25 },
		{ 195, 216, 25 },
		{ 195, 217, 25 },
		{ 195, 218, 25 },
		{ 195, 219, 25 },
		{ 195, 220, 25 },
		{ 195, 221, 25 },
		{ 195, 222, 25 },
		{ 195, 223, 25 },
		{ 195, 224, 25 },
		{ 195, 225, 25 },
		{ 196, 197, 25 },
		{ 196, 198, 25 },
		{ 196, 199, 25 },
		{ 196, 200, 25 },
		{ 196, 201, 25 },
		{ 196, 202, 25 },
		{ 196, 203, 25 },
		{ 196, 204, 25 },
		{ 196, 205, 25 },
		{ 196, 206, 25 },
		{ 196, 207, 25 },
		{ 196, 208, 25 },
		{ 196, 209, 25 },
		{ 196, 210, 25 },
		{ 196, 211, 25 },
		{ 196, 212, 25 },
		{ 196, 213, 25 },
		{ 196, 214, 25 },
		{ 196, 215, 25 },
		{ 196, 216, 25 },
		{ 196, 217, 25 },
		{ 196, 218, 25 },
		{ 196, 219, 25 },
		{ 196, 220, 25 },
		{ 196, 221, 25 },
		{ 196, 222, 25 },
		{ 196, 223, 25 },
		{ 196, 224, 25 },
		{ 196, 225, 25 },
		{ 196, 226, 25 },
		{ 197, 198, 25 },
		{ 197, 199, 25 },
		{ 197, 200, 25 },
		{ 197, 201, 25 },
		{ 197, 202, 25 },
		{ 197, 203, 25 },
		{ 197, 204, 25 },
		{ 197, 205, 25 },
		{ 197, 206, 25 },
		{ 197, 207, 25 },
		{ 197, 208, 25 },
		{ 197, 209, 25 },
		{ 197, 210, 25 },
		{ 197, 211, 25 },
		{ 197, 212, 25 },
		{ 197, 213, 25 },
		{ 197, 214, 25 },
		{ 197, 215, 25 },
		{ 197, 216, 25 },
		{ 197, 217, 25 },
		{ 197, 218, 25 },
		{ 197, 219, 25 },
		{ 197, 220, 25 },
		{ 197, 221, 25 },
		{ 197, 222, 25 },
		{ 197, 223, 25 },
		{ 197, 224, 25 },
		{ 197, 225, 25 },
		{ 197, 226, 25 },
		{ 198, 199, 25 },
		{ 198, 200, 25 },
		{ 198, 201, 25 },
		{ 198, 202, 25 },
		{ 198, 203, 25 },
		{ 198, 204, 25 },
		{ 198, 205, 25 },
		{ 198, 206, 25 },
		{ 198, 207, 25 },
		{ 198, 208, 25 },
		{ 198, 209, 25 },
		{ 198, 210, 25 },
		{ 198, 211, 25 },
		{ 198, 212, 25 },
		{ 198, 213, 25 },
		{ 198, 214, 25 },
		{ 198, 215, 25 },
		{ 198, 216, 25 },
		{ 198, 217, 25 },
		{ 198, 218, 25 },
		{ 198, 219, 25 },
		{ 198, 220, 25 },
		{ 198, 221, 25 },
		{ 198, 222, 25 },
		{ 198, 223, 25 },
		{ 198, 224, 25 },
		{ 198, 225, 25 },
		{ 198, 226, 25 },
		{ 198, 227, 25 },
		{ 199, 200, 25 },
		{ 199, 201, 25 },
		{ 199, 202, 25 },
		{ 199, 203, 25 },
		{ 199, 204, 25 },
		{ 199, 205, 25 },
		{ 199, 206, 25 },
		{ 199, 207, 25 },
		{ 199, 208, 25 },
		{ 199, 209, 25 },
		{ 199, 210, 25 },
		{ 199, 211, 25 },
		{ 199, 212, 25 },
		{ 199, 213, 25 },
		{ 199, 214, 25 },
		{ 199, 215, 25 },
		{ 199, 216, 25 },
		{ 199, 217, 25 },
		{ 199, 218, 25 },
		{ 199, 219, 25 },
		{ 199, 220, 25 },
		{ 199, 221, 25 },
		{ 199, 222, 25 },
		{ 199, 223, 25 },
		{ 199, 224, 25 },
		{ 199, 225, 25 },
		{ 199, 226, 25 },
		{ 199, 227, 25 },
		{ 200, 201, 25 },
		{ 200, 202, 25 },
		{ 200, 203, 25 },
		{ 200, 204, 25 },
		{ 200, 205, 25 },
		{ 200, 206, 25 },
		{ 200, 207, 25 },
		{ 200, 208, 25 },
		{ 200, 209, 25 },
		{ 200, 210, 25 },
		{ 200, 211, 25 },
		{ 200, 212, 25 },
		{ 200, 213, 25 },
		{ 200, 214, 25 },
		{ 200, 215, 25 },
		{ 200, 216, 25 },
		{ 200, 217, 25 },
		{ 200, 218, 25 },
		{ 200, 219, 25 },
		{ 200, 220, 25 },
		{ 200, 221, 25 },
		{ 200, 222, 25 },
		{ 200, 223, 25 },
		{ 200, 224, 25 },
		{ 200, 225, 25 },
		{ 200, 226, 25 },
		{ 200, 227, 25 },
		{ 200, 228, 25 },
		{ 201, 202, 25 },
		{ 201, 203, 25 },
		{ 201, 204, 25 },
		{ 201, 205, 25 },
		{ 201, 206, 25 },
		{ 201, 207, 25 },
		{ 201, 208, 25 },
		{ 201, 209, 25 },
		{ 201, 210, 25 },
		{ 201, 211, 25 },
		{ 201, 212, 25 },
		{ 201, 213, 25 },
		{ 201, 214, 25 },
		{ 201, 215, 25 },
		{ 201, 216, 25 },
		{ 201, 217, 25 },
		{ 201, 218, 25 },
		{ 201, 219, 25 },
		{ 201, 220, 25 },
		{ 201, 221, 25 },
		{ 201, 222, 25 },
		{ 201, 223, 25 },
		{ 201, 224, 25 },
		{ 201, 225, 25 },
		{ 201, 226, 25 },
		{ 201, 227, 25 },
		{ 201, 228, 25 },
		{ 202, 203, 25 },
		{ 202, 204, 25 },
		{ 202, 205, 25 },
		{ 202, 206, 25 },
		{ 202, 207, 25 },
		{ 202, 208, 25 },
		{ 202, 209, 25 },
		{ 202, 210, 25 },
		{ 202, 211, 25 },
		{ 202, 212, 25 },
		{ 202, 213, 25 },
		{ 202, 214, 25 },
		{ 202, 215, 25 },
		{ 202, 216, 25 },
		{ 202, 217, 25 },
		{ 202, 218, 25 },
		{ 202, 219, 25 },
		{ 202, 220, 25 },
		{ 202, 221, 25 },
		{ 202, 222, 25 },
		{ 202, 223, 25 },
		{ 202, 224, 25 },
		{ 202, 225, 25 },
		{ 202, 226, 25 },
		{ 202, 227, 25 },
		{ 202, 228, 25 },
		{ 202, 229, 25 },
		{ 203, 204, 25 },
		{ 203, 205, 25 },
		{ 203, 206, 25 },
		{ 203, 207, 25 },
		{ 203, 208, 25 },
		{ 203, 209, 25 },
		{ 203, 210, 25 },
		{ 203, 211, 25 },
		{ 203, 212, 25 },
		{ 203, 213, 25 },
		{ 203, 214, 25 },
		{ 203, 215, 25 },
		{ 203, 216, 25 },
		{ 203, 217, 25 },
		{ 203, 218, 25 },
		{ 203, 219, 25 },
		{ 203, 220, 25 },
		{ 203, 221, 25 },
		{ 203, 222, 25 },
		{ 203, 223, 25 },
		{ 203, 224, 25 },
		{ 203, 225, 25 },
		{ 203, 226, 25 },
		{ 203, 227, 25 },
		{ 203, 228, 25 },
		{ 203, 229, 25 },
		{ 204, 205, 25 },
		{ 204, 206, 25 },
		{ 204, 207, 25 },
		{ 204, 208, 25 },
		{ 204, 209, 25 },
		{ 204, 210, 25 },
		{ 204, 211, 25 },
		{ 204, 212, 25 },
		{ 204, 213, 25 },
		{ 204, 214, 25 },
		{ 204, 215, 25 },
		{ 204, 216, 25 },
		{ 204, 217, 25 },
		{ 204, 218, 25 },
		{ 204, 219, 25 },
		{ 204, 220, 25 },
		{ 204, 221, 25 },
		{ 204, 222, 25 },
		{ 204, 223, 25 },
		{ 204, 224, 25 },
		{ 204, 225, 25 },
		{ 204, 226, 25 },
		{ 204, 227, 25 },
		{ 204, 228, 25 },
		{ 204, 229, 25 },
		{ 204, 230, 25 },
		{ 205, 206, 25 },
		{ 205, 207, 25 },
		{ 205, 208, 25 },
		{ 205, 209, 25 },
		{ 205, 210, 25 },
		{ 205, 211, 25 },
		{ 205, 212, 25 },
		{ 205, 213, 25 },
		{ 205, 214, 25 },
		{ 205, 215, 25 },
		{ 205, 216, 25 },
		{ 205, 217, 25 },
		{ 205, 218, 25 },
		{ 205, 219, 25 },
		{ 205, 220, 25 },
		{ 205, 221, 25 },
		{ 205, 222, 25 },
		{ 205, 223, 25 },
		{ 205, 224, 25 },
		{ 205, 225, 25 },
		{ 205, 226, 25 },
		{ 205, 227, 25 },
		{ 205, 228, 25 },
		{ 205, 229, 25 },
		{ 205, 230, 25 },
		{ 206, 207, 25 },
		{ 206, 208, 25 },
		{ 206, 209, 25 },
		{ 206, 210, 25 },
		{ 206, 211, 25 },
		{ 206, 212, 25 },
		{ 206, 213, 25 },
		{ 206, 214, 25 },
		{ 206, 215, 25 },
		{ 206, 216, 25 },
		{ 206, 217, 25 },
		{ 206, 218, 25 },
		{ 206, 219, 25 },
		{ 206, 220, 25 },
		{ 206, 221, 25 },
		{ 206, 222, 25 },
		{ 206, 223, 25 },
		{ 206, 224, 25 },
		{ 206, 225, 25 },
		{ 206, 226, 25 },
		{ 206, 227, 25 },
		{ 206, 228, 25 },
		{ 206, 229, 25 },
		{ 206, 230, 25 },
		{ 206, 231, 25 },
		{ 207, 208, 25 },
		{ 207, 209, 25 },
		{ 207, 210, 25 },
		{ 207, 211, 25 },
		{ 207, 212, 25 },
		{ 207, 213, 25 },
		{ 207, 214, 25 },
		{ 207, 215, 25 },
		{ 207, 216, 25 },
		{ 207, 217, 25 },
		{ 207, 218, 25 },
		{ 207, 219, 25 },
		{ 207, 220, 25 },
		{ 207, 221, 25 },
		{ 207, 222, 25 },
		{ 207, 223, 25 },
		{ 207, 224, 25 },
		{ 207, 225, 25 },
		{ 207, 226, 25 },
		{ 207, 227, 25 },
		{ 207, 228, 25 },
		{ 207, 229, 25 },
		{ 207, 230, 25 },
		{ 207, 231, 25 },
		{ 208, 209, 25 },
		{ 208, 210, 25 },
		{ 208, 211, 25 },
		{ 208, 212, 25 },
		{ 208, 213, 25 },
		{ 208, 214, 25 },
		{ 208, 215, 25 },
		{ 208, 216, 25 },
		{ 208, 217, 25 },
		{ 208, 218, 25 },
		{ 208, 219, 25 },
		{ 208, 220, 25 },
		{ 208, 221, 25 },
		{ 208, 222, 25 },
		{ 208, 223, 25 },
		{ 208, 224, 25 },
		{ 208, 225, 25 },
		{ 208, 226, 25 },
		{ 208, 227, 25 },
		{ 208, 228, 25 },
		{ 208, 229, 25 },
		{ 208, 230, 25 },
		{ 208, 231, 25 },
		{ 208, 232, 25 },
		{ 209, 210, 25 },
		{ 209, 211, 25 },
		{ 209, 212, 25 },
		{ 209, 213, 25 },
		{ 209, 214, 25 },
		{ 209, 215, 25 },
		{ 209, 216, 25 },
		{ 209, 217, 25 },
		{ 209, 218, 25 },
		{ 209, 219, 25 },
		{ 209, 220, 25 },
		{ 209, 221, 25 },
		{ 209, 222, 25 },
		{ 209, 223, 25 },
		{ 209, 224, 25 },
		{ 209, 225, 25 },
		{ 209, 226, 25 },
		{ 209, 227, 25 },
		{ 209, 228, 25 },
		{ 209, 229, 25 },
		{ 209, 230, 25 },
		{ 209, 231, 25 },
		{ 209, 232, 25 },
		{ 210, 211, 25 },
		{ 210, 212, 25 },
		{ 210, 213, 25 },
		{ 210, 214, 25 },
		{ 210, 215, 25 },
		{ 210, 216, 25 },
		{ 210, 217, 25 },
		{ 210, 218, 25 },
		{ 210, 219, 25 },
		{ 210, 220, 25 },
		{ 210, 221, 25 },
		{ 210, 222, 25 },
		{ 210, 223, 25 },
		{ 210, 224, 25 },
		{ 210, 225, 25 },
		{ 210, 226, 25 },
		{ 210, 227, 25 },
		{ 210, 228, 25 },
		{ 210, 229, 25 },
		{ 210, 230, 25 },
		{ 210, 231, 25 },
		{ 210, 232, 25 },
		{ 210, 233, 25 },
		{ 211, 212, 25 },
		{ 211, 213, 25 },
		{ 211, 214, 25 },
		{ 211, 215, 25 },
		{ 211, 216, 25 },
		{ 211, 217, 25 },
		{ 211, 218, 25 },
		{ 211, 219, 25 },
		{ 211, 220, 25 },
		{ 211, 221, 25 },
		{ 211, 222, 25 },
		{ 211, 223, 25 },
		{ 211, 224, 25 },
		{ 211, 225, 25 },
		{ 211, 226, 25 },
		{ 211, 227, 25 },
		{ 211, 228, 25 },
		{ 211, 229, 25 },
		{ 211, 230, 25 },
		{ 211, 231, 25 },
		{ 211, 232, 25 },
		{ 211, 233, 25 },
		{ 212, 213, 25 },
		{ 212, 214, 25 },
		{ 212, 215, 25 },
		{ 212, 216, 25 },
		{ 212, 217, 25 },
		{ 212, 218, 25 },
		{ 212, 219, 25 },
		{ 212, 220, 25 },
		{ 212, 221, 25 },
		{ 212, 222, 25 },
		{ 212, 223, 25 },
		{ 212, 224, 25 },
		{ 212, 225, 25 },
		{ 212, 226, 25 },
		{ 212, 227, 25 },
		{ 212, 228, 25 },
		{ 212, 229, 25 },
		{ 212, 230, 25 },
		{ 212, 231, 25 },
		{ 212, 232, 25 },
		{ 212, 233, 25 },
		{ 212, 234, 25 },
		{ 213, 214, 25 },
		{ 213, 215, 25 },
		{ 213, 216, 25 },
		{ 213, 217, 25 },
		{ 213, 218, 25 },
		{ 213, 219, 25 },
		{ 213, 220, 25 },
		{ 213, 221, 25 },
		{ 213, 222, 25 },
		{ 213, 223, 25 },
		{ 213, 224, 25 },
		{ 213, 225, 25 },
		{ 213, 226, 25 },
		{ 213, 227, 25 },
		{ 213, 228, 25 },
		{ 213, 229, 25 },
		{ 213, 230, 25 },
		{ 213, 231, 25 },
		{ 213, 232, 25 },
		{ 213, 233, 25 },
		{ 213, 234, 25 },
		{ 214, 215, 25 },
		{ 214, 216, 25 },
		{ 214, 217, 25 },
		{ 214, 218, 25 },
		{ 214, 219, 25 },
		{ 214, 220, 25 },
		{ 214, 221, 25 },
		{ 214, 222, 25 },
		{ 214, 223, 25 },
		{ 214, 224, 25 },
		{ 214, 225, 25 },
		{ 214, 226, 25 },
		{ 214, 227, 25 },
		{ 214, 228, 25 },
		{ 214, 229, 25 },
		{ 214, 230, 25 },
		{ 214, 231, 25 },
		{ 214, 232, 25 },
		{ 214, 233, 25 },
		{ 214, 234, 25 },
		{ 214, 235, 25 },
		{ 215, 216, 25 },
		{ 215, 217, 25 },
		{ 215, 218, 25 },
		{ 215, 219, 25 },
		{ 215, 220, 25 },
		{ 215, 221, 25 },
		{ 215, 222, 25 },
		{ 215, 223, 25 },
		{ 215, 224, 25 },
		{ 215, 225, 25 },
		{ 215, 226, 25 },
		{ 215, 227, 25 },
		{ 215, 228, 25 },
		{ 215, 229, 25 },
		{ 215, 230, 25 },
		{ 215, 231, 25 },
		{ 215, 232, 25 },
		{ 215, 233, 25 },
		{ 215, 234, 25 },
		{ 215, 235, 25 },
		{ 216, 217, 25 },
		{ 216, 218, 25 },
		{ 216, 219, 25 },
		{ 216, 220, 25 },
		{ 216, 221, 25 },
		{ 216, 222, 25 },
		{ 216, 223, 25 },
		{ 216, 224, 25 },
		{ 216, 225, 25 },
		{ 216, 226, 25 },
		{ 216, 227, 25 },
		{ 216, 228, 25 },
		{ 216, 229, 25 },
		{ 216, 230, 25 },
		{ 216, 231, 25 },
		{ 216, 232, 25 },
		{ 216, 233, 25 },
		{ 216, 234, 25 },
		{ 216, 235, 25 },
		{ 216, 236, 25 },
		{ 217, 218, 25 },
		{ 217, 219, 25 },
		{ 217, 220, 25 },
		{ 217, 221, 25 },
		{ 217, 222, 25 },
		{ 217, 223, 25 },
		{ 217, 224, 25 },
		{ 217, 225, 25 },
		{ 217, 226, 25 },
		{ 217, 227, 25 },
		{ 217, 228, 25 },
		{ 217, 229, 25 },
		{ 217, 230, 25 },
		{ 217, 231, 25 },
		{ 217, 232, 25 },
		{ 217, 233, 25 },
		{ 217, 234, 25 },
		{ 217, 235, 25 },
		{ 217, 236, 25 },
		{ 218, 219, 25 },
		{ 218, 220, 25 },
		{ 218, 221, 25 },
		{ 218, 222, 25 },
		{ 218, 223, 25 },
		{ 218, 224, 25 },
		{ 218, 225, 25 },
		{ 218, 226, 25 },
		{ 218, 227, 25 },
		{ 218, 228, 25 },
		{ 218, 229, 25 },
		{ 218, 230, 25 },
		{ 218, 231, 25 },
		{ 218, 232, 25 },
		{ 218, 233, 25 },
		{ 218, 234, 25 },
		{ 218, 235, 25 },
		{ 218, 236, 25 },
		{ 218, 237, 25 },
		{ 219, 220, 25 },
		{ 219, 221, 25 },
		{ 219, 222, 25 },
		{ 219, 223, 25 },
		{ 219, 224, 25 },
		{ 219, 225, 25 },
		{ 219, 226, 25 },
		{ 219, 227, 25 },
		{ 219, 228, 25 },
		{ 219, 229, 25 },
		{ 219, 230, 25 },
		{ 219, 231, 25 },
		{ 219, 232, 25 },
		{ 219, 233, 25 },
		{ 219, 234, 25 },
		{ 219, 235, 25 },
		{ 219, 236, 25 },
		{ 219, 237, 25 },
		{ 220, 221, 25 },
		{ 220, 222, 25 },
		{ 220, 223, 25 },
		{ 220, 224, 25 },
		{ 220, 225, 25 },
		{ 220, 226, 25 },
		{ 220, 227, 25 },
		{ 220, 228, 25 },
		{ 220, 229, 25 },
		{ 220, 230, 25 },
		{ 220, 231, 25 },
		{ 220, 232, 25 },
		{ 220, 233, 25 },
		{ 220, 234, 25 },
		{ 220, 235, 25 },
		{ 220, 236, 25 },
		{ 220, 237, 25 },
		{ 220, 238, 25 },
		{ 221, 222, 25 },
		{ 221, 223, 25 },
		{ 221, 224, 25 },
		{ 221, 225, 25 },
		{ 221, 226, 25 },
		{ 221, 227, 25 },
		{ 221, 228, 25 },
		{ 221, 229, 25 },
		{ 221, 230, 25 },
		{ 221, 231, 25 },
		{ 221, 232, 25 },
		{ 221, 233, 25 },
		{ 221, 234, 25 },
		{ 221, 235, 25 },
		{ 221, 236, 25 },
		{ 221, 237, 25 },
		{ 221, 238, 25 },
		{ 222, 223, 25 },
		{ 222, 224, 25 },
		{ 222, 225, 25 },
		{ 222, 226, 25 },
		{ 222, 227, 25 },
		{ 222, 228, 25 },
		{ 222, 229, 25 },
		{ 222, 230, 25 },
		{ 222, 231, 25 },
		{ 222, 232, 25 },
		{ 222, 233, 25 },
		{ 222, 234, 25 },
		{ 222, 235, 25 },
		{ 222, 236, 25 },
		{ 222, 237, 25 },
		{ 222, 238, 25 },
		{ 222, 239, 25 },
		{ 223, 224, 25 },
		{ 223, 225, 25 },
		{ 223, 226, 25 },
		{ 223, 227, 25 },
		{ 223, 228, 25 },
		{ 223, 229, 25 },
		{ 223, 230, 25 },
		{ 223, 231, 25 },
		{ 223, 232, 25 },
		{ 223, 233, 25 },
		{ 223, 234, 25 },
		{ 223, 235, 25 },
		{ 223, 236, 25 },
		{ 223, 237, 25 },
		{ 223, 238, 25 },
		{ 223, 239, 25 },
		{ 224, 225, 25 },
		{ 224, 226, 25 },
		{ 224, 227, 25 },
		{ 224, 228, 25 },
		{ 224, 229, 25 },
		{ 224, 230, 25 },
		{ 224, 231, 25 },
		{ 224, 232, 25 },
		{ 224, 233, 25 },
		{ 224, 234, 25 },
		{ 224, 235, 25 },
		{ 224, 236, 25 },
		{ 224, 237, 25 },
		{ 224, 238, 25 },
		{ 224, 239, 25 },
		{ 224, 240, 25 },
		{ 225, 226, 25 },
		{ 225, 227, 25 },
		{ 225, 228, 25 },
		{ 225, 229, 25 },
		{ 225, 230, 25 },
		{ 225, 231, 25 },
		{ 225, 232, 25 },
		{ 225, 233, 25 },
		{ 225, 234, 25 },
		{ 225, 235, 25 },
		{ 225, 236, 25 },
		{ 225, 237, 25 },
		{ 225, 238, 25 },
		{ 225, 239, 25 },
		{ 225, 240, 25 },
		{ 226, 227, 25 },
		{ 226, 228, 25 },
		{ 226, 229, 25 },
		{ 226, 230, 25 },
		{ 226, 231, 25 },
		{ 226, 232, 25 },
		{ 226, 233, 25 },
		{ 226, 234, 25 },
		{ 226, 235, 25 },
		{ 226, 236, 25 },
		{ 226, 237, 25 },
		{ 226, 238, 25 },
		{ 226, 239, 25 },
		{ 226, 240, 25 },
		{ 226, 241, 25 },
		{ 227, 228, 25 },
		{ 227, 229, 25 },
		{ 227, 230, 25 },
		{ 227, 231, 25 },
		{ 227, 232, 25 },
		{ 227, 233, 25 },
		{ 227, 234, 25 },
		{ 227, 235, 25 },
		{ 227, 236, 25 },
		{ 227, 237, 25 },
		{ 227, 238, 25 },
		{ 227, 239, 25 },
		{ 227, 240, 25 },
		{ 227, 241, 25 },
		{ 228, 229, 25 },
		{ 228, 230, 25 },
		{ 228, 231, 25 },
		{ 228, 232, 25 },
		{ 228, 233, 25 },
		{ 228, 234, 25 },
		{ 228, 235, 25 },
		{ 228, 236, 25 },
		{ 228, 237, 25 },
		{ 228, 238, 25 },
		{ 228, 239, 25 },
		{ 228, 240, 25 },
		{ 228, 241, 25 },
		{ 228, 242, 25 },
		{ 229, 230, 25 },
		{ 229, 231, 25 },
		{ 229, 232, 25 },
		{ 229, 233, 25 },
		{ 229, 234, 25 },
		{ 229, 235, 25 },
		{ 229, 236, 25 },
		{ 229, 237, 25 },
		{ 229, 238, 25 },
		{ 229, 239, 25 },
		{ 229, 240, 25 },
		{ 229, 241, 25 },
		{ 229, 242, 25 },
		{ 230, 231, 25 },
		{ 230, 232, 25 },
		{ 230, 233, 25 },
		{ 230, 234, 25 },
		{ 230, 235, 25 },
		{ 230, 236, 25 },
		{ 230, 237, 25 },
		{ 230, 238, 25 },
		{ 230, 239, 25 },
		{ 230, 240, 25 },
		{ 230, 241, 25 },
		{ 230, 242, 25 },
		{ 230, 243, 25 },
		{ 231, 232, 25 },
		{ 231, 233, 25 },
		{ 231, 234, 25 },
		{ 231, 235, 25 },
		{ 231, 236, 25 },
		{ 231, 237, 25 },
		{ 231, 238, 25 },
		{ 231, 239, 25 },
		{ 231, 240, 25 },
		{ 231, 241, 25 },
		{ 231, 242, 25 },
		{ 231, 243, 25 },
		{ 232, 233, 25 },
		{ 232, 234, 25 },
		{ 232, 235, 25 },
		{ 232, 236, 25 },
		{ 232, 237, 25 },
		{ 232, 238, 25 },
		{ 232, 239, 25 },
		{ 232, 240, 25 },
		{ 232, 241, 25 },
		{ 232, 242, 25 },
		{ 232, 243, 25 },
		{ 232, 244, 25 },
		{ 233, 234, 25 },
		{ 233, 235, 25 },
		{ 233, 236, 25 },
		{ 233, 237, 25 },
		{ 233, 238, 25 },
		{ 233, 239, 25 },
		{ 233, 240, 25 },
		{ 233, 241, 25 },
		{ 233, 242, 25 },
		{ 233, 243, 25 },
		{ 233, 244, 25 },
		{ 234, 235, 25 },
		{ 234, 236, 25 },
		{ 234, 237, 25 },
		{ 234, 238, 25 },
		{ 234, 239, 25 },
		{ 234, 240, 25 },
		{ 234, 241, 25 },
		{ 234, 242, 25 },
		{ 234, 243, 25 },
		{ 234, 244, 25 },
		{ 234, 245, 25 },
		{ 235, 236, 25 },
		{ 235, 237, 25 },
		{ 235, 238, 25 },
		{ 235, 239, 25 },
		{ 235, 240, 25 },
		{ 235, 241, 25 },
		{ 235, 242, 25 },
		{ 235, 243, 25 },
		{ 235, 244, 25 },
		{ 235, 245, 25 },
		{ 236, 237, 25 },
		{ 236, 238, 25 },
		{ 236, 239, 25 },
		{ 236, 240, 25 },
		{ 236, 241, 25 },
		{ 236, 242, 25 },
		{ 236, 243, 25 },
		{ 236, 244, 25 },
		{ 236, 245, 25 },
		{ 236, 246, 25 },
		{ 237, 238, 25 },
		{ 237, 239, 25 },
		{ 237, 240, 25 },
		{ 237, 241, 25 },
		{ 237, 242, 25 },
		{ 237, 243, 25 },
		{ 237, 244, 25 },
		{ 237, 245, 25 },
		{ 237, 246, 25 },
		{ 238, 239, 25 },
		{ 238, 240, 25 },
		{ 238, 241, 25 },
		{ 238, 242, 25 },
		{ 238, 243, 25 },
		{ 238, 244, 25 },
		{ 238, 245, 25 },
		{ 238, 246, 25 },
		{ 238, 247, 25 },
		{ 239, 240, 25 },
		{ 239, 241, 25 },
		{ 239, 242, 25 },
		{ 239, 243, 25 },
		{ 239, 244, 25 },
		{ 239, 245, 25 },
		{ 239, 246, 25 },
		{ 239, 247, 25 },
		{ 240, 241, 25 },
		{ 240, 242, 25 },
		{ 240, 243, 25 },
		{ 240, 244, 25 },
		{ 240, 245, 25 },
		{ 240, 246, 25 },
		{ 240, 247, 25 },
		{ 240, 248, 25 },
		{ 274, 266, 25 },
		{ 274, 267, 25 },
		{ 274, 268, 25 },
		{ 274, 269, 25 },
		{ 274, 270, 25 },
		{ 274, 271, 25 },
		{ 274, 272, 25 },
		{ 274, 273, 25 },
		{ 275, 267, 25 },
		{ 275, 268, 25 },
		{ 275, 269, 25 },
		{ 275, 270, 25 },
		{ 275, 271, 25 },
		{ 275, 272, 25 },
		{ 275, 273, 25 },
		{ 275, 274, 25 },
		{ 276, 267, 25 },
		{ 276, 268, 25 },
		{ 276, 269, 25 },
		{ 276, 270, 25 },
		{ 276, 271, 25 },
		{ 276, 272, 25 },
		{ 276, 273, 25 },
		{ 276, 274, 25 },
		{ 276, 275, 25 },
		{ 277, 268, 25 },
		{ 277, 269, 25 },
		{ 277, 270, 25 },
		{ 277, 271, 25 },
		{ 277, 272, 25 },
		{ 277, 273, 25 },
		{ 277, 274, 25 },
		{ 277, 275, 25 },
		{ 277, 276, 25 },
		{ 278, 268, 25 },
		{ 278, 269, 25 },
		{ 278, 270, 25 },
		{ 278, 271, 25 },
		{ 278, 272, 25 },
		{ 278, 273, 25 },
		{ 278, 274, 25 },
		{ 278, 275, 25 },
		{ 278, 276, 25 },
		{ 278, 277, 25 },
		{ 279, 269, 25 },
		{ 279, 270, 25 },
		{ 279, 271, 25 },
		{ 279, 272, 25 },
		{ 279, 273, 25 },
		{ 279, 274, 25 },
		{ 279, 275, 25 },
		{ 279, 276, 25 },
		{ 279, 277, 25 },
		{ 279, 278, 25 },
		{ 280, 269, 25 },
		{ 280, 270, 25 },
		{ 280, 271, 25 },
		{ 280, 272, 25 },
		{ 280, 273, 25 },
		{ 280, 274, 25 },
		{ 280, 275, 25 },
		{ 280, 276, 25 },
		{ 280, 277, 25 },
		{ 280, 278, 25 },
		{ 280, 279, 25 },
		{ 281, 270, 25 },
		{ 281, 271, 25 },
		{ 281, 272, 25 },
		{ 281, 273, 25 },
		{ 281, 274, 25 },
		{ 281, 275, 25 },
		{ 281, 276, 25 },
		{ 281, 277, 25 },
		{ 281, 278, 25 },
		{ 281, 279, 25 },
		{ 281, 280, 25 },
		{ 282, 270, 25 },
		{ 282, 271, 25 },
		{ 282, 272, 25 },
		{ 282, 273, 25 },
		{ 282, 274, 25 },
		{ 282, 275, 25 },
		{ 282, 276, 25 },
		{ 282, 277, 25 },
		{ 282, 278, 25 },
		{ 282, 279, 25 },
		{ 282, 280, 25 },
		{ 282, 281, 25 },
		{ 283, 271, 25 },
		{ 283, 272, 25 },
		{ 283, 273, 25 },
		{ 283, 274, 25 },
		{ 283, 275, 25 },
		{ 283, 276, 25 },
		{ 283, 277, 25 },
		{ 283, 278, 25 },
		{ 283, 279, 25 },
		{ 283, 280, 25 },
		{ 283, 281, 25 },
		{ 283, 282, 25 },
		{ 284, 271, 25 },
		{ 284, 272, 25 },
		{ 284, 273, 25 },
		{ 284, 274, 25 },
		{ 284, 275, 25 },
		{ 284, 276, 25 },
		{ 284, 277, 25 },
		{ 284, 278, 25 },
		{ 284, 279, 25 },
		{ 284, 280, 25 },
		{ 284, 281, 25 },
		{ 284, 282, 25 },
		{ 284, 283, 25 },
		{ 285, 272, 25 },
		{ 285, 273, 25 },
		{ 285, 274, 25 },
		{ 285, 275, 25 },
		{ 285, 276, 25 },
		{ 285, 277, 25 },
		{ 285, 278, 25 },
		{ 285, 279, 25 },
		{ 285, 280, 25 },
		{ 285, 281, 25 },
		{ 285, 282, 25 },
		{ 285, 283, 25 },
		{ 285, 284, 25 },
		{ 286, 272, 25 },
		{ 286, 273, 25 },
		{ 286, 274, 25 },
		{ 286, 275, 25 },
		{ 286, 276, 25 },
		{ 286, 277, 25 },
		{ 286, 278, 25 },
		{ 286, 279, 25 },
		{ 286, 280, 25 },
		{ 286, 281, 25 },
		{ 286, 282, 25 },
		{ 286, 283, 25 },
		{ 286, 284, 25 },
		{ 286, 285, 25 },
		{ 287, 273, 25 },
		{ 287, 274, 25 },
		{ 287, 275, 25 },
		{ 287, 276, 25 },
		{ 287, 277, 25 },
		{ 287, 278, 25 },
		{ 287, 279, 25 },
		{ 287, 280, 25 },
		{ 287, 281, 25 },
		{ 287, 282, 25 },
		{ 287, 283, 25 },
		{ 287, 284, 25 },
		{ 287, 285, 25 },
		{ 287, 286, 25 },
		{ 288, 273, 25 },
		{ 288, 274, 25 },
		{ 288, 275, 25 },
		{ 288, 276, 25 },
		{ 288, 277, 25 },
		{ 288, 278, 25 },
		{ 288, 279, 25 },
		{ 288, 280, 25 },
		{ 288, 281, 25 },
		{ 288, 282, 25 },
		{ 288, 283, 25 },
		{ 288, 284, 25 },
		{ 288, 285, 25 },
		{ 288, 286, 25 },
		{ 288, 287, 25 },
		{ 289, 274, 25 },
		{ 289, 275, 25 },
		{ 289, 276, 25 },
		{ 289, 277, 25 },
		{ 289, 278, 25 },
		{ 289, 279, 25 },
		{ 289, 280, 25 },
		{ 289, 281, 25 },
		{ 289, 282, 25 },
		{ 289, 283, 25 },
		{ 289, 284, 25 },
		{ 289, 285, 25 },
		{ 289, 286, 25 },
		{ 289, 287, 25 },
		{ 289, 288, 25 },
		{ 290, 274, 25 },
		{ 290, 275, 25 },
		{ 290, 276, 25 },
		{ 290, 277, 25 },
		{ 290, 278, 25 },
		{ 290, 279, 25 },
		{ 290, 280, 25 },
		{ 290, 281, 25 },
		{ 290, 282, 25 },
		{ 290, 283, 25 },
		{ 290, 284, 25 },
		{ 290, 285, 25 },
		{ 290, 286, 25 },
		{ 290, 287, 25 },
		{ 290, 288, 25 },
		{ 290, 289, 25 },
		{ 291, 275, 25 },
		{ 291, 276, 25 },
		{ 291, 277, 25 },
		{ 291, 278, 25 },
		{ 291, 279, 25 },
		{ 291, 280, 25 },
		{ 291, 281, 25 },
		{ 291, 282, 25 },
		{ 291, 283, 25 },
		{ 291, 284, 25 },
		{ 291, 285, 25 },
		{ 291, 286, 25 },
		{ 291, 287, 25 },
		{ 291, 288, 25 },
		{ 291, 289, 25 },
		{ 291, 290, 25 },
		{ 292, 275, 25 },
		{ 292, 276, 25 },
		{ 292, 277, 25 },
		{ 292, 278, 25 },
		{ 292, 279, 25 },
		{ 292, 280, 25 },
		{ 292, 281, 25 },
		{ 292, 282, 25 },
		{ 292, 283, 25 },
		{ 292, 284, 25 },
		{ 292, 285, 25 },
		{ 292, 286, 25 },
		{ 292, 287, 25 },
		{ 292, 288, 25 },
		{ 292, 289, 25 },
		{ 292, 290, 25 },
		{ 292, 291, 25 },
		{ 293, 276, 25 },
		{ 293, 277, 25 },
		{ 293, 278, 25 },
		{ 293, 279, 25 },
		{ 293, 280, 25 },
		{ 293, 281, 25 },
		{ 293, 282, 25 },
		{ 293, 283, 25 },
		{ 293, 284, 25 },
		{ 293, 285, 25 },
		{ 293, 286, 25 },
		{ 293, 287, 25 },
		{ 293, 288, 25 },
		{ 293, 289, 25 },
		{ 293, 290, 25 },
		{ 293, 291, 25 },
		{ 293, 292, 25 },
		{ 294, 276, 25 },
		{ 294, 277, 25 },
		{ 294, 278, 25 },
		{ 294, 279, 25 },
		{ 294, 280, 25 },
		{ 294, 281, 25 },
		{ 294, 282, 25 },
		{ 294, 283, 25 },
		{ 294, 284, 25 },
		{ 294, 285, 25 },
		{ 294, 286, 25 },
		{ 294, 287, 25 },
		{ 294, 288, 25 },
		{ 294, 289, 25 },
		{ 294, 290, 25 },
		{ 294, 291, 25 },
		{ 294, 292, 25 },
		{ 294, 293, 25 },
		{ 295, 277, 25 },
		{ 295, 278, 25 },
		{ 295, 279, 25 },
		{ 295, 280, 25 },
		{ 295, 281, 25 },
		{ 295, 282, 25 },
		{ 295, 283, 25 },
		{ 295, 284, 25 },
		{ 295, 285, 25 },
		{ 295, 286, 25 },
		{ 295, 287, 25 },
		{ 295, 288, 25 },
		{ 295, 289, 25 },
		{ 295, 290, 25 },
		{ 295, 291, 25 },
		{ 295, 292, 25 },
		{ 295, 293, 25 },
		{ 295, 294, 25 },
		{ 296, 277, 25 },
		{ 296, 278, 25 },
		{ 296, 279, 25 },
		{ 296, 280, 25 },
		{ 296, 281, 25 },
		{ 296, 282, 25 },
		{ 296, 283, 25 },
		{ 296, 284, 25 },
		{ 296, 285, 25 },
		{ 296, 286, 25 },
		{ 296, 287, 25 },
		{ 296, 288, 25 },
		{ 296, 289, 25 },
		{ 296, 290, 25 },
		{ 296, 291, 25 },
		{ 296, 292, 25 },
		{ 296, 293, 25 },
		{ 296, 294, 25 },
		{ 296, 295, 25 },
		{ 297, 278, 25 },
		{ 297, 279, 25 },
		{ 297, 280, 25 },
		{ 297, 281, 25 },
		{ 297, 282, 25 },
		{ 297, 283, 25 },
		{ 297, 284, 25 },
		{ 297, 285, 25 },
		{ 297, 286, 25 },
		{ 297, 287, 25 },
		{ 297, 288, 25 },
		{ 297, 289, 25 },
		{ 297, 290, 25 },
		{ 297, 291, 25 },
		{ 297, 292, 25 },
		{ 297, 293, 25 },
		{ 297, 294, 25 },
		{ 297, 295, 25 },
		{ 297, 296, 25 },
		{ 298, 278, 25 },
		{ 298, 279, 25 },
		{ 298, 280, 25 },
		{ 298, 281, 25 },
		{ 298, 282, 25 },
		{ 298, 283, 25 },
		{ 298, 284, 25 },
		{ 298, 285, 25 },
		{ 298, 286, 25 },
		{ 298, 287, 25 },
		{ 298, 288, 25 },
		{ 298, 289, 25 },
		{ 298, 290, 25 },
		{ 298, 291, 25 },
		{ 298, 292, 25 },
		{ 298, 293, 25 },
		{ 298, 294, 25 },
		{ 298, 295, 25 },
		{ 298, 296, 25 },
		{ 298, 297, 25 },
		{ 299, 279, 25 },
		{ 299, 280, 25 },
		{ 299, 281, 25 },
		{ 299, 282, 25 },
		{ 299, 283, 25 },
		{ 299, 284, 25 },
		{ 299, 285, 25 },
		{ 299, 286, 25 },
		{ 299, 287, 25 },
		{ 299, 288, 25 },
		{ 299, 289, 25 },
		{ 299, 290, 25 },
		{ 299, 291, 25 },
		{ 299, 292, 25 },
		{ 299, 293, 25 },
		{ 299, 294, 25 },
		{ 299, 295, 25 },
		{ 299, 296, 25 },
		{ 299, 297, 25 },
		{ 299, 298, 25 },
		{ 300, 279, 25 },
		{ 300, 280, 25 },
		{ 300, 281, 25 },
		{ 300, 282, 25 },
		{ 300, 283, 25 },
		{ 300, 284, 25 },
		{ 300, 285, 25 },
		{ 300, 286, 25 },
		{ 300, 287, 25 },
		{ 300, 288, 25 },
		{ 300, 289, 25 },
		{ 300, 290, 25 },
		{ 300, 291, 25 },
		{ 300, 292, 25 },
		{ 300, 293, 25 },
		{ 300, 294, 25 },
		{ 300, 295, 25 },
		{ 300, 296, 25 },
		{ 300, 297, 25 },
		{ 300, 298, 25 },
		{ 300, 299, 25 },
		{ 301, 280, 25 },
		{ 301, 281, 25 },
		{ 301, 282, 25 },
		{ 301, 283, 25 },
		{ 301, 284, 25 },
		{ 301, 285, 25 },
		{ 301, 286, 25 },
		{ 301, 287, 25 },
		{ 301, 288, 25 },
		{ 301, 289, 25 },
		{ 301, 290, 25 },
		{ 301, 291, 25 },
		{ 301, 292, 25 },
		{ 301, 293, 25 },
		{ 301, 294, 25 },
		{ 301, 295, 25 },
		{ 301, 296, 25 },
		{ 301, 297, 25 },
		{ 301, 298, 25 },
		{ 301, 299, 25 },
		{ 301, 300, 25 },
		{ 302, 280, 25 },
		{ 302, 281, 25 },
		{ 302, 282, 25 },
		{ 302, 283, 25 },
		{ 302, 284, 25 },
		{ 302, 285, 25 },
		{ 302, 286, 25 },
		{ 302, 287, 25 },
		{ 302, 288, 25 },
		{ 302, 289, 25 },
		{ 302, 290, 25 },
		{ 302, 291, 25 },
		{ 302, 292, 25 },
		{ 302, 293, 25 },
		{ 302, 294, 25 },
		{ 302, 295, 25 },
		{ 302, 296, 25 },
		{ 302, 297, 25 },
		{ 302, 298, 25 },
		{ 302, 299, 25 },
		{ 302, 300, 25 },
		{ 302, 301, 25 },
		{ 303, 281, 25 },
		{ 303, 282, 25 },
		{ 303, 283, 25 },
		{ 303, 284, 25 },
		{ 303, 285, 25 },
		{ 303, 286, 25 },
		{ 303, 287, 25 },
		{ 303, 288, 25 },
		{ 303, 289, 25 },
		{ 303, 290, 25 },
		{ 303, 291, 25 },
		{ 303, 292, 25 },
		{ 303, 293, 25 },
		{ 303, 294, 25 },
		{ 303, 295, 25 },
		{ 303, 296, 25 },
		{ 303, 297, 25 },
		{ 303, 298, 25 },
		{ 303, 299, 25 },
		{ 303, 300, 25 },
		{ 303, 301, 25 },
		{ 303, 302, 25 },
		{ 304, 281, 25 },
		{ 304, 282, 25 },
		{ 304, 283, 25 },
		{ 304, 284, 25 },
		{ 304, 285, 25 },
		{ 304, 286, 25 },
		{ 304, 287, 25 },
		{ 304, 288, 25 },
		{ 304, 289, 25 },
		{ 304, 290, 25 },
		{ 304, 291, 25 },
		{ 304, 292, 25 },
		{ 304, 293, 25 },
		{ 304, 294, 25 },
		{ 304, 295, 25 },
		{ 304, 296, 25 },
		{ 304, 297, 25 },
		{ 304, 298, 25 },
		{ 304, 299, 25 },
		{ 304, 300, 25 },
		{ 304, 301, 25 },
		{ 304, 302, 25 },
		{ 304, 303, 25 },
		{ 305, 282, 25 },
		{ 305, 283, 25 },
		{ 305, 284, 25 },
		{ 305, 285, 25 },
		{ 305, 286, 25 },
		{ 305, 287, 25 },
		{ 305, 288, 25 },
		{ 305, 289, 25 },
		{ 305, 290, 25 },
		{ 305, 291, 25 },
		{ 305, 292, 25 },
		{ 305, 293, 25 },
		{ 305, 294, 25 },
		{ 305, 295, 25 },
		{ 305, 296, 25 },
		{ 305, 297, 25 },
		{ 305, 298, 25 },
		{ 305, 299, 25 },
		{ 305, 300, 25 },
		{ 305, 301, 25 },
		{ 305, 302, 25 },
		{ 305, 303, 25 },
		{ 305, 304, 25 },
		{ 306, 282, 25 },
		{ 306, 283, 25 },
		{ 306, 284, 25 },
		{ 306, 285, 25 },
		{ 306, 286, 25 },
		{ 306, 287, 25 },
		{ 306, 288, 25 },
		{ 306, 289, 25 },
		{ 306, 290, 25 },
		{ 306, 291, 25 },
		{ 306, 292, 25 },
		{ 306, 293, 25 },
		{ 306, 294, 25 },
		{ 306, 295, 25 },
		{ 306, 296, 25 },
		{ 306, 297, 25 },
		{ 306, 298, 25 },
		{ 306, 299, 25 },
		{ 306, 300, 25 },
		{ 306, 301, 25 },
		{ 306, 302, 25 },
		{ 306, 303, 25 },
		{ 306, 304, 25 },
		{ 306, 305, 25 },
		{ 307, 283, 25 },
		{ 307, 284, 25 },
		{ 307, 285, 25 },
		{ 307, 286, 25 },
		{ 307, 287, 25 },
		{ 307, 288, 25 },
		{ 307, 289, 25 },
		{ 307, 290, 25 },
		{ 307, 291, 25 },
		{ 307, 292, 25 },
		{ 307, 293, 25 },
		{ 307, 294, 25 },
		{ 307, 295, 25 },
		{ 307, 296, 25 },
		{ 307, 297, 25 },
		{ 307, 298, 25 },
		{ 307, 299, 25 },
		{ 307, 300, 25 },
		{ 307, 301, 25 },
		{ 307, 302, 25 },
		{ 307, 303, 25 },
		{ 307, 304, 25 },
		{ 307, 305, 25 },
		{ 307, 306, 25 },
		{ 308, 283, 25 },
		{ 308, 284, 25 },
		{ 308, 285, 25 },
		{ 308, 286, 25 },
		{ 308, 287, 25 },
		{ 308, 288, 25 },
		{ 308, 289, 25 },
		{ 308, 290, 25 },
		{ 308, 291, 25 },
		{ 308, 292, 25 },
		{ 308, 293, 25 },
		{ 308, 294, 25 },
		{ 308, 295, 25 },
		{ 308, 296, 25 },
		{ 308, 297, 25 },
		{ 308, 298, 25 },
		{ 308, 299, 25 },
		{ 308, 300, 25 },
		{ 308, 301, 25 },
		{ 308, 302, 25 },
		{ 308, 303, 25 },
		{ 308, 304, 25 },
		{ 308, 305, 25 },
		{ 308, 306, 25 },
		{ 308, 307, 25 },
		{ 309, 284, 25 },
		{ 309, 285, 25 },
		{ 309, 286, 25 },
		{ 309, 287, 25 },
		{ 309, 288, 25 },
		{ 309, 289, 25 },
		{ 309, 290, 25 },
		{ 309, 291, 25 },
		{ 309, 292, 25 },
		{ 309, 293, 25 },
		{ 309, 294, 25 },
		{ 309, 295, 25 },
		{ 309, 296, 25 },
		{ 309, 297, 25 },
		{ 309, 298, 25 },
		{ 309, 299, 25 },
		{ 309, 300, 25 },
		{ 309, 301, 25 },
		{ 309, 302, 25 },
		{ 309, 303, 25 },
		{ 309, 304, 25 },
		{ 309, 305, 25 },
		{ 309, 306, 25 },
		{ 309, 307, 25 },
		{ 309, 308, 25 },
		{ 310, 284, 25 },
		{ 310, 285, 25 },
		{ 310, 286, 25 },
		{ 310, 287, 25 },
		{ 310, 288, 25 },
		{ 310, 289, 25 },
		{ 310, 290, 25 },
		{ 310, 291, 25 },
		{ 310, 292, 25 },
		{ 310, 293, 25 },
		{ 310, 294, 25 },
		{ 310, 295, 25 },
		{ 310, 296, 25 },
		{ 310, 297, 25 },
		{ 310, 298, 25 },
		{ 310, 299, 25 },
		{ 310, 300, 25 },
		{ 310, 301, 25 },
		{ 310, 302, 25 },
		{ 310, 303, 25 },
		{ 310, 304, 25 },
		{ 310, 305, 25 },
		{ 310, 306, 25 },
		{ 310, 307, 25 },
		{ 310, 308, 25 },
		{ 310, 309, 25 },
		{ 311, 285, 25 },
		{ 311, 286, 25 },
		{ 311, 287, 25 },
		{ 311, 288, 25 },
		{ 311, 289, 25 },
		{ 311, 290, 25 },
		{ 311, 291, 25 },
		{ 311, 292, 25 },
		{ 311, 293, 25 },
		{ 311, 294, 25 },
		{ 311, 295, 25 },
		{ 311, 296, 25 },
		{ 311, 297, 25 },
		{ 311, 298, 25 },
		{ 311, 299, 25 },
		{ 311, 300, 25 },
		{ 311, 301, 25 },
		{ 311, 302, 25 },
		{ 311, 303, 25 },
		{ 311, 304, 25 },
		{ 311, 305, 25 },
		{ 311, 306, 25 },
		{ 311, 307, 25 },
		{ 311, 308, 25 },
		{ 311, 309, 25 },
		{ 311, 310, 25 },
		{ 312, 285, 25 },
		{ 312, 286, 25 },
		{ 312, 287, 25 },
		{ 312, 288, 25 },
		{ 312, 289, 25 },
		{ 312, 290, 25 },
		{ 312, 291, 25 },
		{ 312, 292, 25 },
		{ 312, 293, 25 },
		{ 312, 294, 25 },
		{ 312, 295, 25 },
		{ 312, 296, 25 },
		{ 312, 297, 25 },
		{ 312, 298, 25 },
		{ 312, 299, 25 },
		{ 312, 300, 25 },
		{ 312, 301, 25 },
		{ 312, 302, 25 },
		{ 312, 303, 25 },
		{ 312, 304, 25 },
		{ 312, 305, 25 },
		{ 312, 306, 25 },
		{ 312, 307, 25 },
		{ 312, 308, 25 },
		{ 312, 309, 25 },
		{ 312, 310, 25 },
		{ 312, 311, 25 },
		{ 313, 286, 25 },
		{ 313, 287, 25 },
		{ 313, 288, 25 },
		{ 313, 289, 25 },
		{ 313, 290, 25 },
		{ 313, 291, 25 },
		{ 313, 292, 25 },
		{ 313, 293, 25 },
		{ 313, 294, 25 },
		{ 313, 295, 25 },
		{ 313, 296, 25 },
		{ 313, 297, 25 },
		{ 313, 298, 25 },
		{ 313, 299, 25 },
		{ 313, 300, 25 },
		{ 313, 301, 25 },
		{ 313, 302, 25 },
		{ 313, 303, 25 },
		{ 313, 304, 25 },
		{ 313, 305, 25 },
		{ 313, 306, 25 },
		{ 313, 307, 25 },
		{ 313, 308, 25 },
		{ 313, 309, 25 },
		{ 313, 310, 25 },
		{ 313, 311, 25 },
		{ 313, 312, 25 },
		{ 314, 286, 25 },
		{ 314, 287, 25 },
		{ 314, 288, 25 },
		{ 314, 289, 25 },
		{ 314, 290, 25 },
		{ 314, 291, 25 },
		{ 314, 292, 25 },
		{ 314, 293, 25 },
		{ 314, 294, 25 },
		{ 314, 295, 25 },
		{ 314, 296, 25 },
		{ 314, 297, 25 },
		{ 314, 298, 25 },
		{ 314, 299, 25 },
		{ 314, 300, 25 },
		{ 314, 301, 25 },
		{ 314, 302, 25 },
		{ 314, 303, 25 },
		{ 314, 304, 25 },
		{ 314, 305, 25 },
		{ 314, 306, 25 },
		{ 314, 307, 25 },
		{ 314, 308, 25 },
		{ 314, 309, 25 },
		{ 314, 310, 25 },
		{ 314, 311, 25 },
		{ 314, 312, 25 },
		{ 314, 313, 25 },
		{ 315, 287, 25 },
		{ 315, 288, 25 },
		{ 315, 289, 25 },
		{ 315, 290, 25 },
		{ 315, 291, 25 },
		{ 315, 292, 25 },
		{ 315, 293, 25 },
		{ 315, 294, 25 },
		{ 315, 295, 25 },
		{ 315, 296, 25 },
		{ 315, 297, 25 },
		{ 315, 298, 25 },
		{ 315, 299, 25 },
		{ 315, 300, 25 },
		{ 315, 301, 25 },
		{ 315, 302, 25 },
		{ 315, 303, 25 },
		{ 315, 304, 25 },
		{ 315, 305, 25 },
		{ 315, 306, 25 },
		{ 315, 307, 25 },
		{ 315, 308, 25 },
		{ 315, 309, 25 },
		{ 315, 310, 25 },
		{ 315, 311, 25 },
		{ 315, 312, 25 },
		{ 315, 313, 25 },
		{ 315, 314, 25 },
		{ 316, 287, 25 },
		{ 316, 288, 25 },
		{ 316, 289, 25 },
		{ 316, 290, 25 },
		{ 316, 291, 25 },
		{ 316, 292, 25 },
		{ 316, 293, 25 },
		{ 316, 294, 25 },
		{ 316, 295, 25 },
		{ 316, 296, 25 },
		{ 316, 297, 25 },
		{ 316, 298, 25 },
		{ 316, 299, 25 },
		{ 316, 300, 25 },
		{ 316, 301, 25 },
		{ 316, 302, 25 },
		{ 316, 303, 25 },
		{ 316, 304, 25 },
		{ 316, 305, 25 },
		{ 316, 306, 25 },
		{ 316, 307, 25 },
		{ 316, 308, 25 },
		{ 316, 309, 25 },
		{ 316, 310, 25 },
		{ 316, 311, 25 },
		{ 316, 312, 25 },
		{ 316, 313, 25 },
		{ 316, 314, 25 },
		{ 316, 315, 25 },
		{ 317, 288, 25 },
		{ 317, 289, 25 },
		{ 317, 290, 25 },
		{ 317, 291, 25 },
		{ 317, 292, 25 },
		{ 317, 293, 25 },
		{ 317, 294, 25 },
		{ 317, 295, 25 },
		{ 317, 296, 25 },
		{ 317, 297, 25 },
		{ 317, 298, 25 },
		{ 317, 299, 25 },
		{ 317, 300, 25 },
		{ 317, 301, 25 },
		{ 317, 302, 25 },
		{ 317, 303, 25 },
		{ 317, 304, 25 },
		{ 317, 305, 25 },
		{ 317, 306, 25 },
		{ 317, 307, 25 },
		{ 317, 308, 25 },
		{ 317, 309, 25 },
		{ 317, 310, 25 },
		{ 317, 311, 25 },
		{ 317, 312, 25 },
		{ 317, 313, 25 },
		{ 317, 314, 25 },
		{ 317, 315, 25 },
		{ 317, 316, 25 },
		{ 318, 288, 25 },
		{ 318, 289, 25 },
		{ 318, 290, 25 },
		{ 318, 291, 25 },
		{ 318, 292, 25 },
		{ 318, 293, 25 },
		{ 318, 294, 25 },
		{ 318, 295, 25 },
		{ 318, 296, 25 },
		{ 318, 297, 25 },
		{ 318, 298, 25 },
		{ 318, 299, 25 },
		{ 318, 300, 25 },
		{ 318, 301, 25 },
		{ 318, 302, 25 },
		{ 318, 303, 25 },
		{ 318, 304, 25 },
		{ 318, 305, 25 },
		{ 318, 306, 25 },
		{ 318, 307, 25 },
		{ 318, 308, 25 },
		{ 318, 309, 25 },
		{ 318, 310, 25 },
		{ 318, 311, 25 },
		{ 318, 312, 25 },
		{ 318, 313, 25 },
		{ 318, 314, 25 },
		{ 318, 315, 25 },
		{ 318, 316, 25 },
		{ 318, 317, 25 },
		{ 319, 289, 25 },
		{ 319, 290, 25 },
		{ 319, 291, 25 },
		{ 319, 292, 25 },
		{ 319, 293, 25 },
		{ 319, 294, 25 },
		{ 319, 295, 25 },
		{ 319, 296, 25 },
		{ 319, 297, 25 },
		{ 319, 298, 25 },
		{ 319, 299, 25 },
		{ 319, 300, 25 },
		{ 319, 301, 25 },
		{ 319, 302, 25 },
		{ 319, 303, 25 },
		{ 319, 304, 25 },
		{ 319, 305, 25 },
		{ 319, 306, 25 },
		{ 319, 307, 25 },
		{ 319, 308, 25 },
		{ 319, 309, 25 },
		{ 319, 310, 25 },
		{ 319, 311, 25 },
		{ 319, 312, 25 },
		{ 319, 313, 25 },
		{ 319, 314, 25 },
		{ 319, 315, 25 },
		{ 319, 316, 25 },
		{ 319, 317, 25 },
		{ 319, 318, 25 },
		{ 320, 289, 25 },
		{ 320, 290, 25 },
		{ 320, 291, 25 },
		{ 320, 292, 25 },
		{ 320, 293, 25 },
		{ 320, 294, 25 },
		{ 320, 295, 25 },
		{ 320, 296, 25 },
		{ 320, 297, 25 },
		{ 320, 298, 25 },
		{ 320, 299, 25 },
		{ 320, 300, 25 },
		{ 320, 301, 25 },
		{ 320, 302, 25 },
		{ 320, 303, 25 },
		{ 320, 304, 25 },
		{ 320, 305, 25 },
		{ 320, 306, 25 },
		{ 320, 307, 25 },
		{ 320, 308, 25 },
		{ 320, 309, 25 },
		{ 320, 310, 25 },
		{ 320, 311, 25 },
		{ 320, 312, 25 },
		{ 320, 313, 25 },
		{ 320, 314, 25 },
		{ 320, 315, 25 },
		{ 320, 316, 25 },
		{ 320, 317, 25 },
		{ 320, 318, 25 },
		{ 320, 319, 25 },
		{ 321, 290, 25 },
		{ 321, 291, 25 },
		{ 321, 292, 25 },
		{ 321, 293, 25 },
		{ 321, 294, 25 },
		{ 321, 295, 25 },
		{ 321, 296, 25 },
		{ 321, 297, 25 },
		{ 321, 298, 25 },
		{ 321, 299, 25 },
		{ 321, 300, 25 },
		{ 321, 301, 25 },
		{ 321, 302, 25 },
		{ 321, 303, 25 },
		{ 321, 304, 25 },
		{ 321, 305, 25 },
		{ 321, 306, 25 },
		{ 321, 307, 25 },
		{ 321, 308, 25 },
		{ 321, 309, 25 },
		{ 321, 310, 25 },
		{ 321, 311, 25 },
		{ 321, 312, 25 },
		{ 321, 313, 25 },
		{ 321, 314, 25 },
		{ 321, 315, 25 },
		{ 321, 316, 25 },
		{ 321, 317, 25 },
		{ 321, 318, 25 },
		{ 321, 319, 25 },
		{ 321, 320, 25 },
		{ 322, 290, 25 },
		{ 322, 291, 25 },
		{ 322, 292, 25 },
		{ 322, 293, 25 },
		{ 322, 294, 25 },
		{ 322, 295, 25 },
		{ 322, 296, 25 },
		{ 322, 297, 25 },
		{ 322, 298, 25 },
		{ 322, 299, 25 },
		{ 322, 300, 25 },
		{ 322, 301, 25 },
		{ 322, 302, 25 },
		{ 322, 303, 25 },
		{ 322, 304, 25 },
		{ 322, 305, 25 },
		{ 322, 306, 25 },
		{ 322, 307, 25 },
		{ 322, 308, 25 },
		{ 322, 309, 25 },
		{ 322, 310, 25 },
		{ 322, 311, 25 },
		{ 322, 312, 25 },
		{ 322, 313, 25 },
		{ 322, 314, 25 },
		{ 322, 315, 25 },
		{ 322, 316, 25 },
		{ 322, 317, 25 },
		{ 322, 318, 25 },
		{ 322, 319, 25 },
		{ 322, 320, 25 },
		{ 322, 321, 25 },
		{ 323, 291, 25 },
		{ 323, 292, 25 },
		{ 323, 293, 25 },
		{ 323, 294, 25 },
		{ 323, 295, 25 },
		{ 323, 296, 25 },
		{ 323, 297, 25 },
		{ 323, 298, 25 },
		{ 323, 299, 25 },
		{ 323, 300, 25 },
		{ 323, 301, 25 },
		{ 323, 302, 25 },
		{ 323, 303, 25 },
		{ 323, 304, 25 },
		{ 323, 305, 25 },
		{ 323, 306, 25 },
		{ 323, 307, 25 },
		{ 323, 308, 25 },
		{ 323, 309, 25 },
		{ 323, 310, 25 },
		{ 323, 311, 25 },
		{ 323, 312, 25 },
		{ 323, 313, 25 },
		{ 323, 314, 25 },
		{ 323, 315, 25 },
		{ 323, 316, 25 },
		{ 323, 317, 25 },
		{ 323, 318, 25 },
		{ 323, 319, 25 },
		{ 323, 320, 25 },
		{ 323, 321, 25 },
		{ 323, 322, 25 },
		{ 324, 291, 25 },
		{ 324, 292, 25 },
		{ 324, 293, 25 },
		{ 324, 294, 25 },
		{ 324, 295, 25 },
		{ 324, 296, 25 },
		{ 324, 297, 25 },
		{ 324, 298, 25 },
		{ 324, 299, 25 },
		{ 324, 300, 25 },
		{ 324, 301, 25 },
		{ 324, 302, 25 },
		{ 324, 303, 25 },
		{ 324, 304, 25 },
		{ 324, 305, 25 },
		{ 324, 306, 25 },
		{ 324, 307, 25 },
		{ 324, 308, 25 },
		{ 324, 309, 25 },
		{ 324, 310, 25 },
		{ 324, 311, 25 },
		{ 324, 312, 25 },
		{ 324, 313, 25 },
		{ 324, 314, 25 },
		{ 324, 315, 25 },
		{ 324, 316, 25 },
		{ 324, 317, 25 },
		{ 324, 318, 25 },
		{ 324, 319, 25 },
		{ 324, 320, 25 },
		{ 324, 321, 25 },
		{ 324, 322, 25 },
		{ 324, 323, 25 },
		{ 325, 292, 25 },
		{ 325, 293, 25 },
		{ 325, 294, 25 },
		{ 325, 295, 25 },
		{ 325, 296, 25 },
		{ 325, 297, 25 },
		{ 325, 298, 25 },
		{ 325, 299, 25 },
		{ 325, 300, 25 },
		{ 325, 301, 25 },
		{ 325, 302, 25 },
		{ 325, 303, 25 },
		{ 325, 304, 25 },
		{ 325, 305, 25 },
		{ 325, 306, 25 },
		{ 325, 307, 25 },
		{ 325, 308, 25 },
		{ 325, 309, 25 },
		{ 325, 310, 25 },
		{ 325, 311, 25 },
		{ 325, 312, 25 },
		{ 325, 313, 25 },
		{ 325, 314, 25 },
		{ 325, 315, 25 },
		{ 325, 316, 25 },
		{ 325, 317, 25 },
		{ 325, 318, 25 },
		{ 325, 319, 25 },
		{ 325, 320, 25 },
		{ 325, 321, 25 },
		{ 325, 322, 25 },
		{ 325, 323, 25 },
		{ 325, 324, 25 },
		{ 326, 292, 25 },
		{ 326, 293, 25 },
		{ 326, 294, 25 },
		{ 326, 295, 25 },
		{ 326, 296, 25 },
		{ 326, 297, 25 },
		{ 326, 298, 25 },
		{ 326, 299, 25 },
		{ 326, 300, 25 },
		{ 326, 301, 25 },
		{ 326, 302, 25 },
		{ 326, 303, 25 },
		{ 326, 304, 25 },
		{ 326, 305, 25 },
		{ 326, 306, 25 },
		{ 326, 307, 25 },
		{ 326, 308, 25 },
		{ 326, 309, 25 },
		{ 326, 310, 25 },
		{ 326, 311, 25 },
		{ 326, 312, 25 },
		{ 326, 313, 25 },
		{ 326, 314, 25 },
		{ 326, 315, 25 },
		{ 326, 316, 25 },
		{ 326, 317, 25 },
		{ 326, 318, 25 },
		{ 326, 319, 25 },
		{ 326, 320, 25 },
		{ 326, 321, 25 },
		{ 326, 322, 25 },
		{ 326, 323, 25 },
		{ 326, 324, 25 },
		{ 326, 325, 25 },
		{ 327, 293, 25 },
		{ 327, 294, 25 },
		{ 327, 295, 25 },
		{ 327, 296, 25 },
		{ 327, 297, 25 },
		{ 327, 298, 25 },
		{ 327, 299, 25 },
		{ 327, 300, 25 },
		{ 327, 301, 25 },
		{ 327, 302, 25 },
		{ 327, 303, 25 },
		{ 327, 304, 25 },
		{ 327, 305, 25 },
		{ 327, 306, 25 },
		{ 327, 307, 25 },
		{ 327, 308, 25 },
		{ 327, 309, 25 },
		{ 327, 310, 25 },
		{ 327, 311, 25 },
		{ 327, 312, 25 },
		{ 327, 313, 25 },
		{ 327, 314, 25 },
		{ 327, 315, 25 },
		{ 327, 316, 25 },
		{ 327, 317, 25 },
		{ 327, 318, 25 },
		{ 327, 319, 25 },
		{ 327, 320, 25 },
		{ 327, 321, 25 },
		{ 327, 322, 25 },
		{ 327, 323, 25 },
		{ 327, 324, 25 },
		{ 327, 325, 25 },
		{ 327, 326, 25 },
		{ 328, 293, 25 },
		{ 328, 294, 25 },
		{ 328, 295, 25 },
		{ 328, 296, 25 },
		{ 328, 297, 25 },
		{ 328, 298, 25 },
		{ 328, 299, 25 },
		{ 328, 300, 25 },
		{ 328, 301, 25 },
		{ 328, 302, 25 },
		{ 328, 303, 25 },
		{ 328, 304, 25 },
		{ 328, 305, 25 },
		{ 328, 306, 25 },
		{ 328, 307, 25 },
		{ 328, 308, 25 },
		{ 328, 309, 25 },
		{ 328, 310, 25 },
		{ 328, 311, 25 },
		{ 328, 312, 25 },
		{ 328, 313, 25 },
		{ 328, 314, 25 },
		{ 328, 315, 25 },
		{ 328, 316, 25 },
		{ 328, 317, 25 },
		{ 328, 318, 25 },
		{ 328, 319, 25 },
		{ 328, 320, 25 },
		{ 328, 321, 25 },
		{ 328, 322, 25 },
		{ 328, 323, 25 },
		{ 328, 324, 25 },
		{ 328, 325, 25 },
		{ 328, 326, 25 },
		{ 328, 327, 25 },
		{ 329, 294, 25 },
		{ 329, 295, 25 },
		{ 329, 296, 25 },
		{ 329, 297, 25 },
		{ 329, 298, 25 },
		{ 329, 299, 25 },
		{ 329, 300, 25 },
		{ 329, 301, 25 },
		{ 329, 302, 25 },
		{ 329, 303, 25 },
		{ 329, 304, 25 },
		{ 329, 305, 25 },
		{ 329, 306, 25 },
		{ 329, 307, 25 },
		{ 329, 308, 25 },
		{ 329, 309, 25 },
		{ 329, 310, 25 },
		{ 329, 311, 25 },
		{ 329, 312, 25 },
		{ 329, 313, 25 },
		{ 329, 314, 25 },
		{ 329, 315, 25 },
		{ 329, 316, 25 },
		{ 329, 317, 25 },
		{ 329, 318, 25 },
		{ 329, 319, 25 },
		{ 329, 320, 25 },
		{ 329, 321, 25 },
		{ 329, 322, 25 },
		{ 329, 323, 25 },
		{ 329, 324, 25 },
		{ 329, 325, 25 },
		{ 329, 326, 25 },
		{ 329, 327, 25 },
		{ 329, 328, 25 },
		{ 330, 294, 25 },
		{ 330, 295, 25 },
		{ 330, 296, 25 },
		{ 330, 297, 25 },
		{ 330, 298, 25 },
		{ 330, 299, 25 },
		{ 330, 300, 25 },
		{ 330, 301, 25 },
		{ 330, 302, 25 },
		{ 330, 303, 25 },
		{ 330, 304, 25 },
		{ 330, 305, 25 },
		{ 330, 306, 25 },
		{ 330, 307, 25 },
		{ 330, 308, 25 },
		{ 330, 309, 25 },
		{ 330, 310, 25 },
		{ 330, 311, 25 },
		{ 330, 312, 25 },
		{ 330, 313, 25 },
		{ 330, 314, 25 },
		{ 330, 315, 25 },
		{ 330, 316, 25 },
		{ 330, 317, 25 },
		{ 330, 318, 25 },
		{ 330, 319, 25 },
		{ 330, 320, 25 },
		{ 330, 321, 25 },
		{ 330, 322, 25 },
		{ 330, 323, 25 },
		{ 330, 324, 25 },
		{ 330, 325, 25 },
		{ 330, 326, 25 },
		{ 330, 327, 25 },
		{ 330, 328, 25 },
		{ 330, 329, 25 },
		{ 331, 295, 25 },
		{ 331, 296, 25 },
		{ 331, 297, 25 },
		{ 331, 298, 25 },
		{ 331, 299, 25 },
		{ 331, 300, 25 },
		{ 331, 301, 25 },
		{ 331, 302, 25 },
		{ 331, 303, 25 },
		{ 331, 304, 25 },
		{ 331, 305, 25 },
		{ 331, 306, 25 },
		{ 331, 307, 25 },
		{ 331, 308, 25 },
		{ 331, 309, 25 },
		{ 331, 310, 25 },
		{ 331, 311, 25 },
		{ 331, 312, 25 },
		{ 331, 313, 25 },
		{ 331, 314, 25 },
		{ 331, 315, 25 },
		{ 331, 316, 25 },
		{ 331, 317, 25 },
		{ 331, 318, 25 },
		{ 331, 319, 25 },
		{ 331, 320, 25 },
		{ 331, 321, 25 },
		{ 331, 322, 25 },
		{ 331, 323, 25 },
		{ 331, 324, 25 },
		{ 331, 325, 25 },
		{ 331, 326, 25 },
		{ 331, 327, 25 },
		{ 331, 328, 25 },
		{ 331, 329, 25 },
		{ 331, 330, 25 },
		{ 332, 295, 25 },
		{ 332, 296, 25 },
		{ 332, 297, 25 },
		{ 332, 298, 25 },
		{ 332, 299, 25 },
		{ 332, 300, 25 },
		{ 332, 301, 25 },
		{ 332, 302, 25 },
		{ 332, 303, 25 },
		{ 332, 304, 25 },
		{ 332, 305, 25 },
		{ 332, 306, 25 },
		{ 332, 307, 25 },
		{ 332, 308, 25 },
		{ 332, 309, 25 },
		{ 332, 310, 25 },
		{ 332, 311, 25 },
		{ 332, 312, 25 },
		{ 332, 313, 25 },
		{ 332, 314, 25 },
		{ 332, 315, 25 },
		{ 332, 316, 25 },
		{ 332, 317, 25 },
		{ 332, 318, 25 },
		{ 332, 319, 25 },
		{ 332, 320, 25 },
		{ 332, 321, 25 },
		{ 332, 322, 25 },
		{ 332, 323, 25 },
		{ 332, 324, 25 },
		{ 332, 325, 25 },
		{ 332, 326, 25 },
		{ 332, 327, 25 },
		{ 332, 328, 25 },
		{ 332, 329, 25 },
		{ 332, 330, 25 },
		{ 332, 331, 25 },
		{ 333, 296, 25 },
		{ 333, 297, 25 },
		{ 333, 298, 25 },
		{ 333, 299, 25 },
		{ 333, 300, 25 },
		{ 333, 301, 25 },
		{ 333, 302, 25 },
		{ 333, 303, 25 },
		{ 333, 304, 25 },
		{ 333, 305, 25 },
		{ 333, 306, 25 },
		{ 333, 307, 25 },
		{ 333, 308, 25 },
		{ 333, 309, 25 },
		{ 333, 310, 25 },
		{ 333, 311, 25 },
		{ 333, 312, 25 },
		{ 333, 313, 25 },
		{ 333, 314, 25 },
		{ 333, 315, 25 },
		{ 333, 316, 25 },
		{ 333, 317, 25 },
		{ 333, 318, 25 },
		{ 333, 319, 25 },
		{ 333, 320, 25 },
		{ 333, 321, 25 },
		{ 333, 322, 25 },
		{ 333, 323, 25 },
		{ 333, 324, 25 },
		{ 333, 325, 25 },
		{ 333, 326, 25 },
		{ 333, 327, 25 },
		{ 333, 328, 25 },
		{ 333, 329, 25 },
		{ 333, 330, 25 },
		{ 333, 331, 25 },
		{ 333, 332, 25 },
		{ 334, 296, 25 },
		{ 334, 297, 25 },
		{ 334, 298, 25 },
		{ 334, 299, 25 },
		{ 334, 300, 25 },
		{ 334, 301, 25 },
		{ 334, 302, 25 },
		{ 334, 303, 25 },
		{ 334, 304, 25 },
		{ 334, 305, 25 },
		{ 334, 306, 25 },
		{ 334, 307, 25 },
		{ 334, 308, 25 },
		{ 334, 309, 25 },
		{ 334, 310, 25 },
		{ 334, 311, 25 },
		{ 334, 312, 25 },
		{ 334, 313, 25 },
		{ 334, 314, 25 },
		{ 334, 315, 25 },
		{ 334, 316, 25 },
		{ 334, 317, 25 },
		{ 334, 318, 25 },
		{ 334, 319, 25 },
		{ 334, 320, 25 },
		{ 334, 321, 25 },
		{ 334, 322, 25 },
		{ 334, 323, 25 },
		{ 334, 324, 25 },
		{ 334, 325, 25 },
		{ 334, 326, 25 },
		{ 334, 327, 25 },
		{ 334, 328, 25 },
		{ 334, 329, 25 },
		{ 334, 330, 25 },
		{ 334, 331, 25 },
		{ 334, 332, 25 },
		{ 334, 333, 25 },
		{ 335, 297, 25 },
		{ 335, 298, 25 },
		{ 335, 299, 25 },
		{ 335, 300, 25 },
		{ 335, 301, 25 },
		{ 335, 302, 25 },
		{ 335, 303, 25 },
		{ 335, 304, 25 },
		{ 335, 305, 25 },
		{ 335, 306, 25 },
		{ 335, 307, 25 },
		{ 335, 308, 25 },
		{ 335, 309, 25 },
		{ 335, 310, 25 },
		{ 335, 311, 25 },
		{ 335, 312, 25 },
		{ 335, 313, 25 },
		{ 335, 314, 25 },
		{ 335, 315, 25 },
		{ 335, 316, 25 },
		{ 335, 317, 25 },
		{ 335, 318, 25 },
		{ 335, 319, 25 },
		{ 335, 320, 25 },
		{ 335, 321, 25 },
		{ 335, 322, 25 },
		{ 335, 323, 25 },
		{ 335, 324, 25 },
		{ 335, 325, 25 },
		{ 335, 326, 25 },
		{ 335, 327, 25 },
		{ 335, 328, 25 },
		{ 335, 329, 25 },
		{ 335, 330, 25 },
		{ 335, 331, 25 },
		{ 335, 332, 25 },
		{ 335, 333, 25 },
		{ 335, 334, 25 },
		{ 336, 297, 25 },
		{ 336, 298, 25 },
		{ 336, 299, 25 },
		{ 336, 300, 25 },
		{ 336, 301, 25 },
		{ 336, 302, 25 },
		{ 336, 303, 25 },
		{ 336, 304, 25 },
		{ 336, 305, 25 },
		{ 336, 306, 25 },
		{ 336, 307, 25 },
		{ 336, 308, 25 },
		{ 336, 309, 25 },
		{ 336, 310, 25 },
		{ 336, 311, 25 },
		{ 336, 312, 25 },
		{ 336, 313, 25 },
		{ 336, 314, 25 },
		{ 336, 315, 25 },
		{ 336, 316, 25 },
		{ 336, 317, 25 },
		{ 336, 318, 25 },
		{ 336, 319, 25 },
		{ 336, 320, 25 },
		{ 336, 321, 25 },
		{ 336, 322, 25 },
		{ 336, 323, 25 },
		{ 336, 324, 25 },
		{ 336, 325, 25 },
		{ 336, 326, 25 },
		{ 336, 327, 25 },
		{ 336, 328, 25 },
		{ 336, 329, 25 },
		{ 336, 330, 25 },
		{ 336, 331, 25 },
		{ 336, 332, 25 },
		{ 336, 333, 25 },
		{ 336, 334, 25 },
		{ 336, 335, 25 },
		{ 337, 298, 25 },
		{ 337, 299, 25 },
		{ 337, 300, 25 },
		{ 337, 301, 25 },
		{ 337, 302, 25 },
		{ 337, 303, 25 },
		{ 337, 304, 25 },
		{ 337, 305, 25 },
		{ 337, 306, 25 },
		{ 337, 307, 25 },
		{ 337, 308, 25 },
		{ 337, 309, 25 },
		{ 337, 310, 25 },
		{ 337, 311, 25 },
		{ 337, 312, 25 },
		{ 337, 313, 25 },
		{ 337, 314, 25 },
		{ 337, 315, 25 },
		{ 337, 316, 25 },
		{ 337, 317, 25 },
		{ 337, 318, 25 },
		{ 337, 319, 25 },
		{ 337, 320, 25 },
		{ 337, 321, 25 },
		{ 337, 322, 25 },
		{ 337, 323, 25 },
		{ 337, 324, 25 },
		{ 337, 325, 25 },
		{ 337, 326, 25 },
		{ 337, 327, 25 },
		{ 337, 328, 25 },
		{ 337, 329, 25 },
		{ 337, 330, 25 },
		{ 337, 331, 25 },
		{ 337, 332, 25 },
		{ 337, 333, 25 },
		{ 337, 334, 25 },
		{ 337, 335, 25 },
		{ 337, 336, 25 },
		{ 338, 298, 25 },
		{ 338, 299, 25 },
		{ 338, 300, 25 },
		{ 338, 301, 25 },
		{ 338, 302, 25 },
		{ 338, 303, 25 },
		{ 338, 304, 25 },
		{ 338, 305, 25 },
		{ 338, 306, 25 },
		{ 338, 307, 25 },
		{ 338, 308, 25 },
		{ 338, 309, 25 },
		{ 338, 310, 25 },
		{ 338, 311, 25 },
		{ 338, 312, 25 },
		{ 338, 313, 25 },
		{ 338, 314, 25 },
		{ 338, 315, 25 },
		{ 338, 316, 25 },
		{ 338, 317, 25 },
		{ 338, 318, 25 },
		{ 338, 319, 25 },
		{ 338, 320, 25 },
		{ 338, 321, 25 },
		{ 338, 322, 25 },
		{ 338, 323, 25 },
		{ 338, 324, 25 },
		{ 338, 325, 25 },
		{ 338, 326, 25 },
		{ 338, 327, 25 },
		{ 338, 328, 25 },
		{ 338, 329, 25 },
		{ 338, 330, 25 },
		{ 338, 331, 25 },
		{ 338, 332, 25 },
		{ 338, 333, 25 },
		{ 338, 334, 25 },
		{ 338, 335, 25 },
		{ 338, 336, 25 },
		{ 338, 337, 25 },
		{ 339, 299, 25 },
		{ 339, 300, 25 },
		{ 339, 301, 25 },
		{ 339, 302, 25 },
		{ 339, 303, 25 },
		{ 339, 304, 25 },
		{ 339, 305, 25 },
		{ 339, 306, 25 },
		{ 339, 307, 25 },
		{ 339, 308, 25 },
		{ 339, 309, 25 },
		{ 339, 310, 25 },
		{ 339, 311, 25 },
		{ 339, 312, 25 },
		{ 339, 313, 25 },
		{ 339, 314, 25 },
		{ 339, 315, 25 },
		{ 339, 316, 25 },
		{ 339, 317, 25 },
		{ 339, 318, 25 },
		{ 339, 319, 25 },
		{ 339, 320, 25 },
		{ 339, 321, 25 },
		{ 339, 322, 25 },
		{ 339, 323, 25 },
		{ 339, 324, 25 },
		{ 339, 325, 25 },
		{ 339, 326, 25 },
		{ 339, 327, 25 },
		{ 339, 328, 25 },
		{ 339, 329, 25 },
		{ 339, 330, 25 },
		{ 339, 331, 25 },
		{ 339, 332, 25 },
		{ 339, 333, 25 },
		{ 339, 334, 25 },
		{ 339, 335, 25 },
		{ 339, 336, 25 },
		{ 339, 337, 25 },
		{ 339, 338, 25 },
		{ 340, 299, 25 },
		{ 340, 300, 25 },
		{ 340, 301, 25 },
		{ 340, 302, 25 },
		{ 340, 303, 25 },
		{ 340, 304, 25 },
		{ 340, 305, 25 },
		{ 340, 306, 25 },
		{ 340, 307, 25 },
		{ 340, 308, 25 },
		{ 340, 309, 25 },
		{ 340, 310, 25 },
		{ 340, 311, 25 },
		{ 340, 312, 25 },
		{ 340, 313, 25 },
		{ 340, 314, 25 },
		{ 340, 315, 25 },
		{ 340, 316, 25 },
		{ 340, 317, 25 },
		{ 340, 318, 25 },
		{ 340, 319, 25 },
		{ 340, 320, 25 },
		{ 340, 321, 25 },
		{ 340, 322, 25 },
		{ 340, 323, 25 },
		{ 340, 324, 25 },
		{ 340, 325, 25 },
		{ 340, 326, 25 },
		{ 340, 327, 25 },
		{ 340, 328, 25 },
		{ 340, 329, 25 },
		{ 340, 330, 25 },
		{ 340, 331, 25 },
		{ 340, 332, 25 },
		{ 340, 333, 25 },
		{ 340, 334, 25 },
		{ 340, 335, 25 },
		{ 340, 336, 25 },
		{ 340, 337, 25 },
		{ 340, 338, 25 },
		{ 340, 339, 25 },
		{ 341, 300, 25 },
		{ 341, 301, 25 },
		{ 341, 302, 25 },
		{ 341, 303, 25 },
		{ 341, 304, 25 },
		{ 341, 305, 25 },
		{ 341, 306, 25 },
		{ 341, 307, 25 },
		{ 341, 308, 25 },
		{ 341, 309, 25 },
		{ 341, 310, 25 },
		{ 341, 311, 25 },
		{ 341, 312, 25 },
		{ 341, 313, 25 },
		{ 341, 314, 25 },
		{ 341, 315, 25 },
		{ 341, 316, 25 },
		{ 341, 317, 25 },
		{ 341, 318, 25 },
		{ 341, 319, 25 },
		{ 341, 320, 25 },
		{ 341, 321, 25 },
		{ 341, 322, 25 },
		{ 341, 323, 25 },
		{ 341, 324, 25 },
		{ 341, 325, 25 },
		{ 341, 326, 25 },
		{ 341, 327, 25 },
		{ 341, 328, 25 },
		{ 341, 329, 25 },
		{ 341, 330, 25 },
		{ 341, 331, 25 },
		{ 341, 332, 25 },
		{ 341, 333, 25 },
		{ 341, 334, 25 },
		{ 341, 335, 25 },
		{ 341, 336, 25 },
		{ 341, 337, 25 },
		{ 341, 338, 25 },
		{ 341, 339, 25 },
		{ 341, 340, 25 },
		{ 342, 300, 25 },
		{ 342, 301, 25 },
		{ 342, 302, 25 },
		{ 342, 303, 25 },
		{ 342, 304, 25 },
		{ 342, 305, 25 },
		{ 342, 306, 25 },
		{ 342, 307, 25 },
		{ 342, 308, 25 },
		{ 342, 309, 25 },
		{ 342, 310, 25 },
		{ 342, 311, 25 },
		{ 342, 312, 25 },
		{ 342, 313, 25 },
		{ 342, 314, 25 },
		{ 342, 315, 25 },
		{ 342, 316, 25 },
		{ 342, 317, 25 },
		{ 342, 318, 25 },
		{ 342, 319, 25 },
		{ 342, 320, 25 },
		{ 342, 321, 25 },
		{ 342, 322, 25 },
		{ 342, 323, 25 },
		{ 342, 324, 25 },
		{ 342, 325, 25 },
		{ 342, 326, 25 },
		{ 342, 327, 25 },
		{ 342, 328, 25 },
		{ 342, 329, 25 },
		{ 342, 330, 25 },
		{ 342, 331, 25 },
		{ 342, 332, 25 },
		{ 342, 333, 25 },
		{ 342, 334, 25 },
		{ 342, 335, 25 },
		{ 342, 336, 25 },
		{ 342, 337, 25 },
		{ 342, 338, 25 },
		{ 342, 339, 25 },
		{ 342, 340, 25 },
		{ 342, 341, 25 },
		{ 343, 301, 25 },
		{ 343, 302, 25 },
		{ 343, 303, 25 },
		{ 343, 304, 25 },
		{ 343, 305, 25 },
		{ 343, 306, 25 },
		{ 343, 307, 25 },
		{ 343, 308, 25 },
		{ 343, 309, 25 },
		{ 343, 310, 25 },
		{ 343, 311, 25 },
		{ 343, 312, 25 },
		{ 343, 313, 25 },
		{ 343, 314, 25 },
		{ 343, 315, 25 },
		{ 343, 316, 25 },
		{ 343, 317, 25 },
		{ 343, 318, 25 },
		{ 343, 319, 25 },
		{ 343, 320, 25 },
		{ 343, 321, 25 },
		{ 343, 322, 25 },
		{ 343, 323, 25 },
		{ 343, 324, 25 },
		{ 343, 325, 25 },
		{ 343, 326, 25 },
		{ 343, 327, 25 },
		{ 343, 328, 25 },
		{ 343, 329, 25 },
		{ 343, 330, 25 },
		{ 343, 331, 25 },
		{ 343, 332, 25 },
		{ 343, 333, 25 },
		{ 343, 334, 25 },
		{ 343, 335, 25 },
		{ 343, 336, 25 },
		{ 343, 337, 25 },
		{ 343, 338, 25 },
		{ 343, 339, 25 },
		{ 343, 340, 25 },
		{ 343, 341, 25 },
		{ 343, 342, 25 },
		{ 344, 301, 25 },
		{ 344, 302, 25 },
		{ 344, 303, 25 },
		{ 344, 304, 25 },
		{ 344, 305, 25 },
		{ 344, 306, 25 },
		{ 344, 307, 25 },
		{ 344, 308, 25 },
		{ 344, 309, 25 },
		{ 344, 310, 25 },
		{ 344, 311, 25 },
		{ 344, 312, 25 },
		{ 344, 313, 25 },
		{ 344, 314, 25 },
		{ 344, 315, 25 },
		{ 344, 316, 25 },
		{ 344, 317, 25 },
		{ 344, 318, 25 },
		{ 344, 319, 25 },
		{ 344, 320, 25 },
		{ 344, 321, 25 },
		{ 344, 322, 25 },
		{ 344, 323, 25 },
		{ 344, 324, 25 },
		{ 344, 325, 25 },
		{ 344, 326, 25 },
		{ 344, 327, 25 },
		{ 344, 328, 25 },
		{ 344, 329, 25 },
		{ 344, 330, 25 },
		{ 344, 331, 25 },
		{ 344, 332, 25 },
		{ 344, 333, 25 },
		{ 344, 334, 25 },
		{ 344, 335, 25 },
		{ 344, 336, 25 },
		{ 344, 337, 25 },
		{ 344, 338, 25 },
		{ 344, 339, 25 },
		{ 344, 340, 25 },
		{ 344, 341, 25 },
		{ 344, 342, 25 },
		{ 344, 343, 25 },
		{ 345, 302, 25 },
		{ 345, 303, 25 },
		{ 345, 304, 25 },
		{ 345, 305, 25 },
		{ 345, 306, 25 },
		{ 345, 307, 25 },
		{ 345, 308, 25 },
		{ 345, 309, 25 },
		{ 345, 310, 25 },
		{ 345, 311, 25 },
		{ 345, 312, 25 },
		{ 345, 313, 25 },
		{ 345, 314, 25 },
		{ 345, 315, 25 },
		{ 345, 316, 25 },
		{ 345, 317, 25 },
		{ 345, 318, 25 },
		{ 345, 319, 25 },
		{ 345, 320, 25 },
		{ 345, 321, 25 },
		{ 345, 322, 25 },
		{ 345, 323, 25 },
		{ 345, 324, 25 },
		{ 345, 325, 25 },
		{ 345, 326, 25 },
		{ 345, 327, 25 },
		{ 345, 328, 25 },
		{ 345, 329, 25 },
		{ 345, 330, 25 },
		{ 345, 331, 25 },
		{ 345, 332, 25 },
		{ 345, 333, 25 },
		{ 345, 334, 25 },
		{ 345, 335, 25 },
		{ 345, 336, 25 },
		{ 345, 337, 25 },
		{ 345, 338, 25 },
		{ 345, 339, 25 },
		{ 345, 340, 25 },
		{ 345, 341, 25 },
		{ 345, 342, 25 },
		{ 345, 343, 25 },
		{ 345, 344, 25 },
		{ 346, 302, 25 },
		{ 346, 303, 25 },
		{ 346, 304, 25 },
		{ 346, 305, 25 },
		{ 346, 306, 25 },
		{ 346, 307, 25 },
		{ 346, 308, 25 },
		{ 346, 309, 25 },
		{ 346, 310, 25 },
		{ 346, 311, 25 },
		{ 346, 312, 25 },
		{ 346, 313, 25 },
		{ 346, 314, 25 },
		{ 346, 315, 25 },
		{ 346, 316, 25 },
		{ 346, 317, 25 },
		{ 346, 318, 25 },
		{ 346, 319, 25 },
		{ 346, 320, 25 },
		{ 346, 321, 25 },
		{ 346, 322, 25 },
		{ 346, 323, 25 },
		{ 346, 324, 25 },
		{ 346, 325, 25 },
		{ 346, 326, 25 },
		{ 346, 327, 25 },
		{ 346, 328, 25 },
		{ 346, 329, 25 },
		{ 346, 330, 25 },
		{ 346, 331, 25 },
		{ 346, 332, 25 },
		{ 346, 333, 25 },
		{ 346, 334, 25 },
		{ 346, 335, 25 },
		{ 346, 336, 25 },
		{ 346, 337, 25 },
		{ 346, 338, 25 },
		{ 346, 339, 25 },
		{ 346, 340, 25 },
		{ 346, 341, 25 },
		{ 346, 342, 25 },
		{ 346, 343, 25 },
		{ 346, 344, 25 },
		{ 346, 345, 25 },
		{ 347, 303, 25 },
		{ 347, 304, 25 },
		{ 347, 305, 25 },
		{ 347, 306, 25 },
		{ 347, 307, 25 },
		{ 347, 308, 25 },
		{ 347, 309, 25 },
		{ 347, 310, 25 },
		{ 347, 311, 25 },
		{ 347, 312, 25 },
		{ 347, 313, 25 },
		{ 347, 314, 25 },
		{ 347, 315, 25 },
		{ 347, 316, 25 },
		{ 347, 317, 25 },
		{ 347, 318, 25 },
		{ 347, 319, 25 },
		{ 347, 320, 25 },
		{ 347, 321, 25 },
		{ 347, 322, 25 },
		{ 347, 323, 25 },
		{ 347, 324, 25 },
		{ 347, 325, 25 },
		{ 347, 326, 25 },
		{ 347, 327, 25 },
		{ 347, 328, 25 },
		{ 347, 329, 25 },
		{ 347, 330, 25 },
		{ 347, 331, 25 },
		{ 347, 332, 25 },
		{ 347, 333, 25 },
		{ 347, 334, 25 },
		{ 347, 335, 25 },
		{ 347, 336, 25 },
		{ 347, 337, 25 },
		{ 347, 338, 25 },
		{ 347, 339, 25 },
		{ 347, 340, 25 },
		{ 347, 341, 25 },
		{ 347, 342, 25 },
		{ 347, 343, 25 },
		{ 347, 344, 25 },
		{ 347, 345, 25 },
		{ 347, 346, 25 },
		{ 348, 303, 25 },
		{ 348, 304, 25 },
		{ 348, 305, 25 },
		{ 348, 306, 25 },
		{ 348, 307, 25 },
		{ 348, 308, 25 },
		{ 348, 309, 25 },
		{ 348, 310, 25 },
		{ 348, 311, 25 },
		{ 348, 312, 25 },
		{ 348, 313, 25 },
		{ 348, 314, 25 },
		{ 348, 315, 25 },
		{ 348, 316, 25 },
		{ 348, 317, 25 },
		{ 348, 318, 25 },
		{ 348, 319, 25 },
		{ 348, 320, 25 },
		{ 348, 321, 25 },
		{ 348, 322, 25 },
		{ 348, 323, 25 },
		{ 348, 324, 25 },
		{ 348, 325, 25 },
		{ 348, 326, 25 },
		{ 348, 327, 25 },
		{ 348, 328, 25 },
		{ 348, 329, 25 },
		{ 348, 330, 25 },
		{ 348, 331, 25 },
		{ 348, 332, 25 },
		{ 348, 333, 25 },
		{ 348, 334, 25 },
		{ 348, 335, 25 },
		{ 348, 336, 25 },
		{ 348, 337, 25 },
		{ 348, 338, 25 },
		{ 348, 339, 25 },
		{ 348, 340, 25 },
		{ 348, 341, 25 },
		{ 348, 342, 25 },
		{ 348, 343, 25 },
		{ 348, 344, 25 },
		{ 348, 345, 25 },
		{ 348, 346, 25 },
		{ 348, 347, 25 },
		{ 349, 304, 25 },
		{ 349, 305, 25 },
		{ 349, 306, 25 },
		{ 349, 307, 25 },
		{ 349, 308, 25 },
		{ 349, 309, 25 },
		{ 349, 310, 25 },
		{ 349, 311, 25 },
		{ 349, 312, 25 },
		{ 349, 313, 25 },
		{ 349, 314, 25 },
		{ 349, 315, 25 },
		{ 349, 316, 25 },
		{ 349, 317, 25 },
		{ 349, 318, 25 },
		{ 349, 319, 25 },
		{ 349, 320, 25 },
		{ 349, 321, 25 },
		{ 349, 322, 25 },
		{ 349, 323, 25 },
		{ 349, 324, 25 },
		{ 349, 325, 25 },
		{ 349, 326, 25 },
		{ 349, 327, 25 },
		{ 349, 328, 25 },
		{ 349, 329, 25 },
		{ 349, 330, 25 },
		{ 349, 331, 25 },
		{ 349, 332, 25 },
		{ 349, 333, 25 },
		{ 349, 334, 25 },
		{ 349, 335, 25 },
		{ 349, 336, 25 },
		{ 349, 337, 25 },
		{ 349, 338, 25 },
		{ 349, 339, 25 },
		{ 349, 340, 25 },
		{ 349, 341, 25 },
		{ 349, 342, 25 },
		{ 349, 343, 25 },
		{ 349, 344, 25 },
		{ 349, 345, 25 },
		{ 349, 346, 25 },
		{ 349, 347, 25 },
		{ 349, 348, 25 },
		{ 350, 304, 25 },
		{ 350, 305, 25 },
		{ 350, 306, 25 },
		{ 350, 307, 25 },
		{ 350, 308, 25 },
		{ 350, 309, 25 },
		{ 350, 310, 25 },
		{ 350, 311, 25 },
		{ 350, 312, 25 },
		{ 350, 313, 25 },
		{ 350, 314, 25 },
		{ 350, 315, 25 },
		{ 350, 316, 25 },
		{ 350, 317, 25 },
		{ 350, 318, 25 },
		{ 350, 319, 25 },
		{ 350, 320, 25 },
		{ 350, 321, 25 },
		{ 350, 322, 25 },
		{ 350, 323, 25 },
		{ 350, 324, 25 },
		{ 350, 325, 25 },
		{ 350, 326, 25 },
		{ 350, 327, 25 },
		{ 350, 328, 25 },
		{ 350, 329, 25 },
		{ 350, 330, 25 },
		{ 350, 331, 25 },
		{ 350, 332, 25 },
		{ 350, 333, 25 },
		{ 350, 334, 25 },
		{ 350, 335, 25 },
		{ 350, 336, 25 },
		{ 350, 337, 25 },
		{ 350, 338, 25 },
		{ 350, 339, 25 },
		{ 350, 340, 25 },
		{ 350, 341, 25 },
		{ 350, 342, 25 },
		{ 350, 343, 25 },
		{ 350, 344, 25 },
		{ 350, 345, 25 },
		{ 350, 346, 25 },
		{ 350, 347, 25 },
		{ 350, 348, 25 },
		{ 350, 349, 25 },
		{ 351, 305, 25 },
		{ 351, 306, 25 },
		{ 351, 307, 25 },
		{ 351, 308, 25 },
		{ 351, 309, 25 },
		{ 351, 310, 25 },
		{ 351, 311, 25 },
		{ 351, 312, 25 },
		{ 351, 313, 25 },
		{ 351, 314, 25 },
		{ 351, 315, 25 },
		{ 351, 316, 25 },
		{ 351, 317, 25 },
		{ 351, 318, 25 },
		{ 351, 319, 25 },
		{ 351, 320, 25 },
		{ 351, 321, 25 },
		{ 351, 322, 25 },
		{ 351, 323, 25 },
		{ 351, 324, 25 },
		{ 351, 325, 25 },
		{ 351, 326, 25 },
		{ 351, 327, 25 },
		{ 351, 328, 25 },
		{ 351, 329, 25 },
		{ 351, 330, 25 },
		{ 351, 331, 25 },
		{ 351, 332, 25 },
		{ 351, 333, 25 },
		{ 351, 334, 25 },
		{ 351, 335, 25 },
		{ 351, 336, 25 },
		{ 351, 337, 25 },
		{ 351, 338, 25 },
		{ 351, 339, 25 },
		{ 351, 340, 25 },
		{ 351, 341, 25 },
		{ 351, 342, 25 },
		{ 351, 343, 25 },
		{ 351, 344, 25 },
		{ 351, 345, 25 },
		{ 351, 346, 25 },
		{ 351, 347, 25 },
		{ 351, 348, 25 },
		{ 351, 349, 25 },
		{ 351, 350, 25 },
		{ 352, 305, 25 },
		{ 352, 306, 25 },
		{ 352, 307, 25 },
		{ 352, 308, 25 },
		{ 352, 309, 25 },
		{ 352, 310, 25 },
		{ 352, 311, 25 },
		{ 352, 312, 25 },
		{ 352, 313, 25 },
		{ 352, 314, 25 },
		{ 352, 315, 25 },
		{ 352, 316, 25 },
		{ 352, 317, 25 },
		{ 352, 318, 25 },
		{ 352, 319, 25 },
		{ 352, 320, 25 },
		{ 352, 321, 25 },
		{ 352, 322, 25 },
		{ 352, 323, 25 },
		{ 352, 324, 25 },
		{ 352, 325, 25 },
		{ 352, 326, 25 },
		{ 352, 327, 25 },
		{ 352, 328, 25 },
		{ 352, 329, 25 },
		{ 352, 330, 25 },
		{ 352, 331, 25 },
		{ 352, 332, 25 },
		{ 352, 333, 25 },
		{ 352, 334, 25 },
		{ 352, 335, 25 },
		{ 352, 336, 25 },
		{ 352, 337, 25 },
		{ 352, 338, 25 },
		{ 352, 339, 25 },
		{ 352, 340, 25 },
		{ 352, 341, 25 },
		{ 352, 342, 25 },
		{ 352, 343, 25 },
		{ 352, 344, 25 },
		{ 352, 345, 25 },
		{ 352, 346, 25 },
		{ 352, 347, 25 },
		{ 352, 348, 25 },
		{ 352, 349, 25 },
		{ 352, 350, 25 },
		{ 352, 351, 25 },
		{ 353, 306, 25 },
		{ 353, 307, 25 },
		{ 353, 308, 25 },
		{ 353, 309, 25 },
		{ 353, 310, 25 },
		{ 353, 311, 25 },
		{ 353, 312, 25 },
		{ 353, 313, 25 },
		{ 353, 314, 25 },
		{ 353, 315, 25 },
		{ 353, 316, 25 },
		{ 353, 317, 25 },
		{ 353, 318, 25 },
		{ 353, 319, 25 },
		{ 353, 320, 25 },
		{ 353, 321, 25 },
		{ 353, 322, 25 },
		{ 353, 323, 25 },
		{ 353, 324, 25 },
		{ 353, 325, 25 },
		{ 353, 326, 25 },
		{ 353, 327, 25 },
		{ 353, 328, 25 },
		{ 353, 329, 25 },
		{ 353, 330, 25 },
		{ 353, 331, 25 },
		{ 353, 332, 25 },
		{ 353, 333, 25 },
		{ 353, 334, 25 },
		{ 353, 335, 25 },
		{ 353, 336, 25 },
		{ 353, 337, 25 },
		{ 353, 338, 25 },
		{ 353, 339, 25 },
		{ 353, 340, 25 },
		{ 353, 341, 25 },
		{ 353, 342, 25 },
		{ 353, 343, 25 },
		{ 353, 344, 25 },
		{ 353, 345, 25 },
		{ 353, 346, 25 },
		{ 353, 347, 25 },
		{ 353, 348, 25 },
		{ 353, 349, 25 },
		{ 353, 350, 25 },
		{ 353, 351, 25 },
		{ 353, 352, 25 },
		{ 354, 306, 25 },
		{ 354, 307, 25 },
		{ 354, 308, 25 },
		{ 354, 309, 25 },
		{ 354, 310, 25 },
		{ 354, 311, 25 },
		{ 354, 312, 25 },
		{ 354, 313, 25 },
		{ 354, 314, 25 },
		{ 354, 315, 25 },
		{ 354, 316, 25 },
		{ 354, 317, 25 },
		{ 354, 318, 25 },
		{ 354, 319, 25 },
		{ 354, 320, 25 },
		{ 354, 321, 25 },
		{ 354, 322, 25 },
		{ 354, 323, 25 },
		{ 354, 324, 25 },
		{ 354, 325, 25 },
		{ 354, 326, 25 },
		{ 354, 327, 25 },
		{ 354, 328, 25 },
		{ 354, 329, 25 },
		{ 354, 330, 25 },
		{ 354, 331, 25 },
		{ 354, 332, 25 },
		{ 354, 333, 25 },
		{ 354, 334, 25 },
		{ 354, 335, 25 },
		{ 354, 336, 25 },
		{ 354, 337, 25 },
		{ 354, 338, 25 },
		{ 354, 339, 25 },
		{ 354, 340, 25 },
		{ 354, 341, 25 },
		{ 354, 342, 25 },
		{ 354, 343, 25 },
		{ 354, 344, 25 },
		{ 354, 345, 25 },
		{ 354, 346, 25 },
		{ 354, 347, 25 },
		{ 354, 348, 25 },
		{ 354, 349, 25 },
		{ 354, 350, 25 },
		{ 354, 351, 25 },
		{ 354, 352, 25 },
		{ 354, 353, 25 },
		{ 355, 307, 25 },
		{ 355, 308, 25 },
		{ 355, 309, 25 },
		{ 355, 310, 25 },
		{ 355, 311, 25 },
		{ 355, 312, 25 },
		{ 355, 313, 25 },
		{ 355, 314, 25 },
		{ 355, 315, 25 },
		{ 355, 316, 25 },
		{ 355, 317, 25 },
		{ 355, 318, 25 },
		{ 355, 319, 25 },
		{ 355, 320, 25 },
		{ 355, 321, 25 },
		{ 355, 322, 25 },
		{ 355, 323, 25 },
		{ 355, 324, 25 },
		{ 355, 325, 25 },
		{ 355, 326, 25 },
		{ 355, 327, 25 },
		{ 355, 328, 25 },
		{ 355, 329, 25 },
		{ 355, 330, 25 },
		{ 355, 331, 25 },
		{ 355, 332, 25 },
		{ 355, 333, 25 },
		{ 355, 334, 25 },
		{ 355, 335, 25 },
		{ 355, 336, 25 },
		{ 355, 337, 25 },
		{ 355, 338, 25 },
		{ 355, 339, 25 },
		{ 355, 340, 25 },
		{ 355, 341, 25 },
		{ 355, 342, 25 },
		{ 355, 343, 25 },
		{ 355, 344, 25 },
		{ 355, 345, 25 },
		{ 355, 346, 25 },
		{ 355, 347, 25 },
		{ 355, 348, 25 },
		{ 355, 349, 25 },
		{ 355, 350, 25 },
		{ 355, 351, 25 },
		{ 355, 352, 25 },
		{ 355, 353, 25 },
		{ 355, 354, 25 },
		{ 356, 307, 25 },
		{ 356, 308, 25 },
		{ 356, 309, 25 },
		{ 356, 310, 25 },
		{ 356, 311, 25 },
		{ 356, 312, 25 },
		{ 356, 313, 25 },
		{ 356, 314, 25 },
		{ 356, 315, 25 },
		{ 356, 316, 25 },
		{ 356, 317, 25 },
		{ 356, 318, 25 },
		{ 356, 319, 25 },
		{ 356, 320, 25 },
		{ 356, 321, 25 },
		{ 356, 322, 25 },
		{ 356, 323, 25 },
		{ 356, 324, 25 },
		{ 356, 325, 25 },
		{ 356, 326, 25 },
		{ 356, 327, 25 },
		{ 356, 328, 25 },
		{ 356, 329, 25 },
		{ 356, 330, 25 },
		{ 356, 331, 25 },
		{ 356, 332, 25 },
		{ 356, 333, 25 },
		{ 356, 334, 25 },
		{ 356, 335, 25 },
		{ 356, 336, 25 },
		{ 356, 337, 25 },
		{ 356, 338, 25 },
		{ 356, 339, 25 },
		{ 356, 340, 25 },
		{ 356, 341, 25 },
		{ 356, 342, 25 },
		{ 356, 343, 25 },
		{ 356, 344, 25 },
		{ 356, 345, 25 },
		{ 356, 346, 25 },
		{ 356, 347, 25 },
		{ 356, 348, 25 },
		{ 356, 349, 25 },
		{ 356, 350, 25 },
		{ 356, 351, 25 },
		{ 356, 352, 25 },
		{ 356, 353, 25 },
		{ 356, 354, 25 },
		{ 356, 355, 25 },
		{ 357, 308, 25 },
		{ 357, 309, 25 },
		{ 357, 310, 25 },
		{ 357, 311, 25 },
		{ 357, 312, 25 },
		{ 357, 313, 25 },
		{ 357, 314, 25 },
		{ 357, 315, 25 },
		{ 357, 316, 25 },
		{ 357, 317, 25 },
		{ 357, 318, 25 },
		{ 357, 319, 25 },
		{ 357, 320, 25 },
		{ 357, 321, 25 },
		{ 357, 322, 25 },
		{ 357, 323, 25 },
		{ 357, 324, 25 },
		{ 357, 325, 25 },
		{ 357, 326, 25 },
		{ 357, 327, 25 },
		{ 357, 328, 25 },
		{ 357, 329, 25 },
		{ 357, 330, 25 },
		{ 357, 331, 25 },
		{ 357, 332, 25 },
		{ 357, 333, 25 },
		{ 357, 334, 25 },
		{ 357, 335, 25 },
		{ 357, 336, 25 },
		{ 357, 337, 25 },
		{ 357, 338, 25 },
		{ 357, 339, 25 },
		{ 357, 340, 25 },
		{ 357, 341, 25 },
		{ 357, 342, 25 },
		{ 357, 343, 25 },
		{ 357, 344, 25 },
		{ 357, 345, 25 },
		{ 357, 346, 25 },
		{ 357, 347, 25 },
		{ 357, 348, 25 },
		{ 357, 349, 25 },
		{ 357, 350, 25 },
		{ 357, 351, 25 },
		{ 357, 352, 25 },
		{ 357, 353, 25 },
		{ 357, 354, 25 },
		{ 357, 355, 25 },
		{ 357, 356, 25 },
		{ 358, 308, 25 },
		{ 358, 309, 25 },
		{ 358, 310, 25 },
		{ 358, 311, 25 },
		{ 358, 312, 25 },
		{ 358, 313, 25 },
		{ 358, 314, 25 },
		{ 358, 315, 25 },
		{ 358, 316, 25 },
		{ 358, 317, 25 },
		{ 358, 318, 25 },
		{ 358, 319, 25 },
		{ 358, 320, 25 },
		{ 358, 321, 25 },
		{ 358, 322, 25 },
		{ 358, 323, 25 },
		{ 358, 324, 25 },
		{ 358, 325, 25 },
		{ 358, 326, 25 },
		{ 358, 327, 25 },
		{ 358, 328, 25 },
		{ 358, 329, 25 },
		{ 358, 330, 25 },
		{ 358, 331, 25 },
		{ 358, 332, 25 },
		{ 358, 333, 25 },
		{ 358, 334, 25 },
		{ 358, 335, 25 },
		{ 358, 336, 25 },
		{ 358, 337, 25 },
		{ 358, 338, 25 },
		{ 358, 339, 25 },
		{ 358, 340, 25 },
		{ 358, 341, 25 },
		{ 358, 342, 25 },
		{ 358, 343, 25 },
		{ 358, 344, 25 },
		{ 358, 345, 25 },
		{ 358, 346, 25 },
		{ 358, 347, 25 },
		{ 358, 348, 25 },
		{ 358, 349, 25 },
		{ 358, 350, 25 },
		{ 358, 351, 25 },
		{ 358, 352, 25 },
		{ 358, 353, 25 },
		{ 358, 354, 25 },
		{ 358, 355, 25 },
		{ 358, 356, 25 },
		{ 358, 357, 25 },
		{ 359, 309, 25 },
		{ 359, 310, 25 },
		{ 359, 311, 25 },
		{ 359, 312, 25 },
		{ 359, 313, 25 },
		{ 359, 314, 25 },
		{ 359, 315, 25 },
		{ 359, 316, 25 },
		{ 359, 317, 25 },
		{ 359, 318, 25 },
		{ 359, 319, 25 },
		{ 359, 320, 25 },
		{ 359, 321, 25 },
		{ 359, 322, 25 },
		{ 359, 323, 25 },
		{ 359, 324, 25 },
		{ 359, 325, 25 },
		{ 359, 326, 25 },
		{ 359, 327, 25 },
		{ 359, 328, 25 },
		{ 359, 329, 25 },
		{ 359, 330, 25 },
		{ 359, 331, 25 },
		{ 359, 332, 25 },
		{ 359, 333, 25 },
		{ 359, 334, 25 },
		{ 359, 335, 25 },
		{ 359, 336, 25 },
		{ 359, 337, 25 },
		{ 359, 338, 25 },
		{ 359, 339, 25 },
		{ 359, 340, 25 },
		{ 359, 341, 25 },
		{ 359, 342, 25 },
		{ 359, 343, 25 },
		{ 359, 344, 25 },
		{ 359, 345, 25 },
		{ 359, 346, 25 },
		{ 359, 347, 25 },
		{ 359, 348, 25 },
		{ 359, 349, 25 },
		{ 359, 350, 25 },
		{ 359, 351, 25 },
		{ 359, 352, 25 },
		{ 359, 353, 25 },
		{ 359, 354, 25 },
		{ 359, 355, 25 },
		{ 359, 356, 25 },
		{ 359, 357, 25 },
		{ 359, 358, 25 },
		{ 360, 309, 25 },
		{ 360, 310, 25 },
		{ 360, 311, 25 },
		{ 360, 312, 25 },
		{ 360, 313, 25 },
		{ 360, 314, 25 },
		{ 360, 315, 25 },
		{ 360, 316, 25 },
		{ 360, 317, 25 },
		{ 360, 318, 25 },
		{ 360, 319, 25 },
		{ 360, 320, 25 },
		{ 360, 321, 25 },
		{ 360, 322, 25 },
		{ 360, 323, 25 },
		{ 360, 324, 25 },
		{ 360, 325, 25 },
		{ 360, 326, 25 },
		{ 360, 327, 25 },
		{ 360, 328, 25 },
		{ 360, 329, 25 },
		{ 360, 330, 25 },
		{ 360, 331, 25 },
		{ 360, 332, 25 },
		{ 360, 333, 25 },
		{ 360, 334, 25 },
		{ 360, 335, 25 },
		{ 360, 336, 25 },
		{ 360, 337, 25 },
		{ 360, 338, 25 },
		{ 360, 339, 25 },
		{ 360, 340, 25 },
		{ 360, 341, 25 },
		{ 360, 342, 25 },
		{ 360, 343, 25 },
		{ 360, 344, 25 },
		{ 360, 345, 25 },
		{ 360, 346, 25 },
		{ 360, 347, 25 },
		{ 360, 348, 25 },
		{ 360, 349, 25 },
		{ 360, 350, 25 },
		{ 360, 351, 25 },
		{ 360, 352, 25 },
		{ 360, 353, 25 },
		{ 360, 354, 25 },
		{ 360, 355, 25 },
		{ 360, 356, 25 },
		{ 360, 357, 25 },
		{ 360, 358, 25 },
		{ 360, 359, 25 },
		{ 361, 310, 25 },
		{ 361, 311, 25 },
		{ 361, 312, 25 },
		{ 361, 313, 25 },
		{ 361, 314, 25 },
		{ 361, 315, 25 },
		{ 361, 316, 25 },
		{ 361, 317, 25 },
		{ 361, 318, 25 },
		{ 361, 319, 25 },
		{ 361, 320, 25 },
		{ 361, 321, 25 },
		{ 361, 322, 25 },
		{ 361, 323, 25 },
		{ 361, 324, 25 },
		{ 361, 325, 25 },
		{ 361, 326, 25 },
		{ 361, 327, 25 },
		{ 361, 328, 25 },
		{ 361, 329, 25 },
		{ 361, 330, 25 },
		{ 361, 331, 25 },
		{ 361, 332, 25 },
		{ 361, 333, 25 },
		{ 361, 334, 25 },
		{ 361, 335, 25 },
		{ 361, 336, 25 },
		{ 361, 337, 25 },
		{ 361, 338, 25 },
		{ 361, 339, 25 },
		{ 361, 340, 25 },
		{ 361, 341, 25 },
		{ 361, 342, 25 },
		{ 361, 343, 25 },
		{ 361, 344, 25 },
		{ 361, 345, 25 },
		{ 361, 346, 25 },
		{ 361, 347, 25 },
		{ 361, 348, 25 },
		{ 361, 349, 25 },
		{ 361, 350, 25 },
		{ 361, 351, 25 },
		{ 361, 352, 25 },
		{ 361, 353, 25 },
		{ 361, 354, 25 },
		{ 361, 355, 25 },
		{ 361, 356, 25 },
		{ 361, 357, 25 },
		{ 361, 358, 25 },
		{ 361, 359, 25 },
		{ 361, 360, 25 },
		{ 362, 310, 25 },
		{ 362, 311, 25 },
		{ 362, 312, 25 },
		{ 362, 313, 25 },
		{ 362, 314, 25 },
		{ 362, 315, 25 },
		{ 362, 316, 25 },
		{ 362, 317, 25 },
		{ 362, 318, 25 },
		{ 362, 319, 25 },
		{ 362, 320, 25 },
		{ 362, 321, 25 },
		{ 362, 322, 25 },
		{ 362, 323, 25 },
		{ 362, 324, 25 },
		{ 362, 325, 25 },
		{ 362, 326, 25 },
		{ 362, 327, 25 },
		{ 362, 328, 25 },
		{ 362, 329, 25 },
		{ 362, 330, 25 },
		{ 362, 331, 25 },
		{ 362, 332, 25 },
		{ 362, 333, 25 },
		{ 362, 334, 25 },
		{ 362, 335, 25 },
		{ 362, 336, 25 },
		{ 362, 337, 25 },
		{ 362, 338, 25 },
		{ 362, 339, 25 },
		{ 362, 340, 25 },
		{ 362, 341, 25 },
		{ 362, 342, 25 },
		{ 362, 343, 25 },
		{ 362, 344, 25 },
		{ 362, 345, 25 },
		{ 362, 346, 25 },
		{ 362, 347, 25 },
		{ 362, 348, 25 },
		{ 362, 349, 25 },
		{ 362, 350, 25 },
		{ 362, 351, 25 },
		{ 362, 352, 25 },
		{ 362, 353, 25 },
		{ 362, 354, 25 },
		{ 362, 355, 25 },
		{ 362, 356, 25 },
		{ 362, 357, 25 },
		{ 362, 358, 25 },
		{ 362, 359, 25 },
		{ 362, 360, 25 },
		{ 362, 361, 25 },
		{ 363, 311, 25 },
		{ 363, 312, 25 },
		{ 363, 313, 25 },
		{ 363, 314, 25 },
		{ 363, 315, 25 },
		{ 363, 316, 25 },
		{ 363, 317, 25 },
		{ 363, 318, 25 },
		{ 363, 319, 25 },
		{ 363, 320, 25 },
		{ 363, 321, 25 },
		{ 363, 322, 25 },
		{ 363, 323, 25 },
		{ 363, 324, 25 },
		{ 363, 325, 25 },
		{ 363, 326, 25 },
		{ 363, 327, 25 },
		{ 363, 328, 25 },
		{ 363, 329, 25 },
		{ 363, 330, 25 },
		{ 363, 331, 25 },
		{ 363, 332, 25 },
		{ 363, 333, 25 },
		{ 363, 334, 25 },
		{ 363, 335, 25 },
		{ 363, 336, 25 },
		{ 363, 337, 25 },
		{ 363, 338, 25 },
		{ 363, 339, 25 },
		{ 363, 340, 25 },
		{ 363, 341, 25 },
		{ 363, 342, 25 },
		{ 363, 343, 25 },
		{ 363, 344, 25 },
		{ 363, 345, 25 },
		{ 363, 346, 25 },
		{ 363, 347, 25 },
		{ 363, 348, 25 },
		{ 363, 349, 25 },
		{ 363, 350, 25 },
		{ 363, 351, 25 },
		{ 363, 352, 25 },
		{ 363, 353, 25 },
		{ 363, 354, 25 },
		{ 363, 355, 25 },
		{ 363, 356, 25 },
		{ 363, 357, 25 },
		{ 363, 358, 25 },
		{ 363, 359, 25 },
		{ 363, 360, 25 },
		{ 363, 361, 25 },
		{ 363, 362, 25 },
		{ 364, 311, 25 },
		{ 364, 312, 25 },
		{ 364, 313, 25 },
		{ 364, 314, 25 },
		{ 364, 315, 25 },
		{ 364, 316, 25 },
		{ 364, 317, 25 },
		{ 364, 318, 25 },
		{ 364, 319, 25 },
		{ 364, 320, 25 },
		{ 364, 321, 25 },
		{ 364, 322, 25 },
		{ 364, 323, 25 },
		{ 364, 324, 25 },
		{ 364, 325, 25 },
		{ 364, 326, 25 },
		{ 364, 327, 25 },
		{ 364, 328, 25 },
		{ 364, 329, 25 },
		{ 364, 330, 25 },
		{ 364, 331, 25 },
		{ 364, 332, 25 },
		{ 364, 333, 25 },
		{ 364, 334, 25 },
		{ 364, 335, 25 },
		{ 364, 336, 25 },
		{ 364, 337, 25 },
		{ 364, 338, 25 },
		{ 364, 339, 25 },
		{ 364, 340, 25 },
		{ 364, 341, 25 },
		{ 364, 342, 25 },
		{ 364, 343, 25 },
		{ 364, 344, 25 },
		{ 364, 345, 25 },
		{ 364, 346, 25 },
		{ 364, 347, 25 },
		{ 364, 348, 25 },
		{ 364, 349, 25 },
		{ 364, 350, 25 },
		{ 364, 351, 25 },
		{ 364, 352, 25 },
		{ 364, 353, 25 },
		{ 364, 354, 25 },
		{ 364, 355, 25 },
		{ 364, 356, 25 },
		{ 364, 357, 25 },
		{ 364, 358, 25 },
		{ 364, 359, 25 },
		{ 364, 360, 25 },
		{ 364, 361, 25 },
		{ 364, 362, 25 },
		{ 364, 363, 25 },
		{ 365, 312, 25 },
		{ 365, 313, 25 },
		{ 365, 314, 25 },
		{ 365, 315, 25 },
		{ 365, 316, 25 },
		{ 365, 317, 25 },
		{ 365, 318, 25 },
		{ 365, 319, 25 },
		{ 365, 320, 25 },
		{ 365, 321, 25 },
		{ 365, 322, 25 },
		{ 365, 323, 25 },
		{ 365, 324, 25 },
		{ 365, 325, 25 },
		{ 365, 326, 25 },
		{ 365, 327, 25 },
		{ 365, 328, 25 },
		{ 365, 329, 25 },
		{ 365, 330, 25 },
		{ 365, 331, 25 },
		{ 365, 332, 25 },
		{ 365, 333, 25 },
		{ 365, 334, 25 },
		{ 365, 335, 25 },
		{ 365, 336, 25 },
		{ 365, 337, 25 },
		{ 365, 338, 25 },
		{ 365, 339, 25 },
		{ 365, 340, 25 },
		{ 365, 341, 25 },
		{ 365, 342, 25 },
		{ 365, 343, 25 },
		{ 365, 344, 25 },
		{ 365, 345, 25 },
		{ 365, 346, 25 },
		{ 365, 347, 25 },
		{ 365, 348, 25 },
		{ 365, 349, 25 },
		{ 365, 350, 25 },
		{ 365, 351, 25 },
		{ 365, 352, 25 },
		{ 365, 353, 25 },
		{ 365, 354, 25 },
		{ 365, 355, 25 },
		{ 365, 356, 25 },
		{ 365, 357, 25 },
		{ 365, 358, 25 },
		{ 365, 359, 25 },
		{ 365, 360, 25 },
		{ 365, 361, 25 },
		{ 365, 362, 25 },
		{ 365, 363, 25 },
		{ 365, 364, 25 },
		{ 366, 312, 25 },
		{ 366, 313, 25 },
		{ 366, 314, 25 },
		{ 366, 315, 25 },
		{ 366, 316, 25 },
		{ 366, 317, 25 },
		{ 366, 318, 25 },
		{ 366, 319, 25 },
		{ 366, 320, 25 },
		{ 366, 321, 25 },
		{ 366, 322, 25 },
		{ 366, 323, 25 },
		{ 366, 324, 25 },
		{ 366, 325, 25 },
		{ 366, 326, 25 },
		{ 366, 327, 25 },
		{ 366, 328, 25 },
		{ 366, 329, 25 },
		{ 366, 330, 25 },
		{ 366, 331, 25 },
		{ 366, 332, 25 },
		{ 366, 333, 25 },
		{ 366, 334, 25 },
		{ 366, 335, 25 },
		{ 366, 336, 25 },
		{ 366, 337, 25 },
		{ 366, 338, 25 },
		{ 366, 339, 25 },
		{ 366, 340, 25 },
		{ 366, 341, 25 },
		{ 366, 342, 25 },
		{ 366, 343, 25 },
		{ 366, 344, 25 },
		{ 366, 345, 25 },
		{ 366, 346, 25 },
		{ 366, 347, 25 },
		{ 366, 348, 25 },
		{ 366, 349, 25 },
		{ 366, 350, 25 },
		{ 366, 351, 25 },
		{ 366, 352, 25 },
		{ 366, 353, 25 },
		{ 366, 354, 25 },
		{ 366, 355, 25 },
		{ 366, 356, 25 },
		{ 366, 357, 25 },
		{ 366, 358, 25 },
		{ 366, 359, 25 },
		{ 366, 360, 25 },
		{ 366, 361, 25 },
		{ 366, 362, 25 },
		{ 366, 363, 25 },
		{ 366, 364, 25 },
		{ 366, 365, 25 },
		{ 367, 313, 25 },
		{ 367, 314, 25 },
		{ 367, 315, 25 },
		{ 367, 316, 25 },
		{ 367, 317, 25 },
		{ 367, 318, 25 },
		{ 367, 319, 25 },
		{ 367, 320, 25 },
		{ 367, 321, 25 },
		{ 367, 322, 25 },
		{ 367, 323, 25 },
		{ 367, 324, 25 },
		{ 367, 325, 25 },
		{ 367, 326, 25 },
		{ 367, 327, 25 },
		{ 367, 328, 25 },
		{ 367, 329, 25 },
		{ 367, 330, 25 },
		{ 367, 331, 25 },
		{ 367, 332, 25 },
		{ 367, 333, 25 },
		{ 367, 334, 25 },
		{ 367, 335, 25 },
		{ 367, 336, 25 },
		{ 367, 337, 25 },
		{ 367, 338, 25 },
		{ 367, 339, 25 },
		{ 367, 340, 25 },
		{ 367, 341, 25 },
		{ 367, 342, 25 },
		{ 367, 343, 25 },
		{ 367, 344, 25 },
		{ 367, 345, 25 },
		{ 367, 346, 25 },
		{ 367, 347, 25 },
		{ 367, 348, 25 },
		{ 367, 349, 25 },
		{ 367, 350, 25 },
		{ 367, 351, 25 },
		{ 367, 352, 25 },
		{ 367, 353, 25 },
		{ 367, 354, 25 },
		{ 367, 355, 25 },
		{ 367, 356, 25 },
		{ 367, 357, 25 },
		{ 367, 358, 25 },
		{ 367, 359, 25 },
		{ 367, 360, 25 },
		{ 367, 361, 25 },
		{ 367, 362, 25 },
		{ 367, 363, 25 },
		{ 367, 364, 25 },
		{ 367, 365, 25 },
		{ 367, 366, 25 },
		{ 368, 313, 25 },
		{ 368, 314, 25 },
		{ 368, 315, 25 },
		{ 368, 316, 25 },
		{ 368, 317, 25 },
		{ 368, 318, 25 },
		{ 368, 319, 25 },
		{ 368, 320, 25 },
		{ 368, 321, 25 },
		{ 368, 322, 25 },
		{ 368, 323, 25 },
		{ 368, 324, 25 },
		{ 368, 325, 25 },
		{ 368, 326, 25 },
		{ 368, 327, 25 },
		{ 368, 328, 25 },
		{ 368, 329, 25 },
		{ 368, 330, 25 },
		{ 368, 331, 25 },
		{ 368, 332, 25 },
		{ 368, 333, 25 },
		{ 368, 334, 25 },
		{ 368, 335, 25 },
		{ 368, 336, 25 },
		{ 368, 337, 25 },
		{ 368, 338, 25 },
		{ 368, 339, 25 },
		{ 368, 340, 25 },
		{ 368, 341, 25 },
		{ 368, 342, 25 },
		{ 368, 343, 25 },
		{ 368, 344, 25 },
		{ 368, 345, 25 },
		{ 368, 346, 25 },
		{ 368, 347, 25 },
		{ 368, 348, 25 },
		{ 368, 349, 25 },
		{ 368, 350, 25 },
		{ 368, 351, 25 },
		{ 368, 352, 25 },
		{ 368, 353, 25 },
		{ 368, 354, 25 },
		{ 368, 355, 25 },
		{ 368, 356, 25 },
		{ 368, 357, 25 },
		{ 368, 358, 25 },
		{ 368, 359, 25 },
		{ 368, 360, 25 },
		{ 368, 361, 25 },
		{ 368, 362, 25 },
		{ 368, 363, 25 },
		{ 368, 364, 25 },
		{ 368, 365, 25 },
		{ 368, 366, 25 },
		{ 368, 367, 25 },
		{ 369, 314, 25 },
		{ 369, 315, 25 },
		{ 369, 316, 25 },
		{ 369, 317, 25 },
		{ 369, 318, 25 },
		{ 369, 319, 25 },
		{ 369, 320, 25 },
		{ 369, 321, 25 },
		{ 369, 322, 25 },
		{ 369, 323, 25 },
		{ 369, 324, 25 },
		{ 369, 325, 25 },
		{ 369, 326, 25 },
		{ 369, 327, 25 },
		{ 369, 328, 25 },
		{ 369, 329, 25 },
		{ 369, 330, 25 },
		{ 369, 331, 25 },
		{ 369, 332, 25 },
		{ 369, 333, 25 },
		{ 369, 334, 25 },
		{ 369, 335, 25 },
		{ 369, 336, 25 },
		{ 369, 337, 25 },
		{ 369, 338, 25 },
		{ 369, 339, 25 },
		{ 369, 340, 25 },
		{ 369, 341, 25 },
		{ 369, 342, 25 },
		{ 369, 343, 25 },
		{ 369, 344, 25 },
		{ 369, 345, 25 },
		{ 369, 346, 25 },
		{ 369, 347, 25 },
		{ 369, 348, 25 },
		{ 369, 349, 25 },
		{ 369, 350, 25 },
		{ 369, 351, 25 },
		{ 369, 352, 25 },
		{ 369, 353, 25 },
		{ 369, 354, 25 },
		{ 369, 355, 25 },
		{ 369, 356, 25 },
		{ 369, 357, 25 },
		{ 369, 358, 25 },
		{ 369, 359, 25 },
		{ 369, 360, 25 },
		{ 369, 361, 25 },
		{ 369, 362, 25 },
		{ 369, 363, 25 },
		{ 369, 364, 25 },
		{ 369, 365, 25 },
		{ 369, 366, 25 },
		{ 369, 367, 25 },
		{ 369, 368, 25 },
		{ 370, 314, 25 },
		{ 370, 315, 25 },
		{ 370, 316, 25 },
		{ 370, 317, 25 },
		{ 370, 318, 25 },
		{ 370, 319, 25 },
		{ 370, 320, 25 },
		{ 370, 321, 25 },
		{ 370, 322, 25 },
		{ 370, 323, 25 },
		{ 370, 324, 25 },
		{ 370, 325, 25 },
		{ 370, 326, 25 },
		{ 370, 327, 25 },
		{ 370, 328, 25 },
		{ 370, 329, 25 },
		{ 370, 330, 25 },
		{ 370, 331, 25 },
		{ 370, 332, 25 },
		{ 370, 333, 25 },
		{ 370, 334, 25 },
		{ 370, 335, 25 },
		{ 370, 336, 25 },
		{ 370, 337, 25 },
		{ 370, 338, 25 },
		{ 370, 339, 25 },
		{ 370, 340, 25 },
		{ 370, 341, 25 },
		{ 370, 342, 25 },
		{ 370, 343, 25 },
		{ 370, 344, 25 },
		{ 370, 345, 25 },
		{ 370, 346, 25 },
		{ 370, 347, 25 },
		{ 370, 348, 25 },
		{ 370, 349, 25 },
		{ 370, 350, 25 },
		{ 370, 351, 25 },
		{ 370, 352, 25 },
		{ 370, 353, 25 },
		{ 370, 354, 25 },
		{ 370, 355, 25 },
		{ 370, 356, 25 },
		{ 370, 357, 25 },
		{ 370, 358, 25 },
		{ 370, 359, 25 },
		{ 370, 360, 25 },
		{ 370, 361, 25 },
		{ 370, 362, 25 },
		{ 370, 363, 25 },
		{ 370, 364, 25 },
		{ 370, 365, 25 },
		{ 370, 366, 25 },
		{ 370, 367, 25 },
		{ 370, 368, 25 },
		{ 370, 369, 25 },
		{ 371, 315, 25 },
		{ 371, 316, 25 },
		{ 371, 317, 25 },
		{ 371, 318, 25 },
		{ 371, 319, 25 },
		{ 371, 320, 25 },
		{ 371, 321, 25 },
		{ 371, 322, 25 },
		{ 371, 323, 25 },
		{ 371, 324, 25 },
		{ 371, 325, 25 },
		{ 371, 326, 25 },
		{ 371, 327, 25 },
		{ 371, 328, 25 },
		{ 371, 329, 25 },
		{ 371, 330, 25 },
		{ 371, 331, 25 },
		{ 371, 332, 25 },
		{ 371, 333, 25 },
		{ 371, 334, 25 },
		{ 371, 335, 25 },
		{ 371, 336, 25 },
		{ 371, 337, 25 },
		{ 371, 338, 25 },
		{ 371, 339, 25 },
		{ 371, 340, 25 },
		{ 371, 341, 25 },
		{ 371, 342, 25 },
		{ 371, 343, 25 },
		{ 371, 344, 25 },
		{ 371, 345, 25 },
		{ 371, 346, 25 },
		{ 371, 347, 25 },
		{ 371, 348, 25 },
		{ 371, 349, 25 },
		{ 371, 350, 25 },
		{ 371, 351, 25 },
		{ 371, 352, 25 },
		{ 371, 353, 25 },
		{ 371, 354, 25 },
		{ 371, 355, 25 },
		{ 371, 356, 25 },
		{ 371, 357, 25 },
		{ 371, 358, 25 },
		{ 371, 359, 25 },
		{ 371, 360, 25 },
		{ 371, 361, 25 },
		{ 371, 362, 25 },
		{ 371, 363, 25 },
		{ 371, 364, 25 },
		{ 371, 365, 25 },
		{ 371, 366, 25 },
		{ 371, 367, 25 },
		{ 371, 368, 25 },
		{ 371, 369, 25 },
		{ 371, 370, 25 },
		{ 372, 315, 25 },
		{ 372, 316, 25 },
		{ 372, 317, 25 },
		{ 372, 318, 25 },
		{ 372, 319, 25 },
		{ 372, 320, 25 },
		{ 372, 321, 25 },
		{ 372, 322, 25 },
		{ 372, 323, 25 },
		{ 372, 324, 25 },
		{ 372, 325, 25 },
		{ 372, 326, 25 },
		{ 372, 327, 25 },
		{ 372, 328, 25 },
		{ 372, 329, 25 },
		{ 372, 330, 25 },
		{ 372, 331, 25 },
		{ 372, 332, 25 },
		{ 372, 333, 25 },
		{ 372, 334, 25 },
		{ 372, 335, 25 },
		{ 372, 336, 25 },
		{ 372, 337, 25 },
		{ 372, 338, 25 },
		{ 372, 339, 25 },
		{ 372, 340, 25 },
		{ 372, 341, 25 },
		{ 372, 342, 25 },
		{ 372, 343, 25 },
		{ 372, 344, 25 },
		{ 372, 345, 25 },
		{ 372, 346, 25 },
		{ 372, 347, 25 },
		{ 372, 348, 25 },
		{ 372, 349, 25 },
		{ 372, 350, 25 },
		{ 372, 351, 25 },
		{ 372, 352, 25 },
		{ 372, 353, 25 },
		{ 372, 354, 25 },
		{ 372, 355, 25 },
		{ 372, 356, 25 },
		{ 372, 357, 25 },
		{ 372, 358, 25 },
		{ 372, 359, 25 },
		{ 372, 360, 25 },
		{ 372, 361, 25 },
		{ 372, 362, 25 },
		{ 372, 363, 25 },
		{ 372, 364, 25 },
		{ 372, 365, 25 },
		{ 372, 366, 25 },
		{ 372, 367, 25 },
		{ 372, 368, 25 },
		{ 372, 369, 25 },
		{ 372, 370, 25 },
		{ 372, 371, 25 },
		{ 373, 316, 25 },
		{ 373, 317, 25 },
		{ 373, 318, 25 },
		{ 373, 319, 25 },
		{ 373, 320, 25 },
		{ 373, 321, 25 },
		{ 373, 322, 25 },
		{ 373, 323, 25 },
		{ 373, 324, 25 },
		{ 373, 325, 25 },
		{ 373, 326, 25 },
		{ 373, 327, 25 },
		{ 373, 328, 25 },
		{ 373, 329, 25 },
		{ 373, 330, 25 },
		{ 373, 331, 25 },
		{ 373, 332, 25 },
		{ 373, 333, 25 },
		{ 373, 334, 25 },
		{ 373, 335, 25 },
		{ 373, 336, 25 },
		{ 373, 337, 25 },
		{ 373, 338, 25 },
		{ 373, 339, 25 },
		{ 373, 340, 25 },
		{ 373, 341, 25 },
		{ 373, 342, 25 },
		{ 373, 343, 25 },
		{ 373, 344, 25 },
		{ 373, 345, 25 },
		{ 373, 346, 25 },
		{ 373, 347, 25 },
		{ 373, 348, 25 },
		{ 373, 349, 25 },
		{ 373, 350, 25 },
		{ 373, 351, 25 },
		{ 373, 352, 25 },
		{ 373, 353, 25 },
		{ 373, 354, 25 },
		{ 373, 355, 25 },
		{ 373, 356, 25 },
		{ 373, 357, 25 },
		{ 373, 358, 25 },
		{ 373, 359, 25 },
		{ 373, 360, 25 },
		{ 373, 361, 25 },
		{ 373, 362, 25 },
		{ 373, 363, 25 },
		{ 373, 364, 25 },
		{ 373, 365, 25 },
		{ 373, 366, 25 },
		{ 373, 367, 25 },
		{ 373, 368, 25 },
		{ 373, 369, 25 },
		{ 373, 370, 25 },
		{ 373, 371, 25 },
		{ 373, 372, 25 },
		{ 374, 316, 25 },
		{ 374, 317, 25 },
		{ 374, 318, 25 },
		{ 374, 319, 25 },
		{ 374, 320, 25 },
		{ 374, 321, 25 },
		{ 374, 322, 25 },
		{ 374, 323, 25 },
		{ 374, 324, 25 },
		{ 374, 325, 25 },
		{ 374, 326, 25 },
		{ 374, 327, 25 },
		{ 374, 328, 25 },
		{ 374, 329, 25 },
		{ 374, 330, 25 },
		{ 374, 331, 25 },
		{ 374, 332, 25 },
		{ 374, 333, 25 },
		{ 374, 334, 25 },
		{ 374, 335, 25 },
		{ 374, 336, 25 },
		{ 374, 337, 25 },
		{ 374, 338, 25 },
		{ 374, 339, 25 },
		{ 374, 340, 25 },
		{ 374, 341, 25 },
		{ 374, 342, 25 },
		{ 374, 343, 25 },
		{ 374, 344, 25 },
		{ 374, 345, 25 },
		{ 374, 346, 25 },
		{ 374, 347, 25 },
		{ 374, 348, 25 },
		{ 374, 349, 25 },
		{ 374, 350, 25 },
		{ 374, 351, 25 },
		{ 374, 352, 25 },
		{ 374, 353, 25 },
		{ 374, 354, 25 },
		{ 374, 355, 25 },
		{ 374, 356, 25 },
		{ 374, 357, 25 },
		{ 374, 358, 25 },
		{ 374, 359, 25 },
		{ 374, 360, 25 },
		{ 374, 361, 25 },
		{ 374, 362, 25 },
		{ 374, 363, 25 },
		{ 374, 364, 25 },
		{ 374, 365, 25 },
		{ 374, 366, 25 },
		{ 374, 367, 25 },
		{ 374, 368, 25 },
		{ 374, 369, 25 },
		{ 374, 370, 25 },
		{ 374, 371, 25 },
		{ 374, 372, 25 },
		{ 374, 373, 25 },
		{ 375, 317, 25 },
		{ 375, 318, 25 },
		{ 375, 319, 25 },
		{ 375, 320, 25 },
		{ 375, 321, 25 },
		{ 375, 322, 25 },
		{ 375, 323, 25 },
		{ 375, 324, 25 },
		{ 375, 325, 25 },
		{ 375, 326, 25 },
		{ 375, 327, 25 },
		{ 375, 328, 25 },
		{ 375, 329, 25 },
		{ 375, 330, 25 },
		{ 375, 331, 25 },
		{ 375, 332, 25 },
		{ 375, 333, 25 },
		{ 375, 334, 25 },
		{ 375, 335, 25 },
		{ 375, 336, 25 },
		{ 375, 337, 25 },
		{ 375, 338, 25 },
		{ 375, 339, 25 },
		{ 375, 340, 25 },
		{ 375, 341, 25 },
		{ 375, 342, 25 },
		{ 375, 343, 25 },
		{ 375, 344, 25 },
		{ 375, 345, 25 },
		{ 375, 346, 25 },
		{ 375, 347, 25 },
		{ 375, 348, 25 },
		{ 375, 349, 25 },
		{ 375, 350, 25 },
		{ 375, 351, 25 },
		{ 375, 352, 25 },
		{ 375, 353, 25 },
		{ 375, 354, 25 },
		{ 375, 355, 25 },
		{ 375, 356, 25 },
		{ 375, 357, 25 },
		{ 375, 358, 25 },
		{ 375, 359, 25 },
		{ 375, 360, 25 },
		{ 375, 361, 25 },
		{ 375, 362, 25 },
		{ 375, 363, 25 },
		{ 375, 364, 25 },
		{ 375, 365, 25 },
		{ 375, 366, 25 },
		{ 375, 367, 25 },
		{ 375, 368, 25 },
		{ 375, 369, 25 },
		{ 375, 370, 25 },
		{ 375, 371, 25 },
		{ 375, 372, 25 },
		{ 375, 373, 25 },
		{ 375, 374, 25 },
		{ 376, 317, 25 },
		{ 376, 318, 25 },
		{ 376, 319, 25 },
		{ 376, 320, 25 },
		{ 376, 321, 25 },
		{ 376, 322, 25 },
		{ 376, 323, 25 },
		{ 376, 324, 25 },
		{ 376, 325, 25 },
		{ 376, 326, 25 },
		{ 376, 327, 25 },
		{ 376, 328, 25 },
		{ 376, 329, 25 },
		{ 376, 330, 25 },
		{ 376, 331, 25 },
		{ 376, 332, 25 },
		{ 376, 333, 25 },
		{ 376, 334, 25 },
		{ 376, 335, 25 },
		{ 376, 336, 25 },
		{ 376, 337, 25 },
		{ 376, 338, 25 },
		{ 376, 339, 25 },
		{ 376, 340, 25 },
		{ 376, 341, 25 },
		{ 376, 342, 25 },
		{ 376, 343, 25 },
		{ 376, 344, 25 },
		{ 376, 345, 25 },
		{ 376, 346, 25 },
		{ 376, 347, 25 },
		{ 376, 348, 25 },
		{ 376, 349, 25 },
		{ 376, 350, 25 },
		{ 376, 351, 25 },
		{ 376, 352, 25 },
		{ 376, 353, 25 },
		{ 376, 354, 25 },
		{ 376, 355, 25 },
		{ 376, 356, 25 },
		{ 376, 357, 25 },
		{ 376, 358, 25 },
		{ 376, 359, 25 },
		{ 376, 360, 25 },
		{ 376, 361, 25 },
		{ 376, 362, 25 },
		{ 376, 363, 25 },
		{ 376, 364, 25 },
		{ 376, 365, 25 },
		{ 376, 366, 25 },
		{ 376, 367, 25 },
		{ 376, 368, 25 },
		{ 376, 369, 25 },
		{ 376, 370, 25 },
		{ 376, 371, 25 },
		{ 376, 372, 25 },
		{ 376, 373, 25 },
		{ 376, 374, 25 },
		{ 376, 375, 25 },
		{ 377, 318, 25 },
		{ 377, 319, 25 },
		{ 377, 320, 25 },
		{ 377, 321, 25 },
		{ 377, 322, 25 },
		{ 377, 323, 25 },
		{ 377, 324, 25 },
		{ 377, 325, 25 },
		{ 377, 326, 25 },
		{ 377, 327, 25 },
		{ 377, 328, 25 },
		{ 377, 329, 25 },
		{ 377, 330, 25 },
		{ 377, 331, 25 },
		{ 377, 332, 25 },
		{ 377, 333, 25 },
		{ 377, 334, 25 },
		{ 377, 335, 25 },
		{ 377, 336, 25 },
		{ 377, 337, 25 },
		{ 377, 338, 25 },
		{ 377, 339, 25 },
		{ 377, 340, 25 },
		{ 377, 341, 25 },
		{ 377, 342, 25 },
		{ 377, 343, 25 },
		{ 377, 344, 25 },
		{ 377, 345, 25 },
		{ 377, 346, 25 },
		{ 377, 347, 25 },
		{ 377, 348, 25 },
		{ 377, 349, 25 },
		{ 377, 350, 25 },
		{ 377, 351, 25 },
		{ 377, 352, 25 },
		{ 377, 353, 25 },
		{ 377, 354, 25 },
		{ 377, 355, 25 },
		{ 377, 356, 25 },
		{ 377, 357, 25 },
		{ 377, 358, 25 },
		{ 377, 359, 25 },
		{ 377, 360, 25 },
		{ 377, 361, 25 },
		{ 377, 362, 25 },
		{ 377, 363, 25 },
		{ 377, 364, 25 },
		{ 377, 365, 25 },
		{ 377, 366, 25 },
		{ 377, 367, 25 },
		{ 377, 368, 25 },
		{ 377, 369, 25 },
		{ 377, 370, 25 },
		{ 377, 371, 25 },
		{ 377, 372, 25 },
		{ 377, 373, 25 },
		{ 377, 374, 25 },
		{ 377, 375, 25 },
		{ 377, 376, 25 },
		{ 378, 319, 25 },
		{ 378, 320, 25 },
		{ 378, 321, 25 },
		{ 378, 322, 25 },
		{ 378, 323, 25 },
		{ 378, 324, 25 },
		{ 378, 325, 25 },
		{ 378, 326, 25 },
		{ 378, 327, 25 },
		{ 378, 328, 25 },
		{ 378, 329, 25 },
		{ 378, 330, 25 },
		{ 378, 331, 25 },
		{ 378, 332, 25 },
		{ 378, 333, 25 },
		{ 378, 334, 25 },
		{ 378, 335, 25 },
		{ 378, 336, 25 },
		{ 378, 337, 25 },
		{ 378, 338, 25 },
		{ 378, 339, 25 },
		{ 378, 340, 25 },
		{ 378, 341, 25 },
		{ 378, 342, 25 },
		{ 378, 343, 25 },
		{ 378, 344, 25 },
		{ 378, 345, 25 },
		{ 378, 346, 25 },
		{ 378, 347, 25 },
		{ 378, 348, 25 },
		{ 378, 349, 25 },
		{ 378, 350, 25 },
		{ 378, 351, 25 },
		{ 378, 352, 25 },
		{ 378, 353, 25 },
		{ 378, 354, 25 },
		{ 378, 355, 25 },
		{ 378, 356, 25 },
		{ 378, 357, 25 },
		{ 378, 358, 25 },
		{ 378, 359, 25 },
		{ 378, 360, 25 },
		{ 378, 361, 25 },
		{ 378, 362, 25 },
		{ 378, 363, 25 },
		{ 378, 364, 25 },
		{ 378, 365, 25 },
		{ 378, 366, 25 },
		{ 378, 367, 25 },
		{ 378, 368, 25 },
		{ 378, 369, 25 },
		{ 378, 370, 25 },
		{ 378, 371, 25 },
		{ 378, 372, 25 },
		{ 378, 373, 25 },
		{ 378, 374, 25 },
		{ 378, 375, 25 },
		{ 378, 376, 25 },
		{ 378, 377, 25 },
		{ 379, 319, 25 },
		{ 379, 320, 25 },
		{ 379, 321, 25 },
		{ 379, 322, 25 },
		{ 379, 323, 25 },
		{ 379, 324, 25 },
		{ 379, 325, 25 },
		{ 379, 326, 25 },
		{ 379, 327, 25 },
		{ 379, 328, 25 },
		{ 379, 329, 25 },
		{ 379, 330, 25 },
		{ 379, 331, 25 },
		{ 379, 332, 25 },
		{ 379, 333, 25 },
		{ 379, 334, 25 },
		{ 379, 335, 25 },
		{ 379, 336, 25 },
		{ 379, 337, 25 },
		{ 379, 338, 25 },
		{ 379, 339, 25 },
		{ 379, 340, 25 },
		{ 379, 341, 25 },
		{ 379, 342, 25 },
		{ 379, 343, 25 },
		{ 379, 344, 25 },
		{ 379, 345, 25 },
		{ 379, 346, 25 },
		{ 379, 347, 25 },
		{ 379, 348, 25 },
		{ 379, 349, 25 },
		{ 379, 350, 25 },
		{ 379, 351, 25 },
		{ 379, 352, 25 },
		{ 379, 353, 25 },
		{ 379, 354, 25 },
		{ 379, 355, 25 },
		{ 379, 356, 25 },
		{ 379, 357, 25 },
		{ 379, 358, 25 },
		{ 379, 359, 25 },
		{ 379, 360, 25 },
		{ 379, 361, 25 },
		{ 379, 362, 25 },
		{ 379, 363, 25 },
		{ 379, 364, 25 },
		{ 379, 365, 25 },
		{ 379, 366, 25 },
		{ 379, 367, 25 },
		{ 379, 368, 25 },
		{ 379, 369, 25 },
		{ 379, 370, 25 },
		{ 379, 371, 25 },
		{ 379, 372, 25 },
		{ 379, 373, 25 },
		{ 379, 374, 25 },
		{ 379, 375, 25 },
		{ 379, 376, 25 },
		{ 379, 377, 25 },
		{ 379, 378, 25 },
		{ 380, 320, 25 },
		{ 380, 321, 25 },
		{ 380, 322, 25 },
		{ 380, 323, 25 },
		{ 380, 324, 25 },
		{ 380, 325, 25 },
		{ 380, 326, 25 },
		{ 380, 327, 25 },
		{ 380, 328, 25 },
		{ 380, 329, 25 },
		{ 380, 330, 25 },
		{ 380, 331, 25 },
		{ 380, 332, 25 },
		{ 380, 333, 25 },
		{ 380, 334, 25 },
		{ 380, 335, 25 },
		{ 380, 336, 25 },
		{ 380, 337, 25 },
		{ 380, 338, 25 },
		{ 380, 339, 25 },
		{ 380, 340, 25 },
		{ 380, 341, 25 },
		{ 380, 342, 25 },
		{ 380, 343, 25 },
		{ 380, 344, 25 },
		{ 380, 345, 25 },
		{ 380, 346, 25 },
		{ 380, 347, 25 },
		{ 380, 348, 25 },
		{ 380, 349, 25 },
		{ 380, 350, 25 },
		{ 380, 351, 25 },
		{ 380, 352, 25 },
		{ 380, 353, 25 },
		{ 380, 354, 25 },
		{ 380, 355, 25 },
		{ 380, 356, 25 },
		{ 380, 357, 25 },
		{ 380, 358, 25 },
		{ 380, 359, 25 },
		{ 380, 360, 25 },
		{ 380, 361, 25 },
		{ 380, 362, 25 },
		{ 380, 363, 25 },
		{ 380, 364, 25 },
		{ 380, 365, 25 },
		{ 380, 366, 25 },
		{ 380, 367, 25 },
		{ 380, 368, 25 },
		{ 380, 369, 25 },
		{ 380, 370, 25 },
		{ 380, 371, 25 },
		{ 380, 372, 25 },
		{ 380, 373, 25 },
		{ 380, 374, 25 },
		{ 380, 375, 25 },
		{ 380, 376, 25 },
		{ 380, 377, 25 },
		{ 380, 378, 25 },
		{ 380, 379, 25 },
		{ 381, 321, 25 },
		{ 381, 322, 25 },
		{ 381, 323, 25 },
		{ 381, 324, 25 },
		{ 381, 325, 25 },
		{ 381, 326, 25 },
		{ 381, 327, 25 },
		{ 381, 328, 25 },
		{ 381, 329, 25 },
		{ 381, 330, 25 },
		{ 381, 331, 25 },
		{ 381, 332, 25 },
		{ 381, 333, 25 },
		{ 381, 334, 25 },
		{ 381, 335, 25 },
		{ 381, 336, 25 },
		{ 381, 337, 25 },
		{ 381, 338, 25 },
		{ 381, 339, 25 },
		{ 381, 340, 25 },
		{ 381, 341, 25 },
		{ 381, 342, 25 },
		{ 381, 343, 25 },
		{ 381, 344, 25 },
		{ 381, 345, 25 },
		{ 381, 346, 25 },
		{ 381, 347, 25 },
		{ 381, 348, 25 },
		{ 381, 349, 25 },
		{ 381, 350, 25 },
		{ 381, 351, 25 },
		{ 381, 352, 25 },
		{ 381, 353, 25 },
		{ 381, 354, 25 },
		{ 381, 355, 25 },
		{ 381, 356, 25 },
		{ 381, 357, 25 },
		{ 381, 358, 25 },
		{ 381, 359, 25 },
		{ 381, 360, 25 },
		{ 381, 361, 25 },
		{ 381, 362, 25 },
		{ 381, 363, 25 },
		{ 381, 364, 25 },
		{ 381, 365, 25 },
		{ 381, 366, 25 },
		{ 381, 367, 25 },
		{ 381, 368, 25 },
		{ 381, 369, 25 },
		{ 381, 370, 25 },
		{ 381, 371, 25 },
		{ 381, 372, 25 },
		{ 381, 373, 25 },
		{ 381, 374, 25 },
		{ 381, 375, 25 },
		{ 381, 376, 25 },
		{ 381, 377, 25 },
		{ 381, 378, 25 },
		{ 381, 379, 25 },
		{ 382, 323, 25 },
		{ 382, 324, 25 },
		{ 382, 325, 25 },
		{ 382, 326, 25 },
		{ 382, 327, 25 },
		{ 382, 328, 25 },
		{ 382, 329, 25 },
		{ 382, 330, 25 },
		{ 382, 331, 25 },
		{ 382, 332, 25 },
		{ 382, 333, 25 },
		{ 382, 334, 25 },
		{ 382, 335, 25 },
		{ 382, 336, 25 },
		{ 382, 337, 25 },
		{ 382, 338, 25 },
		{ 382, 339, 25 },
		{ 382, 340, 25 },
		{ 382, 341, 25 },
		{ 382, 342, 25 },
		{ 382, 343, 25 },
		{ 382, 344, 25 },
		{ 382, 345, 25 },
		{ 382, 346, 25 },
		{ 382, 347, 25 },
		{ 382, 348, 25 },
		{ 382, 349, 25 },
		{ 382, 350, 25 },
		{ 382, 351, 25 },
		{ 382, 352, 25 },
		{ 382, 353, 25 },
		{ 382, 354, 25 },
		{ 382, 355, 25 },
		{ 382, 356, 25 },
		{ 382, 357, 25 },
		{ 382, 358, 25 },
		{ 382, 359, 25 },
		{ 382, 360, 25 },
		{ 382, 361, 25 },
		{ 382, 362, 25 },
		{ 382, 363, 25 },
		{ 382, 364, 25 },
		{ 382, 365, 25 },
		{ 382, 366, 25 },
		{ 382, 367, 25 },
		{ 382, 368, 25 },
		{ 382, 369, 25 },
		{ 382, 370, 25 },
		{ 382, 371, 25 },
		{ 382, 372, 25 },
		{ 382, 373, 25 },
		{ 382, 374, 25 },
		{ 382, 375, 25 },
		{ 382, 376, 25 },
		{ 382, 377, 25 },
		{ 382, 378, 25 },
		{ 382, 379, 25 },
		{ 383, 327, 25 },
		{ 383, 328, 25 },
		{ 383, 329, 25 },
		{ 383, 330, 25 },
		{ 383, 331, 25 },
		{ 383, 332, 25 },
		{ 383, 333, 25 },
		{ 383, 334, 25 },
		{ 383, 335, 25 },
		{ 383, 336, 25 },
		{ 383, 337, 25 },
		{ 383, 338, 25 },
		{ 383, 339, 25 },
		{ 383, 340, 25 },
		{ 383, 341, 25 },
		{ 383, 342, 25 },
		{ 383, 343, 25 },
		{ 383, 344, 25 },
		{ 383, 345, 25 },
		{ 383, 346, 25 },
		{ 383, 347, 25 },
		{ 383, 348, 25 },
		{ 383, 349, 25 },
		{ 383, 350, 25 },
		{ 383, 351, 25 },
		{ 383, 352, 25 },
		{ 383, 353, 25 },
		{ 383, 354, 25 },
		{ 383, 355, 25 },
		{ 383, 356, 25 },
		{ 383, 357, 25 },
		{ 383, 358, 25 },
		{ 383, 359, 25 },
		{ 383, 360, 25 },
		{ 383, 361, 25 },
		{ 383, 362, 25 },
		{ 383, 363, 25 },
		{ 383, 364, 25 },
		{ 383, 365, 25 },
		{ 383, 366, 25 },
		{ 383, 367, 25 },
		{ 383, 368, 25 },
		{ 383, 369, 25 },
		{ 383, 370, 25 },
		{ 383, 371, 25 },
		{ 383, 372, 25 },
		{ 383, 373, 25 },
		{ 383, 374, 25 },
		{ 383, 375, 25 },
		{ 383, 376, 25 },
		{ 131, 131, 26 },
		{ 131, 132, 26 },
		{ 131, 133, 26 },
		{ 131, 134, 26 },
		{ 131, 135, 26 },
		{ 131, 136, 26 },
		{ 131, 137, 26 },
		{ 131, 138, 26 },
		{ 131, 139, 26 },
		{ 131, 140, 26 },
		{ 131, 141, 26 },
		{ 131, 142, 26 },
		{ 131, 143, 26 },
		{ 131, 144, 26 },
		{ 131, 145, 26 },
		{ 131, 146, 26 },
		{ 131, 147, 26 },
		{ 131, 148, 26 },
		{ 131, 149, 26 },
		{ 131, 150, 26 },
		{ 131, 151, 26 },
		{ 131, 152, 26 },
		{ 131, 153, 26 },
		{ 131, 154, 26 },
		{ 131, 155, 26 },
		{ 132, 131, 26 },
		{ 132, 132, 26 },
		{ 132, 133, 26 },
		{ 132, 134, 26 },
		{ 132, 135, 26 },
		{ 132, 136, 26 },
		{ 132, 137, 26 },
		{ 132, 138, 26 },
		{ 132, 139, 26 },
		{ 132, 140, 26 },
		{ 132, 141, 26 },
		{ 132, 142, 26 },
		{ 132, 143, 26 },
		{ 132, 144, 26 },
		{ 132, 145, 26 },
		{ 132, 146, 26 },
		{ 132, 147, 26 },
		{ 132, 148, 26 },
		{ 132, 149, 26 },
		{ 132, 150, 26 },
		{ 132, 151, 26 },
		{ 132, 152, 26 },
		{ 132, 153, 26 },
		{ 132, 154, 26 },
		{ 132, 155, 26 },
		{ 132, 156, 26 },
		{ 132, 157, 26 },
		{ 132, 158, 26 },
		{ 132, 159, 26 },
		{ 132, 160, 26 },
		{ 133, 131, 26 },
		{ 133, 132, 26 },
		{ 133, 133, 26 },
		{ 133, 134, 26 },
		{ 133, 135, 26 },
		{ 133, 136, 26 },
		{ 133, 137, 26 },
		{ 133, 138, 26 },
		{ 133, 139, 26 },
		{ 133, 140, 26 },
		{ 133, 141, 26 },
		{ 133, 142, 26 },
		{ 133, 143, 26 },
		{ 133, 144, 26 },
		{ 133, 145, 26 },
		{ 133, 146, 26 },
		{ 133, 147, 26 },
		{ 133, 148, 26 },
		{ 133, 149, 26 },
		{ 133, 150, 26 },
		{ 133, 151, 26 },
		{ 133, 152, 26 },
		{ 133, 153, 26 },
		{ 133, 154, 26 },
		{ 133, 155, 26 },
		{ 133, 156, 26 },
		{ 133, 157, 26 },
		{ 133, 158, 26 },
		{ 133, 159, 26 },
		{ 133, 160, 26 },
		{ 133, 161, 26 },
		{ 133, 162, 26 },
		{ 134, 131, 26 },
		{ 134, 132, 26 },
		{ 134, 133, 26 },
		{ 134, 134, 26 },
		{ 134, 135, 26 },
		{ 134, 136, 26 },
		{ 134, 137, 26 },
		{ 134, 138, 26 },
		{ 134, 139, 26 },
		{ 134, 140, 26 },
		{ 134, 141, 26 },
		{ 134, 142, 26 },
		{ 134, 143, 26 },
		{ 134, 144, 26 },
		{ 134, 145, 26 },
		{ 134, 146, 26 },
		{ 134, 147, 26 },
		{ 134, 148, 26 },
		{ 134, 149, 26 },
		{ 134, 150, 26 },
		{ 134, 151, 26 },
		{ 134, 152, 26 },
		{ 134, 153, 26 },
		{ 134, 154, 26 },
		{ 134, 155, 26 },
		{ 134, 156, 26 },
		{ 134, 157, 26 },
		{ 134, 158, 26 },
		{ 134, 159, 26 },
		{ 134, 160, 26 },
		{ 134, 161, 26 },
		{ 134, 162, 26 },
		{ 134, 163, 26 },
		{ 135, 131, 26 },
		{ 135, 132, 26 },
		{ 135, 133, 26 },
		{ 135, 134, 26 },
		{ 135, 135, 26 },
		{ 135, 136, 26 },
		{ 135, 137, 26 },
		{ 135, 138, 26 },
		{ 135, 139, 26 },
		{ 135, 140, 26 },
		{ 135, 141, 26 },
		{ 135, 142, 26 },
		{ 135, 143, 26 },
		{ 135, 144, 26 },
		{ 135, 145, 26 },
		{ 135, 146, 26 },
		{ 135, 147, 26 },
		{ 135, 148, 26 },
		{ 135, 149, 26 },
		{ 135, 150, 26 },
		{ 135, 151, 26 },
		{ 135, 152, 26 },
		{ 135, 153, 26 },
		{ 135, 154, 26 },
		{ 135, 155, 26 },
		{ 135, 156, 26 },
		{ 135, 157, 26 },
		{ 135, 158, 26 },
		{ 135, 159, 26 },
		{ 135, 160, 26 },
		{ 135, 161, 26 },
		{ 135, 162, 26 },
		{ 135, 163, 26 },
		{ 135, 164, 26 },
		{ 136, 131, 26 },
		{ 136, 132, 26 },
		{ 136, 133, 26 },
		{ 136, 134, 26 },
		{ 136, 135, 26 },
		{ 136, 136, 26 },
		{ 136, 137, 26 },
		{ 136, 138, 26 },
		{ 136, 139, 26 },
		{ 136, 140, 26 },
		{ 136, 141, 26 },
		{ 136, 142, 26 },
		{ 136, 143, 26 },
		{ 136, 144, 26 },
		{ 136, 145, 26 },
		{ 136, 146, 26 },
		{ 136, 147, 26 },
		{ 136, 148, 26 },
		{ 136, 149, 26 },
		{ 136, 150, 26 },
		{ 136, 151, 26 },
		{ 136, 152, 26 },
		{ 136, 153, 26 },
		{ 136, 154, 26 },
		{ 136, 155, 26 },
		{ 136, 156, 26 },
		{ 136, 157, 26 },
		{ 136, 158, 26 },
		{ 136, 159, 26 },
		{ 136, 160, 26 },
		{ 136, 161, 26 },
		{ 136, 162, 26 },
		{ 136, 163, 26 },
		{ 136, 164, 26 },
		{ 136, 165, 26 },
		{ 137, 131, 26 },
		{ 137, 132, 26 },
		{ 137, 133, 26 },
		{ 137, 134, 26 },
		{ 137, 135, 26 },
		{ 137, 136, 26 },
		{ 137, 137, 26 },
		{ 137, 138, 26 },
		{ 137, 139, 26 },
		{ 137, 140, 26 },
		{ 137, 141, 26 },
		{ 137, 142, 26 },
		{ 137, 143, 26 },
		{ 137, 144, 26 },
		{ 137, 145, 26 },
		{ 137, 146, 26 },
		{ 137, 147, 26 },
		{ 137, 148, 26 },
		{ 137, 149, 26 },
		{ 137, 150, 26 },
		{ 137, 151, 26 },
		{ 137, 152, 26 },
		{ 137, 153, 26 },
		{ 137, 154, 26 },
		{ 137, 155, 26 },
		{ 137, 156, 26 },
		{ 137, 157, 26 },
		{ 137, 158, 26 },
		{ 137, 159, 26 },
		{ 137, 160, 26 },
		{ 137, 161, 26 },
		{ 137, 162, 26 },
		{ 137, 163, 26 },
		{ 137, 164, 26 },
		{ 137, 165, 26 },
		{ 137, 166, 26 },
		{ 138, 131, 26 },
		{ 138, 132, 26 },
		{ 138, 133, 26 },
		{ 138, 134, 26 },
		{ 138, 135, 26 },
		{ 138, 136, 26 },
		{ 138, 137, 26 },
		{ 138, 138, 26 },
		{ 138, 139, 26 },
		{ 138, 140, 26 },
		{ 138, 141, 26 },
		{ 138, 142, 26 },
		{ 138, 143, 26 },
		{ 138, 144, 26 },
		{ 138, 145, 26 },
		{ 138, 146, 26 },
		{ 138, 147, 26 },
		{ 138, 148, 26 },
		{ 138, 149, 26 },
		{ 138, 150, 26 },
		{ 138, 151, 26 },
		{ 138, 152, 26 },
		{ 138, 153, 26 },
		{ 138, 154, 26 },
		{ 138, 155, 26 },
		{ 138, 156, 26 },
		{ 138, 157, 26 },
		{ 138, 158, 26 },
		{ 138, 159, 26 },
		{ 138, 160, 26 },
		{ 138, 161, 26 },
		{ 138, 162, 26 },
		{ 138, 163, 26 },
		{ 138, 164, 26 },
		{ 138, 165, 26 },
		{ 138, 166, 26 },
		{ 138, 167, 26 },
		{ 139, 131, 26 },
		{ 139, 132, 26 },
		{ 139, 133, 26 },
		{ 139, 134, 26 },
		{ 139, 135, 26 },
		{ 139, 136, 26 },
		{ 139, 137, 26 },
		{ 139, 138, 26 },
		{ 139, 139, 26 },
		{ 139, 140, 26 },
		{ 139, 141, 26 },
		{ 139, 142, 26 },
		{ 139, 143, 26 },
		{ 139, 144, 26 },
		{ 139, 145, 26 },
		{ 139, 146, 26 },
		{ 139, 147, 26 },
		{ 139, 148, 26 },
		{ 139, 149, 26 },
		{ 139, 150, 26 },
		{ 139, 151, 26 },
		{ 139, 152, 26 },
		{ 139, 153, 26 },
		{ 139, 154, 26 },
		{ 139, 155, 26 },
		{ 139, 156, 26 },
		{ 139, 157, 26 },
		{ 139, 158, 26 },
		{ 139, 159, 26 },
		{ 139, 160, 26 },
		{ 139, 161, 26 },
		{ 139, 162, 26 },
		{ 139, 163, 26 },
		{ 139, 164, 26 },
		{ 139, 165, 26 },
		{ 139, 166, 26 },
		{ 139, 167, 26 },
		{ 139, 168, 26 },
		{ 140, 131, 26 },
		{ 140, 132, 26 },
		{ 140, 133, 26 },
		{ 140, 134, 26 },
		{ 140, 135, 26 },
		{ 140, 136, 26 },
		{ 140, 137, 26 },
		{ 140, 138, 26 },
		{ 140, 139, 26 },
		{ 140, 140, 26 },
		{ 140, 141, 26 },
		{ 140, 142, 26 },
		{ 140, 143, 26 },
		{ 140, 144, 26 },
		{ 140, 145, 26 },
		{ 140, 146, 26 },
		{ 140, 147, 26 },
		{ 140, 148, 26 },
		{ 140, 149, 26 },
		{ 140, 150, 26 },
		{ 140, 151, 26 },
		{ 140, 152, 26 },
		{ 140, 153, 26 },
		{ 140, 154, 26 },
		{ 140, 155, 26 },
		{ 140, 156, 26 },
		{ 140, 157, 26 },
		{ 140, 158, 26 },
		{ 140, 159, 26 },
		{ 140, 160, 26 },
		{ 140, 161, 26 },
		{ 140, 162, 26 },
		{ 140, 163, 26 },
		{ 140, 164, 26 },
		{ 140, 165, 26 },
		{ 140, 166, 26 },
		{ 140, 167, 26 },
		{ 140, 168, 26 },
		{ 140, 169, 26 },
		{ 141, 131, 26 },
		{ 141, 132, 26 },
		{ 141, 133, 26 },
		{ 141, 134, 26 },
		{ 141, 135, 26 },
		{ 141, 136, 26 },
		{ 141, 137, 26 },
		{ 141, 138, 26 },
		{ 141, 139, 26 },
		{ 141, 140, 26 },
		{ 141, 141, 26 },
		{ 141, 142, 26 },
		{ 141, 143, 26 },
		{ 141, 144, 26 },
		{ 141, 145, 26 },
		{ 141, 146, 26 },
		{ 141, 147, 26 },
		{ 141, 148, 26 },
		{ 141, 149, 26 },
		{ 141, 150, 26 },
		{ 141, 151, 26 },
		{ 141, 152, 26 },
		{ 141, 153, 26 },
		{ 141, 154, 26 },
		{ 141, 155, 26 },
		{ 141, 156, 26 },
		{ 141, 157, 26 },
		{ 141, 158, 26 },
		{ 141, 159, 26 },
		{ 141, 160, 26 },
		{ 141, 161, 26 },
		{ 141, 162, 26 },
		{ 141, 163, 26 },
		{ 141, 164, 26 },
		{ 141, 165, 26 },
		{ 141, 166, 26 },
		{ 141, 167, 26 },
		{ 141, 168, 26 },
		{ 141, 169, 26 },
		{ 142, 131, 26 },
		{ 142, 132, 26 },
		{ 142, 133, 26 },
		{ 142, 134, 26 },
		{ 142, 135, 26 },
		{ 142, 136, 26 },
		{ 142, 137, 26 },
		{ 142, 138, 26 },
		{ 142, 139, 26 },
		{ 142, 140, 26 },
		{ 142, 141, 26 },
		{ 142, 142, 26 },
		{ 142, 143, 26 },
		{ 142, 144, 26 },
		{ 142, 145, 26 },
		{ 142, 146, 26 },
		{ 142, 147, 26 },
		{ 142, 148, 26 },
		{ 142, 149, 26 },
		{ 142, 150, 26 },
		{ 142, 151, 26 },
		{ 142, 152, 26 },
		{ 142, 153, 26 },
		{ 142, 154, 26 },
		{ 142, 155, 26 },
		{ 142, 156, 26 },
		{ 142, 157, 26 },
		{ 142, 158, 26 },
		{ 142, 159, 26 },
		{ 142, 160, 26 },
		{ 142, 161, 26 },
		{ 142, 162, 26 },
		{ 142, 163, 26 },
		{ 142, 164, 26 },
		{ 142, 165, 26 },
		{ 142, 166, 26 },
		{ 142, 167, 26 },
		{ 142, 168, 26 },
		{ 142, 169, 26 },
		{ 142, 170, 26 },
		{ 143, 131, 26 },
		{ 143, 132, 26 },
		{ 143, 133, 26 },
		{ 143, 134, 26 },
		{ 143, 135, 26 },
		{ 143, 136, 26 },
		{ 143, 137, 26 },
		{ 143, 138, 26 },
		{ 143, 139, 26 },
		{ 143, 140, 26 },
		{ 143, 141, 26 },
		{ 143, 142, 26 },
		{ 143, 143, 26 },
		{ 143, 144, 26 },
		{ 143, 145, 26 },
		{ 143, 146, 26 },
		{ 143, 147, 26 },
		{ 143, 148, 26 },
		{ 143, 149, 26 },
		{ 143, 150, 26 },
		{ 143, 151, 26 },
		{ 143, 152, 26 },
		{ 143, 153, 26 },
		{ 143, 154, 26 },
		{ 143, 155, 26 },
		{ 143, 156, 26 },
		{ 143, 157, 26 },
		{ 143, 158, 26 },
		{ 143, 159, 26 },
		{ 143, 160, 26 },
		{ 143, 161, 26 },
		{ 143, 162, 26 },
		{ 143, 163, 26 },
		{ 143, 164, 26 },
		{ 143, 165, 26 },
		{ 143, 166, 26 },
		{ 143, 167, 26 },
		{ 143, 168, 26 },
		{ 143, 169, 26 },
		{ 143, 170, 26 },
		{ 143, 171, 26 },
		{ 144, 131, 26 },
		{ 144, 132, 26 },
		{ 144, 133, 26 },
		{ 144, 134, 26 },
		{ 144, 135, 26 },
		{ 144, 136, 26 },
		{ 144, 137, 26 },
		{ 144, 138, 26 },
		{ 144, 139, 26 },
		{ 144, 140, 26 },
		{ 144, 141, 26 },
		{ 144, 142, 26 },
		{ 144, 143, 26 },
		{ 144, 144, 26 },
		{ 144, 145, 26 },
		{ 144, 146, 26 },
		{ 144, 147, 26 },
		{ 144, 148, 26 },
		{ 144, 149, 26 },
		{ 144, 150, 26 },
		{ 144, 151, 26 },
		{ 144, 152, 26 },
		{ 144, 153, 26 },
		{ 144, 154, 26 },
		{ 144, 155, 26 },
		{ 144, 156, 26 },
		{ 144, 157, 26 },
		{ 144, 158, 26 },
		{ 144, 159, 26 },
		{ 144, 160, 26 },
		{ 144, 161, 26 },
		{ 144, 162, 26 },
		{ 144, 163, 26 },
		{ 144, 164, 26 },
		{ 144, 165, 26 },
		{ 144, 166, 26 },
		{ 144, 167, 26 },
		{ 144, 168, 26 },
		{ 144, 169, 26 },
		{ 144, 170, 26 },
		{ 144, 171, 26 },
		{ 144, 172, 26 },
		{ 145, 131, 26 },
		{ 145, 132, 26 },
		{ 145, 133, 26 },
		{ 145, 134, 26 },
		{ 145, 135, 26 },
		{ 145, 136, 26 },
		{ 145, 137, 26 },
		{ 145, 138, 26 },
		{ 145, 139, 26 },
		{ 145, 140, 26 },
		{ 145, 141, 26 },
		{ 145, 142, 26 },
		{ 145, 143, 26 },
		{ 145, 144, 26 },
		{ 145, 145, 26 },
		{ 145, 146, 26 },
		{ 145, 147, 26 },
		{ 145, 148, 26 },
		{ 145, 149, 26 },
		{ 145, 150, 26 },
		{ 145, 151, 26 },
		{ 145, 152, 26 },
		{ 145, 153, 26 },
		{ 145, 154, 26 },
		{ 145, 155, 26 },
		{ 145, 156, 26 },
		{ 145, 157, 26 },
		{ 145, 158, 26 },
		{ 145, 159, 26 },
		{ 145, 160, 26 },
		{ 145, 161, 26 },
		{ 145, 162, 26 },
		{ 145, 163, 26 },
		{ 145, 164, 26 },
		{ 145, 165, 26 },
		{ 145, 166, 26 },
		{ 145, 167, 26 },
		{ 145, 168, 26 },
		{ 145, 169, 26 },
		{ 145, 170, 26 },
		{ 145, 171, 26 },
		{ 145, 172, 26 },
		{ 146, 131, 26 },
		{ 146, 132, 26 },
		{ 146, 133, 26 },
		{ 146, 134, 26 },
		{ 146, 135, 26 },
		{ 146, 136, 26 },
		{ 146, 137, 26 },
		{ 146, 138, 26 },
		{ 146, 139, 26 },
		{ 146, 140, 26 },
		{ 146, 141, 26 },
		{ 146, 142, 26 },
		{ 146, 143, 26 },
		{ 146, 144, 26 },
		{ 146, 145, 26 },
		{ 146, 146, 26 },
		{ 146, 147, 26 },
		{ 146, 148, 26 },
		{ 146, 149, 26 },
		{ 146, 150, 26 },
		{ 146, 151, 26 },
		{ 146, 152, 26 },
		{ 146, 153, 26 },
		{ 146, 154, 26 },
		{ 146, 155, 26 },
		{ 146, 156, 26 },
		{ 146, 157, 26 },
		{ 146, 158, 26 },
		{ 146, 159, 26 },
		{ 146, 160, 26 },
		{ 146, 161, 26 },
		{ 146, 162, 26 },
		{ 146, 163, 26 },
		{ 146, 164, 26 },
		{ 146, 165, 26 },
		{ 146, 166, 26 },
		{ 146, 167, 26 },
		{ 146, 168, 26 },
		{ 146, 169, 26 },
		{ 146, 170, 26 },
		{ 146, 171, 26 },
		{ 146, 172, 26 },
		{ 146, 173, 26 },
		{ 147, 131, 26 },
		{ 147, 132, 26 },
		{ 147, 133, 26 },
		{ 147, 134, 26 },
		{ 147, 135, 26 },
		{ 147, 136, 26 },
		{ 147, 137, 26 },
		{ 147, 138, 26 },
		{ 147, 139, 26 },
		{ 147, 140, 26 },
		{ 147, 141, 26 },
		{ 147, 142, 26 },
		{ 147, 143, 26 },
		{ 147, 144, 26 },
		{ 147, 145, 26 },
		{ 147, 146, 26 },
		{ 147, 147, 26 },
		{ 147, 148, 26 },
		{ 147, 149, 26 },
		{ 147, 150, 26 },
		{ 147, 151, 26 },
		{ 147, 152, 26 },
		{ 147, 153, 26 },
		{ 147, 154, 26 },
		{ 147, 155, 26 },
		{ 147, 156, 26 },
		{ 147, 157, 26 },
		{ 147, 158, 26 },
		{ 147, 159, 26 },
		{ 147, 160, 26 },
		{ 147, 161, 26 },
		{ 147, 162, 26 },
		{ 147, 163, 26 },
		{ 147, 164, 26 },
		{ 147, 165, 26 },
		{ 147, 166, 26 },
		{ 147, 167, 26 },
		{ 147, 168, 26 },
		{ 147, 169, 26 },
		{ 147, 170, 26 },
		{ 147, 171, 26 },
		{ 147, 172, 26 },
		{ 147, 173, 26 },
		{ 147, 174, 26 },
		{ 148, 131, 26 },
		{ 148, 132, 26 },
		{ 148, 133, 26 },
		{ 148, 134, 26 },
		{ 148, 135, 26 },
		{ 148, 136, 26 },
		{ 148, 137, 26 },
		{ 148, 138, 26 },
		{ 148, 139, 26 },
		{ 148, 140, 26 },
		{ 148, 141, 26 },
		{ 148, 142, 26 },
		{ 148, 143, 26 },
		{ 148, 144, 26 },
		{ 148, 145, 26 },
		{ 148, 146, 26 },
		{ 148, 147, 26 },
		{ 148, 148, 26 },
		{ 148, 149, 26 },
		{ 148, 150, 26 },
		{ 148, 151, 26 },
		{ 148, 152, 26 },
		{ 148, 153, 26 },
		{ 148, 154, 26 },
		{ 148, 155, 26 },
		{ 148, 156, 26 },
		{ 148, 157, 26 },
		{ 148, 158, 26 },
		{ 148, 159, 26 },
		{ 148, 160, 26 },
		{ 148, 161, 26 },
		{ 148, 162, 26 },
		{ 148, 163, 26 },
		{ 148, 164, 26 },
		{ 148, 165, 26 },
		{ 148, 166, 26 },
		{ 148, 167, 26 },
		{ 148, 168, 26 },
		{ 148, 169, 26 },
		{ 148, 170, 26 },
		{ 148, 171, 26 },
		{ 148, 172, 26 },
		{ 148, 173, 26 },
		{ 148, 174, 26 },
		{ 148, 175, 26 },
		{ 149, 131, 26 },
		{ 149, 132, 26 },
		{ 149, 133, 26 },
		{ 149, 134, 26 },
		{ 149, 135, 26 },
		{ 149, 136, 26 },
		{ 149, 137, 26 },
		{ 149, 138, 26 },
		{ 149, 139, 26 },
		{ 149, 140, 26 },
		{ 149, 141, 26 },
		{ 149, 142, 26 },
		{ 149, 143, 26 },
		{ 149, 144, 26 },
		{ 149, 145, 26 },
		{ 149, 146, 26 },
		{ 149, 147, 26 },
		{ 149, 148, 26 },
		{ 149, 149, 26 },
		{ 149, 150, 26 },
		{ 149, 151, 26 },
		{ 149, 152, 26 },
		{ 149, 153, 26 },
		{ 149, 154, 26 },
		{ 149, 155, 26 },
		{ 149, 156, 26 },
		{ 149, 157, 26 },
		{ 149, 158, 26 },
		{ 149, 159, 26 },
		{ 149, 160, 26 },
		{ 149, 161, 26 },
		{ 149, 162, 26 },
		{ 149, 163, 26 },
		{ 149, 164, 26 },
		{ 149, 165, 26 },
		{ 149, 166, 26 },
		{ 149, 167, 26 },
		{ 149, 168, 26 },
		{ 149, 169, 26 },
		{ 149, 170, 26 },
		{ 149, 171, 26 },
		{ 149, 172, 26 },
		{ 149, 173, 26 },
		{ 149, 174, 26 },
		{ 149, 175, 26 },
		{ 150, 131, 26 },
		{ 150, 132, 26 },
		{ 150, 133, 26 },
		{ 150, 134, 26 },
		{ 150, 135, 26 },
		{ 150, 136, 26 },
		{ 150, 137, 26 },
		{ 150, 138, 26 },
		{ 150, 139, 26 },
		{ 150, 140, 26 },
		{ 150, 141, 26 },
		{ 150, 142, 26 },
		{ 150, 143, 26 },
		{ 150, 144, 26 },
		{ 150, 145, 26 },
		{ 150, 146, 26 },
		{ 150, 147, 26 },
		{ 150, 148, 26 },
		{ 150, 149, 26 },
		{ 150, 150, 26 },
		{ 150, 151, 26 },
		{ 150, 152, 26 },
		{ 150, 153, 26 },
		{ 150, 154, 26 },
		{ 150, 155, 26 },
		{ 150, 156, 26 },
		{ 150, 157, 26 },
		{ 150, 158, 26 },
		{ 150, 159, 26 },
		{ 150, 160, 26 },
		{ 150, 161, 26 },
		{ 150, 162, 26 },
		{ 150, 163, 26 },
		{ 150, 164, 26 },
		{ 150, 165, 26 },
		{ 150, 166, 26 },
		{ 150, 167, 26 },
		{ 150, 168, 26 },
		{ 150, 169, 26 },
		{ 150, 170, 26 },
		{ 150, 171, 26 },
		{ 150, 172, 26 },
		{ 150, 173, 26 },
		{ 150, 174, 26 },
		{ 150, 175, 26 },
		{ 150, 176, 26 },
		{ 151, 131, 26 },
		{ 151, 132, 26 },
		{ 151, 133, 26 },
		{ 151, 134, 26 },
		{ 151, 135, 26 },
		{ 151, 136, 26 },
		{ 151, 137, 26 },
		{ 151, 138, 26 },
		{ 151, 139, 26 },
		{ 151, 140, 26 },
		{ 151, 141, 26 },
		{ 151, 142, 26 },
		{ 151, 143, 26 },
		{ 151, 144, 26 },
		{ 151, 145, 26 },
		{ 151, 146, 26 },
		{ 151, 147, 26 },
		{ 151, 148, 26 },
		{ 151, 149, 26 },
		{ 151, 150, 26 },
		{ 151, 151, 26 },
		{ 151, 152, 26 },
		{ 151, 153, 26 },
		{ 151, 154, 26 },
		{ 151, 155, 26 },
		{ 151, 156, 26 },
		{ 151, 157, 26 },
		{ 151, 158, 26 },
		{ 151, 159, 26 },
		{ 151, 160, 26 },
		{ 151, 161, 26 },
		{ 151, 162, 26 },
		{ 151, 163, 26 },
		{ 151, 164, 26 },
		{ 151, 165, 26 },
		{ 151, 166, 26 },
		{ 151, 167, 26 },
		{ 151, 168, 26 },
		{ 151, 169, 26 },
		{ 151, 170, 26 },
		{ 151, 171, 26 },
		{ 151, 172, 26 },
		{ 151, 173, 26 },
		{ 151, 174, 26 },
		{ 151, 175, 26 },
		{ 151, 176, 26 },
		{ 151, 177, 26 },
		{ 152, 131, 26 },
		{ 152, 132, 26 },
		{ 152, 133, 26 },
		{ 152, 134, 26 },
		{ 152, 135, 26 },
		{ 152, 136, 26 },
		{ 152, 137, 26 },
		{ 152, 138, 26 },
		{ 152, 139, 26 },
		{ 152, 140, 26 },
		{ 152, 141, 26 },
		{ 152, 142, 26 },
		{ 152, 143, 26 },
		{ 152, 144, 26 },
		{ 152, 145, 26 },
		{ 152, 146, 26 },
		{ 152, 147, 26 },
		{ 152, 148, 26 },
		{ 152, 149, 26 },
		{ 152, 150, 26 },
		{ 152, 151, 26 },
		{ 152, 152, 26 },
		{ 152, 153, 26 },
		{ 152, 154, 26 },
		{ 152, 155, 26 },
		{ 152, 156, 26 },
		{ 152, 157, 26 },
		{ 152, 158, 26 },
		{ 152, 159, 26 },
		{ 152, 160, 26 },
		{ 152, 161, 26 },
		{ 152, 162, 26 },
		{ 152, 163, 26 },
		{ 152, 164, 26 },
		{ 152, 165, 26 },
		{ 152, 166, 26 },
		{ 152, 167, 26 },
		{ 152, 168, 26 },
		{ 152, 169, 26 },
		{ 152, 170, 26 },
		{ 152, 171, 26 },
		{ 152, 172, 26 },
		{ 152, 173, 26 },
		{ 152, 174, 26 },
		{ 152, 175, 26 },
		{ 152, 176, 26 },
		{ 152, 177, 26 },
		{ 152, 178, 26 },
		{ 153, 131, 26 },
		{ 153, 132, 26 },
		{ 153, 133, 26 },
		{ 153, 134, 26 },
		{ 153, 135, 26 },
		{ 153, 136, 26 },
		{ 153, 137, 26 },
		{ 153, 138, 26 },
		{ 153, 139, 26 },
		{ 153, 140, 26 },
		{ 153, 141, 26 },
		{ 153, 142, 26 },
		{ 153, 143, 26 },
		{ 153, 144, 26 },
		{ 153, 145, 26 },
		{ 153, 146, 26 },
		{ 153, 147, 26 },
		{ 153, 148, 26 },
		{ 153, 149, 26 },
		{ 153, 150, 26 },
		{ 153, 151, 26 },
		{ 153, 152, 26 },
		{ 153, 153, 26 },
		{ 153, 154, 26 },
		{ 153, 155, 26 },
		{ 153, 156, 26 },
		{ 153, 157, 26 },
		{ 153, 158, 26 },
		{ 153, 159, 26 },
		{ 153, 160, 26 },
		{ 153, 161, 26 },
		{ 153, 162, 26 },
		{ 153, 163, 26 },
		{ 153, 164, 26 },
		{ 153, 165, 26 },
		{ 153, 166, 26 },
		{ 153, 167, 26 },
		{ 153, 168, 26 },
		{ 153, 169, 26 },
		{ 153, 170, 26 },
		{ 153, 171, 26 },
		{ 153, 172, 26 },
		{ 153, 173, 26 },
		{ 153, 174, 26 },
		{ 153, 175, 26 },
		{ 153, 176, 26 },
		{ 153, 177, 26 },
		{ 153, 178, 26 },
		{ 154, 131, 26 },
		{ 154, 132, 26 },
		{ 154, 133, 26 },
		{ 154, 134, 26 },
		{ 154, 135, 26 },
		{ 154, 136, 26 },
		{ 154, 137, 26 },
		{ 154, 138, 26 },
		{ 154, 139, 26 },
		{ 154, 140, 26 },
		{ 154, 141, 26 },
		{ 154, 142, 26 },
		{ 154, 143, 26 },
		{ 154, 144, 26 },
		{ 154, 145, 26 },
		{ 154, 146, 26 },
		{ 154, 147, 26 },
		{ 154, 148, 26 },
		{ 154, 149, 26 },
		{ 154, 150, 26 },
		{ 154, 151, 26 },
		{ 154, 152, 26 },
		{ 154, 153, 26 },
		{ 154, 154, 26 },
		{ 154, 155, 26 },
		{ 154, 156, 26 },
		{ 154, 157, 26 },
		{ 154, 158, 26 },
		{ 154, 159, 26 },
		{ 154, 160, 26 },
		{ 154, 161, 26 },
		{ 154, 162, 26 },
		{ 154, 163, 26 },
		{ 154, 164, 26 },
		{ 154, 165, 26 },
		{ 154, 166, 26 },
		{ 154, 167, 26 },
		{ 154, 168, 26 },
		{ 154, 169, 26 },
		{ 154, 170, 26 },
		{ 154, 171, 26 },
		{ 154, 172, 26 },
		{ 154, 173, 26 },
		{ 154, 174, 26 },
		{ 154, 175, 26 },
		{ 154, 176, 26 },
		{ 154, 177, 26 },
		{ 154, 178, 26 },
		{ 154, 179, 26 },
		{ 155, 131, 26 },
		{ 155, 132, 26 },
		{ 155, 133, 26 },
		{ 155, 134, 26 },
		{ 155, 135, 26 },
		{ 155, 136, 26 },
		{ 155, 137, 26 },
		{ 155, 138, 26 },
		{ 155, 139, 26 },
		{ 155, 140, 26 },
		{ 155, 141, 26 },
		{ 155, 142, 26 },
		{ 155, 143, 26 },
		{ 155, 144, 26 },
		{ 155, 145, 26 },
		{ 155, 146, 26 },
		{ 155, 147, 26 },
		{ 155, 148, 26 },
		{ 155, 149, 26 },
		{ 155, 150, 26 },
		{ 155, 151, 26 },
		{ 155, 152, 26 },
		{ 155, 153, 26 },
		{ 155, 154, 26 },
		{ 155, 155, 26 },
		{ 155, 156, 26 },
		{ 155, 157, 26 },
		{ 155, 158, 26 },
		{ 155, 159, 26 },
		{ 155, 160, 26 },
		{ 155, 161, 26 },
		{ 155, 162, 26 },
		{ 155, 163, 26 },
		{ 155, 164, 26 },
		{ 155, 165, 26 },
		{ 155, 166, 26 },
		{ 155, 167, 26 },
		{ 155, 168, 26 },
		{ 155, 169, 26 },
		{ 155, 170, 26 },
		{ 155, 171, 26 },
		{ 155, 172, 26 },
		{ 155, 173, 26 },
		{ 155, 174, 26 },
		{ 155, 175, 26 },
		{ 155, 176, 26 },
		{ 155, 177, 26 },
		{ 155, 178, 26 },
		{ 155, 179, 26 },
		{ 155, 180, 26 },
		{ 156, 132, 26 },
		{ 156, 133, 26 },
		{ 156, 134, 26 },
		{ 156, 135, 26 },
		{ 156, 136, 26 },
		{ 156, 137, 26 },
		{ 156, 138, 26 },
		{ 156, 139, 26 },
		{ 156, 140, 26 },
		{ 156, 141, 26 },
		{ 156, 142, 26 },
		{ 156, 143, 26 },
		{ 156, 144, 26 },
		{ 156, 145, 26 },
		{ 156, 146, 26 },
		{ 156, 147, 26 },
		{ 156, 148, 26 },
		{ 156, 149, 26 },
		{ 156, 150, 26 },
		{ 156, 151, 26 },
		{ 156, 152, 26 },
		{ 156, 153, 26 },
		{ 156, 154, 26 },
		{ 156, 155, 26 },
		{ 156, 156, 26 },
		{ 156, 157, 26 },
		{ 156, 158, 26 },
		{ 156, 159, 26 },
		{ 156, 160, 26 },
		{ 156, 161, 26 },
		{ 156, 162, 26 },
		{ 156, 163, 26 },
		{ 156, 164, 26 },
		{ 156, 165, 26 },
		{ 156, 166, 26 },
		{ 156, 167, 26 },
		{ 156, 168, 26 },
		{ 156, 169, 26 },
		{ 156, 170, 26 },
		{ 156, 171, 26 },
		{ 156, 172, 26 },
		{ 156, 173, 26 },
		{ 156, 174, 26 },
		{ 156, 175, 26 },
		{ 156, 176, 26 },
		{ 156, 177, 26 },
		{ 156, 178, 26 },
		{ 156, 179, 26 },
		{ 156, 180, 26 },
		{ 156, 181, 26 },
		{ 157, 132, 26 },
		{ 157, 133, 26 },
		{ 157, 134, 26 },
		{ 157, 135, 26 },
		{ 157, 136, 26 },
		{ 157, 137, 26 },
		{ 157, 138, 26 },
		{ 157, 139, 26 },
		{ 157, 140, 26 },
		{ 157, 141, 26 },
		{ 157, 142, 26 },
		{ 157, 143, 26 },
		{ 157, 144, 26 },
		{ 157, 145, 26 },
		{ 157, 146, 26 },
		{ 157, 147, 26 },
		{ 157, 148, 26 },
		{ 157, 149, 26 },
		{ 157, 150, 26 },
		{ 157, 151, 26 },
		{ 157, 152, 26 },
		{ 157, 153, 26 },
		{ 157, 154, 26 },
		{ 157, 155, 26 },
		{ 157, 156, 26 },
		{ 157, 157, 26 },
		{ 157, 158, 26 },
		{ 157, 159, 26 },
		{ 157, 160, 26 },
		{ 157, 161, 26 },
		{ 157, 162, 26 },
		{ 157, 163, 26 },
		{ 157, 164, 26 },
		{ 157, 165, 26 },
		{ 157, 166, 26 },
		{ 157, 167, 26 },
		{ 157, 168, 26 },
		{ 157, 169, 26 },
		{ 157, 170, 26 },
		{ 157, 171, 26 },
		{ 157, 172, 26 },
		{ 157, 173, 26 },
		{ 157, 174, 26 },
		{ 157, 175, 26 },
		{ 157, 176, 26 },
		{ 157, 177, 26 },
		{ 157, 178, 26 },
		{ 157, 179, 26 },
		{ 157, 180, 26 },
		{ 157, 181, 26 },
		{ 158, 132, 26 },
		{ 158, 133, 26 },
		{ 158, 134, 26 },
		{ 158, 135, 26 },
		{ 158, 136, 26 },
		{ 158, 137, 26 },
		{ 158, 138, 26 },
		{ 158, 139, 26 },
		{ 158, 140, 26 },
		{ 158, 141, 26 },
		{ 158, 142, 26 },
		{ 158, 143, 26 },
		{ 158, 144, 26 },
		{ 158, 145, 26 },
		{ 158, 146, 26 },
		{ 158, 147, 26 },
		{ 158, 148, 26 },
		{ 158, 149, 26 },
		{ 158, 150, 26 },
		{ 158, 151, 26 },
		{ 158, 152, 26 },
		{ 158, 153, 26 },
		{ 158, 154, 26 },
		{ 158, 155, 26 },
		{ 158, 156, 26 },
		{ 158, 157, 26 },
		{ 158, 158, 26 },
		{ 158, 159, 26 },
		{ 158, 160, 26 },
		{ 158, 161, 26 },
		{ 158, 162, 26 },
		{ 158, 163, 26 },
		{ 158, 164, 26 },
		{ 158, 165, 26 },
		{ 158, 166, 26 },
		{ 158, 167, 26 },
		{ 158, 168, 26 },
		{ 158, 169, 26 },
		{ 158, 170, 26 },
		{ 158, 171, 26 },
		{ 158, 172, 26 },
		{ 158, 173, 26 },
		{ 158, 174, 26 },
		{ 158, 175, 26 },
		{ 158, 176, 26 },
		{ 158, 177, 26 },
		{ 158, 178, 26 },
		{ 158, 179, 26 },
		{ 158, 180, 26 },
		{ 158, 181, 26 },
		{ 158, 182, 26 },
		{ 159, 132, 26 },
		{ 159, 133, 26 },
		{ 159, 134, 26 },
		{ 159, 135, 26 },
		{ 159, 136, 26 },
		{ 159, 137, 26 },
		{ 159, 138, 26 },
		{ 159, 139, 26 },
		{ 159, 140, 26 },
		{ 159, 141, 26 },
		{ 159, 142, 26 },
		{ 159, 143, 26 },
		{ 159, 144, 26 },
		{ 159, 145, 26 },
		{ 159, 146, 26 },
		{ 159, 147, 26 },
		{ 159, 148, 26 },
		{ 159, 149, 26 },
		{ 159, 150, 26 },
		{ 159, 151, 26 },
		{ 159, 152, 26 },
		{ 159, 153, 26 },
		{ 159, 154, 26 },
		{ 159, 155, 26 },
		{ 159, 156, 26 },
		{ 159, 157, 26 },
		{ 159, 158, 26 },
		{ 159, 159, 26 },
		{ 159, 160, 26 },
		{ 159, 161, 26 },
		{ 159, 162, 26 },
		{ 159, 163, 26 },
		{ 159, 164, 26 },
		{ 159, 165, 26 },
		{ 159, 166, 26 },
		{ 159, 167, 26 },
		{ 159, 168, 26 },
		{ 159, 169, 26 },
		{ 159, 170, 26 },
		{ 159, 171, 26 },
		{ 159, 172, 26 },
		{ 159, 173, 26 },
		{ 159, 174, 26 },
		{ 159, 175, 26 },
		{ 159, 176, 26 },
		{ 159, 177, 26 },
		{ 159, 178, 26 },
		{ 159, 179, 26 },
		{ 159, 180, 26 },
		{ 159, 181, 26 },
		{ 159, 182, 26 },
		{ 159, 183, 26 },
		{ 160, 132, 26 },
		{ 160, 133, 26 },
		{ 160, 134, 26 },
		{ 160, 135, 26 },
		{ 160, 136, 26 },
		{ 160, 137, 26 },
		{ 160, 138, 26 },
		{ 160, 139, 26 },
		{ 160, 140, 26 },
		{ 160, 141, 26 },
		{ 160, 142, 26 },
		{ 160, 143, 26 },
		{ 160, 144, 26 },
		{ 160, 145, 26 },
		{ 160, 146, 26 },
		{ 160, 147, 26 },
		{ 160, 148, 26 },
		{ 160, 149, 26 },
		{ 160, 150, 26 },
		{ 160, 151, 26 },
		{ 160, 152, 26 },
		{ 160, 153, 26 },
		{ 160, 154, 26 },
		{ 160, 155, 26 },
		{ 160, 156, 26 },
		{ 160, 157, 26 },
		{ 160, 158, 26 },
		{ 160, 159, 26 },
		{ 160, 160, 26 },
		{ 160, 161, 26 },
		{ 160, 162, 26 },
		{ 160, 163, 26 },
		{ 160, 164, 26 },
		{ 160, 165, 26 },
		{ 160, 166, 26 },
		{ 160, 167, 26 },
		{ 160, 168, 26 },
		{ 160, 169, 26 },
		{ 160, 170, 26 },
		{ 160, 171, 26 },
		{ 160, 172, 26 },
		{ 160, 173, 26 },
		{ 160, 174, 26 },
		{ 160, 175, 26 },
		{ 160, 176, 26 },
		{ 160, 177, 26 },
		{ 160, 178, 26 },
		{ 160, 179, 26 },
		{ 160, 180, 26 },
		{ 160, 181, 26 },
		{ 160, 182, 26 },
		{ 160, 183, 26 },
		{ 160, 184, 26 },
		{ 161, 133, 26 },
		{ 161, 134, 26 },
		{ 161, 135, 26 },
		{ 161, 136, 26 },
		{ 161, 137, 26 },
		{ 161, 138, 26 },
		{ 161, 139, 26 },
		{ 161, 140, 26 },
		{ 161, 141, 26 },
		{ 161, 142, 26 },
		{ 161, 143, 26 },
		{ 161, 144, 26 },
		{ 161, 145, 26 },
		{ 161, 146, 26 },
		{ 161, 147, 26 },
		{ 161, 148, 26 },
		{ 161, 149, 26 },
		{ 161, 150, 26 },
		{ 161, 151, 26 },
		{ 161, 152, 26 },
		{ 161, 153, 26 },
		{ 161, 154, 26 },
		{ 161, 155, 26 },
		{ 161, 156, 26 },
		{ 161, 157, 26 },
		{ 161, 158, 26 },
		{ 161, 159, 26 },
		{ 161, 160, 26 },
		{ 161, 161, 26 },
		{ 161, 162, 26 },
		{ 161, 163, 26 },
		{ 161, 164, 26 },
		{ 161, 165, 26 },
		{ 161, 166, 26 },
		{ 161, 167, 26 },
		{ 161, 168, 26 },
		{ 161, 169, 26 },
		{ 161, 170, 26 },
		{ 161, 171, 26 },
		{ 161, 172, 26 },
		{ 161, 173, 26 },
		{ 161, 174, 26 },
		{ 161, 175, 26 },
		{ 161, 176, 26 },
		{ 161, 177, 26 },
		{ 161, 178, 26 },
		{ 161, 179, 26 },
		{ 161, 180, 26 },
		{ 161, 181, 26 },
		{ 161, 182, 26 },
		{ 161, 183, 26 },
		{ 161, 184, 26 },
		{ 162, 133, 26 },
		{ 162, 134, 26 },
		{ 162, 135, 26 },
		{ 162, 136, 26 },
		{ 162, 137, 26 },
		{ 162, 138, 26 },
		{ 162, 139, 26 },
		{ 162, 140, 26 },
		{ 162, 141, 26 },
		{ 162, 142, 26 },
		{ 162, 143, 26 },
		{ 162, 144, 26 },
		{ 162, 145, 26 },
		{ 162, 146, 26 },
		{ 162, 147, 26 },
		{ 162, 148, 26 },
		{ 162, 149, 26 },
		{ 162, 150, 26 },
		{ 162, 151, 26 },
		{ 162, 152, 26 },
		{ 162, 153, 26 },
		{ 162, 154, 26 },
		{ 162, 155, 26 },
		{ 162, 156, 26 },
		{ 162, 157, 26 },
		{ 162, 158, 26 },
		{ 162, 159, 26 },
		{ 162, 160, 26 },
		{ 162, 161, 26 },
		{ 162, 162, 26 },
		{ 162, 163, 26 },
		{ 162, 164, 26 },
		{ 162, 165, 26 },
		{ 162, 166, 26 },
		{ 162, 167, 26 },
		{ 162, 168, 26 },
		{ 162, 169, 26 },
		{ 162, 170, 26 },
		{ 162, 171, 26 },
		{ 162, 172, 26 },
		{ 162, 173, 26 },
		{ 162, 174, 26 },
		{ 162, 175, 26 },
		{ 162, 176, 26 },
		{ 162, 177, 26 },
		{ 162, 178, 26 },
		{ 162, 179, 26 },
		{ 162, 180, 26 },
		{ 162, 181, 26 },
		{ 162, 182, 26 },
		{ 162, 183, 26 },
		{ 162, 184, 26 },
		{ 162, 185, 26 },
		{ 163, 134, 26 },
		{ 163, 135, 26 },
		{ 163, 136, 26 },
		{ 163, 137, 26 },
		{ 163, 138, 26 },
		{ 163, 139, 26 },
		{ 163, 140, 26 },
		{ 163, 141, 26 },
		{ 163, 142, 26 },
		{ 163, 143, 26 },
		{ 163, 144, 26 },
		{ 163, 145, 26 },
		{ 163, 146, 26 },
		{ 163, 147, 26 },
		{ 163, 148, 26 },
		{ 163, 149, 26 },
		{ 163, 150, 26 },
		{ 163, 151, 26 },
		{ 163, 152, 26 },
		{ 163, 153, 26 },
		{ 163, 154, 26 },
		{ 163, 155, 26 },
		{ 163, 156, 26 },
		{ 163, 157, 26 },
		{ 163, 158, 26 },
		{ 163, 159, 26 },
		{ 163, 160, 26 },
		{ 163, 161, 26 },
		{ 163, 162, 26 },
		{ 163, 163, 26 },
		{ 163, 164, 26 },
		{ 163, 165, 26 },
		{ 163, 166, 26 },
		{ 163, 167, 26 },
		{ 163, 168, 26 },
		{ 163, 169, 26 },
		{ 163, 170, 26 },
		{ 163, 171, 26 },
		{ 163, 172, 26 },
		{ 163, 173, 26 },
		{ 163, 174, 26 },
		{ 163, 175, 26 },
		{ 163, 176, 26 },
		{ 163, 177, 26 },
		{ 163, 178, 26 },
		{ 163, 179, 26 },
		{ 163, 180, 26 },
		{ 163, 181, 26 },
		{ 163, 182, 26 },
		{ 163, 183, 26 },
		{ 163, 184, 26 },
		{ 163, 185, 26 },
		{ 163, 186, 26 },
		{ 164, 135, 26 },
		{ 164, 136, 26 },
		{ 164, 137, 26 },
		{ 164, 138, 26 },
		{ 164, 139, 26 },
		{ 164, 140, 26 },
		{ 164, 141, 26 },
		{ 164, 142, 26 },
		{ 164, 143, 26 },
		{ 164, 144, 26 },
		{ 164, 145, 26 },
		{ 164, 146, 26 },
		{ 164, 147, 26 },
		{ 164, 148, 26 },
		{ 164, 149, 26 },
		{ 164, 150, 26 },
		{ 164, 151, 26 },
		{ 164, 152, 26 },
		{ 164, 153, 26 },
		{ 164, 154, 26 },
		{ 164, 155, 26 },
		{ 164, 156, 26 },
		{ 164, 157, 26 },
		{ 164, 158, 26 },
		{ 164, 159, 26 },
		{ 164, 160, 26 },
		{ 164, 161, 26 },
		{ 164, 162, 26 },
		{ 164, 163, 26 },
		{ 164, 164, 26 },
		{ 164, 165, 26 },
		{ 164, 166, 26 },
		{ 164, 167, 26 },
		{ 164, 168, 26 },
		{ 164, 169, 26 },
		{ 164, 170, 26 },
		{ 164, 171, 26 },
		{ 164, 172, 26 },
		{ 164, 173, 26 },
		{ 164, 174, 26 },
		{ 164, 175, 26 },
		{ 164, 176, 26 },
		{ 164, 177, 26 },
		{ 164, 178, 26 },
		{ 164, 179, 26 },
		{ 164, 180, 26 },
		{ 164, 181, 26 },
		{ 164, 182, 26 },
		{ 164, 183, 26 },
		{ 164, 184, 26 },
		{ 164, 185, 26 },
		{ 164, 186, 26 },
		{ 164, 187, 26 },
		{ 165, 136, 26 },
		{ 165, 137, 26 },
		{ 165, 138, 26 },
		{ 165, 139, 26 },
		{ 165, 140, 26 },
		{ 165, 141, 26 },
		{ 165, 142, 26 },
		{ 165, 143, 26 },
		{ 165, 144, 26 },
		{ 165, 145, 26 },
		{ 165, 146, 26 },
		{ 165, 147, 26 },
		{ 165, 148, 26 },
		{ 165, 149, 26 },
		{ 165, 150, 26 },
		{ 165, 151, 26 },
		{ 165, 152, 26 },
		{ 165, 153, 26 },
		{ 165, 154, 26 },
		{ 165, 155, 26 },
		{ 165, 156, 26 },
		{ 165, 157, 26 },
		{ 165, 158, 26 },
		{ 165, 159, 26 },
		{ 165, 160, 26 },
		{ 165, 161, 26 },
		{ 165, 162, 26 },
		{ 165, 163, 26 },
		{ 165, 164, 26 },
		{ 165, 165, 26 },
		{ 165, 166, 26 },
		{ 165, 167, 26 },
		{ 165, 168, 26 },
		{ 165, 169, 26 },
		{ 165, 170, 26 },
		{ 165, 171, 26 },
		{ 165, 172, 26 },
		{ 165, 173, 26 },
		{ 165, 174, 26 },
		{ 165, 175, 26 },
		{ 165, 176, 26 },
		{ 165, 177, 26 },
		{ 165, 178, 26 },
		{ 165, 179, 26 },
		{ 165, 180, 26 },
		{ 165, 181, 26 },
		{ 165, 182, 26 },
		{ 165, 183, 26 },
		{ 165, 184, 26 },
		{ 165, 185, 26 },
		{ 165, 186, 26 },
		{ 165, 187, 26 },
		{ 166, 137, 26 },
		{ 166, 138, 26 },
		{ 166, 139, 26 },
		{ 166, 140, 26 },
		{ 166, 141, 26 },
		{ 166, 142, 26 },
		{ 166, 143, 26 },
		{ 166, 144, 26 },
		{ 166, 145, 26 },
		{ 166, 146, 26 },
		{ 166, 147, 26 },
		{ 166, 148, 26 },
		{ 166, 149, 26 },
		{ 166, 150, 26 },
		{ 166, 151, 26 },
		{ 166, 152, 26 },
		{ 166, 153, 26 },
		{ 166, 154, 26 },
		{ 166, 155, 26 },
		{ 166, 156, 26 },
		{ 166, 157, 26 },
		{ 166, 158, 26 },
		{ 166, 159, 26 },
		{ 166, 160, 26 },
		{ 166, 161, 26 },
		{ 166, 162, 26 },
		{ 166, 163, 26 },
		{ 166, 164, 26 },
		{ 166, 165, 26 },
		{ 166, 166, 26 },
		{ 166, 167, 26 },
		{ 166, 168, 26 },
		{ 166, 169, 26 },
		{ 166, 170, 26 },
		{ 166, 171, 26 },
		{ 166, 172, 26 },
		{ 166, 173, 26 },
		{ 166, 174, 26 },
		{ 166, 175, 26 },
		{ 166, 176, 26 },
		{ 166, 177, 26 },
		{ 166, 178, 26 },
		{ 166, 179, 26 },
		{ 166, 180, 26 },
		{ 166, 181, 26 },
		{ 166, 182, 26 },
		{ 166, 183, 26 },
		{ 166, 184, 26 },
		{ 166, 185, 26 },
		{ 166, 186, 26 },
		{ 166, 187, 26 },
		{ 166, 188, 26 },
		{ 167, 138, 26 },
		{ 167, 139, 26 },
		{ 167, 140, 26 },
		{ 167, 141, 26 },
		{ 167, 142, 26 },
		{ 167, 143, 26 },
		{ 167, 144, 26 },
		{ 167, 145, 26 },
		{ 167, 146, 26 },
		{ 167, 147, 26 },
		{ 167, 148, 26 },
		{ 167, 149, 26 },
		{ 167, 150, 26 },
		{ 167, 151, 26 },
		{ 167, 152, 26 },
		{ 167, 153, 26 },
		{ 167, 154, 26 },
		{ 167, 155, 26 },
		{ 167, 156, 26 },
		{ 167, 157, 26 },
		{ 167, 158, 26 },
		{ 167, 159, 26 },
		{ 167, 160, 26 },
		{ 167, 161, 26 },
		{ 167, 162, 26 },
		{ 167, 163, 26 },
		{ 167, 164, 26 },
		{ 167, 165, 26 },
		{ 167, 166, 26 },
		{ 167, 167, 26 },
		{ 167, 168, 26 },
		{ 167, 169, 26 },
		{ 167, 170, 26 },
		{ 167, 171, 26 },
		{ 167, 172, 26 },
		{ 167, 173, 26 },
		{ 167, 174, 26 },
		{ 167, 175, 26 },
		{ 167, 176, 26 },
		{ 167, 177, 26 },
		{ 167, 178, 26 },
		{ 167, 179, 26 },
		{ 167, 180, 26 },
		{ 167, 181, 26 },
		{ 167, 182, 26 },
		{ 167, 183, 26 },
		{ 167, 184, 26 },
		{ 167, 185, 26 },
		{ 167, 186, 26 },
		{ 167, 187, 26 },
		{ 167, 188, 26 },
		{ 167, 189, 26 },
		{ 168, 139, 26 },
		{ 168, 140, 26 },
		{ 168, 141, 26 },
		{ 168, 142, 26 },
		{ 168, 143, 26 },
		{ 168, 144, 26 },
		{ 168, 145, 26 },
		{ 168, 146, 26 },
		{ 168, 147, 26 },
		{ 168, 148, 26 },
		{ 168, 149, 26 },
		{ 168, 150, 26 },
		{ 168, 151, 26 },
		{ 168, 152, 26 },
		{ 168, 153, 26 },
		{ 168, 154, 26 },
		{ 168, 155, 26 },
		{ 168, 156, 26 },
		{ 168, 157, 26 },
		{ 168, 158, 26 },
		{ 168, 159, 26 },
		{ 168, 160, 26 },
		{ 168, 161, 26 },
		{ 168, 162, 26 },
		{ 168, 163, 26 },
		{ 168, 164, 26 },
		{ 168, 165, 26 },
		{ 168, 166, 26 },
		{ 168, 167, 26 },
		{ 168, 168, 26 },
		{ 168, 169, 26 },
		{ 168, 170, 26 },
		{ 168, 171, 26 },
		{ 168, 172, 26 },
		{ 168, 173, 26 },
		{ 168, 174, 26 },
		{ 168, 175, 26 },
		{ 168, 176, 26 },
		{ 168, 177, 26 },
		{ 168, 178, 26 },
		{ 168, 179, 26 },
		{ 168, 180, 26 },
		{ 168, 181, 26 },
		{ 168, 182, 26 },
		{ 168, 183, 26 },
		{ 168, 184, 26 },
		{ 168, 185, 26 },
		{ 168, 186, 26 },
		{ 168, 187, 26 },
		{ 168, 188, 26 },
		{ 168, 189, 26 },
		{ 168, 190, 26 },
		{ 169, 140, 26 },
		{ 169, 141, 26 },
		{ 169, 142, 26 },
		{ 169, 143, 26 },
		{ 169, 144, 26 },
		{ 169, 145, 26 },
		{ 169, 146, 26 },
		{ 169, 147, 26 },
		{ 169, 148, 26 },
		{ 169, 149, 26 },
		{ 169, 150, 26 },
		{ 169, 151, 26 },
		{ 169, 152, 26 },
		{ 169, 153, 26 },
		{ 169, 154, 26 },
		{ 169, 155, 26 },
		{ 169, 156, 26 },
		{ 169, 157, 26 },
		{ 169, 158, 26 },
		{ 169, 159, 26 },
		{ 169, 160, 26 },
		{ 169, 161, 26 },
		{ 169, 162, 26 },
		{ 169, 163, 26 },
		{ 169, 164, 26 },
		{ 169, 165, 26 },
		{ 169, 166, 26 },
		{ 169, 167, 26 },
		{ 169, 168, 26 },
		{ 169, 169, 26 },
		{ 169, 170, 26 },
		{ 169, 171, 26 },
		{ 169, 172, 26 },
		{ 169, 173, 26 },
		{ 169, 174, 26 },
		{ 169, 175, 26 },
		{ 169, 176, 26 },
		{ 169, 177, 26 },
		{ 169, 178, 26 },
		{ 169, 179, 26 },
		{ 169, 180, 26 },
		{ 169, 181, 26 },
		{ 169, 182, 26 },
		{ 169, 183, 26 },
		{ 169, 184, 26 },
		{ 169, 185, 26 },
		{ 169, 186, 26 },
		{ 169, 187, 26 },
		{ 169, 188, 26 },
		{ 169, 189, 26 },
		{ 169, 190, 26 },
		{ 170, 142, 26 },
		{ 170, 143, 26 },
		{ 170, 144, 26 },
		{ 170, 145, 26 },
		{ 170, 146, 26 },
		{ 170, 147, 26 },
		{ 170, 148, 26 },
		{ 170, 149, 26 },
		{ 170, 150, 26 },
		{ 170, 151, 26 },
		{ 170, 152, 26 },
		{ 170, 153, 26 },
		{ 170, 154, 26 },
		{ 170, 155, 26 },
		{ 170, 156, 26 },
		{ 170, 157, 26 },
		{ 170, 158, 26 },
		{ 170, 159, 26 },
		{ 170, 160, 26 },
		{ 170, 161, 26 },
		{ 170, 162, 26 },
		{ 170, 163, 26 },
		{ 170, 164, 26 },
		{ 170, 165, 26 },
		{ 170, 166, 26 },
		{ 170, 167, 26 },
		{ 170, 168, 26 },
		{ 170, 169, 26 },
		{ 170, 170, 26 },
		{ 170, 171, 26 },
		{ 170, 172, 26 },
		{ 170, 173, 26 },
		{ 170, 174, 26 },
		{ 170, 175, 26 },
		{ 170, 176, 26 },
		{ 170, 177, 26 },
		{ 170, 178, 26 },
		{ 170, 179, 26 },
		{ 170, 180, 26 },
		{ 170, 181, 26 },
		{ 170, 182, 26 },
		{ 170, 183, 26 },
		{ 170, 184, 26 },
		{ 170, 185, 26 },
		{ 170, 186, 26 },
		{ 170, 187, 26 },
		{ 170, 188, 26 },
		{ 170, 189, 26 },
		{ 170, 190, 26 },
		{ 170, 191, 26 },
		{ 171, 143, 26 },
		{ 171, 144, 26 },
		{ 171, 145, 26 },
		{ 171, 146, 26 },
		{ 171, 147, 26 },
		{ 171, 148, 26 },
		{ 171, 149, 26 },
		{ 171, 150, 26 },
		{ 171, 151, 26 },
		{ 171, 152, 26 },
		{ 171, 153, 26 },
		{ 171, 154, 26 },
		{ 171, 155, 26 },
		{ 171, 156, 26 },
		{ 171, 157, 26 },
		{ 171, 158, 26 },
		{ 171, 159, 26 },
		{ 171, 160, 26 },
		{ 171, 161, 26 },
		{ 171, 162, 26 },
		{ 171, 163, 26 },
		{ 171, 164, 26 },
		{ 171, 165, 26 },
		{ 171, 166, 26 },
		{ 171, 167, 26 },
		{ 171, 168, 26 },
		{ 171, 169, 26 },
		{ 171, 170, 26 },
		{ 171, 171, 26 },
		{ 171, 172, 26 },
		{ 171, 173, 26 },
		{ 171, 174, 26 },
		{ 171, 175, 26 },
		{ 171, 176, 26 },
		{ 171, 177, 26 },
		{ 171, 178, 26 },
		{ 171, 179, 26 },
		{ 171, 180, 26 },
		{ 171, 181, 26 },
		{ 171, 182, 26 },
		{ 171, 183, 26 },
		{ 171, 184, 26 },
		{ 171, 185, 26 },
		{ 171, 186, 26 },
		{ 171, 187, 26 },
		{ 171, 188, 26 },
		{ 171, 189, 26 },
		{ 171, 190, 26 },
		{ 171, 191, 26 },
		{ 171, 192, 26 },
		{ 172, 144, 26 },
		{ 172, 145, 26 },
		{ 172, 146, 26 },
		{ 172, 147, 26 },
		{ 172, 148, 26 },
		{ 172, 149, 26 },
		{ 172, 150, 26 },
		{ 172, 151, 26 },
		{ 172, 152, 26 },
		{ 172, 153, 26 },
		{ 172, 154, 26 },
		{ 172, 155, 26 },
		{ 172, 156, 26 },
		{ 172, 157, 26 },
		{ 172, 158, 26 },
		{ 172, 159, 26 },
		{ 172, 160, 26 },
		{ 172, 161, 26 },
		{ 172, 162, 26 },
		{ 172, 163, 26 },
		{ 172, 164, 26 },
		{ 172, 165, 26 },
		{ 172, 166, 26 },
		{ 172, 167, 26 },
		{ 172, 168, 26 },
		{ 172, 169, 26 },
		{ 172, 170, 26 },
		{ 172, 171, 26 },
		{ 172, 172, 26 },
		{ 172, 173, 26 },
		{ 172, 174, 26 },
		{ 172, 175, 26 },
		{ 172, 176, 26 },
		{ 172, 177, 26 },
		{ 172, 178, 26 },
		{ 172, 179, 26 },
		{ 172, 180, 26 },
		{ 172, 181, 26 },
		{ 172, 182, 26 },
		{ 172, 183, 26 },
		{ 172, 184, 26 },
		{ 172, 185, 26 },
		{ 172, 186, 26 },
		{ 172, 187, 26 },
		{ 172, 188, 26 },
		{ 172, 189, 26 },
		{ 172, 190, 26 },
		{ 172, 191, 26 },
		{ 172, 192, 26 },
		{ 172, 193, 26 },
		{ 173, 146, 26 },
		{ 173, 147, 26 },
		{ 173, 148, 26 },
		{ 173, 149, 26 },
		{ 173, 150, 26 },
		{ 173, 151, 26 },
		{ 173, 152, 26 },
		{ 173, 153, 26 },
		{ 173, 154, 26 },
		{ 173, 155, 26 },
		{ 173, 156, 26 },
		{ 173, 157, 26 },
		{ 173, 158, 26 },
		{ 173, 159, 26 },
		{ 173, 160, 26 },
		{ 173, 161, 26 },
		{ 173, 162, 26 },
		{ 173, 163, 26 },
		{ 173, 164, 26 },
		{ 173, 165, 26 },
		{ 173, 166, 26 },
		{ 173, 167, 26 },
		{ 173, 168, 26 },
		{ 173, 169, 26 },
		{ 173, 170, 26 },
		{ 173, 171, 26 },
		{ 173, 172, 26 },
		{ 173, 173, 26 },
		{ 173, 174, 26 },
		{ 173, 175, 26 },
		{ 173, 176, 26 },
		{ 173, 177, 26 },
		{ 173, 178, 26 },
		{ 173, 179, 26 },
		{ 173, 180, 26 },
		{ 173, 181, 26 },
		{ 173, 182, 26 },
		{ 173, 183, 26 },
		{ 173, 184, 26 },
		{ 173, 185, 26 },
		{ 173, 186, 26 },
		{ 173, 187, 26 },
		{ 173, 188, 26 },
		{ 173, 189, 26 },
		{ 173, 190, 26 },
		{ 173, 191, 26 },
		{ 173, 192, 26 },
		{ 173, 193, 26 },
		{ 174, 147, 26 },
		{ 174, 148, 26 },
		{ 174, 149, 26 },
		{ 174, 150, 26 },
		{ 174, 151, 26 },
		{ 174, 152, 26 },
		{ 174, 153, 26 },
		{ 174, 154, 26 },
		{ 174, 155, 26 },
		{ 174, 156, 26 },
		{ 174, 157, 26 },
		{ 174, 158, 26 },
		{ 174, 159, 26 },
		{ 174, 160, 26 },
		{ 174, 161, 26 },
		{ 174, 162, 26 },
		{ 174, 163, 26 },
		{ 174, 164, 26 },
		{ 174, 165, 26 },
		{ 174, 166, 26 },
		{ 174, 167, 26 },
		{ 174, 168, 26 },
		{ 174, 169, 26 },
		{ 174, 170, 26 },
		{ 174, 171, 26 },
		{ 174, 172, 26 },
		{ 174, 173, 26 },
		{ 174, 174, 26 },
		{ 174, 175, 26 },
		{ 174, 176, 26 },
		{ 174, 177, 26 },
		{ 174, 178, 26 },
		{ 174, 179, 26 },
		{ 174, 180, 26 },
		{ 174, 181, 26 },
		{ 174, 182, 26 },
		{ 174, 183, 26 },
		{ 174, 184, 26 },
		{ 174, 185, 26 },
		{ 174, 186, 26 },
		{ 174, 187, 26 },
		{ 174, 188, 26 },
		{ 174, 189, 26 },
		{ 174, 190, 26 },
		{ 174, 191, 26 },
		{ 174, 192, 26 },
		{ 174, 193, 26 },
		{ 174, 194, 26 },
		{ 175, 148, 26 },
		{ 175, 149, 26 },
		{ 175, 150, 26 },
		{ 175, 151, 26 },
		{ 175, 152, 26 },
		{ 175, 153, 26 },
		{ 175, 154, 26 },
		{ 175, 155, 26 },
		{ 175, 156, 26 },
		{ 175, 157, 26 },
		{ 175, 158, 26 },
		{ 175, 159, 26 },
		{ 175, 160, 26 },
		{ 175, 161, 26 },
		{ 175, 162, 26 },
		{ 175, 163, 26 },
		{ 175, 164, 26 },
		{ 175, 165, 26 },
		{ 175, 166, 26 },
		{ 175, 167, 26 },
		{ 175, 168, 26 },
		{ 175, 169, 26 },
		{ 175, 170, 26 },
		{ 175, 171, 26 },
		{ 175, 172, 26 },
		{ 175, 173, 26 },
		{ 175, 174, 26 },
		{ 175, 175, 26 },
		{ 175, 176, 26 },
		{ 175, 177, 26 },
		{ 175, 178, 26 },
		{ 175, 179, 26 },
		{ 175, 180, 26 },
		{ 175, 181, 26 },
		{ 175, 182, 26 },
		{ 175, 183, 26 },
		{ 175, 184, 26 },
		{ 175, 185, 26 },
		{ 175, 186, 26 },
		{ 175, 187, 26 },
		{ 175, 188, 26 },
		{ 175, 189, 26 },
		{ 175, 190, 26 },
		{ 175, 191, 26 },
		{ 175, 192, 26 },
		{ 175, 193, 26 },
		{ 175, 194, 26 },
		{ 175, 195, 26 },
		{ 176, 150, 26 },
		{ 176, 151, 26 },
		{ 176, 152, 26 },
		{ 176, 153, 26 },
		{ 176, 154, 26 },
		{ 176, 155, 26 },
		{ 176, 156, 26 },
		{ 176, 157, 26 },
		{ 176, 158, 26 },
		{ 176, 159, 26 },
		{ 176, 160, 26 },
		{ 176, 161, 26 },
		{ 176, 162, 26 },
		{ 176, 163, 26 },
		{ 176, 164, 26 },
		{ 176, 165, 26 },
		{ 176, 166, 26 },
		{ 176, 167, 26 },
		{ 176, 168, 26 },
		{ 176, 169, 26 },
		{ 176, 170, 26 },
		{ 176, 171, 26 },
		{ 176, 172, 26 },
		{ 176, 173, 26 },
		{ 176, 174, 26 },
		{ 176, 175, 26 },
		{ 176, 176, 26 },
		{ 176, 177, 26 },
		{ 176, 178, 26 },
		{ 176, 179, 26 },
		{ 176, 180, 26 },
		{ 176, 181, 26 },
		{ 176, 182, 26 },
		{ 176, 183, 26 },
		{ 176, 184, 26 },
		{ 176, 185, 26 },
		{ 176, 186, 26 },
		{ 176, 187, 26 },
		{ 176, 188, 26 },
		{ 176, 189, 26 },
		{ 176, 190, 26 },
		{ 176, 191, 26 },
		{ 176, 192, 26 },
		{ 176, 193, 26 },
		{ 176, 194, 26 },
		{ 176, 195, 26 },
		{ 176, 196, 26 },
		{ 177, 151, 26 },
		{ 177, 152, 26 },
		{ 177, 153, 26 },
		{ 177, 154, 26 },
		{ 177, 155, 26 },
		{ 177, 156, 26 },
		{ 177, 157, 26 },
		{ 177, 158, 26 },
		{ 177, 159, 26 },
		{ 177, 160, 26 },
		{ 177, 161, 26 },
		{ 177, 162, 26 },
		{ 177, 163, 26 },
		{ 177, 164, 26 },
		{ 177, 165, 26 },
		{ 177, 166, 26 },
		{ 177, 167, 26 },
		{ 177, 168, 26 },
		{ 177, 169, 26 },
		{ 177, 170, 26 },
		{ 177, 171, 26 },
		{ 177, 172, 26 },
		{ 177, 173, 26 },
		{ 177, 174, 26 },
		{ 177, 175, 26 },
		{ 177, 176, 26 },
		{ 177, 177, 26 },
		{ 177, 178, 26 },
		{ 177, 179, 26 },
		{ 177, 180, 26 },
		{ 177, 181, 26 },
		{ 177, 182, 26 },
		{ 177, 183, 26 },
		{ 177, 184, 26 },
		{ 177, 185, 26 },
		{ 177, 186, 26 },
		{ 177, 187, 26 },
		{ 177, 188, 26 },
		{ 177, 189, 26 },
		{ 177, 190, 26 },
		{ 177, 191, 26 },
		{ 177, 192, 26 },
		{ 177, 193, 26 },
		{ 177, 194, 26 },
		{ 177, 195, 26 },
		{ 177, 196, 26 },
		{ 178, 152, 26 },
		{ 178, 153, 26 },
		{ 178, 154, 26 },
		{ 178, 155, 26 },
		{ 178, 156, 26 },
		{ 178, 157, 26 },
		{ 178, 158, 26 },
		{ 178, 159, 26 },
		{ 178, 160, 26 },
		{ 178, 161, 26 },
		{ 178, 162, 26 },
		{ 178, 163, 26 },
		{ 178, 164, 26 },
		{ 178, 165, 26 },
		{ 178, 166, 26 },
		{ 178, 167, 26 },
		{ 178, 168, 26 },
		{ 178, 169, 26 },
		{ 178, 170, 26 },
		{ 178, 171, 26 },
		{ 178, 172, 26 },
		{ 178, 173, 26 },
		{ 178, 174, 26 },
		{ 178, 175, 26 },
		{ 178, 176, 26 },
		{ 178, 177, 26 },
		{ 178, 178, 26 },
		{ 178, 179, 26 },
		{ 178, 180, 26 },
		{ 178, 181, 26 },
		{ 178, 182, 26 },
		{ 178, 183, 26 },
		{ 178, 184, 26 },
		{ 178, 185, 26 },
		{ 178, 186, 26 },
		{ 178, 187, 26 },
		{ 178, 188, 26 },
		{ 178, 189, 26 },
		{ 178, 190, 26 },
		{ 178, 191, 26 },
		{ 178, 192, 26 },
		{ 178, 193, 26 },
		{ 178, 194, 26 },
		{ 178, 195, 26 },
		{ 178, 196, 26 },
		{ 178, 197, 26 },
		{ 179, 154, 26 },
		{ 179, 155, 26 },
		{ 179, 156, 26 },
		{ 179, 157, 26 },
		{ 179, 158, 26 },
		{ 179, 159, 26 },
		{ 179, 160, 26 },
		{ 179, 161, 26 },
		{ 179, 162, 26 },
		{ 179, 163, 26 },
		{ 179, 164, 26 },
		{ 179, 165, 26 },
		{ 179, 166, 26 },
		{ 179, 167, 26 },
		{ 179, 168, 26 },
		{ 179, 169, 26 },
		{ 179, 170, 26 },
		{ 179, 171, 26 },
		{ 179, 172, 26 },
		{ 179, 173, 26 },
		{ 179, 174, 26 },
		{ 179, 175, 26 },
		{ 179, 176, 26 },
		{ 179, 177, 26 },
		{ 179, 178, 26 },
		{ 179, 179, 26 },
		{ 179, 180, 26 },
		{ 179, 181, 26 },
		{ 179, 182, 26 },
		{ 179, 183, 26 },
		{ 179, 184, 26 },
		{ 179, 185, 26 },
		{ 179, 186, 26 },
		{ 179, 187, 26 },
		{ 179, 188, 26 },
		{ 179, 189, 26 },
		{ 179, 190, 26 },
		{ 179, 191, 26 },
		{ 179, 192, 26 },
		{ 179, 193, 26 },
		{ 179, 194, 26 },
		{ 179, 195, 26 },
		{ 179, 196, 26 },
		{ 179, 197, 26 },
		{ 179, 198, 26 },
		{ 180, 155, 26 },
		{ 180, 156, 26 },
		{ 180, 157, 26 },
		{ 180, 158, 26 },
		{ 180, 159, 26 },
		{ 180, 160, 26 },
		{ 180, 161, 26 },
		{ 180, 162, 26 },
		{ 180, 163, 26 },
		{ 180, 164, 26 },
		{ 180, 165, 26 },
		{ 180, 166, 26 },
		{ 180, 167, 26 },
		{ 180, 168, 26 },
		{ 180, 169, 26 },
		{ 180, 170, 26 },
		{ 180, 171, 26 },
		{ 180, 172, 26 },
		{ 180, 173, 26 },
		{ 180, 174, 26 },
		{ 180, 175, 26 },
		{ 180, 176, 26 },
		{ 180, 177, 26 },
		{ 180, 178, 26 },
		{ 180, 179, 26 },
		{ 180, 180, 26 },
		{ 180, 181, 26 },
		{ 180, 182, 26 },
		{ 180, 183, 26 },
		{ 180, 184, 26 },
		{ 180, 185, 26 },
		{ 180, 186, 26 },
		{ 180, 187, 26 },
		{ 180, 188, 26 },
		{ 180, 189, 26 },
		{ 180, 190, 26 },
		{ 180, 191, 26 },
		{ 180, 192, 26 },
		{ 180, 193, 26 },
		{ 180, 194, 26 },
		{ 180, 195, 26 },
		{ 180, 196, 26 },
		{ 180, 197, 26 },
		{ 180, 198, 26 },
		{ 180, 199, 26 },
		{ 181, 156, 26 },
		{ 181, 157, 26 },
		{ 181, 158, 26 },
		{ 181, 159, 26 },
		{ 181, 160, 26 },
		{ 181, 161, 26 },
		{ 181, 162, 26 },
		{ 181, 163, 26 },
		{ 181, 164, 26 },
		{ 181, 165, 26 },
		{ 181, 166, 26 },
		{ 181, 167, 26 },
		{ 181, 168, 26 },
		{ 181, 169, 26 },
		{ 181, 170, 26 },
		{ 181, 171, 26 },
		{ 181, 172, 26 },
		{ 181, 173, 26 },
		{ 181, 174, 26 },
		{ 181, 175, 26 },
		{ 181, 176, 26 },
		{ 181, 177, 26 },
		{ 181, 178, 26 },
		{ 181, 179, 26 },
		{ 181, 180, 26 },
		{ 181, 181, 26 },
		{ 181, 182, 26 },
		{ 181, 183, 26 },
		{ 181, 184, 26 },
		{ 181, 185, 26 },
		{ 181, 186, 26 },
		{ 181, 187, 26 },
		{ 181, 188, 26 },
		{ 181, 189, 26 },
		{ 181, 190, 26 },
		{ 181, 191, 26 },
		{ 181, 192, 26 },
		{ 181, 193, 26 },
		{ 181, 194, 26 },
		{ 181, 195, 26 },
		{ 181, 196, 26 },
		{ 181, 197, 26 },
		{ 181, 198, 26 },
		{ 181, 199, 26 },
		{ 182, 158, 26 },
		{ 182, 159, 26 },
		{ 182, 160, 26 },
		{ 182, 161, 26 },
		{ 182, 162, 26 },
		{ 182, 163, 26 },
		{ 182, 164, 26 },
		{ 182, 165, 26 },
		{ 182, 166, 26 },
		{ 182, 167, 26 },
		{ 182, 168, 26 },
		{ 182, 169, 26 },
		{ 182, 170, 26 },
		{ 182, 171, 26 },
		{ 182, 172, 26 },
		{ 182, 173, 26 },
		{ 182, 174, 26 },
		{ 182, 175, 26 },
		{ 182, 176, 26 },
		{ 182, 177, 26 },
		{ 182, 178, 26 },
		{ 182, 179, 26 },
		{ 182, 180, 26 },
		{ 182, 181, 26 },
		{ 182, 182, 26 },
		{ 182, 183, 26 },
		{ 182, 184, 26 },
		{ 182, 185, 26 },
		{ 182, 186, 26 },
		{ 182, 187, 26 },
		{ 182, 188, 26 },
		{ 182, 189, 26 },
		{ 182, 190, 26 },
		{ 182, 191, 26 },
		{ 182, 192, 26 },
		{ 182, 193, 26 },
		{ 182, 194, 26 },
		{ 182, 195, 26 },
		{ 182, 196, 26 },
		{ 182, 197, 26 },
		{ 182, 198, 26 },
		{ 182, 199, 26 },
		{ 182, 200, 26 },
		{ 183, 159, 26 },
		{ 183, 160, 26 },
		{ 183, 161, 26 },
		{ 183, 162, 26 },
		{ 183, 163, 26 },
		{ 183, 164, 26 },
		{ 183, 165, 26 },
		{ 183, 166, 26 },
		{ 183, 167, 26 },
		{ 183, 168, 26 },
		{ 183, 169, 26 },
		{ 183, 170, 26 },
		{ 183, 171, 26 },
		{ 183, 172, 26 },
		{ 183, 173, 26 },
		{ 183, 174, 26 },
		{ 183, 175, 26 },
		{ 183, 176, 26 },
		{ 183, 177, 26 },
		{ 183, 178, 26 },
		{ 183, 179, 26 },
		{ 183, 180, 26 },
		{ 183, 181, 26 },
		{ 183, 182, 26 },
		{ 183, 183, 26 },
		{ 183, 184, 26 },
		{ 183, 185, 26 },
		{ 183, 186, 26 },
		{ 183, 187, 26 },
		{ 183, 188, 26 },
		{ 183, 189, 26 },
		{ 183, 190, 26 },
		{ 183, 191, 26 },
		{ 183, 192, 26 },
		{ 183, 193, 26 },
		{ 183, 194, 26 },
		{ 183, 195, 26 },
		{ 183, 196, 26 },
		{ 183, 197, 26 },
		{ 183, 198, 26 },
		{ 183, 199, 26 },
		{ 183, 200, 26 },
		{ 183, 201, 26 },
		{ 184, 160, 26 },
		{ 184, 161, 26 },
		{ 184, 162, 26 },
		{ 184, 163, 26 },
		{ 184, 164, 26 },
		{ 184, 165, 26 },
		{ 184, 166, 26 },
		{ 184, 167, 26 },
		{ 184, 168, 26 },
		{ 184, 169, 26 },
		{ 184, 170, 26 },
		{ 184, 171, 26 },
		{ 184, 172, 26 },
		{ 184, 173, 26 },
		{ 184, 174, 26 },
		{ 184, 175, 26 },
		{ 184, 176, 26 },
		{ 184, 177, 26 },
		{ 184, 178, 26 },
		{ 184, 179, 26 },
		{ 184, 180, 26 },
		{ 184, 181, 26 },
		{ 184, 182, 26 },
		{ 184, 183, 26 },
		{ 184, 184, 26 },
		{ 184, 185, 26 },
		{ 184, 186, 26 },
		{ 184, 187, 26 },
		{ 184, 188, 26 },
		{ 184, 189, 26 },
		{ 184, 190, 26 },
		{ 184, 191, 26 },
		{ 184, 192, 26 },
		{ 184, 193, 26 },
		{ 184, 194, 26 },
		{ 184, 195, 26 },
		{ 184, 196, 26 },
		{ 184, 197, 26 },
		{ 184, 198, 26 },
		{ 184, 199, 26 },
		{ 184, 200, 26 },
		{ 184, 201, 26 },
		{ 184, 202, 26 },
		{ 185, 162, 26 },
		{ 185, 163, 26 },
		{ 185, 164, 26 },
		{ 185, 165, 26 },
		{ 185, 166, 26 },
		{ 185, 167, 26 },
		{ 185, 168, 26 },
		{ 185, 169, 26 },
		{ 185, 170, 26 },
		{ 185, 171, 26 },
		{ 185, 172, 26 },
		{ 185, 173, 26 },
		{ 185, 174, 26 },
		{ 185, 175, 26 },
		{ 185, 176, 26 },
		{ 185, 177, 26 },
		{ 185, 178, 26 },
		{ 185, 179, 26 },
		{ 185, 180, 26 },
		{ 185, 181, 26 },
		{ 185, 182, 26 },
		{ 185, 183, 26 },
		{ 185, 184, 26 },
		{ 185, 185, 26 },
		{ 185, 186, 26 },
		{ 185, 187, 26 },
		{ 185, 188, 26 },
		{ 185, 189, 26 },
		{ 185, 190, 26 },
		{ 185, 191, 26 },
		{ 185, 192, 26 },
		{ 185, 193, 26 },
		{ 185, 194, 26 },
		{ 185, 195, 26 },
		{ 185, 196, 26 },
		{ 185, 197, 26 },
		{ 185, 198, 26 },
		{ 185, 199, 26 },
		{ 185, 200, 26 },
		{ 185, 201, 26 },
		{ 185, 202, 26 },
		{ 186, 163, 26 },
		{ 186, 164, 26 },
		{ 186, 165, 26 },
		{ 186, 166, 26 },
		{ 186, 167, 26 },
		{ 186, 168, 26 },
		{ 186, 169, 26 },
		{ 186, 170, 26 },
		{ 186, 171, 26 },
		{ 186, 172, 26 },
		{ 186, 173, 26 },
		{ 186, 174, 26 },
		{ 186, 175, 26 },
		{ 186, 176, 26 },
		{ 186, 177, 26 },
		{ 186, 178, 26 },
		{ 186, 179, 26 },
		{ 186, 180, 26 },
		{ 186, 181, 26 },
		{ 186, 182, 26 },
		{ 186, 183, 26 },
		{ 186, 184, 26 },
		{ 186, 185, 26 },
		{ 186, 186, 26 },
		{ 186, 187, 26 },
		{ 186, 188, 26 },
		{ 186, 189, 26 },
		{ 186, 190, 26 },
		{ 186, 191, 26 },
		{ 186, 192, 26 },
		{ 186, 193, 26 },
		{ 186, 194, 26 },
		{ 186, 195, 26 },
		{ 186, 196, 26 },
		{ 186, 197, 26 },
		{ 186, 198, 26 },
		{ 186, 199, 26 },
		{ 186, 200, 26 },
		{ 186, 201, 26 },
		{ 186, 202, 26 },
		{ 186, 203, 26 },
		{ 187, 164, 26 },
		{ 187, 165, 26 },
		{ 187, 166, 26 },
		{ 187, 167, 26 },
		{ 187, 168, 26 },
		{ 187, 169, 26 },
		{ 187, 170, 26 },
		{ 187, 171, 26 },
		{ 187, 172, 26 },
		{ 187, 173, 26 },
		{ 187, 174, 26 },
		{ 187, 175, 26 },
		{ 187, 176, 26 },
		{ 187, 177, 26 },
		{ 187, 178, 26 },
		{ 187, 179, 26 },
		{ 187, 180, 26 },
		{ 187, 181, 26 },
		{ 187, 182, 26 },
		{ 187, 183, 26 },
		{ 187, 184, 26 },
		{ 187, 185, 26 },
		{ 187, 186, 26 },
		{ 187, 187, 26 },
		{ 187, 188, 26 },
		{ 187, 189, 26 },
		{ 187, 190, 26 },
		{ 187, 191, 26 },
		{ 187, 192, 26 },
		{ 187, 193, 26 },
		{ 187, 194, 26 },
		{ 187, 195, 26 },
		{ 187, 196, 26 },
		{ 187, 197, 26 },
		{ 187, 198, 26 },
		{ 187, 199, 26 },
		{ 187, 200, 26 },
		{ 187, 201, 26 },
		{ 187, 202, 26 },
		{ 187, 203, 26 },
		{ 187, 204, 26 },
		{ 188, 166, 26 },
		{ 188, 167, 26 },
		{ 188, 168, 26 },
		{ 188, 169, 26 },
		{ 188, 170, 26 },
		{ 188, 171, 26 },
		{ 188, 172, 26 },
		{ 188, 173, 26 },
		{ 188, 174, 26 },
		{ 188, 175, 26 },
		{ 188, 176, 26 },
		{ 188, 177, 26 },
		{ 188, 178, 26 },
		{ 188, 179, 26 },
		{ 188, 180, 26 },
		{ 188, 181, 26 },
		{ 188, 182, 26 },
		{ 188, 183, 26 },
		{ 188, 184, 26 },
		{ 188, 185, 26 },
		{ 188, 186, 26 },
		{ 188, 187, 26 },
		{ 188, 188, 26 },
		{ 188, 189, 26 },
		{ 188, 190, 26 },
		{ 188, 191, 26 },
		{ 188, 192, 26 },
		{ 188, 193, 26 },
		{ 188, 194, 26 },
		{ 188, 195, 26 },
		{ 188, 196, 26 },
		{ 188, 197, 26 },
		{ 188, 198, 26 },
		{ 188, 199, 26 },
		{ 188, 200, 26 },
		{ 188, 201, 26 },
		{ 188, 202, 26 },
		{ 188, 203, 26 },
		{ 188, 204, 26 },
		{ 188, 205, 26 },
		{ 189, 167, 26 },
		{ 189, 168, 26 },
		{ 189, 169, 26 },
		{ 189, 170, 26 },
		{ 189, 171, 26 },
		{ 189, 172, 26 },
		{ 189, 173, 26 },
		{ 189, 174, 26 },
		{ 189, 175, 26 },
		{ 189, 176, 26 },
		{ 189, 177, 26 },
		{ 189, 178, 26 },
		{ 189, 179, 26 },
		{ 189, 180, 26 },
		{ 189, 181, 26 },
		{ 189, 182, 26 },
		{ 189, 183, 26 },
		{ 189, 184, 26 },
		{ 189, 185, 26 },
		{ 189, 186, 26 },
		{ 189, 187, 26 },
		{ 189, 188, 26 },
		{ 189, 189, 26 },
		{ 189, 190, 26 },
		{ 189, 191, 26 },
		{ 189, 192, 26 },
		{ 189, 193, 26 },
		{ 189, 194, 26 },
		{ 189, 195, 26 },
		{ 189, 196, 26 },
		{ 189, 197, 26 },
		{ 189, 198, 26 },
		{ 189, 199, 26 },
		{ 189, 200, 26 },
		{ 189, 201, 26 },
		{ 189, 202, 26 },
		{ 189, 203, 26 },
		{ 189, 204, 26 },
		{ 189, 205, 26 },
		{ 190, 168, 26 },
		{ 190, 169, 26 },
		{ 190, 170, 26 },
		{ 190, 171, 26 },
		{ 190, 172, 26 },
		{ 190, 173, 26 },
		{ 190, 174, 26 },
		{ 190, 175, 26 },
		{ 190, 176, 26 },
		{ 190, 177, 26 },
		{ 190, 178, 26 },
		{ 190, 179, 26 },
		{ 190, 180, 26 },
		{ 190, 181, 26 },
		{ 190, 182, 26 },
		{ 190, 183, 26 },
		{ 190, 184, 26 },
		{ 190, 185, 26 },
		{ 190, 186, 26 },
		{ 190, 187, 26 },
		{ 190, 188, 26 },
		{ 190, 189, 26 },
		{ 190, 190, 26 },
		{ 190, 191, 26 },
		{ 190, 192, 26 },
		{ 190, 193, 26 },
		{ 190, 194, 26 },
		{ 190, 195, 26 },
		{ 190, 196, 26 },
		{ 190, 197, 26 },
		{ 190, 198, 26 },
		{ 190, 199, 26 },
		{ 190, 200, 26 },
		{ 190, 201, 26 },
		{ 190, 202, 26 },
		{ 190, 203, 26 },
		{ 190, 204, 26 },
		{ 190, 205, 26 },
		{ 190, 206, 26 },
		{ 191, 170, 26 },
		{ 191, 171, 26 },
		{ 191, 172, 26 },
		{ 191, 173, 26 },
		{ 191, 174, 26 },
		{ 191, 175, 26 },
		{ 191, 176, 26 },
		{ 191, 177, 26 },
		{ 191, 178, 26 },
		{ 191, 179, 26 },
		{ 191, 180, 26 },
		{ 191, 181, 26 },
		{ 191, 182, 26 },
		{ 191, 183, 26 },
		{ 191, 184, 26 },
		{ 191, 185, 26 },
		{ 191, 186, 26 },
		{ 191, 187, 26 },
		{ 191, 188, 26 },
		{ 191, 189, 26 },
		{ 191, 190, 26 },
		{ 191, 191, 26 },
		{ 191, 192, 26 },
		{ 191, 193, 26 },
		{ 191, 194, 26 },
		{ 191, 195, 26 },
		{ 191, 196, 26 },
		{ 191, 197, 26 },
		{ 191, 198, 26 },
		{ 191, 199, 26 },
		{ 191, 200, 26 },
		{ 191, 201, 26 },
		{ 191, 202, 26 },
		{ 191, 203, 26 },
		{ 191, 204, 26 },
		{ 191, 205, 26 },
		{ 191, 206, 26 },
		{ 191, 207, 26 },
		{ 192, 171, 26 },
		{ 192, 172, 26 },
		{ 192, 173, 26 },
		{ 192, 174, 26 },
		{ 192, 175, 26 },
		{ 192, 176, 26 },
		{ 192, 177, 26 },
		{ 192, 178, 26 },
		{ 192, 179, 26 },
		{ 192, 180, 26 },
		{ 192, 181, 26 },
		{ 192, 182, 26 },
		{ 192, 183, 26 },
		{ 192, 184, 26 },
		{ 192, 185, 26 },
		{ 192, 186, 26 },
		{ 192, 187, 26 },
		{ 192, 188, 26 },
		{ 192, 189, 26 },
		{ 192, 190, 26 },
		{ 192, 191, 26 },
		{ 192, 192, 26 },
		{ 192, 193, 26 },
		{ 192, 194, 26 },
		{ 192, 195, 26 },
		{ 192, 196, 26 },
		{ 192, 197, 26 },
		{ 192, 198, 26 },
		{ 192, 199, 26 },
		{ 192, 200, 26 },
		{ 192, 201, 26 },
		{ 192, 202, 26 },
		{ 192, 203, 26 },
		{ 192, 204, 26 },
		{ 192, 205, 26 },
		{ 192, 206, 26 },
		{ 192, 207, 26 },
		{ 192, 208, 26 },
		{ 193, 172, 26 },
		{ 193, 173, 26 },
		{ 193, 174, 26 },
		{ 193, 175, 26 },
		{ 193, 176, 26 },
		{ 193, 177, 26 },
		{ 193, 178, 26 },
		{ 193, 179, 26 },
		{ 193, 180, 26 },
		{ 193, 181, 26 },
		{ 193, 182, 26 },
		{ 193, 183, 26 },
		{ 193, 184, 26 },
		{ 193, 185, 26 },
		{ 193, 186, 26 },
		{ 193, 187, 26 },
		{ 193, 188, 26 },
		{ 193, 189, 26 },
		{ 193, 190, 26 },
		{ 193, 191, 26 },
		{ 193, 192, 26 },
		{ 193, 193, 26 },
		{ 193, 194, 26 },
		{ 193, 195, 26 },
		{ 193, 196, 26 },
		{ 193, 197, 26 },
		{ 193, 198, 26 },
		{ 193, 199, 26 },
		{ 193, 200, 26 },
		{ 193, 201, 26 },
		{ 193, 202, 26 },
		{ 193, 203, 26 },
		{ 193, 204, 26 },
		{ 193, 205, 26 },
		{ 193, 206, 26 },
		{ 193, 207, 26 },
		{ 193, 208, 26 },
		{ 194, 174, 26 },
		{ 194, 175, 26 },
		{ 194, 176, 26 },
		{ 194, 177, 26 },
		{ 194, 178, 26 },
		{ 194, 179, 26 },
		{ 194, 180, 26 },
		{ 194, 181, 26 },
		{ 194, 182, 26 },
		{ 194, 183, 26 },
		{ 194, 184, 26 },
		{ 194, 185, 26 },
		{ 194, 186, 26 },
		{ 194, 187, 26 },
		{ 194, 188, 26 },
		{ 194, 189, 26 },
		{ 194, 190, 26 },
		{ 194, 191, 26 },
		{ 194, 192, 26 },
		{ 194, 193, 26 },
		{ 194, 194, 26 },
		{ 194, 195, 26 },
		{ 194, 196, 26 },
		{ 194, 197, 26 },
		{ 194, 198, 26 },
		{ 194, 199, 26 },
		{ 194, 200, 26 },
		{ 194, 201, 26 },
		{ 194, 202, 26 },
		{ 194, 203, 26 },
		{ 194, 204, 26 },
		{ 194, 205, 26 },
		{ 194, 206, 26 },
		{ 194, 207, 26 },
		{ 194, 208, 26 },
		{ 194, 209, 26 },
		{ 195, 175, 26 },
		{ 195, 176, 26 },
		{ 195, 177, 26 },
		{ 195, 178, 26 },
		{ 195, 179, 26 },
		{ 195, 180, 26 },
		{ 195, 181, 26 },
		{ 195, 182, 26 },
		{ 195, 183, 26 },
		{ 195, 184, 26 },
		{ 195, 185, 26 },
		{ 195, 186, 26 },
		{ 195, 187, 26 },
		{ 195, 188, 26 },
		{ 195, 189, 26 },
		{ 195, 190, 26 },
		{ 195, 191, 26 },
		{ 195, 192, 26 },
		{ 195, 193, 26 },
		{ 195, 194, 26 },
		{ 195, 195, 26 },
		{ 195, 196, 26 },
		{ 195, 197, 26 },
		{ 195, 198, 26 },
		{ 195, 199, 26 },
		{ 195, 200, 26 },
		{ 195, 201, 26 },
		{ 195, 202, 26 },
		{ 195, 203, 26 },
		{ 195, 204, 26 },
		{ 195, 205, 26 },
		{ 195, 206, 26 },
		{ 195, 207, 26 },
		{ 195, 208, 26 },
		{ 195, 209, 26 },
		{ 195, 210, 26 },
		{ 196, 176, 26 },
		{ 196, 177, 26 },
		{ 196, 178, 26 },
		{ 196, 179, 26 },
		{ 196, 180, 26 },
		{ 196, 181, 26 },
		{ 196, 182, 26 },
		{ 196, 183, 26 },
		{ 196, 184, 26 },
		{ 196, 185, 26 },
		{ 196, 186, 26 },
		{ 196, 187, 26 },
		{ 196, 188, 26 },
		{ 196, 189, 26 },
		{ 196, 190, 26 },
		{ 196, 191, 26 },
		{ 196, 192, 26 },
		{ 196, 193, 26 },
		{ 196, 194, 26 },
		{ 196, 195, 26 },
		{ 196, 196, 26 },
		{ 196, 197, 26 },
		{ 196, 198, 26 },
		{ 196, 199, 26 },
		{ 196, 200, 26 },
		{ 196, 201, 26 },
		{ 196, 202, 26 },
		{ 196, 203, 26 },
		{ 196, 204, 26 },
		{ 196, 205, 26 },
		{ 196, 206, 26 },
		{ 196, 207, 26 },
		{ 196, 208, 26 },
		{ 196, 209, 26 },
		{ 196, 210, 26 },
		{ 196, 211, 26 },
		{ 197, 178, 26 },
		{ 197, 179, 26 },
		{ 197, 180, 26 },
		{ 197, 181, 26 },
		{ 197, 182, 26 },
		{ 197, 183, 26 },
		{ 197, 184, 26 },
		{ 197, 185, 26 },
		{ 197, 186, 26 },
		{ 197, 187, 26 },
		{ 197, 188, 26 },
		{ 197, 189, 26 },
		{ 197, 190, 26 },
		{ 197, 191, 26 },
		{ 197, 192, 26 },
		{ 197, 193, 26 },
		{ 197, 194, 26 },
		{ 197, 195, 26 },
		{ 197, 196, 26 },
		{ 197, 197, 26 },
		{ 197, 198, 26 },
		{ 197, 199, 26 },
		{ 197, 200, 26 },
		{ 197, 201, 26 },
		{ 197, 202, 26 },
		{ 197, 203, 26 },
		{ 197, 204, 26 },
		{ 197, 205, 26 },
		{ 197, 206, 26 },
		{ 197, 207, 26 },
		{ 197, 208, 26 },
		{ 197, 209, 26 },
		{ 197, 210, 26 },
		{ 197, 211, 26 },
		{ 198, 179, 26 },
		{ 198, 180, 26 },
		{ 198, 181, 26 },
		{ 198, 182, 26 },
		{ 198, 183, 26 },
		{ 198, 184, 26 },
		{ 198, 185, 26 },
		{ 198, 186, 26 },
		{ 198, 187, 26 },
		{ 198, 188, 26 },
		{ 198, 189, 26 },
		{ 198, 190, 26 },
		{ 198, 191, 26 },
		{ 198, 192, 26 },
		{ 198, 193, 26 },
		{ 198, 194, 26 },
		{ 198, 195, 26 },
		{ 198, 196, 26 },
		{ 198, 197, 26 },
		{ 198, 198, 26 },
		{ 198, 199, 26 },
		{ 198, 200, 26 },
		{ 198, 201, 26 },
		{ 198, 202, 26 },
		{ 198, 203, 26 },
		{ 198, 204, 26 },
		{ 198, 205, 26 },
		{ 198, 206, 26 },
		{ 198, 207, 26 },
		{ 198, 208, 26 },
		{ 198, 209, 26 },
		{ 198, 210, 26 },
		{ 198, 211, 26 },
		{ 198, 212, 26 },
		{ 199, 180, 26 },
		{ 199, 181, 26 },
		{ 199, 182, 26 },
		{ 199, 183, 26 },
		{ 199, 184, 26 },
		{ 199, 185, 26 },
		{ 199, 186, 26 },
		{ 199, 187, 26 },
		{ 199, 188, 26 },
		{ 199, 189, 26 },
		{ 199, 190, 26 },
		{ 199, 191, 26 },
		{ 199, 192, 26 },
		{ 199, 193, 26 },
		{ 199, 194, 26 },
		{ 199, 195, 26 },
		{ 199, 196, 26 },
		{ 199, 197, 26 },
		{ 199, 198, 26 },
		{ 199, 199, 26 },
		{ 199, 200, 26 },
		{ 199, 201, 26 },
		{ 199, 202, 26 },
		{ 199, 203, 26 },
		{ 199, 204, 26 },
		{ 199, 205, 26 },
		{ 199, 206, 26 },
		{ 199, 207, 26 },
		{ 199, 208, 26 },
		{ 199, 209, 26 },
		{ 199, 210, 26 },
		{ 199, 211, 26 },
		{ 199, 212, 26 },
		{ 199, 213, 26 },
		{ 200, 182, 26 },
		{ 200, 183, 26 },
		{ 200, 184, 26 },
		{ 200, 185, 26 },
		{ 200, 186, 26 },
		{ 200, 187, 26 },
		{ 200, 188, 26 },
		{ 200, 189, 26 },
		{ 200, 190, 26 },
		{ 200, 191, 26 },
		{ 200, 192, 26 },
		{ 200, 193, 26 },
		{ 200, 194, 26 },
		{ 200, 195, 26 },
		{ 200, 196, 26 },
		{ 200, 197, 26 },
		{ 200, 198, 26 },
		{ 200, 199, 26 },
		{ 200, 200, 26 },
		{ 200, 201, 26 },
		{ 200, 202, 26 },
		{ 200, 203, 26 },
		{ 200, 204, 26 },
		{ 200, 205, 26 },
		{ 200, 206, 26 },
		{ 200, 207, 26 },
		{ 200, 208, 26 },
		{ 200, 209, 26 },
		{ 200, 210, 26 },
		{ 200, 211, 26 },
		{ 200, 212, 26 },
		{ 200, 213, 26 },
		{ 200, 214, 26 },
		{ 201, 183, 26 },
		{ 201, 184, 26 },
		{ 201, 185, 26 },
		{ 201, 186, 26 },
		{ 201, 187, 26 },
		{ 201, 188, 26 },
		{ 201, 189, 26 },
		{ 201, 190, 26 },
		{ 201, 191, 26 },
		{ 201, 192, 26 },
		{ 201, 193, 26 },
		{ 201, 194, 26 },
		{ 201, 195, 26 },
		{ 201, 196, 26 },
		{ 201, 197, 26 },
		{ 201, 198, 26 },
		{ 201, 199, 26 },
		{ 201, 200, 26 },
		{ 201, 201, 26 },
		{ 201, 202, 26 },
		{ 201, 203, 26 },
		{ 201, 204, 26 },
		{ 201, 205, 26 },
		{ 201, 206, 26 },
		{ 201, 207, 26 },
		{ 201, 208, 26 },
		{ 201, 209, 26 },
		{ 201, 210, 26 },
		{ 201, 211, 26 },
		{ 201, 212, 26 },
		{ 201, 213, 26 },
		{ 201, 214, 26 },
		{ 202, 184, 26 },
		{ 202, 185, 26 },
		{ 202, 186, 26 },
		{ 202, 187, 26 },
		{ 202, 188, 26 },
		{ 202, 189, 26 },
		{ 202, 190, 26 },
		{ 202, 191, 26 },
		{ 202, 192, 26 },
		{ 202, 193, 26 },
		{ 202, 194, 26 },
		{ 202, 195, 26 },
		{ 202, 196, 26 },
		{ 202, 197, 26 },
		{ 202, 198, 26 },
		{ 202, 199, 26 },
		{ 202, 200, 26 },
		{ 202, 201, 26 },
		{ 202, 202, 26 },
		{ 202, 203, 26 },
		{ 202, 204, 26 },
		{ 202, 205, 26 },
		{ 202, 206, 26 },
		{ 202, 207, 26 },
		{ 202, 208, 26 },
		{ 202, 209, 26 },
		{ 202, 210, 26 },
		{ 202, 211, 26 },
		{ 202, 212, 26 },
		{ 202, 213, 26 },
		{ 202, 214, 26 },
		{ 202, 215, 26 },
		{ 203, 186, 26 },
		{ 203, 187, 26 },
		{ 203, 188, 26 },
		{ 203, 189, 26 },
		{ 203, 190, 26 },
		{ 203, 191, 26 },
		{ 203, 192, 26 },
		{ 203, 193, 26 },
		{ 203, 194, 26 },
		{ 203, 195, 26 },
		{ 203, 196, 26 },
		{ 203, 197, 26 },
		{ 203, 198, 26 },
		{ 203, 199, 26 },
		{ 203, 200, 26 },
		{ 203, 201, 26 },
		{ 203, 202, 26 },
		{ 203, 203, 26 },
		{ 203, 204, 26 },
		{ 203, 205, 26 },
		{ 203, 206, 26 },
		{ 203, 207, 26 },
		{ 203, 208, 26 },
		{ 203, 209, 26 },
		{ 203, 210, 26 },
		{ 203, 211, 26 },
		{ 203, 212, 26 },
		{ 203, 213, 26 },
		{ 203, 214, 26 },
		{ 203, 215, 26 },
		{ 203, 216, 26 },
		{ 204, 187, 26 },
		{ 204, 188, 26 },
		{ 204, 189, 26 },
		{ 204, 190, 26 },
		{ 204, 191, 26 },
		{ 204, 192, 26 },
		{ 204, 193, 26 },
		{ 204, 194, 26 },
		{ 204, 195, 26 },
		{ 204, 196, 26 },
		{ 204, 197, 26 },
		{ 204, 198, 26 },
		{ 204, 199, 26 },
		{ 204, 200, 26 },
		{ 204, 201, 26 },
		{ 204, 202, 26 },
		{ 204, 203, 26 },
		{ 204, 204, 26 },
		{ 204, 205, 26 },
		{ 204, 206, 26 },
		{ 204, 207, 26 },
		{ 204, 208, 26 },
		{ 204, 209, 26 },
		{ 204, 210, 26 },
		{ 204, 211, 26 },
		{ 204, 212, 26 },
		{ 204, 213, 26 },
		{ 204, 214, 26 },
		{ 204, 215, 26 },
		{ 204, 216, 26 },
		{ 204, 217, 26 },
		{ 205, 188, 26 },
		{ 205, 189, 26 },
		{ 205, 190, 26 },
		{ 205, 191, 26 },
		{ 205, 192, 26 },
		{ 205, 193, 26 },
		{ 205, 194, 26 },
		{ 205, 195, 26 },
		{ 205, 196, 26 },
		{ 205, 197, 26 },
		{ 205, 198, 26 },
		{ 205, 199, 26 },
		{ 205, 200, 26 },
		{ 205, 201, 26 },
		{ 205, 202, 26 },
		{ 205, 203, 26 },
		{ 205, 204, 26 },
		{ 205, 205, 26 },
		{ 205, 206, 26 },
		{ 205, 207, 26 },
		{ 205, 208, 26 },
		{ 205, 209, 26 },
		{ 205, 210, 26 },
		{ 205, 211, 26 },
		{ 205, 212, 26 },
		{ 205, 213, 26 },
		{ 205, 214, 26 },
		{ 205, 215, 26 },
		{ 205, 216, 26 },
		{ 205, 217, 26 },
		{ 206, 190, 26 },
		{ 206, 191, 26 },
		{ 206, 192, 26 },
		{ 206, 193, 26 },
		{ 206, 194, 26 },
		{ 206, 195, 26 },
		{ 206, 196, 26 },
		{ 206, 197, 26 },
		{ 206, 198, 26 },
		{ 206, 199, 26 },
		{ 206, 200, 26 },
		{ 206, 201, 26 },
		{ 206, 202, 26 },
		{ 206, 203, 26 },
		{ 206, 204, 26 },
		{ 206, 205, 26 },
		{ 206, 206, 26 },
		{ 206, 207, 26 },
		{ 206, 208, 26 },
		{ 206, 209, 26 },
		{ 206, 210, 26 },
		{ 206, 211, 26 },
		{ 206, 212, 26 },
		{ 206, 213, 26 },
		{ 206, 214, 26 },
		{ 206, 215, 26 },
		{ 206, 216, 26 },
		{ 206, 217, 26 },
		{ 206, 218, 26 },
		{ 207, 191, 26 },
		{ 207, 192, 26 },
		{ 207, 193, 26 },
		{ 207, 194, 26 },
		{ 207, 195, 26 },
		{ 207, 196, 26 },
		{ 207, 197, 26 },
		{ 207, 198, 26 },
		{ 207, 199, 26 },
		{ 207, 200, 26 },
		{ 207, 201, 26 },
		{ 207, 202, 26 },
		{ 207, 203, 26 },
		{ 207, 204, 26 },
		{ 207, 205, 26 },
		{ 207, 206, 26 },
		{ 207, 207, 26 },
		{ 207, 208, 26 },
		{ 207, 209, 26 },
		{ 207, 210, 26 },
		{ 207, 211, 26 },
		{ 207, 212, 26 },
		{ 207, 213, 26 },
		{ 207, 214, 26 },
		{ 207, 215, 26 },
		{ 207, 216, 26 },
		{ 207, 217, 26 },
		{ 207, 218, 26 },
		{ 207, 219, 26 },
		{ 208, 192, 26 },
		{ 208, 193, 26 },
		{ 208, 194, 26 },
		{ 208, 195, 26 },
		{ 208, 196, 26 },
		{ 208, 197, 26 },
		{ 208, 198, 26 },
		{ 208, 199, 26 },
		{ 208, 200, 26 },
		{ 208, 201, 26 },
		{ 208, 202, 26 },
		{ 208, 203, 26 },
		{ 208, 204, 26 },
		{ 208, 205, 26 },
		{ 208, 206, 26 },
		{ 208, 207, 26 },
		{ 208, 208, 26 },
		{ 208, 209, 26 },
		{ 208, 210, 26 },
		{ 208, 211, 26 },
		{ 208, 212, 26 },
		{ 208, 213, 26 },
		{ 208, 214, 26 },
		{ 208, 215, 26 },
		{ 208, 216, 26 },
		{ 208, 217, 26 },
		{ 208, 218, 26 },
		{ 208, 219, 26 },
		{ 208, 220, 26 },
		{ 209, 194, 26 },
		{ 209, 195, 26 },
		{ 209, 196, 26 },
		{ 209, 197, 26 },
		{ 209, 198, 26 },
		{ 209, 199, 26 },
		{ 209, 200, 26 },
		{ 209, 201, 26 },
		{ 209, 202, 26 },
		{ 209, 203, 26 },
		{ 209, 204, 26 },
		{ 209, 205, 26 },
		{ 209, 206, 26 },
		{ 209, 207, 26 },
		{ 209, 208, 26 },
		{ 209, 209, 26 },
		{ 209, 210, 26 },
		{ 209, 211, 26 },
		{ 209, 212, 26 },
		{ 209, 213, 26 },
		{ 209, 214, 26 },
		{ 209, 215, 26 },
		{ 209, 216, 26 },
		{ 209, 217, 26 },
		{ 209, 218, 26 },
		{ 209, 219, 26 },
		{ 209, 220, 26 },
		{ 210, 195, 26 },
		{ 210, 196, 26 },
		{ 210, 197, 26 },
		{ 210, 198, 26 },
		{ 210, 199, 26 },
		{ 210, 200, 26 },
		{ 210, 201, 26 },
		{ 210, 202, 26 },
		{ 210, 203, 26 },
		{ 210, 204, 26 },
		{ 210, 205, 26 },
		{ 210, 206, 26 },
		{ 210, 207, 26 },
		{ 210, 208, 26 },
		{ 210, 209, 26 },
		{ 210, 210, 26 },
		{ 210, 211, 26 },
		{ 210, 212, 26 },
		{ 210, 213, 26 },
		{ 210, 214, 26 },
		{ 210, 215, 26 },
		{ 210, 216, 26 },
		{ 210, 217, 26 },
		{ 210, 218, 26 },
		{ 210, 219, 26 },
		{ 210, 220, 26 },
		{ 210, 221, 26 },
		{ 211, 196, 26 },
		{ 211, 197, 26 },
		{ 211, 198, 26 },
		{ 211, 199, 26 },
		{ 211, 200, 26 },
		{ 211, 201, 26 },
		{ 211, 202, 26 },
		{ 211, 203, 26 },
		{ 211, 204, 26 },
		{ 211, 205, 26 },
		{ 211, 206, 26 },
		{ 211, 207, 26 },
		{ 211, 208, 26 },
		{ 211, 209, 26 },
		{ 211, 210, 26 },
		{ 211, 211, 26 },
		{ 211, 212, 26 },
		{ 211, 213, 26 },
		{ 211, 214, 26 },
		{ 211, 215, 26 },
		{ 211, 216, 26 },
		{ 211, 217, 26 },
		{ 211, 218, 26 },
		{ 211, 219, 26 },
		{ 211, 220, 26 },
		{ 211, 221, 26 },
		{ 211, 222, 26 },
		{ 212, 198, 26 },
		{ 212, 199, 26 },
		{ 212, 200, 26 },
		{ 212, 201, 26 },
		{ 212, 202, 26 },
		{ 212, 203, 26 },
		{ 212, 204, 26 },
		{ 212, 205, 26 },
		{ 212, 206, 26 },
		{ 212, 207, 26 },
		{ 212, 208, 26 },
		{ 212, 209, 26 },
		{ 212, 210, 26 },
		{ 212, 211, 26 },
		{ 212, 212, 26 },
		{ 212, 213, 26 },
		{ 212, 214, 26 },
		{ 212, 215, 26 },
		{ 212, 216, 26 },
		{ 212, 217, 26 },
		{ 212, 218, 26 },
		{ 212, 219, 26 },
		{ 212, 220, 26 },
		{ 212, 221, 26 },
		{ 212, 222, 26 },
		{ 212, 223, 26 },
		{ 213, 199, 26 },
		{ 213, 200, 26 },
		{ 213, 201, 26 },
		{ 213, 202, 26 },
		{ 213, 203, 26 },
		{ 213, 204, 26 },
		{ 213, 205, 26 },
		{ 213, 206, 26 },
		{ 213, 207, 26 },
		{ 213, 208, 26 },
		{ 213, 209, 26 },
		{ 213, 210, 26 },
		{ 213, 211, 26 },
		{ 213, 212, 26 },
		{ 213, 213, 26 },
		{ 213, 214, 26 },
		{ 213, 215, 26 },
		{ 213, 216, 26 },
		{ 213, 217, 26 },
		{ 213, 218, 26 },
		{ 213, 219, 26 },
		{ 213, 220, 26 },
		{ 213, 221, 26 },
		{ 213, 222, 26 },
		{ 213, 223, 26 },
		{ 214, 200, 26 },
		{ 214, 201, 26 },
		{ 214, 202, 26 },
		{ 214, 203, 26 },
		{ 214, 204, 26 },
		{ 214, 205, 26 },
		{ 214, 206, 26 },
		{ 214, 207, 26 },
		{ 214, 208, 26 },
		{ 214, 209, 26 },
		{ 214, 210, 26 },
		{ 214, 211, 26 },
		{ 214, 212, 26 },
		{ 214, 213, 26 },
		{ 214, 214, 26 },
		{ 214, 215, 26 },
		{ 214, 216, 26 },
		{ 214, 217, 26 },
		{ 214, 218, 26 },
		{ 214, 219, 26 },
		{ 214, 220, 26 },
		{ 214, 221, 26 },
		{ 214, 222, 26 },
		{ 214, 223, 26 },
		{ 214, 224, 26 },
		{ 215, 202, 26 },
		{ 215, 203, 26 },
		{ 215, 204, 26 },
		{ 215, 205, 26 },
		{ 215, 206, 26 },
		{ 215, 207, 26 },
		{ 215, 208, 26 },
		{ 215, 209, 26 },
		{ 215, 210, 26 },
		{ 215, 211, 26 },
		{ 215, 212, 26 },
		{ 215, 213, 26 },
		{ 215, 214, 26 },
		{ 215, 215, 26 },
		{ 215, 216, 26 },
		{ 215, 217, 26 },
		{ 215, 218, 26 },
		{ 215, 219, 26 },
		{ 215, 220, 26 },
		{ 215, 221, 26 },
		{ 215, 222, 26 },
		{ 215, 223, 26 },
		{ 215, 224, 26 },
		{ 215, 225, 26 },
		{ 216, 203, 26 },
		{ 216, 204, 26 },
		{ 216, 205, 26 },
		{ 216, 206, 26 },
		{ 216, 207, 26 },
		{ 216, 208, 26 },
		{ 216, 209, 26 },
		{ 216, 210, 26 },
		{ 216, 211, 26 },
		{ 216, 212, 26 },
		{ 216, 213, 26 },
		{ 216, 214, 26 },
		{ 216, 215, 26 },
		{ 216, 216, 26 },
		{ 216, 217, 26 },
		{ 216, 218, 26 },
		{ 216, 219, 26 },
		{ 216, 220, 26 },
		{ 216, 221, 26 },
		{ 216, 222, 26 },
		{ 216, 223, 26 },
		{ 216, 224, 26 },
		{ 216, 225, 26 },
		{ 216, 226, 26 },
		{ 217, 204, 26 },
		{ 217, 205, 26 },
		{ 217, 206, 26 },
		{ 217, 207, 26 },
		{ 217, 208, 26 },
		{ 217, 209, 26 },
		{ 217, 210, 26 },
		{ 217, 211, 26 },
		{ 217, 212, 26 },
		{ 217, 213, 26 },
		{ 217, 214, 26 },
		{ 217, 215, 26 },
		{ 217, 216, 26 },
		{ 217, 217, 26 },
		{ 217, 218, 26 },
		{ 217, 219, 26 },
		{ 217, 220, 26 },
		{ 217, 221, 26 },
		{ 217, 222, 26 },
		{ 217, 223, 26 },
		{ 217, 224, 26 },
		{ 217, 225, 26 },
		{ 217, 226, 26 },
		{ 218, 206, 26 },
		{ 218, 207, 26 },
		{ 218, 208, 26 },
		{ 218, 209, 26 },
		{ 218, 210, 26 },
		{ 218, 211, 26 },
		{ 218, 212, 26 },
		{ 218, 213, 26 },
		{ 218, 214, 26 },
		{ 218, 215, 26 },
		{ 218, 216, 26 },
		{ 218, 217, 26 },
		{ 218, 218, 26 },
		{ 218, 219, 26 },
		{ 218, 220, 26 },
		{ 218, 221, 26 },
		{ 218, 222, 26 },
		{ 218, 223, 26 },
		{ 218, 224, 26 },
		{ 218, 225, 26 },
		{ 218, 226, 26 },
		{ 218, 227, 26 },
		{ 219, 207, 26 },
		{ 219, 208, 26 },
		{ 219, 209, 26 },
		{ 219, 210, 26 },
		{ 219, 211, 26 },
		{ 219, 212, 26 },
		{ 219, 213, 26 },
		{ 219, 214, 26 },
		{ 219, 215, 26 },
		{ 219, 216, 26 },
		{ 219, 217, 26 },
		{ 219, 218, 26 },
		{ 219, 219, 26 },
		{ 219, 220, 26 },
		{ 219, 221, 26 },
		{ 219, 222, 26 },
		{ 219, 223, 26 },
		{ 219, 224, 26 },
		{ 219, 225, 26 },
		{ 219, 226, 26 },
		{ 219, 227, 26 },
		{ 219, 228, 26 },
		{ 220, 208, 26 },
		{ 220, 209, 26 },
		{ 220, 210, 26 },
		{ 220, 211, 26 },
		{ 220, 212, 26 },
		{ 220, 213, 26 },
		{ 220, 214, 26 },
		{ 220, 215, 26 },
		{ 220, 216, 26 },
		{ 220, 217, 26 },
		{ 220, 218, 26 },
		{ 220, 219, 26 },
		{ 220, 220, 26 },
		{ 220, 221, 26 },
		{ 220, 222, 26 },
		{ 220, 223, 26 },
		{ 220, 224, 26 },
		{ 220, 225, 26 },
		{ 220, 226, 26 },
		{ 220, 227, 26 },
		{ 220, 228, 26 },
		{ 220, 229, 26 },
		{ 221, 210, 26 },
		{ 221, 211, 26 },
		{ 221, 212, 26 },
		{ 221, 213, 26 },
		{ 221, 214, 26 },
		{ 221, 215, 26 },
		{ 221, 216, 26 },
		{ 221, 217, 26 },
		{ 221, 218, 26 },
		{ 221, 219, 26 },
		{ 221, 220, 26 },
		{ 221, 221, 26 },
		{ 221, 222, 26 },
		{ 221, 223, 26 },
		{ 221, 224, 26 },
		{ 221, 225, 26 },
		{ 221, 226, 26 },
		{ 221, 227, 26 },
		{ 221, 228, 26 },
		{ 221, 229, 26 },
		{ 222, 211, 26 },
		{ 222, 212, 26 },
		{ 222, 213, 26 },
		{ 222, 214, 26 },
		{ 222, 215, 26 },
		{ 222, 216, 26 },
		{ 222, 217, 26 },
		{ 222, 218, 26 },
		{ 222, 219, 26 },
		{ 222, 220, 26 },
		{ 222, 221, 26 },
		{ 222, 222, 26 },
		{ 222, 223, 26 },
		{ 222, 224, 26 },
		{ 222, 225, 26 },
		{ 222, 226, 26 },
		{ 222, 227, 26 },
		{ 222, 228, 26 },
		{ 222, 229, 26 },
		{ 222, 230, 26 },
		{ 223, 212, 26 },
		{ 223, 213, 26 },
		{ 223, 214, 26 },
		{ 223, 215, 26 },
		{ 223, 216, 26 },
		{ 223, 217, 26 },
		{ 223, 218, 26 },
		{ 223, 219, 26 },
		{ 223, 220, 26 },
		{ 223, 221, 26 },
		{ 223, 222, 26 },
		{ 223, 223, 26 },
		{ 223, 224, 26 },
		{ 223, 225, 26 },
		{ 223, 226, 26 },
		{ 223, 227, 26 },
		{ 223, 228, 26 },
		{ 223, 229, 26 },
		{ 223, 230, 26 },
		{ 223, 231, 26 },
		{ 224, 214, 26 },
		{ 224, 215, 26 },
		{ 224, 216, 26 },
		{ 224, 217, 26 },
		{ 224, 218, 26 },
		{ 224, 219, 26 },
		{ 224, 220, 26 },
		{ 224, 221, 26 },
		{ 224, 222, 26 },
		{ 224, 223, 26 },
		{ 224, 224, 26 },
		{ 224, 225, 26 },
		{ 224, 226, 26 },
		{ 224, 227, 26 },
		{ 224, 228, 26 },
		{ 224, 229, 26 },
		{ 224, 230, 26 },
		{ 224, 231, 26 },
		{ 224, 232, 26 },
		{ 225, 215, 26 },
		{ 225, 216, 26 },
		{ 225, 217, 26 },
		{ 225, 218, 26 },
		{ 225, 219, 26 },
		{ 225, 220, 26 },
		{ 225, 221, 26 },
		{ 225, 222, 26 },
		{ 225, 223, 26 },
		{ 225, 224, 26 },
		{ 225, 225, 26 },
		{ 225, 226, 26 },
		{ 225, 227, 26 },
		{ 225, 228, 26 },
		{ 225, 229, 26 },
		{ 225, 230, 26 },
		{ 225, 231, 26 },
		{ 225, 232, 26 },
		{ 226, 216, 26 },
		{ 226, 217, 26 },
		{ 226, 218, 26 },
		{ 226, 219, 26 },
		{ 226, 220, 26 },
		{ 226, 221, 26 },
		{ 226, 222, 26 },
		{ 226, 223, 26 },
		{ 226, 224, 26 },
		{ 226, 225, 26 },
		{ 226, 226, 26 },
		{ 226, 227, 26 },
		{ 226, 228, 26 },
		{ 226, 229, 26 },
		{ 226, 230, 26 },
		{ 226, 231, 26 },
		{ 226, 232, 26 },
		{ 226, 233, 26 },
		{ 227, 218, 26 },
		{ 227, 219, 26 },
		{ 227, 220, 26 },
		{ 227, 221, 26 },
		{ 227, 222, 26 },
		{ 227, 223, 26 },
		{ 227, 224, 26 },
		{ 227, 225, 26 },
		{ 227, 226, 26 },
		{ 227, 227, 26 },
		{ 227, 228, 26 },
		{ 227, 229, 26 },
		{ 227, 230, 26 },
		{ 227, 231, 26 },
		{ 227, 232, 26 },
		{ 227, 233, 26 },
		{ 227, 234, 26 },
		{ 228, 219, 26 },
		{ 228, 220, 26 },
		{ 228, 221, 26 },
		{ 228, 222, 26 },
		{ 228, 223, 26 },
		{ 228, 224, 26 },
		{ 228, 225, 26 },
		{ 228, 226, 26 },
		{ 228, 227, 26 },
		{ 228, 228, 26 },
		{ 228, 229, 26 },
		{ 228, 230, 26 },
		{ 228, 231, 26 },
		{ 228, 232, 26 },
		{ 228, 233, 26 },
		{ 228, 234, 26 },
		{ 228, 235, 26 },
		{ 229, 220, 26 },
		{ 229, 221, 26 },
		{ 229, 222, 26 },
		{ 229, 223, 26 },
		{ 229, 224, 26 },
		{ 229, 225, 26 },
		{ 229, 226, 26 },
		{ 229, 227, 26 },
		{ 229, 228, 26 },
		{ 229, 229, 26 },
		{ 229, 230, 26 },
		{ 229, 231, 26 },
		{ 229, 232, 26 },
		{ 229, 233, 26 },
		{ 229, 234, 26 },
		{ 229, 235, 26 },
		{ 230, 222, 26 },
		{ 230, 223, 26 },
		{ 230, 224, 26 },
		{ 230, 225, 26 },
		{ 230, 226, 26 },
		{ 230, 227, 26 },
		{ 230, 228, 26 },
		{ 230, 229, 26 },
		{ 230, 230, 26 },
		{ 230, 231, 26 },
		{ 230, 232, 26 },
		{ 230, 233, 26 },
		{ 230, 234, 26 },
		{ 230, 235, 26 },
		{ 230, 236, 26 },
		{ 231, 223, 26 },
		{ 231, 224, 26 },
		{ 231, 225, 26 },
		{ 231, 226, 26 },
		{ 231, 227, 26 },
		{ 231, 228, 26 },
		{ 231, 229, 26 },
		{ 231, 230, 26 },
		{ 231, 231, 26 },
		{ 231, 232, 26 },
		{ 231, 233, 26 },
		{ 231, 234, 26 },
		{ 231, 235, 26 },
		{ 231, 236, 26 },
		{ 231, 237, 26 },
		{ 232, 224, 26 },
		{ 232, 225, 26 },
		{ 232, 226, 26 },
		{ 232, 227, 26 },
		{ 232, 228, 26 },
		{ 232, 229, 26 },
		{ 232, 230, 26 },
		{ 232, 231, 26 },
		{ 232, 232, 26 },
		{ 232, 233, 26 },
		{ 232, 234, 26 },
		{ 232, 235, 26 },
		{ 232, 236, 26 },
		{ 232, 237, 26 },
		{ 232, 238, 26 },
		{ 233, 226, 26 },
		{ 233, 227, 26 },
		{ 233, 228, 26 },
		{ 233, 229, 26 },
		{ 233, 230, 26 },
		{ 233, 231, 26 },
		{ 233, 232, 26 },
		{ 233, 233, 26 },
		{ 233, 234, 26 },
		{ 233, 235, 26 },
		{ 233, 236, 26 },
		{ 233, 237, 26 },
		{ 233, 238, 26 },
		{ 234, 227, 26 },
		{ 234, 228, 26 },
		{ 234, 229, 26 },
		{ 234, 230, 26 },
		{ 234, 231, 26 },
		{ 234, 232, 26 },
		{ 234, 233, 26 },
		{ 234, 234, 26 },
		{ 234, 235, 26 },
		{ 234, 236, 26 },
		{ 234, 237, 26 },
		{ 234, 238, 26 },
		{ 234, 239, 26 },
		{ 235, 228, 26 },
		{ 235, 229, 26 },
		{ 235, 230, 26 },
		{ 235, 231, 26 },
		{ 235, 232, 26 },
		{ 235, 233, 26 },
		{ 235, 234, 26 },
		{ 235, 235, 26 },
		{ 235, 236, 26 },
		{ 235, 237, 26 },
		{ 235, 238, 26 },
		{ 235, 239, 26 },
		{ 235, 240, 26 },
		{ 236, 230, 26 },
		{ 236, 231, 26 },
		{ 236, 232, 26 },
		{ 236, 233, 26 },
		{ 236, 234, 26 },
		{ 236, 235, 26 },
		{ 236, 236, 26 },
		{ 236, 237, 26 },
		{ 236, 238, 26 },
		{ 236, 239, 26 },
		{ 236, 240, 26 },
		{ 236, 241, 26 },
		{ 237, 231, 26 },
		{ 237, 232, 26 },
		{ 237, 233, 26 },
		{ 237, 234, 26 },
		{ 237, 235, 26 },
		{ 237, 236, 26 },
		{ 237, 237, 26 },
		{ 237, 238, 26 },
		{ 237, 239, 26 },
		{ 237, 240, 26 },
		{ 237, 241, 26 },
		{ 238, 232, 26 },
		{ 238, 233, 26 },
		{ 238, 234, 26 },
		{ 238, 235, 26 },
		{ 238, 236, 26 },
		{ 238, 237, 26 },
		{ 238, 238, 26 },
		{ 238, 239, 26 },
		{ 238, 240, 26 },
		{ 238, 241, 26 },
		{ 238, 242, 26 },
		{ 239, 234, 26 },
		{ 239, 235, 26 },
		{ 239, 236, 26 },
		{ 239, 237, 26 },
		{ 239, 238, 26 },
		{ 239, 239, 26 },
		{ 239, 240, 26 },
		{ 239, 241, 26 },
		{ 239, 242, 26 },
		{ 239, 243, 26 },
		{ 240, 235, 26 },
		{ 240, 236, 26 },
		{ 240, 237, 26 },
		{ 240, 238, 26 },
		{ 240, 239, 26 },
		{ 240, 240, 26 },
		{ 240, 241, 26 },
		{ 240, 242, 26 },
		{ 240, 243, 26 },
		{ 240, 244, 26 },
		{ 241, 236, 26 },
		{ 241, 237, 26 },
		{ 241, 238, 26 },
		{ 241, 239, 26 },
		{ 241, 240, 26 },
		{ 242, 238, 26 },
		{ 242, 239, 26 },
		{ 242, 240, 26 },
		{ 243, 239, 26 },
		{ 243, 240, 26 },
		{ 244, 240, 26 },
		{ 270, 274, 26 },
		{ 271, 274, 26 },
		{ 271, 275, 26 },
		{ 272, 274, 26 },
		{ 272, 275, 26 },
		{ 272, 276, 26 },
		{ 273, 274, 26 },
		{ 273, 275, 26 },
		{ 273, 276, 26 },
		{ 273, 277, 26 },
		{ 273, 278, 26 },
		{ 274, 270, 26 },
		{ 274, 271, 26 },
		{ 274, 272, 26 },
		{ 274, 273, 26 },
		{ 274, 274, 26 },
		{ 274, 275, 26 },
		{ 274, 276, 26 },
		{ 274, 277, 26 },
		{ 274, 278, 26 },
		{ 274, 279, 26 },
		{ 275, 271, 26 },
		{ 275, 272, 26 },
		{ 275, 273, 26 },
		{ 275, 274, 26 },
		{ 275, 275, 26 },
		{ 275, 276, 26 },
		{ 275, 277, 26 },
		{ 275, 278, 26 },
		{ 275, 279, 26 },
		{ 275, 280, 26 },
		{ 276, 272, 26 },
		{ 276, 273, 26 },
		{ 276, 274, 26 },
		{ 276, 275, 26 },
		{ 276, 276, 26 },
		{ 276, 277, 26 },
		{ 276, 278, 26 },
		{ 276, 279, 26 },
		{ 276, 280, 26 },
		{ 276, 281, 26 },
		{ 276, 282, 26 },
		{ 277, 273, 26 },
		{ 277, 274, 26 },
		{ 277, 275, 26 },
		{ 277, 276, 26 },
		{ 277, 277, 26 },
		{ 277, 278, 26 },
		{ 277, 279, 26 },
		{ 277, 280, 26 },
		{ 277, 281, 26 },
		{ 277, 282, 26 },
		{ 277, 283, 26 },
		{ 278, 273, 26 },
		{ 278, 274, 26 },
		{ 278, 275, 26 },
		{ 278, 276, 26 },
		{ 278, 277, 26 },
		{ 278, 278, 26 },
		{ 278, 279, 26 },
		{ 278, 280, 26 },
		{ 278, 281, 26 },
		{ 278, 282, 26 },
		{ 278, 283, 26 },
		{ 278, 284, 26 },
		{ 279, 274, 26 },
		{ 279, 275, 26 },
		{ 279, 276, 26 },
		{ 279, 277, 26 },
		{ 279, 278, 26 },
		{ 279, 279, 26 },
		{ 279, 280, 26 },
		{ 279, 281, 26 },
		{ 279, 282, 26 },
		{ 279, 283, 26 },
		{ 279, 284, 26 },
		{ 279, 285, 26 },
		{ 279, 286, 26 },
		{ 280, 275, 26 },
		{ 280, 276, 26 },
		{ 280, 277, 26 },
		{ 280, 278, 26 },
		{ 280, 279, 26 },
		{ 280, 280, 26 },
		{ 280, 281, 26 },
		{ 280, 282, 26 },
		{ 280, 283, 26 },
		{ 280, 284, 26 },
		{ 280, 285, 26 },
		{ 280, 286, 26 },
		{ 280, 287, 26 },
		{ 281, 276, 26 },
		{ 281, 277, 26 },
		{ 281, 278, 26 },
		{ 281, 279, 26 },
		{ 281, 280, 26 },
		{ 281, 281, 26 },
		{ 281, 282, 26 },
		{ 281, 283, 26 },
		{ 281, 284, 26 },
		{ 281, 285, 26 },
		{ 281, 286, 26 },
		{ 281, 287, 26 },
		{ 281, 288, 26 },
		{ 282, 276, 26 },
		{ 282, 277, 26 },
		{ 282, 278, 26 },
		{ 282, 279, 26 },
		{ 282, 280, 26 },
		{ 282, 281, 26 },
		{ 282, 282, 26 },
		{ 282, 283, 26 },
		{ 282, 284, 26 },
		{ 282, 285, 26 },
		{ 282, 286, 26 },
		{ 282, 287, 26 },
		{ 282, 288, 26 },
		{ 282, 289, 26 },
		{ 282, 290, 26 },
		{ 283, 277, 26 },
		{ 283, 278, 26 },
		{ 283, 279, 26 },
		{ 283, 280, 26 },
		{ 283, 281, 26 },
		{ 283, 282, 26 },
		{ 283, 283, 26 },
		{ 283, 284, 26 },
		{ 283, 285, 26 },
		{ 283, 286, 26 },
		{ 283, 287, 26 },
		{ 283, 288, 26 },
		{ 283, 289, 26 },
		{ 283, 290, 26 },
		{ 283, 291, 26 },
		{ 284, 278, 26 },
		{ 284, 279, 26 },
		{ 284, 280, 26 },
		{ 284, 281, 26 },
		{ 284, 282, 26 },
		{ 284, 283, 26 },
		{ 284, 284, 26 },
		{ 284, 285, 26 },
		{ 284, 286, 26 },
		{ 284, 287, 26 },
		{ 284, 288, 26 },
		{ 284, 289, 26 },
		{ 284, 290, 26 },
		{ 284, 291, 26 },
		{ 284, 292, 26 },
		{ 285, 279, 26 },
		{ 285, 280, 26 },
		{ 285, 281, 26 },
		{ 285, 282, 26 },
		{ 285, 283, 26 },
		{ 285, 284, 26 },
		{ 285, 285, 26 },
		{ 285, 286, 26 },
		{ 285, 287, 26 },
		{ 285, 288, 26 },
		{ 285, 289, 26 },
		{ 285, 290, 26 },
		{ 285, 291, 26 },
		{ 285, 292, 26 },
		{ 285, 293, 26 },
		{ 285, 294, 26 },
		{ 286, 279, 26 },
		{ 286, 280, 26 },
		{ 286, 281, 26 },
		{ 286, 282, 26 },
		{ 286, 283, 26 },
		{ 286, 284, 26 },
		{ 286, 285, 26 },
		{ 286, 286, 26 },
		{ 286, 287, 26 },
		{ 286, 288, 26 },
		{ 286, 289, 26 },
		{ 286, 290, 26 },
		{ 286, 291, 26 },
		{ 286, 292, 26 },
		{ 286, 293, 26 },
		{ 286, 294, 26 },
		{ 286, 295, 26 },
		{ 287, 280, 26 },
		{ 287, 281, 26 },
		{ 287, 282, 26 },
		{ 287, 283, 26 },
		{ 287, 284, 26 },
		{ 287, 285, 26 },
		{ 287, 286, 26 },
		{ 287, 287, 26 },
		{ 287, 288, 26 },
		{ 287, 289, 26 },
		{ 287, 290, 26 },
		{ 287, 291, 26 },
		{ 287, 292, 26 },
		{ 287, 293, 26 },
		{ 287, 294, 26 },
		{ 287, 295, 26 },
		{ 287, 296, 26 },
		{ 288, 281, 26 },
		{ 288, 282, 26 },
		{ 288, 283, 26 },
		{ 288, 284, 26 },
		{ 288, 285, 26 },
		{ 288, 286, 26 },
		{ 288, 287, 26 },
		{ 288, 288, 26 },
		{ 288, 289, 26 },
		{ 288, 290, 26 },
		{ 288, 291, 26 },
		{ 288, 292, 26 },
		{ 288, 293, 26 },
		{ 288, 294, 26 },
		{ 288, 295, 26 },
		{ 288, 296, 26 },
		{ 288, 297, 26 },
		{ 288, 298, 26 },
		{ 289, 282, 26 },
		{ 289, 283, 26 },
		{ 289, 284, 26 },
		{ 289, 285, 26 },
		{ 289, 286, 26 },
		{ 289, 287, 26 },
		{ 289, 288, 26 },
		{ 289, 289, 26 },
		{ 289, 290, 26 },
		{ 289, 291, 26 },
		{ 289, 292, 26 },
		{ 289, 293, 26 },
		{ 289, 294, 26 },
		{ 289, 295, 26 },
		{ 289, 296, 26 },
		{ 289, 297, 26 },
		{ 289, 298, 26 },
		{ 289, 299, 26 },
		{ 290, 282, 26 },
		{ 290, 283, 26 },
		{ 290, 284, 26 },
		{ 290, 285, 26 },
		{ 290, 286, 26 },
		{ 290, 287, 26 },
		{ 290, 288, 26 },
		{ 290, 289, 26 },
		{ 290, 290, 26 },
		{ 290, 291, 26 },
		{ 290, 292, 26 },
		{ 290, 293, 26 },
		{ 290, 294, 26 },
		{ 290, 295, 26 },
		{ 290, 296, 26 },
		{ 290, 297, 26 },
		{ 290, 298, 26 },
		{ 290, 299, 26 },
		{ 290, 300, 26 },
		{ 291, 283, 26 },
		{ 291, 284, 26 },
		{ 291, 285, 26 },
		{ 291, 286, 26 },
		{ 291, 287, 26 },
		{ 291, 288, 26 },
		{ 291, 289, 26 },
		{ 291, 290, 26 },
		{ 291, 291, 26 },
		{ 291, 292, 26 },
		{ 291, 293, 26 },
		{ 291, 294, 26 },
		{ 291, 295, 26 },
		{ 291, 296, 26 },
		{ 291, 297, 26 },
		{ 291, 298, 26 },
		{ 291, 299, 26 },
		{ 291, 300, 26 },
		{ 291, 301, 26 },
		{ 291, 302, 26 },
		{ 292, 284, 26 },
		{ 292, 285, 26 },
		{ 292, 286, 26 },
		{ 292, 287, 26 },
		{ 292, 288, 26 },
		{ 292, 289, 26 },
		{ 292, 290, 26 },
		{ 292, 291, 26 },
		{ 292, 292, 26 },
		{ 292, 293, 26 },
		{ 292, 294, 26 },
		{ 292, 295, 26 },
		{ 292, 296, 26 },
		{ 292, 297, 26 },
		{ 292, 298, 26 },
		{ 292, 299, 26 },
		{ 292, 300, 26 },
		{ 292, 301, 26 },
		{ 292, 302, 26 },
		{ 292, 303, 26 },
		{ 293, 285, 26 },
		{ 293, 286, 26 },
		{ 293, 287, 26 },
		{ 293, 288, 26 },
		{ 293, 289, 26 },
		{ 293, 290, 26 },
		{ 293, 291, 26 },
		{ 293, 292, 26 },
		{ 293, 293, 26 },
		{ 293, 294, 26 },
		{ 293, 295, 26 },
		{ 293, 296, 26 },
		{ 293, 297, 26 },
		{ 293, 298, 26 },
		{ 293, 299, 26 },
		{ 293, 300, 26 },
		{ 293, 301, 26 },
		{ 293, 302, 26 },
		{ 293, 303, 26 },
		{ 293, 304, 26 },
		{ 294, 285, 26 },
		{ 294, 286, 26 },
		{ 294, 287, 26 },
		{ 294, 288, 26 },
		{ 294, 289, 26 },
		{ 294, 290, 26 },
		{ 294, 291, 26 },
		{ 294, 292, 26 },
		{ 294, 293, 26 },
		{ 294, 294, 26 },
		{ 294, 295, 26 },
		{ 294, 296, 26 },
		{ 294, 297, 26 },
		{ 294, 298, 26 },
		{ 294, 299, 26 },
		{ 294, 300, 26 },
		{ 294, 301, 26 },
		{ 294, 302, 26 },
		{ 294, 303, 26 },
		{ 294, 304, 26 },
		{ 294, 305, 26 },
		{ 294, 306, 26 },
		{ 295, 286, 26 },
		{ 295, 287, 26 },
		{ 295, 288, 26 },
		{ 295, 289, 26 },
		{ 295, 290, 26 },
		{ 295, 291, 26 },
		{ 295, 292, 26 },
		{ 295, 293, 26 },
		{ 295, 294, 26 },
		{ 295, 295, 26 },
		{ 295, 296, 26 },
		{ 295, 297, 26 },
		{ 295, 298, 26 },
		{ 295, 299, 26 },
		{ 295, 300, 26 },
		{ 295, 301, 26 },
		{ 295, 302, 26 },
		{ 295, 303, 26 },
		{ 295, 304, 26 },
		{ 295, 305, 26 },
		{ 295, 306, 26 },
		{ 295, 307, 26 },
		{ 296, 287, 26 },
		{ 296, 288, 26 },
		{ 296, 289, 26 },
		{ 296, 290, 26 },
		{ 296, 291, 26 },
		{ 296, 292, 26 },
		{ 296, 293, 26 },
		{ 296, 294, 26 },
		{ 296, 295, 26 },
		{ 296, 296, 26 },
		{ 296, 297, 26 },
		{ 296, 298, 26 },
		{ 296, 299, 26 },
		{ 296, 300, 26 },
		{ 296, 301, 26 },
		{ 296, 302, 26 },
		{ 296, 303, 26 },
		{ 296, 304, 26 },
		{ 296, 305, 26 },
		{ 296, 306, 26 },
		{ 296, 307, 26 },
		{ 296, 308, 26 },
		{ 297, 288, 26 },
		{ 297, 289, 26 },
		{ 297, 290, 26 },
		{ 297, 291, 26 },
		{ 297, 292, 26 },
		{ 297, 293, 26 },
		{ 297, 294, 26 },
		{ 297, 295, 26 },
		{ 297, 296, 26 },
		{ 297, 297, 26 },
		{ 297, 298, 26 },
		{ 297, 299, 26 },
		{ 297, 300, 26 },
		{ 297, 301, 26 },
		{ 297, 302, 26 },
		{ 297, 303, 26 },
		{ 297, 304, 26 },
		{ 297, 305, 26 },
		{ 297, 306, 26 },
		{ 297, 307, 26 },
		{ 297, 308, 26 },
		{ 297, 309, 26 },
		{ 297, 310, 26 },
		{ 298, 288, 26 },
		{ 298, 289, 26 },
		{ 298, 290, 26 },
		{ 298, 291, 26 },
		{ 298, 292, 26 },
		{ 298, 293, 26 },
		{ 298, 294, 26 },
		{ 298, 295, 26 },
		{ 298, 296, 26 },
		{ 298, 297, 26 },
		{ 298, 298, 26 },
		{ 298, 299, 26 },
		{ 298, 300, 26 },
		{ 298, 301, 26 },
		{ 298, 302, 26 },
		{ 298, 303, 26 },
		{ 298, 304, 26 },
		{ 298, 305, 26 },
		{ 298, 306, 26 },
		{ 298, 307, 26 },
		{ 298, 308, 26 },
		{ 298, 309, 26 },
		{ 298, 310, 26 },
		{ 298, 311, 26 },
		{ 299, 289, 26 },
		{ 299, 290, 26 },
		{ 299, 291, 26 },
		{ 299, 292, 26 },
		{ 299, 293, 26 },
		{ 299, 294, 26 },
		{ 299, 295, 26 },
		{ 299, 296, 26 },
		{ 299, 297, 26 },
		{ 299, 298, 26 },
		{ 299, 299, 26 },
		{ 299, 300, 26 },
		{ 299, 301, 26 },
		{ 299, 302, 26 },
		{ 299, 303, 26 },
		{ 299, 304, 26 },
		{ 299, 305, 26 },
		{ 299, 306, 26 },
		{ 299, 307, 26 },
		{ 299, 308, 26 },
		{ 299, 309, 26 },
		{ 299, 310, 26 },
		{ 299, 311, 26 },
		{ 299, 312, 26 },
		{ 300, 290, 26 },
		{ 300, 291, 26 },
		{ 300, 292, 26 },
		{ 300, 293, 26 },
		{ 300, 294, 26 },
		{ 300, 295, 26 },
		{ 300, 296, 26 },
		{ 300, 297, 26 },
		{ 300, 298, 26 },
		{ 300, 299, 26 },
		{ 300, 300, 26 },
		{ 300, 301, 26 },
		{ 300, 302, 26 },
		{ 300, 303, 26 },
		{ 300, 304, 26 },
		{ 300, 305, 26 },
		{ 300, 306, 26 },
		{ 300, 307, 26 },
		{ 300, 308, 26 },
		{ 300, 309, 26 },
		{ 300, 310, 26 },
		{ 300, 311, 26 },
		{ 300, 312, 26 },
		{ 300, 313, 26 },
		{ 300, 314, 26 },
		{ 301, 291, 26 },
		{ 301, 292, 26 },
		{ 301, 293, 26 },
		{ 301, 294, 26 },
		{ 301, 295, 26 },
		{ 301, 296, 26 },
		{ 301, 297, 26 },
		{ 301, 298, 26 },
		{ 301, 299, 26 },
		{ 301, 300, 26 },
		{ 301, 301, 26 },
		{ 301, 302, 26 },
		{ 301, 303, 26 },
		{ 301, 304, 26 },
		{ 301, 305, 26 },
		{ 301, 306, 26 },
		{ 301, 307, 26 },
		{ 301, 308, 26 },
		{ 301, 309, 26 },
		{ 301, 310, 26 },
		{ 301, 311, 26 },
		{ 301, 312, 26 },
		{ 301, 313, 26 },
		{ 301, 314, 26 },
		{ 301, 315, 26 },
		{ 302, 291, 26 },
		{ 302, 292, 26 },
		{ 302, 293, 26 },
		{ 302, 294, 26 },
		{ 302, 295, 26 },
		{ 302, 296, 26 },
		{ 302, 297, 26 },
		{ 302, 298, 26 },
		{ 302, 299, 26 },
		{ 302, 300, 26 },
		{ 302, 301, 26 },
		{ 302, 302, 26 },
		{ 302, 303, 26 },
		{ 302, 304, 26 },
		{ 302, 305, 26 },
		{ 302, 306, 26 },
		{ 302, 307, 26 },
		{ 302, 308, 26 },
		{ 302, 309, 26 },
		{ 302, 310, 26 },
		{ 302, 311, 26 },
		{ 302, 312, 26 },
		{ 302, 313, 26 },
		{ 302, 314, 26 },
		{ 302, 315, 26 },
		{ 302, 316, 26 },
		{ 303, 292, 26 },
		{ 303, 293, 26 },
		{ 303, 294, 26 },
		{ 303, 295, 26 },
		{ 303, 296, 26 },
		{ 303, 297, 26 },
		{ 303, 298, 26 },
		{ 303, 299, 26 },
		{ 303, 300, 26 },
		{ 303, 301, 26 },
		{ 303, 302, 26 },
		{ 303, 303, 26 },
		{ 303, 304, 26 },
		{ 303, 305, 26 },
		{ 303, 306, 26 },
		{ 303, 307, 26 },
		{ 303, 308, 26 },
		{ 303, 309, 26 },
		{ 303, 310, 26 },
		{ 303, 311, 26 },
		{ 303, 312, 26 },
		{ 303, 313, 26 },
		{ 303, 314, 26 },
		{ 303, 315, 26 },
		{ 303, 316, 26 },
		{ 303, 317, 26 },
		{ 303, 318, 26 },
		{ 304, 293, 26 },
		{ 304, 294, 26 },
		{ 304, 295, 26 },
		{ 304, 296, 26 },
		{ 304, 297, 26 },
		{ 304, 298, 26 },
		{ 304, 299, 26 },
		{ 304, 300, 26 },
		{ 304, 301, 26 },
		{ 304, 302, 26 },
		{ 304, 303, 26 },
		{ 304, 304, 26 },
		{ 304, 305, 26 },
		{ 304, 306, 26 },
		{ 304, 307, 26 },
		{ 304, 308, 26 },
		{ 304, 309, 26 },
		{ 304, 310, 26 },
		{ 304, 311, 26 },
		{ 304, 312, 26 },
		{ 304, 313, 26 },
		{ 304, 314, 26 },
		{ 304, 315, 26 },
		{ 304, 316, 26 },
		{ 304, 317, 26 },
		{ 304, 318, 26 },
		{ 304, 319, 26 },
		{ 305, 294, 26 },
		{ 305, 295, 26 },
		{ 305, 296, 26 },
		{ 305, 297, 26 },
		{ 305, 298, 26 },
		{ 305, 299, 26 },
		{ 305, 300, 26 },
		{ 305, 301, 26 },
		{ 305, 302, 26 },
		{ 305, 303, 26 },
		{ 305, 304, 26 },
		{ 305, 305, 26 },
		{ 305, 306, 26 },
		{ 305, 307, 26 },
		{ 305, 308, 26 },
		{ 305, 309, 26 },
		{ 305, 310, 26 },
		{ 305, 311, 26 },
		{ 305, 312, 26 },
		{ 305, 313, 26 },
		{ 305, 314, 26 },
		{ 305, 315, 26 },
		{ 305, 316, 26 },
		{ 305, 317, 26 },
		{ 305, 318, 26 },
		{ 305, 319, 26 },
		{ 305, 320, 26 },
		{ 306, 294, 26 },
		{ 306, 295, 26 },
		{ 306, 296, 26 },
		{ 306, 297, 26 },
		{ 306, 298, 26 },
		{ 306, 299, 26 },
		{ 306, 300, 26 },
		{ 306, 301, 26 },
		{ 306, 302, 26 },
		{ 306, 303, 26 },
		{ 306, 304, 26 },
		{ 306, 305, 26 },
		{ 306, 306, 26 },
		{ 306, 307, 26 },
		{ 306, 308, 26 },
		{ 306, 309, 26 },
		{ 306, 310, 26 },
		{ 306, 311, 26 },
		{ 306, 312, 26 },
		{ 306, 313, 26 },
		{ 306, 314, 26 },
		{ 306, 315, 26 },
		{ 306, 316, 26 },
		{ 306, 317, 26 },
		{ 306, 318, 26 },
		{ 306, 319, 26 },
		{ 306, 320, 26 },
		{ 306, 321, 26 },
		{ 306, 322, 26 },
		{ 307, 295, 26 },
		{ 307, 296, 26 },
		{ 307, 297, 26 },
		{ 307, 298, 26 },
		{ 307, 299, 26 },
		{ 307, 300, 26 },
		{ 307, 301, 26 },
		{ 307, 302, 26 },
		{ 307, 303, 26 },
		{ 307, 304, 26 },
		{ 307, 305, 26 },
		{ 307, 306, 26 },
		{ 307, 307, 26 },
		{ 307, 308, 26 },
		{ 307, 309, 26 },
		{ 307, 310, 26 },
		{ 307, 311, 26 },
		{ 307, 312, 26 },
		{ 307, 313, 26 },
		{ 307, 314, 26 },
		{ 307, 315, 26 },
		{ 307, 316, 26 },
		{ 307, 317, 26 },
		{ 307, 318, 26 },
		{ 307, 319, 26 },
		{ 307, 320, 26 },
		{ 307, 321, 26 },
		{ 307, 322, 26 },
		{ 307, 323, 26 },
		{ 308, 296, 26 },
		{ 308, 297, 26 },
		{ 308, 298, 26 },
		{ 308, 299, 26 },
		{ 308, 300, 26 },
		{ 308, 301, 26 },
		{ 308, 302, 26 },
		{ 308, 303, 26 },
		{ 308, 304, 26 },
		{ 308, 305, 26 },
		{ 308, 306, 26 },
		{ 308, 307, 26 },
		{ 308, 308, 26 },
		{ 308, 309, 26 },
		{ 308, 310, 26 },
		{ 308, 311, 26 },
		{ 308, 312, 26 },
		{ 308, 313, 26 },
		{ 308, 314, 26 },
		{ 308, 315, 26 },
		{ 308, 316, 26 },
		{ 308, 317, 26 },
		{ 308, 318, 26 },
		{ 308, 319, 26 },
		{ 308, 320, 26 },
		{ 308, 321, 26 },
		{ 308, 322, 26 },
		{ 308, 323, 26 },
		{ 308, 324, 26 },
		{ 309, 297, 26 },
		{ 309, 298, 26 },
		{ 309, 299, 26 },
		{ 309, 300, 26 },
		{ 309, 301, 26 },
		{ 309, 302, 26 },
		{ 309, 303, 26 },
		{ 309, 304, 26 },
		{ 309, 305, 26 },
		{ 309, 306, 26 },
		{ 309, 307, 26 },
		{ 309, 308, 26 },
		{ 309, 309, 26 },
		{ 309, 310, 26 },
		{ 309, 311, 26 },
		{ 309, 312, 26 },
		{ 309, 313, 26 },
		{ 309, 314, 26 },
		{ 309, 315, 26 },
		{ 309, 316, 26 },
		{ 309, 317, 26 },
		{ 309, 318, 26 },
		{ 309, 319, 26 },
		{ 309, 320, 26 },
		{ 309, 321, 26 },
		{ 309, 322, 26 },
		{ 309, 323, 26 },
		{ 309, 324, 26 },
		{ 309, 325, 26 },
		{ 309, 326, 26 },
		{ 310, 297, 26 },
		{ 310, 298, 26 },
		{ 310, 299, 26 },
		{ 310, 300, 26 },
		{ 310, 301, 26 },
		{ 310, 302, 26 },
		{ 310, 303, 26 },
		{ 310, 304, 26 },
		{ 310, 305, 26 },
		{ 310, 306, 26 },
		{ 310, 307, 26 },
		{ 310, 308, 26 },
		{ 310, 309, 26 },
		{ 310, 310, 26 },
		{ 310, 311, 26 },
		{ 310, 312, 26 },
		{ 310, 313, 26 },
		{ 310, 314, 26 },
		{ 310, 315, 26 },
		{ 310, 316, 26 },
		{ 310, 317, 26 },
		{ 310, 318, 26 },
		{ 310, 319, 26 },
		{ 310, 320, 26 },
		{ 310, 321, 26 },
		{ 310, 322, 26 },
		{ 310, 323, 26 },
		{ 310, 324, 26 },
		{ 310, 325, 26 },
		{ 310, 326, 26 },
		{ 310, 327, 26 },
		{ 311, 298, 26 },
		{ 311, 299, 26 },
		{ 311, 300, 26 },
		{ 311, 301, 26 },
		{ 311, 302, 26 },
		{ 311, 303, 26 },
		{ 311, 304, 26 },
		{ 311, 305, 26 },
		{ 311, 306, 26 },
		{ 311, 307, 26 },
		{ 311, 308, 26 },
		{ 311, 309, 26 },
		{ 311, 310, 26 },
		{ 311, 311, 26 },
		{ 311, 312, 26 },
		{ 311, 313, 26 },
		{ 311, 314, 26 },
		{ 311, 315, 26 },
		{ 311, 316, 26 },
		{ 311, 317, 26 },
		{ 311, 318, 26 },
		{ 311, 319, 26 },
		{ 311, 320, 26 },
		{ 311, 321, 26 },
		{ 311, 322, 26 },
		{ 311, 323, 26 },
		{ 311, 324, 26 },
		{ 311, 325, 26 },
		{ 311, 326, 26 },
		{ 311, 327, 26 },
		{ 311, 328, 26 },
		{ 312, 299, 26 },
		{ 312, 300, 26 },
		{ 312, 301, 26 },
		{ 312, 302, 26 },
		{ 312, 303, 26 },
		{ 312, 304, 26 },
		{ 312, 305, 26 },
		{ 312, 306, 26 },
		{ 312, 307, 26 },
		{ 312, 308, 26 },
		{ 312, 309, 26 },
		{ 312, 310, 26 },
		{ 312, 311, 26 },
		{ 312, 312, 26 },
		{ 312, 313, 26 },
		{ 312, 314, 26 },
		{ 312, 315, 26 },
		{ 312, 316, 26 },
		{ 312, 317, 26 },
		{ 312, 318, 26 },
		{ 312, 319, 26 },
		{ 312, 320, 26 },
		{ 312, 321, 26 },
		{ 312, 322, 26 },
		{ 312, 323, 26 },
		{ 312, 324, 26 },
		{ 312, 325, 26 },
		{ 312, 326, 26 },
		{ 312, 327, 26 },
		{ 312, 328, 26 },
		{ 312, 329, 26 },
		{ 312, 330, 26 },
		{ 313, 300, 26 },
		{ 313, 301, 26 },
		{ 313, 302, 26 },
		{ 313, 303, 26 },
		{ 313, 304, 26 },
		{ 313, 305, 26 },
		{ 313, 306, 26 },
		{ 313, 307, 26 },
		{ 313, 308, 26 },
		{ 313, 309, 26 },
		{ 313, 310, 26 },
		{ 313, 311, 26 },
		{ 313, 312, 26 },
		{ 313, 313, 26 },
		{ 313, 314, 26 },
		{ 313, 315, 26 },
		{ 313, 316, 26 },
		{ 313, 317, 26 },
		{ 313, 318, 26 },
		{ 313, 319, 26 },
		{ 313, 320, 26 },
		{ 313, 321, 26 },
		{ 313, 322, 26 },
		{ 313, 323, 26 },
		{ 313, 324, 26 },
		{ 313, 325, 26 },
		{ 313, 326, 26 },
		{ 313, 327, 26 },
		{ 313, 328, 26 },
		{ 313, 329, 26 },
		{ 313, 330, 26 },
		{ 313, 331, 26 },
		{ 314, 300, 26 },
		{ 314, 301, 26 },
		{ 314, 302, 26 },
		{ 314, 303, 26 },
		{ 314, 304, 26 },
		{ 314, 305, 26 },
		{ 314, 306, 26 },
		{ 314, 307, 26 },
		{ 314, 308, 26 },
		{ 314, 309, 26 },
		{ 314, 310, 26 },
		{ 314, 311, 26 },
		{ 314, 312, 26 },
		{ 314, 313, 26 },
		{ 314, 314, 26 },
		{ 314, 315, 26 },
		{ 314, 316, 26 },
		{ 314, 317, 26 },
		{ 314, 318, 26 },
		{ 314, 319, 26 },
		{ 314, 320, 26 },
		{ 314, 321, 26 },
		{ 314, 322, 26 },
		{ 314, 323, 26 },
		{ 314, 324, 26 },
		{ 314, 325, 26 },
		{ 314, 326, 26 },
		{ 314, 327, 26 },
		{ 314, 328, 26 },
		{ 314, 329, 26 },
		{ 314, 330, 26 },
		{ 314, 331, 26 },
		{ 314, 332, 26 },
		{ 315, 301, 26 },
		{ 315, 302, 26 },
		{ 315, 303, 26 },
		{ 315, 304, 26 },
		{ 315, 305, 26 },
		{ 315, 306, 26 },
		{ 315, 307, 26 },
		{ 315, 308, 26 },
		{ 315, 309, 26 },
		{ 315, 310, 26 },
		{ 315, 311, 26 },
		{ 315, 312, 26 },
		{ 315, 313, 26 },
		{ 315, 314, 26 },
		{ 315, 315, 26 },
		{ 315, 316, 26 },
		{ 315, 317, 26 },
		{ 315, 318, 26 },
		{ 315, 319, 26 },
		{ 315, 320, 26 },
		{ 315, 321, 26 },
		{ 315, 322, 26 },
		{ 315, 323, 26 },
		{ 315, 324, 26 },
		{ 315, 325, 26 },
		{ 315, 326, 26 },
		{ 315, 327, 26 },
		{ 315, 328, 26 },
		{ 315, 329, 26 },
		{ 315, 330, 26 },
		{ 315, 331, 26 },
		{ 315, 332, 26 },
		{ 315, 333, 26 },
		{ 315, 334, 26 },
		{ 316, 302, 26 },
		{ 316, 303, 26 },
		{ 316, 304, 26 },
		{ 316, 305, 26 },
		{ 316, 306, 26 },
		{ 316, 307, 26 },
		{ 316, 308, 26 },
		{ 316, 309, 26 },
		{ 316, 310, 26 },
		{ 316, 311, 26 },
		{ 316, 312, 26 },
		{ 316, 313, 26 },
		{ 316, 314, 26 },
		{ 316, 315, 26 },
		{ 316, 316, 26 },
		{ 316, 317, 26 },
		{ 316, 318, 26 },
		{ 316, 319, 26 },
		{ 316, 320, 26 },
		{ 316, 321, 26 },
		{ 316, 322, 26 },
		{ 316, 323, 26 },
		{ 316, 324, 26 },
		{ 316, 325, 26 },
		{ 316, 326, 26 },
		{ 316, 327, 26 },
		{ 316, 328, 26 },
		{ 316, 329, 26 },
		{ 316, 330, 26 },
		{ 316, 331, 26 },
		{ 316, 332, 26 },
		{ 316, 333, 26 },
		{ 316, 334, 26 },
		{ 316, 335, 26 },
		{ 317, 303, 26 },
		{ 317, 304, 26 },
		{ 317, 305, 26 },
		{ 317, 306, 26 },
		{ 317, 307, 26 },
		{ 317, 308, 26 },
		{ 317, 309, 26 },
		{ 317, 310, 26 },
		{ 317, 311, 26 },
		{ 317, 312, 26 },
		{ 317, 313, 26 },
		{ 317, 314, 26 },
		{ 317, 315, 26 },
		{ 317, 316, 26 },
		{ 317, 317, 26 },
		{ 317, 318, 26 },
		{ 317, 319, 26 },
		{ 317, 320, 26 },
		{ 317, 321, 26 },
		{ 317, 322, 26 },
		{ 317, 323, 26 },
		{ 317, 324, 26 },
		{ 317, 325, 26 },
		{ 317, 326, 26 },
		{ 317, 327, 26 },
		{ 317, 328, 26 },
		{ 317, 329, 26 },
		{ 317, 330, 26 },
		{ 317, 331, 26 },
		{ 317, 332, 26 },
		{ 317, 333, 26 },
		{ 317, 334, 26 },
		{ 317, 335, 26 },
		{ 317, 336, 26 },
		{ 318, 303, 26 },
		{ 318, 304, 26 },
		{ 318, 305, 26 },
		{ 318, 306, 26 },
		{ 318, 307, 26 },
		{ 318, 308, 26 },
		{ 318, 309, 26 },
		{ 318, 310, 26 },
		{ 318, 311, 26 },
		{ 318, 312, 26 },
		{ 318, 313, 26 },
		{ 318, 314, 26 },
		{ 318, 315, 26 },
		{ 318, 316, 26 },
		{ 318, 317, 26 },
		{ 318, 318, 26 },
		{ 318, 319, 26 },
		{ 318, 320, 26 },
		{ 318, 321, 26 },
		{ 318, 322, 26 },
		{ 318, 323, 26 },
		{ 318, 324, 26 },
		{ 318, 325, 26 },
		{ 318, 326, 26 },
		{ 318, 327, 26 },
		{ 318, 328, 26 },
		{ 318, 329, 26 },
		{ 318, 330, 26 },
		{ 318, 331, 26 },
		{ 318, 332, 26 },
		{ 318, 333, 26 },
		{ 318, 334, 26 },
		{ 318, 335, 26 },
		{ 318, 336, 26 },
		{ 318, 337, 26 },
		{ 318, 338, 26 },
		{ 319, 304, 26 },
		{ 319, 305, 26 },
		{ 319, 306, 26 },
		{ 319, 307, 26 },
		{ 319, 308, 26 },
		{ 319, 309, 26 },
		{ 319, 310, 26 },
		{ 319, 311, 26 },
		{ 319, 312, 26 },
		{ 319, 313, 26 },
		{ 319, 314, 26 },
		{ 319, 315, 26 },
		{ 319, 316, 26 },
		{ 319, 317, 26 },
		{ 319, 318, 26 },
		{ 319, 319, 26 },
		{ 319, 320, 26 },
		{ 319, 321, 26 },
		{ 319, 322, 26 },
		{ 319, 323, 26 },
		{ 319, 324, 26 },
		{ 319, 325, 26 },
		{ 319, 326, 26 },
		{ 319, 327, 26 },
		{ 319, 328, 26 },
		{ 319, 329, 26 },
		{ 319, 330, 26 },
		{ 319, 331, 26 },
		{ 319, 332, 26 },
		{ 319, 333, 26 },
		{ 319, 334, 26 },
		{ 319, 335, 26 },
		{ 319, 336, 26 },
		{ 319, 337, 26 },
		{ 319, 338, 26 },
		{ 319, 339, 26 },
		{ 320, 305, 26 },
		{ 320, 306, 26 },
		{ 320, 307, 26 },
		{ 320, 308, 26 },
		{ 320, 309, 26 },
		{ 320, 310, 26 },
		{ 320, 311, 26 },
		{ 320, 312, 26 },
		{ 320, 313, 26 },
		{ 320, 314, 26 },
		{ 320, 315, 26 },
		{ 320, 316, 26 },
		{ 320, 317, 26 },
		{ 320, 318, 26 },
		{ 320, 319, 26 },
		{ 320, 320, 26 },
		{ 320, 321, 26 },
		{ 320, 322, 26 },
		{ 320, 323, 26 },
		{ 320, 324, 26 },
		{ 320, 325, 26 },
		{ 320, 326, 26 },
		{ 320, 327, 26 },
		{ 320, 328, 26 },
		{ 320, 329, 26 },
		{ 320, 330, 26 },
		{ 320, 331, 26 },
		{ 320, 332, 26 },
		{ 320, 333, 26 },
		{ 320, 334, 26 },
		{ 320, 335, 26 },
		{ 320, 336, 26 },
		{ 320, 337, 26 },
		{ 320, 338, 26 },
		{ 320, 339, 26 },
		{ 320, 340, 26 },
		{ 321, 306, 26 },
		{ 321, 307, 26 },
		{ 321, 308, 26 },
		{ 321, 309, 26 },
		{ 321, 310, 26 },
		{ 321, 311, 26 },
		{ 321, 312, 26 },
		{ 321, 313, 26 },
		{ 321, 314, 26 },
		{ 321, 315, 26 },
		{ 321, 316, 26 },
		{ 321, 317, 26 },
		{ 321, 318, 26 },
		{ 321, 319, 26 },
		{ 321, 320, 26 },
		{ 321, 321, 26 },
		{ 321, 322, 26 },
		{ 321, 323, 26 },
		{ 321, 324, 26 },
		{ 321, 325, 26 },
		{ 321, 326, 26 },
		{ 321, 327, 26 },
		{ 321, 328, 26 },
		{ 321, 329, 26 },
		{ 321, 330, 26 },
		{ 321, 331, 26 },
		{ 321, 332, 26 },
		{ 321, 333, 26 },
		{ 321, 334, 26 },
		{ 321, 335, 26 },
		{ 321, 336, 26 },
		{ 321, 337, 26 },
		{ 321, 338, 26 },
		{ 321, 339, 26 },
		{ 321, 340, 26 },
		{ 321, 341, 26 },
		{ 321, 342, 26 },
		{ 322, 306, 26 },
		{ 322, 307, 26 },
		{ 322, 308, 26 },
		{ 322, 309, 26 },
		{ 322, 310, 26 },
		{ 322, 311, 26 },
		{ 322, 312, 26 },
		{ 322, 313, 26 },
		{ 322, 314, 26 },
		{ 322, 315, 26 },
		{ 322, 316, 26 },
		{ 322, 317, 26 },
		{ 322, 318, 26 },
		{ 322, 319, 26 },
		{ 322, 320, 26 },
		{ 322, 321, 26 },
		{ 322, 322, 26 },
		{ 322, 323, 26 },
		{ 322, 324, 26 },
		{ 322, 325, 26 },
		{ 322, 326, 26 },
		{ 322, 327, 26 },
		{ 322, 328, 26 },
		{ 322, 329, 26 },
		{ 322, 330, 26 },
		{ 322, 331, 26 },
		{ 322, 332, 26 },
		{ 322, 333, 26 },
		{ 322, 334, 26 },
		{ 322, 335, 26 },
		{ 322, 336, 26 },
		{ 322, 337, 26 },
		{ 322, 338, 26 },
		{ 322, 339, 26 },
		{ 322, 340, 26 },
		{ 322, 341, 26 },
		{ 322, 342, 26 },
		{ 322, 343, 26 },
		{ 323, 307, 26 },
		{ 323, 308, 26 },
		{ 323, 309, 26 },
		{ 323, 310, 26 },
		{ 323, 311, 26 },
		{ 323, 312, 26 },
		{ 323, 313, 26 },
		{ 323, 314, 26 },
		{ 323, 315, 26 },
		{ 323, 316, 26 },
		{ 323, 317, 26 },
		{ 323, 318, 26 },
		{ 323, 319, 26 },
		{ 323, 320, 26 },
		{ 323, 321, 26 },
		{ 323, 322, 26 },
		{ 323, 323, 26 },
		{ 323, 324, 26 },
		{ 323, 325, 26 },
		{ 323, 326, 26 },
		{ 323, 327, 26 },
		{ 323, 328, 26 },
		{ 323, 329, 26 },
		{ 323, 330, 26 },
		{ 323, 331, 26 },
		{ 323, 332, 26 },
		{ 323, 333, 26 },
		{ 323, 334, 26 },
		{ 323, 335, 26 },
		{ 323, 336, 26 },
		{ 323, 337, 26 },
		{ 323, 338, 26 },
		{ 323, 339, 26 },
		{ 323, 340, 26 },
		{ 323, 341, 26 },
		{ 323, 342, 26 },
		{ 323, 343, 26 },
		{ 323, 344, 26 },
		{ 324, 308, 26 },
		{ 324, 309, 26 },
		{ 324, 310, 26 },
		{ 324, 311, 26 },
		{ 324, 312, 26 },
		{ 324, 313, 26 },
		{ 324, 314, 26 },
		{ 324, 315, 26 },
		{ 324, 316, 26 },
		{ 324, 317, 26 },
		{ 324, 318, 26 },
		{ 324, 319, 26 },
		{ 324, 320, 26 },
		{ 324, 321, 26 },
		{ 324, 322, 26 },
		{ 324, 323, 26 },
		{ 324, 324, 26 },
		{ 324, 325, 26 },
		{ 324, 326, 26 },
		{ 324, 327, 26 },
		{ 324, 328, 26 },
		{ 324, 329, 26 },
		{ 324, 330, 26 },
		{ 324, 331, 26 },
		{ 324, 332, 26 },
		{ 324, 333, 26 },
		{ 324, 334, 26 },
		{ 324, 335, 26 },
		{ 324, 336, 26 },
		{ 324, 337, 26 },
		{ 324, 338, 26 },
		{ 324, 339, 26 },
		{ 324, 340, 26 },
		{ 324, 341, 26 },
		{ 324, 342, 26 },
		{ 324, 343, 26 },
		{ 324, 344, 26 },
		{ 324, 345, 26 },
		{ 324, 346, 26 },
		{ 325, 309, 26 },
		{ 325, 310, 26 },
		{ 325, 311, 26 },
		{ 325, 312, 26 },
		{ 325, 313, 26 },
		{ 325, 314, 26 },
		{ 325, 315, 26 },
		{ 325, 316, 26 },
		{ 325, 317, 26 },
		{ 325, 318, 26 },
		{ 325, 319, 26 },
		{ 325, 320, 26 },
		{ 325, 321, 26 },
		{ 325, 322, 26 },
		{ 325, 323, 26 },
		{ 325, 324, 26 },
		{ 325, 325, 26 },
		{ 325, 326, 26 },
		{ 325, 327, 26 },
		{ 325, 328, 26 },
		{ 325, 329, 26 },
		{ 325, 330, 26 },
		{ 325, 331, 26 },
		{ 325, 332, 26 },
		{ 325, 333, 26 },
		{ 325, 334, 26 },
		{ 325, 335, 26 },
		{ 325, 336, 26 },
		{ 325, 337, 26 },
		{ 325, 338, 26 },
		{ 325, 339, 26 },
		{ 325, 340, 26 },
		{ 325, 341, 26 },
		{ 325, 342, 26 },
		{ 325, 343, 26 },
		{ 325, 344, 26 },
		{ 325, 345, 26 },
		{ 325, 346, 26 },
		{ 325, 347, 26 },
		{ 326, 309, 26 },
		{ 326, 310, 26 },
		{ 326, 311, 26 },
		{ 326, 312, 26 },
		{ 326, 313, 26 },
		{ 326, 314, 26 },
		{ 326, 315, 26 },
		{ 326, 316, 26 },
		{ 326, 317, 26 },
		{ 326, 318, 26 },
		{ 326, 319, 26 },
		{ 326, 320, 26 },
		{ 326, 321, 26 },
		{ 326, 322, 26 },
		{ 326, 323, 26 },
		{ 326, 324, 26 },
		{ 326, 325, 26 },
		{ 326, 326, 26 },
		{ 326, 327, 26 },
		{ 326, 328, 26 },
		{ 326, 329, 26 },
		{ 326, 330, 26 },
		{ 326, 331, 26 },
		{ 326, 332, 26 },
		{ 326, 333, 26 },
		{ 326, 334, 26 },
		{ 326, 335, 26 },
		{ 326, 336, 26 },
		{ 326, 337, 26 },
		{ 326, 338, 26 },
		{ 326, 339, 26 },
		{ 326, 340, 26 },
		{ 326, 341, 26 },
		{ 326, 342, 26 },
		{ 326, 343, 26 },
		{ 326, 344, 26 },
		{ 326, 345, 26 },
		{ 326, 346, 26 },
		{ 326, 347, 26 },
		{ 326, 348, 26 },
		{ 327, 310, 26 },
		{ 327, 311, 26 },
		{ 327, 312, 26 },
		{ 327, 313, 26 },
		{ 327, 314, 26 },
		{ 327, 315, 26 },
		{ 327, 316, 26 },
		{ 327, 317, 26 },
		{ 327, 318, 26 },
		{ 327, 319, 26 },
		{ 327, 320, 26 },
		{ 327, 321, 26 },
		{ 327, 322, 26 },
		{ 327, 323, 26 },
		{ 327, 324, 26 },
		{ 327, 325, 26 },
		{ 327, 326, 26 },
		{ 327, 327, 26 },
		{ 327, 328, 26 },
		{ 327, 329, 26 },
		{ 327, 330, 26 },
		{ 327, 331, 26 },
		{ 327, 332, 26 },
		{ 327, 333, 26 },
		{ 327, 334, 26 },
		{ 327, 335, 26 },
		{ 327, 336, 26 },
		{ 327, 337, 26 },
		{ 327, 338, 26 },
		{ 327, 339, 26 },
		{ 327, 340, 26 },
		{ 327, 341, 26 },
		{ 327, 342, 26 },
		{ 327, 343, 26 },
		{ 327, 344, 26 },
		{ 327, 345, 26 },
		{ 327, 346, 26 },
		{ 327, 347, 26 },
		{ 327, 348, 26 },
		{ 327, 349, 26 },
		{ 327, 350, 26 },
		{ 328, 311, 26 },
		{ 328, 312, 26 },
		{ 328, 313, 26 },
		{ 328, 314, 26 },
		{ 328, 315, 26 },
		{ 328, 316, 26 },
		{ 328, 317, 26 },
		{ 328, 318, 26 },
		{ 328, 319, 26 },
		{ 328, 320, 26 },
		{ 328, 321, 26 },
		{ 328, 322, 26 },
		{ 328, 323, 26 },
		{ 328, 324, 26 },
		{ 328, 325, 26 },
		{ 328, 326, 26 },
		{ 328, 327, 26 },
		{ 328, 328, 26 },
		{ 328, 329, 26 },
		{ 328, 330, 26 },
		{ 328, 331, 26 },
		{ 328, 332, 26 },
		{ 328, 333, 26 },
		{ 328, 334, 26 },
		{ 328, 335, 26 },
		{ 328, 336, 26 },
		{ 328, 337, 26 },
		{ 328, 338, 26 },
		{ 328, 339, 26 },
		{ 328, 340, 26 },
		{ 328, 341, 26 },
		{ 328, 342, 26 },
		{ 328, 343, 26 },
		{ 328, 344, 26 },
		{ 328, 345, 26 },
		{ 328, 346, 26 },
		{ 328, 347, 26 },
		{ 328, 348, 26 },
		{ 328, 349, 26 },
		{ 328, 350, 26 },
		{ 328, 351, 26 },
		{ 329, 312, 26 },
		{ 329, 313, 26 },
		{ 329, 314, 26 },
		{ 329, 315, 26 },
		{ 329, 316, 26 },
		{ 329, 317, 26 },
		{ 329, 318, 26 },
		{ 329, 319, 26 },
		{ 329, 320, 26 },
		{ 329, 321, 26 },
		{ 329, 322, 26 },
		{ 329, 323, 26 },
		{ 329, 324, 26 },
		{ 329, 325, 26 },
		{ 329, 326, 26 },
		{ 329, 327, 26 },
		{ 329, 328, 26 },
		{ 329, 329, 26 },
		{ 329, 330, 26 },
		{ 329, 331, 26 },
		{ 329, 332, 26 },
		{ 329, 333, 26 },
		{ 329, 334, 26 },
		{ 329, 335, 26 },
		{ 329, 336, 26 },
		{ 329, 337, 26 },
		{ 329, 338, 26 },
		{ 329, 339, 26 },
		{ 329, 340, 26 },
		{ 329, 341, 26 },
		{ 329, 342, 26 },
		{ 329, 343, 26 },
		{ 329, 344, 26 },
		{ 329, 345, 26 },
		{ 329, 346, 26 },
		{ 329, 347, 26 },
		{ 329, 348, 26 },
		{ 329, 349, 26 },
		{ 329, 350, 26 },
		{ 329, 351, 26 },
		{ 329, 352, 26 },
		{ 330, 312, 26 },
		{ 330, 313, 26 },
		{ 330, 314, 26 },
		{ 330, 315, 26 },
		{ 330, 316, 26 },
		{ 330, 317, 26 },
		{ 330, 318, 26 },
		{ 330, 319, 26 },
		{ 330, 320, 26 },
		{ 330, 321, 26 },
		{ 330, 322, 26 },
		{ 330, 323, 26 },
		{ 330, 324, 26 },
		{ 330, 325, 26 },
		{ 330, 326, 26 },
		{ 330, 327, 26 },
		{ 330, 328, 26 },
		{ 330, 329, 26 },
		{ 330, 330, 26 },
		{ 330, 331, 26 },
		{ 330, 332, 26 },
		{ 330, 333, 26 },
		{ 330, 334, 26 },
		{ 330, 335, 26 },
		{ 330, 336, 26 },
		{ 330, 337, 26 },
		{ 330, 338, 26 },
		{ 330, 339, 26 },
		{ 330, 340, 26 },
		{ 330, 341, 26 },
		{ 330, 342, 26 },
		{ 330, 343, 26 },
		{ 330, 344, 26 },
		{ 330, 345, 26 },
		{ 330, 346, 26 },
		{ 330, 347, 26 },
		{ 330, 348, 26 },
		{ 330, 349, 26 },
		{ 330, 350, 26 },
		{ 330, 351, 26 },
		{ 330, 352, 26 },
		{ 330, 353, 26 },
		{ 330, 354, 26 },
		{ 331, 313, 26 },
		{ 331, 314, 26 },
		{ 331, 315, 26 },
		{ 331, 316, 26 },
		{ 331, 317, 26 },
		{ 331, 318, 26 },
		{ 331, 319, 26 },
		{ 331, 320, 26 },
		{ 331, 321, 26 },
		{ 331, 322, 26 },
		{ 331, 323, 26 },
		{ 331, 324, 26 },
		{ 331, 325, 26 },
		{ 331, 326, 26 },
		{ 331, 327, 26 },
		{ 331, 328, 26 },
		{ 331, 329, 26 },
		{ 331, 330, 26 },
		{ 331, 331, 26 },
		{ 331, 332, 26 },
		{ 331, 333, 26 },
		{ 331, 334, 26 },
		{ 331, 335, 26 },
		{ 331, 336, 26 },
		{ 331, 337, 26 },
		{ 331, 338, 26 },
		{ 331, 339, 26 },
		{ 331, 340, 26 },
		{ 331, 341, 26 },
		{ 331, 342, 26 },
		{ 331, 343, 26 },
		{ 331, 344, 26 },
		{ 331, 345, 26 },
		{ 331, 346, 26 },
		{ 331, 347, 26 },
		{ 331, 348, 26 },
		{ 331, 349, 26 },
		{ 331, 350, 26 },
		{ 331, 351, 26 },
		{ 331, 352, 26 },
		{ 331, 353, 26 },
		{ 331, 354, 26 },
		{ 331, 355, 26 },
		{ 332, 314, 26 },
		{ 332, 315, 26 },
		{ 332, 316, 26 },
		{ 332, 317, 26 },
		{ 332, 318, 26 },
		{ 332, 319, 26 },
		{ 332, 320, 26 },
		{ 332, 321, 26 },
		{ 332, 322, 26 },
		{ 332, 323, 26 },
		{ 332, 324, 26 },
		{ 332, 325, 26 },
		{ 332, 326, 26 },
		{ 332, 327, 26 },
		{ 332, 328, 26 },
		{ 332, 329, 26 },
		{ 332, 330, 26 },
		{ 332, 331, 26 },
		{ 332, 332, 26 },
		{ 332, 333, 26 },
		{ 332, 334, 26 },
		{ 332, 335, 26 },
		{ 332, 336, 26 },
		{ 332, 337, 26 },
		{ 332, 338, 26 },
		{ 332, 339, 26 },
		{ 332, 340, 26 },
		{ 332, 341, 26 },
		{ 332, 342, 26 },
		{ 332, 343, 26 },
		{ 332, 344, 26 },
		{ 332, 345, 26 },
		{ 332, 346, 26 },
		{ 332, 347, 26 },
		{ 332, 348, 26 },
		{ 332, 349, 26 },
		{ 332, 350, 26 },
		{ 332, 351, 26 },
		{ 332, 352, 26 },
		{ 332, 353, 26 },
		{ 332, 354, 26 },
		{ 332, 355, 26 },
		{ 332, 356, 26 },
		{ 333, 315, 26 },
		{ 333, 316, 26 },
		{ 333, 317, 26 },
		{ 333, 318, 26 },
		{ 333, 319, 26 },
		{ 333, 320, 26 },
		{ 333, 321, 26 },
		{ 333, 322, 26 },
		{ 333, 323, 26 },
		{ 333, 324, 26 },
		{ 333, 325, 26 },
		{ 333, 326, 26 },
		{ 333, 327, 26 },
		{ 333, 328, 26 },
		{ 333, 329, 26 },
		{ 333, 330, 26 },
		{ 333, 331, 26 },
		{ 333, 332, 26 },
		{ 333, 333, 26 },
		{ 333, 334, 26 },
		{ 333, 335, 26 },
		{ 333, 336, 26 },
		{ 333, 337, 26 },
		{ 333, 338, 26 },
		{ 333, 339, 26 },
		{ 333, 340, 26 },
		{ 333, 341, 26 },
		{ 333, 342, 26 },
		{ 333, 343, 26 },
		{ 333, 344, 26 },
		{ 333, 345, 26 },
		{ 333, 346, 26 },
		{ 333, 347, 26 },
		{ 333, 348, 26 },
		{ 333, 349, 26 },
		{ 333, 350, 26 },
		{ 333, 351, 26 },
		{ 333, 352, 26 },
		{ 333, 353, 26 },
		{ 333, 354, 26 },
		{ 333, 355, 26 },
		{ 333, 356, 26 },
		{ 333, 357, 26 },
		{ 333, 358, 26 },
		{ 334, 315, 26 },
		{ 334, 316, 26 },
		{ 334, 317, 26 },
		{ 334, 318, 26 },
		{ 334, 319, 26 },
		{ 334, 320, 26 },
		{ 334, 321, 26 },
		{ 334, 322, 26 },
		{ 334, 323, 26 },
		{ 334, 324, 26 },
		{ 334, 325, 26 },
		{ 334, 326, 26 },
		{ 334, 327, 26 },
		{ 334, 328, 26 },
		{ 334, 329, 26 },
		{ 334, 330, 26 },
		{ 334, 331, 26 },
		{ 334, 332, 26 },
		{ 334, 333, 26 },
		{ 334, 334, 26 },
		{ 334, 335, 26 },
		{ 334, 336, 26 },
		{ 334, 337, 26 },
		{ 334, 338, 26 },
		{ 334, 339, 26 },
		{ 334, 340, 26 },
		{ 334, 341, 26 },
		{ 334, 342, 26 },
		{ 334, 343, 26 },
		{ 334, 344, 26 },
		{ 334, 345, 26 },
		{ 334, 346, 26 },
		{ 334, 347, 26 },
		{ 334, 348, 26 },
		{ 334, 349, 26 },
		{ 334, 350, 26 },
		{ 334, 351, 26 },
		{ 334, 352, 26 },
		{ 334, 353, 26 },
		{ 334, 354, 26 },
		{ 334, 355, 26 },
		{ 334, 356, 26 },
		{ 334, 357, 26 },
		{ 334, 358, 26 },
		{ 334, 359, 26 },
		{ 335, 316, 26 },
		{ 335, 317, 26 },
		{ 335, 318, 26 },
		{ 335, 319, 26 },
		{ 335, 320, 26 },
		{ 335, 321, 26 },
		{ 335, 322, 26 },
		{ 335, 323, 26 },
		{ 335, 324, 26 },
		{ 335, 325, 26 },
		{ 335, 326, 26 },
		{ 335, 327, 26 },
		{ 335, 328, 26 },
		{ 335, 329, 26 },
		{ 335, 330, 26 },
		{ 335, 331, 26 },
		{ 335, 332, 26 },
		{ 335, 333, 26 },
		{ 335, 334, 26 },
		{ 335, 335, 26 },
		{ 335, 336, 26 },
		{ 335, 337, 26 },
		{ 335, 338, 26 },
		{ 335, 339, 26 },
		{ 335, 340, 26 },
		{ 335, 341, 26 },
		{ 335, 342, 26 },
		{ 335, 343, 26 },
		{ 335, 344, 26 },
		{ 335, 345, 26 },
		{ 335, 346, 26 },
		{ 335, 347, 26 },
		{ 335, 348, 26 },
		{ 335, 349, 26 },
		{ 335, 350, 26 },
		{ 335, 351, 26 },
		{ 335, 352, 26 },
		{ 335, 353, 26 },
		{ 335, 354, 26 },
		{ 335, 355, 26 },
		{ 335, 356, 26 },
		{ 335, 357, 26 },
		{ 335, 358, 26 },
		{ 335, 359, 26 },
		{ 335, 360, 26 },
		{ 336, 317, 26 },
		{ 336, 318, 26 },
		{ 336, 319, 26 },
		{ 336, 320, 26 },
		{ 336, 321, 26 },
		{ 336, 322, 26 },
		{ 336, 323, 26 },
		{ 336, 324, 26 },
		{ 336, 325, 26 },
		{ 336, 326, 26 },
		{ 336, 327, 26 },
		{ 336, 328, 26 },
		{ 336, 329, 26 },
		{ 336, 330, 26 },
		{ 336, 331, 26 },
		{ 336, 332, 26 },
		{ 336, 333, 26 },
		{ 336, 334, 26 },
		{ 336, 335, 26 },
		{ 336, 336, 26 },
		{ 336, 337, 26 },
		{ 336, 338, 26 },
		{ 336, 339, 26 },
		{ 336, 340, 26 },
		{ 336, 341, 26 },
		{ 336, 342, 26 },
		{ 336, 343, 26 },
		{ 336, 344, 26 },
		{ 336, 345, 26 },
		{ 336, 346, 26 },
		{ 336, 347, 26 },
		{ 336, 348, 26 },
		{ 336, 349, 26 },
		{ 336, 350, 26 },
		{ 336, 351, 26 },
		{ 336, 352, 26 },
		{ 336, 353, 26 },
		{ 336, 354, 26 },
		{ 336, 355, 26 },
		{ 336, 356, 26 },
		{ 336, 357, 26 },
		{ 336, 358, 26 },
		{ 336, 359, 26 },
		{ 336, 360, 26 },
		{ 336, 361, 26 },
		{ 336, 362, 26 },
		{ 337, 318, 26 },
		{ 337, 319, 26 },
		{ 337, 320, 26 },
		{ 337, 321, 26 },
		{ 337, 322, 26 },
		{ 337, 323, 26 },
		{ 337, 324, 26 },
		{ 337, 325, 26 },
		{ 337, 326, 26 },
		{ 337, 327, 26 },
		{ 337, 328, 26 },
		{ 337, 329, 26 },
		{ 337, 330, 26 },
		{ 337, 331, 26 },
		{ 337, 332, 26 },
		{ 337, 333, 26 },
		{ 337, 334, 26 },
		{ 337, 335, 26 },
		{ 337, 336, 26 },
		{ 337, 337, 26 },
		{ 337, 338, 26 },
		{ 337, 339, 26 },
		{ 337, 340, 26 },
		{ 337, 341, 26 },
		{ 337, 342, 26 },
		{ 337, 343, 26 },
		{ 337, 344, 26 },
		{ 337, 345, 26 },
		{ 337, 346, 26 },
		{ 337, 347, 26 },
		{ 337, 348, 26 },
		{ 337, 349, 26 },
		{ 337, 350, 26 },
		{ 337, 351, 26 },
		{ 337, 352, 26 },
		{ 337, 353, 26 },
		{ 337, 354, 26 },
		{ 337, 355, 26 },
		{ 337, 356, 26 },
		{ 337, 357, 26 },
		{ 337, 358, 26 },
		{ 337, 359, 26 },
		{ 337, 360, 26 },
		{ 337, 361, 26 },
		{ 337, 362, 26 },
		{ 337, 363, 26 },
		{ 338, 318, 26 },
		{ 338, 319, 26 },
		{ 338, 320, 26 },
		{ 338, 321, 26 },
		{ 338, 322, 26 },
		{ 338, 323, 26 },
		{ 338, 324, 26 },
		{ 338, 325, 26 },
		{ 338, 326, 26 },
		{ 338, 327, 26 },
		{ 338, 328, 26 },
		{ 338, 329, 26 },
		{ 338, 330, 26 },
		{ 338, 331, 26 },
		{ 338, 332, 26 },
		{ 338, 333, 26 },
		{ 338, 334, 26 },
		{ 338, 335, 26 },
		{ 338, 336, 26 },
		{ 338, 337, 26 },
		{ 338, 338, 26 },
		{ 338, 339, 26 },
		{ 338, 340, 26 },
		{ 338, 341, 26 },
		{ 338, 342, 26 },
		{ 338, 343, 26 },
		{ 338, 344, 26 },
		{ 338, 345, 26 },
		{ 338, 346, 26 },
		{ 338, 347, 26 },
		{ 338, 348, 26 },
		{ 338, 349, 26 },
		{ 338, 350, 26 },
		{ 338, 351, 26 },
		{ 338, 352, 26 },
		{ 338, 353, 26 },
		{ 338, 354, 26 },
		{ 338, 355, 26 },
		{ 338, 356, 26 },
		{ 338, 357, 26 },
		{ 338, 358, 26 },
		{ 338, 359, 26 },
		{ 338, 360, 26 },
		{ 338, 361, 26 },
		{ 338, 362, 26 },
		{ 338, 363, 26 },
		{ 338, 364, 26 },
		{ 339, 319, 26 },
		{ 339, 320, 26 },
		{ 339, 321, 26 },
		{ 339, 322, 26 },
		{ 339, 323, 26 },
		{ 339, 324, 26 },
		{ 339, 325, 26 },
		{ 339, 326, 26 },
		{ 339, 327, 26 },
		{ 339, 328, 26 },
		{ 339, 329, 26 },
		{ 339, 330, 26 },
		{ 339, 331, 26 },
		{ 339, 332, 26 },
		{ 339, 333, 26 },
		{ 339, 334, 26 },
		{ 339, 335, 26 },
		{ 339, 336, 26 },
		{ 339, 337, 26 },
		{ 339, 338, 26 },
		{ 339, 339, 26 },
		{ 339, 340, 26 },
		{ 339, 341, 26 },
		{ 339, 342, 26 },
		{ 339, 343, 26 },
		{ 339, 344, 26 },
		{ 339, 345, 26 },
		{ 339, 346, 26 },
		{ 339, 347, 26 },
		{ 339, 348, 26 },
		{ 339, 349, 26 },
		{ 339, 350, 26 },
		{ 339, 351, 26 },
		{ 339, 352, 26 },
		{ 339, 353, 26 },
		{ 339, 354, 26 },
		{ 339, 355, 26 },
		{ 339, 356, 26 },
		{ 339, 357, 26 },
		{ 339, 358, 26 },
		{ 339, 359, 26 },
		{ 339, 360, 26 },
		{ 339, 361, 26 },
		{ 339, 362, 26 },
		{ 339, 363, 26 },
		{ 339, 364, 26 },
		{ 339, 365, 26 },
		{ 339, 366, 26 },
		{ 340, 320, 26 },
		{ 340, 321, 26 },
		{ 340, 322, 26 },
		{ 340, 323, 26 },
		{ 340, 324, 26 },
		{ 340, 325, 26 },
		{ 340, 326, 26 },
		{ 340, 327, 26 },
		{ 340, 328, 26 },
		{ 340, 329, 26 },
		{ 340, 330, 26 },
		{ 340, 331, 26 },
		{ 340, 332, 26 },
		{ 340, 333, 26 },
		{ 340, 334, 26 },
		{ 340, 335, 26 },
		{ 340, 336, 26 },
		{ 340, 337, 26 },
		{ 340, 338, 26 },
		{ 340, 339, 26 },
		{ 340, 340, 26 },
		{ 340, 341, 26 },
		{ 340, 342, 26 },
		{ 340, 343, 26 },
		{ 340, 344, 26 },
		{ 340, 345, 26 },
		{ 340, 346, 26 },
		{ 340, 347, 26 },
		{ 340, 348, 26 },
		{ 340, 349, 26 },
		{ 340, 350, 26 },
		{ 340, 351, 26 },
		{ 340, 352, 26 },
		{ 340, 353, 26 },
		{ 340, 354, 26 },
		{ 340, 355, 26 },
		{ 340, 356, 26 },
		{ 340, 357, 26 },
		{ 340, 358, 26 },
		{ 340, 359, 26 },
		{ 340, 360, 26 },
		{ 340, 361, 26 },
		{ 340, 362, 26 },
		{ 340, 363, 26 },
		{ 340, 364, 26 },
		{ 340, 365, 26 },
		{ 340, 366, 26 },
		{ 340, 367, 26 },
		{ 341, 321, 26 },
		{ 341, 322, 26 },
		{ 341, 323, 26 },
		{ 341, 324, 26 },
		{ 341, 325, 26 },
		{ 341, 326, 26 },
		{ 341, 327, 26 },
		{ 341, 328, 26 },
		{ 341, 329, 26 },
		{ 341, 330, 26 },
		{ 341, 331, 26 },
		{ 341, 332, 26 },
		{ 341, 333, 26 },
		{ 341, 334, 26 },
		{ 341, 335, 26 },
		{ 341, 336, 26 },
		{ 341, 337, 26 },
		{ 341, 338, 26 },
		{ 341, 339, 26 },
		{ 341, 340, 26 },
		{ 341, 341, 26 },
		{ 341, 342, 26 },
		{ 341, 343, 26 },
		{ 341, 344, 26 },
		{ 341, 345, 26 },
		{ 341, 346, 26 },
		{ 341, 347, 26 },
		{ 341, 348, 26 },
		{ 341, 349, 26 },
		{ 341, 350, 26 },
		{ 341, 351, 26 },
		{ 341, 352, 26 },
		{ 341, 353, 26 },
		{ 341, 354, 26 },
		{ 341, 355, 26 },
		{ 341, 356, 26 },
		{ 341, 357, 26 },
		{ 341, 358, 26 },
		{ 341, 359, 26 },
		{ 341, 360, 26 },
		{ 341, 361, 26 },
		{ 341, 362, 26 },
		{ 341, 363, 26 },
		{ 341, 364, 26 },
		{ 341, 365, 26 },
		{ 341, 366, 26 },
		{ 341, 367, 26 },
		{ 341, 368, 26 },
		{ 342, 321, 26 },
		{ 342, 322, 26 },
		{ 342, 323, 26 },
		{ 342, 324, 26 },
		{ 342, 325, 26 },
		{ 342, 326, 26 },
		{ 342, 327, 26 },
		{ 342, 328, 26 },
		{ 342, 329, 26 },
		{ 342, 330, 26 },
		{ 342, 331, 26 },
		{ 342, 332, 26 },
		{ 342, 333, 26 },
		{ 342, 334, 26 },
		{ 342, 335, 26 },
		{ 342, 336, 26 },
		{ 342, 337, 26 },
		{ 342, 338, 26 },
		{ 342, 339, 26 },
		{ 342, 340, 26 },
		{ 342, 341, 26 },
		{ 342, 342, 26 },
		{ 342, 343, 26 },
		{ 342, 344, 26 },
		{ 342, 345, 26 },
		{ 342, 346, 26 },
		{ 342, 347, 26 },
		{ 342, 348, 26 },
		{ 342, 349, 26 },
		{ 342, 350, 26 },
		{ 342, 351, 26 },
		{ 342, 352, 26 },
		{ 342, 353, 26 },
		{ 342, 354, 26 },
		{ 342, 355, 26 },
		{ 342, 356, 26 },
		{ 342, 357, 26 },
		{ 342, 358, 26 },
		{ 342, 359, 26 },
		{ 342, 360, 26 },
		{ 342, 361, 26 },
		{ 342, 362, 26 },
		{ 342, 363, 26 },
		{ 342, 364, 26 },
		{ 342, 365, 26 },
		{ 342, 366, 26 },
		{ 342, 367, 26 },
		{ 342, 368, 26 },
		{ 342, 369, 26 },
		{ 342, 370, 26 },
		{ 343, 322, 26 },
		{ 343, 323, 26 },
		{ 343, 324, 26 },
		{ 343, 325, 26 },
		{ 343, 326, 26 },
		{ 343, 327, 26 },
		{ 343, 328, 26 },
		{ 343, 329, 26 },
		{ 343, 330, 26 },
		{ 343, 331, 26 },
		{ 343, 332, 26 },
		{ 343, 333, 26 },
		{ 343, 334, 26 },
		{ 343, 335, 26 },
		{ 343, 336, 26 },
		{ 343, 337, 26 },
		{ 343, 338, 26 },
		{ 343, 339, 26 },
		{ 343, 340, 26 },
		{ 343, 341, 26 },
		{ 343, 342, 26 },
		{ 343, 343, 26 },
		{ 343, 344, 26 },
		{ 343, 345, 26 },
		{ 343, 346, 26 },
		{ 343, 347, 26 },
		{ 343, 348, 26 },
		{ 343, 349, 26 },
		{ 343, 350, 26 },
		{ 343, 351, 26 },
		{ 343, 352, 26 },
		{ 343, 353, 26 },
		{ 343, 354, 26 },
		{ 343, 355, 26 },
		{ 343, 356, 26 },
		{ 343, 357, 26 },
		{ 343, 358, 26 },
		{ 343, 359, 26 },
		{ 343, 360, 26 },
		{ 343, 361, 26 },
		{ 343, 362, 26 },
		{ 343, 363, 26 },
		{ 343, 364, 26 },
		{ 343, 365, 26 },
		{ 343, 366, 26 },
		{ 343, 367, 26 },
		{ 343, 368, 26 },
		{ 343, 369, 26 },
		{ 343, 370, 26 },
		{ 343, 371, 26 },
		{ 344, 323, 26 },
		{ 344, 324, 26 },
		{ 344, 325, 26 },
		{ 344, 326, 26 },
		{ 344, 327, 26 },
		{ 344, 328, 26 },
		{ 344, 329, 26 },
		{ 344, 330, 26 },
		{ 344, 331, 26 },
		{ 344, 332, 26 },
		{ 344, 333, 26 },
		{ 344, 334, 26 },
		{ 344, 335, 26 },
		{ 344, 336, 26 },
		{ 344, 337, 26 },
		{ 344, 338, 26 },
		{ 344, 339, 26 },
		{ 344, 340, 26 },
		{ 344, 341, 26 },
		{ 344, 342, 26 },
		{ 344, 343, 26 },
		{ 344, 344, 26 },
		{ 344, 345, 26 },
		{ 344, 346, 26 },
		{ 344, 347, 26 },
		{ 344, 348, 26 },
		{ 344, 349, 26 },
		{ 344, 350, 26 },
		{ 344, 351, 26 },
		{ 344, 352, 26 },
		{ 344, 353, 26 },
		{ 344, 354, 26 },
		{ 344, 355, 26 },
		{ 344, 356, 26 },
		{ 344, 357, 26 },
		{ 344, 358, 26 },
		{ 344, 359, 26 },
		{ 344, 360, 26 },
		{ 344, 361, 26 },
		{ 344, 362, 26 },
		{ 344, 363, 26 },
		{ 344, 364, 26 },
		{ 344, 365, 26 },
		{ 344, 366, 26 },
		{ 344, 367, 26 },
		{ 344, 368, 26 },
		{ 344, 369, 26 },
		{ 344, 370, 26 },
		{ 344, 371, 26 },
		{ 344, 372, 26 },
		{ 345, 324, 26 },
		{ 345, 325, 26 },
		{ 345, 326, 26 },
		{ 345, 327, 26 },
		{ 345, 328, 26 },
		{ 345, 329, 26 },
		{ 345, 330, 26 },
		{ 345, 331, 26 },
		{ 345, 332, 26 },
		{ 345, 333, 26 },
		{ 345, 334, 26 },
		{ 345, 335, 26 },
		{ 345, 336, 26 },
		{ 345, 337, 26 },
		{ 345, 338, 26 },
		{ 345, 339, 26 },
		{ 345, 340, 26 },
		{ 345, 341, 26 },
		{ 345, 342, 26 },
		{ 345, 343, 26 },
		{ 345, 344, 26 },
		{ 345, 345, 26 },
		{ 345, 346, 26 },
		{ 345, 347, 26 },
		{ 345, 348, 26 },
		{ 345, 349, 26 },
		{ 345, 350, 26 },
		{ 345, 351, 26 },
		{ 345, 352, 26 },
		{ 345, 353, 26 },
		{ 345, 354, 26 },
		{ 345, 355, 26 },
		{ 345, 356, 26 },
		{ 345, 357, 26 },
		{ 345, 358, 26 },
		{ 345, 359, 26 },
		{ 345, 360, 26 },
		{ 345, 361, 26 },
		{ 345, 362, 26 },
		{ 345, 363, 26 },
		{ 345, 364, 26 },
		{ 345, 365, 26 },
		{ 345, 366, 26 },
		{ 345, 367, 26 },
		{ 345, 368, 26 },
		{ 345, 369, 26 },
		{ 345, 370, 26 },
		{ 345, 371, 26 },
		{ 345, 372, 26 },
		{ 345, 373, 26 },
		{ 345, 374, 26 },
		{ 346, 324, 26 },
		{ 346, 325, 26 },
		{ 346, 326, 26 },
		{ 346, 327, 26 },
		{ 346, 328, 26 },
		{ 346, 329, 26 },
		{ 346, 330, 26 },
		{ 346, 331, 26 },
		{ 346, 332, 26 },
		{ 346, 333, 26 },
		{ 346, 334, 26 },
		{ 346, 335, 26 },
		{ 346, 336, 26 },
		{ 346, 337, 26 },
		{ 346, 338, 26 },
		{ 346, 339, 26 },
		{ 346, 340, 26 },
		{ 346, 341, 26 },
		{ 346, 342, 26 },
		{ 346, 343, 26 },
		{ 346, 344, 26 },
		{ 346, 345, 26 },
		{ 346, 346, 26 },
		{ 346, 347, 26 },
		{ 346, 348, 26 },
		{ 346, 349, 26 },
		{ 346, 350, 26 },
		{ 346, 351, 26 },
		{ 346, 352, 26 },
		{ 346, 353, 26 },
		{ 346, 354, 26 },
		{ 346, 355, 26 },
		{ 346, 356, 26 },
		{ 346, 357, 26 },
		{ 346, 358, 26 },
		{ 346, 359, 26 },
		{ 346, 360, 26 },
		{ 346, 361, 26 },
		{ 346, 362, 26 },
		{ 346, 363, 26 },
		{ 346, 364, 26 },
		{ 346, 365, 26 },
		{ 346, 366, 26 },
		{ 346, 367, 26 },
		{ 346, 368, 26 },
		{ 346, 369, 26 },
		{ 346, 370, 26 },
		{ 346, 371, 26 },
		{ 346, 372, 26 },
		{ 346, 373, 26 },
		{ 346, 374, 26 },
		{ 346, 375, 26 },
		{ 347, 325, 26 },
		{ 347, 326, 26 },
		{ 347, 327, 26 },
		{ 347, 328, 26 },
		{ 347, 329, 26 },
		{ 347, 330, 26 },
		{ 347, 331, 26 },
		{ 347, 332, 26 },
		{ 347, 333, 26 },
		{ 347, 334, 26 },
		{ 347, 335, 26 },
		{ 347, 336, 26 },
		{ 347, 337, 26 },
		{ 347, 338, 26 },
		{ 347, 339, 26 },
		{ 347, 340, 26 },
		{ 347, 341, 26 },
		{ 347, 342, 26 },
		{ 347, 343, 26 },
		{ 347, 344, 26 },
		{ 347, 345, 26 },
		{ 347, 346, 26 },
		{ 347, 347, 26 },
		{ 347, 348, 26 },
		{ 347, 349, 26 },
		{ 347, 350, 26 },
		{ 347, 351, 26 },
		{ 347, 352, 26 },
		{ 347, 353, 26 },
		{ 347, 354, 26 },
		{ 347, 355, 26 },
		{ 347, 356, 26 },
		{ 347, 357, 26 },
		{ 347, 358, 26 },
		{ 347, 359, 26 },
		{ 347, 360, 26 },
		{ 347, 361, 26 },
		{ 347, 362, 26 },
		{ 347, 363, 26 },
		{ 347, 364, 26 },
		{ 347, 365, 26 },
		{ 347, 366, 26 },
		{ 347, 367, 26 },
		{ 347, 368, 26 },
		{ 347, 369, 26 },
		{ 347, 370, 26 },
		{ 347, 371, 26 },
		{ 347, 372, 26 },
		{ 347, 373, 26 },
		{ 347, 374, 26 },
		{ 347, 375, 26 },
		{ 347, 376, 26 },
		{ 348, 326, 26 },
		{ 348, 327, 26 },
		{ 348, 328, 26 },
		{ 348, 329, 26 },
		{ 348, 330, 26 },
		{ 348, 331, 26 },
		{ 348, 332, 26 },
		{ 348, 333, 26 },
		{ 348, 334, 26 },
		{ 348, 335, 26 },
		{ 348, 336, 26 },
		{ 348, 337, 26 },
		{ 348, 338, 26 },
		{ 348, 339, 26 },
		{ 348, 340, 26 },
		{ 348, 341, 26 },
		{ 348, 342, 26 },
		{ 348, 343, 26 },
		{ 348, 344, 26 },
		{ 348, 345, 26 },
		{ 348, 346, 26 },
		{ 348, 347, 26 },
		{ 348, 348, 26 },
		{ 348, 349, 26 },
		{ 348, 350, 26 },
		{ 348, 351, 26 },
		{ 348, 352, 26 },
		{ 348, 353, 26 },
		{ 348, 354, 26 },
		{ 348, 355, 26 },
		{ 348, 356, 26 },
		{ 348, 357, 26 },
		{ 348, 358, 26 },
		{ 348, 359, 26 },
		{ 348, 360, 26 },
		{ 348, 361, 26 },
		{ 348, 362, 26 },
		{ 348, 363, 26 },
		{ 348, 364, 26 },
		{ 348, 365, 26 },
		{ 348, 366, 26 },
		{ 348, 367, 26 },
		{ 348, 368, 26 },
		{ 348, 369, 26 },
		{ 348, 370, 26 },
		{ 348, 371, 26 },
		{ 348, 372, 26 },
		{ 348, 373, 26 },
		{ 348, 374, 26 },
		{ 348, 375, 26 },
		{ 348, 376, 26 },
		{ 348, 377, 26 },
		{ 349, 327, 26 },
		{ 349, 328, 26 },
		{ 349, 329, 26 },
		{ 349, 330, 26 },
		{ 349, 331, 26 },
		{ 349, 332, 26 },
		{ 349, 333, 26 },
		{ 349, 334, 26 },
		{ 349, 335, 26 },
		{ 349, 336, 26 },
		{ 349, 337, 26 },
		{ 349, 338, 26 },
		{ 349, 339, 26 },
		{ 349, 340, 26 },
		{ 349, 341, 26 },
		{ 349, 342, 26 },
		{ 349, 343, 26 },
		{ 349, 344, 26 },
		{ 349, 345, 26 },
		{ 349, 346, 26 },
		{ 349, 347, 26 },
		{ 349, 348, 26 },
		{ 349, 349, 26 },
		{ 349, 350, 26 },
		{ 349, 351, 26 },
		{ 349, 352, 26 },
		{ 349, 353, 26 },
		{ 349, 354, 26 },
		{ 349, 355, 26 },
		{ 349, 356, 26 },
		{ 349, 357, 26 },
		{ 349, 358, 26 },
		{ 349, 359, 26 },
		{ 349, 360, 26 },
		{ 349, 361, 26 },
		{ 349, 362, 26 },
		{ 349, 363, 26 },
		{ 349, 364, 26 },
		{ 349, 365, 26 },
		{ 349, 366, 26 },
		{ 349, 367, 26 },
		{ 349, 368, 26 },
		{ 349, 369, 26 },
		{ 349, 370, 26 },
		{ 349, 371, 26 },
		{ 349, 372, 26 },
		{ 349, 373, 26 },
		{ 349, 374, 26 },
		{ 349, 375, 26 },
		{ 349, 376, 26 },
		{ 349, 377, 26 },
		{ 349, 378, 26 },
		{ 350, 327, 26 },
		{ 350, 328, 26 },
		{ 350, 329, 26 },
		{ 350, 330, 26 },
		{ 350, 331, 26 },
		{ 350, 332, 26 },
		{ 350, 333, 26 },
		{ 350, 334, 26 },
		{ 350, 335, 26 },
		{ 350, 336, 26 },
		{ 350, 337, 26 },
		{ 350, 338, 26 },
		{ 350, 339, 26 },
		{ 350, 340, 26 },
		{ 350, 341, 26 },
		{ 350, 342, 26 },
		{ 350, 343, 26 },
		{ 350, 344, 26 },
		{ 350, 345, 26 },
		{ 350, 346, 26 },
		{ 350, 347, 26 },
		{ 350, 348, 26 },
		{ 350, 349, 26 },
		{ 350, 350, 26 },
		{ 350, 351, 26 },
		{ 350, 352, 26 },
		{ 350, 353, 26 },
		{ 350, 354, 26 },
		{ 350, 355, 26 },
		{ 350, 356, 26 },
		{ 350, 357, 26 },
		{ 350, 358, 26 },
		{ 350, 359, 26 },
		{ 350, 360, 26 },
		{ 350, 361, 26 },
		{ 350, 362, 26 },
		{ 350, 363, 26 },
		{ 350, 364, 26 },
		{ 350, 365, 26 },
		{ 350, 366, 26 },
		{ 350, 367, 26 },
		{ 350, 368, 26 },
		{ 350, 369, 26 },
		{ 350, 370, 26 },
		{ 350, 371, 26 },
		{ 350, 372, 26 },
		{ 350, 373, 26 },
		{ 350, 374, 26 },
		{ 350, 375, 26 },
		{ 350, 376, 26 },
		{ 350, 377, 26 },
		{ 350, 378, 26 },
		{ 350, 379, 26 },
		{ 351, 328, 26 },
		{ 351, 329, 26 },
		{ 351, 330, 26 },
		{ 351, 331, 26 },
		{ 351, 332, 26 },
		{ 351, 333, 26 },
		{ 351, 334, 26 },
		{ 351, 335, 26 },
		{ 351, 336, 26 },
		{ 351, 337, 26 },
		{ 351, 338, 26 },
		{ 351, 339, 26 },
		{ 351, 340, 26 },
		{ 351, 341, 26 },
		{ 351, 342, 26 },
		{ 351, 343, 26 },
		{ 351, 344, 26 },
		{ 351, 345, 26 },
		{ 351, 346, 26 },
		{ 351, 347, 26 },
		{ 351, 348, 26 },
		{ 351, 349, 26 },
		{ 351, 350, 26 },
		{ 351, 351, 26 },
		{ 351, 352, 26 },
		{ 351, 353, 26 },
		{ 351, 354, 26 },
		{ 351, 355, 26 },
		{ 351, 356, 26 },
		{ 351, 357, 26 },
		{ 351, 358, 26 },
		{ 351, 359, 26 },
		{ 351, 360, 26 },
		{ 351, 361, 26 },
		{ 351, 362, 26 },
		{ 351, 363, 26 },
		{ 351, 364, 26 },
		{ 351, 365, 26 },
		{ 351, 366, 26 },
		{ 351, 367, 26 },
		{ 351, 368, 26 },
		{ 351, 369, 26 },
		{ 351, 370, 26 },
		{ 351, 371, 26 },
		{ 351, 372, 26 },
		{ 351, 373, 26 },
		{ 351, 374, 26 },
		{ 351, 375, 26 },
		{ 351, 376, 26 },
		{ 351, 377, 26 },
		{ 351, 378, 26 },
		{ 351, 379, 26 },
		{ 351, 380, 26 },
		{ 352, 329, 26 },
		{ 352, 330, 26 },
		{ 352, 331, 26 },
		{ 352, 332, 26 },
		{ 352, 333, 26 },
		{ 352, 334, 26 },
		{ 352, 335, 26 },
		{ 352, 336, 26 },
		{ 352, 337, 26 },
		{ 352, 338, 26 },
		{ 352, 339, 26 },
		{ 352, 340, 26 },
		{ 352, 341, 26 },
		{ 352, 342, 26 },
		{ 352, 343, 26 },
		{ 352, 344, 26 },
		{ 352, 345, 26 },
		{ 352, 346, 26 },
		{ 352, 347, 26 },
		{ 352, 348, 26 },
		{ 352, 349, 26 },
		{ 352, 350, 26 },
		{ 352, 351, 26 },
		{ 352, 352, 26 },
		{ 352, 353, 26 },
		{ 352, 354, 26 },
		{ 352, 355, 26 },
		{ 352, 356, 26 },
		{ 352, 357, 26 },
		{ 352, 358, 26 },
		{ 352, 359, 26 },
		{ 352, 360, 26 },
		{ 352, 361, 26 },
		{ 352, 362, 26 },
		{ 352, 363, 26 },
		{ 352, 364, 26 },
		{ 352, 365, 26 },
		{ 352, 366, 26 },
		{ 352, 367, 26 },
		{ 352, 368, 26 },
		{ 352, 369, 26 },
		{ 352, 370, 26 },
		{ 352, 371, 26 },
		{ 352, 372, 26 },
		{ 352, 373, 26 },
		{ 352, 374, 26 },
		{ 352, 375, 26 },
		{ 352, 376, 26 },
		{ 352, 377, 26 },
		{ 352, 378, 26 },
		{ 352, 379, 26 },
		{ 352, 380, 26 },
		{ 352, 381, 26 },
		{ 353, 330, 26 },
		{ 353, 331, 26 },
		{ 353, 332, 26 },
		{ 353, 333, 26 },
		{ 353, 334, 26 },
		{ 353, 335, 26 },
		{ 353, 336, 26 },
		{ 353, 337, 26 },
		{ 353, 338, 26 },
		{ 353, 339, 26 },
		{ 353, 340, 26 },
		{ 353, 341, 26 },
		{ 353, 342, 26 },
		{ 353, 343, 26 },
		{ 353, 344, 26 },
		{ 353, 345, 26 },
		{ 353, 346, 26 },
		{ 353, 347, 26 },
		{ 353, 348, 26 },
		{ 353, 349, 26 },
		{ 353, 350, 26 },
		{ 353, 351, 26 },
		{ 353, 352, 26 },
		{ 353, 353, 26 },
		{ 353, 354, 26 },
		{ 353, 355, 26 },
		{ 353, 356, 26 },
		{ 353, 357, 26 },
		{ 353, 358, 26 },
		{ 353, 359, 26 },
		{ 353, 360, 26 },
		{ 353, 361, 26 },
		{ 353, 362, 26 },
		{ 353, 363, 26 },
		{ 353, 364, 26 },
		{ 353, 365, 26 },
		{ 353, 366, 26 },
		{ 353, 367, 26 },
		{ 353, 368, 26 },
		{ 353, 369, 26 },
		{ 353, 370, 26 },
		{ 353, 371, 26 },
		{ 353, 372, 26 },
		{ 353, 373, 26 },
		{ 353, 374, 26 },
		{ 353, 375, 26 },
		{ 353, 376, 26 },
		{ 353, 377, 26 },
		{ 353, 378, 26 },
		{ 353, 379, 26 },
		{ 353, 380, 26 },
		{ 353, 381, 26 },
		{ 354, 330, 26 },
		{ 354, 331, 26 },
		{ 354, 332, 26 },
		{ 354, 333, 26 },
		{ 354, 334, 26 },
		{ 354, 335, 26 },
		{ 354, 336, 26 },
		{ 354, 337, 26 },
		{ 354, 338, 26 },
		{ 354, 339, 26 },
		{ 354, 340, 26 },
		{ 354, 341, 26 },
		{ 354, 342, 26 },
		{ 354, 343, 26 },
		{ 354, 344, 26 },
		{ 354, 345, 26 },
		{ 354, 346, 26 },
		{ 354, 347, 26 },
		{ 354, 348, 26 },
		{ 354, 349, 26 },
		{ 354, 350, 26 },
		{ 354, 351, 26 },
		{ 354, 352, 26 },
		{ 354, 353, 26 },
		{ 354, 354, 26 },
		{ 354, 355, 26 },
		{ 354, 356, 26 },
		{ 354, 357, 26 },
		{ 354, 358, 26 },
		{ 354, 359, 26 },
		{ 354, 360, 26 },
		{ 354, 361, 26 },
		{ 354, 362, 26 },
		{ 354, 363, 26 },
		{ 354, 364, 26 },
		{ 354, 365, 26 },
		{ 354, 366, 26 },
		{ 354, 367, 26 },
		{ 354, 368, 26 },
		{ 354, 369, 26 },
		{ 354, 370, 26 },
		{ 354, 371, 26 },
		{ 354, 372, 26 },
		{ 354, 373, 26 },
		{ 354, 374, 26 },
		{ 354, 375, 26 },
		{ 354, 376, 26 },
		{ 354, 377, 26 },
		{ 354, 378, 26 },
		{ 354, 379, 26 },
		{ 354, 380, 26 },
		{ 354, 381, 26 },
		{ 354, 382, 26 },
		{ 355, 331, 26 },
		{ 355, 332, 26 },
		{ 355, 333, 26 },
		{ 355, 334, 26 },
		{ 355, 335, 26 },
		{ 355, 336, 26 },
		{ 355, 337, 26 },
		{ 355, 338, 26 },
		{ 355, 339, 26 },
		{ 355, 340, 26 },
		{ 355, 341, 26 },
		{ 355, 342, 26 },
		{ 355, 343, 26 },
		{ 355, 344, 26 },
		{ 355, 345, 26 },
		{ 355, 346, 26 },
		{ 355, 347, 26 },
		{ 355, 348, 26 },
		{ 355, 349, 26 },
		{ 355, 350, 26 },
		{ 355, 351, 26 },
		{ 355, 352, 26 },
		{ 355, 353, 26 },
		{ 355, 354, 26 },
		{ 355, 355, 26 },
		{ 355, 356, 26 },
		{ 355, 357, 26 },
		{ 355, 358, 26 },
		{ 355, 359, 26 },
		{ 355, 360, 26 },
		{ 355, 361, 26 },
		{ 355, 362, 26 },
		{ 355, 363, 26 },
		{ 355, 364, 26 },
		{ 355, 365, 26 },
		{ 355, 366, 26 },
		{ 355, 367, 26 },
		{ 355, 368, 26 },
		{ 355, 369, 26 },
		{ 355, 370, 26 },
		{ 355, 371, 26 },
		{ 355, 372, 26 },
		{ 355, 373, 26 },
		{ 355, 374, 26 },
		{ 355, 375, 26 },
		{ 355, 376, 26 },
		{ 355, 377, 26 },
		{ 355, 378, 26 },
		{ 355, 379, 26 },
		{ 355, 380, 26 },
		{ 355, 381, 26 },
		{ 355, 382, 26 },
		{ 356, 332, 26 },
		{ 356, 333, 26 },
		{ 356, 334, 26 },
		{ 356, 335, 26 },
		{ 356, 336, 26 },
		{ 356, 337, 26 },
		{ 356, 338, 26 },
		{ 356, 339, 26 },
		{ 356, 340, 26 },
		{ 356, 341, 26 },
		{ 356, 342, 26 },
		{ 356, 343, 26 },
		{ 356, 344, 26 },
		{ 356, 345, 26 },
		{ 356, 346, 26 },
		{ 356, 347, 26 },
		{ 356, 348, 26 },
		{ 356, 349, 26 },
		{ 356, 350, 26 },
		{ 356, 351, 26 },
		{ 356, 352, 26 },
		{ 356, 353, 26 },
		{ 356, 354, 26 },
		{ 356, 355, 26 },
		{ 356, 356, 26 },
		{ 356, 357, 26 },
		{ 356, 358, 26 },
		{ 356, 359, 26 },
		{ 356, 360, 26 },
		{ 356, 361, 26 },
		{ 356, 362, 26 },
		{ 356, 363, 26 },
		{ 356, 364, 26 },
		{ 356, 365, 26 },
		{ 356, 366, 26 },
		{ 356, 367, 26 },
		{ 356, 368, 26 },
		{ 356, 369, 26 },
		{ 356, 370, 26 },
		{ 356, 371, 26 },
		{ 356, 372, 26 },
		{ 356, 373, 26 },
		{ 356, 374, 26 },
		{ 356, 375, 26 },
		{ 356, 376, 26 },
		{ 356, 377, 26 },
		{ 356, 378, 26 },
		{ 356, 379, 26 },
		{ 356, 380, 26 },
		{ 356, 381, 26 },
		{ 356, 382, 26 },
		{ 357, 333, 26 },
		{ 357, 334, 26 },
		{ 357, 335, 26 },
		{ 357, 336, 26 },
		{ 357, 337, 26 },
		{ 357, 338, 26 },
		{ 357, 339, 26 },
		{ 357, 340, 26 },
		{ 357, 341, 26 },
		{ 357, 342, 26 },
		{ 357, 343, 26 },
		{ 357, 344, 26 },
		{ 357, 345, 26 },
		{ 357, 346, 26 },
		{ 357, 347, 26 },
		{ 357, 348, 26 },
		{ 357, 349, 26 },
		{ 357, 350, 26 },
		{ 357, 351, 26 },
		{ 357, 352, 26 },
		{ 357, 353, 26 },
		{ 357, 354, 26 },
		{ 357, 355, 26 },
		{ 357, 356, 26 },
		{ 357, 357, 26 },
		{ 357, 358, 26 },
		{ 357, 359, 26 },
		{ 357, 360, 26 },
		{ 357, 361, 26 },
		{ 357, 362, 26 },
		{ 357, 363, 26 },
		{ 357, 364, 26 },
		{ 357, 365, 26 },
		{ 357, 366, 26 },
		{ 357, 367, 26 },
		{ 357, 368, 26 },
		{ 357, 369, 26 },
		{ 357, 370, 26 },
		{ 357, 371, 26 },
		{ 357, 372, 26 },
		{ 357, 373, 26 },
		{ 357, 374, 26 },
		{ 357, 375, 26 },
		{ 357, 376, 26 },
		{ 357, 377, 26 },
		{ 357, 378, 26 },
		{ 357, 379, 26 },
		{ 357, 380, 26 },
		{ 357, 381, 26 },
		{ 357, 382, 26 },
		{ 358, 333, 26 },
		{ 358, 334, 26 },
		{ 358, 335, 26 },
		{ 358, 336, 26 },
		{ 358, 337, 26 },
		{ 358, 338, 26 },
		{ 358, 339, 26 },
		{ 358, 340, 26 },
		{ 358, 341, 26 },
		{ 358, 342, 26 },
		{ 358, 343, 26 },
		{ 358, 344, 26 },
		{ 358, 345, 26 },
		{ 358, 346, 26 },
		{ 358, 347, 26 },
		{ 358, 348, 26 },
		{ 358, 349, 26 },
		{ 358, 350, 26 },
		{ 358, 351, 26 },
		{ 358, 352, 26 },
		{ 358, 353, 26 },
		{ 358, 354, 26 },
		{ 358, 355, 26 },
		{ 358, 356, 26 },
		{ 358, 357, 26 },
		{ 358, 358, 26 },
		{ 358, 359, 26 },
		{ 358, 360, 26 },
		{ 358, 361, 26 },
		{ 358, 362, 26 },
		{ 358, 363, 26 },
		{ 358, 364, 26 },
		{ 358, 365, 26 },
		{ 358, 366, 26 },
		{ 358, 367, 26 },
		{ 358, 368, 26 },
		{ 358, 369, 26 },
		{ 358, 370, 26 },
		{ 358, 371, 26 },
		{ 358, 372, 26 },
		{ 358, 373, 26 },
		{ 358, 374, 26 },
		{ 358, 375, 26 },
		{ 358, 376, 26 },
		{ 358, 377, 26 },
		{ 358, 378, 26 },
		{ 358, 379, 26 },
		{ 358, 380, 26 },
		{ 358, 381, 26 },
		{ 358, 382, 26 },
		{ 359, 334, 26 },
		{ 359, 335, 26 },
		{ 359, 336, 26 },
		{ 359, 337, 26 },
		{ 359, 338, 26 },
		{ 359, 339, 26 },
		{ 359, 340, 26 },
		{ 359, 341, 26 },
		{ 359, 342, 26 },
		{ 359, 343, 26 },
		{ 359, 344, 26 },
		{ 359, 345, 26 },
		{ 359, 346, 26 },
		{ 359, 347, 26 },
		{ 359, 348, 26 },
		{ 359, 349, 26 },
		{ 359, 350, 26 },
		{ 359, 351, 26 },
		{ 359, 352, 26 },
		{ 359, 353, 26 },
		{ 359, 354, 26 },
		{ 359, 355, 26 },
		{ 359, 356, 26 },
		{ 359, 357, 26 },
		{ 359, 358, 26 },
		{ 359, 359, 26 },
		{ 359, 360, 26 },
		{ 359, 361, 26 },
		{ 359, 362, 26 },
		{ 359, 363, 26 },
		{ 359, 364, 26 },
		{ 359, 365, 26 },
		{ 359, 366, 26 },
		{ 359, 367, 26 },
		{ 359, 368, 26 },
		{ 359, 369, 26 },
		{ 359, 370, 26 },
		{ 359, 371, 26 },
		{ 359, 372, 26 },
		{ 359, 373, 26 },
		{ 359, 374, 26 },
		{ 359, 375, 26 },
		{ 359, 376, 26 },
		{ 359, 377, 26 },
		{ 359, 378, 26 },
		{ 359, 379, 26 },
		{ 359, 380, 26 },
		{ 359, 381, 26 },
		{ 359, 382, 26 },
		{ 359, 383, 26 },
		{ 360, 335, 26 },
		{ 360, 336, 26 },
		{ 360, 337, 26 },
		{ 360, 338, 26 },
		{ 360, 339, 26 },
		{ 360, 340, 26 },
		{ 360, 341, 26 },
		{ 360, 342, 26 },
		{ 360, 343, 26 },
		{ 360, 344, 26 },
		{ 360, 345, 26 },
		{ 360, 346, 26 },
		{ 360, 347, 26 },
		{ 360, 348, 26 },
		{ 360, 349, 26 },
		{ 360, 350, 26 },
		{ 360, 351, 26 },
		{ 360, 352, 26 },
		{ 360, 353, 26 },
		{ 360, 354, 26 },
		{ 360, 355, 26 },
		{ 360, 356, 26 },
		{ 360, 357, 26 },
		{ 360, 358, 26 },
		{ 360, 359, 26 },
		{ 360, 360, 26 },
		{ 360, 361, 26 },
		{ 360, 362, 26 },
		{ 360, 363, 26 },
		{ 360, 364, 26 },
		{ 360, 365, 26 },
		{ 360, 366, 26 },
		{ 360, 367, 26 },
		{ 360, 368, 26 },
		{ 360, 369, 26 },
		{ 360, 370, 26 },
		{ 360, 371, 26 },
		{ 360, 372, 26 },
		{ 360, 373, 26 },
		{ 360, 374, 26 },
		{ 360, 375, 26 },
		{ 360, 376, 26 },
		{ 360, 377, 26 },
		{ 360, 378, 26 },
		{ 360, 379, 26 },
		{ 360, 380, 26 },
		{ 360, 381, 26 },
		{ 360, 382, 26 },
		{ 360, 383, 26 },
		{ 361, 336, 26 },
		{ 361, 337, 26 },
		{ 361, 338, 26 },
		{ 361, 339, 26 },
		{ 361, 340, 26 },
		{ 361, 341, 26 },
		{ 361, 342, 26 },
		{ 361, 343, 26 },
		{ 361, 344, 26 },
		{ 361, 345, 26 },
		{ 361, 346, 26 },
		{ 361, 347, 26 },
		{ 361, 348, 26 },
		{ 361, 349, 26 },
		{ 361, 350, 26 },
		{ 361, 351, 26 },
		{ 361, 352, 26 },
		{ 361, 353, 26 },
		{ 361, 354, 26 },
		{ 361, 355, 26 },
		{ 361, 356, 26 },
		{ 361, 357, 26 },
		{ 361, 358, 26 },
		{ 361, 359, 26 },
		{ 361, 360, 26 },
		{ 361, 361, 26 },
		{ 361, 362, 26 },
		{ 361, 363, 26 },
		{ 361, 364, 26 },
		{ 361, 365, 26 },
		{ 361, 366, 26 },
		{ 361, 367, 26 },
		{ 361, 368, 26 },
		{ 361, 369, 26 },
		{ 361, 370, 26 },
		{ 361, 371, 26 },
		{ 361, 372, 26 },
		{ 361, 373, 26 },
		{ 361, 374, 26 },
		{ 361, 375, 26 },
		{ 361, 376, 26 },
		{ 361, 377, 26 },
		{ 361, 378, 26 },
		{ 361, 379, 26 },
		{ 361, 380, 26 },
		{ 361, 381, 26 },
		{ 361, 382, 26 },
		{ 361, 383, 26 },
		{ 362, 336, 26 },
		{ 362, 337, 26 },
		{ 362, 338, 26 },
		{ 362, 339, 26 },
		{ 362, 340, 26 },
		{ 362, 341, 26 },
		{ 362, 342, 26 },
		{ 362, 343, 26 },
		{ 362, 344, 26 },
		{ 362, 345, 26 },
		{ 362, 346, 26 },
		{ 362, 347, 26 },
		{ 362, 348, 26 },
		{ 362, 349, 26 },
		{ 362, 350, 26 },
		{ 362, 351, 26 },
		{ 362, 352, 26 },
		{ 362, 353, 26 },
		{ 362, 354, 26 },
		{ 362, 355, 26 },
		{ 362, 356, 26 },
		{ 362, 357, 26 },
		{ 362, 358, 26 },
		{ 362, 359, 26 },
		{ 362, 360, 26 },
		{ 362, 361, 26 },
		{ 362, 362, 26 },
		{ 362, 363, 26 },
		{ 362, 364, 26 },
		{ 362, 365, 26 },
		{ 362, 366, 26 },
		{ 362, 367, 26 },
		{ 362, 368, 26 },
		{ 362, 369, 26 },
		{ 362, 370, 26 },
		{ 362, 371, 26 },
		{ 362, 372, 26 },
		{ 362, 373, 26 },
		{ 362, 374, 26 },
		{ 362, 375, 26 },
		{ 362, 376, 26 },
		{ 362, 377, 26 },
		{ 362, 378, 26 },
		{ 362, 379, 26 },
		{ 362, 380, 26 },
		{ 362, 381, 26 },
		{ 362, 382, 26 },
		{ 362, 383, 26 },
		{ 363, 337, 26 },
		{ 363, 338, 26 },
		{ 363, 339, 26 },
		{ 363, 340, 26 },
		{ 363, 341, 26 },
		{ 363, 342, 26 },
		{ 363, 343, 26 },
		{ 363, 344, 26 },
		{ 363, 345, 26 },
		{ 363, 346, 26 },
		{ 363, 347, 26 },
		{ 363, 348, 26 },
		{ 363, 349, 26 },
		{ 363, 350, 26 },
		{ 363, 351, 26 },
		{ 363, 352, 26 },
		{ 363, 353, 26 },
		{ 363, 354, 26 },
		{ 363, 355, 26 },
		{ 363, 356, 26 },
		{ 363, 357, 26 },
		{ 363, 358, 26 },
		{ 363, 359, 26 },
		{ 363, 360, 26 },
		{ 363, 361, 26 },
		{ 363, 362, 26 },
		{ 363, 363, 26 },
		{ 363, 364, 26 },
		{ 363, 365, 26 },
		{ 363, 366, 26 },
		{ 363, 367, 26 },
		{ 363, 368, 26 },
		{ 363, 369, 26 },
		{ 363, 370, 26 },
		{ 363, 371, 26 },
		{ 363, 372, 26 },
		{ 363, 373, 26 },
		{ 363, 374, 26 },
		{ 363, 375, 26 },
		{ 363, 376, 26 },
		{ 363, 377, 26 },
		{ 363, 378, 26 },
		{ 363, 379, 26 },
		{ 363, 380, 26 },
		{ 363, 381, 26 },
		{ 363, 382, 26 },
		{ 363, 383, 26 },
		{ 364, 338, 26 },
		{ 364, 339, 26 },
		{ 364, 340, 26 },
		{ 364, 341, 26 },
		{ 364, 342, 26 },
		{ 364, 343, 26 },
		{ 364, 344, 26 },
		{ 364, 345, 26 },
		{ 364, 346, 26 },
		{ 364, 347, 26 },
		{ 364, 348, 26 },
		{ 364, 349, 26 },
		{ 364, 350, 26 },
		{ 364, 351, 26 },
		{ 364, 352, 26 },
		{ 364, 353, 26 },
		{ 364, 354, 26 },
		{ 364, 355, 26 },
		{ 364, 356, 26 },
		{ 364, 357, 26 },
		{ 364, 358, 26 },
		{ 364, 359, 26 },
		{ 364, 360, 26 },
		{ 364, 361, 26 },
		{ 364, 362, 26 },
		{ 364, 363, 26 },
		{ 364, 364, 26 },
		{ 364, 365, 26 },
		{ 364, 366, 26 },
		{ 364, 367, 26 },
		{ 364, 368, 26 },
		{ 364, 369, 26 },
		{ 364, 370, 26 },
		{ 364, 371, 26 },
		{ 364, 372, 26 },
		{ 364, 373, 26 },
		{ 364, 374, 26 },
		{ 364, 375, 26 },
		{ 364, 376, 26 },
		{ 364, 377, 26 },
		{ 364, 378, 26 },
		{ 364, 379, 26 },
		{ 364, 380, 26 },
		{ 364, 381, 26 },
		{ 364, 382, 26 },
		{ 364, 383, 26 },
		{ 365, 339, 26 },
		{ 365, 340, 26 },
		{ 365, 341, 26 },
		{ 365, 342, 26 },
		{ 365, 343, 26 },
		{ 365, 344, 26 },
		{ 365, 345, 26 },
		{ 365, 346, 26 },
		{ 365, 347, 26 },
		{ 365, 348, 26 },
		{ 365, 349, 26 },
		{ 365, 350, 26 },
		{ 365, 351, 26 },
		{ 365, 352, 26 },
		{ 365, 353, 26 },
		{ 365, 354, 26 },
		{ 365, 355, 26 },
		{ 365, 356, 26 },
		{ 365, 357, 26 },
		{ 365, 358, 26 },
		{ 365, 359, 26 },
		{ 365, 360, 26 },
		{ 365, 361, 26 },
		{ 365, 362, 26 },
		{ 365, 363, 26 },
		{ 365, 364, 26 },
		{ 365, 365, 26 },
		{ 365, 366, 26 },
		{ 365, 367, 26 },
		{ 365, 368, 26 },
		{ 365, 369, 26 },
		{ 365, 370, 26 },
		{ 365, 371, 26 },
		{ 365, 372, 26 },
		{ 365, 373, 26 },
		{ 365, 374, 26 },
		{ 365, 375, 26 },
		{ 365, 376, 26 },
		{ 365, 377, 26 },
		{ 365, 378, 26 },
		{ 365, 379, 26 },
		{ 365, 380, 26 },
		{ 365, 381, 26 },
		{ 365, 382, 26 },
		{ 365, 383, 26 },
		{ 366, 339, 26 },
		{ 366, 340, 26 },
		{ 366, 341, 26 },
		{ 366, 342, 26 },
		{ 366, 343, 26 },
		{ 366, 344, 26 },
		{ 366, 345, 26 },
		{ 366, 346, 26 },
		{ 366, 347, 26 },
		{ 366, 348, 26 },
		{ 366, 349, 26 },
		{ 366, 350, 26 },
		{ 366, 351, 26 },
		{ 366, 352, 26 },
		{ 366, 353, 26 },
		{ 366, 354, 26 },
		{ 366, 355, 26 },
		{ 366, 356, 26 },
		{ 366, 357, 26 },
		{ 366, 358, 26 },
		{ 366, 359, 26 },
		{ 366, 360, 26 },
		{ 366, 361, 26 },
		{ 366, 362, 26 },
		{ 366, 363, 26 },
		{ 366, 364, 26 },
		{ 366, 365, 26 },
		{ 366, 366, 26 },
		{ 366, 367, 26 },
		{ 366, 368, 26 },
		{ 366, 369, 26 },
		{ 366, 370, 26 },
		{ 366, 371, 26 },
		{ 366, 372, 26 },
		{ 366, 373, 26 },
		{ 366, 374, 26 },
		{ 366, 375, 26 },
		{ 366, 376, 26 },
		{ 366, 377, 26 },
		{ 366, 378, 26 },
		{ 366, 379, 26 },
		{ 366, 380, 26 },
		{ 366, 381, 26 },
		{ 366, 382, 26 },
		{ 366, 383, 26 },
		{ 367, 340, 26 },
		{ 367, 341, 26 },
		{ 367, 342, 26 },
		{ 367, 343, 26 },
		{ 367, 344, 26 },
		{ 367, 345, 26 },
		{ 367, 346, 26 },
		{ 367, 347, 26 },
		{ 367, 348, 26 },
		{ 367, 349, 26 },
		{ 367, 350, 26 },
		{ 367, 351, 26 },
		{ 367, 352, 26 },
		{ 367, 353, 26 },
		{ 367, 354, 26 },
		{ 367, 355, 26 },
		{ 367, 356, 26 },
		{ 367, 357, 26 },
		{ 367, 358, 26 },
		{ 367, 359, 26 },
		{ 367, 360, 26 },
		{ 367, 361, 26 },
		{ 367, 362, 26 },
		{ 367, 363, 26 },
		{ 367, 364, 26 },
		{ 367, 365, 26 },
		{ 367, 366, 26 },
		{ 367, 367, 26 },
		{ 367, 368, 26 },
		{ 367, 369, 26 },
		{ 367, 370, 26 },
		{ 367, 371, 26 },
		{ 367, 372, 26 },
		{ 367, 373, 26 },
		{ 367, 374, 26 },
		{ 367, 375, 26 },
		{ 367, 376, 26 },
		{ 367, 377, 26 },
		{ 367, 378, 26 },
		{ 367, 379, 26 },
		{ 367, 380, 26 },
		{ 367, 381, 26 },
		{ 367, 382, 26 },
		{ 367, 383, 26 },
		{ 368, 341, 26 },
		{ 368, 342, 26 },
		{ 368, 343, 26 },
		{ 368, 344, 26 },
		{ 368, 345, 26 },
		{ 368, 346, 26 },
		{ 368, 347, 26 },
		{ 368, 348, 26 },
		{ 368, 349, 26 },
		{ 368, 350, 26 },
		{ 368, 351, 26 },
		{ 368, 352, 26 },
		{ 368, 353, 26 },
		{ 368, 354, 26 },
		{ 368, 355, 26 },
		{ 368, 356, 26 },
		{ 368, 357, 26 },
		{ 368, 358, 26 },
		{ 368, 359, 26 },
		{ 368, 360, 26 },
		{ 368, 361, 26 },
		{ 368, 362, 26 },
		{ 368, 363, 26 },
		{ 368, 364, 26 },
		{ 368, 365, 26 },
		{ 368, 366, 26 },
		{ 368, 367, 26 },
		{ 368, 368, 26 },
		{ 368, 369, 26 },
		{ 368, 370, 26 },
		{ 368, 371, 26 },
		{ 368, 372, 26 },
		{ 368, 373, 26 },
		{ 368, 374, 26 },
		{ 368, 375, 26 },
		{ 368, 376, 26 },
		{ 368, 377, 26 },
		{ 368, 378, 26 },
		{ 368, 379, 26 },
		{ 368, 380, 26 },
		{ 368, 381, 26 },
		{ 368, 382, 26 },
		{ 368, 383, 26 },
		{ 369, 342, 26 },
		{ 369, 343, 26 },
		{ 369, 344, 26 },
		{ 369, 345, 26 },
		{ 369, 346, 26 },
		{ 369, 347, 26 },
		{ 369, 348, 26 },
		{ 369, 349, 26 },
		{ 369, 350, 26 },
		{ 369, 351, 26 },
		{ 369, 352, 26 },
		{ 369, 353, 26 },
		{ 369, 354, 26 },
		{ 369, 355, 26 },
		{ 369, 356, 26 },
		{ 369, 357, 26 },
		{ 369, 358, 26 },
		{ 369, 359, 26 },
		{ 369, 360, 26 },
		{ 369, 361, 26 },
		{ 369, 362, 26 },
		{ 369, 363, 26 },
		{ 369, 364, 26 },
		{ 369, 365, 26 },
		{ 369, 366, 26 },
		{ 369, 367, 26 },
		{ 369, 368, 26 },
		{ 369, 369, 26 },
		{ 369, 370, 26 },
		{ 369, 371, 26 },
		{ 369, 372, 26 },
		{ 369, 373, 26 },
		{ 369, 374, 26 },
		{ 369, 375, 26 },
		{ 369, 376, 26 },
		{ 369, 377, 26 },
		{ 369, 378, 26 },
		{ 369, 379, 26 },
		{ 369, 380, 26 },
		{ 369, 381, 26 },
		{ 369, 382, 26 },
		{ 369, 383, 26 },
		{ 370, 342, 26 },
		{ 370, 343, 26 },
		{ 370, 344, 26 },
		{ 370, 345, 26 },
		{ 370, 346, 26 },
		{ 370, 347, 26 },
		{ 370, 348, 26 },
		{ 370, 349, 26 },
		{ 370, 350, 26 },
		{ 370, 351, 26 },
		{ 370, 352, 26 },
		{ 370, 353, 26 },
		{ 370, 354, 26 },
		{ 370, 355, 26 },
		{ 370, 356, 26 },
		{ 370, 357, 26 },
		{ 370, 358, 26 },
		{ 370, 359, 26 },
		{ 370, 360, 26 },
		{ 370, 361, 26 },
		{ 370, 362, 26 },
		{ 370, 363, 26 },
		{ 370, 364, 26 },
		{ 370, 365, 26 },
		{ 370, 366, 26 },
		{ 370, 367, 26 },
		{ 370, 368, 26 },
		{ 370, 369, 26 },
		{ 370, 370, 26 },
		{ 370, 371, 26 },
		{ 370, 372, 26 },
		{ 370, 373, 26 },
		{ 370, 374, 26 },
		{ 370, 375, 26 },
		{ 370, 376, 26 },
		{ 370, 377, 26 },
		{ 370, 378, 26 },
		{ 370, 379, 26 },
		{ 370, 380, 26 },
		{ 370, 381, 26 },
		{ 370, 382, 26 },
		{ 370, 383, 26 },
		{ 371, 343, 26 },
		{ 371, 344, 26 },
		{ 371, 345, 26 },
		{ 371, 346, 26 },
		{ 371, 347, 26 },
		{ 371, 348, 26 },
		{ 371, 349, 26 },
		{ 371, 350, 26 },
		{ 371, 351, 26 },
		{ 371, 352, 26 },
		{ 371, 353, 26 },
		{ 371, 354, 26 },
		{ 371, 355, 26 },
		{ 371, 356, 26 },
		{ 371, 357, 26 },
		{ 371, 358, 26 },
		{ 371, 359, 26 },
		{ 371, 360, 26 },
		{ 371, 361, 26 },
		{ 371, 362, 26 },
		{ 371, 363, 26 },
		{ 371, 364, 26 },
		{ 371, 365, 26 },
		{ 371, 366, 26 },
		{ 371, 367, 26 },
		{ 371, 368, 26 },
		{ 371, 369, 26 },
		{ 371, 370, 26 },
		{ 371, 371, 26 },
		{ 371, 372, 26 },
		{ 371, 373, 26 },
		{ 371, 374, 26 },
		{ 371, 375, 26 },
		{ 371, 376, 26 },
		{ 371, 377, 26 },
		{ 371, 378, 26 },
		{ 371, 379, 26 },
		{ 371, 380, 26 },
		{ 371, 381, 26 },
		{ 371, 382, 26 },
		{ 371, 383, 26 },
		{ 372, 344, 26 },
		{ 372, 345, 26 },
		{ 372, 346, 26 },
		{ 372, 347, 26 },
		{ 372, 348, 26 },
		{ 372, 349, 26 },
		{ 372, 350, 26 },
		{ 372, 351, 26 },
		{ 372, 352, 26 },
		{ 372, 353, 26 },
		{ 372, 354, 26 },
		{ 372, 355, 26 },
		{ 372, 356, 26 },
		{ 372, 357, 26 },
		{ 372, 358, 26 },
		{ 372, 359, 26 },
		{ 372, 360, 26 },
		{ 372, 361, 26 },
		{ 372, 362, 26 },
		{ 372, 363, 26 },
		{ 372, 364, 26 },
		{ 372, 365, 26 },
		{ 372, 366, 26 },
		{ 372, 367, 26 },
		{ 372, 368, 26 },
		{ 372, 369, 26 },
		{ 372, 370, 26 },
		{ 372, 371, 26 },
		{ 372, 372, 26 },
		{ 372, 373, 26 },
		{ 372, 374, 26 },
		{ 372, 375, 26 },
		{ 372, 376, 26 },
		{ 372, 377, 26 },
		{ 372, 378, 26 },
		{ 372, 379, 26 },
		{ 372, 380, 26 },
		{ 372, 381, 26 },
		{ 372, 382, 26 },
		{ 372, 383, 26 },
		{ 373, 345, 26 },
		{ 373, 346, 26 },
		{ 373, 347, 26 },
		{ 373, 348, 26 },
		{ 373, 349, 26 },
		{ 373, 350, 26 },
		{ 373, 351, 26 },
		{ 373, 352, 26 },
		{ 373, 353, 26 },
		{ 373, 354, 26 },
		{ 373, 355, 26 },
		{ 373, 356, 26 },
		{ 373, 357, 26 },
		{ 373, 358, 26 },
		{ 373, 359, 26 },
		{ 373, 360, 26 },
		{ 373, 361, 26 },
		{ 373, 362, 26 },
		{ 373, 363, 26 },
		{ 373, 364, 26 },
		{ 373, 365, 26 },
		{ 373, 366, 26 },
		{ 373, 367, 26 },
		{ 373, 368, 26 },
		{ 373, 369, 26 },
		{ 373, 370, 26 },
		{ 373, 371, 26 },
		{ 373, 372, 26 },
		{ 373, 373, 26 },
		{ 373, 374, 26 },
		{ 373, 375, 26 },
		{ 373, 376, 26 },
		{ 373, 377, 26 },
		{ 373, 378, 26 },
		{ 373, 379, 26 },
		{ 373, 380, 26 },
		{ 373, 381, 26 },
		{ 373, 382, 26 },
		{ 373, 383, 26 },
		{ 374, 345, 26 },
		{ 374, 346, 26 },
		{ 374, 347, 26 },
		{ 374, 348, 26 },
		{ 374, 349, 26 },
		{ 374, 350, 26 },
		{ 374, 351, 26 },
		{ 374, 352, 26 },
		{ 374, 353, 26 },
		{ 374, 354, 26 },
		{ 374, 355, 26 },
		{ 374, 356, 26 },
		{ 374, 357, 26 },
		{ 374, 358, 26 },
		{ 374, 359, 26 },
		{ 374, 360, 26 },
		{ 374, 361, 26 },
		{ 374, 362, 26 },
		{ 374, 363, 26 },
		{ 374, 364, 26 },
		{ 374, 365, 26 },
		{ 374, 366, 26 },
		{ 374, 367, 26 },
		{ 374, 368, 26 },
		{ 374, 369, 26 },
		{ 374, 370, 26 },
		{ 374, 371, 26 },
		{ 374, 372, 26 },
		{ 374, 373, 26 },
		{ 374, 374, 26 },
		{ 374, 375, 26 },
		{ 374, 376, 26 },
		{ 374, 377, 26 },
		{ 374, 378, 26 },
		{ 374, 379, 26 },
		{ 374, 380, 26 },
		{ 374, 381, 26 },
		{ 374, 382, 26 },
		{ 374, 383, 26 },
		{ 375, 346, 26 },
		{ 375, 347, 26 },
		{ 375, 348, 26 },
		{ 375, 349, 26 },
		{ 375, 350, 26 },
		{ 375, 351, 26 },
		{ 375, 352, 26 },
		{ 375, 353, 26 },
		{ 375, 354, 26 },
		{ 375, 355, 26 },
		{ 375, 356, 26 },
		{ 375, 357, 26 },
		{ 375, 358, 26 },
		{ 375, 359, 26 },
		{ 375, 360, 26 },
		{ 375, 361, 26 },
		{ 375, 362, 26 },
		{ 375, 363, 26 },
		{ 375, 364, 26 },
		{ 375, 365, 26 },
		{ 375, 366, 26 },
		{ 375, 367, 26 },
		{ 375, 368, 26 },
		{ 375, 369, 26 },
		{ 375, 370, 26 },
		{ 375, 371, 26 },
		{ 375, 372, 26 },
		{ 375, 373, 26 },
		{ 375, 374, 26 },
		{ 375, 375, 26 },
		{ 375, 376, 26 },
		{ 375, 377, 26 },
		{ 375, 378, 26 },
		{ 375, 379, 26 },
		{ 375, 380, 26 },
		{ 375, 381, 26 },
		{ 375, 382, 26 },
		{ 375, 383, 26 },
		{ 376, 347, 26 },
		{ 376, 348, 26 },
		{ 376, 349, 26 },
		{ 376, 350, 26 },
		{ 376, 351, 26 },
		{ 376, 352, 26 },
		{ 376, 353, 26 },
		{ 376, 354, 26 },
		{ 376, 355, 26 },
		{ 376, 356, 26 },
		{ 376, 357, 26 },
		{ 376, 358, 26 },
		{ 376, 359, 26 },
		{ 376, 360, 26 },
		{ 376, 361, 26 },
		{ 376, 362, 26 },
		{ 376, 363, 26 },
		{ 376, 364, 26 },
		{ 376, 365, 26 },
		{ 376, 366, 26 },
		{ 376, 367, 26 },
		{ 376, 368, 26 },
		{ 376, 369, 26 },
		{ 376, 370, 26 },
		{ 376, 371, 26 },
		{ 376, 372, 26 },
		{ 376, 373, 26 },
		{ 376, 374, 26 },
		{ 376, 375, 26 },
		{ 376, 376, 26 },
		{ 376, 377, 26 },
		{ 376, 378, 26 },
		{ 376, 379, 26 },
		{ 376, 380, 26 },
		{ 376, 381, 26 },
		{ 376, 382, 26 },
		{ 376, 383, 26 },
		{ 377, 348, 26 },
		{ 377, 349, 26 },
		{ 377, 350, 26 },
		{ 377, 351, 26 },
		{ 377, 352, 26 },
		{ 377, 353, 26 },
		{ 377, 354, 26 },
		{ 377, 355, 26 },
		{ 377, 356, 26 },
		{ 377, 357, 26 },
		{ 377, 358, 26 },
		{ 377, 359, 26 },
		{ 377, 360, 26 },
		{ 377, 361, 26 },
		{ 377, 362, 26 },
		{ 377, 363, 26 },
		{ 377, 364, 26 },
		{ 377, 365, 26 },
		{ 377, 366, 26 },
		{ 377, 367, 26 },
		{ 377, 368, 26 },
		{ 377, 369, 26 },
		{ 377, 370, 26 },
		{ 377, 371, 26 },
		{ 377, 372, 26 },
		{ 377, 373, 26 },
		{ 377, 374, 26 },
		{ 377, 375, 26 },
		{ 377, 376, 26 },
		{ 377, 377, 26 },
		{ 377, 378, 26 },
		{ 377, 379, 26 },
		{ 377, 380, 26 },
		{ 377, 381, 26 },
		{ 377, 382, 26 },
		{ 377, 383, 26 },
		{ 378, 349, 26 },
		{ 378, 350, 26 },
		{ 378, 351, 26 },
		{ 378, 352, 26 },
		{ 378, 353, 26 },
		{ 378, 354, 26 },
		{ 378, 355, 26 },
		{ 378, 356, 26 },
		{ 378, 357, 26 },
		{ 378, 358, 26 },
		{ 378, 359, 26 },
		{ 378, 360, 26 },
		{ 378, 361, 26 },
		{ 378, 362, 26 },
		{ 378, 363, 26 },
		{ 378, 364, 26 },
		{ 378, 365, 26 },
		{ 378, 366, 26 },
		{ 378, 367, 26 },
		{ 378, 368, 26 },
		{ 378, 369, 26 },
		{ 378, 370, 26 },
		{ 378, 371, 26 },
		{ 378, 372, 26 },
		{ 378, 373, 26 },
		{ 378, 374, 26 },
		{ 378, 375, 26 },
		{ 378, 376, 26 },
		{ 378, 377, 26 },
		{ 378, 378, 26 },
		{ 378, 379, 26 },
		{ 378, 380, 26 },
		{ 378, 381, 26 },
		{ 378, 382, 26 },
		{ 378, 383, 26 },
		{ 379, 350, 26 },
		{ 379, 351, 26 },
		{ 379, 352, 26 },
		{ 379, 353, 26 },
		{ 379, 354, 26 },
		{ 379, 355, 26 },
		{ 379, 356, 26 },
		{ 379, 357, 26 },
		{ 379, 358, 26 },
		{ 379, 359, 26 },
		{ 379, 360, 26 },
		{ 379, 361, 26 },
		{ 379, 362, 26 },
		{ 379, 363, 26 },
		{ 379, 364, 26 },
		{ 379, 365, 26 },
		{ 379, 366, 26 },
		{ 379, 367, 26 },
		{ 379, 368, 26 },
		{ 379, 369, 26 },
		{ 379, 370, 26 },
		{ 379, 371, 26 },
		{ 379, 372, 26 },
		{ 379, 373, 26 },
		{ 379, 374, 26 },
		{ 379, 375, 26 },
		{ 379, 376, 26 },
		{ 379, 377, 26 },
		{ 379, 378, 26 },
		{ 379, 379, 26 },
		{ 379, 380, 26 },
		{ 379, 381, 26 },
		{ 379, 382, 26 },
		{ 379, 383, 26 },
		{ 380, 351, 26 },
		{ 380, 352, 26 },
		{ 380, 353, 26 },
		{ 380, 354, 26 },
		{ 380, 355, 26 },
		{ 380, 356, 26 },
		{ 380, 357, 26 },
		{ 380, 358, 26 },
		{ 380, 359, 26 },
		{ 380, 360, 26 },
		{ 380, 361, 26 },
		{ 380, 362, 26 },
		{ 380, 363, 26 },
		{ 380, 364, 26 },
		{ 380, 365, 26 },
		{ 380, 366, 26 },
		{ 380, 367, 26 },
		{ 380, 368, 26 },
		{ 380, 369, 26 },
		{ 380, 370, 26 },
		{ 380, 371, 26 },
		{ 380, 372, 26 },
		{ 380, 373, 26 },
		{ 380, 374, 26 },
		{ 380, 375, 26 },
		{ 380, 376, 26 },
		{ 380, 377, 26 },
		{ 380, 378, 26 },
		{ 380, 379, 26 },
		{ 380, 380, 26 },
		{ 380, 381, 26 },
		{ 380, 382, 26 },
		{ 380, 383, 26 },
		{ 381, 352, 26 },
		{ 381, 353, 26 },
		{ 381, 354, 26 },
		{ 381, 355, 26 },
		{ 381, 356, 26 },
		{ 381, 357, 26 },
		{ 381, 358, 26 },
		{ 381, 359, 26 },
		{ 381, 360, 26 },
		{ 381, 361, 26 },
		{ 381, 362, 26 },
		{ 381, 363, 26 },
		{ 381, 364, 26 },
		{ 381, 365, 26 },
		{ 381, 366, 26 },
		{ 381, 367, 26 },
		{ 381, 368, 26 },
		{ 381, 369, 26 },
		{ 381, 370, 26 },
		{ 381, 371, 26 },
		{ 381, 372, 26 },
		{ 381, 373, 26 },
		{ 381, 374, 26 },
		{ 381, 375, 26 },
		{ 381, 376, 26 },
		{ 381, 377, 26 },
		{ 381, 378, 26 },
		{ 381, 379, 26 },
		{ 381, 380, 26 },
		{ 381, 381, 26 },
		{ 381, 382, 26 },
		{ 381, 383, 26 },
		{ 382, 354, 26 },
		{ 382, 355, 26 },
		{ 382, 356, 26 },
		{ 382, 357, 26 },
		{ 382, 358, 26 },
		{ 382, 359, 26 },
		{ 382, 360, 26 },
		{ 382, 361, 26 },
		{ 382, 362, 26 },
		{ 382, 363, 26 },
		{ 382, 364, 26 },
		{ 382, 365, 26 },
		{ 382, 366, 26 },
		{ 382, 367, 26 },
		{ 382, 368, 26 },
		{ 382, 369, 26 },
		{ 382, 370, 26 },
		{ 382, 371, 26 },
		{ 382, 372, 26 },
		{ 382, 373, 26 },
		{ 382, 374, 26 },
		{ 382, 375, 26 },
		{ 382, 376, 26 },
		{ 382, 377, 26 },
		{ 382, 378, 26 },
		{ 382, 379, 26 },
		{ 382, 380, 26 },
		{ 382, 381, 26 },
		{ 382, 382, 26 },
		{ 382, 383, 26 },
		{ 383, 359, 26 },
		{ 383, 360, 26 },
		{ 383, 361, 26 },
		{ 383, 362, 26 },
		{ 383, 363, 26 },
		{ 383, 364, 26 },
		{ 383, 365, 26 },
		{ 383, 366, 26 },
		{ 383, 367, 26 },
		{ 383, 368, 26 },
		{ 383, 369, 26 },
		{ 383, 370, 26 },
		{ 383, 371, 26 },
		{ 383, 372, 26 },
		{ 383, 373, 26 },
		{ 383, 374, 26 },
		{ 383, 375, 26 },
		{ 383, 376, 26 },
		{ 383, 377, 26 },
		{ 383, 378, 26 },
		{ 383, 379, 26 },
		{ 383, 380, 26 },
		{ 383, 381, 26 },
		{ 383, 382, 26 },
		{ 383, 383, 26 },
		{ 135, 132, 27 },
		{ 135, 133, 27 },
		{ 135, 134, 27 },
		{ 136, 132, 27 },
		{ 136, 133, 27 },
		{ 136, 134, 27 },
		{ 136, 135, 27 },
		{ 137, 132, 27 },
		{ 137, 133, 27 },
		{ 137, 134, 27 },
		{ 137, 135, 27 },
		{ 137, 136, 27 },
		{ 138, 131, 27 },
		{ 138, 132, 27 },
		{ 138, 133, 27 },
		{ 138, 134, 27 },
		{ 138, 135, 27 },
		{ 138, 136, 27 },
		{ 138, 137, 27 },
		{ 139, 131, 27 },
		{ 139, 132, 27 },
		{ 139, 133, 27 },
		{ 139, 134, 27 },
		{ 139, 135, 27 },
		{ 139, 136, 27 },
		{ 139, 137, 27 },
		{ 139, 138, 27 },
		{ 140, 131, 27 },
		{ 140, 132, 27 },
		{ 140, 133, 27 },
		{ 140, 134, 27 },
		{ 140, 135, 27 },
		{ 140, 136, 27 },
		{ 140, 137, 27 },
		{ 140, 138, 27 },
		{ 140, 139, 27 },
		{ 141, 131, 27 },
		{ 141, 132, 27 },
		{ 141, 133, 27 },
		{ 141, 134, 27 },
		{ 141, 135, 27 },
		{ 141, 136, 27 },
		{ 141, 137, 27 },
		{ 141, 138, 27 },
		{ 141, 139, 27 },
		{ 141, 140, 27 },
		{ 142, 131, 27 },
		{ 142, 132, 27 },
		{ 142, 133, 27 },
		{ 142, 134, 27 },
		{ 142, 135, 27 },
		{ 142, 136, 27 },
		{ 142, 137, 27 },
		{ 142, 138, 27 },
		{ 142, 139, 27 },
		{ 142, 140, 27 },
		{ 142, 141, 27 },
		{ 143, 131, 27 },
		{ 143, 132, 27 },
		{ 143, 133, 27 },
		{ 143, 134, 27 },
		{ 143, 135, 27 },
		{ 143, 136, 27 },
		{ 143, 137, 27 },
		{ 143, 138, 27 },
		{ 143, 139, 27 },
		{ 143, 140, 27 },
		{ 143, 141, 27 },
		{ 143, 142, 27 },
		{ 144, 131, 27 },
		{ 144, 132, 27 },
		{ 144, 133, 27 },
		{ 144, 134, 27 },
		{ 144, 135, 27 },
		{ 144, 136, 27 },
		{ 144, 137, 27 },
		{ 144, 138, 27 },
		{ 144, 139, 27 },
		{ 144, 140, 27 },
		{ 144, 141, 27 },
		{ 144, 142, 27 },
		{ 144, 143, 27 },
		{ 145, 131, 27 },
		{ 145, 132, 27 },
		{ 145, 133, 27 },
		{ 145, 134, 27 },
		{ 145, 135, 27 },
		{ 145, 136, 27 },
		{ 145, 137, 27 },
		{ 145, 138, 27 },
		{ 145, 139, 27 },
		{ 145, 140, 27 },
		{ 145, 141, 27 },
		{ 145, 142, 27 },
		{ 145, 143, 27 },
		{ 145, 144, 27 },
		{ 146, 131, 27 },
		{ 146, 132, 27 },
		{ 146, 133, 27 },
		{ 146, 134, 27 },
		{ 146, 135, 27 },
		{ 146, 136, 27 },
		{ 146, 137, 27 },
		{ 146, 138, 27 },
		{ 146, 139, 27 },
		{ 146, 140, 27 },
		{ 146, 141, 27 },
		{ 146, 142, 27 },
		{ 146, 143, 27 },
		{ 146, 144, 27 },
		{ 146, 145, 27 },
		{ 147, 131, 27 },
		{ 147, 132, 27 },
		{ 147, 133, 27 },
		{ 147, 134, 27 },
		{ 147, 135, 27 },
		{ 147, 136, 27 },
		{ 147, 137, 27 },
		{ 147, 138, 27 },
		{ 147, 139, 27 },
		{ 147, 140, 27 },
		{ 147, 141, 27 },
		{ 147, 142, 27 },
		{ 147, 143, 27 },
		{ 147, 144, 27 },
		{ 147, 145, 27 },
		{ 147, 146, 27 },
		{ 148, 131, 27 },
		{ 148, 132, 27 },
		{ 148, 133, 27 },
		{ 148, 134, 27 },
		{ 148, 135, 27 },
		{ 148, 136, 27 },
		{ 148, 137, 27 },
		{ 148, 138, 27 },
		{ 148, 139, 27 },
		{ 148, 140, 27 },
		{ 148, 141, 27 },
		{ 148, 142, 27 },
		{ 148, 143, 27 },
		{ 148, 144, 27 },
		{ 148, 145, 27 },
		{ 148, 146, 27 },
		{ 148, 147, 27 },
		{ 149, 131, 27 },
		{ 149, 132, 27 },
		{ 149, 133, 27 },
		{ 149, 134, 27 },
		{ 149, 135, 27 },
		{ 149, 136, 27 },
		{ 149, 137, 27 },
		{ 149, 138, 27 },
		{ 149, 139, 27 },
		{ 149, 140, 27 },
		{ 149, 141, 27 },
		{ 149, 142, 27 },
		{ 149, 143, 27 },
		{ 149, 144, 27 },
		{ 149, 145, 27 },
		{ 149, 146, 27 },
		{ 149, 147, 27 },
		{ 149, 148, 27 },
		{ 150, 131, 27 },
		{ 150, 132, 27 },
		{ 150, 133, 27 },
		{ 150, 134, 27 },
		{ 150, 135, 27 },
		{ 150, 136, 27 },
		{ 150, 137, 27 },
		{ 150, 138, 27 },
		{ 150, 139, 27 },
		{ 150, 140, 27 },
		{ 150, 141, 27 },
		{ 150, 142, 27 },
		{ 150, 143, 27 },
		{ 150, 144, 27 },
		{ 150, 145, 27 },
		{ 150, 146, 27 },
		{ 150, 147, 27 },
		{ 150, 148, 27 },
		{ 150, 149, 27 },
		{ 151, 131, 27 },
		{ 151, 132, 27 },
		{ 151, 133, 27 },
		{ 151, 134, 27 },
		{ 151, 135, 27 },
		{ 151, 136, 27 },
		{ 151, 137, 27 },
		{ 151, 138, 27 },
		{ 151, 139, 27 },
		{ 151, 140, 27 },
		{ 151, 141, 27 },
		{ 151, 142, 27 },
		{ 151, 143, 27 },
		{ 151, 144, 27 },
		{ 151, 145, 27 },
		{ 151, 146, 27 },
		{ 151, 147, 27 },
		{ 151, 148, 27 },
		{ 151, 149, 27 },
		{ 151, 150, 27 },
		{ 152, 131, 27 },
		{ 152, 132, 27 },
		{ 152, 133, 27 },
		{ 152, 134, 27 },
		{ 152, 135, 27 },
		{ 152, 136, 27 },
		{ 152, 137, 27 },
		{ 152, 138, 27 },
		{ 152, 139, 27 },
		{ 152, 140, 27 },
		{ 152, 141, 27 },
		{ 152, 142, 27 },
		{ 152, 143, 27 },
		{ 152, 144, 27 },
		{ 152, 145, 27 },
		{ 152, 146, 27 },
		{ 152, 147, 27 },
		{ 152, 148, 27 },
		{ 152, 149, 27 },
		{ 152, 150, 27 },
		{ 152, 151, 27 },
		{ 153, 131, 27 },
		{ 153, 132, 27 },
		{ 153, 133, 27 },
		{ 153, 134, 27 },
		{ 153, 135, 27 },
		{ 153, 136, 27 },
		{ 153, 137, 27 },
		{ 153, 138, 27 },
		{ 153, 139, 27 },
		{ 153, 140, 27 },
		{ 153, 141, 27 },
		{ 153, 142, 27 },
		{ 153, 143, 27 },
		{ 153, 144, 27 },
		{ 153, 145, 27 },
		{ 153, 146, 27 },
		{ 153, 147, 27 },
		{ 153, 148, 27 },
		{ 153, 149, 27 },
		{ 153, 150, 27 },
		{ 153, 151, 27 },
		{ 153, 152, 27 },
		{ 154, 131, 27 },
		{ 154, 132, 27 },
		{ 154, 133, 27 },
		{ 154, 134, 27 },
		{ 154, 135, 27 },
		{ 154, 136, 27 },
		{ 154, 137, 27 },
		{ 154, 138, 27 },
		{ 154, 139, 27 },
		{ 154, 140, 27 },
		{ 154, 141, 27 },
		{ 154, 142, 27 },
		{ 154, 143, 27 },
		{ 154, 144, 27 },
		{ 154, 145, 27 },
		{ 154, 146, 27 },
		{ 154, 147, 27 },
		{ 154, 148, 27 },
		{ 154, 149, 27 },
		{ 154, 150, 27 },
		{ 154, 151, 27 },
		{ 154, 152, 27 },
		{ 154, 153, 27 },
		{ 155, 131, 27 },
		{ 155, 132, 27 },
		{ 155, 133, 27 },
		{ 155, 134, 27 },
		{ 155, 135, 27 },
		{ 155, 136, 27 },
		{ 155, 137, 27 },
		{ 155, 138, 27 },
		{ 155, 139, 27 },
		{ 155, 140, 27 },
		{ 155, 141, 27 },
		{ 155, 142, 27 },
		{ 155, 143, 27 },
		{ 155, 144, 27 },
		{ 155, 145, 27 },
		{ 155, 146, 27 },
		{ 155, 147, 27 },
		{ 155, 148, 27 },
		{ 155, 149, 27 },
		{ 155, 150, 27 },
		{ 155, 151, 27 },
		{ 155, 152, 27 },
		{ 155, 153, 27 },
		{ 155, 154, 27 },
		{ 156, 131, 27 },
		{ 156, 132, 27 },
		{ 156, 133, 27 },
		{ 156, 134, 27 },
		{ 156, 135, 27 },
		{ 156, 136, 27 },
		{ 156, 137, 27 },
		{ 156, 138, 27 },
		{ 156, 139, 27 },
		{ 156, 140, 27 },
		{ 156, 141, 27 },
		{ 156, 142, 27 },
		{ 156, 143, 27 },
		{ 156, 144, 27 },
		{ 156, 145, 27 },
		{ 156, 146, 27 },
		{ 156, 147, 27 },
		{ 156, 148, 27 },
		{ 156, 149, 27 },
		{ 156, 150, 27 },
		{ 156, 151, 27 },
		{ 156, 152, 27 },
		{ 156, 153, 27 },
		{ 156, 154, 27 },
		{ 156, 155, 27 },
		{ 157, 131, 27 },
		{ 157, 132, 27 },
		{ 157, 133, 27 },
		{ 157, 134, 27 },
		{ 157, 135, 27 },
		{ 157, 136, 27 },
		{ 157, 137, 27 },
		{ 157, 138, 27 },
		{ 157, 139, 27 },
		{ 157, 140, 27 },
		{ 157, 141, 27 },
		{ 157, 142, 27 },
		{ 157, 143, 27 },
		{ 157, 144, 27 },
		{ 157, 145, 27 },
		{ 157, 146, 27 },
		{ 157, 147, 27 },
		{ 157, 148, 27 },
		{ 157, 149, 27 },
		{ 157, 150, 27 },
		{ 157, 151, 27 },
		{ 157, 152, 27 },
		{ 157, 153, 27 },
		{ 157, 154, 27 },
		{ 157, 155, 27 },
		{ 157, 156, 27 },
		{ 158, 131, 27 },
		{ 158, 132, 27 },
		{ 158, 133, 27 },
		{ 158, 134, 27 },
		{ 158, 135, 27 },
		{ 158, 136, 27 },
		{ 158, 137, 27 },
		{ 158, 138, 27 },
		{ 158, 139, 27 },
		{ 158, 140, 27 },
		{ 158, 141, 27 },
		{ 158, 142, 27 },
		{ 158, 143, 27 },
		{ 158, 144, 27 },
		{ 158, 145, 27 },
		{ 158, 146, 27 },
		{ 158, 147, 27 },
		{ 158, 148, 27 },
		{ 158, 149, 27 },
		{ 158, 150, 27 },
		{ 158, 151, 27 },
		{ 158, 152, 27 },
		{ 158, 153, 27 },
		{ 158, 154, 27 },
		{ 158, 155, 27 },
		{ 158, 156, 27 },
		{ 158, 157, 27 },
		{ 159, 131, 27 },
		{ 159, 132, 27 },
		{ 159, 133, 27 },
		{ 159, 134, 27 },
		{ 159, 135, 27 },
		{ 159, 136, 27 },
		{ 159, 137, 27 },
		{ 159, 138, 27 },
		{ 159, 139, 27 },
		{ 159, 140, 27 },
		{ 159, 141, 27 },
		{ 159, 142, 27 },
		{ 159, 143, 27 },
		{ 159, 144, 27 },
		{ 159, 145, 27 },
		{ 159, 146, 27 },
		{ 159, 147, 27 },
		{ 159, 148, 27 },
		{ 159, 149, 27 },
		{ 159, 150, 27 },
		{ 159, 151, 27 },
		{ 159, 152, 27 },
		{ 159, 153, 27 },
		{ 159, 154, 27 },
		{ 159, 155, 27 },
		{ 159, 156, 27 },
		{ 159, 157, 27 },
		{ 159, 158, 27 },
		{ 160, 131, 27 },
		{ 160, 132, 27 },
		{ 160, 133, 27 },
		{ 160, 134, 27 },
		{ 160, 135, 27 },
		{ 160, 136, 27 },
		{ 160, 137, 27 },
		{ 160, 138, 27 },
		{ 160, 139, 27 },
		{ 160, 140, 27 },
		{ 160, 141, 27 },
		{ 160, 142, 27 },
		{ 160, 143, 27 },
		{ 160, 144, 27 },
		{ 160, 145, 27 },
		{ 160, 146, 27 },
		{ 160, 147, 27 },
		{ 160, 148, 27 },
		{ 160, 149, 27 },
		{ 160, 150, 27 },
		{ 160, 151, 27 },
		{ 160, 152, 27 },
		{ 160, 153, 27 },
		{ 160, 154, 27 },
		{ 160, 155, 27 },
		{ 160, 156, 27 },
		{ 160, 157, 27 },
		{ 160, 158, 27 },
		{ 160, 159, 27 },
		{ 161, 131, 27 },
		{ 161, 132, 27 },
		{ 161, 133, 27 },
		{ 161, 134, 27 },
		{ 161, 135, 27 },
		{ 161, 136, 27 },
		{ 161, 137, 27 },
		{ 161, 138, 27 },
		{ 161, 139, 27 },
		{ 161, 140, 27 },
		{ 161, 141, 27 },
		{ 161, 142, 27 },
		{ 161, 143, 27 },
		{ 161, 144, 27 },
		{ 161, 145, 27 },
		{ 161, 146, 27 },
		{ 161, 147, 27 },
		{ 161, 148, 27 },
		{ 161, 149, 27 },
		{ 161, 150, 27 },
		{ 161, 151, 27 },
		{ 161, 152, 27 },
		{ 161, 153, 27 },
		{ 161, 154, 27 },
		{ 161, 155, 27 },
		{ 161, 156, 27 },
		{ 161, 157, 27 },
		{ 161, 158, 27 },
		{ 161, 159, 27 },
		{ 161, 160, 27 },
		{ 162, 131, 27 },
		{ 162, 132, 27 },
		{ 162, 133, 27 },
		{ 162, 134, 27 },
		{ 162, 135, 27 },
		{ 162, 136, 27 },
		{ 162, 137, 27 },
		{ 162, 138, 27 },
		{ 162, 139, 27 },
		{ 162, 140, 27 },
		{ 162, 141, 27 },
		{ 162, 142, 27 },
		{ 162, 143, 27 },
		{ 162, 144, 27 },
		{ 162, 145, 27 },
		{ 162, 146, 27 },
		{ 162, 147, 27 },
		{ 162, 148, 27 },
		{ 162, 149, 27 },
		{ 162, 150, 27 },
		{ 162, 151, 27 },
		{ 162, 152, 27 },
		{ 162, 153, 27 },
		{ 162, 154, 27 },
		{ 162, 155, 27 },
		{ 162, 156, 27 },
		{ 162, 157, 27 },
		{ 162, 158, 27 },
		{ 162, 159, 27 },
		{ 162, 160, 27 },
		{ 162, 161, 27 },
		{ 163, 131, 27 },
		{ 163, 132, 27 },
		{ 163, 133, 27 },
		{ 163, 134, 27 },
		{ 163, 135, 27 },
		{ 163, 136, 27 },
		{ 163, 137, 27 },
		{ 163, 138, 27 },
		{ 163, 139, 27 },
		{ 163, 140, 27 },
		{ 163, 141, 27 },
		{ 163, 142, 27 },
		{ 163, 143, 27 },
		{ 163, 144, 27 },
		{ 163, 145, 27 },
		{ 163, 146, 27 },
		{ 163, 147, 27 },
		{ 163, 148, 27 },
		{ 163, 149, 27 },
		{ 163, 150, 27 },
		{ 163, 151, 27 },
		{ 163, 152, 27 },
		{ 163, 153, 27 },
		{ 163, 154, 27 },
		{ 163, 155, 27 },
		{ 163, 156, 27 },
		{ 163, 157, 27 },
		{ 163, 158, 27 },
		{ 163, 159, 27 },
		{ 163, 160, 27 },
		{ 163, 161, 27 },
		{ 163, 162, 27 },
		{ 164, 131, 27 },
		{ 164, 132, 27 },
		{ 164, 133, 27 },
		{ 164, 134, 27 },
		{ 164, 135, 27 },
		{ 164, 136, 27 },
		{ 164, 137, 27 },
		{ 164, 138, 27 },
		{ 164, 139, 27 },
		{ 164, 140, 27 },
		{ 164, 141, 27 },
		{ 164, 142, 27 },
		{ 164, 143, 27 },
		{ 164, 144, 27 },
		{ 164, 145, 27 },
		{ 164, 146, 27 },
		{ 164, 147, 27 },
		{ 164, 148, 27 },
		{ 164, 149, 27 },
		{ 164, 150, 27 },
		{ 164, 151, 27 },
		{ 164, 152, 27 },
		{ 164, 153, 27 },
		{ 164, 154, 27 },
		{ 164, 155, 27 },
		{ 164, 156, 27 },
		{ 164, 157, 27 },
		{ 164, 158, 27 },
		{ 164, 159, 27 },
		{ 164, 160, 27 },
		{ 164, 161, 27 },
		{ 164, 162, 27 },
		{ 164, 163, 27 },
		{ 165, 131, 27 },
		{ 165, 132, 27 },
		{ 165, 133, 27 },
		{ 165, 134, 27 },
		{ 165, 135, 27 },
		{ 165, 136, 27 },
		{ 165, 137, 27 },
		{ 165, 138, 27 },
		{ 165, 139, 27 },
		{ 165, 140, 27 },
		{ 165, 141, 27 },
		{ 165, 142, 27 },
		{ 165, 143, 27 },
		{ 165, 144, 27 },
		{ 165, 145, 27 },
		{ 165, 146, 27 },
		{ 165, 147, 27 },
		{ 165, 148, 27 },
		{ 165, 149, 27 },
		{ 165, 150, 27 },
		{ 165, 151, 27 },
		{ 165, 152, 27 },
		{ 165, 153, 27 },
		{ 165, 154, 27 },
		{ 165, 155, 27 },
		{ 165, 156, 27 },
		{ 165, 157, 27 },
		{ 165, 158, 27 },
		{ 165, 159, 27 },
		{ 165, 160, 27 },
		{ 165, 161, 27 },
		{ 165, 162, 27 },
		{ 165, 163, 27 },
		{ 165, 164, 27 },
		{ 166, 131, 27 },
		{ 166, 132, 27 },
		{ 166, 133, 27 },
		{ 166, 134, 27 },
		{ 166, 135, 27 },
		{ 166, 136, 27 },
		{ 166, 137, 27 },
		{ 166, 138, 27 },
		{ 166, 139, 27 },
		{ 166, 140, 27 },
		{ 166, 141, 27 },
		{ 166, 142, 27 },
		{ 166, 143, 27 },
		{ 166, 144, 27 },
		{ 166, 145, 27 },
		{ 166, 146, 27 },
		{ 166, 147, 27 },
		{ 166, 148, 27 },
		{ 166, 149, 27 },
		{ 166, 150, 27 },
		{ 166, 151, 27 },
		{ 166, 152, 27 },
		{ 166, 153, 27 },
		{ 166, 154, 27 },
		{ 166, 155, 27 },
		{ 166, 156, 27 },
		{ 166, 157, 27 },
		{ 166, 158, 27 },
		{ 166, 159, 27 },
		{ 166, 160, 27 },
		{ 166, 161, 27 },
		{ 166, 162, 27 },
		{ 166, 163, 27 },
		{ 166, 164, 27 },
		{ 166, 165, 27 },
		{ 167, 131, 27 },
		{ 167, 132, 27 },
		{ 167, 133, 27 },
		{ 167, 134, 27 },
		{ 167, 135, 27 },
		{ 167, 136, 27 },
		{ 167, 137, 27 },
		{ 167, 138, 27 },
		{ 167, 139, 27 },
		{ 167, 140, 27 },
		{ 167, 141, 27 },
		{ 167, 142, 27 },
		{ 167, 143, 27 },
		{ 167, 144, 27 },
		{ 167, 145, 27 },
		{ 167, 146, 27 },
		{ 167, 147, 27 },
		{ 167, 148, 27 },
		{ 167, 149, 27 },
		{ 167, 150, 27 },
		{ 167, 151, 27 },
		{ 167, 152, 27 },
		{ 167, 153, 27 },
		{ 167, 154, 27 },
		{ 167, 155, 27 },
		{ 167, 156, 27 },
		{ 167, 157, 27 },
		{ 167, 158, 27 },
		{ 167, 159, 27 },
		{ 167, 160, 27 },
		{ 167, 161, 27 },
		{ 167, 162, 27 },
		{ 167, 163, 27 },
		{ 167, 164, 27 },
		{ 167, 165, 27 },
		{ 167, 166, 27 },
		{ 168, 131, 27 },
		{ 168, 132, 27 },
		{ 168, 133, 27 },
		{ 168, 134, 27 },
		{ 168, 135, 27 },
		{ 168, 136, 27 },
		{ 168, 137, 27 },
		{ 168, 138, 27 },
		{ 168, 139, 27 },
		{ 168, 140, 27 },
		{ 168, 141, 27 },
		{ 168, 142, 27 },
		{ 168, 143, 27 },
		{ 168, 144, 27 },
		{ 168, 145, 27 },
		{ 168, 146, 27 },
		{ 168, 147, 27 },
		{ 168, 148, 27 },
		{ 168, 149, 27 },
		{ 168, 150, 27 },
		{ 168, 151, 27 },
		{ 168, 152, 27 },
		{ 168, 153, 27 },
		{ 168, 154, 27 },
		{ 168, 155, 27 },
		{ 168, 156, 27 },
		{ 168, 157, 27 },
		{ 168, 158, 27 },
		{ 168, 159, 27 },
		{ 168, 160, 27 },
		{ 168, 161, 27 },
		{ 168, 162, 27 },
		{ 168, 163, 27 },
		{ 168, 164, 27 },
		{ 168, 165, 27 },
		{ 168, 166, 27 },
		{ 168, 167, 27 },
		{ 169, 131, 27 },
		{ 169, 132, 27 },
		{ 169, 133, 27 },
		{ 169, 134, 27 },
		{ 169, 135, 27 },
		{ 169, 136, 27 },
		{ 169, 137, 27 },
		{ 169, 138, 27 },
		{ 169, 139, 27 },
		{ 169, 140, 27 },
		{ 169, 141, 27 },
		{ 169, 142, 27 },
		{ 169, 143, 27 },
		{ 169, 144, 27 },
		{ 169, 145, 27 },
		{ 169, 146, 27 },
		{ 169, 147, 27 },
		{ 169, 148, 27 },
		{ 169, 149, 27 },
		{ 169, 150, 27 },
		{ 169, 151, 27 },
		{ 169, 152, 27 },
		{ 169, 153, 27 },
		{ 169, 154, 27 },
		{ 169, 155, 27 },
		{ 169, 156, 27 },
		{ 169, 157, 27 },
		{ 169, 158, 27 },
		{ 169, 159, 27 },
		{ 169, 160, 27 },
		{ 169, 161, 27 },
		{ 169, 162, 27 },
		{ 169, 163, 27 },
		{ 169, 164, 27 },
		{ 169, 165, 27 },
		{ 169, 166, 27 },
		{ 169, 167, 27 },
		{ 169, 168, 27 },
		{ 170, 131, 27 },
		{ 170, 132, 27 },
		{ 170, 133, 27 },
		{ 170, 134, 27 },
		{ 170, 135, 27 },
		{ 170, 136, 27 },
		{ 170, 137, 27 },
		{ 170, 138, 27 },
		{ 170, 139, 27 },
		{ 170, 140, 27 },
		{ 170, 141, 27 },
		{ 170, 142, 27 },
		{ 170, 143, 27 },
		{ 170, 144, 27 },
		{ 170, 145, 27 },
		{ 170, 146, 27 },
		{ 170, 147, 27 },
		{ 170, 148, 27 },
		{ 170, 149, 27 },
		{ 170, 150, 27 },
		{ 170, 151, 27 },
		{ 170, 152, 27 },
		{ 170, 153, 27 },
		{ 170, 154, 27 },
		{ 170, 155, 27 },
		{ 170, 156, 27 },
		{ 170, 157, 27 },
		{ 170, 158, 27 },
		{ 170, 159, 27 },
		{ 170, 160, 27 },
		{ 170, 161, 27 },
		{ 170, 162, 27 },
		{ 170, 163, 27 },
		{ 170, 164, 27 },
		{ 170, 165, 27 },
		{ 170, 166, 27 },
		{ 170, 167, 27 },
		{ 170, 168, 27 },
		{ 170, 169, 27 },
		{ 171, 131, 27 },
		{ 171, 132, 27 },
		{ 171, 133, 27 },
		{ 171, 134, 27 },
		{ 171, 135, 27 },
		{ 171, 136, 27 },
		{ 171, 137, 27 },
		{ 171, 138, 27 },
		{ 171, 139, 27 },
		{ 171, 140, 27 },
		{ 171, 141, 27 },
		{ 171, 142, 27 },
		{ 171, 143, 27 },
		{ 171, 144, 27 },
		{ 171, 145, 27 },
		{ 171, 146, 27 },
		{ 171, 147, 27 },
		{ 171, 148, 27 },
		{ 171, 149, 27 },
		{ 171, 150, 27 },
		{ 171, 151, 27 },
		{ 171, 152, 27 },
		{ 171, 153, 27 },
		{ 171, 154, 27 },
		{ 171, 155, 27 },
		{ 171, 156, 27 },
		{ 171, 157, 27 },
		{ 171, 158, 27 },
		{ 171, 159, 27 },
		{ 171, 160, 27 },
		{ 171, 161, 27 },
		{ 171, 162, 27 },
		{ 171, 163, 27 },
		{ 171, 164, 27 },
		{ 171, 165, 27 },
		{ 171, 166, 27 },
		{ 171, 167, 27 },
		{ 171, 168, 27 },
		{ 171, 169, 27 },
		{ 171, 170, 27 },
		{ 172, 131, 27 },
		{ 172, 132, 27 },
		{ 172, 133, 27 },
		{ 172, 134, 27 },
		{ 172, 135, 27 },
		{ 172, 136, 27 },
		{ 172, 137, 27 },
		{ 172, 138, 27 },
		{ 172, 139, 27 },
		{ 172, 140, 27 },
		{ 172, 141, 27 },
		{ 172, 142, 27 },
		{ 172, 143, 27 },
		{ 172, 144, 27 },
		{ 172, 145, 27 },
		{ 172, 146, 27 },
		{ 172, 147, 27 },
		{ 172, 148, 27 },
		{ 172, 149, 27 },
		{ 172, 150, 27 },
		{ 172, 151, 27 },
		{ 172, 152, 27 },
		{ 172, 153, 27 },
		{ 172, 154, 27 },
		{ 172, 155, 27 },
		{ 172, 156, 27 },
		{ 172, 157, 27 },
		{ 172, 158, 27 },
		{ 172, 159, 27 },
		{ 172, 160, 27 },
		{ 172, 161, 27 },
		{ 172, 162, 27 },
		{ 172, 163, 27 },
		{ 172, 164, 27 },
		{ 172, 165, 27 },
		{ 172, 166, 27 },
		{ 172, 167, 27 },
		{ 172, 168, 27 },
		{ 172, 169, 27 },
		{ 172, 170, 27 },
		{ 172, 171, 27 },
		{ 173, 131, 27 },
		{ 173, 132, 27 },
		{ 173, 133, 27 },
		{ 173, 134, 27 },
		{ 173, 135, 27 },
		{ 173, 136, 27 },
		{ 173, 137, 27 },
		{ 173, 138, 27 },
		{ 173, 139, 27 },
		{ 173, 140, 27 },
		{ 173, 141, 27 },
		{ 173, 142, 27 },
		{ 173, 143, 27 },
		{ 173, 144, 27 },
		{ 173, 145, 27 },
		{ 173, 146, 27 },
		{ 173, 147, 27 },
		{ 173, 148, 27 },
		{ 173, 149, 27 },
		{ 173, 150, 27 },
		{ 173, 151, 27 },
		{ 173, 152, 27 },
		{ 173, 153, 27 },
		{ 173, 154, 27 },
		{ 173, 155, 27 },
		{ 173, 156, 27 },
		{ 173, 157, 27 },
		{ 173, 158, 27 },
		{ 173, 159, 27 },
		{ 173, 160, 27 },
		{ 173, 161, 27 },
		{ 173, 162, 27 },
		{ 173, 163, 27 },
		{ 173, 164, 27 },
		{ 173, 165, 27 },
		{ 173, 166, 27 },
		{ 173, 167, 27 },
		{ 173, 168, 27 },
		{ 173, 169, 27 },
		{ 173, 170, 27 },
		{ 173, 171, 27 },
		{ 173, 172, 27 },
		{ 174, 131, 27 },
		{ 174, 132, 27 },
		{ 174, 133, 27 },
		{ 174, 134, 27 },
		{ 174, 135, 27 },
		{ 174, 136, 27 },
		{ 174, 137, 27 },
		{ 174, 138, 27 },
		{ 174, 139, 27 },
		{ 174, 140, 27 },
		{ 174, 141, 27 },
		{ 174, 142, 27 },
		{ 174, 143, 27 },
		{ 174, 144, 27 },
		{ 174, 145, 27 },
		{ 174, 146, 27 },
		{ 174, 147, 27 },
		{ 174, 148, 27 },
		{ 174, 149, 27 },
		{ 174, 150, 27 },
		{ 174, 151, 27 },
		{ 174, 152, 27 },
		{ 174, 153, 27 },
		{ 174, 154, 27 },
		{ 174, 155, 27 },
		{ 174, 156, 27 },
		{ 174, 157, 27 },
		{ 174, 158, 27 },
		{ 174, 159, 27 },
		{ 174, 160, 27 },
		{ 174, 161, 27 },
		{ 174, 162, 27 },
		{ 174, 163, 27 },
		{ 174, 164, 27 },
		{ 174, 165, 27 },
		{ 174, 166, 27 },
		{ 174, 167, 27 },
		{ 174, 168, 27 },
		{ 174, 169, 27 },
		{ 174, 170, 27 },
		{ 174, 171, 27 },
		{ 174, 172, 27 },
		{ 174, 173, 27 },
		{ 175, 131, 27 },
		{ 175, 132, 27 },
		{ 175, 133, 27 },
		{ 175, 134, 27 },
		{ 175, 135, 27 },
		{ 175, 136, 27 },
		{ 175, 137, 27 },
		{ 175, 138, 27 },
		{ 175, 139, 27 },
		{ 175, 140, 27 },
		{ 175, 141, 27 },
		{ 175, 142, 27 },
		{ 175, 143, 27 },
		{ 175, 144, 27 },
		{ 175, 145, 27 },
		{ 175, 146, 27 },
		{ 175, 147, 27 },
		{ 175, 148, 27 },
		{ 175, 149, 27 },
		{ 175, 150, 27 },
		{ 175, 151, 27 },
		{ 175, 152, 27 },
		{ 175, 153, 27 },
		{ 175, 154, 27 },
		{ 175, 155, 27 },
		{ 175, 156, 27 },
		{ 175, 157, 27 },
		{ 175, 158, 27 },
		{ 175, 159, 27 },
		{ 175, 160, 27 },
		{ 175, 161, 27 },
		{ 175, 162, 27 },
		{ 175, 163, 27 },
		{ 175, 164, 27 },
		{ 175, 165, 27 },
		{ 175, 166, 27 },
		{ 175, 167, 27 },
		{ 175, 168, 27 },
		{ 175, 169, 27 },
		{ 175, 170, 27 },
		{ 175, 171, 27 },
		{ 175, 172, 27 },
		{ 175, 173, 27 },
		{ 175, 174, 27 },
		{ 176, 131, 27 },
		{ 176, 132, 27 },
		{ 176, 133, 27 },
		{ 176, 134, 27 },
		{ 176, 135, 27 },
		{ 176, 136, 27 },
		{ 176, 137, 27 },
		{ 176, 138, 27 },
		{ 176, 139, 27 },
		{ 176, 140, 27 },
		{ 176, 141, 27 },
		{ 176, 142, 27 },
		{ 176, 143, 27 },
		{ 176, 144, 27 },
		{ 176, 145, 27 },
		{ 176, 146, 27 },
		{ 176, 147, 27 },
		{ 176, 148, 27 },
		{ 176, 149, 27 },
		{ 176, 150, 27 },
		{ 176, 151, 27 },
		{ 176, 152, 27 },
		{ 176, 153, 27 },
		{ 176, 154, 27 },
		{ 176, 155, 27 },
		{ 176, 156, 27 },
		{ 176, 157, 27 },
		{ 176, 158, 27 },
		{ 176, 159, 27 },
		{ 176, 160, 27 },
		{ 176, 161, 27 },
		{ 176, 162, 27 },
		{ 176, 163, 27 },
		{ 176, 164, 27 },
		{ 176, 165, 27 },
		{ 176, 166, 27 },
		{ 176, 167, 27 },
		{ 176, 168, 27 },
		{ 176, 169, 27 },
		{ 176, 170, 27 },
		{ 176, 171, 27 },
		{ 176, 172, 27 },
		{ 176, 173, 27 },
		{ 176, 174, 27 },
		{ 176, 175, 27 },
		{ 177, 131, 27 },
		{ 177, 132, 27 },
		{ 177, 133, 27 },
		{ 177, 134, 27 },
		{ 177, 135, 27 },
		{ 177, 136, 27 },
		{ 177, 137, 27 },
		{ 177, 138, 27 },
		{ 177, 139, 27 },
		{ 177, 140, 27 },
		{ 177, 141, 27 },
		{ 177, 142, 27 },
		{ 177, 143, 27 },
		{ 177, 144, 27 },
		{ 177, 145, 27 },
		{ 177, 146, 27 },
		{ 177, 147, 27 },
		{ 177, 148, 27 },
		{ 177, 149, 27 },
		{ 177, 150, 27 },
		{ 177, 151, 27 },
		{ 177, 152, 27 },
		{ 177, 153, 27 },
		{ 177, 154, 27 },
		{ 177, 155, 27 },
		{ 177, 156, 27 },
		{ 177, 157, 27 },
		{ 177, 158, 27 },
		{ 177, 159, 27 },
		{ 177, 160, 27 },
		{ 177, 161, 27 },
		{ 177, 162, 27 },
		{ 177, 163, 27 },
		{ 177, 164, 27 },
		{ 177, 165, 27 },
		{ 177, 166, 27 },
		{ 177, 167, 27 },
		{ 177, 168, 27 },
		{ 177, 169, 27 },
		{ 177, 170, 27 },
		{ 177, 171, 27 },
		{ 177, 172, 27 },
		{ 177, 173, 27 },
		{ 177, 174, 27 },
		{ 177, 175, 27 },
		{ 177, 176, 27 },
		{ 178, 131, 27 },
		{ 178, 132, 27 },
		{ 178, 133, 27 },
		{ 178, 134, 27 },
		{ 178, 135, 27 },
		{ 178, 136, 27 },
		{ 178, 137, 27 },
		{ 178, 138, 27 },
		{ 178, 139, 27 },
		{ 178, 140, 27 },
		{ 178, 141, 27 },
		{ 178, 142, 27 },
		{ 178, 143, 27 },
		{ 178, 144, 27 },
		{ 178, 145, 27 },
		{ 178, 146, 27 },
		{ 178, 147, 27 },
		{ 178, 148, 27 },
		{ 178, 149, 27 },
		{ 178, 150, 27 },
		{ 178, 151, 27 },
		{ 178, 152, 27 },
		{ 178, 153, 27 },
		{ 178, 154, 27 },
		{ 178, 155, 27 },
		{ 178, 156, 27 },
		{ 178, 157, 27 },
		{ 178, 158, 27 },
		{ 178, 159, 27 },
		{ 178, 160, 27 },
		{ 178, 161, 27 },
		{ 178, 162, 27 },
		{ 178, 163, 27 },
		{ 178, 164, 27 },
		{ 178, 165, 27 },
		{ 178, 166, 27 },
		{ 178, 167, 27 },
		{ 178, 168, 27 },
		{ 178, 169, 27 },
		{ 178, 170, 27 },
		{ 178, 171, 27 },
		{ 178, 172, 27 },
		{ 178, 173, 27 },
		{ 178, 174, 27 },
		{ 178, 175, 27 },
		{ 178, 176, 27 },
		{ 178, 177, 27 },
		{ 179, 131, 27 },
		{ 179, 132, 27 },
		{ 179, 133, 27 },
		{ 179, 134, 27 },
		{ 179, 135, 27 },
		{ 179, 136, 27 },
		{ 179, 137, 27 },
		{ 179, 138, 27 },
		{ 179, 139, 27 },
		{ 179, 140, 27 },
		{ 179, 141, 27 },
		{ 179, 142, 27 },
		{ 179, 143, 27 },
		{ 179, 144, 27 },
		{ 179, 145, 27 },
		{ 179, 146, 27 },
		{ 179, 147, 27 },
		{ 179, 148, 27 },
		{ 179, 149, 27 },
		{ 179, 150, 27 },
		{ 179, 151, 27 },
		{ 179, 152, 27 },
		{ 179, 153, 27 },
		{ 179, 154, 27 },
		{ 179, 155, 27 },
		{ 179, 156, 27 },
		{ 179, 157, 27 },
		{ 179, 158, 27 },
		{ 179, 159, 27 },
		{ 179, 160, 27 },
		{ 179, 161, 27 },
		{ 179, 162, 27 },
		{ 179, 163, 27 },
		{ 179, 164, 27 },
		{ 179, 165, 27 },
		{ 179, 166, 27 },
		{ 179, 167, 27 },
		{ 179, 168, 27 },
		{ 179, 169, 27 },
		{ 179, 170, 27 },
		{ 179, 171, 27 },
		{ 179, 172, 27 },
		{ 179, 173, 27 },
		{ 179, 174, 27 },
		{ 179, 175, 27 },
		{ 179, 176, 27 },
		{ 179, 177, 27 },
		{ 179, 178, 27 },
		{ 180, 131, 27 },
		{ 180, 132, 27 },
		{ 180, 133, 27 },
		{ 180, 134, 27 },
		{ 180, 135, 27 },
		{ 180, 136, 27 },
		{ 180, 137, 27 },
		{ 180, 138, 27 },
		{ 180, 139, 27 },
		{ 180, 140, 27 },
		{ 180, 141, 27 },
		{ 180, 142, 27 },
		{ 180, 143, 27 },
		{ 180, 144, 27 },
		{ 180, 145, 27 },
		{ 180, 146, 27 },
		{ 180, 147, 27 },
		{ 180, 148, 27 },
		{ 180, 149, 27 },
		{ 180, 150, 27 },
		{ 180, 151, 27 },
		{ 180, 152, 27 },
		{ 180, 153, 27 },
		{ 180, 154, 27 },
		{ 180, 155, 27 },
		{ 180, 156, 27 },
		{ 180, 157, 27 },
		{ 180, 158, 27 },
		{ 180, 159, 27 },
		{ 180, 160, 27 },
		{ 180, 161, 27 },
		{ 180, 162, 27 },
		{ 180, 163, 27 },
		{ 180, 164, 27 },
		{ 180, 165, 27 },
		{ 180, 166, 27 },
		{ 180, 167, 27 },
		{ 180, 168, 27 },
		{ 180, 169, 27 },
		{ 180, 170, 27 },
		{ 180, 171, 27 },
		{ 180, 172, 27 },
		{ 180, 173, 27 },
		{ 180, 174, 27 },
		{ 180, 175, 27 },
		{ 180, 176, 27 },
		{ 180, 177, 27 },
		{ 180, 178, 27 },
		{ 180, 179, 27 },
		{ 181, 131, 27 },
		{ 181, 132, 27 },
		{ 181, 133, 27 },
		{ 181, 134, 27 },
		{ 181, 135, 27 },
		{ 181, 136, 27 },
		{ 181, 137, 27 },
		{ 181, 138, 27 },
		{ 181, 139, 27 },
		{ 181, 140, 27 },
		{ 181, 141, 27 },
		{ 181, 142, 27 },
		{ 181, 143, 27 },
		{ 181, 144, 27 },
		{ 181, 145, 27 },
		{ 181, 146, 27 },
		{ 181, 147, 27 },
		{ 181, 148, 27 },
		{ 181, 149, 27 },
		{ 181, 150, 27 },
		{ 181, 151, 27 },
		{ 181, 152, 27 },
		{ 181, 153, 27 },
		{ 181, 154, 27 },
		{ 181, 155, 27 },
		{ 181, 156, 27 },
		{ 181, 157, 27 },
		{ 181, 158, 27 },
		{ 181, 159, 27 },
		{ 181, 160, 27 },
		{ 181, 161, 27 },
		{ 181, 162, 27 },
		{ 181, 163, 27 },
		{ 181, 164, 27 },
		{ 181, 165, 27 },
		{ 181, 166, 27 },
		{ 181, 167, 27 },
		{ 181, 168, 27 },
		{ 181, 169, 27 },
		{ 181, 170, 27 },
		{ 181, 171, 27 },
		{ 181, 172, 27 },
		{ 181, 173, 27 },
		{ 181, 174, 27 },
		{ 181, 175, 27 },
		{ 181, 176, 27 },
		{ 181, 177, 27 },
		{ 181, 178, 27 },
		{ 181, 179, 27 },
		{ 181, 180, 27 },
		{ 182, 131, 27 },
		{ 182, 132, 27 },
		{ 182, 133, 27 },
		{ 182, 134, 27 },
		{ 182, 135, 27 },
		{ 182, 136, 27 },
		{ 182, 137, 27 },
		{ 182, 138, 27 },
		{ 182, 139, 27 },
		{ 182, 140, 27 },
		{ 182, 141, 27 },
		{ 182, 142, 27 },
		{ 182, 143, 27 },
		{ 182, 144, 27 },
		{ 182, 145, 27 },
		{ 182, 146, 27 },
		{ 182, 147, 27 },
		{ 182, 148, 27 },
		{ 182, 149, 27 },
		{ 182, 150, 27 },
		{ 182, 151, 27 },
		{ 182, 152, 27 },
		{ 182, 153, 27 },
		{ 182, 154, 27 },
		{ 182, 155, 27 },
		{ 182, 156, 27 },
		{ 182, 157, 27 },
		{ 182, 158, 27 },
		{ 182, 159, 27 },
		{ 182, 160, 27 },
		{ 182, 161, 27 },
		{ 182, 162, 27 },
		{ 182, 163, 27 },
		{ 182, 164, 27 },
		{ 182, 165, 27 },
		{ 182, 166, 27 },
		{ 182, 167, 27 },
		{ 182, 168, 27 },
		{ 182, 169, 27 },
		{ 182, 170, 27 },
		{ 182, 171, 27 },
		{ 182, 172, 27 },
		{ 182, 173, 27 },
		{ 182, 174, 27 },
		{ 182, 175, 27 },
		{ 182, 176, 27 },
		{ 182, 177, 27 },
		{ 182, 178, 27 },
		{ 182, 179, 27 },
		{ 182, 180, 27 },
		{ 182, 181, 27 },
		{ 183, 131, 27 },
		{ 183, 132, 27 },
		{ 183, 133, 27 },
		{ 183, 134, 27 },
		{ 183, 135, 27 },
		{ 183, 136, 27 },
		{ 183, 137, 27 },
		{ 183, 138, 27 },
		{ 183, 139, 27 },
		{ 183, 140, 27 },
		{ 183, 141, 27 },
		{ 183, 142, 27 },
		{ 183, 143, 27 },
		{ 183, 144, 27 },
		{ 183, 145, 27 },
		{ 183, 146, 27 },
		{ 183, 147, 27 },
		{ 183, 148, 27 },
		{ 183, 149, 27 },
		{ 183, 150, 27 },
		{ 183, 151, 27 },
		{ 183, 152, 27 },
		{ 183, 153, 27 },
		{ 183, 154, 27 },
		{ 183, 155, 27 },
		{ 183, 156, 27 },
		{ 183, 157, 27 },
		{ 183, 158, 27 },
		{ 183, 159, 27 },
		{ 183, 160, 27 },
		{ 183, 161, 27 },
		{ 183, 162, 27 },
		{ 183, 163, 27 },
		{ 183, 164, 27 },
		{ 183, 165, 27 },
		{ 183, 166, 27 },
		{ 183, 167, 27 },
		{ 183, 168, 27 },
		{ 183, 169, 27 },
		{ 183, 170, 27 },
		{ 183, 171, 27 },
		{ 183, 172, 27 },
		{ 183, 173, 27 },
		{ 183, 174, 27 },
		{ 183, 175, 27 },
		{ 183, 176, 27 },
		{ 183, 177, 27 },
		{ 183, 178, 27 },
		{ 183, 179, 27 },
		{ 183, 180, 27 },
		{ 183, 181, 27 },
		{ 183, 182, 27 },
		{ 184, 131, 27 },
		{ 184, 132, 27 },
		{ 184, 133, 27 },
		{ 184, 134, 27 },
		{ 184, 135, 27 },
		{ 184, 136, 27 },
		{ 184, 137, 27 },
		{ 184, 138, 27 },
		{ 184, 139, 27 },
		{ 184, 140, 27 },
		{ 184, 141, 27 },
		{ 184, 142, 27 },
		{ 184, 143, 27 },
		{ 184, 144, 27 },
		{ 184, 145, 27 },
		{ 184, 146, 27 },
		{ 184, 147, 27 },
		{ 184, 148, 27 },
		{ 184, 149, 27 },
		{ 184, 150, 27 },
		{ 184, 151, 27 },
		{ 184, 152, 27 },
		{ 184, 153, 27 },
		{ 184, 154, 27 },
		{ 184, 155, 27 },
		{ 184, 156, 27 },
		{ 184, 157, 27 },
		{ 184, 158, 27 },
		{ 184, 159, 27 },
		{ 184, 160, 27 },
		{ 184, 161, 27 },
		{ 184, 162, 27 },
		{ 184, 163, 27 },
		{ 184, 164, 27 },
		{ 184, 165, 27 },
		{ 184, 166, 27 },
		{ 184, 167, 27 },
		{ 184, 168, 27 },
		{ 184, 169, 27 },
		{ 184, 170, 27 },
		{ 184, 171, 27 },
		{ 184, 172, 27 },
		{ 184, 173, 27 },
		{ 184, 174, 27 },
		{ 184, 175, 27 },
		{ 184, 176, 27 },
		{ 184, 177, 27 },
		{ 184, 178, 27 },
		{ 184, 179, 27 },
		{ 184, 180, 27 },
		{ 184, 181, 27 },
		{ 184, 182, 27 },
		{ 184, 183, 27 },
		{ 185, 131, 27 },
		{ 185, 132, 27 },
		{ 185, 133, 27 },
		{ 185, 134, 27 },
		{ 185, 135, 27 },
		{ 185, 136, 27 },
		{ 185, 137, 27 },
		{ 185, 138, 27 },
		{ 185, 139, 27 },
		{ 185, 140, 27 },
		{ 185, 141, 27 },
		{ 185, 142, 27 },
		{ 185, 143, 27 },
		{ 185, 144, 27 },
		{ 185, 145, 27 },
		{ 185, 146, 27 },
		{ 185, 147, 27 },
		{ 185, 148, 27 },
		{ 185, 149, 27 },
		{ 185, 150, 27 },
		{ 185, 151, 27 },
		{ 185, 152, 27 },
		{ 185, 153, 27 },
		{ 185, 154, 27 },
		{ 185, 155, 27 },
		{ 185, 156, 27 },
		{ 185, 157, 27 },
		{ 185, 158, 27 },
		{ 185, 159, 27 },
		{ 185, 160, 27 },
		{ 185, 161, 27 },
		{ 185, 162, 27 },
		{ 185, 163, 27 },
		{ 185, 164, 27 },
		{ 185, 165, 27 },
		{ 185, 166, 27 },
		{ 185, 167, 27 },
		{ 185, 168, 27 },
		{ 185, 169, 27 },
		{ 185, 170, 27 },
		{ 185, 171, 27 },
		{ 185, 172, 27 },
		{ 185, 173, 27 },
		{ 185, 174, 27 },
		{ 185, 175, 27 },
		{ 185, 176, 27 },
		{ 185, 177, 27 },
		{ 185, 178, 27 },
		{ 185, 179, 27 },
		{ 185, 180, 27 },
		{ 185, 181, 27 },
		{ 185, 182, 27 },
		{ 185, 183, 27 },
		{ 185, 184, 27 },
		{ 186, 131, 27 },
		{ 186, 132, 27 },
		{ 186, 133, 27 },
		{ 186, 134, 27 },
		{ 186, 135, 27 },
		{ 186, 136, 27 },
		{ 186, 137, 27 },
		{ 186, 138, 27 },
		{ 186, 139, 27 },
		{ 186, 140, 27 },
		{ 186, 141, 27 },
		{ 186, 142, 27 },
		{ 186, 143, 27 },
		{ 186, 144, 27 },
		{ 186, 145, 27 },
		{ 186, 146, 27 },
		{ 186, 147, 27 },
		{ 186, 148, 27 },
		{ 186, 149, 27 },
		{ 186, 150, 27 },
		{ 186, 151, 27 },
		{ 186, 152, 27 },
		{ 186, 153, 27 },
		{ 186, 154, 27 },
		{ 186, 155, 27 },
		{ 186, 156, 27 },
		{ 186, 157, 27 },
		{ 186, 158, 27 },
		{ 186, 159, 27 },
		{ 186, 160, 27 },
		{ 186, 161, 27 },
		{ 186, 162, 27 },
		{ 186, 163, 27 },
		{ 186, 164, 27 },
		{ 186, 165, 27 },
		{ 186, 166, 27 },
		{ 186, 167, 27 },
		{ 186, 168, 27 },
		{ 186, 169, 27 },
		{ 186, 170, 27 },
		{ 186, 171, 27 },
		{ 186, 172, 27 },
		{ 186, 173, 27 },
		{ 186, 174, 27 },
		{ 186, 175, 27 },
		{ 186, 176, 27 },
		{ 186, 177, 27 },
		{ 186, 178, 27 },
		{ 186, 179, 27 },
		{ 186, 180, 27 },
		{ 186, 181, 27 },
		{ 186, 182, 27 },
		{ 186, 183, 27 },
		{ 186, 184, 27 },
		{ 186, 185, 27 },
		{ 187, 131, 27 },
		{ 187, 132, 27 },
		{ 187, 133, 27 },
		{ 187, 134, 27 },
		{ 187, 135, 27 },
		{ 187, 136, 27 },
		{ 187, 137, 27 },
		{ 187, 138, 27 },
		{ 187, 139, 27 },
		{ 187, 140, 27 },
		{ 187, 141, 27 },
		{ 187, 142, 27 },
		{ 187, 143, 27 },
		{ 187, 144, 27 },
		{ 187, 145, 27 },
		{ 187, 146, 27 },
		{ 187, 147, 27 },
		{ 187, 148, 27 },
		{ 187, 149, 27 },
		{ 187, 150, 27 },
		{ 187, 151, 27 },
		{ 187, 152, 27 },
		{ 187, 153, 27 },
		{ 187, 154, 27 },
		{ 187, 155, 27 },
		{ 187, 156, 27 },
		{ 187, 157, 27 },
		{ 187, 158, 27 },
		{ 187, 159, 27 },
		{ 187, 160, 27 },
		{ 187, 161, 27 },
		{ 187, 162, 27 },
		{ 187, 163, 27 },
		{ 187, 164, 27 },
		{ 187, 165, 27 },
		{ 187, 166, 27 },
		{ 187, 167, 27 },
		{ 187, 168, 27 },
		{ 187, 169, 27 },
		{ 187, 170, 27 },
		{ 187, 171, 27 },
		{ 187, 172, 27 },
		{ 187, 173, 27 },
		{ 187, 174, 27 },
		{ 187, 175, 27 },
		{ 187, 176, 27 },
		{ 187, 177, 27 },
		{ 187, 178, 27 },
		{ 187, 179, 27 },
		{ 187, 180, 27 },
		{ 187, 181, 27 },
		{ 187, 182, 27 },
		{ 187, 183, 27 },
		{ 187, 184, 27 },
		{ 187, 185, 27 },
		{ 187, 186, 27 },
		{ 188, 132, 27 },
		{ 188, 133, 27 },
		{ 188, 134, 27 },
		{ 188, 135, 27 },
		{ 188, 136, 27 },
		{ 188, 137, 27 },
		{ 188, 138, 27 },
		{ 188, 139, 27 },
		{ 188, 140, 27 },
		{ 188, 141, 27 },
		{ 188, 142, 27 },
		{ 188, 143, 27 },
		{ 188, 144, 27 },
		{ 188, 145, 27 },
		{ 188, 146, 27 },
		{ 188, 147, 27 },
		{ 188, 148, 27 },
		{ 188, 149, 27 },
		{ 188, 150, 27 },
		{ 188, 151, 27 },
		{ 188, 152, 27 },
		{ 188, 153, 27 },
		{ 188, 154, 27 },
		{ 188, 155, 27 },
		{ 188, 156, 27 },
		{ 188, 157, 27 },
		{ 188, 158, 27 },
		{ 188, 159, 27 },
		{ 188, 160, 27 },
		{ 188, 161, 27 },
		{ 188, 162, 27 },
		{ 188, 163, 27 },
		{ 188, 164, 27 },
		{ 188, 165, 27 },
		{ 188, 166, 27 },
		{ 188, 167, 27 },
		{ 188, 168, 27 },
		{ 188, 169, 27 },
		{ 188, 170, 27 },
		{ 188, 171, 27 },
		{ 188, 172, 27 },
		{ 188, 173, 27 },
		{ 188, 174, 27 },
		{ 188, 175, 27 },
		{ 188, 176, 27 },
		{ 188, 177, 27 },
		{ 188, 178, 27 },
		{ 188, 179, 27 },
		{ 188, 180, 27 },
		{ 188, 181, 27 },
		{ 188, 182, 27 },
		{ 188, 183, 27 },
		{ 188, 184, 27 },
		{ 188, 185, 27 },
		{ 188, 186, 27 },
		{ 188, 187, 27 },
		{ 189, 132, 27 },
		{ 189, 133, 27 },
		{ 189, 134, 27 },
		{ 189, 135, 27 },
		{ 189, 136, 27 },
		{ 189, 137, 27 },
		{ 189, 138, 27 },
		{ 189, 139, 27 },
		{ 189, 140, 27 },
		{ 189, 141, 27 },
		{ 189, 142, 27 },
		{ 189, 143, 27 },
		{ 189, 144, 27 },
		{ 189, 145, 27 },
		{ 189, 146, 27 },
		{ 189, 147, 27 },
		{ 189, 148, 27 },
		{ 189, 149, 27 },
		{ 189, 150, 27 },
		{ 189, 151, 27 },
		{ 189, 152, 27 },
		{ 189, 153, 27 },
		{ 189, 154, 27 },
		{ 189, 155, 27 },
		{ 189, 156, 27 },
		{ 189, 157, 27 },
		{ 189, 158, 27 },
		{ 189, 159, 27 },
		{ 189, 160, 27 },
		{ 189, 161, 27 },
		{ 189, 162, 27 },
		{ 189, 163, 27 },
		{ 189, 164, 27 },
		{ 189, 165, 27 },
		{ 189, 166, 27 },
		{ 189, 167, 27 },
		{ 189, 168, 27 },
		{ 189, 169, 27 },
		{ 189, 170, 27 },
		{ 189, 171, 27 },
		{ 189, 172, 27 },
		{ 189, 173, 27 },
		{ 189, 174, 27 },
		{ 189, 175, 27 },
		{ 189, 176, 27 },
		{ 189, 177, 27 },
		{ 189, 178, 27 },
		{ 189, 179, 27 },
		{ 189, 180, 27 },
		{ 189, 181, 27 },
		{ 189, 182, 27 },
		{ 189, 183, 27 },
		{ 189, 184, 27 },
		{ 189, 185, 27 },
		{ 189, 186, 27 },
		{ 189, 187, 27 },
		{ 189, 188, 27 },
		{ 190, 132, 27 },
		{ 190, 133, 27 },
		{ 190, 134, 27 },
		{ 190, 135, 27 },
		{ 190, 136, 27 },
		{ 190, 137, 27 },
		{ 190, 138, 27 },
		{ 190, 139, 27 },
		{ 190, 140, 27 },
		{ 190, 141, 27 },
		{ 190, 142, 27 },
		{ 190, 143, 27 },
		{ 190, 144, 27 },
		{ 190, 145, 27 },
		{ 190, 146, 27 },
		{ 190, 147, 27 },
		{ 190, 148, 27 },
		{ 190, 149, 27 },
		{ 190, 150, 27 },
		{ 190, 151, 27 },
		{ 190, 152, 27 },
		{ 190, 153, 27 },
		{ 190, 154, 27 },
		{ 190, 155, 27 },
		{ 190, 156, 27 },
		{ 190, 157, 27 },
		{ 190, 158, 27 },
		{ 190, 159, 27 },
		{ 190, 160, 27 },
		{ 190, 161, 27 },
		{ 190, 162, 27 },
		{ 190, 163, 27 },
		{ 190, 164, 27 },
		{ 190, 165, 27 },
		{ 190, 166, 27 },
		{ 190, 167, 27 },
		{ 190, 168, 27 },
		{ 190, 169, 27 },
		{ 190, 170, 27 },
		{ 190, 171, 27 },
		{ 190, 172, 27 },
		{ 190, 173, 27 },
		{ 190, 174, 27 },
		{ 190, 175, 27 },
		{ 190, 176, 27 },
		{ 190, 177, 27 },
		{ 190, 178, 27 },
		{ 190, 179, 27 },
		{ 190, 180, 27 },
		{ 190, 181, 27 },
		{ 190, 182, 27 },
		{ 190, 183, 27 },
		{ 190, 184, 27 },
		{ 190, 185, 27 },
		{ 190, 186, 27 },
		{ 190, 187, 27 },
		{ 190, 188, 27 },
		{ 190, 189, 27 },
		{ 191, 132, 27 },
		{ 191, 133, 27 },
		{ 191, 134, 27 },
		{ 191, 135, 27 },
		{ 191, 136, 27 },
		{ 191, 137, 27 },
		{ 191, 138, 27 },
		{ 191, 139, 27 },
		{ 191, 140, 27 },
		{ 191, 141, 27 },
		{ 191, 142, 27 },
		{ 191, 143, 27 },
		{ 191, 144, 27 },
		{ 191, 145, 27 },
		{ 191, 146, 27 },
		{ 191, 147, 27 },
		{ 191, 148, 27 },
		{ 191, 149, 27 },
		{ 191, 150, 27 },
		{ 191, 151, 27 },
		{ 191, 152, 27 },
		{ 191, 153, 27 },
		{ 191, 154, 27 },
		{ 191, 155, 27 },
		{ 191, 156, 27 },
		{ 191, 157, 27 },
		{ 191, 158, 27 },
		{ 191, 159, 27 },
		{ 191, 160, 27 },
		{ 191, 161, 27 },
		{ 191, 162, 27 },
		{ 191, 163, 27 },
		{ 191, 164, 27 },
		{ 191, 165, 27 },
		{ 191, 166, 27 },
		{ 191, 167, 27 },
		{ 191, 168, 27 },
		{ 191, 169, 27 },
		{ 191, 170, 27 },
		{ 191, 171, 27 },
		{ 191, 172, 27 },
		{ 191, 173, 27 },
		{ 191, 174, 27 },
		{ 191, 175, 27 },
		{ 191, 176, 27 },
		{ 191, 177, 27 },
		{ 191, 178, 27 },
		{ 191, 179, 27 },
		{ 191, 180, 27 },
		{ 191, 181, 27 },
		{ 191, 182, 27 },
		{ 191, 183, 27 },
		{ 191, 184, 27 },
		{ 191, 185, 27 },
		{ 191, 186, 27 },
		{ 191, 187, 27 },
		{ 191, 188, 27 },
		{ 191, 189, 27 },
		{ 191, 190, 27 },
		{ 192, 133, 27 },
		{ 192, 134, 27 },
		{ 192, 135, 27 },
		{ 192, 136, 27 },
		{ 192, 137, 27 },
		{ 192, 138, 27 },
		{ 192, 139, 27 },
		{ 192, 140, 27 },
		{ 192, 141, 27 },
		{ 192, 142, 27 },
		{ 192, 143, 27 },
		{ 192, 144, 27 },
		{ 192, 145, 27 },
		{ 192, 146, 27 },
		{ 192, 147, 27 },
		{ 192, 148, 27 },
		{ 192, 149, 27 },
		{ 192, 150, 27 },
		{ 192, 151, 27 },
		{ 192, 152, 27 },
		{ 192, 153, 27 },
		{ 192, 154, 27 },
		{ 192, 155, 27 },
		{ 192, 156, 27 },
		{ 192, 157, 27 },
		{ 192, 158, 27 },
		{ 192, 159, 27 },
		{ 192, 160, 27 },
		{ 192, 161, 27 },
		{ 192, 162, 27 },
		{ 192, 163, 27 },
		{ 192, 164, 27 },
		{ 192, 165, 27 },
		{ 192, 166, 27 },
		{ 192, 167, 27 },
		{ 192, 168, 27 },
		{ 192, 169, 27 },
		{ 192, 170, 27 },
		{ 192, 171, 27 },
		{ 192, 172, 27 },
		{ 192, 173, 27 },
		{ 192, 174, 27 },
		{ 192, 175, 27 },
		{ 192, 176, 27 },
		{ 192, 177, 27 },
		{ 192, 178, 27 },
		{ 192, 179, 27 },
		{ 192, 180, 27 },
		{ 192, 181, 27 },
		{ 192, 182, 27 },
		{ 192, 183, 27 },
		{ 192, 184, 27 },
		{ 192, 185, 27 },
		{ 192, 186, 27 },
		{ 192, 187, 27 },
		{ 192, 188, 27 },
		{ 192, 189, 27 },
		{ 192, 190, 27 },
		{ 192, 191, 27 },
		{ 193, 133, 27 },
		{ 193, 134, 27 },
		{ 193, 135, 27 },
		{ 193, 136, 27 },
		{ 193, 137, 27 },
		{ 193, 138, 27 },
		{ 193, 139, 27 },
		{ 193, 140, 27 },
		{ 193, 141, 27 },
		{ 193, 142, 27 },
		{ 193, 143, 27 },
		{ 193, 144, 27 },
		{ 193, 145, 27 },
		{ 193, 146, 27 },
		{ 193, 147, 27 },
		{ 193, 148, 27 },
		{ 193, 149, 27 },
		{ 193, 150, 27 },
		{ 193, 151, 27 },
		{ 193, 152, 27 },
		{ 193, 153, 27 },
		{ 193, 154, 27 },
		{ 193, 155, 27 },
		{ 193, 156, 27 },
		{ 193, 157, 27 },
		{ 193, 158, 27 },
		{ 193, 159, 27 },
		{ 193, 160, 27 },
		{ 193, 161, 27 },
		{ 193, 162, 27 },
		{ 193, 163, 27 },
		{ 193, 164, 27 },
		{ 193, 165, 27 },
		{ 193, 166, 27 },
		{ 193, 167, 27 },
		{ 193, 168, 27 },
		{ 193, 169, 27 },
		{ 193, 170, 27 },
		{ 193, 171, 27 },
		{ 193, 172, 27 },
		{ 193, 173, 27 },
		{ 193, 174, 27 },
		{ 193, 175, 27 },
		{ 193, 176, 27 },
		{ 193, 177, 27 },
		{ 193, 178, 27 },
		{ 193, 179, 27 },
		{ 193, 180, 27 },
		{ 193, 181, 27 },
		{ 193, 182, 27 },
		{ 193, 183, 27 },
		{ 193, 184, 27 },
		{ 193, 185, 27 },
		{ 193, 186, 27 },
		{ 193, 187, 27 },
		{ 193, 188, 27 },
		{ 193, 189, 27 },
		{ 193, 190, 27 },
		{ 193, 191, 27 },
		{ 193, 192, 27 },
		{ 194, 134, 27 },
		{ 194, 135, 27 },
		{ 194, 136, 27 },
		{ 194, 137, 27 },
		{ 194, 138, 27 },
		{ 194, 139, 27 },
		{ 194, 140, 27 },
		{ 194, 141, 27 },
		{ 194, 142, 27 },
		{ 194, 143, 27 },
		{ 194, 144, 27 },
		{ 194, 145, 27 },
		{ 194, 146, 27 },
		{ 194, 147, 27 },
		{ 194, 148, 27 },
		{ 194, 149, 27 },
		{ 194, 150, 27 },
		{ 194, 151, 27 },
		{ 194, 152, 27 },
		{ 194, 153, 27 },
		{ 194, 154, 27 },
		{ 194, 155, 27 },
		{ 194, 156, 27 },
		{ 194, 157, 27 },
		{ 194, 158, 27 },
		{ 194, 159, 27 },
		{ 194, 160, 27 },
		{ 194, 161, 27 },
		{ 194, 162, 27 },
		{ 194, 163, 27 },
		{ 194, 164, 27 },
		{ 194, 165, 27 },
		{ 194, 166, 27 },
		{ 194, 167, 27 },
		{ 194, 168, 27 },
		{ 194, 169, 27 },
		{ 194, 170, 27 },
		{ 194, 171, 27 },
		{ 194, 172, 27 },
		{ 194, 173, 27 },
		{ 194, 174, 27 },
		{ 194, 175, 27 },
		{ 194, 176, 27 },
		{ 194, 177, 27 },
		{ 194, 178, 27 },
		{ 194, 179, 27 },
		{ 194, 180, 27 },
		{ 194, 181, 27 },
		{ 194, 182, 27 },
		{ 194, 183, 27 },
		{ 194, 184, 27 },
		{ 194, 185, 27 },
		{ 194, 186, 27 },
		{ 194, 187, 27 },
		{ 194, 188, 27 },
		{ 194, 189, 27 },
		{ 194, 190, 27 },
		{ 194, 191, 27 },
		{ 194, 192, 27 },
		{ 194, 193, 27 },
		{ 195, 135, 27 },
		{ 195, 136, 27 },
		{ 195, 137, 27 },
		{ 195, 138, 27 },
		{ 195, 139, 27 },
		{ 195, 140, 27 },
		{ 195, 141, 27 },
		{ 195, 142, 27 },
		{ 195, 143, 27 },
		{ 195, 144, 27 },
		{ 195, 145, 27 },
		{ 195, 146, 27 },
		{ 195, 147, 27 },
		{ 195, 148, 27 },
		{ 195, 149, 27 },
		{ 195, 150, 27 },
		{ 195, 151, 27 },
		{ 195, 152, 27 },
		{ 195, 153, 27 },
		{ 195, 154, 27 },
		{ 195, 155, 27 },
		{ 195, 156, 27 },
		{ 195, 157, 27 },
		{ 195, 158, 27 },
		{ 195, 159, 27 },
		{ 195, 160, 27 },
		{ 195, 161, 27 },
		{ 195, 162, 27 },
		{ 195, 163, 27 },
		{ 195, 164, 27 },
		{ 195, 165, 27 },
		{ 195, 166, 27 },
		{ 195, 167, 27 },
		{ 195, 168, 27 },
		{ 195, 169, 27 },
		{ 195, 170, 27 },
		{ 195, 171, 27 },
		{ 195, 172, 27 },
		{ 195, 173, 27 },
		{ 195, 174, 27 },
		{ 195, 175, 27 },
		{ 195, 176, 27 },
		{ 195, 177, 27 },
		{ 195, 178, 27 },
		{ 195, 179, 27 },
		{ 195, 180, 27 },
		{ 195, 181, 27 },
		{ 195, 182, 27 },
		{ 195, 183, 27 },
		{ 195, 184, 27 },
		{ 195, 185, 27 },
		{ 195, 186, 27 },
		{ 195, 187, 27 },
		{ 195, 188, 27 },
		{ 195, 189, 27 },
		{ 195, 190, 27 },
		{ 195, 191, 27 },
		{ 195, 192, 27 },
		{ 195, 193, 27 },
		{ 195, 194, 27 },
		{ 196, 137, 27 },
		{ 196, 138, 27 },
		{ 196, 139, 27 },
		{ 196, 140, 27 },
		{ 196, 141, 27 },
		{ 196, 142, 27 },
		{ 196, 143, 27 },
		{ 196, 144, 27 },
		{ 196, 145, 27 },
		{ 196, 146, 27 },
		{ 196, 147, 27 },
		{ 196, 148, 27 },
		{ 196, 149, 27 },
		{ 196, 150, 27 },
		{ 196, 151, 27 },
		{ 196, 152, 27 },
		{ 196, 153, 27 },
		{ 196, 154, 27 },
		{ 196, 155, 27 },
		{ 196, 156, 27 },
		{ 196, 157, 27 },
		{ 196, 158, 27 },
		{ 196, 159, 27 },
		{ 196, 160, 27 },
		{ 196, 161, 27 },
		{ 196, 162, 27 },
		{ 196, 163, 27 },
		{ 196, 164, 27 },
		{ 196, 165, 27 },
		{ 196, 166, 27 },
		{ 196, 167, 27 },
		{ 196, 168, 27 },
		{ 196, 169, 27 },
		{ 196, 170, 27 },
		{ 196, 171, 27 },
		{ 196, 172, 27 },
		{ 196, 173, 27 },
		{ 196, 174, 27 },
		{ 196, 175, 27 },
		{ 196, 176, 27 },
		{ 196, 177, 27 },
		{ 196, 178, 27 },
		{ 196, 179, 27 },
		{ 196, 180, 27 },
		{ 196, 181, 27 },
		{ 196, 182, 27 },
		{ 196, 183, 27 },
		{ 196, 184, 27 },
		{ 196, 185, 27 },
		{ 196, 186, 27 },
		{ 196, 187, 27 },
		{ 196, 188, 27 },
		{ 196, 189, 27 },
		{ 196, 190, 27 },
		{ 196, 191, 27 },
		{ 196, 192, 27 },
		{ 196, 193, 27 },
		{ 196, 194, 27 },
		{ 196, 195, 27 },
		{ 197, 138, 27 },
		{ 197, 139, 27 },
		{ 197, 140, 27 },
		{ 197, 141, 27 },
		{ 197, 142, 27 },
		{ 197, 143, 27 },
		{ 197, 144, 27 },
		{ 197, 145, 27 },
		{ 197, 146, 27 },
		{ 197, 147, 27 },
		{ 197, 148, 27 },
		{ 197, 149, 27 },
		{ 197, 150, 27 },
		{ 197, 151, 27 },
		{ 197, 152, 27 },
		{ 197, 153, 27 },
		{ 197, 154, 27 },
		{ 197, 155, 27 },
		{ 197, 156, 27 },
		{ 197, 157, 27 },
		{ 197, 158, 27 },
		{ 197, 159, 27 },
		{ 197, 160, 27 },
		{ 197, 161, 27 },
		{ 197, 162, 27 },
		{ 197, 163, 27 },
		{ 197, 164, 27 },
		{ 197, 165, 27 },
		{ 197, 166, 27 },
		{ 197, 167, 27 },
		{ 197, 168, 27 },
		{ 197, 169, 27 },
		{ 197, 170, 27 },
		{ 197, 171, 27 },
		{ 197, 172, 27 },
		{ 197, 173, 27 },
		{ 197, 174, 27 },
		{ 197, 175, 27 },
		{ 197, 176, 27 },
		{ 197, 177, 27 },
		{ 197, 178, 27 },
		{ 197, 179, 27 },
		{ 197, 180, 27 },
		{ 197, 181, 27 },
		{ 197, 182, 27 },
		{ 197, 183, 27 },
		{ 197, 184, 27 },
		{ 197, 185, 27 },
		{ 197, 186, 27 },
		{ 197, 187, 27 },
		{ 197, 188, 27 },
		{ 197, 189, 27 },
		{ 197, 190, 27 },
		{ 197, 191, 27 },
		{ 197, 192, 27 },
		{ 197, 193, 27 },
		{ 197, 194, 27 },
		{ 197, 195, 27 },
		{ 197, 196, 27 },
		{ 198, 140, 27 },
		{ 198, 141, 27 },
		{ 198, 142, 27 },
		{ 198, 143, 27 },
		{ 198, 144, 27 },
		{ 198, 145, 27 },
		{ 198, 146, 27 },
		{ 198, 147, 27 },
		{ 198, 148, 27 },
		{ 198, 149, 27 },
		{ 198, 150, 27 },
		{ 198, 151, 27 },
		{ 198, 152, 27 },
		{ 198, 153, 27 },
		{ 198, 154, 27 },
		{ 198, 155, 27 },
		{ 198, 156, 27 },
		{ 198, 157, 27 },
		{ 198, 158, 27 },
		{ 198, 159, 27 },
		{ 198, 160, 27 },
		{ 198, 161, 27 },
		{ 198, 162, 27 },
		{ 198, 163, 27 },
		{ 198, 164, 27 },
		{ 198, 165, 27 },
		{ 198, 166, 27 },
		{ 198, 167, 27 },
		{ 198, 168, 27 },
		{ 198, 169, 27 },
		{ 198, 170, 27 },
		{ 198, 171, 27 },
		{ 198, 172, 27 },
		{ 198, 173, 27 },
		{ 198, 174, 27 },
		{ 198, 175, 27 },
		{ 198, 176, 27 },
		{ 198, 177, 27 },
		{ 198, 178, 27 },
		{ 198, 179, 27 },
		{ 198, 180, 27 },
		{ 198, 181, 27 },
		{ 198, 182, 27 },
		{ 198, 183, 27 },
		{ 198, 184, 27 },
		{ 198, 185, 27 },
		{ 198, 186, 27 },
		{ 198, 187, 27 },
		{ 198, 188, 27 },
		{ 198, 189, 27 },
		{ 198, 190, 27 },
		{ 198, 191, 27 },
		{ 198, 192, 27 },
		{ 198, 193, 27 },
		{ 198, 194, 27 },
		{ 198, 195, 27 },
		{ 198, 196, 27 },
		{ 198, 197, 27 },
		{ 199, 142, 27 },
		{ 199, 143, 27 },
		{ 199, 144, 27 },
		{ 199, 145, 27 },
		{ 199, 146, 27 },
		{ 199, 147, 27 },
		{ 199, 148, 27 },
		{ 199, 149, 27 },
		{ 199, 150, 27 },
		{ 199, 151, 27 },
		{ 199, 152, 27 },
		{ 199, 153, 27 },
		{ 199, 154, 27 },
		{ 199, 155, 27 },
		{ 199, 156, 27 },
		{ 199, 157, 27 },
		{ 199, 158, 27 },
		{ 199, 159, 27 },
		{ 199, 160, 27 },
		{ 199, 161, 27 },
		{ 199, 162, 27 },
		{ 199, 163, 27 },
		{ 199, 164, 27 },
		{ 199, 165, 27 },
		{ 199, 166, 27 },
		{ 199, 167, 27 },
		{ 199, 168, 27 },
		{ 199, 169, 27 },
		{ 199, 170, 27 },
		{ 199, 171, 27 },
		{ 199, 172, 27 },
		{ 199, 173, 27 },
		{ 199, 174, 27 },
		{ 199, 175, 27 },
		{ 199, 176, 27 },
		{ 199, 177, 27 },
		{ 199, 178, 27 },
		{ 199, 179, 27 },
		{ 199, 180, 27 },
		{ 199, 181, 27 },
		{ 199, 182, 27 },
		{ 199, 183, 27 },
		{ 199, 184, 27 },
		{ 199, 185, 27 },
		{ 199, 186, 27 },
		{ 199, 187, 27 },
		{ 199, 188, 27 },
		{ 199, 189, 27 },
		{ 199, 190, 27 },
		{ 199, 191, 27 },
		{ 199, 192, 27 },
		{ 199, 193, 27 },
		{ 199, 194, 27 },
		{ 199, 195, 27 },
		{ 199, 196, 27 },
		{ 199, 197, 27 },
		{ 199, 198, 27 },
		{ 200, 144, 27 },
		{ 200, 145, 27 },
		{ 200, 146, 27 },
		{ 200, 147, 27 },
		{ 200, 148, 27 },
		{ 200, 149, 27 },
		{ 200, 150, 27 },
		{ 200, 151, 27 },
		{ 200, 152, 27 },
		{ 200, 153, 27 },
		{ 200, 154, 27 },
		{ 200, 155, 27 },
		{ 200, 156, 27 },
		{ 200, 157, 27 },
		{ 200, 158, 27 },
		{ 200, 159, 27 },
		{ 200, 160, 27 },
		{ 200, 161, 27 },
		{ 200, 162, 27 },
		{ 200, 163, 27 },
		{ 200, 164, 27 },
		{ 200, 165, 27 },
		{ 200, 166, 27 },
		{ 200, 167, 27 },
		{ 200, 168, 27 },
		{ 200, 169, 27 },
		{ 200, 170, 27 },
		{ 200, 171, 27 },
		{ 200, 172, 27 },
		{ 200, 173, 27 },
		{ 200, 174, 27 },
		{ 200, 175, 27 },
		{ 200, 176, 27 },
		{ 200, 177, 27 },
		{ 200, 178, 27 },
		{ 200, 179, 27 },
		{ 200, 180, 27 },
		{ 200, 181, 27 },
		{ 200, 182, 27 },
		{ 200, 183, 27 },
		{ 200, 184, 27 },
		{ 200, 185, 27 },
		{ 200, 186, 27 },
		{ 200, 187, 27 },
		{ 200, 188, 27 },
		{ 200, 189, 27 },
		{ 200, 190, 27 },
		{ 200, 191, 27 },
		{ 200, 192, 27 },
		{ 200, 193, 27 },
		{ 200, 194, 27 },
		{ 200, 195, 27 },
		{ 200, 196, 27 },
		{ 200, 197, 27 },
		{ 200, 198, 27 },
		{ 200, 199, 27 },
		{ 201, 146, 27 },
		{ 201, 147, 27 },
		{ 201, 148, 27 },
		{ 201, 149, 27 },
		{ 201, 150, 27 },
		{ 201, 151, 27 },
		{ 201, 152, 27 },
		{ 201, 153, 27 },
		{ 201, 154, 27 },
		{ 201, 155, 27 },
		{ 201, 156, 27 },
		{ 201, 157, 27 },
		{ 201, 158, 27 },
		{ 201, 159, 27 },
		{ 201, 160, 27 },
		{ 201, 161, 27 },
		{ 201, 162, 27 },
		{ 201, 163, 27 },
		{ 201, 164, 27 },
		{ 201, 165, 27 },
		{ 201, 166, 27 },
		{ 201, 167, 27 },
		{ 201, 168, 27 },
		{ 201, 169, 27 },
		{ 201, 170, 27 },
		{ 201, 171, 27 },
		{ 201, 172, 27 },
		{ 201, 173, 27 },
		{ 201, 174, 27 },
		{ 201, 175, 27 },
		{ 201, 176, 27 },
		{ 201, 177, 27 },
		{ 201, 178, 27 },
		{ 201, 179, 27 },
		{ 201, 180, 27 },
		{ 201, 181, 27 },
		{ 201, 182, 27 },
		{ 201, 183, 27 },
		{ 201, 184, 27 },
		{ 201, 185, 27 },
		{ 201, 186, 27 },
		{ 201, 187, 27 },
		{ 201, 188, 27 },
		{ 201, 189, 27 },
		{ 201, 190, 27 },
		{ 201, 191, 27 },
		{ 201, 192, 27 },
		{ 201, 193, 27 },
		{ 201, 194, 27 },
		{ 201, 195, 27 },
		{ 201, 196, 27 },
		{ 201, 197, 27 },
		{ 201, 198, 27 },
		{ 201, 199, 27 },
		{ 201, 200, 27 },
		{ 202, 148, 27 },
		{ 202, 149, 27 },
		{ 202, 150, 27 },
		{ 202, 151, 27 },
		{ 202, 152, 27 },
		{ 202, 153, 27 },
		{ 202, 154, 27 },
		{ 202, 155, 27 },
		{ 202, 156, 27 },
		{ 202, 157, 27 },
		{ 202, 158, 27 },
		{ 202, 159, 27 },
		{ 202, 160, 27 },
		{ 202, 161, 27 },
		{ 202, 162, 27 },
		{ 202, 163, 27 },
		{ 202, 164, 27 },
		{ 202, 165, 27 },
		{ 202, 166, 27 },
		{ 202, 167, 27 },
		{ 202, 168, 27 },
		{ 202, 169, 27 },
		{ 202, 170, 27 },
		{ 202, 171, 27 },
		{ 202, 172, 27 },
		{ 202, 173, 27 },
		{ 202, 174, 27 },
		{ 202, 175, 27 },
		{ 202, 176, 27 },
		{ 202, 177, 27 },
		{ 202, 178, 27 },
		{ 202, 179, 27 },
		{ 202, 180, 27 },
		{ 202, 181, 27 },
		{ 202, 182, 27 },
		{ 202, 183, 27 },
		{ 202, 184, 27 },
		{ 202, 185, 27 },
		{ 202, 186, 27 },
		{ 202, 187, 27 },
		{ 202, 188, 27 },
		{ 202, 189, 27 },
		{ 202, 190, 27 },
		{ 202, 191, 27 },
		{ 202, 192, 27 },
		{ 202, 193, 27 },
		{ 202, 194, 27 },
		{ 202, 195, 27 },
		{ 202, 196, 27 },
		{ 202, 197, 27 },
		{ 202, 198, 27 },
		{ 202, 199, 27 },
		{ 202, 200, 27 },
		{ 202, 201, 27 },
		{ 203, 150, 27 },
		{ 203, 151, 27 },
		{ 203, 152, 27 },
		{ 203, 153, 27 },
		{ 203, 154, 27 },
		{ 203, 155, 27 },
		{ 203, 156, 27 },
		{ 203, 157, 27 },
		{ 203, 158, 27 },
		{ 203, 159, 27 },
		{ 203, 160, 27 },
		{ 203, 161, 27 },
		{ 203, 162, 27 },
		{ 203, 163, 27 },
		{ 203, 164, 27 },
		{ 203, 165, 27 },
		{ 203, 166, 27 },
		{ 203, 167, 27 },
		{ 203, 168, 27 },
		{ 203, 169, 27 },
		{ 203, 170, 27 },
		{ 203, 171, 27 },
		{ 203, 172, 27 },
		{ 203, 173, 27 },
		{ 203, 174, 27 },
		{ 203, 175, 27 },
		{ 203, 176, 27 },
		{ 203, 177, 27 },
		{ 203, 178, 27 },
		{ 203, 179, 27 },
		{ 203, 180, 27 },
		{ 203, 181, 27 },
		{ 203, 182, 27 },
		{ 203, 183, 27 },
		{ 203, 184, 27 },
		{ 203, 185, 27 },
		{ 203, 186, 27 },
		{ 203, 187, 27 },
		{ 203, 188, 27 },
		{ 203, 189, 27 },
		{ 203, 190, 27 },
		{ 203, 191, 27 },
		{ 203, 192, 27 },
		{ 203, 193, 27 },
		{ 203, 194, 27 },
		{ 203, 195, 27 },
		{ 203, 196, 27 },
		{ 203, 197, 27 },
		{ 203, 198, 27 },
		{ 203, 199, 27 },
		{ 203, 200, 27 },
		{ 203, 201, 27 },
		{ 203, 202, 27 },
		{ 204, 152, 27 },
		{ 204, 153, 27 },
		{ 204, 154, 27 },
		{ 204, 155, 27 },
		{ 204, 156, 27 },
		{ 204, 157, 27 },
		{ 204, 158, 27 },
		{ 204, 159, 27 },
		{ 204, 160, 27 },
		{ 204, 161, 27 },
		{ 204, 162, 27 },
		{ 204, 163, 27 },
		{ 204, 164, 27 },
		{ 204, 165, 27 },
		{ 204, 166, 27 },
		{ 204, 167, 27 },
		{ 204, 168, 27 },
		{ 204, 169, 27 },
		{ 204, 170, 27 },
		{ 204, 171, 27 },
		{ 204, 172, 27 },
		{ 204, 173, 27 },
		{ 204, 174, 27 },
		{ 204, 175, 27 },
		{ 204, 176, 27 },
		{ 204, 177, 27 },
		{ 204, 178, 27 },
		{ 204, 179, 27 },
		{ 204, 180, 27 },
		{ 204, 181, 27 },
		{ 204, 182, 27 },
		{ 204, 183, 27 },
		{ 204, 184, 27 },
		{ 204, 185, 27 },
		{ 204, 186, 27 },
		{ 204, 187, 27 },
		{ 204, 188, 27 },
		{ 204, 189, 27 },
		{ 204, 190, 27 },
		{ 204, 191, 27 },
		{ 204, 192, 27 },
		{ 204, 193, 27 },
		{ 204, 194, 27 },
		{ 204, 195, 27 },
		{ 204, 196, 27 },
		{ 204, 197, 27 },
		{ 204, 198, 27 },
		{ 204, 199, 27 },
		{ 204, 200, 27 },
		{ 204, 201, 27 },
		{ 204, 202, 27 },
		{ 204, 203, 27 },
		{ 205, 154, 27 },
		{ 205, 155, 27 },
		{ 205, 156, 27 },
		{ 205, 157, 27 },
		{ 205, 158, 27 },
		{ 205, 159, 27 },
		{ 205, 160, 27 },
		{ 205, 161, 27 },
		{ 205, 162, 27 },
		{ 205, 163, 27 },
		{ 205, 164, 27 },
		{ 205, 165, 27 },
		{ 205, 166, 27 },
		{ 205, 167, 27 },
		{ 205, 168, 27 },
		{ 205, 169, 27 },
		{ 205, 170, 27 },
		{ 205, 171, 27 },
		{ 205, 172, 27 },
		{ 205, 173, 27 },
		{ 205, 174, 27 },
		{ 205, 175, 27 },
		{ 205, 176, 27 },
		{ 205, 177, 27 },
		{ 205, 178, 27 },
		{ 205, 179, 27 },
		{ 205, 180, 27 },
		{ 205, 181, 27 },
		{ 205, 182, 27 },
		{ 205, 183, 27 },
		{ 205, 184, 27 },
		{ 205, 185, 27 },
		{ 205, 186, 27 },
		{ 205, 187, 27 },
		{ 205, 188, 27 },
		{ 205, 189, 27 },
		{ 205, 190, 27 },
		{ 205, 191, 27 },
		{ 205, 192, 27 },
		{ 205, 193, 27 },
		{ 205, 194, 27 },
		{ 205, 195, 27 },
		{ 205, 196, 27 },
		{ 205, 197, 27 },
		{ 205, 198, 27 },
		{ 205, 199, 27 },
		{ 205, 200, 27 },
		{ 205, 201, 27 },
		{ 205, 202, 27 },
		{ 205, 203, 27 },
		{ 205, 204, 27 },
		{ 206, 156, 27 },
		{ 206, 157, 27 },
		{ 206, 158, 27 },
		{ 206, 159, 27 },
		{ 206, 160, 27 },
		{ 206, 161, 27 },
		{ 206, 162, 27 },
		{ 206, 163, 27 },
		{ 206, 164, 27 },
		{ 206, 165, 27 },
		{ 206, 166, 27 },
		{ 206, 167, 27 },
		{ 206, 168, 27 },
		{ 206, 169, 27 },
		{ 206, 170, 27 },
		{ 206, 171, 27 },
		{ 206, 172, 27 },
		{ 206, 173, 27 },
		{ 206, 174, 27 },
		{ 206, 175, 27 },
		{ 206, 176, 27 },
		{ 206, 177, 27 },
		{ 206, 178, 27 },
		{ 206, 179, 27 },
		{ 206, 180, 27 },
		{ 206, 181, 27 },
		{ 206, 182, 27 },
		{ 206, 183, 27 },
		{ 206, 184, 27 },
		{ 206, 185, 27 },
		{ 206, 186, 27 },
		{ 206, 187, 27 },
		{ 206, 188, 27 },
		{ 206, 189, 27 },
		{ 206, 190, 27 },
		{ 206, 191, 27 },
		{ 206, 192, 27 },
		{ 206, 193, 27 },
		{ 206, 194, 27 },
		{ 206, 195, 27 },
		{ 206, 196, 27 },
		{ 206, 197, 27 },
		{ 206, 198, 27 },
		{ 206, 199, 27 },
		{ 206, 200, 27 },
		{ 206, 201, 27 },
		{ 206, 202, 27 },
		{ 206, 203, 27 },
		{ 206, 204, 27 },
		{ 206, 205, 27 },
		{ 207, 158, 27 },
		{ 207, 159, 27 },
		{ 207, 160, 27 },
		{ 207, 161, 27 },
		{ 207, 162, 27 },
		{ 207, 163, 27 },
		{ 207, 164, 27 },
		{ 207, 165, 27 },
		{ 207, 166, 27 },
		{ 207, 167, 27 },
		{ 207, 168, 27 },
		{ 207, 169, 27 },
		{ 207, 170, 27 },
		{ 207, 171, 27 },
		{ 207, 172, 27 },
		{ 207, 173, 27 },
		{ 207, 174, 27 },
		{ 207, 175, 27 },
		{ 207, 176, 27 },
		{ 207, 177, 27 },
		{ 207, 178, 27 },
		{ 207, 179, 27 },
		{ 207, 180, 27 },
		{ 207, 181, 27 },
		{ 207, 182, 27 },
		{ 207, 183, 27 },
		{ 207, 184, 27 },
		{ 207, 185, 27 },
		{ 207, 186, 27 },
		{ 207, 187, 27 },
		{ 207, 188, 27 },
		{ 207, 189, 27 },
		{ 207, 190, 27 },
		{ 207, 191, 27 },
		{ 207, 192, 27 },
		{ 207, 193, 27 },
		{ 207, 194, 27 },
		{ 207, 195, 27 },
		{ 207, 196, 27 },
		{ 207, 197, 27 },
		{ 207, 198, 27 },
		{ 207, 199, 27 },
		{ 207, 200, 27 },
		{ 207, 201, 27 },
		{ 207, 202, 27 },
		{ 207, 203, 27 },
		{ 207, 204, 27 },
		{ 207, 205, 27 },
		{ 207, 206, 27 },
		{ 208, 160, 27 },
		{ 208, 161, 27 },
		{ 208, 162, 27 },
		{ 208, 163, 27 },
		{ 208, 164, 27 },
		{ 208, 165, 27 },
		{ 208, 166, 27 },
		{ 208, 167, 27 },
		{ 208, 168, 27 },
		{ 208, 169, 27 },
		{ 208, 170, 27 },
		{ 208, 171, 27 },
		{ 208, 172, 27 },
		{ 208, 173, 27 },
		{ 208, 174, 27 },
		{ 208, 175, 27 },
		{ 208, 176, 27 },
		{ 208, 177, 27 },
		{ 208, 178, 27 },
		{ 208, 179, 27 },
		{ 208, 180, 27 },
		{ 208, 181, 27 },
		{ 208, 182, 27 },
		{ 208, 183, 27 },
		{ 208, 184, 27 },
		{ 208, 185, 27 },
		{ 208, 186, 27 },
		{ 208, 187, 27 },
		{ 208, 188, 27 },
		{ 208, 189, 27 },
		{ 208, 190, 27 },
		{ 208, 191, 27 },
		{ 208, 192, 27 },
		{ 208, 193, 27 },
		{ 208, 194, 27 },
		{ 208, 195, 27 },
		{ 208, 196, 27 },
		{ 208, 197, 27 },
		{ 208, 198, 27 },
		{ 208, 199, 27 },
		{ 208, 200, 27 },
		{ 208, 201, 27 },
		{ 208, 202, 27 },
		{ 208, 203, 27 },
		{ 208, 204, 27 },
		{ 208, 205, 27 },
		{ 208, 206, 27 },
		{ 208, 207, 27 },
		{ 209, 162, 27 },
		{ 209, 163, 27 },
		{ 209, 164, 27 },
		{ 209, 165, 27 },
		{ 209, 166, 27 },
		{ 209, 167, 27 },
		{ 209, 168, 27 },
		{ 209, 169, 27 },
		{ 209, 170, 27 },
		{ 209, 171, 27 },
		{ 209, 172, 27 },
		{ 209, 173, 27 },
		{ 209, 174, 27 },
		{ 209, 175, 27 },
		{ 209, 176, 27 },
		{ 209, 177, 27 },
		{ 209, 178, 27 },
		{ 209, 179, 27 },
		{ 209, 180, 27 },
		{ 209, 181, 27 },
		{ 209, 182, 27 },
		{ 209, 183, 27 },
		{ 209, 184, 27 },
		{ 209, 185, 27 },
		{ 209, 186, 27 },
		{ 209, 187, 27 },
		{ 209, 188, 27 },
		{ 209, 189, 27 },
		{ 209, 190, 27 },
		{ 209, 191, 27 },
		{ 209, 192, 27 },
		{ 209, 193, 27 },
		{ 209, 194, 27 },
		{ 209, 195, 27 },
		{ 209, 196, 27 },
		{ 209, 197, 27 },
		{ 209, 198, 27 },
		{ 209, 199, 27 },
		{ 209, 200, 27 },
		{ 209, 201, 27 },
		{ 209, 202, 27 },
		{ 209, 203, 27 },
		{ 209, 204, 27 },
		{ 209, 205, 27 },
		{ 209, 206, 27 },
		{ 209, 207, 27 },
		{ 209, 208, 27 },
		{ 210, 164, 27 },
		{ 210, 165, 27 },
		{ 210, 166, 27 },
		{ 210, 167, 27 },
		{ 210, 168, 27 },
		{ 210, 169, 27 },
		{ 210, 170, 27 },
		{ 210, 171, 27 },
		{ 210, 172, 27 },
		{ 210, 173, 27 },
		{ 210, 174, 27 },
		{ 210, 175, 27 },
		{ 210, 176, 27 },
		{ 210, 177, 27 },
		{ 210, 178, 27 },
		{ 210, 179, 27 },
		{ 210, 180, 27 },
		{ 210, 181, 27 },
		{ 210, 182, 27 },
		{ 210, 183, 27 },
		{ 210, 184, 27 },
		{ 210, 185, 27 },
		{ 210, 186, 27 },
		{ 210, 187, 27 },
		{ 210, 188, 27 },
		{ 210, 189, 27 },
		{ 210, 190, 27 },
		{ 210, 191, 27 },
		{ 210, 192, 27 },
		{ 210, 193, 27 },
		{ 210, 194, 27 },
		{ 210, 195, 27 },
		{ 210, 196, 27 },
		{ 210, 197, 27 },
		{ 210, 198, 27 },
		{ 210, 199, 27 },
		{ 210, 200, 27 },
		{ 210, 201, 27 },
		{ 210, 202, 27 },
		{ 210, 203, 27 },
		{ 210, 204, 27 },
		{ 210, 205, 27 },
		{ 210, 206, 27 },
		{ 210, 207, 27 },
		{ 210, 208, 27 },
		{ 210, 209, 27 },
		{ 211, 166, 27 },
		{ 211, 167, 27 },
		{ 211, 168, 27 },
		{ 211, 169, 27 },
		{ 211, 170, 27 },
		{ 211, 171, 27 },
		{ 211, 172, 27 },
		{ 211, 173, 27 },
		{ 211, 174, 27 },
		{ 211, 175, 27 },
		{ 211, 176, 27 },
		{ 211, 177, 27 },
		{ 211, 178, 27 },
		{ 211, 179, 27 },
		{ 211, 180, 27 },
		{ 211, 181, 27 },
		{ 211, 182, 27 },
		{ 211, 183, 27 },
		{ 211, 184, 27 },
		{ 211, 185, 27 },
		{ 211, 186, 27 },
		{ 211, 187, 27 },
		{ 211, 188, 27 },
		{ 211, 189, 27 },
		{ 211, 190, 27 },
		{ 211, 191, 27 },
		{ 211, 192, 27 },
		{ 211, 193, 27 },
		{ 211, 194, 27 },
		{ 211, 195, 27 },
		{ 211, 196, 27 },
		{ 211, 197, 27 },
		{ 211, 198, 27 },
		{ 211, 199, 27 },
		{ 211, 200, 27 },
		{ 211, 201, 27 },
		{ 211, 202, 27 },
		{ 211, 203, 27 },
		{ 211, 204, 27 },
		{ 211, 205, 27 },
		{ 211, 206, 27 },
		{ 211, 207, 27 },
		{ 211, 208, 27 },
		{ 211, 209, 27 },
		{ 211, 210, 27 },
		{ 212, 168, 27 },
		{ 212, 169, 27 },
		{ 212, 170, 27 },
		{ 212, 171, 27 },
		{ 212, 172, 27 },
		{ 212, 173, 27 },
		{ 212, 174, 27 },
		{ 212, 175, 27 },
		{ 212, 176, 27 },
		{ 212, 177, 27 },
		{ 212, 178, 27 },
		{ 212, 179, 27 },
		{ 212, 180, 27 },
		{ 212, 181, 27 },
		{ 212, 182, 27 },
		{ 212, 183, 27 },
		{ 212, 184, 27 },
		{ 212, 185, 27 },
		{ 212, 186, 27 },
		{ 212, 187, 27 },
		{ 212, 188, 27 },
		{ 212, 189, 27 },
		{ 212, 190, 27 },
		{ 212, 191, 27 },
		{ 212, 192, 27 },
		{ 212, 193, 27 },
		{ 212, 194, 27 },
		{ 212, 195, 27 },
		{ 212, 196, 27 },
		{ 212, 197, 27 },
		{ 212, 198, 27 },
		{ 212, 199, 27 },
		{ 212, 200, 27 },
		{ 212, 201, 27 },
		{ 212, 202, 27 },
		{ 212, 203, 27 },
		{ 212, 204, 27 },
		{ 212, 205, 27 },
		{ 212, 206, 27 },
		{ 212, 207, 27 },
		{ 212, 208, 27 },
		{ 212, 209, 27 },
		{ 212, 210, 27 },
		{ 212, 211, 27 },
		{ 213, 170, 27 },
		{ 213, 171, 27 },
		{ 213, 172, 27 },
		{ 213, 173, 27 },
		{ 213, 174, 27 },
		{ 213, 175, 27 },
		{ 213, 176, 27 },
		{ 213, 177, 27 },
		{ 213, 178, 27 },
		{ 213, 179, 27 },
		{ 213, 180, 27 },
		{ 213, 181, 27 },
		{ 213, 182, 27 },
		{ 213, 183, 27 },
		{ 213, 184, 27 },
		{ 213, 185, 27 },
		{ 213, 186, 27 },
		{ 213, 187, 27 },
		{ 213, 188, 27 },
		{ 213, 189, 27 },
		{ 213, 190, 27 },
		{ 213, 191, 27 },
		{ 213, 192, 27 },
		{ 213, 193, 27 },
		{ 213, 194, 27 },
		{ 213, 195, 27 },
		{ 213, 196, 27 },
		{ 213, 197, 27 },
		{ 213, 198, 27 },
		{ 213, 199, 27 },
		{ 213, 200, 27 },
		{ 213, 201, 27 },
		{ 213, 202, 27 },
		{ 213, 203, 27 },
		{ 213, 204, 27 },
		{ 213, 205, 27 },
		{ 213, 206, 27 },
		{ 213, 207, 27 },
		{ 213, 208, 27 },
		{ 213, 209, 27 },
		{ 213, 210, 27 },
		{ 213, 211, 27 },
		{ 213, 212, 27 },
		{ 214, 172, 27 },
		{ 214, 173, 27 },
		{ 214, 174, 27 },
		{ 214, 175, 27 },
		{ 214, 176, 27 },
		{ 214, 177, 27 },
		{ 214, 178, 27 },
		{ 214, 179, 27 },
		{ 214, 180, 27 },
		{ 214, 181, 27 },
		{ 214, 182, 27 },
		{ 214, 183, 27 },
		{ 214, 184, 27 },
		{ 214, 185, 27 },
		{ 214, 186, 27 },
		{ 214, 187, 27 },
		{ 214, 188, 27 },
		{ 214, 189, 27 },
		{ 214, 190, 27 },
		{ 214, 191, 27 },
		{ 214, 192, 27 },
		{ 214, 193, 27 },
		{ 214, 194, 27 },
		{ 214, 195, 27 },
		{ 214, 196, 27 },
		{ 214, 197, 27 },
		{ 214, 198, 27 },
		{ 214, 199, 27 },
		{ 214, 200, 27 },
		{ 214, 201, 27 },
		{ 214, 202, 27 },
		{ 214, 203, 27 },
		{ 214, 204, 27 },
		{ 214, 205, 27 },
		{ 214, 206, 27 },
		{ 214, 207, 27 },
		{ 214, 208, 27 },
		{ 214, 209, 27 },
		{ 214, 210, 27 },
		{ 214, 211, 27 },
		{ 214, 212, 27 },
		{ 214, 213, 27 },
		{ 215, 174, 27 },
		{ 215, 175, 27 },
		{ 215, 176, 27 },
		{ 215, 177, 27 },
		{ 215, 178, 27 },
		{ 215, 179, 27 },
		{ 215, 180, 27 },
		{ 215, 181, 27 },
		{ 215, 182, 27 },
		{ 215, 183, 27 },
		{ 215, 184, 27 },
		{ 215, 185, 27 },
		{ 215, 186, 27 },
		{ 215, 187, 27 },
		{ 215, 188, 27 },
		{ 215, 189, 27 },
		{ 215, 190, 27 },
		{ 215, 191, 27 },
		{ 215, 192, 27 },
		{ 215, 193, 27 },
		{ 215, 194, 27 },
		{ 215, 195, 27 },
		{ 215, 196, 27 },
		{ 215, 197, 27 },
		{ 215, 198, 27 },
		{ 215, 199, 27 },
		{ 215, 200, 27 },
		{ 215, 201, 27 },
		{ 215, 202, 27 },
		{ 215, 203, 27 },
		{ 215, 204, 27 },
		{ 215, 205, 27 },
		{ 215, 206, 27 },
		{ 215, 207, 27 },
		{ 215, 208, 27 },
		{ 215, 209, 27 },
		{ 215, 210, 27 },
		{ 215, 211, 27 },
		{ 215, 212, 27 },
		{ 215, 213, 27 },
		{ 215, 214, 27 },
		{ 216, 176, 27 },
		{ 216, 177, 27 },
		{ 216, 178, 27 },
		{ 216, 179, 27 },
		{ 216, 180, 27 },
		{ 216, 181, 27 },
		{ 216, 182, 27 },
		{ 216, 183, 27 },
		{ 216, 184, 27 },
		{ 216, 185, 27 },
		{ 216, 186, 27 },
		{ 216, 187, 27 },
		{ 216, 188, 27 },
		{ 216, 189, 27 },
		{ 216, 190, 27 },
		{ 216, 191, 27 },
		{ 216, 192, 27 },
		{ 216, 193, 27 },
		{ 216, 194, 27 },
		{ 216, 195, 27 },
		{ 216, 196, 27 },
		{ 216, 197, 27 },
		{ 216, 198, 27 },
		{ 216, 199, 27 },
		{ 216, 200, 27 },
		{ 216, 201, 27 },
		{ 216, 202, 27 },
		{ 216, 203, 27 },
		{ 216, 204, 27 },
		{ 216, 205, 27 },
		{ 216, 206, 27 },
		{ 216, 207, 27 },
		{ 216, 208, 27 },
		{ 216, 209, 27 },
		{ 216, 210, 27 },
		{ 216, 211, 27 },
		{ 216, 212, 27 },
		{ 216, 213, 27 },
		{ 216, 214, 27 },
		{ 216, 215, 27 },
		{ 217, 178, 27 },
		{ 217, 179, 27 },
		{ 217, 180, 27 },
		{ 217, 181, 27 },
		{ 217, 182, 27 },
		{ 217, 183, 27 },
		{ 217, 184, 27 },
		{ 217, 185, 27 },
		{ 217, 186, 27 },
		{ 217, 187, 27 },
		{ 217, 188, 27 },
		{ 217, 189, 27 },
		{ 217, 190, 27 },
		{ 217, 191, 27 },
		{ 217, 192, 27 },
		{ 217, 193, 27 },
		{ 217, 194, 27 },
		{ 217, 195, 27 },
		{ 217, 196, 27 },
		{ 217, 197, 27 },
		{ 217, 198, 27 },
		{ 217, 199, 27 },
		{ 217, 200, 27 },
		{ 217, 201, 27 },
		{ 217, 202, 27 },
		{ 217, 203, 27 },
		{ 217, 204, 27 },
		{ 217, 205, 27 },
		{ 217, 206, 27 },
		{ 217, 207, 27 },
		{ 217, 208, 27 },
		{ 217, 209, 27 },
		{ 217, 210, 27 },
		{ 217, 211, 27 },
		{ 217, 212, 27 },
		{ 217, 213, 27 },
		{ 217, 214, 27 },
		{ 217, 215, 27 },
		{ 217, 216, 27 },
		{ 218, 180, 27 },
		{ 218, 181, 27 },
		{ 218, 182, 27 },
		{ 218, 183, 27 },
		{ 218, 184, 27 },
		{ 218, 185, 27 },
		{ 218, 186, 27 },
		{ 218, 187, 27 },
		{ 218, 188, 27 },
		{ 218, 189, 27 },
		{ 218, 190, 27 },
		{ 218, 191, 27 },
		{ 218, 192, 27 },
		{ 218, 193, 27 },
		{ 218, 194, 27 },
		{ 218, 195, 27 },
		{ 218, 196, 27 },
		{ 218, 197, 27 },
		{ 218, 198, 27 },
		{ 218, 199, 27 },
		{ 218, 200, 27 },
		{ 218, 201, 27 },
		{ 218, 202, 27 },
		{ 218, 203, 27 },
		{ 218, 204, 27 },
		{ 218, 205, 27 },
		{ 218, 206, 27 },
		{ 218, 207, 27 },
		{ 218, 208, 27 },
		{ 218, 209, 27 },
		{ 218, 210, 27 },
		{ 218, 211, 27 },
		{ 218, 212, 27 },
		{ 218, 213, 27 },
		{ 218, 214, 27 },
		{ 218, 215, 27 },
		{ 218, 216, 27 },
		{ 218, 217, 27 },
		{ 219, 182, 27 },
		{ 219, 183, 27 },
		{ 219, 184, 27 },
		{ 219, 185, 27 },
		{ 219, 186, 27 },
		{ 219, 187, 27 },
		{ 219, 188, 27 },
		{ 219, 189, 27 },
		{ 219, 190, 27 },
		{ 219, 191, 27 },
		{ 219, 192, 27 },
		{ 219, 193, 27 },
		{ 219, 194, 27 },
		{ 219, 195, 27 },
		{ 219, 196, 27 },
		{ 219, 197, 27 },
		{ 219, 198, 27 },
		{ 219, 199, 27 },
		{ 219, 200, 27 },
		{ 219, 201, 27 },
		{ 219, 202, 27 },
		{ 219, 203, 27 },
		{ 219, 204, 27 },
		{ 219, 205, 27 },
		{ 219, 206, 27 },
		{ 219, 207, 27 },
		{ 219, 208, 27 },
		{ 219, 209, 27 },
		{ 219, 210, 27 },
		{ 219, 211, 27 },
		{ 219, 212, 27 },
		{ 219, 213, 27 },
		{ 219, 214, 27 },
		{ 219, 215, 27 },
		{ 219, 216, 27 },
		{ 219, 217, 27 },
		{ 219, 218, 27 },
		{ 220, 184, 27 },
		{ 220, 185, 27 },
		{ 220, 186, 27 },
		{ 220, 187, 27 },
		{ 220, 188, 27 },
		{ 220, 189, 27 },
		{ 220, 190, 27 },
		{ 220, 191, 27 },
		{ 220, 192, 27 },
		{ 220, 193, 27 },
		{ 220, 194, 27 },
		{ 220, 195, 27 },
		{ 220, 196, 27 },
		{ 220, 197, 27 },
		{ 220, 198, 27 },
		{ 220, 199, 27 },
		{ 220, 200, 27 },
		{ 220, 201, 27 },
		{ 220, 202, 27 },
		{ 220, 203, 27 },
		{ 220, 204, 27 },
		{ 220, 205, 27 },
		{ 220, 206, 27 },
		{ 220, 207, 27 },
		{ 220, 208, 27 },
		{ 220, 209, 27 },
		{ 220, 210, 27 },
		{ 220, 211, 27 },
		{ 220, 212, 27 },
		{ 220, 213, 27 },
		{ 220, 214, 27 },
		{ 220, 215, 27 },
		{ 220, 216, 27 },
		{ 220, 217, 27 },
		{ 220, 218, 27 },
		{ 220, 219, 27 },
		{ 221, 186, 27 },
		{ 221, 187, 27 },
		{ 221, 188, 27 },
		{ 221, 189, 27 },
		{ 221, 190, 27 },
		{ 221, 191, 27 },
		{ 221, 192, 27 },
		{ 221, 193, 27 },
		{ 221, 194, 27 },
		{ 221, 195, 27 },
		{ 221, 196, 27 },
		{ 221, 197, 27 },
		{ 221, 198, 27 },
		{ 221, 199, 27 },
		{ 221, 200, 27 },
		{ 221, 201, 27 },
		{ 221, 202, 27 },
		{ 221, 203, 27 },
		{ 221, 204, 27 },
		{ 221, 205, 27 },
		{ 221, 206, 27 },
		{ 221, 207, 27 },
		{ 221, 208, 27 },
		{ 221, 209, 27 },
		{ 221, 210, 27 },
		{ 221, 211, 27 },
		{ 221, 212, 27 },
		{ 221, 213, 27 },
		{ 221, 214, 27 },
		{ 221, 215, 27 },
		{ 221, 216, 27 },
		{ 221, 217, 27 },
		{ 221, 218, 27 },
		{ 221, 219, 27 },
		{ 221, 220, 27 },
		{ 222, 188, 27 },
		{ 222, 189, 27 },
		{ 222, 190, 27 },
		{ 222, 191, 27 },
		{ 222, 192, 27 },
		{ 222, 193, 27 },
		{ 222, 194, 27 },
		{ 222, 195, 27 },
		{ 222, 196, 27 },
		{ 222, 197, 27 },
		{ 222, 198, 27 },
		{ 222, 199, 27 },
		{ 222, 200, 27 },
		{ 222, 201, 27 },
		{ 222, 202, 27 },
		{ 222, 203, 27 },
		{ 222, 204, 27 },
		{ 222, 205, 27 },
		{ 222, 206, 27 },
		{ 222, 207, 27 },
		{ 222, 208, 27 },
		{ 222, 209, 27 },
		{ 222, 210, 27 },
		{ 222, 211, 27 },
		{ 222, 212, 27 },
		{ 222, 213, 27 },
		{ 222, 214, 27 },
		{ 222, 215, 27 },
		{ 222, 216, 27 },
		{ 222, 217, 27 },
		{ 222, 218, 27 },
		{ 222, 219, 27 },
		{ 222, 220, 27 },
		{ 222, 221, 27 },
		{ 223, 190, 27 },
		{ 223, 191, 27 },
		{ 223, 192, 27 },
		{ 223, 193, 27 },
		{ 223, 194, 27 },
		{ 223, 195, 27 },
		{ 223, 196, 27 },
		{ 223, 197, 27 },
		{ 223, 198, 27 },
		{ 223, 199, 27 },
		{ 223, 200, 27 },
		{ 223, 201, 27 },
		{ 223, 202, 27 },
		{ 223, 203, 27 },
		{ 223, 204, 27 },
		{ 223, 205, 27 },
		{ 223, 206, 27 },
		{ 223, 207, 27 },
		{ 223, 208, 27 },
		{ 223, 209, 27 },
		{ 223, 210, 27 },
		{ 223, 211, 27 },
		{ 223, 212, 27 },
		{ 223, 213, 27 },
		{ 223, 214, 27 },
		{ 223, 215, 27 },
		{ 223, 216, 27 },
		{ 223, 217, 27 },
		{ 223, 218, 27 },
		{ 223, 219, 27 },
		{ 223, 220, 27 },
		{ 223, 221, 27 },
		{ 223, 222, 27 },
		{ 224, 192, 27 },
		{ 224, 193, 27 },
		{ 224, 194, 27 },
		{ 224, 195, 27 },
		{ 224, 196, 27 },
		{ 224, 197, 27 },
		{ 224, 198, 27 },
		{ 224, 199, 27 },
		{ 224, 200, 27 },
		{ 224, 201, 27 },
		{ 224, 202, 27 },
		{ 224, 203, 27 },
		{ 224, 204, 27 },
		{ 224, 205, 27 },
		{ 224, 206, 27 },
		{ 224, 207, 27 },
		{ 224, 208, 27 },
		{ 224, 209, 27 },
		{ 224, 210, 27 },
		{ 224, 211, 27 },
		{ 224, 212, 27 },
		{ 224, 213, 27 },
		{ 224, 214, 27 },
		{ 224, 215, 27 },
		{ 224, 216, 27 },
		{ 224, 217, 27 },
		{ 224, 218, 27 },
		{ 224, 219, 27 },
		{ 224, 220, 27 },
		{ 224, 221, 27 },
		{ 224, 222, 27 },
		{ 224, 223, 27 },
		{ 225, 194, 27 },
		{ 225, 195, 27 },
		{ 225, 196, 27 },
		{ 225, 197, 27 },
		{ 225, 198, 27 },
		{ 225, 199, 27 },
		{ 225, 200, 27 },
		{ 225, 201, 27 },
		{ 225, 202, 27 },
		{ 225, 203, 27 },
		{ 225, 204, 27 },
		{ 225, 205, 27 },
		{ 225, 206, 27 },
		{ 225, 207, 27 },
		{ 225, 208, 27 },
		{ 225, 209, 27 },
		{ 225, 210, 27 },
		{ 225, 211, 27 },
		{ 225, 212, 27 },
		{ 225, 213, 27 },
		{ 225, 214, 27 },
		{ 225, 215, 27 },
		{ 225, 216, 27 },
		{ 225, 217, 27 },
		{ 225, 218, 27 },
		{ 225, 219, 27 },
		{ 225, 220, 27 },
		{ 225, 221, 27 },
		{ 225, 222, 27 },
		{ 225, 223, 27 },
		{ 225, 224, 27 },
		{ 226, 196, 27 },
		{ 226, 197, 27 },
		{ 226, 198, 27 },
		{ 226, 199, 27 },
		{ 226, 200, 27 },
		{ 226, 201, 27 },
		{ 226, 202, 27 },
		{ 226, 203, 27 },
		{ 226, 204, 27 },
		{ 226, 205, 27 },
		{ 226, 206, 27 },
		{ 226, 207, 27 },
		{ 226, 208, 27 },
		{ 226, 209, 27 },
		{ 226, 210, 27 },
		{ 226, 211, 27 },
		{ 226, 212, 27 },
		{ 226, 213, 27 },
		{ 226, 214, 27 },
		{ 226, 215, 27 },
		{ 226, 216, 27 },
		{ 226, 217, 27 },
		{ 226, 218, 27 },
		{ 226, 219, 27 },
		{ 226, 220, 27 },
		{ 226, 221, 27 },
		{ 226, 222, 27 },
		{ 226, 223, 27 },
		{ 226, 224, 27 },
		{ 226, 225, 27 },
		{ 227, 198, 27 },
		{ 227, 199, 27 },
		{ 227, 200, 27 },
		{ 227, 201, 27 },
		{ 227, 202, 27 },
		{ 227, 203, 27 },
		{ 227, 204, 27 },
		{ 227, 205, 27 },
		{ 227, 206, 27 },
		{ 227, 207, 27 },
		{ 227, 208, 27 },
		{ 227, 209, 27 },
		{ 227, 210, 27 },
		{ 227, 211, 27 },
		{ 227, 212, 27 },
		{ 227, 213, 27 },
		{ 227, 214, 27 },
		{ 227, 215, 27 },
		{ 227, 216, 27 },
		{ 227, 217, 27 },
		{ 227, 218, 27 },
		{ 227, 219, 27 },
		{ 227, 220, 27 },
		{ 227, 221, 27 },
		{ 227, 222, 27 },
		{ 227, 223, 27 },
		{ 227, 224, 27 },
		{ 227, 225, 27 },
		{ 227, 226, 27 },
		{ 228, 200, 27 },
		{ 228, 201, 27 },
		{ 228, 202, 27 },
		{ 228, 203, 27 },
		{ 228, 204, 27 },
		{ 228, 205, 27 },
		{ 228, 206, 27 },
		{ 228, 207, 27 },
		{ 228, 208, 27 },
		{ 228, 209, 27 },
		{ 228, 210, 27 },
		{ 228, 211, 27 },
		{ 228, 212, 27 },
		{ 228, 213, 27 },
		{ 228, 214, 27 },
		{ 228, 215, 27 },
		{ 228, 216, 27 },
		{ 228, 217, 27 },
		{ 228, 218, 27 },
		{ 228, 219, 27 },
		{ 228, 220, 27 },
		{ 228, 221, 27 },
		{ 228, 222, 27 },
		{ 228, 223, 27 },
		{ 228, 224, 27 },
		{ 228, 225, 27 },
		{ 228, 226, 27 },
		{ 228, 227, 27 },
		{ 229, 202, 27 },
		{ 229, 203, 27 },
		{ 229, 204, 27 },
		{ 229, 205, 27 },
		{ 229, 206, 27 },
		{ 229, 207, 27 },
		{ 229, 208, 27 },
		{ 229, 209, 27 },
		{ 229, 210, 27 },
		{ 229, 211, 27 },
		{ 229, 212, 27 },
		{ 229, 213, 27 },
		{ 229, 214, 27 },
		{ 229, 215, 27 },
		{ 229, 216, 27 },
		{ 229, 217, 27 },
		{ 229, 218, 27 },
		{ 229, 219, 27 },
		{ 229, 220, 27 },
		{ 229, 221, 27 },
		{ 229, 222, 27 },
		{ 229, 223, 27 },
		{ 229, 224, 27 },
		{ 229, 225, 27 },
		{ 229, 226, 27 },
		{ 229, 227, 27 },
		{ 229, 228, 27 },
		{ 230, 204, 27 },
		{ 230, 205, 27 },
		{ 230, 206, 27 },
		{ 230, 207, 27 },
		{ 230, 208, 27 },
		{ 230, 209, 27 },
		{ 230, 210, 27 },
		{ 230, 211, 27 },
		{ 230, 212, 27 },
		{ 230, 213, 27 },
		{ 230, 214, 27 },
		{ 230, 215, 27 },
		{ 230, 216, 27 },
		{ 230, 217, 27 },
		{ 230, 218, 27 },
		{ 230, 219, 27 },
		{ 230, 220, 27 },
		{ 230, 221, 27 },
		{ 230, 222, 27 },
		{ 230, 223, 27 },
		{ 230, 224, 27 },
		{ 230, 225, 27 },
		{ 230, 226, 27 },
		{ 230, 227, 27 },
		{ 230, 228, 27 },
		{ 230, 229, 27 },
		{ 231, 206, 27 },
		{ 231, 207, 27 },
		{ 231, 208, 27 },
		{ 231, 209, 27 },
		{ 231, 210, 27 },
		{ 231, 211, 27 },
		{ 231, 212, 27 },
		{ 231, 213, 27 },
		{ 231, 214, 27 },
		{ 231, 215, 27 },
		{ 231, 216, 27 },
		{ 231, 217, 27 },
		{ 231, 218, 27 },
		{ 231, 219, 27 },
		{ 231, 220, 27 },
		{ 231, 221, 27 },
		{ 231, 222, 27 },
		{ 231, 223, 27 },
		{ 231, 224, 27 },
		{ 231, 225, 27 },
		{ 231, 226, 27 },
		{ 231, 227, 27 },
		{ 231, 228, 27 },
		{ 231, 229, 27 },
		{ 231, 230, 27 },
		{ 232, 208, 27 },
		{ 232, 209, 27 },
		{ 232, 210, 27 },
		{ 232, 211, 27 },
		{ 232, 212, 27 },
		{ 232, 213, 27 },
		{ 232, 214, 27 },
		{ 232, 215, 27 },
		{ 232, 216, 27 },
		{ 232, 217, 27 },
		{ 232, 218, 27 },
		{ 232, 219, 27 },
		{ 232, 220, 27 },
		{ 232, 221, 27 },
		{ 232, 222, 27 },
		{ 232, 223, 27 },
		{ 232, 224, 27 },
		{ 232, 225, 27 },
		{ 232, 226, 27 },
		{ 232, 227, 27 },
		{ 232, 228, 27 },
		{ 232, 229, 27 },
		{ 232, 230, 27 },
		{ 232, 231, 27 },
		{ 233, 210, 27 },
		{ 233, 211, 27 },
		{ 233, 212, 27 },
		{ 233, 213, 27 },
		{ 233, 214, 27 },
		{ 233, 215, 27 },
		{ 233, 216, 27 },
		{ 233, 217, 27 },
		{ 233, 218, 27 },
		{ 233, 219, 27 },
		{ 233, 220, 27 },
		{ 233, 221, 27 },
		{ 233, 222, 27 },
		{ 233, 223, 27 },
		{ 233, 224, 27 },
		{ 233, 225, 27 },
		{ 233, 226, 27 },
		{ 233, 227, 27 },
		{ 233, 228, 27 },
		{ 233, 229, 27 },
		{ 233, 230, 27 },
		{ 233, 231, 27 },
		{ 233, 232, 27 },
		{ 234, 212, 27 },
		{ 234, 213, 27 },
		{ 234, 214, 27 },
		{ 234, 215, 27 },
		{ 234, 216, 27 },
		{ 234, 217, 27 },
		{ 234, 218, 27 },
		{ 234, 219, 27 },
		{ 234, 220, 27 },
		{ 234, 221, 27 },
		{ 234, 222, 27 },
		{ 234, 223, 27 },
		{ 234, 224, 27 },
		{ 234, 225, 27 },
		{ 234, 226, 27 },
		{ 234, 227, 27 },
		{ 234, 228, 27 },
		{ 234, 229, 27 },
		{ 234, 230, 27 },
		{ 234, 231, 27 },
		{ 234, 232, 27 },
		{ 234, 233, 27 },
		{ 235, 214, 27 },
		{ 235, 215, 27 },
		{ 235, 216, 27 },
		{ 235, 217, 27 },
		{ 235, 218, 27 },
		{ 235, 219, 27 },
		{ 235, 220, 27 },
		{ 235, 221, 27 },
		{ 235, 222, 27 },
		{ 235, 223, 27 },
		{ 235, 224, 27 },
		{ 235, 225, 27 },
		{ 235, 226, 27 },
		{ 235, 227, 27 },
		{ 235, 228, 27 },
		{ 235, 229, 27 },
		{ 235, 230, 27 },
		{ 235, 231, 27 },
		{ 235, 232, 27 },
		{ 235, 233, 27 },
		{ 235, 234, 27 },
		{ 236, 216, 27 },
		{ 236, 217, 27 },
		{ 236, 218, 27 },
		{ 236, 219, 27 },
		{ 236, 220, 27 },
		{ 236, 221, 27 },
		{ 236, 222, 27 },
		{ 236, 223, 27 },
		{ 236, 224, 27 },
		{ 236, 225, 27 },
		{ 236, 226, 27 },
		{ 236, 227, 27 },
		{ 236, 228, 27 },
		{ 236, 229, 27 },
		{ 236, 230, 27 },
		{ 236, 231, 27 },
		{ 236, 232, 27 },
		{ 236, 233, 27 },
		{ 236, 234, 27 },
		{ 236, 235, 27 },
		{ 237, 218, 27 },
		{ 237, 219, 27 },
		{ 237, 220, 27 },
		{ 237, 221, 27 },
		{ 237, 222, 27 },
		{ 237, 223, 27 },
		{ 237, 224, 27 },
		{ 237, 225, 27 },
		{ 237, 226, 27 },
		{ 237, 227, 27 },
		{ 237, 228, 27 },
		{ 237, 229, 27 },
		{ 237, 230, 27 },
		{ 237, 231, 27 },
		{ 237, 232, 27 },
		{ 237, 233, 27 },
		{ 237, 234, 27 },
		{ 237, 235, 27 },
		{ 237, 236, 27 },
		{ 238, 220, 27 },
		{ 238, 221, 27 },
		{ 238, 222, 27 },
		{ 238, 223, 27 },
		{ 238, 224, 27 },
		{ 238, 225, 27 },
		{ 238, 226, 27 },
		{ 238, 227, 27 },
		{ 238, 228, 27 },
		{ 238, 229, 27 },
		{ 238, 230, 27 },
		{ 238, 231, 27 },
		{ 238, 232, 27 },
		{ 238, 233, 27 },
		{ 238, 234, 27 },
		{ 238, 235, 27 },
		{ 238, 236, 27 },
		{ 238, 237, 27 },
		{ 239, 222, 27 },
		{ 239, 223, 27 },
		{ 239, 224, 27 },
		{ 239, 225, 27 },
		{ 239, 226, 27 },
		{ 239, 227, 27 },
		{ 239, 228, 27 },
		{ 239, 229, 27 },
		{ 239, 230, 27 },
		{ 239, 231, 27 },
		{ 239, 232, 27 },
		{ 239, 233, 27 },
		{ 239, 234, 27 },
		{ 239, 235, 27 },
		{ 239, 236, 27 },
		{ 239, 237, 27 },
		{ 239, 238, 27 },
		{ 240, 224, 27 },
		{ 240, 225, 27 },
		{ 240, 226, 27 },
		{ 240, 227, 27 },
		{ 240, 228, 27 },
		{ 240, 229, 27 },
		{ 240, 230, 27 },
		{ 240, 231, 27 },
		{ 240, 232, 27 },
		{ 240, 233, 27 },
		{ 240, 234, 27 },
		{ 240, 235, 27 },
		{ 240, 236, 27 },
		{ 240, 237, 27 },
		{ 240, 238, 27 },
		{ 240, 239, 27 },
		{ 241, 226, 27 },
		{ 241, 227, 27 },
		{ 241, 228, 27 },
		{ 241, 229, 27 },
		{ 241, 230, 27 },
		{ 241, 231, 27 },
		{ 241, 232, 27 },
		{ 241, 233, 27 },
		{ 241, 234, 27 },
		{ 241, 235, 27 },
		{ 241, 236, 27 },
		{ 241, 237, 27 },
		{ 241, 238, 27 },
		{ 241, 239, 27 },
		{ 241, 240, 27 },
		{ 242, 228, 27 },
		{ 242, 229, 27 },
		{ 242, 230, 27 },
		{ 242, 231, 27 },
		{ 242, 232, 27 },
		{ 242, 233, 27 },
		{ 242, 234, 27 },
		{ 242, 235, 27 },
		{ 242, 236, 27 },
		{ 242, 237, 27 },
		{ 242, 238, 27 },
		{ 242, 239, 27 },
		{ 242, 240, 27 },
		{ 243, 230, 27 },
		{ 243, 231, 27 },
		{ 243, 232, 27 },
		{ 243, 233, 27 },
		{ 243, 234, 27 },
		{ 243, 235, 27 },
		{ 243, 236, 27 },
		{ 243, 237, 27 },
		{ 243, 238, 27 },
		{ 243, 239, 27 },
		{ 243, 240, 27 },
		{ 244, 232, 27 },
		{ 244, 233, 27 },
		{ 244, 234, 27 },
		{ 244, 235, 27 },
		{ 244, 236, 27 },
		{ 244, 237, 27 },
		{ 244, 238, 27 },
		{ 244, 239, 27 },
		{ 244, 240, 27 },
		{ 245, 234, 27 },
		{ 245, 235, 27 },
		{ 245, 236, 27 },
		{ 245, 237, 27 },
		{ 245, 238, 27 },
		{ 245, 239, 27 },
		{ 245, 240, 27 },
		{ 246, 236, 27 },
		{ 246, 237, 27 },
		{ 246, 238, 27 },
		{ 246, 239, 27 },
		{ 246, 240, 27 },
		{ 247, 238, 27 },
		{ 247, 239, 27 },
		{ 247, 240, 27 },
		{ 248, 240, 27 },
		{ 266, 274, 27 },
		{ 267, 274, 27 },
		{ 267, 275, 27 },
		{ 267, 276, 27 },
		{ 268, 274, 27 },
		{ 268, 275, 27 },
		{ 268, 276, 27 },
		{ 268, 277, 27 },
		{ 268, 278, 27 },
		{ 269, 274, 27 },
		{ 269, 275, 27 },
		{ 269, 276, 27 },
		{ 269, 277, 27 },
		{ 269, 278, 27 },
		{ 269, 279, 27 },
		{ 269, 280, 27 },
		{ 270, 274, 27 },
		{ 270, 275, 27 },
		{ 270, 276, 27 },
		{ 270, 277, 27 },
		{ 270, 278, 27 },
		{ 270, 279, 27 },
		{ 270, 280, 27 },
		{ 270, 281, 27 },
		{ 270, 282, 27 },
		{ 271, 274, 27 },
		{ 271, 275, 27 },
		{ 271, 276, 27 },
		{ 271, 277, 27 },
		{ 271, 278, 27 },
		{ 271, 279, 27 },
		{ 271, 280, 27 },
		{ 271, 281, 27 },
		{ 271, 282, 27 },
		{ 271, 283, 27 },
		{ 271, 284, 27 },
		{ 272, 274, 27 },
		{ 272, 275, 27 },
		{ 272, 276, 27 },
		{ 272, 277, 27 },
		{ 272, 278, 27 },
		{ 272, 279, 27 },
		{ 272, 280, 27 },
		{ 272, 281, 27 },
		{ 272, 282, 27 },
		{ 272, 283, 27 },
		{ 272, 284, 27 },
		{ 272, 285, 27 },
		{ 272, 286, 27 },
		{ 273, 274, 27 },
		{ 273, 275, 27 },
		{ 273, 276, 27 },
		{ 273, 277, 27 },
		{ 273, 278, 27 },
		{ 273, 279, 27 },
		{ 273, 280, 27 },
		{ 273, 281, 27 },
		{ 273, 282, 27 },
		{ 273, 283, 27 },
		{ 273, 284, 27 },
		{ 273, 285, 27 },
		{ 273, 286, 27 },
		{ 273, 287, 27 },
		{ 273, 288, 27 },
		{ 274, 275, 27 },
		{ 274, 276, 27 },
		{ 274, 277, 27 },
		{ 274, 278, 27 },
		{ 274, 279, 27 },
		{ 274, 280, 27 },
		{ 274, 281, 27 },
		{ 274, 282, 27 },
		{ 274, 283, 27 },
		{ 274, 284, 27 },
		{ 274, 285, 27 },
		{ 274, 286, 27 },
		{ 274, 287, 27 },
		{ 274, 288, 27 },
		{ 274, 289, 27 },
		{ 274, 290, 27 },
		{ 275, 276, 27 },
		{ 275, 277, 27 },
		{ 275, 278, 27 },
		{ 275, 279, 27 },
		{ 275, 280, 27 },
		{ 275, 281, 27 },
		{ 275, 282, 27 },
		{ 275, 283, 27 },
		{ 275, 284, 27 },
		{ 275, 285, 27 },
		{ 275, 286, 27 },
		{ 275, 287, 27 },
		{ 275, 288, 27 },
		{ 275, 289, 27 },
		{ 275, 290, 27 },
		{ 275, 291, 27 },
		{ 275, 292, 27 },
		{ 276, 277, 27 },
		{ 276, 278, 27 },
		{ 276, 279, 27 },
		{ 276, 280, 27 },
		{ 276, 281, 27 },
		{ 276, 282, 27 },
		{ 276, 283, 27 },
		{ 276, 284, 27 },
		{ 276, 285, 27 },
		{ 276, 286, 27 },
		{ 276, 287, 27 },
		{ 276, 288, 27 },
		{ 276, 289, 27 },
		{ 276, 290, 27 },
		{ 276, 291, 27 },
		{ 276, 292, 27 },
		{ 276, 293, 27 },
		{ 276, 294, 27 },
		{ 277, 278, 27 },
		{ 277, 279, 27 },
		{ 277, 280, 27 },
		{ 277, 281, 27 },
		{ 277, 282, 27 },
		{ 277, 283, 27 },
		{ 277, 284, 27 },
		{ 277, 285, 27 },
		{ 277, 286, 27 },
		{ 277, 287, 27 },
		{ 277, 288, 27 },
		{ 277, 289, 27 },
		{ 277, 290, 27 },
		{ 277, 291, 27 },
		{ 277, 292, 27 },
		{ 277, 293, 27 },
		{ 277, 294, 27 },
		{ 277, 295, 27 },
		{ 277, 296, 27 },
		{ 278, 279, 27 },
		{ 278, 280, 27 },
		{ 278, 281, 27 },
		{ 278, 282, 27 },
		{ 278, 283, 27 },
		{ 278, 284, 27 },
		{ 278, 285, 27 },
		{ 278, 286, 27 },
		{ 278, 287, 27 },
		{ 278, 288, 27 },
		{ 278, 289, 27 },
		{ 278, 290, 27 },
		{ 278, 291, 27 },
		{ 278, 292, 27 },
		{ 278, 293, 27 },
		{ 278, 294, 27 },
		{ 278, 295, 27 },
		{ 278, 296, 27 },
		{ 278, 297, 27 },
		{ 278, 298, 27 },
		{ 279, 280, 27 },
		{ 279, 281, 27 },
		{ 279, 282, 27 },
		{ 279, 283, 27 },
		{ 279, 284, 27 },
		{ 279, 285, 27 },
		{ 279, 286, 27 },
		{ 279, 287, 27 },
		{ 279, 288, 27 },
		{ 279, 289, 27 },
		{ 279, 290, 27 },
		{ 279, 291, 27 },
		{ 279, 292, 27 },
		{ 279, 293, 27 },
		{ 279, 294, 27 },
		{ 279, 295, 27 },
		{ 279, 296, 27 },
		{ 279, 297, 27 },
		{ 279, 298, 27 },
		{ 279, 299, 27 },
		{ 279, 300, 27 },
		{ 280, 281, 27 },
		{ 280, 282, 27 },
		{ 280, 283, 27 },
		{ 280, 284, 27 },
		{ 280, 285, 27 },
		{ 280, 286, 27 },
		{ 280, 287, 27 },
		{ 280, 288, 27 },
		{ 280, 289, 27 },
		{ 280, 290, 27 },
		{ 280, 291, 27 },
		{ 280, 292, 27 },
		{ 280, 293, 27 },
		{ 280, 294, 27 },
		{ 280, 295, 27 },
		{ 280, 296, 27 },
		{ 280, 297, 27 },
		{ 280, 298, 27 },
		{ 280, 299, 27 },
		{ 280, 300, 27 },
		{ 280, 301, 27 },
		{ 280, 302, 27 },
		{ 281, 282, 27 },
		{ 281, 283, 27 },
		{ 281, 284, 27 },
		{ 281, 285, 27 },
		{ 281, 286, 27 },
		{ 281, 287, 27 },
		{ 281, 288, 27 },
		{ 281, 289, 27 },
		{ 281, 290, 27 },
		{ 281, 291, 27 },
		{ 281, 292, 27 },
		{ 281, 293, 27 },
		{ 281, 294, 27 },
		{ 281, 295, 27 },
		{ 281, 296, 27 },
		{ 281, 297, 27 },
		{ 281, 298, 27 },
		{ 281, 299, 27 },
		{ 281, 300, 27 },
		{ 281, 301, 27 },
		{ 281, 302, 27 },
		{ 281, 303, 27 },
		{ 281, 304, 27 },
		{ 282, 283, 27 },
		{ 282, 284, 27 },
		{ 282, 285, 27 },
		{ 282, 286, 27 },
		{ 282, 287, 27 },
		{ 282, 288, 27 },
		{ 282, 289, 27 },
		{ 282, 290, 27 },
		{ 282, 291, 27 },
		{ 282, 292, 27 },
		{ 282, 293, 27 },
		{ 282, 294, 27 },
		{ 282, 295, 27 },
		{ 282, 296, 27 },
		{ 282, 297, 27 },
		{ 282, 298, 27 },
		{ 282, 299, 27 },
		{ 282, 300, 27 },
		{ 282, 301, 27 },
		{ 282, 302, 27 },
		{ 282, 303, 27 },
		{ 282, 304, 27 },
		{ 282, 305, 27 },
		{ 282, 306, 27 },
		{ 283, 284, 27 },
		{ 283, 285, 27 },
		{ 283, 286, 27 },
		{ 283, 287, 27 },
		{ 283, 288, 27 },
		{ 283, 289, 27 },
		{ 283, 290, 27 },
		{ 283, 291, 27 },
		{ 283, 292, 27 },
		{ 283, 293, 27 },
		{ 283, 294, 27 },
		{ 283, 295, 27 },
		{ 283, 296, 27 },
		{ 283, 297, 27 },
		{ 283, 298, 27 },
		{ 283, 299, 27 },
		{ 283, 300, 27 },
		{ 283, 301, 27 },
		{ 283, 302, 27 },
		{ 283, 303, 27 },
		{ 283, 304, 27 },
		{ 283, 305, 27 },
		{ 283, 306, 27 },
		{ 283, 307, 27 },
		{ 283, 308, 27 },
		{ 284, 285, 27 },
		{ 284, 286, 27 },
		{ 284, 287, 27 },
		{ 284, 288, 27 },
		{ 284, 289, 27 },
		{ 284, 290, 27 },
		{ 284, 291, 27 },
		{ 284, 292, 27 },
		{ 284, 293, 27 },
		{ 284, 294, 27 },
		{ 284, 295, 27 },
		{ 284, 296, 27 },
		{ 284, 297, 27 },
		{ 284, 298, 27 },
		{ 284, 299, 27 },
		{ 284, 300, 27 },
		{ 284, 301, 27 },
		{ 284, 302, 27 },
		{ 284, 303, 27 },
		{ 284, 304, 27 },
		{ 284, 305, 27 },
		{ 284, 306, 27 },
		{ 284, 307, 27 },
		{ 284, 308, 27 },
		{ 284, 309, 27 },
		{ 284, 310, 27 },
		{ 285, 286, 27 },
		{ 285, 287, 27 },
		{ 285, 288, 27 },
		{ 285, 289, 27 },
		{ 285, 290, 27 },
		{ 285, 291, 27 },
		{ 285, 292, 27 },
		{ 285, 293, 27 },
		{ 285, 294, 27 },
		{ 285, 295, 27 },
		{ 285, 296, 27 },
		{ 285, 297, 27 },
		{ 285, 298, 27 },
		{ 285, 299, 27 },
		{ 285, 300, 27 },
		{ 285, 301, 27 },
		{ 285, 302, 27 },
		{ 285, 303, 27 },
		{ 285, 304, 27 },
		{ 285, 305, 27 },
		{ 285, 306, 27 },
		{ 285, 307, 27 },
		{ 285, 308, 27 },
		{ 285, 309, 27 },
		{ 285, 310, 27 },
		{ 285, 311, 27 },
		{ 285, 312, 27 },
		{ 286, 287, 27 },
		{ 286, 288, 27 },
		{ 286, 289, 27 },
		{ 286, 290, 27 },
		{ 286, 291, 27 },
		{ 286, 292, 27 },
		{ 286, 293, 27 },
		{ 286, 294, 27 },
		{ 286, 295, 27 },
		{ 286, 296, 27 },
		{ 286, 297, 27 },
		{ 286, 298, 27 },
		{ 286, 299, 27 },
		{ 286, 300, 27 },
		{ 286, 301, 27 },
		{ 286, 302, 27 },
		{ 286, 303, 27 },
		{ 286, 304, 27 },
		{ 286, 305, 27 },
		{ 286, 306, 27 },
		{ 286, 307, 27 },
		{ 286, 308, 27 },
		{ 286, 309, 27 },
		{ 286, 310, 27 },
		{ 286, 311, 27 },
		{ 286, 312, 27 },
		{ 286, 313, 27 },
		{ 286, 314, 27 },
		{ 287, 288, 27 },
		{ 287, 289, 27 },
		{ 287, 290, 27 },
		{ 287, 291, 27 },
		{ 287, 292, 27 },
		{ 287, 293, 27 },
		{ 287, 294, 27 },
		{ 287, 295, 27 },
		{ 287, 296, 27 },
		{ 287, 297, 27 },
		{ 287, 298, 27 },
		{ 287, 299, 27 },
		{ 287, 300, 27 },
		{ 287, 301, 27 },
		{ 287, 302, 27 },
		{ 287, 303, 27 },
		{ 287, 304, 27 },
		{ 287, 305, 27 },
		{ 287, 306, 27 },
		{ 287, 307, 27 },
		{ 287, 308, 27 },
		{ 287, 309, 27 },
		{ 287, 310, 27 },
		{ 287, 311, 27 },
		{ 287, 312, 27 },
		{ 287, 313, 27 },
		{ 287, 314, 27 },
		{ 287, 315, 27 },
		{ 287, 316, 27 },
		{ 288, 289, 27 },
		{ 288, 290, 27 },
		{ 288, 291, 27 },
		{ 288, 292, 27 },
		{ 288, 293, 27 },
		{ 288, 294, 27 },
		{ 288, 295, 27 },
		{ 288, 296, 27 },
		{ 288, 297, 27 },
		{ 288, 298, 27 },
		{ 288, 299, 27 },
		{ 288, 300, 27 },
		{ 288, 301, 27 },
		{ 288, 302, 27 },
		{ 288, 303, 27 },
		{ 288, 304, 27 },
		{ 288, 305, 27 },
		{ 288, 306, 27 },
		{ 288, 307, 27 },
		{ 288, 308, 27 },
		{ 288, 309, 27 },
		{ 288, 310, 27 },
		{ 288, 311, 27 },
		{ 288, 312, 27 },
		{ 288, 313, 27 },
		{ 288, 314, 27 },
		{ 288, 315, 27 },
		{ 288, 316, 27 },
		{ 288, 317, 27 },
		{ 288, 318, 27 },
		{ 289, 290, 27 },
		{ 289, 291, 27 },
		{ 289, 292, 27 },
		{ 289, 293, 27 },
		{ 289, 294, 27 },
		{ 289, 295, 27 },
		{ 289, 296, 27 },
		{ 289, 297, 27 },
		{ 289, 298, 27 },
		{ 289, 299, 27 },
		{ 289, 300, 27 },
		{ 289, 301, 27 },
		{ 289, 302, 27 },
		{ 289, 303, 27 },
		{ 289, 304, 27 },
		{ 289, 305, 27 },
		{ 289, 306, 27 },
		{ 289, 307, 27 },
		{ 289, 308, 27 },
		{ 289, 309, 27 },
		{ 289, 310, 27 },
		{ 289, 311, 27 },
		{ 289, 312, 27 },
		{ 289, 313, 27 },
		{ 289, 314, 27 },
		{ 289, 315, 27 },
		{ 289, 316, 27 },
		{ 289, 317, 27 },
		{ 289, 318, 27 },
		{ 289, 319, 27 },
		{ 289, 320, 27 },
		{ 290, 291, 27 },
		{ 290, 292, 27 },
		{ 290, 293, 27 },
		{ 290, 294, 27 },
		{ 290, 295, 27 },
		{ 290, 296, 27 },
		{ 290, 297, 27 },
		{ 290, 298, 27 },
		{ 290, 299, 27 },
		{ 290, 300, 27 },
		{ 290, 301, 27 },
		{ 290, 302, 27 },
		{ 290, 303, 27 },
		{ 290, 304, 27 },
		{ 290, 305, 27 },
		{ 290, 306, 27 },
		{ 290, 307, 27 },
		{ 290, 308, 27 },
		{ 290, 309, 27 },
		{ 290, 310, 27 },
		{ 290, 311, 27 },
		{ 290, 312, 27 },
		{ 290, 313, 27 },
		{ 290, 314, 27 },
		{ 290, 315, 27 },
		{ 290, 316, 27 },
		{ 290, 317, 27 },
		{ 290, 318, 27 },
		{ 290, 319, 27 },
		{ 290, 320, 27 },
		{ 290, 321, 27 },
		{ 290, 322, 27 },
		{ 291, 292, 27 },
		{ 291, 293, 27 },
		{ 291, 294, 27 },
		{ 291, 295, 27 },
		{ 291, 296, 27 },
		{ 291, 297, 27 },
		{ 291, 298, 27 },
		{ 291, 299, 27 },
		{ 291, 300, 27 },
		{ 291, 301, 27 },
		{ 291, 302, 27 },
		{ 291, 303, 27 },
		{ 291, 304, 27 },
		{ 291, 305, 27 },
		{ 291, 306, 27 },
		{ 291, 307, 27 },
		{ 291, 308, 27 },
		{ 291, 309, 27 },
		{ 291, 310, 27 },
		{ 291, 311, 27 },
		{ 291, 312, 27 },
		{ 291, 313, 27 },
		{ 291, 314, 27 },
		{ 291, 315, 27 },
		{ 291, 316, 27 },
		{ 291, 317, 27 },
		{ 291, 318, 27 },
		{ 291, 319, 27 },
		{ 291, 320, 27 },
		{ 291, 321, 27 },
		{ 291, 322, 27 },
		{ 291, 323, 27 },
		{ 291, 324, 27 },
		{ 292, 293, 27 },
		{ 292, 294, 27 },
		{ 292, 295, 27 },
		{ 292, 296, 27 },
		{ 292, 297, 27 },
		{ 292, 298, 27 },
		{ 292, 299, 27 },
		{ 292, 300, 27 },
		{ 292, 301, 27 },
		{ 292, 302, 27 },
		{ 292, 303, 27 },
		{ 292, 304, 27 },
		{ 292, 305, 27 },
		{ 292, 306, 27 },
		{ 292, 307, 27 },
		{ 292, 308, 27 },
		{ 292, 309, 27 },
		{ 292, 310, 27 },
		{ 292, 311, 27 },
		{ 292, 312, 27 },
		{ 292, 313, 27 },
		{ 292, 314, 27 },
		{ 292, 315, 27 },
		{ 292, 316, 27 },
		{ 292, 317, 27 },
		{ 292, 318, 27 },
		{ 292, 319, 27 },
		{ 292, 320, 27 },
		{ 292, 321, 27 },
		{ 292, 322, 27 },
		{ 292, 323, 27 },
		{ 292, 324, 27 },
		{ 292, 325, 27 },
		{ 292, 326, 27 },
		{ 293, 294, 27 },
		{ 293, 295, 27 },
		{ 293, 296, 27 },
		{ 293, 297, 27 },
		{ 293, 298, 27 },
		{ 293, 299, 27 },
		{ 293, 300, 27 },
		{ 293, 301, 27 },
		{ 293, 302, 27 },
		{ 293, 303, 27 },
		{ 293, 304, 27 },
		{ 293, 305, 27 },
		{ 293, 306, 27 },
		{ 293, 307, 27 },
		{ 293, 308, 27 },
		{ 293, 309, 27 },
		{ 293, 310, 27 },
		{ 293, 311, 27 },
		{ 293, 312, 27 },
		{ 293, 313, 27 },
		{ 293, 314, 27 },
		{ 293, 315, 27 },
		{ 293, 316, 27 },
		{ 293, 317, 27 },
		{ 293, 318, 27 },
		{ 293, 319, 27 },
		{ 293, 320, 27 },
		{ 293, 321, 27 },
		{ 293, 322, 27 },
		{ 293, 323, 27 },
		{ 293, 324, 27 },
		{ 293, 325, 27 },
		{ 293, 326, 27 },
		{ 293, 327, 27 },
		{ 293, 328, 27 },
		{ 294, 295, 27 },
		{ 294, 296, 27 },
		{ 294, 297, 27 },
		{ 294, 298, 27 },
		{ 294, 299, 27 },
		{ 294, 300, 27 },
		{ 294, 301, 27 },
		{ 294, 302, 27 },
		{ 294, 303, 27 },
		{ 294, 304, 27 },
		{ 294, 305, 27 },
		{ 294, 306, 27 },
		{ 294, 307, 27 },
		{ 294, 308, 27 },
		{ 294, 309, 27 },
		{ 294, 310, 27 },
		{ 294, 311, 27 },
		{ 294, 312, 27 },
		{ 294, 313, 27 },
		{ 294, 314, 27 },
		{ 294, 315, 27 },
		{ 294, 316, 27 },
		{ 294, 317, 27 },
		{ 294, 318, 27 },
		{ 294, 319, 27 },
		{ 294, 320, 27 },
		{ 294, 321, 27 },
		{ 294, 322, 27 },
		{ 294, 323, 27 },
		{ 294, 324, 27 },
		{ 294, 325, 27 },
		{ 294, 326, 27 },
		{ 294, 327, 27 },
		{ 294, 328, 27 },
		{ 294, 329, 27 },
		{ 294, 330, 27 },
		{ 295, 296, 27 },
		{ 295, 297, 27 },
		{ 295, 298, 27 },
		{ 295, 299, 27 },
		{ 295, 300, 27 },
		{ 295, 301, 27 },
		{ 295, 302, 27 },
		{ 295, 303, 27 },
		{ 295, 304, 27 },
		{ 295, 305, 27 },
		{ 295, 306, 27 },
		{ 295, 307, 27 },
		{ 295, 308, 27 },
		{ 295, 309, 27 },
		{ 295, 310, 27 },
		{ 295, 311, 27 },
		{ 295, 312, 27 },
		{ 295, 313, 27 },
		{ 295, 314, 27 },
		{ 295, 315, 27 },
		{ 295, 316, 27 },
		{ 295, 317, 27 },
		{ 295, 318, 27 },
		{ 295, 319, 27 },
		{ 295, 320, 27 },
		{ 295, 321, 27 },
		{ 295, 322, 27 },
		{ 295, 323, 27 },
		{ 295, 324, 27 },
		{ 295, 325, 27 },
		{ 295, 326, 27 },
		{ 295, 327, 27 },
		{ 295, 328, 27 },
		{ 295, 329, 27 },
		{ 295, 330, 27 },
		{ 295, 331, 27 },
		{ 295, 332, 27 },
		{ 296, 297, 27 },
		{ 296, 298, 27 },
		{ 296, 299, 27 },
		{ 296, 300, 27 },
		{ 296, 301, 27 },
		{ 296, 302, 27 },
		{ 296, 303, 27 },
		{ 296, 304, 27 },
		{ 296, 305, 27 },
		{ 296, 306, 27 },
		{ 296, 307, 27 },
		{ 296, 308, 27 },
		{ 296, 309, 27 },
		{ 296, 310, 27 },
		{ 296, 311, 27 },
		{ 296, 312, 27 },
		{ 296, 313, 27 },
		{ 296, 314, 27 },
		{ 296, 315, 27 },
		{ 296, 316, 27 },
		{ 296, 317, 27 },
		{ 296, 318, 27 },
		{ 296, 319, 27 },
		{ 296, 320, 27 },
		{ 296, 321, 27 },
		{ 296, 322, 27 },
		{ 296, 323, 27 },
		{ 296, 324, 27 },
		{ 296, 325, 27 },
		{ 296, 326, 27 },
		{ 296, 327, 27 },
		{ 296, 328, 27 },
		{ 296, 329, 27 },
		{ 296, 330, 27 },
		{ 296, 331, 27 },
		{ 296, 332, 27 },
		{ 296, 333, 27 },
		{ 296, 334, 27 },
		{ 297, 298, 27 },
		{ 297, 299, 27 },
		{ 297, 300, 27 },
		{ 297, 301, 27 },
		{ 297, 302, 27 },
		{ 297, 303, 27 },
		{ 297, 304, 27 },
		{ 297, 305, 27 },
		{ 297, 306, 27 },
		{ 297, 307, 27 },
		{ 297, 308, 27 },
		{ 297, 309, 27 },
		{ 297, 310, 27 },
		{ 297, 311, 27 },
		{ 297, 312, 27 },
		{ 297, 313, 27 },
		{ 297, 314, 27 },
		{ 297, 315, 27 },
		{ 297, 316, 27 },
		{ 297, 317, 27 },
		{ 297, 318, 27 },
		{ 297, 319, 27 },
		{ 297, 320, 27 },
		{ 297, 321, 27 },
		{ 297, 322, 27 },
		{ 297, 323, 27 },
		{ 297, 324, 27 },
		{ 297, 325, 27 },
		{ 297, 326, 27 },
		{ 297, 327, 27 },
		{ 297, 328, 27 },
		{ 297, 329, 27 },
		{ 297, 330, 27 },
		{ 297, 331, 27 },
		{ 297, 332, 27 },
		{ 297, 333, 27 },
		{ 297, 334, 27 },
		{ 297, 335, 27 },
		{ 297, 336, 27 },
		{ 298, 299, 27 },
		{ 298, 300, 27 },
		{ 298, 301, 27 },
		{ 298, 302, 27 },
		{ 298, 303, 27 },
		{ 298, 304, 27 },
		{ 298, 305, 27 },
		{ 298, 306, 27 },
		{ 298, 307, 27 },
		{ 298, 308, 27 },
		{ 298, 309, 27 },
		{ 298, 310, 27 },
		{ 298, 311, 27 },
		{ 298, 312, 27 },
		{ 298, 313, 27 },
		{ 298, 314, 27 },
		{ 298, 315, 27 },
		{ 298, 316, 27 },
		{ 298, 317, 27 },
		{ 298, 318, 27 },
		{ 298, 319, 27 },
		{ 298, 320, 27 },
		{ 298, 321, 27 },
		{ 298, 322, 27 },
		{ 298, 323, 27 },
		{ 298, 324, 27 },
		{ 298, 325, 27 },
		{ 298, 326, 27 },
		{ 298, 327, 27 },
		{ 298, 328, 27 },
		{ 298, 329, 27 },
		{ 298, 330, 27 },
		{ 298, 331, 27 },
		{ 298, 332, 27 },
		{ 298, 333, 27 },
		{ 298, 334, 27 },
		{ 298, 335, 27 },
		{ 298, 336, 27 },
		{ 298, 337, 27 },
		{ 298, 338, 27 },
		{ 299, 300, 27 },
		{ 299, 301, 27 },
		{ 299, 302, 27 },
		{ 299, 303, 27 },
		{ 299, 304, 27 },
		{ 299, 305, 27 },
		{ 299, 306, 27 },
		{ 299, 307, 27 },
		{ 299, 308, 27 },
		{ 299, 309, 27 },
		{ 299, 310, 27 },
		{ 299, 311, 27 },
		{ 299, 312, 27 },
		{ 299, 313, 27 },
		{ 299, 314, 27 },
		{ 299, 315, 27 },
		{ 299, 316, 27 },
		{ 299, 317, 27 },
		{ 299, 318, 27 },
		{ 299, 319, 27 },
		{ 299, 320, 27 },
		{ 299, 321, 27 },
		{ 299, 322, 27 },
		{ 299, 323, 27 },
		{ 299, 324, 27 },
		{ 299, 325, 27 },
		{ 299, 326, 27 },
		{ 299, 327, 27 },
		{ 299, 328, 27 },
		{ 299, 329, 27 },
		{ 299, 330, 27 },
		{ 299, 331, 27 },
		{ 299, 332, 27 },
		{ 299, 333, 27 },
		{ 299, 334, 27 },
		{ 299, 335, 27 },
		{ 299, 336, 27 },
		{ 299, 337, 27 },
		{ 299, 338, 27 },
		{ 299, 339, 27 },
		{ 299, 340, 27 },
		{ 300, 301, 27 },
		{ 300, 302, 27 },
		{ 300, 303, 27 },
		{ 300, 304, 27 },
		{ 300, 305, 27 },
		{ 300, 306, 27 },
		{ 300, 307, 27 },
		{ 300, 308, 27 },
		{ 300, 309, 27 },
		{ 300, 310, 27 },
		{ 300, 311, 27 },
		{ 300, 312, 27 },
		{ 300, 313, 27 },
		{ 300, 314, 27 },
		{ 300, 315, 27 },
		{ 300, 316, 27 },
		{ 300, 317, 27 },
		{ 300, 318, 27 },
		{ 300, 319, 27 },
		{ 300, 320, 27 },
		{ 300, 321, 27 },
		{ 300, 322, 27 },
		{ 300, 323, 27 },
		{ 300, 324, 27 },
		{ 300, 325, 27 },
		{ 300, 326, 27 },
		{ 300, 327, 27 },
		{ 300, 328, 27 },
		{ 300, 329, 27 },
		{ 300, 330, 27 },
		{ 300, 331, 27 },
		{ 300, 332, 27 },
		{ 300, 333, 27 },
		{ 300, 334, 27 },
		{ 300, 335, 27 },
		{ 300, 336, 27 },
		{ 300, 337, 27 },
		{ 300, 338, 27 },
		{ 300, 339, 27 },
		{ 300, 340, 27 },
		{ 300, 341, 27 },
		{ 300, 342, 27 },
		{ 301, 302, 27 },
		{ 301, 303, 27 },
		{ 301, 304, 27 },
		{ 301, 305, 27 },
		{ 301, 306, 27 },
		{ 301, 307, 27 },
		{ 301, 308, 27 },
		{ 301, 309, 27 },
		{ 301, 310, 27 },
		{ 301, 311, 27 },
		{ 301, 312, 27 },
		{ 301, 313, 27 },
		{ 301, 314, 27 },
		{ 301, 315, 27 },
		{ 301, 316, 27 },
		{ 301, 317, 27 },
		{ 301, 318, 27 },
		{ 301, 319, 27 },
		{ 301, 320, 27 },
		{ 301, 321, 27 },
		{ 301, 322, 27 },
		{ 301, 323, 27 },
		{ 301, 324, 27 },
		{ 301, 325, 27 },
		{ 301, 326, 27 },
		{ 301, 327, 27 },
		{ 301, 328, 27 },
		{ 301, 329, 27 },
		{ 301, 330, 27 },
		{ 301, 331, 27 },
		{ 301, 332, 27 },
		{ 301, 333, 27 },
		{ 301, 334, 27 },
		{ 301, 335, 27 },
		{ 301, 336, 27 },
		{ 301, 337, 27 },
		{ 301, 338, 27 },
		{ 301, 339, 27 },
		{ 301, 340, 27 },
		{ 301, 341, 27 },
		{ 301, 342, 27 },
		{ 301, 343, 27 },
		{ 301, 344, 27 },
		{ 302, 303, 27 },
		{ 302, 304, 27 },
		{ 302, 305, 27 },
		{ 302, 306, 27 },
		{ 302, 307, 27 },
		{ 302, 308, 27 },
		{ 302, 309, 27 },
		{ 302, 310, 27 },
		{ 302, 311, 27 },
		{ 302, 312, 27 },
		{ 302, 313, 27 },
		{ 302, 314, 27 },
		{ 302, 315, 27 },
		{ 302, 316, 27 },
		{ 302, 317, 27 },
		{ 302, 318, 27 },
		{ 302, 319, 27 },
		{ 302, 320, 27 },
		{ 302, 321, 27 },
		{ 302, 322, 27 },
		{ 302, 323, 27 },
		{ 302, 324, 27 },
		{ 302, 325, 27 },
		{ 302, 326, 27 },
		{ 302, 327, 27 },
		{ 302, 328, 27 },
		{ 302, 329, 27 },
		{ 302, 330, 27 },
		{ 302, 331, 27 },
		{ 302, 332, 27 },
		{ 302, 333, 27 },
		{ 302, 334, 27 },
		{ 302, 335, 27 },
		{ 302, 336, 27 },
		{ 302, 337, 27 },
		{ 302, 338, 27 },
		{ 302, 339, 27 },
		{ 302, 340, 27 },
		{ 302, 341, 27 },
		{ 302, 342, 27 },
		{ 302, 343, 27 },
		{ 302, 344, 27 },
		{ 302, 345, 27 },
		{ 302, 346, 27 },
		{ 303, 304, 27 },
		{ 303, 305, 27 },
		{ 303, 306, 27 },
		{ 303, 307, 27 },
		{ 303, 308, 27 },
		{ 303, 309, 27 },
		{ 303, 310, 27 },
		{ 303, 311, 27 },
		{ 303, 312, 27 },
		{ 303, 313, 27 },
		{ 303, 314, 27 },
		{ 303, 315, 27 },
		{ 303, 316, 27 },
		{ 303, 317, 27 },
		{ 303, 318, 27 },
		{ 303, 319, 27 },
		{ 303, 320, 27 },
		{ 303, 321, 27 },
		{ 303, 322, 27 },
		{ 303, 323, 27 },
		{ 303, 324, 27 },
		{ 303, 325, 27 },
		{ 303, 326, 27 },
		{ 303, 327, 27 },
		{ 303, 328, 27 },
		{ 303, 329, 27 },
		{ 303, 330, 27 },
		{ 303, 331, 27 },
		{ 303, 332, 27 },
		{ 303, 333, 27 },
		{ 303, 334, 27 },
		{ 303, 335, 27 },
		{ 303, 336, 27 },
		{ 303, 337, 27 },
		{ 303, 338, 27 },
		{ 303, 339, 27 },
		{ 303, 340, 27 },
		{ 303, 341, 27 },
		{ 303, 342, 27 },
		{ 303, 343, 27 },
		{ 303, 344, 27 },
		{ 303, 345, 27 },
		{ 303, 346, 27 },
		{ 303, 347, 27 },
		{ 303, 348, 27 },
		{ 304, 305, 27 },
		{ 304, 306, 27 },
		{ 304, 307, 27 },
		{ 304, 308, 27 },
		{ 304, 309, 27 },
		{ 304, 310, 27 },
		{ 304, 311, 27 },
		{ 304, 312, 27 },
		{ 304, 313, 27 },
		{ 304, 314, 27 },
		{ 304, 315, 27 },
		{ 304, 316, 27 },
		{ 304, 317, 27 },
		{ 304, 318, 27 },
		{ 304, 319, 27 },
		{ 304, 320, 27 },
		{ 304, 321, 27 },
		{ 304, 322, 27 },
		{ 304, 323, 27 },
		{ 304, 324, 27 },
		{ 304, 325, 27 },
		{ 304, 326, 27 },
		{ 304, 327, 27 },
		{ 304, 328, 27 },
		{ 304, 329, 27 },
		{ 304, 330, 27 },
		{ 304, 331, 27 },
		{ 304, 332, 27 },
		{ 304, 333, 27 },
		{ 304, 334, 27 },
		{ 304, 335, 27 },
		{ 304, 336, 27 },
		{ 304, 337, 27 },
		{ 304, 338, 27 },
		{ 304, 339, 27 },
		{ 304, 340, 27 },
		{ 304, 341, 27 },
		{ 304, 342, 27 },
		{ 304, 343, 27 },
		{ 304, 344, 27 },
		{ 304, 345, 27 },
		{ 304, 346, 27 },
		{ 304, 347, 27 },
		{ 304, 348, 27 },
		{ 304, 349, 27 },
		{ 304, 350, 27 },
		{ 305, 306, 27 },
		{ 305, 307, 27 },
		{ 305, 308, 27 },
		{ 305, 309, 27 },
		{ 305, 310, 27 },
		{ 305, 311, 27 },
		{ 305, 312, 27 },
		{ 305, 313, 27 },
		{ 305, 314, 27 },
		{ 305, 315, 27 },
		{ 305, 316, 27 },
		{ 305, 317, 27 },
		{ 305, 318, 27 },
		{ 305, 319, 27 },
		{ 305, 320, 27 },
		{ 305, 321, 27 },
		{ 305, 322, 27 },
		{ 305, 323, 27 },
		{ 305, 324, 27 },
		{ 305, 325, 27 },
		{ 305, 326, 27 },
		{ 305, 327, 27 },
		{ 305, 328, 27 },
		{ 305, 329, 27 },
		{ 305, 330, 27 },
		{ 305, 331, 27 },
		{ 305, 332, 27 },
		{ 305, 333, 27 },
		{ 305, 334, 27 },
		{ 305, 335, 27 },
		{ 305, 336, 27 },
		{ 305, 337, 27 },
		{ 305, 338, 27 },
		{ 305, 339, 27 },
		{ 305, 340, 27 },
		{ 305, 341, 27 },
		{ 305, 342, 27 },
		{ 305, 343, 27 },
		{ 305, 344, 27 },
		{ 305, 345, 27 },
		{ 305, 346, 27 },
		{ 305, 347, 27 },
		{ 305, 348, 27 },
		{ 305, 349, 27 },
		{ 305, 350, 27 },
		{ 305, 351, 27 },
		{ 305, 352, 27 },
		{ 306, 307, 27 },
		{ 306, 308, 27 },
		{ 306, 309, 27 },
		{ 306, 310, 27 },
		{ 306, 311, 27 },
		{ 306, 312, 27 },
		{ 306, 313, 27 },
		{ 306, 314, 27 },
		{ 306, 315, 27 },
		{ 306, 316, 27 },
		{ 306, 317, 27 },
		{ 306, 318, 27 },
		{ 306, 319, 27 },
		{ 306, 320, 27 },
		{ 306, 321, 27 },
		{ 306, 322, 27 },
		{ 306, 323, 27 },
		{ 306, 324, 27 },
		{ 306, 325, 27 },
		{ 306, 326, 27 },
		{ 306, 327, 27 },
		{ 306, 328, 27 },
		{ 306, 329, 27 },
		{ 306, 330, 27 },
		{ 306, 331, 27 },
		{ 306, 332, 27 },
		{ 306, 333, 27 },
		{ 306, 334, 27 },
		{ 306, 335, 27 },
		{ 306, 336, 27 },
		{ 306, 337, 27 },
		{ 306, 338, 27 },
		{ 306, 339, 27 },
		{ 306, 340, 27 },
		{ 306, 341, 27 },
		{ 306, 342, 27 },
		{ 306, 343, 27 },
		{ 306, 344, 27 },
		{ 306, 345, 27 },
		{ 306, 346, 27 },
		{ 306, 347, 27 },
		{ 306, 348, 27 },
		{ 306, 349, 27 },
		{ 306, 350, 27 },
		{ 306, 351, 27 },
		{ 306, 352, 27 },
		{ 306, 353, 27 },
		{ 306, 354, 27 },
		{ 307, 308, 27 },
		{ 307, 309, 27 },
		{ 307, 310, 27 },
		{ 307, 311, 27 },
		{ 307, 312, 27 },
		{ 307, 313, 27 },
		{ 307, 314, 27 },
		{ 307, 315, 27 },
		{ 307, 316, 27 },
		{ 307, 317, 27 },
		{ 307, 318, 27 },
		{ 307, 319, 27 },
		{ 307, 320, 27 },
		{ 307, 321, 27 },
		{ 307, 322, 27 },
		{ 307, 323, 27 },
		{ 307, 324, 27 },
		{ 307, 325, 27 },
		{ 307, 326, 27 },
		{ 307, 327, 27 },
		{ 307, 328, 27 },
		{ 307, 329, 27 },
		{ 307, 330, 27 },
		{ 307, 331, 27 },
		{ 307, 332, 27 },
		{ 307, 333, 27 },
		{ 307, 334, 27 },
		{ 307, 335, 27 },
		{ 307, 336, 27 },
		{ 307, 337, 27 },
		{ 307, 338, 27 },
		{ 307, 339, 27 },
		{ 307, 340, 27 },
		{ 307, 341, 27 },
		{ 307, 342, 27 },
		{ 307, 343, 27 },
		{ 307, 344, 27 },
		{ 307, 345, 27 },
		{ 307, 346, 27 },
		{ 307, 347, 27 },
		{ 307, 348, 27 },
		{ 307, 349, 27 },
		{ 307, 350, 27 },
		{ 307, 351, 27 },
		{ 307, 352, 27 },
		{ 307, 353, 27 },
		{ 307, 354, 27 },
		{ 307, 355, 27 },
		{ 307, 356, 27 },
		{ 308, 309, 27 },
		{ 308, 310, 27 },
		{ 308, 311, 27 },
		{ 308, 312, 27 },
		{ 308, 313, 27 },
		{ 308, 314, 27 },
		{ 308, 315, 27 },
		{ 308, 316, 27 },
		{ 308, 317, 27 },
		{ 308, 318, 27 },
		{ 308, 319, 27 },
		{ 308, 320, 27 },
		{ 308, 321, 27 },
		{ 308, 322, 27 },
		{ 308, 323, 27 },
		{ 308, 324, 27 },
		{ 308, 325, 27 },
		{ 308, 326, 27 },
		{ 308, 327, 27 },
		{ 308, 328, 27 },
		{ 308, 329, 27 },
		{ 308, 330, 27 },
		{ 308, 331, 27 },
		{ 308, 332, 27 },
		{ 308, 333, 27 },
		{ 308, 334, 27 },
		{ 308, 335, 27 },
		{ 308, 336, 27 },
		{ 308, 337, 27 },
		{ 308, 338, 27 },
		{ 308, 339, 27 },
		{ 308, 340, 27 },
		{ 308, 341, 27 },
		{ 308, 342, 27 },
		{ 308, 343, 27 },
		{ 308, 344, 27 },
		{ 308, 345, 27 },
		{ 308, 346, 27 },
		{ 308, 347, 27 },
		{ 308, 348, 27 },
		{ 308, 349, 27 },
		{ 308, 350, 27 },
		{ 308, 351, 27 },
		{ 308, 352, 27 },
		{ 308, 353, 27 },
		{ 308, 354, 27 },
		{ 308, 355, 27 },
		{ 308, 356, 27 },
		{ 308, 357, 27 },
		{ 308, 358, 27 },
		{ 309, 310, 27 },
		{ 309, 311, 27 },
		{ 309, 312, 27 },
		{ 309, 313, 27 },
		{ 309, 314, 27 },
		{ 309, 315, 27 },
		{ 309, 316, 27 },
		{ 309, 317, 27 },
		{ 309, 318, 27 },
		{ 309, 319, 27 },
		{ 309, 320, 27 },
		{ 309, 321, 27 },
		{ 309, 322, 27 },
		{ 309, 323, 27 },
		{ 309, 324, 27 },
		{ 309, 325, 27 },
		{ 309, 326, 27 },
		{ 309, 327, 27 },
		{ 309, 328, 27 },
		{ 309, 329, 27 },
		{ 309, 330, 27 },
		{ 309, 331, 27 },
		{ 309, 332, 27 },
		{ 309, 333, 27 },
		{ 309, 334, 27 },
		{ 309, 335, 27 },
		{ 309, 336, 27 },
		{ 309, 337, 27 },
		{ 309, 338, 27 },
		{ 309, 339, 27 },
		{ 309, 340, 27 },
		{ 309, 341, 27 },
		{ 309, 342, 27 },
		{ 309, 343, 27 },
		{ 309, 344, 27 },
		{ 309, 345, 27 },
		{ 309, 346, 27 },
		{ 309, 347, 27 },
		{ 309, 348, 27 },
		{ 309, 349, 27 },
		{ 309, 350, 27 },
		{ 309, 351, 27 },
		{ 309, 352, 27 },
		{ 309, 353, 27 },
		{ 309, 354, 27 },
		{ 309, 355, 27 },
		{ 309, 356, 27 },
		{ 309, 357, 27 },
		{ 309, 358, 27 },
		{ 309, 359, 27 },
		{ 309, 360, 27 },
		{ 310, 311, 27 },
		{ 310, 312, 27 },
		{ 310, 313, 27 },
		{ 310, 314, 27 },
		{ 310, 315, 27 },
		{ 310, 316, 27 },
		{ 310, 317, 27 },
		{ 310, 318, 27 },
		{ 310, 319, 27 },
		{ 310, 320, 27 },
		{ 310, 321, 27 },
		{ 310, 322, 27 },
		{ 310, 323, 27 },
		{ 310, 324, 27 },
		{ 310, 325, 27 },
		{ 310, 326, 27 },
		{ 310, 327, 27 },
		{ 310, 328, 27 },
		{ 310, 329, 27 },
		{ 310, 330, 27 },
		{ 310, 331, 27 },
		{ 310, 332, 27 },
		{ 310, 333, 27 },
		{ 310, 334, 27 },
		{ 310, 335, 27 },
		{ 310, 336, 27 },
		{ 310, 337, 27 },
		{ 310, 338, 27 },
		{ 310, 339, 27 },
		{ 310, 340, 27 },
		{ 310, 341, 27 },
		{ 310, 342, 27 },
		{ 310, 343, 27 },
		{ 310, 344, 27 },
		{ 310, 345, 27 },
		{ 310, 346, 27 },
		{ 310, 347, 27 },
		{ 310, 348, 27 },
		{ 310, 349, 27 },
		{ 310, 350, 27 },
		{ 310, 351, 27 },
		{ 310, 352, 27 },
		{ 310, 353, 27 },
		{ 310, 354, 27 },
		{ 310, 355, 27 },
		{ 310, 356, 27 },
		{ 310, 357, 27 },
		{ 310, 358, 27 },
		{ 310, 359, 27 },
		{ 310, 360, 27 },
		{ 310, 361, 27 },
		{ 310, 362, 27 },
		{ 311, 312, 27 },
		{ 311, 313, 27 },
		{ 311, 314, 27 },
		{ 311, 315, 27 },
		{ 311, 316, 27 },
		{ 311, 317, 27 },
		{ 311, 318, 27 },
		{ 311, 319, 27 },
		{ 311, 320, 27 },
		{ 311, 321, 27 },
		{ 311, 322, 27 },
		{ 311, 323, 27 },
		{ 311, 324, 27 },
		{ 311, 325, 27 },
		{ 311, 326, 27 },
		{ 311, 327, 27 },
		{ 311, 328, 27 },
		{ 311, 329, 27 },
		{ 311, 330, 27 },
		{ 311, 331, 27 },
		{ 311, 332, 27 },
		{ 311, 333, 27 },
		{ 311, 334, 27 },
		{ 311, 335, 27 },
		{ 311, 336, 27 },
		{ 311, 337, 27 },
		{ 311, 338, 27 },
		{ 311, 339, 27 },
		{ 311, 340, 27 },
		{ 311, 341, 27 },
		{ 311, 342, 27 },
		{ 311, 343, 27 },
		{ 311, 344, 27 },
		{ 311, 345, 27 },
		{ 311, 346, 27 },
		{ 311, 347, 27 },
		{ 311, 348, 27 },
		{ 311, 349, 27 },
		{ 311, 350, 27 },
		{ 311, 351, 27 },
		{ 311, 352, 27 },
		{ 311, 353, 27 },
		{ 311, 354, 27 },
		{ 311, 355, 27 },
		{ 311, 356, 27 },
		{ 311, 357, 27 },
		{ 311, 358, 27 },
		{ 311, 359, 27 },
		{ 311, 360, 27 },
		{ 311, 361, 27 },
		{ 311, 362, 27 },
		{ 311, 363, 27 },
		{ 311, 364, 27 },
		{ 312, 313, 27 },
		{ 312, 314, 27 },
		{ 312, 315, 27 },
		{ 312, 316, 27 },
		{ 312, 317, 27 },
		{ 312, 318, 27 },
		{ 312, 319, 27 },
		{ 312, 320, 27 },
		{ 312, 321, 27 },
		{ 312, 322, 27 },
		{ 312, 323, 27 },
		{ 312, 324, 27 },
		{ 312, 325, 27 },
		{ 312, 326, 27 },
		{ 312, 327, 27 },
		{ 312, 328, 27 },
		{ 312, 329, 27 },
		{ 312, 330, 27 },
		{ 312, 331, 27 },
		{ 312, 332, 27 },
		{ 312, 333, 27 },
		{ 312, 334, 27 },
		{ 312, 335, 27 },
		{ 312, 336, 27 },
		{ 312, 337, 27 },
		{ 312, 338, 27 },
		{ 312, 339, 27 },
		{ 312, 340, 27 },
		{ 312, 341, 27 },
		{ 312, 342, 27 },
		{ 312, 343, 27 },
		{ 312, 344, 27 },
		{ 312, 345, 27 },
		{ 312, 346, 27 },
		{ 312, 347, 27 },
		{ 312, 348, 27 },
		{ 312, 349, 27 },
		{ 312, 350, 27 },
		{ 312, 351, 27 },
		{ 312, 352, 27 },
		{ 312, 353, 27 },
		{ 312, 354, 27 },
		{ 312, 355, 27 },
		{ 312, 356, 27 },
		{ 312, 357, 27 },
		{ 312, 358, 27 },
		{ 312, 359, 27 },
		{ 312, 360, 27 },
		{ 312, 361, 27 },
		{ 312, 362, 27 },
		{ 312, 363, 27 },
		{ 312, 364, 27 },
		{ 312, 365, 27 },
		{ 312, 366, 27 },
		{ 313, 314, 27 },
		{ 313, 315, 27 },
		{ 313, 316, 27 },
		{ 313, 317, 27 },
		{ 313, 318, 27 },
		{ 313, 319, 27 },
		{ 313, 320, 27 },
		{ 313, 321, 27 },
		{ 313, 322, 27 },
		{ 313, 323, 27 },
		{ 313, 324, 27 },
		{ 313, 325, 27 },
		{ 313, 326, 27 },
		{ 313, 327, 27 },
		{ 313, 328, 27 },
		{ 313, 329, 27 },
		{ 313, 330, 27 },
		{ 313, 331, 27 },
		{ 313, 332, 27 },
		{ 313, 333, 27 },
		{ 313, 334, 27 },
		{ 313, 335, 27 },
		{ 313, 336, 27 },
		{ 313, 337, 27 },
		{ 313, 338, 27 },
		{ 313, 339, 27 },
		{ 313, 340, 27 },
		{ 313, 341, 27 },
		{ 313, 342, 27 },
		{ 313, 343, 27 },
		{ 313, 344, 27 },
		{ 313, 345, 27 },
		{ 313, 346, 27 },
		{ 313, 347, 27 },
		{ 313, 348, 27 },
		{ 313, 349, 27 },
		{ 313, 350, 27 },
		{ 313, 351, 27 },
		{ 313, 352, 27 },
		{ 313, 353, 27 },
		{ 313, 354, 27 },
		{ 313, 355, 27 },
		{ 313, 356, 27 },
		{ 313, 357, 27 },
		{ 313, 358, 27 },
		{ 313, 359, 27 },
		{ 313, 360, 27 },
		{ 313, 361, 27 },
		{ 313, 362, 27 },
		{ 313, 363, 27 },
		{ 313, 364, 27 },
		{ 313, 365, 27 },
		{ 313, 366, 27 },
		{ 313, 367, 27 },
		{ 313, 368, 27 },
		{ 314, 315, 27 },
		{ 314, 316, 27 },
		{ 314, 317, 27 },
		{ 314, 318, 27 },
		{ 314, 319, 27 },
		{ 314, 320, 27 },
		{ 314, 321, 27 },
		{ 314, 322, 27 },
		{ 314, 323, 27 },
		{ 314, 324, 27 },
		{ 314, 325, 27 },
		{ 314, 326, 27 },
		{ 314, 327, 27 },
		{ 314, 328, 27 },
		{ 314, 329, 27 },
		{ 314, 330, 27 },
		{ 314, 331, 27 },
		{ 314, 332, 27 },
		{ 314, 333, 27 },
		{ 314, 334, 27 },
		{ 314, 335, 27 },
		{ 314, 336, 27 },
		{ 314, 337, 27 },
		{ 314, 338, 27 },
		{ 314, 339, 27 },
		{ 314, 340, 27 },
		{ 314, 341, 27 },
		{ 314, 342, 27 },
		{ 314, 343, 27 },
		{ 314, 344, 27 },
		{ 314, 345, 27 },
		{ 314, 346, 27 },
		{ 314, 347, 27 },
		{ 314, 348, 27 },
		{ 314, 349, 27 },
		{ 314, 350, 27 },
		{ 314, 351, 27 },
		{ 314, 352, 27 },
		{ 314, 353, 27 },
		{ 314, 354, 27 },
		{ 314, 355, 27 },
		{ 314, 356, 27 },
		{ 314, 357, 27 },
		{ 314, 358, 27 },
		{ 314, 359, 27 },
		{ 314, 360, 27 },
		{ 314, 361, 27 },
		{ 314, 362, 27 },
		{ 314, 363, 27 },
		{ 314, 364, 27 },
		{ 314, 365, 27 },
		{ 314, 366, 27 },
		{ 314, 367, 27 },
		{ 314, 368, 27 },
		{ 314, 369, 27 },
		{ 314, 370, 27 },
		{ 315, 316, 27 },
		{ 315, 317, 27 },
		{ 315, 318, 27 },
		{ 315, 319, 27 },
		{ 315, 320, 27 },
		{ 315, 321, 27 },
		{ 315, 322, 27 },
		{ 315, 323, 27 },
		{ 315, 324, 27 },
		{ 315, 325, 27 },
		{ 315, 326, 27 },
		{ 315, 327, 27 },
		{ 315, 328, 27 },
		{ 315, 329, 27 },
		{ 315, 330, 27 },
		{ 315, 331, 27 },
		{ 315, 332, 27 },
		{ 315, 333, 27 },
		{ 315, 334, 27 },
		{ 315, 335, 27 },
		{ 315, 336, 27 },
		{ 315, 337, 27 },
		{ 315, 338, 27 },
		{ 315, 339, 27 },
		{ 315, 340, 27 },
		{ 315, 341, 27 },
		{ 315, 342, 27 },
		{ 315, 343, 27 },
		{ 315, 344, 27 },
		{ 315, 345, 27 },
		{ 315, 346, 27 },
		{ 315, 347, 27 },
		{ 315, 348, 27 },
		{ 315, 349, 27 },
		{ 315, 350, 27 },
		{ 315, 351, 27 },
		{ 315, 352, 27 },
		{ 315, 353, 27 },
		{ 315, 354, 27 },
		{ 315, 355, 27 },
		{ 315, 356, 27 },
		{ 315, 357, 27 },
		{ 315, 358, 27 },
		{ 315, 359, 27 },
		{ 315, 360, 27 },
		{ 315, 361, 27 },
		{ 315, 362, 27 },
		{ 315, 363, 27 },
		{ 315, 364, 27 },
		{ 315, 365, 27 },
		{ 315, 366, 27 },
		{ 315, 367, 27 },
		{ 315, 368, 27 },
		{ 315, 369, 27 },
		{ 315, 370, 27 },
		{ 315, 371, 27 },
		{ 315, 372, 27 },
		{ 316, 317, 27 },
		{ 316, 318, 27 },
		{ 316, 319, 27 },
		{ 316, 320, 27 },
		{ 316, 321, 27 },
		{ 316, 322, 27 },
		{ 316, 323, 27 },
		{ 316, 324, 27 },
		{ 316, 325, 27 },
		{ 316, 326, 27 },
		{ 316, 327, 27 },
		{ 316, 328, 27 },
		{ 316, 329, 27 },
		{ 316, 330, 27 },
		{ 316, 331, 27 },
		{ 316, 332, 27 },
		{ 316, 333, 27 },
		{ 316, 334, 27 },
		{ 316, 335, 27 },
		{ 316, 336, 27 },
		{ 316, 337, 27 },
		{ 316, 338, 27 },
		{ 316, 339, 27 },
		{ 316, 340, 27 },
		{ 316, 341, 27 },
		{ 316, 342, 27 },
		{ 316, 343, 27 },
		{ 316, 344, 27 },
		{ 316, 345, 27 },
		{ 316, 346, 27 },
		{ 316, 347, 27 },
		{ 316, 348, 27 },
		{ 316, 349, 27 },
		{ 316, 350, 27 },
		{ 316, 351, 27 },
		{ 316, 352, 27 },
		{ 316, 353, 27 },
		{ 316, 354, 27 },
		{ 316, 355, 27 },
		{ 316, 356, 27 },
		{ 316, 357, 27 },
		{ 316, 358, 27 },
		{ 316, 359, 27 },
		{ 316, 360, 27 },
		{ 316, 361, 27 },
		{ 316, 362, 27 },
		{ 316, 363, 27 },
		{ 316, 364, 27 },
		{ 316, 365, 27 },
		{ 316, 366, 27 },
		{ 316, 367, 27 },
		{ 316, 368, 27 },
		{ 316, 369, 27 },
		{ 316, 370, 27 },
		{ 316, 371, 27 },
		{ 316, 372, 27 },
		{ 316, 373, 27 },
		{ 316, 374, 27 },
		{ 317, 318, 27 },
		{ 317, 319, 27 },
		{ 317, 320, 27 },
		{ 317, 321, 27 },
		{ 317, 322, 27 },
		{ 317, 323, 27 },
		{ 317, 324, 27 },
		{ 317, 325, 27 },
		{ 317, 326, 27 },
		{ 317, 327, 27 },
		{ 317, 328, 27 },
		{ 317, 329, 27 },
		{ 317, 330, 27 },
		{ 317, 331, 27 },
		{ 317, 332, 27 },
		{ 317, 333, 27 },
		{ 317, 334, 27 },
		{ 317, 335, 27 },
		{ 317, 336, 27 },
		{ 317, 337, 27 },
		{ 317, 338, 27 },
		{ 317, 339, 27 },
		{ 317, 340, 27 },
		{ 317, 341, 27 },
		{ 317, 342, 27 },
		{ 317, 343, 27 },
		{ 317, 344, 27 },
		{ 317, 345, 27 },
		{ 317, 346, 27 },
		{ 317, 347, 27 },
		{ 317, 348, 27 },
		{ 317, 349, 27 },
		{ 317, 350, 27 },
		{ 317, 351, 27 },
		{ 317, 352, 27 },
		{ 317, 353, 27 },
		{ 317, 354, 27 },
		{ 317, 355, 27 },
		{ 317, 356, 27 },
		{ 317, 357, 27 },
		{ 317, 358, 27 },
		{ 317, 359, 27 },
		{ 317, 360, 27 },
		{ 317, 361, 27 },
		{ 317, 362, 27 },
		{ 317, 363, 27 },
		{ 317, 364, 27 },
		{ 317, 365, 27 },
		{ 317, 366, 27 },
		{ 317, 367, 27 },
		{ 317, 368, 27 },
		{ 317, 369, 27 },
		{ 317, 370, 27 },
		{ 317, 371, 27 },
		{ 317, 372, 27 },
		{ 317, 373, 27 },
		{ 317, 374, 27 },
		{ 317, 375, 27 },
		{ 317, 376, 27 },
		{ 318, 319, 27 },
		{ 318, 320, 27 },
		{ 318, 321, 27 },
		{ 318, 322, 27 },
		{ 318, 323, 27 },
		{ 318, 324, 27 },
		{ 318, 325, 27 },
		{ 318, 326, 27 },
		{ 318, 327, 27 },
		{ 318, 328, 27 },
		{ 318, 329, 27 },
		{ 318, 330, 27 },
		{ 318, 331, 27 },
		{ 318, 332, 27 },
		{ 318, 333, 27 },
		{ 318, 334, 27 },
		{ 318, 335, 27 },
		{ 318, 336, 27 },
		{ 318, 337, 27 },
		{ 318, 338, 27 },
		{ 318, 339, 27 },
		{ 318, 340, 27 },
		{ 318, 341, 27 },
		{ 318, 342, 27 },
		{ 318, 343, 27 },
		{ 318, 344, 27 },
		{ 318, 345, 27 },
		{ 318, 346, 27 },
		{ 318, 347, 27 },
		{ 318, 348, 27 },
		{ 318, 349, 27 },
		{ 318, 350, 27 },
		{ 318, 351, 27 },
		{ 318, 352, 27 },
		{ 318, 353, 27 },
		{ 318, 354, 27 },
		{ 318, 355, 27 },
		{ 318, 356, 27 },
		{ 318, 357, 27 },
		{ 318, 358, 27 },
		{ 318, 359, 27 },
		{ 318, 360, 27 },
		{ 318, 361, 27 },
		{ 318, 362, 27 },
		{ 318, 363, 27 },
		{ 318, 364, 27 },
		{ 318, 365, 27 },
		{ 318, 366, 27 },
		{ 318, 367, 27 },
		{ 318, 368, 27 },
		{ 318, 369, 27 },
		{ 318, 370, 27 },
		{ 318, 371, 27 },
		{ 318, 372, 27 },
		{ 318, 373, 27 },
		{ 318, 374, 27 },
		{ 318, 375, 27 },
		{ 318, 376, 27 },
		{ 318, 377, 27 },
		{ 319, 320, 27 },
		{ 319, 321, 27 },
		{ 319, 322, 27 },
		{ 319, 323, 27 },
		{ 319, 324, 27 },
		{ 319, 325, 27 },
		{ 319, 326, 27 },
		{ 319, 327, 27 },
		{ 319, 328, 27 },
		{ 319, 329, 27 },
		{ 319, 330, 27 },
		{ 319, 331, 27 },
		{ 319, 332, 27 },
		{ 319, 333, 27 },
		{ 319, 334, 27 },
		{ 319, 335, 27 },
		{ 319, 336, 27 },
		{ 319, 337, 27 },
		{ 319, 338, 27 },
		{ 319, 339, 27 },
		{ 319, 340, 27 },
		{ 319, 341, 27 },
		{ 319, 342, 27 },
		{ 319, 343, 27 },
		{ 319, 344, 27 },
		{ 319, 345, 27 },
		{ 319, 346, 27 },
		{ 319, 347, 27 },
		{ 319, 348, 27 },
		{ 319, 349, 27 },
		{ 319, 350, 27 },
		{ 319, 351, 27 },
		{ 319, 352, 27 },
		{ 319, 353, 27 },
		{ 319, 354, 27 },
		{ 319, 355, 27 },
		{ 319, 356, 27 },
		{ 319, 357, 27 },
		{ 319, 358, 27 },
		{ 319, 359, 27 },
		{ 319, 360, 27 },
		{ 319, 361, 27 },
		{ 319, 362, 27 },
		{ 319, 363, 27 },
		{ 319, 364, 27 },
		{ 319, 365, 27 },
		{ 319, 366, 27 },
		{ 319, 367, 27 },
		{ 319, 368, 27 },
		{ 319, 369, 27 },
		{ 319, 370, 27 },
		{ 319, 371, 27 },
		{ 319, 372, 27 },
		{ 319, 373, 27 },
		{ 319, 374, 27 },
		{ 319, 375, 27 },
		{ 319, 376, 27 },
		{ 319, 377, 27 },
		{ 319, 378, 27 },
		{ 319, 379, 27 },
		{ 320, 321, 27 },
		{ 320, 322, 27 },
		{ 320, 323, 27 },
		{ 320, 324, 27 },
		{ 320, 325, 27 },
		{ 320, 326, 27 },
		{ 320, 327, 27 },
		{ 320, 328, 27 },
		{ 320, 329, 27 },
		{ 320, 330, 27 },
		{ 320, 331, 27 },
		{ 320, 332, 27 },
		{ 320, 333, 27 },
		{ 320, 334, 27 },
		{ 320, 335, 27 },
		{ 320, 336, 27 },
		{ 320, 337, 27 },
		{ 320, 338, 27 },
		{ 320, 339, 27 },
		{ 320, 340, 27 },
		{ 320, 341, 27 },
		{ 320, 342, 27 },
		{ 320, 343, 27 },
		{ 320, 344, 27 },
		{ 320, 345, 27 },
		{ 320, 346, 27 },
		{ 320, 347, 27 },
		{ 320, 348, 27 },
		{ 320, 349, 27 },
		{ 320, 350, 27 },
		{ 320, 351, 27 },
		{ 320, 352, 27 },
		{ 320, 353, 27 },
		{ 320, 354, 27 },
		{ 320, 355, 27 },
		{ 320, 356, 27 },
		{ 320, 357, 27 },
		{ 320, 358, 27 },
		{ 320, 359, 27 },
		{ 320, 360, 27 },
		{ 320, 361, 27 },
		{ 320, 362, 27 },
		{ 320, 363, 27 },
		{ 320, 364, 27 },
		{ 320, 365, 27 },
		{ 320, 366, 27 },
		{ 320, 367, 27 },
		{ 320, 368, 27 },
		{ 320, 369, 27 },
		{ 320, 370, 27 },
		{ 320, 371, 27 },
		{ 320, 372, 27 },
		{ 320, 373, 27 },
		{ 320, 374, 27 },
		{ 320, 375, 27 },
		{ 320, 376, 27 },
		{ 320, 377, 27 },
		{ 320, 378, 27 },
		{ 320, 379, 27 },
		{ 320, 380, 27 },
		{ 321, 322, 27 },
		{ 321, 323, 27 },
		{ 321, 324, 27 },
		{ 321, 325, 27 },
		{ 321, 326, 27 },
		{ 321, 327, 27 },
		{ 321, 328, 27 },
		{ 321, 329, 27 },
		{ 321, 330, 27 },
		{ 321, 331, 27 },
		{ 321, 332, 27 },
		{ 321, 333, 27 },
		{ 321, 334, 27 },
		{ 321, 335, 27 },
		{ 321, 336, 27 },
		{ 321, 337, 27 },
		{ 321, 338, 27 },
		{ 321, 339, 27 },
		{ 321, 340, 27 },
		{ 321, 341, 27 },
		{ 321, 342, 27 },
		{ 321, 343, 27 },
		{ 321, 344, 27 },
		{ 321, 345, 27 },
		{ 321, 346, 27 },
		{ 321, 347, 27 },
		{ 321, 348, 27 },
		{ 321, 349, 27 },
		{ 321, 350, 27 },
		{ 321, 351, 27 },
		{ 321, 352, 27 },
		{ 321, 353, 27 },
		{ 321, 354, 27 },
		{ 321, 355, 27 },
		{ 321, 356, 27 },
		{ 321, 357, 27 },
		{ 321, 358, 27 },
		{ 321, 359, 27 },
		{ 321, 360, 27 },
		{ 321, 361, 27 },
		{ 321, 362, 27 },
		{ 321, 363, 27 },
		{ 321, 364, 27 },
		{ 321, 365, 27 },
		{ 321, 366, 27 },
		{ 321, 367, 27 },
		{ 321, 368, 27 },
		{ 321, 369, 27 },
		{ 321, 370, 27 },
		{ 321, 371, 27 },
		{ 321, 372, 27 },
		{ 321, 373, 27 },
		{ 321, 374, 27 },
		{ 321, 375, 27 },
		{ 321, 376, 27 },
		{ 321, 377, 27 },
		{ 321, 378, 27 },
		{ 321, 379, 27 },
		{ 321, 380, 27 },
		{ 321, 381, 27 },
		{ 322, 323, 27 },
		{ 322, 324, 27 },
		{ 322, 325, 27 },
		{ 322, 326, 27 },
		{ 322, 327, 27 },
		{ 322, 328, 27 },
		{ 322, 329, 27 },
		{ 322, 330, 27 },
		{ 322, 331, 27 },
		{ 322, 332, 27 },
		{ 322, 333, 27 },
		{ 322, 334, 27 },
		{ 322, 335, 27 },
		{ 322, 336, 27 },
		{ 322, 337, 27 },
		{ 322, 338, 27 },
		{ 322, 339, 27 },
		{ 322, 340, 27 },
		{ 322, 341, 27 },
		{ 322, 342, 27 },
		{ 322, 343, 27 },
		{ 322, 344, 27 },
		{ 322, 345, 27 },
		{ 322, 346, 27 },
		{ 322, 347, 27 },
		{ 322, 348, 27 },
		{ 322, 349, 27 },
		{ 322, 350, 27 },
		{ 322, 351, 27 },
		{ 322, 352, 27 },
		{ 322, 353, 27 },
		{ 322, 354, 27 },
		{ 322, 355, 27 },
		{ 322, 356, 27 },
		{ 322, 357, 27 },
		{ 322, 358, 27 },
		{ 322, 359, 27 },
		{ 322, 360, 27 },
		{ 322, 361, 27 },
		{ 322, 362, 27 },
		{ 322, 363, 27 },
		{ 322, 364, 27 },
		{ 322, 365, 27 },
		{ 322, 366, 27 },
		{ 322, 367, 27 },
		{ 322, 368, 27 },
		{ 322, 369, 27 },
		{ 322, 370, 27 },
		{ 322, 371, 27 },
		{ 322, 372, 27 },
		{ 322, 373, 27 },
		{ 322, 374, 27 },
		{ 322, 375, 27 },
		{ 322, 376, 27 },
		{ 322, 377, 27 },
		{ 322, 378, 27 },
		{ 322, 379, 27 },
		{ 322, 380, 27 },
		{ 322, 381, 27 },
		{ 323, 324, 27 },
		{ 323, 325, 27 },
		{ 323, 326, 27 },
		{ 323, 327, 27 },
		{ 323, 328, 27 },
		{ 323, 329, 27 },
		{ 323, 330, 27 },
		{ 323, 331, 27 },
		{ 323, 332, 27 },
		{ 323, 333, 27 },
		{ 323, 334, 27 },
		{ 323, 335, 27 },
		{ 323, 336, 27 },
		{ 323, 337, 27 },
		{ 323, 338, 27 },
		{ 323, 339, 27 },
		{ 323, 340, 27 },
		{ 323, 341, 27 },
		{ 323, 342, 27 },
		{ 323, 343, 27 },
		{ 323, 344, 27 },
		{ 323, 345, 27 },
		{ 323, 346, 27 },
		{ 323, 347, 27 },
		{ 323, 348, 27 },
		{ 323, 349, 27 },
		{ 323, 350, 27 },
		{ 323, 351, 27 },
		{ 323, 352, 27 },
		{ 323, 353, 27 },
		{ 323, 354, 27 },
		{ 323, 355, 27 },
		{ 323, 356, 27 },
		{ 323, 357, 27 },
		{ 323, 358, 27 },
		{ 323, 359, 27 },
		{ 323, 360, 27 },
		{ 323, 361, 27 },
		{ 323, 362, 27 },
		{ 323, 363, 27 },
		{ 323, 364, 27 },
		{ 323, 365, 27 },
		{ 323, 366, 27 },
		{ 323, 367, 27 },
		{ 323, 368, 27 },
		{ 323, 369, 27 },
		{ 323, 370, 27 },
		{ 323, 371, 27 },
		{ 323, 372, 27 },
		{ 323, 373, 27 },
		{ 323, 374, 27 },
		{ 323, 375, 27 },
		{ 323, 376, 27 },
		{ 323, 377, 27 },
		{ 323, 378, 27 },
		{ 323, 379, 27 },
		{ 323, 380, 27 },
		{ 323, 381, 27 },
		{ 323, 382, 27 },
		{ 324, 325, 27 },
		{ 324, 326, 27 },
		{ 324, 327, 27 },
		{ 324, 328, 27 },
		{ 324, 329, 27 },
		{ 324, 330, 27 },
		{ 324, 331, 27 },
		{ 324, 332, 27 },
		{ 324, 333, 27 },
		{ 324, 334, 27 },
		{ 324, 335, 27 },
		{ 324, 336, 27 },
		{ 324, 337, 27 },
		{ 324, 338, 27 },
		{ 324, 339, 27 },
		{ 324, 340, 27 },
		{ 324, 341, 27 },
		{ 324, 342, 27 },
		{ 324, 343, 27 },
		{ 324, 344, 27 },
		{ 324, 345, 27 },
		{ 324, 346, 27 },
		{ 324, 347, 27 },
		{ 324, 348, 27 },
		{ 324, 349, 27 },
		{ 324, 350, 27 },
		{ 324, 351, 27 },
		{ 324, 352, 27 },
		{ 324, 353, 27 },
		{ 324, 354, 27 },
		{ 324, 355, 27 },
		{ 324, 356, 27 },
		{ 324, 357, 27 },
		{ 324, 358, 27 },
		{ 324, 359, 27 },
		{ 324, 360, 27 },
		{ 324, 361, 27 },
		{ 324, 362, 27 },
		{ 324, 363, 27 },
		{ 324, 364, 27 },
		{ 324, 365, 27 },
		{ 324, 366, 27 },
		{ 324, 367, 27 },
		{ 324, 368, 27 },
		{ 324, 369, 27 },
		{ 324, 370, 27 },
		{ 324, 371, 27 },
		{ 324, 372, 27 },
		{ 324, 373, 27 },
		{ 324, 374, 27 },
		{ 324, 375, 27 },
		{ 324, 376, 27 },
		{ 324, 377, 27 },
		{ 324, 378, 27 },
		{ 324, 379, 27 },
		{ 324, 380, 27 },
		{ 324, 381, 27 },
		{ 324, 382, 27 },
		{ 325, 326, 27 },
		{ 325, 327, 27 },
		{ 325, 328, 27 },
		{ 325, 329, 27 },
		{ 325, 330, 27 },
		{ 325, 331, 27 },
		{ 325, 332, 27 },
		{ 325, 333, 27 },
		{ 325, 334, 27 },
		{ 325, 335, 27 },
		{ 325, 336, 27 },
		{ 325, 337, 27 },
		{ 325, 338, 27 },
		{ 325, 339, 27 },
		{ 325, 340, 27 },
		{ 325, 341, 27 },
		{ 325, 342, 27 },
		{ 325, 343, 27 },
		{ 325, 344, 27 },
		{ 325, 345, 27 },
		{ 325, 346, 27 },
		{ 325, 347, 27 },
		{ 325, 348, 27 },
		{ 325, 349, 27 },
		{ 325, 350, 27 },
		{ 325, 351, 27 },
		{ 325, 352, 27 },
		{ 325, 353, 27 },
		{ 325, 354, 27 },
		{ 325, 355, 27 },
		{ 325, 356, 27 },
		{ 325, 357, 27 },
		{ 325, 358, 27 },
		{ 325, 359, 27 },
		{ 325, 360, 27 },
		{ 325, 361, 27 },
		{ 325, 362, 27 },
		{ 325, 363, 27 },
		{ 325, 364, 27 },
		{ 325, 365, 27 },
		{ 325, 366, 27 },
		{ 325, 367, 27 },
		{ 325, 368, 27 },
		{ 325, 369, 27 },
		{ 325, 370, 27 },
		{ 325, 371, 27 },
		{ 325, 372, 27 },
		{ 325, 373, 27 },
		{ 325, 374, 27 },
		{ 325, 375, 27 },
		{ 325, 376, 27 },
		{ 325, 377, 27 },
		{ 325, 378, 27 },
		{ 325, 379, 27 },
		{ 325, 380, 27 },
		{ 325, 381, 27 },
		{ 325, 382, 27 },
		{ 326, 327, 27 },
		{ 326, 328, 27 },
		{ 326, 329, 27 },
		{ 326, 330, 27 },
		{ 326, 331, 27 },
		{ 326, 332, 27 },
		{ 326, 333, 27 },
		{ 326, 334, 27 },
		{ 326, 335, 27 },
		{ 326, 336, 27 },
		{ 326, 337, 27 },
		{ 326, 338, 27 },
		{ 326, 339, 27 },
		{ 326, 340, 27 },
		{ 326, 341, 27 },
		{ 326, 342, 27 },
		{ 326, 343, 27 },
		{ 326, 344, 27 },
		{ 326, 345, 27 },
		{ 326, 346, 27 },
		{ 326, 347, 27 },
		{ 326, 348, 27 },
		{ 326, 349, 27 },
		{ 326, 350, 27 },
		{ 326, 351, 27 },
		{ 326, 352, 27 },
		{ 326, 353, 27 },
		{ 326, 354, 27 },
		{ 326, 355, 27 },
		{ 326, 356, 27 },
		{ 326, 357, 27 },
		{ 326, 358, 27 },
		{ 326, 359, 27 },
		{ 326, 360, 27 },
		{ 326, 361, 27 },
		{ 326, 362, 27 },
		{ 326, 363, 27 },
		{ 326, 364, 27 },
		{ 326, 365, 27 },
		{ 326, 366, 27 },
		{ 326, 367, 27 },
		{ 326, 368, 27 },
		{ 326, 369, 27 },
		{ 326, 370, 27 },
		{ 326, 371, 27 },
		{ 326, 372, 27 },
		{ 326, 373, 27 },
		{ 326, 374, 27 },
		{ 326, 375, 27 },
		{ 326, 376, 27 },
		{ 326, 377, 27 },
		{ 326, 378, 27 },
		{ 326, 379, 27 },
		{ 326, 380, 27 },
		{ 326, 381, 27 },
		{ 326, 382, 27 },
		{ 327, 328, 27 },
		{ 327, 329, 27 },
		{ 327, 330, 27 },
		{ 327, 331, 27 },
		{ 327, 332, 27 },
		{ 327, 333, 27 },
		{ 327, 334, 27 },
		{ 327, 335, 27 },
		{ 327, 336, 27 },
		{ 327, 337, 27 },
		{ 327, 338, 27 },
		{ 327, 339, 27 },
		{ 327, 340, 27 },
		{ 327, 341, 27 },
		{ 327, 342, 27 },
		{ 327, 343, 27 },
		{ 327, 344, 27 },
		{ 327, 345, 27 },
		{ 327, 346, 27 },
		{ 327, 347, 27 },
		{ 327, 348, 27 },
		{ 327, 349, 27 },
		{ 327, 350, 27 },
		{ 327, 351, 27 },
		{ 327, 352, 27 },
		{ 327, 353, 27 },
		{ 327, 354, 27 },
		{ 327, 355, 27 },
		{ 327, 356, 27 },
		{ 327, 357, 27 },
		{ 327, 358, 27 },
		{ 327, 359, 27 },
		{ 327, 360, 27 },
		{ 327, 361, 27 },
		{ 327, 362, 27 },
		{ 327, 363, 27 },
		{ 327, 364, 27 },
		{ 327, 365, 27 },
		{ 327, 366, 27 },
		{ 327, 367, 27 },
		{ 327, 368, 27 },
		{ 327, 369, 27 },
		{ 327, 370, 27 },
		{ 327, 371, 27 },
		{ 327, 372, 27 },
		{ 327, 373, 27 },
		{ 327, 374, 27 },
		{ 327, 375, 27 },
		{ 327, 376, 27 },
		{ 327, 377, 27 },
		{ 327, 378, 27 },
		{ 327, 379, 27 },
		{ 327, 380, 27 },
		{ 327, 381, 27 },
		{ 327, 382, 27 },
		{ 327, 383, 27 },
		{ 328, 329, 27 },
		{ 328, 330, 27 },
		{ 328, 331, 27 },
		{ 328, 332, 27 },
		{ 328, 333, 27 },
		{ 328, 334, 27 },
		{ 328, 335, 27 },
		{ 328, 336, 27 },
		{ 328, 337, 27 },
		{ 328, 338, 27 },
		{ 328, 339, 27 },
		{ 328, 340, 27 },
		{ 328, 341, 27 },
		{ 328, 342, 27 },
		{ 328, 343, 27 },
		{ 328, 344, 27 },
		{ 328, 345, 27 },
		{ 328, 346, 27 },
		{ 328, 347, 27 },
		{ 328, 348, 27 },
		{ 328, 349, 27 },
		{ 328, 350, 27 },
		{ 328, 351, 27 },
		{ 328, 352, 27 },
		{ 328, 353, 27 },
		{ 328, 354, 27 },
		{ 328, 355, 27 },
		{ 328, 356, 27 },
		{ 328, 357, 27 },
		{ 328, 358, 27 },
		{ 328, 359, 27 },
		{ 328, 360, 27 },
		{ 328, 361, 27 },
		{ 328, 362, 27 },
		{ 328, 363, 27 },
		{ 328, 364, 27 },
		{ 328, 365, 27 },
		{ 328, 366, 27 },
		{ 328, 367, 27 },
		{ 328, 368, 27 },
		{ 328, 369, 27 },
		{ 328, 370, 27 },
		{ 328, 371, 27 },
		{ 328, 372, 27 },
		{ 328, 373, 27 },
		{ 328, 374, 27 },
		{ 328, 375, 27 },
		{ 328, 376, 27 },
		{ 328, 377, 27 },
		{ 328, 378, 27 },
		{ 328, 379, 27 },
		{ 328, 380, 27 },
		{ 328, 381, 27 },
		{ 328, 382, 27 },
		{ 328, 383, 27 },
		{ 329, 330, 27 },
		{ 329, 331, 27 },
		{ 329, 332, 27 },
		{ 329, 333, 27 },
		{ 329, 334, 27 },
		{ 329, 335, 27 },
		{ 329, 336, 27 },
		{ 329, 337, 27 },
		{ 329, 338, 27 },
		{ 329, 339, 27 },
		{ 329, 340, 27 },
		{ 329, 341, 27 },
		{ 329, 342, 27 },
		{ 329, 343, 27 },
		{ 329, 344, 27 },
		{ 329, 345, 27 },
		{ 329, 346, 27 },
		{ 329, 347, 27 },
		{ 329, 348, 27 },
		{ 329, 349, 27 },
		{ 329, 350, 27 },
		{ 329, 351, 27 },
		{ 329, 352, 27 },
		{ 329, 353, 27 },
		{ 329, 354, 27 },
		{ 329, 355, 27 },
		{ 329, 356, 27 },
		{ 329, 357, 27 },
		{ 329, 358, 27 },
		{ 329, 359, 27 },
		{ 329, 360, 27 },
		{ 329, 361, 27 },
		{ 329, 362, 27 },
		{ 329, 363, 27 },
		{ 329, 364, 27 },
		{ 329, 365, 27 },
		{ 329, 366, 27 },
		{ 329, 367, 27 },
		{ 329, 368, 27 },
		{ 329, 369, 27 },
		{ 329, 370, 27 },
		{ 329, 371, 27 },
		{ 329, 372, 27 },
		{ 329, 373, 27 },
		{ 329, 374, 27 },
		{ 329, 375, 27 },
		{ 329, 376, 27 },
		{ 329, 377, 27 },
		{ 329, 378, 27 },
		{ 329, 379, 27 },
		{ 329, 380, 27 },
		{ 329, 381, 27 },
		{ 329, 382, 27 },
		{ 329, 383, 27 },
		{ 330, 331, 27 },
		{ 330, 332, 27 },
		{ 330, 333, 27 },
		{ 330, 334, 27 },
		{ 330, 335, 27 },
		{ 330, 336, 27 },
		{ 330, 337, 27 },
		{ 330, 338, 27 },
		{ 330, 339, 27 },
		{ 330, 340, 27 },
		{ 330, 341, 27 },
		{ 330, 342, 27 },
		{ 330, 343, 27 },
		{ 330, 344, 27 },
		{ 330, 345, 27 },
		{ 330, 346, 27 },
		{ 330, 347, 27 },
		{ 330, 348, 27 },
		{ 330, 349, 27 },
		{ 330, 350, 27 },
		{ 330, 351, 27 },
		{ 330, 352, 27 },
		{ 330, 353, 27 },
		{ 330, 354, 27 },
		{ 330, 355, 27 },
		{ 330, 356, 27 },
		{ 330, 357, 27 },
		{ 330, 358, 27 },
		{ 330, 359, 27 },
		{ 330, 360, 27 },
		{ 330, 361, 27 },
		{ 330, 362, 27 },
		{ 330, 363, 27 },
		{ 330, 364, 27 },
		{ 330, 365, 27 },
		{ 330, 366, 27 },
		{ 330, 367, 27 },
		{ 330, 368, 27 },
		{ 330, 369, 27 },
		{ 330, 370, 27 },
		{ 330, 371, 27 },
		{ 330, 372, 27 },
		{ 330, 373, 27 },
		{ 330, 374, 27 },
		{ 330, 375, 27 },
		{ 330, 376, 27 },
		{ 330, 377, 27 },
		{ 330, 378, 27 },
		{ 330, 379, 27 },
		{ 330, 380, 27 },
		{ 330, 381, 27 },
		{ 330, 382, 27 },
		{ 330, 383, 27 },
		{ 331, 332, 27 },
		{ 331, 333, 27 },
		{ 331, 334, 27 },
		{ 331, 335, 27 },
		{ 331, 336, 27 },
		{ 331, 337, 27 },
		{ 331, 338, 27 },
		{ 331, 339, 27 },
		{ 331, 340, 27 },
		{ 331, 341, 27 },
		{ 331, 342, 27 },
		{ 331, 343, 27 },
		{ 331, 344, 27 },
		{ 331, 345, 27 },
		{ 331, 346, 27 },
		{ 331, 347, 27 },
		{ 331, 348, 27 },
		{ 331, 349, 27 },
		{ 331, 350, 27 },
		{ 331, 351, 27 },
		{ 331, 352, 27 },
		{ 331, 353, 27 },
		{ 331, 354, 27 },
		{ 331, 355, 27 },
		{ 331, 356, 27 },
		{ 331, 357, 27 },
		{ 331, 358, 27 },
		{ 331, 359, 27 },
		{ 331, 360, 27 },
		{ 331, 361, 27 },
		{ 331, 362, 27 },
		{ 331, 363, 27 },
		{ 331, 364, 27 },
		{ 331, 365, 27 },
		{ 331, 366, 27 },
		{ 331, 367, 27 },
		{ 331, 368, 27 },
		{ 331, 369, 27 },
		{ 331, 370, 27 },
		{ 331, 371, 27 },
		{ 331, 372, 27 },
		{ 331, 373, 27 },
		{ 331, 374, 27 },
		{ 331, 375, 27 },
		{ 331, 376, 27 },
		{ 331, 377, 27 },
		{ 331, 378, 27 },
		{ 331, 379, 27 },
		{ 331, 380, 27 },
		{ 331, 381, 27 },
		{ 331, 382, 27 },
		{ 331, 383, 27 },
		{ 332, 333, 27 },
		{ 332, 334, 27 },
		{ 332, 335, 27 },
		{ 332, 336, 27 },
		{ 332, 337, 27 },
		{ 332, 338, 27 },
		{ 332, 339, 27 },
		{ 332, 340, 27 },
		{ 332, 341, 27 },
		{ 332, 342, 27 },
		{ 332, 343, 27 },
		{ 332, 344, 27 },
		{ 332, 345, 27 },
		{ 332, 346, 27 },
		{ 332, 347, 27 },
		{ 332, 348, 27 },
		{ 332, 349, 27 },
		{ 332, 350, 27 },
		{ 332, 351, 27 },
		{ 332, 352, 27 },
		{ 332, 353, 27 },
		{ 332, 354, 27 },
		{ 332, 355, 27 },
		{ 332, 356, 27 },
		{ 332, 357, 27 },
		{ 332, 358, 27 },
		{ 332, 359, 27 },
		{ 332, 360, 27 },
		{ 332, 361, 27 },
		{ 332, 362, 27 },
		{ 332, 363, 27 },
		{ 332, 364, 27 },
		{ 332, 365, 27 },
		{ 332, 366, 27 },
		{ 332, 367, 27 },
		{ 332, 368, 27 },
		{ 332, 369, 27 },
		{ 332, 370, 27 },
		{ 332, 371, 27 },
		{ 332, 372, 27 },
		{ 332, 373, 27 },
		{ 332, 374, 27 },
		{ 332, 375, 27 },
		{ 332, 376, 27 },
		{ 332, 377, 27 },
		{ 332, 378, 27 },
		{ 332, 379, 27 },
		{ 332, 380, 27 },
		{ 332, 381, 27 },
		{ 332, 382, 27 },
		{ 332, 383, 27 },
		{ 333, 334, 27 },
		{ 333, 335, 27 },
		{ 333, 336, 27 },
		{ 333, 337, 27 },
		{ 333, 338, 27 },
		{ 333, 339, 27 },
		{ 333, 340, 27 },
		{ 333, 341, 27 },
		{ 333, 342, 27 },
		{ 333, 343, 27 },
		{ 333, 344, 27 },
		{ 333, 345, 27 },
		{ 333, 346, 27 },
		{ 333, 347, 27 },
		{ 333, 348, 27 },
		{ 333, 349, 27 },
		{ 333, 350, 27 },
		{ 333, 351, 27 },
		{ 333, 352, 27 },
		{ 333, 353, 27 },
		{ 333, 354, 27 },
		{ 333, 355, 27 },
		{ 333, 356, 27 },
		{ 333, 357, 27 },
		{ 333, 358, 27 },
		{ 333, 359, 27 },
		{ 333, 360, 27 },
		{ 333, 361, 27 },
		{ 333, 362, 27 },
		{ 333, 363, 27 },
		{ 333, 364, 27 },
		{ 333, 365, 27 },
		{ 333, 366, 27 },
		{ 333, 367, 27 },
		{ 333, 368, 27 },
		{ 333, 369, 27 },
		{ 333, 370, 27 },
		{ 333, 371, 27 },
		{ 333, 372, 27 },
		{ 333, 373, 27 },
		{ 333, 374, 27 },
		{ 333, 375, 27 },
		{ 333, 376, 27 },
		{ 333, 377, 27 },
		{ 333, 378, 27 },
		{ 333, 379, 27 },
		{ 333, 380, 27 },
		{ 333, 381, 27 },
		{ 333, 382, 27 },
		{ 333, 383, 27 },
		{ 334, 335, 27 },
		{ 334, 336, 27 },
		{ 334, 337, 27 },
		{ 334, 338, 27 },
		{ 334, 339, 27 },
		{ 334, 340, 27 },
		{ 334, 341, 27 },
		{ 334, 342, 27 },
		{ 334, 343, 27 },
		{ 334, 344, 27 },
		{ 334, 345, 27 },
		{ 334, 346, 27 },
		{ 334, 347, 27 },
		{ 334, 348, 27 },
		{ 334, 349, 27 },
		{ 334, 350, 27 },
		{ 334, 351, 27 },
		{ 334, 352, 27 },
		{ 334, 353, 27 },
		{ 334, 354, 27 },
		{ 334, 355, 27 },
		{ 334, 356, 27 },
		{ 334, 357, 27 },
		{ 334, 358, 27 },
		{ 334, 359, 27 },
		{ 334, 360, 27 },
		{ 334, 361, 27 },
		{ 334, 362, 27 },
		{ 334, 363, 27 },
		{ 334, 364, 27 },
		{ 334, 365, 27 },
		{ 334, 366, 27 },
		{ 334, 367, 27 },
		{ 334, 368, 27 },
		{ 334, 369, 27 },
		{ 334, 370, 27 },
		{ 334, 371, 27 },
		{ 334, 372, 27 },
		{ 334, 373, 27 },
		{ 334, 374, 27 },
		{ 334, 375, 27 },
		{ 334, 376, 27 },
		{ 334, 377, 27 },
		{ 334, 378, 27 },
		{ 334, 379, 27 },
		{ 334, 380, 27 },
		{ 334, 381, 27 },
		{ 334, 382, 27 },
		{ 334, 383, 27 },
		{ 335, 336, 27 },
		{ 335, 337, 27 },
		{ 335, 338, 27 },
		{ 335, 339, 27 },
		{ 335, 340, 27 },
		{ 335, 341, 27 },
		{ 335, 342, 27 },
		{ 335, 343, 27 },
		{ 335, 344, 27 },
		{ 335, 345, 27 },
		{ 335, 346, 27 },
		{ 335, 347, 27 },
		{ 335, 348, 27 },
		{ 335, 349, 27 },
		{ 335, 350, 27 },
		{ 335, 351, 27 },
		{ 335, 352, 27 },
		{ 335, 353, 27 },
		{ 335, 354, 27 },
		{ 335, 355, 27 },
		{ 335, 356, 27 },
		{ 335, 357, 27 },
		{ 335, 358, 27 },
		{ 335, 359, 27 },
		{ 335, 360, 27 },
		{ 335, 361, 27 },
		{ 335, 362, 27 },
		{ 335, 363, 27 },
		{ 335, 364, 27 },
		{ 335, 365, 27 },
		{ 335, 366, 27 },
		{ 335, 367, 27 },
		{ 335, 368, 27 },
		{ 335, 369, 27 },
		{ 335, 370, 27 },
		{ 335, 371, 27 },
		{ 335, 372, 27 },
		{ 335, 373, 27 },
		{ 335, 374, 27 },
		{ 335, 375, 27 },
		{ 335, 376, 27 },
		{ 335, 377, 27 },
		{ 335, 378, 27 },
		{ 335, 379, 27 },
		{ 335, 380, 27 },
		{ 335, 381, 27 },
		{ 335, 382, 27 },
		{ 335, 383, 27 },
		{ 336, 337, 27 },
		{ 336, 338, 27 },
		{ 336, 339, 27 },
		{ 336, 340, 27 },
		{ 336, 341, 27 },
		{ 336, 342, 27 },
		{ 336, 343, 27 },
		{ 336, 344, 27 },
		{ 336, 345, 27 },
		{ 336, 346, 27 },
		{ 336, 347, 27 },
		{ 336, 348, 27 },
		{ 336, 349, 27 },
		{ 336, 350, 27 },
		{ 336, 351, 27 },
		{ 336, 352, 27 },
		{ 336, 353, 27 },
		{ 336, 354, 27 },
		{ 336, 355, 27 },
		{ 336, 356, 27 },
		{ 336, 357, 27 },
		{ 336, 358, 27 },
		{ 336, 359, 27 },
		{ 336, 360, 27 },
		{ 336, 361, 27 },
		{ 336, 362, 27 },
		{ 336, 363, 27 },
		{ 336, 364, 27 },
		{ 336, 365, 27 },
		{ 336, 366, 27 },
		{ 336, 367, 27 },
		{ 336, 368, 27 },
		{ 336, 369, 27 },
		{ 336, 370, 27 },
		{ 336, 371, 27 },
		{ 336, 372, 27 },
		{ 336, 373, 27 },
		{ 336, 374, 27 },
		{ 336, 375, 27 },
		{ 336, 376, 27 },
		{ 336, 377, 27 },
		{ 336, 378, 27 },
		{ 336, 379, 27 },
		{ 336, 380, 27 },
		{ 336, 381, 27 },
		{ 336, 382, 27 },
		{ 336, 383, 27 },
		{ 337, 338, 27 },
		{ 337, 339, 27 },
		{ 337, 340, 27 },
		{ 337, 341, 27 },
		{ 337, 342, 27 },
		{ 337, 343, 27 },
		{ 337, 344, 27 },
		{ 337, 345, 27 },
		{ 337, 346, 27 },
		{ 337, 347, 27 },
		{ 337, 348, 27 },
		{ 337, 349, 27 },
		{ 337, 350, 27 },
		{ 337, 351, 27 },
		{ 337, 352, 27 },
		{ 337, 353, 27 },
		{ 337, 354, 27 },
		{ 337, 355, 27 },
		{ 337, 356, 27 },
		{ 337, 357, 27 },
		{ 337, 358, 27 },
		{ 337, 359, 27 },
		{ 337, 360, 27 },
		{ 337, 361, 27 },
		{ 337, 362, 27 },
		{ 337, 363, 27 },
		{ 337, 364, 27 },
		{ 337, 365, 27 },
		{ 337, 366, 27 },
		{ 337, 367, 27 },
		{ 337, 368, 27 },
		{ 337, 369, 27 },
		{ 337, 370, 27 },
		{ 337, 371, 27 },
		{ 337, 372, 27 },
		{ 337, 373, 27 },
		{ 337, 374, 27 },
		{ 337, 375, 27 },
		{ 337, 376, 27 },
		{ 337, 377, 27 },
		{ 337, 378, 27 },
		{ 337, 379, 27 },
		{ 337, 380, 27 },
		{ 337, 381, 27 },
		{ 337, 382, 27 },
		{ 337, 383, 27 },
		{ 338, 339, 27 },
		{ 338, 340, 27 },
		{ 338, 341, 27 },
		{ 338, 342, 27 },
		{ 338, 343, 27 },
		{ 338, 344, 27 },
		{ 338, 345, 27 },
		{ 338, 346, 27 },
		{ 338, 347, 27 },
		{ 338, 348, 27 },
		{ 338, 349, 27 },
		{ 338, 350, 27 },
		{ 338, 351, 27 },
		{ 338, 352, 27 },
		{ 338, 353, 27 },
		{ 338, 354, 27 },
		{ 338, 355, 27 },
		{ 338, 356, 27 },
		{ 338, 357, 27 },
		{ 338, 358, 27 },
		{ 338, 359, 27 },
		{ 338, 360, 27 },
		{ 338, 361, 27 },
		{ 338, 362, 27 },
		{ 338, 363, 27 },
		{ 338, 364, 27 },
		{ 338, 365, 27 },
		{ 338, 366, 27 },
		{ 338, 367, 27 },
		{ 338, 368, 27 },
		{ 338, 369, 27 },
		{ 338, 370, 27 },
		{ 338, 371, 27 },
		{ 338, 372, 27 },
		{ 338, 373, 27 },
		{ 338, 374, 27 },
		{ 338, 375, 27 },
		{ 338, 376, 27 },
		{ 338, 377, 27 },
		{ 338, 378, 27 },
		{ 338, 379, 27 },
		{ 338, 380, 27 },
		{ 338, 381, 27 },
		{ 338, 382, 27 },
		{ 338, 383, 27 },
		{ 339, 340, 27 },
		{ 339, 341, 27 },
		{ 339, 342, 27 },
		{ 339, 343, 27 },
		{ 339, 344, 27 },
		{ 339, 345, 27 },
		{ 339, 346, 27 },
		{ 339, 347, 27 },
		{ 339, 348, 27 },
		{ 339, 349, 27 },
		{ 339, 350, 27 },
		{ 339, 351, 27 },
		{ 339, 352, 27 },
		{ 339, 353, 27 },
		{ 339, 354, 27 },
		{ 339, 355, 27 },
		{ 339, 356, 27 },
		{ 339, 357, 27 },
		{ 339, 358, 27 },
		{ 339, 359, 27 },
		{ 339, 360, 27 },
		{ 339, 361, 27 },
		{ 339, 362, 27 },
		{ 339, 363, 27 },
		{ 339, 364, 27 },
		{ 339, 365, 27 },
		{ 339, 366, 27 },
		{ 339, 367, 27 },
		{ 339, 368, 27 },
		{ 339, 369, 27 },
		{ 339, 370, 27 },
		{ 339, 371, 27 },
		{ 339, 372, 27 },
		{ 339, 373, 27 },
		{ 339, 374, 27 },
		{ 339, 375, 27 },
		{ 339, 376, 27 },
		{ 339, 377, 27 },
		{ 339, 378, 27 },
		{ 339, 379, 27 },
		{ 339, 380, 27 },
		{ 339, 381, 27 },
		{ 339, 382, 27 },
		{ 339, 383, 27 },
		{ 340, 341, 27 },
		{ 340, 342, 27 },
		{ 340, 343, 27 },
		{ 340, 344, 27 },
		{ 340, 345, 27 },
		{ 340, 346, 27 },
		{ 340, 347, 27 },
		{ 340, 348, 27 },
		{ 340, 349, 27 },
		{ 340, 350, 27 },
		{ 340, 351, 27 },
		{ 340, 352, 27 },
		{ 340, 353, 27 },
		{ 340, 354, 27 },
		{ 340, 355, 27 },
		{ 340, 356, 27 },
		{ 340, 357, 27 },
		{ 340, 358, 27 },
		{ 340, 359, 27 },
		{ 340, 360, 27 },
		{ 340, 361, 27 },
		{ 340, 362, 27 },
		{ 340, 363, 27 },
		{ 340, 364, 27 },
		{ 340, 365, 27 },
		{ 340, 366, 27 },
		{ 340, 367, 27 },
		{ 340, 368, 27 },
		{ 340, 369, 27 },
		{ 340, 370, 27 },
		{ 340, 371, 27 },
		{ 340, 372, 27 },
		{ 340, 373, 27 },
		{ 340, 374, 27 },
		{ 340, 375, 27 },
		{ 340, 376, 27 },
		{ 340, 377, 27 },
		{ 340, 378, 27 },
		{ 340, 379, 27 },
		{ 340, 380, 27 },
		{ 340, 381, 27 },
		{ 340, 382, 27 },
		{ 340, 383, 27 },
		{ 341, 342, 27 },
		{ 341, 343, 27 },
		{ 341, 344, 27 },
		{ 341, 345, 27 },
		{ 341, 346, 27 },
		{ 341, 347, 27 },
		{ 341, 348, 27 },
		{ 341, 349, 27 },
		{ 341, 350, 27 },
		{ 341, 351, 27 },
		{ 341, 352, 27 },
		{ 341, 353, 27 },
		{ 341, 354, 27 },
		{ 341, 355, 27 },
		{ 341, 356, 27 },
		{ 341, 357, 27 },
		{ 341, 358, 27 },
		{ 341, 359, 27 },
		{ 341, 360, 27 },
		{ 341, 361, 27 },
		{ 341, 362, 27 },
		{ 341, 363, 27 },
		{ 341, 364, 27 },
		{ 341, 365, 27 },
		{ 341, 366, 27 },
		{ 341, 367, 27 },
		{ 341, 368, 27 },
		{ 341, 369, 27 },
		{ 341, 370, 27 },
		{ 341, 371, 27 },
		{ 341, 372, 27 },
		{ 341, 373, 27 },
		{ 341, 374, 27 },
		{ 341, 375, 27 },
		{ 341, 376, 27 },
		{ 341, 377, 27 },
		{ 341, 378, 27 },
		{ 341, 379, 27 },
		{ 341, 380, 27 },
		{ 341, 381, 27 },
		{ 341, 382, 27 },
		{ 341, 383, 27 },
		{ 342, 343, 27 },
		{ 342, 344, 27 },
		{ 342, 345, 27 },
		{ 342, 346, 27 },
		{ 342, 347, 27 },
		{ 342, 348, 27 },
		{ 342, 349, 27 },
		{ 342, 350, 27 },
		{ 342, 351, 27 },
		{ 342, 352, 27 },
		{ 342, 353, 27 },
		{ 342, 354, 27 },
		{ 342, 355, 27 },
		{ 342, 356, 27 },
		{ 342, 357, 27 },
		{ 342, 358, 27 },
		{ 342, 359, 27 },
		{ 342, 360, 27 },
		{ 342, 361, 27 },
		{ 342, 362, 27 },
		{ 342, 363, 27 },
		{ 342, 364, 27 },
		{ 342, 365, 27 },
		{ 342, 366, 27 },
		{ 342, 367, 27 },
		{ 342, 368, 27 },
		{ 342, 369, 27 },
		{ 342, 370, 27 },
		{ 342, 371, 27 },
		{ 342, 372, 27 },
		{ 342, 373, 27 },
		{ 342, 374, 27 },
		{ 342, 375, 27 },
		{ 342, 376, 27 },
		{ 342, 377, 27 },
		{ 342, 378, 27 },
		{ 342, 379, 27 },
		{ 342, 380, 27 },
		{ 342, 381, 27 },
		{ 342, 382, 27 },
		{ 342, 383, 27 },
		{ 343, 344, 27 },
		{ 343, 345, 27 },
		{ 343, 346, 27 },
		{ 343, 347, 27 },
		{ 343, 348, 27 },
		{ 343, 349, 27 },
		{ 343, 350, 27 },
		{ 343, 351, 27 },
		{ 343, 352, 27 },
		{ 343, 353, 27 },
		{ 343, 354, 27 },
		{ 343, 355, 27 },
		{ 343, 356, 27 },
		{ 343, 357, 27 },
		{ 343, 358, 27 },
		{ 343, 359, 27 },
		{ 343, 360, 27 },
		{ 343, 361, 27 },
		{ 343, 362, 27 },
		{ 343, 363, 27 },
		{ 343, 364, 27 },
		{ 343, 365, 27 },
		{ 343, 366, 27 },
		{ 343, 367, 27 },
		{ 343, 368, 27 },
		{ 343, 369, 27 },
		{ 343, 370, 27 },
		{ 343, 371, 27 },
		{ 343, 372, 27 },
		{ 343, 373, 27 },
		{ 343, 374, 27 },
		{ 343, 375, 27 },
		{ 343, 376, 27 },
		{ 343, 377, 27 },
		{ 343, 378, 27 },
		{ 343, 379, 27 },
		{ 343, 380, 27 },
		{ 343, 381, 27 },
		{ 343, 382, 27 },
		{ 343, 383, 27 },
		{ 344, 345, 27 },
		{ 344, 346, 27 },
		{ 344, 347, 27 },
		{ 344, 348, 27 },
		{ 344, 349, 27 },
		{ 344, 350, 27 },
		{ 344, 351, 27 },
		{ 344, 352, 27 },
		{ 344, 353, 27 },
		{ 344, 354, 27 },
		{ 344, 355, 27 },
		{ 344, 356, 27 },
		{ 344, 357, 27 },
		{ 344, 358, 27 },
		{ 344, 359, 27 },
		{ 344, 360, 27 },
		{ 344, 361, 27 },
		{ 344, 362, 27 },
		{ 344, 363, 27 },
		{ 344, 364, 27 },
		{ 344, 365, 27 },
		{ 344, 366, 27 },
		{ 344, 367, 27 },
		{ 344, 368, 27 },
		{ 344, 369, 27 },
		{ 344, 370, 27 },
		{ 344, 371, 27 },
		{ 344, 372, 27 },
		{ 344, 373, 27 },
		{ 344, 374, 27 },
		{ 344, 375, 27 },
		{ 344, 376, 27 },
		{ 344, 377, 27 },
		{ 344, 378, 27 },
		{ 344, 379, 27 },
		{ 344, 380, 27 },
		{ 344, 381, 27 },
		{ 344, 382, 27 },
		{ 344, 383, 27 },
		{ 345, 346, 27 },
		{ 345, 347, 27 },
		{ 345, 348, 27 },
		{ 345, 349, 27 },
		{ 345, 350, 27 },
		{ 345, 351, 27 },
		{ 345, 352, 27 },
		{ 345, 353, 27 },
		{ 345, 354, 27 },
		{ 345, 355, 27 },
		{ 345, 356, 27 },
		{ 345, 357, 27 },
		{ 345, 358, 27 },
		{ 345, 359, 27 },
		{ 345, 360, 27 },
		{ 345, 361, 27 },
		{ 345, 362, 27 },
		{ 345, 363, 27 },
		{ 345, 364, 27 },
		{ 345, 365, 27 },
		{ 345, 366, 27 },
		{ 345, 367, 27 },
		{ 345, 368, 27 },
		{ 345, 369, 27 },
		{ 345, 370, 27 },
		{ 345, 371, 27 },
		{ 345, 372, 27 },
		{ 345, 373, 27 },
		{ 345, 374, 27 },
		{ 345, 375, 27 },
		{ 345, 376, 27 },
		{ 345, 377, 27 },
		{ 345, 378, 27 },
		{ 345, 379, 27 },
		{ 345, 380, 27 },
		{ 345, 381, 27 },
		{ 345, 382, 27 },
		{ 345, 383, 27 },
		{ 346, 347, 27 },
		{ 346, 348, 27 },
		{ 346, 349, 27 },
		{ 346, 350, 27 },
		{ 346, 351, 27 },
		{ 346, 352, 27 },
		{ 346, 353, 27 },
		{ 346, 354, 27 },
		{ 346, 355, 27 },
		{ 346, 356, 27 },
		{ 346, 357, 27 },
		{ 346, 358, 27 },
		{ 346, 359, 27 },
		{ 346, 360, 27 },
		{ 346, 361, 27 },
		{ 346, 362, 27 },
		{ 346, 363, 27 },
		{ 346, 364, 27 },
		{ 346, 365, 27 },
		{ 346, 366, 27 },
		{ 346, 367, 27 },
		{ 346, 368, 27 },
		{ 346, 369, 27 },
		{ 346, 370, 27 },
		{ 346, 371, 27 },
		{ 346, 372, 27 },
		{ 346, 373, 27 },
		{ 346, 374, 27 },
		{ 346, 375, 27 },
		{ 346, 376, 27 },
		{ 346, 377, 27 },
		{ 346, 378, 27 },
		{ 346, 379, 27 },
		{ 346, 380, 27 },
		{ 346, 381, 27 },
		{ 346, 382, 27 },
		{ 346, 383, 27 },
		{ 347, 348, 27 },
		{ 347, 349, 27 },
		{ 347, 350, 27 },
		{ 347, 351, 27 },
		{ 347, 352, 27 },
		{ 347, 353, 27 },
		{ 347, 354, 27 },
		{ 347, 355, 27 },
		{ 347, 356, 27 },
		{ 347, 357, 27 },
		{ 347, 358, 27 },
		{ 347, 359, 27 },
		{ 347, 360, 27 },
		{ 347, 361, 27 },
		{ 347, 362, 27 },
		{ 347, 363, 27 },
		{ 347, 364, 27 },
		{ 347, 365, 27 },
		{ 347, 366, 27 },
		{ 347, 367, 27 },
		{ 347, 368, 27 },
		{ 347, 369, 27 },
		{ 347, 370, 27 },
		{ 347, 371, 27 },
		{ 347, 372, 27 },
		{ 347, 373, 27 },
		{ 347, 374, 27 },
		{ 347, 375, 27 },
		{ 347, 376, 27 },
		{ 347, 377, 27 },
		{ 347, 378, 27 },
		{ 347, 379, 27 },
		{ 347, 380, 27 },
		{ 347, 381, 27 },
		{ 347, 382, 27 },
		{ 347, 383, 27 },
		{ 348, 349, 27 },
		{ 348, 350, 27 },
		{ 348, 351, 27 },
		{ 348, 352, 27 },
		{ 348, 353, 27 },
		{ 348, 354, 27 },
		{ 348, 355, 27 },
		{ 348, 356, 27 },
		{ 348, 357, 27 },
		{ 348, 358, 27 },
		{ 348, 359, 27 },
		{ 348, 360, 27 },
		{ 348, 361, 27 },
		{ 348, 362, 27 },
		{ 348, 363, 27 },
		{ 348, 364, 27 },
		{ 348, 365, 27 },
		{ 348, 366, 27 },
		{ 348, 367, 27 },
		{ 348, 368, 27 },
		{ 348, 369, 27 },
		{ 348, 370, 27 },
		{ 348, 371, 27 },
		{ 348, 372, 27 },
		{ 348, 373, 27 },
		{ 348, 374, 27 },
		{ 348, 375, 27 },
		{ 348, 376, 27 },
		{ 348, 377, 27 },
		{ 348, 378, 27 },
		{ 348, 379, 27 },
		{ 348, 380, 27 },
		{ 348, 381, 27 },
		{ 348, 382, 27 },
		{ 348, 383, 27 },
		{ 349, 350, 27 },
		{ 349, 351, 27 },
		{ 349, 352, 27 },
		{ 349, 353, 27 },
		{ 349, 354, 27 },
		{ 349, 355, 27 },
		{ 349, 356, 27 },
		{ 349, 357, 27 },
		{ 349, 358, 27 },
		{ 349, 359, 27 },
		{ 349, 360, 27 },
		{ 349, 361, 27 },
		{ 349, 362, 27 },
		{ 349, 363, 27 },
		{ 349, 364, 27 },
		{ 349, 365, 27 },
		{ 349, 366, 27 },
		{ 349, 367, 27 },
		{ 349, 368, 27 },
		{ 349, 369, 27 },
		{ 349, 370, 27 },
		{ 349, 371, 27 },
		{ 349, 372, 27 },
		{ 349, 373, 27 },
		{ 349, 374, 27 },
		{ 349, 375, 27 },
		{ 349, 376, 27 },
		{ 349, 377, 27 },
		{ 349, 378, 27 },
		{ 349, 379, 27 },
		{ 349, 380, 27 },
		{ 349, 381, 27 },
		{ 349, 382, 27 },
		{ 349, 383, 27 },
		{ 350, 351, 27 },
		{ 350, 352, 27 },
		{ 350, 353, 27 },
		{ 350, 354, 27 },
		{ 350, 355, 27 },
		{ 350, 356, 27 },
		{ 350, 357, 27 },
		{ 350, 358, 27 },
		{ 350, 359, 27 },
		{ 350, 360, 27 },
		{ 350, 361, 27 },
		{ 350, 362, 27 },
		{ 350, 363, 27 },
		{ 350, 364, 27 },
		{ 350, 365, 27 },
		{ 350, 366, 27 },
		{ 350, 367, 27 },
		{ 350, 368, 27 },
		{ 350, 369, 27 },
		{ 350, 370, 27 },
		{ 350, 371, 27 },
		{ 350, 372, 27 },
		{ 350, 373, 27 },
		{ 350, 374, 27 },
		{ 350, 375, 27 },
		{ 350, 376, 27 },
		{ 350, 377, 27 },
		{ 350, 378, 27 },
		{ 350, 379, 27 },
		{ 350, 380, 27 },
		{ 350, 381, 27 },
		{ 350, 382, 27 },
		{ 350, 383, 27 },
		{ 351, 352, 27 },
		{ 351, 353, 27 },
		{ 351, 354, 27 },
		{ 351, 355, 27 },
		{ 351, 356, 27 },
		{ 351, 357, 27 },
		{ 351, 358, 27 },
		{ 351, 359, 27 },
		{ 351, 360, 27 },
		{ 351, 361, 27 },
		{ 351, 362, 27 },
		{ 351, 363, 27 },
		{ 351, 364, 27 },
		{ 351, 365, 27 },
		{ 351, 366, 27 },
		{ 351, 367, 27 },
		{ 351, 368, 27 },
		{ 351, 369, 27 },
		{ 351, 370, 27 },
		{ 351, 371, 27 },
		{ 351, 372, 27 },
		{ 351, 373, 27 },
		{ 351, 374, 27 },
		{ 351, 375, 27 },
		{ 351, 376, 27 },
		{ 351, 377, 27 },
		{ 351, 378, 27 },
		{ 351, 379, 27 },
		{ 351, 380, 27 },
		{ 351, 381, 27 },
		{ 351, 382, 27 },
		{ 351, 383, 27 },
		{ 352, 353, 27 },
		{ 352, 354, 27 },
		{ 352, 355, 27 },
		{ 352, 356, 27 },
		{ 352, 357, 27 },
		{ 352, 358, 27 },
		{ 352, 359, 27 },
		{ 352, 360, 27 },
		{ 352, 361, 27 },
		{ 352, 362, 27 },
		{ 352, 363, 27 },
		{ 352, 364, 27 },
		{ 352, 365, 27 },
		{ 352, 366, 27 },
		{ 352, 367, 27 },
		{ 352, 368, 27 },
		{ 352, 369, 27 },
		{ 352, 370, 27 },
		{ 352, 371, 27 },
		{ 352, 372, 27 },
		{ 352, 373, 27 },
		{ 352, 374, 27 },
		{ 352, 375, 27 },
		{ 352, 376, 27 },
		{ 352, 377, 27 },
		{ 352, 378, 27 },
		{ 352, 379, 27 },
		{ 352, 380, 27 },
		{ 352, 381, 27 },
		{ 352, 382, 27 },
		{ 352, 383, 27 },
		{ 353, 354, 27 },
		{ 353, 355, 27 },
		{ 353, 356, 27 },
		{ 353, 357, 27 },
		{ 353, 358, 27 },
		{ 353, 359, 27 },
		{ 353, 360, 27 },
		{ 353, 361, 27 },
		{ 353, 362, 27 },
		{ 353, 363, 27 },
		{ 353, 364, 27 },
		{ 353, 365, 27 },
		{ 353, 366, 27 },
		{ 353, 367, 27 },
		{ 353, 368, 27 },
		{ 353, 369, 27 },
		{ 353, 370, 27 },
		{ 353, 371, 27 },
		{ 353, 372, 27 },
		{ 353, 373, 27 },
		{ 353, 374, 27 },
		{ 353, 375, 27 },
		{ 353, 376, 27 },
		{ 353, 377, 27 },
		{ 353, 378, 27 },
		{ 353, 379, 27 },
		{ 353, 380, 27 },
		{ 353, 381, 27 },
		{ 353, 382, 27 },
		{ 353, 383, 27 },
		{ 354, 355, 27 },
		{ 354, 356, 27 },
		{ 354, 357, 27 },
		{ 354, 358, 27 },
		{ 354, 359, 27 },
		{ 354, 360, 27 },
		{ 354, 361, 27 },
		{ 354, 362, 27 },
		{ 354, 363, 27 },
		{ 354, 364, 27 },
		{ 354, 365, 27 },
		{ 354, 366, 27 },
		{ 354, 367, 27 },
		{ 354, 368, 27 },
		{ 354, 369, 27 },
		{ 354, 370, 27 },
		{ 354, 371, 27 },
		{ 354, 372, 27 },
		{ 354, 373, 27 },
		{ 354, 374, 27 },
		{ 354, 375, 27 },
		{ 354, 376, 27 },
		{ 354, 377, 27 },
		{ 354, 378, 27 },
		{ 354, 379, 27 },
		{ 354, 380, 27 },
		{ 354, 381, 27 },
		{ 354, 382, 27 },
		{ 354, 383, 27 },
		{ 355, 356, 27 },
		{ 355, 357, 27 },
		{ 355, 358, 27 },
		{ 355, 359, 27 },
		{ 355, 360, 27 },
		{ 355, 361, 27 },
		{ 355, 362, 27 },
		{ 355, 363, 27 },
		{ 355, 364, 27 },
		{ 355, 365, 27 },
		{ 355, 366, 27 },
		{ 355, 367, 27 },
		{ 355, 368, 27 },
		{ 355, 369, 27 },
		{ 355, 370, 27 },
		{ 355, 371, 27 },
		{ 355, 372, 27 },
		{ 355, 373, 27 },
		{ 355, 374, 27 },
		{ 355, 375, 27 },
		{ 355, 376, 27 },
		{ 355, 377, 27 },
		{ 355, 378, 27 },
		{ 355, 379, 27 },
		{ 355, 380, 27 },
		{ 355, 381, 27 },
		{ 355, 382, 27 },
		{ 355, 383, 27 },
		{ 356, 357, 27 },
		{ 356, 358, 27 },
		{ 356, 359, 27 },
		{ 356, 360, 27 },
		{ 356, 361, 27 },
		{ 356, 362, 27 },
		{ 356, 363, 27 },
		{ 356, 364, 27 },
		{ 356, 365, 27 },
		{ 356, 366, 27 },
		{ 356, 367, 27 },
		{ 356, 368, 27 },
		{ 356, 369, 27 },
		{ 356, 370, 27 },
		{ 356, 371, 27 },
		{ 356, 372, 27 },
		{ 356, 373, 27 },
		{ 356, 374, 27 },
		{ 356, 375, 27 },
		{ 356, 376, 27 },
		{ 356, 377, 27 },
		{ 356, 378, 27 },
		{ 356, 379, 27 },
		{ 356, 380, 27 },
		{ 356, 381, 27 },
		{ 356, 382, 27 },
		{ 356, 383, 27 },
		{ 357, 358, 27 },
		{ 357, 359, 27 },
		{ 357, 360, 27 },
		{ 357, 361, 27 },
		{ 357, 362, 27 },
		{ 357, 363, 27 },
		{ 357, 364, 27 },
		{ 357, 365, 27 },
		{ 357, 366, 27 },
		{ 357, 367, 27 },
		{ 357, 368, 27 },
		{ 357, 369, 27 },
		{ 357, 370, 27 },
		{ 357, 371, 27 },
		{ 357, 372, 27 },
		{ 357, 373, 27 },
		{ 357, 374, 27 },
		{ 357, 375, 27 },
		{ 357, 376, 27 },
		{ 357, 377, 27 },
		{ 357, 378, 27 },
		{ 357, 379, 27 },
		{ 357, 380, 27 },
		{ 357, 381, 27 },
		{ 357, 382, 27 },
		{ 357, 383, 27 },
		{ 358, 359, 27 },
		{ 358, 360, 27 },
		{ 358, 361, 27 },
		{ 358, 362, 27 },
		{ 358, 363, 27 },
		{ 358, 364, 27 },
		{ 358, 365, 27 },
		{ 358, 366, 27 },
		{ 358, 367, 27 },
		{ 358, 368, 27 },
		{ 358, 369, 27 },
		{ 358, 370, 27 },
		{ 358, 371, 27 },
		{ 358, 372, 27 },
		{ 358, 373, 27 },
		{ 358, 374, 27 },
		{ 358, 375, 27 },
		{ 358, 376, 27 },
		{ 358, 377, 27 },
		{ 358, 378, 27 },
		{ 358, 379, 27 },
		{ 358, 380, 27 },
		{ 358, 381, 27 },
		{ 358, 382, 27 },
		{ 358, 383, 27 },
		{ 359, 360, 27 },
		{ 359, 361, 27 },
		{ 359, 362, 27 },
		{ 359, 363, 27 },
		{ 359, 364, 27 },
		{ 359, 365, 27 },
		{ 359, 366, 27 },
		{ 359, 367, 27 },
		{ 359, 368, 27 },
		{ 359, 369, 27 },
		{ 359, 370, 27 },
		{ 359, 371, 27 },
		{ 359, 372, 27 },
		{ 359, 373, 27 },
		{ 359, 374, 27 },
		{ 359, 375, 27 },
		{ 359, 376, 27 },
		{ 359, 377, 27 },
		{ 359, 378, 27 },
		{ 359, 379, 27 },
		{ 359, 380, 27 },
		{ 359, 381, 27 },
		{ 359, 382, 27 },
		{ 359, 383, 27 },
		{ 360, 361, 27 },
		{ 360, 362, 27 },
		{ 360, 363, 27 },
		{ 360, 364, 27 },
		{ 360, 365, 27 },
		{ 360, 366, 27 },
		{ 360, 367, 27 },
		{ 360, 368, 27 },
		{ 360, 369, 27 },
		{ 360, 370, 27 },
		{ 360, 371, 27 },
		{ 360, 372, 27 },
		{ 360, 373, 27 },
		{ 360, 374, 27 },
		{ 360, 375, 27 },
		{ 360, 376, 27 },
		{ 360, 377, 27 },
		{ 360, 378, 27 },
		{ 360, 379, 27 },
		{ 360, 380, 27 },
		{ 360, 381, 27 },
		{ 360, 382, 27 },
		{ 360, 383, 27 },
		{ 361, 362, 27 },
		{ 361, 363, 27 },
		{ 361, 364, 27 },
		{ 361, 365, 27 },
		{ 361, 366, 27 },
		{ 361, 367, 27 },
		{ 361, 368, 27 },
		{ 361, 369, 27 },
		{ 361, 370, 27 },
		{ 361, 371, 27 },
		{ 361, 372, 27 },
		{ 361, 373, 27 },
		{ 361, 374, 27 },
		{ 361, 375, 27 },
		{ 361, 376, 27 },
		{ 361, 377, 27 },
		{ 361, 378, 27 },
		{ 361, 379, 27 },
		{ 361, 380, 27 },
		{ 361, 381, 27 },
		{ 361, 382, 27 },
		{ 361, 383, 27 },
		{ 362, 363, 27 },
		{ 362, 364, 27 },
		{ 362, 365, 27 },
		{ 362, 366, 27 },
		{ 362, 367, 27 },
		{ 362, 368, 27 },
		{ 362, 369, 27 },
		{ 362, 370, 27 },
		{ 362, 371, 27 },
		{ 362, 372, 27 },
		{ 362, 373, 27 },
		{ 362, 374, 27 },
		{ 362, 375, 27 },
		{ 362, 376, 27 },
		{ 362, 377, 27 },
		{ 362, 378, 27 },
		{ 362, 379, 27 },
		{ 362, 380, 27 },
		{ 362, 381, 27 },
		{ 362, 382, 27 },
		{ 362, 383, 27 },
		{ 363, 364, 27 },
		{ 363, 365, 27 },
		{ 363, 366, 27 },
		{ 363, 367, 27 },
		{ 363, 368, 27 },
		{ 363, 369, 27 },
		{ 363, 370, 27 },
		{ 363, 371, 27 },
		{ 363, 372, 27 },
		{ 363, 373, 27 },
		{ 363, 374, 27 },
		{ 363, 375, 27 },
		{ 363, 376, 27 },
		{ 363, 377, 27 },
		{ 363, 378, 27 },
		{ 363, 379, 27 },
		{ 363, 380, 27 },
		{ 363, 381, 27 },
		{ 363, 382, 27 },
		{ 363, 383, 27 },
		{ 364, 365, 27 },
		{ 364, 366, 27 },
		{ 364, 367, 27 },
		{ 364, 368, 27 },
		{ 364, 369, 27 },
		{ 364, 370, 27 },
		{ 364, 371, 27 },
		{ 364, 372, 27 },
		{ 364, 373, 27 },
		{ 364, 374, 27 },
		{ 364, 375, 27 },
		{ 364, 376, 27 },
		{ 364, 377, 27 },
		{ 364, 378, 27 },
		{ 364, 379, 27 },
		{ 364, 380, 27 },
		{ 364, 381, 27 },
		{ 364, 382, 27 },
		{ 364, 383, 27 },
		{ 365, 366, 27 },
		{ 365, 367, 27 },
		{ 365, 368, 27 },
		{ 365, 369, 27 },
		{ 365, 370, 27 },
		{ 365, 371, 27 },
		{ 365, 372, 27 },
		{ 365, 373, 27 },
		{ 365, 374, 27 },
		{ 365, 375, 27 },
		{ 365, 376, 27 },
		{ 365, 377, 27 },
		{ 365, 378, 27 },
		{ 365, 379, 27 },
		{ 365, 380, 27 },
		{ 365, 381, 27 },
		{ 365, 382, 27 },
		{ 365, 383, 27 },
		{ 366, 367, 27 },
		{ 366, 368, 27 },
		{ 366, 369, 27 },
		{ 366, 370, 27 },
		{ 366, 371, 27 },
		{ 366, 372, 27 },
		{ 366, 373, 27 },
		{ 366, 374, 27 },
		{ 366, 375, 27 },
		{ 366, 376, 27 },
		{ 366, 377, 27 },
		{ 366, 378, 27 },
		{ 366, 379, 27 },
		{ 366, 380, 27 },
		{ 366, 381, 27 },
		{ 366, 382, 27 },
		{ 366, 383, 27 },
		{ 367, 368, 27 },
		{ 367, 369, 27 },
		{ 367, 370, 27 },
		{ 367, 371, 27 },
		{ 367, 372, 27 },
		{ 367, 373, 27 },
		{ 367, 374, 27 },
		{ 367, 375, 27 },
		{ 367, 376, 27 },
		{ 367, 377, 27 },
		{ 367, 378, 27 },
		{ 367, 379, 27 },
		{ 367, 380, 27 },
		{ 367, 381, 27 },
		{ 367, 382, 27 },
		{ 367, 383, 27 },
		{ 368, 369, 27 },
		{ 368, 370, 27 },
		{ 368, 371, 27 },
		{ 368, 372, 27 },
		{ 368, 373, 27 },
		{ 368, 374, 27 },
		{ 368, 375, 27 },
		{ 368, 376, 27 },
		{ 368, 377, 27 },
		{ 368, 378, 27 },
		{ 368, 379, 27 },
		{ 368, 380, 27 },
		{ 368, 381, 27 },
		{ 368, 382, 27 },
		{ 368, 383, 27 },
		{ 369, 370, 27 },
		{ 369, 371, 27 },
		{ 369, 372, 27 },
		{ 369, 373, 27 },
		{ 369, 374, 27 },
		{ 369, 375, 27 },
		{ 369, 376, 27 },
		{ 369, 377, 27 },
		{ 369, 378, 27 },
		{ 369, 379, 27 },
		{ 369, 380, 27 },
		{ 369, 381, 27 },
		{ 369, 382, 27 },
		{ 369, 383, 27 },
		{ 370, 371, 27 },
		{ 370, 372, 27 },
		{ 370, 373, 27 },
		{ 370, 374, 27 },
		{ 370, 375, 27 },
		{ 370, 376, 27 },
		{ 370, 377, 27 },
		{ 370, 378, 27 },
		{ 370, 379, 27 },
		{ 370, 380, 27 },
		{ 370, 381, 27 },
		{ 370, 382, 27 },
		{ 370, 383, 27 },
		{ 371, 372, 27 },
		{ 371, 373, 27 },
		{ 371, 374, 27 },
		{ 371, 375, 27 },
		{ 371, 376, 27 },
		{ 371, 377, 27 },
		{ 371, 378, 27 },
		{ 371, 379, 27 },
		{ 371, 380, 27 },
		{ 371, 381, 27 },
		{ 371, 382, 27 },
		{ 371, 383, 27 },
		{ 372, 373, 27 },
		{ 372, 374, 27 },
		{ 372, 375, 27 },
		{ 372, 376, 27 },
		{ 372, 377, 27 },
		{ 372, 378, 27 },
		{ 372, 379, 27 },
		{ 372, 380, 27 },
		{ 372, 381, 27 },
		{ 372, 382, 27 },
		{ 372, 383, 27 },
		{ 373, 374, 27 },
		{ 373, 375, 27 },
		{ 373, 376, 27 },
		{ 373, 377, 27 },
		{ 373, 378, 27 },
		{ 373, 379, 27 },
		{ 373, 380, 27 },
		{ 373, 381, 27 },
		{ 373, 382, 27 },
		{ 373, 383, 27 },
		{ 374, 375, 27 },
		{ 374, 376, 27 },
		{ 374, 377, 27 },
		{ 374, 378, 27 },
		{ 374, 379, 27 },
		{ 374, 380, 27 },
		{ 374, 381, 27 },
		{ 374, 382, 27 },
		{ 374, 383, 27 },
		{ 375, 376, 27 },
		{ 375, 377, 27 },
		{ 375, 378, 27 },
		{ 375, 379, 27 },
		{ 375, 380, 27 },
		{ 375, 381, 27 },
		{ 375, 382, 27 },
		{ 375, 383, 27 },
		{ 376, 377, 27 },
		{ 376, 378, 27 },
		{ 376, 379, 27 },
		{ 376, 380, 27 },
		{ 376, 381, 27 },
		{ 376, 382, 27 },
		{ 376, 383, 27 },
		{ 377, 378, 27 },
		{ 377, 379, 27 },
		{ 377, 380, 27 },
		{ 377, 381, 27 },
		{ 377, 382, 27 },
		{ 378, 379, 27 },
		{ 378, 380, 27 },
		{ 378, 381, 27 },
		{ 378, 382, 27 },
		{ 379, 380, 27 },
		{ 379, 381, 27 },
		{ 379, 382, 27 },
		{ 166, 133, 28 },
		{ 166, 134, 28 },
		{ 167, 132, 28 },
		{ 167, 133, 28 },
		{ 167, 134, 28 },
		{ 167, 135, 28 },
		{ 167, 136, 28 },
		{ 168, 132, 28 },
		{ 168, 133, 28 },
		{ 168, 134, 28 },
		{ 168, 135, 28 },
		{ 168, 136, 28 },
		{ 168, 137, 28 },
		{ 169, 132, 28 },
		{ 169, 133, 28 },
		{ 169, 134, 28 },
		{ 169, 135, 28 },
		{ 169, 136, 28 },
		{ 169, 137, 28 },
		{ 169, 138, 28 },
		{ 169, 139, 28 },
		{ 170, 131, 28 },
		{ 170, 132, 28 },
		{ 170, 133, 28 },
		{ 170, 134, 28 },
		{ 170, 135, 28 },
		{ 170, 136, 28 },
		{ 170, 137, 28 },
		{ 170, 138, 28 },
		{ 170, 139, 28 },
		{ 170, 140, 28 },
		{ 171, 131, 28 },
		{ 171, 132, 28 },
		{ 171, 133, 28 },
		{ 171, 134, 28 },
		{ 171, 135, 28 },
		{ 171, 136, 28 },
		{ 171, 137, 28 },
		{ 171, 138, 28 },
		{ 171, 139, 28 },
		{ 171, 140, 28 },
		{ 171, 141, 28 },
		{ 171, 142, 28 },
		{ 172, 131, 28 },
		{ 172, 132, 28 },
		{ 172, 133, 28 },
		{ 172, 134, 28 },
		{ 172, 135, 28 },
		{ 172, 136, 28 },
		{ 172, 137, 28 },
		{ 172, 138, 28 },
		{ 172, 139, 28 },
		{ 172, 140, 28 },
		{ 172, 141, 28 },
		{ 172, 142, 28 },
		{ 172, 143, 28 },
		{ 173, 131, 28 },
		{ 173, 132, 28 },
		{ 173, 133, 28 },
		{ 173, 134, 28 },
		{ 173, 135, 28 },
		{ 173, 136, 28 },
		{ 173, 137, 28 },
		{ 173, 138, 28 },
		{ 173, 139, 28 },
		{ 173, 140, 28 },
		{ 173, 141, 28 },
		{ 173, 142, 28 },
		{ 173, 143, 28 },
		{ 173, 144, 28 },
		{ 174, 131, 28 },
		{ 174, 132, 28 },
		{ 174, 133, 28 },
		{ 174, 134, 28 },
		{ 174, 135, 28 },
		{ 174, 136, 28 },
		{ 174, 137, 28 },
		{ 174, 138, 28 },
		{ 174, 139, 28 },
		{ 174, 140, 28 },
		{ 174, 141, 28 },
		{ 174, 142, 28 },
		{ 174, 143, 28 },
		{ 174, 144, 28 },
		{ 174, 145, 28 },
		{ 174, 146, 28 },
		{ 175, 131, 28 },
		{ 175, 132, 28 },
		{ 175, 133, 28 },
		{ 175, 134, 28 },
		{ 175, 135, 28 },
		{ 175, 136, 28 },
		{ 175, 137, 28 },
		{ 175, 138, 28 },
		{ 175, 139, 28 },
		{ 175, 140, 28 },
		{ 175, 141, 28 },
		{ 175, 142, 28 },
		{ 175, 143, 28 },
		{ 175, 144, 28 },
		{ 175, 145, 28 },
		{ 175, 146, 28 },
		{ 175, 147, 28 },
		{ 176, 131, 28 },
		{ 176, 132, 28 },
		{ 176, 133, 28 },
		{ 176, 134, 28 },
		{ 176, 135, 28 },
		{ 176, 136, 28 },
		{ 176, 137, 28 },
		{ 176, 138, 28 },
		{ 176, 139, 28 },
		{ 176, 140, 28 },
		{ 176, 141, 28 },
		{ 176, 142, 28 },
		{ 176, 143, 28 },
		{ 176, 144, 28 },
		{ 176, 145, 28 },
		{ 176, 146, 28 },
		{ 176, 147, 28 },
		{ 176, 148, 28 },
		{ 177, 131, 28 },
		{ 177, 132, 28 },
		{ 177, 133, 28 },
		{ 177, 134, 28 },
		{ 177, 135, 28 },
		{ 177, 136, 28 },
		{ 177, 137, 28 },
		{ 177, 138, 28 },
		{ 177, 139, 28 },
		{ 177, 140, 28 },
		{ 177, 141, 28 },
		{ 177, 142, 28 },
		{ 177, 143, 28 },
		{ 177, 144, 28 },
		{ 177, 145, 28 },
		{ 177, 146, 28 },
		{ 177, 147, 28 },
		{ 177, 148, 28 },
		{ 177, 149, 28 },
		{ 177, 150, 28 },
		{ 178, 131, 28 },
		{ 178, 132, 28 },
		{ 178, 133, 28 },
		{ 178, 134, 28 },
		{ 178, 135, 28 },
		{ 178, 136, 28 },
		{ 178, 137, 28 },
		{ 178, 138, 28 },
		{ 178, 139, 28 },
		{ 178, 140, 28 },
		{ 178, 141, 28 },
		{ 178, 142, 28 },
		{ 178, 143, 28 },
		{ 178, 144, 28 },
		{ 178, 145, 28 },
		{ 178, 146, 28 },
		{ 178, 147, 28 },
		{ 178, 148, 28 },
		{ 178, 149, 28 },
		{ 178, 150, 28 },
		{ 178, 151, 28 },
		{ 179, 131, 28 },
		{ 179, 132, 28 },
		{ 179, 133, 28 },
		{ 179, 134, 28 },
		{ 179, 135, 28 },
		{ 179, 136, 28 },
		{ 179, 137, 28 },
		{ 179, 138, 28 },
		{ 179, 139, 28 },
		{ 179, 140, 28 },
		{ 179, 141, 28 },
		{ 179, 142, 28 },
		{ 179, 143, 28 },
		{ 179, 144, 28 },
		{ 179, 145, 28 },
		{ 179, 146, 28 },
		{ 179, 147, 28 },
		{ 179, 148, 28 },
		{ 179, 149, 28 },
		{ 179, 150, 28 },
		{ 179, 151, 28 },
		{ 179, 152, 28 },
		{ 180, 131, 28 },
		{ 180, 132, 28 },
		{ 180, 133, 28 },
		{ 180, 134, 28 },
		{ 180, 135, 28 },
		{ 180, 136, 28 },
		{ 180, 137, 28 },
		{ 180, 138, 28 },
		{ 180, 139, 28 },
		{ 180, 140, 28 },
		{ 180, 141, 28 },
		{ 180, 142, 28 },
		{ 180, 143, 28 },
		{ 180, 144, 28 },
		{ 180, 145, 28 },
		{ 180, 146, 28 },
		{ 180, 147, 28 },
		{ 180, 148, 28 },
		{ 180, 149, 28 },
		{ 180, 150, 28 },
		{ 180, 151, 28 },
		{ 180, 152, 28 },
		{ 180, 153, 28 },
		{ 180, 154, 28 },
		{ 181, 131, 28 },
		{ 181, 132, 28 },
		{ 181, 133, 28 },
		{ 181, 134, 28 },
		{ 181, 135, 28 },
		{ 181, 136, 28 },
		{ 181, 137, 28 },
		{ 181, 138, 28 },
		{ 181, 139, 28 },
		{ 181, 140, 28 },
		{ 181, 141, 28 },
		{ 181, 142, 28 },
		{ 181, 143, 28 },
		{ 181, 144, 28 },
		{ 181, 145, 28 },
		{ 181, 146, 28 },
		{ 181, 147, 28 },
		{ 181, 148, 28 },
		{ 181, 149, 28 },
		{ 181, 150, 28 },
		{ 181, 151, 28 },
		{ 181, 152, 28 },
		{ 181, 153, 28 },
		{ 181, 154, 28 },
		{ 181, 155, 28 },
		{ 182, 131, 28 },
		{ 182, 132, 28 },
		{ 182, 133, 28 },
		{ 182, 134, 28 },
		{ 182, 135, 28 },
		{ 182, 136, 28 },
		{ 182, 137, 28 },
		{ 182, 138, 28 },
		{ 182, 139, 28 },
		{ 182, 140, 28 },
		{ 182, 141, 28 },
		{ 182, 142, 28 },
		{ 182, 143, 28 },
		{ 182, 144, 28 },
		{ 182, 145, 28 },
		{ 182, 146, 28 },
		{ 182, 147, 28 },
		{ 182, 148, 28 },
		{ 182, 149, 28 },
		{ 182, 150, 28 },
		{ 182, 151, 28 },
		{ 182, 152, 28 },
		{ 182, 153, 28 },
		{ 182, 154, 28 },
		{ 182, 155, 28 },
		{ 182, 156, 28 },
		{ 183, 131, 28 },
		{ 183, 132, 28 },
		{ 183, 133, 28 },
		{ 183, 134, 28 },
		{ 183, 135, 28 },
		{ 183, 136, 28 },
		{ 183, 137, 28 },
		{ 183, 138, 28 },
		{ 183, 139, 28 },
		{ 183, 140, 28 },
		{ 183, 141, 28 },
		{ 183, 142, 28 },
		{ 183, 143, 28 },
		{ 183, 144, 28 },
		{ 183, 145, 28 },
		{ 183, 146, 28 },
		{ 183, 147, 28 },
		{ 183, 148, 28 },
		{ 183, 149, 28 },
		{ 183, 150, 28 },
		{ 183, 151, 28 },
		{ 183, 152, 28 },
		{ 183, 153, 28 },
		{ 183, 154, 28 },
		{ 183, 155, 28 },
		{ 183, 156, 28 },
		{ 183, 157, 28 },
		{ 183, 158, 28 },
		{ 184, 131, 28 },
		{ 184, 132, 28 },
		{ 184, 133, 28 },
		{ 184, 134, 28 },
		{ 184, 135, 28 },
		{ 184, 136, 28 },
		{ 184, 137, 28 },
		{ 184, 138, 28 },
		{ 184, 139, 28 },
		{ 184, 140, 28 },
		{ 184, 141, 28 },
		{ 184, 142, 28 },
		{ 184, 143, 28 },
		{ 184, 144, 28 },
		{ 184, 145, 28 },
		{ 184, 146, 28 },
		{ 184, 147, 28 },
		{ 184, 148, 28 },
		{ 184, 149, 28 },
		{ 184, 150, 28 },
		{ 184, 151, 28 },
		{ 184, 152, 28 },
		{ 184, 153, 28 },
		{ 184, 154, 28 },
		{ 184, 155, 28 },
		{ 184, 156, 28 },
		{ 184, 157, 28 },
		{ 184, 158, 28 },
		{ 184, 159, 28 },
		{ 185, 131, 28 },
		{ 185, 132, 28 },
		{ 185, 133, 28 },
		{ 185, 134, 28 },
		{ 185, 135, 28 },
		{ 185, 136, 28 },
		{ 185, 137, 28 },
		{ 185, 138, 28 },
		{ 185, 139, 28 },
		{ 185, 140, 28 },
		{ 185, 141, 28 },
		{ 185, 142, 28 },
		{ 185, 143, 28 },
		{ 185, 144, 28 },
		{ 185, 145, 28 },
		{ 185, 146, 28 },
		{ 185, 147, 28 },
		{ 185, 148, 28 },
		{ 185, 149, 28 },
		{ 185, 150, 28 },
		{ 185, 151, 28 },
		{ 185, 152, 28 },
		{ 185, 153, 28 },
		{ 185, 154, 28 },
		{ 185, 155, 28 },
		{ 185, 156, 28 },
		{ 185, 157, 28 },
		{ 185, 158, 28 },
		{ 185, 159, 28 },
		{ 185, 160, 28 },
		{ 186, 131, 28 },
		{ 186, 132, 28 },
		{ 186, 133, 28 },
		{ 186, 134, 28 },
		{ 186, 135, 28 },
		{ 186, 136, 28 },
		{ 186, 137, 28 },
		{ 186, 138, 28 },
		{ 186, 139, 28 },
		{ 186, 140, 28 },
		{ 186, 141, 28 },
		{ 186, 142, 28 },
		{ 186, 143, 28 },
		{ 186, 144, 28 },
		{ 186, 145, 28 },
		{ 186, 146, 28 },
		{ 186, 147, 28 },
		{ 186, 148, 28 },
		{ 186, 149, 28 },
		{ 186, 150, 28 },
		{ 186, 151, 28 },
		{ 186, 152, 28 },
		{ 186, 153, 28 },
		{ 186, 154, 28 },
		{ 186, 155, 28 },
		{ 186, 156, 28 },
		{ 186, 157, 28 },
		{ 186, 158, 28 },
		{ 186, 159, 28 },
		{ 186, 160, 28 },
		{ 186, 161, 28 },
		{ 186, 162, 28 },
		{ 187, 131, 28 },
		{ 187, 132, 28 },
		{ 187, 133, 28 },
		{ 187, 134, 28 },
		{ 187, 135, 28 },
		{ 187, 136, 28 },
		{ 187, 137, 28 },
		{ 187, 138, 28 },
		{ 187, 139, 28 },
		{ 187, 140, 28 },
		{ 187, 141, 28 },
		{ 187, 142, 28 },
		{ 187, 143, 28 },
		{ 187, 144, 28 },
		{ 187, 145, 28 },
		{ 187, 146, 28 },
		{ 187, 147, 28 },
		{ 187, 148, 28 },
		{ 187, 149, 28 },
		{ 187, 150, 28 },
		{ 187, 151, 28 },
		{ 187, 152, 28 },
		{ 187, 153, 28 },
		{ 187, 154, 28 },
		{ 187, 155, 28 },
		{ 187, 156, 28 },
		{ 187, 157, 28 },
		{ 187, 158, 28 },
		{ 187, 159, 28 },
		{ 187, 160, 28 },
		{ 187, 161, 28 },
		{ 187, 162, 28 },
		{ 187, 163, 28 },
		{ 188, 131, 28 },
		{ 188, 132, 28 },
		{ 188, 133, 28 },
		{ 188, 134, 28 },
		{ 188, 135, 28 },
		{ 188, 136, 28 },
		{ 188, 137, 28 },
		{ 188, 138, 28 },
		{ 188, 139, 28 },
		{ 188, 140, 28 },
		{ 188, 141, 28 },
		{ 188, 142, 28 },
		{ 188, 143, 28 },
		{ 188, 144, 28 },
		{ 188, 145, 28 },
		{ 188, 146, 28 },
		{ 188, 147, 28 },
		{ 188, 148, 28 },
		{ 188, 149, 28 },
		{ 188, 150, 28 },
		{ 188, 151, 28 },
		{ 188, 152, 28 },
		{ 188, 153, 28 },
		{ 188, 154, 28 },
		{ 188, 155, 28 },
		{ 188, 156, 28 },
		{ 188, 157, 28 },
		{ 188, 158, 28 },
		{ 188, 159, 28 },
		{ 188, 160, 28 },
		{ 188, 161, 28 },
		{ 188, 162, 28 },
		{ 188, 163, 28 },
		{ 188, 164, 28 },
		{ 189, 131, 28 },
		{ 189, 132, 28 },
		{ 189, 133, 28 },
		{ 189, 134, 28 },
		{ 189, 135, 28 },
		{ 189, 136, 28 },
		{ 189, 137, 28 },
		{ 189, 138, 28 },
		{ 189, 139, 28 },
		{ 189, 140, 28 },
		{ 189, 141, 28 },
		{ 189, 142, 28 },
		{ 189, 143, 28 },
		{ 189, 144, 28 },
		{ 189, 145, 28 },
		{ 189, 146, 28 },
		{ 189, 147, 28 },
		{ 189, 148, 28 },
		{ 189, 149, 28 },
		{ 189, 150, 28 },
		{ 189, 151, 28 },
		{ 189, 152, 28 },
		{ 189, 153, 28 },
		{ 189, 154, 28 },
		{ 189, 155, 28 },
		{ 189, 156, 28 },
		{ 189, 157, 28 },
		{ 189, 158, 28 },
		{ 189, 159, 28 },
		{ 189, 160, 28 },
		{ 189, 161, 28 },
		{ 189, 162, 28 },
		{ 189, 163, 28 },
		{ 189, 164, 28 },
		{ 189, 165, 28 },
		{ 189, 166, 28 },
		{ 190, 131, 28 },
		{ 190, 132, 28 },
		{ 190, 133, 28 },
		{ 190, 134, 28 },
		{ 190, 135, 28 },
		{ 190, 136, 28 },
		{ 190, 137, 28 },
		{ 190, 138, 28 },
		{ 190, 139, 28 },
		{ 190, 140, 28 },
		{ 190, 141, 28 },
		{ 190, 142, 28 },
		{ 190, 143, 28 },
		{ 190, 144, 28 },
		{ 190, 145, 28 },
		{ 190, 146, 28 },
		{ 190, 147, 28 },
		{ 190, 148, 28 },
		{ 190, 149, 28 },
		{ 190, 150, 28 },
		{ 190, 151, 28 },
		{ 190, 152, 28 },
		{ 190, 153, 28 },
		{ 190, 154, 28 },
		{ 190, 155, 28 },
		{ 190, 156, 28 },
		{ 190, 157, 28 },
		{ 190, 158, 28 },
		{ 190, 159, 28 },
		{ 190, 160, 28 },
		{ 190, 161, 28 },
		{ 190, 162, 28 },
		{ 190, 163, 28 },
		{ 190, 164, 28 },
		{ 190, 165, 28 },
		{ 190, 166, 28 },
		{ 190, 167, 28 },
		{ 191, 131, 28 },
		{ 191, 132, 28 },
		{ 191, 133, 28 },
		{ 191, 134, 28 },
		{ 191, 135, 28 },
		{ 191, 136, 28 },
		{ 191, 137, 28 },
		{ 191, 138, 28 },
		{ 191, 139, 28 },
		{ 191, 140, 28 },
		{ 191, 141, 28 },
		{ 191, 142, 28 },
		{ 191, 143, 28 },
		{ 191, 144, 28 },
		{ 191, 145, 28 },
		{ 191, 146, 28 },
		{ 191, 147, 28 },
		{ 191, 148, 28 },
		{ 191, 149, 28 },
		{ 191, 150, 28 },
		{ 191, 151, 28 },
		{ 191, 152, 28 },
		{ 191, 153, 28 },
		{ 191, 154, 28 },
		{ 191, 155, 28 },
		{ 191, 156, 28 },
		{ 191, 157, 28 },
		{ 191, 158, 28 },
		{ 191, 159, 28 },
		{ 191, 160, 28 },
		{ 191, 161, 28 },
		{ 191, 162, 28 },
		{ 191, 163, 28 },
		{ 191, 164, 28 },
		{ 191, 165, 28 },
		{ 191, 166, 28 },
		{ 191, 167, 28 },
		{ 191, 168, 28 },
		{ 192, 131, 28 },
		{ 192, 132, 28 },
		{ 192, 133, 28 },
		{ 192, 134, 28 },
		{ 192, 135, 28 },
		{ 192, 136, 28 },
		{ 192, 137, 28 },
		{ 192, 138, 28 },
		{ 192, 139, 28 },
		{ 192, 140, 28 },
		{ 192, 141, 28 },
		{ 192, 142, 28 },
		{ 192, 143, 28 },
		{ 192, 144, 28 },
		{ 192, 145, 28 },
		{ 192, 146, 28 },
		{ 192, 147, 28 },
		{ 192, 148, 28 },
		{ 192, 149, 28 },
		{ 192, 150, 28 },
		{ 192, 151, 28 },
		{ 192, 152, 28 },
		{ 192, 153, 28 },
		{ 192, 154, 28 },
		{ 192, 155, 28 },
		{ 192, 156, 28 },
		{ 192, 157, 28 },
		{ 192, 158, 28 },
		{ 192, 159, 28 },
		{ 192, 160, 28 },
		{ 192, 161, 28 },
		{ 192, 162, 28 },
		{ 192, 163, 28 },
		{ 192, 164, 28 },
		{ 192, 165, 28 },
		{ 192, 166, 28 },
		{ 192, 167, 28 },
		{ 192, 168, 28 },
		{ 192, 169, 28 },
		{ 192, 170, 28 },
		{ 193, 131, 28 },
		{ 193, 132, 28 },
		{ 193, 133, 28 },
		{ 193, 134, 28 },
		{ 193, 135, 28 },
		{ 193, 136, 28 },
		{ 193, 137, 28 },
		{ 193, 138, 28 },
		{ 193, 139, 28 },
		{ 193, 140, 28 },
		{ 193, 141, 28 },
		{ 193, 142, 28 },
		{ 193, 143, 28 },
		{ 193, 144, 28 },
		{ 193, 145, 28 },
		{ 193, 146, 28 },
		{ 193, 147, 28 },
		{ 193, 148, 28 },
		{ 193, 149, 28 },
		{ 193, 150, 28 },
		{ 193, 151, 28 },
		{ 193, 152, 28 },
		{ 193, 153, 28 },
		{ 193, 154, 28 },
		{ 193, 155, 28 },
		{ 193, 156, 28 },
		{ 193, 157, 28 },
		{ 193, 158, 28 },
		{ 193, 159, 28 },
		{ 193, 160, 28 },
		{ 193, 161, 28 },
		{ 193, 162, 28 },
		{ 193, 163, 28 },
		{ 193, 164, 28 },
		{ 193, 165, 28 },
		{ 193, 166, 28 },
		{ 193, 167, 28 },
		{ 193, 168, 28 },
		{ 193, 169, 28 },
		{ 193, 170, 28 },
		{ 193, 171, 28 },
		{ 194, 131, 28 },
		{ 194, 132, 28 },
		{ 194, 133, 28 },
		{ 194, 134, 28 },
		{ 194, 135, 28 },
		{ 194, 136, 28 },
		{ 194, 137, 28 },
		{ 194, 138, 28 },
		{ 194, 139, 28 },
		{ 194, 140, 28 },
		{ 194, 141, 28 },
		{ 194, 142, 28 },
		{ 194, 143, 28 },
		{ 194, 144, 28 },
		{ 194, 145, 28 },
		{ 194, 146, 28 },
		{ 194, 147, 28 },
		{ 194, 148, 28 },
		{ 194, 149, 28 },
		{ 194, 150, 28 },
		{ 194, 151, 28 },
		{ 194, 152, 28 },
		{ 194, 153, 28 },
		{ 194, 154, 28 },
		{ 194, 155, 28 },
		{ 194, 156, 28 },
		{ 194, 157, 28 },
		{ 194, 158, 28 },
		{ 194, 159, 28 },
		{ 194, 160, 28 },
		{ 194, 161, 28 },
		{ 194, 162, 28 },
		{ 194, 163, 28 },
		{ 194, 164, 28 },
		{ 194, 165, 28 },
		{ 194, 166, 28 },
		{ 194, 167, 28 },
		{ 194, 168, 28 },
		{ 194, 169, 28 },
		{ 194, 170, 28 },
		{ 194, 171, 28 },
		{ 194, 172, 28 },
		{ 195, 131, 28 },
		{ 195, 132, 28 },
		{ 195, 133, 28 },
		{ 195, 134, 28 },
		{ 195, 135, 28 },
		{ 195, 136, 28 },
		{ 195, 137, 28 },
		{ 195, 138, 28 },
		{ 195, 139, 28 },
		{ 195, 140, 28 },
		{ 195, 141, 28 },
		{ 195, 142, 28 },
		{ 195, 143, 28 },
		{ 195, 144, 28 },
		{ 195, 145, 28 },
		{ 195, 146, 28 },
		{ 195, 147, 28 },
		{ 195, 148, 28 },
		{ 195, 149, 28 },
		{ 195, 150, 28 },
		{ 195, 151, 28 },
		{ 195, 152, 28 },
		{ 195, 153, 28 },
		{ 195, 154, 28 },
		{ 195, 155, 28 },
		{ 195, 156, 28 },
		{ 195, 157, 28 },
		{ 195, 158, 28 },
		{ 195, 159, 28 },
		{ 195, 160, 28 },
		{ 195, 161, 28 },
		{ 195, 162, 28 },
		{ 195, 163, 28 },
		{ 195, 164, 28 },
		{ 195, 165, 28 },
		{ 195, 166, 28 },
		{ 195, 167, 28 },
		{ 195, 168, 28 },
		{ 195, 169, 28 },
		{ 195, 170, 28 },
		{ 195, 171, 28 },
		{ 195, 172, 28 },
		{ 195, 173, 28 },
		{ 195, 174, 28 },
		{ 196, 131, 28 },
		{ 196, 132, 28 },
		{ 196, 133, 28 },
		{ 196, 134, 28 },
		{ 196, 135, 28 },
		{ 196, 136, 28 },
		{ 196, 137, 28 },
		{ 196, 138, 28 },
		{ 196, 139, 28 },
		{ 196, 140, 28 },
		{ 196, 141, 28 },
		{ 196, 142, 28 },
		{ 196, 143, 28 },
		{ 196, 144, 28 },
		{ 196, 145, 28 },
		{ 196, 146, 28 },
		{ 196, 147, 28 },
		{ 196, 148, 28 },
		{ 196, 149, 28 },
		{ 196, 150, 28 },
		{ 196, 151, 28 },
		{ 196, 152, 28 },
		{ 196, 153, 28 },
		{ 196, 154, 28 },
		{ 196, 155, 28 },
		{ 196, 156, 28 },
		{ 196, 157, 28 },
		{ 196, 158, 28 },
		{ 196, 159, 28 },
		{ 196, 160, 28 },
		{ 196, 161, 28 },
		{ 196, 162, 28 },
		{ 196, 163, 28 },
		{ 196, 164, 28 },
		{ 196, 165, 28 },
		{ 196, 166, 28 },
		{ 196, 167, 28 },
		{ 196, 168, 28 },
		{ 196, 169, 28 },
		{ 196, 170, 28 },
		{ 196, 171, 28 },
		{ 196, 172, 28 },
		{ 196, 173, 28 },
		{ 196, 174, 28 },
		{ 196, 175, 28 },
		{ 197, 131, 28 },
		{ 197, 132, 28 },
		{ 197, 133, 28 },
		{ 197, 134, 28 },
		{ 197, 135, 28 },
		{ 197, 136, 28 },
		{ 197, 137, 28 },
		{ 197, 138, 28 },
		{ 197, 139, 28 },
		{ 197, 140, 28 },
		{ 197, 141, 28 },
		{ 197, 142, 28 },
		{ 197, 143, 28 },
		{ 197, 144, 28 },
		{ 197, 145, 28 },
		{ 197, 146, 28 },
		{ 197, 147, 28 },
		{ 197, 148, 28 },
		{ 197, 149, 28 },
		{ 197, 150, 28 },
		{ 197, 151, 28 },
		{ 197, 152, 28 },
		{ 197, 153, 28 },
		{ 197, 154, 28 },
		{ 197, 155, 28 },
		{ 197, 156, 28 },
		{ 197, 157, 28 },
		{ 197, 158, 28 },
		{ 197, 159, 28 },
		{ 197, 160, 28 },
		{ 197, 161, 28 },
		{ 197, 162, 28 },
		{ 197, 163, 28 },
		{ 197, 164, 28 },
		{ 197, 165, 28 },
		{ 197, 166, 28 },
		{ 197, 167, 28 },
		{ 197, 168, 28 },
		{ 197, 169, 28 },
		{ 197, 170, 28 },
		{ 197, 171, 28 },
		{ 197, 172, 28 },
		{ 197, 173, 28 },
		{ 197, 174, 28 },
		{ 197, 175, 28 },
		{ 197, 176, 28 },
		{ 198, 131, 28 },
		{ 198, 132, 28 },
		{ 198, 133, 28 },
		{ 198, 134, 28 },
		{ 198, 135, 28 },
		{ 198, 136, 28 },
		{ 198, 137, 28 },
		{ 198, 138, 28 },
		{ 198, 139, 28 },
		{ 198, 140, 28 },
		{ 198, 141, 28 },
		{ 198, 142, 28 },
		{ 198, 143, 28 },
		{ 198, 144, 28 },
		{ 198, 145, 28 },
		{ 198, 146, 28 },
		{ 198, 147, 28 },
		{ 198, 148, 28 },
		{ 198, 149, 28 },
		{ 198, 150, 28 },
		{ 198, 151, 28 },
		{ 198, 152, 28 },
		{ 198, 153, 28 },
		{ 198, 154, 28 },
		{ 198, 155, 28 },
		{ 198, 156, 28 },
		{ 198, 157, 28 },
		{ 198, 158, 28 },
		{ 198, 159, 28 },
		{ 198, 160, 28 },
		{ 198, 161, 28 },
		{ 198, 162, 28 },
		{ 198, 163, 28 },
		{ 198, 164, 28 },
		{ 198, 165, 28 },
		{ 198, 166, 28 },
		{ 198, 167, 28 },
		{ 198, 168, 28 },
		{ 198, 169, 28 },
		{ 198, 170, 28 },
		{ 198, 171, 28 },
		{ 198, 172, 28 },
		{ 198, 173, 28 },
		{ 198, 174, 28 },
		{ 198, 175, 28 },
		{ 198, 176, 28 },
		{ 198, 177, 28 },
		{ 198, 178, 28 },
		{ 199, 131, 28 },
		{ 199, 132, 28 },
		{ 199, 133, 28 },
		{ 199, 134, 28 },
		{ 199, 135, 28 },
		{ 199, 136, 28 },
		{ 199, 137, 28 },
		{ 199, 138, 28 },
		{ 199, 139, 28 },
		{ 199, 140, 28 },
		{ 199, 141, 28 },
		{ 199, 142, 28 },
		{ 199, 143, 28 },
		{ 199, 144, 28 },
		{ 199, 145, 28 },
		{ 199, 146, 28 },
		{ 199, 147, 28 },
		{ 199, 148, 28 },
		{ 199, 149, 28 },
		{ 199, 150, 28 },
		{ 199, 151, 28 },
		{ 199, 152, 28 },
		{ 199, 153, 28 },
		{ 199, 154, 28 },
		{ 199, 155, 28 },
		{ 199, 156, 28 },
		{ 199, 157, 28 },
		{ 199, 158, 28 },
		{ 199, 159, 28 },
		{ 199, 160, 28 },
		{ 199, 161, 28 },
		{ 199, 162, 28 },
		{ 199, 163, 28 },
		{ 199, 164, 28 },
		{ 199, 165, 28 },
		{ 199, 166, 28 },
		{ 199, 167, 28 },
		{ 199, 168, 28 },
		{ 199, 169, 28 },
		{ 199, 170, 28 },
		{ 199, 171, 28 },
		{ 199, 172, 28 },
		{ 199, 173, 28 },
		{ 199, 174, 28 },
		{ 199, 175, 28 },
		{ 199, 176, 28 },
		{ 199, 177, 28 },
		{ 199, 178, 28 },
		{ 199, 179, 28 },
		{ 200, 131, 28 },
		{ 200, 132, 28 },
		{ 200, 133, 28 },
		{ 200, 134, 28 },
		{ 200, 135, 28 },
		{ 200, 136, 28 },
		{ 200, 137, 28 },
		{ 200, 138, 28 },
		{ 200, 139, 28 },
		{ 200, 140, 28 },
		{ 200, 141, 28 },
		{ 200, 142, 28 },
		{ 200, 143, 28 },
		{ 200, 144, 28 },
		{ 200, 145, 28 },
		{ 200, 146, 28 },
		{ 200, 147, 28 },
		{ 200, 148, 28 },
		{ 200, 149, 28 },
		{ 200, 150, 28 },
		{ 200, 151, 28 },
		{ 200, 152, 28 },
		{ 200, 153, 28 },
		{ 200, 154, 28 },
		{ 200, 155, 28 },
		{ 200, 156, 28 },
		{ 200, 157, 28 },
		{ 200, 158, 28 },
		{ 200, 159, 28 },
		{ 200, 160, 28 },
		{ 200, 161, 28 },
		{ 200, 162, 28 },
		{ 200, 163, 28 },
		{ 200, 164, 28 },
		{ 200, 165, 28 },
		{ 200, 166, 28 },
		{ 200, 167, 28 },
		{ 200, 168, 28 },
		{ 200, 169, 28 },
		{ 200, 170, 28 },
		{ 200, 171, 28 },
		{ 200, 172, 28 },
		{ 200, 173, 28 },
		{ 200, 174, 28 },
		{ 200, 175, 28 },
		{ 200, 176, 28 },
		{ 200, 177, 28 },
		{ 200, 178, 28 },
		{ 200, 179, 28 },
		{ 200, 180, 28 },
		{ 201, 131, 28 },
		{ 201, 132, 28 },
		{ 201, 133, 28 },
		{ 201, 134, 28 },
		{ 201, 135, 28 },
		{ 201, 136, 28 },
		{ 201, 137, 28 },
		{ 201, 138, 28 },
		{ 201, 139, 28 },
		{ 201, 140, 28 },
		{ 201, 141, 28 },
		{ 201, 142, 28 },
		{ 201, 143, 28 },
		{ 201, 144, 28 },
		{ 201, 145, 28 },
		{ 201, 146, 28 },
		{ 201, 147, 28 },
		{ 201, 148, 28 },
		{ 201, 149, 28 },
		{ 201, 150, 28 },
		{ 201, 151, 28 },
		{ 201, 152, 28 },
		{ 201, 153, 28 },
		{ 201, 154, 28 },
		{ 201, 155, 28 },
		{ 201, 156, 28 },
		{ 201, 157, 28 },
		{ 201, 158, 28 },
		{ 201, 159, 28 },
		{ 201, 160, 28 },
		{ 201, 161, 28 },
		{ 201, 162, 28 },
		{ 201, 163, 28 },
		{ 201, 164, 28 },
		{ 201, 165, 28 },
		{ 201, 166, 28 },
		{ 201, 167, 28 },
		{ 201, 168, 28 },
		{ 201, 169, 28 },
		{ 201, 170, 28 },
		{ 201, 171, 28 },
		{ 201, 172, 28 },
		{ 201, 173, 28 },
		{ 201, 174, 28 },
		{ 201, 175, 28 },
		{ 201, 176, 28 },
		{ 201, 177, 28 },
		{ 201, 178, 28 },
		{ 201, 179, 28 },
		{ 201, 180, 28 },
		{ 201, 181, 28 },
		{ 201, 182, 28 },
		{ 202, 131, 28 },
		{ 202, 132, 28 },
		{ 202, 133, 28 },
		{ 202, 134, 28 },
		{ 202, 135, 28 },
		{ 202, 136, 28 },
		{ 202, 137, 28 },
		{ 202, 138, 28 },
		{ 202, 139, 28 },
		{ 202, 140, 28 },
		{ 202, 141, 28 },
		{ 202, 142, 28 },
		{ 202, 143, 28 },
		{ 202, 144, 28 },
		{ 202, 145, 28 },
		{ 202, 146, 28 },
		{ 202, 147, 28 },
		{ 202, 148, 28 },
		{ 202, 149, 28 },
		{ 202, 150, 28 },
		{ 202, 151, 28 },
		{ 202, 152, 28 },
		{ 202, 153, 28 },
		{ 202, 154, 28 },
		{ 202, 155, 28 },
		{ 202, 156, 28 },
		{ 202, 157, 28 },
		{ 202, 158, 28 },
		{ 202, 159, 28 },
		{ 202, 160, 28 },
		{ 202, 161, 28 },
		{ 202, 162, 28 },
		{ 202, 163, 28 },
		{ 202, 164, 28 },
		{ 202, 165, 28 },
		{ 202, 166, 28 },
		{ 202, 167, 28 },
		{ 202, 168, 28 },
		{ 202, 169, 28 },
		{ 202, 170, 28 },
		{ 202, 171, 28 },
		{ 202, 172, 28 },
		{ 202, 173, 28 },
		{ 202, 174, 28 },
		{ 202, 175, 28 },
		{ 202, 176, 28 },
		{ 202, 177, 28 },
		{ 202, 178, 28 },
		{ 202, 179, 28 },
		{ 202, 180, 28 },
		{ 202, 181, 28 },
		{ 202, 182, 28 },
		{ 202, 183, 28 },
		{ 203, 131, 28 },
		{ 203, 132, 28 },
		{ 203, 133, 28 },
		{ 203, 134, 28 },
		{ 203, 135, 28 },
		{ 203, 136, 28 },
		{ 203, 137, 28 },
		{ 203, 138, 28 },
		{ 203, 139, 28 },
		{ 203, 140, 28 },
		{ 203, 141, 28 },
		{ 203, 142, 28 },
		{ 203, 143, 28 },
		{ 203, 144, 28 },
		{ 203, 145, 28 },
		{ 203, 146, 28 },
		{ 203, 147, 28 },
		{ 203, 148, 28 },
		{ 203, 149, 28 },
		{ 203, 150, 28 },
		{ 203, 151, 28 },
		{ 203, 152, 28 },
		{ 203, 153, 28 },
		{ 203, 154, 28 },
		{ 203, 155, 28 },
		{ 203, 156, 28 },
		{ 203, 157, 28 },
		{ 203, 158, 28 },
		{ 203, 159, 28 },
		{ 203, 160, 28 },
		{ 203, 161, 28 },
		{ 203, 162, 28 },
		{ 203, 163, 28 },
		{ 203, 164, 28 },
		{ 203, 165, 28 },
		{ 203, 166, 28 },
		{ 203, 167, 28 },
		{ 203, 168, 28 },
		{ 203, 169, 28 },
		{ 203, 170, 28 },
		{ 203, 171, 28 },
		{ 203, 172, 28 },
		{ 203, 173, 28 },
		{ 203, 174, 28 },
		{ 203, 175, 28 },
		{ 203, 176, 28 },
		{ 203, 177, 28 },
		{ 203, 178, 28 },
		{ 203, 179, 28 },
		{ 203, 180, 28 },
		{ 203, 181, 28 },
		{ 203, 182, 28 },
		{ 203, 183, 28 },
		{ 203, 184, 28 },
		{ 204, 131, 28 },
		{ 204, 132, 28 },
		{ 204, 133, 28 },
		{ 204, 134, 28 },
		{ 204, 135, 28 },
		{ 204, 136, 28 },
		{ 204, 137, 28 },
		{ 204, 138, 28 },
		{ 204, 139, 28 },
		{ 204, 140, 28 },
		{ 204, 141, 28 },
		{ 204, 142, 28 },
		{ 204, 143, 28 },
		{ 204, 144, 28 },
		{ 204, 145, 28 },
		{ 204, 146, 28 },
		{ 204, 147, 28 },
		{ 204, 148, 28 },
		{ 204, 149, 28 },
		{ 204, 150, 28 },
		{ 204, 151, 28 },
		{ 204, 152, 28 },
		{ 204, 153, 28 },
		{ 204, 154, 28 },
		{ 204, 155, 28 },
		{ 204, 156, 28 },
		{ 204, 157, 28 },
		{ 204, 158, 28 },
		{ 204, 159, 28 },
		{ 204, 160, 28 },
		{ 204, 161, 28 },
		{ 204, 162, 28 },
		{ 204, 163, 28 },
		{ 204, 164, 28 },
		{ 204, 165, 28 },
		{ 204, 166, 28 },
		{ 204, 167, 28 },
		{ 204, 168, 28 },
		{ 204, 169, 28 },
		{ 204, 170, 28 },
		{ 204, 171, 28 },
		{ 204, 172, 28 },
		{ 204, 173, 28 },
		{ 204, 174, 28 },
		{ 204, 175, 28 },
		{ 204, 176, 28 },
		{ 204, 177, 28 },
		{ 204, 178, 28 },
		{ 204, 179, 28 },
		{ 204, 180, 28 },
		{ 204, 181, 28 },
		{ 204, 182, 28 },
		{ 204, 183, 28 },
		{ 204, 184, 28 },
		{ 204, 185, 28 },
		{ 204, 186, 28 },
		{ 205, 131, 28 },
		{ 205, 132, 28 },
		{ 205, 133, 28 },
		{ 205, 134, 28 },
		{ 205, 135, 28 },
		{ 205, 136, 28 },
		{ 205, 137, 28 },
		{ 205, 138, 28 },
		{ 205, 139, 28 },
		{ 205, 140, 28 },
		{ 205, 141, 28 },
		{ 205, 142, 28 },
		{ 205, 143, 28 },
		{ 205, 144, 28 },
		{ 205, 145, 28 },
		{ 205, 146, 28 },
		{ 205, 147, 28 },
		{ 205, 148, 28 },
		{ 205, 149, 28 },
		{ 205, 150, 28 },
		{ 205, 151, 28 },
		{ 205, 152, 28 },
		{ 205, 153, 28 },
		{ 205, 154, 28 },
		{ 205, 155, 28 },
		{ 205, 156, 28 },
		{ 205, 157, 28 },
		{ 205, 158, 28 },
		{ 205, 159, 28 },
		{ 205, 160, 28 },
		{ 205, 161, 28 },
		{ 205, 162, 28 },
		{ 205, 163, 28 },
		{ 205, 164, 28 },
		{ 205, 165, 28 },
		{ 205, 166, 28 },
		{ 205, 167, 28 },
		{ 205, 168, 28 },
		{ 205, 169, 28 },
		{ 205, 170, 28 },
		{ 205, 171, 28 },
		{ 205, 172, 28 },
		{ 205, 173, 28 },
		{ 205, 174, 28 },
		{ 205, 175, 28 },
		{ 205, 176, 28 },
		{ 205, 177, 28 },
		{ 205, 178, 28 },
		{ 205, 179, 28 },
		{ 205, 180, 28 },
		{ 205, 181, 28 },
		{ 205, 182, 28 },
		{ 205, 183, 28 },
		{ 205, 184, 28 },
		{ 205, 185, 28 },
		{ 205, 186, 28 },
		{ 205, 187, 28 },
		{ 206, 131, 28 },
		{ 206, 132, 28 },
		{ 206, 133, 28 },
		{ 206, 134, 28 },
		{ 206, 135, 28 },
		{ 206, 136, 28 },
		{ 206, 137, 28 },
		{ 206, 138, 28 },
		{ 206, 139, 28 },
		{ 206, 140, 28 },
		{ 206, 141, 28 },
		{ 206, 142, 28 },
		{ 206, 143, 28 },
		{ 206, 144, 28 },
		{ 206, 145, 28 },
		{ 206, 146, 28 },
		{ 206, 147, 28 },
		{ 206, 148, 28 },
		{ 206, 149, 28 },
		{ 206, 150, 28 },
		{ 206, 151, 28 },
		{ 206, 152, 28 },
		{ 206, 153, 28 },
		{ 206, 154, 28 },
		{ 206, 155, 28 },
		{ 206, 156, 28 },
		{ 206, 157, 28 },
		{ 206, 158, 28 },
		{ 206, 159, 28 },
		{ 206, 160, 28 },
		{ 206, 161, 28 },
		{ 206, 162, 28 },
		{ 206, 163, 28 },
		{ 206, 164, 28 },
		{ 206, 165, 28 },
		{ 206, 166, 28 },
		{ 206, 167, 28 },
		{ 206, 168, 28 },
		{ 206, 169, 28 },
		{ 206, 170, 28 },
		{ 206, 171, 28 },
		{ 206, 172, 28 },
		{ 206, 173, 28 },
		{ 206, 174, 28 },
		{ 206, 175, 28 },
		{ 206, 176, 28 },
		{ 206, 177, 28 },
		{ 206, 178, 28 },
		{ 206, 179, 28 },
		{ 206, 180, 28 },
		{ 206, 181, 28 },
		{ 206, 182, 28 },
		{ 206, 183, 28 },
		{ 206, 184, 28 },
		{ 206, 185, 28 },
		{ 206, 186, 28 },
		{ 206, 187, 28 },
		{ 206, 188, 28 },
		{ 207, 131, 28 },
		{ 207, 132, 28 },
		{ 207, 133, 28 },
		{ 207, 134, 28 },
		{ 207, 135, 28 },
		{ 207, 136, 28 },
		{ 207, 137, 28 },
		{ 207, 138, 28 },
		{ 207, 139, 28 },
		{ 207, 140, 28 },
		{ 207, 141, 28 },
		{ 207, 142, 28 },
		{ 207, 143, 28 },
		{ 207, 144, 28 },
		{ 207, 145, 28 },
		{ 207, 146, 28 },
		{ 207, 147, 28 },
		{ 207, 148, 28 },
		{ 207, 149, 28 },
		{ 207, 150, 28 },
		{ 207, 151, 28 },
		{ 207, 152, 28 },
		{ 207, 153, 28 },
		{ 207, 154, 28 },
		{ 207, 155, 28 },
		{ 207, 156, 28 },
		{ 207, 157, 28 },
		{ 207, 158, 28 },
		{ 207, 159, 28 },
		{ 207, 160, 28 },
		{ 207, 161, 28 },
		{ 207, 162, 28 },
		{ 207, 163, 28 },
		{ 207, 164, 28 },
		{ 207, 165, 28 },
		{ 207, 166, 28 },
		{ 207, 167, 28 },
		{ 207, 168, 28 },
		{ 207, 169, 28 },
		{ 207, 170, 28 },
		{ 207, 171, 28 },
		{ 207, 172, 28 },
		{ 207, 173, 28 },
		{ 207, 174, 28 },
		{ 207, 175, 28 },
		{ 207, 176, 28 },
		{ 207, 177, 28 },
		{ 207, 178, 28 },
		{ 207, 179, 28 },
		{ 207, 180, 28 },
		{ 207, 181, 28 },
		{ 207, 182, 28 },
		{ 207, 183, 28 },
		{ 207, 184, 28 },
		{ 207, 185, 28 },
		{ 207, 186, 28 },
		{ 207, 187, 28 },
		{ 207, 188, 28 },
		{ 207, 189, 28 },
		{ 207, 190, 28 },
		{ 208, 131, 28 },
		{ 208, 132, 28 },
		{ 208, 133, 28 },
		{ 208, 134, 28 },
		{ 208, 135, 28 },
		{ 208, 136, 28 },
		{ 208, 137, 28 },
		{ 208, 138, 28 },
		{ 208, 139, 28 },
		{ 208, 140, 28 },
		{ 208, 141, 28 },
		{ 208, 142, 28 },
		{ 208, 143, 28 },
		{ 208, 144, 28 },
		{ 208, 145, 28 },
		{ 208, 146, 28 },
		{ 208, 147, 28 },
		{ 208, 148, 28 },
		{ 208, 149, 28 },
		{ 208, 150, 28 },
		{ 208, 151, 28 },
		{ 208, 152, 28 },
		{ 208, 153, 28 },
		{ 208, 154, 28 },
		{ 208, 155, 28 },
		{ 208, 156, 28 },
		{ 208, 157, 28 },
		{ 208, 158, 28 },
		{ 208, 159, 28 },
		{ 208, 160, 28 },
		{ 208, 161, 28 },
		{ 208, 162, 28 },
		{ 208, 163, 28 },
		{ 208, 164, 28 },
		{ 208, 165, 28 },
		{ 208, 166, 28 },
		{ 208, 167, 28 },
		{ 208, 168, 28 },
		{ 208, 169, 28 },
		{ 208, 170, 28 },
		{ 208, 171, 28 },
		{ 208, 172, 28 },
		{ 208, 173, 28 },
		{ 208, 174, 28 },
		{ 208, 175, 28 },
		{ 208, 176, 28 },
		{ 208, 177, 28 },
		{ 208, 178, 28 },
		{ 208, 179, 28 },
		{ 208, 180, 28 },
		{ 208, 181, 28 },
		{ 208, 182, 28 },
		{ 208, 183, 28 },
		{ 208, 184, 28 },
		{ 208, 185, 28 },
		{ 208, 186, 28 },
		{ 208, 187, 28 },
		{ 208, 188, 28 },
		{ 208, 189, 28 },
		{ 208, 190, 28 },
		{ 208, 191, 28 },
		{ 209, 131, 28 },
		{ 209, 132, 28 },
		{ 209, 133, 28 },
		{ 209, 134, 28 },
		{ 209, 135, 28 },
		{ 209, 136, 28 },
		{ 209, 137, 28 },
		{ 209, 138, 28 },
		{ 209, 139, 28 },
		{ 209, 140, 28 },
		{ 209, 141, 28 },
		{ 209, 142, 28 },
		{ 209, 143, 28 },
		{ 209, 144, 28 },
		{ 209, 145, 28 },
		{ 209, 146, 28 },
		{ 209, 147, 28 },
		{ 209, 148, 28 },
		{ 209, 149, 28 },
		{ 209, 150, 28 },
		{ 209, 151, 28 },
		{ 209, 152, 28 },
		{ 209, 153, 28 },
		{ 209, 154, 28 },
		{ 209, 155, 28 },
		{ 209, 156, 28 },
		{ 209, 157, 28 },
		{ 209, 158, 28 },
		{ 209, 159, 28 },
		{ 209, 160, 28 },
		{ 209, 161, 28 },
		{ 209, 162, 28 },
		{ 209, 163, 28 },
		{ 209, 164, 28 },
		{ 209, 165, 28 },
		{ 209, 166, 28 },
		{ 209, 167, 28 },
		{ 209, 168, 28 },
		{ 209, 169, 28 },
		{ 209, 170, 28 },
		{ 209, 171, 28 },
		{ 209, 172, 28 },
		{ 209, 173, 28 },
		{ 209, 174, 28 },
		{ 209, 175, 28 },
		{ 209, 176, 28 },
		{ 209, 177, 28 },
		{ 209, 178, 28 },
		{ 209, 179, 28 },
		{ 209, 180, 28 },
		{ 209, 181, 28 },
		{ 209, 182, 28 },
		{ 209, 183, 28 },
		{ 209, 184, 28 },
		{ 209, 185, 28 },
		{ 209, 186, 28 },
		{ 209, 187, 28 },
		{ 209, 188, 28 },
		{ 209, 189, 28 },
		{ 209, 190, 28 },
		{ 209, 191, 28 },
		{ 209, 192, 28 },
		{ 210, 131, 28 },
		{ 210, 132, 28 },
		{ 210, 133, 28 },
		{ 210, 134, 28 },
		{ 210, 135, 28 },
		{ 210, 136, 28 },
		{ 210, 137, 28 },
		{ 210, 138, 28 },
		{ 210, 139, 28 },
		{ 210, 140, 28 },
		{ 210, 141, 28 },
		{ 210, 142, 28 },
		{ 210, 143, 28 },
		{ 210, 144, 28 },
		{ 210, 145, 28 },
		{ 210, 146, 28 },
		{ 210, 147, 28 },
		{ 210, 148, 28 },
		{ 210, 149, 28 },
		{ 210, 150, 28 },
		{ 210, 151, 28 },
		{ 210, 152, 28 },
		{ 210, 153, 28 },
		{ 210, 154, 28 },
		{ 210, 155, 28 },
		{ 210, 156, 28 },
		{ 210, 157, 28 },
		{ 210, 158, 28 },
		{ 210, 159, 28 },
		{ 210, 160, 28 },
		{ 210, 161, 28 },
		{ 210, 162, 28 },
		{ 210, 163, 28 },
		{ 210, 164, 28 },
		{ 210, 165, 28 },
		{ 210, 166, 28 },
		{ 210, 167, 28 },
		{ 210, 168, 28 },
		{ 210, 169, 28 },
		{ 210, 170, 28 },
		{ 210, 171, 28 },
		{ 210, 172, 28 },
		{ 210, 173, 28 },
		{ 210, 174, 28 },
		{ 210, 175, 28 },
		{ 210, 176, 28 },
		{ 210, 177, 28 },
		{ 210, 178, 28 },
		{ 210, 179, 28 },
		{ 210, 180, 28 },
		{ 210, 181, 28 },
		{ 210, 182, 28 },
		{ 210, 183, 28 },
		{ 210, 184, 28 },
		{ 210, 185, 28 },
		{ 210, 186, 28 },
		{ 210, 187, 28 },
		{ 210, 188, 28 },
		{ 210, 189, 28 },
		{ 210, 190, 28 },
		{ 210, 191, 28 },
		{ 210, 192, 28 },
		{ 210, 193, 28 },
		{ 210, 194, 28 },
		{ 211, 131, 28 },
		{ 211, 132, 28 },
		{ 211, 133, 28 },
		{ 211, 134, 28 },
		{ 211, 135, 28 },
		{ 211, 136, 28 },
		{ 211, 137, 28 },
		{ 211, 138, 28 },
		{ 211, 139, 28 },
		{ 211, 140, 28 },
		{ 211, 141, 28 },
		{ 211, 142, 28 },
		{ 211, 143, 28 },
		{ 211, 144, 28 },
		{ 211, 145, 28 },
		{ 211, 146, 28 },
		{ 211, 147, 28 },
		{ 211, 148, 28 },
		{ 211, 149, 28 },
		{ 211, 150, 28 },
		{ 211, 151, 28 },
		{ 211, 152, 28 },
		{ 211, 153, 28 },
		{ 211, 154, 28 },
		{ 211, 155, 28 },
		{ 211, 156, 28 },
		{ 211, 157, 28 },
		{ 211, 158, 28 },
		{ 211, 159, 28 },
		{ 211, 160, 28 },
		{ 211, 161, 28 },
		{ 211, 162, 28 },
		{ 211, 163, 28 },
		{ 211, 164, 28 },
		{ 211, 165, 28 },
		{ 211, 166, 28 },
		{ 211, 167, 28 },
		{ 211, 168, 28 },
		{ 211, 169, 28 },
		{ 211, 170, 28 },
		{ 211, 171, 28 },
		{ 211, 172, 28 },
		{ 211, 173, 28 },
		{ 211, 174, 28 },
		{ 211, 175, 28 },
		{ 211, 176, 28 },
		{ 211, 177, 28 },
		{ 211, 178, 28 },
		{ 211, 179, 28 },
		{ 211, 180, 28 },
		{ 211, 181, 28 },
		{ 211, 182, 28 },
		{ 211, 183, 28 },
		{ 211, 184, 28 },
		{ 211, 185, 28 },
		{ 211, 186, 28 },
		{ 211, 187, 28 },
		{ 211, 188, 28 },
		{ 211, 189, 28 },
		{ 211, 190, 28 },
		{ 211, 191, 28 },
		{ 211, 192, 28 },
		{ 211, 193, 28 },
		{ 211, 194, 28 },
		{ 211, 195, 28 },
		{ 212, 131, 28 },
		{ 212, 132, 28 },
		{ 212, 133, 28 },
		{ 212, 134, 28 },
		{ 212, 135, 28 },
		{ 212, 136, 28 },
		{ 212, 137, 28 },
		{ 212, 138, 28 },
		{ 212, 139, 28 },
		{ 212, 140, 28 },
		{ 212, 141, 28 },
		{ 212, 142, 28 },
		{ 212, 143, 28 },
		{ 212, 144, 28 },
		{ 212, 145, 28 },
		{ 212, 146, 28 },
		{ 212, 147, 28 },
		{ 212, 148, 28 },
		{ 212, 149, 28 },
		{ 212, 150, 28 },
		{ 212, 151, 28 },
		{ 212, 152, 28 },
		{ 212, 153, 28 },
		{ 212, 154, 28 },
		{ 212, 155, 28 },
		{ 212, 156, 28 },
		{ 212, 157, 28 },
		{ 212, 158, 28 },
		{ 212, 159, 28 },
		{ 212, 160, 28 },
		{ 212, 161, 28 },
		{ 212, 162, 28 },
		{ 212, 163, 28 },
		{ 212, 164, 28 },
		{ 212, 165, 28 },
		{ 212, 166, 28 },
		{ 212, 167, 28 },
		{ 212, 168, 28 },
		{ 212, 169, 28 },
		{ 212, 170, 28 },
		{ 212, 171, 28 },
		{ 212, 172, 28 },
		{ 212, 173, 28 },
		{ 212, 174, 28 },
		{ 212, 175, 28 },
		{ 212, 176, 28 },
		{ 212, 177, 28 },
		{ 212, 178, 28 },
		{ 212, 179, 28 },
		{ 212, 180, 28 },
		{ 212, 181, 28 },
		{ 212, 182, 28 },
		{ 212, 183, 28 },
		{ 212, 184, 28 },
		{ 212, 185, 28 },
		{ 212, 186, 28 },
		{ 212, 187, 28 },
		{ 212, 188, 28 },
		{ 212, 189, 28 },
		{ 212, 190, 28 },
		{ 212, 191, 28 },
		{ 212, 192, 28 },
		{ 212, 193, 28 },
		{ 212, 194, 28 },
		{ 212, 195, 28 },
		{ 212, 196, 28 },
		{ 213, 131, 28 },
		{ 213, 132, 28 },
		{ 213, 133, 28 },
		{ 213, 134, 28 },
		{ 213, 135, 28 },
		{ 213, 136, 28 },
		{ 213, 137, 28 },
		{ 213, 138, 28 },
		{ 213, 139, 28 },
		{ 213, 140, 28 },
		{ 213, 141, 28 },
		{ 213, 142, 28 },
		{ 213, 143, 28 },
		{ 213, 144, 28 },
		{ 213, 145, 28 },
		{ 213, 146, 28 },
		{ 213, 147, 28 },
		{ 213, 148, 28 },
		{ 213, 149, 28 },
		{ 213, 150, 28 },
		{ 213, 151, 28 },
		{ 213, 152, 28 },
		{ 213, 153, 28 },
		{ 213, 154, 28 },
		{ 213, 155, 28 },
		{ 213, 156, 28 },
		{ 213, 157, 28 },
		{ 213, 158, 28 },
		{ 213, 159, 28 },
		{ 213, 160, 28 },
		{ 213, 161, 28 },
		{ 213, 162, 28 },
		{ 213, 163, 28 },
		{ 213, 164, 28 },
		{ 213, 165, 28 },
		{ 213, 166, 28 },
		{ 213, 167, 28 },
		{ 213, 168, 28 },
		{ 213, 169, 28 },
		{ 213, 170, 28 },
		{ 213, 171, 28 },
		{ 213, 172, 28 },
		{ 213, 173, 28 },
		{ 213, 174, 28 },
		{ 213, 175, 28 },
		{ 213, 176, 28 },
		{ 213, 177, 28 },
		{ 213, 178, 28 },
		{ 213, 179, 28 },
		{ 213, 180, 28 },
		{ 213, 181, 28 },
		{ 213, 182, 28 },
		{ 213, 183, 28 },
		{ 213, 184, 28 },
		{ 213, 185, 28 },
		{ 213, 186, 28 },
		{ 213, 187, 28 },
		{ 213, 188, 28 },
		{ 213, 189, 28 },
		{ 213, 190, 28 },
		{ 213, 191, 28 },
		{ 213, 192, 28 },
		{ 213, 193, 28 },
		{ 213, 194, 28 },
		{ 213, 195, 28 },
		{ 213, 196, 28 },
		{ 213, 197, 28 },
		{ 213, 198, 28 },
		{ 214, 131, 28 },
		{ 214, 132, 28 },
		{ 214, 133, 28 },
		{ 214, 134, 28 },
		{ 214, 135, 28 },
		{ 214, 136, 28 },
		{ 214, 137, 28 },
		{ 214, 138, 28 },
		{ 214, 139, 28 },
		{ 214, 140, 28 },
		{ 214, 141, 28 },
		{ 214, 142, 28 },
		{ 214, 143, 28 },
		{ 214, 144, 28 },
		{ 214, 145, 28 },
		{ 214, 146, 28 },
		{ 214, 147, 28 },
		{ 214, 148, 28 },
		{ 214, 149, 28 },
		{ 214, 150, 28 },
		{ 214, 151, 28 },
		{ 214, 152, 28 },
		{ 214, 153, 28 },
		{ 214, 154, 28 },
		{ 214, 155, 28 },
		{ 214, 156, 28 },
		{ 214, 157, 28 },
		{ 214, 158, 28 },
		{ 214, 159, 28 },
		{ 214, 160, 28 },
		{ 214, 161, 28 },
		{ 214, 162, 28 },
		{ 214, 163, 28 },
		{ 214, 164, 28 },
		{ 214, 165, 28 },
		{ 214, 166, 28 },
		{ 214, 167, 28 },
		{ 214, 168, 28 },
		{ 214, 169, 28 },
		{ 214, 170, 28 },
		{ 214, 171, 28 },
		{ 214, 172, 28 },
		{ 214, 173, 28 },
		{ 214, 174, 28 },
		{ 214, 175, 28 },
		{ 214, 176, 28 },
		{ 214, 177, 28 },
		{ 214, 178, 28 },
		{ 214, 179, 28 },
		{ 214, 180, 28 },
		{ 214, 181, 28 },
		{ 214, 182, 28 },
		{ 214, 183, 28 },
		{ 214, 184, 28 },
		{ 214, 185, 28 },
		{ 214, 186, 28 },
		{ 214, 187, 28 },
		{ 214, 188, 28 },
		{ 214, 189, 28 },
		{ 214, 190, 28 },
		{ 214, 191, 28 },
		{ 214, 192, 28 },
		{ 214, 193, 28 },
		{ 214, 194, 28 },
		{ 214, 195, 28 },
		{ 214, 196, 28 },
		{ 214, 197, 28 },
		{ 214, 198, 28 },
		{ 214, 199, 28 },
		{ 215, 131, 28 },
		{ 215, 132, 28 },
		{ 215, 133, 28 },
		{ 215, 134, 28 },
		{ 215, 135, 28 },
		{ 215, 136, 28 },
		{ 215, 137, 28 },
		{ 215, 138, 28 },
		{ 215, 139, 28 },
		{ 215, 140, 28 },
		{ 215, 141, 28 },
		{ 215, 142, 28 },
		{ 215, 143, 28 },
		{ 215, 144, 28 },
		{ 215, 145, 28 },
		{ 215, 146, 28 },
		{ 215, 147, 28 },
		{ 215, 148, 28 },
		{ 215, 149, 28 },
		{ 215, 150, 28 },
		{ 215, 151, 28 },
		{ 215, 152, 28 },
		{ 215, 153, 28 },
		{ 215, 154, 28 },
		{ 215, 155, 28 },
		{ 215, 156, 28 },
		{ 215, 157, 28 },
		{ 215, 158, 28 },
		{ 215, 159, 28 },
		{ 215, 160, 28 },
		{ 215, 161, 28 },
		{ 215, 162, 28 },
		{ 215, 163, 28 },
		{ 215, 164, 28 },
		{ 215, 165, 28 },
		{ 215, 166, 28 },
		{ 215, 167, 28 },
		{ 215, 168, 28 },
		{ 215, 169, 28 },
		{ 215, 170, 28 },
		{ 215, 171, 28 },
		{ 215, 172, 28 },
		{ 215, 173, 28 },
		{ 215, 174, 28 },
		{ 215, 175, 28 },
		{ 215, 176, 28 },
		{ 215, 177, 28 },
		{ 215, 178, 28 },
		{ 215, 179, 28 },
		{ 215, 180, 28 },
		{ 215, 181, 28 },
		{ 215, 182, 28 },
		{ 215, 183, 28 },
		{ 215, 184, 28 },
		{ 215, 185, 28 },
		{ 215, 186, 28 },
		{ 215, 187, 28 },
		{ 215, 188, 28 },
		{ 215, 189, 28 },
		{ 215, 190, 28 },
		{ 215, 191, 28 },
		{ 215, 192, 28 },
		{ 215, 193, 28 },
		{ 215, 194, 28 },
		{ 215, 195, 28 },
		{ 215, 196, 28 },
		{ 215, 197, 28 },
		{ 215, 198, 28 },
		{ 215, 199, 28 },
		{ 215, 200, 28 },
		{ 216, 131, 28 },
		{ 216, 132, 28 },
		{ 216, 133, 28 },
		{ 216, 134, 28 },
		{ 216, 135, 28 },
		{ 216, 136, 28 },
		{ 216, 137, 28 },
		{ 216, 138, 28 },
		{ 216, 139, 28 },
		{ 216, 140, 28 },
		{ 216, 141, 28 },
		{ 216, 142, 28 },
		{ 216, 143, 28 },
		{ 216, 144, 28 },
		{ 216, 145, 28 },
		{ 216, 146, 28 },
		{ 216, 147, 28 },
		{ 216, 148, 28 },
		{ 216, 149, 28 },
		{ 216, 150, 28 },
		{ 216, 151, 28 },
		{ 216, 152, 28 },
		{ 216, 153, 28 },
		{ 216, 154, 28 },
		{ 216, 155, 28 },
		{ 216, 156, 28 },
		{ 216, 157, 28 },
		{ 216, 158, 28 },
		{ 216, 159, 28 },
		{ 216, 160, 28 },
		{ 216, 161, 28 },
		{ 216, 162, 28 },
		{ 216, 163, 28 },
		{ 216, 164, 28 },
		{ 216, 165, 28 },
		{ 216, 166, 28 },
		{ 216, 167, 28 },
		{ 216, 168, 28 },
		{ 216, 169, 28 },
		{ 216, 170, 28 },
		{ 216, 171, 28 },
		{ 216, 172, 28 },
		{ 216, 173, 28 },
		{ 216, 174, 28 },
		{ 216, 175, 28 },
		{ 216, 176, 28 },
		{ 216, 177, 28 },
		{ 216, 178, 28 },
		{ 216, 179, 28 },
		{ 216, 180, 28 },
		{ 216, 181, 28 },
		{ 216, 182, 28 },
		{ 216, 183, 28 },
		{ 216, 184, 28 },
		{ 216, 185, 28 },
		{ 216, 186, 28 },
		{ 216, 187, 28 },
		{ 216, 188, 28 },
		{ 216, 189, 28 },
		{ 216, 190, 28 },
		{ 216, 191, 28 },
		{ 216, 192, 28 },
		{ 216, 193, 28 },
		{ 216, 194, 28 },
		{ 216, 195, 28 },
		{ 216, 196, 28 },
		{ 216, 197, 28 },
		{ 216, 198, 28 },
		{ 216, 199, 28 },
		{ 216, 200, 28 },
		{ 216, 201, 28 },
		{ 216, 202, 28 },
		{ 217, 131, 28 },
		{ 217, 132, 28 },
		{ 217, 133, 28 },
		{ 217, 134, 28 },
		{ 217, 135, 28 },
		{ 217, 136, 28 },
		{ 217, 137, 28 },
		{ 217, 138, 28 },
		{ 217, 139, 28 },
		{ 217, 140, 28 },
		{ 217, 141, 28 },
		{ 217, 142, 28 },
		{ 217, 143, 28 },
		{ 217, 144, 28 },
		{ 217, 145, 28 },
		{ 217, 146, 28 },
		{ 217, 147, 28 },
		{ 217, 148, 28 },
		{ 217, 149, 28 },
		{ 217, 150, 28 },
		{ 217, 151, 28 },
		{ 217, 152, 28 },
		{ 217, 153, 28 },
		{ 217, 154, 28 },
		{ 217, 155, 28 },
		{ 217, 156, 28 },
		{ 217, 157, 28 },
		{ 217, 158, 28 },
		{ 217, 159, 28 },
		{ 217, 160, 28 },
		{ 217, 161, 28 },
		{ 217, 162, 28 },
		{ 217, 163, 28 },
		{ 217, 164, 28 },
		{ 217, 165, 28 },
		{ 217, 166, 28 },
		{ 217, 167, 28 },
		{ 217, 168, 28 },
		{ 217, 169, 28 },
		{ 217, 170, 28 },
		{ 217, 171, 28 },
		{ 217, 172, 28 },
		{ 217, 173, 28 },
		{ 217, 174, 28 },
		{ 217, 175, 28 },
		{ 217, 176, 28 },
		{ 217, 177, 28 },
		{ 217, 178, 28 },
		{ 217, 179, 28 },
		{ 217, 180, 28 },
		{ 217, 181, 28 },
		{ 217, 182, 28 },
		{ 217, 183, 28 },
		{ 217, 184, 28 },
		{ 217, 185, 28 },
		{ 217, 186, 28 },
		{ 217, 187, 28 },
		{ 217, 188, 28 },
		{ 217, 189, 28 },
		{ 217, 190, 28 },
		{ 217, 191, 28 },
		{ 217, 192, 28 },
		{ 217, 193, 28 },
		{ 217, 194, 28 },
		{ 217, 195, 28 },
		{ 217, 196, 28 },
		{ 217, 197, 28 },
		{ 217, 198, 28 },
		{ 217, 199, 28 },
		{ 217, 200, 28 },
		{ 217, 201, 28 },
		{ 217, 202, 28 },
		{ 217, 203, 28 },
		{ 218, 131, 28 },
		{ 218, 132, 28 },
		{ 218, 133, 28 },
		{ 218, 134, 28 },
		{ 218, 135, 28 },
		{ 218, 136, 28 },
		{ 218, 137, 28 },
		{ 218, 138, 28 },
		{ 218, 139, 28 },
		{ 218, 140, 28 },
		{ 218, 141, 28 },
		{ 218, 142, 28 },
		{ 218, 143, 28 },
		{ 218, 144, 28 },
		{ 218, 145, 28 },
		{ 218, 146, 28 },
		{ 218, 147, 28 },
		{ 218, 148, 28 },
		{ 218, 149, 28 },
		{ 218, 150, 28 },
		{ 218, 151, 28 },
		{ 218, 152, 28 },
		{ 218, 153, 28 },
		{ 218, 154, 28 },
		{ 218, 155, 28 },
		{ 218, 156, 28 },
		{ 218, 157, 28 },
		{ 218, 158, 28 },
		{ 218, 159, 28 },
		{ 218, 160, 28 },
		{ 218, 161, 28 },
		{ 218, 162, 28 },
		{ 218, 163, 28 },
		{ 218, 164, 28 },
		{ 218, 165, 28 },
		{ 218, 166, 28 },
		{ 218, 167, 28 },
		{ 218, 168, 28 },
		{ 218, 169, 28 },
		{ 218, 170, 28 },
		{ 218, 171, 28 },
		{ 218, 172, 28 },
		{ 218, 173, 28 },
		{ 218, 174, 28 },
		{ 218, 175, 28 },
		{ 218, 176, 28 },
		{ 218, 177, 28 },
		{ 218, 178, 28 },
		{ 218, 179, 28 },
		{ 218, 180, 28 },
		{ 218, 181, 28 },
		{ 218, 182, 28 },
		{ 218, 183, 28 },
		{ 218, 184, 28 },
		{ 218, 185, 28 },
		{ 218, 186, 28 },
		{ 218, 187, 28 },
		{ 218, 188, 28 },
		{ 218, 189, 28 },
		{ 218, 190, 28 },
		{ 218, 191, 28 },
		{ 218, 192, 28 },
		{ 218, 193, 28 },
		{ 218, 194, 28 },
		{ 218, 195, 28 },
		{ 218, 196, 28 },
		{ 218, 197, 28 },
		{ 218, 198, 28 },
		{ 218, 199, 28 },
		{ 218, 200, 28 },
		{ 218, 201, 28 },
		{ 218, 202, 28 },
		{ 218, 203, 28 },
		{ 218, 204, 28 },
		{ 219, 132, 28 },
		{ 219, 133, 28 },
		{ 219, 134, 28 },
		{ 219, 135, 28 },
		{ 219, 136, 28 },
		{ 219, 137, 28 },
		{ 219, 138, 28 },
		{ 219, 139, 28 },
		{ 219, 140, 28 },
		{ 219, 141, 28 },
		{ 219, 142, 28 },
		{ 219, 143, 28 },
		{ 219, 144, 28 },
		{ 219, 145, 28 },
		{ 219, 146, 28 },
		{ 219, 147, 28 },
		{ 219, 148, 28 },
		{ 219, 149, 28 },
		{ 219, 150, 28 },
		{ 219, 151, 28 },
		{ 219, 152, 28 },
		{ 219, 153, 28 },
		{ 219, 154, 28 },
		{ 219, 155, 28 },
		{ 219, 156, 28 },
		{ 219, 157, 28 },
		{ 219, 158, 28 },
		{ 219, 159, 28 },
		{ 219, 160, 28 },
		{ 219, 161, 28 },
		{ 219, 162, 28 },
		{ 219, 163, 28 },
		{ 219, 164, 28 },
		{ 219, 165, 28 },
		{ 219, 166, 28 },
		{ 219, 167, 28 },
		{ 219, 168, 28 },
		{ 219, 169, 28 },
		{ 219, 170, 28 },
		{ 219, 171, 28 },
		{ 219, 172, 28 },
		{ 219, 173, 28 },
		{ 219, 174, 28 },
		{ 219, 175, 28 },
		{ 219, 176, 28 },
		{ 219, 177, 28 },
		{ 219, 178, 28 },
		{ 219, 179, 28 },
		{ 219, 180, 28 },
		{ 219, 181, 28 },
		{ 219, 182, 28 },
		{ 219, 183, 28 },
		{ 219, 184, 28 },
		{ 219, 185, 28 },
		{ 219, 186, 28 },
		{ 219, 187, 28 },
		{ 219, 188, 28 },
		{ 219, 189, 28 },
		{ 219, 190, 28 },
		{ 219, 191, 28 },
		{ 219, 192, 28 },
		{ 219, 193, 28 },
		{ 219, 194, 28 },
		{ 219, 195, 28 },
		{ 219, 196, 28 },
		{ 219, 197, 28 },
		{ 219, 198, 28 },
		{ 219, 199, 28 },
		{ 219, 200, 28 },
		{ 219, 201, 28 },
		{ 219, 202, 28 },
		{ 219, 203, 28 },
		{ 219, 204, 28 },
		{ 219, 205, 28 },
		{ 219, 206, 28 },
		{ 220, 132, 28 },
		{ 220, 133, 28 },
		{ 220, 134, 28 },
		{ 220, 135, 28 },
		{ 220, 136, 28 },
		{ 220, 137, 28 },
		{ 220, 138, 28 },
		{ 220, 139, 28 },
		{ 220, 140, 28 },
		{ 220, 141, 28 },
		{ 220, 142, 28 },
		{ 220, 143, 28 },
		{ 220, 144, 28 },
		{ 220, 145, 28 },
		{ 220, 146, 28 },
		{ 220, 147, 28 },
		{ 220, 148, 28 },
		{ 220, 149, 28 },
		{ 220, 150, 28 },
		{ 220, 151, 28 },
		{ 220, 152, 28 },
		{ 220, 153, 28 },
		{ 220, 154, 28 },
		{ 220, 155, 28 },
		{ 220, 156, 28 },
		{ 220, 157, 28 },
		{ 220, 158, 28 },
		{ 220, 159, 28 },
		{ 220, 160, 28 },
		{ 220, 161, 28 },
		{ 220, 162, 28 },
		{ 220, 163, 28 },
		{ 220, 164, 28 },
		{ 220, 165, 28 },
		{ 220, 166, 28 },
		{ 220, 167, 28 },
		{ 220, 168, 28 },
		{ 220, 169, 28 },
		{ 220, 170, 28 },
		{ 220, 171, 28 },
		{ 220, 172, 28 },
		{ 220, 173, 28 },
		{ 220, 174, 28 },
		{ 220, 175, 28 },
		{ 220, 176, 28 },
		{ 220, 177, 28 },
		{ 220, 178, 28 },
		{ 220, 179, 28 },
		{ 220, 180, 28 },
		{ 220, 181, 28 },
		{ 220, 182, 28 },
		{ 220, 183, 28 },
		{ 220, 184, 28 },
		{ 220, 185, 28 },
		{ 220, 186, 28 },
		{ 220, 187, 28 },
		{ 220, 188, 28 },
		{ 220, 189, 28 },
		{ 220, 190, 28 },
		{ 220, 191, 28 },
		{ 220, 192, 28 },
		{ 220, 193, 28 },
		{ 220, 194, 28 },
		{ 220, 195, 28 },
		{ 220, 196, 28 },
		{ 220, 197, 28 },
		{ 220, 198, 28 },
		{ 220, 199, 28 },
		{ 220, 200, 28 },
		{ 220, 201, 28 },
		{ 220, 202, 28 },
		{ 220, 203, 28 },
		{ 220, 204, 28 },
		{ 220, 205, 28 },
		{ 220, 206, 28 },
		{ 220, 207, 28 },
		{ 221, 132, 28 },
		{ 221, 133, 28 },
		{ 221, 134, 28 },
		{ 221, 135, 28 },
		{ 221, 136, 28 },
		{ 221, 137, 28 },
		{ 221, 138, 28 },
		{ 221, 139, 28 },
		{ 221, 140, 28 },
		{ 221, 141, 28 },
		{ 221, 142, 28 },
		{ 221, 143, 28 },
		{ 221, 144, 28 },
		{ 221, 145, 28 },
		{ 221, 146, 28 },
		{ 221, 147, 28 },
		{ 221, 148, 28 },
		{ 221, 149, 28 },
		{ 221, 150, 28 },
		{ 221, 151, 28 },
		{ 221, 152, 28 },
		{ 221, 153, 28 },
		{ 221, 154, 28 },
		{ 221, 155, 28 },
		{ 221, 156, 28 },
		{ 221, 157, 28 },
		{ 221, 158, 28 },
		{ 221, 159, 28 },
		{ 221, 160, 28 },
		{ 221, 161, 28 },
		{ 221, 162, 28 },
		{ 221, 163, 28 },
		{ 221, 164, 28 },
		{ 221, 165, 28 },
		{ 221, 166, 28 },
		{ 221, 167, 28 },
		{ 221, 168, 28 },
		{ 221, 169, 28 },
		{ 221, 170, 28 },
		{ 221, 171, 28 },
		{ 221, 172, 28 },
		{ 221, 173, 28 },
		{ 221, 174, 28 },
		{ 221, 175, 28 },
		{ 221, 176, 28 },
		{ 221, 177, 28 },
		{ 221, 178, 28 },
		{ 221, 179, 28 },
		{ 221, 180, 28 },
		{ 221, 181, 28 },
		{ 221, 182, 28 },
		{ 221, 183, 28 },
		{ 221, 184, 28 },
		{ 221, 185, 28 },
		{ 221, 186, 28 },
		{ 221, 187, 28 },
		{ 221, 188, 28 },
		{ 221, 189, 28 },
		{ 221, 190, 28 },
		{ 221, 191, 28 },
		{ 221, 192, 28 },
		{ 221, 193, 28 },
		{ 221, 194, 28 },
		{ 221, 195, 28 },
		{ 221, 196, 28 },
		{ 221, 197, 28 },
		{ 221, 198, 28 },
		{ 221, 199, 28 },
		{ 221, 200, 28 },
		{ 221, 201, 28 },
		{ 221, 202, 28 },
		{ 221, 203, 28 },
		{ 221, 204, 28 },
		{ 221, 205, 28 },
		{ 221, 206, 28 },
		{ 221, 207, 28 },
		{ 221, 208, 28 },
		{ 222, 132, 28 },
		{ 222, 133, 28 },
		{ 222, 134, 28 },
		{ 222, 135, 28 },
		{ 222, 136, 28 },
		{ 222, 137, 28 },
		{ 222, 138, 28 },
		{ 222, 139, 28 },
		{ 222, 140, 28 },
		{ 222, 141, 28 },
		{ 222, 142, 28 },
		{ 222, 143, 28 },
		{ 222, 144, 28 },
		{ 222, 145, 28 },
		{ 222, 146, 28 },
		{ 222, 147, 28 },
		{ 222, 148, 28 },
		{ 222, 149, 28 },
		{ 222, 150, 28 },
		{ 222, 151, 28 },
		{ 222, 152, 28 },
		{ 222, 153, 28 },
		{ 222, 154, 28 },
		{ 222, 155, 28 },
		{ 222, 156, 28 },
		{ 222, 157, 28 },
		{ 222, 158, 28 },
		{ 222, 159, 28 },
		{ 222, 160, 28 },
		{ 222, 161, 28 },
		{ 222, 162, 28 },
		{ 222, 163, 28 },
		{ 222, 164, 28 },
		{ 222, 165, 28 },
		{ 222, 166, 28 },
		{ 222, 167, 28 },
		{ 222, 168, 28 },
		{ 222, 169, 28 },
		{ 222, 170, 28 },
		{ 222, 171, 28 },
		{ 222, 172, 28 },
		{ 222, 173, 28 },
		{ 222, 174, 28 },
		{ 222, 175, 28 },
		{ 222, 176, 28 },
		{ 222, 177, 28 },
		{ 222, 178, 28 },
		{ 222, 179, 28 },
		{ 222, 180, 28 },
		{ 222, 181, 28 },
		{ 222, 182, 28 },
		{ 222, 183, 28 },
		{ 222, 184, 28 },
		{ 222, 185, 28 },
		{ 222, 186, 28 },
		{ 222, 187, 28 },
		{ 222, 188, 28 },
		{ 222, 189, 28 },
		{ 222, 190, 28 },
		{ 222, 191, 28 },
		{ 222, 192, 28 },
		{ 222, 193, 28 },
		{ 222, 194, 28 },
		{ 222, 195, 28 },
		{ 222, 196, 28 },
		{ 222, 197, 28 },
		{ 222, 198, 28 },
		{ 222, 199, 28 },
		{ 222, 200, 28 },
		{ 222, 201, 28 },
		{ 222, 202, 28 },
		{ 222, 203, 28 },
		{ 222, 204, 28 },
		{ 222, 205, 28 },
		{ 222, 206, 28 },
		{ 222, 207, 28 },
		{ 222, 208, 28 },
		{ 222, 209, 28 },
		{ 222, 210, 28 },
		{ 223, 133, 28 },
		{ 223, 134, 28 },
		{ 223, 135, 28 },
		{ 223, 136, 28 },
		{ 223, 137, 28 },
		{ 223, 138, 28 },
		{ 223, 139, 28 },
		{ 223, 140, 28 },
		{ 223, 141, 28 },
		{ 223, 142, 28 },
		{ 223, 143, 28 },
		{ 223, 144, 28 },
		{ 223, 145, 28 },
		{ 223, 146, 28 },
		{ 223, 147, 28 },
		{ 223, 148, 28 },
		{ 223, 149, 28 },
		{ 223, 150, 28 },
		{ 223, 151, 28 },
		{ 223, 152, 28 },
		{ 223, 153, 28 },
		{ 223, 154, 28 },
		{ 223, 155, 28 },
		{ 223, 156, 28 },
		{ 223, 157, 28 },
		{ 223, 158, 28 },
		{ 223, 159, 28 },
		{ 223, 160, 28 },
		{ 223, 161, 28 },
		{ 223, 162, 28 },
		{ 223, 163, 28 },
		{ 223, 164, 28 },
		{ 223, 165, 28 },
		{ 223, 166, 28 },
		{ 223, 167, 28 },
		{ 223, 168, 28 },
		{ 223, 169, 28 },
		{ 223, 170, 28 },
		{ 223, 171, 28 },
		{ 223, 172, 28 },
		{ 223, 173, 28 },
		{ 223, 174, 28 },
		{ 223, 175, 28 },
		{ 223, 176, 28 },
		{ 223, 177, 28 },
		{ 223, 178, 28 },
		{ 223, 179, 28 },
		{ 223, 180, 28 },
		{ 223, 181, 28 },
		{ 223, 182, 28 },
		{ 223, 183, 28 },
		{ 223, 184, 28 },
		{ 223, 185, 28 },
		{ 223, 186, 28 },
		{ 223, 187, 28 },
		{ 223, 188, 28 },
		{ 223, 189, 28 },
		{ 223, 190, 28 },
		{ 223, 191, 28 },
		{ 223, 192, 28 },
		{ 223, 193, 28 },
		{ 223, 194, 28 },
		{ 223, 195, 28 },
		{ 223, 196, 28 },
		{ 223, 197, 28 },
		{ 223, 198, 28 },
		{ 223, 199, 28 },
		{ 223, 200, 28 },
		{ 223, 201, 28 },
		{ 223, 202, 28 },
		{ 223, 203, 28 },
		{ 223, 204, 28 },
		{ 223, 205, 28 },
		{ 223, 206, 28 },
		{ 223, 207, 28 },
		{ 223, 208, 28 },
		{ 223, 209, 28 },
		{ 223, 210, 28 },
		{ 223, 211, 28 },
		{ 224, 133, 28 },
		{ 224, 134, 28 },
		{ 224, 135, 28 },
		{ 224, 136, 28 },
		{ 224, 137, 28 },
		{ 224, 138, 28 },
		{ 224, 139, 28 },
		{ 224, 140, 28 },
		{ 224, 141, 28 },
		{ 224, 142, 28 },
		{ 224, 143, 28 },
		{ 224, 144, 28 },
		{ 224, 145, 28 },
		{ 224, 146, 28 },
		{ 224, 147, 28 },
		{ 224, 148, 28 },
		{ 224, 149, 28 },
		{ 224, 150, 28 },
		{ 224, 151, 28 },
		{ 224, 152, 28 },
		{ 224, 153, 28 },
		{ 224, 154, 28 },
		{ 224, 155, 28 },
		{ 224, 156, 28 },
		{ 224, 157, 28 },
		{ 224, 158, 28 },
		{ 224, 159, 28 },
		{ 224, 160, 28 },
		{ 224, 161, 28 },
		{ 224, 162, 28 },
		{ 224, 163, 28 },
		{ 224, 164, 28 },
		{ 224, 165, 28 },
		{ 224, 166, 28 },
		{ 224, 167, 28 },
		{ 224, 168, 28 },
		{ 224, 169, 28 },
		{ 224, 170, 28 },
		{ 224, 171, 28 },
		{ 224, 172, 28 },
		{ 224, 173, 28 },
		{ 224, 174, 28 },
		{ 224, 175, 28 },
		{ 224, 176, 28 },
		{ 224, 177, 28 },
		{ 224, 178, 28 },
		{ 224, 179, 28 },
		{ 224, 180, 28 },
		{ 224, 181, 28 },
		{ 224, 182, 28 },
		{ 224, 183, 28 },
		{ 224, 184, 28 },
		{ 224, 185, 28 },
		{ 224, 186, 28 },
		{ 224, 187, 28 },
		{ 224, 188, 28 },
		{ 224, 189, 28 },
		{ 224, 190, 28 },
		{ 224, 191, 28 },
		{ 224, 192, 28 },
		{ 224, 193, 28 },
		{ 224, 194, 28 },
		{ 224, 195, 28 },
		{ 224, 196, 28 },
		{ 224, 197, 28 },
		{ 224, 198, 28 },
		{ 224, 199, 28 },
		{ 224, 200, 28 },
		{ 224, 201, 28 },
		{ 224, 202, 28 },
		{ 224, 203, 28 },
		{ 224, 204, 28 },
		{ 224, 205, 28 },
		{ 224, 206, 28 },
		{ 224, 207, 28 },
		{ 224, 208, 28 },
		{ 224, 209, 28 },
		{ 224, 210, 28 },
		{ 224, 211, 28 },
		{ 224, 212, 28 },
		{ 225, 134, 28 },
		{ 225, 135, 28 },
		{ 225, 136, 28 },
		{ 225, 137, 28 },
		{ 225, 138, 28 },
		{ 225, 139, 28 },
		{ 225, 140, 28 },
		{ 225, 141, 28 },
		{ 225, 142, 28 },
		{ 225, 143, 28 },
		{ 225, 144, 28 },
		{ 225, 145, 28 },
		{ 225, 146, 28 },
		{ 225, 147, 28 },
		{ 225, 148, 28 },
		{ 225, 149, 28 },
		{ 225, 150, 28 },
		{ 225, 151, 28 },
		{ 225, 152, 28 },
		{ 225, 153, 28 },
		{ 225, 154, 28 },
		{ 225, 155, 28 },
		{ 225, 156, 28 },
		{ 225, 157, 28 },
		{ 225, 158, 28 },
		{ 225, 159, 28 },
		{ 225, 160, 28 },
		{ 225, 161, 28 },
		{ 225, 162, 28 },
		{ 225, 163, 28 },
		{ 225, 164, 28 },
		{ 225, 165, 28 },
		{ 225, 166, 28 },
		{ 225, 167, 28 },
		{ 225, 168, 28 },
		{ 225, 169, 28 },
		{ 225, 170, 28 },
		{ 225, 171, 28 },
		{ 225, 172, 28 },
		{ 225, 173, 28 },
		{ 225, 174, 28 },
		{ 225, 175, 28 },
		{ 225, 176, 28 },
		{ 225, 177, 28 },
		{ 225, 178, 28 },
		{ 225, 179, 28 },
		{ 225, 180, 28 },
		{ 225, 181, 28 },
		{ 225, 182, 28 },
		{ 225, 183, 28 },
		{ 225, 184, 28 },
		{ 225, 185, 28 },
		{ 225, 186, 28 },
		{ 225, 187, 28 },
		{ 225, 188, 28 },
		{ 225, 189, 28 },
		{ 225, 190, 28 },
		{ 225, 191, 28 },
		{ 225, 192, 28 },
		{ 225, 193, 28 },
		{ 225, 194, 28 },
		{ 225, 195, 28 },
		{ 225, 196, 28 },
		{ 225, 197, 28 },
		{ 225, 198, 28 },
		{ 225, 199, 28 },
		{ 225, 200, 28 },
		{ 225, 201, 28 },
		{ 225, 202, 28 },
		{ 225, 203, 28 },
		{ 225, 204, 28 },
		{ 225, 205, 28 },
		{ 225, 206, 28 },
		{ 225, 207, 28 },
		{ 225, 208, 28 },
		{ 225, 209, 28 },
		{ 225, 210, 28 },
		{ 225, 211, 28 },
		{ 225, 212, 28 },
		{ 225, 213, 28 },
		{ 225, 214, 28 },
		{ 226, 136, 28 },
		{ 226, 137, 28 },
		{ 226, 138, 28 },
		{ 226, 139, 28 },
		{ 226, 140, 28 },
		{ 226, 141, 28 },
		{ 226, 142, 28 },
		{ 226, 143, 28 },
		{ 226, 144, 28 },
		{ 226, 145, 28 },
		{ 226, 146, 28 },
		{ 226, 147, 28 },
		{ 226, 148, 28 },
		{ 226, 149, 28 },
		{ 226, 150, 28 },
		{ 226, 151, 28 },
		{ 226, 152, 28 },
		{ 226, 153, 28 },
		{ 226, 154, 28 },
		{ 226, 155, 28 },
		{ 226, 156, 28 },
		{ 226, 157, 28 },
		{ 226, 158, 28 },
		{ 226, 159, 28 },
		{ 226, 160, 28 },
		{ 226, 161, 28 },
		{ 226, 162, 28 },
		{ 226, 163, 28 },
		{ 226, 164, 28 },
		{ 226, 165, 28 },
		{ 226, 166, 28 },
		{ 226, 167, 28 },
		{ 226, 168, 28 },
		{ 226, 169, 28 },
		{ 226, 170, 28 },
		{ 226, 171, 28 },
		{ 226, 172, 28 },
		{ 226, 173, 28 },
		{ 226, 174, 28 },
		{ 226, 175, 28 },
		{ 226, 176, 28 },
		{ 226, 177, 28 },
		{ 226, 178, 28 },
		{ 226, 179, 28 },
		{ 226, 180, 28 },
		{ 226, 181, 28 },
		{ 226, 182, 28 },
		{ 226, 183, 28 },
		{ 226, 184, 28 },
		{ 226, 185, 28 },
		{ 226, 186, 28 },
		{ 226, 187, 28 },
		{ 226, 188, 28 },
		{ 226, 189, 28 },
		{ 226, 190, 28 },
		{ 226, 191, 28 },
		{ 226, 192, 28 },
		{ 226, 193, 28 },
		{ 226, 194, 28 },
		{ 226, 195, 28 },
		{ 226, 196, 28 },
		{ 226, 197, 28 },
		{ 226, 198, 28 },
		{ 226, 199, 28 },
		{ 226, 200, 28 },
		{ 226, 201, 28 },
		{ 226, 202, 28 },
		{ 226, 203, 28 },
		{ 226, 204, 28 },
		{ 226, 205, 28 },
		{ 226, 206, 28 },
		{ 226, 207, 28 },
		{ 226, 208, 28 },
		{ 226, 209, 28 },
		{ 226, 210, 28 },
		{ 226, 211, 28 },
		{ 226, 212, 28 },
		{ 226, 213, 28 },
		{ 226, 214, 28 },
		{ 226, 215, 28 },
		{ 227, 139, 28 },
		{ 227, 140, 28 },
		{ 227, 141, 28 },
		{ 227, 142, 28 },
		{ 227, 143, 28 },
		{ 227, 144, 28 },
		{ 227, 145, 28 },
		{ 227, 146, 28 },
		{ 227, 147, 28 },
		{ 227, 148, 28 },
		{ 227, 149, 28 },
		{ 227, 150, 28 },
		{ 227, 151, 28 },
		{ 227, 152, 28 },
		{ 227, 153, 28 },
		{ 227, 154, 28 },
		{ 227, 155, 28 },
		{ 227, 156, 28 },
		{ 227, 157, 28 },
		{ 227, 158, 28 },
		{ 227, 159, 28 },
		{ 227, 160, 28 },
		{ 227, 161, 28 },
		{ 227, 162, 28 },
		{ 227, 163, 28 },
		{ 227, 164, 28 },
		{ 227, 165, 28 },
		{ 227, 166, 28 },
		{ 227, 167, 28 },
		{ 227, 168, 28 },
		{ 227, 169, 28 },
		{ 227, 170, 28 },
		{ 227, 171, 28 },
		{ 227, 172, 28 },
		{ 227, 173, 28 },
		{ 227, 174, 28 },
		{ 227, 175, 28 },
		{ 227, 176, 28 },
		{ 227, 177, 28 },
		{ 227, 178, 28 },
		{ 227, 179, 28 },
		{ 227, 180, 28 },
		{ 227, 181, 28 },
		{ 227, 182, 28 },
		{ 227, 183, 28 },
		{ 227, 184, 28 },
		{ 227, 185, 28 },
		{ 227, 186, 28 },
		{ 227, 187, 28 },
		{ 227, 188, 28 },
		{ 227, 189, 28 },
		{ 227, 190, 28 },
		{ 227, 191, 28 },
		{ 227, 192, 28 },
		{ 227, 193, 28 },
		{ 227, 194, 28 },
		{ 227, 195, 28 },
		{ 227, 196, 28 },
		{ 227, 197, 28 },
		{ 227, 198, 28 },
		{ 227, 199, 28 },
		{ 227, 200, 28 },
		{ 227, 201, 28 },
		{ 227, 202, 28 },
		{ 227, 203, 28 },
		{ 227, 204, 28 },
		{ 227, 205, 28 },
		{ 227, 206, 28 },
		{ 227, 207, 28 },
		{ 227, 208, 28 },
		{ 227, 209, 28 },
		{ 227, 210, 28 },
		{ 227, 211, 28 },
		{ 227, 212, 28 },
		{ 227, 213, 28 },
		{ 227, 214, 28 },
		{ 227, 215, 28 },
		{ 227, 216, 28 },
		{ 228, 142, 28 },
		{ 228, 143, 28 },
		{ 228, 144, 28 },
		{ 228, 145, 28 },
		{ 228, 146, 28 },
		{ 228, 147, 28 },
		{ 228, 148, 28 },
		{ 228, 149, 28 },
		{ 228, 150, 28 },
		{ 228, 151, 28 },
		{ 228, 152, 28 },
		{ 228, 153, 28 },
		{ 228, 154, 28 },
		{ 228, 155, 28 },
		{ 228, 156, 28 },
		{ 228, 157, 28 },
		{ 228, 158, 28 },
		{ 228, 159, 28 },
		{ 228, 160, 28 },
		{ 228, 161, 28 },
		{ 228, 162, 28 },
		{ 228, 163, 28 },
		{ 228, 164, 28 },
		{ 228, 165, 28 },
		{ 228, 166, 28 },
		{ 228, 167, 28 },
		{ 228, 168, 28 },
		{ 228, 169, 28 },
		{ 228, 170, 28 },
		{ 228, 171, 28 },
		{ 228, 172, 28 },
		{ 228, 173, 28 },
		{ 228, 174, 28 },
		{ 228, 175, 28 },
		{ 228, 176, 28 },
		{ 228, 177, 28 },
		{ 228, 178, 28 },
		{ 228, 179, 28 },
		{ 228, 180, 28 },
		{ 228, 181, 28 },
		{ 228, 182, 28 },
		{ 228, 183, 28 },
		{ 228, 184, 28 },
		{ 228, 185, 28 },
		{ 228, 186, 28 },
		{ 228, 187, 28 },
		{ 228, 188, 28 },
		{ 228, 189, 28 },
		{ 228, 190, 28 },
		{ 228, 191, 28 },
		{ 228, 192, 28 },
		{ 228, 193, 28 },
		{ 228, 194, 28 },
		{ 228, 195, 28 },
		{ 228, 196, 28 },
		{ 228, 197, 28 },
		{ 228, 198, 28 },
		{ 228, 199, 28 },
		{ 228, 200, 28 },
		{ 228, 201, 28 },
		{ 228, 202, 28 },
		{ 228, 203, 28 },
		{ 228, 204, 28 },
		{ 228, 205, 28 },
		{ 228, 206, 28 },
		{ 228, 207, 28 },
		{ 228, 208, 28 },
		{ 228, 209, 28 },
		{ 228, 210, 28 },
		{ 228, 211, 28 },
		{ 228, 212, 28 },
		{ 228, 213, 28 },
		{ 228, 214, 28 },
		{ 228, 215, 28 },
		{ 228, 216, 28 },
		{ 228, 217, 28 },
		{ 228, 218, 28 },
		{ 229, 146, 28 },
		{ 229, 147, 28 },
		{ 229, 148, 28 },
		{ 229, 149, 28 },
		{ 229, 150, 28 },
		{ 229, 151, 28 },
		{ 229, 152, 28 },
		{ 229, 153, 28 },
		{ 229, 154, 28 },
		{ 229, 155, 28 },
		{ 229, 156, 28 },
		{ 229, 157, 28 },
		{ 229, 158, 28 },
		{ 229, 159, 28 },
		{ 229, 160, 28 },
		{ 229, 161, 28 },
		{ 229, 162, 28 },
		{ 229, 163, 28 },
		{ 229, 164, 28 },
		{ 229, 165, 28 },
		{ 229, 166, 28 },
		{ 229, 167, 28 },
		{ 229, 168, 28 },
		{ 229, 169, 28 },
		{ 229, 170, 28 },
		{ 229, 171, 28 },
		{ 229, 172, 28 },
		{ 229, 173, 28 },
		{ 229, 174, 28 },
		{ 229, 175, 28 },
		{ 229, 176, 28 },
		{ 229, 177, 28 },
		{ 229, 178, 28 },
		{ 229, 179, 28 },
		{ 229, 180, 28 },
		{ 229, 181, 28 },
		{ 229, 182, 28 },
		{ 229, 183, 28 },
		{ 229, 184, 28 },
		{ 229, 185, 28 },
		{ 229, 186, 28 },
		{ 229, 187, 28 },
		{ 229, 188, 28 },
		{ 229, 189, 28 },
		{ 229, 190, 28 },
		{ 229, 191, 28 },
		{ 229, 192, 28 },
		{ 229, 193, 28 },
		{ 229, 194, 28 },
		{ 229, 195, 28 },
		{ 229, 196, 28 },
		{ 229, 197, 28 },
		{ 229, 198, 28 },
		{ 229, 199, 28 },
		{ 229, 200, 28 },
		{ 229, 201, 28 },
		{ 229, 202, 28 },
		{ 229, 203, 28 },
		{ 229, 204, 28 },
		{ 229, 205, 28 },
		{ 229, 206, 28 },
		{ 229, 207, 28 },
		{ 229, 208, 28 },
		{ 229, 209, 28 },
		{ 229, 210, 28 },
		{ 229, 211, 28 },
		{ 229, 212, 28 },
		{ 229, 213, 28 },
		{ 229, 214, 28 },
		{ 229, 215, 28 },
		{ 229, 216, 28 },
		{ 229, 217, 28 },
		{ 229, 218, 28 },
		{ 229, 219, 28 },
		{ 230, 150, 28 },
		{ 230, 151, 28 },
		{ 230, 152, 28 },
		{ 230, 153, 28 },
		{ 230, 154, 28 },
		{ 230, 155, 28 },
		{ 230, 156, 28 },
		{ 230, 157, 28 },
		{ 230, 158, 28 },
		{ 230, 159, 28 },
		{ 230, 160, 28 },
		{ 230, 161, 28 },
		{ 230, 162, 28 },
		{ 230, 163, 28 },
		{ 230, 164, 28 },
		{ 230, 165, 28 },
		{ 230, 166, 28 },
		{ 230, 167, 28 },
		{ 230, 168, 28 },
		{ 230, 169, 28 },
		{ 230, 170, 28 },
		{ 230, 171, 28 },
		{ 230, 172, 28 },
		{ 230, 173, 28 },
		{ 230, 174, 28 },
		{ 230, 175, 28 },
		{ 230, 176, 28 },
		{ 230, 177, 28 },
		{ 230, 178, 28 },
		{ 230, 179, 28 },
		{ 230, 180, 28 },
		{ 230, 181, 28 },
		{ 230, 182, 28 },
		{ 230, 183, 28 },
		{ 230, 184, 28 },
		{ 230, 185, 28 },
		{ 230, 186, 28 },
		{ 230, 187, 28 },
		{ 230, 188, 28 },
		{ 230, 189, 28 },
		{ 230, 190, 28 },
		{ 230, 191, 28 },
		{ 230, 192, 28 },
		{ 230, 193, 28 },
		{ 230, 194, 28 },
		{ 230, 195, 28 },
		{ 230, 196, 28 },
		{ 230, 197, 28 },
		{ 230, 198, 28 },
		{ 230, 199, 28 },
		{ 230, 200, 28 },
		{ 230, 201, 28 },
		{ 230, 202, 28 },
		{ 230, 203, 28 },
		{ 230, 204, 28 },
		{ 230, 205, 28 },
		{ 230, 206, 28 },
		{ 230, 207, 28 },
		{ 230, 208, 28 },
		{ 230, 209, 28 },
		{ 230, 210, 28 },
		{ 230, 211, 28 },
		{ 230, 212, 28 },
		{ 230, 213, 28 },
		{ 230, 214, 28 },
		{ 230, 215, 28 },
		{ 230, 216, 28 },
		{ 230, 217, 28 },
		{ 230, 218, 28 },
		{ 230, 219, 28 },
		{ 230, 220, 28 },
		{ 231, 154, 28 },
		{ 231, 155, 28 },
		{ 231, 156, 28 },
		{ 231, 157, 28 },
		{ 231, 158, 28 },
		{ 231, 159, 28 },
		{ 231, 160, 28 },
		{ 231, 161, 28 },
		{ 231, 162, 28 },
		{ 231, 163, 28 },
		{ 231, 164, 28 },
		{ 231, 165, 28 },
		{ 231, 166, 28 },
		{ 231, 167, 28 },
		{ 231, 168, 28 },
		{ 231, 169, 28 },
		{ 231, 170, 28 },
		{ 231, 171, 28 },
		{ 231, 172, 28 },
		{ 231, 173, 28 },
		{ 231, 174, 28 },
		{ 231, 175, 28 },
		{ 231, 176, 28 },
		{ 231, 177, 28 },
		{ 231, 178, 28 },
		{ 231, 179, 28 },
		{ 231, 180, 28 },
		{ 231, 181, 28 },
		{ 231, 182, 28 },
		{ 231, 183, 28 },
		{ 231, 184, 28 },
		{ 231, 185, 28 },
		{ 231, 186, 28 },
		{ 231, 187, 28 },
		{ 231, 188, 28 },
		{ 231, 189, 28 },
		{ 231, 190, 28 },
		{ 231, 191, 28 },
		{ 231, 192, 28 },
		{ 231, 193, 28 },
		{ 231, 194, 28 },
		{ 231, 195, 28 },
		{ 231, 196, 28 },
		{ 231, 197, 28 },
		{ 231, 198, 28 },
		{ 231, 199, 28 },
		{ 231, 200, 28 },
		{ 231, 201, 28 },
		{ 231, 202, 28 },
		{ 231, 203, 28 },
		{ 231, 204, 28 },
		{ 231, 205, 28 },
		{ 231, 206, 28 },
		{ 231, 207, 28 },
		{ 231, 208, 28 },
		{ 231, 209, 28 },
		{ 231, 210, 28 },
		{ 231, 211, 28 },
		{ 231, 212, 28 },
		{ 231, 213, 28 },
		{ 231, 214, 28 },
		{ 231, 215, 28 },
		{ 231, 216, 28 },
		{ 231, 217, 28 },
		{ 231, 218, 28 },
		{ 231, 219, 28 },
		{ 231, 220, 28 },
		{ 231, 221, 28 },
		{ 231, 222, 28 },
		{ 232, 158, 28 },
		{ 232, 159, 28 },
		{ 232, 160, 28 },
		{ 232, 161, 28 },
		{ 232, 162, 28 },
		{ 232, 163, 28 },
		{ 232, 164, 28 },
		{ 232, 165, 28 },
		{ 232, 166, 28 },
		{ 232, 167, 28 },
		{ 232, 168, 28 },
		{ 232, 169, 28 },
		{ 232, 170, 28 },
		{ 232, 171, 28 },
		{ 232, 172, 28 },
		{ 232, 173, 28 },
		{ 232, 174, 28 },
		{ 232, 175, 28 },
		{ 232, 176, 28 },
		{ 232, 177, 28 },
		{ 232, 178, 28 },
		{ 232, 179, 28 },
		{ 232, 180, 28 },
		{ 232, 181, 28 },
		{ 232, 182, 28 },
		{ 232, 183, 28 },
		{ 232, 184, 28 },
		{ 232, 185, 28 },
		{ 232, 186, 28 },
		{ 232, 187, 28 },
		{ 232, 188, 28 },
		{ 232, 189, 28 },
		{ 232, 190, 28 },
		{ 232, 191, 28 },
		{ 232, 192, 28 },
		{ 232, 193, 28 },
		{ 232, 194, 28 },
		{ 232, 195, 28 },
		{ 232, 196, 28 },
		{ 232, 197, 28 },
		{ 232, 198, 28 },
		{ 232, 199, 28 },
		{ 232, 200, 28 },
		{ 232, 201, 28 },
		{ 232, 202, 28 },
		{ 232, 203, 28 },
		{ 232, 204, 28 },
		{ 232, 205, 28 },
		{ 232, 206, 28 },
		{ 232, 207, 28 },
		{ 232, 208, 28 },
		{ 232, 209, 28 },
		{ 232, 210, 28 },
		{ 232, 211, 28 },
		{ 232, 212, 28 },
		{ 232, 213, 28 },
		{ 232, 214, 28 },
		{ 232, 215, 28 },
		{ 232, 216, 28 },
		{ 232, 217, 28 },
		{ 232, 218, 28 },
		{ 232, 219, 28 },
		{ 232, 220, 28 },
		{ 232, 221, 28 },
		{ 232, 222, 28 },
		{ 232, 223, 28 },
		{ 233, 162, 28 },
		{ 233, 163, 28 },
		{ 233, 164, 28 },
		{ 233, 165, 28 },
		{ 233, 166, 28 },
		{ 233, 167, 28 },
		{ 233, 168, 28 },
		{ 233, 169, 28 },
		{ 233, 170, 28 },
		{ 233, 171, 28 },
		{ 233, 172, 28 },
		{ 233, 173, 28 },
		{ 233, 174, 28 },
		{ 233, 175, 28 },
		{ 233, 176, 28 },
		{ 233, 177, 28 },
		{ 233, 178, 28 },
		{ 233, 179, 28 },
		{ 233, 180, 28 },
		{ 233, 181, 28 },
		{ 233, 182, 28 },
		{ 233, 183, 28 },
		{ 233, 184, 28 },
		{ 233, 185, 28 },
		{ 233, 186, 28 },
		{ 233, 187, 28 },
		{ 233, 188, 28 },
		{ 233, 189, 28 },
		{ 233, 190, 28 },
		{ 233, 191, 28 },
		{ 233, 192, 28 },
		{ 233, 193, 28 },
		{ 233, 194, 28 },
		{ 233, 195, 28 },
		{ 233, 196, 28 },
		{ 233, 197, 28 },
		{ 233, 198, 28 },
		{ 233, 199, 28 },
		{ 233, 200, 28 },
		{ 233, 201, 28 },
		{ 233, 202, 28 },
		{ 233, 203, 28 },
		{ 233, 204, 28 },
		{ 233, 205, 28 },
		{ 233, 206, 28 },
		{ 233, 207, 28 },
		{ 233, 208, 28 },
		{ 233, 209, 28 },
		{ 233, 210, 28 },
		{ 233, 211, 28 },
		{ 233, 212, 28 },
		{ 233, 213, 28 },
		{ 233, 214, 28 },
		{ 233, 215, 28 },
		{ 233, 216, 28 },
		{ 233, 217, 28 },
		{ 233, 218, 28 },
		{ 233, 219, 28 },
		{ 233, 220, 28 },
		{ 233, 221, 28 },
		{ 233, 222, 28 },
		{ 233, 223, 28 },
		{ 233, 224, 28 },
		{ 234, 166, 28 },
		{ 234, 167, 28 },
		{ 234, 168, 28 },
		{ 234, 169, 28 },
		{ 234, 170, 28 },
		{ 234, 171, 28 },
		{ 234, 172, 28 },
		{ 234, 173, 28 },
		{ 234, 174, 28 },
		{ 234, 175, 28 },
		{ 234, 176, 28 },
		{ 234, 177, 28 },
		{ 234, 178, 28 },
		{ 234, 179, 28 },
		{ 234, 180, 28 },
		{ 234, 181, 28 },
		{ 234, 182, 28 },
		{ 234, 183, 28 },
		{ 234, 184, 28 },
		{ 234, 185, 28 },
		{ 234, 186, 28 },
		{ 234, 187, 28 },
		{ 234, 188, 28 },
		{ 234, 189, 28 },
		{ 234, 190, 28 },
		{ 234, 191, 28 },
		{ 234, 192, 28 },
		{ 234, 193, 28 },
		{ 234, 194, 28 },
		{ 234, 195, 28 },
		{ 234, 196, 28 },
		{ 234, 197, 28 },
		{ 234, 198, 28 },
		{ 234, 199, 28 },
		{ 234, 200, 28 },
		{ 234, 201, 28 },
		{ 234, 202, 28 },
		{ 234, 203, 28 },
		{ 234, 204, 28 },
		{ 234, 205, 28 },
		{ 234, 206, 28 },
		{ 234, 207, 28 },
		{ 234, 208, 28 },
		{ 234, 209, 28 },
		{ 234, 210, 28 },
		{ 234, 211, 28 },
		{ 234, 212, 28 },
		{ 234, 213, 28 },
		{ 234, 214, 28 },
		{ 234, 215, 28 },
		{ 234, 216, 28 },
		{ 234, 217, 28 },
		{ 234, 218, 28 },
		{ 234, 219, 28 },
		{ 234, 220, 28 },
		{ 234, 221, 28 },
		{ 234, 222, 28 },
		{ 234, 223, 28 },
		{ 234, 224, 28 },
		{ 234, 225, 28 },
		{ 234, 226, 28 },
		{ 235, 170, 28 },
		{ 235, 171, 28 },
		{ 235, 172, 28 },
		{ 235, 173, 28 },
		{ 235, 174, 28 },
		{ 235, 175, 28 },
		{ 235, 176, 28 },
		{ 235, 177, 28 },
		{ 235, 178, 28 },
		{ 235, 179, 28 },
		{ 235, 180, 28 },
		{ 235, 181, 28 },
		{ 235, 182, 28 },
		{ 235, 183, 28 },
		{ 235, 184, 28 },
		{ 235, 185, 28 },
		{ 235, 186, 28 },
		{ 235, 187, 28 },
		{ 235, 188, 28 },
		{ 235, 189, 28 },
		{ 235, 190, 28 },
		{ 235, 191, 28 },
		{ 235, 192, 28 },
		{ 235, 193, 28 },
		{ 235, 194, 28 },
		{ 235, 195, 28 },
		{ 235, 196, 28 },
		{ 235, 197, 28 },
		{ 235, 198, 28 },
		{ 235, 199, 28 },
		{ 235, 200, 28 },
		{ 235, 201, 28 },
		{ 235, 202, 28 },
		{ 235, 203, 28 },
		{ 235, 204, 28 },
		{ 235, 205, 28 },
		{ 235, 206, 28 },
		{ 235, 207, 28 },
		{ 235, 208, 28 },
		{ 235, 209, 28 },
		{ 235, 210, 28 },
		{ 235, 211, 28 },
		{ 235, 212, 28 },
		{ 235, 213, 28 },
		{ 235, 214, 28 },
		{ 235, 215, 28 },
		{ 235, 216, 28 },
		{ 235, 217, 28 },
		{ 235, 218, 28 },
		{ 235, 219, 28 },
		{ 235, 220, 28 },
		{ 235, 221, 28 },
		{ 235, 222, 28 },
		{ 235, 223, 28 },
		{ 235, 224, 28 },
		{ 235, 225, 28 },
		{ 235, 226, 28 },
		{ 235, 227, 28 },
		{ 236, 174, 28 },
		{ 236, 175, 28 },
		{ 236, 176, 28 },
		{ 236, 177, 28 },
		{ 236, 178, 28 },
		{ 236, 179, 28 },
		{ 236, 180, 28 },
		{ 236, 181, 28 },
		{ 236, 182, 28 },
		{ 236, 183, 28 },
		{ 236, 184, 28 },
		{ 236, 185, 28 },
		{ 236, 186, 28 },
		{ 236, 187, 28 },
		{ 236, 188, 28 },
		{ 236, 189, 28 },
		{ 236, 190, 28 },
		{ 236, 191, 28 },
		{ 236, 192, 28 },
		{ 236, 193, 28 },
		{ 236, 194, 28 },
		{ 236, 195, 28 },
		{ 236, 196, 28 },
		{ 236, 197, 28 },
		{ 236, 198, 28 },
		{ 236, 199, 28 },
		{ 236, 200, 28 },
		{ 236, 201, 28 },
		{ 236, 202, 28 },
		{ 236, 203, 28 },
		{ 236, 204, 28 },
		{ 236, 205, 28 },
		{ 236, 206, 28 },
		{ 236, 207, 28 },
		{ 236, 208, 28 },
		{ 236, 209, 28 },
		{ 236, 210, 28 },
		{ 236, 211, 28 },
		{ 236, 212, 28 },
		{ 236, 213, 28 },
		{ 236, 214, 28 },
		{ 236, 215, 28 },
		{ 236, 216, 28 },
		{ 236, 217, 28 },
		{ 236, 218, 28 },
		{ 236, 219, 28 },
		{ 236, 220, 28 },
		{ 236, 221, 28 },
		{ 236, 222, 28 },
		{ 236, 223, 28 },
		{ 236, 224, 28 },
		{ 236, 225, 28 },
		{ 236, 226, 28 },
		{ 236, 227, 28 },
		{ 236, 228, 28 },
		{ 237, 178, 28 },
		{ 237, 179, 28 },
		{ 237, 180, 28 },
		{ 237, 181, 28 },
		{ 237, 182, 28 },
		{ 237, 183, 28 },
		{ 237, 184, 28 },
		{ 237, 185, 28 },
		{ 237, 186, 28 },
		{ 237, 187, 28 },
		{ 237, 188, 28 },
		{ 237, 189, 28 },
		{ 237, 190, 28 },
		{ 237, 191, 28 },
		{ 237, 192, 28 },
		{ 237, 193, 28 },
		{ 237, 194, 28 },
		{ 237, 195, 28 },
		{ 237, 196, 28 },
		{ 237, 197, 28 },
		{ 237, 198, 28 },
		{ 237, 199, 28 },
		{ 237, 200, 28 },
		{ 237, 201, 28 },
		{ 237, 202, 28 },
		{ 237, 203, 28 },
		{ 237, 204, 28 },
		{ 237, 205, 28 },
		{ 237, 206, 28 },
		{ 237, 207, 28 },
		{ 237, 208, 28 },
		{ 237, 209, 28 },
		{ 237, 210, 28 },
		{ 237, 211, 28 },
		{ 237, 212, 28 },
		{ 237, 213, 28 },
		{ 237, 214, 28 },
		{ 237, 215, 28 },
		{ 237, 216, 28 },
		{ 237, 217, 28 },
		{ 237, 218, 28 },
		{ 237, 219, 28 },
		{ 237, 220, 28 },
		{ 237, 221, 28 },
		{ 237, 222, 28 },
		{ 237, 223, 28 },
		{ 237, 224, 28 },
		{ 237, 225, 28 },
		{ 237, 226, 28 },
		{ 237, 227, 28 },
		{ 237, 228, 28 },
		{ 237, 229, 28 },
		{ 237, 230, 28 },
		{ 238, 182, 28 },
		{ 238, 183, 28 },
		{ 238, 184, 28 },
		{ 238, 185, 28 },
		{ 238, 186, 28 },
		{ 238, 187, 28 },
		{ 238, 188, 28 },
		{ 238, 189, 28 },
		{ 238, 190, 28 },
		{ 238, 191, 28 },
		{ 238, 192, 28 },
		{ 238, 193, 28 },
		{ 238, 194, 28 },
		{ 238, 195, 28 },
		{ 238, 196, 28 },
		{ 238, 197, 28 },
		{ 238, 198, 28 },
		{ 238, 199, 28 },
		{ 238, 200, 28 },
		{ 238, 201, 28 },
		{ 238, 202, 28 },
		{ 238, 203, 28 },
		{ 238, 204, 28 },
		{ 238, 205, 28 },
		{ 238, 206, 28 },
		{ 238, 207, 28 },
		{ 238, 208, 28 },
		{ 238, 209, 28 },
		{ 238, 210, 28 },
		{ 238, 211, 28 },
		{ 238, 212, 28 },
		{ 238, 213, 28 },
		{ 238, 214, 28 },
		{ 238, 215, 28 },
		{ 238, 216, 28 },
		{ 238, 217, 28 },
		{ 238, 218, 28 },
		{ 238, 219, 28 },
		{ 238, 220, 28 },
		{ 238, 221, 28 },
		{ 238, 222, 28 },
		{ 238, 223, 28 },
		{ 238, 224, 28 },
		{ 238, 225, 28 },
		{ 238, 226, 28 },
		{ 238, 227, 28 },
		{ 238, 228, 28 },
		{ 238, 229, 28 },
		{ 238, 230, 28 },
		{ 238, 231, 28 },
		{ 239, 186, 28 },
		{ 239, 187, 28 },
		{ 239, 188, 28 },
		{ 239, 189, 28 },
		{ 239, 190, 28 },
		{ 239, 191, 28 },
		{ 239, 192, 28 },
		{ 239, 193, 28 },
		{ 239, 194, 28 },
		{ 239, 195, 28 },
		{ 239, 196, 28 },
		{ 239, 197, 28 },
		{ 239, 198, 28 },
		{ 239, 199, 28 },
		{ 239, 200, 28 },
		{ 239, 201, 28 },
		{ 239, 202, 28 },
		{ 239, 203, 28 },
		{ 239, 204, 28 },
		{ 239, 205, 28 },
		{ 239, 206, 28 },
		{ 239, 207, 28 },
		{ 239, 208, 28 },
		{ 239, 209, 28 },
		{ 239, 210, 28 },
		{ 239, 211, 28 },
		{ 239, 212, 28 },
		{ 239, 213, 28 },
		{ 239, 214, 28 },
		{ 239, 215, 28 },
		{ 239, 216, 28 },
		{ 239, 217, 28 },
		{ 239, 218, 28 },
		{ 239, 219, 28 },
		{ 239, 220, 28 },
		{ 239, 221, 28 },
		{ 239, 222, 28 },
		{ 239, 223, 28 },
		{ 239, 224, 28 },
		{ 239, 225, 28 },
		{ 239, 226, 28 },
		{ 239, 227, 28 },
		{ 239, 228, 28 },
		{ 239, 229, 28 },
		{ 239, 230, 28 },
		{ 239, 231, 28 },
		{ 239, 232, 28 },
		{ 240, 190, 28 },
		{ 240, 191, 28 },
		{ 240, 192, 28 },
		{ 240, 193, 28 },
		{ 240, 194, 28 },
		{ 240, 195, 28 },
		{ 240, 196, 28 },
		{ 240, 197, 28 },
		{ 240, 198, 28 },
		{ 240, 199, 28 },
		{ 240, 200, 28 },
		{ 240, 201, 28 },
		{ 240, 202, 28 },
		{ 240, 203, 28 },
		{ 240, 204, 28 },
		{ 240, 205, 28 },
		{ 240, 206, 28 },
		{ 240, 207, 28 },
		{ 240, 208, 28 },
		{ 240, 209, 28 },
		{ 240, 210, 28 },
		{ 240, 211, 28 },
		{ 240, 212, 28 },
		{ 240, 213, 28 },
		{ 240, 214, 28 },
		{ 240, 215, 28 },
		{ 240, 216, 28 },
		{ 240, 217, 28 },
		{ 240, 218, 28 },
		{ 240, 219, 28 },
		{ 240, 220, 28 },
		{ 240, 221, 28 },
		{ 240, 222, 28 },
		{ 240, 223, 28 },
		{ 240, 224, 28 },
		{ 240, 225, 28 },
		{ 240, 226, 28 },
		{ 240, 227, 28 },
		{ 240, 228, 28 },
		{ 240, 229, 28 },
		{ 240, 230, 28 },
		{ 240, 231, 28 },
		{ 240, 232, 28 },
		{ 240, 233, 28 },
		{ 240, 234, 28 },
		{ 241, 194, 28 },
		{ 241, 195, 28 },
		{ 241, 196, 28 },
		{ 241, 197, 28 },
		{ 241, 198, 28 },
		{ 241, 199, 28 },
		{ 241, 200, 28 },
		{ 241, 201, 28 },
		{ 241, 202, 28 },
		{ 241, 203, 28 },
		{ 241, 204, 28 },
		{ 241, 205, 28 },
		{ 241, 206, 28 },
		{ 241, 207, 28 },
		{ 241, 208, 28 },
		{ 241, 209, 28 },
		{ 241, 210, 28 },
		{ 241, 211, 28 },
		{ 241, 212, 28 },
		{ 241, 213, 28 },
		{ 241, 214, 28 },
		{ 241, 215, 28 },
		{ 241, 216, 28 },
		{ 241, 217, 28 },
		{ 241, 218, 28 },
		{ 241, 219, 28 },
		{ 241, 220, 28 },
		{ 241, 221, 28 },
		{ 241, 222, 28 },
		{ 241, 223, 28 },
		{ 241, 224, 28 },
		{ 241, 225, 28 },
		{ 241, 226, 28 },
		{ 241, 227, 28 },
		{ 241, 228, 28 },
		{ 241, 229, 28 },
		{ 241, 230, 28 },
		{ 241, 231, 28 },
		{ 241, 232, 28 },
		{ 241, 233, 28 },
		{ 241, 234, 28 },
		{ 241, 235, 28 },
		{ 242, 198, 28 },
		{ 242, 199, 28 },
		{ 242, 200, 28 },
		{ 242, 201, 28 },
		{ 242, 202, 28 },
		{ 242, 203, 28 },
		{ 242, 204, 28 },
		{ 242, 205, 28 },
		{ 242, 206, 28 },
		{ 242, 207, 28 },
		{ 242, 208, 28 },
		{ 242, 209, 28 },
		{ 242, 210, 28 },
		{ 242, 211, 28 },
		{ 242, 212, 28 },
		{ 242, 213, 28 },
		{ 242, 214, 28 },
		{ 242, 215, 28 },
		{ 242, 216, 28 },
		{ 242, 217, 28 },
		{ 242, 218, 28 },
		{ 242, 219, 28 },
		{ 242, 220, 28 },
		{ 242, 221, 28 },
		{ 242, 222, 28 },
		{ 242, 223, 28 },
		{ 242, 224, 28 },
		{ 242, 225, 28 },
		{ 242, 226, 28 },
		{ 242, 227, 28 },
		{ 242, 228, 28 },
		{ 242, 229, 28 },
		{ 242, 230, 28 },
		{ 242, 231, 28 },
		{ 242, 232, 28 },
		{ 242, 233, 28 },
		{ 242, 234, 28 },
		{ 242, 235, 28 },
		{ 242, 236, 28 },
		{ 243, 202, 28 },
		{ 243, 203, 28 },
		{ 243, 204, 28 },
		{ 243, 205, 28 },
		{ 243, 206, 28 },
		{ 243, 207, 28 },
		{ 243, 208, 28 },
		{ 243, 209, 28 },
		{ 243, 210, 28 },
		{ 243, 211, 28 },
		{ 243, 212, 28 },
		{ 243, 213, 28 },
		{ 243, 214, 28 },
		{ 243, 215, 28 },
		{ 243, 216, 28 },
		{ 243, 217, 28 },
		{ 243, 218, 28 },
		{ 243, 219, 28 },
		{ 243, 220, 28 },
		{ 243, 221, 28 },
		{ 243, 222, 28 },
		{ 243, 223, 28 },
		{ 243, 224, 28 },
		{ 243, 225, 28 },
		{ 243, 226, 28 },
		{ 243, 227, 28 },
		{ 243, 228, 28 },
		{ 243, 229, 28 },
		{ 243, 230, 28 },
		{ 243, 231, 28 },
		{ 243, 232, 28 },
		{ 243, 233, 28 },
		{ 243, 234, 28 },
		{ 243, 235, 28 },
		{ 243, 236, 28 },
		{ 243, 237, 28 },
		{ 243, 238, 28 },
		{ 244, 206, 28 },
		{ 244, 207, 28 },
		{ 244, 208, 28 },
		{ 244, 209, 28 },
		{ 244, 210, 28 },
		{ 244, 211, 28 },
		{ 244, 212, 28 },
		{ 244, 213, 28 },
		{ 244, 214, 28 },
		{ 244, 215, 28 },
		{ 244, 216, 28 },
		{ 244, 217, 28 },
		{ 244, 218, 28 },
		{ 244, 219, 28 },
		{ 244, 220, 28 },
		{ 244, 221, 28 },
		{ 244, 222, 28 },
		{ 244, 223, 28 },
		{ 244, 224, 28 },
		{ 244, 225, 28 },
		{ 244, 226, 28 },
		{ 244, 227, 28 },
		{ 244, 228, 28 },
		{ 244, 229, 28 },
		{ 244, 230, 28 },
		{ 244, 231, 28 },
		{ 244, 232, 28 },
		{ 244, 233, 28 },
		{ 244, 234, 28 },
		{ 244, 235, 28 },
		{ 244, 236, 28 },
		{ 244, 237, 28 },
		{ 244, 238, 28 },
		{ 244, 239, 28 },
		{ 245, 210, 28 },
		{ 245, 211, 28 },
		{ 245, 212, 28 },
		{ 245, 213, 28 },
		{ 245, 214, 28 },
		{ 245, 215, 28 },
		{ 245, 216, 28 },
		{ 245, 217, 28 },
		{ 245, 218, 28 },
		{ 245, 219, 28 },
		{ 245, 220, 28 },
		{ 245, 221, 28 },
		{ 245, 222, 28 },
		{ 245, 223, 28 },
		{ 245, 224, 28 },
		{ 245, 225, 28 },
		{ 245, 226, 28 },
		{ 245, 227, 28 },
		{ 245, 228, 28 },
		{ 245, 229, 28 },
		{ 245, 230, 28 },
		{ 245, 231, 28 },
		{ 245, 232, 28 },
		{ 245, 233, 28 },
		{ 245, 234, 28 },
		{ 245, 235, 28 },
		{ 245, 236, 28 },
		{ 245, 237, 28 },
		{ 245, 238, 28 },
		{ 245, 239, 28 },
		{ 245, 240, 28 },
		{ 246, 214, 28 },
		{ 246, 215, 28 },
		{ 246, 216, 28 },
		{ 246, 217, 28 },
		{ 246, 218, 28 },
		{ 246, 219, 28 },
		{ 246, 220, 28 },
		{ 246, 221, 28 },
		{ 246, 222, 28 },
		{ 246, 223, 28 },
		{ 246, 224, 28 },
		{ 246, 225, 28 },
		{ 246, 226, 28 },
		{ 246, 227, 28 },
		{ 246, 228, 28 },
		{ 246, 229, 28 },
		{ 246, 230, 28 },
		{ 246, 231, 28 },
		{ 246, 232, 28 },
		{ 246, 233, 28 },
		{ 246, 234, 28 },
		{ 246, 235, 28 },
		{ 246, 236, 28 },
		{ 246, 237, 28 },
		{ 246, 238, 28 },
		{ 246, 239, 28 },
		{ 246, 240, 28 },
		{ 247, 218, 28 },
		{ 247, 219, 28 },
		{ 247, 220, 28 },
		{ 247, 221, 28 },
		{ 247, 222, 28 },
		{ 247, 223, 28 },
		{ 247, 224, 28 },
		{ 247, 225, 28 },
		{ 247, 226, 28 },
		{ 247, 227, 28 },
		{ 247, 228, 28 },
		{ 247, 229, 28 },
		{ 247, 230, 28 },
		{ 247, 231, 28 },
		{ 247, 232, 28 },
		{ 247, 233, 28 },
		{ 247, 234, 28 },
		{ 247, 235, 28 },
		{ 247, 236, 28 },
		{ 247, 237, 28 },
		{ 247, 238, 28 },
		{ 247, 239, 28 },
		{ 247, 240, 28 },
		{ 248, 222, 28 },
		{ 248, 223, 28 },
		{ 248, 224, 28 },
		{ 248, 225, 28 },
		{ 248, 226, 28 },
		{ 248, 227, 28 },
		{ 248, 228, 28 },
		{ 248, 229, 28 },
		{ 248, 230, 28 },
		{ 248, 231, 28 },
		{ 248, 232, 28 },
		{ 248, 233, 28 },
		{ 248, 234, 28 },
		{ 248, 235, 28 },
		{ 248, 236, 28 },
		{ 248, 237, 28 },
		{ 248, 238, 28 },
		{ 248, 239, 28 },
		{ 248, 240, 28 },
		{ 249, 226, 28 },
		{ 249, 227, 28 },
		{ 249, 228, 28 },
		{ 249, 229, 28 },
		{ 249, 230, 28 },
		{ 249, 231, 28 },
		{ 249, 232, 28 },
		{ 249, 233, 28 },
		{ 249, 234, 28 },
		{ 249, 235, 28 },
		{ 249, 236, 28 },
		{ 249, 237, 28 },
		{ 249, 238, 28 },
		{ 249, 239, 28 },
		{ 249, 240, 28 },
		{ 250, 230, 28 },
		{ 250, 231, 28 },
		{ 250, 232, 28 },
		{ 250, 233, 28 },
		{ 250, 234, 28 },
		{ 250, 235, 28 },
		{ 250, 236, 28 },
		{ 250, 237, 28 },
		{ 250, 238, 28 },
		{ 250, 239, 28 },
		{ 250, 240, 28 },
		{ 251, 234, 28 },
		{ 251, 235, 28 },
		{ 251, 236, 28 },
		{ 251, 237, 28 },
		{ 251, 238, 28 },
		{ 251, 239, 28 },
		{ 251, 240, 28 },
		{ 252, 238, 28 },
		{ 252, 239, 28 },
		{ 252, 240, 28 },
		{ 262, 274, 28 },
		{ 262, 275, 28 },
		{ 262, 276, 28 },
		{ 263, 274, 28 },
		{ 263, 275, 28 },
		{ 263, 276, 28 },
		{ 263, 277, 28 },
		{ 263, 278, 28 },
		{ 263, 279, 28 },
		{ 263, 280, 28 },
		{ 264, 274, 28 },
		{ 264, 275, 28 },
		{ 264, 276, 28 },
		{ 264, 277, 28 },
		{ 264, 278, 28 },
		{ 264, 279, 28 },
		{ 264, 280, 28 },
		{ 264, 281, 28 },
		{ 264, 282, 28 },
		{ 264, 283, 28 },
		{ 264, 284, 28 },
		{ 265, 274, 28 },
		{ 265, 275, 28 },
		{ 265, 276, 28 },
		{ 265, 277, 28 },
		{ 265, 278, 28 },
		{ 265, 279, 28 },
		{ 265, 280, 28 },
		{ 265, 281, 28 },
		{ 265, 282, 28 },
		{ 265, 283, 28 },
		{ 265, 284, 28 },
		{ 265, 285, 28 },
		{ 265, 286, 28 },
		{ 265, 287, 28 },
		{ 265, 288, 28 },
		{ 266, 274, 28 },
		{ 266, 275, 28 },
		{ 266, 276, 28 },
		{ 266, 277, 28 },
		{ 266, 278, 28 },
		{ 266, 279, 28 },
		{ 266, 280, 28 },
		{ 266, 281, 28 },
		{ 266, 282, 28 },
		{ 266, 283, 28 },
		{ 266, 284, 28 },
		{ 266, 285, 28 },
		{ 266, 286, 28 },
		{ 266, 287, 28 },
		{ 266, 288, 28 },
		{ 266, 289, 28 },
		{ 266, 290, 28 },
		{ 266, 291, 28 },
		{ 266, 292, 28 },
		{ 267, 274, 28 },
		{ 267, 275, 28 },
		{ 267, 276, 28 },
		{ 267, 277, 28 },
		{ 267, 278, 28 },
		{ 267, 279, 28 },
		{ 267, 280, 28 },
		{ 267, 281, 28 },
		{ 267, 282, 28 },
		{ 267, 283, 28 },
		{ 267, 284, 28 },
		{ 267, 285, 28 },
		{ 267, 286, 28 },
		{ 267, 287, 28 },
		{ 267, 288, 28 },
		{ 267, 289, 28 },
		{ 267, 290, 28 },
		{ 267, 291, 28 },
		{ 267, 292, 28 },
		{ 267, 293, 28 },
		{ 267, 294, 28 },
		{ 267, 295, 28 },
		{ 267, 296, 28 },
		{ 268, 274, 28 },
		{ 268, 275, 28 },
		{ 268, 276, 28 },
		{ 268, 277, 28 },
		{ 268, 278, 28 },
		{ 268, 279, 28 },
		{ 268, 280, 28 },
		{ 268, 281, 28 },
		{ 268, 282, 28 },
		{ 268, 283, 28 },
		{ 268, 284, 28 },
		{ 268, 285, 28 },
		{ 268, 286, 28 },
		{ 268, 287, 28 },
		{ 268, 288, 28 },
		{ 268, 289, 28 },
		{ 268, 290, 28 },
		{ 268, 291, 28 },
		{ 268, 292, 28 },
		{ 268, 293, 28 },
		{ 268, 294, 28 },
		{ 268, 295, 28 },
		{ 268, 296, 28 },
		{ 268, 297, 28 },
		{ 268, 298, 28 },
		{ 268, 299, 28 },
		{ 268, 300, 28 },
		{ 269, 274, 28 },
		{ 269, 275, 28 },
		{ 269, 276, 28 },
		{ 269, 277, 28 },
		{ 269, 278, 28 },
		{ 269, 279, 28 },
		{ 269, 280, 28 },
		{ 269, 281, 28 },
		{ 269, 282, 28 },
		{ 269, 283, 28 },
		{ 269, 284, 28 },
		{ 269, 285, 28 },
		{ 269, 286, 28 },
		{ 269, 287, 28 },
		{ 269, 288, 28 },
		{ 269, 289, 28 },
		{ 269, 290, 28 },
		{ 269, 291, 28 },
		{ 269, 292, 28 },
		{ 269, 293, 28 },
		{ 269, 294, 28 },
		{ 269, 295, 28 },
		{ 269, 296, 28 },
		{ 269, 297, 28 },
		{ 269, 298, 28 },
		{ 269, 299, 28 },
		{ 269, 300, 28 },
		{ 269, 301, 28 },
		{ 269, 302, 28 },
		{ 269, 303, 28 },
		{ 269, 304, 28 },
		{ 270, 275, 28 },
		{ 270, 276, 28 },
		{ 270, 277, 28 },
		{ 270, 278, 28 },
		{ 270, 279, 28 },
		{ 270, 280, 28 },
		{ 270, 281, 28 },
		{ 270, 282, 28 },
		{ 270, 283, 28 },
		{ 270, 284, 28 },
		{ 270, 285, 28 },
		{ 270, 286, 28 },
		{ 270, 287, 28 },
		{ 270, 288, 28 },
		{ 270, 289, 28 },
		{ 270, 290, 28 },
		{ 270, 291, 28 },
		{ 270, 292, 28 },
		{ 270, 293, 28 },
		{ 270, 294, 28 },
		{ 270, 295, 28 },
		{ 270, 296, 28 },
		{ 270, 297, 28 },
		{ 270, 298, 28 },
		{ 270, 299, 28 },
		{ 270, 300, 28 },
		{ 270, 301, 28 },
		{ 270, 302, 28 },
		{ 270, 303, 28 },
		{ 270, 304, 28 },
		{ 270, 305, 28 },
		{ 270, 306, 28 },
		{ 270, 307, 28 },
		{ 270, 308, 28 },
		{ 271, 276, 28 },
		{ 271, 277, 28 },
		{ 271, 278, 28 },
		{ 271, 279, 28 },
		{ 271, 280, 28 },
		{ 271, 281, 28 },
		{ 271, 282, 28 },
		{ 271, 283, 28 },
		{ 271, 284, 28 },
		{ 271, 285, 28 },
		{ 271, 286, 28 },
		{ 271, 287, 28 },
		{ 271, 288, 28 },
		{ 271, 289, 28 },
		{ 271, 290, 28 },
		{ 271, 291, 28 },
		{ 271, 292, 28 },
		{ 271, 293, 28 },
		{ 271, 294, 28 },
		{ 271, 295, 28 },
		{ 271, 296, 28 },
		{ 271, 297, 28 },
		{ 271, 298, 28 },
		{ 271, 299, 28 },
		{ 271, 300, 28 },
		{ 271, 301, 28 },
		{ 271, 302, 28 },
		{ 271, 303, 28 },
		{ 271, 304, 28 },
		{ 271, 305, 28 },
		{ 271, 306, 28 },
		{ 271, 307, 28 },
		{ 271, 308, 28 },
		{ 271, 309, 28 },
		{ 271, 310, 28 },
		{ 271, 311, 28 },
		{ 271, 312, 28 },
		{ 272, 278, 28 },
		{ 272, 279, 28 },
		{ 272, 280, 28 },
		{ 272, 281, 28 },
		{ 272, 282, 28 },
		{ 272, 283, 28 },
		{ 272, 284, 28 },
		{ 272, 285, 28 },
		{ 272, 286, 28 },
		{ 272, 287, 28 },
		{ 272, 288, 28 },
		{ 272, 289, 28 },
		{ 272, 290, 28 },
		{ 272, 291, 28 },
		{ 272, 292, 28 },
		{ 272, 293, 28 },
		{ 272, 294, 28 },
		{ 272, 295, 28 },
		{ 272, 296, 28 },
		{ 272, 297, 28 },
		{ 272, 298, 28 },
		{ 272, 299, 28 },
		{ 272, 300, 28 },
		{ 272, 301, 28 },
		{ 272, 302, 28 },
		{ 272, 303, 28 },
		{ 272, 304, 28 },
		{ 272, 305, 28 },
		{ 272, 306, 28 },
		{ 272, 307, 28 },
		{ 272, 308, 28 },
		{ 272, 309, 28 },
		{ 272, 310, 28 },
		{ 272, 311, 28 },
		{ 272, 312, 28 },
		{ 272, 313, 28 },
		{ 272, 314, 28 },
		{ 272, 315, 28 },
		{ 272, 316, 28 },
		{ 273, 279, 28 },
		{ 273, 280, 28 },
		{ 273, 281, 28 },
		{ 273, 282, 28 },
		{ 273, 283, 28 },
		{ 273, 284, 28 },
		{ 273, 285, 28 },
		{ 273, 286, 28 },
		{ 273, 287, 28 },
		{ 273, 288, 28 },
		{ 273, 289, 28 },
		{ 273, 290, 28 },
		{ 273, 291, 28 },
		{ 273, 292, 28 },
		{ 273, 293, 28 },
		{ 273, 294, 28 },
		{ 273, 295, 28 },
		{ 273, 296, 28 },
		{ 273, 297, 28 },
		{ 273, 298, 28 },
		{ 273, 299, 28 },
		{ 273, 300, 28 },
		{ 273, 301, 28 },
		{ 273, 302, 28 },
		{ 273, 303, 28 },
		{ 273, 304, 28 },
		{ 273, 305, 28 },
		{ 273, 306, 28 },
		{ 273, 307, 28 },
		{ 273, 308, 28 },
		{ 273, 309, 28 },
		{ 273, 310, 28 },
		{ 273, 311, 28 },
		{ 273, 312, 28 },
		{ 273, 313, 28 },
		{ 273, 314, 28 },
		{ 273, 315, 28 },
		{ 273, 316, 28 },
		{ 273, 317, 28 },
		{ 273, 318, 28 },
		{ 273, 319, 28 },
		{ 273, 320, 28 },
		{ 274, 280, 28 },
		{ 274, 281, 28 },
		{ 274, 282, 28 },
		{ 274, 283, 28 },
		{ 274, 284, 28 },
		{ 274, 285, 28 },
		{ 274, 286, 28 },
		{ 274, 287, 28 },
		{ 274, 288, 28 },
		{ 274, 289, 28 },
		{ 274, 290, 28 },
		{ 274, 291, 28 },
		{ 274, 292, 28 },
		{ 274, 293, 28 },
		{ 274, 294, 28 },
		{ 274, 295, 28 },
		{ 274, 296, 28 },
		{ 274, 297, 28 },
		{ 274, 298, 28 },
		{ 274, 299, 28 },
		{ 274, 300, 28 },
		{ 274, 301, 28 },
		{ 274, 302, 28 },
		{ 274, 303, 28 },
		{ 274, 304, 28 },
		{ 274, 305, 28 },
		{ 274, 306, 28 },
		{ 274, 307, 28 },
		{ 274, 308, 28 },
		{ 274, 309, 28 },
		{ 274, 310, 28 },
		{ 274, 311, 28 },
		{ 274, 312, 28 },
		{ 274, 313, 28 },
		{ 274, 314, 28 },
		{ 274, 315, 28 },
		{ 274, 316, 28 },
		{ 274, 317, 28 },
		{ 274, 318, 28 },
		{ 274, 319, 28 },
		{ 274, 320, 28 },
		{ 274, 321, 28 },
		{ 274, 322, 28 },
		{ 274, 323, 28 },
		{ 274, 324, 28 },
		{ 275, 282, 28 },
		{ 275, 283, 28 },
		{ 275, 284, 28 },
		{ 275, 285, 28 },
		{ 275, 286, 28 },
		{ 275, 287, 28 },
		{ 275, 288, 28 },
		{ 275, 289, 28 },
		{ 275, 290, 28 },
		{ 275, 291, 28 },
		{ 275, 292, 28 },
		{ 275, 293, 28 },
		{ 275, 294, 28 },
		{ 275, 295, 28 },
		{ 275, 296, 28 },
		{ 275, 297, 28 },
		{ 275, 298, 28 },
		{ 275, 299, 28 },
		{ 275, 300, 28 },
		{ 275, 301, 28 },
		{ 275, 302, 28 },
		{ 275, 303, 28 },
		{ 275, 304, 28 },
		{ 275, 305, 28 },
		{ 275, 306, 28 },
		{ 275, 307, 28 },
		{ 275, 308, 28 },
		{ 275, 309, 28 },
		{ 275, 310, 28 },
		{ 275, 311, 28 },
		{ 275, 312, 28 },
		{ 275, 313, 28 },
		{ 275, 314, 28 },
		{ 275, 315, 28 },
		{ 275, 316, 28 },
		{ 275, 317, 28 },
		{ 275, 318, 28 },
		{ 275, 319, 28 },
		{ 275, 320, 28 },
		{ 275, 321, 28 },
		{ 275, 322, 28 },
		{ 275, 323, 28 },
		{ 275, 324, 28 },
		{ 275, 325, 28 },
		{ 275, 326, 28 },
		{ 275, 327, 28 },
		{ 275, 328, 28 },
		{ 276, 283, 28 },
		{ 276, 284, 28 },
		{ 276, 285, 28 },
		{ 276, 286, 28 },
		{ 276, 287, 28 },
		{ 276, 288, 28 },
		{ 276, 289, 28 },
		{ 276, 290, 28 },
		{ 276, 291, 28 },
		{ 276, 292, 28 },
		{ 276, 293, 28 },
		{ 276, 294, 28 },
		{ 276, 295, 28 },
		{ 276, 296, 28 },
		{ 276, 297, 28 },
		{ 276, 298, 28 },
		{ 276, 299, 28 },
		{ 276, 300, 28 },
		{ 276, 301, 28 },
		{ 276, 302, 28 },
		{ 276, 303, 28 },
		{ 276, 304, 28 },
		{ 276, 305, 28 },
		{ 276, 306, 28 },
		{ 276, 307, 28 },
		{ 276, 308, 28 },
		{ 276, 309, 28 },
		{ 276, 310, 28 },
		{ 276, 311, 28 },
		{ 276, 312, 28 },
		{ 276, 313, 28 },
		{ 276, 314, 28 },
		{ 276, 315, 28 },
		{ 276, 316, 28 },
		{ 276, 317, 28 },
		{ 276, 318, 28 },
		{ 276, 319, 28 },
		{ 276, 320, 28 },
		{ 276, 321, 28 },
		{ 276, 322, 28 },
		{ 276, 323, 28 },
		{ 276, 324, 28 },
		{ 276, 325, 28 },
		{ 276, 326, 28 },
		{ 276, 327, 28 },
		{ 276, 328, 28 },
		{ 276, 329, 28 },
		{ 276, 330, 28 },
		{ 276, 331, 28 },
		{ 276, 332, 28 },
		{ 277, 284, 28 },
		{ 277, 285, 28 },
		{ 277, 286, 28 },
		{ 277, 287, 28 },
		{ 277, 288, 28 },
		{ 277, 289, 28 },
		{ 277, 290, 28 },
		{ 277, 291, 28 },
		{ 277, 292, 28 },
		{ 277, 293, 28 },
		{ 277, 294, 28 },
		{ 277, 295, 28 },
		{ 277, 296, 28 },
		{ 277, 297, 28 },
		{ 277, 298, 28 },
		{ 277, 299, 28 },
		{ 277, 300, 28 },
		{ 277, 301, 28 },
		{ 277, 302, 28 },
		{ 277, 303, 28 },
		{ 277, 304, 28 },
		{ 277, 305, 28 },
		{ 277, 306, 28 },
		{ 277, 307, 28 },
		{ 277, 308, 28 },
		{ 277, 309, 28 },
		{ 277, 310, 28 },
		{ 277, 311, 28 },
		{ 277, 312, 28 },
		{ 277, 313, 28 },
		{ 277, 314, 28 },
		{ 277, 315, 28 },
		{ 277, 316, 28 },
		{ 277, 317, 28 },
		{ 277, 318, 28 },
		{ 277, 319, 28 },
		{ 277, 320, 28 },
		{ 277, 321, 28 },
		{ 277, 322, 28 },
		{ 277, 323, 28 },
		{ 277, 324, 28 },
		{ 277, 325, 28 },
		{ 277, 326, 28 },
		{ 277, 327, 28 },
		{ 277, 328, 28 },
		{ 277, 329, 28 },
		{ 277, 330, 28 },
		{ 277, 331, 28 },
		{ 277, 332, 28 },
		{ 277, 333, 28 },
		{ 277, 334, 28 },
		{ 277, 335, 28 },
		{ 277, 336, 28 },
		{ 278, 286, 28 },
		{ 278, 287, 28 },
		{ 278, 288, 28 },
		{ 278, 289, 28 },
		{ 278, 290, 28 },
		{ 278, 291, 28 },
		{ 278, 292, 28 },
		{ 278, 293, 28 },
		{ 278, 294, 28 },
		{ 278, 295, 28 },
		{ 278, 296, 28 },
		{ 278, 297, 28 },
		{ 278, 298, 28 },
		{ 278, 299, 28 },
		{ 278, 300, 28 },
		{ 278, 301, 28 },
		{ 278, 302, 28 },
		{ 278, 303, 28 },
		{ 278, 304, 28 },
		{ 278, 305, 28 },
		{ 278, 306, 28 },
		{ 278, 307, 28 },
		{ 278, 308, 28 },
		{ 278, 309, 28 },
		{ 278, 310, 28 },
		{ 278, 311, 28 },
		{ 278, 312, 28 },
		{ 278, 313, 28 },
		{ 278, 314, 28 },
		{ 278, 315, 28 },
		{ 278, 316, 28 },
		{ 278, 317, 28 },
		{ 278, 318, 28 },
		{ 278, 319, 28 },
		{ 278, 320, 28 },
		{ 278, 321, 28 },
		{ 278, 322, 28 },
		{ 278, 323, 28 },
		{ 278, 324, 28 },
		{ 278, 325, 28 },
		{ 278, 326, 28 },
		{ 278, 327, 28 },
		{ 278, 328, 28 },
		{ 278, 329, 28 },
		{ 278, 330, 28 },
		{ 278, 331, 28 },
		{ 278, 332, 28 },
		{ 278, 333, 28 },
		{ 278, 334, 28 },
		{ 278, 335, 28 },
		{ 278, 336, 28 },
		{ 278, 337, 28 },
		{ 278, 338, 28 },
		{ 278, 339, 28 },
		{ 278, 340, 28 },
		{ 279, 287, 28 },
		{ 279, 288, 28 },
		{ 279, 289, 28 },
		{ 279, 290, 28 },
		{ 279, 291, 28 },
		{ 279, 292, 28 },
		{ 279, 293, 28 },
		{ 279, 294, 28 },
		{ 279, 295, 28 },
		{ 279, 296, 28 },
		{ 279, 297, 28 },
		{ 279, 298, 28 },
		{ 279, 299, 28 },
		{ 279, 300, 28 },
		{ 279, 301, 28 },
		{ 279, 302, 28 },
		{ 279, 303, 28 },
		{ 279, 304, 28 },
		{ 279, 305, 28 },
		{ 279, 306, 28 },
		{ 279, 307, 28 },
		{ 279, 308, 28 },
		{ 279, 309, 28 },
		{ 279, 310, 28 },
		{ 279, 311, 28 },
		{ 279, 312, 28 },
		{ 279, 313, 28 },
		{ 279, 314, 28 },
		{ 279, 315, 28 },
		{ 279, 316, 28 },
		{ 279, 317, 28 },
		{ 279, 318, 28 },
		{ 279, 319, 28 },
		{ 279, 320, 28 },
		{ 279, 321, 28 },
		{ 279, 322, 28 },
		{ 279, 323, 28 },
		{ 279, 324, 28 },
		{ 279, 325, 28 },
		{ 279, 326, 28 },
		{ 279, 327, 28 },
		{ 279, 328, 28 },
		{ 279, 329, 28 },
		{ 279, 330, 28 },
		{ 279, 331, 28 },
		{ 279, 332, 28 },
		{ 279, 333, 28 },
		{ 279, 334, 28 },
		{ 279, 335, 28 },
		{ 279, 336, 28 },
		{ 279, 337, 28 },
		{ 279, 338, 28 },
		{ 279, 339, 28 },
		{ 279, 340, 28 },
		{ 279, 341, 28 },
		{ 279, 342, 28 },
		{ 279, 343, 28 },
		{ 279, 344, 28 },
		{ 280, 288, 28 },
		{ 280, 289, 28 },
		{ 280, 290, 28 },
		{ 280, 291, 28 },
		{ 280, 292, 28 },
		{ 280, 293, 28 },
		{ 280, 294, 28 },
		{ 280, 295, 28 },
		{ 280, 296, 28 },
		{ 280, 297, 28 },
		{ 280, 298, 28 },
		{ 280, 299, 28 },
		{ 280, 300, 28 },
		{ 280, 301, 28 },
		{ 280, 302, 28 },
		{ 280, 303, 28 },
		{ 280, 304, 28 },
		{ 280, 305, 28 },
		{ 280, 306, 28 },
		{ 280, 307, 28 },
		{ 280, 308, 28 },
		{ 280, 309, 28 },
		{ 280, 310, 28 },
		{ 280, 311, 28 },
		{ 280, 312, 28 },
		{ 280, 313, 28 },
		{ 280, 314, 28 },
		{ 280, 315, 28 },
		{ 280, 316, 28 },
		{ 280, 317, 28 },
		{ 280, 318, 28 },
		{ 280, 319, 28 },
		{ 280, 320, 28 },
		{ 280, 321, 28 },
		{ 280, 322, 28 },
		{ 280, 323, 28 },
		{ 280, 324, 28 },
		{ 280, 325, 28 },
		{ 280, 326, 28 },
		{ 280, 327, 28 },
		{ 280, 328, 28 },
		{ 280, 329, 28 },
		{ 280, 330, 28 },
		{ 280, 331, 28 },
		{ 280, 332, 28 },
		{ 280, 333, 28 },
		{ 280, 334, 28 },
		{ 280, 335, 28 },
		{ 280, 336, 28 },
		{ 280, 337, 28 },
		{ 280, 338, 28 },
		{ 280, 339, 28 },
		{ 280, 340, 28 },
		{ 280, 341, 28 },
		{ 280, 342, 28 },
		{ 280, 343, 28 },
		{ 280, 344, 28 },
		{ 280, 345, 28 },
		{ 280, 346, 28 },
		{ 280, 347, 28 },
		{ 280, 348, 28 },
		{ 281, 290, 28 },
		{ 281, 291, 28 },
		{ 281, 292, 28 },
		{ 281, 293, 28 },
		{ 281, 294, 28 },
		{ 281, 295, 28 },
		{ 281, 296, 28 },
		{ 281, 297, 28 },
		{ 281, 298, 28 },
		{ 281, 299, 28 },
		{ 281, 300, 28 },
		{ 281, 301, 28 },
		{ 281, 302, 28 },
		{ 281, 303, 28 },
		{ 281, 304, 28 },
		{ 281, 305, 28 },
		{ 281, 306, 28 },
		{ 281, 307, 28 },
		{ 281, 308, 28 },
		{ 281, 309, 28 },
		{ 281, 310, 28 },
		{ 281, 311, 28 },
		{ 281, 312, 28 },
		{ 281, 313, 28 },
		{ 281, 314, 28 },
		{ 281, 315, 28 },
		{ 281, 316, 28 },
		{ 281, 317, 28 },
		{ 281, 318, 28 },
		{ 281, 319, 28 },
		{ 281, 320, 28 },
		{ 281, 321, 28 },
		{ 281, 322, 28 },
		{ 281, 323, 28 },
		{ 281, 324, 28 },
		{ 281, 325, 28 },
		{ 281, 326, 28 },
		{ 281, 327, 28 },
		{ 281, 328, 28 },
		{ 281, 329, 28 },
		{ 281, 330, 28 },
		{ 281, 331, 28 },
		{ 281, 332, 28 },
		{ 281, 333, 28 },
		{ 281, 334, 28 },
		{ 281, 335, 28 },
		{ 281, 336, 28 },
		{ 281, 337, 28 },
		{ 281, 338, 28 },
		{ 281, 339, 28 },
		{ 281, 340, 28 },
		{ 281, 341, 28 },
		{ 281, 342, 28 },
		{ 281, 343, 28 },
		{ 281, 344, 28 },
		{ 281, 345, 28 },
		{ 281, 346, 28 },
		{ 281, 347, 28 },
		{ 281, 348, 28 },
		{ 281, 349, 28 },
		{ 281, 350, 28 },
		{ 281, 351, 28 },
		{ 281, 352, 28 },
		{ 282, 291, 28 },
		{ 282, 292, 28 },
		{ 282, 293, 28 },
		{ 282, 294, 28 },
		{ 282, 295, 28 },
		{ 282, 296, 28 },
		{ 282, 297, 28 },
		{ 282, 298, 28 },
		{ 282, 299, 28 },
		{ 282, 300, 28 },
		{ 282, 301, 28 },
		{ 282, 302, 28 },
		{ 282, 303, 28 },
		{ 282, 304, 28 },
		{ 282, 305, 28 },
		{ 282, 306, 28 },
		{ 282, 307, 28 },
		{ 282, 308, 28 },
		{ 282, 309, 28 },
		{ 282, 310, 28 },
		{ 282, 311, 28 },
		{ 282, 312, 28 },
		{ 282, 313, 28 },
		{ 282, 314, 28 },
		{ 282, 315, 28 },
		{ 282, 316, 28 },
		{ 282, 317, 28 },
		{ 282, 318, 28 },
		{ 282, 319, 28 },
		{ 282, 320, 28 },
		{ 282, 321, 28 },
		{ 282, 322, 28 },
		{ 282, 323, 28 },
		{ 282, 324, 28 },
		{ 282, 325, 28 },
		{ 282, 326, 28 },
		{ 282, 327, 28 },
		{ 282, 328, 28 },
		{ 282, 329, 28 },
		{ 282, 330, 28 },
		{ 282, 331, 28 },
		{ 282, 332, 28 },
		{ 282, 333, 28 },
		{ 282, 334, 28 },
		{ 282, 335, 28 },
		{ 282, 336, 28 },
		{ 282, 337, 28 },
		{ 282, 338, 28 },
		{ 282, 339, 28 },
		{ 282, 340, 28 },
		{ 282, 341, 28 },
		{ 282, 342, 28 },
		{ 282, 343, 28 },
		{ 282, 344, 28 },
		{ 282, 345, 28 },
		{ 282, 346, 28 },
		{ 282, 347, 28 },
		{ 282, 348, 28 },
		{ 282, 349, 28 },
		{ 282, 350, 28 },
		{ 282, 351, 28 },
		{ 282, 352, 28 },
		{ 282, 353, 28 },
		{ 282, 354, 28 },
		{ 282, 355, 28 },
		{ 282, 356, 28 },
		{ 283, 292, 28 },
		{ 283, 293, 28 },
		{ 283, 294, 28 },
		{ 283, 295, 28 },
		{ 283, 296, 28 },
		{ 283, 297, 28 },
		{ 283, 298, 28 },
		{ 283, 299, 28 },
		{ 283, 300, 28 },
		{ 283, 301, 28 },
		{ 283, 302, 28 },
		{ 283, 303, 28 },
		{ 283, 304, 28 },
		{ 283, 305, 28 },
		{ 283, 306, 28 },
		{ 283, 307, 28 },
		{ 283, 308, 28 },
		{ 283, 309, 28 },
		{ 283, 310, 28 },
		{ 283, 311, 28 },
		{ 283, 312, 28 },
		{ 283, 313, 28 },
		{ 283, 314, 28 },
		{ 283, 315, 28 },
		{ 283, 316, 28 },
		{ 283, 317, 28 },
		{ 283, 318, 28 },
		{ 283, 319, 28 },
		{ 283, 320, 28 },
		{ 283, 321, 28 },
		{ 283, 322, 28 },
		{ 283, 323, 28 },
		{ 283, 324, 28 },
		{ 283, 325, 28 },
		{ 283, 326, 28 },
		{ 283, 327, 28 },
		{ 283, 328, 28 },
		{ 283, 329, 28 },
		{ 283, 330, 28 },
		{ 283, 331, 28 },
		{ 283, 332, 28 },
		{ 283, 333, 28 },
		{ 283, 334, 28 },
		{ 283, 335, 28 },
		{ 283, 336, 28 },
		{ 283, 337, 28 },
		{ 283, 338, 28 },
		{ 283, 339, 28 },
		{ 283, 340, 28 },
		{ 283, 341, 28 },
		{ 283, 342, 28 },
		{ 283, 343, 28 },
		{ 283, 344, 28 },
		{ 283, 345, 28 },
		{ 283, 346, 28 },
		{ 283, 347, 28 },
		{ 283, 348, 28 },
		{ 283, 349, 28 },
		{ 283, 350, 28 },
		{ 283, 351, 28 },
		{ 283, 352, 28 },
		{ 283, 353, 28 },
		{ 283, 354, 28 },
		{ 283, 355, 28 },
		{ 283, 356, 28 },
		{ 283, 357, 28 },
		{ 283, 358, 28 },
		{ 283, 359, 28 },
		{ 283, 360, 28 },
		{ 284, 294, 28 },
		{ 284, 295, 28 },
		{ 284, 296, 28 },
		{ 284, 297, 28 },
		{ 284, 298, 28 },
		{ 284, 299, 28 },
		{ 284, 300, 28 },
		{ 284, 301, 28 },
		{ 284, 302, 28 },
		{ 284, 303, 28 },
		{ 284, 304, 28 },
		{ 284, 305, 28 },
		{ 284, 306, 28 },
		{ 284, 307, 28 },
		{ 284, 308, 28 },
		{ 284, 309, 28 },
		{ 284, 310, 28 },
		{ 284, 311, 28 },
		{ 284, 312, 28 },
		{ 284, 313, 28 },
		{ 284, 314, 28 },
		{ 284, 315, 28 },
		{ 284, 316, 28 },
		{ 284, 317, 28 },
		{ 284, 318, 28 },
		{ 284, 319, 28 },
		{ 284, 320, 28 },
		{ 284, 321, 28 },
		{ 284, 322, 28 },
		{ 284, 323, 28 },
		{ 284, 324, 28 },
		{ 284, 325, 28 },
		{ 284, 326, 28 },
		{ 284, 327, 28 },
		{ 284, 328, 28 },
		{ 284, 329, 28 },
		{ 284, 330, 28 },
		{ 284, 331, 28 },
		{ 284, 332, 28 },
		{ 284, 333, 28 },
		{ 284, 334, 28 },
		{ 284, 335, 28 },
		{ 284, 336, 28 },
		{ 284, 337, 28 },
		{ 284, 338, 28 },
		{ 284, 339, 28 },
		{ 284, 340, 28 },
		{ 284, 341, 28 },
		{ 284, 342, 28 },
		{ 284, 343, 28 },
		{ 284, 344, 28 },
		{ 284, 345, 28 },
		{ 284, 346, 28 },
		{ 284, 347, 28 },
		{ 284, 348, 28 },
		{ 284, 349, 28 },
		{ 284, 350, 28 },
		{ 284, 351, 28 },
		{ 284, 352, 28 },
		{ 284, 353, 28 },
		{ 284, 354, 28 },
		{ 284, 355, 28 },
		{ 284, 356, 28 },
		{ 284, 357, 28 },
		{ 284, 358, 28 },
		{ 284, 359, 28 },
		{ 284, 360, 28 },
		{ 284, 361, 28 },
		{ 284, 362, 28 },
		{ 284, 363, 28 },
		{ 284, 364, 28 },
		{ 285, 295, 28 },
		{ 285, 296, 28 },
		{ 285, 297, 28 },
		{ 285, 298, 28 },
		{ 285, 299, 28 },
		{ 285, 300, 28 },
		{ 285, 301, 28 },
		{ 285, 302, 28 },
		{ 285, 303, 28 },
		{ 285, 304, 28 },
		{ 285, 305, 28 },
		{ 285, 306, 28 },
		{ 285, 307, 28 },
		{ 285, 308, 28 },
		{ 285, 309, 28 },
		{ 285, 310, 28 },
		{ 285, 311, 28 },
		{ 285, 312, 28 },
		{ 285, 313, 28 },
		{ 285, 314, 28 },
		{ 285, 315, 28 },
		{ 285, 316, 28 },
		{ 285, 317, 28 },
		{ 285, 318, 28 },
		{ 285, 319, 28 },
		{ 285, 320, 28 },
		{ 285, 321, 28 },
		{ 285, 322, 28 },
		{ 285, 323, 28 },
		{ 285, 324, 28 },
		{ 285, 325, 28 },
		{ 285, 326, 28 },
		{ 285, 327, 28 },
		{ 285, 328, 28 },
		{ 285, 329, 28 },
		{ 285, 330, 28 },
		{ 285, 331, 28 },
		{ 285, 332, 28 },
		{ 285, 333, 28 },
		{ 285, 334, 28 },
		{ 285, 335, 28 },
		{ 285, 336, 28 },
		{ 285, 337, 28 },
		{ 285, 338, 28 },
		{ 285, 339, 28 },
		{ 285, 340, 28 },
		{ 285, 341, 28 },
		{ 285, 342, 28 },
		{ 285, 343, 28 },
		{ 285, 344, 28 },
		{ 285, 345, 28 },
		{ 285, 346, 28 },
		{ 285, 347, 28 },
		{ 285, 348, 28 },
		{ 285, 349, 28 },
		{ 285, 350, 28 },
		{ 285, 351, 28 },
		{ 285, 352, 28 },
		{ 285, 353, 28 },
		{ 285, 354, 28 },
		{ 285, 355, 28 },
		{ 285, 356, 28 },
		{ 285, 357, 28 },
		{ 285, 358, 28 },
		{ 285, 359, 28 },
		{ 285, 360, 28 },
		{ 285, 361, 28 },
		{ 285, 362, 28 },
		{ 285, 363, 28 },
		{ 285, 364, 28 },
		{ 285, 365, 28 },
		{ 285, 366, 28 },
		{ 285, 367, 28 },
		{ 285, 368, 28 },
		{ 286, 296, 28 },
		{ 286, 297, 28 },
		{ 286, 298, 28 },
		{ 286, 299, 28 },
		{ 286, 300, 28 },
		{ 286, 301, 28 },
		{ 286, 302, 28 },
		{ 286, 303, 28 },
		{ 286, 304, 28 },
		{ 286, 305, 28 },
		{ 286, 306, 28 },
		{ 286, 307, 28 },
		{ 286, 308, 28 },
		{ 286, 309, 28 },
		{ 286, 310, 28 },
		{ 286, 311, 28 },
		{ 286, 312, 28 },
		{ 286, 313, 28 },
		{ 286, 314, 28 },
		{ 286, 315, 28 },
		{ 286, 316, 28 },
		{ 286, 317, 28 },
		{ 286, 318, 28 },
		{ 286, 319, 28 },
		{ 286, 320, 28 },
		{ 286, 321, 28 },
		{ 286, 322, 28 },
		{ 286, 323, 28 },
		{ 286, 324, 28 },
		{ 286, 325, 28 },
		{ 286, 326, 28 },
		{ 286, 327, 28 },
		{ 286, 328, 28 },
		{ 286, 329, 28 },
		{ 286, 330, 28 },
		{ 286, 331, 28 },
		{ 286, 332, 28 },
		{ 286, 333, 28 },
		{ 286, 334, 28 },
		{ 286, 335, 28 },
		{ 286, 336, 28 },
		{ 286, 337, 28 },
		{ 286, 338, 28 },
		{ 286, 339, 28 },
		{ 286, 340, 28 },
		{ 286, 341, 28 },
		{ 286, 342, 28 },
		{ 286, 343, 28 },
		{ 286, 344, 28 },
		{ 286, 345, 28 },
		{ 286, 346, 28 },
		{ 286, 347, 28 },
		{ 286, 348, 28 },
		{ 286, 349, 28 },
		{ 286, 350, 28 },
		{ 286, 351, 28 },
		{ 286, 352, 28 },
		{ 286, 353, 28 },
		{ 286, 354, 28 },
		{ 286, 355, 28 },
		{ 286, 356, 28 },
		{ 286, 357, 28 },
		{ 286, 358, 28 },
		{ 286, 359, 28 },
		{ 286, 360, 28 },
		{ 286, 361, 28 },
		{ 286, 362, 28 },
		{ 286, 363, 28 },
		{ 286, 364, 28 },
		{ 286, 365, 28 },
		{ 286, 366, 28 },
		{ 286, 367, 28 },
		{ 286, 368, 28 },
		{ 286, 369, 28 },
		{ 286, 370, 28 },
		{ 286, 371, 28 },
		{ 286, 372, 28 },
		{ 287, 298, 28 },
		{ 287, 299, 28 },
		{ 287, 300, 28 },
		{ 287, 301, 28 },
		{ 287, 302, 28 },
		{ 287, 303, 28 },
		{ 287, 304, 28 },
		{ 287, 305, 28 },
		{ 287, 306, 28 },
		{ 287, 307, 28 },
		{ 287, 308, 28 },
		{ 287, 309, 28 },
		{ 287, 310, 28 },
		{ 287, 311, 28 },
		{ 287, 312, 28 },
		{ 287, 313, 28 },
		{ 287, 314, 28 },
		{ 287, 315, 28 },
		{ 287, 316, 28 },
		{ 287, 317, 28 },
		{ 287, 318, 28 },
		{ 287, 319, 28 },
		{ 287, 320, 28 },
		{ 287, 321, 28 },
		{ 287, 322, 28 },
		{ 287, 323, 28 },
		{ 287, 324, 28 },
		{ 287, 325, 28 },
		{ 287, 326, 28 },
		{ 287, 327, 28 },
		{ 287, 328, 28 },
		{ 287, 329, 28 },
		{ 287, 330, 28 },
		{ 287, 331, 28 },
		{ 287, 332, 28 },
		{ 287, 333, 28 },
		{ 287, 334, 28 },
		{ 287, 335, 28 },
		{ 287, 336, 28 },
		{ 287, 337, 28 },
		{ 287, 338, 28 },
		{ 287, 339, 28 },
		{ 287, 340, 28 },
		{ 287, 341, 28 },
		{ 287, 342, 28 },
		{ 287, 343, 28 },
		{ 287, 344, 28 },
		{ 287, 345, 28 },
		{ 287, 346, 28 },
		{ 287, 347, 28 },
		{ 287, 348, 28 },
		{ 287, 349, 28 },
		{ 287, 350, 28 },
		{ 287, 351, 28 },
		{ 287, 352, 28 },
		{ 287, 353, 28 },
		{ 287, 354, 28 },
		{ 287, 355, 28 },
		{ 287, 356, 28 },
		{ 287, 357, 28 },
		{ 287, 358, 28 },
		{ 287, 359, 28 },
		{ 287, 360, 28 },
		{ 287, 361, 28 },
		{ 287, 362, 28 },
		{ 287, 363, 28 },
		{ 287, 364, 28 },
		{ 287, 365, 28 },
		{ 287, 366, 28 },
		{ 287, 367, 28 },
		{ 287, 368, 28 },
		{ 287, 369, 28 },
		{ 287, 370, 28 },
		{ 287, 371, 28 },
		{ 287, 372, 28 },
		{ 287, 373, 28 },
		{ 287, 374, 28 },
		{ 287, 375, 28 },
		{ 288, 299, 28 },
		{ 288, 300, 28 },
		{ 288, 301, 28 },
		{ 288, 302, 28 },
		{ 288, 303, 28 },
		{ 288, 304, 28 },
		{ 288, 305, 28 },
		{ 288, 306, 28 },
		{ 288, 307, 28 },
		{ 288, 308, 28 },
		{ 288, 309, 28 },
		{ 288, 310, 28 },
		{ 288, 311, 28 },
		{ 288, 312, 28 },
		{ 288, 313, 28 },
		{ 288, 314, 28 },
		{ 288, 315, 28 },
		{ 288, 316, 28 },
		{ 288, 317, 28 },
		{ 288, 318, 28 },
		{ 288, 319, 28 },
		{ 288, 320, 28 },
		{ 288, 321, 28 },
		{ 288, 322, 28 },
		{ 288, 323, 28 },
		{ 288, 324, 28 },
		{ 288, 325, 28 },
		{ 288, 326, 28 },
		{ 288, 327, 28 },
		{ 288, 328, 28 },
		{ 288, 329, 28 },
		{ 288, 330, 28 },
		{ 288, 331, 28 },
		{ 288, 332, 28 },
		{ 288, 333, 28 },
		{ 288, 334, 28 },
		{ 288, 335, 28 },
		{ 288, 336, 28 },
		{ 288, 337, 28 },
		{ 288, 338, 28 },
		{ 288, 339, 28 },
		{ 288, 340, 28 },
		{ 288, 341, 28 },
		{ 288, 342, 28 },
		{ 288, 343, 28 },
		{ 288, 344, 28 },
		{ 288, 345, 28 },
		{ 288, 346, 28 },
		{ 288, 347, 28 },
		{ 288, 348, 28 },
		{ 288, 349, 28 },
		{ 288, 350, 28 },
		{ 288, 351, 28 },
		{ 288, 352, 28 },
		{ 288, 353, 28 },
		{ 288, 354, 28 },
		{ 288, 355, 28 },
		{ 288, 356, 28 },
		{ 288, 357, 28 },
		{ 288, 358, 28 },
		{ 288, 359, 28 },
		{ 288, 360, 28 },
		{ 288, 361, 28 },
		{ 288, 362, 28 },
		{ 288, 363, 28 },
		{ 288, 364, 28 },
		{ 288, 365, 28 },
		{ 288, 366, 28 },
		{ 288, 367, 28 },
		{ 288, 368, 28 },
		{ 288, 369, 28 },
		{ 288, 370, 28 },
		{ 288, 371, 28 },
		{ 288, 372, 28 },
		{ 288, 373, 28 },
		{ 288, 374, 28 },
		{ 288, 375, 28 },
		{ 288, 376, 28 },
		{ 288, 377, 28 },
		{ 288, 378, 28 },
		{ 289, 300, 28 },
		{ 289, 301, 28 },
		{ 289, 302, 28 },
		{ 289, 303, 28 },
		{ 289, 304, 28 },
		{ 289, 305, 28 },
		{ 289, 306, 28 },
		{ 289, 307, 28 },
		{ 289, 308, 28 },
		{ 289, 309, 28 },
		{ 289, 310, 28 },
		{ 289, 311, 28 },
		{ 289, 312, 28 },
		{ 289, 313, 28 },
		{ 289, 314, 28 },
		{ 289, 315, 28 },
		{ 289, 316, 28 },
		{ 289, 317, 28 },
		{ 289, 318, 28 },
		{ 289, 319, 28 },
		{ 289, 320, 28 },
		{ 289, 321, 28 },
		{ 289, 322, 28 },
		{ 289, 323, 28 },
		{ 289, 324, 28 },
		{ 289, 325, 28 },
		{ 289, 326, 28 },
		{ 289, 327, 28 },
		{ 289, 328, 28 },
		{ 289, 329, 28 },
		{ 289, 330, 28 },
		{ 289, 331, 28 },
		{ 289, 332, 28 },
		{ 289, 333, 28 },
		{ 289, 334, 28 },
		{ 289, 335, 28 },
		{ 289, 336, 28 },
		{ 289, 337, 28 },
		{ 289, 338, 28 },
		{ 289, 339, 28 },
		{ 289, 340, 28 },
		{ 289, 341, 28 },
		{ 289, 342, 28 },
		{ 289, 343, 28 },
		{ 289, 344, 28 },
		{ 289, 345, 28 },
		{ 289, 346, 28 },
		{ 289, 347, 28 },
		{ 289, 348, 28 },
		{ 289, 349, 28 },
		{ 289, 350, 28 },
		{ 289, 351, 28 },
		{ 289, 352, 28 },
		{ 289, 353, 28 },
		{ 289, 354, 28 },
		{ 289, 355, 28 },
		{ 289, 356, 28 },
		{ 289, 357, 28 },
		{ 289, 358, 28 },
		{ 289, 359, 28 },
		{ 289, 360, 28 },
		{ 289, 361, 28 },
		{ 289, 362, 28 },
		{ 289, 363, 28 },
		{ 289, 364, 28 },
		{ 289, 365, 28 },
		{ 289, 366, 28 },
		{ 289, 367, 28 },
		{ 289, 368, 28 },
		{ 289, 369, 28 },
		{ 289, 370, 28 },
		{ 289, 371, 28 },
		{ 289, 372, 28 },
		{ 289, 373, 28 },
		{ 289, 374, 28 },
		{ 289, 375, 28 },
		{ 289, 376, 28 },
		{ 289, 377, 28 },
		{ 289, 378, 28 },
		{ 289, 379, 28 },
		{ 289, 380, 28 },
		{ 290, 302, 28 },
		{ 290, 303, 28 },
		{ 290, 304, 28 },
		{ 290, 305, 28 },
		{ 290, 306, 28 },
		{ 290, 307, 28 },
		{ 290, 308, 28 },
		{ 290, 309, 28 },
		{ 290, 310, 28 },
		{ 290, 311, 28 },
		{ 290, 312, 28 },
		{ 290, 313, 28 },
		{ 290, 314, 28 },
		{ 290, 315, 28 },
		{ 290, 316, 28 },
		{ 290, 317, 28 },
		{ 290, 318, 28 },
		{ 290, 319, 28 },
		{ 290, 320, 28 },
		{ 290, 321, 28 },
		{ 290, 322, 28 },
		{ 290, 323, 28 },
		{ 290, 324, 28 },
		{ 290, 325, 28 },
		{ 290, 326, 28 },
		{ 290, 327, 28 },
		{ 290, 328, 28 },
		{ 290, 329, 28 },
		{ 290, 330, 28 },
		{ 290, 331, 28 },
		{ 290, 332, 28 },
		{ 290, 333, 28 },
		{ 290, 334, 28 },
		{ 290, 335, 28 },
		{ 290, 336, 28 },
		{ 290, 337, 28 },
		{ 290, 338, 28 },
		{ 290, 339, 28 },
		{ 290, 340, 28 },
		{ 290, 341, 28 },
		{ 290, 342, 28 },
		{ 290, 343, 28 },
		{ 290, 344, 28 },
		{ 290, 345, 28 },
		{ 290, 346, 28 },
		{ 290, 347, 28 },
		{ 290, 348, 28 },
		{ 290, 349, 28 },
		{ 290, 350, 28 },
		{ 290, 351, 28 },
		{ 290, 352, 28 },
		{ 290, 353, 28 },
		{ 290, 354, 28 },
		{ 290, 355, 28 },
		{ 290, 356, 28 },
		{ 290, 357, 28 },
		{ 290, 358, 28 },
		{ 290, 359, 28 },
		{ 290, 360, 28 },
		{ 290, 361, 28 },
		{ 290, 362, 28 },
		{ 290, 363, 28 },
		{ 290, 364, 28 },
		{ 290, 365, 28 },
		{ 290, 366, 28 },
		{ 290, 367, 28 },
		{ 290, 368, 28 },
		{ 290, 369, 28 },
		{ 290, 370, 28 },
		{ 290, 371, 28 },
		{ 290, 372, 28 },
		{ 290, 373, 28 },
		{ 290, 374, 28 },
		{ 290, 375, 28 },
		{ 290, 376, 28 },
		{ 290, 377, 28 },
		{ 290, 378, 28 },
		{ 290, 379, 28 },
		{ 290, 380, 28 },
		{ 290, 381, 28 },
		{ 291, 303, 28 },
		{ 291, 304, 28 },
		{ 291, 305, 28 },
		{ 291, 306, 28 },
		{ 291, 307, 28 },
		{ 291, 308, 28 },
		{ 291, 309, 28 },
		{ 291, 310, 28 },
		{ 291, 311, 28 },
		{ 291, 312, 28 },
		{ 291, 313, 28 },
		{ 291, 314, 28 },
		{ 291, 315, 28 },
		{ 291, 316, 28 },
		{ 291, 317, 28 },
		{ 291, 318, 28 },
		{ 291, 319, 28 },
		{ 291, 320, 28 },
		{ 291, 321, 28 },
		{ 291, 322, 28 },
		{ 291, 323, 28 },
		{ 291, 324, 28 },
		{ 291, 325, 28 },
		{ 291, 326, 28 },
		{ 291, 327, 28 },
		{ 291, 328, 28 },
		{ 291, 329, 28 },
		{ 291, 330, 28 },
		{ 291, 331, 28 },
		{ 291, 332, 28 },
		{ 291, 333, 28 },
		{ 291, 334, 28 },
		{ 291, 335, 28 },
		{ 291, 336, 28 },
		{ 291, 337, 28 },
		{ 291, 338, 28 },
		{ 291, 339, 28 },
		{ 291, 340, 28 },
		{ 291, 341, 28 },
		{ 291, 342, 28 },
		{ 291, 343, 28 },
		{ 291, 344, 28 },
		{ 291, 345, 28 },
		{ 291, 346, 28 },
		{ 291, 347, 28 },
		{ 291, 348, 28 },
		{ 291, 349, 28 },
		{ 291, 350, 28 },
		{ 291, 351, 28 },
		{ 291, 352, 28 },
		{ 291, 353, 28 },
		{ 291, 354, 28 },
		{ 291, 355, 28 },
		{ 291, 356, 28 },
		{ 291, 357, 28 },
		{ 291, 358, 28 },
		{ 291, 359, 28 },
		{ 291, 360, 28 },
		{ 291, 361, 28 },
		{ 291, 362, 28 },
		{ 291, 363, 28 },
		{ 291, 364, 28 },
		{ 291, 365, 28 },
		{ 291, 366, 28 },
		{ 291, 367, 28 },
		{ 291, 368, 28 },
		{ 291, 369, 28 },
		{ 291, 370, 28 },
		{ 291, 371, 28 },
		{ 291, 372, 28 },
		{ 291, 373, 28 },
		{ 291, 374, 28 },
		{ 291, 375, 28 },
		{ 291, 376, 28 },
		{ 291, 377, 28 },
		{ 291, 378, 28 },
		{ 291, 379, 28 },
		{ 291, 380, 28 },
		{ 291, 381, 28 },
		{ 292, 304, 28 },
		{ 292, 305, 28 },
		{ 292, 306, 28 },
		{ 292, 307, 28 },
		{ 292, 308, 28 },
		{ 292, 309, 28 },
		{ 292, 310, 28 },
		{ 292, 311, 28 },
		{ 292, 312, 28 },
		{ 292, 313, 28 },
		{ 292, 314, 28 },
		{ 292, 315, 28 },
		{ 292, 316, 28 },
		{ 292, 317, 28 },
		{ 292, 318, 28 },
		{ 292, 319, 28 },
		{ 292, 320, 28 },
		{ 292, 321, 28 },
		{ 292, 322, 28 },
		{ 292, 323, 28 },
		{ 292, 324, 28 },
		{ 292, 325, 28 },
		{ 292, 326, 28 },
		{ 292, 327, 28 },
		{ 292, 328, 28 },
		{ 292, 329, 28 },
		{ 292, 330, 28 },
		{ 292, 331, 28 },
		{ 292, 332, 28 },
		{ 292, 333, 28 },
		{ 292, 334, 28 },
		{ 292, 335, 28 },
		{ 292, 336, 28 },
		{ 292, 337, 28 },
		{ 292, 338, 28 },
		{ 292, 339, 28 },
		{ 292, 340, 28 },
		{ 292, 341, 28 },
		{ 292, 342, 28 },
		{ 292, 343, 28 },
		{ 292, 344, 28 },
		{ 292, 345, 28 },
		{ 292, 346, 28 },
		{ 292, 347, 28 },
		{ 292, 348, 28 },
		{ 292, 349, 28 },
		{ 292, 350, 28 },
		{ 292, 351, 28 },
		{ 292, 352, 28 },
		{ 292, 353, 28 },
		{ 292, 354, 28 },
		{ 292, 355, 28 },
		{ 292, 356, 28 },
		{ 292, 357, 28 },
		{ 292, 358, 28 },
		{ 292, 359, 28 },
		{ 292, 360, 28 },
		{ 292, 361, 28 },
		{ 292, 362, 28 },
		{ 292, 363, 28 },
		{ 292, 364, 28 },
		{ 292, 365, 28 },
		{ 292, 366, 28 },
		{ 292, 367, 28 },
		{ 292, 368, 28 },
		{ 292, 369, 28 },
		{ 292, 370, 28 },
		{ 292, 371, 28 },
		{ 292, 372, 28 },
		{ 292, 373, 28 },
		{ 292, 374, 28 },
		{ 292, 375, 28 },
		{ 292, 376, 28 },
		{ 292, 377, 28 },
		{ 292, 378, 28 },
		{ 292, 379, 28 },
		{ 292, 380, 28 },
		{ 292, 381, 28 },
		{ 292, 382, 28 },
		{ 293, 306, 28 },
		{ 293, 307, 28 },
		{ 293, 308, 28 },
		{ 293, 309, 28 },
		{ 293, 310, 28 },
		{ 293, 311, 28 },
		{ 293, 312, 28 },
		{ 293, 313, 28 },
		{ 293, 314, 28 },
		{ 293, 315, 28 },
		{ 293, 316, 28 },
		{ 293, 317, 28 },
		{ 293, 318, 28 },
		{ 293, 319, 28 },
		{ 293, 320, 28 },
		{ 293, 321, 28 },
		{ 293, 322, 28 },
		{ 293, 323, 28 },
		{ 293, 324, 28 },
		{ 293, 325, 28 },
		{ 293, 326, 28 },
		{ 293, 327, 28 },
		{ 293, 328, 28 },
		{ 293, 329, 28 },
		{ 293, 330, 28 },
		{ 293, 331, 28 },
		{ 293, 332, 28 },
		{ 293, 333, 28 },
		{ 293, 334, 28 },
		{ 293, 335, 28 },
		{ 293, 336, 28 },
		{ 293, 337, 28 },
		{ 293, 338, 28 },
		{ 293, 339, 28 },
		{ 293, 340, 28 },
		{ 293, 341, 28 },
		{ 293, 342, 28 },
		{ 293, 343, 28 },
		{ 293, 344, 28 },
		{ 293, 345, 28 },
		{ 293, 346, 28 },
		{ 293, 347, 28 },
		{ 293, 348, 28 },
		{ 293, 349, 28 },
		{ 293, 350, 28 },
		{ 293, 351, 28 },
		{ 293, 352, 28 },
		{ 293, 353, 28 },
		{ 293, 354, 28 },
		{ 293, 355, 28 },
		{ 293, 356, 28 },
		{ 293, 357, 28 },
		{ 293, 358, 28 },
		{ 293, 359, 28 },
		{ 293, 360, 28 },
		{ 293, 361, 28 },
		{ 293, 362, 28 },
		{ 293, 363, 28 },
		{ 293, 364, 28 },
		{ 293, 365, 28 },
		{ 293, 366, 28 },
		{ 293, 367, 28 },
		{ 293, 368, 28 },
		{ 293, 369, 28 },
		{ 293, 370, 28 },
		{ 293, 371, 28 },
		{ 293, 372, 28 },
		{ 293, 373, 28 },
		{ 293, 374, 28 },
		{ 293, 375, 28 },
		{ 293, 376, 28 },
		{ 293, 377, 28 },
		{ 293, 378, 28 },
		{ 293, 379, 28 },
		{ 293, 380, 28 },
		{ 293, 381, 28 },
		{ 293, 382, 28 },
		{ 294, 307, 28 },
		{ 294, 308, 28 },
		{ 294, 309, 28 },
		{ 294, 310, 28 },
		{ 294, 311, 28 },
		{ 294, 312, 28 },
		{ 294, 313, 28 },
		{ 294, 314, 28 },
		{ 294, 315, 28 },
		{ 294, 316, 28 },
		{ 294, 317, 28 },
		{ 294, 318, 28 },
		{ 294, 319, 28 },
		{ 294, 320, 28 },
		{ 294, 321, 28 },
		{ 294, 322, 28 },
		{ 294, 323, 28 },
		{ 294, 324, 28 },
		{ 294, 325, 28 },
		{ 294, 326, 28 },
		{ 294, 327, 28 },
		{ 294, 328, 28 },
		{ 294, 329, 28 },
		{ 294, 330, 28 },
		{ 294, 331, 28 },
		{ 294, 332, 28 },
		{ 294, 333, 28 },
		{ 294, 334, 28 },
		{ 294, 335, 28 },
		{ 294, 336, 28 },
		{ 294, 337, 28 },
		{ 294, 338, 28 },
		{ 294, 339, 28 },
		{ 294, 340, 28 },
		{ 294, 341, 28 },
		{ 294, 342, 28 },
		{ 294, 343, 28 },
		{ 294, 344, 28 },
		{ 294, 345, 28 },
		{ 294, 346, 28 },
		{ 294, 347, 28 },
		{ 294, 348, 28 },
		{ 294, 349, 28 },
		{ 294, 350, 28 },
		{ 294, 351, 28 },
		{ 294, 352, 28 },
		{ 294, 353, 28 },
		{ 294, 354, 28 },
		{ 294, 355, 28 },
		{ 294, 356, 28 },
		{ 294, 357, 28 },
		{ 294, 358, 28 },
		{ 294, 359, 28 },
		{ 294, 360, 28 },
		{ 294, 361, 28 },
		{ 294, 362, 28 },
		{ 294, 363, 28 },
		{ 294, 364, 28 },
		{ 294, 365, 28 },
		{ 294, 366, 28 },
		{ 294, 367, 28 },
		{ 294, 368, 28 },
		{ 294, 369, 28 },
		{ 294, 370, 28 },
		{ 294, 371, 28 },
		{ 294, 372, 28 },
		{ 294, 373, 28 },
		{ 294, 374, 28 },
		{ 294, 375, 28 },
		{ 294, 376, 28 },
		{ 294, 377, 28 },
		{ 294, 378, 28 },
		{ 294, 379, 28 },
		{ 294, 380, 28 },
		{ 294, 381, 28 },
		{ 294, 382, 28 },
		{ 295, 308, 28 },
		{ 295, 309, 28 },
		{ 295, 310, 28 },
		{ 295, 311, 28 },
		{ 295, 312, 28 },
		{ 295, 313, 28 },
		{ 295, 314, 28 },
		{ 295, 315, 28 },
		{ 295, 316, 28 },
		{ 295, 317, 28 },
		{ 295, 318, 28 },
		{ 295, 319, 28 },
		{ 295, 320, 28 },
		{ 295, 321, 28 },
		{ 295, 322, 28 },
		{ 295, 323, 28 },
		{ 295, 324, 28 },
		{ 295, 325, 28 },
		{ 295, 326, 28 },
		{ 295, 327, 28 },
		{ 295, 328, 28 },
		{ 295, 329, 28 },
		{ 295, 330, 28 },
		{ 295, 331, 28 },
		{ 295, 332, 28 },
		{ 295, 333, 28 },
		{ 295, 334, 28 },
		{ 295, 335, 28 },
		{ 295, 336, 28 },
		{ 295, 337, 28 },
		{ 295, 338, 28 },
		{ 295, 339, 28 },
		{ 295, 340, 28 },
		{ 295, 341, 28 },
		{ 295, 342, 28 },
		{ 295, 343, 28 },
		{ 295, 344, 28 },
		{ 295, 345, 28 },
		{ 295, 346, 28 },
		{ 295, 347, 28 },
		{ 295, 348, 28 },
		{ 295, 349, 28 },
		{ 295, 350, 28 },
		{ 295, 351, 28 },
		{ 295, 352, 28 },
		{ 295, 353, 28 },
		{ 295, 354, 28 },
		{ 295, 355, 28 },
		{ 295, 356, 28 },
		{ 295, 357, 28 },
		{ 295, 358, 28 },
		{ 295, 359, 28 },
		{ 295, 360, 28 },
		{ 295, 361, 28 },
		{ 295, 362, 28 },
		{ 295, 363, 28 },
		{ 295, 364, 28 },
		{ 295, 365, 28 },
		{ 295, 366, 28 },
		{ 295, 367, 28 },
		{ 295, 368, 28 },
		{ 295, 369, 28 },
		{ 295, 370, 28 },
		{ 295, 371, 28 },
		{ 295, 372, 28 },
		{ 295, 373, 28 },
		{ 295, 374, 28 },
		{ 295, 375, 28 },
		{ 295, 376, 28 },
		{ 295, 377, 28 },
		{ 295, 378, 28 },
		{ 295, 379, 28 },
		{ 295, 380, 28 },
		{ 295, 381, 28 },
		{ 295, 382, 28 },
		{ 296, 310, 28 },
		{ 296, 311, 28 },
		{ 296, 312, 28 },
		{ 296, 313, 28 },
		{ 296, 314, 28 },
		{ 296, 315, 28 },
		{ 296, 316, 28 },
		{ 296, 317, 28 },
		{ 296, 318, 28 },
		{ 296, 319, 28 },
		{ 296, 320, 28 },
		{ 296, 321, 28 },
		{ 296, 322, 28 },
		{ 296, 323, 28 },
		{ 296, 324, 28 },
		{ 296, 325, 28 },
		{ 296, 326, 28 },
		{ 296, 327, 28 },
		{ 296, 328, 28 },
		{ 296, 329, 28 },
		{ 296, 330, 28 },
		{ 296, 331, 28 },
		{ 296, 332, 28 },
		{ 296, 333, 28 },
		{ 296, 334, 28 },
		{ 296, 335, 28 },
		{ 296, 336, 28 },
		{ 296, 337, 28 },
		{ 296, 338, 28 },
		{ 296, 339, 28 },
		{ 296, 340, 28 },
		{ 296, 341, 28 },
		{ 296, 342, 28 },
		{ 296, 343, 28 },
		{ 296, 344, 28 },
		{ 296, 345, 28 },
		{ 296, 346, 28 },
		{ 296, 347, 28 },
		{ 296, 348, 28 },
		{ 296, 349, 28 },
		{ 296, 350, 28 },
		{ 296, 351, 28 },
		{ 296, 352, 28 },
		{ 296, 353, 28 },
		{ 296, 354, 28 },
		{ 296, 355, 28 },
		{ 296, 356, 28 },
		{ 296, 357, 28 },
		{ 296, 358, 28 },
		{ 296, 359, 28 },
		{ 296, 360, 28 },
		{ 296, 361, 28 },
		{ 296, 362, 28 },
		{ 296, 363, 28 },
		{ 296, 364, 28 },
		{ 296, 365, 28 },
		{ 296, 366, 28 },
		{ 296, 367, 28 },
		{ 296, 368, 28 },
		{ 296, 369, 28 },
		{ 296, 370, 28 },
		{ 296, 371, 28 },
		{ 296, 372, 28 },
		{ 296, 373, 28 },
		{ 296, 374, 28 },
		{ 296, 375, 28 },
		{ 296, 376, 28 },
		{ 296, 377, 28 },
		{ 296, 378, 28 },
		{ 296, 379, 28 },
		{ 296, 380, 28 },
		{ 296, 381, 28 },
		{ 296, 382, 28 },
		{ 296, 383, 28 },
		{ 297, 311, 28 },
		{ 297, 312, 28 },
		{ 297, 313, 28 },
		{ 297, 314, 28 },
		{ 297, 315, 28 },
		{ 297, 316, 28 },
		{ 297, 317, 28 },
		{ 297, 318, 28 },
		{ 297, 319, 28 },
		{ 297, 320, 28 },
		{ 297, 321, 28 },
		{ 297, 322, 28 },
		{ 297, 323, 28 },
		{ 297, 324, 28 },
		{ 297, 325, 28 },
		{ 297, 326, 28 },
		{ 297, 327, 28 },
		{ 297, 328, 28 },
		{ 297, 329, 28 },
		{ 297, 330, 28 },
		{ 297, 331, 28 },
		{ 297, 332, 28 },
		{ 297, 333, 28 },
		{ 297, 334, 28 },
		{ 297, 335, 28 },
		{ 297, 336, 28 },
		{ 297, 337, 28 },
		{ 297, 338, 28 },
		{ 297, 339, 28 },
		{ 297, 340, 28 },
		{ 297, 341, 28 },
		{ 297, 342, 28 },
		{ 297, 343, 28 },
		{ 297, 344, 28 },
		{ 297, 345, 28 },
		{ 297, 346, 28 },
		{ 297, 347, 28 },
		{ 297, 348, 28 },
		{ 297, 349, 28 },
		{ 297, 350, 28 },
		{ 297, 351, 28 },
		{ 297, 352, 28 },
		{ 297, 353, 28 },
		{ 297, 354, 28 },
		{ 297, 355, 28 },
		{ 297, 356, 28 },
		{ 297, 357, 28 },
		{ 297, 358, 28 },
		{ 297, 359, 28 },
		{ 297, 360, 28 },
		{ 297, 361, 28 },
		{ 297, 362, 28 },
		{ 297, 363, 28 },
		{ 297, 364, 28 },
		{ 297, 365, 28 },
		{ 297, 366, 28 },
		{ 297, 367, 28 },
		{ 297, 368, 28 },
		{ 297, 369, 28 },
		{ 297, 370, 28 },
		{ 297, 371, 28 },
		{ 297, 372, 28 },
		{ 297, 373, 28 },
		{ 297, 374, 28 },
		{ 297, 375, 28 },
		{ 297, 376, 28 },
		{ 297, 377, 28 },
		{ 297, 378, 28 },
		{ 297, 379, 28 },
		{ 297, 380, 28 },
		{ 297, 381, 28 },
		{ 297, 382, 28 },
		{ 297, 383, 28 },
		{ 298, 312, 28 },
		{ 298, 313, 28 },
		{ 298, 314, 28 },
		{ 298, 315, 28 },
		{ 298, 316, 28 },
		{ 298, 317, 28 },
		{ 298, 318, 28 },
		{ 298, 319, 28 },
		{ 298, 320, 28 },
		{ 298, 321, 28 },
		{ 298, 322, 28 },
		{ 298, 323, 28 },
		{ 298, 324, 28 },
		{ 298, 325, 28 },
		{ 298, 326, 28 },
		{ 298, 327, 28 },
		{ 298, 328, 28 },
		{ 298, 329, 28 },
		{ 298, 330, 28 },
		{ 298, 331, 28 },
		{ 298, 332, 28 },
		{ 298, 333, 28 },
		{ 298, 334, 28 },
		{ 298, 335, 28 },
		{ 298, 336, 28 },
		{ 298, 337, 28 },
		{ 298, 338, 28 },
		{ 298, 339, 28 },
		{ 298, 340, 28 },
		{ 298, 341, 28 },
		{ 298, 342, 28 },
		{ 298, 343, 28 },
		{ 298, 344, 28 },
		{ 298, 345, 28 },
		{ 298, 346, 28 },
		{ 298, 347, 28 },
		{ 298, 348, 28 },
		{ 298, 349, 28 },
		{ 298, 350, 28 },
		{ 298, 351, 28 },
		{ 298, 352, 28 },
		{ 298, 353, 28 },
		{ 298, 354, 28 },
		{ 298, 355, 28 },
		{ 298, 356, 28 },
		{ 298, 357, 28 },
		{ 298, 358, 28 },
		{ 298, 359, 28 },
		{ 298, 360, 28 },
		{ 298, 361, 28 },
		{ 298, 362, 28 },
		{ 298, 363, 28 },
		{ 298, 364, 28 },
		{ 298, 365, 28 },
		{ 298, 366, 28 },
		{ 298, 367, 28 },
		{ 298, 368, 28 },
		{ 298, 369, 28 },
		{ 298, 370, 28 },
		{ 298, 371, 28 },
		{ 298, 372, 28 },
		{ 298, 373, 28 },
		{ 298, 374, 28 },
		{ 298, 375, 28 },
		{ 298, 376, 28 },
		{ 298, 377, 28 },
		{ 298, 378, 28 },
		{ 298, 379, 28 },
		{ 298, 380, 28 },
		{ 298, 381, 28 },
		{ 298, 382, 28 },
		{ 298, 383, 28 },
		{ 299, 314, 28 },
		{ 299, 315, 28 },
		{ 299, 316, 28 },
		{ 299, 317, 28 },
		{ 299, 318, 28 },
		{ 299, 319, 28 },
		{ 299, 320, 28 },
		{ 299, 321, 28 },
		{ 299, 322, 28 },
		{ 299, 323, 28 },
		{ 299, 324, 28 },
		{ 299, 325, 28 },
		{ 299, 326, 28 },
		{ 299, 327, 28 },
		{ 299, 328, 28 },
		{ 299, 329, 28 },
		{ 299, 330, 28 },
		{ 299, 331, 28 },
		{ 299, 332, 28 },
		{ 299, 333, 28 },
		{ 299, 334, 28 },
		{ 299, 335, 28 },
		{ 299, 336, 28 },
		{ 299, 337, 28 },
		{ 299, 338, 28 },
		{ 299, 339, 28 },
		{ 299, 340, 28 },
		{ 299, 341, 28 },
		{ 299, 342, 28 },
		{ 299, 343, 28 },
		{ 299, 344, 28 },
		{ 299, 345, 28 },
		{ 299, 346, 28 },
		{ 299, 347, 28 },
		{ 299, 348, 28 },
		{ 299, 349, 28 },
		{ 299, 350, 28 },
		{ 299, 351, 28 },
		{ 299, 352, 28 },
		{ 299, 353, 28 },
		{ 299, 354, 28 },
		{ 299, 355, 28 },
		{ 299, 356, 28 },
		{ 299, 357, 28 },
		{ 299, 358, 28 },
		{ 299, 359, 28 },
		{ 299, 360, 28 },
		{ 299, 361, 28 },
		{ 299, 362, 28 },
		{ 299, 363, 28 },
		{ 299, 364, 28 },
		{ 299, 365, 28 },
		{ 299, 366, 28 },
		{ 299, 367, 28 },
		{ 299, 368, 28 },
		{ 299, 369, 28 },
		{ 299, 370, 28 },
		{ 299, 371, 28 },
		{ 299, 372, 28 },
		{ 299, 373, 28 },
		{ 299, 374, 28 },
		{ 299, 375, 28 },
		{ 299, 376, 28 },
		{ 299, 377, 28 },
		{ 299, 378, 28 },
		{ 299, 379, 28 },
		{ 299, 380, 28 },
		{ 299, 381, 28 },
		{ 299, 382, 28 },
		{ 299, 383, 28 },
		{ 300, 315, 28 },
		{ 300, 316, 28 },
		{ 300, 317, 28 },
		{ 300, 318, 28 },
		{ 300, 319, 28 },
		{ 300, 320, 28 },
		{ 300, 321, 28 },
		{ 300, 322, 28 },
		{ 300, 323, 28 },
		{ 300, 324, 28 },
		{ 300, 325, 28 },
		{ 300, 326, 28 },
		{ 300, 327, 28 },
		{ 300, 328, 28 },
		{ 300, 329, 28 },
		{ 300, 330, 28 },
		{ 300, 331, 28 },
		{ 300, 332, 28 },
		{ 300, 333, 28 },
		{ 300, 334, 28 },
		{ 300, 335, 28 },
		{ 300, 336, 28 },
		{ 300, 337, 28 },
		{ 300, 338, 28 },
		{ 300, 339, 28 },
		{ 300, 340, 28 },
		{ 300, 341, 28 },
		{ 300, 342, 28 },
		{ 300, 343, 28 },
		{ 300, 344, 28 },
		{ 300, 345, 28 },
		{ 300, 346, 28 },
		{ 300, 347, 28 },
		{ 300, 348, 28 },
		{ 300, 349, 28 },
		{ 300, 350, 28 },
		{ 300, 351, 28 },
		{ 300, 352, 28 },
		{ 300, 353, 28 },
		{ 300, 354, 28 },
		{ 300, 355, 28 },
		{ 300, 356, 28 },
		{ 300, 357, 28 },
		{ 300, 358, 28 },
		{ 300, 359, 28 },
		{ 300, 360, 28 },
		{ 300, 361, 28 },
		{ 300, 362, 28 },
		{ 300, 363, 28 },
		{ 300, 364, 28 },
		{ 300, 365, 28 },
		{ 300, 366, 28 },
		{ 300, 367, 28 },
		{ 300, 368, 28 },
		{ 300, 369, 28 },
		{ 300, 370, 28 },
		{ 300, 371, 28 },
		{ 300, 372, 28 },
		{ 300, 373, 28 },
		{ 300, 374, 28 },
		{ 300, 375, 28 },
		{ 300, 376, 28 },
		{ 300, 377, 28 },
		{ 300, 378, 28 },
		{ 300, 379, 28 },
		{ 300, 380, 28 },
		{ 300, 381, 28 },
		{ 300, 382, 28 },
		{ 300, 383, 28 },
		{ 301, 316, 28 },
		{ 301, 317, 28 },
		{ 301, 318, 28 },
		{ 301, 319, 28 },
		{ 301, 320, 28 },
		{ 301, 321, 28 },
		{ 301, 322, 28 },
		{ 301, 323, 28 },
		{ 301, 324, 28 },
		{ 301, 325, 28 },
		{ 301, 326, 28 },
		{ 301, 327, 28 },
		{ 301, 328, 28 },
		{ 301, 329, 28 },
		{ 301, 330, 28 },
		{ 301, 331, 28 },
		{ 301, 332, 28 },
		{ 301, 333, 28 },
		{ 301, 334, 28 },
		{ 301, 335, 28 },
		{ 301, 336, 28 },
		{ 301, 337, 28 },
		{ 301, 338, 28 },
		{ 301, 339, 28 },
		{ 301, 340, 28 },
		{ 301, 341, 28 },
		{ 301, 342, 28 },
		{ 301, 343, 28 },
		{ 301, 344, 28 },
		{ 301, 345, 28 },
		{ 301, 346, 28 },
		{ 301, 347, 28 },
		{ 301, 348, 28 },
		{ 301, 349, 28 },
		{ 301, 350, 28 },
		{ 301, 351, 28 },
		{ 301, 352, 28 },
		{ 301, 353, 28 },
		{ 301, 354, 28 },
		{ 301, 355, 28 },
		{ 301, 356, 28 },
		{ 301, 357, 28 },
		{ 301, 358, 28 },
		{ 301, 359, 28 },
		{ 301, 360, 28 },
		{ 301, 361, 28 },
		{ 301, 362, 28 },
		{ 301, 363, 28 },
		{ 301, 364, 28 },
		{ 301, 365, 28 },
		{ 301, 366, 28 },
		{ 301, 367, 28 },
		{ 301, 368, 28 },
		{ 301, 369, 28 },
		{ 301, 370, 28 },
		{ 301, 371, 28 },
		{ 301, 372, 28 },
		{ 301, 373, 28 },
		{ 301, 374, 28 },
		{ 301, 375, 28 },
		{ 301, 376, 28 },
		{ 301, 377, 28 },
		{ 301, 378, 28 },
		{ 301, 379, 28 },
		{ 301, 380, 28 },
		{ 301, 381, 28 },
		{ 301, 382, 28 },
		{ 301, 383, 28 },
		{ 302, 318, 28 },
		{ 302, 319, 28 },
		{ 302, 320, 28 },
		{ 302, 321, 28 },
		{ 302, 322, 28 },
		{ 302, 323, 28 },
		{ 302, 324, 28 },
		{ 302, 325, 28 },
		{ 302, 326, 28 },
		{ 302, 327, 28 },
		{ 302, 328, 28 },
		{ 302, 329, 28 },
		{ 302, 330, 28 },
		{ 302, 331, 28 },
		{ 302, 332, 28 },
		{ 302, 333, 28 },
		{ 302, 334, 28 },
		{ 302, 335, 28 },
		{ 302, 336, 28 },
		{ 302, 337, 28 },
		{ 302, 338, 28 },
		{ 302, 339, 28 },
		{ 302, 340, 28 },
		{ 302, 341, 28 },
		{ 302, 342, 28 },
		{ 302, 343, 28 },
		{ 302, 344, 28 },
		{ 302, 345, 28 },
		{ 302, 346, 28 },
		{ 302, 347, 28 },
		{ 302, 348, 28 },
		{ 302, 349, 28 },
		{ 302, 350, 28 },
		{ 302, 351, 28 },
		{ 302, 352, 28 },
		{ 302, 353, 28 },
		{ 302, 354, 28 },
		{ 302, 355, 28 },
		{ 302, 356, 28 },
		{ 302, 357, 28 },
		{ 302, 358, 28 },
		{ 302, 359, 28 },
		{ 302, 360, 28 },
		{ 302, 361, 28 },
		{ 302, 362, 28 },
		{ 302, 363, 28 },
		{ 302, 364, 28 },
		{ 302, 365, 28 },
		{ 302, 366, 28 },
		{ 302, 367, 28 },
		{ 302, 368, 28 },
		{ 302, 369, 28 },
		{ 302, 370, 28 },
		{ 302, 371, 28 },
		{ 302, 372, 28 },
		{ 302, 373, 28 },
		{ 302, 374, 28 },
		{ 302, 375, 28 },
		{ 302, 376, 28 },
		{ 302, 377, 28 },
		{ 302, 378, 28 },
		{ 302, 379, 28 },
		{ 302, 380, 28 },
		{ 302, 381, 28 },
		{ 302, 382, 28 },
		{ 302, 383, 28 },
		{ 303, 319, 28 },
		{ 303, 320, 28 },
		{ 303, 321, 28 },
		{ 303, 322, 28 },
		{ 303, 323, 28 },
		{ 303, 324, 28 },
		{ 303, 325, 28 },
		{ 303, 326, 28 },
		{ 303, 327, 28 },
		{ 303, 328, 28 },
		{ 303, 329, 28 },
		{ 303, 330, 28 },
		{ 303, 331, 28 },
		{ 303, 332, 28 },
		{ 303, 333, 28 },
		{ 303, 334, 28 },
		{ 303, 335, 28 },
		{ 303, 336, 28 },
		{ 303, 337, 28 },
		{ 303, 338, 28 },
		{ 303, 339, 28 },
		{ 303, 340, 28 },
		{ 303, 341, 28 },
		{ 303, 342, 28 },
		{ 303, 343, 28 },
		{ 303, 344, 28 },
		{ 303, 345, 28 },
		{ 303, 346, 28 },
		{ 303, 347, 28 },
		{ 303, 348, 28 },
		{ 303, 349, 28 },
		{ 303, 350, 28 },
		{ 303, 351, 28 },
		{ 303, 352, 28 },
		{ 303, 353, 28 },
		{ 303, 354, 28 },
		{ 303, 355, 28 },
		{ 303, 356, 28 },
		{ 303, 357, 28 },
		{ 303, 358, 28 },
		{ 303, 359, 28 },
		{ 303, 360, 28 },
		{ 303, 361, 28 },
		{ 303, 362, 28 },
		{ 303, 363, 28 },
		{ 303, 364, 28 },
		{ 303, 365, 28 },
		{ 303, 366, 28 },
		{ 303, 367, 28 },
		{ 303, 368, 28 },
		{ 303, 369, 28 },
		{ 303, 370, 28 },
		{ 303, 371, 28 },
		{ 303, 372, 28 },
		{ 303, 373, 28 },
		{ 303, 374, 28 },
		{ 303, 375, 28 },
		{ 303, 376, 28 },
		{ 303, 377, 28 },
		{ 303, 378, 28 },
		{ 303, 379, 28 },
		{ 303, 380, 28 },
		{ 303, 381, 28 },
		{ 303, 382, 28 },
		{ 303, 383, 28 },
		{ 304, 320, 28 },
		{ 304, 321, 28 },
		{ 304, 322, 28 },
		{ 304, 323, 28 },
		{ 304, 324, 28 },
		{ 304, 325, 28 },
		{ 304, 326, 28 },
		{ 304, 327, 28 },
		{ 304, 328, 28 },
		{ 304, 329, 28 },
		{ 304, 330, 28 },
		{ 304, 331, 28 },
		{ 304, 332, 28 },
		{ 304, 333, 28 },
		{ 304, 334, 28 },
		{ 304, 335, 28 },
		{ 304, 336, 28 },
		{ 304, 337, 28 },
		{ 304, 338, 28 },
		{ 304, 339, 28 },
		{ 304, 340, 28 },
		{ 304, 341, 28 },
		{ 304, 342, 28 },
		{ 304, 343, 28 },
		{ 304, 344, 28 },
		{ 304, 345, 28 },
		{ 304, 346, 28 },
		{ 304, 347, 28 },
		{ 304, 348, 28 },
		{ 304, 349, 28 },
		{ 304, 350, 28 },
		{ 304, 351, 28 },
		{ 304, 352, 28 },
		{ 304, 353, 28 },
		{ 304, 354, 28 },
		{ 304, 355, 28 },
		{ 304, 356, 28 },
		{ 304, 357, 28 },
		{ 304, 358, 28 },
		{ 304, 359, 28 },
		{ 304, 360, 28 },
		{ 304, 361, 28 },
		{ 304, 362, 28 },
		{ 304, 363, 28 },
		{ 304, 364, 28 },
		{ 304, 365, 28 },
		{ 304, 366, 28 },
		{ 304, 367, 28 },
		{ 304, 368, 28 },
		{ 304, 369, 28 },
		{ 304, 370, 28 },
		{ 304, 371, 28 },
		{ 304, 372, 28 },
		{ 304, 373, 28 },
		{ 304, 374, 28 },
		{ 304, 375, 28 },
		{ 304, 376, 28 },
		{ 304, 377, 28 },
		{ 304, 378, 28 },
		{ 304, 379, 28 },
		{ 304, 380, 28 },
		{ 304, 381, 28 },
		{ 304, 382, 28 },
		{ 304, 383, 28 },
		{ 305, 322, 28 },
		{ 305, 323, 28 },
		{ 305, 324, 28 },
		{ 305, 325, 28 },
		{ 305, 326, 28 },
		{ 305, 327, 28 },
		{ 305, 328, 28 },
		{ 305, 329, 28 },
		{ 305, 330, 28 },
		{ 305, 331, 28 },
		{ 305, 332, 28 },
		{ 305, 333, 28 },
		{ 305, 334, 28 },
		{ 305, 335, 28 },
		{ 305, 336, 28 },
		{ 305, 337, 28 },
		{ 305, 338, 28 },
		{ 305, 339, 28 },
		{ 305, 340, 28 },
		{ 305, 341, 28 },
		{ 305, 342, 28 },
		{ 305, 343, 28 },
		{ 305, 344, 28 },
		{ 305, 345, 28 },
		{ 305, 346, 28 },
		{ 305, 347, 28 },
		{ 305, 348, 28 },
		{ 305, 349, 28 },
		{ 305, 350, 28 },
		{ 305, 351, 28 },
		{ 305, 352, 28 },
		{ 305, 353, 28 },
		{ 305, 354, 28 },
		{ 305, 355, 28 },
		{ 305, 356, 28 },
		{ 305, 357, 28 },
		{ 305, 358, 28 },
		{ 305, 359, 28 },
		{ 305, 360, 28 },
		{ 305, 361, 28 },
		{ 305, 362, 28 },
		{ 305, 363, 28 },
		{ 305, 364, 28 },
		{ 305, 365, 28 },
		{ 305, 366, 28 },
		{ 305, 367, 28 },
		{ 305, 368, 28 },
		{ 305, 369, 28 },
		{ 305, 370, 28 },
		{ 305, 371, 28 },
		{ 305, 372, 28 },
		{ 305, 373, 28 },
		{ 305, 374, 28 },
		{ 305, 375, 28 },
		{ 305, 376, 28 },
		{ 305, 377, 28 },
		{ 305, 378, 28 },
		{ 305, 379, 28 },
		{ 305, 380, 28 },
		{ 305, 381, 28 },
		{ 305, 382, 28 },
		{ 305, 383, 28 },
		{ 306, 323, 28 },
		{ 306, 324, 28 },
		{ 306, 325, 28 },
		{ 306, 326, 28 },
		{ 306, 327, 28 },
		{ 306, 328, 28 },
		{ 306, 329, 28 },
		{ 306, 330, 28 },
		{ 306, 331, 28 },
		{ 306, 332, 28 },
		{ 306, 333, 28 },
		{ 306, 334, 28 },
		{ 306, 335, 28 },
		{ 306, 336, 28 },
		{ 306, 337, 28 },
		{ 306, 338, 28 },
		{ 306, 339, 28 },
		{ 306, 340, 28 },
		{ 306, 341, 28 },
		{ 306, 342, 28 },
		{ 306, 343, 28 },
		{ 306, 344, 28 },
		{ 306, 345, 28 },
		{ 306, 346, 28 },
		{ 306, 347, 28 },
		{ 306, 348, 28 },
		{ 306, 349, 28 },
		{ 306, 350, 28 },
		{ 306, 351, 28 },
		{ 306, 352, 28 },
		{ 306, 353, 28 },
		{ 306, 354, 28 },
		{ 306, 355, 28 },
		{ 306, 356, 28 },
		{ 306, 357, 28 },
		{ 306, 358, 28 },
		{ 306, 359, 28 },
		{ 306, 360, 28 },
		{ 306, 361, 28 },
		{ 306, 362, 28 },
		{ 306, 363, 28 },
		{ 306, 364, 28 },
		{ 306, 365, 28 },
		{ 306, 366, 28 },
		{ 306, 367, 28 },
		{ 306, 368, 28 },
		{ 306, 369, 28 },
		{ 306, 370, 28 },
		{ 306, 371, 28 },
		{ 306, 372, 28 },
		{ 306, 373, 28 },
		{ 306, 374, 28 },
		{ 306, 375, 28 },
		{ 306, 376, 28 },
		{ 306, 377, 28 },
		{ 306, 378, 28 },
		{ 306, 379, 28 },
		{ 306, 380, 28 },
		{ 306, 381, 28 },
		{ 306, 382, 28 },
		{ 306, 383, 28 },
		{ 307, 324, 28 },
		{ 307, 325, 28 },
		{ 307, 326, 28 },
		{ 307, 327, 28 },
		{ 307, 328, 28 },
		{ 307, 329, 28 },
		{ 307, 330, 28 },
		{ 307, 331, 28 },
		{ 307, 332, 28 },
		{ 307, 333, 28 },
		{ 307, 334, 28 },
		{ 307, 335, 28 },
		{ 307, 336, 28 },
		{ 307, 337, 28 },
		{ 307, 338, 28 },
		{ 307, 339, 28 },
		{ 307, 340, 28 },
		{ 307, 341, 28 },
		{ 307, 342, 28 },
		{ 307, 343, 28 },
		{ 307, 344, 28 },
		{ 307, 345, 28 },
		{ 307, 346, 28 },
		{ 307, 347, 28 },
		{ 307, 348, 28 },
		{ 307, 349, 28 },
		{ 307, 350, 28 },
		{ 307, 351, 28 },
		{ 307, 352, 28 },
		{ 307, 353, 28 },
		{ 307, 354, 28 },
		{ 307, 355, 28 },
		{ 307, 356, 28 },
		{ 307, 357, 28 },
		{ 307, 358, 28 },
		{ 307, 359, 28 },
		{ 307, 360, 28 },
		{ 307, 361, 28 },
		{ 307, 362, 28 },
		{ 307, 363, 28 },
		{ 307, 364, 28 },
		{ 307, 365, 28 },
		{ 307, 366, 28 },
		{ 307, 367, 28 },
		{ 307, 368, 28 },
		{ 307, 369, 28 },
		{ 307, 370, 28 },
		{ 307, 371, 28 },
		{ 307, 372, 28 },
		{ 307, 373, 28 },
		{ 307, 374, 28 },
		{ 307, 375, 28 },
		{ 307, 376, 28 },
		{ 307, 377, 28 },
		{ 307, 378, 28 },
		{ 307, 379, 28 },
		{ 307, 380, 28 },
		{ 307, 381, 28 },
		{ 307, 382, 28 },
		{ 307, 383, 28 },
		{ 308, 326, 28 },
		{ 308, 327, 28 },
		{ 308, 328, 28 },
		{ 308, 329, 28 },
		{ 308, 330, 28 },
		{ 308, 331, 28 },
		{ 308, 332, 28 },
		{ 308, 333, 28 },
		{ 308, 334, 28 },
		{ 308, 335, 28 },
		{ 308, 336, 28 },
		{ 308, 337, 28 },
		{ 308, 338, 28 },
		{ 308, 339, 28 },
		{ 308, 340, 28 },
		{ 308, 341, 28 },
		{ 308, 342, 28 },
		{ 308, 343, 28 },
		{ 308, 344, 28 },
		{ 308, 345, 28 },
		{ 308, 346, 28 },
		{ 308, 347, 28 },
		{ 308, 348, 28 },
		{ 308, 349, 28 },
		{ 308, 350, 28 },
		{ 308, 351, 28 },
		{ 308, 352, 28 },
		{ 308, 353, 28 },
		{ 308, 354, 28 },
		{ 308, 355, 28 },
		{ 308, 356, 28 },
		{ 308, 357, 28 },
		{ 308, 358, 28 },
		{ 308, 359, 28 },
		{ 308, 360, 28 },
		{ 308, 361, 28 },
		{ 308, 362, 28 },
		{ 308, 363, 28 },
		{ 308, 364, 28 },
		{ 308, 365, 28 },
		{ 308, 366, 28 },
		{ 308, 367, 28 },
		{ 308, 368, 28 },
		{ 308, 369, 28 },
		{ 308, 370, 28 },
		{ 308, 371, 28 },
		{ 308, 372, 28 },
		{ 308, 373, 28 },
		{ 308, 374, 28 },
		{ 308, 375, 28 },
		{ 308, 376, 28 },
		{ 308, 377, 28 },
		{ 308, 378, 28 },
		{ 308, 379, 28 },
		{ 308, 380, 28 },
		{ 308, 381, 28 },
		{ 308, 382, 28 },
		{ 308, 383, 28 },
		{ 309, 327, 28 },
		{ 309, 328, 28 },
		{ 309, 329, 28 },
		{ 309, 330, 28 },
		{ 309, 331, 28 },
		{ 309, 332, 28 },
		{ 309, 333, 28 },
		{ 309, 334, 28 },
		{ 309, 335, 28 },
		{ 309, 336, 28 },
		{ 309, 337, 28 },
		{ 309, 338, 28 },
		{ 309, 339, 28 },
		{ 309, 340, 28 },
		{ 309, 341, 28 },
		{ 309, 342, 28 },
		{ 309, 343, 28 },
		{ 309, 344, 28 },
		{ 309, 345, 28 },
		{ 309, 346, 28 },
		{ 309, 347, 28 },
		{ 309, 348, 28 },
		{ 309, 349, 28 },
		{ 309, 350, 28 },
		{ 309, 351, 28 },
		{ 309, 352, 28 },
		{ 309, 353, 28 },
		{ 309, 354, 28 },
		{ 309, 355, 28 },
		{ 309, 356, 28 },
		{ 309, 357, 28 },
		{ 309, 358, 28 },
		{ 309, 359, 28 },
		{ 309, 360, 28 },
		{ 309, 361, 28 },
		{ 309, 362, 28 },
		{ 309, 363, 28 },
		{ 309, 364, 28 },
		{ 309, 365, 28 },
		{ 309, 366, 28 },
		{ 309, 367, 28 },
		{ 309, 368, 28 },
		{ 309, 369, 28 },
		{ 309, 370, 28 },
		{ 309, 371, 28 },
		{ 309, 372, 28 },
		{ 309, 373, 28 },
		{ 309, 374, 28 },
		{ 309, 375, 28 },
		{ 309, 376, 28 },
		{ 309, 377, 28 },
		{ 309, 378, 28 },
		{ 309, 379, 28 },
		{ 309, 380, 28 },
		{ 309, 381, 28 },
		{ 309, 382, 28 },
		{ 309, 383, 28 },
		{ 310, 328, 28 },
		{ 310, 329, 28 },
		{ 310, 330, 28 },
		{ 310, 331, 28 },
		{ 310, 332, 28 },
		{ 310, 333, 28 },
		{ 310, 334, 28 },
		{ 310, 335, 28 },
		{ 310, 336, 28 },
		{ 310, 337, 28 },
		{ 310, 338, 28 },
		{ 310, 339, 28 },
		{ 310, 340, 28 },
		{ 310, 341, 28 },
		{ 310, 342, 28 },
		{ 310, 343, 28 },
		{ 310, 344, 28 },
		{ 310, 345, 28 },
		{ 310, 346, 28 },
		{ 310, 347, 28 },
		{ 310, 348, 28 },
		{ 310, 349, 28 },
		{ 310, 350, 28 },
		{ 310, 351, 28 },
		{ 310, 352, 28 },
		{ 310, 353, 28 },
		{ 310, 354, 28 },
		{ 310, 355, 28 },
		{ 310, 356, 28 },
		{ 310, 357, 28 },
		{ 310, 358, 28 },
		{ 310, 359, 28 },
		{ 310, 360, 28 },
		{ 310, 361, 28 },
		{ 310, 362, 28 },
		{ 310, 363, 28 },
		{ 310, 364, 28 },
		{ 310, 365, 28 },
		{ 310, 366, 28 },
		{ 310, 367, 28 },
		{ 310, 368, 28 },
		{ 310, 369, 28 },
		{ 310, 370, 28 },
		{ 310, 371, 28 },
		{ 310, 372, 28 },
		{ 310, 373, 28 },
		{ 310, 374, 28 },
		{ 310, 375, 28 },
		{ 310, 376, 28 },
		{ 310, 377, 28 },
		{ 310, 378, 28 },
		{ 310, 379, 28 },
		{ 310, 380, 28 },
		{ 310, 381, 28 },
		{ 310, 382, 28 },
		{ 310, 383, 28 },
		{ 311, 330, 28 },
		{ 311, 331, 28 },
		{ 311, 332, 28 },
		{ 311, 333, 28 },
		{ 311, 334, 28 },
		{ 311, 335, 28 },
		{ 311, 336, 28 },
		{ 311, 337, 28 },
		{ 311, 338, 28 },
		{ 311, 339, 28 },
		{ 311, 340, 28 },
		{ 311, 341, 28 },
		{ 311, 342, 28 },
		{ 311, 343, 28 },
		{ 311, 344, 28 },
		{ 311, 345, 28 },
		{ 311, 346, 28 },
		{ 311, 347, 28 },
		{ 311, 348, 28 },
		{ 311, 349, 28 },
		{ 311, 350, 28 },
		{ 311, 351, 28 },
		{ 311, 352, 28 },
		{ 311, 353, 28 },
		{ 311, 354, 28 },
		{ 311, 355, 28 },
		{ 311, 356, 28 },
		{ 311, 357, 28 },
		{ 311, 358, 28 },
		{ 311, 359, 28 },
		{ 311, 360, 28 },
		{ 311, 361, 28 },
		{ 311, 362, 28 },
		{ 311, 363, 28 },
		{ 311, 364, 28 },
		{ 311, 365, 28 },
		{ 311, 366, 28 },
		{ 311, 367, 28 },
		{ 311, 368, 28 },
		{ 311, 369, 28 },
		{ 311, 370, 28 },
		{ 311, 371, 28 },
		{ 311, 372, 28 },
		{ 311, 373, 28 },
		{ 311, 374, 28 },
		{ 311, 375, 28 },
		{ 311, 376, 28 },
		{ 311, 377, 28 },
		{ 311, 378, 28 },
		{ 311, 379, 28 },
		{ 311, 380, 28 },
		{ 311, 381, 28 },
		{ 311, 382, 28 },
		{ 311, 383, 28 },
		{ 312, 331, 28 },
		{ 312, 332, 28 },
		{ 312, 333, 28 },
		{ 312, 334, 28 },
		{ 312, 335, 28 },
		{ 312, 336, 28 },
		{ 312, 337, 28 },
		{ 312, 338, 28 },
		{ 312, 339, 28 },
		{ 312, 340, 28 },
		{ 312, 341, 28 },
		{ 312, 342, 28 },
		{ 312, 343, 28 },
		{ 312, 344, 28 },
		{ 312, 345, 28 },
		{ 312, 346, 28 },
		{ 312, 347, 28 },
		{ 312, 348, 28 },
		{ 312, 349, 28 },
		{ 312, 350, 28 },
		{ 312, 351, 28 },
		{ 312, 352, 28 },
		{ 312, 353, 28 },
		{ 312, 354, 28 },
		{ 312, 355, 28 },
		{ 312, 356, 28 },
		{ 312, 357, 28 },
		{ 312, 358, 28 },
		{ 312, 359, 28 },
		{ 312, 360, 28 },
		{ 312, 361, 28 },
		{ 312, 362, 28 },
		{ 312, 363, 28 },
		{ 312, 364, 28 },
		{ 312, 365, 28 },
		{ 312, 366, 28 },
		{ 312, 367, 28 },
		{ 312, 368, 28 },
		{ 312, 369, 28 },
		{ 312, 370, 28 },
		{ 312, 371, 28 },
		{ 312, 372, 28 },
		{ 312, 373, 28 },
		{ 312, 374, 28 },
		{ 312, 375, 28 },
		{ 312, 376, 28 },
		{ 312, 377, 28 },
		{ 312, 378, 28 },
		{ 312, 379, 28 },
		{ 312, 380, 28 },
		{ 312, 381, 28 },
		{ 312, 382, 28 },
		{ 312, 383, 28 },
		{ 313, 332, 28 },
		{ 313, 333, 28 },
		{ 313, 334, 28 },
		{ 313, 335, 28 },
		{ 313, 336, 28 },
		{ 313, 337, 28 },
		{ 313, 338, 28 },
		{ 313, 339, 28 },
		{ 313, 340, 28 },
		{ 313, 341, 28 },
		{ 313, 342, 28 },
		{ 313, 343, 28 },
		{ 313, 344, 28 },
		{ 313, 345, 28 },
		{ 313, 346, 28 },
		{ 313, 347, 28 },
		{ 313, 348, 28 },
		{ 313, 349, 28 },
		{ 313, 350, 28 },
		{ 313, 351, 28 },
		{ 313, 352, 28 },
		{ 313, 353, 28 },
		{ 313, 354, 28 },
		{ 313, 355, 28 },
		{ 313, 356, 28 },
		{ 313, 357, 28 },
		{ 313, 358, 28 },
		{ 313, 359, 28 },
		{ 313, 360, 28 },
		{ 313, 361, 28 },
		{ 313, 362, 28 },
		{ 313, 363, 28 },
		{ 313, 364, 28 },
		{ 313, 365, 28 },
		{ 313, 366, 28 },
		{ 313, 367, 28 },
		{ 313, 368, 28 },
		{ 313, 369, 28 },
		{ 313, 370, 28 },
		{ 313, 371, 28 },
		{ 313, 372, 28 },
		{ 313, 373, 28 },
		{ 313, 374, 28 },
		{ 313, 375, 28 },
		{ 313, 376, 28 },
		{ 313, 377, 28 },
		{ 313, 378, 28 },
		{ 313, 379, 28 },
		{ 313, 380, 28 },
		{ 313, 381, 28 },
		{ 313, 382, 28 },
		{ 313, 383, 28 },
		{ 314, 334, 28 },
		{ 314, 335, 28 },
		{ 314, 336, 28 },
		{ 314, 337, 28 },
		{ 314, 338, 28 },
		{ 314, 339, 28 },
		{ 314, 340, 28 },
		{ 314, 341, 28 },
		{ 314, 342, 28 },
		{ 314, 343, 28 },
		{ 314, 344, 28 },
		{ 314, 345, 28 },
		{ 314, 346, 28 },
		{ 314, 347, 28 },
		{ 314, 348, 28 },
		{ 314, 349, 28 },
		{ 314, 350, 28 },
		{ 314, 351, 28 },
		{ 314, 352, 28 },
		{ 314, 353, 28 },
		{ 314, 354, 28 },
		{ 314, 355, 28 },
		{ 314, 356, 28 },
		{ 314, 357, 28 },
		{ 314, 358, 28 },
		{ 314, 359, 28 },
		{ 314, 360, 28 },
		{ 314, 361, 28 },
		{ 314, 362, 28 },
		{ 314, 363, 28 },
		{ 314, 364, 28 },
		{ 314, 365, 28 },
		{ 314, 366, 28 },
		{ 314, 367, 28 },
		{ 314, 368, 28 },
		{ 314, 369, 28 },
		{ 314, 370, 28 },
		{ 314, 371, 28 },
		{ 314, 372, 28 },
		{ 314, 373, 28 },
		{ 314, 374, 28 },
		{ 314, 375, 28 },
		{ 314, 376, 28 },
		{ 314, 377, 28 },
		{ 314, 378, 28 },
		{ 314, 379, 28 },
		{ 314, 380, 28 },
		{ 314, 381, 28 },
		{ 314, 382, 28 },
		{ 314, 383, 28 },
		{ 315, 335, 28 },
		{ 315, 336, 28 },
		{ 315, 337, 28 },
		{ 315, 338, 28 },
		{ 315, 339, 28 },
		{ 315, 340, 28 },
		{ 315, 341, 28 },
		{ 315, 342, 28 },
		{ 315, 343, 28 },
		{ 315, 344, 28 },
		{ 315, 345, 28 },
		{ 315, 346, 28 },
		{ 315, 347, 28 },
		{ 315, 348, 28 },
		{ 315, 349, 28 },
		{ 315, 350, 28 },
		{ 315, 351, 28 },
		{ 315, 352, 28 },
		{ 315, 353, 28 },
		{ 315, 354, 28 },
		{ 315, 355, 28 },
		{ 315, 356, 28 },
		{ 315, 357, 28 },
		{ 315, 358, 28 },
		{ 315, 359, 28 },
		{ 315, 360, 28 },
		{ 315, 361, 28 },
		{ 315, 362, 28 },
		{ 315, 363, 28 },
		{ 315, 364, 28 },
		{ 315, 365, 28 },
		{ 315, 366, 28 },
		{ 315, 367, 28 },
		{ 315, 368, 28 },
		{ 315, 369, 28 },
		{ 315, 370, 28 },
		{ 315, 371, 28 },
		{ 315, 372, 28 },
		{ 315, 373, 28 },
		{ 315, 374, 28 },
		{ 315, 375, 28 },
		{ 315, 376, 28 },
		{ 315, 377, 28 },
		{ 315, 378, 28 },
		{ 315, 379, 28 },
		{ 315, 380, 28 },
		{ 315, 381, 28 },
		{ 315, 382, 28 },
		{ 315, 383, 28 },
		{ 316, 336, 28 },
		{ 316, 337, 28 },
		{ 316, 338, 28 },
		{ 316, 339, 28 },
		{ 316, 340, 28 },
		{ 316, 341, 28 },
		{ 316, 342, 28 },
		{ 316, 343, 28 },
		{ 316, 344, 28 },
		{ 316, 345, 28 },
		{ 316, 346, 28 },
		{ 316, 347, 28 },
		{ 316, 348, 28 },
		{ 316, 349, 28 },
		{ 316, 350, 28 },
		{ 316, 351, 28 },
		{ 316, 352, 28 },
		{ 316, 353, 28 },
		{ 316, 354, 28 },
		{ 316, 355, 28 },
		{ 316, 356, 28 },
		{ 316, 357, 28 },
		{ 316, 358, 28 },
		{ 316, 359, 28 },
		{ 316, 360, 28 },
		{ 316, 361, 28 },
		{ 316, 362, 28 },
		{ 316, 363, 28 },
		{ 316, 364, 28 },
		{ 316, 365, 28 },
		{ 316, 366, 28 },
		{ 316, 367, 28 },
		{ 316, 368, 28 },
		{ 316, 369, 28 },
		{ 316, 370, 28 },
		{ 316, 371, 28 },
		{ 316, 372, 28 },
		{ 316, 373, 28 },
		{ 316, 374, 28 },
		{ 316, 375, 28 },
		{ 316, 376, 28 },
		{ 316, 377, 28 },
		{ 316, 378, 28 },
		{ 316, 379, 28 },
		{ 316, 380, 28 },
		{ 316, 381, 28 },
		{ 316, 382, 28 },
		{ 316, 383, 28 },
		{ 317, 338, 28 },
		{ 317, 339, 28 },
		{ 317, 340, 28 },
		{ 317, 341, 28 },
		{ 317, 342, 28 },
		{ 317, 343, 28 },
		{ 317, 344, 28 },
		{ 317, 345, 28 },
		{ 317, 346, 28 },
		{ 317, 347, 28 },
		{ 317, 348, 28 },
		{ 317, 349, 28 },
		{ 317, 350, 28 },
		{ 317, 351, 28 },
		{ 317, 352, 28 },
		{ 317, 353, 28 },
		{ 317, 354, 28 },
		{ 317, 355, 28 },
		{ 317, 356, 28 },
		{ 317, 357, 28 },
		{ 317, 358, 28 },
		{ 317, 359, 28 },
		{ 317, 360, 28 },
		{ 317, 361, 28 },
		{ 317, 362, 28 },
		{ 317, 363, 28 },
		{ 317, 364, 28 },
		{ 317, 365, 28 },
		{ 317, 366, 28 },
		{ 317, 367, 28 },
		{ 317, 368, 28 },
		{ 317, 369, 28 },
		{ 317, 370, 28 },
		{ 317, 371, 28 },
		{ 317, 372, 28 },
		{ 317, 373, 28 },
		{ 317, 374, 28 },
		{ 317, 375, 28 },
		{ 317, 376, 28 },
		{ 317, 377, 28 },
		{ 317, 378, 28 },
		{ 317, 379, 28 },
		{ 317, 380, 28 },
		{ 317, 381, 28 },
		{ 317, 382, 28 },
		{ 317, 383, 28 },
		{ 318, 339, 28 },
		{ 318, 340, 28 },
		{ 318, 341, 28 },
		{ 318, 342, 28 },
		{ 318, 343, 28 },
		{ 318, 344, 28 },
		{ 318, 345, 28 },
		{ 318, 346, 28 },
		{ 318, 347, 28 },
		{ 318, 348, 28 },
		{ 318, 349, 28 },
		{ 318, 350, 28 },
		{ 318, 351, 28 },
		{ 318, 352, 28 },
		{ 318, 353, 28 },
		{ 318, 354, 28 },
		{ 318, 355, 28 },
		{ 318, 356, 28 },
		{ 318, 357, 28 },
		{ 318, 358, 28 },
		{ 318, 359, 28 },
		{ 318, 360, 28 },
		{ 318, 361, 28 },
		{ 318, 362, 28 },
		{ 318, 363, 28 },
		{ 318, 364, 28 },
		{ 318, 365, 28 },
		{ 318, 366, 28 },
		{ 318, 367, 28 },
		{ 318, 368, 28 },
		{ 318, 369, 28 },
		{ 318, 370, 28 },
		{ 318, 371, 28 },
		{ 318, 372, 28 },
		{ 318, 373, 28 },
		{ 318, 374, 28 },
		{ 318, 375, 28 },
		{ 318, 376, 28 },
		{ 318, 377, 28 },
		{ 318, 378, 28 },
		{ 318, 379, 28 },
		{ 318, 380, 28 },
		{ 318, 381, 28 },
		{ 318, 382, 28 },
		{ 318, 383, 28 },
		{ 319, 340, 28 },
		{ 319, 341, 28 },
		{ 319, 342, 28 },
		{ 319, 343, 28 },
		{ 319, 344, 28 },
		{ 319, 345, 28 },
		{ 319, 346, 28 },
		{ 319, 347, 28 },
		{ 319, 348, 28 },
		{ 319, 349, 28 },
		{ 319, 350, 28 },
		{ 319, 351, 28 },
		{ 319, 352, 28 },
		{ 319, 353, 28 },
		{ 319, 354, 28 },
		{ 319, 355, 28 },
		{ 319, 356, 28 },
		{ 319, 357, 28 },
		{ 319, 358, 28 },
		{ 319, 359, 28 },
		{ 319, 360, 28 },
		{ 319, 361, 28 },
		{ 319, 362, 28 },
		{ 319, 363, 28 },
		{ 319, 364, 28 },
		{ 319, 365, 28 },
		{ 319, 366, 28 },
		{ 319, 367, 28 },
		{ 319, 368, 28 },
		{ 319, 369, 28 },
		{ 319, 370, 28 },
		{ 319, 371, 28 },
		{ 319, 372, 28 },
		{ 319, 373, 28 },
		{ 319, 374, 28 },
		{ 319, 375, 28 },
		{ 319, 376, 28 },
		{ 319, 377, 28 },
		{ 319, 378, 28 },
		{ 319, 379, 28 },
		{ 319, 380, 28 },
		{ 319, 381, 28 },
		{ 319, 382, 28 },
		{ 319, 383, 28 },
		{ 320, 342, 28 },
		{ 320, 343, 28 },
		{ 320, 344, 28 },
		{ 320, 345, 28 },
		{ 320, 346, 28 },
		{ 320, 347, 28 },
		{ 320, 348, 28 },
		{ 320, 349, 28 },
		{ 320, 350, 28 },
		{ 320, 351, 28 },
		{ 320, 352, 28 },
		{ 320, 353, 28 },
		{ 320, 354, 28 },
		{ 320, 355, 28 },
		{ 320, 356, 28 },
		{ 320, 357, 28 },
		{ 320, 358, 28 },
		{ 320, 359, 28 },
		{ 320, 360, 28 },
		{ 320, 361, 28 },
		{ 320, 362, 28 },
		{ 320, 363, 28 },
		{ 320, 364, 28 },
		{ 320, 365, 28 },
		{ 320, 366, 28 },
		{ 320, 367, 28 },
		{ 320, 368, 28 },
		{ 320, 369, 28 },
		{ 320, 370, 28 },
		{ 320, 371, 28 },
		{ 320, 372, 28 },
		{ 320, 373, 28 },
		{ 320, 374, 28 },
		{ 320, 375, 28 },
		{ 320, 376, 28 },
		{ 320, 377, 28 },
		{ 320, 378, 28 },
		{ 320, 379, 28 },
		{ 320, 380, 28 },
		{ 320, 381, 28 },
		{ 320, 382, 28 },
		{ 320, 383, 28 },
		{ 321, 343, 28 },
		{ 321, 344, 28 },
		{ 321, 345, 28 },
		{ 321, 346, 28 },
		{ 321, 347, 28 },
		{ 321, 348, 28 },
		{ 321, 349, 28 },
		{ 321, 350, 28 },
		{ 321, 351, 28 },
		{ 321, 352, 28 },
		{ 321, 353, 28 },
		{ 321, 354, 28 },
		{ 321, 355, 28 },
		{ 321, 356, 28 },
		{ 321, 357, 28 },
		{ 321, 358, 28 },
		{ 321, 359, 28 },
		{ 321, 360, 28 },
		{ 321, 361, 28 },
		{ 321, 362, 28 },
		{ 321, 363, 28 },
		{ 321, 364, 28 },
		{ 321, 365, 28 },
		{ 321, 366, 28 },
		{ 321, 367, 28 },
		{ 321, 368, 28 },
		{ 321, 369, 28 },
		{ 321, 370, 28 },
		{ 321, 371, 28 },
		{ 321, 372, 28 },
		{ 321, 373, 28 },
		{ 321, 374, 28 },
		{ 321, 375, 28 },
		{ 321, 376, 28 },
		{ 321, 377, 28 },
		{ 321, 378, 28 },
		{ 321, 379, 28 },
		{ 321, 380, 28 },
		{ 321, 381, 28 },
		{ 321, 382, 28 },
		{ 321, 383, 28 },
		{ 322, 344, 28 },
		{ 322, 345, 28 },
		{ 322, 346, 28 },
		{ 322, 347, 28 },
		{ 322, 348, 28 },
		{ 322, 349, 28 },
		{ 322, 350, 28 },
		{ 322, 351, 28 },
		{ 322, 352, 28 },
		{ 322, 353, 28 },
		{ 322, 354, 28 },
		{ 322, 355, 28 },
		{ 322, 356, 28 },
		{ 322, 357, 28 },
		{ 322, 358, 28 },
		{ 322, 359, 28 },
		{ 322, 360, 28 },
		{ 322, 361, 28 },
		{ 322, 362, 28 },
		{ 322, 363, 28 },
		{ 322, 364, 28 },
		{ 322, 365, 28 },
		{ 322, 366, 28 },
		{ 322, 367, 28 },
		{ 322, 368, 28 },
		{ 322, 369, 28 },
		{ 322, 370, 28 },
		{ 322, 371, 28 },
		{ 322, 372, 28 },
		{ 322, 373, 28 },
		{ 322, 374, 28 },
		{ 322, 375, 28 },
		{ 322, 376, 28 },
		{ 322, 377, 28 },
		{ 322, 378, 28 },
		{ 322, 379, 28 },
		{ 322, 380, 28 },
		{ 322, 381, 28 },
		{ 322, 382, 28 },
		{ 322, 383, 28 },
		{ 323, 346, 28 },
		{ 323, 347, 28 },
		{ 323, 348, 28 },
		{ 323, 349, 28 },
		{ 323, 350, 28 },
		{ 323, 351, 28 },
		{ 323, 352, 28 },
		{ 323, 353, 28 },
		{ 323, 354, 28 },
		{ 323, 355, 28 },
		{ 323, 356, 28 },
		{ 323, 357, 28 },
		{ 323, 358, 28 },
		{ 323, 359, 28 },
		{ 323, 360, 28 },
		{ 323, 361, 28 },
		{ 323, 362, 28 },
		{ 323, 363, 28 },
		{ 323, 364, 28 },
		{ 323, 365, 28 },
		{ 323, 366, 28 },
		{ 323, 367, 28 },
		{ 323, 368, 28 },
		{ 323, 369, 28 },
		{ 323, 370, 28 },
		{ 323, 371, 28 },
		{ 323, 372, 28 },
		{ 323, 373, 28 },
		{ 323, 374, 28 },
		{ 323, 375, 28 },
		{ 323, 376, 28 },
		{ 323, 377, 28 },
		{ 323, 378, 28 },
		{ 323, 379, 28 },
		{ 323, 380, 28 },
		{ 323, 381, 28 },
		{ 323, 382, 28 },
		{ 323, 383, 28 },
		{ 324, 347, 28 },
		{ 324, 348, 28 },
		{ 324, 349, 28 },
		{ 324, 350, 28 },
		{ 324, 351, 28 },
		{ 324, 352, 28 },
		{ 324, 353, 28 },
		{ 324, 354, 28 },
		{ 324, 355, 28 },
		{ 324, 356, 28 },
		{ 324, 357, 28 },
		{ 324, 358, 28 },
		{ 324, 359, 28 },
		{ 324, 360, 28 },
		{ 324, 361, 28 },
		{ 324, 362, 28 },
		{ 324, 363, 28 },
		{ 324, 364, 28 },
		{ 324, 365, 28 },
		{ 324, 366, 28 },
		{ 324, 367, 28 },
		{ 324, 368, 28 },
		{ 324, 369, 28 },
		{ 324, 370, 28 },
		{ 324, 371, 28 },
		{ 324, 372, 28 },
		{ 324, 373, 28 },
		{ 324, 374, 28 },
		{ 324, 375, 28 },
		{ 324, 376, 28 },
		{ 324, 377, 28 },
		{ 324, 378, 28 },
		{ 324, 379, 28 },
		{ 324, 380, 28 },
		{ 324, 381, 28 },
		{ 324, 382, 28 },
		{ 324, 383, 28 },
		{ 325, 348, 28 },
		{ 325, 349, 28 },
		{ 325, 350, 28 },
		{ 325, 351, 28 },
		{ 325, 352, 28 },
		{ 325, 353, 28 },
		{ 325, 354, 28 },
		{ 325, 355, 28 },
		{ 325, 356, 28 },
		{ 325, 357, 28 },
		{ 325, 358, 28 },
		{ 325, 359, 28 },
		{ 325, 360, 28 },
		{ 325, 361, 28 },
		{ 325, 362, 28 },
		{ 325, 363, 28 },
		{ 325, 364, 28 },
		{ 325, 365, 28 },
		{ 325, 366, 28 },
		{ 325, 367, 28 },
		{ 325, 368, 28 },
		{ 325, 369, 28 },
		{ 325, 370, 28 },
		{ 325, 371, 28 },
		{ 325, 372, 28 },
		{ 325, 373, 28 },
		{ 325, 374, 28 },
		{ 325, 375, 28 },
		{ 325, 376, 28 },
		{ 325, 377, 28 },
		{ 325, 378, 28 },
		{ 325, 379, 28 },
		{ 325, 380, 28 },
		{ 325, 381, 28 },
		{ 325, 382, 28 },
		{ 325, 383, 28 },
		{ 326, 350, 28 },
		{ 326, 351, 28 },
		{ 326, 352, 28 },
		{ 326, 353, 28 },
		{ 326, 354, 28 },
		{ 326, 355, 28 },
		{ 326, 356, 28 },
		{ 326, 357, 28 },
		{ 326, 358, 28 },
		{ 326, 359, 28 },
		{ 326, 360, 28 },
		{ 326, 361, 28 },
		{ 326, 362, 28 },
		{ 326, 363, 28 },
		{ 326, 364, 28 },
		{ 326, 365, 28 },
		{ 326, 366, 28 },
		{ 326, 367, 28 },
		{ 326, 368, 28 },
		{ 326, 369, 28 },
		{ 326, 370, 28 },
		{ 326, 371, 28 },
		{ 326, 372, 28 },
		{ 326, 373, 28 },
		{ 326, 374, 28 },
		{ 326, 375, 28 },
		{ 326, 376, 28 },
		{ 326, 377, 28 },
		{ 326, 378, 28 },
		{ 326, 379, 28 },
		{ 326, 380, 28 },
		{ 326, 381, 28 },
		{ 326, 382, 28 },
		{ 326, 383, 28 },
		{ 327, 351, 28 },
		{ 327, 352, 28 },
		{ 327, 353, 28 },
		{ 327, 354, 28 },
		{ 327, 355, 28 },
		{ 327, 356, 28 },
		{ 327, 357, 28 },
		{ 327, 358, 28 },
		{ 327, 359, 28 },
		{ 327, 360, 28 },
		{ 327, 361, 28 },
		{ 327, 362, 28 },
		{ 327, 363, 28 },
		{ 327, 364, 28 },
		{ 327, 365, 28 },
		{ 327, 366, 28 },
		{ 327, 367, 28 },
		{ 327, 368, 28 },
		{ 327, 369, 28 },
		{ 327, 370, 28 },
		{ 327, 371, 28 },
		{ 327, 372, 28 },
		{ 327, 373, 28 },
		{ 327, 374, 28 },
		{ 327, 375, 28 },
		{ 327, 376, 28 },
		{ 327, 377, 28 },
		{ 327, 378, 28 },
		{ 327, 379, 28 },
		{ 327, 380, 28 },
		{ 327, 381, 28 },
		{ 327, 382, 28 },
		{ 327, 383, 28 },
		{ 328, 352, 28 },
		{ 328, 353, 28 },
		{ 328, 354, 28 },
		{ 328, 355, 28 },
		{ 328, 356, 28 },
		{ 328, 357, 28 },
		{ 328, 358, 28 },
		{ 328, 359, 28 },
		{ 328, 360, 28 },
		{ 328, 361, 28 },
		{ 328, 362, 28 },
		{ 328, 363, 28 },
		{ 328, 364, 28 },
		{ 328, 365, 28 },
		{ 328, 366, 28 },
		{ 328, 367, 28 },
		{ 328, 368, 28 },
		{ 328, 369, 28 },
		{ 328, 370, 28 },
		{ 328, 371, 28 },
		{ 328, 372, 28 },
		{ 328, 373, 28 },
		{ 328, 374, 28 },
		{ 328, 375, 28 },
		{ 328, 376, 28 },
		{ 328, 377, 28 },
		{ 328, 378, 28 },
		{ 328, 379, 28 },
		{ 328, 380, 28 },
		{ 328, 381, 28 },
		{ 328, 382, 28 },
		{ 328, 383, 28 },
		{ 329, 354, 28 },
		{ 329, 355, 28 },
		{ 329, 356, 28 },
		{ 329, 357, 28 },
		{ 329, 358, 28 },
		{ 329, 359, 28 },
		{ 329, 360, 28 },
		{ 329, 361, 28 },
		{ 329, 362, 28 },
		{ 329, 363, 28 },
		{ 329, 364, 28 },
		{ 329, 365, 28 },
		{ 329, 366, 28 },
		{ 329, 367, 28 },
		{ 329, 368, 28 },
		{ 329, 369, 28 },
		{ 329, 370, 28 },
		{ 329, 371, 28 },
		{ 329, 372, 28 },
		{ 329, 373, 28 },
		{ 329, 374, 28 },
		{ 329, 375, 28 },
		{ 329, 376, 28 },
		{ 329, 377, 28 },
		{ 329, 378, 28 },
		{ 329, 379, 28 },
		{ 329, 380, 28 },
		{ 329, 381, 28 },
		{ 329, 382, 28 },
		{ 329, 383, 28 },
		{ 330, 355, 28 },
		{ 330, 356, 28 },
		{ 330, 357, 28 },
		{ 330, 358, 28 },
		{ 330, 359, 28 },
		{ 330, 360, 28 },
		{ 330, 361, 28 },
		{ 330, 362, 28 },
		{ 330, 363, 28 },
		{ 330, 364, 28 },
		{ 330, 365, 28 },
		{ 330, 366, 28 },
		{ 330, 367, 28 },
		{ 330, 368, 28 },
		{ 330, 369, 28 },
		{ 330, 370, 28 },
		{ 330, 371, 28 },
		{ 330, 372, 28 },
		{ 330, 373, 28 },
		{ 330, 374, 28 },
		{ 330, 375, 28 },
		{ 330, 376, 28 },
		{ 330, 377, 28 },
		{ 330, 378, 28 },
		{ 330, 379, 28 },
		{ 330, 380, 28 },
		{ 330, 381, 28 },
		{ 330, 382, 28 },
		{ 330, 383, 28 },
		{ 331, 356, 28 },
		{ 331, 357, 28 },
		{ 331, 358, 28 },
		{ 331, 359, 28 },
		{ 331, 360, 28 },
		{ 331, 361, 28 },
		{ 331, 362, 28 },
		{ 331, 363, 28 },
		{ 331, 364, 28 },
		{ 331, 365, 28 },
		{ 331, 366, 28 },
		{ 331, 367, 28 },
		{ 331, 368, 28 },
		{ 331, 369, 28 },
		{ 331, 370, 28 },
		{ 331, 371, 28 },
		{ 331, 372, 28 },
		{ 331, 373, 28 },
		{ 331, 374, 28 },
		{ 331, 375, 28 },
		{ 331, 376, 28 },
		{ 331, 377, 28 },
		{ 331, 378, 28 },
		{ 331, 379, 28 },
		{ 331, 380, 28 },
		{ 331, 381, 28 },
		{ 331, 382, 28 },
		{ 331, 383, 28 },
		{ 332, 358, 28 },
		{ 332, 359, 28 },
		{ 332, 360, 28 },
		{ 332, 361, 28 },
		{ 332, 362, 28 },
		{ 332, 363, 28 },
		{ 332, 364, 28 },
		{ 332, 365, 28 },
		{ 332, 366, 28 },
		{ 332, 367, 28 },
		{ 332, 368, 28 },
		{ 332, 369, 28 },
		{ 332, 370, 28 },
		{ 332, 371, 28 },
		{ 332, 372, 28 },
		{ 332, 373, 28 },
		{ 332, 374, 28 },
		{ 332, 375, 28 },
		{ 332, 376, 28 },
		{ 332, 377, 28 },
		{ 332, 378, 28 },
		{ 332, 379, 28 },
		{ 332, 380, 28 },
		{ 332, 381, 28 },
		{ 332, 382, 28 },
		{ 332, 383, 28 },
		{ 333, 359, 28 },
		{ 333, 360, 28 },
		{ 333, 361, 28 },
		{ 333, 362, 28 },
		{ 333, 363, 28 },
		{ 333, 364, 28 },
		{ 333, 365, 28 },
		{ 333, 366, 28 },
		{ 333, 367, 28 },
		{ 333, 368, 28 },
		{ 333, 369, 28 },
		{ 333, 370, 28 },
		{ 333, 371, 28 },
		{ 333, 372, 28 },
		{ 333, 373, 28 },
		{ 333, 374, 28 },
		{ 333, 375, 28 },
		{ 333, 376, 28 },
		{ 333, 377, 28 },
		{ 333, 378, 28 },
		{ 333, 379, 28 },
		{ 333, 380, 28 },
		{ 333, 381, 28 },
		{ 333, 382, 28 },
		{ 333, 383, 28 },
		{ 334, 360, 28 },
		{ 334, 361, 28 },
		{ 334, 362, 28 },
		{ 334, 363, 28 },
		{ 334, 364, 28 },
		{ 334, 365, 28 },
		{ 334, 366, 28 },
		{ 334, 367, 28 },
		{ 334, 368, 28 },
		{ 334, 369, 28 },
		{ 334, 370, 28 },
		{ 334, 371, 28 },
		{ 334, 372, 28 },
		{ 334, 373, 28 },
		{ 334, 374, 28 },
		{ 334, 375, 28 },
		{ 334, 376, 28 },
		{ 334, 377, 28 },
		{ 334, 378, 28 },
		{ 334, 379, 28 },
		{ 334, 380, 28 },
		{ 334, 381, 28 },
		{ 334, 382, 28 },
		{ 334, 383, 28 },
		{ 335, 362, 28 },
		{ 335, 363, 28 },
		{ 335, 364, 28 },
		{ 335, 365, 28 },
		{ 335, 366, 28 },
		{ 335, 367, 28 },
		{ 335, 368, 28 },
		{ 335, 369, 28 },
		{ 335, 370, 28 },
		{ 335, 371, 28 },
		{ 335, 372, 28 },
		{ 335, 373, 28 },
		{ 335, 374, 28 },
		{ 335, 375, 28 },
		{ 335, 376, 28 },
		{ 335, 377, 28 },
		{ 335, 378, 28 },
		{ 335, 379, 28 },
		{ 335, 380, 28 },
		{ 335, 381, 28 },
		{ 335, 382, 28 },
		{ 335, 383, 28 },
		{ 336, 363, 28 },
		{ 336, 364, 28 },
		{ 336, 365, 28 },
		{ 336, 366, 28 },
		{ 336, 367, 28 },
		{ 336, 368, 28 },
		{ 336, 369, 28 },
		{ 336, 370, 28 },
		{ 336, 371, 28 },
		{ 336, 372, 28 },
		{ 336, 373, 28 },
		{ 336, 374, 28 },
		{ 336, 375, 28 },
		{ 336, 376, 28 },
		{ 336, 377, 28 },
		{ 336, 378, 28 },
		{ 336, 379, 28 },
		{ 336, 380, 28 },
		{ 336, 381, 28 },
		{ 336, 382, 28 },
		{ 336, 383, 28 },
		{ 337, 364, 28 },
		{ 337, 365, 28 },
		{ 337, 366, 28 },
		{ 337, 367, 28 },
		{ 337, 368, 28 },
		{ 337, 369, 28 },
		{ 337, 370, 28 },
		{ 337, 371, 28 },
		{ 337, 372, 28 },
		{ 337, 373, 28 },
		{ 337, 374, 28 },
		{ 337, 375, 28 },
		{ 337, 376, 28 },
		{ 337, 377, 28 },
		{ 337, 378, 28 },
		{ 337, 379, 28 },
		{ 337, 380, 28 },
		{ 337, 381, 28 },
		{ 337, 382, 28 },
		{ 337, 383, 28 },
		{ 338, 366, 28 },
		{ 338, 367, 28 },
		{ 338, 368, 28 },
		{ 338, 369, 28 },
		{ 338, 370, 28 },
		{ 338, 371, 28 },
		{ 338, 372, 28 },
		{ 338, 373, 28 },
		{ 338, 374, 28 },
		{ 338, 375, 28 },
		{ 338, 376, 28 },
		{ 338, 377, 28 },
		{ 338, 378, 28 },
		{ 338, 379, 28 },
		{ 338, 380, 28 },
		{ 338, 381, 28 },
		{ 338, 382, 28 },
		{ 338, 383, 28 },
		{ 339, 367, 28 },
		{ 339, 368, 28 },
		{ 339, 369, 28 },
		{ 339, 370, 28 },
		{ 339, 371, 28 },
		{ 339, 372, 28 },
		{ 339, 373, 28 },
		{ 339, 374, 28 },
		{ 339, 375, 28 },
		{ 339, 376, 28 },
		{ 339, 377, 28 },
		{ 339, 378, 28 },
		{ 339, 379, 28 },
		{ 339, 380, 28 },
		{ 339, 381, 28 },
		{ 339, 382, 28 },
		{ 339, 383, 28 },
		{ 340, 368, 28 },
		{ 340, 369, 28 },
		{ 340, 370, 28 },
		{ 340, 371, 28 },
		{ 340, 372, 28 },
		{ 340, 373, 28 },
		{ 340, 374, 28 },
		{ 340, 375, 28 },
		{ 340, 376, 28 },
		{ 340, 377, 28 },
		{ 340, 378, 28 },
		{ 340, 379, 28 },
		{ 340, 380, 28 },
		{ 340, 381, 28 },
		{ 340, 382, 28 },
		{ 340, 383, 28 },
		{ 341, 370, 28 },
		{ 341, 371, 28 },
		{ 341, 372, 28 },
		{ 341, 373, 28 },
		{ 341, 374, 28 },
		{ 341, 375, 28 },
		{ 341, 376, 28 },
		{ 341, 377, 28 },
		{ 341, 378, 28 },
		{ 341, 379, 28 },
		{ 341, 380, 28 },
		{ 341, 381, 28 },
		{ 341, 382, 28 },
		{ 341, 383, 28 },
		{ 342, 371, 28 },
		{ 342, 372, 28 },
		{ 342, 373, 28 },
		{ 342, 374, 28 },
		{ 342, 375, 28 },
		{ 342, 376, 28 },
		{ 342, 377, 28 },
		{ 342, 378, 28 },
		{ 342, 379, 28 },
		{ 342, 380, 28 },
		{ 342, 381, 28 },
		{ 342, 382, 28 },
		{ 342, 383, 28 },
		{ 343, 372, 28 },
		{ 343, 373, 28 },
		{ 343, 374, 28 },
		{ 343, 375, 28 },
		{ 343, 376, 28 },
		{ 343, 377, 28 },
		{ 343, 378, 28 },
		{ 343, 379, 28 },
		{ 343, 380, 28 },
		{ 343, 381, 28 },
		{ 343, 382, 28 },
		{ 343, 383, 28 },
		{ 344, 374, 28 },
		{ 344, 375, 28 },
		{ 344, 376, 28 },
		{ 344, 377, 28 },
		{ 344, 378, 28 },
		{ 344, 379, 28 },
		{ 344, 380, 28 },
		{ 344, 381, 28 },
		{ 344, 382, 28 },
		{ 344, 383, 28 },
		{ 345, 375, 28 },
		{ 345, 376, 28 },
		{ 345, 377, 28 },
		{ 345, 378, 28 },
		{ 345, 379, 28 },
		{ 345, 380, 28 },
		{ 345, 381, 28 },
		{ 345, 382, 28 },
		{ 346, 377, 28 },
		{ 346, 378, 28 },
		{ 346, 379, 28 },
		{ 346, 380, 28 },
		{ 346, 381, 28 },
		{ 346, 382, 28 },
		{ 347, 378, 28 },
		{ 347, 379, 28 },
		{ 347, 380, 28 },
		{ 347, 381, 28 },
		{ 347, 382, 28 },
		{ 348, 380, 28 },
		{ 348, 381, 28 },
		{ 197, 133, 29 },
		{ 197, 134, 29 },
		{ 197, 135, 29 },
		{ 198, 132, 29 },
		{ 198, 133, 29 },
		{ 198, 134, 29 },
		{ 198, 135, 29 },
		{ 198, 136, 29 },
		{ 198, 137, 29 },
		{ 198, 138, 29 },
		{ 199, 132, 29 },
		{ 199, 133, 29 },
		{ 199, 134, 29 },
		{ 199, 135, 29 },
		{ 199, 136, 29 },
		{ 199, 137, 29 },
		{ 199, 138, 29 },
		{ 199, 139, 29 },
		{ 199, 140, 29 },
		{ 200, 132, 29 },
		{ 200, 133, 29 },
		{ 200, 134, 29 },
		{ 200, 135, 29 },
		{ 200, 136, 29 },
		{ 200, 137, 29 },
		{ 200, 138, 29 },
		{ 200, 139, 29 },
		{ 200, 140, 29 },
		{ 200, 141, 29 },
		{ 200, 142, 29 },
		{ 201, 131, 29 },
		{ 201, 132, 29 },
		{ 201, 133, 29 },
		{ 201, 134, 29 },
		{ 201, 135, 29 },
		{ 201, 136, 29 },
		{ 201, 137, 29 },
		{ 201, 138, 29 },
		{ 201, 139, 29 },
		{ 201, 140, 29 },
		{ 201, 141, 29 },
		{ 201, 142, 29 },
		{ 201, 143, 29 },
		{ 201, 144, 29 },
		{ 202, 131, 29 },
		{ 202, 132, 29 },
		{ 202, 133, 29 },
		{ 202, 134, 29 },
		{ 202, 135, 29 },
		{ 202, 136, 29 },
		{ 202, 137, 29 },
		{ 202, 138, 29 },
		{ 202, 139, 29 },
		{ 202, 140, 29 },
		{ 202, 141, 29 },
		{ 202, 142, 29 },
		{ 202, 143, 29 },
		{ 202, 144, 29 },
		{ 202, 145, 29 },
		{ 202, 146, 29 },
		{ 203, 131, 29 },
		{ 203, 132, 29 },
		{ 203, 133, 29 },
		{ 203, 134, 29 },
		{ 203, 135, 29 },
		{ 203, 136, 29 },
		{ 203, 137, 29 },
		{ 203, 138, 29 },
		{ 203, 139, 29 },
		{ 203, 140, 29 },
		{ 203, 141, 29 },
		{ 203, 142, 29 },
		{ 203, 143, 29 },
		{ 203, 144, 29 },
		{ 203, 145, 29 },
		{ 203, 146, 29 },
		{ 203, 147, 29 },
		{ 203, 148, 29 },
		{ 204, 131, 29 },
		{ 204, 132, 29 },
		{ 204, 133, 29 },
		{ 204, 134, 29 },
		{ 204, 135, 29 },
		{ 204, 136, 29 },
		{ 204, 137, 29 },
		{ 204, 138, 29 },
		{ 204, 139, 29 },
		{ 204, 140, 29 },
		{ 204, 141, 29 },
		{ 204, 142, 29 },
		{ 204, 143, 29 },
		{ 204, 144, 29 },
		{ 204, 145, 29 },
		{ 204, 146, 29 },
		{ 204, 147, 29 },
		{ 204, 148, 29 },
		{ 204, 149, 29 },
		{ 204, 150, 29 },
		{ 205, 131, 29 },
		{ 205, 132, 29 },
		{ 205, 133, 29 },
		{ 205, 134, 29 },
		{ 205, 135, 29 },
		{ 205, 136, 29 },
		{ 205, 137, 29 },
		{ 205, 138, 29 },
		{ 205, 139, 29 },
		{ 205, 140, 29 },
		{ 205, 141, 29 },
		{ 205, 142, 29 },
		{ 205, 143, 29 },
		{ 205, 144, 29 },
		{ 205, 145, 29 },
		{ 205, 146, 29 },
		{ 205, 147, 29 },
		{ 205, 148, 29 },
		{ 205, 149, 29 },
		{ 205, 150, 29 },
		{ 205, 151, 29 },
		{ 205, 152, 29 },
		{ 206, 131, 29 },
		{ 206, 132, 29 },
		{ 206, 133, 29 },
		{ 206, 134, 29 },
		{ 206, 135, 29 },
		{ 206, 136, 29 },
		{ 206, 137, 29 },
		{ 206, 138, 29 },
		{ 206, 139, 29 },
		{ 206, 140, 29 },
		{ 206, 141, 29 },
		{ 206, 142, 29 },
		{ 206, 143, 29 },
		{ 206, 144, 29 },
		{ 206, 145, 29 },
		{ 206, 146, 29 },
		{ 206, 147, 29 },
		{ 206, 148, 29 },
		{ 206, 149, 29 },
		{ 206, 150, 29 },
		{ 206, 151, 29 },
		{ 206, 152, 29 },
		{ 206, 153, 29 },
		{ 206, 154, 29 },
		{ 207, 131, 29 },
		{ 207, 132, 29 },
		{ 207, 133, 29 },
		{ 207, 134, 29 },
		{ 207, 135, 29 },
		{ 207, 136, 29 },
		{ 207, 137, 29 },
		{ 207, 138, 29 },
		{ 207, 139, 29 },
		{ 207, 140, 29 },
		{ 207, 141, 29 },
		{ 207, 142, 29 },
		{ 207, 143, 29 },
		{ 207, 144, 29 },
		{ 207, 145, 29 },
		{ 207, 146, 29 },
		{ 207, 147, 29 },
		{ 207, 148, 29 },
		{ 207, 149, 29 },
		{ 207, 150, 29 },
		{ 207, 151, 29 },
		{ 207, 152, 29 },
		{ 207, 153, 29 },
		{ 207, 154, 29 },
		{ 207, 155, 29 },
		{ 207, 156, 29 },
		{ 208, 131, 29 },
		{ 208, 132, 29 },
		{ 208, 133, 29 },
		{ 208, 134, 29 },
		{ 208, 135, 29 },
		{ 208, 136, 29 },
		{ 208, 137, 29 },
		{ 208, 138, 29 },
		{ 208, 139, 29 },
		{ 208, 140, 29 },
		{ 208, 141, 29 },
		{ 208, 142, 29 },
		{ 208, 143, 29 },
		{ 208, 144, 29 },
		{ 208, 145, 29 },
		{ 208, 146, 29 },
		{ 208, 147, 29 },
		{ 208, 148, 29 },
		{ 208, 149, 29 },
		{ 208, 150, 29 },
		{ 208, 151, 29 },
		{ 208, 152, 29 },
		{ 208, 153, 29 },
		{ 208, 154, 29 },
		{ 208, 155, 29 },
		{ 208, 156, 29 },
		{ 208, 157, 29 },
		{ 208, 158, 29 },
		{ 209, 131, 29 },
		{ 209, 132, 29 },
		{ 209, 133, 29 },
		{ 209, 134, 29 },
		{ 209, 135, 29 },
		{ 209, 136, 29 },
		{ 209, 137, 29 },
		{ 209, 138, 29 },
		{ 209, 139, 29 },
		{ 209, 140, 29 },
		{ 209, 141, 29 },
		{ 209, 142, 29 },
		{ 209, 143, 29 },
		{ 209, 144, 29 },
		{ 209, 145, 29 },
		{ 209, 146, 29 },
		{ 209, 147, 29 },
		{ 209, 148, 29 },
		{ 209, 149, 29 },
		{ 209, 150, 29 },
		{ 209, 151, 29 },
		{ 209, 152, 29 },
		{ 209, 153, 29 },
		{ 209, 154, 29 },
		{ 209, 155, 29 },
		{ 209, 156, 29 },
		{ 209, 157, 29 },
		{ 209, 158, 29 },
		{ 209, 159, 29 },
		{ 209, 160, 29 },
		{ 210, 131, 29 },
		{ 210, 132, 29 },
		{ 210, 133, 29 },
		{ 210, 134, 29 },
		{ 210, 135, 29 },
		{ 210, 136, 29 },
		{ 210, 137, 29 },
		{ 210, 138, 29 },
		{ 210, 139, 29 },
		{ 210, 140, 29 },
		{ 210, 141, 29 },
		{ 210, 142, 29 },
		{ 210, 143, 29 },
		{ 210, 144, 29 },
		{ 210, 145, 29 },
		{ 210, 146, 29 },
		{ 210, 147, 29 },
		{ 210, 148, 29 },
		{ 210, 149, 29 },
		{ 210, 150, 29 },
		{ 210, 151, 29 },
		{ 210, 152, 29 },
		{ 210, 153, 29 },
		{ 210, 154, 29 },
		{ 210, 155, 29 },
		{ 210, 156, 29 },
		{ 210, 157, 29 },
		{ 210, 158, 29 },
		{ 210, 159, 29 },
		{ 210, 160, 29 },
		{ 210, 161, 29 },
		{ 210, 162, 29 },
		{ 211, 131, 29 },
		{ 211, 132, 29 },
		{ 211, 133, 29 },
		{ 211, 134, 29 },
		{ 211, 135, 29 },
		{ 211, 136, 29 },
		{ 211, 137, 29 },
		{ 211, 138, 29 },
		{ 211, 139, 29 },
		{ 211, 140, 29 },
		{ 211, 141, 29 },
		{ 211, 142, 29 },
		{ 211, 143, 29 },
		{ 211, 144, 29 },
		{ 211, 145, 29 },
		{ 211, 146, 29 },
		{ 211, 147, 29 },
		{ 211, 148, 29 },
		{ 211, 149, 29 },
		{ 211, 150, 29 },
		{ 211, 151, 29 },
		{ 211, 152, 29 },
		{ 211, 153, 29 },
		{ 211, 154, 29 },
		{ 211, 155, 29 },
		{ 211, 156, 29 },
		{ 211, 157, 29 },
		{ 211, 158, 29 },
		{ 211, 159, 29 },
		{ 211, 160, 29 },
		{ 211, 161, 29 },
		{ 211, 162, 29 },
		{ 211, 163, 29 },
		{ 211, 164, 29 },
		{ 212, 131, 29 },
		{ 212, 132, 29 },
		{ 212, 133, 29 },
		{ 212, 134, 29 },
		{ 212, 135, 29 },
		{ 212, 136, 29 },
		{ 212, 137, 29 },
		{ 212, 138, 29 },
		{ 212, 139, 29 },
		{ 212, 140, 29 },
		{ 212, 141, 29 },
		{ 212, 142, 29 },
		{ 212, 143, 29 },
		{ 212, 144, 29 },
		{ 212, 145, 29 },
		{ 212, 146, 29 },
		{ 212, 147, 29 },
		{ 212, 148, 29 },
		{ 212, 149, 29 },
		{ 212, 150, 29 },
		{ 212, 151, 29 },
		{ 212, 152, 29 },
		{ 212, 153, 29 },
		{ 212, 154, 29 },
		{ 212, 155, 29 },
		{ 212, 156, 29 },
		{ 212, 157, 29 },
		{ 212, 158, 29 },
		{ 212, 159, 29 },
		{ 212, 160, 29 },
		{ 212, 161, 29 },
		{ 212, 162, 29 },
		{ 212, 163, 29 },
		{ 212, 164, 29 },
		{ 212, 165, 29 },
		{ 212, 166, 29 },
		{ 213, 131, 29 },
		{ 213, 132, 29 },
		{ 213, 133, 29 },
		{ 213, 134, 29 },
		{ 213, 135, 29 },
		{ 213, 136, 29 },
		{ 213, 137, 29 },
		{ 213, 138, 29 },
		{ 213, 139, 29 },
		{ 213, 140, 29 },
		{ 213, 141, 29 },
		{ 213, 142, 29 },
		{ 213, 143, 29 },
		{ 213, 144, 29 },
		{ 213, 145, 29 },
		{ 213, 146, 29 },
		{ 213, 147, 29 },
		{ 213, 148, 29 },
		{ 213, 149, 29 },
		{ 213, 150, 29 },
		{ 213, 151, 29 },
		{ 213, 152, 29 },
		{ 213, 153, 29 },
		{ 213, 154, 29 },
		{ 213, 155, 29 },
		{ 213, 156, 29 },
		{ 213, 157, 29 },
		{ 213, 158, 29 },
		{ 213, 159, 29 },
		{ 213, 160, 29 },
		{ 213, 161, 29 },
		{ 213, 162, 29 },
		{ 213, 163, 29 },
		{ 213, 164, 29 },
		{ 213, 165, 29 },
		{ 213, 166, 29 },
		{ 213, 167, 29 },
		{ 213, 168, 29 },
		{ 214, 131, 29 },
		{ 214, 132, 29 },
		{ 214, 133, 29 },
		{ 214, 134, 29 },
		{ 214, 135, 29 },
		{ 214, 136, 29 },
		{ 214, 137, 29 },
		{ 214, 138, 29 },
		{ 214, 139, 29 },
		{ 214, 140, 29 },
		{ 214, 141, 29 },
		{ 214, 142, 29 },
		{ 214, 143, 29 },
		{ 214, 144, 29 },
		{ 214, 145, 29 },
		{ 214, 146, 29 },
		{ 214, 147, 29 },
		{ 214, 148, 29 },
		{ 214, 149, 29 },
		{ 214, 150, 29 },
		{ 214, 151, 29 },
		{ 214, 152, 29 },
		{ 214, 153, 29 },
		{ 214, 154, 29 },
		{ 214, 155, 29 },
		{ 214, 156, 29 },
		{ 214, 157, 29 },
		{ 214, 158, 29 },
		{ 214, 159, 29 },
		{ 214, 160, 29 },
		{ 214, 161, 29 },
		{ 214, 162, 29 },
		{ 214, 163, 29 },
		{ 214, 164, 29 },
		{ 214, 165, 29 },
		{ 214, 166, 29 },
		{ 214, 167, 29 },
		{ 214, 168, 29 },
		{ 214, 169, 29 },
		{ 214, 170, 29 },
		{ 215, 131, 29 },
		{ 215, 132, 29 },
		{ 215, 133, 29 },
		{ 215, 134, 29 },
		{ 215, 135, 29 },
		{ 215, 136, 29 },
		{ 215, 137, 29 },
		{ 215, 138, 29 },
		{ 215, 139, 29 },
		{ 215, 140, 29 },
		{ 215, 141, 29 },
		{ 215, 142, 29 },
		{ 215, 143, 29 },
		{ 215, 144, 29 },
		{ 215, 145, 29 },
		{ 215, 146, 29 },
		{ 215, 147, 29 },
		{ 215, 148, 29 },
		{ 215, 149, 29 },
		{ 215, 150, 29 },
		{ 215, 151, 29 },
		{ 215, 152, 29 },
		{ 215, 153, 29 },
		{ 215, 154, 29 },
		{ 215, 155, 29 },
		{ 215, 156, 29 },
		{ 215, 157, 29 },
		{ 215, 158, 29 },
		{ 215, 159, 29 },
		{ 215, 160, 29 },
		{ 215, 161, 29 },
		{ 215, 162, 29 },
		{ 215, 163, 29 },
		{ 215, 164, 29 },
		{ 215, 165, 29 },
		{ 215, 166, 29 },
		{ 215, 167, 29 },
		{ 215, 168, 29 },
		{ 215, 169, 29 },
		{ 215, 170, 29 },
		{ 215, 171, 29 },
		{ 215, 172, 29 },
		{ 216, 131, 29 },
		{ 216, 132, 29 },
		{ 216, 133, 29 },
		{ 216, 134, 29 },
		{ 216, 135, 29 },
		{ 216, 136, 29 },
		{ 216, 137, 29 },
		{ 216, 138, 29 },
		{ 216, 139, 29 },
		{ 216, 140, 29 },
		{ 216, 141, 29 },
		{ 216, 142, 29 },
		{ 216, 143, 29 },
		{ 216, 144, 29 },
		{ 216, 145, 29 },
		{ 216, 146, 29 },
		{ 216, 147, 29 },
		{ 216, 148, 29 },
		{ 216, 149, 29 },
		{ 216, 150, 29 },
		{ 216, 151, 29 },
		{ 216, 152, 29 },
		{ 216, 153, 29 },
		{ 216, 154, 29 },
		{ 216, 155, 29 },
		{ 216, 156, 29 },
		{ 216, 157, 29 },
		{ 216, 158, 29 },
		{ 216, 159, 29 },
		{ 216, 160, 29 },
		{ 216, 161, 29 },
		{ 216, 162, 29 },
		{ 216, 163, 29 },
		{ 216, 164, 29 },
		{ 216, 165, 29 },
		{ 216, 166, 29 },
		{ 216, 167, 29 },
		{ 216, 168, 29 },
		{ 216, 169, 29 },
		{ 216, 170, 29 },
		{ 216, 171, 29 },
		{ 216, 172, 29 },
		{ 216, 173, 29 },
		{ 216, 174, 29 },
		{ 217, 131, 29 },
		{ 217, 132, 29 },
		{ 217, 133, 29 },
		{ 217, 134, 29 },
		{ 217, 135, 29 },
		{ 217, 136, 29 },
		{ 217, 137, 29 },
		{ 217, 138, 29 },
		{ 217, 139, 29 },
		{ 217, 140, 29 },
		{ 217, 141, 29 },
		{ 217, 142, 29 },
		{ 217, 143, 29 },
		{ 217, 144, 29 },
		{ 217, 145, 29 },
		{ 217, 146, 29 },
		{ 217, 147, 29 },
		{ 217, 148, 29 },
		{ 217, 149, 29 },
		{ 217, 150, 29 },
		{ 217, 151, 29 },
		{ 217, 152, 29 },
		{ 217, 153, 29 },
		{ 217, 154, 29 },
		{ 217, 155, 29 },
		{ 217, 156, 29 },
		{ 217, 157, 29 },
		{ 217, 158, 29 },
		{ 217, 159, 29 },
		{ 217, 160, 29 },
		{ 217, 161, 29 },
		{ 217, 162, 29 },
		{ 217, 163, 29 },
		{ 217, 164, 29 },
		{ 217, 165, 29 },
		{ 217, 166, 29 },
		{ 217, 167, 29 },
		{ 217, 168, 29 },
		{ 217, 169, 29 },
		{ 217, 170, 29 },
		{ 217, 171, 29 },
		{ 217, 172, 29 },
		{ 217, 173, 29 },
		{ 217, 174, 29 },
		{ 217, 175, 29 },
		{ 217, 176, 29 },
		{ 218, 131, 29 },
		{ 218, 132, 29 },
		{ 218, 133, 29 },
		{ 218, 134, 29 },
		{ 218, 135, 29 },
		{ 218, 136, 29 },
		{ 218, 137, 29 },
		{ 218, 138, 29 },
		{ 218, 139, 29 },
		{ 218, 140, 29 },
		{ 218, 141, 29 },
		{ 218, 142, 29 },
		{ 218, 143, 29 },
		{ 218, 144, 29 },
		{ 218, 145, 29 },
		{ 218, 146, 29 },
		{ 218, 147, 29 },
		{ 218, 148, 29 },
		{ 218, 149, 29 },
		{ 218, 150, 29 },
		{ 218, 151, 29 },
		{ 218, 152, 29 },
		{ 218, 153, 29 },
		{ 218, 154, 29 },
		{ 218, 155, 29 },
		{ 218, 156, 29 },
		{ 218, 157, 29 },
		{ 218, 158, 29 },
		{ 218, 159, 29 },
		{ 218, 160, 29 },
		{ 218, 161, 29 },
		{ 218, 162, 29 },
		{ 218, 163, 29 },
		{ 218, 164, 29 },
		{ 218, 165, 29 },
		{ 218, 166, 29 },
		{ 218, 167, 29 },
		{ 218, 168, 29 },
		{ 218, 169, 29 },
		{ 218, 170, 29 },
		{ 218, 171, 29 },
		{ 218, 172, 29 },
		{ 218, 173, 29 },
		{ 218, 174, 29 },
		{ 218, 175, 29 },
		{ 218, 176, 29 },
		{ 218, 177, 29 },
		{ 218, 178, 29 },
		{ 219, 131, 29 },
		{ 219, 132, 29 },
		{ 219, 133, 29 },
		{ 219, 134, 29 },
		{ 219, 135, 29 },
		{ 219, 136, 29 },
		{ 219, 137, 29 },
		{ 219, 138, 29 },
		{ 219, 139, 29 },
		{ 219, 140, 29 },
		{ 219, 141, 29 },
		{ 219, 142, 29 },
		{ 219, 143, 29 },
		{ 219, 144, 29 },
		{ 219, 145, 29 },
		{ 219, 146, 29 },
		{ 219, 147, 29 },
		{ 219, 148, 29 },
		{ 219, 149, 29 },
		{ 219, 150, 29 },
		{ 219, 151, 29 },
		{ 219, 152, 29 },
		{ 219, 153, 29 },
		{ 219, 154, 29 },
		{ 219, 155, 29 },
		{ 219, 156, 29 },
		{ 219, 157, 29 },
		{ 219, 158, 29 },
		{ 219, 159, 29 },
		{ 219, 160, 29 },
		{ 219, 161, 29 },
		{ 219, 162, 29 },
		{ 219, 163, 29 },
		{ 219, 164, 29 },
		{ 219, 165, 29 },
		{ 219, 166, 29 },
		{ 219, 167, 29 },
		{ 219, 168, 29 },
		{ 219, 169, 29 },
		{ 219, 170, 29 },
		{ 219, 171, 29 },
		{ 219, 172, 29 },
		{ 219, 173, 29 },
		{ 219, 174, 29 },
		{ 219, 175, 29 },
		{ 219, 176, 29 },
		{ 219, 177, 29 },
		{ 219, 178, 29 },
		{ 219, 179, 29 },
		{ 219, 180, 29 },
		{ 220, 131, 29 },
		{ 220, 132, 29 },
		{ 220, 133, 29 },
		{ 220, 134, 29 },
		{ 220, 135, 29 },
		{ 220, 136, 29 },
		{ 220, 137, 29 },
		{ 220, 138, 29 },
		{ 220, 139, 29 },
		{ 220, 140, 29 },
		{ 220, 141, 29 },
		{ 220, 142, 29 },
		{ 220, 143, 29 },
		{ 220, 144, 29 },
		{ 220, 145, 29 },
		{ 220, 146, 29 },
		{ 220, 147, 29 },
		{ 220, 148, 29 },
		{ 220, 149, 29 },
		{ 220, 150, 29 },
		{ 220, 151, 29 },
		{ 220, 152, 29 },
		{ 220, 153, 29 },
		{ 220, 154, 29 },
		{ 220, 155, 29 },
		{ 220, 156, 29 },
		{ 220, 157, 29 },
		{ 220, 158, 29 },
		{ 220, 159, 29 },
		{ 220, 160, 29 },
		{ 220, 161, 29 },
		{ 220, 162, 29 },
		{ 220, 163, 29 },
		{ 220, 164, 29 },
		{ 220, 165, 29 },
		{ 220, 166, 29 },
		{ 220, 167, 29 },
		{ 220, 168, 29 },
		{ 220, 169, 29 },
		{ 220, 170, 29 },
		{ 220, 171, 29 },
		{ 220, 172, 29 },
		{ 220, 173, 29 },
		{ 220, 174, 29 },
		{ 220, 175, 29 },
		{ 220, 176, 29 },
		{ 220, 177, 29 },
		{ 220, 178, 29 },
		{ 220, 179, 29 },
		{ 220, 180, 29 },
		{ 220, 181, 29 },
		{ 220, 182, 29 },
		{ 221, 131, 29 },
		{ 221, 132, 29 },
		{ 221, 133, 29 },
		{ 221, 134, 29 },
		{ 221, 135, 29 },
		{ 221, 136, 29 },
		{ 221, 137, 29 },
		{ 221, 138, 29 },
		{ 221, 139, 29 },
		{ 221, 140, 29 },
		{ 221, 141, 29 },
		{ 221, 142, 29 },
		{ 221, 143, 29 },
		{ 221, 144, 29 },
		{ 221, 145, 29 },
		{ 221, 146, 29 },
		{ 221, 147, 29 },
		{ 221, 148, 29 },
		{ 221, 149, 29 },
		{ 221, 150, 29 },
		{ 221, 151, 29 },
		{ 221, 152, 29 },
		{ 221, 153, 29 },
		{ 221, 154, 29 },
		{ 221, 155, 29 },
		{ 221, 156, 29 },
		{ 221, 157, 29 },
		{ 221, 158, 29 },
		{ 221, 159, 29 },
		{ 221, 160, 29 },
		{ 221, 161, 29 },
		{ 221, 162, 29 },
		{ 221, 163, 29 },
		{ 221, 164, 29 },
		{ 221, 165, 29 },
		{ 221, 166, 29 },
		{ 221, 167, 29 },
		{ 221, 168, 29 },
		{ 221, 169, 29 },
		{ 221, 170, 29 },
		{ 221, 171, 29 },
		{ 221, 172, 29 },
		{ 221, 173, 29 },
		{ 221, 174, 29 },
		{ 221, 175, 29 },
		{ 221, 176, 29 },
		{ 221, 177, 29 },
		{ 221, 178, 29 },
		{ 221, 179, 29 },
		{ 221, 180, 29 },
		{ 221, 181, 29 },
		{ 221, 182, 29 },
		{ 221, 183, 29 },
		{ 221, 184, 29 },
		{ 222, 131, 29 },
		{ 222, 132, 29 },
		{ 222, 133, 29 },
		{ 222, 134, 29 },
		{ 222, 135, 29 },
		{ 222, 136, 29 },
		{ 222, 137, 29 },
		{ 222, 138, 29 },
		{ 222, 139, 29 },
		{ 222, 140, 29 },
		{ 222, 141, 29 },
		{ 222, 142, 29 },
		{ 222, 143, 29 },
		{ 222, 144, 29 },
		{ 222, 145, 29 },
		{ 222, 146, 29 },
		{ 222, 147, 29 },
		{ 222, 148, 29 },
		{ 222, 149, 29 },
		{ 222, 150, 29 },
		{ 222, 151, 29 },
		{ 222, 152, 29 },
		{ 222, 153, 29 },
		{ 222, 154, 29 },
		{ 222, 155, 29 },
		{ 222, 156, 29 },
		{ 222, 157, 29 },
		{ 222, 158, 29 },
		{ 222, 159, 29 },
		{ 222, 160, 29 },
		{ 222, 161, 29 },
		{ 222, 162, 29 },
		{ 222, 163, 29 },
		{ 222, 164, 29 },
		{ 222, 165, 29 },
		{ 222, 166, 29 },
		{ 222, 167, 29 },
		{ 222, 168, 29 },
		{ 222, 169, 29 },
		{ 222, 170, 29 },
		{ 222, 171, 29 },
		{ 222, 172, 29 },
		{ 222, 173, 29 },
		{ 222, 174, 29 },
		{ 222, 175, 29 },
		{ 222, 176, 29 },
		{ 222, 177, 29 },
		{ 222, 178, 29 },
		{ 222, 179, 29 },
		{ 222, 180, 29 },
		{ 222, 181, 29 },
		{ 222, 182, 29 },
		{ 222, 183, 29 },
		{ 222, 184, 29 },
		{ 222, 185, 29 },
		{ 222, 186, 29 },
		{ 223, 131, 29 },
		{ 223, 132, 29 },
		{ 223, 133, 29 },
		{ 223, 134, 29 },
		{ 223, 135, 29 },
		{ 223, 136, 29 },
		{ 223, 137, 29 },
		{ 223, 138, 29 },
		{ 223, 139, 29 },
		{ 223, 140, 29 },
		{ 223, 141, 29 },
		{ 223, 142, 29 },
		{ 223, 143, 29 },
		{ 223, 144, 29 },
		{ 223, 145, 29 },
		{ 223, 146, 29 },
		{ 223, 147, 29 },
		{ 223, 148, 29 },
		{ 223, 149, 29 },
		{ 223, 150, 29 },
		{ 223, 151, 29 },
		{ 223, 152, 29 },
		{ 223, 153, 29 },
		{ 223, 154, 29 },
		{ 223, 155, 29 },
		{ 223, 156, 29 },
		{ 223, 157, 29 },
		{ 223, 158, 29 },
		{ 223, 159, 29 },
		{ 223, 160, 29 },
		{ 223, 161, 29 },
		{ 223, 162, 29 },
		{ 223, 163, 29 },
		{ 223, 164, 29 },
		{ 223, 165, 29 },
		{ 223, 166, 29 },
		{ 223, 167, 29 },
		{ 223, 168, 29 },
		{ 223, 169, 29 },
		{ 223, 170, 29 },
		{ 223, 171, 29 },
		{ 223, 172, 29 },
		{ 223, 173, 29 },
		{ 223, 174, 29 },
		{ 223, 175, 29 },
		{ 223, 176, 29 },
		{ 223, 177, 29 },
		{ 223, 178, 29 },
		{ 223, 179, 29 },
		{ 223, 180, 29 },
		{ 223, 181, 29 },
		{ 223, 182, 29 },
		{ 223, 183, 29 },
		{ 223, 184, 29 },
		{ 223, 185, 29 },
		{ 223, 186, 29 },
		{ 223, 187, 29 },
		{ 223, 188, 29 },
		{ 224, 131, 29 },
		{ 224, 132, 29 },
		{ 224, 133, 29 },
		{ 224, 134, 29 },
		{ 224, 135, 29 },
		{ 224, 136, 29 },
		{ 224, 137, 29 },
		{ 224, 138, 29 },
		{ 224, 139, 29 },
		{ 224, 140, 29 },
		{ 224, 141, 29 },
		{ 224, 142, 29 },
		{ 224, 143, 29 },
		{ 224, 144, 29 },
		{ 224, 145, 29 },
		{ 224, 146, 29 },
		{ 224, 147, 29 },
		{ 224, 148, 29 },
		{ 224, 149, 29 },
		{ 224, 150, 29 },
		{ 224, 151, 29 },
		{ 224, 152, 29 },
		{ 224, 153, 29 },
		{ 224, 154, 29 },
		{ 224, 155, 29 },
		{ 224, 156, 29 },
		{ 224, 157, 29 },
		{ 224, 158, 29 },
		{ 224, 159, 29 },
		{ 224, 160, 29 },
		{ 224, 161, 29 },
		{ 224, 162, 29 },
		{ 224, 163, 29 },
		{ 224, 164, 29 },
		{ 224, 165, 29 },
		{ 224, 166, 29 },
		{ 224, 167, 29 },
		{ 224, 168, 29 },
		{ 224, 169, 29 },
		{ 224, 170, 29 },
		{ 224, 171, 29 },
		{ 224, 172, 29 },
		{ 224, 173, 29 },
		{ 224, 174, 29 },
		{ 224, 175, 29 },
		{ 224, 176, 29 },
		{ 224, 177, 29 },
		{ 224, 178, 29 },
		{ 224, 179, 29 },
		{ 224, 180, 29 },
		{ 224, 181, 29 },
		{ 224, 182, 29 },
		{ 224, 183, 29 },
		{ 224, 184, 29 },
		{ 224, 185, 29 },
		{ 224, 186, 29 },
		{ 224, 187, 29 },
		{ 224, 188, 29 },
		{ 224, 189, 29 },
		{ 224, 190, 29 },
		{ 225, 131, 29 },
		{ 225, 132, 29 },
		{ 225, 133, 29 },
		{ 225, 134, 29 },
		{ 225, 135, 29 },
		{ 225, 136, 29 },
		{ 225, 137, 29 },
		{ 225, 138, 29 },
		{ 225, 139, 29 },
		{ 225, 140, 29 },
		{ 225, 141, 29 },
		{ 225, 142, 29 },
		{ 225, 143, 29 },
		{ 225, 144, 29 },
		{ 225, 145, 29 },
		{ 225, 146, 29 },
		{ 225, 147, 29 },
		{ 225, 148, 29 },
		{ 225, 149, 29 },
		{ 225, 150, 29 },
		{ 225, 151, 29 },
		{ 225, 152, 29 },
		{ 225, 153, 29 },
		{ 225, 154, 29 },
		{ 225, 155, 29 },
		{ 225, 156, 29 },
		{ 225, 157, 29 },
		{ 225, 158, 29 },
		{ 225, 159, 29 },
		{ 225, 160, 29 },
		{ 225, 161, 29 },
		{ 225, 162, 29 },
		{ 225, 163, 29 },
		{ 225, 164, 29 },
		{ 225, 165, 29 },
		{ 225, 166, 29 },
		{ 225, 167, 29 },
		{ 225, 168, 29 },
		{ 225, 169, 29 },
		{ 225, 170, 29 },
		{ 225, 171, 29 },
		{ 225, 172, 29 },
		{ 225, 173, 29 },
		{ 225, 174, 29 },
		{ 225, 175, 29 },
		{ 225, 176, 29 },
		{ 225, 177, 29 },
		{ 225, 178, 29 },
		{ 225, 179, 29 },
		{ 225, 180, 29 },
		{ 225, 181, 29 },
		{ 225, 182, 29 },
		{ 225, 183, 29 },
		{ 225, 184, 29 },
		{ 225, 185, 29 },
		{ 225, 186, 29 },
		{ 225, 187, 29 },
		{ 225, 188, 29 },
		{ 225, 189, 29 },
		{ 225, 190, 29 },
		{ 225, 191, 29 },
		{ 225, 192, 29 },
		{ 226, 131, 29 },
		{ 226, 132, 29 },
		{ 226, 133, 29 },
		{ 226, 134, 29 },
		{ 226, 135, 29 },
		{ 226, 136, 29 },
		{ 226, 137, 29 },
		{ 226, 138, 29 },
		{ 226, 139, 29 },
		{ 226, 140, 29 },
		{ 226, 141, 29 },
		{ 226, 142, 29 },
		{ 226, 143, 29 },
		{ 226, 144, 29 },
		{ 226, 145, 29 },
		{ 226, 146, 29 },
		{ 226, 147, 29 },
		{ 226, 148, 29 },
		{ 226, 149, 29 },
		{ 226, 150, 29 },
		{ 226, 151, 29 },
		{ 226, 152, 29 },
		{ 226, 153, 29 },
		{ 226, 154, 29 },
		{ 226, 155, 29 },
		{ 226, 156, 29 },
		{ 226, 157, 29 },
		{ 226, 158, 29 },
		{ 226, 159, 29 },
		{ 226, 160, 29 },
		{ 226, 161, 29 },
		{ 226, 162, 29 },
		{ 226, 163, 29 },
		{ 226, 164, 29 },
		{ 226, 165, 29 },
		{ 226, 166, 29 },
		{ 226, 167, 29 },
		{ 226, 168, 29 },
		{ 226, 169, 29 },
		{ 226, 170, 29 },
		{ 226, 171, 29 },
		{ 226, 172, 29 },
		{ 226, 173, 29 },
		{ 226, 174, 29 },
		{ 226, 175, 29 },
		{ 226, 176, 29 },
		{ 226, 177, 29 },
		{ 226, 178, 29 },
		{ 226, 179, 29 },
		{ 226, 180, 29 },
		{ 226, 181, 29 },
		{ 226, 182, 29 },
		{ 226, 183, 29 },
		{ 226, 184, 29 },
		{ 226, 185, 29 },
		{ 226, 186, 29 },
		{ 226, 187, 29 },
		{ 226, 188, 29 },
		{ 226, 189, 29 },
		{ 226, 190, 29 },
		{ 226, 191, 29 },
		{ 226, 192, 29 },
		{ 226, 193, 29 },
		{ 226, 194, 29 },
		{ 227, 131, 29 },
		{ 227, 132, 29 },
		{ 227, 133, 29 },
		{ 227, 134, 29 },
		{ 227, 135, 29 },
		{ 227, 136, 29 },
		{ 227, 137, 29 },
		{ 227, 138, 29 },
		{ 227, 139, 29 },
		{ 227, 140, 29 },
		{ 227, 141, 29 },
		{ 227, 142, 29 },
		{ 227, 143, 29 },
		{ 227, 144, 29 },
		{ 227, 145, 29 },
		{ 227, 146, 29 },
		{ 227, 147, 29 },
		{ 227, 148, 29 },
		{ 227, 149, 29 },
		{ 227, 150, 29 },
		{ 227, 151, 29 },
		{ 227, 152, 29 },
		{ 227, 153, 29 },
		{ 227, 154, 29 },
		{ 227, 155, 29 },
		{ 227, 156, 29 },
		{ 227, 157, 29 },
		{ 227, 158, 29 },
		{ 227, 159, 29 },
		{ 227, 160, 29 },
		{ 227, 161, 29 },
		{ 227, 162, 29 },
		{ 227, 163, 29 },
		{ 227, 164, 29 },
		{ 227, 165, 29 },
		{ 227, 166, 29 },
		{ 227, 167, 29 },
		{ 227, 168, 29 },
		{ 227, 169, 29 },
		{ 227, 170, 29 },
		{ 227, 171, 29 },
		{ 227, 172, 29 },
		{ 227, 173, 29 },
		{ 227, 174, 29 },
		{ 227, 175, 29 },
		{ 227, 176, 29 },
		{ 227, 177, 29 },
		{ 227, 178, 29 },
		{ 227, 179, 29 },
		{ 227, 180, 29 },
		{ 227, 181, 29 },
		{ 227, 182, 29 },
		{ 227, 183, 29 },
		{ 227, 184, 29 },
		{ 227, 185, 29 },
		{ 227, 186, 29 },
		{ 227, 187, 29 },
		{ 227, 188, 29 },
		{ 227, 189, 29 },
		{ 227, 190, 29 },
		{ 227, 191, 29 },
		{ 227, 192, 29 },
		{ 227, 193, 29 },
		{ 227, 194, 29 },
		{ 227, 195, 29 },
		{ 227, 196, 29 },
		{ 228, 131, 29 },
		{ 228, 132, 29 },
		{ 228, 133, 29 },
		{ 228, 134, 29 },
		{ 228, 135, 29 },
		{ 228, 136, 29 },
		{ 228, 137, 29 },
		{ 228, 138, 29 },
		{ 228, 139, 29 },
		{ 228, 140, 29 },
		{ 228, 141, 29 },
		{ 228, 142, 29 },
		{ 228, 143, 29 },
		{ 228, 144, 29 },
		{ 228, 145, 29 },
		{ 228, 146, 29 },
		{ 228, 147, 29 },
		{ 228, 148, 29 },
		{ 228, 149, 29 },
		{ 228, 150, 29 },
		{ 228, 151, 29 },
		{ 228, 152, 29 },
		{ 228, 153, 29 },
		{ 228, 154, 29 },
		{ 228, 155, 29 },
		{ 228, 156, 29 },
		{ 228, 157, 29 },
		{ 228, 158, 29 },
		{ 228, 159, 29 },
		{ 228, 160, 29 },
		{ 228, 161, 29 },
		{ 228, 162, 29 },
		{ 228, 163, 29 },
		{ 228, 164, 29 },
		{ 228, 165, 29 },
		{ 228, 166, 29 },
		{ 228, 167, 29 },
		{ 228, 168, 29 },
		{ 228, 169, 29 },
		{ 228, 170, 29 },
		{ 228, 171, 29 },
		{ 228, 172, 29 },
		{ 228, 173, 29 },
		{ 228, 174, 29 },
		{ 228, 175, 29 },
		{ 228, 176, 29 },
		{ 228, 177, 29 },
		{ 228, 178, 29 },
		{ 228, 179, 29 },
		{ 228, 180, 29 },
		{ 228, 181, 29 },
		{ 228, 182, 29 },
		{ 228, 183, 29 },
		{ 228, 184, 29 },
		{ 228, 185, 29 },
		{ 228, 186, 29 },
		{ 228, 187, 29 },
		{ 228, 188, 29 },
		{ 228, 189, 29 },
		{ 228, 190, 29 },
		{ 228, 191, 29 },
		{ 228, 192, 29 },
		{ 228, 193, 29 },
		{ 228, 194, 29 },
		{ 228, 195, 29 },
		{ 228, 196, 29 },
		{ 228, 197, 29 },
		{ 228, 198, 29 },
		{ 229, 131, 29 },
		{ 229, 132, 29 },
		{ 229, 133, 29 },
		{ 229, 134, 29 },
		{ 229, 135, 29 },
		{ 229, 136, 29 },
		{ 229, 137, 29 },
		{ 229, 138, 29 },
		{ 229, 139, 29 },
		{ 229, 140, 29 },
		{ 229, 141, 29 },
		{ 229, 142, 29 },
		{ 229, 143, 29 },
		{ 229, 144, 29 },
		{ 229, 145, 29 },
		{ 229, 146, 29 },
		{ 229, 147, 29 },
		{ 229, 148, 29 },
		{ 229, 149, 29 },
		{ 229, 150, 29 },
		{ 229, 151, 29 },
		{ 229, 152, 29 },
		{ 229, 153, 29 },
		{ 229, 154, 29 },
		{ 229, 155, 29 },
		{ 229, 156, 29 },
		{ 229, 157, 29 },
		{ 229, 158, 29 },
		{ 229, 159, 29 },
		{ 229, 160, 29 },
		{ 229, 161, 29 },
		{ 229, 162, 29 },
		{ 229, 163, 29 },
		{ 229, 164, 29 },
		{ 229, 165, 29 },
		{ 229, 166, 29 },
		{ 229, 167, 29 },
		{ 229, 168, 29 },
		{ 229, 169, 29 },
		{ 229, 170, 29 },
		{ 229, 171, 29 },
		{ 229, 172, 29 },
		{ 229, 173, 29 },
		{ 229, 174, 29 },
		{ 229, 175, 29 },
		{ 229, 176, 29 },
		{ 229, 177, 29 },
		{ 229, 178, 29 },
		{ 229, 179, 29 },
		{ 229, 180, 29 },
		{ 229, 181, 29 },
		{ 229, 182, 29 },
		{ 229, 183, 29 },
		{ 229, 184, 29 },
		{ 229, 185, 29 },
		{ 229, 186, 29 },
		{ 229, 187, 29 },
		{ 229, 188, 29 },
		{ 229, 189, 29 },
		{ 229, 190, 29 },
		{ 229, 191, 29 },
		{ 229, 192, 29 },
		{ 229, 193, 29 },
		{ 229, 194, 29 },
		{ 229, 195, 29 },
		{ 229, 196, 29 },
		{ 229, 197, 29 },
		{ 229, 198, 29 },
		{ 229, 199, 29 },
		{ 229, 200, 29 },
		{ 230, 131, 29 },
		{ 230, 132, 29 },
		{ 230, 133, 29 },
		{ 230, 134, 29 },
		{ 230, 135, 29 },
		{ 230, 136, 29 },
		{ 230, 137, 29 },
		{ 230, 138, 29 },
		{ 230, 139, 29 },
		{ 230, 140, 29 },
		{ 230, 141, 29 },
		{ 230, 142, 29 },
		{ 230, 143, 29 },
		{ 230, 144, 29 },
		{ 230, 145, 29 },
		{ 230, 146, 29 },
		{ 230, 147, 29 },
		{ 230, 148, 29 },
		{ 230, 149, 29 },
		{ 230, 150, 29 },
		{ 230, 151, 29 },
		{ 230, 152, 29 },
		{ 230, 153, 29 },
		{ 230, 154, 29 },
		{ 230, 155, 29 },
		{ 230, 156, 29 },
		{ 230, 157, 29 },
		{ 230, 158, 29 },
		{ 230, 159, 29 },
		{ 230, 160, 29 },
		{ 230, 161, 29 },
		{ 230, 162, 29 },
		{ 230, 163, 29 },
		{ 230, 164, 29 },
		{ 230, 165, 29 },
		{ 230, 166, 29 },
		{ 230, 167, 29 },
		{ 230, 168, 29 },
		{ 230, 169, 29 },
		{ 230, 170, 29 },
		{ 230, 171, 29 },
		{ 230, 172, 29 },
		{ 230, 173, 29 },
		{ 230, 174, 29 },
		{ 230, 175, 29 },
		{ 230, 176, 29 },
		{ 230, 177, 29 },
		{ 230, 178, 29 },
		{ 230, 179, 29 },
		{ 230, 180, 29 },
		{ 230, 181, 29 },
		{ 230, 182, 29 },
		{ 230, 183, 29 },
		{ 230, 184, 29 },
		{ 230, 185, 29 },
		{ 230, 186, 29 },
		{ 230, 187, 29 },
		{ 230, 188, 29 },
		{ 230, 189, 29 },
		{ 230, 190, 29 },
		{ 230, 191, 29 },
		{ 230, 192, 29 },
		{ 230, 193, 29 },
		{ 230, 194, 29 },
		{ 230, 195, 29 },
		{ 230, 196, 29 },
		{ 230, 197, 29 },
		{ 230, 198, 29 },
		{ 230, 199, 29 },
		{ 230, 200, 29 },
		{ 230, 201, 29 },
		{ 230, 202, 29 },
		{ 231, 131, 29 },
		{ 231, 132, 29 },
		{ 231, 133, 29 },
		{ 231, 134, 29 },
		{ 231, 135, 29 },
		{ 231, 136, 29 },
		{ 231, 137, 29 },
		{ 231, 138, 29 },
		{ 231, 139, 29 },
		{ 231, 140, 29 },
		{ 231, 141, 29 },
		{ 231, 142, 29 },
		{ 231, 143, 29 },
		{ 231, 144, 29 },
		{ 231, 145, 29 },
		{ 231, 146, 29 },
		{ 231, 147, 29 },
		{ 231, 148, 29 },
		{ 231, 149, 29 },
		{ 231, 150, 29 },
		{ 231, 151, 29 },
		{ 231, 152, 29 },
		{ 231, 153, 29 },
		{ 231, 154, 29 },
		{ 231, 155, 29 },
		{ 231, 156, 29 },
		{ 231, 157, 29 },
		{ 231, 158, 29 },
		{ 231, 159, 29 },
		{ 231, 160, 29 },
		{ 231, 161, 29 },
		{ 231, 162, 29 },
		{ 231, 163, 29 },
		{ 231, 164, 29 },
		{ 231, 165, 29 },
		{ 231, 166, 29 },
		{ 231, 167, 29 },
		{ 231, 168, 29 },
		{ 231, 169, 29 },
		{ 231, 170, 29 },
		{ 231, 171, 29 },
		{ 231, 172, 29 },
		{ 231, 173, 29 },
		{ 231, 174, 29 },
		{ 231, 175, 29 },
		{ 231, 176, 29 },
		{ 231, 177, 29 },
		{ 231, 178, 29 },
		{ 231, 179, 29 },
		{ 231, 180, 29 },
		{ 231, 181, 29 },
		{ 231, 182, 29 },
		{ 231, 183, 29 },
		{ 231, 184, 29 },
		{ 231, 185, 29 },
		{ 231, 186, 29 },
		{ 231, 187, 29 },
		{ 231, 188, 29 },
		{ 231, 189, 29 },
		{ 231, 190, 29 },
		{ 231, 191, 29 },
		{ 231, 192, 29 },
		{ 231, 193, 29 },
		{ 231, 194, 29 },
		{ 231, 195, 29 },
		{ 231, 196, 29 },
		{ 231, 197, 29 },
		{ 231, 198, 29 },
		{ 231, 199, 29 },
		{ 231, 200, 29 },
		{ 231, 201, 29 },
		{ 231, 202, 29 },
		{ 231, 203, 29 },
		{ 231, 204, 29 },
		{ 232, 131, 29 },
		{ 232, 132, 29 },
		{ 232, 133, 29 },
		{ 232, 134, 29 },
		{ 232, 135, 29 },
		{ 232, 136, 29 },
		{ 232, 137, 29 },
		{ 232, 138, 29 },
		{ 232, 139, 29 },
		{ 232, 140, 29 },
		{ 232, 141, 29 },
		{ 232, 142, 29 },
		{ 232, 143, 29 },
		{ 232, 144, 29 },
		{ 232, 145, 29 },
		{ 232, 146, 29 },
		{ 232, 147, 29 },
		{ 232, 148, 29 },
		{ 232, 149, 29 },
		{ 232, 150, 29 },
		{ 232, 151, 29 },
		{ 232, 152, 29 },
		{ 232, 153, 29 },
		{ 232, 154, 29 },
		{ 232, 155, 29 },
		{ 232, 156, 29 },
		{ 232, 157, 29 },
		{ 232, 158, 29 },
		{ 232, 159, 29 },
		{ 232, 160, 29 },
		{ 232, 161, 29 },
		{ 232, 162, 29 },
		{ 232, 163, 29 },
		{ 232, 164, 29 },
		{ 232, 165, 29 },
		{ 232, 166, 29 },
		{ 232, 167, 29 },
		{ 232, 168, 29 },
		{ 232, 169, 29 },
		{ 232, 170, 29 },
		{ 232, 171, 29 },
		{ 232, 172, 29 },
		{ 232, 173, 29 },
		{ 232, 174, 29 },
		{ 232, 175, 29 },
		{ 232, 176, 29 },
		{ 232, 177, 29 },
		{ 232, 178, 29 },
		{ 232, 179, 29 },
		{ 232, 180, 29 },
		{ 232, 181, 29 },
		{ 232, 182, 29 },
		{ 232, 183, 29 },
		{ 232, 184, 29 },
		{ 232, 185, 29 },
		{ 232, 186, 29 },
		{ 232, 187, 29 },
		{ 232, 188, 29 },
		{ 232, 189, 29 },
		{ 232, 190, 29 },
		{ 232, 191, 29 },
		{ 232, 192, 29 },
		{ 232, 193, 29 },
		{ 232, 194, 29 },
		{ 232, 195, 29 },
		{ 232, 196, 29 },
		{ 232, 197, 29 },
		{ 232, 198, 29 },
		{ 232, 199, 29 },
		{ 232, 200, 29 },
		{ 232, 201, 29 },
		{ 232, 202, 29 },
		{ 232, 203, 29 },
		{ 232, 204, 29 },
		{ 232, 205, 29 },
		{ 232, 206, 29 },
		{ 233, 131, 29 },
		{ 233, 132, 29 },
		{ 233, 133, 29 },
		{ 233, 134, 29 },
		{ 233, 135, 29 },
		{ 233, 136, 29 },
		{ 233, 137, 29 },
		{ 233, 138, 29 },
		{ 233, 139, 29 },
		{ 233, 140, 29 },
		{ 233, 141, 29 },
		{ 233, 142, 29 },
		{ 233, 143, 29 },
		{ 233, 144, 29 },
		{ 233, 145, 29 },
		{ 233, 146, 29 },
		{ 233, 147, 29 },
		{ 233, 148, 29 },
		{ 233, 149, 29 },
		{ 233, 150, 29 },
		{ 233, 151, 29 },
		{ 233, 152, 29 },
		{ 233, 153, 29 },
		{ 233, 154, 29 },
		{ 233, 155, 29 },
		{ 233, 156, 29 },
		{ 233, 157, 29 },
		{ 233, 158, 29 },
		{ 233, 159, 29 },
		{ 233, 160, 29 },
		{ 233, 161, 29 },
		{ 233, 162, 29 },
		{ 233, 163, 29 },
		{ 233, 164, 29 },
		{ 233, 165, 29 },
		{ 233, 166, 29 },
		{ 233, 167, 29 },
		{ 233, 168, 29 },
		{ 233, 169, 29 },
		{ 233, 170, 29 },
		{ 233, 171, 29 },
		{ 233, 172, 29 },
		{ 233, 173, 29 },
		{ 233, 174, 29 },
		{ 233, 175, 29 },
		{ 233, 176, 29 },
		{ 233, 177, 29 },
		{ 233, 178, 29 },
		{ 233, 179, 29 },
		{ 233, 180, 29 },
		{ 233, 181, 29 },
		{ 233, 182, 29 },
		{ 233, 183, 29 },
		{ 233, 184, 29 },
		{ 233, 185, 29 },
		{ 233, 186, 29 },
		{ 233, 187, 29 },
		{ 233, 188, 29 },
		{ 233, 189, 29 },
		{ 233, 190, 29 },
		{ 233, 191, 29 },
		{ 233, 192, 29 },
		{ 233, 193, 29 },
		{ 233, 194, 29 },
		{ 233, 195, 29 },
		{ 233, 196, 29 },
		{ 233, 197, 29 },
		{ 233, 198, 29 },
		{ 233, 199, 29 },
		{ 233, 200, 29 },
		{ 233, 201, 29 },
		{ 233, 202, 29 },
		{ 233, 203, 29 },
		{ 233, 204, 29 },
		{ 233, 205, 29 },
		{ 233, 206, 29 },
		{ 233, 207, 29 },
		{ 233, 208, 29 },
		{ 234, 131, 29 },
		{ 234, 132, 29 },
		{ 234, 133, 29 },
		{ 234, 134, 29 },
		{ 234, 135, 29 },
		{ 234, 136, 29 },
		{ 234, 137, 29 },
		{ 234, 138, 29 },
		{ 234, 139, 29 },
		{ 234, 140, 29 },
		{ 234, 141, 29 },
		{ 234, 142, 29 },
		{ 234, 143, 29 },
		{ 234, 144, 29 },
		{ 234, 145, 29 },
		{ 234, 146, 29 },
		{ 234, 147, 29 },
		{ 234, 148, 29 },
		{ 234, 149, 29 },
		{ 234, 150, 29 },
		{ 234, 151, 29 },
		{ 234, 152, 29 },
		{ 234, 153, 29 },
		{ 234, 154, 29 },
		{ 234, 155, 29 },
		{ 234, 156, 29 },
		{ 234, 157, 29 },
		{ 234, 158, 29 },
		{ 234, 159, 29 },
		{ 234, 160, 29 },
		{ 234, 161, 29 },
		{ 234, 162, 29 },
		{ 234, 163, 29 },
		{ 234, 164, 29 },
		{ 234, 165, 29 },
		{ 234, 166, 29 },
		{ 234, 167, 29 },
		{ 234, 168, 29 },
		{ 234, 169, 29 },
		{ 234, 170, 29 },
		{ 234, 171, 29 },
		{ 234, 172, 29 },
		{ 234, 173, 29 },
		{ 234, 174, 29 },
		{ 234, 175, 29 },
		{ 234, 176, 29 },
		{ 234, 177, 29 },
		{ 234, 178, 29 },
		{ 234, 179, 29 },
		{ 234, 180, 29 },
		{ 234, 181, 29 },
		{ 234, 182, 29 },
		{ 234, 183, 29 },
		{ 234, 184, 29 },
		{ 234, 185, 29 },
		{ 234, 186, 29 },
		{ 234, 187, 29 },
		{ 234, 188, 29 },
		{ 234, 189, 29 },
		{ 234, 190, 29 },
		{ 234, 191, 29 },
		{ 234, 192, 29 },
		{ 234, 193, 29 },
		{ 234, 194, 29 },
		{ 234, 195, 29 },
		{ 234, 196, 29 },
		{ 234, 197, 29 },
		{ 234, 198, 29 },
		{ 234, 199, 29 },
		{ 234, 200, 29 },
		{ 234, 201, 29 },
		{ 234, 202, 29 },
		{ 234, 203, 29 },
		{ 234, 204, 29 },
		{ 234, 205, 29 },
		{ 234, 206, 29 },
		{ 234, 207, 29 },
		{ 234, 208, 29 },
		{ 234, 209, 29 },
		{ 234, 210, 29 },
		{ 235, 131, 29 },
		{ 235, 132, 29 },
		{ 235, 133, 29 },
		{ 235, 134, 29 },
		{ 235, 135, 29 },
		{ 235, 136, 29 },
		{ 235, 137, 29 },
		{ 235, 138, 29 },
		{ 235, 139, 29 },
		{ 235, 140, 29 },
		{ 235, 141, 29 },
		{ 235, 142, 29 },
		{ 235, 143, 29 },
		{ 235, 144, 29 },
		{ 235, 145, 29 },
		{ 235, 146, 29 },
		{ 235, 147, 29 },
		{ 235, 148, 29 },
		{ 235, 149, 29 },
		{ 235, 150, 29 },
		{ 235, 151, 29 },
		{ 235, 152, 29 },
		{ 235, 153, 29 },
		{ 235, 154, 29 },
		{ 235, 155, 29 },
		{ 235, 156, 29 },
		{ 235, 157, 29 },
		{ 235, 158, 29 },
		{ 235, 159, 29 },
		{ 235, 160, 29 },
		{ 235, 161, 29 },
		{ 235, 162, 29 },
		{ 235, 163, 29 },
		{ 235, 164, 29 },
		{ 235, 165, 29 },
		{ 235, 166, 29 },
		{ 235, 167, 29 },
		{ 235, 168, 29 },
		{ 235, 169, 29 },
		{ 235, 170, 29 },
		{ 235, 171, 29 },
		{ 235, 172, 29 },
		{ 235, 173, 29 },
		{ 235, 174, 29 },
		{ 235, 175, 29 },
		{ 235, 176, 29 },
		{ 235, 177, 29 },
		{ 235, 178, 29 },
		{ 235, 179, 29 },
		{ 235, 180, 29 },
		{ 235, 181, 29 },
		{ 235, 182, 29 },
		{ 235, 183, 29 },
		{ 235, 184, 29 },
		{ 235, 185, 29 },
		{ 235, 186, 29 },
		{ 235, 187, 29 },
		{ 235, 188, 29 },
		{ 235, 189, 29 },
		{ 235, 190, 29 },
		{ 235, 191, 29 },
		{ 235, 192, 29 },
		{ 235, 193, 29 },
		{ 235, 194, 29 },
		{ 235, 195, 29 },
		{ 235, 196, 29 },
		{ 235, 197, 29 },
		{ 235, 198, 29 },
		{ 235, 199, 29 },
		{ 235, 200, 29 },
		{ 235, 201, 29 },
		{ 235, 202, 29 },
		{ 235, 203, 29 },
		{ 235, 204, 29 },
		{ 235, 205, 29 },
		{ 235, 206, 29 },
		{ 235, 207, 29 },
		{ 235, 208, 29 },
		{ 235, 209, 29 },
		{ 235, 210, 29 },
		{ 235, 211, 29 },
		{ 235, 212, 29 },
		{ 236, 131, 29 },
		{ 236, 132, 29 },
		{ 236, 133, 29 },
		{ 236, 134, 29 },
		{ 236, 135, 29 },
		{ 236, 136, 29 },
		{ 236, 137, 29 },
		{ 236, 138, 29 },
		{ 236, 139, 29 },
		{ 236, 140, 29 },
		{ 236, 141, 29 },
		{ 236, 142, 29 },
		{ 236, 143, 29 },
		{ 236, 144, 29 },
		{ 236, 145, 29 },
		{ 236, 146, 29 },
		{ 236, 147, 29 },
		{ 236, 148, 29 },
		{ 236, 149, 29 },
		{ 236, 150, 29 },
		{ 236, 151, 29 },
		{ 236, 152, 29 },
		{ 236, 153, 29 },
		{ 236, 154, 29 },
		{ 236, 155, 29 },
		{ 236, 156, 29 },
		{ 236, 157, 29 },
		{ 236, 158, 29 },
		{ 236, 159, 29 },
		{ 236, 160, 29 },
		{ 236, 161, 29 },
		{ 236, 162, 29 },
		{ 236, 163, 29 },
		{ 236, 164, 29 },
		{ 236, 165, 29 },
		{ 236, 166, 29 },
		{ 236, 167, 29 },
		{ 236, 168, 29 },
		{ 236, 169, 29 },
		{ 236, 170, 29 },
		{ 236, 171, 29 },
		{ 236, 172, 29 },
		{ 236, 173, 29 },
		{ 236, 174, 29 },
		{ 236, 175, 29 },
		{ 236, 176, 29 },
		{ 236, 177, 29 },
		{ 236, 178, 29 },
		{ 236, 179, 29 },
		{ 236, 180, 29 },
		{ 236, 181, 29 },
		{ 236, 182, 29 },
		{ 236, 183, 29 },
		{ 236, 184, 29 },
		{ 236, 185, 29 },
		{ 236, 186, 29 },
		{ 236, 187, 29 },
		{ 236, 188, 29 },
		{ 236, 189, 29 },
		{ 236, 190, 29 },
		{ 236, 191, 29 },
		{ 236, 192, 29 },
		{ 236, 193, 29 },
		{ 236, 194, 29 },
		{ 236, 195, 29 },
		{ 236, 196, 29 },
		{ 236, 197, 29 },
		{ 236, 198, 29 },
		{ 236, 199, 29 },
		{ 236, 200, 29 },
		{ 236, 201, 29 },
		{ 236, 202, 29 },
		{ 236, 203, 29 },
		{ 236, 204, 29 },
		{ 236, 205, 29 },
		{ 236, 206, 29 },
		{ 236, 207, 29 },
		{ 236, 208, 29 },
		{ 236, 209, 29 },
		{ 236, 210, 29 },
		{ 236, 211, 29 },
		{ 236, 212, 29 },
		{ 236, 213, 29 },
		{ 236, 214, 29 },
		{ 237, 131, 29 },
		{ 237, 132, 29 },
		{ 237, 133, 29 },
		{ 237, 134, 29 },
		{ 237, 135, 29 },
		{ 237, 136, 29 },
		{ 237, 137, 29 },
		{ 237, 138, 29 },
		{ 237, 139, 29 },
		{ 237, 140, 29 },
		{ 237, 141, 29 },
		{ 237, 142, 29 },
		{ 237, 143, 29 },
		{ 237, 144, 29 },
		{ 237, 145, 29 },
		{ 237, 146, 29 },
		{ 237, 147, 29 },
		{ 237, 148, 29 },
		{ 237, 149, 29 },
		{ 237, 150, 29 },
		{ 237, 151, 29 },
		{ 237, 152, 29 },
		{ 237, 153, 29 },
		{ 237, 154, 29 },
		{ 237, 155, 29 },
		{ 237, 156, 29 },
		{ 237, 157, 29 },
		{ 237, 158, 29 },
		{ 237, 159, 29 },
		{ 237, 160, 29 },
		{ 237, 161, 29 },
		{ 237, 162, 29 },
		{ 237, 163, 29 },
		{ 237, 164, 29 },
		{ 237, 165, 29 },
		{ 237, 166, 29 },
		{ 237, 167, 29 },
		{ 237, 168, 29 },
		{ 237, 169, 29 },
		{ 237, 170, 29 },
		{ 237, 171, 29 },
		{ 237, 172, 29 },
		{ 237, 173, 29 },
		{ 237, 174, 29 },
		{ 237, 175, 29 },
		{ 237, 176, 29 },
		{ 237, 177, 29 },
		{ 237, 178, 29 },
		{ 237, 179, 29 },
		{ 237, 180, 29 },
		{ 237, 181, 29 },
		{ 237, 182, 29 },
		{ 237, 183, 29 },
		{ 237, 184, 29 },
		{ 237, 185, 29 },
		{ 237, 186, 29 },
		{ 237, 187, 29 },
		{ 237, 188, 29 },
		{ 237, 189, 29 },
		{ 237, 190, 29 },
		{ 237, 191, 29 },
		{ 237, 192, 29 },
		{ 237, 193, 29 },
		{ 237, 194, 29 },
		{ 237, 195, 29 },
		{ 237, 196, 29 },
		{ 237, 197, 29 },
		{ 237, 198, 29 },
		{ 237, 199, 29 },
		{ 237, 200, 29 },
		{ 237, 201, 29 },
		{ 237, 202, 29 },
		{ 237, 203, 29 },
		{ 237, 204, 29 },
		{ 237, 205, 29 },
		{ 237, 206, 29 },
		{ 237, 207, 29 },
		{ 237, 208, 29 },
		{ 237, 209, 29 },
		{ 237, 210, 29 },
		{ 237, 211, 29 },
		{ 237, 212, 29 },
		{ 237, 213, 29 },
		{ 237, 214, 29 },
		{ 237, 215, 29 },
		{ 237, 216, 29 },
		{ 238, 131, 29 },
		{ 238, 132, 29 },
		{ 238, 133, 29 },
		{ 238, 134, 29 },
		{ 238, 135, 29 },
		{ 238, 136, 29 },
		{ 238, 137, 29 },
		{ 238, 138, 29 },
		{ 238, 139, 29 },
		{ 238, 140, 29 },
		{ 238, 141, 29 },
		{ 238, 142, 29 },
		{ 238, 143, 29 },
		{ 238, 144, 29 },
		{ 238, 145, 29 },
		{ 238, 146, 29 },
		{ 238, 147, 29 },
		{ 238, 148, 29 },
		{ 238, 149, 29 },
		{ 238, 150, 29 },
		{ 238, 151, 29 },
		{ 238, 152, 29 },
		{ 238, 153, 29 },
		{ 238, 154, 29 },
		{ 238, 155, 29 },
		{ 238, 156, 29 },
		{ 238, 157, 29 },
		{ 238, 158, 29 },
		{ 238, 159, 29 },
		{ 238, 160, 29 },
		{ 238, 161, 29 },
		{ 238, 162, 29 },
		{ 238, 163, 29 },
		{ 238, 164, 29 },
		{ 238, 165, 29 },
		{ 238, 166, 29 },
		{ 238, 167, 29 },
		{ 238, 168, 29 },
		{ 238, 169, 29 },
		{ 238, 170, 29 },
		{ 238, 171, 29 },
		{ 238, 172, 29 },
		{ 238, 173, 29 },
		{ 238, 174, 29 },
		{ 238, 175, 29 },
		{ 238, 176, 29 },
		{ 238, 177, 29 },
		{ 238, 178, 29 },
		{ 238, 179, 29 },
		{ 238, 180, 29 },
		{ 238, 181, 29 },
		{ 238, 182, 29 },
		{ 238, 183, 29 },
		{ 238, 184, 29 },
		{ 238, 185, 29 },
		{ 238, 186, 29 },
		{ 238, 187, 29 },
		{ 238, 188, 29 },
		{ 238, 189, 29 },
		{ 238, 190, 29 },
		{ 238, 191, 29 },
		{ 238, 192, 29 },
		{ 238, 193, 29 },
		{ 238, 194, 29 },
		{ 238, 195, 29 },
		{ 238, 196, 29 },
		{ 238, 197, 29 },
		{ 238, 198, 29 },
		{ 238, 199, 29 },
		{ 238, 200, 29 },
		{ 238, 201, 29 },
		{ 238, 202, 29 },
		{ 238, 203, 29 },
		{ 238, 204, 29 },
		{ 238, 205, 29 },
		{ 238, 206, 29 },
		{ 238, 207, 29 },
		{ 238, 208, 29 },
		{ 238, 209, 29 },
		{ 238, 210, 29 },
		{ 238, 211, 29 },
		{ 238, 212, 29 },
		{ 238, 213, 29 },
		{ 238, 214, 29 },
		{ 238, 215, 29 },
		{ 238, 216, 29 },
		{ 238, 217, 29 },
		{ 238, 218, 29 },
		{ 239, 131, 29 },
		{ 239, 132, 29 },
		{ 239, 133, 29 },
		{ 239, 134, 29 },
		{ 239, 135, 29 },
		{ 239, 136, 29 },
		{ 239, 137, 29 },
		{ 239, 138, 29 },
		{ 239, 139, 29 },
		{ 239, 140, 29 },
		{ 239, 141, 29 },
		{ 239, 142, 29 },
		{ 239, 143, 29 },
		{ 239, 144, 29 },
		{ 239, 145, 29 },
		{ 239, 146, 29 },
		{ 239, 147, 29 },
		{ 239, 148, 29 },
		{ 239, 149, 29 },
		{ 239, 150, 29 },
		{ 239, 151, 29 },
		{ 239, 152, 29 },
		{ 239, 153, 29 },
		{ 239, 154, 29 },
		{ 239, 155, 29 },
		{ 239, 156, 29 },
		{ 239, 157, 29 },
		{ 239, 158, 29 },
		{ 239, 159, 29 },
		{ 239, 160, 29 },
		{ 239, 161, 29 },
		{ 239, 162, 29 },
		{ 239, 163, 29 },
		{ 239, 164, 29 },
		{ 239, 165, 29 },
		{ 239, 166, 29 },
		{ 239, 167, 29 },
		{ 239, 168, 29 },
		{ 239, 169, 29 },
		{ 239, 170, 29 },
		{ 239, 171, 29 },
		{ 239, 172, 29 },
		{ 239, 173, 29 },
		{ 239, 174, 29 },
		{ 239, 175, 29 },
		{ 239, 176, 29 },
		{ 239, 177, 29 },
		{ 239, 178, 29 },
		{ 239, 179, 29 },
		{ 239, 180, 29 },
		{ 239, 181, 29 },
		{ 239, 182, 29 },
		{ 239, 183, 29 },
		{ 239, 184, 29 },
		{ 239, 185, 29 },
		{ 239, 186, 29 },
		{ 239, 187, 29 },
		{ 239, 188, 29 },
		{ 239, 189, 29 },
		{ 239, 190, 29 },
		{ 239, 191, 29 },
		{ 239, 192, 29 },
		{ 239, 193, 29 },
		{ 239, 194, 29 },
		{ 239, 195, 29 },
		{ 239, 196, 29 },
		{ 239, 197, 29 },
		{ 239, 198, 29 },
		{ 239, 199, 29 },
		{ 239, 200, 29 },
		{ 239, 201, 29 },
		{ 239, 202, 29 },
		{ 239, 203, 29 },
		{ 239, 204, 29 },
		{ 239, 205, 29 },
		{ 239, 206, 29 },
		{ 239, 207, 29 },
		{ 239, 208, 29 },
		{ 239, 209, 29 },
		{ 239, 210, 29 },
		{ 239, 211, 29 },
		{ 239, 212, 29 },
		{ 239, 213, 29 },
		{ 239, 214, 29 },
		{ 239, 215, 29 },
		{ 239, 216, 29 },
		{ 239, 217, 29 },
		{ 239, 218, 29 },
		{ 239, 219, 29 },
		{ 239, 220, 29 },
		{ 240, 131, 29 },
		{ 240, 132, 29 },
		{ 240, 133, 29 },
		{ 240, 134, 29 },
		{ 240, 135, 29 },
		{ 240, 136, 29 },
		{ 240, 137, 29 },
		{ 240, 138, 29 },
		{ 240, 139, 29 },
		{ 240, 140, 29 },
		{ 240, 141, 29 },
		{ 240, 142, 29 },
		{ 240, 143, 29 },
		{ 240, 144, 29 },
		{ 240, 145, 29 },
		{ 240, 146, 29 },
		{ 240, 147, 29 },
		{ 240, 148, 29 },
		{ 240, 149, 29 },
		{ 240, 150, 29 },
		{ 240, 151, 29 },
		{ 240, 152, 29 },
		{ 240, 153, 29 },
		{ 240, 154, 29 },
		{ 240, 155, 29 },
		{ 240, 156, 29 },
		{ 240, 157, 29 },
		{ 240, 158, 29 },
		{ 240, 159, 29 },
		{ 240, 160, 29 },
		{ 240, 161, 29 },
		{ 240, 162, 29 },
		{ 240, 163, 29 },
		{ 240, 164, 29 },
		{ 240, 165, 29 },
		{ 240, 166, 29 },
		{ 240, 167, 29 },
		{ 240, 168, 29 },
		{ 240, 169, 29 },
		{ 240, 170, 29 },
		{ 240, 171, 29 },
		{ 240, 172, 29 },
		{ 240, 173, 29 },
		{ 240, 174, 29 },
		{ 240, 175, 29 },
		{ 240, 176, 29 },
		{ 240, 177, 29 },
		{ 240, 178, 29 },
		{ 240, 179, 29 },
		{ 240, 180, 29 },
		{ 240, 181, 29 },
		{ 240, 182, 29 },
		{ 240, 183, 29 },
		{ 240, 184, 29 },
		{ 240, 185, 29 },
		{ 240, 186, 29 },
		{ 240, 187, 29 },
		{ 240, 188, 29 },
		{ 240, 189, 29 },
		{ 240, 190, 29 },
		{ 240, 191, 29 },
		{ 240, 192, 29 },
		{ 240, 193, 29 },
		{ 240, 194, 29 },
		{ 240, 195, 29 },
		{ 240, 196, 29 },
		{ 240, 197, 29 },
		{ 240, 198, 29 },
		{ 240, 199, 29 },
		{ 240, 200, 29 },
		{ 240, 201, 29 },
		{ 240, 202, 29 },
		{ 240, 203, 29 },
		{ 240, 204, 29 },
		{ 240, 205, 29 },
		{ 240, 206, 29 },
		{ 240, 207, 29 },
		{ 240, 208, 29 },
		{ 240, 209, 29 },
		{ 240, 210, 29 },
		{ 240, 211, 29 },
		{ 240, 212, 29 },
		{ 240, 213, 29 },
		{ 240, 214, 29 },
		{ 240, 215, 29 },
		{ 240, 216, 29 },
		{ 240, 217, 29 },
		{ 240, 218, 29 },
		{ 240, 219, 29 },
		{ 240, 220, 29 },
		{ 240, 221, 29 },
		{ 240, 222, 29 },
		{ 241, 131, 29 },
		{ 241, 132, 29 },
		{ 241, 133, 29 },
		{ 241, 134, 29 },
		{ 241, 135, 29 },
		{ 241, 136, 29 },
		{ 241, 137, 29 },
		{ 241, 138, 29 },
		{ 241, 139, 29 },
		{ 241, 140, 29 },
		{ 241, 141, 29 },
		{ 241, 142, 29 },
		{ 241, 143, 29 },
		{ 241, 144, 29 },
		{ 241, 145, 29 },
		{ 241, 146, 29 },
		{ 241, 147, 29 },
		{ 241, 148, 29 },
		{ 241, 149, 29 },
		{ 241, 150, 29 },
		{ 241, 151, 29 },
		{ 241, 152, 29 },
		{ 241, 153, 29 },
		{ 241, 154, 29 },
		{ 241, 155, 29 },
		{ 241, 156, 29 },
		{ 241, 157, 29 },
		{ 241, 158, 29 },
		{ 241, 159, 29 },
		{ 241, 160, 29 },
		{ 241, 161, 29 },
		{ 241, 162, 29 },
		{ 241, 163, 29 },
		{ 241, 164, 29 },
		{ 241, 165, 29 },
		{ 241, 166, 29 },
		{ 241, 167, 29 },
		{ 241, 168, 29 },
		{ 241, 169, 29 },
		{ 241, 170, 29 },
		{ 241, 171, 29 },
		{ 241, 172, 29 },
		{ 241, 173, 29 },
		{ 241, 174, 29 },
		{ 241, 175, 29 },
		{ 241, 176, 29 },
		{ 241, 177, 29 },
		{ 241, 178, 29 },
		{ 241, 179, 29 },
		{ 241, 180, 29 },
		{ 241, 181, 29 },
		{ 241, 182, 29 },
		{ 241, 183, 29 },
		{ 241, 184, 29 },
		{ 241, 185, 29 },
		{ 241, 186, 29 },
		{ 241, 187, 29 },
		{ 241, 188, 29 },
		{ 241, 189, 29 },
		{ 241, 190, 29 },
		{ 241, 191, 29 },
		{ 241, 192, 29 },
		{ 241, 193, 29 },
		{ 241, 194, 29 },
		{ 241, 195, 29 },
		{ 241, 196, 29 },
		{ 241, 197, 29 },
		{ 241, 198, 29 },
		{ 241, 199, 29 },
		{ 241, 200, 29 },
		{ 241, 201, 29 },
		{ 241, 202, 29 },
		{ 241, 203, 29 },
		{ 241, 204, 29 },
		{ 241, 205, 29 },
		{ 241, 206, 29 },
		{ 241, 207, 29 },
		{ 241, 208, 29 },
		{ 241, 209, 29 },
		{ 241, 210, 29 },
		{ 241, 211, 29 },
		{ 241, 212, 29 },
		{ 241, 213, 29 },
		{ 241, 214, 29 },
		{ 241, 215, 29 },
		{ 241, 216, 29 },
		{ 241, 217, 29 },
		{ 241, 218, 29 },
		{ 241, 219, 29 },
		{ 241, 220, 29 },
		{ 241, 221, 29 },
		{ 241, 222, 29 },
		{ 241, 223, 29 },
		{ 241, 224, 29 },
		{ 242, 131, 29 },
		{ 242, 132, 29 },
		{ 242, 133, 29 },
		{ 242, 134, 29 },
		{ 242, 135, 29 },
		{ 242, 136, 29 },
		{ 242, 137, 29 },
		{ 242, 138, 29 },
		{ 242, 139, 29 },
		{ 242, 140, 29 },
		{ 242, 141, 29 },
		{ 242, 142, 29 },
		{ 242, 143, 29 },
		{ 242, 144, 29 },
		{ 242, 145, 29 },
		{ 242, 146, 29 },
		{ 242, 147, 29 },
		{ 242, 148, 29 },
		{ 242, 149, 29 },
		{ 242, 150, 29 },
		{ 242, 151, 29 },
		{ 242, 152, 29 },
		{ 242, 153, 29 },
		{ 242, 154, 29 },
		{ 242, 155, 29 },
		{ 242, 156, 29 },
		{ 242, 157, 29 },
		{ 242, 158, 29 },
		{ 242, 159, 29 },
		{ 242, 160, 29 },
		{ 242, 161, 29 },
		{ 242, 162, 29 },
		{ 242, 163, 29 },
		{ 242, 164, 29 },
		{ 242, 165, 29 },
		{ 242, 166, 29 },
		{ 242, 167, 29 },
		{ 242, 168, 29 },
		{ 242, 169, 29 },
		{ 242, 170, 29 },
		{ 242, 171, 29 },
		{ 242, 172, 29 },
		{ 242, 173, 29 },
		{ 242, 174, 29 },
		{ 242, 175, 29 },
		{ 242, 176, 29 },
		{ 242, 177, 29 },
		{ 242, 178, 29 },
		{ 242, 179, 29 },
		{ 242, 180, 29 },
		{ 242, 181, 29 },
		{ 242, 182, 29 },
		{ 242, 183, 29 },
		{ 242, 184, 29 },
		{ 242, 185, 29 },
		{ 242, 186, 29 },
		{ 242, 187, 29 },
		{ 242, 188, 29 },
		{ 242, 189, 29 },
		{ 242, 190, 29 },
		{ 242, 191, 29 },
		{ 242, 192, 29 },
		{ 242, 193, 29 },
		{ 242, 194, 29 },
		{ 242, 195, 29 },
		{ 242, 196, 29 },
		{ 242, 197, 29 },
		{ 242, 198, 29 },
		{ 242, 199, 29 },
		{ 242, 200, 29 },
		{ 242, 201, 29 },
		{ 242, 202, 29 },
		{ 242, 203, 29 },
		{ 242, 204, 29 },
		{ 242, 205, 29 },
		{ 242, 206, 29 },
		{ 242, 207, 29 },
		{ 242, 208, 29 },
		{ 242, 209, 29 },
		{ 242, 210, 29 },
		{ 242, 211, 29 },
		{ 242, 212, 29 },
		{ 242, 213, 29 },
		{ 242, 214, 29 },
		{ 242, 215, 29 },
		{ 242, 216, 29 },
		{ 242, 217, 29 },
		{ 242, 218, 29 },
		{ 242, 219, 29 },
		{ 242, 220, 29 },
		{ 242, 221, 29 },
		{ 242, 222, 29 },
		{ 242, 223, 29 },
		{ 242, 224, 29 },
		{ 242, 225, 29 },
		{ 242, 226, 29 },
		{ 243, 131, 29 },
		{ 243, 132, 29 },
		{ 243, 133, 29 },
		{ 243, 134, 29 },
		{ 243, 135, 29 },
		{ 243, 136, 29 },
		{ 243, 137, 29 },
		{ 243, 138, 29 },
		{ 243, 139, 29 },
		{ 243, 140, 29 },
		{ 243, 141, 29 },
		{ 243, 142, 29 },
		{ 243, 143, 29 },
		{ 243, 144, 29 },
		{ 243, 145, 29 },
		{ 243, 146, 29 },
		{ 243, 147, 29 },
		{ 243, 148, 29 },
		{ 243, 149, 29 },
		{ 243, 150, 29 },
		{ 243, 151, 29 },
		{ 243, 152, 29 },
		{ 243, 153, 29 },
		{ 243, 154, 29 },
		{ 243, 155, 29 },
		{ 243, 156, 29 },
		{ 243, 157, 29 },
		{ 243, 158, 29 },
		{ 243, 159, 29 },
		{ 243, 160, 29 },
		{ 243, 161, 29 },
		{ 243, 162, 29 },
		{ 243, 163, 29 },
		{ 243, 164, 29 },
		{ 243, 165, 29 },
		{ 243, 166, 29 },
		{ 243, 167, 29 },
		{ 243, 168, 29 },
		{ 243, 169, 29 },
		{ 243, 170, 29 },
		{ 243, 171, 29 },
		{ 243, 172, 29 },
		{ 243, 173, 29 },
		{ 243, 174, 29 },
		{ 243, 175, 29 },
		{ 243, 176, 29 },
		{ 243, 177, 29 },
		{ 243, 178, 29 },
		{ 243, 179, 29 },
		{ 243, 180, 29 },
		{ 243, 181, 29 },
		{ 243, 182, 29 },
		{ 243, 183, 29 },
		{ 243, 184, 29 },
		{ 243, 185, 29 },
		{ 243, 186, 29 },
		{ 243, 187, 29 },
		{ 243, 188, 29 },
		{ 243, 189, 29 },
		{ 243, 190, 29 },
		{ 243, 191, 29 },
		{ 243, 192, 29 },
		{ 243, 193, 29 },
		{ 243, 194, 29 },
		{ 243, 195, 29 },
		{ 243, 196, 29 },
		{ 243, 197, 29 },
		{ 243, 198, 29 },
		{ 243, 199, 29 },
		{ 243, 200, 29 },
		{ 243, 201, 29 },
		{ 243, 202, 29 },
		{ 243, 203, 29 },
		{ 243, 204, 29 },
		{ 243, 205, 29 },
		{ 243, 206, 29 },
		{ 243, 207, 29 },
		{ 243, 208, 29 },
		{ 243, 209, 29 },
		{ 243, 210, 29 },
		{ 243, 211, 29 },
		{ 243, 212, 29 },
		{ 243, 213, 29 },
		{ 243, 214, 29 },
		{ 243, 215, 29 },
		{ 243, 216, 29 },
		{ 243, 217, 29 },
		{ 243, 218, 29 },
		{ 243, 219, 29 },
		{ 243, 220, 29 },
		{ 243, 221, 29 },
		{ 243, 222, 29 },
		{ 243, 223, 29 },
		{ 243, 224, 29 },
		{ 243, 225, 29 },
		{ 243, 226, 29 },
		{ 243, 227, 29 },
		{ 243, 228, 29 },
		{ 244, 131, 29 },
		{ 244, 132, 29 },
		{ 244, 133, 29 },
		{ 244, 134, 29 },
		{ 244, 135, 29 },
		{ 244, 136, 29 },
		{ 244, 137, 29 },
		{ 244, 138, 29 },
		{ 244, 139, 29 },
		{ 244, 140, 29 },
		{ 244, 141, 29 },
		{ 244, 142, 29 },
		{ 244, 143, 29 },
		{ 244, 144, 29 },
		{ 244, 145, 29 },
		{ 244, 146, 29 },
		{ 244, 147, 29 },
		{ 244, 148, 29 },
		{ 244, 149, 29 },
		{ 244, 150, 29 },
		{ 244, 151, 29 },
		{ 244, 152, 29 },
		{ 244, 153, 29 },
		{ 244, 154, 29 },
		{ 244, 155, 29 },
		{ 244, 156, 29 },
		{ 244, 157, 29 },
		{ 244, 158, 29 },
		{ 244, 159, 29 },
		{ 244, 160, 29 },
		{ 244, 161, 29 },
		{ 244, 162, 29 },
		{ 244, 163, 29 },
		{ 244, 164, 29 },
		{ 244, 165, 29 },
		{ 244, 166, 29 },
		{ 244, 167, 29 },
		{ 244, 168, 29 },
		{ 244, 169, 29 },
		{ 244, 170, 29 },
		{ 244, 171, 29 },
		{ 244, 172, 29 },
		{ 244, 173, 29 },
		{ 244, 174, 29 },
		{ 244, 175, 29 },
		{ 244, 176, 29 },
		{ 244, 177, 29 },
		{ 244, 178, 29 },
		{ 244, 179, 29 },
		{ 244, 180, 29 },
		{ 244, 181, 29 },
		{ 244, 182, 29 },
		{ 244, 183, 29 },
		{ 244, 184, 29 },
		{ 244, 185, 29 },
		{ 244, 186, 29 },
		{ 244, 187, 29 },
		{ 244, 188, 29 },
		{ 244, 189, 29 },
		{ 244, 190, 29 },
		{ 244, 191, 29 },
		{ 244, 192, 29 },
		{ 244, 193, 29 },
		{ 244, 194, 29 },
		{ 244, 195, 29 },
		{ 244, 196, 29 },
		{ 244, 197, 29 },
		{ 244, 198, 29 },
		{ 244, 199, 29 },
		{ 244, 200, 29 },
		{ 244, 201, 29 },
		{ 244, 202, 29 },
		{ 244, 203, 29 },
		{ 244, 204, 29 },
		{ 244, 205, 29 },
		{ 244, 206, 29 },
		{ 244, 207, 29 },
		{ 244, 208, 29 },
		{ 244, 209, 29 },
		{ 244, 210, 29 },
		{ 244, 211, 29 },
		{ 244, 212, 29 },
		{ 244, 213, 29 },
		{ 244, 214, 29 },
		{ 244, 215, 29 },
		{ 244, 216, 29 },
		{ 244, 217, 29 },
		{ 244, 218, 29 },
		{ 244, 219, 29 },
		{ 244, 220, 29 },
		{ 244, 221, 29 },
		{ 244, 222, 29 },
		{ 244, 223, 29 },
		{ 244, 224, 29 },
		{ 244, 225, 29 },
		{ 244, 226, 29 },
		{ 244, 227, 29 },
		{ 244, 228, 29 },
		{ 244, 229, 29 },
		{ 244, 230, 29 },
		{ 245, 131, 29 },
		{ 245, 132, 29 },
		{ 245, 133, 29 },
		{ 245, 134, 29 },
		{ 245, 135, 29 },
		{ 245, 136, 29 },
		{ 245, 137, 29 },
		{ 245, 138, 29 },
		{ 245, 139, 29 },
		{ 245, 140, 29 },
		{ 245, 141, 29 },
		{ 245, 142, 29 },
		{ 245, 143, 29 },
		{ 245, 144, 29 },
		{ 245, 145, 29 },
		{ 245, 146, 29 },
		{ 245, 147, 29 },
		{ 245, 148, 29 },
		{ 245, 149, 29 },
		{ 245, 150, 29 },
		{ 245, 151, 29 },
		{ 245, 152, 29 },
		{ 245, 153, 29 },
		{ 245, 154, 29 },
		{ 245, 155, 29 },
		{ 245, 156, 29 },
		{ 245, 157, 29 },
		{ 245, 158, 29 },
		{ 245, 159, 29 },
		{ 245, 160, 29 },
		{ 245, 161, 29 },
		{ 245, 162, 29 },
		{ 245, 163, 29 },
		{ 245, 164, 29 },
		{ 245, 165, 29 },
		{ 245, 166, 29 },
		{ 245, 167, 29 },
		{ 245, 168, 29 },
		{ 245, 169, 29 },
		{ 245, 170, 29 },
		{ 245, 171, 29 },
		{ 245, 172, 29 },
		{ 245, 173, 29 },
		{ 245, 174, 29 },
		{ 245, 175, 29 },
		{ 245, 176, 29 },
		{ 245, 177, 29 },
		{ 245, 178, 29 },
		{ 245, 179, 29 },
		{ 245, 180, 29 },
		{ 245, 181, 29 },
		{ 245, 182, 29 },
		{ 245, 183, 29 },
		{ 245, 184, 29 },
		{ 245, 185, 29 },
		{ 245, 186, 29 },
		{ 245, 187, 29 },
		{ 245, 188, 29 },
		{ 245, 189, 29 },
		{ 245, 190, 29 },
		{ 245, 191, 29 },
		{ 245, 192, 29 },
		{ 245, 193, 29 },
		{ 245, 194, 29 },
		{ 245, 195, 29 },
		{ 245, 196, 29 },
		{ 245, 197, 29 },
		{ 245, 198, 29 },
		{ 245, 199, 29 },
		{ 245, 200, 29 },
		{ 245, 201, 29 },
		{ 245, 202, 29 },
		{ 245, 203, 29 },
		{ 245, 204, 29 },
		{ 245, 205, 29 },
		{ 245, 206, 29 },
		{ 245, 207, 29 },
		{ 245, 208, 29 },
		{ 245, 209, 29 },
		{ 245, 210, 29 },
		{ 245, 211, 29 },
		{ 245, 212, 29 },
		{ 245, 213, 29 },
		{ 245, 214, 29 },
		{ 245, 215, 29 },
		{ 245, 216, 29 },
		{ 245, 217, 29 },
		{ 245, 218, 29 },
		{ 245, 219, 29 },
		{ 245, 220, 29 },
		{ 245, 221, 29 },
		{ 245, 222, 29 },
		{ 245, 223, 29 },
		{ 245, 224, 29 },
		{ 245, 225, 29 },
		{ 245, 226, 29 },
		{ 245, 227, 29 },
		{ 245, 228, 29 },
		{ 245, 229, 29 },
		{ 245, 230, 29 },
		{ 245, 231, 29 },
		{ 245, 232, 29 },
		{ 246, 131, 29 },
		{ 246, 132, 29 },
		{ 246, 133, 29 },
		{ 246, 134, 29 },
		{ 246, 135, 29 },
		{ 246, 136, 29 },
		{ 246, 137, 29 },
		{ 246, 138, 29 },
		{ 246, 139, 29 },
		{ 246, 140, 29 },
		{ 246, 141, 29 },
		{ 246, 142, 29 },
		{ 246, 143, 29 },
		{ 246, 144, 29 },
		{ 246, 145, 29 },
		{ 246, 146, 29 },
		{ 246, 147, 29 },
		{ 246, 148, 29 },
		{ 246, 149, 29 },
		{ 246, 150, 29 },
		{ 246, 151, 29 },
		{ 246, 152, 29 },
		{ 246, 153, 29 },
		{ 246, 154, 29 },
		{ 246, 155, 29 },
		{ 246, 156, 29 },
		{ 246, 157, 29 },
		{ 246, 158, 29 },
		{ 246, 159, 29 },
		{ 246, 160, 29 },
		{ 246, 161, 29 },
		{ 246, 162, 29 },
		{ 246, 163, 29 },
		{ 246, 164, 29 },
		{ 246, 165, 29 },
		{ 246, 166, 29 },
		{ 246, 167, 29 },
		{ 246, 168, 29 },
		{ 246, 169, 29 },
		{ 246, 170, 29 },
		{ 246, 171, 29 },
		{ 246, 172, 29 },
		{ 246, 173, 29 },
		{ 246, 174, 29 },
		{ 246, 175, 29 },
		{ 246, 176, 29 },
		{ 246, 177, 29 },
		{ 246, 178, 29 },
		{ 246, 179, 29 },
		{ 246, 180, 29 },
		{ 246, 181, 29 },
		{ 246, 182, 29 },
		{ 246, 183, 29 },
		{ 246, 184, 29 },
		{ 246, 185, 29 },
		{ 246, 186, 29 },
		{ 246, 187, 29 },
		{ 246, 188, 29 },
		{ 246, 189, 29 },
		{ 246, 190, 29 },
		{ 246, 191, 29 },
		{ 246, 192, 29 },
		{ 246, 193, 29 },
		{ 246, 194, 29 },
		{ 246, 195, 29 },
		{ 246, 196, 29 },
		{ 246, 197, 29 },
		{ 246, 198, 29 },
		{ 246, 199, 29 },
		{ 246, 200, 29 },
		{ 246, 201, 29 },
		{ 246, 202, 29 },
		{ 246, 203, 29 },
		{ 246, 204, 29 },
		{ 246, 205, 29 },
		{ 246, 206, 29 },
		{ 246, 207, 29 },
		{ 246, 208, 29 },
		{ 246, 209, 29 },
		{ 246, 210, 29 },
		{ 246, 211, 29 },
		{ 246, 212, 29 },
		{ 246, 213, 29 },
		{ 246, 214, 29 },
		{ 246, 215, 29 },
		{ 246, 216, 29 },
		{ 246, 217, 29 },
		{ 246, 218, 29 },
		{ 246, 219, 29 },
		{ 246, 220, 29 },
		{ 246, 221, 29 },
		{ 246, 222, 29 },
		{ 246, 223, 29 },
		{ 246, 224, 29 },
		{ 246, 225, 29 },
		{ 246, 226, 29 },
		{ 246, 227, 29 },
		{ 246, 228, 29 },
		{ 246, 229, 29 },
		{ 246, 230, 29 },
		{ 246, 231, 29 },
		{ 246, 232, 29 },
		{ 246, 233, 29 },
		{ 246, 234, 29 },
		{ 247, 131, 29 },
		{ 247, 132, 29 },
		{ 247, 133, 29 },
		{ 247, 134, 29 },
		{ 247, 135, 29 },
		{ 247, 136, 29 },
		{ 247, 137, 29 },
		{ 247, 138, 29 },
		{ 247, 139, 29 },
		{ 247, 140, 29 },
		{ 247, 141, 29 },
		{ 247, 142, 29 },
		{ 247, 143, 29 },
		{ 247, 144, 29 },
		{ 247, 145, 29 },
		{ 247, 146, 29 },
		{ 247, 147, 29 },
		{ 247, 148, 29 },
		{ 247, 149, 29 },
		{ 247, 150, 29 },
		{ 247, 151, 29 },
		{ 247, 152, 29 },
		{ 247, 153, 29 },
		{ 247, 154, 29 },
		{ 247, 155, 29 },
		{ 247, 156, 29 },
		{ 247, 157, 29 },
		{ 247, 158, 29 },
		{ 247, 159, 29 },
		{ 247, 160, 29 },
		{ 247, 161, 29 },
		{ 247, 162, 29 },
		{ 247, 163, 29 },
		{ 247, 164, 29 },
		{ 247, 165, 29 },
		{ 247, 166, 29 },
		{ 247, 167, 29 },
		{ 247, 168, 29 },
		{ 247, 169, 29 },
		{ 247, 170, 29 },
		{ 247, 171, 29 },
		{ 247, 172, 29 },
		{ 247, 173, 29 },
		{ 247, 174, 29 },
		{ 247, 175, 29 },
		{ 247, 176, 29 },
		{ 247, 177, 29 },
		{ 247, 178, 29 },
		{ 247, 179, 29 },
		{ 247, 180, 29 },
		{ 247, 181, 29 },
		{ 247, 182, 29 },
		{ 247, 183, 29 },
		{ 247, 184, 29 },
		{ 247, 185, 29 },
		{ 247, 186, 29 },
		{ 247, 187, 29 },
		{ 247, 188, 29 },
		{ 247, 189, 29 },
		{ 247, 190, 29 },
		{ 247, 191, 29 },
		{ 247, 192, 29 },
		{ 247, 193, 29 },
		{ 247, 194, 29 },
		{ 247, 195, 29 },
		{ 247, 196, 29 },
		{ 247, 197, 29 },
		{ 247, 198, 29 },
		{ 247, 199, 29 },
		{ 247, 200, 29 },
		{ 247, 201, 29 },
		{ 247, 202, 29 },
		{ 247, 203, 29 },
		{ 247, 204, 29 },
		{ 247, 205, 29 },
		{ 247, 206, 29 },
		{ 247, 207, 29 },
		{ 247, 208, 29 },
		{ 247, 209, 29 },
		{ 247, 210, 29 },
		{ 247, 211, 29 },
		{ 247, 212, 29 },
		{ 247, 213, 29 },
		{ 247, 214, 29 },
		{ 247, 215, 29 },
		{ 247, 216, 29 },
		{ 247, 217, 29 },
		{ 247, 218, 29 },
		{ 247, 219, 29 },
		{ 247, 220, 29 },
		{ 247, 221, 29 },
		{ 247, 222, 29 },
		{ 247, 223, 29 },
		{ 247, 224, 29 },
		{ 247, 225, 29 },
		{ 247, 226, 29 },
		{ 247, 227, 29 },
		{ 247, 228, 29 },
		{ 247, 229, 29 },
		{ 247, 230, 29 },
		{ 247, 231, 29 },
		{ 247, 232, 29 },
		{ 247, 233, 29 },
		{ 247, 234, 29 },
		{ 247, 235, 29 },
		{ 247, 236, 29 },
		{ 248, 131, 29 },
		{ 248, 132, 29 },
		{ 248, 133, 29 },
		{ 248, 134, 29 },
		{ 248, 135, 29 },
		{ 248, 136, 29 },
		{ 248, 137, 29 },
		{ 248, 138, 29 },
		{ 248, 139, 29 },
		{ 248, 140, 29 },
		{ 248, 141, 29 },
		{ 248, 142, 29 },
		{ 248, 143, 29 },
		{ 248, 144, 29 },
		{ 248, 145, 29 },
		{ 248, 146, 29 },
		{ 248, 147, 29 },
		{ 248, 148, 29 },
		{ 248, 149, 29 },
		{ 248, 150, 29 },
		{ 248, 151, 29 },
		{ 248, 152, 29 },
		{ 248, 153, 29 },
		{ 248, 154, 29 },
		{ 248, 155, 29 },
		{ 248, 156, 29 },
		{ 248, 157, 29 },
		{ 248, 158, 29 },
		{ 248, 159, 29 },
		{ 248, 160, 29 },
		{ 248, 161, 29 },
		{ 248, 162, 29 },
		{ 248, 163, 29 },
		{ 248, 164, 29 },
		{ 248, 165, 29 },
		{ 248, 166, 29 },
		{ 248, 167, 29 },
		{ 248, 168, 29 },
		{ 248, 169, 29 },
		{ 248, 170, 29 },
		{ 248, 171, 29 },
		{ 248, 172, 29 },
		{ 248, 173, 29 },
		{ 248, 174, 29 },
		{ 248, 175, 29 },
		{ 248, 176, 29 },
		{ 248, 177, 29 },
		{ 248, 178, 29 },
		{ 248, 179, 29 },
		{ 248, 180, 29 },
		{ 248, 181, 29 },
		{ 248, 182, 29 },
		{ 248, 183, 29 },
		{ 248, 184, 29 },
		{ 248, 185, 29 },
		{ 248, 186, 29 },
		{ 248, 187, 29 },
		{ 248, 188, 29 },
		{ 248, 189, 29 },
		{ 248, 190, 29 },
		{ 248, 191, 29 },
		{ 248, 192, 29 },
		{ 248, 193, 29 },
		{ 248, 194, 29 },
		{ 248, 195, 29 },
		{ 248, 196, 29 },
		{ 248, 197, 29 },
		{ 248, 198, 29 },
		{ 248, 199, 29 },
		{ 248, 200, 29 },
		{ 248, 201, 29 },
		{ 248, 202, 29 },
		{ 248, 203, 29 },
		{ 248, 204, 29 },
		{ 248, 205, 29 },
		{ 248, 206, 29 },
		{ 248, 207, 29 },
		{ 248, 208, 29 },
		{ 248, 209, 29 },
		{ 248, 210, 29 },
		{ 248, 211, 29 },
		{ 248, 212, 29 },
		{ 248, 213, 29 },
		{ 248, 214, 29 },
		{ 248, 215, 29 },
		{ 248, 216, 29 },
		{ 248, 217, 29 },
		{ 248, 218, 29 },
		{ 248, 219, 29 },
		{ 248, 220, 29 },
		{ 248, 221, 29 },
		{ 248, 222, 29 },
		{ 248, 223, 29 },
		{ 248, 224, 29 },
		{ 248, 225, 29 },
		{ 248, 226, 29 },
		{ 248, 227, 29 },
		{ 248, 228, 29 },
		{ 248, 229, 29 },
		{ 248, 230, 29 },
		{ 248, 231, 29 },
		{ 248, 232, 29 },
		{ 248, 233, 29 },
		{ 248, 234, 29 },
		{ 248, 235, 29 },
		{ 248, 236, 29 },
		{ 248, 237, 29 },
		{ 248, 238, 29 },
		{ 249, 131, 29 },
		{ 249, 132, 29 },
		{ 249, 133, 29 },
		{ 249, 134, 29 },
		{ 249, 135, 29 },
		{ 249, 136, 29 },
		{ 249, 137, 29 },
		{ 249, 138, 29 },
		{ 249, 139, 29 },
		{ 249, 140, 29 },
		{ 249, 141, 29 },
		{ 249, 142, 29 },
		{ 249, 143, 29 },
		{ 249, 144, 29 },
		{ 249, 145, 29 },
		{ 249, 146, 29 },
		{ 249, 147, 29 },
		{ 249, 148, 29 },
		{ 249, 149, 29 },
		{ 249, 150, 29 },
		{ 249, 151, 29 },
		{ 249, 152, 29 },
		{ 249, 153, 29 },
		{ 249, 154, 29 },
		{ 249, 155, 29 },
		{ 249, 156, 29 },
		{ 249, 157, 29 },
		{ 249, 158, 29 },
		{ 249, 159, 29 },
		{ 249, 160, 29 },
		{ 249, 161, 29 },
		{ 249, 162, 29 },
		{ 249, 163, 29 },
		{ 249, 164, 29 },
		{ 249, 165, 29 },
		{ 249, 166, 29 },
		{ 249, 167, 29 },
		{ 249, 168, 29 },
		{ 249, 169, 29 },
		{ 249, 170, 29 },
		{ 249, 171, 29 },
		{ 249, 172, 29 },
		{ 249, 173, 29 },
		{ 249, 174, 29 },
		{ 249, 175, 29 },
		{ 249, 176, 29 },
		{ 249, 177, 29 },
		{ 249, 178, 29 },
		{ 249, 179, 29 },
		{ 249, 180, 29 },
		{ 249, 181, 29 },
		{ 249, 182, 29 },
		{ 249, 183, 29 },
		{ 249, 184, 29 },
		{ 249, 185, 29 },
		{ 249, 186, 29 },
		{ 249, 187, 29 },
		{ 249, 188, 29 },
		{ 249, 189, 29 },
		{ 249, 190, 29 },
		{ 249, 191, 29 },
		{ 249, 192, 29 },
		{ 249, 193, 29 },
		{ 249, 194, 29 },
		{ 249, 195, 29 },
		{ 249, 196, 29 },
		{ 249, 197, 29 },
		{ 249, 198, 29 },
		{ 249, 199, 29 },
		{ 249, 200, 29 },
		{ 249, 201, 29 },
		{ 249, 202, 29 },
		{ 249, 203, 29 },
		{ 249, 204, 29 },
		{ 249, 205, 29 },
		{ 249, 206, 29 },
		{ 249, 207, 29 },
		{ 249, 208, 29 },
		{ 249, 209, 29 },
		{ 249, 210, 29 },
		{ 249, 211, 29 },
		{ 249, 212, 29 },
		{ 249, 213, 29 },
		{ 249, 214, 29 },
		{ 249, 215, 29 },
		{ 249, 216, 29 },
		{ 249, 217, 29 },
		{ 249, 218, 29 },
		{ 249, 219, 29 },
		{ 249, 220, 29 },
		{ 249, 221, 29 },
		{ 249, 222, 29 },
		{ 249, 223, 29 },
		{ 249, 224, 29 },
		{ 249, 225, 29 },
		{ 249, 226, 29 },
		{ 249, 227, 29 },
		{ 249, 228, 29 },
		{ 249, 229, 29 },
		{ 249, 230, 29 },
		{ 249, 231, 29 },
		{ 249, 232, 29 },
		{ 249, 233, 29 },
		{ 249, 234, 29 },
		{ 249, 235, 29 },
		{ 249, 236, 29 },
		{ 249, 237, 29 },
		{ 249, 238, 29 },
		{ 249, 239, 29 },
		{ 249, 240, 29 },
		{ 250, 131, 29 },
		{ 250, 132, 29 },
		{ 250, 133, 29 },
		{ 250, 134, 29 },
		{ 250, 135, 29 },
		{ 250, 136, 29 },
		{ 250, 137, 29 },
		{ 250, 138, 29 },
		{ 250, 139, 29 },
		{ 250, 140, 29 },
		{ 250, 141, 29 },
		{ 250, 142, 29 },
		{ 250, 143, 29 },
		{ 250, 144, 29 },
		{ 250, 145, 29 },
		{ 250, 146, 29 },
		{ 250, 147, 29 },
		{ 250, 148, 29 },
		{ 250, 149, 29 },
		{ 250, 150, 29 },
		{ 250, 151, 29 },
		{ 250, 152, 29 },
		{ 250, 153, 29 },
		{ 250, 154, 29 },
		{ 250, 155, 29 },
		{ 250, 156, 29 },
		{ 250, 157, 29 },
		{ 250, 158, 29 },
		{ 250, 159, 29 },
		{ 250, 160, 29 },
		{ 250, 161, 29 },
		{ 250, 162, 29 },
		{ 250, 163, 29 },
		{ 250, 164, 29 },
		{ 250, 165, 29 },
		{ 250, 166, 29 },
		{ 250, 167, 29 },
		{ 250, 168, 29 },
		{ 250, 169, 29 },
		{ 250, 170, 29 },
		{ 250, 171, 29 },
		{ 250, 172, 29 },
		{ 250, 173, 29 },
		{ 250, 174, 29 },
		{ 250, 175, 29 },
		{ 250, 176, 29 },
		{ 250, 177, 29 },
		{ 250, 178, 29 },
		{ 250, 179, 29 },
		{ 250, 180, 29 },
		{ 250, 181, 29 },
		{ 250, 182, 29 },
		{ 250, 183, 29 },
		{ 250, 184, 29 },
		{ 250, 185, 29 },
		{ 250, 186, 29 },
		{ 250, 187, 29 },
		{ 250, 188, 29 },
		{ 250, 189, 29 },
		{ 250, 190, 29 },
		{ 250, 191, 29 },
		{ 250, 192, 29 },
		{ 250, 193, 29 },
		{ 250, 194, 29 },
		{ 250, 195, 29 },
		{ 250, 196, 29 },
		{ 250, 197, 29 },
		{ 250, 198, 29 },
		{ 250, 199, 29 },
		{ 250, 200, 29 },
		{ 250, 201, 29 },
		{ 250, 202, 29 },
		{ 250, 203, 29 },
		{ 250, 204, 29 },
		{ 250, 205, 29 },
		{ 250, 206, 29 },
		{ 250, 207, 29 },
		{ 250, 208, 29 },
		{ 250, 209, 29 },
		{ 250, 210, 29 },
		{ 250, 211, 29 },
		{ 250, 212, 29 },
		{ 250, 213, 29 },
		{ 250, 214, 29 },
		{ 250, 215, 29 },
		{ 250, 216, 29 },
		{ 250, 217, 29 },
		{ 250, 218, 29 },
		{ 250, 219, 29 },
		{ 250, 220, 29 },
		{ 250, 221, 29 },
		{ 250, 222, 29 },
		{ 250, 223, 29 },
		{ 250, 224, 29 },
		{ 250, 225, 29 },
		{ 250, 226, 29 },
		{ 250, 227, 29 },
		{ 250, 228, 29 },
		{ 250, 229, 29 },
		{ 250, 230, 29 },
		{ 250, 231, 29 },
		{ 250, 232, 29 },
		{ 250, 233, 29 },
		{ 250, 234, 29 },
		{ 250, 235, 29 },
		{ 250, 236, 29 },
		{ 250, 237, 29 },
		{ 250, 238, 29 },
		{ 250, 239, 29 },
		{ 250, 240, 29 },
		{ 251, 132, 29 },
		{ 251, 133, 29 },
		{ 251, 134, 29 },
		{ 251, 135, 29 },
		{ 251, 136, 29 },
		{ 251, 137, 29 },
		{ 251, 138, 29 },
		{ 251, 139, 29 },
		{ 251, 140, 29 },
		{ 251, 141, 29 },
		{ 251, 142, 29 },
		{ 251, 143, 29 },
		{ 251, 144, 29 },
		{ 251, 145, 29 },
		{ 251, 146, 29 },
		{ 251, 147, 29 },
		{ 251, 148, 29 },
		{ 251, 149, 29 },
		{ 251, 150, 29 },
		{ 251, 151, 29 },
		{ 251, 152, 29 },
		{ 251, 153, 29 },
		{ 251, 154, 29 },
		{ 251, 155, 29 },
		{ 251, 156, 29 },
		{ 251, 157, 29 },
		{ 251, 158, 29 },
		{ 251, 159, 29 },
		{ 251, 160, 29 },
		{ 251, 161, 29 },
		{ 251, 162, 29 },
		{ 251, 163, 29 },
		{ 251, 164, 29 },
		{ 251, 165, 29 },
		{ 251, 166, 29 },
		{ 251, 167, 29 },
		{ 251, 168, 29 },
		{ 251, 169, 29 },
		{ 251, 170, 29 },
		{ 251, 171, 29 },
		{ 251, 172, 29 },
		{ 251, 173, 29 },
		{ 251, 174, 29 },
		{ 251, 175, 29 },
		{ 251, 176, 29 },
		{ 251, 177, 29 },
		{ 251, 178, 29 },
		{ 251, 179, 29 },
		{ 251, 180, 29 },
		{ 251, 181, 29 },
		{ 251, 182, 29 },
		{ 251, 183, 29 },
		{ 251, 184, 29 },
		{ 251, 185, 29 },
		{ 251, 186, 29 },
		{ 251, 187, 29 },
		{ 251, 188, 29 },
		{ 251, 189, 29 },
		{ 251, 190, 29 },
		{ 251, 191, 29 },
		{ 251, 192, 29 },
		{ 251, 193, 29 },
		{ 251, 194, 29 },
		{ 251, 195, 29 },
		{ 251, 196, 29 },
		{ 251, 197, 29 },
		{ 251, 198, 29 },
		{ 251, 199, 29 },
		{ 251, 200, 29 },
		{ 251, 201, 29 },
		{ 251, 202, 29 },
		{ 251, 203, 29 },
		{ 251, 204, 29 },
		{ 251, 205, 29 },
		{ 251, 206, 29 },
		{ 251, 207, 29 },
		{ 251, 208, 29 },
		{ 251, 209, 29 },
		{ 251, 210, 29 },
		{ 251, 211, 29 },
		{ 251, 212, 29 },
		{ 251, 213, 29 },
		{ 251, 214, 29 },
		{ 251, 215, 29 },
		{ 251, 216, 29 },
		{ 251, 217, 29 },
		{ 251, 218, 29 },
		{ 251, 219, 29 },
		{ 251, 220, 29 },
		{ 251, 221, 29 },
		{ 251, 222, 29 },
		{ 251, 223, 29 },
		{ 251, 224, 29 },
		{ 251, 225, 29 },
		{ 251, 226, 29 },
		{ 251, 227, 29 },
		{ 251, 228, 29 },
		{ 251, 229, 29 },
		{ 251, 230, 29 },
		{ 251, 231, 29 },
		{ 251, 232, 29 },
		{ 251, 233, 29 },
		{ 251, 234, 29 },
		{ 251, 235, 29 },
		{ 251, 236, 29 },
		{ 251, 237, 29 },
		{ 251, 238, 29 },
		{ 251, 239, 29 },
		{ 251, 240, 29 },
		{ 252, 132, 29 },
		{ 252, 133, 29 },
		{ 252, 134, 29 },
		{ 252, 135, 29 },
		{ 252, 136, 29 },
		{ 252, 137, 29 },
		{ 252, 138, 29 },
		{ 252, 139, 29 },
		{ 252, 140, 29 },
		{ 252, 141, 29 },
		{ 252, 142, 29 },
		{ 252, 143, 29 },
		{ 252, 144, 29 },
		{ 252, 145, 29 },
		{ 252, 146, 29 },
		{ 252, 147, 29 },
		{ 252, 148, 29 },
		{ 252, 149, 29 },
		{ 252, 150, 29 },
		{ 252, 151, 29 },
		{ 252, 152, 29 },
		{ 252, 153, 29 },
		{ 252, 154, 29 },
		{ 252, 155, 29 },
		{ 252, 156, 29 },
		{ 252, 157, 29 },
		{ 252, 158, 29 },
		{ 252, 159, 29 },
		{ 252, 160, 29 },
		{ 252, 161, 29 },
		{ 252, 162, 29 },
		{ 252, 163, 29 },
		{ 252, 164, 29 },
		{ 252, 165, 29 },
		{ 252, 166, 29 },
		{ 252, 167, 29 },
		{ 252, 168, 29 },
		{ 252, 169, 29 },
		{ 252, 170, 29 },
		{ 252, 171, 29 },
		{ 252, 172, 29 },
		{ 252, 173, 29 },
		{ 252, 174, 29 },
		{ 252, 175, 29 },
		{ 252, 176, 29 },
		{ 252, 177, 29 },
		{ 252, 178, 29 },
		{ 252, 179, 29 },
		{ 252, 180, 29 },
		{ 252, 181, 29 },
		{ 252, 182, 29 },
		{ 252, 183, 29 },
		{ 252, 184, 29 },
		{ 252, 185, 29 },
		{ 252, 186, 29 },
		{ 252, 187, 29 },
		{ 252, 188, 29 },
		{ 252, 189, 29 },
		{ 252, 190, 29 },
		{ 252, 191, 29 },
		{ 252, 192, 29 },
		{ 252, 193, 29 },
		{ 252, 194, 29 },
		{ 252, 195, 29 },
		{ 252, 196, 29 },
		{ 252, 197, 29 },
		{ 252, 198, 29 },
		{ 252, 199, 29 },
		{ 252, 200, 29 },
		{ 252, 201, 29 },
		{ 252, 202, 29 },
		{ 252, 203, 29 },
		{ 252, 204, 29 },
		{ 252, 205, 29 },
		{ 252, 206, 29 },
		{ 252, 207, 29 },
		{ 252, 208, 29 },
		{ 252, 209, 29 },
		{ 252, 210, 29 },
		{ 252, 211, 29 },
		{ 252, 212, 29 },
		{ 252, 213, 29 },
		{ 252, 214, 29 },
		{ 252, 215, 29 },
		{ 252, 216, 29 },
		{ 252, 217, 29 },
		{ 252, 218, 29 },
		{ 252, 219, 29 },
		{ 252, 220, 29 },
		{ 252, 221, 29 },
		{ 252, 222, 29 },
		{ 252, 223, 29 },
		{ 252, 224, 29 },
		{ 252, 225, 29 },
		{ 252, 226, 29 },
		{ 252, 227, 29 },
		{ 252, 228, 29 },
		{ 252, 229, 29 },
		{ 252, 230, 29 },
		{ 252, 231, 29 },
		{ 252, 232, 29 },
		{ 252, 233, 29 },
		{ 252, 234, 29 },
		{ 252, 235, 29 },
		{ 252, 236, 29 },
		{ 252, 237, 29 },
		{ 252, 238, 29 },
		{ 252, 239, 29 },
		{ 252, 240, 29 },
		{ 253, 132, 29 },
		{ 253, 133, 29 },
		{ 253, 134, 29 },
		{ 253, 135, 29 },
		{ 253, 136, 29 },
		{ 253, 137, 29 },
		{ 253, 138, 29 },
		{ 253, 139, 29 },
		{ 253, 140, 29 },
		{ 253, 141, 29 },
		{ 253, 142, 29 },
		{ 253, 143, 29 },
		{ 253, 144, 29 },
		{ 253, 145, 29 },
		{ 253, 146, 29 },
		{ 253, 147, 29 },
		{ 253, 148, 29 },
		{ 253, 149, 29 },
		{ 253, 150, 29 },
		{ 253, 151, 29 },
		{ 253, 152, 29 },
		{ 253, 153, 29 },
		{ 253, 154, 29 },
		{ 253, 155, 29 },
		{ 253, 156, 29 },
		{ 253, 157, 29 },
		{ 253, 158, 29 },
		{ 253, 159, 29 },
		{ 253, 160, 29 },
		{ 253, 161, 29 },
		{ 253, 162, 29 },
		{ 253, 163, 29 },
		{ 253, 164, 29 },
		{ 253, 165, 29 },
		{ 253, 166, 29 },
		{ 253, 167, 29 },
		{ 253, 168, 29 },
		{ 253, 169, 29 },
		{ 253, 170, 29 },
		{ 253, 171, 29 },
		{ 253, 172, 29 },
		{ 253, 173, 29 },
		{ 253, 174, 29 },
		{ 253, 175, 29 },
		{ 253, 176, 29 },
		{ 253, 177, 29 },
		{ 253, 178, 29 },
		{ 253, 179, 29 },
		{ 253, 180, 29 },
		{ 253, 181, 29 },
		{ 253, 182, 29 },
		{ 253, 183, 29 },
		{ 253, 184, 29 },
		{ 253, 185, 29 },
		{ 253, 186, 29 },
		{ 253, 187, 29 },
		{ 253, 188, 29 },
		{ 253, 189, 29 },
		{ 253, 190, 29 },
		{ 253, 191, 29 },
		{ 253, 192, 29 },
		{ 253, 193, 29 },
		{ 253, 194, 29 },
		{ 253, 195, 29 },
		{ 253, 196, 29 },
		{ 253, 197, 29 },
		{ 253, 198, 29 },
		{ 253, 199, 29 },
		{ 253, 200, 29 },
		{ 253, 201, 29 },
		{ 253, 202, 29 },
		{ 253, 203, 29 },
		{ 253, 204, 29 },
		{ 253, 205, 29 },
		{ 253, 206, 29 },
		{ 253, 207, 29 },
		{ 253, 208, 29 },
		{ 253, 209, 29 },
		{ 253, 210, 29 },
		{ 253, 211, 29 },
		{ 253, 212, 29 },
		{ 253, 213, 29 },
		{ 253, 214, 29 },
		{ 253, 215, 29 },
		{ 253, 216, 29 },
		{ 253, 217, 29 },
		{ 253, 218, 29 },
		{ 253, 219, 29 },
		{ 253, 220, 29 },
		{ 253, 221, 29 },
		{ 253, 222, 29 },
		{ 253, 223, 29 },
		{ 253, 224, 29 },
		{ 253, 225, 29 },
		{ 253, 226, 29 },
		{ 253, 227, 29 },
		{ 253, 228, 29 },
		{ 253, 229, 29 },
		{ 253, 230, 29 },
		{ 253, 231, 29 },
		{ 253, 232, 29 },
		{ 253, 233, 29 },
		{ 253, 234, 29 },
		{ 253, 235, 29 },
		{ 253, 236, 29 },
		{ 253, 237, 29 },
		{ 253, 238, 29 },
		{ 253, 239, 29 },
		{ 253, 240, 29 },
		{ 254, 132, 29 },
		{ 254, 133, 29 },
		{ 254, 134, 29 },
		{ 254, 135, 29 },
		{ 254, 136, 29 },
		{ 254, 137, 29 },
		{ 254, 138, 29 },
		{ 254, 139, 29 },
		{ 254, 140, 29 },
		{ 254, 141, 29 },
		{ 254, 142, 29 },
		{ 254, 143, 29 },
		{ 254, 144, 29 },
		{ 254, 145, 29 },
		{ 254, 146, 29 },
		{ 254, 147, 29 },
		{ 254, 148, 29 },
		{ 254, 149, 29 },
		{ 254, 150, 29 },
		{ 254, 151, 29 },
		{ 254, 152, 29 },
		{ 254, 153, 29 },
		{ 254, 154, 29 },
		{ 254, 155, 29 },
		{ 254, 156, 29 },
		{ 254, 157, 29 },
		{ 254, 158, 29 },
		{ 254, 159, 29 },
		{ 254, 160, 29 },
		{ 254, 161, 29 },
		{ 254, 162, 29 },
		{ 254, 163, 29 },
		{ 254, 164, 29 },
		{ 254, 165, 29 },
		{ 254, 166, 29 },
		{ 254, 167, 29 },
		{ 254, 168, 29 },
		{ 254, 169, 29 },
		{ 254, 170, 29 },
		{ 254, 171, 29 },
		{ 254, 172, 29 },
		{ 254, 173, 29 },
		{ 254, 174, 29 },
		{ 254, 175, 29 },
		{ 254, 176, 29 },
		{ 254, 177, 29 },
		{ 254, 178, 29 },
		{ 254, 179, 29 },
		{ 254, 180, 29 },
		{ 254, 181, 29 },
		{ 254, 182, 29 },
		{ 254, 183, 29 },
		{ 254, 184, 29 },
		{ 254, 185, 29 },
		{ 254, 186, 29 },
		{ 254, 187, 29 },
		{ 254, 188, 29 },
		{ 254, 189, 29 },
		{ 254, 190, 29 },
		{ 254, 191, 29 },
		{ 254, 192, 29 },
		{ 254, 193, 29 },
		{ 254, 194, 29 },
		{ 254, 195, 29 },
		{ 254, 196, 29 },
		{ 254, 197, 29 },
		{ 254, 198, 29 },
		{ 254, 199, 29 },
		{ 254, 200, 29 },
		{ 254, 201, 29 },
		{ 254, 202, 29 },
		{ 254, 203, 29 },
		{ 254, 204, 29 },
		{ 254, 205, 29 },
		{ 254, 206, 29 },
		{ 254, 207, 29 },
		{ 254, 208, 29 },
		{ 254, 209, 29 },
		{ 254, 210, 29 },
		{ 254, 211, 29 },
		{ 254, 212, 29 },
		{ 254, 213, 29 },
		{ 254, 214, 29 },
		{ 254, 215, 29 },
		{ 254, 216, 29 },
		{ 254, 217, 29 },
		{ 254, 218, 29 },
		{ 254, 219, 29 },
		{ 254, 220, 29 },
		{ 254, 221, 29 },
		{ 254, 222, 29 },
		{ 254, 223, 29 },
		{ 254, 224, 29 },
		{ 254, 225, 29 },
		{ 254, 226, 29 },
		{ 254, 227, 29 },
		{ 254, 228, 29 },
		{ 254, 229, 29 },
		{ 254, 230, 29 },
		{ 254, 231, 29 },
		{ 254, 232, 29 },
		{ 254, 233, 29 },
		{ 254, 234, 29 },
		{ 254, 235, 29 },
		{ 254, 236, 29 },
		{ 254, 237, 29 },
		{ 254, 238, 29 },
		{ 254, 239, 29 },
		{ 254, 240, 29 },
		{ 255, 133, 29 },
		{ 255, 134, 29 },
		{ 255, 135, 29 },
		{ 255, 136, 29 },
		{ 255, 137, 29 },
		{ 255, 138, 29 },
		{ 255, 139, 29 },
		{ 255, 140, 29 },
		{ 255, 141, 29 },
		{ 255, 142, 29 },
		{ 255, 143, 29 },
		{ 255, 144, 29 },
		{ 255, 145, 29 },
		{ 255, 146, 29 },
		{ 255, 147, 29 },
		{ 255, 148, 29 },
		{ 255, 149, 29 },
		{ 255, 150, 29 },
		{ 255, 151, 29 },
		{ 255, 152, 29 },
		{ 255, 153, 29 },
		{ 255, 154, 29 },
		{ 255, 155, 29 },
		{ 255, 156, 29 },
		{ 255, 157, 29 },
		{ 255, 158, 29 },
		{ 255, 159, 29 },
		{ 255, 160, 29 },
		{ 255, 161, 29 },
		{ 255, 162, 29 },
		{ 255, 163, 29 },
		{ 255, 164, 29 },
		{ 255, 165, 29 },
		{ 255, 166, 29 },
		{ 255, 167, 29 },
		{ 255, 168, 29 },
		{ 255, 169, 29 },
		{ 255, 170, 29 },
		{ 255, 171, 29 },
		{ 255, 172, 29 },
		{ 255, 173, 29 },
		{ 255, 174, 29 },
		{ 255, 175, 29 },
		{ 255, 176, 29 },
		{ 255, 177, 29 },
		{ 255, 178, 29 },
		{ 255, 179, 29 },
		{ 255, 180, 29 },
		{ 255, 181, 29 },
		{ 255, 182, 29 },
		{ 255, 183, 29 },
		{ 255, 184, 29 },
		{ 255, 185, 29 },
		{ 255, 186, 29 },
		{ 255, 187, 29 },
		{ 255, 188, 29 },
		{ 255, 189, 29 },
		{ 255, 190, 29 },
		{ 255, 191, 29 },
		{ 255, 192, 29 },
		{ 255, 193, 29 },
		{ 255, 194, 29 },
		{ 255, 195, 29 },
		{ 255, 196, 29 },
		{ 255, 197, 29 },
		{ 255, 198, 29 },
		{ 255, 199, 29 },
		{ 255, 200, 29 },
		{ 255, 201, 29 },
		{ 255, 202, 29 },
		{ 255, 203, 29 },
		{ 255, 204, 29 },
		{ 255, 205, 29 },
		{ 255, 206, 29 },
		{ 255, 207, 29 },
		{ 255, 208, 29 },
		{ 255, 209, 29 },
		{ 255, 210, 29 },
		{ 255, 211, 29 },
		{ 255, 212, 29 },
		{ 255, 213, 29 },
		{ 255, 214, 29 },
		{ 255, 215, 29 },
		{ 255, 216, 29 },
		{ 255, 217, 29 },
		{ 255, 218, 29 },
		{ 255, 219, 29 },
		{ 255, 220, 29 },
		{ 255, 221, 29 },
		{ 255, 222, 29 },
		{ 255, 223, 29 },
		{ 255, 224, 29 },
		{ 255, 225, 29 },
		{ 255, 226, 29 },
		{ 255, 227, 29 },
		{ 255, 228, 29 },
		{ 255, 229, 29 },
		{ 255, 230, 29 },
		{ 255, 231, 29 },
		{ 255, 232, 29 },
		{ 255, 233, 29 },
		{ 255, 234, 29 },
		{ 255, 235, 29 },
		{ 255, 236, 29 },
		{ 255, 237, 29 },
		{ 255, 238, 29 },
		{ 255, 239, 29 },
		{ 255, 240, 29 },
		{ 256, 134, 29 },
		{ 256, 135, 29 },
		{ 256, 136, 29 },
		{ 256, 137, 29 },
		{ 256, 138, 29 },
		{ 256, 139, 29 },
		{ 256, 140, 29 },
		{ 256, 141, 29 },
		{ 256, 142, 29 },
		{ 256, 143, 29 },
		{ 256, 144, 29 },
		{ 256, 145, 29 },
		{ 256, 146, 29 },
		{ 256, 147, 29 },
		{ 256, 148, 29 },
		{ 256, 149, 29 },
		{ 256, 150, 29 },
		{ 256, 151, 29 },
		{ 256, 152, 29 },
		{ 256, 153, 29 },
		{ 256, 154, 29 },
		{ 256, 155, 29 },
		{ 256, 156, 29 },
		{ 256, 157, 29 },
		{ 256, 158, 29 },
		{ 256, 159, 29 },
		{ 256, 160, 29 },
		{ 256, 161, 29 },
		{ 256, 162, 29 },
		{ 256, 163, 29 },
		{ 256, 164, 29 },
		{ 256, 165, 29 },
		{ 256, 166, 29 },
		{ 256, 167, 29 },
		{ 256, 168, 29 },
		{ 256, 169, 29 },
		{ 256, 170, 29 },
		{ 256, 171, 29 },
		{ 256, 172, 29 },
		{ 256, 173, 29 },
		{ 256, 174, 29 },
		{ 256, 175, 29 },
		{ 256, 176, 29 },
		{ 256, 177, 29 },
		{ 256, 178, 29 },
		{ 256, 179, 29 },
		{ 256, 180, 29 },
		{ 256, 181, 29 },
		{ 256, 182, 29 },
		{ 256, 183, 29 },
		{ 256, 184, 29 },
		{ 256, 185, 29 },
		{ 256, 186, 29 },
		{ 256, 187, 29 },
		{ 256, 188, 29 },
		{ 256, 189, 29 },
		{ 256, 190, 29 },
		{ 256, 191, 29 },
		{ 256, 192, 29 },
		{ 256, 193, 29 },
		{ 256, 194, 29 },
		{ 256, 195, 29 },
		{ 256, 196, 29 },
		{ 256, 197, 29 },
		{ 256, 198, 29 },
		{ 256, 199, 29 },
		{ 256, 200, 29 },
		{ 256, 201, 29 },
		{ 256, 202, 29 },
		{ 256, 203, 29 },
		{ 256, 204, 29 },
		{ 256, 205, 29 },
		{ 256, 206, 29 },
		{ 256, 207, 29 },
		{ 256, 208, 29 },
		{ 256, 209, 29 },
		{ 256, 210, 29 },
		{ 256, 211, 29 },
		{ 256, 212, 29 },
		{ 256, 213, 29 },
		{ 256, 214, 29 },
		{ 256, 215, 29 },
		{ 256, 216, 29 },
		{ 256, 217, 29 },
		{ 256, 218, 29 },
		{ 256, 219, 29 },
		{ 256, 220, 29 },
		{ 256, 221, 29 },
		{ 256, 222, 29 },
		{ 256, 223, 29 },
		{ 256, 224, 29 },
		{ 256, 225, 29 },
		{ 256, 226, 29 },
		{ 256, 227, 29 },
		{ 256, 228, 29 },
		{ 256, 229, 29 },
		{ 256, 230, 29 },
		{ 256, 231, 29 },
		{ 256, 232, 29 },
		{ 256, 233, 29 },
		{ 256, 234, 29 },
		{ 256, 235, 29 },
		{ 256, 236, 29 },
		{ 256, 237, 29 },
		{ 256, 238, 29 },
		{ 256, 239, 29 },
		{ 256, 240, 29 },
		{ 258, 274, 29 },
		{ 258, 275, 29 },
		{ 258, 276, 29 },
		{ 258, 277, 29 },
		{ 258, 278, 29 },
		{ 258, 279, 29 },
		{ 258, 280, 29 },
		{ 258, 281, 29 },
		{ 258, 282, 29 },
		{ 258, 283, 29 },
		{ 258, 284, 29 },
		{ 258, 285, 29 },
		{ 258, 286, 29 },
		{ 258, 287, 29 },
		{ 258, 288, 29 },
		{ 258, 289, 29 },
		{ 258, 290, 29 },
		{ 258, 291, 29 },
		{ 258, 292, 29 },
		{ 258, 293, 29 },
		{ 258, 294, 29 },
		{ 258, 295, 29 },
		{ 258, 296, 29 },
		{ 258, 297, 29 },
		{ 258, 298, 29 },
		{ 258, 299, 29 },
		{ 258, 300, 29 },
		{ 258, 301, 29 },
		{ 258, 302, 29 },
		{ 258, 303, 29 },
		{ 258, 304, 29 },
		{ 258, 305, 29 },
		{ 258, 306, 29 },
		{ 258, 307, 29 },
		{ 258, 308, 29 },
		{ 258, 309, 29 },
		{ 258, 310, 29 },
		{ 258, 311, 29 },
		{ 258, 312, 29 },
		{ 258, 313, 29 },
		{ 258, 314, 29 },
		{ 258, 315, 29 },
		{ 258, 316, 29 },
		{ 258, 317, 29 },
		{ 258, 318, 29 },
		{ 258, 319, 29 },
		{ 258, 320, 29 },
		{ 258, 321, 29 },
		{ 258, 322, 29 },
		{ 258, 323, 29 },
		{ 258, 324, 29 },
		{ 258, 325, 29 },
		{ 258, 326, 29 },
		{ 258, 327, 29 },
		{ 258, 328, 29 },
		{ 258, 329, 29 },
		{ 258, 330, 29 },
		{ 258, 331, 29 },
		{ 258, 332, 29 },
		{ 258, 333, 29 },
		{ 258, 334, 29 },
		{ 258, 335, 29 },
		{ 258, 336, 29 },
		{ 258, 337, 29 },
		{ 258, 338, 29 },
		{ 258, 339, 29 },
		{ 258, 340, 29 },
		{ 258, 341, 29 },
		{ 258, 342, 29 },
		{ 258, 343, 29 },
		{ 258, 344, 29 },
		{ 258, 345, 29 },
		{ 258, 346, 29 },
		{ 258, 347, 29 },
		{ 258, 348, 29 },
		{ 258, 349, 29 },
		{ 258, 350, 29 },
		{ 258, 351, 29 },
		{ 258, 352, 29 },
		{ 258, 353, 29 },
		{ 258, 354, 29 },
		{ 258, 355, 29 },
		{ 258, 356, 29 },
		{ 258, 357, 29 },
		{ 258, 358, 29 },
		{ 258, 359, 29 },
		{ 258, 360, 29 },
		{ 258, 361, 29 },
		{ 258, 362, 29 },
		{ 258, 363, 29 },
		{ 258, 364, 29 },
		{ 258, 365, 29 },
		{ 258, 366, 29 },
		{ 258, 367, 29 },
		{ 258, 368, 29 },
		{ 258, 369, 29 },
		{ 258, 370, 29 },
		{ 258, 371, 29 },
		{ 258, 372, 29 },
		{ 258, 373, 29 },
		{ 258, 374, 29 },
		{ 258, 375, 29 },
		{ 258, 376, 29 },
		{ 258, 377, 29 },
		{ 258, 378, 29 },
		{ 258, 379, 29 },
		{ 258, 380, 29 },
		{ 259, 274, 29 },
		{ 259, 275, 29 },
		{ 259, 276, 29 },
		{ 259, 277, 29 },
		{ 259, 278, 29 },
		{ 259, 279, 29 },
		{ 259, 280, 29 },
		{ 259, 281, 29 },
		{ 259, 282, 29 },
		{ 259, 283, 29 },
		{ 259, 284, 29 },
		{ 259, 285, 29 },
		{ 259, 286, 29 },
		{ 259, 287, 29 },
		{ 259, 288, 29 },
		{ 259, 289, 29 },
		{ 259, 290, 29 },
		{ 259, 291, 29 },
		{ 259, 292, 29 },
		{ 259, 293, 29 },
		{ 259, 294, 29 },
		{ 259, 295, 29 },
		{ 259, 296, 29 },
		{ 259, 297, 29 },
		{ 259, 298, 29 },
		{ 259, 299, 29 },
		{ 259, 300, 29 },
		{ 259, 301, 29 },
		{ 259, 302, 29 },
		{ 259, 303, 29 },
		{ 259, 304, 29 },
		{ 259, 305, 29 },
		{ 259, 306, 29 },
		{ 259, 307, 29 },
		{ 259, 308, 29 },
		{ 259, 309, 29 },
		{ 259, 310, 29 },
		{ 259, 311, 29 },
		{ 259, 312, 29 },
		{ 259, 313, 29 },
		{ 259, 314, 29 },
		{ 259, 315, 29 },
		{ 259, 316, 29 },
		{ 259, 317, 29 },
		{ 259, 318, 29 },
		{ 259, 319, 29 },
		{ 259, 320, 29 },
		{ 259, 321, 29 },
		{ 259, 322, 29 },
		{ 259, 323, 29 },
		{ 259, 324, 29 },
		{ 259, 325, 29 },
		{ 259, 326, 29 },
		{ 259, 327, 29 },
		{ 259, 328, 29 },
		{ 259, 329, 29 },
		{ 259, 330, 29 },
		{ 259, 331, 29 },
		{ 259, 332, 29 },
		{ 259, 333, 29 },
		{ 259, 334, 29 },
		{ 259, 335, 29 },
		{ 259, 336, 29 },
		{ 259, 337, 29 },
		{ 259, 338, 29 },
		{ 259, 339, 29 },
		{ 259, 340, 29 },
		{ 259, 341, 29 },
		{ 259, 342, 29 },
		{ 259, 343, 29 },
		{ 259, 344, 29 },
		{ 259, 345, 29 },
		{ 259, 346, 29 },
		{ 259, 347, 29 },
		{ 259, 348, 29 },
		{ 259, 349, 29 },
		{ 259, 350, 29 },
		{ 259, 351, 29 },
		{ 259, 352, 29 },
		{ 259, 353, 29 },
		{ 259, 354, 29 },
		{ 259, 355, 29 },
		{ 259, 356, 29 },
		{ 259, 357, 29 },
		{ 259, 358, 29 },
		{ 259, 359, 29 },
		{ 259, 360, 29 },
		{ 259, 361, 29 },
		{ 259, 362, 29 },
		{ 259, 363, 29 },
		{ 259, 364, 29 },
		{ 259, 365, 29 },
		{ 259, 366, 29 },
		{ 259, 367, 29 },
		{ 259, 368, 29 },
		{ 259, 369, 29 },
		{ 259, 370, 29 },
		{ 259, 371, 29 },
		{ 259, 372, 29 },
		{ 259, 373, 29 },
		{ 259, 374, 29 },
		{ 259, 375, 29 },
		{ 259, 376, 29 },
		{ 259, 377, 29 },
		{ 259, 378, 29 },
		{ 259, 379, 29 },
		{ 259, 380, 29 },
		{ 259, 381, 29 },
		{ 260, 274, 29 },
		{ 260, 275, 29 },
		{ 260, 276, 29 },
		{ 260, 277, 29 },
		{ 260, 278, 29 },
		{ 260, 279, 29 },
		{ 260, 280, 29 },
		{ 260, 281, 29 },
		{ 260, 282, 29 },
		{ 260, 283, 29 },
		{ 260, 284, 29 },
		{ 260, 285, 29 },
		{ 260, 286, 29 },
		{ 260, 287, 29 },
		{ 260, 288, 29 },
		{ 260, 289, 29 },
		{ 260, 290, 29 },
		{ 260, 291, 29 },
		{ 260, 292, 29 },
		{ 260, 293, 29 },
		{ 260, 294, 29 },
		{ 260, 295, 29 },
		{ 260, 296, 29 },
		{ 260, 297, 29 },
		{ 260, 298, 29 },
		{ 260, 299, 29 },
		{ 260, 300, 29 },
		{ 260, 301, 29 },
		{ 260, 302, 29 },
		{ 260, 303, 29 },
		{ 260, 304, 29 },
		{ 260, 305, 29 },
		{ 260, 306, 29 },
		{ 260, 307, 29 },
		{ 260, 308, 29 },
		{ 260, 309, 29 },
		{ 260, 310, 29 },
		{ 260, 311, 29 },
		{ 260, 312, 29 },
		{ 260, 313, 29 },
		{ 260, 314, 29 },
		{ 260, 315, 29 },
		{ 260, 316, 29 },
		{ 260, 317, 29 },
		{ 260, 318, 29 },
		{ 260, 319, 29 },
		{ 260, 320, 29 },
		{ 260, 321, 29 },
		{ 260, 322, 29 },
		{ 260, 323, 29 },
		{ 260, 324, 29 },
		{ 260, 325, 29 },
		{ 260, 326, 29 },
		{ 260, 327, 29 },
		{ 260, 328, 29 },
		{ 260, 329, 29 },
		{ 260, 330, 29 },
		{ 260, 331, 29 },
		{ 260, 332, 29 },
		{ 260, 333, 29 },
		{ 260, 334, 29 },
		{ 260, 335, 29 },
		{ 260, 336, 29 },
		{ 260, 337, 29 },
		{ 260, 338, 29 },
		{ 260, 339, 29 },
		{ 260, 340, 29 },
		{ 260, 341, 29 },
		{ 260, 342, 29 },
		{ 260, 343, 29 },
		{ 260, 344, 29 },
		{ 260, 345, 29 },
		{ 260, 346, 29 },
		{ 260, 347, 29 },
		{ 260, 348, 29 },
		{ 260, 349, 29 },
		{ 260, 350, 29 },
		{ 260, 351, 29 },
		{ 260, 352, 29 },
		{ 260, 353, 29 },
		{ 260, 354, 29 },
		{ 260, 355, 29 },
		{ 260, 356, 29 },
		{ 260, 357, 29 },
		{ 260, 358, 29 },
		{ 260, 359, 29 },
		{ 260, 360, 29 },
		{ 260, 361, 29 },
		{ 260, 362, 29 },
		{ 260, 363, 29 },
		{ 260, 364, 29 },
		{ 260, 365, 29 },
		{ 260, 366, 29 },
		{ 260, 367, 29 },
		{ 260, 368, 29 },
		{ 260, 369, 29 },
		{ 260, 370, 29 },
		{ 260, 371, 29 },
		{ 260, 372, 29 },
		{ 260, 373, 29 },
		{ 260, 374, 29 },
		{ 260, 375, 29 },
		{ 260, 376, 29 },
		{ 260, 377, 29 },
		{ 260, 378, 29 },
		{ 260, 379, 29 },
		{ 260, 380, 29 },
		{ 260, 381, 29 },
		{ 260, 382, 29 },
		{ 261, 274, 29 },
		{ 261, 275, 29 },
		{ 261, 276, 29 },
		{ 261, 277, 29 },
		{ 261, 278, 29 },
		{ 261, 279, 29 },
		{ 261, 280, 29 },
		{ 261, 281, 29 },
		{ 261, 282, 29 },
		{ 261, 283, 29 },
		{ 261, 284, 29 },
		{ 261, 285, 29 },
		{ 261, 286, 29 },
		{ 261, 287, 29 },
		{ 261, 288, 29 },
		{ 261, 289, 29 },
		{ 261, 290, 29 },
		{ 261, 291, 29 },
		{ 261, 292, 29 },
		{ 261, 293, 29 },
		{ 261, 294, 29 },
		{ 261, 295, 29 },
		{ 261, 296, 29 },
		{ 261, 297, 29 },
		{ 261, 298, 29 },
		{ 261, 299, 29 },
		{ 261, 300, 29 },
		{ 261, 301, 29 },
		{ 261, 302, 29 },
		{ 261, 303, 29 },
		{ 261, 304, 29 },
		{ 261, 305, 29 },
		{ 261, 306, 29 },
		{ 261, 307, 29 },
		{ 261, 308, 29 },
		{ 261, 309, 29 },
		{ 261, 310, 29 },
		{ 261, 311, 29 },
		{ 261, 312, 29 },
		{ 261, 313, 29 },
		{ 261, 314, 29 },
		{ 261, 315, 29 },
		{ 261, 316, 29 },
		{ 261, 317, 29 },
		{ 261, 318, 29 },
		{ 261, 319, 29 },
		{ 261, 320, 29 },
		{ 261, 321, 29 },
		{ 261, 322, 29 },
		{ 261, 323, 29 },
		{ 261, 324, 29 },
		{ 261, 325, 29 },
		{ 261, 326, 29 },
		{ 261, 327, 29 },
		{ 261, 328, 29 },
		{ 261, 329, 29 },
		{ 261, 330, 29 },
		{ 261, 331, 29 },
		{ 261, 332, 29 },
		{ 261, 333, 29 },
		{ 261, 334, 29 },
		{ 261, 335, 29 },
		{ 261, 336, 29 },
		{ 261, 337, 29 },
		{ 261, 338, 29 },
		{ 261, 339, 29 },
		{ 261, 340, 29 },
		{ 261, 341, 29 },
		{ 261, 342, 29 },
		{ 261, 343, 29 },
		{ 261, 344, 29 },
		{ 261, 345, 29 },
		{ 261, 346, 29 },
		{ 261, 347, 29 },
		{ 261, 348, 29 },
		{ 261, 349, 29 },
		{ 261, 350, 29 },
		{ 261, 351, 29 },
		{ 261, 352, 29 },
		{ 261, 353, 29 },
		{ 261, 354, 29 },
		{ 261, 355, 29 },
		{ 261, 356, 29 },
		{ 261, 357, 29 },
		{ 261, 358, 29 },
		{ 261, 359, 29 },
		{ 261, 360, 29 },
		{ 261, 361, 29 },
		{ 261, 362, 29 },
		{ 261, 363, 29 },
		{ 261, 364, 29 },
		{ 261, 365, 29 },
		{ 261, 366, 29 },
		{ 261, 367, 29 },
		{ 261, 368, 29 },
		{ 261, 369, 29 },
		{ 261, 370, 29 },
		{ 261, 371, 29 },
		{ 261, 372, 29 },
		{ 261, 373, 29 },
		{ 261, 374, 29 },
		{ 261, 375, 29 },
		{ 261, 376, 29 },
		{ 261, 377, 29 },
		{ 261, 378, 29 },
		{ 261, 379, 29 },
		{ 261, 380, 29 },
		{ 261, 381, 29 },
		{ 261, 382, 29 },
		{ 262, 274, 29 },
		{ 262, 275, 29 },
		{ 262, 276, 29 },
		{ 262, 277, 29 },
		{ 262, 278, 29 },
		{ 262, 279, 29 },
		{ 262, 280, 29 },
		{ 262, 281, 29 },
		{ 262, 282, 29 },
		{ 262, 283, 29 },
		{ 262, 284, 29 },
		{ 262, 285, 29 },
		{ 262, 286, 29 },
		{ 262, 287, 29 },
		{ 262, 288, 29 },
		{ 262, 289, 29 },
		{ 262, 290, 29 },
		{ 262, 291, 29 },
		{ 262, 292, 29 },
		{ 262, 293, 29 },
		{ 262, 294, 29 },
		{ 262, 295, 29 },
		{ 262, 296, 29 },
		{ 262, 297, 29 },
		{ 262, 298, 29 },
		{ 262, 299, 29 },
		{ 262, 300, 29 },
		{ 262, 301, 29 },
		{ 262, 302, 29 },
		{ 262, 303, 29 },
		{ 262, 304, 29 },
		{ 262, 305, 29 },
		{ 262, 306, 29 },
		{ 262, 307, 29 },
		{ 262, 308, 29 },
		{ 262, 309, 29 },
		{ 262, 310, 29 },
		{ 262, 311, 29 },
		{ 262, 312, 29 },
		{ 262, 313, 29 },
		{ 262, 314, 29 },
		{ 262, 315, 29 },
		{ 262, 316, 29 },
		{ 262, 317, 29 },
		{ 262, 318, 29 },
		{ 262, 319, 29 },
		{ 262, 320, 29 },
		{ 262, 321, 29 },
		{ 262, 322, 29 },
		{ 262, 323, 29 },
		{ 262, 324, 29 },
		{ 262, 325, 29 },
		{ 262, 326, 29 },
		{ 262, 327, 29 },
		{ 262, 328, 29 },
		{ 262, 329, 29 },
		{ 262, 330, 29 },
		{ 262, 331, 29 },
		{ 262, 332, 29 },
		{ 262, 333, 29 },
		{ 262, 334, 29 },
		{ 262, 335, 29 },
		{ 262, 336, 29 },
		{ 262, 337, 29 },
		{ 262, 338, 29 },
		{ 262, 339, 29 },
		{ 262, 340, 29 },
		{ 262, 341, 29 },
		{ 262, 342, 29 },
		{ 262, 343, 29 },
		{ 262, 344, 29 },
		{ 262, 345, 29 },
		{ 262, 346, 29 },
		{ 262, 347, 29 },
		{ 262, 348, 29 },
		{ 262, 349, 29 },
		{ 262, 350, 29 },
		{ 262, 351, 29 },
		{ 262, 352, 29 },
		{ 262, 353, 29 },
		{ 262, 354, 29 },
		{ 262, 355, 29 },
		{ 262, 356, 29 },
		{ 262, 357, 29 },
		{ 262, 358, 29 },
		{ 262, 359, 29 },
		{ 262, 360, 29 },
		{ 262, 361, 29 },
		{ 262, 362, 29 },
		{ 262, 363, 29 },
		{ 262, 364, 29 },
		{ 262, 365, 29 },
		{ 262, 366, 29 },
		{ 262, 367, 29 },
		{ 262, 368, 29 },
		{ 262, 369, 29 },
		{ 262, 370, 29 },
		{ 262, 371, 29 },
		{ 262, 372, 29 },
		{ 262, 373, 29 },
		{ 262, 374, 29 },
		{ 262, 375, 29 },
		{ 262, 376, 29 },
		{ 262, 377, 29 },
		{ 262, 378, 29 },
		{ 262, 379, 29 },
		{ 262, 380, 29 },
		{ 262, 381, 29 },
		{ 262, 382, 29 },
		{ 263, 274, 29 },
		{ 263, 275, 29 },
		{ 263, 276, 29 },
		{ 263, 277, 29 },
		{ 263, 278, 29 },
		{ 263, 279, 29 },
		{ 263, 280, 29 },
		{ 263, 281, 29 },
		{ 263, 282, 29 },
		{ 263, 283, 29 },
		{ 263, 284, 29 },
		{ 263, 285, 29 },
		{ 263, 286, 29 },
		{ 263, 287, 29 },
		{ 263, 288, 29 },
		{ 263, 289, 29 },
		{ 263, 290, 29 },
		{ 263, 291, 29 },
		{ 263, 292, 29 },
		{ 263, 293, 29 },
		{ 263, 294, 29 },
		{ 263, 295, 29 },
		{ 263, 296, 29 },
		{ 263, 297, 29 },
		{ 263, 298, 29 },
		{ 263, 299, 29 },
		{ 263, 300, 29 },
		{ 263, 301, 29 },
		{ 263, 302, 29 },
		{ 263, 303, 29 },
		{ 263, 304, 29 },
		{ 263, 305, 29 },
		{ 263, 306, 29 },
		{ 263, 307, 29 },
		{ 263, 308, 29 },
		{ 263, 309, 29 },
		{ 263, 310, 29 },
		{ 263, 311, 29 },
		{ 263, 312, 29 },
		{ 263, 313, 29 },
		{ 263, 314, 29 },
		{ 263, 315, 29 },
		{ 263, 316, 29 },
		{ 263, 317, 29 },
		{ 263, 318, 29 },
		{ 263, 319, 29 },
		{ 263, 320, 29 },
		{ 263, 321, 29 },
		{ 263, 322, 29 },
		{ 263, 323, 29 },
		{ 263, 324, 29 },
		{ 263, 325, 29 },
		{ 263, 326, 29 },
		{ 263, 327, 29 },
		{ 263, 328, 29 },
		{ 263, 329, 29 },
		{ 263, 330, 29 },
		{ 263, 331, 29 },
		{ 263, 332, 29 },
		{ 263, 333, 29 },
		{ 263, 334, 29 },
		{ 263, 335, 29 },
		{ 263, 336, 29 },
		{ 263, 337, 29 },
		{ 263, 338, 29 },
		{ 263, 339, 29 },
		{ 263, 340, 29 },
		{ 263, 341, 29 },
		{ 263, 342, 29 },
		{ 263, 343, 29 },
		{ 263, 344, 29 },
		{ 263, 345, 29 },
		{ 263, 346, 29 },
		{ 263, 347, 29 },
		{ 263, 348, 29 },
		{ 263, 349, 29 },
		{ 263, 350, 29 },
		{ 263, 351, 29 },
		{ 263, 352, 29 },
		{ 263, 353, 29 },
		{ 263, 354, 29 },
		{ 263, 355, 29 },
		{ 263, 356, 29 },
		{ 263, 357, 29 },
		{ 263, 358, 29 },
		{ 263, 359, 29 },
		{ 263, 360, 29 },
		{ 263, 361, 29 },
		{ 263, 362, 29 },
		{ 263, 363, 29 },
		{ 263, 364, 29 },
		{ 263, 365, 29 },
		{ 263, 366, 29 },
		{ 263, 367, 29 },
		{ 263, 368, 29 },
		{ 263, 369, 29 },
		{ 263, 370, 29 },
		{ 263, 371, 29 },
		{ 263, 372, 29 },
		{ 263, 373, 29 },
		{ 263, 374, 29 },
		{ 263, 375, 29 },
		{ 263, 376, 29 },
		{ 263, 377, 29 },
		{ 263, 378, 29 },
		{ 263, 379, 29 },
		{ 263, 380, 29 },
		{ 263, 381, 29 },
		{ 263, 382, 29 },
		{ 264, 274, 29 },
		{ 264, 275, 29 },
		{ 264, 276, 29 },
		{ 264, 277, 29 },
		{ 264, 278, 29 },
		{ 264, 279, 29 },
		{ 264, 280, 29 },
		{ 264, 281, 29 },
		{ 264, 282, 29 },
		{ 264, 283, 29 },
		{ 264, 284, 29 },
		{ 264, 285, 29 },
		{ 264, 286, 29 },
		{ 264, 287, 29 },
		{ 264, 288, 29 },
		{ 264, 289, 29 },
		{ 264, 290, 29 },
		{ 264, 291, 29 },
		{ 264, 292, 29 },
		{ 264, 293, 29 },
		{ 264, 294, 29 },
		{ 264, 295, 29 },
		{ 264, 296, 29 },
		{ 264, 297, 29 },
		{ 264, 298, 29 },
		{ 264, 299, 29 },
		{ 264, 300, 29 },
		{ 264, 301, 29 },
		{ 264, 302, 29 },
		{ 264, 303, 29 },
		{ 264, 304, 29 },
		{ 264, 305, 29 },
		{ 264, 306, 29 },
		{ 264, 307, 29 },
		{ 264, 308, 29 },
		{ 264, 309, 29 },
		{ 264, 310, 29 },
		{ 264, 311, 29 },
		{ 264, 312, 29 },
		{ 264, 313, 29 },
		{ 264, 314, 29 },
		{ 264, 315, 29 },
		{ 264, 316, 29 },
		{ 264, 317, 29 },
		{ 264, 318, 29 },
		{ 264, 319, 29 },
		{ 264, 320, 29 },
		{ 264, 321, 29 },
		{ 264, 322, 29 },
		{ 264, 323, 29 },
		{ 264, 324, 29 },
		{ 264, 325, 29 },
		{ 264, 326, 29 },
		{ 264, 327, 29 },
		{ 264, 328, 29 },
		{ 264, 329, 29 },
		{ 264, 330, 29 },
		{ 264, 331, 29 },
		{ 264, 332, 29 },
		{ 264, 333, 29 },
		{ 264, 334, 29 },
		{ 264, 335, 29 },
		{ 264, 336, 29 },
		{ 264, 337, 29 },
		{ 264, 338, 29 },
		{ 264, 339, 29 },
		{ 264, 340, 29 },
		{ 264, 341, 29 },
		{ 264, 342, 29 },
		{ 264, 343, 29 },
		{ 264, 344, 29 },
		{ 264, 345, 29 },
		{ 264, 346, 29 },
		{ 264, 347, 29 },
		{ 264, 348, 29 },
		{ 264, 349, 29 },
		{ 264, 350, 29 },
		{ 264, 351, 29 },
		{ 264, 352, 29 },
		{ 264, 353, 29 },
		{ 264, 354, 29 },
		{ 264, 355, 29 },
		{ 264, 356, 29 },
		{ 264, 357, 29 },
		{ 264, 358, 29 },
		{ 264, 359, 29 },
		{ 264, 360, 29 },
		{ 264, 361, 29 },
		{ 264, 362, 29 },
		{ 264, 363, 29 },
		{ 264, 364, 29 },
		{ 264, 365, 29 },
		{ 264, 366, 29 },
		{ 264, 367, 29 },
		{ 264, 368, 29 },
		{ 264, 369, 29 },
		{ 264, 370, 29 },
		{ 264, 371, 29 },
		{ 264, 372, 29 },
		{ 264, 373, 29 },
		{ 264, 374, 29 },
		{ 264, 375, 29 },
		{ 264, 376, 29 },
		{ 264, 377, 29 },
		{ 264, 378, 29 },
		{ 264, 379, 29 },
		{ 264, 380, 29 },
		{ 264, 381, 29 },
		{ 264, 382, 29 },
		{ 264, 383, 29 },
		{ 265, 274, 29 },
		{ 265, 275, 29 },
		{ 265, 276, 29 },
		{ 265, 277, 29 },
		{ 265, 278, 29 },
		{ 265, 279, 29 },
		{ 265, 280, 29 },
		{ 265, 281, 29 },
		{ 265, 282, 29 },
		{ 265, 283, 29 },
		{ 265, 284, 29 },
		{ 265, 285, 29 },
		{ 265, 286, 29 },
		{ 265, 287, 29 },
		{ 265, 288, 29 },
		{ 265, 289, 29 },
		{ 265, 290, 29 },
		{ 265, 291, 29 },
		{ 265, 292, 29 },
		{ 265, 293, 29 },
		{ 265, 294, 29 },
		{ 265, 295, 29 },
		{ 265, 296, 29 },
		{ 265, 297, 29 },
		{ 265, 298, 29 },
		{ 265, 299, 29 },
		{ 265, 300, 29 },
		{ 265, 301, 29 },
		{ 265, 302, 29 },
		{ 265, 303, 29 },
		{ 265, 304, 29 },
		{ 265, 305, 29 },
		{ 265, 306, 29 },
		{ 265, 307, 29 },
		{ 265, 308, 29 },
		{ 265, 309, 29 },
		{ 265, 310, 29 },
		{ 265, 311, 29 },
		{ 265, 312, 29 },
		{ 265, 313, 29 },
		{ 265, 314, 29 },
		{ 265, 315, 29 },
		{ 265, 316, 29 },
		{ 265, 317, 29 },
		{ 265, 318, 29 },
		{ 265, 319, 29 },
		{ 265, 320, 29 },
		{ 265, 321, 29 },
		{ 265, 322, 29 },
		{ 265, 323, 29 },
		{ 265, 324, 29 },
		{ 265, 325, 29 },
		{ 265, 326, 29 },
		{ 265, 327, 29 },
		{ 265, 328, 29 },
		{ 265, 329, 29 },
		{ 265, 330, 29 },
		{ 265, 331, 29 },
		{ 265, 332, 29 },
		{ 265, 333, 29 },
		{ 265, 334, 29 },
		{ 265, 335, 29 },
		{ 265, 336, 29 },
		{ 265, 337, 29 },
		{ 265, 338, 29 },
		{ 265, 339, 29 },
		{ 265, 340, 29 },
		{ 265, 341, 29 },
		{ 265, 342, 29 },
		{ 265, 343, 29 },
		{ 265, 344, 29 },
		{ 265, 345, 29 },
		{ 265, 346, 29 },
		{ 265, 347, 29 },
		{ 265, 348, 29 },
		{ 265, 349, 29 },
		{ 265, 350, 29 },
		{ 265, 351, 29 },
		{ 265, 352, 29 },
		{ 265, 353, 29 },
		{ 265, 354, 29 },
		{ 265, 355, 29 },
		{ 265, 356, 29 },
		{ 265, 357, 29 },
		{ 265, 358, 29 },
		{ 265, 359, 29 },
		{ 265, 360, 29 },
		{ 265, 361, 29 },
		{ 265, 362, 29 },
		{ 265, 363, 29 },
		{ 265, 364, 29 },
		{ 265, 365, 29 },
		{ 265, 366, 29 },
		{ 265, 367, 29 },
		{ 265, 368, 29 },
		{ 265, 369, 29 },
		{ 265, 370, 29 },
		{ 265, 371, 29 },
		{ 265, 372, 29 },
		{ 265, 373, 29 },
		{ 265, 374, 29 },
		{ 265, 375, 29 },
		{ 265, 376, 29 },
		{ 265, 377, 29 },
		{ 265, 378, 29 },
		{ 265, 379, 29 },
		{ 265, 380, 29 },
		{ 265, 381, 29 },
		{ 265, 382, 29 },
		{ 265, 383, 29 },
		{ 266, 276, 29 },
		{ 266, 277, 29 },
		{ 266, 278, 29 },
		{ 266, 279, 29 },
		{ 266, 280, 29 },
		{ 266, 281, 29 },
		{ 266, 282, 29 },
		{ 266, 283, 29 },
		{ 266, 284, 29 },
		{ 266, 285, 29 },
		{ 266, 286, 29 },
		{ 266, 287, 29 },
		{ 266, 288, 29 },
		{ 266, 289, 29 },
		{ 266, 290, 29 },
		{ 266, 291, 29 },
		{ 266, 292, 29 },
		{ 266, 293, 29 },
		{ 266, 294, 29 },
		{ 266, 295, 29 },
		{ 266, 296, 29 },
		{ 266, 297, 29 },
		{ 266, 298, 29 },
		{ 266, 299, 29 },
		{ 266, 300, 29 },
		{ 266, 301, 29 },
		{ 266, 302, 29 },
		{ 266, 303, 29 },
		{ 266, 304, 29 },
		{ 266, 305, 29 },
		{ 266, 306, 29 },
		{ 266, 307, 29 },
		{ 266, 308, 29 },
		{ 266, 309, 29 },
		{ 266, 310, 29 },
		{ 266, 311, 29 },
		{ 266, 312, 29 },
		{ 266, 313, 29 },
		{ 266, 314, 29 },
		{ 266, 315, 29 },
		{ 266, 316, 29 },
		{ 266, 317, 29 },
		{ 266, 318, 29 },
		{ 266, 319, 29 },
		{ 266, 320, 29 },
		{ 266, 321, 29 },
		{ 266, 322, 29 },
		{ 266, 323, 29 },
		{ 266, 324, 29 },
		{ 266, 325, 29 },
		{ 266, 326, 29 },
		{ 266, 327, 29 },
		{ 266, 328, 29 },
		{ 266, 329, 29 },
		{ 266, 330, 29 },
		{ 266, 331, 29 },
		{ 266, 332, 29 },
		{ 266, 333, 29 },
		{ 266, 334, 29 },
		{ 266, 335, 29 },
		{ 266, 336, 29 },
		{ 266, 337, 29 },
		{ 266, 338, 29 },
		{ 266, 339, 29 },
		{ 266, 340, 29 },
		{ 266, 341, 29 },
		{ 266, 342, 29 },
		{ 266, 343, 29 },
		{ 266, 344, 29 },
		{ 266, 345, 29 },
		{ 266, 346, 29 },
		{ 266, 347, 29 },
		{ 266, 348, 29 },
		{ 266, 349, 29 },
		{ 266, 350, 29 },
		{ 266, 351, 29 },
		{ 266, 352, 29 },
		{ 266, 353, 29 },
		{ 266, 354, 29 },
		{ 266, 355, 29 },
		{ 266, 356, 29 },
		{ 266, 357, 29 },
		{ 266, 358, 29 },
		{ 266, 359, 29 },
		{ 266, 360, 29 },
		{ 266, 361, 29 },
		{ 266, 362, 29 },
		{ 266, 363, 29 },
		{ 266, 364, 29 },
		{ 266, 365, 29 },
		{ 266, 366, 29 },
		{ 266, 367, 29 },
		{ 266, 368, 29 },
		{ 266, 369, 29 },
		{ 266, 370, 29 },
		{ 266, 371, 29 },
		{ 266, 372, 29 },
		{ 266, 373, 29 },
		{ 266, 374, 29 },
		{ 266, 375, 29 },
		{ 266, 376, 29 },
		{ 266, 377, 29 },
		{ 266, 378, 29 },
		{ 266, 379, 29 },
		{ 266, 380, 29 },
		{ 266, 381, 29 },
		{ 266, 382, 29 },
		{ 266, 383, 29 },
		{ 267, 278, 29 },
		{ 267, 279, 29 },
		{ 267, 280, 29 },
		{ 267, 281, 29 },
		{ 267, 282, 29 },
		{ 267, 283, 29 },
		{ 267, 284, 29 },
		{ 267, 285, 29 },
		{ 267, 286, 29 },
		{ 267, 287, 29 },
		{ 267, 288, 29 },
		{ 267, 289, 29 },
		{ 267, 290, 29 },
		{ 267, 291, 29 },
		{ 267, 292, 29 },
		{ 267, 293, 29 },
		{ 267, 294, 29 },
		{ 267, 295, 29 },
		{ 267, 296, 29 },
		{ 267, 297, 29 },
		{ 267, 298, 29 },
		{ 267, 299, 29 },
		{ 267, 300, 29 },
		{ 267, 301, 29 },
		{ 267, 302, 29 },
		{ 267, 303, 29 },
		{ 267, 304, 29 },
		{ 267, 305, 29 },
		{ 267, 306, 29 },
		{ 267, 307, 29 },
		{ 267, 308, 29 },
		{ 267, 309, 29 },
		{ 267, 310, 29 },
		{ 267, 311, 29 },
		{ 267, 312, 29 },
		{ 267, 313, 29 },
		{ 267, 314, 29 },
		{ 267, 315, 29 },
		{ 267, 316, 29 },
		{ 267, 317, 29 },
		{ 267, 318, 29 },
		{ 267, 319, 29 },
		{ 267, 320, 29 },
		{ 267, 321, 29 },
		{ 267, 322, 29 },
		{ 267, 323, 29 },
		{ 267, 324, 29 },
		{ 267, 325, 29 },
		{ 267, 326, 29 },
		{ 267, 327, 29 },
		{ 267, 328, 29 },
		{ 267, 329, 29 },
		{ 267, 330, 29 },
		{ 267, 331, 29 },
		{ 267, 332, 29 },
		{ 267, 333, 29 },
		{ 267, 334, 29 },
		{ 267, 335, 29 },
		{ 267, 336, 29 },
		{ 267, 337, 29 },
		{ 267, 338, 29 },
		{ 267, 339, 29 },
		{ 267, 340, 29 },
		{ 267, 341, 29 },
		{ 267, 342, 29 },
		{ 267, 343, 29 },
		{ 267, 344, 29 },
		{ 267, 345, 29 },
		{ 267, 346, 29 },
		{ 267, 347, 29 },
		{ 267, 348, 29 },
		{ 267, 349, 29 },
		{ 267, 350, 29 },
		{ 267, 351, 29 },
		{ 267, 352, 29 },
		{ 267, 353, 29 },
		{ 267, 354, 29 },
		{ 267, 355, 29 },
		{ 267, 356, 29 },
		{ 267, 357, 29 },
		{ 267, 358, 29 },
		{ 267, 359, 29 },
		{ 267, 360, 29 },
		{ 267, 361, 29 },
		{ 267, 362, 29 },
		{ 267, 363, 29 },
		{ 267, 364, 29 },
		{ 267, 365, 29 },
		{ 267, 366, 29 },
		{ 267, 367, 29 },
		{ 267, 368, 29 },
		{ 267, 369, 29 },
		{ 267, 370, 29 },
		{ 267, 371, 29 },
		{ 267, 372, 29 },
		{ 267, 373, 29 },
		{ 267, 374, 29 },
		{ 267, 375, 29 },
		{ 267, 376, 29 },
		{ 267, 377, 29 },
		{ 267, 378, 29 },
		{ 267, 379, 29 },
		{ 267, 380, 29 },
		{ 267, 381, 29 },
		{ 267, 382, 29 },
		{ 267, 383, 29 },
		{ 268, 280, 29 },
		{ 268, 281, 29 },
		{ 268, 282, 29 },
		{ 268, 283, 29 },
		{ 268, 284, 29 },
		{ 268, 285, 29 },
		{ 268, 286, 29 },
		{ 268, 287, 29 },
		{ 268, 288, 29 },
		{ 268, 289, 29 },
		{ 268, 290, 29 },
		{ 268, 291, 29 },
		{ 268, 292, 29 },
		{ 268, 293, 29 },
		{ 268, 294, 29 },
		{ 268, 295, 29 },
		{ 268, 296, 29 },
		{ 268, 297, 29 },
		{ 268, 298, 29 },
		{ 268, 299, 29 },
		{ 268, 300, 29 },
		{ 268, 301, 29 },
		{ 268, 302, 29 },
		{ 268, 303, 29 },
		{ 268, 304, 29 },
		{ 268, 305, 29 },
		{ 268, 306, 29 },
		{ 268, 307, 29 },
		{ 268, 308, 29 },
		{ 268, 309, 29 },
		{ 268, 310, 29 },
		{ 268, 311, 29 },
		{ 268, 312, 29 },
		{ 268, 313, 29 },
		{ 268, 314, 29 },
		{ 268, 315, 29 },
		{ 268, 316, 29 },
		{ 268, 317, 29 },
		{ 268, 318, 29 },
		{ 268, 319, 29 },
		{ 268, 320, 29 },
		{ 268, 321, 29 },
		{ 268, 322, 29 },
		{ 268, 323, 29 },
		{ 268, 324, 29 },
		{ 268, 325, 29 },
		{ 268, 326, 29 },
		{ 268, 327, 29 },
		{ 268, 328, 29 },
		{ 268, 329, 29 },
		{ 268, 330, 29 },
		{ 268, 331, 29 },
		{ 268, 332, 29 },
		{ 268, 333, 29 },
		{ 268, 334, 29 },
		{ 268, 335, 29 },
		{ 268, 336, 29 },
		{ 268, 337, 29 },
		{ 268, 338, 29 },
		{ 268, 339, 29 },
		{ 268, 340, 29 },
		{ 268, 341, 29 },
		{ 268, 342, 29 },
		{ 268, 343, 29 },
		{ 268, 344, 29 },
		{ 268, 345, 29 },
		{ 268, 346, 29 },
		{ 268, 347, 29 },
		{ 268, 348, 29 },
		{ 268, 349, 29 },
		{ 268, 350, 29 },
		{ 268, 351, 29 },
		{ 268, 352, 29 },
		{ 268, 353, 29 },
		{ 268, 354, 29 },
		{ 268, 355, 29 },
		{ 268, 356, 29 },
		{ 268, 357, 29 },
		{ 268, 358, 29 },
		{ 268, 359, 29 },
		{ 268, 360, 29 },
		{ 268, 361, 29 },
		{ 268, 362, 29 },
		{ 268, 363, 29 },
		{ 268, 364, 29 },
		{ 268, 365, 29 },
		{ 268, 366, 29 },
		{ 268, 367, 29 },
		{ 268, 368, 29 },
		{ 268, 369, 29 },
		{ 268, 370, 29 },
		{ 268, 371, 29 },
		{ 268, 372, 29 },
		{ 268, 373, 29 },
		{ 268, 374, 29 },
		{ 268, 375, 29 },
		{ 268, 376, 29 },
		{ 268, 377, 29 },
		{ 268, 378, 29 },
		{ 268, 379, 29 },
		{ 268, 380, 29 },
		{ 268, 381, 29 },
		{ 268, 382, 29 },
		{ 268, 383, 29 },
		{ 269, 282, 29 },
		{ 269, 283, 29 },
		{ 269, 284, 29 },
		{ 269, 285, 29 },
		{ 269, 286, 29 },
		{ 269, 287, 29 },
		{ 269, 288, 29 },
		{ 269, 289, 29 },
		{ 269, 290, 29 },
		{ 269, 291, 29 },
		{ 269, 292, 29 },
		{ 269, 293, 29 },
		{ 269, 294, 29 },
		{ 269, 295, 29 },
		{ 269, 296, 29 },
		{ 269, 297, 29 },
		{ 269, 298, 29 },
		{ 269, 299, 29 },
		{ 269, 300, 29 },
		{ 269, 301, 29 },
		{ 269, 302, 29 },
		{ 269, 303, 29 },
		{ 269, 304, 29 },
		{ 269, 305, 29 },
		{ 269, 306, 29 },
		{ 269, 307, 29 },
		{ 269, 308, 29 },
		{ 269, 309, 29 },
		{ 269, 310, 29 },
		{ 269, 311, 29 },
		{ 269, 312, 29 },
		{ 269, 313, 29 },
		{ 269, 314, 29 },
		{ 269, 315, 29 },
		{ 269, 316, 29 },
		{ 269, 317, 29 },
		{ 269, 318, 29 },
		{ 269, 319, 29 },
		{ 269, 320, 29 },
		{ 269, 321, 29 },
		{ 269, 322, 29 },
		{ 269, 323, 29 },
		{ 269, 324, 29 },
		{ 269, 325, 29 },
		{ 269, 326, 29 },
		{ 269, 327, 29 },
		{ 269, 328, 29 },
		{ 269, 329, 29 },
		{ 269, 330, 29 },
		{ 269, 331, 29 },
		{ 269, 332, 29 },
		{ 269, 333, 29 },
		{ 269, 334, 29 },
		{ 269, 335, 29 },
		{ 269, 336, 29 },
		{ 269, 337, 29 },
		{ 269, 338, 29 },
		{ 269, 339, 29 },
		{ 269, 340, 29 },
		{ 269, 341, 29 },
		{ 269, 342, 29 },
		{ 269, 343, 29 },
		{ 269, 344, 29 },
		{ 269, 345, 29 },
		{ 269, 346, 29 },
		{ 269, 347, 29 },
		{ 269, 348, 29 },
		{ 269, 349, 29 },
		{ 269, 350, 29 },
		{ 269, 351, 29 },
		{ 269, 352, 29 },
		{ 269, 353, 29 },
		{ 269, 354, 29 },
		{ 269, 355, 29 },
		{ 269, 356, 29 },
		{ 269, 357, 29 },
		{ 269, 358, 29 },
		{ 269, 359, 29 },
		{ 269, 360, 29 },
		{ 269, 361, 29 },
		{ 269, 362, 29 },
		{ 269, 363, 29 },
		{ 269, 364, 29 },
		{ 269, 365, 29 },
		{ 269, 366, 29 },
		{ 269, 367, 29 },
		{ 269, 368, 29 },
		{ 269, 369, 29 },
		{ 269, 370, 29 },
		{ 269, 371, 29 },
		{ 269, 372, 29 },
		{ 269, 373, 29 },
		{ 269, 374, 29 },
		{ 269, 375, 29 },
		{ 269, 376, 29 },
		{ 269, 377, 29 },
		{ 269, 378, 29 },
		{ 269, 379, 29 },
		{ 269, 380, 29 },
		{ 269, 381, 29 },
		{ 269, 382, 29 },
		{ 269, 383, 29 },
		{ 270, 284, 29 },
		{ 270, 285, 29 },
		{ 270, 286, 29 },
		{ 270, 287, 29 },
		{ 270, 288, 29 },
		{ 270, 289, 29 },
		{ 270, 290, 29 },
		{ 270, 291, 29 },
		{ 270, 292, 29 },
		{ 270, 293, 29 },
		{ 270, 294, 29 },
		{ 270, 295, 29 },
		{ 270, 296, 29 },
		{ 270, 297, 29 },
		{ 270, 298, 29 },
		{ 270, 299, 29 },
		{ 270, 300, 29 },
		{ 270, 301, 29 },
		{ 270, 302, 29 },
		{ 270, 303, 29 },
		{ 270, 304, 29 },
		{ 270, 305, 29 },
		{ 270, 306, 29 },
		{ 270, 307, 29 },
		{ 270, 308, 29 },
		{ 270, 309, 29 },
		{ 270, 310, 29 },
		{ 270, 311, 29 },
		{ 270, 312, 29 },
		{ 270, 313, 29 },
		{ 270, 314, 29 },
		{ 270, 315, 29 },
		{ 270, 316, 29 },
		{ 270, 317, 29 },
		{ 270, 318, 29 },
		{ 270, 319, 29 },
		{ 270, 320, 29 },
		{ 270, 321, 29 },
		{ 270, 322, 29 },
		{ 270, 323, 29 },
		{ 270, 324, 29 },
		{ 270, 325, 29 },
		{ 270, 326, 29 },
		{ 270, 327, 29 },
		{ 270, 328, 29 },
		{ 270, 329, 29 },
		{ 270, 330, 29 },
		{ 270, 331, 29 },
		{ 270, 332, 29 },
		{ 270, 333, 29 },
		{ 270, 334, 29 },
		{ 270, 335, 29 },
		{ 270, 336, 29 },
		{ 270, 337, 29 },
		{ 270, 338, 29 },
		{ 270, 339, 29 },
		{ 270, 340, 29 },
		{ 270, 341, 29 },
		{ 270, 342, 29 },
		{ 270, 343, 29 },
		{ 270, 344, 29 },
		{ 270, 345, 29 },
		{ 270, 346, 29 },
		{ 270, 347, 29 },
		{ 270, 348, 29 },
		{ 270, 349, 29 },
		{ 270, 350, 29 },
		{ 270, 351, 29 },
		{ 270, 352, 29 },
		{ 270, 353, 29 },
		{ 270, 354, 29 },
		{ 270, 355, 29 },
		{ 270, 356, 29 },
		{ 270, 357, 29 },
		{ 270, 358, 29 },
		{ 270, 359, 29 },
		{ 270, 360, 29 },
		{ 270, 361, 29 },
		{ 270, 362, 29 },
		{ 270, 363, 29 },
		{ 270, 364, 29 },
		{ 270, 365, 29 },
		{ 270, 366, 29 },
		{ 270, 367, 29 },
		{ 270, 368, 29 },
		{ 270, 369, 29 },
		{ 270, 370, 29 },
		{ 270, 371, 29 },
		{ 270, 372, 29 },
		{ 270, 373, 29 },
		{ 270, 374, 29 },
		{ 270, 375, 29 },
		{ 270, 376, 29 },
		{ 270, 377, 29 },
		{ 270, 378, 29 },
		{ 270, 379, 29 },
		{ 270, 380, 29 },
		{ 270, 381, 29 },
		{ 270, 382, 29 },
		{ 270, 383, 29 },
		{ 271, 286, 29 },
		{ 271, 287, 29 },
		{ 271, 288, 29 },
		{ 271, 289, 29 },
		{ 271, 290, 29 },
		{ 271, 291, 29 },
		{ 271, 292, 29 },
		{ 271, 293, 29 },
		{ 271, 294, 29 },
		{ 271, 295, 29 },
		{ 271, 296, 29 },
		{ 271, 297, 29 },
		{ 271, 298, 29 },
		{ 271, 299, 29 },
		{ 271, 300, 29 },
		{ 271, 301, 29 },
		{ 271, 302, 29 },
		{ 271, 303, 29 },
		{ 271, 304, 29 },
		{ 271, 305, 29 },
		{ 271, 306, 29 },
		{ 271, 307, 29 },
		{ 271, 308, 29 },
		{ 271, 309, 29 },
		{ 271, 310, 29 },
		{ 271, 311, 29 },
		{ 271, 312, 29 },
		{ 271, 313, 29 },
		{ 271, 314, 29 },
		{ 271, 315, 29 },
		{ 271, 316, 29 },
		{ 271, 317, 29 },
		{ 271, 318, 29 },
		{ 271, 319, 29 },
		{ 271, 320, 29 },
		{ 271, 321, 29 },
		{ 271, 322, 29 },
		{ 271, 323, 29 },
		{ 271, 324, 29 },
		{ 271, 325, 29 },
		{ 271, 326, 29 },
		{ 271, 327, 29 },
		{ 271, 328, 29 },
		{ 271, 329, 29 },
		{ 271, 330, 29 },
		{ 271, 331, 29 },
		{ 271, 332, 29 },
		{ 271, 333, 29 },
		{ 271, 334, 29 },
		{ 271, 335, 29 },
		{ 271, 336, 29 },
		{ 271, 337, 29 },
		{ 271, 338, 29 },
		{ 271, 339, 29 },
		{ 271, 340, 29 },
		{ 271, 341, 29 },
		{ 271, 342, 29 },
		{ 271, 343, 29 },
		{ 271, 344, 29 },
		{ 271, 345, 29 },
		{ 271, 346, 29 },
		{ 271, 347, 29 },
		{ 271, 348, 29 },
		{ 271, 349, 29 },
		{ 271, 350, 29 },
		{ 271, 351, 29 },
		{ 271, 352, 29 },
		{ 271, 353, 29 },
		{ 271, 354, 29 },
		{ 271, 355, 29 },
		{ 271, 356, 29 },
		{ 271, 357, 29 },
		{ 271, 358, 29 },
		{ 271, 359, 29 },
		{ 271, 360, 29 },
		{ 271, 361, 29 },
		{ 271, 362, 29 },
		{ 271, 363, 29 },
		{ 271, 364, 29 },
		{ 271, 365, 29 },
		{ 271, 366, 29 },
		{ 271, 367, 29 },
		{ 271, 368, 29 },
		{ 271, 369, 29 },
		{ 271, 370, 29 },
		{ 271, 371, 29 },
		{ 271, 372, 29 },
		{ 271, 373, 29 },
		{ 271, 374, 29 },
		{ 271, 375, 29 },
		{ 271, 376, 29 },
		{ 271, 377, 29 },
		{ 271, 378, 29 },
		{ 271, 379, 29 },
		{ 271, 380, 29 },
		{ 271, 381, 29 },
		{ 271, 382, 29 },
		{ 271, 383, 29 },
		{ 272, 288, 29 },
		{ 272, 289, 29 },
		{ 272, 290, 29 },
		{ 272, 291, 29 },
		{ 272, 292, 29 },
		{ 272, 293, 29 },
		{ 272, 294, 29 },
		{ 272, 295, 29 },
		{ 272, 296, 29 },
		{ 272, 297, 29 },
		{ 272, 298, 29 },
		{ 272, 299, 29 },
		{ 272, 300, 29 },
		{ 272, 301, 29 },
		{ 272, 302, 29 },
		{ 272, 303, 29 },
		{ 272, 304, 29 },
		{ 272, 305, 29 },
		{ 272, 306, 29 },
		{ 272, 307, 29 },
		{ 272, 308, 29 },
		{ 272, 309, 29 },
		{ 272, 310, 29 },
		{ 272, 311, 29 },
		{ 272, 312, 29 },
		{ 272, 313, 29 },
		{ 272, 314, 29 },
		{ 272, 315, 29 },
		{ 272, 316, 29 },
		{ 272, 317, 29 },
		{ 272, 318, 29 },
		{ 272, 319, 29 },
		{ 272, 320, 29 },
		{ 272, 321, 29 },
		{ 272, 322, 29 },
		{ 272, 323, 29 },
		{ 272, 324, 29 },
		{ 272, 325, 29 },
		{ 272, 326, 29 },
		{ 272, 327, 29 },
		{ 272, 328, 29 },
		{ 272, 329, 29 },
		{ 272, 330, 29 },
		{ 272, 331, 29 },
		{ 272, 332, 29 },
		{ 272, 333, 29 },
		{ 272, 334, 29 },
		{ 272, 335, 29 },
		{ 272, 336, 29 },
		{ 272, 337, 29 },
		{ 272, 338, 29 },
		{ 272, 339, 29 },
		{ 272, 340, 29 },
		{ 272, 341, 29 },
		{ 272, 342, 29 },
		{ 272, 343, 29 },
		{ 272, 344, 29 },
		{ 272, 345, 29 },
		{ 272, 346, 29 },
		{ 272, 347, 29 },
		{ 272, 348, 29 },
		{ 272, 349, 29 },
		{ 272, 350, 29 },
		{ 272, 351, 29 },
		{ 272, 352, 29 },
		{ 272, 353, 29 },
		{ 272, 354, 29 },
		{ 272, 355, 29 },
		{ 272, 356, 29 },
		{ 272, 357, 29 },
		{ 272, 358, 29 },
		{ 272, 359, 29 },
		{ 272, 360, 29 },
		{ 272, 361, 29 },
		{ 272, 362, 29 },
		{ 272, 363, 29 },
		{ 272, 364, 29 },
		{ 272, 365, 29 },
		{ 272, 366, 29 },
		{ 272, 367, 29 },
		{ 272, 368, 29 },
		{ 272, 369, 29 },
		{ 272, 370, 29 },
		{ 272, 371, 29 },
		{ 272, 372, 29 },
		{ 272, 373, 29 },
		{ 272, 374, 29 },
		{ 272, 375, 29 },
		{ 272, 376, 29 },
		{ 272, 377, 29 },
		{ 272, 378, 29 },
		{ 272, 379, 29 },
		{ 272, 380, 29 },
		{ 272, 381, 29 },
		{ 272, 382, 29 },
		{ 272, 383, 29 },
		{ 273, 290, 29 },
		{ 273, 291, 29 },
		{ 273, 292, 29 },
		{ 273, 293, 29 },
		{ 273, 294, 29 },
		{ 273, 295, 29 },
		{ 273, 296, 29 },
		{ 273, 297, 29 },
		{ 273, 298, 29 },
		{ 273, 299, 29 },
		{ 273, 300, 29 },
		{ 273, 301, 29 },
		{ 273, 302, 29 },
		{ 273, 303, 29 },
		{ 273, 304, 29 },
		{ 273, 305, 29 },
		{ 273, 306, 29 },
		{ 273, 307, 29 },
		{ 273, 308, 29 },
		{ 273, 309, 29 },
		{ 273, 310, 29 },
		{ 273, 311, 29 },
		{ 273, 312, 29 },
		{ 273, 313, 29 },
		{ 273, 314, 29 },
		{ 273, 315, 29 },
		{ 273, 316, 29 },
		{ 273, 317, 29 },
		{ 273, 318, 29 },
		{ 273, 319, 29 },
		{ 273, 320, 29 },
		{ 273, 321, 29 },
		{ 273, 322, 29 },
		{ 273, 323, 29 },
		{ 273, 324, 29 },
		{ 273, 325, 29 },
		{ 273, 326, 29 },
		{ 273, 327, 29 },
		{ 273, 328, 29 },
		{ 273, 329, 29 },
		{ 273, 330, 29 },
		{ 273, 331, 29 },
		{ 273, 332, 29 },
		{ 273, 333, 29 },
		{ 273, 334, 29 },
		{ 273, 335, 29 },
		{ 273, 336, 29 },
		{ 273, 337, 29 },
		{ 273, 338, 29 },
		{ 273, 339, 29 },
		{ 273, 340, 29 },
		{ 273, 341, 29 },
		{ 273, 342, 29 },
		{ 273, 343, 29 },
		{ 273, 344, 29 },
		{ 273, 345, 29 },
		{ 273, 346, 29 },
		{ 273, 347, 29 },
		{ 273, 348, 29 },
		{ 273, 349, 29 },
		{ 273, 350, 29 },
		{ 273, 351, 29 },
		{ 273, 352, 29 },
		{ 273, 353, 29 },
		{ 273, 354, 29 },
		{ 273, 355, 29 },
		{ 273, 356, 29 },
		{ 273, 357, 29 },
		{ 273, 358, 29 },
		{ 273, 359, 29 },
		{ 273, 360, 29 },
		{ 273, 361, 29 },
		{ 273, 362, 29 },
		{ 273, 363, 29 },
		{ 273, 364, 29 },
		{ 273, 365, 29 },
		{ 273, 366, 29 },
		{ 273, 367, 29 },
		{ 273, 368, 29 },
		{ 273, 369, 29 },
		{ 273, 370, 29 },
		{ 273, 371, 29 },
		{ 273, 372, 29 },
		{ 273, 373, 29 },
		{ 273, 374, 29 },
		{ 273, 375, 29 },
		{ 273, 376, 29 },
		{ 273, 377, 29 },
		{ 273, 378, 29 },
		{ 273, 379, 29 },
		{ 273, 380, 29 },
		{ 273, 381, 29 },
		{ 273, 382, 29 },
		{ 273, 383, 29 },
		{ 274, 292, 29 },
		{ 274, 293, 29 },
		{ 274, 294, 29 },
		{ 274, 295, 29 },
		{ 274, 296, 29 },
		{ 274, 297, 29 },
		{ 274, 298, 29 },
		{ 274, 299, 29 },
		{ 274, 300, 29 },
		{ 274, 301, 29 },
		{ 274, 302, 29 },
		{ 274, 303, 29 },
		{ 274, 304, 29 },
		{ 274, 305, 29 },
		{ 274, 306, 29 },
		{ 274, 307, 29 },
		{ 274, 308, 29 },
		{ 274, 309, 29 },
		{ 274, 310, 29 },
		{ 274, 311, 29 },
		{ 274, 312, 29 },
		{ 274, 313, 29 },
		{ 274, 314, 29 },
		{ 274, 315, 29 },
		{ 274, 316, 29 },
		{ 274, 317, 29 },
		{ 274, 318, 29 },
		{ 274, 319, 29 },
		{ 274, 320, 29 },
		{ 274, 321, 29 },
		{ 274, 322, 29 },
		{ 274, 323, 29 },
		{ 274, 324, 29 },
		{ 274, 325, 29 },
		{ 274, 326, 29 },
		{ 274, 327, 29 },
		{ 274, 328, 29 },
		{ 274, 329, 29 },
		{ 274, 330, 29 },
		{ 274, 331, 29 },
		{ 274, 332, 29 },
		{ 274, 333, 29 },
		{ 274, 334, 29 },
		{ 274, 335, 29 },
		{ 274, 336, 29 },
		{ 274, 337, 29 },
		{ 274, 338, 29 },
		{ 274, 339, 29 },
		{ 274, 340, 29 },
		{ 274, 341, 29 },
		{ 274, 342, 29 },
		{ 274, 343, 29 },
		{ 274, 344, 29 },
		{ 274, 345, 29 },
		{ 274, 346, 29 },
		{ 274, 347, 29 },
		{ 274, 348, 29 },
		{ 274, 349, 29 },
		{ 274, 350, 29 },
		{ 274, 351, 29 },
		{ 274, 352, 29 },
		{ 274, 353, 29 },
		{ 274, 354, 29 },
		{ 274, 355, 29 },
		{ 274, 356, 29 },
		{ 274, 357, 29 },
		{ 274, 358, 29 },
		{ 274, 359, 29 },
		{ 274, 360, 29 },
		{ 274, 361, 29 },
		{ 274, 362, 29 },
		{ 274, 363, 29 },
		{ 274, 364, 29 },
		{ 274, 365, 29 },
		{ 274, 366, 29 },
		{ 274, 367, 29 },
		{ 274, 368, 29 },
		{ 274, 369, 29 },
		{ 274, 370, 29 },
		{ 274, 371, 29 },
		{ 274, 372, 29 },
		{ 274, 373, 29 },
		{ 274, 374, 29 },
		{ 274, 375, 29 },
		{ 274, 376, 29 },
		{ 274, 377, 29 },
		{ 274, 378, 29 },
		{ 274, 379, 29 },
		{ 274, 380, 29 },
		{ 274, 381, 29 },
		{ 274, 382, 29 },
		{ 274, 383, 29 },
		{ 275, 294, 29 },
		{ 275, 295, 29 },
		{ 275, 296, 29 },
		{ 275, 297, 29 },
		{ 275, 298, 29 },
		{ 275, 299, 29 },
		{ 275, 300, 29 },
		{ 275, 301, 29 },
		{ 275, 302, 29 },
		{ 275, 303, 29 },
		{ 275, 304, 29 },
		{ 275, 305, 29 },
		{ 275, 306, 29 },
		{ 275, 307, 29 },
		{ 275, 308, 29 },
		{ 275, 309, 29 },
		{ 275, 310, 29 },
		{ 275, 311, 29 },
		{ 275, 312, 29 },
		{ 275, 313, 29 },
		{ 275, 314, 29 },
		{ 275, 315, 29 },
		{ 275, 316, 29 },
		{ 275, 317, 29 },
		{ 275, 318, 29 },
		{ 275, 319, 29 },
		{ 275, 320, 29 },
		{ 275, 321, 29 },
		{ 275, 322, 29 },
		{ 275, 323, 29 },
		{ 275, 324, 29 },
		{ 275, 325, 29 },
		{ 275, 326, 29 },
		{ 275, 327, 29 },
		{ 275, 328, 29 },
		{ 275, 329, 29 },
		{ 275, 330, 29 },
		{ 275, 331, 29 },
		{ 275, 332, 29 },
		{ 275, 333, 29 },
		{ 275, 334, 29 },
		{ 275, 335, 29 },
		{ 275, 336, 29 },
		{ 275, 337, 29 },
		{ 275, 338, 29 },
		{ 275, 339, 29 },
		{ 275, 340, 29 },
		{ 275, 341, 29 },
		{ 275, 342, 29 },
		{ 275, 343, 29 },
		{ 275, 344, 29 },
		{ 275, 345, 29 },
		{ 275, 346, 29 },
		{ 275, 347, 29 },
		{ 275, 348, 29 },
		{ 275, 349, 29 },
		{ 275, 350, 29 },
		{ 275, 351, 29 },
		{ 275, 352, 29 },
		{ 275, 353, 29 },
		{ 275, 354, 29 },
		{ 275, 355, 29 },
		{ 275, 356, 29 },
		{ 275, 357, 29 },
		{ 275, 358, 29 },
		{ 275, 359, 29 },
		{ 275, 360, 29 },
		{ 275, 361, 29 },
		{ 275, 362, 29 },
		{ 275, 363, 29 },
		{ 275, 364, 29 },
		{ 275, 365, 29 },
		{ 275, 366, 29 },
		{ 275, 367, 29 },
		{ 275, 368, 29 },
		{ 275, 369, 29 },
		{ 275, 370, 29 },
		{ 275, 371, 29 },
		{ 275, 372, 29 },
		{ 275, 373, 29 },
		{ 275, 374, 29 },
		{ 275, 375, 29 },
		{ 275, 376, 29 },
		{ 275, 377, 29 },
		{ 275, 378, 29 },
		{ 275, 379, 29 },
		{ 275, 380, 29 },
		{ 275, 381, 29 },
		{ 275, 382, 29 },
		{ 275, 383, 29 },
		{ 276, 296, 29 },
		{ 276, 297, 29 },
		{ 276, 298, 29 },
		{ 276, 299, 29 },
		{ 276, 300, 29 },
		{ 276, 301, 29 },
		{ 276, 302, 29 },
		{ 276, 303, 29 },
		{ 276, 304, 29 },
		{ 276, 305, 29 },
		{ 276, 306, 29 },
		{ 276, 307, 29 },
		{ 276, 308, 29 },
		{ 276, 309, 29 },
		{ 276, 310, 29 },
		{ 276, 311, 29 },
		{ 276, 312, 29 },
		{ 276, 313, 29 },
		{ 276, 314, 29 },
		{ 276, 315, 29 },
		{ 276, 316, 29 },
		{ 276, 317, 29 },
		{ 276, 318, 29 },
		{ 276, 319, 29 },
		{ 276, 320, 29 },
		{ 276, 321, 29 },
		{ 276, 322, 29 },
		{ 276, 323, 29 },
		{ 276, 324, 29 },
		{ 276, 325, 29 },
		{ 276, 326, 29 },
		{ 276, 327, 29 },
		{ 276, 328, 29 },
		{ 276, 329, 29 },
		{ 276, 330, 29 },
		{ 276, 331, 29 },
		{ 276, 332, 29 },
		{ 276, 333, 29 },
		{ 276, 334, 29 },
		{ 276, 335, 29 },
		{ 276, 336, 29 },
		{ 276, 337, 29 },
		{ 276, 338, 29 },
		{ 276, 339, 29 },
		{ 276, 340, 29 },
		{ 276, 341, 29 },
		{ 276, 342, 29 },
		{ 276, 343, 29 },
		{ 276, 344, 29 },
		{ 276, 345, 29 },
		{ 276, 346, 29 },
		{ 276, 347, 29 },
		{ 276, 348, 29 },
		{ 276, 349, 29 },
		{ 276, 350, 29 },
		{ 276, 351, 29 },
		{ 276, 352, 29 },
		{ 276, 353, 29 },
		{ 276, 354, 29 },
		{ 276, 355, 29 },
		{ 276, 356, 29 },
		{ 276, 357, 29 },
		{ 276, 358, 29 },
		{ 276, 359, 29 },
		{ 276, 360, 29 },
		{ 276, 361, 29 },
		{ 276, 362, 29 },
		{ 276, 363, 29 },
		{ 276, 364, 29 },
		{ 276, 365, 29 },
		{ 276, 366, 29 },
		{ 276, 367, 29 },
		{ 276, 368, 29 },
		{ 276, 369, 29 },
		{ 276, 370, 29 },
		{ 276, 371, 29 },
		{ 276, 372, 29 },
		{ 276, 373, 29 },
		{ 276, 374, 29 },
		{ 276, 375, 29 },
		{ 276, 376, 29 },
		{ 276, 377, 29 },
		{ 276, 378, 29 },
		{ 276, 379, 29 },
		{ 276, 380, 29 },
		{ 276, 381, 29 },
		{ 276, 382, 29 },
		{ 276, 383, 29 },
		{ 277, 298, 29 },
		{ 277, 299, 29 },
		{ 277, 300, 29 },
		{ 277, 301, 29 },
		{ 277, 302, 29 },
		{ 277, 303, 29 },
		{ 277, 304, 29 },
		{ 277, 305, 29 },
		{ 277, 306, 29 },
		{ 277, 307, 29 },
		{ 277, 308, 29 },
		{ 277, 309, 29 },
		{ 277, 310, 29 },
		{ 277, 311, 29 },
		{ 277, 312, 29 },
		{ 277, 313, 29 },
		{ 277, 314, 29 },
		{ 277, 315, 29 },
		{ 277, 316, 29 },
		{ 277, 317, 29 },
		{ 277, 318, 29 },
		{ 277, 319, 29 },
		{ 277, 320, 29 },
		{ 277, 321, 29 },
		{ 277, 322, 29 },
		{ 277, 323, 29 },
		{ 277, 324, 29 },
		{ 277, 325, 29 },
		{ 277, 326, 29 },
		{ 277, 327, 29 },
		{ 277, 328, 29 },
		{ 277, 329, 29 },
		{ 277, 330, 29 },
		{ 277, 331, 29 },
		{ 277, 332, 29 },
		{ 277, 333, 29 },
		{ 277, 334, 29 },
		{ 277, 335, 29 },
		{ 277, 336, 29 },
		{ 277, 337, 29 },
		{ 277, 338, 29 },
		{ 277, 339, 29 },
		{ 277, 340, 29 },
		{ 277, 341, 29 },
		{ 277, 342, 29 },
		{ 277, 343, 29 },
		{ 277, 344, 29 },
		{ 277, 345, 29 },
		{ 277, 346, 29 },
		{ 277, 347, 29 },
		{ 277, 348, 29 },
		{ 277, 349, 29 },
		{ 277, 350, 29 },
		{ 277, 351, 29 },
		{ 277, 352, 29 },
		{ 277, 353, 29 },
		{ 277, 354, 29 },
		{ 277, 355, 29 },
		{ 277, 356, 29 },
		{ 277, 357, 29 },
		{ 277, 358, 29 },
		{ 277, 359, 29 },
		{ 277, 360, 29 },
		{ 277, 361, 29 },
		{ 277, 362, 29 },
		{ 277, 363, 29 },
		{ 277, 364, 29 },
		{ 277, 365, 29 },
		{ 277, 366, 29 },
		{ 277, 367, 29 },
		{ 277, 368, 29 },
		{ 277, 369, 29 },
		{ 277, 370, 29 },
		{ 277, 371, 29 },
		{ 277, 372, 29 },
		{ 277, 373, 29 },
		{ 277, 374, 29 },
		{ 277, 375, 29 },
		{ 277, 376, 29 },
		{ 277, 377, 29 },
		{ 277, 378, 29 },
		{ 277, 379, 29 },
		{ 277, 380, 29 },
		{ 277, 381, 29 },
		{ 277, 382, 29 },
		{ 277, 383, 29 },
		{ 278, 300, 29 },
		{ 278, 301, 29 },
		{ 278, 302, 29 },
		{ 278, 303, 29 },
		{ 278, 304, 29 },
		{ 278, 305, 29 },
		{ 278, 306, 29 },
		{ 278, 307, 29 },
		{ 278, 308, 29 },
		{ 278, 309, 29 },
		{ 278, 310, 29 },
		{ 278, 311, 29 },
		{ 278, 312, 29 },
		{ 278, 313, 29 },
		{ 278, 314, 29 },
		{ 278, 315, 29 },
		{ 278, 316, 29 },
		{ 278, 317, 29 },
		{ 278, 318, 29 },
		{ 278, 319, 29 },
		{ 278, 320, 29 },
		{ 278, 321, 29 },
		{ 278, 322, 29 },
		{ 278, 323, 29 },
		{ 278, 324, 29 },
		{ 278, 325, 29 },
		{ 278, 326, 29 },
		{ 278, 327, 29 },
		{ 278, 328, 29 },
		{ 278, 329, 29 },
		{ 278, 330, 29 },
		{ 278, 331, 29 },
		{ 278, 332, 29 },
		{ 278, 333, 29 },
		{ 278, 334, 29 },
		{ 278, 335, 29 },
		{ 278, 336, 29 },
		{ 278, 337, 29 },
		{ 278, 338, 29 },
		{ 278, 339, 29 },
		{ 278, 340, 29 },
		{ 278, 341, 29 },
		{ 278, 342, 29 },
		{ 278, 343, 29 },
		{ 278, 344, 29 },
		{ 278, 345, 29 },
		{ 278, 346, 29 },
		{ 278, 347, 29 },
		{ 278, 348, 29 },
		{ 278, 349, 29 },
		{ 278, 350, 29 },
		{ 278, 351, 29 },
		{ 278, 352, 29 },
		{ 278, 353, 29 },
		{ 278, 354, 29 },
		{ 278, 355, 29 },
		{ 278, 356, 29 },
		{ 278, 357, 29 },
		{ 278, 358, 29 },
		{ 278, 359, 29 },
		{ 278, 360, 29 },
		{ 278, 361, 29 },
		{ 278, 362, 29 },
		{ 278, 363, 29 },
		{ 278, 364, 29 },
		{ 278, 365, 29 },
		{ 278, 366, 29 },
		{ 278, 367, 29 },
		{ 278, 368, 29 },
		{ 278, 369, 29 },
		{ 278, 370, 29 },
		{ 278, 371, 29 },
		{ 278, 372, 29 },
		{ 278, 373, 29 },
		{ 278, 374, 29 },
		{ 278, 375, 29 },
		{ 278, 376, 29 },
		{ 278, 377, 29 },
		{ 278, 378, 29 },
		{ 278, 379, 29 },
		{ 278, 380, 29 },
		{ 278, 381, 29 },
		{ 278, 382, 29 },
		{ 278, 383, 29 },
		{ 279, 302, 29 },
		{ 279, 303, 29 },
		{ 279, 304, 29 },
		{ 279, 305, 29 },
		{ 279, 306, 29 },
		{ 279, 307, 29 },
		{ 279, 308, 29 },
		{ 279, 309, 29 },
		{ 279, 310, 29 },
		{ 279, 311, 29 },
		{ 279, 312, 29 },
		{ 279, 313, 29 },
		{ 279, 314, 29 },
		{ 279, 315, 29 },
		{ 279, 316, 29 },
		{ 279, 317, 29 },
		{ 279, 318, 29 },
		{ 279, 319, 29 },
		{ 279, 320, 29 },
		{ 279, 321, 29 },
		{ 279, 322, 29 },
		{ 279, 323, 29 },
		{ 279, 324, 29 },
		{ 279, 325, 29 },
		{ 279, 326, 29 },
		{ 279, 327, 29 },
		{ 279, 328, 29 },
		{ 279, 329, 29 },
		{ 279, 330, 29 },
		{ 279, 331, 29 },
		{ 279, 332, 29 },
		{ 279, 333, 29 },
		{ 279, 334, 29 },
		{ 279, 335, 29 },
		{ 279, 336, 29 },
		{ 279, 337, 29 },
		{ 279, 338, 29 },
		{ 279, 339, 29 },
		{ 279, 340, 29 },
		{ 279, 341, 29 },
		{ 279, 342, 29 },
		{ 279, 343, 29 },
		{ 279, 344, 29 },
		{ 279, 345, 29 },
		{ 279, 346, 29 },
		{ 279, 347, 29 },
		{ 279, 348, 29 },
		{ 279, 349, 29 },
		{ 279, 350, 29 },
		{ 279, 351, 29 },
		{ 279, 352, 29 },
		{ 279, 353, 29 },
		{ 279, 354, 29 },
		{ 279, 355, 29 },
		{ 279, 356, 29 },
		{ 279, 357, 29 },
		{ 279, 358, 29 },
		{ 279, 359, 29 },
		{ 279, 360, 29 },
		{ 279, 361, 29 },
		{ 279, 362, 29 },
		{ 279, 363, 29 },
		{ 279, 364, 29 },
		{ 279, 365, 29 },
		{ 279, 366, 29 },
		{ 279, 367, 29 },
		{ 279, 368, 29 },
		{ 279, 369, 29 },
		{ 279, 370, 29 },
		{ 279, 371, 29 },
		{ 279, 372, 29 },
		{ 279, 373, 29 },
		{ 279, 374, 29 },
		{ 279, 375, 29 },
		{ 279, 376, 29 },
		{ 279, 377, 29 },
		{ 279, 378, 29 },
		{ 279, 379, 29 },
		{ 279, 380, 29 },
		{ 279, 381, 29 },
		{ 279, 382, 29 },
		{ 279, 383, 29 },
		{ 280, 304, 29 },
		{ 280, 305, 29 },
		{ 280, 306, 29 },
		{ 280, 307, 29 },
		{ 280, 308, 29 },
		{ 280, 309, 29 },
		{ 280, 310, 29 },
		{ 280, 311, 29 },
		{ 280, 312, 29 },
		{ 280, 313, 29 },
		{ 280, 314, 29 },
		{ 280, 315, 29 },
		{ 280, 316, 29 },
		{ 280, 317, 29 },
		{ 280, 318, 29 },
		{ 280, 319, 29 },
		{ 280, 320, 29 },
		{ 280, 321, 29 },
		{ 280, 322, 29 },
		{ 280, 323, 29 },
		{ 280, 324, 29 },
		{ 280, 325, 29 },
		{ 280, 326, 29 },
		{ 280, 327, 29 },
		{ 280, 328, 29 },
		{ 280, 329, 29 },
		{ 280, 330, 29 },
		{ 280, 331, 29 },
		{ 280, 332, 29 },
		{ 280, 333, 29 },
		{ 280, 334, 29 },
		{ 280, 335, 29 },
		{ 280, 336, 29 },
		{ 280, 337, 29 },
		{ 280, 338, 29 },
		{ 280, 339, 29 },
		{ 280, 340, 29 },
		{ 280, 341, 29 },
		{ 280, 342, 29 },
		{ 280, 343, 29 },
		{ 280, 344, 29 },
		{ 280, 345, 29 },
		{ 280, 346, 29 },
		{ 280, 347, 29 },
		{ 280, 348, 29 },
		{ 280, 349, 29 },
		{ 280, 350, 29 },
		{ 280, 351, 29 },
		{ 280, 352, 29 },
		{ 280, 353, 29 },
		{ 280, 354, 29 },
		{ 280, 355, 29 },
		{ 280, 356, 29 },
		{ 280, 357, 29 },
		{ 280, 358, 29 },
		{ 280, 359, 29 },
		{ 280, 360, 29 },
		{ 280, 361, 29 },
		{ 280, 362, 29 },
		{ 280, 363, 29 },
		{ 280, 364, 29 },
		{ 280, 365, 29 },
		{ 280, 366, 29 },
		{ 280, 367, 29 },
		{ 280, 368, 29 },
		{ 280, 369, 29 },
		{ 280, 370, 29 },
		{ 280, 371, 29 },
		{ 280, 372, 29 },
		{ 280, 373, 29 },
		{ 280, 374, 29 },
		{ 280, 375, 29 },
		{ 280, 376, 29 },
		{ 280, 377, 29 },
		{ 280, 378, 29 },
		{ 280, 379, 29 },
		{ 280, 380, 29 },
		{ 280, 381, 29 },
		{ 280, 382, 29 },
		{ 280, 383, 29 },
		{ 281, 306, 29 },
		{ 281, 307, 29 },
		{ 281, 308, 29 },
		{ 281, 309, 29 },
		{ 281, 310, 29 },
		{ 281, 311, 29 },
		{ 281, 312, 29 },
		{ 281, 313, 29 },
		{ 281, 314, 29 },
		{ 281, 315, 29 },
		{ 281, 316, 29 },
		{ 281, 317, 29 },
		{ 281, 318, 29 },
		{ 281, 319, 29 },
		{ 281, 320, 29 },
		{ 281, 321, 29 },
		{ 281, 322, 29 },
		{ 281, 323, 29 },
		{ 281, 324, 29 },
		{ 281, 325, 29 },
		{ 281, 326, 29 },
		{ 281, 327, 29 },
		{ 281, 328, 29 },
		{ 281, 329, 29 },
		{ 281, 330, 29 },
		{ 281, 331, 29 },
		{ 281, 332, 29 },
		{ 281, 333, 29 },
		{ 281, 334, 29 },
		{ 281, 335, 29 },
		{ 281, 336, 29 },
		{ 281, 337, 29 },
		{ 281, 338, 29 },
		{ 281, 339, 29 },
		{ 281, 340, 29 },
		{ 281, 341, 29 },
		{ 281, 342, 29 },
		{ 281, 343, 29 },
		{ 281, 344, 29 },
		{ 281, 345, 29 },
		{ 281, 346, 29 },
		{ 281, 347, 29 },
		{ 281, 348, 29 },
		{ 281, 349, 29 },
		{ 281, 350, 29 },
		{ 281, 351, 29 },
		{ 281, 352, 29 },
		{ 281, 353, 29 },
		{ 281, 354, 29 },
		{ 281, 355, 29 },
		{ 281, 356, 29 },
		{ 281, 357, 29 },
		{ 281, 358, 29 },
		{ 281, 359, 29 },
		{ 281, 360, 29 },
		{ 281, 361, 29 },
		{ 281, 362, 29 },
		{ 281, 363, 29 },
		{ 281, 364, 29 },
		{ 281, 365, 29 },
		{ 281, 366, 29 },
		{ 281, 367, 29 },
		{ 281, 368, 29 },
		{ 281, 369, 29 },
		{ 281, 370, 29 },
		{ 281, 371, 29 },
		{ 281, 372, 29 },
		{ 281, 373, 29 },
		{ 281, 374, 29 },
		{ 281, 375, 29 },
		{ 281, 376, 29 },
		{ 281, 377, 29 },
		{ 281, 378, 29 },
		{ 281, 379, 29 },
		{ 281, 380, 29 },
		{ 281, 381, 29 },
		{ 281, 382, 29 },
		{ 281, 383, 29 },
		{ 282, 308, 29 },
		{ 282, 309, 29 },
		{ 282, 310, 29 },
		{ 282, 311, 29 },
		{ 282, 312, 29 },
		{ 282, 313, 29 },
		{ 282, 314, 29 },
		{ 282, 315, 29 },
		{ 282, 316, 29 },
		{ 282, 317, 29 },
		{ 282, 318, 29 },
		{ 282, 319, 29 },
		{ 282, 320, 29 },
		{ 282, 321, 29 },
		{ 282, 322, 29 },
		{ 282, 323, 29 },
		{ 282, 324, 29 },
		{ 282, 325, 29 },
		{ 282, 326, 29 },
		{ 282, 327, 29 },
		{ 282, 328, 29 },
		{ 282, 329, 29 },
		{ 282, 330, 29 },
		{ 282, 331, 29 },
		{ 282, 332, 29 },
		{ 282, 333, 29 },
		{ 282, 334, 29 },
		{ 282, 335, 29 },
		{ 282, 336, 29 },
		{ 282, 337, 29 },
		{ 282, 338, 29 },
		{ 282, 339, 29 },
		{ 282, 340, 29 },
		{ 282, 341, 29 },
		{ 282, 342, 29 },
		{ 282, 343, 29 },
		{ 282, 344, 29 },
		{ 282, 345, 29 },
		{ 282, 346, 29 },
		{ 282, 347, 29 },
		{ 282, 348, 29 },
		{ 282, 349, 29 },
		{ 282, 350, 29 },
		{ 282, 351, 29 },
		{ 282, 352, 29 },
		{ 282, 353, 29 },
		{ 282, 354, 29 },
		{ 282, 355, 29 },
		{ 282, 356, 29 },
		{ 282, 357, 29 },
		{ 282, 358, 29 },
		{ 282, 359, 29 },
		{ 282, 360, 29 },
		{ 282, 361, 29 },
		{ 282, 362, 29 },
		{ 282, 363, 29 },
		{ 282, 364, 29 },
		{ 282, 365, 29 },
		{ 282, 366, 29 },
		{ 282, 367, 29 },
		{ 282, 368, 29 },
		{ 282, 369, 29 },
		{ 282, 370, 29 },
		{ 282, 371, 29 },
		{ 282, 372, 29 },
		{ 282, 373, 29 },
		{ 282, 374, 29 },
		{ 282, 375, 29 },
		{ 282, 376, 29 },
		{ 282, 377, 29 },
		{ 282, 378, 29 },
		{ 282, 379, 29 },
		{ 282, 380, 29 },
		{ 282, 381, 29 },
		{ 282, 382, 29 },
		{ 282, 383, 29 },
		{ 283, 310, 29 },
		{ 283, 311, 29 },
		{ 283, 312, 29 },
		{ 283, 313, 29 },
		{ 283, 314, 29 },
		{ 283, 315, 29 },
		{ 283, 316, 29 },
		{ 283, 317, 29 },
		{ 283, 318, 29 },
		{ 283, 319, 29 },
		{ 283, 320, 29 },
		{ 283, 321, 29 },
		{ 283, 322, 29 },
		{ 283, 323, 29 },
		{ 283, 324, 29 },
		{ 283, 325, 29 },
		{ 283, 326, 29 },
		{ 283, 327, 29 },
		{ 283, 328, 29 },
		{ 283, 329, 29 },
		{ 283, 330, 29 },
		{ 283, 331, 29 },
		{ 283, 332, 29 },
		{ 283, 333, 29 },
		{ 283, 334, 29 },
		{ 283, 335, 29 },
		{ 283, 336, 29 },
		{ 283, 337, 29 },
		{ 283, 338, 29 },
		{ 283, 339, 29 },
		{ 283, 340, 29 },
		{ 283, 341, 29 },
		{ 283, 342, 29 },
		{ 283, 343, 29 },
		{ 283, 344, 29 },
		{ 283, 345, 29 },
		{ 283, 346, 29 },
		{ 283, 347, 29 },
		{ 283, 348, 29 },
		{ 283, 349, 29 },
		{ 283, 350, 29 },
		{ 283, 351, 29 },
		{ 283, 352, 29 },
		{ 283, 353, 29 },
		{ 283, 354, 29 },
		{ 283, 355, 29 },
		{ 283, 356, 29 },
		{ 283, 357, 29 },
		{ 283, 358, 29 },
		{ 283, 359, 29 },
		{ 283, 360, 29 },
		{ 283, 361, 29 },
		{ 283, 362, 29 },
		{ 283, 363, 29 },
		{ 283, 364, 29 },
		{ 283, 365, 29 },
		{ 283, 366, 29 },
		{ 283, 367, 29 },
		{ 283, 368, 29 },
		{ 283, 369, 29 },
		{ 283, 370, 29 },
		{ 283, 371, 29 },
		{ 283, 372, 29 },
		{ 283, 373, 29 },
		{ 283, 374, 29 },
		{ 283, 375, 29 },
		{ 283, 376, 29 },
		{ 283, 377, 29 },
		{ 283, 378, 29 },
		{ 283, 379, 29 },
		{ 283, 380, 29 },
		{ 283, 381, 29 },
		{ 283, 382, 29 },
		{ 283, 383, 29 },
		{ 284, 312, 29 },
		{ 284, 313, 29 },
		{ 284, 314, 29 },
		{ 284, 315, 29 },
		{ 284, 316, 29 },
		{ 284, 317, 29 },
		{ 284, 318, 29 },
		{ 284, 319, 29 },
		{ 284, 320, 29 },
		{ 284, 321, 29 },
		{ 284, 322, 29 },
		{ 284, 323, 29 },
		{ 284, 324, 29 },
		{ 284, 325, 29 },
		{ 284, 326, 29 },
		{ 284, 327, 29 },
		{ 284, 328, 29 },
		{ 284, 329, 29 },
		{ 284, 330, 29 },
		{ 284, 331, 29 },
		{ 284, 332, 29 },
		{ 284, 333, 29 },
		{ 284, 334, 29 },
		{ 284, 335, 29 },
		{ 284, 336, 29 },
		{ 284, 337, 29 },
		{ 284, 338, 29 },
		{ 284, 339, 29 },
		{ 284, 340, 29 },
		{ 284, 341, 29 },
		{ 284, 342, 29 },
		{ 284, 343, 29 },
		{ 284, 344, 29 },
		{ 284, 345, 29 },
		{ 284, 346, 29 },
		{ 284, 347, 29 },
		{ 284, 348, 29 },
		{ 284, 349, 29 },
		{ 284, 350, 29 },
		{ 284, 351, 29 },
		{ 284, 352, 29 },
		{ 284, 353, 29 },
		{ 284, 354, 29 },
		{ 284, 355, 29 },
		{ 284, 356, 29 },
		{ 284, 357, 29 },
		{ 284, 358, 29 },
		{ 284, 359, 29 },
		{ 284, 360, 29 },
		{ 284, 361, 29 },
		{ 284, 362, 29 },
		{ 284, 363, 29 },
		{ 284, 364, 29 },
		{ 284, 365, 29 },
		{ 284, 366, 29 },
		{ 284, 367, 29 },
		{ 284, 368, 29 },
		{ 284, 369, 29 },
		{ 284, 370, 29 },
		{ 284, 371, 29 },
		{ 284, 372, 29 },
		{ 284, 373, 29 },
		{ 284, 374, 29 },
		{ 284, 375, 29 },
		{ 284, 376, 29 },
		{ 284, 377, 29 },
		{ 284, 378, 29 },
		{ 284, 379, 29 },
		{ 284, 380, 29 },
		{ 284, 381, 29 },
		{ 284, 382, 29 },
		{ 284, 383, 29 },
		{ 285, 314, 29 },
		{ 285, 315, 29 },
		{ 285, 316, 29 },
		{ 285, 317, 29 },
		{ 285, 318, 29 },
		{ 285, 319, 29 },
		{ 285, 320, 29 },
		{ 285, 321, 29 },
		{ 285, 322, 29 },
		{ 285, 323, 29 },
		{ 285, 324, 29 },
		{ 285, 325, 29 },
		{ 285, 326, 29 },
		{ 285, 327, 29 },
		{ 285, 328, 29 },
		{ 285, 329, 29 },
		{ 285, 330, 29 },
		{ 285, 331, 29 },
		{ 285, 332, 29 },
		{ 285, 333, 29 },
		{ 285, 334, 29 },
		{ 285, 335, 29 },
		{ 285, 336, 29 },
		{ 285, 337, 29 },
		{ 285, 338, 29 },
		{ 285, 339, 29 },
		{ 285, 340, 29 },
		{ 285, 341, 29 },
		{ 285, 342, 29 },
		{ 285, 343, 29 },
		{ 285, 344, 29 },
		{ 285, 345, 29 },
		{ 285, 346, 29 },
		{ 285, 347, 29 },
		{ 285, 348, 29 },
		{ 285, 349, 29 },
		{ 285, 350, 29 },
		{ 285, 351, 29 },
		{ 285, 352, 29 },
		{ 285, 353, 29 },
		{ 285, 354, 29 },
		{ 285, 355, 29 },
		{ 285, 356, 29 },
		{ 285, 357, 29 },
		{ 285, 358, 29 },
		{ 285, 359, 29 },
		{ 285, 360, 29 },
		{ 285, 361, 29 },
		{ 285, 362, 29 },
		{ 285, 363, 29 },
		{ 285, 364, 29 },
		{ 285, 365, 29 },
		{ 285, 366, 29 },
		{ 285, 367, 29 },
		{ 285, 368, 29 },
		{ 285, 369, 29 },
		{ 285, 370, 29 },
		{ 285, 371, 29 },
		{ 285, 372, 29 },
		{ 285, 373, 29 },
		{ 285, 374, 29 },
		{ 285, 375, 29 },
		{ 285, 376, 29 },
		{ 285, 377, 29 },
		{ 285, 378, 29 },
		{ 285, 379, 29 },
		{ 285, 380, 29 },
		{ 285, 381, 29 },
		{ 285, 382, 29 },
		{ 285, 383, 29 },
		{ 286, 316, 29 },
		{ 286, 317, 29 },
		{ 286, 318, 29 },
		{ 286, 319, 29 },
		{ 286, 320, 29 },
		{ 286, 321, 29 },
		{ 286, 322, 29 },
		{ 286, 323, 29 },
		{ 286, 324, 29 },
		{ 286, 325, 29 },
		{ 286, 326, 29 },
		{ 286, 327, 29 },
		{ 286, 328, 29 },
		{ 286, 329, 29 },
		{ 286, 330, 29 },
		{ 286, 331, 29 },
		{ 286, 332, 29 },
		{ 286, 333, 29 },
		{ 286, 334, 29 },
		{ 286, 335, 29 },
		{ 286, 336, 29 },
		{ 286, 337, 29 },
		{ 286, 338, 29 },
		{ 286, 339, 29 },
		{ 286, 340, 29 },
		{ 286, 341, 29 },
		{ 286, 342, 29 },
		{ 286, 343, 29 },
		{ 286, 344, 29 },
		{ 286, 345, 29 },
		{ 286, 346, 29 },
		{ 286, 347, 29 },
		{ 286, 348, 29 },
		{ 286, 349, 29 },
		{ 286, 350, 29 },
		{ 286, 351, 29 },
		{ 286, 352, 29 },
		{ 286, 353, 29 },
		{ 286, 354, 29 },
		{ 286, 355, 29 },
		{ 286, 356, 29 },
		{ 286, 357, 29 },
		{ 286, 358, 29 },
		{ 286, 359, 29 },
		{ 286, 360, 29 },
		{ 286, 361, 29 },
		{ 286, 362, 29 },
		{ 286, 363, 29 },
		{ 286, 364, 29 },
		{ 286, 365, 29 },
		{ 286, 366, 29 },
		{ 286, 367, 29 },
		{ 286, 368, 29 },
		{ 286, 369, 29 },
		{ 286, 370, 29 },
		{ 286, 371, 29 },
		{ 286, 372, 29 },
		{ 286, 373, 29 },
		{ 286, 374, 29 },
		{ 286, 375, 29 },
		{ 286, 376, 29 },
		{ 286, 377, 29 },
		{ 286, 378, 29 },
		{ 286, 379, 29 },
		{ 286, 380, 29 },
		{ 286, 381, 29 },
		{ 286, 382, 29 },
		{ 286, 383, 29 },
		{ 287, 318, 29 },
		{ 287, 319, 29 },
		{ 287, 320, 29 },
		{ 287, 321, 29 },
		{ 287, 322, 29 },
		{ 287, 323, 29 },
		{ 287, 324, 29 },
		{ 287, 325, 29 },
		{ 287, 326, 29 },
		{ 287, 327, 29 },
		{ 287, 328, 29 },
		{ 287, 329, 29 },
		{ 287, 330, 29 },
		{ 287, 331, 29 },
		{ 287, 332, 29 },
		{ 287, 333, 29 },
		{ 287, 334, 29 },
		{ 287, 335, 29 },
		{ 287, 336, 29 },
		{ 287, 337, 29 },
		{ 287, 338, 29 },
		{ 287, 339, 29 },
		{ 287, 340, 29 },
		{ 287, 341, 29 },
		{ 287, 342, 29 },
		{ 287, 343, 29 },
		{ 287, 344, 29 },
		{ 287, 345, 29 },
		{ 287, 346, 29 },
		{ 287, 347, 29 },
		{ 287, 348, 29 },
		{ 287, 349, 29 },
		{ 287, 350, 29 },
		{ 287, 351, 29 },
		{ 287, 352, 29 },
		{ 287, 353, 29 },
		{ 287, 354, 29 },
		{ 287, 355, 29 },
		{ 287, 356, 29 },
		{ 287, 357, 29 },
		{ 287, 358, 29 },
		{ 287, 359, 29 },
		{ 287, 360, 29 },
		{ 287, 361, 29 },
		{ 287, 362, 29 },
		{ 287, 363, 29 },
		{ 287, 364, 29 },
		{ 287, 365, 29 },
		{ 287, 366, 29 },
		{ 287, 367, 29 },
		{ 287, 368, 29 },
		{ 287, 369, 29 },
		{ 287, 370, 29 },
		{ 287, 371, 29 },
		{ 287, 372, 29 },
		{ 287, 373, 29 },
		{ 287, 374, 29 },
		{ 287, 375, 29 },
		{ 287, 376, 29 },
		{ 287, 377, 29 },
		{ 287, 378, 29 },
		{ 287, 379, 29 },
		{ 287, 380, 29 },
		{ 287, 381, 29 },
		{ 287, 382, 29 },
		{ 287, 383, 29 },
		{ 288, 320, 29 },
		{ 288, 321, 29 },
		{ 288, 322, 29 },
		{ 288, 323, 29 },
		{ 288, 324, 29 },
		{ 288, 325, 29 },
		{ 288, 326, 29 },
		{ 288, 327, 29 },
		{ 288, 328, 29 },
		{ 288, 329, 29 },
		{ 288, 330, 29 },
		{ 288, 331, 29 },
		{ 288, 332, 29 },
		{ 288, 333, 29 },
		{ 288, 334, 29 },
		{ 288, 335, 29 },
		{ 288, 336, 29 },
		{ 288, 337, 29 },
		{ 288, 338, 29 },
		{ 288, 339, 29 },
		{ 288, 340, 29 },
		{ 288, 341, 29 },
		{ 288, 342, 29 },
		{ 288, 343, 29 },
		{ 288, 344, 29 },
		{ 288, 345, 29 },
		{ 288, 346, 29 },
		{ 288, 347, 29 },
		{ 288, 348, 29 },
		{ 288, 349, 29 },
		{ 288, 350, 29 },
		{ 288, 351, 29 },
		{ 288, 352, 29 },
		{ 288, 353, 29 },
		{ 288, 354, 29 },
		{ 288, 355, 29 },
		{ 288, 356, 29 },
		{ 288, 357, 29 },
		{ 288, 358, 29 },
		{ 288, 359, 29 },
		{ 288, 360, 29 },
		{ 288, 361, 29 },
		{ 288, 362, 29 },
		{ 288, 363, 29 },
		{ 288, 364, 29 },
		{ 288, 365, 29 },
		{ 288, 366, 29 },
		{ 288, 367, 29 },
		{ 288, 368, 29 },
		{ 288, 369, 29 },
		{ 288, 370, 29 },
		{ 288, 371, 29 },
		{ 288, 372, 29 },
		{ 288, 373, 29 },
		{ 288, 374, 29 },
		{ 288, 375, 29 },
		{ 288, 376, 29 },
		{ 288, 377, 29 },
		{ 288, 378, 29 },
		{ 288, 379, 29 },
		{ 288, 380, 29 },
		{ 288, 381, 29 },
		{ 288, 382, 29 },
		{ 288, 383, 29 },
		{ 289, 322, 29 },
		{ 289, 323, 29 },
		{ 289, 324, 29 },
		{ 289, 325, 29 },
		{ 289, 326, 29 },
		{ 289, 327, 29 },
		{ 289, 328, 29 },
		{ 289, 329, 29 },
		{ 289, 330, 29 },
		{ 289, 331, 29 },
		{ 289, 332, 29 },
		{ 289, 333, 29 },
		{ 289, 334, 29 },
		{ 289, 335, 29 },
		{ 289, 336, 29 },
		{ 289, 337, 29 },
		{ 289, 338, 29 },
		{ 289, 339, 29 },
		{ 289, 340, 29 },
		{ 289, 341, 29 },
		{ 289, 342, 29 },
		{ 289, 343, 29 },
		{ 289, 344, 29 },
		{ 289, 345, 29 },
		{ 289, 346, 29 },
		{ 289, 347, 29 },
		{ 289, 348, 29 },
		{ 289, 349, 29 },
		{ 289, 350, 29 },
		{ 289, 351, 29 },
		{ 289, 352, 29 },
		{ 289, 353, 29 },
		{ 289, 354, 29 },
		{ 289, 355, 29 },
		{ 289, 356, 29 },
		{ 289, 357, 29 },
		{ 289, 358, 29 },
		{ 289, 359, 29 },
		{ 289, 360, 29 },
		{ 289, 361, 29 },
		{ 289, 362, 29 },
		{ 289, 363, 29 },
		{ 289, 364, 29 },
		{ 289, 365, 29 },
		{ 289, 366, 29 },
		{ 289, 367, 29 },
		{ 289, 368, 29 },
		{ 289, 369, 29 },
		{ 289, 370, 29 },
		{ 289, 371, 29 },
		{ 289, 372, 29 },
		{ 289, 373, 29 },
		{ 289, 374, 29 },
		{ 289, 375, 29 },
		{ 289, 376, 29 },
		{ 289, 377, 29 },
		{ 289, 378, 29 },
		{ 289, 379, 29 },
		{ 289, 380, 29 },
		{ 289, 381, 29 },
		{ 289, 382, 29 },
		{ 289, 383, 29 },
		{ 290, 324, 29 },
		{ 290, 325, 29 },
		{ 290, 326, 29 },
		{ 290, 327, 29 },
		{ 290, 328, 29 },
		{ 290, 329, 29 },
		{ 290, 330, 29 },
		{ 290, 331, 29 },
		{ 290, 332, 29 },
		{ 290, 333, 29 },
		{ 290, 334, 29 },
		{ 290, 335, 29 },
		{ 290, 336, 29 },
		{ 290, 337, 29 },
		{ 290, 338, 29 },
		{ 290, 339, 29 },
		{ 290, 340, 29 },
		{ 290, 341, 29 },
		{ 290, 342, 29 },
		{ 290, 343, 29 },
		{ 290, 344, 29 },
		{ 290, 345, 29 },
		{ 290, 346, 29 },
		{ 290, 347, 29 },
		{ 290, 348, 29 },
		{ 290, 349, 29 },
		{ 290, 350, 29 },
		{ 290, 351, 29 },
		{ 290, 352, 29 },
		{ 290, 353, 29 },
		{ 290, 354, 29 },
		{ 290, 355, 29 },
		{ 290, 356, 29 },
		{ 290, 357, 29 },
		{ 290, 358, 29 },
		{ 290, 359, 29 },
		{ 290, 360, 29 },
		{ 290, 361, 29 },
		{ 290, 362, 29 },
		{ 290, 363, 29 },
		{ 290, 364, 29 },
		{ 290, 365, 29 },
		{ 290, 366, 29 },
		{ 290, 367, 29 },
		{ 290, 368, 29 },
		{ 290, 369, 29 },
		{ 290, 370, 29 },
		{ 290, 371, 29 },
		{ 290, 372, 29 },
		{ 290, 373, 29 },
		{ 290, 374, 29 },
		{ 290, 375, 29 },
		{ 290, 376, 29 },
		{ 290, 377, 29 },
		{ 290, 378, 29 },
		{ 290, 379, 29 },
		{ 290, 380, 29 },
		{ 290, 381, 29 },
		{ 290, 382, 29 },
		{ 290, 383, 29 },
		{ 291, 326, 29 },
		{ 291, 327, 29 },
		{ 291, 328, 29 },
		{ 291, 329, 29 },
		{ 291, 330, 29 },
		{ 291, 331, 29 },
		{ 291, 332, 29 },
		{ 291, 333, 29 },
		{ 291, 334, 29 },
		{ 291, 335, 29 },
		{ 291, 336, 29 },
		{ 291, 337, 29 },
		{ 291, 338, 29 },
		{ 291, 339, 29 },
		{ 291, 340, 29 },
		{ 291, 341, 29 },
		{ 291, 342, 29 },
		{ 291, 343, 29 },
		{ 291, 344, 29 },
		{ 291, 345, 29 },
		{ 291, 346, 29 },
		{ 291, 347, 29 },
		{ 291, 348, 29 },
		{ 291, 349, 29 },
		{ 291, 350, 29 },
		{ 291, 351, 29 },
		{ 291, 352, 29 },
		{ 291, 353, 29 },
		{ 291, 354, 29 },
		{ 291, 355, 29 },
		{ 291, 356, 29 },
		{ 291, 357, 29 },
		{ 291, 358, 29 },
		{ 291, 359, 29 },
		{ 291, 360, 29 },
		{ 291, 361, 29 },
		{ 291, 362, 29 },
		{ 291, 363, 29 },
		{ 291, 364, 29 },
		{ 291, 365, 29 },
		{ 291, 366, 29 },
		{ 291, 367, 29 },
		{ 291, 368, 29 },
		{ 291, 369, 29 },
		{ 291, 370, 29 },
		{ 291, 371, 29 },
		{ 291, 372, 29 },
		{ 291, 373, 29 },
		{ 291, 374, 29 },
		{ 291, 375, 29 },
		{ 291, 376, 29 },
		{ 291, 377, 29 },
		{ 291, 378, 29 },
		{ 291, 379, 29 },
		{ 291, 380, 29 },
		{ 291, 381, 29 },
		{ 291, 382, 29 },
		{ 291, 383, 29 },
		{ 292, 328, 29 },
		{ 292, 329, 29 },
		{ 292, 330, 29 },
		{ 292, 331, 29 },
		{ 292, 332, 29 },
		{ 292, 333, 29 },
		{ 292, 334, 29 },
		{ 292, 335, 29 },
		{ 292, 336, 29 },
		{ 292, 337, 29 },
		{ 292, 338, 29 },
		{ 292, 339, 29 },
		{ 292, 340, 29 },
		{ 292, 341, 29 },
		{ 292, 342, 29 },
		{ 292, 343, 29 },
		{ 292, 344, 29 },
		{ 292, 345, 29 },
		{ 292, 346, 29 },
		{ 292, 347, 29 },
		{ 292, 348, 29 },
		{ 292, 349, 29 },
		{ 292, 350, 29 },
		{ 292, 351, 29 },
		{ 292, 352, 29 },
		{ 292, 353, 29 },
		{ 292, 354, 29 },
		{ 292, 355, 29 },
		{ 292, 356, 29 },
		{ 292, 357, 29 },
		{ 292, 358, 29 },
		{ 292, 359, 29 },
		{ 292, 360, 29 },
		{ 292, 361, 29 },
		{ 292, 362, 29 },
		{ 292, 363, 29 },
		{ 292, 364, 29 },
		{ 292, 365, 29 },
		{ 292, 366, 29 },
		{ 292, 367, 29 },
		{ 292, 368, 29 },
		{ 292, 369, 29 },
		{ 292, 370, 29 },
		{ 292, 371, 29 },
		{ 292, 372, 29 },
		{ 292, 373, 29 },
		{ 292, 374, 29 },
		{ 292, 375, 29 },
		{ 292, 376, 29 },
		{ 292, 377, 29 },
		{ 292, 378, 29 },
		{ 292, 379, 29 },
		{ 292, 380, 29 },
		{ 292, 381, 29 },
		{ 292, 382, 29 },
		{ 292, 383, 29 },
		{ 293, 330, 29 },
		{ 293, 331, 29 },
		{ 293, 332, 29 },
		{ 293, 333, 29 },
		{ 293, 334, 29 },
		{ 293, 335, 29 },
		{ 293, 336, 29 },
		{ 293, 337, 29 },
		{ 293, 338, 29 },
		{ 293, 339, 29 },
		{ 293, 340, 29 },
		{ 293, 341, 29 },
		{ 293, 342, 29 },
		{ 293, 343, 29 },
		{ 293, 344, 29 },
		{ 293, 345, 29 },
		{ 293, 346, 29 },
		{ 293, 347, 29 },
		{ 293, 348, 29 },
		{ 293, 349, 29 },
		{ 293, 350, 29 },
		{ 293, 351, 29 },
		{ 293, 352, 29 },
		{ 293, 353, 29 },
		{ 293, 354, 29 },
		{ 293, 355, 29 },
		{ 293, 356, 29 },
		{ 293, 357, 29 },
		{ 293, 358, 29 },
		{ 293, 359, 29 },
		{ 293, 360, 29 },
		{ 293, 361, 29 },
		{ 293, 362, 29 },
		{ 293, 363, 29 },
		{ 293, 364, 29 },
		{ 293, 365, 29 },
		{ 293, 366, 29 },
		{ 293, 367, 29 },
		{ 293, 368, 29 },
		{ 293, 369, 29 },
		{ 293, 370, 29 },
		{ 293, 371, 29 },
		{ 293, 372, 29 },
		{ 293, 373, 29 },
		{ 293, 374, 29 },
		{ 293, 375, 29 },
		{ 293, 376, 29 },
		{ 293, 377, 29 },
		{ 293, 378, 29 },
		{ 293, 379, 29 },
		{ 293, 380, 29 },
		{ 293, 381, 29 },
		{ 293, 382, 29 },
		{ 293, 383, 29 },
		{ 294, 332, 29 },
		{ 294, 333, 29 },
		{ 294, 334, 29 },
		{ 294, 335, 29 },
		{ 294, 336, 29 },
		{ 294, 337, 29 },
		{ 294, 338, 29 },
		{ 294, 339, 29 },
		{ 294, 340, 29 },
		{ 294, 341, 29 },
		{ 294, 342, 29 },
		{ 294, 343, 29 },
		{ 294, 344, 29 },
		{ 294, 345, 29 },
		{ 294, 346, 29 },
		{ 294, 347, 29 },
		{ 294, 348, 29 },
		{ 294, 349, 29 },
		{ 294, 350, 29 },
		{ 294, 351, 29 },
		{ 294, 352, 29 },
		{ 294, 353, 29 },
		{ 294, 354, 29 },
		{ 294, 355, 29 },
		{ 294, 356, 29 },
		{ 294, 357, 29 },
		{ 294, 358, 29 },
		{ 294, 359, 29 },
		{ 294, 360, 29 },
		{ 294, 361, 29 },
		{ 294, 362, 29 },
		{ 294, 363, 29 },
		{ 294, 364, 29 },
		{ 294, 365, 29 },
		{ 294, 366, 29 },
		{ 294, 367, 29 },
		{ 294, 368, 29 },
		{ 294, 369, 29 },
		{ 294, 370, 29 },
		{ 294, 371, 29 },
		{ 294, 372, 29 },
		{ 294, 373, 29 },
		{ 294, 374, 29 },
		{ 294, 375, 29 },
		{ 294, 376, 29 },
		{ 294, 377, 29 },
		{ 294, 378, 29 },
		{ 294, 379, 29 },
		{ 294, 380, 29 },
		{ 294, 381, 29 },
		{ 294, 382, 29 },
		{ 294, 383, 29 },
		{ 295, 334, 29 },
		{ 295, 335, 29 },
		{ 295, 336, 29 },
		{ 295, 337, 29 },
		{ 295, 338, 29 },
		{ 295, 339, 29 },
		{ 295, 340, 29 },
		{ 295, 341, 29 },
		{ 295, 342, 29 },
		{ 295, 343, 29 },
		{ 295, 344, 29 },
		{ 295, 345, 29 },
		{ 295, 346, 29 },
		{ 295, 347, 29 },
		{ 295, 348, 29 },
		{ 295, 349, 29 },
		{ 295, 350, 29 },
		{ 295, 351, 29 },
		{ 295, 352, 29 },
		{ 295, 353, 29 },
		{ 295, 354, 29 },
		{ 295, 355, 29 },
		{ 295, 356, 29 },
		{ 295, 357, 29 },
		{ 295, 358, 29 },
		{ 295, 359, 29 },
		{ 295, 360, 29 },
		{ 295, 361, 29 },
		{ 295, 362, 29 },
		{ 295, 363, 29 },
		{ 295, 364, 29 },
		{ 295, 365, 29 },
		{ 295, 366, 29 },
		{ 295, 367, 29 },
		{ 295, 368, 29 },
		{ 295, 369, 29 },
		{ 295, 370, 29 },
		{ 295, 371, 29 },
		{ 295, 372, 29 },
		{ 295, 373, 29 },
		{ 295, 374, 29 },
		{ 295, 375, 29 },
		{ 295, 376, 29 },
		{ 295, 377, 29 },
		{ 295, 378, 29 },
		{ 295, 379, 29 },
		{ 295, 380, 29 },
		{ 295, 381, 29 },
		{ 295, 382, 29 },
		{ 295, 383, 29 },
		{ 296, 336, 29 },
		{ 296, 337, 29 },
		{ 296, 338, 29 },
		{ 296, 339, 29 },
		{ 296, 340, 29 },
		{ 296, 341, 29 },
		{ 296, 342, 29 },
		{ 296, 343, 29 },
		{ 296, 344, 29 },
		{ 296, 345, 29 },
		{ 296, 346, 29 },
		{ 296, 347, 29 },
		{ 296, 348, 29 },
		{ 296, 349, 29 },
		{ 296, 350, 29 },
		{ 296, 351, 29 },
		{ 296, 352, 29 },
		{ 296, 353, 29 },
		{ 296, 354, 29 },
		{ 296, 355, 29 },
		{ 296, 356, 29 },
		{ 296, 357, 29 },
		{ 296, 358, 29 },
		{ 296, 359, 29 },
		{ 296, 360, 29 },
		{ 296, 361, 29 },
		{ 296, 362, 29 },
		{ 296, 363, 29 },
		{ 296, 364, 29 },
		{ 296, 365, 29 },
		{ 296, 366, 29 },
		{ 296, 367, 29 },
		{ 296, 368, 29 },
		{ 296, 369, 29 },
		{ 296, 370, 29 },
		{ 296, 371, 29 },
		{ 296, 372, 29 },
		{ 296, 373, 29 },
		{ 296, 374, 29 },
		{ 296, 375, 29 },
		{ 296, 376, 29 },
		{ 296, 377, 29 },
		{ 296, 378, 29 },
		{ 296, 379, 29 },
		{ 296, 380, 29 },
		{ 296, 381, 29 },
		{ 296, 382, 29 },
		{ 296, 383, 29 },
		{ 297, 338, 29 },
		{ 297, 339, 29 },
		{ 297, 340, 29 },
		{ 297, 341, 29 },
		{ 297, 342, 29 },
		{ 297, 343, 29 },
		{ 297, 344, 29 },
		{ 297, 345, 29 },
		{ 297, 346, 29 },
		{ 297, 347, 29 },
		{ 297, 348, 29 },
		{ 297, 349, 29 },
		{ 297, 350, 29 },
		{ 297, 351, 29 },
		{ 297, 352, 29 },
		{ 297, 353, 29 },
		{ 297, 354, 29 },
		{ 297, 355, 29 },
		{ 297, 356, 29 },
		{ 297, 357, 29 },
		{ 297, 358, 29 },
		{ 297, 359, 29 },
		{ 297, 360, 29 },
		{ 297, 361, 29 },
		{ 297, 362, 29 },
		{ 297, 363, 29 },
		{ 297, 364, 29 },
		{ 297, 365, 29 },
		{ 297, 366, 29 },
		{ 297, 367, 29 },
		{ 297, 368, 29 },
		{ 297, 369, 29 },
		{ 297, 370, 29 },
		{ 297, 371, 29 },
		{ 297, 372, 29 },
		{ 297, 373, 29 },
		{ 297, 374, 29 },
		{ 297, 375, 29 },
		{ 297, 376, 29 },
		{ 297, 377, 29 },
		{ 297, 378, 29 },
		{ 297, 379, 29 },
		{ 297, 380, 29 },
		{ 297, 381, 29 },
		{ 297, 382, 29 },
		{ 297, 383, 29 },
		{ 298, 340, 29 },
		{ 298, 341, 29 },
		{ 298, 342, 29 },
		{ 298, 343, 29 },
		{ 298, 344, 29 },
		{ 298, 345, 29 },
		{ 298, 346, 29 },
		{ 298, 347, 29 },
		{ 298, 348, 29 },
		{ 298, 349, 29 },
		{ 298, 350, 29 },
		{ 298, 351, 29 },
		{ 298, 352, 29 },
		{ 298, 353, 29 },
		{ 298, 354, 29 },
		{ 298, 355, 29 },
		{ 298, 356, 29 },
		{ 298, 357, 29 },
		{ 298, 358, 29 },
		{ 298, 359, 29 },
		{ 298, 360, 29 },
		{ 298, 361, 29 },
		{ 298, 362, 29 },
		{ 298, 363, 29 },
		{ 298, 364, 29 },
		{ 298, 365, 29 },
		{ 298, 366, 29 },
		{ 298, 367, 29 },
		{ 298, 368, 29 },
		{ 298, 369, 29 },
		{ 298, 370, 29 },
		{ 298, 371, 29 },
		{ 298, 372, 29 },
		{ 298, 373, 29 },
		{ 298, 374, 29 },
		{ 298, 375, 29 },
		{ 298, 376, 29 },
		{ 298, 377, 29 },
		{ 298, 378, 29 },
		{ 298, 379, 29 },
		{ 298, 380, 29 },
		{ 298, 381, 29 },
		{ 298, 382, 29 },
		{ 298, 383, 29 },
		{ 299, 342, 29 },
		{ 299, 343, 29 },
		{ 299, 344, 29 },
		{ 299, 345, 29 },
		{ 299, 346, 29 },
		{ 299, 347, 29 },
		{ 299, 348, 29 },
		{ 299, 349, 29 },
		{ 299, 350, 29 },
		{ 299, 351, 29 },
		{ 299, 352, 29 },
		{ 299, 353, 29 },
		{ 299, 354, 29 },
		{ 299, 355, 29 },
		{ 299, 356, 29 },
		{ 299, 357, 29 },
		{ 299, 358, 29 },
		{ 299, 359, 29 },
		{ 299, 360, 29 },
		{ 299, 361, 29 },
		{ 299, 362, 29 },
		{ 299, 363, 29 },
		{ 299, 364, 29 },
		{ 299, 365, 29 },
		{ 299, 366, 29 },
		{ 299, 367, 29 },
		{ 299, 368, 29 },
		{ 299, 369, 29 },
		{ 299, 370, 29 },
		{ 299, 371, 29 },
		{ 299, 372, 29 },
		{ 299, 373, 29 },
		{ 299, 374, 29 },
		{ 299, 375, 29 },
		{ 299, 376, 29 },
		{ 299, 377, 29 },
		{ 299, 378, 29 },
		{ 299, 379, 29 },
		{ 299, 380, 29 },
		{ 299, 381, 29 },
		{ 299, 382, 29 },
		{ 299, 383, 29 },
		{ 300, 344, 29 },
		{ 300, 345, 29 },
		{ 300, 346, 29 },
		{ 300, 347, 29 },
		{ 300, 348, 29 },
		{ 300, 349, 29 },
		{ 300, 350, 29 },
		{ 300, 351, 29 },
		{ 300, 352, 29 },
		{ 300, 353, 29 },
		{ 300, 354, 29 },
		{ 300, 355, 29 },
		{ 300, 356, 29 },
		{ 300, 357, 29 },
		{ 300, 358, 29 },
		{ 300, 359, 29 },
		{ 300, 360, 29 },
		{ 300, 361, 29 },
		{ 300, 362, 29 },
		{ 300, 363, 29 },
		{ 300, 364, 29 },
		{ 300, 365, 29 },
		{ 300, 366, 29 },
		{ 300, 367, 29 },
		{ 300, 368, 29 },
		{ 300, 369, 29 },
		{ 300, 370, 29 },
		{ 300, 371, 29 },
		{ 300, 372, 29 },
		{ 300, 373, 29 },
		{ 300, 374, 29 },
		{ 300, 375, 29 },
		{ 300, 376, 29 },
		{ 300, 377, 29 },
		{ 300, 378, 29 },
		{ 300, 379, 29 },
		{ 300, 380, 29 },
		{ 300, 381, 29 },
		{ 300, 382, 29 },
		{ 300, 383, 29 },
		{ 301, 346, 29 },
		{ 301, 347, 29 },
		{ 301, 348, 29 },
		{ 301, 349, 29 },
		{ 301, 350, 29 },
		{ 301, 351, 29 },
		{ 301, 352, 29 },
		{ 301, 353, 29 },
		{ 301, 354, 29 },
		{ 301, 355, 29 },
		{ 301, 356, 29 },
		{ 301, 357, 29 },
		{ 301, 358, 29 },
		{ 301, 359, 29 },
		{ 301, 360, 29 },
		{ 301, 361, 29 },
		{ 301, 362, 29 },
		{ 301, 363, 29 },
		{ 301, 364, 29 },
		{ 301, 365, 29 },
		{ 301, 366, 29 },
		{ 301, 367, 29 },
		{ 301, 368, 29 },
		{ 301, 369, 29 },
		{ 301, 370, 29 },
		{ 301, 371, 29 },
		{ 301, 372, 29 },
		{ 301, 373, 29 },
		{ 301, 374, 29 },
		{ 301, 375, 29 },
		{ 301, 376, 29 },
		{ 301, 377, 29 },
		{ 301, 378, 29 },
		{ 301, 379, 29 },
		{ 301, 380, 29 },
		{ 301, 381, 29 },
		{ 301, 382, 29 },
		{ 301, 383, 29 },
		{ 302, 348, 29 },
		{ 302, 349, 29 },
		{ 302, 350, 29 },
		{ 302, 351, 29 },
		{ 302, 352, 29 },
		{ 302, 353, 29 },
		{ 302, 354, 29 },
		{ 302, 355, 29 },
		{ 302, 356, 29 },
		{ 302, 357, 29 },
		{ 302, 358, 29 },
		{ 302, 359, 29 },
		{ 302, 360, 29 },
		{ 302, 361, 29 },
		{ 302, 362, 29 },
		{ 302, 363, 29 },
		{ 302, 364, 29 },
		{ 302, 365, 29 },
		{ 302, 366, 29 },
		{ 302, 367, 29 },
		{ 302, 368, 29 },
		{ 302, 369, 29 },
		{ 302, 370, 29 },
		{ 302, 371, 29 },
		{ 302, 372, 29 },
		{ 302, 373, 29 },
		{ 302, 374, 29 },
		{ 302, 375, 29 },
		{ 302, 376, 29 },
		{ 302, 377, 29 },
		{ 302, 378, 29 },
		{ 302, 379, 29 },
		{ 302, 380, 29 },
		{ 302, 381, 29 },
		{ 302, 382, 29 },
		{ 302, 383, 29 },
		{ 303, 350, 29 },
		{ 303, 351, 29 },
		{ 303, 352, 29 },
		{ 303, 353, 29 },
		{ 303, 354, 29 },
		{ 303, 355, 29 },
		{ 303, 356, 29 },
		{ 303, 357, 29 },
		{ 303, 358, 29 },
		{ 303, 359, 29 },
		{ 303, 360, 29 },
		{ 303, 361, 29 },
		{ 303, 362, 29 },
		{ 303, 363, 29 },
		{ 303, 364, 29 },
		{ 303, 365, 29 },
		{ 303, 366, 29 },
		{ 303, 367, 29 },
		{ 303, 368, 29 },
		{ 303, 369, 29 },
		{ 303, 370, 29 },
		{ 303, 371, 29 },
		{ 303, 372, 29 },
		{ 303, 373, 29 },
		{ 303, 374, 29 },
		{ 303, 375, 29 },
		{ 303, 376, 29 },
		{ 303, 377, 29 },
		{ 303, 378, 29 },
		{ 303, 379, 29 },
		{ 303, 380, 29 },
		{ 303, 381, 29 },
		{ 303, 382, 29 },
		{ 303, 383, 29 },
		{ 304, 352, 29 },
		{ 304, 353, 29 },
		{ 304, 354, 29 },
		{ 304, 355, 29 },
		{ 304, 356, 29 },
		{ 304, 357, 29 },
		{ 304, 358, 29 },
		{ 304, 359, 29 },
		{ 304, 360, 29 },
		{ 304, 361, 29 },
		{ 304, 362, 29 },
		{ 304, 363, 29 },
		{ 304, 364, 29 },
		{ 304, 365, 29 },
		{ 304, 366, 29 },
		{ 304, 367, 29 },
		{ 304, 368, 29 },
		{ 304, 369, 29 },
		{ 304, 370, 29 },
		{ 304, 371, 29 },
		{ 304, 372, 29 },
		{ 304, 373, 29 },
		{ 304, 374, 29 },
		{ 304, 375, 29 },
		{ 304, 376, 29 },
		{ 304, 377, 29 },
		{ 304, 378, 29 },
		{ 304, 379, 29 },
		{ 304, 380, 29 },
		{ 304, 381, 29 },
		{ 304, 382, 29 },
		{ 304, 383, 29 },
		{ 305, 354, 29 },
		{ 305, 355, 29 },
		{ 305, 356, 29 },
		{ 305, 357, 29 },
		{ 305, 358, 29 },
		{ 305, 359, 29 },
		{ 305, 360, 29 },
		{ 305, 361, 29 },
		{ 305, 362, 29 },
		{ 305, 363, 29 },
		{ 305, 364, 29 },
		{ 305, 365, 29 },
		{ 305, 366, 29 },
		{ 305, 367, 29 },
		{ 305, 368, 29 },
		{ 305, 369, 29 },
		{ 305, 370, 29 },
		{ 305, 371, 29 },
		{ 305, 372, 29 },
		{ 305, 373, 29 },
		{ 305, 374, 29 },
		{ 305, 375, 29 },
		{ 305, 376, 29 },
		{ 305, 377, 29 },
		{ 305, 378, 29 },
		{ 305, 379, 29 },
		{ 305, 380, 29 },
		{ 305, 381, 29 },
		{ 305, 382, 29 },
		{ 305, 383, 29 },
		{ 306, 356, 29 },
		{ 306, 357, 29 },
		{ 306, 358, 29 },
		{ 306, 359, 29 },
		{ 306, 360, 29 },
		{ 306, 361, 29 },
		{ 306, 362, 29 },
		{ 306, 363, 29 },
		{ 306, 364, 29 },
		{ 306, 365, 29 },
		{ 306, 366, 29 },
		{ 306, 367, 29 },
		{ 306, 368, 29 },
		{ 306, 369, 29 },
		{ 306, 370, 29 },
		{ 306, 371, 29 },
		{ 306, 372, 29 },
		{ 306, 373, 29 },
		{ 306, 374, 29 },
		{ 306, 375, 29 },
		{ 306, 376, 29 },
		{ 306, 377, 29 },
		{ 306, 378, 29 },
		{ 306, 379, 29 },
		{ 306, 380, 29 },
		{ 306, 381, 29 },
		{ 306, 382, 29 },
		{ 306, 383, 29 },
		{ 307, 358, 29 },
		{ 307, 359, 29 },
		{ 307, 360, 29 },
		{ 307, 361, 29 },
		{ 307, 362, 29 },
		{ 307, 363, 29 },
		{ 307, 364, 29 },
		{ 307, 365, 29 },
		{ 307, 366, 29 },
		{ 307, 367, 29 },
		{ 307, 368, 29 },
		{ 307, 369, 29 },
		{ 307, 370, 29 },
		{ 307, 371, 29 },
		{ 307, 372, 29 },
		{ 307, 373, 29 },
		{ 307, 374, 29 },
		{ 307, 375, 29 },
		{ 307, 376, 29 },
		{ 307, 377, 29 },
		{ 307, 378, 29 },
		{ 307, 379, 29 },
		{ 307, 380, 29 },
		{ 307, 381, 29 },
		{ 307, 382, 29 },
		{ 307, 383, 29 },
		{ 308, 360, 29 },
		{ 308, 361, 29 },
		{ 308, 362, 29 },
		{ 308, 363, 29 },
		{ 308, 364, 29 },
		{ 308, 365, 29 },
		{ 308, 366, 29 },
		{ 308, 367, 29 },
		{ 308, 368, 29 },
		{ 308, 369, 29 },
		{ 308, 370, 29 },
		{ 308, 371, 29 },
		{ 308, 372, 29 },
		{ 308, 373, 29 },
		{ 308, 374, 29 },
		{ 308, 375, 29 },
		{ 308, 376, 29 },
		{ 308, 377, 29 },
		{ 308, 378, 29 },
		{ 308, 379, 29 },
		{ 308, 380, 29 },
		{ 308, 381, 29 },
		{ 308, 382, 29 },
		{ 308, 383, 29 },
		{ 309, 362, 29 },
		{ 309, 363, 29 },
		{ 309, 364, 29 },
		{ 309, 365, 29 },
		{ 309, 366, 29 },
		{ 309, 367, 29 },
		{ 309, 368, 29 },
		{ 309, 369, 29 },
		{ 309, 370, 29 },
		{ 309, 371, 29 },
		{ 309, 372, 29 },
		{ 309, 373, 29 },
		{ 309, 374, 29 },
		{ 309, 375, 29 },
		{ 309, 376, 29 },
		{ 309, 377, 29 },
		{ 309, 378, 29 },
		{ 309, 379, 29 },
		{ 309, 380, 29 },
		{ 309, 381, 29 },
		{ 309, 382, 29 },
		{ 309, 383, 29 },
		{ 310, 364, 29 },
		{ 310, 365, 29 },
		{ 310, 366, 29 },
		{ 310, 367, 29 },
		{ 310, 368, 29 },
		{ 310, 369, 29 },
		{ 310, 370, 29 },
		{ 310, 371, 29 },
		{ 310, 372, 29 },
		{ 310, 373, 29 },
		{ 310, 374, 29 },
		{ 310, 375, 29 },
		{ 310, 376, 29 },
		{ 310, 377, 29 },
		{ 310, 378, 29 },
		{ 310, 379, 29 },
		{ 310, 380, 29 },
		{ 310, 381, 29 },
		{ 310, 382, 29 },
		{ 310, 383, 29 },
		{ 311, 366, 29 },
		{ 311, 367, 29 },
		{ 311, 368, 29 },
		{ 311, 369, 29 },
		{ 311, 370, 29 },
		{ 311, 371, 29 },
		{ 311, 372, 29 },
		{ 311, 373, 29 },
		{ 311, 374, 29 },
		{ 311, 375, 29 },
		{ 311, 376, 29 },
		{ 311, 377, 29 },
		{ 311, 378, 29 },
		{ 311, 379, 29 },
		{ 311, 380, 29 },
		{ 311, 381, 29 },
		{ 311, 382, 29 },
		{ 311, 383, 29 },
		{ 312, 368, 29 },
		{ 312, 369, 29 },
		{ 312, 370, 29 },
		{ 312, 371, 29 },
		{ 312, 372, 29 },
		{ 312, 373, 29 },
		{ 312, 374, 29 },
		{ 312, 375, 29 },
		{ 312, 376, 29 },
		{ 312, 377, 29 },
		{ 312, 378, 29 },
		{ 312, 379, 29 },
		{ 312, 380, 29 },
		{ 312, 381, 29 },
		{ 312, 382, 29 },
		{ 312, 383, 29 },
		{ 313, 370, 29 },
		{ 313, 371, 29 },
		{ 313, 372, 29 },
		{ 313, 373, 29 },
		{ 313, 374, 29 },
		{ 313, 375, 29 },
		{ 313, 376, 29 },
		{ 313, 377, 29 },
		{ 313, 378, 29 },
		{ 313, 379, 29 },
		{ 313, 380, 29 },
		{ 313, 381, 29 },
		{ 313, 382, 29 },
		{ 313, 383, 29 },
		{ 314, 372, 29 },
		{ 314, 373, 29 },
		{ 314, 374, 29 },
		{ 314, 375, 29 },
		{ 314, 376, 29 },
		{ 314, 377, 29 },
		{ 314, 378, 29 },
		{ 314, 379, 29 },
		{ 314, 380, 29 },
		{ 314, 381, 29 },
		{ 314, 382, 29 },
		{ 315, 374, 29 },
		{ 315, 375, 29 },
		{ 315, 376, 29 },
		{ 315, 377, 29 },
		{ 315, 378, 29 },
		{ 315, 379, 29 },
		{ 315, 380, 29 },
		{ 315, 381, 29 },
		{ 315, 382, 29 },
		{ 316, 376, 29 },
		{ 316, 377, 29 },
		{ 316, 378, 29 },
		{ 316, 379, 29 },
		{ 316, 380, 29 },
		{ 316, 381, 29 },
		{ 316, 382, 29 },
		{ 317, 379, 29 },
		{ 317, 380, 29 },
		{ 317, 381, 29 },
		{ 226, 1, 30 },
		{ 226, 2, 30 },
		{ 226, 3, 30 },
		{ 226, 4, 30 },
		{ 226, 5, 30 },
		{ 226, 6, 30 },
		{ 226, 7, 30 },
		{ 226, 8, 30 },
		{ 226, 506, 30 },
		{ 226, 507, 30 },
		{ 226, 508, 30 },
		{ 226, 509, 30 },
		{ 226, 510, 30 },
		{ 226, 511, 30 },
		{ 226, 512, 30 },
		{ 227, 1, 30 },
		{ 227, 2, 30 },
		{ 227, 3, 30 },
		{ 227, 4, 30 },
		{ 227, 5, 30 },
		{ 227, 6, 30 },
		{ 227, 7, 30 },
		{ 227, 8, 30 },
		{ 227, 9, 30 },
		{ 227, 10, 30 },
		{ 227, 11, 30 },
		{ 227, 12, 30 },
		{ 227, 13, 30 },
		{ 227, 14, 30 },
		{ 227, 15, 30 },
		{ 227, 16, 30 },
		{ 227, 498, 30 },
		{ 227, 499, 30 },
		{ 227, 500, 30 },
		{ 227, 501, 30 },
		{ 227, 502, 30 },
		{ 227, 503, 30 },
		{ 227, 504, 30 },
		{ 227, 505, 30 },
		{ 227, 506, 30 },
		{ 227, 507, 30 },
		{ 227, 508, 30 },
		{ 227, 509, 30 },
		{ 227, 510, 30 },
		{ 227, 511, 30 },
		{ 227, 512, 30 },
		{ 228, 1, 30 },
		{ 228, 2, 30 },
		{ 228, 3, 30 },
		{ 228, 4, 30 },
		{ 228, 5, 30 },
		{ 228, 6, 30 },
		{ 228, 7, 30 },
		{ 228, 8, 30 },
		{ 228, 9, 30 },
		{ 228, 10, 30 },
		{ 228, 11, 30 },
		{ 228, 12, 30 },
		{ 228, 13, 30 },
		{ 228, 14, 30 },
		{ 228, 15, 30 },
		{ 228, 16, 30 },
		{ 228, 17, 30 },
		{ 228, 18, 30 },
		{ 228, 19, 30 },
		{ 228, 20, 30 },
		{ 228, 21, 30 },
		{ 228, 22, 30 },
		{ 228, 23, 30 },
		{ 228, 24, 30 },
		{ 228, 490, 30 },
		{ 228, 491, 30 },
		{ 228, 492, 30 },
		{ 228, 493, 30 },
		{ 228, 494, 30 },
		{ 228, 495, 30 },
		{ 228, 496, 30 },
		{ 228, 497, 30 },
		{ 228, 498, 30 },
		{ 228, 499, 30 },
		{ 228, 500, 30 },
		{ 228, 501, 30 },
		{ 228, 502, 30 },
		{ 228, 503, 30 },
		{ 228, 504, 30 },
		{ 228, 505, 30 },
		{ 228, 506, 30 },
		{ 228, 507, 30 },
		{ 228, 508, 30 },
		{ 228, 509, 30 },
		{ 228, 510, 30 },
		{ 228, 511, 30 },
		{ 228, 512, 30 },
		{ 229, 1, 30 },
		{ 229, 2, 30 },
		{ 229, 3, 30 },
		{ 229, 4, 30 },
		{ 229, 5, 30 },
		{ 229, 6, 30 },
		{ 229, 7, 30 },
		{ 229, 8, 30 },
		{ 229, 9, 30 },
		{ 229, 10, 30 },
		{ 229, 11, 30 },
		{ 229, 12, 30 },
		{ 229, 13, 30 },
		{ 229, 14, 30 },
		{ 229, 15, 30 },
		{ 229, 16, 30 },
		{ 229, 17, 30 },
		{ 229, 18, 30 },
		{ 229, 19, 30 },
		{ 229, 20, 30 },
		{ 229, 21, 30 },
		{ 229, 22, 30 },
		{ 229, 23, 30 },
		{ 229, 24, 30 },
		{ 229, 25, 30 },
		{ 229, 26, 30 },
		{ 229, 27, 30 },
		{ 229, 28, 30 },
		{ 229, 29, 30 },
		{ 229, 30, 30 },
		{ 229, 31, 30 },
		{ 229, 32, 30 },
		{ 229, 482, 30 },
		{ 229, 483, 30 },
		{ 229, 484, 30 },
		{ 229, 485, 30 },
		{ 229, 486, 30 },
		{ 229, 487, 30 },
		{ 229, 488, 30 },
		{ 229, 489, 30 },
		{ 229, 490, 30 },
		{ 229, 491, 30 },
		{ 229, 492, 30 },
		{ 229, 493, 30 },
		{ 229, 494, 30 },
		{ 229, 495, 30 },
		{ 229, 496, 30 },
		{ 229, 497, 30 },
		{ 229, 498, 30 },
		{ 229, 499, 30 },
		{ 229, 500, 30 },
		{ 229, 501, 30 },
		{ 229, 502, 30 },
		{ 229, 503, 30 },
		{ 229, 504, 30 },
		{ 229, 505, 30 },
		{ 229, 506, 30 },
		{ 229, 507, 30 },
		{ 229, 508, 30 },
		{ 229, 509, 30 },
		{ 229, 510, 30 },
		{ 229, 511, 30 },
		{ 229, 512, 30 },
		{ 230, 1, 30 },
		{ 230, 2, 30 },
		{ 230, 3, 30 },
		{ 230, 4, 30 },
		{ 230, 5, 30 },
		{ 230, 6, 30 },
		{ 230, 7, 30 },
		{ 230, 8, 30 },
		{ 230, 9, 30 },
		{ 230, 10, 30 },
		{ 230, 11, 30 },
		{ 230, 12, 30 },
		{ 230, 13, 30 },
		{ 230, 14, 30 },
		{ 230, 15, 30 },
		{ 230, 16, 30 },
		{ 230, 17, 30 },
		{ 230, 18, 30 },
		{ 230, 19, 30 },
		{ 230, 20, 30 },
		{ 230, 21, 30 },
		{ 230, 22, 30 },
		{ 230, 23, 30 },
		{ 230, 24, 30 },
		{ 230, 25, 30 },
		{ 230, 26, 30 },
		{ 230, 27, 30 },
		{ 230, 28, 30 },
		{ 230, 29, 30 },
		{ 230, 30, 30 },
		{ 230, 31, 30 },
		{ 230, 32, 30 },
		{ 230, 33, 30 },
		{ 230, 34, 30 },
		{ 230, 35, 30 },
		{ 230, 36, 30 },
		{ 230, 37, 30 },
		{ 230, 38, 30 },
		{ 230, 39, 30 },
		{ 230, 40, 30 },
		{ 230, 474, 30 },
		{ 230, 475, 30 },
		{ 230, 476, 30 },
		{ 230, 477, 30 },
		{ 230, 478, 30 },
		{ 230, 479, 30 },
		{ 230, 480, 30 },
		{ 230, 481, 30 },
		{ 230, 482, 30 },
		{ 230, 483, 30 },
		{ 230, 484, 30 },
		{ 230, 485, 30 },
		{ 230, 486, 30 },
		{ 230, 487, 30 },
		{ 230, 488, 30 },
		{ 230, 489, 30 },
		{ 230, 490, 30 },
		{ 230, 491, 30 },
		{ 230, 492, 30 },
		{ 230, 493, 30 },
		{ 230, 494, 30 },
		{ 230, 495, 30 },
		{ 230, 496, 30 },
		{ 230, 497, 30 },
		{ 230, 498, 30 },
		{ 230, 499, 30 },
		{ 230, 500, 30 },
		{ 230, 501, 30 },
		{ 230, 502, 30 },
		{ 230, 503, 30 },
		{ 230, 504, 30 },
		{ 230, 505, 30 },
		{ 230, 506, 30 },
		{ 230, 507, 30 },
		{ 230, 508, 30 },
		{ 230, 509, 30 },
		{ 230, 510, 30 },
		{ 230, 511, 30 },
		{ 230, 512, 30 },
		{ 231, 1, 30 },
		{ 231, 2, 30 },
		{ 231, 3, 30 },
		{ 231, 4, 30 },
		{ 231, 5, 30 },
		{ 231, 6, 30 },
		{ 231, 7, 30 },
		{ 231, 8, 30 },
		{ 231, 9, 30 },
		{ 231, 10, 30 },
		{ 231, 11, 30 },
		{ 231, 12, 30 },
		{ 231, 13, 30 },
		{ 231, 14, 30 },
		{ 231, 15, 30 },
		{ 231, 16, 30 },
		{ 231, 17, 30 },
		{ 231, 18, 30 },
		{ 231, 19, 30 },
		{ 231, 20, 30 },
		{ 231, 21, 30 },
		{ 231, 22, 30 },
		{ 231, 23, 30 },
		{ 231, 24, 30 },
		{ 231, 25, 30 },
		{ 231, 26, 30 },
		{ 231, 27, 30 },
		{ 231, 28, 30 },
		{ 231, 29, 30 },
		{ 231, 30, 30 },
		{ 231, 31, 30 },
		{ 231, 32, 30 },
		{ 231, 33, 30 },
		{ 231, 34, 30 },
		{ 231, 35, 30 },
		{ 231, 36, 30 },
		{ 231, 37, 30 },
		{ 231, 38, 30 },
		{ 231, 39, 30 },
		{ 231, 40, 30 },
		{ 231, 41, 30 },
		{ 231, 42, 30 },
		{ 231, 43, 30 },
		{ 231, 44, 30 },
		{ 231, 45, 30 },
		{ 231, 46, 30 },
		{ 231, 47, 30 },
		{ 231, 48, 30 },
		{ 231, 466, 30 },
		{ 231, 467, 30 },
		{ 231, 468, 30 },
		{ 231, 469, 30 },
		{ 231, 470, 30 },
		{ 231, 471, 30 },
		{ 231, 472, 30 },
		{ 231, 473, 30 },
		{ 231, 474, 30 },
		{ 231, 475, 30 },
		{ 231, 476, 30 },
		{ 231, 477, 30 },
		{ 231, 478, 30 },
		{ 231, 479, 30 },
		{ 231, 480, 30 },
		{ 231, 481, 30 },
		{ 231, 482, 30 },
		{ 231, 483, 30 },
		{ 231, 484, 30 },
		{ 231, 485, 30 },
		{ 231, 486, 30 },
		{ 231, 487, 30 },
		{ 231, 488, 30 },
		{ 231, 489, 30 },
		{ 231, 490, 30 },
		{ 231, 491, 30 },
		{ 231, 492, 30 },
		{ 231, 493, 30 },
		{ 231, 494, 30 },
		{ 231, 495, 30 },
		{ 231, 496, 30 },
		{ 231, 497, 30 },
		{ 231, 498, 30 },
		{ 231, 499, 30 },
		{ 231, 500, 30 },
		{ 231, 501, 30 },
		{ 231, 502, 30 },
		{ 231, 503, 30 },
		{ 231, 504, 30 },
		{ 231, 505, 30 },
		{ 231, 506, 30 },
		{ 231, 507, 30 },
		{ 231, 508, 30 },
		{ 231, 509, 30 },
		{ 231, 510, 30 },
		{ 231, 511, 30 },
		{ 231, 512, 30 },
		{ 232, 1, 30 },
		{ 232, 2, 30 },
		{ 232, 3, 30 },
		{ 232, 4, 30 },
		{ 232, 5, 30 },
		{ 232, 6, 30 },
		{ 232, 7, 30 },
		{ 232, 8, 30 },
		{ 232, 9, 30 },
		{ 232, 10, 30 },
		{ 232, 11, 30 },
		{ 232, 12, 30 },
		{ 232, 13, 30 },
		{ 232, 14, 30 },
		{ 232, 15, 30 },
		{ 232, 16, 30 },
		{ 232, 17, 30 },
		{ 232, 18, 30 },
		{ 232, 19, 30 },
		{ 232, 20, 30 },
		{ 232, 21, 30 },
		{ 232, 22, 30 },
		{ 232, 23, 30 },
		{ 232, 24, 30 },
		{ 232, 25, 30 },
		{ 232, 26, 30 },
		{ 232, 27, 30 },
		{ 232, 28, 30 },
		{ 232, 29, 30 },
		{ 232, 30, 30 },
		{ 232, 31, 30 },
		{ 232, 32, 30 },
		{ 232, 33, 30 },
		{ 232, 34, 30 },
		{ 232, 35, 30 },
		{ 232, 36, 30 },
		{ 232, 37, 30 },
		{ 232, 38, 30 },
		{ 232, 39, 30 },
		{ 232, 40, 30 },
		{ 232, 41, 30 },
		{ 232, 42, 30 },
		{ 232, 43, 30 },
		{ 232, 44, 30 },
		{ 232, 45, 30 },
		{ 232, 46, 30 },
		{ 232, 47, 30 },
		{ 232, 48, 30 },
		{ 232, 49, 30 },
		{ 232, 50, 30 },
		{ 232, 51, 30 },
		{ 232, 52, 30 },
		{ 232, 53, 30 },
		{ 232, 54, 30 },
		{ 232, 55, 30 },
		{ 232, 56, 30 },
		{ 232, 458, 30 },
		{ 232, 459, 30 },
		{ 232, 460, 30 },
		{ 232, 461, 30 },
		{ 232, 462, 30 },
		{ 232, 463, 30 },
		{ 232, 464, 30 },
		{ 232, 465, 30 },
		{ 232, 466, 30 },
		{ 232, 467, 30 },
		{ 232, 468, 30 },
		{ 232, 469, 30 },
		{ 232, 470, 30 },
		{ 232, 471, 30 },
		{ 232, 472, 30 },
		{ 232, 473, 30 },
		{ 232, 474, 30 },
		{ 232, 475, 30 },
		{ 232, 476, 30 },
		{ 232, 477, 30 },
		{ 232, 478, 30 },
		{ 232, 479, 30 },
		{ 232, 480, 30 },
		{ 232, 481, 30 },
		{ 232, 482, 30 },
		{ 232, 483, 30 },
		{ 232, 484, 30 },
		{ 232, 485, 30 },
		{ 232, 486, 30 },
		{ 232, 487, 30 },
		{ 232, 488, 30 },
		{ 232, 489, 30 },
		{ 232, 490, 30 },
		{ 232, 491, 30 },
		{ 232, 492, 30 },
		{ 232, 493, 30 },
		{ 232, 494, 30 },
		{ 232, 495, 30 },
		{ 232, 496, 30 },
		{ 232, 497, 30 },
		{ 232, 498, 30 },
		{ 232, 499, 30 },
		{ 232, 500, 30 },
		{ 232, 501, 30 },
		{ 232, 502, 30 },
		{ 232, 503, 30 },
		{ 232, 504, 30 },
		{ 232, 505, 30 },
		{ 232, 506, 30 },
		{ 232, 507, 30 },
		{ 232, 508, 30 },
		{ 232, 509, 30 },
		{ 232, 510, 30 },
		{ 232, 511, 30 },
		{ 232, 512, 30 },
		{ 233, 1, 30 },
		{ 233, 2, 30 },
		{ 233, 3, 30 },
		{ 233, 4, 30 },
		{ 233, 5, 30 },
		{ 233, 6, 30 },
		{ 233, 7, 30 },
		{ 233, 8, 30 },
		{ 233, 9, 30 },
		{ 233, 10, 30 },
		{ 233, 11, 30 },
		{ 233, 12, 30 },
		{ 233, 13, 30 },
		{ 233, 14, 30 },
		{ 233, 15, 30 },
		{ 233, 16, 30 },
		{ 233, 17, 30 },
		{ 233, 18, 30 },
		{ 233, 19, 30 },
		{ 233, 20, 30 },
		{ 233, 21, 30 },
		{ 233, 22, 30 },
		{ 233, 23, 30 },
		{ 233, 24, 30 },
		{ 233, 25, 30 },
		{ 233, 26, 30 },
		{ 233, 27, 30 },
		{ 233, 28, 30 },
		{ 233, 29, 30 },
		{ 233, 30, 30 },
		{ 233, 31, 30 },
		{ 233, 32, 30 },
		{ 233, 33, 30 },
		{ 233, 34, 30 },
		{ 233, 35, 30 },
		{ 233, 36, 30 },
		{ 233, 37, 30 },
		{ 233, 38, 30 },
		{ 233, 39, 30 },
		{ 233, 40, 30 },
		{ 233, 41, 30 },
		{ 233, 42, 30 },
		{ 233, 43, 30 },
		{ 233, 44, 30 },
		{ 233, 45, 30 },
		{ 233, 46, 30 },
		{ 233, 47, 30 },
		{ 233, 48, 30 },
		{ 233, 49, 30 },
		{ 233, 50, 30 },
		{ 233, 51, 30 },
		{ 233, 52, 30 },
		{ 233, 53, 30 },
		{ 233, 54, 30 },
		{ 233, 55, 30 },
		{ 233, 56, 30 },
		{ 233, 57, 30 },
		{ 233, 58, 30 },
		{ 233, 59, 30 },
		{ 233, 60, 30 },
		{ 233, 61, 30 },
		{ 233, 62, 30 },
		{ 233, 63, 30 },
		{ 233, 64, 30 },
		{ 233, 450, 30 },
		{ 233, 451, 30 },
		{ 233, 452, 30 },
		{ 233, 453, 30 },
		{ 233, 454, 30 },
		{ 233, 455, 30 },
		{ 233, 456, 30 },
		{ 233, 457, 30 },
		{ 233, 458, 30 },
		{ 233, 459, 30 },
		{ 233, 460, 30 },
		{ 233, 461, 30 },
		{ 233, 462, 30 },
		{ 233, 463, 30 },
		{ 233, 464, 30 },
		{ 233, 465, 30 },
		{ 233, 466, 30 },
		{ 233, 467, 30 },
		{ 233, 468, 30 },
		{ 233, 469, 30 },
		{ 233, 470, 30 },
		{ 233, 471, 30 },
		{ 233, 472, 30 },
		{ 233, 473, 30 },
		{ 233, 474, 30 },
		{ 233, 475, 30 },
		{ 233, 476, 30 },
		{ 233, 477, 30 },
		{ 233, 478, 30 },
		{ 233, 479, 30 },
		{ 233, 480, 30 },
		{ 233, 481, 30 },
		{ 233, 482, 30 },
		{ 233, 483, 30 },
		{ 233, 484, 30 },
		{ 233, 485, 30 },
		{ 233, 486, 30 },
		{ 233, 487, 30 },
		{ 233, 488, 30 },
		{ 233, 489, 30 },
		{ 233, 490, 30 },
		{ 233, 491, 30 },
		{ 233, 492, 30 },
		{ 233, 493, 30 },
		{ 233, 494, 30 },
		{ 233, 495, 30 },
		{ 233, 496, 30 },
		{ 233, 497, 30 },
		{ 233, 498, 30 },
		{ 233, 499, 30 },
		{ 233, 500, 30 },
		{ 233, 501, 30 },
		{ 233, 502, 30 },
		{ 233, 503, 30 },
		{ 233, 504, 30 },
		{ 233, 505, 30 },
		{ 233, 506, 30 },
		{ 233, 507, 30 },
		{ 233, 508, 30 },
		{ 233, 509, 30 },
		{ 233, 510, 30 },
		{ 233, 511, 30 },
		{ 233, 512, 30 },
		{ 234, 1, 30 },
		{ 234, 2, 30 },
		{ 234, 3, 30 },
		{ 234, 4, 30 },
		{ 234, 5, 30 },
		{ 234, 6, 30 },
		{ 234, 7, 30 },
		{ 234, 8, 30 },
		{ 234, 9, 30 },
		{ 234, 10, 30 },
		{ 234, 11, 30 },
		{ 234, 12, 30 },
		{ 234, 13, 30 },
		{ 234, 14, 30 },
		{ 234, 15, 30 },
		{ 234, 16, 30 },
		{ 234, 17, 30 },
		{ 234, 18, 30 },
		{ 234, 19, 30 },
		{ 234, 20, 30 },
		{ 234, 21, 30 },
		{ 234, 22, 30 },
		{ 234, 23, 30 },
		{ 234, 24, 30 },
		{ 234, 25, 30 },
		{ 234, 26, 30 },
		{ 234, 27, 30 },
		{ 234, 28, 30 },
		{ 234, 29, 30 },
		{ 234, 30, 30 },
		{ 234, 31, 30 },
		{ 234, 32, 30 },
		{ 234, 33, 30 },
		{ 234, 34, 30 },
		{ 234, 35, 30 },
		{ 234, 36, 30 },
		{ 234, 37, 30 },
		{ 234, 38, 30 },
		{ 234, 39, 30 },
		{ 234, 40, 30 },
		{ 234, 41, 30 },
		{ 234, 42, 30 },
		{ 234, 43, 30 },
		{ 234, 44, 30 },
		{ 234, 45, 30 },
		{ 234, 46, 30 },
		{ 234, 47, 30 },
		{ 234, 48, 30 },
		{ 234, 49, 30 },
		{ 234, 50, 30 },
		{ 234, 51, 30 },
		{ 234, 52, 30 },
		{ 234, 53, 30 },
		{ 234, 54, 30 },
		{ 234, 55, 30 },
		{ 234, 56, 30 },
		{ 234, 57, 30 },
		{ 234, 58, 30 },
		{ 234, 59, 30 },
		{ 234, 60, 30 },
		{ 234, 61, 30 },
		{ 234, 62, 30 },
		{ 234, 63, 30 },
		{ 234, 64, 30 },
		{ 234, 65, 30 },
		{ 234, 66, 30 },
		{ 234, 67, 30 },
		{ 234, 68, 30 },
		{ 234, 69, 30 },
		{ 234, 70, 30 },
		{ 234, 71, 30 },
		{ 234, 72, 30 },
		{ 234, 442, 30 },
		{ 234, 443, 30 },
		{ 234, 444, 30 },
		{ 234, 445, 30 },
		{ 234, 446, 30 },
		{ 234, 447, 30 },
		{ 234, 448, 30 },
		{ 234, 449, 30 },
		{ 234, 450, 30 },
		{ 234, 451, 30 },
		{ 234, 452, 30 },
		{ 234, 453, 30 },
		{ 234, 454, 30 },
		{ 234, 455, 30 },
		{ 234, 456, 30 },
		{ 234, 457, 30 },
		{ 234, 458, 30 },
		{ 234, 459, 30 },
		{ 234, 460, 30 },
		{ 234, 461, 30 },
		{ 234, 462, 30 },
		{ 234, 463, 30 },
		{ 234, 464, 30 },
		{ 234, 465, 30 },
		{ 234, 466, 30 },
		{ 234, 467, 30 },
		{ 234, 468, 30 },
		{ 234, 469, 30 },
		{ 234, 470, 30 },
		{ 234, 471, 30 },
		{ 234, 472, 30 },
		{ 234, 473, 30 },
		{ 234, 474, 30 },
		{ 234, 475, 30 },
		{ 234, 476, 30 },
		{ 234, 477, 30 },
		{ 234, 478, 30 },
		{ 234, 479, 30 },
		{ 234, 480, 30 },
		{ 234, 481, 30 },
		{ 234, 482, 30 },
		{ 234, 483, 30 },
		{ 234, 484, 30 },
		{ 234, 485, 30 },
		{ 234, 486, 30 },
		{ 234, 487, 30 },
		{ 234, 488, 30 },
		{ 234, 489, 30 },
		{ 234, 490, 30 },
		{ 234, 491, 30 },
		{ 234, 492, 30 },
		{ 234, 493, 30 },
		{ 234, 494, 30 },
		{ 234, 495, 30 },
		{ 234, 496, 30 },
		{ 234, 497, 30 },
		{ 234, 498, 30 },
		{ 234, 499, 30 },
		{ 234, 500, 30 },
		{ 234, 501, 30 },
		{ 234, 502, 30 },
		{ 234, 503, 30 },
		{ 234, 504, 30 },
		{ 234, 505, 30 },
		{ 234, 506, 30 },
		{ 234, 507, 30 },
		{ 234, 508, 30 },
		{ 234, 509, 30 },
		{ 234, 510, 30 },
		{ 234, 511, 30 },
		{ 234, 512, 30 },
		{ 235, 1, 30 },
		{ 235, 2, 30 },
		{ 235, 3, 30 },
		{ 235, 4, 30 },
		{ 235, 5, 30 },
		{ 235, 6, 30 },
		{ 235, 7, 30 },
		{ 235, 8, 30 },
		{ 235, 9, 30 },
		{ 235, 10, 30 },
		{ 235, 11, 30 },
		{ 235, 12, 30 },
		{ 235, 13, 30 },
		{ 235, 14, 30 },
		{ 235, 15, 30 },
		{ 235, 16, 30 },
		{ 235, 17, 30 },
		{ 235, 18, 30 },
		{ 235, 19, 30 },
		{ 235, 20, 30 },
		{ 235, 21, 30 },
		{ 235, 22, 30 },
		{ 235, 23, 30 },
		{ 235, 24, 30 },
		{ 235, 25, 30 },
		{ 235, 26, 30 },
		{ 235, 27, 30 },
		{ 235, 28, 30 },
		{ 235, 29, 30 },
		{ 235, 30, 30 },
		{ 235, 31, 30 },
		{ 235, 32, 30 },
		{ 235, 33, 30 },
		{ 235, 34, 30 },
		{ 235, 35, 30 },
		{ 235, 36, 30 },
		{ 235, 37, 30 },
		{ 235, 38, 30 },
		{ 235, 39, 30 },
		{ 235, 40, 30 },
		{ 235, 41, 30 },
		{ 235, 42, 30 },
		{ 235, 43, 30 },
		{ 235, 44, 30 },
		{ 235, 45, 30 },
		{ 235, 46, 30 },
		{ 235, 47, 30 },
		{ 235, 48, 30 },
		{ 235, 49, 30 },
		{ 235, 50, 30 },
		{ 235, 51, 30 },
		{ 235, 52, 30 },
		{ 235, 53, 30 },
		{ 235, 54, 30 },
		{ 235, 55, 30 },
		{ 235, 56, 30 },
		{ 235, 57, 30 },
		{ 235, 58, 30 },
		{ 235, 59, 30 },
		{ 235, 60, 30 },
		{ 235, 61, 30 },
		{ 235, 62, 30 },
		{ 235, 63, 30 },
		{ 235, 64, 30 },
		{ 235, 65, 30 },
		{ 235, 66, 30 },
		{ 235, 67, 30 },
		{ 235, 68, 30 },
		{ 235, 69, 30 },
		{ 235, 70, 30 },
		{ 235, 71, 30 },
		{ 235, 72, 30 },
		{ 235, 73, 30 },
		{ 235, 74, 30 },
		{ 235, 75, 30 },
		{ 235, 76, 30 },
		{ 235, 77, 30 },
		{ 235, 78, 30 },
		{ 235, 79, 30 },
		{ 235, 80, 30 },
		{ 235, 434, 30 },
		{ 235, 435, 30 },
		{ 235, 436, 30 },
		{ 235, 437, 30 },
		{ 235, 438, 30 },
		{ 235, 439, 30 },
		{ 235, 440, 30 },
		{ 235, 441, 30 },
		{ 235, 442, 30 },
		{ 235, 443, 30 },
		{ 235, 444, 30 },
		{ 235, 445, 30 },
		{ 235, 446, 30 },
		{ 235, 447, 30 },
		{ 235, 448, 30 },
		{ 235, 449, 30 },
		{ 235, 450, 30 },
		{ 235, 451, 30 },
		{ 235, 452, 30 },
		{ 235, 453, 30 },
		{ 235, 454, 30 },
		{ 235, 455, 30 },
		{ 235, 456, 30 },
		{ 235, 457, 30 },
		{ 235, 458, 30 },
		{ 235, 459, 30 },
		{ 235, 460, 30 },
		{ 235, 461, 30 },
		{ 235, 462, 30 },
		{ 235, 463, 30 },
		{ 235, 464, 30 },
		{ 235, 465, 30 },
		{ 235, 466, 30 },
		{ 235, 467, 30 },
		{ 235, 468, 30 },
		{ 235, 469, 30 },
		{ 235, 470, 30 },
		{ 235, 471, 30 },
		{ 235, 472, 30 },
		{ 235, 473, 30 },
		{ 235, 474, 30 },
		{ 235, 475, 30 },
		{ 235, 476, 30 },
		{ 235, 477, 30 },
		{ 235, 478, 30 },
		{ 235, 479, 30 },
		{ 235, 480, 30 },
		{ 235, 481, 30 },
		{ 235, 482, 30 },
		{ 235, 483, 30 },
		{ 235, 484, 30 },
		{ 235, 485, 30 },
		{ 235, 486, 30 },
		{ 235, 487, 30 },
		{ 235, 488, 30 },
		{ 235, 489, 30 },
		{ 235, 490, 30 },
		{ 235, 491, 30 },
		{ 235, 492, 30 },
		{ 235, 493, 30 },
		{ 235, 494, 30 },
		{ 235, 495, 30 },
		{ 235, 496, 30 },
		{ 235, 497, 30 },
		{ 235, 498, 30 },
		{ 235, 499, 30 },
		{ 235, 500, 30 },
		{ 235, 501, 30 },
		{ 235, 502, 30 },
		{ 235, 503, 30 },
		{ 235, 504, 30 },
		{ 235, 505, 30 },
		{ 235, 506, 30 },
		{ 235, 507, 30 },
		{ 235, 508, 30 },
		{ 235, 509, 30 },
		{ 235, 510, 30 },
		{ 235, 511, 30 },
		{ 235, 512, 30 },
		{ 236, 1, 30 },
		{ 236, 2, 30 },
		{ 236, 3, 30 },
		{ 236, 4, 30 },
		{ 236, 5, 30 },
		{ 236, 6, 30 },
		{ 236, 7, 30 },
		{ 236, 8, 30 },
		{ 236, 9, 30 },
		{ 236, 10, 30 },
		{ 236, 11, 30 },
		{ 236, 12, 30 },
		{ 236, 13, 30 },
		{ 236, 14, 30 },
		{ 236, 15, 30 },
		{ 236, 16, 30 },
		{ 236, 17, 30 },
		{ 236, 18, 30 },
		{ 236, 19, 30 },
		{ 236, 20, 30 },
		{ 236, 21, 30 },
		{ 236, 22, 30 },
		{ 236, 23, 30 },
		{ 236, 24, 30 },
		{ 236, 25, 30 },
		{ 236, 26, 30 },
		{ 236, 27, 30 },
		{ 236, 28, 30 },
		{ 236, 29, 30 },
		{ 236, 30, 30 },
		{ 236, 31, 30 },
		{ 236, 32, 30 },
		{ 236, 33, 30 },
		{ 236, 34, 30 },
		{ 236, 35, 30 },
		{ 236, 36, 30 },
		{ 236, 37, 30 },
		{ 236, 38, 30 },
		{ 236, 39, 30 },
		{ 236, 40, 30 },
		{ 236, 41, 30 },
		{ 236, 42, 30 },
		{ 236, 43, 30 },
		{ 236, 44, 30 },
		{ 236, 45, 30 },
		{ 236, 46, 30 },
		{ 236, 47, 30 },
		{ 236, 48, 30 },
		{ 236, 49, 30 },
		{ 236, 50, 30 },
		{ 236, 51, 30 },
		{ 236, 52, 30 },
		{ 236, 53, 30 },
		{ 236, 54, 30 },
		{ 236, 55, 30 },
		{ 236, 56, 30 },
		{ 236, 57, 30 },
		{ 236, 58, 30 },
		{ 236, 59, 30 },
		{ 236, 60, 30 },
		{ 236, 61, 30 },
		{ 236, 62, 30 },
		{ 236, 63, 30 },
		{ 236, 64, 30 },
		{ 236, 65, 30 },
		{ 236, 66, 30 },
		{ 236, 67, 30 },
		{ 236, 68, 30 },
		{ 236, 69, 30 },
		{ 236, 70, 30 },
		{ 236, 71, 30 },
		{ 236, 72, 30 },
		{ 236, 73, 30 },
		{ 236, 74, 30 },
		{ 236, 75, 30 },
		{ 236, 76, 30 },
		{ 236, 77, 30 },
		{ 236, 78, 30 },
		{ 236, 79, 30 },
		{ 236, 80, 30 },
		{ 236, 81, 30 },
		{ 236, 82, 30 },
		{ 236, 83, 30 },
		{ 236, 84, 30 },
		{ 236, 85, 30 },
		{ 236, 86, 30 },
		{ 236, 87, 30 },
		{ 236, 88, 30 },
		{ 236, 426, 30 },
		{ 236, 427, 30 },
		{ 236, 428, 30 },
		{ 236, 429, 30 },
		{ 236, 430, 30 },
		{ 236, 431, 30 },
		{ 236, 432, 30 },
		{ 236, 433, 30 },
		{ 236, 434, 30 },
		{ 236, 435, 30 },
		{ 236, 436, 30 },
		{ 236, 437, 30 },
		{ 236, 438, 30 },
		{ 236, 439, 30 },
		{ 236, 440, 30 },
		{ 236, 441, 30 },
		{ 236, 442, 30 },
		{ 236, 443, 30 },
		{ 236, 444, 30 },
		{ 236, 445, 30 },
		{ 236, 446, 30 },
		{ 236, 447, 30 },
		{ 236, 448, 30 },
		{ 236, 449, 30 },
		{ 236, 450, 30 },
		{ 236, 451, 30 },
		{ 236, 452, 30 },
		{ 236, 453, 30 },
		{ 236, 454, 30 },
		{ 236, 455, 30 },
		{ 236, 456, 30 },
		{ 236, 457, 30 },
		{ 236, 458, 30 },
		{ 236, 459, 30 },
		{ 236, 460, 30 },
		{ 236, 461, 30 },
		{ 236, 462, 30 },
		{ 236, 463, 30 },
		{ 236, 464, 30 },
		{ 236, 465, 30 },
		{ 236, 466, 30 },
		{ 236, 467, 30 },
		{ 236, 468, 30 },
		{ 236, 469, 30 },
		{ 236, 470, 30 },
		{ 236, 471, 30 },
		{ 236, 472, 30 },
		{ 236, 473, 30 },
		{ 236, 474, 30 },
		{ 236, 475, 30 },
		{ 236, 476, 30 },
		{ 236, 477, 30 },
		{ 236, 478, 30 },
		{ 236, 479, 30 },
		{ 236, 480, 30 },
		{ 236, 481, 30 },
		{ 236, 482, 30 },
		{ 236, 483, 30 },
		{ 236, 484, 30 },
		{ 236, 485, 30 },
		{ 236, 486, 30 },
		{ 236, 487, 30 },
		{ 236, 488, 30 },
		{ 236, 489, 30 },
		{ 236, 490, 30 },
		{ 236, 491, 30 },
		{ 236, 492, 30 },
		{ 236, 493, 30 },
		{ 236, 494, 30 },
		{ 236, 495, 30 },
		{ 236, 496, 30 },
		{ 236, 497, 30 },
		{ 236, 498, 30 },
		{ 236, 499, 30 },
		{ 236, 500, 30 },
		{ 236, 501, 30 },
		{ 236, 502, 30 },
		{ 236, 503, 30 },
		{ 236, 504, 30 },
		{ 236, 505, 30 },
		{ 236, 506, 30 },
		{ 236, 507, 30 },
		{ 236, 508, 30 },
		{ 236, 509, 30 },
		{ 236, 510, 30 },
		{ 236, 511, 30 },
		{ 236, 512, 30 },
		{ 237, 1, 30 },
		{ 237, 2, 30 },
		{ 237, 3, 30 },
		{ 237, 4, 30 },
		{ 237, 5, 30 },
		{ 237, 6, 30 },
		{ 237, 7, 30 },
		{ 237, 8, 30 },
		{ 237, 9, 30 },
		{ 237, 10, 30 },
		{ 237, 11, 30 },
		{ 237, 12, 30 },
		{ 237, 13, 30 },
		{ 237, 14, 30 },
		{ 237, 15, 30 },
		{ 237, 16, 30 },
		{ 237, 17, 30 },
		{ 237, 18, 30 },
		{ 237, 19, 30 },
		{ 237, 20, 30 },
		{ 237, 21, 30 },
		{ 237, 22, 30 },
		{ 237, 23, 30 },
		{ 237, 24, 30 },
		{ 237, 25, 30 },
		{ 237, 26, 30 },
		{ 237, 27, 30 },
		{ 237, 28, 30 },
		{ 237, 29, 30 },
		{ 237, 30, 30 },
		{ 237, 31, 30 },
		{ 237, 32, 30 },
		{ 237, 33, 30 },
		{ 237, 34, 30 },
		{ 237, 35, 30 },
		{ 237, 36, 30 },
		{ 237, 37, 30 },
		{ 237, 38, 30 },
		{ 237, 39, 30 },
		{ 237, 40, 30 },
		{ 237, 41, 30 },
		{ 237, 42, 30 },
		{ 237, 43, 30 },
		{ 237, 44, 30 },
		{ 237, 45, 30 },
		{ 237, 46, 30 },
		{ 237, 47, 30 },
		{ 237, 48, 30 },
		{ 237, 49, 30 },
		{ 237, 50, 30 },
		{ 237, 51, 30 },
		{ 237, 52, 30 },
		{ 237, 53, 30 },
		{ 237, 54, 30 },
		{ 237, 55, 30 },
		{ 237, 56, 30 },
		{ 237, 57, 30 },
		{ 237, 58, 30 },
		{ 237, 59, 30 },
		{ 237, 60, 30 },
		{ 237, 61, 30 },
		{ 237, 62, 30 },
		{ 237, 63, 30 },
		{ 237, 64, 30 },
		{ 237, 65, 30 },
		{ 237, 66, 30 },
		{ 237, 67, 30 },
		{ 237, 68, 30 },
		{ 237, 69, 30 },
		{ 237, 70, 30 },
		{ 237, 71, 30 },
		{ 237, 72, 30 },
		{ 237, 73, 30 },
		{ 237, 74, 30 },
		{ 237, 75, 30 },
		{ 237, 76, 30 },
		{ 237, 77, 30 },
		{ 237, 78, 30 },
		{ 237, 79, 30 },
		{ 237, 80, 30 },
		{ 237, 81, 30 },
		{ 237, 82, 30 },
		{ 237, 83, 30 },
		{ 237, 84, 30 },
		{ 237, 85, 30 },
		{ 237, 86, 30 },
		{ 237, 87, 30 },
		{ 237, 88, 30 },
		{ 237, 89, 30 },
		{ 237, 90, 30 },
		{ 237, 91, 30 },
		{ 237, 92, 30 },
		{ 237, 93, 30 },
		{ 237, 94, 30 },
		{ 237, 95, 30 },
		{ 237, 96, 30 },
		{ 237, 418, 30 },
		{ 237, 419, 30 },
		{ 237, 420, 30 },
		{ 237, 421, 30 },
		{ 237, 422, 30 },
		{ 237, 423, 30 },
		{ 237, 424, 30 },
		{ 237, 425, 30 },
		{ 237, 426, 30 },
		{ 237, 427, 30 },
		{ 237, 428, 30 },
		{ 237, 429, 30 },
		{ 237, 430, 30 },
		{ 237, 431, 30 },
		{ 237, 432, 30 },
		{ 237, 433, 30 },
		{ 237, 434, 30 },
		{ 237, 435, 30 },
		{ 237, 436, 30 },
		{ 237, 437, 30 },
		{ 237, 438, 30 },
		{ 237, 439, 30 },
		{ 237, 440, 30 },
		{ 237, 441, 30 },
		{ 237, 442, 30 },
		{ 237, 443, 30 },
		{ 237, 444, 30 },
		{ 237, 445, 30 },
		{ 237, 446, 30 },
		{ 237, 447, 30 },
		{ 237, 448, 30 },
		{ 237, 449, 30 },
		{ 237, 450, 30 },
		{ 237, 451, 30 },
		{ 237, 452, 30 },
		{ 237, 453, 30 },
		{ 237, 454, 30 },
		{ 237, 455, 30 },
		{ 237, 456, 30 },
		{ 237, 457, 30 },
		{ 237, 458, 30 },
		{ 237, 459, 30 },
		{ 237, 460, 30 },
		{ 237, 461, 30 },
		{ 237, 462, 30 },
		{ 237, 463, 30 },
		{ 237, 464, 30 },
		{ 237, 465, 30 },
		{ 237, 466, 30 },
		{ 237, 467, 30 },
		{ 237, 468, 30 },
		{ 237, 469, 30 },
		{ 237, 470, 30 },
		{ 237, 471, 30 },
		{ 237, 472, 30 },
		{ 237, 473, 30 },
		{ 237, 474, 30 },
		{ 237, 475, 30 },
		{ 237, 476, 30 },
		{ 237, 477, 30 },
		{ 237, 478, 30 },
		{ 237, 479, 30 },
		{ 237, 480, 30 },
		{ 237, 481, 30 },
		{ 237, 482, 30 },
		{ 237, 483, 30 },
		{ 237, 484, 30 },
		{ 237, 485, 30 },
		{ 237, 486, 30 },
		{ 237, 487, 30 },
		{ 237, 488, 30 },
		{ 237, 489, 30 },
		{ 237, 490, 30 },
		{ 237, 491, 30 },
		{ 237, 492, 30 },
		{ 237, 493, 30 },
		{ 237, 494, 30 },
		{ 237, 495, 30 },
		{ 237, 496, 30 },
		{ 237, 497, 30 },
		{ 237, 498, 30 },
		{ 237, 499, 30 },
		{ 237, 500, 30 },
		{ 237, 501, 30 },
		{ 237, 502, 30 },
		{ 237, 503, 30 },
		{ 237, 504, 30 },
		{ 237, 505, 30 },
		{ 237, 506, 30 },
		{ 237, 507, 30 },
		{ 237, 508, 30 },
		{ 237, 509, 30 },
		{ 237, 510, 30 },
		{ 237, 511, 30 },
		{ 237, 512, 30 },
		{ 238, 1, 30 },
		{ 238, 2, 30 },
		{ 238, 3, 30 },
		{ 238, 4, 30 },
		{ 238, 5, 30 },
		{ 238, 6, 30 },
		{ 238, 7, 30 },
		{ 238, 8, 30 },
		{ 238, 9, 30 },
		{ 238, 10, 30 },
		{ 238, 11, 30 },
		{ 238, 12, 30 },
		{ 238, 13, 30 },
		{ 238, 14, 30 },
		{ 238, 15, 30 },
		{ 238, 16, 30 },
		{ 238, 17, 30 },
		{ 238, 18, 30 },
		{ 238, 19, 30 },
		{ 238, 20, 30 },
		{ 238, 21, 30 },
		{ 238, 22, 30 },
		{ 238, 23, 30 },
		{ 238, 24, 30 },
		{ 238, 25, 30 },
		{ 238, 26, 30 },
		{ 238, 27, 30 },
		{ 238, 28, 30 },
		{ 238, 29, 30 },
		{ 238, 30, 30 },
		{ 238, 31, 30 },
		{ 238, 32, 30 },
		{ 238, 33, 30 },
		{ 238, 34, 30 },
		{ 238, 35, 30 },
		{ 238, 36, 30 },
		{ 238, 37, 30 },
		{ 238, 38, 30 },
		{ 238, 39, 30 },
		{ 238, 40, 30 },
		{ 238, 41, 30 },
		{ 238, 42, 30 },
		{ 238, 43, 30 },
		{ 238, 44, 30 },
		{ 238, 45, 30 },
		{ 238, 46, 30 },
		{ 238, 47, 30 },
		{ 238, 48, 30 },
		{ 238, 49, 30 },
		{ 238, 50, 30 },
		{ 238, 51, 30 },
		{ 238, 52, 30 },
		{ 238, 53, 30 },
		{ 238, 54, 30 },
		{ 238, 55, 30 },
		{ 238, 56, 30 },
		{ 238, 57, 30 },
		{ 238, 58, 30 },
		{ 238, 59, 30 },
		{ 238, 60, 30 },
		{ 238, 61, 30 },
		{ 238, 62, 30 },
		{ 238, 63, 30 },
		{ 238, 64, 30 },
		{ 238, 65, 30 },
		{ 238, 66, 30 },
		{ 238, 67, 30 },
		{ 238, 68, 30 },
		{ 238, 69, 30 },
		{ 238, 70, 30 },
		{ 238, 71, 30 },
		{ 238, 72, 30 },
		{ 238, 73, 30 },
		{ 238, 74, 30 },
		{ 238, 75, 30 },
		{ 238, 76, 30 },
		{ 238, 77, 30 },
		{ 238, 78, 30 },
		{ 238, 79, 30 },
		{ 238, 80, 30 },
		{ 238, 81, 30 },
		{ 238, 82, 30 },
		{ 238, 83, 30 },
		{ 238, 84, 30 },
		{ 238, 85, 30 },
		{ 238, 86, 30 },
		{ 238, 87, 30 },
		{ 238, 88, 30 },
		{ 238, 89, 30 },
		{ 238, 90, 30 },
		{ 238, 91, 30 },
		{ 238, 92, 30 },
		{ 238, 93, 30 },
		{ 238, 94, 30 },
		{ 238, 95, 30 },
		{ 238, 96, 30 },
		{ 238, 97, 30 },
		{ 238, 98, 30 },
		{ 238, 99, 30 },
		{ 238, 100, 30 },
		{ 238, 101, 30 },
		{ 238, 102, 30 },
		{ 238, 103, 30 },
		{ 238, 104, 30 },
		{ 238, 410, 30 },
		{ 238, 411, 30 },
		{ 238, 412, 30 },
		{ 238, 413, 30 },
		{ 238, 414, 30 },
		{ 238, 415, 30 },
		{ 238, 416, 30 },
		{ 238, 417, 30 },
		{ 238, 418, 30 },
		{ 238, 419, 30 },
		{ 238, 420, 30 },
		{ 238, 421, 30 },
		{ 238, 422, 30 },
		{ 238, 423, 30 },
		{ 238, 424, 30 },
		{ 238, 425, 30 },
		{ 238, 426, 30 },
		{ 238, 427, 30 },
		{ 238, 428, 30 },
		{ 238, 429, 30 },
		{ 238, 430, 30 },
		{ 238, 431, 30 },
		{ 238, 432, 30 },
		{ 238, 433, 30 },
		{ 238, 434, 30 },
		{ 238, 435, 30 },
		{ 238, 436, 30 },
		{ 238, 437, 30 },
		{ 238, 438, 30 },
		{ 238, 439, 30 },
		{ 238, 440, 30 },
		{ 238, 441, 30 },
		{ 238, 442, 30 },
		{ 238, 443, 30 },
		{ 238, 444, 30 },
		{ 238, 445, 30 },
		{ 238, 446, 30 },
		{ 238, 447, 30 },
		{ 238, 448, 30 },
		{ 238, 449, 30 },
		{ 238, 450, 30 },
		{ 238, 451, 30 },
		{ 238, 452, 30 },
		{ 238, 453, 30 },
		{ 238, 454, 30 },
		{ 238, 455, 30 },
		{ 238, 456, 30 },
		{ 238, 457, 30 },
		{ 238, 458, 30 },
		{ 238, 459, 30 },
		{ 238, 460, 30 },
		{ 238, 461, 30 },
		{ 238, 462, 30 },
		{ 238, 463, 30 },
		{ 238, 464, 30 },
		{ 238, 465, 30 },
		{ 238, 466, 30 },
		{ 238, 467, 30 },
		{ 238, 468, 30 },
		{ 238, 469, 30 },
		{ 238, 470, 30 },
		{ 238, 471, 30 },
		{ 238, 472, 30 },
		{ 238, 473, 30 },
		{ 238, 474, 30 },
		{ 238, 475, 30 },
		{ 238, 476, 30 },
		{ 238, 477, 30 },
		{ 238, 478, 30 },
		{ 238, 479, 30 },
		{ 238, 480, 30 },
		{ 238, 481, 30 },
		{ 238, 482, 30 },
		{ 238, 483, 30 },
		{ 238, 484, 30 },
		{ 238, 485, 30 },
		{ 238, 486, 30 },
		{ 238, 487, 30 },
		{ 238, 488, 30 },
		{ 238, 489, 30 },
		{ 238, 490, 30 },
		{ 238, 491, 30 },
		{ 238, 492, 30 },
		{ 238, 493, 30 },
		{ 238, 494, 30 },
		{ 238, 495, 30 },
		{ 238, 496, 30 },
		{ 238, 497, 30 },
		{ 238, 498, 30 },
		{ 238, 499, 30 },
		{ 238, 500, 30 },
		{ 238, 501, 30 },
		{ 238, 502, 30 },
		{ 238, 503, 30 },
		{ 238, 504, 30 },
		{ 238, 505, 30 },
		{ 238, 506, 30 },
		{ 238, 507, 30 },
		{ 238, 508, 30 },
		{ 238, 509, 30 },
		{ 238, 510, 30 },
		{ 238, 511, 30 },
		{ 238, 512, 30 },
		{ 239, 1, 30 },
		{ 239, 2, 30 },
		{ 239, 3, 30 },
		{ 239, 4, 30 },
		{ 239, 5, 30 },
		{ 239, 6, 30 },
		{ 239, 7, 30 },
		{ 239, 8, 30 },
		{ 239, 9, 30 },
		{ 239, 10, 30 },
		{ 239, 11, 30 },
		{ 239, 12, 30 },
		{ 239, 13, 30 },
		{ 239, 14, 30 },
		{ 239, 15, 30 },
		{ 239, 16, 30 },
		{ 239, 17, 30 },
		{ 239, 18, 30 },
		{ 239, 19, 30 },
		{ 239, 20, 30 },
		{ 239, 21, 30 },
		{ 239, 22, 30 },
		{ 239, 23, 30 },
		{ 239, 24, 30 },
		{ 239, 25, 30 },
		{ 239, 26, 30 },
		{ 239, 27, 30 },
		{ 239, 28, 30 },
		{ 239, 29, 30 },
		{ 239, 30, 30 },
		{ 239, 31, 30 },
		{ 239, 32, 30 },
		{ 239, 33, 30 },
		{ 239, 34, 30 },
		{ 239, 35, 30 },
		{ 239, 36, 30 },
		{ 239, 37, 30 },
		{ 239, 38, 30 },
		{ 239, 39, 30 },
		{ 239, 40, 30 },
		{ 239, 41, 30 },
		{ 239, 42, 30 },
		{ 239, 43, 30 },
		{ 239, 44, 30 },
		{ 239, 45, 30 },
		{ 239, 46, 30 },
		{ 239, 47, 30 },
		{ 239, 48, 30 },
		{ 239, 49, 30 },
		{ 239, 50, 30 },
		{ 239, 51, 30 },
		{ 239, 52, 30 },
		{ 239, 53, 30 },
		{ 239, 54, 30 },
		{ 239, 55, 30 },
		{ 239, 56, 30 },
		{ 239, 57, 30 },
		{ 239, 58, 30 },
		{ 239, 59, 30 },
		{ 239, 60, 30 },
		{ 239, 61, 30 },
		{ 239, 62, 30 },
		{ 239, 63, 30 },
		{ 239, 64, 30 },
		{ 239, 65, 30 },
		{ 239, 66, 30 },
		{ 239, 67, 30 },
		{ 239, 68, 30 },
		{ 239, 69, 30 },
		{ 239, 70, 30 },
		{ 239, 71, 30 },
		{ 239, 72, 30 },
		{ 239, 73, 30 },
		{ 239, 74, 30 },
		{ 239, 75, 30 },
		{ 239, 76, 30 },
		{ 239, 77, 30 },
		{ 239, 78, 30 },
		{ 239, 79, 30 },
		{ 239, 80, 30 },
		{ 239, 81, 30 },
		{ 239, 82, 30 },
		{ 239, 83, 30 },
		{ 239, 84, 30 },
		{ 239, 85, 30 },
		{ 239, 86, 30 },
		{ 239, 87, 30 },
		{ 239, 88, 30 },
		{ 239, 89, 30 },
		{ 239, 90, 30 },
		{ 239, 91, 30 },
		{ 239, 92, 30 },
		{ 239, 93, 30 },
		{ 239, 94, 30 },
		{ 239, 95, 30 },
		{ 239, 96, 30 },
		{ 239, 97, 30 },
		{ 239, 98, 30 },
		{ 239, 99, 30 },
		{ 239, 100, 30 },
		{ 239, 101, 30 },
		{ 239, 102, 30 },
		{ 239, 103, 30 },
		{ 239, 104, 30 },
		{ 239, 105, 30 },
		{ 239, 106, 30 },
		{ 239, 107, 30 },
		{ 239, 108, 30 },
		{ 239, 109, 30 },
		{ 239, 110, 30 },
		{ 239, 111, 30 },
		{ 239, 112, 30 },
		{ 239, 402, 30 },
		{ 239, 403, 30 },
		{ 239, 404, 30 },
		{ 239, 405, 30 },
		{ 239, 406, 30 },
		{ 239, 407, 30 },
		{ 239, 408, 30 },
		{ 239, 409, 30 },
		{ 239, 410, 30 },
		{ 239, 411, 30 },
		{ 239, 412, 30 },
		{ 239, 413, 30 },
		{ 239, 414, 30 },
		{ 239, 415, 30 },
		{ 239, 416, 30 },
		{ 239, 417, 30 },
		{ 239, 418, 30 },
		{ 239, 419, 30 },
		{ 239, 420, 30 },
		{ 239, 421, 30 },
		{ 239, 422, 30 },
		{ 239, 423, 30 },
		{ 239, 424, 30 },
		{ 239, 425, 30 },
		{ 239, 426, 30 },
		{ 239, 427, 30 },
		{ 239, 428, 30 },
		{ 239, 429, 30 },
		{ 239, 430, 30 },
		{ 239, 431, 30 },
		{ 239, 432, 30 },
		{ 239, 433, 30 },
		{ 239, 434, 30 },
		{ 239, 435, 30 },
		{ 239, 436, 30 },
		{ 239, 437, 30 },
		{ 239, 438, 30 },
		{ 239, 439, 30 },
		{ 239, 440, 30 },
		{ 239, 441, 30 },
		{ 239, 442, 30 },
		{ 239, 443, 30 },
		{ 239, 444, 30 },
		{ 239, 445, 30 },
		{ 239, 446, 30 },
		{ 239, 447, 30 },
		{ 239, 448, 30 },
		{ 239, 449, 30 },
		{ 239, 450, 30 },
		{ 239, 451, 30 },
		{ 239, 452, 30 },
		{ 239, 453, 30 },
		{ 239, 454, 30 },
		{ 239, 455, 30 },
		{ 239, 456, 30 },
		{ 239, 457, 30 },
		{ 239, 458, 30 },
		{ 239, 459, 30 },
		{ 239, 460, 30 },
		{ 239, 461, 30 },
		{ 239, 462, 30 },
		{ 239, 463, 30 },
		{ 239, 464, 30 },
		{ 239, 465, 30 },
		{ 239, 466, 30 },
		{ 239, 467, 30 },
		{ 239, 468, 30 },
		{ 239, 469, 30 },
		{ 239, 470, 30 },
		{ 239, 471, 30 },
		{ 239, 472, 30 },
		{ 239, 473, 30 },
		{ 239, 474, 30 },
		{ 239, 475, 30 },
		{ 239, 476, 30 },
		{ 239, 477, 30 },
		{ 239, 478, 30 },
		{ 239, 479, 30 },
		{ 239, 480, 30 },
		{ 239, 481, 30 },
		{ 239, 482, 30 },
		{ 239, 483, 30 },
		{ 239, 484, 30 },
		{ 239, 485, 30 },
		{ 239, 486, 30 },
		{ 239, 487, 30 },
		{ 239, 488, 30 },
		{ 239, 489, 30 },
		{ 239, 490, 30 },
		{ 239, 491, 30 },
		{ 239, 492, 30 },
		{ 239, 493, 30 },
		{ 239, 494, 30 },
		{ 239, 495, 30 },
		{ 239, 496, 30 },
		{ 239, 497, 30 },
		{ 239, 498, 30 },
		{ 239, 499, 30 },
		{ 239, 500, 30 },
		{ 239, 501, 30 },
		{ 239, 502, 30 },
		{ 239, 503, 30 },
		{ 239, 504, 30 },
		{ 239, 505, 30 },
		{ 239, 506, 30 },
		{ 239, 507, 30 },
		{ 239, 508, 30 },
		{ 239, 509, 30 },
		{ 239, 510, 30 },
		{ 239, 511, 30 },
		{ 239, 512, 30 },
		{ 240, 1, 30 },
		{ 240, 2, 30 },
		{ 240, 3, 30 },
		{ 240, 4, 30 },
		{ 240, 5, 30 },
		{ 240, 6, 30 },
		{ 240, 7, 30 },
		{ 240, 8, 30 },
		{ 240, 9, 30 },
		{ 240, 10, 30 },
		{ 240, 11, 30 },
		{ 240, 12, 30 },
		{ 240, 13, 30 },
		{ 240, 14, 30 },
		{ 240, 15, 30 },
		{ 240, 16, 30 },
		{ 240, 17, 30 },
		{ 240, 18, 30 },
		{ 240, 19, 30 },
		{ 240, 20, 30 },
		{ 240, 21, 30 },
		{ 240, 22, 30 },
		{ 240, 23, 30 },
		{ 240, 24, 30 },
		{ 240, 25, 30 },
		{ 240, 26, 30 },
		{ 240, 27, 30 },
		{ 240, 28, 30 },
		{ 240, 29, 30 },
		{ 240, 30, 30 },
		{ 240, 31, 30 },
		{ 240, 32, 30 },
		{ 240, 33, 30 },
		{ 240, 34, 30 },
		{ 240, 35, 30 },
		{ 240, 36, 30 },
		{ 240, 37, 30 },
		{ 240, 38, 30 },
		{ 240, 39, 30 },
		{ 240, 40, 30 },
		{ 240, 41, 30 },
		{ 240, 42, 30 },
		{ 240, 43, 30 },
		{ 240, 44, 30 },
		{ 240, 45, 30 },
		{ 240, 46, 30 },
		{ 240, 47, 30 },
		{ 240, 48, 30 },
		{ 240, 49, 30 },
		{ 240, 50, 30 },
		{ 240, 51, 30 },
		{ 240, 52, 30 },
		{ 240, 53, 30 },
		{ 240, 54, 30 },
		{ 240, 55, 30 },
		{ 240, 56, 30 },
		{ 240, 57, 30 },
		{ 240, 58, 30 },
		{ 240, 59, 30 },
		{ 240, 60, 30 },
		{ 240, 61, 30 },
		{ 240, 62, 30 },
		{ 240, 63, 30 },
		{ 240, 64, 30 },
		{ 240, 65, 30 },
		{ 240, 66, 30 },
		{ 240, 67, 30 },
		{ 240, 68, 30 },
		{ 240, 69, 30 },
		{ 240, 70, 30 },
		{ 240, 71, 30 },
		{ 240, 72, 30 },
		{ 240, 73, 30 },
		{ 240, 74, 30 },
		{ 240, 75, 30 },
		{ 240, 76, 30 },
		{ 240, 77, 30 },
		{ 240, 78, 30 },
		{ 240, 79, 30 },
		{ 240, 80, 30 },
		{ 240, 81, 30 },
		{ 240, 82, 30 },
		{ 240, 83, 30 },
		{ 240, 84, 30 },
		{ 240, 85, 30 },
		{ 240, 86, 30 },
		{ 240, 87, 30 },
		{ 240, 88, 30 },
		{ 240, 89, 30 },
		{ 240, 90, 30 },
		{ 240, 91, 30 },
		{ 240, 92, 30 },
		{ 240, 93, 30 },
		{ 240, 94, 30 },
		{ 240, 95, 30 },
		{ 240, 96, 30 },
		{ 240, 97, 30 },
		{ 240, 98, 30 },
		{ 240, 99, 30 },
		{ 240, 100, 30 },
		{ 240, 101, 30 },
		{ 240, 102, 30 },
		{ 240, 103, 30 },
		{ 240, 104, 30 },
		{ 240, 105, 30 },
		{ 240, 106, 30 },
		{ 240, 107, 30 },
		{ 240, 108, 30 },
		{ 240, 109, 30 },
		{ 240, 110, 30 },
		{ 240, 111, 30 },
		{ 240, 112, 30 },
		{ 240, 113, 30 },
		{ 240, 114, 30 },
		{ 240, 115, 30 },
		{ 240, 116, 30 },
		{ 240, 117, 30 },
		{ 240, 118, 30 },
		{ 240, 119, 30 },
		{ 240, 120, 30 },
		{ 240, 394, 30 },
		{ 240, 395, 30 },
		{ 240, 396, 30 },
		{ 240, 397, 30 },
		{ 240, 398, 30 },
		{ 240, 399, 30 },
		{ 240, 400, 30 },
		{ 240, 401, 30 },
		{ 240, 402, 30 },
		{ 240, 403, 30 },
		{ 240, 404, 30 },
		{ 240, 405, 30 },
		{ 240, 406, 30 },
		{ 240, 407, 30 },
		{ 240, 408, 30 },
		{ 240, 409, 30 },
		{ 240, 410, 30 },
		{ 240, 411, 30 },
		{ 240, 412, 30 },
		{ 240, 413, 30 },
		{ 240, 414, 30 },
		{ 240, 415, 30 },
		{ 240, 416, 30 },
		{ 240, 417, 30 },
		{ 240, 418, 30 },
		{ 240, 419, 30 },
		{ 240, 420, 30 },
		{ 240, 421, 30 },
		{ 240, 422, 30 },
		{ 240, 423, 30 },
		{ 240, 424, 30 },
		{ 240, 425, 30 },
		{ 240, 426, 30 },
		{ 240, 427, 30 },
		{ 240, 428, 30 },
		{ 240, 429, 30 },
		{ 240, 430, 30 },
		{ 240, 431, 30 },
		{ 240, 432, 30 },
		{ 240, 433, 30 },
		{ 240, 434, 30 },
		{ 240, 435, 30 },
		{ 240, 436, 30 },
		{ 240, 437, 30 },
		{ 240, 438, 30 },
		{ 240, 439, 30 },
		{ 240, 440, 30 },
		{ 240, 441, 30 },
		{ 240, 442, 30 },
		{ 240, 443, 30 },
		{ 240, 444, 30 },
		{ 240, 445, 30 },
		{ 240, 446, 30 },
		{ 240, 447, 30 },
		{ 240, 448, 30 },
		{ 240, 449, 30 },
		{ 240, 450, 30 },
		{ 240, 451, 30 },
		{ 240, 452, 30 },
		{ 240, 453, 30 },
		{ 240, 454, 30 },
		{ 240, 455, 30 },
		{ 240, 456, 30 },
		{ 240, 457, 30 },
		{ 240, 458, 30 },
		{ 240, 459, 30 },
		{ 240, 460, 30 },
		{ 240, 461, 30 },
		{ 240, 462, 30 },
		{ 240, 463, 30 },
		{ 240, 464, 30 },
		{ 240, 465, 30 },
		{ 240, 466, 30 },
		{ 240, 467, 30 },
		{ 240, 468, 30 },
		{ 240, 469, 30 },
		{ 240, 470, 30 },
		{ 240, 471, 30 },
		{ 240, 472, 30 },
		{ 240, 473, 30 },
		{ 240, 474, 30 },
		{ 240, 475, 30 },
		{ 240, 476, 30 },
		{ 240, 477, 30 },
		{ 240, 478, 30 },
		{ 240, 479, 30 },
		{ 240, 480, 30 },
		{ 240, 481, 30 },
		{ 240, 482, 30 },
		{ 240, 483, 30 },
		{ 240, 484, 30 },
		{ 240, 485, 30 },
		{ 240, 486, 30 },
		{ 240, 487, 30 },
		{ 240, 488, 30 },
		{ 240, 489, 30 },
		{ 240, 490, 30 },
		{ 240, 491, 30 },
		{ 240, 492, 30 },
		{ 240, 493, 30 },
		{ 240, 494, 30 },
		{ 240, 495, 30 },
		{ 240, 496, 30 },
		{ 240, 497, 30 },
		{ 240, 498, 30 },
		{ 240, 499, 30 },
		{ 240, 500, 30 },
		{ 240, 501, 30 },
		{ 240, 502, 30 },
		{ 240, 503, 30 },
		{ 240, 504, 30 },
		{ 240, 505, 30 },
		{ 240, 506, 30 },
		{ 240, 507, 30 },
		{ 240, 508, 30 },
		{ 240, 509, 30 },
		{ 240, 510, 30 },
		{ 240, 511, 30 },
		{ 240, 512, 30 },
		{ 241, 1, 30 },
		{ 241, 2, 30 },
		{ 241, 3, 30 },
		{ 241, 4, 30 },
		{ 241, 5, 30 },
		{ 241, 6, 30 },
		{ 241, 7, 30 },
		{ 241, 8, 30 },
		{ 241, 9, 30 },
		{ 241, 10, 30 },
		{ 241, 11, 30 },
		{ 241, 12, 30 },
		{ 241, 13, 30 },
		{ 241, 14, 30 },
		{ 241, 15, 30 },
		{ 241, 16, 30 },
		{ 241, 17, 30 },
		{ 241, 18, 30 },
		{ 241, 19, 30 },
		{ 241, 20, 30 },
		{ 241, 21, 30 },
		{ 241, 22, 30 },
		{ 241, 23, 30 },
		{ 241, 24, 30 },
		{ 241, 25, 30 },
		{ 241, 26, 30 },
		{ 241, 27, 30 },
		{ 241, 28, 30 },
		{ 241, 29, 30 },
		{ 241, 30, 30 },
		{ 241, 31, 30 },
		{ 241, 32, 30 },
		{ 241, 33, 30 },
		{ 241, 34, 30 },
		{ 241, 35, 30 },
		{ 241, 36, 30 },
		{ 241, 37, 30 },
		{ 241, 38, 30 },
		{ 241, 39, 30 },
		{ 241, 40, 30 },
		{ 241, 41, 30 },
		{ 241, 42, 30 },
		{ 241, 43, 30 },
		{ 241, 44, 30 },
		{ 241, 45, 30 },
		{ 241, 46, 30 },
		{ 241, 47, 30 },
		{ 241, 48, 30 },
		{ 241, 49, 30 },
		{ 241, 50, 30 },
		{ 241, 51, 30 },
		{ 241, 52, 30 },
		{ 241, 53, 30 },
		{ 241, 54, 30 },
		{ 241, 55, 30 },
		{ 241, 56, 30 },
		{ 241, 57, 30 },
		{ 241, 58, 30 },
		{ 241, 59, 30 },
		{ 241, 60, 30 },
		{ 241, 61, 30 },
		{ 241, 62, 30 },
		{ 241, 63, 30 },
		{ 241, 64, 30 },
		{ 241, 65, 30 },
		{ 241, 66, 30 },
		{ 241, 67, 30 },
		{ 241, 68, 30 },
		{ 241, 69, 30 },
		{ 241, 70, 30 },
		{ 241, 71, 30 },
		{ 241, 72, 30 },
		{ 241, 73, 30 },
		{ 241, 74, 30 },
		{ 241, 75, 30 },
		{ 241, 76, 30 },
		{ 241, 77, 30 },
		{ 241, 78, 30 },
		{ 241, 79, 30 },
		{ 241, 80, 30 },
		{ 241, 81, 30 },
		{ 241, 82, 30 },
		{ 241, 83, 30 },
		{ 241, 84, 30 },
		{ 241, 85, 30 },
		{ 241, 86, 30 },
		{ 241, 87, 30 },
		{ 241, 88, 30 },
		{ 241, 89, 30 },
		{ 241, 90, 30 },
		{ 241, 91, 30 },
		{ 241, 92, 30 },
		{ 241, 93, 30 },
		{ 241, 94, 30 },
		{ 241, 95, 30 },
		{ 241, 96, 30 },
		{ 241, 97, 30 },
		{ 241, 98, 30 },
		{ 241, 99, 30 },
		{ 241, 100, 30 },
		{ 241, 101, 30 },
		{ 241, 102, 30 },
		{ 241, 103, 30 },
		{ 241, 104, 30 },
		{ 241, 105, 30 },
		{ 241, 106, 30 },
		{ 241, 107, 30 },
		{ 241, 108, 30 },
		{ 241, 109, 30 },
		{ 241, 110, 30 },
		{ 241, 111, 30 },
		{ 241, 112, 30 },
		{ 241, 113, 30 },
		{ 241, 114, 30 },
		{ 241, 115, 30 },
		{ 241, 116, 30 },
		{ 241, 117, 30 },
		{ 241, 118, 30 },
		{ 241, 119, 30 },
		{ 241, 120, 30 },
		{ 241, 121, 30 },
		{ 241, 122, 30 },
		{ 241, 123, 30 },
		{ 241, 124, 30 },
		{ 241, 125, 30 },
		{ 241, 126, 30 },
		{ 241, 127, 30 },
		{ 241, 128, 30 },
		{ 241, 386, 30 },
		{ 241, 387, 30 },
		{ 241, 388, 30 },
		{ 241, 389, 30 },
		{ 241, 390, 30 },
		{ 241, 391, 30 },
		{ 241, 392, 30 },
		{ 241, 393, 30 },
		{ 241, 394, 30 },
		{ 241, 395, 30 },
		{ 241, 396, 30 },
		{ 241, 397, 30 },
		{ 241, 398, 30 },
		{ 241, 399, 30 },
		{ 241, 400, 30 },
		{ 241, 401, 30 },
		{ 241, 402, 30 },
		{ 241, 403, 30 },
		{ 241, 404, 30 },
		{ 241, 405, 30 },
		{ 241, 406, 30 },
		{ 241, 407, 30 },
		{ 241, 408, 30 },
		{ 241, 409, 30 },
		{ 241, 410, 30 },
		{ 241, 411, 30 },
		{ 241, 412, 30 },
		{ 241, 413, 30 },
		{ 241, 414, 30 },
		{ 241, 415, 30 },
		{ 241, 416, 30 },
		{ 241, 417, 30 },
		{ 241, 418, 30 },
		{ 241, 419, 30 },
		{ 241, 420, 30 },
		{ 241, 421, 30 },
		{ 241, 422, 30 },
		{ 241, 423, 30 },
		{ 241, 424, 30 },
		{ 241, 425, 30 },
		{ 241, 426, 30 },
		{ 241, 427, 30 },
		{ 241, 428, 30 },
		{ 241, 429, 30 },
		{ 241, 430, 30 },
		{ 241, 431, 30 },
		{ 241, 432, 30 },
		{ 241, 433, 30 },
		{ 241, 434, 30 },
		{ 241, 435, 30 },
		{ 241, 436, 30 },
		{ 241, 437, 30 },
		{ 241, 438, 30 },
		{ 241, 439, 30 },
		{ 241, 440, 30 },
		{ 241, 441, 30 },
		{ 241, 442, 30 },
		{ 241, 443, 30 },
		{ 241, 444, 30 },
		{ 241, 445, 30 },
		{ 241, 446, 30 },
		{ 241, 447, 30 },
		{ 241, 448, 30 },
		{ 241, 449, 30 },
		{ 241, 450, 30 },
		{ 241, 451, 30 },
		{ 241, 452, 30 },
		{ 241, 453, 30 },
		{ 241, 454, 30 },
		{ 241, 455, 30 },
		{ 241, 456, 30 },
		{ 241, 457, 30 },
		{ 241, 458, 30 },
		{ 241, 459, 30 },
		{ 241, 460, 30 },
		{ 241, 461, 30 },
		{ 241, 462, 30 },
		{ 241, 463, 30 },
		{ 241, 464, 30 },
		{ 241, 465, 30 },
		{ 241, 466, 30 },
		{ 241, 467, 30 },
		{ 241, 468, 30 },
		{ 241, 469, 30 },
		{ 241, 470, 30 },
		{ 241, 471, 30 },
		{ 241, 472, 30 },
		{ 241, 473, 30 },
		{ 241, 474, 30 },
		{ 241, 475, 30 },
		{ 241, 476, 30 },
		{ 241, 477, 30 },
		{ 241, 478, 30 },
		{ 241, 479, 30 },
		{ 241, 480, 30 },
		{ 241, 481, 30 },
		{ 241, 482, 30 },
		{ 241, 483, 30 },
		{ 241, 484, 30 },
		{ 241, 485, 30 },
		{ 241, 486, 30 },
		{ 241, 487, 30 },
		{ 241, 488, 30 },
		{ 241, 489, 30 },
		{ 241, 490, 30 },
		{ 241, 491, 30 },
		{ 241, 492, 30 },
		{ 241, 493, 30 },
		{ 241, 494, 30 },
		{ 241, 495, 30 },
		{ 241, 496, 30 },
		{ 241, 497, 30 },
		{ 241, 498, 30 },
		{ 241, 499, 30 },
		{ 241, 500, 30 },
		{ 241, 501, 30 },
		{ 241, 502, 30 },
		{ 241, 503, 30 },
		{ 241, 504, 30 },
		{ 241, 505, 30 },
		{ 241, 506, 30 },
		{ 241, 507, 30 },
		{ 241, 508, 30 },
		{ 241, 509, 30 },
		{ 241, 510, 30 },
		{ 241, 511, 30 },
		{ 241, 512, 30 },
		{ 242, 1, 30 },
		{ 242, 2, 30 },
		{ 242, 3, 30 },
		{ 242, 4, 30 },
		{ 242, 5, 30 },
		{ 242, 6, 30 },
		{ 242, 7, 30 },
		{ 242, 8, 30 },
		{ 242, 9, 30 },
		{ 242, 10, 30 },
		{ 242, 11, 30 },
		{ 242, 12, 30 },
		{ 242, 13, 30 },
		{ 242, 14, 30 },
		{ 242, 15, 30 },
		{ 242, 16, 30 },
		{ 242, 17, 30 },
		{ 242, 18, 30 },
		{ 242, 19, 30 },
		{ 242, 20, 30 },
		{ 242, 21, 30 },
		{ 242, 22, 30 },
		{ 242, 23, 30 },
		{ 242, 24, 30 },
		{ 242, 25, 30 },
		{ 242, 26, 30 },
		{ 242, 27, 30 },
		{ 242, 28, 30 },
		{ 242, 29, 30 },
		{ 242, 30, 30 },
		{ 242, 31, 30 },
		{ 242, 32, 30 },
		{ 242, 33, 30 },
		{ 242, 34, 30 },
		{ 242, 35, 30 },
		{ 242, 36, 30 },
		{ 242, 37, 30 },
		{ 242, 38, 30 },
		{ 242, 39, 30 },
		{ 242, 40, 30 },
		{ 242, 41, 30 },
		{ 242, 42, 30 },
		{ 242, 43, 30 },
		{ 242, 44, 30 },
		{ 242, 45, 30 },
		{ 242, 46, 30 },
		{ 242, 47, 30 },
		{ 242, 48, 30 },
		{ 242, 49, 30 },
		{ 242, 50, 30 },
		{ 242, 51, 30 },
		{ 242, 52, 30 },
		{ 242, 53, 30 },
		{ 242, 54, 30 },
		{ 242, 55, 30 },
		{ 242, 56, 30 },
		{ 242, 57, 30 },
		{ 242, 58, 30 },
		{ 242, 59, 30 },
		{ 242, 60, 30 },
		{ 242, 61, 30 },
		{ 242, 62, 30 },
		{ 242, 63, 30 },
		{ 242, 64, 30 },
		{ 242, 65, 30 },
		{ 242, 66, 30 },
		{ 242, 67, 30 },
		{ 242, 68, 30 },
		{ 242, 69, 30 },
		{ 242, 70, 30 },
		{ 242, 71, 30 },
		{ 242, 72, 30 },
		{ 242, 73, 30 },
		{ 242, 74, 30 },
		{ 242, 75, 30 },
		{ 242, 76, 30 },
		{ 242, 77, 30 },
		{ 242, 78, 30 },
		{ 242, 79, 30 },
		{ 242, 80, 30 },
		{ 242, 81, 30 },
		{ 242, 82, 30 },
		{ 242, 83, 30 },
		{ 242, 84, 30 },
		{ 242, 85, 30 },
		{ 242, 86, 30 },
		{ 242, 87, 30 },
		{ 242, 88, 30 },
		{ 242, 89, 30 },
		{ 242, 90, 30 },
		{ 242, 91, 30 },
		{ 242, 92, 30 },
		{ 242, 93, 30 },
		{ 242, 94, 30 },
		{ 242, 95, 30 },
		{ 242, 96, 30 },
		{ 242, 97, 30 },
		{ 242, 98, 30 },
		{ 242, 99, 30 },
		{ 242, 100, 30 },
		{ 242, 101, 30 },
		{ 242, 102, 30 },
		{ 242, 103, 30 },
		{ 242, 104, 30 },
		{ 242, 105, 30 },
		{ 242, 106, 30 },
		{ 242, 107, 30 },
		{ 242, 108, 30 },
		{ 242, 109, 30 },
		{ 242, 110, 30 },
		{ 242, 111, 30 },
		{ 242, 112, 30 },
		{ 242, 113, 30 },
		{ 242, 114, 30 },
		{ 242, 115, 30 },
		{ 242, 116, 30 },
		{ 242, 117, 30 },
		{ 242, 118, 30 },
		{ 242, 119, 30 },
		{ 242, 120, 30 },
		{ 242, 121, 30 },
		{ 242, 122, 30 },
		{ 242, 123, 30 },
		{ 242, 124, 30 },
		{ 242, 125, 30 },
		{ 242, 126, 30 },
		{ 242, 127, 30 },
		{ 242, 128, 30 },
		{ 242, 129, 30 },
		{ 242, 130, 30 },
		{ 242, 131, 30 },
		{ 242, 132, 30 },
		{ 242, 133, 30 },
		{ 242, 134, 30 },
		{ 242, 135, 30 },
		{ 242, 136, 30 },
		{ 242, 378, 30 },
		{ 242, 379, 30 },
		{ 242, 380, 30 },
		{ 242, 381, 30 },
		{ 242, 382, 30 },
		{ 242, 383, 30 },
		{ 242, 384, 30 },
		{ 242, 385, 30 },
		{ 242, 386, 30 },
		{ 242, 387, 30 },
		{ 242, 388, 30 },
		{ 242, 389, 30 },
		{ 242, 390, 30 },
		{ 242, 391, 30 },
		{ 242, 392, 30 },
		{ 242, 393, 30 },
		{ 242, 394, 30 },
		{ 242, 395, 30 },
		{ 242, 396, 30 },
		{ 242, 397, 30 },
		{ 242, 398, 30 },
		{ 242, 399, 30 },
		{ 242, 400, 30 },
		{ 242, 401, 30 },
		{ 242, 402, 30 },
		{ 242, 403, 30 },
		{ 242, 404, 30 },
		{ 242, 405, 30 },
		{ 242, 406, 30 },
		{ 242, 407, 30 },
		{ 242, 408, 30 },
		{ 242, 409, 30 },
		{ 242, 410, 30 },
		{ 242, 411, 30 },
		{ 242, 412, 30 },
		{ 242, 413, 30 },
		{ 242, 414, 30 },
		{ 242, 415, 30 },
		{ 242, 416, 30 },
		{ 242, 417, 30 },
		{ 242, 418, 30 },
		{ 242, 419, 30 },
		{ 242, 420, 30 },
		{ 242, 421, 30 },
		{ 242, 422, 30 },
		{ 242, 423, 30 },
		{ 242, 424, 30 },
		{ 242, 425, 30 },
		{ 242, 426, 30 },
		{ 242, 427, 30 },
		{ 242, 428, 30 },
		{ 242, 429, 30 },
		{ 242, 430, 30 },
		{ 242, 431, 30 },
		{ 242, 432, 30 },
		{ 242, 433, 30 },
		{ 242, 434, 30 },
		{ 242, 435, 30 },
		{ 242, 436, 30 },
		{ 242, 437, 30 },
		{ 242, 438, 30 },
		{ 242, 439, 30 },
		{ 242, 440, 30 },
		{ 242, 441, 30 },
		{ 242, 442, 30 },
		{ 242, 443, 30 },
		{ 242, 444, 30 },
		{ 242, 445, 30 },
		{ 242, 446, 30 },
		{ 242, 447, 30 },
		{ 242, 448, 30 },
		{ 242, 449, 30 },
		{ 242, 450, 30 },
		{ 242, 451, 30 },
		{ 242, 452, 30 },
		{ 242, 453, 30 },
		{ 242, 454, 30 },
		{ 242, 455, 30 },
		{ 242, 456, 30 },
		{ 242, 457, 30 },
		{ 242, 458, 30 },
		{ 242, 459, 30 },
		{ 242, 460, 30 },
		{ 242, 461, 30 },
		{ 242, 462, 30 },
		{ 242, 463, 30 },
		{ 242, 464, 30 },
		{ 242, 465, 30 },
		{ 242, 466, 30 },
		{ 242, 467, 30 },
		{ 242, 468, 30 },
		{ 242, 469, 30 },
		{ 242, 470, 30 },
		{ 242, 471, 30 },
		{ 242, 472, 30 },
		{ 242, 473, 30 },
		{ 242, 474, 30 },
		{ 242, 475, 30 },
		{ 242, 476, 30 },
		{ 242, 477, 30 },
		{ 242, 478, 30 },
		{ 242, 479, 30 },
		{ 242, 480, 30 },
		{ 242, 481, 30 },
		{ 242, 482, 30 },
		{ 242, 483, 30 },
		{ 242, 484, 30 },
		{ 242, 485, 30 },
		{ 242, 486, 30 },
		{ 242, 487, 30 },
		{ 242, 488, 30 },
		{ 242, 489, 30 },
		{ 242, 490, 30 },
		{ 242, 491, 30 },
		{ 242, 492, 30 },
		{ 242, 493, 30 },
		{ 242, 494, 30 },
		{ 242, 495, 30 },
		{ 242, 496, 30 },
		{ 242, 497, 30 },
		{ 242, 498, 30 },
		{ 242, 499, 30 },
		{ 242, 500, 30 },
		{ 242, 501, 30 },
		{ 242, 502, 30 },
		{ 242, 503, 30 },
		{ 242, 504, 30 },
		{ 242, 505, 30 },
		{ 242, 506, 30 },
		{ 242, 507, 30 },
		{ 242, 508, 30 },
		{ 242, 509, 30 },
		{ 242, 510, 30 },
		{ 242, 511, 30 },
		{ 242, 512, 30 },
		{ 243, 1, 30 },
		{ 243, 2, 30 },
		{ 243, 3, 30 },
		{ 243, 4, 30 },
		{ 243, 5, 30 },
		{ 243, 6, 30 },
		{ 243, 7, 30 },
		{ 243, 8, 30 },
		{ 243, 9, 30 },
		{ 243, 10, 30 },
		{ 243, 11, 30 },
		{ 243, 12, 30 },
		{ 243, 13, 30 },
		{ 243, 14, 30 },
		{ 243, 15, 30 },
		{ 243, 16, 30 },
		{ 243, 17, 30 },
		{ 243, 18, 30 },
		{ 243, 19, 30 },
		{ 243, 20, 30 },
		{ 243, 21, 30 },
		{ 243, 22, 30 },
		{ 243, 23, 30 },
		{ 243, 24, 30 },
		{ 243, 25, 30 },
		{ 243, 26, 30 },
		{ 243, 27, 30 },
		{ 243, 28, 30 },
		{ 243, 29, 30 },
		{ 243, 30, 30 },
		{ 243, 31, 30 },
		{ 243, 32, 30 },
		{ 243, 33, 30 },
		{ 243, 34, 30 },
		{ 243, 35, 30 },
		{ 243, 36, 30 },
		{ 243, 37, 30 },
		{ 243, 38, 30 },
		{ 243, 39, 30 },
		{ 243, 40, 30 },
		{ 243, 41, 30 },
		{ 243, 42, 30 },
		{ 243, 43, 30 },
		{ 243, 44, 30 },
		{ 243, 45, 30 },
		{ 243, 46, 30 },
		{ 243, 47, 30 },
		{ 243, 48, 30 },
		{ 243, 49, 30 },
		{ 243, 50, 30 },
		{ 243, 51, 30 },
		{ 243, 52, 30 },
		{ 243, 53, 30 },
		{ 243, 54, 30 },
		{ 243, 55, 30 },
		{ 243, 56, 30 },
		{ 243, 57, 30 },
		{ 243, 58, 30 },
		{ 243, 59, 30 },
		{ 243, 60, 30 },
		{ 243, 61, 30 },
		{ 243, 62, 30 },
		{ 243, 63, 30 },
		{ 243, 64, 30 },
		{ 243, 65, 30 },
		{ 243, 66, 30 },
		{ 243, 67, 30 },
		{ 243, 68, 30 },
		{ 243, 69, 30 },
		{ 243, 70, 30 },
		{ 243, 71, 30 },
		{ 243, 72, 30 },
		{ 243, 73, 30 },
		{ 243, 74, 30 },
		{ 243, 75, 30 },
		{ 243, 76, 30 },
		{ 243, 77, 30 },
		{ 243, 78, 30 },
		{ 243, 79, 30 },
		{ 243, 80, 30 },
		{ 243, 81, 30 },
		{ 243, 82, 30 },
		{ 243, 83, 30 },
		{ 243, 84, 30 },
		{ 243, 85, 30 },
		{ 243, 86, 30 },
		{ 243, 87, 30 },
		{ 243, 88, 30 },
		{ 243, 89, 30 },
		{ 243, 90, 30 },
		{ 243, 91, 30 },
		{ 243, 92, 30 },
		{ 243, 93, 30 },
		{ 243, 94, 30 },
		{ 243, 95, 30 },
		{ 243, 96, 30 },
		{ 243, 97, 30 },
		{ 243, 98, 30 },
		{ 243, 99, 30 },
		{ 243, 100, 30 },
		{ 243, 101, 30 },
		{ 243, 102, 30 },
		{ 243, 103, 30 },
		{ 243, 104, 30 },
		{ 243, 105, 30 },
		{ 243, 106, 30 },
		{ 243, 107, 30 },
		{ 243, 108, 30 },
		{ 243, 109, 30 },
		{ 243, 110, 30 },
		{ 243, 111, 30 },
		{ 243, 112, 30 },
		{ 243, 113, 30 },
		{ 243, 114, 30 },
		{ 243, 115, 30 },
		{ 243, 116, 30 },
		{ 243, 117, 30 },
		{ 243, 118, 30 },
		{ 243, 119, 30 },
		{ 243, 120, 30 },
		{ 243, 121, 30 },
		{ 243, 122, 30 },
		{ 243, 123, 30 },
		{ 243, 124, 30 },
		{ 243, 125, 30 },
		{ 243, 126, 30 },
		{ 243, 127, 30 },
		{ 243, 128, 30 },
		{ 243, 129, 30 },
		{ 243, 130, 30 },
		{ 243, 131, 30 },
		{ 243, 132, 30 },
		{ 243, 133, 30 },
		{ 243, 134, 30 },
		{ 243, 135, 30 },
		{ 243, 136, 30 },
		{ 243, 137, 30 },
		{ 243, 138, 30 },
		{ 243, 139, 30 },
		{ 243, 140, 30 },
		{ 243, 141, 30 },
		{ 243, 142, 30 },
		{ 243, 143, 30 },
		{ 243, 144, 30 },
		{ 243, 370, 30 },
		{ 243, 371, 30 },
		{ 243, 372, 30 },
		{ 243, 373, 30 },
		{ 243, 374, 30 },
		{ 243, 375, 30 },
		{ 243, 376, 30 },
		{ 243, 377, 30 },
		{ 243, 378, 30 },
		{ 243, 379, 30 },
		{ 243, 380, 30 },
		{ 243, 381, 30 },
		{ 243, 382, 30 },
		{ 243, 383, 30 },
		{ 243, 384, 30 },
		{ 243, 385, 30 },
		{ 243, 386, 30 },
		{ 243, 387, 30 },
		{ 243, 388, 30 },
		{ 243, 389, 30 },
		{ 243, 390, 30 },
		{ 243, 391, 30 },
		{ 243, 392, 30 },
		{ 243, 393, 30 },
		{ 243, 394, 30 },
		{ 243, 395, 30 },
		{ 243, 396, 30 },
		{ 243, 397, 30 },
		{ 243, 398, 30 },
		{ 243, 399, 30 },
		{ 243, 400, 30 },
		{ 243, 401, 30 },
		{ 243, 402, 30 },
		{ 243, 403, 30 },
		{ 243, 404, 30 },
		{ 243, 405, 30 },
		{ 243, 406, 30 },
		{ 243, 407, 30 },
		{ 243, 408, 30 },
		{ 243, 409, 30 },
		{ 243, 410, 30 },
		{ 243, 411, 30 },
		{ 243, 412, 30 },
		{ 243, 413, 30 },
		{ 243, 414, 30 },
		{ 243, 415, 30 },
		{ 243, 416, 30 },
		{ 243, 417, 30 },
		{ 243, 418, 30 },
		{ 243, 419, 30 },
		{ 243, 420, 30 },
		{ 243, 421, 30 },
		{ 243, 422, 30 },
		{ 243, 423, 30 },
		{ 243, 424, 30 },
		{ 243, 425, 30 },
		{ 243, 426, 30 },
		{ 243, 427, 30 },
		{ 243, 428, 30 },
		{ 243, 429, 30 },
		{ 243, 430, 30 },
		{ 243, 431, 30 },
		{ 243, 432, 30 },
		{ 243, 433, 30 },
		{ 243, 434, 30 },
		{ 243, 435, 30 },
		{ 243, 436, 30 },
		{ 243, 437, 30 },
		{ 243, 438, 30 },
		{ 243, 439, 30 },
		{ 243, 440, 30 },
		{ 243, 441, 30 },
		{ 243, 442, 30 },
		{ 243, 443, 30 },
		{ 243, 444, 30 },
		{ 243, 445, 30 },
		{ 243, 446, 30 },
		{ 243, 447, 30 },
		{ 243, 448, 30 },
		{ 243, 449, 30 },
		{ 243, 450, 30 },
		{ 243, 451, 30 },
		{ 243, 452, 30 },
		{ 243, 453, 30 },
		{ 243, 454, 30 },
		{ 243, 455, 30 },
		{ 243, 456, 30 },
		{ 243, 457, 30 },
		{ 243, 458, 30 },
		{ 243, 459, 30 },
		{ 243, 460, 30 },
		{ 243, 461, 30 },
		{ 243, 462, 30 },
		{ 243, 463, 30 },
		{ 243, 464, 30 },
		{ 243, 465, 30 },
		{ 243, 466, 30 },
		{ 243, 467, 30 },
		{ 243, 468, 30 },
		{ 243, 469, 30 },
		{ 243, 470, 30 },
		{ 243, 471, 30 },
		{ 243, 472, 30 },
		{ 243, 473, 30 },
		{ 243, 474, 30 },
		{ 243, 475, 30 },
		{ 243, 476, 30 },
		{ 243, 477, 30 },
		{ 243, 478, 30 },
		{ 243, 479, 30 },
		{ 243, 480, 30 },
		{ 243, 481, 30 },
		{ 243, 482, 30 },
		{ 243, 483, 30 },
		{ 243, 484, 30 },
		{ 243, 485, 30 },
		{ 243, 486, 30 },
		{ 243, 487, 30 },
		{ 243, 488, 30 },
		{ 243, 489, 30 },
		{ 243, 490, 30 },
		{ 243, 491, 30 },
		{ 243, 492, 30 },
		{ 243, 493, 30 },
		{ 243, 494, 30 },
		{ 243, 495, 30 },
		{ 243, 496, 30 },
		{ 243, 497, 30 },
		{ 243, 498, 30 },
		{ 243, 499, 30 },
		{ 243, 500, 30 },
		{ 243, 501, 30 },
		{ 243, 502, 30 },
		{ 243, 503, 30 },
		{ 243, 504, 30 },
		{ 243, 505, 30 },
		{ 243, 506, 30 },
		{ 243, 507, 30 },
		{ 243, 508, 30 },
		{ 243, 509, 30 },
		{ 243, 510, 30 },
		{ 243, 511, 30 },
		{ 243, 512, 30 },
		{ 244, 1, 30 },
		{ 244, 2, 30 },
		{ 244, 3, 30 },
		{ 244, 4, 30 },
		{ 244, 5, 30 },
		{ 244, 6, 30 },
		{ 244, 7, 30 },
		{ 244, 8, 30 },
		{ 244, 9, 30 },
		{ 244, 10, 30 },
		{ 244, 11, 30 },
		{ 244, 12, 30 },
		{ 244, 13, 30 },
		{ 244, 14, 30 },
		{ 244, 15, 30 },
		{ 244, 16, 30 },
		{ 244, 17, 30 },
		{ 244, 18, 30 },
		{ 244, 19, 30 },
		{ 244, 20, 30 },
		{ 244, 21, 30 },
		{ 244, 22, 30 },
		{ 244, 23, 30 },
		{ 244, 24, 30 },
		{ 244, 25, 30 },
		{ 244, 26, 30 },
		{ 244, 27, 30 },
		{ 244, 28, 30 },
		{ 244, 29, 30 },
		{ 244, 30, 30 },
		{ 244, 31, 30 },
		{ 244, 32, 30 },
		{ 244, 33, 30 },
		{ 244, 34, 30 },
		{ 244, 35, 30 },
		{ 244, 36, 30 },
		{ 244, 37, 30 },
		{ 244, 38, 30 },
		{ 244, 39, 30 },
		{ 244, 40, 30 },
		{ 244, 41, 30 },
		{ 244, 42, 30 },
		{ 244, 43, 30 },
		{ 244, 44, 30 },
		{ 244, 45, 30 },
		{ 244, 46, 30 },
		{ 244, 47, 30 },
		{ 244, 48, 30 },
		{ 244, 49, 30 },
		{ 244, 50, 30 },
		{ 244, 51, 30 },
		{ 244, 52, 30 },
		{ 244, 53, 30 },
		{ 244, 54, 30 },
		{ 244, 55, 30 },
		{ 244, 56, 30 },
		{ 244, 57, 30 },
		{ 244, 58, 30 },
		{ 244, 59, 30 },
		{ 244, 60, 30 },
		{ 244, 61, 30 },
		{ 244, 62, 30 },
		{ 244, 63, 30 },
		{ 244, 64, 30 },
		{ 244, 65, 30 },
		{ 244, 66, 30 },
		{ 244, 67, 30 },
		{ 244, 68, 30 },
		{ 244, 69, 30 },
		{ 244, 70, 30 },
		{ 244, 71, 30 },
		{ 244, 72, 30 },
		{ 244, 73, 30 },
		{ 244, 74, 30 },
		{ 244, 75, 30 },
		{ 244, 76, 30 },
		{ 244, 77, 30 },
		{ 244, 78, 30 },
		{ 244, 79, 30 },
		{ 244, 80, 30 },
		{ 244, 81, 30 },
		{ 244, 82, 30 },
		{ 244, 83, 30 },
		{ 244, 84, 30 },
		{ 244, 85, 30 },
		{ 244, 86, 30 },
		{ 244, 87, 30 },
		{ 244, 88, 30 },
		{ 244, 89, 30 },
		{ 244, 90, 30 },
		{ 244, 91, 30 },
		{ 244, 92, 30 },
		{ 244, 93, 30 },
		{ 244, 94, 30 },
		{ 244, 95, 30 },
		{ 244, 96, 30 },
		{ 244, 97, 30 },
		{ 244, 98, 30 },
		{ 244, 99, 30 },
		{ 244, 100, 30 },
		{ 244, 101, 30 },
		{ 244, 102, 30 },
		{ 244, 103, 30 },
		{ 244, 104, 30 },
		{ 244, 105, 30 },
		{ 244, 106, 30 },
		{ 244, 107, 30 },
		{ 244, 108, 30 },
		{ 244, 109, 30 },
		{ 244, 110, 30 },
		{ 244, 111, 30 },
		{ 244, 112, 30 },
		{ 244, 113, 30 },
		{ 244, 114, 30 },
		{ 244, 115, 30 },
		{ 244, 116, 30 },
		{ 244, 117, 30 },
		{ 244, 118, 30 },
		{ 244, 119, 30 },
		{ 244, 120, 30 },
		{ 244, 121, 30 },
		{ 244, 122, 30 },
		{ 244, 123, 30 },
		{ 244, 124, 30 },
		{ 244, 125, 30 },
		{ 244, 126, 30 },
		{ 244, 127, 30 },
		{ 244, 128, 30 },
		{ 244, 129, 30 },
		{ 244, 130, 30 },
		{ 244, 131, 30 },
		{ 244, 132, 30 },
		{ 244, 133, 30 },
		{ 244, 134, 30 },
		{ 244, 135, 30 },
		{ 244, 136, 30 },
		{ 244, 137, 30 },
		{ 244, 138, 30 },
		{ 244, 139, 30 },
		{ 244, 140, 30 },
		{ 244, 141, 30 },
		{ 244, 142, 30 },
		{ 244, 143, 30 },
		{ 244, 144, 30 },
		{ 244, 145, 30 },
		{ 244, 146, 30 },
		{ 244, 147, 30 },
		{ 244, 148, 30 },
		{ 244, 149, 30 },
		{ 244, 150, 30 },
		{ 244, 151, 30 },
		{ 244, 152, 30 },
		{ 244, 362, 30 },
		{ 244, 363, 30 },
		{ 244, 364, 30 },
		{ 244, 365, 30 },
		{ 244, 366, 30 },
		{ 244, 367, 30 },
		{ 244, 368, 30 },
		{ 244, 369, 30 },
		{ 244, 370, 30 },
		{ 244, 371, 30 },
		{ 244, 372, 30 },
		{ 244, 373, 30 },
		{ 244, 374, 30 },
		{ 244, 375, 30 },
		{ 244, 376, 30 },
		{ 244, 377, 30 },
		{ 244, 378, 30 },
		{ 244, 379, 30 },
		{ 244, 380, 30 },
		{ 244, 381, 30 },
		{ 244, 382, 30 },
		{ 244, 383, 30 },
		{ 244, 384, 30 },
		{ 244, 385, 30 },
		{ 244, 386, 30 },
		{ 244, 387, 30 },
		{ 244, 388, 30 },
		{ 244, 389, 30 },
		{ 244, 390, 30 },
		{ 244, 391, 30 },
		{ 244, 392, 30 },
		{ 244, 393, 30 },
		{ 244, 394, 30 },
		{ 244, 395, 30 },
		{ 244, 396, 30 },
		{ 244, 397, 30 },
		{ 244, 398, 30 },
		{ 244, 399, 30 },
		{ 244, 400, 30 },
		{ 244, 401, 30 },
		{ 244, 402, 30 },
		{ 244, 403, 30 },
		{ 244, 404, 30 },
		{ 244, 405, 30 },
		{ 244, 406, 30 },
		{ 244, 407, 30 },
		{ 244, 408, 30 },
		{ 244, 409, 30 },
		{ 244, 410, 30 },
		{ 244, 411, 30 },
		{ 244, 412, 30 },
		{ 244, 413, 30 },
		{ 244, 414, 30 },
		{ 244, 415, 30 },
		{ 244, 416, 30 },
		{ 244, 417, 30 },
		{ 244, 418, 30 },
		{ 244, 419, 30 },
		{ 244, 420, 30 },
		{ 244, 421, 30 },
		{ 244, 422, 30 },
		{ 244, 423, 30 },
		{ 244, 424, 30 },
		{ 244, 425, 30 },
		{ 244, 426, 30 },
		{ 244, 427, 30 },
		{ 244, 428, 30 },
		{ 244, 429, 30 },
		{ 244, 430, 30 },
		{ 244, 431, 30 },
		{ 244, 432, 30 },
		{ 244, 433, 30 },
		{ 244, 434, 30 },
		{ 244, 435, 30 },
		{ 244, 436, 30 },
		{ 244, 437, 30 },
		{ 244, 438, 30 },
		{ 244, 439, 30 },
		{ 244, 440, 30 },
		{ 244, 441, 30 },
		{ 244, 442, 30 },
		{ 244, 443, 30 },
		{ 244, 444, 30 },
		{ 244, 445, 30 },
		{ 244, 446, 30 },
		{ 244, 447, 30 },
		{ 244, 448, 30 },
		{ 244, 449, 30 },
		{ 244, 450, 30 },
		{ 244, 451, 30 },
		{ 244, 452, 30 },
		{ 244, 453, 30 },
		{ 244, 454, 30 },
		{ 244, 455, 30 },
		{ 244, 456, 30 },
		{ 244, 457, 30 },
		{ 244, 458, 30 },
		{ 244, 459, 30 },
		{ 244, 460, 30 },
		{ 244, 461, 30 },
		{ 244, 462, 30 },
		{ 244, 463, 30 },
		{ 244, 464, 30 },
		{ 244, 465, 30 },
		{ 244, 466, 30 },
		{ 244, 467, 30 },
		{ 244, 468, 30 },
		{ 244, 469, 30 },
		{ 244, 470, 30 },
		{ 244, 471, 30 },
		{ 244, 472, 30 },
		{ 244, 473, 30 },
		{ 244, 474, 30 },
		{ 244, 475, 30 },
		{ 244, 476, 30 },
		{ 244, 477, 30 },
		{ 244, 478, 30 },
		{ 244, 479, 30 },
		{ 244, 480, 30 },
		{ 244, 481, 30 },
		{ 244, 482, 30 },
		{ 244, 483, 30 },
		{ 244, 484, 30 },
		{ 244, 485, 30 },
		{ 244, 486, 30 },
		{ 244, 487, 30 },
		{ 244, 488, 30 },
		{ 244, 489, 30 },
		{ 244, 490, 30 },
		{ 244, 491, 30 },
		{ 244, 492, 30 },
		{ 244, 493, 30 },
		{ 244, 494, 30 },
		{ 244, 495, 30 },
		{ 244, 496, 30 },
		{ 244, 497, 30 },
		{ 244, 498, 30 },
		{ 244, 499, 30 },
		{ 244, 500, 30 },
		{ 244, 501, 30 },
		{ 244, 502, 30 },
		{ 244, 503, 30 },
		{ 244, 504, 30 },
		{ 244, 505, 30 },
		{ 244, 506, 30 },
		{ 244, 507, 30 },
		{ 244, 508, 30 },
		{ 244, 509, 30 },
		{ 244, 510, 30 },
		{ 244, 511, 30 },
		{ 244, 512, 30 },
		{ 245, 1, 30 },
		{ 245, 2, 30 },
		{ 245, 3, 30 },
		{ 245, 4, 30 },
		{ 245, 5, 30 },
		{ 245, 6, 30 },
		{ 245, 7, 30 },
		{ 245, 8, 30 },
		{ 245, 9, 30 },
		{ 245, 10, 30 },
		{ 245, 11, 30 },
		{ 245, 12, 30 },
		{ 245, 13, 30 },
		{ 245, 14, 30 },
		{ 245, 15, 30 },
		{ 245, 16, 30 },
		{ 245, 17, 30 },
		{ 245, 18, 30 },
		{ 245, 19, 30 },
		{ 245, 20, 30 },
		{ 245, 21, 30 },
		{ 245, 22, 30 },
		{ 245, 23, 30 },
		{ 245, 24, 30 },
		{ 245, 25, 30 },
		{ 245, 26, 30 },
		{ 245, 27, 30 },
		{ 245, 28, 30 },
		{ 245, 29, 30 },
		{ 245, 30, 30 },
		{ 245, 31, 30 },
		{ 245, 32, 30 },
		{ 245, 33, 30 },
		{ 245, 34, 30 },
		{ 245, 35, 30 },
		{ 245, 36, 30 },
		{ 245, 37, 30 },
		{ 245, 38, 30 },
		{ 245, 39, 30 },
		{ 245, 40, 30 },
		{ 245, 41, 30 },
		{ 245, 42, 30 },
		{ 245, 43, 30 },
		{ 245, 44, 30 },
		{ 245, 45, 30 },
		{ 245, 46, 30 },
		{ 245, 47, 30 },
		{ 245, 48, 30 },
		{ 245, 49, 30 },
		{ 245, 50, 30 },
		{ 245, 51, 30 },
		{ 245, 52, 30 },
		{ 245, 53, 30 },
		{ 245, 54, 30 },
		{ 245, 55, 30 },
		{ 245, 56, 30 },
		{ 245, 57, 30 },
		{ 245, 58, 30 },
		{ 245, 59, 30 },
		{ 245, 60, 30 },
		{ 245, 61, 30 },
		{ 245, 62, 30 },
		{ 245, 63, 30 },
		{ 245, 64, 30 },
		{ 245, 65, 30 },
		{ 245, 66, 30 },
		{ 245, 67, 30 },
		{ 245, 68, 30 },
		{ 245, 69, 30 },
		{ 245, 70, 30 },
		{ 245, 71, 30 },
		{ 245, 72, 30 },
		{ 245, 73, 30 },
		{ 245, 74, 30 },
		{ 245, 75, 30 },
		{ 245, 76, 30 },
		{ 245, 77, 30 },
		{ 245, 78, 30 },
		{ 245, 79, 30 },
		{ 245, 80, 30 },
		{ 245, 81, 30 },
		{ 245, 82, 30 },
		{ 245, 83, 30 },
		{ 245, 84, 30 },
		{ 245, 85, 30 },
		{ 245, 86, 30 },
		{ 245, 87, 30 },
		{ 245, 88, 30 },
		{ 245, 89, 30 },
		{ 245, 90, 30 },
		{ 245, 91, 30 },
		{ 245, 92, 30 },
		{ 245, 93, 30 },
		{ 245, 94, 30 },
		{ 245, 95, 30 },
		{ 245, 96, 30 },
		{ 245, 97, 30 },
		{ 245, 98, 30 },
		{ 245, 99, 30 },
		{ 245, 100, 30 },
		{ 245, 101, 30 },
		{ 245, 102, 30 },
		{ 245, 103, 30 },
		{ 245, 104, 30 },
		{ 245, 105, 30 },
		{ 245, 106, 30 },
		{ 245, 107, 30 },
		{ 245, 108, 30 },
		{ 245, 109, 30 },
		{ 245, 110, 30 },
		{ 245, 111, 30 },
		{ 245, 112, 30 },
		{ 245, 113, 30 },
		{ 245, 114, 30 },
		{ 245, 115, 30 },
		{ 245, 116, 30 },
		{ 245, 117, 30 },
		{ 245, 118, 30 },
		{ 245, 119, 30 },
		{ 245, 120, 30 },
		{ 245, 121, 30 },
		{ 245, 122, 30 },
		{ 245, 123, 30 },
		{ 245, 124, 30 },
		{ 245, 125, 30 },
		{ 245, 126, 30 },
		{ 245, 127, 30 },
		{ 245, 128, 30 },
		{ 245, 129, 30 },
		{ 245, 130, 30 },
		{ 245, 131, 30 },
		{ 245, 132, 30 },
		{ 245, 133, 30 },
		{ 245, 134, 30 },
		{ 245, 135, 30 },
		{ 245, 136, 30 },
		{ 245, 137, 30 },
		{ 245, 138, 30 },
		{ 245, 139, 30 },
		{ 245, 140, 30 },
		{ 245, 141, 30 },
		{ 245, 142, 30 },
		{ 245, 143, 30 },
		{ 245, 144, 30 },
		{ 245, 145, 30 },
		{ 245, 146, 30 },
		{ 245, 147, 30 },
		{ 245, 148, 30 },
		{ 245, 149, 30 },
		{ 245, 150, 30 },
		{ 245, 151, 30 },
		{ 245, 152, 30 },
		{ 245, 153, 30 },
		{ 245, 154, 30 },
		{ 245, 155, 30 },
		{ 245, 156, 30 },
		{ 245, 157, 30 },
		{ 245, 158, 30 },
		{ 245, 159, 30 },
		{ 245, 160, 30 },
		{ 245, 354, 30 },
		{ 245, 355, 30 },
		{ 245, 356, 30 },
		{ 245, 357, 30 },
		{ 245, 358, 30 },
		{ 245, 359, 30 },
		{ 245, 360, 30 },
		{ 245, 361, 30 },
		{ 245, 362, 30 },
		{ 245, 363, 30 },
		{ 245, 364, 30 },
		{ 245, 365, 30 },
		{ 245, 366, 30 },
		{ 245, 367, 30 },
		{ 245, 368, 30 },
		{ 245, 369, 30 },
		{ 245, 370, 30 },
		{ 245, 371, 30 },
		{ 245, 372, 30 },
		{ 245, 373, 30 },
		{ 245, 374, 30 },
		{ 245, 375, 30 },
		{ 245, 376, 30 },
		{ 245, 377, 30 },
		{ 245, 378, 30 },
		{ 245, 379, 30 },
		{ 245, 380, 30 },
		{ 245, 381, 30 },
		{ 245, 382, 30 },
		{ 245, 383, 30 },
		{ 245, 384, 30 },
		{ 245, 385, 30 },
		{ 245, 386, 30 },
		{ 245, 387, 30 },
		{ 245, 388, 30 },
		{ 245, 389, 30 },
		{ 245, 390, 30 },
		{ 245, 391, 30 },
		{ 245, 392, 30 },
		{ 245, 393, 30 },
		{ 245, 394, 30 },
		{ 245, 395, 30 },
		{ 245, 396, 30 },
		{ 245, 397, 30 },
		{ 245, 398, 30 },
		{ 245, 399, 30 },
		{ 245, 400, 30 },
		{ 245, 401, 30 },
		{ 245, 402, 30 },
		{ 245, 403, 30 },
		{ 245, 404, 30 },
		{ 245, 405, 30 },
		{ 245, 406, 30 },
		{ 245, 407, 30 },
		{ 245, 408, 30 },
		{ 245, 409, 30 },
		{ 245, 410, 30 },
		{ 245, 411, 30 },
		{ 245, 412, 30 },
		{ 245, 413, 30 },
		{ 245, 414, 30 },
		{ 245, 415, 30 },
		{ 245, 416, 30 },
		{ 245, 417, 30 },
		{ 245, 418, 30 },
		{ 245, 419, 30 },
		{ 245, 420, 30 },
		{ 245, 421, 30 },
		{ 245, 422, 30 },
		{ 245, 423, 30 },
		{ 245, 424, 30 },
		{ 245, 425, 30 },
		{ 245, 426, 30 },
		{ 245, 427, 30 },
		{ 245, 428, 30 },
		{ 245, 429, 30 },
		{ 245, 430, 30 },
		{ 245, 431, 30 },
		{ 245, 432, 30 },
		{ 245, 433, 30 },
		{ 245, 434, 30 },
		{ 245, 435, 30 },
		{ 245, 436, 30 },
		{ 245, 437, 30 },
		{ 245, 438, 30 },
		{ 245, 439, 30 },
		{ 245, 440, 30 },
		{ 245, 441, 30 },
		{ 245, 442, 30 },
		{ 245, 443, 30 },
		{ 245, 444, 30 },
		{ 245, 445, 30 },
		{ 245, 446, 30 },
		{ 245, 447, 30 },
		{ 245, 448, 30 },
		{ 245, 449, 30 },
		{ 245, 450, 30 },
		{ 245, 451, 30 },
		{ 245, 452, 30 },
		{ 245, 453, 30 },
		{ 245, 454, 30 },
		{ 245, 455, 30 },
		{ 245, 456, 30 },
		{ 245, 457, 30 },
		{ 245, 458, 30 },
		{ 245, 459, 30 },
		{ 245, 460, 30 },
		{ 245, 461, 30 },
		{ 245, 462, 30 },
		{ 245, 463, 30 },
		{ 245, 464, 30 },
		{ 245, 465, 30 },
		{ 245, 466, 30 },
		{ 245, 467, 30 },
		{ 245, 468, 30 },
		{ 245, 469, 30 },
		{ 245, 470, 30 },
		{ 245, 471, 30 },
		{ 245, 472, 30 },
		{ 245, 473, 30 },
		{ 245, 474, 30 },
		{ 245, 475, 30 },
		{ 245, 476, 30 },
		{ 245, 477, 30 },
		{ 245, 478, 30 },
		{ 245, 479, 30 },
		{ 245, 480, 30 },
		{ 245, 481, 30 },
		{ 245, 482, 30 },
		{ 245, 483, 30 },
		{ 245, 484, 30 },
		{ 245, 485, 30 },
		{ 245, 486, 30 },
		{ 245, 487, 30 },
		{ 245, 488, 30 },
		{ 245, 489, 30 },
		{ 245, 490, 30 },
		{ 245, 491, 30 },
		{ 245, 492, 30 },
		{ 245, 493, 30 },
		{ 245, 494, 30 },
		{ 245, 495, 30 },
		{ 245, 496, 30 },
		{ 245, 497, 30 },
		{ 245, 498, 30 },
		{ 245, 499, 30 },
		{ 245, 500, 30 },
		{ 245, 501, 30 },
		{ 245, 502, 30 },
		{ 245, 503, 30 },
		{ 245, 504, 30 },
		{ 245, 505, 30 },
		{ 245, 506, 30 },
		{ 245, 507, 30 },
		{ 245, 508, 30 },
		{ 245, 509, 30 },
		{ 245, 510, 30 },
		{ 245, 511, 30 },
		{ 245, 512, 30 },
		{ 246, 1, 30 },
		{ 246, 2, 30 },
		{ 246, 3, 30 },
		{ 246, 4, 30 },
		{ 246, 5, 30 },
		{ 246, 6, 30 },
		{ 246, 7, 30 },
		{ 246, 8, 30 },
		{ 246, 9, 30 },
		{ 246, 10, 30 },
		{ 246, 11, 30 },
		{ 246, 12, 30 },
		{ 246, 13, 30 },
		{ 246, 14, 30 },
		{ 246, 15, 30 },
		{ 246, 16, 30 },
		{ 246, 17, 30 },
		{ 246, 18, 30 },
		{ 246, 19, 30 },
		{ 246, 20, 30 },
		{ 246, 21, 30 },
		{ 246, 22, 30 },
		{ 246, 23, 30 },
		{ 246, 24, 30 },
		{ 246, 25, 30 },
		{ 246, 26, 30 },
		{ 246, 27, 30 },
		{ 246, 28, 30 },
		{ 246, 29, 30 },
		{ 246, 30, 30 },
		{ 246, 31, 30 },
		{ 246, 32, 30 },
		{ 246, 33, 30 },
		{ 246, 34, 30 },
		{ 246, 35, 30 },
		{ 246, 36, 30 },
		{ 246, 37, 30 },
		{ 246, 38, 30 },
		{ 246, 39, 30 },
		{ 246, 40, 30 },
		{ 246, 41, 30 },
		{ 246, 42, 30 },
		{ 246, 43, 30 },
		{ 246, 44, 30 },
		{ 246, 45, 30 },
		{ 246, 46, 30 },
		{ 246, 47, 30 },
		{ 246, 48, 30 },
		{ 246, 49, 30 },
		{ 246, 50, 30 },
		{ 246, 51, 30 },
		{ 246, 52, 30 },
		{ 246, 53, 30 },
		{ 246, 54, 30 },
		{ 246, 55, 30 },
		{ 246, 56, 30 },
		{ 246, 57, 30 },
		{ 246, 58, 30 },
		{ 246, 59, 30 },
		{ 246, 60, 30 },
		{ 246, 61, 30 },
		{ 246, 62, 30 },
		{ 246, 63, 30 },
		{ 246, 64, 30 },
		{ 246, 65, 30 },
		{ 246, 66, 30 },
		{ 246, 67, 30 },
		{ 246, 68, 30 },
		{ 246, 69, 30 },
		{ 246, 70, 30 },
		{ 246, 71, 30 },
		{ 246, 72, 30 },
		{ 246, 73, 30 },
		{ 246, 74, 30 },
		{ 246, 75, 30 },
		{ 246, 76, 30 },
		{ 246, 77, 30 },
		{ 246, 78, 30 },
		{ 246, 79, 30 },
		{ 246, 80, 30 },
		{ 246, 81, 30 },
		{ 246, 82, 30 },
		{ 246, 83, 30 },
		{ 246, 84, 30 },
		{ 246, 85, 30 },
		{ 246, 86, 30 },
		{ 246, 87, 30 },
		{ 246, 88, 30 },
		{ 246, 89, 30 },
		{ 246, 90, 30 },
		{ 246, 91, 30 },
		{ 246, 92, 30 },
		{ 246, 93, 30 },
		{ 246, 94, 30 },
		{ 246, 95, 30 },
		{ 246, 96, 30 },
		{ 246, 97, 30 },
		{ 246, 98, 30 },
		{ 246, 99, 30 },
		{ 246, 100, 30 },
		{ 246, 101, 30 },
		{ 246, 102, 30 },
		{ 246, 103, 30 },
		{ 246, 104, 30 },
		{ 246, 105, 30 },
		{ 246, 106, 30 },
		{ 246, 107, 30 },
		{ 246, 108, 30 },
		{ 246, 109, 30 },
		{ 246, 110, 30 },
		{ 246, 111, 30 },
		{ 246, 112, 30 },
		{ 246, 113, 30 },
		{ 246, 114, 30 },
		{ 246, 115, 30 },
		{ 246, 116, 30 },
		{ 246, 117, 30 },
		{ 246, 118, 30 },
		{ 246, 119, 30 },
		{ 246, 120, 30 },
		{ 246, 121, 30 },
		{ 246, 122, 30 },
		{ 246, 123, 30 },
		{ 246, 124, 30 },
		{ 246, 125, 30 },
		{ 246, 126, 30 },
		{ 246, 127, 30 },
		{ 246, 128, 30 },
		{ 246, 129, 30 },
		{ 246, 130, 30 },
		{ 246, 131, 30 },
		{ 246, 132, 30 },
		{ 246, 133, 30 },
		{ 246, 134, 30 },
		{ 246, 135, 30 },
		{ 246, 136, 30 },
		{ 246, 137, 30 },
		{ 246, 138, 30 },
		{ 246, 139, 30 },
		{ 246, 140, 30 },
		{ 246, 141, 30 },
		{ 246, 142, 30 },
		{ 246, 143, 30 },
		{ 246, 144, 30 },
		{ 246, 145, 30 },
		{ 246, 146, 30 },
		{ 246, 147, 30 },
		{ 246, 148, 30 },
		{ 246, 149, 30 },
		{ 246, 150, 30 },
		{ 246, 151, 30 },
		{ 246, 152, 30 },
		{ 246, 153, 30 },
		{ 246, 154, 30 },
		{ 246, 155, 30 },
		{ 246, 156, 30 },
		{ 246, 157, 30 },
		{ 246, 158, 30 },
		{ 246, 159, 30 },
		{ 246, 160, 30 },
		{ 246, 161, 30 },
		{ 246, 162, 30 },
		{ 246, 163, 30 },
		{ 246, 164, 30 },
		{ 246, 165, 30 },
		{ 246, 166, 30 },
		{ 246, 167, 30 },
		{ 246, 168, 30 },
		{ 246, 346, 30 },
		{ 246, 347, 30 },
		{ 246, 348, 30 },
		{ 246, 349, 30 },
		{ 246, 350, 30 },
		{ 246, 351, 30 },
		{ 246, 352, 30 },
		{ 246, 353, 30 },
		{ 246, 354, 30 },
		{ 246, 355, 30 },
		{ 246, 356, 30 },
		{ 246, 357, 30 },
		{ 246, 358, 30 },
		{ 246, 359, 30 },
		{ 246, 360, 30 },
		{ 246, 361, 30 },
		{ 246, 362, 30 },
		{ 246, 363, 30 },
		{ 246, 364, 30 },
		{ 246, 365, 30 },
		{ 246, 366, 30 },
		{ 246, 367, 30 },
		{ 246, 368, 30 },
		{ 246, 369, 30 },
		{ 246, 370, 30 },
		{ 246, 371, 30 },
		{ 246, 372, 30 },
		{ 246, 373, 30 },
		{ 246, 374, 30 },
		{ 246, 375, 30 },
		{ 246, 376, 30 },
		{ 246, 377, 30 },
		{ 246, 378, 30 },
		{ 246, 379, 30 },
		{ 246, 380, 30 },
		{ 246, 381, 30 },
		{ 246, 382, 30 },
		{ 246, 383, 30 },
		{ 246, 384, 30 },
		{ 246, 385, 30 },
		{ 246, 386, 30 },
		{ 246, 387, 30 },
		{ 246, 388, 30 },
		{ 246, 389, 30 },
		{ 246, 390, 30 },
		{ 246, 391, 30 },
		{ 246, 392, 30 },
		{ 246, 393, 30 },
		{ 246, 394, 30 },
		{ 246, 395, 30 },
		{ 246, 396, 30 },
		{ 246, 397, 30 },
		{ 246, 398, 30 },
		{ 246, 399, 30 },
		{ 246, 400, 30 },
		{ 246, 401, 30 },
		{ 246, 402, 30 },
		{ 246, 403, 30 },
		{ 246, 404, 30 },
		{ 246, 405, 30 },
		{ 246, 406, 30 },
		{ 246, 407, 30 },
		{ 246, 408, 30 },
		{ 246, 409, 30 },
		{ 246, 410, 30 },
		{ 246, 411, 30 },
		{ 246, 412, 30 },
		{ 246, 413, 30 },
		{ 246, 414, 30 },
		{ 246, 415, 30 },
		{ 246, 416, 30 },
		{ 246, 417, 30 },
		{ 246, 418, 30 },
		{ 246, 419, 30 },
		{ 246, 420, 30 },
		{ 246, 421, 30 },
		{ 246, 422, 30 },
		{ 246, 423, 30 },
		{ 246, 424, 30 },
		{ 246, 425, 30 },
		{ 246, 426, 30 },
		{ 246, 427, 30 },
		{ 246, 428, 30 },
		{ 246, 429, 30 },
		{ 246, 430, 30 },
		{ 246, 431, 30 },
		{ 246, 432, 30 },
		{ 246, 433, 30 },
		{ 246, 434, 30 },
		{ 246, 435, 30 },
		{ 246, 436, 30 },
		{ 246, 437, 30 },
		{ 246, 438, 30 },
		{ 246, 439, 30 },
		{ 246, 440, 30 },
		{ 246, 441, 30 },
		{ 246, 442, 30 },
		{ 246, 443, 30 },
		{ 246, 444, 30 },
		{ 246, 445, 30 },
		{ 246, 446, 30 },
		{ 246, 447, 30 },
		{ 246, 448, 30 },
		{ 246, 449, 30 },
		{ 246, 450, 30 },
		{ 246, 451, 30 },
		{ 246, 452, 30 },
		{ 246, 453, 30 },
		{ 246, 454, 30 },
		{ 246, 455, 30 },
		{ 246, 456, 30 },
		{ 246, 457, 30 },
		{ 246, 458, 30 },
		{ 246, 459, 30 },
		{ 246, 460, 30 },
		{ 246, 461, 30 },
		{ 246, 462, 30 },
		{ 246, 463, 30 },
		{ 246, 464, 30 },
		{ 246, 465, 30 },
		{ 246, 466, 30 },
		{ 246, 467, 30 },
		{ 246, 468, 30 },
		{ 246, 469, 30 },
		{ 246, 470, 30 },
		{ 246, 471, 30 },
		{ 246, 472, 30 },
		{ 246, 473, 30 },
		{ 246, 474, 30 },
		{ 246, 475, 30 },
		{ 246, 476, 30 },
		{ 246, 477, 30 },
		{ 246, 478, 30 },
		{ 246, 479, 30 },
		{ 246, 480, 30 },
		{ 246, 481, 30 },
		{ 246, 482, 30 },
		{ 246, 483, 30 },
		{ 246, 484, 30 },
		{ 246, 485, 30 },
		{ 246, 486, 30 },
		{ 246, 487, 30 },
		{ 246, 488, 30 },
		{ 246, 489, 30 },
		{ 246, 490, 30 },
		{ 246, 491, 30 },
		{ 246, 492, 30 },
		{ 246, 493, 30 },
		{ 246, 494, 30 },
		{ 246, 495, 30 },
		{ 246, 496, 30 },
		{ 246, 497, 30 },
		{ 246, 498, 30 },
		{ 246, 499, 30 },
		{ 246, 500, 30 },
		{ 246, 501, 30 },
		{ 246, 502, 30 },
		{ 246, 503, 30 },
		{ 246, 504, 30 },
		{ 246, 505, 30 },
		{ 246, 506, 30 },
		{ 246, 507, 30 },
		{ 246, 508, 30 },
		{ 246, 509, 30 },
		{ 246, 510, 30 },
		{ 246, 511, 30 },
		{ 246, 512, 30 },
		{ 247, 1, 30 },
		{ 247, 2, 30 },
		{ 247, 3, 30 },
		{ 247, 4, 30 },
		{ 247, 5, 30 },
		{ 247, 6, 30 },
		{ 247, 7, 30 },
		{ 247, 8, 30 },
		{ 247, 9, 30 },
		{ 247, 10, 30 },
		{ 247, 11, 30 },
		{ 247, 12, 30 },
		{ 247, 13, 30 },
		{ 247, 14, 30 },
		{ 247, 15, 30 },
		{ 247, 16, 30 },
		{ 247, 17, 30 },
		{ 247, 18, 30 },
		{ 247, 19, 30 },
		{ 247, 20, 30 },
		{ 247, 21, 30 },
		{ 247, 22, 30 },
		{ 247, 23, 30 },
		{ 247, 24, 30 },
		{ 247, 25, 30 },
		{ 247, 26, 30 },
		{ 247, 27, 30 },
		{ 247, 28, 30 },
		{ 247, 29, 30 },
		{ 247, 30, 30 },
		{ 247, 31, 30 },
		{ 247, 32, 30 },
		{ 247, 33, 30 },
		{ 247, 34, 30 },
		{ 247, 35, 30 },
		{ 247, 36, 30 },
		{ 247, 37, 30 },
		{ 247, 38, 30 },
		{ 247, 39, 30 },
		{ 247, 40, 30 },
		{ 247, 41, 30 },
		{ 247, 42, 30 },
		{ 247, 43, 30 },
		{ 247, 44, 30 },
		{ 247, 45, 30 },
		{ 247, 46, 30 },
		{ 247, 47, 30 },
		{ 247, 48, 30 },
		{ 247, 49, 30 },
		{ 247, 50, 30 },
		{ 247, 51, 30 },
		{ 247, 52, 30 },
		{ 247, 53, 30 },
		{ 247, 54, 30 },
		{ 247, 55, 30 },
		{ 247, 56, 30 },
		{ 247, 57, 30 },
		{ 247, 58, 30 },
		{ 247, 59, 30 },
		{ 247, 60, 30 },
		{ 247, 61, 30 },
		{ 247, 62, 30 },
		{ 247, 63, 30 },
		{ 247, 64, 30 },
		{ 247, 65, 30 },
		{ 247, 66, 30 },
		{ 247, 67, 30 },
		{ 247, 68, 30 },
		{ 247, 69, 30 },
		{ 247, 70, 30 },
		{ 247, 71, 30 },
		{ 247, 72, 30 },
		{ 247, 73, 30 },
		{ 247, 74, 30 },
		{ 247, 75, 30 },
		{ 247, 76, 30 },
		{ 247, 77, 30 },
		{ 247, 78, 30 },
		{ 247, 79, 30 },
		{ 247, 80, 30 },
		{ 247, 81, 30 },
		{ 247, 82, 30 },
		{ 247, 83, 30 },
		{ 247, 84, 30 },
		{ 247, 85, 30 },
		{ 247, 86, 30 },
		{ 247, 87, 30 },
		{ 247, 88, 30 },
		{ 247, 89, 30 },
		{ 247, 90, 30 },
		{ 247, 91, 30 },
		{ 247, 92, 30 },
		{ 247, 93, 30 },
		{ 247, 94, 30 },
		{ 247, 95, 30 },
		{ 247, 96, 30 },
		{ 247, 97, 30 },
		{ 247, 98, 30 },
		{ 247, 99, 30 },
		{ 247, 100, 30 },
		{ 247, 101, 30 },
		{ 247, 102, 30 },
		{ 247, 103, 30 },
		{ 247, 104, 30 },
		{ 247, 105, 30 },
		{ 247, 106, 30 },
		{ 247, 107, 30 },
		{ 247, 108, 30 },
		{ 247, 109, 30 },
		{ 247, 110, 30 },
		{ 247, 111, 30 },
		{ 247, 112, 30 },
		{ 247, 113, 30 },
		{ 247, 114, 30 },
		{ 247, 115, 30 },
		{ 247, 116, 30 },
		{ 247, 117, 30 },
		{ 247, 118, 30 },
		{ 247, 119, 30 },
		{ 247, 120, 30 },
		{ 247, 121, 30 },
		{ 247, 122, 30 },
		{ 247, 123, 30 },
		{ 247, 124, 30 },
		{ 247, 125, 30 },
		{ 247, 126, 30 },
		{ 247, 127, 30 },
		{ 247, 128, 30 },
		{ 247, 129, 30 },
		{ 247, 130, 30 },
		{ 247, 131, 30 },
		{ 247, 132, 30 },
		{ 247, 133, 30 },
		{ 247, 134, 30 },
		{ 247, 135, 30 },
		{ 247, 136, 30 },
		{ 247, 137, 30 },
		{ 247, 138, 30 },
		{ 247, 139, 30 },
		{ 247, 140, 30 },
		{ 247, 141, 30 },
		{ 247, 142, 30 },
		{ 247, 143, 30 },
		{ 247, 144, 30 },
		{ 247, 145, 30 },
		{ 247, 146, 30 },
		{ 247, 147, 30 },
		{ 247, 148, 30 },
		{ 247, 149, 30 },
		{ 247, 150, 30 },
		{ 247, 151, 30 },
		{ 247, 152, 30 },
		{ 247, 153, 30 },
		{ 247, 154, 30 },
		{ 247, 155, 30 },
		{ 247, 156, 30 },
		{ 247, 157, 30 },
		{ 247, 158, 30 },
		{ 247, 159, 30 },
		{ 247, 160, 30 },
		{ 247, 161, 30 },
		{ 247, 162, 30 },
		{ 247, 163, 30 },
		{ 247, 164, 30 },
		{ 247, 165, 30 },
		{ 247, 166, 30 },
		{ 247, 167, 30 },
		{ 247, 168, 30 },
		{ 247, 169, 30 },
		{ 247, 170, 30 },
		{ 247, 171, 30 },
		{ 247, 172, 30 },
		{ 247, 173, 30 },
		{ 247, 174, 30 },
		{ 247, 175, 30 },
		{ 247, 176, 30 },
		{ 247, 338, 30 },
		{ 247, 339, 30 },
		{ 247, 340, 30 },
		{ 247, 341, 30 },
		{ 247, 342, 30 },
		{ 247, 343, 30 },
		{ 247, 344, 30 },
		{ 247, 345, 30 },
		{ 247, 346, 30 },
		{ 247, 347, 30 },
		{ 247, 348, 30 },
		{ 247, 349, 30 },
		{ 247, 350, 30 },
		{ 247, 351, 30 },
		{ 247, 352, 30 },
		{ 247, 353, 30 },
		{ 247, 354, 30 },
		{ 247, 355, 30 },
		{ 247, 356, 30 },
		{ 247, 357, 30 },
		{ 247, 358, 30 },
		{ 247, 359, 30 },
		{ 247, 360, 30 },
		{ 247, 361, 30 },
		{ 247, 362, 30 },
		{ 247, 363, 30 },
		{ 247, 364, 30 },
		{ 247, 365, 30 },
		{ 247, 366, 30 },
		{ 247, 367, 30 },
		{ 247, 368, 30 },
		{ 247, 369, 30 },
		{ 247, 370, 30 },
		{ 247, 371, 30 },
		{ 247, 372, 30 },
		{ 247, 373, 30 },
		{ 247, 374, 30 },
		{ 247, 375, 30 },
		{ 247, 376, 30 },
		{ 247, 377, 30 },
		{ 247, 378, 30 },
		{ 247, 379, 30 },
		{ 247, 380, 30 },
		{ 247, 381, 30 },
		{ 247, 382, 30 },
		{ 247, 383, 30 },
		{ 247, 384, 30 },
		{ 247, 385, 30 },
		{ 247, 386, 30 },
		{ 247, 387, 30 },
		{ 247, 388, 30 },
		{ 247, 389, 30 },
		{ 247, 390, 30 },
		{ 247, 391, 30 },
		{ 247, 392, 30 },
		{ 247, 393, 30 },
		{ 247, 394, 30 },
		{ 247, 395, 30 },
		{ 247, 396, 30 },
		{ 247, 397, 30 },
		{ 247, 398, 30 },
		{ 247, 399, 30 },
		{ 247, 400, 30 },
		{ 247, 401, 30 },
		{ 247, 402, 30 },
		{ 247, 403, 30 },
		{ 247, 404, 30 },
		{ 247, 405, 30 },
		{ 247, 406, 30 },
		{ 247, 407, 30 },
		{ 247, 408, 30 },
		{ 247, 409, 30 },
		{ 247, 410, 30 },
		{ 247, 411, 30 },
		{ 247, 412, 30 },
		{ 247, 413, 30 },
		{ 247, 414, 30 },
		{ 247, 415, 30 },
		{ 247, 416, 30 },
		{ 247, 417, 30 },
		{ 247, 418, 30 },
		{ 247, 419, 30 },
		{ 247, 420, 30 },
		{ 247, 421, 30 },
		{ 247, 422, 30 },
		{ 247, 423, 30 },
		{ 247, 424, 30 },
		{ 247, 425, 30 },
		{ 247, 426, 30 },
		{ 247, 427, 30 },
		{ 247, 428, 30 },
		{ 247, 429, 30 },
		{ 247, 430, 30 },
		{ 247, 431, 30 },
		{ 247, 432, 30 },
		{ 247, 433, 30 },
		{ 247, 434, 30 },
		{ 247, 435, 30 },
		{ 247, 436, 30 },
		{ 247, 437, 30 },
		{ 247, 438, 30 },
		{ 247, 439, 30 },
		{ 247, 440, 30 },
		{ 247, 441, 30 },
		{ 247, 442, 30 },
		{ 247, 443, 30 },
		{ 247, 444, 30 },
		{ 247, 445, 30 },
		{ 247, 446, 30 },
		{ 247, 447, 30 },
		{ 247, 448, 30 },
		{ 247, 449, 30 },
		{ 247, 450, 30 },
		{ 247, 451, 30 },
		{ 247, 452, 30 },
		{ 247, 453, 30 },
		{ 247, 454, 30 },
		{ 247, 455, 30 },
		{ 247, 456, 30 },
		{ 247, 457, 30 },
		{ 247, 458, 30 },
		{ 247, 459, 30 },
		{ 247, 460, 30 },
		{ 247, 461, 30 },
		{ 247, 462, 30 },
		{ 247, 463, 30 },
		{ 247, 464, 30 },
		{ 247, 465, 30 },
		{ 247, 466, 30 },
		{ 247, 467, 30 },
		{ 247, 468, 30 },
		{ 247, 469, 30 },
		{ 247, 470, 30 },
		{ 247, 471, 30 },
		{ 247, 472, 30 },
		{ 247, 473, 30 },
		{ 247, 474, 30 },
		{ 247, 475, 30 },
		{ 247, 476, 30 },
		{ 247, 477, 30 },
		{ 247, 478, 30 },
		{ 247, 479, 30 },
		{ 247, 480, 30 },
		{ 247, 481, 30 },
		{ 247, 482, 30 },
		{ 247, 483, 30 },
		{ 247, 484, 30 },
		{ 247, 485, 30 },
		{ 247, 486, 30 },
		{ 247, 487, 30 },
		{ 247, 488, 30 },
		{ 247, 489, 30 },
		{ 247, 490, 30 },
		{ 247, 491, 30 },
		{ 247, 492, 30 },
		{ 247, 493, 30 },
		{ 247, 494, 30 },
		{ 247, 495, 30 },
		{ 247, 496, 30 },
		{ 247, 497, 30 },
		{ 247, 498, 30 },
		{ 247, 499, 30 },
		{ 247, 500, 30 },
		{ 247, 501, 30 },
		{ 247, 502, 30 },
		{ 247, 503, 30 },
		{ 247, 504, 30 },
		{ 247, 505, 30 },
		{ 247, 506, 30 },
		{ 247, 507, 30 },
		{ 247, 508, 30 },
		{ 247, 509, 30 },
		{ 247, 510, 30 },
		{ 247, 511, 30 },
		{ 247, 512, 30 },
		{ 248, 1, 30 },
		{ 248, 2, 30 },
		{ 248, 3, 30 },
		{ 248, 4, 30 },
		{ 248, 5, 30 },
		{ 248, 6, 30 },
		{ 248, 7, 30 },
		{ 248, 8, 30 },
		{ 248, 9, 30 },
		{ 248, 10, 30 },
		{ 248, 11, 30 },
		{ 248, 12, 30 },
		{ 248, 13, 30 },
		{ 248, 14, 30 },
		{ 248, 15, 30 },
		{ 248, 16, 30 },
		{ 248, 17, 30 },
		{ 248, 18, 30 },
		{ 248, 19, 30 },
		{ 248, 20, 30 },
		{ 248, 21, 30 },
		{ 248, 22, 30 },
		{ 248, 23, 30 },
		{ 248, 24, 30 },
		{ 248, 25, 30 },
		{ 248, 26, 30 },
		{ 248, 27, 30 },
		{ 248, 28, 30 },
		{ 248, 29, 30 },
		{ 248, 30, 30 },
		{ 248, 31, 30 },
		{ 248, 32, 30 },
		{ 248, 33, 30 },
		{ 248, 34, 30 },
		{ 248, 35, 30 },
		{ 248, 36, 30 },
		{ 248, 37, 30 },
		{ 248, 38, 30 },
		{ 248, 39, 30 },
		{ 248, 40, 30 },
		{ 248, 41, 30 },
		{ 248, 42, 30 },
		{ 248, 43, 30 },
		{ 248, 44, 30 },
		{ 248, 45, 30 },
		{ 248, 46, 30 },
		{ 248, 47, 30 },
		{ 248, 48, 30 },
		{ 248, 49, 30 },
		{ 248, 50, 30 },
		{ 248, 51, 30 },
		{ 248, 52, 30 },
		{ 248, 53, 30 },
		{ 248, 54, 30 },
		{ 248, 55, 30 },
		{ 248, 56, 30 },
		{ 248, 57, 30 },
		{ 248, 58, 30 },
		{ 248, 59, 30 },
		{ 248, 60, 30 },
		{ 248, 61, 30 },
		{ 248, 62, 30 },
		{ 248, 63, 30 },
		{ 248, 64, 30 },
		{ 248, 65, 30 },
		{ 248, 66, 30 },
		{ 248, 67, 30 },
		{ 248, 68, 30 },
		{ 248, 69, 30 },
		{ 248, 70, 30 },
		{ 248, 71, 30 },
		{ 248, 72, 30 },
		{ 248, 73, 30 },
		{ 248, 74, 30 },
		{ 248, 75, 30 },
		{ 248, 76, 30 },
		{ 248, 77, 30 },
		{ 248, 78, 30 },
		{ 248, 79, 30 },
		{ 248, 80, 30 },
		{ 248, 81, 30 },
		{ 248, 82, 30 },
		{ 248, 83, 30 },
		{ 248, 84, 30 },
		{ 248, 85, 30 },
		{ 248, 86, 30 },
		{ 248, 87, 30 },
		{ 248, 88, 30 },
		{ 248, 89, 30 },
		{ 248, 90, 30 },
		{ 248, 91, 30 },
		{ 248, 92, 30 },
		{ 248, 93, 30 },
		{ 248, 94, 30 },
		{ 248, 95, 30 },
		{ 248, 96, 30 },
		{ 248, 97, 30 },
		{ 248, 98, 30 },
		{ 248, 99, 30 },
		{ 248, 100, 30 },
		{ 248, 101, 30 },
		{ 248, 102, 30 },
		{ 248, 103, 30 },
		{ 248, 104, 30 },
		{ 248, 105, 30 },
		{ 248, 106, 30 },
		{ 248, 107, 30 },
		{ 248, 108, 30 },
		{ 248, 109, 30 },
		{ 248, 110, 30 },
		{ 248, 111, 30 },
		{ 248, 112, 30 },
		{ 248, 113, 30 },
		{ 248, 114, 30 },
		{ 248, 115, 30 },
		{ 248, 116, 30 },
		{ 248, 117, 30 },
		{ 248, 118, 30 },
		{ 248, 119, 30 },
		{ 248, 120, 30 },
		{ 248, 121, 30 },
		{ 248, 122, 30 },
		{ 248, 123, 30 },
		{ 248, 124, 30 },
		{ 248, 125, 30 },
		{ 248, 126, 30 },
		{ 248, 127, 30 },
		{ 248, 128, 30 },
		{ 248, 129, 30 },
		{ 248, 130, 30 },
		{ 248, 131, 30 },
		{ 248, 132, 30 },
		{ 248, 133, 30 },
		{ 248, 134, 30 },
		{ 248, 135, 30 },
		{ 248, 136, 30 },
		{ 248, 137, 30 },
		{ 248, 138, 30 },
		{ 248, 139, 30 },
		{ 248, 140, 30 },
		{ 248, 141, 30 },
		{ 248, 142, 30 },
		{ 248, 143, 30 },
		{ 248, 144, 30 },
		{ 248, 145, 30 },
		{ 248, 146, 30 },
		{ 248, 147, 30 },
		{ 248, 148, 30 },
		{ 248, 149, 30 },
		{ 248, 150, 30 },
		{ 248, 151, 30 },
		{ 248, 152, 30 },
		{ 248, 153, 30 },
		{ 248, 154, 30 },
		{ 248, 155, 30 },
		{ 248, 156, 30 },
		{ 248, 157, 30 },
		{ 248, 158, 30 },
		{ 248, 159, 30 },
		{ 248, 160, 30 },
		{ 248, 161, 30 },
		{ 248, 162, 30 },
		{ 248, 163, 30 },
		{ 248, 164, 30 },
		{ 248, 165, 30 },
		{ 248, 166, 30 },
		{ 248, 167, 30 },
		{ 248, 168, 30 },
		{ 248, 169, 30 },
		{ 248, 170, 30 },
		{ 248, 171, 30 },
		{ 248, 172, 30 },
		{ 248, 173, 30 },
		{ 248, 174, 30 },
		{ 248, 175, 30 },
		{ 248, 176, 30 },
		{ 248, 177, 30 },
		{ 248, 178, 30 },
		{ 248, 179, 30 },
		{ 248, 180, 30 },
		{ 248, 181, 30 },
		{ 248, 182, 30 },
		{ 248, 183, 30 },
		{ 248, 184, 30 },
		{ 248, 330, 30 },
		{ 248, 331, 30 },
		{ 248, 332, 30 },
		{ 248, 333, 30 },
		{ 248, 334, 30 },
		{ 248, 335, 30 },
		{ 248, 336, 30 },
		{ 248, 337, 30 },
		{ 248, 338, 30 },
		{ 248, 339, 30 },
		{ 248, 340, 30 },
		{ 248, 341, 30 },
		{ 248, 342, 30 },
		{ 248, 343, 30 },
		{ 248, 344, 30 },
		{ 248, 345, 30 },
		{ 248, 346, 30 },
		{ 248, 347, 30 },
		{ 248, 348, 30 },
		{ 248, 349, 30 },
		{ 248, 350, 30 },
		{ 248, 351, 30 },
		{ 248, 352, 30 },
		{ 248, 353, 30 },
		{ 248, 354, 30 },
		{ 248, 355, 30 },
		{ 248, 356, 30 },
		{ 248, 357, 30 },
		{ 248, 358, 30 },
		{ 248, 359, 30 },
		{ 248, 360, 30 },
		{ 248, 361, 30 },
		{ 248, 362, 30 },
		{ 248, 363, 30 },
		{ 248, 364, 30 },
		{ 248, 365, 30 },
		{ 248, 366, 30 },
		{ 248, 367, 30 },
		{ 248, 368, 30 },
		{ 248, 369, 30 },
		{ 248, 370, 30 },
		{ 248, 371, 30 },
		{ 248, 372, 30 },
		{ 248, 373, 30 },
		{ 248, 374, 30 },
		{ 248, 375, 30 },
		{ 248, 376, 30 },
		{ 248, 377, 30 },
		{ 248, 378, 30 },
		{ 248, 379, 30 },
		{ 248, 380, 30 },
		{ 248, 381, 30 },
		{ 248, 382, 30 },
		{ 248, 383, 30 },
		{ 248, 384, 30 },
		{ 248, 385, 30 },
		{ 248, 386, 30 },
		{ 248, 387, 30 },
		{ 248, 388, 30 },
		{ 248, 389, 30 },
		{ 248, 390, 30 },
		{ 248, 391, 30 },
		{ 248, 392, 30 },
		{ 248, 393, 30 },
		{ 248, 394, 30 },
		{ 248, 395, 30 },
		{ 248, 396, 30 },
		{ 248, 397, 30 },
		{ 248, 398, 30 },
		{ 248, 399, 30 },
		{ 248, 400, 30 },
		{ 248, 401, 30 },
		{ 248, 402, 30 },
		{ 248, 403, 30 },
		{ 248, 404, 30 },
		{ 248, 405, 30 },
		{ 248, 406, 30 },
		{ 248, 407, 30 },
		{ 248, 408, 30 },
		{ 248, 409, 30 },
		{ 248, 410, 30 },
		{ 248, 411, 30 },
		{ 248, 412, 30 },
		{ 248, 413, 30 },
		{ 248, 414, 30 },
		{ 248, 415, 30 },
		{ 248, 416, 30 },
		{ 248, 417, 30 },
		{ 248, 418, 30 },
		{ 248, 419, 30 },
		{ 248, 420, 30 },
		{ 248, 421, 30 },
		{ 248, 422, 30 },
		{ 248, 423, 30 },
		{ 248, 424, 30 },
		{ 248, 425, 30 },
		{ 248, 426, 30 },
		{ 248, 427, 30 },
		{ 248, 428, 30 },
		{ 248, 429, 30 },
		{ 248, 430, 30 },
		{ 248, 431, 30 },
		{ 248, 432, 30 },
		{ 248, 433, 30 },
		{ 248, 434, 30 },
		{ 248, 435, 30 },
		{ 248, 436, 30 },
		{ 248, 437, 30 },
		{ 248, 438, 30 },
		{ 248, 439, 30 },
		{ 248, 440, 30 },
		{ 248, 441, 30 },
		{ 248, 442, 30 },
		{ 248, 443, 30 },
		{ 248, 444, 30 },
		{ 248, 445, 30 },
		{ 248, 446, 30 },
		{ 248, 447, 30 },
		{ 248, 448, 30 },
		{ 248, 449, 30 },
		{ 248, 450, 30 },
		{ 248, 451, 30 },
		{ 248, 452, 30 },
		{ 248, 453, 30 },
		{ 248, 454, 30 },
		{ 248, 455, 30 },
		{ 248, 456, 30 },
		{ 248, 457, 30 },
		{ 248, 458, 30 },
		{ 248, 459, 30 },
		{ 248, 460, 30 },
		{ 248, 461, 30 },
		{ 248, 462, 30 },
		{ 248, 463, 30 },
		{ 248, 464, 30 },
		{ 248, 465, 30 },
		{ 248, 466, 30 },
		{ 248, 467, 30 },
		{ 248, 468, 30 },
		{ 248, 469, 30 },
		{ 248, 470, 30 },
		{ 248, 471, 30 },
		{ 248, 472, 30 },
		{ 248, 473, 30 },
		{ 248, 474, 30 },
		{ 248, 475, 30 },
		{ 248, 476, 30 },
		{ 248, 477, 30 },
		{ 248, 478, 30 },
		{ 248, 479, 30 },
		{ 248, 480, 30 },
		{ 248, 481, 30 },
		{ 248, 482, 30 },
		{ 248, 483, 30 },
		{ 248, 484, 30 },
		{ 248, 485, 30 },
		{ 248, 486, 30 },
		{ 248, 487, 30 },
		{ 248, 488, 30 },
		{ 248, 489, 30 },
		{ 248, 490, 30 },
		{ 248, 491, 30 },
		{ 248, 492, 30 },
		{ 248, 493, 30 },
		{ 248, 494, 30 },
		{ 248, 495, 30 },
		{ 248, 496, 30 },
		{ 248, 497, 30 },
		{ 248, 498, 30 },
		{ 248, 499, 30 },
		{ 248, 500, 30 },
		{ 248, 501, 30 },
		{ 248, 502, 30 },
		{ 248, 503, 30 },
		{ 248, 504, 30 },
		{ 248, 505, 30 },
		{ 248, 506, 30 },
		{ 248, 507, 30 },
		{ 248, 508, 30 },
		{ 248, 509, 30 },
		{ 248, 510, 30 },
		{ 248, 511, 30 },
		{ 248, 512, 30 },
		{ 249, 1, 30 },
		{ 249, 2, 30 },
		{ 249, 3, 30 },
		{ 249, 4, 30 },
		{ 249, 5, 30 },
		{ 249, 6, 30 },
		{ 249, 7, 30 },
		{ 249, 8, 30 },
		{ 249, 9, 30 },
		{ 249, 10, 30 },
		{ 249, 11, 30 },
		{ 249, 12, 30 },
		{ 249, 13, 30 },
		{ 249, 14, 30 },
		{ 249, 15, 30 },
		{ 249, 16, 30 },
		{ 249, 17, 30 },
		{ 249, 18, 30 },
		{ 249, 19, 30 },
		{ 249, 20, 30 },
		{ 249, 21, 30 },
		{ 249, 22, 30 },
		{ 249, 23, 30 },
		{ 249, 24, 30 },
		{ 249, 25, 30 },
		{ 249, 26, 30 },
		{ 249, 27, 30 },
		{ 249, 28, 30 },
		{ 249, 29, 30 },
		{ 249, 30, 30 },
		{ 249, 31, 30 },
		{ 249, 32, 30 },
		{ 249, 33, 30 },
		{ 249, 34, 30 },
		{ 249, 35, 30 },
		{ 249, 36, 30 },
		{ 249, 37, 30 },
		{ 249, 38, 30 },
		{ 249, 39, 30 },
		{ 249, 40, 30 },
		{ 249, 41, 30 },
		{ 249, 42, 30 },
		{ 249, 43, 30 },
		{ 249, 44, 30 },
		{ 249, 45, 30 },
		{ 249, 46, 30 },
		{ 249, 47, 30 },
		{ 249, 48, 30 },
		{ 249, 49, 30 },
		{ 249, 50, 30 },
		{ 249, 51, 30 },
		{ 249, 52, 30 },
		{ 249, 53, 30 },
		{ 249, 54, 30 },
		{ 249, 55, 30 },
		{ 249, 56, 30 },
		{ 249, 57, 30 },
		{ 249, 58, 30 },
		{ 249, 59, 30 },
		{ 249, 60, 30 },
		{ 249, 61, 30 },
		{ 249, 62, 30 },
		{ 249, 63, 30 },
		{ 249, 64, 30 },
		{ 249, 65, 30 },
		{ 249, 66, 30 },
		{ 249, 67, 30 },
		{ 249, 68, 30 },
		{ 249, 69, 30 },
		{ 249, 70, 30 },
		{ 249, 71, 30 },
		{ 249, 72, 30 },
		{ 249, 73, 30 },
		{ 249, 74, 30 },
		{ 249, 75, 30 },
		{ 249, 76, 30 },
		{ 249, 77, 30 },
		{ 249, 78, 30 },
		{ 249, 79, 30 },
		{ 249, 80, 30 },
		{ 249, 81, 30 },
		{ 249, 82, 30 },
		{ 249, 83, 30 },
		{ 249, 84, 30 },
		{ 249, 85, 30 },
		{ 249, 86, 30 },
		{ 249, 87, 30 },
		{ 249, 88, 30 },
		{ 249, 89, 30 },
		{ 249, 90, 30 },
		{ 249, 91, 30 },
		{ 249, 92, 30 },
		{ 249, 93, 30 },
		{ 249, 94, 30 },
		{ 249, 95, 30 },
		{ 249, 96, 30 },
		{ 249, 97, 30 },
		{ 249, 98, 30 },
		{ 249, 99, 30 },
		{ 249, 100, 30 },
		{ 249, 101, 30 },
		{ 249, 102, 30 },
		{ 249, 103, 30 },
		{ 249, 104, 30 },
		{ 249, 105, 30 },
		{ 249, 106, 30 },
		{ 249, 107, 30 },
		{ 249, 108, 30 },
		{ 249, 109, 30 },
		{ 249, 110, 30 },
		{ 249, 111, 30 },
		{ 249, 112, 30 },
		{ 249, 113, 30 },
		{ 249, 114, 30 },
		{ 249, 115, 30 },
		{ 249, 116, 30 },
		{ 249, 117, 30 },
		{ 249, 118, 30 },
		{ 249, 119, 30 },
		{ 249, 120, 30 },
		{ 249, 121, 30 },
		{ 249, 122, 30 },
		{ 249, 123, 30 },
		{ 249, 124, 30 },
		{ 249, 125, 30 },
		{ 249, 126, 30 },
		{ 249, 127, 30 },
		{ 249, 128, 30 },
		{ 249, 129, 30 },
		{ 249, 130, 30 },
		{ 249, 131, 30 },
		{ 249, 132, 30 },
		{ 249, 133, 30 },
		{ 249, 134, 30 },
		{ 249, 135, 30 },
		{ 249, 136, 30 },
		{ 249, 137, 30 },
		{ 249, 138, 30 },
		{ 249, 139, 30 },
		{ 249, 140, 30 },
		{ 249, 141, 30 },
		{ 249, 142, 30 },
		{ 249, 143, 30 },
		{ 249, 144, 30 },
		{ 249, 145, 30 },
		{ 249, 146, 30 },
		{ 249, 147, 30 },
		{ 249, 148, 30 },
		{ 249, 149, 30 },
		{ 249, 150, 30 },
		{ 249, 151, 30 },
		{ 249, 152, 30 },
		{ 249, 153, 30 },
		{ 249, 154, 30 },
		{ 249, 155, 30 },
		{ 249, 156, 30 },
		{ 249, 157, 30 },
		{ 249, 158, 30 },
		{ 249, 159, 30 },
		{ 249, 160, 30 },
		{ 249, 161, 30 },
		{ 249, 162, 30 },
		{ 249, 163, 30 },
		{ 249, 164, 30 },
		{ 249, 165, 30 },
		{ 249, 166, 30 },
		{ 249, 167, 30 },
		{ 249, 168, 30 },
		{ 249, 169, 30 },
		{ 249, 170, 30 },
		{ 249, 171, 30 },
		{ 249, 172, 30 },
		{ 249, 173, 30 },
		{ 249, 174, 30 },
		{ 249, 175, 30 },
		{ 249, 176, 30 },
		{ 249, 177, 30 },
		{ 249, 178, 30 },
		{ 249, 179, 30 },
		{ 249, 180, 30 },
		{ 249, 181, 30 },
		{ 249, 182, 30 },
		{ 249, 183, 30 },
		{ 249, 184, 30 },
		{ 249, 185, 30 },
		{ 249, 186, 30 },
		{ 249, 187, 30 },
		{ 249, 188, 30 },
		{ 249, 189, 30 },
		{ 249, 325, 30 },
		{ 249, 326, 30 },
		{ 249, 327, 30 },
		{ 249, 328, 30 },
		{ 249, 329, 30 },
		{ 249, 330, 30 },
		{ 249, 331, 30 },
		{ 249, 332, 30 },
		{ 249, 333, 30 },
		{ 249, 334, 30 },
		{ 249, 335, 30 },
		{ 249, 336, 30 },
		{ 249, 337, 30 },
		{ 249, 338, 30 },
		{ 249, 339, 30 },
		{ 249, 340, 30 },
		{ 249, 341, 30 },
		{ 249, 342, 30 },
		{ 249, 343, 30 },
		{ 249, 344, 30 },
		{ 249, 345, 30 },
		{ 249, 346, 30 },
		{ 249, 347, 30 },
		{ 249, 348, 30 },
		{ 249, 349, 30 },
		{ 249, 350, 30 },
		{ 249, 351, 30 },
		{ 249, 352, 30 },
		{ 249, 353, 30 },
		{ 249, 354, 30 },
		{ 249, 355, 30 },
		{ 249, 356, 30 },
		{ 249, 357, 30 },
		{ 249, 358, 30 },
		{ 249, 359, 30 },
		{ 249, 360, 30 },
		{ 249, 361, 30 },
		{ 249, 362, 30 },
		{ 249, 363, 30 },
		{ 249, 364, 30 },
		{ 249, 365, 30 },
		{ 249, 366, 30 },
		{ 249, 367, 30 },
		{ 249, 368, 30 },
		{ 249, 369, 30 },
		{ 249, 370, 30 },
		{ 249, 371, 30 },
		{ 249, 372, 30 },
		{ 249, 373, 30 },
		{ 249, 374, 30 },
		{ 249, 375, 30 },
		{ 249, 376, 30 },
		{ 249, 377, 30 },
		{ 249, 378, 30 },
		{ 249, 379, 30 },
		{ 249, 380, 30 },
		{ 249, 381, 30 },
		{ 249, 382, 30 },
		{ 249, 383, 30 },
		{ 249, 384, 30 },
		{ 249, 385, 30 },
		{ 249, 386, 30 },
		{ 249, 387, 30 },
		{ 249, 388, 30 },
		{ 249, 389, 30 },
		{ 249, 390, 30 },
		{ 249, 391, 30 },
		{ 249, 392, 30 },
		{ 249, 393, 30 },
		{ 249, 394, 30 },
		{ 249, 395, 30 },
		{ 249, 396, 30 },
		{ 249, 397, 30 },
		{ 249, 398, 30 },
		{ 249, 399, 30 },
		{ 249, 400, 30 },
		{ 249, 401, 30 },
		{ 249, 402, 30 },
		{ 249, 403, 30 },
		{ 249, 404, 30 },
		{ 249, 405, 30 },
		{ 249, 406, 30 },
		{ 249, 407, 30 },
		{ 249, 408, 30 },
		{ 249, 409, 30 },
		{ 249, 410, 30 },
		{ 249, 411, 30 },
		{ 249, 412, 30 },
		{ 249, 413, 30 },
		{ 249, 414, 30 },
		{ 249, 415, 30 },
		{ 249, 416, 30 },
		{ 249, 417, 30 },
		{ 249, 418, 30 },
		{ 249, 419, 30 },
		{ 249, 420, 30 },
		{ 249, 421, 30 },
		{ 249, 422, 30 },
		{ 249, 423, 30 },
		{ 249, 424, 30 },
		{ 249, 425, 30 },
		{ 249, 426, 30 },
		{ 249, 427, 30 },
		{ 249, 428, 30 },
		{ 249, 429, 30 },
		{ 249, 430, 30 },
		{ 249, 431, 30 },
		{ 249, 432, 30 },
		{ 249, 433, 30 },
		{ 249, 434, 30 },
		{ 249, 435, 30 },
		{ 249, 436, 30 },
		{ 249, 437, 30 },
		{ 249, 438, 30 },
		{ 249, 439, 30 },
		{ 249, 440, 30 },
		{ 249, 441, 30 },
		{ 249, 442, 30 },
		{ 249, 443, 30 },
		{ 249, 444, 30 },
		{ 249, 445, 30 },
		{ 249, 446, 30 },
		{ 249, 447, 30 },
		{ 249, 448, 30 },
		{ 249, 449, 30 },
		{ 249, 450, 30 },
		{ 249, 451, 30 },
		{ 249, 452, 30 },
		{ 249, 453, 30 },
		{ 249, 454, 30 },
		{ 249, 455, 30 },
		{ 249, 456, 30 },
		{ 249, 457, 30 },
		{ 249, 458, 30 },
		{ 249, 459, 30 },
		{ 249, 460, 30 },
		{ 249, 461, 30 },
		{ 249, 462, 30 },
		{ 249, 463, 30 },
		{ 249, 464, 30 },
		{ 249, 465, 30 },
		{ 249, 466, 30 },
		{ 249, 467, 30 },
		{ 249, 468, 30 },
		{ 249, 469, 30 },
		{ 249, 470, 30 },
		{ 249, 471, 30 },
		{ 249, 472, 30 },
		{ 249, 473, 30 },
		{ 249, 474, 30 },
		{ 249, 475, 30 },
		{ 249, 476, 30 },
		{ 249, 477, 30 },
		{ 249, 478, 30 },
		{ 249, 479, 30 },
		{ 249, 480, 30 },
		{ 249, 481, 30 },
		{ 249, 482, 30 },
		{ 249, 483, 30 },
		{ 249, 484, 30 },
		{ 249, 485, 30 },
		{ 249, 486, 30 },
		{ 249, 487, 30 },
		{ 249, 488, 30 },
		{ 249, 489, 30 },
		{ 249, 490, 30 },
		{ 249, 491, 30 },
		{ 249, 492, 30 },
		{ 249, 493, 30 },
		{ 249, 494, 30 },
		{ 249, 495, 30 },
		{ 249, 496, 30 },
		{ 249, 497, 30 },
		{ 249, 498, 30 },
		{ 249, 499, 30 },
		{ 249, 500, 30 },
		{ 249, 501, 30 },
		{ 249, 502, 30 },
		{ 249, 503, 30 },
		{ 249, 504, 30 },
		{ 249, 505, 30 },
		{ 249, 506, 30 },
		{ 249, 507, 30 },
		{ 249, 508, 30 },
		{ 249, 509, 30 },
		{ 249, 510, 30 },
		{ 249, 511, 30 },
		{ 249, 512, 30 },
		{ 250, 1, 30 },
		{ 250, 2, 30 },
		{ 250, 3, 30 },
		{ 250, 4, 30 },
		{ 250, 5, 30 },
		{ 250, 6, 30 },
		{ 250, 7, 30 },
		{ 250, 8, 30 },
		{ 250, 9, 30 },
		{ 250, 10, 30 },
		{ 250, 11, 30 },
		{ 250, 12, 30 },
		{ 250, 13, 30 },
		{ 250, 14, 30 },
		{ 250, 15, 30 },
		{ 250, 16, 30 },
		{ 250, 17, 30 },
		{ 250, 18, 30 },
		{ 250, 19, 30 },
		{ 250, 20, 30 },
		{ 250, 21, 30 },
		{ 250, 22, 30 },
		{ 250, 23, 30 },
		{ 250, 24, 30 },
		{ 250, 25, 30 },
		{ 250, 26, 30 },
		{ 250, 27, 30 },
		{ 250, 28, 30 },
		{ 250, 29, 30 },
		{ 250, 30, 30 },
		{ 250, 31, 30 },
		{ 250, 32, 30 },
		{ 250, 33, 30 },
		{ 250, 34, 30 },
		{ 250, 35, 30 },
		{ 250, 36, 30 },
		{ 250, 37, 30 },
		{ 250, 38, 30 },
		{ 250, 39, 30 },
		{ 250, 40, 30 },
		{ 250, 41, 30 },
		{ 250, 42, 30 },
		{ 250, 43, 30 },
		{ 250, 44, 30 },
		{ 250, 45, 30 },
		{ 250, 46, 30 },
		{ 250, 47, 30 },
		{ 250, 48, 30 },
		{ 250, 49, 30 },
		{ 250, 50, 30 },
		{ 250, 51, 30 },
		{ 250, 52, 30 },
		{ 250, 53, 30 },
		{ 250, 54, 30 },
		{ 250, 55, 30 },
		{ 250, 56, 30 },
		{ 250, 57, 30 },
		{ 250, 58, 30 },
		{ 250, 59, 30 },
		{ 250, 60, 30 },
		{ 250, 61, 30 },
		{ 250, 62, 30 },
		{ 250, 63, 30 },
		{ 250, 64, 30 },
		{ 250, 65, 30 },
		{ 250, 66, 30 },
		{ 250, 67, 30 },
		{ 250, 68, 30 },
		{ 250, 69, 30 },
		{ 250, 70, 30 },
		{ 250, 71, 30 },
		{ 250, 72, 30 },
		{ 250, 73, 30 },
		{ 250, 74, 30 },
		{ 250, 75, 30 },
		{ 250, 76, 30 },
		{ 250, 77, 30 },
		{ 250, 78, 30 },
		{ 250, 79, 30 },
		{ 250, 80, 30 },
		{ 250, 81, 30 },
		{ 250, 82, 30 },
		{ 250, 83, 30 },
		{ 250, 84, 30 },
		{ 250, 85, 30 },
		{ 250, 86, 30 },
		{ 250, 87, 30 },
		{ 250, 88, 30 },
		{ 250, 89, 30 },
		{ 250, 90, 30 },
		{ 250, 91, 30 },
		{ 250, 92, 30 },
		{ 250, 93, 30 },
		{ 250, 94, 30 },
		{ 250, 95, 30 },
		{ 250, 96, 30 },
		{ 250, 97, 30 },
		{ 250, 98, 30 },
		{ 250, 99, 30 },
		{ 250, 100, 30 },
		{ 250, 101, 30 },
		{ 250, 102, 30 },
		{ 250, 103, 30 },
		{ 250, 104, 30 },
		{ 250, 105, 30 },
		{ 250, 106, 30 },
		{ 250, 107, 30 },
		{ 250, 108, 30 },
		{ 250, 109, 30 },
		{ 250, 110, 30 },
		{ 250, 111, 30 },
		{ 250, 112, 30 },
		{ 250, 113, 30 },
		{ 250, 114, 30 },
		{ 250, 115, 30 },
		{ 250, 116, 30 },
		{ 250, 117, 30 },
		{ 250, 118, 30 },
		{ 250, 119, 30 },
		{ 250, 120, 30 },
		{ 250, 121, 30 },
		{ 250, 122, 30 },
		{ 250, 123, 30 },
		{ 250, 124, 30 },
		{ 250, 125, 30 },
		{ 250, 126, 30 },
		{ 250, 127, 30 },
		{ 250, 128, 30 },
		{ 250, 129, 30 },
		{ 250, 130, 30 },
		{ 250, 131, 30 },
		{ 250, 132, 30 },
		{ 250, 133, 30 },
		{ 250, 134, 30 },
		{ 250, 135, 30 },
		{ 250, 136, 30 },
		{ 250, 137, 30 },
		{ 250, 138, 30 },
		{ 250, 139, 30 },
		{ 250, 140, 30 },
		{ 250, 141, 30 },
		{ 250, 142, 30 },
		{ 250, 143, 30 },
		{ 250, 144, 30 },
		{ 250, 145, 30 },
		{ 250, 146, 30 },
		{ 250, 147, 30 },
		{ 250, 148, 30 },
		{ 250, 149, 30 },
		{ 250, 150, 30 },
		{ 250, 151, 30 },
		{ 250, 152, 30 },
		{ 250, 153, 30 },
		{ 250, 154, 30 },
		{ 250, 155, 30 },
		{ 250, 156, 30 },
		{ 250, 157, 30 },
		{ 250, 158, 30 },
		{ 250, 159, 30 },
		{ 250, 160, 30 },
		{ 250, 161, 30 },
		{ 250, 162, 30 },
		{ 250, 163, 30 },
		{ 250, 164, 30 },
		{ 250, 165, 30 },
		{ 250, 166, 30 },
		{ 250, 167, 30 },
		{ 250, 168, 30 },
		{ 250, 169, 30 },
		{ 250, 170, 30 },
		{ 250, 171, 30 },
		{ 250, 172, 30 },
		{ 250, 173, 30 },
		{ 250, 174, 30 },
		{ 250, 175, 30 },
		{ 250, 176, 30 },
		{ 250, 177, 30 },
		{ 250, 178, 30 },
		{ 250, 179, 30 },
		{ 250, 180, 30 },
		{ 250, 181, 30 },
		{ 250, 182, 30 },
		{ 250, 183, 30 },
		{ 250, 184, 30 },
		{ 250, 185, 30 },
		{ 250, 186, 30 },
		{ 250, 187, 30 },
		{ 250, 188, 30 },
		{ 250, 189, 30 },
		{ 250, 190, 30 },
		{ 250, 324, 30 },
		{ 250, 325, 30 },
		{ 250, 326, 30 },
		{ 250, 327, 30 },
		{ 250, 328, 30 },
		{ 250, 329, 30 },
		{ 250, 330, 30 },
		{ 250, 331, 30 },
		{ 250, 332, 30 },
		{ 250, 333, 30 },
		{ 250, 334, 30 },
		{ 250, 335, 30 },
		{ 250, 336, 30 },
		{ 250, 337, 30 },
		{ 250, 338, 30 },
		{ 250, 339, 30 },
		{ 250, 340, 30 },
		{ 250, 341, 30 },
		{ 250, 342, 30 },
		{ 250, 343, 30 },
		{ 250, 344, 30 },
		{ 250, 345, 30 },
		{ 250, 346, 30 },
		{ 250, 347, 30 },
		{ 250, 348, 30 },
		{ 250, 349, 30 },
		{ 250, 350, 30 },
		{ 250, 351, 30 },
		{ 250, 352, 30 },
		{ 250, 353, 30 },
		{ 250, 354, 30 },
		{ 250, 355, 30 },
		{ 250, 356, 30 },
		{ 250, 357, 30 },
		{ 250, 358, 30 },
		{ 250, 359, 30 },
		{ 250, 360, 30 },
		{ 250, 361, 30 },
		{ 250, 362, 30 },
		{ 250, 363, 30 },
		{ 250, 364, 30 },
		{ 250, 365, 30 },
		{ 250, 366, 30 },
		{ 250, 367, 30 },
		{ 250, 368, 30 },
		{ 250, 369, 30 },
		{ 250, 370, 30 },
		{ 250, 371, 30 },
		{ 250, 372, 30 },
		{ 250, 373, 30 },
		{ 250, 374, 30 },
		{ 250, 375, 30 },
		{ 250, 376, 30 },
		{ 250, 377, 30 },
		{ 250, 378, 30 },
		{ 250, 379, 30 },
		{ 250, 380, 30 },
		{ 250, 381, 30 },
		{ 250, 382, 30 },
		{ 250, 383, 30 },
		{ 250, 384, 30 },
		{ 250, 385, 30 },
		{ 250, 386, 30 },
		{ 250, 387, 30 },
		{ 250, 388, 30 },
		{ 250, 389, 30 },
		{ 250, 390, 30 },
		{ 250, 391, 30 },
		{ 250, 392, 30 },
		{ 250, 393, 30 },
		{ 250, 394, 30 },
		{ 250, 395, 30 },
		{ 250, 396, 30 },
		{ 250, 397, 30 },
		{ 250, 398, 30 },
		{ 250, 399, 30 },
		{ 250, 400, 30 },
		{ 250, 401, 30 },
		{ 250, 402, 30 },
		{ 250, 403, 30 },
		{ 250, 404, 30 },
		{ 250, 405, 30 },
		{ 250, 406, 30 },
		{ 250, 407, 30 },
		{ 250, 408, 30 },
		{ 250, 409, 30 },
		{ 250, 410, 30 },
		{ 250, 411, 30 },
		{ 250, 412, 30 },
		{ 250, 413, 30 },
		{ 250, 414, 30 },
		{ 250, 415, 30 },
		{ 250, 416, 30 },
		{ 250, 417, 30 },
		{ 250, 418, 30 },
		{ 250, 419, 30 },
		{ 250, 420, 30 },
		{ 250, 421, 30 },
		{ 250, 422, 30 },
		{ 250, 423, 30 },
		{ 250, 424, 30 },
		{ 250, 425, 30 },
		{ 250, 426, 30 },
		{ 250, 427, 30 },
		{ 250, 428, 30 },
		{ 250, 429, 30 },
		{ 250, 430, 30 },
		{ 250, 431, 30 },
		{ 250, 432, 30 },
		{ 250, 433, 30 },
		{ 250, 434, 30 },
		{ 250, 435, 30 },
		{ 250, 436, 30 },
		{ 250, 437, 30 },
		{ 250, 438, 30 },
		{ 250, 439, 30 },
		{ 250, 440, 30 },
		{ 250, 441, 30 },
		{ 250, 442, 30 },
		{ 250, 443, 30 },
		{ 250, 444, 30 },
		{ 250, 445, 30 },
		{ 250, 446, 30 },
		{ 250, 447, 30 },
		{ 250, 448, 30 },
		{ 250, 449, 30 },
		{ 250, 450, 30 },
		{ 250, 451, 30 },
		{ 250, 452, 30 },
		{ 250, 453, 30 },
		{ 250, 454, 30 },
		{ 250, 455, 30 },
		{ 250, 456, 30 },
		{ 250, 457, 30 },
		{ 250, 458, 30 },
		{ 250, 459, 30 },
		{ 250, 460, 30 },
		{ 250, 461, 30 },
		{ 250, 462, 30 },
		{ 250, 463, 30 },
		{ 250, 464, 30 },
		{ 250, 465, 30 },
		{ 250, 466, 30 },
		{ 250, 467, 30 },
		{ 250, 468, 30 },
		{ 250, 469, 30 },
		{ 250, 470, 30 },
		{ 250, 471, 30 },
		{ 250, 472, 30 },
		{ 250, 473, 30 },
		{ 250, 474, 30 },
		{ 250, 475, 30 },
		{ 250, 476, 30 },
		{ 250, 477, 30 },
		{ 250, 478, 30 },
		{ 250, 479, 30 },
		{ 250, 480, 30 },
		{ 250, 481, 30 },
		{ 250, 482, 30 },
		{ 250, 483, 30 },
		{ 250, 484, 30 },
		{ 250, 485, 30 },
		{ 250, 486, 30 },
		{ 250, 487, 30 },
		{ 250, 488, 30 },
		{ 250, 489, 30 },
		{ 250, 490, 30 },
		{ 250, 491, 30 },
		{ 250, 492, 30 },
		{ 250, 493, 30 },
		{ 250, 494, 30 },
		{ 250, 495, 30 },
		{ 250, 496, 30 },
		{ 250, 497, 30 },
		{ 250, 498, 30 },
		{ 250, 499, 30 },
		{ 250, 500, 30 },
		{ 250, 501, 30 },
		{ 250, 502, 30 },
		{ 250, 503, 30 },
		{ 250, 504, 30 },
		{ 250, 505, 30 },
		{ 250, 506, 30 },
		{ 250, 507, 30 },
		{ 250, 508, 30 },
		{ 250, 509, 30 },
		{ 250, 510, 30 },
		{ 250, 511, 30 },
		{ 250, 512, 30 },
		{ 251, 1, 30 },
		{ 251, 2, 30 },
		{ 251, 3, 30 },
		{ 251, 4, 30 },
		{ 251, 5, 30 },
		{ 251, 6, 30 },
		{ 251, 7, 30 },
		{ 251, 8, 30 },
		{ 251, 9, 30 },
		{ 251, 10, 30 },
		{ 251, 11, 30 },
		{ 251, 12, 30 },
		{ 251, 13, 30 },
		{ 251, 14, 30 },
		{ 251, 15, 30 },
		{ 251, 16, 30 },
		{ 251, 17, 30 },
		{ 251, 18, 30 },
		{ 251, 19, 30 },
		{ 251, 20, 30 },
		{ 251, 21, 30 },
		{ 251, 22, 30 },
		{ 251, 23, 30 },
		{ 251, 24, 30 },
		{ 251, 25, 30 },
		{ 251, 26, 30 },
		{ 251, 27, 30 },
		{ 251, 28, 30 },
		{ 251, 29, 30 },
		{ 251, 30, 30 },
		{ 251, 31, 30 },
		{ 251, 32, 30 },
		{ 251, 33, 30 },
		{ 251, 34, 30 },
		{ 251, 35, 30 },
		{ 251, 36, 30 },
		{ 251, 37, 30 },
		{ 251, 38, 30 },
		{ 251, 39, 30 },
		{ 251, 40, 30 },
		{ 251, 41, 30 },
		{ 251, 42, 30 },
		{ 251, 43, 30 },
		{ 251, 44, 30 },
		{ 251, 45, 30 },
		{ 251, 46, 30 },
		{ 251, 47, 30 },
		{ 251, 48, 30 },
		{ 251, 49, 30 },
		{ 251, 50, 30 },
		{ 251, 51, 30 },
		{ 251, 52, 30 },
		{ 251, 53, 30 },
		{ 251, 54, 30 },
		{ 251, 55, 30 },
		{ 251, 56, 30 },
		{ 251, 57, 30 },
		{ 251, 58, 30 },
		{ 251, 59, 30 },
		{ 251, 60, 30 },
		{ 251, 61, 30 },
		{ 251, 62, 30 },
		{ 251, 63, 30 },
		{ 251, 64, 30 },
		{ 251, 65, 30 },
		{ 251, 66, 30 },
		{ 251, 67, 30 },
		{ 251, 68, 30 },
		{ 251, 69, 30 },
		{ 251, 70, 30 },
		{ 251, 71, 30 },
		{ 251, 72, 30 },
		{ 251, 73, 30 },
		{ 251, 74, 30 },
		{ 251, 75, 30 },
		{ 251, 76, 30 },
		{ 251, 77, 30 },
		{ 251, 78, 30 },
		{ 251, 79, 30 },
		{ 251, 80, 30 },
		{ 251, 81, 30 },
		{ 251, 82, 30 },
		{ 251, 83, 30 },
		{ 251, 84, 30 },
		{ 251, 85, 30 },
		{ 251, 86, 30 },
		{ 251, 87, 30 },
		{ 251, 88, 30 },
		{ 251, 89, 30 },
		{ 251, 90, 30 },
		{ 251, 91, 30 },
		{ 251, 92, 30 },
		{ 251, 93, 30 },
		{ 251, 94, 30 },
		{ 251, 95, 30 },
		{ 251, 96, 30 },
		{ 251, 97, 30 },
		{ 251, 98, 30 },
		{ 251, 99, 30 },
		{ 251, 100, 30 },
		{ 251, 101, 30 },
		{ 251, 102, 30 },
		{ 251, 103, 30 },
		{ 251, 104, 30 },
		{ 251, 105, 30 },
		{ 251, 106, 30 },
		{ 251, 107, 30 },
		{ 251, 108, 30 },
		{ 251, 109, 30 },
		{ 251, 110, 30 },
		{ 251, 111, 30 },
		{ 251, 112, 30 },
		{ 251, 113, 30 },
		{ 251, 114, 30 },
		{ 251, 115, 30 },
		{ 251, 116, 30 },
		{ 251, 117, 30 },
		{ 251, 118, 30 },
		{ 251, 119, 30 },
		{ 251, 120, 30 },
		{ 251, 121, 30 },
		{ 251, 122, 30 },
		{ 251, 123, 30 },
		{ 251, 124, 30 },
		{ 251, 125, 30 },
		{ 251, 126, 30 },
		{ 251, 127, 30 },
		{ 251, 128, 30 },
		{ 251, 129, 30 },
		{ 251, 130, 30 },
		{ 251, 131, 30 },
		{ 251, 132, 30 },
		{ 251, 133, 30 },
		{ 251, 134, 30 },
		{ 251, 135, 30 },
		{ 251, 136, 30 },
		{ 251, 137, 30 },
		{ 251, 138, 30 },
		{ 251, 139, 30 },
		{ 251, 140, 30 },
		{ 251, 141, 30 },
		{ 251, 142, 30 },
		{ 251, 143, 30 },
		{ 251, 144, 30 },
		{ 251, 145, 30 },
		{ 251, 146, 30 },
		{ 251, 147, 30 },
		{ 251, 148, 30 },
		{ 251, 149, 30 },
		{ 251, 150, 30 },
		{ 251, 151, 30 },
		{ 251, 152, 30 },
		{ 251, 153, 30 },
		{ 251, 154, 30 },
		{ 251, 155, 30 },
		{ 251, 156, 30 },
		{ 251, 157, 30 },
		{ 251, 158, 30 },
		{ 251, 159, 30 },
		{ 251, 160, 30 },
		{ 251, 161, 30 },
		{ 251, 162, 30 },
		{ 251, 163, 30 },
		{ 251, 164, 30 },
		{ 251, 165, 30 },
		{ 251, 166, 30 },
		{ 251, 167, 30 },
		{ 251, 168, 30 },
		{ 251, 169, 30 },
		{ 251, 170, 30 },
		{ 251, 171, 30 },
		{ 251, 172, 30 },
		{ 251, 173, 30 },
		{ 251, 174, 30 },
		{ 251, 175, 30 },
		{ 251, 176, 30 },
		{ 251, 177, 30 },
		{ 251, 178, 30 },
		{ 251, 179, 30 },
		{ 251, 180, 30 },
		{ 251, 181, 30 },
		{ 251, 182, 30 },
		{ 251, 183, 30 },
		{ 251, 184, 30 },
		{ 251, 185, 30 },
		{ 251, 186, 30 },
		{ 251, 187, 30 },
		{ 251, 188, 30 },
		{ 251, 189, 30 },
		{ 251, 190, 30 },
		{ 251, 191, 30 },
		{ 251, 323, 30 },
		{ 251, 324, 30 },
		{ 251, 325, 30 },
		{ 251, 326, 30 },
		{ 251, 327, 30 },
		{ 251, 328, 30 },
		{ 251, 329, 30 },
		{ 251, 330, 30 },
		{ 251, 331, 30 },
		{ 251, 332, 30 },
		{ 251, 333, 30 },
		{ 251, 334, 30 },
		{ 251, 335, 30 },
		{ 251, 336, 30 },
		{ 251, 337, 30 },
		{ 251, 338, 30 },
		{ 251, 339, 30 },
		{ 251, 340, 30 },
		{ 251, 341, 30 },
		{ 251, 342, 30 },
		{ 251, 343, 30 },
		{ 251, 344, 30 },
		{ 251, 345, 30 },
		{ 251, 346, 30 },
		{ 251, 347, 30 },
		{ 251, 348, 30 },
		{ 251, 349, 30 },
		{ 251, 350, 30 },
		{ 251, 351, 30 },
		{ 251, 352, 30 },
		{ 251, 353, 30 },
		{ 251, 354, 30 },
		{ 251, 355, 30 },
		{ 251, 356, 30 },
		{ 251, 357, 30 },
		{ 251, 358, 30 },
		{ 251, 359, 30 },
		{ 251, 360, 30 },
		{ 251, 361, 30 },
		{ 251, 362, 30 },
		{ 251, 363, 30 },
		{ 251, 364, 30 },
		{ 251, 365, 30 },
		{ 251, 366, 30 },
		{ 251, 367, 30 },
		{ 251, 368, 30 },
		{ 251, 369, 30 },
		{ 251, 370, 30 },
		{ 251, 371, 30 },
		{ 251, 372, 30 },
		{ 251, 373, 30 },
		{ 251, 374, 30 },
		{ 251, 375, 30 },
		{ 251, 376, 30 },
		{ 251, 377, 30 },
		{ 251, 378, 30 },
		{ 251, 379, 30 },
		{ 251, 380, 30 },
		{ 251, 381, 30 },
		{ 251, 382, 30 },
		{ 251, 383, 30 },
		{ 251, 384, 30 },
		{ 251, 385, 30 },
		{ 251, 386, 30 },
		{ 251, 387, 30 },
		{ 251, 388, 30 },
		{ 251, 389, 30 },
		{ 251, 390, 30 },
		{ 251, 391, 30 },
		{ 251, 392, 30 },
		{ 251, 393, 30 },
		{ 251, 394, 30 },
		{ 251, 395, 30 },
		{ 251, 396, 30 },
		{ 251, 397, 30 },
		{ 251, 398, 30 },
		{ 251, 399, 30 },
		{ 251, 400, 30 },
		{ 251, 401, 30 },
		{ 251, 402, 30 },
		{ 251, 403, 30 },
		{ 251, 404, 30 },
		{ 251, 405, 30 },
		{ 251, 406, 30 },
		{ 251, 407, 30 },
		{ 251, 408, 30 },
		{ 251, 409, 30 },
		{ 251, 410, 30 },
		{ 251, 411, 30 },
		{ 251, 412, 30 },
		{ 251, 413, 30 },
		{ 251, 414, 30 },
		{ 251, 415, 30 },
		{ 251, 416, 30 },
		{ 251, 417, 30 },
		{ 251, 418, 30 },
		{ 251, 419, 30 },
		{ 251, 420, 30 },
		{ 251, 421, 30 },
		{ 251, 422, 30 },
		{ 251, 423, 30 },
		{ 251, 424, 30 },
		{ 251, 425, 30 },
		{ 251, 426, 30 },
		{ 251, 427, 30 },
		{ 251, 428, 30 },
		{ 251, 429, 30 },
		{ 251, 430, 30 },
		{ 251, 431, 30 },
		{ 251, 432, 30 },
		{ 251, 433, 30 },
		{ 251, 434, 30 },
		{ 251, 435, 30 },
		{ 251, 436, 30 },
		{ 251, 437, 30 },
		{ 251, 438, 30 },
		{ 251, 439, 30 },
		{ 251, 440, 30 },
		{ 251, 441, 30 },
		{ 251, 442, 30 },
		{ 251, 443, 30 },
		{ 251, 444, 30 },
		{ 251, 445, 30 },
		{ 251, 446, 30 },
		{ 251, 447, 30 },
		{ 251, 448, 30 },
		{ 251, 449, 30 },
		{ 251, 450, 30 },
		{ 251, 451, 30 },
		{ 251, 452, 30 },
		{ 251, 453, 30 },
		{ 251, 454, 30 },
		{ 251, 455, 30 },
		{ 251, 456, 30 },
		{ 251, 457, 30 },
		{ 251, 458, 30 },
		{ 251, 459, 30 },
		{ 251, 460, 30 },
		{ 251, 461, 30 },
		{ 251, 462, 30 },
		{ 251, 463, 30 },
		{ 251, 464, 30 },
		{ 251, 465, 30 },
		{ 251, 466, 30 },
		{ 251, 467, 30 },
		{ 251, 468, 30 },
		{ 251, 469, 30 },
		{ 251, 470, 30 },
		{ 251, 471, 30 },
		{ 251, 472, 30 },
		{ 251, 473, 30 },
		{ 251, 474, 30 },
		{ 251, 475, 30 },
		{ 251, 476, 30 },
		{ 251, 477, 30 },
		{ 251, 478, 30 },
		{ 251, 479, 30 },
		{ 251, 480, 30 },
		{ 251, 481, 30 },
		{ 251, 482, 30 },
		{ 251, 483, 30 },
		{ 251, 484, 30 },
		{ 251, 485, 30 },
		{ 251, 486, 30 },
		{ 251, 487, 30 },
		{ 251, 488, 30 },
		{ 251, 489, 30 },
		{ 251, 490, 30 },
		{ 251, 491, 30 },
		{ 251, 492, 30 },
		{ 251, 493, 30 },
		{ 251, 494, 30 },
		{ 251, 495, 30 },
		{ 251, 496, 30 },
		{ 251, 497, 30 },
		{ 251, 498, 30 },
		{ 251, 499, 30 },
		{ 251, 500, 30 },
		{ 251, 501, 30 },
		{ 251, 502, 30 },
		{ 251, 503, 30 },
		{ 251, 504, 30 },
		{ 251, 505, 30 },
		{ 251, 506, 30 },
		{ 251, 507, 30 },
		{ 251, 508, 30 },
		{ 251, 509, 30 },
		{ 251, 510, 30 },
		{ 251, 511, 30 },
		{ 251, 512, 30 },
		{ 252, 1, 30 },
		{ 252, 2, 30 },
		{ 252, 3, 30 },
		{ 252, 4, 30 },
		{ 252, 5, 30 },
		{ 252, 6, 30 },
		{ 252, 7, 30 },
		{ 252, 8, 30 },
		{ 252, 9, 30 },
		{ 252, 10, 30 },
		{ 252, 11, 30 },
		{ 252, 12, 30 },
		{ 252, 13, 30 },
		{ 252, 14, 30 },
		{ 252, 15, 30 },
		{ 252, 16, 30 },
		{ 252, 17, 30 },
		{ 252, 18, 30 },
		{ 252, 19, 30 },
		{ 252, 20, 30 },
		{ 252, 21, 30 },
		{ 252, 22, 30 },
		{ 252, 23, 30 },
		{ 252, 24, 30 },
		{ 252, 25, 30 },
		{ 252, 26, 30 },
		{ 252, 27, 30 },
		{ 252, 28, 30 },
		{ 252, 29, 30 },
		{ 252, 30, 30 },
		{ 252, 31, 30 },
		{ 252, 32, 30 },
		{ 252, 33, 30 },
		{ 252, 34, 30 },
		{ 252, 35, 30 },
		{ 252, 36, 30 },
		{ 252, 37, 30 },
		{ 252, 38, 30 },
		{ 252, 39, 30 },
		{ 252, 40, 30 },
		{ 252, 41, 30 },
		{ 252, 42, 30 },
		{ 252, 43, 30 },
		{ 252, 44, 30 },
		{ 252, 45, 30 },
		{ 252, 46, 30 },
		{ 252, 47, 30 },
		{ 252, 48, 30 },
		{ 252, 49, 30 },
		{ 252, 50, 30 },
		{ 252, 51, 30 },
		{ 252, 52, 30 },
		{ 252, 53, 30 },
		{ 252, 54, 30 },
		{ 252, 55, 30 },
		{ 252, 56, 30 },
		{ 252, 57, 30 },
		{ 252, 58, 30 },
		{ 252, 59, 30 },
		{ 252, 60, 30 },
		{ 252, 61, 30 },
		{ 252, 62, 30 },
		{ 252, 63, 30 },
		{ 252, 64, 30 },
		{ 252, 65, 30 },
		{ 252, 66, 30 },
		{ 252, 67, 30 },
		{ 252, 68, 30 },
		{ 252, 69, 30 },
		{ 252, 70, 30 },
		{ 252, 71, 30 },
		{ 252, 72, 30 },
		{ 252, 73, 30 },
		{ 252, 74, 30 },
		{ 252, 75, 30 },
		{ 252, 76, 30 },
		{ 252, 77, 30 },
		{ 252, 78, 30 },
		{ 252, 79, 30 },
		{ 252, 80, 30 },
		{ 252, 81, 30 },
		{ 252, 82, 30 },
		{ 252, 83, 30 },
		{ 252, 84, 30 },
		{ 252, 85, 30 },
		{ 252, 86, 30 },
		{ 252, 87, 30 },
		{ 252, 88, 30 },
		{ 252, 89, 30 },
		{ 252, 90, 30 },
		{ 252, 91, 30 },
		{ 252, 92, 30 },
		{ 252, 93, 30 },
		{ 252, 94, 30 },
		{ 252, 95, 30 },
		{ 252, 96, 30 },
		{ 252, 97, 30 },
		{ 252, 98, 30 },
		{ 252, 99, 30 },
		{ 252, 100, 30 },
		{ 252, 101, 30 },
		{ 252, 102, 30 },
		{ 252, 103, 30 },
		{ 252, 104, 30 },
		{ 252, 105, 30 },
		{ 252, 106, 30 },
		{ 252, 107, 30 },
		{ 252, 108, 30 },
		{ 252, 109, 30 },
		{ 252, 110, 30 },
		{ 252, 111, 30 },
		{ 252, 112, 30 },
		{ 252, 113, 30 },
		{ 252, 114, 30 },
		{ 252, 115, 30 },
		{ 252, 116, 30 },
		{ 252, 117, 30 },
		{ 252, 118, 30 },
		{ 252, 119, 30 },
		{ 252, 120, 30 },
		{ 252, 121, 30 },
		{ 252, 122, 30 },
		{ 252, 123, 30 },
		{ 252, 124, 30 },
		{ 252, 125, 30 },
		{ 252, 126, 30 },
		{ 252, 127, 30 },
		{ 252, 128, 30 },
		{ 252, 129, 30 },
		{ 252, 130, 30 },
		{ 252, 131, 30 },
		{ 252, 132, 30 },
		{ 252, 133, 30 },
		{ 252, 134, 30 },
		{ 252, 135, 30 },
		{ 252, 136, 30 },
		{ 252, 137, 30 },
		{ 252, 138, 30 },
		{ 252, 139, 30 },
		{ 252, 140, 30 },
		{ 252, 141, 30 },
		{ 252, 142, 30 },
		{ 252, 143, 30 },
		{ 252, 144, 30 },
		{ 252, 145, 30 },
		{ 252, 146, 30 },
		{ 252, 147, 30 },
		{ 252, 148, 30 },
		{ 252, 149, 30 },
		{ 252, 150, 30 },
		{ 252, 151, 30 },
		{ 252, 152, 30 },
		{ 252, 153, 30 },
		{ 252, 154, 30 },
		{ 252, 155, 30 },
		{ 252, 156, 30 },
		{ 252, 157, 30 },
		{ 252, 158, 30 },
		{ 252, 159, 30 },
		{ 252, 160, 30 },
		{ 252, 161, 30 },
		{ 252, 162, 30 },
		{ 252, 163, 30 },
		{ 252, 164, 30 },
		{ 252, 165, 30 },
		{ 252, 166, 30 },
		{ 252, 167, 30 },
		{ 252, 168, 30 },
		{ 252, 169, 30 },
		{ 252, 170, 30 },
		{ 252, 171, 30 },
		{ 252, 172, 30 },
		{ 252, 173, 30 },
		{ 252, 174, 30 },
		{ 252, 175, 30 },
		{ 252, 176, 30 },
		{ 252, 177, 30 },
		{ 252, 178, 30 },
		{ 252, 179, 30 },
		{ 252, 180, 30 },
		{ 252, 181, 30 },
		{ 252, 182, 30 },
		{ 252, 183, 30 },
		{ 252, 184, 30 },
		{ 252, 185, 30 },
		{ 252, 186, 30 },
		{ 252, 187, 30 },
		{ 252, 188, 30 },
		{ 252, 189, 30 },
		{ 252, 190, 30 },
		{ 252, 191, 30 },
		{ 252, 323, 30 },
		{ 252, 324, 30 },
		{ 252, 325, 30 },
		{ 252, 326, 30 },
		{ 252, 327, 30 },
		{ 252, 328, 30 },
		{ 252, 329, 30 },
		{ 252, 330, 30 },
		{ 252, 331, 30 },
		{ 252, 332, 30 },
		{ 252, 333, 30 },
		{ 252, 334, 30 },
		{ 252, 335, 30 },
		{ 252, 336, 30 },
		{ 252, 337, 30 },
		{ 252, 338, 30 },
		{ 252, 339, 30 },
		{ 252, 340, 30 },
		{ 252, 341, 30 },
		{ 252, 342, 30 },
		{ 252, 343, 30 },
		{ 252, 344, 30 },
		{ 252, 345, 30 },
		{ 252, 346, 30 },
		{ 252, 347, 30 },
		{ 252, 348, 30 },
		{ 252, 349, 30 },
		{ 252, 350, 30 },
		{ 252, 351, 30 },
		{ 252, 352, 30 },
		{ 252, 353, 30 },
		{ 252, 354, 30 },
		{ 252, 355, 30 },
		{ 252, 356, 30 },
		{ 252, 357, 30 },
		{ 252, 358, 30 },
		{ 252, 359, 30 },
		{ 252, 360, 30 },
		{ 252, 361, 30 },
		{ 252, 362, 30 },
		{ 252, 363, 30 },
		{ 252, 364, 30 },
		{ 252, 365, 30 },
		{ 252, 366, 30 },
		{ 252, 367, 30 },
		{ 252, 368, 30 },
		{ 252, 369, 30 },
		{ 252, 370, 30 },
		{ 252, 371, 30 },
		{ 252, 372, 30 },
		{ 252, 373, 30 },
		{ 252, 374, 30 },
		{ 252, 375, 30 },
		{ 252, 376, 30 },
		{ 252, 377, 30 },
		{ 252, 378, 30 },
		{ 252, 379, 30 },
		{ 252, 380, 30 },
		{ 252, 381, 30 },
		{ 252, 382, 30 },
		{ 252, 383, 30 },
		{ 252, 384, 30 },
		{ 252, 385, 30 },
		{ 252, 386, 30 },
		{ 252, 387, 30 },
		{ 252, 388, 30 },
		{ 252, 389, 30 },
		{ 252, 390, 30 },
		{ 252, 391, 30 },
		{ 252, 392, 30 },
		{ 252, 393, 30 },
		{ 252, 394, 30 },
		{ 252, 395, 30 },
		{ 252, 396, 30 },
		{ 252, 397, 30 },
		{ 252, 398, 30 },
		{ 252, 399, 30 },
		{ 252, 400, 30 },
		{ 252, 401, 30 },
		{ 252, 402, 30 },
		{ 252, 403, 30 },
		{ 252, 404, 30 },
		{ 252, 405, 30 },
		{ 252, 406, 30 },
		{ 252, 407, 30 },
		{ 252, 408, 30 },
		{ 252, 409, 30 },
		{ 252, 410, 30 },
		{ 252, 411, 30 },
		{ 252, 412, 30 },
		{ 252, 413, 30 },
		{ 252, 414, 30 },
		{ 252, 415, 30 },
		{ 252, 416, 30 },
		{ 252, 417, 30 },
		{ 252, 418, 30 },
		{ 252, 419, 30 },
		{ 252, 420, 30 },
		{ 252, 421, 30 },
		{ 252, 422, 30 },
		{ 252, 423, 30 },
		{ 252, 424, 30 },
		{ 252, 425, 30 },
		{ 252, 426, 30 },
		{ 252, 427, 30 },
		{ 252, 428, 30 },
		{ 252, 429, 30 },
		{ 252, 430, 30 },
		{ 252, 431, 30 },
		{ 252, 432, 30 },
		{ 252, 433, 30 },
		{ 252, 434, 30 },
		{ 252, 435, 30 },
		{ 252, 436, 30 },
		{ 252, 437, 30 },
		{ 252, 438, 30 },
		{ 252, 439, 30 },
		{ 252, 440, 30 },
		{ 252, 441, 30 },
		{ 252, 442, 30 },
		{ 252, 443, 30 },
		{ 252, 444, 30 },
		{ 252, 445, 30 },
		{ 252, 446, 30 },
		{ 252, 447, 30 },
		{ 252, 448, 30 },
		{ 252, 449, 30 },
		{ 252, 450, 30 },
		{ 252, 451, 30 },
		{ 252, 452, 30 },
		{ 252, 453, 30 },
		{ 252, 454, 30 },
		{ 252, 455, 30 },
		{ 252, 456, 30 },
		{ 252, 457, 30 },
		{ 252, 458, 30 },
		{ 252, 459, 30 },
		{ 252, 460, 30 },
		{ 252, 461, 30 },
		{ 252, 462, 30 },
		{ 252, 463, 30 },
		{ 252, 464, 30 },
		{ 252, 465, 30 },
		{ 252, 466, 30 },
		{ 252, 467, 30 },
		{ 252, 468, 30 },
		{ 252, 469, 30 },
		{ 252, 470, 30 },
		{ 252, 471, 30 },
		{ 252, 472, 30 },
		{ 252, 473, 30 },
		{ 252, 474, 30 },
		{ 252, 475, 30 },
		{ 252, 476, 30 },
		{ 252, 477, 30 },
		{ 252, 478, 30 },
		{ 252, 479, 30 },
		{ 252, 480, 30 },
		{ 252, 481, 30 },
		{ 252, 482, 30 },
		{ 252, 483, 30 },
		{ 252, 484, 30 },
		{ 252, 485, 30 },
		{ 252, 486, 30 },
		{ 252, 487, 30 },
		{ 252, 488, 30 },
		{ 252, 489, 30 },
		{ 252, 490, 30 },
		{ 252, 491, 30 },
		{ 252, 492, 30 },
		{ 252, 493, 30 },
		{ 252, 494, 30 },
		{ 252, 495, 30 },
		{ 252, 496, 30 },
		{ 252, 497, 30 },
		{ 252, 498, 30 },
		{ 252, 499, 30 },
		{ 252, 500, 30 },
		{ 252, 501, 30 },
		{ 252, 502, 30 },
		{ 252, 503, 30 },
		{ 252, 504, 30 },
		{ 252, 505, 30 },
		{ 252, 506, 30 },
		{ 252, 507, 30 },
		{ 252, 508, 30 },
		{ 252, 509, 30 },
		{ 252, 510, 30 },
		{ 252, 511, 30 },
		{ 252, 512, 30 },
		{ 253, 1, 30 },
		{ 253, 2, 30 },
		{ 253, 3, 30 },
		{ 253, 4, 30 },
		{ 253, 5, 30 },
		{ 253, 6, 30 },
		{ 253, 7, 30 },
		{ 253, 8, 30 },
		{ 253, 9, 30 },
		{ 253, 10, 30 },
		{ 253, 11, 30 },
		{ 253, 12, 30 },
		{ 253, 13, 30 },
		{ 253, 14, 30 },
		{ 253, 15, 30 },
		{ 253, 16, 30 },
		{ 253, 17, 30 },
		{ 253, 18, 30 },
		{ 253, 19, 30 },
		{ 253, 20, 30 },
		{ 253, 21, 30 },
		{ 253, 22, 30 },
		{ 253, 23, 30 },
		{ 253, 24, 30 },
		{ 253, 25, 30 },
		{ 253, 26, 30 },
		{ 253, 27, 30 },
		{ 253, 28, 30 },
		{ 253, 29, 30 },
		{ 253, 30, 30 },
		{ 253, 31, 30 },
		{ 253, 32, 30 },
		{ 253, 33, 30 },
		{ 253, 34, 30 },
		{ 253, 35, 30 },
		{ 253, 36, 30 },
		{ 253, 37, 30 },
		{ 253, 38, 30 },
		{ 253, 39, 30 },
		{ 253, 40, 30 },
		{ 253, 41, 30 },
		{ 253, 42, 30 },
		{ 253, 43, 30 },
		{ 253, 44, 30 },
		{ 253, 45, 30 },
		{ 253, 46, 30 },
		{ 253, 47, 30 },
		{ 253, 48, 30 },
		{ 253, 49, 30 },
		{ 253, 50, 30 },
		{ 253, 51, 30 },
		{ 253, 52, 30 },
		{ 253, 53, 30 },
		{ 253, 54, 30 },
		{ 253, 55, 30 },
		{ 253, 56, 30 },
		{ 253, 57, 30 },
		{ 253, 58, 30 },
		{ 253, 59, 30 },
		{ 253, 60, 30 },
		{ 253, 61, 30 },
		{ 253, 62, 30 },
		{ 253, 63, 30 },
		{ 253, 64, 30 },
		{ 253, 65, 30 },
		{ 253, 66, 30 },
		{ 253, 67, 30 },
		{ 253, 68, 30 },
		{ 253, 69, 30 },
		{ 253, 70, 30 },
		{ 253, 71, 30 },
		{ 253, 72, 30 },
		{ 253, 73, 30 },
		{ 253, 74, 30 },
		{ 253, 75, 30 },
		{ 253, 76, 30 },
		{ 253, 77, 30 },
		{ 253, 78, 30 },
		{ 253, 79, 30 },
		{ 253, 80, 30 },
		{ 253, 81, 30 },
		{ 253, 82, 30 },
		{ 253, 83, 30 },
		{ 253, 84, 30 },
		{ 253, 85, 30 },
		{ 253, 86, 30 },
		{ 253, 87, 30 },
		{ 253, 88, 30 },
		{ 253, 89, 30 },
		{ 253, 90, 30 },
		{ 253, 91, 30 },
		{ 253, 92, 30 },
		{ 253, 93, 30 },
		{ 253, 94, 30 },
		{ 253, 95, 30 },
		{ 253, 96, 30 },
		{ 253, 97, 30 },
		{ 253, 98, 30 },
		{ 253, 99, 30 },
		{ 253, 100, 30 },
		{ 253, 101, 30 },
		{ 253, 102, 30 },
		{ 253, 103, 30 },
		{ 253, 104, 30 },
		{ 253, 105, 30 },
		{ 253, 106, 30 },
		{ 253, 107, 30 },
		{ 253, 108, 30 },
		{ 253, 109, 30 },
		{ 253, 110, 30 },
		{ 253, 111, 30 },
		{ 253, 112, 30 },
		{ 253, 113, 30 },
		{ 253, 114, 30 },
		{ 253, 115, 30 },
		{ 253, 116, 30 },
		{ 253, 117, 30 },
		{ 253, 118, 30 },
		{ 253, 119, 30 },
		{ 253, 120, 30 },
		{ 253, 121, 30 },
		{ 253, 122, 30 },
		{ 253, 123, 30 },
		{ 253, 124, 30 },
		{ 253, 125, 30 },
		{ 253, 126, 30 },
		{ 253, 127, 30 },
		{ 253, 128, 30 },
		{ 253, 129, 30 },
		{ 253, 130, 30 },
		{ 253, 131, 30 },
		{ 253, 132, 30 },
		{ 253, 133, 30 },
		{ 253, 134, 30 },
		{ 253, 135, 30 },
		{ 253, 136, 30 },
		{ 253, 137, 30 },
		{ 253, 138, 30 },
		{ 253, 139, 30 },
		{ 253, 140, 30 },
		{ 253, 141, 30 },
		{ 253, 142, 30 },
		{ 253, 143, 30 },
		{ 253, 144, 30 },
		{ 253, 145, 30 },
		{ 253, 146, 30 },
		{ 253, 147, 30 },
		{ 253, 148, 30 },
		{ 253, 149, 30 },
		{ 253, 150, 30 },
		{ 253, 151, 30 },
		{ 253, 152, 30 },
		{ 253, 153, 30 },
		{ 253, 154, 30 },
		{ 253, 155, 30 },
		{ 253, 156, 30 },
		{ 253, 157, 30 },
		{ 253, 158, 30 },
		{ 253, 159, 30 },
		{ 253, 160, 30 },
		{ 253, 161, 30 },
		{ 253, 162, 30 },
		{ 253, 163, 30 },
		{ 253, 164, 30 },
		{ 253, 165, 30 },
		{ 253, 166, 30 },
		{ 253, 167, 30 },
		{ 253, 168, 30 },
		{ 253, 169, 30 },
		{ 253, 170, 30 },
		{ 253, 171, 30 },
		{ 253, 172, 30 },
		{ 253, 173, 30 },
		{ 253, 174, 30 },
		{ 253, 175, 30 },
		{ 253, 176, 30 },
		{ 253, 177, 30 },
		{ 253, 178, 30 },
		{ 253, 179, 30 },
		{ 253, 180, 30 },
		{ 253, 181, 30 },
		{ 253, 182, 30 },
		{ 253, 183, 30 },
		{ 253, 184, 30 },
		{ 253, 185, 30 },
		{ 253, 186, 30 },
		{ 253, 187, 30 },
		{ 253, 188, 30 },
		{ 253, 189, 30 },
		{ 253, 190, 30 },
		{ 253, 191, 30 },
		{ 253, 323, 30 },
		{ 253, 324, 30 },
		{ 253, 325, 30 },
		{ 253, 326, 30 },
		{ 253, 327, 30 },
		{ 253, 328, 30 },
		{ 253, 329, 30 },
		{ 253, 330, 30 },
		{ 253, 331, 30 },
		{ 253, 332, 30 },
		{ 253, 333, 30 },
		{ 253, 334, 30 },
		{ 253, 335, 30 },
		{ 253, 336, 30 },
		{ 253, 337, 30 },
		{ 253, 338, 30 },
		{ 253, 339, 30 },
		{ 253, 340, 30 },
		{ 253, 341, 30 },
		{ 253, 342, 30 },
		{ 253, 343, 30 },
		{ 253, 344, 30 },
		{ 253, 345, 30 },
		{ 253, 346, 30 },
		{ 253, 347, 30 },
		{ 253, 348, 30 },
		{ 253, 349, 30 },
		{ 253, 350, 30 },
		{ 253, 351, 30 },
		{ 253, 352, 30 },
		{ 253, 353, 30 },
		{ 253, 354, 30 },
		{ 253, 355, 30 },
		{ 253, 356, 30 },
		{ 253, 357, 30 },
		{ 253, 358, 30 },
		{ 253, 359, 30 },
		{ 253, 360, 30 },
		{ 253, 361, 30 },
		{ 253, 362, 30 },
		{ 253, 363, 30 },
		{ 253, 364, 30 },
		{ 253, 365, 30 },
		{ 253, 366, 30 },
		{ 253, 367, 30 },
		{ 253, 368, 30 },
		{ 253, 369, 30 },
		{ 253, 370, 30 },
		{ 253, 371, 30 },
		{ 253, 372, 30 },
		{ 253, 373, 30 },
		{ 253, 374, 30 },
		{ 253, 375, 30 },
		{ 253, 376, 30 },
		{ 253, 377, 30 },
		{ 253, 378, 30 },
		{ 253, 379, 30 },
		{ 253, 380, 30 },
		{ 253, 381, 30 },
		{ 253, 382, 30 },
		{ 253, 383, 30 },
		{ 253, 384, 30 },
		{ 253, 385, 30 },
		{ 253, 386, 30 },
		{ 253, 387, 30 },
		{ 253, 388, 30 },
		{ 253, 389, 30 },
		{ 253, 390, 30 },
		{ 253, 391, 30 },
		{ 253, 392, 30 },
		{ 253, 393, 30 },
		{ 253, 394, 30 },
		{ 253, 395, 30 },
		{ 253, 396, 30 },
		{ 253, 397, 30 },
		{ 253, 398, 30 },
		{ 253, 399, 30 },
		{ 253, 400, 30 },
		{ 253, 401, 30 },
		{ 253, 402, 30 },
		{ 253, 403, 30 },
		{ 253, 404, 30 },
		{ 253, 405, 30 },
		{ 253, 406, 30 },
		{ 253, 407, 30 },
		{ 253, 408, 30 },
		{ 253, 409, 30 },
		{ 253, 410, 30 },
		{ 253, 411, 30 },
		{ 253, 412, 30 },
		{ 253, 413, 30 },
		{ 253, 414, 30 },
		{ 253, 415, 30 },
		{ 253, 416, 30 },
		{ 253, 417, 30 },
		{ 253, 418, 30 },
		{ 253, 419, 30 },
		{ 253, 420, 30 },
		{ 253, 421, 30 },
		{ 253, 422, 30 },
		{ 253, 423, 30 },
		{ 253, 424, 30 },
		{ 253, 425, 30 },
		{ 253, 426, 30 },
		{ 253, 427, 30 },
		{ 253, 428, 30 },
		{ 253, 429, 30 },
		{ 253, 430, 30 },
		{ 253, 431, 30 },
		{ 253, 432, 30 },
		{ 253, 433, 30 },
		{ 253, 434, 30 },
		{ 253, 435, 30 },
		{ 253, 436, 30 },
		{ 253, 437, 30 },
		{ 253, 438, 30 },
		{ 253, 439, 30 },
		{ 253, 440, 30 },
		{ 253, 441, 30 },
		{ 253, 442, 30 },
		{ 253, 443, 30 },
		{ 253, 444, 30 },
		{ 253, 445, 30 },
		{ 253, 446, 30 },
		{ 253, 447, 30 },
		{ 253, 448, 30 },
		{ 253, 449, 30 },
		{ 253, 450, 30 },
		{ 253, 451, 30 },
		{ 253, 452, 30 },
		{ 253, 453, 30 },
		{ 253, 454, 30 },
		{ 253, 455, 30 },
		{ 253, 456, 30 },
		{ 253, 457, 30 },
		{ 253, 458, 30 },
		{ 253, 459, 30 },
		{ 253, 460, 30 },
		{ 253, 461, 30 },
		{ 253, 462, 30 },
		{ 253, 463, 30 },
		{ 253, 464, 30 },
		{ 253, 465, 30 },
		{ 253, 466, 30 },
		{ 253, 467, 30 },
		{ 253, 468, 30 },
		{ 253, 469, 30 },
		{ 253, 470, 30 },
		{ 253, 471, 30 },
		{ 253, 472, 30 },
		{ 253, 473, 30 },
		{ 253, 474, 30 },
		{ 253, 475, 30 },
		{ 253, 476, 30 },
		{ 253, 477, 30 },
		{ 253, 478, 30 },
		{ 253, 479, 30 },
		{ 253, 480, 30 },
		{ 253, 481, 30 },
		{ 253, 482, 30 },
		{ 253, 483, 30 },
		{ 253, 484, 30 },
		{ 253, 485, 30 },
		{ 253, 486, 30 },
		{ 253, 487, 30 },
		{ 253, 488, 30 },
		{ 253, 489, 30 },
		{ 253, 490, 30 },
		{ 253, 491, 30 },
		{ 253, 492, 30 },
		{ 253, 493, 30 },
		{ 253, 494, 30 },
		{ 253, 495, 30 },
		{ 253, 496, 30 },
		{ 253, 497, 30 },
		{ 253, 498, 30 },
		{ 253, 499, 30 },
		{ 253, 500, 30 },
		{ 253, 501, 30 },
		{ 253, 502, 30 },
		{ 253, 503, 30 },
		{ 253, 504, 30 },
		{ 253, 505, 30 },
		{ 253, 506, 30 },
		{ 253, 507, 30 },
		{ 253, 508, 30 },
		{ 253, 509, 30 },
		{ 253, 510, 30 },
		{ 253, 511, 30 },
		{ 253, 512, 30 },
		{ 254, 1, 30 },
		{ 254, 2, 30 },
		{ 254, 3, 30 },
		{ 254, 4, 30 },
		{ 254, 5, 30 },
		{ 254, 6, 30 },
		{ 254, 7, 30 },
		{ 254, 8, 30 },
		{ 254, 9, 30 },
		{ 254, 10, 30 },
		{ 254, 11, 30 },
		{ 254, 12, 30 },
		{ 254, 13, 30 },
		{ 254, 14, 30 },
		{ 254, 15, 30 },
		{ 254, 16, 30 },
		{ 254, 17, 30 },
		{ 254, 18, 30 },
		{ 254, 19, 30 },
		{ 254, 20, 30 },
		{ 254, 21, 30 },
		{ 254, 22, 30 },
		{ 254, 23, 30 },
		{ 254, 24, 30 },
		{ 254, 25, 30 },
		{ 254, 26, 30 },
		{ 254, 27, 30 },
		{ 254, 28, 30 },
		{ 254, 29, 30 },
		{ 254, 30, 30 },
		{ 254, 31, 30 },
		{ 254, 32, 30 },
		{ 254, 33, 30 },
		{ 254, 34, 30 },
		{ 254, 35, 30 },
		{ 254, 36, 30 },
		{ 254, 37, 30 },
		{ 254, 38, 30 },
		{ 254, 39, 30 },
		{ 254, 40, 30 },
		{ 254, 41, 30 },
		{ 254, 42, 30 },
		{ 254, 43, 30 },
		{ 254, 44, 30 },
		{ 254, 45, 30 },
		{ 254, 46, 30 },
		{ 254, 47, 30 },
		{ 254, 48, 30 },
		{ 254, 49, 30 },
		{ 254, 50, 30 },
		{ 254, 51, 30 },
		{ 254, 52, 30 },
		{ 254, 53, 30 },
		{ 254, 54, 30 },
		{ 254, 55, 30 },
		{ 254, 56, 30 },
		{ 254, 57, 30 },
		{ 254, 58, 30 },
		{ 254, 59, 30 },
		{ 254, 60, 30 },
		{ 254, 61, 30 },
		{ 254, 62, 30 },
		{ 254, 63, 30 },
		{ 254, 64, 30 },
		{ 254, 65, 30 },
		{ 254, 66, 30 },
		{ 254, 67, 30 },
		{ 254, 68, 30 },
		{ 254, 69, 30 },
		{ 254, 70, 30 },
		{ 254, 71, 30 },
		{ 254, 72, 30 },
		{ 254, 73, 30 },
		{ 254, 74, 30 },
		{ 254, 75, 30 },
		{ 254, 76, 30 },
		{ 254, 77, 30 },
		{ 254, 78, 30 },
		{ 254, 79, 30 },
		{ 254, 80, 30 },
		{ 254, 81, 30 },
		{ 254, 82, 30 },
		{ 254, 83, 30 },
		{ 254, 84, 30 },
		{ 254, 85, 30 },
		{ 254, 86, 30 },
		{ 254, 87, 30 },
		{ 254, 88, 30 },
		{ 254, 89, 30 },
		{ 254, 90, 30 },
		{ 254, 91, 30 },
		{ 254, 92, 30 },
		{ 254, 93, 30 },
		{ 254, 94, 30 },
		{ 254, 95, 30 },
		{ 254, 96, 30 },
		{ 254, 97, 30 },
		{ 254, 98, 30 },
		{ 254, 99, 30 },
		{ 254, 100, 30 },
		{ 254, 101, 30 },
		{ 254, 102, 30 },
		{ 254, 103, 30 },
		{ 254, 104, 30 },
		{ 254, 105, 30 },
		{ 254, 106, 30 },
		{ 254, 107, 30 },
		{ 254, 108, 30 },
		{ 254, 109, 30 },
		{ 254, 110, 30 },
		{ 254, 111, 30 },
		{ 254, 112, 30 },
		{ 254, 113, 30 },
		{ 254, 114, 30 },
		{ 254, 115, 30 },
		{ 254, 116, 30 },
		{ 254, 117, 30 },
		{ 254, 118, 30 },
		{ 254, 119, 30 },
		{ 254, 120, 30 },
		{ 254, 121, 30 },
		{ 254, 122, 30 },
		{ 254, 123, 30 },
		{ 254, 124, 30 },
		{ 254, 125, 30 },
		{ 254, 126, 30 },
		{ 254, 127, 30 },
		{ 254, 128, 30 },
		{ 254, 129, 30 },
		{ 254, 130, 30 },
		{ 254, 131, 30 },
		{ 254, 132, 30 },
		{ 254, 133, 30 },
		{ 254, 134, 30 },
		{ 254, 135, 30 },
		{ 254, 136, 30 },
		{ 254, 137, 30 },
		{ 254, 138, 30 },
		{ 254, 139, 30 },
		{ 254, 140, 30 },
		{ 254, 141, 30 },
		{ 254, 142, 30 },
		{ 254, 143, 30 },
		{ 254, 144, 30 },
		{ 254, 145, 30 },
		{ 254, 146, 30 },
		{ 254, 147, 30 },
		{ 254, 148, 30 },
		{ 254, 149, 30 },
		{ 254, 150, 30 },
		{ 254, 151, 30 },
		{ 254, 152, 30 },
		{ 254, 153, 30 },
		{ 254, 154, 30 },
		{ 254, 155, 30 },
		{ 254, 156, 30 },
		{ 254, 157, 30 },
		{ 254, 158, 30 },
		{ 254, 159, 30 },
		{ 254, 160, 30 },
		{ 254, 161, 30 },
		{ 254, 162, 30 },
		{ 254, 163, 30 },
		{ 254, 164, 30 },
		{ 254, 165, 30 },
		{ 254, 166, 30 },
		{ 254, 167, 30 },
		{ 254, 168, 30 },
		{ 254, 169, 30 },
		{ 254, 170, 30 },
		{ 254, 171, 30 },
		{ 254, 172, 30 },
		{ 254, 173, 30 },
		{ 254, 174, 30 },
		{ 254, 175, 30 },
		{ 254, 176, 30 },
		{ 254, 177, 30 },
		{ 254, 178, 30 },
		{ 254, 179, 30 },
		{ 254, 180, 30 },
		{ 254, 181, 30 },
		{ 254, 182, 30 },
		{ 254, 183, 30 },
		{ 254, 184, 30 },
		{ 254, 185, 30 },
		{ 254, 186, 30 },
		{ 254, 187, 30 },
		{ 254, 188, 30 },
		{ 254, 189, 30 },
		{ 254, 190, 30 },
		{ 254, 191, 30 },
		{ 254, 323, 30 },
		{ 254, 324, 30 },
		{ 254, 325, 30 },
		{ 254, 326, 30 },
		{ 254, 327, 30 },
		{ 254, 328, 30 },
		{ 254, 329, 30 },
		{ 254, 330, 30 },
		{ 254, 331, 30 },
		{ 254, 332, 30 },
		{ 254, 333, 30 },
		{ 254, 334, 30 },
		{ 254, 335, 30 },
		{ 254, 336, 30 },
		{ 254, 337, 30 },
		{ 254, 338, 30 },
		{ 254, 339, 30 },
		{ 254, 340, 30 },
		{ 254, 341, 30 },
		{ 254, 342, 30 },
		{ 254, 343, 30 },
		{ 254, 344, 30 },
		{ 254, 345, 30 },
		{ 254, 346, 30 },
		{ 254, 347, 30 },
		{ 254, 348, 30 },
		{ 254, 349, 30 },
		{ 254, 350, 30 },
		{ 254, 351, 30 },
		{ 254, 352, 30 },
		{ 254, 353, 30 },
		{ 254, 354, 30 },
		{ 254, 355, 30 },
		{ 254, 356, 30 },
		{ 254, 357, 30 },
		{ 254, 358, 30 },
		{ 254, 359, 30 },
		{ 254, 360, 30 },
		{ 254, 361, 30 },
		{ 254, 362, 30 },
		{ 254, 363, 30 },
		{ 254, 364, 30 },
		{ 254, 365, 30 },
		{ 254, 366, 30 },
		{ 254, 367, 30 },
		{ 254, 368, 30 },
		{ 254, 369, 30 },
		{ 254, 370, 30 },
		{ 254, 371, 30 },
		{ 254, 372, 30 },
		{ 254, 373, 30 },
		{ 254, 374, 30 },
		{ 254, 375, 30 },
		{ 254, 376, 30 },
		{ 254, 377, 30 },
		{ 254, 378, 30 },
		{ 254, 379, 30 },
		{ 254, 380, 30 },
		{ 254, 381, 30 },
		{ 254, 382, 30 },
		{ 254, 383, 30 },
		{ 254, 384, 30 },
		{ 254, 385, 30 },
		{ 254, 386, 30 },
		{ 254, 387, 30 },
		{ 254, 388, 30 },
		{ 254, 389, 30 },
		{ 254, 390, 30 },
		{ 254, 391, 30 },
		{ 254, 392, 30 },
		{ 254, 393, 30 },
		{ 254, 394, 30 },
		{ 254, 395, 30 },
		{ 254, 396, 30 },
		{ 254, 397, 30 },
		{ 254, 398, 30 },
		{ 254, 399, 30 },
		{ 254, 400, 30 },
		{ 254, 401, 30 },
		{ 254, 402, 30 },
		{ 254, 403, 30 },
		{ 254, 404, 30 },
		{ 254, 405, 30 },
		{ 254, 406, 30 },
		{ 254, 407, 30 },
		{ 254, 408, 30 },
		{ 254, 409, 30 },
		{ 254, 410, 30 },
		{ 254, 411, 30 },
		{ 254, 412, 30 },
		{ 254, 413, 30 },
		{ 254, 414, 30 },
		{ 254, 415, 30 },
		{ 254, 416, 30 },
		{ 254, 417, 30 },
		{ 254, 418, 30 },
		{ 254, 419, 30 },
		{ 254, 420, 30 },
		{ 254, 421, 30 },
		{ 254, 422, 30 },
		{ 254, 423, 30 },
		{ 254, 424, 30 },
		{ 254, 425, 30 },
		{ 254, 426, 30 },
		{ 254, 427, 30 },
		{ 254, 428, 30 },
		{ 254, 429, 30 },
		{ 254, 430, 30 },
		{ 254, 431, 30 },
		{ 254, 432, 30 },
		{ 254, 433, 30 },
		{ 254, 434, 30 },
		{ 254, 435, 30 },
		{ 254, 436, 30 },
		{ 254, 437, 30 },
		{ 254, 438, 30 },
		{ 254, 439, 30 },
		{ 254, 440, 30 },
		{ 254, 441, 30 },
		{ 254, 442, 30 },
		{ 254, 443, 30 },
		{ 254, 444, 30 },
		{ 254, 445, 30 },
		{ 254, 446, 30 },
		{ 254, 447, 30 },
		{ 254, 448, 30 },
		{ 254, 449, 30 },
		{ 254, 450, 30 },
		{ 254, 451, 30 },
		{ 254, 452, 30 },
		{ 254, 453, 30 },
		{ 254, 454, 30 },
		{ 254, 455, 30 },
		{ 254, 456, 30 },
		{ 254, 457, 30 },
		{ 254, 458, 30 },
		{ 254, 459, 30 },
		{ 254, 460, 30 },
		{ 254, 461, 30 },
		{ 254, 462, 30 },
		{ 254, 463, 30 },
		{ 254, 464, 30 },
		{ 254, 465, 30 },
		{ 254, 466, 30 },
		{ 254, 467, 30 },
		{ 254, 468, 30 },
		{ 254, 469, 30 },
		{ 254, 470, 30 },
		{ 254, 471, 30 },
		{ 254, 472, 30 },
		{ 254, 473, 30 },
		{ 254, 474, 30 },
		{ 254, 475, 30 },
		{ 254, 476, 30 },
		{ 254, 477, 30 },
		{ 254, 478, 30 },
		{ 254, 479, 30 },
		{ 254, 480, 30 },
		{ 254, 481, 30 },
		{ 254, 482, 30 },
		{ 254, 483, 30 },
		{ 254, 484, 30 },
		{ 254, 485, 30 },
		{ 254, 486, 30 },
		{ 254, 487, 30 },
		{ 254, 488, 30 },
		{ 254, 489, 30 },
		{ 254, 490, 30 },
		{ 254, 491, 30 },
		{ 254, 492, 30 },
		{ 254, 493, 30 },
		{ 254, 494, 30 },
		{ 254, 495, 30 },
		{ 254, 496, 30 },
		{ 254, 497, 30 },
		{ 254, 498, 30 },
		{ 254, 499, 30 },
		{ 254, 500, 30 },
		{ 254, 501, 30 },
		{ 254, 502, 30 },
		{ 254, 503, 30 },
		{ 254, 504, 30 },
		{ 254, 505, 30 },
		{ 254, 506, 30 },
		{ 254, 507, 30 },
		{ 254, 508, 30 },
		{ 254, 509, 30 },
		{ 254, 510, 30 },
		{ 254, 511, 30 },
		{ 254, 512, 30 },
		{ 255, 1, 30 },
		{ 255, 2, 30 },
		{ 255, 3, 30 },
		{ 255, 4, 30 },
		{ 255, 5, 30 },
		{ 255, 6, 30 },
		{ 255, 7, 30 },
		{ 255, 8, 30 },
		{ 255, 9, 30 },
		{ 255, 10, 30 },
		{ 255, 11, 30 },
		{ 255, 12, 30 },
		{ 255, 13, 30 },
		{ 255, 14, 30 },
		{ 255, 15, 30 },
		{ 255, 16, 30 },
		{ 255, 17, 30 },
		{ 255, 18, 30 },
		{ 255, 19, 30 },
		{ 255, 20, 30 },
		{ 255, 21, 30 },
		{ 255, 22, 30 },
		{ 255, 23, 30 },
		{ 255, 24, 30 },
		{ 255, 25, 30 },
		{ 255, 26, 30 },
		{ 255, 27, 30 },
		{ 255, 28, 30 },
		{ 255, 29, 30 },
		{ 255, 30, 30 },
		{ 255, 31, 30 },
		{ 255, 32, 30 },
		{ 255, 33, 30 },
		{ 255, 34, 30 },
		{ 255, 35, 30 },
		{ 255, 36, 30 },
		{ 255, 37, 30 },
		{ 255, 38, 30 },
		{ 255, 39, 30 },
		{ 255, 40, 30 },
		{ 255, 41, 30 },
		{ 255, 42, 30 },
		{ 255, 43, 30 },
		{ 255, 44, 30 },
		{ 255, 45, 30 },
		{ 255, 46, 30 },
		{ 255, 47, 30 },
		{ 255, 48, 30 },
		{ 255, 49, 30 },
		{ 255, 50, 30 },
		{ 255, 51, 30 },
		{ 255, 52, 30 },
		{ 255, 53, 30 },
		{ 255, 54, 30 },
		{ 255, 55, 30 },
		{ 255, 56, 30 },
		{ 255, 57, 30 },
		{ 255, 58, 30 },
		{ 255, 59, 30 },
		{ 255, 60, 30 },
		{ 255, 61, 30 },
		{ 255, 62, 30 },
		{ 255, 63, 30 },
		{ 255, 64, 30 },
		{ 255, 65, 30 },
		{ 255, 66, 30 },
		{ 255, 67, 30 },
		{ 255, 68, 30 },
		{ 255, 69, 30 },
		{ 255, 70, 30 },
		{ 255, 71, 30 },
		{ 255, 72, 30 },
		{ 255, 73, 30 },
		{ 255, 74, 30 },
		{ 255, 75, 30 },
		{ 255, 76, 30 },
		{ 255, 77, 30 },
		{ 255, 78, 30 },
		{ 255, 79, 30 },
		{ 255, 80, 30 },
		{ 255, 81, 30 },
		{ 255, 82, 30 },
		{ 255, 83, 30 },
		{ 255, 84, 30 },
		{ 255, 85, 30 },
		{ 255, 86, 30 },
		{ 255, 87, 30 },
		{ 255, 88, 30 },
		{ 255, 89, 30 },
		{ 255, 90, 30 },
		{ 255, 91, 30 },
		{ 255, 92, 30 },
		{ 255, 93, 30 },
		{ 255, 94, 30 },
		{ 255, 95, 30 },
		{ 255, 96, 30 },
		{ 255, 97, 30 },
		{ 255, 98, 30 },
		{ 255, 99, 30 },
		{ 255, 100, 30 },
		{ 255, 101, 30 },
		{ 255, 102, 30 },
		{ 255, 103, 30 },
		{ 255, 104, 30 },
		{ 255, 105, 30 },
		{ 255, 106, 30 },
		{ 255, 107, 30 },
		{ 255, 108, 30 },
		{ 255, 109, 30 },
		{ 255, 110, 30 },
		{ 255, 111, 30 },
		{ 255, 112, 30 },
		{ 255, 113, 30 },
		{ 255, 114, 30 },
		{ 255, 115, 30 },
		{ 255, 116, 30 },
		{ 255, 117, 30 },
		{ 255, 118, 30 },
		{ 255, 119, 30 },
		{ 255, 120, 30 },
		{ 255, 121, 30 },
		{ 255, 122, 30 },
		{ 255, 123, 30 },
		{ 255, 124, 30 },
		{ 255, 125, 30 },
		{ 255, 126, 30 },
		{ 255, 127, 30 },
		{ 255, 128, 30 },
		{ 255, 129, 30 },
		{ 255, 130, 30 },
		{ 255, 131, 30 },
		{ 255, 132, 30 },
		{ 255, 133, 30 },
		{ 255, 134, 30 },
		{ 255, 135, 30 },
		{ 255, 136, 30 },
		{ 255, 137, 30 },
		{ 255, 138, 30 },
		{ 255, 139, 30 },
		{ 255, 140, 30 },
		{ 255, 141, 30 },
		{ 255, 142, 30 },
		{ 255, 143, 30 },
		{ 255, 144, 30 },
		{ 255, 145, 30 },
		{ 255, 146, 30 },
		{ 255, 147, 30 },
		{ 255, 148, 30 },
		{ 255, 149, 30 },
		{ 255, 150, 30 },
		{ 255, 151, 30 },
		{ 255, 152, 30 },
		{ 255, 153, 30 },
		{ 255, 154, 30 },
		{ 255, 155, 30 },
		{ 255, 156, 30 },
		{ 255, 157, 30 },
		{ 255, 158, 30 },
		{ 255, 159, 30 },
		{ 255, 160, 30 },
		{ 255, 161, 30 },
		{ 255, 162, 30 },
		{ 255, 163, 30 },
		{ 255, 164, 30 },
		{ 255, 165, 30 },
		{ 255, 166, 30 },
		{ 255, 167, 30 },
		{ 255, 168, 30 },
		{ 255, 169, 30 },
		{ 255, 170, 30 },
		{ 255, 171, 30 },
		{ 255, 172, 30 },
		{ 255, 173, 30 },
		{ 255, 174, 30 },
		{ 255, 175, 30 },
		{ 255, 176, 30 },
		{ 255, 177, 30 },
		{ 255, 178, 30 },
		{ 255, 179, 30 },
		{ 255, 180, 30 },
		{ 255, 181, 30 },
		{ 255, 182, 30 },
		{ 255, 183, 30 },
		{ 255, 184, 30 },
		{ 255, 185, 30 },
		{ 255, 186, 30 },
		{ 255, 187, 30 },
		{ 255, 188, 30 },
		{ 255, 189, 30 },
		{ 255, 190, 30 },
		{ 255, 191, 30 },
		{ 255, 323, 30 },
		{ 255, 324, 30 },
		{ 255, 325, 30 },
		{ 255, 326, 30 },
		{ 255, 327, 30 },
		{ 255, 328, 30 },
		{ 255, 329, 30 },
		{ 255, 330, 30 },
		{ 255, 331, 30 },
		{ 255, 332, 30 },
		{ 255, 333, 30 },
		{ 255, 334, 30 },
		{ 255, 335, 30 },
		{ 255, 336, 30 },
		{ 255, 337, 30 },
		{ 255, 338, 30 },
		{ 255, 339, 30 },
		{ 255, 340, 30 },
		{ 255, 341, 30 },
		{ 255, 342, 30 },
		{ 255, 343, 30 },
		{ 255, 344, 30 },
		{ 255, 345, 30 },
		{ 255, 346, 30 },
		{ 255, 347, 30 },
		{ 255, 348, 30 },
		{ 255, 349, 30 },
		{ 255, 350, 30 },
		{ 255, 351, 30 },
		{ 255, 352, 30 },
		{ 255, 353, 30 },
		{ 255, 354, 30 },
		{ 255, 355, 30 },
		{ 255, 356, 30 },
		{ 255, 357, 30 },
		{ 255, 358, 30 },
		{ 255, 359, 30 },
		{ 255, 360, 30 },
		{ 255, 361, 30 },
		{ 255, 362, 30 },
		{ 255, 363, 30 },
		{ 255, 364, 30 },
		{ 255, 365, 30 },
		{ 255, 366, 30 },
		{ 255, 367, 30 },
		{ 255, 368, 30 },
		{ 255, 369, 30 },
		{ 255, 370, 30 },
		{ 255, 371, 30 },
		{ 255, 372, 30 },
		{ 255, 373, 30 },
		{ 255, 374, 30 },
		{ 255, 375, 30 },
		{ 255, 376, 30 },
		{ 255, 377, 30 },
		{ 255, 378, 30 },
		{ 255, 379, 30 },
		{ 255, 380, 30 },
		{ 255, 381, 30 },
		{ 255, 382, 30 },
		{ 255, 383, 30 },
		{ 255, 384, 30 },
		{ 255, 385, 30 },
		{ 255, 386, 30 },
		{ 255, 387, 30 },
		{ 255, 388, 30 },
		{ 255, 389, 30 },
		{ 255, 390, 30 },
		{ 255, 391, 30 },
		{ 255, 392, 30 },
		{ 255, 393, 30 },
		{ 255, 394, 30 },
		{ 255, 395, 30 },
		{ 255, 396, 30 },
		{ 255, 397, 30 },
		{ 255, 398, 30 },
		{ 255, 399, 30 },
		{ 255, 400, 30 },
		{ 255, 401, 30 },
		{ 255, 402, 30 },
		{ 255, 403, 30 },
		{ 255, 404, 30 },
		{ 255, 405, 30 },
		{ 255, 406, 30 },
		{ 255, 407, 30 },
		{ 255, 408, 30 },
		{ 255, 409, 30 },
		{ 255, 410, 30 },
		{ 255, 411, 30 },
		{ 255, 412, 30 },
		{ 255, 413, 30 },
		{ 255, 414, 30 },
		{ 255, 415, 30 },
		{ 255, 416, 30 },
		{ 255, 417, 30 },
		{ 255, 418, 30 },
		{ 255, 419, 30 },
		{ 255, 420, 30 },
		{ 255, 421, 30 },
		{ 255, 422, 30 },
		{ 255, 423, 30 },
		{ 255, 424, 30 },
		{ 255, 425, 30 },
		{ 255, 426, 30 },
		{ 255, 427, 30 },
		{ 255, 428, 30 },
		{ 255, 429, 30 },
		{ 255, 430, 30 },
		{ 255, 431, 30 },
		{ 255, 432, 30 },
		{ 255, 433, 30 },
		{ 255, 434, 30 },
		{ 255, 435, 30 },
		{ 255, 436, 30 },
		{ 255, 437, 30 },
		{ 255, 438, 30 },
		{ 255, 439, 30 },
		{ 255, 440, 30 },
		{ 255, 441, 30 },
		{ 255, 442, 30 },
		{ 255, 443, 30 },
		{ 255, 444, 30 },
		{ 255, 445, 30 },
		{ 255, 446, 30 },
		{ 255, 447, 30 },
		{ 255, 448, 30 },
		{ 255, 449, 30 },
		{ 255, 450, 30 },
		{ 255, 451, 30 },
		{ 255, 452, 30 },
		{ 255, 453, 30 },
		{ 255, 454, 30 },
		{ 255, 455, 30 },
		{ 255, 456, 30 },
		{ 255, 457, 30 },
		{ 255, 458, 30 },
		{ 255, 459, 30 },
		{ 255, 460, 30 },
		{ 255, 461, 30 },
		{ 255, 462, 30 },
		{ 255, 463, 30 },
		{ 255, 464, 30 },
		{ 255, 465, 30 },
		{ 255, 466, 30 },
		{ 255, 467, 30 },
		{ 255, 468, 30 },
		{ 255, 469, 30 },
		{ 255, 470, 30 },
		{ 255, 471, 30 },
		{ 255, 472, 30 },
		{ 255, 473, 30 },
		{ 255, 474, 30 },
		{ 255, 475, 30 },
		{ 255, 476, 30 },
		{ 255, 477, 30 },
		{ 255, 478, 30 },
		{ 255, 479, 30 },
		{ 255, 480, 30 },
		{ 255, 481, 30 },
		{ 255, 482, 30 },
		{ 255, 483, 30 },
		{ 255, 484, 30 },
		{ 255, 485, 30 },
		{ 255, 486, 30 },
		{ 255, 487, 30 },
		{ 255, 488, 30 },
		{ 255, 489, 30 },
		{ 255, 490, 30 },
		{ 255, 491, 30 },
		{ 255, 492, 30 },
		{ 255, 493, 30 },
		{ 255, 494, 30 },
		{ 255, 495, 30 },
		{ 255, 496, 30 },
		{ 255, 497, 30 },
		{ 255, 498, 30 },
		{ 255, 499, 30 },
		{ 255, 500, 30 },
		{ 255, 501, 30 },
		{ 255, 502, 30 },
		{ 255, 503, 30 },
		{ 255, 504, 30 },
		{ 255, 505, 30 },
		{ 255, 506, 30 },
		{ 255, 507, 30 },
		{ 255, 508, 30 },
		{ 255, 509, 30 },
		{ 255, 510, 30 },
		{ 255, 511, 30 },
		{ 255, 512, 30 },
		{ 256, 1, 30 },
		{ 256, 2, 30 },
		{ 256, 3, 30 },
		{ 256, 4, 30 },
		{ 256, 5, 30 },
		{ 256, 6, 30 },
		{ 256, 7, 30 },
		{ 256, 8, 30 },
		{ 256, 9, 30 },
		{ 256, 10, 30 },
		{ 256, 11, 30 },
		{ 256, 12, 30 },
		{ 256, 13, 30 },
		{ 256, 14, 30 },
		{ 256, 15, 30 },
		{ 256, 16, 30 },
		{ 256, 17, 30 },
		{ 256, 18, 30 },
		{ 256, 19, 30 },
		{ 256, 20, 30 },
		{ 256, 21, 30 },
		{ 256, 22, 30 },
		{ 256, 23, 30 },
		{ 256, 24, 30 },
		{ 256, 25, 30 },
		{ 256, 26, 30 },
		{ 256, 27, 30 },
		{ 256, 28, 30 },
		{ 256, 29, 30 },
		{ 256, 30, 30 },
		{ 256, 31, 30 },
		{ 256, 32, 30 },
		{ 256, 33, 30 },
		{ 256, 34, 30 },
		{ 256, 35, 30 },
		{ 256, 36, 30 },
		{ 256, 37, 30 },
		{ 256, 38, 30 },
		{ 256, 39, 30 },
		{ 256, 40, 30 },
		{ 256, 41, 30 },
		{ 256, 42, 30 },
		{ 256, 43, 30 },
		{ 256, 44, 30 },
		{ 256, 45, 30 },
		{ 256, 46, 30 },
		{ 256, 47, 30 },
		{ 256, 48, 30 },
		{ 256, 49, 30 },
		{ 256, 50, 30 },
		{ 256, 51, 30 },
		{ 256, 52, 30 },
		{ 256, 53, 30 },
		{ 256, 54, 30 },
		{ 256, 55, 30 },
		{ 256, 56, 30 },
		{ 256, 57, 30 },
		{ 256, 58, 30 },
		{ 256, 59, 30 },
		{ 256, 60, 30 },
		{ 256, 61, 30 },
		{ 256, 62, 30 },
		{ 256, 63, 30 },
		{ 256, 64, 30 },
		{ 256, 65, 30 },
		{ 256, 66, 30 },
		{ 256, 67, 30 },
		{ 256, 68, 30 },
		{ 256, 69, 30 },
		{ 256, 70, 30 },
		{ 256, 71, 30 },
		{ 256, 72, 30 },
		{ 256, 73, 30 },
		{ 256, 74, 30 },
		{ 256, 75, 30 },
		{ 256, 76, 30 },
		{ 256, 77, 30 },
		{ 256, 78, 30 },
		{ 256, 79, 30 },
		{ 256, 80, 30 },
		{ 256, 81, 30 },
		{ 256, 82, 30 },
		{ 256, 83, 30 },
		{ 256, 84, 30 },
		{ 256, 85, 30 },
		{ 256, 86, 30 },
		{ 256, 87, 30 },
		{ 256, 88, 30 },
		{ 256, 89, 30 },
		{ 256, 90, 30 },
		{ 256, 91, 30 },
		{ 256, 92, 30 },
		{ 256, 93, 30 },
		{ 256, 94, 30 },
		{ 256, 95, 30 },
		{ 256, 96, 30 },
		{ 256, 97, 30 },
		{ 256, 98, 30 },
		{ 256, 99, 30 },
		{ 256, 100, 30 },
		{ 256, 101, 30 },
		{ 256, 102, 30 },
		{ 256, 103, 30 },
		{ 256, 104, 30 },
		{ 256, 105, 30 },
		{ 256, 106, 30 },
		{ 256, 107, 30 },
		{ 256, 108, 30 },
		{ 256, 109, 30 },
		{ 256, 110, 30 },
		{ 256, 111, 30 },
		{ 256, 112, 30 },
		{ 256, 113, 30 },
		{ 256, 114, 30 },
		{ 256, 115, 30 },
		{ 256, 116, 30 },
		{ 256, 117, 30 },
		{ 256, 118, 30 },
		{ 256, 119, 30 },
		{ 256, 120, 30 },
		{ 256, 121, 30 },
		{ 256, 122, 30 },
		{ 256, 123, 30 },
		{ 256, 124, 30 },
		{ 256, 125, 30 },
		{ 256, 126, 30 },
		{ 256, 127, 30 },
		{ 256, 128, 30 },
		{ 256, 129, 30 },
		{ 256, 130, 30 },
		{ 256, 131, 30 },
		{ 256, 132, 30 },
		{ 256, 133, 30 },
		{ 256, 134, 30 },
		{ 256, 135, 30 },
		{ 256, 136, 30 },
		{ 256, 137, 30 },
		{ 256, 138, 30 },
		{ 256, 139, 30 },
		{ 256, 140, 30 },
		{ 256, 141, 30 },
		{ 256, 142, 30 },
		{ 256, 143, 30 },
		{ 256, 144, 30 },
		{ 256, 145, 30 },
		{ 256, 146, 30 },
		{ 256, 147, 30 },
		{ 256, 148, 30 },
		{ 256, 149, 30 },
		{ 256, 150, 30 },
		{ 256, 151, 30 },
		{ 256, 152, 30 },
		{ 256, 153, 30 },
		{ 256, 154, 30 },
		{ 256, 155, 30 },
		{ 256, 156, 30 },
		{ 256, 157, 30 },
		{ 256, 158, 30 },
		{ 256, 159, 30 },
		{ 256, 160, 30 },
		{ 256, 161, 30 },
		{ 256, 162, 30 },
		{ 256, 163, 30 },
		{ 256, 164, 30 },
		{ 256, 165, 30 },
		{ 256, 166, 30 },
		{ 256, 167, 30 },
		{ 256, 168, 30 },
		{ 256, 169, 30 },
		{ 256, 170, 30 },
		{ 256, 171, 30 },
		{ 256, 172, 30 },
		{ 256, 173, 30 },
		{ 256, 174, 30 },
		{ 256, 175, 30 },
		{ 256, 176, 30 },
		{ 256, 177, 30 },
		{ 256, 178, 30 },
		{ 256, 179, 30 },
		{ 256, 180, 30 },
		{ 256, 181, 30 },
		{ 256, 182, 30 },
		{ 256, 183, 30 },
		{ 256, 184, 30 },
		{ 256, 185, 30 },
		{ 256, 186, 30 },
		{ 256, 187, 30 },
		{ 256, 188, 30 },
		{ 256, 189, 30 },
		{ 256, 190, 30 },
		{ 256, 191, 30 },
		{ 256, 323, 30 },
		{ 256, 324, 30 },
		{ 256, 325, 30 },
		{ 256, 326, 30 },
		{ 256, 327, 30 },
		{ 256, 328, 30 },
		{ 256, 329, 30 },
		{ 256, 330, 30 },
		{ 256, 331, 30 },
		{ 256, 332, 30 },
		{ 256, 333, 30 },
		{ 256, 334, 30 },
		{ 256, 335, 30 },
		{ 256, 336, 30 },
		{ 256, 337, 30 },
		{ 256, 338, 30 },
		{ 256, 339, 30 },
		{ 256, 340, 30 },
		{ 256, 341, 30 },
		{ 256, 342, 30 },
		{ 256, 343, 30 },
		{ 256, 344, 30 },
		{ 256, 345, 30 },
		{ 256, 346, 30 },
		{ 256, 347, 30 },
		{ 256, 348, 30 },
		{ 256, 349, 30 },
		{ 256, 350, 30 },
		{ 256, 351, 30 },
		{ 256, 352, 30 },
		{ 256, 353, 30 },
		{ 256, 354, 30 },
		{ 256, 355, 30 },
		{ 256, 356, 30 },
		{ 256, 357, 30 },
		{ 256, 358, 30 },
		{ 256, 359, 30 },
		{ 256, 360, 30 },
		{ 256, 361, 30 },
		{ 256, 362, 30 },
		{ 256, 363, 30 },
		{ 256, 364, 30 },
		{ 256, 365, 30 },
		{ 256, 366, 30 },
		{ 256, 367, 30 },
		{ 256, 368, 30 },
		{ 256, 369, 30 },
		{ 256, 370, 30 },
		{ 256, 371, 30 },
		{ 256, 372, 30 },
		{ 256, 373, 30 },
		{ 256, 374, 30 },
		{ 256, 375, 30 },
		{ 256, 376, 30 },
		{ 256, 377, 30 },
		{ 256, 378, 30 },
		{ 256, 379, 30 },
		{ 256, 380, 30 },
		{ 256, 381, 30 },
		{ 256, 382, 30 },
		{ 256, 383, 30 },
		{ 256, 384, 30 },
		{ 256, 385, 30 },
		{ 256, 386, 30 },
		{ 256, 387, 30 },
		{ 256, 388, 30 },
		{ 256, 389, 30 },
		{ 256, 390, 30 },
		{ 256, 391, 30 },
		{ 256, 392, 30 },
		{ 256, 393, 30 },
		{ 256, 394, 30 },
		{ 256, 395, 30 },
		{ 256, 396, 30 },
		{ 256, 397, 30 },
		{ 256, 398, 30 },
		{ 256, 399, 30 },
		{ 256, 400, 30 },
		{ 256, 401, 30 },
		{ 256, 402, 30 },
		{ 256, 403, 30 },
		{ 256, 404, 30 },
		{ 256, 405, 30 },
		{ 256, 406, 30 },
		{ 256, 407, 30 },
		{ 256, 408, 30 },
		{ 256, 409, 30 },
		{ 256, 410, 30 },
		{ 256, 411, 30 },
		{ 256, 412, 30 },
		{ 256, 413, 30 },
		{ 256, 414, 30 },
		{ 256, 415, 30 },
		{ 256, 416, 30 },
		{ 256, 417, 30 },
		{ 256, 418, 30 },
		{ 256, 419, 30 },
		{ 256, 420, 30 },
		{ 256, 421, 30 },
		{ 256, 422, 30 },
		{ 256, 423, 30 },
		{ 256, 424, 30 },
		{ 256, 425, 30 },
		{ 256, 426, 30 },
		{ 256, 427, 30 },
		{ 256, 428, 30 },
		{ 256, 429, 30 },
		{ 256, 430, 30 },
		{ 256, 431, 30 },
		{ 256, 432, 30 },
		{ 256, 433, 30 },
		{ 256, 434, 30 },
		{ 256, 435, 30 },
		{ 256, 436, 30 },
		{ 256, 437, 30 },
		{ 256, 438, 30 },
		{ 256, 439, 30 },
		{ 256, 440, 30 },
		{ 256, 441, 30 },
		{ 256, 442, 30 },
		{ 256, 443, 30 },
		{ 256, 444, 30 },
		{ 256, 445, 30 },
		{ 256, 446, 30 },
		{ 256, 447, 30 },
		{ 256, 448, 30 },
		{ 256, 449, 30 },
		{ 256, 450, 30 },
		{ 256, 451, 30 },
		{ 256, 452, 30 },
		{ 256, 453, 30 },
		{ 256, 454, 30 },
		{ 256, 455, 30 },
		{ 256, 456, 30 },
		{ 256, 457, 30 },
		{ 256, 458, 30 },
		{ 256, 459, 30 },
		{ 256, 460, 30 },
		{ 256, 461, 30 },
		{ 256, 462, 30 },
		{ 256, 463, 30 },
		{ 256, 464, 30 },
		{ 256, 465, 30 },
		{ 256, 466, 30 },
		{ 256, 467, 30 },
		{ 256, 468, 30 },
		{ 256, 469, 30 },
		{ 256, 470, 30 },
		{ 256, 471, 30 },
		{ 256, 472, 30 },
		{ 256, 473, 30 },
		{ 256, 474, 30 },
		{ 256, 475, 30 },
		{ 256, 476, 30 },
		{ 256, 477, 30 },
		{ 256, 478, 30 },
		{ 256, 479, 30 },
		{ 256, 480, 30 },
		{ 256, 481, 30 },
		{ 256, 482, 30 },
		{ 256, 483, 30 },
		{ 256, 484, 30 },
		{ 256, 485, 30 },
		{ 256, 486, 30 },
		{ 256, 487, 30 },
		{ 256, 488, 30 },
		{ 256, 489, 30 },
		{ 256, 490, 30 },
		{ 256, 491, 30 },
		{ 256, 492, 30 },
		{ 256, 493, 30 },
		{ 256, 494, 30 },
		{ 256, 495, 30 },
		{ 256, 496, 30 },
		{ 256, 497, 30 },
		{ 256, 498, 30 },
		{ 256, 499, 30 },
		{ 256, 500, 30 },
		{ 256, 501, 30 },
		{ 256, 502, 30 },
		{ 256, 503, 30 },
		{ 256, 504, 30 },
		{ 256, 505, 30 },
		{ 256, 506, 30 },
		{ 256, 507, 30 },
		{ 256, 508, 30 },
		{ 256, 509, 30 },
		{ 256, 510, 30 },
		{ 256, 511, 30 },
		{ 256, 512, 30 },
		{ 257, 1, 30 },
		{ 257, 2, 30 },
		{ 257, 3, 30 },
		{ 257, 4, 30 },
		{ 257, 5, 30 },
		{ 257, 6, 30 },
		{ 257, 7, 30 },
		{ 257, 8, 30 },
		{ 257, 9, 30 },
		{ 257, 10, 30 },
		{ 257, 11, 30 },
		{ 257, 12, 30 },
		{ 257, 13, 30 },
		{ 257, 14, 30 },
		{ 257, 15, 30 },
		{ 257, 16, 30 },
		{ 257, 17, 30 },
		{ 257, 18, 30 },
		{ 257, 19, 30 },
		{ 257, 20, 30 },
		{ 257, 21, 30 },
		{ 257, 22, 30 },
		{ 257, 23, 30 },
		{ 257, 24, 30 },
		{ 257, 25, 30 },
		{ 257, 26, 30 },
		{ 257, 27, 30 },
		{ 257, 28, 30 },
		{ 257, 29, 30 },
		{ 257, 30, 30 },
		{ 257, 31, 30 },
		{ 257, 32, 30 },
		{ 257, 33, 30 },
		{ 257, 34, 30 },
		{ 257, 35, 30 },
		{ 257, 36, 30 },
		{ 257, 37, 30 },
		{ 257, 38, 30 },
		{ 257, 39, 30 },
		{ 257, 40, 30 },
		{ 257, 41, 30 },
		{ 257, 42, 30 },
		{ 257, 43, 30 },
		{ 257, 44, 30 },
		{ 257, 45, 30 },
		{ 257, 46, 30 },
		{ 257, 47, 30 },
		{ 257, 48, 30 },
		{ 257, 49, 30 },
		{ 257, 50, 30 },
		{ 257, 51, 30 },
		{ 257, 52, 30 },
		{ 257, 53, 30 },
		{ 257, 54, 30 },
		{ 257, 55, 30 },
		{ 257, 56, 30 },
		{ 257, 57, 30 },
		{ 257, 58, 30 },
		{ 257, 59, 30 },
		{ 257, 60, 30 },
		{ 257, 61, 30 },
		{ 257, 62, 30 },
		{ 257, 63, 30 },
		{ 257, 64, 30 },
		{ 257, 65, 30 },
		{ 257, 66, 30 },
		{ 257, 67, 30 },
		{ 257, 68, 30 },
		{ 257, 69, 30 },
		{ 257, 70, 30 },
		{ 257, 71, 30 },
		{ 257, 72, 30 },
		{ 257, 73, 30 },
		{ 257, 74, 30 },
		{ 257, 75, 30 },
		{ 257, 76, 30 },
		{ 257, 77, 30 },
		{ 257, 78, 30 },
		{ 257, 79, 30 },
		{ 257, 80, 30 },
		{ 257, 81, 30 },
		{ 257, 82, 30 },
		{ 257, 83, 30 },
		{ 257, 84, 30 },
		{ 257, 85, 30 },
		{ 257, 86, 30 },
		{ 257, 87, 30 },
		{ 257, 88, 30 },
		{ 257, 89, 30 },
		{ 257, 90, 30 },
		{ 257, 91, 30 },
		{ 257, 92, 30 },
		{ 257, 93, 30 },
		{ 257, 94, 30 },
		{ 257, 95, 30 },
		{ 257, 96, 30 },
		{ 257, 97, 30 },
		{ 257, 98, 30 },
		{ 257, 99, 30 },
		{ 257, 100, 30 },
		{ 257, 101, 30 },
		{ 257, 102, 30 },
		{ 257, 103, 30 },
		{ 257, 104, 30 },
		{ 257, 105, 30 },
		{ 257, 106, 30 },
		{ 257, 107, 30 },
		{ 257, 108, 30 },
		{ 257, 109, 30 },
		{ 257, 110, 30 },
		{ 257, 111, 30 },
		{ 257, 112, 30 },
		{ 257, 113, 30 },
		{ 257, 114, 30 },
		{ 257, 115, 30 },
		{ 257, 116, 30 },
		{ 257, 117, 30 },
		{ 257, 118, 30 },
		{ 257, 119, 30 },
		{ 257, 120, 30 },
		{ 257, 121, 30 },
		{ 257, 122, 30 },
		{ 257, 123, 30 },
		{ 257, 124, 30 },
		{ 257, 125, 30 },
		{ 257, 126, 30 },
		{ 257, 127, 30 },
		{ 257, 128, 30 },
		{ 257, 129, 30 },
		{ 257, 130, 30 },
		{ 257, 131, 30 },
		{ 257, 132, 30 },
		{ 257, 133, 30 },
		{ 257, 134, 30 },
		{ 257, 135, 30 },
		{ 257, 136, 30 },
		{ 257, 137, 30 },
		{ 257, 138, 30 },
		{ 257, 139, 30 },
		{ 257, 140, 30 },
		{ 257, 141, 30 },
		{ 257, 142, 30 },
		{ 257, 143, 30 },
		{ 257, 144, 30 },
		{ 257, 145, 30 },
		{ 257, 146, 30 },
		{ 257, 147, 30 },
		{ 257, 148, 30 },
		{ 257, 149, 30 },
		{ 257, 150, 30 },
		{ 257, 151, 30 },
		{ 257, 152, 30 },
		{ 257, 153, 30 },
		{ 257, 154, 30 },
		{ 257, 155, 30 },
		{ 257, 156, 30 },
		{ 257, 157, 30 },
		{ 257, 158, 30 },
		{ 257, 159, 30 },
		{ 257, 160, 30 },
		{ 257, 161, 30 },
		{ 257, 162, 30 },
		{ 257, 163, 30 },
		{ 257, 164, 30 },
		{ 257, 165, 30 },
		{ 257, 166, 30 },
		{ 257, 167, 30 },
		{ 257, 168, 30 },
		{ 257, 169, 30 },
		{ 257, 170, 30 },
		{ 257, 171, 30 },
		{ 257, 172, 30 },
		{ 257, 173, 30 },
		{ 257, 174, 30 },
		{ 257, 175, 30 },
		{ 257, 176, 30 },
		{ 257, 177, 30 },
		{ 257, 178, 30 },
		{ 257, 179, 30 },
		{ 257, 180, 30 },
		{ 257, 181, 30 },
		{ 257, 182, 30 },
		{ 257, 183, 30 },
		{ 257, 184, 30 },
		{ 257, 185, 30 },
		{ 257, 186, 30 },
		{ 257, 187, 30 },
		{ 257, 188, 30 },
		{ 257, 189, 30 },
		{ 257, 190, 30 },
		{ 257, 191, 30 },
		{ 257, 323, 30 },
		{ 257, 324, 30 },
		{ 257, 325, 30 },
		{ 257, 326, 30 },
		{ 257, 327, 30 },
		{ 257, 328, 30 },
		{ 257, 329, 30 },
		{ 257, 330, 30 },
		{ 257, 331, 30 },
		{ 257, 332, 30 },
		{ 257, 333, 30 },
		{ 257, 334, 30 },
		{ 257, 335, 30 },
		{ 257, 336, 30 },
		{ 257, 337, 30 },
		{ 257, 338, 30 },
		{ 257, 339, 30 },
		{ 257, 340, 30 },
		{ 257, 341, 30 },
		{ 257, 342, 30 },
		{ 257, 343, 30 },
		{ 257, 344, 30 },
		{ 257, 345, 30 },
		{ 257, 346, 30 },
		{ 257, 347, 30 },
		{ 257, 348, 30 },
		{ 257, 349, 30 },
		{ 257, 350, 30 },
		{ 257, 351, 30 },
		{ 257, 352, 30 },
		{ 257, 353, 30 },
		{ 257, 354, 30 },
		{ 257, 355, 30 },
		{ 257, 356, 30 },
		{ 257, 357, 30 },
		{ 257, 358, 30 },
		{ 257, 359, 30 },
		{ 257, 360, 30 },
		{ 257, 361, 30 },
		{ 257, 362, 30 },
		{ 257, 363, 30 },
		{ 257, 364, 30 },
		{ 257, 365, 30 },
		{ 257, 366, 30 },
		{ 257, 367, 30 },
		{ 257, 368, 30 },
		{ 257, 369, 30 },
		{ 257, 370, 30 },
		{ 257, 371, 30 },
		{ 257, 372, 30 },
		{ 257, 373, 30 },
		{ 257, 374, 30 },
		{ 257, 375, 30 },
		{ 257, 376, 30 },
		{ 257, 377, 30 },
		{ 257, 378, 30 },
		{ 257, 379, 30 },
		{ 257, 380, 30 },
		{ 257, 381, 30 },
		{ 257, 382, 30 },
		{ 257, 383, 30 },
		{ 257, 384, 30 },
		{ 257, 385, 30 },
		{ 257, 386, 30 },
		{ 257, 387, 30 },
		{ 257, 388, 30 },
		{ 257, 389, 30 },
		{ 257, 390, 30 },
		{ 257, 391, 30 },
		{ 257, 392, 30 },
		{ 257, 393, 30 },
		{ 257, 394, 30 },
		{ 257, 395, 30 },
		{ 257, 396, 30 },
		{ 257, 397, 30 },
		{ 257, 398, 30 },
		{ 257, 399, 30 },
		{ 257, 400, 30 },
		{ 257, 401, 30 },
		{ 257, 402, 30 },
		{ 257, 403, 30 },
		{ 257, 404, 30 },
		{ 257, 405, 30 },
		{ 257, 406, 30 },
		{ 257, 407, 30 },
		{ 257, 408, 30 },
		{ 257, 409, 30 },
		{ 257, 410, 30 },
		{ 257, 411, 30 },
		{ 257, 412, 30 },
		{ 257, 413, 30 },
		{ 257, 414, 30 },
		{ 257, 415, 30 },
		{ 257, 416, 30 },
		{ 257, 417, 30 },
		{ 257, 418, 30 },
		{ 257, 419, 30 },
		{ 257, 420, 30 },
		{ 257, 421, 30 },
		{ 257, 422, 30 },
		{ 257, 423, 30 },
		{ 257, 424, 30 },
		{ 257, 425, 30 },
		{ 257, 426, 30 },
		{ 257, 427, 30 },
		{ 257, 428, 30 },
		{ 257, 429, 30 },
		{ 257, 430, 30 },
		{ 257, 431, 30 },
		{ 257, 432, 30 },
		{ 257, 433, 30 },
		{ 257, 434, 30 },
		{ 257, 435, 30 },
		{ 257, 436, 30 },
		{ 257, 437, 30 },
		{ 257, 438, 30 },
		{ 257, 439, 30 },
		{ 257, 440, 30 },
		{ 257, 441, 30 },
		{ 257, 442, 30 },
		{ 257, 443, 30 },
		{ 257, 444, 30 },
		{ 257, 445, 30 },
		{ 257, 446, 30 },
		{ 257, 447, 30 },
		{ 257, 448, 30 },
		{ 257, 449, 30 },
		{ 257, 450, 30 },
		{ 257, 451, 30 },
		{ 257, 452, 30 },
		{ 257, 453, 30 },
		{ 257, 454, 30 },
		{ 257, 455, 30 },
		{ 257, 456, 30 },
		{ 257, 457, 30 },
		{ 257, 458, 30 },
		{ 257, 459, 30 },
		{ 257, 460, 30 },
		{ 257, 461, 30 },
		{ 257, 462, 30 },
		{ 257, 463, 30 },
		{ 257, 464, 30 },
		{ 257, 465, 30 },
		{ 257, 466, 30 },
		{ 257, 467, 30 },
		{ 257, 468, 30 },
		{ 257, 469, 30 },
		{ 257, 470, 30 },
		{ 257, 471, 30 },
		{ 257, 472, 30 },
		{ 257, 473, 30 },
		{ 257, 474, 30 },
		{ 257, 475, 30 },
		{ 257, 476, 30 },
		{ 257, 477, 30 },
		{ 257, 478, 30 },
		{ 257, 479, 30 },
		{ 257, 480, 30 },
		{ 257, 481, 30 },
		{ 257, 482, 30 },
		{ 257, 483, 30 },
		{ 257, 484, 30 },
		{ 257, 485, 30 },
		{ 257, 486, 30 },
		{ 257, 487, 30 },
		{ 257, 488, 30 },
		{ 257, 489, 30 },
		{ 257, 490, 30 },
		{ 257, 491, 30 },
		{ 257, 492, 30 },
		{ 257, 493, 30 },
		{ 257, 494, 30 },
		{ 257, 495, 30 },
		{ 257, 496, 30 },
		{ 257, 497, 30 },
		{ 257, 498, 30 },
		{ 257, 499, 30 },
		{ 257, 500, 30 },
		{ 257, 501, 30 },
		{ 257, 502, 30 },
		{ 257, 503, 30 },
		{ 257, 504, 30 },
		{ 257, 505, 30 },
		{ 257, 506, 30 },
		{ 257, 507, 30 },
		{ 257, 508, 30 },
		{ 257, 509, 30 },
		{ 257, 510, 30 },
		{ 257, 511, 30 },
		{ 257, 512, 30 },
		{ 258, 1, 30 },
		{ 258, 2, 30 },
		{ 258, 3, 30 },
		{ 258, 4, 30 },
		{ 258, 5, 30 },
		{ 258, 6, 30 },
		{ 258, 7, 30 },
		{ 258, 8, 30 },
		{ 258, 9, 30 },
		{ 258, 10, 30 },
		{ 258, 11, 30 },
		{ 258, 12, 30 },
		{ 258, 13, 30 },
		{ 258, 14, 30 },
		{ 258, 15, 30 },
		{ 258, 16, 30 },
		{ 258, 17, 30 },
		{ 258, 18, 30 },
		{ 258, 19, 30 },
		{ 258, 20, 30 },
		{ 258, 21, 30 },
		{ 258, 22, 30 },
		{ 258, 23, 30 },
		{ 258, 24, 30 },
		{ 258, 25, 30 },
		{ 258, 26, 30 },
		{ 258, 27, 30 },
		{ 258, 28, 30 },
		{ 258, 29, 30 },
		{ 258, 30, 30 },
		{ 258, 31, 30 },
		{ 258, 32, 30 },
		{ 258, 33, 30 },
		{ 258, 34, 30 },
		{ 258, 35, 30 },
		{ 258, 36, 30 },
		{ 258, 37, 30 },
		{ 258, 38, 30 },
		{ 258, 39, 30 },
		{ 258, 40, 30 },
		{ 258, 41, 30 },
		{ 258, 42, 30 },
		{ 258, 43, 30 },
		{ 258, 44, 30 },
		{ 258, 45, 30 },
		{ 258, 46, 30 },
		{ 258, 47, 30 },
		{ 258, 48, 30 },
		{ 258, 49, 30 },
		{ 258, 50, 30 },
		{ 258, 51, 30 },
		{ 258, 52, 30 },
		{ 258, 53, 30 },
		{ 258, 54, 30 },
		{ 258, 55, 30 },
		{ 258, 56, 30 },
		{ 258, 57, 30 },
		{ 258, 58, 30 },
		{ 258, 59, 30 },
		{ 258, 60, 30 },
		{ 258, 61, 30 },
		{ 258, 62, 30 },
		{ 258, 63, 30 },
		{ 258, 64, 30 },
		{ 258, 65, 30 },
		{ 258, 66, 30 },
		{ 258, 67, 30 },
		{ 258, 68, 30 },
		{ 258, 69, 30 },
		{ 258, 70, 30 },
		{ 258, 71, 30 },
		{ 258, 72, 30 },
		{ 258, 73, 30 },
		{ 258, 74, 30 },
		{ 258, 75, 30 },
		{ 258, 76, 30 },
		{ 258, 77, 30 },
		{ 258, 78, 30 },
		{ 258, 79, 30 },
		{ 258, 80, 30 },
		{ 258, 81, 30 },
		{ 258, 82, 30 },
		{ 258, 83, 30 },
		{ 258, 84, 30 },
		{ 258, 85, 30 },
		{ 258, 86, 30 },
		{ 258, 87, 30 },
		{ 258, 88, 30 },
		{ 258, 89, 30 },
		{ 258, 90, 30 },
		{ 258, 91, 30 },
		{ 258, 92, 30 },
		{ 258, 93, 30 },
		{ 258, 94, 30 },
		{ 258, 95, 30 },
		{ 258, 96, 30 },
		{ 258, 97, 30 },
		{ 258, 98, 30 },
		{ 258, 99, 30 },
		{ 258, 100, 30 },
		{ 258, 101, 30 },
		{ 258, 102, 30 },
		{ 258, 103, 30 },
		{ 258, 104, 30 },
		{ 258, 105, 30 },
		{ 258, 106, 30 },
		{ 258, 107, 30 },
		{ 258, 108, 30 },
		{ 258, 109, 30 },
		{ 258, 110, 30 },
		{ 258, 111, 30 },
		{ 258, 112, 30 },
		{ 258, 113, 30 },
		{ 258, 114, 30 },
		{ 258, 115, 30 },
		{ 258, 116, 30 },
		{ 258, 117, 30 },
		{ 258, 118, 30 },
		{ 258, 119, 30 },
		{ 258, 120, 30 },
		{ 258, 121, 30 },
		{ 258, 122, 30 },
		{ 258, 123, 30 },
		{ 258, 124, 30 },
		{ 258, 125, 30 },
		{ 258, 126, 30 },
		{ 258, 127, 30 },
		{ 258, 128, 30 },
		{ 258, 129, 30 },
		{ 258, 130, 30 },
		{ 258, 131, 30 },
		{ 258, 132, 30 },
		{ 258, 133, 30 },
		{ 258, 134, 30 },
		{ 258, 135, 30 },
		{ 258, 136, 30 },
		{ 258, 137, 30 },
		{ 258, 138, 30 },
		{ 258, 139, 30 },
		{ 258, 140, 30 },
		{ 258, 141, 30 },
		{ 258, 142, 30 },
		{ 258, 143, 30 },
		{ 258, 144, 30 },
		{ 258, 145, 30 },
		{ 258, 146, 30 },
		{ 258, 147, 30 },
		{ 258, 148, 30 },
		{ 258, 149, 30 },
		{ 258, 150, 30 },
		{ 258, 151, 30 },
		{ 258, 152, 30 },
		{ 258, 153, 30 },
		{ 258, 154, 30 },
		{ 258, 155, 30 },
		{ 258, 156, 30 },
		{ 258, 157, 30 },
		{ 258, 158, 30 },
		{ 258, 159, 30 },
		{ 258, 160, 30 },
		{ 258, 161, 30 },
		{ 258, 162, 30 },
		{ 258, 163, 30 },
		{ 258, 164, 30 },
		{ 258, 165, 30 },
		{ 258, 166, 30 },
		{ 258, 167, 30 },
		{ 258, 168, 30 },
		{ 258, 169, 30 },
		{ 258, 170, 30 },
		{ 258, 171, 30 },
		{ 258, 172, 30 },
		{ 258, 173, 30 },
		{ 258, 174, 30 },
		{ 258, 175, 30 },
		{ 258, 176, 30 },
		{ 258, 177, 30 },
		{ 258, 178, 30 },
		{ 258, 179, 30 },
		{ 258, 180, 30 },
		{ 258, 181, 30 },
		{ 258, 182, 30 },
		{ 258, 183, 30 },
		{ 258, 184, 30 },
		{ 258, 185, 30 },
		{ 258, 186, 30 },
		{ 258, 187, 30 },
		{ 258, 188, 30 },
		{ 258, 189, 30 },
		{ 258, 190, 30 },
		{ 258, 191, 30 },
		{ 258, 323, 30 },
		{ 258, 324, 30 },
		{ 258, 325, 30 },
		{ 258, 326, 30 },
		{ 258, 327, 30 },
		{ 258, 328, 30 },
		{ 258, 329, 30 },
		{ 258, 330, 30 },
		{ 258, 331, 30 },
		{ 258, 332, 30 },
		{ 258, 333, 30 },
		{ 258, 334, 30 },
		{ 258, 335, 30 },
		{ 258, 336, 30 },
		{ 258, 337, 30 },
		{ 258, 338, 30 },
		{ 258, 339, 30 },
		{ 258, 340, 30 },
		{ 258, 341, 30 },
		{ 258, 342, 30 },
		{ 258, 343, 30 },
		{ 258, 344, 30 },
		{ 258, 345, 30 },
		{ 258, 346, 30 },
		{ 258, 347, 30 },
		{ 258, 348, 30 },
		{ 258, 349, 30 },
		{ 258, 350, 30 },
		{ 258, 351, 30 },
		{ 258, 352, 30 },
		{ 258, 353, 30 },
		{ 258, 354, 30 },
		{ 258, 355, 30 },
		{ 258, 356, 30 },
		{ 258, 357, 30 },
		{ 258, 358, 30 },
		{ 258, 359, 30 },
		{ 258, 360, 30 },
		{ 258, 361, 30 },
		{ 258, 362, 30 },
		{ 258, 363, 30 },
		{ 258, 364, 30 },
		{ 258, 365, 30 },
		{ 258, 366, 30 },
		{ 258, 367, 30 },
		{ 258, 368, 30 },
		{ 258, 369, 30 },
		{ 258, 370, 30 },
		{ 258, 371, 30 },
		{ 258, 372, 30 },
		{ 258, 373, 30 },
		{ 258, 374, 30 },
		{ 258, 375, 30 },
		{ 258, 376, 30 },
		{ 258, 377, 30 },
		{ 258, 378, 30 },
		{ 258, 379, 30 },
		{ 258, 380, 30 },
		{ 258, 381, 30 },
		{ 258, 382, 30 },
		{ 258, 383, 30 },
		{ 258, 384, 30 },
		{ 258, 385, 30 },
		{ 258, 386, 30 },
		{ 258, 387, 30 },
		{ 258, 388, 30 },
		{ 258, 389, 30 },
		{ 258, 390, 30 },
		{ 258, 391, 30 },
		{ 258, 392, 30 },
		{ 258, 393, 30 },
		{ 258, 394, 30 },
		{ 258, 395, 30 },
		{ 258, 396, 30 },
		{ 258, 397, 30 },
		{ 258, 398, 30 },
		{ 258, 399, 30 },
		{ 258, 400, 30 },
		{ 258, 401, 30 },
		{ 258, 402, 30 },
		{ 258, 403, 30 },
		{ 258, 404, 30 },
		{ 258, 405, 30 },
		{ 258, 406, 30 },
		{ 258, 407, 30 },
		{ 258, 408, 30 },
		{ 258, 409, 30 },
		{ 258, 410, 30 },
		{ 258, 411, 30 },
		{ 258, 412, 30 },
		{ 258, 413, 30 },
		{ 258, 414, 30 },
		{ 258, 415, 30 },
		{ 258, 416, 30 },
		{ 258, 417, 30 },
		{ 258, 418, 30 },
		{ 258, 419, 30 },
		{ 258, 420, 30 },
		{ 258, 421, 30 },
		{ 258, 422, 30 },
		{ 258, 423, 30 },
		{ 258, 424, 30 },
		{ 258, 425, 30 },
		{ 258, 426, 30 },
		{ 258, 427, 30 },
		{ 258, 428, 30 },
		{ 258, 429, 30 },
		{ 258, 430, 30 },
		{ 258, 431, 30 },
		{ 258, 432, 30 },
		{ 258, 433, 30 },
		{ 258, 434, 30 },
		{ 258, 435, 30 },
		{ 258, 436, 30 },
		{ 258, 437, 30 },
		{ 258, 438, 30 },
		{ 258, 439, 30 },
		{ 258, 440, 30 },
		{ 258, 441, 30 },
		{ 258, 442, 30 },
		{ 258, 443, 30 },
		{ 258, 444, 30 },
		{ 258, 445, 30 },
		{ 258, 446, 30 },
		{ 258, 447, 30 },
		{ 258, 448, 30 },
		{ 258, 449, 30 },
		{ 258, 450, 30 },
		{ 258, 451, 30 },
		{ 258, 452, 30 },
		{ 258, 453, 30 },
		{ 258, 454, 30 },
		{ 258, 455, 30 },
		{ 258, 456, 30 },
		{ 258, 457, 30 },
		{ 258, 458, 30 },
		{ 258, 459, 30 },
		{ 258, 460, 30 },
		{ 258, 461, 30 },
		{ 258, 462, 30 },
		{ 258, 463, 30 },
		{ 258, 464, 30 },
		{ 258, 465, 30 },
		{ 258, 466, 30 },
		{ 258, 467, 30 },
		{ 258, 468, 30 },
		{ 258, 469, 30 },
		{ 258, 470, 30 },
		{ 258, 471, 30 },
		{ 258, 472, 30 },
		{ 258, 473, 30 },
		{ 258, 474, 30 },
		{ 258, 475, 30 },
		{ 258, 476, 30 },
		{ 258, 477, 30 },
		{ 258, 478, 30 },
		{ 258, 479, 30 },
		{ 258, 480, 30 },
		{ 258, 481, 30 },
		{ 258, 482, 30 },
		{ 258, 483, 30 },
		{ 258, 484, 30 },
		{ 258, 485, 30 },
		{ 258, 486, 30 },
		{ 258, 487, 30 },
		{ 258, 488, 30 },
		{ 258, 489, 30 },
		{ 258, 490, 30 },
		{ 258, 491, 30 },
		{ 258, 492, 30 },
		{ 258, 493, 30 },
		{ 258, 494, 30 },
		{ 258, 495, 30 },
		{ 258, 496, 30 },
		{ 258, 497, 30 },
		{ 258, 498, 30 },
		{ 258, 499, 30 },
		{ 258, 500, 30 },
		{ 258, 501, 30 },
		{ 258, 502, 30 },
		{ 258, 503, 30 },
		{ 258, 504, 30 },
		{ 258, 505, 30 },
		{ 258, 506, 30 },
		{ 258, 507, 30 },
		{ 258, 508, 30 },
		{ 258, 509, 30 },
		{ 258, 510, 30 },
		{ 258, 511, 30 },
		{ 258, 512, 30 },
		{ 259, 1, 30 },
		{ 259, 2, 30 },
		{ 259, 3, 30 },
		{ 259, 4, 30 },
		{ 259, 5, 30 },
		{ 259, 6, 30 },
		{ 259, 7, 30 },
		{ 259, 8, 30 },
		{ 259, 9, 30 },
		{ 259, 10, 30 },
		{ 259, 11, 30 },
		{ 259, 12, 30 },
		{ 259, 13, 30 },
		{ 259, 14, 30 },
		{ 259, 15, 30 },
		{ 259, 16, 30 },
		{ 259, 17, 30 },
		{ 259, 18, 30 },
		{ 259, 19, 30 },
		{ 259, 20, 30 },
		{ 259, 21, 30 },
		{ 259, 22, 30 },
		{ 259, 23, 30 },
		{ 259, 24, 30 },
		{ 259, 25, 30 },
		{ 259, 26, 30 },
		{ 259, 27, 30 },
		{ 259, 28, 30 },
		{ 259, 29, 30 },
		{ 259, 30, 30 },
		{ 259, 31, 30 },
		{ 259, 32, 30 },
		{ 259, 33, 30 },
		{ 259, 34, 30 },
		{ 259, 35, 30 },
		{ 259, 36, 30 },
		{ 259, 37, 30 },
		{ 259, 38, 30 },
		{ 259, 39, 30 },
		{ 259, 40, 30 },
		{ 259, 41, 30 },
		{ 259, 42, 30 },
		{ 259, 43, 30 },
		{ 259, 44, 30 },
		{ 259, 45, 30 },
		{ 259, 46, 30 },
		{ 259, 47, 30 },
		{ 259, 48, 30 },
		{ 259, 49, 30 },
		{ 259, 50, 30 },
		{ 259, 51, 30 },
		{ 259, 52, 30 },
		{ 259, 53, 30 },
		{ 259, 54, 30 },
		{ 259, 55, 30 },
		{ 259, 56, 30 },
		{ 259, 57, 30 },
		{ 259, 58, 30 },
		{ 259, 59, 30 },
		{ 259, 60, 30 },
		{ 259, 61, 30 },
		{ 259, 62, 30 },
		{ 259, 63, 30 },
		{ 259, 64, 30 },
		{ 259, 65, 30 },
		{ 259, 66, 30 },
		{ 259, 67, 30 },
		{ 259, 68, 30 },
		{ 259, 69, 30 },
		{ 259, 70, 30 },
		{ 259, 71, 30 },
		{ 259, 72, 30 },
		{ 259, 73, 30 },
		{ 259, 74, 30 },
		{ 259, 75, 30 },
		{ 259, 76, 30 },
		{ 259, 77, 30 },
		{ 259, 78, 30 },
		{ 259, 79, 30 },
		{ 259, 80, 30 },
		{ 259, 81, 30 },
		{ 259, 82, 30 },
		{ 259, 83, 30 },
		{ 259, 84, 30 },
		{ 259, 85, 30 },
		{ 259, 86, 30 },
		{ 259, 87, 30 },
		{ 259, 88, 30 },
		{ 259, 89, 30 },
		{ 259, 90, 30 },
		{ 259, 91, 30 },
		{ 259, 92, 30 },
		{ 259, 93, 30 },
		{ 259, 94, 30 },
		{ 259, 95, 30 },
		{ 259, 96, 30 },
		{ 259, 97, 30 },
		{ 259, 98, 30 },
		{ 259, 99, 30 },
		{ 259, 100, 30 },
		{ 259, 101, 30 },
		{ 259, 102, 30 },
		{ 259, 103, 30 },
		{ 259, 104, 30 },
		{ 259, 105, 30 },
		{ 259, 106, 30 },
		{ 259, 107, 30 },
		{ 259, 108, 30 },
		{ 259, 109, 30 },
		{ 259, 110, 30 },
		{ 259, 111, 30 },
		{ 259, 112, 30 },
		{ 259, 113, 30 },
		{ 259, 114, 30 },
		{ 259, 115, 30 },
		{ 259, 116, 30 },
		{ 259, 117, 30 },
		{ 259, 118, 30 },
		{ 259, 119, 30 },
		{ 259, 120, 30 },
		{ 259, 121, 30 },
		{ 259, 122, 30 },
		{ 259, 123, 30 },
		{ 259, 124, 30 },
		{ 259, 125, 30 },
		{ 259, 126, 30 },
		{ 259, 127, 30 },
		{ 259, 128, 30 },
		{ 259, 129, 30 },
		{ 259, 130, 30 },
		{ 259, 131, 30 },
		{ 259, 132, 30 },
		{ 259, 133, 30 },
		{ 259, 134, 30 },
		{ 259, 135, 30 },
		{ 259, 136, 30 },
		{ 259, 137, 30 },
		{ 259, 138, 30 },
		{ 259, 139, 30 },
		{ 259, 140, 30 },
		{ 259, 141, 30 },
		{ 259, 142, 30 },
		{ 259, 143, 30 },
		{ 259, 144, 30 },
		{ 259, 145, 30 },
		{ 259, 146, 30 },
		{ 259, 147, 30 },
		{ 259, 148, 30 },
		{ 259, 149, 30 },
		{ 259, 150, 30 },
		{ 259, 151, 30 },
		{ 259, 152, 30 },
		{ 259, 153, 30 },
		{ 259, 154, 30 },
		{ 259, 155, 30 },
		{ 259, 156, 30 },
		{ 259, 157, 30 },
		{ 259, 158, 30 },
		{ 259, 159, 30 },
		{ 259, 160, 30 },
		{ 259, 161, 30 },
		{ 259, 162, 30 },
		{ 259, 163, 30 },
		{ 259, 164, 30 },
		{ 259, 165, 30 },
		{ 259, 166, 30 },
		{ 259, 167, 30 },
		{ 259, 168, 30 },
		{ 259, 169, 30 },
		{ 259, 170, 30 },
		{ 259, 171, 30 },
		{ 259, 172, 30 },
		{ 259, 173, 30 },
		{ 259, 174, 30 },
		{ 259, 175, 30 },
		{ 259, 176, 30 },
		{ 259, 177, 30 },
		{ 259, 178, 30 },
		{ 259, 179, 30 },
		{ 259, 180, 30 },
		{ 259, 181, 30 },
		{ 259, 182, 30 },
		{ 259, 183, 30 },
		{ 259, 184, 30 },
		{ 259, 185, 30 },
		{ 259, 186, 30 },
		{ 259, 187, 30 },
		{ 259, 188, 30 },
		{ 259, 189, 30 },
		{ 259, 190, 30 },
		{ 259, 191, 30 },
		{ 259, 323, 30 },
		{ 259, 324, 30 },
		{ 259, 325, 30 },
		{ 259, 326, 30 },
		{ 259, 327, 30 },
		{ 259, 328, 30 },
		{ 259, 329, 30 },
		{ 259, 330, 30 },
		{ 259, 331, 30 },
		{ 259, 332, 30 },
		{ 259, 333, 30 },
		{ 259, 334, 30 },
		{ 259, 335, 30 },
		{ 259, 336, 30 },
		{ 259, 337, 30 },
		{ 259, 338, 30 },
		{ 259, 339, 30 },
		{ 259, 340, 30 },
		{ 259, 341, 30 },
		{ 259, 342, 30 },
		{ 259, 343, 30 },
		{ 259, 344, 30 },
		{ 259, 345, 30 },
		{ 259, 346, 30 },
		{ 259, 347, 30 },
		{ 259, 348, 30 },
		{ 259, 349, 30 },
		{ 259, 350, 30 },
		{ 259, 351, 30 },
		{ 259, 352, 30 },
		{ 259, 353, 30 },
		{ 259, 354, 30 },
		{ 259, 355, 30 },
		{ 259, 356, 30 },
		{ 259, 357, 30 },
		{ 259, 358, 30 },
		{ 259, 359, 30 },
		{ 259, 360, 30 },
		{ 259, 361, 30 },
		{ 259, 362, 30 },
		{ 259, 363, 30 },
		{ 259, 364, 30 },
		{ 259, 365, 30 },
		{ 259, 366, 30 },
		{ 259, 367, 30 },
		{ 259, 368, 30 },
		{ 259, 369, 30 },
		{ 259, 370, 30 },
		{ 259, 371, 30 },
		{ 259, 372, 30 },
		{ 259, 373, 30 },
		{ 259, 374, 30 },
		{ 259, 375, 30 },
		{ 259, 376, 30 },
		{ 259, 377, 30 },
		{ 259, 378, 30 },
		{ 259, 379, 30 },
		{ 259, 380, 30 },
		{ 259, 381, 30 },
		{ 259, 382, 30 },
		{ 259, 383, 30 },
		{ 259, 384, 30 },
		{ 259, 385, 30 },
		{ 259, 386, 30 },
		{ 259, 387, 30 },
		{ 259, 388, 30 },
		{ 259, 389, 30 },
		{ 259, 390, 30 },
		{ 259, 391, 30 },
		{ 259, 392, 30 },
		{ 259, 393, 30 },
		{ 259, 394, 30 },
		{ 259, 395, 30 },
		{ 259, 396, 30 },
		{ 259, 397, 30 },
		{ 259, 398, 30 },
		{ 259, 399, 30 },
		{ 259, 400, 30 },
		{ 259, 401, 30 },
		{ 259, 402, 30 },
		{ 259, 403, 30 },
		{ 259, 404, 30 },
		{ 259, 405, 30 },
		{ 259, 406, 30 },
		{ 259, 407, 30 },
		{ 259, 408, 30 },
		{ 259, 409, 30 },
		{ 259, 410, 30 },
		{ 259, 411, 30 },
		{ 259, 412, 30 },
		{ 259, 413, 30 },
		{ 259, 414, 30 },
		{ 259, 415, 30 },
		{ 259, 416, 30 },
		{ 259, 417, 30 },
		{ 259, 418, 30 },
		{ 259, 419, 30 },
		{ 259, 420, 30 },
		{ 259, 421, 30 },
		{ 259, 422, 30 },
		{ 259, 423, 30 },
		{ 259, 424, 30 },
		{ 259, 425, 30 },
		{ 259, 426, 30 },
		{ 259, 427, 30 },
		{ 259, 428, 30 },
		{ 259, 429, 30 },
		{ 259, 430, 30 },
		{ 259, 431, 30 },
		{ 259, 432, 30 },
		{ 259, 433, 30 },
		{ 259, 434, 30 },
		{ 259, 435, 30 },
		{ 259, 436, 30 },
		{ 259, 437, 30 },
		{ 259, 438, 30 },
		{ 259, 439, 30 },
		{ 259, 440, 30 },
		{ 259, 441, 30 },
		{ 259, 442, 30 },
		{ 259, 443, 30 },
		{ 259, 444, 30 },
		{ 259, 445, 30 },
		{ 259, 446, 30 },
		{ 259, 447, 30 },
		{ 259, 448, 30 },
		{ 259, 449, 30 },
		{ 259, 450, 30 },
		{ 259, 451, 30 },
		{ 259, 452, 30 },
		{ 259, 453, 30 },
		{ 259, 454, 30 },
		{ 259, 455, 30 },
		{ 259, 456, 30 },
		{ 259, 457, 30 },
		{ 259, 458, 30 },
		{ 259, 459, 30 },
		{ 259, 460, 30 },
		{ 259, 461, 30 },
		{ 259, 462, 30 },
		{ 259, 463, 30 },
		{ 259, 464, 30 },
		{ 259, 465, 30 },
		{ 259, 466, 30 },
		{ 259, 467, 30 },
		{ 259, 468, 30 },
		{ 259, 469, 30 },
		{ 259, 470, 30 },
		{ 259, 471, 30 },
		{ 259, 472, 30 },
		{ 259, 473, 30 },
		{ 259, 474, 30 },
		{ 259, 475, 30 },
		{ 259, 476, 30 },
		{ 259, 477, 30 },
		{ 259, 478, 30 },
		{ 259, 479, 30 },
		{ 259, 480, 30 },
		{ 259, 481, 30 },
		{ 259, 482, 30 },
		{ 259, 483, 30 },
		{ 259, 484, 30 },
		{ 259, 485, 30 },
		{ 259, 486, 30 },
		{ 259, 487, 30 },
		{ 259, 488, 30 },
		{ 259, 489, 30 },
		{ 259, 490, 30 },
		{ 259, 491, 30 },
		{ 259, 492, 30 },
		{ 259, 493, 30 },
		{ 259, 494, 30 },
		{ 259, 495, 30 },
		{ 259, 496, 30 },
		{ 259, 497, 30 },
		{ 259, 498, 30 },
		{ 259, 499, 30 },
		{ 259, 500, 30 },
		{ 259, 501, 30 },
		{ 259, 502, 30 },
		{ 259, 503, 30 },
		{ 259, 504, 30 },
		{ 259, 505, 30 },
		{ 259, 506, 30 },
		{ 259, 507, 30 },
		{ 259, 508, 30 },
		{ 259, 509, 30 },
		{ 259, 510, 30 },
		{ 259, 511, 30 },
		{ 259, 512, 30 },
		{ 260, 1, 30 },
		{ 260, 2, 30 },
		{ 260, 3, 30 },
		{ 260, 4, 30 },
		{ 260, 5, 30 },
		{ 260, 6, 30 },
		{ 260, 7, 30 },
		{ 260, 8, 30 },
		{ 260, 9, 30 },
		{ 260, 10, 30 },
		{ 260, 11, 30 },
		{ 260, 12, 30 },
		{ 260, 13, 30 },
		{ 260, 14, 30 },
		{ 260, 15, 30 },
		{ 260, 16, 30 },
		{ 260, 17, 30 },
		{ 260, 18, 30 },
		{ 260, 19, 30 },
		{ 260, 20, 30 },
		{ 260, 21, 30 },
		{ 260, 22, 30 },
		{ 260, 23, 30 },
		{ 260, 24, 30 },
		{ 260, 25, 30 },
		{ 260, 26, 30 },
		{ 260, 27, 30 },
		{ 260, 28, 30 },
		{ 260, 29, 30 },
		{ 260, 30, 30 },
		{ 260, 31, 30 },
		{ 260, 32, 30 },
		{ 260, 33, 30 },
		{ 260, 34, 30 },
		{ 260, 35, 30 },
		{ 260, 36, 30 },
		{ 260, 37, 30 },
		{ 260, 38, 30 },
		{ 260, 39, 30 },
		{ 260, 40, 30 },
		{ 260, 41, 30 },
		{ 260, 42, 30 },
		{ 260, 43, 30 },
		{ 260, 44, 30 },
		{ 260, 45, 30 },
		{ 260, 46, 30 },
		{ 260, 47, 30 },
		{ 260, 48, 30 },
		{ 260, 49, 30 },
		{ 260, 50, 30 },
		{ 260, 51, 30 },
		{ 260, 52, 30 },
		{ 260, 53, 30 },
		{ 260, 54, 30 },
		{ 260, 55, 30 },
		{ 260, 56, 30 },
		{ 260, 57, 30 },
		{ 260, 58, 30 },
		{ 260, 59, 30 },
		{ 260, 60, 30 },
		{ 260, 61, 30 },
		{ 260, 62, 30 },
		{ 260, 63, 30 },
		{ 260, 64, 30 },
		{ 260, 65, 30 },
		{ 260, 66, 30 },
		{ 260, 67, 30 },
		{ 260, 68, 30 },
		{ 260, 69, 30 },
		{ 260, 70, 30 },
		{ 260, 71, 30 },
		{ 260, 72, 30 },
		{ 260, 73, 30 },
		{ 260, 74, 30 },
		{ 260, 75, 30 },
		{ 260, 76, 30 },
		{ 260, 77, 30 },
		{ 260, 78, 30 },
		{ 260, 79, 30 },
		{ 260, 80, 30 },
		{ 260, 81, 30 },
		{ 260, 82, 30 },
		{ 260, 83, 30 },
		{ 260, 84, 30 },
		{ 260, 85, 30 },
		{ 260, 86, 30 },
		{ 260, 87, 30 },
		{ 260, 88, 30 },
		{ 260, 89, 30 },
		{ 260, 90, 30 },
		{ 260, 91, 30 },
		{ 260, 92, 30 },
		{ 260, 93, 30 },
		{ 260, 94, 30 },
		{ 260, 95, 30 },
		{ 260, 96, 30 },
		{ 260, 97, 30 },
		{ 260, 98, 30 },
		{ 260, 99, 30 },
		{ 260, 100, 30 },
		{ 260, 101, 30 },
		{ 260, 102, 30 },
		{ 260, 103, 30 },
		{ 260, 104, 30 },
		{ 260, 105, 30 },
		{ 260, 106, 30 },
		{ 260, 107, 30 },
		{ 260, 108, 30 },
		{ 260, 109, 30 },
		{ 260, 110, 30 },
		{ 260, 111, 30 },
		{ 260, 112, 30 },
		{ 260, 113, 30 },
		{ 260, 114, 30 },
		{ 260, 115, 30 },
		{ 260, 116, 30 },
		{ 260, 117, 30 },
		{ 260, 118, 30 },
		{ 260, 119, 30 },
		{ 260, 120, 30 },
		{ 260, 121, 30 },
		{ 260, 122, 30 },
		{ 260, 123, 30 },
		{ 260, 124, 30 },
		{ 260, 125, 30 },
		{ 260, 126, 30 },
		{ 260, 127, 30 },
		{ 260, 128, 30 },
		{ 260, 129, 30 },
		{ 260, 130, 30 },
		{ 260, 131, 30 },
		{ 260, 132, 30 },
		{ 260, 133, 30 },
		{ 260, 134, 30 },
		{ 260, 135, 30 },
		{ 260, 136, 30 },
		{ 260, 137, 30 },
		{ 260, 138, 30 },
		{ 260, 139, 30 },
		{ 260, 140, 30 },
		{ 260, 141, 30 },
		{ 260, 142, 30 },
		{ 260, 143, 30 },
		{ 260, 144, 30 },
		{ 260, 145, 30 },
		{ 260, 146, 30 },
		{ 260, 147, 30 },
		{ 260, 148, 30 },
		{ 260, 149, 30 },
		{ 260, 150, 30 },
		{ 260, 151, 30 },
		{ 260, 152, 30 },
		{ 260, 153, 30 },
		{ 260, 154, 30 },
		{ 260, 155, 30 },
		{ 260, 156, 30 },
		{ 260, 157, 30 },
		{ 260, 158, 30 },
		{ 260, 159, 30 },
		{ 260, 160, 30 },
		{ 260, 161, 30 },
		{ 260, 162, 30 },
		{ 260, 163, 30 },
		{ 260, 164, 30 },
		{ 260, 165, 30 },
		{ 260, 166, 30 },
		{ 260, 167, 30 },
		{ 260, 168, 30 },
		{ 260, 169, 30 },
		{ 260, 170, 30 },
		{ 260, 171, 30 },
		{ 260, 172, 30 },
		{ 260, 173, 30 },
		{ 260, 174, 30 },
		{ 260, 175, 30 },
		{ 260, 176, 30 },
		{ 260, 177, 30 },
		{ 260, 178, 30 },
		{ 260, 179, 30 },
		{ 260, 180, 30 },
		{ 260, 181, 30 },
		{ 260, 182, 30 },
		{ 260, 183, 30 },
		{ 260, 184, 30 },
		{ 260, 185, 30 },
		{ 260, 186, 30 },
		{ 260, 187, 30 },
		{ 260, 188, 30 },
		{ 260, 189, 30 },
		{ 260, 190, 30 },
		{ 260, 191, 30 },
		{ 260, 323, 30 },
		{ 260, 324, 30 },
		{ 260, 325, 30 },
		{ 260, 326, 30 },
		{ 260, 327, 30 },
		{ 260, 328, 30 },
		{ 260, 329, 30 },
		{ 260, 330, 30 },
		{ 260, 331, 30 },
		{ 260, 332, 30 },
		{ 260, 333, 30 },
		{ 260, 334, 30 },
		{ 260, 335, 30 },
		{ 260, 336, 30 },
		{ 260, 337, 30 },
		{ 260, 338, 30 },
		{ 260, 339, 30 },
		{ 260, 340, 30 },
		{ 260, 341, 30 },
		{ 260, 342, 30 },
		{ 260, 343, 30 },
		{ 260, 344, 30 },
		{ 260, 345, 30 },
		{ 260, 346, 30 },
		{ 260, 347, 30 },
		{ 260, 348, 30 },
		{ 260, 349, 30 },
		{ 260, 350, 30 },
		{ 260, 351, 30 },
		{ 260, 352, 30 },
		{ 260, 353, 30 },
		{ 260, 354, 30 },
		{ 260, 355, 30 },
		{ 260, 356, 30 },
		{ 260, 357, 30 },
		{ 260, 358, 30 },
		{ 260, 359, 30 },
		{ 260, 360, 30 },
		{ 260, 361, 30 },
		{ 260, 362, 30 },
		{ 260, 363, 30 },
		{ 260, 364, 30 },
		{ 260, 365, 30 },
		{ 260, 366, 30 },
		{ 260, 367, 30 },
		{ 260, 368, 30 },
		{ 260, 369, 30 },
		{ 260, 370, 30 },
		{ 260, 371, 30 },
		{ 260, 372, 30 },
		{ 260, 373, 30 },
		{ 260, 374, 30 },
		{ 260, 375, 30 },
		{ 260, 376, 30 },
		{ 260, 377, 30 },
		{ 260, 378, 30 },
		{ 260, 379, 30 },
		{ 260, 380, 30 },
		{ 260, 381, 30 },
		{ 260, 382, 30 },
		{ 260, 383, 30 },
		{ 260, 384, 30 },
		{ 260, 385, 30 },
		{ 260, 386, 30 },
		{ 260, 387, 30 },
		{ 260, 388, 30 },
		{ 260, 389, 30 },
		{ 260, 390, 30 },
		{ 260, 391, 30 },
		{ 260, 392, 30 },
		{ 260, 393, 30 },
		{ 260, 394, 30 },
		{ 260, 395, 30 },
		{ 260, 396, 30 },
		{ 260, 397, 30 },
		{ 260, 398, 30 },
		{ 260, 399, 30 },
		{ 260, 400, 30 },
		{ 260, 401, 30 },
		{ 260, 402, 30 },
		{ 260, 403, 30 },
		{ 260, 404, 30 },
		{ 260, 405, 30 },
		{ 260, 406, 30 },
		{ 260, 407, 30 },
		{ 260, 408, 30 },
		{ 260, 409, 30 },
		{ 260, 410, 30 },
		{ 260, 411, 30 },
		{ 260, 412, 30 },
		{ 260, 413, 30 },
		{ 260, 414, 30 },
		{ 260, 415, 30 },
		{ 260, 416, 30 },
		{ 260, 417, 30 },
		{ 260, 418, 30 },
		{ 260, 419, 30 },
		{ 260, 420, 30 },
		{ 260, 421, 30 },
		{ 260, 422, 30 },
		{ 260, 423, 30 },
		{ 260, 424, 30 },
		{ 260, 425, 30 },
		{ 260, 426, 30 },
		{ 260, 427, 30 },
		{ 260, 428, 30 },
		{ 260, 429, 30 },
		{ 260, 430, 30 },
		{ 260, 431, 30 },
		{ 260, 432, 30 },
		{ 260, 433, 30 },
		{ 260, 434, 30 },
		{ 260, 435, 30 },
		{ 260, 436, 30 },
		{ 260, 437, 30 },
		{ 260, 438, 30 },
		{ 260, 439, 30 },
		{ 260, 440, 30 },
		{ 260, 441, 30 },
		{ 260, 442, 30 },
		{ 260, 443, 30 },
		{ 260, 444, 30 },
		{ 260, 445, 30 },
		{ 260, 446, 30 },
		{ 260, 447, 30 },
		{ 260, 448, 30 },
		{ 260, 449, 30 },
		{ 260, 450, 30 },
		{ 260, 451, 30 },
		{ 260, 452, 30 },
		{ 260, 453, 30 },
		{ 260, 454, 30 },
		{ 260, 455, 30 },
		{ 260, 456, 30 },
		{ 260, 457, 30 },
		{ 260, 458, 30 },
		{ 260, 459, 30 },
		{ 260, 460, 30 },
		{ 260, 461, 30 },
		{ 260, 462, 30 },
		{ 260, 463, 30 },
		{ 260, 464, 30 },
		{ 260, 465, 30 },
		{ 260, 466, 30 },
		{ 260, 467, 30 },
		{ 260, 468, 30 },
		{ 260, 469, 30 },
		{ 260, 470, 30 },
		{ 260, 471, 30 },
		{ 260, 472, 30 },
		{ 260, 473, 30 },
		{ 260, 474, 30 },
		{ 260, 475, 30 },
		{ 260, 476, 30 },
		{ 260, 477, 30 },
		{ 260, 478, 30 },
		{ 260, 479, 30 },
		{ 260, 480, 30 },
		{ 260, 481, 30 },
		{ 260, 482, 30 },
		{ 260, 483, 30 },
		{ 260, 484, 30 },
		{ 260, 485, 30 },
		{ 260, 486, 30 },
		{ 260, 487, 30 },
		{ 260, 488, 30 },
		{ 260, 489, 30 },
		{ 260, 490, 30 },
		{ 260, 491, 30 },
		{ 260, 492, 30 },
		{ 260, 493, 30 },
		{ 260, 494, 30 },
		{ 260, 495, 30 },
		{ 260, 496, 30 },
		{ 260, 497, 30 },
		{ 260, 498, 30 },
		{ 260, 499, 30 },
		{ 260, 500, 30 },
		{ 260, 501, 30 },
		{ 260, 502, 30 },
		{ 260, 503, 30 },
		{ 260, 504, 30 },
		{ 260, 505, 30 },
		{ 260, 506, 30 },
		{ 260, 507, 30 },
		{ 260, 508, 30 },
		{ 260, 509, 30 },
		{ 260, 510, 30 },
		{ 260, 511, 30 },
		{ 260, 512, 30 },
		{ 261, 1, 30 },
		{ 261, 2, 30 },
		{ 261, 3, 30 },
		{ 261, 4, 30 },
		{ 261, 5, 30 },
		{ 261, 6, 30 },
		{ 261, 7, 30 },
		{ 261, 8, 30 },
		{ 261, 9, 30 },
		{ 261, 10, 30 },
		{ 261, 11, 30 },
		{ 261, 12, 30 },
		{ 261, 13, 30 },
		{ 261, 14, 30 },
		{ 261, 15, 30 },
		{ 261, 16, 30 },
		{ 261, 17, 30 },
		{ 261, 18, 30 },
		{ 261, 19, 30 },
		{ 261, 20, 30 },
		{ 261, 21, 30 },
		{ 261, 22, 30 },
		{ 261, 23, 30 },
		{ 261, 24, 30 },
		{ 261, 25, 30 },
		{ 261, 26, 30 },
		{ 261, 27, 30 },
		{ 261, 28, 30 },
		{ 261, 29, 30 },
		{ 261, 30, 30 },
		{ 261, 31, 30 },
		{ 261, 32, 30 },
		{ 261, 33, 30 },
		{ 261, 34, 30 },
		{ 261, 35, 30 },
		{ 261, 36, 30 },
		{ 261, 37, 30 },
		{ 261, 38, 30 },
		{ 261, 39, 30 },
		{ 261, 40, 30 },
		{ 261, 41, 30 },
		{ 261, 42, 30 },
		{ 261, 43, 30 },
		{ 261, 44, 30 },
		{ 261, 45, 30 },
		{ 261, 46, 30 },
		{ 261, 47, 30 },
		{ 261, 48, 30 },
		{ 261, 49, 30 },
		{ 261, 50, 30 },
		{ 261, 51, 30 },
		{ 261, 52, 30 },
		{ 261, 53, 30 },
		{ 261, 54, 30 },
		{ 261, 55, 30 },
		{ 261, 56, 30 },
		{ 261, 57, 30 },
		{ 261, 58, 30 },
		{ 261, 59, 30 },
		{ 261, 60, 30 },
		{ 261, 61, 30 },
		{ 261, 62, 30 },
		{ 261, 63, 30 },
		{ 261, 64, 30 },
		{ 261, 65, 30 },
		{ 261, 66, 30 },
		{ 261, 67, 30 },
		{ 261, 68, 30 },
		{ 261, 69, 30 },
		{ 261, 70, 30 },
		{ 261, 71, 30 },
		{ 261, 72, 30 },
		{ 261, 73, 30 },
		{ 261, 74, 30 },
		{ 261, 75, 30 },
		{ 261, 76, 30 },
		{ 261, 77, 30 },
		{ 261, 78, 30 },
		{ 261, 79, 30 },
		{ 261, 80, 30 },
		{ 261, 81, 30 },
		{ 261, 82, 30 },
		{ 261, 83, 30 },
		{ 261, 84, 30 },
		{ 261, 85, 30 },
		{ 261, 86, 30 },
		{ 261, 87, 30 },
		{ 261, 88, 30 },
		{ 261, 89, 30 },
		{ 261, 90, 30 },
		{ 261, 91, 30 },
		{ 261, 92, 30 },
		{ 261, 93, 30 },
		{ 261, 94, 30 },
		{ 261, 95, 30 },
		{ 261, 96, 30 },
		{ 261, 97, 30 },
		{ 261, 98, 30 },
		{ 261, 99, 30 },
		{ 261, 100, 30 },
		{ 261, 101, 30 },
		{ 261, 102, 30 },
		{ 261, 103, 30 },
		{ 261, 104, 30 },
		{ 261, 105, 30 },
		{ 261, 106, 30 },
		{ 261, 107, 30 },
		{ 261, 108, 30 },
		{ 261, 109, 30 },
		{ 261, 110, 30 },
		{ 261, 111, 30 },
		{ 261, 112, 30 },
		{ 261, 113, 30 },
		{ 261, 114, 30 },
		{ 261, 115, 30 },
		{ 261, 116, 30 },
		{ 261, 117, 30 },
		{ 261, 118, 30 },
		{ 261, 119, 30 },
		{ 261, 120, 30 },
		{ 261, 121, 30 },
		{ 261, 122, 30 },
		{ 261, 123, 30 },
		{ 261, 124, 30 },
		{ 261, 125, 30 },
		{ 261, 126, 30 },
		{ 261, 127, 30 },
		{ 261, 128, 30 },
		{ 261, 129, 30 },
		{ 261, 130, 30 },
		{ 261, 131, 30 },
		{ 261, 132, 30 },
		{ 261, 133, 30 },
		{ 261, 134, 30 },
		{ 261, 135, 30 },
		{ 261, 136, 30 },
		{ 261, 137, 30 },
		{ 261, 138, 30 },
		{ 261, 139, 30 },
		{ 261, 140, 30 },
		{ 261, 141, 30 },
		{ 261, 142, 30 },
		{ 261, 143, 30 },
		{ 261, 144, 30 },
		{ 261, 145, 30 },
		{ 261, 146, 30 },
		{ 261, 147, 30 },
		{ 261, 148, 30 },
		{ 261, 149, 30 },
		{ 261, 150, 30 },
		{ 261, 151, 30 },
		{ 261, 152, 30 },
		{ 261, 153, 30 },
		{ 261, 154, 30 },
		{ 261, 155, 30 },
		{ 261, 156, 30 },
		{ 261, 157, 30 },
		{ 261, 158, 30 },
		{ 261, 159, 30 },
		{ 261, 160, 30 },
		{ 261, 161, 30 },
		{ 261, 162, 30 },
		{ 261, 163, 30 },
		{ 261, 164, 30 },
		{ 261, 165, 30 },
		{ 261, 166, 30 },
		{ 261, 167, 30 },
		{ 261, 168, 30 },
		{ 261, 169, 30 },
		{ 261, 170, 30 },
		{ 261, 171, 30 },
		{ 261, 172, 30 },
		{ 261, 173, 30 },
		{ 261, 174, 30 },
		{ 261, 175, 30 },
		{ 261, 176, 30 },
		{ 261, 177, 30 },
		{ 261, 178, 30 },
		{ 261, 179, 30 },
		{ 261, 180, 30 },
		{ 261, 181, 30 },
		{ 261, 182, 30 },
		{ 261, 183, 30 },
		{ 261, 184, 30 },
		{ 261, 185, 30 },
		{ 261, 186, 30 },
		{ 261, 187, 30 },
		{ 261, 188, 30 },
		{ 261, 189, 30 },
		{ 261, 190, 30 },
		{ 261, 191, 30 },
		{ 261, 323, 30 },
		{ 261, 324, 30 },
		{ 261, 325, 30 },
		{ 261, 326, 30 },
		{ 261, 327, 30 },
		{ 261, 328, 30 },
		{ 261, 329, 30 },
		{ 261, 330, 30 },
		{ 261, 331, 30 },
		{ 261, 332, 30 },
		{ 261, 333, 30 },
		{ 261, 334, 30 },
		{ 261, 335, 30 },
		{ 261, 336, 30 },
		{ 261, 337, 30 },
		{ 261, 338, 30 },
		{ 261, 339, 30 },
		{ 261, 340, 30 },
		{ 261, 341, 30 },
		{ 261, 342, 30 },
		{ 261, 343, 30 },
		{ 261, 344, 30 },
		{ 261, 345, 30 },
		{ 261, 346, 30 },
		{ 261, 347, 30 },
		{ 261, 348, 30 },
		{ 261, 349, 30 },
		{ 261, 350, 30 },
		{ 261, 351, 30 },
		{ 261, 352, 30 },
		{ 261, 353, 30 },
		{ 261, 354, 30 },
		{ 261, 355, 30 },
		{ 261, 356, 30 },
		{ 261, 357, 30 },
		{ 261, 358, 30 },
		{ 261, 359, 30 },
		{ 261, 360, 30 },
		{ 261, 361, 30 },
		{ 261, 362, 30 },
		{ 261, 363, 30 },
		{ 261, 364, 30 },
		{ 261, 365, 30 },
		{ 261, 366, 30 },
		{ 261, 367, 30 },
		{ 261, 368, 30 },
		{ 261, 369, 30 },
		{ 261, 370, 30 },
		{ 261, 371, 30 },
		{ 261, 372, 30 },
		{ 261, 373, 30 },
		{ 261, 374, 30 },
		{ 261, 375, 30 },
		{ 261, 376, 30 },
		{ 261, 377, 30 },
		{ 261, 378, 30 },
		{ 261, 379, 30 },
		{ 261, 380, 30 },
		{ 261, 381, 30 },
		{ 261, 382, 30 },
		{ 261, 383, 30 },
		{ 261, 384, 30 },
		{ 261, 385, 30 },
		{ 261, 386, 30 },
		{ 261, 387, 30 },
		{ 261, 388, 30 },
		{ 261, 389, 30 },
		{ 261, 390, 30 },
		{ 261, 391, 30 },
		{ 261, 392, 30 },
		{ 261, 393, 30 },
		{ 261, 394, 30 },
		{ 261, 395, 30 },
		{ 261, 396, 30 },
		{ 261, 397, 30 },
		{ 261, 398, 30 },
		{ 261, 399, 30 },
		{ 261, 400, 30 },
		{ 261, 401, 30 },
		{ 261, 402, 30 },
		{ 261, 403, 30 },
		{ 261, 404, 30 },
		{ 261, 405, 30 },
		{ 261, 406, 30 },
		{ 261, 407, 30 },
		{ 261, 408, 30 },
		{ 261, 409, 30 },
		{ 261, 410, 30 },
		{ 261, 411, 30 },
		{ 261, 412, 30 },
		{ 261, 413, 30 },
		{ 261, 414, 30 },
		{ 261, 415, 30 },
		{ 261, 416, 30 },
		{ 261, 417, 30 },
		{ 261, 418, 30 },
		{ 261, 419, 30 },
		{ 261, 420, 30 },
		{ 261, 421, 30 },
		{ 261, 422, 30 },
		{ 261, 423, 30 },
		{ 261, 424, 30 },
		{ 261, 425, 30 },
		{ 261, 426, 30 },
		{ 261, 427, 30 },
		{ 261, 428, 30 },
		{ 261, 429, 30 },
		{ 261, 430, 30 },
		{ 261, 431, 30 },
		{ 261, 432, 30 },
		{ 261, 433, 30 },
		{ 261, 434, 30 },
		{ 261, 435, 30 },
		{ 261, 436, 30 },
		{ 261, 437, 30 },
		{ 261, 438, 30 },
		{ 261, 439, 30 },
		{ 261, 440, 30 },
		{ 261, 441, 30 },
		{ 261, 442, 30 },
		{ 261, 443, 30 },
		{ 261, 444, 30 },
		{ 261, 445, 30 },
		{ 261, 446, 30 },
		{ 261, 447, 30 },
		{ 261, 448, 30 },
		{ 261, 449, 30 },
		{ 261, 450, 30 },
		{ 261, 451, 30 },
		{ 261, 452, 30 },
		{ 261, 453, 30 },
		{ 261, 454, 30 },
		{ 261, 455, 30 },
		{ 261, 456, 30 },
		{ 261, 457, 30 },
		{ 261, 458, 30 },
		{ 261, 459, 30 },
		{ 261, 460, 30 },
		{ 261, 461, 30 },
		{ 261, 462, 30 },
		{ 261, 463, 30 },
		{ 261, 464, 30 },
		{ 261, 465, 30 },
		{ 261, 466, 30 },
		{ 261, 467, 30 },
		{ 261, 468, 30 },
		{ 261, 469, 30 },
		{ 261, 470, 30 },
		{ 261, 471, 30 },
		{ 261, 472, 30 },
		{ 261, 473, 30 },
		{ 261, 474, 30 },
		{ 261, 475, 30 },
		{ 261, 476, 30 },
		{ 261, 477, 30 },
		{ 261, 478, 30 },
		{ 261, 479, 30 },
		{ 261, 480, 30 },
		{ 261, 481, 30 },
		{ 261, 482, 30 },
		{ 261, 483, 30 },
		{ 261, 484, 30 },
		{ 261, 485, 30 },
		{ 261, 486, 30 },
		{ 261, 487, 30 },
		{ 261, 488, 30 },
		{ 261, 489, 30 },
		{ 261, 490, 30 },
		{ 261, 491, 30 },
		{ 261, 492, 30 },
		{ 261, 493, 30 },
		{ 261, 494, 30 },
		{ 261, 495, 30 },
		{ 261, 496, 30 },
		{ 261, 497, 30 },
		{ 261, 498, 30 },
		{ 261, 499, 30 },
		{ 261, 500, 30 },
		{ 261, 501, 30 },
		{ 261, 502, 30 },
		{ 261, 503, 30 },
		{ 261, 504, 30 },
		{ 261, 505, 30 },
		{ 261, 506, 30 },
		{ 261, 507, 30 },
		{ 261, 508, 30 },
		{ 261, 509, 30 },
		{ 261, 510, 30 },
		{ 261, 511, 30 },
		{ 261, 512, 30 },
		{ 262, 1, 30 },
		{ 262, 2, 30 },
		{ 262, 3, 30 },
		{ 262, 4, 30 },
		{ 262, 5, 30 },
		{ 262, 6, 30 },
		{ 262, 7, 30 },
		{ 262, 8, 30 },
		{ 262, 9, 30 },
		{ 262, 10, 30 },
		{ 262, 11, 30 },
		{ 262, 12, 30 },
		{ 262, 13, 30 },
		{ 262, 14, 30 },
		{ 262, 15, 30 },
		{ 262, 16, 30 },
		{ 262, 17, 30 },
		{ 262, 18, 30 },
		{ 262, 19, 30 },
		{ 262, 20, 30 },
		{ 262, 21, 30 },
		{ 262, 22, 30 },
		{ 262, 23, 30 },
		{ 262, 24, 30 },
		{ 262, 25, 30 },
		{ 262, 26, 30 },
		{ 262, 27, 30 },
		{ 262, 28, 30 },
		{ 262, 29, 30 },
		{ 262, 30, 30 },
		{ 262, 31, 30 },
		{ 262, 32, 30 },
		{ 262, 33, 30 },
		{ 262, 34, 30 },
		{ 262, 35, 30 },
		{ 262, 36, 30 },
		{ 262, 37, 30 },
		{ 262, 38, 30 },
		{ 262, 39, 30 },
		{ 262, 40, 30 },
		{ 262, 41, 30 },
		{ 262, 42, 30 },
		{ 262, 43, 30 },
		{ 262, 44, 30 },
		{ 262, 45, 30 },
		{ 262, 46, 30 },
		{ 262, 47, 30 },
		{ 262, 48, 30 },
		{ 262, 49, 30 },
		{ 262, 50, 30 },
		{ 262, 51, 30 },
		{ 262, 52, 30 },
		{ 262, 53, 30 },
		{ 262, 54, 30 },
		{ 262, 55, 30 },
		{ 262, 56, 30 },
		{ 262, 57, 30 },
		{ 262, 58, 30 },
		{ 262, 59, 30 },
		{ 262, 60, 30 },
		{ 262, 61, 30 },
		{ 262, 62, 30 },
		{ 262, 63, 30 },
		{ 262, 64, 30 },
		{ 262, 65, 30 },
		{ 262, 66, 30 },
		{ 262, 67, 30 },
		{ 262, 68, 30 },
		{ 262, 69, 30 },
		{ 262, 70, 30 },
		{ 262, 71, 30 },
		{ 262, 72, 30 },
		{ 262, 73, 30 },
		{ 262, 74, 30 },
		{ 262, 75, 30 },
		{ 262, 76, 30 },
		{ 262, 77, 30 },
		{ 262, 78, 30 },
		{ 262, 79, 30 },
		{ 262, 80, 30 },
		{ 262, 81, 30 },
		{ 262, 82, 30 },
		{ 262, 83, 30 },
		{ 262, 84, 30 },
		{ 262, 85, 30 },
		{ 262, 86, 30 },
		{ 262, 87, 30 },
		{ 262, 88, 30 },
		{ 262, 89, 30 },
		{ 262, 90, 30 },
		{ 262, 91, 30 },
		{ 262, 92, 30 },
		{ 262, 93, 30 },
		{ 262, 94, 30 },
		{ 262, 95, 30 },
		{ 262, 96, 30 },
		{ 262, 97, 30 },
		{ 262, 98, 30 },
		{ 262, 99, 30 },
		{ 262, 100, 30 },
		{ 262, 101, 30 },
		{ 262, 102, 30 },
		{ 262, 103, 30 },
		{ 262, 104, 30 },
		{ 262, 105, 30 },
		{ 262, 106, 30 },
		{ 262, 107, 30 },
		{ 262, 108, 30 },
		{ 262, 109, 30 },
		{ 262, 110, 30 },
		{ 262, 111, 30 },
		{ 262, 112, 30 },
		{ 262, 113, 30 },
		{ 262, 114, 30 },
		{ 262, 115, 30 },
		{ 262, 116, 30 },
		{ 262, 117, 30 },
		{ 262, 118, 30 },
		{ 262, 119, 30 },
		{ 262, 120, 30 },
		{ 262, 121, 30 },
		{ 262, 122, 30 },
		{ 262, 123, 30 },
		{ 262, 124, 30 },
		{ 262, 125, 30 },
		{ 262, 126, 30 },
		{ 262, 127, 30 },
		{ 262, 128, 30 },
		{ 262, 129, 30 },
		{ 262, 130, 30 },
		{ 262, 131, 30 },
		{ 262, 132, 30 },
		{ 262, 133, 30 },
		{ 262, 134, 30 },
		{ 262, 135, 30 },
		{ 262, 136, 30 },
		{ 262, 137, 30 },
		{ 262, 138, 30 },
		{ 262, 139, 30 },
		{ 262, 140, 30 },
		{ 262, 141, 30 },
		{ 262, 142, 30 },
		{ 262, 143, 30 },
		{ 262, 144, 30 },
		{ 262, 145, 30 },
		{ 262, 146, 30 },
		{ 262, 147, 30 },
		{ 262, 148, 30 },
		{ 262, 149, 30 },
		{ 262, 150, 30 },
		{ 262, 151, 30 },
		{ 262, 152, 30 },
		{ 262, 153, 30 },
		{ 262, 154, 30 },
		{ 262, 155, 30 },
		{ 262, 156, 30 },
		{ 262, 157, 30 },
		{ 262, 158, 30 },
		{ 262, 159, 30 },
		{ 262, 160, 30 },
		{ 262, 161, 30 },
		{ 262, 162, 30 },
		{ 262, 163, 30 },
		{ 262, 164, 30 },
		{ 262, 165, 30 },
		{ 262, 166, 30 },
		{ 262, 167, 30 },
		{ 262, 168, 30 },
		{ 262, 169, 30 },
		{ 262, 170, 30 },
		{ 262, 171, 30 },
		{ 262, 172, 30 },
		{ 262, 173, 30 },
		{ 262, 174, 30 },
		{ 262, 175, 30 },
		{ 262, 176, 30 },
		{ 262, 177, 30 },
		{ 262, 178, 30 },
		{ 262, 179, 30 },
		{ 262, 180, 30 },
		{ 262, 181, 30 },
		{ 262, 182, 30 },
		{ 262, 183, 30 },
		{ 262, 184, 30 },
		{ 262, 185, 30 },
		{ 262, 186, 30 },
		{ 262, 187, 30 },
		{ 262, 188, 30 },
		{ 262, 189, 30 },
		{ 262, 190, 30 },
		{ 262, 191, 30 },
		{ 262, 323, 30 },
		{ 262, 324, 30 },
		{ 262, 325, 30 },
		{ 262, 326, 30 },
		{ 262, 327, 30 },
		{ 262, 328, 30 },
		{ 262, 329, 30 },
		{ 262, 330, 30 },
		{ 262, 331, 30 },
		{ 262, 332, 30 },
		{ 262, 333, 30 },
		{ 262, 334, 30 },
		{ 262, 335, 30 },
		{ 262, 336, 30 },
		{ 262, 337, 30 },
		{ 262, 338, 30 },
		{ 262, 339, 30 },
		{ 262, 340, 30 },
		{ 262, 341, 30 },
		{ 262, 342, 30 },
		{ 262, 343, 30 },
		{ 262, 344, 30 },
		{ 262, 345, 30 },
		{ 262, 346, 30 },
		{ 262, 347, 30 },
		{ 262, 348, 30 },
		{ 262, 349, 30 },
		{ 262, 350, 30 },
		{ 262, 351, 30 },
		{ 262, 352, 30 },
		{ 262, 353, 30 },
		{ 262, 354, 30 },
		{ 262, 355, 30 },
		{ 262, 356, 30 },
		{ 262, 357, 30 },
		{ 262, 358, 30 },
		{ 262, 359, 30 },
		{ 262, 360, 30 },
		{ 262, 361, 30 },
		{ 262, 362, 30 },
		{ 262, 363, 30 },
		{ 262, 364, 30 },
		{ 262, 365, 30 },
		{ 262, 366, 30 },
		{ 262, 367, 30 },
		{ 262, 368, 30 },
		{ 262, 369, 30 },
		{ 262, 370, 30 },
		{ 262, 371, 30 },
		{ 262, 372, 30 },
		{ 262, 373, 30 },
		{ 262, 374, 30 },
		{ 262, 375, 30 },
		{ 262, 376, 30 },
		{ 262, 377, 30 },
		{ 262, 378, 30 },
		{ 262, 379, 30 },
		{ 262, 380, 30 },
		{ 262, 381, 30 },
		{ 262, 382, 30 },
		{ 262, 383, 30 },
		{ 262, 384, 30 },
		{ 262, 385, 30 },
		{ 262, 386, 30 },
		{ 262, 387, 30 },
		{ 262, 388, 30 },
		{ 262, 389, 30 },
		{ 262, 390, 30 },
		{ 262, 391, 30 },
		{ 262, 392, 30 },
		{ 262, 393, 30 },
		{ 262, 394, 30 },
		{ 262, 395, 30 },
		{ 262, 396, 30 },
		{ 262, 397, 30 },
		{ 262, 398, 30 },
		{ 262, 399, 30 },
		{ 262, 400, 30 },
		{ 262, 401, 30 },
		{ 262, 402, 30 },
		{ 262, 403, 30 },
		{ 262, 404, 30 },
		{ 262, 405, 30 },
		{ 262, 406, 30 },
		{ 262, 407, 30 },
		{ 262, 408, 30 },
		{ 262, 409, 30 },
		{ 262, 410, 30 },
		{ 262, 411, 30 },
		{ 262, 412, 30 },
		{ 262, 413, 30 },
		{ 262, 414, 30 },
		{ 262, 415, 30 },
		{ 262, 416, 30 },
		{ 262, 417, 30 },
		{ 262, 418, 30 },
		{ 262, 419, 30 },
		{ 262, 420, 30 },
		{ 262, 421, 30 },
		{ 262, 422, 30 },
		{ 262, 423, 30 },
		{ 262, 424, 30 },
		{ 262, 425, 30 },
		{ 262, 426, 30 },
		{ 262, 427, 30 },
		{ 262, 428, 30 },
		{ 262, 429, 30 },
		{ 262, 430, 30 },
		{ 262, 431, 30 },
		{ 262, 432, 30 },
		{ 262, 433, 30 },
		{ 262, 434, 30 },
		{ 262, 435, 30 },
		{ 262, 436, 30 },
		{ 262, 437, 30 },
		{ 262, 438, 30 },
		{ 262, 439, 30 },
		{ 262, 440, 30 },
		{ 262, 441, 30 },
		{ 262, 442, 30 },
		{ 262, 443, 30 },
		{ 262, 444, 30 },
		{ 262, 445, 30 },
		{ 262, 446, 30 },
		{ 262, 447, 30 },
		{ 262, 448, 30 },
		{ 262, 449, 30 },
		{ 262, 450, 30 },
		{ 262, 451, 30 },
		{ 262, 452, 30 },
		{ 262, 453, 30 },
		{ 262, 454, 30 },
		{ 262, 455, 30 },
		{ 262, 456, 30 },
		{ 262, 457, 30 },
		{ 262, 458, 30 },
		{ 262, 459, 30 },
		{ 262, 460, 30 },
		{ 262, 461, 30 },
		{ 262, 462, 30 },
		{ 262, 463, 30 },
		{ 262, 464, 30 },
		{ 262, 465, 30 },
		{ 262, 466, 30 },
		{ 262, 467, 30 },
		{ 262, 468, 30 },
		{ 262, 469, 30 },
		{ 262, 470, 30 },
		{ 262, 471, 30 },
		{ 262, 472, 30 },
		{ 262, 473, 30 },
		{ 262, 474, 30 },
		{ 262, 475, 30 },
		{ 262, 476, 30 },
		{ 262, 477, 30 },
		{ 262, 478, 30 },
		{ 262, 479, 30 },
		{ 262, 480, 30 },
		{ 262, 481, 30 },
		{ 262, 482, 30 },
		{ 262, 483, 30 },
		{ 262, 484, 30 },
		{ 262, 485, 30 },
		{ 262, 486, 30 },
		{ 262, 487, 30 },
		{ 262, 488, 30 },
		{ 262, 489, 30 },
		{ 262, 490, 30 },
		{ 262, 491, 30 },
		{ 262, 492, 30 },
		{ 262, 493, 30 },
		{ 262, 494, 30 },
		{ 262, 495, 30 },
		{ 262, 496, 30 },
		{ 262, 497, 30 },
		{ 262, 498, 30 },
		{ 262, 499, 30 },
		{ 262, 500, 30 },
		{ 262, 501, 30 },
		{ 262, 502, 30 },
		{ 262, 503, 30 },
		{ 262, 504, 30 },
		{ 262, 505, 30 },
		{ 262, 506, 30 },
		{ 262, 507, 30 },
		{ 262, 508, 30 },
		{ 262, 509, 30 },
		{ 262, 510, 30 },
		{ 262, 511, 30 },
		{ 262, 512, 30 },
		{ 263, 1, 30 },
		{ 263, 2, 30 },
		{ 263, 3, 30 },
		{ 263, 4, 30 },
		{ 263, 5, 30 },
		{ 263, 6, 30 },
		{ 263, 7, 30 },
		{ 263, 8, 30 },
		{ 263, 9, 30 },
		{ 263, 10, 30 },
		{ 263, 11, 30 },
		{ 263, 12, 30 },
		{ 263, 13, 30 },
		{ 263, 14, 30 },
		{ 263, 15, 30 },
		{ 263, 16, 30 },
		{ 263, 17, 30 },
		{ 263, 18, 30 },
		{ 263, 19, 30 },
		{ 263, 20, 30 },
		{ 263, 21, 30 },
		{ 263, 22, 30 },
		{ 263, 23, 30 },
		{ 263, 24, 30 },
		{ 263, 25, 30 },
		{ 263, 26, 30 },
		{ 263, 27, 30 },
		{ 263, 28, 30 },
		{ 263, 29, 30 },
		{ 263, 30, 30 },
		{ 263, 31, 30 },
		{ 263, 32, 30 },
		{ 263, 33, 30 },
		{ 263, 34, 30 },
		{ 263, 35, 30 },
		{ 263, 36, 30 },
		{ 263, 37, 30 },
		{ 263, 38, 30 },
		{ 263, 39, 30 },
		{ 263, 40, 30 },
		{ 263, 41, 30 },
		{ 263, 42, 30 },
		{ 263, 43, 30 },
		{ 263, 44, 30 },
		{ 263, 45, 30 },
		{ 263, 46, 30 },
		{ 263, 47, 30 },
		{ 263, 48, 30 },
		{ 263, 49, 30 },
		{ 263, 50, 30 },
		{ 263, 51, 30 },
		{ 263, 52, 30 },
		{ 263, 53, 30 },
		{ 263, 54, 30 },
		{ 263, 55, 30 },
		{ 263, 56, 30 },
		{ 263, 57, 30 },
		{ 263, 58, 30 },
		{ 263, 59, 30 },
		{ 263, 60, 30 },
		{ 263, 61, 30 },
		{ 263, 62, 30 },
		{ 263, 63, 30 },
		{ 263, 64, 30 },
		{ 263, 65, 30 },
		{ 263, 66, 30 },
		{ 263, 67, 30 },
		{ 263, 68, 30 },
		{ 263, 69, 30 },
		{ 263, 70, 30 },
		{ 263, 71, 30 },
		{ 263, 72, 30 },
		{ 263, 73, 30 },
		{ 263, 74, 30 },
		{ 263, 75, 30 },
		{ 263, 76, 30 },
		{ 263, 77, 30 },
		{ 263, 78, 30 },
		{ 263, 79, 30 },
		{ 263, 80, 30 },
		{ 263, 81, 30 },
		{ 263, 82, 30 },
		{ 263, 83, 30 },
		{ 263, 84, 30 },
		{ 263, 85, 30 },
		{ 263, 86, 30 },
		{ 263, 87, 30 },
		{ 263, 88, 30 },
		{ 263, 89, 30 },
		{ 263, 90, 30 },
		{ 263, 91, 30 },
		{ 263, 92, 30 },
		{ 263, 93, 30 },
		{ 263, 94, 30 },
		{ 263, 95, 30 },
		{ 263, 96, 30 },
		{ 263, 97, 30 },
		{ 263, 98, 30 },
		{ 263, 99, 30 },
		{ 263, 100, 30 },
		{ 263, 101, 30 },
		{ 263, 102, 30 },
		{ 263, 103, 30 },
		{ 263, 104, 30 },
		{ 263, 105, 30 },
		{ 263, 106, 30 },
		{ 263, 107, 30 },
		{ 263, 108, 30 },
		{ 263, 109, 30 },
		{ 263, 110, 30 },
		{ 263, 111, 30 },
		{ 263, 112, 30 },
		{ 263, 113, 30 },
		{ 263, 114, 30 },
		{ 263, 115, 30 },
		{ 263, 116, 30 },
		{ 263, 117, 30 },
		{ 263, 118, 30 },
		{ 263, 119, 30 },
		{ 263, 120, 30 },
		{ 263, 121, 30 },
		{ 263, 122, 30 },
		{ 263, 123, 30 },
		{ 263, 124, 30 },
		{ 263, 125, 30 },
		{ 263, 126, 30 },
		{ 263, 127, 30 },
		{ 263, 128, 30 },
		{ 263, 129, 30 },
		{ 263, 130, 30 },
		{ 263, 131, 30 },
		{ 263, 132, 30 },
		{ 263, 133, 30 },
		{ 263, 134, 30 },
		{ 263, 135, 30 },
		{ 263, 136, 30 },
		{ 263, 137, 30 },
		{ 263, 138, 30 },
		{ 263, 139, 30 },
		{ 263, 140, 30 },
		{ 263, 141, 30 },
		{ 263, 142, 30 },
		{ 263, 143, 30 },
		{ 263, 144, 30 },
		{ 263, 145, 30 },
		{ 263, 146, 30 },
		{ 263, 147, 30 },
		{ 263, 148, 30 },
		{ 263, 149, 30 },
		{ 263, 150, 30 },
		{ 263, 151, 30 },
		{ 263, 152, 30 },
		{ 263, 153, 30 },
		{ 263, 154, 30 },
		{ 263, 155, 30 },
		{ 263, 156, 30 },
		{ 263, 157, 30 },
		{ 263, 158, 30 },
		{ 263, 159, 30 },
		{ 263, 160, 30 },
		{ 263, 161, 30 },
		{ 263, 162, 30 },
		{ 263, 163, 30 },
		{ 263, 164, 30 },
		{ 263, 165, 30 },
		{ 263, 166, 30 },
		{ 263, 167, 30 },
		{ 263, 168, 30 },
		{ 263, 169, 30 },
		{ 263, 170, 30 },
		{ 263, 171, 30 },
		{ 263, 172, 30 },
		{ 263, 173, 30 },
		{ 263, 174, 30 },
		{ 263, 175, 30 },
		{ 263, 176, 30 },
		{ 263, 177, 30 },
		{ 263, 178, 30 },
		{ 263, 179, 30 },
		{ 263, 180, 30 },
		{ 263, 181, 30 },
		{ 263, 182, 30 },
		{ 263, 183, 30 },
		{ 263, 184, 30 },
		{ 263, 185, 30 },
		{ 263, 186, 30 },
		{ 263, 187, 30 },
		{ 263, 188, 30 },
		{ 263, 189, 30 },
		{ 263, 190, 30 },
		{ 263, 191, 30 },
		{ 263, 323, 30 },
		{ 263, 324, 30 },
		{ 263, 325, 30 },
		{ 263, 326, 30 },
		{ 263, 327, 30 },
		{ 263, 328, 30 },
		{ 263, 329, 30 },
		{ 263, 330, 30 },
		{ 263, 331, 30 },
		{ 263, 332, 30 },
		{ 263, 333, 30 },
		{ 263, 334, 30 },
		{ 263, 335, 30 },
		{ 263, 336, 30 },
		{ 263, 337, 30 },
		{ 263, 338, 30 },
		{ 263, 339, 30 },
		{ 263, 340, 30 },
		{ 263, 341, 30 },
		{ 263, 342, 30 },
		{ 263, 343, 30 },
		{ 263, 344, 30 },
		{ 263, 345, 30 },
		{ 263, 346, 30 },
		{ 263, 347, 30 },
		{ 263, 348, 30 },
		{ 263, 349, 30 },
		{ 263, 350, 30 },
		{ 263, 351, 30 },
		{ 263, 352, 30 },
		{ 263, 353, 30 },
		{ 263, 354, 30 },
		{ 263, 355, 30 },
		{ 263, 356, 30 },
		{ 263, 357, 30 },
		{ 263, 358, 30 },
		{ 263, 359, 30 },
		{ 263, 360, 30 },
		{ 263, 361, 30 },
		{ 263, 362, 30 },
		{ 263, 363, 30 },
		{ 263, 364, 30 },
		{ 263, 365, 30 },
		{ 263, 366, 30 },
		{ 263, 367, 30 },
		{ 263, 368, 30 },
		{ 263, 369, 30 },
		{ 263, 370, 30 },
		{ 263, 371, 30 },
		{ 263, 372, 30 },
		{ 263, 373, 30 },
		{ 263, 374, 30 },
		{ 263, 375, 30 },
		{ 263, 376, 30 },
		{ 263, 377, 30 },
		{ 263, 378, 30 },
		{ 263, 379, 30 },
		{ 263, 380, 30 },
		{ 263, 381, 30 },
		{ 263, 382, 30 },
		{ 263, 383, 30 },
		{ 263, 384, 30 },
		{ 263, 385, 30 },
		{ 263, 386, 30 },
		{ 263, 387, 30 },
		{ 263, 388, 30 },
		{ 263, 389, 30 },
		{ 263, 390, 30 },
		{ 263, 391, 30 },
		{ 263, 392, 30 },
		{ 263, 393, 30 },
		{ 263, 394, 30 },
		{ 263, 395, 30 },
		{ 263, 396, 30 },
		{ 263, 397, 30 },
		{ 263, 398, 30 },
		{ 263, 399, 30 },
		{ 263, 400, 30 },
		{ 263, 401, 30 },
		{ 263, 402, 30 },
		{ 263, 403, 30 },
		{ 263, 404, 30 },
		{ 263, 405, 30 },
		{ 263, 406, 30 },
		{ 263, 407, 30 },
		{ 263, 408, 30 },
		{ 263, 409, 30 },
		{ 263, 410, 30 },
		{ 263, 411, 30 },
		{ 263, 412, 30 },
		{ 263, 413, 30 },
		{ 263, 414, 30 },
		{ 263, 415, 30 },
		{ 263, 416, 30 },
		{ 263, 417, 30 },
		{ 263, 418, 30 },
		{ 263, 419, 30 },
		{ 263, 420, 30 },
		{ 263, 421, 30 },
		{ 263, 422, 30 },
		{ 263, 423, 30 },
		{ 263, 424, 30 },
		{ 263, 425, 30 },
		{ 263, 426, 30 },
		{ 263, 427, 30 },
		{ 263, 428, 30 },
		{ 263, 429, 30 },
		{ 263, 430, 30 },
		{ 263, 431, 30 },
		{ 263, 432, 30 },
		{ 263, 433, 30 },
		{ 263, 434, 30 },
		{ 263, 435, 30 },
		{ 263, 436, 30 },
		{ 263, 437, 30 },
		{ 263, 438, 30 },
		{ 263, 439, 30 },
		{ 263, 440, 30 },
		{ 263, 441, 30 },
		{ 263, 442, 30 },
		{ 263, 443, 30 },
		{ 263, 444, 30 },
		{ 263, 445, 30 },
		{ 263, 446, 30 },
		{ 263, 447, 30 },
		{ 263, 448, 30 },
		{ 263, 449, 30 },
		{ 263, 450, 30 },
		{ 263, 451, 30 },
		{ 263, 452, 30 },
		{ 263, 453, 30 },
		{ 263, 454, 30 },
		{ 263, 455, 30 },
		{ 263, 456, 30 },
		{ 263, 457, 30 },
		{ 263, 458, 30 },
		{ 263, 459, 30 },
		{ 263, 460, 30 },
		{ 263, 461, 30 },
		{ 263, 462, 30 },
		{ 263, 463, 30 },
		{ 263, 464, 30 },
		{ 263, 465, 30 },
		{ 263, 466, 30 },
		{ 263, 467, 30 },
		{ 263, 468, 30 },
		{ 263, 469, 30 },
		{ 263, 470, 30 },
		{ 263, 471, 30 },
		{ 263, 472, 30 },
		{ 263, 473, 30 },
		{ 263, 474, 30 },
		{ 263, 475, 30 },
		{ 263, 476, 30 },
		{ 263, 477, 30 },
		{ 263, 478, 30 },
		{ 263, 479, 30 },
		{ 263, 480, 30 },
		{ 263, 481, 30 },
		{ 263, 482, 30 },
		{ 263, 483, 30 },
		{ 263, 484, 30 },
		{ 263, 485, 30 },
		{ 263, 486, 30 },
		{ 263, 487, 30 },
		{ 263, 488, 30 },
		{ 263, 489, 30 },
		{ 263, 490, 30 },
		{ 263, 491, 30 },
		{ 263, 492, 30 },
		{ 263, 493, 30 },
		{ 263, 494, 30 },
		{ 263, 495, 30 },
		{ 263, 496, 30 },
		{ 263, 497, 30 },
		{ 263, 498, 30 },
		{ 263, 499, 30 },
		{ 263, 500, 30 },
		{ 263, 501, 30 },
		{ 263, 502, 30 },
		{ 263, 503, 30 },
		{ 263, 504, 30 },
		{ 263, 505, 30 },
		{ 263, 506, 30 },
		{ 263, 507, 30 },
		{ 263, 508, 30 },
		{ 263, 509, 30 },
		{ 263, 510, 30 },
		{ 263, 511, 30 },
		{ 263, 512, 30 },
		{ 264, 1, 30 },
		{ 264, 2, 30 },
		{ 264, 3, 30 },
		{ 264, 4, 30 },
		{ 264, 5, 30 },
		{ 264, 6, 30 },
		{ 264, 7, 30 },
		{ 264, 8, 30 },
		{ 264, 9, 30 },
		{ 264, 10, 30 },
		{ 264, 11, 30 },
		{ 264, 12, 30 },
		{ 264, 13, 30 },
		{ 264, 14, 30 },
		{ 264, 15, 30 },
		{ 264, 16, 30 },
		{ 264, 17, 30 },
		{ 264, 18, 30 },
		{ 264, 19, 30 },
		{ 264, 20, 30 },
		{ 264, 21, 30 },
		{ 264, 22, 30 },
		{ 264, 23, 30 },
		{ 264, 24, 30 },
		{ 264, 25, 30 },
		{ 264, 26, 30 },
		{ 264, 27, 30 },
		{ 264, 28, 30 },
		{ 264, 29, 30 },
		{ 264, 30, 30 },
		{ 264, 31, 30 },
		{ 264, 32, 30 },
		{ 264, 33, 30 },
		{ 264, 34, 30 },
		{ 264, 35, 30 },
		{ 264, 36, 30 },
		{ 264, 37, 30 },
		{ 264, 38, 30 },
		{ 264, 39, 30 },
		{ 264, 40, 30 },
		{ 264, 41, 30 },
		{ 264, 42, 30 },
		{ 264, 43, 30 },
		{ 264, 44, 30 },
		{ 264, 45, 30 },
		{ 264, 46, 30 },
		{ 264, 47, 30 },
		{ 264, 48, 30 },
		{ 264, 49, 30 },
		{ 264, 50, 30 },
		{ 264, 51, 30 },
		{ 264, 52, 30 },
		{ 264, 53, 30 },
		{ 264, 54, 30 },
		{ 264, 55, 30 },
		{ 264, 56, 30 },
		{ 264, 57, 30 },
		{ 264, 58, 30 },
		{ 264, 59, 30 },
		{ 264, 60, 30 },
		{ 264, 61, 30 },
		{ 264, 62, 30 },
		{ 264, 63, 30 },
		{ 264, 64, 30 },
		{ 264, 65, 30 },
		{ 264, 66, 30 },
		{ 264, 67, 30 },
		{ 264, 68, 30 },
		{ 264, 69, 30 },
		{ 264, 70, 30 },
		{ 264, 71, 30 },
		{ 264, 72, 30 },
		{ 264, 73, 30 },
		{ 264, 74, 30 },
		{ 264, 75, 30 },
		{ 264, 76, 30 },
		{ 264, 77, 30 },
		{ 264, 78, 30 },
		{ 264, 79, 30 },
		{ 264, 80, 30 },
		{ 264, 81, 30 },
		{ 264, 82, 30 },
		{ 264, 83, 30 },
		{ 264, 84, 30 },
		{ 264, 85, 30 },
		{ 264, 86, 30 },
		{ 264, 87, 30 },
		{ 264, 88, 30 },
		{ 264, 89, 30 },
		{ 264, 90, 30 },
		{ 264, 91, 30 },
		{ 264, 92, 30 },
		{ 264, 93, 30 },
		{ 264, 94, 30 },
		{ 264, 95, 30 },
		{ 264, 96, 30 },
		{ 264, 97, 30 },
		{ 264, 98, 30 },
		{ 264, 99, 30 },
		{ 264, 100, 30 },
		{ 264, 101, 30 },
		{ 264, 102, 30 },
		{ 264, 103, 30 },
		{ 264, 104, 30 },
		{ 264, 105, 30 },
		{ 264, 106, 30 },
		{ 264, 107, 30 },
		{ 264, 108, 30 },
		{ 264, 109, 30 },
		{ 264, 110, 30 },
		{ 264, 111, 30 },
		{ 264, 112, 30 },
		{ 264, 113, 30 },
		{ 264, 114, 30 },
		{ 264, 115, 30 },
		{ 264, 116, 30 },
		{ 264, 117, 30 },
		{ 264, 118, 30 },
		{ 264, 119, 30 },
		{ 264, 120, 30 },
		{ 264, 121, 30 },
		{ 264, 122, 30 },
		{ 264, 123, 30 },
		{ 264, 124, 30 },
		{ 264, 125, 30 },
		{ 264, 126, 30 },
		{ 264, 127, 30 },
		{ 264, 128, 30 },
		{ 264, 129, 30 },
		{ 264, 130, 30 },
		{ 264, 131, 30 },
		{ 264, 132, 30 },
		{ 264, 133, 30 },
		{ 264, 134, 30 },
		{ 264, 135, 30 },
		{ 264, 136, 30 },
		{ 264, 137, 30 },
		{ 264, 138, 30 },
		{ 264, 139, 30 },
		{ 264, 140, 30 },
		{ 264, 141, 30 },
		{ 264, 142, 30 },
		{ 264, 143, 30 },
		{ 264, 144, 30 },
		{ 264, 145, 30 },
		{ 264, 146, 30 },
		{ 264, 147, 30 },
		{ 264, 148, 30 },
		{ 264, 149, 30 },
		{ 264, 150, 30 },
		{ 264, 151, 30 },
		{ 264, 152, 30 },
		{ 264, 153, 30 },
		{ 264, 154, 30 },
		{ 264, 155, 30 },
		{ 264, 156, 30 },
		{ 264, 157, 30 },
		{ 264, 158, 30 },
		{ 264, 159, 30 },
		{ 264, 160, 30 },
		{ 264, 161, 30 },
		{ 264, 162, 30 },
		{ 264, 163, 30 },
		{ 264, 164, 30 },
		{ 264, 165, 30 },
		{ 264, 166, 30 },
		{ 264, 167, 30 },
		{ 264, 168, 30 },
		{ 264, 169, 30 },
		{ 264, 170, 30 },
		{ 264, 171, 30 },
		{ 264, 172, 30 },
		{ 264, 173, 30 },
		{ 264, 174, 30 },
		{ 264, 175, 30 },
		{ 264, 176, 30 },
		{ 264, 177, 30 },
		{ 264, 178, 30 },
		{ 264, 179, 30 },
		{ 264, 180, 30 },
		{ 264, 181, 30 },
		{ 264, 182, 30 },
		{ 264, 183, 30 },
		{ 264, 184, 30 },
		{ 264, 185, 30 },
		{ 264, 186, 30 },
		{ 264, 187, 30 },
		{ 264, 188, 30 },
		{ 264, 189, 30 },
		{ 264, 190, 30 },
		{ 264, 324, 30 },
		{ 264, 325, 30 },
		{ 264, 326, 30 },
		{ 264, 327, 30 },
		{ 264, 328, 30 },
		{ 264, 329, 30 },
		{ 264, 330, 30 },
		{ 264, 331, 30 },
		{ 264, 332, 30 },
		{ 264, 333, 30 },
		{ 264, 334, 30 },
		{ 264, 335, 30 },
		{ 264, 336, 30 },
		{ 264, 337, 30 },
		{ 264, 338, 30 },
		{ 264, 339, 30 },
		{ 264, 340, 30 },
		{ 264, 341, 30 },
		{ 264, 342, 30 },
		{ 264, 343, 30 },
		{ 264, 344, 30 },
		{ 264, 345, 30 },
		{ 264, 346, 30 },
		{ 264, 347, 30 },
		{ 264, 348, 30 },
		{ 264, 349, 30 },
		{ 264, 350, 30 },
		{ 264, 351, 30 },
		{ 264, 352, 30 },
		{ 264, 353, 30 },
		{ 264, 354, 30 },
		{ 264, 355, 30 },
		{ 264, 356, 30 },
		{ 264, 357, 30 },
		{ 264, 358, 30 },
		{ 264, 359, 30 },
		{ 264, 360, 30 },
		{ 264, 361, 30 },
		{ 264, 362, 30 },
		{ 264, 363, 30 },
		{ 264, 364, 30 },
		{ 264, 365, 30 },
		{ 264, 366, 30 },
		{ 264, 367, 30 },
		{ 264, 368, 30 },
		{ 264, 369, 30 },
		{ 264, 370, 30 },
		{ 264, 371, 30 },
		{ 264, 372, 30 },
		{ 264, 373, 30 },
		{ 264, 374, 30 },
		{ 264, 375, 30 },
		{ 264, 376, 30 },
		{ 264, 377, 30 },
		{ 264, 378, 30 },
		{ 264, 379, 30 },
		{ 264, 380, 30 },
		{ 264, 381, 30 },
		{ 264, 382, 30 },
		{ 264, 383, 30 },
		{ 264, 384, 30 },
		{ 264, 385, 30 },
		{ 264, 386, 30 },
		{ 264, 387, 30 },
		{ 264, 388, 30 },
		{ 264, 389, 30 },
		{ 264, 390, 30 },
		{ 264, 391, 30 },
		{ 264, 392, 30 },
		{ 264, 393, 30 },
		{ 264, 394, 30 },
		{ 264, 395, 30 },
		{ 264, 396, 30 },
		{ 264, 397, 30 },
		{ 264, 398, 30 },
		{ 264, 399, 30 },
		{ 264, 400, 30 },
		{ 264, 401, 30 },
		{ 264, 402, 30 },
		{ 264, 403, 30 },
		{ 264, 404, 30 },
		{ 264, 405, 30 },
		{ 264, 406, 30 },
		{ 264, 407, 30 },
		{ 264, 408, 30 },
		{ 264, 409, 30 },
		{ 264, 410, 30 },
		{ 264, 411, 30 },
		{ 264, 412, 30 },
		{ 264, 413, 30 },
		{ 264, 414, 30 },
		{ 264, 415, 30 },
		{ 264, 416, 30 },
		{ 264, 417, 30 },
		{ 264, 418, 30 },
		{ 264, 419, 30 },
		{ 264, 420, 30 },
		{ 264, 421, 30 },
		{ 264, 422, 30 },
		{ 264, 423, 30 },
		{ 264, 424, 30 },
		{ 264, 425, 30 },
		{ 264, 426, 30 },
		{ 264, 427, 30 },
		{ 264, 428, 30 },
		{ 264, 429, 30 },
		{ 264, 430, 30 },
		{ 264, 431, 30 },
		{ 264, 432, 30 },
		{ 264, 433, 30 },
		{ 264, 434, 30 },
		{ 264, 435, 30 },
		{ 264, 436, 30 },
		{ 264, 437, 30 },
		{ 264, 438, 30 },
		{ 264, 439, 30 },
		{ 264, 440, 30 },
		{ 264, 441, 30 },
		{ 264, 442, 30 },
		{ 264, 443, 30 },
		{ 264, 444, 30 },
		{ 264, 445, 30 },
		{ 264, 446, 30 },
		{ 264, 447, 30 },
		{ 264, 448, 30 },
		{ 264, 449, 30 },
		{ 264, 450, 30 },
		{ 264, 451, 30 },
		{ 264, 452, 30 },
		{ 264, 453, 30 },
		{ 264, 454, 30 },
		{ 264, 455, 30 },
		{ 264, 456, 30 },
		{ 264, 457, 30 },
		{ 264, 458, 30 },
		{ 264, 459, 30 },
		{ 264, 460, 30 },
		{ 264, 461, 30 },
		{ 264, 462, 30 },
		{ 264, 463, 30 },
		{ 264, 464, 30 },
		{ 264, 465, 30 },
		{ 264, 466, 30 },
		{ 264, 467, 30 },
		{ 264, 468, 30 },
		{ 264, 469, 30 },
		{ 264, 470, 30 },
		{ 264, 471, 30 },
		{ 264, 472, 30 },
		{ 264, 473, 30 },
		{ 264, 474, 30 },
		{ 264, 475, 30 },
		{ 264, 476, 30 },
		{ 264, 477, 30 },
		{ 264, 478, 30 },
		{ 264, 479, 30 },
		{ 264, 480, 30 },
		{ 264, 481, 30 },
		{ 264, 482, 30 },
		{ 264, 483, 30 },
		{ 264, 484, 30 },
		{ 264, 485, 30 },
		{ 264, 486, 30 },
		{ 264, 487, 30 },
		{ 264, 488, 30 },
		{ 264, 489, 30 },
		{ 264, 490, 30 },
		{ 264, 491, 30 },
		{ 264, 492, 30 },
		{ 264, 493, 30 },
		{ 264, 494, 30 },
		{ 264, 495, 30 },
		{ 264, 496, 30 },
		{ 264, 497, 30 },
		{ 264, 498, 30 },
		{ 264, 499, 30 },
		{ 264, 500, 30 },
		{ 264, 501, 30 },
		{ 264, 502, 30 },
		{ 264, 503, 30 },
		{ 264, 504, 30 },
		{ 264, 505, 30 },
		{ 264, 506, 30 },
		{ 264, 507, 30 },
		{ 264, 508, 30 },
		{ 264, 509, 30 },
		{ 264, 510, 30 },
		{ 264, 511, 30 },
		{ 264, 512, 30 },
		{ 265, 1, 30 },
		{ 265, 2, 30 },
		{ 265, 3, 30 },
		{ 265, 4, 30 },
		{ 265, 5, 30 },
		{ 265, 6, 30 },
		{ 265, 7, 30 },
		{ 265, 8, 30 },
		{ 265, 9, 30 },
		{ 265, 10, 30 },
		{ 265, 11, 30 },
		{ 265, 12, 30 },
		{ 265, 13, 30 },
		{ 265, 14, 30 },
		{ 265, 15, 30 },
		{ 265, 16, 30 },
		{ 265, 17, 30 },
		{ 265, 18, 30 },
		{ 265, 19, 30 },
		{ 265, 20, 30 },
		{ 265, 21, 30 },
		{ 265, 22, 30 },
		{ 265, 23, 30 },
		{ 265, 24, 30 },
		{ 265, 25, 30 },
		{ 265, 26, 30 },
		{ 265, 27, 30 },
		{ 265, 28, 30 },
		{ 265, 29, 30 },
		{ 265, 30, 30 },
		{ 265, 31, 30 },
		{ 265, 32, 30 },
		{ 265, 33, 30 },
		{ 265, 34, 30 },
		{ 265, 35, 30 },
		{ 265, 36, 30 },
		{ 265, 37, 30 },
		{ 265, 38, 30 },
		{ 265, 39, 30 },
		{ 265, 40, 30 },
		{ 265, 41, 30 },
		{ 265, 42, 30 },
		{ 265, 43, 30 },
		{ 265, 44, 30 },
		{ 265, 45, 30 },
		{ 265, 46, 30 },
		{ 265, 47, 30 },
		{ 265, 48, 30 },
		{ 265, 49, 30 },
		{ 265, 50, 30 },
		{ 265, 51, 30 },
		{ 265, 52, 30 },
		{ 265, 53, 30 },
		{ 265, 54, 30 },
		{ 265, 55, 30 },
		{ 265, 56, 30 },
		{ 265, 57, 30 },
		{ 265, 58, 30 },
		{ 265, 59, 30 },
		{ 265, 60, 30 },
		{ 265, 61, 30 },
		{ 265, 62, 30 },
		{ 265, 63, 30 },
		{ 265, 64, 30 },
		{ 265, 65, 30 },
		{ 265, 66, 30 },
		{ 265, 67, 30 },
		{ 265, 68, 30 },
		{ 265, 69, 30 },
		{ 265, 70, 30 },
		{ 265, 71, 30 },
		{ 265, 72, 30 },
		{ 265, 73, 30 },
		{ 265, 74, 30 },
		{ 265, 75, 30 },
		{ 265, 76, 30 },
		{ 265, 77, 30 },
		{ 265, 78, 30 },
		{ 265, 79, 30 },
		{ 265, 80, 30 },
		{ 265, 81, 30 },
		{ 265, 82, 30 },
		{ 265, 83, 30 },
		{ 265, 84, 30 },
		{ 265, 85, 30 },
		{ 265, 86, 30 },
		{ 265, 87, 30 },
		{ 265, 88, 30 },
		{ 265, 89, 30 },
		{ 265, 90, 30 },
		{ 265, 91, 30 },
		{ 265, 92, 30 },
		{ 265, 93, 30 },
		{ 265, 94, 30 },
		{ 265, 95, 30 },
		{ 265, 96, 30 },
		{ 265, 97, 30 },
		{ 265, 98, 30 },
		{ 265, 99, 30 },
		{ 265, 100, 30 },
		{ 265, 101, 30 },
		{ 265, 102, 30 },
		{ 265, 103, 30 },
		{ 265, 104, 30 },
		{ 265, 105, 30 },
		{ 265, 106, 30 },
		{ 265, 107, 30 },
		{ 265, 108, 30 },
		{ 265, 109, 30 },
		{ 265, 110, 30 },
		{ 265, 111, 30 },
		{ 265, 112, 30 },
		{ 265, 113, 30 },
		{ 265, 114, 30 },
		{ 265, 115, 30 },
		{ 265, 116, 30 },
		{ 265, 117, 30 },
		{ 265, 118, 30 },
		{ 265, 119, 30 },
		{ 265, 120, 30 },
		{ 265, 121, 30 },
		{ 265, 122, 30 },
		{ 265, 123, 30 },
		{ 265, 124, 30 },
		{ 265, 125, 30 },
		{ 265, 126, 30 },
		{ 265, 127, 30 },
		{ 265, 128, 30 },
		{ 265, 129, 30 },
		{ 265, 130, 30 },
		{ 265, 131, 30 },
		{ 265, 132, 30 },
		{ 265, 133, 30 },
		{ 265, 134, 30 },
		{ 265, 135, 30 },
		{ 265, 136, 30 },
		{ 265, 137, 30 },
		{ 265, 138, 30 },
		{ 265, 139, 30 },
		{ 265, 140, 30 },
		{ 265, 141, 30 },
		{ 265, 142, 30 },
		{ 265, 143, 30 },
		{ 265, 144, 30 },
		{ 265, 145, 30 },
		{ 265, 146, 30 },
		{ 265, 147, 30 },
		{ 265, 148, 30 },
		{ 265, 149, 30 },
		{ 265, 150, 30 },
		{ 265, 151, 30 },
		{ 265, 152, 30 },
		{ 265, 153, 30 },
		{ 265, 154, 30 },
		{ 265, 155, 30 },
		{ 265, 156, 30 },
		{ 265, 157, 30 },
		{ 265, 158, 30 },
		{ 265, 159, 30 },
		{ 265, 160, 30 },
		{ 265, 161, 30 },
		{ 265, 162, 30 },
		{ 265, 163, 30 },
		{ 265, 164, 30 },
		{ 265, 165, 30 },
		{ 265, 166, 30 },
		{ 265, 167, 30 },
		{ 265, 168, 30 },
		{ 265, 169, 30 },
		{ 265, 170, 30 },
		{ 265, 171, 30 },
		{ 265, 172, 30 },
		{ 265, 173, 30 },
		{ 265, 174, 30 },
		{ 265, 175, 30 },
		{ 265, 176, 30 },
		{ 265, 177, 30 },
		{ 265, 178, 30 },
		{ 265, 179, 30 },
		{ 265, 180, 30 },
		{ 265, 181, 30 },
		{ 265, 182, 30 },
		{ 265, 183, 30 },
		{ 265, 184, 30 },
		{ 265, 185, 30 },
		{ 265, 186, 30 },
		{ 265, 187, 30 },
		{ 265, 188, 30 },
		{ 265, 189, 30 },
		{ 265, 325, 30 },
		{ 265, 326, 30 },
		{ 265, 327, 30 },
		{ 265, 328, 30 },
		{ 265, 329, 30 },
		{ 265, 330, 30 },
		{ 265, 331, 30 },
		{ 265, 332, 30 },
		{ 265, 333, 30 },
		{ 265, 334, 30 },
		{ 265, 335, 30 },
		{ 265, 336, 30 },
		{ 265, 337, 30 },
		{ 265, 338, 30 },
		{ 265, 339, 30 },
		{ 265, 340, 30 },
		{ 265, 341, 30 },
		{ 265, 342, 30 },
		{ 265, 343, 30 },
		{ 265, 344, 30 },
		{ 265, 345, 30 },
		{ 265, 346, 30 },
		{ 265, 347, 30 },
		{ 265, 348, 30 },
		{ 265, 349, 30 },
		{ 265, 350, 30 },
		{ 265, 351, 30 },
		{ 265, 352, 30 },
		{ 265, 353, 30 },
		{ 265, 354, 30 },
		{ 265, 355, 30 },
		{ 265, 356, 30 },
		{ 265, 357, 30 },
		{ 265, 358, 30 },
		{ 265, 359, 30 },
		{ 265, 360, 30 },
		{ 265, 361, 30 },
		{ 265, 362, 30 },
		{ 265, 363, 30 },
		{ 265, 364, 30 },
		{ 265, 365, 30 },
		{ 265, 366, 30 },
		{ 265, 367, 30 },
		{ 265, 368, 30 },
		{ 265, 369, 30 },
		{ 265, 370, 30 },
		{ 265, 371, 30 },
		{ 265, 372, 30 },
		{ 265, 373, 30 },
		{ 265, 374, 30 },
		{ 265, 375, 30 },
		{ 265, 376, 30 },
		{ 265, 377, 30 },
		{ 265, 378, 30 },
		{ 265, 379, 30 },
		{ 265, 380, 30 },
		{ 265, 381, 30 },
		{ 265, 382, 30 },
		{ 265, 383, 30 },
		{ 265, 384, 30 },
		{ 265, 385, 30 },
		{ 265, 386, 30 },
		{ 265, 387, 30 },
		{ 265, 388, 30 },
		{ 265, 389, 30 },
		{ 265, 390, 30 },
		{ 265, 391, 30 },
		{ 265, 392, 30 },
		{ 265, 393, 30 },
		{ 265, 394, 30 },
		{ 265, 395, 30 },
		{ 265, 396, 30 },
		{ 265, 397, 30 },
		{ 265, 398, 30 },
		{ 265, 399, 30 },
		{ 265, 400, 30 },
		{ 265, 401, 30 },
		{ 265, 402, 30 },
		{ 265, 403, 30 },
		{ 265, 404, 30 },
		{ 265, 405, 30 },
		{ 265, 406, 30 },
		{ 265, 407, 30 },
		{ 265, 408, 30 },
		{ 265, 409, 30 },
		{ 265, 410, 30 },
		{ 265, 411, 30 },
		{ 265, 412, 30 },
		{ 265, 413, 30 },
		{ 265, 414, 30 },
		{ 265, 415, 30 },
		{ 265, 416, 30 },
		{ 265, 417, 30 },
		{ 265, 418, 30 },
		{ 265, 419, 30 },
		{ 265, 420, 30 },
		{ 265, 421, 30 },
		{ 265, 422, 30 },
		{ 265, 423, 30 },
		{ 265, 424, 30 },
		{ 265, 425, 30 },
		{ 265, 426, 30 },
		{ 265, 427, 30 },
		{ 265, 428, 30 },
		{ 265, 429, 30 },
		{ 265, 430, 30 },
		{ 265, 431, 30 },
		{ 265, 432, 30 },
		{ 265, 433, 30 },
		{ 265, 434, 30 },
		{ 265, 435, 30 },
		{ 265, 436, 30 },
		{ 265, 437, 30 },
		{ 265, 438, 30 },
		{ 265, 439, 30 },
		{ 265, 440, 30 },
		{ 265, 441, 30 },
		{ 265, 442, 30 },
		{ 265, 443, 30 },
		{ 265, 444, 30 },
		{ 265, 445, 30 },
		{ 265, 446, 30 },
		{ 265, 447, 30 },
		{ 265, 448, 30 },
		{ 265, 449, 30 },
		{ 265, 450, 30 },
		{ 265, 451, 30 },
		{ 265, 452, 30 },
		{ 265, 453, 30 },
		{ 265, 454, 30 },
		{ 265, 455, 30 },
		{ 265, 456, 30 },
		{ 265, 457, 30 },
		{ 265, 458, 30 },
		{ 265, 459, 30 },
		{ 265, 460, 30 },
		{ 265, 461, 30 },
		{ 265, 462, 30 },
		{ 265, 463, 30 },
		{ 265, 464, 30 },
		{ 265, 465, 30 },
		{ 265, 466, 30 },
		{ 265, 467, 30 },
		{ 265, 468, 30 },
		{ 265, 469, 30 },
		{ 265, 470, 30 },
		{ 265, 471, 30 },
		{ 265, 472, 30 },
		{ 265, 473, 30 },
		{ 265, 474, 30 },
		{ 265, 475, 30 },
		{ 265, 476, 30 },
		{ 265, 477, 30 },
		{ 265, 478, 30 },
		{ 265, 479, 30 },
		{ 265, 480, 30 },
		{ 265, 481, 30 },
		{ 265, 482, 30 },
		{ 265, 483, 30 },
		{ 265, 484, 30 },
		{ 265, 485, 30 },
		{ 265, 486, 30 },
		{ 265, 487, 30 },
		{ 265, 488, 30 },
		{ 265, 489, 30 },
		{ 265, 490, 30 },
		{ 265, 491, 30 },
		{ 265, 492, 30 },
		{ 265, 493, 30 },
		{ 265, 494, 30 },
		{ 265, 495, 30 },
		{ 265, 496, 30 },
		{ 265, 497, 30 },
		{ 265, 498, 30 },
		{ 265, 499, 30 },
		{ 265, 500, 30 },
		{ 265, 501, 30 },
		{ 265, 502, 30 },
		{ 265, 503, 30 },
		{ 265, 504, 30 },
		{ 265, 505, 30 },
		{ 265, 506, 30 },
		{ 265, 507, 30 },
		{ 265, 508, 30 },
		{ 265, 509, 30 },
		{ 265, 510, 30 },
		{ 265, 511, 30 },
		{ 265, 512, 30 },
		{ 266, 1, 30 },
		{ 266, 2, 30 },
		{ 266, 3, 30 },
		{ 266, 4, 30 },
		{ 266, 5, 30 },
		{ 266, 6, 30 },
		{ 266, 7, 30 },
		{ 266, 8, 30 },
		{ 266, 9, 30 },
		{ 266, 10, 30 },
		{ 266, 11, 30 },
		{ 266, 12, 30 },
		{ 266, 13, 30 },
		{ 266, 14, 30 },
		{ 266, 15, 30 },
		{ 266, 16, 30 },
		{ 266, 17, 30 },
		{ 266, 18, 30 },
		{ 266, 19, 30 },
		{ 266, 20, 30 },
		{ 266, 21, 30 },
		{ 266, 22, 30 },
		{ 266, 23, 30 },
		{ 266, 24, 30 },
		{ 266, 25, 30 },
		{ 266, 26, 30 },
		{ 266, 27, 30 },
		{ 266, 28, 30 },
		{ 266, 29, 30 },
		{ 266, 30, 30 },
		{ 266, 31, 30 },
		{ 266, 32, 30 },
		{ 266, 33, 30 },
		{ 266, 34, 30 },
		{ 266, 35, 30 },
		{ 266, 36, 30 },
		{ 266, 37, 30 },
		{ 266, 38, 30 },
		{ 266, 39, 30 },
		{ 266, 40, 30 },
		{ 266, 41, 30 },
		{ 266, 42, 30 },
		{ 266, 43, 30 },
		{ 266, 44, 30 },
		{ 266, 45, 30 },
		{ 266, 46, 30 },
		{ 266, 47, 30 },
		{ 266, 48, 30 },
		{ 266, 49, 30 },
		{ 266, 50, 30 },
		{ 266, 51, 30 },
		{ 266, 52, 30 },
		{ 266, 53, 30 },
		{ 266, 54, 30 },
		{ 266, 55, 30 },
		{ 266, 56, 30 },
		{ 266, 57, 30 },
		{ 266, 58, 30 },
		{ 266, 59, 30 },
		{ 266, 60, 30 },
		{ 266, 61, 30 },
		{ 266, 62, 30 },
		{ 266, 63, 30 },
		{ 266, 64, 30 },
		{ 266, 65, 30 },
		{ 266, 66, 30 },
		{ 266, 67, 30 },
		{ 266, 68, 30 },
		{ 266, 69, 30 },
		{ 266, 70, 30 },
		{ 266, 71, 30 },
		{ 266, 72, 30 },
		{ 266, 73, 30 },
		{ 266, 74, 30 },
		{ 266, 75, 30 },
		{ 266, 76, 30 },
		{ 266, 77, 30 },
		{ 266, 78, 30 },
		{ 266, 79, 30 },
		{ 266, 80, 30 },
		{ 266, 81, 30 },
		{ 266, 82, 30 },
		{ 266, 83, 30 },
		{ 266, 84, 30 },
		{ 266, 85, 30 },
		{ 266, 86, 30 },
		{ 266, 87, 30 },
		{ 266, 88, 30 },
		{ 266, 89, 30 },
		{ 266, 90, 30 },
		{ 266, 91, 30 },
		{ 266, 92, 30 },
		{ 266, 93, 30 },
		{ 266, 94, 30 },
		{ 266, 95, 30 },
		{ 266, 96, 30 },
		{ 266, 97, 30 },
		{ 266, 98, 30 },
		{ 266, 99, 30 },
		{ 266, 100, 30 },
		{ 266, 101, 30 },
		{ 266, 102, 30 },
		{ 266, 103, 30 },
		{ 266, 104, 30 },
		{ 266, 105, 30 },
		{ 266, 106, 30 },
		{ 266, 107, 30 },
		{ 266, 108, 30 },
		{ 266, 109, 30 },
		{ 266, 110, 30 },
		{ 266, 111, 30 },
		{ 266, 112, 30 },
		{ 266, 113, 30 },
		{ 266, 114, 30 },
		{ 266, 115, 30 },
		{ 266, 116, 30 },
		{ 266, 117, 30 },
		{ 266, 118, 30 },
		{ 266, 119, 30 },
		{ 266, 120, 30 },
		{ 266, 121, 30 },
		{ 266, 122, 30 },
		{ 266, 123, 30 },
		{ 266, 124, 30 },
		{ 266, 125, 30 },
		{ 266, 126, 30 },
		{ 266, 127, 30 },
		{ 266, 128, 30 },
		{ 266, 129, 30 },
		{ 266, 130, 30 },
		{ 266, 131, 30 },
		{ 266, 132, 30 },
		{ 266, 133, 30 },
		{ 266, 134, 30 },
		{ 266, 135, 30 },
		{ 266, 136, 30 },
		{ 266, 137, 30 },
		{ 266, 138, 30 },
		{ 266, 139, 30 },
		{ 266, 140, 30 },
		{ 266, 141, 30 },
		{ 266, 142, 30 },
		{ 266, 143, 30 },
		{ 266, 144, 30 },
		{ 266, 145, 30 },
		{ 266, 146, 30 },
		{ 266, 147, 30 },
		{ 266, 148, 30 },
		{ 266, 149, 30 },
		{ 266, 150, 30 },
		{ 266, 151, 30 },
		{ 266, 152, 30 },
		{ 266, 153, 30 },
		{ 266, 154, 30 },
		{ 266, 155, 30 },
		{ 266, 156, 30 },
		{ 266, 157, 30 },
		{ 266, 158, 30 },
		{ 266, 159, 30 },
		{ 266, 160, 30 },
		{ 266, 161, 30 },
		{ 266, 162, 30 },
		{ 266, 163, 30 },
		{ 266, 164, 30 },
		{ 266, 165, 30 },
		{ 266, 166, 30 },
		{ 266, 167, 30 },
		{ 266, 168, 30 },
		{ 266, 169, 30 },
		{ 266, 170, 30 },
		{ 266, 171, 30 },
		{ 266, 172, 30 },
		{ 266, 173, 30 },
		{ 266, 174, 30 },
		{ 266, 175, 30 },
		{ 266, 176, 30 },
		{ 266, 177, 30 },
		{ 266, 178, 30 },
		{ 266, 179, 30 },
		{ 266, 180, 30 },
		{ 266, 181, 30 },
		{ 266, 182, 30 },
		{ 266, 183, 30 },
		{ 266, 184, 30 },
		{ 266, 330, 30 },
		{ 266, 331, 30 },
		{ 266, 332, 30 },
		{ 266, 333, 30 },
		{ 266, 334, 30 },
		{ 266, 335, 30 },
		{ 266, 336, 30 },
		{ 266, 337, 30 },
		{ 266, 338, 30 },
		{ 266, 339, 30 },
		{ 266, 340, 30 },
		{ 266, 341, 30 },
		{ 266, 342, 30 },
		{ 266, 343, 30 },
		{ 266, 344, 30 },
		{ 266, 345, 30 },
		{ 266, 346, 30 },
		{ 266, 347, 30 },
		{ 266, 348, 30 },
		{ 266, 349, 30 },
		{ 266, 350, 30 },
		{ 266, 351, 30 },
		{ 266, 352, 30 },
		{ 266, 353, 30 },
		{ 266, 354, 30 },
		{ 266, 355, 30 },
		{ 266, 356, 30 },
		{ 266, 357, 30 },
		{ 266, 358, 30 },
		{ 266, 359, 30 },
		{ 266, 360, 30 },
		{ 266, 361, 30 },
		{ 266, 362, 30 },
		{ 266, 363, 30 },
		{ 266, 364, 30 },
		{ 266, 365, 30 },
		{ 266, 366, 30 },
		{ 266, 367, 30 },
		{ 266, 368, 30 },
		{ 266, 369, 30 },
		{ 266, 370, 30 },
		{ 266, 371, 30 },
		{ 266, 372, 30 },
		{ 266, 373, 30 },
		{ 266, 374, 30 },
		{ 266, 375, 30 },
		{ 266, 376, 30 },
		{ 266, 377, 30 },
		{ 266, 378, 30 },
		{ 266, 379, 30 },
		{ 266, 380, 30 },
		{ 266, 381, 30 },
		{ 266, 382, 30 },
		{ 266, 383, 30 },
		{ 266, 384, 30 },
		{ 266, 385, 30 },
		{ 266, 386, 30 },
		{ 266, 387, 30 },
		{ 266, 388, 30 },
		{ 266, 389, 30 },
		{ 266, 390, 30 },
		{ 266, 391, 30 },
		{ 266, 392, 30 },
		{ 266, 393, 30 },
		{ 266, 394, 30 },
		{ 266, 395, 30 },
		{ 266, 396, 30 },
		{ 266, 397, 30 },
		{ 266, 398, 30 },
		{ 266, 399, 30 },
		{ 266, 400, 30 },
		{ 266, 401, 30 },
		{ 266, 402, 30 },
		{ 266, 403, 30 },
		{ 266, 404, 30 },
		{ 266, 405, 30 },
		{ 266, 406, 30 },
		{ 266, 407, 30 },
		{ 266, 408, 30 },
		{ 266, 409, 30 },
		{ 266, 410, 30 },
		{ 266, 411, 30 },
		{ 266, 412, 30 },
		{ 266, 413, 30 },
		{ 266, 414, 30 },
		{ 266, 415, 30 },
		{ 266, 416, 30 },
		{ 266, 417, 30 },
		{ 266, 418, 30 },
		{ 266, 419, 30 },
		{ 266, 420, 30 },
		{ 266, 421, 30 },
		{ 266, 422, 30 },
		{ 266, 423, 30 },
		{ 266, 424, 30 },
		{ 266, 425, 30 },
		{ 266, 426, 30 },
		{ 266, 427, 30 },
		{ 266, 428, 30 },
		{ 266, 429, 30 },
		{ 266, 430, 30 },
		{ 266, 431, 30 },
		{ 266, 432, 30 },
		{ 266, 433, 30 },
		{ 266, 434, 30 },
		{ 266, 435, 30 },
		{ 266, 436, 30 },
		{ 266, 437, 30 },
		{ 266, 438, 30 },
		{ 266, 439, 30 },
		{ 266, 440, 30 },
		{ 266, 441, 30 },
		{ 266, 442, 30 },
		{ 266, 443, 30 },
		{ 266, 444, 30 },
		{ 266, 445, 30 },
		{ 266, 446, 30 },
		{ 266, 447, 30 },
		{ 266, 448, 30 },
		{ 266, 449, 30 },
		{ 266, 450, 30 },
		{ 266, 451, 30 },
		{ 266, 452, 30 },
		{ 266, 453, 30 },
		{ 266, 454, 30 },
		{ 266, 455, 30 },
		{ 266, 456, 30 },
		{ 266, 457, 30 },
		{ 266, 458, 30 },
		{ 266, 459, 30 },
		{ 266, 460, 30 },
		{ 266, 461, 30 },
		{ 266, 462, 30 },
		{ 266, 463, 30 },
		{ 266, 464, 30 },
		{ 266, 465, 30 },
		{ 266, 466, 30 },
		{ 266, 467, 30 },
		{ 266, 468, 30 },
		{ 266, 469, 30 },
		{ 266, 470, 30 },
		{ 266, 471, 30 },
		{ 266, 472, 30 },
		{ 266, 473, 30 },
		{ 266, 474, 30 },
		{ 266, 475, 30 },
		{ 266, 476, 30 },
		{ 266, 477, 30 },
		{ 266, 478, 30 },
		{ 266, 479, 30 },
		{ 266, 480, 30 },
		{ 266, 481, 30 },
		{ 266, 482, 30 },
		{ 266, 483, 30 },
		{ 266, 484, 30 },
		{ 266, 485, 30 },
		{ 266, 486, 30 },
		{ 266, 487, 30 },
		{ 266, 488, 30 },
		{ 266, 489, 30 },
		{ 266, 490, 30 },
		{ 266, 491, 30 },
		{ 266, 492, 30 },
		{ 266, 493, 30 },
		{ 266, 494, 30 },
		{ 266, 495, 30 },
		{ 266, 496, 30 },
		{ 266, 497, 30 },
		{ 266, 498, 30 },
		{ 266, 499, 30 },
		{ 266, 500, 30 },
		{ 266, 501, 30 },
		{ 266, 502, 30 },
		{ 266, 503, 30 },
		{ 266, 504, 30 },
		{ 266, 505, 30 },
		{ 266, 506, 30 },
		{ 266, 507, 30 },
		{ 266, 508, 30 },
		{ 266, 509, 30 },
		{ 266, 510, 30 },
		{ 266, 511, 30 },
		{ 266, 512, 30 },
		{ 267, 1, 30 },
		{ 267, 2, 30 },
		{ 267, 3, 30 },
		{ 267, 4, 30 },
		{ 267, 5, 30 },
		{ 267, 6, 30 },
		{ 267, 7, 30 },
		{ 267, 8, 30 },
		{ 267, 9, 30 },
		{ 267, 10, 30 },
		{ 267, 11, 30 },
		{ 267, 12, 30 },
		{ 267, 13, 30 },
		{ 267, 14, 30 },
		{ 267, 15, 30 },
		{ 267, 16, 30 },
		{ 267, 17, 30 },
		{ 267, 18, 30 },
		{ 267, 19, 30 },
		{ 267, 20, 30 },
		{ 267, 21, 30 },
		{ 267, 22, 30 },
		{ 267, 23, 30 },
		{ 267, 24, 30 },
		{ 267, 25, 30 },
		{ 267, 26, 30 },
		{ 267, 27, 30 },
		{ 267, 28, 30 },
		{ 267, 29, 30 },
		{ 267, 30, 30 },
		{ 267, 31, 30 },
		{ 267, 32, 30 },
		{ 267, 33, 30 },
		{ 267, 34, 30 },
		{ 267, 35, 30 },
		{ 267, 36, 30 },
		{ 267, 37, 30 },
		{ 267, 38, 30 },
		{ 267, 39, 30 },
		{ 267, 40, 30 },
		{ 267, 41, 30 },
		{ 267, 42, 30 },
		{ 267, 43, 30 },
		{ 267, 44, 30 },
		{ 267, 45, 30 },
		{ 267, 46, 30 },
		{ 267, 47, 30 },
		{ 267, 48, 30 },
		{ 267, 49, 30 },
		{ 267, 50, 30 },
		{ 267, 51, 30 },
		{ 267, 52, 30 },
		{ 267, 53, 30 },
		{ 267, 54, 30 },
		{ 267, 55, 30 },
		{ 267, 56, 30 },
		{ 267, 57, 30 },
		{ 267, 58, 30 },
		{ 267, 59, 30 },
		{ 267, 60, 30 },
		{ 267, 61, 30 },
		{ 267, 62, 30 },
		{ 267, 63, 30 },
		{ 267, 64, 30 },
		{ 267, 65, 30 },
		{ 267, 66, 30 },
		{ 267, 67, 30 },
		{ 267, 68, 30 },
		{ 267, 69, 30 },
		{ 267, 70, 30 },
		{ 267, 71, 30 },
		{ 267, 72, 30 },
		{ 267, 73, 30 },
		{ 267, 74, 30 },
		{ 267, 75, 30 },
		{ 267, 76, 30 },
		{ 267, 77, 30 },
		{ 267, 78, 30 },
		{ 267, 79, 30 },
		{ 267, 80, 30 },
		{ 267, 81, 30 },
		{ 267, 82, 30 },
		{ 267, 83, 30 },
		{ 267, 84, 30 },
		{ 267, 85, 30 },
		{ 267, 86, 30 },
		{ 267, 87, 30 },
		{ 267, 88, 30 },
		{ 267, 89, 30 },
		{ 267, 90, 30 },
		{ 267, 91, 30 },
		{ 267, 92, 30 },
		{ 267, 93, 30 },
		{ 267, 94, 30 },
		{ 267, 95, 30 },
		{ 267, 96, 30 },
		{ 267, 97, 30 },
		{ 267, 98, 30 },
		{ 267, 99, 30 },
		{ 267, 100, 30 },
		{ 267, 101, 30 },
		{ 267, 102, 30 },
		{ 267, 103, 30 },
		{ 267, 104, 30 },
		{ 267, 105, 30 },
		{ 267, 106, 30 },
		{ 267, 107, 30 },
		{ 267, 108, 30 },
		{ 267, 109, 30 },
		{ 267, 110, 30 },
		{ 267, 111, 30 },
		{ 267, 112, 30 },
		{ 267, 113, 30 },
		{ 267, 114, 30 },
		{ 267, 115, 30 },
		{ 267, 116, 30 },
		{ 267, 117, 30 },
		{ 267, 118, 30 },
		{ 267, 119, 30 },
		{ 267, 120, 30 },
		{ 267, 121, 30 },
		{ 267, 122, 30 },
		{ 267, 123, 30 },
		{ 267, 124, 30 },
		{ 267, 125, 30 },
		{ 267, 126, 30 },
		{ 267, 127, 30 },
		{ 267, 128, 30 },
		{ 267, 129, 30 },
		{ 267, 130, 30 },
		{ 267, 131, 30 },
		{ 267, 132, 30 },
		{ 267, 133, 30 },
		{ 267, 134, 30 },
		{ 267, 135, 30 },
		{ 267, 136, 30 },
		{ 267, 137, 30 },
		{ 267, 138, 30 },
		{ 267, 139, 30 },
		{ 267, 140, 30 },
		{ 267, 141, 30 },
		{ 267, 142, 30 },
		{ 267, 143, 30 },
		{ 267, 144, 30 },
		{ 267, 145, 30 },
		{ 267, 146, 30 },
		{ 267, 147, 30 },
		{ 267, 148, 30 },
		{ 267, 149, 30 },
		{ 267, 150, 30 },
		{ 267, 151, 30 },
		{ 267, 152, 30 },
		{ 267, 153, 30 },
		{ 267, 154, 30 },
		{ 267, 155, 30 },
		{ 267, 156, 30 },
		{ 267, 157, 30 },
		{ 267, 158, 30 },
		{ 267, 159, 30 },
		{ 267, 160, 30 },
		{ 267, 161, 30 },
		{ 267, 162, 30 },
		{ 267, 163, 30 },
		{ 267, 164, 30 },
		{ 267, 165, 30 },
		{ 267, 166, 30 },
		{ 267, 167, 30 },
		{ 267, 168, 30 },
		{ 267, 169, 30 },
		{ 267, 170, 30 },
		{ 267, 171, 30 },
		{ 267, 172, 30 },
		{ 267, 173, 30 },
		{ 267, 174, 30 },
		{ 267, 175, 30 },
		{ 267, 176, 30 },
		{ 267, 338, 30 },
		{ 267, 339, 30 },
		{ 267, 340, 30 },
		{ 267, 341, 30 },
		{ 267, 342, 30 },
		{ 267, 343, 30 },
		{ 267, 344, 30 },
		{ 267, 345, 30 },
		{ 267, 346, 30 },
		{ 267, 347, 30 },
		{ 267, 348, 30 },
		{ 267, 349, 30 },
		{ 267, 350, 30 },
		{ 267, 351, 30 },
		{ 267, 352, 30 },
		{ 267, 353, 30 },
		{ 267, 354, 30 },
		{ 267, 355, 30 },
		{ 267, 356, 30 },
		{ 267, 357, 30 },
		{ 267, 358, 30 },
		{ 267, 359, 30 },
		{ 267, 360, 30 },
		{ 267, 361, 30 },
		{ 267, 362, 30 },
		{ 267, 363, 30 },
		{ 267, 364, 30 },
		{ 267, 365, 30 },
		{ 267, 366, 30 },
		{ 267, 367, 30 },
		{ 267, 368, 30 },
		{ 267, 369, 30 },
		{ 267, 370, 30 },
		{ 267, 371, 30 },
		{ 267, 372, 30 },
		{ 267, 373, 30 },
		{ 267, 374, 30 },
		{ 267, 375, 30 },
		{ 267, 376, 30 },
		{ 267, 377, 30 },
		{ 267, 378, 30 },
		{ 267, 379, 30 },
		{ 267, 380, 30 },
		{ 267, 381, 30 },
		{ 267, 382, 30 },
		{ 267, 383, 30 },
		{ 267, 384, 30 },
		{ 267, 385, 30 },
		{ 267, 386, 30 },
		{ 267, 387, 30 },
		{ 267, 388, 30 },
		{ 267, 389, 30 },
		{ 267, 390, 30 },
		{ 267, 391, 30 },
		{ 267, 392, 30 },
		{ 267, 393, 30 },
		{ 267, 394, 30 },
		{ 267, 395, 30 },
		{ 267, 396, 30 },
		{ 267, 397, 30 },
		{ 267, 398, 30 },
		{ 267, 399, 30 },
		{ 267, 400, 30 },
		{ 267, 401, 30 },
		{ 267, 402, 30 },
		{ 267, 403, 30 },
		{ 267, 404, 30 },
		{ 267, 405, 30 },
		{ 267, 406, 30 },
		{ 267, 407, 30 },
		{ 267, 408, 30 },
		{ 267, 409, 30 },
		{ 267, 410, 30 },
		{ 267, 411, 30 },
		{ 267, 412, 30 },
		{ 267, 413, 30 },
		{ 267, 414, 30 },
		{ 267, 415, 30 },
		{ 267, 416, 30 },
		{ 267, 417, 30 },
		{ 267, 418, 30 },
		{ 267, 419, 30 },
		{ 267, 420, 30 },
		{ 267, 421, 30 },
		{ 267, 422, 30 },
		{ 267, 423, 30 },
		{ 267, 424, 30 },
		{ 267, 425, 30 },
		{ 267, 426, 30 },
		{ 267, 427, 30 },
		{ 267, 428, 30 },
		{ 267, 429, 30 },
		{ 267, 430, 30 },
		{ 267, 431, 30 },
		{ 267, 432, 30 },
		{ 267, 433, 30 },
		{ 267, 434, 30 },
		{ 267, 435, 30 },
		{ 267, 436, 30 },
		{ 267, 437, 30 },
		{ 267, 438, 30 },
		{ 267, 439, 30 },
		{ 267, 440, 30 },
		{ 267, 441, 30 },
		{ 267, 442, 30 },
		{ 267, 443, 30 },
		{ 267, 444, 30 },
		{ 267, 445, 30 },
		{ 267, 446, 30 },
		{ 267, 447, 30 },
		{ 267, 448, 30 },
		{ 267, 449, 30 },
		{ 267, 450, 30 },
		{ 267, 451, 30 },
		{ 267, 452, 30 },
		{ 267, 453, 30 },
		{ 267, 454, 30 },
		{ 267, 455, 30 },
		{ 267, 456, 30 },
		{ 267, 457, 30 },
		{ 267, 458, 30 },
		{ 267, 459, 30 },
		{ 267, 460, 30 },
		{ 267, 461, 30 },
		{ 267, 462, 30 },
		{ 267, 463, 30 },
		{ 267, 464, 30 },
		{ 267, 465, 30 },
		{ 267, 466, 30 },
		{ 267, 467, 30 },
		{ 267, 468, 30 },
		{ 267, 469, 30 },
		{ 267, 470, 30 },
		{ 267, 471, 30 },
		{ 267, 472, 30 },
		{ 267, 473, 30 },
		{ 267, 474, 30 },
		{ 267, 475, 30 },
		{ 267, 476, 30 },
		{ 267, 477, 30 },
		{ 267, 478, 30 },
		{ 267, 479, 30 },
		{ 267, 480, 30 },
		{ 267, 481, 30 },
		{ 267, 482, 30 },
		{ 267, 483, 30 },
		{ 267, 484, 30 },
		{ 267, 485, 30 },
		{ 267, 486, 30 },
		{ 267, 487, 30 },
		{ 267, 488, 30 },
		{ 267, 489, 30 },
		{ 267, 490, 30 },
		{ 267, 491, 30 },
		{ 267, 492, 30 },
		{ 267, 493, 30 },
		{ 267, 494, 30 },
		{ 267, 495, 30 },
		{ 267, 496, 30 },
		{ 267, 497, 30 },
		{ 267, 498, 30 },
		{ 267, 499, 30 },
		{ 267, 500, 30 },
		{ 267, 501, 30 },
		{ 267, 502, 30 },
		{ 267, 503, 30 },
		{ 267, 504, 30 },
		{ 267, 505, 30 },
		{ 267, 506, 30 },
		{ 267, 507, 30 },
		{ 267, 508, 30 },
		{ 267, 509, 30 },
		{ 267, 510, 30 },
		{ 267, 511, 30 },
		{ 267, 512, 30 },
		{ 268, 1, 30 },
		{ 268, 2, 30 },
		{ 268, 3, 30 },
		{ 268, 4, 30 },
		{ 268, 5, 30 },
		{ 268, 6, 30 },
		{ 268, 7, 30 },
		{ 268, 8, 30 },
		{ 268, 9, 30 },
		{ 268, 10, 30 },
		{ 268, 11, 30 },
		{ 268, 12, 30 },
		{ 268, 13, 30 },
		{ 268, 14, 30 },
		{ 268, 15, 30 },
		{ 268, 16, 30 },
		{ 268, 17, 30 },
		{ 268, 18, 30 },
		{ 268, 19, 30 },
		{ 268, 20, 30 },
		{ 268, 21, 30 },
		{ 268, 22, 30 },
		{ 268, 23, 30 },
		{ 268, 24, 30 },
		{ 268, 25, 30 },
		{ 268, 26, 30 },
		{ 268, 27, 30 },
		{ 268, 28, 30 },
		{ 268, 29, 30 },
		{ 268, 30, 30 },
		{ 268, 31, 30 },
		{ 268, 32, 30 },
		{ 268, 33, 30 },
		{ 268, 34, 30 },
		{ 268, 35, 30 },
		{ 268, 36, 30 },
		{ 268, 37, 30 },
		{ 268, 38, 30 },
		{ 268, 39, 30 },
		{ 268, 40, 30 },
		{ 268, 41, 30 },
		{ 268, 42, 30 },
		{ 268, 43, 30 },
		{ 268, 44, 30 },
		{ 268, 45, 30 },
		{ 268, 46, 30 },
		{ 268, 47, 30 },
		{ 268, 48, 30 },
		{ 268, 49, 30 },
		{ 268, 50, 30 },
		{ 268, 51, 30 },
		{ 268, 52, 30 },
		{ 268, 53, 30 },
		{ 268, 54, 30 },
		{ 268, 55, 30 },
		{ 268, 56, 30 },
		{ 268, 57, 30 },
		{ 268, 58, 30 },
		{ 268, 59, 30 },
		{ 268, 60, 30 },
		{ 268, 61, 30 },
		{ 268, 62, 30 },
		{ 268, 63, 30 },
		{ 268, 64, 30 },
		{ 268, 65, 30 },
		{ 268, 66, 30 },
		{ 268, 67, 30 },
		{ 268, 68, 30 },
		{ 268, 69, 30 },
		{ 268, 70, 30 },
		{ 268, 71, 30 },
		{ 268, 72, 30 },
		{ 268, 73, 30 },
		{ 268, 74, 30 },
		{ 268, 75, 30 },
		{ 268, 76, 30 },
		{ 268, 77, 30 },
		{ 268, 78, 30 },
		{ 268, 79, 30 },
		{ 268, 80, 30 },
		{ 268, 81, 30 },
		{ 268, 82, 30 },
		{ 268, 83, 30 },
		{ 268, 84, 30 },
		{ 268, 85, 30 },
		{ 268, 86, 30 },
		{ 268, 87, 30 },
		{ 268, 88, 30 },
		{ 268, 89, 30 },
		{ 268, 90, 30 },
		{ 268, 91, 30 },
		{ 268, 92, 30 },
		{ 268, 93, 30 },
		{ 268, 94, 30 },
		{ 268, 95, 30 },
		{ 268, 96, 30 },
		{ 268, 97, 30 },
		{ 268, 98, 30 },
		{ 268, 99, 30 },
		{ 268, 100, 30 },
		{ 268, 101, 30 },
		{ 268, 102, 30 },
		{ 268, 103, 30 },
		{ 268, 104, 30 },
		{ 268, 105, 30 },
		{ 268, 106, 30 },
		{ 268, 107, 30 },
		{ 268, 108, 30 },
		{ 268, 109, 30 },
		{ 268, 110, 30 },
		{ 268, 111, 30 },
		{ 268, 112, 30 },
		{ 268, 113, 30 },
		{ 268, 114, 30 },
		{ 268, 115, 30 },
		{ 268, 116, 30 },
		{ 268, 117, 30 },
		{ 268, 118, 30 },
		{ 268, 119, 30 },
		{ 268, 120, 30 },
		{ 268, 121, 30 },
		{ 268, 122, 30 },
		{ 268, 123, 30 },
		{ 268, 124, 30 },
		{ 268, 125, 30 },
		{ 268, 126, 30 },
		{ 268, 127, 30 },
		{ 268, 128, 30 },
		{ 268, 129, 30 },
		{ 268, 130, 30 },
		{ 268, 131, 30 },
		{ 268, 132, 30 },
		{ 268, 133, 30 },
		{ 268, 134, 30 },
		{ 268, 135, 30 },
		{ 268, 136, 30 },
		{ 268, 137, 30 },
		{ 268, 138, 30 },
		{ 268, 139, 30 },
		{ 268, 140, 30 },
		{ 268, 141, 30 },
		{ 268, 142, 30 },
		{ 268, 143, 30 },
		{ 268, 144, 30 },
		{ 268, 145, 30 },
		{ 268, 146, 30 },
		{ 268, 147, 30 },
		{ 268, 148, 30 },
		{ 268, 149, 30 },
		{ 268, 150, 30 },
		{ 268, 151, 30 },
		{ 268, 152, 30 },
		{ 268, 153, 30 },
		{ 268, 154, 30 },
		{ 268, 155, 30 },
		{ 268, 156, 30 },
		{ 268, 157, 30 },
		{ 268, 158, 30 },
		{ 268, 159, 30 },
		{ 268, 160, 30 },
		{ 268, 161, 30 },
		{ 268, 162, 30 },
		{ 268, 163, 30 },
		{ 268, 164, 30 },
		{ 268, 165, 30 },
		{ 268, 166, 30 },
		{ 268, 167, 30 },
		{ 268, 168, 30 },
		{ 268, 346, 30 },
		{ 268, 347, 30 },
		{ 268, 348, 30 },
		{ 268, 349, 30 },
		{ 268, 350, 30 },
		{ 268, 351, 30 },
		{ 268, 352, 30 },
		{ 268, 353, 30 },
		{ 268, 354, 30 },
		{ 268, 355, 30 },
		{ 268, 356, 30 },
		{ 268, 357, 30 },
		{ 268, 358, 30 },
		{ 268, 359, 30 },
		{ 268, 360, 30 },
		{ 268, 361, 30 },
		{ 268, 362, 30 },
		{ 268, 363, 30 },
		{ 268, 364, 30 },
		{ 268, 365, 30 },
		{ 268, 366, 30 },
		{ 268, 367, 30 },
		{ 268, 368, 30 },
		{ 268, 369, 30 },
		{ 268, 370, 30 },
		{ 268, 371, 30 },
		{ 268, 372, 30 },
		{ 268, 373, 30 },
		{ 268, 374, 30 },
		{ 268, 375, 30 },
		{ 268, 376, 30 },
		{ 268, 377, 30 },
		{ 268, 378, 30 },
		{ 268, 379, 30 },
		{ 268, 380, 30 },
		{ 268, 381, 30 },
		{ 268, 382, 30 },
		{ 268, 383, 30 },
		{ 268, 384, 30 },
		{ 268, 385, 30 },
		{ 268, 386, 30 },
		{ 268, 387, 30 },
		{ 268, 388, 30 },
		{ 268, 389, 30 },
		{ 268, 390, 30 },
		{ 268, 391, 30 },
		{ 268, 392, 30 },
		{ 268, 393, 30 },
		{ 268, 394, 30 },
		{ 268, 395, 30 },
		{ 268, 396, 30 },
		{ 268, 397, 30 },
		{ 268, 398, 30 },
		{ 268, 399, 30 },
		{ 268, 400, 30 },
		{ 268, 401, 30 },
		{ 268, 402, 30 },
		{ 268, 403, 30 },
		{ 268, 404, 30 },
		{ 268, 405, 30 },
		{ 268, 406, 30 },
		{ 268, 407, 30 },
		{ 268, 408, 30 },
		{ 268, 409, 30 },
		{ 268, 410, 30 },
		{ 268, 411, 30 },
		{ 268, 412, 30 },
		{ 268, 413, 30 },
		{ 268, 414, 30 },
		{ 268, 415, 30 },
		{ 268, 416, 30 },
		{ 268, 417, 30 },
		{ 268, 418, 30 },
		{ 268, 419, 30 },
		{ 268, 420, 30 },
		{ 268, 421, 30 },
		{ 268, 422, 30 },
		{ 268, 423, 30 },
		{ 268, 424, 30 },
		{ 268, 425, 30 },
		{ 268, 426, 30 },
		{ 268, 427, 30 },
		{ 268, 428, 30 },
		{ 268, 429, 30 },
		{ 268, 430, 30 },
		{ 268, 431, 30 },
		{ 268, 432, 30 },
		{ 268, 433, 30 },
		{ 268, 434, 30 },
		{ 268, 435, 30 },
		{ 268, 436, 30 },
		{ 268, 437, 30 },
		{ 268, 438, 30 },
		{ 268, 439, 30 },
		{ 268, 440, 30 },
		{ 268, 441, 30 },
		{ 268, 442, 30 },
		{ 268, 443, 30 },
		{ 268, 444, 30 },
		{ 268, 445, 30 },
		{ 268, 446, 30 },
		{ 268, 447, 30 },
		{ 268, 448, 30 },
		{ 268, 449, 30 },
		{ 268, 450, 30 },
		{ 268, 451, 30 },
		{ 268, 452, 30 },
		{ 268, 453, 30 },
		{ 268, 454, 30 },
		{ 268, 455, 30 },
		{ 268, 456, 30 },
		{ 268, 457, 30 },
		{ 268, 458, 30 },
		{ 268, 459, 30 },
		{ 268, 460, 30 },
		{ 268, 461, 30 },
		{ 268, 462, 30 },
		{ 268, 463, 30 },
		{ 268, 464, 30 },
		{ 268, 465, 30 },
		{ 268, 466, 30 },
		{ 268, 467, 30 },
		{ 268, 468, 30 },
		{ 268, 469, 30 },
		{ 268, 470, 30 },
		{ 268, 471, 30 },
		{ 268, 472, 30 },
		{ 268, 473, 30 },
		{ 268, 474, 30 },
		{ 268, 475, 30 },
		{ 268, 476, 30 },
		{ 268, 477, 30 },
		{ 268, 478, 30 },
		{ 268, 479, 30 },
		{ 268, 480, 30 },
		{ 268, 481, 30 },
		{ 268, 482, 30 },
		{ 268, 483, 30 },
		{ 268, 484, 30 },
		{ 268, 485, 30 },
		{ 268, 486, 30 },
		{ 268, 487, 30 },
		{ 268, 488, 30 },
		{ 268, 489, 30 },
		{ 268, 490, 30 },
		{ 268, 491, 30 },
		{ 268, 492, 30 },
		{ 268, 493, 30 },
		{ 268, 494, 30 },
		{ 268, 495, 30 },
		{ 268, 496, 30 },
		{ 268, 497, 30 },
		{ 268, 498, 30 },
		{ 268, 499, 30 },
		{ 268, 500, 30 },
		{ 268, 501, 30 },
		{ 268, 502, 30 },
		{ 268, 503, 30 },
		{ 268, 504, 30 },
		{ 268, 505, 30 },
		{ 268, 506, 30 },
		{ 268, 507, 30 },
		{ 268, 508, 30 },
		{ 268, 509, 30 },
		{ 268, 510, 30 },
		{ 268, 511, 30 },
		{ 268, 512, 30 },
		{ 269, 1, 30 },
		{ 269, 2, 30 },
		{ 269, 3, 30 },
		{ 269, 4, 30 },
		{ 269, 5, 30 },
		{ 269, 6, 30 },
		{ 269, 7, 30 },
		{ 269, 8, 30 },
		{ 269, 9, 30 },
		{ 269, 10, 30 },
		{ 269, 11, 30 },
		{ 269, 12, 30 },
		{ 269, 13, 30 },
		{ 269, 14, 30 },
		{ 269, 15, 30 },
		{ 269, 16, 30 },
		{ 269, 17, 30 },
		{ 269, 18, 30 },
		{ 269, 19, 30 },
		{ 269, 20, 30 },
		{ 269, 21, 30 },
		{ 269, 22, 30 },
		{ 269, 23, 30 },
		{ 269, 24, 30 },
		{ 269, 25, 30 },
		{ 269, 26, 30 },
		{ 269, 27, 30 },
		{ 269, 28, 30 },
		{ 269, 29, 30 },
		{ 269, 30, 30 },
		{ 269, 31, 30 },
		{ 269, 32, 30 },
		{ 269, 33, 30 },
		{ 269, 34, 30 },
		{ 269, 35, 30 },
		{ 269, 36, 30 },
		{ 269, 37, 30 },
		{ 269, 38, 30 },
		{ 269, 39, 30 },
		{ 269, 40, 30 },
		{ 269, 41, 30 },
		{ 269, 42, 30 },
		{ 269, 43, 30 },
		{ 269, 44, 30 },
		{ 269, 45, 30 },
		{ 269, 46, 30 },
		{ 269, 47, 30 },
		{ 269, 48, 30 },
		{ 269, 49, 30 },
		{ 269, 50, 30 },
		{ 269, 51, 30 },
		{ 269, 52, 30 },
		{ 269, 53, 30 },
		{ 269, 54, 30 },
		{ 269, 55, 30 },
		{ 269, 56, 30 },
		{ 269, 57, 30 },
		{ 269, 58, 30 },
		{ 269, 59, 30 },
		{ 269, 60, 30 },
		{ 269, 61, 30 },
		{ 269, 62, 30 },
		{ 269, 63, 30 },
		{ 269, 64, 30 },
		{ 269, 65, 30 },
		{ 269, 66, 30 },
		{ 269, 67, 30 },
		{ 269, 68, 30 },
		{ 269, 69, 30 },
		{ 269, 70, 30 },
		{ 269, 71, 30 },
		{ 269, 72, 30 },
		{ 269, 73, 30 },
		{ 269, 74, 30 },
		{ 269, 75, 30 },
		{ 269, 76, 30 },
		{ 269, 77, 30 },
		{ 269, 78, 30 },
		{ 269, 79, 30 },
		{ 269, 80, 30 },
		{ 269, 81, 30 },
		{ 269, 82, 30 },
		{ 269, 83, 30 },
		{ 269, 84, 30 },
		{ 269, 85, 30 },
		{ 269, 86, 30 },
		{ 269, 87, 30 },
		{ 269, 88, 30 },
		{ 269, 89, 30 },
		{ 269, 90, 30 },
		{ 269, 91, 30 },
		{ 269, 92, 30 },
		{ 269, 93, 30 },
		{ 269, 94, 30 },
		{ 269, 95, 30 },
		{ 269, 96, 30 },
		{ 269, 97, 30 },
		{ 269, 98, 30 },
		{ 269, 99, 30 },
		{ 269, 100, 30 },
		{ 269, 101, 30 },
		{ 269, 102, 30 },
		{ 269, 103, 30 },
		{ 269, 104, 30 },
		{ 269, 105, 30 },
		{ 269, 106, 30 },
		{ 269, 107, 30 },
		{ 269, 108, 30 },
		{ 269, 109, 30 },
		{ 269, 110, 30 },
		{ 269, 111, 30 },
		{ 269, 112, 30 },
		{ 269, 113, 30 },
		{ 269, 114, 30 },
		{ 269, 115, 30 },
		{ 269, 116, 30 },
		{ 269, 117, 30 },
		{ 269, 118, 30 },
		{ 269, 119, 30 },
		{ 269, 120, 30 },
		{ 269, 121, 30 },
		{ 269, 122, 30 },
		{ 269, 123, 30 },
		{ 269, 124, 30 },
		{ 269, 125, 30 },
		{ 269, 126, 30 },
		{ 269, 127, 30 },
		{ 269, 128, 30 },
		{ 269, 129, 30 },
		{ 269, 130, 30 },
		{ 269, 131, 30 },
		{ 269, 132, 30 },
		{ 269, 133, 30 },
		{ 269, 134, 30 },
		{ 269, 135, 30 },
		{ 269, 136, 30 },
		{ 269, 137, 30 },
		{ 269, 138, 30 },
		{ 269, 139, 30 },
		{ 269, 140, 30 },
		{ 269, 141, 30 },
		{ 269, 142, 30 },
		{ 269, 143, 30 },
		{ 269, 144, 30 },
		{ 269, 145, 30 },
		{ 269, 146, 30 },
		{ 269, 147, 30 },
		{ 269, 148, 30 },
		{ 269, 149, 30 },
		{ 269, 150, 30 },
		{ 269, 151, 30 },
		{ 269, 152, 30 },
		{ 269, 153, 30 },
		{ 269, 154, 30 },
		{ 269, 155, 30 },
		{ 269, 156, 30 },
		{ 269, 157, 30 },
		{ 269, 158, 30 },
		{ 269, 159, 30 },
		{ 269, 160, 30 },
		{ 269, 354, 30 },
		{ 269, 355, 30 },
		{ 269, 356, 30 },
		{ 269, 357, 30 },
		{ 269, 358, 30 },
		{ 269, 359, 30 },
		{ 269, 360, 30 },
		{ 269, 361, 30 },
		{ 269, 362, 30 },
		{ 269, 363, 30 },
		{ 269, 364, 30 },
		{ 269, 365, 30 },
		{ 269, 366, 30 },
		{ 269, 367, 30 },
		{ 269, 368, 30 },
		{ 269, 369, 30 },
		{ 269, 370, 30 },
		{ 269, 371, 30 },
		{ 269, 372, 30 },
		{ 269, 373, 30 },
		{ 269, 374, 30 },
		{ 269, 375, 30 },
		{ 269, 376, 30 },
		{ 269, 377, 30 },
		{ 269, 378, 30 },
		{ 269, 379, 30 },
		{ 269, 380, 30 },
		{ 269, 381, 30 },
		{ 269, 382, 30 },
		{ 269, 383, 30 },
		{ 269, 384, 30 },
		{ 269, 385, 30 },
		{ 269, 386, 30 },
		{ 269, 387, 30 },
		{ 269, 388, 30 },
		{ 269, 389, 30 },
		{ 269, 390, 30 },
		{ 269, 391, 30 },
		{ 269, 392, 30 },
		{ 269, 393, 30 },
		{ 269, 394, 30 },
		{ 269, 395, 30 },
		{ 269, 396, 30 },
		{ 269, 397, 30 },
		{ 269, 398, 30 },
		{ 269, 399, 30 },
		{ 269, 400, 30 },
		{ 269, 401, 30 },
		{ 269, 402, 30 },
		{ 269, 403, 30 },
		{ 269, 404, 30 },
		{ 269, 405, 30 },
		{ 269, 406, 30 },
		{ 269, 407, 30 },
		{ 269, 408, 30 },
		{ 269, 409, 30 },
		{ 269, 410, 30 },
		{ 269, 411, 30 },
		{ 269, 412, 30 },
		{ 269, 413, 30 },
		{ 269, 414, 30 },
		{ 269, 415, 30 },
		{ 269, 416, 30 },
		{ 269, 417, 30 },
		{ 269, 418, 30 },
		{ 269, 419, 30 },
		{ 269, 420, 30 },
		{ 269, 421, 30 },
		{ 269, 422, 30 },
		{ 269, 423, 30 },
		{ 269, 424, 30 },
		{ 269, 425, 30 },
		{ 269, 426, 30 },
		{ 269, 427, 30 },
		{ 269, 428, 30 },
		{ 269, 429, 30 },
		{ 269, 430, 30 },
		{ 269, 431, 30 },
		{ 269, 432, 30 },
		{ 269, 433, 30 },
		{ 269, 434, 30 },
		{ 269, 435, 30 },
		{ 269, 436, 30 },
		{ 269, 437, 30 },
		{ 269, 438, 30 },
		{ 269, 439, 30 },
		{ 269, 440, 30 },
		{ 269, 441, 30 },
		{ 269, 442, 30 },
		{ 269, 443, 30 },
		{ 269, 444, 30 },
		{ 269, 445, 30 },
		{ 269, 446, 30 },
		{ 269, 447, 30 },
		{ 269, 448, 30 },
		{ 269, 449, 30 },
		{ 269, 450, 30 },
		{ 269, 451, 30 },
		{ 269, 452, 30 },
		{ 269, 453, 30 },
		{ 269, 454, 30 },
		{ 269, 455, 30 },
		{ 269, 456, 30 },
		{ 269, 457, 30 },
		{ 269, 458, 30 },
		{ 269, 459, 30 },
		{ 269, 460, 30 },
		{ 269, 461, 30 },
		{ 269, 462, 30 },
		{ 269, 463, 30 },
		{ 269, 464, 30 },
		{ 269, 465, 30 },
		{ 269, 466, 30 },
		{ 269, 467, 30 },
		{ 269, 468, 30 },
		{ 269, 469, 30 },
		{ 269, 470, 30 },
		{ 269, 471, 30 },
		{ 269, 472, 30 },
		{ 269, 473, 30 },
		{ 269, 474, 30 },
		{ 269, 475, 30 },
		{ 269, 476, 30 },
		{ 269, 477, 30 },
		{ 269, 478, 30 },
		{ 269, 479, 30 },
		{ 269, 480, 30 },
		{ 269, 481, 30 },
		{ 269, 482, 30 },
		{ 269, 483, 30 },
		{ 269, 484, 30 },
		{ 269, 485, 30 },
		{ 269, 486, 30 },
		{ 269, 487, 30 },
		{ 269, 488, 30 },
		{ 269, 489, 30 },
		{ 269, 490, 30 },
		{ 269, 491, 30 },
		{ 269, 492, 30 },
		{ 269, 493, 30 },
		{ 269, 494, 30 },
		{ 269, 495, 30 },
		{ 269, 496, 30 },
		{ 269, 497, 30 },
		{ 269, 498, 30 },
		{ 269, 499, 30 },
		{ 269, 500, 30 },
		{ 269, 501, 30 },
		{ 269, 502, 30 },
		{ 269, 503, 30 },
		{ 269, 504, 30 },
		{ 269, 505, 30 },
		{ 269, 506, 30 },
		{ 269, 507, 30 },
		{ 269, 508, 30 },
		{ 269, 509, 30 },
		{ 269, 510, 30 },
		{ 269, 511, 30 },
		{ 269, 512, 30 },
		{ 270, 1, 30 },
		{ 270, 2, 30 },
		{ 270, 3, 30 },
		{ 270, 4, 30 },
		{ 270, 5, 30 },
		{ 270, 6, 30 },
		{ 270, 7, 30 },
		{ 270, 8, 30 },
		{ 270, 9, 30 },
		{ 270, 10, 30 },
		{ 270, 11, 30 },
		{ 270, 12, 30 },
		{ 270, 13, 30 },
		{ 270, 14, 30 },
		{ 270, 15, 30 },
		{ 270, 16, 30 },
		{ 270, 17, 30 },
		{ 270, 18, 30 },
		{ 270, 19, 30 },
		{ 270, 20, 30 },
		{ 270, 21, 30 },
		{ 270, 22, 30 },
		{ 270, 23, 30 },
		{ 270, 24, 30 },
		{ 270, 25, 30 },
		{ 270, 26, 30 },
		{ 270, 27, 30 },
		{ 270, 28, 30 },
		{ 270, 29, 30 },
		{ 270, 30, 30 },
		{ 270, 31, 30 },
		{ 270, 32, 30 },
		{ 270, 33, 30 },
		{ 270, 34, 30 },
		{ 270, 35, 30 },
		{ 270, 36, 30 },
		{ 270, 37, 30 },
		{ 270, 38, 30 },
		{ 270, 39, 30 },
		{ 270, 40, 30 },
		{ 270, 41, 30 },
		{ 270, 42, 30 },
		{ 270, 43, 30 },
		{ 270, 44, 30 },
		{ 270, 45, 30 },
		{ 270, 46, 30 },
		{ 270, 47, 30 },
		{ 270, 48, 30 },
		{ 270, 49, 30 },
		{ 270, 50, 30 },
		{ 270, 51, 30 },
		{ 270, 52, 30 },
		{ 270, 53, 30 },
		{ 270, 54, 30 },
		{ 270, 55, 30 },
		{ 270, 56, 30 },
		{ 270, 57, 30 },
		{ 270, 58, 30 },
		{ 270, 59, 30 },
		{ 270, 60, 30 },
		{ 270, 61, 30 },
		{ 270, 62, 30 },
		{ 270, 63, 30 },
		{ 270, 64, 30 },
		{ 270, 65, 30 },
		{ 270, 66, 30 },
		{ 270, 67, 30 },
		{ 270, 68, 30 },
		{ 270, 69, 30 },
		{ 270, 70, 30 },
		{ 270, 71, 30 },
		{ 270, 72, 30 },
		{ 270, 73, 30 },
		{ 270, 74, 30 },
		{ 270, 75, 30 },
		{ 270, 76, 30 },
		{ 270, 77, 30 },
		{ 270, 78, 30 },
		{ 270, 79, 30 },
		{ 270, 80, 30 },
		{ 270, 81, 30 },
		{ 270, 82, 30 },
		{ 270, 83, 30 },
		{ 270, 84, 30 },
		{ 270, 85, 30 },
		{ 270, 86, 30 },
		{ 270, 87, 30 },
		{ 270, 88, 30 },
		{ 270, 89, 30 },
		{ 270, 90, 30 },
		{ 270, 91, 30 },
		{ 270, 92, 30 },
		{ 270, 93, 30 },
		{ 270, 94, 30 },
		{ 270, 95, 30 },
		{ 270, 96, 30 },
		{ 270, 97, 30 },
		{ 270, 98, 30 },
		{ 270, 99, 30 },
		{ 270, 100, 30 },
		{ 270, 101, 30 },
		{ 270, 102, 30 },
		{ 270, 103, 30 },
		{ 270, 104, 30 },
		{ 270, 105, 30 },
		{ 270, 106, 30 },
		{ 270, 107, 30 },
		{ 270, 108, 30 },
		{ 270, 109, 30 },
		{ 270, 110, 30 },
		{ 270, 111, 30 },
		{ 270, 112, 30 },
		{ 270, 113, 30 },
		{ 270, 114, 30 },
		{ 270, 115, 30 },
		{ 270, 116, 30 },
		{ 270, 117, 30 },
		{ 270, 118, 30 },
		{ 270, 119, 30 },
		{ 270, 120, 30 },
		{ 270, 121, 30 },
		{ 270, 122, 30 },
		{ 270, 123, 30 },
		{ 270, 124, 30 },
		{ 270, 125, 30 },
		{ 270, 126, 30 },
		{ 270, 127, 30 },
		{ 270, 128, 30 },
		{ 270, 129, 30 },
		{ 270, 130, 30 },
		{ 270, 131, 30 },
		{ 270, 132, 30 },
		{ 270, 133, 30 },
		{ 270, 134, 30 },
		{ 270, 135, 30 },
		{ 270, 136, 30 },
		{ 270, 137, 30 },
		{ 270, 138, 30 },
		{ 270, 139, 30 },
		{ 270, 140, 30 },
		{ 270, 141, 30 },
		{ 270, 142, 30 },
		{ 270, 143, 30 },
		{ 270, 144, 30 },
		{ 270, 145, 30 },
		{ 270, 146, 30 },
		{ 270, 147, 30 },
		{ 270, 148, 30 },
		{ 270, 149, 30 },
		{ 270, 150, 30 },
		{ 270, 151, 30 },
		{ 270, 152, 30 },
		{ 270, 362, 30 },
		{ 270, 363, 30 },
		{ 270, 364, 30 },
		{ 270, 365, 30 },
		{ 270, 366, 30 },
		{ 270, 367, 30 },
		{ 270, 368, 30 },
		{ 270, 369, 30 },
		{ 270, 370, 30 },
		{ 270, 371, 30 },
		{ 270, 372, 30 },
		{ 270, 373, 30 },
		{ 270, 374, 30 },
		{ 270, 375, 30 },
		{ 270, 376, 30 },
		{ 270, 377, 30 },
		{ 270, 378, 30 },
		{ 270, 379, 30 },
		{ 270, 380, 30 },
		{ 270, 381, 30 },
		{ 270, 382, 30 },
		{ 270, 383, 30 },
		{ 270, 384, 30 },
		{ 270, 385, 30 },
		{ 270, 386, 30 },
		{ 270, 387, 30 },
		{ 270, 388, 30 },
		{ 270, 389, 30 },
		{ 270, 390, 30 },
		{ 270, 391, 30 },
		{ 270, 392, 30 },
		{ 270, 393, 30 },
		{ 270, 394, 30 },
		{ 270, 395, 30 },
		{ 270, 396, 30 },
		{ 270, 397, 30 },
		{ 270, 398, 30 },
		{ 270, 399, 30 },
		{ 270, 400, 30 },
		{ 270, 401, 30 },
		{ 270, 402, 30 },
		{ 270, 403, 30 },
		{ 270, 404, 30 },
		{ 270, 405, 30 },
		{ 270, 406, 30 },
		{ 270, 407, 30 },
		{ 270, 408, 30 },
		{ 270, 409, 30 },
		{ 270, 410, 30 },
		{ 270, 411, 30 },
		{ 270, 412, 30 },
		{ 270, 413, 30 },
		{ 270, 414, 30 },
		{ 270, 415, 30 },
		{ 270, 416, 30 },
		{ 270, 417, 30 },
		{ 270, 418, 30 },
		{ 270, 419, 30 },
		{ 270, 420, 30 },
		{ 270, 421, 30 },
		{ 270, 422, 30 },
		{ 270, 423, 30 },
		{ 270, 424, 30 },
		{ 270, 425, 30 },
		{ 270, 426, 30 },
		{ 270, 427, 30 },
		{ 270, 428, 30 },
		{ 270, 429, 30 },
		{ 270, 430, 30 },
		{ 270, 431, 30 },
		{ 270, 432, 30 },
		{ 270, 433, 30 },
		{ 270, 434, 30 },
		{ 270, 435, 30 },
		{ 270, 436, 30 },
		{ 270, 437, 30 },
		{ 270, 438, 30 },
		{ 270, 439, 30 },
		{ 270, 440, 30 },
		{ 270, 441, 30 },
		{ 270, 442, 30 },
		{ 270, 443, 30 },
		{ 270, 444, 30 },
		{ 270, 445, 30 },
		{ 270, 446, 30 },
		{ 270, 447, 30 },
		{ 270, 448, 30 },
		{ 270, 449, 30 },
		{ 270, 450, 30 },
		{ 270, 451, 30 },
		{ 270, 452, 30 },
		{ 270, 453, 30 },
		{ 270, 454, 30 },
		{ 270, 455, 30 },
		{ 270, 456, 30 },
		{ 270, 457, 30 },
		{ 270, 458, 30 },
		{ 270, 459, 30 },
		{ 270, 460, 30 },
		{ 270, 461, 30 },
		{ 270, 462, 30 },
		{ 270, 463, 30 },
		{ 270, 464, 30 },
		{ 270, 465, 30 },
		{ 270, 466, 30 },
		{ 270, 467, 30 },
		{ 270, 468, 30 },
		{ 270, 469, 30 },
		{ 270, 470, 30 },
		{ 270, 471, 30 },
		{ 270, 472, 30 },
		{ 270, 473, 30 },
		{ 270, 474, 30 },
		{ 270, 475, 30 },
		{ 270, 476, 30 },
		{ 270, 477, 30 },
		{ 270, 478, 30 },
		{ 270, 479, 30 },
		{ 270, 480, 30 },
		{ 270, 481, 30 },
		{ 270, 482, 30 },
		{ 270, 483, 30 },
		{ 270, 484, 30 },
		{ 270, 485, 30 },
		{ 270, 486, 30 },
		{ 270, 487, 30 },
		{ 270, 488, 30 },
		{ 270, 489, 30 },
		{ 270, 490, 30 },
		{ 270, 491, 30 },
		{ 270, 492, 30 },
		{ 270, 493, 30 },
		{ 270, 494, 30 },
		{ 270, 495, 30 },
		{ 270, 496, 30 },
		{ 270, 497, 30 },
		{ 270, 498, 30 },
		{ 270, 499, 30 },
		{ 270, 500, 30 },
		{ 270, 501, 30 },
		{ 270, 502, 30 },
		{ 270, 503, 30 },
		{ 270, 504, 30 },
		{ 270, 505, 30 },
		{ 270, 506, 30 },
		{ 270, 507, 30 },
		{ 270, 508, 30 },
		{ 270, 509, 30 },
		{ 270, 510, 30 },
		{ 270, 511, 30 },
		{ 270, 512, 30 },
		{ 271, 1, 30 },
		{ 271, 2, 30 },
		{ 271, 3, 30 },
		{ 271, 4, 30 },
		{ 271, 5, 30 },
		{ 271, 6, 30 },
		{ 271, 7, 30 },
		{ 271, 8, 30 },
		{ 271, 9, 30 },
		{ 271, 10, 30 },
		{ 271, 11, 30 },
		{ 271, 12, 30 },
		{ 271, 13, 30 },
		{ 271, 14, 30 },
		{ 271, 15, 30 },
		{ 271, 16, 30 },
		{ 271, 17, 30 },
		{ 271, 18, 30 },
		{ 271, 19, 30 },
		{ 271, 20, 30 },
		{ 271, 21, 30 },
		{ 271, 22, 30 },
		{ 271, 23, 30 },
		{ 271, 24, 30 },
		{ 271, 25, 30 },
		{ 271, 26, 30 },
		{ 271, 27, 30 },
		{ 271, 28, 30 },
		{ 271, 29, 30 },
		{ 271, 30, 30 },
		{ 271, 31, 30 },
		{ 271, 32, 30 },
		{ 271, 33, 30 },
		{ 271, 34, 30 },
		{ 271, 35, 30 },
		{ 271, 36, 30 },
		{ 271, 37, 30 },
		{ 271, 38, 30 },
		{ 271, 39, 30 },
		{ 271, 40, 30 },
		{ 271, 41, 30 },
		{ 271, 42, 30 },
		{ 271, 43, 30 },
		{ 271, 44, 30 },
		{ 271, 45, 30 },
		{ 271, 46, 30 },
		{ 271, 47, 30 },
		{ 271, 48, 30 },
		{ 271, 49, 30 },
		{ 271, 50, 30 },
		{ 271, 51, 30 },
		{ 271, 52, 30 },
		{ 271, 53, 30 },
		{ 271, 54, 30 },
		{ 271, 55, 30 },
		{ 271, 56, 30 },
		{ 271, 57, 30 },
		{ 271, 58, 30 },
		{ 271, 59, 30 },
		{ 271, 60, 30 },
		{ 271, 61, 30 },
		{ 271, 62, 30 },
		{ 271, 63, 30 },
		{ 271, 64, 30 },
		{ 271, 65, 30 },
		{ 271, 66, 30 },
		{ 271, 67, 30 },
		{ 271, 68, 30 },
		{ 271, 69, 30 },
		{ 271, 70, 30 },
		{ 271, 71, 30 },
		{ 271, 72, 30 },
		{ 271, 73, 30 },
		{ 271, 74, 30 },
		{ 271, 75, 30 },
		{ 271, 76, 30 },
		{ 271, 77, 30 },
		{ 271, 78, 30 },
		{ 271, 79, 30 },
		{ 271, 80, 30 },
		{ 271, 81, 30 },
		{ 271, 82, 30 },
		{ 271, 83, 30 },
		{ 271, 84, 30 },
		{ 271, 85, 30 },
		{ 271, 86, 30 },
		{ 271, 87, 30 },
		{ 271, 88, 30 },
		{ 271, 89, 30 },
		{ 271, 90, 30 },
		{ 271, 91, 30 },
		{ 271, 92, 30 },
		{ 271, 93, 30 },
		{ 271, 94, 30 },
		{ 271, 95, 30 },
		{ 271, 96, 30 },
		{ 271, 97, 30 },
		{ 271, 98, 30 },
		{ 271, 99, 30 },
		{ 271, 100, 30 },
		{ 271, 101, 30 },
		{ 271, 102, 30 },
		{ 271, 103, 30 },
		{ 271, 104, 30 },
		{ 271, 105, 30 },
		{ 271, 106, 30 },
		{ 271, 107, 30 },
		{ 271, 108, 30 },
		{ 271, 109, 30 },
		{ 271, 110, 30 },
		{ 271, 111, 30 },
		{ 271, 112, 30 },
		{ 271, 113, 30 },
		{ 271, 114, 30 },
		{ 271, 115, 30 },
		{ 271, 116, 30 },
		{ 271, 117, 30 },
		{ 271, 118, 30 },
		{ 271, 119, 30 },
		{ 271, 120, 30 },
		{ 271, 121, 30 },
		{ 271, 122, 30 },
		{ 271, 123, 30 },
		{ 271, 124, 30 },
		{ 271, 125, 30 },
		{ 271, 126, 30 },
		{ 271, 127, 30 },
		{ 271, 128, 30 },
		{ 271, 129, 30 },
		{ 271, 130, 30 },
		{ 271, 131, 30 },
		{ 271, 132, 30 },
		{ 271, 133, 30 },
		{ 271, 134, 30 },
		{ 271, 135, 30 },
		{ 271, 136, 30 },
		{ 271, 137, 30 },
		{ 271, 138, 30 },
		{ 271, 139, 30 },
		{ 271, 140, 30 },
		{ 271, 141, 30 },
		{ 271, 142, 30 },
		{ 271, 143, 30 },
		{ 271, 144, 30 },
		{ 271, 370, 30 },
		{ 271, 371, 30 },
		{ 271, 372, 30 },
		{ 271, 373, 30 },
		{ 271, 374, 30 },
		{ 271, 375, 30 },
		{ 271, 376, 30 },
		{ 271, 377, 30 },
		{ 271, 378, 30 },
		{ 271, 379, 30 },
		{ 271, 380, 30 },
		{ 271, 381, 30 },
		{ 271, 382, 30 },
		{ 271, 383, 30 },
		{ 271, 384, 30 },
		{ 271, 385, 30 },
		{ 271, 386, 30 },
		{ 271, 387, 30 },
		{ 271, 388, 30 },
		{ 271, 389, 30 },
		{ 271, 390, 30 },
		{ 271, 391, 30 },
		{ 271, 392, 30 },
		{ 271, 393, 30 },
		{ 271, 394, 30 },
		{ 271, 395, 30 },
		{ 271, 396, 30 },
		{ 271, 397, 30 },
		{ 271, 398, 30 },
		{ 271, 399, 30 },
		{ 271, 400, 30 },
		{ 271, 401, 30 },
		{ 271, 402, 30 },
		{ 271, 403, 30 },
		{ 271, 404, 30 },
		{ 271, 405, 30 },
		{ 271, 406, 30 },
		{ 271, 407, 30 },
		{ 271, 408, 30 },
		{ 271, 409, 30 },
		{ 271, 410, 30 },
		{ 271, 411, 30 },
		{ 271, 412, 30 },
		{ 271, 413, 30 },
		{ 271, 414, 30 },
		{ 271, 415, 30 },
		{ 271, 416, 30 },
		{ 271, 417, 30 },
		{ 271, 418, 30 },
		{ 271, 419, 30 },
		{ 271, 420, 30 },
		{ 271, 421, 30 },
		{ 271, 422, 30 },
		{ 271, 423, 30 },
		{ 271, 424, 30 },
		{ 271, 425, 30 },
		{ 271, 426, 30 },
		{ 271, 427, 30 },
		{ 271, 428, 30 },
		{ 271, 429, 30 },
		{ 271, 430, 30 },
		{ 271, 431, 30 },
		{ 271, 432, 30 },
		{ 271, 433, 30 },
		{ 271, 434, 30 },
		{ 271, 435, 30 },
		{ 271, 436, 30 },
		{ 271, 437, 30 },
		{ 271, 438, 30 },
		{ 271, 439, 30 },
		{ 271, 440, 30 },
		{ 271, 441, 30 },
		{ 271, 442, 30 },
		{ 271, 443, 30 },
		{ 271, 444, 30 },
		{ 271, 445, 30 },
		{ 271, 446, 30 },
		{ 271, 447, 30 },
		{ 271, 448, 30 },
		{ 271, 449, 30 },
		{ 271, 450, 30 },
		{ 271, 451, 30 },
		{ 271, 452, 30 },
		{ 271, 453, 30 },
		{ 271, 454, 30 },
		{ 271, 455, 30 },
		{ 271, 456, 30 },
		{ 271, 457, 30 },
		{ 271, 458, 30 },
		{ 271, 459, 30 },
		{ 271, 460, 30 },
		{ 271, 461, 30 },
		{ 271, 462, 30 },
		{ 271, 463, 30 },
		{ 271, 464, 30 },
		{ 271, 465, 30 },
		{ 271, 466, 30 },
		{ 271, 467, 30 },
		{ 271, 468, 30 },
		{ 271, 469, 30 },
		{ 271, 470, 30 },
		{ 271, 471, 30 },
		{ 271, 472, 30 },
		{ 271, 473, 30 },
		{ 271, 474, 30 },
		{ 271, 475, 30 },
		{ 271, 476, 30 },
		{ 271, 477, 30 },
		{ 271, 478, 30 },
		{ 271, 479, 30 },
		{ 271, 480, 30 },
		{ 271, 481, 30 },
		{ 271, 482, 30 },
		{ 271, 483, 30 },
		{ 271, 484, 30 },
		{ 271, 485, 30 },
		{ 271, 486, 30 },
		{ 271, 487, 30 },
		{ 271, 488, 30 },
		{ 271, 489, 30 },
		{ 271, 490, 30 },
		{ 271, 491, 30 },
		{ 271, 492, 30 },
		{ 271, 493, 30 },
		{ 271, 494, 30 },
		{ 271, 495, 30 },
		{ 271, 496, 30 },
		{ 271, 497, 30 },
		{ 271, 498, 30 },
		{ 271, 499, 30 },
		{ 271, 500, 30 },
		{ 271, 501, 30 },
		{ 271, 502, 30 },
		{ 271, 503, 30 },
		{ 271, 504, 30 },
		{ 271, 505, 30 },
		{ 271, 506, 30 },
		{ 271, 507, 30 },
		{ 271, 508, 30 },
		{ 271, 509, 30 },
		{ 271, 510, 30 },
		{ 271, 511, 30 },
		{ 271, 512, 30 },
		{ 272, 1, 30 },
		{ 272, 2, 30 },
		{ 272, 3, 30 },
		{ 272, 4, 30 },
		{ 272, 5, 30 },
		{ 272, 6, 30 },
		{ 272, 7, 30 },
		{ 272, 8, 30 },
		{ 272, 9, 30 },
		{ 272, 10, 30 },
		{ 272, 11, 30 },
		{ 272, 12, 30 },
		{ 272, 13, 30 },
		{ 272, 14, 30 },
		{ 272, 15, 30 },
		{ 272, 16, 30 },
		{ 272, 17, 30 },
		{ 272, 18, 30 },
		{ 272, 19, 30 },
		{ 272, 20, 30 },
		{ 272, 21, 30 },
		{ 272, 22, 30 },
		{ 272, 23, 30 },
		{ 272, 24, 30 },
		{ 272, 25, 30 },
		{ 272, 26, 30 },
		{ 272, 27, 30 },
		{ 272, 28, 30 },
		{ 272, 29, 30 },
		{ 272, 30, 30 },
		{ 272, 31, 30 },
		{ 272, 32, 30 },
		{ 272, 33, 30 },
		{ 272, 34, 30 },
		{ 272, 35, 30 },
		{ 272, 36, 30 },
		{ 272, 37, 30 },
		{ 272, 38, 30 },
		{ 272, 39, 30 },
		{ 272, 40, 30 },
		{ 272, 41, 30 },
		{ 272, 42, 30 },
		{ 272, 43, 30 },
		{ 272, 44, 30 },
		{ 272, 45, 30 },
		{ 272, 46, 30 },
		{ 272, 47, 30 },
		{ 272, 48, 30 },
		{ 272, 49, 30 },
		{ 272, 50, 30 },
		{ 272, 51, 30 },
		{ 272, 52, 30 },
		{ 272, 53, 30 },
		{ 272, 54, 30 },
		{ 272, 55, 30 },
		{ 272, 56, 30 },
		{ 272, 57, 30 },
		{ 272, 58, 30 },
		{ 272, 59, 30 },
		{ 272, 60, 30 },
		{ 272, 61, 30 },
		{ 272, 62, 30 },
		{ 272, 63, 30 },
		{ 272, 64, 30 },
		{ 272, 65, 30 },
		{ 272, 66, 30 },
		{ 272, 67, 30 },
		{ 272, 68, 30 },
		{ 272, 69, 30 },
		{ 272, 70, 30 },
		{ 272, 71, 30 },
		{ 272, 72, 30 },
		{ 272, 73, 30 },
		{ 272, 74, 30 },
		{ 272, 75, 30 },
		{ 272, 76, 30 },
		{ 272, 77, 30 },
		{ 272, 78, 30 },
		{ 272, 79, 30 },
		{ 272, 80, 30 },
		{ 272, 81, 30 },
		{ 272, 82, 30 },
		{ 272, 83, 30 },
		{ 272, 84, 30 },
		{ 272, 85, 30 },
		{ 272, 86, 30 },
		{ 272, 87, 30 },
		{ 272, 88, 30 },
		{ 272, 89, 30 },
		{ 272, 90, 30 },
		{ 272, 91, 30 },
		{ 272, 92, 30 },
		{ 272, 93, 30 },
		{ 272, 94, 30 },
		{ 272, 95, 30 },
		{ 272, 96, 30 },
		{ 272, 97, 30 },
		{ 272, 98, 30 },
		{ 272, 99, 30 },
		{ 272, 100, 30 },
		{ 272, 101, 30 },
		{ 272, 102, 30 },
		{ 272, 103, 30 },
		{ 272, 104, 30 },
		{ 272, 105, 30 },
		{ 272, 106, 30 },
		{ 272, 107, 30 },
		{ 272, 108, 30 },
		{ 272, 109, 30 },
		{ 272, 110, 30 },
		{ 272, 111, 30 },
		{ 272, 112, 30 },
		{ 272, 113, 30 },
		{ 272, 114, 30 },
		{ 272, 115, 30 },
		{ 272, 116, 30 },
		{ 272, 117, 30 },
		{ 272, 118, 30 },
		{ 272, 119, 30 },
		{ 272, 120, 30 },
		{ 272, 121, 30 },
		{ 272, 122, 30 },
		{ 272, 123, 30 },
		{ 272, 124, 30 },
		{ 272, 125, 30 },
		{ 272, 126, 30 },
		{ 272, 127, 30 },
		{ 272, 128, 30 },
		{ 272, 129, 30 },
		{ 272, 130, 30 },
		{ 272, 131, 30 },
		{ 272, 132, 30 },
		{ 272, 133, 30 },
		{ 272, 134, 30 },
		{ 272, 135, 30 },
		{ 272, 136, 30 },
		{ 272, 378, 30 },
		{ 272, 379, 30 },
		{ 272, 380, 30 },
		{ 272, 381, 30 },
		{ 272, 382, 30 },
		{ 272, 383, 30 },
		{ 272, 384, 30 },
		{ 272, 385, 30 },
		{ 272, 386, 30 },
		{ 272, 387, 30 },
		{ 272, 388, 30 },
		{ 272, 389, 30 },
		{ 272, 390, 30 },
		{ 272, 391, 30 },
		{ 272, 392, 30 },
		{ 272, 393, 30 },
		{ 272, 394, 30 },
		{ 272, 395, 30 },
		{ 272, 396, 30 },
		{ 272, 397, 30 },
		{ 272, 398, 30 },
		{ 272, 399, 30 },
		{ 272, 400, 30 },
		{ 272, 401, 30 },
		{ 272, 402, 30 },
		{ 272, 403, 30 },
		{ 272, 404, 30 },
		{ 272, 405, 30 },
		{ 272, 406, 30 },
		{ 272, 407, 30 },
		{ 272, 408, 30 },
		{ 272, 409, 30 },
		{ 272, 410, 30 },
		{ 272, 411, 30 },
		{ 272, 412, 30 },
		{ 272, 413, 30 },
		{ 272, 414, 30 },
		{ 272, 415, 30 },
		{ 272, 416, 30 },
		{ 272, 417, 30 },
		{ 272, 418, 30 },
		{ 272, 419, 30 },
		{ 272, 420, 30 },
		{ 272, 421, 30 },
		{ 272, 422, 30 },
		{ 272, 423, 30 },
		{ 272, 424, 30 },
		{ 272, 425, 30 },
		{ 272, 426, 30 },
		{ 272, 427, 30 },
		{ 272, 428, 30 },
		{ 272, 429, 30 },
		{ 272, 430, 30 },
		{ 272, 431, 30 },
		{ 272, 432, 30 },
		{ 272, 433, 30 },
		{ 272, 434, 30 },
		{ 272, 435, 30 },
		{ 272, 436, 30 },
		{ 272, 437, 30 },
		{ 272, 438, 30 },
		{ 272, 439, 30 },
		{ 272, 440, 30 },
		{ 272, 441, 30 },
		{ 272, 442, 30 },
		{ 272, 443, 30 },
		{ 272, 444, 30 },
		{ 272, 445, 30 },
		{ 272, 446, 30 },
		{ 272, 447, 30 },
		{ 272, 448, 30 },
		{ 272, 449, 30 },
		{ 272, 450, 30 },
		{ 272, 451, 30 },
		{ 272, 452, 30 },
		{ 272, 453, 30 },
		{ 272, 454, 30 },
		{ 272, 455, 30 },
		{ 272, 456, 30 },
		{ 272, 457, 30 },
		{ 272, 458, 30 },
		{ 272, 459, 30 },
		{ 272, 460, 30 },
		{ 272, 461, 30 },
		{ 272, 462, 30 },
		{ 272, 463, 30 },
		{ 272, 464, 30 },
		{ 272, 465, 30 },
		{ 272, 466, 30 },
		{ 272, 467, 30 },
		{ 272, 468, 30 },
		{ 272, 469, 30 },
		{ 272, 470, 30 },
		{ 272, 471, 30 },
		{ 272, 472, 30 },
		{ 272, 473, 30 },
		{ 272, 474, 30 },
		{ 272, 475, 30 },
		{ 272, 476, 30 },
		{ 272, 477, 30 },
		{ 272, 478, 30 },
		{ 272, 479, 30 },
		{ 272, 480, 30 },
		{ 272, 481, 30 },
		{ 272, 482, 30 },
		{ 272, 483, 30 },
		{ 272, 484, 30 },
		{ 272, 485, 30 },
		{ 272, 486, 30 },
		{ 272, 487, 30 },
		{ 272, 488, 30 },
		{ 272, 489, 30 },
		{ 272, 490, 30 },
		{ 272, 491, 30 },
		{ 272, 492, 30 },
		{ 272, 493, 30 },
		{ 272, 494, 30 },
		{ 272, 495, 30 },
		{ 272, 496, 30 },
		{ 272, 497, 30 },
		{ 272, 498, 30 },
		{ 272, 499, 30 },
		{ 272, 500, 30 },
		{ 272, 501, 30 },
		{ 272, 502, 30 },
		{ 272, 503, 30 },
		{ 272, 504, 30 },
		{ 272, 505, 30 },
		{ 272, 506, 30 },
		{ 272, 507, 30 },
		{ 272, 508, 30 },
		{ 272, 509, 30 },
		{ 272, 510, 30 },
		{ 272, 511, 30 },
		{ 272, 512, 30 },
		{ 273, 1, 30 },
		{ 273, 2, 30 },
		{ 273, 3, 30 },
		{ 273, 4, 30 },
		{ 273, 5, 30 },
		{ 273, 6, 30 },
		{ 273, 7, 30 },
		{ 273, 8, 30 },
		{ 273, 9, 30 },
		{ 273, 10, 30 },
		{ 273, 11, 30 },
		{ 273, 12, 30 },
		{ 273, 13, 30 },
		{ 273, 14, 30 },
		{ 273, 15, 30 },
		{ 273, 16, 30 },
		{ 273, 17, 30 },
		{ 273, 18, 30 },
		{ 273, 19, 30 },
		{ 273, 20, 30 },
		{ 273, 21, 30 },
		{ 273, 22, 30 },
		{ 273, 23, 30 },
		{ 273, 24, 30 },
		{ 273, 25, 30 },
		{ 273, 26, 30 },
		{ 273, 27, 30 },
		{ 273, 28, 30 },
		{ 273, 29, 30 },
		{ 273, 30, 30 },
		{ 273, 31, 30 },
		{ 273, 32, 30 },
		{ 273, 33, 30 },
		{ 273, 34, 30 },
		{ 273, 35, 30 },
		{ 273, 36, 30 },
		{ 273, 37, 30 },
		{ 273, 38, 30 },
		{ 273, 39, 30 },
		{ 273, 40, 30 },
		{ 273, 41, 30 },
		{ 273, 42, 30 },
		{ 273, 43, 30 },
		{ 273, 44, 30 },
		{ 273, 45, 30 },
		{ 273, 46, 30 },
		{ 273, 47, 30 },
		{ 273, 48, 30 },
		{ 273, 49, 30 },
		{ 273, 50, 30 },
		{ 273, 51, 30 },
		{ 273, 52, 30 },
		{ 273, 53, 30 },
		{ 273, 54, 30 },
		{ 273, 55, 30 },
		{ 273, 56, 30 },
		{ 273, 57, 30 },
		{ 273, 58, 30 },
		{ 273, 59, 30 },
		{ 273, 60, 30 },
		{ 273, 61, 30 },
		{ 273, 62, 30 },
		{ 273, 63, 30 },
		{ 273, 64, 30 },
		{ 273, 65, 30 },
		{ 273, 66, 30 },
		{ 273, 67, 30 },
		{ 273, 68, 30 },
		{ 273, 69, 30 },
		{ 273, 70, 30 },
		{ 273, 71, 30 },
		{ 273, 72, 30 },
		{ 273, 73, 30 },
		{ 273, 74, 30 },
		{ 273, 75, 30 },
		{ 273, 76, 30 },
		{ 273, 77, 30 },
		{ 273, 78, 30 },
		{ 273, 79, 30 },
		{ 273, 80, 30 },
		{ 273, 81, 30 },
		{ 273, 82, 30 },
		{ 273, 83, 30 },
		{ 273, 84, 30 },
		{ 273, 85, 30 },
		{ 273, 86, 30 },
		{ 273, 87, 30 },
		{ 273, 88, 30 },
		{ 273, 89, 30 },
		{ 273, 90, 30 },
		{ 273, 91, 30 },
		{ 273, 92, 30 },
		{ 273, 93, 30 },
		{ 273, 94, 30 },
		{ 273, 95, 30 },
		{ 273, 96, 30 },
		{ 273, 97, 30 },
		{ 273, 98, 30 },
		{ 273, 99, 30 },
		{ 273, 100, 30 },
		{ 273, 101, 30 },
		{ 273, 102, 30 },
		{ 273, 103, 30 },
		{ 273, 104, 30 },
		{ 273, 105, 30 },
		{ 273, 106, 30 },
		{ 273, 107, 30 },
		{ 273, 108, 30 },
		{ 273, 109, 30 },
		{ 273, 110, 30 },
		{ 273, 111, 30 },
		{ 273, 112, 30 },
		{ 273, 113, 30 },
		{ 273, 114, 30 },
		{ 273, 115, 30 },
		{ 273, 116, 30 },
		{ 273, 117, 30 },
		{ 273, 118, 30 },
		{ 273, 119, 30 },
		{ 273, 120, 30 },
		{ 273, 121, 30 },
		{ 273, 122, 30 },
		{ 273, 123, 30 },
		{ 273, 124, 30 },
		{ 273, 125, 30 },
		{ 273, 126, 30 },
		{ 273, 127, 30 },
		{ 273, 128, 30 },
		{ 273, 386, 30 },
		{ 273, 387, 30 },
		{ 273, 388, 30 },
		{ 273, 389, 30 },
		{ 273, 390, 30 },
		{ 273, 391, 30 },
		{ 273, 392, 30 },
		{ 273, 393, 30 },
		{ 273, 394, 30 },
		{ 273, 395, 30 },
		{ 273, 396, 30 },
		{ 273, 397, 30 },
		{ 273, 398, 30 },
		{ 273, 399, 30 },
		{ 273, 400, 30 },
		{ 273, 401, 30 },
		{ 273, 402, 30 },
		{ 273, 403, 30 },
		{ 273, 404, 30 },
		{ 273, 405, 30 },
		{ 273, 406, 30 },
		{ 273, 407, 30 },
		{ 273, 408, 30 },
		{ 273, 409, 30 },
		{ 273, 410, 30 },
		{ 273, 411, 30 },
		{ 273, 412, 30 },
		{ 273, 413, 30 },
		{ 273, 414, 30 },
		{ 273, 415, 30 },
		{ 273, 416, 30 },
		{ 273, 417, 30 },
		{ 273, 418, 30 },
		{ 273, 419, 30 },
		{ 273, 420, 30 },
		{ 273, 421, 30 },
		{ 273, 422, 30 },
		{ 273, 423, 30 },
		{ 273, 424, 30 },
		{ 273, 425, 30 },
		{ 273, 426, 30 },
		{ 273, 427, 30 },
		{ 273, 428, 30 },
		{ 273, 429, 30 },
		{ 273, 430, 30 },
		{ 273, 431, 30 },
		{ 273, 432, 30 },
		{ 273, 433, 30 },
		{ 273, 434, 30 },
		{ 273, 435, 30 },
		{ 273, 436, 30 },
		{ 273, 437, 30 },
		{ 273, 438, 30 },
		{ 273, 439, 30 },
		{ 273, 440, 30 },
		{ 273, 441, 30 },
		{ 273, 442, 30 },
		{ 273, 443, 30 },
		{ 273, 444, 30 },
		{ 273, 445, 30 },
		{ 273, 446, 30 },
		{ 273, 447, 30 },
		{ 273, 448, 30 },
		{ 273, 449, 30 },
		{ 273, 450, 30 },
		{ 273, 451, 30 },
		{ 273, 452, 30 },
		{ 273, 453, 30 },
		{ 273, 454, 30 },
		{ 273, 455, 30 },
		{ 273, 456, 30 },
		{ 273, 457, 30 },
		{ 273, 458, 30 },
		{ 273, 459, 30 },
		{ 273, 460, 30 },
		{ 273, 461, 30 },
		{ 273, 462, 30 },
		{ 273, 463, 30 },
		{ 273, 464, 30 },
		{ 273, 465, 30 },
		{ 273, 466, 30 },
		{ 273, 467, 30 },
		{ 273, 468, 30 },
		{ 273, 469, 30 },
		{ 273, 470, 30 },
		{ 273, 471, 30 },
		{ 273, 472, 30 },
		{ 273, 473, 30 },
		{ 273, 474, 30 },
		{ 273, 475, 30 },
		{ 273, 476, 30 },
		{ 273, 477, 30 },
		{ 273, 478, 30 },
		{ 273, 479, 30 },
		{ 273, 480, 30 },
		{ 273, 481, 30 },
		{ 273, 482, 30 },
		{ 273, 483, 30 },
		{ 273, 484, 30 },
		{ 273, 485, 30 },
		{ 273, 486, 30 },
		{ 273, 487, 30 },
		{ 273, 488, 30 },
		{ 273, 489, 30 },
		{ 273, 490, 30 },
		{ 273, 491, 30 },
		{ 273, 492, 30 },
		{ 273, 493, 30 },
		{ 273, 494, 30 },
		{ 273, 495, 30 },
		{ 273, 496, 30 },
		{ 273, 497, 30 },
		{ 273, 498, 30 },
		{ 273, 499, 30 },
		{ 273, 500, 30 },
		{ 273, 501, 30 },
		{ 273, 502, 30 },
		{ 273, 503, 30 },
		{ 273, 504, 30 },
		{ 273, 505, 30 },
		{ 273, 506, 30 },
		{ 273, 507, 30 },
		{ 273, 508, 30 },
		{ 273, 509, 30 },
		{ 273, 510, 30 },
		{ 273, 511, 30 },
		{ 273, 512, 30 },
		{ 274, 1, 30 },
		{ 274, 2, 30 },
		{ 274, 3, 30 },
		{ 274, 4, 30 },
		{ 274, 5, 30 },
		{ 274, 6, 30 },
		{ 274, 7, 30 },
		{ 274, 8, 30 },
		{ 274, 9, 30 },
		{ 274, 10, 30 },
		{ 274, 11, 30 },
		{ 274, 12, 30 },
		{ 274, 13, 30 },
		{ 274, 14, 30 },
		{ 274, 15, 30 },
		{ 274, 16, 30 },
		{ 274, 17, 30 },
		{ 274, 18, 30 },
		{ 274, 19, 30 },
		{ 274, 20, 30 },
		{ 274, 21, 30 },
		{ 274, 22, 30 },
		{ 274, 23, 30 },
		{ 274, 24, 30 },
		{ 274, 25, 30 },
		{ 274, 26, 30 },
		{ 274, 27, 30 },
		{ 274, 28, 30 },
		{ 274, 29, 30 },
		{ 274, 30, 30 },
		{ 274, 31, 30 },
		{ 274, 32, 30 },
		{ 274, 33, 30 },
		{ 274, 34, 30 },
		{ 274, 35, 30 },
		{ 274, 36, 30 },
		{ 274, 37, 30 },
		{ 274, 38, 30 },
		{ 274, 39, 30 },
		{ 274, 40, 30 },
		{ 274, 41, 30 },
		{ 274, 42, 30 },
		{ 274, 43, 30 },
		{ 274, 44, 30 },
		{ 274, 45, 30 },
		{ 274, 46, 30 },
		{ 274, 47, 30 },
		{ 274, 48, 30 },
		{ 274, 49, 30 },
		{ 274, 50, 30 },
		{ 274, 51, 30 },
		{ 274, 52, 30 },
		{ 274, 53, 30 },
		{ 274, 54, 30 },
		{ 274, 55, 30 },
		{ 274, 56, 30 },
		{ 274, 57, 30 },
		{ 274, 58, 30 },
		{ 274, 59, 30 },
		{ 274, 60, 30 },
		{ 274, 61, 30 },
		{ 274, 62, 30 },
		{ 274, 63, 30 },
		{ 274, 64, 30 },
		{ 274, 65, 30 },
		{ 274, 66, 30 },
		{ 274, 67, 30 },
		{ 274, 68, 30 },
		{ 274, 69, 30 },
		{ 274, 70, 30 },
		{ 274, 71, 30 },
		{ 274, 72, 30 },
		{ 274, 73, 30 },
		{ 274, 74, 30 },
		{ 274, 75, 30 },
		{ 274, 76, 30 },
		{ 274, 77, 30 },
		{ 274, 78, 30 },
		{ 274, 79, 30 },
		{ 274, 80, 30 },
		{ 274, 81, 30 },
		{ 274, 82, 30 },
		{ 274, 83, 30 },
		{ 274, 84, 30 },
		{ 274, 85, 30 },
		{ 274, 86, 30 },
		{ 274, 87, 30 },
		{ 274, 88, 30 },
		{ 274, 89, 30 },
		{ 274, 90, 30 },
		{ 274, 91, 30 },
		{ 274, 92, 30 },
		{ 274, 93, 30 },
		{ 274, 94, 30 },
		{ 274, 95, 30 },
		{ 274, 96, 30 },
		{ 274, 97, 30 },
		{ 274, 98, 30 },
		{ 274, 99, 30 },
		{ 274, 100, 30 },
		{ 274, 101, 30 },
		{ 274, 102, 30 },
		{ 274, 103, 30 },
		{ 274, 104, 30 },
		{ 274, 105, 30 },
		{ 274, 106, 30 },
		{ 274, 107, 30 },
		{ 274, 108, 30 },
		{ 274, 109, 30 },
		{ 274, 110, 30 },
		{ 274, 111, 30 },
		{ 274, 112, 30 },
		{ 274, 113, 30 },
		{ 274, 114, 30 },
		{ 274, 115, 30 },
		{ 274, 116, 30 },
		{ 274, 117, 30 },
		{ 274, 118, 30 },
		{ 274, 119, 30 },
		{ 274, 120, 30 },
		{ 274, 394, 30 },
		{ 274, 395, 30 },
		{ 274, 396, 30 },
		{ 274, 397, 30 },
		{ 274, 398, 30 },
		{ 274, 399, 30 },
		{ 274, 400, 30 },
		{ 274, 401, 30 },
		{ 274, 402, 30 },
		{ 274, 403, 30 },
		{ 274, 404, 30 },
		{ 274, 405, 30 },
		{ 274, 406, 30 },
		{ 274, 407, 30 },
		{ 274, 408, 30 },
		{ 274, 409, 30 },
		{ 274, 410, 30 },
		{ 274, 411, 30 },
		{ 274, 412, 30 },
		{ 274, 413, 30 },
		{ 274, 414, 30 },
		{ 274, 415, 30 },
		{ 274, 416, 30 },
		{ 274, 417, 30 },
		{ 274, 418, 30 },
		{ 274, 419, 30 },
		{ 274, 420, 30 },
		{ 274, 421, 30 },
		{ 274, 422, 30 },
		{ 274, 423, 30 },
		{ 274, 424, 30 },
		{ 274, 425, 30 },
		{ 274, 426, 30 },
		{ 274, 427, 30 },
		{ 274, 428, 30 },
		{ 274, 429, 30 },
		{ 274, 430, 30 },
		{ 274, 431, 30 },
		{ 274, 432, 30 },
		{ 274, 433, 30 },
		{ 274, 434, 30 },
		{ 274, 435, 30 },
		{ 274, 436, 30 },
		{ 274, 437, 30 },
		{ 274, 438, 30 },
		{ 274, 439, 30 },
		{ 274, 440, 30 },
		{ 274, 441, 30 },
		{ 274, 442, 30 },
		{ 274, 443, 30 },
		{ 274, 444, 30 },
		{ 274, 445, 30 },
		{ 274, 446, 30 },
		{ 274, 447, 30 },
		{ 274, 448, 30 },
		{ 274, 449, 30 },
		{ 274, 450, 30 },
		{ 274, 451, 30 },
		{ 274, 452, 30 },
		{ 274, 453, 30 },
		{ 274, 454, 30 },
		{ 274, 455, 30 },
		{ 274, 456, 30 },
		{ 274, 457, 30 },
		{ 274, 458, 30 },
		{ 274, 459, 30 },
		{ 274, 460, 30 },
		{ 274, 461, 30 },
		{ 274, 462, 30 },
		{ 274, 463, 30 },
		{ 274, 464, 30 },
		{ 274, 465, 30 },
		{ 274, 466, 30 },
		{ 274, 467, 30 },
		{ 274, 468, 30 },
		{ 274, 469, 30 },
		{ 274, 470, 30 },
		{ 274, 471, 30 },
		{ 274, 472, 30 },
		{ 274, 473, 30 },
		{ 274, 474, 30 },
		{ 274, 475, 30 },
		{ 274, 476, 30 },
		{ 274, 477, 30 },
		{ 274, 478, 30 },
		{ 274, 479, 30 },
		{ 274, 480, 30 },
		{ 274, 481, 30 },
		{ 274, 482, 30 },
		{ 274, 483, 30 },
		{ 274, 484, 30 },
		{ 274, 485, 30 },
		{ 274, 486, 30 },
		{ 274, 487, 30 },
		{ 274, 488, 30 },
		{ 274, 489, 30 },
		{ 274, 490, 30 },
		{ 274, 491, 30 },
		{ 274, 492, 30 },
		{ 274, 493, 30 },
		{ 274, 494, 30 },
		{ 274, 495, 30 },
		{ 274, 496, 30 },
		{ 274, 497, 30 },
		{ 274, 498, 30 },
		{ 274, 499, 30 },
		{ 274, 500, 30 },
		{ 274, 501, 30 },
		{ 274, 502, 30 },
		{ 274, 503, 30 },
		{ 274, 504, 30 },
		{ 274, 505, 30 },
		{ 274, 506, 30 },
		{ 274, 507, 30 },
		{ 274, 508, 30 },
		{ 274, 509, 30 },
		{ 274, 510, 30 },
		{ 274, 511, 30 },
		{ 274, 512, 30 },
		{ 275, 1, 30 },
		{ 275, 2, 30 },
		{ 275, 3, 30 },
		{ 275, 4, 30 },
		{ 275, 5, 30 },
		{ 275, 6, 30 },
		{ 275, 7, 30 },
		{ 275, 8, 30 },
		{ 275, 9, 30 },
		{ 275, 10, 30 },
		{ 275, 11, 30 },
		{ 275, 12, 30 },
		{ 275, 13, 30 },
		{ 275, 14, 30 },
		{ 275, 15, 30 },
		{ 275, 16, 30 },
		{ 275, 17, 30 },
		{ 275, 18, 30 },
		{ 275, 19, 30 },
		{ 275, 20, 30 },
		{ 275, 21, 30 },
		{ 275, 22, 30 },
		{ 275, 23, 30 },
		{ 275, 24, 30 },
		{ 275, 25, 30 },
		{ 275, 26, 30 },
		{ 275, 27, 30 },
		{ 275, 28, 30 },
		{ 275, 29, 30 },
		{ 275, 30, 30 },
		{ 275, 31, 30 },
		{ 275, 32, 30 },
		{ 275, 33, 30 },
		{ 275, 34, 30 },
		{ 275, 35, 30 },
		{ 275, 36, 30 },
		{ 275, 37, 30 },
		{ 275, 38, 30 },
		{ 275, 39, 30 },
		{ 275, 40, 30 },
		{ 275, 41, 30 },
		{ 275, 42, 30 },
		{ 275, 43, 30 },
		{ 275, 44, 30 },
		{ 275, 45, 30 },
		{ 275, 46, 30 },
		{ 275, 47, 30 },
		{ 275, 48, 30 },
		{ 275, 49, 30 },
		{ 275, 50, 30 },
		{ 275, 51, 30 },
		{ 275, 52, 30 },
		{ 275, 53, 30 },
		{ 275, 54, 30 },
		{ 275, 55, 30 },
		{ 275, 56, 30 },
		{ 275, 57, 30 },
		{ 275, 58, 30 },
		{ 275, 59, 30 },
		{ 275, 60, 30 },
		{ 275, 61, 30 },
		{ 275, 62, 30 },
		{ 275, 63, 30 },
		{ 275, 64, 30 },
		{ 275, 65, 30 },
		{ 275, 66, 30 },
		{ 275, 67, 30 },
		{ 275, 68, 30 },
		{ 275, 69, 30 },
		{ 275, 70, 30 },
		{ 275, 71, 30 },
		{ 275, 72, 30 },
		{ 275, 73, 30 },
		{ 275, 74, 30 },
		{ 275, 75, 30 },
		{ 275, 76, 30 },
		{ 275, 77, 30 },
		{ 275, 78, 30 },
		{ 275, 79, 30 },
		{ 275, 80, 30 },
		{ 275, 81, 30 },
		{ 275, 82, 30 },
		{ 275, 83, 30 },
		{ 275, 84, 30 },
		{ 275, 85, 30 },
		{ 275, 86, 30 },
		{ 275, 87, 30 },
		{ 275, 88, 30 },
		{ 275, 89, 30 },
		{ 275, 90, 30 },
		{ 275, 91, 30 },
		{ 275, 92, 30 },
		{ 275, 93, 30 },
		{ 275, 94, 30 },
		{ 275, 95, 30 },
		{ 275, 96, 30 },
		{ 275, 97, 30 },
		{ 275, 98, 30 },
		{ 275, 99, 30 },
		{ 275, 100, 30 },
		{ 275, 101, 30 },
		{ 275, 102, 30 },
		{ 275, 103, 30 },
		{ 275, 104, 30 },
		{ 275, 105, 30 },
		{ 275, 106, 30 },
		{ 275, 107, 30 },
		{ 275, 108, 30 },
		{ 275, 109, 30 },
		{ 275, 110, 30 },
		{ 275, 111, 30 },
		{ 275, 112, 30 },
		{ 275, 402, 30 },
		{ 275, 403, 30 },
		{ 275, 404, 30 },
		{ 275, 405, 30 },
		{ 275, 406, 30 },
		{ 275, 407, 30 },
		{ 275, 408, 30 },
		{ 275, 409, 30 },
		{ 275, 410, 30 },
		{ 275, 411, 30 },
		{ 275, 412, 30 },
		{ 275, 413, 30 },
		{ 275, 414, 30 },
		{ 275, 415, 30 },
		{ 275, 416, 30 },
		{ 275, 417, 30 },
		{ 275, 418, 30 },
		{ 275, 419, 30 },
		{ 275, 420, 30 },
		{ 275, 421, 30 },
		{ 275, 422, 30 },
		{ 275, 423, 30 },
		{ 275, 424, 30 },
		{ 275, 425, 30 },
		{ 275, 426, 30 },
		{ 275, 427, 30 },
		{ 275, 428, 30 },
		{ 275, 429, 30 },
		{ 275, 430, 30 },
		{ 275, 431, 30 },
		{ 275, 432, 30 },
		{ 275, 433, 30 },
		{ 275, 434, 30 },
		{ 275, 435, 30 },
		{ 275, 436, 30 },
		{ 275, 437, 30 },
		{ 275, 438, 30 },
		{ 275, 439, 30 },
		{ 275, 440, 30 },
		{ 275, 441, 30 },
		{ 275, 442, 30 },
		{ 275, 443, 30 },
		{ 275, 444, 30 },
		{ 275, 445, 30 },
		{ 275, 446, 30 },
		{ 275, 447, 30 },
		{ 275, 448, 30 },
		{ 275, 449, 30 },
		{ 275, 450, 30 },
		{ 275, 451, 30 },
		{ 275, 452, 30 },
		{ 275, 453, 30 },
		{ 275, 454, 30 },
		{ 275, 455, 30 },
		{ 275, 456, 30 },
		{ 275, 457, 30 },
		{ 275, 458, 30 },
		{ 275, 459, 30 },
		{ 275, 460, 30 },
		{ 275, 461, 30 },
		{ 275, 462, 30 },
		{ 275, 463, 30 },
		{ 275, 464, 30 },
		{ 275, 465, 30 },
		{ 275, 466, 30 },
		{ 275, 467, 30 },
		{ 275, 468, 30 },
		{ 275, 469, 30 },
		{ 275, 470, 30 },
		{ 275, 471, 30 },
		{ 275, 472, 30 },
		{ 275, 473, 30 },
		{ 275, 474, 30 },
		{ 275, 475, 30 },
		{ 275, 476, 30 },
		{ 275, 477, 30 },
		{ 275, 478, 30 },
		{ 275, 479, 30 },
		{ 275, 480, 30 },
		{ 275, 481, 30 },
		{ 275, 482, 30 },
		{ 275, 483, 30 },
		{ 275, 484, 30 },
		{ 275, 485, 30 },
		{ 275, 486, 30 },
		{ 275, 487, 30 },
		{ 275, 488, 30 },
		{ 275, 489, 30 },
		{ 275, 490, 30 },
		{ 275, 491, 30 },
		{ 275, 492, 30 },
		{ 275, 493, 30 },
		{ 275, 494, 30 },
		{ 275, 495, 30 },
		{ 275, 496, 30 },
		{ 275, 497, 30 },
		{ 275, 498, 30 },
		{ 275, 499, 30 },
		{ 275, 500, 30 },
		{ 275, 501, 30 },
		{ 275, 502, 30 },
		{ 275, 503, 30 },
		{ 275, 504, 30 },
		{ 275, 505, 30 },
		{ 275, 506, 30 },
		{ 275, 507, 30 },
		{ 275, 508, 30 },
		{ 275, 509, 30 },
		{ 275, 510, 30 },
		{ 275, 511, 30 },
		{ 275, 512, 30 },
		{ 276, 1, 30 },
		{ 276, 2, 30 },
		{ 276, 3, 30 },
		{ 276, 4, 30 },
		{ 276, 5, 30 },
		{ 276, 6, 30 },
		{ 276, 7, 30 },
		{ 276, 8, 30 },
		{ 276, 9, 30 },
		{ 276, 10, 30 },
		{ 276, 11, 30 },
		{ 276, 12, 30 },
		{ 276, 13, 30 },
		{ 276, 14, 30 },
		{ 276, 15, 30 },
		{ 276, 16, 30 },
		{ 276, 17, 30 },
		{ 276, 18, 30 },
		{ 276, 19, 30 },
		{ 276, 20, 30 },
		{ 276, 21, 30 },
		{ 276, 22, 30 },
		{ 276, 23, 30 },
		{ 276, 24, 30 },
		{ 276, 25, 30 },
		{ 276, 26, 30 },
		{ 276, 27, 30 },
		{ 276, 28, 30 },
		{ 276, 29, 30 },
		{ 276, 30, 30 },
		{ 276, 31, 30 },
		{ 276, 32, 30 },
		{ 276, 33, 30 },
		{ 276, 34, 30 },
		{ 276, 35, 30 },
		{ 276, 36, 30 },
		{ 276, 37, 30 },
		{ 276, 38, 30 },
		{ 276, 39, 30 },
		{ 276, 40, 30 },
		{ 276, 41, 30 },
		{ 276, 42, 30 },
		{ 276, 43, 30 },
		{ 276, 44, 30 },
		{ 276, 45, 30 },
		{ 276, 46, 30 },
		{ 276, 47, 30 },
		{ 276, 48, 30 },
		{ 276, 49, 30 },
		{ 276, 50, 30 },
		{ 276, 51, 30 },
		{ 276, 52, 30 },
		{ 276, 53, 30 },
		{ 276, 54, 30 },
		{ 276, 55, 30 },
		{ 276, 56, 30 },
		{ 276, 57, 30 },
		{ 276, 58, 30 },
		{ 276, 59, 30 },
		{ 276, 60, 30 },
		{ 276, 61, 30 },
		{ 276, 62, 30 },
		{ 276, 63, 30 },
		{ 276, 64, 30 },
		{ 276, 65, 30 },
		{ 276, 66, 30 },
		{ 276, 67, 30 },
		{ 276, 68, 30 },
		{ 276, 69, 30 },
		{ 276, 70, 30 },
		{ 276, 71, 30 },
		{ 276, 72, 30 },
		{ 276, 73, 30 },
		{ 276, 74, 30 },
		{ 276, 75, 30 },
		{ 276, 76, 30 },
		{ 276, 77, 30 },
		{ 276, 78, 30 },
		{ 276, 79, 30 },
		{ 276, 80, 30 },
		{ 276, 81, 30 },
		{ 276, 82, 30 },
		{ 276, 83, 30 },
		{ 276, 84, 30 },
		{ 276, 85, 30 },
		{ 276, 86, 30 },
		{ 276, 87, 30 },
		{ 276, 88, 30 },
		{ 276, 89, 30 },
		{ 276, 90, 30 },
		{ 276, 91, 30 },
		{ 276, 92, 30 },
		{ 276, 93, 30 },
		{ 276, 94, 30 },
		{ 276, 95, 30 },
		{ 276, 96, 30 },
		{ 276, 97, 30 },
		{ 276, 98, 30 },
		{ 276, 99, 30 },
		{ 276, 100, 30 },
		{ 276, 101, 30 },
		{ 276, 102, 30 },
		{ 276, 103, 30 },
		{ 276, 104, 30 },
		{ 276, 410, 30 },
		{ 276, 411, 30 },
		{ 276, 412, 30 },
		{ 276, 413, 30 },
		{ 276, 414, 30 },
		{ 276, 415, 30 },
		{ 276, 416, 30 },
		{ 276, 417, 30 },
		{ 276, 418, 30 },
		{ 276, 419, 30 },
		{ 276, 420, 30 },
		{ 276, 421, 30 },
		{ 276, 422, 30 },
		{ 276, 423, 30 },
		{ 276, 424, 30 },
		{ 276, 425, 30 },
		{ 276, 426, 30 },
		{ 276, 427, 30 },
		{ 276, 428, 30 },
		{ 276, 429, 30 },
		{ 276, 430, 30 },
		{ 276, 431, 30 },
		{ 276, 432, 30 },
		{ 276, 433, 30 },
		{ 276, 434, 30 },
		{ 276, 435, 30 },
		{ 276, 436, 30 },
		{ 276, 437, 30 },
		{ 276, 438, 30 },
		{ 276, 439, 30 },
		{ 276, 440, 30 },
		{ 276, 441, 30 },
		{ 276, 442, 30 },
		{ 276, 443, 30 },
		{ 276, 444, 30 },
		{ 276, 445, 30 },
		{ 276, 446, 30 },
		{ 276, 447, 30 },
		{ 276, 448, 30 },
		{ 276, 449, 30 },
		{ 276, 450, 30 },
		{ 276, 451, 30 },
		{ 276, 452, 30 },
		{ 276, 453, 30 },
		{ 276, 454, 30 },
		{ 276, 455, 30 },
		{ 276, 456, 30 },
		{ 276, 457, 30 },
		{ 276, 458, 30 },
		{ 276, 459, 30 },
		{ 276, 460, 30 },
		{ 276, 461, 30 },
		{ 276, 462, 30 },
		{ 276, 463, 30 },
		{ 276, 464, 30 },
		{ 276, 465, 30 },
		{ 276, 466, 30 },
		{ 276, 467, 30 },
		{ 276, 468, 30 },
		{ 276, 469, 30 },
		{ 276, 470, 30 },
		{ 276, 471, 30 },
		{ 276, 472, 30 },
		{ 276, 473, 30 },
		{ 276, 474, 30 },
		{ 276, 475, 30 },
		{ 276, 476, 30 },
		{ 276, 477, 30 },
		{ 276, 478, 30 },
		{ 276, 479, 30 },
		{ 276, 480, 30 },
		{ 276, 481, 30 },
		{ 276, 482, 30 },
		{ 276, 483, 30 },
		{ 276, 484, 30 },
		{ 276, 485, 30 },
		{ 276, 486, 30 },
		{ 276, 487, 30 },
		{ 276, 488, 30 },
		{ 276, 489, 30 },
		{ 276, 490, 30 },
		{ 276, 491, 30 },
		{ 276, 492, 30 },
		{ 276, 493, 30 },
		{ 276, 494, 30 },
		{ 276, 495, 30 },
		{ 276, 496, 30 },
		{ 276, 497, 30 },
		{ 276, 498, 30 },
		{ 276, 499, 30 },
		{ 276, 500, 30 },
		{ 276, 501, 30 },
		{ 276, 502, 30 },
		{ 276, 503, 30 },
		{ 276, 504, 30 },
		{ 276, 505, 30 },
		{ 276, 506, 30 },
		{ 276, 507, 30 },
		{ 276, 508, 30 },
		{ 276, 509, 30 },
		{ 276, 510, 30 },
		{ 276, 511, 30 },
		{ 276, 512, 30 },
		{ 277, 1, 30 },
		{ 277, 2, 30 },
		{ 277, 3, 30 },
		{ 277, 4, 30 },
		{ 277, 5, 30 },
		{ 277, 6, 30 },
		{ 277, 7, 30 },
		{ 277, 8, 30 },
		{ 277, 9, 30 },
		{ 277, 10, 30 },
		{ 277, 11, 30 },
		{ 277, 12, 30 },
		{ 277, 13, 30 },
		{ 277, 14, 30 },
		{ 277, 15, 30 },
		{ 277, 16, 30 },
		{ 277, 17, 30 },
		{ 277, 18, 30 },
		{ 277, 19, 30 },
		{ 277, 20, 30 },
		{ 277, 21, 30 },
		{ 277, 22, 30 },
		{ 277, 23, 30 },
		{ 277, 24, 30 },
		{ 277, 25, 30 },
		{ 277, 26, 30 },
		{ 277, 27, 30 },
		{ 277, 28, 30 },
		{ 277, 29, 30 },
		{ 277, 30, 30 },
		{ 277, 31, 30 },
		{ 277, 32, 30 },
		{ 277, 33, 30 },
		{ 277, 34, 30 },
		{ 277, 35, 30 },
		{ 277, 36, 30 },
		{ 277, 37, 30 },
		{ 277, 38, 30 },
		{ 277, 39, 30 },
		{ 277, 40, 30 },
		{ 277, 41, 30 },
		{ 277, 42, 30 },
		{ 277, 43, 30 },
		{ 277, 44, 30 },
		{ 277, 45, 30 },
		{ 277, 46, 30 },
		{ 277, 47, 30 },
		{ 277, 48, 30 },
		{ 277, 49, 30 },
		{ 277, 50, 30 },
		{ 277, 51, 30 },
		{ 277, 52, 30 },
		{ 277, 53, 30 },
		{ 277, 54, 30 },
		{ 277, 55, 30 },
		{ 277, 56, 30 },
		{ 277, 57, 30 },
		{ 277, 58, 30 },
		{ 277, 59, 30 },
		{ 277, 60, 30 },
		{ 277, 61, 30 },
		{ 277, 62, 30 },
		{ 277, 63, 30 },
		{ 277, 64, 30 },
		{ 277, 65, 30 },
		{ 277, 66, 30 },
		{ 277, 67, 30 },
		{ 277, 68, 30 },
		{ 277, 69, 30 },
		{ 277, 70, 30 },
		{ 277, 71, 30 },
		{ 277, 72, 30 },
		{ 277, 73, 30 },
		{ 277, 74, 30 },
		{ 277, 75, 30 },
		{ 277, 76, 30 },
		{ 277, 77, 30 },
		{ 277, 78, 30 },
		{ 277, 79, 30 },
		{ 277, 80, 30 },
		{ 277, 81, 30 },
		{ 277, 82, 30 },
		{ 277, 83, 30 },
		{ 277, 84, 30 },
		{ 277, 85, 30 },
		{ 277, 86, 30 },
		{ 277, 87, 30 },
		{ 277, 88, 30 },
		{ 277, 89, 30 },
		{ 277, 90, 30 },
		{ 277, 91, 30 },
		{ 277, 92, 30 },
		{ 277, 93, 30 },
		{ 277, 94, 30 },
		{ 277, 95, 30 },
		{ 277, 96, 30 },
		{ 277, 418, 30 },
		{ 277, 419, 30 },
		{ 277, 420, 30 },
		{ 277, 421, 30 },
		{ 277, 422, 30 },
		{ 277, 423, 30 },
		{ 277, 424, 30 },
		{ 277, 425, 30 },
		{ 277, 426, 30 },
		{ 277, 427, 30 },
		{ 277, 428, 30 },
		{ 277, 429, 30 },
		{ 277, 430, 30 },
		{ 277, 431, 30 },
		{ 277, 432, 30 },
		{ 277, 433, 30 },
		{ 277, 434, 30 },
		{ 277, 435, 30 },
		{ 277, 436, 30 },
		{ 277, 437, 30 },
		{ 277, 438, 30 },
		{ 277, 439, 30 },
		{ 277, 440, 30 },
		{ 277, 441, 30 },
		{ 277, 442, 30 },
		{ 277, 443, 30 },
		{ 277, 444, 30 },
		{ 277, 445, 30 },
		{ 277, 446, 30 },
		{ 277, 447, 30 },
		{ 277, 448, 30 },
		{ 277, 449, 30 },
		{ 277, 450, 30 },
		{ 277, 451, 30 },
		{ 277, 452, 30 },
		{ 277, 453, 30 },
		{ 277, 454, 30 },
		{ 277, 455, 30 },
		{ 277, 456, 30 },
		{ 277, 457, 30 },
		{ 277, 458, 30 },
		{ 277, 459, 30 },
		{ 277, 460, 30 },
		{ 277, 461, 30 },
		{ 277, 462, 30 },
		{ 277, 463, 30 },
		{ 277, 464, 30 },
		{ 277, 465, 30 },
		{ 277, 466, 30 },
		{ 277, 467, 30 },
		{ 277, 468, 30 },
		{ 277, 469, 30 },
		{ 277, 470, 30 },
		{ 277, 471, 30 },
		{ 277, 472, 30 },
		{ 277, 473, 30 },
		{ 277, 474, 30 },
		{ 277, 475, 30 },
		{ 277, 476, 30 },
		{ 277, 477, 30 },
		{ 277, 478, 30 },
		{ 277, 479, 30 },
		{ 277, 480, 30 },
		{ 277, 481, 30 },
		{ 277, 482, 30 },
		{ 277, 483, 30 },
		{ 277, 484, 30 },
		{ 277, 485, 30 },
		{ 277, 486, 30 },
		{ 277, 487, 30 },
		{ 277, 488, 30 },
		{ 277, 489, 30 },
		{ 277, 490, 30 },
		{ 277, 491, 30 },
		{ 277, 492, 30 },
		{ 277, 493, 30 },
		{ 277, 494, 30 },
		{ 277, 495, 30 },
		{ 277, 496, 30 },
		{ 277, 497, 30 },
		{ 277, 498, 30 },
		{ 277, 499, 30 },
		{ 277, 500, 30 },
		{ 277, 501, 30 },
		{ 277, 502, 30 },
		{ 277, 503, 30 },
		{ 277, 504, 30 },
		{ 277, 505, 30 },
		{ 277, 506, 30 },
		{ 277, 507, 30 },
		{ 277, 508, 30 },
		{ 277, 509, 30 },
		{ 277, 510, 30 },
		{ 277, 511, 30 },
		{ 277, 512, 30 },
		{ 278, 1, 30 },
		{ 278, 2, 30 },
		{ 278, 3, 30 },
		{ 278, 4, 30 },
		{ 278, 5, 30 },
		{ 278, 6, 30 },
		{ 278, 7, 30 },
		{ 278, 8, 30 },
		{ 278, 9, 30 },
		{ 278, 10, 30 },
		{ 278, 11, 30 },
		{ 278, 12, 30 },
		{ 278, 13, 30 },
		{ 278, 14, 30 },
		{ 278, 15, 30 },
		{ 278, 16, 30 },
		{ 278, 17, 30 },
		{ 278, 18, 30 },
		{ 278, 19, 30 },
		{ 278, 20, 30 },
		{ 278, 21, 30 },
		{ 278, 22, 30 },
		{ 278, 23, 30 },
		{ 278, 24, 30 },
		{ 278, 25, 30 },
		{ 278, 26, 30 },
		{ 278, 27, 30 },
		{ 278, 28, 30 },
		{ 278, 29, 30 },
		{ 278, 30, 30 },
		{ 278, 31, 30 },
		{ 278, 32, 30 },
		{ 278, 33, 30 },
		{ 278, 34, 30 },
		{ 278, 35, 30 },
		{ 278, 36, 30 },
		{ 278, 37, 30 },
		{ 278, 38, 30 },
		{ 278, 39, 30 },
		{ 278, 40, 30 },
		{ 278, 41, 30 },
		{ 278, 42, 30 },
		{ 278, 43, 30 },
		{ 278, 44, 30 },
		{ 278, 45, 30 },
		{ 278, 46, 30 },
		{ 278, 47, 30 },
		{ 278, 48, 30 },
		{ 278, 49, 30 },
		{ 278, 50, 30 },
		{ 278, 51, 30 },
		{ 278, 52, 30 },
		{ 278, 53, 30 },
		{ 278, 54, 30 },
		{ 278, 55, 30 },
		{ 278, 56, 30 },
		{ 278, 57, 30 },
		{ 278, 58, 30 },
		{ 278, 59, 30 },
		{ 278, 60, 30 },
		{ 278, 61, 30 },
		{ 278, 62, 30 },
		{ 278, 63, 30 },
		{ 278, 64, 30 },
		{ 278, 65, 30 },
		{ 278, 66, 30 },
		{ 278, 67, 30 },
		{ 278, 68, 30 },
		{ 278, 69, 30 },
		{ 278, 70, 30 },
		{ 278, 71, 30 },
		{ 278, 72, 30 },
		{ 278, 73, 30 },
		{ 278, 74, 30 },
		{ 278, 75, 30 },
		{ 278, 76, 30 },
		{ 278, 77, 30 },
		{ 278, 78, 30 },
		{ 278, 79, 30 },
		{ 278, 80, 30 },
		{ 278, 81, 30 },
		{ 278, 82, 30 },
		{ 278, 83, 30 },
		{ 278, 84, 30 },
		{ 278, 85, 30 },
		{ 278, 86, 30 },
		{ 278, 87, 30 },
		{ 278, 88, 30 },
		{ 278, 426, 30 },
		{ 278, 427, 30 },
		{ 278, 428, 30 },
		{ 278, 429, 30 },
		{ 278, 430, 30 },
		{ 278, 431, 30 },
		{ 278, 432, 30 },
		{ 278, 433, 30 },
		{ 278, 434, 30 },
		{ 278, 435, 30 },
		{ 278, 436, 30 },
		{ 278, 437, 30 },
		{ 278, 438, 30 },
		{ 278, 439, 30 },
		{ 278, 440, 30 },
		{ 278, 441, 30 },
		{ 278, 442, 30 },
		{ 278, 443, 30 },
		{ 278, 444, 30 },
		{ 278, 445, 30 },
		{ 278, 446, 30 },
		{ 278, 447, 30 },
		{ 278, 448, 30 },
		{ 278, 449, 30 },
		{ 278, 450, 30 },
		{ 278, 451, 30 },
		{ 278, 452, 30 },
		{ 278, 453, 30 },
		{ 278, 454, 30 },
		{ 278, 455, 30 },
		{ 278, 456, 30 },
		{ 278, 457, 30 },
		{ 278, 458, 30 },
		{ 278, 459, 30 },
		{ 278, 460, 30 },
		{ 278, 461, 30 },
		{ 278, 462, 30 },
		{ 278, 463, 30 },
		{ 278, 464, 30 },
		{ 278, 465, 30 },
		{ 278, 466, 30 },
		{ 278, 467, 30 },
		{ 278, 468, 30 },
		{ 278, 469, 30 },
		{ 278, 470, 30 },
		{ 278, 471, 30 },
		{ 278, 472, 30 },
		{ 278, 473, 30 },
		{ 278, 474, 30 },
		{ 278, 475, 30 },
		{ 278, 476, 30 },
		{ 278, 477, 30 },
		{ 278, 478, 30 },
		{ 278, 479, 30 },
		{ 278, 480, 30 },
		{ 278, 481, 30 },
		{ 278, 482, 30 },
		{ 278, 483, 30 },
		{ 278, 484, 30 },
		{ 278, 485, 30 },
		{ 278, 486, 30 },
		{ 278, 487, 30 },
		{ 278, 488, 30 },
		{ 278, 489, 30 },
		{ 278, 490, 30 },
		{ 278, 491, 30 },
		{ 278, 492, 30 },
		{ 278, 493, 30 },
		{ 278, 494, 30 },
		{ 278, 495, 30 },
		{ 278, 496, 30 },
		{ 278, 497, 30 },
		{ 278, 498, 30 },
		{ 278, 499, 30 },
		{ 278, 500, 30 },
		{ 278, 501, 30 },
		{ 278, 502, 30 },
		{ 278, 503, 30 },
		{ 278, 504, 30 },
		{ 278, 505, 30 },
		{ 278, 506, 30 },
		{ 278, 507, 30 },
		{ 278, 508, 30 },
		{ 278, 509, 30 },
		{ 278, 510, 30 },
		{ 278, 511, 30 },
		{ 278, 512, 30 },
		{ 279, 1, 30 },
		{ 279, 2, 30 },
		{ 279, 3, 30 },
		{ 279, 4, 30 },
		{ 279, 5, 30 },
		{ 279, 6, 30 },
		{ 279, 7, 30 },
		{ 279, 8, 30 },
		{ 279, 9, 30 },
		{ 279, 10, 30 },
		{ 279, 11, 30 },
		{ 279, 12, 30 },
		{ 279, 13, 30 },
		{ 279, 14, 30 },
		{ 279, 15, 30 },
		{ 279, 16, 30 },
		{ 279, 17, 30 },
		{ 279, 18, 30 },
		{ 279, 19, 30 },
		{ 279, 20, 30 },
		{ 279, 21, 30 },
		{ 279, 22, 30 },
		{ 279, 23, 30 },
		{ 279, 24, 30 },
		{ 279, 25, 30 },
		{ 279, 26, 30 },
		{ 279, 27, 30 },
		{ 279, 28, 30 },
		{ 279, 29, 30 },
		{ 279, 30, 30 },
		{ 279, 31, 30 },
		{ 279, 32, 30 },
		{ 279, 33, 30 },
		{ 279, 34, 30 },
		{ 279, 35, 30 },
		{ 279, 36, 30 },
		{ 279, 37, 30 },
		{ 279, 38, 30 },
		{ 279, 39, 30 },
		{ 279, 40, 30 },
		{ 279, 41, 30 },
		{ 279, 42, 30 },
		{ 279, 43, 30 },
		{ 279, 44, 30 },
		{ 279, 45, 30 },
		{ 279, 46, 30 },
		{ 279, 47, 30 },
		{ 279, 48, 30 },
		{ 279, 49, 30 },
		{ 279, 50, 30 },
		{ 279, 51, 30 },
		{ 279, 52, 30 },
		{ 279, 53, 30 },
		{ 279, 54, 30 },
		{ 279, 55, 30 },
		{ 279, 56, 30 },
		{ 279, 57, 30 },
		{ 279, 58, 30 },
		{ 279, 59, 30 },
		{ 279, 60, 30 },
		{ 279, 61, 30 },
		{ 279, 62, 30 },
		{ 279, 63, 30 },
		{ 279, 64, 30 },
		{ 279, 65, 30 },
		{ 279, 66, 30 },
		{ 279, 67, 30 },
		{ 279, 68, 30 },
		{ 279, 69, 30 },
		{ 279, 70, 30 },
		{ 279, 71, 30 },
		{ 279, 72, 30 },
		{ 279, 73, 30 },
		{ 279, 74, 30 },
		{ 279, 75, 30 },
		{ 279, 76, 30 },
		{ 279, 77, 30 },
		{ 279, 78, 30 },
		{ 279, 79, 30 },
		{ 279, 80, 30 },
		{ 279, 434, 30 },
		{ 279, 435, 30 },
		{ 279, 436, 30 },
		{ 279, 437, 30 },
		{ 279, 438, 30 },
		{ 279, 439, 30 },
		{ 279, 440, 30 },
		{ 279, 441, 30 },
		{ 279, 442, 30 },
		{ 279, 443, 30 },
		{ 279, 444, 30 },
		{ 279, 445, 30 },
		{ 279, 446, 30 },
		{ 279, 447, 30 },
		{ 279, 448, 30 },
		{ 279, 449, 30 },
		{ 279, 450, 30 },
		{ 279, 451, 30 },
		{ 279, 452, 30 },
		{ 279, 453, 30 },
		{ 279, 454, 30 },
		{ 279, 455, 30 },
		{ 279, 456, 30 },
		{ 279, 457, 30 },
		{ 279, 458, 30 },
		{ 279, 459, 30 },
		{ 279, 460, 30 },
		{ 279, 461, 30 },
		{ 279, 462, 30 },
		{ 279, 463, 30 },
		{ 279, 464, 30 },
		{ 279, 465, 30 },
		{ 279, 466, 30 },
		{ 279, 467, 30 },
		{ 279, 468, 30 },
		{ 279, 469, 30 },
		{ 279, 470, 30 },
		{ 279, 471, 30 },
		{ 279, 472, 30 },
		{ 279, 473, 30 },
		{ 279, 474, 30 },
		{ 279, 475, 30 },
		{ 279, 476, 30 },
		{ 279, 477, 30 },
		{ 279, 478, 30 },
		{ 279, 479, 30 },
		{ 279, 480, 30 },
		{ 279, 481, 30 },
		{ 279, 482, 30 },
		{ 279, 483, 30 },
		{ 279, 484, 30 },
		{ 279, 485, 30 },
		{ 279, 486, 30 },
		{ 279, 487, 30 },
		{ 279, 488, 30 },
		{ 279, 489, 30 },
		{ 279, 490, 30 },
		{ 279, 491, 30 },
		{ 279, 492, 30 },
		{ 279, 493, 30 },
		{ 279, 494, 30 },
		{ 279, 495, 30 },
		{ 279, 496, 30 },
		{ 279, 497, 30 },
		{ 279, 498, 30 },
		{ 279, 499, 30 },
		{ 279, 500, 30 },
		{ 279, 501, 30 },
		{ 279, 502, 30 },
		{ 279, 503, 30 },
		{ 279, 504, 30 },
		{ 279, 505, 30 },
		{ 279, 506, 30 },
		{ 279, 507, 30 },
		{ 279, 508, 30 },
		{ 279, 509, 30 },
		{ 279, 510, 30 },
		{ 279, 511, 30 },
		{ 279, 512, 30 },
		{ 280, 1, 30 },
		{ 280, 2, 30 },
		{ 280, 3, 30 },
		{ 280, 4, 30 },
		{ 280, 5, 30 },
		{ 280, 6, 30 },
		{ 280, 7, 30 },
		{ 280, 8, 30 },
		{ 280, 9, 30 },
		{ 280, 10, 30 },
		{ 280, 11, 30 },
		{ 280, 12, 30 },
		{ 280, 13, 30 },
		{ 280, 14, 30 },
		{ 280, 15, 30 },
		{ 280, 16, 30 },
		{ 280, 17, 30 },
		{ 280, 18, 30 },
		{ 280, 19, 30 },
		{ 280, 20, 30 },
		{ 280, 21, 30 },
		{ 280, 22, 30 },
		{ 280, 23, 30 },
		{ 280, 24, 30 },
		{ 280, 25, 30 },
		{ 280, 26, 30 },
		{ 280, 27, 30 },
		{ 280, 28, 30 },
		{ 280, 29, 30 },
		{ 280, 30, 30 },
		{ 280, 31, 30 },
		{ 280, 32, 30 },
		{ 280, 33, 30 },
		{ 280, 34, 30 },
		{ 280, 35, 30 },
		{ 280, 36, 30 },
		{ 280, 37, 30 },
		{ 280, 38, 30 },
		{ 280, 39, 30 },
		{ 280, 40, 30 },
		{ 280, 41, 30 },
		{ 280, 42, 30 },
		{ 280, 43, 30 },
		{ 280, 44, 30 },
		{ 280, 45, 30 },
		{ 280, 46, 30 },
		{ 280, 47, 30 },
		{ 280, 48, 30 },
		{ 280, 49, 30 },
		{ 280, 50, 30 },
		{ 280, 51, 30 },
		{ 280, 52, 30 },
		{ 280, 53, 30 },
		{ 280, 54, 30 },
		{ 280, 55, 30 },
		{ 280, 56, 30 },
		{ 280, 57, 30 },
		{ 280, 58, 30 },
		{ 280, 59, 30 },
		{ 280, 60, 30 },
		{ 280, 61, 30 },
		{ 280, 62, 30 },
		{ 280, 63, 30 },
		{ 280, 64, 30 },
		{ 280, 65, 30 },
		{ 280, 66, 30 },
		{ 280, 67, 30 },
		{ 280, 68, 30 },
		{ 280, 69, 30 },
		{ 280, 70, 30 },
		{ 280, 71, 30 },
		{ 280, 72, 30 },
		{ 280, 442, 30 },
		{ 280, 443, 30 },
		{ 280, 444, 30 },
		{ 280, 445, 30 },
		{ 280, 446, 30 },
		{ 280, 447, 30 },
		{ 280, 448, 30 },
		{ 280, 449, 30 },
		{ 280, 450, 30 },
		{ 280, 451, 30 },
		{ 280, 452, 30 },
		{ 280, 453, 30 },
		{ 280, 454, 30 },
		{ 280, 455, 30 },
		{ 280, 456, 30 },
		{ 280, 457, 30 },
		{ 280, 458, 30 },
		{ 280, 459, 30 },
		{ 280, 460, 30 },
		{ 280, 461, 30 },
		{ 280, 462, 30 },
		{ 280, 463, 30 },
		{ 280, 464, 30 },
		{ 280, 465, 30 },
		{ 280, 466, 30 },
		{ 280, 467, 30 },
		{ 280, 468, 30 },
		{ 280, 469, 30 },
		{ 280, 470, 30 },
		{ 280, 471, 30 },
		{ 280, 472, 30 },
		{ 280, 473, 30 },
		{ 280, 474, 30 },
		{ 280, 475, 30 },
		{ 280, 476, 30 },
		{ 280, 477, 30 },
		{ 280, 478, 30 },
		{ 280, 479, 30 },
		{ 280, 480, 30 },
		{ 280, 481, 30 },
		{ 280, 482, 30 },
		{ 280, 483, 30 },
		{ 280, 484, 30 },
		{ 280, 485, 30 },
		{ 280, 486, 30 },
		{ 280, 487, 30 },
		{ 280, 488, 30 },
		{ 280, 489, 30 },
		{ 280, 490, 30 },
		{ 280, 491, 30 },
		{ 280, 492, 30 },
		{ 280, 493, 30 },
		{ 280, 494, 30 },
		{ 280, 495, 30 },
		{ 280, 496, 30 },
		{ 280, 497, 30 },
		{ 280, 498, 30 },
		{ 280, 499, 30 },
		{ 280, 500, 30 },
		{ 280, 501, 30 },
		{ 280, 502, 30 },
		{ 280, 503, 30 },
		{ 280, 504, 30 },
		{ 280, 505, 30 },
		{ 280, 506, 30 },
		{ 280, 507, 30 },
		{ 280, 508, 30 },
		{ 280, 509, 30 },
		{ 280, 510, 30 },
		{ 280, 511, 30 },
		{ 280, 512, 30 },
		{ 281, 1, 30 },
		{ 281, 2, 30 },
		{ 281, 3, 30 },
		{ 281, 4, 30 },
		{ 281, 5, 30 },
		{ 281, 6, 30 },
		{ 281, 7, 30 },
		{ 281, 8, 30 },
		{ 281, 9, 30 },
		{ 281, 10, 30 },
		{ 281, 11, 30 },
		{ 281, 12, 30 },
		{ 281, 13, 30 },
		{ 281, 14, 30 },
		{ 281, 15, 30 },
		{ 281, 16, 30 },
		{ 281, 17, 30 },
		{ 281, 18, 30 },
		{ 281, 19, 30 },
		{ 281, 20, 30 },
		{ 281, 21, 30 },
		{ 281, 22, 30 },
		{ 281, 23, 30 },
		{ 281, 24, 30 },
		{ 281, 25, 30 },
		{ 281, 26, 30 },
		{ 281, 27, 30 },
		{ 281, 28, 30 },
		{ 281, 29, 30 },
		{ 281, 30, 30 },
		{ 281, 31, 30 },
		{ 281, 32, 30 },
		{ 281, 33, 30 },
		{ 281, 34, 30 },
		{ 281, 35, 30 },
		{ 281, 36, 30 },
		{ 281, 37, 30 },
		{ 281, 38, 30 },
		{ 281, 39, 30 },
		{ 281, 40, 30 },
		{ 281, 41, 30 },
		{ 281, 42, 30 },
		{ 281, 43, 30 },
		{ 281, 44, 30 },
		{ 281, 45, 30 },
		{ 281, 46, 30 },
		{ 281, 47, 30 },
		{ 281, 48, 30 },
		{ 281, 49, 30 },
		{ 281, 50, 30 },
		{ 281, 51, 30 },
		{ 281, 52, 30 },
		{ 281, 53, 30 },
		{ 281, 54, 30 },
		{ 281, 55, 30 },
		{ 281, 56, 30 },
		{ 281, 57, 30 },
		{ 281, 58, 30 },
		{ 281, 59, 30 },
		{ 281, 60, 30 },
		{ 281, 61, 30 },
		{ 281, 62, 30 },
		{ 281, 63, 30 },
		{ 281, 64, 30 },
		{ 281, 450, 30 },
		{ 281, 451, 30 },
		{ 281, 452, 30 },
		{ 281, 453, 30 },
		{ 281, 454, 30 },
		{ 281, 455, 30 },
		{ 281, 456, 30 },
		{ 281, 457, 30 },
		{ 281, 458, 30 },
		{ 281, 459, 30 },
		{ 281, 460, 30 },
		{ 281, 461, 30 },
		{ 281, 462, 30 },
		{ 281, 463, 30 },
		{ 281, 464, 30 },
		{ 281, 465, 30 },
		{ 281, 466, 30 },
		{ 281, 467, 30 },
		{ 281, 468, 30 },
		{ 281, 469, 30 },
		{ 281, 470, 30 },
		{ 281, 471, 30 },
		{ 281, 472, 30 },
		{ 281, 473, 30 },
		{ 281, 474, 30 },
		{ 281, 475, 30 },
		{ 281, 476, 30 },
		{ 281, 477, 30 },
		{ 281, 478, 30 },
		{ 281, 479, 30 },
		{ 281, 480, 30 },
		{ 281, 481, 30 },
		{ 281, 482, 30 },
		{ 281, 483, 30 },
		{ 281, 484, 30 },
		{ 281, 485, 30 },
		{ 281, 486, 30 },
		{ 281, 487, 30 },
		{ 281, 488, 30 },
		{ 281, 489, 30 },
		{ 281, 490, 30 },
		{ 281, 491, 30 },
		{ 281, 492, 30 },
		{ 281, 493, 30 },
		{ 281, 494, 30 },
		{ 281, 495, 30 },
		{ 281, 496, 30 },
		{ 281, 497, 30 },
		{ 281, 498, 30 },
		{ 281, 499, 30 },
		{ 281, 500, 30 },
		{ 281, 501, 30 },
		{ 281, 502, 30 },
		{ 281, 503, 30 },
		{ 281, 504, 30 },
		{ 281, 505, 30 },
		{ 281, 506, 30 },
		{ 281, 507, 30 },
		{ 281, 508, 30 },
		{ 281, 509, 30 },
		{ 281, 510, 30 },
		{ 281, 511, 30 },
		{ 281, 512, 30 },
		{ 282, 1, 30 },
		{ 282, 2, 30 },
		{ 282, 3, 30 },
		{ 282, 4, 30 },
		{ 282, 5, 30 },
		{ 282, 6, 30 },
		{ 282, 7, 30 },
		{ 282, 8, 30 },
		{ 282, 9, 30 },
		{ 282, 10, 30 },
		{ 282, 11, 30 },
		{ 282, 12, 30 },
		{ 282, 13, 30 },
		{ 282, 14, 30 },
		{ 282, 15, 30 },
		{ 282, 16, 30 },
		{ 282, 17, 30 },
		{ 282, 18, 30 },
		{ 282, 19, 30 },
		{ 282, 20, 30 },
		{ 282, 21, 30 },
		{ 282, 22, 30 },
		{ 282, 23, 30 },
		{ 282, 24, 30 },
		{ 282, 25, 30 },
		{ 282, 26, 30 },
		{ 282, 27, 30 },
		{ 282, 28, 30 },
		{ 282, 29, 30 },
		{ 282, 30, 30 },
		{ 282, 31, 30 },
		{ 282, 32, 30 },
		{ 282, 33, 30 },
		{ 282, 34, 30 },
		{ 282, 35, 30 },
		{ 282, 36, 30 },
		{ 282, 37, 30 },
		{ 282, 38, 30 },
		{ 282, 39, 30 },
		{ 282, 40, 30 },
		{ 282, 41, 30 },
		{ 282, 42, 30 },
		{ 282, 43, 30 },
		{ 282, 44, 30 },
		{ 282, 45, 30 },
		{ 282, 46, 30 },
		{ 282, 47, 30 },
		{ 282, 48, 30 },
		{ 282, 49, 30 },
		{ 282, 50, 30 },
		{ 282, 51, 30 },
		{ 282, 52, 30 },
		{ 282, 53, 30 },
		{ 282, 54, 30 },
		{ 282, 55, 30 },
		{ 282, 56, 30 },
		{ 282, 458, 30 },
		{ 282, 459, 30 },
		{ 282, 460, 30 },
		{ 282, 461, 30 },
		{ 282, 462, 30 },
		{ 282, 463, 30 },
		{ 282, 464, 30 },
		{ 282, 465, 30 },
		{ 282, 466, 30 },
		{ 282, 467, 30 },
		{ 282, 468, 30 },
		{ 282, 469, 30 },
		{ 282, 470, 30 },
		{ 282, 471, 30 },
		{ 282, 472, 30 },
		{ 282, 473, 30 },
		{ 282, 474, 30 },
		{ 282, 475, 30 },
		{ 282, 476, 30 },
		{ 282, 477, 30 },
		{ 282, 478, 30 },
		{ 282, 479, 30 },
		{ 282, 480, 30 },
		{ 282, 481, 30 },
		{ 282, 482, 30 },
		{ 282, 483, 30 },
		{ 282, 484, 30 },
		{ 282, 485, 30 },
		{ 282, 486, 30 },
		{ 282, 487, 30 },
		{ 282, 488, 30 },
		{ 282, 489, 30 },
		{ 282, 490, 30 },
		{ 282, 491, 30 },
		{ 282, 492, 30 },
		{ 282, 493, 30 },
		{ 282, 494, 30 },
		{ 282, 495, 30 },
		{ 282, 496, 30 },
		{ 282, 497, 30 },
		{ 282, 498, 30 },
		{ 282, 499, 30 },
		{ 282, 500, 30 },
		{ 282, 501, 30 },
		{ 282, 502, 30 },
		{ 282, 503, 30 },
		{ 282, 504, 30 },
		{ 282, 505, 30 },
		{ 282, 506, 30 },
		{ 282, 507, 30 },
		{ 282, 508, 30 },
		{ 282, 509, 30 },
		{ 282, 510, 30 },
		{ 282, 511, 30 },
		{ 282, 512, 30 },
		{ 283, 1, 30 },
		{ 283, 2, 30 },
		{ 283, 3, 30 },
		{ 283, 4, 30 },
		{ 283, 5, 30 },
		{ 283, 6, 30 },
		{ 283, 7, 30 },
		{ 283, 8, 30 },
		{ 283, 9, 30 },
		{ 283, 10, 30 },
		{ 283, 11, 30 },
		{ 283, 12, 30 },
		{ 283, 13, 30 },
		{ 283, 14, 30 },
		{ 283, 15, 30 },
		{ 283, 16, 30 },
		{ 283, 17, 30 },
		{ 283, 18, 30 },
		{ 283, 19, 30 },
		{ 283, 20, 30 },
		{ 283, 21, 30 },
		{ 283, 22, 30 },
		{ 283, 23, 30 },
		{ 283, 24, 30 },
		{ 283, 25, 30 },
		{ 283, 26, 30 },
		{ 283, 27, 30 },
		{ 283, 28, 30 },
		{ 283, 29, 30 },
		{ 283, 30, 30 },
		{ 283, 31, 30 },
		{ 283, 32, 30 },
		{ 283, 33, 30 },
		{ 283, 34, 30 },
		{ 283, 35, 30 },
		{ 283, 36, 30 },
		{ 283, 37, 30 },
		{ 283, 38, 30 },
		{ 283, 39, 30 },
		{ 283, 40, 30 },
		{ 283, 41, 30 },
		{ 283, 42, 30 },
		{ 283, 43, 30 },
		{ 283, 44, 30 },
		{ 283, 45, 30 },
		{ 283, 46, 30 },
		{ 283, 47, 30 },
		{ 283, 48, 30 },
		{ 283, 466, 30 },
		{ 283, 467, 30 },
		{ 283, 468, 30 },
		{ 283, 469, 30 },
		{ 283, 470, 30 },
		{ 283, 471, 30 },
		{ 283, 472, 30 },
		{ 283, 473, 30 },
		{ 283, 474, 30 },
		{ 283, 475, 30 },
		{ 283, 476, 30 },
		{ 283, 477, 30 },
		{ 283, 478, 30 },
		{ 283, 479, 30 },
		{ 283, 480, 30 },
		{ 283, 481, 30 },
		{ 283, 482, 30 },
		{ 283, 483, 30 },
		{ 283, 484, 30 },
		{ 283, 485, 30 },
		{ 283, 486, 30 },
		{ 283, 487, 30 },
		{ 283, 488, 30 },
		{ 283, 489, 30 },
		{ 283, 490, 30 },
		{ 283, 491, 30 },
		{ 283, 492, 30 },
		{ 283, 493, 30 },
		{ 283, 494, 30 },
		{ 283, 495, 30 },
		{ 283, 496, 30 },
		{ 283, 497, 30 },
		{ 283, 498, 30 },
		{ 283, 499, 30 },
		{ 283, 500, 30 },
		{ 283, 501, 30 },
		{ 283, 502, 30 },
		{ 283, 503, 30 },
		{ 283, 504, 30 },
		{ 283, 505, 30 },
		{ 283, 506, 30 },
		{ 283, 507, 30 },
		{ 283, 508, 30 },
		{ 283, 509, 30 },
		{ 283, 510, 30 },
		{ 283, 511, 30 },
		{ 283, 512, 30 },
		{ 284, 1, 30 },
		{ 284, 2, 30 },
		{ 284, 3, 30 },
		{ 284, 4, 30 },
		{ 284, 5, 30 },
		{ 284, 6, 30 },
		{ 284, 7, 30 },
		{ 284, 8, 30 },
		{ 284, 9, 30 },
		{ 284, 10, 30 },
		{ 284, 11, 30 },
		{ 284, 12, 30 },
		{ 284, 13, 30 },
		{ 284, 14, 30 },
		{ 284, 15, 30 },
		{ 284, 16, 30 },
		{ 284, 17, 30 },
		{ 284, 18, 30 },
		{ 284, 19, 30 },
		{ 284, 20, 30 },
		{ 284, 21, 30 },
		{ 284, 22, 30 },
		{ 284, 23, 30 },
		{ 284, 24, 30 },
		{ 284, 25, 30 },
		{ 284, 26, 30 },
		{ 284, 27, 30 },
		{ 284, 28, 30 },
		{ 284, 29, 30 },
		{ 284, 30, 30 },
		{ 284, 31, 30 },
		{ 284, 32, 30 },
		{ 284, 33, 30 },
		{ 284, 34, 30 },
		{ 284, 35, 30 },
		{ 284, 36, 30 },
		{ 284, 37, 30 },
		{ 284, 38, 30 },
		{ 284, 39, 30 },
		{ 284, 40, 30 },
		{ 284, 474, 30 },
		{ 284, 475, 30 },
		{ 284, 476, 30 },
		{ 284, 477, 30 },
		{ 284, 478, 30 },
		{ 284, 479, 30 },
		{ 284, 480, 30 },
		{ 284, 481, 30 },
		{ 284, 482, 30 },
		{ 284, 483, 30 },
		{ 284, 484, 30 },
		{ 284, 485, 30 },
		{ 284, 486, 30 },
		{ 284, 487, 30 },
		{ 284, 488, 30 },
		{ 284, 489, 30 },
		{ 284, 490, 30 },
		{ 284, 491, 30 },
		{ 284, 492, 30 },
		{ 284, 493, 30 },
		{ 284, 494, 30 },
		{ 284, 495, 30 },
		{ 284, 496, 30 },
		{ 284, 497, 30 },
		{ 284, 498, 30 },
		{ 284, 499, 30 },
		{ 284, 500, 30 },
		{ 284, 501, 30 },
		{ 284, 502, 30 },
		{ 284, 503, 30 },
		{ 284, 504, 30 },
		{ 284, 505, 30 },
		{ 284, 506, 30 },
		{ 284, 507, 30 },
		{ 284, 508, 30 },
		{ 284, 509, 30 },
		{ 284, 510, 30 },
		{ 284, 511, 30 },
		{ 284, 512, 30 },
		{ 285, 1, 30 },
		{ 285, 2, 30 },
		{ 285, 3, 30 },
		{ 285, 4, 30 },
		{ 285, 5, 30 },
		{ 285, 6, 30 },
		{ 285, 7, 30 },
		{ 285, 8, 30 },
		{ 285, 9, 30 },
		{ 285, 10, 30 },
		{ 285, 11, 30 },
		{ 285, 12, 30 },
		{ 285, 13, 30 },
		{ 285, 14, 30 },
		{ 285, 15, 30 },
		{ 285, 16, 30 },
		{ 285, 17, 30 },
		{ 285, 18, 30 },
		{ 285, 19, 30 },
		{ 285, 20, 30 },
		{ 285, 21, 30 },
		{ 285, 22, 30 },
		{ 285, 23, 30 },
		{ 285, 24, 30 },
		{ 285, 25, 30 },
		{ 285, 26, 30 },
		{ 285, 27, 30 },
		{ 285, 28, 30 },
		{ 285, 29, 30 },
		{ 285, 30, 30 },
		{ 285, 31, 30 },
		{ 285, 32, 30 },
		{ 285, 482, 30 },
		{ 285, 483, 30 },
		{ 285, 484, 30 },
		{ 285, 485, 30 },
		{ 285, 486, 30 },
		{ 285, 487, 30 },
		{ 285, 488, 30 },
		{ 285, 489, 30 },
		{ 285, 490, 30 },
		{ 285, 491, 30 },
		{ 285, 492, 30 },
		{ 285, 493, 30 },
		{ 285, 494, 30 },
		{ 285, 495, 30 },
		{ 285, 496, 30 },
		{ 285, 497, 30 },
		{ 285, 498, 30 },
		{ 285, 499, 30 },
		{ 285, 500, 30 },
		{ 285, 501, 30 },
		{ 285, 502, 30 },
		{ 285, 503, 30 },
		{ 285, 504, 30 },
		{ 285, 505, 30 },
		{ 285, 506, 30 },
		{ 285, 507, 30 },
		{ 285, 508, 30 },
		{ 285, 509, 30 },
		{ 285, 510, 30 },
		{ 285, 511, 30 },
		{ 285, 512, 30 },
		{ 286, 1, 30 },
		{ 286, 2, 30 },
		{ 286, 3, 30 },
		{ 286, 4, 30 },
		{ 286, 5, 30 },
		{ 286, 6, 30 },
		{ 286, 7, 30 },
		{ 286, 8, 30 },
		{ 286, 9, 30 },
		{ 286, 10, 30 },
		{ 286, 11, 30 },
		{ 286, 12, 30 },
		{ 286, 13, 30 },
		{ 286, 14, 30 },
		{ 286, 15, 30 },
		{ 286, 16, 30 },
		{ 286, 17, 30 },
		{ 286, 18, 30 },
		{ 286, 19, 30 },
		{ 286, 20, 30 },
		{ 286, 21, 30 },
		{ 286, 22, 30 },
		{ 286, 23, 30 },
		{ 286, 24, 30 },
		{ 286, 490, 30 },
		{ 286, 491, 30 },
		{ 286, 492, 30 },
		{ 286, 493, 30 },
		{ 286, 494, 30 },
		{ 286, 495, 30 },
		{ 286, 496, 30 },
		{ 286, 497, 30 },
		{ 286, 498, 30 },
		{ 286, 499, 30 },
		{ 286, 500, 30 },
		{ 286, 501, 30 },
		{ 286, 502, 30 },
		{ 286, 503, 30 },
		{ 286, 504, 30 },
		{ 286, 505, 30 },
		{ 286, 506, 30 },
		{ 286, 507, 30 },
		{ 286, 508, 30 },
		{ 286, 509, 30 },
		{ 286, 510, 30 },
		{ 286, 511, 30 },
		{ 286, 512, 30 },
		{ 287, 1, 30 },
		{ 287, 2, 30 },
		{ 287, 3, 30 },
		{ 287, 4, 30 },
		{ 287, 5, 30 },
		{ 287, 6, 30 },
		{ 287, 7, 30 },
		{ 287, 8, 30 },
		{ 287, 9, 30 },
		{ 287, 10, 30 },
		{ 287, 11, 30 },
		{ 287, 12, 30 },
		{ 287, 13, 30 },
		{ 287, 14, 30 },
		{ 287, 15, 30 },
		{ 287, 16, 30 },
		{ 287, 498, 30 },
		{ 287, 499, 30 },
		{ 287, 500, 30 },
		{ 287, 501, 30 },
		{ 287, 502, 30 },
		{ 287, 503, 30 },
		{ 287, 504, 30 },
		{ 287, 505, 30 },
		{ 287, 506, 30 },
		{ 287, 507, 30 },
		{ 287, 508, 30 },
		{ 287, 509, 30 },
		{ 287, 510, 30 },
		{ 287, 511, 30 },
		{ 287, 512, 30 },
		{ 288, 1, 30 },
		{ 288, 2, 30 },
		{ 288, 3, 30 },
		{ 288, 4, 30 },
		{ 288, 5, 30 },
		{ 288, 6, 30 },
		{ 288, 7, 30 },
		{ 288, 8, 30 },
		{ 288, 506, 30 },
		{ 288, 507, 30 },
		{ 288, 508, 30 },
		{ 288, 509, 30 },
		{ 288, 510, 30 },
		{ 288, 511, 30 },
		{ 288, 512, 30 },
		{ 194, 510, 31 },
		{ 194, 511, 31 },
		{ 194, 512, 31 },
		{ 195, 506, 31 },
		{ 195, 507, 31 },
		{ 195, 508, 31 },
		{ 195, 509, 31 },
		{ 195, 510, 31 },
		{ 195, 511, 31 },
		{ 195, 512, 31 },
		{ 196, 502, 31 },
		{ 196, 503, 31 },
		{ 196, 504, 31 },
		{ 196, 505, 31 },
		{ 196, 506, 31 },
		{ 196, 507, 31 },
		{ 196, 508, 31 },
		{ 196, 509, 31 },
		{ 196, 510, 31 },
		{ 196, 511, 31 },
		{ 196, 512, 31 },
		{ 197, 498, 31 },
		{ 197, 499, 31 },
		{ 197, 500, 31 },
		{ 197, 501, 31 },
		{ 197, 502, 31 },
		{ 197, 503, 31 },
		{ 197, 504, 31 },
		{ 197, 505, 31 },
		{ 197, 506, 31 },
		{ 197, 507, 31 },
		{ 197, 508, 31 },
		{ 197, 509, 31 },
		{ 197, 510, 31 },
		{ 197, 511, 31 },
		{ 197, 512, 31 },
		{ 198, 494, 31 },
		{ 198, 495, 31 },
		{ 198, 496, 31 },
		{ 198, 497, 31 },
		{ 198, 498, 31 },
		{ 198, 499, 31 },
		{ 198, 500, 31 },
		{ 198, 501, 31 },
		{ 198, 502, 31 },
		{ 198, 503, 31 },
		{ 198, 504, 31 },
		{ 198, 505, 31 },
		{ 198, 506, 31 },
		{ 198, 507, 31 },
		{ 198, 508, 31 },
		{ 198, 509, 31 },
		{ 198, 510, 31 },
		{ 198, 511, 31 },
		{ 198, 512, 31 },
		{ 199, 490, 31 },
		{ 199, 491, 31 },
		{ 199, 492, 31 },
		{ 199, 493, 31 },
		{ 199, 494, 31 },
		{ 199, 495, 31 },
		{ 199, 496, 31 },
		{ 199, 497, 31 },
		{ 199, 498, 31 },
		{ 199, 499, 31 },
		{ 199, 500, 31 },
		{ 199, 501, 31 },
		{ 199, 502, 31 },
		{ 199, 503, 31 },
		{ 199, 504, 31 },
		{ 199, 505, 31 },
		{ 199, 506, 31 },
		{ 199, 507, 31 },
		{ 199, 508, 31 },
		{ 199, 509, 31 },
		{ 199, 510, 31 },
		{ 199, 511, 31 },
		{ 199, 512, 31 },
		{ 200, 486, 31 },
		{ 200, 487, 31 },
		{ 200, 488, 31 },
		{ 200, 489, 31 },
		{ 200, 490, 31 },
		{ 200, 491, 31 },
		{ 200, 492, 31 },
		{ 200, 493, 31 },
		{ 200, 494, 31 },
		{ 200, 495, 31 },
		{ 200, 496, 31 },
		{ 200, 497, 31 },
		{ 200, 498, 31 },
		{ 200, 499, 31 },
		{ 200, 500, 31 },
		{ 200, 501, 31 },
		{ 200, 502, 31 },
		{ 200, 503, 31 },
		{ 200, 504, 31 },
		{ 200, 505, 31 },
		{ 200, 506, 31 },
		{ 200, 507, 31 },
		{ 200, 508, 31 },
		{ 200, 509, 31 },
		{ 200, 510, 31 },
		{ 200, 511, 31 },
		{ 200, 512, 31 },
		{ 201, 482, 31 },
		{ 201, 483, 31 },
		{ 201, 484, 31 },
		{ 201, 485, 31 },
		{ 201, 486, 31 },
		{ 201, 487, 31 },
		{ 201, 488, 31 },
		{ 201, 489, 31 },
		{ 201, 490, 31 },
		{ 201, 491, 31 },
		{ 201, 492, 31 },
		{ 201, 493, 31 },
		{ 201, 494, 31 },
		{ 201, 495, 31 },
		{ 201, 496, 31 },
		{ 201, 497, 31 },
		{ 201, 498, 31 },
		{ 201, 499, 31 },
		{ 201, 500, 31 },
		{ 201, 501, 31 },
		{ 201, 502, 31 },
		{ 201, 503, 31 },
		{ 201, 504, 31 },
		{ 201, 505, 31 },
		{ 201, 506, 31 },
		{ 201, 507, 31 },
		{ 201, 508, 31 },
		{ 201, 509, 31 },
		{ 201, 510, 31 },
		{ 201, 511, 31 },
		{ 201, 512, 31 },
		{ 202, 478, 31 },
		{ 202, 479, 31 },
		{ 202, 480, 31 },
		{ 202, 481, 31 },
		{ 202, 482, 31 },
		{ 202, 483, 31 },
		{ 202, 484, 31 },
		{ 202, 485, 31 },
		{ 202, 486, 31 },
		{ 202, 487, 31 },
		{ 202, 488, 31 },
		{ 202, 489, 31 },
		{ 202, 490, 31 },
		{ 202, 491, 31 },
		{ 202, 492, 31 },
		{ 202, 493, 31 },
		{ 202, 494, 31 },
		{ 202, 495, 31 },
		{ 202, 496, 31 },
		{ 202, 497, 31 },
		{ 202, 498, 31 },
		{ 202, 499, 31 },
		{ 202, 500, 31 },
		{ 202, 501, 31 },
		{ 202, 502, 31 },
		{ 202, 503, 31 },
		{ 202, 504, 31 },
		{ 202, 505, 31 },
		{ 202, 506, 31 },
		{ 202, 507, 31 },
		{ 202, 508, 31 },
		{ 202, 509, 31 },
		{ 202, 510, 31 },
		{ 202, 511, 31 },
		{ 202, 512, 31 },
		{ 203, 474, 31 },
		{ 203, 475, 31 },
		{ 203, 476, 31 },
		{ 203, 477, 31 },
		{ 203, 478, 31 },
		{ 203, 479, 31 },
		{ 203, 480, 31 },
		{ 203, 481, 31 },
		{ 203, 482, 31 },
		{ 203, 483, 31 },
		{ 203, 484, 31 },
		{ 203, 485, 31 },
		{ 203, 486, 31 },
		{ 203, 487, 31 },
		{ 203, 488, 31 },
		{ 203, 489, 31 },
		{ 203, 490, 31 },
		{ 203, 491, 31 },
		{ 203, 492, 31 },
		{ 203, 493, 31 },
		{ 203, 494, 31 },
		{ 203, 495, 31 },
		{ 203, 496, 31 },
		{ 203, 497, 31 },
		{ 203, 498, 31 },
		{ 203, 499, 31 },
		{ 203, 500, 31 },
		{ 203, 501, 31 },
		{ 203, 502, 31 },
		{ 203, 503, 31 },
		{ 203, 504, 31 },
		{ 203, 505, 31 },
		{ 203, 506, 31 },
		{ 203, 507, 31 },
		{ 203, 508, 31 },
		{ 203, 509, 31 },
		{ 203, 510, 31 },
		{ 203, 511, 31 },
		{ 203, 512, 31 },
		{ 204, 470, 31 },
		{ 204, 471, 31 },
		{ 204, 472, 31 },
		{ 204, 473, 31 },
		{ 204, 474, 31 },
		{ 204, 475, 31 },
		{ 204, 476, 31 },
		{ 204, 477, 31 },
		{ 204, 478, 31 },
		{ 204, 479, 31 },
		{ 204, 480, 31 },
		{ 204, 481, 31 },
		{ 204, 482, 31 },
		{ 204, 483, 31 },
		{ 204, 484, 31 },
		{ 204, 485, 31 },
		{ 204, 486, 31 },
		{ 204, 487, 31 },
		{ 204, 488, 31 },
		{ 204, 489, 31 },
		{ 204, 490, 31 },
		{ 204, 491, 31 },
		{ 204, 492, 31 },
		{ 204, 493, 31 },
		{ 204, 494, 31 },
		{ 204, 495, 31 },
		{ 204, 496, 31 },
		{ 204, 497, 31 },
		{ 204, 498, 31 },
		{ 204, 499, 31 },
		{ 204, 500, 31 },
		{ 204, 501, 31 },
		{ 204, 502, 31 },
		{ 204, 503, 31 },
		{ 204, 504, 31 },
		{ 204, 505, 31 },
		{ 204, 506, 31 },
		{ 204, 507, 31 },
		{ 204, 508, 31 },
		{ 204, 509, 31 },
		{ 204, 510, 31 },
		{ 204, 511, 31 },
		{ 204, 512, 31 },
		{ 205, 466, 31 },
		{ 205, 467, 31 },
		{ 205, 468, 31 },
		{ 205, 469, 31 },
		{ 205, 470, 31 },
		{ 205, 471, 31 },
		{ 205, 472, 31 },
		{ 205, 473, 31 },
		{ 205, 474, 31 },
		{ 205, 475, 31 },
		{ 205, 476, 31 },
		{ 205, 477, 31 },
		{ 205, 478, 31 },
		{ 205, 479, 31 },
		{ 205, 480, 31 },
		{ 205, 481, 31 },
		{ 205, 482, 31 },
		{ 205, 483, 31 },
		{ 205, 484, 31 },
		{ 205, 485, 31 },
		{ 205, 486, 31 },
		{ 205, 487, 31 },
		{ 205, 488, 31 },
		{ 205, 489, 31 },
		{ 205, 490, 31 },
		{ 205, 491, 31 },
		{ 205, 492, 31 },
		{ 205, 493, 31 },
		{ 205, 494, 31 },
		{ 205, 495, 31 },
		{ 205, 496, 31 },
		{ 205, 497, 31 },
		{ 205, 498, 31 },
		{ 205, 499, 31 },
		{ 205, 500, 31 },
		{ 205, 501, 31 },
		{ 205, 502, 31 },
		{ 205, 503, 31 },
		{ 205, 504, 31 },
		{ 205, 505, 31 },
		{ 205, 506, 31 },
		{ 205, 507, 31 },
		{ 205, 508, 31 },
		{ 205, 509, 31 },
		{ 205, 510, 31 },
		{ 205, 511, 31 },
		{ 205, 512, 31 },
		{ 206, 462, 31 },
		{ 206, 463, 31 },
		{ 206, 464, 31 },
		{ 206, 465, 31 },
		{ 206, 466, 31 },
		{ 206, 467, 31 },
		{ 206, 468, 31 },
		{ 206, 469, 31 },
		{ 206, 470, 31 },
		{ 206, 471, 31 },
		{ 206, 472, 31 },
		{ 206, 473, 31 },
		{ 206, 474, 31 },
		{ 206, 475, 31 },
		{ 206, 476, 31 },
		{ 206, 477, 31 },
		{ 206, 478, 31 },
		{ 206, 479, 31 },
		{ 206, 480, 31 },
		{ 206, 481, 31 },
		{ 206, 482, 31 },
		{ 206, 483, 31 },
		{ 206, 484, 31 },
		{ 206, 485, 31 },
		{ 206, 486, 31 },
		{ 206, 487, 31 },
		{ 206, 488, 31 },
		{ 206, 489, 31 },
		{ 206, 490, 31 },
		{ 206, 491, 31 },
		{ 206, 492, 31 },
		{ 206, 493, 31 },
		{ 206, 494, 31 },
		{ 206, 495, 31 },
		{ 206, 496, 31 },
		{ 206, 497, 31 },
		{ 206, 498, 31 },
		{ 206, 499, 31 },
		{ 206, 500, 31 },
		{ 206, 501, 31 },
		{ 206, 502, 31 },
		{ 206, 503, 31 },
		{ 206, 504, 31 },
		{ 206, 505, 31 },
		{ 206, 506, 31 },
		{ 206, 507, 31 },
		{ 206, 508, 31 },
		{ 206, 509, 31 },
		{ 206, 510, 31 },
		{ 206, 511, 31 },
		{ 206, 512, 31 },
		{ 207, 458, 31 },
		{ 207, 459, 31 },
		{ 207, 460, 31 },
		{ 207, 461, 31 },
		{ 207, 462, 31 },
		{ 207, 463, 31 },
		{ 207, 464, 31 },
		{ 207, 465, 31 },
		{ 207, 466, 31 },
		{ 207, 467, 31 },
		{ 207, 468, 31 },
		{ 207, 469, 31 },
		{ 207, 470, 31 },
		{ 207, 471, 31 },
		{ 207, 472, 31 },
		{ 207, 473, 31 },
		{ 207, 474, 31 },
		{ 207, 475, 31 },
		{ 207, 476, 31 },
		{ 207, 477, 31 },
		{ 207, 478, 31 },
		{ 207, 479, 31 },
		{ 207, 480, 31 },
		{ 207, 481, 31 },
		{ 207, 482, 31 },
		{ 207, 483, 31 },
		{ 207, 484, 31 },
		{ 207, 485, 31 },
		{ 207, 486, 31 },
		{ 207, 487, 31 },
		{ 207, 488, 31 },
		{ 207, 489, 31 },
		{ 207, 490, 31 },
		{ 207, 491, 31 },
		{ 207, 492, 31 },
		{ 207, 493, 31 },
		{ 207, 494, 31 },
		{ 207, 495, 31 },
		{ 207, 496, 31 },
		{ 207, 497, 31 },
		{ 207, 498, 31 },
		{ 207, 499, 31 },
		{ 207, 500, 31 },
		{ 207, 501, 31 },
		{ 207, 502, 31 },
		{ 207, 503, 31 },
		{ 207, 504, 31 },
		{ 207, 505, 31 },
		{ 207, 506, 31 },
		{ 207, 507, 31 },
		{ 207, 508, 31 },
		{ 207, 509, 31 },
		{ 207, 510, 31 },
		{ 207, 511, 31 },
		{ 207, 512, 31 },
		{ 208, 454, 31 },
		{ 208, 455, 31 },
		{ 208, 456, 31 },
		{ 208, 457, 31 },
		{ 208, 458, 31 },
		{ 208, 459, 31 },
		{ 208, 460, 31 },
		{ 208, 461, 31 },
		{ 208, 462, 31 },
		{ 208, 463, 31 },
		{ 208, 464, 31 },
		{ 208, 465, 31 },
		{ 208, 466, 31 },
		{ 208, 467, 31 },
		{ 208, 468, 31 },
		{ 208, 469, 31 },
		{ 208, 470, 31 },
		{ 208, 471, 31 },
		{ 208, 472, 31 },
		{ 208, 473, 31 },
		{ 208, 474, 31 },
		{ 208, 475, 31 },
		{ 208, 476, 31 },
		{ 208, 477, 31 },
		{ 208, 478, 31 },
		{ 208, 479, 31 },
		{ 208, 480, 31 },
		{ 208, 481, 31 },
		{ 208, 482, 31 },
		{ 208, 483, 31 },
		{ 208, 484, 31 },
		{ 208, 485, 31 },
		{ 208, 486, 31 },
		{ 208, 487, 31 },
		{ 208, 488, 31 },
		{ 208, 489, 31 },
		{ 208, 490, 31 },
		{ 208, 491, 31 },
		{ 208, 492, 31 },
		{ 208, 493, 31 },
		{ 208, 494, 31 },
		{ 208, 495, 31 },
		{ 208, 496, 31 },
		{ 208, 497, 31 },
		{ 208, 498, 31 },
		{ 208, 499, 31 },
		{ 208, 500, 31 },
		{ 208, 501, 31 },
		{ 208, 502, 31 },
		{ 208, 503, 31 },
		{ 208, 504, 31 },
		{ 208, 505, 31 },
		{ 208, 506, 31 },
		{ 208, 507, 31 },
		{ 208, 508, 31 },
		{ 208, 509, 31 },
		{ 208, 510, 31 },
		{ 208, 511, 31 },
		{ 208, 512, 31 },
		{ 209, 450, 31 },
		{ 209, 451, 31 },
		{ 209, 452, 31 },
		{ 209, 453, 31 },
		{ 209, 454, 31 },
		{ 209, 455, 31 },
		{ 209, 456, 31 },
		{ 209, 457, 31 },
		{ 209, 458, 31 },
		{ 209, 459, 31 },
		{ 209, 460, 31 },
		{ 209, 461, 31 },
		{ 209, 462, 31 },
		{ 209, 463, 31 },
		{ 209, 464, 31 },
		{ 209, 465, 31 },
		{ 209, 466, 31 },
		{ 209, 467, 31 },
		{ 209, 468, 31 },
		{ 209, 469, 31 },
		{ 209, 470, 31 },
		{ 209, 471, 31 },
		{ 209, 472, 31 },
		{ 209, 473, 31 },
		{ 209, 474, 31 },
		{ 209, 475, 31 },
		{ 209, 476, 31 },
		{ 209, 477, 31 },
		{ 209, 478, 31 },
		{ 209, 479, 31 },
		{ 209, 480, 31 },
		{ 209, 481, 31 },
		{ 209, 482, 31 },
		{ 209, 483, 31 },
		{ 209, 484, 31 },
		{ 209, 485, 31 },
		{ 209, 486, 31 },
		{ 209, 487, 31 },
		{ 209, 488, 31 },
		{ 209, 489, 31 },
		{ 209, 490, 31 },
		{ 209, 491, 31 },
		{ 209, 492, 31 },
		{ 209, 493, 31 },
		{ 209, 494, 31 },
		{ 209, 495, 31 },
		{ 209, 496, 31 },
		{ 209, 497, 31 },
		{ 209, 498, 31 },
		{ 209, 499, 31 },
		{ 209, 500, 31 },
		{ 209, 501, 31 },
		{ 209, 502, 31 },
		{ 209, 503, 31 },
		{ 209, 504, 31 },
		{ 209, 505, 31 },
		{ 209, 506, 31 },
		{ 209, 507, 31 },
		{ 209, 508, 31 },
		{ 209, 509, 31 },
		{ 209, 510, 31 },
		{ 209, 511, 31 },
		{ 209, 512, 31 },
		{ 210, 446, 31 },
		{ 210, 447, 31 },
		{ 210, 448, 31 },
		{ 210, 449, 31 },
		{ 210, 450, 31 },
		{ 210, 451, 31 },
		{ 210, 452, 31 },
		{ 210, 453, 31 },
		{ 210, 454, 31 },
		{ 210, 455, 31 },
		{ 210, 456, 31 },
		{ 210, 457, 31 },
		{ 210, 458, 31 },
		{ 210, 459, 31 },
		{ 210, 460, 31 },
		{ 210, 461, 31 },
		{ 210, 462, 31 },
		{ 210, 463, 31 },
		{ 210, 464, 31 },
		{ 210, 465, 31 },
		{ 210, 466, 31 },
		{ 210, 467, 31 },
		{ 210, 468, 31 },
		{ 210, 469, 31 },
		{ 210, 470, 31 },
		{ 210, 471, 31 },
		{ 210, 472, 31 },
		{ 210, 473, 31 },
		{ 210, 474, 31 },
		{ 210, 475, 31 },
		{ 210, 476, 31 },
		{ 210, 477, 31 },
		{ 210, 478, 31 },
		{ 210, 479, 31 },
		{ 210, 480, 31 },
		{ 210, 481, 31 },
		{ 210, 482, 31 },
		{ 210, 483, 31 },
		{ 210, 484, 31 },
		{ 210, 485, 31 },
		{ 210, 486, 31 },
		{ 210, 487, 31 },
		{ 210, 488, 31 },
		{ 210, 489, 31 },
		{ 210, 490, 31 },
		{ 210, 491, 31 },
		{ 210, 492, 31 },
		{ 210, 493, 31 },
		{ 210, 494, 31 },
		{ 210, 495, 31 },
		{ 210, 496, 31 },
		{ 210, 497, 31 },
		{ 210, 498, 31 },
		{ 210, 499, 31 },
		{ 210, 500, 31 },
		{ 210, 501, 31 },
		{ 210, 502, 31 },
		{ 210, 503, 31 },
		{ 210, 504, 31 },
		{ 210, 505, 31 },
		{ 210, 506, 31 },
		{ 210, 507, 31 },
		{ 210, 508, 31 },
		{ 210, 509, 31 },
		{ 210, 510, 31 },
		{ 210, 511, 31 },
		{ 210, 512, 31 },
		{ 211, 442, 31 },
		{ 211, 443, 31 },
		{ 211, 444, 31 },
		{ 211, 445, 31 },
		{ 211, 446, 31 },
		{ 211, 447, 31 },
		{ 211, 448, 31 },
		{ 211, 449, 31 },
		{ 211, 450, 31 },
		{ 211, 451, 31 },
		{ 211, 452, 31 },
		{ 211, 453, 31 },
		{ 211, 454, 31 },
		{ 211, 455, 31 },
		{ 211, 456, 31 },
		{ 211, 457, 31 },
		{ 211, 458, 31 },
		{ 211, 459, 31 },
		{ 211, 460, 31 },
		{ 211, 461, 31 },
		{ 211, 462, 31 },
		{ 211, 463, 31 },
		{ 211, 464, 31 },
		{ 211, 465, 31 },
		{ 211, 466, 31 },
		{ 211, 467, 31 },
		{ 211, 468, 31 },
		{ 211, 469, 31 },
		{ 211, 470, 31 },
		{ 211, 471, 31 },
		{ 211, 472, 31 },
		{ 211, 473, 31 },
		{ 211, 474, 31 },
		{ 211, 475, 31 },
		{ 211, 476, 31 },
		{ 211, 477, 31 },
		{ 211, 478, 31 },
		{ 211, 479, 31 },
		{ 211, 480, 31 },
		{ 211, 481, 31 },
		{ 211, 482, 31 },
		{ 211, 483, 31 },
		{ 211, 484, 31 },
		{ 211, 485, 31 },
		{ 211, 486, 31 },
		{ 211, 487, 31 },
		{ 211, 488, 31 },
		{ 211, 489, 31 },
		{ 211, 490, 31 },
		{ 211, 491, 31 },
		{ 211, 492, 31 },
		{ 211, 493, 31 },
		{ 211, 494, 31 },
		{ 211, 495, 31 },
		{ 211, 496, 31 },
		{ 211, 497, 31 },
		{ 211, 498, 31 },
		{ 211, 499, 31 },
		{ 211, 500, 31 },
		{ 211, 501, 31 },
		{ 211, 502, 31 },
		{ 211, 503, 31 },
		{ 211, 504, 31 },
		{ 211, 505, 31 },
		{ 211, 506, 31 },
		{ 211, 507, 31 },
		{ 211, 508, 31 },
		{ 211, 509, 31 },
		{ 211, 510, 31 },
		{ 211, 511, 31 },
		{ 211, 512, 31 },
		{ 212, 438, 31 },
		{ 212, 439, 31 },
		{ 212, 440, 31 },
		{ 212, 441, 31 },
		{ 212, 442, 31 },
		{ 212, 443, 31 },
		{ 212, 444, 31 },
		{ 212, 445, 31 },
		{ 212, 446, 31 },
		{ 212, 447, 31 },
		{ 212, 448, 31 },
		{ 212, 449, 31 },
		{ 212, 450, 31 },
		{ 212, 451, 31 },
		{ 212, 452, 31 },
		{ 212, 453, 31 },
		{ 212, 454, 31 },
		{ 212, 455, 31 },
		{ 212, 456, 31 },
		{ 212, 457, 31 },
		{ 212, 458, 31 },
		{ 212, 459, 31 },
		{ 212, 460, 31 },
		{ 212, 461, 31 },
		{ 212, 462, 31 },
		{ 212, 463, 31 },
		{ 212, 464, 31 },
		{ 212, 465, 31 },
		{ 212, 466, 31 },
		{ 212, 467, 31 },
		{ 212, 468, 31 },
		{ 212, 469, 31 },
		{ 212, 470, 31 },
		{ 212, 471, 31 },
		{ 212, 472, 31 },
		{ 212, 473, 31 },
		{ 212, 474, 31 },
		{ 212, 475, 31 },
		{ 212, 476, 31 },
		{ 212, 477, 31 },
		{ 212, 478, 31 },
		{ 212, 479, 31 },
		{ 212, 480, 31 },
		{ 212, 481, 31 },
		{ 212, 482, 31 },
		{ 212, 483, 31 },
		{ 212, 484, 31 },
		{ 212, 485, 31 },
		{ 212, 486, 31 },
		{ 212, 487, 31 },
		{ 212, 488, 31 },
		{ 212, 489, 31 },
		{ 212, 490, 31 },
		{ 212, 491, 31 },
		{ 212, 492, 31 },
		{ 212, 493, 31 },
		{ 212, 494, 31 },
		{ 212, 495, 31 },
		{ 212, 496, 31 },
		{ 212, 497, 31 },
		{ 212, 498, 31 },
		{ 212, 499, 31 },
		{ 212, 500, 31 },
		{ 212, 501, 31 },
		{ 212, 502, 31 },
		{ 212, 503, 31 },
		{ 212, 504, 31 },
		{ 212, 505, 31 },
		{ 212, 506, 31 },
		{ 212, 507, 31 },
		{ 212, 508, 31 },
		{ 212, 509, 31 },
		{ 212, 510, 31 },
		{ 212, 511, 31 },
		{ 212, 512, 31 },
		{ 213, 434, 31 },
		{ 213, 435, 31 },
		{ 213, 436, 31 },
		{ 213, 437, 31 },
		{ 213, 438, 31 },
		{ 213, 439, 31 },
		{ 213, 440, 31 },
		{ 213, 441, 31 },
		{ 213, 442, 31 },
		{ 213, 443, 31 },
		{ 213, 444, 31 },
		{ 213, 445, 31 },
		{ 213, 446, 31 },
		{ 213, 447, 31 },
		{ 213, 448, 31 },
		{ 213, 449, 31 },
		{ 213, 450, 31 },
		{ 213, 451, 31 },
		{ 213, 452, 31 },
		{ 213, 453, 31 },
		{ 213, 454, 31 },
		{ 213, 455, 31 },
		{ 213, 456, 31 },
		{ 213, 457, 31 },
		{ 213, 458, 31 },
		{ 213, 459, 31 },
		{ 213, 460, 31 },
		{ 213, 461, 31 },
		{ 213, 462, 31 },
		{ 213, 463, 31 },
		{ 213, 464, 31 },
		{ 213, 465, 31 },
		{ 213, 466, 31 },
		{ 213, 467, 31 },
		{ 213, 468, 31 },
		{ 213, 469, 31 },
		{ 213, 470, 31 },
		{ 213, 471, 31 },
		{ 213, 472, 31 },
		{ 213, 473, 31 },
		{ 213, 474, 31 },
		{ 213, 475, 31 },
		{ 213, 476, 31 },
		{ 213, 477, 31 },
		{ 213, 478, 31 },
		{ 213, 479, 31 },
		{ 213, 480, 31 },
		{ 213, 481, 31 },
		{ 213, 482, 31 },
		{ 213, 483, 31 },
		{ 213, 484, 31 },
		{ 213, 485, 31 },
		{ 213, 486, 31 },
		{ 213, 487, 31 },
		{ 213, 488, 31 },
		{ 213, 489, 31 },
		{ 213, 490, 31 },
		{ 213, 491, 31 },
		{ 213, 492, 31 },
		{ 213, 493, 31 },
		{ 213, 494, 31 },
		{ 213, 495, 31 },
		{ 213, 496, 31 },
		{ 213, 497, 31 },
		{ 213, 498, 31 },
		{ 213, 499, 31 },
		{ 213, 500, 31 },
		{ 213, 501, 31 },
		{ 213, 502, 31 },
		{ 213, 503, 31 },
		{ 213, 504, 31 },
		{ 213, 505, 31 },
		{ 213, 506, 31 },
		{ 213, 507, 31 },
		{ 213, 508, 31 },
		{ 213, 509, 31 },
		{ 213, 510, 31 },
		{ 213, 511, 31 },
		{ 213, 512, 31 },
		{ 214, 430, 31 },
		{ 214, 431, 31 },
		{ 214, 432, 31 },
		{ 214, 433, 31 },
		{ 214, 434, 31 },
		{ 214, 435, 31 },
		{ 214, 436, 31 },
		{ 214, 437, 31 },
		{ 214, 438, 31 },
		{ 214, 439, 31 },
		{ 214, 440, 31 },
		{ 214, 441, 31 },
		{ 214, 442, 31 },
		{ 214, 443, 31 },
		{ 214, 444, 31 },
		{ 214, 445, 31 },
		{ 214, 446, 31 },
		{ 214, 447, 31 },
		{ 214, 448, 31 },
		{ 214, 449, 31 },
		{ 214, 450, 31 },
		{ 214, 451, 31 },
		{ 214, 452, 31 },
		{ 214, 453, 31 },
		{ 214, 454, 31 },
		{ 214, 455, 31 },
		{ 214, 456, 31 },
		{ 214, 457, 31 },
		{ 214, 458, 31 },
		{ 214, 459, 31 },
		{ 214, 460, 31 },
		{ 214, 461, 31 },
		{ 214, 462, 31 },
		{ 214, 463, 31 },
		{ 214, 464, 31 },
		{ 214, 465, 31 },
		{ 214, 466, 31 },
		{ 214, 467, 31 },
		{ 214, 468, 31 },
		{ 214, 469, 31 },
		{ 214, 470, 31 },
		{ 214, 471, 31 },
		{ 214, 472, 31 },
		{ 214, 473, 31 },
		{ 214, 474, 31 },
		{ 214, 475, 31 },
		{ 214, 476, 31 },
		{ 214, 477, 31 },
		{ 214, 478, 31 },
		{ 214, 479, 31 },
		{ 214, 480, 31 },
		{ 214, 481, 31 },
		{ 214, 482, 31 },
		{ 214, 483, 31 },
		{ 214, 484, 31 },
		{ 214, 485, 31 },
		{ 214, 486, 31 },
		{ 214, 487, 31 },
		{ 214, 488, 31 },
		{ 214, 489, 31 },
		{ 214, 490, 31 },
		{ 214, 491, 31 },
		{ 214, 492, 31 },
		{ 214, 493, 31 },
		{ 214, 494, 31 },
		{ 214, 495, 31 },
		{ 214, 496, 31 },
		{ 214, 497, 31 },
		{ 214, 498, 31 },
		{ 214, 499, 31 },
		{ 214, 500, 31 },
		{ 214, 501, 31 },
		{ 214, 502, 31 },
		{ 214, 503, 31 },
		{ 214, 504, 31 },
		{ 214, 505, 31 },
		{ 214, 506, 31 },
		{ 214, 507, 31 },
		{ 214, 508, 31 },
		{ 214, 509, 31 },
		{ 214, 510, 31 },
		{ 214, 511, 31 },
		{ 214, 512, 31 },
		{ 215, 426, 31 },
		{ 215, 427, 31 },
		{ 215, 428, 31 },
		{ 215, 429, 31 },
		{ 215, 430, 31 },
		{ 215, 431, 31 },
		{ 215, 432, 31 },
		{ 215, 433, 31 },
		{ 215, 434, 31 },
		{ 215, 435, 31 },
		{ 215, 436, 31 },
		{ 215, 437, 31 },
		{ 215, 438, 31 },
		{ 215, 439, 31 },
		{ 215, 440, 31 },
		{ 215, 441, 31 },
		{ 215, 442, 31 },
		{ 215, 443, 31 },
		{ 215, 444, 31 },
		{ 215, 445, 31 },
		{ 215, 446, 31 },
		{ 215, 447, 31 },
		{ 215, 448, 31 },
		{ 215, 449, 31 },
		{ 215, 450, 31 },
		{ 215, 451, 31 },
		{ 215, 452, 31 },
		{ 215, 453, 31 },
		{ 215, 454, 31 },
		{ 215, 455, 31 },
		{ 215, 456, 31 },
		{ 215, 457, 31 },
		{ 215, 458, 31 },
		{ 215, 459, 31 },
		{ 215, 460, 31 },
		{ 215, 461, 31 },
		{ 215, 462, 31 },
		{ 215, 463, 31 },
		{ 215, 464, 31 },
		{ 215, 465, 31 },
		{ 215, 466, 31 },
		{ 215, 467, 31 },
		{ 215, 468, 31 },
		{ 215, 469, 31 },
		{ 215, 470, 31 },
		{ 215, 471, 31 },
		{ 215, 472, 31 },
		{ 215, 473, 31 },
		{ 215, 474, 31 },
		{ 215, 475, 31 },
		{ 215, 476, 31 },
		{ 215, 477, 31 },
		{ 215, 478, 31 },
		{ 215, 479, 31 },
		{ 215, 480, 31 },
		{ 215, 481, 31 },
		{ 215, 482, 31 },
		{ 215, 483, 31 },
		{ 215, 484, 31 },
		{ 215, 485, 31 },
		{ 215, 486, 31 },
		{ 215, 487, 31 },
		{ 215, 488, 31 },
		{ 215, 489, 31 },
		{ 215, 490, 31 },
		{ 215, 491, 31 },
		{ 215, 492, 31 },
		{ 215, 493, 31 },
		{ 215, 494, 31 },
		{ 215, 495, 31 },
		{ 215, 496, 31 },
		{ 215, 497, 31 },
		{ 215, 498, 31 },
		{ 215, 499, 31 },
		{ 215, 500, 31 },
		{ 215, 501, 31 },
		{ 215, 502, 31 },
		{ 215, 503, 31 },
		{ 215, 504, 31 },
		{ 215, 505, 31 },
		{ 215, 506, 31 },
		{ 215, 507, 31 },
		{ 215, 508, 31 },
		{ 215, 509, 31 },
		{ 215, 510, 31 },
		{ 215, 511, 31 },
		{ 215, 512, 31 },
		{ 216, 422, 31 },
		{ 216, 423, 31 },
		{ 216, 424, 31 },
		{ 216, 425, 31 },
		{ 216, 426, 31 },
		{ 216, 427, 31 },
		{ 216, 428, 31 },
		{ 216, 429, 31 },
		{ 216, 430, 31 },
		{ 216, 431, 31 },
		{ 216, 432, 31 },
		{ 216, 433, 31 },
		{ 216, 434, 31 },
		{ 216, 435, 31 },
		{ 216, 436, 31 },
		{ 216, 437, 31 },
		{ 216, 438, 31 },
		{ 216, 439, 31 },
		{ 216, 440, 31 },
		{ 216, 441, 31 },
		{ 216, 442, 31 },
		{ 216, 443, 31 },
		{ 216, 444, 31 },
		{ 216, 445, 31 },
		{ 216, 446, 31 },
		{ 216, 447, 31 },
		{ 216, 448, 31 },
		{ 216, 449, 31 },
		{ 216, 450, 31 },
		{ 216, 451, 31 },
		{ 216, 452, 31 },
		{ 216, 453, 31 },
		{ 216, 454, 31 },
		{ 216, 455, 31 },
		{ 216, 456, 31 },
		{ 216, 457, 31 },
		{ 216, 458, 31 },
		{ 216, 459, 31 },
		{ 216, 460, 31 },
		{ 216, 461, 31 },
		{ 216, 462, 31 },
		{ 216, 463, 31 },
		{ 216, 464, 31 },
		{ 216, 465, 31 },
		{ 216, 466, 31 },
		{ 216, 467, 31 },
		{ 216, 468, 31 },
		{ 216, 469, 31 },
		{ 216, 470, 31 },
		{ 216, 471, 31 },
		{ 216, 472, 31 },
		{ 216, 473, 31 },
		{ 216, 474, 31 },
		{ 216, 475, 31 },
		{ 216, 476, 31 },
		{ 216, 477, 31 },
		{ 216, 478, 31 },
		{ 216, 479, 31 },
		{ 216, 480, 31 },
		{ 216, 481, 31 },
		{ 216, 482, 31 },
		{ 216, 483, 31 },
		{ 216, 484, 31 },
		{ 216, 485, 31 },
		{ 216, 486, 31 },
		{ 216, 487, 31 },
		{ 216, 488, 31 },
		{ 216, 489, 31 },
		{ 216, 490, 31 },
		{ 216, 491, 31 },
		{ 216, 492, 31 },
		{ 216, 493, 31 },
		{ 216, 494, 31 },
		{ 216, 495, 31 },
		{ 216, 496, 31 },
		{ 216, 497, 31 },
		{ 216, 498, 31 },
		{ 216, 499, 31 },
		{ 216, 500, 31 },
		{ 216, 501, 31 },
		{ 216, 502, 31 },
		{ 216, 503, 31 },
		{ 216, 504, 31 },
		{ 216, 505, 31 },
		{ 216, 506, 31 },
		{ 216, 507, 31 },
		{ 216, 508, 31 },
		{ 216, 509, 31 },
		{ 216, 510, 31 },
		{ 216, 511, 31 },
		{ 216, 512, 31 },
		{ 217, 418, 31 },
		{ 217, 419, 31 },
		{ 217, 420, 31 },
		{ 217, 421, 31 },
		{ 217, 422, 31 },
		{ 217, 423, 31 },
		{ 217, 424, 31 },
		{ 217, 425, 31 },
		{ 217, 426, 31 },
		{ 217, 427, 31 },
		{ 217, 428, 31 },
		{ 217, 429, 31 },
		{ 217, 430, 31 },
		{ 217, 431, 31 },
		{ 217, 432, 31 },
		{ 217, 433, 31 },
		{ 217, 434, 31 },
		{ 217, 435, 31 },
		{ 217, 436, 31 },
		{ 217, 437, 31 },
		{ 217, 438, 31 },
		{ 217, 439, 31 },
		{ 217, 440, 31 },
		{ 217, 441, 31 },
		{ 217, 442, 31 },
		{ 217, 443, 31 },
		{ 217, 444, 31 },
		{ 217, 445, 31 },
		{ 217, 446, 31 },
		{ 217, 447, 31 },
		{ 217, 448, 31 },
		{ 217, 449, 31 },
		{ 217, 450, 31 },
		{ 217, 451, 31 },
		{ 217, 452, 31 },
		{ 217, 453, 31 },
		{ 217, 454, 31 },
		{ 217, 455, 31 },
		{ 217, 456, 31 },
		{ 217, 457, 31 },
		{ 217, 458, 31 },
		{ 217, 459, 31 },
		{ 217, 460, 31 },
		{ 217, 461, 31 },
		{ 217, 462, 31 },
		{ 217, 463, 31 },
		{ 217, 464, 31 },
		{ 217, 465, 31 },
		{ 217, 466, 31 },
		{ 217, 467, 31 },
		{ 217, 468, 31 },
		{ 217, 469, 31 },
		{ 217, 470, 31 },
		{ 217, 471, 31 },
		{ 217, 472, 31 },
		{ 217, 473, 31 },
		{ 217, 474, 31 },
		{ 217, 475, 31 },
		{ 217, 476, 31 },
		{ 217, 477, 31 },
		{ 217, 478, 31 },
		{ 217, 479, 31 },
		{ 217, 480, 31 },
		{ 217, 481, 31 },
		{ 217, 482, 31 },
		{ 217, 483, 31 },
		{ 217, 484, 31 },
		{ 217, 485, 31 },
		{ 217, 486, 31 },
		{ 217, 487, 31 },
		{ 217, 488, 31 },
		{ 217, 489, 31 },
		{ 217, 490, 31 },
		{ 217, 491, 31 },
		{ 217, 492, 31 },
		{ 217, 493, 31 },
		{ 217, 494, 31 },
		{ 217, 495, 31 },
		{ 217, 496, 31 },
		{ 217, 497, 31 },
		{ 217, 498, 31 },
		{ 217, 499, 31 },
		{ 217, 500, 31 },
		{ 217, 501, 31 },
		{ 217, 502, 31 },
		{ 217, 503, 31 },
		{ 217, 504, 31 },
		{ 217, 505, 31 },
		{ 217, 506, 31 },
		{ 217, 507, 31 },
		{ 217, 508, 31 },
		{ 217, 509, 31 },
		{ 217, 510, 31 },
		{ 217, 511, 31 },
		{ 217, 512, 31 },
		{ 218, 414, 31 },
		{ 218, 415, 31 },
		{ 218, 416, 31 },
		{ 218, 417, 31 },
		{ 218, 418, 31 },
		{ 218, 419, 31 },
		{ 218, 420, 31 },
		{ 218, 421, 31 },
		{ 218, 422, 31 },
		{ 218, 423, 31 },
		{ 218, 424, 31 },
		{ 218, 425, 31 },
		{ 218, 426, 31 },
		{ 218, 427, 31 },
		{ 218, 428, 31 },
		{ 218, 429, 31 },
		{ 218, 430, 31 },
		{ 218, 431, 31 },
		{ 218, 432, 31 },
		{ 218, 433, 31 },
		{ 218, 434, 31 },
		{ 218, 435, 31 },
		{ 218, 436, 31 },
		{ 218, 437, 31 },
		{ 218, 438, 31 },
		{ 218, 439, 31 },
		{ 218, 440, 31 },
		{ 218, 441, 31 },
		{ 218, 442, 31 },
		{ 218, 443, 31 },
		{ 218, 444, 31 },
		{ 218, 445, 31 },
		{ 218, 446, 31 },
		{ 218, 447, 31 },
		{ 218, 448, 31 },
		{ 218, 449, 31 },
		{ 218, 450, 31 },
		{ 218, 451, 31 },
		{ 218, 452, 31 },
		{ 218, 453, 31 },
		{ 218, 454, 31 },
		{ 218, 455, 31 },
		{ 218, 456, 31 },
		{ 218, 457, 31 },
		{ 218, 458, 31 },
		{ 218, 459, 31 },
		{ 218, 460, 31 },
		{ 218, 461, 31 },
		{ 218, 462, 31 },
		{ 218, 463, 31 },
		{ 218, 464, 31 },
		{ 218, 465, 31 },
		{ 218, 466, 31 },
		{ 218, 467, 31 },
		{ 218, 468, 31 },
		{ 218, 469, 31 },
		{ 218, 470, 31 },
		{ 218, 471, 31 },
		{ 218, 472, 31 },
		{ 218, 473, 31 },
		{ 218, 474, 31 },
		{ 218, 475, 31 },
		{ 218, 476, 31 },
		{ 218, 477, 31 },
		{ 218, 478, 31 },
		{ 218, 479, 31 },
		{ 218, 480, 31 },
		{ 218, 481, 31 },
		{ 218, 482, 31 },
		{ 218, 483, 31 },
		{ 218, 484, 31 },
		{ 218, 485, 31 },
		{ 218, 486, 31 },
		{ 218, 487, 31 },
		{ 218, 488, 31 },
		{ 218, 489, 31 },
		{ 218, 490, 31 },
		{ 218, 491, 31 },
		{ 218, 492, 31 },
		{ 218, 493, 31 },
		{ 218, 494, 31 },
		{ 218, 495, 31 },
		{ 218, 496, 31 },
		{ 218, 497, 31 },
		{ 218, 498, 31 },
		{ 218, 499, 31 },
		{ 218, 500, 31 },
		{ 218, 501, 31 },
		{ 218, 502, 31 },
		{ 218, 503, 31 },
		{ 218, 504, 31 },
		{ 218, 505, 31 },
		{ 218, 506, 31 },
		{ 218, 507, 31 },
		{ 218, 508, 31 },
		{ 218, 509, 31 },
		{ 218, 510, 31 },
		{ 218, 511, 31 },
		{ 218, 512, 31 },
		{ 219, 410, 31 },
		{ 219, 411, 31 },
		{ 219, 412, 31 },
		{ 219, 413, 31 },
		{ 219, 414, 31 },
		{ 219, 415, 31 },
		{ 219, 416, 31 },
		{ 219, 417, 31 },
		{ 219, 418, 31 },
		{ 219, 419, 31 },
		{ 219, 420, 31 },
		{ 219, 421, 31 },
		{ 219, 422, 31 },
		{ 219, 423, 31 },
		{ 219, 424, 31 },
		{ 219, 425, 31 },
		{ 219, 426, 31 },
		{ 219, 427, 31 },
		{ 219, 428, 31 },
		{ 219, 429, 31 },
		{ 219, 430, 31 },
		{ 219, 431, 31 },
		{ 219, 432, 31 },
		{ 219, 433, 31 },
		{ 219, 434, 31 },
		{ 219, 435, 31 },
		{ 219, 436, 31 },
		{ 219, 437, 31 },
		{ 219, 438, 31 },
		{ 219, 439, 31 },
		{ 219, 440, 31 },
		{ 219, 441, 31 },
		{ 219, 442, 31 },
		{ 219, 443, 31 },
		{ 219, 444, 31 },
		{ 219, 445, 31 },
		{ 219, 446, 31 },
		{ 219, 447, 31 },
		{ 219, 448, 31 },
		{ 219, 449, 31 },
		{ 219, 450, 31 },
		{ 219, 451, 31 },
		{ 219, 452, 31 },
		{ 219, 453, 31 },
		{ 219, 454, 31 },
		{ 219, 455, 31 },
		{ 219, 456, 31 },
		{ 219, 457, 31 },
		{ 219, 458, 31 },
		{ 219, 459, 31 },
		{ 219, 460, 31 },
		{ 219, 461, 31 },
		{ 219, 462, 31 },
		{ 219, 463, 31 },
		{ 219, 464, 31 },
		{ 219, 465, 31 },
		{ 219, 466, 31 },
		{ 219, 467, 31 },
		{ 219, 468, 31 },
		{ 219, 469, 31 },
		{ 219, 470, 31 },
		{ 219, 471, 31 },
		{ 219, 472, 31 },
		{ 219, 473, 31 },
		{ 219, 474, 31 },
		{ 219, 475, 31 },
		{ 219, 476, 31 },
		{ 219, 477, 31 },
		{ 219, 478, 31 },
		{ 219, 479, 31 },
		{ 219, 480, 31 },
		{ 219, 481, 31 },
		{ 219, 482, 31 },
		{ 219, 483, 31 },
		{ 219, 484, 31 },
		{ 219, 485, 31 },
		{ 219, 486, 31 },
		{ 219, 487, 31 },
		{ 219, 488, 31 },
		{ 219, 489, 31 },
		{ 219, 490, 31 },
		{ 219, 491, 31 },
		{ 219, 492, 31 },
		{ 219, 493, 31 },
		{ 219, 494, 31 },
		{ 219, 495, 31 },
		{ 219, 496, 31 },
		{ 219, 497, 31 },
		{ 219, 498, 31 },
		{ 219, 499, 31 },
		{ 219, 500, 31 },
		{ 219, 501, 31 },
		{ 219, 502, 31 },
		{ 219, 503, 31 },
		{ 219, 504, 31 },
		{ 219, 505, 31 },
		{ 219, 506, 31 },
		{ 219, 507, 31 },
		{ 219, 508, 31 },
		{ 219, 509, 31 },
		{ 219, 510, 31 },
		{ 219, 511, 31 },
		{ 219, 512, 31 },
		{ 220, 406, 31 },
		{ 220, 407, 31 },
		{ 220, 408, 31 },
		{ 220, 409, 31 },
		{ 220, 410, 31 },
		{ 220, 411, 31 },
		{ 220, 412, 31 },
		{ 220, 413, 31 },
		{ 220, 414, 31 },
		{ 220, 415, 31 },
		{ 220, 416, 31 },
		{ 220, 417, 31 },
		{ 220, 418, 31 },
		{ 220, 419, 31 },
		{ 220, 420, 31 },
		{ 220, 421, 31 },
		{ 220, 422, 31 },
		{ 220, 423, 31 },
		{ 220, 424, 31 },
		{ 220, 425, 31 },
		{ 220, 426, 31 },
		{ 220, 427, 31 },
		{ 220, 428, 31 },
		{ 220, 429, 31 },
		{ 220, 430, 31 },
		{ 220, 431, 31 },
		{ 220, 432, 31 },
		{ 220, 433, 31 },
		{ 220, 434, 31 },
		{ 220, 435, 31 },
		{ 220, 436, 31 },
		{ 220, 437, 31 },
		{ 220, 438, 31 },
		{ 220, 439, 31 },
		{ 220, 440, 31 },
		{ 220, 441, 31 },
		{ 220, 442, 31 },
		{ 220, 443, 31 },
		{ 220, 444, 31 },
		{ 220, 445, 31 },
		{ 220, 446, 31 },
		{ 220, 447, 31 },
		{ 220, 448, 31 },
		{ 220, 449, 31 },
		{ 220, 450, 31 },
		{ 220, 451, 31 },
		{ 220, 452, 31 },
		{ 220, 453, 31 },
		{ 220, 454, 31 },
		{ 220, 455, 31 },
		{ 220, 456, 31 },
		{ 220, 457, 31 },
		{ 220, 458, 31 },
		{ 220, 459, 31 },
		{ 220, 460, 31 },
		{ 220, 461, 31 },
		{ 220, 462, 31 },
		{ 220, 463, 31 },
		{ 220, 464, 31 },
		{ 220, 465, 31 },
		{ 220, 466, 31 },
		{ 220, 467, 31 },
		{ 220, 468, 31 },
		{ 220, 469, 31 },
		{ 220, 470, 31 },
		{ 220, 471, 31 },
		{ 220, 472, 31 },
		{ 220, 473, 31 },
		{ 220, 474, 31 },
		{ 220, 475, 31 },
		{ 220, 476, 31 },
		{ 220, 477, 31 },
		{ 220, 478, 31 },
		{ 220, 479, 31 },
		{ 220, 480, 31 },
		{ 220, 481, 31 },
		{ 220, 482, 31 },
		{ 220, 483, 31 },
		{ 220, 484, 31 },
		{ 220, 485, 31 },
		{ 220, 486, 31 },
		{ 220, 487, 31 },
		{ 220, 488, 31 },
		{ 220, 489, 31 },
		{ 220, 490, 31 },
		{ 220, 491, 31 },
		{ 220, 492, 31 },
		{ 220, 493, 31 },
		{ 220, 494, 31 },
		{ 220, 495, 31 },
		{ 220, 496, 31 },
		{ 220, 497, 31 },
		{ 220, 498, 31 },
		{ 220, 499, 31 },
		{ 220, 500, 31 },
		{ 220, 501, 31 },
		{ 220, 502, 31 },
		{ 220, 503, 31 },
		{ 220, 504, 31 },
		{ 220, 505, 31 },
		{ 220, 506, 31 },
		{ 220, 507, 31 },
		{ 220, 508, 31 },
		{ 220, 509, 31 },
		{ 220, 510, 31 },
		{ 220, 511, 31 },
		{ 220, 512, 31 },
		{ 221, 402, 31 },
		{ 221, 403, 31 },
		{ 221, 404, 31 },
		{ 221, 405, 31 },
		{ 221, 406, 31 },
		{ 221, 407, 31 },
		{ 221, 408, 31 },
		{ 221, 409, 31 },
		{ 221, 410, 31 },
		{ 221, 411, 31 },
		{ 221, 412, 31 },
		{ 221, 413, 31 },
		{ 221, 414, 31 },
		{ 221, 415, 31 },
		{ 221, 416, 31 },
		{ 221, 417, 31 },
		{ 221, 418, 31 },
		{ 221, 419, 31 },
		{ 221, 420, 31 },
		{ 221, 421, 31 },
		{ 221, 422, 31 },
		{ 221, 423, 31 },
		{ 221, 424, 31 },
		{ 221, 425, 31 },
		{ 221, 426, 31 },
		{ 221, 427, 31 },
		{ 221, 428, 31 },
		{ 221, 429, 31 },
		{ 221, 430, 31 },
		{ 221, 431, 31 },
		{ 221, 432, 31 },
		{ 221, 433, 31 },
		{ 221, 434, 31 },
		{ 221, 435, 31 },
		{ 221, 436, 31 },
		{ 221, 437, 31 },
		{ 221, 438, 31 },
		{ 221, 439, 31 },
		{ 221, 440, 31 },
		{ 221, 441, 31 },
		{ 221, 442, 31 },
		{ 221, 443, 31 },
		{ 221, 444, 31 },
		{ 221, 445, 31 },
		{ 221, 446, 31 },
		{ 221, 447, 31 },
		{ 221, 448, 31 },
		{ 221, 449, 31 },
		{ 221, 450, 31 },
		{ 221, 451, 31 },
		{ 221, 452, 31 },
		{ 221, 453, 31 },
		{ 221, 454, 31 },
		{ 221, 455, 31 },
		{ 221, 456, 31 },
		{ 221, 457, 31 },
		{ 221, 458, 31 },
		{ 221, 459, 31 },
		{ 221, 460, 31 },
		{ 221, 461, 31 },
		{ 221, 462, 31 },
		{ 221, 463, 31 },
		{ 221, 464, 31 },
		{ 221, 465, 31 },
		{ 221, 466, 31 },
		{ 221, 467, 31 },
		{ 221, 468, 31 },
		{ 221, 469, 31 },
		{ 221, 470, 31 },
		{ 221, 471, 31 },
		{ 221, 472, 31 },
		{ 221, 473, 31 },
		{ 221, 474, 31 },
		{ 221, 475, 31 },
		{ 221, 476, 31 },
		{ 221, 477, 31 },
		{ 221, 478, 31 },
		{ 221, 479, 31 },
		{ 221, 480, 31 },
		{ 221, 481, 31 },
		{ 221, 482, 31 },
		{ 221, 483, 31 },
		{ 221, 484, 31 },
		{ 221, 485, 31 },
		{ 221, 486, 31 },
		{ 221, 487, 31 },
		{ 221, 488, 31 },
		{ 221, 489, 31 },
		{ 221, 490, 31 },
		{ 221, 491, 31 },
		{ 221, 492, 31 },
		{ 221, 493, 31 },
		{ 221, 494, 31 },
		{ 221, 495, 31 },
		{ 221, 496, 31 },
		{ 221, 497, 31 },
		{ 221, 498, 31 },
		{ 221, 499, 31 },
		{ 221, 500, 31 },
		{ 221, 501, 31 },
		{ 221, 502, 31 },
		{ 221, 503, 31 },
		{ 221, 504, 31 },
		{ 221, 505, 31 },
		{ 221, 506, 31 },
		{ 221, 507, 31 },
		{ 221, 508, 31 },
		{ 221, 509, 31 },
		{ 221, 510, 31 },
		{ 221, 511, 31 },
		{ 221, 512, 31 },
		{ 222, 398, 31 },
		{ 222, 399, 31 },
		{ 222, 400, 31 },
		{ 222, 401, 31 },
		{ 222, 402, 31 },
		{ 222, 403, 31 },
		{ 222, 404, 31 },
		{ 222, 405, 31 },
		{ 222, 406, 31 },
		{ 222, 407, 31 },
		{ 222, 408, 31 },
		{ 222, 409, 31 },
		{ 222, 410, 31 },
		{ 222, 411, 31 },
		{ 222, 412, 31 },
		{ 222, 413, 31 },
		{ 222, 414, 31 },
		{ 222, 415, 31 },
		{ 222, 416, 31 },
		{ 222, 417, 31 },
		{ 222, 418, 31 },
		{ 222, 419, 31 },
		{ 222, 420, 31 },
		{ 222, 421, 31 },
		{ 222, 422, 31 },
		{ 222, 423, 31 },
		{ 222, 424, 31 },
		{ 222, 425, 31 },
		{ 222, 426, 31 },
		{ 222, 427, 31 },
		{ 222, 428, 31 },
		{ 222, 429, 31 },
		{ 222, 430, 31 },
		{ 222, 431, 31 },
		{ 222, 432, 31 },
		{ 222, 433, 31 },
		{ 222, 434, 31 },
		{ 222, 435, 31 },
		{ 222, 436, 31 },
		{ 222, 437, 31 },
		{ 222, 438, 31 },
		{ 222, 439, 31 },
		{ 222, 440, 31 },
		{ 222, 441, 31 },
		{ 222, 442, 31 },
		{ 222, 443, 31 },
		{ 222, 444, 31 },
		{ 222, 445, 31 },
		{ 222, 446, 31 },
		{ 222, 447, 31 },
		{ 222, 448, 31 },
		{ 222, 449, 31 },
		{ 222, 450, 31 },
		{ 222, 451, 31 },
		{ 222, 452, 31 },
		{ 222, 453, 31 },
		{ 222, 454, 31 },
		{ 222, 455, 31 },
		{ 222, 456, 31 },
		{ 222, 457, 31 },
		{ 222, 458, 31 },
		{ 222, 459, 31 },
		{ 222, 460, 31 },
		{ 222, 461, 31 },
		{ 222, 462, 31 },
		{ 222, 463, 31 },
		{ 222, 464, 31 },
		{ 222, 465, 31 },
		{ 222, 466, 31 },
		{ 222, 467, 31 },
		{ 222, 468, 31 },
		{ 222, 469, 31 },
		{ 222, 470, 31 },
		{ 222, 471, 31 },
		{ 222, 472, 31 },
		{ 222, 473, 31 },
		{ 222, 474, 31 },
		{ 222, 475, 31 },
		{ 222, 476, 31 },
		{ 222, 477, 31 },
		{ 222, 478, 31 },
		{ 222, 479, 31 },
		{ 222, 480, 31 },
		{ 222, 481, 31 },
		{ 222, 482, 31 },
		{ 222, 483, 31 },
		{ 222, 484, 31 },
		{ 222, 485, 31 },
		{ 222, 486, 31 },
		{ 222, 487, 31 },
		{ 222, 488, 31 },
		{ 222, 489, 31 },
		{ 222, 490, 31 },
		{ 222, 491, 31 },
		{ 222, 492, 31 },
		{ 222, 493, 31 },
		{ 222, 494, 31 },
		{ 222, 495, 31 },
		{ 222, 496, 31 },
		{ 222, 497, 31 },
		{ 222, 498, 31 },
		{ 222, 499, 31 },
		{ 222, 500, 31 },
		{ 222, 501, 31 },
		{ 222, 502, 31 },
		{ 222, 503, 31 },
		{ 222, 504, 31 },
		{ 222, 505, 31 },
		{ 222, 506, 31 },
		{ 222, 507, 31 },
		{ 222, 508, 31 },
		{ 222, 509, 31 },
		{ 222, 510, 31 },
		{ 222, 511, 31 },
		{ 222, 512, 31 },
		{ 223, 394, 31 },
		{ 223, 395, 31 },
		{ 223, 396, 31 },
		{ 223, 397, 31 },
		{ 223, 398, 31 },
		{ 223, 399, 31 },
		{ 223, 400, 31 },
		{ 223, 401, 31 },
		{ 223, 402, 31 },
		{ 223, 403, 31 },
		{ 223, 404, 31 },
		{ 223, 405, 31 },
		{ 223, 406, 31 },
		{ 223, 407, 31 },
		{ 223, 408, 31 },
		{ 223, 409, 31 },
		{ 223, 410, 31 },
		{ 223, 411, 31 },
		{ 223, 412, 31 },
		{ 223, 413, 31 },
		{ 223, 414, 31 },
		{ 223, 415, 31 },
		{ 223, 416, 31 },
		{ 223, 417, 31 },
		{ 223, 418, 31 },
		{ 223, 419, 31 },
		{ 223, 420, 31 },
		{ 223, 421, 31 },
		{ 223, 422, 31 },
		{ 223, 423, 31 },
		{ 223, 424, 31 },
		{ 223, 425, 31 },
		{ 223, 426, 31 },
		{ 223, 427, 31 },
		{ 223, 428, 31 },
		{ 223, 429, 31 },
		{ 223, 430, 31 },
		{ 223, 431, 31 },
		{ 223, 432, 31 },
		{ 223, 433, 31 },
		{ 223, 434, 31 },
		{ 223, 435, 31 },
		{ 223, 436, 31 },
		{ 223, 437, 31 },
		{ 223, 438, 31 },
		{ 223, 439, 31 },
		{ 223, 440, 31 },
		{ 223, 441, 31 },
		{ 223, 442, 31 },
		{ 223, 443, 31 },
		{ 223, 444, 31 },
		{ 223, 445, 31 },
		{ 223, 446, 31 },
		{ 223, 447, 31 },
		{ 223, 448, 31 },
		{ 223, 449, 31 },
		{ 223, 450, 31 },
		{ 223, 451, 31 },
		{ 223, 452, 31 },
		{ 223, 453, 31 },
		{ 223, 454, 31 },
		{ 223, 455, 31 },
		{ 223, 456, 31 },
		{ 223, 457, 31 },
		{ 223, 458, 31 },
		{ 223, 459, 31 },
		{ 223, 460, 31 },
		{ 223, 461, 31 },
		{ 223, 462, 31 },
		{ 223, 463, 31 },
		{ 223, 464, 31 },
		{ 223, 465, 31 },
		{ 223, 466, 31 },
		{ 223, 467, 31 },
		{ 223, 468, 31 },
		{ 223, 469, 31 },
		{ 223, 470, 31 },
		{ 223, 471, 31 },
		{ 223, 472, 31 },
		{ 223, 473, 31 },
		{ 223, 474, 31 },
		{ 223, 475, 31 },
		{ 223, 476, 31 },
		{ 223, 477, 31 },
		{ 223, 478, 31 },
		{ 223, 479, 31 },
		{ 223, 480, 31 },
		{ 223, 481, 31 },
		{ 223, 482, 31 },
		{ 223, 483, 31 },
		{ 223, 484, 31 },
		{ 223, 485, 31 },
		{ 223, 486, 31 },
		{ 223, 487, 31 },
		{ 223, 488, 31 },
		{ 223, 489, 31 },
		{ 223, 490, 31 },
		{ 223, 491, 31 },
		{ 223, 492, 31 },
		{ 223, 493, 31 },
		{ 223, 494, 31 },
		{ 223, 495, 31 },
		{ 223, 496, 31 },
		{ 223, 497, 31 },
		{ 223, 498, 31 },
		{ 223, 499, 31 },
		{ 223, 500, 31 },
		{ 223, 501, 31 },
		{ 223, 502, 31 },
		{ 223, 503, 31 },
		{ 223, 504, 31 },
		{ 223, 505, 31 },
		{ 223, 506, 31 },
		{ 223, 507, 31 },
		{ 223, 508, 31 },
		{ 223, 509, 31 },
		{ 223, 510, 31 },
		{ 223, 511, 31 },
		{ 223, 512, 31 },
		{ 224, 390, 31 },
		{ 224, 391, 31 },
		{ 224, 392, 31 },
		{ 224, 393, 31 },
		{ 224, 394, 31 },
		{ 224, 395, 31 },
		{ 224, 396, 31 },
		{ 224, 397, 31 },
		{ 224, 398, 31 },
		{ 224, 399, 31 },
		{ 224, 400, 31 },
		{ 224, 401, 31 },
		{ 224, 402, 31 },
		{ 224, 403, 31 },
		{ 224, 404, 31 },
		{ 224, 405, 31 },
		{ 224, 406, 31 },
		{ 224, 407, 31 },
		{ 224, 408, 31 },
		{ 224, 409, 31 },
		{ 224, 410, 31 },
		{ 224, 411, 31 },
		{ 224, 412, 31 },
		{ 224, 413, 31 },
		{ 224, 414, 31 },
		{ 224, 415, 31 },
		{ 224, 416, 31 },
		{ 224, 417, 31 },
		{ 224, 418, 31 },
		{ 224, 419, 31 },
		{ 224, 420, 31 },
		{ 224, 421, 31 },
		{ 224, 422, 31 },
		{ 224, 423, 31 },
		{ 224, 424, 31 },
		{ 224, 425, 31 },
		{ 224, 426, 31 },
		{ 224, 427, 31 },
		{ 224, 428, 31 },
		{ 224, 429, 31 },
		{ 224, 430, 31 },
		{ 224, 431, 31 },
		{ 224, 432, 31 },
		{ 224, 433, 31 },
		{ 224, 434, 31 },
		{ 224, 435, 31 },
		{ 224, 436, 31 },
		{ 224, 437, 31 },
		{ 224, 438, 31 },
		{ 224, 439, 31 },
		{ 224, 440, 31 },
		{ 224, 441, 31 },
		{ 224, 442, 31 },
		{ 224, 443, 31 },
		{ 224, 444, 31 },
		{ 224, 445, 31 },
		{ 224, 446, 31 },
		{ 224, 447, 31 },
		{ 224, 448, 31 },
		{ 224, 449, 31 },
		{ 224, 450, 31 },
		{ 224, 451, 31 },
		{ 224, 452, 31 },
		{ 224, 453, 31 },
		{ 224, 454, 31 },
		{ 224, 455, 31 },
		{ 224, 456, 31 },
		{ 224, 457, 31 },
		{ 224, 458, 31 },
		{ 224, 459, 31 },
		{ 224, 460, 31 },
		{ 224, 461, 31 },
		{ 224, 462, 31 },
		{ 224, 463, 31 },
		{ 224, 464, 31 },
		{ 224, 465, 31 },
		{ 224, 466, 31 },
		{ 224, 467, 31 },
		{ 224, 468, 31 },
		{ 224, 469, 31 },
		{ 224, 470, 31 },
		{ 224, 471, 31 },
		{ 224, 472, 31 },
		{ 224, 473, 31 },
		{ 224, 474, 31 },
		{ 224, 475, 31 },
		{ 224, 476, 31 },
		{ 224, 477, 31 },
		{ 224, 478, 31 },
		{ 224, 479, 31 },
		{ 224, 480, 31 },
		{ 224, 481, 31 },
		{ 224, 482, 31 },
		{ 224, 483, 31 },
		{ 224, 484, 31 },
		{ 224, 485, 31 },
		{ 224, 486, 31 },
		{ 224, 487, 31 },
		{ 224, 488, 31 },
		{ 224, 489, 31 },
		{ 224, 490, 31 },
		{ 224, 491, 31 },
		{ 224, 492, 31 },
		{ 224, 493, 31 },
		{ 224, 494, 31 },
		{ 224, 495, 31 },
		{ 224, 496, 31 },
		{ 224, 497, 31 },
		{ 224, 498, 31 },
		{ 224, 499, 31 },
		{ 224, 500, 31 },
		{ 224, 501, 31 },
		{ 224, 502, 31 },
		{ 224, 503, 31 },
		{ 224, 504, 31 },
		{ 224, 505, 31 },
		{ 224, 506, 31 },
		{ 224, 507, 31 },
		{ 224, 508, 31 },
		{ 224, 509, 31 },
		{ 224, 510, 31 },
		{ 224, 511, 31 },
		{ 224, 512, 31 },
		{ 225, 386, 31 },
		{ 225, 387, 31 },
		{ 225, 388, 31 },
		{ 225, 389, 31 },
		{ 225, 390, 31 },
		{ 225, 391, 31 },
		{ 225, 392, 31 },
		{ 225, 393, 31 },
		{ 225, 394, 31 },
		{ 225, 395, 31 },
		{ 225, 396, 31 },
		{ 225, 397, 31 },
		{ 225, 398, 31 },
		{ 225, 399, 31 },
		{ 225, 400, 31 },
		{ 225, 401, 31 },
		{ 225, 402, 31 },
		{ 225, 403, 31 },
		{ 225, 404, 31 },
		{ 225, 405, 31 },
		{ 225, 406, 31 },
		{ 225, 407, 31 },
		{ 225, 408, 31 },
		{ 225, 409, 31 },
		{ 225, 410, 31 },
		{ 225, 411, 31 },
		{ 225, 412, 31 },
		{ 225, 413, 31 },
		{ 225, 414, 31 },
		{ 225, 415, 31 },
		{ 225, 416, 31 },
		{ 225, 417, 31 },
		{ 225, 418, 31 },
		{ 225, 419, 31 },
		{ 225, 420, 31 },
		{ 225, 421, 31 },
		{ 225, 422, 31 },
		{ 225, 423, 31 },
		{ 225, 424, 31 },
		{ 225, 425, 31 },
		{ 225, 426, 31 },
		{ 225, 427, 31 },
		{ 225, 428, 31 },
		{ 225, 429, 31 },
		{ 225, 430, 31 },
		{ 225, 431, 31 },
		{ 225, 432, 31 },
		{ 225, 433, 31 },
		{ 225, 434, 31 },
		{ 225, 435, 31 },
		{ 225, 436, 31 },
		{ 225, 437, 31 },
		{ 225, 438, 31 },
		{ 225, 439, 31 },
		{ 225, 440, 31 },
		{ 225, 441, 31 },
		{ 225, 442, 31 },
		{ 225, 443, 31 },
		{ 225, 444, 31 },
		{ 225, 445, 31 },
		{ 225, 446, 31 },
		{ 225, 447, 31 },
		{ 225, 448, 31 },
		{ 225, 449, 31 },
		{ 225, 450, 31 },
		{ 225, 451, 31 },
		{ 225, 452, 31 },
		{ 225, 453, 31 },
		{ 225, 454, 31 },
		{ 225, 455, 31 },
		{ 225, 456, 31 },
		{ 225, 457, 31 },
		{ 225, 458, 31 },
		{ 225, 459, 31 },
		{ 225, 460, 31 },
		{ 225, 461, 31 },
		{ 225, 462, 31 },
		{ 225, 463, 31 },
		{ 225, 464, 31 },
		{ 225, 465, 31 },
		{ 225, 466, 31 },
		{ 225, 467, 31 },
		{ 225, 468, 31 },
		{ 225, 469, 31 },
		{ 225, 470, 31 },
		{ 225, 471, 31 },
		{ 225, 472, 31 },
		{ 225, 473, 31 },
		{ 225, 474, 31 },
		{ 225, 475, 31 },
		{ 225, 476, 31 },
		{ 225, 477, 31 },
		{ 225, 478, 31 },
		{ 225, 479, 31 },
		{ 225, 480, 31 },
		{ 225, 481, 31 },
		{ 225, 482, 31 },
		{ 225, 483, 31 },
		{ 225, 484, 31 },
		{ 225, 485, 31 },
		{ 225, 486, 31 },
		{ 225, 487, 31 },
		{ 225, 488, 31 },
		{ 225, 489, 31 },
		{ 225, 490, 31 },
		{ 225, 491, 31 },
		{ 225, 492, 31 },
		{ 225, 493, 31 },
		{ 225, 494, 31 },
		{ 225, 495, 31 },
		{ 225, 496, 31 },
		{ 225, 497, 31 },
		{ 225, 498, 31 },
		{ 225, 499, 31 },
		{ 225, 500, 31 },
		{ 225, 501, 31 },
		{ 225, 502, 31 },
		{ 225, 503, 31 },
		{ 225, 504, 31 },
		{ 225, 505, 31 },
		{ 225, 506, 31 },
		{ 225, 507, 31 },
		{ 225, 508, 31 },
		{ 225, 509, 31 },
		{ 225, 510, 31 },
		{ 225, 511, 31 },
		{ 225, 512, 31 },
		{ 226, 382, 31 },
		{ 226, 383, 31 },
		{ 226, 384, 31 },
		{ 226, 385, 31 },
		{ 226, 386, 31 },
		{ 226, 387, 31 },
		{ 226, 388, 31 },
		{ 226, 389, 31 },
		{ 226, 390, 31 },
		{ 226, 391, 31 },
		{ 226, 392, 31 },
		{ 226, 393, 31 },
		{ 226, 394, 31 },
		{ 226, 395, 31 },
		{ 226, 396, 31 },
		{ 226, 397, 31 },
		{ 226, 398, 31 },
		{ 226, 399, 31 },
		{ 226, 400, 31 },
		{ 226, 401, 31 },
		{ 226, 402, 31 },
		{ 226, 403, 31 },
		{ 226, 404, 31 },
		{ 226, 405, 31 },
		{ 226, 406, 31 },
		{ 226, 407, 31 },
		{ 226, 408, 31 },
		{ 226, 409, 31 },
		{ 226, 410, 31 },
		{ 226, 411, 31 },
		{ 226, 412, 31 },
		{ 226, 413, 31 },
		{ 226, 414, 31 },
		{ 226, 415, 31 },
		{ 226, 416, 31 },
		{ 226, 417, 31 },
		{ 226, 418, 31 },
		{ 226, 419, 31 },
		{ 226, 420, 31 },
		{ 226, 421, 31 },
		{ 226, 422, 31 },
		{ 226, 423, 31 },
		{ 226, 424, 31 },
		{ 226, 425, 31 },
		{ 226, 426, 31 },
		{ 226, 427, 31 },
		{ 226, 428, 31 },
		{ 226, 429, 31 },
		{ 226, 430, 31 },
		{ 226, 431, 31 },
		{ 226, 432, 31 },
		{ 226, 433, 31 },
		{ 226, 434, 31 },
		{ 226, 435, 31 },
		{ 226, 436, 31 },
		{ 226, 437, 31 },
		{ 226, 438, 31 },
		{ 226, 439, 31 },
		{ 226, 440, 31 },
		{ 226, 441, 31 },
		{ 226, 442, 31 },
		{ 226, 443, 31 },
		{ 226, 444, 31 },
		{ 226, 445, 31 },
		{ 226, 446, 31 },
		{ 226, 447, 31 },
		{ 226, 448, 31 },
		{ 226, 449, 31 },
		{ 226, 450, 31 },
		{ 226, 451, 31 },
		{ 226, 452, 31 },
		{ 226, 453, 31 },
		{ 226, 454, 31 },
		{ 226, 455, 31 },
		{ 226, 456, 31 },
		{ 226, 457, 31 },
		{ 226, 458, 31 },
		{ 226, 459, 31 },
		{ 226, 460, 31 },
		{ 226, 461, 31 },
		{ 226, 462, 31 },
		{ 226, 463, 31 },
		{ 226, 464, 31 },
		{ 226, 465, 31 },
		{ 226, 466, 31 },
		{ 226, 467, 31 },
		{ 226, 468, 31 },
		{ 226, 469, 31 },
		{ 226, 470, 31 },
		{ 226, 471, 31 },
		{ 226, 472, 31 },
		{ 226, 473, 31 },
		{ 226, 474, 31 },
		{ 226, 475, 31 },
		{ 226, 476, 31 },
		{ 226, 477, 31 },
		{ 226, 478, 31 },
		{ 226, 479, 31 },
		{ 226, 480, 31 },
		{ 226, 481, 31 },
		{ 226, 482, 31 },
		{ 226, 483, 31 },
		{ 226, 484, 31 },
		{ 226, 485, 31 },
		{ 226, 486, 31 },
		{ 226, 487, 31 },
		{ 226, 488, 31 },
		{ 226, 489, 31 },
		{ 226, 490, 31 },
		{ 226, 491, 31 },
		{ 226, 492, 31 },
		{ 226, 493, 31 },
		{ 226, 494, 31 },
		{ 226, 495, 31 },
		{ 226, 496, 31 },
		{ 226, 497, 31 },
		{ 226, 498, 31 },
		{ 226, 499, 31 },
		{ 226, 500, 31 },
		{ 226, 501, 31 },
		{ 226, 502, 31 },
		{ 226, 503, 31 },
		{ 226, 504, 31 },
		{ 226, 505, 31 },
		{ 226, 506, 31 },
		{ 226, 507, 31 },
		{ 226, 508, 31 },
		{ 226, 509, 31 },
		{ 226, 510, 31 },
		{ 226, 511, 31 },
		{ 226, 512, 31 },
		{ 227, 378, 31 },
		{ 227, 379, 31 },
		{ 227, 380, 31 },
		{ 227, 381, 31 },
		{ 227, 382, 31 },
		{ 227, 383, 31 },
		{ 227, 384, 31 },
		{ 227, 385, 31 },
		{ 227, 386, 31 },
		{ 227, 387, 31 },
		{ 227, 388, 31 },
		{ 227, 389, 31 },
		{ 227, 390, 31 },
		{ 227, 391, 31 },
		{ 227, 392, 31 },
		{ 227, 393, 31 },
		{ 227, 394, 31 },
		{ 227, 395, 31 },
		{ 227, 396, 31 },
		{ 227, 397, 31 },
		{ 227, 398, 31 },
		{ 227, 399, 31 },
		{ 227, 400, 31 },
		{ 227, 401, 31 },
		{ 227, 402, 31 },
		{ 227, 403, 31 },
		{ 227, 404, 31 },
		{ 227, 405, 31 },
		{ 227, 406, 31 },
		{ 227, 407, 31 },
		{ 227, 408, 31 },
		{ 227, 409, 31 },
		{ 227, 410, 31 },
		{ 227, 411, 31 },
		{ 227, 412, 31 },
		{ 227, 413, 31 },
		{ 227, 414, 31 },
		{ 227, 415, 31 },
		{ 227, 416, 31 },
		{ 227, 417, 31 },
		{ 227, 418, 31 },
		{ 227, 419, 31 },
		{ 227, 420, 31 },
		{ 227, 421, 31 },
		{ 227, 422, 31 },
		{ 227, 423, 31 },
		{ 227, 424, 31 },
		{ 227, 425, 31 },
		{ 227, 426, 31 },
		{ 227, 427, 31 },
		{ 227, 428, 31 },
		{ 227, 429, 31 },
		{ 227, 430, 31 },
		{ 227, 431, 31 },
		{ 227, 432, 31 },
		{ 227, 433, 31 },
		{ 227, 434, 31 },
		{ 227, 435, 31 },
		{ 227, 436, 31 },
		{ 227, 437, 31 },
		{ 227, 438, 31 },
		{ 227, 439, 31 },
		{ 227, 440, 31 },
		{ 227, 441, 31 },
		{ 227, 442, 31 },
		{ 227, 443, 31 },
		{ 227, 444, 31 },
		{ 227, 445, 31 },
		{ 227, 446, 31 },
		{ 227, 447, 31 },
		{ 227, 448, 31 },
		{ 227, 449, 31 },
		{ 227, 450, 31 },
		{ 227, 451, 31 },
		{ 227, 452, 31 },
		{ 227, 453, 31 },
		{ 227, 454, 31 },
		{ 227, 455, 31 },
		{ 227, 456, 31 },
		{ 227, 457, 31 },
		{ 227, 458, 31 },
		{ 227, 459, 31 },
		{ 227, 460, 31 },
		{ 227, 461, 31 },
		{ 227, 462, 31 },
		{ 227, 463, 31 },
		{ 227, 464, 31 },
		{ 227, 465, 31 },
		{ 227, 466, 31 },
		{ 227, 467, 31 },
		{ 227, 468, 31 },
		{ 227, 469, 31 },
		{ 227, 470, 31 },
		{ 227, 471, 31 },
		{ 227, 472, 31 },
		{ 227, 473, 31 },
		{ 227, 474, 31 },
		{ 227, 475, 31 },
		{ 227, 476, 31 },
		{ 227, 477, 31 },
		{ 227, 478, 31 },
		{ 227, 479, 31 },
		{ 227, 480, 31 },
		{ 227, 481, 31 },
		{ 227, 482, 31 },
		{ 227, 483, 31 },
		{ 227, 484, 31 },
		{ 227, 485, 31 },
		{ 227, 486, 31 },
		{ 227, 487, 31 },
		{ 227, 488, 31 },
		{ 227, 489, 31 },
		{ 227, 490, 31 },
		{ 227, 491, 31 },
		{ 227, 492, 31 },
		{ 227, 493, 31 },
		{ 227, 494, 31 },
		{ 227, 495, 31 },
		{ 227, 496, 31 },
		{ 227, 497, 31 },
		{ 227, 498, 31 },
		{ 227, 499, 31 },
		{ 227, 500, 31 },
		{ 227, 501, 31 },
		{ 227, 502, 31 },
		{ 227, 503, 31 },
		{ 227, 504, 31 },
		{ 227, 505, 31 },
		{ 227, 506, 31 },
		{ 227, 507, 31 },
		{ 227, 508, 31 },
		{ 227, 509, 31 },
		{ 227, 510, 31 },
		{ 227, 511, 31 },
		{ 227, 512, 31 },
		{ 228, 374, 31 },
		{ 228, 375, 31 },
		{ 228, 376, 31 },
		{ 228, 377, 31 },
		{ 228, 378, 31 },
		{ 228, 379, 31 },
		{ 228, 380, 31 },
		{ 228, 381, 31 },
		{ 228, 382, 31 },
		{ 228, 383, 31 },
		{ 228, 384, 31 },
		{ 228, 385, 31 },
		{ 228, 386, 31 },
		{ 228, 387, 31 },
		{ 228, 388, 31 },
		{ 228, 389, 31 },
		{ 228, 390, 31 },
		{ 228, 391, 31 },
		{ 228, 392, 31 },
		{ 228, 393, 31 },
		{ 228, 394, 31 },
		{ 228, 395, 31 },
		{ 228, 396, 31 },
		{ 228, 397, 31 },
		{ 228, 398, 31 },
		{ 228, 399, 31 },
		{ 228, 400, 31 },
		{ 228, 401, 31 },
		{ 228, 402, 31 },
		{ 228, 403, 31 },
		{ 228, 404, 31 },
		{ 228, 405, 31 },
		{ 228, 406, 31 },
		{ 228, 407, 31 },
		{ 228, 408, 31 },
		{ 228, 409, 31 },
		{ 228, 410, 31 },
		{ 228, 411, 31 },
		{ 228, 412, 31 },
		{ 228, 413, 31 },
		{ 228, 414, 31 },
		{ 228, 415, 31 },
		{ 228, 416, 31 },
		{ 228, 417, 31 },
		{ 228, 418, 31 },
		{ 228, 419, 31 },
		{ 228, 420, 31 },
		{ 228, 421, 31 },
		{ 228, 422, 31 },
		{ 228, 423, 31 },
		{ 228, 424, 31 },
		{ 228, 425, 31 },
		{ 228, 426, 31 },
		{ 228, 427, 31 },
		{ 228, 428, 31 },
		{ 228, 429, 31 },
		{ 228, 430, 31 },
		{ 228, 431, 31 },
		{ 228, 432, 31 },
		{ 228, 433, 31 },
		{ 228, 434, 31 },
		{ 228, 435, 31 },
		{ 228, 436, 31 },
		{ 228, 437, 31 },
		{ 228, 438, 31 },
		{ 228, 439, 31 },
		{ 228, 440, 31 },
		{ 228, 441, 31 },
		{ 228, 442, 31 },
		{ 228, 443, 31 },
		{ 228, 444, 31 },
		{ 228, 445, 31 },
		{ 228, 446, 31 },
		{ 228, 447, 31 },
		{ 228, 448, 31 },
		{ 228, 449, 31 },
		{ 228, 450, 31 },
		{ 228, 451, 31 },
		{ 228, 452, 31 },
		{ 228, 453, 31 },
		{ 228, 454, 31 },
		{ 228, 455, 31 },
		{ 228, 456, 31 },
		{ 228, 457, 31 },
		{ 228, 458, 31 },
		{ 228, 459, 31 },
		{ 228, 460, 31 },
		{ 228, 461, 31 },
		{ 228, 462, 31 },
		{ 228, 463, 31 },
		{ 228, 464, 31 },
		{ 228, 465, 31 },
		{ 228, 466, 31 },
		{ 228, 467, 31 },
		{ 228, 468, 31 },
		{ 228, 469, 31 },
		{ 228, 470, 31 },
		{ 228, 471, 31 },
		{ 228, 472, 31 },
		{ 228, 473, 31 },
		{ 228, 474, 31 },
		{ 228, 475, 31 },
		{ 228, 476, 31 },
		{ 228, 477, 31 },
		{ 228, 478, 31 },
		{ 228, 479, 31 },
		{ 228, 480, 31 },
		{ 228, 481, 31 },
		{ 228, 482, 31 },
		{ 228, 483, 31 },
		{ 228, 484, 31 },
		{ 228, 485, 31 },
		{ 228, 486, 31 },
		{ 228, 487, 31 },
		{ 228, 488, 31 },
		{ 228, 489, 31 },
		{ 228, 490, 31 },
		{ 228, 491, 31 },
		{ 228, 492, 31 },
		{ 228, 493, 31 },
		{ 228, 494, 31 },
		{ 228, 495, 31 },
		{ 228, 496, 31 },
		{ 228, 497, 31 },
		{ 228, 498, 31 },
		{ 228, 499, 31 },
		{ 228, 500, 31 },
		{ 228, 501, 31 },
		{ 228, 502, 31 },
		{ 228, 503, 31 },
		{ 228, 504, 31 },
		{ 228, 505, 31 },
		{ 228, 506, 31 },
		{ 228, 507, 31 },
		{ 228, 508, 31 },
		{ 228, 509, 31 },
		{ 228, 510, 31 },
		{ 228, 511, 31 },
		{ 228, 512, 31 },
		{ 229, 370, 31 },
		{ 229, 371, 31 },
		{ 229, 372, 31 },
		{ 229, 373, 31 },
		{ 229, 374, 31 },
		{ 229, 375, 31 },
		{ 229, 376, 31 },
		{ 229, 377, 31 },
		{ 229, 378, 31 },
		{ 229, 379, 31 },
		{ 229, 380, 31 },
		{ 229, 381, 31 },
		{ 229, 382, 31 },
		{ 229, 383, 31 },
		{ 229, 384, 31 },
		{ 229, 385, 31 },
		{ 229, 386, 31 },
		{ 229, 387, 31 },
		{ 229, 388, 31 },
		{ 229, 389, 31 },
		{ 229, 390, 31 },
		{ 229, 391, 31 },
		{ 229, 392, 31 },
		{ 229, 393, 31 },
		{ 229, 394, 31 },
		{ 229, 395, 31 },
		{ 229, 396, 31 },
		{ 229, 397, 31 },
		{ 229, 398, 31 },
		{ 229, 399, 31 },
		{ 229, 400, 31 },
		{ 229, 401, 31 },
		{ 229, 402, 31 },
		{ 229, 403, 31 },
		{ 229, 404, 31 },
		{ 229, 405, 31 },
		{ 229, 406, 31 },
		{ 229, 407, 31 },
		{ 229, 408, 31 },
		{ 229, 409, 31 },
		{ 229, 410, 31 },
		{ 229, 411, 31 },
		{ 229, 412, 31 },
		{ 229, 413, 31 },
		{ 229, 414, 31 },
		{ 229, 415, 31 },
		{ 229, 416, 31 },
		{ 229, 417, 31 },
		{ 229, 418, 31 },
		{ 229, 419, 31 },
		{ 229, 420, 31 },
		{ 229, 421, 31 },
		{ 229, 422, 31 },
		{ 229, 423, 31 },
		{ 229, 424, 31 },
		{ 229, 425, 31 },
		{ 229, 426, 31 },
		{ 229, 427, 31 },
		{ 229, 428, 31 },
		{ 229, 429, 31 },
		{ 229, 430, 31 },
		{ 229, 431, 31 },
		{ 229, 432, 31 },
		{ 229, 433, 31 },
		{ 229, 434, 31 },
		{ 229, 435, 31 },
		{ 229, 436, 31 },
		{ 229, 437, 31 },
		{ 229, 438, 31 },
		{ 229, 439, 31 },
		{ 229, 440, 31 },
		{ 229, 441, 31 },
		{ 229, 442, 31 },
		{ 229, 443, 31 },
		{ 229, 444, 31 },
		{ 229, 445, 31 },
		{ 229, 446, 31 },
		{ 229, 447, 31 },
		{ 229, 448, 31 },
		{ 229, 449, 31 },
		{ 229, 450, 31 },
		{ 229, 451, 31 },
		{ 229, 452, 31 },
		{ 229, 453, 31 },
		{ 229, 454, 31 },
		{ 229, 455, 31 },
		{ 229, 456, 31 },
		{ 229, 457, 31 },
		{ 229, 458, 31 },
		{ 229, 459, 31 },
		{ 229, 460, 31 },
		{ 229, 461, 31 },
		{ 229, 462, 31 },
		{ 229, 463, 31 },
		{ 229, 464, 31 },
		{ 229, 465, 31 },
		{ 229, 466, 31 },
		{ 229, 467, 31 },
		{ 229, 468, 31 },
		{ 229, 469, 31 },
		{ 229, 470, 31 },
		{ 229, 471, 31 },
		{ 229, 472, 31 },
		{ 229, 473, 31 },
		{ 229, 474, 31 },
		{ 229, 475, 31 },
		{ 229, 476, 31 },
		{ 229, 477, 31 },
		{ 229, 478, 31 },
		{ 229, 479, 31 },
		{ 229, 480, 31 },
		{ 229, 481, 31 },
		{ 229, 482, 31 },
		{ 229, 483, 31 },
		{ 229, 484, 31 },
		{ 229, 485, 31 },
		{ 229, 486, 31 },
		{ 229, 487, 31 },
		{ 229, 488, 31 },
		{ 229, 489, 31 },
		{ 229, 490, 31 },
		{ 229, 491, 31 },
		{ 229, 492, 31 },
		{ 229, 493, 31 },
		{ 229, 494, 31 },
		{ 229, 495, 31 },
		{ 229, 496, 31 },
		{ 229, 497, 31 },
		{ 229, 498, 31 },
		{ 229, 499, 31 },
		{ 229, 500, 31 },
		{ 229, 501, 31 },
		{ 229, 502, 31 },
		{ 229, 503, 31 },
		{ 229, 504, 31 },
		{ 229, 505, 31 },
		{ 229, 506, 31 },
		{ 229, 507, 31 },
		{ 229, 508, 31 },
		{ 229, 509, 31 },
		{ 229, 510, 31 },
		{ 229, 511, 31 },
		{ 229, 512, 31 },
		{ 230, 366, 31 },
		{ 230, 367, 31 },
		{ 230, 368, 31 },
		{ 230, 369, 31 },
		{ 230, 370, 31 },
		{ 230, 371, 31 },
		{ 230, 372, 31 },
		{ 230, 373, 31 },
		{ 230, 374, 31 },
		{ 230, 375, 31 },
		{ 230, 376, 31 },
		{ 230, 377, 31 },
		{ 230, 378, 31 },
		{ 230, 379, 31 },
		{ 230, 380, 31 },
		{ 230, 381, 31 },
		{ 230, 382, 31 },
		{ 230, 383, 31 },
		{ 230, 384, 31 },
		{ 230, 385, 31 },
		{ 230, 386, 31 },
		{ 230, 387, 31 },
		{ 230, 388, 31 },
		{ 230, 389, 31 },
		{ 230, 390, 31 },
		{ 230, 391, 31 },
		{ 230, 392, 31 },
		{ 230, 393, 31 },
		{ 230, 394, 31 },
		{ 230, 395, 31 },
		{ 230, 396, 31 },
		{ 230, 397, 31 },
		{ 230, 398, 31 },
		{ 230, 399, 31 },
		{ 230, 400, 31 },
		{ 230, 401, 31 },
		{ 230, 402, 31 },
		{ 230, 403, 31 },
		{ 230, 404, 31 },
		{ 230, 405, 31 },
		{ 230, 406, 31 },
		{ 230, 407, 31 },
		{ 230, 408, 31 },
		{ 230, 409, 31 },
		{ 230, 410, 31 },
		{ 230, 411, 31 },
		{ 230, 412, 31 },
		{ 230, 413, 31 },
		{ 230, 414, 31 },
		{ 230, 415, 31 },
		{ 230, 416, 31 },
		{ 230, 417, 31 },
		{ 230, 418, 31 },
		{ 230, 419, 31 },
		{ 230, 420, 31 },
		{ 230, 421, 31 },
		{ 230, 422, 31 },
		{ 230, 423, 31 },
		{ 230, 424, 31 },
		{ 230, 425, 31 },
		{ 230, 426, 31 },
		{ 230, 427, 31 },
		{ 230, 428, 31 },
		{ 230, 429, 31 },
		{ 230, 430, 31 },
		{ 230, 431, 31 },
		{ 230, 432, 31 },
		{ 230, 433, 31 },
		{ 230, 434, 31 },
		{ 230, 435, 31 },
		{ 230, 436, 31 },
		{ 230, 437, 31 },
		{ 230, 438, 31 },
		{ 230, 439, 31 },
		{ 230, 440, 31 },
		{ 230, 441, 31 },
		{ 230, 442, 31 },
		{ 230, 443, 31 },
		{ 230, 444, 31 },
		{ 230, 445, 31 },
		{ 230, 446, 31 },
		{ 230, 447, 31 },
		{ 230, 448, 31 },
		{ 230, 449, 31 },
		{ 230, 450, 31 },
		{ 230, 451, 31 },
		{ 230, 452, 31 },
		{ 230, 453, 31 },
		{ 230, 454, 31 },
		{ 230, 455, 31 },
		{ 230, 456, 31 },
		{ 230, 457, 31 },
		{ 230, 458, 31 },
		{ 230, 459, 31 },
		{ 230, 460, 31 },
		{ 230, 461, 31 },
		{ 230, 462, 31 },
		{ 230, 463, 31 },
		{ 230, 464, 31 },
		{ 230, 465, 31 },
		{ 230, 466, 31 },
		{ 230, 467, 31 },
		{ 230, 468, 31 },
		{ 230, 469, 31 },
		{ 230, 470, 31 },
		{ 230, 471, 31 },
		{ 230, 472, 31 },
		{ 230, 473, 31 },
		{ 230, 474, 31 },
		{ 230, 475, 31 },
		{ 230, 476, 31 },
		{ 230, 477, 31 },
		{ 230, 478, 31 },
		{ 230, 479, 31 },
		{ 230, 480, 31 },
		{ 230, 481, 31 },
		{ 230, 482, 31 },
		{ 230, 483, 31 },
		{ 230, 484, 31 },
		{ 230, 485, 31 },
		{ 230, 486, 31 },
		{ 230, 487, 31 },
		{ 230, 488, 31 },
		{ 230, 489, 31 },
		{ 230, 490, 31 },
		{ 230, 491, 31 },
		{ 230, 492, 31 },
		{ 230, 493, 31 },
		{ 230, 494, 31 },
		{ 230, 495, 31 },
		{ 230, 496, 31 },
		{ 230, 497, 31 },
		{ 230, 498, 31 },
		{ 230, 499, 31 },
		{ 230, 500, 31 },
		{ 230, 501, 31 },
		{ 230, 502, 31 },
		{ 230, 503, 31 },
		{ 230, 504, 31 },
		{ 230, 505, 31 },
		{ 230, 506, 31 },
		{ 230, 507, 31 },
		{ 230, 508, 31 },
		{ 230, 509, 31 },
		{ 230, 510, 31 },
		{ 230, 511, 31 },
		{ 230, 512, 31 },
		{ 231, 362, 31 },
		{ 231, 363, 31 },
		{ 231, 364, 31 },
		{ 231, 365, 31 },
		{ 231, 366, 31 },
		{ 231, 367, 31 },
		{ 231, 368, 31 },
		{ 231, 369, 31 },
		{ 231, 370, 31 },
		{ 231, 371, 31 },
		{ 231, 372, 31 },
		{ 231, 373, 31 },
		{ 231, 374, 31 },
		{ 231, 375, 31 },
		{ 231, 376, 31 },
		{ 231, 377, 31 },
		{ 231, 378, 31 },
		{ 231, 379, 31 },
		{ 231, 380, 31 },
		{ 231, 381, 31 },
		{ 231, 382, 31 },
		{ 231, 383, 31 },
		{ 231, 384, 31 },
		{ 231, 385, 31 },
		{ 231, 386, 31 },
		{ 231, 387, 31 },
		{ 231, 388, 31 },
		{ 231, 389, 31 },
		{ 231, 390, 31 },
		{ 231, 391, 31 },
		{ 231, 392, 31 },
		{ 231, 393, 31 },
		{ 231, 394, 31 },
		{ 231, 395, 31 },
		{ 231, 396, 31 },
		{ 231, 397, 31 },
		{ 231, 398, 31 },
		{ 231, 399, 31 },
		{ 231, 400, 31 },
		{ 231, 401, 31 },
		{ 231, 402, 31 },
		{ 231, 403, 31 },
		{ 231, 404, 31 },
		{ 231, 405, 31 },
		{ 231, 406, 31 },
		{ 231, 407, 31 },
		{ 231, 408, 31 },
		{ 231, 409, 31 },
		{ 231, 410, 31 },
		{ 231, 411, 31 },
		{ 231, 412, 31 },
		{ 231, 413, 31 },
		{ 231, 414, 31 },
		{ 231, 415, 31 },
		{ 231, 416, 31 },
		{ 231, 417, 31 },
		{ 231, 418, 31 },
		{ 231, 419, 31 },
		{ 231, 420, 31 },
		{ 231, 421, 31 },
		{ 231, 422, 31 },
		{ 231, 423, 31 },
		{ 231, 424, 31 },
		{ 231, 425, 31 },
		{ 231, 426, 31 },
		{ 231, 427, 31 },
		{ 231, 428, 31 },
		{ 231, 429, 31 },
		{ 231, 430, 31 },
		{ 231, 431, 31 },
		{ 231, 432, 31 },
		{ 231, 433, 31 },
		{ 231, 434, 31 },
		{ 231, 435, 31 },
		{ 231, 436, 31 },
		{ 231, 437, 31 },
		{ 231, 438, 31 },
		{ 231, 439, 31 },
		{ 231, 440, 31 },
		{ 231, 441, 31 },
		{ 231, 442, 31 },
		{ 231, 443, 31 },
		{ 231, 444, 31 },
		{ 231, 445, 31 },
		{ 231, 446, 31 },
		{ 231, 447, 31 },
		{ 231, 448, 31 },
		{ 231, 449, 31 },
		{ 231, 450, 31 },
		{ 231, 451, 31 },
		{ 231, 452, 31 },
		{ 231, 453, 31 },
		{ 231, 454, 31 },
		{ 231, 455, 31 },
		{ 231, 456, 31 },
		{ 231, 457, 31 },
		{ 231, 458, 31 },
		{ 231, 459, 31 },
		{ 231, 460, 31 },
		{ 231, 461, 31 },
		{ 231, 462, 31 },
		{ 231, 463, 31 },
		{ 231, 464, 31 },
		{ 231, 465, 31 },
		{ 231, 466, 31 },
		{ 231, 467, 31 },
		{ 231, 468, 31 },
		{ 231, 469, 31 },
		{ 231, 470, 31 },
		{ 231, 471, 31 },
		{ 231, 472, 31 },
		{ 231, 473, 31 },
		{ 231, 474, 31 },
		{ 231, 475, 31 },
		{ 231, 476, 31 },
		{ 231, 477, 31 },
		{ 231, 478, 31 },
		{ 231, 479, 31 },
		{ 231, 480, 31 },
		{ 231, 481, 31 },
		{ 231, 482, 31 },
		{ 231, 483, 31 },
		{ 231, 484, 31 },
		{ 231, 485, 31 },
		{ 231, 486, 31 },
		{ 231, 487, 31 },
		{ 231, 488, 31 },
		{ 231, 489, 31 },
		{ 231, 490, 31 },
		{ 231, 491, 31 },
		{ 231, 492, 31 },
		{ 231, 493, 31 },
		{ 231, 494, 31 },
		{ 231, 495, 31 },
		{ 231, 496, 31 },
		{ 231, 497, 31 },
		{ 231, 498, 31 },
		{ 231, 499, 31 },
		{ 231, 500, 31 },
		{ 231, 501, 31 },
		{ 231, 502, 31 },
		{ 231, 503, 31 },
		{ 231, 504, 31 },
		{ 231, 505, 31 },
		{ 231, 506, 31 },
		{ 231, 507, 31 },
		{ 231, 508, 31 },
		{ 231, 509, 31 },
		{ 231, 510, 31 },
		{ 231, 511, 31 },
		{ 231, 512, 31 },
		{ 232, 358, 31 },
		{ 232, 359, 31 },
		{ 232, 360, 31 },
		{ 232, 361, 31 },
		{ 232, 362, 31 },
		{ 232, 363, 31 },
		{ 232, 364, 31 },
		{ 232, 365, 31 },
		{ 232, 366, 31 },
		{ 232, 367, 31 },
		{ 232, 368, 31 },
		{ 232, 369, 31 },
		{ 232, 370, 31 },
		{ 232, 371, 31 },
		{ 232, 372, 31 },
		{ 232, 373, 31 },
		{ 232, 374, 31 },
		{ 232, 375, 31 },
		{ 232, 376, 31 },
		{ 232, 377, 31 },
		{ 232, 378, 31 },
		{ 232, 379, 31 },
		{ 232, 380, 31 },
		{ 232, 381, 31 },
		{ 232, 382, 31 },
		{ 232, 383, 31 },
		{ 232, 384, 31 },
		{ 232, 385, 31 },
		{ 232, 386, 31 },
		{ 232, 387, 31 },
		{ 232, 388, 31 },
		{ 232, 389, 31 },
		{ 232, 390, 31 },
		{ 232, 391, 31 },
		{ 232, 392, 31 },
		{ 232, 393, 31 },
		{ 232, 394, 31 },
		{ 232, 395, 31 },
		{ 232, 396, 31 },
		{ 232, 397, 31 },
		{ 232, 398, 31 },
		{ 232, 399, 31 },
		{ 232, 400, 31 },
		{ 232, 401, 31 },
		{ 232, 402, 31 },
		{ 232, 403, 31 },
		{ 232, 404, 31 },
		{ 232, 405, 31 },
		{ 232, 406, 31 },
		{ 232, 407, 31 },
		{ 232, 408, 31 },
		{ 232, 409, 31 },
		{ 232, 410, 31 },
		{ 232, 411, 31 },
		{ 232, 412, 31 },
		{ 232, 413, 31 },
		{ 232, 414, 31 },
		{ 232, 415, 31 },
		{ 232, 416, 31 },
		{ 232, 417, 31 },
		{ 232, 418, 31 },
		{ 232, 419, 31 },
		{ 232, 420, 31 },
		{ 232, 421, 31 },
		{ 232, 422, 31 },
		{ 232, 423, 31 },
		{ 232, 424, 31 },
		{ 232, 425, 31 },
		{ 232, 426, 31 },
		{ 232, 427, 31 },
		{ 232, 428, 31 },
		{ 232, 429, 31 },
		{ 232, 430, 31 },
		{ 232, 431, 31 },
		{ 232, 432, 31 },
		{ 232, 433, 31 },
		{ 232, 434, 31 },
		{ 232, 435, 31 },
		{ 232, 436, 31 },
		{ 232, 437, 31 },
		{ 232, 438, 31 },
		{ 232, 439, 31 },
		{ 232, 440, 31 },
		{ 232, 441, 31 },
		{ 232, 442, 31 },
		{ 232, 443, 31 },
		{ 232, 444, 31 },
		{ 232, 445, 31 },
		{ 232, 446, 31 },
		{ 232, 447, 31 },
		{ 232, 448, 31 },
		{ 232, 449, 31 },
		{ 232, 450, 31 },
		{ 232, 451, 31 },
		{ 232, 452, 31 },
		{ 232, 453, 31 },
		{ 232, 454, 31 },
		{ 232, 455, 31 },
		{ 232, 456, 31 },
		{ 232, 457, 31 },
		{ 232, 458, 31 },
		{ 232, 459, 31 },
		{ 232, 460, 31 },
		{ 232, 461, 31 },
		{ 232, 462, 31 },
		{ 232, 463, 31 },
		{ 232, 464, 31 },
		{ 232, 465, 31 },
		{ 232, 466, 31 },
		{ 232, 467, 31 },
		{ 232, 468, 31 },
		{ 232, 469, 31 },
		{ 232, 470, 31 },
		{ 232, 471, 31 },
		{ 232, 472, 31 },
		{ 232, 473, 31 },
		{ 232, 474, 31 },
		{ 232, 475, 31 },
		{ 232, 476, 31 },
		{ 232, 477, 31 },
		{ 232, 478, 31 },
		{ 232, 479, 31 },
		{ 232, 480, 31 },
		{ 232, 481, 31 },
		{ 232, 482, 31 },
		{ 232, 483, 31 },
		{ 232, 484, 31 },
		{ 232, 485, 31 },
		{ 232, 486, 31 },
		{ 232, 487, 31 },
		{ 232, 488, 31 },
		{ 232, 489, 31 },
		{ 232, 490, 31 },
		{ 232, 491, 31 },
		{ 232, 492, 31 },
		{ 232, 493, 31 },
		{ 232, 494, 31 },
		{ 232, 495, 31 },
		{ 232, 496, 31 },
		{ 232, 497, 31 },
		{ 232, 498, 31 },
		{ 232, 499, 31 },
		{ 232, 500, 31 },
		{ 232, 501, 31 },
		{ 232, 502, 31 },
		{ 232, 503, 31 },
		{ 232, 504, 31 },
		{ 232, 505, 31 },
		{ 232, 506, 31 },
		{ 232, 507, 31 },
		{ 232, 508, 31 },
		{ 232, 509, 31 },
		{ 232, 510, 31 },
		{ 232, 511, 31 },
		{ 232, 512, 31 },
		{ 233, 354, 31 },
		{ 233, 355, 31 },
		{ 233, 356, 31 },
		{ 233, 357, 31 },
		{ 233, 358, 31 },
		{ 233, 359, 31 },
		{ 233, 360, 31 },
		{ 233, 361, 31 },
		{ 233, 362, 31 },
		{ 233, 363, 31 },
		{ 233, 364, 31 },
		{ 233, 365, 31 },
		{ 233, 366, 31 },
		{ 233, 367, 31 },
		{ 233, 368, 31 },
		{ 233, 369, 31 },
		{ 233, 370, 31 },
		{ 233, 371, 31 },
		{ 233, 372, 31 },
		{ 233, 373, 31 },
		{ 233, 374, 31 },
		{ 233, 375, 31 },
		{ 233, 376, 31 },
		{ 233, 377, 31 },
		{ 233, 378, 31 },
		{ 233, 379, 31 },
		{ 233, 380, 31 },
		{ 233, 381, 31 },
		{ 233, 382, 31 },
		{ 233, 383, 31 },
		{ 233, 384, 31 },
		{ 233, 385, 31 },
		{ 233, 386, 31 },
		{ 233, 387, 31 },
		{ 233, 388, 31 },
		{ 233, 389, 31 },
		{ 233, 390, 31 },
		{ 233, 391, 31 },
		{ 233, 392, 31 },
		{ 233, 393, 31 },
		{ 233, 394, 31 },
		{ 233, 395, 31 },
		{ 233, 396, 31 },
		{ 233, 397, 31 },
		{ 233, 398, 31 },
		{ 233, 399, 31 },
		{ 233, 400, 31 },
		{ 233, 401, 31 },
		{ 233, 402, 31 },
		{ 233, 403, 31 },
		{ 233, 404, 31 },
		{ 233, 405, 31 },
		{ 233, 406, 31 },
		{ 233, 407, 31 },
		{ 233, 408, 31 },
		{ 233, 409, 31 },
		{ 233, 410, 31 },
		{ 233, 411, 31 },
		{ 233, 412, 31 },
		{ 233, 413, 31 },
		{ 233, 414, 31 },
		{ 233, 415, 31 },
		{ 233, 416, 31 },
		{ 233, 417, 31 },
		{ 233, 418, 31 },
		{ 233, 419, 31 },
		{ 233, 420, 31 },
		{ 233, 421, 31 },
		{ 233, 422, 31 },
		{ 233, 423, 31 },
		{ 233, 424, 31 },
		{ 233, 425, 31 },
		{ 233, 426, 31 },
		{ 233, 427, 31 },
		{ 233, 428, 31 },
		{ 233, 429, 31 },
		{ 233, 430, 31 },
		{ 233, 431, 31 },
		{ 233, 432, 31 },
		{ 233, 433, 31 },
		{ 233, 434, 31 },
		{ 233, 435, 31 },
		{ 233, 436, 31 },
		{ 233, 437, 31 },
		{ 233, 438, 31 },
		{ 233, 439, 31 },
		{ 233, 440, 31 },
		{ 233, 441, 31 },
		{ 233, 442, 31 },
		{ 233, 443, 31 },
		{ 233, 444, 31 },
		{ 233, 445, 31 },
		{ 233, 446, 31 },
		{ 233, 447, 31 },
		{ 233, 448, 31 },
		{ 233, 449, 31 },
		{ 233, 450, 31 },
		{ 233, 451, 31 },
		{ 233, 452, 31 },
		{ 233, 453, 31 },
		{ 233, 454, 31 },
		{ 233, 455, 31 },
		{ 233, 456, 31 },
		{ 233, 457, 31 },
		{ 233, 458, 31 },
		{ 233, 459, 31 },
		{ 233, 460, 31 },
		{ 233, 461, 31 },
		{ 233, 462, 31 },
		{ 233, 463, 31 },
		{ 233, 464, 31 },
		{ 233, 465, 31 },
		{ 233, 466, 31 },
		{ 233, 467, 31 },
		{ 233, 468, 31 },
		{ 233, 469, 31 },
		{ 233, 470, 31 },
		{ 233, 471, 31 },
		{ 233, 472, 31 },
		{ 233, 473, 31 },
		{ 233, 474, 31 },
		{ 233, 475, 31 },
		{ 233, 476, 31 },
		{ 233, 477, 31 },
		{ 233, 478, 31 },
		{ 233, 479, 31 },
		{ 233, 480, 31 },
		{ 233, 481, 31 },
		{ 233, 482, 31 },
		{ 233, 483, 31 },
		{ 233, 484, 31 },
		{ 233, 485, 31 },
		{ 233, 486, 31 },
		{ 233, 487, 31 },
		{ 233, 488, 31 },
		{ 233, 489, 31 },
		{ 233, 490, 31 },
		{ 233, 491, 31 },
		{ 233, 492, 31 },
		{ 233, 493, 31 },
		{ 233, 494, 31 },
		{ 233, 495, 31 },
		{ 233, 496, 31 },
		{ 233, 497, 31 },
		{ 233, 498, 31 },
		{ 233, 499, 31 },
		{ 233, 500, 31 },
		{ 233, 501, 31 },
		{ 233, 502, 31 },
		{ 233, 503, 31 },
		{ 233, 504, 31 },
		{ 233, 505, 31 },
		{ 233, 506, 31 },
		{ 233, 507, 31 },
		{ 233, 508, 31 },
		{ 233, 509, 31 },
		{ 233, 510, 31 },
		{ 233, 511, 31 },
		{ 233, 512, 31 },
		{ 234, 350, 31 },
		{ 234, 351, 31 },
		{ 234, 352, 31 },
		{ 234, 353, 31 },
		{ 234, 354, 31 },
		{ 234, 355, 31 },
		{ 234, 356, 31 },
		{ 234, 357, 31 },
		{ 234, 358, 31 },
		{ 234, 359, 31 },
		{ 234, 360, 31 },
		{ 234, 361, 31 },
		{ 234, 362, 31 },
		{ 234, 363, 31 },
		{ 234, 364, 31 },
		{ 234, 365, 31 },
		{ 234, 366, 31 },
		{ 234, 367, 31 },
		{ 234, 368, 31 },
		{ 234, 369, 31 },
		{ 234, 370, 31 },
		{ 234, 371, 31 },
		{ 234, 372, 31 },
		{ 234, 373, 31 },
		{ 234, 374, 31 },
		{ 234, 375, 31 },
		{ 234, 376, 31 },
		{ 234, 377, 31 },
		{ 234, 378, 31 },
		{ 234, 379, 31 },
		{ 234, 380, 31 },
		{ 234, 381, 31 },
		{ 234, 382, 31 },
		{ 234, 383, 31 },
		{ 234, 384, 31 },
		{ 234, 385, 31 },
		{ 234, 386, 31 },
		{ 234, 387, 31 },
		{ 234, 388, 31 },
		{ 234, 389, 31 },
		{ 234, 390, 31 },
		{ 234, 391, 31 },
		{ 234, 392, 31 },
		{ 234, 393, 31 },
		{ 234, 394, 31 },
		{ 234, 395, 31 },
		{ 234, 396, 31 },
		{ 234, 397, 31 },
		{ 234, 398, 31 },
		{ 234, 399, 31 },
		{ 234, 400, 31 },
		{ 234, 401, 31 },
		{ 234, 402, 31 },
		{ 234, 403, 31 },
		{ 234, 404, 31 },
		{ 234, 405, 31 },
		{ 234, 406, 31 },
		{ 234, 407, 31 },
		{ 234, 408, 31 },
		{ 234, 409, 31 },
		{ 234, 410, 31 },
		{ 234, 411, 31 },
		{ 234, 412, 31 },
		{ 234, 413, 31 },
		{ 234, 414, 31 },
		{ 234, 415, 31 },
		{ 234, 416, 31 },
		{ 234, 417, 31 },
		{ 234, 418, 31 },
		{ 234, 419, 31 },
		{ 234, 420, 31 },
		{ 234, 421, 31 },
		{ 234, 422, 31 },
		{ 234, 423, 31 },
		{ 234, 424, 31 },
		{ 234, 425, 31 },
		{ 234, 426, 31 },
		{ 234, 427, 31 },
		{ 234, 428, 31 },
		{ 234, 429, 31 },
		{ 234, 430, 31 },
		{ 234, 431, 31 },
		{ 234, 432, 31 },
		{ 234, 433, 31 },
		{ 234, 434, 31 },
		{ 234, 435, 31 },
		{ 234, 436, 31 },
		{ 234, 437, 31 },
		{ 234, 438, 31 },
		{ 234, 439, 31 },
		{ 234, 440, 31 },
		{ 234, 441, 31 },
		{ 234, 442, 31 },
		{ 234, 443, 31 },
		{ 234, 444, 31 },
		{ 234, 445, 31 },
		{ 234, 446, 31 },
		{ 234, 447, 31 },
		{ 234, 448, 31 },
		{ 234, 449, 31 },
		{ 234, 450, 31 },
		{ 234, 451, 31 },
		{ 234, 452, 31 },
		{ 234, 453, 31 },
		{ 234, 454, 31 },
		{ 234, 455, 31 },
		{ 234, 456, 31 },
		{ 234, 457, 31 },
		{ 234, 458, 31 },
		{ 234, 459, 31 },
		{ 234, 460, 31 },
		{ 234, 461, 31 },
		{ 234, 462, 31 },
		{ 234, 463, 31 },
		{ 234, 464, 31 },
		{ 234, 465, 31 },
		{ 234, 466, 31 },
		{ 234, 467, 31 },
		{ 234, 468, 31 },
		{ 234, 469, 31 },
		{ 234, 470, 31 },
		{ 234, 471, 31 },
		{ 234, 472, 31 },
		{ 234, 473, 31 },
		{ 234, 474, 31 },
		{ 234, 475, 31 },
		{ 234, 476, 31 },
		{ 234, 477, 31 },
		{ 234, 478, 31 },
		{ 234, 479, 31 },
		{ 234, 480, 31 },
		{ 234, 481, 31 },
		{ 234, 482, 31 },
		{ 234, 483, 31 },
		{ 234, 484, 31 },
		{ 234, 485, 31 },
		{ 234, 486, 31 },
		{ 234, 487, 31 },
		{ 234, 488, 31 },
		{ 234, 489, 31 },
		{ 234, 490, 31 },
		{ 234, 491, 31 },
		{ 234, 492, 31 },
		{ 234, 493, 31 },
		{ 234, 494, 31 },
		{ 234, 495, 31 },
		{ 234, 496, 31 },
		{ 234, 497, 31 },
		{ 234, 498, 31 },
		{ 234, 499, 31 },
		{ 234, 500, 31 },
		{ 234, 501, 31 },
		{ 234, 502, 31 },
		{ 234, 503, 31 },
		{ 234, 504, 31 },
		{ 234, 505, 31 },
		{ 234, 506, 31 },
		{ 234, 507, 31 },
		{ 234, 508, 31 },
		{ 234, 509, 31 },
		{ 234, 510, 31 },
		{ 234, 511, 31 },
		{ 234, 512, 31 },
		{ 235, 346, 31 },
		{ 235, 347, 31 },
		{ 235, 348, 31 },
		{ 235, 349, 31 },
		{ 235, 350, 31 },
		{ 235, 351, 31 },
		{ 235, 352, 31 },
		{ 235, 353, 31 },
		{ 235, 354, 31 },
		{ 235, 355, 31 },
		{ 235, 356, 31 },
		{ 235, 357, 31 },
		{ 235, 358, 31 },
		{ 235, 359, 31 },
		{ 235, 360, 31 },
		{ 235, 361, 31 },
		{ 235, 362, 31 },
		{ 235, 363, 31 },
		{ 235, 364, 31 },
		{ 235, 365, 31 },
		{ 235, 366, 31 },
		{ 235, 367, 31 },
		{ 235, 368, 31 },
		{ 235, 369, 31 },
		{ 235, 370, 31 },
		{ 235, 371, 31 },
		{ 235, 372, 31 },
		{ 235, 373, 31 },
		{ 235, 374, 31 },
		{ 235, 375, 31 },
		{ 235, 376, 31 },
		{ 235, 377, 31 },
		{ 235, 378, 31 },
		{ 235, 379, 31 },
		{ 235, 380, 31 },
		{ 235, 381, 31 },
		{ 235, 382, 31 },
		{ 235, 383, 31 },
		{ 235, 384, 31 },
		{ 235, 385, 31 },
		{ 235, 386, 31 },
		{ 235, 387, 31 },
		{ 235, 388, 31 },
		{ 235, 389, 31 },
		{ 235, 390, 31 },
		{ 235, 391, 31 },
		{ 235, 392, 31 },
		{ 235, 393, 31 },
		{ 235, 394, 31 },
		{ 235, 395, 31 },
		{ 235, 396, 31 },
		{ 235, 397, 31 },
		{ 235, 398, 31 },
		{ 235, 399, 31 },
		{ 235, 400, 31 },
		{ 235, 401, 31 },
		{ 235, 402, 31 },
		{ 235, 403, 31 },
		{ 235, 404, 31 },
		{ 235, 405, 31 },
		{ 235, 406, 31 },
		{ 235, 407, 31 },
		{ 235, 408, 31 },
		{ 235, 409, 31 },
		{ 235, 410, 31 },
		{ 235, 411, 31 },
		{ 235, 412, 31 },
		{ 235, 413, 31 },
		{ 235, 414, 31 },
		{ 235, 415, 31 },
		{ 235, 416, 31 },
		{ 235, 417, 31 },
		{ 235, 418, 31 },
		{ 235, 419, 31 },
		{ 235, 420, 31 },
		{ 235, 421, 31 },
		{ 235, 422, 31 },
		{ 235, 423, 31 },
		{ 235, 424, 31 },
		{ 235, 425, 31 },
		{ 235, 426, 31 },
		{ 235, 427, 31 },
		{ 235, 428, 31 },
		{ 235, 429, 31 },
		{ 235, 430, 31 },
		{ 235, 431, 31 },
		{ 235, 432, 31 },
		{ 235, 433, 31 },
		{ 235, 434, 31 },
		{ 235, 435, 31 },
		{ 235, 436, 31 },
		{ 235, 437, 31 },
		{ 235, 438, 31 },
		{ 235, 439, 31 },
		{ 235, 440, 31 },
		{ 235, 441, 31 },
		{ 235, 442, 31 },
		{ 235, 443, 31 },
		{ 235, 444, 31 },
		{ 235, 445, 31 },
		{ 235, 446, 31 },
		{ 235, 447, 31 },
		{ 235, 448, 31 },
		{ 235, 449, 31 },
		{ 235, 450, 31 },
		{ 235, 451, 31 },
		{ 235, 452, 31 },
		{ 235, 453, 31 },
		{ 235, 454, 31 },
		{ 235, 455, 31 },
		{ 235, 456, 31 },
		{ 235, 457, 31 },
		{ 235, 458, 31 },
		{ 235, 459, 31 },
		{ 235, 460, 31 },
		{ 235, 461, 31 },
		{ 235, 462, 31 },
		{ 235, 463, 31 },
		{ 235, 464, 31 },
		{ 235, 465, 31 },
		{ 235, 466, 31 },
		{ 235, 467, 31 },
		{ 235, 468, 31 },
		{ 235, 469, 31 },
		{ 235, 470, 31 },
		{ 235, 471, 31 },
		{ 235, 472, 31 },
		{ 235, 473, 31 },
		{ 235, 474, 31 },
		{ 235, 475, 31 },
		{ 235, 476, 31 },
		{ 235, 477, 31 },
		{ 235, 478, 31 },
		{ 235, 479, 31 },
		{ 235, 480, 31 },
		{ 235, 481, 31 },
		{ 235, 482, 31 },
		{ 235, 483, 31 },
		{ 235, 484, 31 },
		{ 235, 485, 31 },
		{ 235, 486, 31 },
		{ 235, 487, 31 },
		{ 235, 488, 31 },
		{ 235, 489, 31 },
		{ 235, 490, 31 },
		{ 235, 491, 31 },
		{ 235, 492, 31 },
		{ 235, 493, 31 },
		{ 235, 494, 31 },
		{ 235, 495, 31 },
		{ 235, 496, 31 },
		{ 235, 497, 31 },
		{ 235, 498, 31 },
		{ 235, 499, 31 },
		{ 235, 500, 31 },
		{ 235, 501, 31 },
		{ 235, 502, 31 },
		{ 235, 503, 31 },
		{ 235, 504, 31 },
		{ 235, 505, 31 },
		{ 235, 506, 31 },
		{ 235, 507, 31 },
		{ 235, 508, 31 },
		{ 235, 509, 31 },
		{ 235, 510, 31 },
		{ 235, 511, 31 },
		{ 235, 512, 31 },
		{ 236, 342, 31 },
		{ 236, 343, 31 },
		{ 236, 344, 31 },
		{ 236, 345, 31 },
		{ 236, 346, 31 },
		{ 236, 347, 31 },
		{ 236, 348, 31 },
		{ 236, 349, 31 },
		{ 236, 350, 31 },
		{ 236, 351, 31 },
		{ 236, 352, 31 },
		{ 236, 353, 31 },
		{ 236, 354, 31 },
		{ 236, 355, 31 },
		{ 236, 356, 31 },
		{ 236, 357, 31 },
		{ 236, 358, 31 },
		{ 236, 359, 31 },
		{ 236, 360, 31 },
		{ 236, 361, 31 },
		{ 236, 362, 31 },
		{ 236, 363, 31 },
		{ 236, 364, 31 },
		{ 236, 365, 31 },
		{ 236, 366, 31 },
		{ 236, 367, 31 },
		{ 236, 368, 31 },
		{ 236, 369, 31 },
		{ 236, 370, 31 },
		{ 236, 371, 31 },
		{ 236, 372, 31 },
		{ 236, 373, 31 },
		{ 236, 374, 31 },
		{ 236, 375, 31 },
		{ 236, 376, 31 },
		{ 236, 377, 31 },
		{ 236, 378, 31 },
		{ 236, 379, 31 },
		{ 236, 380, 31 },
		{ 236, 381, 31 },
		{ 236, 382, 31 },
		{ 236, 383, 31 },
		{ 236, 384, 31 },
		{ 236, 385, 31 },
		{ 236, 386, 31 },
		{ 236, 387, 31 },
		{ 236, 388, 31 },
		{ 236, 389, 31 },
		{ 236, 390, 31 },
		{ 236, 391, 31 },
		{ 236, 392, 31 },
		{ 236, 393, 31 },
		{ 236, 394, 31 },
		{ 236, 395, 31 },
		{ 236, 396, 31 },
		{ 236, 397, 31 },
		{ 236, 398, 31 },
		{ 236, 399, 31 },
		{ 236, 400, 31 },
		{ 236, 401, 31 },
		{ 236, 402, 31 },
		{ 236, 403, 31 },
		{ 236, 404, 31 },
		{ 236, 405, 31 },
		{ 236, 406, 31 },
		{ 236, 407, 31 },
		{ 236, 408, 31 },
		{ 236, 409, 31 },
		{ 236, 410, 31 },
		{ 236, 411, 31 },
		{ 236, 412, 31 },
		{ 236, 413, 31 },
		{ 236, 414, 31 },
		{ 236, 415, 31 },
		{ 236, 416, 31 },
		{ 236, 417, 31 },
		{ 236, 418, 31 },
		{ 236, 419, 31 },
		{ 236, 420, 31 },
		{ 236, 421, 31 },
		{ 236, 422, 31 },
		{ 236, 423, 31 },
		{ 236, 424, 31 },
		{ 236, 425, 31 },
		{ 236, 426, 31 },
		{ 236, 427, 31 },
		{ 236, 428, 31 },
		{ 236, 429, 31 },
		{ 236, 430, 31 },
		{ 236, 431, 31 },
		{ 236, 432, 31 },
		{ 236, 433, 31 },
		{ 236, 434, 31 },
		{ 236, 435, 31 },
		{ 236, 436, 31 },
		{ 236, 437, 31 },
		{ 236, 438, 31 },
		{ 236, 439, 31 },
		{ 236, 440, 31 },
		{ 236, 441, 31 },
		{ 236, 442, 31 },
		{ 236, 443, 31 },
		{ 236, 444, 31 },
		{ 236, 445, 31 },
		{ 236, 446, 31 },
		{ 236, 447, 31 },
		{ 236, 448, 31 },
		{ 236, 449, 31 },
		{ 236, 450, 31 },
		{ 236, 451, 31 },
		{ 236, 452, 31 },
		{ 236, 453, 31 },
		{ 236, 454, 31 },
		{ 236, 455, 31 },
		{ 236, 456, 31 },
		{ 236, 457, 31 },
		{ 236, 458, 31 },
		{ 236, 459, 31 },
		{ 236, 460, 31 },
		{ 236, 461, 31 },
		{ 236, 462, 31 },
		{ 236, 463, 31 },
		{ 236, 464, 31 },
		{ 236, 465, 31 },
		{ 236, 466, 31 },
		{ 236, 467, 31 },
		{ 236, 468, 31 },
		{ 236, 469, 31 },
		{ 236, 470, 31 },
		{ 236, 471, 31 },
		{ 236, 472, 31 },
		{ 236, 473, 31 },
		{ 236, 474, 31 },
		{ 236, 475, 31 },
		{ 236, 476, 31 },
		{ 236, 477, 31 },
		{ 236, 478, 31 },
		{ 236, 479, 31 },
		{ 236, 480, 31 },
		{ 236, 481, 31 },
		{ 236, 482, 31 },
		{ 236, 483, 31 },
		{ 236, 484, 31 },
		{ 236, 485, 31 },
		{ 236, 486, 31 },
		{ 236, 487, 31 },
		{ 236, 488, 31 },
		{ 236, 489, 31 },
		{ 236, 490, 31 },
		{ 236, 491, 31 },
		{ 236, 492, 31 },
		{ 236, 493, 31 },
		{ 236, 494, 31 },
		{ 236, 495, 31 },
		{ 236, 496, 31 },
		{ 236, 497, 31 },
		{ 236, 498, 31 },
		{ 236, 499, 31 },
		{ 236, 500, 31 },
		{ 236, 501, 31 },
		{ 236, 502, 31 },
		{ 236, 503, 31 },
		{ 236, 504, 31 },
		{ 236, 505, 31 },
		{ 236, 506, 31 },
		{ 236, 507, 31 },
		{ 236, 508, 31 },
		{ 236, 509, 31 },
		{ 236, 510, 31 },
		{ 236, 511, 31 },
		{ 236, 512, 31 },
		{ 237, 338, 31 },
		{ 237, 339, 31 },
		{ 237, 340, 31 },
		{ 237, 341, 31 },
		{ 237, 342, 31 },
		{ 237, 343, 31 },
		{ 237, 344, 31 },
		{ 237, 345, 31 },
		{ 237, 346, 31 },
		{ 237, 347, 31 },
		{ 237, 348, 31 },
		{ 237, 349, 31 },
		{ 237, 350, 31 },
		{ 237, 351, 31 },
		{ 237, 352, 31 },
		{ 237, 353, 31 },
		{ 237, 354, 31 },
		{ 237, 355, 31 },
		{ 237, 356, 31 },
		{ 237, 357, 31 },
		{ 237, 358, 31 },
		{ 237, 359, 31 },
		{ 237, 360, 31 },
		{ 237, 361, 31 },
		{ 237, 362, 31 },
		{ 237, 363, 31 },
		{ 237, 364, 31 },
		{ 237, 365, 31 },
		{ 237, 366, 31 },
		{ 237, 367, 31 },
		{ 237, 368, 31 },
		{ 237, 369, 31 },
		{ 237, 370, 31 },
		{ 237, 371, 31 },
		{ 237, 372, 31 },
		{ 237, 373, 31 },
		{ 237, 374, 31 },
		{ 237, 375, 31 },
		{ 237, 376, 31 },
		{ 237, 377, 31 },
		{ 237, 378, 31 },
		{ 237, 379, 31 },
		{ 237, 380, 31 },
		{ 237, 381, 31 },
		{ 237, 382, 31 },
		{ 237, 383, 31 },
		{ 237, 384, 31 },
		{ 237, 385, 31 },
		{ 237, 386, 31 },
		{ 237, 387, 31 },
		{ 237, 388, 31 },
		{ 237, 389, 31 },
		{ 237, 390, 31 },
		{ 237, 391, 31 },
		{ 237, 392, 31 },
		{ 237, 393, 31 },
		{ 237, 394, 31 },
		{ 237, 395, 31 },
		{ 237, 396, 31 },
		{ 237, 397, 31 },
		{ 237, 398, 31 },
		{ 237, 399, 31 },
		{ 237, 400, 31 },
		{ 237, 401, 31 },
		{ 237, 402, 31 },
		{ 237, 403, 31 },
		{ 237, 404, 31 },
		{ 237, 405, 31 },
		{ 237, 406, 31 },
		{ 237, 407, 31 },
		{ 237, 408, 31 },
		{ 237, 409, 31 },
		{ 237, 410, 31 },
		{ 237, 411, 31 },
		{ 237, 412, 31 },
		{ 237, 413, 31 },
		{ 237, 414, 31 },
		{ 237, 415, 31 },
		{ 237, 416, 31 },
		{ 237, 417, 31 },
		{ 237, 418, 31 },
		{ 237, 419, 31 },
		{ 237, 420, 31 },
		{ 237, 421, 31 },
		{ 237, 422, 31 },
		{ 237, 423, 31 },
		{ 237, 424, 31 },
		{ 237, 425, 31 },
		{ 237, 426, 31 },
		{ 237, 427, 31 },
		{ 237, 428, 31 },
		{ 237, 429, 31 },
		{ 237, 430, 31 },
		{ 237, 431, 31 },
		{ 237, 432, 31 },
		{ 237, 433, 31 },
		{ 237, 434, 31 },
		{ 237, 435, 31 },
		{ 237, 436, 31 },
		{ 237, 437, 31 },
		{ 237, 438, 31 },
		{ 237, 439, 31 },
		{ 237, 440, 31 },
		{ 237, 441, 31 },
		{ 237, 442, 31 },
		{ 237, 443, 31 },
		{ 237, 444, 31 },
		{ 237, 445, 31 },
		{ 237, 446, 31 },
		{ 237, 447, 31 },
		{ 237, 448, 31 },
		{ 237, 449, 31 },
		{ 237, 450, 31 },
		{ 237, 451, 31 },
		{ 237, 452, 31 },
		{ 237, 453, 31 },
		{ 237, 454, 31 },
		{ 237, 455, 31 },
		{ 237, 456, 31 },
		{ 237, 457, 31 },
		{ 237, 458, 31 },
		{ 237, 459, 31 },
		{ 237, 460, 31 },
		{ 237, 461, 31 },
		{ 237, 462, 31 },
		{ 237, 463, 31 },
		{ 237, 464, 31 },
		{ 237, 465, 31 },
		{ 237, 466, 31 },
		{ 237, 467, 31 },
		{ 237, 468, 31 },
		{ 237, 469, 31 },
		{ 237, 470, 31 },
		{ 237, 471, 31 },
		{ 237, 472, 31 },
		{ 237, 473, 31 },
		{ 237, 474, 31 },
		{ 237, 475, 31 },
		{ 237, 476, 31 },
		{ 237, 477, 31 },
		{ 237, 478, 31 },
		{ 237, 479, 31 },
		{ 237, 480, 31 },
		{ 237, 481, 31 },
		{ 237, 482, 31 },
		{ 237, 483, 31 },
		{ 237, 484, 31 },
		{ 237, 485, 31 },
		{ 237, 486, 31 },
		{ 237, 487, 31 },
		{ 237, 488, 31 },
		{ 237, 489, 31 },
		{ 237, 490, 31 },
		{ 237, 491, 31 },
		{ 237, 492, 31 },
		{ 237, 493, 31 },
		{ 237, 494, 31 },
		{ 237, 495, 31 },
		{ 237, 496, 31 },
		{ 237, 497, 31 },
		{ 237, 498, 31 },
		{ 237, 499, 31 },
		{ 237, 500, 31 },
		{ 237, 501, 31 },
		{ 237, 502, 31 },
		{ 237, 503, 31 },
		{ 237, 504, 31 },
		{ 237, 505, 31 },
		{ 237, 506, 31 },
		{ 237, 507, 31 },
		{ 237, 508, 31 },
		{ 237, 509, 31 },
		{ 237, 510, 31 },
		{ 237, 511, 31 },
		{ 237, 512, 31 },
		{ 238, 334, 31 },
		{ 238, 335, 31 },
		{ 238, 336, 31 },
		{ 238, 337, 31 },
		{ 238, 338, 31 },
		{ 238, 339, 31 },
		{ 238, 340, 31 },
		{ 238, 341, 31 },
		{ 238, 342, 31 },
		{ 238, 343, 31 },
		{ 238, 344, 31 },
		{ 238, 345, 31 },
		{ 238, 346, 31 },
		{ 238, 347, 31 },
		{ 238, 348, 31 },
		{ 238, 349, 31 },
		{ 238, 350, 31 },
		{ 238, 351, 31 },
		{ 238, 352, 31 },
		{ 238, 353, 31 },
		{ 238, 354, 31 },
		{ 238, 355, 31 },
		{ 238, 356, 31 },
		{ 238, 357, 31 },
		{ 238, 358, 31 },
		{ 238, 359, 31 },
		{ 238, 360, 31 },
		{ 238, 361, 31 },
		{ 238, 362, 31 },
		{ 238, 363, 31 },
		{ 238, 364, 31 },
		{ 238, 365, 31 },
		{ 238, 366, 31 },
		{ 238, 367, 31 },
		{ 238, 368, 31 },
		{ 238, 369, 31 },
		{ 238, 370, 31 },
		{ 238, 371, 31 },
		{ 238, 372, 31 },
		{ 238, 373, 31 },
		{ 238, 374, 31 },
		{ 238, 375, 31 },
		{ 238, 376, 31 },
		{ 238, 377, 31 },
		{ 238, 378, 31 },
		{ 238, 379, 31 },
		{ 238, 380, 31 },
		{ 238, 381, 31 },
		{ 238, 382, 31 },
		{ 238, 383, 31 },
		{ 238, 384, 31 },
		{ 238, 385, 31 },
		{ 238, 386, 31 },
		{ 238, 387, 31 },
		{ 238, 388, 31 },
		{ 238, 389, 31 },
		{ 238, 390, 31 },
		{ 238, 391, 31 },
		{ 238, 392, 31 },
		{ 238, 393, 31 },
		{ 238, 394, 31 },
		{ 238, 395, 31 },
		{ 238, 396, 31 },
		{ 238, 397, 31 },
		{ 238, 398, 31 },
		{ 238, 399, 31 },
		{ 238, 400, 31 },
		{ 238, 401, 31 },
		{ 238, 402, 31 },
		{ 238, 403, 31 },
		{ 238, 404, 31 },
		{ 238, 405, 31 },
		{ 238, 406, 31 },
		{ 238, 407, 31 },
		{ 238, 408, 31 },
		{ 238, 409, 31 },
		{ 238, 410, 31 },
		{ 238, 411, 31 },
		{ 238, 412, 31 },
		{ 238, 413, 31 },
		{ 238, 414, 31 },
		{ 238, 415, 31 },
		{ 238, 416, 31 },
		{ 238, 417, 31 },
		{ 238, 418, 31 },
		{ 238, 419, 31 },
		{ 238, 420, 31 },
		{ 238, 421, 31 },
		{ 238, 422, 31 },
		{ 238, 423, 31 },
		{ 238, 424, 31 },
		{ 238, 425, 31 },
		{ 238, 426, 31 },
		{ 238, 427, 31 },
		{ 238, 428, 31 },
		{ 238, 429, 31 },
		{ 238, 430, 31 },
		{ 238, 431, 31 },
		{ 238, 432, 31 },
		{ 238, 433, 31 },
		{ 238, 434, 31 },
		{ 238, 435, 31 },
		{ 238, 436, 31 },
		{ 238, 437, 31 },
		{ 238, 438, 31 },
		{ 238, 439, 31 },
		{ 238, 440, 31 },
		{ 238, 441, 31 },
		{ 238, 442, 31 },
		{ 238, 443, 31 },
		{ 238, 444, 31 },
		{ 238, 445, 31 },
		{ 238, 446, 31 },
		{ 238, 447, 31 },
		{ 238, 448, 31 },
		{ 238, 449, 31 },
		{ 238, 450, 31 },
		{ 238, 451, 31 },
		{ 238, 452, 31 },
		{ 238, 453, 31 },
		{ 238, 454, 31 },
		{ 238, 455, 31 },
		{ 238, 456, 31 },
		{ 238, 457, 31 },
		{ 238, 458, 31 },
		{ 238, 459, 31 },
		{ 238, 460, 31 },
		{ 238, 461, 31 },
		{ 238, 462, 31 },
		{ 238, 463, 31 },
		{ 238, 464, 31 },
		{ 238, 465, 31 },
		{ 238, 466, 31 },
		{ 238, 467, 31 },
		{ 238, 468, 31 },
		{ 238, 469, 31 },
		{ 238, 470, 31 },
		{ 238, 471, 31 },
		{ 238, 472, 31 },
		{ 238, 473, 31 },
		{ 238, 474, 31 },
		{ 238, 475, 31 },
		{ 238, 476, 31 },
		{ 238, 477, 31 },
		{ 238, 478, 31 },
		{ 238, 479, 31 },
		{ 238, 480, 31 },
		{ 238, 481, 31 },
		{ 238, 482, 31 },
		{ 238, 483, 31 },
		{ 238, 484, 31 },
		{ 238, 485, 31 },
		{ 238, 486, 31 },
		{ 238, 487, 31 },
		{ 238, 488, 31 },
		{ 238, 489, 31 },
		{ 238, 490, 31 },
		{ 238, 491, 31 },
		{ 238, 492, 31 },
		{ 238, 493, 31 },
		{ 238, 494, 31 },
		{ 238, 495, 31 },
		{ 238, 496, 31 },
		{ 238, 497, 31 },
		{ 238, 498, 31 },
		{ 238, 499, 31 },
		{ 238, 500, 31 },
		{ 238, 501, 31 },
		{ 238, 502, 31 },
		{ 238, 503, 31 },
		{ 238, 504, 31 },
		{ 238, 505, 31 },
		{ 238, 506, 31 },
		{ 238, 507, 31 },
		{ 238, 508, 31 },
		{ 238, 509, 31 },
		{ 238, 510, 31 },
		{ 238, 511, 31 },
		{ 238, 512, 31 },
		{ 239, 330, 31 },
		{ 239, 331, 31 },
		{ 239, 332, 31 },
		{ 239, 333, 31 },
		{ 239, 334, 31 },
		{ 239, 335, 31 },
		{ 239, 336, 31 },
		{ 239, 337, 31 },
		{ 239, 338, 31 },
		{ 239, 339, 31 },
		{ 239, 340, 31 },
		{ 239, 341, 31 },
		{ 239, 342, 31 },
		{ 239, 343, 31 },
		{ 239, 344, 31 },
		{ 239, 345, 31 },
		{ 239, 346, 31 },
		{ 239, 347, 31 },
		{ 239, 348, 31 },
		{ 239, 349, 31 },
		{ 239, 350, 31 },
		{ 239, 351, 31 },
		{ 239, 352, 31 },
		{ 239, 353, 31 },
		{ 239, 354, 31 },
		{ 239, 355, 31 },
		{ 239, 356, 31 },
		{ 239, 357, 31 },
		{ 239, 358, 31 },
		{ 239, 359, 31 },
		{ 239, 360, 31 },
		{ 239, 361, 31 },
		{ 239, 362, 31 },
		{ 239, 363, 31 },
		{ 239, 364, 31 },
		{ 239, 365, 31 },
		{ 239, 366, 31 },
		{ 239, 367, 31 },
		{ 239, 368, 31 },
		{ 239, 369, 31 },
		{ 239, 370, 31 },
		{ 239, 371, 31 },
		{ 239, 372, 31 },
		{ 239, 373, 31 },
		{ 239, 374, 31 },
		{ 239, 375, 31 },
		{ 239, 376, 31 },
		{ 239, 377, 31 },
		{ 239, 378, 31 },
		{ 239, 379, 31 },
		{ 239, 380, 31 },
		{ 239, 381, 31 },
		{ 239, 382, 31 },
		{ 239, 383, 31 },
		{ 239, 384, 31 },
		{ 239, 385, 31 },
		{ 239, 386, 31 },
		{ 239, 387, 31 },
		{ 239, 388, 31 },
		{ 239, 389, 31 },
		{ 239, 390, 31 },
		{ 239, 391, 31 },
		{ 239, 392, 31 },
		{ 239, 393, 31 },
		{ 239, 394, 31 },
		{ 239, 395, 31 },
		{ 239, 396, 31 },
		{ 239, 397, 31 },
		{ 239, 398, 31 },
		{ 239, 399, 31 },
		{ 239, 400, 31 },
		{ 239, 401, 31 },
		{ 239, 402, 31 },
		{ 239, 403, 31 },
		{ 239, 404, 31 },
		{ 239, 405, 31 },
		{ 239, 406, 31 },
		{ 239, 407, 31 },
		{ 239, 408, 31 },
		{ 239, 409, 31 },
		{ 239, 410, 31 },
		{ 239, 411, 31 },
		{ 239, 412, 31 },
		{ 239, 413, 31 },
		{ 239, 414, 31 },
		{ 239, 415, 31 },
		{ 239, 416, 31 },
		{ 239, 417, 31 },
		{ 239, 418, 31 },
		{ 239, 419, 31 },
		{ 239, 420, 31 },
		{ 239, 421, 31 },
		{ 239, 422, 31 },
		{ 239, 423, 31 },
		{ 239, 424, 31 },
		{ 239, 425, 31 },
		{ 239, 426, 31 },
		{ 239, 427, 31 },
		{ 239, 428, 31 },
		{ 239, 429, 31 },
		{ 239, 430, 31 },
		{ 239, 431, 31 },
		{ 239, 432, 31 },
		{ 239, 433, 31 },
		{ 239, 434, 31 },
		{ 239, 435, 31 },
		{ 239, 436, 31 },
		{ 239, 437, 31 },
		{ 239, 438, 31 },
		{ 239, 439, 31 },
		{ 239, 440, 31 },
		{ 239, 441, 31 },
		{ 239, 442, 31 },
		{ 239, 443, 31 },
		{ 239, 444, 31 },
		{ 239, 445, 31 },
		{ 239, 446, 31 },
		{ 239, 447, 31 },
		{ 239, 448, 31 },
		{ 239, 449, 31 },
		{ 239, 450, 31 },
		{ 239, 451, 31 },
		{ 239, 452, 31 },
		{ 239, 453, 31 },
		{ 239, 454, 31 },
		{ 239, 455, 31 },
		{ 239, 456, 31 },
		{ 239, 457, 31 },
		{ 239, 458, 31 },
		{ 239, 459, 31 },
		{ 239, 460, 31 },
		{ 239, 461, 31 },
		{ 239, 462, 31 },
		{ 239, 463, 31 },
		{ 239, 464, 31 },
		{ 239, 465, 31 },
		{ 239, 466, 31 },
		{ 239, 467, 31 },
		{ 239, 468, 31 },
		{ 239, 469, 31 },
		{ 239, 470, 31 },
		{ 239, 471, 31 },
		{ 239, 472, 31 },
		{ 239, 473, 31 },
		{ 239, 474, 31 },
		{ 239, 475, 31 },
		{ 239, 476, 31 },
		{ 239, 477, 31 },
		{ 239, 478, 31 },
		{ 239, 479, 31 },
		{ 239, 480, 31 },
		{ 239, 481, 31 },
		{ 239, 482, 31 },
		{ 239, 483, 31 },
		{ 239, 484, 31 },
		{ 239, 485, 31 },
		{ 239, 486, 31 },
		{ 239, 487, 31 },
		{ 239, 488, 31 },
		{ 239, 489, 31 },
		{ 239, 490, 31 },
		{ 239, 491, 31 },
		{ 239, 492, 31 },
		{ 239, 493, 31 },
		{ 239, 494, 31 },
		{ 239, 495, 31 },
		{ 239, 496, 31 },
		{ 239, 497, 31 },
		{ 239, 498, 31 },
		{ 239, 499, 31 },
		{ 239, 500, 31 },
		{ 239, 501, 31 },
		{ 239, 502, 31 },
		{ 239, 503, 31 },
		{ 239, 504, 31 },
		{ 239, 505, 31 },
		{ 239, 506, 31 },
		{ 239, 507, 31 },
		{ 239, 508, 31 },
		{ 239, 509, 31 },
		{ 239, 510, 31 },
		{ 239, 511, 31 },
		{ 239, 512, 31 },
		{ 240, 327, 31 },
		{ 240, 328, 31 },
		{ 240, 329, 31 },
		{ 240, 330, 31 },
		{ 240, 331, 31 },
		{ 240, 332, 31 },
		{ 240, 333, 31 },
		{ 240, 334, 31 },
		{ 240, 335, 31 },
		{ 240, 336, 31 },
		{ 240, 337, 31 },
		{ 240, 338, 31 },
		{ 240, 339, 31 },
		{ 240, 340, 31 },
		{ 240, 341, 31 },
		{ 240, 342, 31 },
		{ 240, 343, 31 },
		{ 240, 344, 31 },
		{ 240, 345, 31 },
		{ 240, 346, 31 },
		{ 240, 347, 31 },
		{ 240, 348, 31 },
		{ 240, 349, 31 },
		{ 240, 350, 31 },
		{ 240, 351, 31 },
		{ 240, 352, 31 },
		{ 240, 353, 31 },
		{ 240, 354, 31 },
		{ 240, 355, 31 },
		{ 240, 356, 31 },
		{ 240, 357, 31 },
		{ 240, 358, 31 },
		{ 240, 359, 31 },
		{ 240, 360, 31 },
		{ 240, 361, 31 },
		{ 240, 362, 31 },
		{ 240, 363, 31 },
		{ 240, 364, 31 },
		{ 240, 365, 31 },
		{ 240, 366, 31 },
		{ 240, 367, 31 },
		{ 240, 368, 31 },
		{ 240, 369, 31 },
		{ 240, 370, 31 },
		{ 240, 371, 31 },
		{ 240, 372, 31 },
		{ 240, 373, 31 },
		{ 240, 374, 31 },
		{ 240, 375, 31 },
		{ 240, 376, 31 },
		{ 240, 377, 31 },
		{ 240, 378, 31 },
		{ 240, 379, 31 },
		{ 240, 380, 31 },
		{ 240, 381, 31 },
		{ 240, 382, 31 },
		{ 240, 383, 31 },
		{ 240, 384, 31 },
		{ 240, 385, 31 },
		{ 240, 386, 31 },
		{ 240, 387, 31 },
		{ 240, 388, 31 },
		{ 240, 389, 31 },
		{ 240, 390, 31 },
		{ 240, 391, 31 },
		{ 240, 392, 31 },
		{ 240, 393, 31 },
		{ 240, 394, 31 },
		{ 240, 395, 31 },
		{ 240, 396, 31 },
		{ 240, 397, 31 },
		{ 240, 398, 31 },
		{ 240, 399, 31 },
		{ 240, 400, 31 },
		{ 240, 401, 31 },
		{ 240, 402, 31 },
		{ 240, 403, 31 },
		{ 240, 404, 31 },
		{ 240, 405, 31 },
		{ 240, 406, 31 },
		{ 240, 407, 31 },
		{ 240, 408, 31 },
		{ 240, 409, 31 },
		{ 240, 410, 31 },
		{ 240, 411, 31 },
		{ 240, 412, 31 },
		{ 240, 413, 31 },
		{ 240, 414, 31 },
		{ 240, 415, 31 },
		{ 240, 416, 31 },
		{ 240, 417, 31 },
		{ 240, 418, 31 },
		{ 240, 419, 31 },
		{ 240, 420, 31 },
		{ 240, 421, 31 },
		{ 240, 422, 31 },
		{ 240, 423, 31 },
		{ 240, 424, 31 },
		{ 240, 425, 31 },
		{ 240, 426, 31 },
		{ 240, 427, 31 },
		{ 240, 428, 31 },
		{ 240, 429, 31 },
		{ 240, 430, 31 },
		{ 240, 431, 31 },
		{ 240, 432, 31 },
		{ 240, 433, 31 },
		{ 240, 434, 31 },
		{ 240, 435, 31 },
		{ 240, 436, 31 },
		{ 240, 437, 31 },
		{ 240, 438, 31 },
		{ 240, 439, 31 },
		{ 240, 440, 31 },
		{ 240, 441, 31 },
		{ 240, 442, 31 },
		{ 240, 443, 31 },
		{ 240, 444, 31 },
		{ 240, 445, 31 },
		{ 240, 446, 31 },
		{ 240, 447, 31 },
		{ 240, 448, 31 },
		{ 240, 449, 31 },
		{ 240, 450, 31 },
		{ 240, 451, 31 },
		{ 240, 452, 31 },
		{ 240, 453, 31 },
		{ 240, 454, 31 },
		{ 240, 455, 31 },
		{ 240, 456, 31 },
		{ 240, 457, 31 },
		{ 240, 458, 31 },
		{ 240, 459, 31 },
		{ 240, 460, 31 },
		{ 240, 461, 31 },
		{ 240, 462, 31 },
		{ 240, 463, 31 },
		{ 240, 464, 31 },
		{ 240, 465, 31 },
		{ 240, 466, 31 },
		{ 240, 467, 31 },
		{ 240, 468, 31 },
		{ 240, 469, 31 },
		{ 240, 470, 31 },
		{ 240, 471, 31 },
		{ 240, 472, 31 },
		{ 240, 473, 31 },
		{ 240, 474, 31 },
		{ 240, 475, 31 },
		{ 240, 476, 31 },
		{ 240, 477, 31 },
		{ 240, 478, 31 },
		{ 240, 479, 31 },
		{ 240, 480, 31 },
		{ 240, 481, 31 },
		{ 240, 482, 31 },
		{ 240, 483, 31 },
		{ 240, 484, 31 },
		{ 240, 485, 31 },
		{ 240, 486, 31 },
		{ 240, 487, 31 },
		{ 240, 488, 31 },
		{ 240, 489, 31 },
		{ 240, 490, 31 },
		{ 240, 491, 31 },
		{ 240, 492, 31 },
		{ 240, 493, 31 },
		{ 240, 494, 31 },
		{ 240, 495, 31 },
		{ 240, 496, 31 },
		{ 240, 497, 31 },
		{ 240, 498, 31 },
		{ 240, 499, 31 },
		{ 240, 500, 31 },
		{ 240, 501, 31 },
		{ 240, 502, 31 },
		{ 240, 503, 31 },
		{ 240, 504, 31 },
		{ 240, 505, 31 },
		{ 240, 506, 31 },
		{ 240, 507, 31 },
		{ 240, 508, 31 },
		{ 240, 509, 31 },
		{ 240, 510, 31 },
		{ 240, 511, 31 },
		{ 240, 512, 31 },
		{ 241, 324, 31 },
		{ 241, 325, 31 },
		{ 241, 326, 31 },
		{ 241, 327, 31 },
		{ 241, 328, 31 },
		{ 241, 329, 31 },
		{ 241, 330, 31 },
		{ 241, 331, 31 },
		{ 241, 332, 31 },
		{ 241, 333, 31 },
		{ 241, 334, 31 },
		{ 241, 335, 31 },
		{ 241, 336, 31 },
		{ 241, 337, 31 },
		{ 241, 338, 31 },
		{ 241, 339, 31 },
		{ 241, 340, 31 },
		{ 241, 341, 31 },
		{ 241, 342, 31 },
		{ 241, 343, 31 },
		{ 241, 344, 31 },
		{ 241, 345, 31 },
		{ 241, 346, 31 },
		{ 241, 347, 31 },
		{ 241, 348, 31 },
		{ 241, 349, 31 },
		{ 241, 350, 31 },
		{ 241, 351, 31 },
		{ 241, 352, 31 },
		{ 241, 353, 31 },
		{ 241, 354, 31 },
		{ 241, 355, 31 },
		{ 241, 356, 31 },
		{ 241, 357, 31 },
		{ 241, 358, 31 },
		{ 241, 359, 31 },
		{ 241, 360, 31 },
		{ 241, 361, 31 },
		{ 241, 362, 31 },
		{ 241, 363, 31 },
		{ 241, 364, 31 },
		{ 241, 365, 31 },
		{ 241, 366, 31 },
		{ 241, 367, 31 },
		{ 241, 368, 31 },
		{ 241, 369, 31 },
		{ 241, 370, 31 },
		{ 241, 371, 31 },
		{ 241, 372, 31 },
		{ 241, 373, 31 },
		{ 241, 374, 31 },
		{ 241, 375, 31 },
		{ 241, 376, 31 },
		{ 241, 377, 31 },
		{ 241, 378, 31 },
		{ 241, 379, 31 },
		{ 241, 380, 31 },
		{ 241, 381, 31 },
		{ 241, 382, 31 },
		{ 241, 383, 31 },
		{ 241, 384, 31 },
		{ 241, 385, 31 },
		{ 241, 386, 31 },
		{ 241, 387, 31 },
		{ 241, 388, 31 },
		{ 241, 389, 31 },
		{ 241, 390, 31 },
		{ 241, 391, 31 },
		{ 241, 392, 31 },
		{ 241, 393, 31 },
		{ 241, 394, 31 },
		{ 241, 395, 31 },
		{ 241, 396, 31 },
		{ 241, 397, 31 },
		{ 241, 398, 31 },
		{ 241, 399, 31 },
		{ 241, 400, 31 },
		{ 241, 401, 31 },
		{ 241, 402, 31 },
		{ 241, 403, 31 },
		{ 241, 404, 31 },
		{ 241, 405, 31 },
		{ 241, 406, 31 },
		{ 241, 407, 31 },
		{ 241, 408, 31 },
		{ 241, 409, 31 },
		{ 241, 410, 31 },
		{ 241, 411, 31 },
		{ 241, 412, 31 },
		{ 241, 413, 31 },
		{ 241, 414, 31 },
		{ 241, 415, 31 },
		{ 241, 416, 31 },
		{ 241, 417, 31 },
		{ 241, 418, 31 },
		{ 241, 419, 31 },
		{ 241, 420, 31 },
		{ 241, 421, 31 },
		{ 241, 422, 31 },
		{ 241, 423, 31 },
		{ 241, 424, 31 },
		{ 241, 425, 31 },
		{ 241, 426, 31 },
		{ 241, 427, 31 },
		{ 241, 428, 31 },
		{ 241, 429, 31 },
		{ 241, 430, 31 },
		{ 241, 431, 31 },
		{ 241, 432, 31 },
		{ 241, 433, 31 },
		{ 241, 434, 31 },
		{ 241, 435, 31 },
		{ 241, 436, 31 },
		{ 241, 437, 31 },
		{ 241, 438, 31 },
		{ 241, 439, 31 },
		{ 241, 440, 31 },
		{ 241, 441, 31 },
		{ 241, 442, 31 },
		{ 241, 443, 31 },
		{ 241, 444, 31 },
		{ 241, 445, 31 },
		{ 241, 446, 31 },
		{ 241, 447, 31 },
		{ 241, 448, 31 },
		{ 241, 449, 31 },
		{ 241, 450, 31 },
		{ 241, 451, 31 },
		{ 241, 452, 31 },
		{ 241, 453, 31 },
		{ 241, 454, 31 },
		{ 241, 455, 31 },
		{ 241, 456, 31 },
		{ 241, 457, 31 },
		{ 241, 458, 31 },
		{ 241, 459, 31 },
		{ 241, 460, 31 },
		{ 241, 461, 31 },
		{ 241, 462, 31 },
		{ 241, 463, 31 },
		{ 241, 464, 31 },
		{ 241, 465, 31 },
		{ 241, 466, 31 },
		{ 241, 467, 31 },
		{ 241, 468, 31 },
		{ 241, 469, 31 },
		{ 241, 470, 31 },
		{ 241, 471, 31 },
		{ 241, 472, 31 },
		{ 241, 473, 31 },
		{ 241, 474, 31 },
		{ 241, 475, 31 },
		{ 241, 476, 31 },
		{ 241, 477, 31 },
		{ 241, 478, 31 },
		{ 241, 479, 31 },
		{ 241, 480, 31 },
		{ 241, 481, 31 },
		{ 241, 482, 31 },
		{ 241, 483, 31 },
		{ 241, 484, 31 },
		{ 241, 485, 31 },
		{ 241, 486, 31 },
		{ 241, 487, 31 },
		{ 241, 488, 31 },
		{ 241, 489, 31 },
		{ 241, 490, 31 },
		{ 241, 491, 31 },
		{ 241, 492, 31 },
		{ 241, 493, 31 },
		{ 241, 494, 31 },
		{ 241, 495, 31 },
		{ 241, 496, 31 },
		{ 241, 497, 31 },
		{ 241, 498, 31 },
		{ 241, 499, 31 },
		{ 241, 500, 31 },
		{ 241, 501, 31 },
		{ 241, 502, 31 },
		{ 241, 503, 31 },
		{ 241, 504, 31 },
		{ 241, 505, 31 },
		{ 241, 506, 31 },
		{ 241, 507, 31 },
		{ 241, 508, 31 },
		{ 241, 509, 31 },
		{ 241, 510, 31 },
		{ 241, 511, 31 },
		{ 241, 512, 31 },
		{ 242, 324, 31 },
		{ 242, 325, 31 },
		{ 242, 326, 31 },
		{ 242, 327, 31 },
		{ 242, 328, 31 },
		{ 242, 329, 31 },
		{ 242, 330, 31 },
		{ 242, 331, 31 },
		{ 242, 332, 31 },
		{ 242, 333, 31 },
		{ 242, 334, 31 },
		{ 242, 335, 31 },
		{ 242, 336, 31 },
		{ 242, 337, 31 },
		{ 242, 338, 31 },
		{ 242, 339, 31 },
		{ 242, 340, 31 },
		{ 242, 341, 31 },
		{ 242, 342, 31 },
		{ 242, 343, 31 },
		{ 242, 344, 31 },
		{ 242, 345, 31 },
		{ 242, 346, 31 },
		{ 242, 347, 31 },
		{ 242, 348, 31 },
		{ 242, 349, 31 },
		{ 242, 350, 31 },
		{ 242, 351, 31 },
		{ 242, 352, 31 },
		{ 242, 353, 31 },
		{ 242, 354, 31 },
		{ 242, 355, 31 },
		{ 242, 356, 31 },
		{ 242, 357, 31 },
		{ 242, 358, 31 },
		{ 242, 359, 31 },
		{ 242, 360, 31 },
		{ 242, 361, 31 },
		{ 242, 362, 31 },
		{ 242, 363, 31 },
		{ 242, 364, 31 },
		{ 242, 365, 31 },
		{ 242, 366, 31 },
		{ 242, 367, 31 },
		{ 242, 368, 31 },
		{ 242, 369, 31 },
		{ 242, 370, 31 },
		{ 242, 371, 31 },
		{ 242, 372, 31 },
		{ 242, 373, 31 },
		{ 242, 374, 31 },
		{ 242, 375, 31 },
		{ 242, 376, 31 },
		{ 242, 377, 31 },
		{ 242, 378, 31 },
		{ 242, 379, 31 },
		{ 242, 380, 31 },
		{ 242, 381, 31 },
		{ 242, 382, 31 },
		{ 242, 383, 31 },
		{ 242, 384, 31 },
		{ 242, 385, 31 },
		{ 242, 386, 31 },
		{ 242, 387, 31 },
		{ 242, 388, 31 },
		{ 242, 389, 31 },
		{ 242, 390, 31 },
		{ 242, 391, 31 },
		{ 242, 392, 31 },
		{ 242, 393, 31 },
		{ 242, 394, 31 },
		{ 242, 395, 31 },
		{ 242, 396, 31 },
		{ 242, 397, 31 },
		{ 242, 398, 31 },
		{ 242, 399, 31 },
		{ 242, 400, 31 },
		{ 242, 401, 31 },
		{ 242, 402, 31 },
		{ 242, 403, 31 },
		{ 242, 404, 31 },
		{ 242, 405, 31 },
		{ 242, 406, 31 },
		{ 242, 407, 31 },
		{ 242, 408, 31 },
		{ 242, 409, 31 },
		{ 242, 410, 31 },
		{ 242, 411, 31 },
		{ 242, 412, 31 },
		{ 242, 413, 31 },
		{ 242, 414, 31 },
		{ 242, 415, 31 },
		{ 242, 416, 31 },
		{ 242, 417, 31 },
		{ 242, 418, 31 },
		{ 242, 419, 31 },
		{ 242, 420, 31 },
		{ 242, 421, 31 },
		{ 242, 422, 31 },
		{ 242, 423, 31 },
		{ 242, 424, 31 },
		{ 242, 425, 31 },
		{ 242, 426, 31 },
		{ 242, 427, 31 },
		{ 242, 428, 31 },
		{ 242, 429, 31 },
		{ 242, 430, 31 },
		{ 242, 431, 31 },
		{ 242, 432, 31 },
		{ 242, 433, 31 },
		{ 242, 434, 31 },
		{ 242, 435, 31 },
		{ 242, 436, 31 },
		{ 242, 437, 31 },
		{ 242, 438, 31 },
		{ 242, 439, 31 },
		{ 242, 440, 31 },
		{ 242, 441, 31 },
		{ 242, 442, 31 },
		{ 242, 443, 31 },
		{ 242, 444, 31 },
		{ 242, 445, 31 },
		{ 242, 446, 31 },
		{ 242, 447, 31 },
		{ 242, 448, 31 },
		{ 242, 449, 31 },
		{ 242, 450, 31 },
		{ 242, 451, 31 },
		{ 242, 452, 31 },
		{ 242, 453, 31 },
		{ 242, 454, 31 },
		{ 242, 455, 31 },
		{ 242, 456, 31 },
		{ 242, 457, 31 },
		{ 242, 458, 31 },
		{ 242, 459, 31 },
		{ 242, 460, 31 },
		{ 242, 461, 31 },
		{ 242, 462, 31 },
		{ 242, 463, 31 },
		{ 242, 464, 31 },
		{ 242, 465, 31 },
		{ 242, 466, 31 },
		{ 242, 467, 31 },
		{ 242, 468, 31 },
		{ 242, 469, 31 },
		{ 242, 470, 31 },
		{ 242, 471, 31 },
		{ 242, 472, 31 },
		{ 242, 473, 31 },
		{ 242, 474, 31 },
		{ 242, 475, 31 },
		{ 242, 476, 31 },
		{ 242, 477, 31 },
		{ 242, 478, 31 },
		{ 242, 479, 31 },
		{ 242, 480, 31 },
		{ 242, 481, 31 },
		{ 242, 482, 31 },
		{ 242, 483, 31 },
		{ 242, 484, 31 },
		{ 242, 485, 31 },
		{ 242, 486, 31 },
		{ 242, 487, 31 },
		{ 242, 488, 31 },
		{ 242, 489, 31 },
		{ 242, 490, 31 },
		{ 242, 491, 31 },
		{ 242, 492, 31 },
		{ 242, 493, 31 },
		{ 242, 494, 31 },
		{ 242, 495, 31 },
		{ 242, 496, 31 },
		{ 242, 497, 31 },
		{ 242, 498, 31 },
		{ 242, 499, 31 },
		{ 242, 500, 31 },
		{ 242, 501, 31 },
		{ 242, 502, 31 },
		{ 242, 503, 31 },
		{ 242, 504, 31 },
		{ 242, 505, 31 },
		{ 242, 506, 31 },
		{ 242, 507, 31 },
		{ 242, 508, 31 },
		{ 242, 509, 31 },
		{ 242, 510, 31 },
		{ 242, 511, 31 },
		{ 242, 512, 31 },
		{ 243, 323, 31 },
		{ 243, 324, 31 },
		{ 243, 325, 31 },
		{ 243, 326, 31 },
		{ 243, 327, 31 },
		{ 243, 328, 31 },
		{ 243, 329, 31 },
		{ 243, 330, 31 },
		{ 243, 331, 31 },
		{ 243, 332, 31 },
		{ 243, 333, 31 },
		{ 243, 334, 31 },
		{ 243, 335, 31 },
		{ 243, 336, 31 },
		{ 243, 337, 31 },
		{ 243, 338, 31 },
		{ 243, 339, 31 },
		{ 243, 340, 31 },
		{ 243, 341, 31 },
		{ 243, 342, 31 },
		{ 243, 343, 31 },
		{ 243, 344, 31 },
		{ 243, 345, 31 },
		{ 243, 346, 31 },
		{ 243, 347, 31 },
		{ 243, 348, 31 },
		{ 243, 349, 31 },
		{ 243, 350, 31 },
		{ 243, 351, 31 },
		{ 243, 352, 31 },
		{ 243, 353, 31 },
		{ 243, 354, 31 },
		{ 243, 355, 31 },
		{ 243, 356, 31 },
		{ 243, 357, 31 },
		{ 243, 358, 31 },
		{ 243, 359, 31 },
		{ 243, 360, 31 },
		{ 243, 361, 31 },
		{ 243, 362, 31 },
		{ 243, 363, 31 },
		{ 243, 364, 31 },
		{ 243, 365, 31 },
		{ 243, 366, 31 },
		{ 243, 367, 31 },
		{ 243, 368, 31 },
		{ 243, 369, 31 },
		{ 243, 370, 31 },
		{ 243, 371, 31 },
		{ 243, 372, 31 },
		{ 243, 373, 31 },
		{ 243, 374, 31 },
		{ 243, 375, 31 },
		{ 243, 376, 31 },
		{ 243, 377, 31 },
		{ 243, 378, 31 },
		{ 243, 379, 31 },
		{ 243, 380, 31 },
		{ 243, 381, 31 },
		{ 243, 382, 31 },
		{ 243, 383, 31 },
		{ 243, 384, 31 },
		{ 243, 385, 31 },
		{ 243, 386, 31 },
		{ 243, 387, 31 },
		{ 243, 388, 31 },
		{ 243, 389, 31 },
		{ 243, 390, 31 },
		{ 243, 391, 31 },
		{ 243, 392, 31 },
		{ 243, 393, 31 },
		{ 243, 394, 31 },
		{ 243, 395, 31 },
		{ 243, 396, 31 },
		{ 243, 397, 31 },
		{ 243, 398, 31 },
		{ 243, 399, 31 },
		{ 243, 400, 31 },
		{ 243, 401, 31 },
		{ 243, 402, 31 },
		{ 243, 403, 31 },
		{ 243, 404, 31 },
		{ 243, 405, 31 },
		{ 243, 406, 31 },
		{ 243, 407, 31 },
		{ 243, 408, 31 },
		{ 243, 409, 31 },
		{ 243, 410, 31 },
		{ 243, 411, 31 },
		{ 243, 412, 31 },
		{ 243, 413, 31 },
		{ 243, 414, 31 },
		{ 243, 415, 31 },
		{ 243, 416, 31 },
		{ 243, 417, 31 },
		{ 243, 418, 31 },
		{ 243, 419, 31 },
		{ 243, 420, 31 },
		{ 243, 421, 31 },
		{ 243, 422, 31 },
		{ 243, 423, 31 },
		{ 243, 424, 31 },
		{ 243, 425, 31 },
		{ 243, 426, 31 },
		{ 243, 427, 31 },
		{ 243, 428, 31 },
		{ 243, 429, 31 },
		{ 243, 430, 31 },
		{ 243, 431, 31 },
		{ 243, 432, 31 },
		{ 243, 433, 31 },
		{ 243, 434, 31 },
		{ 243, 435, 31 },
		{ 243, 436, 31 },
		{ 243, 437, 31 },
		{ 243, 438, 31 },
		{ 243, 439, 31 },
		{ 243, 440, 31 },
		{ 243, 441, 31 },
		{ 243, 442, 31 },
		{ 243, 443, 31 },
		{ 243, 444, 31 },
		{ 243, 445, 31 },
		{ 243, 446, 31 },
		{ 243, 447, 31 },
		{ 243, 448, 31 },
		{ 243, 449, 31 },
		{ 243, 450, 31 },
		{ 243, 451, 31 },
		{ 243, 452, 31 },
		{ 243, 453, 31 },
		{ 243, 454, 31 },
		{ 243, 455, 31 },
		{ 243, 456, 31 },
		{ 243, 457, 31 },
		{ 243, 458, 31 },
		{ 243, 459, 31 },
		{ 243, 460, 31 },
		{ 243, 461, 31 },
		{ 243, 462, 31 },
		{ 243, 463, 31 },
		{ 243, 464, 31 },
		{ 243, 465, 31 },
		{ 243, 466, 31 },
		{ 243, 467, 31 },
		{ 243, 468, 31 },
		{ 243, 469, 31 },
		{ 243, 470, 31 },
		{ 243, 471, 31 },
		{ 243, 472, 31 },
		{ 243, 473, 31 },
		{ 243, 474, 31 },
		{ 243, 475, 31 },
		{ 243, 476, 31 },
		{ 243, 477, 31 },
		{ 243, 478, 31 },
		{ 243, 479, 31 },
		{ 243, 480, 31 },
		{ 243, 481, 31 },
		{ 243, 482, 31 },
		{ 243, 483, 31 },
		{ 243, 484, 31 },
		{ 243, 485, 31 },
		{ 243, 486, 31 },
		{ 243, 487, 31 },
		{ 243, 488, 31 },
		{ 243, 489, 31 },
		{ 243, 490, 31 },
		{ 243, 491, 31 },
		{ 243, 492, 31 },
		{ 243, 493, 31 },
		{ 243, 494, 31 },
		{ 243, 495, 31 },
		{ 243, 496, 31 },
		{ 243, 497, 31 },
		{ 243, 498, 31 },
		{ 243, 499, 31 },
		{ 243, 500, 31 },
		{ 243, 501, 31 },
		{ 243, 502, 31 },
		{ 243, 503, 31 },
		{ 243, 504, 31 },
		{ 243, 505, 31 },
		{ 243, 506, 31 },
		{ 243, 507, 31 },
		{ 243, 508, 31 },
		{ 243, 509, 31 },
		{ 243, 510, 31 },
		{ 243, 511, 31 },
		{ 243, 512, 31 },
		{ 244, 323, 31 },
		{ 244, 324, 31 },
		{ 244, 325, 31 },
		{ 244, 326, 31 },
		{ 244, 327, 31 },
		{ 244, 328, 31 },
		{ 244, 329, 31 },
		{ 244, 330, 31 },
		{ 244, 331, 31 },
		{ 244, 332, 31 },
		{ 244, 333, 31 },
		{ 244, 334, 31 },
		{ 244, 335, 31 },
		{ 244, 336, 31 },
		{ 244, 337, 31 },
		{ 244, 338, 31 },
		{ 244, 339, 31 },
		{ 244, 340, 31 },
		{ 244, 341, 31 },
		{ 244, 342, 31 },
		{ 244, 343, 31 },
		{ 244, 344, 31 },
		{ 244, 345, 31 },
		{ 244, 346, 31 },
		{ 244, 347, 31 },
		{ 244, 348, 31 },
		{ 244, 349, 31 },
		{ 244, 350, 31 },
		{ 244, 351, 31 },
		{ 244, 352, 31 },
		{ 244, 353, 31 },
		{ 244, 354, 31 },
		{ 244, 355, 31 },
		{ 244, 356, 31 },
		{ 244, 357, 31 },
		{ 244, 358, 31 },
		{ 244, 359, 31 },
		{ 244, 360, 31 },
		{ 244, 361, 31 },
		{ 244, 362, 31 },
		{ 244, 363, 31 },
		{ 244, 364, 31 },
		{ 244, 365, 31 },
		{ 244, 366, 31 },
		{ 244, 367, 31 },
		{ 244, 368, 31 },
		{ 244, 369, 31 },
		{ 244, 370, 31 },
		{ 244, 371, 31 },
		{ 244, 372, 31 },
		{ 244, 373, 31 },
		{ 244, 374, 31 },
		{ 244, 375, 31 },
		{ 244, 376, 31 },
		{ 244, 377, 31 },
		{ 244, 378, 31 },
		{ 244, 379, 31 },
		{ 244, 380, 31 },
		{ 244, 381, 31 },
		{ 244, 382, 31 },
		{ 244, 383, 31 },
		{ 244, 384, 31 },
		{ 244, 385, 31 },
		{ 244, 386, 31 },
		{ 244, 387, 31 },
		{ 244, 388, 31 },
		{ 244, 389, 31 },
		{ 244, 390, 31 },
		{ 244, 391, 31 },
		{ 244, 392, 31 },
		{ 244, 393, 31 },
		{ 244, 394, 31 },
		{ 244, 395, 31 },
		{ 244, 396, 31 },
		{ 244, 397, 31 },
		{ 244, 398, 31 },
		{ 244, 399, 31 },
		{ 244, 400, 31 },
		{ 244, 401, 31 },
		{ 244, 402, 31 },
		{ 244, 403, 31 },
		{ 244, 404, 31 },
		{ 244, 405, 31 },
		{ 244, 406, 31 },
		{ 244, 407, 31 },
		{ 244, 408, 31 },
		{ 244, 409, 31 },
		{ 244, 410, 31 },
		{ 244, 411, 31 },
		{ 244, 412, 31 },
		{ 244, 413, 31 },
		{ 244, 414, 31 },
		{ 244, 415, 31 },
		{ 244, 416, 31 },
		{ 244, 417, 31 },
		{ 244, 418, 31 },
		{ 244, 419, 31 },
		{ 244, 420, 31 },
		{ 244, 421, 31 },
		{ 244, 422, 31 },
		{ 244, 423, 31 },
		{ 244, 424, 31 },
		{ 244, 425, 31 },
		{ 244, 426, 31 },
		{ 244, 427, 31 },
		{ 244, 428, 31 },
		{ 244, 429, 31 },
		{ 244, 430, 31 },
		{ 244, 431, 31 },
		{ 244, 432, 31 },
		{ 244, 433, 31 },
		{ 244, 434, 31 },
		{ 244, 435, 31 },
		{ 244, 436, 31 },
		{ 244, 437, 31 },
		{ 244, 438, 31 },
		{ 244, 439, 31 },
		{ 244, 440, 31 },
		{ 244, 441, 31 },
		{ 244, 442, 31 },
		{ 244, 443, 31 },
		{ 244, 444, 31 },
		{ 244, 445, 31 },
		{ 244, 446, 31 },
		{ 244, 447, 31 },
		{ 244, 448, 31 },
		{ 244, 449, 31 },
		{ 244, 450, 31 },
		{ 244, 451, 31 },
		{ 244, 452, 31 },
		{ 244, 453, 31 },
		{ 244, 454, 31 },
		{ 244, 455, 31 },
		{ 244, 456, 31 },
		{ 244, 457, 31 },
		{ 244, 458, 31 },
		{ 244, 459, 31 },
		{ 244, 460, 31 },
		{ 244, 461, 31 },
		{ 244, 462, 31 },
		{ 244, 463, 31 },
		{ 244, 464, 31 },
		{ 244, 465, 31 },
		{ 244, 466, 31 },
		{ 244, 467, 31 },
		{ 244, 468, 31 },
		{ 244, 469, 31 },
		{ 244, 470, 31 },
		{ 244, 471, 31 },
		{ 244, 472, 31 },
		{ 244, 473, 31 },
		{ 244, 474, 31 },
		{ 244, 475, 31 },
		{ 244, 476, 31 },
		{ 244, 477, 31 },
		{ 244, 478, 31 },
		{ 244, 479, 31 },
		{ 244, 480, 31 },
		{ 244, 481, 31 },
		{ 244, 482, 31 },
		{ 244, 483, 31 },
		{ 244, 484, 31 },
		{ 244, 485, 31 },
		{ 244, 486, 31 },
		{ 244, 487, 31 },
		{ 244, 488, 31 },
		{ 244, 489, 31 },
		{ 244, 490, 31 },
		{ 244, 491, 31 },
		{ 244, 492, 31 },
		{ 244, 493, 31 },
		{ 244, 494, 31 },
		{ 244, 495, 31 },
		{ 244, 496, 31 },
		{ 244, 497, 31 },
		{ 244, 498, 31 },
		{ 244, 499, 31 },
		{ 244, 500, 31 },
		{ 244, 501, 31 },
		{ 244, 502, 31 },
		{ 244, 503, 31 },
		{ 244, 504, 31 },
		{ 244, 505, 31 },
		{ 244, 506, 31 },
		{ 244, 507, 31 },
		{ 244, 508, 31 },
		{ 244, 509, 31 },
		{ 244, 510, 31 },
		{ 244, 511, 31 },
		{ 244, 512, 31 },
		{ 245, 323, 31 },
		{ 245, 324, 31 },
		{ 245, 325, 31 },
		{ 245, 326, 31 },
		{ 245, 327, 31 },
		{ 245, 328, 31 },
		{ 245, 329, 31 },
		{ 245, 330, 31 },
		{ 245, 331, 31 },
		{ 245, 332, 31 },
		{ 245, 333, 31 },
		{ 245, 334, 31 },
		{ 245, 335, 31 },
		{ 245, 336, 31 },
		{ 245, 337, 31 },
		{ 245, 338, 31 },
		{ 245, 339, 31 },
		{ 245, 340, 31 },
		{ 245, 341, 31 },
		{ 245, 342, 31 },
		{ 245, 343, 31 },
		{ 245, 344, 31 },
		{ 245, 345, 31 },
		{ 245, 346, 31 },
		{ 245, 347, 31 },
		{ 245, 348, 31 },
		{ 245, 349, 31 },
		{ 245, 350, 31 },
		{ 245, 351, 31 },
		{ 245, 352, 31 },
		{ 245, 353, 31 },
		{ 245, 354, 31 },
		{ 245, 355, 31 },
		{ 245, 356, 31 },
		{ 245, 357, 31 },
		{ 245, 358, 31 },
		{ 245, 359, 31 },
		{ 245, 360, 31 },
		{ 245, 361, 31 },
		{ 245, 362, 31 },
		{ 245, 363, 31 },
		{ 245, 364, 31 },
		{ 245, 365, 31 },
		{ 245, 366, 31 },
		{ 245, 367, 31 },
		{ 245, 368, 31 },
		{ 245, 369, 31 },
		{ 245, 370, 31 },
		{ 245, 371, 31 },
		{ 245, 372, 31 },
		{ 245, 373, 31 },
		{ 245, 374, 31 },
		{ 245, 375, 31 },
		{ 245, 376, 31 },
		{ 245, 377, 31 },
		{ 245, 378, 31 },
		{ 245, 379, 31 },
		{ 245, 380, 31 },
		{ 245, 381, 31 },
		{ 245, 382, 31 },
		{ 245, 383, 31 },
		{ 245, 384, 31 },
		{ 245, 385, 31 },
		{ 245, 386, 31 },
		{ 245, 387, 31 },
		{ 245, 388, 31 },
		{ 245, 389, 31 },
		{ 245, 390, 31 },
		{ 245, 391, 31 },
		{ 245, 392, 31 },
		{ 245, 393, 31 },
		{ 245, 394, 31 },
		{ 245, 395, 31 },
		{ 245, 396, 31 },
		{ 245, 397, 31 },
		{ 245, 398, 31 },
		{ 245, 399, 31 },
		{ 245, 400, 31 },
		{ 245, 401, 31 },
		{ 245, 402, 31 },
		{ 245, 403, 31 },
		{ 245, 404, 31 },
		{ 245, 405, 31 },
		{ 245, 406, 31 },
		{ 245, 407, 31 },
		{ 245, 408, 31 },
		{ 245, 409, 31 },
		{ 245, 410, 31 },
		{ 245, 411, 31 },
		{ 245, 412, 31 },
		{ 245, 413, 31 },
		{ 245, 414, 31 },
		{ 245, 415, 31 },
		{ 245, 416, 31 },
		{ 245, 417, 31 },
		{ 245, 418, 31 },
		{ 245, 419, 31 },
		{ 245, 420, 31 },
		{ 245, 421, 31 },
		{ 245, 422, 31 },
		{ 245, 423, 31 },
		{ 245, 424, 31 },
		{ 245, 425, 31 },
		{ 245, 426, 31 },
		{ 245, 427, 31 },
		{ 245, 428, 31 },
		{ 245, 429, 31 },
		{ 245, 430, 31 },
		{ 245, 431, 31 },
		{ 245, 432, 31 },
		{ 245, 433, 31 },
		{ 245, 434, 31 },
		{ 245, 435, 31 },
		{ 245, 436, 31 },
		{ 245, 437, 31 },
		{ 245, 438, 31 },
		{ 245, 439, 31 },
		{ 245, 440, 31 },
		{ 245, 441, 31 },
		{ 245, 442, 31 },
		{ 245, 443, 31 },
		{ 245, 444, 31 },
		{ 245, 445, 31 },
		{ 245, 446, 31 },
		{ 245, 447, 31 },
		{ 245, 448, 31 },
		{ 245, 449, 31 },
		{ 245, 450, 31 },
		{ 245, 451, 31 },
		{ 245, 452, 31 },
		{ 245, 453, 31 },
		{ 245, 454, 31 },
		{ 245, 455, 31 },
		{ 245, 456, 31 },
		{ 245, 457, 31 },
		{ 245, 458, 31 },
		{ 245, 459, 31 },
		{ 245, 460, 31 },
		{ 245, 461, 31 },
		{ 245, 462, 31 },
		{ 245, 463, 31 },
		{ 245, 464, 31 },
		{ 245, 465, 31 },
		{ 245, 466, 31 },
		{ 245, 467, 31 },
		{ 245, 468, 31 },
		{ 245, 469, 31 },
		{ 245, 470, 31 },
		{ 245, 471, 31 },
		{ 245, 472, 31 },
		{ 245, 473, 31 },
		{ 245, 474, 31 },
		{ 245, 475, 31 },
		{ 245, 476, 31 },
		{ 245, 477, 31 },
		{ 245, 478, 31 },
		{ 245, 479, 31 },
		{ 245, 480, 31 },
		{ 245, 481, 31 },
		{ 245, 482, 31 },
		{ 245, 483, 31 },
		{ 245, 484, 31 },
		{ 245, 485, 31 },
		{ 245, 486, 31 },
		{ 245, 487, 31 },
		{ 245, 488, 31 },
		{ 245, 489, 31 },
		{ 245, 490, 31 },
		{ 245, 491, 31 },
		{ 245, 492, 31 },
		{ 245, 493, 31 },
		{ 245, 494, 31 },
		{ 245, 495, 31 },
		{ 245, 496, 31 },
		{ 245, 497, 31 },
		{ 245, 498, 31 },
		{ 245, 499, 31 },
		{ 245, 500, 31 },
		{ 245, 501, 31 },
		{ 245, 502, 31 },
		{ 245, 503, 31 },
		{ 245, 504, 31 },
		{ 245, 505, 31 },
		{ 245, 506, 31 },
		{ 245, 507, 31 },
		{ 245, 508, 31 },
		{ 245, 509, 31 },
		{ 245, 510, 31 },
		{ 245, 511, 31 },
		{ 245, 512, 31 },
		{ 246, 323, 31 },
		{ 246, 324, 31 },
		{ 246, 325, 31 },
		{ 246, 326, 31 },
		{ 246, 327, 31 },
		{ 246, 328, 31 },
		{ 246, 329, 31 },
		{ 246, 330, 31 },
		{ 246, 331, 31 },
		{ 246, 332, 31 },
		{ 246, 333, 31 },
		{ 246, 334, 31 },
		{ 246, 335, 31 },
		{ 246, 336, 31 },
		{ 246, 337, 31 },
		{ 246, 338, 31 },
		{ 246, 339, 31 },
		{ 246, 340, 31 },
		{ 246, 341, 31 },
		{ 246, 342, 31 },
		{ 246, 343, 31 },
		{ 246, 344, 31 },
		{ 246, 345, 31 },
		{ 246, 346, 31 },
		{ 246, 347, 31 },
		{ 246, 348, 31 },
		{ 246, 349, 31 },
		{ 246, 350, 31 },
		{ 246, 351, 31 },
		{ 246, 352, 31 },
		{ 246, 353, 31 },
		{ 246, 354, 31 },
		{ 246, 355, 31 },
		{ 246, 356, 31 },
		{ 246, 357, 31 },
		{ 246, 358, 31 },
		{ 246, 359, 31 },
		{ 246, 360, 31 },
		{ 246, 361, 31 },
		{ 246, 362, 31 },
		{ 246, 363, 31 },
		{ 246, 364, 31 },
		{ 246, 365, 31 },
		{ 246, 366, 31 },
		{ 246, 367, 31 },
		{ 246, 368, 31 },
		{ 246, 369, 31 },
		{ 246, 370, 31 },
		{ 246, 371, 31 },
		{ 246, 372, 31 },
		{ 246, 373, 31 },
		{ 246, 374, 31 },
		{ 246, 375, 31 },
		{ 246, 376, 31 },
		{ 246, 377, 31 },
		{ 246, 378, 31 },
		{ 246, 379, 31 },
		{ 246, 380, 31 },
		{ 246, 381, 31 },
		{ 246, 382, 31 },
		{ 246, 383, 31 },
		{ 246, 384, 31 },
		{ 246, 385, 31 },
		{ 246, 386, 31 },
		{ 246, 387, 31 },
		{ 246, 388, 31 },
		{ 246, 389, 31 },
		{ 246, 390, 31 },
		{ 246, 391, 31 },
		{ 246, 392, 31 },
		{ 246, 393, 31 },
		{ 246, 394, 31 },
		{ 246, 395, 31 },
		{ 246, 396, 31 },
		{ 246, 397, 31 },
		{ 246, 398, 31 },
		{ 246, 399, 31 },
		{ 246, 400, 31 },
		{ 246, 401, 31 },
		{ 246, 402, 31 },
		{ 246, 403, 31 },
		{ 246, 404, 31 },
		{ 246, 405, 31 },
		{ 246, 406, 31 },
		{ 246, 407, 31 },
		{ 246, 408, 31 },
		{ 246, 409, 31 },
		{ 246, 410, 31 },
		{ 246, 411, 31 },
		{ 246, 412, 31 },
		{ 246, 413, 31 },
		{ 246, 414, 31 },
		{ 246, 415, 31 },
		{ 246, 416, 31 },
		{ 246, 417, 31 },
		{ 246, 418, 31 },
		{ 246, 419, 31 },
		{ 246, 420, 31 },
		{ 246, 421, 31 },
		{ 246, 422, 31 },
		{ 246, 423, 31 },
		{ 246, 424, 31 },
		{ 246, 425, 31 },
		{ 246, 426, 31 },
		{ 246, 427, 31 },
		{ 246, 428, 31 },
		{ 246, 429, 31 },
		{ 246, 430, 31 },
		{ 246, 431, 31 },
		{ 246, 432, 31 },
		{ 246, 433, 31 },
		{ 246, 434, 31 },
		{ 246, 435, 31 },
		{ 246, 436, 31 },
		{ 246, 437, 31 },
		{ 246, 438, 31 },
		{ 246, 439, 31 },
		{ 246, 440, 31 },
		{ 246, 441, 31 },
		{ 246, 442, 31 },
		{ 246, 443, 31 },
		{ 246, 444, 31 },
		{ 246, 445, 31 },
		{ 246, 446, 31 },
		{ 246, 447, 31 },
		{ 246, 448, 31 },
		{ 246, 449, 31 },
		{ 246, 450, 31 },
		{ 246, 451, 31 },
		{ 246, 452, 31 },
		{ 246, 453, 31 },
		{ 246, 454, 31 },
		{ 246, 455, 31 },
		{ 246, 456, 31 },
		{ 246, 457, 31 },
		{ 246, 458, 31 },
		{ 246, 459, 31 },
		{ 246, 460, 31 },
		{ 246, 461, 31 },
		{ 246, 462, 31 },
		{ 246, 463, 31 },
		{ 246, 464, 31 },
		{ 246, 465, 31 },
		{ 246, 466, 31 },
		{ 246, 467, 31 },
		{ 246, 468, 31 },
		{ 246, 469, 31 },
		{ 246, 470, 31 },
		{ 246, 471, 31 },
		{ 246, 472, 31 },
		{ 246, 473, 31 },
		{ 246, 474, 31 },
		{ 246, 475, 31 },
		{ 246, 476, 31 },
		{ 246, 477, 31 },
		{ 246, 478, 31 },
		{ 246, 479, 31 },
		{ 246, 480, 31 },
		{ 246, 481, 31 },
		{ 246, 482, 31 },
		{ 246, 483, 31 },
		{ 246, 484, 31 },
		{ 246, 485, 31 },
		{ 246, 486, 31 },
		{ 246, 487, 31 },
		{ 246, 488, 31 },
		{ 246, 489, 31 },
		{ 246, 490, 31 },
		{ 246, 491, 31 },
		{ 246, 492, 31 },
		{ 246, 493, 31 },
		{ 246, 494, 31 },
		{ 246, 495, 31 },
		{ 246, 496, 31 },
		{ 246, 497, 31 },
		{ 246, 498, 31 },
		{ 246, 499, 31 },
		{ 246, 500, 31 },
		{ 246, 501, 31 },
		{ 246, 502, 31 },
		{ 246, 503, 31 },
		{ 246, 504, 31 },
		{ 246, 505, 31 },
		{ 246, 506, 31 },
		{ 246, 507, 31 },
		{ 246, 508, 31 },
		{ 246, 509, 31 },
		{ 246, 510, 31 },
		{ 246, 511, 31 },
		{ 246, 512, 31 },
		{ 247, 323, 31 },
		{ 247, 324, 31 },
		{ 247, 325, 31 },
		{ 247, 326, 31 },
		{ 247, 327, 31 },
		{ 247, 328, 31 },
		{ 247, 329, 31 },
		{ 247, 330, 31 },
		{ 247, 331, 31 },
		{ 247, 332, 31 },
		{ 247, 333, 31 },
		{ 247, 334, 31 },
		{ 247, 335, 31 },
		{ 247, 336, 31 },
		{ 247, 337, 31 },
		{ 247, 338, 31 },
		{ 247, 339, 31 },
		{ 247, 340, 31 },
		{ 247, 341, 31 },
		{ 247, 342, 31 },
		{ 247, 343, 31 },
		{ 247, 344, 31 },
		{ 247, 345, 31 },
		{ 247, 346, 31 },
		{ 247, 347, 31 },
		{ 247, 348, 31 },
		{ 247, 349, 31 },
		{ 247, 350, 31 },
		{ 247, 351, 31 },
		{ 247, 352, 31 },
		{ 247, 353, 31 },
		{ 247, 354, 31 },
		{ 247, 355, 31 },
		{ 247, 356, 31 },
		{ 247, 357, 31 },
		{ 247, 358, 31 },
		{ 247, 359, 31 },
		{ 247, 360, 31 },
		{ 247, 361, 31 },
		{ 247, 362, 31 },
		{ 247, 363, 31 },
		{ 247, 364, 31 },
		{ 247, 365, 31 },
		{ 247, 366, 31 },
		{ 247, 367, 31 },
		{ 247, 368, 31 },
		{ 247, 369, 31 },
		{ 247, 370, 31 },
		{ 247, 371, 31 },
		{ 247, 372, 31 },
		{ 247, 373, 31 },
		{ 247, 374, 31 },
		{ 247, 375, 31 },
		{ 247, 376, 31 },
		{ 247, 377, 31 },
		{ 247, 378, 31 },
		{ 247, 379, 31 },
		{ 247, 380, 31 },
		{ 247, 381, 31 },
		{ 247, 382, 31 },
		{ 247, 383, 31 },
		{ 247, 384, 31 },
		{ 247, 385, 31 },
		{ 247, 386, 31 },
		{ 247, 387, 31 },
		{ 247, 388, 31 },
		{ 247, 389, 31 },
		{ 247, 390, 31 },
		{ 247, 391, 31 },
		{ 247, 392, 31 },
		{ 247, 393, 31 },
		{ 247, 394, 31 },
		{ 247, 395, 31 },
		{ 247, 396, 31 },
		{ 247, 397, 31 },
		{ 247, 398, 31 },
		{ 247, 399, 31 },
		{ 247, 400, 31 },
		{ 247, 401, 31 },
		{ 247, 402, 31 },
		{ 247, 403, 31 },
		{ 247, 404, 31 },
		{ 247, 405, 31 },
		{ 247, 406, 31 },
		{ 247, 407, 31 },
		{ 247, 408, 31 },
		{ 247, 409, 31 },
		{ 247, 410, 31 },
		{ 247, 411, 31 },
		{ 247, 412, 31 },
		{ 247, 413, 31 },
		{ 247, 414, 31 },
		{ 247, 415, 31 },
		{ 247, 416, 31 },
		{ 247, 417, 31 },
		{ 247, 418, 31 },
		{ 247, 419, 31 },
		{ 247, 420, 31 },
		{ 247, 421, 31 },
		{ 247, 422, 31 },
		{ 247, 423, 31 },
		{ 247, 424, 31 },
		{ 247, 425, 31 },
		{ 247, 426, 31 },
		{ 247, 427, 31 },
		{ 247, 428, 31 },
		{ 247, 429, 31 },
		{ 247, 430, 31 },
		{ 247, 431, 31 },
		{ 247, 432, 31 },
		{ 247, 433, 31 },
		{ 247, 434, 31 },
		{ 247, 435, 31 },
		{ 247, 436, 31 },
		{ 247, 437, 31 },
		{ 247, 438, 31 },
		{ 247, 439, 31 },
		{ 247, 440, 31 },
		{ 247, 441, 31 },
		{ 247, 442, 31 },
		{ 247, 443, 31 },
		{ 247, 444, 31 },
		{ 247, 445, 31 },
		{ 247, 446, 31 },
		{ 247, 447, 31 },
		{ 247, 448, 31 },
		{ 247, 449, 31 },
		{ 247, 450, 31 },
		{ 247, 451, 31 },
		{ 247, 452, 31 },
		{ 247, 453, 31 },
		{ 247, 454, 31 },
		{ 247, 455, 31 },
		{ 247, 456, 31 },
		{ 247, 457, 31 },
		{ 247, 458, 31 },
		{ 247, 459, 31 },
		{ 247, 460, 31 },
		{ 247, 461, 31 },
		{ 247, 462, 31 },
		{ 247, 463, 31 },
		{ 247, 464, 31 },
		{ 247, 465, 31 },
		{ 247, 466, 31 },
		{ 247, 467, 31 },
		{ 247, 468, 31 },
		{ 247, 469, 31 },
		{ 247, 470, 31 },
		{ 247, 471, 31 },
		{ 247, 472, 31 },
		{ 247, 473, 31 },
		{ 247, 474, 31 },
		{ 247, 475, 31 },
		{ 247, 476, 31 },
		{ 247, 477, 31 },
		{ 247, 478, 31 },
		{ 247, 479, 31 },
		{ 247, 480, 31 },
		{ 247, 481, 31 },
		{ 247, 482, 31 },
		{ 247, 483, 31 },
		{ 247, 484, 31 },
		{ 247, 485, 31 },
		{ 247, 486, 31 },
		{ 247, 487, 31 },
		{ 247, 488, 31 },
		{ 247, 489, 31 },
		{ 247, 490, 31 },
		{ 247, 491, 31 },
		{ 247, 492, 31 },
		{ 247, 493, 31 },
		{ 247, 494, 31 },
		{ 247, 495, 31 },
		{ 247, 496, 31 },
		{ 247, 497, 31 },
		{ 247, 498, 31 },
		{ 247, 499, 31 },
		{ 247, 500, 31 },
		{ 247, 501, 31 },
		{ 247, 502, 31 },
		{ 247, 503, 31 },
		{ 247, 504, 31 },
		{ 247, 505, 31 },
		{ 247, 506, 31 },
		{ 247, 507, 31 },
		{ 247, 508, 31 },
		{ 247, 509, 31 },
		{ 247, 510, 31 },
		{ 247, 511, 31 },
		{ 247, 512, 31 },
		{ 248, 323, 31 },
		{ 248, 324, 31 },
		{ 248, 325, 31 },
		{ 248, 326, 31 },
		{ 248, 327, 31 },
		{ 248, 328, 31 },
		{ 248, 329, 31 },
		{ 248, 330, 31 },
		{ 248, 331, 31 },
		{ 248, 332, 31 },
		{ 248, 333, 31 },
		{ 248, 334, 31 },
		{ 248, 335, 31 },
		{ 248, 336, 31 },
		{ 248, 337, 31 },
		{ 248, 338, 31 },
		{ 248, 339, 31 },
		{ 248, 340, 31 },
		{ 248, 341, 31 },
		{ 248, 342, 31 },
		{ 248, 343, 31 },
		{ 248, 344, 31 },
		{ 248, 345, 31 },
		{ 248, 346, 31 },
		{ 248, 347, 31 },
		{ 248, 348, 31 },
		{ 248, 349, 31 },
		{ 248, 350, 31 },
		{ 248, 351, 31 },
		{ 248, 352, 31 },
		{ 248, 353, 31 },
		{ 248, 354, 31 },
		{ 248, 355, 31 },
		{ 248, 356, 31 },
		{ 248, 357, 31 },
		{ 248, 358, 31 },
		{ 248, 359, 31 },
		{ 248, 360, 31 },
		{ 248, 361, 31 },
		{ 248, 362, 31 },
		{ 248, 363, 31 },
		{ 248, 364, 31 },
		{ 248, 365, 31 },
		{ 248, 366, 31 },
		{ 248, 367, 31 },
		{ 248, 368, 31 },
		{ 248, 369, 31 },
		{ 248, 370, 31 },
		{ 248, 371, 31 },
		{ 248, 372, 31 },
		{ 248, 373, 31 },
		{ 248, 374, 31 },
		{ 248, 375, 31 },
		{ 248, 376, 31 },
		{ 248, 377, 31 },
		{ 248, 378, 31 },
		{ 248, 379, 31 },
		{ 248, 380, 31 },
		{ 248, 381, 31 },
		{ 248, 382, 31 },
		{ 248, 383, 31 },
		{ 248, 384, 31 },
		{ 248, 385, 31 },
		{ 248, 386, 31 },
		{ 248, 387, 31 },
		{ 248, 388, 31 },
		{ 248, 389, 31 },
		{ 248, 390, 31 },
		{ 248, 391, 31 },
		{ 248, 392, 31 },
		{ 248, 393, 31 },
		{ 248, 394, 31 },
		{ 248, 395, 31 },
		{ 248, 396, 31 },
		{ 248, 397, 31 },
		{ 248, 398, 31 },
		{ 248, 399, 31 },
		{ 248, 400, 31 },
		{ 248, 401, 31 },
		{ 248, 402, 31 },
		{ 248, 403, 31 },
		{ 248, 404, 31 },
		{ 248, 405, 31 },
		{ 248, 406, 31 },
		{ 248, 407, 31 },
		{ 248, 408, 31 },
		{ 248, 409, 31 },
		{ 248, 410, 31 },
		{ 248, 411, 31 },
		{ 248, 412, 31 },
		{ 248, 413, 31 },
		{ 248, 414, 31 },
		{ 248, 415, 31 },
		{ 248, 416, 31 },
		{ 248, 417, 31 },
		{ 248, 418, 31 },
		{ 248, 419, 31 },
		{ 248, 420, 31 },
		{ 248, 421, 31 },
		{ 248, 422, 31 },
		{ 248, 423, 31 },
		{ 248, 424, 31 },
		{ 248, 425, 31 },
		{ 248, 426, 31 },
		{ 248, 427, 31 },
		{ 248, 428, 31 },
		{ 248, 429, 31 },
		{ 248, 430, 31 },
		{ 248, 431, 31 },
		{ 248, 432, 31 },
		{ 248, 433, 31 },
		{ 248, 434, 31 },
		{ 248, 435, 31 },
		{ 248, 436, 31 },
		{ 248, 437, 31 },
		{ 248, 438, 31 },
		{ 248, 439, 31 },
		{ 248, 440, 31 },
		{ 248, 441, 31 },
		{ 248, 442, 31 },
		{ 248, 443, 31 },
		{ 248, 444, 31 },
		{ 248, 445, 31 },
		{ 248, 446, 31 },
		{ 248, 447, 31 },
		{ 248, 448, 31 },
		{ 248, 449, 31 },
		{ 248, 450, 31 },
		{ 248, 451, 31 },
		{ 248, 452, 31 },
		{ 248, 453, 31 },
		{ 248, 454, 31 },
		{ 248, 455, 31 },
		{ 248, 456, 31 },
		{ 248, 457, 31 },
		{ 248, 458, 31 },
		{ 248, 459, 31 },
		{ 248, 460, 31 },
		{ 248, 461, 31 },
		{ 248, 462, 31 },
		{ 248, 463, 31 },
		{ 248, 464, 31 },
		{ 248, 465, 31 },
		{ 248, 466, 31 },
		{ 248, 467, 31 },
		{ 248, 468, 31 },
		{ 248, 469, 31 },
		{ 248, 470, 31 },
		{ 248, 471, 31 },
		{ 248, 472, 31 },
		{ 248, 473, 31 },
		{ 248, 474, 31 },
		{ 248, 475, 31 },
		{ 248, 476, 31 },
		{ 248, 477, 31 },
		{ 248, 478, 31 },
		{ 248, 479, 31 },
		{ 248, 480, 31 },
		{ 248, 481, 31 },
		{ 248, 482, 31 },
		{ 248, 483, 31 },
		{ 248, 484, 31 },
		{ 248, 485, 31 },
		{ 248, 486, 31 },
		{ 248, 487, 31 },
		{ 248, 488, 31 },
		{ 248, 489, 31 },
		{ 248, 490, 31 },
		{ 248, 491, 31 },
		{ 248, 492, 31 },
		{ 248, 493, 31 },
		{ 248, 494, 31 },
		{ 248, 495, 31 },
		{ 248, 496, 31 },
		{ 248, 497, 31 },
		{ 248, 498, 31 },
		{ 248, 499, 31 },
		{ 248, 500, 31 },
		{ 248, 501, 31 },
		{ 248, 502, 31 },
		{ 248, 503, 31 },
		{ 248, 504, 31 },
		{ 248, 505, 31 },
		{ 248, 506, 31 },
		{ 248, 507, 31 },
		{ 248, 508, 31 },
		{ 248, 509, 31 },
		{ 248, 510, 31 },
		{ 248, 511, 31 },
		{ 248, 512, 31 },
		{ 249, 323, 31 },
		{ 249, 324, 31 },
		{ 249, 325, 31 },
		{ 249, 326, 31 },
		{ 249, 327, 31 },
		{ 249, 328, 31 },
		{ 249, 329, 31 },
		{ 249, 330, 31 },
		{ 249, 331, 31 },
		{ 249, 332, 31 },
		{ 249, 333, 31 },
		{ 249, 334, 31 },
		{ 249, 335, 31 },
		{ 249, 336, 31 },
		{ 249, 337, 31 },
		{ 249, 338, 31 },
		{ 249, 339, 31 },
		{ 249, 340, 31 },
		{ 249, 341, 31 },
		{ 249, 342, 31 },
		{ 249, 343, 31 },
		{ 249, 344, 31 },
		{ 249, 345, 31 },
		{ 249, 346, 31 },
		{ 249, 347, 31 },
		{ 249, 348, 31 },
		{ 249, 349, 31 },
		{ 249, 350, 31 },
		{ 249, 351, 31 },
		{ 249, 352, 31 },
		{ 249, 353, 31 },
		{ 249, 354, 31 },
		{ 249, 355, 31 },
		{ 249, 356, 31 },
		{ 249, 357, 31 },
		{ 249, 358, 31 },
		{ 249, 359, 31 },
		{ 249, 360, 31 },
		{ 249, 361, 31 },
		{ 249, 362, 31 },
		{ 249, 363, 31 },
		{ 249, 364, 31 },
		{ 249, 365, 31 },
		{ 249, 366, 31 },
		{ 249, 367, 31 },
		{ 249, 368, 31 },
		{ 249, 369, 31 },
		{ 249, 370, 31 },
		{ 249, 371, 31 },
		{ 249, 372, 31 },
		{ 249, 373, 31 },
		{ 249, 374, 31 },
		{ 249, 375, 31 },
		{ 249, 376, 31 },
		{ 249, 377, 31 },
		{ 249, 378, 31 },
		{ 249, 379, 31 },
		{ 249, 380, 31 },
		{ 249, 381, 31 },
		{ 249, 382, 31 },
		{ 249, 383, 31 },
		{ 249, 384, 31 },
		{ 249, 385, 31 },
		{ 249, 386, 31 },
		{ 249, 387, 31 },
		{ 249, 388, 31 },
		{ 249, 389, 31 },
		{ 249, 390, 31 },
		{ 249, 391, 31 },
		{ 249, 392, 31 },
		{ 249, 393, 31 },
		{ 249, 394, 31 },
		{ 249, 395, 31 },
		{ 249, 396, 31 },
		{ 249, 397, 31 },
		{ 249, 398, 31 },
		{ 249, 399, 31 },
		{ 249, 400, 31 },
		{ 249, 401, 31 },
		{ 249, 402, 31 },
		{ 249, 403, 31 },
		{ 249, 404, 31 },
		{ 249, 405, 31 },
		{ 249, 406, 31 },
		{ 249, 407, 31 },
		{ 249, 408, 31 },
		{ 249, 409, 31 },
		{ 249, 410, 31 },
		{ 249, 411, 31 },
		{ 249, 412, 31 },
		{ 249, 413, 31 },
		{ 249, 414, 31 },
		{ 249, 415, 31 },
		{ 249, 416, 31 },
		{ 249, 417, 31 },
		{ 249, 418, 31 },
		{ 249, 419, 31 },
		{ 249, 420, 31 },
		{ 249, 421, 31 },
		{ 249, 422, 31 },
		{ 249, 423, 31 },
		{ 249, 424, 31 },
		{ 249, 425, 31 },
		{ 249, 426, 31 },
		{ 249, 427, 31 },
		{ 249, 428, 31 },
		{ 249, 429, 31 },
		{ 249, 430, 31 },
		{ 249, 431, 31 },
		{ 249, 432, 31 },
		{ 249, 433, 31 },
		{ 249, 434, 31 },
		{ 249, 435, 31 },
		{ 249, 436, 31 },
		{ 249, 437, 31 },
		{ 249, 438, 31 },
		{ 249, 439, 31 },
		{ 249, 440, 31 },
		{ 249, 441, 31 },
		{ 249, 442, 31 },
		{ 249, 443, 31 },
		{ 249, 444, 31 },
		{ 249, 445, 31 },
		{ 249, 446, 31 },
		{ 249, 447, 31 },
		{ 249, 448, 31 },
		{ 249, 449, 31 },
		{ 249, 450, 31 },
		{ 249, 451, 31 },
		{ 249, 452, 31 },
		{ 249, 453, 31 },
		{ 249, 454, 31 },
		{ 249, 455, 31 },
		{ 249, 456, 31 },
		{ 249, 457, 31 },
		{ 249, 458, 31 },
		{ 249, 459, 31 },
		{ 249, 460, 31 },
		{ 249, 461, 31 },
		{ 249, 462, 31 },
		{ 249, 463, 31 },
		{ 249, 464, 31 },
		{ 249, 465, 31 },
		{ 249, 466, 31 },
		{ 249, 467, 31 },
		{ 249, 468, 31 },
		{ 249, 469, 31 },
		{ 249, 470, 31 },
		{ 249, 471, 31 },
		{ 249, 472, 31 },
		{ 249, 473, 31 },
		{ 249, 474, 31 },
		{ 249, 475, 31 },
		{ 249, 476, 31 },
		{ 249, 477, 31 },
		{ 249, 478, 31 },
		{ 249, 479, 31 },
		{ 249, 480, 31 },
		{ 249, 481, 31 },
		{ 249, 482, 31 },
		{ 249, 483, 31 },
		{ 249, 484, 31 },
		{ 249, 485, 31 },
		{ 249, 486, 31 },
		{ 249, 487, 31 },
		{ 249, 488, 31 },
		{ 249, 489, 31 },
		{ 249, 490, 31 },
		{ 249, 491, 31 },
		{ 249, 492, 31 },
		{ 249, 493, 31 },
		{ 249, 494, 31 },
		{ 249, 495, 31 },
		{ 249, 496, 31 },
		{ 249, 497, 31 },
		{ 249, 498, 31 },
		{ 249, 499, 31 },
		{ 249, 500, 31 },
		{ 249, 501, 31 },
		{ 249, 502, 31 },
		{ 249, 503, 31 },
		{ 249, 504, 31 },
		{ 249, 505, 31 },
		{ 249, 506, 31 },
		{ 249, 507, 31 },
		{ 249, 508, 31 },
		{ 249, 509, 31 },
		{ 249, 510, 31 },
		{ 249, 511, 31 },
		{ 249, 512, 31 },
		{ 250, 323, 31 },
		{ 250, 324, 31 },
		{ 250, 325, 31 },
		{ 250, 326, 31 },
		{ 250, 327, 31 },
		{ 250, 328, 31 },
		{ 250, 329, 31 },
		{ 250, 330, 31 },
		{ 250, 331, 31 },
		{ 250, 332, 31 },
		{ 250, 333, 31 },
		{ 250, 334, 31 },
		{ 250, 335, 31 },
		{ 250, 336, 31 },
		{ 250, 337, 31 },
		{ 250, 338, 31 },
		{ 250, 339, 31 },
		{ 250, 340, 31 },
		{ 250, 341, 31 },
		{ 250, 342, 31 },
		{ 250, 343, 31 },
		{ 250, 344, 31 },
		{ 250, 345, 31 },
		{ 250, 346, 31 },
		{ 250, 347, 31 },
		{ 250, 348, 31 },
		{ 250, 349, 31 },
		{ 250, 350, 31 },
		{ 250, 351, 31 },
		{ 250, 352, 31 },
		{ 250, 353, 31 },
		{ 250, 354, 31 },
		{ 250, 355, 31 },
		{ 250, 356, 31 },
		{ 250, 357, 31 },
		{ 250, 358, 31 },
		{ 250, 359, 31 },
		{ 250, 360, 31 },
		{ 250, 361, 31 },
		{ 250, 362, 31 },
		{ 250, 363, 31 },
		{ 250, 364, 31 },
		{ 250, 365, 31 },
		{ 250, 366, 31 },
		{ 250, 367, 31 },
		{ 250, 368, 31 },
		{ 250, 369, 31 },
		{ 250, 370, 31 },
		{ 250, 371, 31 },
		{ 250, 372, 31 },
		{ 250, 373, 31 },
		{ 250, 374, 31 },
		{ 250, 375, 31 },
		{ 250, 376, 31 },
		{ 250, 377, 31 },
		{ 250, 378, 31 },
		{ 250, 379, 31 },
		{ 250, 380, 31 },
		{ 250, 381, 31 },
		{ 250, 382, 31 },
		{ 250, 383, 31 },
		{ 250, 384, 31 },
		{ 250, 385, 31 },
		{ 250, 386, 31 },
		{ 250, 387, 31 },
		{ 250, 388, 31 },
		{ 250, 389, 31 },
		{ 250, 390, 31 },
		{ 250, 391, 31 },
		{ 250, 392, 31 },
		{ 250, 393, 31 },
		{ 250, 394, 31 },
		{ 250, 395, 31 },
		{ 250, 396, 31 },
		{ 250, 397, 31 },
		{ 250, 398, 31 },
		{ 250, 399, 31 },
		{ 250, 400, 31 },
		{ 250, 401, 31 },
		{ 250, 402, 31 },
		{ 250, 403, 31 },
		{ 250, 404, 31 },
		{ 250, 405, 31 },
		{ 250, 406, 31 },
		{ 250, 407, 31 },
		{ 250, 408, 31 },
		{ 250, 409, 31 },
		{ 250, 410, 31 },
		{ 250, 411, 31 },
		{ 250, 412, 31 },
		{ 250, 413, 31 },
		{ 250, 414, 31 },
		{ 250, 415, 31 },
		{ 250, 416, 31 },
		{ 250, 417, 31 },
		{ 250, 418, 31 },
		{ 250, 419, 31 },
		{ 250, 420, 31 },
		{ 250, 421, 31 },
		{ 250, 422, 31 },
		{ 250, 423, 31 },
		{ 250, 424, 31 },
		{ 250, 425, 31 },
		{ 250, 426, 31 },
		{ 250, 427, 31 },
		{ 250, 428, 31 },
		{ 250, 429, 31 },
		{ 250, 430, 31 },
		{ 250, 431, 31 },
		{ 250, 432, 31 },
		{ 250, 433, 31 },
		{ 250, 434, 31 },
		{ 250, 435, 31 },
		{ 250, 436, 31 },
		{ 250, 437, 31 },
		{ 250, 438, 31 },
		{ 250, 439, 31 },
		{ 250, 440, 31 },
		{ 250, 441, 31 },
		{ 250, 442, 31 },
		{ 250, 443, 31 },
		{ 250, 444, 31 },
		{ 250, 445, 31 },
		{ 250, 446, 31 },
		{ 250, 447, 31 },
		{ 250, 448, 31 },
		{ 250, 449, 31 },
		{ 250, 450, 31 },
		{ 250, 451, 31 },
		{ 250, 452, 31 },
		{ 250, 453, 31 },
		{ 250, 454, 31 },
		{ 250, 455, 31 },
		{ 250, 456, 31 },
		{ 250, 457, 31 },
		{ 250, 458, 31 },
		{ 250, 459, 31 },
		{ 250, 460, 31 },
		{ 250, 461, 31 },
		{ 250, 462, 31 },
		{ 250, 463, 31 },
		{ 250, 464, 31 },
		{ 250, 465, 31 },
		{ 250, 466, 31 },
		{ 250, 467, 31 },
		{ 250, 468, 31 },
		{ 250, 469, 31 },
		{ 250, 470, 31 },
		{ 250, 471, 31 },
		{ 250, 472, 31 },
		{ 250, 473, 31 },
		{ 250, 474, 31 },
		{ 250, 475, 31 },
		{ 250, 476, 31 },
		{ 250, 477, 31 },
		{ 250, 478, 31 },
		{ 250, 479, 31 },
		{ 250, 480, 31 },
		{ 250, 481, 31 },
		{ 250, 482, 31 },
		{ 250, 483, 31 },
		{ 250, 484, 31 },
		{ 250, 485, 31 },
		{ 250, 486, 31 },
		{ 250, 487, 31 },
		{ 250, 488, 31 },
		{ 250, 489, 31 },
		{ 250, 490, 31 },
		{ 250, 491, 31 },
		{ 250, 492, 31 },
		{ 250, 493, 31 },
		{ 250, 494, 31 },
		{ 250, 495, 31 },
		{ 250, 496, 31 },
		{ 250, 497, 31 },
		{ 250, 498, 31 },
		{ 250, 499, 31 },
		{ 250, 500, 31 },
		{ 250, 501, 31 },
		{ 250, 502, 31 },
		{ 250, 503, 31 },
		{ 250, 504, 31 },
		{ 250, 505, 31 },
		{ 250, 506, 31 },
		{ 250, 507, 31 },
		{ 250, 508, 31 },
		{ 250, 509, 31 },
		{ 250, 510, 31 },
		{ 250, 511, 31 },
		{ 250, 512, 31 },
		{ 251, 323, 31 },
		{ 251, 324, 31 },
		{ 251, 325, 31 },
		{ 251, 326, 31 },
		{ 251, 327, 31 },
		{ 251, 328, 31 },
		{ 251, 329, 31 },
		{ 251, 330, 31 },
		{ 251, 331, 31 },
		{ 251, 332, 31 },
		{ 251, 333, 31 },
		{ 251, 334, 31 },
		{ 251, 335, 31 },
		{ 251, 336, 31 },
		{ 251, 337, 31 },
		{ 251, 338, 31 },
		{ 251, 339, 31 },
		{ 251, 340, 31 },
		{ 251, 341, 31 },
		{ 251, 342, 31 },
		{ 251, 343, 31 },
		{ 251, 344, 31 },
		{ 251, 345, 31 },
		{ 251, 346, 31 },
		{ 251, 347, 31 },
		{ 251, 348, 31 },
		{ 251, 349, 31 },
		{ 251, 350, 31 },
		{ 251, 351, 31 },
		{ 251, 352, 31 },
		{ 251, 353, 31 },
		{ 251, 354, 31 },
		{ 251, 355, 31 },
		{ 251, 356, 31 },
		{ 251, 357, 31 },
		{ 251, 358, 31 },
		{ 251, 359, 31 },
		{ 251, 360, 31 },
		{ 251, 361, 31 },
		{ 251, 362, 31 },
		{ 251, 363, 31 },
		{ 251, 364, 31 },
		{ 251, 365, 31 },
		{ 251, 366, 31 },
		{ 251, 367, 31 },
		{ 251, 368, 31 },
		{ 251, 369, 31 },
		{ 251, 370, 31 },
		{ 251, 371, 31 },
		{ 251, 372, 31 },
		{ 251, 373, 31 },
		{ 251, 374, 31 },
		{ 251, 375, 31 },
		{ 251, 376, 31 },
		{ 251, 377, 31 },
		{ 251, 378, 31 },
		{ 251, 379, 31 },
		{ 251, 380, 31 },
		{ 251, 381, 31 },
		{ 251, 382, 31 },
		{ 251, 383, 31 },
		{ 251, 384, 31 },
		{ 251, 385, 31 },
		{ 251, 386, 31 },
		{ 251, 387, 31 },
		{ 251, 388, 31 },
		{ 251, 389, 31 },
		{ 251, 390, 31 },
		{ 251, 391, 31 },
		{ 251, 392, 31 },
		{ 251, 393, 31 },
		{ 251, 394, 31 },
		{ 251, 395, 31 },
		{ 251, 396, 31 },
		{ 251, 397, 31 },
		{ 251, 398, 31 },
		{ 251, 399, 31 },
		{ 251, 400, 31 },
		{ 251, 401, 31 },
		{ 251, 402, 31 },
		{ 251, 403, 31 },
		{ 251, 404, 31 },
		{ 251, 405, 31 },
		{ 251, 406, 31 },
		{ 251, 407, 31 },
		{ 251, 408, 31 },
		{ 251, 409, 31 },
		{ 251, 410, 31 },
		{ 251, 411, 31 },
		{ 251, 412, 31 },
		{ 251, 413, 31 },
		{ 251, 414, 31 },
		{ 251, 415, 31 },
		{ 251, 416, 31 },
		{ 251, 417, 31 },
		{ 251, 418, 31 },
		{ 251, 419, 31 },
		{ 251, 420, 31 },
		{ 251, 421, 31 },
		{ 251, 422, 31 },
		{ 251, 423, 31 },
		{ 251, 424, 31 },
		{ 251, 425, 31 },
		{ 251, 426, 31 },
		{ 251, 427, 31 },
		{ 251, 428, 31 },
		{ 251, 429, 31 },
		{ 251, 430, 31 },
		{ 251, 431, 31 },
		{ 251, 432, 31 },
		{ 251, 433, 31 },
		{ 251, 434, 31 },
		{ 251, 435, 31 },
		{ 251, 436, 31 },
		{ 251, 437, 31 },
		{ 251, 438, 31 },
		{ 251, 439, 31 },
		{ 251, 440, 31 },
		{ 251, 441, 31 },
		{ 251, 442, 31 },
		{ 251, 443, 31 },
		{ 251, 444, 31 },
		{ 251, 445, 31 },
		{ 251, 446, 31 },
		{ 251, 447, 31 },
		{ 251, 448, 31 },
		{ 251, 449, 31 },
		{ 251, 450, 31 },
		{ 251, 451, 31 },
		{ 251, 452, 31 },
		{ 251, 453, 31 },
		{ 251, 454, 31 },
		{ 251, 455, 31 },
		{ 251, 456, 31 },
		{ 251, 457, 31 },
		{ 251, 458, 31 },
		{ 251, 459, 31 },
		{ 251, 460, 31 },
		{ 251, 461, 31 },
		{ 251, 462, 31 },
		{ 251, 463, 31 },
		{ 251, 464, 31 },
		{ 251, 465, 31 },
		{ 251, 466, 31 },
		{ 251, 467, 31 },
		{ 251, 468, 31 },
		{ 251, 469, 31 },
		{ 251, 470, 31 },
		{ 251, 471, 31 },
		{ 251, 472, 31 },
		{ 251, 473, 31 },
		{ 251, 474, 31 },
		{ 251, 475, 31 },
		{ 251, 476, 31 },
		{ 251, 477, 31 },
		{ 251, 478, 31 },
		{ 251, 479, 31 },
		{ 251, 480, 31 },
		{ 251, 481, 31 },
		{ 251, 482, 31 },
		{ 251, 483, 31 },
		{ 251, 484, 31 },
		{ 251, 485, 31 },
		{ 251, 486, 31 },
		{ 251, 487, 31 },
		{ 251, 488, 31 },
		{ 251, 489, 31 },
		{ 251, 490, 31 },
		{ 251, 491, 31 },
		{ 251, 492, 31 },
		{ 251, 493, 31 },
		{ 251, 494, 31 },
		{ 251, 495, 31 },
		{ 251, 496, 31 },
		{ 251, 497, 31 },
		{ 251, 498, 31 },
		{ 251, 499, 31 },
		{ 251, 500, 31 },
		{ 251, 501, 31 },
		{ 251, 502, 31 },
		{ 251, 503, 31 },
		{ 251, 504, 31 },
		{ 251, 505, 31 },
		{ 251, 506, 31 },
		{ 251, 507, 31 },
		{ 251, 508, 31 },
		{ 251, 509, 31 },
		{ 251, 510, 31 },
		{ 251, 511, 31 },
		{ 251, 512, 31 },
		{ 252, 323, 31 },
		{ 252, 324, 31 },
		{ 252, 325, 31 },
		{ 252, 326, 31 },
		{ 252, 327, 31 },
		{ 252, 328, 31 },
		{ 252, 329, 31 },
		{ 252, 330, 31 },
		{ 252, 331, 31 },
		{ 252, 332, 31 },
		{ 252, 333, 31 },
		{ 252, 334, 31 },
		{ 252, 335, 31 },
		{ 252, 336, 31 },
		{ 252, 337, 31 },
		{ 252, 338, 31 },
		{ 252, 339, 31 },
		{ 252, 340, 31 },
		{ 252, 341, 31 },
		{ 252, 342, 31 },
		{ 252, 343, 31 },
		{ 252, 344, 31 },
		{ 252, 345, 31 },
		{ 252, 346, 31 },
		{ 252, 347, 31 },
		{ 252, 348, 31 },
		{ 252, 349, 31 },
		{ 252, 350, 31 },
		{ 252, 351, 31 },
		{ 252, 352, 31 },
		{ 252, 353, 31 },
		{ 252, 354, 31 },
		{ 252, 355, 31 },
		{ 252, 356, 31 },
		{ 252, 357, 31 },
		{ 252, 358, 31 },
		{ 252, 359, 31 },
		{ 252, 360, 31 },
		{ 252, 361, 31 },
		{ 252, 362, 31 },
		{ 252, 363, 31 },
		{ 252, 364, 31 },
		{ 252, 365, 31 },
		{ 252, 366, 31 },
		{ 252, 367, 31 },
		{ 252, 368, 31 },
		{ 252, 369, 31 },
		{ 252, 370, 31 },
		{ 252, 371, 31 },
		{ 252, 372, 31 },
		{ 252, 373, 31 },
		{ 252, 374, 31 },
		{ 252, 375, 31 },
		{ 252, 376, 31 },
		{ 252, 377, 31 },
		{ 252, 378, 31 },
		{ 252, 379, 31 },
		{ 252, 380, 31 },
		{ 252, 381, 31 },
		{ 252, 382, 31 },
		{ 252, 383, 31 },
		{ 252, 384, 31 },
		{ 252, 385, 31 },
		{ 252, 386, 31 },
		{ 252, 387, 31 },
		{ 252, 388, 31 },
		{ 252, 389, 31 },
		{ 252, 390, 31 },
		{ 252, 391, 31 },
		{ 252, 392, 31 },
		{ 252, 393, 31 },
		{ 252, 394, 31 },
		{ 252, 395, 31 },
		{ 252, 396, 31 },
		{ 252, 397, 31 },
		{ 252, 398, 31 },
		{ 252, 399, 31 },
		{ 252, 400, 31 },
		{ 252, 401, 31 },
		{ 252, 402, 31 },
		{ 252, 403, 31 },
		{ 252, 404, 31 },
		{ 252, 405, 31 },
		{ 252, 406, 31 },
		{ 252, 407, 31 },
		{ 252, 408, 31 },
		{ 252, 409, 31 },
		{ 252, 410, 31 },
		{ 252, 411, 31 },
		{ 252, 412, 31 },
		{ 252, 413, 31 },
		{ 252, 414, 31 },
		{ 252, 415, 31 },
		{ 252, 416, 31 },
		{ 252, 417, 31 },
		{ 252, 418, 31 },
		{ 252, 419, 31 },
		{ 252, 420, 31 },
		{ 252, 421, 31 },
		{ 252, 422, 31 },
		{ 252, 423, 31 },
		{ 252, 424, 31 },
		{ 252, 425, 31 },
		{ 252, 426, 31 },
		{ 252, 427, 31 },
		{ 252, 428, 31 },
		{ 252, 429, 31 },
		{ 252, 430, 31 },
		{ 252, 431, 31 },
		{ 252, 432, 31 },
		{ 252, 433, 31 },
		{ 252, 434, 31 },
		{ 252, 435, 31 },
		{ 252, 436, 31 },
		{ 252, 437, 31 },
		{ 252, 438, 31 },
		{ 252, 439, 31 },
		{ 252, 440, 31 },
		{ 252, 441, 31 },
		{ 252, 442, 31 },
		{ 252, 443, 31 },
		{ 252, 444, 31 },
		{ 252, 445, 31 },
		{ 252, 446, 31 },
		{ 252, 447, 31 },
		{ 252, 448, 31 },
		{ 252, 449, 31 },
		{ 252, 450, 31 },
		{ 252, 451, 31 },
		{ 252, 452, 31 },
		{ 252, 453, 31 },
		{ 252, 454, 31 },
		{ 252, 455, 31 },
		{ 252, 456, 31 },
		{ 252, 457, 31 },
		{ 252, 458, 31 },
		{ 252, 459, 31 },
		{ 252, 460, 31 },
		{ 252, 461, 31 },
		{ 252, 462, 31 },
		{ 252, 463, 31 },
		{ 252, 464, 31 },
		{ 252, 465, 31 },
		{ 252, 466, 31 },
		{ 252, 467, 31 },
		{ 252, 468, 31 },
		{ 252, 469, 31 },
		{ 252, 470, 31 },
		{ 252, 471, 31 },
		{ 252, 472, 31 },
		{ 252, 473, 31 },
		{ 252, 474, 31 },
		{ 252, 475, 31 },
		{ 252, 476, 31 },
		{ 252, 477, 31 },
		{ 252, 478, 31 },
		{ 252, 479, 31 },
		{ 252, 480, 31 },
		{ 252, 481, 31 },
		{ 252, 482, 31 },
		{ 252, 483, 31 },
		{ 252, 484, 31 },
		{ 252, 485, 31 },
		{ 252, 486, 31 },
		{ 252, 487, 31 },
		{ 252, 488, 31 },
		{ 252, 489, 31 },
		{ 252, 490, 31 },
		{ 252, 491, 31 },
		{ 252, 492, 31 },
		{ 252, 493, 31 },
		{ 252, 494, 31 },
		{ 252, 495, 31 },
		{ 252, 496, 31 },
		{ 252, 497, 31 },
		{ 252, 498, 31 },
		{ 252, 499, 31 },
		{ 252, 500, 31 },
		{ 252, 501, 31 },
		{ 252, 502, 31 },
		{ 252, 503, 31 },
		{ 252, 504, 31 },
		{ 252, 505, 31 },
		{ 252, 506, 31 },
		{ 252, 507, 31 },
		{ 252, 508, 31 },
		{ 252, 509, 31 },
		{ 252, 510, 31 },
		{ 252, 511, 31 },
		{ 252, 512, 31 },
		{ 253, 323, 31 },
		{ 253, 324, 31 },
		{ 253, 325, 31 },
		{ 253, 326, 31 },
		{ 253, 327, 31 },
		{ 253, 328, 31 },
		{ 253, 329, 31 },
		{ 253, 330, 31 },
		{ 253, 331, 31 },
		{ 253, 332, 31 },
		{ 253, 333, 31 },
		{ 253, 334, 31 },
		{ 253, 335, 31 },
		{ 253, 336, 31 },
		{ 253, 337, 31 },
		{ 253, 338, 31 },
		{ 253, 339, 31 },
		{ 253, 340, 31 },
		{ 253, 341, 31 },
		{ 253, 342, 31 },
		{ 253, 343, 31 },
		{ 253, 344, 31 },
		{ 253, 345, 31 },
		{ 253, 346, 31 },
		{ 253, 347, 31 },
		{ 253, 348, 31 },
		{ 253, 349, 31 },
		{ 253, 350, 31 },
		{ 253, 351, 31 },
		{ 253, 352, 31 },
		{ 253, 353, 31 },
		{ 253, 354, 31 },
		{ 253, 355, 31 },
		{ 253, 356, 31 },
		{ 253, 357, 31 },
		{ 253, 358, 31 },
		{ 253, 359, 31 },
		{ 253, 360, 31 },
		{ 253, 361, 31 },
		{ 253, 362, 31 },
		{ 253, 363, 31 },
		{ 253, 364, 31 },
		{ 253, 365, 31 },
		{ 253, 366, 31 },
		{ 253, 367, 31 },
		{ 253, 368, 31 },
		{ 253, 369, 31 },
		{ 253, 370, 31 },
		{ 253, 371, 31 },
		{ 253, 372, 31 },
		{ 253, 373, 31 },
		{ 253, 374, 31 },
		{ 253, 375, 31 },
		{ 253, 376, 31 },
		{ 253, 377, 31 },
		{ 253, 378, 31 },
		{ 253, 379, 31 },
		{ 253, 380, 31 },
		{ 253, 381, 31 },
		{ 253, 382, 31 },
		{ 253, 383, 31 },
		{ 253, 384, 31 },
		{ 253, 385, 31 },
		{ 253, 386, 31 },
		{ 253, 387, 31 },
		{ 253, 388, 31 },
		{ 253, 389, 31 },
		{ 253, 390, 31 },
		{ 253, 391, 31 },
		{ 253, 392, 31 },
		{ 253, 393, 31 },
		{ 253, 394, 31 },
		{ 253, 395, 31 },
		{ 253, 396, 31 },
		{ 253, 397, 31 },
		{ 253, 398, 31 },
		{ 253, 399, 31 },
		{ 253, 400, 31 },
		{ 253, 401, 31 },
		{ 253, 402, 31 },
		{ 253, 403, 31 },
		{ 253, 404, 31 },
		{ 253, 405, 31 },
		{ 253, 406, 31 },
		{ 253, 407, 31 },
		{ 253, 408, 31 },
		{ 253, 409, 31 },
		{ 253, 410, 31 },
		{ 253, 411, 31 },
		{ 253, 412, 31 },
		{ 253, 413, 31 },
		{ 253, 414, 31 },
		{ 253, 415, 31 },
		{ 253, 416, 31 },
		{ 253, 417, 31 },
		{ 253, 418, 31 },
		{ 253, 419, 31 },
		{ 253, 420, 31 },
		{ 253, 421, 31 },
		{ 253, 422, 31 },
		{ 253, 423, 31 },
		{ 253, 424, 31 },
		{ 253, 425, 31 },
		{ 253, 426, 31 },
		{ 253, 427, 31 },
		{ 253, 428, 31 },
		{ 253, 429, 31 },
		{ 253, 430, 31 },
		{ 253, 431, 31 },
		{ 253, 432, 31 },
		{ 253, 433, 31 },
		{ 253, 434, 31 },
		{ 253, 435, 31 },
		{ 253, 436, 31 },
		{ 253, 437, 31 },
		{ 253, 438, 31 },
		{ 253, 439, 31 },
		{ 253, 440, 31 },
		{ 253, 441, 31 },
		{ 253, 442, 31 },
		{ 253, 443, 31 },
		{ 253, 444, 31 },
		{ 253, 445, 31 },
		{ 253, 446, 31 },
		{ 253, 447, 31 },
		{ 253, 448, 31 },
		{ 253, 449, 31 },
		{ 253, 450, 31 },
		{ 253, 451, 31 },
		{ 253, 452, 31 },
		{ 253, 453, 31 },
		{ 253, 454, 31 },
		{ 253, 455, 31 },
		{ 253, 456, 31 },
		{ 253, 457, 31 },
		{ 253, 458, 31 },
		{ 253, 459, 31 },
		{ 253, 460, 31 },
		{ 253, 461, 31 },
		{ 253, 462, 31 },
		{ 253, 463, 31 },
		{ 253, 464, 31 },
		{ 253, 465, 31 },
		{ 253, 466, 31 },
		{ 253, 467, 31 },
		{ 253, 468, 31 },
		{ 253, 469, 31 },
		{ 253, 470, 31 },
		{ 253, 471, 31 },
		{ 253, 472, 31 },
		{ 253, 473, 31 },
		{ 253, 474, 31 },
		{ 253, 475, 31 },
		{ 253, 476, 31 },
		{ 253, 477, 31 },
		{ 253, 478, 31 },
		{ 253, 479, 31 },
		{ 253, 480, 31 },
		{ 253, 481, 31 },
		{ 253, 482, 31 },
		{ 253, 483, 31 },
		{ 253, 484, 31 },
		{ 253, 485, 31 },
		{ 253, 486, 31 },
		{ 253, 487, 31 },
		{ 253, 488, 31 },
		{ 253, 489, 31 },
		{ 253, 490, 31 },
		{ 253, 491, 31 },
		{ 253, 492, 31 },
		{ 253, 493, 31 },
		{ 253, 494, 31 },
		{ 253, 495, 31 },
		{ 253, 496, 31 },
		{ 253, 497, 31 },
		{ 253, 498, 31 },
		{ 253, 499, 31 },
		{ 253, 500, 31 },
		{ 253, 501, 31 },
		{ 253, 502, 31 },
		{ 253, 503, 31 },
		{ 253, 504, 31 },
		{ 253, 505, 31 },
		{ 253, 506, 31 },
		{ 253, 507, 31 },
		{ 253, 508, 31 },
		{ 253, 509, 31 },
		{ 253, 510, 31 },
		{ 253, 511, 31 },
		{ 253, 512, 31 },
		{ 254, 323, 31 },
		{ 254, 324, 31 },
		{ 254, 325, 31 },
		{ 254, 326, 31 },
		{ 254, 327, 31 },
		{ 254, 328, 31 },
		{ 254, 329, 31 },
		{ 254, 330, 31 },
		{ 254, 331, 31 },
		{ 254, 332, 31 },
		{ 254, 333, 31 },
		{ 254, 334, 31 },
		{ 254, 335, 31 },
		{ 254, 336, 31 },
		{ 254, 337, 31 },
		{ 254, 338, 31 },
		{ 254, 339, 31 },
		{ 254, 340, 31 },
		{ 254, 341, 31 },
		{ 254, 342, 31 },
		{ 254, 343, 31 },
		{ 254, 344, 31 },
		{ 254, 345, 31 },
		{ 254, 346, 31 },
		{ 254, 347, 31 },
		{ 254, 348, 31 },
		{ 254, 349, 31 },
		{ 254, 350, 31 },
		{ 254, 351, 31 },
		{ 254, 352, 31 },
		{ 254, 353, 31 },
		{ 254, 354, 31 },
		{ 254, 355, 31 },
		{ 254, 356, 31 },
		{ 254, 357, 31 },
		{ 254, 358, 31 },
		{ 254, 359, 31 },
		{ 254, 360, 31 },
		{ 254, 361, 31 },
		{ 254, 362, 31 },
		{ 254, 363, 31 },
		{ 254, 364, 31 },
		{ 254, 365, 31 },
		{ 254, 366, 31 },
		{ 254, 367, 31 },
		{ 254, 368, 31 },
		{ 254, 369, 31 },
		{ 254, 370, 31 },
		{ 254, 371, 31 },
		{ 254, 372, 31 },
		{ 254, 373, 31 },
		{ 254, 374, 31 },
		{ 254, 375, 31 },
		{ 254, 376, 31 },
		{ 254, 377, 31 },
		{ 254, 378, 31 },
		{ 254, 379, 31 },
		{ 254, 380, 31 },
		{ 254, 381, 31 },
		{ 254, 382, 31 },
		{ 254, 383, 31 },
		{ 254, 384, 31 },
		{ 254, 385, 31 },
		{ 254, 386, 31 },
		{ 254, 387, 31 },
		{ 254, 388, 31 },
		{ 254, 389, 31 },
		{ 254, 390, 31 },
		{ 254, 391, 31 },
		{ 254, 392, 31 },
		{ 254, 393, 31 },
		{ 254, 394, 31 },
		{ 254, 395, 31 },
		{ 254, 396, 31 },
		{ 254, 397, 31 },
		{ 254, 398, 31 },
		{ 254, 399, 31 },
		{ 254, 400, 31 },
		{ 254, 401, 31 },
		{ 254, 402, 31 },
		{ 254, 403, 31 },
		{ 254, 404, 31 },
		{ 254, 405, 31 },
		{ 254, 406, 31 },
		{ 254, 407, 31 },
		{ 254, 408, 31 },
		{ 254, 409, 31 },
		{ 254, 410, 31 },
		{ 254, 411, 31 },
		{ 254, 412, 31 },
		{ 254, 413, 31 },
		{ 254, 414, 31 },
		{ 254, 415, 31 },
		{ 254, 416, 31 },
		{ 254, 417, 31 },
		{ 254, 418, 31 },
		{ 254, 419, 31 },
		{ 254, 420, 31 },
		{ 254, 421, 31 },
		{ 254, 422, 31 },
		{ 254, 423, 31 },
		{ 254, 424, 31 },
		{ 254, 425, 31 },
		{ 254, 426, 31 },
		{ 254, 427, 31 },
		{ 254, 428, 31 },
		{ 254, 429, 31 },
		{ 254, 430, 31 },
		{ 254, 431, 31 },
		{ 254, 432, 31 },
		{ 254, 433, 31 },
		{ 254, 434, 31 },
		{ 254, 435, 31 },
		{ 254, 436, 31 },
		{ 254, 437, 31 },
		{ 254, 438, 31 },
		{ 254, 439, 31 },
		{ 254, 440, 31 },
		{ 254, 441, 31 },
		{ 254, 442, 31 },
		{ 254, 443, 31 },
		{ 254, 444, 31 },
		{ 254, 445, 31 },
		{ 254, 446, 31 },
		{ 254, 447, 31 },
		{ 254, 448, 31 },
		{ 254, 449, 31 },
		{ 254, 450, 31 },
		{ 254, 451, 31 },
		{ 254, 452, 31 },
		{ 254, 453, 31 },
		{ 254, 454, 31 },
		{ 254, 455, 31 },
		{ 254, 456, 31 },
		{ 254, 457, 31 },
		{ 254, 458, 31 },
		{ 254, 459, 31 },
		{ 254, 460, 31 },
		{ 254, 461, 31 },
		{ 254, 462, 31 },
		{ 254, 463, 31 },
		{ 254, 464, 31 },
		{ 254, 465, 31 },
		{ 254, 466, 31 },
		{ 254, 467, 31 },
		{ 254, 468, 31 },
		{ 254, 469, 31 },
		{ 254, 470, 31 },
		{ 254, 471, 31 },
		{ 254, 472, 31 },
		{ 254, 473, 31 },
		{ 254, 474, 31 },
		{ 254, 475, 31 },
		{ 254, 476, 31 },
		{ 254, 477, 31 },
		{ 254, 478, 31 },
		{ 254, 479, 31 },
		{ 254, 480, 31 },
		{ 254, 481, 31 },
		{ 254, 482, 31 },
		{ 254, 483, 31 },
		{ 254, 484, 31 },
		{ 254, 485, 31 },
		{ 254, 486, 31 },
		{ 254, 487, 31 },
		{ 254, 488, 31 },
		{ 254, 489, 31 },
		{ 254, 490, 31 },
		{ 254, 491, 31 },
		{ 254, 492, 31 },
		{ 254, 493, 31 },
		{ 254, 494, 31 },
		{ 254, 495, 31 },
		{ 254, 496, 31 },
		{ 254, 497, 31 },
		{ 254, 498, 31 },
		{ 254, 499, 31 },
		{ 254, 500, 31 },
		{ 254, 501, 31 },
		{ 254, 502, 31 },
		{ 254, 503, 31 },
		{ 254, 504, 31 },
		{ 254, 505, 31 },
		{ 254, 506, 31 },
		{ 254, 507, 31 },
		{ 254, 508, 31 },
		{ 254, 509, 31 },
		{ 254, 510, 31 },
		{ 254, 511, 31 },
		{ 254, 512, 31 },
		{ 255, 323, 31 },
		{ 255, 324, 31 },
		{ 255, 325, 31 },
		{ 255, 326, 31 },
		{ 255, 327, 31 },
		{ 255, 328, 31 },
		{ 255, 329, 31 },
		{ 255, 330, 31 },
		{ 255, 331, 31 },
		{ 255, 332, 31 },
		{ 255, 333, 31 },
		{ 255, 334, 31 },
		{ 255, 335, 31 },
		{ 255, 336, 31 },
		{ 255, 337, 31 },
		{ 255, 338, 31 },
		{ 255, 339, 31 },
		{ 255, 340, 31 },
		{ 255, 341, 31 },
		{ 255, 342, 31 },
		{ 255, 343, 31 },
		{ 255, 344, 31 },
		{ 255, 345, 31 },
		{ 255, 346, 31 },
		{ 255, 347, 31 },
		{ 255, 348, 31 },
		{ 255, 349, 31 },
		{ 255, 350, 31 },
		{ 255, 351, 31 },
		{ 255, 352, 31 },
		{ 255, 353, 31 },
		{ 255, 354, 31 },
		{ 255, 355, 31 },
		{ 255, 356, 31 },
		{ 255, 357, 31 },
		{ 255, 358, 31 },
		{ 255, 359, 31 },
		{ 255, 360, 31 },
		{ 255, 361, 31 },
		{ 255, 362, 31 },
		{ 255, 363, 31 },
		{ 255, 364, 31 },
		{ 255, 365, 31 },
		{ 255, 366, 31 },
		{ 255, 367, 31 },
		{ 255, 368, 31 },
		{ 255, 369, 31 },
		{ 255, 370, 31 },
		{ 255, 371, 31 },
		{ 255, 372, 31 },
		{ 255, 373, 31 },
		{ 255, 374, 31 },
		{ 255, 375, 31 },
		{ 255, 376, 31 },
		{ 255, 377, 31 },
		{ 255, 378, 31 },
		{ 255, 379, 31 },
		{ 255, 380, 31 },
		{ 255, 381, 31 },
		{ 255, 382, 31 },
		{ 255, 383, 31 },
		{ 255, 384, 31 },
		{ 255, 385, 31 },
		{ 255, 386, 31 },
		{ 255, 387, 31 },
		{ 255, 388, 31 },
		{ 255, 389, 31 },
		{ 255, 390, 31 },
		{ 255, 391, 31 },
		{ 255, 392, 31 },
		{ 255, 393, 31 },
		{ 255, 394, 31 },
		{ 255, 395, 31 },
		{ 255, 396, 31 },
		{ 255, 397, 31 },
		{ 255, 398, 31 },
		{ 255, 399, 31 },
		{ 255, 400, 31 },
		{ 255, 401, 31 },
		{ 255, 402, 31 },
		{ 255, 403, 31 },
		{ 255, 404, 31 },
		{ 255, 405, 31 },
		{ 255, 406, 31 },
		{ 255, 407, 31 },
		{ 255, 408, 31 },
		{ 255, 409, 31 },
		{ 255, 410, 31 },
		{ 255, 411, 31 },
		{ 255, 412, 31 },
		{ 255, 413, 31 },
		{ 255, 414, 31 },
		{ 255, 415, 31 },
		{ 255, 416, 31 },
		{ 255, 417, 31 },
		{ 255, 418, 31 },
		{ 255, 419, 31 },
		{ 255, 420, 31 },
		{ 255, 421, 31 },
		{ 255, 422, 31 },
		{ 255, 423, 31 },
		{ 255, 424, 31 },
		{ 255, 425, 31 },
		{ 255, 426, 31 },
		{ 255, 427, 31 },
		{ 255, 428, 31 },
		{ 255, 429, 31 },
		{ 255, 430, 31 },
		{ 255, 431, 31 },
		{ 255, 432, 31 },
		{ 255, 433, 31 },
		{ 255, 434, 31 },
		{ 255, 435, 31 },
		{ 255, 436, 31 },
		{ 255, 437, 31 },
		{ 255, 438, 31 },
		{ 255, 439, 31 },
		{ 255, 440, 31 },
		{ 255, 441, 31 },
		{ 255, 442, 31 },
		{ 255, 443, 31 },
		{ 255, 444, 31 },
		{ 255, 445, 31 },
		{ 255, 446, 31 },
		{ 255, 447, 31 },
		{ 255, 448, 31 },
		{ 255, 449, 31 },
		{ 255, 450, 31 },
		{ 255, 451, 31 },
		{ 255, 452, 31 },
		{ 255, 453, 31 },
		{ 255, 454, 31 },
		{ 255, 455, 31 },
		{ 255, 456, 31 },
		{ 255, 457, 31 },
		{ 255, 458, 31 },
		{ 255, 459, 31 },
		{ 255, 460, 31 },
		{ 255, 461, 31 },
		{ 255, 462, 31 },
		{ 255, 463, 31 },
		{ 255, 464, 31 },
		{ 255, 465, 31 },
		{ 255, 466, 31 },
		{ 255, 467, 31 },
		{ 255, 468, 31 },
		{ 255, 469, 31 },
		{ 255, 470, 31 },
		{ 255, 471, 31 },
		{ 255, 472, 31 },
		{ 255, 473, 31 },
		{ 255, 474, 31 },
		{ 255, 475, 31 },
		{ 255, 476, 31 },
		{ 255, 477, 31 },
		{ 255, 478, 31 },
		{ 255, 479, 31 },
		{ 255, 480, 31 },
		{ 255, 481, 31 },
		{ 255, 482, 31 },
		{ 255, 483, 31 },
		{ 255, 484, 31 },
		{ 255, 485, 31 },
		{ 255, 486, 31 },
		{ 255, 487, 31 },
		{ 255, 488, 31 },
		{ 255, 489, 31 },
		{ 255, 490, 31 },
		{ 255, 491, 31 },
		{ 255, 492, 31 },
		{ 255, 493, 31 },
		{ 255, 494, 31 },
		{ 255, 495, 31 },
		{ 255, 496, 31 },
		{ 255, 497, 31 },
		{ 255, 498, 31 },
		{ 255, 499, 31 },
		{ 255, 500, 31 },
		{ 255, 501, 31 },
		{ 255, 502, 31 },
		{ 255, 503, 31 },
		{ 255, 504, 31 },
		{ 255, 505, 31 },
		{ 255, 506, 31 },
		{ 255, 507, 31 },
		{ 255, 508, 31 },
		{ 255, 509, 31 },
		{ 255, 510, 31 },
		{ 255, 511, 31 },
		{ 255, 512, 31 },
		{ 256, 324, 31 },
		{ 256, 325, 31 },
		{ 256, 326, 31 },
		{ 256, 327, 31 },
		{ 256, 328, 31 },
		{ 256, 329, 31 },
		{ 256, 330, 31 },
		{ 256, 331, 31 },
		{ 256, 332, 31 },
		{ 256, 333, 31 },
		{ 256, 334, 31 },
		{ 256, 335, 31 },
		{ 256, 336, 31 },
		{ 256, 337, 31 },
		{ 256, 338, 31 },
		{ 256, 339, 31 },
		{ 256, 340, 31 },
		{ 256, 341, 31 },
		{ 256, 342, 31 },
		{ 256, 343, 31 },
		{ 256, 344, 31 },
		{ 256, 345, 31 },
		{ 256, 346, 31 },
		{ 256, 347, 31 },
		{ 256, 348, 31 },
		{ 256, 349, 31 },
		{ 256, 350, 31 },
		{ 256, 351, 31 },
		{ 256, 352, 31 },
		{ 256, 353, 31 },
		{ 256, 354, 31 },
		{ 256, 355, 31 },
		{ 256, 356, 31 },
		{ 256, 357, 31 },
		{ 256, 358, 31 },
		{ 256, 359, 31 },
		{ 256, 360, 31 },
		{ 256, 361, 31 },
		{ 256, 362, 31 },
		{ 256, 363, 31 },
		{ 256, 364, 31 },
		{ 256, 365, 31 },
		{ 256, 366, 31 },
		{ 256, 367, 31 },
		{ 256, 368, 31 },
		{ 256, 369, 31 },
		{ 256, 370, 31 },
		{ 256, 371, 31 },
		{ 256, 372, 31 },
		{ 256, 373, 31 },
		{ 256, 374, 31 },
		{ 256, 375, 31 },
		{ 256, 376, 31 },
		{ 256, 377, 31 },
		{ 256, 378, 31 },
		{ 256, 379, 31 },
		{ 256, 380, 31 },
		{ 256, 381, 31 },
		{ 256, 382, 31 },
		{ 256, 383, 31 },
		{ 256, 384, 31 },
		{ 256, 385, 31 },
		{ 256, 386, 31 },
		{ 256, 387, 31 },
		{ 256, 388, 31 },
		{ 256, 389, 31 },
		{ 256, 390, 31 },
		{ 256, 391, 31 },
		{ 256, 392, 31 },
		{ 256, 393, 31 },
		{ 256, 394, 31 },
		{ 256, 395, 31 },
		{ 256, 396, 31 },
		{ 256, 397, 31 },
		{ 256, 398, 31 },
		{ 256, 399, 31 },
		{ 256, 400, 31 },
		{ 256, 401, 31 },
		{ 256, 402, 31 },
		{ 256, 403, 31 },
		{ 256, 404, 31 },
		{ 256, 405, 31 },
		{ 256, 406, 31 },
		{ 256, 407, 31 },
		{ 256, 408, 31 },
		{ 256, 409, 31 },
		{ 256, 410, 31 },
		{ 256, 411, 31 },
		{ 256, 412, 31 },
		{ 256, 413, 31 },
		{ 256, 414, 31 },
		{ 256, 415, 31 },
		{ 256, 416, 31 },
		{ 256, 417, 31 },
		{ 256, 418, 31 },
		{ 256, 419, 31 },
		{ 256, 420, 31 },
		{ 256, 421, 31 },
		{ 256, 422, 31 },
		{ 256, 423, 31 },
		{ 256, 424, 31 },
		{ 256, 425, 31 },
		{ 256, 426, 31 },
		{ 256, 427, 31 },
		{ 256, 428, 31 },
		{ 256, 429, 31 },
		{ 256, 430, 31 },
		{ 256, 431, 31 },
		{ 256, 432, 31 },
		{ 256, 433, 31 },
		{ 256, 434, 31 },
		{ 256, 435, 31 },
		{ 256, 436, 31 },
		{ 256, 437, 31 },
		{ 256, 438, 31 },
		{ 256, 439, 31 },
		{ 256, 440, 31 },
		{ 256, 441, 31 },
		{ 256, 442, 31 },
		{ 256, 443, 31 },
		{ 256, 444, 31 },
		{ 256, 445, 31 },
		{ 256, 446, 31 },
		{ 256, 447, 31 },
		{ 256, 448, 31 },
		{ 256, 449, 31 },
		{ 256, 450, 31 },
		{ 256, 451, 31 },
		{ 256, 452, 31 },
		{ 256, 453, 31 },
		{ 256, 454, 31 },
		{ 256, 455, 31 },
		{ 256, 456, 31 },
		{ 256, 457, 31 },
		{ 256, 458, 31 },
		{ 256, 459, 31 },
		{ 256, 460, 31 },
		{ 256, 461, 31 },
		{ 256, 462, 31 },
		{ 256, 463, 31 },
		{ 256, 464, 31 },
		{ 256, 465, 31 },
		{ 256, 466, 31 },
		{ 256, 467, 31 },
		{ 256, 468, 31 },
		{ 256, 469, 31 },
		{ 256, 470, 31 },
		{ 256, 471, 31 },
		{ 256, 472, 31 },
		{ 256, 473, 31 },
		{ 256, 474, 31 },
		{ 256, 475, 31 },
		{ 256, 476, 31 },
		{ 256, 477, 31 },
		{ 256, 478, 31 },
		{ 256, 479, 31 },
		{ 256, 480, 31 },
		{ 256, 481, 31 },
		{ 256, 482, 31 },
		{ 256, 483, 31 },
		{ 256, 484, 31 },
		{ 256, 485, 31 },
		{ 256, 486, 31 },
		{ 256, 487, 31 },
		{ 256, 488, 31 },
		{ 256, 489, 31 },
		{ 256, 490, 31 },
		{ 256, 491, 31 },
		{ 256, 492, 31 },
		{ 256, 493, 31 },
		{ 256, 494, 31 },
		{ 256, 495, 31 },
		{ 256, 496, 31 },
		{ 256, 497, 31 },
		{ 256, 498, 31 },
		{ 256, 499, 31 },
		{ 256, 500, 31 },
		{ 256, 501, 31 },
		{ 256, 502, 31 },
		{ 256, 503, 31 },
		{ 256, 504, 31 },
		{ 256, 505, 31 },
		{ 256, 506, 31 },
		{ 256, 507, 31 },
		{ 256, 508, 31 },
		{ 256, 509, 31 },
		{ 256, 510, 31 },
		{ 256, 511, 31 },
		{ 256, 512, 31 },
		{ 258, 1, 31 },
		{ 258, 2, 31 },
		{ 258, 3, 31 },
		{ 258, 4, 31 },
		{ 258, 5, 31 },
		{ 258, 6, 31 },
		{ 258, 7, 31 },
		{ 258, 8, 31 },
		{ 258, 9, 31 },
		{ 258, 10, 31 },
		{ 258, 11, 31 },
		{ 258, 12, 31 },
		{ 258, 13, 31 },
		{ 258, 14, 31 },
		{ 258, 15, 31 },
		{ 258, 16, 31 },
		{ 258, 17, 31 },
		{ 258, 18, 31 },
		{ 258, 19, 31 },
		{ 258, 20, 31 },
		{ 258, 21, 31 },
		{ 258, 22, 31 },
		{ 258, 23, 31 },
		{ 258, 24, 31 },
		{ 258, 25, 31 },
		{ 258, 26, 31 },
		{ 258, 27, 31 },
		{ 258, 28, 31 },
		{ 258, 29, 31 },
		{ 258, 30, 31 },
		{ 258, 31, 31 },
		{ 258, 32, 31 },
		{ 258, 33, 31 },
		{ 258, 34, 31 },
		{ 258, 35, 31 },
		{ 258, 36, 31 },
		{ 258, 37, 31 },
		{ 258, 38, 31 },
		{ 258, 39, 31 },
		{ 258, 40, 31 },
		{ 258, 41, 31 },
		{ 258, 42, 31 },
		{ 258, 43, 31 },
		{ 258, 44, 31 },
		{ 258, 45, 31 },
		{ 258, 46, 31 },
		{ 258, 47, 31 },
		{ 258, 48, 31 },
		{ 258, 49, 31 },
		{ 258, 50, 31 },
		{ 258, 51, 31 },
		{ 258, 52, 31 },
		{ 258, 53, 31 },
		{ 258, 54, 31 },
		{ 258, 55, 31 },
		{ 258, 56, 31 },
		{ 258, 57, 31 },
		{ 258, 58, 31 },
		{ 258, 59, 31 },
		{ 258, 60, 31 },
		{ 258, 61, 31 },
		{ 258, 62, 31 },
		{ 258, 63, 31 },
		{ 258, 64, 31 },
		{ 258, 65, 31 },
		{ 258, 66, 31 },
		{ 258, 67, 31 },
		{ 258, 68, 31 },
		{ 258, 69, 31 },
		{ 258, 70, 31 },
		{ 258, 71, 31 },
		{ 258, 72, 31 },
		{ 258, 73, 31 },
		{ 258, 74, 31 },
		{ 258, 75, 31 },
		{ 258, 76, 31 },
		{ 258, 77, 31 },
		{ 258, 78, 31 },
		{ 258, 79, 31 },
		{ 258, 80, 31 },
		{ 258, 81, 31 },
		{ 258, 82, 31 },
		{ 258, 83, 31 },
		{ 258, 84, 31 },
		{ 258, 85, 31 },
		{ 258, 86, 31 },
		{ 258, 87, 31 },
		{ 258, 88, 31 },
		{ 258, 89, 31 },
		{ 258, 90, 31 },
		{ 258, 91, 31 },
		{ 258, 92, 31 },
		{ 258, 93, 31 },
		{ 258, 94, 31 },
		{ 258, 95, 31 },
		{ 258, 96, 31 },
		{ 258, 97, 31 },
		{ 258, 98, 31 },
		{ 258, 99, 31 },
		{ 258, 100, 31 },
		{ 258, 101, 31 },
		{ 258, 102, 31 },
		{ 258, 103, 31 },
		{ 258, 104, 31 },
		{ 258, 105, 31 },
		{ 258, 106, 31 },
		{ 258, 107, 31 },
		{ 258, 108, 31 },
		{ 258, 109, 31 },
		{ 258, 110, 31 },
		{ 258, 111, 31 },
		{ 258, 112, 31 },
		{ 258, 113, 31 },
		{ 258, 114, 31 },
		{ 258, 115, 31 },
		{ 258, 116, 31 },
		{ 258, 117, 31 },
		{ 258, 118, 31 },
		{ 258, 119, 31 },
		{ 258, 120, 31 },
		{ 258, 121, 31 },
		{ 258, 122, 31 },
		{ 258, 123, 31 },
		{ 258, 124, 31 },
		{ 258, 125, 31 },
		{ 258, 126, 31 },
		{ 258, 127, 31 },
		{ 258, 128, 31 },
		{ 258, 129, 31 },
		{ 258, 130, 31 },
		{ 258, 131, 31 },
		{ 258, 132, 31 },
		{ 258, 133, 31 },
		{ 258, 134, 31 },
		{ 258, 135, 31 },
		{ 258, 136, 31 },
		{ 258, 137, 31 },
		{ 258, 138, 31 },
		{ 258, 139, 31 },
		{ 258, 140, 31 },
		{ 258, 141, 31 },
		{ 258, 142, 31 },
		{ 258, 143, 31 },
		{ 258, 144, 31 },
		{ 258, 145, 31 },
		{ 258, 146, 31 },
		{ 258, 147, 31 },
		{ 258, 148, 31 },
		{ 258, 149, 31 },
		{ 258, 150, 31 },
		{ 258, 151, 31 },
		{ 258, 152, 31 },
		{ 258, 153, 31 },
		{ 258, 154, 31 },
		{ 258, 155, 31 },
		{ 258, 156, 31 },
		{ 258, 157, 31 },
		{ 258, 158, 31 },
		{ 258, 159, 31 },
		{ 258, 160, 31 },
		{ 258, 161, 31 },
		{ 258, 162, 31 },
		{ 258, 163, 31 },
		{ 258, 164, 31 },
		{ 258, 165, 31 },
		{ 258, 166, 31 },
		{ 258, 167, 31 },
		{ 258, 168, 31 },
		{ 258, 169, 31 },
		{ 258, 170, 31 },
		{ 258, 171, 31 },
		{ 258, 172, 31 },
		{ 258, 173, 31 },
		{ 258, 174, 31 },
		{ 258, 175, 31 },
		{ 258, 176, 31 },
		{ 258, 177, 31 },
		{ 258, 178, 31 },
		{ 258, 179, 31 },
		{ 258, 180, 31 },
		{ 258, 181, 31 },
		{ 258, 182, 31 },
		{ 258, 183, 31 },
		{ 258, 184, 31 },
		{ 258, 185, 31 },
		{ 258, 186, 31 },
		{ 258, 187, 31 },
		{ 258, 188, 31 },
		{ 258, 189, 31 },
		{ 258, 190, 31 },
		{ 259, 1, 31 },
		{ 259, 2, 31 },
		{ 259, 3, 31 },
		{ 259, 4, 31 },
		{ 259, 5, 31 },
		{ 259, 6, 31 },
		{ 259, 7, 31 },
		{ 259, 8, 31 },
		{ 259, 9, 31 },
		{ 259, 10, 31 },
		{ 259, 11, 31 },
		{ 259, 12, 31 },
		{ 259, 13, 31 },
		{ 259, 14, 31 },
		{ 259, 15, 31 },
		{ 259, 16, 31 },
		{ 259, 17, 31 },
		{ 259, 18, 31 },
		{ 259, 19, 31 },
		{ 259, 20, 31 },
		{ 259, 21, 31 },
		{ 259, 22, 31 },
		{ 259, 23, 31 },
		{ 259, 24, 31 },
		{ 259, 25, 31 },
		{ 259, 26, 31 },
		{ 259, 27, 31 },
		{ 259, 28, 31 },
		{ 259, 29, 31 },
		{ 259, 30, 31 },
		{ 259, 31, 31 },
		{ 259, 32, 31 },
		{ 259, 33, 31 },
		{ 259, 34, 31 },
		{ 259, 35, 31 },
		{ 259, 36, 31 },
		{ 259, 37, 31 },
		{ 259, 38, 31 },
		{ 259, 39, 31 },
		{ 259, 40, 31 },
		{ 259, 41, 31 },
		{ 259, 42, 31 },
		{ 259, 43, 31 },
		{ 259, 44, 31 },
		{ 259, 45, 31 },
		{ 259, 46, 31 },
		{ 259, 47, 31 },
		{ 259, 48, 31 },
		{ 259, 49, 31 },
		{ 259, 50, 31 },
		{ 259, 51, 31 },
		{ 259, 52, 31 },
		{ 259, 53, 31 },
		{ 259, 54, 31 },
		{ 259, 55, 31 },
		{ 259, 56, 31 },
		{ 259, 57, 31 },
		{ 259, 58, 31 },
		{ 259, 59, 31 },
		{ 259, 60, 31 },
		{ 259, 61, 31 },
		{ 259, 62, 31 },
		{ 259, 63, 31 },
		{ 259, 64, 31 },
		{ 259, 65, 31 },
		{ 259, 66, 31 },
		{ 259, 67, 31 },
		{ 259, 68, 31 },
		{ 259, 69, 31 },
		{ 259, 70, 31 },
		{ 259, 71, 31 },
		{ 259, 72, 31 },
		{ 259, 73, 31 },
		{ 259, 74, 31 },
		{ 259, 75, 31 },
		{ 259, 76, 31 },
		{ 259, 77, 31 },
		{ 259, 78, 31 },
		{ 259, 79, 31 },
		{ 259, 80, 31 },
		{ 259, 81, 31 },
		{ 259, 82, 31 },
		{ 259, 83, 31 },
		{ 259, 84, 31 },
		{ 259, 85, 31 },
		{ 259, 86, 31 },
		{ 259, 87, 31 },
		{ 259, 88, 31 },
		{ 259, 89, 31 },
		{ 259, 90, 31 },
		{ 259, 91, 31 },
		{ 259, 92, 31 },
		{ 259, 93, 31 },
		{ 259, 94, 31 },
		{ 259, 95, 31 },
		{ 259, 96, 31 },
		{ 259, 97, 31 },
		{ 259, 98, 31 },
		{ 259, 99, 31 },
		{ 259, 100, 31 },
		{ 259, 101, 31 },
		{ 259, 102, 31 },
		{ 259, 103, 31 },
		{ 259, 104, 31 },
		{ 259, 105, 31 },
		{ 259, 106, 31 },
		{ 259, 107, 31 },
		{ 259, 108, 31 },
		{ 259, 109, 31 },
		{ 259, 110, 31 },
		{ 259, 111, 31 },
		{ 259, 112, 31 },
		{ 259, 113, 31 },
		{ 259, 114, 31 },
		{ 259, 115, 31 },
		{ 259, 116, 31 },
		{ 259, 117, 31 },
		{ 259, 118, 31 },
		{ 259, 119, 31 },
		{ 259, 120, 31 },
		{ 259, 121, 31 },
		{ 259, 122, 31 },
		{ 259, 123, 31 },
		{ 259, 124, 31 },
		{ 259, 125, 31 },
		{ 259, 126, 31 },
		{ 259, 127, 31 },
		{ 259, 128, 31 },
		{ 259, 129, 31 },
		{ 259, 130, 31 },
		{ 259, 131, 31 },
		{ 259, 132, 31 },
		{ 259, 133, 31 },
		{ 259, 134, 31 },
		{ 259, 135, 31 },
		{ 259, 136, 31 },
		{ 259, 137, 31 },
		{ 259, 138, 31 },
		{ 259, 139, 31 },
		{ 259, 140, 31 },
		{ 259, 141, 31 },
		{ 259, 142, 31 },
		{ 259, 143, 31 },
		{ 259, 144, 31 },
		{ 259, 145, 31 },
		{ 259, 146, 31 },
		{ 259, 147, 31 },
		{ 259, 148, 31 },
		{ 259, 149, 31 },
		{ 259, 150, 31 },
		{ 259, 151, 31 },
		{ 259, 152, 31 },
		{ 259, 153, 31 },
		{ 259, 154, 31 },
		{ 259, 155, 31 },
		{ 259, 156, 31 },
		{ 259, 157, 31 },
		{ 259, 158, 31 },
		{ 259, 159, 31 },
		{ 259, 160, 31 },
		{ 259, 161, 31 },
		{ 259, 162, 31 },
		{ 259, 163, 31 },
		{ 259, 164, 31 },
		{ 259, 165, 31 },
		{ 259, 166, 31 },
		{ 259, 167, 31 },
		{ 259, 168, 31 },
		{ 259, 169, 31 },
		{ 259, 170, 31 },
		{ 259, 171, 31 },
		{ 259, 172, 31 },
		{ 259, 173, 31 },
		{ 259, 174, 31 },
		{ 259, 175, 31 },
		{ 259, 176, 31 },
		{ 259, 177, 31 },
		{ 259, 178, 31 },
		{ 259, 179, 31 },
		{ 259, 180, 31 },
		{ 259, 181, 31 },
		{ 259, 182, 31 },
		{ 259, 183, 31 },
		{ 259, 184, 31 },
		{ 259, 185, 31 },
		{ 259, 186, 31 },
		{ 259, 187, 31 },
		{ 259, 188, 31 },
		{ 259, 189, 31 },
		{ 259, 190, 31 },
		{ 259, 191, 31 },
		{ 260, 1, 31 },
		{ 260, 2, 31 },
		{ 260, 3, 31 },
		{ 260, 4, 31 },
		{ 260, 5, 31 },
		{ 260, 6, 31 },
		{ 260, 7, 31 },
		{ 260, 8, 31 },
		{ 260, 9, 31 },
		{ 260, 10, 31 },
		{ 260, 11, 31 },
		{ 260, 12, 31 },
		{ 260, 13, 31 },
		{ 260, 14, 31 },
		{ 260, 15, 31 },
		{ 260, 16, 31 },
		{ 260, 17, 31 },
		{ 260, 18, 31 },
		{ 260, 19, 31 },
		{ 260, 20, 31 },
		{ 260, 21, 31 },
		{ 260, 22, 31 },
		{ 260, 23, 31 },
		{ 260, 24, 31 },
		{ 260, 25, 31 },
		{ 260, 26, 31 },
		{ 260, 27, 31 },
		{ 260, 28, 31 },
		{ 260, 29, 31 },
		{ 260, 30, 31 },
		{ 260, 31, 31 },
		{ 260, 32, 31 },
		{ 260, 33, 31 },
		{ 260, 34, 31 },
		{ 260, 35, 31 },
		{ 260, 36, 31 },
		{ 260, 37, 31 },
		{ 260, 38, 31 },
		{ 260, 39, 31 },
		{ 260, 40, 31 },
		{ 260, 41, 31 },
		{ 260, 42, 31 },
		{ 260, 43, 31 },
		{ 260, 44, 31 },
		{ 260, 45, 31 },
		{ 260, 46, 31 },
		{ 260, 47, 31 },
		{ 260, 48, 31 },
		{ 260, 49, 31 },
		{ 260, 50, 31 },
		{ 260, 51, 31 },
		{ 260, 52, 31 },
		{ 260, 53, 31 },
		{ 260, 54, 31 },
		{ 260, 55, 31 },
		{ 260, 56, 31 },
		{ 260, 57, 31 },
		{ 260, 58, 31 },
		{ 260, 59, 31 },
		{ 260, 60, 31 },
		{ 260, 61, 31 },
		{ 260, 62, 31 },
		{ 260, 63, 31 },
		{ 260, 64, 31 },
		{ 260, 65, 31 },
		{ 260, 66, 31 },
		{ 260, 67, 31 },
		{ 260, 68, 31 },
		{ 260, 69, 31 },
		{ 260, 70, 31 },
		{ 260, 71, 31 },
		{ 260, 72, 31 },
		{ 260, 73, 31 },
		{ 260, 74, 31 },
		{ 260, 75, 31 },
		{ 260, 76, 31 },
		{ 260, 77, 31 },
		{ 260, 78, 31 },
		{ 260, 79, 31 },
		{ 260, 80, 31 },
		{ 260, 81, 31 },
		{ 260, 82, 31 },
		{ 260, 83, 31 },
		{ 260, 84, 31 },
		{ 260, 85, 31 },
		{ 260, 86, 31 },
		{ 260, 87, 31 },
		{ 260, 88, 31 },
		{ 260, 89, 31 },
		{ 260, 90, 31 },
		{ 260, 91, 31 },
		{ 260, 92, 31 },
		{ 260, 93, 31 },
		{ 260, 94, 31 },
		{ 260, 95, 31 },
		{ 260, 96, 31 },
		{ 260, 97, 31 },
		{ 260, 98, 31 },
		{ 260, 99, 31 },
		{ 260, 100, 31 },
		{ 260, 101, 31 },
		{ 260, 102, 31 },
		{ 260, 103, 31 },
		{ 260, 104, 31 },
		{ 260, 105, 31 },
		{ 260, 106, 31 },
		{ 260, 107, 31 },
		{ 260, 108, 31 },
		{ 260, 109, 31 },
		{ 260, 110, 31 },
		{ 260, 111, 31 },
		{ 260, 112, 31 },
		{ 260, 113, 31 },
		{ 260, 114, 31 },
		{ 260, 115, 31 },
		{ 260, 116, 31 },
		{ 260, 117, 31 },
		{ 260, 118, 31 },
		{ 260, 119, 31 },
		{ 260, 120, 31 },
		{ 260, 121, 31 },
		{ 260, 122, 31 },
		{ 260, 123, 31 },
		{ 260, 124, 31 },
		{ 260, 125, 31 },
		{ 260, 126, 31 },
		{ 260, 127, 31 },
		{ 260, 128, 31 },
		{ 260, 129, 31 },
		{ 260, 130, 31 },
		{ 260, 131, 31 },
		{ 260, 132, 31 },
		{ 260, 133, 31 },
		{ 260, 134, 31 },
		{ 260, 135, 31 },
		{ 260, 136, 31 },
		{ 260, 137, 31 },
		{ 260, 138, 31 },
		{ 260, 139, 31 },
		{ 260, 140, 31 },
		{ 260, 141, 31 },
		{ 260, 142, 31 },
		{ 260, 143, 31 },
		{ 260, 144, 31 },
		{ 260, 145, 31 },
		{ 260, 146, 31 },
		{ 260, 147, 31 },
		{ 260, 148, 31 },
		{ 260, 149, 31 },
		{ 260, 150, 31 },
		{ 260, 151, 31 },
		{ 260, 152, 31 },
		{ 260, 153, 31 },
		{ 260, 154, 31 },
		{ 260, 155, 31 },
		{ 260, 156, 31 },
		{ 260, 157, 31 },
		{ 260, 158, 31 },
		{ 260, 159, 31 },
		{ 260, 160, 31 },
		{ 260, 161, 31 },
		{ 260, 162, 31 },
		{ 260, 163, 31 },
		{ 260, 164, 31 },
		{ 260, 165, 31 },
		{ 260, 166, 31 },
		{ 260, 167, 31 },
		{ 260, 168, 31 },
		{ 260, 169, 31 },
		{ 260, 170, 31 },
		{ 260, 171, 31 },
		{ 260, 172, 31 },
		{ 260, 173, 31 },
		{ 260, 174, 31 },
		{ 260, 175, 31 },
		{ 260, 176, 31 },
		{ 260, 177, 31 },
		{ 260, 178, 31 },
		{ 260, 179, 31 },
		{ 260, 180, 31 },
		{ 260, 181, 31 },
		{ 260, 182, 31 },
		{ 260, 183, 31 },
		{ 260, 184, 31 },
		{ 260, 185, 31 },
		{ 260, 186, 31 },
		{ 260, 187, 31 },
		{ 260, 188, 31 },
		{ 260, 189, 31 },
		{ 260, 190, 31 },
		{ 260, 191, 31 },
		{ 261, 1, 31 },
		{ 261, 2, 31 },
		{ 261, 3, 31 },
		{ 261, 4, 31 },
		{ 261, 5, 31 },
		{ 261, 6, 31 },
		{ 261, 7, 31 },
		{ 261, 8, 31 },
		{ 261, 9, 31 },
		{ 261, 10, 31 },
		{ 261, 11, 31 },
		{ 261, 12, 31 },
		{ 261, 13, 31 },
		{ 261, 14, 31 },
		{ 261, 15, 31 },
		{ 261, 16, 31 },
		{ 261, 17, 31 },
		{ 261, 18, 31 },
		{ 261, 19, 31 },
		{ 261, 20, 31 },
		{ 261, 21, 31 },
		{ 261, 22, 31 },
		{ 261, 23, 31 },
		{ 261, 24, 31 },
		{ 261, 25, 31 },
		{ 261, 26, 31 },
		{ 261, 27, 31 },
		{ 261, 28, 31 },
		{ 261, 29, 31 },
		{ 261, 30, 31 },
		{ 261, 31, 31 },
		{ 261, 32, 31 },
		{ 261, 33, 31 },
		{ 261, 34, 31 },
		{ 261, 35, 31 },
		{ 261, 36, 31 },
		{ 261, 37, 31 },
		{ 261, 38, 31 },
		{ 261, 39, 31 },
		{ 261, 40, 31 },
		{ 261, 41, 31 },
		{ 261, 42, 31 },
		{ 261, 43, 31 },
		{ 261, 44, 31 },
		{ 261, 45, 31 },
		{ 261, 46, 31 },
		{ 261, 47, 31 },
		{ 261, 48, 31 },
		{ 261, 49, 31 },
		{ 261, 50, 31 },
		{ 261, 51, 31 },
		{ 261, 52, 31 },
		{ 261, 53, 31 },
		{ 261, 54, 31 },
		{ 261, 55, 31 },
		{ 261, 56, 31 },
		{ 261, 57, 31 },
		{ 261, 58, 31 },
		{ 261, 59, 31 },
		{ 261, 60, 31 },
		{ 261, 61, 31 },
		{ 261, 62, 31 },
		{ 261, 63, 31 },
		{ 261, 64, 31 },
		{ 261, 65, 31 },
		{ 261, 66, 31 },
		{ 261, 67, 31 },
		{ 261, 68, 31 },
		{ 261, 69, 31 },
		{ 261, 70, 31 },
		{ 261, 71, 31 },
		{ 261, 72, 31 },
		{ 261, 73, 31 },
		{ 261, 74, 31 },
		{ 261, 75, 31 },
		{ 261, 76, 31 },
		{ 261, 77, 31 },
		{ 261, 78, 31 },
		{ 261, 79, 31 },
		{ 261, 80, 31 },
		{ 261, 81, 31 },
		{ 261, 82, 31 },
		{ 261, 83, 31 },
		{ 261, 84, 31 },
		{ 261, 85, 31 },
		{ 261, 86, 31 },
		{ 261, 87, 31 },
		{ 261, 88, 31 },
		{ 261, 89, 31 },
		{ 261, 90, 31 },
		{ 261, 91, 31 },
		{ 261, 92, 31 },
		{ 261, 93, 31 },
		{ 261, 94, 31 },
		{ 261, 95, 31 },
		{ 261, 96, 31 },
		{ 261, 97, 31 },
		{ 261, 98, 31 },
		{ 261, 99, 31 },
		{ 261, 100, 31 },
		{ 261, 101, 31 },
		{ 261, 102, 31 },
		{ 261, 103, 31 },
		{ 261, 104, 31 },
		{ 261, 105, 31 },
		{ 261, 106, 31 },
		{ 261, 107, 31 },
		{ 261, 108, 31 },
		{ 261, 109, 31 },
		{ 261, 110, 31 },
		{ 261, 111, 31 },
		{ 261, 112, 31 },
		{ 261, 113, 31 },
		{ 261, 114, 31 },
		{ 261, 115, 31 },
		{ 261, 116, 31 },
		{ 261, 117, 31 },
		{ 261, 118, 31 },
		{ 261, 119, 31 },
		{ 261, 120, 31 },
		{ 261, 121, 31 },
		{ 261, 122, 31 },
		{ 261, 123, 31 },
		{ 261, 124, 31 },
		{ 261, 125, 31 },
		{ 261, 126, 31 },
		{ 261, 127, 31 },
		{ 261, 128, 31 },
		{ 261, 129, 31 },
		{ 261, 130, 31 },
		{ 261, 131, 31 },
		{ 261, 132, 31 },
		{ 261, 133, 31 },
		{ 261, 134, 31 },
		{ 261, 135, 31 },
		{ 261, 136, 31 },
		{ 261, 137, 31 },
		{ 261, 138, 31 },
		{ 261, 139, 31 },
		{ 261, 140, 31 },
		{ 261, 141, 31 },
		{ 261, 142, 31 },
		{ 261, 143, 31 },
		{ 261, 144, 31 },
		{ 261, 145, 31 },
		{ 261, 146, 31 },
		{ 261, 147, 31 },
		{ 261, 148, 31 },
		{ 261, 149, 31 },
		{ 261, 150, 31 },
		{ 261, 151, 31 },
		{ 261, 152, 31 },
		{ 261, 153, 31 },
		{ 261, 154, 31 },
		{ 261, 155, 31 },
		{ 261, 156, 31 },
		{ 261, 157, 31 },
		{ 261, 158, 31 },
		{ 261, 159, 31 },
		{ 261, 160, 31 },
		{ 261, 161, 31 },
		{ 261, 162, 31 },
		{ 261, 163, 31 },
		{ 261, 164, 31 },
		{ 261, 165, 31 },
		{ 261, 166, 31 },
		{ 261, 167, 31 },
		{ 261, 168, 31 },
		{ 261, 169, 31 },
		{ 261, 170, 31 },
		{ 261, 171, 31 },
		{ 261, 172, 31 },
		{ 261, 173, 31 },
		{ 261, 174, 31 },
		{ 261, 175, 31 },
		{ 261, 176, 31 },
		{ 261, 177, 31 },
		{ 261, 178, 31 },
		{ 261, 179, 31 },
		{ 261, 180, 31 },
		{ 261, 181, 31 },
		{ 261, 182, 31 },
		{ 261, 183, 31 },
		{ 261, 184, 31 },
		{ 261, 185, 31 },
		{ 261, 186, 31 },
		{ 261, 187, 31 },
		{ 261, 188, 31 },
		{ 261, 189, 31 },
		{ 261, 190, 31 },
		{ 261, 191, 31 },
		{ 262, 1, 31 },
		{ 262, 2, 31 },
		{ 262, 3, 31 },
		{ 262, 4, 31 },
		{ 262, 5, 31 },
		{ 262, 6, 31 },
		{ 262, 7, 31 },
		{ 262, 8, 31 },
		{ 262, 9, 31 },
		{ 262, 10, 31 },
		{ 262, 11, 31 },
		{ 262, 12, 31 },
		{ 262, 13, 31 },
		{ 262, 14, 31 },
		{ 262, 15, 31 },
		{ 262, 16, 31 },
		{ 262, 17, 31 },
		{ 262, 18, 31 },
		{ 262, 19, 31 },
		{ 262, 20, 31 },
		{ 262, 21, 31 },
		{ 262, 22, 31 },
		{ 262, 23, 31 },
		{ 262, 24, 31 },
		{ 262, 25, 31 },
		{ 262, 26, 31 },
		{ 262, 27, 31 },
		{ 262, 28, 31 },
		{ 262, 29, 31 },
		{ 262, 30, 31 },
		{ 262, 31, 31 },
		{ 262, 32, 31 },
		{ 262, 33, 31 },
		{ 262, 34, 31 },
		{ 262, 35, 31 },
		{ 262, 36, 31 },
		{ 262, 37, 31 },
		{ 262, 38, 31 },
		{ 262, 39, 31 },
		{ 262, 40, 31 },
		{ 262, 41, 31 },
		{ 262, 42, 31 },
		{ 262, 43, 31 },
		{ 262, 44, 31 },
		{ 262, 45, 31 },
		{ 262, 46, 31 },
		{ 262, 47, 31 },
		{ 262, 48, 31 },
		{ 262, 49, 31 },
		{ 262, 50, 31 },
		{ 262, 51, 31 },
		{ 262, 52, 31 },
		{ 262, 53, 31 },
		{ 262, 54, 31 },
		{ 262, 55, 31 },
		{ 262, 56, 31 },
		{ 262, 57, 31 },
		{ 262, 58, 31 },
		{ 262, 59, 31 },
		{ 262, 60, 31 },
		{ 262, 61, 31 },
		{ 262, 62, 31 },
		{ 262, 63, 31 },
		{ 262, 64, 31 },
		{ 262, 65, 31 },
		{ 262, 66, 31 },
		{ 262, 67, 31 },
		{ 262, 68, 31 },
		{ 262, 69, 31 },
		{ 262, 70, 31 },
		{ 262, 71, 31 },
		{ 262, 72, 31 },
		{ 262, 73, 31 },
		{ 262, 74, 31 },
		{ 262, 75, 31 },
		{ 262, 76, 31 },
		{ 262, 77, 31 },
		{ 262, 78, 31 },
		{ 262, 79, 31 },
		{ 262, 80, 31 },
		{ 262, 81, 31 },
		{ 262, 82, 31 },
		{ 262, 83, 31 },
		{ 262, 84, 31 },
		{ 262, 85, 31 },
		{ 262, 86, 31 },
		{ 262, 87, 31 },
		{ 262, 88, 31 },
		{ 262, 89, 31 },
		{ 262, 90, 31 },
		{ 262, 91, 31 },
		{ 262, 92, 31 },
		{ 262, 93, 31 },
		{ 262, 94, 31 },
		{ 262, 95, 31 },
		{ 262, 96, 31 },
		{ 262, 97, 31 },
		{ 262, 98, 31 },
		{ 262, 99, 31 },
		{ 262, 100, 31 },
		{ 262, 101, 31 },
		{ 262, 102, 31 },
		{ 262, 103, 31 },
		{ 262, 104, 31 },
		{ 262, 105, 31 },
		{ 262, 106, 31 },
		{ 262, 107, 31 },
		{ 262, 108, 31 },
		{ 262, 109, 31 },
		{ 262, 110, 31 },
		{ 262, 111, 31 },
		{ 262, 112, 31 },
		{ 262, 113, 31 },
		{ 262, 114, 31 },
		{ 262, 115, 31 },
		{ 262, 116, 31 },
		{ 262, 117, 31 },
		{ 262, 118, 31 },
		{ 262, 119, 31 },
		{ 262, 120, 31 },
		{ 262, 121, 31 },
		{ 262, 122, 31 },
		{ 262, 123, 31 },
		{ 262, 124, 31 },
		{ 262, 125, 31 },
		{ 262, 126, 31 },
		{ 262, 127, 31 },
		{ 262, 128, 31 },
		{ 262, 129, 31 },
		{ 262, 130, 31 },
		{ 262, 131, 31 },
		{ 262, 132, 31 },
		{ 262, 133, 31 },
		{ 262, 134, 31 },
		{ 262, 135, 31 },
		{ 262, 136, 31 },
		{ 262, 137, 31 },
		{ 262, 138, 31 },
		{ 262, 139, 31 },
		{ 262, 140, 31 },
		{ 262, 141, 31 },
		{ 262, 142, 31 },
		{ 262, 143, 31 },
		{ 262, 144, 31 },
		{ 262, 145, 31 },
		{ 262, 146, 31 },
		{ 262, 147, 31 },
		{ 262, 148, 31 },
		{ 262, 149, 31 },
		{ 262, 150, 31 },
		{ 262, 151, 31 },
		{ 262, 152, 31 },
		{ 262, 153, 31 },
		{ 262, 154, 31 },
		{ 262, 155, 31 },
		{ 262, 156, 31 },
		{ 262, 157, 31 },
		{ 262, 158, 31 },
		{ 262, 159, 31 },
		{ 262, 160, 31 },
		{ 262, 161, 31 },
		{ 262, 162, 31 },
		{ 262, 163, 31 },
		{ 262, 164, 31 },
		{ 262, 165, 31 },
		{ 262, 166, 31 },
		{ 262, 167, 31 },
		{ 262, 168, 31 },
		{ 262, 169, 31 },
		{ 262, 170, 31 },
		{ 262, 171, 31 },
		{ 262, 172, 31 },
		{ 262, 173, 31 },
		{ 262, 174, 31 },
		{ 262, 175, 31 },
		{ 262, 176, 31 },
		{ 262, 177, 31 },
		{ 262, 178, 31 },
		{ 262, 179, 31 },
		{ 262, 180, 31 },
		{ 262, 181, 31 },
		{ 262, 182, 31 },
		{ 262, 183, 31 },
		{ 262, 184, 31 },
		{ 262, 185, 31 },
		{ 262, 186, 31 },
		{ 262, 187, 31 },
		{ 262, 188, 31 },
		{ 262, 189, 31 },
		{ 262, 190, 31 },
		{ 262, 191, 31 },
		{ 263, 1, 31 },
		{ 263, 2, 31 },
		{ 263, 3, 31 },
		{ 263, 4, 31 },
		{ 263, 5, 31 },
		{ 263, 6, 31 },
		{ 263, 7, 31 },
		{ 263, 8, 31 },
		{ 263, 9, 31 },
		{ 263, 10, 31 },
		{ 263, 11, 31 },
		{ 263, 12, 31 },
		{ 263, 13, 31 },
		{ 263, 14, 31 },
		{ 263, 15, 31 },
		{ 263, 16, 31 },
		{ 263, 17, 31 },
		{ 263, 18, 31 },
		{ 263, 19, 31 },
		{ 263, 20, 31 },
		{ 263, 21, 31 },
		{ 263, 22, 31 },
		{ 263, 23, 31 },
		{ 263, 24, 31 },
		{ 263, 25, 31 },
		{ 263, 26, 31 },
		{ 263, 27, 31 },
		{ 263, 28, 31 },
		{ 263, 29, 31 },
		{ 263, 30, 31 },
		{ 263, 31, 31 },
		{ 263, 32, 31 },
		{ 263, 33, 31 },
		{ 263, 34, 31 },
		{ 263, 35, 31 },
		{ 263, 36, 31 },
		{ 263, 37, 31 },
		{ 263, 38, 31 },
		{ 263, 39, 31 },
		{ 263, 40, 31 },
		{ 263, 41, 31 },
		{ 263, 42, 31 },
		{ 263, 43, 31 },
		{ 263, 44, 31 },
		{ 263, 45, 31 },
		{ 263, 46, 31 },
		{ 263, 47, 31 },
		{ 263, 48, 31 },
		{ 263, 49, 31 },
		{ 263, 50, 31 },
		{ 263, 51, 31 },
		{ 263, 52, 31 },
		{ 263, 53, 31 },
		{ 263, 54, 31 },
		{ 263, 55, 31 },
		{ 263, 56, 31 },
		{ 263, 57, 31 },
		{ 263, 58, 31 },
		{ 263, 59, 31 },
		{ 263, 60, 31 },
		{ 263, 61, 31 },
		{ 263, 62, 31 },
		{ 263, 63, 31 },
		{ 263, 64, 31 },
		{ 263, 65, 31 },
		{ 263, 66, 31 },
		{ 263, 67, 31 },
		{ 263, 68, 31 },
		{ 263, 69, 31 },
		{ 263, 70, 31 },
		{ 263, 71, 31 },
		{ 263, 72, 31 },
		{ 263, 73, 31 },
		{ 263, 74, 31 },
		{ 263, 75, 31 },
		{ 263, 76, 31 },
		{ 263, 77, 31 },
		{ 263, 78, 31 },
		{ 263, 79, 31 },
		{ 263, 80, 31 },
		{ 263, 81, 31 },
		{ 263, 82, 31 },
		{ 263, 83, 31 },
		{ 263, 84, 31 },
		{ 263, 85, 31 },
		{ 263, 86, 31 },
		{ 263, 87, 31 },
		{ 263, 88, 31 },
		{ 263, 89, 31 },
		{ 263, 90, 31 },
		{ 263, 91, 31 },
		{ 263, 92, 31 },
		{ 263, 93, 31 },
		{ 263, 94, 31 },
		{ 263, 95, 31 },
		{ 263, 96, 31 },
		{ 263, 97, 31 },
		{ 263, 98, 31 },
		{ 263, 99, 31 },
		{ 263, 100, 31 },
		{ 263, 101, 31 },
		{ 263, 102, 31 },
		{ 263, 103, 31 },
		{ 263, 104, 31 },
		{ 263, 105, 31 },
		{ 263, 106, 31 },
		{ 263, 107, 31 },
		{ 263, 108, 31 },
		{ 263, 109, 31 },
		{ 263, 110, 31 },
		{ 263, 111, 31 },
		{ 263, 112, 31 },
		{ 263, 113, 31 },
		{ 263, 114, 31 },
		{ 263, 115, 31 },
		{ 263, 116, 31 },
		{ 263, 117, 31 },
		{ 263, 118, 31 },
		{ 263, 119, 31 },
		{ 263, 120, 31 },
		{ 263, 121, 31 },
		{ 263, 122, 31 },
		{ 263, 123, 31 },
		{ 263, 124, 31 },
		{ 263, 125, 31 },
		{ 263, 126, 31 },
		{ 263, 127, 31 },
		{ 263, 128, 31 },
		{ 263, 129, 31 },
		{ 263, 130, 31 },
		{ 263, 131, 31 },
		{ 263, 132, 31 },
		{ 263, 133, 31 },
		{ 263, 134, 31 },
		{ 263, 135, 31 },
		{ 263, 136, 31 },
		{ 263, 137, 31 },
		{ 263, 138, 31 },
		{ 263, 139, 31 },
		{ 263, 140, 31 },
		{ 263, 141, 31 },
		{ 263, 142, 31 },
		{ 263, 143, 31 },
		{ 263, 144, 31 },
		{ 263, 145, 31 },
		{ 263, 146, 31 },
		{ 263, 147, 31 },
		{ 263, 148, 31 },
		{ 263, 149, 31 },
		{ 263, 150, 31 },
		{ 263, 151, 31 },
		{ 263, 152, 31 },
		{ 263, 153, 31 },
		{ 263, 154, 31 },
		{ 263, 155, 31 },
		{ 263, 156, 31 },
		{ 263, 157, 31 },
		{ 263, 158, 31 },
		{ 263, 159, 31 },
		{ 263, 160, 31 },
		{ 263, 161, 31 },
		{ 263, 162, 31 },
		{ 263, 163, 31 },
		{ 263, 164, 31 },
		{ 263, 165, 31 },
		{ 263, 166, 31 },
		{ 263, 167, 31 },
		{ 263, 168, 31 },
		{ 263, 169, 31 },
		{ 263, 170, 31 },
		{ 263, 171, 31 },
		{ 263, 172, 31 },
		{ 263, 173, 31 },
		{ 263, 174, 31 },
		{ 263, 175, 31 },
		{ 263, 176, 31 },
		{ 263, 177, 31 },
		{ 263, 178, 31 },
		{ 263, 179, 31 },
		{ 263, 180, 31 },
		{ 263, 181, 31 },
		{ 263, 182, 31 },
		{ 263, 183, 31 },
		{ 263, 184, 31 },
		{ 263, 185, 31 },
		{ 263, 186, 31 },
		{ 263, 187, 31 },
		{ 263, 188, 31 },
		{ 263, 189, 31 },
		{ 263, 190, 31 },
		{ 263, 191, 31 },
		{ 264, 1, 31 },
		{ 264, 2, 31 },
		{ 264, 3, 31 },
		{ 264, 4, 31 },
		{ 264, 5, 31 },
		{ 264, 6, 31 },
		{ 264, 7, 31 },
		{ 264, 8, 31 },
		{ 264, 9, 31 },
		{ 264, 10, 31 },
		{ 264, 11, 31 },
		{ 264, 12, 31 },
		{ 264, 13, 31 },
		{ 264, 14, 31 },
		{ 264, 15, 31 },
		{ 264, 16, 31 },
		{ 264, 17, 31 },
		{ 264, 18, 31 },
		{ 264, 19, 31 },
		{ 264, 20, 31 },
		{ 264, 21, 31 },
		{ 264, 22, 31 },
		{ 264, 23, 31 },
		{ 264, 24, 31 },
		{ 264, 25, 31 },
		{ 264, 26, 31 },
		{ 264, 27, 31 },
		{ 264, 28, 31 },
		{ 264, 29, 31 },
		{ 264, 30, 31 },
		{ 264, 31, 31 },
		{ 264, 32, 31 },
		{ 264, 33, 31 },
		{ 264, 34, 31 },
		{ 264, 35, 31 },
		{ 264, 36, 31 },
		{ 264, 37, 31 },
		{ 264, 38, 31 },
		{ 264, 39, 31 },
		{ 264, 40, 31 },
		{ 264, 41, 31 },
		{ 264, 42, 31 },
		{ 264, 43, 31 },
		{ 264, 44, 31 },
		{ 264, 45, 31 },
		{ 264, 46, 31 },
		{ 264, 47, 31 },
		{ 264, 48, 31 },
		{ 264, 49, 31 },
		{ 264, 50, 31 },
		{ 264, 51, 31 },
		{ 264, 52, 31 },
		{ 264, 53, 31 },
		{ 264, 54, 31 },
		{ 264, 55, 31 },
		{ 264, 56, 31 },
		{ 264, 57, 31 },
		{ 264, 58, 31 },
		{ 264, 59, 31 },
		{ 264, 60, 31 },
		{ 264, 61, 31 },
		{ 264, 62, 31 },
		{ 264, 63, 31 },
		{ 264, 64, 31 },
		{ 264, 65, 31 },
		{ 264, 66, 31 },
		{ 264, 67, 31 },
		{ 264, 68, 31 },
		{ 264, 69, 31 },
		{ 264, 70, 31 },
		{ 264, 71, 31 },
		{ 264, 72, 31 },
		{ 264, 73, 31 },
		{ 264, 74, 31 },
		{ 264, 75, 31 },
		{ 264, 76, 31 },
		{ 264, 77, 31 },
		{ 264, 78, 31 },
		{ 264, 79, 31 },
		{ 264, 80, 31 },
		{ 264, 81, 31 },
		{ 264, 82, 31 },
		{ 264, 83, 31 },
		{ 264, 84, 31 },
		{ 264, 85, 31 },
		{ 264, 86, 31 },
		{ 264, 87, 31 },
		{ 264, 88, 31 },
		{ 264, 89, 31 },
		{ 264, 90, 31 },
		{ 264, 91, 31 },
		{ 264, 92, 31 },
		{ 264, 93, 31 },
		{ 264, 94, 31 },
		{ 264, 95, 31 },
		{ 264, 96, 31 },
		{ 264, 97, 31 },
		{ 264, 98, 31 },
		{ 264, 99, 31 },
		{ 264, 100, 31 },
		{ 264, 101, 31 },
		{ 264, 102, 31 },
		{ 264, 103, 31 },
		{ 264, 104, 31 },
		{ 264, 105, 31 },
		{ 264, 106, 31 },
		{ 264, 107, 31 },
		{ 264, 108, 31 },
		{ 264, 109, 31 },
		{ 264, 110, 31 },
		{ 264, 111, 31 },
		{ 264, 112, 31 },
		{ 264, 113, 31 },
		{ 264, 114, 31 },
		{ 264, 115, 31 },
		{ 264, 116, 31 },
		{ 264, 117, 31 },
		{ 264, 118, 31 },
		{ 264, 119, 31 },
		{ 264, 120, 31 },
		{ 264, 121, 31 },
		{ 264, 122, 31 },
		{ 264, 123, 31 },
		{ 264, 124, 31 },
		{ 264, 125, 31 },
		{ 264, 126, 31 },
		{ 264, 127, 31 },
		{ 264, 128, 31 },
		{ 264, 129, 31 },
		{ 264, 130, 31 },
		{ 264, 131, 31 },
		{ 264, 132, 31 },
		{ 264, 133, 31 },
		{ 264, 134, 31 },
		{ 264, 135, 31 },
		{ 264, 136, 31 },
		{ 264, 137, 31 },
		{ 264, 138, 31 },
		{ 264, 139, 31 },
		{ 264, 140, 31 },
		{ 264, 141, 31 },
		{ 264, 142, 31 },
		{ 264, 143, 31 },
		{ 264, 144, 31 },
		{ 264, 145, 31 },
		{ 264, 146, 31 },
		{ 264, 147, 31 },
		{ 264, 148, 31 },
		{ 264, 149, 31 },
		{ 264, 150, 31 },
		{ 264, 151, 31 },
		{ 264, 152, 31 },
		{ 264, 153, 31 },
		{ 264, 154, 31 },
		{ 264, 155, 31 },
		{ 264, 156, 31 },
		{ 264, 157, 31 },
		{ 264, 158, 31 },
		{ 264, 159, 31 },
		{ 264, 160, 31 },
		{ 264, 161, 31 },
		{ 264, 162, 31 },
		{ 264, 163, 31 },
		{ 264, 164, 31 },
		{ 264, 165, 31 },
		{ 264, 166, 31 },
		{ 264, 167, 31 },
		{ 264, 168, 31 },
		{ 264, 169, 31 },
		{ 264, 170, 31 },
		{ 264, 171, 31 },
		{ 264, 172, 31 },
		{ 264, 173, 31 },
		{ 264, 174, 31 },
		{ 264, 175, 31 },
		{ 264, 176, 31 },
		{ 264, 177, 31 },
		{ 264, 178, 31 },
		{ 264, 179, 31 },
		{ 264, 180, 31 },
		{ 264, 181, 31 },
		{ 264, 182, 31 },
		{ 264, 183, 31 },
		{ 264, 184, 31 },
		{ 264, 185, 31 },
		{ 264, 186, 31 },
		{ 264, 187, 31 },
		{ 264, 188, 31 },
		{ 264, 189, 31 },
		{ 264, 190, 31 },
		{ 264, 191, 31 },
		{ 265, 1, 31 },
		{ 265, 2, 31 },
		{ 265, 3, 31 },
		{ 265, 4, 31 },
		{ 265, 5, 31 },
		{ 265, 6, 31 },
		{ 265, 7, 31 },
		{ 265, 8, 31 },
		{ 265, 9, 31 },
		{ 265, 10, 31 },
		{ 265, 11, 31 },
		{ 265, 12, 31 },
		{ 265, 13, 31 },
		{ 265, 14, 31 },
		{ 265, 15, 31 },
		{ 265, 16, 31 },
		{ 265, 17, 31 },
		{ 265, 18, 31 },
		{ 265, 19, 31 },
		{ 265, 20, 31 },
		{ 265, 21, 31 },
		{ 265, 22, 31 },
		{ 265, 23, 31 },
		{ 265, 24, 31 },
		{ 265, 25, 31 },
		{ 265, 26, 31 },
		{ 265, 27, 31 },
		{ 265, 28, 31 },
		{ 265, 29, 31 },
		{ 265, 30, 31 },
		{ 265, 31, 31 },
		{ 265, 32, 31 },
		{ 265, 33, 31 },
		{ 265, 34, 31 },
		{ 265, 35, 31 },
		{ 265, 36, 31 },
		{ 265, 37, 31 },
		{ 265, 38, 31 },
		{ 265, 39, 31 },
		{ 265, 40, 31 },
		{ 265, 41, 31 },
		{ 265, 42, 31 },
		{ 265, 43, 31 },
		{ 265, 44, 31 },
		{ 265, 45, 31 },
		{ 265, 46, 31 },
		{ 265, 47, 31 },
		{ 265, 48, 31 },
		{ 265, 49, 31 },
		{ 265, 50, 31 },
		{ 265, 51, 31 },
		{ 265, 52, 31 },
		{ 265, 53, 31 },
		{ 265, 54, 31 },
		{ 265, 55, 31 },
		{ 265, 56, 31 },
		{ 265, 57, 31 },
		{ 265, 58, 31 },
		{ 265, 59, 31 },
		{ 265, 60, 31 },
		{ 265, 61, 31 },
		{ 265, 62, 31 },
		{ 265, 63, 31 },
		{ 265, 64, 31 },
		{ 265, 65, 31 },
		{ 265, 66, 31 },
		{ 265, 67, 31 },
		{ 265, 68, 31 },
		{ 265, 69, 31 },
		{ 265, 70, 31 },
		{ 265, 71, 31 },
		{ 265, 72, 31 },
		{ 265, 73, 31 },
		{ 265, 74, 31 },
		{ 265, 75, 31 },
		{ 265, 76, 31 },
		{ 265, 77, 31 },
		{ 265, 78, 31 },
		{ 265, 79, 31 },
		{ 265, 80, 31 },
		{ 265, 81, 31 },
		{ 265, 82, 31 },
		{ 265, 83, 31 },
		{ 265, 84, 31 },
		{ 265, 85, 31 },
		{ 265, 86, 31 },
		{ 265, 87, 31 },
		{ 265, 88, 31 },
		{ 265, 89, 31 },
		{ 265, 90, 31 },
		{ 265, 91, 31 },
		{ 265, 92, 31 },
		{ 265, 93, 31 },
		{ 265, 94, 31 },
		{ 265, 95, 31 },
		{ 265, 96, 31 },
		{ 265, 97, 31 },
		{ 265, 98, 31 },
		{ 265, 99, 31 },
		{ 265, 100, 31 },
		{ 265, 101, 31 },
		{ 265, 102, 31 },
		{ 265, 103, 31 },
		{ 265, 104, 31 },
		{ 265, 105, 31 },
		{ 265, 106, 31 },
		{ 265, 107, 31 },
		{ 265, 108, 31 },
		{ 265, 109, 31 },
		{ 265, 110, 31 },
		{ 265, 111, 31 },
		{ 265, 112, 31 },
		{ 265, 113, 31 },
		{ 265, 114, 31 },
		{ 265, 115, 31 },
		{ 265, 116, 31 },
		{ 265, 117, 31 },
		{ 265, 118, 31 },
		{ 265, 119, 31 },
		{ 265, 120, 31 },
		{ 265, 121, 31 },
		{ 265, 122, 31 },
		{ 265, 123, 31 },
		{ 265, 124, 31 },
		{ 265, 125, 31 },
		{ 265, 126, 31 },
		{ 265, 127, 31 },
		{ 265, 128, 31 },
		{ 265, 129, 31 },
		{ 265, 130, 31 },
		{ 265, 131, 31 },
		{ 265, 132, 31 },
		{ 265, 133, 31 },
		{ 265, 134, 31 },
		{ 265, 135, 31 },
		{ 265, 136, 31 },
		{ 265, 137, 31 },
		{ 265, 138, 31 },
		{ 265, 139, 31 },
		{ 265, 140, 31 },
		{ 265, 141, 31 },
		{ 265, 142, 31 },
		{ 265, 143, 31 },
		{ 265, 144, 31 },
		{ 265, 145, 31 },
		{ 265, 146, 31 },
		{ 265, 147, 31 },
		{ 265, 148, 31 },
		{ 265, 149, 31 },
		{ 265, 150, 31 },
		{ 265, 151, 31 },
		{ 265, 152, 31 },
		{ 265, 153, 31 },
		{ 265, 154, 31 },
		{ 265, 155, 31 },
		{ 265, 156, 31 },
		{ 265, 157, 31 },
		{ 265, 158, 31 },
		{ 265, 159, 31 },
		{ 265, 160, 31 },
		{ 265, 161, 31 },
		{ 265, 162, 31 },
		{ 265, 163, 31 },
		{ 265, 164, 31 },
		{ 265, 165, 31 },
		{ 265, 166, 31 },
		{ 265, 167, 31 },
		{ 265, 168, 31 },
		{ 265, 169, 31 },
		{ 265, 170, 31 },
		{ 265, 171, 31 },
		{ 265, 172, 31 },
		{ 265, 173, 31 },
		{ 265, 174, 31 },
		{ 265, 175, 31 },
		{ 265, 176, 31 },
		{ 265, 177, 31 },
		{ 265, 178, 31 },
		{ 265, 179, 31 },
		{ 265, 180, 31 },
		{ 265, 181, 31 },
		{ 265, 182, 31 },
		{ 265, 183, 31 },
		{ 265, 184, 31 },
		{ 265, 185, 31 },
		{ 265, 186, 31 },
		{ 265, 187, 31 },
		{ 265, 188, 31 },
		{ 265, 189, 31 },
		{ 265, 190, 31 },
		{ 265, 191, 31 },
		{ 266, 1, 31 },
		{ 266, 2, 31 },
		{ 266, 3, 31 },
		{ 266, 4, 31 },
		{ 266, 5, 31 },
		{ 266, 6, 31 },
		{ 266, 7, 31 },
		{ 266, 8, 31 },
		{ 266, 9, 31 },
		{ 266, 10, 31 },
		{ 266, 11, 31 },
		{ 266, 12, 31 },
		{ 266, 13, 31 },
		{ 266, 14, 31 },
		{ 266, 15, 31 },
		{ 266, 16, 31 },
		{ 266, 17, 31 },
		{ 266, 18, 31 },
		{ 266, 19, 31 },
		{ 266, 20, 31 },
		{ 266, 21, 31 },
		{ 266, 22, 31 },
		{ 266, 23, 31 },
		{ 266, 24, 31 },
		{ 266, 25, 31 },
		{ 266, 26, 31 },
		{ 266, 27, 31 },
		{ 266, 28, 31 },
		{ 266, 29, 31 },
		{ 266, 30, 31 },
		{ 266, 31, 31 },
		{ 266, 32, 31 },
		{ 266, 33, 31 },
		{ 266, 34, 31 },
		{ 266, 35, 31 },
		{ 266, 36, 31 },
		{ 266, 37, 31 },
		{ 266, 38, 31 },
		{ 266, 39, 31 },
		{ 266, 40, 31 },
		{ 266, 41, 31 },
		{ 266, 42, 31 },
		{ 266, 43, 31 },
		{ 266, 44, 31 },
		{ 266, 45, 31 },
		{ 266, 46, 31 },
		{ 266, 47, 31 },
		{ 266, 48, 31 },
		{ 266, 49, 31 },
		{ 266, 50, 31 },
		{ 266, 51, 31 },
		{ 266, 52, 31 },
		{ 266, 53, 31 },
		{ 266, 54, 31 },
		{ 266, 55, 31 },
		{ 266, 56, 31 },
		{ 266, 57, 31 },
		{ 266, 58, 31 },
		{ 266, 59, 31 },
		{ 266, 60, 31 },
		{ 266, 61, 31 },
		{ 266, 62, 31 },
		{ 266, 63, 31 },
		{ 266, 64, 31 },
		{ 266, 65, 31 },
		{ 266, 66, 31 },
		{ 266, 67, 31 },
		{ 266, 68, 31 },
		{ 266, 69, 31 },
		{ 266, 70, 31 },
		{ 266, 71, 31 },
		{ 266, 72, 31 },
		{ 266, 73, 31 },
		{ 266, 74, 31 },
		{ 266, 75, 31 },
		{ 266, 76, 31 },
		{ 266, 77, 31 },
		{ 266, 78, 31 },
		{ 266, 79, 31 },
		{ 266, 80, 31 },
		{ 266, 81, 31 },
		{ 266, 82, 31 },
		{ 266, 83, 31 },
		{ 266, 84, 31 },
		{ 266, 85, 31 },
		{ 266, 86, 31 },
		{ 266, 87, 31 },
		{ 266, 88, 31 },
		{ 266, 89, 31 },
		{ 266, 90, 31 },
		{ 266, 91, 31 },
		{ 266, 92, 31 },
		{ 266, 93, 31 },
		{ 266, 94, 31 },
		{ 266, 95, 31 },
		{ 266, 96, 31 },
		{ 266, 97, 31 },
		{ 266, 98, 31 },
		{ 266, 99, 31 },
		{ 266, 100, 31 },
		{ 266, 101, 31 },
		{ 266, 102, 31 },
		{ 266, 103, 31 },
		{ 266, 104, 31 },
		{ 266, 105, 31 },
		{ 266, 106, 31 },
		{ 266, 107, 31 },
		{ 266, 108, 31 },
		{ 266, 109, 31 },
		{ 266, 110, 31 },
		{ 266, 111, 31 },
		{ 266, 112, 31 },
		{ 266, 113, 31 },
		{ 266, 114, 31 },
		{ 266, 115, 31 },
		{ 266, 116, 31 },
		{ 266, 117, 31 },
		{ 266, 118, 31 },
		{ 266, 119, 31 },
		{ 266, 120, 31 },
		{ 266, 121, 31 },
		{ 266, 122, 31 },
		{ 266, 123, 31 },
		{ 266, 124, 31 },
		{ 266, 125, 31 },
		{ 266, 126, 31 },
		{ 266, 127, 31 },
		{ 266, 128, 31 },
		{ 266, 129, 31 },
		{ 266, 130, 31 },
		{ 266, 131, 31 },
		{ 266, 132, 31 },
		{ 266, 133, 31 },
		{ 266, 134, 31 },
		{ 266, 135, 31 },
		{ 266, 136, 31 },
		{ 266, 137, 31 },
		{ 266, 138, 31 },
		{ 266, 139, 31 },
		{ 266, 140, 31 },
		{ 266, 141, 31 },
		{ 266, 142, 31 },
		{ 266, 143, 31 },
		{ 266, 144, 31 },
		{ 266, 145, 31 },
		{ 266, 146, 31 },
		{ 266, 147, 31 },
		{ 266, 148, 31 },
		{ 266, 149, 31 },
		{ 266, 150, 31 },
		{ 266, 151, 31 },
		{ 266, 152, 31 },
		{ 266, 153, 31 },
		{ 266, 154, 31 },
		{ 266, 155, 31 },
		{ 266, 156, 31 },
		{ 266, 157, 31 },
		{ 266, 158, 31 },
		{ 266, 159, 31 },
		{ 266, 160, 31 },
		{ 266, 161, 31 },
		{ 266, 162, 31 },
		{ 266, 163, 31 },
		{ 266, 164, 31 },
		{ 266, 165, 31 },
		{ 266, 166, 31 },
		{ 266, 167, 31 },
		{ 266, 168, 31 },
		{ 266, 169, 31 },
		{ 266, 170, 31 },
		{ 266, 171, 31 },
		{ 266, 172, 31 },
		{ 266, 173, 31 },
		{ 266, 174, 31 },
		{ 266, 175, 31 },
		{ 266, 176, 31 },
		{ 266, 177, 31 },
		{ 266, 178, 31 },
		{ 266, 179, 31 },
		{ 266, 180, 31 },
		{ 266, 181, 31 },
		{ 266, 182, 31 },
		{ 266, 183, 31 },
		{ 266, 184, 31 },
		{ 266, 185, 31 },
		{ 266, 186, 31 },
		{ 266, 187, 31 },
		{ 266, 188, 31 },
		{ 266, 189, 31 },
		{ 266, 190, 31 },
		{ 266, 191, 31 },
		{ 267, 1, 31 },
		{ 267, 2, 31 },
		{ 267, 3, 31 },
		{ 267, 4, 31 },
		{ 267, 5, 31 },
		{ 267, 6, 31 },
		{ 267, 7, 31 },
		{ 267, 8, 31 },
		{ 267, 9, 31 },
		{ 267, 10, 31 },
		{ 267, 11, 31 },
		{ 267, 12, 31 },
		{ 267, 13, 31 },
		{ 267, 14, 31 },
		{ 267, 15, 31 },
		{ 267, 16, 31 },
		{ 267, 17, 31 },
		{ 267, 18, 31 },
		{ 267, 19, 31 },
		{ 267, 20, 31 },
		{ 267, 21, 31 },
		{ 267, 22, 31 },
		{ 267, 23, 31 },
		{ 267, 24, 31 },
		{ 267, 25, 31 },
		{ 267, 26, 31 },
		{ 267, 27, 31 },
		{ 267, 28, 31 },
		{ 267, 29, 31 },
		{ 267, 30, 31 },
		{ 267, 31, 31 },
		{ 267, 32, 31 },
		{ 267, 33, 31 },
		{ 267, 34, 31 },
		{ 267, 35, 31 },
		{ 267, 36, 31 },
		{ 267, 37, 31 },
		{ 267, 38, 31 },
		{ 267, 39, 31 },
		{ 267, 40, 31 },
		{ 267, 41, 31 },
		{ 267, 42, 31 },
		{ 267, 43, 31 },
		{ 267, 44, 31 },
		{ 267, 45, 31 },
		{ 267, 46, 31 },
		{ 267, 47, 31 },
		{ 267, 48, 31 },
		{ 267, 49, 31 },
		{ 267, 50, 31 },
		{ 267, 51, 31 },
		{ 267, 52, 31 },
		{ 267, 53, 31 },
		{ 267, 54, 31 },
		{ 267, 55, 31 },
		{ 267, 56, 31 },
		{ 267, 57, 31 },
		{ 267, 58, 31 },
		{ 267, 59, 31 },
		{ 267, 60, 31 },
		{ 267, 61, 31 },
		{ 267, 62, 31 },
		{ 267, 63, 31 },
		{ 267, 64, 31 },
		{ 267, 65, 31 },
		{ 267, 66, 31 },
		{ 267, 67, 31 },
		{ 267, 68, 31 },
		{ 267, 69, 31 },
		{ 267, 70, 31 },
		{ 267, 71, 31 },
		{ 267, 72, 31 },
		{ 267, 73, 31 },
		{ 267, 74, 31 },
		{ 267, 75, 31 },
		{ 267, 76, 31 },
		{ 267, 77, 31 },
		{ 267, 78, 31 },
		{ 267, 79, 31 },
		{ 267, 80, 31 },
		{ 267, 81, 31 },
		{ 267, 82, 31 },
		{ 267, 83, 31 },
		{ 267, 84, 31 },
		{ 267, 85, 31 },
		{ 267, 86, 31 },
		{ 267, 87, 31 },
		{ 267, 88, 31 },
		{ 267, 89, 31 },
		{ 267, 90, 31 },
		{ 267, 91, 31 },
		{ 267, 92, 31 },
		{ 267, 93, 31 },
		{ 267, 94, 31 },
		{ 267, 95, 31 },
		{ 267, 96, 31 },
		{ 267, 97, 31 },
		{ 267, 98, 31 },
		{ 267, 99, 31 },
		{ 267, 100, 31 },
		{ 267, 101, 31 },
		{ 267, 102, 31 },
		{ 267, 103, 31 },
		{ 267, 104, 31 },
		{ 267, 105, 31 },
		{ 267, 106, 31 },
		{ 267, 107, 31 },
		{ 267, 108, 31 },
		{ 267, 109, 31 },
		{ 267, 110, 31 },
		{ 267, 111, 31 },
		{ 267, 112, 31 },
		{ 267, 113, 31 },
		{ 267, 114, 31 },
		{ 267, 115, 31 },
		{ 267, 116, 31 },
		{ 267, 117, 31 },
		{ 267, 118, 31 },
		{ 267, 119, 31 },
		{ 267, 120, 31 },
		{ 267, 121, 31 },
		{ 267, 122, 31 },
		{ 267, 123, 31 },
		{ 267, 124, 31 },
		{ 267, 125, 31 },
		{ 267, 126, 31 },
		{ 267, 127, 31 },
		{ 267, 128, 31 },
		{ 267, 129, 31 },
		{ 267, 130, 31 },
		{ 267, 131, 31 },
		{ 267, 132, 31 },
		{ 267, 133, 31 },
		{ 267, 134, 31 },
		{ 267, 135, 31 },
		{ 267, 136, 31 },
		{ 267, 137, 31 },
		{ 267, 138, 31 },
		{ 267, 139, 31 },
		{ 267, 140, 31 },
		{ 267, 141, 31 },
		{ 267, 142, 31 },
		{ 267, 143, 31 },
		{ 267, 144, 31 },
		{ 267, 145, 31 },
		{ 267, 146, 31 },
		{ 267, 147, 31 },
		{ 267, 148, 31 },
		{ 267, 149, 31 },
		{ 267, 150, 31 },
		{ 267, 151, 31 },
		{ 267, 152, 31 },
		{ 267, 153, 31 },
		{ 267, 154, 31 },
		{ 267, 155, 31 },
		{ 267, 156, 31 },
		{ 267, 157, 31 },
		{ 267, 158, 31 },
		{ 267, 159, 31 },
		{ 267, 160, 31 },
		{ 267, 161, 31 },
		{ 267, 162, 31 },
		{ 267, 163, 31 },
		{ 267, 164, 31 },
		{ 267, 165, 31 },
		{ 267, 166, 31 },
		{ 267, 167, 31 },
		{ 267, 168, 31 },
		{ 267, 169, 31 },
		{ 267, 170, 31 },
		{ 267, 171, 31 },
		{ 267, 172, 31 },
		{ 267, 173, 31 },
		{ 267, 174, 31 },
		{ 267, 175, 31 },
		{ 267, 176, 31 },
		{ 267, 177, 31 },
		{ 267, 178, 31 },
		{ 267, 179, 31 },
		{ 267, 180, 31 },
		{ 267, 181, 31 },
		{ 267, 182, 31 },
		{ 267, 183, 31 },
		{ 267, 184, 31 },
		{ 267, 185, 31 },
		{ 267, 186, 31 },
		{ 267, 187, 31 },
		{ 267, 188, 31 },
		{ 267, 189, 31 },
		{ 267, 190, 31 },
		{ 267, 191, 31 },
		{ 268, 1, 31 },
		{ 268, 2, 31 },
		{ 268, 3, 31 },
		{ 268, 4, 31 },
		{ 268, 5, 31 },
		{ 268, 6, 31 },
		{ 268, 7, 31 },
		{ 268, 8, 31 },
		{ 268, 9, 31 },
		{ 268, 10, 31 },
		{ 268, 11, 31 },
		{ 268, 12, 31 },
		{ 268, 13, 31 },
		{ 268, 14, 31 },
		{ 268, 15, 31 },
		{ 268, 16, 31 },
		{ 268, 17, 31 },
		{ 268, 18, 31 },
		{ 268, 19, 31 },
		{ 268, 20, 31 },
		{ 268, 21, 31 },
		{ 268, 22, 31 },
		{ 268, 23, 31 },
		{ 268, 24, 31 },
		{ 268, 25, 31 },
		{ 268, 26, 31 },
		{ 268, 27, 31 },
		{ 268, 28, 31 },
		{ 268, 29, 31 },
		{ 268, 30, 31 },
		{ 268, 31, 31 },
		{ 268, 32, 31 },
		{ 268, 33, 31 },
		{ 268, 34, 31 },
		{ 268, 35, 31 },
		{ 268, 36, 31 },
		{ 268, 37, 31 },
		{ 268, 38, 31 },
		{ 268, 39, 31 },
		{ 268, 40, 31 },
		{ 268, 41, 31 },
		{ 268, 42, 31 },
		{ 268, 43, 31 },
		{ 268, 44, 31 },
		{ 268, 45, 31 },
		{ 268, 46, 31 },
		{ 268, 47, 31 },
		{ 268, 48, 31 },
		{ 268, 49, 31 },
		{ 268, 50, 31 },
		{ 268, 51, 31 },
		{ 268, 52, 31 },
		{ 268, 53, 31 },
		{ 268, 54, 31 },
		{ 268, 55, 31 },
		{ 268, 56, 31 },
		{ 268, 57, 31 },
		{ 268, 58, 31 },
		{ 268, 59, 31 },
		{ 268, 60, 31 },
		{ 268, 61, 31 },
		{ 268, 62, 31 },
		{ 268, 63, 31 },
		{ 268, 64, 31 },
		{ 268, 65, 31 },
		{ 268, 66, 31 },
		{ 268, 67, 31 },
		{ 268, 68, 31 },
		{ 268, 69, 31 },
		{ 268, 70, 31 },
		{ 268, 71, 31 },
		{ 268, 72, 31 },
		{ 268, 73, 31 },
		{ 268, 74, 31 },
		{ 268, 75, 31 },
		{ 268, 76, 31 },
		{ 268, 77, 31 },
		{ 268, 78, 31 },
		{ 268, 79, 31 },
		{ 268, 80, 31 },
		{ 268, 81, 31 },
		{ 268, 82, 31 },
		{ 268, 83, 31 },
		{ 268, 84, 31 },
		{ 268, 85, 31 },
		{ 268, 86, 31 },
		{ 268, 87, 31 },
		{ 268, 88, 31 },
		{ 268, 89, 31 },
		{ 268, 90, 31 },
		{ 268, 91, 31 },
		{ 268, 92, 31 },
		{ 268, 93, 31 },
		{ 268, 94, 31 },
		{ 268, 95, 31 },
		{ 268, 96, 31 },
		{ 268, 97, 31 },
		{ 268, 98, 31 },
		{ 268, 99, 31 },
		{ 268, 100, 31 },
		{ 268, 101, 31 },
		{ 268, 102, 31 },
		{ 268, 103, 31 },
		{ 268, 104, 31 },
		{ 268, 105, 31 },
		{ 268, 106, 31 },
		{ 268, 107, 31 },
		{ 268, 108, 31 },
		{ 268, 109, 31 },
		{ 268, 110, 31 },
		{ 268, 111, 31 },
		{ 268, 112, 31 },
		{ 268, 113, 31 },
		{ 268, 114, 31 },
		{ 268, 115, 31 },
		{ 268, 116, 31 },
		{ 268, 117, 31 },
		{ 268, 118, 31 },
		{ 268, 119, 31 },
		{ 268, 120, 31 },
		{ 268, 121, 31 },
		{ 268, 122, 31 },
		{ 268, 123, 31 },
		{ 268, 124, 31 },
		{ 268, 125, 31 },
		{ 268, 126, 31 },
		{ 268, 127, 31 },
		{ 268, 128, 31 },
		{ 268, 129, 31 },
		{ 268, 130, 31 },
		{ 268, 131, 31 },
		{ 268, 132, 31 },
		{ 268, 133, 31 },
		{ 268, 134, 31 },
		{ 268, 135, 31 },
		{ 268, 136, 31 },
		{ 268, 137, 31 },
		{ 268, 138, 31 },
		{ 268, 139, 31 },
		{ 268, 140, 31 },
		{ 268, 141, 31 },
		{ 268, 142, 31 },
		{ 268, 143, 31 },
		{ 268, 144, 31 },
		{ 268, 145, 31 },
		{ 268, 146, 31 },
		{ 268, 147, 31 },
		{ 268, 148, 31 },
		{ 268, 149, 31 },
		{ 268, 150, 31 },
		{ 268, 151, 31 },
		{ 268, 152, 31 },
		{ 268, 153, 31 },
		{ 268, 154, 31 },
		{ 268, 155, 31 },
		{ 268, 156, 31 },
		{ 268, 157, 31 },
		{ 268, 158, 31 },
		{ 268, 159, 31 },
		{ 268, 160, 31 },
		{ 268, 161, 31 },
		{ 268, 162, 31 },
		{ 268, 163, 31 },
		{ 268, 164, 31 },
		{ 268, 165, 31 },
		{ 268, 166, 31 },
		{ 268, 167, 31 },
		{ 268, 168, 31 },
		{ 268, 169, 31 },
		{ 268, 170, 31 },
		{ 268, 171, 31 },
		{ 268, 172, 31 },
		{ 268, 173, 31 },
		{ 268, 174, 31 },
		{ 268, 175, 31 },
		{ 268, 176, 31 },
		{ 268, 177, 31 },
		{ 268, 178, 31 },
		{ 268, 179, 31 },
		{ 268, 180, 31 },
		{ 268, 181, 31 },
		{ 268, 182, 31 },
		{ 268, 183, 31 },
		{ 268, 184, 31 },
		{ 268, 185, 31 },
		{ 268, 186, 31 },
		{ 268, 187, 31 },
		{ 268, 188, 31 },
		{ 268, 189, 31 },
		{ 268, 190, 31 },
		{ 268, 191, 31 },
		{ 269, 1, 31 },
		{ 269, 2, 31 },
		{ 269, 3, 31 },
		{ 269, 4, 31 },
		{ 269, 5, 31 },
		{ 269, 6, 31 },
		{ 269, 7, 31 },
		{ 269, 8, 31 },
		{ 269, 9, 31 },
		{ 269, 10, 31 },
		{ 269, 11, 31 },
		{ 269, 12, 31 },
		{ 269, 13, 31 },
		{ 269, 14, 31 },
		{ 269, 15, 31 },
		{ 269, 16, 31 },
		{ 269, 17, 31 },
		{ 269, 18, 31 },
		{ 269, 19, 31 },
		{ 269, 20, 31 },
		{ 269, 21, 31 },
		{ 269, 22, 31 },
		{ 269, 23, 31 },
		{ 269, 24, 31 },
		{ 269, 25, 31 },
		{ 269, 26, 31 },
		{ 269, 27, 31 },
		{ 269, 28, 31 },
		{ 269, 29, 31 },
		{ 269, 30, 31 },
		{ 269, 31, 31 },
		{ 269, 32, 31 },
		{ 269, 33, 31 },
		{ 269, 34, 31 },
		{ 269, 35, 31 },
		{ 269, 36, 31 },
		{ 269, 37, 31 },
		{ 269, 38, 31 },
		{ 269, 39, 31 },
		{ 269, 40, 31 },
		{ 269, 41, 31 },
		{ 269, 42, 31 },
		{ 269, 43, 31 },
		{ 269, 44, 31 },
		{ 269, 45, 31 },
		{ 269, 46, 31 },
		{ 269, 47, 31 },
		{ 269, 48, 31 },
		{ 269, 49, 31 },
		{ 269, 50, 31 },
		{ 269, 51, 31 },
		{ 269, 52, 31 },
		{ 269, 53, 31 },
		{ 269, 54, 31 },
		{ 269, 55, 31 },
		{ 269, 56, 31 },
		{ 269, 57, 31 },
		{ 269, 58, 31 },
		{ 269, 59, 31 },
		{ 269, 60, 31 },
		{ 269, 61, 31 },
		{ 269, 62, 31 },
		{ 269, 63, 31 },
		{ 269, 64, 31 },
		{ 269, 65, 31 },
		{ 269, 66, 31 },
		{ 269, 67, 31 },
		{ 269, 68, 31 },
		{ 269, 69, 31 },
		{ 269, 70, 31 },
		{ 269, 71, 31 },
		{ 269, 72, 31 },
		{ 269, 73, 31 },
		{ 269, 74, 31 },
		{ 269, 75, 31 },
		{ 269, 76, 31 },
		{ 269, 77, 31 },
		{ 269, 78, 31 },
		{ 269, 79, 31 },
		{ 269, 80, 31 },
		{ 269, 81, 31 },
		{ 269, 82, 31 },
		{ 269, 83, 31 },
		{ 269, 84, 31 },
		{ 269, 85, 31 },
		{ 269, 86, 31 },
		{ 269, 87, 31 },
		{ 269, 88, 31 },
		{ 269, 89, 31 },
		{ 269, 90, 31 },
		{ 269, 91, 31 },
		{ 269, 92, 31 },
		{ 269, 93, 31 },
		{ 269, 94, 31 },
		{ 269, 95, 31 },
		{ 269, 96, 31 },
		{ 269, 97, 31 },
		{ 269, 98, 31 },
		{ 269, 99, 31 },
		{ 269, 100, 31 },
		{ 269, 101, 31 },
		{ 269, 102, 31 },
		{ 269, 103, 31 },
		{ 269, 104, 31 },
		{ 269, 105, 31 },
		{ 269, 106, 31 },
		{ 269, 107, 31 },
		{ 269, 108, 31 },
		{ 269, 109, 31 },
		{ 269, 110, 31 },
		{ 269, 111, 31 },
		{ 269, 112, 31 },
		{ 269, 113, 31 },
		{ 269, 114, 31 },
		{ 269, 115, 31 },
		{ 269, 116, 31 },
		{ 269, 117, 31 },
		{ 269, 118, 31 },
		{ 269, 119, 31 },
		{ 269, 120, 31 },
		{ 269, 121, 31 },
		{ 269, 122, 31 },
		{ 269, 123, 31 },
		{ 269, 124, 31 },
		{ 269, 125, 31 },
		{ 269, 126, 31 },
		{ 269, 127, 31 },
		{ 269, 128, 31 },
		{ 269, 129, 31 },
		{ 269, 130, 31 },
		{ 269, 131, 31 },
		{ 269, 132, 31 },
		{ 269, 133, 31 },
		{ 269, 134, 31 },
		{ 269, 135, 31 },
		{ 269, 136, 31 },
		{ 269, 137, 31 },
		{ 269, 138, 31 },
		{ 269, 139, 31 },
		{ 269, 140, 31 },
		{ 269, 141, 31 },
		{ 269, 142, 31 },
		{ 269, 143, 31 },
		{ 269, 144, 31 },
		{ 269, 145, 31 },
		{ 269, 146, 31 },
		{ 269, 147, 31 },
		{ 269, 148, 31 },
		{ 269, 149, 31 },
		{ 269, 150, 31 },
		{ 269, 151, 31 },
		{ 269, 152, 31 },
		{ 269, 153, 31 },
		{ 269, 154, 31 },
		{ 269, 155, 31 },
		{ 269, 156, 31 },
		{ 269, 157, 31 },
		{ 269, 158, 31 },
		{ 269, 159, 31 },
		{ 269, 160, 31 },
		{ 269, 161, 31 },
		{ 269, 162, 31 },
		{ 269, 163, 31 },
		{ 269, 164, 31 },
		{ 269, 165, 31 },
		{ 269, 166, 31 },
		{ 269, 167, 31 },
		{ 269, 168, 31 },
		{ 269, 169, 31 },
		{ 269, 170, 31 },
		{ 269, 171, 31 },
		{ 269, 172, 31 },
		{ 269, 173, 31 },
		{ 269, 174, 31 },
		{ 269, 175, 31 },
		{ 269, 176, 31 },
		{ 269, 177, 31 },
		{ 269, 178, 31 },
		{ 269, 179, 31 },
		{ 269, 180, 31 },
		{ 269, 181, 31 },
		{ 269, 182, 31 },
		{ 269, 183, 31 },
		{ 269, 184, 31 },
		{ 269, 185, 31 },
		{ 269, 186, 31 },
		{ 269, 187, 31 },
		{ 269, 188, 31 },
		{ 269, 189, 31 },
		{ 269, 190, 31 },
		{ 269, 191, 31 },
		{ 270, 1, 31 },
		{ 270, 2, 31 },
		{ 270, 3, 31 },
		{ 270, 4, 31 },
		{ 270, 5, 31 },
		{ 270, 6, 31 },
		{ 270, 7, 31 },
		{ 270, 8, 31 },
		{ 270, 9, 31 },
		{ 270, 10, 31 },
		{ 270, 11, 31 },
		{ 270, 12, 31 },
		{ 270, 13, 31 },
		{ 270, 14, 31 },
		{ 270, 15, 31 },
		{ 270, 16, 31 },
		{ 270, 17, 31 },
		{ 270, 18, 31 },
		{ 270, 19, 31 },
		{ 270, 20, 31 },
		{ 270, 21, 31 },
		{ 270, 22, 31 },
		{ 270, 23, 31 },
		{ 270, 24, 31 },
		{ 270, 25, 31 },
		{ 270, 26, 31 },
		{ 270, 27, 31 },
		{ 270, 28, 31 },
		{ 270, 29, 31 },
		{ 270, 30, 31 },
		{ 270, 31, 31 },
		{ 270, 32, 31 },
		{ 270, 33, 31 },
		{ 270, 34, 31 },
		{ 270, 35, 31 },
		{ 270, 36, 31 },
		{ 270, 37, 31 },
		{ 270, 38, 31 },
		{ 270, 39, 31 },
		{ 270, 40, 31 },
		{ 270, 41, 31 },
		{ 270, 42, 31 },
		{ 270, 43, 31 },
		{ 270, 44, 31 },
		{ 270, 45, 31 },
		{ 270, 46, 31 },
		{ 270, 47, 31 },
		{ 270, 48, 31 },
		{ 270, 49, 31 },
		{ 270, 50, 31 },
		{ 270, 51, 31 },
		{ 270, 52, 31 },
		{ 270, 53, 31 },
		{ 270, 54, 31 },
		{ 270, 55, 31 },
		{ 270, 56, 31 },
		{ 270, 57, 31 },
		{ 270, 58, 31 },
		{ 270, 59, 31 },
		{ 270, 60, 31 },
		{ 270, 61, 31 },
		{ 270, 62, 31 },
		{ 270, 63, 31 },
		{ 270, 64, 31 },
		{ 270, 65, 31 },
		{ 270, 66, 31 },
		{ 270, 67, 31 },
		{ 270, 68, 31 },
		{ 270, 69, 31 },
		{ 270, 70, 31 },
		{ 270, 71, 31 },
		{ 270, 72, 31 },
		{ 270, 73, 31 },
		{ 270, 74, 31 },
		{ 270, 75, 31 },
		{ 270, 76, 31 },
		{ 270, 77, 31 },
		{ 270, 78, 31 },
		{ 270, 79, 31 },
		{ 270, 80, 31 },
		{ 270, 81, 31 },
		{ 270, 82, 31 },
		{ 270, 83, 31 },
		{ 270, 84, 31 },
		{ 270, 85, 31 },
		{ 270, 86, 31 },
		{ 270, 87, 31 },
		{ 270, 88, 31 },
		{ 270, 89, 31 },
		{ 270, 90, 31 },
		{ 270, 91, 31 },
		{ 270, 92, 31 },
		{ 270, 93, 31 },
		{ 270, 94, 31 },
		{ 270, 95, 31 },
		{ 270, 96, 31 },
		{ 270, 97, 31 },
		{ 270, 98, 31 },
		{ 270, 99, 31 },
		{ 270, 100, 31 },
		{ 270, 101, 31 },
		{ 270, 102, 31 },
		{ 270, 103, 31 },
		{ 270, 104, 31 },
		{ 270, 105, 31 },
		{ 270, 106, 31 },
		{ 270, 107, 31 },
		{ 270, 108, 31 },
		{ 270, 109, 31 },
		{ 270, 110, 31 },
		{ 270, 111, 31 },
		{ 270, 112, 31 },
		{ 270, 113, 31 },
		{ 270, 114, 31 },
		{ 270, 115, 31 },
		{ 270, 116, 31 },
		{ 270, 117, 31 },
		{ 270, 118, 31 },
		{ 270, 119, 31 },
		{ 270, 120, 31 },
		{ 270, 121, 31 },
		{ 270, 122, 31 },
		{ 270, 123, 31 },
		{ 270, 124, 31 },
		{ 270, 125, 31 },
		{ 270, 126, 31 },
		{ 270, 127, 31 },
		{ 270, 128, 31 },
		{ 270, 129, 31 },
		{ 270, 130, 31 },
		{ 270, 131, 31 },
		{ 270, 132, 31 },
		{ 270, 133, 31 },
		{ 270, 134, 31 },
		{ 270, 135, 31 },
		{ 270, 136, 31 },
		{ 270, 137, 31 },
		{ 270, 138, 31 },
		{ 270, 139, 31 },
		{ 270, 140, 31 },
		{ 270, 141, 31 },
		{ 270, 142, 31 },
		{ 270, 143, 31 },
		{ 270, 144, 31 },
		{ 270, 145, 31 },
		{ 270, 146, 31 },
		{ 270, 147, 31 },
		{ 270, 148, 31 },
		{ 270, 149, 31 },
		{ 270, 150, 31 },
		{ 270, 151, 31 },
		{ 270, 152, 31 },
		{ 270, 153, 31 },
		{ 270, 154, 31 },
		{ 270, 155, 31 },
		{ 270, 156, 31 },
		{ 270, 157, 31 },
		{ 270, 158, 31 },
		{ 270, 159, 31 },
		{ 270, 160, 31 },
		{ 270, 161, 31 },
		{ 270, 162, 31 },
		{ 270, 163, 31 },
		{ 270, 164, 31 },
		{ 270, 165, 31 },
		{ 270, 166, 31 },
		{ 270, 167, 31 },
		{ 270, 168, 31 },
		{ 270, 169, 31 },
		{ 270, 170, 31 },
		{ 270, 171, 31 },
		{ 270, 172, 31 },
		{ 270, 173, 31 },
		{ 270, 174, 31 },
		{ 270, 175, 31 },
		{ 270, 176, 31 },
		{ 270, 177, 31 },
		{ 270, 178, 31 },
		{ 270, 179, 31 },
		{ 270, 180, 31 },
		{ 270, 181, 31 },
		{ 270, 182, 31 },
		{ 270, 183, 31 },
		{ 270, 184, 31 },
		{ 270, 185, 31 },
		{ 270, 186, 31 },
		{ 270, 187, 31 },
		{ 270, 188, 31 },
		{ 270, 189, 31 },
		{ 270, 190, 31 },
		{ 270, 191, 31 },
		{ 271, 1, 31 },
		{ 271, 2, 31 },
		{ 271, 3, 31 },
		{ 271, 4, 31 },
		{ 271, 5, 31 },
		{ 271, 6, 31 },
		{ 271, 7, 31 },
		{ 271, 8, 31 },
		{ 271, 9, 31 },
		{ 271, 10, 31 },
		{ 271, 11, 31 },
		{ 271, 12, 31 },
		{ 271, 13, 31 },
		{ 271, 14, 31 },
		{ 271, 15, 31 },
		{ 271, 16, 31 },
		{ 271, 17, 31 },
		{ 271, 18, 31 },
		{ 271, 19, 31 },
		{ 271, 20, 31 },
		{ 271, 21, 31 },
		{ 271, 22, 31 },
		{ 271, 23, 31 },
		{ 271, 24, 31 },
		{ 271, 25, 31 },
		{ 271, 26, 31 },
		{ 271, 27, 31 },
		{ 271, 28, 31 },
		{ 271, 29, 31 },
		{ 271, 30, 31 },
		{ 271, 31, 31 },
		{ 271, 32, 31 },
		{ 271, 33, 31 },
		{ 271, 34, 31 },
		{ 271, 35, 31 },
		{ 271, 36, 31 },
		{ 271, 37, 31 },
		{ 271, 38, 31 },
		{ 271, 39, 31 },
		{ 271, 40, 31 },
		{ 271, 41, 31 },
		{ 271, 42, 31 },
		{ 271, 43, 31 },
		{ 271, 44, 31 },
		{ 271, 45, 31 },
		{ 271, 46, 31 },
		{ 271, 47, 31 },
		{ 271, 48, 31 },
		{ 271, 49, 31 },
		{ 271, 50, 31 },
		{ 271, 51, 31 },
		{ 271, 52, 31 },
		{ 271, 53, 31 },
		{ 271, 54, 31 },
		{ 271, 55, 31 },
		{ 271, 56, 31 },
		{ 271, 57, 31 },
		{ 271, 58, 31 },
		{ 271, 59, 31 },
		{ 271, 60, 31 },
		{ 271, 61, 31 },
		{ 271, 62, 31 },
		{ 271, 63, 31 },
		{ 271, 64, 31 },
		{ 271, 65, 31 },
		{ 271, 66, 31 },
		{ 271, 67, 31 },
		{ 271, 68, 31 },
		{ 271, 69, 31 },
		{ 271, 70, 31 },
		{ 271, 71, 31 },
		{ 271, 72, 31 },
		{ 271, 73, 31 },
		{ 271, 74, 31 },
		{ 271, 75, 31 },
		{ 271, 76, 31 },
		{ 271, 77, 31 },
		{ 271, 78, 31 },
		{ 271, 79, 31 },
		{ 271, 80, 31 },
		{ 271, 81, 31 },
		{ 271, 82, 31 },
		{ 271, 83, 31 },
		{ 271, 84, 31 },
		{ 271, 85, 31 },
		{ 271, 86, 31 },
		{ 271, 87, 31 },
		{ 271, 88, 31 },
		{ 271, 89, 31 },
		{ 271, 90, 31 },
		{ 271, 91, 31 },
		{ 271, 92, 31 },
		{ 271, 93, 31 },
		{ 271, 94, 31 },
		{ 271, 95, 31 },
		{ 271, 96, 31 },
		{ 271, 97, 31 },
		{ 271, 98, 31 },
		{ 271, 99, 31 },
		{ 271, 100, 31 },
		{ 271, 101, 31 },
		{ 271, 102, 31 },
		{ 271, 103, 31 },
		{ 271, 104, 31 },
		{ 271, 105, 31 },
		{ 271, 106, 31 },
		{ 271, 107, 31 },
		{ 271, 108, 31 },
		{ 271, 109, 31 },
		{ 271, 110, 31 },
		{ 271, 111, 31 },
		{ 271, 112, 31 },
		{ 271, 113, 31 },
		{ 271, 114, 31 },
		{ 271, 115, 31 },
		{ 271, 116, 31 },
		{ 271, 117, 31 },
		{ 271, 118, 31 },
		{ 271, 119, 31 },
		{ 271, 120, 31 },
		{ 271, 121, 31 },
		{ 271, 122, 31 },
		{ 271, 123, 31 },
		{ 271, 124, 31 },
		{ 271, 125, 31 },
		{ 271, 126, 31 },
		{ 271, 127, 31 },
		{ 271, 128, 31 },
		{ 271, 129, 31 },
		{ 271, 130, 31 },
		{ 271, 131, 31 },
		{ 271, 132, 31 },
		{ 271, 133, 31 },
		{ 271, 134, 31 },
		{ 271, 135, 31 },
		{ 271, 136, 31 },
		{ 271, 137, 31 },
		{ 271, 138, 31 },
		{ 271, 139, 31 },
		{ 271, 140, 31 },
		{ 271, 141, 31 },
		{ 271, 142, 31 },
		{ 271, 143, 31 },
		{ 271, 144, 31 },
		{ 271, 145, 31 },
		{ 271, 146, 31 },
		{ 271, 147, 31 },
		{ 271, 148, 31 },
		{ 271, 149, 31 },
		{ 271, 150, 31 },
		{ 271, 151, 31 },
		{ 271, 152, 31 },
		{ 271, 153, 31 },
		{ 271, 154, 31 },
		{ 271, 155, 31 },
		{ 271, 156, 31 },
		{ 271, 157, 31 },
		{ 271, 158, 31 },
		{ 271, 159, 31 },
		{ 271, 160, 31 },
		{ 271, 161, 31 },
		{ 271, 162, 31 },
		{ 271, 163, 31 },
		{ 271, 164, 31 },
		{ 271, 165, 31 },
		{ 271, 166, 31 },
		{ 271, 167, 31 },
		{ 271, 168, 31 },
		{ 271, 169, 31 },
		{ 271, 170, 31 },
		{ 271, 171, 31 },
		{ 271, 172, 31 },
		{ 271, 173, 31 },
		{ 271, 174, 31 },
		{ 271, 175, 31 },
		{ 271, 176, 31 },
		{ 271, 177, 31 },
		{ 271, 178, 31 },
		{ 271, 179, 31 },
		{ 271, 180, 31 },
		{ 271, 181, 31 },
		{ 271, 182, 31 },
		{ 271, 183, 31 },
		{ 271, 184, 31 },
		{ 271, 185, 31 },
		{ 271, 186, 31 },
		{ 271, 187, 31 },
		{ 271, 188, 31 },
		{ 271, 189, 31 },
		{ 271, 190, 31 },
		{ 271, 191, 31 },
		{ 272, 1, 31 },
		{ 272, 2, 31 },
		{ 272, 3, 31 },
		{ 272, 4, 31 },
		{ 272, 5, 31 },
		{ 272, 6, 31 },
		{ 272, 7, 31 },
		{ 272, 8, 31 },
		{ 272, 9, 31 },
		{ 272, 10, 31 },
		{ 272, 11, 31 },
		{ 272, 12, 31 },
		{ 272, 13, 31 },
		{ 272, 14, 31 },
		{ 272, 15, 31 },
		{ 272, 16, 31 },
		{ 272, 17, 31 },
		{ 272, 18, 31 },
		{ 272, 19, 31 },
		{ 272, 20, 31 },
		{ 272, 21, 31 },
		{ 272, 22, 31 },
		{ 272, 23, 31 },
		{ 272, 24, 31 },
		{ 272, 25, 31 },
		{ 272, 26, 31 },
		{ 272, 27, 31 },
		{ 272, 28, 31 },
		{ 272, 29, 31 },
		{ 272, 30, 31 },
		{ 272, 31, 31 },
		{ 272, 32, 31 },
		{ 272, 33, 31 },
		{ 272, 34, 31 },
		{ 272, 35, 31 },
		{ 272, 36, 31 },
		{ 272, 37, 31 },
		{ 272, 38, 31 },
		{ 272, 39, 31 },
		{ 272, 40, 31 },
		{ 272, 41, 31 },
		{ 272, 42, 31 },
		{ 272, 43, 31 },
		{ 272, 44, 31 },
		{ 272, 45, 31 },
		{ 272, 46, 31 },
		{ 272, 47, 31 },
		{ 272, 48, 31 },
		{ 272, 49, 31 },
		{ 272, 50, 31 },
		{ 272, 51, 31 },
		{ 272, 52, 31 },
		{ 272, 53, 31 },
		{ 272, 54, 31 },
		{ 272, 55, 31 },
		{ 272, 56, 31 },
		{ 272, 57, 31 },
		{ 272, 58, 31 },
		{ 272, 59, 31 },
		{ 272, 60, 31 },
		{ 272, 61, 31 },
		{ 272, 62, 31 },
		{ 272, 63, 31 },
		{ 272, 64, 31 },
		{ 272, 65, 31 },
		{ 272, 66, 31 },
		{ 272, 67, 31 },
		{ 272, 68, 31 },
		{ 272, 69, 31 },
		{ 272, 70, 31 },
		{ 272, 71, 31 },
		{ 272, 72, 31 },
		{ 272, 73, 31 },
		{ 272, 74, 31 },
		{ 272, 75, 31 },
		{ 272, 76, 31 },
		{ 272, 77, 31 },
		{ 272, 78, 31 },
		{ 272, 79, 31 },
		{ 272, 80, 31 },
		{ 272, 81, 31 },
		{ 272, 82, 31 },
		{ 272, 83, 31 },
		{ 272, 84, 31 },
		{ 272, 85, 31 },
		{ 272, 86, 31 },
		{ 272, 87, 31 },
		{ 272, 88, 31 },
		{ 272, 89, 31 },
		{ 272, 90, 31 },
		{ 272, 91, 31 },
		{ 272, 92, 31 },
		{ 272, 93, 31 },
		{ 272, 94, 31 },
		{ 272, 95, 31 },
		{ 272, 96, 31 },
		{ 272, 97, 31 },
		{ 272, 98, 31 },
		{ 272, 99, 31 },
		{ 272, 100, 31 },
		{ 272, 101, 31 },
		{ 272, 102, 31 },
		{ 272, 103, 31 },
		{ 272, 104, 31 },
		{ 272, 105, 31 },
		{ 272, 106, 31 },
		{ 272, 107, 31 },
		{ 272, 108, 31 },
		{ 272, 109, 31 },
		{ 272, 110, 31 },
		{ 272, 111, 31 },
		{ 272, 112, 31 },
		{ 272, 113, 31 },
		{ 272, 114, 31 },
		{ 272, 115, 31 },
		{ 272, 116, 31 },
		{ 272, 117, 31 },
		{ 272, 118, 31 },
		{ 272, 119, 31 },
		{ 272, 120, 31 },
		{ 272, 121, 31 },
		{ 272, 122, 31 },
		{ 272, 123, 31 },
		{ 272, 124, 31 },
		{ 272, 125, 31 },
		{ 272, 126, 31 },
		{ 272, 127, 31 },
		{ 272, 128, 31 },
		{ 272, 129, 31 },
		{ 272, 130, 31 },
		{ 272, 131, 31 },
		{ 272, 132, 31 },
		{ 272, 133, 31 },
		{ 272, 134, 31 },
		{ 272, 135, 31 },
		{ 272, 136, 31 },
		{ 272, 137, 31 },
		{ 272, 138, 31 },
		{ 272, 139, 31 },
		{ 272, 140, 31 },
		{ 272, 141, 31 },
		{ 272, 142, 31 },
		{ 272, 143, 31 },
		{ 272, 144, 31 },
		{ 272, 145, 31 },
		{ 272, 146, 31 },
		{ 272, 147, 31 },
		{ 272, 148, 31 },
		{ 272, 149, 31 },
		{ 272, 150, 31 },
		{ 272, 151, 31 },
		{ 272, 152, 31 },
		{ 272, 153, 31 },
		{ 272, 154, 31 },
		{ 272, 155, 31 },
		{ 272, 156, 31 },
		{ 272, 157, 31 },
		{ 272, 158, 31 },
		{ 272, 159, 31 },
		{ 272, 160, 31 },
		{ 272, 161, 31 },
		{ 272, 162, 31 },
		{ 272, 163, 31 },
		{ 272, 164, 31 },
		{ 272, 165, 31 },
		{ 272, 166, 31 },
		{ 272, 167, 31 },
		{ 272, 168, 31 },
		{ 272, 169, 31 },
		{ 272, 170, 31 },
		{ 272, 171, 31 },
		{ 272, 172, 31 },
		{ 272, 173, 31 },
		{ 272, 174, 31 },
		{ 272, 175, 31 },
		{ 272, 176, 31 },
		{ 272, 177, 31 },
		{ 272, 178, 31 },
		{ 272, 179, 31 },
		{ 272, 180, 31 },
		{ 272, 181, 31 },
		{ 272, 182, 31 },
		{ 272, 183, 31 },
		{ 272, 184, 31 },
		{ 272, 185, 31 },
		{ 272, 186, 31 },
		{ 272, 187, 31 },
		{ 272, 188, 31 },
		{ 272, 189, 31 },
		{ 272, 190, 31 },
		{ 273, 1, 31 },
		{ 273, 2, 31 },
		{ 273, 3, 31 },
		{ 273, 4, 31 },
		{ 273, 5, 31 },
		{ 273, 6, 31 },
		{ 273, 7, 31 },
		{ 273, 8, 31 },
		{ 273, 9, 31 },
		{ 273, 10, 31 },
		{ 273, 11, 31 },
		{ 273, 12, 31 },
		{ 273, 13, 31 },
		{ 273, 14, 31 },
		{ 273, 15, 31 },
		{ 273, 16, 31 },
		{ 273, 17, 31 },
		{ 273, 18, 31 },
		{ 273, 19, 31 },
		{ 273, 20, 31 },
		{ 273, 21, 31 },
		{ 273, 22, 31 },
		{ 273, 23, 31 },
		{ 273, 24, 31 },
		{ 273, 25, 31 },
		{ 273, 26, 31 },
		{ 273, 27, 31 },
		{ 273, 28, 31 },
		{ 273, 29, 31 },
		{ 273, 30, 31 },
		{ 273, 31, 31 },
		{ 273, 32, 31 },
		{ 273, 33, 31 },
		{ 273, 34, 31 },
		{ 273, 35, 31 },
		{ 273, 36, 31 },
		{ 273, 37, 31 },
		{ 273, 38, 31 },
		{ 273, 39, 31 },
		{ 273, 40, 31 },
		{ 273, 41, 31 },
		{ 273, 42, 31 },
		{ 273, 43, 31 },
		{ 273, 44, 31 },
		{ 273, 45, 31 },
		{ 273, 46, 31 },
		{ 273, 47, 31 },
		{ 273, 48, 31 },
		{ 273, 49, 31 },
		{ 273, 50, 31 },
		{ 273, 51, 31 },
		{ 273, 52, 31 },
		{ 273, 53, 31 },
		{ 273, 54, 31 },
		{ 273, 55, 31 },
		{ 273, 56, 31 },
		{ 273, 57, 31 },
		{ 273, 58, 31 },
		{ 273, 59, 31 },
		{ 273, 60, 31 },
		{ 273, 61, 31 },
		{ 273, 62, 31 },
		{ 273, 63, 31 },
		{ 273, 64, 31 },
		{ 273, 65, 31 },
		{ 273, 66, 31 },
		{ 273, 67, 31 },
		{ 273, 68, 31 },
		{ 273, 69, 31 },
		{ 273, 70, 31 },
		{ 273, 71, 31 },
		{ 273, 72, 31 },
		{ 273, 73, 31 },
		{ 273, 74, 31 },
		{ 273, 75, 31 },
		{ 273, 76, 31 },
		{ 273, 77, 31 },
		{ 273, 78, 31 },
		{ 273, 79, 31 },
		{ 273, 80, 31 },
		{ 273, 81, 31 },
		{ 273, 82, 31 },
		{ 273, 83, 31 },
		{ 273, 84, 31 },
		{ 273, 85, 31 },
		{ 273, 86, 31 },
		{ 273, 87, 31 },
		{ 273, 88, 31 },
		{ 273, 89, 31 },
		{ 273, 90, 31 },
		{ 273, 91, 31 },
		{ 273, 92, 31 },
		{ 273, 93, 31 },
		{ 273, 94, 31 },
		{ 273, 95, 31 },
		{ 273, 96, 31 },
		{ 273, 97, 31 },
		{ 273, 98, 31 },
		{ 273, 99, 31 },
		{ 273, 100, 31 },
		{ 273, 101, 31 },
		{ 273, 102, 31 },
		{ 273, 103, 31 },
		{ 273, 104, 31 },
		{ 273, 105, 31 },
		{ 273, 106, 31 },
		{ 273, 107, 31 },
		{ 273, 108, 31 },
		{ 273, 109, 31 },
		{ 273, 110, 31 },
		{ 273, 111, 31 },
		{ 273, 112, 31 },
		{ 273, 113, 31 },
		{ 273, 114, 31 },
		{ 273, 115, 31 },
		{ 273, 116, 31 },
		{ 273, 117, 31 },
		{ 273, 118, 31 },
		{ 273, 119, 31 },
		{ 273, 120, 31 },
		{ 273, 121, 31 },
		{ 273, 122, 31 },
		{ 273, 123, 31 },
		{ 273, 124, 31 },
		{ 273, 125, 31 },
		{ 273, 126, 31 },
		{ 273, 127, 31 },
		{ 273, 128, 31 },
		{ 273, 129, 31 },
		{ 273, 130, 31 },
		{ 273, 131, 31 },
		{ 273, 132, 31 },
		{ 273, 133, 31 },
		{ 273, 134, 31 },
		{ 273, 135, 31 },
		{ 273, 136, 31 },
		{ 273, 137, 31 },
		{ 273, 138, 31 },
		{ 273, 139, 31 },
		{ 273, 140, 31 },
		{ 273, 141, 31 },
		{ 273, 142, 31 },
		{ 273, 143, 31 },
		{ 273, 144, 31 },
		{ 273, 145, 31 },
		{ 273, 146, 31 },
		{ 273, 147, 31 },
		{ 273, 148, 31 },
		{ 273, 149, 31 },
		{ 273, 150, 31 },
		{ 273, 151, 31 },
		{ 273, 152, 31 },
		{ 273, 153, 31 },
		{ 273, 154, 31 },
		{ 273, 155, 31 },
		{ 273, 156, 31 },
		{ 273, 157, 31 },
		{ 273, 158, 31 },
		{ 273, 159, 31 },
		{ 273, 160, 31 },
		{ 273, 161, 31 },
		{ 273, 162, 31 },
		{ 273, 163, 31 },
		{ 273, 164, 31 },
		{ 273, 165, 31 },
		{ 273, 166, 31 },
		{ 273, 167, 31 },
		{ 273, 168, 31 },
		{ 273, 169, 31 },
		{ 273, 170, 31 },
		{ 273, 171, 31 },
		{ 273, 172, 31 },
		{ 273, 173, 31 },
		{ 273, 174, 31 },
		{ 273, 175, 31 },
		{ 273, 176, 31 },
		{ 273, 177, 31 },
		{ 273, 178, 31 },
		{ 273, 179, 31 },
		{ 273, 180, 31 },
		{ 273, 181, 31 },
		{ 273, 182, 31 },
		{ 273, 183, 31 },
		{ 273, 184, 31 },
		{ 273, 185, 31 },
		{ 273, 186, 31 },
		{ 273, 187, 31 },
		{ 273, 188, 31 },
		{ 273, 189, 31 },
		{ 273, 190, 31 },
		{ 274, 1, 31 },
		{ 274, 2, 31 },
		{ 274, 3, 31 },
		{ 274, 4, 31 },
		{ 274, 5, 31 },
		{ 274, 6, 31 },
		{ 274, 7, 31 },
		{ 274, 8, 31 },
		{ 274, 9, 31 },
		{ 274, 10, 31 },
		{ 274, 11, 31 },
		{ 274, 12, 31 },
		{ 274, 13, 31 },
		{ 274, 14, 31 },
		{ 274, 15, 31 },
		{ 274, 16, 31 },
		{ 274, 17, 31 },
		{ 274, 18, 31 },
		{ 274, 19, 31 },
		{ 274, 20, 31 },
		{ 274, 21, 31 },
		{ 274, 22, 31 },
		{ 274, 23, 31 },
		{ 274, 24, 31 },
		{ 274, 25, 31 },
		{ 274, 26, 31 },
		{ 274, 27, 31 },
		{ 274, 28, 31 },
		{ 274, 29, 31 },
		{ 274, 30, 31 },
		{ 274, 31, 31 },
		{ 274, 32, 31 },
		{ 274, 33, 31 },
		{ 274, 34, 31 },
		{ 274, 35, 31 },
		{ 274, 36, 31 },
		{ 274, 37, 31 },
		{ 274, 38, 31 },
		{ 274, 39, 31 },
		{ 274, 40, 31 },
		{ 274, 41, 31 },
		{ 274, 42, 31 },
		{ 274, 43, 31 },
		{ 274, 44, 31 },
		{ 274, 45, 31 },
		{ 274, 46, 31 },
		{ 274, 47, 31 },
		{ 274, 48, 31 },
		{ 274, 49, 31 },
		{ 274, 50, 31 },
		{ 274, 51, 31 },
		{ 274, 52, 31 },
		{ 274, 53, 31 },
		{ 274, 54, 31 },
		{ 274, 55, 31 },
		{ 274, 56, 31 },
		{ 274, 57, 31 },
		{ 274, 58, 31 },
		{ 274, 59, 31 },
		{ 274, 60, 31 },
		{ 274, 61, 31 },
		{ 274, 62, 31 },
		{ 274, 63, 31 },
		{ 274, 64, 31 },
		{ 274, 65, 31 },
		{ 274, 66, 31 },
		{ 274, 67, 31 },
		{ 274, 68, 31 },
		{ 274, 69, 31 },
		{ 274, 70, 31 },
		{ 274, 71, 31 },
		{ 274, 72, 31 },
		{ 274, 73, 31 },
		{ 274, 74, 31 },
		{ 274, 75, 31 },
		{ 274, 76, 31 },
		{ 274, 77, 31 },
		{ 274, 78, 31 },
		{ 274, 79, 31 },
		{ 274, 80, 31 },
		{ 274, 81, 31 },
		{ 274, 82, 31 },
		{ 274, 83, 31 },
		{ 274, 84, 31 },
		{ 274, 85, 31 },
		{ 274, 86, 31 },
		{ 274, 87, 31 },
		{ 274, 88, 31 },
		{ 274, 89, 31 },
		{ 274, 90, 31 },
		{ 274, 91, 31 },
		{ 274, 92, 31 },
		{ 274, 93, 31 },
		{ 274, 94, 31 },
		{ 274, 95, 31 },
		{ 274, 96, 31 },
		{ 274, 97, 31 },
		{ 274, 98, 31 },
		{ 274, 99, 31 },
		{ 274, 100, 31 },
		{ 274, 101, 31 },
		{ 274, 102, 31 },
		{ 274, 103, 31 },
		{ 274, 104, 31 },
		{ 274, 105, 31 },
		{ 274, 106, 31 },
		{ 274, 107, 31 },
		{ 274, 108, 31 },
		{ 274, 109, 31 },
		{ 274, 110, 31 },
		{ 274, 111, 31 },
		{ 274, 112, 31 },
		{ 274, 113, 31 },
		{ 274, 114, 31 },
		{ 274, 115, 31 },
		{ 274, 116, 31 },
		{ 274, 117, 31 },
		{ 274, 118, 31 },
		{ 274, 119, 31 },
		{ 274, 120, 31 },
		{ 274, 121, 31 },
		{ 274, 122, 31 },
		{ 274, 123, 31 },
		{ 274, 124, 31 },
		{ 274, 125, 31 },
		{ 274, 126, 31 },
		{ 274, 127, 31 },
		{ 274, 128, 31 },
		{ 274, 129, 31 },
		{ 274, 130, 31 },
		{ 274, 131, 31 },
		{ 274, 132, 31 },
		{ 274, 133, 31 },
		{ 274, 134, 31 },
		{ 274, 135, 31 },
		{ 274, 136, 31 },
		{ 274, 137, 31 },
		{ 274, 138, 31 },
		{ 274, 139, 31 },
		{ 274, 140, 31 },
		{ 274, 141, 31 },
		{ 274, 142, 31 },
		{ 274, 143, 31 },
		{ 274, 144, 31 },
		{ 274, 145, 31 },
		{ 274, 146, 31 },
		{ 274, 147, 31 },
		{ 274, 148, 31 },
		{ 274, 149, 31 },
		{ 274, 150, 31 },
		{ 274, 151, 31 },
		{ 274, 152, 31 },
		{ 274, 153, 31 },
		{ 274, 154, 31 },
		{ 274, 155, 31 },
		{ 274, 156, 31 },
		{ 274, 157, 31 },
		{ 274, 158, 31 },
		{ 274, 159, 31 },
		{ 274, 160, 31 },
		{ 274, 161, 31 },
		{ 274, 162, 31 },
		{ 274, 163, 31 },
		{ 274, 164, 31 },
		{ 274, 165, 31 },
		{ 274, 166, 31 },
		{ 274, 167, 31 },
		{ 274, 168, 31 },
		{ 274, 169, 31 },
		{ 274, 170, 31 },
		{ 274, 171, 31 },
		{ 274, 172, 31 },
		{ 274, 173, 31 },
		{ 274, 174, 31 },
		{ 274, 175, 31 },
		{ 274, 176, 31 },
		{ 274, 177, 31 },
		{ 274, 178, 31 },
		{ 274, 179, 31 },
		{ 274, 180, 31 },
		{ 274, 181, 31 },
		{ 274, 182, 31 },
		{ 274, 183, 31 },
		{ 274, 184, 31 },
		{ 274, 185, 31 },
		{ 274, 186, 31 },
		{ 274, 187, 31 },
		{ 275, 1, 31 },
		{ 275, 2, 31 },
		{ 275, 3, 31 },
		{ 275, 4, 31 },
		{ 275, 5, 31 },
		{ 275, 6, 31 },
		{ 275, 7, 31 },
		{ 275, 8, 31 },
		{ 275, 9, 31 },
		{ 275, 10, 31 },
		{ 275, 11, 31 },
		{ 275, 12, 31 },
		{ 275, 13, 31 },
		{ 275, 14, 31 },
		{ 275, 15, 31 },
		{ 275, 16, 31 },
		{ 275, 17, 31 },
		{ 275, 18, 31 },
		{ 275, 19, 31 },
		{ 275, 20, 31 },
		{ 275, 21, 31 },
		{ 275, 22, 31 },
		{ 275, 23, 31 },
		{ 275, 24, 31 },
		{ 275, 25, 31 },
		{ 275, 26, 31 },
		{ 275, 27, 31 },
		{ 275, 28, 31 },
		{ 275, 29, 31 },
		{ 275, 30, 31 },
		{ 275, 31, 31 },
		{ 275, 32, 31 },
		{ 275, 33, 31 },
		{ 275, 34, 31 },
		{ 275, 35, 31 },
		{ 275, 36, 31 },
		{ 275, 37, 31 },
		{ 275, 38, 31 },
		{ 275, 39, 31 },
		{ 275, 40, 31 },
		{ 275, 41, 31 },
		{ 275, 42, 31 },
		{ 275, 43, 31 },
		{ 275, 44, 31 },
		{ 275, 45, 31 },
		{ 275, 46, 31 },
		{ 275, 47, 31 },
		{ 275, 48, 31 },
		{ 275, 49, 31 },
		{ 275, 50, 31 },
		{ 275, 51, 31 },
		{ 275, 52, 31 },
		{ 275, 53, 31 },
		{ 275, 54, 31 },
		{ 275, 55, 31 },
		{ 275, 56, 31 },
		{ 275, 57, 31 },
		{ 275, 58, 31 },
		{ 275, 59, 31 },
		{ 275, 60, 31 },
		{ 275, 61, 31 },
		{ 275, 62, 31 },
		{ 275, 63, 31 },
		{ 275, 64, 31 },
		{ 275, 65, 31 },
		{ 275, 66, 31 },
		{ 275, 67, 31 },
		{ 275, 68, 31 },
		{ 275, 69, 31 },
		{ 275, 70, 31 },
		{ 275, 71, 31 },
		{ 275, 72, 31 },
		{ 275, 73, 31 },
		{ 275, 74, 31 },
		{ 275, 75, 31 },
		{ 275, 76, 31 },
		{ 275, 77, 31 },
		{ 275, 78, 31 },
		{ 275, 79, 31 },
		{ 275, 80, 31 },
		{ 275, 81, 31 },
		{ 275, 82, 31 },
		{ 275, 83, 31 },
		{ 275, 84, 31 },
		{ 275, 85, 31 },
		{ 275, 86, 31 },
		{ 275, 87, 31 },
		{ 275, 88, 31 },
		{ 275, 89, 31 },
		{ 275, 90, 31 },
		{ 275, 91, 31 },
		{ 275, 92, 31 },
		{ 275, 93, 31 },
		{ 275, 94, 31 },
		{ 275, 95, 31 },
		{ 275, 96, 31 },
		{ 275, 97, 31 },
		{ 275, 98, 31 },
		{ 275, 99, 31 },
		{ 275, 100, 31 },
		{ 275, 101, 31 },
		{ 275, 102, 31 },
		{ 275, 103, 31 },
		{ 275, 104, 31 },
		{ 275, 105, 31 },
		{ 275, 106, 31 },
		{ 275, 107, 31 },
		{ 275, 108, 31 },
		{ 275, 109, 31 },
		{ 275, 110, 31 },
		{ 275, 111, 31 },
		{ 275, 112, 31 },
		{ 275, 113, 31 },
		{ 275, 114, 31 },
		{ 275, 115, 31 },
		{ 275, 116, 31 },
		{ 275, 117, 31 },
		{ 275, 118, 31 },
		{ 275, 119, 31 },
		{ 275, 120, 31 },
		{ 275, 121, 31 },
		{ 275, 122, 31 },
		{ 275, 123, 31 },
		{ 275, 124, 31 },
		{ 275, 125, 31 },
		{ 275, 126, 31 },
		{ 275, 127, 31 },
		{ 275, 128, 31 },
		{ 275, 129, 31 },
		{ 275, 130, 31 },
		{ 275, 131, 31 },
		{ 275, 132, 31 },
		{ 275, 133, 31 },
		{ 275, 134, 31 },
		{ 275, 135, 31 },
		{ 275, 136, 31 },
		{ 275, 137, 31 },
		{ 275, 138, 31 },
		{ 275, 139, 31 },
		{ 275, 140, 31 },
		{ 275, 141, 31 },
		{ 275, 142, 31 },
		{ 275, 143, 31 },
		{ 275, 144, 31 },
		{ 275, 145, 31 },
		{ 275, 146, 31 },
		{ 275, 147, 31 },
		{ 275, 148, 31 },
		{ 275, 149, 31 },
		{ 275, 150, 31 },
		{ 275, 151, 31 },
		{ 275, 152, 31 },
		{ 275, 153, 31 },
		{ 275, 154, 31 },
		{ 275, 155, 31 },
		{ 275, 156, 31 },
		{ 275, 157, 31 },
		{ 275, 158, 31 },
		{ 275, 159, 31 },
		{ 275, 160, 31 },
		{ 275, 161, 31 },
		{ 275, 162, 31 },
		{ 275, 163, 31 },
		{ 275, 164, 31 },
		{ 275, 165, 31 },
		{ 275, 166, 31 },
		{ 275, 167, 31 },
		{ 275, 168, 31 },
		{ 275, 169, 31 },
		{ 275, 170, 31 },
		{ 275, 171, 31 },
		{ 275, 172, 31 },
		{ 275, 173, 31 },
		{ 275, 174, 31 },
		{ 275, 175, 31 },
		{ 275, 176, 31 },
		{ 275, 177, 31 },
		{ 275, 178, 31 },
		{ 275, 179, 31 },
		{ 275, 180, 31 },
		{ 275, 181, 31 },
		{ 275, 182, 31 },
		{ 275, 183, 31 },
		{ 275, 184, 31 },
		{ 276, 1, 31 },
		{ 276, 2, 31 },
		{ 276, 3, 31 },
		{ 276, 4, 31 },
		{ 276, 5, 31 },
		{ 276, 6, 31 },
		{ 276, 7, 31 },
		{ 276, 8, 31 },
		{ 276, 9, 31 },
		{ 276, 10, 31 },
		{ 276, 11, 31 },
		{ 276, 12, 31 },
		{ 276, 13, 31 },
		{ 276, 14, 31 },
		{ 276, 15, 31 },
		{ 276, 16, 31 },
		{ 276, 17, 31 },
		{ 276, 18, 31 },
		{ 276, 19, 31 },
		{ 276, 20, 31 },
		{ 276, 21, 31 },
		{ 276, 22, 31 },
		{ 276, 23, 31 },
		{ 276, 24, 31 },
		{ 276, 25, 31 },
		{ 276, 26, 31 },
		{ 276, 27, 31 },
		{ 276, 28, 31 },
		{ 276, 29, 31 },
		{ 276, 30, 31 },
		{ 276, 31, 31 },
		{ 276, 32, 31 },
		{ 276, 33, 31 },
		{ 276, 34, 31 },
		{ 276, 35, 31 },
		{ 276, 36, 31 },
		{ 276, 37, 31 },
		{ 276, 38, 31 },
		{ 276, 39, 31 },
		{ 276, 40, 31 },
		{ 276, 41, 31 },
		{ 276, 42, 31 },
		{ 276, 43, 31 },
		{ 276, 44, 31 },
		{ 276, 45, 31 },
		{ 276, 46, 31 },
		{ 276, 47, 31 },
		{ 276, 48, 31 },
		{ 276, 49, 31 },
		{ 276, 50, 31 },
		{ 276, 51, 31 },
		{ 276, 52, 31 },
		{ 276, 53, 31 },
		{ 276, 54, 31 },
		{ 276, 55, 31 },
		{ 276, 56, 31 },
		{ 276, 57, 31 },
		{ 276, 58, 31 },
		{ 276, 59, 31 },
		{ 276, 60, 31 },
		{ 276, 61, 31 },
		{ 276, 62, 31 },
		{ 276, 63, 31 },
		{ 276, 64, 31 },
		{ 276, 65, 31 },
		{ 276, 66, 31 },
		{ 276, 67, 31 },
		{ 276, 68, 31 },
		{ 276, 69, 31 },
		{ 276, 70, 31 },
		{ 276, 71, 31 },
		{ 276, 72, 31 },
		{ 276, 73, 31 },
		{ 276, 74, 31 },
		{ 276, 75, 31 },
		{ 276, 76, 31 },
		{ 276, 77, 31 },
		{ 276, 78, 31 },
		{ 276, 79, 31 },
		{ 276, 80, 31 },
		{ 276, 81, 31 },
		{ 276, 82, 31 },
		{ 276, 83, 31 },
		{ 276, 84, 31 },
		{ 276, 85, 31 },
		{ 276, 86, 31 },
		{ 276, 87, 31 },
		{ 276, 88, 31 },
		{ 276, 89, 31 },
		{ 276, 90, 31 },
		{ 276, 91, 31 },
		{ 276, 92, 31 },
		{ 276, 93, 31 },
		{ 276, 94, 31 },
		{ 276, 95, 31 },
		{ 276, 96, 31 },
		{ 276, 97, 31 },
		{ 276, 98, 31 },
		{ 276, 99, 31 },
		{ 276, 100, 31 },
		{ 276, 101, 31 },
		{ 276, 102, 31 },
		{ 276, 103, 31 },
		{ 276, 104, 31 },
		{ 276, 105, 31 },
		{ 276, 106, 31 },
		{ 276, 107, 31 },
		{ 276, 108, 31 },
		{ 276, 109, 31 },
		{ 276, 110, 31 },
		{ 276, 111, 31 },
		{ 276, 112, 31 },
		{ 276, 113, 31 },
		{ 276, 114, 31 },
		{ 276, 115, 31 },
		{ 276, 116, 31 },
		{ 276, 117, 31 },
		{ 276, 118, 31 },
		{ 276, 119, 31 },
		{ 276, 120, 31 },
		{ 276, 121, 31 },
		{ 276, 122, 31 },
		{ 276, 123, 31 },
		{ 276, 124, 31 },
		{ 276, 125, 31 },
		{ 276, 126, 31 },
		{ 276, 127, 31 },
		{ 276, 128, 31 },
		{ 276, 129, 31 },
		{ 276, 130, 31 },
		{ 276, 131, 31 },
		{ 276, 132, 31 },
		{ 276, 133, 31 },
		{ 276, 134, 31 },
		{ 276, 135, 31 },
		{ 276, 136, 31 },
		{ 276, 137, 31 },
		{ 276, 138, 31 },
		{ 276, 139, 31 },
		{ 276, 140, 31 },
		{ 276, 141, 31 },
		{ 276, 142, 31 },
		{ 276, 143, 31 },
		{ 276, 144, 31 },
		{ 276, 145, 31 },
		{ 276, 146, 31 },
		{ 276, 147, 31 },
		{ 276, 148, 31 },
		{ 276, 149, 31 },
		{ 276, 150, 31 },
		{ 276, 151, 31 },
		{ 276, 152, 31 },
		{ 276, 153, 31 },
		{ 276, 154, 31 },
		{ 276, 155, 31 },
		{ 276, 156, 31 },
		{ 276, 157, 31 },
		{ 276, 158, 31 },
		{ 276, 159, 31 },
		{ 276, 160, 31 },
		{ 276, 161, 31 },
		{ 276, 162, 31 },
		{ 276, 163, 31 },
		{ 276, 164, 31 },
		{ 276, 165, 31 },
		{ 276, 166, 31 },
		{ 276, 167, 31 },
		{ 276, 168, 31 },
		{ 276, 169, 31 },
		{ 276, 170, 31 },
		{ 276, 171, 31 },
		{ 276, 172, 31 },
		{ 276, 173, 31 },
		{ 276, 174, 31 },
		{ 276, 175, 31 },
		{ 276, 176, 31 },
		{ 276, 177, 31 },
		{ 276, 178, 31 },
		{ 276, 179, 31 },
		{ 276, 180, 31 },
		{ 277, 1, 31 },
		{ 277, 2, 31 },
		{ 277, 3, 31 },
		{ 277, 4, 31 },
		{ 277, 5, 31 },
		{ 277, 6, 31 },
		{ 277, 7, 31 },
		{ 277, 8, 31 },
		{ 277, 9, 31 },
		{ 277, 10, 31 },
		{ 277, 11, 31 },
		{ 277, 12, 31 },
		{ 277, 13, 31 },
		{ 277, 14, 31 },
		{ 277, 15, 31 },
		{ 277, 16, 31 },
		{ 277, 17, 31 },
		{ 277, 18, 31 },
		{ 277, 19, 31 },
		{ 277, 20, 31 },
		{ 277, 21, 31 },
		{ 277, 22, 31 },
		{ 277, 23, 31 },
		{ 277, 24, 31 },
		{ 277, 25, 31 },
		{ 277, 26, 31 },
		{ 277, 27, 31 },
		{ 277, 28, 31 },
		{ 277, 29, 31 },
		{ 277, 30, 31 },
		{ 277, 31, 31 },
		{ 277, 32, 31 },
		{ 277, 33, 31 },
		{ 277, 34, 31 },
		{ 277, 35, 31 },
		{ 277, 36, 31 },
		{ 277, 37, 31 },
		{ 277, 38, 31 },
		{ 277, 39, 31 },
		{ 277, 40, 31 },
		{ 277, 41, 31 },
		{ 277, 42, 31 },
		{ 277, 43, 31 },
		{ 277, 44, 31 },
		{ 277, 45, 31 },
		{ 277, 46, 31 },
		{ 277, 47, 31 },
		{ 277, 48, 31 },
		{ 277, 49, 31 },
		{ 277, 50, 31 },
		{ 277, 51, 31 },
		{ 277, 52, 31 },
		{ 277, 53, 31 },
		{ 277, 54, 31 },
		{ 277, 55, 31 },
		{ 277, 56, 31 },
		{ 277, 57, 31 },
		{ 277, 58, 31 },
		{ 277, 59, 31 },
		{ 277, 60, 31 },
		{ 277, 61, 31 },
		{ 277, 62, 31 },
		{ 277, 63, 31 },
		{ 277, 64, 31 },
		{ 277, 65, 31 },
		{ 277, 66, 31 },
		{ 277, 67, 31 },
		{ 277, 68, 31 },
		{ 277, 69, 31 },
		{ 277, 70, 31 },
		{ 277, 71, 31 },
		{ 277, 72, 31 },
		{ 277, 73, 31 },
		{ 277, 74, 31 },
		{ 277, 75, 31 },
		{ 277, 76, 31 },
		{ 277, 77, 31 },
		{ 277, 78, 31 },
		{ 277, 79, 31 },
		{ 277, 80, 31 },
		{ 277, 81, 31 },
		{ 277, 82, 31 },
		{ 277, 83, 31 },
		{ 277, 84, 31 },
		{ 277, 85, 31 },
		{ 277, 86, 31 },
		{ 277, 87, 31 },
		{ 277, 88, 31 },
		{ 277, 89, 31 },
		{ 277, 90, 31 },
		{ 277, 91, 31 },
		{ 277, 92, 31 },
		{ 277, 93, 31 },
		{ 277, 94, 31 },
		{ 277, 95, 31 },
		{ 277, 96, 31 },
		{ 277, 97, 31 },
		{ 277, 98, 31 },
		{ 277, 99, 31 },
		{ 277, 100, 31 },
		{ 277, 101, 31 },
		{ 277, 102, 31 },
		{ 277, 103, 31 },
		{ 277, 104, 31 },
		{ 277, 105, 31 },
		{ 277, 106, 31 },
		{ 277, 107, 31 },
		{ 277, 108, 31 },
		{ 277, 109, 31 },
		{ 277, 110, 31 },
		{ 277, 111, 31 },
		{ 277, 112, 31 },
		{ 277, 113, 31 },
		{ 277, 114, 31 },
		{ 277, 115, 31 },
		{ 277, 116, 31 },
		{ 277, 117, 31 },
		{ 277, 118, 31 },
		{ 277, 119, 31 },
		{ 277, 120, 31 },
		{ 277, 121, 31 },
		{ 277, 122, 31 },
		{ 277, 123, 31 },
		{ 277, 124, 31 },
		{ 277, 125, 31 },
		{ 277, 126, 31 },
		{ 277, 127, 31 },
		{ 277, 128, 31 },
		{ 277, 129, 31 },
		{ 277, 130, 31 },
		{ 277, 131, 31 },
		{ 277, 132, 31 },
		{ 277, 133, 31 },
		{ 277, 134, 31 },
		{ 277, 135, 31 },
		{ 277, 136, 31 },
		{ 277, 137, 31 },
		{ 277, 138, 31 },
		{ 277, 139, 31 },
		{ 277, 140, 31 },
		{ 277, 141, 31 },
		{ 277, 142, 31 },
		{ 277, 143, 31 },
		{ 277, 144, 31 },
		{ 277, 145, 31 },
		{ 277, 146, 31 },
		{ 277, 147, 31 },
		{ 277, 148, 31 },
		{ 277, 149, 31 },
		{ 277, 150, 31 },
		{ 277, 151, 31 },
		{ 277, 152, 31 },
		{ 277, 153, 31 },
		{ 277, 154, 31 },
		{ 277, 155, 31 },
		{ 277, 156, 31 },
		{ 277, 157, 31 },
		{ 277, 158, 31 },
		{ 277, 159, 31 },
		{ 277, 160, 31 },
		{ 277, 161, 31 },
		{ 277, 162, 31 },
		{ 277, 163, 31 },
		{ 277, 164, 31 },
		{ 277, 165, 31 },
		{ 277, 166, 31 },
		{ 277, 167, 31 },
		{ 277, 168, 31 },
		{ 277, 169, 31 },
		{ 277, 170, 31 },
		{ 277, 171, 31 },
		{ 277, 172, 31 },
		{ 277, 173, 31 },
		{ 277, 174, 31 },
		{ 277, 175, 31 },
		{ 277, 176, 31 },
		{ 278, 1, 31 },
		{ 278, 2, 31 },
		{ 278, 3, 31 },
		{ 278, 4, 31 },
		{ 278, 5, 31 },
		{ 278, 6, 31 },
		{ 278, 7, 31 },
		{ 278, 8, 31 },
		{ 278, 9, 31 },
		{ 278, 10, 31 },
		{ 278, 11, 31 },
		{ 278, 12, 31 },
		{ 278, 13, 31 },
		{ 278, 14, 31 },
		{ 278, 15, 31 },
		{ 278, 16, 31 },
		{ 278, 17, 31 },
		{ 278, 18, 31 },
		{ 278, 19, 31 },
		{ 278, 20, 31 },
		{ 278, 21, 31 },
		{ 278, 22, 31 },
		{ 278, 23, 31 },
		{ 278, 24, 31 },
		{ 278, 25, 31 },
		{ 278, 26, 31 },
		{ 278, 27, 31 },
		{ 278, 28, 31 },
		{ 278, 29, 31 },
		{ 278, 30, 31 },
		{ 278, 31, 31 },
		{ 278, 32, 31 },
		{ 278, 33, 31 },
		{ 278, 34, 31 },
		{ 278, 35, 31 },
		{ 278, 36, 31 },
		{ 278, 37, 31 },
		{ 278, 38, 31 },
		{ 278, 39, 31 },
		{ 278, 40, 31 },
		{ 278, 41, 31 },
		{ 278, 42, 31 },
		{ 278, 43, 31 },
		{ 278, 44, 31 },
		{ 278, 45, 31 },
		{ 278, 46, 31 },
		{ 278, 47, 31 },
		{ 278, 48, 31 },
		{ 278, 49, 31 },
		{ 278, 50, 31 },
		{ 278, 51, 31 },
		{ 278, 52, 31 },
		{ 278, 53, 31 },
		{ 278, 54, 31 },
		{ 278, 55, 31 },
		{ 278, 56, 31 },
		{ 278, 57, 31 },
		{ 278, 58, 31 },
		{ 278, 59, 31 },
		{ 278, 60, 31 },
		{ 278, 61, 31 },
		{ 278, 62, 31 },
		{ 278, 63, 31 },
		{ 278, 64, 31 },
		{ 278, 65, 31 },
		{ 278, 66, 31 },
		{ 278, 67, 31 },
		{ 278, 68, 31 },
		{ 278, 69, 31 },
		{ 278, 70, 31 },
		{ 278, 71, 31 },
		{ 278, 72, 31 },
		{ 278, 73, 31 },
		{ 278, 74, 31 },
		{ 278, 75, 31 },
		{ 278, 76, 31 },
		{ 278, 77, 31 },
		{ 278, 78, 31 },
		{ 278, 79, 31 },
		{ 278, 80, 31 },
		{ 278, 81, 31 },
		{ 278, 82, 31 },
		{ 278, 83, 31 },
		{ 278, 84, 31 },
		{ 278, 85, 31 },
		{ 278, 86, 31 },
		{ 278, 87, 31 },
		{ 278, 88, 31 },
		{ 278, 89, 31 },
		{ 278, 90, 31 },
		{ 278, 91, 31 },
		{ 278, 92, 31 },
		{ 278, 93, 31 },
		{ 278, 94, 31 },
		{ 278, 95, 31 },
		{ 278, 96, 31 },
		{ 278, 97, 31 },
		{ 278, 98, 31 },
		{ 278, 99, 31 },
		{ 278, 100, 31 },
		{ 278, 101, 31 },
		{ 278, 102, 31 },
		{ 278, 103, 31 },
		{ 278, 104, 31 },
		{ 278, 105, 31 },
		{ 278, 106, 31 },
		{ 278, 107, 31 },
		{ 278, 108, 31 },
		{ 278, 109, 31 },
		{ 278, 110, 31 },
		{ 278, 111, 31 },
		{ 278, 112, 31 },
		{ 278, 113, 31 },
		{ 278, 114, 31 },
		{ 278, 115, 31 },
		{ 278, 116, 31 },
		{ 278, 117, 31 },
		{ 278, 118, 31 },
		{ 278, 119, 31 },
		{ 278, 120, 31 },
		{ 278, 121, 31 },
		{ 278, 122, 31 },
		{ 278, 123, 31 },
		{ 278, 124, 31 },
		{ 278, 125, 31 },
		{ 278, 126, 31 },
		{ 278, 127, 31 },
		{ 278, 128, 31 },
		{ 278, 129, 31 },
		{ 278, 130, 31 },
		{ 278, 131, 31 },
		{ 278, 132, 31 },
		{ 278, 133, 31 },
		{ 278, 134, 31 },
		{ 278, 135, 31 },
		{ 278, 136, 31 },
		{ 278, 137, 31 },
		{ 278, 138, 31 },
		{ 278, 139, 31 },
		{ 278, 140, 31 },
		{ 278, 141, 31 },
		{ 278, 142, 31 },
		{ 278, 143, 31 },
		{ 278, 144, 31 },
		{ 278, 145, 31 },
		{ 278, 146, 31 },
		{ 278, 147, 31 },
		{ 278, 148, 31 },
		{ 278, 149, 31 },
		{ 278, 150, 31 },
		{ 278, 151, 31 },
		{ 278, 152, 31 },
		{ 278, 153, 31 },
		{ 278, 154, 31 },
		{ 278, 155, 31 },
		{ 278, 156, 31 },
		{ 278, 157, 31 },
		{ 278, 158, 31 },
		{ 278, 159, 31 },
		{ 278, 160, 31 },
		{ 278, 161, 31 },
		{ 278, 162, 31 },
		{ 278, 163, 31 },
		{ 278, 164, 31 },
		{ 278, 165, 31 },
		{ 278, 166, 31 },
		{ 278, 167, 31 },
		{ 278, 168, 31 },
		{ 278, 169, 31 },
		{ 278, 170, 31 },
		{ 278, 171, 31 },
		{ 278, 172, 31 },
		{ 279, 1, 31 },
		{ 279, 2, 31 },
		{ 279, 3, 31 },
		{ 279, 4, 31 },
		{ 279, 5, 31 },
		{ 279, 6, 31 },
		{ 279, 7, 31 },
		{ 279, 8, 31 },
		{ 279, 9, 31 },
		{ 279, 10, 31 },
		{ 279, 11, 31 },
		{ 279, 12, 31 },
		{ 279, 13, 31 },
		{ 279, 14, 31 },
		{ 279, 15, 31 },
		{ 279, 16, 31 },
		{ 279, 17, 31 },
		{ 279, 18, 31 },
		{ 279, 19, 31 },
		{ 279, 20, 31 },
		{ 279, 21, 31 },
		{ 279, 22, 31 },
		{ 279, 23, 31 },
		{ 279, 24, 31 },
		{ 279, 25, 31 },
		{ 279, 26, 31 },
		{ 279, 27, 31 },
		{ 279, 28, 31 },
		{ 279, 29, 31 },
		{ 279, 30, 31 },
		{ 279, 31, 31 },
		{ 279, 32, 31 },
		{ 279, 33, 31 },
		{ 279, 34, 31 },
		{ 279, 35, 31 },
		{ 279, 36, 31 },
		{ 279, 37, 31 },
		{ 279, 38, 31 },
		{ 279, 39, 31 },
		{ 279, 40, 31 },
		{ 279, 41, 31 },
		{ 279, 42, 31 },
		{ 279, 43, 31 },
		{ 279, 44, 31 },
		{ 279, 45, 31 },
		{ 279, 46, 31 },
		{ 279, 47, 31 },
		{ 279, 48, 31 },
		{ 279, 49, 31 },
		{ 279, 50, 31 },
		{ 279, 51, 31 },
		{ 279, 52, 31 },
		{ 279, 53, 31 },
		{ 279, 54, 31 },
		{ 279, 55, 31 },
		{ 279, 56, 31 },
		{ 279, 57, 31 },
		{ 279, 58, 31 },
		{ 279, 59, 31 },
		{ 279, 60, 31 },
		{ 279, 61, 31 },
		{ 279, 62, 31 },
		{ 279, 63, 31 },
		{ 279, 64, 31 },
		{ 279, 65, 31 },
		{ 279, 66, 31 },
		{ 279, 67, 31 },
		{ 279, 68, 31 },
		{ 279, 69, 31 },
		{ 279, 70, 31 },
		{ 279, 71, 31 },
		{ 279, 72, 31 },
		{ 279, 73, 31 },
		{ 279, 74, 31 },
		{ 279, 75, 31 },
		{ 279, 76, 31 },
		{ 279, 77, 31 },
		{ 279, 78, 31 },
		{ 279, 79, 31 },
		{ 279, 80, 31 },
		{ 279, 81, 31 },
		{ 279, 82, 31 },
		{ 279, 83, 31 },
		{ 279, 84, 31 },
		{ 279, 85, 31 },
		{ 279, 86, 31 },
		{ 279, 87, 31 },
		{ 279, 88, 31 },
		{ 279, 89, 31 },
		{ 279, 90, 31 },
		{ 279, 91, 31 },
		{ 279, 92, 31 },
		{ 279, 93, 31 },
		{ 279, 94, 31 },
		{ 279, 95, 31 },
		{ 279, 96, 31 },
		{ 279, 97, 31 },
		{ 279, 98, 31 },
		{ 279, 99, 31 },
		{ 279, 100, 31 },
		{ 279, 101, 31 },
		{ 279, 102, 31 },
		{ 279, 103, 31 },
		{ 279, 104, 31 },
		{ 279, 105, 31 },
		{ 279, 106, 31 },
		{ 279, 107, 31 },
		{ 279, 108, 31 },
		{ 279, 109, 31 },
		{ 279, 110, 31 },
		{ 279, 111, 31 },
		{ 279, 112, 31 },
		{ 279, 113, 31 },
		{ 279, 114, 31 },
		{ 279, 115, 31 },
		{ 279, 116, 31 },
		{ 279, 117, 31 },
		{ 279, 118, 31 },
		{ 279, 119, 31 },
		{ 279, 120, 31 },
		{ 279, 121, 31 },
		{ 279, 122, 31 },
		{ 279, 123, 31 },
		{ 279, 124, 31 },
		{ 279, 125, 31 },
		{ 279, 126, 31 },
		{ 279, 127, 31 },
		{ 279, 128, 31 },
		{ 279, 129, 31 },
		{ 279, 130, 31 },
		{ 279, 131, 31 },
		{ 279, 132, 31 },
		{ 279, 133, 31 },
		{ 279, 134, 31 },
		{ 279, 135, 31 },
		{ 279, 136, 31 },
		{ 279, 137, 31 },
		{ 279, 138, 31 },
		{ 279, 139, 31 },
		{ 279, 140, 31 },
		{ 279, 141, 31 },
		{ 279, 142, 31 },
		{ 279, 143, 31 },
		{ 279, 144, 31 },
		{ 279, 145, 31 },
		{ 279, 146, 31 },
		{ 279, 147, 31 },
		{ 279, 148, 31 },
		{ 279, 149, 31 },
		{ 279, 150, 31 },
		{ 279, 151, 31 },
		{ 279, 152, 31 },
		{ 279, 153, 31 },
		{ 279, 154, 31 },
		{ 279, 155, 31 },
		{ 279, 156, 31 },
		{ 279, 157, 31 },
		{ 279, 158, 31 },
		{ 279, 159, 31 },
		{ 279, 160, 31 },
		{ 279, 161, 31 },
		{ 279, 162, 31 },
		{ 279, 163, 31 },
		{ 279, 164, 31 },
		{ 279, 165, 31 },
		{ 279, 166, 31 },
		{ 279, 167, 31 },
		{ 279, 168, 31 },
		{ 280, 1, 31 },
		{ 280, 2, 31 },
		{ 280, 3, 31 },
		{ 280, 4, 31 },
		{ 280, 5, 31 },
		{ 280, 6, 31 },
		{ 280, 7, 31 },
		{ 280, 8, 31 },
		{ 280, 9, 31 },
		{ 280, 10, 31 },
		{ 280, 11, 31 },
		{ 280, 12, 31 },
		{ 280, 13, 31 },
		{ 280, 14, 31 },
		{ 280, 15, 31 },
		{ 280, 16, 31 },
		{ 280, 17, 31 },
		{ 280, 18, 31 },
		{ 280, 19, 31 },
		{ 280, 20, 31 },
		{ 280, 21, 31 },
		{ 280, 22, 31 },
		{ 280, 23, 31 },
		{ 280, 24, 31 },
		{ 280, 25, 31 },
		{ 280, 26, 31 },
		{ 280, 27, 31 },
		{ 280, 28, 31 },
		{ 280, 29, 31 },
		{ 280, 30, 31 },
		{ 280, 31, 31 },
		{ 280, 32, 31 },
		{ 280, 33, 31 },
		{ 280, 34, 31 },
		{ 280, 35, 31 },
		{ 280, 36, 31 },
		{ 280, 37, 31 },
		{ 280, 38, 31 },
		{ 280, 39, 31 },
		{ 280, 40, 31 },
		{ 280, 41, 31 },
		{ 280, 42, 31 },
		{ 280, 43, 31 },
		{ 280, 44, 31 },
		{ 280, 45, 31 },
		{ 280, 46, 31 },
		{ 280, 47, 31 },
		{ 280, 48, 31 },
		{ 280, 49, 31 },
		{ 280, 50, 31 },
		{ 280, 51, 31 },
		{ 280, 52, 31 },
		{ 280, 53, 31 },
		{ 280, 54, 31 },
		{ 280, 55, 31 },
		{ 280, 56, 31 },
		{ 280, 57, 31 },
		{ 280, 58, 31 },
		{ 280, 59, 31 },
		{ 280, 60, 31 },
		{ 280, 61, 31 },
		{ 280, 62, 31 },
		{ 280, 63, 31 },
		{ 280, 64, 31 },
		{ 280, 65, 31 },
		{ 280, 66, 31 },
		{ 280, 67, 31 },
		{ 280, 68, 31 },
		{ 280, 69, 31 },
		{ 280, 70, 31 },
		{ 280, 71, 31 },
		{ 280, 72, 31 },
		{ 280, 73, 31 },
		{ 280, 74, 31 },
		{ 280, 75, 31 },
		{ 280, 76, 31 },
		{ 280, 77, 31 },
		{ 280, 78, 31 },
		{ 280, 79, 31 },
		{ 280, 80, 31 },
		{ 280, 81, 31 },
		{ 280, 82, 31 },
		{ 280, 83, 31 },
		{ 280, 84, 31 },
		{ 280, 85, 31 },
		{ 280, 86, 31 },
		{ 280, 87, 31 },
		{ 280, 88, 31 },
		{ 280, 89, 31 },
		{ 280, 90, 31 },
		{ 280, 91, 31 },
		{ 280, 92, 31 },
		{ 280, 93, 31 },
		{ 280, 94, 31 },
		{ 280, 95, 31 },
		{ 280, 96, 31 },
		{ 280, 97, 31 },
		{ 280, 98, 31 },
		{ 280, 99, 31 },
		{ 280, 100, 31 },
		{ 280, 101, 31 },
		{ 280, 102, 31 },
		{ 280, 103, 31 },
		{ 280, 104, 31 },
		{ 280, 105, 31 },
		{ 280, 106, 31 },
		{ 280, 107, 31 },
		{ 280, 108, 31 },
		{ 280, 109, 31 },
		{ 280, 110, 31 },
		{ 280, 111, 31 },
		{ 280, 112, 31 },
		{ 280, 113, 31 },
		{ 280, 114, 31 },
		{ 280, 115, 31 },
		{ 280, 116, 31 },
		{ 280, 117, 31 },
		{ 280, 118, 31 },
		{ 280, 119, 31 },
		{ 280, 120, 31 },
		{ 280, 121, 31 },
		{ 280, 122, 31 },
		{ 280, 123, 31 },
		{ 280, 124, 31 },
		{ 280, 125, 31 },
		{ 280, 126, 31 },
		{ 280, 127, 31 },
		{ 280, 128, 31 },
		{ 280, 129, 31 },
		{ 280, 130, 31 },
		{ 280, 131, 31 },
		{ 280, 132, 31 },
		{ 280, 133, 31 },
		{ 280, 134, 31 },
		{ 280, 135, 31 },
		{ 280, 136, 31 },
		{ 280, 137, 31 },
		{ 280, 138, 31 },
		{ 280, 139, 31 },
		{ 280, 140, 31 },
		{ 280, 141, 31 },
		{ 280, 142, 31 },
		{ 280, 143, 31 },
		{ 280, 144, 31 },
		{ 280, 145, 31 },
		{ 280, 146, 31 },
		{ 280, 147, 31 },
		{ 280, 148, 31 },
		{ 280, 149, 31 },
		{ 280, 150, 31 },
		{ 280, 151, 31 },
		{ 280, 152, 31 },
		{ 280, 153, 31 },
		{ 280, 154, 31 },
		{ 280, 155, 31 },
		{ 280, 156, 31 },
		{ 280, 157, 31 },
		{ 280, 158, 31 },
		{ 280, 159, 31 },
		{ 280, 160, 31 },
		{ 280, 161, 31 },
		{ 280, 162, 31 },
		{ 280, 163, 31 },
		{ 280, 164, 31 },
		{ 281, 1, 31 },
		{ 281, 2, 31 },
		{ 281, 3, 31 },
		{ 281, 4, 31 },
		{ 281, 5, 31 },
		{ 281, 6, 31 },
		{ 281, 7, 31 },
		{ 281, 8, 31 },
		{ 281, 9, 31 },
		{ 281, 10, 31 },
		{ 281, 11, 31 },
		{ 281, 12, 31 },
		{ 281, 13, 31 },
		{ 281, 14, 31 },
		{ 281, 15, 31 },
		{ 281, 16, 31 },
		{ 281, 17, 31 },
		{ 281, 18, 31 },
		{ 281, 19, 31 },
		{ 281, 20, 31 },
		{ 281, 21, 31 },
		{ 281, 22, 31 },
		{ 281, 23, 31 },
		{ 281, 24, 31 },
		{ 281, 25, 31 },
		{ 281, 26, 31 },
		{ 281, 27, 31 },
		{ 281, 28, 31 },
		{ 281, 29, 31 },
		{ 281, 30, 31 },
		{ 281, 31, 31 },
		{ 281, 32, 31 },
		{ 281, 33, 31 },
		{ 281, 34, 31 },
		{ 281, 35, 31 },
		{ 281, 36, 31 },
		{ 281, 37, 31 },
		{ 281, 38, 31 },
		{ 281, 39, 31 },
		{ 281, 40, 31 },
		{ 281, 41, 31 },
		{ 281, 42, 31 },
		{ 281, 43, 31 },
		{ 281, 44, 31 },
		{ 281, 45, 31 },
		{ 281, 46, 31 },
		{ 281, 47, 31 },
		{ 281, 48, 31 },
		{ 281, 49, 31 },
		{ 281, 50, 31 },
		{ 281, 51, 31 },
		{ 281, 52, 31 },
		{ 281, 53, 31 },
		{ 281, 54, 31 },
		{ 281, 55, 31 },
		{ 281, 56, 31 },
		{ 281, 57, 31 },
		{ 281, 58, 31 },
		{ 281, 59, 31 },
		{ 281, 60, 31 },
		{ 281, 61, 31 },
		{ 281, 62, 31 },
		{ 281, 63, 31 },
		{ 281, 64, 31 },
		{ 281, 65, 31 },
		{ 281, 66, 31 },
		{ 281, 67, 31 },
		{ 281, 68, 31 },
		{ 281, 69, 31 },
		{ 281, 70, 31 },
		{ 281, 71, 31 },
		{ 281, 72, 31 },
		{ 281, 73, 31 },
		{ 281, 74, 31 },
		{ 281, 75, 31 },
		{ 281, 76, 31 },
		{ 281, 77, 31 },
		{ 281, 78, 31 },
		{ 281, 79, 31 },
		{ 281, 80, 31 },
		{ 281, 81, 31 },
		{ 281, 82, 31 },
		{ 281, 83, 31 },
		{ 281, 84, 31 },
		{ 281, 85, 31 },
		{ 281, 86, 31 },
		{ 281, 87, 31 },
		{ 281, 88, 31 },
		{ 281, 89, 31 },
		{ 281, 90, 31 },
		{ 281, 91, 31 },
		{ 281, 92, 31 },
		{ 281, 93, 31 },
		{ 281, 94, 31 },
		{ 281, 95, 31 },
		{ 281, 96, 31 },
		{ 281, 97, 31 },
		{ 281, 98, 31 },
		{ 281, 99, 31 },
		{ 281, 100, 31 },
		{ 281, 101, 31 },
		{ 281, 102, 31 },
		{ 281, 103, 31 },
		{ 281, 104, 31 },
		{ 281, 105, 31 },
		{ 281, 106, 31 },
		{ 281, 107, 31 },
		{ 281, 108, 31 },
		{ 281, 109, 31 },
		{ 281, 110, 31 },
		{ 281, 111, 31 },
		{ 281, 112, 31 },
		{ 281, 113, 31 },
		{ 281, 114, 31 },
		{ 281, 115, 31 },
		{ 281, 116, 31 },
		{ 281, 117, 31 },
		{ 281, 118, 31 },
		{ 281, 119, 31 },
		{ 281, 120, 31 },
		{ 281, 121, 31 },
		{ 281, 122, 31 },
		{ 281, 123, 31 },
		{ 281, 124, 31 },
		{ 281, 125, 31 },
		{ 281, 126, 31 },
		{ 281, 127, 31 },
		{ 281, 128, 31 },
		{ 281, 129, 31 },
		{ 281, 130, 31 },
		{ 281, 131, 31 },
		{ 281, 132, 31 },
		{ 281, 133, 31 },
		{ 281, 134, 31 },
		{ 281, 135, 31 },
		{ 281, 136, 31 },
		{ 281, 137, 31 },
		{ 281, 138, 31 },
		{ 281, 139, 31 },
		{ 281, 140, 31 },
		{ 281, 141, 31 },
		{ 281, 142, 31 },
		{ 281, 143, 31 },
		{ 281, 144, 31 },
		{ 281, 145, 31 },
		{ 281, 146, 31 },
		{ 281, 147, 31 },
		{ 281, 148, 31 },
		{ 281, 149, 31 },
		{ 281, 150, 31 },
		{ 281, 151, 31 },
		{ 281, 152, 31 },
		{ 281, 153, 31 },
		{ 281, 154, 31 },
		{ 281, 155, 31 },
		{ 281, 156, 31 },
		{ 281, 157, 31 },
		{ 281, 158, 31 },
		{ 281, 159, 31 },
		{ 281, 160, 31 },
		{ 282, 1, 31 },
		{ 282, 2, 31 },
		{ 282, 3, 31 },
		{ 282, 4, 31 },
		{ 282, 5, 31 },
		{ 282, 6, 31 },
		{ 282, 7, 31 },
		{ 282, 8, 31 },
		{ 282, 9, 31 },
		{ 282, 10, 31 },
		{ 282, 11, 31 },
		{ 282, 12, 31 },
		{ 282, 13, 31 },
		{ 282, 14, 31 },
		{ 282, 15, 31 },
		{ 282, 16, 31 },
		{ 282, 17, 31 },
		{ 282, 18, 31 },
		{ 282, 19, 31 },
		{ 282, 20, 31 },
		{ 282, 21, 31 },
		{ 282, 22, 31 },
		{ 282, 23, 31 },
		{ 282, 24, 31 },
		{ 282, 25, 31 },
		{ 282, 26, 31 },
		{ 282, 27, 31 },
		{ 282, 28, 31 },
		{ 282, 29, 31 },
		{ 282, 30, 31 },
		{ 282, 31, 31 },
		{ 282, 32, 31 },
		{ 282, 33, 31 },
		{ 282, 34, 31 },
		{ 282, 35, 31 },
		{ 282, 36, 31 },
		{ 282, 37, 31 },
		{ 282, 38, 31 },
		{ 282, 39, 31 },
		{ 282, 40, 31 },
		{ 282, 41, 31 },
		{ 282, 42, 31 },
		{ 282, 43, 31 },
		{ 282, 44, 31 },
		{ 282, 45, 31 },
		{ 282, 46, 31 },
		{ 282, 47, 31 },
		{ 282, 48, 31 },
		{ 282, 49, 31 },
		{ 282, 50, 31 },
		{ 282, 51, 31 },
		{ 282, 52, 31 },
		{ 282, 53, 31 },
		{ 282, 54, 31 },
		{ 282, 55, 31 },
		{ 282, 56, 31 },
		{ 282, 57, 31 },
		{ 282, 58, 31 },
		{ 282, 59, 31 },
		{ 282, 60, 31 },
		{ 282, 61, 31 },
		{ 282, 62, 31 },
		{ 282, 63, 31 },
		{ 282, 64, 31 },
		{ 282, 65, 31 },
		{ 282, 66, 31 },
		{ 282, 67, 31 },
		{ 282, 68, 31 },
		{ 282, 69, 31 },
		{ 282, 70, 31 },
		{ 282, 71, 31 },
		{ 282, 72, 31 },
		{ 282, 73, 31 },
		{ 282, 74, 31 },
		{ 282, 75, 31 },
		{ 282, 76, 31 },
		{ 282, 77, 31 },
		{ 282, 78, 31 },
		{ 282, 79, 31 },
		{ 282, 80, 31 },
		{ 282, 81, 31 },
		{ 282, 82, 31 },
		{ 282, 83, 31 },
		{ 282, 84, 31 },
		{ 282, 85, 31 },
		{ 282, 86, 31 },
		{ 282, 87, 31 },
		{ 282, 88, 31 },
		{ 282, 89, 31 },
		{ 282, 90, 31 },
		{ 282, 91, 31 },
		{ 282, 92, 31 },
		{ 282, 93, 31 },
		{ 282, 94, 31 },
		{ 282, 95, 31 },
		{ 282, 96, 31 },
		{ 282, 97, 31 },
		{ 282, 98, 31 },
		{ 282, 99, 31 },
		{ 282, 100, 31 },
		{ 282, 101, 31 },
		{ 282, 102, 31 },
		{ 282, 103, 31 },
		{ 282, 104, 31 },
		{ 282, 105, 31 },
		{ 282, 106, 31 },
		{ 282, 107, 31 },
		{ 282, 108, 31 },
		{ 282, 109, 31 },
		{ 282, 110, 31 },
		{ 282, 111, 31 },
		{ 282, 112, 31 },
		{ 282, 113, 31 },
		{ 282, 114, 31 },
		{ 282, 115, 31 },
		{ 282, 116, 31 },
		{ 282, 117, 31 },
		{ 282, 118, 31 },
		{ 282, 119, 31 },
		{ 282, 120, 31 },
		{ 282, 121, 31 },
		{ 282, 122, 31 },
		{ 282, 123, 31 },
		{ 282, 124, 31 },
		{ 282, 125, 31 },
		{ 282, 126, 31 },
		{ 282, 127, 31 },
		{ 282, 128, 31 },
		{ 282, 129, 31 },
		{ 282, 130, 31 },
		{ 282, 131, 31 },
		{ 282, 132, 31 },
		{ 282, 133, 31 },
		{ 282, 134, 31 },
		{ 282, 135, 31 },
		{ 282, 136, 31 },
		{ 282, 137, 31 },
		{ 282, 138, 31 },
		{ 282, 139, 31 },
		{ 282, 140, 31 },
		{ 282, 141, 31 },
		{ 282, 142, 31 },
		{ 282, 143, 31 },
		{ 282, 144, 31 },
		{ 282, 145, 31 },
		{ 282, 146, 31 },
		{ 282, 147, 31 },
		{ 282, 148, 31 },
		{ 282, 149, 31 },
		{ 282, 150, 31 },
		{ 282, 151, 31 },
		{ 282, 152, 31 },
		{ 282, 153, 31 },
		{ 282, 154, 31 },
		{ 282, 155, 31 },
		{ 282, 156, 31 },
		{ 283, 1, 31 },
		{ 283, 2, 31 },
		{ 283, 3, 31 },
		{ 283, 4, 31 },
		{ 283, 5, 31 },
		{ 283, 6, 31 },
		{ 283, 7, 31 },
		{ 283, 8, 31 },
		{ 283, 9, 31 },
		{ 283, 10, 31 },
		{ 283, 11, 31 },
		{ 283, 12, 31 },
		{ 283, 13, 31 },
		{ 283, 14, 31 },
		{ 283, 15, 31 },
		{ 283, 16, 31 },
		{ 283, 17, 31 },
		{ 283, 18, 31 },
		{ 283, 19, 31 },
		{ 283, 20, 31 },
		{ 283, 21, 31 },
		{ 283, 22, 31 },
		{ 283, 23, 31 },
		{ 283, 24, 31 },
		{ 283, 25, 31 },
		{ 283, 26, 31 },
		{ 283, 27, 31 },
		{ 283, 28, 31 },
		{ 283, 29, 31 },
		{ 283, 30, 31 },
		{ 283, 31, 31 },
		{ 283, 32, 31 },
		{ 283, 33, 31 },
		{ 283, 34, 31 },
		{ 283, 35, 31 },
		{ 283, 36, 31 },
		{ 283, 37, 31 },
		{ 283, 38, 31 },
		{ 283, 39, 31 },
		{ 283, 40, 31 },
		{ 283, 41, 31 },
		{ 283, 42, 31 },
		{ 283, 43, 31 },
		{ 283, 44, 31 },
		{ 283, 45, 31 },
		{ 283, 46, 31 },
		{ 283, 47, 31 },
		{ 283, 48, 31 },
		{ 283, 49, 31 },
		{ 283, 50, 31 },
		{ 283, 51, 31 },
		{ 283, 52, 31 },
		{ 283, 53, 31 },
		{ 283, 54, 31 },
		{ 283, 55, 31 },
		{ 283, 56, 31 },
		{ 283, 57, 31 },
		{ 283, 58, 31 },
		{ 283, 59, 31 },
		{ 283, 60, 31 },
		{ 283, 61, 31 },
		{ 283, 62, 31 },
		{ 283, 63, 31 },
		{ 283, 64, 31 },
		{ 283, 65, 31 },
		{ 283, 66, 31 },
		{ 283, 67, 31 },
		{ 283, 68, 31 },
		{ 283, 69, 31 },
		{ 283, 70, 31 },
		{ 283, 71, 31 },
		{ 283, 72, 31 },
		{ 283, 73, 31 },
		{ 283, 74, 31 },
		{ 283, 75, 31 },
		{ 283, 76, 31 },
		{ 283, 77, 31 },
		{ 283, 78, 31 },
		{ 283, 79, 31 },
		{ 283, 80, 31 },
		{ 283, 81, 31 },
		{ 283, 82, 31 },
		{ 283, 83, 31 },
		{ 283, 84, 31 },
		{ 283, 85, 31 },
		{ 283, 86, 31 },
		{ 283, 87, 31 },
		{ 283, 88, 31 },
		{ 283, 89, 31 },
		{ 283, 90, 31 },
		{ 283, 91, 31 },
		{ 283, 92, 31 },
		{ 283, 93, 31 },
		{ 283, 94, 31 },
		{ 283, 95, 31 },
		{ 283, 96, 31 },
		{ 283, 97, 31 },
		{ 283, 98, 31 },
		{ 283, 99, 31 },
		{ 283, 100, 31 },
		{ 283, 101, 31 },
		{ 283, 102, 31 },
		{ 283, 103, 31 },
		{ 283, 104, 31 },
		{ 283, 105, 31 },
		{ 283, 106, 31 },
		{ 283, 107, 31 },
		{ 283, 108, 31 },
		{ 283, 109, 31 },
		{ 283, 110, 31 },
		{ 283, 111, 31 },
		{ 283, 112, 31 },
		{ 283, 113, 31 },
		{ 283, 114, 31 },
		{ 283, 115, 31 },
		{ 283, 116, 31 },
		{ 283, 117, 31 },
		{ 283, 118, 31 },
		{ 283, 119, 31 },
		{ 283, 120, 31 },
		{ 283, 121, 31 },
		{ 283, 122, 31 },
		{ 283, 123, 31 },
		{ 283, 124, 31 },
		{ 283, 125, 31 },
		{ 283, 126, 31 },
		{ 283, 127, 31 },
		{ 283, 128, 31 },
		{ 283, 129, 31 },
		{ 283, 130, 31 },
		{ 283, 131, 31 },
		{ 283, 132, 31 },
		{ 283, 133, 31 },
		{ 283, 134, 31 },
		{ 283, 135, 31 },
		{ 283, 136, 31 },
		{ 283, 137, 31 },
		{ 283, 138, 31 },
		{ 283, 139, 31 },
		{ 283, 140, 31 },
		{ 283, 141, 31 },
		{ 283, 142, 31 },
		{ 283, 143, 31 },
		{ 283, 144, 31 },
		{ 283, 145, 31 },
		{ 283, 146, 31 },
		{ 283, 147, 31 },
		{ 283, 148, 31 },
		{ 283, 149, 31 },
		{ 283, 150, 31 },
		{ 283, 151, 31 },
		{ 283, 152, 31 },
		{ 284, 1, 31 },
		{ 284, 2, 31 },
		{ 284, 3, 31 },
		{ 284, 4, 31 },
		{ 284, 5, 31 },
		{ 284, 6, 31 },
		{ 284, 7, 31 },
		{ 284, 8, 31 },
		{ 284, 9, 31 },
		{ 284, 10, 31 },
		{ 284, 11, 31 },
		{ 284, 12, 31 },
		{ 284, 13, 31 },
		{ 284, 14, 31 },
		{ 284, 15, 31 },
		{ 284, 16, 31 },
		{ 284, 17, 31 },
		{ 284, 18, 31 },
		{ 284, 19, 31 },
		{ 284, 20, 31 },
		{ 284, 21, 31 },
		{ 284, 22, 31 },
		{ 284, 23, 31 },
		{ 284, 24, 31 },
		{ 284, 25, 31 },
		{ 284, 26, 31 },
		{ 284, 27, 31 },
		{ 284, 28, 31 },
		{ 284, 29, 31 },
		{ 284, 30, 31 },
		{ 284, 31, 31 },
		{ 284, 32, 31 },
		{ 284, 33, 31 },
		{ 284, 34, 31 },
		{ 284, 35, 31 },
		{ 284, 36, 31 },
		{ 284, 37, 31 },
		{ 284, 38, 31 },
		{ 284, 39, 31 },
		{ 284, 40, 31 },
		{ 284, 41, 31 },
		{ 284, 42, 31 },
		{ 284, 43, 31 },
		{ 284, 44, 31 },
		{ 284, 45, 31 },
		{ 284, 46, 31 },
		{ 284, 47, 31 },
		{ 284, 48, 31 },
		{ 284, 49, 31 },
		{ 284, 50, 31 },
		{ 284, 51, 31 },
		{ 284, 52, 31 },
		{ 284, 53, 31 },
		{ 284, 54, 31 },
		{ 284, 55, 31 },
		{ 284, 56, 31 },
		{ 284, 57, 31 },
		{ 284, 58, 31 },
		{ 284, 59, 31 },
		{ 284, 60, 31 },
		{ 284, 61, 31 },
		{ 284, 62, 31 },
		{ 284, 63, 31 },
		{ 284, 64, 31 },
		{ 284, 65, 31 },
		{ 284, 66, 31 },
		{ 284, 67, 31 },
		{ 284, 68, 31 },
		{ 284, 69, 31 },
		{ 284, 70, 31 },
		{ 284, 71, 31 },
		{ 284, 72, 31 },
		{ 284, 73, 31 },
		{ 284, 74, 31 },
		{ 284, 75, 31 },
		{ 284, 76, 31 },
		{ 284, 77, 31 },
		{ 284, 78, 31 },
		{ 284, 79, 31 },
		{ 284, 80, 31 },
		{ 284, 81, 31 },
		{ 284, 82, 31 },
		{ 284, 83, 31 },
		{ 284, 84, 31 },
		{ 284, 85, 31 },
		{ 284, 86, 31 },
		{ 284, 87, 31 },
		{ 284, 88, 31 },
		{ 284, 89, 31 },
		{ 284, 90, 31 },
		{ 284, 91, 31 },
		{ 284, 92, 31 },
		{ 284, 93, 31 },
		{ 284, 94, 31 },
		{ 284, 95, 31 },
		{ 284, 96, 31 },
		{ 284, 97, 31 },
		{ 284, 98, 31 },
		{ 284, 99, 31 },
		{ 284, 100, 31 },
		{ 284, 101, 31 },
		{ 284, 102, 31 },
		{ 284, 103, 31 },
		{ 284, 104, 31 },
		{ 284, 105, 31 },
		{ 284, 106, 31 },
		{ 284, 107, 31 },
		{ 284, 108, 31 },
		{ 284, 109, 31 },
		{ 284, 110, 31 },
		{ 284, 111, 31 },
		{ 284, 112, 31 },
		{ 284, 113, 31 },
		{ 284, 114, 31 },
		{ 284, 115, 31 },
		{ 284, 116, 31 },
		{ 284, 117, 31 },
		{ 284, 118, 31 },
		{ 284, 119, 31 },
		{ 284, 120, 31 },
		{ 284, 121, 31 },
		{ 284, 122, 31 },
		{ 284, 123, 31 },
		{ 284, 124, 31 },
		{ 284, 125, 31 },
		{ 284, 126, 31 },
		{ 284, 127, 31 },
		{ 284, 128, 31 },
		{ 284, 129, 31 },
		{ 284, 130, 31 },
		{ 284, 131, 31 },
		{ 284, 132, 31 },
		{ 284, 133, 31 },
		{ 284, 134, 31 },
		{ 284, 135, 31 },
		{ 284, 136, 31 },
		{ 284, 137, 31 },
		{ 284, 138, 31 },
		{ 284, 139, 31 },
		{ 284, 140, 31 },
		{ 284, 141, 31 },
		{ 284, 142, 31 },
		{ 284, 143, 31 },
		{ 284, 144, 31 },
		{ 284, 145, 31 },
		{ 284, 146, 31 },
		{ 284, 147, 31 },
		{ 284, 148, 31 },
		{ 285, 1, 31 },
		{ 285, 2, 31 },
		{ 285, 3, 31 },
		{ 285, 4, 31 },
		{ 285, 5, 31 },
		{ 285, 6, 31 },
		{ 285, 7, 31 },
		{ 285, 8, 31 },
		{ 285, 9, 31 },
		{ 285, 10, 31 },
		{ 285, 11, 31 },
		{ 285, 12, 31 },
		{ 285, 13, 31 },
		{ 285, 14, 31 },
		{ 285, 15, 31 },
		{ 285, 16, 31 },
		{ 285, 17, 31 },
		{ 285, 18, 31 },
		{ 285, 19, 31 },
		{ 285, 20, 31 },
		{ 285, 21, 31 },
		{ 285, 22, 31 },
		{ 285, 23, 31 },
		{ 285, 24, 31 },
		{ 285, 25, 31 },
		{ 285, 26, 31 },
		{ 285, 27, 31 },
		{ 285, 28, 31 },
		{ 285, 29, 31 },
		{ 285, 30, 31 },
		{ 285, 31, 31 },
		{ 285, 32, 31 },
		{ 285, 33, 31 },
		{ 285, 34, 31 },
		{ 285, 35, 31 },
		{ 285, 36, 31 },
		{ 285, 37, 31 },
		{ 285, 38, 31 },
		{ 285, 39, 31 },
		{ 285, 40, 31 },
		{ 285, 41, 31 },
		{ 285, 42, 31 },
		{ 285, 43, 31 },
		{ 285, 44, 31 },
		{ 285, 45, 31 },
		{ 285, 46, 31 },
		{ 285, 47, 31 },
		{ 285, 48, 31 },
		{ 285, 49, 31 },
		{ 285, 50, 31 },
		{ 285, 51, 31 },
		{ 285, 52, 31 },
		{ 285, 53, 31 },
		{ 285, 54, 31 },
		{ 285, 55, 31 },
		{ 285, 56, 31 },
		{ 285, 57, 31 },
		{ 285, 58, 31 },
		{ 285, 59, 31 },
		{ 285, 60, 31 },
		{ 285, 61, 31 },
		{ 285, 62, 31 },
		{ 285, 63, 31 },
		{ 285, 64, 31 },
		{ 285, 65, 31 },
		{ 285, 66, 31 },
		{ 285, 67, 31 },
		{ 285, 68, 31 },
		{ 285, 69, 31 },
		{ 285, 70, 31 },
		{ 285, 71, 31 },
		{ 285, 72, 31 },
		{ 285, 73, 31 },
		{ 285, 74, 31 },
		{ 285, 75, 31 },
		{ 285, 76, 31 },
		{ 285, 77, 31 },
		{ 285, 78, 31 },
		{ 285, 79, 31 },
		{ 285, 80, 31 },
		{ 285, 81, 31 },
		{ 285, 82, 31 },
		{ 285, 83, 31 },
		{ 285, 84, 31 },
		{ 285, 85, 31 },
		{ 285, 86, 31 },
		{ 285, 87, 31 },
		{ 285, 88, 31 },
		{ 285, 89, 31 },
		{ 285, 90, 31 },
		{ 285, 91, 31 },
		{ 285, 92, 31 },
		{ 285, 93, 31 },
		{ 285, 94, 31 },
		{ 285, 95, 31 },
		{ 285, 96, 31 },
		{ 285, 97, 31 },
		{ 285, 98, 31 },
		{ 285, 99, 31 },
		{ 285, 100, 31 },
		{ 285, 101, 31 },
		{ 285, 102, 31 },
		{ 285, 103, 31 },
		{ 285, 104, 31 },
		{ 285, 105, 31 },
		{ 285, 106, 31 },
		{ 285, 107, 31 },
		{ 285, 108, 31 },
		{ 285, 109, 31 },
		{ 285, 110, 31 },
		{ 285, 111, 31 },
		{ 285, 112, 31 },
		{ 285, 113, 31 },
		{ 285, 114, 31 },
		{ 285, 115, 31 },
		{ 285, 116, 31 },
		{ 285, 117, 31 },
		{ 285, 118, 31 },
		{ 285, 119, 31 },
		{ 285, 120, 31 },
		{ 285, 121, 31 },
		{ 285, 122, 31 },
		{ 285, 123, 31 },
		{ 285, 124, 31 },
		{ 285, 125, 31 },
		{ 285, 126, 31 },
		{ 285, 127, 31 },
		{ 285, 128, 31 },
		{ 285, 129, 31 },
		{ 285, 130, 31 },
		{ 285, 131, 31 },
		{ 285, 132, 31 },
		{ 285, 133, 31 },
		{ 285, 134, 31 },
		{ 285, 135, 31 },
		{ 285, 136, 31 },
		{ 285, 137, 31 },
		{ 285, 138, 31 },
		{ 285, 139, 31 },
		{ 285, 140, 31 },
		{ 285, 141, 31 },
		{ 285, 142, 31 },
		{ 285, 143, 31 },
		{ 285, 144, 31 },
		{ 286, 1, 31 },
		{ 286, 2, 31 },
		{ 286, 3, 31 },
		{ 286, 4, 31 },
		{ 286, 5, 31 },
		{ 286, 6, 31 },
		{ 286, 7, 31 },
		{ 286, 8, 31 },
		{ 286, 9, 31 },
		{ 286, 10, 31 },
		{ 286, 11, 31 },
		{ 286, 12, 31 },
		{ 286, 13, 31 },
		{ 286, 14, 31 },
		{ 286, 15, 31 },
		{ 286, 16, 31 },
		{ 286, 17, 31 },
		{ 286, 18, 31 },
		{ 286, 19, 31 },
		{ 286, 20, 31 },
		{ 286, 21, 31 },
		{ 286, 22, 31 },
		{ 286, 23, 31 },
		{ 286, 24, 31 },
		{ 286, 25, 31 },
		{ 286, 26, 31 },
		{ 286, 27, 31 },
		{ 286, 28, 31 },
		{ 286, 29, 31 },
		{ 286, 30, 31 },
		{ 286, 31, 31 },
		{ 286, 32, 31 },
		{ 286, 33, 31 },
		{ 286, 34, 31 },
		{ 286, 35, 31 },
		{ 286, 36, 31 },
		{ 286, 37, 31 },
		{ 286, 38, 31 },
		{ 286, 39, 31 },
		{ 286, 40, 31 },
		{ 286, 41, 31 },
		{ 286, 42, 31 },
		{ 286, 43, 31 },
		{ 286, 44, 31 },
		{ 286, 45, 31 },
		{ 286, 46, 31 },
		{ 286, 47, 31 },
		{ 286, 48, 31 },
		{ 286, 49, 31 },
		{ 286, 50, 31 },
		{ 286, 51, 31 },
		{ 286, 52, 31 },
		{ 286, 53, 31 },
		{ 286, 54, 31 },
		{ 286, 55, 31 },
		{ 286, 56, 31 },
		{ 286, 57, 31 },
		{ 286, 58, 31 },
		{ 286, 59, 31 },
		{ 286, 60, 31 },
		{ 286, 61, 31 },
		{ 286, 62, 31 },
		{ 286, 63, 31 },
		{ 286, 64, 31 },
		{ 286, 65, 31 },
		{ 286, 66, 31 },
		{ 286, 67, 31 },
		{ 286, 68, 31 },
		{ 286, 69, 31 },
		{ 286, 70, 31 },
		{ 286, 71, 31 },
		{ 286, 72, 31 },
		{ 286, 73, 31 },
		{ 286, 74, 31 },
		{ 286, 75, 31 },
		{ 286, 76, 31 },
		{ 286, 77, 31 },
		{ 286, 78, 31 },
		{ 286, 79, 31 },
		{ 286, 80, 31 },
		{ 286, 81, 31 },
		{ 286, 82, 31 },
		{ 286, 83, 31 },
		{ 286, 84, 31 },
		{ 286, 85, 31 },
		{ 286, 86, 31 },
		{ 286, 87, 31 },
		{ 286, 88, 31 },
		{ 286, 89, 31 },
		{ 286, 90, 31 },
		{ 286, 91, 31 },
		{ 286, 92, 31 },
		{ 286, 93, 31 },
		{ 286, 94, 31 },
		{ 286, 95, 31 },
		{ 286, 96, 31 },
		{ 286, 97, 31 },
		{ 286, 98, 31 },
		{ 286, 99, 31 },
		{ 286, 100, 31 },
		{ 286, 101, 31 },
		{ 286, 102, 31 },
		{ 286, 103, 31 },
		{ 286, 104, 31 },
		{ 286, 105, 31 },
		{ 286, 106, 31 },
		{ 286, 107, 31 },
		{ 286, 108, 31 },
		{ 286, 109, 31 },
		{ 286, 110, 31 },
		{ 286, 111, 31 },
		{ 286, 112, 31 },
		{ 286, 113, 31 },
		{ 286, 114, 31 },
		{ 286, 115, 31 },
		{ 286, 116, 31 },
		{ 286, 117, 31 },
		{ 286, 118, 31 },
		{ 286, 119, 31 },
		{ 286, 120, 31 },
		{ 286, 121, 31 },
		{ 286, 122, 31 },
		{ 286, 123, 31 },
		{ 286, 124, 31 },
		{ 286, 125, 31 },
		{ 286, 126, 31 },
		{ 286, 127, 31 },
		{ 286, 128, 31 },
		{ 286, 129, 31 },
		{ 286, 130, 31 },
		{ 286, 131, 31 },
		{ 286, 132, 31 },
		{ 286, 133, 31 },
		{ 286, 134, 31 },
		{ 286, 135, 31 },
		{ 286, 136, 31 },
		{ 286, 137, 31 },
		{ 286, 138, 31 },
		{ 286, 139, 31 },
		{ 286, 140, 31 },
		{ 287, 1, 31 },
		{ 287, 2, 31 },
		{ 287, 3, 31 },
		{ 287, 4, 31 },
		{ 287, 5, 31 },
		{ 287, 6, 31 },
		{ 287, 7, 31 },
		{ 287, 8, 31 },
		{ 287, 9, 31 },
		{ 287, 10, 31 },
		{ 287, 11, 31 },
		{ 287, 12, 31 },
		{ 287, 13, 31 },
		{ 287, 14, 31 },
		{ 287, 15, 31 },
		{ 287, 16, 31 },
		{ 287, 17, 31 },
		{ 287, 18, 31 },
		{ 287, 19, 31 },
		{ 287, 20, 31 },
		{ 287, 21, 31 },
		{ 287, 22, 31 },
		{ 287, 23, 31 },
		{ 287, 24, 31 },
		{ 287, 25, 31 },
		{ 287, 26, 31 },
		{ 287, 27, 31 },
		{ 287, 28, 31 },
		{ 287, 29, 31 },
		{ 287, 30, 31 },
		{ 287, 31, 31 },
		{ 287, 32, 31 },
		{ 287, 33, 31 },
		{ 287, 34, 31 },
		{ 287, 35, 31 },
		{ 287, 36, 31 },
		{ 287, 37, 31 },
		{ 287, 38, 31 },
		{ 287, 39, 31 },
		{ 287, 40, 31 },
		{ 287, 41, 31 },
		{ 287, 42, 31 },
		{ 287, 43, 31 },
		{ 287, 44, 31 },
		{ 287, 45, 31 },
		{ 287, 46, 31 },
		{ 287, 47, 31 },
		{ 287, 48, 31 },
		{ 287, 49, 31 },
		{ 287, 50, 31 },
		{ 287, 51, 31 },
		{ 287, 52, 31 },
		{ 287, 53, 31 },
		{ 287, 54, 31 },
		{ 287, 55, 31 },
		{ 287, 56, 31 },
		{ 287, 57, 31 },
		{ 287, 58, 31 },
		{ 287, 59, 31 },
		{ 287, 60, 31 },
		{ 287, 61, 31 },
		{ 287, 62, 31 },
		{ 287, 63, 31 },
		{ 287, 64, 31 },
		{ 287, 65, 31 },
		{ 287, 66, 31 },
		{ 287, 67, 31 },
		{ 287, 68, 31 },
		{ 287, 69, 31 },
		{ 287, 70, 31 },
		{ 287, 71, 31 },
		{ 287, 72, 31 },
		{ 287, 73, 31 },
		{ 287, 74, 31 },
		{ 287, 75, 31 },
		{ 287, 76, 31 },
		{ 287, 77, 31 },
		{ 287, 78, 31 },
		{ 287, 79, 31 },
		{ 287, 80, 31 },
		{ 287, 81, 31 },
		{ 287, 82, 31 },
		{ 287, 83, 31 },
		{ 287, 84, 31 },
		{ 287, 85, 31 },
		{ 287, 86, 31 },
		{ 287, 87, 31 },
		{ 287, 88, 31 },
		{ 287, 89, 31 },
		{ 287, 90, 31 },
		{ 287, 91, 31 },
		{ 287, 92, 31 },
		{ 287, 93, 31 },
		{ 287, 94, 31 },
		{ 287, 95, 31 },
		{ 287, 96, 31 },
		{ 287, 97, 31 },
		{ 287, 98, 31 },
		{ 287, 99, 31 },
		{ 287, 100, 31 },
		{ 287, 101, 31 },
		{ 287, 102, 31 },
		{ 287, 103, 31 },
		{ 287, 104, 31 },
		{ 287, 105, 31 },
		{ 287, 106, 31 },
		{ 287, 107, 31 },
		{ 287, 108, 31 },
		{ 287, 109, 31 },
		{ 287, 110, 31 },
		{ 287, 111, 31 },
		{ 287, 112, 31 },
		{ 287, 113, 31 },
		{ 287, 114, 31 },
		{ 287, 115, 31 },
		{ 287, 116, 31 },
		{ 287, 117, 31 },
		{ 287, 118, 31 },
		{ 287, 119, 31 },
		{ 287, 120, 31 },
		{ 287, 121, 31 },
		{ 287, 122, 31 },
		{ 287, 123, 31 },
		{ 287, 124, 31 },
		{ 287, 125, 31 },
		{ 287, 126, 31 },
		{ 287, 127, 31 },
		{ 287, 128, 31 },
		{ 287, 129, 31 },
		{ 287, 130, 31 },
		{ 287, 131, 31 },
		{ 287, 132, 31 },
		{ 287, 133, 31 },
		{ 287, 134, 31 },
		{ 287, 135, 31 },
		{ 287, 136, 31 },
		{ 288, 1, 31 },
		{ 288, 2, 31 },
		{ 288, 3, 31 },
		{ 288, 4, 31 },
		{ 288, 5, 31 },
		{ 288, 6, 31 },
		{ 288, 7, 31 },
		{ 288, 8, 31 },
		{ 288, 9, 31 },
		{ 288, 10, 31 },
		{ 288, 11, 31 },
		{ 288, 12, 31 },
		{ 288, 13, 31 },
		{ 288, 14, 31 },
		{ 288, 15, 31 },
		{ 288, 16, 31 },
		{ 288, 17, 31 },
		{ 288, 18, 31 },
		{ 288, 19, 31 },
		{ 288, 20, 31 },
		{ 288, 21, 31 },
		{ 288, 22, 31 },
		{ 288, 23, 31 },
		{ 288, 24, 31 },
		{ 288, 25, 31 },
		{ 288, 26, 31 },
		{ 288, 27, 31 },
		{ 288, 28, 31 },
		{ 288, 29, 31 },
		{ 288, 30, 31 },
		{ 288, 31, 31 },
		{ 288, 32, 31 },
		{ 288, 33, 31 },
		{ 288, 34, 31 },
		{ 288, 35, 31 },
		{ 288, 36, 31 },
		{ 288, 37, 31 },
		{ 288, 38, 31 },
		{ 288, 39, 31 },
		{ 288, 40, 31 },
		{ 288, 41, 31 },
		{ 288, 42, 31 },
		{ 288, 43, 31 },
		{ 288, 44, 31 },
		{ 288, 45, 31 },
		{ 288, 46, 31 },
		{ 288, 47, 31 },
		{ 288, 48, 31 },
		{ 288, 49, 31 },
		{ 288, 50, 31 },
		{ 288, 51, 31 },
		{ 288, 52, 31 },
		{ 288, 53, 31 },
		{ 288, 54, 31 },
		{ 288, 55, 31 },
		{ 288, 56, 31 },
		{ 288, 57, 31 },
		{ 288, 58, 31 },
		{ 288, 59, 31 },
		{ 288, 60, 31 },
		{ 288, 61, 31 },
		{ 288, 62, 31 },
		{ 288, 63, 31 },
		{ 288, 64, 31 },
		{ 288, 65, 31 },
		{ 288, 66, 31 },
		{ 288, 67, 31 },
		{ 288, 68, 31 },
		{ 288, 69, 31 },
		{ 288, 70, 31 },
		{ 288, 71, 31 },
		{ 288, 72, 31 },
		{ 288, 73, 31 },
		{ 288, 74, 31 },
		{ 288, 75, 31 },
		{ 288, 76, 31 },
		{ 288, 77, 31 },
		{ 288, 78, 31 },
		{ 288, 79, 31 },
		{ 288, 80, 31 },
		{ 288, 81, 31 },
		{ 288, 82, 31 },
		{ 288, 83, 31 },
		{ 288, 84, 31 },
		{ 288, 85, 31 },
		{ 288, 86, 31 },
		{ 288, 87, 31 },
		{ 288, 88, 31 },
		{ 288, 89, 31 },
		{ 288, 90, 31 },
		{ 288, 91, 31 },
		{ 288, 92, 31 },
		{ 288, 93, 31 },
		{ 288, 94, 31 },
		{ 288, 95, 31 },
		{ 288, 96, 31 },
		{ 288, 97, 31 },
		{ 288, 98, 31 },
		{ 288, 99, 31 },
		{ 288, 100, 31 },
		{ 288, 101, 31 },
		{ 288, 102, 31 },
		{ 288, 103, 31 },
		{ 288, 104, 31 },
		{ 288, 105, 31 },
		{ 288, 106, 31 },
		{ 288, 107, 31 },
		{ 288, 108, 31 },
		{ 288, 109, 31 },
		{ 288, 110, 31 },
		{ 288, 111, 31 },
		{ 288, 112, 31 },
		{ 288, 113, 31 },
		{ 288, 114, 31 },
		{ 288, 115, 31 },
		{ 288, 116, 31 },
		{ 288, 117, 31 },
		{ 288, 118, 31 },
		{ 288, 119, 31 },
		{ 288, 120, 31 },
		{ 288, 121, 31 },
		{ 288, 122, 31 },
		{ 288, 123, 31 },
		{ 288, 124, 31 },
		{ 288, 125, 31 },
		{ 288, 126, 31 },
		{ 288, 127, 31 },
		{ 288, 128, 31 },
		{ 288, 129, 31 },
		{ 288, 130, 31 },
		{ 288, 131, 31 },
		{ 288, 132, 31 },
		{ 289, 1, 31 },
		{ 289, 2, 31 },
		{ 289, 3, 31 },
		{ 289, 4, 31 },
		{ 289, 5, 31 },
		{ 289, 6, 31 },
		{ 289, 7, 31 },
		{ 289, 8, 31 },
		{ 289, 9, 31 },
		{ 289, 10, 31 },
		{ 289, 11, 31 },
		{ 289, 12, 31 },
		{ 289, 13, 31 },
		{ 289, 14, 31 },
		{ 289, 15, 31 },
		{ 289, 16, 31 },
		{ 289, 17, 31 },
		{ 289, 18, 31 },
		{ 289, 19, 31 },
		{ 289, 20, 31 },
		{ 289, 21, 31 },
		{ 289, 22, 31 },
		{ 289, 23, 31 },
		{ 289, 24, 31 },
		{ 289, 25, 31 },
		{ 289, 26, 31 },
		{ 289, 27, 31 },
		{ 289, 28, 31 },
		{ 289, 29, 31 },
		{ 289, 30, 31 },
		{ 289, 31, 31 },
		{ 289, 32, 31 },
		{ 289, 33, 31 },
		{ 289, 34, 31 },
		{ 289, 35, 31 },
		{ 289, 36, 31 },
		{ 289, 37, 31 },
		{ 289, 38, 31 },
		{ 289, 39, 31 },
		{ 289, 40, 31 },
		{ 289, 41, 31 },
		{ 289, 42, 31 },
		{ 289, 43, 31 },
		{ 289, 44, 31 },
		{ 289, 45, 31 },
		{ 289, 46, 31 },
		{ 289, 47, 31 },
		{ 289, 48, 31 },
		{ 289, 49, 31 },
		{ 289, 50, 31 },
		{ 289, 51, 31 },
		{ 289, 52, 31 },
		{ 289, 53, 31 },
		{ 289, 54, 31 },
		{ 289, 55, 31 },
		{ 289, 56, 31 },
		{ 289, 57, 31 },
		{ 289, 58, 31 },
		{ 289, 59, 31 },
		{ 289, 60, 31 },
		{ 289, 61, 31 },
		{ 289, 62, 31 },
		{ 289, 63, 31 },
		{ 289, 64, 31 },
		{ 289, 65, 31 },
		{ 289, 66, 31 },
		{ 289, 67, 31 },
		{ 289, 68, 31 },
		{ 289, 69, 31 },
		{ 289, 70, 31 },
		{ 289, 71, 31 },
		{ 289, 72, 31 },
		{ 289, 73, 31 },
		{ 289, 74, 31 },
		{ 289, 75, 31 },
		{ 289, 76, 31 },
		{ 289, 77, 31 },
		{ 289, 78, 31 },
		{ 289, 79, 31 },
		{ 289, 80, 31 },
		{ 289, 81, 31 },
		{ 289, 82, 31 },
		{ 289, 83, 31 },
		{ 289, 84, 31 },
		{ 289, 85, 31 },
		{ 289, 86, 31 },
		{ 289, 87, 31 },
		{ 289, 88, 31 },
		{ 289, 89, 31 },
		{ 289, 90, 31 },
		{ 289, 91, 31 },
		{ 289, 92, 31 },
		{ 289, 93, 31 },
		{ 289, 94, 31 },
		{ 289, 95, 31 },
		{ 289, 96, 31 },
		{ 289, 97, 31 },
		{ 289, 98, 31 },
		{ 289, 99, 31 },
		{ 289, 100, 31 },
		{ 289, 101, 31 },
		{ 289, 102, 31 },
		{ 289, 103, 31 },
		{ 289, 104, 31 },
		{ 289, 105, 31 },
		{ 289, 106, 31 },
		{ 289, 107, 31 },
		{ 289, 108, 31 },
		{ 289, 109, 31 },
		{ 289, 110, 31 },
		{ 289, 111, 31 },
		{ 289, 112, 31 },
		{ 289, 113, 31 },
		{ 289, 114, 31 },
		{ 289, 115, 31 },
		{ 289, 116, 31 },
		{ 289, 117, 31 },
		{ 289, 118, 31 },
		{ 289, 119, 31 },
		{ 289, 120, 31 },
		{ 289, 121, 31 },
		{ 289, 122, 31 },
		{ 289, 123, 31 },
		{ 289, 124, 31 },
		{ 289, 125, 31 },
		{ 289, 126, 31 },
		{ 289, 127, 31 },
		{ 289, 128, 31 },
		{ 290, 1, 31 },
		{ 290, 2, 31 },
		{ 290, 3, 31 },
		{ 290, 4, 31 },
		{ 290, 5, 31 },
		{ 290, 6, 31 },
		{ 290, 7, 31 },
		{ 290, 8, 31 },
		{ 290, 9, 31 },
		{ 290, 10, 31 },
		{ 290, 11, 31 },
		{ 290, 12, 31 },
		{ 290, 13, 31 },
		{ 290, 14, 31 },
		{ 290, 15, 31 },
		{ 290, 16, 31 },
		{ 290, 17, 31 },
		{ 290, 18, 31 },
		{ 290, 19, 31 },
		{ 290, 20, 31 },
		{ 290, 21, 31 },
		{ 290, 22, 31 },
		{ 290, 23, 31 },
		{ 290, 24, 31 },
		{ 290, 25, 31 },
		{ 290, 26, 31 },
		{ 290, 27, 31 },
		{ 290, 28, 31 },
		{ 290, 29, 31 },
		{ 290, 30, 31 },
		{ 290, 31, 31 },
		{ 290, 32, 31 },
		{ 290, 33, 31 },
		{ 290, 34, 31 },
		{ 290, 35, 31 },
		{ 290, 36, 31 },
		{ 290, 37, 31 },
		{ 290, 38, 31 },
		{ 290, 39, 31 },
		{ 290, 40, 31 },
		{ 290, 41, 31 },
		{ 290, 42, 31 },
		{ 290, 43, 31 },
		{ 290, 44, 31 },
		{ 290, 45, 31 },
		{ 290, 46, 31 },
		{ 290, 47, 31 },
		{ 290, 48, 31 },
		{ 290, 49, 31 },
		{ 290, 50, 31 },
		{ 290, 51, 31 },
		{ 290, 52, 31 },
		{ 290, 53, 31 },
		{ 290, 54, 31 },
		{ 290, 55, 31 },
		{ 290, 56, 31 },
		{ 290, 57, 31 },
		{ 290, 58, 31 },
		{ 290, 59, 31 },
		{ 290, 60, 31 },
		{ 290, 61, 31 },
		{ 290, 62, 31 },
		{ 290, 63, 31 },
		{ 290, 64, 31 },
		{ 290, 65, 31 },
		{ 290, 66, 31 },
		{ 290, 67, 31 },
		{ 290, 68, 31 },
		{ 290, 69, 31 },
		{ 290, 70, 31 },
		{ 290, 71, 31 },
		{ 290, 72, 31 },
		{ 290, 73, 31 },
		{ 290, 74, 31 },
		{ 290, 75, 31 },
		{ 290, 76, 31 },
		{ 290, 77, 31 },
		{ 290, 78, 31 },
		{ 290, 79, 31 },
		{ 290, 80, 31 },
		{ 290, 81, 31 },
		{ 290, 82, 31 },
		{ 290, 83, 31 },
		{ 290, 84, 31 },
		{ 290, 85, 31 },
		{ 290, 86, 31 },
		{ 290, 87, 31 },
		{ 290, 88, 31 },
		{ 290, 89, 31 },
		{ 290, 90, 31 },
		{ 290, 91, 31 },
		{ 290, 92, 31 },
		{ 290, 93, 31 },
		{ 290, 94, 31 },
		{ 290, 95, 31 },
		{ 290, 96, 31 },
		{ 290, 97, 31 },
		{ 290, 98, 31 },
		{ 290, 99, 31 },
		{ 290, 100, 31 },
		{ 290, 101, 31 },
		{ 290, 102, 31 },
		{ 290, 103, 31 },
		{ 290, 104, 31 },
		{ 290, 105, 31 },
		{ 290, 106, 31 },
		{ 290, 107, 31 },
		{ 290, 108, 31 },
		{ 290, 109, 31 },
		{ 290, 110, 31 },
		{ 290, 111, 31 },
		{ 290, 112, 31 },
		{ 290, 113, 31 },
		{ 290, 114, 31 },
		{ 290, 115, 31 },
		{ 290, 116, 31 },
		{ 290, 117, 31 },
		{ 290, 118, 31 },
		{ 290, 119, 31 },
		{ 290, 120, 31 },
		{ 290, 121, 31 },
		{ 290, 122, 31 },
		{ 290, 123, 31 },
		{ 290, 124, 31 },
		{ 291, 1, 31 },
		{ 291, 2, 31 },
		{ 291, 3, 31 },
		{ 291, 4, 31 },
		{ 291, 5, 31 },
		{ 291, 6, 31 },
		{ 291, 7, 31 },
		{ 291, 8, 31 },
		{ 291, 9, 31 },
		{ 291, 10, 31 },
		{ 291, 11, 31 },
		{ 291, 12, 31 },
		{ 291, 13, 31 },
		{ 291, 14, 31 },
		{ 291, 15, 31 },
		{ 291, 16, 31 },
		{ 291, 17, 31 },
		{ 291, 18, 31 },
		{ 291, 19, 31 },
		{ 291, 20, 31 },
		{ 291, 21, 31 },
		{ 291, 22, 31 },
		{ 291, 23, 31 },
		{ 291, 24, 31 },
		{ 291, 25, 31 },
		{ 291, 26, 31 },
		{ 291, 27, 31 },
		{ 291, 28, 31 },
		{ 291, 29, 31 },
		{ 291, 30, 31 },
		{ 291, 31, 31 },
		{ 291, 32, 31 },
		{ 291, 33, 31 },
		{ 291, 34, 31 },
		{ 291, 35, 31 },
		{ 291, 36, 31 },
		{ 291, 37, 31 },
		{ 291, 38, 31 },
		{ 291, 39, 31 },
		{ 291, 40, 31 },
		{ 291, 41, 31 },
		{ 291, 42, 31 },
		{ 291, 43, 31 },
		{ 291, 44, 31 },
		{ 291, 45, 31 },
		{ 291, 46, 31 },
		{ 291, 47, 31 },
		{ 291, 48, 31 },
		{ 291, 49, 31 },
		{ 291, 50, 31 },
		{ 291, 51, 31 },
		{ 291, 52, 31 },
		{ 291, 53, 31 },
		{ 291, 54, 31 },
		{ 291, 55, 31 },
		{ 291, 56, 31 },
		{ 291, 57, 31 },
		{ 291, 58, 31 },
		{ 291, 59, 31 },
		{ 291, 60, 31 },
		{ 291, 61, 31 },
		{ 291, 62, 31 },
		{ 291, 63, 31 },
		{ 291, 64, 31 },
		{ 291, 65, 31 },
		{ 291, 66, 31 },
		{ 291, 67, 31 },
		{ 291, 68, 31 },
		{ 291, 69, 31 },
		{ 291, 70, 31 },
		{ 291, 71, 31 },
		{ 291, 72, 31 },
		{ 291, 73, 31 },
		{ 291, 74, 31 },
		{ 291, 75, 31 },
		{ 291, 76, 31 },
		{ 291, 77, 31 },
		{ 291, 78, 31 },
		{ 291, 79, 31 },
		{ 291, 80, 31 },
		{ 291, 81, 31 },
		{ 291, 82, 31 },
		{ 291, 83, 31 },
		{ 291, 84, 31 },
		{ 291, 85, 31 },
		{ 291, 86, 31 },
		{ 291, 87, 31 },
		{ 291, 88, 31 },
		{ 291, 89, 31 },
		{ 291, 90, 31 },
		{ 291, 91, 31 },
		{ 291, 92, 31 },
		{ 291, 93, 31 },
		{ 291, 94, 31 },
		{ 291, 95, 31 },
		{ 291, 96, 31 },
		{ 291, 97, 31 },
		{ 291, 98, 31 },
		{ 291, 99, 31 },
		{ 291, 100, 31 },
		{ 291, 101, 31 },
		{ 291, 102, 31 },
		{ 291, 103, 31 },
		{ 291, 104, 31 },
		{ 291, 105, 31 },
		{ 291, 106, 31 },
		{ 291, 107, 31 },
		{ 291, 108, 31 },
		{ 291, 109, 31 },
		{ 291, 110, 31 },
		{ 291, 111, 31 },
		{ 291, 112, 31 },
		{ 291, 113, 31 },
		{ 291, 114, 31 },
		{ 291, 115, 31 },
		{ 291, 116, 31 },
		{ 291, 117, 31 },
		{ 291, 118, 31 },
		{ 291, 119, 31 },
		{ 291, 120, 31 },
		{ 292, 1, 31 },
		{ 292, 2, 31 },
		{ 292, 3, 31 },
		{ 292, 4, 31 },
		{ 292, 5, 31 },
		{ 292, 6, 31 },
		{ 292, 7, 31 },
		{ 292, 8, 31 },
		{ 292, 9, 31 },
		{ 292, 10, 31 },
		{ 292, 11, 31 },
		{ 292, 12, 31 },
		{ 292, 13, 31 },
		{ 292, 14, 31 },
		{ 292, 15, 31 },
		{ 292, 16, 31 },
		{ 292, 17, 31 },
		{ 292, 18, 31 },
		{ 292, 19, 31 },
		{ 292, 20, 31 },
		{ 292, 21, 31 },
		{ 292, 22, 31 },
		{ 292, 23, 31 },
		{ 292, 24, 31 },
		{ 292, 25, 31 },
		{ 292, 26, 31 },
		{ 292, 27, 31 },
		{ 292, 28, 31 },
		{ 292, 29, 31 },
		{ 292, 30, 31 },
		{ 292, 31, 31 },
		{ 292, 32, 31 },
		{ 292, 33, 31 },
		{ 292, 34, 31 },
		{ 292, 35, 31 },
		{ 292, 36, 31 },
		{ 292, 37, 31 },
		{ 292, 38, 31 },
		{ 292, 39, 31 },
		{ 292, 40, 31 },
		{ 292, 41, 31 },
		{ 292, 42, 31 },
		{ 292, 43, 31 },
		{ 292, 44, 31 },
		{ 292, 45, 31 },
		{ 292, 46, 31 },
		{ 292, 47, 31 },
		{ 292, 48, 31 },
		{ 292, 49, 31 },
		{ 292, 50, 31 },
		{ 292, 51, 31 },
		{ 292, 52, 31 },
		{ 292, 53, 31 },
		{ 292, 54, 31 },
		{ 292, 55, 31 },
		{ 292, 56, 31 },
		{ 292, 57, 31 },
		{ 292, 58, 31 },
		{ 292, 59, 31 },
		{ 292, 60, 31 },
		{ 292, 61, 31 },
		{ 292, 62, 31 },
		{ 292, 63, 31 },
		{ 292, 64, 31 },
		{ 292, 65, 31 },
		{ 292, 66, 31 },
		{ 292, 67, 31 },
		{ 292, 68, 31 },
		{ 292, 69, 31 },
		{ 292, 70, 31 },
		{ 292, 71, 31 },
		{ 292, 72, 31 },
		{ 292, 73, 31 },
		{ 292, 74, 31 },
		{ 292, 75, 31 },
		{ 292, 76, 31 },
		{ 292, 77, 31 },
		{ 292, 78, 31 },
		{ 292, 79, 31 },
		{ 292, 80, 31 },
		{ 292, 81, 31 },
		{ 292, 82, 31 },
		{ 292, 83, 31 },
		{ 292, 84, 31 },
		{ 292, 85, 31 },
		{ 292, 86, 31 },
		{ 292, 87, 31 },
		{ 292, 88, 31 },
		{ 292, 89, 31 },
		{ 292, 90, 31 },
		{ 292, 91, 31 },
		{ 292, 92, 31 },
		{ 292, 93, 31 },
		{ 292, 94, 31 },
		{ 292, 95, 31 },
		{ 292, 96, 31 },
		{ 292, 97, 31 },
		{ 292, 98, 31 },
		{ 292, 99, 31 },
		{ 292, 100, 31 },
		{ 292, 101, 31 },
		{ 292, 102, 31 },
		{ 292, 103, 31 },
		{ 292, 104, 31 },
		{ 292, 105, 31 },
		{ 292, 106, 31 },
		{ 292, 107, 31 },
		{ 292, 108, 31 },
		{ 292, 109, 31 },
		{ 292, 110, 31 },
		{ 292, 111, 31 },
		{ 292, 112, 31 },
		{ 292, 113, 31 },
		{ 292, 114, 31 },
		{ 292, 115, 31 },
		{ 292, 116, 31 },
		{ 293, 1, 31 },
		{ 293, 2, 31 },
		{ 293, 3, 31 },
		{ 293, 4, 31 },
		{ 293, 5, 31 },
		{ 293, 6, 31 },
		{ 293, 7, 31 },
		{ 293, 8, 31 },
		{ 293, 9, 31 },
		{ 293, 10, 31 },
		{ 293, 11, 31 },
		{ 293, 12, 31 },
		{ 293, 13, 31 },
		{ 293, 14, 31 },
		{ 293, 15, 31 },
		{ 293, 16, 31 },
		{ 293, 17, 31 },
		{ 293, 18, 31 },
		{ 293, 19, 31 },
		{ 293, 20, 31 },
		{ 293, 21, 31 },
		{ 293, 22, 31 },
		{ 293, 23, 31 },
		{ 293, 24, 31 },
		{ 293, 25, 31 },
		{ 293, 26, 31 },
		{ 293, 27, 31 },
		{ 293, 28, 31 },
		{ 293, 29, 31 },
		{ 293, 30, 31 },
		{ 293, 31, 31 },
		{ 293, 32, 31 },
		{ 293, 33, 31 },
		{ 293, 34, 31 },
		{ 293, 35, 31 },
		{ 293, 36, 31 },
		{ 293, 37, 31 },
		{ 293, 38, 31 },
		{ 293, 39, 31 },
		{ 293, 40, 31 },
		{ 293, 41, 31 },
		{ 293, 42, 31 },
		{ 293, 43, 31 },
		{ 293, 44, 31 },
		{ 293, 45, 31 },
		{ 293, 46, 31 },
		{ 293, 47, 31 },
		{ 293, 48, 31 },
		{ 293, 49, 31 },
		{ 293, 50, 31 },
		{ 293, 51, 31 },
		{ 293, 52, 31 },
		{ 293, 53, 31 },
		{ 293, 54, 31 },
		{ 293, 55, 31 },
		{ 293, 56, 31 },
		{ 293, 57, 31 },
		{ 293, 58, 31 },
		{ 293, 59, 31 },
		{ 293, 60, 31 },
		{ 293, 61, 31 },
		{ 293, 62, 31 },
		{ 293, 63, 31 },
		{ 293, 64, 31 },
		{ 293, 65, 31 },
		{ 293, 66, 31 },
		{ 293, 67, 31 },
		{ 293, 68, 31 },
		{ 293, 69, 31 },
		{ 293, 70, 31 },
		{ 293, 71, 31 },
		{ 293, 72, 31 },
		{ 293, 73, 31 },
		{ 293, 74, 31 },
		{ 293, 75, 31 },
		{ 293, 76, 31 },
		{ 293, 77, 31 },
		{ 293, 78, 31 },
		{ 293, 79, 31 },
		{ 293, 80, 31 },
		{ 293, 81, 31 },
		{ 293, 82, 31 },
		{ 293, 83, 31 },
		{ 293, 84, 31 },
		{ 293, 85, 31 },
		{ 293, 86, 31 },
		{ 293, 87, 31 },
		{ 293, 88, 31 },
		{ 293, 89, 31 },
		{ 293, 90, 31 },
		{ 293, 91, 31 },
		{ 293, 92, 31 },
		{ 293, 93, 31 },
		{ 293, 94, 31 },
		{ 293, 95, 31 },
		{ 293, 96, 31 },
		{ 293, 97, 31 },
		{ 293, 98, 31 },
		{ 293, 99, 31 },
		{ 293, 100, 31 },
		{ 293, 101, 31 },
		{ 293, 102, 31 },
		{ 293, 103, 31 },
		{ 293, 104, 31 },
		{ 293, 105, 31 },
		{ 293, 106, 31 },
		{ 293, 107, 31 },
		{ 293, 108, 31 },
		{ 293, 109, 31 },
		{ 293, 110, 31 },
		{ 293, 111, 31 },
		{ 293, 112, 31 },
		{ 294, 1, 31 },
		{ 294, 2, 31 },
		{ 294, 3, 31 },
		{ 294, 4, 31 },
		{ 294, 5, 31 },
		{ 294, 6, 31 },
		{ 294, 7, 31 },
		{ 294, 8, 31 },
		{ 294, 9, 31 },
		{ 294, 10, 31 },
		{ 294, 11, 31 },
		{ 294, 12, 31 },
		{ 294, 13, 31 },
		{ 294, 14, 31 },
		{ 294, 15, 31 },
		{ 294, 16, 31 },
		{ 294, 17, 31 },
		{ 294, 18, 31 },
		{ 294, 19, 31 },
		{ 294, 20, 31 },
		{ 294, 21, 31 },
		{ 294, 22, 31 },
		{ 294, 23, 31 },
		{ 294, 24, 31 },
		{ 294, 25, 31 },
		{ 294, 26, 31 },
		{ 294, 27, 31 },
		{ 294, 28, 31 },
		{ 294, 29, 31 },
		{ 294, 30, 31 },
		{ 294, 31, 31 },
		{ 294, 32, 31 },
		{ 294, 33, 31 },
		{ 294, 34, 31 },
		{ 294, 35, 31 },
		{ 294, 36, 31 },
		{ 294, 37, 31 },
		{ 294, 38, 31 },
		{ 294, 39, 31 },
		{ 294, 40, 31 },
		{ 294, 41, 31 },
		{ 294, 42, 31 },
		{ 294, 43, 31 },
		{ 294, 44, 31 },
		{ 294, 45, 31 },
		{ 294, 46, 31 },
		{ 294, 47, 31 },
		{ 294, 48, 31 },
		{ 294, 49, 31 },
		{ 294, 50, 31 },
		{ 294, 51, 31 },
		{ 294, 52, 31 },
		{ 294, 53, 31 },
		{ 294, 54, 31 },
		{ 294, 55, 31 },
		{ 294, 56, 31 },
		{ 294, 57, 31 },
		{ 294, 58, 31 },
		{ 294, 59, 31 },
		{ 294, 60, 31 },
		{ 294, 61, 31 },
		{ 294, 62, 31 },
		{ 294, 63, 31 },
		{ 294, 64, 31 },
		{ 294, 65, 31 },
		{ 294, 66, 31 },
		{ 294, 67, 31 },
		{ 294, 68, 31 },
		{ 294, 69, 31 },
		{ 294, 70, 31 },
		{ 294, 71, 31 },
		{ 294, 72, 31 },
		{ 294, 73, 31 },
		{ 294, 74, 31 },
		{ 294, 75, 31 },
		{ 294, 76, 31 },
		{ 294, 77, 31 },
		{ 294, 78, 31 },
		{ 294, 79, 31 },
		{ 294, 80, 31 },
		{ 294, 81, 31 },
		{ 294, 82, 31 },
		{ 294, 83, 31 },
		{ 294, 84, 31 },
		{ 294, 85, 31 },
		{ 294, 86, 31 },
		{ 294, 87, 31 },
		{ 294, 88, 31 },
		{ 294, 89, 31 },
		{ 294, 90, 31 },
		{ 294, 91, 31 },
		{ 294, 92, 31 },
		{ 294, 93, 31 },
		{ 294, 94, 31 },
		{ 294, 95, 31 },
		{ 294, 96, 31 },
		{ 294, 97, 31 },
		{ 294, 98, 31 },
		{ 294, 99, 31 },
		{ 294, 100, 31 },
		{ 294, 101, 31 },
		{ 294, 102, 31 },
		{ 294, 103, 31 },
		{ 294, 104, 31 },
		{ 294, 105, 31 },
		{ 294, 106, 31 },
		{ 294, 107, 31 },
		{ 294, 108, 31 },
		{ 295, 1, 31 },
		{ 295, 2, 31 },
		{ 295, 3, 31 },
		{ 295, 4, 31 },
		{ 295, 5, 31 },
		{ 295, 6, 31 },
		{ 295, 7, 31 },
		{ 295, 8, 31 },
		{ 295, 9, 31 },
		{ 295, 10, 31 },
		{ 295, 11, 31 },
		{ 295, 12, 31 },
		{ 295, 13, 31 },
		{ 295, 14, 31 },
		{ 295, 15, 31 },
		{ 295, 16, 31 },
		{ 295, 17, 31 },
		{ 295, 18, 31 },
		{ 295, 19, 31 },
		{ 295, 20, 31 },
		{ 295, 21, 31 },
		{ 295, 22, 31 },
		{ 295, 23, 31 },
		{ 295, 24, 31 },
		{ 295, 25, 31 },
		{ 295, 26, 31 },
		{ 295, 27, 31 },
		{ 295, 28, 31 },
		{ 295, 29, 31 },
		{ 295, 30, 31 },
		{ 295, 31, 31 },
		{ 295, 32, 31 },
		{ 295, 33, 31 },
		{ 295, 34, 31 },
		{ 295, 35, 31 },
		{ 295, 36, 31 },
		{ 295, 37, 31 },
		{ 295, 38, 31 },
		{ 295, 39, 31 },
		{ 295, 40, 31 },
		{ 295, 41, 31 },
		{ 295, 42, 31 },
		{ 295, 43, 31 },
		{ 295, 44, 31 },
		{ 295, 45, 31 },
		{ 295, 46, 31 },
		{ 295, 47, 31 },
		{ 295, 48, 31 },
		{ 295, 49, 31 },
		{ 295, 50, 31 },
		{ 295, 51, 31 },
		{ 295, 52, 31 },
		{ 295, 53, 31 },
		{ 295, 54, 31 },
		{ 295, 55, 31 },
		{ 295, 56, 31 },
		{ 295, 57, 31 },
		{ 295, 58, 31 },
		{ 295, 59, 31 },
		{ 295, 60, 31 },
		{ 295, 61, 31 },
		{ 295, 62, 31 },
		{ 295, 63, 31 },
		{ 295, 64, 31 },
		{ 295, 65, 31 },
		{ 295, 66, 31 },
		{ 295, 67, 31 },
		{ 295, 68, 31 },
		{ 295, 69, 31 },
		{ 295, 70, 31 },
		{ 295, 71, 31 },
		{ 295, 72, 31 },
		{ 295, 73, 31 },
		{ 295, 74, 31 },
		{ 295, 75, 31 },
		{ 295, 76, 31 },
		{ 295, 77, 31 },
		{ 295, 78, 31 },
		{ 295, 79, 31 },
		{ 295, 80, 31 },
		{ 295, 81, 31 },
		{ 295, 82, 31 },
		{ 295, 83, 31 },
		{ 295, 84, 31 },
		{ 295, 85, 31 },
		{ 295, 86, 31 },
		{ 295, 87, 31 },
		{ 295, 88, 31 },
		{ 295, 89, 31 },
		{ 295, 90, 31 },
		{ 295, 91, 31 },
		{ 295, 92, 31 },
		{ 295, 93, 31 },
		{ 295, 94, 31 },
		{ 295, 95, 31 },
		{ 295, 96, 31 },
		{ 295, 97, 31 },
		{ 295, 98, 31 },
		{ 295, 99, 31 },
		{ 295, 100, 31 },
		{ 295, 101, 31 },
		{ 295, 102, 31 },
		{ 295, 103, 31 },
		{ 295, 104, 31 },
		{ 296, 1, 31 },
		{ 296, 2, 31 },
		{ 296, 3, 31 },
		{ 296, 4, 31 },
		{ 296, 5, 31 },
		{ 296, 6, 31 },
		{ 296, 7, 31 },
		{ 296, 8, 31 },
		{ 296, 9, 31 },
		{ 296, 10, 31 },
		{ 296, 11, 31 },
		{ 296, 12, 31 },
		{ 296, 13, 31 },
		{ 296, 14, 31 },
		{ 296, 15, 31 },
		{ 296, 16, 31 },
		{ 296, 17, 31 },
		{ 296, 18, 31 },
		{ 296, 19, 31 },
		{ 296, 20, 31 },
		{ 296, 21, 31 },
		{ 296, 22, 31 },
		{ 296, 23, 31 },
		{ 296, 24, 31 },
		{ 296, 25, 31 },
		{ 296, 26, 31 },
		{ 296, 27, 31 },
		{ 296, 28, 31 },
		{ 296, 29, 31 },
		{ 296, 30, 31 },
		{ 296, 31, 31 },
		{ 296, 32, 31 },
		{ 296, 33, 31 },
		{ 296, 34, 31 },
		{ 296, 35, 31 },
		{ 296, 36, 31 },
		{ 296, 37, 31 },
		{ 296, 38, 31 },
		{ 296, 39, 31 },
		{ 296, 40, 31 },
		{ 296, 41, 31 },
		{ 296, 42, 31 },
		{ 296, 43, 31 },
		{ 296, 44, 31 },
		{ 296, 45, 31 },
		{ 296, 46, 31 },
		{ 296, 47, 31 },
		{ 296, 48, 31 },
		{ 296, 49, 31 },
		{ 296, 50, 31 },
		{ 296, 51, 31 },
		{ 296, 52, 31 },
		{ 296, 53, 31 },
		{ 296, 54, 31 },
		{ 296, 55, 31 },
		{ 296, 56, 31 },
		{ 296, 57, 31 },
		{ 296, 58, 31 },
		{ 296, 59, 31 },
		{ 296, 60, 31 },
		{ 296, 61, 31 },
		{ 296, 62, 31 },
		{ 296, 63, 31 },
		{ 296, 64, 31 },
		{ 296, 65, 31 },
		{ 296, 66, 31 },
		{ 296, 67, 31 },
		{ 296, 68, 31 },
		{ 296, 69, 31 },
		{ 296, 70, 31 },
		{ 296, 71, 31 },
		{ 296, 72, 31 },
		{ 296, 73, 31 },
		{ 296, 74, 31 },
		{ 296, 75, 31 },
		{ 296, 76, 31 },
		{ 296, 77, 31 },
		{ 296, 78, 31 },
		{ 296, 79, 31 },
		{ 296, 80, 31 },
		{ 296, 81, 31 },
		{ 296, 82, 31 },
		{ 296, 83, 31 },
		{ 296, 84, 31 },
		{ 296, 85, 31 },
		{ 296, 86, 31 },
		{ 296, 87, 31 },
		{ 296, 88, 31 },
		{ 296, 89, 31 },
		{ 296, 90, 31 },
		{ 296, 91, 31 },
		{ 296, 92, 31 },
		{ 296, 93, 31 },
		{ 296, 94, 31 },
		{ 296, 95, 31 },
		{ 296, 96, 31 },
		{ 296, 97, 31 },
		{ 296, 98, 31 },
		{ 296, 99, 31 },
		{ 296, 100, 31 },
		{ 297, 1, 31 },
		{ 297, 2, 31 },
		{ 297, 3, 31 },
		{ 297, 4, 31 },
		{ 297, 5, 31 },
		{ 297, 6, 31 },
		{ 297, 7, 31 },
		{ 297, 8, 31 },
		{ 297, 9, 31 },
		{ 297, 10, 31 },
		{ 297, 11, 31 },
		{ 297, 12, 31 },
		{ 297, 13, 31 },
		{ 297, 14, 31 },
		{ 297, 15, 31 },
		{ 297, 16, 31 },
		{ 297, 17, 31 },
		{ 297, 18, 31 },
		{ 297, 19, 31 },
		{ 297, 20, 31 },
		{ 297, 21, 31 },
		{ 297, 22, 31 },
		{ 297, 23, 31 },
		{ 297, 24, 31 },
		{ 297, 25, 31 },
		{ 297, 26, 31 },
		{ 297, 27, 31 },
		{ 297, 28, 31 },
		{ 297, 29, 31 },
		{ 297, 30, 31 },
		{ 297, 31, 31 },
		{ 297, 32, 31 },
		{ 297, 33, 31 },
		{ 297, 34, 31 },
		{ 297, 35, 31 },
		{ 297, 36, 31 },
		{ 297, 37, 31 },
		{ 297, 38, 31 },
		{ 297, 39, 31 },
		{ 297, 40, 31 },
		{ 297, 41, 31 },
		{ 297, 42, 31 },
		{ 297, 43, 31 },
		{ 297, 44, 31 },
		{ 297, 45, 31 },
		{ 297, 46, 31 },
		{ 297, 47, 31 },
		{ 297, 48, 31 },
		{ 297, 49, 31 },
		{ 297, 50, 31 },
		{ 297, 51, 31 },
		{ 297, 52, 31 },
		{ 297, 53, 31 },
		{ 297, 54, 31 },
		{ 297, 55, 31 },
		{ 297, 56, 31 },
		{ 297, 57, 31 },
		{ 297, 58, 31 },
		{ 297, 59, 31 },
		{ 297, 60, 31 },
		{ 297, 61, 31 },
		{ 297, 62, 31 },
		{ 297, 63, 31 },
		{ 297, 64, 31 },
		{ 297, 65, 31 },
		{ 297, 66, 31 },
		{ 297, 67, 31 },
		{ 297, 68, 31 },
		{ 297, 69, 31 },
		{ 297, 70, 31 },
		{ 297, 71, 31 },
		{ 297, 72, 31 },
		{ 297, 73, 31 },
		{ 297, 74, 31 },
		{ 297, 75, 31 },
		{ 297, 76, 31 },
		{ 297, 77, 31 },
		{ 297, 78, 31 },
		{ 297, 79, 31 },
		{ 297, 80, 31 },
		{ 297, 81, 31 },
		{ 297, 82, 31 },
		{ 297, 83, 31 },
		{ 297, 84, 31 },
		{ 297, 85, 31 },
		{ 297, 86, 31 },
		{ 297, 87, 31 },
		{ 297, 88, 31 },
		{ 297, 89, 31 },
		{ 297, 90, 31 },
		{ 297, 91, 31 },
		{ 297, 92, 31 },
		{ 297, 93, 31 },
		{ 297, 94, 31 },
		{ 297, 95, 31 },
		{ 297, 96, 31 },
		{ 298, 1, 31 },
		{ 298, 2, 31 },
		{ 298, 3, 31 },
		{ 298, 4, 31 },
		{ 298, 5, 31 },
		{ 298, 6, 31 },
		{ 298, 7, 31 },
		{ 298, 8, 31 },
		{ 298, 9, 31 },
		{ 298, 10, 31 },
		{ 298, 11, 31 },
		{ 298, 12, 31 },
		{ 298, 13, 31 },
		{ 298, 14, 31 },
		{ 298, 15, 31 },
		{ 298, 16, 31 },
		{ 298, 17, 31 },
		{ 298, 18, 31 },
		{ 298, 19, 31 },
		{ 298, 20, 31 },
		{ 298, 21, 31 },
		{ 298, 22, 31 },
		{ 298, 23, 31 },
		{ 298, 24, 31 },
		{ 298, 25, 31 },
		{ 298, 26, 31 },
		{ 298, 27, 31 },
		{ 298, 28, 31 },
		{ 298, 29, 31 },
		{ 298, 30, 31 },
		{ 298, 31, 31 },
		{ 298, 32, 31 },
		{ 298, 33, 31 },
		{ 298, 34, 31 },
		{ 298, 35, 31 },
		{ 298, 36, 31 },
		{ 298, 37, 31 },
		{ 298, 38, 31 },
		{ 298, 39, 31 },
		{ 298, 40, 31 },
		{ 298, 41, 31 },
		{ 298, 42, 31 },
		{ 298, 43, 31 },
		{ 298, 44, 31 },
		{ 298, 45, 31 },
		{ 298, 46, 31 },
		{ 298, 47, 31 },
		{ 298, 48, 31 },
		{ 298, 49, 31 },
		{ 298, 50, 31 },
		{ 298, 51, 31 },
		{ 298, 52, 31 },
		{ 298, 53, 31 },
		{ 298, 54, 31 },
		{ 298, 55, 31 },
		{ 298, 56, 31 },
		{ 298, 57, 31 },
		{ 298, 58, 31 },
		{ 298, 59, 31 },
		{ 298, 60, 31 },
		{ 298, 61, 31 },
		{ 298, 62, 31 },
		{ 298, 63, 31 },
		{ 298, 64, 31 },
		{ 298, 65, 31 },
		{ 298, 66, 31 },
		{ 298, 67, 31 },
		{ 298, 68, 31 },
		{ 298, 69, 31 },
		{ 298, 70, 31 },
		{ 298, 71, 31 },
		{ 298, 72, 31 },
		{ 298, 73, 31 },
		{ 298, 74, 31 },
		{ 298, 75, 31 },
		{ 298, 76, 31 },
		{ 298, 77, 31 },
		{ 298, 78, 31 },
		{ 298, 79, 31 },
		{ 298, 80, 31 },
		{ 298, 81, 31 },
		{ 298, 82, 31 },
		{ 298, 83, 31 },
		{ 298, 84, 31 },
		{ 298, 85, 31 },
		{ 298, 86, 31 },
		{ 298, 87, 31 },
		{ 298, 88, 31 },
		{ 298, 89, 31 },
		{ 298, 90, 31 },
		{ 298, 91, 31 },
		{ 298, 92, 31 },
		{ 299, 1, 31 },
		{ 299, 2, 31 },
		{ 299, 3, 31 },
		{ 299, 4, 31 },
		{ 299, 5, 31 },
		{ 299, 6, 31 },
		{ 299, 7, 31 },
		{ 299, 8, 31 },
		{ 299, 9, 31 },
		{ 299, 10, 31 },
		{ 299, 11, 31 },
		{ 299, 12, 31 },
		{ 299, 13, 31 },
		{ 299, 14, 31 },
		{ 299, 15, 31 },
		{ 299, 16, 31 },
		{ 299, 17, 31 },
		{ 299, 18, 31 },
		{ 299, 19, 31 },
		{ 299, 20, 31 },
		{ 299, 21, 31 },
		{ 299, 22, 31 },
		{ 299, 23, 31 },
		{ 299, 24, 31 },
		{ 299, 25, 31 },
		{ 299, 26, 31 },
		{ 299, 27, 31 },
		{ 299, 28, 31 },
		{ 299, 29, 31 },
		{ 299, 30, 31 },
		{ 299, 31, 31 },
		{ 299, 32, 31 },
		{ 299, 33, 31 },
		{ 299, 34, 31 },
		{ 299, 35, 31 },
		{ 299, 36, 31 },
		{ 299, 37, 31 },
		{ 299, 38, 31 },
		{ 299, 39, 31 },
		{ 299, 40, 31 },
		{ 299, 41, 31 },
		{ 299, 42, 31 },
		{ 299, 43, 31 },
		{ 299, 44, 31 },
		{ 299, 45, 31 },
		{ 299, 46, 31 },
		{ 299, 47, 31 },
		{ 299, 48, 31 },
		{ 299, 49, 31 },
		{ 299, 50, 31 },
		{ 299, 51, 31 },
		{ 299, 52, 31 },
		{ 299, 53, 31 },
		{ 299, 54, 31 },
		{ 299, 55, 31 },
		{ 299, 56, 31 },
		{ 299, 57, 31 },
		{ 299, 58, 31 },
		{ 299, 59, 31 },
		{ 299, 60, 31 },
		{ 299, 61, 31 },
		{ 299, 62, 31 },
		{ 299, 63, 31 },
		{ 299, 64, 31 },
		{ 299, 65, 31 },
		{ 299, 66, 31 },
		{ 299, 67, 31 },
		{ 299, 68, 31 },
		{ 299, 69, 31 },
		{ 299, 70, 31 },
		{ 299, 71, 31 },
		{ 299, 72, 31 },
		{ 299, 73, 31 },
		{ 299, 74, 31 },
		{ 299, 75, 31 },
		{ 299, 76, 31 },
		{ 299, 77, 31 },
		{ 299, 78, 31 },
		{ 299, 79, 31 },
		{ 299, 80, 31 },
		{ 299, 81, 31 },
		{ 299, 82, 31 },
		{ 299, 83, 31 },
		{ 299, 84, 31 },
		{ 299, 85, 31 },
		{ 299, 86, 31 },
		{ 299, 87, 31 },
		{ 299, 88, 31 },
		{ 300, 1, 31 },
		{ 300, 2, 31 },
		{ 300, 3, 31 },
		{ 300, 4, 31 },
		{ 300, 5, 31 },
		{ 300, 6, 31 },
		{ 300, 7, 31 },
		{ 300, 8, 31 },
		{ 300, 9, 31 },
		{ 300, 10, 31 },
		{ 300, 11, 31 },
		{ 300, 12, 31 },
		{ 300, 13, 31 },
		{ 300, 14, 31 },
		{ 300, 15, 31 },
		{ 300, 16, 31 },
		{ 300, 17, 31 },
		{ 300, 18, 31 },
		{ 300, 19, 31 },
		{ 300, 20, 31 },
		{ 300, 21, 31 },
		{ 300, 22, 31 },
		{ 300, 23, 31 },
		{ 300, 24, 31 },
		{ 300, 25, 31 },
		{ 300, 26, 31 },
		{ 300, 27, 31 },
		{ 300, 28, 31 },
		{ 300, 29, 31 },
		{ 300, 30, 31 },
		{ 300, 31, 31 },
		{ 300, 32, 31 },
		{ 300, 33, 31 },
		{ 300, 34, 31 },
		{ 300, 35, 31 },
		{ 300, 36, 31 },
		{ 300, 37, 31 },
		{ 300, 38, 31 },
		{ 300, 39, 31 },
		{ 300, 40, 31 },
		{ 300, 41, 31 },
		{ 300, 42, 31 },
		{ 300, 43, 31 },
		{ 300, 44, 31 },
		{ 300, 45, 31 },
		{ 300, 46, 31 },
		{ 300, 47, 31 },
		{ 300, 48, 31 },
		{ 300, 49, 31 },
		{ 300, 50, 31 },
		{ 300, 51, 31 },
		{ 300, 52, 31 },
		{ 300, 53, 31 },
		{ 300, 54, 31 },
		{ 300, 55, 31 },
		{ 300, 56, 31 },
		{ 300, 57, 31 },
		{ 300, 58, 31 },
		{ 300, 59, 31 },
		{ 300, 60, 31 },
		{ 300, 61, 31 },
		{ 300, 62, 31 },
		{ 300, 63, 31 },
		{ 300, 64, 31 },
		{ 300, 65, 31 },
		{ 300, 66, 31 },
		{ 300, 67, 31 },
		{ 300, 68, 31 },
		{ 300, 69, 31 },
		{ 300, 70, 31 },
		{ 300, 71, 31 },
		{ 300, 72, 31 },
		{ 300, 73, 31 },
		{ 300, 74, 31 },
		{ 300, 75, 31 },
		{ 300, 76, 31 },
		{ 300, 77, 31 },
		{ 300, 78, 31 },
		{ 300, 79, 31 },
		{ 300, 80, 31 },
		{ 300, 81, 31 },
		{ 300, 82, 31 },
		{ 300, 83, 31 },
		{ 300, 84, 31 },
		{ 301, 1, 31 },
		{ 301, 2, 31 },
		{ 301, 3, 31 },
		{ 301, 4, 31 },
		{ 301, 5, 31 },
		{ 301, 6, 31 },
		{ 301, 7, 31 },
		{ 301, 8, 31 },
		{ 301, 9, 31 },
		{ 301, 10, 31 },
		{ 301, 11, 31 },
		{ 301, 12, 31 },
		{ 301, 13, 31 },
		{ 301, 14, 31 },
		{ 301, 15, 31 },
		{ 301, 16, 31 },
		{ 301, 17, 31 },
		{ 301, 18, 31 },
		{ 301, 19, 31 },
		{ 301, 20, 31 },
		{ 301, 21, 31 },
		{ 301, 22, 31 },
		{ 301, 23, 31 },
		{ 301, 24, 31 },
		{ 301, 25, 31 },
		{ 301, 26, 31 },
		{ 301, 27, 31 },
		{ 301, 28, 31 },
		{ 301, 29, 31 },
		{ 301, 30, 31 },
		{ 301, 31, 31 },
		{ 301, 32, 31 },
		{ 301, 33, 31 },
		{ 301, 34, 31 },
		{ 301, 35, 31 },
		{ 301, 36, 31 },
		{ 301, 37, 31 },
		{ 301, 38, 31 },
		{ 301, 39, 31 },
		{ 301, 40, 31 },
		{ 301, 41, 31 },
		{ 301, 42, 31 },
		{ 301, 43, 31 },
		{ 301, 44, 31 },
		{ 301, 45, 31 },
		{ 301, 46, 31 },
		{ 301, 47, 31 },
		{ 301, 48, 31 },
		{ 301, 49, 31 },
		{ 301, 50, 31 },
		{ 301, 51, 31 },
		{ 301, 52, 31 },
		{ 301, 53, 31 },
		{ 301, 54, 31 },
		{ 301, 55, 31 },
		{ 301, 56, 31 },
		{ 301, 57, 31 },
		{ 301, 58, 31 },
		{ 301, 59, 31 },
		{ 301, 60, 31 },
		{ 301, 61, 31 },
		{ 301, 62, 31 },
		{ 301, 63, 31 },
		{ 301, 64, 31 },
		{ 301, 65, 31 },
		{ 301, 66, 31 },
		{ 301, 67, 31 },
		{ 301, 68, 31 },
		{ 301, 69, 31 },
		{ 301, 70, 31 },
		{ 301, 71, 31 },
		{ 301, 72, 31 },
		{ 301, 73, 31 },
		{ 301, 74, 31 },
		{ 301, 75, 31 },
		{ 301, 76, 31 },
		{ 301, 77, 31 },
		{ 301, 78, 31 },
		{ 301, 79, 31 },
		{ 301, 80, 31 },
		{ 302, 1, 31 },
		{ 302, 2, 31 },
		{ 302, 3, 31 },
		{ 302, 4, 31 },
		{ 302, 5, 31 },
		{ 302, 6, 31 },
		{ 302, 7, 31 },
		{ 302, 8, 31 },
		{ 302, 9, 31 },
		{ 302, 10, 31 },
		{ 302, 11, 31 },
		{ 302, 12, 31 },
		{ 302, 13, 31 },
		{ 302, 14, 31 },
		{ 302, 15, 31 },
		{ 302, 16, 31 },
		{ 302, 17, 31 },
		{ 302, 18, 31 },
		{ 302, 19, 31 },
		{ 302, 20, 31 },
		{ 302, 21, 31 },
		{ 302, 22, 31 },
		{ 302, 23, 31 },
		{ 302, 24, 31 },
		{ 302, 25, 31 },
		{ 302, 26, 31 },
		{ 302, 27, 31 },
		{ 302, 28, 31 },
		{ 302, 29, 31 },
		{ 302, 30, 31 },
		{ 302, 31, 31 },
		{ 302, 32, 31 },
		{ 302, 33, 31 },
		{ 302, 34, 31 },
		{ 302, 35, 31 },
		{ 302, 36, 31 },
		{ 302, 37, 31 },
		{ 302, 38, 31 },
		{ 302, 39, 31 },
		{ 302, 40, 31 },
		{ 302, 41, 31 },
		{ 302, 42, 31 },
		{ 302, 43, 31 },
		{ 302, 44, 31 },
		{ 302, 45, 31 },
		{ 302, 46, 31 },
		{ 302, 47, 31 },
		{ 302, 48, 31 },
		{ 302, 49, 31 },
		{ 302, 50, 31 },
		{ 302, 51, 31 },
		{ 302, 52, 31 },
		{ 302, 53, 31 },
		{ 302, 54, 31 },
		{ 302, 55, 31 },
		{ 302, 56, 31 },
		{ 302, 57, 31 },
		{ 302, 58, 31 },
		{ 302, 59, 31 },
		{ 302, 60, 31 },
		{ 302, 61, 31 },
		{ 302, 62, 31 },
		{ 302, 63, 31 },
		{ 302, 64, 31 },
		{ 302, 65, 31 },
		{ 302, 66, 31 },
		{ 302, 67, 31 },
		{ 302, 68, 31 },
		{ 302, 69, 31 },
		{ 302, 70, 31 },
		{ 302, 71, 31 },
		{ 302, 72, 31 },
		{ 302, 73, 31 },
		{ 302, 74, 31 },
		{ 302, 75, 31 },
		{ 302, 76, 31 },
		{ 303, 1, 31 },
		{ 303, 2, 31 },
		{ 303, 3, 31 },
		{ 303, 4, 31 },
		{ 303, 5, 31 },
		{ 303, 6, 31 },
		{ 303, 7, 31 },
		{ 303, 8, 31 },
		{ 303, 9, 31 },
		{ 303, 10, 31 },
		{ 303, 11, 31 },
		{ 303, 12, 31 },
		{ 303, 13, 31 },
		{ 303, 14, 31 },
		{ 303, 15, 31 },
		{ 303, 16, 31 },
		{ 303, 17, 31 },
		{ 303, 18, 31 },
		{ 303, 19, 31 },
		{ 303, 20, 31 },
		{ 303, 21, 31 },
		{ 303, 22, 31 },
		{ 303, 23, 31 },
		{ 303, 24, 31 },
		{ 303, 25, 31 },
		{ 303, 26, 31 },
		{ 303, 27, 31 },
		{ 303, 28, 31 },
		{ 303, 29, 31 },
		{ 303, 30, 31 },
		{ 303, 31, 31 },
		{ 303, 32, 31 },
		{ 303, 33, 31 },
		{ 303, 34, 31 },
		{ 303, 35, 31 },
		{ 303, 36, 31 },
		{ 303, 37, 31 },
		{ 303, 38, 31 },
		{ 303, 39, 31 },
		{ 303, 40, 31 },
		{ 303, 41, 31 },
		{ 303, 42, 31 },
		{ 303, 43, 31 },
		{ 303, 44, 31 },
		{ 303, 45, 31 },
		{ 303, 46, 31 },
		{ 303, 47, 31 },
		{ 303, 48, 31 },
		{ 303, 49, 31 },
		{ 303, 50, 31 },
		{ 303, 51, 31 },
		{ 303, 52, 31 },
		{ 303, 53, 31 },
		{ 303, 54, 31 },
		{ 303, 55, 31 },
		{ 303, 56, 31 },
		{ 303, 57, 31 },
		{ 303, 58, 31 },
		{ 303, 59, 31 },
		{ 303, 60, 31 },
		{ 303, 61, 31 },
		{ 303, 62, 31 },
		{ 303, 63, 31 },
		{ 303, 64, 31 },
		{ 303, 65, 31 },
		{ 303, 66, 31 },
		{ 303, 67, 31 },
		{ 303, 68, 31 },
		{ 303, 69, 31 },
		{ 303, 70, 31 },
		{ 303, 71, 31 },
		{ 303, 72, 31 },
		{ 304, 1, 31 },
		{ 304, 2, 31 },
		{ 304, 3, 31 },
		{ 304, 4, 31 },
		{ 304, 5, 31 },
		{ 304, 6, 31 },
		{ 304, 7, 31 },
		{ 304, 8, 31 },
		{ 304, 9, 31 },
		{ 304, 10, 31 },
		{ 304, 11, 31 },
		{ 304, 12, 31 },
		{ 304, 13, 31 },
		{ 304, 14, 31 },
		{ 304, 15, 31 },
		{ 304, 16, 31 },
		{ 304, 17, 31 },
		{ 304, 18, 31 },
		{ 304, 19, 31 },
		{ 304, 20, 31 },
		{ 304, 21, 31 },
		{ 304, 22, 31 },
		{ 304, 23, 31 },
		{ 304, 24, 31 },
		{ 304, 25, 31 },
		{ 304, 26, 31 },
		{ 304, 27, 31 },
		{ 304, 28, 31 },
		{ 304, 29, 31 },
		{ 304, 30, 31 },
		{ 304, 31, 31 },
		{ 304, 32, 31 },
		{ 304, 33, 31 },
		{ 304, 34, 31 },
		{ 304, 35, 31 },
		{ 304, 36, 31 },
		{ 304, 37, 31 },
		{ 304, 38, 31 },
		{ 304, 39, 31 },
		{ 304, 40, 31 },
		{ 304, 41, 31 },
		{ 304, 42, 31 },
		{ 304, 43, 31 },
		{ 304, 44, 31 },
		{ 304, 45, 31 },
		{ 304, 46, 31 },
		{ 304, 47, 31 },
		{ 304, 48, 31 },
		{ 304, 49, 31 },
		{ 304, 50, 31 },
		{ 304, 51, 31 },
		{ 304, 52, 31 },
		{ 304, 53, 31 },
		{ 304, 54, 31 },
		{ 304, 55, 31 },
		{ 304, 56, 31 },
		{ 304, 57, 31 },
		{ 304, 58, 31 },
		{ 304, 59, 31 },
		{ 304, 60, 31 },
		{ 304, 61, 31 },
		{ 304, 62, 31 },
		{ 304, 63, 31 },
		{ 304, 64, 31 },
		{ 304, 65, 31 },
		{ 304, 66, 31 },
		{ 304, 67, 31 },
		{ 304, 68, 31 },
		{ 305, 1, 31 },
		{ 305, 2, 31 },
		{ 305, 3, 31 },
		{ 305, 4, 31 },
		{ 305, 5, 31 },
		{ 305, 6, 31 },
		{ 305, 7, 31 },
		{ 305, 8, 31 },
		{ 305, 9, 31 },
		{ 305, 10, 31 },
		{ 305, 11, 31 },
		{ 305, 12, 31 },
		{ 305, 13, 31 },
		{ 305, 14, 31 },
		{ 305, 15, 31 },
		{ 305, 16, 31 },
		{ 305, 17, 31 },
		{ 305, 18, 31 },
		{ 305, 19, 31 },
		{ 305, 20, 31 },
		{ 305, 21, 31 },
		{ 305, 22, 31 },
		{ 305, 23, 31 },
		{ 305, 24, 31 },
		{ 305, 25, 31 },
		{ 305, 26, 31 },
		{ 305, 27, 31 },
		{ 305, 28, 31 },
		{ 305, 29, 31 },
		{ 305, 30, 31 },
		{ 305, 31, 31 },
		{ 305, 32, 31 },
		{ 305, 33, 31 },
		{ 305, 34, 31 },
		{ 305, 35, 31 },
		{ 305, 36, 31 },
		{ 305, 37, 31 },
		{ 305, 38, 31 },
		{ 305, 39, 31 },
		{ 305, 40, 31 },
		{ 305, 41, 31 },
		{ 305, 42, 31 },
		{ 305, 43, 31 },
		{ 305, 44, 31 },
		{ 305, 45, 31 },
		{ 305, 46, 31 },
		{ 305, 47, 31 },
		{ 305, 48, 31 },
		{ 305, 49, 31 },
		{ 305, 50, 31 },
		{ 305, 51, 31 },
		{ 305, 52, 31 },
		{ 305, 53, 31 },
		{ 305, 54, 31 },
		{ 305, 55, 31 },
		{ 305, 56, 31 },
		{ 305, 57, 31 },
		{ 305, 58, 31 },
		{ 305, 59, 31 },
		{ 305, 60, 31 },
		{ 305, 61, 31 },
		{ 305, 62, 31 },
		{ 305, 63, 31 },
		{ 305, 64, 31 },
		{ 306, 1, 31 },
		{ 306, 2, 31 },
		{ 306, 3, 31 },
		{ 306, 4, 31 },
		{ 306, 5, 31 },
		{ 306, 6, 31 },
		{ 306, 7, 31 },
		{ 306, 8, 31 },
		{ 306, 9, 31 },
		{ 306, 10, 31 },
		{ 306, 11, 31 },
		{ 306, 12, 31 },
		{ 306, 13, 31 },
		{ 306, 14, 31 },
		{ 306, 15, 31 },
		{ 306, 16, 31 },
		{ 306, 17, 31 },
		{ 306, 18, 31 },
		{ 306, 19, 31 },
		{ 306, 20, 31 },
		{ 306, 21, 31 },
		{ 306, 22, 31 },
		{ 306, 23, 31 },
		{ 306, 24, 31 },
		{ 306, 25, 31 },
		{ 306, 26, 31 },
		{ 306, 27, 31 },
		{ 306, 28, 31 },
		{ 306, 29, 31 },
		{ 306, 30, 31 },
		{ 306, 31, 31 },
		{ 306, 32, 31 },
		{ 306, 33, 31 },
		{ 306, 34, 31 },
		{ 306, 35, 31 },
		{ 306, 36, 31 },
		{ 306, 37, 31 },
		{ 306, 38, 31 },
		{ 306, 39, 31 },
		{ 306, 40, 31 },
		{ 306, 41, 31 },
		{ 306, 42, 31 },
		{ 306, 43, 31 },
		{ 306, 44, 31 },
		{ 306, 45, 31 },
		{ 306, 46, 31 },
		{ 306, 47, 31 },
		{ 306, 48, 31 },
		{ 306, 49, 31 },
		{ 306, 50, 31 },
		{ 306, 51, 31 },
		{ 306, 52, 31 },
		{ 306, 53, 31 },
		{ 306, 54, 31 },
		{ 306, 55, 31 },
		{ 306, 56, 31 },
		{ 306, 57, 31 },
		{ 306, 58, 31 },
		{ 306, 59, 31 },
		{ 306, 60, 31 },
		{ 307, 1, 31 },
		{ 307, 2, 31 },
		{ 307, 3, 31 },
		{ 307, 4, 31 },
		{ 307, 5, 31 },
		{ 307, 6, 31 },
		{ 307, 7, 31 },
		{ 307, 8, 31 },
		{ 307, 9, 31 },
		{ 307, 10, 31 },
		{ 307, 11, 31 },
		{ 307, 12, 31 },
		{ 307, 13, 31 },
		{ 307, 14, 31 },
		{ 307, 15, 31 },
		{ 307, 16, 31 },
		{ 307, 17, 31 },
		{ 307, 18, 31 },
		{ 307, 19, 31 },
		{ 307, 20, 31 },
		{ 307, 21, 31 },
		{ 307, 22, 31 },
		{ 307, 23, 31 },
		{ 307, 24, 31 },
		{ 307, 25, 31 },
		{ 307, 26, 31 },
		{ 307, 27, 31 },
		{ 307, 28, 31 },
		{ 307, 29, 31 },
		{ 307, 30, 31 },
		{ 307, 31, 31 },
		{ 307, 32, 31 },
		{ 307, 33, 31 },
		{ 307, 34, 31 },
		{ 307, 35, 31 },
		{ 307, 36, 31 },
		{ 307, 37, 31 },
		{ 307, 38, 31 },
		{ 307, 39, 31 },
		{ 307, 40, 31 },
		{ 307, 41, 31 },
		{ 307, 42, 31 },
		{ 307, 43, 31 },
		{ 307, 44, 31 },
		{ 307, 45, 31 },
		{ 307, 46, 31 },
		{ 307, 47, 31 },
		{ 307, 48, 31 },
		{ 307, 49, 31 },
		{ 307, 50, 31 },
		{ 307, 51, 31 },
		{ 307, 52, 31 },
		{ 307, 53, 31 },
		{ 307, 54, 31 },
		{ 307, 55, 31 },
		{ 307, 56, 31 },
		{ 308, 1, 31 },
		{ 308, 2, 31 },
		{ 308, 3, 31 },
		{ 308, 4, 31 },
		{ 308, 5, 31 },
		{ 308, 6, 31 },
		{ 308, 7, 31 },
		{ 308, 8, 31 },
		{ 308, 9, 31 },
		{ 308, 10, 31 },
		{ 308, 11, 31 },
		{ 308, 12, 31 },
		{ 308, 13, 31 },
		{ 308, 14, 31 },
		{ 308, 15, 31 },
		{ 308, 16, 31 },
		{ 308, 17, 31 },
		{ 308, 18, 31 },
		{ 308, 19, 31 },
		{ 308, 20, 31 },
		{ 308, 21, 31 },
		{ 308, 22, 31 },
		{ 308, 23, 31 },
		{ 308, 24, 31 },
		{ 308, 25, 31 },
		{ 308, 26, 31 },
		{ 308, 27, 31 },
		{ 308, 28, 31 },
		{ 308, 29, 31 },
		{ 308, 30, 31 },
		{ 308, 31, 31 },
		{ 308, 32, 31 },
		{ 308, 33, 31 },
		{ 308, 34, 31 },
		{ 308, 35, 31 },
		{ 308, 36, 31 },
		{ 308, 37, 31 },
		{ 308, 38, 31 },
		{ 308, 39, 31 },
		{ 308, 40, 31 },
		{ 308, 41, 31 },
		{ 308, 42, 31 },
		{ 308, 43, 31 },
		{ 308, 44, 31 },
		{ 308, 45, 31 },
		{ 308, 46, 31 },
		{ 308, 47, 31 },
		{ 308, 48, 31 },
		{ 308, 49, 31 },
		{ 308, 50, 31 },
		{ 308, 51, 31 },
		{ 308, 52, 31 },
		{ 309, 1, 31 },
		{ 309, 2, 31 },
		{ 309, 3, 31 },
		{ 309, 4, 31 },
		{ 309, 5, 31 },
		{ 309, 6, 31 },
		{ 309, 7, 31 },
		{ 309, 8, 31 },
		{ 309, 9, 31 },
		{ 309, 10, 31 },
		{ 309, 11, 31 },
		{ 309, 12, 31 },
		{ 309, 13, 31 },
		{ 309, 14, 31 },
		{ 309, 15, 31 },
		{ 309, 16, 31 },
		{ 309, 17, 31 },
		{ 309, 18, 31 },
		{ 309, 19, 31 },
		{ 309, 20, 31 },
		{ 309, 21, 31 },
		{ 309, 22, 31 },
		{ 309, 23, 31 },
		{ 309, 24, 31 },
		{ 309, 25, 31 },
		{ 309, 26, 31 },
		{ 309, 27, 31 },
		{ 309, 28, 31 },
		{ 309, 29, 31 },
		{ 309, 30, 31 },
		{ 309, 31, 31 },
		{ 309, 32, 31 },
		{ 309, 33, 31 },
		{ 309, 34, 31 },
		{ 309, 35, 31 },
		{ 309, 36, 31 },
		{ 309, 37, 31 },
		{ 309, 38, 31 },
		{ 309, 39, 31 },
		{ 309, 40, 31 },
		{ 309, 41, 31 },
		{ 309, 42, 31 },
		{ 309, 43, 31 },
		{ 309, 44, 31 },
		{ 309, 45, 31 },
		{ 309, 46, 31 },
		{ 309, 47, 31 },
		{ 309, 48, 31 },
		{ 310, 1, 31 },
		{ 310, 2, 31 },
		{ 310, 3, 31 },
		{ 310, 4, 31 },
		{ 310, 5, 31 },
		{ 310, 6, 31 },
		{ 310, 7, 31 },
		{ 310, 8, 31 },
		{ 310, 9, 31 },
		{ 310, 10, 31 },
		{ 310, 11, 31 },
		{ 310, 12, 31 },
		{ 310, 13, 31 },
		{ 310, 14, 31 },
		{ 310, 15, 31 },
		{ 310, 16, 31 },
		{ 310, 17, 31 },
		{ 310, 18, 31 },
		{ 310, 19, 31 },
		{ 310, 20, 31 },
		{ 310, 21, 31 },
		{ 310, 22, 31 },
		{ 310, 23, 31 },
		{ 310, 24, 31 },
		{ 310, 25, 31 },
		{ 310, 26, 31 },
		{ 310, 27, 31 },
		{ 310, 28, 31 },
		{ 310, 29, 31 },
		{ 310, 30, 31 },
		{ 310, 31, 31 },
		{ 310, 32, 31 },
		{ 310, 33, 31 },
		{ 310, 34, 31 },
		{ 310, 35, 31 },
		{ 310, 36, 31 },
		{ 310, 37, 31 },
		{ 310, 38, 31 },
		{ 310, 39, 31 },
		{ 310, 40, 31 },
		{ 310, 41, 31 },
		{ 310, 42, 31 },
		{ 310, 43, 31 },
		{ 310, 44, 31 },
		{ 311, 1, 31 },
		{ 311, 2, 31 },
		{ 311, 3, 31 },
		{ 311, 4, 31 },
		{ 311, 5, 31 },
		{ 311, 6, 31 },
		{ 311, 7, 31 },
		{ 311, 8, 31 },
		{ 311, 9, 31 },
		{ 311, 10, 31 },
		{ 311, 11, 31 },
		{ 311, 12, 31 },
		{ 311, 13, 31 },
		{ 311, 14, 31 },
		{ 311, 15, 31 },
		{ 311, 16, 31 },
		{ 311, 17, 31 },
		{ 311, 18, 31 },
		{ 311, 19, 31 },
		{ 311, 20, 31 },
		{ 311, 21, 31 },
		{ 311, 22, 31 },
		{ 311, 23, 31 },
		{ 311, 24, 31 },
		{ 311, 25, 31 },
		{ 311, 26, 31 },
		{ 311, 27, 31 },
		{ 311, 28, 31 },
		{ 311, 29, 31 },
		{ 311, 30, 31 },
		{ 311, 31, 31 },
		{ 311, 32, 31 },
		{ 311, 33, 31 },
		{ 311, 34, 31 },
		{ 311, 35, 31 },
		{ 311, 36, 31 },
		{ 311, 37, 31 },
		{ 311, 38, 31 },
		{ 311, 39, 31 },
		{ 311, 40, 31 },
		{ 312, 1, 31 },
		{ 312, 2, 31 },
		{ 312, 3, 31 },
		{ 312, 4, 31 },
		{ 312, 5, 31 },
		{ 312, 6, 31 },
		{ 312, 7, 31 },
		{ 312, 8, 31 },
		{ 312, 9, 31 },
		{ 312, 10, 31 },
		{ 312, 11, 31 },
		{ 312, 12, 31 },
		{ 312, 13, 31 },
		{ 312, 14, 31 },
		{ 312, 15, 31 },
		{ 312, 16, 31 },
		{ 312, 17, 31 },
		{ 312, 18, 31 },
		{ 312, 19, 31 },
		{ 312, 20, 31 },
		{ 312, 21, 31 },
		{ 312, 22, 31 },
		{ 312, 23, 31 },
		{ 312, 24, 31 },
		{ 312, 25, 31 },
		{ 312, 26, 31 },
		{ 312, 27, 31 },
		{ 312, 28, 31 },
		{ 312, 29, 31 },
		{ 312, 30, 31 },
		{ 312, 31, 31 },
		{ 312, 32, 31 },
		{ 312, 33, 31 },
		{ 312, 34, 31 },
		{ 312, 35, 31 },
		{ 312, 36, 31 },
		{ 313, 1, 31 },
		{ 313, 2, 31 },
		{ 313, 3, 31 },
		{ 313, 4, 31 },
		{ 313, 5, 31 },
		{ 313, 6, 31 },
		{ 313, 7, 31 },
		{ 313, 8, 31 },
		{ 313, 9, 31 },
		{ 313, 10, 31 },
		{ 313, 11, 31 },
		{ 313, 12, 31 },
		{ 313, 13, 31 },
		{ 313, 14, 31 },
		{ 313, 15, 31 },
		{ 313, 16, 31 },
		{ 313, 17, 31 },
		{ 313, 18, 31 },
		{ 313, 19, 31 },
		{ 313, 20, 31 },
		{ 313, 21, 31 },
		{ 313, 22, 31 },
		{ 313, 23, 31 },
		{ 313, 24, 31 },
		{ 313, 25, 31 },
		{ 313, 26, 31 },
		{ 313, 27, 31 },
		{ 313, 28, 31 },
		{ 313, 29, 31 },
		{ 313, 30, 31 },
		{ 313, 31, 31 },
		{ 313, 32, 31 },
		{ 314, 1, 31 },
		{ 314, 2, 31 },
		{ 314, 3, 31 },
		{ 314, 4, 31 },
		{ 314, 5, 31 },
		{ 314, 6, 31 },
		{ 314, 7, 31 },
		{ 314, 8, 31 },
		{ 314, 9, 31 },
		{ 314, 10, 31 },
		{ 314, 11, 31 },
		{ 314, 12, 31 },
		{ 314, 13, 31 },
		{ 314, 14, 31 },
		{ 314, 15, 31 },
		{ 314, 16, 31 },
		{ 314, 17, 31 },
		{ 314, 18, 31 },
		{ 314, 19, 31 },
		{ 314, 20, 31 },
		{ 314, 21, 31 },
		{ 314, 22, 31 },
		{ 314, 23, 31 },
		{ 314, 24, 31 },
		{ 314, 25, 31 },
		{ 314, 26, 31 },
		{ 314, 27, 31 },
		{ 314, 28, 31 },
		{ 315, 1, 31 },
		{ 315, 2, 31 },
		{ 315, 3, 31 },
		{ 315, 4, 31 },
		{ 315, 5, 31 },
		{ 315, 6, 31 },
		{ 315, 7, 31 },
		{ 315, 8, 31 },
		{ 315, 9, 31 },
		{ 315, 10, 31 },
		{ 315, 11, 31 },
		{ 315, 12, 31 },
		{ 315, 13, 31 },
		{ 315, 14, 31 },
		{ 315, 15, 31 },
		{ 315, 16, 31 },
		{ 315, 17, 31 },
		{ 315, 18, 31 },
		{ 315, 19, 31 },
		{ 315, 20, 31 },
		{ 315, 21, 31 },
		{ 315, 22, 31 },
		{ 315, 23, 31 },
		{ 315, 24, 31 },
		{ 316, 1, 31 },
		{ 316, 2, 31 },
		{ 316, 3, 31 },
		{ 316, 4, 31 },
		{ 316, 5, 31 },
		{ 316, 6, 31 },
		{ 316, 7, 31 },
		{ 316, 8, 31 },
		{ 316, 9, 31 },
		{ 316, 10, 31 },
		{ 316, 11, 31 },
		{ 316, 12, 31 },
		{ 316, 13, 31 },
		{ 316, 14, 31 },
		{ 316, 15, 31 },
		{ 316, 16, 31 },
		{ 316, 17, 31 },
		{ 316, 18, 31 },
		{ 316, 19, 31 },
		{ 316, 20, 31 },
		{ 317, 1, 31 },
		{ 317, 2, 31 },
		{ 317, 3, 31 },
		{ 317, 4, 31 },
		{ 317, 5, 31 },
		{ 317, 6, 31 },
		{ 317, 7, 31 },
		{ 317, 8, 31 },
		{ 317, 9, 31 },
		{ 317, 10, 31 },
		{ 317, 11, 31 },
		{ 317, 12, 31 },
		{ 317, 13, 31 },
		{ 317, 14, 31 },
		{ 317, 15, 31 },
		{ 317, 16, 31 },
		{ 318, 1, 31 },
		{ 318, 2, 31 },
		{ 318, 3, 31 },
		{ 318, 4, 31 },
		{ 318, 5, 31 },
		{ 318, 6, 31 },
		{ 318, 7, 31 },
		{ 318, 8, 31 },
		{ 318, 9, 31 },
		{ 318, 10, 31 },
		{ 318, 11, 31 },
		{ 318, 12, 31 },
		{ 319, 1, 31 },
		{ 319, 2, 31 },
		{ 319, 3, 31 },
		{ 319, 4, 31 },
		{ 319, 5, 31 },
		{ 319, 6, 31 },
		{ 319, 7, 31 },
		{ 319, 8, 31 },
		{ 320, 1, 31 },
		{ 320, 2, 31 },
		{ 320, 3, 31 },
		{ 320, 4, 31 },
		{ 162, 511, 32 },
		{ 162, 512, 32 },
		{ 163, 508, 32 },
		{ 163, 509, 32 },
		{ 163, 510, 32 },
		{ 163, 511, 32 },
		{ 163, 512, 32 },
		{ 164, 506, 32 },
		{ 164, 507, 32 },
		{ 164, 508, 32 },
		{ 164, 509, 32 },
		{ 164, 510, 32 },
		{ 164, 511, 32 },
		{ 164, 512, 32 },
		{ 165, 503, 32 },
		{ 165, 504, 32 },
		{ 165, 505, 32 },
		{ 165, 506, 32 },
		{ 165, 507, 32 },
		{ 165, 508, 32 },
		{ 165, 509, 32 },
		{ 165, 510, 32 },
		{ 165, 511, 32 },
		{ 165, 512, 32 },
		{ 166, 500, 32 },
		{ 166, 501, 32 },
		{ 166, 502, 32 },
		{ 166, 503, 32 },
		{ 166, 504, 32 },
		{ 166, 505, 32 },
		{ 166, 506, 32 },
		{ 166, 507, 32 },
		{ 166, 508, 32 },
		{ 166, 509, 32 },
		{ 166, 510, 32 },
		{ 166, 511, 32 },
		{ 166, 512, 32 },
		{ 167, 498, 32 },
		{ 167, 499, 32 },
		{ 167, 500, 32 },
		{ 167, 501, 32 },
		{ 167, 502, 32 },
		{ 167, 503, 32 },
		{ 167, 504, 32 },
		{ 167, 505, 32 },
		{ 167, 506, 32 },
		{ 167, 507, 32 },
		{ 167, 508, 32 },
		{ 167, 509, 32 },
		{ 167, 510, 32 },
		{ 167, 511, 32 },
		{ 167, 512, 32 },
		{ 168, 495, 32 },
		{ 168, 496, 32 },
		{ 168, 497, 32 },
		{ 168, 498, 32 },
		{ 168, 499, 32 },
		{ 168, 500, 32 },
		{ 168, 501, 32 },
		{ 168, 502, 32 },
		{ 168, 503, 32 },
		{ 168, 504, 32 },
		{ 168, 505, 32 },
		{ 168, 506, 32 },
		{ 168, 507, 32 },
		{ 168, 508, 32 },
		{ 168, 509, 32 },
		{ 168, 510, 32 },
		{ 168, 511, 32 },
		{ 168, 512, 32 },
		{ 169, 492, 32 },
		{ 169, 493, 32 },
		{ 169, 494, 32 },
		{ 169, 495, 32 },
		{ 169, 496, 32 },
		{ 169, 497, 32 },
		{ 169, 498, 32 },
		{ 169, 499, 32 },
		{ 169, 500, 32 },
		{ 169, 501, 32 },
		{ 169, 502, 32 },
		{ 169, 503, 32 },
		{ 169, 504, 32 },
		{ 169, 505, 32 },
		{ 169, 506, 32 },
		{ 169, 507, 32 },
		{ 169, 508, 32 },
		{ 169, 509, 32 },
		{ 169, 510, 32 },
		{ 169, 511, 32 },
		{ 169, 512, 32 },
		{ 170, 490, 32 },
		{ 170, 491, 32 },
		{ 170, 492, 32 },
		{ 170, 493, 32 },
		{ 170, 494, 32 },
		{ 170, 495, 32 },
		{ 170, 496, 32 },
		{ 170, 497, 32 },
		{ 170, 498, 32 },
		{ 170, 499, 32 },
		{ 170, 500, 32 },
		{ 170, 501, 32 },
		{ 170, 502, 32 },
		{ 170, 503, 32 },
		{ 170, 504, 32 },
		{ 170, 505, 32 },
		{ 170, 506, 32 },
		{ 170, 507, 32 },
		{ 170, 508, 32 },
		{ 170, 509, 32 },
		{ 170, 510, 32 },
		{ 170, 511, 32 },
		{ 170, 512, 32 },
		{ 171, 487, 32 },
		{ 171, 488, 32 },
		{ 171, 489, 32 },
		{ 171, 490, 32 },
		{ 171, 491, 32 },
		{ 171, 492, 32 },
		{ 171, 493, 32 },
		{ 171, 494, 32 },
		{ 171, 495, 32 },
		{ 171, 496, 32 },
		{ 171, 497, 32 },
		{ 171, 498, 32 },
		{ 171, 499, 32 },
		{ 171, 500, 32 },
		{ 171, 501, 32 },
		{ 171, 502, 32 },
		{ 171, 503, 32 },
		{ 171, 504, 32 },
		{ 171, 505, 32 },
		{ 171, 506, 32 },
		{ 171, 507, 32 },
		{ 171, 508, 32 },
		{ 171, 509, 32 },
		{ 171, 510, 32 },
		{ 171, 511, 32 },
		{ 171, 512, 32 },
		{ 172, 484, 32 },
		{ 172, 485, 32 },
		{ 172, 486, 32 },
		{ 172, 487, 32 },
		{ 172, 488, 32 },
		{ 172, 489, 32 },
		{ 172, 490, 32 },
		{ 172, 491, 32 },
		{ 172, 492, 32 },
		{ 172, 493, 32 },
		{ 172, 494, 32 },
		{ 172, 495, 32 },
		{ 172, 496, 32 },
		{ 172, 497, 32 },
		{ 172, 498, 32 },
		{ 172, 499, 32 },
		{ 172, 500, 32 },
		{ 172, 501, 32 },
		{ 172, 502, 32 },
		{ 172, 503, 32 },
		{ 172, 504, 32 },
		{ 172, 505, 32 },
		{ 172, 506, 32 },
		{ 172, 507, 32 },
		{ 172, 508, 32 },
		{ 172, 509, 32 },
		{ 172, 510, 32 },
		{ 172, 511, 32 },
		{ 172, 512, 32 },
		{ 173, 482, 32 },
		{ 173, 483, 32 },
		{ 173, 484, 32 },
		{ 173, 485, 32 },
		{ 173, 486, 32 },
		{ 173, 487, 32 },
		{ 173, 488, 32 },
		{ 173, 489, 32 },
		{ 173, 490, 32 },
		{ 173, 491, 32 },
		{ 173, 492, 32 },
		{ 173, 493, 32 },
		{ 173, 494, 32 },
		{ 173, 495, 32 },
		{ 173, 496, 32 },
		{ 173, 497, 32 },
		{ 173, 498, 32 },
		{ 173, 499, 32 },
		{ 173, 500, 32 },
		{ 173, 501, 32 },
		{ 173, 502, 32 },
		{ 173, 503, 32 },
		{ 173, 504, 32 },
		{ 173, 505, 32 },
		{ 173, 506, 32 },
		{ 173, 507, 32 },
		{ 173, 508, 32 },
		{ 173, 509, 32 },
		{ 173, 510, 32 },
		{ 173, 511, 32 },
		{ 173, 512, 32 },
		{ 174, 479, 32 },
		{ 174, 480, 32 },
		{ 174, 481, 32 },
		{ 174, 482, 32 },
		{ 174, 483, 32 },
		{ 174, 484, 32 },
		{ 174, 485, 32 },
		{ 174, 486, 32 },
		{ 174, 487, 32 },
		{ 174, 488, 32 },
		{ 174, 489, 32 },
		{ 174, 490, 32 },
		{ 174, 491, 32 },
		{ 174, 492, 32 },
		{ 174, 493, 32 },
		{ 174, 494, 32 },
		{ 174, 495, 32 },
		{ 174, 496, 32 },
		{ 174, 497, 32 },
		{ 174, 498, 32 },
		{ 174, 499, 32 },
		{ 174, 500, 32 },
		{ 174, 501, 32 },
		{ 174, 502, 32 },
		{ 174, 503, 32 },
		{ 174, 504, 32 },
		{ 174, 505, 32 },
		{ 174, 506, 32 },
		{ 174, 507, 32 },
		{ 174, 508, 32 },
		{ 174, 509, 32 },
		{ 174, 510, 32 },
		{ 174, 511, 32 },
		{ 174, 512, 32 },
		{ 175, 476, 32 },
		{ 175, 477, 32 },
		{ 175, 478, 32 },
		{ 175, 479, 32 },
		{ 175, 480, 32 },
		{ 175, 481, 32 },
		{ 175, 482, 32 },
		{ 175, 483, 32 },
		{ 175, 484, 32 },
		{ 175, 485, 32 },
		{ 175, 486, 32 },
		{ 175, 487, 32 },
		{ 175, 488, 32 },
		{ 175, 489, 32 },
		{ 175, 490, 32 },
		{ 175, 491, 32 },
		{ 175, 492, 32 },
		{ 175, 493, 32 },
		{ 175, 494, 32 },
		{ 175, 495, 32 },
		{ 175, 496, 32 },
		{ 175, 497, 32 },
		{ 175, 498, 32 },
		{ 175, 499, 32 },
		{ 175, 500, 32 },
		{ 175, 501, 32 },
		{ 175, 502, 32 },
		{ 175, 503, 32 },
		{ 175, 504, 32 },
		{ 175, 505, 32 },
		{ 175, 506, 32 },
		{ 175, 507, 32 },
		{ 175, 508, 32 },
		{ 175, 509, 32 },
		{ 175, 510, 32 },
		{ 175, 511, 32 },
		{ 175, 512, 32 },
		{ 176, 474, 32 },
		{ 176, 475, 32 },
		{ 176, 476, 32 },
		{ 176, 477, 32 },
		{ 176, 478, 32 },
		{ 176, 479, 32 },
		{ 176, 480, 32 },
		{ 176, 481, 32 },
		{ 176, 482, 32 },
		{ 176, 483, 32 },
		{ 176, 484, 32 },
		{ 176, 485, 32 },
		{ 176, 486, 32 },
		{ 176, 487, 32 },
		{ 176, 488, 32 },
		{ 176, 489, 32 },
		{ 176, 490, 32 },
		{ 176, 491, 32 },
		{ 176, 492, 32 },
		{ 176, 493, 32 },
		{ 176, 494, 32 },
		{ 176, 495, 32 },
		{ 176, 496, 32 },
		{ 176, 497, 32 },
		{ 176, 498, 32 },
		{ 176, 499, 32 },
		{ 176, 500, 32 },
		{ 176, 501, 32 },
		{ 176, 502, 32 },
		{ 176, 503, 32 },
		{ 176, 504, 32 },
		{ 176, 505, 32 },
		{ 176, 506, 32 },
		{ 176, 507, 32 },
		{ 176, 508, 32 },
		{ 176, 509, 32 },
		{ 176, 510, 32 },
		{ 176, 511, 32 },
		{ 176, 512, 32 },
		{ 177, 471, 32 },
		{ 177, 472, 32 },
		{ 177, 473, 32 },
		{ 177, 474, 32 },
		{ 177, 475, 32 },
		{ 177, 476, 32 },
		{ 177, 477, 32 },
		{ 177, 478, 32 },
		{ 177, 479, 32 },
		{ 177, 480, 32 },
		{ 177, 481, 32 },
		{ 177, 482, 32 },
		{ 177, 483, 32 },
		{ 177, 484, 32 },
		{ 177, 485, 32 },
		{ 177, 486, 32 },
		{ 177, 487, 32 },
		{ 177, 488, 32 },
		{ 177, 489, 32 },
		{ 177, 490, 32 },
		{ 177, 491, 32 },
		{ 177, 492, 32 },
		{ 177, 493, 32 },
		{ 177, 494, 32 },
		{ 177, 495, 32 },
		{ 177, 496, 32 },
		{ 177, 497, 32 },
		{ 177, 498, 32 },
		{ 177, 499, 32 },
		{ 177, 500, 32 },
		{ 177, 501, 32 },
		{ 177, 502, 32 },
		{ 177, 503, 32 },
		{ 177, 504, 32 },
		{ 177, 505, 32 },
		{ 177, 506, 32 },
		{ 177, 507, 32 },
		{ 177, 508, 32 },
		{ 177, 509, 32 },
		{ 177, 510, 32 },
		{ 177, 511, 32 },
		{ 177, 512, 32 },
		{ 178, 468, 32 },
		{ 178, 469, 32 },
		{ 178, 470, 32 },
		{ 178, 471, 32 },
		{ 178, 472, 32 },
		{ 178, 473, 32 },
		{ 178, 474, 32 },
		{ 178, 475, 32 },
		{ 178, 476, 32 },
		{ 178, 477, 32 },
		{ 178, 478, 32 },
		{ 178, 479, 32 },
		{ 178, 480, 32 },
		{ 178, 481, 32 },
		{ 178, 482, 32 },
		{ 178, 483, 32 },
		{ 178, 484, 32 },
		{ 178, 485, 32 },
		{ 178, 486, 32 },
		{ 178, 487, 32 },
		{ 178, 488, 32 },
		{ 178, 489, 32 },
		{ 178, 490, 32 },
		{ 178, 491, 32 },
		{ 178, 492, 32 },
		{ 178, 493, 32 },
		{ 178, 494, 32 },
		{ 178, 495, 32 },
		{ 178, 496, 32 },
		{ 178, 497, 32 },
		{ 178, 498, 32 },
		{ 178, 499, 32 },
		{ 178, 500, 32 },
		{ 178, 501, 32 },
		{ 178, 502, 32 },
		{ 178, 503, 32 },
		{ 178, 504, 32 },
		{ 178, 505, 32 },
		{ 178, 506, 32 },
		{ 178, 507, 32 },
		{ 178, 508, 32 },
		{ 178, 509, 32 },
		{ 178, 510, 32 },
		{ 178, 511, 32 },
		{ 178, 512, 32 },
		{ 179, 466, 32 },
		{ 179, 467, 32 },
		{ 179, 468, 32 },
		{ 179, 469, 32 },
		{ 179, 470, 32 },
		{ 179, 471, 32 },
		{ 179, 472, 32 },
		{ 179, 473, 32 },
		{ 179, 474, 32 },
		{ 179, 475, 32 },
		{ 179, 476, 32 },
		{ 179, 477, 32 },
		{ 179, 478, 32 },
		{ 179, 479, 32 },
		{ 179, 480, 32 },
		{ 179, 481, 32 },
		{ 179, 482, 32 },
		{ 179, 483, 32 },
		{ 179, 484, 32 },
		{ 179, 485, 32 },
		{ 179, 486, 32 },
		{ 179, 487, 32 },
		{ 179, 488, 32 },
		{ 179, 489, 32 },
		{ 179, 490, 32 },
		{ 179, 491, 32 },
		{ 179, 492, 32 },
		{ 179, 493, 32 },
		{ 179, 494, 32 },
		{ 179, 495, 32 },
		{ 179, 496, 32 },
		{ 179, 497, 32 },
		{ 179, 498, 32 },
		{ 179, 499, 32 },
		{ 179, 500, 32 },
		{ 179, 501, 32 },
		{ 179, 502, 32 },
		{ 179, 503, 32 },
		{ 179, 504, 32 },
		{ 179, 505, 32 },
		{ 179, 506, 32 },
		{ 179, 507, 32 },
		{ 179, 508, 32 },
		{ 179, 509, 32 },
		{ 179, 510, 32 },
		{ 179, 511, 32 },
		{ 179, 512, 32 },
		{ 180, 463, 32 },
		{ 180, 464, 32 },
		{ 180, 465, 32 },
		{ 180, 466, 32 },
		{ 180, 467, 32 },
		{ 180, 468, 32 },
		{ 180, 469, 32 },
		{ 180, 470, 32 },
		{ 180, 471, 32 },
		{ 180, 472, 32 },
		{ 180, 473, 32 },
		{ 180, 474, 32 },
		{ 180, 475, 32 },
		{ 180, 476, 32 },
		{ 180, 477, 32 },
		{ 180, 478, 32 },
		{ 180, 479, 32 },
		{ 180, 480, 32 },
		{ 180, 481, 32 },
		{ 180, 482, 32 },
		{ 180, 483, 32 },
		{ 180, 484, 32 },
		{ 180, 485, 32 },
		{ 180, 486, 32 },
		{ 180, 487, 32 },
		{ 180, 488, 32 },
		{ 180, 489, 32 },
		{ 180, 490, 32 },
		{ 180, 491, 32 },
		{ 180, 492, 32 },
		{ 180, 493, 32 },
		{ 180, 494, 32 },
		{ 180, 495, 32 },
		{ 180, 496, 32 },
		{ 180, 497, 32 },
		{ 180, 498, 32 },
		{ 180, 499, 32 },
		{ 180, 500, 32 },
		{ 180, 501, 32 },
		{ 180, 502, 32 },
		{ 180, 503, 32 },
		{ 180, 504, 32 },
		{ 180, 505, 32 },
		{ 180, 506, 32 },
		{ 180, 507, 32 },
		{ 180, 508, 32 },
		{ 180, 509, 32 },
		{ 180, 510, 32 },
		{ 180, 511, 32 },
		{ 180, 512, 32 },
		{ 181, 460, 32 },
		{ 181, 461, 32 },
		{ 181, 462, 32 },
		{ 181, 463, 32 },
		{ 181, 464, 32 },
		{ 181, 465, 32 },
		{ 181, 466, 32 },
		{ 181, 467, 32 },
		{ 181, 468, 32 },
		{ 181, 469, 32 },
		{ 181, 470, 32 },
		{ 181, 471, 32 },
		{ 181, 472, 32 },
		{ 181, 473, 32 },
		{ 181, 474, 32 },
		{ 181, 475, 32 },
		{ 181, 476, 32 },
		{ 181, 477, 32 },
		{ 181, 478, 32 },
		{ 181, 479, 32 },
		{ 181, 480, 32 },
		{ 181, 481, 32 },
		{ 181, 482, 32 },
		{ 181, 483, 32 },
		{ 181, 484, 32 },
		{ 181, 485, 32 },
		{ 181, 486, 32 },
		{ 181, 487, 32 },
		{ 181, 488, 32 },
		{ 181, 489, 32 },
		{ 181, 490, 32 },
		{ 181, 491, 32 },
		{ 181, 492, 32 },
		{ 181, 493, 32 },
		{ 181, 494, 32 },
		{ 181, 495, 32 },
		{ 181, 496, 32 },
		{ 181, 497, 32 },
		{ 181, 498, 32 },
		{ 181, 499, 32 },
		{ 181, 500, 32 },
		{ 181, 501, 32 },
		{ 181, 502, 32 },
		{ 181, 503, 32 },
		{ 181, 504, 32 },
		{ 181, 505, 32 },
		{ 181, 506, 32 },
		{ 181, 507, 32 },
		{ 181, 508, 32 },
		{ 181, 509, 32 },
		{ 181, 510, 32 },
		{ 181, 511, 32 },
		{ 181, 512, 32 },
		{ 182, 458, 32 },
		{ 182, 459, 32 },
		{ 182, 460, 32 },
		{ 182, 461, 32 },
		{ 182, 462, 32 },
		{ 182, 463, 32 },
		{ 182, 464, 32 },
		{ 182, 465, 32 },
		{ 182, 466, 32 },
		{ 182, 467, 32 },
		{ 182, 468, 32 },
		{ 182, 469, 32 },
		{ 182, 470, 32 },
		{ 182, 471, 32 },
		{ 182, 472, 32 },
		{ 182, 473, 32 },
		{ 182, 474, 32 },
		{ 182, 475, 32 },
		{ 182, 476, 32 },
		{ 182, 477, 32 },
		{ 182, 478, 32 },
		{ 182, 479, 32 },
		{ 182, 480, 32 },
		{ 182, 481, 32 },
		{ 182, 482, 32 },
		{ 182, 483, 32 },
		{ 182, 484, 32 },
		{ 182, 485, 32 },
		{ 182, 486, 32 },
		{ 182, 487, 32 },
		{ 182, 488, 32 },
		{ 182, 489, 32 },
		{ 182, 490, 32 },
		{ 182, 491, 32 },
		{ 182, 492, 32 },
		{ 182, 493, 32 },
		{ 182, 494, 32 },
		{ 182, 495, 32 },
		{ 182, 496, 32 },
		{ 182, 497, 32 },
		{ 182, 498, 32 },
		{ 182, 499, 32 },
		{ 182, 500, 32 },
		{ 182, 501, 32 },
		{ 182, 502, 32 },
		{ 182, 503, 32 },
		{ 182, 504, 32 },
		{ 182, 505, 32 },
		{ 182, 506, 32 },
		{ 182, 507, 32 },
		{ 182, 508, 32 },
		{ 182, 509, 32 },
		{ 182, 510, 32 },
		{ 182, 511, 32 },
		{ 182, 512, 32 },
		{ 183, 455, 32 },
		{ 183, 456, 32 },
		{ 183, 457, 32 },
		{ 183, 458, 32 },
		{ 183, 459, 32 },
		{ 183, 460, 32 },
		{ 183, 461, 32 },
		{ 183, 462, 32 },
		{ 183, 463, 32 },
		{ 183, 464, 32 },
		{ 183, 465, 32 },
		{ 183, 466, 32 },
		{ 183, 467, 32 },
		{ 183, 468, 32 },
		{ 183, 469, 32 },
		{ 183, 470, 32 },
		{ 183, 471, 32 },
		{ 183, 472, 32 },
		{ 183, 473, 32 },
		{ 183, 474, 32 },
		{ 183, 475, 32 },
		{ 183, 476, 32 },
		{ 183, 477, 32 },
		{ 183, 478, 32 },
		{ 183, 479, 32 },
		{ 183, 480, 32 },
		{ 183, 481, 32 },
		{ 183, 482, 32 },
		{ 183, 483, 32 },
		{ 183, 484, 32 },
		{ 183, 485, 32 },
		{ 183, 486, 32 },
		{ 183, 487, 32 },
		{ 183, 488, 32 },
		{ 183, 489, 32 },
		{ 183, 490, 32 },
		{ 183, 491, 32 },
		{ 183, 492, 32 },
		{ 183, 493, 32 },
		{ 183, 494, 32 },
		{ 183, 495, 32 },
		{ 183, 496, 32 },
		{ 183, 497, 32 },
		{ 183, 498, 32 },
		{ 183, 499, 32 },
		{ 183, 500, 32 },
		{ 183, 501, 32 },
		{ 183, 502, 32 },
		{ 183, 503, 32 },
		{ 183, 504, 32 },
		{ 183, 505, 32 },
		{ 183, 506, 32 },
		{ 183, 507, 32 },
		{ 183, 508, 32 },
		{ 183, 509, 32 },
		{ 183, 510, 32 },
		{ 183, 511, 32 },
		{ 183, 512, 32 },
		{ 184, 452, 32 },
		{ 184, 453, 32 },
		{ 184, 454, 32 },
		{ 184, 455, 32 },
		{ 184, 456, 32 },
		{ 184, 457, 32 },
		{ 184, 458, 32 },
		{ 184, 459, 32 },
		{ 184, 460, 32 },
		{ 184, 461, 32 },
		{ 184, 462, 32 },
		{ 184, 463, 32 },
		{ 184, 464, 32 },
		{ 184, 465, 32 },
		{ 184, 466, 32 },
		{ 184, 467, 32 },
		{ 184, 468, 32 },
		{ 184, 469, 32 },
		{ 184, 470, 32 },
		{ 184, 471, 32 },
		{ 184, 472, 32 },
		{ 184, 473, 32 },
		{ 184, 474, 32 },
		{ 184, 475, 32 },
		{ 184, 476, 32 },
		{ 184, 477, 32 },
		{ 184, 478, 32 },
		{ 184, 479, 32 },
		{ 184, 480, 32 },
		{ 184, 481, 32 },
		{ 184, 482, 32 },
		{ 184, 483, 32 },
		{ 184, 484, 32 },
		{ 184, 485, 32 },
		{ 184, 486, 32 },
		{ 184, 487, 32 },
		{ 184, 488, 32 },
		{ 184, 489, 32 },
		{ 184, 490, 32 },
		{ 184, 491, 32 },
		{ 184, 492, 32 },
		{ 184, 493, 32 },
		{ 184, 494, 32 },
		{ 184, 495, 32 },
		{ 184, 496, 32 },
		{ 184, 497, 32 },
		{ 184, 498, 32 },
		{ 184, 499, 32 },
		{ 184, 500, 32 },
		{ 184, 501, 32 },
		{ 184, 502, 32 },
		{ 184, 503, 32 },
		{ 184, 504, 32 },
		{ 184, 505, 32 },
		{ 184, 506, 32 },
		{ 184, 507, 32 },
		{ 184, 508, 32 },
		{ 184, 509, 32 },
		{ 184, 510, 32 },
		{ 184, 511, 32 },
		{ 184, 512, 32 },
		{ 185, 450, 32 },
		{ 185, 451, 32 },
		{ 185, 452, 32 },
		{ 185, 453, 32 },
		{ 185, 454, 32 },
		{ 185, 455, 32 },
		{ 185, 456, 32 },
		{ 185, 457, 32 },
		{ 185, 458, 32 },
		{ 185, 459, 32 },
		{ 185, 460, 32 },
		{ 185, 461, 32 },
		{ 185, 462, 32 },
		{ 185, 463, 32 },
		{ 185, 464, 32 },
		{ 185, 465, 32 },
		{ 185, 466, 32 },
		{ 185, 467, 32 },
		{ 185, 468, 32 },
		{ 185, 469, 32 },
		{ 185, 470, 32 },
		{ 185, 471, 32 },
		{ 185, 472, 32 },
		{ 185, 473, 32 },
		{ 185, 474, 32 },
		{ 185, 475, 32 },
		{ 185, 476, 32 },
		{ 185, 477, 32 },
		{ 185, 478, 32 },
		{ 185, 479, 32 },
		{ 185, 480, 32 },
		{ 185, 481, 32 },
		{ 185, 482, 32 },
		{ 185, 483, 32 },
		{ 185, 484, 32 },
		{ 185, 485, 32 },
		{ 185, 486, 32 },
		{ 185, 487, 32 },
		{ 185, 488, 32 },
		{ 185, 489, 32 },
		{ 185, 490, 32 },
		{ 185, 491, 32 },
		{ 185, 492, 32 },
		{ 185, 493, 32 },
		{ 185, 494, 32 },
		{ 185, 495, 32 },
		{ 185, 496, 32 },
		{ 185, 497, 32 },
		{ 185, 498, 32 },
		{ 185, 499, 32 },
		{ 185, 500, 32 },
		{ 185, 501, 32 },
		{ 185, 502, 32 },
		{ 185, 503, 32 },
		{ 185, 504, 32 },
		{ 185, 505, 32 },
		{ 185, 506, 32 },
		{ 185, 507, 32 },
		{ 185, 508, 32 },
		{ 185, 509, 32 },
		{ 185, 510, 32 },
		{ 185, 511, 32 },
		{ 185, 512, 32 },
		{ 186, 447, 32 },
		{ 186, 448, 32 },
		{ 186, 449, 32 },
		{ 186, 450, 32 },
		{ 186, 451, 32 },
		{ 186, 452, 32 },
		{ 186, 453, 32 },
		{ 186, 454, 32 },
		{ 186, 455, 32 },
		{ 186, 456, 32 },
		{ 186, 457, 32 },
		{ 186, 458, 32 },
		{ 186, 459, 32 },
		{ 186, 460, 32 },
		{ 186, 461, 32 },
		{ 186, 462, 32 },
		{ 186, 463, 32 },
		{ 186, 464, 32 },
		{ 186, 465, 32 },
		{ 186, 466, 32 },
		{ 186, 467, 32 },
		{ 186, 468, 32 },
		{ 186, 469, 32 },
		{ 186, 470, 32 },
		{ 186, 471, 32 },
		{ 186, 472, 32 },
		{ 186, 473, 32 },
		{ 186, 474, 32 },
		{ 186, 475, 32 },
		{ 186, 476, 32 },
		{ 186, 477, 32 },
		{ 186, 478, 32 },
		{ 186, 479, 32 },
		{ 186, 480, 32 },
		{ 186, 481, 32 },
		{ 186, 482, 32 },
		{ 186, 483, 32 },
		{ 186, 484, 32 },
		{ 186, 485, 32 },
		{ 186, 486, 32 },
		{ 186, 487, 32 },
		{ 186, 488, 32 },
		{ 186, 489, 32 },
		{ 186, 490, 32 },
		{ 186, 491, 32 },
		{ 186, 492, 32 },
		{ 186, 493, 32 },
		{ 186, 494, 32 },
		{ 186, 495, 32 },
		{ 186, 496, 32 },
		{ 186, 497, 32 },
		{ 186, 498, 32 },
		{ 186, 499, 32 },
		{ 186, 500, 32 },
		{ 186, 501, 32 },
		{ 186, 502, 32 },
		{ 186, 503, 32 },
		{ 186, 504, 32 },
		{ 186, 505, 32 },
		{ 186, 506, 32 },
		{ 186, 507, 32 },
		{ 186, 508, 32 },
		{ 186, 509, 32 },
		{ 186, 510, 32 },
		{ 186, 511, 32 },
		{ 186, 512, 32 },
		{ 187, 444, 32 },
		{ 187, 445, 32 },
		{ 187, 446, 32 },
		{ 187, 447, 32 },
		{ 187, 448, 32 },
		{ 187, 449, 32 },
		{ 187, 450, 32 },
		{ 187, 451, 32 },
		{ 187, 452, 32 },
		{ 187, 453, 32 },
		{ 187, 454, 32 },
		{ 187, 455, 32 },
		{ 187, 456, 32 },
		{ 187, 457, 32 },
		{ 187, 458, 32 },
		{ 187, 459, 32 },
		{ 187, 460, 32 },
		{ 187, 461, 32 },
		{ 187, 462, 32 },
		{ 187, 463, 32 },
		{ 187, 464, 32 },
		{ 187, 465, 32 },
		{ 187, 466, 32 },
		{ 187, 467, 32 },
		{ 187, 468, 32 },
		{ 187, 469, 32 },
		{ 187, 470, 32 },
		{ 187, 471, 32 },
		{ 187, 472, 32 },
		{ 187, 473, 32 },
		{ 187, 474, 32 },
		{ 187, 475, 32 },
		{ 187, 476, 32 },
		{ 187, 477, 32 },
		{ 187, 478, 32 },
		{ 187, 479, 32 },
		{ 187, 480, 32 },
		{ 187, 481, 32 },
		{ 187, 482, 32 },
		{ 187, 483, 32 },
		{ 187, 484, 32 },
		{ 187, 485, 32 },
		{ 187, 486, 32 },
		{ 187, 487, 32 },
		{ 187, 488, 32 },
		{ 187, 489, 32 },
		{ 187, 490, 32 },
		{ 187, 491, 32 },
		{ 187, 492, 32 },
		{ 187, 493, 32 },
		{ 187, 494, 32 },
		{ 187, 495, 32 },
		{ 187, 496, 32 },
		{ 187, 497, 32 },
		{ 187, 498, 32 },
		{ 187, 499, 32 },
		{ 187, 500, 32 },
		{ 187, 501, 32 },
		{ 187, 502, 32 },
		{ 187, 503, 32 },
		{ 187, 504, 32 },
		{ 187, 505, 32 },
		{ 187, 506, 32 },
		{ 187, 507, 32 },
		{ 187, 508, 32 },
		{ 187, 509, 32 },
		{ 187, 510, 32 },
		{ 187, 511, 32 },
		{ 187, 512, 32 },
		{ 188, 442, 32 },
		{ 188, 443, 32 },
		{ 188, 444, 32 },
		{ 188, 445, 32 },
		{ 188, 446, 32 },
		{ 188, 447, 32 },
		{ 188, 448, 32 },
		{ 188, 449, 32 },
		{ 188, 450, 32 },
		{ 188, 451, 32 },
		{ 188, 452, 32 },
		{ 188, 453, 32 },
		{ 188, 454, 32 },
		{ 188, 455, 32 },
		{ 188, 456, 32 },
		{ 188, 457, 32 },
		{ 188, 458, 32 },
		{ 188, 459, 32 },
		{ 188, 460, 32 },
		{ 188, 461, 32 },
		{ 188, 462, 32 },
		{ 188, 463, 32 },
		{ 188, 464, 32 },
		{ 188, 465, 32 },
		{ 188, 466, 32 },
		{ 188, 467, 32 },
		{ 188, 468, 32 },
		{ 188, 469, 32 },
		{ 188, 470, 32 },
		{ 188, 471, 32 },
		{ 188, 472, 32 },
		{ 188, 473, 32 },
		{ 188, 474, 32 },
		{ 188, 475, 32 },
		{ 188, 476, 32 },
		{ 188, 477, 32 },
		{ 188, 478, 32 },
		{ 188, 479, 32 },
		{ 188, 480, 32 },
		{ 188, 481, 32 },
		{ 188, 482, 32 },
		{ 188, 483, 32 },
		{ 188, 484, 32 },
		{ 188, 485, 32 },
		{ 188, 486, 32 },
		{ 188, 487, 32 },
		{ 188, 488, 32 },
		{ 188, 489, 32 },
		{ 188, 490, 32 },
		{ 188, 491, 32 },
		{ 188, 492, 32 },
		{ 188, 493, 32 },
		{ 188, 494, 32 },
		{ 188, 495, 32 },
		{ 188, 496, 32 },
		{ 188, 497, 32 },
		{ 188, 498, 32 },
		{ 188, 499, 32 },
		{ 188, 500, 32 },
		{ 188, 501, 32 },
		{ 188, 502, 32 },
		{ 188, 503, 32 },
		{ 188, 504, 32 },
		{ 188, 505, 32 },
		{ 188, 506, 32 },
		{ 188, 507, 32 },
		{ 188, 508, 32 },
		{ 188, 509, 32 },
		{ 188, 510, 32 },
		{ 188, 511, 32 },
		{ 188, 512, 32 },
		{ 189, 439, 32 },
		{ 189, 440, 32 },
		{ 189, 441, 32 },
		{ 189, 442, 32 },
		{ 189, 443, 32 },
		{ 189, 444, 32 },
		{ 189, 445, 32 },
		{ 189, 446, 32 },
		{ 189, 447, 32 },
		{ 189, 448, 32 },
		{ 189, 449, 32 },
		{ 189, 450, 32 },
		{ 189, 451, 32 },
		{ 189, 452, 32 },
		{ 189, 453, 32 },
		{ 189, 454, 32 },
		{ 189, 455, 32 },
		{ 189, 456, 32 },
		{ 189, 457, 32 },
		{ 189, 458, 32 },
		{ 189, 459, 32 },
		{ 189, 460, 32 },
		{ 189, 461, 32 },
		{ 189, 462, 32 },
		{ 189, 463, 32 },
		{ 189, 464, 32 },
		{ 189, 465, 32 },
		{ 189, 466, 32 },
		{ 189, 467, 32 },
		{ 189, 468, 32 },
		{ 189, 469, 32 },
		{ 189, 470, 32 },
		{ 189, 471, 32 },
		{ 189, 472, 32 },
		{ 189, 473, 32 },
		{ 189, 474, 32 },
		{ 189, 475, 32 },
		{ 189, 476, 32 },
		{ 189, 477, 32 },
		{ 189, 478, 32 },
		{ 189, 479, 32 },
		{ 189, 480, 32 },
		{ 189, 481, 32 },
		{ 189, 482, 32 },
		{ 189, 483, 32 },
		{ 189, 484, 32 },
		{ 189, 485, 32 },
		{ 189, 486, 32 },
		{ 189, 487, 32 },
		{ 189, 488, 32 },
		{ 189, 489, 32 },
		{ 189, 490, 32 },
		{ 189, 491, 32 },
		{ 189, 492, 32 },
		{ 189, 493, 32 },
		{ 189, 494, 32 },
		{ 189, 495, 32 },
		{ 189, 496, 32 },
		{ 189, 497, 32 },
		{ 189, 498, 32 },
		{ 189, 499, 32 },
		{ 189, 500, 32 },
		{ 189, 501, 32 },
		{ 189, 502, 32 },
		{ 189, 503, 32 },
		{ 189, 504, 32 },
		{ 189, 505, 32 },
		{ 189, 506, 32 },
		{ 189, 507, 32 },
		{ 189, 508, 32 },
		{ 189, 509, 32 },
		{ 189, 510, 32 },
		{ 189, 511, 32 },
		{ 189, 512, 32 },
		{ 190, 436, 32 },
		{ 190, 437, 32 },
		{ 190, 438, 32 },
		{ 190, 439, 32 },
		{ 190, 440, 32 },
		{ 190, 441, 32 },
		{ 190, 442, 32 },
		{ 190, 443, 32 },
		{ 190, 444, 32 },
		{ 190, 445, 32 },
		{ 190, 446, 32 },
		{ 190, 447, 32 },
		{ 190, 448, 32 },
		{ 190, 449, 32 },
		{ 190, 450, 32 },
		{ 190, 451, 32 },
		{ 190, 452, 32 },
		{ 190, 453, 32 },
		{ 190, 454, 32 },
		{ 190, 455, 32 },
		{ 190, 456, 32 },
		{ 190, 457, 32 },
		{ 190, 458, 32 },
		{ 190, 459, 32 },
		{ 190, 460, 32 },
		{ 190, 461, 32 },
		{ 190, 462, 32 },
		{ 190, 463, 32 },
		{ 190, 464, 32 },
		{ 190, 465, 32 },
		{ 190, 466, 32 },
		{ 190, 467, 32 },
		{ 190, 468, 32 },
		{ 190, 469, 32 },
		{ 190, 470, 32 },
		{ 190, 471, 32 },
		{ 190, 472, 32 },
		{ 190, 473, 32 },
		{ 190, 474, 32 },
		{ 190, 475, 32 },
		{ 190, 476, 32 },
		{ 190, 477, 32 },
		{ 190, 478, 32 },
		{ 190, 479, 32 },
		{ 190, 480, 32 },
		{ 190, 481, 32 },
		{ 190, 482, 32 },
		{ 190, 483, 32 },
		{ 190, 484, 32 },
		{ 190, 485, 32 },
		{ 190, 486, 32 },
		{ 190, 487, 32 },
		{ 190, 488, 32 },
		{ 190, 489, 32 },
		{ 190, 490, 32 },
		{ 190, 491, 32 },
		{ 190, 492, 32 },
		{ 190, 493, 32 },
		{ 190, 494, 32 },
		{ 190, 495, 32 },
		{ 190, 496, 32 },
		{ 190, 497, 32 },
		{ 190, 498, 32 },
		{ 190, 499, 32 },
		{ 190, 500, 32 },
		{ 190, 501, 32 },
		{ 190, 502, 32 },
		{ 190, 503, 32 },
		{ 190, 504, 32 },
		{ 190, 505, 32 },
		{ 190, 506, 32 },
		{ 190, 507, 32 },
		{ 190, 508, 32 },
		{ 190, 509, 32 },
		{ 190, 510, 32 },
		{ 190, 511, 32 },
		{ 190, 512, 32 },
		{ 191, 434, 32 },
		{ 191, 435, 32 },
		{ 191, 436, 32 },
		{ 191, 437, 32 },
		{ 191, 438, 32 },
		{ 191, 439, 32 },
		{ 191, 440, 32 },
		{ 191, 441, 32 },
		{ 191, 442, 32 },
		{ 191, 443, 32 },
		{ 191, 444, 32 },
		{ 191, 445, 32 },
		{ 191, 446, 32 },
		{ 191, 447, 32 },
		{ 191, 448, 32 },
		{ 191, 449, 32 },
		{ 191, 450, 32 },
		{ 191, 451, 32 },
		{ 191, 452, 32 },
		{ 191, 453, 32 },
		{ 191, 454, 32 },
		{ 191, 455, 32 },
		{ 191, 456, 32 },
		{ 191, 457, 32 },
		{ 191, 458, 32 },
		{ 191, 459, 32 },
		{ 191, 460, 32 },
		{ 191, 461, 32 },
		{ 191, 462, 32 },
		{ 191, 463, 32 },
		{ 191, 464, 32 },
		{ 191, 465, 32 },
		{ 191, 466, 32 },
		{ 191, 467, 32 },
		{ 191, 468, 32 },
		{ 191, 469, 32 },
		{ 191, 470, 32 },
		{ 191, 471, 32 },
		{ 191, 472, 32 },
		{ 191, 473, 32 },
		{ 191, 474, 32 },
		{ 191, 475, 32 },
		{ 191, 476, 32 },
		{ 191, 477, 32 },
		{ 191, 478, 32 },
		{ 191, 479, 32 },
		{ 191, 480, 32 },
		{ 191, 481, 32 },
		{ 191, 482, 32 },
		{ 191, 483, 32 },
		{ 191, 484, 32 },
		{ 191, 485, 32 },
		{ 191, 486, 32 },
		{ 191, 487, 32 },
		{ 191, 488, 32 },
		{ 191, 489, 32 },
		{ 191, 490, 32 },
		{ 191, 491, 32 },
		{ 191, 492, 32 },
		{ 191, 493, 32 },
		{ 191, 494, 32 },
		{ 191, 495, 32 },
		{ 191, 496, 32 },
		{ 191, 497, 32 },
		{ 191, 498, 32 },
		{ 191, 499, 32 },
		{ 191, 500, 32 },
		{ 191, 501, 32 },
		{ 191, 502, 32 },
		{ 191, 503, 32 },
		{ 191, 504, 32 },
		{ 191, 505, 32 },
		{ 191, 506, 32 },
		{ 191, 507, 32 },
		{ 191, 508, 32 },
		{ 191, 509, 32 },
		{ 191, 510, 32 },
		{ 191, 511, 32 },
		{ 191, 512, 32 },
		{ 192, 431, 32 },
		{ 192, 432, 32 },
		{ 192, 433, 32 },
		{ 192, 434, 32 },
		{ 192, 435, 32 },
		{ 192, 436, 32 },
		{ 192, 437, 32 },
		{ 192, 438, 32 },
		{ 192, 439, 32 },
		{ 192, 440, 32 },
		{ 192, 441, 32 },
		{ 192, 442, 32 },
		{ 192, 443, 32 },
		{ 192, 444, 32 },
		{ 192, 445, 32 },
		{ 192, 446, 32 },
		{ 192, 447, 32 },
		{ 192, 448, 32 },
		{ 192, 449, 32 },
		{ 192, 450, 32 },
		{ 192, 451, 32 },
		{ 192, 452, 32 },
		{ 192, 453, 32 },
		{ 192, 454, 32 },
		{ 192, 455, 32 },
		{ 192, 456, 32 },
		{ 192, 457, 32 },
		{ 192, 458, 32 },
		{ 192, 459, 32 },
		{ 192, 460, 32 },
		{ 192, 461, 32 },
		{ 192, 462, 32 },
		{ 192, 463, 32 },
		{ 192, 464, 32 },
		{ 192, 465, 32 },
		{ 192, 466, 32 },
		{ 192, 467, 32 },
		{ 192, 468, 32 },
		{ 192, 469, 32 },
		{ 192, 470, 32 },
		{ 192, 471, 32 },
		{ 192, 472, 32 },
		{ 192, 473, 32 },
		{ 192, 474, 32 },
		{ 192, 475, 32 },
		{ 192, 476, 32 },
		{ 192, 477, 32 },
		{ 192, 478, 32 },
		{ 192, 479, 32 },
		{ 192, 480, 32 },
		{ 192, 481, 32 },
		{ 192, 482, 32 },
		{ 192, 483, 32 },
		{ 192, 484, 32 },
		{ 192, 485, 32 },
		{ 192, 486, 32 },
		{ 192, 487, 32 },
		{ 192, 488, 32 },
		{ 192, 489, 32 },
		{ 192, 490, 32 },
		{ 192, 491, 32 },
		{ 192, 492, 32 },
		{ 192, 493, 32 },
		{ 192, 494, 32 },
		{ 192, 495, 32 },
		{ 192, 496, 32 },
		{ 192, 497, 32 },
		{ 192, 498, 32 },
		{ 192, 499, 32 },
		{ 192, 500, 32 },
		{ 192, 501, 32 },
		{ 192, 502, 32 },
		{ 192, 503, 32 },
		{ 192, 504, 32 },
		{ 192, 505, 32 },
		{ 192, 506, 32 },
		{ 192, 507, 32 },
		{ 192, 508, 32 },
		{ 192, 509, 32 },
		{ 192, 510, 32 },
		{ 192, 511, 32 },
		{ 192, 512, 32 },
		{ 193, 428, 32 },
		{ 193, 429, 32 },
		{ 193, 430, 32 },
		{ 193, 431, 32 },
		{ 193, 432, 32 },
		{ 193, 433, 32 },
		{ 193, 434, 32 },
		{ 193, 435, 32 },
		{ 193, 436, 32 },
		{ 193, 437, 32 },
		{ 193, 438, 32 },
		{ 193, 439, 32 },
		{ 193, 440, 32 },
		{ 193, 441, 32 },
		{ 193, 442, 32 },
		{ 193, 443, 32 },
		{ 193, 444, 32 },
		{ 193, 445, 32 },
		{ 193, 446, 32 },
		{ 193, 447, 32 },
		{ 193, 448, 32 },
		{ 193, 449, 32 },
		{ 193, 450, 32 },
		{ 193, 451, 32 },
		{ 193, 452, 32 },
		{ 193, 453, 32 },
		{ 193, 454, 32 },
		{ 193, 455, 32 },
		{ 193, 456, 32 },
		{ 193, 457, 32 },
		{ 193, 458, 32 },
		{ 193, 459, 32 },
		{ 193, 460, 32 },
		{ 193, 461, 32 },
		{ 193, 462, 32 },
		{ 193, 463, 32 },
		{ 193, 464, 32 },
		{ 193, 465, 32 },
		{ 193, 466, 32 },
		{ 193, 467, 32 },
		{ 193, 468, 32 },
		{ 193, 469, 32 },
		{ 193, 470, 32 },
		{ 193, 471, 32 },
		{ 193, 472, 32 },
		{ 193, 473, 32 },
		{ 193, 474, 32 },
		{ 193, 475, 32 },
		{ 193, 476, 32 },
		{ 193, 477, 32 },
		{ 193, 478, 32 },
		{ 193, 479, 32 },
		{ 193, 480, 32 },
		{ 193, 481, 32 },
		{ 193, 482, 32 },
		{ 193, 483, 32 },
		{ 193, 484, 32 },
		{ 193, 485, 32 },
		{ 193, 486, 32 },
		{ 193, 487, 32 },
		{ 193, 488, 32 },
		{ 193, 489, 32 },
		{ 193, 490, 32 },
		{ 193, 491, 32 },
		{ 193, 492, 32 },
		{ 193, 493, 32 },
		{ 193, 494, 32 },
		{ 193, 495, 32 },
		{ 193, 496, 32 },
		{ 193, 497, 32 },
		{ 193, 498, 32 },
		{ 193, 499, 32 },
		{ 193, 500, 32 },
		{ 193, 501, 32 },
		{ 193, 502, 32 },
		{ 193, 503, 32 },
		{ 193, 504, 32 },
		{ 193, 505, 32 },
		{ 193, 506, 32 },
		{ 193, 507, 32 },
		{ 193, 508, 32 },
		{ 193, 509, 32 },
		{ 193, 510, 32 },
		{ 193, 511, 32 },
		{ 193, 512, 32 },
		{ 194, 426, 32 },
		{ 194, 427, 32 },
		{ 194, 428, 32 },
		{ 194, 429, 32 },
		{ 194, 430, 32 },
		{ 194, 431, 32 },
		{ 194, 432, 32 },
		{ 194, 433, 32 },
		{ 194, 434, 32 },
		{ 194, 435, 32 },
		{ 194, 436, 32 },
		{ 194, 437, 32 },
		{ 194, 438, 32 },
		{ 194, 439, 32 },
		{ 194, 440, 32 },
		{ 194, 441, 32 },
		{ 194, 442, 32 },
		{ 194, 443, 32 },
		{ 194, 444, 32 },
		{ 194, 445, 32 },
		{ 194, 446, 32 },
		{ 194, 447, 32 },
		{ 194, 448, 32 },
		{ 194, 449, 32 },
		{ 194, 450, 32 },
		{ 194, 451, 32 },
		{ 194, 452, 32 },
		{ 194, 453, 32 },
		{ 194, 454, 32 },
		{ 194, 455, 32 },
		{ 194, 456, 32 },
		{ 194, 457, 32 },
		{ 194, 458, 32 },
		{ 194, 459, 32 },
		{ 194, 460, 32 },
		{ 194, 461, 32 },
		{ 194, 462, 32 },
		{ 194, 463, 32 },
		{ 194, 464, 32 },
		{ 194, 465, 32 },
		{ 194, 466, 32 },
		{ 194, 467, 32 },
		{ 194, 468, 32 },
		{ 194, 469, 32 },
		{ 194, 470, 32 },
		{ 194, 471, 32 },
		{ 194, 472, 32 },
		{ 194, 473, 32 },
		{ 194, 474, 32 },
		{ 194, 475, 32 },
		{ 194, 476, 32 },
		{ 194, 477, 32 },
		{ 194, 478, 32 },
		{ 194, 479, 32 },
		{ 194, 480, 32 },
		{ 194, 481, 32 },
		{ 194, 482, 32 },
		{ 194, 483, 32 },
		{ 194, 484, 32 },
		{ 194, 485, 32 },
		{ 194, 486, 32 },
		{ 194, 487, 32 },
		{ 194, 488, 32 },
		{ 194, 489, 32 },
		{ 194, 490, 32 },
		{ 194, 491, 32 },
		{ 194, 492, 32 },
		{ 194, 493, 32 },
		{ 194, 494, 32 },
		{ 194, 495, 32 },
		{ 194, 496, 32 },
		{ 194, 497, 32 },
		{ 194, 498, 32 },
		{ 194, 499, 32 },
		{ 194, 500, 32 },
		{ 194, 501, 32 },
		{ 194, 502, 32 },
		{ 194, 503, 32 },
		{ 194, 504, 32 },
		{ 194, 505, 32 },
		{ 194, 506, 32 },
		{ 194, 507, 32 },
		{ 194, 508, 32 },
		{ 194, 509, 32 },
		{ 194, 510, 32 },
		{ 194, 511, 32 },
		{ 194, 512, 32 },
		{ 195, 423, 32 },
		{ 195, 424, 32 },
		{ 195, 425, 32 },
		{ 195, 426, 32 },
		{ 195, 427, 32 },
		{ 195, 428, 32 },
		{ 195, 429, 32 },
		{ 195, 430, 32 },
		{ 195, 431, 32 },
		{ 195, 432, 32 },
		{ 195, 433, 32 },
		{ 195, 434, 32 },
		{ 195, 435, 32 },
		{ 195, 436, 32 },
		{ 195, 437, 32 },
		{ 195, 438, 32 },
		{ 195, 439, 32 },
		{ 195, 440, 32 },
		{ 195, 441, 32 },
		{ 195, 442, 32 },
		{ 195, 443, 32 },
		{ 195, 444, 32 },
		{ 195, 445, 32 },
		{ 195, 446, 32 },
		{ 195, 447, 32 },
		{ 195, 448, 32 },
		{ 195, 449, 32 },
		{ 195, 450, 32 },
		{ 195, 451, 32 },
		{ 195, 452, 32 },
		{ 195, 453, 32 },
		{ 195, 454, 32 },
		{ 195, 455, 32 },
		{ 195, 456, 32 },
		{ 195, 457, 32 },
		{ 195, 458, 32 },
		{ 195, 459, 32 },
		{ 195, 460, 32 },
		{ 195, 461, 32 },
		{ 195, 462, 32 },
		{ 195, 463, 32 },
		{ 195, 464, 32 },
		{ 195, 465, 32 },
		{ 195, 466, 32 },
		{ 195, 467, 32 },
		{ 195, 468, 32 },
		{ 195, 469, 32 },
		{ 195, 470, 32 },
		{ 195, 471, 32 },
		{ 195, 472, 32 },
		{ 195, 473, 32 },
		{ 195, 474, 32 },
		{ 195, 475, 32 },
		{ 195, 476, 32 },
		{ 195, 477, 32 },
		{ 195, 478, 32 },
		{ 195, 479, 32 },
		{ 195, 480, 32 },
		{ 195, 481, 32 },
		{ 195, 482, 32 },
		{ 195, 483, 32 },
		{ 195, 484, 32 },
		{ 195, 485, 32 },
		{ 195, 486, 32 },
		{ 195, 487, 32 },
		{ 195, 488, 32 },
		{ 195, 489, 32 },
		{ 195, 490, 32 },
		{ 195, 491, 32 },
		{ 195, 492, 32 },
		{ 195, 493, 32 },
		{ 195, 494, 32 },
		{ 195, 495, 32 },
		{ 195, 496, 32 },
		{ 195, 497, 32 },
		{ 195, 498, 32 },
		{ 195, 499, 32 },
		{ 195, 500, 32 },
		{ 195, 501, 32 },
		{ 195, 502, 32 },
		{ 195, 503, 32 },
		{ 195, 504, 32 },
		{ 195, 505, 32 },
		{ 195, 506, 32 },
		{ 195, 507, 32 },
		{ 195, 508, 32 },
		{ 195, 509, 32 },
		{ 195, 510, 32 },
		{ 195, 511, 32 },
		{ 195, 512, 32 },
		{ 196, 420, 32 },
		{ 196, 421, 32 },
		{ 196, 422, 32 },
		{ 196, 423, 32 },
		{ 196, 424, 32 },
		{ 196, 425, 32 },
		{ 196, 426, 32 },
		{ 196, 427, 32 },
		{ 196, 428, 32 },
		{ 196, 429, 32 },
		{ 196, 430, 32 },
		{ 196, 431, 32 },
		{ 196, 432, 32 },
		{ 196, 433, 32 },
		{ 196, 434, 32 },
		{ 196, 435, 32 },
		{ 196, 436, 32 },
		{ 196, 437, 32 },
		{ 196, 438, 32 },
		{ 196, 439, 32 },
		{ 196, 440, 32 },
		{ 196, 441, 32 },
		{ 196, 442, 32 },
		{ 196, 443, 32 },
		{ 196, 444, 32 },
		{ 196, 445, 32 },
		{ 196, 446, 32 },
		{ 196, 447, 32 },
		{ 196, 448, 32 },
		{ 196, 449, 32 },
		{ 196, 450, 32 },
		{ 196, 451, 32 },
		{ 196, 452, 32 },
		{ 196, 453, 32 },
		{ 196, 454, 32 },
		{ 196, 455, 32 },
		{ 196, 456, 32 },
		{ 196, 457, 32 },
		{ 196, 458, 32 },
		{ 196, 459, 32 },
		{ 196, 460, 32 },
		{ 196, 461, 32 },
		{ 196, 462, 32 },
		{ 196, 463, 32 },
		{ 196, 464, 32 },
		{ 196, 465, 32 },
		{ 196, 466, 32 },
		{ 196, 467, 32 },
		{ 196, 468, 32 },
		{ 196, 469, 32 },
		{ 196, 470, 32 },
		{ 196, 471, 32 },
		{ 196, 472, 32 },
		{ 196, 473, 32 },
		{ 196, 474, 32 },
		{ 196, 475, 32 },
		{ 196, 476, 32 },
		{ 196, 477, 32 },
		{ 196, 478, 32 },
		{ 196, 479, 32 },
		{ 196, 480, 32 },
		{ 196, 481, 32 },
		{ 196, 482, 32 },
		{ 196, 483, 32 },
		{ 196, 484, 32 },
		{ 196, 485, 32 },
		{ 196, 486, 32 },
		{ 196, 487, 32 },
		{ 196, 488, 32 },
		{ 196, 489, 32 },
		{ 196, 490, 32 },
		{ 196, 491, 32 },
		{ 196, 492, 32 },
		{ 196, 493, 32 },
		{ 196, 494, 32 },
		{ 196, 495, 32 },
		{ 196, 496, 32 },
		{ 196, 497, 32 },
		{ 196, 498, 32 },
		{ 196, 499, 32 },
		{ 196, 500, 32 },
		{ 196, 501, 32 },
		{ 196, 502, 32 },
		{ 196, 503, 32 },
		{ 196, 504, 32 },
		{ 196, 505, 32 },
		{ 196, 506, 32 },
		{ 196, 507, 32 },
		{ 196, 508, 32 },
		{ 196, 509, 32 },
		{ 196, 510, 32 },
		{ 196, 511, 32 },
		{ 196, 512, 32 },
		{ 197, 418, 32 },
		{ 197, 419, 32 },
		{ 197, 420, 32 },
		{ 197, 421, 32 },
		{ 197, 422, 32 },
		{ 197, 423, 32 },
		{ 197, 424, 32 },
		{ 197, 425, 32 },
		{ 197, 426, 32 },
		{ 197, 427, 32 },
		{ 197, 428, 32 },
		{ 197, 429, 32 },
		{ 197, 430, 32 },
		{ 197, 431, 32 },
		{ 197, 432, 32 },
		{ 197, 433, 32 },
		{ 197, 434, 32 },
		{ 197, 435, 32 },
		{ 197, 436, 32 },
		{ 197, 437, 32 },
		{ 197, 438, 32 },
		{ 197, 439, 32 },
		{ 197, 440, 32 },
		{ 197, 441, 32 },
		{ 197, 442, 32 },
		{ 197, 443, 32 },
		{ 197, 444, 32 },
		{ 197, 445, 32 },
		{ 197, 446, 32 },
		{ 197, 447, 32 },
		{ 197, 448, 32 },
		{ 197, 449, 32 },
		{ 197, 450, 32 },
		{ 197, 451, 32 },
		{ 197, 452, 32 },
		{ 197, 453, 32 },
		{ 197, 454, 32 },
		{ 197, 455, 32 },
		{ 197, 456, 32 },
		{ 197, 457, 32 },
		{ 197, 458, 32 },
		{ 197, 459, 32 },
		{ 197, 460, 32 },
		{ 197, 461, 32 },
		{ 197, 462, 32 },
		{ 197, 463, 32 },
		{ 197, 464, 32 },
		{ 197, 465, 32 },
		{ 197, 466, 32 },
		{ 197, 467, 32 },
		{ 197, 468, 32 },
		{ 197, 469, 32 },
		{ 197, 470, 32 },
		{ 197, 471, 32 },
		{ 197, 472, 32 },
		{ 197, 473, 32 },
		{ 197, 474, 32 },
		{ 197, 475, 32 },
		{ 197, 476, 32 },
		{ 197, 477, 32 },
		{ 197, 478, 32 },
		{ 197, 479, 32 },
		{ 197, 480, 32 },
		{ 197, 481, 32 },
		{ 197, 482, 32 },
		{ 197, 483, 32 },
		{ 197, 484, 32 },
		{ 197, 485, 32 },
		{ 197, 486, 32 },
		{ 197, 487, 32 },
		{ 197, 488, 32 },
		{ 197, 489, 32 },
		{ 197, 490, 32 },
		{ 197, 491, 32 },
		{ 197, 492, 32 },
		{ 197, 493, 32 },
		{ 197, 494, 32 },
		{ 197, 495, 32 },
		{ 197, 496, 32 },
		{ 197, 497, 32 },
		{ 197, 498, 32 },
		{ 197, 499, 32 },
		{ 197, 500, 32 },
		{ 197, 501, 32 },
		{ 197, 502, 32 },
		{ 197, 503, 32 },
		{ 197, 504, 32 },
		{ 197, 505, 32 },
		{ 197, 506, 32 },
		{ 197, 507, 32 },
		{ 197, 508, 32 },
		{ 197, 509, 32 },
		{ 197, 510, 32 },
		{ 197, 511, 32 },
		{ 197, 512, 32 },
		{ 198, 415, 32 },
		{ 198, 416, 32 },
		{ 198, 417, 32 },
		{ 198, 418, 32 },
		{ 198, 419, 32 },
		{ 198, 420, 32 },
		{ 198, 421, 32 },
		{ 198, 422, 32 },
		{ 198, 423, 32 },
		{ 198, 424, 32 },
		{ 198, 425, 32 },
		{ 198, 426, 32 },
		{ 198, 427, 32 },
		{ 198, 428, 32 },
		{ 198, 429, 32 },
		{ 198, 430, 32 },
		{ 198, 431, 32 },
		{ 198, 432, 32 },
		{ 198, 433, 32 },
		{ 198, 434, 32 },
		{ 198, 435, 32 },
		{ 198, 436, 32 },
		{ 198, 437, 32 },
		{ 198, 438, 32 },
		{ 198, 439, 32 },
		{ 198, 440, 32 },
		{ 198, 441, 32 },
		{ 198, 442, 32 },
		{ 198, 443, 32 },
		{ 198, 444, 32 },
		{ 198, 445, 32 },
		{ 198, 446, 32 },
		{ 198, 447, 32 },
		{ 198, 448, 32 },
		{ 198, 449, 32 },
		{ 198, 450, 32 },
		{ 198, 451, 32 },
		{ 198, 452, 32 },
		{ 198, 453, 32 },
		{ 198, 454, 32 },
		{ 198, 455, 32 },
		{ 198, 456, 32 },
		{ 198, 457, 32 },
		{ 198, 458, 32 },
		{ 198, 459, 32 },
		{ 198, 460, 32 },
		{ 198, 461, 32 },
		{ 198, 462, 32 },
		{ 198, 463, 32 },
		{ 198, 464, 32 },
		{ 198, 465, 32 },
		{ 198, 466, 32 },
		{ 198, 467, 32 },
		{ 198, 468, 32 },
		{ 198, 469, 32 },
		{ 198, 470, 32 },
		{ 198, 471, 32 },
		{ 198, 472, 32 },
		{ 198, 473, 32 },
		{ 198, 474, 32 },
		{ 198, 475, 32 },
		{ 198, 476, 32 },
		{ 198, 477, 32 },
		{ 198, 478, 32 },
		{ 198, 479, 32 },
		{ 198, 480, 32 },
		{ 198, 481, 32 },
		{ 198, 482, 32 },
		{ 198, 483, 32 },
		{ 198, 484, 32 },
		{ 198, 485, 32 },
		{ 198, 486, 32 },
		{ 198, 487, 32 },
		{ 198, 488, 32 },
		{ 198, 489, 32 },
		{ 198, 490, 32 },
		{ 198, 491, 32 },
		{ 198, 492, 32 },
		{ 198, 493, 32 },
		{ 198, 494, 32 },
		{ 198, 495, 32 },
		{ 198, 496, 32 },
		{ 198, 497, 32 },
		{ 198, 498, 32 },
		{ 198, 499, 32 },
		{ 198, 500, 32 },
		{ 198, 501, 32 },
		{ 198, 502, 32 },
		{ 198, 503, 32 },
		{ 198, 504, 32 },
		{ 198, 505, 32 },
		{ 198, 506, 32 },
		{ 198, 507, 32 },
		{ 198, 508, 32 },
		{ 198, 509, 32 },
		{ 198, 510, 32 },
		{ 198, 511, 32 },
		{ 198, 512, 32 },
		{ 199, 412, 32 },
		{ 199, 413, 32 },
		{ 199, 414, 32 },
		{ 199, 415, 32 },
		{ 199, 416, 32 },
		{ 199, 417, 32 },
		{ 199, 418, 32 },
		{ 199, 419, 32 },
		{ 199, 420, 32 },
		{ 199, 421, 32 },
		{ 199, 422, 32 },
		{ 199, 423, 32 },
		{ 199, 424, 32 },
		{ 199, 425, 32 },
		{ 199, 426, 32 },
		{ 199, 427, 32 },
		{ 199, 428, 32 },
		{ 199, 429, 32 },
		{ 199, 430, 32 },
		{ 199, 431, 32 },
		{ 199, 432, 32 },
		{ 199, 433, 32 },
		{ 199, 434, 32 },
		{ 199, 435, 32 },
		{ 199, 436, 32 },
		{ 199, 437, 32 },
		{ 199, 438, 32 },
		{ 199, 439, 32 },
		{ 199, 440, 32 },
		{ 199, 441, 32 },
		{ 199, 442, 32 },
		{ 199, 443, 32 },
		{ 199, 444, 32 },
		{ 199, 445, 32 },
		{ 199, 446, 32 },
		{ 199, 447, 32 },
		{ 199, 448, 32 },
		{ 199, 449, 32 },
		{ 199, 450, 32 },
		{ 199, 451, 32 },
		{ 199, 452, 32 },
		{ 199, 453, 32 },
		{ 199, 454, 32 },
		{ 199, 455, 32 },
		{ 199, 456, 32 },
		{ 199, 457, 32 },
		{ 199, 458, 32 },
		{ 199, 459, 32 },
		{ 199, 460, 32 },
		{ 199, 461, 32 },
		{ 199, 462, 32 },
		{ 199, 463, 32 },
		{ 199, 464, 32 },
		{ 199, 465, 32 },
		{ 199, 466, 32 },
		{ 199, 467, 32 },
		{ 199, 468, 32 },
		{ 199, 469, 32 },
		{ 199, 470, 32 },
		{ 199, 471, 32 },
		{ 199, 472, 32 },
		{ 199, 473, 32 },
		{ 199, 474, 32 },
		{ 199, 475, 32 },
		{ 199, 476, 32 },
		{ 199, 477, 32 },
		{ 199, 478, 32 },
		{ 199, 479, 32 },
		{ 199, 480, 32 },
		{ 199, 481, 32 },
		{ 199, 482, 32 },
		{ 199, 483, 32 },
		{ 199, 484, 32 },
		{ 199, 485, 32 },
		{ 199, 486, 32 },
		{ 199, 487, 32 },
		{ 199, 488, 32 },
		{ 199, 489, 32 },
		{ 199, 490, 32 },
		{ 199, 491, 32 },
		{ 199, 492, 32 },
		{ 199, 493, 32 },
		{ 199, 494, 32 },
		{ 199, 495, 32 },
		{ 199, 496, 32 },
		{ 199, 497, 32 },
		{ 199, 498, 32 },
		{ 199, 499, 32 },
		{ 199, 500, 32 },
		{ 199, 501, 32 },
		{ 199, 502, 32 },
		{ 199, 503, 32 },
		{ 199, 504, 32 },
		{ 199, 505, 32 },
		{ 199, 506, 32 },
		{ 199, 507, 32 },
		{ 199, 508, 32 },
		{ 199, 509, 32 },
		{ 199, 510, 32 },
		{ 199, 511, 32 },
		{ 199, 512, 32 },
		{ 200, 410, 32 },
		{ 200, 411, 32 },
		{ 200, 412, 32 },
		{ 200, 413, 32 },
		{ 200, 414, 32 },
		{ 200, 415, 32 },
		{ 200, 416, 32 },
		{ 200, 417, 32 },
		{ 200, 418, 32 },
		{ 200, 419, 32 },
		{ 200, 420, 32 },
		{ 200, 421, 32 },
		{ 200, 422, 32 },
		{ 200, 423, 32 },
		{ 200, 424, 32 },
		{ 200, 425, 32 },
		{ 200, 426, 32 },
		{ 200, 427, 32 },
		{ 200, 428, 32 },
		{ 200, 429, 32 },
		{ 200, 430, 32 },
		{ 200, 431, 32 },
		{ 200, 432, 32 },
		{ 200, 433, 32 },
		{ 200, 434, 32 },
		{ 200, 435, 32 },
		{ 200, 436, 32 },
		{ 200, 437, 32 },
		{ 200, 438, 32 },
		{ 200, 439, 32 },
		{ 200, 440, 32 },
		{ 200, 441, 32 },
		{ 200, 442, 32 },
		{ 200, 443, 32 },
		{ 200, 444, 32 },
		{ 200, 445, 32 },
		{ 200, 446, 32 },
		{ 200, 447, 32 },
		{ 200, 448, 32 },
		{ 200, 449, 32 },
		{ 200, 450, 32 },
		{ 200, 451, 32 },
		{ 200, 452, 32 },
		{ 200, 453, 32 },
		{ 200, 454, 32 },
		{ 200, 455, 32 },
		{ 200, 456, 32 },
		{ 200, 457, 32 },
		{ 200, 458, 32 },
		{ 200, 459, 32 },
		{ 200, 460, 32 },
		{ 200, 461, 32 },
		{ 200, 462, 32 },
		{ 200, 463, 32 },
		{ 200, 464, 32 },
		{ 200, 465, 32 },
		{ 200, 466, 32 },
		{ 200, 467, 32 },
		{ 200, 468, 32 },
		{ 200, 469, 32 },
		{ 200, 470, 32 },
		{ 200, 471, 32 },
		{ 200, 472, 32 },
		{ 200, 473, 32 },
		{ 200, 474, 32 },
		{ 200, 475, 32 },
		{ 200, 476, 32 },
		{ 200, 477, 32 },
		{ 200, 478, 32 },
		{ 200, 479, 32 },
		{ 200, 480, 32 },
		{ 200, 481, 32 },
		{ 200, 482, 32 },
		{ 200, 483, 32 },
		{ 200, 484, 32 },
		{ 200, 485, 32 },
		{ 200, 486, 32 },
		{ 200, 487, 32 },
		{ 200, 488, 32 },
		{ 200, 489, 32 },
		{ 200, 490, 32 },
		{ 200, 491, 32 },
		{ 200, 492, 32 },
		{ 200, 493, 32 },
		{ 200, 494, 32 },
		{ 200, 495, 32 },
		{ 200, 496, 32 },
		{ 200, 497, 32 },
		{ 200, 498, 32 },
		{ 200, 499, 32 },
		{ 200, 500, 32 },
		{ 200, 501, 32 },
		{ 200, 502, 32 },
		{ 200, 503, 32 },
		{ 200, 504, 32 },
		{ 200, 505, 32 },
		{ 200, 506, 32 },
		{ 200, 507, 32 },
		{ 200, 508, 32 },
		{ 200, 509, 32 },
		{ 200, 510, 32 },
		{ 200, 511, 32 },
		{ 200, 512, 32 },
		{ 201, 407, 32 },
		{ 201, 408, 32 },
		{ 201, 409, 32 },
		{ 201, 410, 32 },
		{ 201, 411, 32 },
		{ 201, 412, 32 },
		{ 201, 413, 32 },
		{ 201, 414, 32 },
		{ 201, 415, 32 },
		{ 201, 416, 32 },
		{ 201, 417, 32 },
		{ 201, 418, 32 },
		{ 201, 419, 32 },
		{ 201, 420, 32 },
		{ 201, 421, 32 },
		{ 201, 422, 32 },
		{ 201, 423, 32 },
		{ 201, 424, 32 },
		{ 201, 425, 32 },
		{ 201, 426, 32 },
		{ 201, 427, 32 },
		{ 201, 428, 32 },
		{ 201, 429, 32 },
		{ 201, 430, 32 },
		{ 201, 431, 32 },
		{ 201, 432, 32 },
		{ 201, 433, 32 },
		{ 201, 434, 32 },
		{ 201, 435, 32 },
		{ 201, 436, 32 },
		{ 201, 437, 32 },
		{ 201, 438, 32 },
		{ 201, 439, 32 },
		{ 201, 440, 32 },
		{ 201, 441, 32 },
		{ 201, 442, 32 },
		{ 201, 443, 32 },
		{ 201, 444, 32 },
		{ 201, 445, 32 },
		{ 201, 446, 32 },
		{ 201, 447, 32 },
		{ 201, 448, 32 },
		{ 201, 449, 32 },
		{ 201, 450, 32 },
		{ 201, 451, 32 },
		{ 201, 452, 32 },
		{ 201, 453, 32 },
		{ 201, 454, 32 },
		{ 201, 455, 32 },
		{ 201, 456, 32 },
		{ 201, 457, 32 },
		{ 201, 458, 32 },
		{ 201, 459, 32 },
		{ 201, 460, 32 },
		{ 201, 461, 32 },
		{ 201, 462, 32 },
		{ 201, 463, 32 },
		{ 201, 464, 32 },
		{ 201, 465, 32 },
		{ 201, 466, 32 },
		{ 201, 467, 32 },
		{ 201, 468, 32 },
		{ 201, 469, 32 },
		{ 201, 470, 32 },
		{ 201, 471, 32 },
		{ 201, 472, 32 },
		{ 201, 473, 32 },
		{ 201, 474, 32 },
		{ 201, 475, 32 },
		{ 201, 476, 32 },
		{ 201, 477, 32 },
		{ 201, 478, 32 },
		{ 201, 479, 32 },
		{ 201, 480, 32 },
		{ 201, 481, 32 },
		{ 201, 482, 32 },
		{ 201, 483, 32 },
		{ 201, 484, 32 },
		{ 201, 485, 32 },
		{ 201, 486, 32 },
		{ 201, 487, 32 },
		{ 201, 488, 32 },
		{ 201, 489, 32 },
		{ 201, 490, 32 },
		{ 201, 491, 32 },
		{ 201, 492, 32 },
		{ 201, 493, 32 },
		{ 201, 494, 32 },
		{ 201, 495, 32 },
		{ 201, 496, 32 },
		{ 201, 497, 32 },
		{ 201, 498, 32 },
		{ 201, 499, 32 },
		{ 201, 500, 32 },
		{ 201, 501, 32 },
		{ 201, 502, 32 },
		{ 201, 503, 32 },
		{ 201, 504, 32 },
		{ 201, 505, 32 },
		{ 201, 506, 32 },
		{ 201, 507, 32 },
		{ 201, 508, 32 },
		{ 201, 509, 32 },
		{ 201, 510, 32 },
		{ 201, 511, 32 },
		{ 201, 512, 32 },
		{ 202, 404, 32 },
		{ 202, 405, 32 },
		{ 202, 406, 32 },
		{ 202, 407, 32 },
		{ 202, 408, 32 },
		{ 202, 409, 32 },
		{ 202, 410, 32 },
		{ 202, 411, 32 },
		{ 202, 412, 32 },
		{ 202, 413, 32 },
		{ 202, 414, 32 },
		{ 202, 415, 32 },
		{ 202, 416, 32 },
		{ 202, 417, 32 },
		{ 202, 418, 32 },
		{ 202, 419, 32 },
		{ 202, 420, 32 },
		{ 202, 421, 32 },
		{ 202, 422, 32 },
		{ 202, 423, 32 },
		{ 202, 424, 32 },
		{ 202, 425, 32 },
		{ 202, 426, 32 },
		{ 202, 427, 32 },
		{ 202, 428, 32 },
		{ 202, 429, 32 },
		{ 202, 430, 32 },
		{ 202, 431, 32 },
		{ 202, 432, 32 },
		{ 202, 433, 32 },
		{ 202, 434, 32 },
		{ 202, 435, 32 },
		{ 202, 436, 32 },
		{ 202, 437, 32 },
		{ 202, 438, 32 },
		{ 202, 439, 32 },
		{ 202, 440, 32 },
		{ 202, 441, 32 },
		{ 202, 442, 32 },
		{ 202, 443, 32 },
		{ 202, 444, 32 },
		{ 202, 445, 32 },
		{ 202, 446, 32 },
		{ 202, 447, 32 },
		{ 202, 448, 32 },
		{ 202, 449, 32 },
		{ 202, 450, 32 },
		{ 202, 451, 32 },
		{ 202, 452, 32 },
		{ 202, 453, 32 },
		{ 202, 454, 32 },
		{ 202, 455, 32 },
		{ 202, 456, 32 },
		{ 202, 457, 32 },
		{ 202, 458, 32 },
		{ 202, 459, 32 },
		{ 202, 460, 32 },
		{ 202, 461, 32 },
		{ 202, 462, 32 },
		{ 202, 463, 32 },
		{ 202, 464, 32 },
		{ 202, 465, 32 },
		{ 202, 466, 32 },
		{ 202, 467, 32 },
		{ 202, 468, 32 },
		{ 202, 469, 32 },
		{ 202, 470, 32 },
		{ 202, 471, 32 },
		{ 202, 472, 32 },
		{ 202, 473, 32 },
		{ 202, 474, 32 },
		{ 202, 475, 32 },
		{ 202, 476, 32 },
		{ 202, 477, 32 },
		{ 202, 478, 32 },
		{ 202, 479, 32 },
		{ 202, 480, 32 },
		{ 202, 481, 32 },
		{ 202, 482, 32 },
		{ 202, 483, 32 },
		{ 202, 484, 32 },
		{ 202, 485, 32 },
		{ 202, 486, 32 },
		{ 202, 487, 32 },
		{ 202, 488, 32 },
		{ 202, 489, 32 },
		{ 202, 490, 32 },
		{ 202, 491, 32 },
		{ 202, 492, 32 },
		{ 202, 493, 32 },
		{ 202, 494, 32 },
		{ 202, 495, 32 },
		{ 202, 496, 32 },
		{ 202, 497, 32 },
		{ 202, 498, 32 },
		{ 202, 499, 32 },
		{ 202, 500, 32 },
		{ 202, 501, 32 },
		{ 202, 502, 32 },
		{ 202, 503, 32 },
		{ 202, 504, 32 },
		{ 202, 505, 32 },
		{ 202, 506, 32 },
		{ 202, 507, 32 },
		{ 202, 508, 32 },
		{ 202, 509, 32 },
		{ 202, 510, 32 },
		{ 202, 511, 32 },
		{ 202, 512, 32 },
		{ 203, 402, 32 },
		{ 203, 403, 32 },
		{ 203, 404, 32 },
		{ 203, 405, 32 },
		{ 203, 406, 32 },
		{ 203, 407, 32 },
		{ 203, 408, 32 },
		{ 203, 409, 32 },
		{ 203, 410, 32 },
		{ 203, 411, 32 },
		{ 203, 412, 32 },
		{ 203, 413, 32 },
		{ 203, 414, 32 },
		{ 203, 415, 32 },
		{ 203, 416, 32 },
		{ 203, 417, 32 },
		{ 203, 418, 32 },
		{ 203, 419, 32 },
		{ 203, 420, 32 },
		{ 203, 421, 32 },
		{ 203, 422, 32 },
		{ 203, 423, 32 },
		{ 203, 424, 32 },
		{ 203, 425, 32 },
		{ 203, 426, 32 },
		{ 203, 427, 32 },
		{ 203, 428, 32 },
		{ 203, 429, 32 },
		{ 203, 430, 32 },
		{ 203, 431, 32 },
		{ 203, 432, 32 },
		{ 203, 433, 32 },
		{ 203, 434, 32 },
		{ 203, 435, 32 },
		{ 203, 436, 32 },
		{ 203, 437, 32 },
		{ 203, 438, 32 },
		{ 203, 439, 32 },
		{ 203, 440, 32 },
		{ 203, 441, 32 },
		{ 203, 442, 32 },
		{ 203, 443, 32 },
		{ 203, 444, 32 },
		{ 203, 445, 32 },
		{ 203, 446, 32 },
		{ 203, 447, 32 },
		{ 203, 448, 32 },
		{ 203, 449, 32 },
		{ 203, 450, 32 },
		{ 203, 451, 32 },
		{ 203, 452, 32 },
		{ 203, 453, 32 },
		{ 203, 454, 32 },
		{ 203, 455, 32 },
		{ 203, 456, 32 },
		{ 203, 457, 32 },
		{ 203, 458, 32 },
		{ 203, 459, 32 },
		{ 203, 460, 32 },
		{ 203, 461, 32 },
		{ 203, 462, 32 },
		{ 203, 463, 32 },
		{ 203, 464, 32 },
		{ 203, 465, 32 },
		{ 203, 466, 32 },
		{ 203, 467, 32 },
		{ 203, 468, 32 },
		{ 203, 469, 32 },
		{ 203, 470, 32 },
		{ 203, 471, 32 },
		{ 203, 472, 32 },
		{ 203, 473, 32 },
		{ 203, 474, 32 },
		{ 203, 475, 32 },
		{ 203, 476, 32 },
		{ 203, 477, 32 },
		{ 203, 478, 32 },
		{ 203, 479, 32 },
		{ 203, 480, 32 },
		{ 203, 481, 32 },
		{ 203, 482, 32 },
		{ 203, 483, 32 },
		{ 203, 484, 32 },
		{ 203, 485, 32 },
		{ 203, 486, 32 },
		{ 203, 487, 32 },
		{ 203, 488, 32 },
		{ 203, 489, 32 },
		{ 203, 490, 32 },
		{ 203, 491, 32 },
		{ 203, 492, 32 },
		{ 203, 493, 32 },
		{ 203, 494, 32 },
		{ 203, 495, 32 },
		{ 203, 496, 32 },
		{ 203, 497, 32 },
		{ 203, 498, 32 },
		{ 203, 499, 32 },
		{ 203, 500, 32 },
		{ 203, 501, 32 },
		{ 203, 502, 32 },
		{ 203, 503, 32 },
		{ 203, 504, 32 },
		{ 203, 505, 32 },
		{ 203, 506, 32 },
		{ 203, 507, 32 },
		{ 203, 508, 32 },
		{ 203, 509, 32 },
		{ 203, 510, 32 },
		{ 203, 511, 32 },
		{ 203, 512, 32 },
		{ 204, 399, 32 },
		{ 204, 400, 32 },
		{ 204, 401, 32 },
		{ 204, 402, 32 },
		{ 204, 403, 32 },
		{ 204, 404, 32 },
		{ 204, 405, 32 },
		{ 204, 406, 32 },
		{ 204, 407, 32 },
		{ 204, 408, 32 },
		{ 204, 409, 32 },
		{ 204, 410, 32 },
		{ 204, 411, 32 },
		{ 204, 412, 32 },
		{ 204, 413, 32 },
		{ 204, 414, 32 },
		{ 204, 415, 32 },
		{ 204, 416, 32 },
		{ 204, 417, 32 },
		{ 204, 418, 32 },
		{ 204, 419, 32 },
		{ 204, 420, 32 },
		{ 204, 421, 32 },
		{ 204, 422, 32 },
		{ 204, 423, 32 },
		{ 204, 424, 32 },
		{ 204, 425, 32 },
		{ 204, 426, 32 },
		{ 204, 427, 32 },
		{ 204, 428, 32 },
		{ 204, 429, 32 },
		{ 204, 430, 32 },
		{ 204, 431, 32 },
		{ 204, 432, 32 },
		{ 204, 433, 32 },
		{ 204, 434, 32 },
		{ 204, 435, 32 },
		{ 204, 436, 32 },
		{ 204, 437, 32 },
		{ 204, 438, 32 },
		{ 204, 439, 32 },
		{ 204, 440, 32 },
		{ 204, 441, 32 },
		{ 204, 442, 32 },
		{ 204, 443, 32 },
		{ 204, 444, 32 },
		{ 204, 445, 32 },
		{ 204, 446, 32 },
		{ 204, 447, 32 },
		{ 204, 448, 32 },
		{ 204, 449, 32 },
		{ 204, 450, 32 },
		{ 204, 451, 32 },
		{ 204, 452, 32 },
		{ 204, 453, 32 },
		{ 204, 454, 32 },
		{ 204, 455, 32 },
		{ 204, 456, 32 },
		{ 204, 457, 32 },
		{ 204, 458, 32 },
		{ 204, 459, 32 },
		{ 204, 460, 32 },
		{ 204, 461, 32 },
		{ 204, 462, 32 },
		{ 204, 463, 32 },
		{ 204, 464, 32 },
		{ 204, 465, 32 },
		{ 204, 466, 32 },
		{ 204, 467, 32 },
		{ 204, 468, 32 },
		{ 204, 469, 32 },
		{ 204, 470, 32 },
		{ 204, 471, 32 },
		{ 204, 472, 32 },
		{ 204, 473, 32 },
		{ 204, 474, 32 },
		{ 204, 475, 32 },
		{ 204, 476, 32 },
		{ 204, 477, 32 },
		{ 204, 478, 32 },
		{ 204, 479, 32 },
		{ 204, 480, 32 },
		{ 204, 481, 32 },
		{ 204, 482, 32 },
		{ 204, 483, 32 },
		{ 204, 484, 32 },
		{ 204, 485, 32 },
		{ 204, 486, 32 },
		{ 204, 487, 32 },
		{ 204, 488, 32 },
		{ 204, 489, 32 },
		{ 204, 490, 32 },
		{ 204, 491, 32 },
		{ 204, 492, 32 },
		{ 204, 493, 32 },
		{ 204, 494, 32 },
		{ 204, 495, 32 },
		{ 204, 496, 32 },
		{ 204, 497, 32 },
		{ 204, 498, 32 },
		{ 204, 499, 32 },
		{ 204, 500, 32 },
		{ 204, 501, 32 },
		{ 204, 502, 32 },
		{ 204, 503, 32 },
		{ 204, 504, 32 },
		{ 204, 505, 32 },
		{ 204, 506, 32 },
		{ 204, 507, 32 },
		{ 204, 508, 32 },
		{ 204, 509, 32 },
		{ 204, 510, 32 },
		{ 204, 511, 32 },
		{ 204, 512, 32 },
		{ 205, 396, 32 },
		{ 205, 397, 32 },
		{ 205, 398, 32 },
		{ 205, 399, 32 },
		{ 205, 400, 32 },
		{ 205, 401, 32 },
		{ 205, 402, 32 },
		{ 205, 403, 32 },
		{ 205, 404, 32 },
		{ 205, 405, 32 },
		{ 205, 406, 32 },
		{ 205, 407, 32 },
		{ 205, 408, 32 },
		{ 205, 409, 32 },
		{ 205, 410, 32 },
		{ 205, 411, 32 },
		{ 205, 412, 32 },
		{ 205, 413, 32 },
		{ 205, 414, 32 },
		{ 205, 415, 32 },
		{ 205, 416, 32 },
		{ 205, 417, 32 },
		{ 205, 418, 32 },
		{ 205, 419, 32 },
		{ 205, 420, 32 },
		{ 205, 421, 32 },
		{ 205, 422, 32 },
		{ 205, 423, 32 },
		{ 205, 424, 32 },
		{ 205, 425, 32 },
		{ 205, 426, 32 },
		{ 205, 427, 32 },
		{ 205, 428, 32 },
		{ 205, 429, 32 },
		{ 205, 430, 32 },
		{ 205, 431, 32 },
		{ 205, 432, 32 },
		{ 205, 433, 32 },
		{ 205, 434, 32 },
		{ 205, 435, 32 },
		{ 205, 436, 32 },
		{ 205, 437, 32 },
		{ 205, 438, 32 },
		{ 205, 439, 32 },
		{ 205, 440, 32 },
		{ 205, 441, 32 },
		{ 205, 442, 32 },
		{ 205, 443, 32 },
		{ 205, 444, 32 },
		{ 205, 445, 32 },
		{ 205, 446, 32 },
		{ 205, 447, 32 },
		{ 205, 448, 32 },
		{ 205, 449, 32 },
		{ 205, 450, 32 },
		{ 205, 451, 32 },
		{ 205, 452, 32 },
		{ 205, 453, 32 },
		{ 205, 454, 32 },
		{ 205, 455, 32 },
		{ 205, 456, 32 },
		{ 205, 457, 32 },
		{ 205, 458, 32 },
		{ 205, 459, 32 },
		{ 205, 460, 32 },
		{ 205, 461, 32 },
		{ 205, 462, 32 },
		{ 205, 463, 32 },
		{ 205, 464, 32 },
		{ 205, 465, 32 },
		{ 205, 466, 32 },
		{ 205, 467, 32 },
		{ 205, 468, 32 },
		{ 205, 469, 32 },
		{ 205, 470, 32 },
		{ 205, 471, 32 },
		{ 205, 472, 32 },
		{ 205, 473, 32 },
		{ 205, 474, 32 },
		{ 205, 475, 32 },
		{ 205, 476, 32 },
		{ 205, 477, 32 },
		{ 205, 478, 32 },
		{ 205, 479, 32 },
		{ 205, 480, 32 },
		{ 205, 481, 32 },
		{ 205, 482, 32 },
		{ 205, 483, 32 },
		{ 205, 484, 32 },
		{ 205, 485, 32 },
		{ 205, 486, 32 },
		{ 205, 487, 32 },
		{ 205, 488, 32 },
		{ 205, 489, 32 },
		{ 205, 490, 32 },
		{ 205, 491, 32 },
		{ 205, 492, 32 },
		{ 205, 493, 32 },
		{ 205, 494, 32 },
		{ 205, 495, 32 },
		{ 205, 496, 32 },
		{ 205, 497, 32 },
		{ 205, 498, 32 },
		{ 205, 499, 32 },
		{ 205, 500, 32 },
		{ 205, 501, 32 },
		{ 205, 502, 32 },
		{ 205, 503, 32 },
		{ 205, 504, 32 },
		{ 205, 505, 32 },
		{ 205, 506, 32 },
		{ 205, 507, 32 },
		{ 205, 508, 32 },
		{ 205, 509, 32 },
		{ 205, 510, 32 },
		{ 205, 511, 32 },
		{ 205, 512, 32 },
		{ 206, 394, 32 },
		{ 206, 395, 32 },
		{ 206, 396, 32 },
		{ 206, 397, 32 },
		{ 206, 398, 32 },
		{ 206, 399, 32 },
		{ 206, 400, 32 },
		{ 206, 401, 32 },
		{ 206, 402, 32 },
		{ 206, 403, 32 },
		{ 206, 404, 32 },
		{ 206, 405, 32 },
		{ 206, 406, 32 },
		{ 206, 407, 32 },
		{ 206, 408, 32 },
		{ 206, 409, 32 },
		{ 206, 410, 32 },
		{ 206, 411, 32 },
		{ 206, 412, 32 },
		{ 206, 413, 32 },
		{ 206, 414, 32 },
		{ 206, 415, 32 },
		{ 206, 416, 32 },
		{ 206, 417, 32 },
		{ 206, 418, 32 },
		{ 206, 419, 32 },
		{ 206, 420, 32 },
		{ 206, 421, 32 },
		{ 206, 422, 32 },
		{ 206, 423, 32 },
		{ 206, 424, 32 },
		{ 206, 425, 32 },
		{ 206, 426, 32 },
		{ 206, 427, 32 },
		{ 206, 428, 32 },
		{ 206, 429, 32 },
		{ 206, 430, 32 },
		{ 206, 431, 32 },
		{ 206, 432, 32 },
		{ 206, 433, 32 },
		{ 206, 434, 32 },
		{ 206, 435, 32 },
		{ 206, 436, 32 },
		{ 206, 437, 32 },
		{ 206, 438, 32 },
		{ 206, 439, 32 },
		{ 206, 440, 32 },
		{ 206, 441, 32 },
		{ 206, 442, 32 },
		{ 206, 443, 32 },
		{ 206, 444, 32 },
		{ 206, 445, 32 },
		{ 206, 446, 32 },
		{ 206, 447, 32 },
		{ 206, 448, 32 },
		{ 206, 449, 32 },
		{ 206, 450, 32 },
		{ 206, 451, 32 },
		{ 206, 452, 32 },
		{ 206, 453, 32 },
		{ 206, 454, 32 },
		{ 206, 455, 32 },
		{ 206, 456, 32 },
		{ 206, 457, 32 },
		{ 206, 458, 32 },
		{ 206, 459, 32 },
		{ 206, 460, 32 },
		{ 206, 461, 32 },
		{ 206, 462, 32 },
		{ 206, 463, 32 },
		{ 206, 464, 32 },
		{ 206, 465, 32 },
		{ 206, 466, 32 },
		{ 206, 467, 32 },
		{ 206, 468, 32 },
		{ 206, 469, 32 },
		{ 206, 470, 32 },
		{ 206, 471, 32 },
		{ 206, 472, 32 },
		{ 206, 473, 32 },
		{ 206, 474, 32 },
		{ 206, 475, 32 },
		{ 206, 476, 32 },
		{ 206, 477, 32 },
		{ 206, 478, 32 },
		{ 206, 479, 32 },
		{ 206, 480, 32 },
		{ 206, 481, 32 },
		{ 206, 482, 32 },
		{ 206, 483, 32 },
		{ 206, 484, 32 },
		{ 206, 485, 32 },
		{ 206, 486, 32 },
		{ 206, 487, 32 },
		{ 206, 488, 32 },
		{ 206, 489, 32 },
		{ 206, 490, 32 },
		{ 206, 491, 32 },
		{ 206, 492, 32 },
		{ 206, 493, 32 },
		{ 206, 494, 32 },
		{ 206, 495, 32 },
		{ 206, 496, 32 },
		{ 206, 497, 32 },
		{ 206, 498, 32 },
		{ 206, 499, 32 },
		{ 206, 500, 32 },
		{ 206, 501, 32 },
		{ 206, 502, 32 },
		{ 206, 503, 32 },
		{ 206, 504, 32 },
		{ 206, 505, 32 },
		{ 206, 506, 32 },
		{ 206, 507, 32 },
		{ 206, 508, 32 },
		{ 206, 509, 32 },
		{ 206, 510, 32 },
		{ 206, 511, 32 },
		{ 206, 512, 32 },
		{ 207, 391, 32 },
		{ 207, 392, 32 },
		{ 207, 393, 32 },
		{ 207, 394, 32 },
		{ 207, 395, 32 },
		{ 207, 396, 32 },
		{ 207, 397, 32 },
		{ 207, 398, 32 },
		{ 207, 399, 32 },
		{ 207, 400, 32 },
		{ 207, 401, 32 },
		{ 207, 402, 32 },
		{ 207, 403, 32 },
		{ 207, 404, 32 },
		{ 207, 405, 32 },
		{ 207, 406, 32 },
		{ 207, 407, 32 },
		{ 207, 408, 32 },
		{ 207, 409, 32 },
		{ 207, 410, 32 },
		{ 207, 411, 32 },
		{ 207, 412, 32 },
		{ 207, 413, 32 },
		{ 207, 414, 32 },
		{ 207, 415, 32 },
		{ 207, 416, 32 },
		{ 207, 417, 32 },
		{ 207, 418, 32 },
		{ 207, 419, 32 },
		{ 207, 420, 32 },
		{ 207, 421, 32 },
		{ 207, 422, 32 },
		{ 207, 423, 32 },
		{ 207, 424, 32 },
		{ 207, 425, 32 },
		{ 207, 426, 32 },
		{ 207, 427, 32 },
		{ 207, 428, 32 },
		{ 207, 429, 32 },
		{ 207, 430, 32 },
		{ 207, 431, 32 },
		{ 207, 432, 32 },
		{ 207, 433, 32 },
		{ 207, 434, 32 },
		{ 207, 435, 32 },
		{ 207, 436, 32 },
		{ 207, 437, 32 },
		{ 207, 438, 32 },
		{ 207, 439, 32 },
		{ 207, 440, 32 },
		{ 207, 441, 32 },
		{ 207, 442, 32 },
		{ 207, 443, 32 },
		{ 207, 444, 32 },
		{ 207, 445, 32 },
		{ 207, 446, 32 },
		{ 207, 447, 32 },
		{ 207, 448, 32 },
		{ 207, 449, 32 },
		{ 207, 450, 32 },
		{ 207, 451, 32 },
		{ 207, 452, 32 },
		{ 207, 453, 32 },
		{ 207, 454, 32 },
		{ 207, 455, 32 },
		{ 207, 456, 32 },
		{ 207, 457, 32 },
		{ 207, 458, 32 },
		{ 207, 459, 32 },
		{ 207, 460, 32 },
		{ 207, 461, 32 },
		{ 207, 462, 32 },
		{ 207, 463, 32 },
		{ 207, 464, 32 },
		{ 207, 465, 32 },
		{ 207, 466, 32 },
		{ 207, 467, 32 },
		{ 207, 468, 32 },
		{ 207, 469, 32 },
		{ 207, 470, 32 },
		{ 207, 471, 32 },
		{ 207, 472, 32 },
		{ 207, 473, 32 },
		{ 207, 474, 32 },
		{ 207, 475, 32 },
		{ 207, 476, 32 },
		{ 207, 477, 32 },
		{ 207, 478, 32 },
		{ 207, 479, 32 },
		{ 207, 480, 32 },
		{ 207, 481, 32 },
		{ 207, 482, 32 },
		{ 207, 483, 32 },
		{ 207, 484, 32 },
		{ 207, 485, 32 },
		{ 207, 486, 32 },
		{ 207, 487, 32 },
		{ 207, 488, 32 },
		{ 207, 489, 32 },
		{ 207, 490, 32 },
		{ 207, 491, 32 },
		{ 207, 492, 32 },
		{ 207, 493, 32 },
		{ 207, 494, 32 },
		{ 207, 495, 32 },
		{ 207, 496, 32 },
		{ 207, 497, 32 },
		{ 207, 498, 32 },
		{ 207, 499, 32 },
		{ 207, 500, 32 },
		{ 207, 501, 32 },
		{ 207, 502, 32 },
		{ 207, 503, 32 },
		{ 207, 504, 32 },
		{ 207, 505, 32 },
		{ 207, 506, 32 },
		{ 207, 507, 32 },
		{ 207, 508, 32 },
		{ 207, 509, 32 },
		{ 207, 510, 32 },
		{ 207, 511, 32 },
		{ 207, 512, 32 },
		{ 208, 388, 32 },
		{ 208, 389, 32 },
		{ 208, 390, 32 },
		{ 208, 391, 32 },
		{ 208, 392, 32 },
		{ 208, 393, 32 },
		{ 208, 394, 32 },
		{ 208, 395, 32 },
		{ 208, 396, 32 },
		{ 208, 397, 32 },
		{ 208, 398, 32 },
		{ 208, 399, 32 },
		{ 208, 400, 32 },
		{ 208, 401, 32 },
		{ 208, 402, 32 },
		{ 208, 403, 32 },
		{ 208, 404, 32 },
		{ 208, 405, 32 },
		{ 208, 406, 32 },
		{ 208, 407, 32 },
		{ 208, 408, 32 },
		{ 208, 409, 32 },
		{ 208, 410, 32 },
		{ 208, 411, 32 },
		{ 208, 412, 32 },
		{ 208, 413, 32 },
		{ 208, 414, 32 },
		{ 208, 415, 32 },
		{ 208, 416, 32 },
		{ 208, 417, 32 },
		{ 208, 418, 32 },
		{ 208, 419, 32 },
		{ 208, 420, 32 },
		{ 208, 421, 32 },
		{ 208, 422, 32 },
		{ 208, 423, 32 },
		{ 208, 424, 32 },
		{ 208, 425, 32 },
		{ 208, 426, 32 },
		{ 208, 427, 32 },
		{ 208, 428, 32 },
		{ 208, 429, 32 },
		{ 208, 430, 32 },
		{ 208, 431, 32 },
		{ 208, 432, 32 },
		{ 208, 433, 32 },
		{ 208, 434, 32 },
		{ 208, 435, 32 },
		{ 208, 436, 32 },
		{ 208, 437, 32 },
		{ 208, 438, 32 },
		{ 208, 439, 32 },
		{ 208, 440, 32 },
		{ 208, 441, 32 },
		{ 208, 442, 32 },
		{ 208, 443, 32 },
		{ 208, 444, 32 },
		{ 208, 445, 32 },
		{ 208, 446, 32 },
		{ 208, 447, 32 },
		{ 208, 448, 32 },
		{ 208, 449, 32 },
		{ 208, 450, 32 },
		{ 208, 451, 32 },
		{ 208, 452, 32 },
		{ 208, 453, 32 },
		{ 208, 454, 32 },
		{ 208, 455, 32 },
		{ 208, 456, 32 },
		{ 208, 457, 32 },
		{ 208, 458, 32 },
		{ 208, 459, 32 },
		{ 208, 460, 32 },
		{ 208, 461, 32 },
		{ 208, 462, 32 },
		{ 208, 463, 32 },
		{ 208, 464, 32 },
		{ 208, 465, 32 },
		{ 208, 466, 32 },
		{ 208, 467, 32 },
		{ 208, 468, 32 },
		{ 208, 469, 32 },
		{ 208, 470, 32 },
		{ 208, 471, 32 },
		{ 208, 472, 32 },
		{ 208, 473, 32 },
		{ 208, 474, 32 },
		{ 208, 475, 32 },
		{ 208, 476, 32 },
		{ 208, 477, 32 },
		{ 208, 478, 32 },
		{ 208, 479, 32 },
		{ 208, 480, 32 },
		{ 208, 481, 32 },
		{ 208, 482, 32 },
		{ 208, 483, 32 },
		{ 208, 484, 32 },
		{ 208, 485, 32 },
		{ 208, 486, 32 },
		{ 208, 487, 32 },
		{ 208, 488, 32 },
		{ 208, 489, 32 },
		{ 208, 490, 32 },
		{ 208, 491, 32 },
		{ 208, 492, 32 },
		{ 208, 493, 32 },
		{ 208, 494, 32 },
		{ 208, 495, 32 },
		{ 208, 496, 32 },
		{ 208, 497, 32 },
		{ 208, 498, 32 },
		{ 208, 499, 32 },
		{ 208, 500, 32 },
		{ 208, 501, 32 },
		{ 208, 502, 32 },
		{ 208, 503, 32 },
		{ 208, 504, 32 },
		{ 208, 505, 32 },
		{ 208, 506, 32 },
		{ 208, 507, 32 },
		{ 208, 508, 32 },
		{ 208, 509, 32 },
		{ 208, 510, 32 },
		{ 208, 511, 32 },
		{ 208, 512, 32 },
		{ 209, 386, 32 },
		{ 209, 387, 32 },
		{ 209, 388, 32 },
		{ 209, 389, 32 },
		{ 209, 390, 32 },
		{ 209, 391, 32 },
		{ 209, 392, 32 },
		{ 209, 393, 32 },
		{ 209, 394, 32 },
		{ 209, 395, 32 },
		{ 209, 396, 32 },
		{ 209, 397, 32 },
		{ 209, 398, 32 },
		{ 209, 399, 32 },
		{ 209, 400, 32 },
		{ 209, 401, 32 },
		{ 209, 402, 32 },
		{ 209, 403, 32 },
		{ 209, 404, 32 },
		{ 209, 405, 32 },
		{ 209, 406, 32 },
		{ 209, 407, 32 },
		{ 209, 408, 32 },
		{ 209, 409, 32 },
		{ 209, 410, 32 },
		{ 209, 411, 32 },
		{ 209, 412, 32 },
		{ 209, 413, 32 },
		{ 209, 414, 32 },
		{ 209, 415, 32 },
		{ 209, 416, 32 },
		{ 209, 417, 32 },
		{ 209, 418, 32 },
		{ 209, 419, 32 },
		{ 209, 420, 32 },
		{ 209, 421, 32 },
		{ 209, 422, 32 },
		{ 209, 423, 32 },
		{ 209, 424, 32 },
		{ 209, 425, 32 },
		{ 209, 426, 32 },
		{ 209, 427, 32 },
		{ 209, 428, 32 },
		{ 209, 429, 32 },
		{ 209, 430, 32 },
		{ 209, 431, 32 },
		{ 209, 432, 32 },
		{ 209, 433, 32 },
		{ 209, 434, 32 },
		{ 209, 435, 32 },
		{ 209, 436, 32 },
		{ 209, 437, 32 },
		{ 209, 438, 32 },
		{ 209, 439, 32 },
		{ 209, 440, 32 },
		{ 209, 441, 32 },
		{ 209, 442, 32 },
		{ 209, 443, 32 },
		{ 209, 444, 32 },
		{ 209, 445, 32 },
		{ 209, 446, 32 },
		{ 209, 447, 32 },
		{ 209, 448, 32 },
		{ 209, 449, 32 },
		{ 209, 450, 32 },
		{ 209, 451, 32 },
		{ 209, 452, 32 },
		{ 209, 453, 32 },
		{ 209, 454, 32 },
		{ 209, 455, 32 },
		{ 209, 456, 32 },
		{ 209, 457, 32 },
		{ 209, 458, 32 },
		{ 209, 459, 32 },
		{ 209, 460, 32 },
		{ 209, 461, 32 },
		{ 209, 462, 32 },
		{ 209, 463, 32 },
		{ 209, 464, 32 },
		{ 209, 465, 32 },
		{ 209, 466, 32 },
		{ 209, 467, 32 },
		{ 209, 468, 32 },
		{ 209, 469, 32 },
		{ 209, 470, 32 },
		{ 209, 471, 32 },
		{ 209, 472, 32 },
		{ 209, 473, 32 },
		{ 209, 474, 32 },
		{ 209, 475, 32 },
		{ 209, 476, 32 },
		{ 209, 477, 32 },
		{ 209, 478, 32 },
		{ 209, 479, 32 },
		{ 209, 480, 32 },
		{ 209, 481, 32 },
		{ 209, 482, 32 },
		{ 209, 483, 32 },
		{ 209, 484, 32 },
		{ 209, 485, 32 },
		{ 209, 486, 32 },
		{ 209, 487, 32 },
		{ 209, 488, 32 },
		{ 209, 489, 32 },
		{ 209, 490, 32 },
		{ 209, 491, 32 },
		{ 209, 492, 32 },
		{ 209, 493, 32 },
		{ 209, 494, 32 },
		{ 209, 495, 32 },
		{ 209, 496, 32 },
		{ 209, 497, 32 },
		{ 209, 498, 32 },
		{ 209, 499, 32 },
		{ 209, 500, 32 },
		{ 209, 501, 32 },
		{ 209, 502, 32 },
		{ 209, 503, 32 },
		{ 209, 504, 32 },
		{ 209, 505, 32 },
		{ 209, 506, 32 },
		{ 209, 507, 32 },
		{ 209, 508, 32 },
		{ 209, 509, 32 },
		{ 209, 510, 32 },
		{ 209, 511, 32 },
		{ 209, 512, 32 },
		{ 210, 383, 32 },
		{ 210, 384, 32 },
		{ 210, 385, 32 },
		{ 210, 386, 32 },
		{ 210, 387, 32 },
		{ 210, 388, 32 },
		{ 210, 389, 32 },
		{ 210, 390, 32 },
		{ 210, 391, 32 },
		{ 210, 392, 32 },
		{ 210, 393, 32 },
		{ 210, 394, 32 },
		{ 210, 395, 32 },
		{ 210, 396, 32 },
		{ 210, 397, 32 },
		{ 210, 398, 32 },
		{ 210, 399, 32 },
		{ 210, 400, 32 },
		{ 210, 401, 32 },
		{ 210, 402, 32 },
		{ 210, 403, 32 },
		{ 210, 404, 32 },
		{ 210, 405, 32 },
		{ 210, 406, 32 },
		{ 210, 407, 32 },
		{ 210, 408, 32 },
		{ 210, 409, 32 },
		{ 210, 410, 32 },
		{ 210, 411, 32 },
		{ 210, 412, 32 },
		{ 210, 413, 32 },
		{ 210, 414, 32 },
		{ 210, 415, 32 },
		{ 210, 416, 32 },
		{ 210, 417, 32 },
		{ 210, 418, 32 },
		{ 210, 419, 32 },
		{ 210, 420, 32 },
		{ 210, 421, 32 },
		{ 210, 422, 32 },
		{ 210, 423, 32 },
		{ 210, 424, 32 },
		{ 210, 425, 32 },
		{ 210, 426, 32 },
		{ 210, 427, 32 },
		{ 210, 428, 32 },
		{ 210, 429, 32 },
		{ 210, 430, 32 },
		{ 210, 431, 32 },
		{ 210, 432, 32 },
		{ 210, 433, 32 },
		{ 210, 434, 32 },
		{ 210, 435, 32 },
		{ 210, 436, 32 },
		{ 210, 437, 32 },
		{ 210, 438, 32 },
		{ 210, 439, 32 },
		{ 210, 440, 32 },
		{ 210, 441, 32 },
		{ 210, 442, 32 },
		{ 210, 443, 32 },
		{ 210, 444, 32 },
		{ 210, 445, 32 },
		{ 210, 446, 32 },
		{ 210, 447, 32 },
		{ 210, 448, 32 },
		{ 210, 449, 32 },
		{ 210, 450, 32 },
		{ 210, 451, 32 },
		{ 210, 452, 32 },
		{ 210, 453, 32 },
		{ 210, 454, 32 },
		{ 210, 455, 32 },
		{ 210, 456, 32 },
		{ 210, 457, 32 },
		{ 210, 458, 32 },
		{ 210, 459, 32 },
		{ 210, 460, 32 },
		{ 210, 461, 32 },
		{ 210, 462, 32 },
		{ 210, 463, 32 },
		{ 210, 464, 32 },
		{ 210, 465, 32 },
		{ 210, 466, 32 },
		{ 210, 467, 32 },
		{ 210, 468, 32 },
		{ 210, 469, 32 },
		{ 210, 470, 32 },
		{ 210, 471, 32 },
		{ 210, 472, 32 },
		{ 210, 473, 32 },
		{ 210, 474, 32 },
		{ 210, 475, 32 },
		{ 210, 476, 32 },
		{ 210, 477, 32 },
		{ 210, 478, 32 },
		{ 210, 479, 32 },
		{ 210, 480, 32 },
		{ 210, 481, 32 },
		{ 210, 482, 32 },
		{ 210, 483, 32 },
		{ 210, 484, 32 },
		{ 210, 485, 32 },
		{ 210, 486, 32 },
		{ 210, 487, 32 },
		{ 210, 488, 32 },
		{ 210, 489, 32 },
		{ 210, 490, 32 },
		{ 210, 491, 32 },
		{ 210, 492, 32 },
		{ 210, 493, 32 },
		{ 210, 494, 32 },
		{ 210, 495, 32 },
		{ 210, 496, 32 },
		{ 210, 497, 32 },
		{ 210, 498, 32 },
		{ 210, 499, 32 },
		{ 210, 500, 32 },
		{ 210, 501, 32 },
		{ 210, 502, 32 },
		{ 210, 503, 32 },
		{ 210, 504, 32 },
		{ 210, 505, 32 },
		{ 210, 506, 32 },
		{ 210, 507, 32 },
		{ 210, 508, 32 },
		{ 210, 509, 32 },
		{ 210, 510, 32 },
		{ 210, 511, 32 },
		{ 210, 512, 32 },
		{ 211, 380, 32 },
		{ 211, 381, 32 },
		{ 211, 382, 32 },
		{ 211, 383, 32 },
		{ 211, 384, 32 },
		{ 211, 385, 32 },
		{ 211, 386, 32 },
		{ 211, 387, 32 },
		{ 211, 388, 32 },
		{ 211, 389, 32 },
		{ 211, 390, 32 },
		{ 211, 391, 32 },
		{ 211, 392, 32 },
		{ 211, 393, 32 },
		{ 211, 394, 32 },
		{ 211, 395, 32 },
		{ 211, 396, 32 },
		{ 211, 397, 32 },
		{ 211, 398, 32 },
		{ 211, 399, 32 },
		{ 211, 400, 32 },
		{ 211, 401, 32 },
		{ 211, 402, 32 },
		{ 211, 403, 32 },
		{ 211, 404, 32 },
		{ 211, 405, 32 },
		{ 211, 406, 32 },
		{ 211, 407, 32 },
		{ 211, 408, 32 },
		{ 211, 409, 32 },
		{ 211, 410, 32 },
		{ 211, 411, 32 },
		{ 211, 412, 32 },
		{ 211, 413, 32 },
		{ 211, 414, 32 },
		{ 211, 415, 32 },
		{ 211, 416, 32 },
		{ 211, 417, 32 },
		{ 211, 418, 32 },
		{ 211, 419, 32 },
		{ 211, 420, 32 },
		{ 211, 421, 32 },
		{ 211, 422, 32 },
		{ 211, 423, 32 },
		{ 211, 424, 32 },
		{ 211, 425, 32 },
		{ 211, 426, 32 },
		{ 211, 427, 32 },
		{ 211, 428, 32 },
		{ 211, 429, 32 },
		{ 211, 430, 32 },
		{ 211, 431, 32 },
		{ 211, 432, 32 },
		{ 211, 433, 32 },
		{ 211, 434, 32 },
		{ 211, 435, 32 },
		{ 211, 436, 32 },
		{ 211, 437, 32 },
		{ 211, 438, 32 },
		{ 211, 439, 32 },
		{ 211, 440, 32 },
		{ 211, 441, 32 },
		{ 211, 442, 32 },
		{ 211, 443, 32 },
		{ 211, 444, 32 },
		{ 211, 445, 32 },
		{ 211, 446, 32 },
		{ 211, 447, 32 },
		{ 211, 448, 32 },
		{ 211, 449, 32 },
		{ 211, 450, 32 },
		{ 211, 451, 32 },
		{ 211, 452, 32 },
		{ 211, 453, 32 },
		{ 211, 454, 32 },
		{ 211, 455, 32 },
		{ 211, 456, 32 },
		{ 211, 457, 32 },
		{ 211, 458, 32 },
		{ 211, 459, 32 },
		{ 211, 460, 32 },
		{ 211, 461, 32 },
		{ 211, 462, 32 },
		{ 211, 463, 32 },
		{ 211, 464, 32 },
		{ 211, 465, 32 },
		{ 211, 466, 32 },
		{ 211, 467, 32 },
		{ 211, 468, 32 },
		{ 211, 469, 32 },
		{ 211, 470, 32 },
		{ 211, 471, 32 },
		{ 211, 472, 32 },
		{ 211, 473, 32 },
		{ 211, 474, 32 },
		{ 211, 475, 32 },
		{ 211, 476, 32 },
		{ 211, 477, 32 },
		{ 211, 478, 32 },
		{ 211, 479, 32 },
		{ 211, 480, 32 },
		{ 211, 481, 32 },
		{ 211, 482, 32 },
		{ 211, 483, 32 },
		{ 211, 484, 32 },
		{ 211, 485, 32 },
		{ 211, 486, 32 },
		{ 211, 487, 32 },
		{ 211, 488, 32 },
		{ 211, 489, 32 },
		{ 211, 490, 32 },
		{ 211, 491, 32 },
		{ 211, 492, 32 },
		{ 211, 493, 32 },
		{ 211, 494, 32 },
		{ 211, 495, 32 },
		{ 211, 496, 32 },
		{ 211, 497, 32 },
		{ 211, 498, 32 },
		{ 211, 499, 32 },
		{ 211, 500, 32 },
		{ 211, 501, 32 },
		{ 211, 502, 32 },
		{ 211, 503, 32 },
		{ 211, 504, 32 },
		{ 211, 505, 32 },
		{ 211, 506, 32 },
		{ 211, 507, 32 },
		{ 211, 508, 32 },
		{ 211, 509, 32 },
		{ 211, 510, 32 },
		{ 211, 511, 32 },
		{ 211, 512, 32 },
		{ 212, 378, 32 },
		{ 212, 379, 32 },
		{ 212, 380, 32 },
		{ 212, 381, 32 },
		{ 212, 382, 32 },
		{ 212, 383, 32 },
		{ 212, 384, 32 },
		{ 212, 385, 32 },
		{ 212, 386, 32 },
		{ 212, 387, 32 },
		{ 212, 388, 32 },
		{ 212, 389, 32 },
		{ 212, 390, 32 },
		{ 212, 391, 32 },
		{ 212, 392, 32 },
		{ 212, 393, 32 },
		{ 212, 394, 32 },
		{ 212, 395, 32 },
		{ 212, 396, 32 },
		{ 212, 397, 32 },
		{ 212, 398, 32 },
		{ 212, 399, 32 },
		{ 212, 400, 32 },
		{ 212, 401, 32 },
		{ 212, 402, 32 },
		{ 212, 403, 32 },
		{ 212, 404, 32 },
		{ 212, 405, 32 },
		{ 212, 406, 32 },
		{ 212, 407, 32 },
		{ 212, 408, 32 },
		{ 212, 409, 32 },
		{ 212, 410, 32 },
		{ 212, 411, 32 },
		{ 212, 412, 32 },
		{ 212, 413, 32 },
		{ 212, 414, 32 },
		{ 212, 415, 32 },
		{ 212, 416, 32 },
		{ 212, 417, 32 },
		{ 212, 418, 32 },
		{ 212, 419, 32 },
		{ 212, 420, 32 },
		{ 212, 421, 32 },
		{ 212, 422, 32 },
		{ 212, 423, 32 },
		{ 212, 424, 32 },
		{ 212, 425, 32 },
		{ 212, 426, 32 },
		{ 212, 427, 32 },
		{ 212, 428, 32 },
		{ 212, 429, 32 },
		{ 212, 430, 32 },
		{ 212, 431, 32 },
		{ 212, 432, 32 },
		{ 212, 433, 32 },
		{ 212, 434, 32 },
		{ 212, 435, 32 },
		{ 212, 436, 32 },
		{ 212, 437, 32 },
		{ 212, 438, 32 },
		{ 212, 439, 32 },
		{ 212, 440, 32 },
		{ 212, 441, 32 },
		{ 212, 442, 32 },
		{ 212, 443, 32 },
		{ 212, 444, 32 },
		{ 212, 445, 32 },
		{ 212, 446, 32 },
		{ 212, 447, 32 },
		{ 212, 448, 32 },
		{ 212, 449, 32 },
		{ 212, 450, 32 },
		{ 212, 451, 32 },
		{ 212, 452, 32 },
		{ 212, 453, 32 },
		{ 212, 454, 32 },
		{ 212, 455, 32 },
		{ 212, 456, 32 },
		{ 212, 457, 32 },
		{ 212, 458, 32 },
		{ 212, 459, 32 },
		{ 212, 460, 32 },
		{ 212, 461, 32 },
		{ 212, 462, 32 },
		{ 212, 463, 32 },
		{ 212, 464, 32 },
		{ 212, 465, 32 },
		{ 212, 466, 32 },
		{ 212, 467, 32 },
		{ 212, 468, 32 },
		{ 212, 469, 32 },
		{ 212, 470, 32 },
		{ 212, 471, 32 },
		{ 212, 472, 32 },
		{ 212, 473, 32 },
		{ 212, 474, 32 },
		{ 212, 475, 32 },
		{ 212, 476, 32 },
		{ 212, 477, 32 },
		{ 212, 478, 32 },
		{ 212, 479, 32 },
		{ 212, 480, 32 },
		{ 212, 481, 32 },
		{ 212, 482, 32 },
		{ 212, 483, 32 },
		{ 212, 484, 32 },
		{ 212, 485, 32 },
		{ 212, 486, 32 },
		{ 212, 487, 32 },
		{ 212, 488, 32 },
		{ 212, 489, 32 },
		{ 212, 490, 32 },
		{ 212, 491, 32 },
		{ 212, 492, 32 },
		{ 212, 493, 32 },
		{ 212, 494, 32 },
		{ 212, 495, 32 },
		{ 212, 496, 32 },
		{ 212, 497, 32 },
		{ 212, 498, 32 },
		{ 212, 499, 32 },
		{ 212, 500, 32 },
		{ 212, 501, 32 },
		{ 212, 502, 32 },
		{ 212, 503, 32 },
		{ 212, 504, 32 },
		{ 212, 505, 32 },
		{ 212, 506, 32 },
		{ 212, 507, 32 },
		{ 212, 508, 32 },
		{ 212, 509, 32 },
		{ 212, 510, 32 },
		{ 212, 511, 32 },
		{ 212, 512, 32 },
		{ 213, 375, 32 },
		{ 213, 376, 32 },
		{ 213, 377, 32 },
		{ 213, 378, 32 },
		{ 213, 379, 32 },
		{ 213, 380, 32 },
		{ 213, 381, 32 },
		{ 213, 382, 32 },
		{ 213, 383, 32 },
		{ 213, 384, 32 },
		{ 213, 385, 32 },
		{ 213, 386, 32 },
		{ 213, 387, 32 },
		{ 213, 388, 32 },
		{ 213, 389, 32 },
		{ 213, 390, 32 },
		{ 213, 391, 32 },
		{ 213, 392, 32 },
		{ 213, 393, 32 },
		{ 213, 394, 32 },
		{ 213, 395, 32 },
		{ 213, 396, 32 },
		{ 213, 397, 32 },
		{ 213, 398, 32 },
		{ 213, 399, 32 },
		{ 213, 400, 32 },
		{ 213, 401, 32 },
		{ 213, 402, 32 },
		{ 213, 403, 32 },
		{ 213, 404, 32 },
		{ 213, 405, 32 },
		{ 213, 406, 32 },
		{ 213, 407, 32 },
		{ 213, 408, 32 },
		{ 213, 409, 32 },
		{ 213, 410, 32 },
		{ 213, 411, 32 },
		{ 213, 412, 32 },
		{ 213, 413, 32 },
		{ 213, 414, 32 },
		{ 213, 415, 32 },
		{ 213, 416, 32 },
		{ 213, 417, 32 },
		{ 213, 418, 32 },
		{ 213, 419, 32 },
		{ 213, 420, 32 },
		{ 213, 421, 32 },
		{ 213, 422, 32 },
		{ 213, 423, 32 },
		{ 213, 424, 32 },
		{ 213, 425, 32 },
		{ 213, 426, 32 },
		{ 213, 427, 32 },
		{ 213, 428, 32 },
		{ 213, 429, 32 },
		{ 213, 430, 32 },
		{ 213, 431, 32 },
		{ 213, 432, 32 },
		{ 213, 433, 32 },
		{ 213, 434, 32 },
		{ 213, 435, 32 },
		{ 213, 436, 32 },
		{ 213, 437, 32 },
		{ 213, 438, 32 },
		{ 213, 439, 32 },
		{ 213, 440, 32 },
		{ 213, 441, 32 },
		{ 213, 442, 32 },
		{ 213, 443, 32 },
		{ 213, 444, 32 },
		{ 213, 445, 32 },
		{ 213, 446, 32 },
		{ 213, 447, 32 },
		{ 213, 448, 32 },
		{ 213, 449, 32 },
		{ 213, 450, 32 },
		{ 213, 451, 32 },
		{ 213, 452, 32 },
		{ 213, 453, 32 },
		{ 213, 454, 32 },
		{ 213, 455, 32 },
		{ 213, 456, 32 },
		{ 213, 457, 32 },
		{ 213, 458, 32 },
		{ 213, 459, 32 },
		{ 213, 460, 32 },
		{ 213, 461, 32 },
		{ 213, 462, 32 },
		{ 213, 463, 32 },
		{ 213, 464, 32 },
		{ 213, 465, 32 },
		{ 213, 466, 32 },
		{ 213, 467, 32 },
		{ 213, 468, 32 },
		{ 213, 469, 32 },
		{ 213, 470, 32 },
		{ 213, 471, 32 },
		{ 213, 472, 32 },
		{ 213, 473, 32 },
		{ 213, 474, 32 },
		{ 213, 475, 32 },
		{ 213, 476, 32 },
		{ 213, 477, 32 },
		{ 213, 478, 32 },
		{ 213, 479, 32 },
		{ 213, 480, 32 },
		{ 213, 481, 32 },
		{ 213, 482, 32 },
		{ 213, 483, 32 },
		{ 213, 484, 32 },
		{ 213, 485, 32 },
		{ 213, 486, 32 },
		{ 213, 487, 32 },
		{ 213, 488, 32 },
		{ 213, 489, 32 },
		{ 213, 490, 32 },
		{ 213, 491, 32 },
		{ 213, 492, 32 },
		{ 213, 493, 32 },
		{ 213, 494, 32 },
		{ 213, 495, 32 },
		{ 213, 496, 32 },
		{ 213, 497, 32 },
		{ 213, 498, 32 },
		{ 213, 499, 32 },
		{ 213, 500, 32 },
		{ 213, 501, 32 },
		{ 213, 502, 32 },
		{ 213, 503, 32 },
		{ 213, 504, 32 },
		{ 213, 505, 32 },
		{ 213, 506, 32 },
		{ 213, 507, 32 },
		{ 213, 508, 32 },
		{ 213, 509, 32 },
		{ 213, 510, 32 },
		{ 213, 511, 32 },
		{ 213, 512, 32 },
		{ 214, 372, 32 },
		{ 214, 373, 32 },
		{ 214, 374, 32 },
		{ 214, 375, 32 },
		{ 214, 376, 32 },
		{ 214, 377, 32 },
		{ 214, 378, 32 },
		{ 214, 379, 32 },
		{ 214, 380, 32 },
		{ 214, 381, 32 },
		{ 214, 382, 32 },
		{ 214, 383, 32 },
		{ 214, 384, 32 },
		{ 214, 385, 32 },
		{ 214, 386, 32 },
		{ 214, 387, 32 },
		{ 214, 388, 32 },
		{ 214, 389, 32 },
		{ 214, 390, 32 },
		{ 214, 391, 32 },
		{ 214, 392, 32 },
		{ 214, 393, 32 },
		{ 214, 394, 32 },
		{ 214, 395, 32 },
		{ 214, 396, 32 },
		{ 214, 397, 32 },
		{ 214, 398, 32 },
		{ 214, 399, 32 },
		{ 214, 400, 32 },
		{ 214, 401, 32 },
		{ 214, 402, 32 },
		{ 214, 403, 32 },
		{ 214, 404, 32 },
		{ 214, 405, 32 },
		{ 214, 406, 32 },
		{ 214, 407, 32 },
		{ 214, 408, 32 },
		{ 214, 409, 32 },
		{ 214, 410, 32 },
		{ 214, 411, 32 },
		{ 214, 412, 32 },
		{ 214, 413, 32 },
		{ 214, 414, 32 },
		{ 214, 415, 32 },
		{ 214, 416, 32 },
		{ 214, 417, 32 },
		{ 214, 418, 32 },
		{ 214, 419, 32 },
		{ 214, 420, 32 },
		{ 214, 421, 32 },
		{ 214, 422, 32 },
		{ 214, 423, 32 },
		{ 214, 424, 32 },
		{ 214, 425, 32 },
		{ 214, 426, 32 },
		{ 214, 427, 32 },
		{ 214, 428, 32 },
		{ 214, 429, 32 },
		{ 214, 430, 32 },
		{ 214, 431, 32 },
		{ 214, 432, 32 },
		{ 214, 433, 32 },
		{ 214, 434, 32 },
		{ 214, 435, 32 },
		{ 214, 436, 32 },
		{ 214, 437, 32 },
		{ 214, 438, 32 },
		{ 214, 439, 32 },
		{ 214, 440, 32 },
		{ 214, 441, 32 },
		{ 214, 442, 32 },
		{ 214, 443, 32 },
		{ 214, 444, 32 },
		{ 214, 445, 32 },
		{ 214, 446, 32 },
		{ 214, 447, 32 },
		{ 214, 448, 32 },
		{ 214, 449, 32 },
		{ 214, 450, 32 },
		{ 214, 451, 32 },
		{ 214, 452, 32 },
		{ 214, 453, 32 },
		{ 214, 454, 32 },
		{ 214, 455, 32 },
		{ 214, 456, 32 },
		{ 214, 457, 32 },
		{ 214, 458, 32 },
		{ 214, 459, 32 },
		{ 214, 460, 32 },
		{ 214, 461, 32 },
		{ 214, 462, 32 },
		{ 214, 463, 32 },
		{ 214, 464, 32 },
		{ 214, 465, 32 },
		{ 214, 466, 32 },
		{ 214, 467, 32 },
		{ 214, 468, 32 },
		{ 214, 469, 32 },
		{ 214, 470, 32 },
		{ 214, 471, 32 },
		{ 214, 472, 32 },
		{ 214, 473, 32 },
		{ 214, 474, 32 },
		{ 214, 475, 32 },
		{ 214, 476, 32 },
		{ 214, 477, 32 },
		{ 214, 478, 32 },
		{ 214, 479, 32 },
		{ 214, 480, 32 },
		{ 214, 481, 32 },
		{ 214, 482, 32 },
		{ 214, 483, 32 },
		{ 214, 484, 32 },
		{ 214, 485, 32 },
		{ 214, 486, 32 },
		{ 214, 487, 32 },
		{ 214, 488, 32 },
		{ 214, 489, 32 },
		{ 214, 490, 32 },
		{ 214, 491, 32 },
		{ 214, 492, 32 },
		{ 214, 493, 32 },
		{ 214, 494, 32 },
		{ 214, 495, 32 },
		{ 214, 496, 32 },
		{ 214, 497, 32 },
		{ 214, 498, 32 },
		{ 214, 499, 32 },
		{ 214, 500, 32 },
		{ 214, 501, 32 },
		{ 214, 502, 32 },
		{ 214, 503, 32 },
		{ 214, 504, 32 },
		{ 214, 505, 32 },
		{ 214, 506, 32 },
		{ 214, 507, 32 },
		{ 214, 508, 32 },
		{ 214, 509, 32 },
		{ 214, 510, 32 },
		{ 214, 511, 32 },
		{ 214, 512, 32 },
		{ 215, 370, 32 },
		{ 215, 371, 32 },
		{ 215, 372, 32 },
		{ 215, 373, 32 },
		{ 215, 374, 32 },
		{ 215, 375, 32 },
		{ 215, 376, 32 },
		{ 215, 377, 32 },
		{ 215, 378, 32 },
		{ 215, 379, 32 },
		{ 215, 380, 32 },
		{ 215, 381, 32 },
		{ 215, 382, 32 },
		{ 215, 383, 32 },
		{ 215, 384, 32 },
		{ 215, 385, 32 },
		{ 215, 386, 32 },
		{ 215, 387, 32 },
		{ 215, 388, 32 },
		{ 215, 389, 32 },
		{ 215, 390, 32 },
		{ 215, 391, 32 },
		{ 215, 392, 32 },
		{ 215, 393, 32 },
		{ 215, 394, 32 },
		{ 215, 395, 32 },
		{ 215, 396, 32 },
		{ 215, 397, 32 },
		{ 215, 398, 32 },
		{ 215, 399, 32 },
		{ 215, 400, 32 },
		{ 215, 401, 32 },
		{ 215, 402, 32 },
		{ 215, 403, 32 },
		{ 215, 404, 32 },
		{ 215, 405, 32 },
		{ 215, 406, 32 },
		{ 215, 407, 32 },
		{ 215, 408, 32 },
		{ 215, 409, 32 },
		{ 215, 410, 32 },
		{ 215, 411, 32 },
		{ 215, 412, 32 },
		{ 215, 413, 32 },
		{ 215, 414, 32 },
		{ 215, 415, 32 },
		{ 215, 416, 32 },
		{ 215, 417, 32 },
		{ 215, 418, 32 },
		{ 215, 419, 32 },
		{ 215, 420, 32 },
		{ 215, 421, 32 },
		{ 215, 422, 32 },
		{ 215, 423, 32 },
		{ 215, 424, 32 },
		{ 215, 425, 32 },
		{ 215, 426, 32 },
		{ 215, 427, 32 },
		{ 215, 428, 32 },
		{ 215, 429, 32 },
		{ 215, 430, 32 },
		{ 215, 431, 32 },
		{ 215, 432, 32 },
		{ 215, 433, 32 },
		{ 215, 434, 32 },
		{ 215, 435, 32 },
		{ 215, 436, 32 },
		{ 215, 437, 32 },
		{ 215, 438, 32 },
		{ 215, 439, 32 },
		{ 215, 440, 32 },
		{ 215, 441, 32 },
		{ 215, 442, 32 },
		{ 215, 443, 32 },
		{ 215, 444, 32 },
		{ 215, 445, 32 },
		{ 215, 446, 32 },
		{ 215, 447, 32 },
		{ 215, 448, 32 },
		{ 215, 449, 32 },
		{ 215, 450, 32 },
		{ 215, 451, 32 },
		{ 215, 452, 32 },
		{ 215, 453, 32 },
		{ 215, 454, 32 },
		{ 215, 455, 32 },
		{ 215, 456, 32 },
		{ 215, 457, 32 },
		{ 215, 458, 32 },
		{ 215, 459, 32 },
		{ 215, 460, 32 },
		{ 215, 461, 32 },
		{ 215, 462, 32 },
		{ 215, 463, 32 },
		{ 215, 464, 32 },
		{ 215, 465, 32 },
		{ 215, 466, 32 },
		{ 215, 467, 32 },
		{ 215, 468, 32 },
		{ 215, 469, 32 },
		{ 215, 470, 32 },
		{ 215, 471, 32 },
		{ 215, 472, 32 },
		{ 215, 473, 32 },
		{ 215, 474, 32 },
		{ 215, 475, 32 },
		{ 215, 476, 32 },
		{ 215, 477, 32 },
		{ 215, 478, 32 },
		{ 215, 479, 32 },
		{ 215, 480, 32 },
		{ 215, 481, 32 },
		{ 215, 482, 32 },
		{ 215, 483, 32 },
		{ 215, 484, 32 },
		{ 215, 485, 32 },
		{ 215, 486, 32 },
		{ 215, 487, 32 },
		{ 215, 488, 32 },
		{ 215, 489, 32 },
		{ 215, 490, 32 },
		{ 215, 491, 32 },
		{ 215, 492, 32 },
		{ 215, 493, 32 },
		{ 215, 494, 32 },
		{ 215, 495, 32 },
		{ 215, 496, 32 },
		{ 215, 497, 32 },
		{ 215, 498, 32 },
		{ 215, 499, 32 },
		{ 215, 500, 32 },
		{ 215, 501, 32 },
		{ 215, 502, 32 },
		{ 215, 503, 32 },
		{ 215, 504, 32 },
		{ 215, 505, 32 },
		{ 215, 506, 32 },
		{ 215, 507, 32 },
		{ 215, 508, 32 },
		{ 215, 509, 32 },
		{ 215, 510, 32 },
		{ 215, 511, 32 },
		{ 215, 512, 32 },
		{ 216, 367, 32 },
		{ 216, 368, 32 },
		{ 216, 369, 32 },
		{ 216, 370, 32 },
		{ 216, 371, 32 },
		{ 216, 372, 32 },
		{ 216, 373, 32 },
		{ 216, 374, 32 },
		{ 216, 375, 32 },
		{ 216, 376, 32 },
		{ 216, 377, 32 },
		{ 216, 378, 32 },
		{ 216, 379, 32 },
		{ 216, 380, 32 },
		{ 216, 381, 32 },
		{ 216, 382, 32 },
		{ 216, 383, 32 },
		{ 216, 384, 32 },
		{ 216, 385, 32 },
		{ 216, 386, 32 },
		{ 216, 387, 32 },
		{ 216, 388, 32 },
		{ 216, 389, 32 },
		{ 216, 390, 32 },
		{ 216, 391, 32 },
		{ 216, 392, 32 },
		{ 216, 393, 32 },
		{ 216, 394, 32 },
		{ 216, 395, 32 },
		{ 216, 396, 32 },
		{ 216, 397, 32 },
		{ 216, 398, 32 },
		{ 216, 399, 32 },
		{ 216, 400, 32 },
		{ 216, 401, 32 },
		{ 216, 402, 32 },
		{ 216, 403, 32 },
		{ 216, 404, 32 },
		{ 216, 405, 32 },
		{ 216, 406, 32 },
		{ 216, 407, 32 },
		{ 216, 408, 32 },
		{ 216, 409, 32 },
		{ 216, 410, 32 },
		{ 216, 411, 32 },
		{ 216, 412, 32 },
		{ 216, 413, 32 },
		{ 216, 414, 32 },
		{ 216, 415, 32 },
		{ 216, 416, 32 },
		{ 216, 417, 32 },
		{ 216, 418, 32 },
		{ 216, 419, 32 },
		{ 216, 420, 32 },
		{ 216, 421, 32 },
		{ 216, 422, 32 },
		{ 216, 423, 32 },
		{ 216, 424, 32 },
		{ 216, 425, 32 },
		{ 216, 426, 32 },
		{ 216, 427, 32 },
		{ 216, 428, 32 },
		{ 216, 429, 32 },
		{ 216, 430, 32 },
		{ 216, 431, 32 },
		{ 216, 432, 32 },
		{ 216, 433, 32 },
		{ 216, 434, 32 },
		{ 216, 435, 32 },
		{ 216, 436, 32 },
		{ 216, 437, 32 },
		{ 216, 438, 32 },
		{ 216, 439, 32 },
		{ 216, 440, 32 },
		{ 216, 441, 32 },
		{ 216, 442, 32 },
		{ 216, 443, 32 },
		{ 216, 444, 32 },
		{ 216, 445, 32 },
		{ 216, 446, 32 },
		{ 216, 447, 32 },
		{ 216, 448, 32 },
		{ 216, 449, 32 },
		{ 216, 450, 32 },
		{ 216, 451, 32 },
		{ 216, 452, 32 },
		{ 216, 453, 32 },
		{ 216, 454, 32 },
		{ 216, 455, 32 },
		{ 216, 456, 32 },
		{ 216, 457, 32 },
		{ 216, 458, 32 },
		{ 216, 459, 32 },
		{ 216, 460, 32 },
		{ 216, 461, 32 },
		{ 216, 462, 32 },
		{ 216, 463, 32 },
		{ 216, 464, 32 },
		{ 216, 465, 32 },
		{ 216, 466, 32 },
		{ 216, 467, 32 },
		{ 216, 468, 32 },
		{ 216, 469, 32 },
		{ 216, 470, 32 },
		{ 216, 471, 32 },
		{ 216, 472, 32 },
		{ 216, 473, 32 },
		{ 216, 474, 32 },
		{ 216, 475, 32 },
		{ 216, 476, 32 },
		{ 216, 477, 32 },
		{ 216, 478, 32 },
		{ 216, 479, 32 },
		{ 216, 480, 32 },
		{ 216, 481, 32 },
		{ 216, 482, 32 },
		{ 216, 483, 32 },
		{ 216, 484, 32 },
		{ 216, 485, 32 },
		{ 216, 486, 32 },
		{ 216, 487, 32 },
		{ 216, 488, 32 },
		{ 216, 489, 32 },
		{ 216, 490, 32 },
		{ 216, 491, 32 },
		{ 216, 492, 32 },
		{ 216, 493, 32 },
		{ 216, 494, 32 },
		{ 216, 495, 32 },
		{ 216, 496, 32 },
		{ 216, 497, 32 },
		{ 216, 498, 32 },
		{ 216, 499, 32 },
		{ 216, 500, 32 },
		{ 216, 501, 32 },
		{ 216, 502, 32 },
		{ 216, 503, 32 },
		{ 216, 504, 32 },
		{ 216, 505, 32 },
		{ 216, 506, 32 },
		{ 216, 507, 32 },
		{ 216, 508, 32 },
		{ 216, 509, 32 },
		{ 216, 510, 32 },
		{ 216, 511, 32 },
		{ 216, 512, 32 },
		{ 217, 364, 32 },
		{ 217, 365, 32 },
		{ 217, 366, 32 },
		{ 217, 367, 32 },
		{ 217, 368, 32 },
		{ 217, 369, 32 },
		{ 217, 370, 32 },
		{ 217, 371, 32 },
		{ 217, 372, 32 },
		{ 217, 373, 32 },
		{ 217, 374, 32 },
		{ 217, 375, 32 },
		{ 217, 376, 32 },
		{ 217, 377, 32 },
		{ 217, 378, 32 },
		{ 217, 379, 32 },
		{ 217, 380, 32 },
		{ 217, 381, 32 },
		{ 217, 382, 32 },
		{ 217, 383, 32 },
		{ 217, 384, 32 },
		{ 217, 385, 32 },
		{ 217, 386, 32 },
		{ 217, 387, 32 },
		{ 217, 388, 32 },
		{ 217, 389, 32 },
		{ 217, 390, 32 },
		{ 217, 391, 32 },
		{ 217, 392, 32 },
		{ 217, 393, 32 },
		{ 217, 394, 32 },
		{ 217, 395, 32 },
		{ 217, 396, 32 },
		{ 217, 397, 32 },
		{ 217, 398, 32 },
		{ 217, 399, 32 },
		{ 217, 400, 32 },
		{ 217, 401, 32 },
		{ 217, 402, 32 },
		{ 217, 403, 32 },
		{ 217, 404, 32 },
		{ 217, 405, 32 },
		{ 217, 406, 32 },
		{ 217, 407, 32 },
		{ 217, 408, 32 },
		{ 217, 409, 32 },
		{ 217, 410, 32 },
		{ 217, 411, 32 },
		{ 217, 412, 32 },
		{ 217, 413, 32 },
		{ 217, 414, 32 },
		{ 217, 415, 32 },
		{ 217, 416, 32 },
		{ 217, 417, 32 },
		{ 217, 418, 32 },
		{ 217, 419, 32 },
		{ 217, 420, 32 },
		{ 217, 421, 32 },
		{ 217, 422, 32 },
		{ 217, 423, 32 },
		{ 217, 424, 32 },
		{ 217, 425, 32 },
		{ 217, 426, 32 },
		{ 217, 427, 32 },
		{ 217, 428, 32 },
		{ 217, 429, 32 },
		{ 217, 430, 32 },
		{ 217, 431, 32 },
		{ 217, 432, 32 },
		{ 217, 433, 32 },
		{ 217, 434, 32 },
		{ 217, 435, 32 },
		{ 217, 436, 32 },
		{ 217, 437, 32 },
		{ 217, 438, 32 },
		{ 217, 439, 32 },
		{ 217, 440, 32 },
		{ 217, 441, 32 },
		{ 217, 442, 32 },
		{ 217, 443, 32 },
		{ 217, 444, 32 },
		{ 217, 445, 32 },
		{ 217, 446, 32 },
		{ 217, 447, 32 },
		{ 217, 448, 32 },
		{ 217, 449, 32 },
		{ 217, 450, 32 },
		{ 217, 451, 32 },
		{ 217, 452, 32 },
		{ 217, 453, 32 },
		{ 217, 454, 32 },
		{ 217, 455, 32 },
		{ 217, 456, 32 },
		{ 217, 457, 32 },
		{ 217, 458, 32 },
		{ 217, 459, 32 },
		{ 217, 460, 32 },
		{ 217, 461, 32 },
		{ 217, 462, 32 },
		{ 217, 463, 32 },
		{ 217, 464, 32 },
		{ 217, 465, 32 },
		{ 217, 466, 32 },
		{ 217, 467, 32 },
		{ 217, 468, 32 },
		{ 217, 469, 32 },
		{ 217, 470, 32 },
		{ 217, 471, 32 },
		{ 217, 472, 32 },
		{ 217, 473, 32 },
		{ 217, 474, 32 },
		{ 217, 475, 32 },
		{ 217, 476, 32 },
		{ 217, 477, 32 },
		{ 217, 478, 32 },
		{ 217, 479, 32 },
		{ 217, 480, 32 },
		{ 217, 481, 32 },
		{ 217, 482, 32 },
		{ 217, 483, 32 },
		{ 217, 484, 32 },
		{ 217, 485, 32 },
		{ 217, 486, 32 },
		{ 217, 487, 32 },
		{ 217, 488, 32 },
		{ 217, 489, 32 },
		{ 217, 490, 32 },
		{ 217, 491, 32 },
		{ 217, 492, 32 },
		{ 217, 493, 32 },
		{ 217, 494, 32 },
		{ 217, 495, 32 },
		{ 217, 496, 32 },
		{ 217, 497, 32 },
		{ 217, 498, 32 },
		{ 217, 499, 32 },
		{ 217, 500, 32 },
		{ 217, 501, 32 },
		{ 217, 502, 32 },
		{ 217, 503, 32 },
		{ 217, 504, 32 },
		{ 217, 505, 32 },
		{ 217, 506, 32 },
		{ 217, 507, 32 },
		{ 217, 508, 32 },
		{ 217, 509, 32 },
		{ 217, 510, 32 },
		{ 217, 511, 32 },
		{ 217, 512, 32 },
		{ 218, 362, 32 },
		{ 218, 363, 32 },
		{ 218, 364, 32 },
		{ 218, 365, 32 },
		{ 218, 366, 32 },
		{ 218, 367, 32 },
		{ 218, 368, 32 },
		{ 218, 369, 32 },
		{ 218, 370, 32 },
		{ 218, 371, 32 },
		{ 218, 372, 32 },
		{ 218, 373, 32 },
		{ 218, 374, 32 },
		{ 218, 375, 32 },
		{ 218, 376, 32 },
		{ 218, 377, 32 },
		{ 218, 378, 32 },
		{ 218, 379, 32 },
		{ 218, 380, 32 },
		{ 218, 381, 32 },
		{ 218, 382, 32 },
		{ 218, 383, 32 },
		{ 218, 384, 32 },
		{ 218, 385, 32 },
		{ 218, 386, 32 },
		{ 218, 387, 32 },
		{ 218, 388, 32 },
		{ 218, 389, 32 },
		{ 218, 390, 32 },
		{ 218, 391, 32 },
		{ 218, 392, 32 },
		{ 218, 393, 32 },
		{ 218, 394, 32 },
		{ 218, 395, 32 },
		{ 218, 396, 32 },
		{ 218, 397, 32 },
		{ 218, 398, 32 },
		{ 218, 399, 32 },
		{ 218, 400, 32 },
		{ 218, 401, 32 },
		{ 218, 402, 32 },
		{ 218, 403, 32 },
		{ 218, 404, 32 },
		{ 218, 405, 32 },
		{ 218, 406, 32 },
		{ 218, 407, 32 },
		{ 218, 408, 32 },
		{ 218, 409, 32 },
		{ 218, 410, 32 },
		{ 218, 411, 32 },
		{ 218, 412, 32 },
		{ 218, 413, 32 },
		{ 218, 414, 32 },
		{ 218, 415, 32 },
		{ 218, 416, 32 },
		{ 218, 417, 32 },
		{ 218, 418, 32 },
		{ 218, 419, 32 },
		{ 218, 420, 32 },
		{ 218, 421, 32 },
		{ 218, 422, 32 },
		{ 218, 423, 32 },
		{ 218, 424, 32 },
		{ 218, 425, 32 },
		{ 218, 426, 32 },
		{ 218, 427, 32 },
		{ 218, 428, 32 },
		{ 218, 429, 32 },
		{ 218, 430, 32 },
		{ 218, 431, 32 },
		{ 218, 432, 32 },
		{ 218, 433, 32 },
		{ 218, 434, 32 },
		{ 218, 435, 32 },
		{ 218, 436, 32 },
		{ 218, 437, 32 },
		{ 218, 438, 32 },
		{ 218, 439, 32 },
		{ 218, 440, 32 },
		{ 218, 441, 32 },
		{ 218, 442, 32 },
		{ 218, 443, 32 },
		{ 218, 444, 32 },
		{ 218, 445, 32 },
		{ 218, 446, 32 },
		{ 218, 447, 32 },
		{ 218, 448, 32 },
		{ 218, 449, 32 },
		{ 218, 450, 32 },
		{ 218, 451, 32 },
		{ 218, 452, 32 },
		{ 218, 453, 32 },
		{ 218, 454, 32 },
		{ 218, 455, 32 },
		{ 218, 456, 32 },
		{ 218, 457, 32 },
		{ 218, 458, 32 },
		{ 218, 459, 32 },
		{ 218, 460, 32 },
		{ 218, 461, 32 },
		{ 218, 462, 32 },
		{ 218, 463, 32 },
		{ 218, 464, 32 },
		{ 218, 465, 32 },
		{ 218, 466, 32 },
		{ 218, 467, 32 },
		{ 218, 468, 32 },
		{ 218, 469, 32 },
		{ 218, 470, 32 },
		{ 218, 471, 32 },
		{ 218, 472, 32 },
		{ 218, 473, 32 },
		{ 218, 474, 32 },
		{ 218, 475, 32 },
		{ 218, 476, 32 },
		{ 218, 477, 32 },
		{ 218, 478, 32 },
		{ 218, 479, 32 },
		{ 218, 480, 32 },
		{ 218, 481, 32 },
		{ 218, 482, 32 },
		{ 218, 483, 32 },
		{ 218, 484, 32 },
		{ 218, 485, 32 },
		{ 218, 486, 32 },
		{ 218, 487, 32 },
		{ 218, 488, 32 },
		{ 218, 489, 32 },
		{ 218, 490, 32 },
		{ 218, 491, 32 },
		{ 218, 492, 32 },
		{ 218, 493, 32 },
		{ 218, 494, 32 },
		{ 218, 495, 32 },
		{ 218, 496, 32 },
		{ 218, 497, 32 },
		{ 218, 498, 32 },
		{ 218, 499, 32 },
		{ 218, 500, 32 },
		{ 218, 501, 32 },
		{ 218, 502, 32 },
		{ 218, 503, 32 },
		{ 218, 504, 32 },
		{ 218, 505, 32 },
		{ 218, 506, 32 },
		{ 218, 507, 32 },
		{ 218, 508, 32 },
		{ 218, 509, 32 },
		{ 218, 510, 32 },
		{ 218, 511, 32 },
		{ 218, 512, 32 },
		{ 219, 359, 32 },
		{ 219, 360, 32 },
		{ 219, 361, 32 },
		{ 219, 362, 32 },
		{ 219, 363, 32 },
		{ 219, 364, 32 },
		{ 219, 365, 32 },
		{ 219, 366, 32 },
		{ 219, 367, 32 },
		{ 219, 368, 32 },
		{ 219, 369, 32 },
		{ 219, 370, 32 },
		{ 219, 371, 32 },
		{ 219, 372, 32 },
		{ 219, 373, 32 },
		{ 219, 374, 32 },
		{ 219, 375, 32 },
		{ 219, 376, 32 },
		{ 219, 377, 32 },
		{ 219, 378, 32 },
		{ 219, 379, 32 },
		{ 219, 380, 32 },
		{ 219, 381, 32 },
		{ 219, 382, 32 },
		{ 219, 383, 32 },
		{ 219, 384, 32 },
		{ 219, 385, 32 },
		{ 219, 386, 32 },
		{ 219, 387, 32 },
		{ 219, 388, 32 },
		{ 219, 389, 32 },
		{ 219, 390, 32 },
		{ 219, 391, 32 },
		{ 219, 392, 32 },
		{ 219, 393, 32 },
		{ 219, 394, 32 },
		{ 219, 395, 32 },
		{ 219, 396, 32 },
		{ 219, 397, 32 },
		{ 219, 398, 32 },
		{ 219, 399, 32 },
		{ 219, 400, 32 },
		{ 219, 401, 32 },
		{ 219, 402, 32 },
		{ 219, 403, 32 },
		{ 219, 404, 32 },
		{ 219, 405, 32 },
		{ 219, 406, 32 },
		{ 219, 407, 32 },
		{ 219, 408, 32 },
		{ 219, 409, 32 },
		{ 219, 410, 32 },
		{ 219, 411, 32 },
		{ 219, 412, 32 },
		{ 219, 413, 32 },
		{ 219, 414, 32 },
		{ 219, 415, 32 },
		{ 219, 416, 32 },
		{ 219, 417, 32 },
		{ 219, 418, 32 },
		{ 219, 419, 32 },
		{ 219, 420, 32 },
		{ 219, 421, 32 },
		{ 219, 422, 32 },
		{ 219, 423, 32 },
		{ 219, 424, 32 },
		{ 219, 425, 32 },
		{ 219, 426, 32 },
		{ 219, 427, 32 },
		{ 219, 428, 32 },
		{ 219, 429, 32 },
		{ 219, 430, 32 },
		{ 219, 431, 32 },
		{ 219, 432, 32 },
		{ 219, 433, 32 },
		{ 219, 434, 32 },
		{ 219, 435, 32 },
		{ 219, 436, 32 },
		{ 219, 437, 32 },
		{ 219, 438, 32 },
		{ 219, 439, 32 },
		{ 219, 440, 32 },
		{ 219, 441, 32 },
		{ 219, 442, 32 },
		{ 219, 443, 32 },
		{ 219, 444, 32 },
		{ 219, 445, 32 },
		{ 219, 446, 32 },
		{ 219, 447, 32 },
		{ 219, 448, 32 },
		{ 219, 449, 32 },
		{ 219, 450, 32 },
		{ 219, 451, 32 },
		{ 219, 452, 32 },
		{ 219, 453, 32 },
		{ 219, 454, 32 },
		{ 219, 455, 32 },
		{ 219, 456, 32 },
		{ 219, 457, 32 },
		{ 219, 458, 32 },
		{ 219, 459, 32 },
		{ 219, 460, 32 },
		{ 219, 461, 32 },
		{ 219, 462, 32 },
		{ 219, 463, 32 },
		{ 219, 464, 32 },
		{ 219, 465, 32 },
		{ 219, 466, 32 },
		{ 219, 467, 32 },
		{ 219, 468, 32 },
		{ 219, 469, 32 },
		{ 219, 470, 32 },
		{ 219, 471, 32 },
		{ 219, 472, 32 },
		{ 219, 473, 32 },
		{ 219, 474, 32 },
		{ 219, 475, 32 },
		{ 219, 476, 32 },
		{ 219, 477, 32 },
		{ 219, 478, 32 },
		{ 219, 479, 32 },
		{ 219, 480, 32 },
		{ 219, 481, 32 },
		{ 219, 482, 32 },
		{ 219, 483, 32 },
		{ 219, 484, 32 },
		{ 219, 485, 32 },
		{ 219, 486, 32 },
		{ 219, 487, 32 },
		{ 219, 488, 32 },
		{ 219, 489, 32 },
		{ 219, 490, 32 },
		{ 219, 491, 32 },
		{ 219, 492, 32 },
		{ 219, 493, 32 },
		{ 219, 494, 32 },
		{ 219, 495, 32 },
		{ 219, 496, 32 },
		{ 219, 497, 32 },
		{ 219, 498, 32 },
		{ 219, 499, 32 },
		{ 219, 500, 32 },
		{ 219, 501, 32 },
		{ 219, 502, 32 },
		{ 219, 503, 32 },
		{ 219, 504, 32 },
		{ 219, 505, 32 },
		{ 219, 506, 32 },
		{ 219, 507, 32 },
		{ 219, 508, 32 },
		{ 219, 509, 32 },
		{ 219, 510, 32 },
		{ 219, 511, 32 },
		{ 219, 512, 32 },
		{ 220, 356, 32 },
		{ 220, 357, 32 },
		{ 220, 358, 32 },
		{ 220, 359, 32 },
		{ 220, 360, 32 },
		{ 220, 361, 32 },
		{ 220, 362, 32 },
		{ 220, 363, 32 },
		{ 220, 364, 32 },
		{ 220, 365, 32 },
		{ 220, 366, 32 },
		{ 220, 367, 32 },
		{ 220, 368, 32 },
		{ 220, 369, 32 },
		{ 220, 370, 32 },
		{ 220, 371, 32 },
		{ 220, 372, 32 },
		{ 220, 373, 32 },
		{ 220, 374, 32 },
		{ 220, 375, 32 },
		{ 220, 376, 32 },
		{ 220, 377, 32 },
		{ 220, 378, 32 },
		{ 220, 379, 32 },
		{ 220, 380, 32 },
		{ 220, 381, 32 },
		{ 220, 382, 32 },
		{ 220, 383, 32 },
		{ 220, 384, 32 },
		{ 220, 385, 32 },
		{ 220, 386, 32 },
		{ 220, 387, 32 },
		{ 220, 388, 32 },
		{ 220, 389, 32 },
		{ 220, 390, 32 },
		{ 220, 391, 32 },
		{ 220, 392, 32 },
		{ 220, 393, 32 },
		{ 220, 394, 32 },
		{ 220, 395, 32 },
		{ 220, 396, 32 },
		{ 220, 397, 32 },
		{ 220, 398, 32 },
		{ 220, 399, 32 },
		{ 220, 400, 32 },
		{ 220, 401, 32 },
		{ 220, 402, 32 },
		{ 220, 403, 32 },
		{ 220, 404, 32 },
		{ 220, 405, 32 },
		{ 220, 406, 32 },
		{ 220, 407, 32 },
		{ 220, 408, 32 },
		{ 220, 409, 32 },
		{ 220, 410, 32 },
		{ 220, 411, 32 },
		{ 220, 412, 32 },
		{ 220, 413, 32 },
		{ 220, 414, 32 },
		{ 220, 415, 32 },
		{ 220, 416, 32 },
		{ 220, 417, 32 },
		{ 220, 418, 32 },
		{ 220, 419, 32 },
		{ 220, 420, 32 },
		{ 220, 421, 32 },
		{ 220, 422, 32 },
		{ 220, 423, 32 },
		{ 220, 424, 32 },
		{ 220, 425, 32 },
		{ 220, 426, 32 },
		{ 220, 427, 32 },
		{ 220, 428, 32 },
		{ 220, 429, 32 },
		{ 220, 430, 32 },
		{ 220, 431, 32 },
		{ 220, 432, 32 },
		{ 220, 433, 32 },
		{ 220, 434, 32 },
		{ 220, 435, 32 },
		{ 220, 436, 32 },
		{ 220, 437, 32 },
		{ 220, 438, 32 },
		{ 220, 439, 32 },
		{ 220, 440, 32 },
		{ 220, 441, 32 },
		{ 220, 442, 32 },
		{ 220, 443, 32 },
		{ 220, 444, 32 },
		{ 220, 445, 32 },
		{ 220, 446, 32 },
		{ 220, 447, 32 },
		{ 220, 448, 32 },
		{ 220, 449, 32 },
		{ 220, 450, 32 },
		{ 220, 451, 32 },
		{ 220, 452, 32 },
		{ 220, 453, 32 },
		{ 220, 454, 32 },
		{ 220, 455, 32 },
		{ 220, 456, 32 },
		{ 220, 457, 32 },
		{ 220, 458, 32 },
		{ 220, 459, 32 },
		{ 220, 460, 32 },
		{ 220, 461, 32 },
		{ 220, 462, 32 },
		{ 220, 463, 32 },
		{ 220, 464, 32 },
		{ 220, 465, 32 },
		{ 220, 466, 32 },
		{ 220, 467, 32 },
		{ 220, 468, 32 },
		{ 220, 469, 32 },
		{ 220, 470, 32 },
		{ 220, 471, 32 },
		{ 220, 472, 32 },
		{ 220, 473, 32 },
		{ 220, 474, 32 },
		{ 220, 475, 32 },
		{ 220, 476, 32 },
		{ 220, 477, 32 },
		{ 220, 478, 32 },
		{ 220, 479, 32 },
		{ 220, 480, 32 },
		{ 220, 481, 32 },
		{ 220, 482, 32 },
		{ 220, 483, 32 },
		{ 220, 484, 32 },
		{ 220, 485, 32 },
		{ 220, 486, 32 },
		{ 220, 487, 32 },
		{ 220, 488, 32 },
		{ 220, 489, 32 },
		{ 220, 490, 32 },
		{ 220, 491, 32 },
		{ 220, 492, 32 },
		{ 220, 493, 32 },
		{ 220, 494, 32 },
		{ 220, 495, 32 },
		{ 220, 496, 32 },
		{ 220, 497, 32 },
		{ 220, 498, 32 },
		{ 220, 499, 32 },
		{ 220, 500, 32 },
		{ 220, 501, 32 },
		{ 220, 502, 32 },
		{ 220, 503, 32 },
		{ 220, 504, 32 },
		{ 220, 505, 32 },
		{ 220, 506, 32 },
		{ 220, 507, 32 },
		{ 220, 508, 32 },
		{ 220, 509, 32 },
		{ 220, 510, 32 },
		{ 220, 511, 32 },
		{ 220, 512, 32 },
		{ 221, 354, 32 },
		{ 221, 355, 32 },
		{ 221, 356, 32 },
		{ 221, 357, 32 },
		{ 221, 358, 32 },
		{ 221, 359, 32 },
		{ 221, 360, 32 },
		{ 221, 361, 32 },
		{ 221, 362, 32 },
		{ 221, 363, 32 },
		{ 221, 364, 32 },
		{ 221, 365, 32 },
		{ 221, 366, 32 },
		{ 221, 367, 32 },
		{ 221, 368, 32 },
		{ 221, 369, 32 },
		{ 221, 370, 32 },
		{ 221, 371, 32 },
		{ 221, 372, 32 },
		{ 221, 373, 32 },
		{ 221, 374, 32 },
		{ 221, 375, 32 },
		{ 221, 376, 32 },
		{ 221, 377, 32 },
		{ 221, 378, 32 },
		{ 221, 379, 32 },
		{ 221, 380, 32 },
		{ 221, 381, 32 },
		{ 221, 382, 32 },
		{ 221, 383, 32 },
		{ 221, 384, 32 },
		{ 221, 385, 32 },
		{ 221, 386, 32 },
		{ 221, 387, 32 },
		{ 221, 388, 32 },
		{ 221, 389, 32 },
		{ 221, 390, 32 },
		{ 221, 391, 32 },
		{ 221, 392, 32 },
		{ 221, 393, 32 },
		{ 221, 394, 32 },
		{ 221, 395, 32 },
		{ 221, 396, 32 },
		{ 221, 397, 32 },
		{ 221, 398, 32 },
		{ 221, 399, 32 },
		{ 221, 400, 32 },
		{ 221, 401, 32 },
		{ 221, 402, 32 },
		{ 221, 403, 32 },
		{ 221, 404, 32 },
		{ 221, 405, 32 },
		{ 221, 406, 32 },
		{ 221, 407, 32 },
		{ 221, 408, 32 },
		{ 221, 409, 32 },
		{ 221, 410, 32 },
		{ 221, 411, 32 },
		{ 221, 412, 32 },
		{ 221, 413, 32 },
		{ 221, 414, 32 },
		{ 221, 415, 32 },
		{ 221, 416, 32 },
		{ 221, 417, 32 },
		{ 221, 418, 32 },
		{ 221, 419, 32 },
		{ 221, 420, 32 },
		{ 221, 421, 32 },
		{ 221, 422, 32 },
		{ 221, 423, 32 },
		{ 221, 424, 32 },
		{ 221, 425, 32 },
		{ 221, 426, 32 },
		{ 221, 427, 32 },
		{ 221, 428, 32 },
		{ 221, 429, 32 },
		{ 221, 430, 32 },
		{ 221, 431, 32 },
		{ 221, 432, 32 },
		{ 221, 433, 32 },
		{ 221, 434, 32 },
		{ 221, 435, 32 },
		{ 221, 436, 32 },
		{ 221, 437, 32 },
		{ 221, 438, 32 },
		{ 221, 439, 32 },
		{ 221, 440, 32 },
		{ 221, 441, 32 },
		{ 221, 442, 32 },
		{ 221, 443, 32 },
		{ 221, 444, 32 },
		{ 221, 445, 32 },
		{ 221, 446, 32 },
		{ 221, 447, 32 },
		{ 221, 448, 32 },
		{ 221, 449, 32 },
		{ 221, 450, 32 },
		{ 221, 451, 32 },
		{ 221, 452, 32 },
		{ 221, 453, 32 },
		{ 221, 454, 32 },
		{ 221, 455, 32 },
		{ 221, 456, 32 },
		{ 221, 457, 32 },
		{ 221, 458, 32 },
		{ 221, 459, 32 },
		{ 221, 460, 32 },
		{ 221, 461, 32 },
		{ 221, 462, 32 },
		{ 221, 463, 32 },
		{ 221, 464, 32 },
		{ 221, 465, 32 },
		{ 221, 466, 32 },
		{ 221, 467, 32 },
		{ 221, 468, 32 },
		{ 221, 469, 32 },
		{ 221, 470, 32 },
		{ 221, 471, 32 },
		{ 221, 472, 32 },
		{ 221, 473, 32 },
		{ 221, 474, 32 },
		{ 221, 475, 32 },
		{ 221, 476, 32 },
		{ 221, 477, 32 },
		{ 221, 478, 32 },
		{ 221, 479, 32 },
		{ 221, 480, 32 },
		{ 221, 481, 32 },
		{ 221, 482, 32 },
		{ 221, 483, 32 },
		{ 221, 484, 32 },
		{ 221, 485, 32 },
		{ 221, 486, 32 },
		{ 221, 487, 32 },
		{ 221, 488, 32 },
		{ 221, 489, 32 },
		{ 221, 490, 32 },
		{ 221, 491, 32 },
		{ 221, 492, 32 },
		{ 221, 493, 32 },
		{ 221, 494, 32 },
		{ 221, 495, 32 },
		{ 221, 496, 32 },
		{ 221, 497, 32 },
		{ 221, 498, 32 },
		{ 221, 499, 32 },
		{ 221, 500, 32 },
		{ 221, 501, 32 },
		{ 221, 502, 32 },
		{ 221, 503, 32 },
		{ 221, 504, 32 },
		{ 221, 505, 32 },
		{ 221, 506, 32 },
		{ 221, 507, 32 },
		{ 221, 508, 32 },
		{ 221, 509, 32 },
		{ 221, 510, 32 },
		{ 221, 511, 32 },
		{ 221, 512, 32 },
		{ 222, 351, 32 },
		{ 222, 352, 32 },
		{ 222, 353, 32 },
		{ 222, 354, 32 },
		{ 222, 355, 32 },
		{ 222, 356, 32 },
		{ 222, 357, 32 },
		{ 222, 358, 32 },
		{ 222, 359, 32 },
		{ 222, 360, 32 },
		{ 222, 361, 32 },
		{ 222, 362, 32 },
		{ 222, 363, 32 },
		{ 222, 364, 32 },
		{ 222, 365, 32 },
		{ 222, 366, 32 },
		{ 222, 367, 32 },
		{ 222, 368, 32 },
		{ 222, 369, 32 },
		{ 222, 370, 32 },
		{ 222, 371, 32 },
		{ 222, 372, 32 },
		{ 222, 373, 32 },
		{ 222, 374, 32 },
		{ 222, 375, 32 },
		{ 222, 376, 32 },
		{ 222, 377, 32 },
		{ 222, 378, 32 },
		{ 222, 379, 32 },
		{ 222, 380, 32 },
		{ 222, 381, 32 },
		{ 222, 382, 32 },
		{ 222, 383, 32 },
		{ 222, 384, 32 },
		{ 222, 385, 32 },
		{ 222, 386, 32 },
		{ 222, 387, 32 },
		{ 222, 388, 32 },
		{ 222, 389, 32 },
		{ 222, 390, 32 },
		{ 222, 391, 32 },
		{ 222, 392, 32 },
		{ 222, 393, 32 },
		{ 222, 394, 32 },
		{ 222, 395, 32 },
		{ 222, 396, 32 },
		{ 222, 397, 32 },
		{ 222, 398, 32 },
		{ 222, 399, 32 },
		{ 222, 400, 32 },
		{ 222, 401, 32 },
		{ 222, 402, 32 },
		{ 222, 403, 32 },
		{ 222, 404, 32 },
		{ 222, 405, 32 },
		{ 222, 406, 32 },
		{ 222, 407, 32 },
		{ 222, 408, 32 },
		{ 222, 409, 32 },
		{ 222, 410, 32 },
		{ 222, 411, 32 },
		{ 222, 412, 32 },
		{ 222, 413, 32 },
		{ 222, 414, 32 },
		{ 222, 415, 32 },
		{ 222, 416, 32 },
		{ 222, 417, 32 },
		{ 222, 418, 32 },
		{ 222, 419, 32 },
		{ 222, 420, 32 },
		{ 222, 421, 32 },
		{ 222, 422, 32 },
		{ 222, 423, 32 },
		{ 222, 424, 32 },
		{ 222, 425, 32 },
		{ 222, 426, 32 },
		{ 222, 427, 32 },
		{ 222, 428, 32 },
		{ 222, 429, 32 },
		{ 222, 430, 32 },
		{ 222, 431, 32 },
		{ 222, 432, 32 },
		{ 222, 433, 32 },
		{ 222, 434, 32 },
		{ 222, 435, 32 },
		{ 222, 436, 32 },
		{ 222, 437, 32 },
		{ 222, 438, 32 },
		{ 222, 439, 32 },
		{ 222, 440, 32 },
		{ 222, 441, 32 },
		{ 222, 442, 32 },
		{ 222, 443, 32 },
		{ 222, 444, 32 },
		{ 222, 445, 32 },
		{ 222, 446, 32 },
		{ 222, 447, 32 },
		{ 222, 448, 32 },
		{ 222, 449, 32 },
		{ 222, 450, 32 },
		{ 222, 451, 32 },
		{ 222, 452, 32 },
		{ 222, 453, 32 },
		{ 222, 454, 32 },
		{ 222, 455, 32 },
		{ 222, 456, 32 },
		{ 222, 457, 32 },
		{ 222, 458, 32 },
		{ 222, 459, 32 },
		{ 222, 460, 32 },
		{ 222, 461, 32 },
		{ 222, 462, 32 },
		{ 222, 463, 32 },
		{ 222, 464, 32 },
		{ 222, 465, 32 },
		{ 222, 466, 32 },
		{ 222, 467, 32 },
		{ 222, 468, 32 },
		{ 222, 469, 32 },
		{ 222, 470, 32 },
		{ 222, 471, 32 },
		{ 222, 472, 32 },
		{ 222, 473, 32 },
		{ 222, 474, 32 },
		{ 222, 475, 32 },
		{ 222, 476, 32 },
		{ 222, 477, 32 },
		{ 222, 478, 32 },
		{ 222, 479, 32 },
		{ 222, 480, 32 },
		{ 222, 481, 32 },
		{ 222, 482, 32 },
		{ 222, 483, 32 },
		{ 222, 484, 32 },
		{ 222, 485, 32 },
		{ 222, 486, 32 },
		{ 222, 487, 32 },
		{ 222, 488, 32 },
		{ 222, 489, 32 },
		{ 222, 490, 32 },
		{ 222, 491, 32 },
		{ 222, 492, 32 },
		{ 222, 493, 32 },
		{ 222, 494, 32 },
		{ 222, 495, 32 },
		{ 222, 496, 32 },
		{ 222, 497, 32 },
		{ 222, 498, 32 },
		{ 222, 499, 32 },
		{ 222, 500, 32 },
		{ 222, 501, 32 },
		{ 222, 502, 32 },
		{ 222, 503, 32 },
		{ 222, 504, 32 },
		{ 222, 505, 32 },
		{ 222, 506, 32 },
		{ 222, 507, 32 },
		{ 222, 508, 32 },
		{ 222, 509, 32 },
		{ 222, 510, 32 },
		{ 222, 511, 32 },
		{ 222, 512, 32 },
		{ 223, 348, 32 },
		{ 223, 349, 32 },
		{ 223, 350, 32 },
		{ 223, 351, 32 },
		{ 223, 352, 32 },
		{ 223, 353, 32 },
		{ 223, 354, 32 },
		{ 223, 355, 32 },
		{ 223, 356, 32 },
		{ 223, 357, 32 },
		{ 223, 358, 32 },
		{ 223, 359, 32 },
		{ 223, 360, 32 },
		{ 223, 361, 32 },
		{ 223, 362, 32 },
		{ 223, 363, 32 },
		{ 223, 364, 32 },
		{ 223, 365, 32 },
		{ 223, 366, 32 },
		{ 223, 367, 32 },
		{ 223, 368, 32 },
		{ 223, 369, 32 },
		{ 223, 370, 32 },
		{ 223, 371, 32 },
		{ 223, 372, 32 },
		{ 223, 373, 32 },
		{ 223, 374, 32 },
		{ 223, 375, 32 },
		{ 223, 376, 32 },
		{ 223, 377, 32 },
		{ 223, 378, 32 },
		{ 223, 379, 32 },
		{ 223, 380, 32 },
		{ 223, 381, 32 },
		{ 223, 382, 32 },
		{ 223, 383, 32 },
		{ 223, 384, 32 },
		{ 223, 385, 32 },
		{ 223, 386, 32 },
		{ 223, 387, 32 },
		{ 223, 388, 32 },
		{ 223, 389, 32 },
		{ 223, 390, 32 },
		{ 223, 391, 32 },
		{ 223, 392, 32 },
		{ 223, 393, 32 },
		{ 223, 394, 32 },
		{ 223, 395, 32 },
		{ 223, 396, 32 },
		{ 223, 397, 32 },
		{ 223, 398, 32 },
		{ 223, 399, 32 },
		{ 223, 400, 32 },
		{ 223, 401, 32 },
		{ 223, 402, 32 },
		{ 223, 403, 32 },
		{ 223, 404, 32 },
		{ 223, 405, 32 },
		{ 223, 406, 32 },
		{ 223, 407, 32 },
		{ 223, 408, 32 },
		{ 223, 409, 32 },
		{ 223, 410, 32 },
		{ 223, 411, 32 },
		{ 223, 412, 32 },
		{ 223, 413, 32 },
		{ 223, 414, 32 },
		{ 223, 415, 32 },
		{ 223, 416, 32 },
		{ 223, 417, 32 },
		{ 223, 418, 32 },
		{ 223, 419, 32 },
		{ 223, 420, 32 },
		{ 223, 421, 32 },
		{ 223, 422, 32 },
		{ 223, 423, 32 },
		{ 223, 424, 32 },
		{ 223, 425, 32 },
		{ 223, 426, 32 },
		{ 223, 427, 32 },
		{ 223, 428, 32 },
		{ 223, 429, 32 },
		{ 223, 430, 32 },
		{ 223, 431, 32 },
		{ 223, 432, 32 },
		{ 223, 433, 32 },
		{ 223, 434, 32 },
		{ 223, 435, 32 },
		{ 223, 436, 32 },
		{ 223, 437, 32 },
		{ 223, 438, 32 },
		{ 223, 439, 32 },
		{ 223, 440, 32 },
		{ 223, 441, 32 },
		{ 223, 442, 32 },
		{ 223, 443, 32 },
		{ 223, 444, 32 },
		{ 223, 445, 32 },
		{ 223, 446, 32 },
		{ 223, 447, 32 },
		{ 223, 448, 32 },
		{ 223, 449, 32 },
		{ 223, 450, 32 },
		{ 223, 451, 32 },
		{ 223, 452, 32 },
		{ 223, 453, 32 },
		{ 223, 454, 32 },
		{ 223, 455, 32 },
		{ 223, 456, 32 },
		{ 223, 457, 32 },
		{ 223, 458, 32 },
		{ 223, 459, 32 },
		{ 223, 460, 32 },
		{ 223, 461, 32 },
		{ 223, 462, 32 },
		{ 223, 463, 32 },
		{ 223, 464, 32 },
		{ 223, 465, 32 },
		{ 223, 466, 32 },
		{ 223, 467, 32 },
		{ 223, 468, 32 },
		{ 223, 469, 32 },
		{ 223, 470, 32 },
		{ 223, 471, 32 },
		{ 223, 472, 32 },
		{ 223, 473, 32 },
		{ 223, 474, 32 },
		{ 223, 475, 32 },
		{ 223, 476, 32 },
		{ 223, 477, 32 },
		{ 223, 478, 32 },
		{ 223, 479, 32 },
		{ 223, 480, 32 },
		{ 223, 481, 32 },
		{ 223, 482, 32 },
		{ 223, 483, 32 },
		{ 223, 484, 32 },
		{ 223, 485, 32 },
		{ 223, 486, 32 },
		{ 223, 487, 32 },
		{ 223, 488, 32 },
		{ 223, 489, 32 },
		{ 223, 490, 32 },
		{ 223, 491, 32 },
		{ 223, 492, 32 },
		{ 223, 493, 32 },
		{ 223, 494, 32 },
		{ 223, 495, 32 },
		{ 223, 496, 32 },
		{ 223, 497, 32 },
		{ 223, 498, 32 },
		{ 223, 499, 32 },
		{ 223, 500, 32 },
		{ 223, 501, 32 },
		{ 223, 502, 32 },
		{ 223, 503, 32 },
		{ 223, 504, 32 },
		{ 223, 505, 32 },
		{ 223, 506, 32 },
		{ 223, 507, 32 },
		{ 223, 508, 32 },
		{ 223, 509, 32 },
		{ 223, 510, 32 },
		{ 223, 511, 32 },
		{ 223, 512, 32 },
		{ 224, 346, 32 },
		{ 224, 347, 32 },
		{ 224, 348, 32 },
		{ 224, 349, 32 },
		{ 224, 350, 32 },
		{ 224, 351, 32 },
		{ 224, 352, 32 },
		{ 224, 353, 32 },
		{ 224, 354, 32 },
		{ 224, 355, 32 },
		{ 224, 356, 32 },
		{ 224, 357, 32 },
		{ 224, 358, 32 },
		{ 224, 359, 32 },
		{ 224, 360, 32 },
		{ 224, 361, 32 },
		{ 224, 362, 32 },
		{ 224, 363, 32 },
		{ 224, 364, 32 },
		{ 224, 365, 32 },
		{ 224, 366, 32 },
		{ 224, 367, 32 },
		{ 224, 368, 32 },
		{ 224, 369, 32 },
		{ 224, 370, 32 },
		{ 224, 371, 32 },
		{ 224, 372, 32 },
		{ 224, 373, 32 },
		{ 224, 374, 32 },
		{ 224, 375, 32 },
		{ 224, 376, 32 },
		{ 224, 377, 32 },
		{ 224, 378, 32 },
		{ 224, 379, 32 },
		{ 224, 380, 32 },
		{ 224, 381, 32 },
		{ 224, 382, 32 },
		{ 224, 383, 32 },
		{ 224, 384, 32 },
		{ 224, 385, 32 },
		{ 224, 386, 32 },
		{ 224, 387, 32 },
		{ 224, 388, 32 },
		{ 224, 389, 32 },
		{ 224, 390, 32 },
		{ 224, 391, 32 },
		{ 224, 392, 32 },
		{ 224, 393, 32 },
		{ 224, 394, 32 },
		{ 224, 395, 32 },
		{ 224, 396, 32 },
		{ 224, 397, 32 },
		{ 224, 398, 32 },
		{ 224, 399, 32 },
		{ 224, 400, 32 },
		{ 224, 401, 32 },
		{ 224, 402, 32 },
		{ 224, 403, 32 },
		{ 224, 404, 32 },
		{ 224, 405, 32 },
		{ 224, 406, 32 },
		{ 224, 407, 32 },
		{ 224, 408, 32 },
		{ 224, 409, 32 },
		{ 224, 410, 32 },
		{ 224, 411, 32 },
		{ 224, 412, 32 },
		{ 224, 413, 32 },
		{ 224, 414, 32 },
		{ 224, 415, 32 },
		{ 224, 416, 32 },
		{ 224, 417, 32 },
		{ 224, 418, 32 },
		{ 224, 419, 32 },
		{ 224, 420, 32 },
		{ 224, 421, 32 },
		{ 224, 422, 32 },
		{ 224, 423, 32 },
		{ 224, 424, 32 },
		{ 224, 425, 32 },
		{ 224, 426, 32 },
		{ 224, 427, 32 },
		{ 224, 428, 32 },
		{ 224, 429, 32 },
		{ 224, 430, 32 },
		{ 224, 431, 32 },
		{ 224, 432, 32 },
		{ 224, 433, 32 },
		{ 224, 434, 32 },
		{ 224, 435, 32 },
		{ 224, 436, 32 },
		{ 224, 437, 32 },
		{ 224, 438, 32 },
		{ 224, 439, 32 },
		{ 224, 440, 32 },
		{ 224, 441, 32 },
		{ 224, 442, 32 },
		{ 224, 443, 32 },
		{ 224, 444, 32 },
		{ 224, 445, 32 },
		{ 224, 446, 32 },
		{ 224, 447, 32 },
		{ 224, 448, 32 },
		{ 224, 449, 32 },
		{ 224, 450, 32 },
		{ 224, 451, 32 },
		{ 224, 452, 32 },
		{ 224, 453, 32 },
		{ 224, 454, 32 },
		{ 224, 455, 32 },
		{ 224, 456, 32 },
		{ 224, 457, 32 },
		{ 224, 458, 32 },
		{ 224, 459, 32 },
		{ 224, 460, 32 },
		{ 224, 461, 32 },
		{ 224, 462, 32 },
		{ 224, 463, 32 },
		{ 224, 464, 32 },
		{ 224, 465, 32 },
		{ 224, 466, 32 },
		{ 224, 467, 32 },
		{ 224, 468, 32 },
		{ 224, 469, 32 },
		{ 224, 470, 32 },
		{ 224, 471, 32 },
		{ 224, 472, 32 },
		{ 224, 473, 32 },
		{ 224, 474, 32 },
		{ 224, 475, 32 },
		{ 224, 476, 32 },
		{ 224, 477, 32 },
		{ 224, 478, 32 },
		{ 224, 479, 32 },
		{ 224, 480, 32 },
		{ 224, 481, 32 },
		{ 224, 482, 32 },
		{ 224, 483, 32 },
		{ 224, 484, 32 },
		{ 224, 485, 32 },
		{ 224, 486, 32 },
		{ 224, 487, 32 },
		{ 224, 488, 32 },
		{ 224, 489, 32 },
		{ 224, 490, 32 },
		{ 224, 491, 32 },
		{ 224, 492, 32 },
		{ 224, 493, 32 },
		{ 224, 494, 32 },
		{ 224, 495, 32 },
		{ 224, 496, 32 },
		{ 224, 497, 32 },
		{ 224, 498, 32 },
		{ 224, 499, 32 },
		{ 224, 500, 32 },
		{ 224, 501, 32 },
		{ 224, 502, 32 },
		{ 224, 503, 32 },
		{ 224, 504, 32 },
		{ 224, 505, 32 },
		{ 224, 506, 32 },
		{ 224, 507, 32 },
		{ 224, 508, 32 },
		{ 224, 509, 32 },
		{ 224, 510, 32 },
		{ 224, 511, 32 },
		{ 224, 512, 32 },
		{ 225, 343, 32 },
		{ 225, 344, 32 },
		{ 225, 345, 32 },
		{ 225, 346, 32 },
		{ 225, 347, 32 },
		{ 225, 348, 32 },
		{ 225, 349, 32 },
		{ 225, 350, 32 },
		{ 225, 351, 32 },
		{ 225, 352, 32 },
		{ 225, 353, 32 },
		{ 225, 354, 32 },
		{ 225, 355, 32 },
		{ 225, 356, 32 },
		{ 225, 357, 32 },
		{ 225, 358, 32 },
		{ 225, 359, 32 },
		{ 225, 360, 32 },
		{ 225, 361, 32 },
		{ 225, 362, 32 },
		{ 225, 363, 32 },
		{ 225, 364, 32 },
		{ 225, 365, 32 },
		{ 225, 366, 32 },
		{ 225, 367, 32 },
		{ 225, 368, 32 },
		{ 225, 369, 32 },
		{ 225, 370, 32 },
		{ 225, 371, 32 },
		{ 225, 372, 32 },
		{ 225, 373, 32 },
		{ 225, 374, 32 },
		{ 225, 375, 32 },
		{ 225, 376, 32 },
		{ 225, 377, 32 },
		{ 225, 378, 32 },
		{ 225, 379, 32 },
		{ 225, 380, 32 },
		{ 225, 381, 32 },
		{ 225, 382, 32 },
		{ 225, 383, 32 },
		{ 225, 384, 32 },
		{ 225, 385, 32 },
		{ 225, 386, 32 },
		{ 225, 387, 32 },
		{ 225, 388, 32 },
		{ 225, 389, 32 },
		{ 225, 390, 32 },
		{ 225, 391, 32 },
		{ 225, 392, 32 },
		{ 225, 393, 32 },
		{ 225, 394, 32 },
		{ 225, 395, 32 },
		{ 225, 396, 32 },
		{ 225, 397, 32 },
		{ 225, 398, 32 },
		{ 225, 399, 32 },
		{ 225, 400, 32 },
		{ 225, 401, 32 },
		{ 225, 402, 32 },
		{ 225, 403, 32 },
		{ 225, 404, 32 },
		{ 225, 405, 32 },
		{ 225, 406, 32 },
		{ 225, 407, 32 },
		{ 225, 408, 32 },
		{ 225, 409, 32 },
		{ 225, 410, 32 },
		{ 225, 411, 32 },
		{ 225, 412, 32 },
		{ 225, 413, 32 },
		{ 225, 414, 32 },
		{ 225, 415, 32 },
		{ 225, 416, 32 },
		{ 225, 417, 32 },
		{ 225, 418, 32 },
		{ 225, 419, 32 },
		{ 225, 420, 32 },
		{ 225, 421, 32 },
		{ 225, 422, 32 },
		{ 225, 423, 32 },
		{ 225, 424, 32 },
		{ 225, 425, 32 },
		{ 225, 426, 32 },
		{ 225, 427, 32 },
		{ 225, 428, 32 },
		{ 225, 429, 32 },
		{ 225, 430, 32 },
		{ 225, 431, 32 },
		{ 225, 432, 32 },
		{ 225, 433, 32 },
		{ 225, 434, 32 },
		{ 225, 435, 32 },
		{ 225, 436, 32 },
		{ 225, 437, 32 },
		{ 225, 438, 32 },
		{ 225, 439, 32 },
		{ 225, 440, 32 },
		{ 225, 441, 32 },
		{ 225, 442, 32 },
		{ 225, 443, 32 },
		{ 225, 444, 32 },
		{ 225, 445, 32 },
		{ 225, 446, 32 },
		{ 225, 447, 32 },
		{ 225, 448, 32 },
		{ 225, 449, 32 },
		{ 225, 450, 32 },
		{ 225, 451, 32 },
		{ 225, 452, 32 },
		{ 225, 453, 32 },
		{ 225, 454, 32 },
		{ 225, 455, 32 },
		{ 225, 456, 32 },
		{ 225, 457, 32 },
		{ 225, 458, 32 },
		{ 225, 459, 32 },
		{ 225, 460, 32 },
		{ 225, 461, 32 },
		{ 225, 462, 32 },
		{ 225, 463, 32 },
		{ 225, 464, 32 },
		{ 225, 465, 32 },
		{ 225, 466, 32 },
		{ 225, 467, 32 },
		{ 225, 468, 32 },
		{ 225, 469, 32 },
		{ 225, 470, 32 },
		{ 225, 471, 32 },
		{ 225, 472, 32 },
		{ 225, 473, 32 },
		{ 225, 474, 32 },
		{ 225, 475, 32 },
		{ 225, 476, 32 },
		{ 225, 477, 32 },
		{ 225, 478, 32 },
		{ 225, 479, 32 },
		{ 225, 480, 32 },
		{ 225, 481, 32 },
		{ 225, 482, 32 },
		{ 225, 483, 32 },
		{ 225, 484, 32 },
		{ 225, 485, 32 },
		{ 225, 486, 32 },
		{ 225, 487, 32 },
		{ 225, 488, 32 },
		{ 225, 489, 32 },
		{ 225, 490, 32 },
		{ 225, 491, 32 },
		{ 225, 492, 32 },
		{ 225, 493, 32 },
		{ 225, 494, 32 },
		{ 225, 495, 32 },
		{ 225, 496, 32 },
		{ 225, 497, 32 },
		{ 225, 498, 32 },
		{ 225, 499, 32 },
		{ 225, 500, 32 },
		{ 225, 501, 32 },
		{ 225, 502, 32 },
		{ 225, 503, 32 },
		{ 225, 504, 32 },
		{ 225, 505, 32 },
		{ 225, 506, 32 },
		{ 225, 507, 32 },
		{ 225, 508, 32 },
		{ 225, 509, 32 },
		{ 225, 510, 32 },
		{ 225, 511, 32 },
		{ 225, 512, 32 },
		{ 226, 340, 32 },
		{ 226, 341, 32 },
		{ 226, 342, 32 },
		{ 226, 343, 32 },
		{ 226, 344, 32 },
		{ 226, 345, 32 },
		{ 226, 346, 32 },
		{ 226, 347, 32 },
		{ 226, 348, 32 },
		{ 226, 349, 32 },
		{ 226, 350, 32 },
		{ 226, 351, 32 },
		{ 226, 352, 32 },
		{ 226, 353, 32 },
		{ 226, 354, 32 },
		{ 226, 355, 32 },
		{ 226, 356, 32 },
		{ 226, 357, 32 },
		{ 226, 358, 32 },
		{ 226, 359, 32 },
		{ 226, 360, 32 },
		{ 226, 361, 32 },
		{ 226, 362, 32 },
		{ 226, 363, 32 },
		{ 226, 364, 32 },
		{ 226, 365, 32 },
		{ 226, 366, 32 },
		{ 226, 367, 32 },
		{ 226, 368, 32 },
		{ 226, 369, 32 },
		{ 226, 370, 32 },
		{ 226, 371, 32 },
		{ 226, 372, 32 },
		{ 226, 373, 32 },
		{ 226, 374, 32 },
		{ 226, 375, 32 },
		{ 226, 376, 32 },
		{ 226, 377, 32 },
		{ 226, 378, 32 },
		{ 226, 379, 32 },
		{ 226, 380, 32 },
		{ 226, 381, 32 },
		{ 226, 382, 32 },
		{ 226, 383, 32 },
		{ 226, 384, 32 },
		{ 226, 385, 32 },
		{ 226, 386, 32 },
		{ 226, 387, 32 },
		{ 226, 388, 32 },
		{ 226, 389, 32 },
		{ 226, 390, 32 },
		{ 226, 391, 32 },
		{ 226, 392, 32 },
		{ 226, 393, 32 },
		{ 226, 394, 32 },
		{ 226, 395, 32 },
		{ 226, 396, 32 },
		{ 226, 397, 32 },
		{ 226, 398, 32 },
		{ 226, 399, 32 },
		{ 226, 400, 32 },
		{ 226, 401, 32 },
		{ 226, 402, 32 },
		{ 226, 403, 32 },
		{ 226, 404, 32 },
		{ 226, 405, 32 },
		{ 226, 406, 32 },
		{ 226, 407, 32 },
		{ 226, 408, 32 },
		{ 226, 409, 32 },
		{ 226, 410, 32 },
		{ 226, 411, 32 },
		{ 226, 412, 32 },
		{ 226, 413, 32 },
		{ 226, 414, 32 },
		{ 226, 415, 32 },
		{ 226, 416, 32 },
		{ 226, 417, 32 },
		{ 226, 418, 32 },
		{ 226, 419, 32 },
		{ 226, 420, 32 },
		{ 226, 421, 32 },
		{ 226, 422, 32 },
		{ 226, 423, 32 },
		{ 226, 424, 32 },
		{ 226, 425, 32 },
		{ 226, 426, 32 },
		{ 226, 427, 32 },
		{ 226, 428, 32 },
		{ 226, 429, 32 },
		{ 226, 430, 32 },
		{ 226, 431, 32 },
		{ 226, 432, 32 },
		{ 226, 433, 32 },
		{ 226, 434, 32 },
		{ 226, 435, 32 },
		{ 226, 436, 32 },
		{ 226, 437, 32 },
		{ 226, 438, 32 },
		{ 226, 439, 32 },
		{ 226, 440, 32 },
		{ 226, 441, 32 },
		{ 226, 442, 32 },
		{ 226, 443, 32 },
		{ 226, 444, 32 },
		{ 226, 445, 32 },
		{ 226, 446, 32 },
		{ 226, 447, 32 },
		{ 226, 448, 32 },
		{ 226, 449, 32 },
		{ 226, 450, 32 },
		{ 226, 451, 32 },
		{ 226, 452, 32 },
		{ 226, 453, 32 },
		{ 226, 454, 32 },
		{ 226, 455, 32 },
		{ 226, 456, 32 },
		{ 226, 457, 32 },
		{ 226, 458, 32 },
		{ 226, 459, 32 },
		{ 226, 460, 32 },
		{ 226, 461, 32 },
		{ 226, 462, 32 },
		{ 226, 463, 32 },
		{ 226, 464, 32 },
		{ 226, 465, 32 },
		{ 226, 466, 32 },
		{ 226, 467, 32 },
		{ 226, 468, 32 },
		{ 226, 469, 32 },
		{ 226, 470, 32 },
		{ 226, 471, 32 },
		{ 226, 472, 32 },
		{ 226, 473, 32 },
		{ 226, 474, 32 },
		{ 226, 475, 32 },
		{ 226, 476, 32 },
		{ 226, 477, 32 },
		{ 226, 478, 32 },
		{ 226, 479, 32 },
		{ 226, 480, 32 },
		{ 226, 481, 32 },
		{ 226, 482, 32 },
		{ 226, 483, 32 },
		{ 226, 484, 32 },
		{ 226, 485, 32 },
		{ 226, 486, 32 },
		{ 226, 487, 32 },
		{ 226, 488, 32 },
		{ 226, 489, 32 },
		{ 226, 490, 32 },
		{ 226, 491, 32 },
		{ 226, 492, 32 },
		{ 226, 493, 32 },
		{ 226, 494, 32 },
		{ 226, 495, 32 },
		{ 226, 496, 32 },
		{ 226, 497, 32 },
		{ 226, 498, 32 },
		{ 226, 499, 32 },
		{ 226, 500, 32 },
		{ 226, 501, 32 },
		{ 226, 502, 32 },
		{ 226, 503, 32 },
		{ 226, 504, 32 },
		{ 227, 338, 32 },
		{ 227, 339, 32 },
		{ 227, 340, 32 },
		{ 227, 341, 32 },
		{ 227, 342, 32 },
		{ 227, 343, 32 },
		{ 227, 344, 32 },
		{ 227, 345, 32 },
		{ 227, 346, 32 },
		{ 227, 347, 32 },
		{ 227, 348, 32 },
		{ 227, 349, 32 },
		{ 227, 350, 32 },
		{ 227, 351, 32 },
		{ 227, 352, 32 },
		{ 227, 353, 32 },
		{ 227, 354, 32 },
		{ 227, 355, 32 },
		{ 227, 356, 32 },
		{ 227, 357, 32 },
		{ 227, 358, 32 },
		{ 227, 359, 32 },
		{ 227, 360, 32 },
		{ 227, 361, 32 },
		{ 227, 362, 32 },
		{ 227, 363, 32 },
		{ 227, 364, 32 },
		{ 227, 365, 32 },
		{ 227, 366, 32 },
		{ 227, 367, 32 },
		{ 227, 368, 32 },
		{ 227, 369, 32 },
		{ 227, 370, 32 },
		{ 227, 371, 32 },
		{ 227, 372, 32 },
		{ 227, 373, 32 },
		{ 227, 374, 32 },
		{ 227, 375, 32 },
		{ 227, 376, 32 },
		{ 227, 377, 32 },
		{ 227, 378, 32 },
		{ 227, 379, 32 },
		{ 227, 380, 32 },
		{ 227, 381, 32 },
		{ 227, 382, 32 },
		{ 227, 383, 32 },
		{ 227, 384, 32 },
		{ 227, 385, 32 },
		{ 227, 386, 32 },
		{ 227, 387, 32 },
		{ 227, 388, 32 },
		{ 227, 389, 32 },
		{ 227, 390, 32 },
		{ 227, 391, 32 },
		{ 227, 392, 32 },
		{ 227, 393, 32 },
		{ 227, 394, 32 },
		{ 227, 395, 32 },
		{ 227, 396, 32 },
		{ 227, 397, 32 },
		{ 227, 398, 32 },
		{ 227, 399, 32 },
		{ 227, 400, 32 },
		{ 227, 401, 32 },
		{ 227, 402, 32 },
		{ 227, 403, 32 },
		{ 227, 404, 32 },
		{ 227, 405, 32 },
		{ 227, 406, 32 },
		{ 227, 407, 32 },
		{ 227, 408, 32 },
		{ 227, 409, 32 },
		{ 227, 410, 32 },
		{ 227, 411, 32 },
		{ 227, 412, 32 },
		{ 227, 413, 32 },
		{ 227, 414, 32 },
		{ 227, 415, 32 },
		{ 227, 416, 32 },
		{ 227, 417, 32 },
		{ 227, 418, 32 },
		{ 227, 419, 32 },
		{ 227, 420, 32 },
		{ 227, 421, 32 },
		{ 227, 422, 32 },
		{ 227, 423, 32 },
		{ 227, 424, 32 },
		{ 227, 425, 32 },
		{ 227, 426, 32 },
		{ 227, 427, 32 },
		{ 227, 428, 32 },
		{ 227, 429, 32 },
		{ 227, 430, 32 },
		{ 227, 431, 32 },
		{ 227, 432, 32 },
		{ 227, 433, 32 },
		{ 227, 434, 32 },
		{ 227, 435, 32 },
		{ 227, 436, 32 },
		{ 227, 437, 32 },
		{ 227, 438, 32 },
		{ 227, 439, 32 },
		{ 227, 440, 32 },
		{ 227, 441, 32 },
		{ 227, 442, 32 },
		{ 227, 443, 32 },
		{ 227, 444, 32 },
		{ 227, 445, 32 },
		{ 227, 446, 32 },
		{ 227, 447, 32 },
		{ 227, 448, 32 },
		{ 227, 449, 32 },
		{ 227, 450, 32 },
		{ 227, 451, 32 },
		{ 227, 452, 32 },
		{ 227, 453, 32 },
		{ 227, 454, 32 },
		{ 227, 455, 32 },
		{ 227, 456, 32 },
		{ 227, 457, 32 },
		{ 227, 458, 32 },
		{ 227, 459, 32 },
		{ 227, 460, 32 },
		{ 227, 461, 32 },
		{ 227, 462, 32 },
		{ 227, 463, 32 },
		{ 227, 464, 32 },
		{ 227, 465, 32 },
		{ 227, 466, 32 },
		{ 227, 467, 32 },
		{ 227, 468, 32 },
		{ 227, 469, 32 },
		{ 227, 470, 32 },
		{ 227, 471, 32 },
		{ 227, 472, 32 },
		{ 227, 473, 32 },
		{ 227, 474, 32 },
		{ 227, 475, 32 },
		{ 227, 476, 32 },
		{ 227, 477, 32 },
		{ 227, 478, 32 },
		{ 227, 479, 32 },
		{ 227, 480, 32 },
		{ 227, 481, 32 },
		{ 227, 482, 32 },
		{ 227, 483, 32 },
		{ 227, 484, 32 },
		{ 227, 485, 32 },
		{ 227, 486, 32 },
		{ 227, 487, 32 },
		{ 227, 488, 32 },
		{ 227, 489, 32 },
		{ 227, 490, 32 },
		{ 227, 491, 32 },
		{ 227, 492, 32 },
		{ 227, 493, 32 },
		{ 227, 494, 32 },
		{ 227, 495, 32 },
		{ 227, 496, 32 },
		{ 228, 335, 32 },
		{ 228, 336, 32 },
		{ 228, 337, 32 },
		{ 228, 338, 32 },
		{ 228, 339, 32 },
		{ 228, 340, 32 },
		{ 228, 341, 32 },
		{ 228, 342, 32 },
		{ 228, 343, 32 },
		{ 228, 344, 32 },
		{ 228, 345, 32 },
		{ 228, 346, 32 },
		{ 228, 347, 32 },
		{ 228, 348, 32 },
		{ 228, 349, 32 },
		{ 228, 350, 32 },
		{ 228, 351, 32 },
		{ 228, 352, 32 },
		{ 228, 353, 32 },
		{ 228, 354, 32 },
		{ 228, 355, 32 },
		{ 228, 356, 32 },
		{ 228, 357, 32 },
		{ 228, 358, 32 },
		{ 228, 359, 32 },
		{ 228, 360, 32 },
		{ 228, 361, 32 },
		{ 228, 362, 32 },
		{ 228, 363, 32 },
		{ 228, 364, 32 },
		{ 228, 365, 32 },
		{ 228, 366, 32 },
		{ 228, 367, 32 },
		{ 228, 368, 32 },
		{ 228, 369, 32 },
		{ 228, 370, 32 },
		{ 228, 371, 32 },
		{ 228, 372, 32 },
		{ 228, 373, 32 },
		{ 228, 374, 32 },
		{ 228, 375, 32 },
		{ 228, 376, 32 },
		{ 228, 377, 32 },
		{ 228, 378, 32 },
		{ 228, 379, 32 },
		{ 228, 380, 32 },
		{ 228, 381, 32 },
		{ 228, 382, 32 },
		{ 228, 383, 32 },
		{ 228, 384, 32 },
		{ 228, 385, 32 },
		{ 228, 386, 32 },
		{ 228, 387, 32 },
		{ 228, 388, 32 },
		{ 228, 389, 32 },
		{ 228, 390, 32 },
		{ 228, 391, 32 },
		{ 228, 392, 32 },
		{ 228, 393, 32 },
		{ 228, 394, 32 },
		{ 228, 395, 32 },
		{ 228, 396, 32 },
		{ 228, 397, 32 },
		{ 228, 398, 32 },
		{ 228, 399, 32 },
		{ 228, 400, 32 },
		{ 228, 401, 32 },
		{ 228, 402, 32 },
		{ 228, 403, 32 },
		{ 228, 404, 32 },
		{ 228, 405, 32 },
		{ 228, 406, 32 },
		{ 228, 407, 32 },
		{ 228, 408, 32 },
		{ 228, 409, 32 },
		{ 228, 410, 32 },
		{ 228, 411, 32 },
		{ 228, 412, 32 },
		{ 228, 413, 32 },
		{ 228, 414, 32 },
		{ 228, 415, 32 },
		{ 228, 416, 32 },
		{ 228, 417, 32 },
		{ 228, 418, 32 },
		{ 228, 419, 32 },
		{ 228, 420, 32 },
		{ 228, 421, 32 },
		{ 228, 422, 32 },
		{ 228, 423, 32 },
		{ 228, 424, 32 },
		{ 228, 425, 32 },
		{ 228, 426, 32 },
		{ 228, 427, 32 },
		{ 228, 428, 32 },
		{ 228, 429, 32 },
		{ 228, 430, 32 },
		{ 228, 431, 32 },
		{ 228, 432, 32 },
		{ 228, 433, 32 },
		{ 228, 434, 32 },
		{ 228, 435, 32 },
		{ 228, 436, 32 },
		{ 228, 437, 32 },
		{ 228, 438, 32 },
		{ 228, 439, 32 },
		{ 228, 440, 32 },
		{ 228, 441, 32 },
		{ 228, 442, 32 },
		{ 228, 443, 32 },
		{ 228, 444, 32 },
		{ 228, 445, 32 },
		{ 228, 446, 32 },
		{ 228, 447, 32 },
		{ 228, 448, 32 },
		{ 228, 449, 32 },
		{ 228, 450, 32 },
		{ 228, 451, 32 },
		{ 228, 452, 32 },
		{ 228, 453, 32 },
		{ 228, 454, 32 },
		{ 228, 455, 32 },
		{ 228, 456, 32 },
		{ 228, 457, 32 },
		{ 228, 458, 32 },
		{ 228, 459, 32 },
		{ 228, 460, 32 },
		{ 228, 461, 32 },
		{ 228, 462, 32 },
		{ 228, 463, 32 },
		{ 228, 464, 32 },
		{ 228, 465, 32 },
		{ 228, 466, 32 },
		{ 228, 467, 32 },
		{ 228, 468, 32 },
		{ 228, 469, 32 },
		{ 228, 470, 32 },
		{ 228, 471, 32 },
		{ 228, 472, 32 },
		{ 228, 473, 32 },
		{ 228, 474, 32 },
		{ 228, 475, 32 },
		{ 228, 476, 32 },
		{ 228, 477, 32 },
		{ 228, 478, 32 },
		{ 228, 479, 32 },
		{ 228, 480, 32 },
		{ 228, 481, 32 },
		{ 228, 482, 32 },
		{ 228, 483, 32 },
		{ 228, 484, 32 },
		{ 228, 485, 32 },
		{ 228, 486, 32 },
		{ 228, 487, 32 },
		{ 228, 488, 32 },
		{ 229, 332, 32 },
		{ 229, 333, 32 },
		{ 229, 334, 32 },
		{ 229, 335, 32 },
		{ 229, 336, 32 },
		{ 229, 337, 32 },
		{ 229, 338, 32 },
		{ 229, 339, 32 },
		{ 229, 340, 32 },
		{ 229, 341, 32 },
		{ 229, 342, 32 },
		{ 229, 343, 32 },
		{ 229, 344, 32 },
		{ 229, 345, 32 },
		{ 229, 346, 32 },
		{ 229, 347, 32 },
		{ 229, 348, 32 },
		{ 229, 349, 32 },
		{ 229, 350, 32 },
		{ 229, 351, 32 },
		{ 229, 352, 32 },
		{ 229, 353, 32 },
		{ 229, 354, 32 },
		{ 229, 355, 32 },
		{ 229, 356, 32 },
		{ 229, 357, 32 },
		{ 229, 358, 32 },
		{ 229, 359, 32 },
		{ 229, 360, 32 },
		{ 229, 361, 32 },
		{ 229, 362, 32 },
		{ 229, 363, 32 },
		{ 229, 364, 32 },
		{ 229, 365, 32 },
		{ 229, 366, 32 },
		{ 229, 367, 32 },
		{ 229, 368, 32 },
		{ 229, 369, 32 },
		{ 229, 370, 32 },
		{ 229, 371, 32 },
		{ 229, 372, 32 },
		{ 229, 373, 32 },
		{ 229, 374, 32 },
		{ 229, 375, 32 },
		{ 229, 376, 32 },
		{ 229, 377, 32 },
		{ 229, 378, 32 },
		{ 229, 379, 32 },
		{ 229, 380, 32 },
		{ 229, 381, 32 },
		{ 229, 382, 32 },
		{ 229, 383, 32 },
		{ 229, 384, 32 },
		{ 229, 385, 32 },
		{ 229, 386, 32 },
		{ 229, 387, 32 },
		{ 229, 388, 32 },
		{ 229, 389, 32 },
		{ 229, 390, 32 },
		{ 229, 391, 32 },
		{ 229, 392, 32 },
		{ 229, 393, 32 },
		{ 229, 394, 32 },
		{ 229, 395, 32 },
		{ 229, 396, 32 },
		{ 229, 397, 32 },
		{ 229, 398, 32 },
		{ 229, 399, 32 },
		{ 229, 400, 32 },
		{ 229, 401, 32 },
		{ 229, 402, 32 },
		{ 229, 403, 32 },
		{ 229, 404, 32 },
		{ 229, 405, 32 },
		{ 229, 406, 32 },
		{ 229, 407, 32 },
		{ 229, 408, 32 },
		{ 229, 409, 32 },
		{ 229, 410, 32 },
		{ 229, 411, 32 },
		{ 229, 412, 32 },
		{ 229, 413, 32 },
		{ 229, 414, 32 },
		{ 229, 415, 32 },
		{ 229, 416, 32 },
		{ 229, 417, 32 },
		{ 229, 418, 32 },
		{ 229, 419, 32 },
		{ 229, 420, 32 },
		{ 229, 421, 32 },
		{ 229, 422, 32 },
		{ 229, 423, 32 },
		{ 229, 424, 32 },
		{ 229, 425, 32 },
		{ 229, 426, 32 },
		{ 229, 427, 32 },
		{ 229, 428, 32 },
		{ 229, 429, 32 },
		{ 229, 430, 32 },
		{ 229, 431, 32 },
		{ 229, 432, 32 },
		{ 229, 433, 32 },
		{ 229, 434, 32 },
		{ 229, 435, 32 },
		{ 229, 436, 32 },
		{ 229, 437, 32 },
		{ 229, 438, 32 },
		{ 229, 439, 32 },
		{ 229, 440, 32 },
		{ 229, 441, 32 },
		{ 229, 442, 32 },
		{ 229, 443, 32 },
		{ 229, 444, 32 },
		{ 229, 445, 32 },
		{ 229, 446, 32 },
		{ 229, 447, 32 },
		{ 229, 448, 32 },
		{ 229, 449, 32 },
		{ 229, 450, 32 },
		{ 229, 451, 32 },
		{ 229, 452, 32 },
		{ 229, 453, 32 },
		{ 229, 454, 32 },
		{ 229, 455, 32 },
		{ 229, 456, 32 },
		{ 229, 457, 32 },
		{ 229, 458, 32 },
		{ 229, 459, 32 },
		{ 229, 460, 32 },
		{ 229, 461, 32 },
		{ 229, 462, 32 },
		{ 229, 463, 32 },
		{ 229, 464, 32 },
		{ 229, 465, 32 },
		{ 229, 466, 32 },
		{ 229, 467, 32 },
		{ 229, 468, 32 },
		{ 229, 469, 32 },
		{ 229, 470, 32 },
		{ 229, 471, 32 },
		{ 229, 472, 32 },
		{ 229, 473, 32 },
		{ 229, 474, 32 },
		{ 229, 475, 32 },
		{ 229, 476, 32 },
		{ 229, 477, 32 },
		{ 229, 478, 32 },
		{ 229, 479, 32 },
		{ 229, 480, 32 },
		{ 230, 330, 32 },
		{ 230, 331, 32 },
		{ 230, 332, 32 },
		{ 230, 333, 32 },
		{ 230, 334, 32 },
		{ 230, 335, 32 },
		{ 230, 336, 32 },
		{ 230, 337, 32 },
		{ 230, 338, 32 },
		{ 230, 339, 32 },
		{ 230, 340, 32 },
		{ 230, 341, 32 },
		{ 230, 342, 32 },
		{ 230, 343, 32 },
		{ 230, 344, 32 },
		{ 230, 345, 32 },
		{ 230, 346, 32 },
		{ 230, 347, 32 },
		{ 230, 348, 32 },
		{ 230, 349, 32 },
		{ 230, 350, 32 },
		{ 230, 351, 32 },
		{ 230, 352, 32 },
		{ 230, 353, 32 },
		{ 230, 354, 32 },
		{ 230, 355, 32 },
		{ 230, 356, 32 },
		{ 230, 357, 32 },
		{ 230, 358, 32 },
		{ 230, 359, 32 },
		{ 230, 360, 32 },
		{ 230, 361, 32 },
		{ 230, 362, 32 },
		{ 230, 363, 32 },
		{ 230, 364, 32 },
		{ 230, 365, 32 },
		{ 230, 366, 32 },
		{ 230, 367, 32 },
		{ 230, 368, 32 },
		{ 230, 369, 32 },
		{ 230, 370, 32 },
		{ 230, 371, 32 },
		{ 230, 372, 32 },
		{ 230, 373, 32 },
		{ 230, 374, 32 },
		{ 230, 375, 32 },
		{ 230, 376, 32 },
		{ 230, 377, 32 },
		{ 230, 378, 32 },
		{ 230, 379, 32 },
		{ 230, 380, 32 },
		{ 230, 381, 32 },
		{ 230, 382, 32 },
		{ 230, 383, 32 },
		{ 230, 384, 32 },
		{ 230, 385, 32 },
		{ 230, 386, 32 },
		{ 230, 387, 32 },
		{ 230, 388, 32 },
		{ 230, 389, 32 },
		{ 230, 390, 32 },
		{ 230, 391, 32 },
		{ 230, 392, 32 },
		{ 230, 393, 32 },
		{ 230, 394, 32 },
		{ 230, 395, 32 },
		{ 230, 396, 32 },
		{ 230, 397, 32 },
		{ 230, 398, 32 },
		{ 230, 399, 32 },
		{ 230, 400, 32 },
		{ 230, 401, 32 },
		{ 230, 402, 32 },
		{ 230, 403, 32 },
		{ 230, 404, 32 },
		{ 230, 405, 32 },
		{ 230, 406, 32 },
		{ 230, 407, 32 },
		{ 230, 408, 32 },
		{ 230, 409, 32 },
		{ 230, 410, 32 },
		{ 230, 411, 32 },
		{ 230, 412, 32 },
		{ 230, 413, 32 },
		{ 230, 414, 32 },
		{ 230, 415, 32 },
		{ 230, 416, 32 },
		{ 230, 417, 32 },
		{ 230, 418, 32 },
		{ 230, 419, 32 },
		{ 230, 420, 32 },
		{ 230, 421, 32 },
		{ 230, 422, 32 },
		{ 230, 423, 32 },
		{ 230, 424, 32 },
		{ 230, 425, 32 },
		{ 230, 426, 32 },
		{ 230, 427, 32 },
		{ 230, 428, 32 },
		{ 230, 429, 32 },
		{ 230, 430, 32 },
		{ 230, 431, 32 },
		{ 230, 432, 32 },
		{ 230, 433, 32 },
		{ 230, 434, 32 },
		{ 230, 435, 32 },
		{ 230, 436, 32 },
		{ 230, 437, 32 },
		{ 230, 438, 32 },
		{ 230, 439, 32 },
		{ 230, 440, 32 },
		{ 230, 441, 32 },
		{ 230, 442, 32 },
		{ 230, 443, 32 },
		{ 230, 444, 32 },
		{ 230, 445, 32 },
		{ 230, 446, 32 },
		{ 230, 447, 32 },
		{ 230, 448, 32 },
		{ 230, 449, 32 },
		{ 230, 450, 32 },
		{ 230, 451, 32 },
		{ 230, 452, 32 },
		{ 230, 453, 32 },
		{ 230, 454, 32 },
		{ 230, 455, 32 },
		{ 230, 456, 32 },
		{ 230, 457, 32 },
		{ 230, 458, 32 },
		{ 230, 459, 32 },
		{ 230, 460, 32 },
		{ 230, 461, 32 },
		{ 230, 462, 32 },
		{ 230, 463, 32 },
		{ 230, 464, 32 },
		{ 230, 465, 32 },
		{ 230, 466, 32 },
		{ 230, 467, 32 },
		{ 230, 468, 32 },
		{ 230, 469, 32 },
		{ 230, 470, 32 },
		{ 230, 471, 32 },
		{ 230, 472, 32 },
		{ 231, 327, 32 },
		{ 231, 328, 32 },
		{ 231, 329, 32 },
		{ 231, 330, 32 },
		{ 231, 331, 32 },
		{ 231, 332, 32 },
		{ 231, 333, 32 },
		{ 231, 334, 32 },
		{ 231, 335, 32 },
		{ 231, 336, 32 },
		{ 231, 337, 32 },
		{ 231, 338, 32 },
		{ 231, 339, 32 },
		{ 231, 340, 32 },
		{ 231, 341, 32 },
		{ 231, 342, 32 },
		{ 231, 343, 32 },
		{ 231, 344, 32 },
		{ 231, 345, 32 },
		{ 231, 346, 32 },
		{ 231, 347, 32 },
		{ 231, 348, 32 },
		{ 231, 349, 32 },
		{ 231, 350, 32 },
		{ 231, 351, 32 },
		{ 231, 352, 32 },
		{ 231, 353, 32 },
		{ 231, 354, 32 },
		{ 231, 355, 32 },
		{ 231, 356, 32 },
		{ 231, 357, 32 },
		{ 231, 358, 32 },
		{ 231, 359, 32 },
		{ 231, 360, 32 },
		{ 231, 361, 32 },
		{ 231, 362, 32 },
		{ 231, 363, 32 },
		{ 231, 364, 32 },
		{ 231, 365, 32 },
		{ 231, 366, 32 },
		{ 231, 367, 32 },
		{ 231, 368, 32 },
		{ 231, 369, 32 },
		{ 231, 370, 32 },
		{ 231, 371, 32 },
		{ 231, 372, 32 },
		{ 231, 373, 32 },
		{ 231, 374, 32 },
		{ 231, 375, 32 },
		{ 231, 376, 32 },
		{ 231, 377, 32 },
		{ 231, 378, 32 },
		{ 231, 379, 32 },
		{ 231, 380, 32 },
		{ 231, 381, 32 },
		{ 231, 382, 32 },
		{ 231, 383, 32 },
		{ 231, 384, 32 },
		{ 231, 385, 32 },
		{ 231, 386, 32 },
		{ 231, 387, 32 },
		{ 231, 388, 32 },
		{ 231, 389, 32 },
		{ 231, 390, 32 },
		{ 231, 391, 32 },
		{ 231, 392, 32 },
		{ 231, 393, 32 },
		{ 231, 394, 32 },
		{ 231, 395, 32 },
		{ 231, 396, 32 },
		{ 231, 397, 32 },
		{ 231, 398, 32 },
		{ 231, 399, 32 },
		{ 231, 400, 32 },
		{ 231, 401, 32 },
		{ 231, 402, 32 },
		{ 231, 403, 32 },
		{ 231, 404, 32 },
		{ 231, 405, 32 },
		{ 231, 406, 32 },
		{ 231, 407, 32 },
		{ 231, 408, 32 },
		{ 231, 409, 32 },
		{ 231, 410, 32 },
		{ 231, 411, 32 },
		{ 231, 412, 32 },
		{ 231, 413, 32 },
		{ 231, 414, 32 },
		{ 231, 415, 32 },
		{ 231, 416, 32 },
		{ 231, 417, 32 },
		{ 231, 418, 32 },
		{ 231, 419, 32 },
		{ 231, 420, 32 },
		{ 231, 421, 32 },
		{ 231, 422, 32 },
		{ 231, 423, 32 },
		{ 231, 424, 32 },
		{ 231, 425, 32 },
		{ 231, 426, 32 },
		{ 231, 427, 32 },
		{ 231, 428, 32 },
		{ 231, 429, 32 },
		{ 231, 430, 32 },
		{ 231, 431, 32 },
		{ 231, 432, 32 },
		{ 231, 433, 32 },
		{ 231, 434, 32 },
		{ 231, 435, 32 },
		{ 231, 436, 32 },
		{ 231, 437, 32 },
		{ 231, 438, 32 },
		{ 231, 439, 32 },
		{ 231, 440, 32 },
		{ 231, 441, 32 },
		{ 231, 442, 32 },
		{ 231, 443, 32 },
		{ 231, 444, 32 },
		{ 231, 445, 32 },
		{ 231, 446, 32 },
		{ 231, 447, 32 },
		{ 231, 448, 32 },
		{ 231, 449, 32 },
		{ 231, 450, 32 },
		{ 231, 451, 32 },
		{ 231, 452, 32 },
		{ 231, 453, 32 },
		{ 231, 454, 32 },
		{ 231, 455, 32 },
		{ 231, 456, 32 },
		{ 231, 457, 32 },
		{ 231, 458, 32 },
		{ 231, 459, 32 },
		{ 231, 460, 32 },
		{ 231, 461, 32 },
		{ 231, 462, 32 },
		{ 231, 463, 32 },
		{ 231, 464, 32 },
		{ 232, 325, 32 },
		{ 232, 326, 32 },
		{ 232, 327, 32 },
		{ 232, 328, 32 },
		{ 232, 329, 32 },
		{ 232, 330, 32 },
		{ 232, 331, 32 },
		{ 232, 332, 32 },
		{ 232, 333, 32 },
		{ 232, 334, 32 },
		{ 232, 335, 32 },
		{ 232, 336, 32 },
		{ 232, 337, 32 },
		{ 232, 338, 32 },
		{ 232, 339, 32 },
		{ 232, 340, 32 },
		{ 232, 341, 32 },
		{ 232, 342, 32 },
		{ 232, 343, 32 },
		{ 232, 344, 32 },
		{ 232, 345, 32 },
		{ 232, 346, 32 },
		{ 232, 347, 32 },
		{ 232, 348, 32 },
		{ 232, 349, 32 },
		{ 232, 350, 32 },
		{ 232, 351, 32 },
		{ 232, 352, 32 },
		{ 232, 353, 32 },
		{ 232, 354, 32 },
		{ 232, 355, 32 },
		{ 232, 356, 32 },
		{ 232, 357, 32 },
		{ 232, 358, 32 },
		{ 232, 359, 32 },
		{ 232, 360, 32 },
		{ 232, 361, 32 },
		{ 232, 362, 32 },
		{ 232, 363, 32 },
		{ 232, 364, 32 },
		{ 232, 365, 32 },
		{ 232, 366, 32 },
		{ 232, 367, 32 },
		{ 232, 368, 32 },
		{ 232, 369, 32 },
		{ 232, 370, 32 },
		{ 232, 371, 32 },
		{ 232, 372, 32 },
		{ 232, 373, 32 },
		{ 232, 374, 32 },
		{ 232, 375, 32 },
		{ 232, 376, 32 },
		{ 232, 377, 32 },
		{ 232, 378, 32 },
		{ 232, 379, 32 },
		{ 232, 380, 32 },
		{ 232, 381, 32 },
		{ 232, 382, 32 },
		{ 232, 383, 32 },
		{ 232, 384, 32 },
		{ 232, 385, 32 },
		{ 232, 386, 32 },
		{ 232, 387, 32 },
		{ 232, 388, 32 },
		{ 232, 389, 32 },
		{ 232, 390, 32 },
		{ 232, 391, 32 },
		{ 232, 392, 32 },
		{ 232, 393, 32 },
		{ 232, 394, 32 },
		{ 232, 395, 32 },
		{ 232, 396, 32 },
		{ 232, 397, 32 },
		{ 232, 398, 32 },
		{ 232, 399, 32 },
		{ 232, 400, 32 },
		{ 232, 401, 32 },
		{ 232, 402, 32 },
		{ 232, 403, 32 },
		{ 232, 404, 32 },
		{ 232, 405, 32 },
		{ 232, 406, 32 },
		{ 232, 407, 32 },
		{ 232, 408, 32 },
		{ 232, 409, 32 },
		{ 232, 410, 32 },
		{ 232, 411, 32 },
		{ 232, 412, 32 },
		{ 232, 413, 32 },
		{ 232, 414, 32 },
		{ 232, 415, 32 },
		{ 232, 416, 32 },
		{ 232, 417, 32 },
		{ 232, 418, 32 },
		{ 232, 419, 32 },
		{ 232, 420, 32 },
		{ 232, 421, 32 },
		{ 232, 422, 32 },
		{ 232, 423, 32 },
		{ 232, 424, 32 },
		{ 232, 425, 32 },
		{ 232, 426, 32 },
		{ 232, 427, 32 },
		{ 232, 428, 32 },
		{ 232, 429, 32 },
		{ 232, 430, 32 },
		{ 232, 431, 32 },
		{ 232, 432, 32 },
		{ 232, 433, 32 },
		{ 232, 434, 32 },
		{ 232, 435, 32 },
		{ 232, 436, 32 },
		{ 232, 437, 32 },
		{ 232, 438, 32 },
		{ 232, 439, 32 },
		{ 232, 440, 32 },
		{ 232, 441, 32 },
		{ 232, 442, 32 },
		{ 232, 443, 32 },
		{ 232, 444, 32 },
		{ 232, 445, 32 },
		{ 232, 446, 32 },
		{ 232, 447, 32 },
		{ 232, 448, 32 },
		{ 232, 449, 32 },
		{ 232, 450, 32 },
		{ 232, 451, 32 },
		{ 232, 452, 32 },
		{ 232, 453, 32 },
		{ 232, 454, 32 },
		{ 232, 455, 32 },
		{ 232, 456, 32 },
		{ 233, 324, 32 },
		{ 233, 325, 32 },
		{ 233, 326, 32 },
		{ 233, 327, 32 },
		{ 233, 328, 32 },
		{ 233, 329, 32 },
		{ 233, 330, 32 },
		{ 233, 331, 32 },
		{ 233, 332, 32 },
		{ 233, 333, 32 },
		{ 233, 334, 32 },
		{ 233, 335, 32 },
		{ 233, 336, 32 },
		{ 233, 337, 32 },
		{ 233, 338, 32 },
		{ 233, 339, 32 },
		{ 233, 340, 32 },
		{ 233, 341, 32 },
		{ 233, 342, 32 },
		{ 233, 343, 32 },
		{ 233, 344, 32 },
		{ 233, 345, 32 },
		{ 233, 346, 32 },
		{ 233, 347, 32 },
		{ 233, 348, 32 },
		{ 233, 349, 32 },
		{ 233, 350, 32 },
		{ 233, 351, 32 },
		{ 233, 352, 32 },
		{ 233, 353, 32 },
		{ 233, 354, 32 },
		{ 233, 355, 32 },
		{ 233, 356, 32 },
		{ 233, 357, 32 },
		{ 233, 358, 32 },
		{ 233, 359, 32 },
		{ 233, 360, 32 },
		{ 233, 361, 32 },
		{ 233, 362, 32 },
		{ 233, 363, 32 },
		{ 233, 364, 32 },
		{ 233, 365, 32 },
		{ 233, 366, 32 },
		{ 233, 367, 32 },
		{ 233, 368, 32 },
		{ 233, 369, 32 },
		{ 233, 370, 32 },
		{ 233, 371, 32 },
		{ 233, 372, 32 },
		{ 233, 373, 32 },
		{ 233, 374, 32 },
		{ 233, 375, 32 },
		{ 233, 376, 32 },
		{ 233, 377, 32 },
		{ 233, 378, 32 },
		{ 233, 379, 32 },
		{ 233, 380, 32 },
		{ 233, 381, 32 },
		{ 233, 382, 32 },
		{ 233, 383, 32 },
		{ 233, 384, 32 },
		{ 233, 385, 32 },
		{ 233, 386, 32 },
		{ 233, 387, 32 },
		{ 233, 388, 32 },
		{ 233, 389, 32 },
		{ 233, 390, 32 },
		{ 233, 391, 32 },
		{ 233, 392, 32 },
		{ 233, 393, 32 },
		{ 233, 394, 32 },
		{ 233, 395, 32 },
		{ 233, 396, 32 },
		{ 233, 397, 32 },
		{ 233, 398, 32 },
		{ 233, 399, 32 },
		{ 233, 400, 32 },
		{ 233, 401, 32 },
		{ 233, 402, 32 },
		{ 233, 403, 32 },
		{ 233, 404, 32 },
		{ 233, 405, 32 },
		{ 233, 406, 32 },
		{ 233, 407, 32 },
		{ 233, 408, 32 },
		{ 233, 409, 32 },
		{ 233, 410, 32 },
		{ 233, 411, 32 },
		{ 233, 412, 32 },
		{ 233, 413, 32 },
		{ 233, 414, 32 },
		{ 233, 415, 32 },
		{ 233, 416, 32 },
		{ 233, 417, 32 },
		{ 233, 418, 32 },
		{ 233, 419, 32 },
		{ 233, 420, 32 },
		{ 233, 421, 32 },
		{ 233, 422, 32 },
		{ 233, 423, 32 },
		{ 233, 424, 32 },
		{ 233, 425, 32 },
		{ 233, 426, 32 },
		{ 233, 427, 32 },
		{ 233, 428, 32 },
		{ 233, 429, 32 },
		{ 233, 430, 32 },
		{ 233, 431, 32 },
		{ 233, 432, 32 },
		{ 233, 433, 32 },
		{ 233, 434, 32 },
		{ 233, 435, 32 },
		{ 233, 436, 32 },
		{ 233, 437, 32 },
		{ 233, 438, 32 },
		{ 233, 439, 32 },
		{ 233, 440, 32 },
		{ 233, 441, 32 },
		{ 233, 442, 32 },
		{ 233, 443, 32 },
		{ 233, 444, 32 },
		{ 233, 445, 32 },
		{ 233, 446, 32 },
		{ 233, 447, 32 },
		{ 233, 448, 32 },
		{ 234, 324, 32 },
		{ 234, 325, 32 },
		{ 234, 326, 32 },
		{ 234, 327, 32 },
		{ 234, 328, 32 },
		{ 234, 329, 32 },
		{ 234, 330, 32 },
		{ 234, 331, 32 },
		{ 234, 332, 32 },
		{ 234, 333, 32 },
		{ 234, 334, 32 },
		{ 234, 335, 32 },
		{ 234, 336, 32 },
		{ 234, 337, 32 },
		{ 234, 338, 32 },
		{ 234, 339, 32 },
		{ 234, 340, 32 },
		{ 234, 341, 32 },
		{ 234, 342, 32 },
		{ 234, 343, 32 },
		{ 234, 344, 32 },
		{ 234, 345, 32 },
		{ 234, 346, 32 },
		{ 234, 347, 32 },
		{ 234, 348, 32 },
		{ 234, 349, 32 },
		{ 234, 350, 32 },
		{ 234, 351, 32 },
		{ 234, 352, 32 },
		{ 234, 353, 32 },
		{ 234, 354, 32 },
		{ 234, 355, 32 },
		{ 234, 356, 32 },
		{ 234, 357, 32 },
		{ 234, 358, 32 },
		{ 234, 359, 32 },
		{ 234, 360, 32 },
		{ 234, 361, 32 },
		{ 234, 362, 32 },
		{ 234, 363, 32 },
		{ 234, 364, 32 },
		{ 234, 365, 32 },
		{ 234, 366, 32 },
		{ 234, 367, 32 },
		{ 234, 368, 32 },
		{ 234, 369, 32 },
		{ 234, 370, 32 },
		{ 234, 371, 32 },
		{ 234, 372, 32 },
		{ 234, 373, 32 },
		{ 234, 374, 32 },
		{ 234, 375, 32 },
		{ 234, 376, 32 },
		{ 234, 377, 32 },
		{ 234, 378, 32 },
		{ 234, 379, 32 },
		{ 234, 380, 32 },
		{ 234, 381, 32 },
		{ 234, 382, 32 },
		{ 234, 383, 32 },
		{ 234, 384, 32 },
		{ 234, 385, 32 },
		{ 234, 386, 32 },
		{ 234, 387, 32 },
		{ 234, 388, 32 },
		{ 234, 389, 32 },
		{ 234, 390, 32 },
		{ 234, 391, 32 },
		{ 234, 392, 32 },
		{ 234, 393, 32 },
		{ 234, 394, 32 },
		{ 234, 395, 32 },
		{ 234, 396, 32 },
		{ 234, 397, 32 },
		{ 234, 398, 32 },
		{ 234, 399, 32 },
		{ 234, 400, 32 },
		{ 234, 401, 32 },
		{ 234, 402, 32 },
		{ 234, 403, 32 },
		{ 234, 404, 32 },
		{ 234, 405, 32 },
		{ 234, 406, 32 },
		{ 234, 407, 32 },
		{ 234, 408, 32 },
		{ 234, 409, 32 },
		{ 234, 410, 32 },
		{ 234, 411, 32 },
		{ 234, 412, 32 },
		{ 234, 413, 32 },
		{ 234, 414, 32 },
		{ 234, 415, 32 },
		{ 234, 416, 32 },
		{ 234, 417, 32 },
		{ 234, 418, 32 },
		{ 234, 419, 32 },
		{ 234, 420, 32 },
		{ 234, 421, 32 },
		{ 234, 422, 32 },
		{ 234, 423, 32 },
		{ 234, 424, 32 },
		{ 234, 425, 32 },
		{ 234, 426, 32 },
		{ 234, 427, 32 },
		{ 234, 428, 32 },
		{ 234, 429, 32 },
		{ 234, 430, 32 },
		{ 234, 431, 32 },
		{ 234, 432, 32 },
		{ 234, 433, 32 },
		{ 234, 434, 32 },
		{ 234, 435, 32 },
		{ 234, 436, 32 },
		{ 234, 437, 32 },
		{ 234, 438, 32 },
		{ 234, 439, 32 },
		{ 234, 440, 32 },
		{ 235, 323, 32 },
		{ 235, 324, 32 },
		{ 235, 325, 32 },
		{ 235, 326, 32 },
		{ 235, 327, 32 },
		{ 235, 328, 32 },
		{ 235, 329, 32 },
		{ 235, 330, 32 },
		{ 235, 331, 32 },
		{ 235, 332, 32 },
		{ 235, 333, 32 },
		{ 235, 334, 32 },
		{ 235, 335, 32 },
		{ 235, 336, 32 },
		{ 235, 337, 32 },
		{ 235, 338, 32 },
		{ 235, 339, 32 },
		{ 235, 340, 32 },
		{ 235, 341, 32 },
		{ 235, 342, 32 },
		{ 235, 343, 32 },
		{ 235, 344, 32 },
		{ 235, 345, 32 },
		{ 235, 346, 32 },
		{ 235, 347, 32 },
		{ 235, 348, 32 },
		{ 235, 349, 32 },
		{ 235, 350, 32 },
		{ 235, 351, 32 },
		{ 235, 352, 32 },
		{ 235, 353, 32 },
		{ 235, 354, 32 },
		{ 235, 355, 32 },
		{ 235, 356, 32 },
		{ 235, 357, 32 },
		{ 235, 358, 32 },
		{ 235, 359, 32 },
		{ 235, 360, 32 },
		{ 235, 361, 32 },
		{ 235, 362, 32 },
		{ 235, 363, 32 },
		{ 235, 364, 32 },
		{ 235, 365, 32 },
		{ 235, 366, 32 },
		{ 235, 367, 32 },
		{ 235, 368, 32 },
		{ 235, 369, 32 },
		{ 235, 370, 32 },
		{ 235, 371, 32 },
		{ 235, 372, 32 },
		{ 235, 373, 32 },
		{ 235, 374, 32 },
		{ 235, 375, 32 },
		{ 235, 376, 32 },
		{ 235, 377, 32 },
		{ 235, 378, 32 },
		{ 235, 379, 32 },
		{ 235, 380, 32 },
		{ 235, 381, 32 },
		{ 235, 382, 32 },
		{ 235, 383, 32 },
		{ 235, 384, 32 },
		{ 235, 385, 32 },
		{ 235, 386, 32 },
		{ 235, 387, 32 },
		{ 235, 388, 32 },
		{ 235, 389, 32 },
		{ 235, 390, 32 },
		{ 235, 391, 32 },
		{ 235, 392, 32 },
		{ 235, 393, 32 },
		{ 235, 394, 32 },
		{ 235, 395, 32 },
		{ 235, 396, 32 },
		{ 235, 397, 32 },
		{ 235, 398, 32 },
		{ 235, 399, 32 },
		{ 235, 400, 32 },
		{ 235, 401, 32 },
		{ 235, 402, 32 },
		{ 235, 403, 32 },
		{ 235, 404, 32 },
		{ 235, 405, 32 },
		{ 235, 406, 32 },
		{ 235, 407, 32 },
		{ 235, 408, 32 },
		{ 235, 409, 32 },
		{ 235, 410, 32 },
		{ 235, 411, 32 },
		{ 235, 412, 32 },
		{ 235, 413, 32 },
		{ 235, 414, 32 },
		{ 235, 415, 32 },
		{ 235, 416, 32 },
		{ 235, 417, 32 },
		{ 235, 418, 32 },
		{ 235, 419, 32 },
		{ 235, 420, 32 },
		{ 235, 421, 32 },
		{ 235, 422, 32 },
		{ 235, 423, 32 },
		{ 235, 424, 32 },
		{ 235, 425, 32 },
		{ 235, 426, 32 },
		{ 235, 427, 32 },
		{ 235, 428, 32 },
		{ 235, 429, 32 },
		{ 235, 430, 32 },
		{ 235, 431, 32 },
		{ 235, 432, 32 },
		{ 236, 323, 32 },
		{ 236, 324, 32 },
		{ 236, 325, 32 },
		{ 236, 326, 32 },
		{ 236, 327, 32 },
		{ 236, 328, 32 },
		{ 236, 329, 32 },
		{ 236, 330, 32 },
		{ 236, 331, 32 },
		{ 236, 332, 32 },
		{ 236, 333, 32 },
		{ 236, 334, 32 },
		{ 236, 335, 32 },
		{ 236, 336, 32 },
		{ 236, 337, 32 },
		{ 236, 338, 32 },
		{ 236, 339, 32 },
		{ 236, 340, 32 },
		{ 236, 341, 32 },
		{ 236, 342, 32 },
		{ 236, 343, 32 },
		{ 236, 344, 32 },
		{ 236, 345, 32 },
		{ 236, 346, 32 },
		{ 236, 347, 32 },
		{ 236, 348, 32 },
		{ 236, 349, 32 },
		{ 236, 350, 32 },
		{ 236, 351, 32 },
		{ 236, 352, 32 },
		{ 236, 353, 32 },
		{ 236, 354, 32 },
		{ 236, 355, 32 },
		{ 236, 356, 32 },
		{ 236, 357, 32 },
		{ 236, 358, 32 },
		{ 236, 359, 32 },
		{ 236, 360, 32 },
		{ 236, 361, 32 },
		{ 236, 362, 32 },
		{ 236, 363, 32 },
		{ 236, 364, 32 },
		{ 236, 365, 32 },
		{ 236, 366, 32 },
		{ 236, 367, 32 },
		{ 236, 368, 32 },
		{ 236, 369, 32 },
		{ 236, 370, 32 },
		{ 236, 371, 32 },
		{ 236, 372, 32 },
		{ 236, 373, 32 },
		{ 236, 374, 32 },
		{ 236, 375, 32 },
		{ 236, 376, 32 },
		{ 236, 377, 32 },
		{ 236, 378, 32 },
		{ 236, 379, 32 },
		{ 236, 380, 32 },
		{ 236, 381, 32 },
		{ 236, 382, 32 },
		{ 236, 383, 32 },
		{ 236, 384, 32 },
		{ 236, 385, 32 },
		{ 236, 386, 32 },
		{ 236, 387, 32 },
		{ 236, 388, 32 },
		{ 236, 389, 32 },
		{ 236, 390, 32 },
		{ 236, 391, 32 },
		{ 236, 392, 32 },
		{ 236, 393, 32 },
		{ 236, 394, 32 },
		{ 236, 395, 32 },
		{ 236, 396, 32 },
		{ 236, 397, 32 },
		{ 236, 398, 32 },
		{ 236, 399, 32 },
		{ 236, 400, 32 },
		{ 236, 401, 32 },
		{ 236, 402, 32 },
		{ 236, 403, 32 },
		{ 236, 404, 32 },
		{ 236, 405, 32 },
		{ 236, 406, 32 },
		{ 236, 407, 32 },
		{ 236, 408, 32 },
		{ 236, 409, 32 },
		{ 236, 410, 32 },
		{ 236, 411, 32 },
		{ 236, 412, 32 },
		{ 236, 413, 32 },
		{ 236, 414, 32 },
		{ 236, 415, 32 },
		{ 236, 416, 32 },
		{ 236, 417, 32 },
		{ 236, 418, 32 },
		{ 236, 419, 32 },
		{ 236, 420, 32 },
		{ 236, 421, 32 },
		{ 236, 422, 32 },
		{ 236, 423, 32 },
		{ 236, 424, 32 },
		{ 237, 323, 32 },
		{ 237, 324, 32 },
		{ 237, 325, 32 },
		{ 237, 326, 32 },
		{ 237, 327, 32 },
		{ 237, 328, 32 },
		{ 237, 329, 32 },
		{ 237, 330, 32 },
		{ 237, 331, 32 },
		{ 237, 332, 32 },
		{ 237, 333, 32 },
		{ 237, 334, 32 },
		{ 237, 335, 32 },
		{ 237, 336, 32 },
		{ 237, 337, 32 },
		{ 237, 338, 32 },
		{ 237, 339, 32 },
		{ 237, 340, 32 },
		{ 237, 341, 32 },
		{ 237, 342, 32 },
		{ 237, 343, 32 },
		{ 237, 344, 32 },
		{ 237, 345, 32 },
		{ 237, 346, 32 },
		{ 237, 347, 32 },
		{ 237, 348, 32 },
		{ 237, 349, 32 },
		{ 237, 350, 32 },
		{ 237, 351, 32 },
		{ 237, 352, 32 },
		{ 237, 353, 32 },
		{ 237, 354, 32 },
		{ 237, 355, 32 },
		{ 237, 356, 32 },
		{ 237, 357, 32 },
		{ 237, 358, 32 },
		{ 237, 359, 32 },
		{ 237, 360, 32 },
		{ 237, 361, 32 },
		{ 237, 362, 32 },
		{ 237, 363, 32 },
		{ 237, 364, 32 },
		{ 237, 365, 32 },
		{ 237, 366, 32 },
		{ 237, 367, 32 },
		{ 237, 368, 32 },
		{ 237, 369, 32 },
		{ 237, 370, 32 },
		{ 237, 371, 32 },
		{ 237, 372, 32 },
		{ 237, 373, 32 },
		{ 237, 374, 32 },
		{ 237, 375, 32 },
		{ 237, 376, 32 },
		{ 237, 377, 32 },
		{ 237, 378, 32 },
		{ 237, 379, 32 },
		{ 237, 380, 32 },
		{ 237, 381, 32 },
		{ 237, 382, 32 },
		{ 237, 383, 32 },
		{ 237, 384, 32 },
		{ 237, 385, 32 },
		{ 237, 386, 32 },
		{ 237, 387, 32 },
		{ 237, 388, 32 },
		{ 237, 389, 32 },
		{ 237, 390, 32 },
		{ 237, 391, 32 },
		{ 237, 392, 32 },
		{ 237, 393, 32 },
		{ 237, 394, 32 },
		{ 237, 395, 32 },
		{ 237, 396, 32 },
		{ 237, 397, 32 },
		{ 237, 398, 32 },
		{ 237, 399, 32 },
		{ 237, 400, 32 },
		{ 237, 401, 32 },
		{ 237, 402, 32 },
		{ 237, 403, 32 },
		{ 237, 404, 32 },
		{ 237, 405, 32 },
		{ 237, 406, 32 },
		{ 237, 407, 32 },
		{ 237, 408, 32 },
		{ 237, 409, 32 },
		{ 237, 410, 32 },
		{ 237, 411, 32 },
		{ 237, 412, 32 },
		{ 237, 413, 32 },
		{ 237, 414, 32 },
		{ 237, 415, 32 },
		{ 237, 416, 32 },
		{ 238, 323, 32 },
		{ 238, 324, 32 },
		{ 238, 325, 32 },
		{ 238, 326, 32 },
		{ 238, 327, 32 },
		{ 238, 328, 32 },
		{ 238, 329, 32 },
		{ 238, 330, 32 },
		{ 238, 331, 32 },
		{ 238, 332, 32 },
		{ 238, 333, 32 },
		{ 238, 334, 32 },
		{ 238, 335, 32 },
		{ 238, 336, 32 },
		{ 238, 337, 32 },
		{ 238, 338, 32 },
		{ 238, 339, 32 },
		{ 238, 340, 32 },
		{ 238, 341, 32 },
		{ 238, 342, 32 },
		{ 238, 343, 32 },
		{ 238, 344, 32 },
		{ 238, 345, 32 },
		{ 238, 346, 32 },
		{ 238, 347, 32 },
		{ 238, 348, 32 },
		{ 238, 349, 32 },
		{ 238, 350, 32 },
		{ 238, 351, 32 },
		{ 238, 352, 32 },
		{ 238, 353, 32 },
		{ 238, 354, 32 },
		{ 238, 355, 32 },
		{ 238, 356, 32 },
		{ 238, 357, 32 },
		{ 238, 358, 32 },
		{ 238, 359, 32 },
		{ 238, 360, 32 },
		{ 238, 361, 32 },
		{ 238, 362, 32 },
		{ 238, 363, 32 },
		{ 238, 364, 32 },
		{ 238, 365, 32 },
		{ 238, 366, 32 },
		{ 238, 367, 32 },
		{ 238, 368, 32 },
		{ 238, 369, 32 },
		{ 238, 370, 32 },
		{ 238, 371, 32 },
		{ 238, 372, 32 },
		{ 238, 373, 32 },
		{ 238, 374, 32 },
		{ 238, 375, 32 },
		{ 238, 376, 32 },
		{ 238, 377, 32 },
		{ 238, 378, 32 },
		{ 238, 379, 32 },
		{ 238, 380, 32 },
		{ 238, 381, 32 },
		{ 238, 382, 32 },
		{ 238, 383, 32 },
		{ 238, 384, 32 },
		{ 238, 385, 32 },
		{ 238, 386, 32 },
		{ 238, 387, 32 },
		{ 238, 388, 32 },
		{ 238, 389, 32 },
		{ 238, 390, 32 },
		{ 238, 391, 32 },
		{ 238, 392, 32 },
		{ 238, 393, 32 },
		{ 238, 394, 32 },
		{ 238, 395, 32 },
		{ 238, 396, 32 },
		{ 238, 397, 32 },
		{ 238, 398, 32 },
		{ 238, 399, 32 },
		{ 238, 400, 32 },
		{ 238, 401, 32 },
		{ 238, 402, 32 },
		{ 238, 403, 32 },
		{ 238, 404, 32 },
		{ 238, 405, 32 },
		{ 238, 406, 32 },
		{ 238, 407, 32 },
		{ 238, 408, 32 },
		{ 239, 323, 32 },
		{ 239, 324, 32 },
		{ 239, 325, 32 },
		{ 239, 326, 32 },
		{ 239, 327, 32 },
		{ 239, 328, 32 },
		{ 239, 329, 32 },
		{ 239, 330, 32 },
		{ 239, 331, 32 },
		{ 239, 332, 32 },
		{ 239, 333, 32 },
		{ 239, 334, 32 },
		{ 239, 335, 32 },
		{ 239, 336, 32 },
		{ 239, 337, 32 },
		{ 239, 338, 32 },
		{ 239, 339, 32 },
		{ 239, 340, 32 },
		{ 239, 341, 32 },
		{ 239, 342, 32 },
		{ 239, 343, 32 },
		{ 239, 344, 32 },
		{ 239, 345, 32 },
		{ 239, 346, 32 },
		{ 239, 347, 32 },
		{ 239, 348, 32 },
		{ 239, 349, 32 },
		{ 239, 350, 32 },
		{ 239, 351, 32 },
		{ 239, 352, 32 },
		{ 239, 353, 32 },
		{ 239, 354, 32 },
		{ 239, 355, 32 },
		{ 239, 356, 32 },
		{ 239, 357, 32 },
		{ 239, 358, 32 },
		{ 239, 359, 32 },
		{ 239, 360, 32 },
		{ 239, 361, 32 },
		{ 239, 362, 32 },
		{ 239, 363, 32 },
		{ 239, 364, 32 },
		{ 239, 365, 32 },
		{ 239, 366, 32 },
		{ 239, 367, 32 },
		{ 239, 368, 32 },
		{ 239, 369, 32 },
		{ 239, 370, 32 },
		{ 239, 371, 32 },
		{ 239, 372, 32 },
		{ 239, 373, 32 },
		{ 239, 374, 32 },
		{ 239, 375, 32 },
		{ 239, 376, 32 },
		{ 239, 377, 32 },
		{ 239, 378, 32 },
		{ 239, 379, 32 },
		{ 239, 380, 32 },
		{ 239, 381, 32 },
		{ 239, 382, 32 },
		{ 239, 383, 32 },
		{ 239, 384, 32 },
		{ 239, 385, 32 },
		{ 239, 386, 32 },
		{ 239, 387, 32 },
		{ 239, 388, 32 },
		{ 239, 389, 32 },
		{ 239, 390, 32 },
		{ 239, 391, 32 },
		{ 239, 392, 32 },
		{ 239, 393, 32 },
		{ 239, 394, 32 },
		{ 239, 395, 32 },
		{ 239, 396, 32 },
		{ 239, 397, 32 },
		{ 239, 398, 32 },
		{ 239, 399, 32 },
		{ 239, 400, 32 },
		{ 240, 323, 32 },
		{ 240, 324, 32 },
		{ 240, 325, 32 },
		{ 240, 326, 32 },
		{ 240, 327, 32 },
		{ 240, 328, 32 },
		{ 240, 329, 32 },
		{ 240, 330, 32 },
		{ 240, 331, 32 },
		{ 240, 332, 32 },
		{ 240, 333, 32 },
		{ 240, 334, 32 },
		{ 240, 335, 32 },
		{ 240, 336, 32 },
		{ 240, 337, 32 },
		{ 240, 338, 32 },
		{ 240, 339, 32 },
		{ 240, 340, 32 },
		{ 240, 341, 32 },
		{ 240, 342, 32 },
		{ 240, 343, 32 },
		{ 240, 344, 32 },
		{ 240, 345, 32 },
		{ 240, 346, 32 },
		{ 240, 347, 32 },
		{ 240, 348, 32 },
		{ 240, 349, 32 },
		{ 240, 350, 32 },
		{ 240, 351, 32 },
		{ 240, 352, 32 },
		{ 240, 353, 32 },
		{ 240, 354, 32 },
		{ 240, 355, 32 },
		{ 240, 356, 32 },
		{ 240, 357, 32 },
		{ 240, 358, 32 },
		{ 240, 359, 32 },
		{ 240, 360, 32 },
		{ 240, 361, 32 },
		{ 240, 362, 32 },
		{ 240, 363, 32 },
		{ 240, 364, 32 },
		{ 240, 365, 32 },
		{ 240, 366, 32 },
		{ 240, 367, 32 },
		{ 240, 368, 32 },
		{ 240, 369, 32 },
		{ 240, 370, 32 },
		{ 240, 371, 32 },
		{ 240, 372, 32 },
		{ 240, 373, 32 },
		{ 240, 374, 32 },
		{ 240, 375, 32 },
		{ 240, 376, 32 },
		{ 240, 377, 32 },
		{ 240, 378, 32 },
		{ 240, 379, 32 },
		{ 240, 380, 32 },
		{ 240, 381, 32 },
		{ 240, 382, 32 },
		{ 240, 383, 32 },
		{ 240, 384, 32 },
		{ 240, 385, 32 },
		{ 240, 386, 32 },
		{ 240, 387, 32 },
		{ 240, 388, 32 },
		{ 240, 389, 32 },
		{ 240, 390, 32 },
		{ 240, 391, 32 },
		{ 240, 392, 32 },
		{ 241, 323, 32 },
		{ 241, 324, 32 },
		{ 241, 325, 32 },
		{ 241, 326, 32 },
		{ 241, 327, 32 },
		{ 241, 328, 32 },
		{ 241, 329, 32 },
		{ 241, 330, 32 },
		{ 241, 331, 32 },
		{ 241, 332, 32 },
		{ 241, 333, 32 },
		{ 241, 334, 32 },
		{ 241, 335, 32 },
		{ 241, 336, 32 },
		{ 241, 337, 32 },
		{ 241, 338, 32 },
		{ 241, 339, 32 },
		{ 241, 340, 32 },
		{ 241, 341, 32 },
		{ 241, 342, 32 },
		{ 241, 343, 32 },
		{ 241, 344, 32 },
		{ 241, 345, 32 },
		{ 241, 346, 32 },
		{ 241, 347, 32 },
		{ 241, 348, 32 },
		{ 241, 349, 32 },
		{ 241, 350, 32 },
		{ 241, 351, 32 },
		{ 241, 352, 32 },
		{ 241, 353, 32 },
		{ 241, 354, 32 },
		{ 241, 355, 32 },
		{ 241, 356, 32 },
		{ 241, 357, 32 },
		{ 241, 358, 32 },
		{ 241, 359, 32 },
		{ 241, 360, 32 },
		{ 241, 361, 32 },
		{ 241, 362, 32 },
		{ 241, 363, 32 },
		{ 241, 364, 32 },
		{ 241, 365, 32 },
		{ 241, 366, 32 },
		{ 241, 367, 32 },
		{ 241, 368, 32 },
		{ 241, 369, 32 },
		{ 241, 370, 32 },
		{ 241, 371, 32 },
		{ 241, 372, 32 },
		{ 241, 373, 32 },
		{ 241, 374, 32 },
		{ 241, 375, 32 },
		{ 241, 376, 32 },
		{ 241, 377, 32 },
		{ 241, 378, 32 },
		{ 241, 379, 32 },
		{ 241, 380, 32 },
		{ 241, 381, 32 },
		{ 241, 382, 32 },
		{ 241, 383, 32 },
		{ 241, 384, 32 },
		{ 242, 323, 32 },
		{ 242, 324, 32 },
		{ 242, 325, 32 },
		{ 242, 326, 32 },
		{ 242, 327, 32 },
		{ 242, 328, 32 },
		{ 242, 329, 32 },
		{ 242, 330, 32 },
		{ 242, 331, 32 },
		{ 242, 332, 32 },
		{ 242, 333, 32 },
		{ 242, 334, 32 },
		{ 242, 335, 32 },
		{ 242, 336, 32 },
		{ 242, 337, 32 },
		{ 242, 338, 32 },
		{ 242, 339, 32 },
		{ 242, 340, 32 },
		{ 242, 341, 32 },
		{ 242, 342, 32 },
		{ 242, 343, 32 },
		{ 242, 344, 32 },
		{ 242, 345, 32 },
		{ 242, 346, 32 },
		{ 242, 347, 32 },
		{ 242, 348, 32 },
		{ 242, 349, 32 },
		{ 242, 350, 32 },
		{ 242, 351, 32 },
		{ 242, 352, 32 },
		{ 242, 353, 32 },
		{ 242, 354, 32 },
		{ 242, 355, 32 },
		{ 242, 356, 32 },
		{ 242, 357, 32 },
		{ 242, 358, 32 },
		{ 242, 359, 32 },
		{ 242, 360, 32 },
		{ 242, 361, 32 },
		{ 242, 362, 32 },
		{ 242, 363, 32 },
		{ 242, 364, 32 },
		{ 242, 365, 32 },
		{ 242, 366, 32 },
		{ 242, 367, 32 },
		{ 242, 368, 32 },
		{ 242, 369, 32 },
		{ 242, 370, 32 },
		{ 242, 371, 32 },
		{ 242, 372, 32 },
		{ 242, 373, 32 },
		{ 242, 374, 32 },
		{ 242, 375, 32 },
		{ 242, 376, 32 },
		{ 243, 323, 32 },
		{ 243, 324, 32 },
		{ 243, 325, 32 },
		{ 243, 326, 32 },
		{ 243, 327, 32 },
		{ 243, 328, 32 },
		{ 243, 329, 32 },
		{ 243, 330, 32 },
		{ 243, 331, 32 },
		{ 243, 332, 32 },
		{ 243, 333, 32 },
		{ 243, 334, 32 },
		{ 243, 335, 32 },
		{ 243, 336, 32 },
		{ 243, 337, 32 },
		{ 243, 338, 32 },
		{ 243, 339, 32 },
		{ 243, 340, 32 },
		{ 243, 341, 32 },
		{ 243, 342, 32 },
		{ 243, 343, 32 },
		{ 243, 344, 32 },
		{ 243, 345, 32 },
		{ 243, 346, 32 },
		{ 243, 347, 32 },
		{ 243, 348, 32 },
		{ 243, 349, 32 },
		{ 243, 350, 32 },
		{ 243, 351, 32 },
		{ 243, 352, 32 },
		{ 243, 353, 32 },
		{ 243, 354, 32 },
		{ 243, 355, 32 },
		{ 243, 356, 32 },
		{ 243, 357, 32 },
		{ 243, 358, 32 },
		{ 243, 359, 32 },
		{ 243, 360, 32 },
		{ 243, 361, 32 },
		{ 243, 362, 32 },
		{ 243, 363, 32 },
		{ 243, 364, 32 },
		{ 243, 365, 32 },
		{ 243, 366, 32 },
		{ 243, 367, 32 },
		{ 243, 368, 32 },
		{ 244, 323, 32 },
		{ 244, 324, 32 },
		{ 244, 325, 32 },
		{ 244, 326, 32 },
		{ 244, 327, 32 },
		{ 244, 328, 32 },
		{ 244, 329, 32 },
		{ 244, 330, 32 },
		{ 244, 331, 32 },
		{ 244, 332, 32 },
		{ 244, 333, 32 },
		{ 244, 334, 32 },
		{ 244, 335, 32 },
		{ 244, 336, 32 },
		{ 244, 337, 32 },
		{ 244, 338, 32 },
		{ 244, 339, 32 },
		{ 244, 340, 32 },
		{ 244, 341, 32 },
		{ 244, 342, 32 },
		{ 244, 343, 32 },
		{ 244, 344, 32 },
		{ 244, 345, 32 },
		{ 244, 346, 32 },
		{ 244, 347, 32 },
		{ 244, 348, 32 },
		{ 244, 349, 32 },
		{ 244, 350, 32 },
		{ 244, 351, 32 },
		{ 244, 352, 32 },
		{ 244, 353, 32 },
		{ 244, 354, 32 },
		{ 244, 355, 32 },
		{ 244, 356, 32 },
		{ 244, 357, 32 },
		{ 244, 358, 32 },
		{ 244, 359, 32 },
		{ 244, 360, 32 },
		{ 245, 323, 32 },
		{ 245, 324, 32 },
		{ 245, 325, 32 },
		{ 245, 326, 32 },
		{ 245, 327, 32 },
		{ 245, 328, 32 },
		{ 245, 329, 32 },
		{ 245, 330, 32 },
		{ 245, 331, 32 },
		{ 245, 332, 32 },
		{ 245, 333, 32 },
		{ 245, 334, 32 },
		{ 245, 335, 32 },
		{ 245, 336, 32 },
		{ 245, 337, 32 },
		{ 245, 338, 32 },
		{ 245, 339, 32 },
		{ 245, 340, 32 },
		{ 245, 341, 32 },
		{ 245, 342, 32 },
		{ 245, 343, 32 },
		{ 245, 344, 32 },
		{ 245, 345, 32 },
		{ 245, 346, 32 },
		{ 245, 347, 32 },
		{ 245, 348, 32 },
		{ 245, 349, 32 },
		{ 245, 350, 32 },
		{ 245, 351, 32 },
		{ 245, 352, 32 },
		{ 246, 323, 32 },
		{ 246, 324, 32 },
		{ 246, 325, 32 },
		{ 246, 326, 32 },
		{ 246, 327, 32 },
		{ 246, 328, 32 },
		{ 246, 329, 32 },
		{ 246, 330, 32 },
		{ 246, 331, 32 },
		{ 246, 332, 32 },
		{ 246, 333, 32 },
		{ 246, 334, 32 },
		{ 246, 335, 32 },
		{ 246, 336, 32 },
		{ 246, 337, 32 },
		{ 246, 338, 32 },
		{ 246, 339, 32 },
		{ 246, 340, 32 },
		{ 246, 341, 32 },
		{ 246, 342, 32 },
		{ 246, 343, 32 },
		{ 246, 344, 32 },
		{ 247, 324, 32 },
		{ 247, 325, 32 },
		{ 247, 326, 32 },
		{ 247, 327, 32 },
		{ 247, 328, 32 },
		{ 247, 329, 32 },
		{ 247, 330, 32 },
		{ 247, 331, 32 },
		{ 247, 332, 32 },
		{ 247, 333, 32 },
		{ 247, 334, 32 },
		{ 247, 335, 32 },
		{ 247, 336, 32 },
		{ 248, 325, 32 },
		{ 248, 326, 32 },
		{ 248, 327, 32 },
		{ 266, 187, 32 },
		{ 266, 188, 32 },
		{ 266, 189, 32 },
		{ 267, 178, 32 },
		{ 267, 179, 32 },
		{ 267, 180, 32 },
		{ 267, 181, 32 },
		{ 267, 182, 32 },
		{ 267, 183, 32 },
		{ 267, 184, 32 },
		{ 267, 185, 32 },
		{ 267, 186, 32 },
		{ 267, 187, 32 },
		{ 267, 188, 32 },
		{ 267, 189, 32 },
		{ 267, 190, 32 },
		{ 268, 170, 32 },
		{ 268, 171, 32 },
		{ 268, 172, 32 },
		{ 268, 173, 32 },
		{ 268, 174, 32 },
		{ 268, 175, 32 },
		{ 268, 176, 32 },
		{ 268, 177, 32 },
		{ 268, 178, 32 },
		{ 268, 179, 32 },
		{ 268, 180, 32 },
		{ 268, 181, 32 },
		{ 268, 182, 32 },
		{ 268, 183, 32 },
		{ 268, 184, 32 },
		{ 268, 185, 32 },
		{ 268, 186, 32 },
		{ 268, 187, 32 },
		{ 268, 188, 32 },
		{ 268, 189, 32 },
		{ 268, 190, 32 },
		{ 268, 191, 32 },
		{ 269, 162, 32 },
		{ 269, 163, 32 },
		{ 269, 164, 32 },
		{ 269, 165, 32 },
		{ 269, 166, 32 },
		{ 269, 167, 32 },
		{ 269, 168, 32 },
		{ 269, 169, 32 },
		{ 269, 170, 32 },
		{ 269, 171, 32 },
		{ 269, 172, 32 },
		{ 269, 173, 32 },
		{ 269, 174, 32 },
		{ 269, 175, 32 },
		{ 269, 176, 32 },
		{ 269, 177, 32 },
		{ 269, 178, 32 },
		{ 269, 179, 32 },
		{ 269, 180, 32 },
		{ 269, 181, 32 },
		{ 269, 182, 32 },
		{ 269, 183, 32 },
		{ 269, 184, 32 },
		{ 269, 185, 32 },
		{ 269, 186, 32 },
		{ 269, 187, 32 },
		{ 269, 188, 32 },
		{ 269, 189, 32 },
		{ 269, 190, 32 },
		{ 269, 191, 32 },
		{ 270, 154, 32 },
		{ 270, 155, 32 },
		{ 270, 156, 32 },
		{ 270, 157, 32 },
		{ 270, 158, 32 },
		{ 270, 159, 32 },
		{ 270, 160, 32 },
		{ 270, 161, 32 },
		{ 270, 162, 32 },
		{ 270, 163, 32 },
		{ 270, 164, 32 },
		{ 270, 165, 32 },
		{ 270, 166, 32 },
		{ 270, 167, 32 },
		{ 270, 168, 32 },
		{ 270, 169, 32 },
		{ 270, 170, 32 },
		{ 270, 171, 32 },
		{ 270, 172, 32 },
		{ 270, 173, 32 },
		{ 270, 174, 32 },
		{ 270, 175, 32 },
		{ 270, 176, 32 },
		{ 270, 177, 32 },
		{ 270, 178, 32 },
		{ 270, 179, 32 },
		{ 270, 180, 32 },
		{ 270, 181, 32 },
		{ 270, 182, 32 },
		{ 270, 183, 32 },
		{ 270, 184, 32 },
		{ 270, 185, 32 },
		{ 270, 186, 32 },
		{ 270, 187, 32 },
		{ 270, 188, 32 },
		{ 270, 189, 32 },
		{ 270, 190, 32 },
		{ 270, 191, 32 },
		{ 271, 146, 32 },
		{ 271, 147, 32 },
		{ 271, 148, 32 },
		{ 271, 149, 32 },
		{ 271, 150, 32 },
		{ 271, 151, 32 },
		{ 271, 152, 32 },
		{ 271, 153, 32 },
		{ 271, 154, 32 },
		{ 271, 155, 32 },
		{ 271, 156, 32 },
		{ 271, 157, 32 },
		{ 271, 158, 32 },
		{ 271, 159, 32 },
		{ 271, 160, 32 },
		{ 271, 161, 32 },
		{ 271, 162, 32 },
		{ 271, 163, 32 },
		{ 271, 164, 32 },
		{ 271, 165, 32 },
		{ 271, 166, 32 },
		{ 271, 167, 32 },
		{ 271, 168, 32 },
		{ 271, 169, 32 },
		{ 271, 170, 32 },
		{ 271, 171, 32 },
		{ 271, 172, 32 },
		{ 271, 173, 32 },
		{ 271, 174, 32 },
		{ 271, 175, 32 },
		{ 271, 176, 32 },
		{ 271, 177, 32 },
		{ 271, 178, 32 },
		{ 271, 179, 32 },
		{ 271, 180, 32 },
		{ 271, 181, 32 },
		{ 271, 182, 32 },
		{ 271, 183, 32 },
		{ 271, 184, 32 },
		{ 271, 185, 32 },
		{ 271, 186, 32 },
		{ 271, 187, 32 },
		{ 271, 188, 32 },
		{ 271, 189, 32 },
		{ 271, 190, 32 },
		{ 271, 191, 32 },
		{ 272, 138, 32 },
		{ 272, 139, 32 },
		{ 272, 140, 32 },
		{ 272, 141, 32 },
		{ 272, 142, 32 },
		{ 272, 143, 32 },
		{ 272, 144, 32 },
		{ 272, 145, 32 },
		{ 272, 146, 32 },
		{ 272, 147, 32 },
		{ 272, 148, 32 },
		{ 272, 149, 32 },
		{ 272, 150, 32 },
		{ 272, 151, 32 },
		{ 272, 152, 32 },
		{ 272, 153, 32 },
		{ 272, 154, 32 },
		{ 272, 155, 32 },
		{ 272, 156, 32 },
		{ 272, 157, 32 },
		{ 272, 158, 32 },
		{ 272, 159, 32 },
		{ 272, 160, 32 },
		{ 272, 161, 32 },
		{ 272, 162, 32 },
		{ 272, 163, 32 },
		{ 272, 164, 32 },
		{ 272, 165, 32 },
		{ 272, 166, 32 },
		{ 272, 167, 32 },
		{ 272, 168, 32 },
		{ 272, 169, 32 },
		{ 272, 170, 32 },
		{ 272, 171, 32 },
		{ 272, 172, 32 },
		{ 272, 173, 32 },
		{ 272, 174, 32 },
		{ 272, 175, 32 },
		{ 272, 176, 32 },
		{ 272, 177, 32 },
		{ 272, 178, 32 },
		{ 272, 179, 32 },
		{ 272, 180, 32 },
		{ 272, 181, 32 },
		{ 272, 182, 32 },
		{ 272, 183, 32 },
		{ 272, 184, 32 },
		{ 272, 185, 32 },
		{ 272, 186, 32 },
		{ 272, 187, 32 },
		{ 272, 188, 32 },
		{ 272, 189, 32 },
		{ 272, 190, 32 },
		{ 272, 191, 32 },
		{ 273, 130, 32 },
		{ 273, 131, 32 },
		{ 273, 132, 32 },
		{ 273, 133, 32 },
		{ 273, 134, 32 },
		{ 273, 135, 32 },
		{ 273, 136, 32 },
		{ 273, 137, 32 },
		{ 273, 138, 32 },
		{ 273, 139, 32 },
		{ 273, 140, 32 },
		{ 273, 141, 32 },
		{ 273, 142, 32 },
		{ 273, 143, 32 },
		{ 273, 144, 32 },
		{ 273, 145, 32 },
		{ 273, 146, 32 },
		{ 273, 147, 32 },
		{ 273, 148, 32 },
		{ 273, 149, 32 },
		{ 273, 150, 32 },
		{ 273, 151, 32 },
		{ 273, 152, 32 },
		{ 273, 153, 32 },
		{ 273, 154, 32 },
		{ 273, 155, 32 },
		{ 273, 156, 32 },
		{ 273, 157, 32 },
		{ 273, 158, 32 },
		{ 273, 159, 32 },
		{ 273, 160, 32 },
		{ 273, 161, 32 },
		{ 273, 162, 32 },
		{ 273, 163, 32 },
		{ 273, 164, 32 },
		{ 273, 165, 32 },
		{ 273, 166, 32 },
		{ 273, 167, 32 },
		{ 273, 168, 32 },
		{ 273, 169, 32 },
		{ 273, 170, 32 },
		{ 273, 171, 32 },
		{ 273, 172, 32 },
		{ 273, 173, 32 },
		{ 273, 174, 32 },
		{ 273, 175, 32 },
		{ 273, 176, 32 },
		{ 273, 177, 32 },
		{ 273, 178, 32 },
		{ 273, 179, 32 },
		{ 273, 180, 32 },
		{ 273, 181, 32 },
		{ 273, 182, 32 },
		{ 273, 183, 32 },
		{ 273, 184, 32 },
		{ 273, 185, 32 },
		{ 273, 186, 32 },
		{ 273, 187, 32 },
		{ 273, 188, 32 },
		{ 273, 189, 32 },
		{ 273, 190, 32 },
		{ 273, 191, 32 },
		{ 274, 122, 32 },
		{ 274, 123, 32 },
		{ 274, 124, 32 },
		{ 274, 125, 32 },
		{ 274, 126, 32 },
		{ 274, 127, 32 },
		{ 274, 128, 32 },
		{ 274, 129, 32 },
		{ 274, 130, 32 },
		{ 274, 131, 32 },
		{ 274, 132, 32 },
		{ 274, 133, 32 },
		{ 274, 134, 32 },
		{ 274, 135, 32 },
		{ 274, 136, 32 },
		{ 274, 137, 32 },
		{ 274, 138, 32 },
		{ 274, 139, 32 },
		{ 274, 140, 32 },
		{ 274, 141, 32 },
		{ 274, 142, 32 },
		{ 274, 143, 32 },
		{ 274, 144, 32 },
		{ 274, 145, 32 },
		{ 274, 146, 32 },
		{ 274, 147, 32 },
		{ 274, 148, 32 },
		{ 274, 149, 32 },
		{ 274, 150, 32 },
		{ 274, 151, 32 },
		{ 274, 152, 32 },
		{ 274, 153, 32 },
		{ 274, 154, 32 },
		{ 274, 155, 32 },
		{ 274, 156, 32 },
		{ 274, 157, 32 },
		{ 274, 158, 32 },
		{ 274, 159, 32 },
		{ 274, 160, 32 },
		{ 274, 161, 32 },
		{ 274, 162, 32 },
		{ 274, 163, 32 },
		{ 274, 164, 32 },
		{ 274, 165, 32 },
		{ 274, 166, 32 },
		{ 274, 167, 32 },
		{ 274, 168, 32 },
		{ 274, 169, 32 },
		{ 274, 170, 32 },
		{ 274, 171, 32 },
		{ 274, 172, 32 },
		{ 274, 173, 32 },
		{ 274, 174, 32 },
		{ 274, 175, 32 },
		{ 274, 176, 32 },
		{ 274, 177, 32 },
		{ 274, 178, 32 },
		{ 274, 179, 32 },
		{ 274, 180, 32 },
		{ 274, 181, 32 },
		{ 274, 182, 32 },
		{ 274, 183, 32 },
		{ 274, 184, 32 },
		{ 274, 185, 32 },
		{ 274, 186, 32 },
		{ 274, 187, 32 },
		{ 274, 188, 32 },
		{ 274, 189, 32 },
		{ 274, 190, 32 },
		{ 274, 191, 32 },
		{ 275, 114, 32 },
		{ 275, 115, 32 },
		{ 275, 116, 32 },
		{ 275, 117, 32 },
		{ 275, 118, 32 },
		{ 275, 119, 32 },
		{ 275, 120, 32 },
		{ 275, 121, 32 },
		{ 275, 122, 32 },
		{ 275, 123, 32 },
		{ 275, 124, 32 },
		{ 275, 125, 32 },
		{ 275, 126, 32 },
		{ 275, 127, 32 },
		{ 275, 128, 32 },
		{ 275, 129, 32 },
		{ 275, 130, 32 },
		{ 275, 131, 32 },
		{ 275, 132, 32 },
		{ 275, 133, 32 },
		{ 275, 134, 32 },
		{ 275, 135, 32 },
		{ 275, 136, 32 },
		{ 275, 137, 32 },
		{ 275, 138, 32 },
		{ 275, 139, 32 },
		{ 275, 140, 32 },
		{ 275, 141, 32 },
		{ 275, 142, 32 },
		{ 275, 143, 32 },
		{ 275, 144, 32 },
		{ 275, 145, 32 },
		{ 275, 146, 32 },
		{ 275, 147, 32 },
		{ 275, 148, 32 },
		{ 275, 149, 32 },
		{ 275, 150, 32 },
		{ 275, 151, 32 },
		{ 275, 152, 32 },
		{ 275, 153, 32 },
		{ 275, 154, 32 },
		{ 275, 155, 32 },
		{ 275, 156, 32 },
		{ 275, 157, 32 },
		{ 275, 158, 32 },
		{ 275, 159, 32 },
		{ 275, 160, 32 },
		{ 275, 161, 32 },
		{ 275, 162, 32 },
		{ 275, 163, 32 },
		{ 275, 164, 32 },
		{ 275, 165, 32 },
		{ 275, 166, 32 },
		{ 275, 167, 32 },
		{ 275, 168, 32 },
		{ 275, 169, 32 },
		{ 275, 170, 32 },
		{ 275, 171, 32 },
		{ 275, 172, 32 },
		{ 275, 173, 32 },
		{ 275, 174, 32 },
		{ 275, 175, 32 },
		{ 275, 176, 32 },
		{ 275, 177, 32 },
		{ 275, 178, 32 },
		{ 275, 179, 32 },
		{ 275, 180, 32 },
		{ 275, 181, 32 },
		{ 275, 182, 32 },
		{ 275, 183, 32 },
		{ 275, 184, 32 },
		{ 275, 185, 32 },
		{ 275, 186, 32 },
		{ 275, 187, 32 },
		{ 275, 188, 32 },
		{ 275, 189, 32 },
		{ 275, 190, 32 },
		{ 275, 191, 32 },
		{ 276, 106, 32 },
		{ 276, 107, 32 },
		{ 276, 108, 32 },
		{ 276, 109, 32 },
		{ 276, 110, 32 },
		{ 276, 111, 32 },
		{ 276, 112, 32 },
		{ 276, 113, 32 },
		{ 276, 114, 32 },
		{ 276, 115, 32 },
		{ 276, 116, 32 },
		{ 276, 117, 32 },
		{ 276, 118, 32 },
		{ 276, 119, 32 },
		{ 276, 120, 32 },
		{ 276, 121, 32 },
		{ 276, 122, 32 },
		{ 276, 123, 32 },
		{ 276, 124, 32 },
		{ 276, 125, 32 },
		{ 276, 126, 32 },
		{ 276, 127, 32 },
		{ 276, 128, 32 },
		{ 276, 129, 32 },
		{ 276, 130, 32 },
		{ 276, 131, 32 },
		{ 276, 132, 32 },
		{ 276, 133, 32 },
		{ 276, 134, 32 },
		{ 276, 135, 32 },
		{ 276, 136, 32 },
		{ 276, 137, 32 },
		{ 276, 138, 32 },
		{ 276, 139, 32 },
		{ 276, 140, 32 },
		{ 276, 141, 32 },
		{ 276, 142, 32 },
		{ 276, 143, 32 },
		{ 276, 144, 32 },
		{ 276, 145, 32 },
		{ 276, 146, 32 },
		{ 276, 147, 32 },
		{ 276, 148, 32 },
		{ 276, 149, 32 },
		{ 276, 150, 32 },
		{ 276, 151, 32 },
		{ 276, 152, 32 },
		{ 276, 153, 32 },
		{ 276, 154, 32 },
		{ 276, 155, 32 },
		{ 276, 156, 32 },
		{ 276, 157, 32 },
		{ 276, 158, 32 },
		{ 276, 159, 32 },
		{ 276, 160, 32 },
		{ 276, 161, 32 },
		{ 276, 162, 32 },
		{ 276, 163, 32 },
		{ 276, 164, 32 },
		{ 276, 165, 32 },
		{ 276, 166, 32 },
		{ 276, 167, 32 },
		{ 276, 168, 32 },
		{ 276, 169, 32 },
		{ 276, 170, 32 },
		{ 276, 171, 32 },
		{ 276, 172, 32 },
		{ 276, 173, 32 },
		{ 276, 174, 32 },
		{ 276, 175, 32 },
		{ 276, 176, 32 },
		{ 276, 177, 32 },
		{ 276, 178, 32 },
		{ 276, 179, 32 },
		{ 276, 180, 32 },
		{ 276, 181, 32 },
		{ 276, 182, 32 },
		{ 276, 183, 32 },
		{ 276, 184, 32 },
		{ 276, 185, 32 },
		{ 276, 186, 32 },
		{ 276, 187, 32 },
		{ 276, 188, 32 },
		{ 276, 189, 32 },
		{ 276, 190, 32 },
		{ 276, 191, 32 },
		{ 277, 98, 32 },
		{ 277, 99, 32 },
		{ 277, 100, 32 },
		{ 277, 101, 32 },
		{ 277, 102, 32 },
		{ 277, 103, 32 },
		{ 277, 104, 32 },
		{ 277, 105, 32 },
		{ 277, 106, 32 },
		{ 277, 107, 32 },
		{ 277, 108, 32 },
		{ 277, 109, 32 },
		{ 277, 110, 32 },
		{ 277, 111, 32 },
		{ 277, 112, 32 },
		{ 277, 113, 32 },
		{ 277, 114, 32 },
		{ 277, 115, 32 },
		{ 277, 116, 32 },
		{ 277, 117, 32 },
		{ 277, 118, 32 },
		{ 277, 119, 32 },
		{ 277, 120, 32 },
		{ 277, 121, 32 },
		{ 277, 122, 32 },
		{ 277, 123, 32 },
		{ 277, 124, 32 },
		{ 277, 125, 32 },
		{ 277, 126, 32 },
		{ 277, 127, 32 },
		{ 277, 128, 32 },
		{ 277, 129, 32 },
		{ 277, 130, 32 },
		{ 277, 131, 32 },
		{ 277, 132, 32 },
		{ 277, 133, 32 },
		{ 277, 134, 32 },
		{ 277, 135, 32 },
		{ 277, 136, 32 },
		{ 277, 137, 32 },
		{ 277, 138, 32 },
		{ 277, 139, 32 },
		{ 277, 140, 32 },
		{ 277, 141, 32 },
		{ 277, 142, 32 },
		{ 277, 143, 32 },
		{ 277, 144, 32 },
		{ 277, 145, 32 },
		{ 277, 146, 32 },
		{ 277, 147, 32 },
		{ 277, 148, 32 },
		{ 277, 149, 32 },
		{ 277, 150, 32 },
		{ 277, 151, 32 },
		{ 277, 152, 32 },
		{ 277, 153, 32 },
		{ 277, 154, 32 },
		{ 277, 155, 32 },
		{ 277, 156, 32 },
		{ 277, 157, 32 },
		{ 277, 158, 32 },
		{ 277, 159, 32 },
		{ 277, 160, 32 },
		{ 277, 161, 32 },
		{ 277, 162, 32 },
		{ 277, 163, 32 },
		{ 277, 164, 32 },
		{ 277, 165, 32 },
		{ 277, 166, 32 },
		{ 277, 167, 32 },
		{ 277, 168, 32 },
		{ 277, 169, 32 },
		{ 277, 170, 32 },
		{ 277, 171, 32 },
		{ 277, 172, 32 },
		{ 277, 173, 32 },
		{ 277, 174, 32 },
		{ 277, 175, 32 },
		{ 277, 176, 32 },
		{ 277, 177, 32 },
		{ 277, 178, 32 },
		{ 277, 179, 32 },
		{ 277, 180, 32 },
		{ 277, 181, 32 },
		{ 277, 182, 32 },
		{ 277, 183, 32 },
		{ 277, 184, 32 },
		{ 277, 185, 32 },
		{ 277, 186, 32 },
		{ 277, 187, 32 },
		{ 277, 188, 32 },
		{ 277, 189, 32 },
		{ 277, 190, 32 },
		{ 277, 191, 32 },
		{ 278, 90, 32 },
		{ 278, 91, 32 },
		{ 278, 92, 32 },
		{ 278, 93, 32 },
		{ 278, 94, 32 },
		{ 278, 95, 32 },
		{ 278, 96, 32 },
		{ 278, 97, 32 },
		{ 278, 98, 32 },
		{ 278, 99, 32 },
		{ 278, 100, 32 },
		{ 278, 101, 32 },
		{ 278, 102, 32 },
		{ 278, 103, 32 },
		{ 278, 104, 32 },
		{ 278, 105, 32 },
		{ 278, 106, 32 },
		{ 278, 107, 32 },
		{ 278, 108, 32 },
		{ 278, 109, 32 },
		{ 278, 110, 32 },
		{ 278, 111, 32 },
		{ 278, 112, 32 },
		{ 278, 113, 32 },
		{ 278, 114, 32 },
		{ 278, 115, 32 },
		{ 278, 116, 32 },
		{ 278, 117, 32 },
		{ 278, 118, 32 },
		{ 278, 119, 32 },
		{ 278, 120, 32 },
		{ 278, 121, 32 },
		{ 278, 122, 32 },
		{ 278, 123, 32 },
		{ 278, 124, 32 },
		{ 278, 125, 32 },
		{ 278, 126, 32 },
		{ 278, 127, 32 },
		{ 278, 128, 32 },
		{ 278, 129, 32 },
		{ 278, 130, 32 },
		{ 278, 131, 32 },
		{ 278, 132, 32 },
		{ 278, 133, 32 },
		{ 278, 134, 32 },
		{ 278, 135, 32 },
		{ 278, 136, 32 },
		{ 278, 137, 32 },
		{ 278, 138, 32 },
		{ 278, 139, 32 },
		{ 278, 140, 32 },
		{ 278, 141, 32 },
		{ 278, 142, 32 },
		{ 278, 143, 32 },
		{ 278, 144, 32 },
		{ 278, 145, 32 },
		{ 278, 146, 32 },
		{ 278, 147, 32 },
		{ 278, 148, 32 },
		{ 278, 149, 32 },
		{ 278, 150, 32 },
		{ 278, 151, 32 },
		{ 278, 152, 32 },
		{ 278, 153, 32 },
		{ 278, 154, 32 },
		{ 278, 155, 32 },
		{ 278, 156, 32 },
		{ 278, 157, 32 },
		{ 278, 158, 32 },
		{ 278, 159, 32 },
		{ 278, 160, 32 },
		{ 278, 161, 32 },
		{ 278, 162, 32 },
		{ 278, 163, 32 },
		{ 278, 164, 32 },
		{ 278, 165, 32 },
		{ 278, 166, 32 },
		{ 278, 167, 32 },
		{ 278, 168, 32 },
		{ 278, 169, 32 },
		{ 278, 170, 32 },
		{ 278, 171, 32 },
		{ 278, 172, 32 },
		{ 278, 173, 32 },
		{ 278, 174, 32 },
		{ 278, 175, 32 },
		{ 278, 176, 32 },
		{ 278, 177, 32 },
		{ 278, 178, 32 },
		{ 278, 179, 32 },
		{ 278, 180, 32 },
		{ 278, 181, 32 },
		{ 278, 182, 32 },
		{ 278, 183, 32 },
		{ 278, 184, 32 },
		{ 278, 185, 32 },
		{ 278, 186, 32 },
		{ 278, 187, 32 },
		{ 278, 188, 32 },
		{ 278, 189, 32 },
		{ 278, 190, 32 },
		{ 278, 191, 32 },
		{ 279, 82, 32 },
		{ 279, 83, 32 },
		{ 279, 84, 32 },
		{ 279, 85, 32 },
		{ 279, 86, 32 },
		{ 279, 87, 32 },
		{ 279, 88, 32 },
		{ 279, 89, 32 },
		{ 279, 90, 32 },
		{ 279, 91, 32 },
		{ 279, 92, 32 },
		{ 279, 93, 32 },
		{ 279, 94, 32 },
		{ 279, 95, 32 },
		{ 279, 96, 32 },
		{ 279, 97, 32 },
		{ 279, 98, 32 },
		{ 279, 99, 32 },
		{ 279, 100, 32 },
		{ 279, 101, 32 },
		{ 279, 102, 32 },
		{ 279, 103, 32 },
		{ 279, 104, 32 },
		{ 279, 105, 32 },
		{ 279, 106, 32 },
		{ 279, 107, 32 },
		{ 279, 108, 32 },
		{ 279, 109, 32 },
		{ 279, 110, 32 },
		{ 279, 111, 32 },
		{ 279, 112, 32 },
		{ 279, 113, 32 },
		{ 279, 114, 32 },
		{ 279, 115, 32 },
		{ 279, 116, 32 },
		{ 279, 117, 32 },
		{ 279, 118, 32 },
		{ 279, 119, 32 },
		{ 279, 120, 32 },
		{ 279, 121, 32 },
		{ 279, 122, 32 },
		{ 279, 123, 32 },
		{ 279, 124, 32 },
		{ 279, 125, 32 },
		{ 279, 126, 32 },
		{ 279, 127, 32 },
		{ 279, 128, 32 },
		{ 279, 129, 32 },
		{ 279, 130, 32 },
		{ 279, 131, 32 },
		{ 279, 132, 32 },
		{ 279, 133, 32 },
		{ 279, 134, 32 },
		{ 279, 135, 32 },
		{ 279, 136, 32 },
		{ 279, 137, 32 },
		{ 279, 138, 32 },
		{ 279, 139, 32 },
		{ 279, 140, 32 },
		{ 279, 141, 32 },
		{ 279, 142, 32 },
		{ 279, 143, 32 },
		{ 279, 144, 32 },
		{ 279, 145, 32 },
		{ 279, 146, 32 },
		{ 279, 147, 32 },
		{ 279, 148, 32 },
		{ 279, 149, 32 },
		{ 279, 150, 32 },
		{ 279, 151, 32 },
		{ 279, 152, 32 },
		{ 279, 153, 32 },
		{ 279, 154, 32 },
		{ 279, 155, 32 },
		{ 279, 156, 32 },
		{ 279, 157, 32 },
		{ 279, 158, 32 },
		{ 279, 159, 32 },
		{ 279, 160, 32 },
		{ 279, 161, 32 },
		{ 279, 162, 32 },
		{ 279, 163, 32 },
		{ 279, 164, 32 },
		{ 279, 165, 32 },
		{ 279, 166, 32 },
		{ 279, 167, 32 },
		{ 279, 168, 32 },
		{ 279, 169, 32 },
		{ 279, 170, 32 },
		{ 279, 171, 32 },
		{ 279, 172, 32 },
		{ 279, 173, 32 },
		{ 279, 174, 32 },
		{ 279, 175, 32 },
		{ 279, 176, 32 },
		{ 279, 177, 32 },
		{ 279, 178, 32 },
		{ 279, 179, 32 },
		{ 279, 180, 32 },
		{ 279, 181, 32 },
		{ 279, 182, 32 },
		{ 279, 183, 32 },
		{ 279, 184, 32 },
		{ 279, 185, 32 },
		{ 279, 186, 32 },
		{ 279, 187, 32 },
		{ 279, 188, 32 },
		{ 279, 189, 32 },
		{ 279, 190, 32 },
		{ 279, 191, 32 },
		{ 280, 74, 32 },
		{ 280, 75, 32 },
		{ 280, 76, 32 },
		{ 280, 77, 32 },
		{ 280, 78, 32 },
		{ 280, 79, 32 },
		{ 280, 80, 32 },
		{ 280, 81, 32 },
		{ 280, 82, 32 },
		{ 280, 83, 32 },
		{ 280, 84, 32 },
		{ 280, 85, 32 },
		{ 280, 86, 32 },
		{ 280, 87, 32 },
		{ 280, 88, 32 },
		{ 280, 89, 32 },
		{ 280, 90, 32 },
		{ 280, 91, 32 },
		{ 280, 92, 32 },
		{ 280, 93, 32 },
		{ 280, 94, 32 },
		{ 280, 95, 32 },
		{ 280, 96, 32 },
		{ 280, 97, 32 },
		{ 280, 98, 32 },
		{ 280, 99, 32 },
		{ 280, 100, 32 },
		{ 280, 101, 32 },
		{ 280, 102, 32 },
		{ 280, 103, 32 },
		{ 280, 104, 32 },
		{ 280, 105, 32 },
		{ 280, 106, 32 },
		{ 280, 107, 32 },
		{ 280, 108, 32 },
		{ 280, 109, 32 },
		{ 280, 110, 32 },
		{ 280, 111, 32 },
		{ 280, 112, 32 },
		{ 280, 113, 32 },
		{ 280, 114, 32 },
		{ 280, 115, 32 },
		{ 280, 116, 32 },
		{ 280, 117, 32 },
		{ 280, 118, 32 },
		{ 280, 119, 32 },
		{ 280, 120, 32 },
		{ 280, 121, 32 },
		{ 280, 122, 32 },
		{ 280, 123, 32 },
		{ 280, 124, 32 },
		{ 280, 125, 32 },
		{ 280, 126, 32 },
		{ 280, 127, 32 },
		{ 280, 128, 32 },
		{ 280, 129, 32 },
		{ 280, 130, 32 },
		{ 280, 131, 32 },
		{ 280, 132, 32 },
		{ 280, 133, 32 },
		{ 280, 134, 32 },
		{ 280, 135, 32 },
		{ 280, 136, 32 },
		{ 280, 137, 32 },
		{ 280, 138, 32 },
		{ 280, 139, 32 },
		{ 280, 140, 32 },
		{ 280, 141, 32 },
		{ 280, 142, 32 },
		{ 280, 143, 32 },
		{ 280, 144, 32 },
		{ 280, 145, 32 },
		{ 280, 146, 32 },
		{ 280, 147, 32 },
		{ 280, 148, 32 },
		{ 280, 149, 32 },
		{ 280, 150, 32 },
		{ 280, 151, 32 },
		{ 280, 152, 32 },
		{ 280, 153, 32 },
		{ 280, 154, 32 },
		{ 280, 155, 32 },
		{ 280, 156, 32 },
		{ 280, 157, 32 },
		{ 280, 158, 32 },
		{ 280, 159, 32 },
		{ 280, 160, 32 },
		{ 280, 161, 32 },
		{ 280, 162, 32 },
		{ 280, 163, 32 },
		{ 280, 164, 32 },
		{ 280, 165, 32 },
		{ 280, 166, 32 },
		{ 280, 167, 32 },
		{ 280, 168, 32 },
		{ 280, 169, 32 },
		{ 280, 170, 32 },
		{ 280, 171, 32 },
		{ 280, 172, 32 },
		{ 280, 173, 32 },
		{ 280, 174, 32 },
		{ 280, 175, 32 },
		{ 280, 176, 32 },
		{ 280, 177, 32 },
		{ 280, 178, 32 },
		{ 280, 179, 32 },
		{ 280, 180, 32 },
		{ 280, 181, 32 },
		{ 280, 182, 32 },
		{ 280, 183, 32 },
		{ 280, 184, 32 },
		{ 280, 185, 32 },
		{ 280, 186, 32 },
		{ 280, 187, 32 },
		{ 280, 188, 32 },
		{ 280, 189, 32 },
		{ 280, 190, 32 },
		{ 281, 66, 32 },
		{ 281, 67, 32 },
		{ 281, 68, 32 },
		{ 281, 69, 32 },
		{ 281, 70, 32 },
		{ 281, 71, 32 },
		{ 281, 72, 32 },
		{ 281, 73, 32 },
		{ 281, 74, 32 },
		{ 281, 75, 32 },
		{ 281, 76, 32 },
		{ 281, 77, 32 },
		{ 281, 78, 32 },
		{ 281, 79, 32 },
		{ 281, 80, 32 },
		{ 281, 81, 32 },
		{ 281, 82, 32 },
		{ 281, 83, 32 },
		{ 281, 84, 32 },
		{ 281, 85, 32 },
		{ 281, 86, 32 },
		{ 281, 87, 32 },
		{ 281, 88, 32 },
		{ 281, 89, 32 },
		{ 281, 90, 32 },
		{ 281, 91, 32 },
		{ 281, 92, 32 },
		{ 281, 93, 32 },
		{ 281, 94, 32 },
		{ 281, 95, 32 },
		{ 281, 96, 32 },
		{ 281, 97, 32 },
		{ 281, 98, 32 },
		{ 281, 99, 32 },
		{ 281, 100, 32 },
		{ 281, 101, 32 },
		{ 281, 102, 32 },
		{ 281, 103, 32 },
		{ 281, 104, 32 },
		{ 281, 105, 32 },
		{ 281, 106, 32 },
		{ 281, 107, 32 },
		{ 281, 108, 32 },
		{ 281, 109, 32 },
		{ 281, 110, 32 },
		{ 281, 111, 32 },
		{ 281, 112, 32 },
		{ 281, 113, 32 },
		{ 281, 114, 32 },
		{ 281, 115, 32 },
		{ 281, 116, 32 },
		{ 281, 117, 32 },
		{ 281, 118, 32 },
		{ 281, 119, 32 },
		{ 281, 120, 32 },
		{ 281, 121, 32 },
		{ 281, 122, 32 },
		{ 281, 123, 32 },
		{ 281, 124, 32 },
		{ 281, 125, 32 },
		{ 281, 126, 32 },
		{ 281, 127, 32 },
		{ 281, 128, 32 },
		{ 281, 129, 32 },
		{ 281, 130, 32 },
		{ 281, 131, 32 },
		{ 281, 132, 32 },
		{ 281, 133, 32 },
		{ 281, 134, 32 },
		{ 281, 135, 32 },
		{ 281, 136, 32 },
		{ 281, 137, 32 },
		{ 281, 138, 32 },
		{ 281, 139, 32 },
		{ 281, 140, 32 },
		{ 281, 141, 32 },
		{ 281, 142, 32 },
		{ 281, 143, 32 },
		{ 281, 144, 32 },
		{ 281, 145, 32 },
		{ 281, 146, 32 },
		{ 281, 147, 32 },
		{ 281, 148, 32 },
		{ 281, 149, 32 },
		{ 281, 150, 32 },
		{ 281, 151, 32 },
		{ 281, 152, 32 },
		{ 281, 153, 32 },
		{ 281, 154, 32 },
		{ 281, 155, 32 },
		{ 281, 156, 32 },
		{ 281, 157, 32 },
		{ 281, 158, 32 },
		{ 281, 159, 32 },
		{ 281, 160, 32 },
		{ 281, 161, 32 },
		{ 281, 162, 32 },
		{ 281, 163, 32 },
		{ 281, 164, 32 },
		{ 281, 165, 32 },
		{ 281, 166, 32 },
		{ 281, 167, 32 },
		{ 281, 168, 32 },
		{ 281, 169, 32 },
		{ 281, 170, 32 },
		{ 281, 171, 32 },
		{ 281, 172, 32 },
		{ 281, 173, 32 },
		{ 281, 174, 32 },
		{ 281, 175, 32 },
		{ 281, 176, 32 },
		{ 281, 177, 32 },
		{ 281, 178, 32 },
		{ 281, 179, 32 },
		{ 281, 180, 32 },
		{ 281, 181, 32 },
		{ 281, 182, 32 },
		{ 281, 183, 32 },
		{ 281, 184, 32 },
		{ 281, 185, 32 },
		{ 281, 186, 32 },
		{ 281, 187, 32 },
		{ 281, 188, 32 },
		{ 281, 189, 32 },
		{ 281, 190, 32 },
		{ 282, 58, 32 },
		{ 282, 59, 32 },
		{ 282, 60, 32 },
		{ 282, 61, 32 },
		{ 282, 62, 32 },
		{ 282, 63, 32 },
		{ 282, 64, 32 },
		{ 282, 65, 32 },
		{ 282, 66, 32 },
		{ 282, 67, 32 },
		{ 282, 68, 32 },
		{ 282, 69, 32 },
		{ 282, 70, 32 },
		{ 282, 71, 32 },
		{ 282, 72, 32 },
		{ 282, 73, 32 },
		{ 282, 74, 32 },
		{ 282, 75, 32 },
		{ 282, 76, 32 },
		{ 282, 77, 32 },
		{ 282, 78, 32 },
		{ 282, 79, 32 },
		{ 282, 80, 32 },
		{ 282, 81, 32 },
		{ 282, 82, 32 },
		{ 282, 83, 32 },
		{ 282, 84, 32 },
		{ 282, 85, 32 },
		{ 282, 86, 32 },
		{ 282, 87, 32 },
		{ 282, 88, 32 },
		{ 282, 89, 32 },
		{ 282, 90, 32 },
		{ 282, 91, 32 },
		{ 282, 92, 32 },
		{ 282, 93, 32 },
		{ 282, 94, 32 },
		{ 282, 95, 32 },
		{ 282, 96, 32 },
		{ 282, 97, 32 },
		{ 282, 98, 32 },
		{ 282, 99, 32 },
		{ 282, 100, 32 },
		{ 282, 101, 32 },
		{ 282, 102, 32 },
		{ 282, 103, 32 },
		{ 282, 104, 32 },
		{ 282, 105, 32 },
		{ 282, 106, 32 },
		{ 282, 107, 32 },
		{ 282, 108, 32 },
		{ 282, 109, 32 },
		{ 282, 110, 32 },
		{ 282, 111, 32 },
		{ 282, 112, 32 },
		{ 282, 113, 32 },
		{ 282, 114, 32 },
		{ 282, 115, 32 },
		{ 282, 116, 32 },
		{ 282, 117, 32 },
		{ 282, 118, 32 },
		{ 282, 119, 32 },
		{ 282, 120, 32 },
		{ 282, 121, 32 },
		{ 282, 122, 32 },
		{ 282, 123, 32 },
		{ 282, 124, 32 },
		{ 282, 125, 32 },
		{ 282, 126, 32 },
		{ 282, 127, 32 },
		{ 282, 128, 32 },
		{ 282, 129, 32 },
		{ 282, 130, 32 },
		{ 282, 131, 32 },
		{ 282, 132, 32 },
		{ 282, 133, 32 },
		{ 282, 134, 32 },
		{ 282, 135, 32 },
		{ 282, 136, 32 },
		{ 282, 137, 32 },
		{ 282, 138, 32 },
		{ 282, 139, 32 },
		{ 282, 140, 32 },
		{ 282, 141, 32 },
		{ 282, 142, 32 },
		{ 282, 143, 32 },
		{ 282, 144, 32 },
		{ 282, 145, 32 },
		{ 282, 146, 32 },
		{ 282, 147, 32 },
		{ 282, 148, 32 },
		{ 282, 149, 32 },
		{ 282, 150, 32 },
		{ 282, 151, 32 },
		{ 282, 152, 32 },
		{ 282, 153, 32 },
		{ 282, 154, 32 },
		{ 282, 155, 32 },
		{ 282, 156, 32 },
		{ 282, 157, 32 },
		{ 282, 158, 32 },
		{ 282, 159, 32 },
		{ 282, 160, 32 },
		{ 282, 161, 32 },
		{ 282, 162, 32 },
		{ 282, 163, 32 },
		{ 282, 164, 32 },
		{ 282, 165, 32 },
		{ 282, 166, 32 },
		{ 282, 167, 32 },
		{ 282, 168, 32 },
		{ 282, 169, 32 },
		{ 282, 170, 32 },
		{ 282, 171, 32 },
		{ 282, 172, 32 },
		{ 282, 173, 32 },
		{ 282, 174, 32 },
		{ 282, 175, 32 },
		{ 282, 176, 32 },
		{ 282, 177, 32 },
		{ 282, 178, 32 },
		{ 282, 179, 32 },
		{ 282, 180, 32 },
		{ 282, 181, 32 },
		{ 282, 182, 32 },
		{ 282, 183, 32 },
		{ 282, 184, 32 },
		{ 282, 185, 32 },
		{ 282, 186, 32 },
		{ 282, 187, 32 },
		{ 282, 188, 32 },
		{ 282, 189, 32 },
		{ 283, 50, 32 },
		{ 283, 51, 32 },
		{ 283, 52, 32 },
		{ 283, 53, 32 },
		{ 283, 54, 32 },
		{ 283, 55, 32 },
		{ 283, 56, 32 },
		{ 283, 57, 32 },
		{ 283, 58, 32 },
		{ 283, 59, 32 },
		{ 283, 60, 32 },
		{ 283, 61, 32 },
		{ 283, 62, 32 },
		{ 283, 63, 32 },
		{ 283, 64, 32 },
		{ 283, 65, 32 },
		{ 283, 66, 32 },
		{ 283, 67, 32 },
		{ 283, 68, 32 },
		{ 283, 69, 32 },
		{ 283, 70, 32 },
		{ 283, 71, 32 },
		{ 283, 72, 32 },
		{ 283, 73, 32 },
		{ 283, 74, 32 },
		{ 283, 75, 32 },
		{ 283, 76, 32 },
		{ 283, 77, 32 },
		{ 283, 78, 32 },
		{ 283, 79, 32 },
		{ 283, 80, 32 },
		{ 283, 81, 32 },
		{ 283, 82, 32 },
		{ 283, 83, 32 },
		{ 283, 84, 32 },
		{ 283, 85, 32 },
		{ 283, 86, 32 },
		{ 283, 87, 32 },
		{ 283, 88, 32 },
		{ 283, 89, 32 },
		{ 283, 90, 32 },
		{ 283, 91, 32 },
		{ 283, 92, 32 },
		{ 283, 93, 32 },
		{ 283, 94, 32 },
		{ 283, 95, 32 },
		{ 283, 96, 32 },
		{ 283, 97, 32 },
		{ 283, 98, 32 },
		{ 283, 99, 32 },
		{ 283, 100, 32 },
		{ 283, 101, 32 },
		{ 283, 102, 32 },
		{ 283, 103, 32 },
		{ 283, 104, 32 },
		{ 283, 105, 32 },
		{ 283, 106, 32 },
		{ 283, 107, 32 },
		{ 283, 108, 32 },
		{ 283, 109, 32 },
		{ 283, 110, 32 },
		{ 283, 111, 32 },
		{ 283, 112, 32 },
		{ 283, 113, 32 },
		{ 283, 114, 32 },
		{ 283, 115, 32 },
		{ 283, 116, 32 },
		{ 283, 117, 32 },
		{ 283, 118, 32 },
		{ 283, 119, 32 },
		{ 283, 120, 32 },
		{ 283, 121, 32 },
		{ 283, 122, 32 },
		{ 283, 123, 32 },
		{ 283, 124, 32 },
		{ 283, 125, 32 },
		{ 283, 126, 32 },
		{ 283, 127, 32 },
		{ 283, 128, 32 },
		{ 283, 129, 32 },
		{ 283, 130, 32 },
		{ 283, 131, 32 },
		{ 283, 132, 32 },
		{ 283, 133, 32 },
		{ 283, 134, 32 },
		{ 283, 135, 32 },
		{ 283, 136, 32 },
		{ 283, 137, 32 },
		{ 283, 138, 32 },
		{ 283, 139, 32 },
		{ 283, 140, 32 },
		{ 283, 141, 32 },
		{ 283, 142, 32 },
		{ 283, 143, 32 },
		{ 283, 144, 32 },
		{ 283, 145, 32 },
		{ 283, 146, 32 },
		{ 283, 147, 32 },
		{ 283, 148, 32 },
		{ 283, 149, 32 },
		{ 283, 150, 32 },
		{ 283, 151, 32 },
		{ 283, 152, 32 },
		{ 283, 153, 32 },
		{ 283, 154, 32 },
		{ 283, 155, 32 },
		{ 283, 156, 32 },
		{ 283, 157, 32 },
		{ 283, 158, 32 },
		{ 283, 159, 32 },
		{ 283, 160, 32 },
		{ 283, 161, 32 },
		{ 283, 162, 32 },
		{ 283, 163, 32 },
		{ 283, 164, 32 },
		{ 283, 165, 32 },
		{ 283, 166, 32 },
		{ 283, 167, 32 },
		{ 283, 168, 32 },
		{ 283, 169, 32 },
		{ 283, 170, 32 },
		{ 283, 171, 32 },
		{ 283, 172, 32 },
		{ 283, 173, 32 },
		{ 283, 174, 32 },
		{ 283, 175, 32 },
		{ 283, 176, 32 },
		{ 283, 177, 32 },
		{ 283, 178, 32 },
		{ 283, 179, 32 },
		{ 283, 180, 32 },
		{ 283, 181, 32 },
		{ 283, 182, 32 },
		{ 283, 183, 32 },
		{ 283, 184, 32 },
		{ 283, 185, 32 },
		{ 283, 186, 32 },
		{ 283, 187, 32 },
		{ 284, 42, 32 },
		{ 284, 43, 32 },
		{ 284, 44, 32 },
		{ 284, 45, 32 },
		{ 284, 46, 32 },
		{ 284, 47, 32 },
		{ 284, 48, 32 },
		{ 284, 49, 32 },
		{ 284, 50, 32 },
		{ 284, 51, 32 },
		{ 284, 52, 32 },
		{ 284, 53, 32 },
		{ 284, 54, 32 },
		{ 284, 55, 32 },
		{ 284, 56, 32 },
		{ 284, 57, 32 },
		{ 284, 58, 32 },
		{ 284, 59, 32 },
		{ 284, 60, 32 },
		{ 284, 61, 32 },
		{ 284, 62, 32 },
		{ 284, 63, 32 },
		{ 284, 64, 32 },
		{ 284, 65, 32 },
		{ 284, 66, 32 },
		{ 284, 67, 32 },
		{ 284, 68, 32 },
		{ 284, 69, 32 },
		{ 284, 70, 32 },
		{ 284, 71, 32 },
		{ 284, 72, 32 },
		{ 284, 73, 32 },
		{ 284, 74, 32 },
		{ 284, 75, 32 },
		{ 284, 76, 32 },
		{ 284, 77, 32 },
		{ 284, 78, 32 },
		{ 284, 79, 32 },
		{ 284, 80, 32 },
		{ 284, 81, 32 },
		{ 284, 82, 32 },
		{ 284, 83, 32 },
		{ 284, 84, 32 },
		{ 284, 85, 32 },
		{ 284, 86, 32 },
		{ 284, 87, 32 },
		{ 284, 88, 32 },
		{ 284, 89, 32 },
		{ 284, 90, 32 },
		{ 284, 91, 32 },
		{ 284, 92, 32 },
		{ 284, 93, 32 },
		{ 284, 94, 32 },
		{ 284, 95, 32 },
		{ 284, 96, 32 },
		{ 284, 97, 32 },
		{ 284, 98, 32 },
		{ 284, 99, 32 },
		{ 284, 100, 32 },
		{ 284, 101, 32 },
		{ 284, 102, 32 },
		{ 284, 103, 32 },
		{ 284, 104, 32 },
		{ 284, 105, 32 },
		{ 284, 106, 32 },
		{ 284, 107, 32 },
		{ 284, 108, 32 },
		{ 284, 109, 32 },
		{ 284, 110, 32 },
		{ 284, 111, 32 },
		{ 284, 112, 32 },
		{ 284, 113, 32 },
		{ 284, 114, 32 },
		{ 284, 115, 32 },
		{ 284, 116, 32 },
		{ 284, 117, 32 },
		{ 284, 118, 32 },
		{ 284, 119, 32 },
		{ 284, 120, 32 },
		{ 284, 121, 32 },
		{ 284, 122, 32 },
		{ 284, 123, 32 },
		{ 284, 124, 32 },
		{ 284, 125, 32 },
		{ 284, 126, 32 },
		{ 284, 127, 32 },
		{ 284, 128, 32 },
		{ 284, 129, 32 },
		{ 284, 130, 32 },
		{ 284, 131, 32 },
		{ 284, 132, 32 },
		{ 284, 133, 32 },
		{ 284, 134, 32 },
		{ 284, 135, 32 },
		{ 284, 136, 32 },
		{ 284, 137, 32 },
		{ 284, 138, 32 },
		{ 284, 139, 32 },
		{ 284, 140, 32 },
		{ 284, 141, 32 },
		{ 284, 142, 32 },
		{ 284, 143, 32 },
		{ 284, 144, 32 },
		{ 284, 145, 32 },
		{ 284, 146, 32 },
		{ 284, 147, 32 },
		{ 284, 148, 32 },
		{ 284, 149, 32 },
		{ 284, 150, 32 },
		{ 284, 151, 32 },
		{ 284, 152, 32 },
		{ 284, 153, 32 },
		{ 284, 154, 32 },
		{ 284, 155, 32 },
		{ 284, 156, 32 },
		{ 284, 157, 32 },
		{ 284, 158, 32 },
		{ 284, 159, 32 },
		{ 284, 160, 32 },
		{ 284, 161, 32 },
		{ 284, 162, 32 },
		{ 284, 163, 32 },
		{ 284, 164, 32 },
		{ 284, 165, 32 },
		{ 284, 166, 32 },
		{ 284, 167, 32 },
		{ 284, 168, 32 },
		{ 284, 169, 32 },
		{ 284, 170, 32 },
		{ 284, 171, 32 },
		{ 284, 172, 32 },
		{ 284, 173, 32 },
		{ 284, 174, 32 },
		{ 284, 175, 32 },
		{ 284, 176, 32 },
		{ 284, 177, 32 },
		{ 284, 178, 32 },
		{ 284, 179, 32 },
		{ 284, 180, 32 },
		{ 284, 181, 32 },
		{ 284, 182, 32 },
		{ 284, 183, 32 },
		{ 284, 184, 32 },
		{ 285, 34, 32 },
		{ 285, 35, 32 },
		{ 285, 36, 32 },
		{ 285, 37, 32 },
		{ 285, 38, 32 },
		{ 285, 39, 32 },
		{ 285, 40, 32 },
		{ 285, 41, 32 },
		{ 285, 42, 32 },
		{ 285, 43, 32 },
		{ 285, 44, 32 },
		{ 285, 45, 32 },
		{ 285, 46, 32 },
		{ 285, 47, 32 },
		{ 285, 48, 32 },
		{ 285, 49, 32 },
		{ 285, 50, 32 },
		{ 285, 51, 32 },
		{ 285, 52, 32 },
		{ 285, 53, 32 },
		{ 285, 54, 32 },
		{ 285, 55, 32 },
		{ 285, 56, 32 },
		{ 285, 57, 32 },
		{ 285, 58, 32 },
		{ 285, 59, 32 },
		{ 285, 60, 32 },
		{ 285, 61, 32 },
		{ 285, 62, 32 },
		{ 285, 63, 32 },
		{ 285, 64, 32 },
		{ 285, 65, 32 },
		{ 285, 66, 32 },
		{ 285, 67, 32 },
		{ 285, 68, 32 },
		{ 285, 69, 32 },
		{ 285, 70, 32 },
		{ 285, 71, 32 },
		{ 285, 72, 32 },
		{ 285, 73, 32 },
		{ 285, 74, 32 },
		{ 285, 75, 32 },
		{ 285, 76, 32 },
		{ 285, 77, 32 },
		{ 285, 78, 32 },
		{ 285, 79, 32 },
		{ 285, 80, 32 },
		{ 285, 81, 32 },
		{ 285, 82, 32 },
		{ 285, 83, 32 },
		{ 285, 84, 32 },
		{ 285, 85, 32 },
		{ 285, 86, 32 },
		{ 285, 87, 32 },
		{ 285, 88, 32 },
		{ 285, 89, 32 },
		{ 285, 90, 32 },
		{ 285, 91, 32 },
		{ 285, 92, 32 },
		{ 285, 93, 32 },
		{ 285, 94, 32 },
		{ 285, 95, 32 },
		{ 285, 96, 32 },
		{ 285, 97, 32 },
		{ 285, 98, 32 },
		{ 285, 99, 32 },
		{ 285, 100, 32 },
		{ 285, 101, 32 },
		{ 285, 102, 32 },
		{ 285, 103, 32 },
		{ 285, 104, 32 },
		{ 285, 105, 32 },
		{ 285, 106, 32 },
		{ 285, 107, 32 },
		{ 285, 108, 32 },
		{ 285, 109, 32 },
		{ 285, 110, 32 },
		{ 285, 111, 32 },
		{ 285, 112, 32 },
		{ 285, 113, 32 },
		{ 285, 114, 32 },
		{ 285, 115, 32 },
		{ 285, 116, 32 },
		{ 285, 117, 32 },
		{ 285, 118, 32 },
		{ 285, 119, 32 },
		{ 285, 120, 32 },
		{ 285, 121, 32 },
		{ 285, 122, 32 },
		{ 285, 123, 32 },
		{ 285, 124, 32 },
		{ 285, 125, 32 },
		{ 285, 126, 32 },
		{ 285, 127, 32 },
		{ 285, 128, 32 },
		{ 285, 129, 32 },
		{ 285, 130, 32 },
		{ 285, 131, 32 },
		{ 285, 132, 32 },
		{ 285, 133, 32 },
		{ 285, 134, 32 },
		{ 285, 135, 32 },
		{ 285, 136, 32 },
		{ 285, 137, 32 },
		{ 285, 138, 32 },
		{ 285, 139, 32 },
		{ 285, 140, 32 },
		{ 285, 141, 32 },
		{ 285, 142, 32 },
		{ 285, 143, 32 },
		{ 285, 144, 32 },
		{ 285, 145, 32 },
		{ 285, 146, 32 },
		{ 285, 147, 32 },
		{ 285, 148, 32 },
		{ 285, 149, 32 },
		{ 285, 150, 32 },
		{ 285, 151, 32 },
		{ 285, 152, 32 },
		{ 285, 153, 32 },
		{ 285, 154, 32 },
		{ 285, 155, 32 },
		{ 285, 156, 32 },
		{ 285, 157, 32 },
		{ 285, 158, 32 },
		{ 285, 159, 32 },
		{ 285, 160, 32 },
		{ 285, 161, 32 },
		{ 285, 162, 32 },
		{ 285, 163, 32 },
		{ 285, 164, 32 },
		{ 285, 165, 32 },
		{ 285, 166, 32 },
		{ 285, 167, 32 },
		{ 285, 168, 32 },
		{ 285, 169, 32 },
		{ 285, 170, 32 },
		{ 285, 171, 32 },
		{ 285, 172, 32 },
		{ 285, 173, 32 },
		{ 285, 174, 32 },
		{ 285, 175, 32 },
		{ 285, 176, 32 },
		{ 285, 177, 32 },
		{ 285, 178, 32 },
		{ 285, 179, 32 },
		{ 285, 180, 32 },
		{ 285, 181, 32 },
		{ 285, 182, 32 },
		{ 286, 26, 32 },
		{ 286, 27, 32 },
		{ 286, 28, 32 },
		{ 286, 29, 32 },
		{ 286, 30, 32 },
		{ 286, 31, 32 },
		{ 286, 32, 32 },
		{ 286, 33, 32 },
		{ 286, 34, 32 },
		{ 286, 35, 32 },
		{ 286, 36, 32 },
		{ 286, 37, 32 },
		{ 286, 38, 32 },
		{ 286, 39, 32 },
		{ 286, 40, 32 },
		{ 286, 41, 32 },
		{ 286, 42, 32 },
		{ 286, 43, 32 },
		{ 286, 44, 32 },
		{ 286, 45, 32 },
		{ 286, 46, 32 },
		{ 286, 47, 32 },
		{ 286, 48, 32 },
		{ 286, 49, 32 },
		{ 286, 50, 32 },
		{ 286, 51, 32 },
		{ 286, 52, 32 },
		{ 286, 53, 32 },
		{ 286, 54, 32 },
		{ 286, 55, 32 },
		{ 286, 56, 32 },
		{ 286, 57, 32 },
		{ 286, 58, 32 },
		{ 286, 59, 32 },
		{ 286, 60, 32 },
		{ 286, 61, 32 },
		{ 286, 62, 32 },
		{ 286, 63, 32 },
		{ 286, 64, 32 },
		{ 286, 65, 32 },
		{ 286, 66, 32 },
		{ 286, 67, 32 },
		{ 286, 68, 32 },
		{ 286, 69, 32 },
		{ 286, 70, 32 },
		{ 286, 71, 32 },
		{ 286, 72, 32 },
		{ 286, 73, 32 },
		{ 286, 74, 32 },
		{ 286, 75, 32 },
		{ 286, 76, 32 },
		{ 286, 77, 32 },
		{ 286, 78, 32 },
		{ 286, 79, 32 },
		{ 286, 80, 32 },
		{ 286, 81, 32 },
		{ 286, 82, 32 },
		{ 286, 83, 32 },
		{ 286, 84, 32 },
		{ 286, 85, 32 },
		{ 286, 86, 32 },
		{ 286, 87, 32 },
		{ 286, 88, 32 },
		{ 286, 89, 32 },
		{ 286, 90, 32 },
		{ 286, 91, 32 },
		{ 286, 92, 32 },
		{ 286, 93, 32 },
		{ 286, 94, 32 },
		{ 286, 95, 32 },
		{ 286, 96, 32 },
		{ 286, 97, 32 },
		{ 286, 98, 32 },
		{ 286, 99, 32 },
		{ 286, 100, 32 },
		{ 286, 101, 32 },
		{ 286, 102, 32 },
		{ 286, 103, 32 },
		{ 286, 104, 32 },
		{ 286, 105, 32 },
		{ 286, 106, 32 },
		{ 286, 107, 32 },
		{ 286, 108, 32 },
		{ 286, 109, 32 },
		{ 286, 110, 32 },
		{ 286, 111, 32 },
		{ 286, 112, 32 },
		{ 286, 113, 32 },
		{ 286, 114, 32 },
		{ 286, 115, 32 },
		{ 286, 116, 32 },
		{ 286, 117, 32 },
		{ 286, 118, 32 },
		{ 286, 119, 32 },
		{ 286, 120, 32 },
		{ 286, 121, 32 },
		{ 286, 122, 32 },
		{ 286, 123, 32 },
		{ 286, 124, 32 },
		{ 286, 125, 32 },
		{ 286, 126, 32 },
		{ 286, 127, 32 },
		{ 286, 128, 32 },
		{ 286, 129, 32 },
		{ 286, 130, 32 },
		{ 286, 131, 32 },
		{ 286, 132, 32 },
		{ 286, 133, 32 },
		{ 286, 134, 32 },
		{ 286, 135, 32 },
		{ 286, 136, 32 },
		{ 286, 137, 32 },
		{ 286, 138, 32 },
		{ 286, 139, 32 },
		{ 286, 140, 32 },
		{ 286, 141, 32 },
		{ 286, 142, 32 },
		{ 286, 143, 32 },
		{ 286, 144, 32 },
		{ 286, 145, 32 },
		{ 286, 146, 32 },
		{ 286, 147, 32 },
		{ 286, 148, 32 },
		{ 286, 149, 32 },
		{ 286, 150, 32 },
		{ 286, 151, 32 },
		{ 286, 152, 32 },
		{ 286, 153, 32 },
		{ 286, 154, 32 },
		{ 286, 155, 32 },
		{ 286, 156, 32 },
		{ 286, 157, 32 },
		{ 286, 158, 32 },
		{ 286, 159, 32 },
		{ 286, 160, 32 },
		{ 286, 161, 32 },
		{ 286, 162, 32 },
		{ 286, 163, 32 },
		{ 286, 164, 32 },
		{ 286, 165, 32 },
		{ 286, 166, 32 },
		{ 286, 167, 32 },
		{ 286, 168, 32 },
		{ 286, 169, 32 },
		{ 286, 170, 32 },
		{ 286, 171, 32 },
		{ 286, 172, 32 },
		{ 286, 173, 32 },
		{ 286, 174, 32 },
		{ 286, 175, 32 },
		{ 286, 176, 32 },
		{ 286, 177, 32 },
		{ 286, 178, 32 },
		{ 286, 179, 32 },
		{ 287, 18, 32 },
		{ 287, 19, 32 },
		{ 287, 20, 32 },
		{ 287, 21, 32 },
		{ 287, 22, 32 },
		{ 287, 23, 32 },
		{ 287, 24, 32 },
		{ 287, 25, 32 },
		{ 287, 26, 32 },
		{ 287, 27, 32 },
		{ 287, 28, 32 },
		{ 287, 29, 32 },
		{ 287, 30, 32 },
		{ 287, 31, 32 },
		{ 287, 32, 32 },
		{ 287, 33, 32 },
		{ 287, 34, 32 },
		{ 287, 35, 32 },
		{ 287, 36, 32 },
		{ 287, 37, 32 },
		{ 287, 38, 32 },
		{ 287, 39, 32 },
		{ 287, 40, 32 },
		{ 287, 41, 32 },
		{ 287, 42, 32 },
		{ 287, 43, 32 },
		{ 287, 44, 32 },
		{ 287, 45, 32 },
		{ 287, 46, 32 },
		{ 287, 47, 32 },
		{ 287, 48, 32 },
		{ 287, 49, 32 },
		{ 287, 50, 32 },
		{ 287, 51, 32 },
		{ 287, 52, 32 },
		{ 287, 53, 32 },
		{ 287, 54, 32 },
		{ 287, 55, 32 },
		{ 287, 56, 32 },
		{ 287, 57, 32 },
		{ 287, 58, 32 },
		{ 287, 59, 32 },
		{ 287, 60, 32 },
		{ 287, 61, 32 },
		{ 287, 62, 32 },
		{ 287, 63, 32 },
		{ 287, 64, 32 },
		{ 287, 65, 32 },
		{ 287, 66, 32 },
		{ 287, 67, 32 },
		{ 287, 68, 32 },
		{ 287, 69, 32 },
		{ 287, 70, 32 },
		{ 287, 71, 32 },
		{ 287, 72, 32 },
		{ 287, 73, 32 },
		{ 287, 74, 32 },
		{ 287, 75, 32 },
		{ 287, 76, 32 },
		{ 287, 77, 32 },
		{ 287, 78, 32 },
		{ 287, 79, 32 },
		{ 287, 80, 32 },
		{ 287, 81, 32 },
		{ 287, 82, 32 },
		{ 287, 83, 32 },
		{ 287, 84, 32 },
		{ 287, 85, 32 },
		{ 287, 86, 32 },
		{ 287, 87, 32 },
		{ 287, 88, 32 },
		{ 287, 89, 32 },
		{ 287, 90, 32 },
		{ 287, 91, 32 },
		{ 287, 92, 32 },
		{ 287, 93, 32 },
		{ 287, 94, 32 },
		{ 287, 95, 32 },
		{ 287, 96, 32 },
		{ 287, 97, 32 },
		{ 287, 98, 32 },
		{ 287, 99, 32 },
		{ 287, 100, 32 },
		{ 287, 101, 32 },
		{ 287, 102, 32 },
		{ 287, 103, 32 },
		{ 287, 104, 32 },
		{ 287, 105, 32 },
		{ 287, 106, 32 },
		{ 287, 107, 32 },
		{ 287, 108, 32 },
		{ 287, 109, 32 },
		{ 287, 110, 32 },
		{ 287, 111, 32 },
		{ 287, 112, 32 },
		{ 287, 113, 32 },
		{ 287, 114, 32 },
		{ 287, 115, 32 },
		{ 287, 116, 32 },
		{ 287, 117, 32 },
		{ 287, 118, 32 },
		{ 287, 119, 32 },
		{ 287, 120, 32 },
		{ 287, 121, 32 },
		{ 287, 122, 32 },
		{ 287, 123, 32 },
		{ 287, 124, 32 },
		{ 287, 125, 32 },
		{ 287, 126, 32 },
		{ 287, 127, 32 },
		{ 287, 128, 32 },
		{ 287, 129, 32 },
		{ 287, 130, 32 },
		{ 287, 131, 32 },
		{ 287, 132, 32 },
		{ 287, 133, 32 },
		{ 287, 134, 32 },
		{ 287, 135, 32 },
		{ 287, 136, 32 },
		{ 287, 137, 32 },
		{ 287, 138, 32 },
		{ 287, 139, 32 },
		{ 287, 140, 32 },
		{ 287, 141, 32 },
		{ 287, 142, 32 },
		{ 287, 143, 32 },
		{ 287, 144, 32 },
		{ 287, 145, 32 },
		{ 287, 146, 32 },
		{ 287, 147, 32 },
		{ 287, 148, 32 },
		{ 287, 149, 32 },
		{ 287, 150, 32 },
		{ 287, 151, 32 },
		{ 287, 152, 32 },
		{ 287, 153, 32 },
		{ 287, 154, 32 },
		{ 287, 155, 32 },
		{ 287, 156, 32 },
		{ 287, 157, 32 },
		{ 287, 158, 32 },
		{ 287, 159, 32 },
		{ 287, 160, 32 },
		{ 287, 161, 32 },
		{ 287, 162, 32 },
		{ 287, 163, 32 },
		{ 287, 164, 32 },
		{ 287, 165, 32 },
		{ 287, 166, 32 },
		{ 287, 167, 32 },
		{ 287, 168, 32 },
		{ 287, 169, 32 },
		{ 287, 170, 32 },
		{ 287, 171, 32 },
		{ 287, 172, 32 },
		{ 287, 173, 32 },
		{ 287, 174, 32 },
		{ 287, 175, 32 },
		{ 287, 176, 32 },
		{ 288, 10, 32 },
		{ 288, 11, 32 },
		{ 288, 12, 32 },
		{ 288, 13, 32 },
		{ 288, 14, 32 },
		{ 288, 15, 32 },
		{ 288, 16, 32 },
		{ 288, 17, 32 },
		{ 288, 18, 32 },
		{ 288, 19, 32 },
		{ 288, 20, 32 },
		{ 288, 21, 32 },
		{ 288, 22, 32 },
		{ 288, 23, 32 },
		{ 288, 24, 32 },
		{ 288, 25, 32 },
		{ 288, 26, 32 },
		{ 288, 27, 32 },
		{ 288, 28, 32 },
		{ 288, 29, 32 },
		{ 288, 30, 32 },
		{ 288, 31, 32 },
		{ 288, 32, 32 },
		{ 288, 33, 32 },
		{ 288, 34, 32 },
		{ 288, 35, 32 },
		{ 288, 36, 32 },
		{ 288, 37, 32 },
		{ 288, 38, 32 },
		{ 288, 39, 32 },
		{ 288, 40, 32 },
		{ 288, 41, 32 },
		{ 288, 42, 32 },
		{ 288, 43, 32 },
		{ 288, 44, 32 },
		{ 288, 45, 32 },
		{ 288, 46, 32 },
		{ 288, 47, 32 },
		{ 288, 48, 32 },
		{ 288, 49, 32 },
		{ 288, 50, 32 },
		{ 288, 51, 32 },
		{ 288, 52, 32 },
		{ 288, 53, 32 },
		{ 288, 54, 32 },
		{ 288, 55, 32 },
		{ 288, 56, 32 },
		{ 288, 57, 32 },
		{ 288, 58, 32 },
		{ 288, 59, 32 },
		{ 288, 60, 32 },
		{ 288, 61, 32 },
		{ 288, 62, 32 },
		{ 288, 63, 32 },
		{ 288, 64, 32 },
		{ 288, 65, 32 },
		{ 288, 66, 32 },
		{ 288, 67, 32 },
		{ 288, 68, 32 },
		{ 288, 69, 32 },
		{ 288, 70, 32 },
		{ 288, 71, 32 },
		{ 288, 72, 32 },
		{ 288, 73, 32 },
		{ 288, 74, 32 },
		{ 288, 75, 32 },
		{ 288, 76, 32 },
		{ 288, 77, 32 },
		{ 288, 78, 32 },
		{ 288, 79, 32 },
		{ 288, 80, 32 },
		{ 288, 81, 32 },
		{ 288, 82, 32 },
		{ 288, 83, 32 },
		{ 288, 84, 32 },
		{ 288, 85, 32 },
		{ 288, 86, 32 },
		{ 288, 87, 32 },
		{ 288, 88, 32 },
		{ 288, 89, 32 },
		{ 288, 90, 32 },
		{ 288, 91, 32 },
		{ 288, 92, 32 },
		{ 288, 93, 32 },
		{ 288, 94, 32 },
		{ 288, 95, 32 },
		{ 288, 96, 32 },
		{ 288, 97, 32 },
		{ 288, 98, 32 },
		{ 288, 99, 32 },
		{ 288, 100, 32 },
		{ 288, 101, 32 },
		{ 288, 102, 32 },
		{ 288, 103, 32 },
		{ 288, 104, 32 },
		{ 288, 105, 32 },
		{ 288, 106, 32 },
		{ 288, 107, 32 },
		{ 288, 108, 32 },
		{ 288, 109, 32 },
		{ 288, 110, 32 },
		{ 288, 111, 32 },
		{ 288, 112, 32 },
		{ 288, 113, 32 },
		{ 288, 114, 32 },
		{ 288, 115, 32 },
		{ 288, 116, 32 },
		{ 288, 117, 32 },
		{ 288, 118, 32 },
		{ 288, 119, 32 },
		{ 288, 120, 32 },
		{ 288, 121, 32 },
		{ 288, 122, 32 },
		{ 288, 123, 32 },
		{ 288, 124, 32 },
		{ 288, 125, 32 },
		{ 288, 126, 32 },
		{ 288, 127, 32 },
		{ 288, 128, 32 },
		{ 288, 129, 32 },
		{ 288, 130, 32 },
		{ 288, 131, 32 },
		{ 288, 132, 32 },
		{ 288, 133, 32 },
		{ 288, 134, 32 },
		{ 288, 135, 32 },
		{ 288, 136, 32 },
		{ 288, 137, 32 },
		{ 288, 138, 32 },
		{ 288, 139, 32 },
		{ 288, 140, 32 },
		{ 288, 141, 32 },
		{ 288, 142, 32 },
		{ 288, 143, 32 },
		{ 288, 144, 32 },
		{ 288, 145, 32 },
		{ 288, 146, 32 },
		{ 288, 147, 32 },
		{ 288, 148, 32 },
		{ 288, 149, 32 },
		{ 288, 150, 32 },
		{ 288, 151, 32 },
		{ 288, 152, 32 },
		{ 288, 153, 32 },
		{ 288, 154, 32 },
		{ 288, 155, 32 },
		{ 288, 156, 32 },
		{ 288, 157, 32 },
		{ 288, 158, 32 },
		{ 288, 159, 32 },
		{ 288, 160, 32 },
		{ 288, 161, 32 },
		{ 288, 162, 32 },
		{ 288, 163, 32 },
		{ 288, 164, 32 },
		{ 288, 165, 32 },
		{ 288, 166, 32 },
		{ 288, 167, 32 },
		{ 288, 168, 32 },
		{ 288, 169, 32 },
		{ 288, 170, 32 },
		{ 288, 171, 32 },
		{ 288, 172, 32 },
		{ 288, 173, 32 },
		{ 288, 174, 32 },
		{ 289, 2, 32 },
		{ 289, 3, 32 },
		{ 289, 4, 32 },
		{ 289, 5, 32 },
		{ 289, 6, 32 },
		{ 289, 7, 32 },
		{ 289, 8, 32 },
		{ 289, 9, 32 },
		{ 289, 10, 32 },
		{ 289, 11, 32 },
		{ 289, 12, 32 },
		{ 289, 13, 32 },
		{ 289, 14, 32 },
		{ 289, 15, 32 },
		{ 289, 16, 32 },
		{ 289, 17, 32 },
		{ 289, 18, 32 },
		{ 289, 19, 32 },
		{ 289, 20, 32 },
		{ 289, 21, 32 },
		{ 289, 22, 32 },
		{ 289, 23, 32 },
		{ 289, 24, 32 },
		{ 289, 25, 32 },
		{ 289, 26, 32 },
		{ 289, 27, 32 },
		{ 289, 28, 32 },
		{ 289, 29, 32 },
		{ 289, 30, 32 },
		{ 289, 31, 32 },
		{ 289, 32, 32 },
		{ 289, 33, 32 },
		{ 289, 34, 32 },
		{ 289, 35, 32 },
		{ 289, 36, 32 },
		{ 289, 37, 32 },
		{ 289, 38, 32 },
		{ 289, 39, 32 },
		{ 289, 40, 32 },
		{ 289, 41, 32 },
		{ 289, 42, 32 },
		{ 289, 43, 32 },
		{ 289, 44, 32 },
		{ 289, 45, 32 },
		{ 289, 46, 32 },
		{ 289, 47, 32 },
		{ 289, 48, 32 },
		{ 289, 49, 32 },
		{ 289, 50, 32 },
		{ 289, 51, 32 },
		{ 289, 52, 32 },
		{ 289, 53, 32 },
		{ 289, 54, 32 },
		{ 289, 55, 32 },
		{ 289, 56, 32 },
		{ 289, 57, 32 },
		{ 289, 58, 32 },
		{ 289, 59, 32 },
		{ 289, 60, 32 },
		{ 289, 61, 32 },
		{ 289, 62, 32 },
		{ 289, 63, 32 },
		{ 289, 64, 32 },
		{ 289, 65, 32 },
		{ 289, 66, 32 },
		{ 289, 67, 32 },
		{ 289, 68, 32 },
		{ 289, 69, 32 },
		{ 289, 70, 32 },
		{ 289, 71, 32 },
		{ 289, 72, 32 },
		{ 289, 73, 32 },
		{ 289, 74, 32 },
		{ 289, 75, 32 },
		{ 289, 76, 32 },
		{ 289, 77, 32 },
		{ 289, 78, 32 },
		{ 289, 79, 32 },
		{ 289, 80, 32 },
		{ 289, 81, 32 },
		{ 289, 82, 32 },
		{ 289, 83, 32 },
		{ 289, 84, 32 },
		{ 289, 85, 32 },
		{ 289, 86, 32 },
		{ 289, 87, 32 },
		{ 289, 88, 32 },
		{ 289, 89, 32 },
		{ 289, 90, 32 },
		{ 289, 91, 32 },
		{ 289, 92, 32 },
		{ 289, 93, 32 },
		{ 289, 94, 32 },
		{ 289, 95, 32 },
		{ 289, 96, 32 },
		{ 289, 97, 32 },
		{ 289, 98, 32 },
		{ 289, 99, 32 },
		{ 289, 100, 32 },
		{ 289, 101, 32 },
		{ 289, 102, 32 },
		{ 289, 103, 32 },
		{ 289, 104, 32 },
		{ 289, 105, 32 },
		{ 289, 106, 32 },
		{ 289, 107, 32 },
		{ 289, 108, 32 },
		{ 289, 109, 32 },
		{ 289, 110, 32 },
		{ 289, 111, 32 },
		{ 289, 112, 32 },
		{ 289, 113, 32 },
		{ 289, 114, 32 },
		{ 289, 115, 32 },
		{ 289, 116, 32 },
		{ 289, 117, 32 },
		{ 289, 118, 32 },
		{ 289, 119, 32 },
		{ 289, 120, 32 },
		{ 289, 121, 32 },
		{ 289, 122, 32 },
		{ 289, 123, 32 },
		{ 289, 124, 32 },
		{ 289, 125, 32 },
		{ 289, 126, 32 },
		{ 289, 127, 32 },
		{ 289, 128, 32 },
		{ 289, 129, 32 },
		{ 289, 130, 32 },
		{ 289, 131, 32 },
		{ 289, 132, 32 },
		{ 289, 133, 32 },
		{ 289, 134, 32 },
		{ 289, 135, 32 },
		{ 289, 136, 32 },
		{ 289, 137, 32 },
		{ 289, 138, 32 },
		{ 289, 139, 32 },
		{ 289, 140, 32 },
		{ 289, 141, 32 },
		{ 289, 142, 32 },
		{ 289, 143, 32 },
		{ 289, 144, 32 },
		{ 289, 145, 32 },
		{ 289, 146, 32 },
		{ 289, 147, 32 },
		{ 289, 148, 32 },
		{ 289, 149, 32 },
		{ 289, 150, 32 },
		{ 289, 151, 32 },
		{ 289, 152, 32 },
		{ 289, 153, 32 },
		{ 289, 154, 32 },
		{ 289, 155, 32 },
		{ 289, 156, 32 },
		{ 289, 157, 32 },
		{ 289, 158, 32 },
		{ 289, 159, 32 },
		{ 289, 160, 32 },
		{ 289, 161, 32 },
		{ 289, 162, 32 },
		{ 289, 163, 32 },
		{ 289, 164, 32 },
		{ 289, 165, 32 },
		{ 289, 166, 32 },
		{ 289, 167, 32 },
		{ 289, 168, 32 },
		{ 289, 169, 32 },
		{ 289, 170, 32 },
		{ 289, 171, 32 },
		{ 290, 1, 32 },
		{ 290, 2, 32 },
		{ 290, 3, 32 },
		{ 290, 4, 32 },
		{ 290, 5, 32 },
		{ 290, 6, 32 },
		{ 290, 7, 32 },
		{ 290, 8, 32 },
		{ 290, 9, 32 },
		{ 290, 10, 32 },
		{ 290, 11, 32 },
		{ 290, 12, 32 },
		{ 290, 13, 32 },
		{ 290, 14, 32 },
		{ 290, 15, 32 },
		{ 290, 16, 32 },
		{ 290, 17, 32 },
		{ 290, 18, 32 },
		{ 290, 19, 32 },
		{ 290, 20, 32 },
		{ 290, 21, 32 },
		{ 290, 22, 32 },
		{ 290, 23, 32 },
		{ 290, 24, 32 },
		{ 290, 25, 32 },
		{ 290, 26, 32 },
		{ 290, 27, 32 },
		{ 290, 28, 32 },
		{ 290, 29, 32 },
		{ 290, 30, 32 },
		{ 290, 31, 32 },
		{ 290, 32, 32 },
		{ 290, 33, 32 },
		{ 290, 34, 32 },
		{ 290, 35, 32 },
		{ 290, 36, 32 },
		{ 290, 37, 32 },
		{ 290, 38, 32 },
		{ 290, 39, 32 },
		{ 290, 40, 32 },
		{ 290, 41, 32 },
		{ 290, 42, 32 },
		{ 290, 43, 32 },
		{ 290, 44, 32 },
		{ 290, 45, 32 },
		{ 290, 46, 32 },
		{ 290, 47, 32 },
		{ 290, 48, 32 },
		{ 290, 49, 32 },
		{ 290, 50, 32 },
		{ 290, 51, 32 },
		{ 290, 52, 32 },
		{ 290, 53, 32 },
		{ 290, 54, 32 },
		{ 290, 55, 32 },
		{ 290, 56, 32 },
		{ 290, 57, 32 },
		{ 290, 58, 32 },
		{ 290, 59, 32 },
		{ 290, 60, 32 },
		{ 290, 61, 32 },
		{ 290, 62, 32 },
		{ 290, 63, 32 },
		{ 290, 64, 32 },
		{ 290, 65, 32 },
		{ 290, 66, 32 },
		{ 290, 67, 32 },
		{ 290, 68, 32 },
		{ 290, 69, 32 },
		{ 290, 70, 32 },
		{ 290, 71, 32 },
		{ 290, 72, 32 },
		{ 290, 73, 32 },
		{ 290, 74, 32 },
		{ 290, 75, 32 },
		{ 290, 76, 32 },
		{ 290, 77, 32 },
		{ 290, 78, 32 },
		{ 290, 79, 32 },
		{ 290, 80, 32 },
		{ 290, 81, 32 },
		{ 290, 82, 32 },
		{ 290, 83, 32 },
		{ 290, 84, 32 },
		{ 290, 85, 32 },
		{ 290, 86, 32 },
		{ 290, 87, 32 },
		{ 290, 88, 32 },
		{ 290, 89, 32 },
		{ 290, 90, 32 },
		{ 290, 91, 32 },
		{ 290, 92, 32 },
		{ 290, 93, 32 },
		{ 290, 94, 32 },
		{ 290, 95, 32 },
		{ 290, 96, 32 },
		{ 290, 97, 32 },
		{ 290, 98, 32 },
		{ 290, 99, 32 },
		{ 290, 100, 32 },
		{ 290, 101, 32 },
		{ 290, 102, 32 },
		{ 290, 103, 32 },
		{ 290, 104, 32 },
		{ 290, 105, 32 },
		{ 290, 106, 32 },
		{ 290, 107, 32 },
		{ 290, 108, 32 },
		{ 290, 109, 32 },
		{ 290, 110, 32 },
		{ 290, 111, 32 },
		{ 290, 112, 32 },
		{ 290, 113, 32 },
		{ 290, 114, 32 },
		{ 290, 115, 32 },
		{ 290, 116, 32 },
		{ 290, 117, 32 },
		{ 290, 118, 32 },
		{ 290, 119, 32 },
		{ 290, 120, 32 },
		{ 290, 121, 32 },
		{ 290, 122, 32 },
		{ 290, 123, 32 },
		{ 290, 124, 32 },
		{ 290, 125, 32 },
		{ 290, 126, 32 },
		{ 290, 127, 32 },
		{ 290, 128, 32 },
		{ 290, 129, 32 },
		{ 290, 130, 32 },
		{ 290, 131, 32 },
		{ 290, 132, 32 },
		{ 290, 133, 32 },
		{ 290, 134, 32 },
		{ 290, 135, 32 },
		{ 290, 136, 32 },
		{ 290, 137, 32 },
		{ 290, 138, 32 },
		{ 290, 139, 32 },
		{ 290, 140, 32 },
		{ 290, 141, 32 },
		{ 290, 142, 32 },
		{ 290, 143, 32 },
		{ 290, 144, 32 },
		{ 290, 145, 32 },
		{ 290, 146, 32 },
		{ 290, 147, 32 },
		{ 290, 148, 32 },
		{ 290, 149, 32 },
		{ 290, 150, 32 },
		{ 290, 151, 32 },
		{ 290, 152, 32 },
		{ 290, 153, 32 },
		{ 290, 154, 32 },
		{ 290, 155, 32 },
		{ 290, 156, 32 },
		{ 290, 157, 32 },
		{ 290, 158, 32 },
		{ 290, 159, 32 },
		{ 290, 160, 32 },
		{ 290, 161, 32 },
		{ 290, 162, 32 },
		{ 290, 163, 32 },
		{ 290, 164, 32 },
		{ 290, 165, 32 },
		{ 290, 166, 32 },
		{ 290, 167, 32 },
		{ 290, 168, 32 },
		{ 291, 1, 32 },
		{ 291, 2, 32 },
		{ 291, 3, 32 },
		{ 291, 4, 32 },
		{ 291, 5, 32 },
		{ 291, 6, 32 },
		{ 291, 7, 32 },
		{ 291, 8, 32 },
		{ 291, 9, 32 },
		{ 291, 10, 32 },
		{ 291, 11, 32 },
		{ 291, 12, 32 },
		{ 291, 13, 32 },
		{ 291, 14, 32 },
		{ 291, 15, 32 },
		{ 291, 16, 32 },
		{ 291, 17, 32 },
		{ 291, 18, 32 },
		{ 291, 19, 32 },
		{ 291, 20, 32 },
		{ 291, 21, 32 },
		{ 291, 22, 32 },
		{ 291, 23, 32 },
		{ 291, 24, 32 },
		{ 291, 25, 32 },
		{ 291, 26, 32 },
		{ 291, 27, 32 },
		{ 291, 28, 32 },
		{ 291, 29, 32 },
		{ 291, 30, 32 },
		{ 291, 31, 32 },
		{ 291, 32, 32 },
		{ 291, 33, 32 },
		{ 291, 34, 32 },
		{ 291, 35, 32 },
		{ 291, 36, 32 },
		{ 291, 37, 32 },
		{ 291, 38, 32 },
		{ 291, 39, 32 },
		{ 291, 40, 32 },
		{ 291, 41, 32 },
		{ 291, 42, 32 },
		{ 291, 43, 32 },
		{ 291, 44, 32 },
		{ 291, 45, 32 },
		{ 291, 46, 32 },
		{ 291, 47, 32 },
		{ 291, 48, 32 },
		{ 291, 49, 32 },
		{ 291, 50, 32 },
		{ 291, 51, 32 },
		{ 291, 52, 32 },
		{ 291, 53, 32 },
		{ 291, 54, 32 },
		{ 291, 55, 32 },
		{ 291, 56, 32 },
		{ 291, 57, 32 },
		{ 291, 58, 32 },
		{ 291, 59, 32 },
		{ 291, 60, 32 },
		{ 291, 61, 32 },
		{ 291, 62, 32 },
		{ 291, 63, 32 },
		{ 291, 64, 32 },
		{ 291, 65, 32 },
		{ 291, 66, 32 },
		{ 291, 67, 32 },
		{ 291, 68, 32 },
		{ 291, 69, 32 },
		{ 291, 70, 32 },
		{ 291, 71, 32 },
		{ 291, 72, 32 },
		{ 291, 73, 32 },
		{ 291, 74, 32 },
		{ 291, 75, 32 },
		{ 291, 76, 32 },
		{ 291, 77, 32 },
		{ 291, 78, 32 },
		{ 291, 79, 32 },
		{ 291, 80, 32 },
		{ 291, 81, 32 },
		{ 291, 82, 32 },
		{ 291, 83, 32 },
		{ 291, 84, 32 },
		{ 291, 85, 32 },
		{ 291, 86, 32 },
		{ 291, 87, 32 },
		{ 291, 88, 32 },
		{ 291, 89, 32 },
		{ 291, 90, 32 },
		{ 291, 91, 32 },
		{ 291, 92, 32 },
		{ 291, 93, 32 },
		{ 291, 94, 32 },
		{ 291, 95, 32 },
		{ 291, 96, 32 },
		{ 291, 97, 32 },
		{ 291, 98, 32 },
		{ 291, 99, 32 },
		{ 291, 100, 32 },
		{ 291, 101, 32 },
		{ 291, 102, 32 },
		{ 291, 103, 32 },
		{ 291, 104, 32 },
		{ 291, 105, 32 },
		{ 291, 106, 32 },
		{ 291, 107, 32 },
		{ 291, 108, 32 },
		{ 291, 109, 32 },
		{ 291, 110, 32 },
		{ 291, 111, 32 },
		{ 291, 112, 32 },
		{ 291, 113, 32 },
		{ 291, 114, 32 },
		{ 291, 115, 32 },
		{ 291, 116, 32 },
		{ 291, 117, 32 },
		{ 291, 118, 32 },
		{ 291, 119, 32 },
		{ 291, 120, 32 },
		{ 291, 121, 32 },
		{ 291, 122, 32 },
		{ 291, 123, 32 },
		{ 291, 124, 32 },
		{ 291, 125, 32 },
		{ 291, 126, 32 },
		{ 291, 127, 32 },
		{ 291, 128, 32 },
		{ 291, 129, 32 },
		{ 291, 130, 32 },
		{ 291, 131, 32 },
		{ 291, 132, 32 },
		{ 291, 133, 32 },
		{ 291, 134, 32 },
		{ 291, 135, 32 },
		{ 291, 136, 32 },
		{ 291, 137, 32 },
		{ 291, 138, 32 },
		{ 291, 139, 32 },
		{ 291, 140, 32 },
		{ 291, 141, 32 },
		{ 291, 142, 32 },
		{ 291, 143, 32 },
		{ 291, 144, 32 },
		{ 291, 145, 32 },
		{ 291, 146, 32 },
		{ 291, 147, 32 },
		{ 291, 148, 32 },
		{ 291, 149, 32 },
		{ 291, 150, 32 },
		{ 291, 151, 32 },
		{ 291, 152, 32 },
		{ 291, 153, 32 },
		{ 291, 154, 32 },
		{ 291, 155, 32 },
		{ 291, 156, 32 },
		{ 291, 157, 32 },
		{ 291, 158, 32 },
		{ 291, 159, 32 },
		{ 291, 160, 32 },
		{ 291, 161, 32 },
		{ 291, 162, 32 },
		{ 291, 163, 32 },
		{ 291, 164, 32 },
		{ 291, 165, 32 },
		{ 291, 166, 32 },
		{ 292, 1, 32 },
		{ 292, 2, 32 },
		{ 292, 3, 32 },
		{ 292, 4, 32 },
		{ 292, 5, 32 },
		{ 292, 6, 32 },
		{ 292, 7, 32 },
		{ 292, 8, 32 },
		{ 292, 9, 32 },
		{ 292, 10, 32 },
		{ 292, 11, 32 },
		{ 292, 12, 32 },
		{ 292, 13, 32 },
		{ 292, 14, 32 },
		{ 292, 15, 32 },
		{ 292, 16, 32 },
		{ 292, 17, 32 },
		{ 292, 18, 32 },
		{ 292, 19, 32 },
		{ 292, 20, 32 },
		{ 292, 21, 32 },
		{ 292, 22, 32 },
		{ 292, 23, 32 },
		{ 292, 24, 32 },
		{ 292, 25, 32 },
		{ 292, 26, 32 },
		{ 292, 27, 32 },
		{ 292, 28, 32 },
		{ 292, 29, 32 },
		{ 292, 30, 32 },
		{ 292, 31, 32 },
		{ 292, 32, 32 },
		{ 292, 33, 32 },
		{ 292, 34, 32 },
		{ 292, 35, 32 },
		{ 292, 36, 32 },
		{ 292, 37, 32 },
		{ 292, 38, 32 },
		{ 292, 39, 32 },
		{ 292, 40, 32 },
		{ 292, 41, 32 },
		{ 292, 42, 32 },
		{ 292, 43, 32 },
		{ 292, 44, 32 },
		{ 292, 45, 32 },
		{ 292, 46, 32 },
		{ 292, 47, 32 },
		{ 292, 48, 32 },
		{ 292, 49, 32 },
		{ 292, 50, 32 },
		{ 292, 51, 32 },
		{ 292, 52, 32 },
		{ 292, 53, 32 },
		{ 292, 54, 32 },
		{ 292, 55, 32 },
		{ 292, 56, 32 },
		{ 292, 57, 32 },
		{ 292, 58, 32 },
		{ 292, 59, 32 },
		{ 292, 60, 32 },
		{ 292, 61, 32 },
		{ 292, 62, 32 },
		{ 292, 63, 32 },
		{ 292, 64, 32 },
		{ 292, 65, 32 },
		{ 292, 66, 32 },
		{ 292, 67, 32 },
		{ 292, 68, 32 },
		{ 292, 69, 32 },
		{ 292, 70, 32 },
		{ 292, 71, 32 },
		{ 292, 72, 32 },
		{ 292, 73, 32 },
		{ 292, 74, 32 },
		{ 292, 75, 32 },
		{ 292, 76, 32 },
		{ 292, 77, 32 },
		{ 292, 78, 32 },
		{ 292, 79, 32 },
		{ 292, 80, 32 },
		{ 292, 81, 32 },
		{ 292, 82, 32 },
		{ 292, 83, 32 },
		{ 292, 84, 32 },
		{ 292, 85, 32 },
		{ 292, 86, 32 },
		{ 292, 87, 32 },
		{ 292, 88, 32 },
		{ 292, 89, 32 },
		{ 292, 90, 32 },
		{ 292, 91, 32 },
		{ 292, 92, 32 },
		{ 292, 93, 32 },
		{ 292, 94, 32 },
		{ 292, 95, 32 },
		{ 292, 96, 32 },
		{ 292, 97, 32 },
		{ 292, 98, 32 },
		{ 292, 99, 32 },
		{ 292, 100, 32 },
		{ 292, 101, 32 },
		{ 292, 102, 32 },
		{ 292, 103, 32 },
		{ 292, 104, 32 },
		{ 292, 105, 32 },
		{ 292, 106, 32 },
		{ 292, 107, 32 },
		{ 292, 108, 32 },
		{ 292, 109, 32 },
		{ 292, 110, 32 },
		{ 292, 111, 32 },
		{ 292, 112, 32 },
		{ 292, 113, 32 },
		{ 292, 114, 32 },
		{ 292, 115, 32 },
		{ 292, 116, 32 },
		{ 292, 117, 32 },
		{ 292, 118, 32 },
		{ 292, 119, 32 },
		{ 292, 120, 32 },
		{ 292, 121, 32 },
		{ 292, 122, 32 },
		{ 292, 123, 32 },
		{ 292, 124, 32 },
		{ 292, 125, 32 },
		{ 292, 126, 32 },
		{ 292, 127, 32 },
		{ 292, 128, 32 },
		{ 292, 129, 32 },
		{ 292, 130, 32 },
		{ 292, 131, 32 },
		{ 292, 132, 32 },
		{ 292, 133, 32 },
		{ 292, 134, 32 },
		{ 292, 135, 32 },
		{ 292, 136, 32 },
		{ 292, 137, 32 },
		{ 292, 138, 32 },
		{ 292, 139, 32 },
		{ 292, 140, 32 },
		{ 292, 141, 32 },
		{ 292, 142, 32 },
		{ 292, 143, 32 },
		{ 292, 144, 32 },
		{ 292, 145, 32 },
		{ 292, 146, 32 },
		{ 292, 147, 32 },
		{ 292, 148, 32 },
		{ 292, 149, 32 },
		{ 292, 150, 32 },
		{ 292, 151, 32 },
		{ 292, 152, 32 },
		{ 292, 153, 32 },
		{ 292, 154, 32 },
		{ 292, 155, 32 },
		{ 292, 156, 32 },
		{ 292, 157, 32 },
		{ 292, 158, 32 },
		{ 292, 159, 32 },
		{ 292, 160, 32 },
		{ 292, 161, 32 },
		{ 292, 162, 32 },
		{ 292, 163, 32 },
		{ 293, 1, 32 },
		{ 293, 2, 32 },
		{ 293, 3, 32 },
		{ 293, 4, 32 },
		{ 293, 5, 32 },
		{ 293, 6, 32 },
		{ 293, 7, 32 },
		{ 293, 8, 32 },
		{ 293, 9, 32 },
		{ 293, 10, 32 },
		{ 293, 11, 32 },
		{ 293, 12, 32 },
		{ 293, 13, 32 },
		{ 293, 14, 32 },
		{ 293, 15, 32 },
		{ 293, 16, 32 },
		{ 293, 17, 32 },
		{ 293, 18, 32 },
		{ 293, 19, 32 },
		{ 293, 20, 32 },
		{ 293, 21, 32 },
		{ 293, 22, 32 },
		{ 293, 23, 32 },
		{ 293, 24, 32 },
		{ 293, 25, 32 },
		{ 293, 26, 32 },
		{ 293, 27, 32 },
		{ 293, 28, 32 },
		{ 293, 29, 32 },
		{ 293, 30, 32 },
		{ 293, 31, 32 },
		{ 293, 32, 32 },
		{ 293, 33, 32 },
		{ 293, 34, 32 },
		{ 293, 35, 32 },
		{ 293, 36, 32 },
		{ 293, 37, 32 },
		{ 293, 38, 32 },
		{ 293, 39, 32 },
		{ 293, 40, 32 },
		{ 293, 41, 32 },
		{ 293, 42, 32 },
		{ 293, 43, 32 },
		{ 293, 44, 32 },
		{ 293, 45, 32 },
		{ 293, 46, 32 },
		{ 293, 47, 32 },
		{ 293, 48, 32 },
		{ 293, 49, 32 },
		{ 293, 50, 32 },
		{ 293, 51, 32 },
		{ 293, 52, 32 },
		{ 293, 53, 32 },
		{ 293, 54, 32 },
		{ 293, 55, 32 },
		{ 293, 56, 32 },
		{ 293, 57, 32 },
		{ 293, 58, 32 },
		{ 293, 59, 32 },
		{ 293, 60, 32 },
		{ 293, 61, 32 },
		{ 293, 62, 32 },
		{ 293, 63, 32 },
		{ 293, 64, 32 },
		{ 293, 65, 32 },
		{ 293, 66, 32 },
		{ 293, 67, 32 },
		{ 293, 68, 32 },
		{ 293, 69, 32 },
		{ 293, 70, 32 },
		{ 293, 71, 32 },
		{ 293, 72, 32 },
		{ 293, 73, 32 },
		{ 293, 74, 32 },
		{ 293, 75, 32 },
		{ 293, 76, 32 },
		{ 293, 77, 32 },
		{ 293, 78, 32 },
		{ 293, 79, 32 },
		{ 293, 80, 32 },
		{ 293, 81, 32 },
		{ 293, 82, 32 },
		{ 293, 83, 32 },
		{ 293, 84, 32 },
		{ 293, 85, 32 },
		{ 293, 86, 32 },
		{ 293, 87, 32 },
		{ 293, 88, 32 },
		{ 293, 89, 32 },
		{ 293, 90, 32 },
		{ 293, 91, 32 },
		{ 293, 92, 32 },
		{ 293, 93, 32 },
		{ 293, 94, 32 },
		{ 293, 95, 32 },
		{ 293, 96, 32 },
		{ 293, 97, 32 },
		{ 293, 98, 32 },
		{ 293, 99, 32 },
		{ 293, 100, 32 },
		{ 293, 101, 32 },
		{ 293, 102, 32 },
		{ 293, 103, 32 },
		{ 293, 104, 32 },
		{ 293, 105, 32 },
		{ 293, 106, 32 },
		{ 293, 107, 32 },
		{ 293, 108, 32 },
		{ 293, 109, 32 },
		{ 293, 110, 32 },
		{ 293, 111, 32 },
		{ 293, 112, 32 },
		{ 293, 113, 32 },
		{ 293, 114, 32 },
		{ 293, 115, 32 },
		{ 293, 116, 32 },
		{ 293, 117, 32 },
		{ 293, 118, 32 },
		{ 293, 119, 32 },
		{ 293, 120, 32 },
		{ 293, 121, 32 },
		{ 293, 122, 32 },
		{ 293, 123, 32 },
		{ 293, 124, 32 },
		{ 293, 125, 32 },
		{ 293, 126, 32 },
		{ 293, 127, 32 },
		{ 293, 128, 32 },
		{ 293, 129, 32 },
		{ 293, 130, 32 },
		{ 293, 131, 32 },
		{ 293, 132, 32 },
		{ 293, 133, 32 },
		{ 293, 134, 32 },
		{ 293, 135, 32 },
		{ 293, 136, 32 },
		{ 293, 137, 32 },
		{ 293, 138, 32 },
		{ 293, 139, 32 },
		{ 293, 140, 32 },
		{ 293, 141, 32 },
		{ 293, 142, 32 },
		{ 293, 143, 32 },
		{ 293, 144, 32 },
		{ 293, 145, 32 },
		{ 293, 146, 32 },
		{ 293, 147, 32 },
		{ 293, 148, 32 },
		{ 293, 149, 32 },
		{ 293, 150, 32 },
		{ 293, 151, 32 },
		{ 293, 152, 32 },
		{ 293, 153, 32 },
		{ 293, 154, 32 },
		{ 293, 155, 32 },
		{ 293, 156, 32 },
		{ 293, 157, 32 },
		{ 293, 158, 32 },
		{ 293, 159, 32 },
		{ 293, 160, 32 },
		{ 294, 1, 32 },
		{ 294, 2, 32 },
		{ 294, 3, 32 },
		{ 294, 4, 32 },
		{ 294, 5, 32 },
		{ 294, 6, 32 },
		{ 294, 7, 32 },
		{ 294, 8, 32 },
		{ 294, 9, 32 },
		{ 294, 10, 32 },
		{ 294, 11, 32 },
		{ 294, 12, 32 },
		{ 294, 13, 32 },
		{ 294, 14, 32 },
		{ 294, 15, 32 },
		{ 294, 16, 32 },
		{ 294, 17, 32 },
		{ 294, 18, 32 },
		{ 294, 19, 32 },
		{ 294, 20, 32 },
		{ 294, 21, 32 },
		{ 294, 22, 32 },
		{ 294, 23, 32 },
		{ 294, 24, 32 },
		{ 294, 25, 32 },
		{ 294, 26, 32 },
		{ 294, 27, 32 },
		{ 294, 28, 32 },
		{ 294, 29, 32 },
		{ 294, 30, 32 },
		{ 294, 31, 32 },
		{ 294, 32, 32 },
		{ 294, 33, 32 },
		{ 294, 34, 32 },
		{ 294, 35, 32 },
		{ 294, 36, 32 },
		{ 294, 37, 32 },
		{ 294, 38, 32 },
		{ 294, 39, 32 },
		{ 294, 40, 32 },
		{ 294, 41, 32 },
		{ 294, 42, 32 },
		{ 294, 43, 32 },
		{ 294, 44, 32 },
		{ 294, 45, 32 },
		{ 294, 46, 32 },
		{ 294, 47, 32 },
		{ 294, 48, 32 },
		{ 294, 49, 32 },
		{ 294, 50, 32 },
		{ 294, 51, 32 },
		{ 294, 52, 32 },
		{ 294, 53, 32 },
		{ 294, 54, 32 },
		{ 294, 55, 32 },
		{ 294, 56, 32 },
		{ 294, 57, 32 },
		{ 294, 58, 32 },
		{ 294, 59, 32 },
		{ 294, 60, 32 },
		{ 294, 61, 32 },
		{ 294, 62, 32 },
		{ 294, 63, 32 },
		{ 294, 64, 32 },
		{ 294, 65, 32 },
		{ 294, 66, 32 },
		{ 294, 67, 32 },
		{ 294, 68, 32 },
		{ 294, 69, 32 },
		{ 294, 70, 32 },
		{ 294, 71, 32 },
		{ 294, 72, 32 },
		{ 294, 73, 32 },
		{ 294, 74, 32 },
		{ 294, 75, 32 },
		{ 294, 76, 32 },
		{ 294, 77, 32 },
		{ 294, 78, 32 },
		{ 294, 79, 32 },
		{ 294, 80, 32 },
		{ 294, 81, 32 },
		{ 294, 82, 32 },
		{ 294, 83, 32 },
		{ 294, 84, 32 },
		{ 294, 85, 32 },
		{ 294, 86, 32 },
		{ 294, 87, 32 },
		{ 294, 88, 32 },
		{ 294, 89, 32 },
		{ 294, 90, 32 },
		{ 294, 91, 32 },
		{ 294, 92, 32 },
		{ 294, 93, 32 },
		{ 294, 94, 32 },
		{ 294, 95, 32 },
		{ 294, 96, 32 },
		{ 294, 97, 32 },
		{ 294, 98, 32 },
		{ 294, 99, 32 },
		{ 294, 100, 32 },
		{ 294, 101, 32 },
		{ 294, 102, 32 },
		{ 294, 103, 32 },
		{ 294, 104, 32 },
		{ 294, 105, 32 },
		{ 294, 106, 32 },
		{ 294, 107, 32 },
		{ 294, 108, 32 },
		{ 294, 109, 32 },
		{ 294, 110, 32 },
		{ 294, 111, 32 },
		{ 294, 112, 32 },
		{ 294, 113, 32 },
		{ 294, 114, 32 },
		{ 294, 115, 32 },
		{ 294, 116, 32 },
		{ 294, 117, 32 },
		{ 294, 118, 32 },
		{ 294, 119, 32 },
		{ 294, 120, 32 },
		{ 294, 121, 32 },
		{ 294, 122, 32 },
		{ 294, 123, 32 },
		{ 294, 124, 32 },
		{ 294, 125, 32 },
		{ 294, 126, 32 },
		{ 294, 127, 32 },
		{ 294, 128, 32 },
		{ 294, 129, 32 },
		{ 294, 130, 32 },
		{ 294, 131, 32 },
		{ 294, 132, 32 },
		{ 294, 133, 32 },
		{ 294, 134, 32 },
		{ 294, 135, 32 },
		{ 294, 136, 32 },
		{ 294, 137, 32 },
		{ 294, 138, 32 },
		{ 294, 139, 32 },
		{ 294, 140, 32 },
		{ 294, 141, 32 },
		{ 294, 142, 32 },
		{ 294, 143, 32 },
		{ 294, 144, 32 },
		{ 294, 145, 32 },
		{ 294, 146, 32 },
		{ 294, 147, 32 },
		{ 294, 148, 32 },
		{ 294, 149, 32 },
		{ 294, 150, 32 },
		{ 294, 151, 32 },
		{ 294, 152, 32 },
		{ 294, 153, 32 },
		{ 294, 154, 32 },
		{ 294, 155, 32 },
		{ 294, 156, 32 },
		{ 294, 157, 32 },
		{ 294, 158, 32 },
		{ 295, 1, 32 },
		{ 295, 2, 32 },
		{ 295, 3, 32 },
		{ 295, 4, 32 },
		{ 295, 5, 32 },
		{ 295, 6, 32 },
		{ 295, 7, 32 },
		{ 295, 8, 32 },
		{ 295, 9, 32 },
		{ 295, 10, 32 },
		{ 295, 11, 32 },
		{ 295, 12, 32 },
		{ 295, 13, 32 },
		{ 295, 14, 32 },
		{ 295, 15, 32 },
		{ 295, 16, 32 },
		{ 295, 17, 32 },
		{ 295, 18, 32 },
		{ 295, 19, 32 },
		{ 295, 20, 32 },
		{ 295, 21, 32 },
		{ 295, 22, 32 },
		{ 295, 23, 32 },
		{ 295, 24, 32 },
		{ 295, 25, 32 },
		{ 295, 26, 32 },
		{ 295, 27, 32 },
		{ 295, 28, 32 },
		{ 295, 29, 32 },
		{ 295, 30, 32 },
		{ 295, 31, 32 },
		{ 295, 32, 32 },
		{ 295, 33, 32 },
		{ 295, 34, 32 },
		{ 295, 35, 32 },
		{ 295, 36, 32 },
		{ 295, 37, 32 },
		{ 295, 38, 32 },
		{ 295, 39, 32 },
		{ 295, 40, 32 },
		{ 295, 41, 32 },
		{ 295, 42, 32 },
		{ 295, 43, 32 },
		{ 295, 44, 32 },
		{ 295, 45, 32 },
		{ 295, 46, 32 },
		{ 295, 47, 32 },
		{ 295, 48, 32 },
		{ 295, 49, 32 },
		{ 295, 50, 32 },
		{ 295, 51, 32 },
		{ 295, 52, 32 },
		{ 295, 53, 32 },
		{ 295, 54, 32 },
		{ 295, 55, 32 },
		{ 295, 56, 32 },
		{ 295, 57, 32 },
		{ 295, 58, 32 },
		{ 295, 59, 32 },
		{ 295, 60, 32 },
		{ 295, 61, 32 },
		{ 295, 62, 32 },
		{ 295, 63, 32 },
		{ 295, 64, 32 },
		{ 295, 65, 32 },
		{ 295, 66, 32 },
		{ 295, 67, 32 },
		{ 295, 68, 32 },
		{ 295, 69, 32 },
		{ 295, 70, 32 },
		{ 295, 71, 32 },
		{ 295, 72, 32 },
		{ 295, 73, 32 },
		{ 295, 74, 32 },
		{ 295, 75, 32 },
		{ 295, 76, 32 },
		{ 295, 77, 32 },
		{ 295, 78, 32 },
		{ 295, 79, 32 },
		{ 295, 80, 32 },
		{ 295, 81, 32 },
		{ 295, 82, 32 },
		{ 295, 83, 32 },
		{ 295, 84, 32 },
		{ 295, 85, 32 },
		{ 295, 86, 32 },
		{ 295, 87, 32 },
		{ 295, 88, 32 },
		{ 295, 89, 32 },
		{ 295, 90, 32 },
		{ 295, 91, 32 },
		{ 295, 92, 32 },
		{ 295, 93, 32 },
		{ 295, 94, 32 },
		{ 295, 95, 32 },
		{ 295, 96, 32 },
		{ 295, 97, 32 },
		{ 295, 98, 32 },
		{ 295, 99, 32 },
		{ 295, 100, 32 },
		{ 295, 101, 32 },
		{ 295, 102, 32 },
		{ 295, 103, 32 },
		{ 295, 104, 32 },
		{ 295, 105, 32 },
		{ 295, 106, 32 },
		{ 295, 107, 32 },
		{ 295, 108, 32 },
		{ 295, 109, 32 },
		{ 295, 110, 32 },
		{ 295, 111, 32 },
		{ 295, 112, 32 },
		{ 295, 113, 32 },
		{ 295, 114, 32 },
		{ 295, 115, 32 },
		{ 295, 116, 32 },
		{ 295, 117, 32 },
		{ 295, 118, 32 },
		{ 295, 119, 32 },
		{ 295, 120, 32 },
		{ 295, 121, 32 },
		{ 295, 122, 32 },
		{ 295, 123, 32 },
		{ 295, 124, 32 },
		{ 295, 125, 32 },
		{ 295, 126, 32 },
		{ 295, 127, 32 },
		{ 295, 128, 32 },
		{ 295, 129, 32 },
		{ 295, 130, 32 },
		{ 295, 131, 32 },
		{ 295, 132, 32 },
		{ 295, 133, 32 },
		{ 295, 134, 32 },
		{ 295, 135, 32 },
		{ 295, 136, 32 },
		{ 295, 137, 32 },
		{ 295, 138, 32 },
		{ 295, 139, 32 },
		{ 295, 140, 32 },
		{ 295, 141, 32 },
		{ 295, 142, 32 },
		{ 295, 143, 32 },
		{ 295, 144, 32 },
		{ 295, 145, 32 },
		{ 295, 146, 32 },
		{ 295, 147, 32 },
		{ 295, 148, 32 },
		{ 295, 149, 32 },
		{ 295, 150, 32 },
		{ 295, 151, 32 },
		{ 295, 152, 32 },
		{ 295, 153, 32 },
		{ 295, 154, 32 },
		{ 295, 155, 32 },
		{ 296, 1, 32 },
		{ 296, 2, 32 },
		{ 296, 3, 32 },
		{ 296, 4, 32 },
		{ 296, 5, 32 },
		{ 296, 6, 32 },
		{ 296, 7, 32 },
		{ 296, 8, 32 },
		{ 296, 9, 32 },
		{ 296, 10, 32 },
		{ 296, 11, 32 },
		{ 296, 12, 32 },
		{ 296, 13, 32 },
		{ 296, 14, 32 },
		{ 296, 15, 32 },
		{ 296, 16, 32 },
		{ 296, 17, 32 },
		{ 296, 18, 32 },
		{ 296, 19, 32 },
		{ 296, 20, 32 },
		{ 296, 21, 32 },
		{ 296, 22, 32 },
		{ 296, 23, 32 },
		{ 296, 24, 32 },
		{ 296, 25, 32 },
		{ 296, 26, 32 },
		{ 296, 27, 32 },
		{ 296, 28, 32 },
		{ 296, 29, 32 },
		{ 296, 30, 32 },
		{ 296, 31, 32 },
		{ 296, 32, 32 },
		{ 296, 33, 32 },
		{ 296, 34, 32 },
		{ 296, 35, 32 },
		{ 296, 36, 32 },
		{ 296, 37, 32 },
		{ 296, 38, 32 },
		{ 296, 39, 32 },
		{ 296, 40, 32 },
		{ 296, 41, 32 },
		{ 296, 42, 32 },
		{ 296, 43, 32 },
		{ 296, 44, 32 },
		{ 296, 45, 32 },
		{ 296, 46, 32 },
		{ 296, 47, 32 },
		{ 296, 48, 32 },
		{ 296, 49, 32 },
		{ 296, 50, 32 },
		{ 296, 51, 32 },
		{ 296, 52, 32 },
		{ 296, 53, 32 },
		{ 296, 54, 32 },
		{ 296, 55, 32 },
		{ 296, 56, 32 },
		{ 296, 57, 32 },
		{ 296, 58, 32 },
		{ 296, 59, 32 },
		{ 296, 60, 32 },
		{ 296, 61, 32 },
		{ 296, 62, 32 },
		{ 296, 63, 32 },
		{ 296, 64, 32 },
		{ 296, 65, 32 },
		{ 296, 66, 32 },
		{ 296, 67, 32 },
		{ 296, 68, 32 },
		{ 296, 69, 32 },
		{ 296, 70, 32 },
		{ 296, 71, 32 },
		{ 296, 72, 32 },
		{ 296, 73, 32 },
		{ 296, 74, 32 },
		{ 296, 75, 32 },
		{ 296, 76, 32 },
		{ 296, 77, 32 },
		{ 296, 78, 32 },
		{ 296, 79, 32 },
		{ 296, 80, 32 },
		{ 296, 81, 32 },
		{ 296, 82, 32 },
		{ 296, 83, 32 },
		{ 296, 84, 32 },
		{ 296, 85, 32 },
		{ 296, 86, 32 },
		{ 296, 87, 32 },
		{ 296, 88, 32 },
		{ 296, 89, 32 },
		{ 296, 90, 32 },
		{ 296, 91, 32 },
		{ 296, 92, 32 },
		{ 296, 93, 32 },
		{ 296, 94, 32 },
		{ 296, 95, 32 },
		{ 296, 96, 32 },
		{ 296, 97, 32 },
		{ 296, 98, 32 },
		{ 296, 99, 32 },
		{ 296, 100, 32 },
		{ 296, 101, 32 },
		{ 296, 102, 32 },
		{ 296, 103, 32 },
		{ 296, 104, 32 },
		{ 296, 105, 32 },
		{ 296, 106, 32 },
		{ 296, 107, 32 },
		{ 296, 108, 32 },
		{ 296, 109, 32 },
		{ 296, 110, 32 },
		{ 296, 111, 32 },
		{ 296, 112, 32 },
		{ 296, 113, 32 },
		{ 296, 114, 32 },
		{ 296, 115, 32 },
		{ 296, 116, 32 },
		{ 296, 117, 32 },
		{ 296, 118, 32 },
		{ 296, 119, 32 },
		{ 296, 120, 32 },
		{ 296, 121, 32 },
		{ 296, 122, 32 },
		{ 296, 123, 32 },
		{ 296, 124, 32 },
		{ 296, 125, 32 },
		{ 296, 126, 32 },
		{ 296, 127, 32 },
		{ 296, 128, 32 },
		{ 296, 129, 32 },
		{ 296, 130, 32 },
		{ 296, 131, 32 },
		{ 296, 132, 32 },
		{ 296, 133, 32 },
		{ 296, 134, 32 },
		{ 296, 135, 32 },
		{ 296, 136, 32 },
		{ 296, 137, 32 },
		{ 296, 138, 32 },
		{ 296, 139, 32 },
		{ 296, 140, 32 },
		{ 296, 141, 32 },
		{ 296, 142, 32 },
		{ 296, 143, 32 },
		{ 296, 144, 32 },
		{ 296, 145, 32 },
		{ 296, 146, 32 },
		{ 296, 147, 32 },
		{ 296, 148, 32 },
		{ 296, 149, 32 },
		{ 296, 150, 32 },
		{ 296, 151, 32 },
		{ 296, 152, 32 },
		{ 297, 1, 32 },
		{ 297, 2, 32 },
		{ 297, 3, 32 },
		{ 297, 4, 32 },
		{ 297, 5, 32 },
		{ 297, 6, 32 },
		{ 297, 7, 32 },
		{ 297, 8, 32 },
		{ 297, 9, 32 },
		{ 297, 10, 32 },
		{ 297, 11, 32 },
		{ 297, 12, 32 },
		{ 297, 13, 32 },
		{ 297, 14, 32 },
		{ 297, 15, 32 },
		{ 297, 16, 32 },
		{ 297, 17, 32 },
		{ 297, 18, 32 },
		{ 297, 19, 32 },
		{ 297, 20, 32 },
		{ 297, 21, 32 },
		{ 297, 22, 32 },
		{ 297, 23, 32 },
		{ 297, 24, 32 },
		{ 297, 25, 32 },
		{ 297, 26, 32 },
		{ 297, 27, 32 },
		{ 297, 28, 32 },
		{ 297, 29, 32 },
		{ 297, 30, 32 },
		{ 297, 31, 32 },
		{ 297, 32, 32 },
		{ 297, 33, 32 },
		{ 297, 34, 32 },
		{ 297, 35, 32 },
		{ 297, 36, 32 },
		{ 297, 37, 32 },
		{ 297, 38, 32 },
		{ 297, 39, 32 },
		{ 297, 40, 32 },
		{ 297, 41, 32 },
		{ 297, 42, 32 },
		{ 297, 43, 32 },
		{ 297, 44, 32 },
		{ 297, 45, 32 },
		{ 297, 46, 32 },
		{ 297, 47, 32 },
		{ 297, 48, 32 },
		{ 297, 49, 32 },
		{ 297, 50, 32 },
		{ 297, 51, 32 },
		{ 297, 52, 32 },
		{ 297, 53, 32 },
		{ 297, 54, 32 },
		{ 297, 55, 32 },
		{ 297, 56, 32 },
		{ 297, 57, 32 },
		{ 297, 58, 32 },
		{ 297, 59, 32 },
		{ 297, 60, 32 },
		{ 297, 61, 32 },
		{ 297, 62, 32 },
		{ 297, 63, 32 },
		{ 297, 64, 32 },
		{ 297, 65, 32 },
		{ 297, 66, 32 },
		{ 297, 67, 32 },
		{ 297, 68, 32 },
		{ 297, 69, 32 },
		{ 297, 70, 32 },
		{ 297, 71, 32 },
		{ 297, 72, 32 },
		{ 297, 73, 32 },
		{ 297, 74, 32 },
		{ 297, 75, 32 },
		{ 297, 76, 32 },
		{ 297, 77, 32 },
		{ 297, 78, 32 },
		{ 297, 79, 32 },
		{ 297, 80, 32 },
		{ 297, 81, 32 },
		{ 297, 82, 32 },
		{ 297, 83, 32 },
		{ 297, 84, 32 },
		{ 297, 85, 32 },
		{ 297, 86, 32 },
		{ 297, 87, 32 },
		{ 297, 88, 32 },
		{ 297, 89, 32 },
		{ 297, 90, 32 },
		{ 297, 91, 32 },
		{ 297, 92, 32 },
		{ 297, 93, 32 },
		{ 297, 94, 32 },
		{ 297, 95, 32 },
		{ 297, 96, 32 },
		{ 297, 97, 32 },
		{ 297, 98, 32 },
		{ 297, 99, 32 },
		{ 297, 100, 32 },
		{ 297, 101, 32 },
		{ 297, 102, 32 },
		{ 297, 103, 32 },
		{ 297, 104, 32 },
		{ 297, 105, 32 },
		{ 297, 106, 32 },
		{ 297, 107, 32 },
		{ 297, 108, 32 },
		{ 297, 109, 32 },
		{ 297, 110, 32 },
		{ 297, 111, 32 },
		{ 297, 112, 32 },
		{ 297, 113, 32 },
		{ 297, 114, 32 },
		{ 297, 115, 32 },
		{ 297, 116, 32 },
		{ 297, 117, 32 },
		{ 297, 118, 32 },
		{ 297, 119, 32 },
		{ 297, 120, 32 },
		{ 297, 121, 32 },
		{ 297, 122, 32 },
		{ 297, 123, 32 },
		{ 297, 124, 32 },
		{ 297, 125, 32 },
		{ 297, 126, 32 },
		{ 297, 127, 32 },
		{ 297, 128, 32 },
		{ 297, 129, 32 },
		{ 297, 130, 32 },
		{ 297, 131, 32 },
		{ 297, 132, 32 },
		{ 297, 133, 32 },
		{ 297, 134, 32 },
		{ 297, 135, 32 },
		{ 297, 136, 32 },
		{ 297, 137, 32 },
		{ 297, 138, 32 },
		{ 297, 139, 32 },
		{ 297, 140, 32 },
		{ 297, 141, 32 },
		{ 297, 142, 32 },
		{ 297, 143, 32 },
		{ 297, 144, 32 },
		{ 297, 145, 32 },
		{ 297, 146, 32 },
		{ 297, 147, 32 },
		{ 297, 148, 32 },
		{ 297, 149, 32 },
		{ 297, 150, 32 },
		{ 298, 1, 32 },
		{ 298, 2, 32 },
		{ 298, 3, 32 },
		{ 298, 4, 32 },
		{ 298, 5, 32 },
		{ 298, 6, 32 },
		{ 298, 7, 32 },
		{ 298, 8, 32 },
		{ 298, 9, 32 },
		{ 298, 10, 32 },
		{ 298, 11, 32 },
		{ 298, 12, 32 },
		{ 298, 13, 32 },
		{ 298, 14, 32 },
		{ 298, 15, 32 },
		{ 298, 16, 32 },
		{ 298, 17, 32 },
		{ 298, 18, 32 },
		{ 298, 19, 32 },
		{ 298, 20, 32 },
		{ 298, 21, 32 },
		{ 298, 22, 32 },
		{ 298, 23, 32 },
		{ 298, 24, 32 },
		{ 298, 25, 32 },
		{ 298, 26, 32 },
		{ 298, 27, 32 },
		{ 298, 28, 32 },
		{ 298, 29, 32 },
		{ 298, 30, 32 },
		{ 298, 31, 32 },
		{ 298, 32, 32 },
		{ 298, 33, 32 },
		{ 298, 34, 32 },
		{ 298, 35, 32 },
		{ 298, 36, 32 },
		{ 298, 37, 32 },
		{ 298, 38, 32 },
		{ 298, 39, 32 },
		{ 298, 40, 32 },
		{ 298, 41, 32 },
		{ 298, 42, 32 },
		{ 298, 43, 32 },
		{ 298, 44, 32 },
		{ 298, 45, 32 },
		{ 298, 46, 32 },
		{ 298, 47, 32 },
		{ 298, 48, 32 },
		{ 298, 49, 32 },
		{ 298, 50, 32 },
		{ 298, 51, 32 },
		{ 298, 52, 32 },
		{ 298, 53, 32 },
		{ 298, 54, 32 },
		{ 298, 55, 32 },
		{ 298, 56, 32 },
		{ 298, 57, 32 },
		{ 298, 58, 32 },
		{ 298, 59, 32 },
		{ 298, 60, 32 },
		{ 298, 61, 32 },
		{ 298, 62, 32 },
		{ 298, 63, 32 },
		{ 298, 64, 32 },
		{ 298, 65, 32 },
		{ 298, 66, 32 },
		{ 298, 67, 32 },
		{ 298, 68, 32 },
		{ 298, 69, 32 },
		{ 298, 70, 32 },
		{ 298, 71, 32 },
		{ 298, 72, 32 },
		{ 298, 73, 32 },
		{ 298, 74, 32 },
		{ 298, 75, 32 },
		{ 298, 76, 32 },
		{ 298, 77, 32 },
		{ 298, 78, 32 },
		{ 298, 79, 32 },
		{ 298, 80, 32 },
		{ 298, 81, 32 },
		{ 298, 82, 32 },
		{ 298, 83, 32 },
		{ 298, 84, 32 },
		{ 298, 85, 32 },
		{ 298, 86, 32 },
		{ 298, 87, 32 },
		{ 298, 88, 32 },
		{ 298, 89, 32 },
		{ 298, 90, 32 },
		{ 298, 91, 32 },
		{ 298, 92, 32 },
		{ 298, 93, 32 },
		{ 298, 94, 32 },
		{ 298, 95, 32 },
		{ 298, 96, 32 },
		{ 298, 97, 32 },
		{ 298, 98, 32 },
		{ 298, 99, 32 },
		{ 298, 100, 32 },
		{ 298, 101, 32 },
		{ 298, 102, 32 },
		{ 298, 103, 32 },
		{ 298, 104, 32 },
		{ 298, 105, 32 },
		{ 298, 106, 32 },
		{ 298, 107, 32 },
		{ 298, 108, 32 },
		{ 298, 109, 32 },
		{ 298, 110, 32 },
		{ 298, 111, 32 },
		{ 298, 112, 32 },
		{ 298, 113, 32 },
		{ 298, 114, 32 },
		{ 298, 115, 32 },
		{ 298, 116, 32 },
		{ 298, 117, 32 },
		{ 298, 118, 32 },
		{ 298, 119, 32 },
		{ 298, 120, 32 },
		{ 298, 121, 32 },
		{ 298, 122, 32 },
		{ 298, 123, 32 },
		{ 298, 124, 32 },
		{ 298, 125, 32 },
		{ 298, 126, 32 },
		{ 298, 127, 32 },
		{ 298, 128, 32 },
		{ 298, 129, 32 },
		{ 298, 130, 32 },
		{ 298, 131, 32 },
		{ 298, 132, 32 },
		{ 298, 133, 32 },
		{ 298, 134, 32 },
		{ 298, 135, 32 },
		{ 298, 136, 32 },
		{ 298, 137, 32 },
		{ 298, 138, 32 },
		{ 298, 139, 32 },
		{ 298, 140, 32 },
		{ 298, 141, 32 },
		{ 298, 142, 32 },
		{ 298, 143, 32 },
		{ 298, 144, 32 },
		{ 298, 145, 32 },
		{ 298, 146, 32 },
		{ 298, 147, 32 },
		{ 299, 1, 32 },
		{ 299, 2, 32 },
		{ 299, 3, 32 },
		{ 299, 4, 32 },
		{ 299, 5, 32 },
		{ 299, 6, 32 },
		{ 299, 7, 32 },
		{ 299, 8, 32 },
		{ 299, 9, 32 },
		{ 299, 10, 32 },
		{ 299, 11, 32 },
		{ 299, 12, 32 },
		{ 299, 13, 32 },
		{ 299, 14, 32 },
		{ 299, 15, 32 },
		{ 299, 16, 32 },
		{ 299, 17, 32 },
		{ 299, 18, 32 },
		{ 299, 19, 32 },
		{ 299, 20, 32 },
		{ 299, 21, 32 },
		{ 299, 22, 32 },
		{ 299, 23, 32 },
		{ 299, 24, 32 },
		{ 299, 25, 32 },
		{ 299, 26, 32 },
		{ 299, 27, 32 },
		{ 299, 28, 32 },
		{ 299, 29, 32 },
		{ 299, 30, 32 },
		{ 299, 31, 32 },
		{ 299, 32, 32 },
		{ 299, 33, 32 },
		{ 299, 34, 32 },
		{ 299, 35, 32 },
		{ 299, 36, 32 },
		{ 299, 37, 32 },
		{ 299, 38, 32 },
		{ 299, 39, 32 },
		{ 299, 40, 32 },
		{ 299, 41, 32 },
		{ 299, 42, 32 },
		{ 299, 43, 32 },
		{ 299, 44, 32 },
		{ 299, 45, 32 },
		{ 299, 46, 32 },
		{ 299, 47, 32 },
		{ 299, 48, 32 },
		{ 299, 49, 32 },
		{ 299, 50, 32 },
		{ 299, 51, 32 },
		{ 299, 52, 32 },
		{ 299, 53, 32 },
		{ 299, 54, 32 },
		{ 299, 55, 32 },
		{ 299, 56, 32 },
		{ 299, 57, 32 },
		{ 299, 58, 32 },
		{ 299, 59, 32 },
		{ 299, 60, 32 },
		{ 299, 61, 32 },
		{ 299, 62, 32 },
		{ 299, 63, 32 },
		{ 299, 64, 32 },
		{ 299, 65, 32 },
		{ 299, 66, 32 },
		{ 299, 67, 32 },
		{ 299, 68, 32 },
		{ 299, 69, 32 },
		{ 299, 70, 32 },
		{ 299, 71, 32 },
		{ 299, 72, 32 },
		{ 299, 73, 32 },
		{ 299, 74, 32 },
		{ 299, 75, 32 },
		{ 299, 76, 32 },
		{ 299, 77, 32 },
		{ 299, 78, 32 },
		{ 299, 79, 32 },
		{ 299, 80, 32 },
		{ 299, 81, 32 },
		{ 299, 82, 32 },
		{ 299, 83, 32 },
		{ 299, 84, 32 },
		{ 299, 85, 32 },
		{ 299, 86, 32 },
		{ 299, 87, 32 },
		{ 299, 88, 32 },
		{ 299, 89, 32 },
		{ 299, 90, 32 },
		{ 299, 91, 32 },
		{ 299, 92, 32 },
		{ 299, 93, 32 },
		{ 299, 94, 32 },
		{ 299, 95, 32 },
		{ 299, 96, 32 },
		{ 299, 97, 32 },
		{ 299, 98, 32 },
		{ 299, 99, 32 },
		{ 299, 100, 32 },
		{ 299, 101, 32 },
		{ 299, 102, 32 },
		{ 299, 103, 32 },
		{ 299, 104, 32 },
		{ 299, 105, 32 },
		{ 299, 106, 32 },
		{ 299, 107, 32 },
		{ 299, 108, 32 },
		{ 299, 109, 32 },
		{ 299, 110, 32 },
		{ 299, 111, 32 },
		{ 299, 112, 32 },
		{ 299, 113, 32 },
		{ 299, 114, 32 },
		{ 299, 115, 32 },
		{ 299, 116, 32 },
		{ 299, 117, 32 },
		{ 299, 118, 32 },
		{ 299, 119, 32 },
		{ 299, 120, 32 },
		{ 299, 121, 32 },
		{ 299, 122, 32 },
		{ 299, 123, 32 },
		{ 299, 124, 32 },
		{ 299, 125, 32 },
		{ 299, 126, 32 },
		{ 299, 127, 32 },
		{ 299, 128, 32 },
		{ 299, 129, 32 },
		{ 299, 130, 32 },
		{ 299, 131, 32 },
		{ 299, 132, 32 },
		{ 299, 133, 32 },
		{ 299, 134, 32 },
		{ 299, 135, 32 },
		{ 299, 136, 32 },
		{ 299, 137, 32 },
		{ 299, 138, 32 },
		{ 299, 139, 32 },
		{ 299, 140, 32 },
		{ 299, 141, 32 },
		{ 299, 142, 32 },
		{ 299, 143, 32 },
		{ 299, 144, 32 },
		{ 300, 1, 32 },
		{ 300, 2, 32 },
		{ 300, 3, 32 },
		{ 300, 4, 32 },
		{ 300, 5, 32 },
		{ 300, 6, 32 },
		{ 300, 7, 32 },
		{ 300, 8, 32 },
		{ 300, 9, 32 },
		{ 300, 10, 32 },
		{ 300, 11, 32 },
		{ 300, 12, 32 },
		{ 300, 13, 32 },
		{ 300, 14, 32 },
		{ 300, 15, 32 },
		{ 300, 16, 32 },
		{ 300, 17, 32 },
		{ 300, 18, 32 },
		{ 300, 19, 32 },
		{ 300, 20, 32 },
		{ 300, 21, 32 },
		{ 300, 22, 32 },
		{ 300, 23, 32 },
		{ 300, 24, 32 },
		{ 300, 25, 32 },
		{ 300, 26, 32 },
		{ 300, 27, 32 },
		{ 300, 28, 32 },
		{ 300, 29, 32 },
		{ 300, 30, 32 },
		{ 300, 31, 32 },
		{ 300, 32, 32 },
		{ 300, 33, 32 },
		{ 300, 34, 32 },
		{ 300, 35, 32 },
		{ 300, 36, 32 },
		{ 300, 37, 32 },
		{ 300, 38, 32 },
		{ 300, 39, 32 },
		{ 300, 40, 32 },
		{ 300, 41, 32 },
		{ 300, 42, 32 },
		{ 300, 43, 32 },
		{ 300, 44, 32 },
		{ 300, 45, 32 },
		{ 300, 46, 32 },
		{ 300, 47, 32 },
		{ 300, 48, 32 },
		{ 300, 49, 32 },
		{ 300, 50, 32 },
		{ 300, 51, 32 },
		{ 300, 52, 32 },
		{ 300, 53, 32 },
		{ 300, 54, 32 },
		{ 300, 55, 32 },
		{ 300, 56, 32 },
		{ 300, 57, 32 },
		{ 300, 58, 32 },
		{ 300, 59, 32 },
		{ 300, 60, 32 },
		{ 300, 61, 32 },
		{ 300, 62, 32 },
		{ 300, 63, 32 },
		{ 300, 64, 32 },
		{ 300, 65, 32 },
		{ 300, 66, 32 },
		{ 300, 67, 32 },
		{ 300, 68, 32 },
		{ 300, 69, 32 },
		{ 300, 70, 32 },
		{ 300, 71, 32 },
		{ 300, 72, 32 },
		{ 300, 73, 32 },
		{ 300, 74, 32 },
		{ 300, 75, 32 },
		{ 300, 76, 32 },
		{ 300, 77, 32 },
		{ 300, 78, 32 },
		{ 300, 79, 32 },
		{ 300, 80, 32 },
		{ 300, 81, 32 },
		{ 300, 82, 32 },
		{ 300, 83, 32 },
		{ 300, 84, 32 },
		{ 300, 85, 32 },
		{ 300, 86, 32 },
		{ 300, 87, 32 },
		{ 300, 88, 32 },
		{ 300, 89, 32 },
		{ 300, 90, 32 },
		{ 300, 91, 32 },
		{ 300, 92, 32 },
		{ 300, 93, 32 },
		{ 300, 94, 32 },
		{ 300, 95, 32 },
		{ 300, 96, 32 },
		{ 300, 97, 32 },
		{ 300, 98, 32 },
		{ 300, 99, 32 },
		{ 300, 100, 32 },
		{ 300, 101, 32 },
		{ 300, 102, 32 },
		{ 300, 103, 32 },
		{ 300, 104, 32 },
		{ 300, 105, 32 },
		{ 300, 106, 32 },
		{ 300, 107, 32 },
		{ 300, 108, 32 },
		{ 300, 109, 32 },
		{ 300, 110, 32 },
		{ 300, 111, 32 },
		{ 300, 112, 32 },
		{ 300, 113, 32 },
		{ 300, 114, 32 },
		{ 300, 115, 32 },
		{ 300, 116, 32 },
		{ 300, 117, 32 },
		{ 300, 118, 32 },
		{ 300, 119, 32 },
		{ 300, 120, 32 },
		{ 300, 121, 32 },
		{ 300, 122, 32 },
		{ 300, 123, 32 },
		{ 300, 124, 32 },
		{ 300, 125, 32 },
		{ 300, 126, 32 },
		{ 300, 127, 32 },
		{ 300, 128, 32 },
		{ 300, 129, 32 },
		{ 300, 130, 32 },
		{ 300, 131, 32 },
		{ 300, 132, 32 },
		{ 300, 133, 32 },
		{ 300, 134, 32 },
		{ 300, 135, 32 },
		{ 300, 136, 32 },
		{ 300, 137, 32 },
		{ 300, 138, 32 },
		{ 300, 139, 32 },
		{ 300, 140, 32 },
		{ 300, 141, 32 },
		{ 300, 142, 32 },
		{ 301, 1, 32 },
		{ 301, 2, 32 },
		{ 301, 3, 32 },
		{ 301, 4, 32 },
		{ 301, 5, 32 },
		{ 301, 6, 32 },
		{ 301, 7, 32 },
		{ 301, 8, 32 },
		{ 301, 9, 32 },
		{ 301, 10, 32 },
		{ 301, 11, 32 },
		{ 301, 12, 32 },
		{ 301, 13, 32 },
		{ 301, 14, 32 },
		{ 301, 15, 32 },
		{ 301, 16, 32 },
		{ 301, 17, 32 },
		{ 301, 18, 32 },
		{ 301, 19, 32 },
		{ 301, 20, 32 },
		{ 301, 21, 32 },
		{ 301, 22, 32 },
		{ 301, 23, 32 },
		{ 301, 24, 32 },
		{ 301, 25, 32 },
		{ 301, 26, 32 },
		{ 301, 27, 32 },
		{ 301, 28, 32 },
		{ 301, 29, 32 },
		{ 301, 30, 32 },
		{ 301, 31, 32 },
		{ 301, 32, 32 },
		{ 301, 33, 32 },
		{ 301, 34, 32 },
		{ 301, 35, 32 },
		{ 301, 36, 32 },
		{ 301, 37, 32 },
		{ 301, 38, 32 },
		{ 301, 39, 32 },
		{ 301, 40, 32 },
		{ 301, 41, 32 },
		{ 301, 42, 32 },
		{ 301, 43, 32 },
		{ 301, 44, 32 },
		{ 301, 45, 32 },
		{ 301, 46, 32 },
		{ 301, 47, 32 },
		{ 301, 48, 32 },
		{ 301, 49, 32 },
		{ 301, 50, 32 },
		{ 301, 51, 32 },
		{ 301, 52, 32 },
		{ 301, 53, 32 },
		{ 301, 54, 32 },
		{ 301, 55, 32 },
		{ 301, 56, 32 },
		{ 301, 57, 32 },
		{ 301, 58, 32 },
		{ 301, 59, 32 },
		{ 301, 60, 32 },
		{ 301, 61, 32 },
		{ 301, 62, 32 },
		{ 301, 63, 32 },
		{ 301, 64, 32 },
		{ 301, 65, 32 },
		{ 301, 66, 32 },
		{ 301, 67, 32 },
		{ 301, 68, 32 },
		{ 301, 69, 32 },
		{ 301, 70, 32 },
		{ 301, 71, 32 },
		{ 301, 72, 32 },
		{ 301, 73, 32 },
		{ 301, 74, 32 },
		{ 301, 75, 32 },
		{ 301, 76, 32 },
		{ 301, 77, 32 },
		{ 301, 78, 32 },
		{ 301, 79, 32 },
		{ 301, 80, 32 },
		{ 301, 81, 32 },
		{ 301, 82, 32 },
		{ 301, 83, 32 },
		{ 301, 84, 32 },
		{ 301, 85, 32 },
		{ 301, 86, 32 },
		{ 301, 87, 32 },
		{ 301, 88, 32 },
		{ 301, 89, 32 },
		{ 301, 90, 32 },
		{ 301, 91, 32 },
		{ 301, 92, 32 },
		{ 301, 93, 32 },
		{ 301, 94, 32 },
		{ 301, 95, 32 },
		{ 301, 96, 32 },
		{ 301, 97, 32 },
		{ 301, 98, 32 },
		{ 301, 99, 32 },
		{ 301, 100, 32 },
		{ 301, 101, 32 },
		{ 301, 102, 32 },
		{ 301, 103, 32 },
		{ 301, 104, 32 },
		{ 301, 105, 32 },
		{ 301, 106, 32 },
		{ 301, 107, 32 },
		{ 301, 108, 32 },
		{ 301, 109, 32 },
		{ 301, 110, 32 },
		{ 301, 111, 32 },
		{ 301, 112, 32 },
		{ 301, 113, 32 },
		{ 301, 114, 32 },
		{ 301, 115, 32 },
		{ 301, 116, 32 },
		{ 301, 117, 32 },
		{ 301, 118, 32 },
		{ 301, 119, 32 },
		{ 301, 120, 32 },
		{ 301, 121, 32 },
		{ 301, 122, 32 },
		{ 301, 123, 32 },
		{ 301, 124, 32 },
		{ 301, 125, 32 },
		{ 301, 126, 32 },
		{ 301, 127, 32 },
		{ 301, 128, 32 },
		{ 301, 129, 32 },
		{ 301, 130, 32 },
		{ 301, 131, 32 },
		{ 301, 132, 32 },
		{ 301, 133, 32 },
		{ 301, 134, 32 },
		{ 301, 135, 32 },
		{ 301, 136, 32 },
		{ 301, 137, 32 },
		{ 301, 138, 32 },
		{ 301, 139, 32 },
		{ 302, 1, 32 },
		{ 302, 2, 32 },
		{ 302, 3, 32 },
		{ 302, 4, 32 },
		{ 302, 5, 32 },
		{ 302, 6, 32 },
		{ 302, 7, 32 },
		{ 302, 8, 32 },
		{ 302, 9, 32 },
		{ 302, 10, 32 },
		{ 302, 11, 32 },
		{ 302, 12, 32 },
		{ 302, 13, 32 },
		{ 302, 14, 32 },
		{ 302, 15, 32 },
		{ 302, 16, 32 },
		{ 302, 17, 32 },
		{ 302, 18, 32 },
		{ 302, 19, 32 },
		{ 302, 20, 32 },
		{ 302, 21, 32 },
		{ 302, 22, 32 },
		{ 302, 23, 32 },
		{ 302, 24, 32 },
		{ 302, 25, 32 },
		{ 302, 26, 32 },
		{ 302, 27, 32 },
		{ 302, 28, 32 },
		{ 302, 29, 32 },
		{ 302, 30, 32 },
		{ 302, 31, 32 },
		{ 302, 32, 32 },
		{ 302, 33, 32 },
		{ 302, 34, 32 },
		{ 302, 35, 32 },
		{ 302, 36, 32 },
		{ 302, 37, 32 },
		{ 302, 38, 32 },
		{ 302, 39, 32 },
		{ 302, 40, 32 },
		{ 302, 41, 32 },
		{ 302, 42, 32 },
		{ 302, 43, 32 },
		{ 302, 44, 32 },
		{ 302, 45, 32 },
		{ 302, 46, 32 },
		{ 302, 47, 32 },
		{ 302, 48, 32 },
		{ 302, 49, 32 },
		{ 302, 50, 32 },
		{ 302, 51, 32 },
		{ 302, 52, 32 },
		{ 302, 53, 32 },
		{ 302, 54, 32 },
		{ 302, 55, 32 },
		{ 302, 56, 32 },
		{ 302, 57, 32 },
		{ 302, 58, 32 },
		{ 302, 59, 32 },
		{ 302, 60, 32 },
		{ 302, 61, 32 },
		{ 302, 62, 32 },
		{ 302, 63, 32 },
		{ 302, 64, 32 },
		{ 302, 65, 32 },
		{ 302, 66, 32 },
		{ 302, 67, 32 },
		{ 302, 68, 32 },
		{ 302, 69, 32 },
		{ 302, 70, 32 },
		{ 302, 71, 32 },
		{ 302, 72, 32 },
		{ 302, 73, 32 },
		{ 302, 74, 32 },
		{ 302, 75, 32 },
		{ 302, 76, 32 },
		{ 302, 77, 32 },
		{ 302, 78, 32 },
		{ 302, 79, 32 },
		{ 302, 80, 32 },
		{ 302, 81, 32 },
		{ 302, 82, 32 },
		{ 302, 83, 32 },
		{ 302, 84, 32 },
		{ 302, 85, 32 },
		{ 302, 86, 32 },
		{ 302, 87, 32 },
		{ 302, 88, 32 },
		{ 302, 89, 32 },
		{ 302, 90, 32 },
		{ 302, 91, 32 },
		{ 302, 92, 32 },
		{ 302, 93, 32 },
		{ 302, 94, 32 },
		{ 302, 95, 32 },
		{ 302, 96, 32 },
		{ 302, 97, 32 },
		{ 302, 98, 32 },
		{ 302, 99, 32 },
		{ 302, 100, 32 },
		{ 302, 101, 32 },
		{ 302, 102, 32 },
		{ 302, 103, 32 },
		{ 302, 104, 32 },
		{ 302, 105, 32 },
		{ 302, 106, 32 },
		{ 302, 107, 32 },
		{ 302, 108, 32 },
		{ 302, 109, 32 },
		{ 302, 110, 32 },
		{ 302, 111, 32 },
		{ 302, 112, 32 },
		{ 302, 113, 32 },
		{ 302, 114, 32 },
		{ 302, 115, 32 },
		{ 302, 116, 32 },
		{ 302, 117, 32 },
		{ 302, 118, 32 },
		{ 302, 119, 32 },
		{ 302, 120, 32 },
		{ 302, 121, 32 },
		{ 302, 122, 32 },
		{ 302, 123, 32 },
		{ 302, 124, 32 },
		{ 302, 125, 32 },
		{ 302, 126, 32 },
		{ 302, 127, 32 },
		{ 302, 128, 32 },
		{ 302, 129, 32 },
		{ 302, 130, 32 },
		{ 302, 131, 32 },
		{ 302, 132, 32 },
		{ 302, 133, 32 },
		{ 302, 134, 32 },
		{ 302, 135, 32 },
		{ 302, 136, 32 },
		{ 303, 1, 32 },
		{ 303, 2, 32 },
		{ 303, 3, 32 },
		{ 303, 4, 32 },
		{ 303, 5, 32 },
		{ 303, 6, 32 },
		{ 303, 7, 32 },
		{ 303, 8, 32 },
		{ 303, 9, 32 },
		{ 303, 10, 32 },
		{ 303, 11, 32 },
		{ 303, 12, 32 },
		{ 303, 13, 32 },
		{ 303, 14, 32 },
		{ 303, 15, 32 },
		{ 303, 16, 32 },
		{ 303, 17, 32 },
		{ 303, 18, 32 },
		{ 303, 19, 32 },
		{ 303, 20, 32 },
		{ 303, 21, 32 },
		{ 303, 22, 32 },
		{ 303, 23, 32 },
		{ 303, 24, 32 },
		{ 303, 25, 32 },
		{ 303, 26, 32 },
		{ 303, 27, 32 },
		{ 303, 28, 32 },
		{ 303, 29, 32 },
		{ 303, 30, 32 },
		{ 303, 31, 32 },
		{ 303, 32, 32 },
		{ 303, 33, 32 },
		{ 303, 34, 32 },
		{ 303, 35, 32 },
		{ 303, 36, 32 },
		{ 303, 37, 32 },
		{ 303, 38, 32 },
		{ 303, 39, 32 },
		{ 303, 40, 32 },
		{ 303, 41, 32 },
		{ 303, 42, 32 },
		{ 303, 43, 32 },
		{ 303, 44, 32 },
		{ 303, 45, 32 },
		{ 303, 46, 32 },
		{ 303, 47, 32 },
		{ 303, 48, 32 },
		{ 303, 49, 32 },
		{ 303, 50, 32 },
		{ 303, 51, 32 },
		{ 303, 52, 32 },
		{ 303, 53, 32 },
		{ 303, 54, 32 },
		{ 303, 55, 32 },
		{ 303, 56, 32 },
		{ 303, 57, 32 },
		{ 303, 58, 32 },
		{ 303, 59, 32 },
		{ 303, 60, 32 },
		{ 303, 61, 32 },
		{ 303, 62, 32 },
		{ 303, 63, 32 },
		{ 303, 64, 32 },
		{ 303, 65, 32 },
		{ 303, 66, 32 },
		{ 303, 67, 32 },
		{ 303, 68, 32 },
		{ 303, 69, 32 },
		{ 303, 70, 32 },
		{ 303, 71, 32 },
		{ 303, 72, 32 },
		{ 303, 73, 32 },
		{ 303, 74, 32 },
		{ 303, 75, 32 },
		{ 303, 76, 32 },
		{ 303, 77, 32 },
		{ 303, 78, 32 },
		{ 303, 79, 32 },
		{ 303, 80, 32 },
		{ 303, 81, 32 },
		{ 303, 82, 32 },
		{ 303, 83, 32 },
		{ 303, 84, 32 },
		{ 303, 85, 32 },
		{ 303, 86, 32 },
		{ 303, 87, 32 },
		{ 303, 88, 32 },
		{ 303, 89, 32 },
		{ 303, 90, 32 },
		{ 303, 91, 32 },
		{ 303, 92, 32 },
		{ 303, 93, 32 },
		{ 303, 94, 32 },
		{ 303, 95, 32 },
		{ 303, 96, 32 },
		{ 303, 97, 32 },
		{ 303, 98, 32 },
		{ 303, 99, 32 },
		{ 303, 100, 32 },
		{ 303, 101, 32 },
		{ 303, 102, 32 },
		{ 303, 103, 32 },
		{ 303, 104, 32 },
		{ 303, 105, 32 },
		{ 303, 106, 32 },
		{ 303, 107, 32 },
		{ 303, 108, 32 },
		{ 303, 109, 32 },
		{ 303, 110, 32 },
		{ 303, 111, 32 },
		{ 303, 112, 32 },
		{ 303, 113, 32 },
		{ 303, 114, 32 },
		{ 303, 115, 32 },
		{ 303, 116, 32 },
		{ 303, 117, 32 },
		{ 303, 118, 32 },
		{ 303, 119, 32 },
		{ 303, 120, 32 },
		{ 303, 121, 32 },
		{ 303, 122, 32 },
		{ 303, 123, 32 },
		{ 303, 124, 32 },
		{ 303, 125, 32 },
		{ 303, 126, 32 },
		{ 303, 127, 32 },
		{ 303, 128, 32 },
		{ 303, 129, 32 },
		{ 303, 130, 32 },
		{ 303, 131, 32 },
		{ 303, 132, 32 },
		{ 303, 133, 32 },
		{ 303, 134, 32 },
		{ 304, 1, 32 },
		{ 304, 2, 32 },
		{ 304, 3, 32 },
		{ 304, 4, 32 },
		{ 304, 5, 32 },
		{ 304, 6, 32 },
		{ 304, 7, 32 },
		{ 304, 8, 32 },
		{ 304, 9, 32 },
		{ 304, 10, 32 },
		{ 304, 11, 32 },
		{ 304, 12, 32 },
		{ 304, 13, 32 },
		{ 304, 14, 32 },
		{ 304, 15, 32 },
		{ 304, 16, 32 },
		{ 304, 17, 32 },
		{ 304, 18, 32 },
		{ 304, 19, 32 },
		{ 304, 20, 32 },
		{ 304, 21, 32 },
		{ 304, 22, 32 },
		{ 304, 23, 32 },
		{ 304, 24, 32 },
		{ 304, 25, 32 },
		{ 304, 26, 32 },
		{ 304, 27, 32 },
		{ 304, 28, 32 },
		{ 304, 29, 32 },
		{ 304, 30, 32 },
		{ 304, 31, 32 },
		{ 304, 32, 32 },
		{ 304, 33, 32 },
		{ 304, 34, 32 },
		{ 304, 35, 32 },
		{ 304, 36, 32 },
		{ 304, 37, 32 },
		{ 304, 38, 32 },
		{ 304, 39, 32 },
		{ 304, 40, 32 },
		{ 304, 41, 32 },
		{ 304, 42, 32 },
		{ 304, 43, 32 },
		{ 304, 44, 32 },
		{ 304, 45, 32 },
		{ 304, 46, 32 },
		{ 304, 47, 32 },
		{ 304, 48, 32 },
		{ 304, 49, 32 },
		{ 304, 50, 32 },
		{ 304, 51, 32 },
		{ 304, 52, 32 },
		{ 304, 53, 32 },
		{ 304, 54, 32 },
		{ 304, 55, 32 },
		{ 304, 56, 32 },
		{ 304, 57, 32 },
		{ 304, 58, 32 },
		{ 304, 59, 32 },
		{ 304, 60, 32 },
		{ 304, 61, 32 },
		{ 304, 62, 32 },
		{ 304, 63, 32 },
		{ 304, 64, 32 },
		{ 304, 65, 32 },
		{ 304, 66, 32 },
		{ 304, 67, 32 },
		{ 304, 68, 32 },
		{ 304, 69, 32 },
		{ 304, 70, 32 },
		{ 304, 71, 32 },
		{ 304, 72, 32 },
		{ 304, 73, 32 },
		{ 304, 74, 32 },
		{ 304, 75, 32 },
		{ 304, 76, 32 },
		{ 304, 77, 32 },
		{ 304, 78, 32 },
		{ 304, 79, 32 },
		{ 304, 80, 32 },
		{ 304, 81, 32 },
		{ 304, 82, 32 },
		{ 304, 83, 32 },
		{ 304, 84, 32 },
		{ 304, 85, 32 },
		{ 304, 86, 32 },
		{ 304, 87, 32 },
		{ 304, 88, 32 },
		{ 304, 89, 32 },
		{ 304, 90, 32 },
		{ 304, 91, 32 },
		{ 304, 92, 32 },
		{ 304, 93, 32 },
		{ 304, 94, 32 },
		{ 304, 95, 32 },
		{ 304, 96, 32 },
		{ 304, 97, 32 },
		{ 304, 98, 32 },
		{ 304, 99, 32 },
		{ 304, 100, 32 },
		{ 304, 101, 32 },
		{ 304, 102, 32 },
		{ 304, 103, 32 },
		{ 304, 104, 32 },
		{ 304, 105, 32 },
		{ 304, 106, 32 },
		{ 304, 107, 32 },
		{ 304, 108, 32 },
		{ 304, 109, 32 },
		{ 304, 110, 32 },
		{ 304, 111, 32 },
		{ 304, 112, 32 },
		{ 304, 113, 32 },
		{ 304, 114, 32 },
		{ 304, 115, 32 },
		{ 304, 116, 32 },
		{ 304, 117, 32 },
		{ 304, 118, 32 },
		{ 304, 119, 32 },
		{ 304, 120, 32 },
		{ 304, 121, 32 },
		{ 304, 122, 32 },
		{ 304, 123, 32 },
		{ 304, 124, 32 },
		{ 304, 125, 32 },
		{ 304, 126, 32 },
		{ 304, 127, 32 },
		{ 304, 128, 32 },
		{ 304, 129, 32 },
		{ 304, 130, 32 },
		{ 304, 131, 32 },
		{ 305, 1, 32 },
		{ 305, 2, 32 },
		{ 305, 3, 32 },
		{ 305, 4, 32 },
		{ 305, 5, 32 },
		{ 305, 6, 32 },
		{ 305, 7, 32 },
		{ 305, 8, 32 },
		{ 305, 9, 32 },
		{ 305, 10, 32 },
		{ 305, 11, 32 },
		{ 305, 12, 32 },
		{ 305, 13, 32 },
		{ 305, 14, 32 },
		{ 305, 15, 32 },
		{ 305, 16, 32 },
		{ 305, 17, 32 },
		{ 305, 18, 32 },
		{ 305, 19, 32 },
		{ 305, 20, 32 },
		{ 305, 21, 32 },
		{ 305, 22, 32 },
		{ 305, 23, 32 },
		{ 305, 24, 32 },
		{ 305, 25, 32 },
		{ 305, 26, 32 },
		{ 305, 27, 32 },
		{ 305, 28, 32 },
		{ 305, 29, 32 },
		{ 305, 30, 32 },
		{ 305, 31, 32 },
		{ 305, 32, 32 },
		{ 305, 33, 32 },
		{ 305, 34, 32 },
		{ 305, 35, 32 },
		{ 305, 36, 32 },
		{ 305, 37, 32 },
		{ 305, 38, 32 },
		{ 305, 39, 32 },
		{ 305, 40, 32 },
		{ 305, 41, 32 },
		{ 305, 42, 32 },
		{ 305, 43, 32 },
		{ 305, 44, 32 },
		{ 305, 45, 32 },
		{ 305, 46, 32 },
		{ 305, 47, 32 },
		{ 305, 48, 32 },
		{ 305, 49, 32 },
		{ 305, 50, 32 },
		{ 305, 51, 32 },
		{ 305, 52, 32 },
		{ 305, 53, 32 },
		{ 305, 54, 32 },
		{ 305, 55, 32 },
		{ 305, 56, 32 },
		{ 305, 57, 32 },
		{ 305, 58, 32 },
		{ 305, 59, 32 },
		{ 305, 60, 32 },
		{ 305, 61, 32 },
		{ 305, 62, 32 },
		{ 305, 63, 32 },
		{ 305, 64, 32 },
		{ 305, 65, 32 },
		{ 305, 66, 32 },
		{ 305, 67, 32 },
		{ 305, 68, 32 },
		{ 305, 69, 32 },
		{ 305, 70, 32 },
		{ 305, 71, 32 },
		{ 305, 72, 32 },
		{ 305, 73, 32 },
		{ 305, 74, 32 },
		{ 305, 75, 32 },
		{ 305, 76, 32 },
		{ 305, 77, 32 },
		{ 305, 78, 32 },
		{ 305, 79, 32 },
		{ 305, 80, 32 },
		{ 305, 81, 32 },
		{ 305, 82, 32 },
		{ 305, 83, 32 },
		{ 305, 84, 32 },
		{ 305, 85, 32 },
		{ 305, 86, 32 },
		{ 305, 87, 32 },
		{ 305, 88, 32 },
		{ 305, 89, 32 },
		{ 305, 90, 32 },
		{ 305, 91, 32 },
		{ 305, 92, 32 },
		{ 305, 93, 32 },
		{ 305, 94, 32 },
		{ 305, 95, 32 },
		{ 305, 96, 32 },
		{ 305, 97, 32 },
		{ 305, 98, 32 },
		{ 305, 99, 32 },
		{ 305, 100, 32 },
		{ 305, 101, 32 },
		{ 305, 102, 32 },
		{ 305, 103, 32 },
		{ 305, 104, 32 },
		{ 305, 105, 32 },
		{ 305, 106, 32 },
		{ 305, 107, 32 },
		{ 305, 108, 32 },
		{ 305, 109, 32 },
		{ 305, 110, 32 },
		{ 305, 111, 32 },
		{ 305, 112, 32 },
		{ 305, 113, 32 },
		{ 305, 114, 32 },
		{ 305, 115, 32 },
		{ 305, 116, 32 },
		{ 305, 117, 32 },
		{ 305, 118, 32 },
		{ 305, 119, 32 },
		{ 305, 120, 32 },
		{ 305, 121, 32 },
		{ 305, 122, 32 },
		{ 305, 123, 32 },
		{ 305, 124, 32 },
		{ 305, 125, 32 },
		{ 305, 126, 32 },
		{ 305, 127, 32 },
		{ 305, 128, 32 },
		{ 306, 1, 32 },
		{ 306, 2, 32 },
		{ 306, 3, 32 },
		{ 306, 4, 32 },
		{ 306, 5, 32 },
		{ 306, 6, 32 },
		{ 306, 7, 32 },
		{ 306, 8, 32 },
		{ 306, 9, 32 },
		{ 306, 10, 32 },
		{ 306, 11, 32 },
		{ 306, 12, 32 },
		{ 306, 13, 32 },
		{ 306, 14, 32 },
		{ 306, 15, 32 },
		{ 306, 16, 32 },
		{ 306, 17, 32 },
		{ 306, 18, 32 },
		{ 306, 19, 32 },
		{ 306, 20, 32 },
		{ 306, 21, 32 },
		{ 306, 22, 32 },
		{ 306, 23, 32 },
		{ 306, 24, 32 },
		{ 306, 25, 32 },
		{ 306, 26, 32 },
		{ 306, 27, 32 },
		{ 306, 28, 32 },
		{ 306, 29, 32 },
		{ 306, 30, 32 },
		{ 306, 31, 32 },
		{ 306, 32, 32 },
		{ 306, 33, 32 },
		{ 306, 34, 32 },
		{ 306, 35, 32 },
		{ 306, 36, 32 },
		{ 306, 37, 32 },
		{ 306, 38, 32 },
		{ 306, 39, 32 },
		{ 306, 40, 32 },
		{ 306, 41, 32 },
		{ 306, 42, 32 },
		{ 306, 43, 32 },
		{ 306, 44, 32 },
		{ 306, 45, 32 },
		{ 306, 46, 32 },
		{ 306, 47, 32 },
		{ 306, 48, 32 },
		{ 306, 49, 32 },
		{ 306, 50, 32 },
		{ 306, 51, 32 },
		{ 306, 52, 32 },
		{ 306, 53, 32 },
		{ 306, 54, 32 },
		{ 306, 55, 32 },
		{ 306, 56, 32 },
		{ 306, 57, 32 },
		{ 306, 58, 32 },
		{ 306, 59, 32 },
		{ 306, 60, 32 },
		{ 306, 61, 32 },
		{ 306, 62, 32 },
		{ 306, 63, 32 },
		{ 306, 64, 32 },
		{ 306, 65, 32 },
		{ 306, 66, 32 },
		{ 306, 67, 32 },
		{ 306, 68, 32 },
		{ 306, 69, 32 },
		{ 306, 70, 32 },
		{ 306, 71, 32 },
		{ 306, 72, 32 },
		{ 306, 73, 32 },
		{ 306, 74, 32 },
		{ 306, 75, 32 },
		{ 306, 76, 32 },
		{ 306, 77, 32 },
		{ 306, 78, 32 },
		{ 306, 79, 32 },
		{ 306, 80, 32 },
		{ 306, 81, 32 },
		{ 306, 82, 32 },
		{ 306, 83, 32 },
		{ 306, 84, 32 },
		{ 306, 85, 32 },
		{ 306, 86, 32 },
		{ 306, 87, 32 },
		{ 306, 88, 32 },
		{ 306, 89, 32 },
		{ 306, 90, 32 },
		{ 306, 91, 32 },
		{ 306, 92, 32 },
		{ 306, 93, 32 },
		{ 306, 94, 32 },
		{ 306, 95, 32 },
		{ 306, 96, 32 },
		{ 306, 97, 32 },
		{ 306, 98, 32 },
		{ 306, 99, 32 },
		{ 306, 100, 32 },
		{ 306, 101, 32 },
		{ 306, 102, 32 },
		{ 306, 103, 32 },
		{ 306, 104, 32 },
		{ 306, 105, 32 },
		{ 306, 106, 32 },
		{ 306, 107, 32 },
		{ 306, 108, 32 },
		{ 306, 109, 32 },
		{ 306, 110, 32 },
		{ 306, 111, 32 },
		{ 306, 112, 32 },
		{ 306, 113, 32 },
		{ 306, 114, 32 },
		{ 306, 115, 32 },
		{ 306, 116, 32 },
		{ 306, 117, 32 },
		{ 306, 118, 32 },
		{ 306, 119, 32 },
		{ 306, 120, 32 },
		{ 306, 121, 32 },
		{ 306, 122, 32 },
		{ 306, 123, 32 },
		{ 306, 124, 32 },
		{ 306, 125, 32 },
		{ 306, 126, 32 },
		{ 307, 1, 32 },
		{ 307, 2, 32 },
		{ 307, 3, 32 },
		{ 307, 4, 32 },
		{ 307, 5, 32 },
		{ 307, 6, 32 },
		{ 307, 7, 32 },
		{ 307, 8, 32 },
		{ 307, 9, 32 },
		{ 307, 10, 32 },
		{ 307, 11, 32 },
		{ 307, 12, 32 },
		{ 307, 13, 32 },
		{ 307, 14, 32 },
		{ 307, 15, 32 },
		{ 307, 16, 32 },
		{ 307, 17, 32 },
		{ 307, 18, 32 },
		{ 307, 19, 32 },
		{ 307, 20, 32 },
		{ 307, 21, 32 },
		{ 307, 22, 32 },
		{ 307, 23, 32 },
		{ 307, 24, 32 },
		{ 307, 25, 32 },
		{ 307, 26, 32 },
		{ 307, 27, 32 },
		{ 307, 28, 32 },
		{ 307, 29, 32 },
		{ 307, 30, 32 },
		{ 307, 31, 32 },
		{ 307, 32, 32 },
		{ 307, 33, 32 },
		{ 307, 34, 32 },
		{ 307, 35, 32 },
		{ 307, 36, 32 },
		{ 307, 37, 32 },
		{ 307, 38, 32 },
		{ 307, 39, 32 },
		{ 307, 40, 32 },
		{ 307, 41, 32 },
		{ 307, 42, 32 },
		{ 307, 43, 32 },
		{ 307, 44, 32 },
		{ 307, 45, 32 },
		{ 307, 46, 32 },
		{ 307, 47, 32 },
		{ 307, 48, 32 },
		{ 307, 49, 32 },
		{ 307, 50, 32 },
		{ 307, 51, 32 },
		{ 307, 52, 32 },
		{ 307, 53, 32 },
		{ 307, 54, 32 },
		{ 307, 55, 32 },
		{ 307, 56, 32 },
		{ 307, 57, 32 },
		{ 307, 58, 32 },
		{ 307, 59, 32 },
		{ 307, 60, 32 },
		{ 307, 61, 32 },
		{ 307, 62, 32 },
		{ 307, 63, 32 },
		{ 307, 64, 32 },
		{ 307, 65, 32 },
		{ 307, 66, 32 },
		{ 307, 67, 32 },
		{ 307, 68, 32 },
		{ 307, 69, 32 },
		{ 307, 70, 32 },
		{ 307, 71, 32 },
		{ 307, 72, 32 },
		{ 307, 73, 32 },
		{ 307, 74, 32 },
		{ 307, 75, 32 },
		{ 307, 76, 32 },
		{ 307, 77, 32 },
		{ 307, 78, 32 },
		{ 307, 79, 32 },
		{ 307, 80, 32 },
		{ 307, 81, 32 },
		{ 307, 82, 32 },
		{ 307, 83, 32 },
		{ 307, 84, 32 },
		{ 307, 85, 32 },
		{ 307, 86, 32 },
		{ 307, 87, 32 },
		{ 307, 88, 32 },
		{ 307, 89, 32 },
		{ 307, 90, 32 },
		{ 307, 91, 32 },
		{ 307, 92, 32 },
		{ 307, 93, 32 },
		{ 307, 94, 32 },
		{ 307, 95, 32 },
		{ 307, 96, 32 },
		{ 307, 97, 32 },
		{ 307, 98, 32 },
		{ 307, 99, 32 },
		{ 307, 100, 32 },
		{ 307, 101, 32 },
		{ 307, 102, 32 },
		{ 307, 103, 32 },
		{ 307, 104, 32 },
		{ 307, 105, 32 },
		{ 307, 106, 32 },
		{ 307, 107, 32 },
		{ 307, 108, 32 },
		{ 307, 109, 32 },
		{ 307, 110, 32 },
		{ 307, 111, 32 },
		{ 307, 112, 32 },
		{ 307, 113, 32 },
		{ 307, 114, 32 },
		{ 307, 115, 32 },
		{ 307, 116, 32 },
		{ 307, 117, 32 },
		{ 307, 118, 32 },
		{ 307, 119, 32 },
		{ 307, 120, 32 },
		{ 307, 121, 32 },
		{ 307, 122, 32 },
		{ 307, 123, 32 },
		{ 308, 1, 32 },
		{ 308, 2, 32 },
		{ 308, 3, 32 },
		{ 308, 4, 32 },
		{ 308, 5, 32 },
		{ 308, 6, 32 },
		{ 308, 7, 32 },
		{ 308, 8, 32 },
		{ 308, 9, 32 },
		{ 308, 10, 32 },
		{ 308, 11, 32 },
		{ 308, 12, 32 },
		{ 308, 13, 32 },
		{ 308, 14, 32 },
		{ 308, 15, 32 },
		{ 308, 16, 32 },
		{ 308, 17, 32 },
		{ 308, 18, 32 },
		{ 308, 19, 32 },
		{ 308, 20, 32 },
		{ 308, 21, 32 },
		{ 308, 22, 32 },
		{ 308, 23, 32 },
		{ 308, 24, 32 },
		{ 308, 25, 32 },
		{ 308, 26, 32 },
		{ 308, 27, 32 },
		{ 308, 28, 32 },
		{ 308, 29, 32 },
		{ 308, 30, 32 },
		{ 308, 31, 32 },
		{ 308, 32, 32 },
		{ 308, 33, 32 },
		{ 308, 34, 32 },
		{ 308, 35, 32 },
		{ 308, 36, 32 },
		{ 308, 37, 32 },
		{ 308, 38, 32 },
		{ 308, 39, 32 },
		{ 308, 40, 32 },
		{ 308, 41, 32 },
		{ 308, 42, 32 },
		{ 308, 43, 32 },
		{ 308, 44, 32 },
		{ 308, 45, 32 },
		{ 308, 46, 32 },
		{ 308, 47, 32 },
		{ 308, 48, 32 },
		{ 308, 49, 32 },
		{ 308, 50, 32 },
		{ 308, 51, 32 },
		{ 308, 52, 32 },
		{ 308, 53, 32 },
		{ 308, 54, 32 },
		{ 308, 55, 32 },
		{ 308, 56, 32 },
		{ 308, 57, 32 },
		{ 308, 58, 32 },
		{ 308, 59, 32 },
		{ 308, 60, 32 },
		{ 308, 61, 32 },
		{ 308, 62, 32 },
		{ 308, 63, 32 },
		{ 308, 64, 32 },
		{ 308, 65, 32 },
		{ 308, 66, 32 },
		{ 308, 67, 32 },
		{ 308, 68, 32 },
		{ 308, 69, 32 },
		{ 308, 70, 32 },
		{ 308, 71, 32 },
		{ 308, 72, 32 },
		{ 308, 73, 32 },
		{ 308, 74, 32 },
		{ 308, 75, 32 },
		{ 308, 76, 32 },
		{ 308, 77, 32 },
		{ 308, 78, 32 },
		{ 308, 79, 32 },
		{ 308, 80, 32 },
		{ 308, 81, 32 },
		{ 308, 82, 32 },
		{ 308, 83, 32 },
		{ 308, 84, 32 },
		{ 308, 85, 32 },
		{ 308, 86, 32 },
		{ 308, 87, 32 },
		{ 308, 88, 32 },
		{ 308, 89, 32 },
		{ 308, 90, 32 },
		{ 308, 91, 32 },
		{ 308, 92, 32 },
		{ 308, 93, 32 },
		{ 308, 94, 32 },
		{ 308, 95, 32 },
		{ 308, 96, 32 },
		{ 308, 97, 32 },
		{ 308, 98, 32 },
		{ 308, 99, 32 },
		{ 308, 100, 32 },
		{ 308, 101, 32 },
		{ 308, 102, 32 },
		{ 308, 103, 32 },
		{ 308, 104, 32 },
		{ 308, 105, 32 },
		{ 308, 106, 32 },
		{ 308, 107, 32 },
		{ 308, 108, 32 },
		{ 308, 109, 32 },
		{ 308, 110, 32 },
		{ 308, 111, 32 },
		{ 308, 112, 32 },
		{ 308, 113, 32 },
		{ 308, 114, 32 },
		{ 308, 115, 32 },
		{ 308, 116, 32 },
		{ 308, 117, 32 },
		{ 308, 118, 32 },
		{ 308, 119, 32 },
		{ 308, 120, 32 },
		{ 309, 1, 32 },
		{ 309, 2, 32 },
		{ 309, 3, 32 },
		{ 309, 4, 32 },
		{ 309, 5, 32 },
		{ 309, 6, 32 },
		{ 309, 7, 32 },
		{ 309, 8, 32 },
		{ 309, 9, 32 },
		{ 309, 10, 32 },
		{ 309, 11, 32 },
		{ 309, 12, 32 },
		{ 309, 13, 32 },
		{ 309, 14, 32 },
		{ 309, 15, 32 },
		{ 309, 16, 32 },
		{ 309, 17, 32 },
		{ 309, 18, 32 },
		{ 309, 19, 32 },
		{ 309, 20, 32 },
		{ 309, 21, 32 },
		{ 309, 22, 32 },
		{ 309, 23, 32 },
		{ 309, 24, 32 },
		{ 309, 25, 32 },
		{ 309, 26, 32 },
		{ 309, 27, 32 },
		{ 309, 28, 32 },
		{ 309, 29, 32 },
		{ 309, 30, 32 },
		{ 309, 31, 32 },
		{ 309, 32, 32 },
		{ 309, 33, 32 },
		{ 309, 34, 32 },
		{ 309, 35, 32 },
		{ 309, 36, 32 },
		{ 309, 37, 32 },
		{ 309, 38, 32 },
		{ 309, 39, 32 },
		{ 309, 40, 32 },
		{ 309, 41, 32 },
		{ 309, 42, 32 },
		{ 309, 43, 32 },
		{ 309, 44, 32 },
		{ 309, 45, 32 },
		{ 309, 46, 32 },
		{ 309, 47, 32 },
		{ 309, 48, 32 },
		{ 309, 49, 32 },
		{ 309, 50, 32 },
		{ 309, 51, 32 },
		{ 309, 52, 32 },
		{ 309, 53, 32 },
		{ 309, 54, 32 },
		{ 309, 55, 32 },
		{ 309, 56, 32 },
		{ 309, 57, 32 },
		{ 309, 58, 32 },
		{ 309, 59, 32 },
		{ 309, 60, 32 },
		{ 309, 61, 32 },
		{ 309, 62, 32 },
		{ 309, 63, 32 },
		{ 309, 64, 32 },
		{ 309, 65, 32 },
		{ 309, 66, 32 },
		{ 309, 67, 32 },
		{ 309, 68, 32 },
		{ 309, 69, 32 },
		{ 309, 70, 32 },
		{ 309, 71, 32 },
		{ 309, 72, 32 },
		{ 309, 73, 32 },
		{ 309, 74, 32 },
		{ 309, 75, 32 },
		{ 309, 76, 32 },
		{ 309, 77, 32 },
		{ 309, 78, 32 },
		{ 309, 79, 32 },
		{ 309, 80, 32 },
		{ 309, 81, 32 },
		{ 309, 82, 32 },
		{ 309, 83, 32 },
		{ 309, 84, 32 },
		{ 309, 85, 32 },
		{ 309, 86, 32 },
		{ 309, 87, 32 },
		{ 309, 88, 32 },
		{ 309, 89, 32 },
		{ 309, 90, 32 },
		{ 309, 91, 32 },
		{ 309, 92, 32 },
		{ 309, 93, 32 },
		{ 309, 94, 32 },
		{ 309, 95, 32 },
		{ 309, 96, 32 },
		{ 309, 97, 32 },
		{ 309, 98, 32 },
		{ 309, 99, 32 },
		{ 309, 100, 32 },
		{ 309, 101, 32 },
		{ 309, 102, 32 },
		{ 309, 103, 32 },
		{ 309, 104, 32 },
		{ 309, 105, 32 },
		{ 309, 106, 32 },
		{ 309, 107, 32 },
		{ 309, 108, 32 },
		{ 309, 109, 32 },
		{ 309, 110, 32 },
		{ 309, 111, 32 },
		{ 309, 112, 32 },
		{ 309, 113, 32 },
		{ 309, 114, 32 },
		{ 309, 115, 32 },
		{ 309, 116, 32 },
		{ 309, 117, 32 },
		{ 309, 118, 32 },
		{ 310, 1, 32 },
		{ 310, 2, 32 },
		{ 310, 3, 32 },
		{ 310, 4, 32 },
		{ 310, 5, 32 },
		{ 310, 6, 32 },
		{ 310, 7, 32 },
		{ 310, 8, 32 },
		{ 310, 9, 32 },
		{ 310, 10, 32 },
		{ 310, 11, 32 },
		{ 310, 12, 32 },
		{ 310, 13, 32 },
		{ 310, 14, 32 },
		{ 310, 15, 32 },
		{ 310, 16, 32 },
		{ 310, 17, 32 },
		{ 310, 18, 32 },
		{ 310, 19, 32 },
		{ 310, 20, 32 },
		{ 310, 21, 32 },
		{ 310, 22, 32 },
		{ 310, 23, 32 },
		{ 310, 24, 32 },
		{ 310, 25, 32 },
		{ 310, 26, 32 },
		{ 310, 27, 32 },
		{ 310, 28, 32 },
		{ 310, 29, 32 },
		{ 310, 30, 32 },
		{ 310, 31, 32 },
		{ 310, 32, 32 },
		{ 310, 33, 32 },
		{ 310, 34, 32 },
		{ 310, 35, 32 },
		{ 310, 36, 32 },
		{ 310, 37, 32 },
		{ 310, 38, 32 },
		{ 310, 39, 32 },
		{ 310, 40, 32 },
		{ 310, 41, 32 },
		{ 310, 42, 32 },
		{ 310, 43, 32 },
		{ 310, 44, 32 },
		{ 310, 45, 32 },
		{ 310, 46, 32 },
		{ 310, 47, 32 },
		{ 310, 48, 32 },
		{ 310, 49, 32 },
		{ 310, 50, 32 },
		{ 310, 51, 32 },
		{ 310, 52, 32 },
		{ 310, 53, 32 },
		{ 310, 54, 32 },
		{ 310, 55, 32 },
		{ 310, 56, 32 },
		{ 310, 57, 32 },
		{ 310, 58, 32 },
		{ 310, 59, 32 },
		{ 310, 60, 32 },
		{ 310, 61, 32 },
		{ 310, 62, 32 },
		{ 310, 63, 32 },
		{ 310, 64, 32 },
		{ 310, 65, 32 },
		{ 310, 66, 32 },
		{ 310, 67, 32 },
		{ 310, 68, 32 },
		{ 310, 69, 32 },
		{ 310, 70, 32 },
		{ 310, 71, 32 },
		{ 310, 72, 32 },
		{ 310, 73, 32 },
		{ 310, 74, 32 },
		{ 310, 75, 32 },
		{ 310, 76, 32 },
		{ 310, 77, 32 },
		{ 310, 78, 32 },
		{ 310, 79, 32 },
		{ 310, 80, 32 },
		{ 310, 81, 32 },
		{ 310, 82, 32 },
		{ 310, 83, 32 },
		{ 310, 84, 32 },
		{ 310, 85, 32 },
		{ 310, 86, 32 },
		{ 310, 87, 32 },
		{ 310, 88, 32 },
		{ 310, 89, 32 },
		{ 310, 90, 32 },
		{ 310, 91, 32 },
		{ 310, 92, 32 },
		{ 310, 93, 32 },
		{ 310, 94, 32 },
		{ 310, 95, 32 },
		{ 310, 96, 32 },
		{ 310, 97, 32 },
		{ 310, 98, 32 },
		{ 310, 99, 32 },
		{ 310, 100, 32 },
		{ 310, 101, 32 },
		{ 310, 102, 32 },
		{ 310, 103, 32 },
		{ 310, 104, 32 },
		{ 310, 105, 32 },
		{ 310, 106, 32 },
		{ 310, 107, 32 },
		{ 310, 108, 32 },
		{ 310, 109, 32 },
		{ 310, 110, 32 },
		{ 310, 111, 32 },
		{ 310, 112, 32 },
		{ 310, 113, 32 },
		{ 310, 114, 32 },
		{ 310, 115, 32 },
		{ 311, 1, 32 },
		{ 311, 2, 32 },
		{ 311, 3, 32 },
		{ 311, 4, 32 },
		{ 311, 5, 32 },
		{ 311, 6, 32 },
		{ 311, 7, 32 },
		{ 311, 8, 32 },
		{ 311, 9, 32 },
		{ 311, 10, 32 },
		{ 311, 11, 32 },
		{ 311, 12, 32 },
		{ 311, 13, 32 },
		{ 311, 14, 32 },
		{ 311, 15, 32 },
		{ 311, 16, 32 },
		{ 311, 17, 32 },
		{ 311, 18, 32 },
		{ 311, 19, 32 },
		{ 311, 20, 32 },
		{ 311, 21, 32 },
		{ 311, 22, 32 },
		{ 311, 23, 32 },
		{ 311, 24, 32 },
		{ 311, 25, 32 },
		{ 311, 26, 32 },
		{ 311, 27, 32 },
		{ 311, 28, 32 },
		{ 311, 29, 32 },
		{ 311, 30, 32 },
		{ 311, 31, 32 },
		{ 311, 32, 32 },
		{ 311, 33, 32 },
		{ 311, 34, 32 },
		{ 311, 35, 32 },
		{ 311, 36, 32 },
		{ 311, 37, 32 },
		{ 311, 38, 32 },
		{ 311, 39, 32 },
		{ 311, 40, 32 },
		{ 311, 41, 32 },
		{ 311, 42, 32 },
		{ 311, 43, 32 },
		{ 311, 44, 32 },
		{ 311, 45, 32 },
		{ 311, 46, 32 },
		{ 311, 47, 32 },
		{ 311, 48, 32 },
		{ 311, 49, 32 },
		{ 311, 50, 32 },
		{ 311, 51, 32 },
		{ 311, 52, 32 },
		{ 311, 53, 32 },
		{ 311, 54, 32 },
		{ 311, 55, 32 },
		{ 311, 56, 32 },
		{ 311, 57, 32 },
		{ 311, 58, 32 },
		{ 311, 59, 32 },
		{ 311, 60, 32 },
		{ 311, 61, 32 },
		{ 311, 62, 32 },
		{ 311, 63, 32 },
		{ 311, 64, 32 },
		{ 311, 65, 32 },
		{ 311, 66, 32 },
		{ 311, 67, 32 },
		{ 311, 68, 32 },
		{ 311, 69, 32 },
		{ 311, 70, 32 },
		{ 311, 71, 32 },
		{ 311, 72, 32 },
		{ 311, 73, 32 },
		{ 311, 74, 32 },
		{ 311, 75, 32 },
		{ 311, 76, 32 },
		{ 311, 77, 32 },
		{ 311, 78, 32 },
		{ 311, 79, 32 },
		{ 311, 80, 32 },
		{ 311, 81, 32 },
		{ 311, 82, 32 },
		{ 311, 83, 32 },
		{ 311, 84, 32 },
		{ 311, 85, 32 },
		{ 311, 86, 32 },
		{ 311, 87, 32 },
		{ 311, 88, 32 },
		{ 311, 89, 32 },
		{ 311, 90, 32 },
		{ 311, 91, 32 },
		{ 311, 92, 32 },
		{ 311, 93, 32 },
		{ 311, 94, 32 },
		{ 311, 95, 32 },
		{ 311, 96, 32 },
		{ 311, 97, 32 },
		{ 311, 98, 32 },
		{ 311, 99, 32 },
		{ 311, 100, 32 },
		{ 311, 101, 32 },
		{ 311, 102, 32 },
		{ 311, 103, 32 },
		{ 311, 104, 32 },
		{ 311, 105, 32 },
		{ 311, 106, 32 },
		{ 311, 107, 32 },
		{ 311, 108, 32 },
		{ 311, 109, 32 },
		{ 311, 110, 32 },
		{ 311, 111, 32 },
		{ 311, 112, 32 },
		{ 312, 1, 32 },
		{ 312, 2, 32 },
		{ 312, 3, 32 },
		{ 312, 4, 32 },
		{ 312, 5, 32 },
		{ 312, 6, 32 },
		{ 312, 7, 32 },
		{ 312, 8, 32 },
		{ 312, 9, 32 },
		{ 312, 10, 32 },
		{ 312, 11, 32 },
		{ 312, 12, 32 },
		{ 312, 13, 32 },
		{ 312, 14, 32 },
		{ 312, 15, 32 },
		{ 312, 16, 32 },
		{ 312, 17, 32 },
		{ 312, 18, 32 },
		{ 312, 19, 32 },
		{ 312, 20, 32 },
		{ 312, 21, 32 },
		{ 312, 22, 32 },
		{ 312, 23, 32 },
		{ 312, 24, 32 },
		{ 312, 25, 32 },
		{ 312, 26, 32 },
		{ 312, 27, 32 },
		{ 312, 28, 32 },
		{ 312, 29, 32 },
		{ 312, 30, 32 },
		{ 312, 31, 32 },
		{ 312, 32, 32 },
		{ 312, 33, 32 },
		{ 312, 34, 32 },
		{ 312, 35, 32 },
		{ 312, 36, 32 },
		{ 312, 37, 32 },
		{ 312, 38, 32 },
		{ 312, 39, 32 },
		{ 312, 40, 32 },
		{ 312, 41, 32 },
		{ 312, 42, 32 },
		{ 312, 43, 32 },
		{ 312, 44, 32 },
		{ 312, 45, 32 },
		{ 312, 46, 32 },
		{ 312, 47, 32 },
		{ 312, 48, 32 },
		{ 312, 49, 32 },
		{ 312, 50, 32 },
		{ 312, 51, 32 },
		{ 312, 52, 32 },
		{ 312, 53, 32 },
		{ 312, 54, 32 },
		{ 312, 55, 32 },
		{ 312, 56, 32 },
		{ 312, 57, 32 },
		{ 312, 58, 32 },
		{ 312, 59, 32 },
		{ 312, 60, 32 },
		{ 312, 61, 32 },
		{ 312, 62, 32 },
		{ 312, 63, 32 },
		{ 312, 64, 32 },
		{ 312, 65, 32 },
		{ 312, 66, 32 },
		{ 312, 67, 32 },
		{ 312, 68, 32 },
		{ 312, 69, 32 },
		{ 312, 70, 32 },
		{ 312, 71, 32 },
		{ 312, 72, 32 },
		{ 312, 73, 32 },
		{ 312, 74, 32 },
		{ 312, 75, 32 },
		{ 312, 76, 32 },
		{ 312, 77, 32 },
		{ 312, 78, 32 },
		{ 312, 79, 32 },
		{ 312, 80, 32 },
		{ 312, 81, 32 },
		{ 312, 82, 32 },
		{ 312, 83, 32 },
		{ 312, 84, 32 },
		{ 312, 85, 32 },
		{ 312, 86, 32 },
		{ 312, 87, 32 },
		{ 312, 88, 32 },
		{ 312, 89, 32 },
		{ 312, 90, 32 },
		{ 312, 91, 32 },
		{ 312, 92, 32 },
		{ 312, 93, 32 },
		{ 312, 94, 32 },
		{ 312, 95, 32 },
		{ 312, 96, 32 },
		{ 312, 97, 32 },
		{ 312, 98, 32 },
		{ 312, 99, 32 },
		{ 312, 100, 32 },
		{ 312, 101, 32 },
		{ 312, 102, 32 },
		{ 312, 103, 32 },
		{ 312, 104, 32 },
		{ 312, 105, 32 },
		{ 312, 106, 32 },
		{ 312, 107, 32 },
		{ 312, 108, 32 },
		{ 312, 109, 32 },
		{ 312, 110, 32 },
		{ 313, 1, 32 },
		{ 313, 2, 32 },
		{ 313, 3, 32 },
		{ 313, 4, 32 },
		{ 313, 5, 32 },
		{ 313, 6, 32 },
		{ 313, 7, 32 },
		{ 313, 8, 32 },
		{ 313, 9, 32 },
		{ 313, 10, 32 },
		{ 313, 11, 32 },
		{ 313, 12, 32 },
		{ 313, 13, 32 },
		{ 313, 14, 32 },
		{ 313, 15, 32 },
		{ 313, 16, 32 },
		{ 313, 17, 32 },
		{ 313, 18, 32 },
		{ 313, 19, 32 },
		{ 313, 20, 32 },
		{ 313, 21, 32 },
		{ 313, 22, 32 },
		{ 313, 23, 32 },
		{ 313, 24, 32 },
		{ 313, 25, 32 },
		{ 313, 26, 32 },
		{ 313, 27, 32 },
		{ 313, 28, 32 },
		{ 313, 29, 32 },
		{ 313, 30, 32 },
		{ 313, 31, 32 },
		{ 313, 32, 32 },
		{ 313, 33, 32 },
		{ 313, 34, 32 },
		{ 313, 35, 32 },
		{ 313, 36, 32 },
		{ 313, 37, 32 },
		{ 313, 38, 32 },
		{ 313, 39, 32 },
		{ 313, 40, 32 },
		{ 313, 41, 32 },
		{ 313, 42, 32 },
		{ 313, 43, 32 },
		{ 313, 44, 32 },
		{ 313, 45, 32 },
		{ 313, 46, 32 },
		{ 313, 47, 32 },
		{ 313, 48, 32 },
		{ 313, 49, 32 },
		{ 313, 50, 32 },
		{ 313, 51, 32 },
		{ 313, 52, 32 },
		{ 313, 53, 32 },
		{ 313, 54, 32 },
		{ 313, 55, 32 },
		{ 313, 56, 32 },
		{ 313, 57, 32 },
		{ 313, 58, 32 },
		{ 313, 59, 32 },
		{ 313, 60, 32 },
		{ 313, 61, 32 },
		{ 313, 62, 32 },
		{ 313, 63, 32 },
		{ 313, 64, 32 },
		{ 313, 65, 32 },
		{ 313, 66, 32 },
		{ 313, 67, 32 },
		{ 313, 68, 32 },
		{ 313, 69, 32 },
		{ 313, 70, 32 },
		{ 313, 71, 32 },
		{ 313, 72, 32 },
		{ 313, 73, 32 },
		{ 313, 74, 32 },
		{ 313, 75, 32 },
		{ 313, 76, 32 },
		{ 313, 77, 32 },
		{ 313, 78, 32 },
		{ 313, 79, 32 },
		{ 313, 80, 32 },
		{ 313, 81, 32 },
		{ 313, 82, 32 },
		{ 313, 83, 32 },
		{ 313, 84, 32 },
		{ 313, 85, 32 },
		{ 313, 86, 32 },
		{ 313, 87, 32 },
		{ 313, 88, 32 },
		{ 313, 89, 32 },
		{ 313, 90, 32 },
		{ 313, 91, 32 },
		{ 313, 92, 32 },
		{ 313, 93, 32 },
		{ 313, 94, 32 },
		{ 313, 95, 32 },
		{ 313, 96, 32 },
		{ 313, 97, 32 },
		{ 313, 98, 32 },
		{ 313, 99, 32 },
		{ 313, 100, 32 },
		{ 313, 101, 32 },
		{ 313, 102, 32 },
		{ 313, 103, 32 },
		{ 313, 104, 32 },
		{ 313, 105, 32 },
		{ 313, 106, 32 },
		{ 313, 107, 32 },
		{ 314, 1, 32 },
		{ 314, 2, 32 },
		{ 314, 3, 32 },
		{ 314, 4, 32 },
		{ 314, 5, 32 },
		{ 314, 6, 32 },
		{ 314, 7, 32 },
		{ 314, 8, 32 },
		{ 314, 9, 32 },
		{ 314, 10, 32 },
		{ 314, 11, 32 },
		{ 314, 12, 32 },
		{ 314, 13, 32 },
		{ 314, 14, 32 },
		{ 314, 15, 32 },
		{ 314, 16, 32 },
		{ 314, 17, 32 },
		{ 314, 18, 32 },
		{ 314, 19, 32 },
		{ 314, 20, 32 },
		{ 314, 21, 32 },
		{ 314, 22, 32 },
		{ 314, 23, 32 },
		{ 314, 24, 32 },
		{ 314, 25, 32 },
		{ 314, 26, 32 },
		{ 314, 27, 32 },
		{ 314, 28, 32 },
		{ 314, 29, 32 },
		{ 314, 30, 32 },
		{ 314, 31, 32 },
		{ 314, 32, 32 },
		{ 314, 33, 32 },
		{ 314, 34, 32 },
		{ 314, 35, 32 },
		{ 314, 36, 32 },
		{ 314, 37, 32 },
		{ 314, 38, 32 },
		{ 314, 39, 32 },
		{ 314, 40, 32 },
		{ 314, 41, 32 },
		{ 314, 42, 32 },
		{ 314, 43, 32 },
		{ 314, 44, 32 },
		{ 314, 45, 32 },
		{ 314, 46, 32 },
		{ 314, 47, 32 },
		{ 314, 48, 32 },
		{ 314, 49, 32 },
		{ 314, 50, 32 },
		{ 314, 51, 32 },
		{ 314, 52, 32 },
		{ 314, 53, 32 },
		{ 314, 54, 32 },
		{ 314, 55, 32 },
		{ 314, 56, 32 },
		{ 314, 57, 32 },
		{ 314, 58, 32 },
		{ 314, 59, 32 },
		{ 314, 60, 32 },
		{ 314, 61, 32 },
		{ 314, 62, 32 },
		{ 314, 63, 32 },
		{ 314, 64, 32 },
		{ 314, 65, 32 },
		{ 314, 66, 32 },
		{ 314, 67, 32 },
		{ 314, 68, 32 },
		{ 314, 69, 32 },
		{ 314, 70, 32 },
		{ 314, 71, 32 },
		{ 314, 72, 32 },
		{ 314, 73, 32 },
		{ 314, 74, 32 },
		{ 314, 75, 32 },
		{ 314, 76, 32 },
		{ 314, 77, 32 },
		{ 314, 78, 32 },
		{ 314, 79, 32 },
		{ 314, 80, 32 },
		{ 314, 81, 32 },
		{ 314, 82, 32 },
		{ 314, 83, 32 },
		{ 314, 84, 32 },
		{ 314, 85, 32 },
		{ 314, 86, 32 },
		{ 314, 87, 32 },
		{ 314, 88, 32 },
		{ 314, 89, 32 },
		{ 314, 90, 32 },
		{ 314, 91, 32 },
		{ 314, 92, 32 },
		{ 314, 93, 32 },
		{ 314, 94, 32 },
		{ 314, 95, 32 },
		{ 314, 96, 32 },
		{ 314, 97, 32 },
		{ 314, 98, 32 },
		{ 314, 99, 32 },
		{ 314, 100, 32 },
		{ 314, 101, 32 },
		{ 314, 102, 32 },
		{ 314, 103, 32 },
		{ 314, 104, 32 },
		{ 315, 1, 32 },
		{ 315, 2, 32 },
		{ 315, 3, 32 },
		{ 315, 4, 32 },
		{ 315, 5, 32 },
		{ 315, 6, 32 },
		{ 315, 7, 32 },
		{ 315, 8, 32 },
		{ 315, 9, 32 },
		{ 315, 10, 32 },
		{ 315, 11, 32 },
		{ 315, 12, 32 },
		{ 315, 13, 32 },
		{ 315, 14, 32 },
		{ 315, 15, 32 },
		{ 315, 16, 32 },
		{ 315, 17, 32 },
		{ 315, 18, 32 },
		{ 315, 19, 32 },
		{ 315, 20, 32 },
		{ 315, 21, 32 },
		{ 315, 22, 32 },
		{ 315, 23, 32 },
		{ 315, 24, 32 },
		{ 315, 25, 32 },
		{ 315, 26, 32 },
		{ 315, 27, 32 },
		{ 315, 28, 32 },
		{ 315, 29, 32 },
		{ 315, 30, 32 },
		{ 315, 31, 32 },
		{ 315, 32, 32 },
		{ 315, 33, 32 },
		{ 315, 34, 32 },
		{ 315, 35, 32 },
		{ 315, 36, 32 },
		{ 315, 37, 32 },
		{ 315, 38, 32 },
		{ 315, 39, 32 },
		{ 315, 40, 32 },
		{ 315, 41, 32 },
		{ 315, 42, 32 },
		{ 315, 43, 32 },
		{ 315, 44, 32 },
		{ 315, 45, 32 },
		{ 315, 46, 32 },
		{ 315, 47, 32 },
		{ 315, 48, 32 },
		{ 315, 49, 32 },
		{ 315, 50, 32 },
		{ 315, 51, 32 },
		{ 315, 52, 32 },
		{ 315, 53, 32 },
		{ 315, 54, 32 },
		{ 315, 55, 32 },
		{ 315, 56, 32 },
		{ 315, 57, 32 },
		{ 315, 58, 32 },
		{ 315, 59, 32 },
		{ 315, 60, 32 },
		{ 315, 61, 32 },
		{ 315, 62, 32 },
		{ 315, 63, 32 },
		{ 315, 64, 32 },
		{ 315, 65, 32 },
		{ 315, 66, 32 },
		{ 315, 67, 32 },
		{ 315, 68, 32 },
		{ 315, 69, 32 },
		{ 315, 70, 32 },
		{ 315, 71, 32 },
		{ 315, 72, 32 },
		{ 315, 73, 32 },
		{ 315, 74, 32 },
		{ 315, 75, 32 },
		{ 315, 76, 32 },
		{ 315, 77, 32 },
		{ 315, 78, 32 },
		{ 315, 79, 32 },
		{ 315, 80, 32 },
		{ 315, 81, 32 },
		{ 315, 82, 32 },
		{ 315, 83, 32 },
		{ 315, 84, 32 },
		{ 315, 85, 32 },
		{ 315, 86, 32 },
		{ 315, 87, 32 },
		{ 315, 88, 32 },
		{ 315, 89, 32 },
		{ 315, 90, 32 },
		{ 315, 91, 32 },
		{ 315, 92, 32 },
		{ 315, 93, 32 },
		{ 315, 94, 32 },
		{ 315, 95, 32 },
		{ 315, 96, 32 },
		{ 315, 97, 32 },
		{ 315, 98, 32 },
		{ 315, 99, 32 },
		{ 315, 100, 32 },
		{ 315, 101, 32 },
		{ 315, 102, 32 },
		{ 316, 1, 32 },
		{ 316, 2, 32 },
		{ 316, 3, 32 },
		{ 316, 4, 32 },
		{ 316, 5, 32 },
		{ 316, 6, 32 },
		{ 316, 7, 32 },
		{ 316, 8, 32 },
		{ 316, 9, 32 },
		{ 316, 10, 32 },
		{ 316, 11, 32 },
		{ 316, 12, 32 },
		{ 316, 13, 32 },
		{ 316, 14, 32 },
		{ 316, 15, 32 },
		{ 316, 16, 32 },
		{ 316, 17, 32 },
		{ 316, 18, 32 },
		{ 316, 19, 32 },
		{ 316, 20, 32 },
		{ 316, 21, 32 },
		{ 316, 22, 32 },
		{ 316, 23, 32 },
		{ 316, 24, 32 },
		{ 316, 25, 32 },
		{ 316, 26, 32 },
		{ 316, 27, 32 },
		{ 316, 28, 32 },
		{ 316, 29, 32 },
		{ 316, 30, 32 },
		{ 316, 31, 32 },
		{ 316, 32, 32 },
		{ 316, 33, 32 },
		{ 316, 34, 32 },
		{ 316, 35, 32 },
		{ 316, 36, 32 },
		{ 316, 37, 32 },
		{ 316, 38, 32 },
		{ 316, 39, 32 },
		{ 316, 40, 32 },
		{ 316, 41, 32 },
		{ 316, 42, 32 },
		{ 316, 43, 32 },
		{ 316, 44, 32 },
		{ 316, 45, 32 },
		{ 316, 46, 32 },
		{ 316, 47, 32 },
		{ 316, 48, 32 },
		{ 316, 49, 32 },
		{ 316, 50, 32 },
		{ 316, 51, 32 },
		{ 316, 52, 32 },
		{ 316, 53, 32 },
		{ 316, 54, 32 },
		{ 316, 55, 32 },
		{ 316, 56, 32 },
		{ 316, 57, 32 },
		{ 316, 58, 32 },
		{ 316, 59, 32 },
		{ 316, 60, 32 },
		{ 316, 61, 32 },
		{ 316, 62, 32 },
		{ 316, 63, 32 },
		{ 316, 64, 32 },
		{ 316, 65, 32 },
		{ 316, 66, 32 },
		{ 316, 67, 32 },
		{ 316, 68, 32 },
		{ 316, 69, 32 },
		{ 316, 70, 32 },
		{ 316, 71, 32 },
		{ 316, 72, 32 },
		{ 316, 73, 32 },
		{ 316, 74, 32 },
		{ 316, 75, 32 },
		{ 316, 76, 32 },
		{ 316, 77, 32 },
		{ 316, 78, 32 },
		{ 316, 79, 32 },
		{ 316, 80, 32 },
		{ 316, 81, 32 },
		{ 316, 82, 32 },
		{ 316, 83, 32 },
		{ 316, 84, 32 },
		{ 316, 85, 32 },
		{ 316, 86, 32 },
		{ 316, 87, 32 },
		{ 316, 88, 32 },
		{ 316, 89, 32 },
		{ 316, 90, 32 },
		{ 316, 91, 32 },
		{ 316, 92, 32 },
		{ 316, 93, 32 },
		{ 316, 94, 32 },
		{ 316, 95, 32 },
		{ 316, 96, 32 },
		{ 316, 97, 32 },
		{ 316, 98, 32 },
		{ 316, 99, 32 },
		{ 317, 1, 32 },
		{ 317, 2, 32 },
		{ 317, 3, 32 },
		{ 317, 4, 32 },
		{ 317, 5, 32 },
		{ 317, 6, 32 },
		{ 317, 7, 32 },
		{ 317, 8, 32 },
		{ 317, 9, 32 },
		{ 317, 10, 32 },
		{ 317, 11, 32 },
		{ 317, 12, 32 },
		{ 317, 13, 32 },
		{ 317, 14, 32 },
		{ 317, 15, 32 },
		{ 317, 16, 32 },
		{ 317, 17, 32 },
		{ 317, 18, 32 },
		{ 317, 19, 32 },
		{ 317, 20, 32 },
		{ 317, 21, 32 },
		{ 317, 22, 32 },
		{ 317, 23, 32 },
		{ 317, 24, 32 },
		{ 317, 25, 32 },
		{ 317, 26, 32 },
		{ 317, 27, 32 },
		{ 317, 28, 32 },
		{ 317, 29, 32 },
		{ 317, 30, 32 },
		{ 317, 31, 32 },
		{ 317, 32, 32 },
		{ 317, 33, 32 },
		{ 317, 34, 32 },
		{ 317, 35, 32 },
		{ 317, 36, 32 },
		{ 317, 37, 32 },
		{ 317, 38, 32 },
		{ 317, 39, 32 },
		{ 317, 40, 32 },
		{ 317, 41, 32 },
		{ 317, 42, 32 },
		{ 317, 43, 32 },
		{ 317, 44, 32 },
		{ 317, 45, 32 },
		{ 317, 46, 32 },
		{ 317, 47, 32 },
		{ 317, 48, 32 },
		{ 317, 49, 32 },
		{ 317, 50, 32 },
		{ 317, 51, 32 },
		{ 317, 52, 32 },
		{ 317, 53, 32 },
		{ 317, 54, 32 },
		{ 317, 55, 32 },
		{ 317, 56, 32 },
		{ 317, 57, 32 },
		{ 317, 58, 32 },
		{ 317, 59, 32 },
		{ 317, 60, 32 },
		{ 317, 61, 32 },
		{ 317, 62, 32 },
		{ 317, 63, 32 },
		{ 317, 64, 32 },
		{ 317, 65, 32 },
		{ 317, 66, 32 },
		{ 317, 67, 32 },
		{ 317, 68, 32 },
		{ 317, 69, 32 },
		{ 317, 70, 32 },
		{ 317, 71, 32 },
		{ 317, 72, 32 },
		{ 317, 73, 32 },
		{ 317, 74, 32 },
		{ 317, 75, 32 },
		{ 317, 76, 32 },
		{ 317, 77, 32 },
		{ 317, 78, 32 },
		{ 317, 79, 32 },
		{ 317, 80, 32 },
		{ 317, 81, 32 },
		{ 317, 82, 32 },
		{ 317, 83, 32 },
		{ 317, 84, 32 },
		{ 317, 85, 32 },
		{ 317, 86, 32 },
		{ 317, 87, 32 },
		{ 317, 88, 32 },
		{ 317, 89, 32 },
		{ 317, 90, 32 },
		{ 317, 91, 32 },
		{ 317, 92, 32 },
		{ 317, 93, 32 },
		{ 317, 94, 32 },
		{ 317, 95, 32 },
		{ 317, 96, 32 },
		{ 318, 1, 32 },
		{ 318, 2, 32 },
		{ 318, 3, 32 },
		{ 318, 4, 32 },
		{ 318, 5, 32 },
		{ 318, 6, 32 },
		{ 318, 7, 32 },
		{ 318, 8, 32 },
		{ 318, 9, 32 },
		{ 318, 10, 32 },
		{ 318, 11, 32 },
		{ 318, 12, 32 },
		{ 318, 13, 32 },
		{ 318, 14, 32 },
		{ 318, 15, 32 },
		{ 318, 16, 32 },
		{ 318, 17, 32 },
		{ 318, 18, 32 },
		{ 318, 19, 32 },
		{ 318, 20, 32 },
		{ 318, 21, 32 },
		{ 318, 22, 32 },
		{ 318, 23, 32 },
		{ 318, 24, 32 },
		{ 318, 25, 32 },
		{ 318, 26, 32 },
		{ 318, 27, 32 },
		{ 318, 28, 32 },
		{ 318, 29, 32 },
		{ 318, 30, 32 },
		{ 318, 31, 32 },
		{ 318, 32, 32 },
		{ 318, 33, 32 },
		{ 318, 34, 32 },
		{ 318, 35, 32 },
		{ 318, 36, 32 },
		{ 318, 37, 32 },
		{ 318, 38, 32 },
		{ 318, 39, 32 },
		{ 318, 40, 32 },
		{ 318, 41, 32 },
		{ 318, 42, 32 },
		{ 318, 43, 32 },
		{ 318, 44, 32 },
		{ 318, 45, 32 },
		{ 318, 46, 32 },
		{ 318, 47, 32 },
		{ 318, 48, 32 },
		{ 318, 49, 32 },
		{ 318, 50, 32 },
		{ 318, 51, 32 },
		{ 318, 52, 32 },
		{ 318, 53, 32 },
		{ 318, 54, 32 },
		{ 318, 55, 32 },
		{ 318, 56, 32 },
		{ 318, 57, 32 },
		{ 318, 58, 32 },
		{ 318, 59, 32 },
		{ 318, 60, 32 },
		{ 318, 61, 32 },
		{ 318, 62, 32 },
		{ 318, 63, 32 },
		{ 318, 64, 32 },
		{ 318, 65, 32 },
		{ 318, 66, 32 },
		{ 318, 67, 32 },
		{ 318, 68, 32 },
		{ 318, 69, 32 },
		{ 318, 70, 32 },
		{ 318, 71, 32 },
		{ 318, 72, 32 },
		{ 318, 73, 32 },
		{ 318, 74, 32 },
		{ 318, 75, 32 },
		{ 318, 76, 32 },
		{ 318, 77, 32 },
		{ 318, 78, 32 },
		{ 318, 79, 32 },
		{ 318, 80, 32 },
		{ 318, 81, 32 },
		{ 318, 82, 32 },
		{ 318, 83, 32 },
		{ 318, 84, 32 },
		{ 318, 85, 32 },
		{ 318, 86, 32 },
		{ 318, 87, 32 },
		{ 318, 88, 32 },
		{ 318, 89, 32 },
		{ 318, 90, 32 },
		{ 318, 91, 32 },
		{ 318, 92, 32 },
		{ 318, 93, 32 },
		{ 318, 94, 32 },
		{ 319, 1, 32 },
		{ 319, 2, 32 },
		{ 319, 3, 32 },
		{ 319, 4, 32 },
		{ 319, 5, 32 },
		{ 319, 6, 32 },
		{ 319, 7, 32 },
		{ 319, 8, 32 },
		{ 319, 9, 32 },
		{ 319, 10, 32 },
		{ 319, 11, 32 },
		{ 319, 12, 32 },
		{ 319, 13, 32 },
		{ 319, 14, 32 },
		{ 319, 15, 32 },
		{ 319, 16, 32 },
		{ 319, 17, 32 },
		{ 319, 18, 32 },
		{ 319, 19, 32 },
		{ 319, 20, 32 },
		{ 319, 21, 32 },
		{ 319, 22, 32 },
		{ 319, 23, 32 },
		{ 319, 24, 32 },
		{ 319, 25, 32 },
		{ 319, 26, 32 },
		{ 319, 27, 32 },
		{ 319, 28, 32 },
		{ 319, 29, 32 },
		{ 319, 30, 32 },
		{ 319, 31, 32 },
		{ 319, 32, 32 },
		{ 319, 33, 32 },
		{ 319, 34, 32 },
		{ 319, 35, 32 },
		{ 319, 36, 32 },
		{ 319, 37, 32 },
		{ 319, 38, 32 },
		{ 319, 39, 32 },
		{ 319, 40, 32 },
		{ 319, 41, 32 },
		{ 319, 42, 32 },
		{ 319, 43, 32 },
		{ 319, 44, 32 },
		{ 319, 45, 32 },
		{ 319, 46, 32 },
		{ 319, 47, 32 },
		{ 319, 48, 32 },
		{ 319, 49, 32 },
		{ 319, 50, 32 },
		{ 319, 51, 32 },
		{ 319, 52, 32 },
		{ 319, 53, 32 },
		{ 319, 54, 32 },
		{ 319, 55, 32 },
		{ 319, 56, 32 },
		{ 319, 57, 32 },
		{ 319, 58, 32 },
		{ 319, 59, 32 },
		{ 319, 60, 32 },
		{ 319, 61, 32 },
		{ 319, 62, 32 },
		{ 319, 63, 32 },
		{ 319, 64, 32 },
		{ 319, 65, 32 },
		{ 319, 66, 32 },
		{ 319, 67, 32 },
		{ 319, 68, 32 },
		{ 319, 69, 32 },
		{ 319, 70, 32 },
		{ 319, 71, 32 },
		{ 319, 72, 32 },
		{ 319, 73, 32 },
		{ 319, 74, 32 },
		{ 319, 75, 32 },
		{ 319, 76, 32 },
		{ 319, 77, 32 },
		{ 319, 78, 32 },
		{ 319, 79, 32 },
		{ 319, 80, 32 },
		{ 319, 81, 32 },
		{ 319, 82, 32 },
		{ 319, 83, 32 },
		{ 319, 84, 32 },
		{ 319, 85, 32 },
		{ 319, 86, 32 },
		{ 319, 87, 32 },
		{ 319, 88, 32 },
		{ 319, 89, 32 },
		{ 319, 90, 32 },
		{ 319, 91, 32 },
		{ 320, 1, 32 },
		{ 320, 2, 32 },
		{ 320, 3, 32 },
		{ 320, 4, 32 },
		{ 320, 5, 32 },
		{ 320, 6, 32 },
		{ 320, 7, 32 },
		{ 320, 8, 32 },
		{ 320, 9, 32 },
		{ 320, 10, 32 },
		{ 320, 11, 32 },
		{ 320, 12, 32 },
		{ 320, 13, 32 },
		{ 320, 14, 32 },
		{ 320, 15, 32 },
		{ 320, 16, 32 },
		{ 320, 17, 32 },
		{ 320, 18, 32 },
		{ 320, 19, 32 },
		{ 320, 20, 32 },
		{ 320, 21, 32 },
		{ 320, 22, 32 },
		{ 320, 23, 32 },
		{ 320, 24, 32 },
		{ 320, 25, 32 },
		{ 320, 26, 32 },
		{ 320, 27, 32 },
		{ 320, 28, 32 },
		{ 320, 29, 32 },
		{ 320, 30, 32 },
		{ 320, 31, 32 },
		{ 320, 32, 32 },
		{ 320, 33, 32 },
		{ 320, 34, 32 },
		{ 320, 35, 32 },
		{ 320, 36, 32 },
		{ 320, 37, 32 },
		{ 320, 38, 32 },
		{ 320, 39, 32 },
		{ 320, 40, 32 },
		{ 320, 41, 32 },
		{ 320, 42, 32 },
		{ 320, 43, 32 },
		{ 320, 44, 32 },
		{ 320, 45, 32 },
		{ 320, 46, 32 },
		{ 320, 47, 32 },
		{ 320, 48, 32 },
		{ 320, 49, 32 },
		{ 320, 50, 32 },
		{ 320, 51, 32 },
		{ 320, 52, 32 },
		{ 320, 53, 32 },
		{ 320, 54, 32 },
		{ 320, 55, 32 },
		{ 320, 56, 32 },
		{ 320, 57, 32 },
		{ 320, 58, 32 },
		{ 320, 59, 32 },
		{ 320, 60, 32 },
		{ 320, 61, 32 },
		{ 320, 62, 32 },
		{ 320, 63, 32 },
		{ 320, 64, 32 },
		{ 320, 65, 32 },
		{ 320, 66, 32 },
		{ 320, 67, 32 },
		{ 320, 68, 32 },
		{ 320, 69, 32 },
		{ 320, 70, 32 },
		{ 320, 71, 32 },
		{ 320, 72, 32 },
		{ 320, 73, 32 },
		{ 320, 74, 32 },
		{ 320, 75, 32 },
		{ 320, 76, 32 },
		{ 320, 77, 32 },
		{ 320, 78, 32 },
		{ 320, 79, 32 },
		{ 320, 80, 32 },
		{ 320, 81, 32 },
		{ 320, 82, 32 },
		{ 320, 83, 32 },
		{ 320, 84, 32 },
		{ 320, 85, 32 },
		{ 320, 86, 32 },
		{ 320, 87, 32 },
		{ 320, 88, 32 },
		{ 321, 1, 32 },
		{ 321, 2, 32 },
		{ 321, 3, 32 },
		{ 321, 4, 32 },
		{ 321, 5, 32 },
		{ 321, 6, 32 },
		{ 321, 7, 32 },
		{ 321, 8, 32 },
		{ 321, 9, 32 },
		{ 321, 10, 32 },
		{ 321, 11, 32 },
		{ 321, 12, 32 },
		{ 321, 13, 32 },
		{ 321, 14, 32 },
		{ 321, 15, 32 },
		{ 321, 16, 32 },
		{ 321, 17, 32 },
		{ 321, 18, 32 },
		{ 321, 19, 32 },
		{ 321, 20, 32 },
		{ 321, 21, 32 },
		{ 321, 22, 32 },
		{ 321, 23, 32 },
		{ 321, 24, 32 },
		{ 321, 25, 32 },
		{ 321, 26, 32 },
		{ 321, 27, 32 },
		{ 321, 28, 32 },
		{ 321, 29, 32 },
		{ 321, 30, 32 },
		{ 321, 31, 32 },
		{ 321, 32, 32 },
		{ 321, 33, 32 },
		{ 321, 34, 32 },
		{ 321, 35, 32 },
		{ 321, 36, 32 },
		{ 321, 37, 32 },
		{ 321, 38, 32 },
		{ 321, 39, 32 },
		{ 321, 40, 32 },
		{ 321, 41, 32 },
		{ 321, 42, 32 },
		{ 321, 43, 32 },
		{ 321, 44, 32 },
		{ 321, 45, 32 },
		{ 321, 46, 32 },
		{ 321, 47, 32 },
		{ 321, 48, 32 },
		{ 321, 49, 32 },
		{ 321, 50, 32 },
		{ 321, 51, 32 },
		{ 321, 52, 32 },
		{ 321, 53, 32 },
		{ 321, 54, 32 },
		{ 321, 55, 32 },
		{ 321, 56, 32 },
		{ 321, 57, 32 },
		{ 321, 58, 32 },
		{ 321, 59, 32 },
		{ 321, 60, 32 },
		{ 321, 61, 32 },
		{ 321, 62, 32 },
		{ 321, 63, 32 },
		{ 321, 64, 32 },
		{ 321, 65, 32 },
		{ 321, 66, 32 },
		{ 321, 67, 32 },
		{ 321, 68, 32 },
		{ 321, 69, 32 },
		{ 321, 70, 32 },
		{ 321, 71, 32 },
		{ 321, 72, 32 },
		{ 321, 73, 32 },
		{ 321, 74, 32 },
		{ 321, 75, 32 },
		{ 321, 76, 32 },
		{ 321, 77, 32 },
		{ 321, 78, 32 },
		{ 321, 79, 32 },
		{ 321, 80, 32 },
		{ 321, 81, 32 },
		{ 321, 82, 32 },
		{ 321, 83, 32 },
		{ 321, 84, 32 },
		{ 321, 85, 32 },
		{ 321, 86, 32 },
		{ 322, 1, 32 },
		{ 322, 2, 32 },
		{ 322, 3, 32 },
		{ 322, 4, 32 },
		{ 322, 5, 32 },
		{ 322, 6, 32 },
		{ 322, 7, 32 },
		{ 322, 8, 32 },
		{ 322, 9, 32 },
		{ 322, 10, 32 },
		{ 322, 11, 32 },
		{ 322, 12, 32 },
		{ 322, 13, 32 },
		{ 322, 14, 32 },
		{ 322, 15, 32 },
		{ 322, 16, 32 },
		{ 322, 17, 32 },
		{ 322, 18, 32 },
		{ 322, 19, 32 },
		{ 322, 20, 32 },
		{ 322, 21, 32 },
		{ 322, 22, 32 },
		{ 322, 23, 32 },
		{ 322, 24, 32 },
		{ 322, 25, 32 },
		{ 322, 26, 32 },
		{ 322, 27, 32 },
		{ 322, 28, 32 },
		{ 322, 29, 32 },
		{ 322, 30, 32 },
		{ 322, 31, 32 },
		{ 322, 32, 32 },
		{ 322, 33, 32 },
		{ 322, 34, 32 },
		{ 322, 35, 32 },
		{ 322, 36, 32 },
		{ 322, 37, 32 },
		{ 322, 38, 32 },
		{ 322, 39, 32 },
		{ 322, 40, 32 },
		{ 322, 41, 32 },
		{ 322, 42, 32 },
		{ 322, 43, 32 },
		{ 322, 44, 32 },
		{ 322, 45, 32 },
		{ 322, 46, 32 },
		{ 322, 47, 32 },
		{ 322, 48, 32 },
		{ 322, 49, 32 },
		{ 322, 50, 32 },
		{ 322, 51, 32 },
		{ 322, 52, 32 },
		{ 322, 53, 32 },
		{ 322, 54, 32 },
		{ 322, 55, 32 },
		{ 322, 56, 32 },
		{ 322, 57, 32 },
		{ 322, 58, 32 },
		{ 322, 59, 32 },
		{ 322, 60, 32 },
		{ 322, 61, 32 },
		{ 322, 62, 32 },
		{ 322, 63, 32 },
		{ 322, 64, 32 },
		{ 322, 65, 32 },
		{ 322, 66, 32 },
		{ 322, 67, 32 },
		{ 322, 68, 32 },
		{ 322, 69, 32 },
		{ 322, 70, 32 },
		{ 322, 71, 32 },
		{ 322, 72, 32 },
		{ 322, 73, 32 },
		{ 322, 74, 32 },
		{ 322, 75, 32 },
		{ 322, 76, 32 },
		{ 322, 77, 32 },
		{ 322, 78, 32 },
		{ 322, 79, 32 },
		{ 322, 80, 32 },
		{ 322, 81, 32 },
		{ 322, 82, 32 },
		{ 322, 83, 32 },
		{ 323, 1, 32 },
		{ 323, 2, 32 },
		{ 323, 3, 32 },
		{ 323, 4, 32 },
		{ 323, 5, 32 },
		{ 323, 6, 32 },
		{ 323, 7, 32 },
		{ 323, 8, 32 },
		{ 323, 9, 32 },
		{ 323, 10, 32 },
		{ 323, 11, 32 },
		{ 323, 12, 32 },
		{ 323, 13, 32 },
		{ 323, 14, 32 },
		{ 323, 15, 32 },
		{ 323, 16, 32 },
		{ 323, 17, 32 },
		{ 323, 18, 32 },
		{ 323, 19, 32 },
		{ 323, 20, 32 },
		{ 323, 21, 32 },
		{ 323, 22, 32 },
		{ 323, 23, 32 },
		{ 323, 24, 32 },
		{ 323, 25, 32 },
		{ 323, 26, 32 },
		{ 323, 27, 32 },
		{ 323, 28, 32 },
		{ 323, 29, 32 },
		{ 323, 30, 32 },
		{ 323, 31, 32 },
		{ 323, 32, 32 },
		{ 323, 33, 32 },
		{ 323, 34, 32 },
		{ 323, 35, 32 },
		{ 323, 36, 32 },
		{ 323, 37, 32 },
		{ 323, 38, 32 },
		{ 323, 39, 32 },
		{ 323, 40, 32 },
		{ 323, 41, 32 },
		{ 323, 42, 32 },
		{ 323, 43, 32 },
		{ 323, 44, 32 },
		{ 323, 45, 32 },
		{ 323, 46, 32 },
		{ 323, 47, 32 },
		{ 323, 48, 32 },
		{ 323, 49, 32 },
		{ 323, 50, 32 },
		{ 323, 51, 32 },
		{ 323, 52, 32 },
		{ 323, 53, 32 },
		{ 323, 54, 32 },
		{ 323, 55, 32 },
		{ 323, 56, 32 },
		{ 323, 57, 32 },
		{ 323, 58, 32 },
		{ 323, 59, 32 },
		{ 323, 60, 32 },
		{ 323, 61, 32 },
		{ 323, 62, 32 },
		{ 323, 63, 32 },
		{ 323, 64, 32 },
		{ 323, 65, 32 },
		{ 323, 66, 32 },
		{ 323, 67, 32 },
		{ 323, 68, 32 },
		{ 323, 69, 32 },
		{ 323, 70, 32 },
		{ 323, 71, 32 },
		{ 323, 72, 32 },
		{ 323, 73, 32 },
		{ 323, 74, 32 },
		{ 323, 75, 32 },
		{ 323, 76, 32 },
		{ 323, 77, 32 },
		{ 323, 78, 32 },
		{ 323, 79, 32 },
		{ 323, 80, 32 },
		{ 324, 1, 32 },
		{ 324, 2, 32 },
		{ 324, 3, 32 },
		{ 324, 4, 32 },
		{ 324, 5, 32 },
		{ 324, 6, 32 },
		{ 324, 7, 32 },
		{ 324, 8, 32 },
		{ 324, 9, 32 },
		{ 324, 10, 32 },
		{ 324, 11, 32 },
		{ 324, 12, 32 },
		{ 324, 13, 32 },
		{ 324, 14, 32 },
		{ 324, 15, 32 },
		{ 324, 16, 32 },
		{ 324, 17, 32 },
		{ 324, 18, 32 },
		{ 324, 19, 32 },
		{ 324, 20, 32 },
		{ 324, 21, 32 },
		{ 324, 22, 32 },
		{ 324, 23, 32 },
		{ 324, 24, 32 },
		{ 324, 25, 32 },
		{ 324, 26, 32 },
		{ 324, 27, 32 },
		{ 324, 28, 32 },
		{ 324, 29, 32 },
		{ 324, 30, 32 },
		{ 324, 31, 32 },
		{ 324, 32, 32 },
		{ 324, 33, 32 },
		{ 324, 34, 32 },
		{ 324, 35, 32 },
		{ 324, 36, 32 },
		{ 324, 37, 32 },
		{ 324, 38, 32 },
		{ 324, 39, 32 },
		{ 324, 40, 32 },
		{ 324, 41, 32 },
		{ 324, 42, 32 },
		{ 324, 43, 32 },
		{ 324, 44, 32 },
		{ 324, 45, 32 },
		{ 324, 46, 32 },
		{ 324, 47, 32 },
		{ 324, 48, 32 },
		{ 324, 49, 32 },
		{ 324, 50, 32 },
		{ 324, 51, 32 },
		{ 324, 52, 32 },
		{ 324, 53, 32 },
		{ 324, 54, 32 },
		{ 324, 55, 32 },
		{ 324, 56, 32 },
		{ 324, 57, 32 },
		{ 324, 58, 32 },
		{ 324, 59, 32 },
		{ 324, 60, 32 },
		{ 324, 61, 32 },
		{ 324, 62, 32 },
		{ 324, 63, 32 },
		{ 324, 64, 32 },
		{ 324, 65, 32 },
		{ 324, 66, 32 },
		{ 324, 67, 32 },
		{ 324, 68, 32 },
		{ 324, 69, 32 },
		{ 324, 70, 32 },
		{ 324, 71, 32 },
		{ 324, 72, 32 },
		{ 324, 73, 32 },
		{ 324, 74, 32 },
		{ 324, 75, 32 },
		{ 324, 76, 32 },
		{ 324, 77, 32 },
		{ 324, 78, 32 },
		{ 325, 1, 32 },
		{ 325, 2, 32 },
		{ 325, 3, 32 },
		{ 325, 4, 32 },
		{ 325, 5, 32 },
		{ 325, 6, 32 },
		{ 325, 7, 32 },
		{ 325, 8, 32 },
		{ 325, 9, 32 },
		{ 325, 10, 32 },
		{ 325, 11, 32 },
		{ 325, 12, 32 },
		{ 325, 13, 32 },
		{ 325, 14, 32 },
		{ 325, 15, 32 },
		{ 325, 16, 32 },
		{ 325, 17, 32 },
		{ 325, 18, 32 },
		{ 325, 19, 32 },
		{ 325, 20, 32 },
		{ 325, 21, 32 },
		{ 325, 22, 32 },
		{ 325, 23, 32 },
		{ 325, 24, 32 },
		{ 325, 25, 32 },
		{ 325, 26, 32 },
		{ 325, 27, 32 },
		{ 325, 28, 32 },
		{ 325, 29, 32 },
		{ 325, 30, 32 },
		{ 325, 31, 32 },
		{ 325, 32, 32 },
		{ 325, 33, 32 },
		{ 325, 34, 32 },
		{ 325, 35, 32 },
		{ 325, 36, 32 },
		{ 325, 37, 32 },
		{ 325, 38, 32 },
		{ 325, 39, 32 },
		{ 325, 40, 32 },
		{ 325, 41, 32 },
		{ 325, 42, 32 },
		{ 325, 43, 32 },
		{ 325, 44, 32 },
		{ 325, 45, 32 },
		{ 325, 46, 32 },
		{ 325, 47, 32 },
		{ 325, 48, 32 },
		{ 325, 49, 32 },
		{ 325, 50, 32 },
		{ 325, 51, 32 },
		{ 325, 52, 32 },
		{ 325, 53, 32 },
		{ 325, 54, 32 },
		{ 325, 55, 32 },
		{ 325, 56, 32 },
		{ 325, 57, 32 },
		{ 325, 58, 32 },
		{ 325, 59, 32 },
		{ 325, 60, 32 },
		{ 325, 61, 32 },
		{ 325, 62, 32 },
		{ 325, 63, 32 },
		{ 325, 64, 32 },
		{ 325, 65, 32 },
		{ 325, 66, 32 },
		{ 325, 67, 32 },
		{ 325, 68, 32 },
		{ 325, 69, 32 },
		{ 325, 70, 32 },
		{ 325, 71, 32 },
		{ 325, 72, 32 },
		{ 325, 73, 32 },
		{ 325, 74, 32 },
		{ 325, 75, 32 },
		{ 326, 1, 32 },
		{ 326, 2, 32 },
		{ 326, 3, 32 },
		{ 326, 4, 32 },
		{ 326, 5, 32 },
		{ 326, 6, 32 },
		{ 326, 7, 32 },
		{ 326, 8, 32 },
		{ 326, 9, 32 },
		{ 326, 10, 32 },
		{ 326, 11, 32 },
		{ 326, 12, 32 },
		{ 326, 13, 32 },
		{ 326, 14, 32 },
		{ 326, 15, 32 },
		{ 326, 16, 32 },
		{ 326, 17, 32 },
		{ 326, 18, 32 },
		{ 326, 19, 32 },
		{ 326, 20, 32 },
		{ 326, 21, 32 },
		{ 326, 22, 32 },
		{ 326, 23, 32 },
		{ 326, 24, 32 },
		{ 326, 25, 32 },
		{ 326, 26, 32 },
		{ 326, 27, 32 },
		{ 326, 28, 32 },
		{ 326, 29, 32 },
		{ 326, 30, 32 },
		{ 326, 31, 32 },
		{ 326, 32, 32 },
		{ 326, 33, 32 },
		{ 326, 34, 32 },
		{ 326, 35, 32 },
		{ 326, 36, 32 },
		{ 326, 37, 32 },
		{ 326, 38, 32 },
		{ 326, 39, 32 },
		{ 326, 40, 32 },
		{ 326, 41, 32 },
		{ 326, 42, 32 },
		{ 326, 43, 32 },
		{ 326, 44, 32 },
		{ 326, 45, 32 },
		{ 326, 46, 32 },
		{ 326, 47, 32 },
		{ 326, 48, 32 },
		{ 326, 49, 32 },
		{ 326, 50, 32 },
		{ 326, 51, 32 },
		{ 326, 52, 32 },
		{ 326, 53, 32 },
		{ 326, 54, 32 },
		{ 326, 55, 32 },
		{ 326, 56, 32 },
		{ 326, 57, 32 },
		{ 326, 58, 32 },
		{ 326, 59, 32 },
		{ 326, 60, 32 },
		{ 326, 61, 32 },
		{ 326, 62, 32 },
		{ 326, 63, 32 },
		{ 326, 64, 32 },
		{ 326, 65, 32 },
		{ 326, 66, 32 },
		{ 326, 67, 32 },
		{ 326, 68, 32 },
		{ 326, 69, 32 },
		{ 326, 70, 32 },
		{ 326, 71, 32 },
		{ 326, 72, 32 },
		{ 327, 1, 32 },
		{ 327, 2, 32 },
		{ 327, 3, 32 },
		{ 327, 4, 32 },
		{ 327, 5, 32 },
		{ 327, 6, 32 },
		{ 327, 7, 32 },
		{ 327, 8, 32 },
		{ 327, 9, 32 },
		{ 327, 10, 32 },
		{ 327, 11, 32 },
		{ 327, 12, 32 },
		{ 327, 13, 32 },
		{ 327, 14, 32 },
		{ 327, 15, 32 },
		{ 327, 16, 32 },
		{ 327, 17, 32 },
		{ 327, 18, 32 },
		{ 327, 19, 32 },
		{ 327, 20, 32 },
		{ 327, 21, 32 },
		{ 327, 22, 32 },
		{ 327, 23, 32 },
		{ 327, 24, 32 },
		{ 327, 25, 32 },
		{ 327, 26, 32 },
		{ 327, 27, 32 },
		{ 327, 28, 32 },
		{ 327, 29, 32 },
		{ 327, 30, 32 },
		{ 327, 31, 32 },
		{ 327, 32, 32 },
		{ 327, 33, 32 },
		{ 327, 34, 32 },
		{ 327, 35, 32 },
		{ 327, 36, 32 },
		{ 327, 37, 32 },
		{ 327, 38, 32 },
		{ 327, 39, 32 },
		{ 327, 40, 32 },
		{ 327, 41, 32 },
		{ 327, 42, 32 },
		{ 327, 43, 32 },
		{ 327, 44, 32 },
		{ 327, 45, 32 },
		{ 327, 46, 32 },
		{ 327, 47, 32 },
		{ 327, 48, 32 },
		{ 327, 49, 32 },
		{ 327, 50, 32 },
		{ 327, 51, 32 },
		{ 327, 52, 32 },
		{ 327, 53, 32 },
		{ 327, 54, 32 },
		{ 327, 55, 32 },
		{ 327, 56, 32 },
		{ 327, 57, 32 },
		{ 327, 58, 32 },
		{ 327, 59, 32 },
		{ 327, 60, 32 },
		{ 327, 61, 32 },
		{ 327, 62, 32 },
		{ 327, 63, 32 },
		{ 327, 64, 32 },
		{ 327, 65, 32 },
		{ 327, 66, 32 },
		{ 327, 67, 32 },
		{ 327, 68, 32 },
		{ 327, 69, 32 },
		{ 327, 70, 32 },
		{ 328, 1, 32 },
		{ 328, 2, 32 },
		{ 328, 3, 32 },
		{ 328, 4, 32 },
		{ 328, 5, 32 },
		{ 328, 6, 32 },
		{ 328, 7, 32 },
		{ 328, 8, 32 },
		{ 328, 9, 32 },
		{ 328, 10, 32 },
		{ 328, 11, 32 },
		{ 328, 12, 32 },
		{ 328, 13, 32 },
		{ 328, 14, 32 },
		{ 328, 15, 32 },
		{ 328, 16, 32 },
		{ 328, 17, 32 },
		{ 328, 18, 32 },
		{ 328, 19, 32 },
		{ 328, 20, 32 },
		{ 328, 21, 32 },
		{ 328, 22, 32 },
		{ 328, 23, 32 },
		{ 328, 24, 32 },
		{ 328, 25, 32 },
		{ 328, 26, 32 },
		{ 328, 27, 32 },
		{ 328, 28, 32 },
		{ 328, 29, 32 },
		{ 328, 30, 32 },
		{ 328, 31, 32 },
		{ 328, 32, 32 },
		{ 328, 33, 32 },
		{ 328, 34, 32 },
		{ 328, 35, 32 },
		{ 328, 36, 32 },
		{ 328, 37, 32 },
		{ 328, 38, 32 },
		{ 328, 39, 32 },
		{ 328, 40, 32 },
		{ 328, 41, 32 },
		{ 328, 42, 32 },
		{ 328, 43, 32 },
		{ 328, 44, 32 },
		{ 328, 45, 32 },
		{ 328, 46, 32 },
		{ 328, 47, 32 },
		{ 328, 48, 32 },
		{ 328, 49, 32 },
		{ 328, 50, 32 },
		{ 328, 51, 32 },
		{ 328, 52, 32 },
		{ 328, 53, 32 },
		{ 328, 54, 32 },
		{ 328, 55, 32 },
		{ 328, 56, 32 },
		{ 328, 57, 32 },
		{ 328, 58, 32 },
		{ 328, 59, 32 },
		{ 328, 60, 32 },
		{ 328, 61, 32 },
		{ 328, 62, 32 },
		{ 328, 63, 32 },
		{ 328, 64, 32 },
		{ 328, 65, 32 },
		{ 328, 66, 32 },
		{ 328, 67, 32 },
		{ 329, 1, 32 },
		{ 329, 2, 32 },
		{ 329, 3, 32 },
		{ 329, 4, 32 },
		{ 329, 5, 32 },
		{ 329, 6, 32 },
		{ 329, 7, 32 },
		{ 329, 8, 32 },
		{ 329, 9, 32 },
		{ 329, 10, 32 },
		{ 329, 11, 32 },
		{ 329, 12, 32 },
		{ 329, 13, 32 },
		{ 329, 14, 32 },
		{ 329, 15, 32 },
		{ 329, 16, 32 },
		{ 329, 17, 32 },
		{ 329, 18, 32 },
		{ 329, 19, 32 },
		{ 329, 20, 32 },
		{ 329, 21, 32 },
		{ 329, 22, 32 },
		{ 329, 23, 32 },
		{ 329, 24, 32 },
		{ 329, 25, 32 },
		{ 329, 26, 32 },
		{ 329, 27, 32 },
		{ 329, 28, 32 },
		{ 329, 29, 32 },
		{ 329, 30, 32 },
		{ 329, 31, 32 },
		{ 329, 32, 32 },
		{ 329, 33, 32 },
		{ 329, 34, 32 },
		{ 329, 35, 32 },
		{ 329, 36, 32 },
		{ 329, 37, 32 },
		{ 329, 38, 32 },
		{ 329, 39, 32 },
		{ 329, 40, 32 },
		{ 329, 41, 32 },
		{ 329, 42, 32 },
		{ 329, 43, 32 },
		{ 329, 44, 32 },
		{ 329, 45, 32 },
		{ 329, 46, 32 },
		{ 329, 47, 32 },
		{ 329, 48, 32 },
		{ 329, 49, 32 },
		{ 329, 50, 32 },
		{ 329, 51, 32 },
		{ 329, 52, 32 },
		{ 329, 53, 32 },
		{ 329, 54, 32 },
		{ 329, 55, 32 },
		{ 329, 56, 32 },
		{ 329, 57, 32 },
		{ 329, 58, 32 },
		{ 329, 59, 32 },
		{ 329, 60, 32 },
		{ 329, 61, 32 },
		{ 329, 62, 32 },
		{ 329, 63, 32 },
		{ 329, 64, 32 },
		{ 330, 1, 32 },
		{ 330, 2, 32 },
		{ 330, 3, 32 },
		{ 330, 4, 32 },
		{ 330, 5, 32 },
		{ 330, 6, 32 },
		{ 330, 7, 32 },
		{ 330, 8, 32 },
		{ 330, 9, 32 },
		{ 330, 10, 32 },
		{ 330, 11, 32 },
		{ 330, 12, 32 },
		{ 330, 13, 32 },
		{ 330, 14, 32 },
		{ 330, 15, 32 },
		{ 330, 16, 32 },
		{ 330, 17, 32 },
		{ 330, 18, 32 },
		{ 330, 19, 32 },
		{ 330, 20, 32 },
		{ 330, 21, 32 },
		{ 330, 22, 32 },
		{ 330, 23, 32 },
		{ 330, 24, 32 },
		{ 330, 25, 32 },
		{ 330, 26, 32 },
		{ 330, 27, 32 },
		{ 330, 28, 32 },
		{ 330, 29, 32 },
		{ 330, 30, 32 },
		{ 330, 31, 32 },
		{ 330, 32, 32 },
		{ 330, 33, 32 },
		{ 330, 34, 32 },
		{ 330, 35, 32 },
		{ 330, 36, 32 },
		{ 330, 37, 32 },
		{ 330, 38, 32 },
		{ 330, 39, 32 },
		{ 330, 40, 32 },
		{ 330, 41, 32 },
		{ 330, 42, 32 },
		{ 330, 43, 32 },
		{ 330, 44, 32 },
		{ 330, 45, 32 },
		{ 330, 46, 32 },
		{ 330, 47, 32 },
		{ 330, 48, 32 },
		{ 330, 49, 32 },
		{ 330, 50, 32 },
		{ 330, 51, 32 },
		{ 330, 52, 32 },
		{ 330, 53, 32 },
		{ 330, 54, 32 },
		{ 330, 55, 32 },
		{ 330, 56, 32 },
		{ 330, 57, 32 },
		{ 330, 58, 32 },
		{ 330, 59, 32 },
		{ 330, 60, 32 },
		{ 330, 61, 32 },
		{ 330, 62, 32 },
		{ 331, 1, 32 },
		{ 331, 2, 32 },
		{ 331, 3, 32 },
		{ 331, 4, 32 },
		{ 331, 5, 32 },
		{ 331, 6, 32 },
		{ 331, 7, 32 },
		{ 331, 8, 32 },
		{ 331, 9, 32 },
		{ 331, 10, 32 },
		{ 331, 11, 32 },
		{ 331, 12, 32 },
		{ 331, 13, 32 },
		{ 331, 14, 32 },
		{ 331, 15, 32 },
		{ 331, 16, 32 },
		{ 331, 17, 32 },
		{ 331, 18, 32 },
		{ 331, 19, 32 },
		{ 331, 20, 32 },
		{ 331, 21, 32 },
		{ 331, 22, 32 },
		{ 331, 23, 32 },
		{ 331, 24, 32 },
		{ 331, 25, 32 },
		{ 331, 26, 32 },
		{ 331, 27, 32 },
		{ 331, 28, 32 },
		{ 331, 29, 32 },
		{ 331, 30, 32 },
		{ 331, 31, 32 },
		{ 331, 32, 32 },
		{ 331, 33, 32 },
		{ 331, 34, 32 },
		{ 331, 35, 32 },
		{ 331, 36, 32 },
		{ 331, 37, 32 },
		{ 331, 38, 32 },
		{ 331, 39, 32 },
		{ 331, 40, 32 },
		{ 331, 41, 32 },
		{ 331, 42, 32 },
		{ 331, 43, 32 },
		{ 331, 44, 32 },
		{ 331, 45, 32 },
		{ 331, 46, 32 },
		{ 331, 47, 32 },
		{ 331, 48, 32 },
		{ 331, 49, 32 },
		{ 331, 50, 32 },
		{ 331, 51, 32 },
		{ 331, 52, 32 },
		{ 331, 53, 32 },
		{ 331, 54, 32 },
		{ 331, 55, 32 },
		{ 331, 56, 32 },
		{ 331, 57, 32 },
		{ 331, 58, 32 },
		{ 331, 59, 32 },
		{ 332, 1, 32 },
		{ 332, 2, 32 },
		{ 332, 3, 32 },
		{ 332, 4, 32 },
		{ 332, 5, 32 },
		{ 332, 6, 32 },
		{ 332, 7, 32 },
		{ 332, 8, 32 },
		{ 332, 9, 32 },
		{ 332, 10, 32 },
		{ 332, 11, 32 },
		{ 332, 12, 32 },
		{ 332, 13, 32 },
		{ 332, 14, 32 },
		{ 332, 15, 32 },
		{ 332, 16, 32 },
		{ 332, 17, 32 },
		{ 332, 18, 32 },
		{ 332, 19, 32 },
		{ 332, 20, 32 },
		{ 332, 21, 32 },
		{ 332, 22, 32 },
		{ 332, 23, 32 },
		{ 332, 24, 32 },
		{ 332, 25, 32 },
		{ 332, 26, 32 },
		{ 332, 27, 32 },
		{ 332, 28, 32 },
		{ 332, 29, 32 },
		{ 332, 30, 32 },
		{ 332, 31, 32 },
		{ 332, 32, 32 },
		{ 332, 33, 32 },
		{ 332, 34, 32 },
		{ 332, 35, 32 },
		{ 332, 36, 32 },
		{ 332, 37, 32 },
		{ 332, 38, 32 },
		{ 332, 39, 32 },
		{ 332, 40, 32 },
		{ 332, 41, 32 },
		{ 332, 42, 32 },
		{ 332, 43, 32 },
		{ 332, 44, 32 },
		{ 332, 45, 32 },
		{ 332, 46, 32 },
		{ 332, 47, 32 },
		{ 332, 48, 32 },
		{ 332, 49, 32 },
		{ 332, 50, 32 },
		{ 332, 51, 32 },
		{ 332, 52, 32 },
		{ 332, 53, 32 },
		{ 332, 54, 32 },
		{ 332, 55, 32 },
		{ 332, 56, 32 },
		{ 333, 1, 32 },
		{ 333, 2, 32 },
		{ 333, 3, 32 },
		{ 333, 4, 32 },
		{ 333, 5, 32 },
		{ 333, 6, 32 },
		{ 333, 7, 32 },
		{ 333, 8, 32 },
		{ 333, 9, 32 },
		{ 333, 10, 32 },
		{ 333, 11, 32 },
		{ 333, 12, 32 },
		{ 333, 13, 32 },
		{ 333, 14, 32 },
		{ 333, 15, 32 },
		{ 333, 16, 32 },
		{ 333, 17, 32 },
		{ 333, 18, 32 },
		{ 333, 19, 32 },
		{ 333, 20, 32 },
		{ 333, 21, 32 },
		{ 333, 22, 32 },
		{ 333, 23, 32 },
		{ 333, 24, 32 },
		{ 333, 25, 32 },
		{ 333, 26, 32 },
		{ 333, 27, 32 },
		{ 333, 28, 32 },
		{ 333, 29, 32 },
		{ 333, 30, 32 },
		{ 333, 31, 32 },
		{ 333, 32, 32 },
		{ 333, 33, 32 },
		{ 333, 34, 32 },
		{ 333, 35, 32 },
		{ 333, 36, 32 },
		{ 333, 37, 32 },
		{ 333, 38, 32 },
		{ 333, 39, 32 },
		{ 333, 40, 32 },
		{ 333, 41, 32 },
		{ 333, 42, 32 },
		{ 333, 43, 32 },
		{ 333, 44, 32 },
		{ 333, 45, 32 },
		{ 333, 46, 32 },
		{ 333, 47, 32 },
		{ 333, 48, 32 },
		{ 333, 49, 32 },
		{ 333, 50, 32 },
		{ 333, 51, 32 },
		{ 333, 52, 32 },
		{ 333, 53, 32 },
		{ 333, 54, 32 },
		{ 334, 1, 32 },
		{ 334, 2, 32 },
		{ 334, 3, 32 },
		{ 334, 4, 32 },
		{ 334, 5, 32 },
		{ 334, 6, 32 },
		{ 334, 7, 32 },
		{ 334, 8, 32 },
		{ 334, 9, 32 },
		{ 334, 10, 32 },
		{ 334, 11, 32 },
		{ 334, 12, 32 },
		{ 334, 13, 32 },
		{ 334, 14, 32 },
		{ 334, 15, 32 },
		{ 334, 16, 32 },
		{ 334, 17, 32 },
		{ 334, 18, 32 },
		{ 334, 19, 32 },
		{ 334, 20, 32 },
		{ 334, 21, 32 },
		{ 334, 22, 32 },
		{ 334, 23, 32 },
		{ 334, 24, 32 },
		{ 334, 25, 32 },
		{ 334, 26, 32 },
		{ 334, 27, 32 },
		{ 334, 28, 32 },
		{ 334, 29, 32 },
		{ 334, 30, 32 },
		{ 334, 31, 32 },
		{ 334, 32, 32 },
		{ 334, 33, 32 },
		{ 334, 34, 32 },
		{ 334, 35, 32 },
		{ 334, 36, 32 },
		{ 334, 37, 32 },
		{ 334, 38, 32 },
		{ 334, 39, 32 },
		{ 334, 40, 32 },
		{ 334, 41, 32 },
		{ 334, 42, 32 },
		{ 334, 43, 32 },
		{ 334, 44, 32 },
		{ 334, 45, 32 },
		{ 334, 46, 32 },
		{ 334, 47, 32 },
		{ 334, 48, 32 },
		{ 334, 49, 32 },
		{ 334, 50, 32 },
		{ 334, 51, 32 },
		{ 335, 1, 32 },
		{ 335, 2, 32 },
		{ 335, 3, 32 },
		{ 335, 4, 32 },
		{ 335, 5, 32 },
		{ 335, 6, 32 },
		{ 335, 7, 32 },
		{ 335, 8, 32 },
		{ 335, 9, 32 },
		{ 335, 10, 32 },
		{ 335, 11, 32 },
		{ 335, 12, 32 },
		{ 335, 13, 32 },
		{ 335, 14, 32 },
		{ 335, 15, 32 },
		{ 335, 16, 32 },
		{ 335, 17, 32 },
		{ 335, 18, 32 },
		{ 335, 19, 32 },
		{ 335, 20, 32 },
		{ 335, 21, 32 },
		{ 335, 22, 32 },
		{ 335, 23, 32 },
		{ 335, 24, 32 },
		{ 335, 25, 32 },
		{ 335, 26, 32 },
		{ 335, 27, 32 },
		{ 335, 28, 32 },
		{ 335, 29, 32 },
		{ 335, 30, 32 },
		{ 335, 31, 32 },
		{ 335, 32, 32 },
		{ 335, 33, 32 },
		{ 335, 34, 32 },
		{ 335, 35, 32 },
		{ 335, 36, 32 },
		{ 335, 37, 32 },
		{ 335, 38, 32 },
		{ 335, 39, 32 },
		{ 335, 40, 32 },
		{ 335, 41, 32 },
		{ 335, 42, 32 },
		{ 335, 43, 32 },
		{ 335, 44, 32 },
		{ 335, 45, 32 },
		{ 335, 46, 32 },
		{ 335, 47, 32 },
		{ 335, 48, 32 },
		{ 336, 1, 32 },
		{ 336, 2, 32 },
		{ 336, 3, 32 },
		{ 336, 4, 32 },
		{ 336, 5, 32 },
		{ 336, 6, 32 },
		{ 336, 7, 32 },
		{ 336, 8, 32 },
		{ 336, 9, 32 },
		{ 336, 10, 32 },
		{ 336, 11, 32 },
		{ 336, 12, 32 },
		{ 336, 13, 32 },
		{ 336, 14, 32 },
		{ 336, 15, 32 },
		{ 336, 16, 32 },
		{ 336, 17, 32 },
		{ 336, 18, 32 },
		{ 336, 19, 32 },
		{ 336, 20, 32 },
		{ 336, 21, 32 },
		{ 336, 22, 32 },
		{ 336, 23, 32 },
		{ 336, 24, 32 },
		{ 336, 25, 32 },
		{ 336, 26, 32 },
		{ 336, 27, 32 },
		{ 336, 28, 32 },
		{ 336, 29, 32 },
		{ 336, 30, 32 },
		{ 336, 31, 32 },
		{ 336, 32, 32 },
		{ 336, 33, 32 },
		{ 336, 34, 32 },
		{ 336, 35, 32 },
		{ 336, 36, 32 },
		{ 336, 37, 32 },
		{ 336, 38, 32 },
		{ 336, 39, 32 },
		{ 336, 40, 32 },
		{ 336, 41, 32 },
		{ 336, 42, 32 },
		{ 336, 43, 32 },
		{ 336, 44, 32 },
		{ 336, 45, 32 },
		{ 336, 46, 32 },
		{ 337, 1, 32 },
		{ 337, 2, 32 },
		{ 337, 3, 32 },
		{ 337, 4, 32 },
		{ 337, 5, 32 },
		{ 337, 6, 32 },
		{ 337, 7, 32 },
		{ 337, 8, 32 },
		{ 337, 9, 32 },
		{ 337, 10, 32 },
		{ 337, 11, 32 },
		{ 337, 12, 32 },
		{ 337, 13, 32 },
		{ 337, 14, 32 },
		{ 337, 15, 32 },
		{ 337, 16, 32 },
		{ 337, 17, 32 },
		{ 337, 18, 32 },
		{ 337, 19, 32 },
		{ 337, 20, 32 },
		{ 337, 21, 32 },
		{ 337, 22, 32 },
		{ 337, 23, 32 },
		{ 337, 24, 32 },
		{ 337, 25, 32 },
		{ 337, 26, 32 },
		{ 337, 27, 32 },
		{ 337, 28, 32 },
		{ 337, 29, 32 },
		{ 337, 30, 32 },
		{ 337, 31, 32 },
		{ 337, 32, 32 },
		{ 337, 33, 32 },
		{ 337, 34, 32 },
		{ 337, 35, 32 },
		{ 337, 36, 32 },
		{ 337, 37, 32 },
		{ 337, 38, 32 },
		{ 337, 39, 32 },
		{ 337, 40, 32 },
		{ 337, 41, 32 },
		{ 337, 42, 32 },
		{ 337, 43, 32 },
		{ 338, 1, 32 },
		{ 338, 2, 32 },
		{ 338, 3, 32 },
		{ 338, 4, 32 },
		{ 338, 5, 32 },
		{ 338, 6, 32 },
		{ 338, 7, 32 },
		{ 338, 8, 32 },
		{ 338, 9, 32 },
		{ 338, 10, 32 },
		{ 338, 11, 32 },
		{ 338, 12, 32 },
		{ 338, 13, 32 },
		{ 338, 14, 32 },
		{ 338, 15, 32 },
		{ 338, 16, 32 },
		{ 338, 17, 32 },
		{ 338, 18, 32 },
		{ 338, 19, 32 },
		{ 338, 20, 32 },
		{ 338, 21, 32 },
		{ 338, 22, 32 },
		{ 338, 23, 32 },
		{ 338, 24, 32 },
		{ 338, 25, 32 },
		{ 338, 26, 32 },
		{ 338, 27, 32 },
		{ 338, 28, 32 },
		{ 338, 29, 32 },
		{ 338, 30, 32 },
		{ 338, 31, 32 },
		{ 338, 32, 32 },
		{ 338, 33, 32 },
		{ 338, 34, 32 },
		{ 338, 35, 32 },
		{ 338, 36, 32 },
		{ 338, 37, 32 },
		{ 338, 38, 32 },
		{ 338, 39, 32 },
		{ 338, 40, 32 },
		{ 339, 1, 32 },
		{ 339, 2, 32 },
		{ 339, 3, 32 },
		{ 339, 4, 32 },
		{ 339, 5, 32 },
		{ 339, 6, 32 },
		{ 339, 7, 32 },
		{ 339, 8, 32 },
		{ 339, 9, 32 },
		{ 339, 10, 32 },
		{ 339, 11, 32 },
		{ 339, 12, 32 },
		{ 339, 13, 32 },
		{ 339, 14, 32 },
		{ 339, 15, 32 },
		{ 339, 16, 32 },
		{ 339, 17, 32 },
		{ 339, 18, 32 },
		{ 339, 19, 32 },
		{ 339, 20, 32 },
		{ 339, 21, 32 },
		{ 339, 22, 32 },
		{ 339, 23, 32 },
		{ 339, 24, 32 },
		{ 339, 25, 32 },
		{ 339, 26, 32 },
		{ 339, 27, 32 },
		{ 339, 28, 32 },
		{ 339, 29, 32 },
		{ 339, 30, 32 },
		{ 339, 31, 32 },
		{ 339, 32, 32 },
		{ 339, 33, 32 },
		{ 339, 34, 32 },
		{ 339, 35, 32 },
		{ 339, 36, 32 },
		{ 339, 37, 32 },
		{ 339, 38, 32 },
		{ 340, 1, 32 },
		{ 340, 2, 32 },
		{ 340, 3, 32 },
		{ 340, 4, 32 },
		{ 340, 5, 32 },
		{ 340, 6, 32 },
		{ 340, 7, 32 },
		{ 340, 8, 32 },
		{ 340, 9, 32 },
		{ 340, 10, 32 },
		{ 340, 11, 32 },
		{ 340, 12, 32 },
		{ 340, 13, 32 },
		{ 340, 14, 32 },
		{ 340, 15, 32 },
		{ 340, 16, 32 },
		{ 340, 17, 32 },
		{ 340, 18, 32 },
		{ 340, 19, 32 },
		{ 340, 20, 32 },
		{ 340, 21, 32 },
		{ 340, 22, 32 },
		{ 340, 23, 32 },
		{ 340, 24, 32 },
		{ 340, 25, 32 },
		{ 340, 26, 32 },
		{ 340, 27, 32 },
		{ 340, 28, 32 },
		{ 340, 29, 32 },
		{ 340, 30, 32 },
		{ 340, 31, 32 },
		{ 340, 32, 32 },
		{ 340, 33, 32 },
		{ 340, 34, 32 },
		{ 340, 35, 32 },
		{ 341, 1, 32 },
		{ 341, 2, 32 },
		{ 341, 3, 32 },
		{ 341, 4, 32 },
		{ 341, 5, 32 },
		{ 341, 6, 32 },
		{ 341, 7, 32 },
		{ 341, 8, 32 },
		{ 341, 9, 32 },
		{ 341, 10, 32 },
		{ 341, 11, 32 },
		{ 341, 12, 32 },
		{ 341, 13, 32 },
		{ 341, 14, 32 },
		{ 341, 15, 32 },
		{ 341, 16, 32 },
		{ 341, 17, 32 },
		{ 341, 18, 32 },
		{ 341, 19, 32 },
		{ 341, 20, 32 },
		{ 341, 21, 32 },
		{ 341, 22, 32 },
		{ 341, 23, 32 },
		{ 341, 24, 32 },
		{ 341, 25, 32 },
		{ 341, 26, 32 },
		{ 341, 27, 32 },
		{ 341, 28, 32 },
		{ 341, 29, 32 },
		{ 341, 30, 32 },
		{ 341, 31, 32 },
		{ 341, 32, 32 },
		{ 342, 1, 32 },
		{ 342, 2, 32 },
		{ 342, 3, 32 },
		{ 342, 4, 32 },
		{ 342, 5, 32 },
		{ 342, 6, 32 },
		{ 342, 7, 32 },
		{ 342, 8, 32 },
		{ 342, 9, 32 },
		{ 342, 10, 32 },
		{ 342, 11, 32 },
		{ 342, 12, 32 },
		{ 342, 13, 32 },
		{ 342, 14, 32 },
		{ 342, 15, 32 },
		{ 342, 16, 32 },
		{ 342, 17, 32 },
		{ 342, 18, 32 },
		{ 342, 19, 32 },
		{ 342, 20, 32 },
		{ 342, 21, 32 },
		{ 342, 22, 32 },
		{ 342, 23, 32 },
		{ 342, 24, 32 },
		{ 342, 25, 32 },
		{ 342, 26, 32 },
		{ 342, 27, 32 },
		{ 342, 28, 32 },
		{ 342, 29, 32 },
		{ 342, 30, 32 },
		{ 343, 1, 32 },
		{ 343, 2, 32 },
		{ 343, 3, 32 },
		{ 343, 4, 32 },
		{ 343, 5, 32 },
		{ 343, 6, 32 },
		{ 343, 7, 32 },
		{ 343, 8, 32 },
		{ 343, 9, 32 },
		{ 343, 10, 32 },
		{ 343, 11, 32 },
		{ 343, 12, 32 },
		{ 343, 13, 32 },
		{ 343, 14, 32 },
		{ 343, 15, 32 },
		{ 343, 16, 32 },
		{ 343, 17, 32 },
		{ 343, 18, 32 },
		{ 343, 19, 32 },
		{ 343, 20, 32 },
		{ 343, 21, 32 },
		{ 343, 22, 32 },
		{ 343, 23, 32 },
		{ 343, 24, 32 },
		{ 343, 25, 32 },
		{ 343, 26, 32 },
		{ 343, 27, 32 },
		{ 344, 1, 32 },
		{ 344, 2, 32 },
		{ 344, 3, 32 },
		{ 344, 4, 32 },
		{ 344, 5, 32 },
		{ 344, 6, 32 },
		{ 344, 7, 32 },
		{ 344, 8, 32 },
		{ 344, 9, 32 },
		{ 344, 10, 32 },
		{ 344, 11, 32 },
		{ 344, 12, 32 },
		{ 344, 13, 32 },
		{ 344, 14, 32 },
		{ 344, 15, 32 },
		{ 344, 16, 32 },
		{ 344, 17, 32 },
		{ 344, 18, 32 },
		{ 344, 19, 32 },
		{ 344, 20, 32 },
		{ 344, 21, 32 },
		{ 344, 22, 32 },
		{ 344, 23, 32 },
		{ 344, 24, 32 },
		{ 345, 1, 32 },
		{ 345, 2, 32 },
		{ 345, 3, 32 },
		{ 345, 4, 32 },
		{ 345, 5, 32 },
		{ 345, 6, 32 },
		{ 345, 7, 32 },
		{ 345, 8, 32 },
		{ 345, 9, 32 },
		{ 345, 10, 32 },
		{ 345, 11, 32 },
		{ 345, 12, 32 },
		{ 345, 13, 32 },
		{ 345, 14, 32 },
		{ 345, 15, 32 },
		{ 345, 16, 32 },
		{ 345, 17, 32 },
		{ 345, 18, 32 },
		{ 345, 19, 32 },
		{ 345, 20, 32 },
		{ 345, 21, 32 },
		{ 345, 22, 32 },
		{ 346, 1, 32 },
		{ 346, 2, 32 },
		{ 346, 3, 32 },
		{ 346, 4, 32 },
		{ 346, 5, 32 },
		{ 346, 6, 32 },
		{ 346, 7, 32 },
		{ 346, 8, 32 },
		{ 346, 9, 32 },
		{ 346, 10, 32 },
		{ 346, 11, 32 },
		{ 346, 12, 32 },
		{ 346, 13, 32 },
		{ 346, 14, 32 },
		{ 346, 15, 32 },
		{ 346, 16, 32 },
		{ 346, 17, 32 },
		{ 346, 18, 32 },
		{ 346, 19, 32 },
		{ 347, 1, 32 },
		{ 347, 2, 32 },
		{ 347, 3, 32 },
		{ 347, 4, 32 },
		{ 347, 5, 32 },
		{ 347, 6, 32 },
		{ 347, 7, 32 },
		{ 347, 8, 32 },
		{ 347, 9, 32 },
		{ 347, 10, 32 },
		{ 347, 11, 32 },
		{ 347, 12, 32 },
		{ 347, 13, 32 },
		{ 347, 14, 32 },
		{ 347, 15, 32 },
		{ 347, 16, 32 },
		{ 348, 1, 32 },
		{ 348, 2, 32 },
		{ 348, 3, 32 },
		{ 348, 4, 32 },
		{ 348, 5, 32 },
		{ 348, 6, 32 },
		{ 348, 7, 32 },
		{ 348, 8, 32 },
		{ 348, 9, 32 },
		{ 348, 10, 32 },
		{ 348, 11, 32 },
		{ 348, 12, 32 },
		{ 348, 13, 32 },
		{ 348, 14, 32 },
		{ 349, 1, 32 },
		{ 349, 2, 32 },
		{ 349, 3, 32 },
		{ 349, 4, 32 },
		{ 349, 5, 32 },
		{ 349, 6, 32 },
		{ 349, 7, 32 },
		{ 349, 8, 32 },
		{ 349, 9, 32 },
		{ 349, 10, 32 },
		{ 349, 11, 32 },
		{ 350, 1, 32 },
		{ 350, 2, 32 },
		{ 350, 3, 32 },
		{ 350, 4, 32 },
		{ 350, 5, 32 },
		{ 350, 6, 32 },
		{ 350, 7, 32 },
		{ 350, 8, 32 },
		{ 351, 1, 32 },
		{ 351, 2, 32 },
		{ 351, 3, 32 },
		{ 351, 4, 32 },
		{ 351, 5, 32 },
		{ 351, 6, 32 },
		{ 352, 1, 32 },
		{ 352, 2, 32 },
		{ 352, 3, 32 },
		{ 130, 512, 33 },
		{ 131, 510, 33 },
		{ 131, 511, 33 },
		{ 131, 512, 33 },
		{ 132, 508, 33 },
		{ 132, 509, 33 },
		{ 132, 510, 33 },
		{ 132, 511, 33 },
		{ 132, 512, 33 },
		{ 133, 506, 33 },
		{ 133, 507, 33 },
		{ 133, 508, 33 },
		{ 133, 509, 33 },
		{ 133, 510, 33 },
		{ 133, 511, 33 },
		{ 133, 512, 33 },
		{ 134, 504, 33 },
		{ 134, 505, 33 },
		{ 134, 506, 33 },
		{ 134, 507, 33 },
		{ 134, 508, 33 },
		{ 134, 509, 33 },
		{ 134, 510, 33 },
		{ 134, 511, 33 },
		{ 134, 512, 33 },
		{ 135, 502, 33 },
		{ 135, 503, 33 },
		{ 135, 504, 33 },
		{ 135, 505, 33 },
		{ 135, 506, 33 },
		{ 135, 507, 33 },
		{ 135, 508, 33 },
		{ 135, 509, 33 },
		{ 135, 510, 33 },
		{ 135, 511, 33 },
		{ 135, 512, 33 },
		{ 136, 500, 33 },
		{ 136, 501, 33 },
		{ 136, 502, 33 },
		{ 136, 503, 33 },
		{ 136, 504, 33 },
		{ 136, 505, 33 },
		{ 136, 506, 33 },
		{ 136, 507, 33 },
		{ 136, 508, 33 },
		{ 136, 509, 33 },
		{ 136, 510, 33 },
		{ 136, 511, 33 },
		{ 136, 512, 33 },
		{ 137, 498, 33 },
		{ 137, 499, 33 },
		{ 137, 500, 33 },
		{ 137, 501, 33 },
		{ 137, 502, 33 },
		{ 137, 503, 33 },
		{ 137, 504, 33 },
		{ 137, 505, 33 },
		{ 137, 506, 33 },
		{ 137, 507, 33 },
		{ 137, 508, 33 },
		{ 137, 509, 33 },
		{ 137, 510, 33 },
		{ 137, 511, 33 },
		{ 137, 512, 33 },
		{ 138, 496, 33 },
		{ 138, 497, 33 },
		{ 138, 498, 33 },
		{ 138, 499, 33 },
		{ 138, 500, 33 },
		{ 138, 501, 33 },
		{ 138, 502, 33 },
		{ 138, 503, 33 },
		{ 138, 504, 33 },
		{ 138, 505, 33 },
		{ 138, 506, 33 },
		{ 138, 507, 33 },
		{ 138, 508, 33 },
		{ 138, 509, 33 },
		{ 138, 510, 33 },
		{ 138, 511, 33 },
		{ 138, 512, 33 },
		{ 139, 494, 33 },
		{ 139, 495, 33 },
		{ 139, 496, 33 },
		{ 139, 497, 33 },
		{ 139, 498, 33 },
		{ 139, 499, 33 },
		{ 139, 500, 33 },
		{ 139, 501, 33 },
		{ 139, 502, 33 },
		{ 139, 503, 33 },
		{ 139, 504, 33 },
		{ 139, 505, 33 },
		{ 139, 506, 33 },
		{ 139, 507, 33 },
		{ 139, 508, 33 },
		{ 139, 509, 33 },
		{ 139, 510, 33 },
		{ 139, 511, 33 },
		{ 139, 512, 33 },
		{ 140, 492, 33 },
		{ 140, 493, 33 },
		{ 140, 494, 33 },
		{ 140, 495, 33 },
		{ 140, 496, 33 },
		{ 140, 497, 33 },
		{ 140, 498, 33 },
		{ 140, 499, 33 },
		{ 140, 500, 33 },
		{ 140, 501, 33 },
		{ 140, 502, 33 },
		{ 140, 503, 33 },
		{ 140, 504, 33 },
		{ 140, 505, 33 },
		{ 140, 506, 33 },
		{ 140, 507, 33 },
		{ 140, 508, 33 },
		{ 140, 509, 33 },
		{ 140, 510, 33 },
		{ 140, 511, 33 },
		{ 140, 512, 33 },
		{ 141, 490, 33 },
		{ 141, 491, 33 },
		{ 141, 492, 33 },
		{ 141, 493, 33 },
		{ 141, 494, 33 },
		{ 141, 495, 33 },
		{ 141, 496, 33 },
		{ 141, 497, 33 },
		{ 141, 498, 33 },
		{ 141, 499, 33 },
		{ 141, 500, 33 },
		{ 141, 501, 33 },
		{ 141, 502, 33 },
		{ 141, 503, 33 },
		{ 141, 504, 33 },
		{ 141, 505, 33 },
		{ 141, 506, 33 },
		{ 141, 507, 33 },
		{ 141, 508, 33 },
		{ 141, 509, 33 },
		{ 141, 510, 33 },
		{ 141, 511, 33 },
		{ 141, 512, 33 },
		{ 142, 488, 33 },
		{ 142, 489, 33 },
		{ 142, 490, 33 },
		{ 142, 491, 33 },
		{ 142, 492, 33 },
		{ 142, 493, 33 },
		{ 142, 494, 33 },
		{ 142, 495, 33 },
		{ 142, 496, 33 },
		{ 142, 497, 33 },
		{ 142, 498, 33 },
		{ 142, 499, 33 },
		{ 142, 500, 33 },
		{ 142, 501, 33 },
		{ 142, 502, 33 },
		{ 142, 503, 33 },
		{ 142, 504, 33 },
		{ 142, 505, 33 },
		{ 142, 506, 33 },
		{ 142, 507, 33 },
		{ 142, 508, 33 },
		{ 142, 509, 33 },
		{ 142, 510, 33 },
		{ 142, 511, 33 },
		{ 142, 512, 33 },
		{ 143, 486, 33 },
		{ 143, 487, 33 },
		{ 143, 488, 33 },
		{ 143, 489, 33 },
		{ 143, 490, 33 },
		{ 143, 491, 33 },
		{ 143, 492, 33 },
		{ 143, 493, 33 },
		{ 143, 494, 33 },
		{ 143, 495, 33 },
		{ 143, 496, 33 },
		{ 143, 497, 33 },
		{ 143, 498, 33 },
		{ 143, 499, 33 },
		{ 143, 500, 33 },
		{ 143, 501, 33 },
		{ 143, 502, 33 },
		{ 143, 503, 33 },
		{ 143, 504, 33 },
		{ 143, 505, 33 },
		{ 143, 506, 33 },
		{ 143, 507, 33 },
		{ 143, 508, 33 },
		{ 143, 509, 33 },
		{ 143, 510, 33 },
		{ 143, 511, 33 },
		{ 143, 512, 33 },
		{ 144, 484, 33 },
		{ 144, 485, 33 },
		{ 144, 486, 33 },
		{ 144, 487, 33 },
		{ 144, 488, 33 },
		{ 144, 489, 33 },
		{ 144, 490, 33 },
		{ 144, 491, 33 },
		{ 144, 492, 33 },
		{ 144, 493, 33 },
		{ 144, 494, 33 },
		{ 144, 495, 33 },
		{ 144, 496, 33 },
		{ 144, 497, 33 },
		{ 144, 498, 33 },
		{ 144, 499, 33 },
		{ 144, 500, 33 },
		{ 144, 501, 33 },
		{ 144, 502, 33 },
		{ 144, 503, 33 },
		{ 144, 504, 33 },
		{ 144, 505, 33 },
		{ 144, 506, 33 },
		{ 144, 507, 33 },
		{ 144, 508, 33 },
		{ 144, 509, 33 },
		{ 144, 510, 33 },
		{ 144, 511, 33 },
		{ 144, 512, 33 },
		{ 145, 482, 33 },
		{ 145, 483, 33 },
		{ 145, 484, 33 },
		{ 145, 485, 33 },
		{ 145, 486, 33 },
		{ 145, 487, 33 },
		{ 145, 488, 33 },
		{ 145, 489, 33 },
		{ 145, 490, 33 },
		{ 145, 491, 33 },
		{ 145, 492, 33 },
		{ 145, 493, 33 },
		{ 145, 494, 33 },
		{ 145, 495, 33 },
		{ 145, 496, 33 },
		{ 145, 497, 33 },
		{ 145, 498, 33 },
		{ 145, 499, 33 },
		{ 145, 500, 33 },
		{ 145, 501, 33 },
		{ 145, 502, 33 },
		{ 145, 503, 33 },
		{ 145, 504, 33 },
		{ 145, 505, 33 },
		{ 145, 506, 33 },
		{ 145, 507, 33 },
		{ 145, 508, 33 },
		{ 145, 509, 33 },
		{ 145, 510, 33 },
		{ 145, 511, 33 },
		{ 145, 512, 33 },
		{ 146, 480, 33 },
		{ 146, 481, 33 },
		{ 146, 482, 33 },
		{ 146, 483, 33 },
		{ 146, 484, 33 },
		{ 146, 485, 33 },
		{ 146, 486, 33 },
		{ 146, 487, 33 },
		{ 146, 488, 33 },
		{ 146, 489, 33 },
		{ 146, 490, 33 },
		{ 146, 491, 33 },
		{ 146, 492, 33 },
		{ 146, 493, 33 },
		{ 146, 494, 33 },
		{ 146, 495, 33 },
		{ 146, 496, 33 },
		{ 146, 497, 33 },
		{ 146, 498, 33 },
		{ 146, 499, 33 },
		{ 146, 500, 33 },
		{ 146, 501, 33 },
		{ 146, 502, 33 },
		{ 146, 503, 33 },
		{ 146, 504, 33 },
		{ 146, 505, 33 },
		{ 146, 506, 33 },
		{ 146, 507, 33 },
		{ 146, 508, 33 },
		{ 146, 509, 33 },
		{ 146, 510, 33 },
		{ 146, 511, 33 },
		{ 146, 512, 33 },
		{ 147, 478, 33 },
		{ 147, 479, 33 },
		{ 147, 480, 33 },
		{ 147, 481, 33 },
		{ 147, 482, 33 },
		{ 147, 483, 33 },
		{ 147, 484, 33 },
		{ 147, 485, 33 },
		{ 147, 486, 33 },
		{ 147, 487, 33 },
		{ 147, 488, 33 },
		{ 147, 489, 33 },
		{ 147, 490, 33 },
		{ 147, 491, 33 },
		{ 147, 492, 33 },
		{ 147, 493, 33 },
		{ 147, 494, 33 },
		{ 147, 495, 33 },
		{ 147, 496, 33 },
		{ 147, 497, 33 },
		{ 147, 498, 33 },
		{ 147, 499, 33 },
		{ 147, 500, 33 },
		{ 147, 501, 33 },
		{ 147, 502, 33 },
		{ 147, 503, 33 },
		{ 147, 504, 33 },
		{ 147, 505, 33 },
		{ 147, 506, 33 },
		{ 147, 507, 33 },
		{ 147, 508, 33 },
		{ 147, 509, 33 },
		{ 147, 510, 33 },
		{ 147, 511, 33 },
		{ 147, 512, 33 },
		{ 148, 476, 33 },
		{ 148, 477, 33 },
		{ 148, 478, 33 },
		{ 148, 479, 33 },
		{ 148, 480, 33 },
		{ 148, 481, 33 },
		{ 148, 482, 33 },
		{ 148, 483, 33 },
		{ 148, 484, 33 },
		{ 148, 485, 33 },
		{ 148, 486, 33 },
		{ 148, 487, 33 },
		{ 148, 488, 33 },
		{ 148, 489, 33 },
		{ 148, 490, 33 },
		{ 148, 491, 33 },
		{ 148, 492, 33 },
		{ 148, 493, 33 },
		{ 148, 494, 33 },
		{ 148, 495, 33 },
		{ 148, 496, 33 },
		{ 148, 497, 33 },
		{ 148, 498, 33 },
		{ 148, 499, 33 },
		{ 148, 500, 33 },
		{ 148, 501, 33 },
		{ 148, 502, 33 },
		{ 148, 503, 33 },
		{ 148, 504, 33 },
		{ 148, 505, 33 },
		{ 148, 506, 33 },
		{ 148, 507, 33 },
		{ 148, 508, 33 },
		{ 148, 509, 33 },
		{ 148, 510, 33 },
		{ 148, 511, 33 },
		{ 148, 512, 33 },
		{ 149, 474, 33 },
		{ 149, 475, 33 },
		{ 149, 476, 33 },
		{ 149, 477, 33 },
		{ 149, 478, 33 },
		{ 149, 479, 33 },
		{ 149, 480, 33 },
		{ 149, 481, 33 },
		{ 149, 482, 33 },
		{ 149, 483, 33 },
		{ 149, 484, 33 },
		{ 149, 485, 33 },
		{ 149, 486, 33 },
		{ 149, 487, 33 },
		{ 149, 488, 33 },
		{ 149, 489, 33 },
		{ 149, 490, 33 },
		{ 149, 491, 33 },
		{ 149, 492, 33 },
		{ 149, 493, 33 },
		{ 149, 494, 33 },
		{ 149, 495, 33 },
		{ 149, 496, 33 },
		{ 149, 497, 33 },
		{ 149, 498, 33 },
		{ 149, 499, 33 },
		{ 149, 500, 33 },
		{ 149, 501, 33 },
		{ 149, 502, 33 },
		{ 149, 503, 33 },
		{ 149, 504, 33 },
		{ 149, 505, 33 },
		{ 149, 506, 33 },
		{ 149, 507, 33 },
		{ 149, 508, 33 },
		{ 149, 509, 33 },
		{ 149, 510, 33 },
		{ 149, 511, 33 },
		{ 149, 512, 33 },
		{ 150, 472, 33 },
		{ 150, 473, 33 },
		{ 150, 474, 33 },
		{ 150, 475, 33 },
		{ 150, 476, 33 },
		{ 150, 477, 33 },
		{ 150, 478, 33 },
		{ 150, 479, 33 },
		{ 150, 480, 33 },
		{ 150, 481, 33 },
		{ 150, 482, 33 },
		{ 150, 483, 33 },
		{ 150, 484, 33 },
		{ 150, 485, 33 },
		{ 150, 486, 33 },
		{ 150, 487, 33 },
		{ 150, 488, 33 },
		{ 150, 489, 33 },
		{ 150, 490, 33 },
		{ 150, 491, 33 },
		{ 150, 492, 33 },
		{ 150, 493, 33 },
		{ 150, 494, 33 },
		{ 150, 495, 33 },
		{ 150, 496, 33 },
		{ 150, 497, 33 },
		{ 150, 498, 33 },
		{ 150, 499, 33 },
		{ 150, 500, 33 },
		{ 150, 501, 33 },
		{ 150, 502, 33 },
		{ 150, 503, 33 },
		{ 150, 504, 33 },
		{ 150, 505, 33 },
		{ 150, 506, 33 },
		{ 150, 507, 33 },
		{ 150, 508, 33 },
		{ 150, 509, 33 },
		{ 150, 510, 33 },
		{ 150, 511, 33 },
		{ 150, 512, 33 },
		{ 151, 470, 33 },
		{ 151, 471, 33 },
		{ 151, 472, 33 },
		{ 151, 473, 33 },
		{ 151, 474, 33 },
		{ 151, 475, 33 },
		{ 151, 476, 33 },
		{ 151, 477, 33 },
		{ 151, 478, 33 },
		{ 151, 479, 33 },
		{ 151, 480, 33 },
		{ 151, 481, 33 },
		{ 151, 482, 33 },
		{ 151, 483, 33 },
		{ 151, 484, 33 },
		{ 151, 485, 33 },
		{ 151, 486, 33 },
		{ 151, 487, 33 },
		{ 151, 488, 33 },
		{ 151, 489, 33 },
		{ 151, 490, 33 },
		{ 151, 491, 33 },
		{ 151, 492, 33 },
		{ 151, 493, 33 },
		{ 151, 494, 33 },
		{ 151, 495, 33 },
		{ 151, 496, 33 },
		{ 151, 497, 33 },
		{ 151, 498, 33 },
		{ 151, 499, 33 },
		{ 151, 500, 33 },
		{ 151, 501, 33 },
		{ 151, 502, 33 },
		{ 151, 503, 33 },
		{ 151, 504, 33 },
		{ 151, 505, 33 },
		{ 151, 506, 33 },
		{ 151, 507, 33 },
		{ 151, 508, 33 },
		{ 151, 509, 33 },
		{ 151, 510, 33 },
		{ 151, 511, 33 },
		{ 151, 512, 33 },
		{ 152, 468, 33 },
		{ 152, 469, 33 },
		{ 152, 470, 33 },
		{ 152, 471, 33 },
		{ 152, 472, 33 },
		{ 152, 473, 33 },
		{ 152, 474, 33 },
		{ 152, 475, 33 },
		{ 152, 476, 33 },
		{ 152, 477, 33 },
		{ 152, 478, 33 },
		{ 152, 479, 33 },
		{ 152, 480, 33 },
		{ 152, 481, 33 },
		{ 152, 482, 33 },
		{ 152, 483, 33 },
		{ 152, 484, 33 },
		{ 152, 485, 33 },
		{ 152, 486, 33 },
		{ 152, 487, 33 },
		{ 152, 488, 33 },
		{ 152, 489, 33 },
		{ 152, 490, 33 },
		{ 152, 491, 33 },
		{ 152, 492, 33 },
		{ 152, 493, 33 },
		{ 152, 494, 33 },
		{ 152, 495, 33 },
		{ 152, 496, 33 },
		{ 152, 497, 33 },
		{ 152, 498, 33 },
		{ 152, 499, 33 },
		{ 152, 500, 33 },
		{ 152, 501, 33 },
		{ 152, 502, 33 },
		{ 152, 503, 33 },
		{ 152, 504, 33 },
		{ 152, 505, 33 },
		{ 152, 506, 33 },
		{ 152, 507, 33 },
		{ 152, 508, 33 },
		{ 152, 509, 33 },
		{ 152, 510, 33 },
		{ 152, 511, 33 },
		{ 152, 512, 33 },
		{ 153, 466, 33 },
		{ 153, 467, 33 },
		{ 153, 468, 33 },
		{ 153, 469, 33 },
		{ 153, 470, 33 },
		{ 153, 471, 33 },
		{ 153, 472, 33 },
		{ 153, 473, 33 },
		{ 153, 474, 33 },
		{ 153, 475, 33 },
		{ 153, 476, 33 },
		{ 153, 477, 33 },
		{ 153, 478, 33 },
		{ 153, 479, 33 },
		{ 153, 480, 33 },
		{ 153, 481, 33 },
		{ 153, 482, 33 },
		{ 153, 483, 33 },
		{ 153, 484, 33 },
		{ 153, 485, 33 },
		{ 153, 486, 33 },
		{ 153, 487, 33 },
		{ 153, 488, 33 },
		{ 153, 489, 33 },
		{ 153, 490, 33 },
		{ 153, 491, 33 },
		{ 153, 492, 33 },
		{ 153, 493, 33 },
		{ 153, 494, 33 },
		{ 153, 495, 33 },
		{ 153, 496, 33 },
		{ 153, 497, 33 },
		{ 153, 498, 33 },
		{ 153, 499, 33 },
		{ 153, 500, 33 },
		{ 153, 501, 33 },
		{ 153, 502, 33 },
		{ 153, 503, 33 },
		{ 153, 504, 33 },
		{ 153, 505, 33 },
		{ 153, 506, 33 },
		{ 153, 507, 33 },
		{ 153, 508, 33 },
		{ 153, 509, 33 },
		{ 153, 510, 33 },
		{ 153, 511, 33 },
		{ 153, 512, 33 },
		{ 154, 464, 33 },
		{ 154, 465, 33 },
		{ 154, 466, 33 },
		{ 154, 467, 33 },
		{ 154, 468, 33 },
		{ 154, 469, 33 },
		{ 154, 470, 33 },
		{ 154, 471, 33 },
		{ 154, 472, 33 },
		{ 154, 473, 33 },
		{ 154, 474, 33 },
		{ 154, 475, 33 },
		{ 154, 476, 33 },
		{ 154, 477, 33 },
		{ 154, 478, 33 },
		{ 154, 479, 33 },
		{ 154, 480, 33 },
		{ 154, 481, 33 },
		{ 154, 482, 33 },
		{ 154, 483, 33 },
		{ 154, 484, 33 },
		{ 154, 485, 33 },
		{ 154, 486, 33 },
		{ 154, 487, 33 },
		{ 154, 488, 33 },
		{ 154, 489, 33 },
		{ 154, 490, 33 },
		{ 154, 491, 33 },
		{ 154, 492, 33 },
		{ 154, 493, 33 },
		{ 154, 494, 33 },
		{ 154, 495, 33 },
		{ 154, 496, 33 },
		{ 154, 497, 33 },
		{ 154, 498, 33 },
		{ 154, 499, 33 },
		{ 154, 500, 33 },
		{ 154, 501, 33 },
		{ 154, 502, 33 },
		{ 154, 503, 33 },
		{ 154, 504, 33 },
		{ 154, 505, 33 },
		{ 154, 506, 33 },
		{ 154, 507, 33 },
		{ 154, 508, 33 },
		{ 154, 509, 33 },
		{ 154, 510, 33 },
		{ 154, 511, 33 },
		{ 154, 512, 33 },
		{ 155, 462, 33 },
		{ 155, 463, 33 },
		{ 155, 464, 33 },
		{ 155, 465, 33 },
		{ 155, 466, 33 },
		{ 155, 467, 33 },
		{ 155, 468, 33 },
		{ 155, 469, 33 },
		{ 155, 470, 33 },
		{ 155, 471, 33 },
		{ 155, 472, 33 },
		{ 155, 473, 33 },
		{ 155, 474, 33 },
		{ 155, 475, 33 },
		{ 155, 476, 33 },
		{ 155, 477, 33 },
		{ 155, 478, 33 },
		{ 155, 479, 33 },
		{ 155, 480, 33 },
		{ 155, 481, 33 },
		{ 155, 482, 33 },
		{ 155, 483, 33 },
		{ 155, 484, 33 },
		{ 155, 485, 33 },
		{ 155, 486, 33 },
		{ 155, 487, 33 },
		{ 155, 488, 33 },
		{ 155, 489, 33 },
		{ 155, 490, 33 },
		{ 155, 491, 33 },
		{ 155, 492, 33 },
		{ 155, 493, 33 },
		{ 155, 494, 33 },
		{ 155, 495, 33 },
		{ 155, 496, 33 },
		{ 155, 497, 33 },
		{ 155, 498, 33 },
		{ 155, 499, 33 },
		{ 155, 500, 33 },
		{ 155, 501, 33 },
		{ 155, 502, 33 },
		{ 155, 503, 33 },
		{ 155, 504, 33 },
		{ 155, 505, 33 },
		{ 155, 506, 33 },
		{ 155, 507, 33 },
		{ 155, 508, 33 },
		{ 155, 509, 33 },
		{ 155, 510, 33 },
		{ 155, 511, 33 },
		{ 155, 512, 33 },
		{ 156, 460, 33 },
		{ 156, 461, 33 },
		{ 156, 462, 33 },
		{ 156, 463, 33 },
		{ 156, 464, 33 },
		{ 156, 465, 33 },
		{ 156, 466, 33 },
		{ 156, 467, 33 },
		{ 156, 468, 33 },
		{ 156, 469, 33 },
		{ 156, 470, 33 },
		{ 156, 471, 33 },
		{ 156, 472, 33 },
		{ 156, 473, 33 },
		{ 156, 474, 33 },
		{ 156, 475, 33 },
		{ 156, 476, 33 },
		{ 156, 477, 33 },
		{ 156, 478, 33 },
		{ 156, 479, 33 },
		{ 156, 480, 33 },
		{ 156, 481, 33 },
		{ 156, 482, 33 },
		{ 156, 483, 33 },
		{ 156, 484, 33 },
		{ 156, 485, 33 },
		{ 156, 486, 33 },
		{ 156, 487, 33 },
		{ 156, 488, 33 },
		{ 156, 489, 33 },
		{ 156, 490, 33 },
		{ 156, 491, 33 },
		{ 156, 492, 33 },
		{ 156, 493, 33 },
		{ 156, 494, 33 },
		{ 156, 495, 33 },
		{ 156, 496, 33 },
		{ 156, 497, 33 },
		{ 156, 498, 33 },
		{ 156, 499, 33 },
		{ 156, 500, 33 },
		{ 156, 501, 33 },
		{ 156, 502, 33 },
		{ 156, 503, 33 },
		{ 156, 504, 33 },
		{ 156, 505, 33 },
		{ 156, 506, 33 },
		{ 156, 507, 33 },
		{ 156, 508, 33 },
		{ 156, 509, 33 },
		{ 156, 510, 33 },
		{ 156, 511, 33 },
		{ 156, 512, 33 },
		{ 157, 458, 33 },
		{ 157, 459, 33 },
		{ 157, 460, 33 },
		{ 157, 461, 33 },
		{ 157, 462, 33 },
		{ 157, 463, 33 },
		{ 157, 464, 33 },
		{ 157, 465, 33 },
		{ 157, 466, 33 },
		{ 157, 467, 33 },
		{ 157, 468, 33 },
		{ 157, 469, 33 },
		{ 157, 470, 33 },
		{ 157, 471, 33 },
		{ 157, 472, 33 },
		{ 157, 473, 33 },
		{ 157, 474, 33 },
		{ 157, 475, 33 },
		{ 157, 476, 33 },
		{ 157, 477, 33 },
		{ 157, 478, 33 },
		{ 157, 479, 33 },
		{ 157, 480, 33 },
		{ 157, 481, 33 },
		{ 157, 482, 33 },
		{ 157, 483, 33 },
		{ 157, 484, 33 },
		{ 157, 485, 33 },
		{ 157, 486, 33 },
		{ 157, 487, 33 },
		{ 157, 488, 33 },
		{ 157, 489, 33 },
		{ 157, 490, 33 },
		{ 157, 491, 33 },
		{ 157, 492, 33 },
		{ 157, 493, 33 },
		{ 157, 494, 33 },
		{ 157, 495, 33 },
		{ 157, 496, 33 },
		{ 157, 497, 33 },
		{ 157, 498, 33 },
		{ 157, 499, 33 },
		{ 157, 500, 33 },
		{ 157, 501, 33 },
		{ 157, 502, 33 },
		{ 157, 503, 33 },
		{ 157, 504, 33 },
		{ 157, 505, 33 },
		{ 157, 506, 33 },
		{ 157, 507, 33 },
		{ 157, 508, 33 },
		{ 157, 509, 33 },
		{ 157, 510, 33 },
		{ 157, 511, 33 },
		{ 157, 512, 33 },
		{ 158, 456, 33 },
		{ 158, 457, 33 },
		{ 158, 458, 33 },
		{ 158, 459, 33 },
		{ 158, 460, 33 },
		{ 158, 461, 33 },
		{ 158, 462, 33 },
		{ 158, 463, 33 },
		{ 158, 464, 33 },
		{ 158, 465, 33 },
		{ 158, 466, 33 },
		{ 158, 467, 33 },
		{ 158, 468, 33 },
		{ 158, 469, 33 },
		{ 158, 470, 33 },
		{ 158, 471, 33 },
		{ 158, 472, 33 },
		{ 158, 473, 33 },
		{ 158, 474, 33 },
		{ 158, 475, 33 },
		{ 158, 476, 33 },
		{ 158, 477, 33 },
		{ 158, 478, 33 },
		{ 158, 479, 33 },
		{ 158, 480, 33 },
		{ 158, 481, 33 },
		{ 158, 482, 33 },
		{ 158, 483, 33 },
		{ 158, 484, 33 },
		{ 158, 485, 33 },
		{ 158, 486, 33 },
		{ 158, 487, 33 },
		{ 158, 488, 33 },
		{ 158, 489, 33 },
		{ 158, 490, 33 },
		{ 158, 491, 33 },
		{ 158, 492, 33 },
		{ 158, 493, 33 },
		{ 158, 494, 33 },
		{ 158, 495, 33 },
		{ 158, 496, 33 },
		{ 158, 497, 33 },
		{ 158, 498, 33 },
		{ 158, 499, 33 },
		{ 158, 500, 33 },
		{ 158, 501, 33 },
		{ 158, 502, 33 },
		{ 158, 503, 33 },
		{ 158, 504, 33 },
		{ 158, 505, 33 },
		{ 158, 506, 33 },
		{ 158, 507, 33 },
		{ 158, 508, 33 },
		{ 158, 509, 33 },
		{ 158, 510, 33 },
		{ 158, 511, 33 },
		{ 158, 512, 33 },
		{ 159, 454, 33 },
		{ 159, 455, 33 },
		{ 159, 456, 33 },
		{ 159, 457, 33 },
		{ 159, 458, 33 },
		{ 159, 459, 33 },
		{ 159, 460, 33 },
		{ 159, 461, 33 },
		{ 159, 462, 33 },
		{ 159, 463, 33 },
		{ 159, 464, 33 },
		{ 159, 465, 33 },
		{ 159, 466, 33 },
		{ 159, 467, 33 },
		{ 159, 468, 33 },
		{ 159, 469, 33 },
		{ 159, 470, 33 },
		{ 159, 471, 33 },
		{ 159, 472, 33 },
		{ 159, 473, 33 },
		{ 159, 474, 33 },
		{ 159, 475, 33 },
		{ 159, 476, 33 },
		{ 159, 477, 33 },
		{ 159, 478, 33 },
		{ 159, 479, 33 },
		{ 159, 480, 33 },
		{ 159, 481, 33 },
		{ 159, 482, 33 },
		{ 159, 483, 33 },
		{ 159, 484, 33 },
		{ 159, 485, 33 },
		{ 159, 486, 33 },
		{ 159, 487, 33 },
		{ 159, 488, 33 },
		{ 159, 489, 33 },
		{ 159, 490, 33 },
		{ 159, 491, 33 },
		{ 159, 492, 33 },
		{ 159, 493, 33 },
		{ 159, 494, 33 },
		{ 159, 495, 33 },
		{ 159, 496, 33 },
		{ 159, 497, 33 },
		{ 159, 498, 33 },
		{ 159, 499, 33 },
		{ 159, 500, 33 },
		{ 159, 501, 33 },
		{ 159, 502, 33 },
		{ 159, 503, 33 },
		{ 159, 504, 33 },
		{ 159, 505, 33 },
		{ 159, 506, 33 },
		{ 159, 507, 33 },
		{ 159, 508, 33 },
		{ 159, 509, 33 },
		{ 159, 510, 33 },
		{ 159, 511, 33 },
		{ 159, 512, 33 },
		{ 160, 452, 33 },
		{ 160, 453, 33 },
		{ 160, 454, 33 },
		{ 160, 455, 33 },
		{ 160, 456, 33 },
		{ 160, 457, 33 },
		{ 160, 458, 33 },
		{ 160, 459, 33 },
		{ 160, 460, 33 },
		{ 160, 461, 33 },
		{ 160, 462, 33 },
		{ 160, 463, 33 },
		{ 160, 464, 33 },
		{ 160, 465, 33 },
		{ 160, 466, 33 },
		{ 160, 467, 33 },
		{ 160, 468, 33 },
		{ 160, 469, 33 },
		{ 160, 470, 33 },
		{ 160, 471, 33 },
		{ 160, 472, 33 },
		{ 160, 473, 33 },
		{ 160, 474, 33 },
		{ 160, 475, 33 },
		{ 160, 476, 33 },
		{ 160, 477, 33 },
		{ 160, 478, 33 },
		{ 160, 479, 33 },
		{ 160, 480, 33 },
		{ 160, 481, 33 },
		{ 160, 482, 33 },
		{ 160, 483, 33 },
		{ 160, 484, 33 },
		{ 160, 485, 33 },
		{ 160, 486, 33 },
		{ 160, 487, 33 },
		{ 160, 488, 33 },
		{ 160, 489, 33 },
		{ 160, 490, 33 },
		{ 160, 491, 33 },
		{ 160, 492, 33 },
		{ 160, 493, 33 },
		{ 160, 494, 33 },
		{ 160, 495, 33 },
		{ 160, 496, 33 },
		{ 160, 497, 33 },
		{ 160, 498, 33 },
		{ 160, 499, 33 },
		{ 160, 500, 33 },
		{ 160, 501, 33 },
		{ 160, 502, 33 },
		{ 160, 503, 33 },
		{ 160, 504, 33 },
		{ 160, 505, 33 },
		{ 160, 506, 33 },
		{ 160, 507, 33 },
		{ 160, 508, 33 },
		{ 160, 509, 33 },
		{ 160, 510, 33 },
		{ 160, 511, 33 },
		{ 160, 512, 33 },
		{ 161, 450, 33 },
		{ 161, 451, 33 },
		{ 161, 452, 33 },
		{ 161, 453, 33 },
		{ 161, 454, 33 },
		{ 161, 455, 33 },
		{ 161, 456, 33 },
		{ 161, 457, 33 },
		{ 161, 458, 33 },
		{ 161, 459, 33 },
		{ 161, 460, 33 },
		{ 161, 461, 33 },
		{ 161, 462, 33 },
		{ 161, 463, 33 },
		{ 161, 464, 33 },
		{ 161, 465, 33 },
		{ 161, 466, 33 },
		{ 161, 467, 33 },
		{ 161, 468, 33 },
		{ 161, 469, 33 },
		{ 161, 470, 33 },
		{ 161, 471, 33 },
		{ 161, 472, 33 },
		{ 161, 473, 33 },
		{ 161, 474, 33 },
		{ 161, 475, 33 },
		{ 161, 476, 33 },
		{ 161, 477, 33 },
		{ 161, 478, 33 },
		{ 161, 479, 33 },
		{ 161, 480, 33 },
		{ 161, 481, 33 },
		{ 161, 482, 33 },
		{ 161, 483, 33 },
		{ 161, 484, 33 },
		{ 161, 485, 33 },
		{ 161, 486, 33 },
		{ 161, 487, 33 },
		{ 161, 488, 33 },
		{ 161, 489, 33 },
		{ 161, 490, 33 },
		{ 161, 491, 33 },
		{ 161, 492, 33 },
		{ 161, 493, 33 },
		{ 161, 494, 33 },
		{ 161, 495, 33 },
		{ 161, 496, 33 },
		{ 161, 497, 33 },
		{ 161, 498, 33 },
		{ 161, 499, 33 },
		{ 161, 500, 33 },
		{ 161, 501, 33 },
		{ 161, 502, 33 },
		{ 161, 503, 33 },
		{ 161, 504, 33 },
		{ 161, 505, 33 },
		{ 161, 506, 33 },
		{ 161, 507, 33 },
		{ 161, 508, 33 },
		{ 161, 509, 33 },
		{ 161, 510, 33 },
		{ 161, 511, 33 },
		{ 161, 512, 33 },
		{ 162, 448, 33 },
		{ 162, 449, 33 },
		{ 162, 450, 33 },
		{ 162, 451, 33 },
		{ 162, 452, 33 },
		{ 162, 453, 33 },
		{ 162, 454, 33 },
		{ 162, 455, 33 },
		{ 162, 456, 33 },
		{ 162, 457, 33 },
		{ 162, 458, 33 },
		{ 162, 459, 33 },
		{ 162, 460, 33 },
		{ 162, 461, 33 },
		{ 162, 462, 33 },
		{ 162, 463, 33 },
		{ 162, 464, 33 },
		{ 162, 465, 33 },
		{ 162, 466, 33 },
		{ 162, 467, 33 },
		{ 162, 468, 33 },
		{ 162, 469, 33 },
		{ 162, 470, 33 },
		{ 162, 471, 33 },
		{ 162, 472, 33 },
		{ 162, 473, 33 },
		{ 162, 474, 33 },
		{ 162, 475, 33 },
		{ 162, 476, 33 },
		{ 162, 477, 33 },
		{ 162, 478, 33 },
		{ 162, 479, 33 },
		{ 162, 480, 33 },
		{ 162, 481, 33 },
		{ 162, 482, 33 },
		{ 162, 483, 33 },
		{ 162, 484, 33 },
		{ 162, 485, 33 },
		{ 162, 486, 33 },
		{ 162, 487, 33 },
		{ 162, 488, 33 },
		{ 162, 489, 33 },
		{ 162, 490, 33 },
		{ 162, 491, 33 },
		{ 162, 492, 33 },
		{ 162, 493, 33 },
		{ 162, 494, 33 },
		{ 162, 495, 33 },
		{ 162, 496, 33 },
		{ 162, 497, 33 },
		{ 162, 498, 33 },
		{ 162, 499, 33 },
		{ 162, 500, 33 },
		{ 162, 501, 33 },
		{ 162, 502, 33 },
		{ 162, 503, 33 },
		{ 162, 504, 33 },
		{ 162, 505, 33 },
		{ 162, 506, 33 },
		{ 162, 507, 33 },
		{ 162, 508, 33 },
		{ 162, 509, 33 },
		{ 162, 510, 33 },
		{ 162, 511, 33 },
		{ 162, 512, 33 },
		{ 163, 446, 33 },
		{ 163, 447, 33 },
		{ 163, 448, 33 },
		{ 163, 449, 33 },
		{ 163, 450, 33 },
		{ 163, 451, 33 },
		{ 163, 452, 33 },
		{ 163, 453, 33 },
		{ 163, 454, 33 },
		{ 163, 455, 33 },
		{ 163, 456, 33 },
		{ 163, 457, 33 },
		{ 163, 458, 33 },
		{ 163, 459, 33 },
		{ 163, 460, 33 },
		{ 163, 461, 33 },
		{ 163, 462, 33 },
		{ 163, 463, 33 },
		{ 163, 464, 33 },
		{ 163, 465, 33 },
		{ 163, 466, 33 },
		{ 163, 467, 33 },
		{ 163, 468, 33 },
		{ 163, 469, 33 },
		{ 163, 470, 33 },
		{ 163, 471, 33 },
		{ 163, 472, 33 },
		{ 163, 473, 33 },
		{ 163, 474, 33 },
		{ 163, 475, 33 },
		{ 163, 476, 33 },
		{ 163, 477, 33 },
		{ 163, 478, 33 },
		{ 163, 479, 33 },
		{ 163, 480, 33 },
		{ 163, 481, 33 },
		{ 163, 482, 33 },
		{ 163, 483, 33 },
		{ 163, 484, 33 },
		{ 163, 485, 33 },
		{ 163, 486, 33 },
		{ 163, 487, 33 },
		{ 163, 488, 33 },
		{ 163, 489, 33 },
		{ 163, 490, 33 },
		{ 163, 491, 33 },
		{ 163, 492, 33 },
		{ 163, 493, 33 },
		{ 163, 494, 33 },
		{ 163, 495, 33 },
		{ 163, 496, 33 },
		{ 163, 497, 33 },
		{ 163, 498, 33 },
		{ 163, 499, 33 },
		{ 163, 500, 33 },
		{ 163, 501, 33 },
		{ 163, 502, 33 },
		{ 163, 503, 33 },
		{ 163, 504, 33 },
		{ 163, 505, 33 },
		{ 163, 506, 33 },
		{ 163, 507, 33 },
		{ 163, 508, 33 },
		{ 163, 509, 33 },
		{ 163, 510, 33 },
		{ 163, 511, 33 },
		{ 163, 512, 33 },
		{ 164, 444, 33 },
		{ 164, 445, 33 },
		{ 164, 446, 33 },
		{ 164, 447, 33 },
		{ 164, 448, 33 },
		{ 164, 449, 33 },
		{ 164, 450, 33 },
		{ 164, 451, 33 },
		{ 164, 452, 33 },
		{ 164, 453, 33 },
		{ 164, 454, 33 },
		{ 164, 455, 33 },
		{ 164, 456, 33 },
		{ 164, 457, 33 },
		{ 164, 458, 33 },
		{ 164, 459, 33 },
		{ 164, 460, 33 },
		{ 164, 461, 33 },
		{ 164, 462, 33 },
		{ 164, 463, 33 },
		{ 164, 464, 33 },
		{ 164, 465, 33 },
		{ 164, 466, 33 },
		{ 164, 467, 33 },
		{ 164, 468, 33 },
		{ 164, 469, 33 },
		{ 164, 470, 33 },
		{ 164, 471, 33 },
		{ 164, 472, 33 },
		{ 164, 473, 33 },
		{ 164, 474, 33 },
		{ 164, 475, 33 },
		{ 164, 476, 33 },
		{ 164, 477, 33 },
		{ 164, 478, 33 },
		{ 164, 479, 33 },
		{ 164, 480, 33 },
		{ 164, 481, 33 },
		{ 164, 482, 33 },
		{ 164, 483, 33 },
		{ 164, 484, 33 },
		{ 164, 485, 33 },
		{ 164, 486, 33 },
		{ 164, 487, 33 },
		{ 164, 488, 33 },
		{ 164, 489, 33 },
		{ 164, 490, 33 },
		{ 164, 491, 33 },
		{ 164, 492, 33 },
		{ 164, 493, 33 },
		{ 164, 494, 33 },
		{ 164, 495, 33 },
		{ 164, 496, 33 },
		{ 164, 497, 33 },
		{ 164, 498, 33 },
		{ 164, 499, 33 },
		{ 164, 500, 33 },
		{ 164, 501, 33 },
		{ 164, 502, 33 },
		{ 164, 503, 33 },
		{ 164, 504, 33 },
		{ 164, 505, 33 },
		{ 164, 506, 33 },
		{ 164, 507, 33 },
		{ 164, 508, 33 },
		{ 164, 509, 33 },
		{ 164, 510, 33 },
		{ 164, 511, 33 },
		{ 164, 512, 33 },
		{ 165, 442, 33 },
		{ 165, 443, 33 },
		{ 165, 444, 33 },
		{ 165, 445, 33 },
		{ 165, 446, 33 },
		{ 165, 447, 33 },
		{ 165, 448, 33 },
		{ 165, 449, 33 },
		{ 165, 450, 33 },
		{ 165, 451, 33 },
		{ 165, 452, 33 },
		{ 165, 453, 33 },
		{ 165, 454, 33 },
		{ 165, 455, 33 },
		{ 165, 456, 33 },
		{ 165, 457, 33 },
		{ 165, 458, 33 },
		{ 165, 459, 33 },
		{ 165, 460, 33 },
		{ 165, 461, 33 },
		{ 165, 462, 33 },
		{ 165, 463, 33 },
		{ 165, 464, 33 },
		{ 165, 465, 33 },
		{ 165, 466, 33 },
		{ 165, 467, 33 },
		{ 165, 468, 33 },
		{ 165, 469, 33 },
		{ 165, 470, 33 },
		{ 165, 471, 33 },
		{ 165, 472, 33 },
		{ 165, 473, 33 },
		{ 165, 474, 33 },
		{ 165, 475, 33 },
		{ 165, 476, 33 },
		{ 165, 477, 33 },
		{ 165, 478, 33 },
		{ 165, 479, 33 },
		{ 165, 480, 33 },
		{ 165, 481, 33 },
		{ 165, 482, 33 },
		{ 165, 483, 33 },
		{ 165, 484, 33 },
		{ 165, 485, 33 },
		{ 165, 486, 33 },
		{ 165, 487, 33 },
		{ 165, 488, 33 },
		{ 165, 489, 33 },
		{ 165, 490, 33 },
		{ 165, 491, 33 },
		{ 165, 492, 33 },
		{ 165, 493, 33 },
		{ 165, 494, 33 },
		{ 165, 495, 33 },
		{ 165, 496, 33 },
		{ 165, 497, 33 },
		{ 165, 498, 33 },
		{ 165, 499, 33 },
		{ 165, 500, 33 },
		{ 165, 501, 33 },
		{ 165, 502, 33 },
		{ 165, 503, 33 },
		{ 165, 504, 33 },
		{ 165, 505, 33 },
		{ 165, 506, 33 },
		{ 165, 507, 33 },
		{ 165, 508, 33 },
		{ 165, 509, 33 },
		{ 165, 510, 33 },
		{ 165, 511, 33 },
		{ 165, 512, 33 },
		{ 166, 440, 33 },
		{ 166, 441, 33 },
		{ 166, 442, 33 },
		{ 166, 443, 33 },
		{ 166, 444, 33 },
		{ 166, 445, 33 },
		{ 166, 446, 33 },
		{ 166, 447, 33 },
		{ 166, 448, 33 },
		{ 166, 449, 33 },
		{ 166, 450, 33 },
		{ 166, 451, 33 },
		{ 166, 452, 33 },
		{ 166, 453, 33 },
		{ 166, 454, 33 },
		{ 166, 455, 33 },
		{ 166, 456, 33 },
		{ 166, 457, 33 },
		{ 166, 458, 33 },
		{ 166, 459, 33 },
		{ 166, 460, 33 },
		{ 166, 461, 33 },
		{ 166, 462, 33 },
		{ 166, 463, 33 },
		{ 166, 464, 33 },
		{ 166, 465, 33 },
		{ 166, 466, 33 },
		{ 166, 467, 33 },
		{ 166, 468, 33 },
		{ 166, 469, 33 },
		{ 166, 470, 33 },
		{ 166, 471, 33 },
		{ 166, 472, 33 },
		{ 166, 473, 33 },
		{ 166, 474, 33 },
		{ 166, 475, 33 },
		{ 166, 476, 33 },
		{ 166, 477, 33 },
		{ 166, 478, 33 },
		{ 166, 479, 33 },
		{ 166, 480, 33 },
		{ 166, 481, 33 },
		{ 166, 482, 33 },
		{ 166, 483, 33 },
		{ 166, 484, 33 },
		{ 166, 485, 33 },
		{ 166, 486, 33 },
		{ 166, 487, 33 },
		{ 166, 488, 33 },
		{ 166, 489, 33 },
		{ 166, 490, 33 },
		{ 166, 491, 33 },
		{ 166, 492, 33 },
		{ 166, 493, 33 },
		{ 166, 494, 33 },
		{ 166, 495, 33 },
		{ 166, 496, 33 },
		{ 166, 497, 33 },
		{ 166, 498, 33 },
		{ 166, 499, 33 },
		{ 166, 500, 33 },
		{ 166, 501, 33 },
		{ 166, 502, 33 },
		{ 166, 503, 33 },
		{ 166, 504, 33 },
		{ 166, 505, 33 },
		{ 166, 506, 33 },
		{ 166, 507, 33 },
		{ 166, 508, 33 },
		{ 166, 509, 33 },
		{ 166, 510, 33 },
		{ 166, 511, 33 },
		{ 166, 512, 33 },
		{ 167, 438, 33 },
		{ 167, 439, 33 },
		{ 167, 440, 33 },
		{ 167, 441, 33 },
		{ 167, 442, 33 },
		{ 167, 443, 33 },
		{ 167, 444, 33 },
		{ 167, 445, 33 },
		{ 167, 446, 33 },
		{ 167, 447, 33 },
		{ 167, 448, 33 },
		{ 167, 449, 33 },
		{ 167, 450, 33 },
		{ 167, 451, 33 },
		{ 167, 452, 33 },
		{ 167, 453, 33 },
		{ 167, 454, 33 },
		{ 167, 455, 33 },
		{ 167, 456, 33 },
		{ 167, 457, 33 },
		{ 167, 458, 33 },
		{ 167, 459, 33 },
		{ 167, 460, 33 },
		{ 167, 461, 33 },
		{ 167, 462, 33 },
		{ 167, 463, 33 },
		{ 167, 464, 33 },
		{ 167, 465, 33 },
		{ 167, 466, 33 },
		{ 167, 467, 33 },
		{ 167, 468, 33 },
		{ 167, 469, 33 },
		{ 167, 470, 33 },
		{ 167, 471, 33 },
		{ 167, 472, 33 },
		{ 167, 473, 33 },
		{ 167, 474, 33 },
		{ 167, 475, 33 },
		{ 167, 476, 33 },
		{ 167, 477, 33 },
		{ 167, 478, 33 },
		{ 167, 479, 33 },
		{ 167, 480, 33 },
		{ 167, 481, 33 },
		{ 167, 482, 33 },
		{ 167, 483, 33 },
		{ 167, 484, 33 },
		{ 167, 485, 33 },
		{ 167, 486, 33 },
		{ 167, 487, 33 },
		{ 167, 488, 33 },
		{ 167, 489, 33 },
		{ 167, 490, 33 },
		{ 167, 491, 33 },
		{ 167, 492, 33 },
		{ 167, 493, 33 },
		{ 167, 494, 33 },
		{ 167, 495, 33 },
		{ 167, 496, 33 },
		{ 167, 497, 33 },
		{ 167, 498, 33 },
		{ 167, 499, 33 },
		{ 167, 500, 33 },
		{ 167, 501, 33 },
		{ 167, 502, 33 },
		{ 167, 503, 33 },
		{ 167, 504, 33 },
		{ 167, 505, 33 },
		{ 167, 506, 33 },
		{ 167, 507, 33 },
		{ 167, 508, 33 },
		{ 167, 509, 33 },
		{ 167, 510, 33 },
		{ 167, 511, 33 },
		{ 167, 512, 33 },
		{ 168, 436, 33 },
		{ 168, 437, 33 },
		{ 168, 438, 33 },
		{ 168, 439, 33 },
		{ 168, 440, 33 },
		{ 168, 441, 33 },
		{ 168, 442, 33 },
		{ 168, 443, 33 },
		{ 168, 444, 33 },
		{ 168, 445, 33 },
		{ 168, 446, 33 },
		{ 168, 447, 33 },
		{ 168, 448, 33 },
		{ 168, 449, 33 },
		{ 168, 450, 33 },
		{ 168, 451, 33 },
		{ 168, 452, 33 },
		{ 168, 453, 33 },
		{ 168, 454, 33 },
		{ 168, 455, 33 },
		{ 168, 456, 33 },
		{ 168, 457, 33 },
		{ 168, 458, 33 },
		{ 168, 459, 33 },
		{ 168, 460, 33 },
		{ 168, 461, 33 },
		{ 168, 462, 33 },
		{ 168, 463, 33 },
		{ 168, 464, 33 },
		{ 168, 465, 33 },
		{ 168, 466, 33 },
		{ 168, 467, 33 },
		{ 168, 468, 33 },
		{ 168, 469, 33 },
		{ 168, 470, 33 },
		{ 168, 471, 33 },
		{ 168, 472, 33 },
		{ 168, 473, 33 },
		{ 168, 474, 33 },
		{ 168, 475, 33 },
		{ 168, 476, 33 },
		{ 168, 477, 33 },
		{ 168, 478, 33 },
		{ 168, 479, 33 },
		{ 168, 480, 33 },
		{ 168, 481, 33 },
		{ 168, 482, 33 },
		{ 168, 483, 33 },
		{ 168, 484, 33 },
		{ 168, 485, 33 },
		{ 168, 486, 33 },
		{ 168, 487, 33 },
		{ 168, 488, 33 },
		{ 168, 489, 33 },
		{ 168, 490, 33 },
		{ 168, 491, 33 },
		{ 168, 492, 33 },
		{ 168, 493, 33 },
		{ 168, 494, 33 },
		{ 168, 495, 33 },
		{ 168, 496, 33 },
		{ 168, 497, 33 },
		{ 168, 498, 33 },
		{ 168, 499, 33 },
		{ 168, 500, 33 },
		{ 168, 501, 33 },
		{ 168, 502, 33 },
		{ 168, 503, 33 },
		{ 168, 504, 33 },
		{ 168, 505, 33 },
		{ 168, 506, 33 },
		{ 168, 507, 33 },
		{ 168, 508, 33 },
		{ 168, 509, 33 },
		{ 168, 510, 33 },
		{ 168, 511, 33 },
		{ 168, 512, 33 },
		{ 169, 434, 33 },
		{ 169, 435, 33 },
		{ 169, 436, 33 },
		{ 169, 437, 33 },
		{ 169, 438, 33 },
		{ 169, 439, 33 },
		{ 169, 440, 33 },
		{ 169, 441, 33 },
		{ 169, 442, 33 },
		{ 169, 443, 33 },
		{ 169, 444, 33 },
		{ 169, 445, 33 },
		{ 169, 446, 33 },
		{ 169, 447, 33 },
		{ 169, 448, 33 },
		{ 169, 449, 33 },
		{ 169, 450, 33 },
		{ 169, 451, 33 },
		{ 169, 452, 33 },
		{ 169, 453, 33 },
		{ 169, 454, 33 },
		{ 169, 455, 33 },
		{ 169, 456, 33 },
		{ 169, 457, 33 },
		{ 169, 458, 33 },
		{ 169, 459, 33 },
		{ 169, 460, 33 },
		{ 169, 461, 33 },
		{ 169, 462, 33 },
		{ 169, 463, 33 },
		{ 169, 464, 33 },
		{ 169, 465, 33 },
		{ 169, 466, 33 },
		{ 169, 467, 33 },
		{ 169, 468, 33 },
		{ 169, 469, 33 },
		{ 169, 470, 33 },
		{ 169, 471, 33 },
		{ 169, 472, 33 },
		{ 169, 473, 33 },
		{ 169, 474, 33 },
		{ 169, 475, 33 },
		{ 169, 476, 33 },
		{ 169, 477, 33 },
		{ 169, 478, 33 },
		{ 169, 479, 33 },
		{ 169, 480, 33 },
		{ 169, 481, 33 },
		{ 169, 482, 33 },
		{ 169, 483, 33 },
		{ 169, 484, 33 },
		{ 169, 485, 33 },
		{ 169, 486, 33 },
		{ 169, 487, 33 },
		{ 169, 488, 33 },
		{ 169, 489, 33 },
		{ 169, 490, 33 },
		{ 169, 491, 33 },
		{ 169, 492, 33 },
		{ 169, 493, 33 },
		{ 169, 494, 33 },
		{ 169, 495, 33 },
		{ 169, 496, 33 },
		{ 169, 497, 33 },
		{ 169, 498, 33 },
		{ 169, 499, 33 },
		{ 169, 500, 33 },
		{ 169, 501, 33 },
		{ 169, 502, 33 },
		{ 169, 503, 33 },
		{ 169, 504, 33 },
		{ 169, 505, 33 },
		{ 169, 506, 33 },
		{ 169, 507, 33 },
		{ 169, 508, 33 },
		{ 169, 509, 33 },
		{ 169, 510, 33 },
		{ 169, 511, 33 },
		{ 169, 512, 33 },
		{ 170, 432, 33 },
		{ 170, 433, 33 },
		{ 170, 434, 33 },
		{ 170, 435, 33 },
		{ 170, 436, 33 },
		{ 170, 437, 33 },
		{ 170, 438, 33 },
		{ 170, 439, 33 },
		{ 170, 440, 33 },
		{ 170, 441, 33 },
		{ 170, 442, 33 },
		{ 170, 443, 33 },
		{ 170, 444, 33 },
		{ 170, 445, 33 },
		{ 170, 446, 33 },
		{ 170, 447, 33 },
		{ 170, 448, 33 },
		{ 170, 449, 33 },
		{ 170, 450, 33 },
		{ 170, 451, 33 },
		{ 170, 452, 33 },
		{ 170, 453, 33 },
		{ 170, 454, 33 },
		{ 170, 455, 33 },
		{ 170, 456, 33 },
		{ 170, 457, 33 },
		{ 170, 458, 33 },
		{ 170, 459, 33 },
		{ 170, 460, 33 },
		{ 170, 461, 33 },
		{ 170, 462, 33 },
		{ 170, 463, 33 },
		{ 170, 464, 33 },
		{ 170, 465, 33 },
		{ 170, 466, 33 },
		{ 170, 467, 33 },
		{ 170, 468, 33 },
		{ 170, 469, 33 },
		{ 170, 470, 33 },
		{ 170, 471, 33 },
		{ 170, 472, 33 },
		{ 170, 473, 33 },
		{ 170, 474, 33 },
		{ 170, 475, 33 },
		{ 170, 476, 33 },
		{ 170, 477, 33 },
		{ 170, 478, 33 },
		{ 170, 479, 33 },
		{ 170, 480, 33 },
		{ 170, 481, 33 },
		{ 170, 482, 33 },
		{ 170, 483, 33 },
		{ 170, 484, 33 },
		{ 170, 485, 33 },
		{ 170, 486, 33 },
		{ 170, 487, 33 },
		{ 170, 488, 33 },
		{ 170, 489, 33 },
		{ 170, 490, 33 },
		{ 170, 491, 33 },
		{ 170, 492, 33 },
		{ 170, 493, 33 },
		{ 170, 494, 33 },
		{ 170, 495, 33 },
		{ 170, 496, 33 },
		{ 170, 497, 33 },
		{ 170, 498, 33 },
		{ 170, 499, 33 },
		{ 170, 500, 33 },
		{ 170, 501, 33 },
		{ 170, 502, 33 },
		{ 170, 503, 33 },
		{ 170, 504, 33 },
		{ 170, 505, 33 },
		{ 170, 506, 33 },
		{ 170, 507, 33 },
		{ 170, 508, 33 },
		{ 170, 509, 33 },
		{ 170, 510, 33 },
		{ 170, 511, 33 },
		{ 170, 512, 33 },
		{ 171, 430, 33 },
		{ 171, 431, 33 },
		{ 171, 432, 33 },
		{ 171, 433, 33 },
		{ 171, 434, 33 },
		{ 171, 435, 33 },
		{ 171, 436, 33 },
		{ 171, 437, 33 },
		{ 171, 438, 33 },
		{ 171, 439, 33 },
		{ 171, 440, 33 },
		{ 171, 441, 33 },
		{ 171, 442, 33 },
		{ 171, 443, 33 },
		{ 171, 444, 33 },
		{ 171, 445, 33 },
		{ 171, 446, 33 },
		{ 171, 447, 33 },
		{ 171, 448, 33 },
		{ 171, 449, 33 },
		{ 171, 450, 33 },
		{ 171, 451, 33 },
		{ 171, 452, 33 },
		{ 171, 453, 33 },
		{ 171, 454, 33 },
		{ 171, 455, 33 },
		{ 171, 456, 33 },
		{ 171, 457, 33 },
		{ 171, 458, 33 },
		{ 171, 459, 33 },
		{ 171, 460, 33 },
		{ 171, 461, 33 },
		{ 171, 462, 33 },
		{ 171, 463, 33 },
		{ 171, 464, 33 },
		{ 171, 465, 33 },
		{ 171, 466, 33 },
		{ 171, 467, 33 },
		{ 171, 468, 33 },
		{ 171, 469, 33 },
		{ 171, 470, 33 },
		{ 171, 471, 33 },
		{ 171, 472, 33 },
		{ 171, 473, 33 },
		{ 171, 474, 33 },
		{ 171, 475, 33 },
		{ 171, 476, 33 },
		{ 171, 477, 33 },
		{ 171, 478, 33 },
		{ 171, 479, 33 },
		{ 171, 480, 33 },
		{ 171, 481, 33 },
		{ 171, 482, 33 },
		{ 171, 483, 33 },
		{ 171, 484, 33 },
		{ 171, 485, 33 },
		{ 171, 486, 33 },
		{ 171, 487, 33 },
		{ 171, 488, 33 },
		{ 171, 489, 33 },
		{ 171, 490, 33 },
		{ 171, 491, 33 },
		{ 171, 492, 33 },
		{ 171, 493, 33 },
		{ 171, 494, 33 },
		{ 171, 495, 33 },
		{ 171, 496, 33 },
		{ 171, 497, 33 },
		{ 171, 498, 33 },
		{ 171, 499, 33 },
		{ 171, 500, 33 },
		{ 171, 501, 33 },
		{ 171, 502, 33 },
		{ 171, 503, 33 },
		{ 171, 504, 33 },
		{ 171, 505, 33 },
		{ 171, 506, 33 },
		{ 171, 507, 33 },
		{ 171, 508, 33 },
		{ 171, 509, 33 },
		{ 171, 510, 33 },
		{ 171, 511, 33 },
		{ 171, 512, 33 },
		{ 172, 428, 33 },
		{ 172, 429, 33 },
		{ 172, 430, 33 },
		{ 172, 431, 33 },
		{ 172, 432, 33 },
		{ 172, 433, 33 },
		{ 172, 434, 33 },
		{ 172, 435, 33 },
		{ 172, 436, 33 },
		{ 172, 437, 33 },
		{ 172, 438, 33 },
		{ 172, 439, 33 },
		{ 172, 440, 33 },
		{ 172, 441, 33 },
		{ 172, 442, 33 },
		{ 172, 443, 33 },
		{ 172, 444, 33 },
		{ 172, 445, 33 },
		{ 172, 446, 33 },
		{ 172, 447, 33 },
		{ 172, 448, 33 },
		{ 172, 449, 33 },
		{ 172, 450, 33 },
		{ 172, 451, 33 },
		{ 172, 452, 33 },
		{ 172, 453, 33 },
		{ 172, 454, 33 },
		{ 172, 455, 33 },
		{ 172, 456, 33 },
		{ 172, 457, 33 },
		{ 172, 458, 33 },
		{ 172, 459, 33 },
		{ 172, 460, 33 },
		{ 172, 461, 33 },
		{ 172, 462, 33 },
		{ 172, 463, 33 },
		{ 172, 464, 33 },
		{ 172, 465, 33 },
		{ 172, 466, 33 },
		{ 172, 467, 33 },
		{ 172, 468, 33 },
		{ 172, 469, 33 },
		{ 172, 470, 33 },
		{ 172, 471, 33 },
		{ 172, 472, 33 },
		{ 172, 473, 33 },
		{ 172, 474, 33 },
		{ 172, 475, 33 },
		{ 172, 476, 33 },
		{ 172, 477, 33 },
		{ 172, 478, 33 },
		{ 172, 479, 33 },
		{ 172, 480, 33 },
		{ 172, 481, 33 },
		{ 172, 482, 33 },
		{ 172, 483, 33 },
		{ 172, 484, 33 },
		{ 172, 485, 33 },
		{ 172, 486, 33 },
		{ 172, 487, 33 },
		{ 172, 488, 33 },
		{ 172, 489, 33 },
		{ 172, 490, 33 },
		{ 172, 491, 33 },
		{ 172, 492, 33 },
		{ 172, 493, 33 },
		{ 172, 494, 33 },
		{ 172, 495, 33 },
		{ 172, 496, 33 },
		{ 172, 497, 33 },
		{ 172, 498, 33 },
		{ 172, 499, 33 },
		{ 172, 500, 33 },
		{ 172, 501, 33 },
		{ 172, 502, 33 },
		{ 172, 503, 33 },
		{ 172, 504, 33 },
		{ 172, 505, 33 },
		{ 172, 506, 33 },
		{ 172, 507, 33 },
		{ 172, 508, 33 },
		{ 172, 509, 33 },
		{ 172, 510, 33 },
		{ 172, 511, 33 },
		{ 172, 512, 33 },
		{ 173, 426, 33 },
		{ 173, 427, 33 },
		{ 173, 428, 33 },
		{ 173, 429, 33 },
		{ 173, 430, 33 },
		{ 173, 431, 33 },
		{ 173, 432, 33 },
		{ 173, 433, 33 },
		{ 173, 434, 33 },
		{ 173, 435, 33 },
		{ 173, 436, 33 },
		{ 173, 437, 33 },
		{ 173, 438, 33 },
		{ 173, 439, 33 },
		{ 173, 440, 33 },
		{ 173, 441, 33 },
		{ 173, 442, 33 },
		{ 173, 443, 33 },
		{ 173, 444, 33 },
		{ 173, 445, 33 },
		{ 173, 446, 33 },
		{ 173, 447, 33 },
		{ 173, 448, 33 },
		{ 173, 449, 33 },
		{ 173, 450, 33 },
		{ 173, 451, 33 },
		{ 173, 452, 33 },
		{ 173, 453, 33 },
		{ 173, 454, 33 },
		{ 173, 455, 33 },
		{ 173, 456, 33 },
		{ 173, 457, 33 },
		{ 173, 458, 33 },
		{ 173, 459, 33 },
		{ 173, 460, 33 },
		{ 173, 461, 33 },
		{ 173, 462, 33 },
		{ 173, 463, 33 },
		{ 173, 464, 33 },
		{ 173, 465, 33 },
		{ 173, 466, 33 },
		{ 173, 467, 33 },
		{ 173, 468, 33 },
		{ 173, 469, 33 },
		{ 173, 470, 33 },
		{ 173, 471, 33 },
		{ 173, 472, 33 },
		{ 173, 473, 33 },
		{ 173, 474, 33 },
		{ 173, 475, 33 },
		{ 173, 476, 33 },
		{ 173, 477, 33 },
		{ 173, 478, 33 },
		{ 173, 479, 33 },
		{ 173, 480, 33 },
		{ 173, 481, 33 },
		{ 173, 482, 33 },
		{ 173, 483, 33 },
		{ 173, 484, 33 },
		{ 173, 485, 33 },
		{ 173, 486, 33 },
		{ 173, 487, 33 },
		{ 173, 488, 33 },
		{ 173, 489, 33 },
		{ 173, 490, 33 },
		{ 173, 491, 33 },
		{ 173, 492, 33 },
		{ 173, 493, 33 },
		{ 173, 494, 33 },
		{ 173, 495, 33 },
		{ 173, 496, 33 },
		{ 173, 497, 33 },
		{ 173, 498, 33 },
		{ 173, 499, 33 },
		{ 173, 500, 33 },
		{ 173, 501, 33 },
		{ 173, 502, 33 },
		{ 173, 503, 33 },
		{ 173, 504, 33 },
		{ 173, 505, 33 },
		{ 173, 506, 33 },
		{ 173, 507, 33 },
		{ 173, 508, 33 },
		{ 173, 509, 33 },
		{ 173, 510, 33 },
		{ 173, 511, 33 },
		{ 173, 512, 33 },
		{ 174, 424, 33 },
		{ 174, 425, 33 },
		{ 174, 426, 33 },
		{ 174, 427, 33 },
		{ 174, 428, 33 },
		{ 174, 429, 33 },
		{ 174, 430, 33 },
		{ 174, 431, 33 },
		{ 174, 432, 33 },
		{ 174, 433, 33 },
		{ 174, 434, 33 },
		{ 174, 435, 33 },
		{ 174, 436, 33 },
		{ 174, 437, 33 },
		{ 174, 438, 33 },
		{ 174, 439, 33 },
		{ 174, 440, 33 },
		{ 174, 441, 33 },
		{ 174, 442, 33 },
		{ 174, 443, 33 },
		{ 174, 444, 33 },
		{ 174, 445, 33 },
		{ 174, 446, 33 },
		{ 174, 447, 33 },
		{ 174, 448, 33 },
		{ 174, 449, 33 },
		{ 174, 450, 33 },
		{ 174, 451, 33 },
		{ 174, 452, 33 },
		{ 174, 453, 33 },
		{ 174, 454, 33 },
		{ 174, 455, 33 },
		{ 174, 456, 33 },
		{ 174, 457, 33 },
		{ 174, 458, 33 },
		{ 174, 459, 33 },
		{ 174, 460, 33 },
		{ 174, 461, 33 },
		{ 174, 462, 33 },
		{ 174, 463, 33 },
		{ 174, 464, 33 },
		{ 174, 465, 33 },
		{ 174, 466, 33 },
		{ 174, 467, 33 },
		{ 174, 468, 33 },
		{ 174, 469, 33 },
		{ 174, 470, 33 },
		{ 174, 471, 33 },
		{ 174, 472, 33 },
		{ 174, 473, 33 },
		{ 174, 474, 33 },
		{ 174, 475, 33 },
		{ 174, 476, 33 },
		{ 174, 477, 33 },
		{ 174, 478, 33 },
		{ 174, 479, 33 },
		{ 174, 480, 33 },
		{ 174, 481, 33 },
		{ 174, 482, 33 },
		{ 174, 483, 33 },
		{ 174, 484, 33 },
		{ 174, 485, 33 },
		{ 174, 486, 33 },
		{ 174, 487, 33 },
		{ 174, 488, 33 },
		{ 174, 489, 33 },
		{ 174, 490, 33 },
		{ 174, 491, 33 },
		{ 174, 492, 33 },
		{ 174, 493, 33 },
		{ 174, 494, 33 },
		{ 174, 495, 33 },
		{ 174, 496, 33 },
		{ 174, 497, 33 },
		{ 174, 498, 33 },
		{ 174, 499, 33 },
		{ 174, 500, 33 },
		{ 174, 501, 33 },
		{ 174, 502, 33 },
		{ 174, 503, 33 },
		{ 174, 504, 33 },
		{ 174, 505, 33 },
		{ 174, 506, 33 },
		{ 174, 507, 33 },
		{ 174, 508, 33 },
		{ 174, 509, 33 },
		{ 174, 510, 33 },
		{ 174, 511, 33 },
		{ 174, 512, 33 },
		{ 175, 422, 33 },
		{ 175, 423, 33 },
		{ 175, 424, 33 },
		{ 175, 425, 33 },
		{ 175, 426, 33 },
		{ 175, 427, 33 },
		{ 175, 428, 33 },
		{ 175, 429, 33 },
		{ 175, 430, 33 },
		{ 175, 431, 33 },
		{ 175, 432, 33 },
		{ 175, 433, 33 },
		{ 175, 434, 33 },
		{ 175, 435, 33 },
		{ 175, 436, 33 },
		{ 175, 437, 33 },
		{ 175, 438, 33 },
		{ 175, 439, 33 },
		{ 175, 440, 33 },
		{ 175, 441, 33 },
		{ 175, 442, 33 },
		{ 175, 443, 33 },
		{ 175, 444, 33 },
		{ 175, 445, 33 },
		{ 175, 446, 33 },
		{ 175, 447, 33 },
		{ 175, 448, 33 },
		{ 175, 449, 33 },
		{ 175, 450, 33 },
		{ 175, 451, 33 },
		{ 175, 452, 33 },
		{ 175, 453, 33 },
		{ 175, 454, 33 },
		{ 175, 455, 33 },
		{ 175, 456, 33 },
		{ 175, 457, 33 },
		{ 175, 458, 33 },
		{ 175, 459, 33 },
		{ 175, 460, 33 },
		{ 175, 461, 33 },
		{ 175, 462, 33 },
		{ 175, 463, 33 },
		{ 175, 464, 33 },
		{ 175, 465, 33 },
		{ 175, 466, 33 },
		{ 175, 467, 33 },
		{ 175, 468, 33 },
		{ 175, 469, 33 },
		{ 175, 470, 33 },
		{ 175, 471, 33 },
		{ 175, 472, 33 },
		{ 175, 473, 33 },
		{ 175, 474, 33 },
		{ 175, 475, 33 },
		{ 175, 476, 33 },
		{ 175, 477, 33 },
		{ 175, 478, 33 },
		{ 175, 479, 33 },
		{ 175, 480, 33 },
		{ 175, 481, 33 },
		{ 175, 482, 33 },
		{ 175, 483, 33 },
		{ 175, 484, 33 },
		{ 175, 485, 33 },
		{ 175, 486, 33 },
		{ 175, 487, 33 },
		{ 175, 488, 33 },
		{ 175, 489, 33 },
		{ 175, 490, 33 },
		{ 175, 491, 33 },
		{ 175, 492, 33 },
		{ 175, 493, 33 },
		{ 175, 494, 33 },
		{ 175, 495, 33 },
		{ 175, 496, 33 },
		{ 175, 497, 33 },
		{ 175, 498, 33 },
		{ 175, 499, 33 },
		{ 175, 500, 33 },
		{ 175, 501, 33 },
		{ 175, 502, 33 },
		{ 175, 503, 33 },
		{ 175, 504, 33 },
		{ 175, 505, 33 },
		{ 175, 506, 33 },
		{ 175, 507, 33 },
		{ 175, 508, 33 },
		{ 175, 509, 33 },
		{ 175, 510, 33 },
		{ 175, 511, 33 },
		{ 175, 512, 33 },
		{ 176, 420, 33 },
		{ 176, 421, 33 },
		{ 176, 422, 33 },
		{ 176, 423, 33 },
		{ 176, 424, 33 },
		{ 176, 425, 33 },
		{ 176, 426, 33 },
		{ 176, 427, 33 },
		{ 176, 428, 33 },
		{ 176, 429, 33 },
		{ 176, 430, 33 },
		{ 176, 431, 33 },
		{ 176, 432, 33 },
		{ 176, 433, 33 },
		{ 176, 434, 33 },
		{ 176, 435, 33 },
		{ 176, 436, 33 },
		{ 176, 437, 33 },
		{ 176, 438, 33 },
		{ 176, 439, 33 },
		{ 176, 440, 33 },
		{ 176, 441, 33 },
		{ 176, 442, 33 },
		{ 176, 443, 33 },
		{ 176, 444, 33 },
		{ 176, 445, 33 },
		{ 176, 446, 33 },
		{ 176, 447, 33 },
		{ 176, 448, 33 },
		{ 176, 449, 33 },
		{ 176, 450, 33 },
		{ 176, 451, 33 },
		{ 176, 452, 33 },
		{ 176, 453, 33 },
		{ 176, 454, 33 },
		{ 176, 455, 33 },
		{ 176, 456, 33 },
		{ 176, 457, 33 },
		{ 176, 458, 33 },
		{ 176, 459, 33 },
		{ 176, 460, 33 },
		{ 176, 461, 33 },
		{ 176, 462, 33 },
		{ 176, 463, 33 },
		{ 176, 464, 33 },
		{ 176, 465, 33 },
		{ 176, 466, 33 },
		{ 176, 467, 33 },
		{ 176, 468, 33 },
		{ 176, 469, 33 },
		{ 176, 470, 33 },
		{ 176, 471, 33 },
		{ 176, 472, 33 },
		{ 176, 473, 33 },
		{ 176, 474, 33 },
		{ 176, 475, 33 },
		{ 176, 476, 33 },
		{ 176, 477, 33 },
		{ 176, 478, 33 },
		{ 176, 479, 33 },
		{ 176, 480, 33 },
		{ 176, 481, 33 },
		{ 176, 482, 33 },
		{ 176, 483, 33 },
		{ 176, 484, 33 },
		{ 176, 485, 33 },
		{ 176, 486, 33 },
		{ 176, 487, 33 },
		{ 176, 488, 33 },
		{ 176, 489, 33 },
		{ 176, 490, 33 },
		{ 176, 491, 33 },
		{ 176, 492, 33 },
		{ 176, 493, 33 },
		{ 176, 494, 33 },
		{ 176, 495, 33 },
		{ 176, 496, 33 },
		{ 176, 497, 33 },
		{ 176, 498, 33 },
		{ 176, 499, 33 },
		{ 176, 500, 33 },
		{ 176, 501, 33 },
		{ 176, 502, 33 },
		{ 176, 503, 33 },
		{ 176, 504, 33 },
		{ 176, 505, 33 },
		{ 176, 506, 33 },
		{ 176, 507, 33 },
		{ 176, 508, 33 },
		{ 176, 509, 33 },
		{ 176, 510, 33 },
		{ 176, 511, 33 },
		{ 176, 512, 33 },
		{ 177, 418, 33 },
		{ 177, 419, 33 },
		{ 177, 420, 33 },
		{ 177, 421, 33 },
		{ 177, 422, 33 },
		{ 177, 423, 33 },
		{ 177, 424, 33 },
		{ 177, 425, 33 },
		{ 177, 426, 33 },
		{ 177, 427, 33 },
		{ 177, 428, 33 },
		{ 177, 429, 33 },
		{ 177, 430, 33 },
		{ 177, 431, 33 },
		{ 177, 432, 33 },
		{ 177, 433, 33 },
		{ 177, 434, 33 },
		{ 177, 435, 33 },
		{ 177, 436, 33 },
		{ 177, 437, 33 },
		{ 177, 438, 33 },
		{ 177, 439, 33 },
		{ 177, 440, 33 },
		{ 177, 441, 33 },
		{ 177, 442, 33 },
		{ 177, 443, 33 },
		{ 177, 444, 33 },
		{ 177, 445, 33 },
		{ 177, 446, 33 },
		{ 177, 447, 33 },
		{ 177, 448, 33 },
		{ 177, 449, 33 },
		{ 177, 450, 33 },
		{ 177, 451, 33 },
		{ 177, 452, 33 },
		{ 177, 453, 33 },
		{ 177, 454, 33 },
		{ 177, 455, 33 },
		{ 177, 456, 33 },
		{ 177, 457, 33 },
		{ 177, 458, 33 },
		{ 177, 459, 33 },
		{ 177, 460, 33 },
		{ 177, 461, 33 },
		{ 177, 462, 33 },
		{ 177, 463, 33 },
		{ 177, 464, 33 },
		{ 177, 465, 33 },
		{ 177, 466, 33 },
		{ 177, 467, 33 },
		{ 177, 468, 33 },
		{ 177, 469, 33 },
		{ 177, 470, 33 },
		{ 177, 471, 33 },
		{ 177, 472, 33 },
		{ 177, 473, 33 },
		{ 177, 474, 33 },
		{ 177, 475, 33 },
		{ 177, 476, 33 },
		{ 177, 477, 33 },
		{ 177, 478, 33 },
		{ 177, 479, 33 },
		{ 177, 480, 33 },
		{ 177, 481, 33 },
		{ 177, 482, 33 },
		{ 177, 483, 33 },
		{ 177, 484, 33 },
		{ 177, 485, 33 },
		{ 177, 486, 33 },
		{ 177, 487, 33 },
		{ 177, 488, 33 },
		{ 177, 489, 33 },
		{ 177, 490, 33 },
		{ 177, 491, 33 },
		{ 177, 492, 33 },
		{ 177, 493, 33 },
		{ 177, 494, 33 },
		{ 177, 495, 33 },
		{ 177, 496, 33 },
		{ 177, 497, 33 },
		{ 177, 498, 33 },
		{ 177, 499, 33 },
		{ 177, 500, 33 },
		{ 177, 501, 33 },
		{ 177, 502, 33 },
		{ 177, 503, 33 },
		{ 177, 504, 33 },
		{ 177, 505, 33 },
		{ 177, 506, 33 },
		{ 177, 507, 33 },
		{ 177, 508, 33 },
		{ 177, 509, 33 },
		{ 177, 510, 33 },
		{ 177, 511, 33 },
		{ 177, 512, 33 },
		{ 178, 416, 33 },
		{ 178, 417, 33 },
		{ 178, 418, 33 },
		{ 178, 419, 33 },
		{ 178, 420, 33 },
		{ 178, 421, 33 },
		{ 178, 422, 33 },
		{ 178, 423, 33 },
		{ 178, 424, 33 },
		{ 178, 425, 33 },
		{ 178, 426, 33 },
		{ 178, 427, 33 },
		{ 178, 428, 33 },
		{ 178, 429, 33 },
		{ 178, 430, 33 },
		{ 178, 431, 33 },
		{ 178, 432, 33 },
		{ 178, 433, 33 },
		{ 178, 434, 33 },
		{ 178, 435, 33 },
		{ 178, 436, 33 },
		{ 178, 437, 33 },
		{ 178, 438, 33 },
		{ 178, 439, 33 },
		{ 178, 440, 33 },
		{ 178, 441, 33 },
		{ 178, 442, 33 },
		{ 178, 443, 33 },
		{ 178, 444, 33 },
		{ 178, 445, 33 },
		{ 178, 446, 33 },
		{ 178, 447, 33 },
		{ 178, 448, 33 },
		{ 178, 449, 33 },
		{ 178, 450, 33 },
		{ 178, 451, 33 },
		{ 178, 452, 33 },
		{ 178, 453, 33 },
		{ 178, 454, 33 },
		{ 178, 455, 33 },
		{ 178, 456, 33 },
		{ 178, 457, 33 },
		{ 178, 458, 33 },
		{ 178, 459, 33 },
		{ 178, 460, 33 },
		{ 178, 461, 33 },
		{ 178, 462, 33 },
		{ 178, 463, 33 },
		{ 178, 464, 33 },
		{ 178, 465, 33 },
		{ 178, 466, 33 },
		{ 178, 467, 33 },
		{ 178, 468, 33 },
		{ 178, 469, 33 },
		{ 178, 470, 33 },
		{ 178, 471, 33 },
		{ 178, 472, 33 },
		{ 178, 473, 33 },
		{ 178, 474, 33 },
		{ 178, 475, 33 },
		{ 178, 476, 33 },
		{ 178, 477, 33 },
		{ 178, 478, 33 },
		{ 178, 479, 33 },
		{ 178, 480, 33 },
		{ 178, 481, 33 },
		{ 178, 482, 33 },
		{ 178, 483, 33 },
		{ 178, 484, 33 },
		{ 178, 485, 33 },
		{ 178, 486, 33 },
		{ 178, 487, 33 },
		{ 178, 488, 33 },
		{ 178, 489, 33 },
		{ 178, 490, 33 },
		{ 178, 491, 33 },
		{ 178, 492, 33 },
		{ 178, 493, 33 },
		{ 178, 494, 33 },
		{ 178, 495, 33 },
		{ 178, 496, 33 },
		{ 178, 497, 33 },
		{ 178, 498, 33 },
		{ 178, 499, 33 },
		{ 178, 500, 33 },
		{ 178, 501, 33 },
		{ 178, 502, 33 },
		{ 178, 503, 33 },
		{ 178, 504, 33 },
		{ 178, 505, 33 },
		{ 178, 506, 33 },
		{ 178, 507, 33 },
		{ 178, 508, 33 },
		{ 178, 509, 33 },
		{ 178, 510, 33 },
		{ 178, 511, 33 },
		{ 178, 512, 33 },
		{ 179, 414, 33 },
		{ 179, 415, 33 },
		{ 179, 416, 33 },
		{ 179, 417, 33 },
		{ 179, 418, 33 },
		{ 179, 419, 33 },
		{ 179, 420, 33 },
		{ 179, 421, 33 },
		{ 179, 422, 33 },
		{ 179, 423, 33 },
		{ 179, 424, 33 },
		{ 179, 425, 33 },
		{ 179, 426, 33 },
		{ 179, 427, 33 },
		{ 179, 428, 33 },
		{ 179, 429, 33 },
		{ 179, 430, 33 },
		{ 179, 431, 33 },
		{ 179, 432, 33 },
		{ 179, 433, 33 },
		{ 179, 434, 33 },
		{ 179, 435, 33 },
		{ 179, 436, 33 },
		{ 179, 437, 33 },
		{ 179, 438, 33 },
		{ 179, 439, 33 },
		{ 179, 440, 33 },
		{ 179, 441, 33 },
		{ 179, 442, 33 },
		{ 179, 443, 33 },
		{ 179, 444, 33 },
		{ 179, 445, 33 },
		{ 179, 446, 33 },
		{ 179, 447, 33 },
		{ 179, 448, 33 },
		{ 179, 449, 33 },
		{ 179, 450, 33 },
		{ 179, 451, 33 },
		{ 179, 452, 33 },
		{ 179, 453, 33 },
		{ 179, 454, 33 },
		{ 179, 455, 33 },
		{ 179, 456, 33 },
		{ 179, 457, 33 },
		{ 179, 458, 33 },
		{ 179, 459, 33 },
		{ 179, 460, 33 },
		{ 179, 461, 33 },
		{ 179, 462, 33 },
		{ 179, 463, 33 },
		{ 179, 464, 33 },
		{ 179, 465, 33 },
		{ 179, 466, 33 },
		{ 179, 467, 33 },
		{ 179, 468, 33 },
		{ 179, 469, 33 },
		{ 179, 470, 33 },
		{ 179, 471, 33 },
		{ 179, 472, 33 },
		{ 179, 473, 33 },
		{ 179, 474, 33 },
		{ 179, 475, 33 },
		{ 179, 476, 33 },
		{ 179, 477, 33 },
		{ 179, 478, 33 },
		{ 179, 479, 33 },
		{ 179, 480, 33 },
		{ 179, 481, 33 },
		{ 179, 482, 33 },
		{ 179, 483, 33 },
		{ 179, 484, 33 },
		{ 179, 485, 33 },
		{ 179, 486, 33 },
		{ 179, 487, 33 },
		{ 179, 488, 33 },
		{ 179, 489, 33 },
		{ 179, 490, 33 },
		{ 179, 491, 33 },
		{ 179, 492, 33 },
		{ 179, 493, 33 },
		{ 179, 494, 33 },
		{ 179, 495, 33 },
		{ 179, 496, 33 },
		{ 179, 497, 33 },
		{ 179, 498, 33 },
		{ 179, 499, 33 },
		{ 179, 500, 33 },
		{ 179, 501, 33 },
		{ 179, 502, 33 },
		{ 179, 503, 33 },
		{ 179, 504, 33 },
		{ 179, 505, 33 },
		{ 179, 506, 33 },
		{ 179, 507, 33 },
		{ 179, 508, 33 },
		{ 179, 509, 33 },
		{ 179, 510, 33 },
		{ 179, 511, 33 },
		{ 179, 512, 33 },
		{ 180, 412, 33 },
		{ 180, 413, 33 },
		{ 180, 414, 33 },
		{ 180, 415, 33 },
		{ 180, 416, 33 },
		{ 180, 417, 33 },
		{ 180, 418, 33 },
		{ 180, 419, 33 },
		{ 180, 420, 33 },
		{ 180, 421, 33 },
		{ 180, 422, 33 },
		{ 180, 423, 33 },
		{ 180, 424, 33 },
		{ 180, 425, 33 },
		{ 180, 426, 33 },
		{ 180, 427, 33 },
		{ 180, 428, 33 },
		{ 180, 429, 33 },
		{ 180, 430, 33 },
		{ 180, 431, 33 },
		{ 180, 432, 33 },
		{ 180, 433, 33 },
		{ 180, 434, 33 },
		{ 180, 435, 33 },
		{ 180, 436, 33 },
		{ 180, 437, 33 },
		{ 180, 438, 33 },
		{ 180, 439, 33 },
		{ 180, 440, 33 },
		{ 180, 441, 33 },
		{ 180, 442, 33 },
		{ 180, 443, 33 },
		{ 180, 444, 33 },
		{ 180, 445, 33 },
		{ 180, 446, 33 },
		{ 180, 447, 33 },
		{ 180, 448, 33 },
		{ 180, 449, 33 },
		{ 180, 450, 33 },
		{ 180, 451, 33 },
		{ 180, 452, 33 },
		{ 180, 453, 33 },
		{ 180, 454, 33 },
		{ 180, 455, 33 },
		{ 180, 456, 33 },
		{ 180, 457, 33 },
		{ 180, 458, 33 },
		{ 180, 459, 33 },
		{ 180, 460, 33 },
		{ 180, 461, 33 },
		{ 180, 462, 33 },
		{ 180, 463, 33 },
		{ 180, 464, 33 },
		{ 180, 465, 33 },
		{ 180, 466, 33 },
		{ 180, 467, 33 },
		{ 180, 468, 33 },
		{ 180, 469, 33 },
		{ 180, 470, 33 },
		{ 180, 471, 33 },
		{ 180, 472, 33 },
		{ 180, 473, 33 },
		{ 180, 474, 33 },
		{ 180, 475, 33 },
		{ 180, 476, 33 },
		{ 180, 477, 33 },
		{ 180, 478, 33 },
		{ 180, 479, 33 },
		{ 180, 480, 33 },
		{ 180, 481, 33 },
		{ 180, 482, 33 },
		{ 180, 483, 33 },
		{ 180, 484, 33 },
		{ 180, 485, 33 },
		{ 180, 486, 33 },
		{ 180, 487, 33 },
		{ 180, 488, 33 },
		{ 180, 489, 33 },
		{ 180, 490, 33 },
		{ 180, 491, 33 },
		{ 180, 492, 33 },
		{ 180, 493, 33 },
		{ 180, 494, 33 },
		{ 180, 495, 33 },
		{ 180, 496, 33 },
		{ 180, 497, 33 },
		{ 180, 498, 33 },
		{ 180, 499, 33 },
		{ 180, 500, 33 },
		{ 180, 501, 33 },
		{ 180, 502, 33 },
		{ 180, 503, 33 },
		{ 180, 504, 33 },
		{ 180, 505, 33 },
		{ 180, 506, 33 },
		{ 180, 507, 33 },
		{ 180, 508, 33 },
		{ 180, 509, 33 },
		{ 180, 510, 33 },
		{ 180, 511, 33 },
		{ 180, 512, 33 },
		{ 181, 410, 33 },
		{ 181, 411, 33 },
		{ 181, 412, 33 },
		{ 181, 413, 33 },
		{ 181, 414, 33 },
		{ 181, 415, 33 },
		{ 181, 416, 33 },
		{ 181, 417, 33 },
		{ 181, 418, 33 },
		{ 181, 419, 33 },
		{ 181, 420, 33 },
		{ 181, 421, 33 },
		{ 181, 422, 33 },
		{ 181, 423, 33 },
		{ 181, 424, 33 },
		{ 181, 425, 33 },
		{ 181, 426, 33 },
		{ 181, 427, 33 },
		{ 181, 428, 33 },
		{ 181, 429, 33 },
		{ 181, 430, 33 },
		{ 181, 431, 33 },
		{ 181, 432, 33 },
		{ 181, 433, 33 },
		{ 181, 434, 33 },
		{ 181, 435, 33 },
		{ 181, 436, 33 },
		{ 181, 437, 33 },
		{ 181, 438, 33 },
		{ 181, 439, 33 },
		{ 181, 440, 33 },
		{ 181, 441, 33 },
		{ 181, 442, 33 },
		{ 181, 443, 33 },
		{ 181, 444, 33 },
		{ 181, 445, 33 },
		{ 181, 446, 33 },
		{ 181, 447, 33 },
		{ 181, 448, 33 },
		{ 181, 449, 33 },
		{ 181, 450, 33 },
		{ 181, 451, 33 },
		{ 181, 452, 33 },
		{ 181, 453, 33 },
		{ 181, 454, 33 },
		{ 181, 455, 33 },
		{ 181, 456, 33 },
		{ 181, 457, 33 },
		{ 181, 458, 33 },
		{ 181, 459, 33 },
		{ 181, 460, 33 },
		{ 181, 461, 33 },
		{ 181, 462, 33 },
		{ 181, 463, 33 },
		{ 181, 464, 33 },
		{ 181, 465, 33 },
		{ 181, 466, 33 },
		{ 181, 467, 33 },
		{ 181, 468, 33 },
		{ 181, 469, 33 },
		{ 181, 470, 33 },
		{ 181, 471, 33 },
		{ 181, 472, 33 },
		{ 181, 473, 33 },
		{ 181, 474, 33 },
		{ 181, 475, 33 },
		{ 181, 476, 33 },
		{ 181, 477, 33 },
		{ 181, 478, 33 },
		{ 181, 479, 33 },
		{ 181, 480, 33 },
		{ 181, 481, 33 },
		{ 181, 482, 33 },
		{ 181, 483, 33 },
		{ 181, 484, 33 },
		{ 181, 485, 33 },
		{ 181, 486, 33 },
		{ 181, 487, 33 },
		{ 181, 488, 33 },
		{ 181, 489, 33 },
		{ 181, 490, 33 },
		{ 181, 491, 33 },
		{ 181, 492, 33 },
		{ 181, 493, 33 },
		{ 181, 494, 33 },
		{ 181, 495, 33 },
		{ 181, 496, 33 },
		{ 181, 497, 33 },
		{ 181, 498, 33 },
		{ 181, 499, 33 },
		{ 181, 500, 33 },
		{ 181, 501, 33 },
		{ 181, 502, 33 },
		{ 181, 503, 33 },
		{ 181, 504, 33 },
		{ 181, 505, 33 },
		{ 181, 506, 33 },
		{ 181, 507, 33 },
		{ 181, 508, 33 },
		{ 181, 509, 33 },
		{ 181, 510, 33 },
		{ 181, 511, 33 },
		{ 181, 512, 33 },
		{ 182, 408, 33 },
		{ 182, 409, 33 },
		{ 182, 410, 33 },
		{ 182, 411, 33 },
		{ 182, 412, 33 },
		{ 182, 413, 33 },
		{ 182, 414, 33 },
		{ 182, 415, 33 },
		{ 182, 416, 33 },
		{ 182, 417, 33 },
		{ 182, 418, 33 },
		{ 182, 419, 33 },
		{ 182, 420, 33 },
		{ 182, 421, 33 },
		{ 182, 422, 33 },
		{ 182, 423, 33 },
		{ 182, 424, 33 },
		{ 182, 425, 33 },
		{ 182, 426, 33 },
		{ 182, 427, 33 },
		{ 182, 428, 33 },
		{ 182, 429, 33 },
		{ 182, 430, 33 },
		{ 182, 431, 33 },
		{ 182, 432, 33 },
		{ 182, 433, 33 },
		{ 182, 434, 33 },
		{ 182, 435, 33 },
		{ 182, 436, 33 },
		{ 182, 437, 33 },
		{ 182, 438, 33 },
		{ 182, 439, 33 },
		{ 182, 440, 33 },
		{ 182, 441, 33 },
		{ 182, 442, 33 },
		{ 182, 443, 33 },
		{ 182, 444, 33 },
		{ 182, 445, 33 },
		{ 182, 446, 33 },
		{ 182, 447, 33 },
		{ 182, 448, 33 },
		{ 182, 449, 33 },
		{ 182, 450, 33 },
		{ 182, 451, 33 },
		{ 182, 452, 33 },
		{ 182, 453, 33 },
		{ 182, 454, 33 },
		{ 182, 455, 33 },
		{ 182, 456, 33 },
		{ 182, 457, 33 },
		{ 182, 458, 33 },
		{ 182, 459, 33 },
		{ 182, 460, 33 },
		{ 182, 461, 33 },
		{ 182, 462, 33 },
		{ 182, 463, 33 },
		{ 182, 464, 33 },
		{ 182, 465, 33 },
		{ 182, 466, 33 },
		{ 182, 467, 33 },
		{ 182, 468, 33 },
		{ 182, 469, 33 },
		{ 182, 470, 33 },
		{ 182, 471, 33 },
		{ 182, 472, 33 },
		{ 182, 473, 33 },
		{ 182, 474, 33 },
		{ 182, 475, 33 },
		{ 182, 476, 33 },
		{ 182, 477, 33 },
		{ 182, 478, 33 },
		{ 182, 479, 33 },
		{ 182, 480, 33 },
		{ 182, 481, 33 },
		{ 182, 482, 33 },
		{ 182, 483, 33 },
		{ 182, 484, 33 },
		{ 182, 485, 33 },
		{ 182, 486, 33 },
		{ 182, 487, 33 },
		{ 182, 488, 33 },
		{ 182, 489, 33 },
		{ 182, 490, 33 },
		{ 182, 491, 33 },
		{ 182, 492, 33 },
		{ 182, 493, 33 },
		{ 182, 494, 33 },
		{ 182, 495, 33 },
		{ 182, 496, 33 },
		{ 182, 497, 33 },
		{ 182, 498, 33 },
		{ 182, 499, 33 },
		{ 182, 500, 33 },
		{ 182, 501, 33 },
		{ 182, 502, 33 },
		{ 182, 503, 33 },
		{ 182, 504, 33 },
		{ 182, 505, 33 },
		{ 182, 506, 33 },
		{ 182, 507, 33 },
		{ 182, 508, 33 },
		{ 182, 509, 33 },
		{ 182, 510, 33 },
		{ 182, 511, 33 },
		{ 182, 512, 33 },
		{ 183, 406, 33 },
		{ 183, 407, 33 },
		{ 183, 408, 33 },
		{ 183, 409, 33 },
		{ 183, 410, 33 },
		{ 183, 411, 33 },
		{ 183, 412, 33 },
		{ 183, 413, 33 },
		{ 183, 414, 33 },
		{ 183, 415, 33 },
		{ 183, 416, 33 },
		{ 183, 417, 33 },
		{ 183, 418, 33 },
		{ 183, 419, 33 },
		{ 183, 420, 33 },
		{ 183, 421, 33 },
		{ 183, 422, 33 },
		{ 183, 423, 33 },
		{ 183, 424, 33 },
		{ 183, 425, 33 },
		{ 183, 426, 33 },
		{ 183, 427, 33 },
		{ 183, 428, 33 },
		{ 183, 429, 33 },
		{ 183, 430, 33 },
		{ 183, 431, 33 },
		{ 183, 432, 33 },
		{ 183, 433, 33 },
		{ 183, 434, 33 },
		{ 183, 435, 33 },
		{ 183, 436, 33 },
		{ 183, 437, 33 },
		{ 183, 438, 33 },
		{ 183, 439, 33 },
		{ 183, 440, 33 },
		{ 183, 441, 33 },
		{ 183, 442, 33 },
		{ 183, 443, 33 },
		{ 183, 444, 33 },
		{ 183, 445, 33 },
		{ 183, 446, 33 },
		{ 183, 447, 33 },
		{ 183, 448, 33 },
		{ 183, 449, 33 },
		{ 183, 450, 33 },
		{ 183, 451, 33 },
		{ 183, 452, 33 },
		{ 183, 453, 33 },
		{ 183, 454, 33 },
		{ 183, 455, 33 },
		{ 183, 456, 33 },
		{ 183, 457, 33 },
		{ 183, 458, 33 },
		{ 183, 459, 33 },
		{ 183, 460, 33 },
		{ 183, 461, 33 },
		{ 183, 462, 33 },
		{ 183, 463, 33 },
		{ 183, 464, 33 },
		{ 183, 465, 33 },
		{ 183, 466, 33 },
		{ 183, 467, 33 },
		{ 183, 468, 33 },
		{ 183, 469, 33 },
		{ 183, 470, 33 },
		{ 183, 471, 33 },
		{ 183, 472, 33 },
		{ 183, 473, 33 },
		{ 183, 474, 33 },
		{ 183, 475, 33 },
		{ 183, 476, 33 },
		{ 183, 477, 33 },
		{ 183, 478, 33 },
		{ 183, 479, 33 },
		{ 183, 480, 33 },
		{ 183, 481, 33 },
		{ 183, 482, 33 },
		{ 183, 483, 33 },
		{ 183, 484, 33 },
		{ 183, 485, 33 },
		{ 183, 486, 33 },
		{ 183, 487, 33 },
		{ 183, 488, 33 },
		{ 183, 489, 33 },
		{ 183, 490, 33 },
		{ 183, 491, 33 },
		{ 183, 492, 33 },
		{ 183, 493, 33 },
		{ 183, 494, 33 },
		{ 183, 495, 33 },
		{ 183, 496, 33 },
		{ 183, 497, 33 },
		{ 183, 498, 33 },
		{ 183, 499, 33 },
		{ 183, 500, 33 },
		{ 183, 501, 33 },
		{ 183, 502, 33 },
		{ 183, 503, 33 },
		{ 183, 504, 33 },
		{ 183, 505, 33 },
		{ 183, 506, 33 },
		{ 183, 507, 33 },
		{ 183, 508, 33 },
		{ 183, 509, 33 },
		{ 183, 510, 33 },
		{ 183, 511, 33 },
		{ 183, 512, 33 },
		{ 184, 404, 33 },
		{ 184, 405, 33 },
		{ 184, 406, 33 },
		{ 184, 407, 33 },
		{ 184, 408, 33 },
		{ 184, 409, 33 },
		{ 184, 410, 33 },
		{ 184, 411, 33 },
		{ 184, 412, 33 },
		{ 184, 413, 33 },
		{ 184, 414, 33 },
		{ 184, 415, 33 },
		{ 184, 416, 33 },
		{ 184, 417, 33 },
		{ 184, 418, 33 },
		{ 184, 419, 33 },
		{ 184, 420, 33 },
		{ 184, 421, 33 },
		{ 184, 422, 33 },
		{ 184, 423, 33 },
		{ 184, 424, 33 },
		{ 184, 425, 33 },
		{ 184, 426, 33 },
		{ 184, 427, 33 },
		{ 184, 428, 33 },
		{ 184, 429, 33 },
		{ 184, 430, 33 },
		{ 184, 431, 33 },
		{ 184, 432, 33 },
		{ 184, 433, 33 },
		{ 184, 434, 33 },
		{ 184, 435, 33 },
		{ 184, 436, 33 },
		{ 184, 437, 33 },
		{ 184, 438, 33 },
		{ 184, 439, 33 },
		{ 184, 440, 33 },
		{ 184, 441, 33 },
		{ 184, 442, 33 },
		{ 184, 443, 33 },
		{ 184, 444, 33 },
		{ 184, 445, 33 },
		{ 184, 446, 33 },
		{ 184, 447, 33 },
		{ 184, 448, 33 },
		{ 184, 449, 33 },
		{ 184, 450, 33 },
		{ 184, 451, 33 },
		{ 184, 452, 33 },
		{ 184, 453, 33 },
		{ 184, 454, 33 },
		{ 184, 455, 33 },
		{ 184, 456, 33 },
		{ 184, 457, 33 },
		{ 184, 458, 33 },
		{ 184, 459, 33 },
		{ 184, 460, 33 },
		{ 184, 461, 33 },
		{ 184, 462, 33 },
		{ 184, 463, 33 },
		{ 184, 464, 33 },
		{ 184, 465, 33 },
		{ 184, 466, 33 },
		{ 184, 467, 33 },
		{ 184, 468, 33 },
		{ 184, 469, 33 },
		{ 184, 470, 33 },
		{ 184, 471, 33 },
		{ 184, 472, 33 },
		{ 184, 473, 33 },
		{ 184, 474, 33 },
		{ 184, 475, 33 },
		{ 184, 476, 33 },
		{ 184, 477, 33 },
		{ 184, 478, 33 },
		{ 184, 479, 33 },
		{ 184, 480, 33 },
		{ 184, 481, 33 },
		{ 184, 482, 33 },
		{ 184, 483, 33 },
		{ 184, 484, 33 },
		{ 184, 485, 33 },
		{ 184, 486, 33 },
		{ 184, 487, 33 },
		{ 184, 488, 33 },
		{ 184, 489, 33 },
		{ 184, 490, 33 },
		{ 184, 491, 33 },
		{ 184, 492, 33 },
		{ 184, 493, 33 },
		{ 184, 494, 33 },
		{ 184, 495, 33 },
		{ 184, 496, 33 },
		{ 184, 497, 33 },
		{ 184, 498, 33 },
		{ 184, 499, 33 },
		{ 184, 500, 33 },
		{ 184, 501, 33 },
		{ 184, 502, 33 },
		{ 184, 503, 33 },
		{ 184, 504, 33 },
		{ 184, 505, 33 },
		{ 184, 506, 33 },
		{ 184, 507, 33 },
		{ 184, 508, 33 },
		{ 184, 509, 33 },
		{ 184, 510, 33 },
		{ 184, 511, 33 },
		{ 184, 512, 33 },
		{ 185, 402, 33 },
		{ 185, 403, 33 },
		{ 185, 404, 33 },
		{ 185, 405, 33 },
		{ 185, 406, 33 },
		{ 185, 407, 33 },
		{ 185, 408, 33 },
		{ 185, 409, 33 },
		{ 185, 410, 33 },
		{ 185, 411, 33 },
		{ 185, 412, 33 },
		{ 185, 413, 33 },
		{ 185, 414, 33 },
		{ 185, 415, 33 },
		{ 185, 416, 33 },
		{ 185, 417, 33 },
		{ 185, 418, 33 },
		{ 185, 419, 33 },
		{ 185, 420, 33 },
		{ 185, 421, 33 },
		{ 185, 422, 33 },
		{ 185, 423, 33 },
		{ 185, 424, 33 },
		{ 185, 425, 33 },
		{ 185, 426, 33 },
		{ 185, 427, 33 },
		{ 185, 428, 33 },
		{ 185, 429, 33 },
		{ 185, 430, 33 },
		{ 185, 431, 33 },
		{ 185, 432, 33 },
		{ 185, 433, 33 },
		{ 185, 434, 33 },
		{ 185, 435, 33 },
		{ 185, 436, 33 },
		{ 185, 437, 33 },
		{ 185, 438, 33 },
		{ 185, 439, 33 },
		{ 185, 440, 33 },
		{ 185, 441, 33 },
		{ 185, 442, 33 },
		{ 185, 443, 33 },
		{ 185, 444, 33 },
		{ 185, 445, 33 },
		{ 185, 446, 33 },
		{ 185, 447, 33 },
		{ 185, 448, 33 },
		{ 185, 449, 33 },
		{ 185, 450, 33 },
		{ 185, 451, 33 },
		{ 185, 452, 33 },
		{ 185, 453, 33 },
		{ 185, 454, 33 },
		{ 185, 455, 33 },
		{ 185, 456, 33 },
		{ 185, 457, 33 },
		{ 185, 458, 33 },
		{ 185, 459, 33 },
		{ 185, 460, 33 },
		{ 185, 461, 33 },
		{ 185, 462, 33 },
		{ 185, 463, 33 },
		{ 185, 464, 33 },
		{ 185, 465, 33 },
		{ 185, 466, 33 },
		{ 185, 467, 33 },
		{ 185, 468, 33 },
		{ 185, 469, 33 },
		{ 185, 470, 33 },
		{ 185, 471, 33 },
		{ 185, 472, 33 },
		{ 185, 473, 33 },
		{ 185, 474, 33 },
		{ 185, 475, 33 },
		{ 185, 476, 33 },
		{ 185, 477, 33 },
		{ 185, 478, 33 },
		{ 185, 479, 33 },
		{ 185, 480, 33 },
		{ 185, 481, 33 },
		{ 185, 482, 33 },
		{ 185, 483, 33 },
		{ 185, 484, 33 },
		{ 185, 485, 33 },
		{ 185, 486, 33 },
		{ 185, 487, 33 },
		{ 185, 488, 33 },
		{ 185, 489, 33 },
		{ 185, 490, 33 },
		{ 185, 491, 33 },
		{ 185, 492, 33 },
		{ 185, 493, 33 },
		{ 185, 494, 33 },
		{ 185, 495, 33 },
		{ 185, 496, 33 },
		{ 185, 497, 33 },
		{ 185, 498, 33 },
		{ 185, 499, 33 },
		{ 185, 500, 33 },
		{ 185, 501, 33 },
		{ 185, 502, 33 },
		{ 185, 503, 33 },
		{ 185, 504, 33 },
		{ 185, 505, 33 },
		{ 185, 506, 33 },
		{ 185, 507, 33 },
		{ 185, 508, 33 },
		{ 185, 509, 33 },
		{ 185, 510, 33 },
		{ 185, 511, 33 },
		{ 185, 512, 33 },
		{ 186, 400, 33 },
		{ 186, 401, 33 },
		{ 186, 402, 33 },
		{ 186, 403, 33 },
		{ 186, 404, 33 },
		{ 186, 405, 33 },
		{ 186, 406, 33 },
		{ 186, 407, 33 },
		{ 186, 408, 33 },
		{ 186, 409, 33 },
		{ 186, 410, 33 },
		{ 186, 411, 33 },
		{ 186, 412, 33 },
		{ 186, 413, 33 },
		{ 186, 414, 33 },
		{ 186, 415, 33 },
		{ 186, 416, 33 },
		{ 186, 417, 33 },
		{ 186, 418, 33 },
		{ 186, 419, 33 },
		{ 186, 420, 33 },
		{ 186, 421, 33 },
		{ 186, 422, 33 },
		{ 186, 423, 33 },
		{ 186, 424, 33 },
		{ 186, 425, 33 },
		{ 186, 426, 33 },
		{ 186, 427, 33 },
		{ 186, 428, 33 },
		{ 186, 429, 33 },
		{ 186, 430, 33 },
		{ 186, 431, 33 },
		{ 186, 432, 33 },
		{ 186, 433, 33 },
		{ 186, 434, 33 },
		{ 186, 435, 33 },
		{ 186, 436, 33 },
		{ 186, 437, 33 },
		{ 186, 438, 33 },
		{ 186, 439, 33 },
		{ 186, 440, 33 },
		{ 186, 441, 33 },
		{ 186, 442, 33 },
		{ 186, 443, 33 },
		{ 186, 444, 33 },
		{ 186, 445, 33 },
		{ 186, 446, 33 },
		{ 186, 447, 33 },
		{ 186, 448, 33 },
		{ 186, 449, 33 },
		{ 186, 450, 33 },
		{ 186, 451, 33 },
		{ 186, 452, 33 },
		{ 186, 453, 33 },
		{ 186, 454, 33 },
		{ 186, 455, 33 },
		{ 186, 456, 33 },
		{ 186, 457, 33 },
		{ 186, 458, 33 },
		{ 186, 459, 33 },
		{ 186, 460, 33 },
		{ 186, 461, 33 },
		{ 186, 462, 33 },
		{ 186, 463, 33 },
		{ 186, 464, 33 },
		{ 186, 465, 33 },
		{ 186, 466, 33 },
		{ 186, 467, 33 },
		{ 186, 468, 33 },
		{ 186, 469, 33 },
		{ 186, 470, 33 },
		{ 186, 471, 33 },
		{ 186, 472, 33 },
		{ 186, 473, 33 },
		{ 186, 474, 33 },
		{ 186, 475, 33 },
		{ 186, 476, 33 },
		{ 186, 477, 33 },
		{ 186, 478, 33 },
		{ 186, 479, 33 },
		{ 186, 480, 33 },
		{ 186, 481, 33 },
		{ 186, 482, 33 },
		{ 186, 483, 33 },
		{ 186, 484, 33 },
		{ 186, 485, 33 },
		{ 186, 486, 33 },
		{ 186, 487, 33 },
		{ 186, 488, 33 },
		{ 186, 489, 33 },
		{ 186, 490, 33 },
		{ 186, 491, 33 },
		{ 186, 492, 33 },
		{ 186, 493, 33 },
		{ 186, 494, 33 },
		{ 186, 495, 33 },
		{ 186, 496, 33 },
		{ 186, 497, 33 },
		{ 186, 498, 33 },
		{ 186, 499, 33 },
		{ 186, 500, 33 },
		{ 186, 501, 33 },
		{ 186, 502, 33 },
		{ 186, 503, 33 },
		{ 186, 504, 33 },
		{ 186, 505, 33 },
		{ 186, 506, 33 },
		{ 186, 507, 33 },
		{ 186, 508, 33 },
		{ 186, 509, 33 },
		{ 186, 510, 33 },
		{ 186, 511, 33 },
		{ 186, 512, 33 },
		{ 187, 398, 33 },
		{ 187, 399, 33 },
		{ 187, 400, 33 },
		{ 187, 401, 33 },
		{ 187, 402, 33 },
		{ 187, 403, 33 },
		{ 187, 404, 33 },
		{ 187, 405, 33 },
		{ 187, 406, 33 },
		{ 187, 407, 33 },
		{ 187, 408, 33 },
		{ 187, 409, 33 },
		{ 187, 410, 33 },
		{ 187, 411, 33 },
		{ 187, 412, 33 },
		{ 187, 413, 33 },
		{ 187, 414, 33 },
		{ 187, 415, 33 },
		{ 187, 416, 33 },
		{ 187, 417, 33 },
		{ 187, 418, 33 },
		{ 187, 419, 33 },
		{ 187, 420, 33 },
		{ 187, 421, 33 },
		{ 187, 422, 33 },
		{ 187, 423, 33 },
		{ 187, 424, 33 },
		{ 187, 425, 33 },
		{ 187, 426, 33 },
		{ 187, 427, 33 },
		{ 187, 428, 33 },
		{ 187, 429, 33 },
		{ 187, 430, 33 },
		{ 187, 431, 33 },
		{ 187, 432, 33 },
		{ 187, 433, 33 },
		{ 187, 434, 33 },
		{ 187, 435, 33 },
		{ 187, 436, 33 },
		{ 187, 437, 33 },
		{ 187, 438, 33 },
		{ 187, 439, 33 },
		{ 187, 440, 33 },
		{ 187, 441, 33 },
		{ 187, 442, 33 },
		{ 187, 443, 33 },
		{ 187, 444, 33 },
		{ 187, 445, 33 },
		{ 187, 446, 33 },
		{ 187, 447, 33 },
		{ 187, 448, 33 },
		{ 187, 449, 33 },
		{ 187, 450, 33 },
		{ 187, 451, 33 },
		{ 187, 452, 33 },
		{ 187, 453, 33 },
		{ 187, 454, 33 },
		{ 187, 455, 33 },
		{ 187, 456, 33 },
		{ 187, 457, 33 },
		{ 187, 458, 33 },
		{ 187, 459, 33 },
		{ 187, 460, 33 },
		{ 187, 461, 33 },
		{ 187, 462, 33 },
		{ 187, 463, 33 },
		{ 187, 464, 33 },
		{ 187, 465, 33 },
		{ 187, 466, 33 },
		{ 187, 467, 33 },
		{ 187, 468, 33 },
		{ 187, 469, 33 },
		{ 187, 470, 33 },
		{ 187, 471, 33 },
		{ 187, 472, 33 },
		{ 187, 473, 33 },
		{ 187, 474, 33 },
		{ 187, 475, 33 },
		{ 187, 476, 33 },
		{ 187, 477, 33 },
		{ 187, 478, 33 },
		{ 187, 479, 33 },
		{ 187, 480, 33 },
		{ 187, 481, 33 },
		{ 187, 482, 33 },
		{ 187, 483, 33 },
		{ 187, 484, 33 },
		{ 187, 485, 33 },
		{ 187, 486, 33 },
		{ 187, 487, 33 },
		{ 187, 488, 33 },
		{ 187, 489, 33 },
		{ 187, 490, 33 },
		{ 187, 491, 33 },
		{ 187, 492, 33 },
		{ 187, 493, 33 },
		{ 187, 494, 33 },
		{ 187, 495, 33 },
		{ 187, 496, 33 },
		{ 187, 497, 33 },
		{ 187, 498, 33 },
		{ 187, 499, 33 },
		{ 187, 500, 33 },
		{ 187, 501, 33 },
		{ 187, 502, 33 },
		{ 187, 503, 33 },
		{ 187, 504, 33 },
		{ 187, 505, 33 },
		{ 187, 506, 33 },
		{ 187, 507, 33 },
		{ 187, 508, 33 },
		{ 187, 509, 33 },
		{ 187, 510, 33 },
		{ 187, 511, 33 },
		{ 187, 512, 33 },
		{ 188, 396, 33 },
		{ 188, 397, 33 },
		{ 188, 398, 33 },
		{ 188, 399, 33 },
		{ 188, 400, 33 },
		{ 188, 401, 33 },
		{ 188, 402, 33 },
		{ 188, 403, 33 },
		{ 188, 404, 33 },
		{ 188, 405, 33 },
		{ 188, 406, 33 },
		{ 188, 407, 33 },
		{ 188, 408, 33 },
		{ 188, 409, 33 },
		{ 188, 410, 33 },
		{ 188, 411, 33 },
		{ 188, 412, 33 },
		{ 188, 413, 33 },
		{ 188, 414, 33 },
		{ 188, 415, 33 },
		{ 188, 416, 33 },
		{ 188, 417, 33 },
		{ 188, 418, 33 },
		{ 188, 419, 33 },
		{ 188, 420, 33 },
		{ 188, 421, 33 },
		{ 188, 422, 33 },
		{ 188, 423, 33 },
		{ 188, 424, 33 },
		{ 188, 425, 33 },
		{ 188, 426, 33 },
		{ 188, 427, 33 },
		{ 188, 428, 33 },
		{ 188, 429, 33 },
		{ 188, 430, 33 },
		{ 188, 431, 33 },
		{ 188, 432, 33 },
		{ 188, 433, 33 },
		{ 188, 434, 33 },
		{ 188, 435, 33 },
		{ 188, 436, 33 },
		{ 188, 437, 33 },
		{ 188, 438, 33 },
		{ 188, 439, 33 },
		{ 188, 440, 33 },
		{ 188, 441, 33 },
		{ 188, 442, 33 },
		{ 188, 443, 33 },
		{ 188, 444, 33 },
		{ 188, 445, 33 },
		{ 188, 446, 33 },
		{ 188, 447, 33 },
		{ 188, 448, 33 },
		{ 188, 449, 33 },
		{ 188, 450, 33 },
		{ 188, 451, 33 },
		{ 188, 452, 33 },
		{ 188, 453, 33 },
		{ 188, 454, 33 },
		{ 188, 455, 33 },
		{ 188, 456, 33 },
		{ 188, 457, 33 },
		{ 188, 458, 33 },
		{ 188, 459, 33 },
		{ 188, 460, 33 },
		{ 188, 461, 33 },
		{ 188, 462, 33 },
		{ 188, 463, 33 },
		{ 188, 464, 33 },
		{ 188, 465, 33 },
		{ 188, 466, 33 },
		{ 188, 467, 33 },
		{ 188, 468, 33 },
		{ 188, 469, 33 },
		{ 188, 470, 33 },
		{ 188, 471, 33 },
		{ 188, 472, 33 },
		{ 188, 473, 33 },
		{ 188, 474, 33 },
		{ 188, 475, 33 },
		{ 188, 476, 33 },
		{ 188, 477, 33 },
		{ 188, 478, 33 },
		{ 188, 479, 33 },
		{ 188, 480, 33 },
		{ 188, 481, 33 },
		{ 188, 482, 33 },
		{ 188, 483, 33 },
		{ 188, 484, 33 },
		{ 188, 485, 33 },
		{ 188, 486, 33 },
		{ 188, 487, 33 },
		{ 188, 488, 33 },
		{ 188, 489, 33 },
		{ 188, 490, 33 },
		{ 188, 491, 33 },
		{ 188, 492, 33 },
		{ 188, 493, 33 },
		{ 188, 494, 33 },
		{ 188, 495, 33 },
		{ 188, 496, 33 },
		{ 188, 497, 33 },
		{ 188, 498, 33 },
		{ 188, 499, 33 },
		{ 188, 500, 33 },
		{ 188, 501, 33 },
		{ 188, 502, 33 },
		{ 188, 503, 33 },
		{ 188, 504, 33 },
		{ 188, 505, 33 },
		{ 188, 506, 33 },
		{ 188, 507, 33 },
		{ 188, 508, 33 },
		{ 188, 509, 33 },
		{ 188, 510, 33 },
		{ 188, 511, 33 },
		{ 188, 512, 33 },
		{ 189, 394, 33 },
		{ 189, 395, 33 },
		{ 189, 396, 33 },
		{ 189, 397, 33 },
		{ 189, 398, 33 },
		{ 189, 399, 33 },
		{ 189, 400, 33 },
		{ 189, 401, 33 },
		{ 189, 402, 33 },
		{ 189, 403, 33 },
		{ 189, 404, 33 },
		{ 189, 405, 33 },
		{ 189, 406, 33 },
		{ 189, 407, 33 },
		{ 189, 408, 33 },
		{ 189, 409, 33 },
		{ 189, 410, 33 },
		{ 189, 411, 33 },
		{ 189, 412, 33 },
		{ 189, 413, 33 },
		{ 189, 414, 33 },
		{ 189, 415, 33 },
		{ 189, 416, 33 },
		{ 189, 417, 33 },
		{ 189, 418, 33 },
		{ 189, 419, 33 },
		{ 189, 420, 33 },
		{ 189, 421, 33 },
		{ 189, 422, 33 },
		{ 189, 423, 33 },
		{ 189, 424, 33 },
		{ 189, 425, 33 },
		{ 189, 426, 33 },
		{ 189, 427, 33 },
		{ 189, 428, 33 },
		{ 189, 429, 33 },
		{ 189, 430, 33 },
		{ 189, 431, 33 },
		{ 189, 432, 33 },
		{ 189, 433, 33 },
		{ 189, 434, 33 },
		{ 189, 435, 33 },
		{ 189, 436, 33 },
		{ 189, 437, 33 },
		{ 189, 438, 33 },
		{ 189, 439, 33 },
		{ 189, 440, 33 },
		{ 189, 441, 33 },
		{ 189, 442, 33 },
		{ 189, 443, 33 },
		{ 189, 444, 33 },
		{ 189, 445, 33 },
		{ 189, 446, 33 },
		{ 189, 447, 33 },
		{ 189, 448, 33 },
		{ 189, 449, 33 },
		{ 189, 450, 33 },
		{ 189, 451, 33 },
		{ 189, 452, 33 },
		{ 189, 453, 33 },
		{ 189, 454, 33 },
		{ 189, 455, 33 },
		{ 189, 456, 33 },
		{ 189, 457, 33 },
		{ 189, 458, 33 },
		{ 189, 459, 33 },
		{ 189, 460, 33 },
		{ 189, 461, 33 },
		{ 189, 462, 33 },
		{ 189, 463, 33 },
		{ 189, 464, 33 },
		{ 189, 465, 33 },
		{ 189, 466, 33 },
		{ 189, 467, 33 },
		{ 189, 468, 33 },
		{ 189, 469, 33 },
		{ 189, 470, 33 },
		{ 189, 471, 33 },
		{ 189, 472, 33 },
		{ 189, 473, 33 },
		{ 189, 474, 33 },
		{ 189, 475, 33 },
		{ 189, 476, 33 },
		{ 189, 477, 33 },
		{ 189, 478, 33 },
		{ 189, 479, 33 },
		{ 189, 480, 33 },
		{ 189, 481, 33 },
		{ 189, 482, 33 },
		{ 189, 483, 33 },
		{ 189, 484, 33 },
		{ 189, 485, 33 },
		{ 189, 486, 33 },
		{ 189, 487, 33 },
		{ 189, 488, 33 },
		{ 189, 489, 33 },
		{ 189, 490, 33 },
		{ 189, 491, 33 },
		{ 189, 492, 33 },
		{ 189, 493, 33 },
		{ 189, 494, 33 },
		{ 189, 495, 33 },
		{ 189, 496, 33 },
		{ 189, 497, 33 },
		{ 189, 498, 33 },
		{ 189, 499, 33 },
		{ 189, 500, 33 },
		{ 189, 501, 33 },
		{ 189, 502, 33 },
		{ 189, 503, 33 },
		{ 189, 504, 33 },
		{ 189, 505, 33 },
		{ 189, 506, 33 },
		{ 189, 507, 33 },
		{ 189, 508, 33 },
		{ 189, 509, 33 },
		{ 189, 510, 33 },
		{ 189, 511, 33 },
		{ 189, 512, 33 },
		{ 190, 392, 33 },
		{ 190, 393, 33 },
		{ 190, 394, 33 },
		{ 190, 395, 33 },
		{ 190, 396, 33 },
		{ 190, 397, 33 },
		{ 190, 398, 33 },
		{ 190, 399, 33 },
		{ 190, 400, 33 },
		{ 190, 401, 33 },
		{ 190, 402, 33 },
		{ 190, 403, 33 },
		{ 190, 404, 33 },
		{ 190, 405, 33 },
		{ 190, 406, 33 },
		{ 190, 407, 33 },
		{ 190, 408, 33 },
		{ 190, 409, 33 },
		{ 190, 410, 33 },
		{ 190, 411, 33 },
		{ 190, 412, 33 },
		{ 190, 413, 33 },
		{ 190, 414, 33 },
		{ 190, 415, 33 },
		{ 190, 416, 33 },
		{ 190, 417, 33 },
		{ 190, 418, 33 },
		{ 190, 419, 33 },
		{ 190, 420, 33 },
		{ 190, 421, 33 },
		{ 190, 422, 33 },
		{ 190, 423, 33 },
		{ 190, 424, 33 },
		{ 190, 425, 33 },
		{ 190, 426, 33 },
		{ 190, 427, 33 },
		{ 190, 428, 33 },
		{ 190, 429, 33 },
		{ 190, 430, 33 },
		{ 190, 431, 33 },
		{ 190, 432, 33 },
		{ 190, 433, 33 },
		{ 190, 434, 33 },
		{ 190, 435, 33 },
		{ 190, 436, 33 },
		{ 190, 437, 33 },
		{ 190, 438, 33 },
		{ 190, 439, 33 },
		{ 190, 440, 33 },
		{ 190, 441, 33 },
		{ 190, 442, 33 },
		{ 190, 443, 33 },
		{ 190, 444, 33 },
		{ 190, 445, 33 },
		{ 190, 446, 33 },
		{ 190, 447, 33 },
		{ 190, 448, 33 },
		{ 190, 449, 33 },
		{ 190, 450, 33 },
		{ 190, 451, 33 },
		{ 190, 452, 33 },
		{ 190, 453, 33 },
		{ 190, 454, 33 },
		{ 190, 455, 33 },
		{ 190, 456, 33 },
		{ 190, 457, 33 },
		{ 190, 458, 33 },
		{ 190, 459, 33 },
		{ 190, 460, 33 },
		{ 190, 461, 33 },
		{ 190, 462, 33 },
		{ 190, 463, 33 },
		{ 190, 464, 33 },
		{ 190, 465, 33 },
		{ 190, 466, 33 },
		{ 190, 467, 33 },
		{ 190, 468, 33 },
		{ 190, 469, 33 },
		{ 190, 470, 33 },
		{ 190, 471, 33 },
		{ 190, 472, 33 },
		{ 190, 473, 33 },
		{ 190, 474, 33 },
		{ 190, 475, 33 },
		{ 190, 476, 33 },
		{ 190, 477, 33 },
		{ 190, 478, 33 },
		{ 190, 479, 33 },
		{ 190, 480, 33 },
		{ 190, 481, 33 },
		{ 190, 482, 33 },
		{ 190, 483, 33 },
		{ 190, 484, 33 },
		{ 190, 485, 33 },
		{ 190, 486, 33 },
		{ 190, 487, 33 },
		{ 190, 488, 33 },
		{ 190, 489, 33 },
		{ 190, 490, 33 },
		{ 190, 491, 33 },
		{ 190, 492, 33 },
		{ 190, 493, 33 },
		{ 190, 494, 33 },
		{ 190, 495, 33 },
		{ 190, 496, 33 },
		{ 190, 497, 33 },
		{ 190, 498, 33 },
		{ 190, 499, 33 },
		{ 190, 500, 33 },
		{ 190, 501, 33 },
		{ 190, 502, 33 },
		{ 190, 503, 33 },
		{ 190, 504, 33 },
		{ 190, 505, 33 },
		{ 190, 506, 33 },
		{ 190, 507, 33 },
		{ 190, 508, 33 },
		{ 190, 509, 33 },
		{ 190, 510, 33 },
		{ 190, 511, 33 },
		{ 190, 512, 33 },
		{ 191, 390, 33 },
		{ 191, 391, 33 },
		{ 191, 392, 33 },
		{ 191, 393, 33 },
		{ 191, 394, 33 },
		{ 191, 395, 33 },
		{ 191, 396, 33 },
		{ 191, 397, 33 },
		{ 191, 398, 33 },
		{ 191, 399, 33 },
		{ 191, 400, 33 },
		{ 191, 401, 33 },
		{ 191, 402, 33 },
		{ 191, 403, 33 },
		{ 191, 404, 33 },
		{ 191, 405, 33 },
		{ 191, 406, 33 },
		{ 191, 407, 33 },
		{ 191, 408, 33 },
		{ 191, 409, 33 },
		{ 191, 410, 33 },
		{ 191, 411, 33 },
		{ 191, 412, 33 },
		{ 191, 413, 33 },
		{ 191, 414, 33 },
		{ 191, 415, 33 },
		{ 191, 416, 33 },
		{ 191, 417, 33 },
		{ 191, 418, 33 },
		{ 191, 419, 33 },
		{ 191, 420, 33 },
		{ 191, 421, 33 },
		{ 191, 422, 33 },
		{ 191, 423, 33 },
		{ 191, 424, 33 },
		{ 191, 425, 33 },
		{ 191, 426, 33 },
		{ 191, 427, 33 },
		{ 191, 428, 33 },
		{ 191, 429, 33 },
		{ 191, 430, 33 },
		{ 191, 431, 33 },
		{ 191, 432, 33 },
		{ 191, 433, 33 },
		{ 191, 434, 33 },
		{ 191, 435, 33 },
		{ 191, 436, 33 },
		{ 191, 437, 33 },
		{ 191, 438, 33 },
		{ 191, 439, 33 },
		{ 191, 440, 33 },
		{ 191, 441, 33 },
		{ 191, 442, 33 },
		{ 191, 443, 33 },
		{ 191, 444, 33 },
		{ 191, 445, 33 },
		{ 191, 446, 33 },
		{ 191, 447, 33 },
		{ 191, 448, 33 },
		{ 191, 449, 33 },
		{ 191, 450, 33 },
		{ 191, 451, 33 },
		{ 191, 452, 33 },
		{ 191, 453, 33 },
		{ 191, 454, 33 },
		{ 191, 455, 33 },
		{ 191, 456, 33 },
		{ 191, 457, 33 },
		{ 191, 458, 33 },
		{ 191, 459, 33 },
		{ 191, 460, 33 },
		{ 191, 461, 33 },
		{ 191, 462, 33 },
		{ 191, 463, 33 },
		{ 191, 464, 33 },
		{ 191, 465, 33 },
		{ 191, 466, 33 },
		{ 191, 467, 33 },
		{ 191, 468, 33 },
		{ 191, 469, 33 },
		{ 191, 470, 33 },
		{ 191, 471, 33 },
		{ 191, 472, 33 },
		{ 191, 473, 33 },
		{ 191, 474, 33 },
		{ 191, 475, 33 },
		{ 191, 476, 33 },
		{ 191, 477, 33 },
		{ 191, 478, 33 },
		{ 191, 479, 33 },
		{ 191, 480, 33 },
		{ 191, 481, 33 },
		{ 191, 482, 33 },
		{ 191, 483, 33 },
		{ 191, 484, 33 },
		{ 191, 485, 33 },
		{ 191, 486, 33 },
		{ 191, 487, 33 },
		{ 191, 488, 33 },
		{ 191, 489, 33 },
		{ 191, 490, 33 },
		{ 191, 491, 33 },
		{ 191, 492, 33 },
		{ 191, 493, 33 },
		{ 191, 494, 33 },
		{ 191, 495, 33 },
		{ 191, 496, 33 },
		{ 191, 497, 33 },
		{ 191, 498, 33 },
		{ 191, 499, 33 },
		{ 191, 500, 33 },
		{ 191, 501, 33 },
		{ 191, 502, 33 },
		{ 191, 503, 33 },
		{ 191, 504, 33 },
		{ 191, 505, 33 },
		{ 191, 506, 33 },
		{ 191, 507, 33 },
		{ 191, 508, 33 },
		{ 191, 509, 33 },
		{ 191, 510, 33 },
		{ 191, 511, 33 },
		{ 191, 512, 33 },
		{ 192, 388, 33 },
		{ 192, 389, 33 },
		{ 192, 390, 33 },
		{ 192, 391, 33 },
		{ 192, 392, 33 },
		{ 192, 393, 33 },
		{ 192, 394, 33 },
		{ 192, 395, 33 },
		{ 192, 396, 33 },
		{ 192, 397, 33 },
		{ 192, 398, 33 },
		{ 192, 399, 33 },
		{ 192, 400, 33 },
		{ 192, 401, 33 },
		{ 192, 402, 33 },
		{ 192, 403, 33 },
		{ 192, 404, 33 },
		{ 192, 405, 33 },
		{ 192, 406, 33 },
		{ 192, 407, 33 },
		{ 192, 408, 33 },
		{ 192, 409, 33 },
		{ 192, 410, 33 },
		{ 192, 411, 33 },
		{ 192, 412, 33 },
		{ 192, 413, 33 },
		{ 192, 414, 33 },
		{ 192, 415, 33 },
		{ 192, 416, 33 },
		{ 192, 417, 33 },
		{ 192, 418, 33 },
		{ 192, 419, 33 },
		{ 192, 420, 33 },
		{ 192, 421, 33 },
		{ 192, 422, 33 },
		{ 192, 423, 33 },
		{ 192, 424, 33 },
		{ 192, 425, 33 },
		{ 192, 426, 33 },
		{ 192, 427, 33 },
		{ 192, 428, 33 },
		{ 192, 429, 33 },
		{ 192, 430, 33 },
		{ 192, 431, 33 },
		{ 192, 432, 33 },
		{ 192, 433, 33 },
		{ 192, 434, 33 },
		{ 192, 435, 33 },
		{ 192, 436, 33 },
		{ 192, 437, 33 },
		{ 192, 438, 33 },
		{ 192, 439, 33 },
		{ 192, 440, 33 },
		{ 192, 441, 33 },
		{ 192, 442, 33 },
		{ 192, 443, 33 },
		{ 192, 444, 33 },
		{ 192, 445, 33 },
		{ 192, 446, 33 },
		{ 192, 447, 33 },
		{ 192, 448, 33 },
		{ 192, 449, 33 },
		{ 192, 450, 33 },
		{ 192, 451, 33 },
		{ 192, 452, 33 },
		{ 192, 453, 33 },
		{ 192, 454, 33 },
		{ 192, 455, 33 },
		{ 192, 456, 33 },
		{ 192, 457, 33 },
		{ 192, 458, 33 },
		{ 192, 459, 33 },
		{ 192, 460, 33 },
		{ 192, 461, 33 },
		{ 192, 462, 33 },
		{ 192, 463, 33 },
		{ 192, 464, 33 },
		{ 192, 465, 33 },
		{ 192, 466, 33 },
		{ 192, 467, 33 },
		{ 192, 468, 33 },
		{ 192, 469, 33 },
		{ 192, 470, 33 },
		{ 192, 471, 33 },
		{ 192, 472, 33 },
		{ 192, 473, 33 },
		{ 192, 474, 33 },
		{ 192, 475, 33 },
		{ 192, 476, 33 },
		{ 192, 477, 33 },
		{ 192, 478, 33 },
		{ 192, 479, 33 },
		{ 192, 480, 33 },
		{ 192, 481, 33 },
		{ 192, 482, 33 },
		{ 192, 483, 33 },
		{ 192, 484, 33 },
		{ 192, 485, 33 },
		{ 192, 486, 33 },
		{ 192, 487, 33 },
		{ 192, 488, 33 },
		{ 192, 489, 33 },
		{ 192, 490, 33 },
		{ 192, 491, 33 },
		{ 192, 492, 33 },
		{ 192, 493, 33 },
		{ 192, 494, 33 },
		{ 192, 495, 33 },
		{ 192, 496, 33 },
		{ 192, 497, 33 },
		{ 192, 498, 33 },
		{ 192, 499, 33 },
		{ 192, 500, 33 },
		{ 192, 501, 33 },
		{ 192, 502, 33 },
		{ 192, 503, 33 },
		{ 192, 504, 33 },
		{ 192, 505, 33 },
		{ 192, 506, 33 },
		{ 192, 507, 33 },
		{ 192, 508, 33 },
		{ 192, 509, 33 },
		{ 192, 510, 33 },
		{ 192, 511, 33 },
		{ 192, 512, 33 },
		{ 193, 386, 33 },
		{ 193, 387, 33 },
		{ 193, 388, 33 },
		{ 193, 389, 33 },
		{ 193, 390, 33 },
		{ 193, 391, 33 },
		{ 193, 392, 33 },
		{ 193, 393, 33 },
		{ 193, 394, 33 },
		{ 193, 395, 33 },
		{ 193, 396, 33 },
		{ 193, 397, 33 },
		{ 193, 398, 33 },
		{ 193, 399, 33 },
		{ 193, 400, 33 },
		{ 193, 401, 33 },
		{ 193, 402, 33 },
		{ 193, 403, 33 },
		{ 193, 404, 33 },
		{ 193, 405, 33 },
		{ 193, 406, 33 },
		{ 193, 407, 33 },
		{ 193, 408, 33 },
		{ 193, 409, 33 },
		{ 193, 410, 33 },
		{ 193, 411, 33 },
		{ 193, 412, 33 },
		{ 193, 413, 33 },
		{ 193, 414, 33 },
		{ 193, 415, 33 },
		{ 193, 416, 33 },
		{ 193, 417, 33 },
		{ 193, 418, 33 },
		{ 193, 419, 33 },
		{ 193, 420, 33 },
		{ 193, 421, 33 },
		{ 193, 422, 33 },
		{ 193, 423, 33 },
		{ 193, 424, 33 },
		{ 193, 425, 33 },
		{ 193, 426, 33 },
		{ 193, 427, 33 },
		{ 193, 428, 33 },
		{ 193, 429, 33 },
		{ 193, 430, 33 },
		{ 193, 431, 33 },
		{ 193, 432, 33 },
		{ 193, 433, 33 },
		{ 193, 434, 33 },
		{ 193, 435, 33 },
		{ 193, 436, 33 },
		{ 193, 437, 33 },
		{ 193, 438, 33 },
		{ 193, 439, 33 },
		{ 193, 440, 33 },
		{ 193, 441, 33 },
		{ 193, 442, 33 },
		{ 193, 443, 33 },
		{ 193, 444, 33 },
		{ 193, 445, 33 },
		{ 193, 446, 33 },
		{ 193, 447, 33 },
		{ 193, 448, 33 },
		{ 193, 449, 33 },
		{ 193, 450, 33 },
		{ 193, 451, 33 },
		{ 193, 452, 33 },
		{ 193, 453, 33 },
		{ 193, 454, 33 },
		{ 193, 455, 33 },
		{ 193, 456, 33 },
		{ 193, 457, 33 },
		{ 193, 458, 33 },
		{ 193, 459, 33 },
		{ 193, 460, 33 },
		{ 193, 461, 33 },
		{ 193, 462, 33 },
		{ 193, 463, 33 },
		{ 193, 464, 33 },
		{ 193, 465, 33 },
		{ 193, 466, 33 },
		{ 193, 467, 33 },
		{ 193, 468, 33 },
		{ 193, 469, 33 },
		{ 193, 470, 33 },
		{ 193, 471, 33 },
		{ 193, 472, 33 },
		{ 193, 473, 33 },
		{ 193, 474, 33 },
		{ 193, 475, 33 },
		{ 193, 476, 33 },
		{ 193, 477, 33 },
		{ 193, 478, 33 },
		{ 193, 479, 33 },
		{ 193, 480, 33 },
		{ 193, 481, 33 },
		{ 193, 482, 33 },
		{ 193, 483, 33 },
		{ 193, 484, 33 },
		{ 193, 485, 33 },
		{ 193, 486, 33 },
		{ 193, 487, 33 },
		{ 193, 488, 33 },
		{ 193, 489, 33 },
		{ 193, 490, 33 },
		{ 193, 491, 33 },
		{ 193, 492, 33 },
		{ 193, 493, 33 },
		{ 193, 494, 33 },
		{ 193, 495, 33 },
		{ 193, 496, 33 },
		{ 193, 497, 33 },
		{ 193, 498, 33 },
		{ 193, 499, 33 },
		{ 193, 500, 33 },
		{ 193, 501, 33 },
		{ 193, 502, 33 },
		{ 193, 503, 33 },
		{ 193, 504, 33 },
		{ 193, 505, 33 },
		{ 193, 506, 33 },
		{ 193, 507, 33 },
		{ 193, 508, 33 },
		{ 193, 509, 33 },
		{ 193, 510, 33 },
		{ 193, 511, 33 },
		{ 193, 512, 33 },
		{ 194, 384, 33 },
		{ 194, 385, 33 },
		{ 194, 386, 33 },
		{ 194, 387, 33 },
		{ 194, 388, 33 },
		{ 194, 389, 33 },
		{ 194, 390, 33 },
		{ 194, 391, 33 },
		{ 194, 392, 33 },
		{ 194, 393, 33 },
		{ 194, 394, 33 },
		{ 194, 395, 33 },
		{ 194, 396, 33 },
		{ 194, 397, 33 },
		{ 194, 398, 33 },
		{ 194, 399, 33 },
		{ 194, 400, 33 },
		{ 194, 401, 33 },
		{ 194, 402, 33 },
		{ 194, 403, 33 },
		{ 194, 404, 33 },
		{ 194, 405, 33 },
		{ 194, 406, 33 },
		{ 194, 407, 33 },
		{ 194, 408, 33 },
		{ 194, 409, 33 },
		{ 194, 410, 33 },
		{ 194, 411, 33 },
		{ 194, 412, 33 },
		{ 194, 413, 33 },
		{ 194, 414, 33 },
		{ 194, 415, 33 },
		{ 194, 416, 33 },
		{ 194, 417, 33 },
		{ 194, 418, 33 },
		{ 194, 419, 33 },
		{ 194, 420, 33 },
		{ 194, 421, 33 },
		{ 194, 422, 33 },
		{ 194, 423, 33 },
		{ 194, 424, 33 },
		{ 194, 425, 33 },
		{ 194, 426, 33 },
		{ 194, 427, 33 },
		{ 194, 428, 33 },
		{ 194, 429, 33 },
		{ 194, 430, 33 },
		{ 194, 431, 33 },
		{ 194, 432, 33 },
		{ 194, 433, 33 },
		{ 194, 434, 33 },
		{ 194, 435, 33 },
		{ 194, 436, 33 },
		{ 194, 437, 33 },
		{ 194, 438, 33 },
		{ 194, 439, 33 },
		{ 194, 440, 33 },
		{ 194, 441, 33 },
		{ 194, 442, 33 },
		{ 194, 443, 33 },
		{ 194, 444, 33 },
		{ 194, 445, 33 },
		{ 194, 446, 33 },
		{ 194, 447, 33 },
		{ 194, 448, 33 },
		{ 194, 449, 33 },
		{ 194, 450, 33 },
		{ 194, 451, 33 },
		{ 194, 452, 33 },
		{ 194, 453, 33 },
		{ 194, 454, 33 },
		{ 194, 455, 33 },
		{ 194, 456, 33 },
		{ 194, 457, 33 },
		{ 194, 458, 33 },
		{ 194, 459, 33 },
		{ 194, 460, 33 },
		{ 194, 461, 33 },
		{ 194, 462, 33 },
		{ 194, 463, 33 },
		{ 194, 464, 33 },
		{ 194, 465, 33 },
		{ 194, 466, 33 },
		{ 194, 467, 33 },
		{ 194, 468, 33 },
		{ 194, 469, 33 },
		{ 194, 470, 33 },
		{ 194, 471, 33 },
		{ 194, 472, 33 },
		{ 194, 473, 33 },
		{ 194, 474, 33 },
		{ 194, 475, 33 },
		{ 194, 476, 33 },
		{ 194, 477, 33 },
		{ 194, 478, 33 },
		{ 194, 479, 33 },
		{ 194, 480, 33 },
		{ 194, 481, 33 },
		{ 194, 482, 33 },
		{ 194, 483, 33 },
		{ 194, 484, 33 },
		{ 194, 485, 33 },
		{ 194, 486, 33 },
		{ 194, 487, 33 },
		{ 194, 488, 33 },
		{ 194, 489, 33 },
		{ 194, 490, 33 },
		{ 194, 491, 33 },
		{ 194, 492, 33 },
		{ 194, 493, 33 },
		{ 194, 494, 33 },
		{ 194, 495, 33 },
		{ 194, 496, 33 },
		{ 194, 497, 33 },
		{ 194, 498, 33 },
		{ 194, 499, 33 },
		{ 194, 500, 33 },
		{ 194, 501, 33 },
		{ 194, 502, 33 },
		{ 194, 503, 33 },
		{ 194, 504, 33 },
		{ 194, 505, 33 },
		{ 194, 506, 33 },
		{ 194, 507, 33 },
		{ 194, 508, 33 },
		{ 195, 382, 33 },
		{ 195, 383, 33 },
		{ 195, 384, 33 },
		{ 195, 385, 33 },
		{ 195, 386, 33 },
		{ 195, 387, 33 },
		{ 195, 388, 33 },
		{ 195, 389, 33 },
		{ 195, 390, 33 },
		{ 195, 391, 33 },
		{ 195, 392, 33 },
		{ 195, 393, 33 },
		{ 195, 394, 33 },
		{ 195, 395, 33 },
		{ 195, 396, 33 },
		{ 195, 397, 33 },
		{ 195, 398, 33 },
		{ 195, 399, 33 },
		{ 195, 400, 33 },
		{ 195, 401, 33 },
		{ 195, 402, 33 },
		{ 195, 403, 33 },
		{ 195, 404, 33 },
		{ 195, 405, 33 },
		{ 195, 406, 33 },
		{ 195, 407, 33 },
		{ 195, 408, 33 },
		{ 195, 409, 33 },
		{ 195, 410, 33 },
		{ 195, 411, 33 },
		{ 195, 412, 33 },
		{ 195, 413, 33 },
		{ 195, 414, 33 },
		{ 195, 415, 33 },
		{ 195, 416, 33 },
		{ 195, 417, 33 },
		{ 195, 418, 33 },
		{ 195, 419, 33 },
		{ 195, 420, 33 },
		{ 195, 421, 33 },
		{ 195, 422, 33 },
		{ 195, 423, 33 },
		{ 195, 424, 33 },
		{ 195, 425, 33 },
		{ 195, 426, 33 },
		{ 195, 427, 33 },
		{ 195, 428, 33 },
		{ 195, 429, 33 },
		{ 195, 430, 33 },
		{ 195, 431, 33 },
		{ 195, 432, 33 },
		{ 195, 433, 33 },
		{ 195, 434, 33 },
		{ 195, 435, 33 },
		{ 195, 436, 33 },
		{ 195, 437, 33 },
		{ 195, 438, 33 },
		{ 195, 439, 33 },
		{ 195, 440, 33 },
		{ 195, 441, 33 },
		{ 195, 442, 33 },
		{ 195, 443, 33 },
		{ 195, 444, 33 },
		{ 195, 445, 33 },
		{ 195, 446, 33 },
		{ 195, 447, 33 },
		{ 195, 448, 33 },
		{ 195, 449, 33 },
		{ 195, 450, 33 },
		{ 195, 451, 33 },
		{ 195, 452, 33 },
		{ 195, 453, 33 },
		{ 195, 454, 33 },
		{ 195, 455, 33 },
		{ 195, 456, 33 },
		{ 195, 457, 33 },
		{ 195, 458, 33 },
		{ 195, 459, 33 },
		{ 195, 460, 33 },
		{ 195, 461, 33 },
		{ 195, 462, 33 },
		{ 195, 463, 33 },
		{ 195, 464, 33 },
		{ 195, 465, 33 },
		{ 195, 466, 33 },
		{ 195, 467, 33 },
		{ 195, 468, 33 },
		{ 195, 469, 33 },
		{ 195, 470, 33 },
		{ 195, 471, 33 },
		{ 195, 472, 33 },
		{ 195, 473, 33 },
		{ 195, 474, 33 },
		{ 195, 475, 33 },
		{ 195, 476, 33 },
		{ 195, 477, 33 },
		{ 195, 478, 33 },
		{ 195, 479, 33 },
		{ 195, 480, 33 },
		{ 195, 481, 33 },
		{ 195, 482, 33 },
		{ 195, 483, 33 },
		{ 195, 484, 33 },
		{ 195, 485, 33 },
		{ 195, 486, 33 },
		{ 195, 487, 33 },
		{ 195, 488, 33 },
		{ 195, 489, 33 },
		{ 195, 490, 33 },
		{ 195, 491, 33 },
		{ 195, 492, 33 },
		{ 195, 493, 33 },
		{ 195, 494, 33 },
		{ 195, 495, 33 },
		{ 195, 496, 33 },
		{ 195, 497, 33 },
		{ 195, 498, 33 },
		{ 195, 499, 33 },
		{ 195, 500, 33 },
		{ 195, 501, 33 },
		{ 195, 502, 33 },
		{ 195, 503, 33 },
		{ 195, 504, 33 },
		{ 196, 380, 33 },
		{ 196, 381, 33 },
		{ 196, 382, 33 },
		{ 196, 383, 33 },
		{ 196, 384, 33 },
		{ 196, 385, 33 },
		{ 196, 386, 33 },
		{ 196, 387, 33 },
		{ 196, 388, 33 },
		{ 196, 389, 33 },
		{ 196, 390, 33 },
		{ 196, 391, 33 },
		{ 196, 392, 33 },
		{ 196, 393, 33 },
		{ 196, 394, 33 },
		{ 196, 395, 33 },
		{ 196, 396, 33 },
		{ 196, 397, 33 },
		{ 196, 398, 33 },
		{ 196, 399, 33 },
		{ 196, 400, 33 },
		{ 196, 401, 33 },
		{ 196, 402, 33 },
		{ 196, 403, 33 },
		{ 196, 404, 33 },
		{ 196, 405, 33 },
		{ 196, 406, 33 },
		{ 196, 407, 33 },
		{ 196, 408, 33 },
		{ 196, 409, 33 },
		{ 196, 410, 33 },
		{ 196, 411, 33 },
		{ 196, 412, 33 },
		{ 196, 413, 33 },
		{ 196, 414, 33 },
		{ 196, 415, 33 },
		{ 196, 416, 33 },
		{ 196, 417, 33 },
		{ 196, 418, 33 },
		{ 196, 419, 33 },
		{ 196, 420, 33 },
		{ 196, 421, 33 },
		{ 196, 422, 33 },
		{ 196, 423, 33 },
		{ 196, 424, 33 },
		{ 196, 425, 33 },
		{ 196, 426, 33 },
		{ 196, 427, 33 },
		{ 196, 428, 33 },
		{ 196, 429, 33 },
		{ 196, 430, 33 },
		{ 196, 431, 33 },
		{ 196, 432, 33 },
		{ 196, 433, 33 },
		{ 196, 434, 33 },
		{ 196, 435, 33 },
		{ 196, 436, 33 },
		{ 196, 437, 33 },
		{ 196, 438, 33 },
		{ 196, 439, 33 },
		{ 196, 440, 33 },
		{ 196, 441, 33 },
		{ 196, 442, 33 },
		{ 196, 443, 33 },
		{ 196, 444, 33 },
		{ 196, 445, 33 },
		{ 196, 446, 33 },
		{ 196, 447, 33 },
		{ 196, 448, 33 },
		{ 196, 449, 33 },
		{ 196, 450, 33 },
		{ 196, 451, 33 },
		{ 196, 452, 33 },
		{ 196, 453, 33 },
		{ 196, 454, 33 },
		{ 196, 455, 33 },
		{ 196, 456, 33 },
		{ 196, 457, 33 },
		{ 196, 458, 33 },
		{ 196, 459, 33 },
		{ 196, 460, 33 },
		{ 196, 461, 33 },
		{ 196, 462, 33 },
		{ 196, 463, 33 },
		{ 196, 464, 33 },
		{ 196, 465, 33 },
		{ 196, 466, 33 },
		{ 196, 467, 33 },
		{ 196, 468, 33 },
		{ 196, 469, 33 },
		{ 196, 470, 33 },
		{ 196, 471, 33 },
		{ 196, 472, 33 },
		{ 196, 473, 33 },
		{ 196, 474, 33 },
		{ 196, 475, 33 },
		{ 196, 476, 33 },
		{ 196, 477, 33 },
		{ 196, 478, 33 },
		{ 196, 479, 33 },
		{ 196, 480, 33 },
		{ 196, 481, 33 },
		{ 196, 482, 33 },
		{ 196, 483, 33 },
		{ 196, 484, 33 },
		{ 196, 485, 33 },
		{ 196, 486, 33 },
		{ 196, 487, 33 },
		{ 196, 488, 33 },
		{ 196, 489, 33 },
		{ 196, 490, 33 },
		{ 196, 491, 33 },
		{ 196, 492, 33 },
		{ 196, 493, 33 },
		{ 196, 494, 33 },
		{ 196, 495, 33 },
		{ 196, 496, 33 },
		{ 196, 497, 33 },
		{ 196, 498, 33 },
		{ 196, 499, 33 },
		{ 196, 500, 33 },
		{ 197, 378, 33 },
		{ 197, 379, 33 },
		{ 197, 380, 33 },
		{ 197, 381, 33 },
		{ 197, 382, 33 },
		{ 197, 383, 33 },
		{ 197, 384, 33 },
		{ 197, 385, 33 },
		{ 197, 386, 33 },
		{ 197, 387, 33 },
		{ 197, 388, 33 },
		{ 197, 389, 33 },
		{ 197, 390, 33 },
		{ 197, 391, 33 },
		{ 197, 392, 33 },
		{ 197, 393, 33 },
		{ 197, 394, 33 },
		{ 197, 395, 33 },
		{ 197, 396, 33 },
		{ 197, 397, 33 },
		{ 197, 398, 33 },
		{ 197, 399, 33 },
		{ 197, 400, 33 },
		{ 197, 401, 33 },
		{ 197, 402, 33 },
		{ 197, 403, 33 },
		{ 197, 404, 33 },
		{ 197, 405, 33 },
		{ 197, 406, 33 },
		{ 197, 407, 33 },
		{ 197, 408, 33 },
		{ 197, 409, 33 },
		{ 197, 410, 33 },
		{ 197, 411, 33 },
		{ 197, 412, 33 },
		{ 197, 413, 33 },
		{ 197, 414, 33 },
		{ 197, 415, 33 },
		{ 197, 416, 33 },
		{ 197, 417, 33 },
		{ 197, 418, 33 },
		{ 197, 419, 33 },
		{ 197, 420, 33 },
		{ 197, 421, 33 },
		{ 197, 422, 33 },
		{ 197, 423, 33 },
		{ 197, 424, 33 },
		{ 197, 425, 33 },
		{ 197, 426, 33 },
		{ 197, 427, 33 },
		{ 197, 428, 33 },
		{ 197, 429, 33 },
		{ 197, 430, 33 },
		{ 197, 431, 33 },
		{ 197, 432, 33 },
		{ 197, 433, 33 },
		{ 197, 434, 33 },
		{ 197, 435, 33 },
		{ 197, 436, 33 },
		{ 197, 437, 33 },
		{ 197, 438, 33 },
		{ 197, 439, 33 },
		{ 197, 440, 33 },
		{ 197, 441, 33 },
		{ 197, 442, 33 },
		{ 197, 443, 33 },
		{ 197, 444, 33 },
		{ 197, 445, 33 },
		{ 197, 446, 33 },
		{ 197, 447, 33 },
		{ 197, 448, 33 },
		{ 197, 449, 33 },
		{ 197, 450, 33 },
		{ 197, 451, 33 },
		{ 197, 452, 33 },
		{ 197, 453, 33 },
		{ 197, 454, 33 },
		{ 197, 455, 33 },
		{ 197, 456, 33 },
		{ 197, 457, 33 },
		{ 197, 458, 33 },
		{ 197, 459, 33 },
		{ 197, 460, 33 },
		{ 197, 461, 33 },
		{ 197, 462, 33 },
		{ 197, 463, 33 },
		{ 197, 464, 33 },
		{ 197, 465, 33 },
		{ 197, 466, 33 },
		{ 197, 467, 33 },
		{ 197, 468, 33 },
		{ 197, 469, 33 },
		{ 197, 470, 33 },
		{ 197, 471, 33 },
		{ 197, 472, 33 },
		{ 197, 473, 33 },
		{ 197, 474, 33 },
		{ 197, 475, 33 },
		{ 197, 476, 33 },
		{ 197, 477, 33 },
		{ 197, 478, 33 },
		{ 197, 479, 33 },
		{ 197, 480, 33 },
		{ 197, 481, 33 },
		{ 197, 482, 33 },
		{ 197, 483, 33 },
		{ 197, 484, 33 },
		{ 197, 485, 33 },
		{ 197, 486, 33 },
		{ 197, 487, 33 },
		{ 197, 488, 33 },
		{ 197, 489, 33 },
		{ 197, 490, 33 },
		{ 197, 491, 33 },
		{ 197, 492, 33 },
		{ 197, 493, 33 },
		{ 197, 494, 33 },
		{ 197, 495, 33 },
		{ 197, 496, 33 },
		{ 198, 376, 33 },
		{ 198, 377, 33 },
		{ 198, 378, 33 },
		{ 198, 379, 33 },
		{ 198, 380, 33 },
		{ 198, 381, 33 },
		{ 198, 382, 33 },
		{ 198, 383, 33 },
		{ 198, 384, 33 },
		{ 198, 385, 33 },
		{ 198, 386, 33 },
		{ 198, 387, 33 },
		{ 198, 388, 33 },
		{ 198, 389, 33 },
		{ 198, 390, 33 },
		{ 198, 391, 33 },
		{ 198, 392, 33 },
		{ 198, 393, 33 },
		{ 198, 394, 33 },
		{ 198, 395, 33 },
		{ 198, 396, 33 },
		{ 198, 397, 33 },
		{ 198, 398, 33 },
		{ 198, 399, 33 },
		{ 198, 400, 33 },
		{ 198, 401, 33 },
		{ 198, 402, 33 },
		{ 198, 403, 33 },
		{ 198, 404, 33 },
		{ 198, 405, 33 },
		{ 198, 406, 33 },
		{ 198, 407, 33 },
		{ 198, 408, 33 },
		{ 198, 409, 33 },
		{ 198, 410, 33 },
		{ 198, 411, 33 },
		{ 198, 412, 33 },
		{ 198, 413, 33 },
		{ 198, 414, 33 },
		{ 198, 415, 33 },
		{ 198, 416, 33 },
		{ 198, 417, 33 },
		{ 198, 418, 33 },
		{ 198, 419, 33 },
		{ 198, 420, 33 },
		{ 198, 421, 33 },
		{ 198, 422, 33 },
		{ 198, 423, 33 },
		{ 198, 424, 33 },
		{ 198, 425, 33 },
		{ 198, 426, 33 },
		{ 198, 427, 33 },
		{ 198, 428, 33 },
		{ 198, 429, 33 },
		{ 198, 430, 33 },
		{ 198, 431, 33 },
		{ 198, 432, 33 },
		{ 198, 433, 33 },
		{ 198, 434, 33 },
		{ 198, 435, 33 },
		{ 198, 436, 33 },
		{ 198, 437, 33 },
		{ 198, 438, 33 },
		{ 198, 439, 33 },
		{ 198, 440, 33 },
		{ 198, 441, 33 },
		{ 198, 442, 33 },
		{ 198, 443, 33 },
		{ 198, 444, 33 },
		{ 198, 445, 33 },
		{ 198, 446, 33 },
		{ 198, 447, 33 },
		{ 198, 448, 33 },
		{ 198, 449, 33 },
		{ 198, 450, 33 },
		{ 198, 451, 33 },
		{ 198, 452, 33 },
		{ 198, 453, 33 },
		{ 198, 454, 33 },
		{ 198, 455, 33 },
		{ 198, 456, 33 },
		{ 198, 457, 33 },
		{ 198, 458, 33 },
		{ 198, 459, 33 },
		{ 198, 460, 33 },
		{ 198, 461, 33 },
		{ 198, 462, 33 },
		{ 198, 463, 33 },
		{ 198, 464, 33 },
		{ 198, 465, 33 },
		{ 198, 466, 33 },
		{ 198, 467, 33 },
		{ 198, 468, 33 },
		{ 198, 469, 33 },
		{ 198, 470, 33 },
		{ 198, 471, 33 },
		{ 198, 472, 33 },
		{ 198, 473, 33 },
		{ 198, 474, 33 },
		{ 198, 475, 33 },
		{ 198, 476, 33 },
		{ 198, 477, 33 },
		{ 198, 478, 33 },
		{ 198, 479, 33 },
		{ 198, 480, 33 },
		{ 198, 481, 33 },
		{ 198, 482, 33 },
		{ 198, 483, 33 },
		{ 198, 484, 33 },
		{ 198, 485, 33 },
		{ 198, 486, 33 },
		{ 198, 487, 33 },
		{ 198, 488, 33 },
		{ 198, 489, 33 },
		{ 198, 490, 33 },
		{ 198, 491, 33 },
		{ 198, 492, 33 },
		{ 199, 374, 33 },
		{ 199, 375, 33 },
		{ 199, 376, 33 },
		{ 199, 377, 33 },
		{ 199, 378, 33 },
		{ 199, 379, 33 },
		{ 199, 380, 33 },
		{ 199, 381, 33 },
		{ 199, 382, 33 },
		{ 199, 383, 33 },
		{ 199, 384, 33 },
		{ 199, 385, 33 },
		{ 199, 386, 33 },
		{ 199, 387, 33 },
		{ 199, 388, 33 },
		{ 199, 389, 33 },
		{ 199, 390, 33 },
		{ 199, 391, 33 },
		{ 199, 392, 33 },
		{ 199, 393, 33 },
		{ 199, 394, 33 },
		{ 199, 395, 33 },
		{ 199, 396, 33 },
		{ 199, 397, 33 },
		{ 199, 398, 33 },
		{ 199, 399, 33 },
		{ 199, 400, 33 },
		{ 199, 401, 33 },
		{ 199, 402, 33 },
		{ 199, 403, 33 },
		{ 199, 404, 33 },
		{ 199, 405, 33 },
		{ 199, 406, 33 },
		{ 199, 407, 33 },
		{ 199, 408, 33 },
		{ 199, 409, 33 },
		{ 199, 410, 33 },
		{ 199, 411, 33 },
		{ 199, 412, 33 },
		{ 199, 413, 33 },
		{ 199, 414, 33 },
		{ 199, 415, 33 },
		{ 199, 416, 33 },
		{ 199, 417, 33 },
		{ 199, 418, 33 },
		{ 199, 419, 33 },
		{ 199, 420, 33 },
		{ 199, 421, 33 },
		{ 199, 422, 33 },
		{ 199, 423, 33 },
		{ 199, 424, 33 },
		{ 199, 425, 33 },
		{ 199, 426, 33 },
		{ 199, 427, 33 },
		{ 199, 428, 33 },
		{ 199, 429, 33 },
		{ 199, 430, 33 },
		{ 199, 431, 33 },
		{ 199, 432, 33 },
		{ 199, 433, 33 },
		{ 199, 434, 33 },
		{ 199, 435, 33 },
		{ 199, 436, 33 },
		{ 199, 437, 33 },
		{ 199, 438, 33 },
		{ 199, 439, 33 },
		{ 199, 440, 33 },
		{ 199, 441, 33 },
		{ 199, 442, 33 },
		{ 199, 443, 33 },
		{ 199, 444, 33 },
		{ 199, 445, 33 },
		{ 199, 446, 33 },
		{ 199, 447, 33 },
		{ 199, 448, 33 },
		{ 199, 449, 33 },
		{ 199, 450, 33 },
		{ 199, 451, 33 },
		{ 199, 452, 33 },
		{ 199, 453, 33 },
		{ 199, 454, 33 },
		{ 199, 455, 33 },
		{ 199, 456, 33 },
		{ 199, 457, 33 },
		{ 199, 458, 33 },
		{ 199, 459, 33 },
		{ 199, 460, 33 },
		{ 199, 461, 33 },
		{ 199, 462, 33 },
		{ 199, 463, 33 },
		{ 199, 464, 33 },
		{ 199, 465, 33 },
		{ 199, 466, 33 },
		{ 199, 467, 33 },
		{ 199, 468, 33 },
		{ 199, 469, 33 },
		{ 199, 470, 33 },
		{ 199, 471, 33 },
		{ 199, 472, 33 },
		{ 199, 473, 33 },
		{ 199, 474, 33 },
		{ 199, 475, 33 },
		{ 199, 476, 33 },
		{ 199, 477, 33 },
		{ 199, 478, 33 },
		{ 199, 479, 33 },
		{ 199, 480, 33 },
		{ 199, 481, 33 },
		{ 199, 482, 33 },
		{ 199, 483, 33 },
		{ 199, 484, 33 },
		{ 199, 485, 33 },
		{ 199, 486, 33 },
		{ 199, 487, 33 },
		{ 199, 488, 33 },
		{ 200, 372, 33 },
		{ 200, 373, 33 },
		{ 200, 374, 33 },
		{ 200, 375, 33 },
		{ 200, 376, 33 },
		{ 200, 377, 33 },
		{ 200, 378, 33 },
		{ 200, 379, 33 },
		{ 200, 380, 33 },
		{ 200, 381, 33 },
		{ 200, 382, 33 },
		{ 200, 383, 33 },
		{ 200, 384, 33 },
		{ 200, 385, 33 },
		{ 200, 386, 33 },
		{ 200, 387, 33 },
		{ 200, 388, 33 },
		{ 200, 389, 33 },
		{ 200, 390, 33 },
		{ 200, 391, 33 },
		{ 200, 392, 33 },
		{ 200, 393, 33 },
		{ 200, 394, 33 },
		{ 200, 395, 33 },
		{ 200, 396, 33 },
		{ 200, 397, 33 },
		{ 200, 398, 33 },
		{ 200, 399, 33 },
		{ 200, 400, 33 },
		{ 200, 401, 33 },
		{ 200, 402, 33 },
		{ 200, 403, 33 },
		{ 200, 404, 33 },
		{ 200, 405, 33 },
		{ 200, 406, 33 },
		{ 200, 407, 33 },
		{ 200, 408, 33 },
		{ 200, 409, 33 },
		{ 200, 410, 33 },
		{ 200, 411, 33 },
		{ 200, 412, 33 },
		{ 200, 413, 33 },
		{ 200, 414, 33 },
		{ 200, 415, 33 },
		{ 200, 416, 33 },
		{ 200, 417, 33 },
		{ 200, 418, 33 },
		{ 200, 419, 33 },
		{ 200, 420, 33 },
		{ 200, 421, 33 },
		{ 200, 422, 33 },
		{ 200, 423, 33 },
		{ 200, 424, 33 },
		{ 200, 425, 33 },
		{ 200, 426, 33 },
		{ 200, 427, 33 },
		{ 200, 428, 33 },
		{ 200, 429, 33 },
		{ 200, 430, 33 },
		{ 200, 431, 33 },
		{ 200, 432, 33 },
		{ 200, 433, 33 },
		{ 200, 434, 33 },
		{ 200, 435, 33 },
		{ 200, 436, 33 },
		{ 200, 437, 33 },
		{ 200, 438, 33 },
		{ 200, 439, 33 },
		{ 200, 440, 33 },
		{ 200, 441, 33 },
		{ 200, 442, 33 },
		{ 200, 443, 33 },
		{ 200, 444, 33 },
		{ 200, 445, 33 },
		{ 200, 446, 33 },
		{ 200, 447, 33 },
		{ 200, 448, 33 },
		{ 200, 449, 33 },
		{ 200, 450, 33 },
		{ 200, 451, 33 },
		{ 200, 452, 33 },
		{ 200, 453, 33 },
		{ 200, 454, 33 },
		{ 200, 455, 33 },
		{ 200, 456, 33 },
		{ 200, 457, 33 },
		{ 200, 458, 33 },
		{ 200, 459, 33 },
		{ 200, 460, 33 },
		{ 200, 461, 33 },
		{ 200, 462, 33 },
		{ 200, 463, 33 },
		{ 200, 464, 33 },
		{ 200, 465, 33 },
		{ 200, 466, 33 },
		{ 200, 467, 33 },
		{ 200, 468, 33 },
		{ 200, 469, 33 },
		{ 200, 470, 33 },
		{ 200, 471, 33 },
		{ 200, 472, 33 },
		{ 200, 473, 33 },
		{ 200, 474, 33 },
		{ 200, 475, 33 },
		{ 200, 476, 33 },
		{ 200, 477, 33 },
		{ 200, 478, 33 },
		{ 200, 479, 33 },
		{ 200, 480, 33 },
		{ 200, 481, 33 },
		{ 200, 482, 33 },
		{ 200, 483, 33 },
		{ 200, 484, 33 },
		{ 201, 370, 33 },
		{ 201, 371, 33 },
		{ 201, 372, 33 },
		{ 201, 373, 33 },
		{ 201, 374, 33 },
		{ 201, 375, 33 },
		{ 201, 376, 33 },
		{ 201, 377, 33 },
		{ 201, 378, 33 },
		{ 201, 379, 33 },
		{ 201, 380, 33 },
		{ 201, 381, 33 },
		{ 201, 382, 33 },
		{ 201, 383, 33 },
		{ 201, 384, 33 },
		{ 201, 385, 33 },
		{ 201, 386, 33 },
		{ 201, 387, 33 },
		{ 201, 388, 33 },
		{ 201, 389, 33 },
		{ 201, 390, 33 },
		{ 201, 391, 33 },
		{ 201, 392, 33 },
		{ 201, 393, 33 },
		{ 201, 394, 33 },
		{ 201, 395, 33 },
		{ 201, 396, 33 },
		{ 201, 397, 33 },
		{ 201, 398, 33 },
		{ 201, 399, 33 },
		{ 201, 400, 33 },
		{ 201, 401, 33 },
		{ 201, 402, 33 },
		{ 201, 403, 33 },
		{ 201, 404, 33 },
		{ 201, 405, 33 },
		{ 201, 406, 33 },
		{ 201, 407, 33 },
		{ 201, 408, 33 },
		{ 201, 409, 33 },
		{ 201, 410, 33 },
		{ 201, 411, 33 },
		{ 201, 412, 33 },
		{ 201, 413, 33 },
		{ 201, 414, 33 },
		{ 201, 415, 33 },
		{ 201, 416, 33 },
		{ 201, 417, 33 },
		{ 201, 418, 33 },
		{ 201, 419, 33 },
		{ 201, 420, 33 },
		{ 201, 421, 33 },
		{ 201, 422, 33 },
		{ 201, 423, 33 },
		{ 201, 424, 33 },
		{ 201, 425, 33 },
		{ 201, 426, 33 },
		{ 201, 427, 33 },
		{ 201, 428, 33 },
		{ 201, 429, 33 },
		{ 201, 430, 33 },
		{ 201, 431, 33 },
		{ 201, 432, 33 },
		{ 201, 433, 33 },
		{ 201, 434, 33 },
		{ 201, 435, 33 },
		{ 201, 436, 33 },
		{ 201, 437, 33 },
		{ 201, 438, 33 },
		{ 201, 439, 33 },
		{ 201, 440, 33 },
		{ 201, 441, 33 },
		{ 201, 442, 33 },
		{ 201, 443, 33 },
		{ 201, 444, 33 },
		{ 201, 445, 33 },
		{ 201, 446, 33 },
		{ 201, 447, 33 },
		{ 201, 448, 33 },
		{ 201, 449, 33 },
		{ 201, 450, 33 },
		{ 201, 451, 33 },
		{ 201, 452, 33 },
		{ 201, 453, 33 },
		{ 201, 454, 33 },
		{ 201, 455, 33 },
		{ 201, 456, 33 },
		{ 201, 457, 33 },
		{ 201, 458, 33 },
		{ 201, 459, 33 },
		{ 201, 460, 33 },
		{ 201, 461, 33 },
		{ 201, 462, 33 },
		{ 201, 463, 33 },
		{ 201, 464, 33 },
		{ 201, 465, 33 },
		{ 201, 466, 33 },
		{ 201, 467, 33 },
		{ 201, 468, 33 },
		{ 201, 469, 33 },
		{ 201, 470, 33 },
		{ 201, 471, 33 },
		{ 201, 472, 33 },
		{ 201, 473, 33 },
		{ 201, 474, 33 },
		{ 201, 475, 33 },
		{ 201, 476, 33 },
		{ 201, 477, 33 },
		{ 201, 478, 33 },
		{ 201, 479, 33 },
		{ 201, 480, 33 },
		{ 202, 368, 33 },
		{ 202, 369, 33 },
		{ 202, 370, 33 },
		{ 202, 371, 33 },
		{ 202, 372, 33 },
		{ 202, 373, 33 },
		{ 202, 374, 33 },
		{ 202, 375, 33 },
		{ 202, 376, 33 },
		{ 202, 377, 33 },
		{ 202, 378, 33 },
		{ 202, 379, 33 },
		{ 202, 380, 33 },
		{ 202, 381, 33 },
		{ 202, 382, 33 },
		{ 202, 383, 33 },
		{ 202, 384, 33 },
		{ 202, 385, 33 },
		{ 202, 386, 33 },
		{ 202, 387, 33 },
		{ 202, 388, 33 },
		{ 202, 389, 33 },
		{ 202, 390, 33 },
		{ 202, 391, 33 },
		{ 202, 392, 33 },
		{ 202, 393, 33 },
		{ 202, 394, 33 },
		{ 202, 395, 33 },
		{ 202, 396, 33 },
		{ 202, 397, 33 },
		{ 202, 398, 33 },
		{ 202, 399, 33 },
		{ 202, 400, 33 },
		{ 202, 401, 33 },
		{ 202, 402, 33 },
		{ 202, 403, 33 },
		{ 202, 404, 33 },
		{ 202, 405, 33 },
		{ 202, 406, 33 },
		{ 202, 407, 33 },
		{ 202, 408, 33 },
		{ 202, 409, 33 },
		{ 202, 410, 33 },
		{ 202, 411, 33 },
		{ 202, 412, 33 },
		{ 202, 413, 33 },
		{ 202, 414, 33 },
		{ 202, 415, 33 },
		{ 202, 416, 33 },
		{ 202, 417, 33 },
		{ 202, 418, 33 },
		{ 202, 419, 33 },
		{ 202, 420, 33 },
		{ 202, 421, 33 },
		{ 202, 422, 33 },
		{ 202, 423, 33 },
		{ 202, 424, 33 },
		{ 202, 425, 33 },
		{ 202, 426, 33 },
		{ 202, 427, 33 },
		{ 202, 428, 33 },
		{ 202, 429, 33 },
		{ 202, 430, 33 },
		{ 202, 431, 33 },
		{ 202, 432, 33 },
		{ 202, 433, 33 },
		{ 202, 434, 33 },
		{ 202, 435, 33 },
		{ 202, 436, 33 },
		{ 202, 437, 33 },
		{ 202, 438, 33 },
		{ 202, 439, 33 },
		{ 202, 440, 33 },
		{ 202, 441, 33 },
		{ 202, 442, 33 },
		{ 202, 443, 33 },
		{ 202, 444, 33 },
		{ 202, 445, 33 },
		{ 202, 446, 33 },
		{ 202, 447, 33 },
		{ 202, 448, 33 },
		{ 202, 449, 33 },
		{ 202, 450, 33 },
		{ 202, 451, 33 },
		{ 202, 452, 33 },
		{ 202, 453, 33 },
		{ 202, 454, 33 },
		{ 202, 455, 33 },
		{ 202, 456, 33 },
		{ 202, 457, 33 },
		{ 202, 458, 33 },
		{ 202, 459, 33 },
		{ 202, 460, 33 },
		{ 202, 461, 33 },
		{ 202, 462, 33 },
		{ 202, 463, 33 },
		{ 202, 464, 33 },
		{ 202, 465, 33 },
		{ 202, 466, 33 },
		{ 202, 467, 33 },
		{ 202, 468, 33 },
		{ 202, 469, 33 },
		{ 202, 470, 33 },
		{ 202, 471, 33 },
		{ 202, 472, 33 },
		{ 202, 473, 33 },
		{ 202, 474, 33 },
		{ 202, 475, 33 },
		{ 202, 476, 33 },
		{ 203, 366, 33 },
		{ 203, 367, 33 },
		{ 203, 368, 33 },
		{ 203, 369, 33 },
		{ 203, 370, 33 },
		{ 203, 371, 33 },
		{ 203, 372, 33 },
		{ 203, 373, 33 },
		{ 203, 374, 33 },
		{ 203, 375, 33 },
		{ 203, 376, 33 },
		{ 203, 377, 33 },
		{ 203, 378, 33 },
		{ 203, 379, 33 },
		{ 203, 380, 33 },
		{ 203, 381, 33 },
		{ 203, 382, 33 },
		{ 203, 383, 33 },
		{ 203, 384, 33 },
		{ 203, 385, 33 },
		{ 203, 386, 33 },
		{ 203, 387, 33 },
		{ 203, 388, 33 },
		{ 203, 389, 33 },
		{ 203, 390, 33 },
		{ 203, 391, 33 },
		{ 203, 392, 33 },
		{ 203, 393, 33 },
		{ 203, 394, 33 },
		{ 203, 395, 33 },
		{ 203, 396, 33 },
		{ 203, 397, 33 },
		{ 203, 398, 33 },
		{ 203, 399, 33 },
		{ 203, 400, 33 },
		{ 203, 401, 33 },
		{ 203, 402, 33 },
		{ 203, 403, 33 },
		{ 203, 404, 33 },
		{ 203, 405, 33 },
		{ 203, 406, 33 },
		{ 203, 407, 33 },
		{ 203, 408, 33 },
		{ 203, 409, 33 },
		{ 203, 410, 33 },
		{ 203, 411, 33 },
		{ 203, 412, 33 },
		{ 203, 413, 33 },
		{ 203, 414, 33 },
		{ 203, 415, 33 },
		{ 203, 416, 33 },
		{ 203, 417, 33 },
		{ 203, 418, 33 },
		{ 203, 419, 33 },
		{ 203, 420, 33 },
		{ 203, 421, 33 },
		{ 203, 422, 33 },
		{ 203, 423, 33 },
		{ 203, 424, 33 },
		{ 203, 425, 33 },
		{ 203, 426, 33 },
		{ 203, 427, 33 },
		{ 203, 428, 33 },
		{ 203, 429, 33 },
		{ 203, 430, 33 },
		{ 203, 431, 33 },
		{ 203, 432, 33 },
		{ 203, 433, 33 },
		{ 203, 434, 33 },
		{ 203, 435, 33 },
		{ 203, 436, 33 },
		{ 203, 437, 33 },
		{ 203, 438, 33 },
		{ 203, 439, 33 },
		{ 203, 440, 33 },
		{ 203, 441, 33 },
		{ 203, 442, 33 },
		{ 203, 443, 33 },
		{ 203, 444, 33 },
		{ 203, 445, 33 },
		{ 203, 446, 33 },
		{ 203, 447, 33 },
		{ 203, 448, 33 },
		{ 203, 449, 33 },
		{ 203, 450, 33 },
		{ 203, 451, 33 },
		{ 203, 452, 33 },
		{ 203, 453, 33 },
		{ 203, 454, 33 },
		{ 203, 455, 33 },
		{ 203, 456, 33 },
		{ 203, 457, 33 },
		{ 203, 458, 33 },
		{ 203, 459, 33 },
		{ 203, 460, 33 },
		{ 203, 461, 33 },
		{ 203, 462, 33 },
		{ 203, 463, 33 },
		{ 203, 464, 33 },
		{ 203, 465, 33 },
		{ 203, 466, 33 },
		{ 203, 467, 33 },
		{ 203, 468, 33 },
		{ 203, 469, 33 },
		{ 203, 470, 33 },
		{ 203, 471, 33 },
		{ 203, 472, 33 },
		{ 204, 364, 33 },
		{ 204, 365, 33 },
		{ 204, 366, 33 },
		{ 204, 367, 33 },
		{ 204, 368, 33 },
		{ 204, 369, 33 },
		{ 204, 370, 33 },
		{ 204, 371, 33 },
		{ 204, 372, 33 },
		{ 204, 373, 33 },
		{ 204, 374, 33 },
		{ 204, 375, 33 },
		{ 204, 376, 33 },
		{ 204, 377, 33 },
		{ 204, 378, 33 },
		{ 204, 379, 33 },
		{ 204, 380, 33 },
		{ 204, 381, 33 },
		{ 204, 382, 33 },
		{ 204, 383, 33 },
		{ 204, 384, 33 },
		{ 204, 385, 33 },
		{ 204, 386, 33 },
		{ 204, 387, 33 },
		{ 204, 388, 33 },
		{ 204, 389, 33 },
		{ 204, 390, 33 },
		{ 204, 391, 33 },
		{ 204, 392, 33 },
		{ 204, 393, 33 },
		{ 204, 394, 33 },
		{ 204, 395, 33 },
		{ 204, 396, 33 },
		{ 204, 397, 33 },
		{ 204, 398, 33 },
		{ 204, 399, 33 },
		{ 204, 400, 33 },
		{ 204, 401, 33 },
		{ 204, 402, 33 },
		{ 204, 403, 33 },
		{ 204, 404, 33 },
		{ 204, 405, 33 },
		{ 204, 406, 33 },
		{ 204, 407, 33 },
		{ 204, 408, 33 },
		{ 204, 409, 33 },
		{ 204, 410, 33 },
		{ 204, 411, 33 },
		{ 204, 412, 33 },
		{ 204, 413, 33 },
		{ 204, 414, 33 },
		{ 204, 415, 33 },
		{ 204, 416, 33 },
		{ 204, 417, 33 },
		{ 204, 418, 33 },
		{ 204, 419, 33 },
		{ 204, 420, 33 },
		{ 204, 421, 33 },
		{ 204, 422, 33 },
		{ 204, 423, 33 },
		{ 204, 424, 33 },
		{ 204, 425, 33 },
		{ 204, 426, 33 },
		{ 204, 427, 33 },
		{ 204, 428, 33 },
		{ 204, 429, 33 },
		{ 204, 430, 33 },
		{ 204, 431, 33 },
		{ 204, 432, 33 },
		{ 204, 433, 33 },
		{ 204, 434, 33 },
		{ 204, 435, 33 },
		{ 204, 436, 33 },
		{ 204, 437, 33 },
		{ 204, 438, 33 },
		{ 204, 439, 33 },
		{ 204, 440, 33 },
		{ 204, 441, 33 },
		{ 204, 442, 33 },
		{ 204, 443, 33 },
		{ 204, 444, 33 },
		{ 204, 445, 33 },
		{ 204, 446, 33 },
		{ 204, 447, 33 },
		{ 204, 448, 33 },
		{ 204, 449, 33 },
		{ 204, 450, 33 },
		{ 204, 451, 33 },
		{ 204, 452, 33 },
		{ 204, 453, 33 },
		{ 204, 454, 33 },
		{ 204, 455, 33 },
		{ 204, 456, 33 },
		{ 204, 457, 33 },
		{ 204, 458, 33 },
		{ 204, 459, 33 },
		{ 204, 460, 33 },
		{ 204, 461, 33 },
		{ 204, 462, 33 },
		{ 204, 463, 33 },
		{ 204, 464, 33 },
		{ 204, 465, 33 },
		{ 204, 466, 33 },
		{ 204, 467, 33 },
		{ 204, 468, 33 },
		{ 205, 362, 33 },
		{ 205, 363, 33 },
		{ 205, 364, 33 },
		{ 205, 365, 33 },
		{ 205, 366, 33 },
		{ 205, 367, 33 },
		{ 205, 368, 33 },
		{ 205, 369, 33 },
		{ 205, 370, 33 },
		{ 205, 371, 33 },
		{ 205, 372, 33 },
		{ 205, 373, 33 },
		{ 205, 374, 33 },
		{ 205, 375, 33 },
		{ 205, 376, 33 },
		{ 205, 377, 33 },
		{ 205, 378, 33 },
		{ 205, 379, 33 },
		{ 205, 380, 33 },
		{ 205, 381, 33 },
		{ 205, 382, 33 },
		{ 205, 383, 33 },
		{ 205, 384, 33 },
		{ 205, 385, 33 },
		{ 205, 386, 33 },
		{ 205, 387, 33 },
		{ 205, 388, 33 },
		{ 205, 389, 33 },
		{ 205, 390, 33 },
		{ 205, 391, 33 },
		{ 205, 392, 33 },
		{ 205, 393, 33 },
		{ 205, 394, 33 },
		{ 205, 395, 33 },
		{ 205, 396, 33 },
		{ 205, 397, 33 },
		{ 205, 398, 33 },
		{ 205, 399, 33 },
		{ 205, 400, 33 },
		{ 205, 401, 33 },
		{ 205, 402, 33 },
		{ 205, 403, 33 },
		{ 205, 404, 33 },
		{ 205, 405, 33 },
		{ 205, 406, 33 },
		{ 205, 407, 33 },
		{ 205, 408, 33 },
		{ 205, 409, 33 },
		{ 205, 410, 33 },
		{ 205, 411, 33 },
		{ 205, 412, 33 },
		{ 205, 413, 33 },
		{ 205, 414, 33 },
		{ 205, 415, 33 },
		{ 205, 416, 33 },
		{ 205, 417, 33 },
		{ 205, 418, 33 },
		{ 205, 419, 33 },
		{ 205, 420, 33 },
		{ 205, 421, 33 },
		{ 205, 422, 33 },
		{ 205, 423, 33 },
		{ 205, 424, 33 },
		{ 205, 425, 33 },
		{ 205, 426, 33 },
		{ 205, 427, 33 },
		{ 205, 428, 33 },
		{ 205, 429, 33 },
		{ 205, 430, 33 },
		{ 205, 431, 33 },
		{ 205, 432, 33 },
		{ 205, 433, 33 },
		{ 205, 434, 33 },
		{ 205, 435, 33 },
		{ 205, 436, 33 },
		{ 205, 437, 33 },
		{ 205, 438, 33 },
		{ 205, 439, 33 },
		{ 205, 440, 33 },
		{ 205, 441, 33 },
		{ 205, 442, 33 },
		{ 205, 443, 33 },
		{ 205, 444, 33 },
		{ 205, 445, 33 },
		{ 205, 446, 33 },
		{ 205, 447, 33 },
		{ 205, 448, 33 },
		{ 205, 449, 33 },
		{ 205, 450, 33 },
		{ 205, 451, 33 },
		{ 205, 452, 33 },
		{ 205, 453, 33 },
		{ 205, 454, 33 },
		{ 205, 455, 33 },
		{ 205, 456, 33 },
		{ 205, 457, 33 },
		{ 205, 458, 33 },
		{ 205, 459, 33 },
		{ 205, 460, 33 },
		{ 205, 461, 33 },
		{ 205, 462, 33 },
		{ 205, 463, 33 },
		{ 205, 464, 33 },
		{ 206, 360, 33 },
		{ 206, 361, 33 },
		{ 206, 362, 33 },
		{ 206, 363, 33 },
		{ 206, 364, 33 },
		{ 206, 365, 33 },
		{ 206, 366, 33 },
		{ 206, 367, 33 },
		{ 206, 368, 33 },
		{ 206, 369, 33 },
		{ 206, 370, 33 },
		{ 206, 371, 33 },
		{ 206, 372, 33 },
		{ 206, 373, 33 },
		{ 206, 374, 33 },
		{ 206, 375, 33 },
		{ 206, 376, 33 },
		{ 206, 377, 33 },
		{ 206, 378, 33 },
		{ 206, 379, 33 },
		{ 206, 380, 33 },
		{ 206, 381, 33 },
		{ 206, 382, 33 },
		{ 206, 383, 33 },
		{ 206, 384, 33 },
		{ 206, 385, 33 },
		{ 206, 386, 33 },
		{ 206, 387, 33 },
		{ 206, 388, 33 },
		{ 206, 389, 33 },
		{ 206, 390, 33 },
		{ 206, 391, 33 },
		{ 206, 392, 33 },
		{ 206, 393, 33 },
		{ 206, 394, 33 },
		{ 206, 395, 33 },
		{ 206, 396, 33 },
		{ 206, 397, 33 },
		{ 206, 398, 33 },
		{ 206, 399, 33 },
		{ 206, 400, 33 },
		{ 206, 401, 33 },
		{ 206, 402, 33 },
		{ 206, 403, 33 },
		{ 206, 404, 33 },
		{ 206, 405, 33 },
		{ 206, 406, 33 },
		{ 206, 407, 33 },
		{ 206, 408, 33 },
		{ 206, 409, 33 },
		{ 206, 410, 33 },
		{ 206, 411, 33 },
		{ 206, 412, 33 },
		{ 206, 413, 33 },
		{ 206, 414, 33 },
		{ 206, 415, 33 },
		{ 206, 416, 33 },
		{ 206, 417, 33 },
		{ 206, 418, 33 },
		{ 206, 419, 33 },
		{ 206, 420, 33 },
		{ 206, 421, 33 },
		{ 206, 422, 33 },
		{ 206, 423, 33 },
		{ 206, 424, 33 },
		{ 206, 425, 33 },
		{ 206, 426, 33 },
		{ 206, 427, 33 },
		{ 206, 428, 33 },
		{ 206, 429, 33 },
		{ 206, 430, 33 },
		{ 206, 431, 33 },
		{ 206, 432, 33 },
		{ 206, 433, 33 },
		{ 206, 434, 33 },
		{ 206, 435, 33 },
		{ 206, 436, 33 },
		{ 206, 437, 33 },
		{ 206, 438, 33 },
		{ 206, 439, 33 },
		{ 206, 440, 33 },
		{ 206, 441, 33 },
		{ 206, 442, 33 },
		{ 206, 443, 33 },
		{ 206, 444, 33 },
		{ 206, 445, 33 },
		{ 206, 446, 33 },
		{ 206, 447, 33 },
		{ 206, 448, 33 },
		{ 206, 449, 33 },
		{ 206, 450, 33 },
		{ 206, 451, 33 },
		{ 206, 452, 33 },
		{ 206, 453, 33 },
		{ 206, 454, 33 },
		{ 206, 455, 33 },
		{ 206, 456, 33 },
		{ 206, 457, 33 },
		{ 206, 458, 33 },
		{ 206, 459, 33 },
		{ 206, 460, 33 },
		{ 207, 358, 33 },
		{ 207, 359, 33 },
		{ 207, 360, 33 },
		{ 207, 361, 33 },
		{ 207, 362, 33 },
		{ 207, 363, 33 },
		{ 207, 364, 33 },
		{ 207, 365, 33 },
		{ 207, 366, 33 },
		{ 207, 367, 33 },
		{ 207, 368, 33 },
		{ 207, 369, 33 },
		{ 207, 370, 33 },
		{ 207, 371, 33 },
		{ 207, 372, 33 },
		{ 207, 373, 33 },
		{ 207, 374, 33 },
		{ 207, 375, 33 },
		{ 207, 376, 33 },
		{ 207, 377, 33 },
		{ 207, 378, 33 },
		{ 207, 379, 33 },
		{ 207, 380, 33 },
		{ 207, 381, 33 },
		{ 207, 382, 33 },
		{ 207, 383, 33 },
		{ 207, 384, 33 },
		{ 207, 385, 33 },
		{ 207, 386, 33 },
		{ 207, 387, 33 },
		{ 207, 388, 33 },
		{ 207, 389, 33 },
		{ 207, 390, 33 },
		{ 207, 391, 33 },
		{ 207, 392, 33 },
		{ 207, 393, 33 },
		{ 207, 394, 33 },
		{ 207, 395, 33 },
		{ 207, 396, 33 },
		{ 207, 397, 33 },
		{ 207, 398, 33 },
		{ 207, 399, 33 },
		{ 207, 400, 33 },
		{ 207, 401, 33 },
		{ 207, 402, 33 },
		{ 207, 403, 33 },
		{ 207, 404, 33 },
		{ 207, 405, 33 },
		{ 207, 406, 33 },
		{ 207, 407, 33 },
		{ 207, 408, 33 },
		{ 207, 409, 33 },
		{ 207, 410, 33 },
		{ 207, 411, 33 },
		{ 207, 412, 33 },
		{ 207, 413, 33 },
		{ 207, 414, 33 },
		{ 207, 415, 33 },
		{ 207, 416, 33 },
		{ 207, 417, 33 },
		{ 207, 418, 33 },
		{ 207, 419, 33 },
		{ 207, 420, 33 },
		{ 207, 421, 33 },
		{ 207, 422, 33 },
		{ 207, 423, 33 },
		{ 207, 424, 33 },
		{ 207, 425, 33 },
		{ 207, 426, 33 },
		{ 207, 427, 33 },
		{ 207, 428, 33 },
		{ 207, 429, 33 },
		{ 207, 430, 33 },
		{ 207, 431, 33 },
		{ 207, 432, 33 },
		{ 207, 433, 33 },
		{ 207, 434, 33 },
		{ 207, 435, 33 },
		{ 207, 436, 33 },
		{ 207, 437, 33 },
		{ 207, 438, 33 },
		{ 207, 439, 33 },
		{ 207, 440, 33 },
		{ 207, 441, 33 },
		{ 207, 442, 33 },
		{ 207, 443, 33 },
		{ 207, 444, 33 },
		{ 207, 445, 33 },
		{ 207, 446, 33 },
		{ 207, 447, 33 },
		{ 207, 448, 33 },
		{ 207, 449, 33 },
		{ 207, 450, 33 },
		{ 207, 451, 33 },
		{ 207, 452, 33 },
		{ 207, 453, 33 },
		{ 207, 454, 33 },
		{ 207, 455, 33 },
		{ 207, 456, 33 },
		{ 208, 356, 33 },
		{ 208, 357, 33 },
		{ 208, 358, 33 },
		{ 208, 359, 33 },
		{ 208, 360, 33 },
		{ 208, 361, 33 },
		{ 208, 362, 33 },
		{ 208, 363, 33 },
		{ 208, 364, 33 },
		{ 208, 365, 33 },
		{ 208, 366, 33 },
		{ 208, 367, 33 },
		{ 208, 368, 33 },
		{ 208, 369, 33 },
		{ 208, 370, 33 },
		{ 208, 371, 33 },
		{ 208, 372, 33 },
		{ 208, 373, 33 },
		{ 208, 374, 33 },
		{ 208, 375, 33 },
		{ 208, 376, 33 },
		{ 208, 377, 33 },
		{ 208, 378, 33 },
		{ 208, 379, 33 },
		{ 208, 380, 33 },
		{ 208, 381, 33 },
		{ 208, 382, 33 },
		{ 208, 383, 33 },
		{ 208, 384, 33 },
		{ 208, 385, 33 },
		{ 208, 386, 33 },
		{ 208, 387, 33 },
		{ 208, 388, 33 },
		{ 208, 389, 33 },
		{ 208, 390, 33 },
		{ 208, 391, 33 },
		{ 208, 392, 33 },
		{ 208, 393, 33 },
		{ 208, 394, 33 },
		{ 208, 395, 33 },
		{ 208, 396, 33 },
		{ 208, 397, 33 },
		{ 208, 398, 33 },
		{ 208, 399, 33 },
		{ 208, 400, 33 },
		{ 208, 401, 33 },
		{ 208, 402, 33 },
		{ 208, 403, 33 },
		{ 208, 404, 33 },
		{ 208, 405, 33 },
		{ 208, 406, 33 },
		{ 208, 407, 33 },
		{ 208, 408, 33 },
		{ 208, 409, 33 },
		{ 208, 410, 33 },
		{ 208, 411, 33 },
		{ 208, 412, 33 },
		{ 208, 413, 33 },
		{ 208, 414, 33 },
		{ 208, 415, 33 },
		{ 208, 416, 33 },
		{ 208, 417, 33 },
		{ 208, 418, 33 },
		{ 208, 419, 33 },
		{ 208, 420, 33 },
		{ 208, 421, 33 },
		{ 208, 422, 33 },
		{ 208, 423, 33 },
		{ 208, 424, 33 },
		{ 208, 425, 33 },
		{ 208, 426, 33 },
		{ 208, 427, 33 },
		{ 208, 428, 33 },
		{ 208, 429, 33 },
		{ 208, 430, 33 },
		{ 208, 431, 33 },
		{ 208, 432, 33 },
		{ 208, 433, 33 },
		{ 208, 434, 33 },
		{ 208, 435, 33 },
		{ 208, 436, 33 },
		{ 208, 437, 33 },
		{ 208, 438, 33 },
		{ 208, 439, 33 },
		{ 208, 440, 33 },
		{ 208, 441, 33 },
		{ 208, 442, 33 },
		{ 208, 443, 33 },
		{ 208, 444, 33 },
		{ 208, 445, 33 },
		{ 208, 446, 33 },
		{ 208, 447, 33 },
		{ 208, 448, 33 },
		{ 208, 449, 33 },
		{ 208, 450, 33 },
		{ 208, 451, 33 },
		{ 208, 452, 33 },
		{ 209, 354, 33 },
		{ 209, 355, 33 },
		{ 209, 356, 33 },
		{ 209, 357, 33 },
		{ 209, 358, 33 },
		{ 209, 359, 33 },
		{ 209, 360, 33 },
		{ 209, 361, 33 },
		{ 209, 362, 33 },
		{ 209, 363, 33 },
		{ 209, 364, 33 },
		{ 209, 365, 33 },
		{ 209, 366, 33 },
		{ 209, 367, 33 },
		{ 209, 368, 33 },
		{ 209, 369, 33 },
		{ 209, 370, 33 },
		{ 209, 371, 33 },
		{ 209, 372, 33 },
		{ 209, 373, 33 },
		{ 209, 374, 33 },
		{ 209, 375, 33 },
		{ 209, 376, 33 },
		{ 209, 377, 33 },
		{ 209, 378, 33 },
		{ 209, 379, 33 },
		{ 209, 380, 33 },
		{ 209, 381, 33 },
		{ 209, 382, 33 },
		{ 209, 383, 33 },
		{ 209, 384, 33 },
		{ 209, 385, 33 },
		{ 209, 386, 33 },
		{ 209, 387, 33 },
		{ 209, 388, 33 },
		{ 209, 389, 33 },
		{ 209, 390, 33 },
		{ 209, 391, 33 },
		{ 209, 392, 33 },
		{ 209, 393, 33 },
		{ 209, 394, 33 },
		{ 209, 395, 33 },
		{ 209, 396, 33 },
		{ 209, 397, 33 },
		{ 209, 398, 33 },
		{ 209, 399, 33 },
		{ 209, 400, 33 },
		{ 209, 401, 33 },
		{ 209, 402, 33 },
		{ 209, 403, 33 },
		{ 209, 404, 33 },
		{ 209, 405, 33 },
		{ 209, 406, 33 },
		{ 209, 407, 33 },
		{ 209, 408, 33 },
		{ 209, 409, 33 },
		{ 209, 410, 33 },
		{ 209, 411, 33 },
		{ 209, 412, 33 },
		{ 209, 413, 33 },
		{ 209, 414, 33 },
		{ 209, 415, 33 },
		{ 209, 416, 33 },
		{ 209, 417, 33 },
		{ 209, 418, 33 },
		{ 209, 419, 33 },
		{ 209, 420, 33 },
		{ 209, 421, 33 },
		{ 209, 422, 33 },
		{ 209, 423, 33 },
		{ 209, 424, 33 },
		{ 209, 425, 33 },
		{ 209, 426, 33 },
		{ 209, 427, 33 },
		{ 209, 428, 33 },
		{ 209, 429, 33 },
		{ 209, 430, 33 },
		{ 209, 431, 33 },
		{ 209, 432, 33 },
		{ 209, 433, 33 },
		{ 209, 434, 33 },
		{ 209, 435, 33 },
		{ 209, 436, 33 },
		{ 209, 437, 33 },
		{ 209, 438, 33 },
		{ 209, 439, 33 },
		{ 209, 440, 33 },
		{ 209, 441, 33 },
		{ 209, 442, 33 },
		{ 209, 443, 33 },
		{ 209, 444, 33 },
		{ 209, 445, 33 },
		{ 209, 446, 33 },
		{ 209, 447, 33 },
		{ 209, 448, 33 },
		{ 210, 352, 33 },
		{ 210, 353, 33 },
		{ 210, 354, 33 },
		{ 210, 355, 33 },
		{ 210, 356, 33 },
		{ 210, 357, 33 },
		{ 210, 358, 33 },
		{ 210, 359, 33 },
		{ 210, 360, 33 },
		{ 210, 361, 33 },
		{ 210, 362, 33 },
		{ 210, 363, 33 },
		{ 210, 364, 33 },
		{ 210, 365, 33 },
		{ 210, 366, 33 },
		{ 210, 367, 33 },
		{ 210, 368, 33 },
		{ 210, 369, 33 },
		{ 210, 370, 33 },
		{ 210, 371, 33 },
		{ 210, 372, 33 },
		{ 210, 373, 33 },
		{ 210, 374, 33 },
		{ 210, 375, 33 },
		{ 210, 376, 33 },
		{ 210, 377, 33 },
		{ 210, 378, 33 },
		{ 210, 379, 33 },
		{ 210, 380, 33 },
		{ 210, 381, 33 },
		{ 210, 382, 33 },
		{ 210, 383, 33 },
		{ 210, 384, 33 },
		{ 210, 385, 33 },
		{ 210, 386, 33 },
		{ 210, 387, 33 },
		{ 210, 388, 33 },
		{ 210, 389, 33 },
		{ 210, 390, 33 },
		{ 210, 391, 33 },
		{ 210, 392, 33 },
		{ 210, 393, 33 },
		{ 210, 394, 33 },
		{ 210, 395, 33 },
		{ 210, 396, 33 },
		{ 210, 397, 33 },
		{ 210, 398, 33 },
		{ 210, 399, 33 },
		{ 210, 400, 33 },
		{ 210, 401, 33 },
		{ 210, 402, 33 },
		{ 210, 403, 33 },
		{ 210, 404, 33 },
		{ 210, 405, 33 },
		{ 210, 406, 33 },
		{ 210, 407, 33 },
		{ 210, 408, 33 },
		{ 210, 409, 33 },
		{ 210, 410, 33 },
		{ 210, 411, 33 },
		{ 210, 412, 33 },
		{ 210, 413, 33 },
		{ 210, 414, 33 },
		{ 210, 415, 33 },
		{ 210, 416, 33 },
		{ 210, 417, 33 },
		{ 210, 418, 33 },
		{ 210, 419, 33 },
		{ 210, 420, 33 },
		{ 210, 421, 33 },
		{ 210, 422, 33 },
		{ 210, 423, 33 },
		{ 210, 424, 33 },
		{ 210, 425, 33 },
		{ 210, 426, 33 },
		{ 210, 427, 33 },
		{ 210, 428, 33 },
		{ 210, 429, 33 },
		{ 210, 430, 33 },
		{ 210, 431, 33 },
		{ 210, 432, 33 },
		{ 210, 433, 33 },
		{ 210, 434, 33 },
		{ 210, 435, 33 },
		{ 210, 436, 33 },
		{ 210, 437, 33 },
		{ 210, 438, 33 },
		{ 210, 439, 33 },
		{ 210, 440, 33 },
		{ 210, 441, 33 },
		{ 210, 442, 33 },
		{ 210, 443, 33 },
		{ 210, 444, 33 },
		{ 211, 350, 33 },
		{ 211, 351, 33 },
		{ 211, 352, 33 },
		{ 211, 353, 33 },
		{ 211, 354, 33 },
		{ 211, 355, 33 },
		{ 211, 356, 33 },
		{ 211, 357, 33 },
		{ 211, 358, 33 },
		{ 211, 359, 33 },
		{ 211, 360, 33 },
		{ 211, 361, 33 },
		{ 211, 362, 33 },
		{ 211, 363, 33 },
		{ 211, 364, 33 },
		{ 211, 365, 33 },
		{ 211, 366, 33 },
		{ 211, 367, 33 },
		{ 211, 368, 33 },
		{ 211, 369, 33 },
		{ 211, 370, 33 },
		{ 211, 371, 33 },
		{ 211, 372, 33 },
		{ 211, 373, 33 },
		{ 211, 374, 33 },
		{ 211, 375, 33 },
		{ 211, 376, 33 },
		{ 211, 377, 33 },
		{ 211, 378, 33 },
		{ 211, 379, 33 },
		{ 211, 380, 33 },
		{ 211, 381, 33 },
		{ 211, 382, 33 },
		{ 211, 383, 33 },
		{ 211, 384, 33 },
		{ 211, 385, 33 },
		{ 211, 386, 33 },
		{ 211, 387, 33 },
		{ 211, 388, 33 },
		{ 211, 389, 33 },
		{ 211, 390, 33 },
		{ 211, 391, 33 },
		{ 211, 392, 33 },
		{ 211, 393, 33 },
		{ 211, 394, 33 },
		{ 211, 395, 33 },
		{ 211, 396, 33 },
		{ 211, 397, 33 },
		{ 211, 398, 33 },
		{ 211, 399, 33 },
		{ 211, 400, 33 },
		{ 211, 401, 33 },
		{ 211, 402, 33 },
		{ 211, 403, 33 },
		{ 211, 404, 33 },
		{ 211, 405, 33 },
		{ 211, 406, 33 },
		{ 211, 407, 33 },
		{ 211, 408, 33 },
		{ 211, 409, 33 },
		{ 211, 410, 33 },
		{ 211, 411, 33 },
		{ 211, 412, 33 },
		{ 211, 413, 33 },
		{ 211, 414, 33 },
		{ 211, 415, 33 },
		{ 211, 416, 33 },
		{ 211, 417, 33 },
		{ 211, 418, 33 },
		{ 211, 419, 33 },
		{ 211, 420, 33 },
		{ 211, 421, 33 },
		{ 211, 422, 33 },
		{ 211, 423, 33 },
		{ 211, 424, 33 },
		{ 211, 425, 33 },
		{ 211, 426, 33 },
		{ 211, 427, 33 },
		{ 211, 428, 33 },
		{ 211, 429, 33 },
		{ 211, 430, 33 },
		{ 211, 431, 33 },
		{ 211, 432, 33 },
		{ 211, 433, 33 },
		{ 211, 434, 33 },
		{ 211, 435, 33 },
		{ 211, 436, 33 },
		{ 211, 437, 33 },
		{ 211, 438, 33 },
		{ 211, 439, 33 },
		{ 211, 440, 33 },
		{ 212, 348, 33 },
		{ 212, 349, 33 },
		{ 212, 350, 33 },
		{ 212, 351, 33 },
		{ 212, 352, 33 },
		{ 212, 353, 33 },
		{ 212, 354, 33 },
		{ 212, 355, 33 },
		{ 212, 356, 33 },
		{ 212, 357, 33 },
		{ 212, 358, 33 },
		{ 212, 359, 33 },
		{ 212, 360, 33 },
		{ 212, 361, 33 },
		{ 212, 362, 33 },
		{ 212, 363, 33 },
		{ 212, 364, 33 },
		{ 212, 365, 33 },
		{ 212, 366, 33 },
		{ 212, 367, 33 },
		{ 212, 368, 33 },
		{ 212, 369, 33 },
		{ 212, 370, 33 },
		{ 212, 371, 33 },
		{ 212, 372, 33 },
		{ 212, 373, 33 },
		{ 212, 374, 33 },
		{ 212, 375, 33 },
		{ 212, 376, 33 },
		{ 212, 377, 33 },
		{ 212, 378, 33 },
		{ 212, 379, 33 },
		{ 212, 380, 33 },
		{ 212, 381, 33 },
		{ 212, 382, 33 },
		{ 212, 383, 33 },
		{ 212, 384, 33 },
		{ 212, 385, 33 },
		{ 212, 386, 33 },
		{ 212, 387, 33 },
		{ 212, 388, 33 },
		{ 212, 389, 33 },
		{ 212, 390, 33 },
		{ 212, 391, 33 },
		{ 212, 392, 33 },
		{ 212, 393, 33 },
		{ 212, 394, 33 },
		{ 212, 395, 33 },
		{ 212, 396, 33 },
		{ 212, 397, 33 },
		{ 212, 398, 33 },
		{ 212, 399, 33 },
		{ 212, 400, 33 },
		{ 212, 401, 33 },
		{ 212, 402, 33 },
		{ 212, 403, 33 },
		{ 212, 404, 33 },
		{ 212, 405, 33 },
		{ 212, 406, 33 },
		{ 212, 407, 33 },
		{ 212, 408, 33 },
		{ 212, 409, 33 },
		{ 212, 410, 33 },
		{ 212, 411, 33 },
		{ 212, 412, 33 },
		{ 212, 413, 33 },
		{ 212, 414, 33 },
		{ 212, 415, 33 },
		{ 212, 416, 33 },
		{ 212, 417, 33 },
		{ 212, 418, 33 },
		{ 212, 419, 33 },
		{ 212, 420, 33 },
		{ 212, 421, 33 },
		{ 212, 422, 33 },
		{ 212, 423, 33 },
		{ 212, 424, 33 },
		{ 212, 425, 33 },
		{ 212, 426, 33 },
		{ 212, 427, 33 },
		{ 212, 428, 33 },
		{ 212, 429, 33 },
		{ 212, 430, 33 },
		{ 212, 431, 33 },
		{ 212, 432, 33 },
		{ 212, 433, 33 },
		{ 212, 434, 33 },
		{ 212, 435, 33 },
		{ 212, 436, 33 },
		{ 213, 346, 33 },
		{ 213, 347, 33 },
		{ 213, 348, 33 },
		{ 213, 349, 33 },
		{ 213, 350, 33 },
		{ 213, 351, 33 },
		{ 213, 352, 33 },
		{ 213, 353, 33 },
		{ 213, 354, 33 },
		{ 213, 355, 33 },
		{ 213, 356, 33 },
		{ 213, 357, 33 },
		{ 213, 358, 33 },
		{ 213, 359, 33 },
		{ 213, 360, 33 },
		{ 213, 361, 33 },
		{ 213, 362, 33 },
		{ 213, 363, 33 },
		{ 213, 364, 33 },
		{ 213, 365, 33 },
		{ 213, 366, 33 },
		{ 213, 367, 33 },
		{ 213, 368, 33 },
		{ 213, 369, 33 },
		{ 213, 370, 33 },
		{ 213, 371, 33 },
		{ 213, 372, 33 },
		{ 213, 373, 33 },
		{ 213, 374, 33 },
		{ 213, 375, 33 },
		{ 213, 376, 33 },
		{ 213, 377, 33 },
		{ 213, 378, 33 },
		{ 213, 379, 33 },
		{ 213, 380, 33 },
		{ 213, 381, 33 },
		{ 213, 382, 33 },
		{ 213, 383, 33 },
		{ 213, 384, 33 },
		{ 213, 385, 33 },
		{ 213, 386, 33 },
		{ 213, 387, 33 },
		{ 213, 388, 33 },
		{ 213, 389, 33 },
		{ 213, 390, 33 },
		{ 213, 391, 33 },
		{ 213, 392, 33 },
		{ 213, 393, 33 },
		{ 213, 394, 33 },
		{ 213, 395, 33 },
		{ 213, 396, 33 },
		{ 213, 397, 33 },
		{ 213, 398, 33 },
		{ 213, 399, 33 },
		{ 213, 400, 33 },
		{ 213, 401, 33 },
		{ 213, 402, 33 },
		{ 213, 403, 33 },
		{ 213, 404, 33 },
		{ 213, 405, 33 },
		{ 213, 406, 33 },
		{ 213, 407, 33 },
		{ 213, 408, 33 },
		{ 213, 409, 33 },
		{ 213, 410, 33 },
		{ 213, 411, 33 },
		{ 213, 412, 33 },
		{ 213, 413, 33 },
		{ 213, 414, 33 },
		{ 213, 415, 33 },
		{ 213, 416, 33 },
		{ 213, 417, 33 },
		{ 213, 418, 33 },
		{ 213, 419, 33 },
		{ 213, 420, 33 },
		{ 213, 421, 33 },
		{ 213, 422, 33 },
		{ 213, 423, 33 },
		{ 213, 424, 33 },
		{ 213, 425, 33 },
		{ 213, 426, 33 },
		{ 213, 427, 33 },
		{ 213, 428, 33 },
		{ 213, 429, 33 },
		{ 213, 430, 33 },
		{ 213, 431, 33 },
		{ 213, 432, 33 },
		{ 214, 344, 33 },
		{ 214, 345, 33 },
		{ 214, 346, 33 },
		{ 214, 347, 33 },
		{ 214, 348, 33 },
		{ 214, 349, 33 },
		{ 214, 350, 33 },
		{ 214, 351, 33 },
		{ 214, 352, 33 },
		{ 214, 353, 33 },
		{ 214, 354, 33 },
		{ 214, 355, 33 },
		{ 214, 356, 33 },
		{ 214, 357, 33 },
		{ 214, 358, 33 },
		{ 214, 359, 33 },
		{ 214, 360, 33 },
		{ 214, 361, 33 },
		{ 214, 362, 33 },
		{ 214, 363, 33 },
		{ 214, 364, 33 },
		{ 214, 365, 33 },
		{ 214, 366, 33 },
		{ 214, 367, 33 },
		{ 214, 368, 33 },
		{ 214, 369, 33 },
		{ 214, 370, 33 },
		{ 214, 371, 33 },
		{ 214, 372, 33 },
		{ 214, 373, 33 },
		{ 214, 374, 33 },
		{ 214, 375, 33 },
		{ 214, 376, 33 },
		{ 214, 377, 33 },
		{ 214, 378, 33 },
		{ 214, 379, 33 },
		{ 214, 380, 33 },
		{ 214, 381, 33 },
		{ 214, 382, 33 },
		{ 214, 383, 33 },
		{ 214, 384, 33 },
		{ 214, 385, 33 },
		{ 214, 386, 33 },
		{ 214, 387, 33 },
		{ 214, 388, 33 },
		{ 214, 389, 33 },
		{ 214, 390, 33 },
		{ 214, 391, 33 },
		{ 214, 392, 33 },
		{ 214, 393, 33 },
		{ 214, 394, 33 },
		{ 214, 395, 33 },
		{ 214, 396, 33 },
		{ 214, 397, 33 },
		{ 214, 398, 33 },
		{ 214, 399, 33 },
		{ 214, 400, 33 },
		{ 214, 401, 33 },
		{ 214, 402, 33 },
		{ 214, 403, 33 },
		{ 214, 404, 33 },
		{ 214, 405, 33 },
		{ 214, 406, 33 },
		{ 214, 407, 33 },
		{ 214, 408, 33 },
		{ 214, 409, 33 },
		{ 214, 410, 33 },
		{ 214, 411, 33 },
		{ 214, 412, 33 },
		{ 214, 413, 33 },
		{ 214, 414, 33 },
		{ 214, 415, 33 },
		{ 214, 416, 33 },
		{ 214, 417, 33 },
		{ 214, 418, 33 },
		{ 214, 419, 33 },
		{ 214, 420, 33 },
		{ 214, 421, 33 },
		{ 214, 422, 33 },
		{ 214, 423, 33 },
		{ 214, 424, 33 },
		{ 214, 425, 33 },
		{ 214, 426, 33 },
		{ 214, 427, 33 },
		{ 214, 428, 33 },
		{ 215, 342, 33 },
		{ 215, 343, 33 },
		{ 215, 344, 33 },
		{ 215, 345, 33 },
		{ 215, 346, 33 },
		{ 215, 347, 33 },
		{ 215, 348, 33 },
		{ 215, 349, 33 },
		{ 215, 350, 33 },
		{ 215, 351, 33 },
		{ 215, 352, 33 },
		{ 215, 353, 33 },
		{ 215, 354, 33 },
		{ 215, 355, 33 },
		{ 215, 356, 33 },
		{ 215, 357, 33 },
		{ 215, 358, 33 },
		{ 215, 359, 33 },
		{ 215, 360, 33 },
		{ 215, 361, 33 },
		{ 215, 362, 33 },
		{ 215, 363, 33 },
		{ 215, 364, 33 },
		{ 215, 365, 33 },
		{ 215, 366, 33 },
		{ 215, 367, 33 },
		{ 215, 368, 33 },
		{ 215, 369, 33 },
		{ 215, 370, 33 },
		{ 215, 371, 33 },
		{ 215, 372, 33 },
		{ 215, 373, 33 },
		{ 215, 374, 33 },
		{ 215, 375, 33 },
		{ 215, 376, 33 },
		{ 215, 377, 33 },
		{ 215, 378, 33 },
		{ 215, 379, 33 },
		{ 215, 380, 33 },
		{ 215, 381, 33 },
		{ 215, 382, 33 },
		{ 215, 383, 33 },
		{ 215, 384, 33 },
		{ 215, 385, 33 },
		{ 215, 386, 33 },
		{ 215, 387, 33 },
		{ 215, 388, 33 },
		{ 215, 389, 33 },
		{ 215, 390, 33 },
		{ 215, 391, 33 },
		{ 215, 392, 33 },
		{ 215, 393, 33 },
		{ 215, 394, 33 },
		{ 215, 395, 33 },
		{ 215, 396, 33 },
		{ 215, 397, 33 },
		{ 215, 398, 33 },
		{ 215, 399, 33 },
		{ 215, 400, 33 },
		{ 215, 401, 33 },
		{ 215, 402, 33 },
		{ 215, 403, 33 },
		{ 215, 404, 33 },
		{ 215, 405, 33 },
		{ 215, 406, 33 },
		{ 215, 407, 33 },
		{ 215, 408, 33 },
		{ 215, 409, 33 },
		{ 215, 410, 33 },
		{ 215, 411, 33 },
		{ 215, 412, 33 },
		{ 215, 413, 33 },
		{ 215, 414, 33 },
		{ 215, 415, 33 },
		{ 215, 416, 33 },
		{ 215, 417, 33 },
		{ 215, 418, 33 },
		{ 215, 419, 33 },
		{ 215, 420, 33 },
		{ 215, 421, 33 },
		{ 215, 422, 33 },
		{ 215, 423, 33 },
		{ 215, 424, 33 },
		{ 216, 340, 33 },
		{ 216, 341, 33 },
		{ 216, 342, 33 },
		{ 216, 343, 33 },
		{ 216, 344, 33 },
		{ 216, 345, 33 },
		{ 216, 346, 33 },
		{ 216, 347, 33 },
		{ 216, 348, 33 },
		{ 216, 349, 33 },
		{ 216, 350, 33 },
		{ 216, 351, 33 },
		{ 216, 352, 33 },
		{ 216, 353, 33 },
		{ 216, 354, 33 },
		{ 216, 355, 33 },
		{ 216, 356, 33 },
		{ 216, 357, 33 },
		{ 216, 358, 33 },
		{ 216, 359, 33 },
		{ 216, 360, 33 },
		{ 216, 361, 33 },
		{ 216, 362, 33 },
		{ 216, 363, 33 },
		{ 216, 364, 33 },
		{ 216, 365, 33 },
		{ 216, 366, 33 },
		{ 216, 367, 33 },
		{ 216, 368, 33 },
		{ 216, 369, 33 },
		{ 216, 370, 33 },
		{ 216, 371, 33 },
		{ 216, 372, 33 },
		{ 216, 373, 33 },
		{ 216, 374, 33 },
		{ 216, 375, 33 },
		{ 216, 376, 33 },
		{ 216, 377, 33 },
		{ 216, 378, 33 },
		{ 216, 379, 33 },
		{ 216, 380, 33 },
		{ 216, 381, 33 },
		{ 216, 382, 33 },
		{ 216, 383, 33 },
		{ 216, 384, 33 },
		{ 216, 385, 33 },
		{ 216, 386, 33 },
		{ 216, 387, 33 },
		{ 216, 388, 33 },
		{ 216, 389, 33 },
		{ 216, 390, 33 },
		{ 216, 391, 33 },
		{ 216, 392, 33 },
		{ 216, 393, 33 },
		{ 216, 394, 33 },
		{ 216, 395, 33 },
		{ 216, 396, 33 },
		{ 216, 397, 33 },
		{ 216, 398, 33 },
		{ 216, 399, 33 },
		{ 216, 400, 33 },
		{ 216, 401, 33 },
		{ 216, 402, 33 },
		{ 216, 403, 33 },
		{ 216, 404, 33 },
		{ 216, 405, 33 },
		{ 216, 406, 33 },
		{ 216, 407, 33 },
		{ 216, 408, 33 },
		{ 216, 409, 33 },
		{ 216, 410, 33 },
		{ 216, 411, 33 },
		{ 216, 412, 33 },
		{ 216, 413, 33 },
		{ 216, 414, 33 },
		{ 216, 415, 33 },
		{ 216, 416, 33 },
		{ 216, 417, 33 },
		{ 216, 418, 33 },
		{ 216, 419, 33 },
		{ 216, 420, 33 },
		{ 217, 338, 33 },
		{ 217, 339, 33 },
		{ 217, 340, 33 },
		{ 217, 341, 33 },
		{ 217, 342, 33 },
		{ 217, 343, 33 },
		{ 217, 344, 33 },
		{ 217, 345, 33 },
		{ 217, 346, 33 },
		{ 217, 347, 33 },
		{ 217, 348, 33 },
		{ 217, 349, 33 },
		{ 217, 350, 33 },
		{ 217, 351, 33 },
		{ 217, 352, 33 },
		{ 217, 353, 33 },
		{ 217, 354, 33 },
		{ 217, 355, 33 },
		{ 217, 356, 33 },
		{ 217, 357, 33 },
		{ 217, 358, 33 },
		{ 217, 359, 33 },
		{ 217, 360, 33 },
		{ 217, 361, 33 },
		{ 217, 362, 33 },
		{ 217, 363, 33 },
		{ 217, 364, 33 },
		{ 217, 365, 33 },
		{ 217, 366, 33 },
		{ 217, 367, 33 },
		{ 217, 368, 33 },
		{ 217, 369, 33 },
		{ 217, 370, 33 },
		{ 217, 371, 33 },
		{ 217, 372, 33 },
		{ 217, 373, 33 },
		{ 217, 374, 33 },
		{ 217, 375, 33 },
		{ 217, 376, 33 },
		{ 217, 377, 33 },
		{ 217, 378, 33 },
		{ 217, 379, 33 },
		{ 217, 380, 33 },
		{ 217, 381, 33 },
		{ 217, 382, 33 },
		{ 217, 383, 33 },
		{ 217, 384, 33 },
		{ 217, 385, 33 },
		{ 217, 386, 33 },
		{ 217, 387, 33 },
		{ 217, 388, 33 },
		{ 217, 389, 33 },
		{ 217, 390, 33 },
		{ 217, 391, 33 },
		{ 217, 392, 33 },
		{ 217, 393, 33 },
		{ 217, 394, 33 },
		{ 217, 395, 33 },
		{ 217, 396, 33 },
		{ 217, 397, 33 },
		{ 217, 398, 33 },
		{ 217, 399, 33 },
		{ 217, 400, 33 },
		{ 217, 401, 33 },
		{ 217, 402, 33 },
		{ 217, 403, 33 },
		{ 217, 404, 33 },
		{ 217, 405, 33 },
		{ 217, 406, 33 },
		{ 217, 407, 33 },
		{ 217, 408, 33 },
		{ 217, 409, 33 },
		{ 217, 410, 33 },
		{ 217, 411, 33 },
		{ 217, 412, 33 },
		{ 217, 413, 33 },
		{ 217, 414, 33 },
		{ 217, 415, 33 },
		{ 217, 416, 33 },
		{ 218, 336, 33 },
		{ 218, 337, 33 },
		{ 218, 338, 33 },
		{ 218, 339, 33 },
		{ 218, 340, 33 },
		{ 218, 341, 33 },
		{ 218, 342, 33 },
		{ 218, 343, 33 },
		{ 218, 344, 33 },
		{ 218, 345, 33 },
		{ 218, 346, 33 },
		{ 218, 347, 33 },
		{ 218, 348, 33 },
		{ 218, 349, 33 },
		{ 218, 350, 33 },
		{ 218, 351, 33 },
		{ 218, 352, 33 },
		{ 218, 353, 33 },
		{ 218, 354, 33 },
		{ 218, 355, 33 },
		{ 218, 356, 33 },
		{ 218, 357, 33 },
		{ 218, 358, 33 },
		{ 218, 359, 33 },
		{ 218, 360, 33 },
		{ 218, 361, 33 },
		{ 218, 362, 33 },
		{ 218, 363, 33 },
		{ 218, 364, 33 },
		{ 218, 365, 33 },
		{ 218, 366, 33 },
		{ 218, 367, 33 },
		{ 218, 368, 33 },
		{ 218, 369, 33 },
		{ 218, 370, 33 },
		{ 218, 371, 33 },
		{ 218, 372, 33 },
		{ 218, 373, 33 },
		{ 218, 374, 33 },
		{ 218, 375, 33 },
		{ 218, 376, 33 },
		{ 218, 377, 33 },
		{ 218, 378, 33 },
		{ 218, 379, 33 },
		{ 218, 380, 33 },
		{ 218, 381, 33 },
		{ 218, 382, 33 },
		{ 218, 383, 33 },
		{ 218, 384, 33 },
		{ 218, 385, 33 },
		{ 218, 386, 33 },
		{ 218, 387, 33 },
		{ 218, 388, 33 },
		{ 218, 389, 33 },
		{ 218, 390, 33 },
		{ 218, 391, 33 },
		{ 218, 392, 33 },
		{ 218, 393, 33 },
		{ 218, 394, 33 },
		{ 218, 395, 33 },
		{ 218, 396, 33 },
		{ 218, 397, 33 },
		{ 218, 398, 33 },
		{ 218, 399, 33 },
		{ 218, 400, 33 },
		{ 218, 401, 33 },
		{ 218, 402, 33 },
		{ 218, 403, 33 },
		{ 218, 404, 33 },
		{ 218, 405, 33 },
		{ 218, 406, 33 },
		{ 218, 407, 33 },
		{ 218, 408, 33 },
		{ 218, 409, 33 },
		{ 218, 410, 33 },
		{ 218, 411, 33 },
		{ 218, 412, 33 },
		{ 219, 334, 33 },
		{ 219, 335, 33 },
		{ 219, 336, 33 },
		{ 219, 337, 33 },
		{ 219, 338, 33 },
		{ 219, 339, 33 },
		{ 219, 340, 33 },
		{ 219, 341, 33 },
		{ 219, 342, 33 },
		{ 219, 343, 33 },
		{ 219, 344, 33 },
		{ 219, 345, 33 },
		{ 219, 346, 33 },
		{ 219, 347, 33 },
		{ 219, 348, 33 },
		{ 219, 349, 33 },
		{ 219, 350, 33 },
		{ 219, 351, 33 },
		{ 219, 352, 33 },
		{ 219, 353, 33 },
		{ 219, 354, 33 },
		{ 219, 355, 33 },
		{ 219, 356, 33 },
		{ 219, 357, 33 },
		{ 219, 358, 33 },
		{ 219, 359, 33 },
		{ 219, 360, 33 },
		{ 219, 361, 33 },
		{ 219, 362, 33 },
		{ 219, 363, 33 },
		{ 219, 364, 33 },
		{ 219, 365, 33 },
		{ 219, 366, 33 },
		{ 219, 367, 33 },
		{ 219, 368, 33 },
		{ 219, 369, 33 },
		{ 219, 370, 33 },
		{ 219, 371, 33 },
		{ 219, 372, 33 },
		{ 219, 373, 33 },
		{ 219, 374, 33 },
		{ 219, 375, 33 },
		{ 219, 376, 33 },
		{ 219, 377, 33 },
		{ 219, 378, 33 },
		{ 219, 379, 33 },
		{ 219, 380, 33 },
		{ 219, 381, 33 },
		{ 219, 382, 33 },
		{ 219, 383, 33 },
		{ 219, 384, 33 },
		{ 219, 385, 33 },
		{ 219, 386, 33 },
		{ 219, 387, 33 },
		{ 219, 388, 33 },
		{ 219, 389, 33 },
		{ 219, 390, 33 },
		{ 219, 391, 33 },
		{ 219, 392, 33 },
		{ 219, 393, 33 },
		{ 219, 394, 33 },
		{ 219, 395, 33 },
		{ 219, 396, 33 },
		{ 219, 397, 33 },
		{ 219, 398, 33 },
		{ 219, 399, 33 },
		{ 219, 400, 33 },
		{ 219, 401, 33 },
		{ 219, 402, 33 },
		{ 219, 403, 33 },
		{ 219, 404, 33 },
		{ 219, 405, 33 },
		{ 219, 406, 33 },
		{ 219, 407, 33 },
		{ 219, 408, 33 },
		{ 220, 332, 33 },
		{ 220, 333, 33 },
		{ 220, 334, 33 },
		{ 220, 335, 33 },
		{ 220, 336, 33 },
		{ 220, 337, 33 },
		{ 220, 338, 33 },
		{ 220, 339, 33 },
		{ 220, 340, 33 },
		{ 220, 341, 33 },
		{ 220, 342, 33 },
		{ 220, 343, 33 },
		{ 220, 344, 33 },
		{ 220, 345, 33 },
		{ 220, 346, 33 },
		{ 220, 347, 33 },
		{ 220, 348, 33 },
		{ 220, 349, 33 },
		{ 220, 350, 33 },
		{ 220, 351, 33 },
		{ 220, 352, 33 },
		{ 220, 353, 33 },
		{ 220, 354, 33 },
		{ 220, 355, 33 },
		{ 220, 356, 33 },
		{ 220, 357, 33 },
		{ 220, 358, 33 },
		{ 220, 359, 33 },
		{ 220, 360, 33 },
		{ 220, 361, 33 },
		{ 220, 362, 33 },
		{ 220, 363, 33 },
		{ 220, 364, 33 },
		{ 220, 365, 33 },
		{ 220, 366, 33 },
		{ 220, 367, 33 },
		{ 220, 368, 33 },
		{ 220, 369, 33 },
		{ 220, 370, 33 },
		{ 220, 371, 33 },
		{ 220, 372, 33 },
		{ 220, 373, 33 },
		{ 220, 374, 33 },
		{ 220, 375, 33 },
		{ 220, 376, 33 },
		{ 220, 377, 33 },
		{ 220, 378, 33 },
		{ 220, 379, 33 },
		{ 220, 380, 33 },
		{ 220, 381, 33 },
		{ 220, 382, 33 },
		{ 220, 383, 33 },
		{ 220, 384, 33 },
		{ 220, 385, 33 },
		{ 220, 386, 33 },
		{ 220, 387, 33 },
		{ 220, 388, 33 },
		{ 220, 389, 33 },
		{ 220, 390, 33 },
		{ 220, 391, 33 },
		{ 220, 392, 33 },
		{ 220, 393, 33 },
		{ 220, 394, 33 },
		{ 220, 395, 33 },
		{ 220, 396, 33 },
		{ 220, 397, 33 },
		{ 220, 398, 33 },
		{ 220, 399, 33 },
		{ 220, 400, 33 },
		{ 220, 401, 33 },
		{ 220, 402, 33 },
		{ 220, 403, 33 },
		{ 220, 404, 33 },
		{ 221, 330, 33 },
		{ 221, 331, 33 },
		{ 221, 332, 33 },
		{ 221, 333, 33 },
		{ 221, 334, 33 },
		{ 221, 335, 33 },
		{ 221, 336, 33 },
		{ 221, 337, 33 },
		{ 221, 338, 33 },
		{ 221, 339, 33 },
		{ 221, 340, 33 },
		{ 221, 341, 33 },
		{ 221, 342, 33 },
		{ 221, 343, 33 },
		{ 221, 344, 33 },
		{ 221, 345, 33 },
		{ 221, 346, 33 },
		{ 221, 347, 33 },
		{ 221, 348, 33 },
		{ 221, 349, 33 },
		{ 221, 350, 33 },
		{ 221, 351, 33 },
		{ 221, 352, 33 },
		{ 221, 353, 33 },
		{ 221, 354, 33 },
		{ 221, 355, 33 },
		{ 221, 356, 33 },
		{ 221, 357, 33 },
		{ 221, 358, 33 },
		{ 221, 359, 33 },
		{ 221, 360, 33 },
		{ 221, 361, 33 },
		{ 221, 362, 33 },
		{ 221, 363, 33 },
		{ 221, 364, 33 },
		{ 221, 365, 33 },
		{ 221, 366, 33 },
		{ 221, 367, 33 },
		{ 221, 368, 33 },
		{ 221, 369, 33 },
		{ 221, 370, 33 },
		{ 221, 371, 33 },
		{ 221, 372, 33 },
		{ 221, 373, 33 },
		{ 221, 374, 33 },
		{ 221, 375, 33 },
		{ 221, 376, 33 },
		{ 221, 377, 33 },
		{ 221, 378, 33 },
		{ 221, 379, 33 },
		{ 221, 380, 33 },
		{ 221, 381, 33 },
		{ 221, 382, 33 },
		{ 221, 383, 33 },
		{ 221, 384, 33 },
		{ 221, 385, 33 },
		{ 221, 386, 33 },
		{ 221, 387, 33 },
		{ 221, 388, 33 },
		{ 221, 389, 33 },
		{ 221, 390, 33 },
		{ 221, 391, 33 },
		{ 221, 392, 33 },
		{ 221, 393, 33 },
		{ 221, 394, 33 },
		{ 221, 395, 33 },
		{ 221, 396, 33 },
		{ 221, 397, 33 },
		{ 221, 398, 33 },
		{ 221, 399, 33 },
		{ 221, 400, 33 },
		{ 222, 328, 33 },
		{ 222, 329, 33 },
		{ 222, 330, 33 },
		{ 222, 331, 33 },
		{ 222, 332, 33 },
		{ 222, 333, 33 },
		{ 222, 334, 33 },
		{ 222, 335, 33 },
		{ 222, 336, 33 },
		{ 222, 337, 33 },
		{ 222, 338, 33 },
		{ 222, 339, 33 },
		{ 222, 340, 33 },
		{ 222, 341, 33 },
		{ 222, 342, 33 },
		{ 222, 343, 33 },
		{ 222, 344, 33 },
		{ 222, 345, 33 },
		{ 222, 346, 33 },
		{ 222, 347, 33 },
		{ 222, 348, 33 },
		{ 222, 349, 33 },
		{ 222, 350, 33 },
		{ 222, 351, 33 },
		{ 222, 352, 33 },
		{ 222, 353, 33 },
		{ 222, 354, 33 },
		{ 222, 355, 33 },
		{ 222, 356, 33 },
		{ 222, 357, 33 },
		{ 222, 358, 33 },
		{ 222, 359, 33 },
		{ 222, 360, 33 },
		{ 222, 361, 33 },
		{ 222, 362, 33 },
		{ 222, 363, 33 },
		{ 222, 364, 33 },
		{ 222, 365, 33 },
		{ 222, 366, 33 },
		{ 222, 367, 33 },
		{ 222, 368, 33 },
		{ 222, 369, 33 },
		{ 222, 370, 33 },
		{ 222, 371, 33 },
		{ 222, 372, 33 },
		{ 222, 373, 33 },
		{ 222, 374, 33 },
		{ 222, 375, 33 },
		{ 222, 376, 33 },
		{ 222, 377, 33 },
		{ 222, 378, 33 },
		{ 222, 379, 33 },
		{ 222, 380, 33 },
		{ 222, 381, 33 },
		{ 222, 382, 33 },
		{ 222, 383, 33 },
		{ 222, 384, 33 },
		{ 222, 385, 33 },
		{ 222, 386, 33 },
		{ 222, 387, 33 },
		{ 222, 388, 33 },
		{ 222, 389, 33 },
		{ 222, 390, 33 },
		{ 222, 391, 33 },
		{ 222, 392, 33 },
		{ 222, 393, 33 },
		{ 222, 394, 33 },
		{ 222, 395, 33 },
		{ 222, 396, 33 },
		{ 223, 326, 33 },
		{ 223, 327, 33 },
		{ 223, 328, 33 },
		{ 223, 329, 33 },
		{ 223, 330, 33 },
		{ 223, 331, 33 },
		{ 223, 332, 33 },
		{ 223, 333, 33 },
		{ 223, 334, 33 },
		{ 223, 335, 33 },
		{ 223, 336, 33 },
		{ 223, 337, 33 },
		{ 223, 338, 33 },
		{ 223, 339, 33 },
		{ 223, 340, 33 },
		{ 223, 341, 33 },
		{ 223, 342, 33 },
		{ 223, 343, 33 },
		{ 223, 344, 33 },
		{ 223, 345, 33 },
		{ 223, 346, 33 },
		{ 223, 347, 33 },
		{ 223, 348, 33 },
		{ 223, 349, 33 },
		{ 223, 350, 33 },
		{ 223, 351, 33 },
		{ 223, 352, 33 },
		{ 223, 353, 33 },
		{ 223, 354, 33 },
		{ 223, 355, 33 },
		{ 223, 356, 33 },
		{ 223, 357, 33 },
		{ 223, 358, 33 },
		{ 223, 359, 33 },
		{ 223, 360, 33 },
		{ 223, 361, 33 },
		{ 223, 362, 33 },
		{ 223, 363, 33 },
		{ 223, 364, 33 },
		{ 223, 365, 33 },
		{ 223, 366, 33 },
		{ 223, 367, 33 },
		{ 223, 368, 33 },
		{ 223, 369, 33 },
		{ 223, 370, 33 },
		{ 223, 371, 33 },
		{ 223, 372, 33 },
		{ 223, 373, 33 },
		{ 223, 374, 33 },
		{ 223, 375, 33 },
		{ 223, 376, 33 },
		{ 223, 377, 33 },
		{ 223, 378, 33 },
		{ 223, 379, 33 },
		{ 223, 380, 33 },
		{ 223, 381, 33 },
		{ 223, 382, 33 },
		{ 223, 383, 33 },
		{ 223, 384, 33 },
		{ 223, 385, 33 },
		{ 223, 386, 33 },
		{ 223, 387, 33 },
		{ 223, 388, 33 },
		{ 223, 389, 33 },
		{ 223, 390, 33 },
		{ 223, 391, 33 },
		{ 223, 392, 33 },
		{ 224, 325, 33 },
		{ 224, 326, 33 },
		{ 224, 327, 33 },
		{ 224, 328, 33 },
		{ 224, 329, 33 },
		{ 224, 330, 33 },
		{ 224, 331, 33 },
		{ 224, 332, 33 },
		{ 224, 333, 33 },
		{ 224, 334, 33 },
		{ 224, 335, 33 },
		{ 224, 336, 33 },
		{ 224, 337, 33 },
		{ 224, 338, 33 },
		{ 224, 339, 33 },
		{ 224, 340, 33 },
		{ 224, 341, 33 },
		{ 224, 342, 33 },
		{ 224, 343, 33 },
		{ 224, 344, 33 },
		{ 224, 345, 33 },
		{ 224, 346, 33 },
		{ 224, 347, 33 },
		{ 224, 348, 33 },
		{ 224, 349, 33 },
		{ 224, 350, 33 },
		{ 224, 351, 33 },
		{ 224, 352, 33 },
		{ 224, 353, 33 },
		{ 224, 354, 33 },
		{ 224, 355, 33 },
		{ 224, 356, 33 },
		{ 224, 357, 33 },
		{ 224, 358, 33 },
		{ 224, 359, 33 },
		{ 224, 360, 33 },
		{ 224, 361, 33 },
		{ 224, 362, 33 },
		{ 224, 363, 33 },
		{ 224, 364, 33 },
		{ 224, 365, 33 },
		{ 224, 366, 33 },
		{ 224, 367, 33 },
		{ 224, 368, 33 },
		{ 224, 369, 33 },
		{ 224, 370, 33 },
		{ 224, 371, 33 },
		{ 224, 372, 33 },
		{ 224, 373, 33 },
		{ 224, 374, 33 },
		{ 224, 375, 33 },
		{ 224, 376, 33 },
		{ 224, 377, 33 },
		{ 224, 378, 33 },
		{ 224, 379, 33 },
		{ 224, 380, 33 },
		{ 224, 381, 33 },
		{ 224, 382, 33 },
		{ 224, 383, 33 },
		{ 224, 384, 33 },
		{ 224, 385, 33 },
		{ 224, 386, 33 },
		{ 224, 387, 33 },
		{ 224, 388, 33 },
		{ 225, 324, 33 },
		{ 225, 325, 33 },
		{ 225, 326, 33 },
		{ 225, 327, 33 },
		{ 225, 328, 33 },
		{ 225, 329, 33 },
		{ 225, 330, 33 },
		{ 225, 331, 33 },
		{ 225, 332, 33 },
		{ 225, 333, 33 },
		{ 225, 334, 33 },
		{ 225, 335, 33 },
		{ 225, 336, 33 },
		{ 225, 337, 33 },
		{ 225, 338, 33 },
		{ 225, 339, 33 },
		{ 225, 340, 33 },
		{ 225, 341, 33 },
		{ 225, 342, 33 },
		{ 225, 343, 33 },
		{ 225, 344, 33 },
		{ 225, 345, 33 },
		{ 225, 346, 33 },
		{ 225, 347, 33 },
		{ 225, 348, 33 },
		{ 225, 349, 33 },
		{ 225, 350, 33 },
		{ 225, 351, 33 },
		{ 225, 352, 33 },
		{ 225, 353, 33 },
		{ 225, 354, 33 },
		{ 225, 355, 33 },
		{ 225, 356, 33 },
		{ 225, 357, 33 },
		{ 225, 358, 33 },
		{ 225, 359, 33 },
		{ 225, 360, 33 },
		{ 225, 361, 33 },
		{ 225, 362, 33 },
		{ 225, 363, 33 },
		{ 225, 364, 33 },
		{ 225, 365, 33 },
		{ 225, 366, 33 },
		{ 225, 367, 33 },
		{ 225, 368, 33 },
		{ 225, 369, 33 },
		{ 225, 370, 33 },
		{ 225, 371, 33 },
		{ 225, 372, 33 },
		{ 225, 373, 33 },
		{ 225, 374, 33 },
		{ 225, 375, 33 },
		{ 225, 376, 33 },
		{ 225, 377, 33 },
		{ 225, 378, 33 },
		{ 225, 379, 33 },
		{ 225, 380, 33 },
		{ 225, 381, 33 },
		{ 225, 382, 33 },
		{ 225, 383, 33 },
		{ 225, 384, 33 },
		{ 226, 323, 33 },
		{ 226, 324, 33 },
		{ 226, 325, 33 },
		{ 226, 326, 33 },
		{ 226, 327, 33 },
		{ 226, 328, 33 },
		{ 226, 329, 33 },
		{ 226, 330, 33 },
		{ 226, 331, 33 },
		{ 226, 332, 33 },
		{ 226, 333, 33 },
		{ 226, 334, 33 },
		{ 226, 335, 33 },
		{ 226, 336, 33 },
		{ 226, 337, 33 },
		{ 226, 338, 33 },
		{ 226, 339, 33 },
		{ 226, 340, 33 },
		{ 226, 341, 33 },
		{ 226, 342, 33 },
		{ 226, 343, 33 },
		{ 226, 344, 33 },
		{ 226, 345, 33 },
		{ 226, 346, 33 },
		{ 226, 347, 33 },
		{ 226, 348, 33 },
		{ 226, 349, 33 },
		{ 226, 350, 33 },
		{ 226, 351, 33 },
		{ 226, 352, 33 },
		{ 226, 353, 33 },
		{ 226, 354, 33 },
		{ 226, 355, 33 },
		{ 226, 356, 33 },
		{ 226, 357, 33 },
		{ 226, 358, 33 },
		{ 226, 359, 33 },
		{ 226, 360, 33 },
		{ 226, 361, 33 },
		{ 226, 362, 33 },
		{ 226, 363, 33 },
		{ 226, 364, 33 },
		{ 226, 365, 33 },
		{ 226, 366, 33 },
		{ 226, 367, 33 },
		{ 226, 368, 33 },
		{ 226, 369, 33 },
		{ 226, 370, 33 },
		{ 226, 371, 33 },
		{ 226, 372, 33 },
		{ 226, 373, 33 },
		{ 226, 374, 33 },
		{ 226, 375, 33 },
		{ 226, 376, 33 },
		{ 226, 377, 33 },
		{ 226, 378, 33 },
		{ 226, 379, 33 },
		{ 226, 380, 33 },
		{ 227, 323, 33 },
		{ 227, 324, 33 },
		{ 227, 325, 33 },
		{ 227, 326, 33 },
		{ 227, 327, 33 },
		{ 227, 328, 33 },
		{ 227, 329, 33 },
		{ 227, 330, 33 },
		{ 227, 331, 33 },
		{ 227, 332, 33 },
		{ 227, 333, 33 },
		{ 227, 334, 33 },
		{ 227, 335, 33 },
		{ 227, 336, 33 },
		{ 227, 337, 33 },
		{ 227, 338, 33 },
		{ 227, 339, 33 },
		{ 227, 340, 33 },
		{ 227, 341, 33 },
		{ 227, 342, 33 },
		{ 227, 343, 33 },
		{ 227, 344, 33 },
		{ 227, 345, 33 },
		{ 227, 346, 33 },
		{ 227, 347, 33 },
		{ 227, 348, 33 },
		{ 227, 349, 33 },
		{ 227, 350, 33 },
		{ 227, 351, 33 },
		{ 227, 352, 33 },
		{ 227, 353, 33 },
		{ 227, 354, 33 },
		{ 227, 355, 33 },
		{ 227, 356, 33 },
		{ 227, 357, 33 },
		{ 227, 358, 33 },
		{ 227, 359, 33 },
		{ 227, 360, 33 },
		{ 227, 361, 33 },
		{ 227, 362, 33 },
		{ 227, 363, 33 },
		{ 227, 364, 33 },
		{ 227, 365, 33 },
		{ 227, 366, 33 },
		{ 227, 367, 33 },
		{ 227, 368, 33 },
		{ 227, 369, 33 },
		{ 227, 370, 33 },
		{ 227, 371, 33 },
		{ 227, 372, 33 },
		{ 227, 373, 33 },
		{ 227, 374, 33 },
		{ 227, 375, 33 },
		{ 227, 376, 33 },
		{ 228, 323, 33 },
		{ 228, 324, 33 },
		{ 228, 325, 33 },
		{ 228, 326, 33 },
		{ 228, 327, 33 },
		{ 228, 328, 33 },
		{ 228, 329, 33 },
		{ 228, 330, 33 },
		{ 228, 331, 33 },
		{ 228, 332, 33 },
		{ 228, 333, 33 },
		{ 228, 334, 33 },
		{ 228, 335, 33 },
		{ 228, 336, 33 },
		{ 228, 337, 33 },
		{ 228, 338, 33 },
		{ 228, 339, 33 },
		{ 228, 340, 33 },
		{ 228, 341, 33 },
		{ 228, 342, 33 },
		{ 228, 343, 33 },
		{ 228, 344, 33 },
		{ 228, 345, 33 },
		{ 228, 346, 33 },
		{ 228, 347, 33 },
		{ 228, 348, 33 },
		{ 228, 349, 33 },
		{ 228, 350, 33 },
		{ 228, 351, 33 },
		{ 228, 352, 33 },
		{ 228, 353, 33 },
		{ 228, 354, 33 },
		{ 228, 355, 33 },
		{ 228, 356, 33 },
		{ 228, 357, 33 },
		{ 228, 358, 33 },
		{ 228, 359, 33 },
		{ 228, 360, 33 },
		{ 228, 361, 33 },
		{ 228, 362, 33 },
		{ 228, 363, 33 },
		{ 228, 364, 33 },
		{ 228, 365, 33 },
		{ 228, 366, 33 },
		{ 228, 367, 33 },
		{ 228, 368, 33 },
		{ 228, 369, 33 },
		{ 228, 370, 33 },
		{ 228, 371, 33 },
		{ 228, 372, 33 },
		{ 229, 323, 33 },
		{ 229, 324, 33 },
		{ 229, 325, 33 },
		{ 229, 326, 33 },
		{ 229, 327, 33 },
		{ 229, 328, 33 },
		{ 229, 329, 33 },
		{ 229, 330, 33 },
		{ 229, 331, 33 },
		{ 229, 332, 33 },
		{ 229, 333, 33 },
		{ 229, 334, 33 },
		{ 229, 335, 33 },
		{ 229, 336, 33 },
		{ 229, 337, 33 },
		{ 229, 338, 33 },
		{ 229, 339, 33 },
		{ 229, 340, 33 },
		{ 229, 341, 33 },
		{ 229, 342, 33 },
		{ 229, 343, 33 },
		{ 229, 344, 33 },
		{ 229, 345, 33 },
		{ 229, 346, 33 },
		{ 229, 347, 33 },
		{ 229, 348, 33 },
		{ 229, 349, 33 },
		{ 229, 350, 33 },
		{ 229, 351, 33 },
		{ 229, 352, 33 },
		{ 229, 353, 33 },
		{ 229, 354, 33 },
		{ 229, 355, 33 },
		{ 229, 356, 33 },
		{ 229, 357, 33 },
		{ 229, 358, 33 },
		{ 229, 359, 33 },
		{ 229, 360, 33 },
		{ 229, 361, 33 },
		{ 229, 362, 33 },
		{ 229, 363, 33 },
		{ 229, 364, 33 },
		{ 229, 365, 33 },
		{ 229, 366, 33 },
		{ 229, 367, 33 },
		{ 229, 368, 33 },
		{ 230, 323, 33 },
		{ 230, 324, 33 },
		{ 230, 325, 33 },
		{ 230, 326, 33 },
		{ 230, 327, 33 },
		{ 230, 328, 33 },
		{ 230, 329, 33 },
		{ 230, 330, 33 },
		{ 230, 331, 33 },
		{ 230, 332, 33 },
		{ 230, 333, 33 },
		{ 230, 334, 33 },
		{ 230, 335, 33 },
		{ 230, 336, 33 },
		{ 230, 337, 33 },
		{ 230, 338, 33 },
		{ 230, 339, 33 },
		{ 230, 340, 33 },
		{ 230, 341, 33 },
		{ 230, 342, 33 },
		{ 230, 343, 33 },
		{ 230, 344, 33 },
		{ 230, 345, 33 },
		{ 230, 346, 33 },
		{ 230, 347, 33 },
		{ 230, 348, 33 },
		{ 230, 349, 33 },
		{ 230, 350, 33 },
		{ 230, 351, 33 },
		{ 230, 352, 33 },
		{ 230, 353, 33 },
		{ 230, 354, 33 },
		{ 230, 355, 33 },
		{ 230, 356, 33 },
		{ 230, 357, 33 },
		{ 230, 358, 33 },
		{ 230, 359, 33 },
		{ 230, 360, 33 },
		{ 230, 361, 33 },
		{ 230, 362, 33 },
		{ 230, 363, 33 },
		{ 230, 364, 33 },
		{ 231, 323, 33 },
		{ 231, 324, 33 },
		{ 231, 325, 33 },
		{ 231, 326, 33 },
		{ 231, 327, 33 },
		{ 231, 328, 33 },
		{ 231, 329, 33 },
		{ 231, 330, 33 },
		{ 231, 331, 33 },
		{ 231, 332, 33 },
		{ 231, 333, 33 },
		{ 231, 334, 33 },
		{ 231, 335, 33 },
		{ 231, 336, 33 },
		{ 231, 337, 33 },
		{ 231, 338, 33 },
		{ 231, 339, 33 },
		{ 231, 340, 33 },
		{ 231, 341, 33 },
		{ 231, 342, 33 },
		{ 231, 343, 33 },
		{ 231, 344, 33 },
		{ 231, 345, 33 },
		{ 231, 346, 33 },
		{ 231, 347, 33 },
		{ 231, 348, 33 },
		{ 231, 349, 33 },
		{ 231, 350, 33 },
		{ 231, 351, 33 },
		{ 231, 352, 33 },
		{ 231, 353, 33 },
		{ 231, 354, 33 },
		{ 231, 355, 33 },
		{ 231, 356, 33 },
		{ 231, 357, 33 },
		{ 231, 358, 33 },
		{ 231, 359, 33 },
		{ 231, 360, 33 },
		{ 232, 323, 33 },
		{ 232, 324, 33 },
		{ 232, 325, 33 },
		{ 232, 326, 33 },
		{ 232, 327, 33 },
		{ 232, 328, 33 },
		{ 232, 329, 33 },
		{ 232, 330, 33 },
		{ 232, 331, 33 },
		{ 232, 332, 33 },
		{ 232, 333, 33 },
		{ 232, 334, 33 },
		{ 232, 335, 33 },
		{ 232, 336, 33 },
		{ 232, 337, 33 },
		{ 232, 338, 33 },
		{ 232, 339, 33 },
		{ 232, 340, 33 },
		{ 232, 341, 33 },
		{ 232, 342, 33 },
		{ 232, 343, 33 },
		{ 232, 344, 33 },
		{ 232, 345, 33 },
		{ 232, 346, 33 },
		{ 232, 347, 33 },
		{ 232, 348, 33 },
		{ 232, 349, 33 },
		{ 232, 350, 33 },
		{ 232, 351, 33 },
		{ 232, 352, 33 },
		{ 232, 353, 33 },
		{ 232, 354, 33 },
		{ 232, 355, 33 },
		{ 232, 356, 33 },
		{ 233, 323, 33 },
		{ 233, 324, 33 },
		{ 233, 325, 33 },
		{ 233, 326, 33 },
		{ 233, 327, 33 },
		{ 233, 328, 33 },
		{ 233, 329, 33 },
		{ 233, 330, 33 },
		{ 233, 331, 33 },
		{ 233, 332, 33 },
		{ 233, 333, 33 },
		{ 233, 334, 33 },
		{ 233, 335, 33 },
		{ 233, 336, 33 },
		{ 233, 337, 33 },
		{ 233, 338, 33 },
		{ 233, 339, 33 },
		{ 233, 340, 33 },
		{ 233, 341, 33 },
		{ 233, 342, 33 },
		{ 233, 343, 33 },
		{ 233, 344, 33 },
		{ 233, 345, 33 },
		{ 233, 346, 33 },
		{ 233, 347, 33 },
		{ 233, 348, 33 },
		{ 233, 349, 33 },
		{ 233, 350, 33 },
		{ 233, 351, 33 },
		{ 233, 352, 33 },
		{ 234, 323, 33 },
		{ 234, 324, 33 },
		{ 234, 325, 33 },
		{ 234, 326, 33 },
		{ 234, 327, 33 },
		{ 234, 328, 33 },
		{ 234, 329, 33 },
		{ 234, 330, 33 },
		{ 234, 331, 33 },
		{ 234, 332, 33 },
		{ 234, 333, 33 },
		{ 234, 334, 33 },
		{ 234, 335, 33 },
		{ 234, 336, 33 },
		{ 234, 337, 33 },
		{ 234, 338, 33 },
		{ 234, 339, 33 },
		{ 234, 340, 33 },
		{ 234, 341, 33 },
		{ 234, 342, 33 },
		{ 234, 343, 33 },
		{ 234, 344, 33 },
		{ 234, 345, 33 },
		{ 234, 346, 33 },
		{ 234, 347, 33 },
		{ 234, 348, 33 },
		{ 235, 323, 33 },
		{ 235, 324, 33 },
		{ 235, 325, 33 },
		{ 235, 326, 33 },
		{ 235, 327, 33 },
		{ 235, 328, 33 },
		{ 235, 329, 33 },
		{ 235, 330, 33 },
		{ 235, 331, 33 },
		{ 235, 332, 33 },
		{ 235, 333, 33 },
		{ 235, 334, 33 },
		{ 235, 335, 33 },
		{ 235, 336, 33 },
		{ 235, 337, 33 },
		{ 235, 338, 33 },
		{ 235, 339, 33 },
		{ 235, 340, 33 },
		{ 235, 341, 33 },
		{ 235, 342, 33 },
		{ 235, 343, 33 },
		{ 235, 344, 33 },
		{ 236, 323, 33 },
		{ 236, 324, 33 },
		{ 236, 325, 33 },
		{ 236, 326, 33 },
		{ 236, 327, 33 },
		{ 236, 328, 33 },
		{ 236, 329, 33 },
		{ 236, 330, 33 },
		{ 236, 331, 33 },
		{ 236, 332, 33 },
		{ 236, 333, 33 },
		{ 236, 334, 33 },
		{ 236, 335, 33 },
		{ 236, 336, 33 },
		{ 236, 337, 33 },
		{ 236, 338, 33 },
		{ 236, 339, 33 },
		{ 236, 340, 33 },
		{ 237, 323, 33 },
		{ 237, 324, 33 },
		{ 237, 325, 33 },
		{ 237, 326, 33 },
		{ 237, 327, 33 },
		{ 237, 328, 33 },
		{ 237, 329, 33 },
		{ 237, 330, 33 },
		{ 237, 331, 33 },
		{ 237, 332, 33 },
		{ 237, 333, 33 },
		{ 237, 334, 33 },
		{ 237, 335, 33 },
		{ 237, 336, 33 },
		{ 238, 323, 33 },
		{ 238, 324, 33 },
		{ 238, 325, 33 },
		{ 238, 326, 33 },
		{ 238, 327, 33 },
		{ 238, 328, 33 },
		{ 238, 329, 33 },
		{ 238, 330, 33 },
		{ 238, 331, 33 },
		{ 238, 332, 33 },
		{ 239, 324, 33 },
		{ 239, 325, 33 },
		{ 239, 326, 33 },
		{ 239, 327, 33 },
		{ 239, 328, 33 },
		{ 275, 186, 33 },
		{ 275, 187, 33 },
		{ 275, 188, 33 },
		{ 275, 189, 33 },
		{ 275, 190, 33 },
		{ 276, 182, 33 },
		{ 276, 183, 33 },
		{ 276, 184, 33 },
		{ 276, 185, 33 },
		{ 276, 186, 33 },
		{ 276, 187, 33 },
		{ 276, 188, 33 },
		{ 276, 189, 33 },
		{ 276, 190, 33 },
		{ 276, 191, 33 },
		{ 277, 178, 33 },
		{ 277, 179, 33 },
		{ 277, 180, 33 },
		{ 277, 181, 33 },
		{ 277, 182, 33 },
		{ 277, 183, 33 },
		{ 277, 184, 33 },
		{ 277, 185, 33 },
		{ 277, 186, 33 },
		{ 277, 187, 33 },
		{ 277, 188, 33 },
		{ 277, 189, 33 },
		{ 277, 190, 33 },
		{ 277, 191, 33 },
		{ 278, 174, 33 },
		{ 278, 175, 33 },
		{ 278, 176, 33 },
		{ 278, 177, 33 },
		{ 278, 178, 33 },
		{ 278, 179, 33 },
		{ 278, 180, 33 },
		{ 278, 181, 33 },
		{ 278, 182, 33 },
		{ 278, 183, 33 },
		{ 278, 184, 33 },
		{ 278, 185, 33 },
		{ 278, 186, 33 },
		{ 278, 187, 33 },
		{ 278, 188, 33 },
		{ 278, 189, 33 },
		{ 278, 190, 33 },
		{ 278, 191, 33 },
		{ 279, 170, 33 },
		{ 279, 171, 33 },
		{ 279, 172, 33 },
		{ 279, 173, 33 },
		{ 279, 174, 33 },
		{ 279, 175, 33 },
		{ 279, 176, 33 },
		{ 279, 177, 33 },
		{ 279, 178, 33 },
		{ 279, 179, 33 },
		{ 279, 180, 33 },
		{ 279, 181, 33 },
		{ 279, 182, 33 },
		{ 279, 183, 33 },
		{ 279, 184, 33 },
		{ 279, 185, 33 },
		{ 279, 186, 33 },
		{ 279, 187, 33 },
		{ 279, 188, 33 },
		{ 279, 189, 33 },
		{ 279, 190, 33 },
		{ 279, 191, 33 },
		{ 280, 166, 33 },
		{ 280, 167, 33 },
		{ 280, 168, 33 },
		{ 280, 169, 33 },
		{ 280, 170, 33 },
		{ 280, 171, 33 },
		{ 280, 172, 33 },
		{ 280, 173, 33 },
		{ 280, 174, 33 },
		{ 280, 175, 33 },
		{ 280, 176, 33 },
		{ 280, 177, 33 },
		{ 280, 178, 33 },
		{ 280, 179, 33 },
		{ 280, 180, 33 },
		{ 280, 181, 33 },
		{ 280, 182, 33 },
		{ 280, 183, 33 },
		{ 280, 184, 33 },
		{ 280, 185, 33 },
		{ 280, 186, 33 },
		{ 280, 187, 33 },
		{ 280, 188, 33 },
		{ 280, 189, 33 },
		{ 280, 190, 33 },
		{ 280, 191, 33 },
		{ 281, 162, 33 },
		{ 281, 163, 33 },
		{ 281, 164, 33 },
		{ 281, 165, 33 },
		{ 281, 166, 33 },
		{ 281, 167, 33 },
		{ 281, 168, 33 },
		{ 281, 169, 33 },
		{ 281, 170, 33 },
		{ 281, 171, 33 },
		{ 281, 172, 33 },
		{ 281, 173, 33 },
		{ 281, 174, 33 },
		{ 281, 175, 33 },
		{ 281, 176, 33 },
		{ 281, 177, 33 },
		{ 281, 178, 33 },
		{ 281, 179, 33 },
		{ 281, 180, 33 },
		{ 281, 181, 33 },
		{ 281, 182, 33 },
		{ 281, 183, 33 },
		{ 281, 184, 33 },
		{ 281, 185, 33 },
		{ 281, 186, 33 },
		{ 281, 187, 33 },
		{ 281, 188, 33 },
		{ 281, 189, 33 },
		{ 281, 190, 33 },
		{ 281, 191, 33 },
		{ 282, 158, 33 },
		{ 282, 159, 33 },
		{ 282, 160, 33 },
		{ 282, 161, 33 },
		{ 282, 162, 33 },
		{ 282, 163, 33 },
		{ 282, 164, 33 },
		{ 282, 165, 33 },
		{ 282, 166, 33 },
		{ 282, 167, 33 },
		{ 282, 168, 33 },
		{ 282, 169, 33 },
		{ 282, 170, 33 },
		{ 282, 171, 33 },
		{ 282, 172, 33 },
		{ 282, 173, 33 },
		{ 282, 174, 33 },
		{ 282, 175, 33 },
		{ 282, 176, 33 },
		{ 282, 177, 33 },
		{ 282, 178, 33 },
		{ 282, 179, 33 },
		{ 282, 180, 33 },
		{ 282, 181, 33 },
		{ 282, 182, 33 },
		{ 282, 183, 33 },
		{ 282, 184, 33 },
		{ 282, 185, 33 },
		{ 282, 186, 33 },
		{ 282, 187, 33 },
		{ 282, 188, 33 },
		{ 282, 189, 33 },
		{ 282, 190, 33 },
		{ 282, 191, 33 },
		{ 283, 154, 33 },
		{ 283, 155, 33 },
		{ 283, 156, 33 },
		{ 283, 157, 33 },
		{ 283, 158, 33 },
		{ 283, 159, 33 },
		{ 283, 160, 33 },
		{ 283, 161, 33 },
		{ 283, 162, 33 },
		{ 283, 163, 33 },
		{ 283, 164, 33 },
		{ 283, 165, 33 },
		{ 283, 166, 33 },
		{ 283, 167, 33 },
		{ 283, 168, 33 },
		{ 283, 169, 33 },
		{ 283, 170, 33 },
		{ 283, 171, 33 },
		{ 283, 172, 33 },
		{ 283, 173, 33 },
		{ 283, 174, 33 },
		{ 283, 175, 33 },
		{ 283, 176, 33 },
		{ 283, 177, 33 },
		{ 283, 178, 33 },
		{ 283, 179, 33 },
		{ 283, 180, 33 },
		{ 283, 181, 33 },
		{ 283, 182, 33 },
		{ 283, 183, 33 },
		{ 283, 184, 33 },
		{ 283, 185, 33 },
		{ 283, 186, 33 },
		{ 283, 187, 33 },
		{ 283, 188, 33 },
		{ 283, 189, 33 },
		{ 283, 190, 33 },
		{ 283, 191, 33 },
		{ 284, 150, 33 },
		{ 284, 151, 33 },
		{ 284, 152, 33 },
		{ 284, 153, 33 },
		{ 284, 154, 33 },
		{ 284, 155, 33 },
		{ 284, 156, 33 },
		{ 284, 157, 33 },
		{ 284, 158, 33 },
		{ 284, 159, 33 },
		{ 284, 160, 33 },
		{ 284, 161, 33 },
		{ 284, 162, 33 },
		{ 284, 163, 33 },
		{ 284, 164, 33 },
		{ 284, 165, 33 },
		{ 284, 166, 33 },
		{ 284, 167, 33 },
		{ 284, 168, 33 },
		{ 284, 169, 33 },
		{ 284, 170, 33 },
		{ 284, 171, 33 },
		{ 284, 172, 33 },
		{ 284, 173, 33 },
		{ 284, 174, 33 },
		{ 284, 175, 33 },
		{ 284, 176, 33 },
		{ 284, 177, 33 },
		{ 284, 178, 33 },
		{ 284, 179, 33 },
		{ 284, 180, 33 },
		{ 284, 181, 33 },
		{ 284, 182, 33 },
		{ 284, 183, 33 },
		{ 284, 184, 33 },
		{ 284, 185, 33 },
		{ 284, 186, 33 },
		{ 284, 187, 33 },
		{ 284, 188, 33 },
		{ 284, 189, 33 },
		{ 284, 190, 33 },
		{ 284, 191, 33 },
		{ 285, 146, 33 },
		{ 285, 147, 33 },
		{ 285, 148, 33 },
		{ 285, 149, 33 },
		{ 285, 150, 33 },
		{ 285, 151, 33 },
		{ 285, 152, 33 },
		{ 285, 153, 33 },
		{ 285, 154, 33 },
		{ 285, 155, 33 },
		{ 285, 156, 33 },
		{ 285, 157, 33 },
		{ 285, 158, 33 },
		{ 285, 159, 33 },
		{ 285, 160, 33 },
		{ 285, 161, 33 },
		{ 285, 162, 33 },
		{ 285, 163, 33 },
		{ 285, 164, 33 },
		{ 285, 165, 33 },
		{ 285, 166, 33 },
		{ 285, 167, 33 },
		{ 285, 168, 33 },
		{ 285, 169, 33 },
		{ 285, 170, 33 },
		{ 285, 171, 33 },
		{ 285, 172, 33 },
		{ 285, 173, 33 },
		{ 285, 174, 33 },
		{ 285, 175, 33 },
		{ 285, 176, 33 },
		{ 285, 177, 33 },
		{ 285, 178, 33 },
		{ 285, 179, 33 },
		{ 285, 180, 33 },
		{ 285, 181, 33 },
		{ 285, 182, 33 },
		{ 285, 183, 33 },
		{ 285, 184, 33 },
		{ 285, 185, 33 },
		{ 285, 186, 33 },
		{ 285, 187, 33 },
		{ 285, 188, 33 },
		{ 285, 189, 33 },
		{ 285, 190, 33 },
		{ 285, 191, 33 },
		{ 286, 142, 33 },
		{ 286, 143, 33 },
		{ 286, 144, 33 },
		{ 286, 145, 33 },
		{ 286, 146, 33 },
		{ 286, 147, 33 },
		{ 286, 148, 33 },
		{ 286, 149, 33 },
		{ 286, 150, 33 },
		{ 286, 151, 33 },
		{ 286, 152, 33 },
		{ 286, 153, 33 },
		{ 286, 154, 33 },
		{ 286, 155, 33 },
		{ 286, 156, 33 },
		{ 286, 157, 33 },
		{ 286, 158, 33 },
		{ 286, 159, 33 },
		{ 286, 160, 33 },
		{ 286, 161, 33 },
		{ 286, 162, 33 },
		{ 286, 163, 33 },
		{ 286, 164, 33 },
		{ 286, 165, 33 },
		{ 286, 166, 33 },
		{ 286, 167, 33 },
		{ 286, 168, 33 },
		{ 286, 169, 33 },
		{ 286, 170, 33 },
		{ 286, 171, 33 },
		{ 286, 172, 33 },
		{ 286, 173, 33 },
		{ 286, 174, 33 },
		{ 286, 175, 33 },
		{ 286, 176, 33 },
		{ 286, 177, 33 },
		{ 286, 178, 33 },
		{ 286, 179, 33 },
		{ 286, 180, 33 },
		{ 286, 181, 33 },
		{ 286, 182, 33 },
		{ 286, 183, 33 },
		{ 286, 184, 33 },
		{ 286, 185, 33 },
		{ 286, 186, 33 },
		{ 286, 187, 33 },
		{ 286, 188, 33 },
		{ 286, 189, 33 },
		{ 286, 190, 33 },
		{ 286, 191, 33 },
		{ 287, 138, 33 },
		{ 287, 139, 33 },
		{ 287, 140, 33 },
		{ 287, 141, 33 },
		{ 287, 142, 33 },
		{ 287, 143, 33 },
		{ 287, 144, 33 },
		{ 287, 145, 33 },
		{ 287, 146, 33 },
		{ 287, 147, 33 },
		{ 287, 148, 33 },
		{ 287, 149, 33 },
		{ 287, 150, 33 },
		{ 287, 151, 33 },
		{ 287, 152, 33 },
		{ 287, 153, 33 },
		{ 287, 154, 33 },
		{ 287, 155, 33 },
		{ 287, 156, 33 },
		{ 287, 157, 33 },
		{ 287, 158, 33 },
		{ 287, 159, 33 },
		{ 287, 160, 33 },
		{ 287, 161, 33 },
		{ 287, 162, 33 },
		{ 287, 163, 33 },
		{ 287, 164, 33 },
		{ 287, 165, 33 },
		{ 287, 166, 33 },
		{ 287, 167, 33 },
		{ 287, 168, 33 },
		{ 287, 169, 33 },
		{ 287, 170, 33 },
		{ 287, 171, 33 },
		{ 287, 172, 33 },
		{ 287, 173, 33 },
		{ 287, 174, 33 },
		{ 287, 175, 33 },
		{ 287, 176, 33 },
		{ 287, 177, 33 },
		{ 287, 178, 33 },
		{ 287, 179, 33 },
		{ 287, 180, 33 },
		{ 287, 181, 33 },
		{ 287, 182, 33 },
		{ 287, 183, 33 },
		{ 287, 184, 33 },
		{ 287, 185, 33 },
		{ 287, 186, 33 },
		{ 287, 187, 33 },
		{ 287, 188, 33 },
		{ 287, 189, 33 },
		{ 287, 190, 33 },
		{ 287, 191, 33 },
		{ 288, 134, 33 },
		{ 288, 135, 33 },
		{ 288, 136, 33 },
		{ 288, 137, 33 },
		{ 288, 138, 33 },
		{ 288, 139, 33 },
		{ 288, 140, 33 },
		{ 288, 141, 33 },
		{ 288, 142, 33 },
		{ 288, 143, 33 },
		{ 288, 144, 33 },
		{ 288, 145, 33 },
		{ 288, 146, 33 },
		{ 288, 147, 33 },
		{ 288, 148, 33 },
		{ 288, 149, 33 },
		{ 288, 150, 33 },
		{ 288, 151, 33 },
		{ 288, 152, 33 },
		{ 288, 153, 33 },
		{ 288, 154, 33 },
		{ 288, 155, 33 },
		{ 288, 156, 33 },
		{ 288, 157, 33 },
		{ 288, 158, 33 },
		{ 288, 159, 33 },
		{ 288, 160, 33 },
		{ 288, 161, 33 },
		{ 288, 162, 33 },
		{ 288, 163, 33 },
		{ 288, 164, 33 },
		{ 288, 165, 33 },
		{ 288, 166, 33 },
		{ 288, 167, 33 },
		{ 288, 168, 33 },
		{ 288, 169, 33 },
		{ 288, 170, 33 },
		{ 288, 171, 33 },
		{ 288, 172, 33 },
		{ 288, 173, 33 },
		{ 288, 174, 33 },
		{ 288, 175, 33 },
		{ 288, 176, 33 },
		{ 288, 177, 33 },
		{ 288, 178, 33 },
		{ 288, 179, 33 },
		{ 288, 180, 33 },
		{ 288, 181, 33 },
		{ 288, 182, 33 },
		{ 288, 183, 33 },
		{ 288, 184, 33 },
		{ 288, 185, 33 },
		{ 288, 186, 33 },
		{ 288, 187, 33 },
		{ 288, 188, 33 },
		{ 288, 189, 33 },
		{ 288, 190, 33 },
		{ 288, 191, 33 },
		{ 289, 130, 33 },
		{ 289, 131, 33 },
		{ 289, 132, 33 },
		{ 289, 133, 33 },
		{ 289, 134, 33 },
		{ 289, 135, 33 },
		{ 289, 136, 33 },
		{ 289, 137, 33 },
		{ 289, 138, 33 },
		{ 289, 139, 33 },
		{ 289, 140, 33 },
		{ 289, 141, 33 },
		{ 289, 142, 33 },
		{ 289, 143, 33 },
		{ 289, 144, 33 },
		{ 289, 145, 33 },
		{ 289, 146, 33 },
		{ 289, 147, 33 },
		{ 289, 148, 33 },
		{ 289, 149, 33 },
		{ 289, 150, 33 },
		{ 289, 151, 33 },
		{ 289, 152, 33 },
		{ 289, 153, 33 },
		{ 289, 154, 33 },
		{ 289, 155, 33 },
		{ 289, 156, 33 },
		{ 289, 157, 33 },
		{ 289, 158, 33 },
		{ 289, 159, 33 },
		{ 289, 160, 33 },
		{ 289, 161, 33 },
		{ 289, 162, 33 },
		{ 289, 163, 33 },
		{ 289, 164, 33 },
		{ 289, 165, 33 },
		{ 289, 166, 33 },
		{ 289, 167, 33 },
		{ 289, 168, 33 },
		{ 289, 169, 33 },
		{ 289, 170, 33 },
		{ 289, 171, 33 },
		{ 289, 172, 33 },
		{ 289, 173, 33 },
		{ 289, 174, 33 },
		{ 289, 175, 33 },
		{ 289, 176, 33 },
		{ 289, 177, 33 },
		{ 289, 178, 33 },
		{ 289, 179, 33 },
		{ 289, 180, 33 },
		{ 289, 181, 33 },
		{ 289, 182, 33 },
		{ 289, 183, 33 },
		{ 289, 184, 33 },
		{ 289, 185, 33 },
		{ 289, 186, 33 },
		{ 289, 187, 33 },
		{ 289, 188, 33 },
		{ 289, 189, 33 },
		{ 289, 190, 33 },
		{ 290, 126, 33 },
		{ 290, 127, 33 },
		{ 290, 128, 33 },
		{ 290, 129, 33 },
		{ 290, 130, 33 },
		{ 290, 131, 33 },
		{ 290, 132, 33 },
		{ 290, 133, 33 },
		{ 290, 134, 33 },
		{ 290, 135, 33 },
		{ 290, 136, 33 },
		{ 290, 137, 33 },
		{ 290, 138, 33 },
		{ 290, 139, 33 },
		{ 290, 140, 33 },
		{ 290, 141, 33 },
		{ 290, 142, 33 },
		{ 290, 143, 33 },
		{ 290, 144, 33 },
		{ 290, 145, 33 },
		{ 290, 146, 33 },
		{ 290, 147, 33 },
		{ 290, 148, 33 },
		{ 290, 149, 33 },
		{ 290, 150, 33 },
		{ 290, 151, 33 },
		{ 290, 152, 33 },
		{ 290, 153, 33 },
		{ 290, 154, 33 },
		{ 290, 155, 33 },
		{ 290, 156, 33 },
		{ 290, 157, 33 },
		{ 290, 158, 33 },
		{ 290, 159, 33 },
		{ 290, 160, 33 },
		{ 290, 161, 33 },
		{ 290, 162, 33 },
		{ 290, 163, 33 },
		{ 290, 164, 33 },
		{ 290, 165, 33 },
		{ 290, 166, 33 },
		{ 290, 167, 33 },
		{ 290, 168, 33 },
		{ 290, 169, 33 },
		{ 290, 170, 33 },
		{ 290, 171, 33 },
		{ 290, 172, 33 },
		{ 290, 173, 33 },
		{ 290, 174, 33 },
		{ 290, 175, 33 },
		{ 290, 176, 33 },
		{ 290, 177, 33 },
		{ 290, 178, 33 },
		{ 290, 179, 33 },
		{ 290, 180, 33 },
		{ 290, 181, 33 },
		{ 290, 182, 33 },
		{ 290, 183, 33 },
		{ 290, 184, 33 },
		{ 290, 185, 33 },
		{ 290, 186, 33 },
		{ 290, 187, 33 },
		{ 290, 188, 33 },
		{ 290, 189, 33 },
		{ 291, 122, 33 },
		{ 291, 123, 33 },
		{ 291, 124, 33 },
		{ 291, 125, 33 },
		{ 291, 126, 33 },
		{ 291, 127, 33 },
		{ 291, 128, 33 },
		{ 291, 129, 33 },
		{ 291, 130, 33 },
		{ 291, 131, 33 },
		{ 291, 132, 33 },
		{ 291, 133, 33 },
		{ 291, 134, 33 },
		{ 291, 135, 33 },
		{ 291, 136, 33 },
		{ 291, 137, 33 },
		{ 291, 138, 33 },
		{ 291, 139, 33 },
		{ 291, 140, 33 },
		{ 291, 141, 33 },
		{ 291, 142, 33 },
		{ 291, 143, 33 },
		{ 291, 144, 33 },
		{ 291, 145, 33 },
		{ 291, 146, 33 },
		{ 291, 147, 33 },
		{ 291, 148, 33 },
		{ 291, 149, 33 },
		{ 291, 150, 33 },
		{ 291, 151, 33 },
		{ 291, 152, 33 },
		{ 291, 153, 33 },
		{ 291, 154, 33 },
		{ 291, 155, 33 },
		{ 291, 156, 33 },
		{ 291, 157, 33 },
		{ 291, 158, 33 },
		{ 291, 159, 33 },
		{ 291, 160, 33 },
		{ 291, 161, 33 },
		{ 291, 162, 33 },
		{ 291, 163, 33 },
		{ 291, 164, 33 },
		{ 291, 165, 33 },
		{ 291, 166, 33 },
		{ 291, 167, 33 },
		{ 291, 168, 33 },
		{ 291, 169, 33 },
		{ 291, 170, 33 },
		{ 291, 171, 33 },
		{ 291, 172, 33 },
		{ 291, 173, 33 },
		{ 291, 174, 33 },
		{ 291, 175, 33 },
		{ 291, 176, 33 },
		{ 291, 177, 33 },
		{ 291, 178, 33 },
		{ 291, 179, 33 },
		{ 291, 180, 33 },
		{ 291, 181, 33 },
		{ 291, 182, 33 },
		{ 291, 183, 33 },
		{ 291, 184, 33 },
		{ 291, 185, 33 },
		{ 291, 186, 33 },
		{ 291, 187, 33 },
		{ 291, 188, 33 },
		{ 292, 118, 33 },
		{ 292, 119, 33 },
		{ 292, 120, 33 },
		{ 292, 121, 33 },
		{ 292, 122, 33 },
		{ 292, 123, 33 },
		{ 292, 124, 33 },
		{ 292, 125, 33 },
		{ 292, 126, 33 },
		{ 292, 127, 33 },
		{ 292, 128, 33 },
		{ 292, 129, 33 },
		{ 292, 130, 33 },
		{ 292, 131, 33 },
		{ 292, 132, 33 },
		{ 292, 133, 33 },
		{ 292, 134, 33 },
		{ 292, 135, 33 },
		{ 292, 136, 33 },
		{ 292, 137, 33 },
		{ 292, 138, 33 },
		{ 292, 139, 33 },
		{ 292, 140, 33 },
		{ 292, 141, 33 },
		{ 292, 142, 33 },
		{ 292, 143, 33 },
		{ 292, 144, 33 },
		{ 292, 145, 33 },
		{ 292, 146, 33 },
		{ 292, 147, 33 },
		{ 292, 148, 33 },
		{ 292, 149, 33 },
		{ 292, 150, 33 },
		{ 292, 151, 33 },
		{ 292, 152, 33 },
		{ 292, 153, 33 },
		{ 292, 154, 33 },
		{ 292, 155, 33 },
		{ 292, 156, 33 },
		{ 292, 157, 33 },
		{ 292, 158, 33 },
		{ 292, 159, 33 },
		{ 292, 160, 33 },
		{ 292, 161, 33 },
		{ 292, 162, 33 },
		{ 292, 163, 33 },
		{ 292, 164, 33 },
		{ 292, 165, 33 },
		{ 292, 166, 33 },
		{ 292, 167, 33 },
		{ 292, 168, 33 },
		{ 292, 169, 33 },
		{ 292, 170, 33 },
		{ 292, 171, 33 },
		{ 292, 172, 33 },
		{ 292, 173, 33 },
		{ 292, 174, 33 },
		{ 292, 175, 33 },
		{ 292, 176, 33 },
		{ 292, 177, 33 },
		{ 292, 178, 33 },
		{ 292, 179, 33 },
		{ 292, 180, 33 },
		{ 292, 181, 33 },
		{ 292, 182, 33 },
		{ 292, 183, 33 },
		{ 292, 184, 33 },
		{ 292, 185, 33 },
		{ 292, 186, 33 },
		{ 293, 114, 33 },
		{ 293, 115, 33 },
		{ 293, 116, 33 },
		{ 293, 117, 33 },
		{ 293, 118, 33 },
		{ 293, 119, 33 },
		{ 293, 120, 33 },
		{ 293, 121, 33 },
		{ 293, 122, 33 },
		{ 293, 123, 33 },
		{ 293, 124, 33 },
		{ 293, 125, 33 },
		{ 293, 126, 33 },
		{ 293, 127, 33 },
		{ 293, 128, 33 },
		{ 293, 129, 33 },
		{ 293, 130, 33 },
		{ 293, 131, 33 },
		{ 293, 132, 33 },
		{ 293, 133, 33 },
		{ 293, 134, 33 },
		{ 293, 135, 33 },
		{ 293, 136, 33 },
		{ 293, 137, 33 },
		{ 293, 138, 33 },
		{ 293, 139, 33 },
		{ 293, 140, 33 },
		{ 293, 141, 33 },
		{ 293, 142, 33 },
		{ 293, 143, 33 },
		{ 293, 144, 33 },
		{ 293, 145, 33 },
		{ 293, 146, 33 },
		{ 293, 147, 33 },
		{ 293, 148, 33 },
		{ 293, 149, 33 },
		{ 293, 150, 33 },
		{ 293, 151, 33 },
		{ 293, 152, 33 },
		{ 293, 153, 33 },
		{ 293, 154, 33 },
		{ 293, 155, 33 },
		{ 293, 156, 33 },
		{ 293, 157, 33 },
		{ 293, 158, 33 },
		{ 293, 159, 33 },
		{ 293, 160, 33 },
		{ 293, 161, 33 },
		{ 293, 162, 33 },
		{ 293, 163, 33 },
		{ 293, 164, 33 },
		{ 293, 165, 33 },
		{ 293, 166, 33 },
		{ 293, 167, 33 },
		{ 293, 168, 33 },
		{ 293, 169, 33 },
		{ 293, 170, 33 },
		{ 293, 171, 33 },
		{ 293, 172, 33 },
		{ 293, 173, 33 },
		{ 293, 174, 33 },
		{ 293, 175, 33 },
		{ 293, 176, 33 },
		{ 293, 177, 33 },
		{ 293, 178, 33 },
		{ 293, 179, 33 },
		{ 293, 180, 33 },
		{ 293, 181, 33 },
		{ 293, 182, 33 },
		{ 293, 183, 33 },
		{ 293, 184, 33 },
		{ 294, 110, 33 },
		{ 294, 111, 33 },
		{ 294, 112, 33 },
		{ 294, 113, 33 },
		{ 294, 114, 33 },
		{ 294, 115, 33 },
		{ 294, 116, 33 },
		{ 294, 117, 33 },
		{ 294, 118, 33 },
		{ 294, 119, 33 },
		{ 294, 120, 33 },
		{ 294, 121, 33 },
		{ 294, 122, 33 },
		{ 294, 123, 33 },
		{ 294, 124, 33 },
		{ 294, 125, 33 },
		{ 294, 126, 33 },
		{ 294, 127, 33 },
		{ 294, 128, 33 },
		{ 294, 129, 33 },
		{ 294, 130, 33 },
		{ 294, 131, 33 },
		{ 294, 132, 33 },
		{ 294, 133, 33 },
		{ 294, 134, 33 },
		{ 294, 135, 33 },
		{ 294, 136, 33 },
		{ 294, 137, 33 },
		{ 294, 138, 33 },
		{ 294, 139, 33 },
		{ 294, 140, 33 },
		{ 294, 141, 33 },
		{ 294, 142, 33 },
		{ 294, 143, 33 },
		{ 294, 144, 33 },
		{ 294, 145, 33 },
		{ 294, 146, 33 },
		{ 294, 147, 33 },
		{ 294, 148, 33 },
		{ 294, 149, 33 },
		{ 294, 150, 33 },
		{ 294, 151, 33 },
		{ 294, 152, 33 },
		{ 294, 153, 33 },
		{ 294, 154, 33 },
		{ 294, 155, 33 },
		{ 294, 156, 33 },
		{ 294, 157, 33 },
		{ 294, 158, 33 },
		{ 294, 159, 33 },
		{ 294, 160, 33 },
		{ 294, 161, 33 },
		{ 294, 162, 33 },
		{ 294, 163, 33 },
		{ 294, 164, 33 },
		{ 294, 165, 33 },
		{ 294, 166, 33 },
		{ 294, 167, 33 },
		{ 294, 168, 33 },
		{ 294, 169, 33 },
		{ 294, 170, 33 },
		{ 294, 171, 33 },
		{ 294, 172, 33 },
		{ 294, 173, 33 },
		{ 294, 174, 33 },
		{ 294, 175, 33 },
		{ 294, 176, 33 },
		{ 294, 177, 33 },
		{ 294, 178, 33 },
		{ 294, 179, 33 },
		{ 294, 180, 33 },
		{ 294, 181, 33 },
		{ 294, 182, 33 },
		{ 295, 106, 33 },
		{ 295, 107, 33 },
		{ 295, 108, 33 },
		{ 295, 109, 33 },
		{ 295, 110, 33 },
		{ 295, 111, 33 },
		{ 295, 112, 33 },
		{ 295, 113, 33 },
		{ 295, 114, 33 },
		{ 295, 115, 33 },
		{ 295, 116, 33 },
		{ 295, 117, 33 },
		{ 295, 118, 33 },
		{ 295, 119, 33 },
		{ 295, 120, 33 },
		{ 295, 121, 33 },
		{ 295, 122, 33 },
		{ 295, 123, 33 },
		{ 295, 124, 33 },
		{ 295, 125, 33 },
		{ 295, 126, 33 },
		{ 295, 127, 33 },
		{ 295, 128, 33 },
		{ 295, 129, 33 },
		{ 295, 130, 33 },
		{ 295, 131, 33 },
		{ 295, 132, 33 },
		{ 295, 133, 33 },
		{ 295, 134, 33 },
		{ 295, 135, 33 },
		{ 295, 136, 33 },
		{ 295, 137, 33 },
		{ 295, 138, 33 },
		{ 295, 139, 33 },
		{ 295, 140, 33 },
		{ 295, 141, 33 },
		{ 295, 142, 33 },
		{ 295, 143, 33 },
		{ 295, 144, 33 },
		{ 295, 145, 33 },
		{ 295, 146, 33 },
		{ 295, 147, 33 },
		{ 295, 148, 33 },
		{ 295, 149, 33 },
		{ 295, 150, 33 },
		{ 295, 151, 33 },
		{ 295, 152, 33 },
		{ 295, 153, 33 },
		{ 295, 154, 33 },
		{ 295, 155, 33 },
		{ 295, 156, 33 },
		{ 295, 157, 33 },
		{ 295, 158, 33 },
		{ 295, 159, 33 },
		{ 295, 160, 33 },
		{ 295, 161, 33 },
		{ 295, 162, 33 },
		{ 295, 163, 33 },
		{ 295, 164, 33 },
		{ 295, 165, 33 },
		{ 295, 166, 33 },
		{ 295, 167, 33 },
		{ 295, 168, 33 },
		{ 295, 169, 33 },
		{ 295, 170, 33 },
		{ 295, 171, 33 },
		{ 295, 172, 33 },
		{ 295, 173, 33 },
		{ 295, 174, 33 },
		{ 295, 175, 33 },
		{ 295, 176, 33 },
		{ 295, 177, 33 },
		{ 295, 178, 33 },
		{ 295, 179, 33 },
		{ 295, 180, 33 },
		{ 296, 102, 33 },
		{ 296, 103, 33 },
		{ 296, 104, 33 },
		{ 296, 105, 33 },
		{ 296, 106, 33 },
		{ 296, 107, 33 },
		{ 296, 108, 33 },
		{ 296, 109, 33 },
		{ 296, 110, 33 },
		{ 296, 111, 33 },
		{ 296, 112, 33 },
		{ 296, 113, 33 },
		{ 296, 114, 33 },
		{ 296, 115, 33 },
		{ 296, 116, 33 },
		{ 296, 117, 33 },
		{ 296, 118, 33 },
		{ 296, 119, 33 },
		{ 296, 120, 33 },
		{ 296, 121, 33 },
		{ 296, 122, 33 },
		{ 296, 123, 33 },
		{ 296, 124, 33 },
		{ 296, 125, 33 },
		{ 296, 126, 33 },
		{ 296, 127, 33 },
		{ 296, 128, 33 },
		{ 296, 129, 33 },
		{ 296, 130, 33 },
		{ 296, 131, 33 },
		{ 296, 132, 33 },
		{ 296, 133, 33 },
		{ 296, 134, 33 },
		{ 296, 135, 33 },
		{ 296, 136, 33 },
		{ 296, 137, 33 },
		{ 296, 138, 33 },
		{ 296, 139, 33 },
		{ 296, 140, 33 },
		{ 296, 141, 33 },
		{ 296, 142, 33 },
		{ 296, 143, 33 },
		{ 296, 144, 33 },
		{ 296, 145, 33 },
		{ 296, 146, 33 },
		{ 296, 147, 33 },
		{ 296, 148, 33 },
		{ 296, 149, 33 },
		{ 296, 150, 33 },
		{ 296, 151, 33 },
		{ 296, 152, 33 },
		{ 296, 153, 33 },
		{ 296, 154, 33 },
		{ 296, 155, 33 },
		{ 296, 156, 33 },
		{ 296, 157, 33 },
		{ 296, 158, 33 },
		{ 296, 159, 33 },
		{ 296, 160, 33 },
		{ 296, 161, 33 },
		{ 296, 162, 33 },
		{ 296, 163, 33 },
		{ 296, 164, 33 },
		{ 296, 165, 33 },
		{ 296, 166, 33 },
		{ 296, 167, 33 },
		{ 296, 168, 33 },
		{ 296, 169, 33 },
		{ 296, 170, 33 },
		{ 296, 171, 33 },
		{ 296, 172, 33 },
		{ 296, 173, 33 },
		{ 296, 174, 33 },
		{ 296, 175, 33 },
		{ 296, 176, 33 },
		{ 296, 177, 33 },
		{ 296, 178, 33 },
		{ 297, 98, 33 },
		{ 297, 99, 33 },
		{ 297, 100, 33 },
		{ 297, 101, 33 },
		{ 297, 102, 33 },
		{ 297, 103, 33 },
		{ 297, 104, 33 },
		{ 297, 105, 33 },
		{ 297, 106, 33 },
		{ 297, 107, 33 },
		{ 297, 108, 33 },
		{ 297, 109, 33 },
		{ 297, 110, 33 },
		{ 297, 111, 33 },
		{ 297, 112, 33 },
		{ 297, 113, 33 },
		{ 297, 114, 33 },
		{ 297, 115, 33 },
		{ 297, 116, 33 },
		{ 297, 117, 33 },
		{ 297, 118, 33 },
		{ 297, 119, 33 },
		{ 297, 120, 33 },
		{ 297, 121, 33 },
		{ 297, 122, 33 },
		{ 297, 123, 33 },
		{ 297, 124, 33 },
		{ 297, 125, 33 },
		{ 297, 126, 33 },
		{ 297, 127, 33 },
		{ 297, 128, 33 },
		{ 297, 129, 33 },
		{ 297, 130, 33 },
		{ 297, 131, 33 },
		{ 297, 132, 33 },
		{ 297, 133, 33 },
		{ 297, 134, 33 },
		{ 297, 135, 33 },
		{ 297, 136, 33 },
		{ 297, 137, 33 },
		{ 297, 138, 33 },
		{ 297, 139, 33 },
		{ 297, 140, 33 },
		{ 297, 141, 33 },
		{ 297, 142, 33 },
		{ 297, 143, 33 },
		{ 297, 144, 33 },
		{ 297, 145, 33 },
		{ 297, 146, 33 },
		{ 297, 147, 33 },
		{ 297, 148, 33 },
		{ 297, 149, 33 },
		{ 297, 150, 33 },
		{ 297, 151, 33 },
		{ 297, 152, 33 },
		{ 297, 153, 33 },
		{ 297, 154, 33 },
		{ 297, 155, 33 },
		{ 297, 156, 33 },
		{ 297, 157, 33 },
		{ 297, 158, 33 },
		{ 297, 159, 33 },
		{ 297, 160, 33 },
		{ 297, 161, 33 },
		{ 297, 162, 33 },
		{ 297, 163, 33 },
		{ 297, 164, 33 },
		{ 297, 165, 33 },
		{ 297, 166, 33 },
		{ 297, 167, 33 },
		{ 297, 168, 33 },
		{ 297, 169, 33 },
		{ 297, 170, 33 },
		{ 297, 171, 33 },
		{ 297, 172, 33 },
		{ 297, 173, 33 },
		{ 297, 174, 33 },
		{ 297, 175, 33 },
		{ 297, 176, 33 },
		{ 298, 94, 33 },
		{ 298, 95, 33 },
		{ 298, 96, 33 },
		{ 298, 97, 33 },
		{ 298, 98, 33 },
		{ 298, 99, 33 },
		{ 298, 100, 33 },
		{ 298, 101, 33 },
		{ 298, 102, 33 },
		{ 298, 103, 33 },
		{ 298, 104, 33 },
		{ 298, 105, 33 },
		{ 298, 106, 33 },
		{ 298, 107, 33 },
		{ 298, 108, 33 },
		{ 298, 109, 33 },
		{ 298, 110, 33 },
		{ 298, 111, 33 },
		{ 298, 112, 33 },
		{ 298, 113, 33 },
		{ 298, 114, 33 },
		{ 298, 115, 33 },
		{ 298, 116, 33 },
		{ 298, 117, 33 },
		{ 298, 118, 33 },
		{ 298, 119, 33 },
		{ 298, 120, 33 },
		{ 298, 121, 33 },
		{ 298, 122, 33 },
		{ 298, 123, 33 },
		{ 298, 124, 33 },
		{ 298, 125, 33 },
		{ 298, 126, 33 },
		{ 298, 127, 33 },
		{ 298, 128, 33 },
		{ 298, 129, 33 },
		{ 298, 130, 33 },
		{ 298, 131, 33 },
		{ 298, 132, 33 },
		{ 298, 133, 33 },
		{ 298, 134, 33 },
		{ 298, 135, 33 },
		{ 298, 136, 33 },
		{ 298, 137, 33 },
		{ 298, 138, 33 },
		{ 298, 139, 33 },
		{ 298, 140, 33 },
		{ 298, 141, 33 },
		{ 298, 142, 33 },
		{ 298, 143, 33 },
		{ 298, 144, 33 },
		{ 298, 145, 33 },
		{ 298, 146, 33 },
		{ 298, 147, 33 },
		{ 298, 148, 33 },
		{ 298, 149, 33 },
		{ 298, 150, 33 },
		{ 298, 151, 33 },
		{ 298, 152, 33 },
		{ 298, 153, 33 },
		{ 298, 154, 33 },
		{ 298, 155, 33 },
		{ 298, 156, 33 },
		{ 298, 157, 33 },
		{ 298, 158, 33 },
		{ 298, 159, 33 },
		{ 298, 160, 33 },
		{ 298, 161, 33 },
		{ 298, 162, 33 },
		{ 298, 163, 33 },
		{ 298, 164, 33 },
		{ 298, 165, 33 },
		{ 298, 166, 33 },
		{ 298, 167, 33 },
		{ 298, 168, 33 },
		{ 298, 169, 33 },
		{ 298, 170, 33 },
		{ 298, 171, 33 },
		{ 298, 172, 33 },
		{ 298, 173, 33 },
		{ 298, 174, 33 },
		{ 299, 90, 33 },
		{ 299, 91, 33 },
		{ 299, 92, 33 },
		{ 299, 93, 33 },
		{ 299, 94, 33 },
		{ 299, 95, 33 },
		{ 299, 96, 33 },
		{ 299, 97, 33 },
		{ 299, 98, 33 },
		{ 299, 99, 33 },
		{ 299, 100, 33 },
		{ 299, 101, 33 },
		{ 299, 102, 33 },
		{ 299, 103, 33 },
		{ 299, 104, 33 },
		{ 299, 105, 33 },
		{ 299, 106, 33 },
		{ 299, 107, 33 },
		{ 299, 108, 33 },
		{ 299, 109, 33 },
		{ 299, 110, 33 },
		{ 299, 111, 33 },
		{ 299, 112, 33 },
		{ 299, 113, 33 },
		{ 299, 114, 33 },
		{ 299, 115, 33 },
		{ 299, 116, 33 },
		{ 299, 117, 33 },
		{ 299, 118, 33 },
		{ 299, 119, 33 },
		{ 299, 120, 33 },
		{ 299, 121, 33 },
		{ 299, 122, 33 },
		{ 299, 123, 33 },
		{ 299, 124, 33 },
		{ 299, 125, 33 },
		{ 299, 126, 33 },
		{ 299, 127, 33 },
		{ 299, 128, 33 },
		{ 299, 129, 33 },
		{ 299, 130, 33 },
		{ 299, 131, 33 },
		{ 299, 132, 33 },
		{ 299, 133, 33 },
		{ 299, 134, 33 },
		{ 299, 135, 33 },
		{ 299, 136, 33 },
		{ 299, 137, 33 },
		{ 299, 138, 33 },
		{ 299, 139, 33 },
		{ 299, 140, 33 },
		{ 299, 141, 33 },
		{ 299, 142, 33 },
		{ 299, 143, 33 },
		{ 299, 144, 33 },
		{ 299, 145, 33 },
		{ 299, 146, 33 },
		{ 299, 147, 33 },
		{ 299, 148, 33 },
		{ 299, 149, 33 },
		{ 299, 150, 33 },
		{ 299, 151, 33 },
		{ 299, 152, 33 },
		{ 299, 153, 33 },
		{ 299, 154, 33 },
		{ 299, 155, 33 },
		{ 299, 156, 33 },
		{ 299, 157, 33 },
		{ 299, 158, 33 },
		{ 299, 159, 33 },
		{ 299, 160, 33 },
		{ 299, 161, 33 },
		{ 299, 162, 33 },
		{ 299, 163, 33 },
		{ 299, 164, 33 },
		{ 299, 165, 33 },
		{ 299, 166, 33 },
		{ 299, 167, 33 },
		{ 299, 168, 33 },
		{ 299, 169, 33 },
		{ 299, 170, 33 },
		{ 299, 171, 33 },
		{ 299, 172, 33 },
		{ 300, 86, 33 },
		{ 300, 87, 33 },
		{ 300, 88, 33 },
		{ 300, 89, 33 },
		{ 300, 90, 33 },
		{ 300, 91, 33 },
		{ 300, 92, 33 },
		{ 300, 93, 33 },
		{ 300, 94, 33 },
		{ 300, 95, 33 },
		{ 300, 96, 33 },
		{ 300, 97, 33 },
		{ 300, 98, 33 },
		{ 300, 99, 33 },
		{ 300, 100, 33 },
		{ 300, 101, 33 },
		{ 300, 102, 33 },
		{ 300, 103, 33 },
		{ 300, 104, 33 },
		{ 300, 105, 33 },
		{ 300, 106, 33 },
		{ 300, 107, 33 },
		{ 300, 108, 33 },
		{ 300, 109, 33 },
		{ 300, 110, 33 },
		{ 300, 111, 33 },
		{ 300, 112, 33 },
		{ 300, 113, 33 },
		{ 300, 114, 33 },
		{ 300, 115, 33 },
		{ 300, 116, 33 },
		{ 300, 117, 33 },
		{ 300, 118, 33 },
		{ 300, 119, 33 },
		{ 300, 120, 33 },
		{ 300, 121, 33 },
		{ 300, 122, 33 },
		{ 300, 123, 33 },
		{ 300, 124, 33 },
		{ 300, 125, 33 },
		{ 300, 126, 33 },
		{ 300, 127, 33 },
		{ 300, 128, 33 },
		{ 300, 129, 33 },
		{ 300, 130, 33 },
		{ 300, 131, 33 },
		{ 300, 132, 33 },
		{ 300, 133, 33 },
		{ 300, 134, 33 },
		{ 300, 135, 33 },
		{ 300, 136, 33 },
		{ 300, 137, 33 },
		{ 300, 138, 33 },
		{ 300, 139, 33 },
		{ 300, 140, 33 },
		{ 300, 141, 33 },
		{ 300, 142, 33 },
		{ 300, 143, 33 },
		{ 300, 144, 33 },
		{ 300, 145, 33 },
		{ 300, 146, 33 },
		{ 300, 147, 33 },
		{ 300, 148, 33 },
		{ 300, 149, 33 },
		{ 300, 150, 33 },
		{ 300, 151, 33 },
		{ 300, 152, 33 },
		{ 300, 153, 33 },
		{ 300, 154, 33 },
		{ 300, 155, 33 },
		{ 300, 156, 33 },
		{ 300, 157, 33 },
		{ 300, 158, 33 },
		{ 300, 159, 33 },
		{ 300, 160, 33 },
		{ 300, 161, 33 },
		{ 300, 162, 33 },
		{ 300, 163, 33 },
		{ 300, 164, 33 },
		{ 300, 165, 33 },
		{ 300, 166, 33 },
		{ 300, 167, 33 },
		{ 300, 168, 33 },
		{ 300, 169, 33 },
		{ 300, 170, 33 },
		{ 301, 82, 33 },
		{ 301, 83, 33 },
		{ 301, 84, 33 },
		{ 301, 85, 33 },
		{ 301, 86, 33 },
		{ 301, 87, 33 },
		{ 301, 88, 33 },
		{ 301, 89, 33 },
		{ 301, 90, 33 },
		{ 301, 91, 33 },
		{ 301, 92, 33 },
		{ 301, 93, 33 },
		{ 301, 94, 33 },
		{ 301, 95, 33 },
		{ 301, 96, 33 },
		{ 301, 97, 33 },
		{ 301, 98, 33 },
		{ 301, 99, 33 },
		{ 301, 100, 33 },
		{ 301, 101, 33 },
		{ 301, 102, 33 },
		{ 301, 103, 33 },
		{ 301, 104, 33 },
		{ 301, 105, 33 },
		{ 301, 106, 33 },
		{ 301, 107, 33 },
		{ 301, 108, 33 },
		{ 301, 109, 33 },
		{ 301, 110, 33 },
		{ 301, 111, 33 },
		{ 301, 112, 33 },
		{ 301, 113, 33 },
		{ 301, 114, 33 },
		{ 301, 115, 33 },
		{ 301, 116, 33 },
		{ 301, 117, 33 },
		{ 301, 118, 33 },
		{ 301, 119, 33 },
		{ 301, 120, 33 },
		{ 301, 121, 33 },
		{ 301, 122, 33 },
		{ 301, 123, 33 },
		{ 301, 124, 33 },
		{ 301, 125, 33 },
		{ 301, 126, 33 },
		{ 301, 127, 33 },
		{ 301, 128, 33 },
		{ 301, 129, 33 },
		{ 301, 130, 33 },
		{ 301, 131, 33 },
		{ 301, 132, 33 },
		{ 301, 133, 33 },
		{ 301, 134, 33 },
		{ 301, 135, 33 },
		{ 301, 136, 33 },
		{ 301, 137, 33 },
		{ 301, 138, 33 },
		{ 301, 139, 33 },
		{ 301, 140, 33 },
		{ 301, 141, 33 },
		{ 301, 142, 33 },
		{ 301, 143, 33 },
		{ 301, 144, 33 },
		{ 301, 145, 33 },
		{ 301, 146, 33 },
		{ 301, 147, 33 },
		{ 301, 148, 33 },
		{ 301, 149, 33 },
		{ 301, 150, 33 },
		{ 301, 151, 33 },
		{ 301, 152, 33 },
		{ 301, 153, 33 },
		{ 301, 154, 33 },
		{ 301, 155, 33 },
		{ 301, 156, 33 },
		{ 301, 157, 33 },
		{ 301, 158, 33 },
		{ 301, 159, 33 },
		{ 301, 160, 33 },
		{ 301, 161, 33 },
		{ 301, 162, 33 },
		{ 301, 163, 33 },
		{ 301, 164, 33 },
		{ 301, 165, 33 },
		{ 301, 166, 33 },
		{ 301, 167, 33 },
		{ 301, 168, 33 },
		{ 302, 78, 33 },
		{ 302, 79, 33 },
		{ 302, 80, 33 },
		{ 302, 81, 33 },
		{ 302, 82, 33 },
		{ 302, 83, 33 },
		{ 302, 84, 33 },
		{ 302, 85, 33 },
		{ 302, 86, 33 },
		{ 302, 87, 33 },
		{ 302, 88, 33 },
		{ 302, 89, 33 },
		{ 302, 90, 33 },
		{ 302, 91, 33 },
		{ 302, 92, 33 },
		{ 302, 93, 33 },
		{ 302, 94, 33 },
		{ 302, 95, 33 },
		{ 302, 96, 33 },
		{ 302, 97, 33 },
		{ 302, 98, 33 },
		{ 302, 99, 33 },
		{ 302, 100, 33 },
		{ 302, 101, 33 },
		{ 302, 102, 33 },
		{ 302, 103, 33 },
		{ 302, 104, 33 },
		{ 302, 105, 33 },
		{ 302, 106, 33 },
		{ 302, 107, 33 },
		{ 302, 108, 33 },
		{ 302, 109, 33 },
		{ 302, 110, 33 },
		{ 302, 111, 33 },
		{ 302, 112, 33 },
		{ 302, 113, 33 },
		{ 302, 114, 33 },
		{ 302, 115, 33 },
		{ 302, 116, 33 },
		{ 302, 117, 33 },
		{ 302, 118, 33 },
		{ 302, 119, 33 },
		{ 302, 120, 33 },
		{ 302, 121, 33 },
		{ 302, 122, 33 },
		{ 302, 123, 33 },
		{ 302, 124, 33 },
		{ 302, 125, 33 },
		{ 302, 126, 33 },
		{ 302, 127, 33 },
		{ 302, 128, 33 },
		{ 302, 129, 33 },
		{ 302, 130, 33 },
		{ 302, 131, 33 },
		{ 302, 132, 33 },
		{ 302, 133, 33 },
		{ 302, 134, 33 },
		{ 302, 135, 33 },
		{ 302, 136, 33 },
		{ 302, 137, 33 },
		{ 302, 138, 33 },
		{ 302, 139, 33 },
		{ 302, 140, 33 },
		{ 302, 141, 33 },
		{ 302, 142, 33 },
		{ 302, 143, 33 },
		{ 302, 144, 33 },
		{ 302, 145, 33 },
		{ 302, 146, 33 },
		{ 302, 147, 33 },
		{ 302, 148, 33 },
		{ 302, 149, 33 },
		{ 302, 150, 33 },
		{ 302, 151, 33 },
		{ 302, 152, 33 },
		{ 302, 153, 33 },
		{ 302, 154, 33 },
		{ 302, 155, 33 },
		{ 302, 156, 33 },
		{ 302, 157, 33 },
		{ 302, 158, 33 },
		{ 302, 159, 33 },
		{ 302, 160, 33 },
		{ 302, 161, 33 },
		{ 302, 162, 33 },
		{ 302, 163, 33 },
		{ 302, 164, 33 },
		{ 302, 165, 33 },
		{ 302, 166, 33 },
		{ 303, 74, 33 },
		{ 303, 75, 33 },
		{ 303, 76, 33 },
		{ 303, 77, 33 },
		{ 303, 78, 33 },
		{ 303, 79, 33 },
		{ 303, 80, 33 },
		{ 303, 81, 33 },
		{ 303, 82, 33 },
		{ 303, 83, 33 },
		{ 303, 84, 33 },
		{ 303, 85, 33 },
		{ 303, 86, 33 },
		{ 303, 87, 33 },
		{ 303, 88, 33 },
		{ 303, 89, 33 },
		{ 303, 90, 33 },
		{ 303, 91, 33 },
		{ 303, 92, 33 },
		{ 303, 93, 33 },
		{ 303, 94, 33 },
		{ 303, 95, 33 },
		{ 303, 96, 33 },
		{ 303, 97, 33 },
		{ 303, 98, 33 },
		{ 303, 99, 33 },
		{ 303, 100, 33 },
		{ 303, 101, 33 },
		{ 303, 102, 33 },
		{ 303, 103, 33 },
		{ 303, 104, 33 },
		{ 303, 105, 33 },
		{ 303, 106, 33 },
		{ 303, 107, 33 },
		{ 303, 108, 33 },
		{ 303, 109, 33 },
		{ 303, 110, 33 },
		{ 303, 111, 33 },
		{ 303, 112, 33 },
		{ 303, 113, 33 },
		{ 303, 114, 33 },
		{ 303, 115, 33 },
		{ 303, 116, 33 },
		{ 303, 117, 33 },
		{ 303, 118, 33 },
		{ 303, 119, 33 },
		{ 303, 120, 33 },
		{ 303, 121, 33 },
		{ 303, 122, 33 },
		{ 303, 123, 33 },
		{ 303, 124, 33 },
		{ 303, 125, 33 },
		{ 303, 126, 33 },
		{ 303, 127, 33 },
		{ 303, 128, 33 },
		{ 303, 129, 33 },
		{ 303, 130, 33 },
		{ 303, 131, 33 },
		{ 303, 132, 33 },
		{ 303, 133, 33 },
		{ 303, 134, 33 },
		{ 303, 135, 33 },
		{ 303, 136, 33 },
		{ 303, 137, 33 },
		{ 303, 138, 33 },
		{ 303, 139, 33 },
		{ 303, 140, 33 },
		{ 303, 141, 33 },
		{ 303, 142, 33 },
		{ 303, 143, 33 },
		{ 303, 144, 33 },
		{ 303, 145, 33 },
		{ 303, 146, 33 },
		{ 303, 147, 33 },
		{ 303, 148, 33 },
		{ 303, 149, 33 },
		{ 303, 150, 33 },
		{ 303, 151, 33 },
		{ 303, 152, 33 },
		{ 303, 153, 33 },
		{ 303, 154, 33 },
		{ 303, 155, 33 },
		{ 303, 156, 33 },
		{ 303, 157, 33 },
		{ 303, 158, 33 },
		{ 303, 159, 33 },
		{ 303, 160, 33 },
		{ 303, 161, 33 },
		{ 303, 162, 33 },
		{ 303, 163, 33 },
		{ 303, 164, 33 },
		{ 304, 70, 33 },
		{ 304, 71, 33 },
		{ 304, 72, 33 },
		{ 304, 73, 33 },
		{ 304, 74, 33 },
		{ 304, 75, 33 },
		{ 304, 76, 33 },
		{ 304, 77, 33 },
		{ 304, 78, 33 },
		{ 304, 79, 33 },
		{ 304, 80, 33 },
		{ 304, 81, 33 },
		{ 304, 82, 33 },
		{ 304, 83, 33 },
		{ 304, 84, 33 },
		{ 304, 85, 33 },
		{ 304, 86, 33 },
		{ 304, 87, 33 },
		{ 304, 88, 33 },
		{ 304, 89, 33 },
		{ 304, 90, 33 },
		{ 304, 91, 33 },
		{ 304, 92, 33 },
		{ 304, 93, 33 },
		{ 304, 94, 33 },
		{ 304, 95, 33 },
		{ 304, 96, 33 },
		{ 304, 97, 33 },
		{ 304, 98, 33 },
		{ 304, 99, 33 },
		{ 304, 100, 33 },
		{ 304, 101, 33 },
		{ 304, 102, 33 },
		{ 304, 103, 33 },
		{ 304, 104, 33 },
		{ 304, 105, 33 },
		{ 304, 106, 33 },
		{ 304, 107, 33 },
		{ 304, 108, 33 },
		{ 304, 109, 33 },
		{ 304, 110, 33 },
		{ 304, 111, 33 },
		{ 304, 112, 33 },
		{ 304, 113, 33 },
		{ 304, 114, 33 },
		{ 304, 115, 33 },
		{ 304, 116, 33 },
		{ 304, 117, 33 },
		{ 304, 118, 33 },
		{ 304, 119, 33 },
		{ 304, 120, 33 },
		{ 304, 121, 33 },
		{ 304, 122, 33 },
		{ 304, 123, 33 },
		{ 304, 124, 33 },
		{ 304, 125, 33 },
		{ 304, 126, 33 },
		{ 304, 127, 33 },
		{ 304, 128, 33 },
		{ 304, 129, 33 },
		{ 304, 130, 33 },
		{ 304, 131, 33 },
		{ 304, 132, 33 },
		{ 304, 133, 33 },
		{ 304, 134, 33 },
		{ 304, 135, 33 },
		{ 304, 136, 33 },
		{ 304, 137, 33 },
		{ 304, 138, 33 },
		{ 304, 139, 33 },
		{ 304, 140, 33 },
		{ 304, 141, 33 },
		{ 304, 142, 33 },
		{ 304, 143, 33 },
		{ 304, 144, 33 },
		{ 304, 145, 33 },
		{ 304, 146, 33 },
		{ 304, 147, 33 },
		{ 304, 148, 33 },
		{ 304, 149, 33 },
		{ 304, 150, 33 },
		{ 304, 151, 33 },
		{ 304, 152, 33 },
		{ 304, 153, 33 },
		{ 304, 154, 33 },
		{ 304, 155, 33 },
		{ 304, 156, 33 },
		{ 304, 157, 33 },
		{ 304, 158, 33 },
		{ 304, 159, 33 },
		{ 304, 160, 33 },
		{ 304, 161, 33 },
		{ 304, 162, 33 },
		{ 305, 66, 33 },
		{ 305, 67, 33 },
		{ 305, 68, 33 },
		{ 305, 69, 33 },
		{ 305, 70, 33 },
		{ 305, 71, 33 },
		{ 305, 72, 33 },
		{ 305, 73, 33 },
		{ 305, 74, 33 },
		{ 305, 75, 33 },
		{ 305, 76, 33 },
		{ 305, 77, 33 },
		{ 305, 78, 33 },
		{ 305, 79, 33 },
		{ 305, 80, 33 },
		{ 305, 81, 33 },
		{ 305, 82, 33 },
		{ 305, 83, 33 },
		{ 305, 84, 33 },
		{ 305, 85, 33 },
		{ 305, 86, 33 },
		{ 305, 87, 33 },
		{ 305, 88, 33 },
		{ 305, 89, 33 },
		{ 305, 90, 33 },
		{ 305, 91, 33 },
		{ 305, 92, 33 },
		{ 305, 93, 33 },
		{ 305, 94, 33 },
		{ 305, 95, 33 },
		{ 305, 96, 33 },
		{ 305, 97, 33 },
		{ 305, 98, 33 },
		{ 305, 99, 33 },
		{ 305, 100, 33 },
		{ 305, 101, 33 },
		{ 305, 102, 33 },
		{ 305, 103, 33 },
		{ 305, 104, 33 },
		{ 305, 105, 33 },
		{ 305, 106, 33 },
		{ 305, 107, 33 },
		{ 305, 108, 33 },
		{ 305, 109, 33 },
		{ 305, 110, 33 },
		{ 305, 111, 33 },
		{ 305, 112, 33 },
		{ 305, 113, 33 },
		{ 305, 114, 33 },
		{ 305, 115, 33 },
		{ 305, 116, 33 },
		{ 305, 117, 33 },
		{ 305, 118, 33 },
		{ 305, 119, 33 },
		{ 305, 120, 33 },
		{ 305, 121, 33 },
		{ 305, 122, 33 },
		{ 305, 123, 33 },
		{ 305, 124, 33 },
		{ 305, 125, 33 },
		{ 305, 126, 33 },
		{ 305, 127, 33 },
		{ 305, 128, 33 },
		{ 305, 129, 33 },
		{ 305, 130, 33 },
		{ 305, 131, 33 },
		{ 305, 132, 33 },
		{ 305, 133, 33 },
		{ 305, 134, 33 },
		{ 305, 135, 33 },
		{ 305, 136, 33 },
		{ 305, 137, 33 },
		{ 305, 138, 33 },
		{ 305, 139, 33 },
		{ 305, 140, 33 },
		{ 305, 141, 33 },
		{ 305, 142, 33 },
		{ 305, 143, 33 },
		{ 305, 144, 33 },
		{ 305, 145, 33 },
		{ 305, 146, 33 },
		{ 305, 147, 33 },
		{ 305, 148, 33 },
		{ 305, 149, 33 },
		{ 305, 150, 33 },
		{ 305, 151, 33 },
		{ 305, 152, 33 },
		{ 305, 153, 33 },
		{ 305, 154, 33 },
		{ 305, 155, 33 },
		{ 305, 156, 33 },
		{ 305, 157, 33 },
		{ 305, 158, 33 },
		{ 305, 159, 33 },
		{ 305, 160, 33 },
		{ 306, 62, 33 },
		{ 306, 63, 33 },
		{ 306, 64, 33 },
		{ 306, 65, 33 },
		{ 306, 66, 33 },
		{ 306, 67, 33 },
		{ 306, 68, 33 },
		{ 306, 69, 33 },
		{ 306, 70, 33 },
		{ 306, 71, 33 },
		{ 306, 72, 33 },
		{ 306, 73, 33 },
		{ 306, 74, 33 },
		{ 306, 75, 33 },
		{ 306, 76, 33 },
		{ 306, 77, 33 },
		{ 306, 78, 33 },
		{ 306, 79, 33 },
		{ 306, 80, 33 },
		{ 306, 81, 33 },
		{ 306, 82, 33 },
		{ 306, 83, 33 },
		{ 306, 84, 33 },
		{ 306, 85, 33 },
		{ 306, 86, 33 },
		{ 306, 87, 33 },
		{ 306, 88, 33 },
		{ 306, 89, 33 },
		{ 306, 90, 33 },
		{ 306, 91, 33 },
		{ 306, 92, 33 },
		{ 306, 93, 33 },
		{ 306, 94, 33 },
		{ 306, 95, 33 },
		{ 306, 96, 33 },
		{ 306, 97, 33 },
		{ 306, 98, 33 },
		{ 306, 99, 33 },
		{ 306, 100, 33 },
		{ 306, 101, 33 },
		{ 306, 102, 33 },
		{ 306, 103, 33 },
		{ 306, 104, 33 },
		{ 306, 105, 33 },
		{ 306, 106, 33 },
		{ 306, 107, 33 },
		{ 306, 108, 33 },
		{ 306, 109, 33 },
		{ 306, 110, 33 },
		{ 306, 111, 33 },
		{ 306, 112, 33 },
		{ 306, 113, 33 },
		{ 306, 114, 33 },
		{ 306, 115, 33 },
		{ 306, 116, 33 },
		{ 306, 117, 33 },
		{ 306, 118, 33 },
		{ 306, 119, 33 },
		{ 306, 120, 33 },
		{ 306, 121, 33 },
		{ 306, 122, 33 },
		{ 306, 123, 33 },
		{ 306, 124, 33 },
		{ 306, 125, 33 },
		{ 306, 126, 33 },
		{ 306, 127, 33 },
		{ 306, 128, 33 },
		{ 306, 129, 33 },
		{ 306, 130, 33 },
		{ 306, 131, 33 },
		{ 306, 132, 33 },
		{ 306, 133, 33 },
		{ 306, 134, 33 },
		{ 306, 135, 33 },
		{ 306, 136, 33 },
		{ 306, 137, 33 },
		{ 306, 138, 33 },
		{ 306, 139, 33 },
		{ 306, 140, 33 },
		{ 306, 141, 33 },
		{ 306, 142, 33 },
		{ 306, 143, 33 },
		{ 306, 144, 33 },
		{ 306, 145, 33 },
		{ 306, 146, 33 },
		{ 306, 147, 33 },
		{ 306, 148, 33 },
		{ 306, 149, 33 },
		{ 306, 150, 33 },
		{ 306, 151, 33 },
		{ 306, 152, 33 },
		{ 306, 153, 33 },
		{ 306, 154, 33 },
		{ 306, 155, 33 },
		{ 306, 156, 33 },
		{ 306, 157, 33 },
		{ 306, 158, 33 },
		{ 307, 58, 33 },
		{ 307, 59, 33 },
		{ 307, 60, 33 },
		{ 307, 61, 33 },
		{ 307, 62, 33 },
		{ 307, 63, 33 },
		{ 307, 64, 33 },
		{ 307, 65, 33 },
		{ 307, 66, 33 },
		{ 307, 67, 33 },
		{ 307, 68, 33 },
		{ 307, 69, 33 },
		{ 307, 70, 33 },
		{ 307, 71, 33 },
		{ 307, 72, 33 },
		{ 307, 73, 33 },
		{ 307, 74, 33 },
		{ 307, 75, 33 },
		{ 307, 76, 33 },
		{ 307, 77, 33 },
		{ 307, 78, 33 },
		{ 307, 79, 33 },
		{ 307, 80, 33 },
		{ 307, 81, 33 },
		{ 307, 82, 33 },
		{ 307, 83, 33 },
		{ 307, 84, 33 },
		{ 307, 85, 33 },
		{ 307, 86, 33 },
		{ 307, 87, 33 },
		{ 307, 88, 33 },
		{ 307, 89, 33 },
		{ 307, 90, 33 },
		{ 307, 91, 33 },
		{ 307, 92, 33 },
		{ 307, 93, 33 },
		{ 307, 94, 33 },
		{ 307, 95, 33 },
		{ 307, 96, 33 },
		{ 307, 97, 33 },
		{ 307, 98, 33 },
		{ 307, 99, 33 },
		{ 307, 100, 33 },
		{ 307, 101, 33 },
		{ 307, 102, 33 },
		{ 307, 103, 33 },
		{ 307, 104, 33 },
		{ 307, 105, 33 },
		{ 307, 106, 33 },
		{ 307, 107, 33 },
		{ 307, 108, 33 },
		{ 307, 109, 33 },
		{ 307, 110, 33 },
		{ 307, 111, 33 },
		{ 307, 112, 33 },
		{ 307, 113, 33 },
		{ 307, 114, 33 },
		{ 307, 115, 33 },
		{ 307, 116, 33 },
		{ 307, 117, 33 },
		{ 307, 118, 33 },
		{ 307, 119, 33 },
		{ 307, 120, 33 },
		{ 307, 121, 33 },
		{ 307, 122, 33 },
		{ 307, 123, 33 },
		{ 307, 124, 33 },
		{ 307, 125, 33 },
		{ 307, 126, 33 },
		{ 307, 127, 33 },
		{ 307, 128, 33 },
		{ 307, 129, 33 },
		{ 307, 130, 33 },
		{ 307, 131, 33 },
		{ 307, 132, 33 },
		{ 307, 133, 33 },
		{ 307, 134, 33 },
		{ 307, 135, 33 },
		{ 307, 136, 33 },
		{ 307, 137, 33 },
		{ 307, 138, 33 },
		{ 307, 139, 33 },
		{ 307, 140, 33 },
		{ 307, 141, 33 },
		{ 307, 142, 33 },
		{ 307, 143, 33 },
		{ 307, 144, 33 },
		{ 307, 145, 33 },
		{ 307, 146, 33 },
		{ 307, 147, 33 },
		{ 307, 148, 33 },
		{ 307, 149, 33 },
		{ 307, 150, 33 },
		{ 307, 151, 33 },
		{ 307, 152, 33 },
		{ 307, 153, 33 },
		{ 307, 154, 33 },
		{ 307, 155, 33 },
		{ 307, 156, 33 },
		{ 308, 54, 33 },
		{ 308, 55, 33 },
		{ 308, 56, 33 },
		{ 308, 57, 33 },
		{ 308, 58, 33 },
		{ 308, 59, 33 },
		{ 308, 60, 33 },
		{ 308, 61, 33 },
		{ 308, 62, 33 },
		{ 308, 63, 33 },
		{ 308, 64, 33 },
		{ 308, 65, 33 },
		{ 308, 66, 33 },
		{ 308, 67, 33 },
		{ 308, 68, 33 },
		{ 308, 69, 33 },
		{ 308, 70, 33 },
		{ 308, 71, 33 },
		{ 308, 72, 33 },
		{ 308, 73, 33 },
		{ 308, 74, 33 },
		{ 308, 75, 33 },
		{ 308, 76, 33 },
		{ 308, 77, 33 },
		{ 308, 78, 33 },
		{ 308, 79, 33 },
		{ 308, 80, 33 },
		{ 308, 81, 33 },
		{ 308, 82, 33 },
		{ 308, 83, 33 },
		{ 308, 84, 33 },
		{ 308, 85, 33 },
		{ 308, 86, 33 },
		{ 308, 87, 33 },
		{ 308, 88, 33 },
		{ 308, 89, 33 },
		{ 308, 90, 33 },
		{ 308, 91, 33 },
		{ 308, 92, 33 },
		{ 308, 93, 33 },
		{ 308, 94, 33 },
		{ 308, 95, 33 },
		{ 308, 96, 33 },
		{ 308, 97, 33 },
		{ 308, 98, 33 },
		{ 308, 99, 33 },
		{ 308, 100, 33 },
		{ 308, 101, 33 },
		{ 308, 102, 33 },
		{ 308, 103, 33 },
		{ 308, 104, 33 },
		{ 308, 105, 33 },
		{ 308, 106, 33 },
		{ 308, 107, 33 },
		{ 308, 108, 33 },
		{ 308, 109, 33 },
		{ 308, 110, 33 },
		{ 308, 111, 33 },
		{ 308, 112, 33 },
		{ 308, 113, 33 },
		{ 308, 114, 33 },
		{ 308, 115, 33 },
		{ 308, 116, 33 },
		{ 308, 117, 33 },
		{ 308, 118, 33 },
		{ 308, 119, 33 },
		{ 308, 120, 33 },
		{ 308, 121, 33 },
		{ 308, 122, 33 },
		{ 308, 123, 33 },
		{ 308, 124, 33 },
		{ 308, 125, 33 },
		{ 308, 126, 33 },
		{ 308, 127, 33 },
		{ 308, 128, 33 },
		{ 308, 129, 33 },
		{ 308, 130, 33 },
		{ 308, 131, 33 },
		{ 308, 132, 33 },
		{ 308, 133, 33 },
		{ 308, 134, 33 },
		{ 308, 135, 33 },
		{ 308, 136, 33 },
		{ 308, 137, 33 },
		{ 308, 138, 33 },
		{ 308, 139, 33 },
		{ 308, 140, 33 },
		{ 308, 141, 33 },
		{ 308, 142, 33 },
		{ 308, 143, 33 },
		{ 308, 144, 33 },
		{ 308, 145, 33 },
		{ 308, 146, 33 },
		{ 308, 147, 33 },
		{ 308, 148, 33 },
		{ 308, 149, 33 },
		{ 308, 150, 33 },
		{ 308, 151, 33 },
		{ 308, 152, 33 },
		{ 308, 153, 33 },
		{ 308, 154, 33 },
		{ 309, 50, 33 },
		{ 309, 51, 33 },
		{ 309, 52, 33 },
		{ 309, 53, 33 },
		{ 309, 54, 33 },
		{ 309, 55, 33 },
		{ 309, 56, 33 },
		{ 309, 57, 33 },
		{ 309, 58, 33 },
		{ 309, 59, 33 },
		{ 309, 60, 33 },
		{ 309, 61, 33 },
		{ 309, 62, 33 },
		{ 309, 63, 33 },
		{ 309, 64, 33 },
		{ 309, 65, 33 },
		{ 309, 66, 33 },
		{ 309, 67, 33 },
		{ 309, 68, 33 },
		{ 309, 69, 33 },
		{ 309, 70, 33 },
		{ 309, 71, 33 },
		{ 309, 72, 33 },
		{ 309, 73, 33 },
		{ 309, 74, 33 },
		{ 309, 75, 33 },
		{ 309, 76, 33 },
		{ 309, 77, 33 },
		{ 309, 78, 33 },
		{ 309, 79, 33 },
		{ 309, 80, 33 },
		{ 309, 81, 33 },
		{ 309, 82, 33 },
		{ 309, 83, 33 },
		{ 309, 84, 33 },
		{ 309, 85, 33 },
		{ 309, 86, 33 },
		{ 309, 87, 33 },
		{ 309, 88, 33 },
		{ 309, 89, 33 },
		{ 309, 90, 33 },
		{ 309, 91, 33 },
		{ 309, 92, 33 },
		{ 309, 93, 33 },
		{ 309, 94, 33 },
		{ 309, 95, 33 },
		{ 309, 96, 33 },
		{ 309, 97, 33 },
		{ 309, 98, 33 },
		{ 309, 99, 33 },
		{ 309, 100, 33 },
		{ 309, 101, 33 },
		{ 309, 102, 33 },
		{ 309, 103, 33 },
		{ 309, 104, 33 },
		{ 309, 105, 33 },
		{ 309, 106, 33 },
		{ 309, 107, 33 },
		{ 309, 108, 33 },
		{ 309, 109, 33 },
		{ 309, 110, 33 },
		{ 309, 111, 33 },
		{ 309, 112, 33 },
		{ 309, 113, 33 },
		{ 309, 114, 33 },
		{ 309, 115, 33 },
		{ 309, 116, 33 },
		{ 309, 117, 33 },
		{ 309, 118, 33 },
		{ 309, 119, 33 },
		{ 309, 120, 33 },
		{ 309, 121, 33 },
		{ 309, 122, 33 },
		{ 309, 123, 33 },
		{ 309, 124, 33 },
		{ 309, 125, 33 },
		{ 309, 126, 33 },
		{ 309, 127, 33 },
		{ 309, 128, 33 },
		{ 309, 129, 33 },
		{ 309, 130, 33 },
		{ 309, 131, 33 },
		{ 309, 132, 33 },
		{ 309, 133, 33 },
		{ 309, 134, 33 },
		{ 309, 135, 33 },
		{ 309, 136, 33 },
		{ 309, 137, 33 },
		{ 309, 138, 33 },
		{ 309, 139, 33 },
		{ 309, 140, 33 },
		{ 309, 141, 33 },
		{ 309, 142, 33 },
		{ 309, 143, 33 },
		{ 309, 144, 33 },
		{ 309, 145, 33 },
		{ 309, 146, 33 },
		{ 309, 147, 33 },
		{ 309, 148, 33 },
		{ 309, 149, 33 },
		{ 309, 150, 33 },
		{ 309, 151, 33 },
		{ 309, 152, 33 },
		{ 310, 46, 33 },
		{ 310, 47, 33 },
		{ 310, 48, 33 },
		{ 310, 49, 33 },
		{ 310, 50, 33 },
		{ 310, 51, 33 },
		{ 310, 52, 33 },
		{ 310, 53, 33 },
		{ 310, 54, 33 },
		{ 310, 55, 33 },
		{ 310, 56, 33 },
		{ 310, 57, 33 },
		{ 310, 58, 33 },
		{ 310, 59, 33 },
		{ 310, 60, 33 },
		{ 310, 61, 33 },
		{ 310, 62, 33 },
		{ 310, 63, 33 },
		{ 310, 64, 33 },
		{ 310, 65, 33 },
		{ 310, 66, 33 },
		{ 310, 67, 33 },
		{ 310, 68, 33 },
		{ 310, 69, 33 },
		{ 310, 70, 33 },
		{ 310, 71, 33 },
		{ 310, 72, 33 },
		{ 310, 73, 33 },
		{ 310, 74, 33 },
		{ 310, 75, 33 },
		{ 310, 76, 33 },
		{ 310, 77, 33 },
		{ 310, 78, 33 },
		{ 310, 79, 33 },
		{ 310, 80, 33 },
		{ 310, 81, 33 },
		{ 310, 82, 33 },
		{ 310, 83, 33 },
		{ 310, 84, 33 },
		{ 310, 85, 33 },
		{ 310, 86, 33 },
		{ 310, 87, 33 },
		{ 310, 88, 33 },
		{ 310, 89, 33 },
		{ 310, 90, 33 },
		{ 310, 91, 33 },
		{ 310, 92, 33 },
		{ 310, 93, 33 },
		{ 310, 94, 33 },
		{ 310, 95, 33 },
		{ 310, 96, 33 },
		{ 310, 97, 33 },
		{ 310, 98, 33 },
		{ 310, 99, 33 },
		{ 310, 100, 33 },
		{ 310, 101, 33 },
		{ 310, 102, 33 },
		{ 310, 103, 33 },
		{ 310, 104, 33 },
		{ 310, 105, 33 },
		{ 310, 106, 33 },
		{ 310, 107, 33 },
		{ 310, 108, 33 },
		{ 310, 109, 33 },
		{ 310, 110, 33 },
		{ 310, 111, 33 },
		{ 310, 112, 33 },
		{ 310, 113, 33 },
		{ 310, 114, 33 },
		{ 310, 115, 33 },
		{ 310, 116, 33 },
		{ 310, 117, 33 },
		{ 310, 118, 33 },
		{ 310, 119, 33 },
		{ 310, 120, 33 },
		{ 310, 121, 33 },
		{ 310, 122, 33 },
		{ 310, 123, 33 },
		{ 310, 124, 33 },
		{ 310, 125, 33 },
		{ 310, 126, 33 },
		{ 310, 127, 33 },
		{ 310, 128, 33 },
		{ 310, 129, 33 },
		{ 310, 130, 33 },
		{ 310, 131, 33 },
		{ 310, 132, 33 },
		{ 310, 133, 33 },
		{ 310, 134, 33 },
		{ 310, 135, 33 },
		{ 310, 136, 33 },
		{ 310, 137, 33 },
		{ 310, 138, 33 },
		{ 310, 139, 33 },
		{ 310, 140, 33 },
		{ 310, 141, 33 },
		{ 310, 142, 33 },
		{ 310, 143, 33 },
		{ 310, 144, 33 },
		{ 310, 145, 33 },
		{ 310, 146, 33 },
		{ 310, 147, 33 },
		{ 310, 148, 33 },
		{ 310, 149, 33 },
		{ 310, 150, 33 },
		{ 311, 42, 33 },
		{ 311, 43, 33 },
		{ 311, 44, 33 },
		{ 311, 45, 33 },
		{ 311, 46, 33 },
		{ 311, 47, 33 },
		{ 311, 48, 33 },
		{ 311, 49, 33 },
		{ 311, 50, 33 },
		{ 311, 51, 33 },
		{ 311, 52, 33 },
		{ 311, 53, 33 },
		{ 311, 54, 33 },
		{ 311, 55, 33 },
		{ 311, 56, 33 },
		{ 311, 57, 33 },
		{ 311, 58, 33 },
		{ 311, 59, 33 },
		{ 311, 60, 33 },
		{ 311, 61, 33 },
		{ 311, 62, 33 },
		{ 311, 63, 33 },
		{ 311, 64, 33 },
		{ 311, 65, 33 },
		{ 311, 66, 33 },
		{ 311, 67, 33 },
		{ 311, 68, 33 },
		{ 311, 69, 33 },
		{ 311, 70, 33 },
		{ 311, 71, 33 },
		{ 311, 72, 33 },
		{ 311, 73, 33 },
		{ 311, 74, 33 },
		{ 311, 75, 33 },
		{ 311, 76, 33 },
		{ 311, 77, 33 },
		{ 311, 78, 33 },
		{ 311, 79, 33 },
		{ 311, 80, 33 },
		{ 311, 81, 33 },
		{ 311, 82, 33 },
		{ 311, 83, 33 },
		{ 311, 84, 33 },
		{ 311, 85, 33 },
		{ 311, 86, 33 },
		{ 311, 87, 33 },
		{ 311, 88, 33 },
		{ 311, 89, 33 },
		{ 311, 90, 33 },
		{ 311, 91, 33 },
		{ 311, 92, 33 },
		{ 311, 93, 33 },
		{ 311, 94, 33 },
		{ 311, 95, 33 },
		{ 311, 96, 33 },
		{ 311, 97, 33 },
		{ 311, 98, 33 },
		{ 311, 99, 33 },
		{ 311, 100, 33 },
		{ 311, 101, 33 },
		{ 311, 102, 33 },
		{ 311, 103, 33 },
		{ 311, 104, 33 },
		{ 311, 105, 33 },
		{ 311, 106, 33 },
		{ 311, 107, 33 },
		{ 311, 108, 33 },
		{ 311, 109, 33 },
		{ 311, 110, 33 },
		{ 311, 111, 33 },
		{ 311, 112, 33 },
		{ 311, 113, 33 },
		{ 311, 114, 33 },
		{ 311, 115, 33 },
		{ 311, 116, 33 },
		{ 311, 117, 33 },
		{ 311, 118, 33 },
		{ 311, 119, 33 },
		{ 311, 120, 33 },
		{ 311, 121, 33 },
		{ 311, 122, 33 },
		{ 311, 123, 33 },
		{ 311, 124, 33 },
		{ 311, 125, 33 },
		{ 311, 126, 33 },
		{ 311, 127, 33 },
		{ 311, 128, 33 },
		{ 311, 129, 33 },
		{ 311, 130, 33 },
		{ 311, 131, 33 },
		{ 311, 132, 33 },
		{ 311, 133, 33 },
		{ 311, 134, 33 },
		{ 311, 135, 33 },
		{ 311, 136, 33 },
		{ 311, 137, 33 },
		{ 311, 138, 33 },
		{ 311, 139, 33 },
		{ 311, 140, 33 },
		{ 311, 141, 33 },
		{ 311, 142, 33 },
		{ 311, 143, 33 },
		{ 311, 144, 33 },
		{ 311, 145, 33 },
		{ 311, 146, 33 },
		{ 311, 147, 33 },
		{ 311, 148, 33 },
		{ 312, 38, 33 },
		{ 312, 39, 33 },
		{ 312, 40, 33 },
		{ 312, 41, 33 },
		{ 312, 42, 33 },
		{ 312, 43, 33 },
		{ 312, 44, 33 },
		{ 312, 45, 33 },
		{ 312, 46, 33 },
		{ 312, 47, 33 },
		{ 312, 48, 33 },
		{ 312, 49, 33 },
		{ 312, 50, 33 },
		{ 312, 51, 33 },
		{ 312, 52, 33 },
		{ 312, 53, 33 },
		{ 312, 54, 33 },
		{ 312, 55, 33 },
		{ 312, 56, 33 },
		{ 312, 57, 33 },
		{ 312, 58, 33 },
		{ 312, 59, 33 },
		{ 312, 60, 33 },
		{ 312, 61, 33 },
		{ 312, 62, 33 },
		{ 312, 63, 33 },
		{ 312, 64, 33 },
		{ 312, 65, 33 },
		{ 312, 66, 33 },
		{ 312, 67, 33 },
		{ 312, 68, 33 },
		{ 312, 69, 33 },
		{ 312, 70, 33 },
		{ 312, 71, 33 },
		{ 312, 72, 33 },
		{ 312, 73, 33 },
		{ 312, 74, 33 },
		{ 312, 75, 33 },
		{ 312, 76, 33 },
		{ 312, 77, 33 },
		{ 312, 78, 33 },
		{ 312, 79, 33 },
		{ 312, 80, 33 },
		{ 312, 81, 33 },
		{ 312, 82, 33 },
		{ 312, 83, 33 },
		{ 312, 84, 33 },
		{ 312, 85, 33 },
		{ 312, 86, 33 },
		{ 312, 87, 33 },
		{ 312, 88, 33 },
		{ 312, 89, 33 },
		{ 312, 90, 33 },
		{ 312, 91, 33 },
		{ 312, 92, 33 },
		{ 312, 93, 33 },
		{ 312, 94, 33 },
		{ 312, 95, 33 },
		{ 312, 96, 33 },
		{ 312, 97, 33 },
		{ 312, 98, 33 },
		{ 312, 99, 33 },
		{ 312, 100, 33 },
		{ 312, 101, 33 },
		{ 312, 102, 33 },
		{ 312, 103, 33 },
		{ 312, 104, 33 },
		{ 312, 105, 33 },
		{ 312, 106, 33 },
		{ 312, 107, 33 },
		{ 312, 108, 33 },
		{ 312, 109, 33 },
		{ 312, 110, 33 },
		{ 312, 111, 33 },
		{ 312, 112, 33 },
		{ 312, 113, 33 },
		{ 312, 114, 33 },
		{ 312, 115, 33 },
		{ 312, 116, 33 },
		{ 312, 117, 33 },
		{ 312, 118, 33 },
		{ 312, 119, 33 },
		{ 312, 120, 33 },
		{ 312, 121, 33 },
		{ 312, 122, 33 },
		{ 312, 123, 33 },
		{ 312, 124, 33 },
		{ 312, 125, 33 },
		{ 312, 126, 33 },
		{ 312, 127, 33 },
		{ 312, 128, 33 },
		{ 312, 129, 33 },
		{ 312, 130, 33 },
		{ 312, 131, 33 },
		{ 312, 132, 33 },
		{ 312, 133, 33 },
		{ 312, 134, 33 },
		{ 312, 135, 33 },
		{ 312, 136, 33 },
		{ 312, 137, 33 },
		{ 312, 138, 33 },
		{ 312, 139, 33 },
		{ 312, 140, 33 },
		{ 312, 141, 33 },
		{ 312, 142, 33 },
		{ 312, 143, 33 },
		{ 312, 144, 33 },
		{ 312, 145, 33 },
		{ 312, 146, 33 },
		{ 313, 34, 33 },
		{ 313, 35, 33 },
		{ 313, 36, 33 },
		{ 313, 37, 33 },
		{ 313, 38, 33 },
		{ 313, 39, 33 },
		{ 313, 40, 33 },
		{ 313, 41, 33 },
		{ 313, 42, 33 },
		{ 313, 43, 33 },
		{ 313, 44, 33 },
		{ 313, 45, 33 },
		{ 313, 46, 33 },
		{ 313, 47, 33 },
		{ 313, 48, 33 },
		{ 313, 49, 33 },
		{ 313, 50, 33 },
		{ 313, 51, 33 },
		{ 313, 52, 33 },
		{ 313, 53, 33 },
		{ 313, 54, 33 },
		{ 313, 55, 33 },
		{ 313, 56, 33 },
		{ 313, 57, 33 },
		{ 313, 58, 33 },
		{ 313, 59, 33 },
		{ 313, 60, 33 },
		{ 313, 61, 33 },
		{ 313, 62, 33 },
		{ 313, 63, 33 },
		{ 313, 64, 33 },
		{ 313, 65, 33 },
		{ 313, 66, 33 },
		{ 313, 67, 33 },
		{ 313, 68, 33 },
		{ 313, 69, 33 },
		{ 313, 70, 33 },
		{ 313, 71, 33 },
		{ 313, 72, 33 },
		{ 313, 73, 33 },
		{ 313, 74, 33 },
		{ 313, 75, 33 },
		{ 313, 76, 33 },
		{ 313, 77, 33 },
		{ 313, 78, 33 },
		{ 313, 79, 33 },
		{ 313, 80, 33 },
		{ 313, 81, 33 },
		{ 313, 82, 33 },
		{ 313, 83, 33 },
		{ 313, 84, 33 },
		{ 313, 85, 33 },
		{ 313, 86, 33 },
		{ 313, 87, 33 },
		{ 313, 88, 33 },
		{ 313, 89, 33 },
		{ 313, 90, 33 },
		{ 313, 91, 33 },
		{ 313, 92, 33 },
		{ 313, 93, 33 },
		{ 313, 94, 33 },
		{ 313, 95, 33 },
		{ 313, 96, 33 },
		{ 313, 97, 33 },
		{ 313, 98, 33 },
		{ 313, 99, 33 },
		{ 313, 100, 33 },
		{ 313, 101, 33 },
		{ 313, 102, 33 },
		{ 313, 103, 33 },
		{ 313, 104, 33 },
		{ 313, 105, 33 },
		{ 313, 106, 33 },
		{ 313, 107, 33 },
		{ 313, 108, 33 },
		{ 313, 109, 33 },
		{ 313, 110, 33 },
		{ 313, 111, 33 },
		{ 313, 112, 33 },
		{ 313, 113, 33 },
		{ 313, 114, 33 },
		{ 313, 115, 33 },
		{ 313, 116, 33 },
		{ 313, 117, 33 },
		{ 313, 118, 33 },
		{ 313, 119, 33 },
		{ 313, 120, 33 },
		{ 313, 121, 33 },
		{ 313, 122, 33 },
		{ 313, 123, 33 },
		{ 313, 124, 33 },
		{ 313, 125, 33 },
		{ 313, 126, 33 },
		{ 313, 127, 33 },
		{ 313, 128, 33 },
		{ 313, 129, 33 },
		{ 313, 130, 33 },
		{ 313, 131, 33 },
		{ 313, 132, 33 },
		{ 313, 133, 33 },
		{ 313, 134, 33 },
		{ 313, 135, 33 },
		{ 313, 136, 33 },
		{ 313, 137, 33 },
		{ 313, 138, 33 },
		{ 313, 139, 33 },
		{ 313, 140, 33 },
		{ 313, 141, 33 },
		{ 313, 142, 33 },
		{ 313, 143, 33 },
		{ 313, 144, 33 },
		{ 314, 30, 33 },
		{ 314, 31, 33 },
		{ 314, 32, 33 },
		{ 314, 33, 33 },
		{ 314, 34, 33 },
		{ 314, 35, 33 },
		{ 314, 36, 33 },
		{ 314, 37, 33 },
		{ 314, 38, 33 },
		{ 314, 39, 33 },
		{ 314, 40, 33 },
		{ 314, 41, 33 },
		{ 314, 42, 33 },
		{ 314, 43, 33 },
		{ 314, 44, 33 },
		{ 314, 45, 33 },
		{ 314, 46, 33 },
		{ 314, 47, 33 },
		{ 314, 48, 33 },
		{ 314, 49, 33 },
		{ 314, 50, 33 },
		{ 314, 51, 33 },
		{ 314, 52, 33 },
		{ 314, 53, 33 },
		{ 314, 54, 33 },
		{ 314, 55, 33 },
		{ 314, 56, 33 },
		{ 314, 57, 33 },
		{ 314, 58, 33 },
		{ 314, 59, 33 },
		{ 314, 60, 33 },
		{ 314, 61, 33 },
		{ 314, 62, 33 },
		{ 314, 63, 33 },
		{ 314, 64, 33 },
		{ 314, 65, 33 },
		{ 314, 66, 33 },
		{ 314, 67, 33 },
		{ 314, 68, 33 },
		{ 314, 69, 33 },
		{ 314, 70, 33 },
		{ 314, 71, 33 },
		{ 314, 72, 33 },
		{ 314, 73, 33 },
		{ 314, 74, 33 },
		{ 314, 75, 33 },
		{ 314, 76, 33 },
		{ 314, 77, 33 },
		{ 314, 78, 33 },
		{ 314, 79, 33 },
		{ 314, 80, 33 },
		{ 314, 81, 33 },
		{ 314, 82, 33 },
		{ 314, 83, 33 },
		{ 314, 84, 33 },
		{ 314, 85, 33 },
		{ 314, 86, 33 },
		{ 314, 87, 33 },
		{ 314, 88, 33 },
		{ 314, 89, 33 },
		{ 314, 90, 33 },
		{ 314, 91, 33 },
		{ 314, 92, 33 },
		{ 314, 93, 33 },
		{ 314, 94, 33 },
		{ 314, 95, 33 },
		{ 314, 96, 33 },
		{ 314, 97, 33 },
		{ 314, 98, 33 },
		{ 314, 99, 33 },
		{ 314, 100, 33 },
		{ 314, 101, 33 },
		{ 314, 102, 33 },
		{ 314, 103, 33 },
		{ 314, 104, 33 },
		{ 314, 105, 33 },
		{ 314, 106, 33 },
		{ 314, 107, 33 },
		{ 314, 108, 33 },
		{ 314, 109, 33 },
		{ 314, 110, 33 },
		{ 314, 111, 33 },
		{ 314, 112, 33 },
		{ 314, 113, 33 },
		{ 314, 114, 33 },
		{ 314, 115, 33 },
		{ 314, 116, 33 },
		{ 314, 117, 33 },
		{ 314, 118, 33 },
		{ 314, 119, 33 },
		{ 314, 120, 33 },
		{ 314, 121, 33 },
		{ 314, 122, 33 },
		{ 314, 123, 33 },
		{ 314, 124, 33 },
		{ 314, 125, 33 },
		{ 314, 126, 33 },
		{ 314, 127, 33 },
		{ 314, 128, 33 },
		{ 314, 129, 33 },
		{ 314, 130, 33 },
		{ 314, 131, 33 },
		{ 314, 132, 33 },
		{ 314, 133, 33 },
		{ 314, 134, 33 },
		{ 314, 135, 33 },
		{ 314, 136, 33 },
		{ 314, 137, 33 },
		{ 314, 138, 33 },
		{ 314, 139, 33 },
		{ 314, 140, 33 },
		{ 314, 141, 33 },
		{ 314, 142, 33 },
		{ 315, 26, 33 },
		{ 315, 27, 33 },
		{ 315, 28, 33 },
		{ 315, 29, 33 },
		{ 315, 30, 33 },
		{ 315, 31, 33 },
		{ 315, 32, 33 },
		{ 315, 33, 33 },
		{ 315, 34, 33 },
		{ 315, 35, 33 },
		{ 315, 36, 33 },
		{ 315, 37, 33 },
		{ 315, 38, 33 },
		{ 315, 39, 33 },
		{ 315, 40, 33 },
		{ 315, 41, 33 },
		{ 315, 42, 33 },
		{ 315, 43, 33 },
		{ 315, 44, 33 },
		{ 315, 45, 33 },
		{ 315, 46, 33 },
		{ 315, 47, 33 },
		{ 315, 48, 33 },
		{ 315, 49, 33 },
		{ 315, 50, 33 },
		{ 315, 51, 33 },
		{ 315, 52, 33 },
		{ 315, 53, 33 },
		{ 315, 54, 33 },
		{ 315, 55, 33 },
		{ 315, 56, 33 },
		{ 315, 57, 33 },
		{ 315, 58, 33 },
		{ 315, 59, 33 },
		{ 315, 60, 33 },
		{ 315, 61, 33 },
		{ 315, 62, 33 },
		{ 315, 63, 33 },
		{ 315, 64, 33 },
		{ 315, 65, 33 },
		{ 315, 66, 33 },
		{ 315, 67, 33 },
		{ 315, 68, 33 },
		{ 315, 69, 33 },
		{ 315, 70, 33 },
		{ 315, 71, 33 },
		{ 315, 72, 33 },
		{ 315, 73, 33 },
		{ 315, 74, 33 },
		{ 315, 75, 33 },
		{ 315, 76, 33 },
		{ 315, 77, 33 },
		{ 315, 78, 33 },
		{ 315, 79, 33 },
		{ 315, 80, 33 },
		{ 315, 81, 33 },
		{ 315, 82, 33 },
		{ 315, 83, 33 },
		{ 315, 84, 33 },
		{ 315, 85, 33 },
		{ 315, 86, 33 },
		{ 315, 87, 33 },
		{ 315, 88, 33 },
		{ 315, 89, 33 },
		{ 315, 90, 33 },
		{ 315, 91, 33 },
		{ 315, 92, 33 },
		{ 315, 93, 33 },
		{ 315, 94, 33 },
		{ 315, 95, 33 },
		{ 315, 96, 33 },
		{ 315, 97, 33 },
		{ 315, 98, 33 },
		{ 315, 99, 33 },
		{ 315, 100, 33 },
		{ 315, 101, 33 },
		{ 315, 102, 33 },
		{ 315, 103, 33 },
		{ 315, 104, 33 },
		{ 315, 105, 33 },
		{ 315, 106, 33 },
		{ 315, 107, 33 },
		{ 315, 108, 33 },
		{ 315, 109, 33 },
		{ 315, 110, 33 },
		{ 315, 111, 33 },
		{ 315, 112, 33 },
		{ 315, 113, 33 },
		{ 315, 114, 33 },
		{ 315, 115, 33 },
		{ 315, 116, 33 },
		{ 315, 117, 33 },
		{ 315, 118, 33 },
		{ 315, 119, 33 },
		{ 315, 120, 33 },
		{ 315, 121, 33 },
		{ 315, 122, 33 },
		{ 315, 123, 33 },
		{ 315, 124, 33 },
		{ 315, 125, 33 },
		{ 315, 126, 33 },
		{ 315, 127, 33 },
		{ 315, 128, 33 },
		{ 315, 129, 33 },
		{ 315, 130, 33 },
		{ 315, 131, 33 },
		{ 315, 132, 33 },
		{ 315, 133, 33 },
		{ 315, 134, 33 },
		{ 315, 135, 33 },
		{ 315, 136, 33 },
		{ 315, 137, 33 },
		{ 315, 138, 33 },
		{ 315, 139, 33 },
		{ 315, 140, 33 },
		{ 316, 22, 33 },
		{ 316, 23, 33 },
		{ 316, 24, 33 },
		{ 316, 25, 33 },
		{ 316, 26, 33 },
		{ 316, 27, 33 },
		{ 316, 28, 33 },
		{ 316, 29, 33 },
		{ 316, 30, 33 },
		{ 316, 31, 33 },
		{ 316, 32, 33 },
		{ 316, 33, 33 },
		{ 316, 34, 33 },
		{ 316, 35, 33 },
		{ 316, 36, 33 },
		{ 316, 37, 33 },
		{ 316, 38, 33 },
		{ 316, 39, 33 },
		{ 316, 40, 33 },
		{ 316, 41, 33 },
		{ 316, 42, 33 },
		{ 316, 43, 33 },
		{ 316, 44, 33 },
		{ 316, 45, 33 },
		{ 316, 46, 33 },
		{ 316, 47, 33 },
		{ 316, 48, 33 },
		{ 316, 49, 33 },
		{ 316, 50, 33 },
		{ 316, 51, 33 },
		{ 316, 52, 33 },
		{ 316, 53, 33 },
		{ 316, 54, 33 },
		{ 316, 55, 33 },
		{ 316, 56, 33 },
		{ 316, 57, 33 },
		{ 316, 58, 33 },
		{ 316, 59, 33 },
		{ 316, 60, 33 },
		{ 316, 61, 33 },
		{ 316, 62, 33 },
		{ 316, 63, 33 },
		{ 316, 64, 33 },
		{ 316, 65, 33 },
		{ 316, 66, 33 },
		{ 316, 67, 33 },
		{ 316, 68, 33 },
		{ 316, 69, 33 },
		{ 316, 70, 33 },
		{ 316, 71, 33 },
		{ 316, 72, 33 },
		{ 316, 73, 33 },
		{ 316, 74, 33 },
		{ 316, 75, 33 },
		{ 316, 76, 33 },
		{ 316, 77, 33 },
		{ 316, 78, 33 },
		{ 316, 79, 33 },
		{ 316, 80, 33 },
		{ 316, 81, 33 },
		{ 316, 82, 33 },
		{ 316, 83, 33 },
		{ 316, 84, 33 },
		{ 316, 85, 33 },
		{ 316, 86, 33 },
		{ 316, 87, 33 },
		{ 316, 88, 33 },
		{ 316, 89, 33 },
		{ 316, 90, 33 },
		{ 316, 91, 33 },
		{ 316, 92, 33 },
		{ 316, 93, 33 },
		{ 316, 94, 33 },
		{ 316, 95, 33 },
		{ 316, 96, 33 },
		{ 316, 97, 33 },
		{ 316, 98, 33 },
		{ 316, 99, 33 },
		{ 316, 100, 33 },
		{ 316, 101, 33 },
		{ 316, 102, 33 },
		{ 316, 103, 33 },
		{ 316, 104, 33 },
		{ 316, 105, 33 },
		{ 316, 106, 33 },
		{ 316, 107, 33 },
		{ 316, 108, 33 },
		{ 316, 109, 33 },
		{ 316, 110, 33 },
		{ 316, 111, 33 },
		{ 316, 112, 33 },
		{ 316, 113, 33 },
		{ 316, 114, 33 },
		{ 316, 115, 33 },
		{ 316, 116, 33 },
		{ 316, 117, 33 },
		{ 316, 118, 33 },
		{ 316, 119, 33 },
		{ 316, 120, 33 },
		{ 316, 121, 33 },
		{ 316, 122, 33 },
		{ 316, 123, 33 },
		{ 316, 124, 33 },
		{ 316, 125, 33 },
		{ 316, 126, 33 },
		{ 316, 127, 33 },
		{ 316, 128, 33 },
		{ 316, 129, 33 },
		{ 316, 130, 33 },
		{ 316, 131, 33 },
		{ 316, 132, 33 },
		{ 316, 133, 33 },
		{ 316, 134, 33 },
		{ 316, 135, 33 },
		{ 316, 136, 33 },
		{ 316, 137, 33 },
		{ 316, 138, 33 },
		{ 317, 18, 33 },
		{ 317, 19, 33 },
		{ 317, 20, 33 },
		{ 317, 21, 33 },
		{ 317, 22, 33 },
		{ 317, 23, 33 },
		{ 317, 24, 33 },
		{ 317, 25, 33 },
		{ 317, 26, 33 },
		{ 317, 27, 33 },
		{ 317, 28, 33 },
		{ 317, 29, 33 },
		{ 317, 30, 33 },
		{ 317, 31, 33 },
		{ 317, 32, 33 },
		{ 317, 33, 33 },
		{ 317, 34, 33 },
		{ 317, 35, 33 },
		{ 317, 36, 33 },
		{ 317, 37, 33 },
		{ 317, 38, 33 },
		{ 317, 39, 33 },
		{ 317, 40, 33 },
		{ 317, 41, 33 },
		{ 317, 42, 33 },
		{ 317, 43, 33 },
		{ 317, 44, 33 },
		{ 317, 45, 33 },
		{ 317, 46, 33 },
		{ 317, 47, 33 },
		{ 317, 48, 33 },
		{ 317, 49, 33 },
		{ 317, 50, 33 },
		{ 317, 51, 33 },
		{ 317, 52, 33 },
		{ 317, 53, 33 },
		{ 317, 54, 33 },
		{ 317, 55, 33 },
		{ 317, 56, 33 },
		{ 317, 57, 33 },
		{ 317, 58, 33 },
		{ 317, 59, 33 },
		{ 317, 60, 33 },
		{ 317, 61, 33 },
		{ 317, 62, 33 },
		{ 317, 63, 33 },
		{ 317, 64, 33 },
		{ 317, 65, 33 },
		{ 317, 66, 33 },
		{ 317, 67, 33 },
		{ 317, 68, 33 },
		{ 317, 69, 33 },
		{ 317, 70, 33 },
		{ 317, 71, 33 },
		{ 317, 72, 33 },
		{ 317, 73, 33 },
		{ 317, 74, 33 },
		{ 317, 75, 33 },
		{ 317, 76, 33 },
		{ 317, 77, 33 },
		{ 317, 78, 33 },
		{ 317, 79, 33 },
		{ 317, 80, 33 },
		{ 317, 81, 33 },
		{ 317, 82, 33 },
		{ 317, 83, 33 },
		{ 317, 84, 33 },
		{ 317, 85, 33 },
		{ 317, 86, 33 },
		{ 317, 87, 33 },
		{ 317, 88, 33 },
		{ 317, 89, 33 },
		{ 317, 90, 33 },
		{ 317, 91, 33 },
		{ 317, 92, 33 },
		{ 317, 93, 33 },
		{ 317, 94, 33 },
		{ 317, 95, 33 },
		{ 317, 96, 33 },
		{ 317, 97, 33 },
		{ 317, 98, 33 },
		{ 317, 99, 33 },
		{ 317, 100, 33 },
		{ 317, 101, 33 },
		{ 317, 102, 33 },
		{ 317, 103, 33 },
		{ 317, 104, 33 },
		{ 317, 105, 33 },
		{ 317, 106, 33 },
		{ 317, 107, 33 },
		{ 317, 108, 33 },
		{ 317, 109, 33 },
		{ 317, 110, 33 },
		{ 317, 111, 33 },
		{ 317, 112, 33 },
		{ 317, 113, 33 },
		{ 317, 114, 33 },
		{ 317, 115, 33 },
		{ 317, 116, 33 },
		{ 317, 117, 33 },
		{ 317, 118, 33 },
		{ 317, 119, 33 },
		{ 317, 120, 33 },
		{ 317, 121, 33 },
		{ 317, 122, 33 },
		{ 317, 123, 33 },
		{ 317, 124, 33 },
		{ 317, 125, 33 },
		{ 317, 126, 33 },
		{ 317, 127, 33 },
		{ 317, 128, 33 },
		{ 317, 129, 33 },
		{ 317, 130, 33 },
		{ 317, 131, 33 },
		{ 317, 132, 33 },
		{ 317, 133, 33 },
		{ 317, 134, 33 },
		{ 317, 135, 33 },
		{ 317, 136, 33 },
		{ 318, 14, 33 },
		{ 318, 15, 33 },
		{ 318, 16, 33 },
		{ 318, 17, 33 },
		{ 318, 18, 33 },
		{ 318, 19, 33 },
		{ 318, 20, 33 },
		{ 318, 21, 33 },
		{ 318, 22, 33 },
		{ 318, 23, 33 },
		{ 318, 24, 33 },
		{ 318, 25, 33 },
		{ 318, 26, 33 },
		{ 318, 27, 33 },
		{ 318, 28, 33 },
		{ 318, 29, 33 },
		{ 318, 30, 33 },
		{ 318, 31, 33 },
		{ 318, 32, 33 },
		{ 318, 33, 33 },
		{ 318, 34, 33 },
		{ 318, 35, 33 },
		{ 318, 36, 33 },
		{ 318, 37, 33 },
		{ 318, 38, 33 },
		{ 318, 39, 33 },
		{ 318, 40, 33 },
		{ 318, 41, 33 },
		{ 318, 42, 33 },
		{ 318, 43, 33 },
		{ 318, 44, 33 },
		{ 318, 45, 33 },
		{ 318, 46, 33 },
		{ 318, 47, 33 },
		{ 318, 48, 33 },
		{ 318, 49, 33 },
		{ 318, 50, 33 },
		{ 318, 51, 33 },
		{ 318, 52, 33 },
		{ 318, 53, 33 },
		{ 318, 54, 33 },
		{ 318, 55, 33 },
		{ 318, 56, 33 },
		{ 318, 57, 33 },
		{ 318, 58, 33 },
		{ 318, 59, 33 },
		{ 318, 60, 33 },
		{ 318, 61, 33 },
		{ 318, 62, 33 },
		{ 318, 63, 33 },
		{ 318, 64, 33 },
		{ 318, 65, 33 },
		{ 318, 66, 33 },
		{ 318, 67, 33 },
		{ 318, 68, 33 },
		{ 318, 69, 33 },
		{ 318, 70, 33 },
		{ 318, 71, 33 },
		{ 318, 72, 33 },
		{ 318, 73, 33 },
		{ 318, 74, 33 },
		{ 318, 75, 33 },
		{ 318, 76, 33 },
		{ 318, 77, 33 },
		{ 318, 78, 33 },
		{ 318, 79, 33 },
		{ 318, 80, 33 },
		{ 318, 81, 33 },
		{ 318, 82, 33 },
		{ 318, 83, 33 },
		{ 318, 84, 33 },
		{ 318, 85, 33 },
		{ 318, 86, 33 },
		{ 318, 87, 33 },
		{ 318, 88, 33 },
		{ 318, 89, 33 },
		{ 318, 90, 33 },
		{ 318, 91, 33 },
		{ 318, 92, 33 },
		{ 318, 93, 33 },
		{ 318, 94, 33 },
		{ 318, 95, 33 },
		{ 318, 96, 33 },
		{ 318, 97, 33 },
		{ 318, 98, 33 },
		{ 318, 99, 33 },
		{ 318, 100, 33 },
		{ 318, 101, 33 },
		{ 318, 102, 33 },
		{ 318, 103, 33 },
		{ 318, 104, 33 },
		{ 318, 105, 33 },
		{ 318, 106, 33 },
		{ 318, 107, 33 },
		{ 318, 108, 33 },
		{ 318, 109, 33 },
		{ 318, 110, 33 },
		{ 318, 111, 33 },
		{ 318, 112, 33 },
		{ 318, 113, 33 },
		{ 318, 114, 33 },
		{ 318, 115, 33 },
		{ 318, 116, 33 },
		{ 318, 117, 33 },
		{ 318, 118, 33 },
		{ 318, 119, 33 },
		{ 318, 120, 33 },
		{ 318, 121, 33 },
		{ 318, 122, 33 },
		{ 318, 123, 33 },
		{ 318, 124, 33 },
		{ 318, 125, 33 },
		{ 318, 126, 33 },
		{ 318, 127, 33 },
		{ 318, 128, 33 },
		{ 318, 129, 33 },
		{ 318, 130, 33 },
		{ 318, 131, 33 },
		{ 318, 132, 33 },
		{ 318, 133, 33 },
		{ 318, 134, 33 },
		{ 319, 10, 33 },
		{ 319, 11, 33 },
		{ 319, 12, 33 },
		{ 319, 13, 33 },
		{ 319, 14, 33 },
		{ 319, 15, 33 },
		{ 319, 16, 33 },
		{ 319, 17, 33 },
		{ 319, 18, 33 },
		{ 319, 19, 33 },
		{ 319, 20, 33 },
		{ 319, 21, 33 },
		{ 319, 22, 33 },
		{ 319, 23, 33 },
		{ 319, 24, 33 },
		{ 319, 25, 33 },
		{ 319, 26, 33 },
		{ 319, 27, 33 },
		{ 319, 28, 33 },
		{ 319, 29, 33 },
		{ 319, 30, 33 },
		{ 319, 31, 33 },
		{ 319, 32, 33 },
		{ 319, 33, 33 },
		{ 319, 34, 33 },
		{ 319, 35, 33 },
		{ 319, 36, 33 },
		{ 319, 37, 33 },
		{ 319, 38, 33 },
		{ 319, 39, 33 },
		{ 319, 40, 33 },
		{ 319, 41, 33 },
		{ 319, 42, 33 },
		{ 319, 43, 33 },
		{ 319, 44, 33 },
		{ 319, 45, 33 },
		{ 319, 46, 33 },
		{ 319, 47, 33 },
		{ 319, 48, 33 },
		{ 319, 49, 33 },
		{ 319, 50, 33 },
		{ 319, 51, 33 },
		{ 319, 52, 33 },
		{ 319, 53, 33 },
		{ 319, 54, 33 },
		{ 319, 55, 33 },
		{ 319, 56, 33 },
		{ 319, 57, 33 },
		{ 319, 58, 33 },
		{ 319, 59, 33 },
		{ 319, 60, 33 },
		{ 319, 61, 33 },
		{ 319, 62, 33 },
		{ 319, 63, 33 },
		{ 319, 64, 33 },
		{ 319, 65, 33 },
		{ 319, 66, 33 },
		{ 319, 67, 33 },
		{ 319, 68, 33 },
		{ 319, 69, 33 },
		{ 319, 70, 33 },
		{ 319, 71, 33 },
		{ 319, 72, 33 },
		{ 319, 73, 33 },
		{ 319, 74, 33 },
		{ 319, 75, 33 },
		{ 319, 76, 33 },
		{ 319, 77, 33 },
		{ 319, 78, 33 },
		{ 319, 79, 33 },
		{ 319, 80, 33 },
		{ 319, 81, 33 },
		{ 319, 82, 33 },
		{ 319, 83, 33 },
		{ 319, 84, 33 },
		{ 319, 85, 33 },
		{ 319, 86, 33 },
		{ 319, 87, 33 },
		{ 319, 88, 33 },
		{ 319, 89, 33 },
		{ 319, 90, 33 },
		{ 319, 91, 33 },
		{ 319, 92, 33 },
		{ 319, 93, 33 },
		{ 319, 94, 33 },
		{ 319, 95, 33 },
		{ 319, 96, 33 },
		{ 319, 97, 33 },
		{ 319, 98, 33 },
		{ 319, 99, 33 },
		{ 319, 100, 33 },
		{ 319, 101, 33 },
		{ 319, 102, 33 },
		{ 319, 103, 33 },
		{ 319, 104, 33 },
		{ 319, 105, 33 },
		{ 319, 106, 33 },
		{ 319, 107, 33 },
		{ 319, 108, 33 },
		{ 319, 109, 33 },
		{ 319, 110, 33 },
		{ 319, 111, 33 },
		{ 319, 112, 33 },
		{ 319, 113, 33 },
		{ 319, 114, 33 },
		{ 319, 115, 33 },
		{ 319, 116, 33 },
		{ 319, 117, 33 },
		{ 319, 118, 33 },
		{ 319, 119, 33 },
		{ 319, 120, 33 },
		{ 319, 121, 33 },
		{ 319, 122, 33 },
		{ 319, 123, 33 },
		{ 319, 124, 33 },
		{ 319, 125, 33 },
		{ 319, 126, 33 },
		{ 319, 127, 33 },
		{ 319, 128, 33 },
		{ 319, 129, 33 },
		{ 319, 130, 33 },
		{ 319, 131, 33 },
		{ 319, 132, 33 },
		{ 320, 6, 33 },
		{ 320, 7, 33 },
		{ 320, 8, 33 },
		{ 320, 9, 33 },
		{ 320, 10, 33 },
		{ 320, 11, 33 },
		{ 320, 12, 33 },
		{ 320, 13, 33 },
		{ 320, 14, 33 },
		{ 320, 15, 33 },
		{ 320, 16, 33 },
		{ 320, 17, 33 },
		{ 320, 18, 33 },
		{ 320, 19, 33 },
		{ 320, 20, 33 },
		{ 320, 21, 33 },
		{ 320, 22, 33 },
		{ 320, 23, 33 },
		{ 320, 24, 33 },
		{ 320, 25, 33 },
		{ 320, 26, 33 },
		{ 320, 27, 33 },
		{ 320, 28, 33 },
		{ 320, 29, 33 },
		{ 320, 30, 33 },
		{ 320, 31, 33 },
		{ 320, 32, 33 },
		{ 320, 33, 33 },
		{ 320, 34, 33 },
		{ 320, 35, 33 },
		{ 320, 36, 33 },
		{ 320, 37, 33 },
		{ 320, 38, 33 },
		{ 320, 39, 33 },
		{ 320, 40, 33 },
		{ 320, 41, 33 },
		{ 320, 42, 33 },
		{ 320, 43, 33 },
		{ 320, 44, 33 },
		{ 320, 45, 33 },
		{ 320, 46, 33 },
		{ 320, 47, 33 },
		{ 320, 48, 33 },
		{ 320, 49, 33 },
		{ 320, 50, 33 },
		{ 320, 51, 33 },
		{ 320, 52, 33 },
		{ 320, 53, 33 },
		{ 320, 54, 33 },
		{ 320, 55, 33 },
		{ 320, 56, 33 },
		{ 320, 57, 33 },
		{ 320, 58, 33 },
		{ 320, 59, 33 },
		{ 320, 60, 33 },
		{ 320, 61, 33 },
		{ 320, 62, 33 },
		{ 320, 63, 33 },
		{ 320, 64, 33 },
		{ 320, 65, 33 },
		{ 320, 66, 33 },
		{ 320, 67, 33 },
		{ 320, 68, 33 },
		{ 320, 69, 33 },
		{ 320, 70, 33 },
		{ 320, 71, 33 },
		{ 320, 72, 33 },
		{ 320, 73, 33 },
		{ 320, 74, 33 },
		{ 320, 75, 33 },
		{ 320, 76, 33 },
		{ 320, 77, 33 },
		{ 320, 78, 33 },
		{ 320, 79, 33 },
		{ 320, 80, 33 },
		{ 320, 81, 33 },
		{ 320, 82, 33 },
		{ 320, 83, 33 },
		{ 320, 84, 33 },
		{ 320, 85, 33 },
		{ 320, 86, 33 },
		{ 320, 87, 33 },
		{ 320, 88, 33 },
		{ 320, 89, 33 },
		{ 320, 90, 33 },
		{ 320, 91, 33 },
		{ 320, 92, 33 },
		{ 320, 93, 33 },
		{ 320, 94, 33 },
		{ 320, 95, 33 },
		{ 320, 96, 33 },
		{ 320, 97, 33 },
		{ 320, 98, 33 },
		{ 320, 99, 33 },
		{ 320, 100, 33 },
		{ 320, 101, 33 },
		{ 320, 102, 33 },
		{ 320, 103, 33 },
		{ 320, 104, 33 },
		{ 320, 105, 33 },
		{ 320, 106, 33 },
		{ 320, 107, 33 },
		{ 320, 108, 33 },
		{ 320, 109, 33 },
		{ 320, 110, 33 },
		{ 320, 111, 33 },
		{ 320, 112, 33 },
		{ 320, 113, 33 },
		{ 320, 114, 33 },
		{ 320, 115, 33 },
		{ 320, 116, 33 },
		{ 320, 117, 33 },
		{ 320, 118, 33 },
		{ 320, 119, 33 },
		{ 320, 120, 33 },
		{ 320, 121, 33 },
		{ 320, 122, 33 },
		{ 320, 123, 33 },
		{ 320, 124, 33 },
		{ 320, 125, 33 },
		{ 320, 126, 33 },
		{ 320, 127, 33 },
		{ 320, 128, 33 },
		{ 320, 129, 33 },
		{ 320, 130, 33 },
		{ 321, 2, 33 },
		{ 321, 3, 33 },
		{ 321, 4, 33 },
		{ 321, 5, 33 },
		{ 321, 6, 33 },
		{ 321, 7, 33 },
		{ 321, 8, 33 },
		{ 321, 9, 33 },
		{ 321, 10, 33 },
		{ 321, 11, 33 },
		{ 321, 12, 33 },
		{ 321, 13, 33 },
		{ 321, 14, 33 },
		{ 321, 15, 33 },
		{ 321, 16, 33 },
		{ 321, 17, 33 },
		{ 321, 18, 33 },
		{ 321, 19, 33 },
		{ 321, 20, 33 },
		{ 321, 21, 33 },
		{ 321, 22, 33 },
		{ 321, 23, 33 },
		{ 321, 24, 33 },
		{ 321, 25, 33 },
		{ 321, 26, 33 },
		{ 321, 27, 33 },
		{ 321, 28, 33 },
		{ 321, 29, 33 },
		{ 321, 30, 33 },
		{ 321, 31, 33 },
		{ 321, 32, 33 },
		{ 321, 33, 33 },
		{ 321, 34, 33 },
		{ 321, 35, 33 },
		{ 321, 36, 33 },
		{ 321, 37, 33 },
		{ 321, 38, 33 },
		{ 321, 39, 33 },
		{ 321, 40, 33 },
		{ 321, 41, 33 },
		{ 321, 42, 33 },
		{ 321, 43, 33 },
		{ 321, 44, 33 },
		{ 321, 45, 33 },
		{ 321, 46, 33 },
		{ 321, 47, 33 },
		{ 321, 48, 33 },
		{ 321, 49, 33 },
		{ 321, 50, 33 },
		{ 321, 51, 33 },
		{ 321, 52, 33 },
		{ 321, 53, 33 },
		{ 321, 54, 33 },
		{ 321, 55, 33 },
		{ 321, 56, 33 },
		{ 321, 57, 33 },
		{ 321, 58, 33 },
		{ 321, 59, 33 },
		{ 321, 60, 33 },
		{ 321, 61, 33 },
		{ 321, 62, 33 },
		{ 321, 63, 33 },
		{ 321, 64, 33 },
		{ 321, 65, 33 },
		{ 321, 66, 33 },
		{ 321, 67, 33 },
		{ 321, 68, 33 },
		{ 321, 69, 33 },
		{ 321, 70, 33 },
		{ 321, 71, 33 },
		{ 321, 72, 33 },
		{ 321, 73, 33 },
		{ 321, 74, 33 },
		{ 321, 75, 33 },
		{ 321, 76, 33 },
		{ 321, 77, 33 },
		{ 321, 78, 33 },
		{ 321, 79, 33 },
		{ 321, 80, 33 },
		{ 321, 81, 33 },
		{ 321, 82, 33 },
		{ 321, 83, 33 },
		{ 321, 84, 33 },
		{ 321, 85, 33 },
		{ 321, 86, 33 },
		{ 321, 87, 33 },
		{ 321, 88, 33 },
		{ 321, 89, 33 },
		{ 321, 90, 33 },
		{ 321, 91, 33 },
		{ 321, 92, 33 },
		{ 321, 93, 33 },
		{ 321, 94, 33 },
		{ 321, 95, 33 },
		{ 321, 96, 33 },
		{ 321, 97, 33 },
		{ 321, 98, 33 },
		{ 321, 99, 33 },
		{ 321, 100, 33 },
		{ 321, 101, 33 },
		{ 321, 102, 33 },
		{ 321, 103, 33 },
		{ 321, 104, 33 },
		{ 321, 105, 33 },
		{ 321, 106, 33 },
		{ 321, 107, 33 },
		{ 321, 108, 33 },
		{ 321, 109, 33 },
		{ 321, 110, 33 },
		{ 321, 111, 33 },
		{ 321, 112, 33 },
		{ 321, 113, 33 },
		{ 321, 114, 33 },
		{ 321, 115, 33 },
		{ 321, 116, 33 },
		{ 321, 117, 33 },
		{ 321, 118, 33 },
		{ 321, 119, 33 },
		{ 321, 120, 33 },
		{ 321, 121, 33 },
		{ 321, 122, 33 },
		{ 321, 123, 33 },
		{ 321, 124, 33 },
		{ 321, 125, 33 },
		{ 321, 126, 33 },
		{ 321, 127, 33 },
		{ 321, 128, 33 },
		{ 322, 1, 33 },
		{ 322, 2, 33 },
		{ 322, 3, 33 },
		{ 322, 4, 33 },
		{ 322, 5, 33 },
		{ 322, 6, 33 },
		{ 322, 7, 33 },
		{ 322, 8, 33 },
		{ 322, 9, 33 },
		{ 322, 10, 33 },
		{ 322, 11, 33 },
		{ 322, 12, 33 },
		{ 322, 13, 33 },
		{ 322, 14, 33 },
		{ 322, 15, 33 },
		{ 322, 16, 33 },
		{ 322, 17, 33 },
		{ 322, 18, 33 },
		{ 322, 19, 33 },
		{ 322, 20, 33 },
		{ 322, 21, 33 },
		{ 322, 22, 33 },
		{ 322, 23, 33 },
		{ 322, 24, 33 },
		{ 322, 25, 33 },
		{ 322, 26, 33 },
		{ 322, 27, 33 },
		{ 322, 28, 33 },
		{ 322, 29, 33 },
		{ 322, 30, 33 },
		{ 322, 31, 33 },
		{ 322, 32, 33 },
		{ 322, 33, 33 },
		{ 322, 34, 33 },
		{ 322, 35, 33 },
		{ 322, 36, 33 },
		{ 322, 37, 33 },
		{ 322, 38, 33 },
		{ 322, 39, 33 },
		{ 322, 40, 33 },
		{ 322, 41, 33 },
		{ 322, 42, 33 },
		{ 322, 43, 33 },
		{ 322, 44, 33 },
		{ 322, 45, 33 },
		{ 322, 46, 33 },
		{ 322, 47, 33 },
		{ 322, 48, 33 },
		{ 322, 49, 33 },
		{ 322, 50, 33 },
		{ 322, 51, 33 },
		{ 322, 52, 33 },
		{ 322, 53, 33 },
		{ 322, 54, 33 },
		{ 322, 55, 33 },
		{ 322, 56, 33 },
		{ 322, 57, 33 },
		{ 322, 58, 33 },
		{ 322, 59, 33 },
		{ 322, 60, 33 },
		{ 322, 61, 33 },
		{ 322, 62, 33 },
		{ 322, 63, 33 },
		{ 322, 64, 33 },
		{ 322, 65, 33 },
		{ 322, 66, 33 },
		{ 322, 67, 33 },
		{ 322, 68, 33 },
		{ 322, 69, 33 },
		{ 322, 70, 33 },
		{ 322, 71, 33 },
		{ 322, 72, 33 },
		{ 322, 73, 33 },
		{ 322, 74, 33 },
		{ 322, 75, 33 },
		{ 322, 76, 33 },
		{ 322, 77, 33 },
		{ 322, 78, 33 },
		{ 322, 79, 33 },
		{ 322, 80, 33 },
		{ 322, 81, 33 },
		{ 322, 82, 33 },
		{ 322, 83, 33 },
		{ 322, 84, 33 },
		{ 322, 85, 33 },
		{ 322, 86, 33 },
		{ 322, 87, 33 },
		{ 322, 88, 33 },
		{ 322, 89, 33 },
		{ 322, 90, 33 },
		{ 322, 91, 33 },
		{ 322, 92, 33 },
		{ 322, 93, 33 },
		{ 322, 94, 33 },
		{ 322, 95, 33 },
		{ 322, 96, 33 },
		{ 322, 97, 33 },
		{ 322, 98, 33 },
		{ 322, 99, 33 },
		{ 322, 100, 33 },
		{ 322, 101, 33 },
		{ 322, 102, 33 },
		{ 322, 103, 33 },
		{ 322, 104, 33 },
		{ 322, 105, 33 },
		{ 322, 106, 33 },
		{ 322, 107, 33 },
		{ 322, 108, 33 },
		{ 322, 109, 33 },
		{ 322, 110, 33 },
		{ 322, 111, 33 },
		{ 322, 112, 33 },
		{ 322, 113, 33 },
		{ 322, 114, 33 },
		{ 322, 115, 33 },
		{ 322, 116, 33 },
		{ 322, 117, 33 },
		{ 322, 118, 33 },
		{ 322, 119, 33 },
		{ 322, 120, 33 },
		{ 322, 121, 33 },
		{ 322, 122, 33 },
		{ 322, 123, 33 },
		{ 322, 124, 33 },
		{ 322, 125, 33 },
		{ 322, 126, 33 },
		{ 323, 1, 33 },
		{ 323, 2, 33 },
		{ 323, 3, 33 },
		{ 323, 4, 33 },
		{ 323, 5, 33 },
		{ 323, 6, 33 },
		{ 323, 7, 33 },
		{ 323, 8, 33 },
		{ 323, 9, 33 },
		{ 323, 10, 33 },
		{ 323, 11, 33 },
		{ 323, 12, 33 },
		{ 323, 13, 33 },
		{ 323, 14, 33 },
		{ 323, 15, 33 },
		{ 323, 16, 33 },
		{ 323, 17, 33 },
		{ 323, 18, 33 },
		{ 323, 19, 33 },
		{ 323, 20, 33 },
		{ 323, 21, 33 },
		{ 323, 22, 33 },
		{ 323, 23, 33 },
		{ 323, 24, 33 },
		{ 323, 25, 33 },
		{ 323, 26, 33 },
		{ 323, 27, 33 },
		{ 323, 28, 33 },
		{ 323, 29, 33 },
		{ 323, 30, 33 },
		{ 323, 31, 33 },
		{ 323, 32, 33 },
		{ 323, 33, 33 },
		{ 323, 34, 33 },
		{ 323, 35, 33 },
		{ 323, 36, 33 },
		{ 323, 37, 33 },
		{ 323, 38, 33 },
		{ 323, 39, 33 },
		{ 323, 40, 33 },
		{ 323, 41, 33 },
		{ 323, 42, 33 },
		{ 323, 43, 33 },
		{ 323, 44, 33 },
		{ 323, 45, 33 },
		{ 323, 46, 33 },
		{ 323, 47, 33 },
		{ 323, 48, 33 },
		{ 323, 49, 33 },
		{ 323, 50, 33 },
		{ 323, 51, 33 },
		{ 323, 52, 33 },
		{ 323, 53, 33 },
		{ 323, 54, 33 },
		{ 323, 55, 33 },
		{ 323, 56, 33 },
		{ 323, 57, 33 },
		{ 323, 58, 33 },
		{ 323, 59, 33 },
		{ 323, 60, 33 },
		{ 323, 61, 33 },
		{ 323, 62, 33 },
		{ 323, 63, 33 },
		{ 323, 64, 33 },
		{ 323, 65, 33 },
		{ 323, 66, 33 },
		{ 323, 67, 33 },
		{ 323, 68, 33 },
		{ 323, 69, 33 },
		{ 323, 70, 33 },
		{ 323, 71, 33 },
		{ 323, 72, 33 },
		{ 323, 73, 33 },
		{ 323, 74, 33 },
		{ 323, 75, 33 },
		{ 323, 76, 33 },
		{ 323, 77, 33 },
		{ 323, 78, 33 },
		{ 323, 79, 33 },
		{ 323, 80, 33 },
		{ 323, 81, 33 },
		{ 323, 82, 33 },
		{ 323, 83, 33 },
		{ 323, 84, 33 },
		{ 323, 85, 33 },
		{ 323, 86, 33 },
		{ 323, 87, 33 },
		{ 323, 88, 33 },
		{ 323, 89, 33 },
		{ 323, 90, 33 },
		{ 323, 91, 33 },
		{ 323, 92, 33 },
		{ 323, 93, 33 },
		{ 323, 94, 33 },
		{ 323, 95, 33 },
		{ 323, 96, 33 },
		{ 323, 97, 33 },
		{ 323, 98, 33 },
		{ 323, 99, 33 },
		{ 323, 100, 33 },
		{ 323, 101, 33 },
		{ 323, 102, 33 },
		{ 323, 103, 33 },
		{ 323, 104, 33 },
		{ 323, 105, 33 },
		{ 323, 106, 33 },
		{ 323, 107, 33 },
		{ 323, 108, 33 },
		{ 323, 109, 33 },
		{ 323, 110, 33 },
		{ 323, 111, 33 },
		{ 323, 112, 33 },
		{ 323, 113, 33 },
		{ 323, 114, 33 },
		{ 323, 115, 33 },
		{ 323, 116, 33 },
		{ 323, 117, 33 },
		{ 323, 118, 33 },
		{ 323, 119, 33 },
		{ 323, 120, 33 },
		{ 323, 121, 33 },
		{ 323, 122, 33 },
		{ 323, 123, 33 },
		{ 323, 124, 33 },
		{ 324, 1, 33 },
		{ 324, 2, 33 },
		{ 324, 3, 33 },
		{ 324, 4, 33 },
		{ 324, 5, 33 },
		{ 324, 6, 33 },
		{ 324, 7, 33 },
		{ 324, 8, 33 },
		{ 324, 9, 33 },
		{ 324, 10, 33 },
		{ 324, 11, 33 },
		{ 324, 12, 33 },
		{ 324, 13, 33 },
		{ 324, 14, 33 },
		{ 324, 15, 33 },
		{ 324, 16, 33 },
		{ 324, 17, 33 },
		{ 324, 18, 33 },
		{ 324, 19, 33 },
		{ 324, 20, 33 },
		{ 324, 21, 33 },
		{ 324, 22, 33 },
		{ 324, 23, 33 },
		{ 324, 24, 33 },
		{ 324, 25, 33 },
		{ 324, 26, 33 },
		{ 324, 27, 33 },
		{ 324, 28, 33 },
		{ 324, 29, 33 },
		{ 324, 30, 33 },
		{ 324, 31, 33 },
		{ 324, 32, 33 },
		{ 324, 33, 33 },
		{ 324, 34, 33 },
		{ 324, 35, 33 },
		{ 324, 36, 33 },
		{ 324, 37, 33 },
		{ 324, 38, 33 },
		{ 324, 39, 33 },
		{ 324, 40, 33 },
		{ 324, 41, 33 },
		{ 324, 42, 33 },
		{ 324, 43, 33 },
		{ 324, 44, 33 },
		{ 324, 45, 33 },
		{ 324, 46, 33 },
		{ 324, 47, 33 },
		{ 324, 48, 33 },
		{ 324, 49, 33 },
		{ 324, 50, 33 },
		{ 324, 51, 33 },
		{ 324, 52, 33 },
		{ 324, 53, 33 },
		{ 324, 54, 33 },
		{ 324, 55, 33 },
		{ 324, 56, 33 },
		{ 324, 57, 33 },
		{ 324, 58, 33 },
		{ 324, 59, 33 },
		{ 324, 60, 33 },
		{ 324, 61, 33 },
		{ 324, 62, 33 },
		{ 324, 63, 33 },
		{ 324, 64, 33 },
		{ 324, 65, 33 },
		{ 324, 66, 33 },
		{ 324, 67, 33 },
		{ 324, 68, 33 },
		{ 324, 69, 33 },
		{ 324, 70, 33 },
		{ 324, 71, 33 },
		{ 324, 72, 33 },
		{ 324, 73, 33 },
		{ 324, 74, 33 },
		{ 324, 75, 33 },
		{ 324, 76, 33 },
		{ 324, 77, 33 },
		{ 324, 78, 33 },
		{ 324, 79, 33 },
		{ 324, 80, 33 },
		{ 324, 81, 33 },
		{ 324, 82, 33 },
		{ 324, 83, 33 },
		{ 324, 84, 33 },
		{ 324, 85, 33 },
		{ 324, 86, 33 },
		{ 324, 87, 33 },
		{ 324, 88, 33 },
		{ 324, 89, 33 },
		{ 324, 90, 33 },
		{ 324, 91, 33 },
		{ 324, 92, 33 },
		{ 324, 93, 33 },
		{ 324, 94, 33 },
		{ 324, 95, 33 },
		{ 324, 96, 33 },
		{ 324, 97, 33 },
		{ 324, 98, 33 },
		{ 324, 99, 33 },
		{ 324, 100, 33 },
		{ 324, 101, 33 },
		{ 324, 102, 33 },
		{ 324, 103, 33 },
		{ 324, 104, 33 },
		{ 324, 105, 33 },
		{ 324, 106, 33 },
		{ 324, 107, 33 },
		{ 324, 108, 33 },
		{ 324, 109, 33 },
		{ 324, 110, 33 },
		{ 324, 111, 33 },
		{ 324, 112, 33 },
		{ 324, 113, 33 },
		{ 324, 114, 33 },
		{ 324, 115, 33 },
		{ 324, 116, 33 },
		{ 324, 117, 33 },
		{ 324, 118, 33 },
		{ 324, 119, 33 },
		{ 324, 120, 33 },
		{ 324, 121, 33 },
		{ 324, 122, 33 },
		{ 325, 1, 33 },
		{ 325, 2, 33 },
		{ 325, 3, 33 },
		{ 325, 4, 33 },
		{ 325, 5, 33 },
		{ 325, 6, 33 },
		{ 325, 7, 33 },
		{ 325, 8, 33 },
		{ 325, 9, 33 },
		{ 325, 10, 33 },
		{ 325, 11, 33 },
		{ 325, 12, 33 },
		{ 325, 13, 33 },
		{ 325, 14, 33 },
		{ 325, 15, 33 },
		{ 325, 16, 33 },
		{ 325, 17, 33 },
		{ 325, 18, 33 },
		{ 325, 19, 33 },
		{ 325, 20, 33 },
		{ 325, 21, 33 },
		{ 325, 22, 33 },
		{ 325, 23, 33 },
		{ 325, 24, 33 },
		{ 325, 25, 33 },
		{ 325, 26, 33 },
		{ 325, 27, 33 },
		{ 325, 28, 33 },
		{ 325, 29, 33 },
		{ 325, 30, 33 },
		{ 325, 31, 33 },
		{ 325, 32, 33 },
		{ 325, 33, 33 },
		{ 325, 34, 33 },
		{ 325, 35, 33 },
		{ 325, 36, 33 },
		{ 325, 37, 33 },
		{ 325, 38, 33 },
		{ 325, 39, 33 },
		{ 325, 40, 33 },
		{ 325, 41, 33 },
		{ 325, 42, 33 },
		{ 325, 43, 33 },
		{ 325, 44, 33 },
		{ 325, 45, 33 },
		{ 325, 46, 33 },
		{ 325, 47, 33 },
		{ 325, 48, 33 },
		{ 325, 49, 33 },
		{ 325, 50, 33 },
		{ 325, 51, 33 },
		{ 325, 52, 33 },
		{ 325, 53, 33 },
		{ 325, 54, 33 },
		{ 325, 55, 33 },
		{ 325, 56, 33 },
		{ 325, 57, 33 },
		{ 325, 58, 33 },
		{ 325, 59, 33 },
		{ 325, 60, 33 },
		{ 325, 61, 33 },
		{ 325, 62, 33 },
		{ 325, 63, 33 },
		{ 325, 64, 33 },
		{ 325, 65, 33 },
		{ 325, 66, 33 },
		{ 325, 67, 33 },
		{ 325, 68, 33 },
		{ 325, 69, 33 },
		{ 325, 70, 33 },
		{ 325, 71, 33 },
		{ 325, 72, 33 },
		{ 325, 73, 33 },
		{ 325, 74, 33 },
		{ 325, 75, 33 },
		{ 325, 76, 33 },
		{ 325, 77, 33 },
		{ 325, 78, 33 },
		{ 325, 79, 33 },
		{ 325, 80, 33 },
		{ 325, 81, 33 },
		{ 325, 82, 33 },
		{ 325, 83, 33 },
		{ 325, 84, 33 },
		{ 325, 85, 33 },
		{ 325, 86, 33 },
		{ 325, 87, 33 },
		{ 325, 88, 33 },
		{ 325, 89, 33 },
		{ 325, 90, 33 },
		{ 325, 91, 33 },
		{ 325, 92, 33 },
		{ 325, 93, 33 },
		{ 325, 94, 33 },
		{ 325, 95, 33 },
		{ 325, 96, 33 },
		{ 325, 97, 33 },
		{ 325, 98, 33 },
		{ 325, 99, 33 },
		{ 325, 100, 33 },
		{ 325, 101, 33 },
		{ 325, 102, 33 },
		{ 325, 103, 33 },
		{ 325, 104, 33 },
		{ 325, 105, 33 },
		{ 325, 106, 33 },
		{ 325, 107, 33 },
		{ 325, 108, 33 },
		{ 325, 109, 33 },
		{ 325, 110, 33 },
		{ 325, 111, 33 },
		{ 325, 112, 33 },
		{ 325, 113, 33 },
		{ 325, 114, 33 },
		{ 325, 115, 33 },
		{ 325, 116, 33 },
		{ 325, 117, 33 },
		{ 325, 118, 33 },
		{ 325, 119, 33 },
		{ 325, 120, 33 },
		{ 326, 1, 33 },
		{ 326, 2, 33 },
		{ 326, 3, 33 },
		{ 326, 4, 33 },
		{ 326, 5, 33 },
		{ 326, 6, 33 },
		{ 326, 7, 33 },
		{ 326, 8, 33 },
		{ 326, 9, 33 },
		{ 326, 10, 33 },
		{ 326, 11, 33 },
		{ 326, 12, 33 },
		{ 326, 13, 33 },
		{ 326, 14, 33 },
		{ 326, 15, 33 },
		{ 326, 16, 33 },
		{ 326, 17, 33 },
		{ 326, 18, 33 },
		{ 326, 19, 33 },
		{ 326, 20, 33 },
		{ 326, 21, 33 },
		{ 326, 22, 33 },
		{ 326, 23, 33 },
		{ 326, 24, 33 },
		{ 326, 25, 33 },
		{ 326, 26, 33 },
		{ 326, 27, 33 },
		{ 326, 28, 33 },
		{ 326, 29, 33 },
		{ 326, 30, 33 },
		{ 326, 31, 33 },
		{ 326, 32, 33 },
		{ 326, 33, 33 },
		{ 326, 34, 33 },
		{ 326, 35, 33 },
		{ 326, 36, 33 },
		{ 326, 37, 33 },
		{ 326, 38, 33 },
		{ 326, 39, 33 },
		{ 326, 40, 33 },
		{ 326, 41, 33 },
		{ 326, 42, 33 },
		{ 326, 43, 33 },
		{ 326, 44, 33 },
		{ 326, 45, 33 },
		{ 326, 46, 33 },
		{ 326, 47, 33 },
		{ 326, 48, 33 },
		{ 326, 49, 33 },
		{ 326, 50, 33 },
		{ 326, 51, 33 },
		{ 326, 52, 33 },
		{ 326, 53, 33 },
		{ 326, 54, 33 },
		{ 326, 55, 33 },
		{ 326, 56, 33 },
		{ 326, 57, 33 },
		{ 326, 58, 33 },
		{ 326, 59, 33 },
		{ 326, 60, 33 },
		{ 326, 61, 33 },
		{ 326, 62, 33 },
		{ 326, 63, 33 },
		{ 326, 64, 33 },
		{ 326, 65, 33 },
		{ 326, 66, 33 },
		{ 326, 67, 33 },
		{ 326, 68, 33 },
		{ 326, 69, 33 },
		{ 326, 70, 33 },
		{ 326, 71, 33 },
		{ 326, 72, 33 },
		{ 326, 73, 33 },
		{ 326, 74, 33 },
		{ 326, 75, 33 },
		{ 326, 76, 33 },
		{ 326, 77, 33 },
		{ 326, 78, 33 },
		{ 326, 79, 33 },
		{ 326, 80, 33 },
		{ 326, 81, 33 },
		{ 326, 82, 33 },
		{ 326, 83, 33 },
		{ 326, 84, 33 },
		{ 326, 85, 33 },
		{ 326, 86, 33 },
		{ 326, 87, 33 },
		{ 326, 88, 33 },
		{ 326, 89, 33 },
		{ 326, 90, 33 },
		{ 326, 91, 33 },
		{ 326, 92, 33 },
		{ 326, 93, 33 },
		{ 326, 94, 33 },
		{ 326, 95, 33 },
		{ 326, 96, 33 },
		{ 326, 97, 33 },
		{ 326, 98, 33 },
		{ 326, 99, 33 },
		{ 326, 100, 33 },
		{ 326, 101, 33 },
		{ 326, 102, 33 },
		{ 326, 103, 33 },
		{ 326, 104, 33 },
		{ 326, 105, 33 },
		{ 326, 106, 33 },
		{ 326, 107, 33 },
		{ 326, 108, 33 },
		{ 326, 109, 33 },
		{ 326, 110, 33 },
		{ 326, 111, 33 },
		{ 326, 112, 33 },
		{ 326, 113, 33 },
		{ 326, 114, 33 },
		{ 326, 115, 33 },
		{ 326, 116, 33 },
		{ 326, 117, 33 },
		{ 326, 118, 33 },
		{ 327, 1, 33 },
		{ 327, 2, 33 },
		{ 327, 3, 33 },
		{ 327, 4, 33 },
		{ 327, 5, 33 },
		{ 327, 6, 33 },
		{ 327, 7, 33 },
		{ 327, 8, 33 },
		{ 327, 9, 33 },
		{ 327, 10, 33 },
		{ 327, 11, 33 },
		{ 327, 12, 33 },
		{ 327, 13, 33 },
		{ 327, 14, 33 },
		{ 327, 15, 33 },
		{ 327, 16, 33 },
		{ 327, 17, 33 },
		{ 327, 18, 33 },
		{ 327, 19, 33 },
		{ 327, 20, 33 },
		{ 327, 21, 33 },
		{ 327, 22, 33 },
		{ 327, 23, 33 },
		{ 327, 24, 33 },
		{ 327, 25, 33 },
		{ 327, 26, 33 },
		{ 327, 27, 33 },
		{ 327, 28, 33 },
		{ 327, 29, 33 },
		{ 327, 30, 33 },
		{ 327, 31, 33 },
		{ 327, 32, 33 },
		{ 327, 33, 33 },
		{ 327, 34, 33 },
		{ 327, 35, 33 },
		{ 327, 36, 33 },
		{ 327, 37, 33 },
		{ 327, 38, 33 },
		{ 327, 39, 33 },
		{ 327, 40, 33 },
		{ 327, 41, 33 },
		{ 327, 42, 33 },
		{ 327, 43, 33 },
		{ 327, 44, 33 },
		{ 327, 45, 33 },
		{ 327, 46, 33 },
		{ 327, 47, 33 },
		{ 327, 48, 33 },
		{ 327, 49, 33 },
		{ 327, 50, 33 },
		{ 327, 51, 33 },
		{ 327, 52, 33 },
		{ 327, 53, 33 },
		{ 327, 54, 33 },
		{ 327, 55, 33 },
		{ 327, 56, 33 },
		{ 327, 57, 33 },
		{ 327, 58, 33 },
		{ 327, 59, 33 },
		{ 327, 60, 33 },
		{ 327, 61, 33 },
		{ 327, 62, 33 },
		{ 327, 63, 33 },
		{ 327, 64, 33 },
		{ 327, 65, 33 },
		{ 327, 66, 33 },
		{ 327, 67, 33 },
		{ 327, 68, 33 },
		{ 327, 69, 33 },
		{ 327, 70, 33 },
		{ 327, 71, 33 },
		{ 327, 72, 33 },
		{ 327, 73, 33 },
		{ 327, 74, 33 },
		{ 327, 75, 33 },
		{ 327, 76, 33 },
		{ 327, 77, 33 },
		{ 327, 78, 33 },
		{ 327, 79, 33 },
		{ 327, 80, 33 },
		{ 327, 81, 33 },
		{ 327, 82, 33 },
		{ 327, 83, 33 },
		{ 327, 84, 33 },
		{ 327, 85, 33 },
		{ 327, 86, 33 },
		{ 327, 87, 33 },
		{ 327, 88, 33 },
		{ 327, 89, 33 },
		{ 327, 90, 33 },
		{ 327, 91, 33 },
		{ 327, 92, 33 },
		{ 327, 93, 33 },
		{ 327, 94, 33 },
		{ 327, 95, 33 },
		{ 327, 96, 33 },
		{ 327, 97, 33 },
		{ 327, 98, 33 },
		{ 327, 99, 33 },
		{ 327, 100, 33 },
		{ 327, 101, 33 },
		{ 327, 102, 33 },
		{ 327, 103, 33 },
		{ 327, 104, 33 },
		{ 327, 105, 33 },
		{ 327, 106, 33 },
		{ 327, 107, 33 },
		{ 327, 108, 33 },
		{ 327, 109, 33 },
		{ 327, 110, 33 },
		{ 327, 111, 33 },
		{ 327, 112, 33 },
		{ 327, 113, 33 },
		{ 327, 114, 33 },
		{ 327, 115, 33 },
		{ 327, 116, 33 },
		{ 328, 1, 33 },
		{ 328, 2, 33 },
		{ 328, 3, 33 },
		{ 328, 4, 33 },
		{ 328, 5, 33 },
		{ 328, 6, 33 },
		{ 328, 7, 33 },
		{ 328, 8, 33 },
		{ 328, 9, 33 },
		{ 328, 10, 33 },
		{ 328, 11, 33 },
		{ 328, 12, 33 },
		{ 328, 13, 33 },
		{ 328, 14, 33 },
		{ 328, 15, 33 },
		{ 328, 16, 33 },
		{ 328, 17, 33 },
		{ 328, 18, 33 },
		{ 328, 19, 33 },
		{ 328, 20, 33 },
		{ 328, 21, 33 },
		{ 328, 22, 33 },
		{ 328, 23, 33 },
		{ 328, 24, 33 },
		{ 328, 25, 33 },
		{ 328, 26, 33 },
		{ 328, 27, 33 },
		{ 328, 28, 33 },
		{ 328, 29, 33 },
		{ 328, 30, 33 },
		{ 328, 31, 33 },
		{ 328, 32, 33 },
		{ 328, 33, 33 },
		{ 328, 34, 33 },
		{ 328, 35, 33 },
		{ 328, 36, 33 },
		{ 328, 37, 33 },
		{ 328, 38, 33 },
		{ 328, 39, 33 },
		{ 328, 40, 33 },
		{ 328, 41, 33 },
		{ 328, 42, 33 },
		{ 328, 43, 33 },
		{ 328, 44, 33 },
		{ 328, 45, 33 },
		{ 328, 46, 33 },
		{ 328, 47, 33 },
		{ 328, 48, 33 },
		{ 328, 49, 33 },
		{ 328, 50, 33 },
		{ 328, 51, 33 },
		{ 328, 52, 33 },
		{ 328, 53, 33 },
		{ 328, 54, 33 },
		{ 328, 55, 33 },
		{ 328, 56, 33 },
		{ 328, 57, 33 },
		{ 328, 58, 33 },
		{ 328, 59, 33 },
		{ 328, 60, 33 },
		{ 328, 61, 33 },
		{ 328, 62, 33 },
		{ 328, 63, 33 },
		{ 328, 64, 33 },
		{ 328, 65, 33 },
		{ 328, 66, 33 },
		{ 328, 67, 33 },
		{ 328, 68, 33 },
		{ 328, 69, 33 },
		{ 328, 70, 33 },
		{ 328, 71, 33 },
		{ 328, 72, 33 },
		{ 328, 73, 33 },
		{ 328, 74, 33 },
		{ 328, 75, 33 },
		{ 328, 76, 33 },
		{ 328, 77, 33 },
		{ 328, 78, 33 },
		{ 328, 79, 33 },
		{ 328, 80, 33 },
		{ 328, 81, 33 },
		{ 328, 82, 33 },
		{ 328, 83, 33 },
		{ 328, 84, 33 },
		{ 328, 85, 33 },
		{ 328, 86, 33 },
		{ 328, 87, 33 },
		{ 328, 88, 33 },
		{ 328, 89, 33 },
		{ 328, 90, 33 },
		{ 328, 91, 33 },
		{ 328, 92, 33 },
		{ 328, 93, 33 },
		{ 328, 94, 33 },
		{ 328, 95, 33 },
		{ 328, 96, 33 },
		{ 328, 97, 33 },
		{ 328, 98, 33 },
		{ 328, 99, 33 },
		{ 328, 100, 33 },
		{ 328, 101, 33 },
		{ 328, 102, 33 },
		{ 328, 103, 33 },
		{ 328, 104, 33 },
		{ 328, 105, 33 },
		{ 328, 106, 33 },
		{ 328, 107, 33 },
		{ 328, 108, 33 },
		{ 328, 109, 33 },
		{ 328, 110, 33 },
		{ 328, 111, 33 },
		{ 328, 112, 33 },
		{ 328, 113, 33 },
		{ 328, 114, 33 },
		{ 329, 1, 33 },
		{ 329, 2, 33 },
		{ 329, 3, 33 },
		{ 329, 4, 33 },
		{ 329, 5, 33 },
		{ 329, 6, 33 },
		{ 329, 7, 33 },
		{ 329, 8, 33 },
		{ 329, 9, 33 },
		{ 329, 10, 33 },
		{ 329, 11, 33 },
		{ 329, 12, 33 },
		{ 329, 13, 33 },
		{ 329, 14, 33 },
		{ 329, 15, 33 },
		{ 329, 16, 33 },
		{ 329, 17, 33 },
		{ 329, 18, 33 },
		{ 329, 19, 33 },
		{ 329, 20, 33 },
		{ 329, 21, 33 },
		{ 329, 22, 33 },
		{ 329, 23, 33 },
		{ 329, 24, 33 },
		{ 329, 25, 33 },
		{ 329, 26, 33 },
		{ 329, 27, 33 },
		{ 329, 28, 33 },
		{ 329, 29, 33 },
		{ 329, 30, 33 },
		{ 329, 31, 33 },
		{ 329, 32, 33 },
		{ 329, 33, 33 },
		{ 329, 34, 33 },
		{ 329, 35, 33 },
		{ 329, 36, 33 },
		{ 329, 37, 33 },
		{ 329, 38, 33 },
		{ 329, 39, 33 },
		{ 329, 40, 33 },
		{ 329, 41, 33 },
		{ 329, 42, 33 },
		{ 329, 43, 33 },
		{ 329, 44, 33 },
		{ 329, 45, 33 },
		{ 329, 46, 33 },
		{ 329, 47, 33 },
		{ 329, 48, 33 },
		{ 329, 49, 33 },
		{ 329, 50, 33 },
		{ 329, 51, 33 },
		{ 329, 52, 33 },
		{ 329, 53, 33 },
		{ 329, 54, 33 },
		{ 329, 55, 33 },
		{ 329, 56, 33 },
		{ 329, 57, 33 },
		{ 329, 58, 33 },
		{ 329, 59, 33 },
		{ 329, 60, 33 },
		{ 329, 61, 33 },
		{ 329, 62, 33 },
		{ 329, 63, 33 },
		{ 329, 64, 33 },
		{ 329, 65, 33 },
		{ 329, 66, 33 },
		{ 329, 67, 33 },
		{ 329, 68, 33 },
		{ 329, 69, 33 },
		{ 329, 70, 33 },
		{ 329, 71, 33 },
		{ 329, 72, 33 },
		{ 329, 73, 33 },
		{ 329, 74, 33 },
		{ 329, 75, 33 },
		{ 329, 76, 33 },
		{ 329, 77, 33 },
		{ 329, 78, 33 },
		{ 329, 79, 33 },
		{ 329, 80, 33 },
		{ 329, 81, 33 },
		{ 329, 82, 33 },
		{ 329, 83, 33 },
		{ 329, 84, 33 },
		{ 329, 85, 33 },
		{ 329, 86, 33 },
		{ 329, 87, 33 },
		{ 329, 88, 33 },
		{ 329, 89, 33 },
		{ 329, 90, 33 },
		{ 329, 91, 33 },
		{ 329, 92, 33 },
		{ 329, 93, 33 },
		{ 329, 94, 33 },
		{ 329, 95, 33 },
		{ 329, 96, 33 },
		{ 329, 97, 33 },
		{ 329, 98, 33 },
		{ 329, 99, 33 },
		{ 329, 100, 33 },
		{ 329, 101, 33 },
		{ 329, 102, 33 },
		{ 329, 103, 33 },
		{ 329, 104, 33 },
		{ 329, 105, 33 },
		{ 329, 106, 33 },
		{ 329, 107, 33 },
		{ 329, 108, 33 },
		{ 329, 109, 33 },
		{ 329, 110, 33 },
		{ 329, 111, 33 },
		{ 329, 112, 33 },
		{ 330, 1, 33 },
		{ 330, 2, 33 },
		{ 330, 3, 33 },
		{ 330, 4, 33 },
		{ 330, 5, 33 },
		{ 330, 6, 33 },
		{ 330, 7, 33 },
		{ 330, 8, 33 },
		{ 330, 9, 33 },
		{ 330, 10, 33 },
		{ 330, 11, 33 },
		{ 330, 12, 33 },
		{ 330, 13, 33 },
		{ 330, 14, 33 },
		{ 330, 15, 33 },
		{ 330, 16, 33 },
		{ 330, 17, 33 },
		{ 330, 18, 33 },
		{ 330, 19, 33 },
		{ 330, 20, 33 },
		{ 330, 21, 33 },
		{ 330, 22, 33 },
		{ 330, 23, 33 },
		{ 330, 24, 33 },
		{ 330, 25, 33 },
		{ 330, 26, 33 },
		{ 330, 27, 33 },
		{ 330, 28, 33 },
		{ 330, 29, 33 },
		{ 330, 30, 33 },
		{ 330, 31, 33 },
		{ 330, 32, 33 },
		{ 330, 33, 33 },
		{ 330, 34, 33 },
		{ 330, 35, 33 },
		{ 330, 36, 33 },
		{ 330, 37, 33 },
		{ 330, 38, 33 },
		{ 330, 39, 33 },
		{ 330, 40, 33 },
		{ 330, 41, 33 },
		{ 330, 42, 33 },
		{ 330, 43, 33 },
		{ 330, 44, 33 },
		{ 330, 45, 33 },
		{ 330, 46, 33 },
		{ 330, 47, 33 },
		{ 330, 48, 33 },
		{ 330, 49, 33 },
		{ 330, 50, 33 },
		{ 330, 51, 33 },
		{ 330, 52, 33 },
		{ 330, 53, 33 },
		{ 330, 54, 33 },
		{ 330, 55, 33 },
		{ 330, 56, 33 },
		{ 330, 57, 33 },
		{ 330, 58, 33 },
		{ 330, 59, 33 },
		{ 330, 60, 33 },
		{ 330, 61, 33 },
		{ 330, 62, 33 },
		{ 330, 63, 33 },
		{ 330, 64, 33 },
		{ 330, 65, 33 },
		{ 330, 66, 33 },
		{ 330, 67, 33 },
		{ 330, 68, 33 },
		{ 330, 69, 33 },
		{ 330, 70, 33 },
		{ 330, 71, 33 },
		{ 330, 72, 33 },
		{ 330, 73, 33 },
		{ 330, 74, 33 },
		{ 330, 75, 33 },
		{ 330, 76, 33 },
		{ 330, 77, 33 },
		{ 330, 78, 33 },
		{ 330, 79, 33 },
		{ 330, 80, 33 },
		{ 330, 81, 33 },
		{ 330, 82, 33 },
		{ 330, 83, 33 },
		{ 330, 84, 33 },
		{ 330, 85, 33 },
		{ 330, 86, 33 },
		{ 330, 87, 33 },
		{ 330, 88, 33 },
		{ 330, 89, 33 },
		{ 330, 90, 33 },
		{ 330, 91, 33 },
		{ 330, 92, 33 },
		{ 330, 93, 33 },
		{ 330, 94, 33 },
		{ 330, 95, 33 },
		{ 330, 96, 33 },
		{ 330, 97, 33 },
		{ 330, 98, 33 },
		{ 330, 99, 33 },
		{ 330, 100, 33 },
		{ 330, 101, 33 },
		{ 330, 102, 33 },
		{ 330, 103, 33 },
		{ 330, 104, 33 },
		{ 330, 105, 33 },
		{ 330, 106, 33 },
		{ 330, 107, 33 },
		{ 330, 108, 33 },
		{ 330, 109, 33 },
		{ 330, 110, 33 },
		{ 331, 1, 33 },
		{ 331, 2, 33 },
		{ 331, 3, 33 },
		{ 331, 4, 33 },
		{ 331, 5, 33 },
		{ 331, 6, 33 },
		{ 331, 7, 33 },
		{ 331, 8, 33 },
		{ 331, 9, 33 },
		{ 331, 10, 33 },
		{ 331, 11, 33 },
		{ 331, 12, 33 },
		{ 331, 13, 33 },
		{ 331, 14, 33 },
		{ 331, 15, 33 },
		{ 331, 16, 33 },
		{ 331, 17, 33 },
		{ 331, 18, 33 },
		{ 331, 19, 33 },
		{ 331, 20, 33 },
		{ 331, 21, 33 },
		{ 331, 22, 33 },
		{ 331, 23, 33 },
		{ 331, 24, 33 },
		{ 331, 25, 33 },
		{ 331, 26, 33 },
		{ 331, 27, 33 },
		{ 331, 28, 33 },
		{ 331, 29, 33 },
		{ 331, 30, 33 },
		{ 331, 31, 33 },
		{ 331, 32, 33 },
		{ 331, 33, 33 },
		{ 331, 34, 33 },
		{ 331, 35, 33 },
		{ 331, 36, 33 },
		{ 331, 37, 33 },
		{ 331, 38, 33 },
		{ 331, 39, 33 },
		{ 331, 40, 33 },
		{ 331, 41, 33 },
		{ 331, 42, 33 },
		{ 331, 43, 33 },
		{ 331, 44, 33 },
		{ 331, 45, 33 },
		{ 331, 46, 33 },
		{ 331, 47, 33 },
		{ 331, 48, 33 },
		{ 331, 49, 33 },
		{ 331, 50, 33 },
		{ 331, 51, 33 },
		{ 331, 52, 33 },
		{ 331, 53, 33 },
		{ 331, 54, 33 },
		{ 331, 55, 33 },
		{ 331, 56, 33 },
		{ 331, 57, 33 },
		{ 331, 58, 33 },
		{ 331, 59, 33 },
		{ 331, 60, 33 },
		{ 331, 61, 33 },
		{ 331, 62, 33 },
		{ 331, 63, 33 },
		{ 331, 64, 33 },
		{ 331, 65, 33 },
		{ 331, 66, 33 },
		{ 331, 67, 33 },
		{ 331, 68, 33 },
		{ 331, 69, 33 },
		{ 331, 70, 33 },
		{ 331, 71, 33 },
		{ 331, 72, 33 },
		{ 331, 73, 33 },
		{ 331, 74, 33 },
		{ 331, 75, 33 },
		{ 331, 76, 33 },
		{ 331, 77, 33 },
		{ 331, 78, 33 },
		{ 331, 79, 33 },
		{ 331, 80, 33 },
		{ 331, 81, 33 },
		{ 331, 82, 33 },
		{ 331, 83, 33 },
		{ 331, 84, 33 },
		{ 331, 85, 33 },
		{ 331, 86, 33 },
		{ 331, 87, 33 },
		{ 331, 88, 33 },
		{ 331, 89, 33 },
		{ 331, 90, 33 },
		{ 331, 91, 33 },
		{ 331, 92, 33 },
		{ 331, 93, 33 },
		{ 331, 94, 33 },
		{ 331, 95, 33 },
		{ 331, 96, 33 },
		{ 331, 97, 33 },
		{ 331, 98, 33 },
		{ 331, 99, 33 },
		{ 331, 100, 33 },
		{ 331, 101, 33 },
		{ 331, 102, 33 },
		{ 331, 103, 33 },
		{ 331, 104, 33 },
		{ 331, 105, 33 },
		{ 331, 106, 33 },
		{ 331, 107, 33 },
		{ 331, 108, 33 },
		{ 332, 1, 33 },
		{ 332, 2, 33 },
		{ 332, 3, 33 },
		{ 332, 4, 33 },
		{ 332, 5, 33 },
		{ 332, 6, 33 },
		{ 332, 7, 33 },
		{ 332, 8, 33 },
		{ 332, 9, 33 },
		{ 332, 10, 33 },
		{ 332, 11, 33 },
		{ 332, 12, 33 },
		{ 332, 13, 33 },
		{ 332, 14, 33 },
		{ 332, 15, 33 },
		{ 332, 16, 33 },
		{ 332, 17, 33 },
		{ 332, 18, 33 },
		{ 332, 19, 33 },
		{ 332, 20, 33 },
		{ 332, 21, 33 },
		{ 332, 22, 33 },
		{ 332, 23, 33 },
		{ 332, 24, 33 },
		{ 332, 25, 33 },
		{ 332, 26, 33 },
		{ 332, 27, 33 },
		{ 332, 28, 33 },
		{ 332, 29, 33 },
		{ 332, 30, 33 },
		{ 332, 31, 33 },
		{ 332, 32, 33 },
		{ 332, 33, 33 },
		{ 332, 34, 33 },
		{ 332, 35, 33 },
		{ 332, 36, 33 },
		{ 332, 37, 33 },
		{ 332, 38, 33 },
		{ 332, 39, 33 },
		{ 332, 40, 33 },
		{ 332, 41, 33 },
		{ 332, 42, 33 },
		{ 332, 43, 33 },
		{ 332, 44, 33 },
		{ 332, 45, 33 },
		{ 332, 46, 33 },
		{ 332, 47, 33 },
		{ 332, 48, 33 },
		{ 332, 49, 33 },
		{ 332, 50, 33 },
		{ 332, 51, 33 },
		{ 332, 52, 33 },
		{ 332, 53, 33 },
		{ 332, 54, 33 },
		{ 332, 55, 33 },
		{ 332, 56, 33 },
		{ 332, 57, 33 },
		{ 332, 58, 33 },
		{ 332, 59, 33 },
		{ 332, 60, 33 },
		{ 332, 61, 33 },
		{ 332, 62, 33 },
		{ 332, 63, 33 },
		{ 332, 64, 33 },
		{ 332, 65, 33 },
		{ 332, 66, 33 },
		{ 332, 67, 33 },
		{ 332, 68, 33 },
		{ 332, 69, 33 },
		{ 332, 70, 33 },
		{ 332, 71, 33 },
		{ 332, 72, 33 },
		{ 332, 73, 33 },
		{ 332, 74, 33 },
		{ 332, 75, 33 },
		{ 332, 76, 33 },
		{ 332, 77, 33 },
		{ 332, 78, 33 },
		{ 332, 79, 33 },
		{ 332, 80, 33 },
		{ 332, 81, 33 },
		{ 332, 82, 33 },
		{ 332, 83, 33 },
		{ 332, 84, 33 },
		{ 332, 85, 33 },
		{ 332, 86, 33 },
		{ 332, 87, 33 },
		{ 332, 88, 33 },
		{ 332, 89, 33 },
		{ 332, 90, 33 },
		{ 332, 91, 33 },
		{ 332, 92, 33 },
		{ 332, 93, 33 },
		{ 332, 94, 33 },
		{ 332, 95, 33 },
		{ 332, 96, 33 },
		{ 332, 97, 33 },
		{ 332, 98, 33 },
		{ 332, 99, 33 },
		{ 332, 100, 33 },
		{ 332, 101, 33 },
		{ 332, 102, 33 },
		{ 332, 103, 33 },
		{ 332, 104, 33 },
		{ 332, 105, 33 },
		{ 332, 106, 33 },
		{ 333, 1, 33 },
		{ 333, 2, 33 },
		{ 333, 3, 33 },
		{ 333, 4, 33 },
		{ 333, 5, 33 },
		{ 333, 6, 33 },
		{ 333, 7, 33 },
		{ 333, 8, 33 },
		{ 333, 9, 33 },
		{ 333, 10, 33 },
		{ 333, 11, 33 },
		{ 333, 12, 33 },
		{ 333, 13, 33 },
		{ 333, 14, 33 },
		{ 333, 15, 33 },
		{ 333, 16, 33 },
		{ 333, 17, 33 },
		{ 333, 18, 33 },
		{ 333, 19, 33 },
		{ 333, 20, 33 },
		{ 333, 21, 33 },
		{ 333, 22, 33 },
		{ 333, 23, 33 },
		{ 333, 24, 33 },
		{ 333, 25, 33 },
		{ 333, 26, 33 },
		{ 333, 27, 33 },
		{ 333, 28, 33 },
		{ 333, 29, 33 },
		{ 333, 30, 33 },
		{ 333, 31, 33 },
		{ 333, 32, 33 },
		{ 333, 33, 33 },
		{ 333, 34, 33 },
		{ 333, 35, 33 },
		{ 333, 36, 33 },
		{ 333, 37, 33 },
		{ 333, 38, 33 },
		{ 333, 39, 33 },
		{ 333, 40, 33 },
		{ 333, 41, 33 },
		{ 333, 42, 33 },
		{ 333, 43, 33 },
		{ 333, 44, 33 },
		{ 333, 45, 33 },
		{ 333, 46, 33 },
		{ 333, 47, 33 },
		{ 333, 48, 33 },
		{ 333, 49, 33 },
		{ 333, 50, 33 },
		{ 333, 51, 33 },
		{ 333, 52, 33 },
		{ 333, 53, 33 },
		{ 333, 54, 33 },
		{ 333, 55, 33 },
		{ 333, 56, 33 },
		{ 333, 57, 33 },
		{ 333, 58, 33 },
		{ 333, 59, 33 },
		{ 333, 60, 33 },
		{ 333, 61, 33 },
		{ 333, 62, 33 },
		{ 333, 63, 33 },
		{ 333, 64, 33 },
		{ 333, 65, 33 },
		{ 333, 66, 33 },
		{ 333, 67, 33 },
		{ 333, 68, 33 },
		{ 333, 69, 33 },
		{ 333, 70, 33 },
		{ 333, 71, 33 },
		{ 333, 72, 33 },
		{ 333, 73, 33 },
		{ 333, 74, 33 },
		{ 333, 75, 33 },
		{ 333, 76, 33 },
		{ 333, 77, 33 },
		{ 333, 78, 33 },
		{ 333, 79, 33 },
		{ 333, 80, 33 },
		{ 333, 81, 33 },
		{ 333, 82, 33 },
		{ 333, 83, 33 },
		{ 333, 84, 33 },
		{ 333, 85, 33 },
		{ 333, 86, 33 },
		{ 333, 87, 33 },
		{ 333, 88, 33 },
		{ 333, 89, 33 },
		{ 333, 90, 33 },
		{ 333, 91, 33 },
		{ 333, 92, 33 },
		{ 333, 93, 33 },
		{ 333, 94, 33 },
		{ 333, 95, 33 },
		{ 333, 96, 33 },
		{ 333, 97, 33 },
		{ 333, 98, 33 },
		{ 333, 99, 33 },
		{ 333, 100, 33 },
		{ 333, 101, 33 },
		{ 333, 102, 33 },
		{ 333, 103, 33 },
		{ 333, 104, 33 },
		{ 334, 1, 33 },
		{ 334, 2, 33 },
		{ 334, 3, 33 },
		{ 334, 4, 33 },
		{ 334, 5, 33 },
		{ 334, 6, 33 },
		{ 334, 7, 33 },
		{ 334, 8, 33 },
		{ 334, 9, 33 },
		{ 334, 10, 33 },
		{ 334, 11, 33 },
		{ 334, 12, 33 },
		{ 334, 13, 33 },
		{ 334, 14, 33 },
		{ 334, 15, 33 },
		{ 334, 16, 33 },
		{ 334, 17, 33 },
		{ 334, 18, 33 },
		{ 334, 19, 33 },
		{ 334, 20, 33 },
		{ 334, 21, 33 },
		{ 334, 22, 33 },
		{ 334, 23, 33 },
		{ 334, 24, 33 },
		{ 334, 25, 33 },
		{ 334, 26, 33 },
		{ 334, 27, 33 },
		{ 334, 28, 33 },
		{ 334, 29, 33 },
		{ 334, 30, 33 },
		{ 334, 31, 33 },
		{ 334, 32, 33 },
		{ 334, 33, 33 },
		{ 334, 34, 33 },
		{ 334, 35, 33 },
		{ 334, 36, 33 },
		{ 334, 37, 33 },
		{ 334, 38, 33 },
		{ 334, 39, 33 },
		{ 334, 40, 33 },
		{ 334, 41, 33 },
		{ 334, 42, 33 },
		{ 334, 43, 33 },
		{ 334, 44, 33 },
		{ 334, 45, 33 },
		{ 334, 46, 33 },
		{ 334, 47, 33 },
		{ 334, 48, 33 },
		{ 334, 49, 33 },
		{ 334, 50, 33 },
		{ 334, 51, 33 },
		{ 334, 52, 33 },
		{ 334, 53, 33 },
		{ 334, 54, 33 },
		{ 334, 55, 33 },
		{ 334, 56, 33 },
		{ 334, 57, 33 },
		{ 334, 58, 33 },
		{ 334, 59, 33 },
		{ 334, 60, 33 },
		{ 334, 61, 33 },
		{ 334, 62, 33 },
		{ 334, 63, 33 },
		{ 334, 64, 33 },
		{ 334, 65, 33 },
		{ 334, 66, 33 },
		{ 334, 67, 33 },
		{ 334, 68, 33 },
		{ 334, 69, 33 },
		{ 334, 70, 33 },
		{ 334, 71, 33 },
		{ 334, 72, 33 },
		{ 334, 73, 33 },
		{ 334, 74, 33 },
		{ 334, 75, 33 },
		{ 334, 76, 33 },
		{ 334, 77, 33 },
		{ 334, 78, 33 },
		{ 334, 79, 33 },
		{ 334, 80, 33 },
		{ 334, 81, 33 },
		{ 334, 82, 33 },
		{ 334, 83, 33 },
		{ 334, 84, 33 },
		{ 334, 85, 33 },
		{ 334, 86, 33 },
		{ 334, 87, 33 },
		{ 334, 88, 33 },
		{ 334, 89, 33 },
		{ 334, 90, 33 },
		{ 334, 91, 33 },
		{ 334, 92, 33 },
		{ 334, 93, 33 },
		{ 334, 94, 33 },
		{ 334, 95, 33 },
		{ 334, 96, 33 },
		{ 334, 97, 33 },
		{ 334, 98, 33 },
		{ 334, 99, 33 },
		{ 334, 100, 33 },
		{ 334, 101, 33 },
		{ 334, 102, 33 },
		{ 335, 1, 33 },
		{ 335, 2, 33 },
		{ 335, 3, 33 },
		{ 335, 4, 33 },
		{ 335, 5, 33 },
		{ 335, 6, 33 },
		{ 335, 7, 33 },
		{ 335, 8, 33 },
		{ 335, 9, 33 },
		{ 335, 10, 33 },
		{ 335, 11, 33 },
		{ 335, 12, 33 },
		{ 335, 13, 33 },
		{ 335, 14, 33 },
		{ 335, 15, 33 },
		{ 335, 16, 33 },
		{ 335, 17, 33 },
		{ 335, 18, 33 },
		{ 335, 19, 33 },
		{ 335, 20, 33 },
		{ 335, 21, 33 },
		{ 335, 22, 33 },
		{ 335, 23, 33 },
		{ 335, 24, 33 },
		{ 335, 25, 33 },
		{ 335, 26, 33 },
		{ 335, 27, 33 },
		{ 335, 28, 33 },
		{ 335, 29, 33 },
		{ 335, 30, 33 },
		{ 335, 31, 33 },
		{ 335, 32, 33 },
		{ 335, 33, 33 },
		{ 335, 34, 33 },
		{ 335, 35, 33 },
		{ 335, 36, 33 },
		{ 335, 37, 33 },
		{ 335, 38, 33 },
		{ 335, 39, 33 },
		{ 335, 40, 33 },
		{ 335, 41, 33 },
		{ 335, 42, 33 },
		{ 335, 43, 33 },
		{ 335, 44, 33 },
		{ 335, 45, 33 },
		{ 335, 46, 33 },
		{ 335, 47, 33 },
		{ 335, 48, 33 },
		{ 335, 49, 33 },
		{ 335, 50, 33 },
		{ 335, 51, 33 },
		{ 335, 52, 33 },
		{ 335, 53, 33 },
		{ 335, 54, 33 },
		{ 335, 55, 33 },
		{ 335, 56, 33 },
		{ 335, 57, 33 },
		{ 335, 58, 33 },
		{ 335, 59, 33 },
		{ 335, 60, 33 },
		{ 335, 61, 33 },
		{ 335, 62, 33 },
		{ 335, 63, 33 },
		{ 335, 64, 33 },
		{ 335, 65, 33 },
		{ 335, 66, 33 },
		{ 335, 67, 33 },
		{ 335, 68, 33 },
		{ 335, 69, 33 },
		{ 335, 70, 33 },
		{ 335, 71, 33 },
		{ 335, 72, 33 },
		{ 335, 73, 33 },
		{ 335, 74, 33 },
		{ 335, 75, 33 },
		{ 335, 76, 33 },
		{ 335, 77, 33 },
		{ 335, 78, 33 },
		{ 335, 79, 33 },
		{ 335, 80, 33 },
		{ 335, 81, 33 },
		{ 335, 82, 33 },
		{ 335, 83, 33 },
		{ 335, 84, 33 },
		{ 335, 85, 33 },
		{ 335, 86, 33 },
		{ 335, 87, 33 },
		{ 335, 88, 33 },
		{ 335, 89, 33 },
		{ 335, 90, 33 },
		{ 335, 91, 33 },
		{ 335, 92, 33 },
		{ 335, 93, 33 },
		{ 335, 94, 33 },
		{ 335, 95, 33 },
		{ 335, 96, 33 },
		{ 335, 97, 33 },
		{ 335, 98, 33 },
		{ 335, 99, 33 },
		{ 335, 100, 33 },
		{ 336, 1, 33 },
		{ 336, 2, 33 },
		{ 336, 3, 33 },
		{ 336, 4, 33 },
		{ 336, 5, 33 },
		{ 336, 6, 33 },
		{ 336, 7, 33 },
		{ 336, 8, 33 },
		{ 336, 9, 33 },
		{ 336, 10, 33 },
		{ 336, 11, 33 },
		{ 336, 12, 33 },
		{ 336, 13, 33 },
		{ 336, 14, 33 },
		{ 336, 15, 33 },
		{ 336, 16, 33 },
		{ 336, 17, 33 },
		{ 336, 18, 33 },
		{ 336, 19, 33 },
		{ 336, 20, 33 },
		{ 336, 21, 33 },
		{ 336, 22, 33 },
		{ 336, 23, 33 },
		{ 336, 24, 33 },
		{ 336, 25, 33 },
		{ 336, 26, 33 },
		{ 336, 27, 33 },
		{ 336, 28, 33 },
		{ 336, 29, 33 },
		{ 336, 30, 33 },
		{ 336, 31, 33 },
		{ 336, 32, 33 },
		{ 336, 33, 33 },
		{ 336, 34, 33 },
		{ 336, 35, 33 },
		{ 336, 36, 33 },
		{ 336, 37, 33 },
		{ 336, 38, 33 },
		{ 336, 39, 33 },
		{ 336, 40, 33 },
		{ 336, 41, 33 },
		{ 336, 42, 33 },
		{ 336, 43, 33 },
		{ 336, 44, 33 },
		{ 336, 45, 33 },
		{ 336, 46, 33 },
		{ 336, 47, 33 },
		{ 336, 48, 33 },
		{ 336, 49, 33 },
		{ 336, 50, 33 },
		{ 336, 51, 33 },
		{ 336, 52, 33 },
		{ 336, 53, 33 },
		{ 336, 54, 33 },
		{ 336, 55, 33 },
		{ 336, 56, 33 },
		{ 336, 57, 33 },
		{ 336, 58, 33 },
		{ 336, 59, 33 },
		{ 336, 60, 33 },
		{ 336, 61, 33 },
		{ 336, 62, 33 },
		{ 336, 63, 33 },
		{ 336, 64, 33 },
		{ 336, 65, 33 },
		{ 336, 66, 33 },
		{ 336, 67, 33 },
		{ 336, 68, 33 },
		{ 336, 69, 33 },
		{ 336, 70, 33 },
		{ 336, 71, 33 },
		{ 336, 72, 33 },
		{ 336, 73, 33 },
		{ 336, 74, 33 },
		{ 336, 75, 33 },
		{ 336, 76, 33 },
		{ 336, 77, 33 },
		{ 336, 78, 33 },
		{ 336, 79, 33 },
		{ 336, 80, 33 },
		{ 336, 81, 33 },
		{ 336, 82, 33 },
		{ 336, 83, 33 },
		{ 336, 84, 33 },
		{ 336, 85, 33 },
		{ 336, 86, 33 },
		{ 336, 87, 33 },
		{ 336, 88, 33 },
		{ 336, 89, 33 },
		{ 336, 90, 33 },
		{ 336, 91, 33 },
		{ 336, 92, 33 },
		{ 336, 93, 33 },
		{ 336, 94, 33 },
		{ 336, 95, 33 },
		{ 336, 96, 33 },
		{ 336, 97, 33 },
		{ 336, 98, 33 },
		{ 337, 1, 33 },
		{ 337, 2, 33 },
		{ 337, 3, 33 },
		{ 337, 4, 33 },
		{ 337, 5, 33 },
		{ 337, 6, 33 },
		{ 337, 7, 33 },
		{ 337, 8, 33 },
		{ 337, 9, 33 },
		{ 337, 10, 33 },
		{ 337, 11, 33 },
		{ 337, 12, 33 },
		{ 337, 13, 33 },
		{ 337, 14, 33 },
		{ 337, 15, 33 },
		{ 337, 16, 33 },
		{ 337, 17, 33 },
		{ 337, 18, 33 },
		{ 337, 19, 33 },
		{ 337, 20, 33 },
		{ 337, 21, 33 },
		{ 337, 22, 33 },
		{ 337, 23, 33 },
		{ 337, 24, 33 },
		{ 337, 25, 33 },
		{ 337, 26, 33 },
		{ 337, 27, 33 },
		{ 337, 28, 33 },
		{ 337, 29, 33 },
		{ 337, 30, 33 },
		{ 337, 31, 33 },
		{ 337, 32, 33 },
		{ 337, 33, 33 },
		{ 337, 34, 33 },
		{ 337, 35, 33 },
		{ 337, 36, 33 },
		{ 337, 37, 33 },
		{ 337, 38, 33 },
		{ 337, 39, 33 },
		{ 337, 40, 33 },
		{ 337, 41, 33 },
		{ 337, 42, 33 },
		{ 337, 43, 33 },
		{ 337, 44, 33 },
		{ 337, 45, 33 },
		{ 337, 46, 33 },
		{ 337, 47, 33 },
		{ 337, 48, 33 },
		{ 337, 49, 33 },
		{ 337, 50, 33 },
		{ 337, 51, 33 },
		{ 337, 52, 33 },
		{ 337, 53, 33 },
		{ 337, 54, 33 },
		{ 337, 55, 33 },
		{ 337, 56, 33 },
		{ 337, 57, 33 },
		{ 337, 58, 33 },
		{ 337, 59, 33 },
		{ 337, 60, 33 },
		{ 337, 61, 33 },
		{ 337, 62, 33 },
		{ 337, 63, 33 },
		{ 337, 64, 33 },
		{ 337, 65, 33 },
		{ 337, 66, 33 },
		{ 337, 67, 33 },
		{ 337, 68, 33 },
		{ 337, 69, 33 },
		{ 337, 70, 33 },
		{ 337, 71, 33 },
		{ 337, 72, 33 },
		{ 337, 73, 33 },
		{ 337, 74, 33 },
		{ 337, 75, 33 },
		{ 337, 76, 33 },
		{ 337, 77, 33 },
		{ 337, 78, 33 },
		{ 337, 79, 33 },
		{ 337, 80, 33 },
		{ 337, 81, 33 },
		{ 337, 82, 33 },
		{ 337, 83, 33 },
		{ 337, 84, 33 },
		{ 337, 85, 33 },
		{ 337, 86, 33 },
		{ 337, 87, 33 },
		{ 337, 88, 33 },
		{ 337, 89, 33 },
		{ 337, 90, 33 },
		{ 337, 91, 33 },
		{ 337, 92, 33 },
		{ 337, 93, 33 },
		{ 337, 94, 33 },
		{ 337, 95, 33 },
		{ 337, 96, 33 },
		{ 338, 1, 33 },
		{ 338, 2, 33 },
		{ 338, 3, 33 },
		{ 338, 4, 33 },
		{ 338, 5, 33 },
		{ 338, 6, 33 },
		{ 338, 7, 33 },
		{ 338, 8, 33 },
		{ 338, 9, 33 },
		{ 338, 10, 33 },
		{ 338, 11, 33 },
		{ 338, 12, 33 },
		{ 338, 13, 33 },
		{ 338, 14, 33 },
		{ 338, 15, 33 },
		{ 338, 16, 33 },
		{ 338, 17, 33 },
		{ 338, 18, 33 },
		{ 338, 19, 33 },
		{ 338, 20, 33 },
		{ 338, 21, 33 },
		{ 338, 22, 33 },
		{ 338, 23, 33 },
		{ 338, 24, 33 },
		{ 338, 25, 33 },
		{ 338, 26, 33 },
		{ 338, 27, 33 },
		{ 338, 28, 33 },
		{ 338, 29, 33 },
		{ 338, 30, 33 },
		{ 338, 31, 33 },
		{ 338, 32, 33 },
		{ 338, 33, 33 },
		{ 338, 34, 33 },
		{ 338, 35, 33 },
		{ 338, 36, 33 },
		{ 338, 37, 33 },
		{ 338, 38, 33 },
		{ 338, 39, 33 },
		{ 338, 40, 33 },
		{ 338, 41, 33 },
		{ 338, 42, 33 },
		{ 338, 43, 33 },
		{ 338, 44, 33 },
		{ 338, 45, 33 },
		{ 338, 46, 33 },
		{ 338, 47, 33 },
		{ 338, 48, 33 },
		{ 338, 49, 33 },
		{ 338, 50, 33 },
		{ 338, 51, 33 },
		{ 338, 52, 33 },
		{ 338, 53, 33 },
		{ 338, 54, 33 },
		{ 338, 55, 33 },
		{ 338, 56, 33 },
		{ 338, 57, 33 },
		{ 338, 58, 33 },
		{ 338, 59, 33 },
		{ 338, 60, 33 },
		{ 338, 61, 33 },
		{ 338, 62, 33 },
		{ 338, 63, 33 },
		{ 338, 64, 33 },
		{ 338, 65, 33 },
		{ 338, 66, 33 },
		{ 338, 67, 33 },
		{ 338, 68, 33 },
		{ 338, 69, 33 },
		{ 338, 70, 33 },
		{ 338, 71, 33 },
		{ 338, 72, 33 },
		{ 338, 73, 33 },
		{ 338, 74, 33 },
		{ 338, 75, 33 },
		{ 338, 76, 33 },
		{ 338, 77, 33 },
		{ 338, 78, 33 },
		{ 338, 79, 33 },
		{ 338, 80, 33 },
		{ 338, 81, 33 },
		{ 338, 82, 33 },
		{ 338, 83, 33 },
		{ 338, 84, 33 },
		{ 338, 85, 33 },
		{ 338, 86, 33 },
		{ 338, 87, 33 },
		{ 338, 88, 33 },
		{ 338, 89, 33 },
		{ 338, 90, 33 },
		{ 338, 91, 33 },
		{ 338, 92, 33 },
		{ 338, 93, 33 },
		{ 338, 94, 33 },
		{ 339, 1, 33 },
		{ 339, 2, 33 },
		{ 339, 3, 33 },
		{ 339, 4, 33 },
		{ 339, 5, 33 },
		{ 339, 6, 33 },
		{ 339, 7, 33 },
		{ 339, 8, 33 },
		{ 339, 9, 33 },
		{ 339, 10, 33 },
		{ 339, 11, 33 },
		{ 339, 12, 33 },
		{ 339, 13, 33 },
		{ 339, 14, 33 },
		{ 339, 15, 33 },
		{ 339, 16, 33 },
		{ 339, 17, 33 },
		{ 339, 18, 33 },
		{ 339, 19, 33 },
		{ 339, 20, 33 },
		{ 339, 21, 33 },
		{ 339, 22, 33 },
		{ 339, 23, 33 },
		{ 339, 24, 33 },
		{ 339, 25, 33 },
		{ 339, 26, 33 },
		{ 339, 27, 33 },
		{ 339, 28, 33 },
		{ 339, 29, 33 },
		{ 339, 30, 33 },
		{ 339, 31, 33 },
		{ 339, 32, 33 },
		{ 339, 33, 33 },
		{ 339, 34, 33 },
		{ 339, 35, 33 },
		{ 339, 36, 33 },
		{ 339, 37, 33 },
		{ 339, 38, 33 },
		{ 339, 39, 33 },
		{ 339, 40, 33 },
		{ 339, 41, 33 },
		{ 339, 42, 33 },
		{ 339, 43, 33 },
		{ 339, 44, 33 },
		{ 339, 45, 33 },
		{ 339, 46, 33 },
		{ 339, 47, 33 },
		{ 339, 48, 33 },
		{ 339, 49, 33 },
		{ 339, 50, 33 },
		{ 339, 51, 33 },
		{ 339, 52, 33 },
		{ 339, 53, 33 },
		{ 339, 54, 33 },
		{ 339, 55, 33 },
		{ 339, 56, 33 },
		{ 339, 57, 33 },
		{ 339, 58, 33 },
		{ 339, 59, 33 },
		{ 339, 60, 33 },
		{ 339, 61, 33 },
		{ 339, 62, 33 },
		{ 339, 63, 33 },
		{ 339, 64, 33 },
		{ 339, 65, 33 },
		{ 339, 66, 33 },
		{ 339, 67, 33 },
		{ 339, 68, 33 },
		{ 339, 69, 33 },
		{ 339, 70, 33 },
		{ 339, 71, 33 },
		{ 339, 72, 33 },
		{ 339, 73, 33 },
		{ 339, 74, 33 },
		{ 339, 75, 33 },
		{ 339, 76, 33 },
		{ 339, 77, 33 },
		{ 339, 78, 33 },
		{ 339, 79, 33 },
		{ 339, 80, 33 },
		{ 339, 81, 33 },
		{ 339, 82, 33 },
		{ 339, 83, 33 },
		{ 339, 84, 33 },
		{ 339, 85, 33 },
		{ 339, 86, 33 },
		{ 339, 87, 33 },
		{ 339, 88, 33 },
		{ 339, 89, 33 },
		{ 339, 90, 33 },
		{ 339, 91, 33 },
		{ 339, 92, 33 },
		{ 340, 1, 33 },
		{ 340, 2, 33 },
		{ 340, 3, 33 },
		{ 340, 4, 33 },
		{ 340, 5, 33 },
		{ 340, 6, 33 },
		{ 340, 7, 33 },
		{ 340, 8, 33 },
		{ 340, 9, 33 },
		{ 340, 10, 33 },
		{ 340, 11, 33 },
		{ 340, 12, 33 },
		{ 340, 13, 33 },
		{ 340, 14, 33 },
		{ 340, 15, 33 },
		{ 340, 16, 33 },
		{ 340, 17, 33 },
		{ 340, 18, 33 },
		{ 340, 19, 33 },
		{ 340, 20, 33 },
		{ 340, 21, 33 },
		{ 340, 22, 33 },
		{ 340, 23, 33 },
		{ 340, 24, 33 },
		{ 340, 25, 33 },
		{ 340, 26, 33 },
		{ 340, 27, 33 },
		{ 340, 28, 33 },
		{ 340, 29, 33 },
		{ 340, 30, 33 },
		{ 340, 31, 33 },
		{ 340, 32, 33 },
		{ 340, 33, 33 },
		{ 340, 34, 33 },
		{ 340, 35, 33 },
		{ 340, 36, 33 },
		{ 340, 37, 33 },
		{ 340, 38, 33 },
		{ 340, 39, 33 },
		{ 340, 40, 33 },
		{ 340, 41, 33 },
		{ 340, 42, 33 },
		{ 340, 43, 33 },
		{ 340, 44, 33 },
		{ 340, 45, 33 },
		{ 340, 46, 33 },
		{ 340, 47, 33 },
		{ 340, 48, 33 },
		{ 340, 49, 33 },
		{ 340, 50, 33 },
		{ 340, 51, 33 },
		{ 340, 52, 33 },
		{ 340, 53, 33 },
		{ 340, 54, 33 },
		{ 340, 55, 33 },
		{ 340, 56, 33 },
		{ 340, 57, 33 },
		{ 340, 58, 33 },
		{ 340, 59, 33 },
		{ 340, 60, 33 },
		{ 340, 61, 33 },
		{ 340, 62, 33 },
		{ 340, 63, 33 },
		{ 340, 64, 33 },
		{ 340, 65, 33 },
		{ 340, 66, 33 },
		{ 340, 67, 33 },
		{ 340, 68, 33 },
		{ 340, 69, 33 },
		{ 340, 70, 33 },
		{ 340, 71, 33 },
		{ 340, 72, 33 },
		{ 340, 73, 33 },
		{ 340, 74, 33 },
		{ 340, 75, 33 },
		{ 340, 76, 33 },
		{ 340, 77, 33 },
		{ 340, 78, 33 },
		{ 340, 79, 33 },
		{ 340, 80, 33 },
		{ 340, 81, 33 },
		{ 340, 82, 33 },
		{ 340, 83, 33 },
		{ 340, 84, 33 },
		{ 340, 85, 33 },
		{ 340, 86, 33 },
		{ 340, 87, 33 },
		{ 340, 88, 33 },
		{ 340, 89, 33 },
		{ 340, 90, 33 },
		{ 341, 1, 33 },
		{ 341, 2, 33 },
		{ 341, 3, 33 },
		{ 341, 4, 33 },
		{ 341, 5, 33 },
		{ 341, 6, 33 },
		{ 341, 7, 33 },
		{ 341, 8, 33 },
		{ 341, 9, 33 },
		{ 341, 10, 33 },
		{ 341, 11, 33 },
		{ 341, 12, 33 },
		{ 341, 13, 33 },
		{ 341, 14, 33 },
		{ 341, 15, 33 },
		{ 341, 16, 33 },
		{ 341, 17, 33 },
		{ 341, 18, 33 },
		{ 341, 19, 33 },
		{ 341, 20, 33 },
		{ 341, 21, 33 },
		{ 341, 22, 33 },
		{ 341, 23, 33 },
		{ 341, 24, 33 },
		{ 341, 25, 33 },
		{ 341, 26, 33 },
		{ 341, 27, 33 },
		{ 341, 28, 33 },
		{ 341, 29, 33 },
		{ 341, 30, 33 },
		{ 341, 31, 33 },
		{ 341, 32, 33 },
		{ 341, 33, 33 },
		{ 341, 34, 33 },
		{ 341, 35, 33 },
		{ 341, 36, 33 },
		{ 341, 37, 33 },
		{ 341, 38, 33 },
		{ 341, 39, 33 },
		{ 341, 40, 33 },
		{ 341, 41, 33 },
		{ 341, 42, 33 },
		{ 341, 43, 33 },
		{ 341, 44, 33 },
		{ 341, 45, 33 },
		{ 341, 46, 33 },
		{ 341, 47, 33 },
		{ 341, 48, 33 },
		{ 341, 49, 33 },
		{ 341, 50, 33 },
		{ 341, 51, 33 },
		{ 341, 52, 33 },
		{ 341, 53, 33 },
		{ 341, 54, 33 },
		{ 341, 55, 33 },
		{ 341, 56, 33 },
		{ 341, 57, 33 },
		{ 341, 58, 33 },
		{ 341, 59, 33 },
		{ 341, 60, 33 },
		{ 341, 61, 33 },
		{ 341, 62, 33 },
		{ 341, 63, 33 },
		{ 341, 64, 33 },
		{ 341, 65, 33 },
		{ 341, 66, 33 },
		{ 341, 67, 33 },
		{ 341, 68, 33 },
		{ 341, 69, 33 },
		{ 341, 70, 33 },
		{ 341, 71, 33 },
		{ 341, 72, 33 },
		{ 341, 73, 33 },
		{ 341, 74, 33 },
		{ 341, 75, 33 },
		{ 341, 76, 33 },
		{ 341, 77, 33 },
		{ 341, 78, 33 },
		{ 341, 79, 33 },
		{ 341, 80, 33 },
		{ 341, 81, 33 },
		{ 341, 82, 33 },
		{ 341, 83, 33 },
		{ 341, 84, 33 },
		{ 341, 85, 33 },
		{ 341, 86, 33 },
		{ 341, 87, 33 },
		{ 341, 88, 33 },
		{ 342, 1, 33 },
		{ 342, 2, 33 },
		{ 342, 3, 33 },
		{ 342, 4, 33 },
		{ 342, 5, 33 },
		{ 342, 6, 33 },
		{ 342, 7, 33 },
		{ 342, 8, 33 },
		{ 342, 9, 33 },
		{ 342, 10, 33 },
		{ 342, 11, 33 },
		{ 342, 12, 33 },
		{ 342, 13, 33 },
		{ 342, 14, 33 },
		{ 342, 15, 33 },
		{ 342, 16, 33 },
		{ 342, 17, 33 },
		{ 342, 18, 33 },
		{ 342, 19, 33 },
		{ 342, 20, 33 },
		{ 342, 21, 33 },
		{ 342, 22, 33 },
		{ 342, 23, 33 },
		{ 342, 24, 33 },
		{ 342, 25, 33 },
		{ 342, 26, 33 },
		{ 342, 27, 33 },
		{ 342, 28, 33 },
		{ 342, 29, 33 },
		{ 342, 30, 33 },
		{ 342, 31, 33 },
		{ 342, 32, 33 },
		{ 342, 33, 33 },
		{ 342, 34, 33 },
		{ 342, 35, 33 },
		{ 342, 36, 33 },
		{ 342, 37, 33 },
		{ 342, 38, 33 },
		{ 342, 39, 33 },
		{ 342, 40, 33 },
		{ 342, 41, 33 },
		{ 342, 42, 33 },
		{ 342, 43, 33 },
		{ 342, 44, 33 },
		{ 342, 45, 33 },
		{ 342, 46, 33 },
		{ 342, 47, 33 },
		{ 342, 48, 33 },
		{ 342, 49, 33 },
		{ 342, 50, 33 },
		{ 342, 51, 33 },
		{ 342, 52, 33 },
		{ 342, 53, 33 },
		{ 342, 54, 33 },
		{ 342, 55, 33 },
		{ 342, 56, 33 },
		{ 342, 57, 33 },
		{ 342, 58, 33 },
		{ 342, 59, 33 },
		{ 342, 60, 33 },
		{ 342, 61, 33 },
		{ 342, 62, 33 },
		{ 342, 63, 33 },
		{ 342, 64, 33 },
		{ 342, 65, 33 },
		{ 342, 66, 33 },
		{ 342, 67, 33 },
		{ 342, 68, 33 },
		{ 342, 69, 33 },
		{ 342, 70, 33 },
		{ 342, 71, 33 },
		{ 342, 72, 33 },
		{ 342, 73, 33 },
		{ 342, 74, 33 },
		{ 342, 75, 33 },
		{ 342, 76, 33 },
		{ 342, 77, 33 },
		{ 342, 78, 33 },
		{ 342, 79, 33 },
		{ 342, 80, 33 },
		{ 342, 81, 33 },
		{ 342, 82, 33 },
		{ 342, 83, 33 },
		{ 342, 84, 33 },
		{ 342, 85, 33 },
		{ 342, 86, 33 },
		{ 343, 1, 33 },
		{ 343, 2, 33 },
		{ 343, 3, 33 },
		{ 343, 4, 33 },
		{ 343, 5, 33 },
		{ 343, 6, 33 },
		{ 343, 7, 33 },
		{ 343, 8, 33 },
		{ 343, 9, 33 },
		{ 343, 10, 33 },
		{ 343, 11, 33 },
		{ 343, 12, 33 },
		{ 343, 13, 33 },
		{ 343, 14, 33 },
		{ 343, 15, 33 },
		{ 343, 16, 33 },
		{ 343, 17, 33 },
		{ 343, 18, 33 },
		{ 343, 19, 33 },
		{ 343, 20, 33 },
		{ 343, 21, 33 },
		{ 343, 22, 33 },
		{ 343, 23, 33 },
		{ 343, 24, 33 },
		{ 343, 25, 33 },
		{ 343, 26, 33 },
		{ 343, 27, 33 },
		{ 343, 28, 33 },
		{ 343, 29, 33 },
		{ 343, 30, 33 },
		{ 343, 31, 33 },
		{ 343, 32, 33 },
		{ 343, 33, 33 },
		{ 343, 34, 33 },
		{ 343, 35, 33 },
		{ 343, 36, 33 },
		{ 343, 37, 33 },
		{ 343, 38, 33 },
		{ 343, 39, 33 },
		{ 343, 40, 33 },
		{ 343, 41, 33 },
		{ 343, 42, 33 },
		{ 343, 43, 33 },
		{ 343, 44, 33 },
		{ 343, 45, 33 },
		{ 343, 46, 33 },
		{ 343, 47, 33 },
		{ 343, 48, 33 },
		{ 343, 49, 33 },
		{ 343, 50, 33 },
		{ 343, 51, 33 },
		{ 343, 52, 33 },
		{ 343, 53, 33 },
		{ 343, 54, 33 },
		{ 343, 55, 33 },
		{ 343, 56, 33 },
		{ 343, 57, 33 },
		{ 343, 58, 33 },
		{ 343, 59, 33 },
		{ 343, 60, 33 },
		{ 343, 61, 33 },
		{ 343, 62, 33 },
		{ 343, 63, 33 },
		{ 343, 64, 33 },
		{ 343, 65, 33 },
		{ 343, 66, 33 },
		{ 343, 67, 33 },
		{ 343, 68, 33 },
		{ 343, 69, 33 },
		{ 343, 70, 33 },
		{ 343, 71, 33 },
		{ 343, 72, 33 },
		{ 343, 73, 33 },
		{ 343, 74, 33 },
		{ 343, 75, 33 },
		{ 343, 76, 33 },
		{ 343, 77, 33 },
		{ 343, 78, 33 },
		{ 343, 79, 33 },
		{ 343, 80, 33 },
		{ 343, 81, 33 },
		{ 343, 82, 33 },
		{ 343, 83, 33 },
		{ 343, 84, 33 },
		{ 344, 1, 33 },
		{ 344, 2, 33 },
		{ 344, 3, 33 },
		{ 344, 4, 33 },
		{ 344, 5, 33 },
		{ 344, 6, 33 },
		{ 344, 7, 33 },
		{ 344, 8, 33 },
		{ 344, 9, 33 },
		{ 344, 10, 33 },
		{ 344, 11, 33 },
		{ 344, 12, 33 },
		{ 344, 13, 33 },
		{ 344, 14, 33 },
		{ 344, 15, 33 },
		{ 344, 16, 33 },
		{ 344, 17, 33 },
		{ 344, 18, 33 },
		{ 344, 19, 33 },
		{ 344, 20, 33 },
		{ 344, 21, 33 },
		{ 344, 22, 33 },
		{ 344, 23, 33 },
		{ 344, 24, 33 },
		{ 344, 25, 33 },
		{ 344, 26, 33 },
		{ 344, 27, 33 },
		{ 344, 28, 33 },
		{ 344, 29, 33 },
		{ 344, 30, 33 },
		{ 344, 31, 33 },
		{ 344, 32, 33 },
		{ 344, 33, 33 },
		{ 344, 34, 33 },
		{ 344, 35, 33 },
		{ 344, 36, 33 },
		{ 344, 37, 33 },
		{ 344, 38, 33 },
		{ 344, 39, 33 },
		{ 344, 40, 33 },
		{ 344, 41, 33 },
		{ 344, 42, 33 },
		{ 344, 43, 33 },
		{ 344, 44, 33 },
		{ 344, 45, 33 },
		{ 344, 46, 33 },
		{ 344, 47, 33 },
		{ 344, 48, 33 },
		{ 344, 49, 33 },
		{ 344, 50, 33 },
		{ 344, 51, 33 },
		{ 344, 52, 33 },
		{ 344, 53, 33 },
		{ 344, 54, 33 },
		{ 344, 55, 33 },
		{ 344, 56, 33 },
		{ 344, 57, 33 },
		{ 344, 58, 33 },
		{ 344, 59, 33 },
		{ 344, 60, 33 },
		{ 344, 61, 33 },
		{ 344, 62, 33 },
		{ 344, 63, 33 },
		{ 344, 64, 33 },
		{ 344, 65, 33 },
		{ 344, 66, 33 },
		{ 344, 67, 33 },
		{ 344, 68, 33 },
		{ 344, 69, 33 },
		{ 344, 70, 33 },
		{ 344, 71, 33 },
		{ 344, 72, 33 },
		{ 344, 73, 33 },
		{ 344, 74, 33 },
		{ 344, 75, 33 },
		{ 344, 76, 33 },
		{ 344, 77, 33 },
		{ 344, 78, 33 },
		{ 344, 79, 33 },
		{ 344, 80, 33 },
		{ 344, 81, 33 },
		{ 344, 82, 33 },
		{ 345, 1, 33 },
		{ 345, 2, 33 },
		{ 345, 3, 33 },
		{ 345, 4, 33 },
		{ 345, 5, 33 },
		{ 345, 6, 33 },
		{ 345, 7, 33 },
		{ 345, 8, 33 },
		{ 345, 9, 33 },
		{ 345, 10, 33 },
		{ 345, 11, 33 },
		{ 345, 12, 33 },
		{ 345, 13, 33 },
		{ 345, 14, 33 },
		{ 345, 15, 33 },
		{ 345, 16, 33 },
		{ 345, 17, 33 },
		{ 345, 18, 33 },
		{ 345, 19, 33 },
		{ 345, 20, 33 },
		{ 345, 21, 33 },
		{ 345, 22, 33 },
		{ 345, 23, 33 },
		{ 345, 24, 33 },
		{ 345, 25, 33 },
		{ 345, 26, 33 },
		{ 345, 27, 33 },
		{ 345, 28, 33 },
		{ 345, 29, 33 },
		{ 345, 30, 33 },
		{ 345, 31, 33 },
		{ 345, 32, 33 },
		{ 345, 33, 33 },
		{ 345, 34, 33 },
		{ 345, 35, 33 },
		{ 345, 36, 33 },
		{ 345, 37, 33 },
		{ 345, 38, 33 },
		{ 345, 39, 33 },
		{ 345, 40, 33 },
		{ 345, 41, 33 },
		{ 345, 42, 33 },
		{ 345, 43, 33 },
		{ 345, 44, 33 },
		{ 345, 45, 33 },
		{ 345, 46, 33 },
		{ 345, 47, 33 },
		{ 345, 48, 33 },
		{ 345, 49, 33 },
		{ 345, 50, 33 },
		{ 345, 51, 33 },
		{ 345, 52, 33 },
		{ 345, 53, 33 },
		{ 345, 54, 33 },
		{ 345, 55, 33 },
		{ 345, 56, 33 },
		{ 345, 57, 33 },
		{ 345, 58, 33 },
		{ 345, 59, 33 },
		{ 345, 60, 33 },
		{ 345, 61, 33 },
		{ 345, 62, 33 },
		{ 345, 63, 33 },
		{ 345, 64, 33 },
		{ 345, 65, 33 },
		{ 345, 66, 33 },
		{ 345, 67, 33 },
		{ 345, 68, 33 },
		{ 345, 69, 33 },
		{ 345, 70, 33 },
		{ 345, 71, 33 },
		{ 345, 72, 33 },
		{ 345, 73, 33 },
		{ 345, 74, 33 },
		{ 345, 75, 33 },
		{ 345, 76, 33 },
		{ 345, 77, 33 },
		{ 345, 78, 33 },
		{ 345, 79, 33 },
		{ 345, 80, 33 },
		{ 346, 1, 33 },
		{ 346, 2, 33 },
		{ 346, 3, 33 },
		{ 346, 4, 33 },
		{ 346, 5, 33 },
		{ 346, 6, 33 },
		{ 346, 7, 33 },
		{ 346, 8, 33 },
		{ 346, 9, 33 },
		{ 346, 10, 33 },
		{ 346, 11, 33 },
		{ 346, 12, 33 },
		{ 346, 13, 33 },
		{ 346, 14, 33 },
		{ 346, 15, 33 },
		{ 346, 16, 33 },
		{ 346, 17, 33 },
		{ 346, 18, 33 },
		{ 346, 19, 33 },
		{ 346, 20, 33 },
		{ 346, 21, 33 },
		{ 346, 22, 33 },
		{ 346, 23, 33 },
		{ 346, 24, 33 },
		{ 346, 25, 33 },
		{ 346, 26, 33 },
		{ 346, 27, 33 },
		{ 346, 28, 33 },
		{ 346, 29, 33 },
		{ 346, 30, 33 },
		{ 346, 31, 33 },
		{ 346, 32, 33 },
		{ 346, 33, 33 },
		{ 346, 34, 33 },
		{ 346, 35, 33 },
		{ 346, 36, 33 },
		{ 346, 37, 33 },
		{ 346, 38, 33 },
		{ 346, 39, 33 },
		{ 346, 40, 33 },
		{ 346, 41, 33 },
		{ 346, 42, 33 },
		{ 346, 43, 33 },
		{ 346, 44, 33 },
		{ 346, 45, 33 },
		{ 346, 46, 33 },
		{ 346, 47, 33 },
		{ 346, 48, 33 },
		{ 346, 49, 33 },
		{ 346, 50, 33 },
		{ 346, 51, 33 },
		{ 346, 52, 33 },
		{ 346, 53, 33 },
		{ 346, 54, 33 },
		{ 346, 55, 33 },
		{ 346, 56, 33 },
		{ 346, 57, 33 },
		{ 346, 58, 33 },
		{ 346, 59, 33 },
		{ 346, 60, 33 },
		{ 346, 61, 33 },
		{ 346, 62, 33 },
		{ 346, 63, 33 },
		{ 346, 64, 33 },
		{ 346, 65, 33 },
		{ 346, 66, 33 },
		{ 346, 67, 33 },
		{ 346, 68, 33 },
		{ 346, 69, 33 },
		{ 346, 70, 33 },
		{ 346, 71, 33 },
		{ 346, 72, 33 },
		{ 346, 73, 33 },
		{ 346, 74, 33 },
		{ 346, 75, 33 },
		{ 346, 76, 33 },
		{ 346, 77, 33 },
		{ 346, 78, 33 },
		{ 347, 1, 33 },
		{ 347, 2, 33 },
		{ 347, 3, 33 },
		{ 347, 4, 33 },
		{ 347, 5, 33 },
		{ 347, 6, 33 },
		{ 347, 7, 33 },
		{ 347, 8, 33 },
		{ 347, 9, 33 },
		{ 347, 10, 33 },
		{ 347, 11, 33 },
		{ 347, 12, 33 },
		{ 347, 13, 33 },
		{ 347, 14, 33 },
		{ 347, 15, 33 },
		{ 347, 16, 33 },
		{ 347, 17, 33 },
		{ 347, 18, 33 },
		{ 347, 19, 33 },
		{ 347, 20, 33 },
		{ 347, 21, 33 },
		{ 347, 22, 33 },
		{ 347, 23, 33 },
		{ 347, 24, 33 },
		{ 347, 25, 33 },
		{ 347, 26, 33 },
		{ 347, 27, 33 },
		{ 347, 28, 33 },
		{ 347, 29, 33 },
		{ 347, 30, 33 },
		{ 347, 31, 33 },
		{ 347, 32, 33 },
		{ 347, 33, 33 },
		{ 347, 34, 33 },
		{ 347, 35, 33 },
		{ 347, 36, 33 },
		{ 347, 37, 33 },
		{ 347, 38, 33 },
		{ 347, 39, 33 },
		{ 347, 40, 33 },
		{ 347, 41, 33 },
		{ 347, 42, 33 },
		{ 347, 43, 33 },
		{ 347, 44, 33 },
		{ 347, 45, 33 },
		{ 347, 46, 33 },
		{ 347, 47, 33 },
		{ 347, 48, 33 },
		{ 347, 49, 33 },
		{ 347, 50, 33 },
		{ 347, 51, 33 },
		{ 347, 52, 33 },
		{ 347, 53, 33 },
		{ 347, 54, 33 },
		{ 347, 55, 33 },
		{ 347, 56, 33 },
		{ 347, 57, 33 },
		{ 347, 58, 33 },
		{ 347, 59, 33 },
		{ 347, 60, 33 },
		{ 347, 61, 33 },
		{ 347, 62, 33 },
		{ 347, 63, 33 },
		{ 347, 64, 33 },
		{ 347, 65, 33 },
		{ 347, 66, 33 },
		{ 347, 67, 33 },
		{ 347, 68, 33 },
		{ 347, 69, 33 },
		{ 347, 70, 33 },
		{ 347, 71, 33 },
		{ 347, 72, 33 },
		{ 347, 73, 33 },
		{ 347, 74, 33 },
		{ 347, 75, 33 },
		{ 347, 76, 33 },
		{ 348, 1, 33 },
		{ 348, 2, 33 },
		{ 348, 3, 33 },
		{ 348, 4, 33 },
		{ 348, 5, 33 },
		{ 348, 6, 33 },
		{ 348, 7, 33 },
		{ 348, 8, 33 },
		{ 348, 9, 33 },
		{ 348, 10, 33 },
		{ 348, 11, 33 },
		{ 348, 12, 33 },
		{ 348, 13, 33 },
		{ 348, 14, 33 },
		{ 348, 15, 33 },
		{ 348, 16, 33 },
		{ 348, 17, 33 },
		{ 348, 18, 33 },
		{ 348, 19, 33 },
		{ 348, 20, 33 },
		{ 348, 21, 33 },
		{ 348, 22, 33 },
		{ 348, 23, 33 },
		{ 348, 24, 33 },
		{ 348, 25, 33 },
		{ 348, 26, 33 },
		{ 348, 27, 33 },
		{ 348, 28, 33 },
		{ 348, 29, 33 },
		{ 348, 30, 33 },
		{ 348, 31, 33 },
		{ 348, 32, 33 },
		{ 348, 33, 33 },
		{ 348, 34, 33 },
		{ 348, 35, 33 },
		{ 348, 36, 33 },
		{ 348, 37, 33 },
		{ 348, 38, 33 },
		{ 348, 39, 33 },
		{ 348, 40, 33 },
		{ 348, 41, 33 },
		{ 348, 42, 33 },
		{ 348, 43, 33 },
		{ 348, 44, 33 },
		{ 348, 45, 33 },
		{ 348, 46, 33 },
		{ 348, 47, 33 },
		{ 348, 48, 33 },
		{ 348, 49, 33 },
		{ 348, 50, 33 },
		{ 348, 51, 33 },
		{ 348, 52, 33 },
		{ 348, 53, 33 },
		{ 348, 54, 33 },
		{ 348, 55, 33 },
		{ 348, 56, 33 },
		{ 348, 57, 33 },
		{ 348, 58, 33 },
		{ 348, 59, 33 },
		{ 348, 60, 33 },
		{ 348, 61, 33 },
		{ 348, 62, 33 },
		{ 348, 63, 33 },
		{ 348, 64, 33 },
		{ 348, 65, 33 },
		{ 348, 66, 33 },
		{ 348, 67, 33 },
		{ 348, 68, 33 },
		{ 348, 69, 33 },
		{ 348, 70, 33 },
		{ 348, 71, 33 },
		{ 348, 72, 33 },
		{ 348, 73, 33 },
		{ 348, 74, 33 },
		{ 349, 1, 33 },
		{ 349, 2, 33 },
		{ 349, 3, 33 },
		{ 349, 4, 33 },
		{ 349, 5, 33 },
		{ 349, 6, 33 },
		{ 349, 7, 33 },
		{ 349, 8, 33 },
		{ 349, 9, 33 },
		{ 349, 10, 33 },
		{ 349, 11, 33 },
		{ 349, 12, 33 },
		{ 349, 13, 33 },
		{ 349, 14, 33 },
		{ 349, 15, 33 },
		{ 349, 16, 33 },
		{ 349, 17, 33 },
		{ 349, 18, 33 },
		{ 349, 19, 33 },
		{ 349, 20, 33 },
		{ 349, 21, 33 },
		{ 349, 22, 33 },
		{ 349, 23, 33 },
		{ 349, 24, 33 },
		{ 349, 25, 33 },
		{ 349, 26, 33 },
		{ 349, 27, 33 },
		{ 349, 28, 33 },
		{ 349, 29, 33 },
		{ 349, 30, 33 },
		{ 349, 31, 33 },
		{ 349, 32, 33 },
		{ 349, 33, 33 },
		{ 349, 34, 33 },
		{ 349, 35, 33 },
		{ 349, 36, 33 },
		{ 349, 37, 33 },
		{ 349, 38, 33 },
		{ 349, 39, 33 },
		{ 349, 40, 33 },
		{ 349, 41, 33 },
		{ 349, 42, 33 },
		{ 349, 43, 33 },
		{ 349, 44, 33 },
		{ 349, 45, 33 },
		{ 349, 46, 33 },
		{ 349, 47, 33 },
		{ 349, 48, 33 },
		{ 349, 49, 33 },
		{ 349, 50, 33 },
		{ 349, 51, 33 },
		{ 349, 52, 33 },
		{ 349, 53, 33 },
		{ 349, 54, 33 },
		{ 349, 55, 33 },
		{ 349, 56, 33 },
		{ 349, 57, 33 },
		{ 349, 58, 33 },
		{ 349, 59, 33 },
		{ 349, 60, 33 },
		{ 349, 61, 33 },
		{ 349, 62, 33 },
		{ 349, 63, 33 },
		{ 349, 64, 33 },
		{ 349, 65, 33 },
		{ 349, 66, 33 },
		{ 349, 67, 33 },
		{ 349, 68, 33 },
		{ 349, 69, 33 },
		{ 349, 70, 33 },
		{ 349, 71, 33 },
		{ 349, 72, 33 },
		{ 350, 1, 33 },
		{ 350, 2, 33 },
		{ 350, 3, 33 },
		{ 350, 4, 33 },
		{ 350, 5, 33 },
		{ 350, 6, 33 },
		{ 350, 7, 33 },
		{ 350, 8, 33 },
		{ 350, 9, 33 },
		{ 350, 10, 33 },
		{ 350, 11, 33 },
		{ 350, 12, 33 },
		{ 350, 13, 33 },
		{ 350, 14, 33 },
		{ 350, 15, 33 },
		{ 350, 16, 33 },
		{ 350, 17, 33 },
		{ 350, 18, 33 },
		{ 350, 19, 33 },
		{ 350, 20, 33 },
		{ 350, 21, 33 },
		{ 350, 22, 33 },
		{ 350, 23, 33 },
		{ 350, 24, 33 },
		{ 350, 25, 33 },
		{ 350, 26, 33 },
		{ 350, 27, 33 },
		{ 350, 28, 33 },
		{ 350, 29, 33 },
		{ 350, 30, 33 },
		{ 350, 31, 33 },
		{ 350, 32, 33 },
		{ 350, 33, 33 },
		{ 350, 34, 33 },
		{ 350, 35, 33 },
		{ 350, 36, 33 },
		{ 350, 37, 33 },
		{ 350, 38, 33 },
		{ 350, 39, 33 },
		{ 350, 40, 33 },
		{ 350, 41, 33 },
		{ 350, 42, 33 },
		{ 350, 43, 33 },
		{ 350, 44, 33 },
		{ 350, 45, 33 },
		{ 350, 46, 33 },
		{ 350, 47, 33 },
		{ 350, 48, 33 },
		{ 350, 49, 33 },
		{ 350, 50, 33 },
		{ 350, 51, 33 },
		{ 350, 52, 33 },
		{ 350, 53, 33 },
		{ 350, 54, 33 },
		{ 350, 55, 33 },
		{ 350, 56, 33 },
		{ 350, 57, 33 },
		{ 350, 58, 33 },
		{ 350, 59, 33 },
		{ 350, 60, 33 },
		{ 350, 61, 33 },
		{ 350, 62, 33 },
		{ 350, 63, 33 },
		{ 350, 64, 33 },
		{ 350, 65, 33 },
		{ 350, 66, 33 },
		{ 350, 67, 33 },
		{ 350, 68, 33 },
		{ 350, 69, 33 },
		{ 350, 70, 33 },
		{ 351, 1, 33 },
		{ 351, 2, 33 },
		{ 351, 3, 33 },
		{ 351, 4, 33 },
		{ 351, 5, 33 },
		{ 351, 6, 33 },
		{ 351, 7, 33 },
		{ 351, 8, 33 },
		{ 351, 9, 33 },
		{ 351, 10, 33 },
		{ 351, 11, 33 },
		{ 351, 12, 33 },
		{ 351, 13, 33 },
		{ 351, 14, 33 },
		{ 351, 15, 33 },
		{ 351, 16, 33 },
		{ 351, 17, 33 },
		{ 351, 18, 33 },
		{ 351, 19, 33 },
		{ 351, 20, 33 },
		{ 351, 21, 33 },
		{ 351, 22, 33 },
		{ 351, 23, 33 },
		{ 351, 24, 33 },
		{ 351, 25, 33 },
		{ 351, 26, 33 },
		{ 351, 27, 33 },
		{ 351, 28, 33 },
		{ 351, 29, 33 },
		{ 351, 30, 33 },
		{ 351, 31, 33 },
		{ 351, 32, 33 },
		{ 351, 33, 33 },
		{ 351, 34, 33 },
		{ 351, 35, 33 },
		{ 351, 36, 33 },
		{ 351, 37, 33 },
		{ 351, 38, 33 },
		{ 351, 39, 33 },
		{ 351, 40, 33 },
		{ 351, 41, 33 },
		{ 351, 42, 33 },
		{ 351, 43, 33 },
		{ 351, 44, 33 },
		{ 351, 45, 33 },
		{ 351, 46, 33 },
		{ 351, 47, 33 },
		{ 351, 48, 33 },
		{ 351, 49, 33 },
		{ 351, 50, 33 },
		{ 351, 51, 33 },
		{ 351, 52, 33 },
		{ 351, 53, 33 },
		{ 351, 54, 33 },
		{ 351, 55, 33 },
		{ 351, 56, 33 },
		{ 351, 57, 33 },
		{ 351, 58, 33 },
		{ 351, 59, 33 },
		{ 351, 60, 33 },
		{ 351, 61, 33 },
		{ 351, 62, 33 },
		{ 351, 63, 33 },
		{ 351, 64, 33 },
		{ 351, 65, 33 },
		{ 351, 66, 33 },
		{ 351, 67, 33 },
		{ 351, 68, 33 },
		{ 352, 1, 33 },
		{ 352, 2, 33 },
		{ 352, 3, 33 },
		{ 352, 4, 33 },
		{ 352, 5, 33 },
		{ 352, 6, 33 },
		{ 352, 7, 33 },
		{ 352, 8, 33 },
		{ 352, 9, 33 },
		{ 352, 10, 33 },
		{ 352, 11, 33 },
		{ 352, 12, 33 },
		{ 352, 13, 33 },
		{ 352, 14, 33 },
		{ 352, 15, 33 },
		{ 352, 16, 33 },
		{ 352, 17, 33 },
		{ 352, 18, 33 },
		{ 352, 19, 33 },
		{ 352, 20, 33 },
		{ 352, 21, 33 },
		{ 352, 22, 33 },
		{ 352, 23, 33 },
		{ 352, 24, 33 },
		{ 352, 25, 33 },
		{ 352, 26, 33 },
		{ 352, 27, 33 },
		{ 352, 28, 33 },
		{ 352, 29, 33 },
		{ 352, 30, 33 },
		{ 352, 31, 33 },
		{ 352, 32, 33 },
		{ 352, 33, 33 },
		{ 352, 34, 33 },
		{ 352, 35, 33 },
		{ 352, 36, 33 },
		{ 352, 37, 33 },
		{ 352, 38, 33 },
		{ 352, 39, 33 },
		{ 352, 40, 33 },
		{ 352, 41, 33 },
		{ 352, 42, 33 },
		{ 352, 43, 33 },
		{ 352, 44, 33 },
		{ 352, 45, 33 },
		{ 352, 46, 33 },
		{ 352, 47, 33 },
		{ 352, 48, 33 },
		{ 352, 49, 33 },
		{ 352, 50, 33 },
		{ 352, 51, 33 },
		{ 352, 52, 33 },
		{ 352, 53, 33 },
		{ 352, 54, 33 },
		{ 352, 55, 33 },
		{ 352, 56, 33 },
		{ 352, 57, 33 },
		{ 352, 58, 33 },
		{ 352, 59, 33 },
		{ 352, 60, 33 },
		{ 352, 61, 33 },
		{ 352, 62, 33 },
		{ 352, 63, 33 },
		{ 352, 64, 33 },
		{ 352, 65, 33 },
		{ 352, 66, 33 },
		{ 353, 1, 33 },
		{ 353, 2, 33 },
		{ 353, 3, 33 },
		{ 353, 4, 33 },
		{ 353, 5, 33 },
		{ 353, 6, 33 },
		{ 353, 7, 33 },
		{ 353, 8, 33 },
		{ 353, 9, 33 },
		{ 353, 10, 33 },
		{ 353, 11, 33 },
		{ 353, 12, 33 },
		{ 353, 13, 33 },
		{ 353, 14, 33 },
		{ 353, 15, 33 },
		{ 353, 16, 33 },
		{ 353, 17, 33 },
		{ 353, 18, 33 },
		{ 353, 19, 33 },
		{ 353, 20, 33 },
		{ 353, 21, 33 },
		{ 353, 22, 33 },
		{ 353, 23, 33 },
		{ 353, 24, 33 },
		{ 353, 25, 33 },
		{ 353, 26, 33 },
		{ 353, 27, 33 },
		{ 353, 28, 33 },
		{ 353, 29, 33 },
		{ 353, 30, 33 },
		{ 353, 31, 33 },
		{ 353, 32, 33 },
		{ 353, 33, 33 },
		{ 353, 34, 33 },
		{ 353, 35, 33 },
		{ 353, 36, 33 },
		{ 353, 37, 33 },
		{ 353, 38, 33 },
		{ 353, 39, 33 },
		{ 353, 40, 33 },
		{ 353, 41, 33 },
		{ 353, 42, 33 },
		{ 353, 43, 33 },
		{ 353, 44, 33 },
		{ 353, 45, 33 },
		{ 353, 46, 33 },
		{ 353, 47, 33 },
		{ 353, 48, 33 },
		{ 353, 49, 33 },
		{ 353, 50, 33 },
		{ 353, 51, 33 },
		{ 353, 52, 33 },
		{ 353, 53, 33 },
		{ 353, 54, 33 },
		{ 353, 55, 33 },
		{ 353, 56, 33 },
		{ 353, 57, 33 },
		{ 353, 58, 33 },
		{ 353, 59, 33 },
		{ 353, 60, 33 },
		{ 353, 61, 33 },
		{ 353, 62, 33 },
		{ 353, 63, 33 },
		{ 353, 64, 33 },
		{ 354, 1, 33 },
		{ 354, 2, 33 },
		{ 354, 3, 33 },
		{ 354, 4, 33 },
		{ 354, 5, 33 },
		{ 354, 6, 33 },
		{ 354, 7, 33 },
		{ 354, 8, 33 },
		{ 354, 9, 33 },
		{ 354, 10, 33 },
		{ 354, 11, 33 },
		{ 354, 12, 33 },
		{ 354, 13, 33 },
		{ 354, 14, 33 },
		{ 354, 15, 33 },
		{ 354, 16, 33 },
		{ 354, 17, 33 },
		{ 354, 18, 33 },
		{ 354, 19, 33 },
		{ 354, 20, 33 },
		{ 354, 21, 33 },
		{ 354, 22, 33 },
		{ 354, 23, 33 },
		{ 354, 24, 33 },
		{ 354, 25, 33 },
		{ 354, 26, 33 },
		{ 354, 27, 33 },
		{ 354, 28, 33 },
		{ 354, 29, 33 },
		{ 354, 30, 33 },
		{ 354, 31, 33 },
		{ 354, 32, 33 },
		{ 354, 33, 33 },
		{ 354, 34, 33 },
		{ 354, 35, 33 },
		{ 354, 36, 33 },
		{ 354, 37, 33 },
		{ 354, 38, 33 },
		{ 354, 39, 33 },
		{ 354, 40, 33 },
		{ 354, 41, 33 },
		{ 354, 42, 33 },
		{ 354, 43, 33 },
		{ 354, 44, 33 },
		{ 354, 45, 33 },
		{ 354, 46, 33 },
		{ 354, 47, 33 },
		{ 354, 48, 33 },
		{ 354, 49, 33 },
		{ 354, 50, 33 },
		{ 354, 51, 33 },
		{ 354, 52, 33 },
		{ 354, 53, 33 },
		{ 354, 54, 33 },
		{ 354, 55, 33 },
		{ 354, 56, 33 },
		{ 354, 57, 33 },
		{ 354, 58, 33 },
		{ 354, 59, 33 },
		{ 354, 60, 33 },
		{ 354, 61, 33 },
		{ 354, 62, 33 },
		{ 355, 1, 33 },
		{ 355, 2, 33 },
		{ 355, 3, 33 },
		{ 355, 4, 33 },
		{ 355, 5, 33 },
		{ 355, 6, 33 },
		{ 355, 7, 33 },
		{ 355, 8, 33 },
		{ 355, 9, 33 },
		{ 355, 10, 33 },
		{ 355, 11, 33 },
		{ 355, 12, 33 },
		{ 355, 13, 33 },
		{ 355, 14, 33 },
		{ 355, 15, 33 },
		{ 355, 16, 33 },
		{ 355, 17, 33 },
		{ 355, 18, 33 },
		{ 355, 19, 33 },
		{ 355, 20, 33 },
		{ 355, 21, 33 },
		{ 355, 22, 33 },
		{ 355, 23, 33 },
		{ 355, 24, 33 },
		{ 355, 25, 33 },
		{ 355, 26, 33 },
		{ 355, 27, 33 },
		{ 355, 28, 33 },
		{ 355, 29, 33 },
		{ 355, 30, 33 },
		{ 355, 31, 33 },
		{ 355, 32, 33 },
		{ 355, 33, 33 },
		{ 355, 34, 33 },
		{ 355, 35, 33 },
		{ 355, 36, 33 },
		{ 355, 37, 33 },
		{ 355, 38, 33 },
		{ 355, 39, 33 },
		{ 355, 40, 33 },
		{ 355, 41, 33 },
		{ 355, 42, 33 },
		{ 355, 43, 33 },
		{ 355, 44, 33 },
		{ 355, 45, 33 },
		{ 355, 46, 33 },
		{ 355, 47, 33 },
		{ 355, 48, 33 },
		{ 355, 49, 33 },
		{ 355, 50, 33 },
		{ 355, 51, 33 },
		{ 355, 52, 33 },
		{ 355, 53, 33 },
		{ 355, 54, 33 },
		{ 355, 55, 33 },
		{ 355, 56, 33 },
		{ 355, 57, 33 },
		{ 355, 58, 33 },
		{ 355, 59, 33 },
		{ 355, 60, 33 },
		{ 356, 1, 33 },
		{ 356, 2, 33 },
		{ 356, 3, 33 },
		{ 356, 4, 33 },
		{ 356, 5, 33 },
		{ 356, 6, 33 },
		{ 356, 7, 33 },
		{ 356, 8, 33 },
		{ 356, 9, 33 },
		{ 356, 10, 33 },
		{ 356, 11, 33 },
		{ 356, 12, 33 },
		{ 356, 13, 33 },
		{ 356, 14, 33 },
		{ 356, 15, 33 },
		{ 356, 16, 33 },
		{ 356, 17, 33 },
		{ 356, 18, 33 },
		{ 356, 19, 33 },
		{ 356, 20, 33 },
		{ 356, 21, 33 },
		{ 356, 22, 33 },
		{ 356, 23, 33 },
		{ 356, 24, 33 },
		{ 356, 25, 33 },
		{ 356, 26, 33 },
		{ 356, 27, 33 },
		{ 356, 28, 33 },
		{ 356, 29, 33 },
		{ 356, 30, 33 },
		{ 356, 31, 33 },
		{ 356, 32, 33 },
		{ 356, 33, 33 },
		{ 356, 34, 33 },
		{ 356, 35, 33 },
		{ 356, 36, 33 },
		{ 356, 37, 33 },
		{ 356, 38, 33 },
		{ 356, 39, 33 },
		{ 356, 40, 33 },
		{ 356, 41, 33 },
		{ 356, 42, 33 },
		{ 356, 43, 33 },
		{ 356, 44, 33 },
		{ 356, 45, 33 },
		{ 356, 46, 33 },
		{ 356, 47, 33 },
		{ 356, 48, 33 },
		{ 356, 49, 33 },
		{ 356, 50, 33 },
		{ 356, 51, 33 },
		{ 356, 52, 33 },
		{ 356, 53, 33 },
		{ 356, 54, 33 },
		{ 356, 55, 33 },
		{ 356, 56, 33 },
		{ 356, 57, 33 },
		{ 356, 58, 33 },
		{ 357, 1, 33 },
		{ 357, 2, 33 },
		{ 357, 3, 33 },
		{ 357, 4, 33 },
		{ 357, 5, 33 },
		{ 357, 6, 33 },
		{ 357, 7, 33 },
		{ 357, 8, 33 },
		{ 357, 9, 33 },
		{ 357, 10, 33 },
		{ 357, 11, 33 },
		{ 357, 12, 33 },
		{ 357, 13, 33 },
		{ 357, 14, 33 },
		{ 357, 15, 33 },
		{ 357, 16, 33 },
		{ 357, 17, 33 },
		{ 357, 18, 33 },
		{ 357, 19, 33 },
		{ 357, 20, 33 },
		{ 357, 21, 33 },
		{ 357, 22, 33 },
		{ 357, 23, 33 },
		{ 357, 24, 33 },
		{ 357, 25, 33 },
		{ 357, 26, 33 },
		{ 357, 27, 33 },
		{ 357, 28, 33 },
		{ 357, 29, 33 },
		{ 357, 30, 33 },
		{ 357, 31, 33 },
		{ 357, 32, 33 },
		{ 357, 33, 33 },
		{ 357, 34, 33 },
		{ 357, 35, 33 },
		{ 357, 36, 33 },
		{ 357, 37, 33 },
		{ 357, 38, 33 },
		{ 357, 39, 33 },
		{ 357, 40, 33 },
		{ 357, 41, 33 },
		{ 357, 42, 33 },
		{ 357, 43, 33 },
		{ 357, 44, 33 },
		{ 357, 45, 33 },
		{ 357, 46, 33 },
		{ 357, 47, 33 },
		{ 357, 48, 33 },
		{ 357, 49, 33 },
		{ 357, 50, 33 },
		{ 357, 51, 33 },
		{ 357, 52, 33 },
		{ 357, 53, 33 },
		{ 357, 54, 33 },
		{ 357, 55, 33 },
		{ 357, 56, 33 },
		{ 358, 1, 33 },
		{ 358, 2, 33 },
		{ 358, 3, 33 },
		{ 358, 4, 33 },
		{ 358, 5, 33 },
		{ 358, 6, 33 },
		{ 358, 7, 33 },
		{ 358, 8, 33 },
		{ 358, 9, 33 },
		{ 358, 10, 33 },
		{ 358, 11, 33 },
		{ 358, 12, 33 },
		{ 358, 13, 33 },
		{ 358, 14, 33 },
		{ 358, 15, 33 },
		{ 358, 16, 33 },
		{ 358, 17, 33 },
		{ 358, 18, 33 },
		{ 358, 19, 33 },
		{ 358, 20, 33 },
		{ 358, 21, 33 },
		{ 358, 22, 33 },
		{ 358, 23, 33 },
		{ 358, 24, 33 },
		{ 358, 25, 33 },
		{ 358, 26, 33 },
		{ 358, 27, 33 },
		{ 358, 28, 33 },
		{ 358, 29, 33 },
		{ 358, 30, 33 },
		{ 358, 31, 33 },
		{ 358, 32, 33 },
		{ 358, 33, 33 },
		{ 358, 34, 33 },
		{ 358, 35, 33 },
		{ 358, 36, 33 },
		{ 358, 37, 33 },
		{ 358, 38, 33 },
		{ 358, 39, 33 },
		{ 358, 40, 33 },
		{ 358, 41, 33 },
		{ 358, 42, 33 },
		{ 358, 43, 33 },
		{ 358, 44, 33 },
		{ 358, 45, 33 },
		{ 358, 46, 33 },
		{ 358, 47, 33 },
		{ 358, 48, 33 },
		{ 358, 49, 33 },
		{ 358, 50, 33 },
		{ 358, 51, 33 },
		{ 358, 52, 33 },
		{ 358, 53, 33 },
		{ 358, 54, 33 },
		{ 359, 1, 33 },
		{ 359, 2, 33 },
		{ 359, 3, 33 },
		{ 359, 4, 33 },
		{ 359, 5, 33 },
		{ 359, 6, 33 },
		{ 359, 7, 33 },
		{ 359, 8, 33 },
		{ 359, 9, 33 },
		{ 359, 10, 33 },
		{ 359, 11, 33 },
		{ 359, 12, 33 },
		{ 359, 13, 33 },
		{ 359, 14, 33 },
		{ 359, 15, 33 },
		{ 359, 16, 33 },
		{ 359, 17, 33 },
		{ 359, 18, 33 },
		{ 359, 19, 33 },
		{ 359, 20, 33 },
		{ 359, 21, 33 },
		{ 359, 22, 33 },
		{ 359, 23, 33 },
		{ 359, 24, 33 },
		{ 359, 25, 33 },
		{ 359, 26, 33 },
		{ 359, 27, 33 },
		{ 359, 28, 33 },
		{ 359, 29, 33 },
		{ 359, 30, 33 },
		{ 359, 31, 33 },
		{ 359, 32, 33 },
		{ 359, 33, 33 },
		{ 359, 34, 33 },
		{ 359, 35, 33 },
		{ 359, 36, 33 },
		{ 359, 37, 33 },
		{ 359, 38, 33 },
		{ 359, 39, 33 },
		{ 359, 40, 33 },
		{ 359, 41, 33 },
		{ 359, 42, 33 },
		{ 359, 43, 33 },
		{ 359, 44, 33 },
		{ 359, 45, 33 },
		{ 359, 46, 33 },
		{ 359, 47, 33 },
		{ 359, 48, 33 },
		{ 359, 49, 33 },
		{ 359, 50, 33 },
		{ 359, 51, 33 },
		{ 359, 52, 33 },
		{ 360, 1, 33 },
		{ 360, 2, 33 },
		{ 360, 3, 33 },
		{ 360, 4, 33 },
		{ 360, 5, 33 },
		{ 360, 6, 33 },
		{ 360, 7, 33 },
		{ 360, 8, 33 },
		{ 360, 9, 33 },
		{ 360, 10, 33 },
		{ 360, 11, 33 },
		{ 360, 12, 33 },
		{ 360, 13, 33 },
		{ 360, 14, 33 },
		{ 360, 15, 33 },
		{ 360, 16, 33 },
		{ 360, 17, 33 },
		{ 360, 18, 33 },
		{ 360, 19, 33 },
		{ 360, 20, 33 },
		{ 360, 21, 33 },
		{ 360, 22, 33 },
		{ 360, 23, 33 },
		{ 360, 24, 33 },
		{ 360, 25, 33 },
		{ 360, 26, 33 },
		{ 360, 27, 33 },
		{ 360, 28, 33 },
		{ 360, 29, 33 },
		{ 360, 30, 33 },
		{ 360, 31, 33 },
		{ 360, 32, 33 },
		{ 360, 33, 33 },
		{ 360, 34, 33 },
		{ 360, 35, 33 },
		{ 360, 36, 33 },
		{ 360, 37, 33 },
		{ 360, 38, 33 },
		{ 360, 39, 33 },
		{ 360, 40, 33 },
		{ 360, 41, 33 },
		{ 360, 42, 33 },
		{ 360, 43, 33 },
		{ 360, 44, 33 },
		{ 360, 45, 33 },
		{ 360, 46, 33 },
		{ 360, 47, 33 },
		{ 360, 48, 33 },
		{ 360, 49, 33 },
		{ 360, 50, 33 },
		{ 361, 1, 33 },
		{ 361, 2, 33 },
		{ 361, 3, 33 },
		{ 361, 4, 33 },
		{ 361, 5, 33 },
		{ 361, 6, 33 },
		{ 361, 7, 33 },
		{ 361, 8, 33 },
		{ 361, 9, 33 },
		{ 361, 10, 33 },
		{ 361, 11, 33 },
		{ 361, 12, 33 },
		{ 361, 13, 33 },
		{ 361, 14, 33 },
		{ 361, 15, 33 },
		{ 361, 16, 33 },
		{ 361, 17, 33 },
		{ 361, 18, 33 },
		{ 361, 19, 33 },
		{ 361, 20, 33 },
		{ 361, 21, 33 },
		{ 361, 22, 33 },
		{ 361, 23, 33 },
		{ 361, 24, 33 },
		{ 361, 25, 33 },
		{ 361, 26, 33 },
		{ 361, 27, 33 },
		{ 361, 28, 33 },
		{ 361, 29, 33 },
		{ 361, 30, 33 },
		{ 361, 31, 33 },
		{ 361, 32, 33 },
		{ 361, 33, 33 },
		{ 361, 34, 33 },
		{ 361, 35, 33 },
		{ 361, 36, 33 },
		{ 361, 37, 33 },
		{ 361, 38, 33 },
		{ 361, 39, 33 },
		{ 361, 40, 33 },
		{ 361, 41, 33 },
		{ 361, 42, 33 },
		{ 361, 43, 33 },
		{ 361, 44, 33 },
		{ 361, 45, 33 },
		{ 361, 46, 33 },
		{ 361, 47, 33 },
		{ 361, 48, 33 },
		{ 362, 1, 33 },
		{ 362, 2, 33 },
		{ 362, 3, 33 },
		{ 362, 4, 33 },
		{ 362, 5, 33 },
		{ 362, 6, 33 },
		{ 362, 7, 33 },
		{ 362, 8, 33 },
		{ 362, 9, 33 },
		{ 362, 10, 33 },
		{ 362, 11, 33 },
		{ 362, 12, 33 },
		{ 362, 13, 33 },
		{ 362, 14, 33 },
		{ 362, 15, 33 },
		{ 362, 16, 33 },
		{ 362, 17, 33 },
		{ 362, 18, 33 },
		{ 362, 19, 33 },
		{ 362, 20, 33 },
		{ 362, 21, 33 },
		{ 362, 22, 33 },
		{ 362, 23, 33 },
		{ 362, 24, 33 },
		{ 362, 25, 33 },
		{ 362, 26, 33 },
		{ 362, 27, 33 },
		{ 362, 28, 33 },
		{ 362, 29, 33 },
		{ 362, 30, 33 },
		{ 362, 31, 33 },
		{ 362, 32, 33 },
		{ 362, 33, 33 },
		{ 362, 34, 33 },
		{ 362, 35, 33 },
		{ 362, 36, 33 },
		{ 362, 37, 33 },
		{ 362, 38, 33 },
		{ 362, 39, 33 },
		{ 362, 40, 33 },
		{ 362, 41, 33 },
		{ 362, 42, 33 },
		{ 362, 43, 33 },
		{ 362, 44, 33 },
		{ 362, 45, 33 },
		{ 362, 46, 33 },
		{ 363, 1, 33 },
		{ 363, 2, 33 },
		{ 363, 3, 33 },
		{ 363, 4, 33 },
		{ 363, 5, 33 },
		{ 363, 6, 33 },
		{ 363, 7, 33 },
		{ 363, 8, 33 },
		{ 363, 9, 33 },
		{ 363, 10, 33 },
		{ 363, 11, 33 },
		{ 363, 12, 33 },
		{ 363, 13, 33 },
		{ 363, 14, 33 },
		{ 363, 15, 33 },
		{ 363, 16, 33 },
		{ 363, 17, 33 },
		{ 363, 18, 33 },
		{ 363, 19, 33 },
		{ 363, 20, 33 },
		{ 363, 21, 33 },
		{ 363, 22, 33 },
		{ 363, 23, 33 },
		{ 363, 24, 33 },
		{ 363, 25, 33 },
		{ 363, 26, 33 },
		{ 363, 27, 33 },
		{ 363, 28, 33 },
		{ 363, 29, 33 },
		{ 363, 30, 33 },
		{ 363, 31, 33 },
		{ 363, 32, 33 },
		{ 363, 33, 33 },
		{ 363, 34, 33 },
		{ 363, 35, 33 },
		{ 363, 36, 33 },
		{ 363, 37, 33 },
		{ 363, 38, 33 },
		{ 363, 39, 33 },
		{ 363, 40, 33 },
		{ 363, 41, 33 },
		{ 363, 42, 33 },
		{ 363, 43, 33 },
		{ 363, 44, 33 },
		{ 364, 1, 33 },
		{ 364, 2, 33 },
		{ 364, 3, 33 },
		{ 364, 4, 33 },
		{ 364, 5, 33 },
		{ 364, 6, 33 },
		{ 364, 7, 33 },
		{ 364, 8, 33 },
		{ 364, 9, 33 },
		{ 364, 10, 33 },
		{ 364, 11, 33 },
		{ 364, 12, 33 },
		{ 364, 13, 33 },
		{ 364, 14, 33 },
		{ 364, 15, 33 },
		{ 364, 16, 33 },
		{ 364, 17, 33 },
		{ 364, 18, 33 },
		{ 364, 19, 33 },
		{ 364, 20, 33 },
		{ 364, 21, 33 },
		{ 364, 22, 33 },
		{ 364, 23, 33 },
		{ 364, 24, 33 },
		{ 364, 25, 33 },
		{ 364, 26, 33 },
		{ 364, 27, 33 },
		{ 364, 28, 33 },
		{ 364, 29, 33 },
		{ 364, 30, 33 },
		{ 364, 31, 33 },
		{ 364, 32, 33 },
		{ 364, 33, 33 },
		{ 364, 34, 33 },
		{ 364, 35, 33 },
		{ 364, 36, 33 },
		{ 364, 37, 33 },
		{ 364, 38, 33 },
		{ 364, 39, 33 },
		{ 364, 40, 33 },
		{ 364, 41, 33 },
		{ 364, 42, 33 },
		{ 365, 1, 33 },
		{ 365, 2, 33 },
		{ 365, 3, 33 },
		{ 365, 4, 33 },
		{ 365, 5, 33 },
		{ 365, 6, 33 },
		{ 365, 7, 33 },
		{ 365, 8, 33 },
		{ 365, 9, 33 },
		{ 365, 10, 33 },
		{ 365, 11, 33 },
		{ 365, 12, 33 },
		{ 365, 13, 33 },
		{ 365, 14, 33 },
		{ 365, 15, 33 },
		{ 365, 16, 33 },
		{ 365, 17, 33 },
		{ 365, 18, 33 },
		{ 365, 19, 33 },
		{ 365, 20, 33 },
		{ 365, 21, 33 },
		{ 365, 22, 33 },
		{ 365, 23, 33 },
		{ 365, 24, 33 },
		{ 365, 25, 33 },
		{ 365, 26, 33 },
		{ 365, 27, 33 },
		{ 365, 28, 33 },
		{ 365, 29, 33 },
		{ 365, 30, 33 },
		{ 365, 31, 33 },
		{ 365, 32, 33 },
		{ 365, 33, 33 },
		{ 365, 34, 33 },
		{ 365, 35, 33 },
		{ 365, 36, 33 },
		{ 365, 37, 33 },
		{ 365, 38, 33 },
		{ 365, 39, 33 },
		{ 365, 40, 33 },
		{ 366, 1, 33 },
		{ 366, 2, 33 },
		{ 366, 3, 33 },
		{ 366, 4, 33 },
		{ 366, 5, 33 },
		{ 366, 6, 33 },
		{ 366, 7, 33 },
		{ 366, 8, 33 },
		{ 366, 9, 33 },
		{ 366, 10, 33 },
		{ 366, 11, 33 },
		{ 366, 12, 33 },
		{ 366, 13, 33 },
		{ 366, 14, 33 },
		{ 366, 15, 33 },
		{ 366, 16, 33 },
		{ 366, 17, 33 },
		{ 366, 18, 33 },
		{ 366, 19, 33 },
		{ 366, 20, 33 },
		{ 366, 21, 33 },
		{ 366, 22, 33 },
		{ 366, 23, 33 },
		{ 366, 24, 33 },
		{ 366, 25, 33 },
		{ 366, 26, 33 },
		{ 366, 27, 33 },
		{ 366, 28, 33 },
		{ 366, 29, 33 },
		{ 366, 30, 33 },
		{ 366, 31, 33 },
		{ 366, 32, 33 },
		{ 366, 33, 33 },
		{ 366, 34, 33 },
		{ 366, 35, 33 },
		{ 366, 36, 33 },
		{ 366, 37, 33 },
		{ 366, 38, 33 },
		{ 367, 1, 33 },
		{ 367, 2, 33 },
		{ 367, 3, 33 },
		{ 367, 4, 33 },
		{ 367, 5, 33 },
		{ 367, 6, 33 },
		{ 367, 7, 33 },
		{ 367, 8, 33 },
		{ 367, 9, 33 },
		{ 367, 10, 33 },
		{ 367, 11, 33 },
		{ 367, 12, 33 },
		{ 367, 13, 33 },
		{ 367, 14, 33 },
		{ 367, 15, 33 },
		{ 367, 16, 33 },
		{ 367, 17, 33 },
		{ 367, 18, 33 },
		{ 367, 19, 33 },
		{ 367, 20, 33 },
		{ 367, 21, 33 },
		{ 367, 22, 33 },
		{ 367, 23, 33 },
		{ 367, 24, 33 },
		{ 367, 25, 33 },
		{ 367, 26, 33 },
		{ 367, 27, 33 },
		{ 367, 28, 33 },
		{ 367, 29, 33 },
		{ 367, 30, 33 },
		{ 367, 31, 33 },
		{ 367, 32, 33 },
		{ 367, 33, 33 },
		{ 367, 34, 33 },
		{ 367, 35, 33 },
		{ 367, 36, 33 },
		{ 368, 1, 33 },
		{ 368, 2, 33 },
		{ 368, 3, 33 },
		{ 368, 4, 33 },
		{ 368, 5, 33 },
		{ 368, 6, 33 },
		{ 368, 7, 33 },
		{ 368, 8, 33 },
		{ 368, 9, 33 },
		{ 368, 10, 33 },
		{ 368, 11, 33 },
		{ 368, 12, 33 },
		{ 368, 13, 33 },
		{ 368, 14, 33 },
		{ 368, 15, 33 },
		{ 368, 16, 33 },
		{ 368, 17, 33 },
		{ 368, 18, 33 },
		{ 368, 19, 33 },
		{ 368, 20, 33 },
		{ 368, 21, 33 },
		{ 368, 22, 33 },
		{ 368, 23, 33 },
		{ 368, 24, 33 },
		{ 368, 25, 33 },
		{ 368, 26, 33 },
		{ 368, 27, 33 },
		{ 368, 28, 33 },
		{ 368, 29, 33 },
		{ 368, 30, 33 },
		{ 368, 31, 33 },
		{ 368, 32, 33 },
		{ 368, 33, 33 },
		{ 368, 34, 33 },
		{ 369, 1, 33 },
		{ 369, 2, 33 },
		{ 369, 3, 33 },
		{ 369, 4, 33 },
		{ 369, 5, 33 },
		{ 369, 6, 33 },
		{ 369, 7, 33 },
		{ 369, 8, 33 },
		{ 369, 9, 33 },
		{ 369, 10, 33 },
		{ 369, 11, 33 },
		{ 369, 12, 33 },
		{ 369, 13, 33 },
		{ 369, 14, 33 },
		{ 369, 15, 33 },
		{ 369, 16, 33 },
		{ 369, 17, 33 },
		{ 369, 18, 33 },
		{ 369, 19, 33 },
		{ 369, 20, 33 },
		{ 369, 21, 33 },
		{ 369, 22, 33 },
		{ 369, 23, 33 },
		{ 369, 24, 33 },
		{ 369, 25, 33 },
		{ 369, 26, 33 },
		{ 369, 27, 33 },
		{ 369, 28, 33 },
		{ 369, 29, 33 },
		{ 369, 30, 33 },
		{ 369, 31, 33 },
		{ 369, 32, 33 },
		{ 370, 1, 33 },
		{ 370, 2, 33 },
		{ 370, 3, 33 },
		{ 370, 4, 33 },
		{ 370, 5, 33 },
		{ 370, 6, 33 },
		{ 370, 7, 33 },
		{ 370, 8, 33 },
		{ 370, 9, 33 },
		{ 370, 10, 33 },
		{ 370, 11, 33 },
		{ 370, 12, 33 },
		{ 370, 13, 33 },
		{ 370, 14, 33 },
		{ 370, 15, 33 },
		{ 370, 16, 33 },
		{ 370, 17, 33 },
		{ 370, 18, 33 },
		{ 370, 19, 33 },
		{ 370, 20, 33 },
		{ 370, 21, 33 },
		{ 370, 22, 33 },
		{ 370, 23, 33 },
		{ 370, 24, 33 },
		{ 370, 25, 33 },
		{ 370, 26, 33 },
		{ 370, 27, 33 },
		{ 370, 28, 33 },
		{ 370, 29, 33 },
		{ 370, 30, 33 },
		{ 371, 1, 33 },
		{ 371, 2, 33 },
		{ 371, 3, 33 },
		{ 371, 4, 33 },
		{ 371, 5, 33 },
		{ 371, 6, 33 },
		{ 371, 7, 33 },
		{ 371, 8, 33 },
		{ 371, 9, 33 },
		{ 371, 10, 33 },
		{ 371, 11, 33 },
		{ 371, 12, 33 },
		{ 371, 13, 33 },
		{ 371, 14, 33 },
		{ 371, 15, 33 },
		{ 371, 16, 33 },
		{ 371, 17, 33 },
		{ 371, 18, 33 },
		{ 371, 19, 33 },
		{ 371, 20, 33 },
		{ 371, 21, 33 },
		{ 371, 22, 33 },
		{ 371, 23, 33 },
		{ 371, 24, 33 },
		{ 371, 25, 33 },
		{ 371, 26, 33 },
		{ 371, 27, 33 },
		{ 371, 28, 33 },
		{ 372, 1, 33 },
		{ 372, 2, 33 },
		{ 372, 3, 33 },
		{ 372, 4, 33 },
		{ 372, 5, 33 },
		{ 372, 6, 33 },
		{ 372, 7, 33 },
		{ 372, 8, 33 },
		{ 372, 9, 33 },
		{ 372, 10, 33 },
		{ 372, 11, 33 },
		{ 372, 12, 33 },
		{ 372, 13, 33 },
		{ 372, 14, 33 },
		{ 372, 15, 33 },
		{ 372, 16, 33 },
		{ 372, 17, 33 },
		{ 372, 18, 33 },
		{ 372, 19, 33 },
		{ 372, 20, 33 },
		{ 372, 21, 33 },
		{ 372, 22, 33 },
		{ 372, 23, 33 },
		{ 372, 24, 33 },
		{ 372, 25, 33 },
		{ 372, 26, 33 },
		{ 373, 1, 33 },
		{ 373, 2, 33 },
		{ 373, 3, 33 },
		{ 373, 4, 33 },
		{ 373, 5, 33 },
		{ 373, 6, 33 },
		{ 373, 7, 33 },
		{ 373, 8, 33 },
		{ 373, 9, 33 },
		{ 373, 10, 33 },
		{ 373, 11, 33 },
		{ 373, 12, 33 },
		{ 373, 13, 33 },
		{ 373, 14, 33 },
		{ 373, 15, 33 },
		{ 373, 16, 33 },
		{ 373, 17, 33 },
		{ 373, 18, 33 },
		{ 373, 19, 33 },
		{ 373, 20, 33 },
		{ 373, 21, 33 },
		{ 373, 22, 33 },
		{ 373, 23, 33 },
		{ 373, 24, 33 },
		{ 374, 1, 33 },
		{ 374, 2, 33 },
		{ 374, 3, 33 },
		{ 374, 4, 33 },
		{ 374, 5, 33 },
		{ 374, 6, 33 },
		{ 374, 7, 33 },
		{ 374, 8, 33 },
		{ 374, 9, 33 },
		{ 374, 10, 33 },
		{ 374, 11, 33 },
		{ 374, 12, 33 },
		{ 374, 13, 33 },
		{ 374, 14, 33 },
		{ 374, 15, 33 },
		{ 374, 16, 33 },
		{ 374, 17, 33 },
		{ 374, 18, 33 },
		{ 374, 19, 33 },
		{ 374, 20, 33 },
		{ 374, 21, 33 },
		{ 374, 22, 33 },
		{ 375, 1, 33 },
		{ 375, 2, 33 },
		{ 375, 3, 33 },
		{ 375, 4, 33 },
		{ 375, 5, 33 },
		{ 375, 6, 33 },
		{ 375, 7, 33 },
		{ 375, 8, 33 },
		{ 375, 9, 33 },
		{ 375, 10, 33 },
		{ 375, 11, 33 },
		{ 375, 12, 33 },
		{ 375, 13, 33 },
		{ 375, 14, 33 },
		{ 375, 15, 33 },
		{ 375, 16, 33 },
		{ 375, 17, 33 },
		{ 375, 18, 33 },
		{ 375, 19, 33 },
		{ 375, 20, 33 },
		{ 376, 1, 33 },
		{ 376, 2, 33 },
		{ 376, 3, 33 },
		{ 376, 4, 33 },
		{ 376, 5, 33 },
		{ 376, 6, 33 },
		{ 376, 7, 33 },
		{ 376, 8, 33 },
		{ 376, 9, 33 },
		{ 376, 10, 33 },
		{ 376, 11, 33 },
		{ 376, 12, 33 },
		{ 376, 13, 33 },
		{ 376, 14, 33 },
		{ 376, 15, 33 },
		{ 376, 16, 33 },
		{ 376, 17, 33 },
		{ 376, 18, 33 },
		{ 377, 1, 33 },
		{ 377, 2, 33 },
		{ 377, 3, 33 },
		{ 377, 4, 33 },
		{ 377, 5, 33 },
		{ 377, 6, 33 },
		{ 377, 7, 33 },
		{ 377, 8, 33 },
		{ 377, 9, 33 },
		{ 377, 10, 33 },
		{ 377, 11, 33 },
		{ 377, 12, 33 },
		{ 377, 13, 33 },
		{ 377, 14, 33 },
		{ 377, 15, 33 },
		{ 377, 16, 33 },
		{ 378, 1, 33 },
		{ 378, 2, 33 },
		{ 378, 3, 33 },
		{ 378, 4, 33 },
		{ 378, 5, 33 },
		{ 378, 6, 33 },
		{ 378, 7, 33 },
		{ 378, 8, 33 },
		{ 378, 9, 33 },
		{ 378, 10, 33 },
		{ 378, 11, 33 },
		{ 378, 12, 33 },
		{ 378, 13, 33 },
		{ 378, 14, 33 },
		{ 379, 1, 33 },
		{ 379, 2, 33 },
		{ 379, 3, 33 },
		{ 379, 4, 33 },
		{ 379, 5, 33 },
		{ 379, 6, 33 },
		{ 379, 7, 33 },
		{ 379, 8, 33 },
		{ 379, 9, 33 },
		{ 379, 10, 33 },
		{ 379, 11, 33 },
		{ 379, 12, 33 },
		{ 380, 1, 33 },
		{ 380, 2, 33 },
		{ 380, 3, 33 },
		{ 380, 4, 33 },
		{ 380, 5, 33 },
		{ 380, 6, 33 },
		{ 380, 7, 33 },
		{ 380, 8, 33 },
		{ 380, 9, 33 },
		{ 380, 10, 33 },
		{ 381, 1, 33 },
		{ 381, 2, 33 },
		{ 381, 3, 33 },
		{ 381, 4, 33 },
		{ 381, 5, 33 },
		{ 381, 6, 33 },
		{ 381, 7, 33 },
		{ 381, 8, 33 },
		{ 382, 1, 33 },
		{ 382, 2, 33 },
		{ 382, 3, 33 },
		{ 382, 4, 33 },
		{ 382, 5, 33 },
		{ 382, 6, 33 },
		{ 383, 1, 33 },
		{ 383, 2, 33 },
		{ 383, 3, 33 },
		{ 383, 4, 33 },
		{ 384, 1, 33 },
		{ 384, 2, 33 },
		{ 98, 512, 34 },
		{ 99, 510, 34 },
		{ 99, 511, 34 },
		{ 99, 512, 34 },
		{ 100, 509, 34 },
		{ 100, 510, 34 },
		{ 100, 511, 34 },
		{ 100, 512, 34 },
		{ 101, 507, 34 },
		{ 101, 508, 34 },
		{ 101, 509, 34 },
		{ 101, 510, 34 },
		{ 101, 511, 34 },
		{ 101, 512, 34 },
		{ 102, 506, 34 },
		{ 102, 507, 34 },
		{ 102, 508, 34 },
		{ 102, 509, 34 },
		{ 102, 510, 34 },
		{ 102, 511, 34 },
		{ 102, 512, 34 },
		{ 103, 504, 34 },
		{ 103, 505, 34 },
		{ 103, 506, 34 },
		{ 103, 507, 34 },
		{ 103, 508, 34 },
		{ 103, 509, 34 },
		{ 103, 510, 34 },
		{ 103, 511, 34 },
		{ 103, 512, 34 },
		{ 104, 502, 34 },
		{ 104, 503, 34 },
		{ 104, 504, 34 },
		{ 104, 505, 34 },
		{ 104, 506, 34 },
		{ 104, 507, 34 },
		{ 104, 508, 34 },
		{ 104, 509, 34 },
		{ 104, 510, 34 },
		{ 104, 511, 34 },
		{ 104, 512, 34 },
		{ 105, 501, 34 },
		{ 105, 502, 34 },
		{ 105, 503, 34 },
		{ 105, 504, 34 },
		{ 105, 505, 34 },
		{ 105, 506, 34 },
		{ 105, 507, 34 },
		{ 105, 508, 34 },
		{ 105, 509, 34 },
		{ 105, 510, 34 },
		{ 105, 511, 34 },
		{ 105, 512, 34 },
		{ 106, 499, 34 },
		{ 106, 500, 34 },
		{ 106, 501, 34 },
		{ 106, 502, 34 },
		{ 106, 503, 34 },
		{ 106, 504, 34 },
		{ 106, 505, 34 },
		{ 106, 506, 34 },
		{ 106, 507, 34 },
		{ 106, 508, 34 },
		{ 106, 509, 34 },
		{ 106, 510, 34 },
		{ 106, 511, 34 },
		{ 106, 512, 34 },
		{ 107, 498, 34 },
		{ 107, 499, 34 },
		{ 107, 500, 34 },
		{ 107, 501, 34 },
		{ 107, 502, 34 },
		{ 107, 503, 34 },
		{ 107, 504, 34 },
		{ 107, 505, 34 },
		{ 107, 506, 34 },
		{ 107, 507, 34 },
		{ 107, 508, 34 },
		{ 107, 509, 34 },
		{ 107, 510, 34 },
		{ 107, 511, 34 },
		{ 107, 512, 34 },
		{ 108, 496, 34 },
		{ 108, 497, 34 },
		{ 108, 498, 34 },
		{ 108, 499, 34 },
		{ 108, 500, 34 },
		{ 108, 501, 34 },
		{ 108, 502, 34 },
		{ 108, 503, 34 },
		{ 108, 504, 34 },
		{ 108, 505, 34 },
		{ 108, 506, 34 },
		{ 108, 507, 34 },
		{ 108, 508, 34 },
		{ 108, 509, 34 },
		{ 108, 510, 34 },
		{ 108, 511, 34 },
		{ 108, 512, 34 },
		{ 109, 494, 34 },
		{ 109, 495, 34 },
		{ 109, 496, 34 },
		{ 109, 497, 34 },
		{ 109, 498, 34 },
		{ 109, 499, 34 },
		{ 109, 500, 34 },
		{ 109, 501, 34 },
		{ 109, 502, 34 },
		{ 109, 503, 34 },
		{ 109, 504, 34 },
		{ 109, 505, 34 },
		{ 109, 506, 34 },
		{ 109, 507, 34 },
		{ 109, 508, 34 },
		{ 109, 509, 34 },
		{ 109, 510, 34 },
		{ 109, 511, 34 },
		{ 109, 512, 34 },
		{ 110, 493, 34 },
		{ 110, 494, 34 },
		{ 110, 495, 34 },
		{ 110, 496, 34 },
		{ 110, 497, 34 },
		{ 110, 498, 34 },
		{ 110, 499, 34 },
		{ 110, 500, 34 },
		{ 110, 501, 34 },
		{ 110, 502, 34 },
		{ 110, 503, 34 },
		{ 110, 504, 34 },
		{ 110, 505, 34 },
		{ 110, 506, 34 },
		{ 110, 507, 34 },
		{ 110, 508, 34 },
		{ 110, 509, 34 },
		{ 110, 510, 34 },
		{ 110, 511, 34 },
		{ 110, 512, 34 },
		{ 111, 491, 34 },
		{ 111, 492, 34 },
		{ 111, 493, 34 },
		{ 111, 494, 34 },
		{ 111, 495, 34 },
		{ 111, 496, 34 },
		{ 111, 497, 34 },
		{ 111, 498, 34 },
		{ 111, 499, 34 },
		{ 111, 500, 34 },
		{ 111, 501, 34 },
		{ 111, 502, 34 },
		{ 111, 503, 34 },
		{ 111, 504, 34 },
		{ 111, 505, 34 },
		{ 111, 506, 34 },
		{ 111, 507, 34 },
		{ 111, 508, 34 },
		{ 111, 509, 34 },
		{ 111, 510, 34 },
		{ 111, 511, 34 },
		{ 111, 512, 34 },
		{ 112, 490, 34 },
		{ 112, 491, 34 },
		{ 112, 492, 34 },
		{ 112, 493, 34 },
		{ 112, 494, 34 },
		{ 112, 495, 34 },
		{ 112, 496, 34 },
		{ 112, 497, 34 },
		{ 112, 498, 34 },
		{ 112, 499, 34 },
		{ 112, 500, 34 },
		{ 112, 501, 34 },
		{ 112, 502, 34 },
		{ 112, 503, 34 },
		{ 112, 504, 34 },
		{ 112, 505, 34 },
		{ 112, 506, 34 },
		{ 112, 507, 34 },
		{ 112, 508, 34 },
		{ 112, 509, 34 },
		{ 112, 510, 34 },
		{ 112, 511, 34 },
		{ 112, 512, 34 },
		{ 113, 488, 34 },
		{ 113, 489, 34 },
		{ 113, 490, 34 },
		{ 113, 491, 34 },
		{ 113, 492, 34 },
		{ 113, 493, 34 },
		{ 113, 494, 34 },
		{ 113, 495, 34 },
		{ 113, 496, 34 },
		{ 113, 497, 34 },
		{ 113, 498, 34 },
		{ 113, 499, 34 },
		{ 113, 500, 34 },
		{ 113, 501, 34 },
		{ 113, 502, 34 },
		{ 113, 503, 34 },
		{ 113, 504, 34 },
		{ 113, 505, 34 },
		{ 113, 506, 34 },
		{ 113, 507, 34 },
		{ 113, 508, 34 },
		{ 113, 509, 34 },
		{ 113, 510, 34 },
		{ 113, 511, 34 },
		{ 113, 512, 34 },
		{ 114, 486, 34 },
		{ 114, 487, 34 },
		{ 114, 488, 34 },
		{ 114, 489, 34 },
		{ 114, 490, 34 },
		{ 114, 491, 34 },
		{ 114, 492, 34 },
		{ 114, 493, 34 },
		{ 114, 494, 34 },
		{ 114, 495, 34 },
		{ 114, 496, 34 },
		{ 114, 497, 34 },
		{ 114, 498, 34 },
		{ 114, 499, 34 },
		{ 114, 500, 34 },
		{ 114, 501, 34 },
		{ 114, 502, 34 },
		{ 114, 503, 34 },
		{ 114, 504, 34 },
		{ 114, 505, 34 },
		{ 114, 506, 34 },
		{ 114, 507, 34 },
		{ 114, 508, 34 },
		{ 114, 509, 34 },
		{ 114, 510, 34 },
		{ 114, 511, 34 },
		{ 114, 512, 34 },
		{ 115, 485, 34 },
		{ 115, 486, 34 },
		{ 115, 487, 34 },
		{ 115, 488, 34 },
		{ 115, 489, 34 },
		{ 115, 490, 34 },
		{ 115, 491, 34 },
		{ 115, 492, 34 },
		{ 115, 493, 34 },
		{ 115, 494, 34 },
		{ 115, 495, 34 },
		{ 115, 496, 34 },
		{ 115, 497, 34 },
		{ 115, 498, 34 },
		{ 115, 499, 34 },
		{ 115, 500, 34 },
		{ 115, 501, 34 },
		{ 115, 502, 34 },
		{ 115, 503, 34 },
		{ 115, 504, 34 },
		{ 115, 505, 34 },
		{ 115, 506, 34 },
		{ 115, 507, 34 },
		{ 115, 508, 34 },
		{ 115, 509, 34 },
		{ 115, 510, 34 },
		{ 115, 511, 34 },
		{ 115, 512, 34 },
		{ 116, 483, 34 },
		{ 116, 484, 34 },
		{ 116, 485, 34 },
		{ 116, 486, 34 },
		{ 116, 487, 34 },
		{ 116, 488, 34 },
		{ 116, 489, 34 },
		{ 116, 490, 34 },
		{ 116, 491, 34 },
		{ 116, 492, 34 },
		{ 116, 493, 34 },
		{ 116, 494, 34 },
		{ 116, 495, 34 },
		{ 116, 496, 34 },
		{ 116, 497, 34 },
		{ 116, 498, 34 },
		{ 116, 499, 34 },
		{ 116, 500, 34 },
		{ 116, 501, 34 },
		{ 116, 502, 34 },
		{ 116, 503, 34 },
		{ 116, 504, 34 },
		{ 116, 505, 34 },
		{ 116, 506, 34 },
		{ 116, 507, 34 },
		{ 116, 508, 34 },
		{ 116, 509, 34 },
		{ 116, 510, 34 },
		{ 116, 511, 34 },
		{ 116, 512, 34 },
		{ 117, 482, 34 },
		{ 117, 483, 34 },
		{ 117, 484, 34 },
		{ 117, 485, 34 },
		{ 117, 486, 34 },
		{ 117, 487, 34 },
		{ 117, 488, 34 },
		{ 117, 489, 34 },
		{ 117, 490, 34 },
		{ 117, 491, 34 },
		{ 117, 492, 34 },
		{ 117, 493, 34 },
		{ 117, 494, 34 },
		{ 117, 495, 34 },
		{ 117, 496, 34 },
		{ 117, 497, 34 },
		{ 117, 498, 34 },
		{ 117, 499, 34 },
		{ 117, 500, 34 },
		{ 117, 501, 34 },
		{ 117, 502, 34 },
		{ 117, 503, 34 },
		{ 117, 504, 34 },
		{ 117, 505, 34 },
		{ 117, 506, 34 },
		{ 117, 507, 34 },
		{ 117, 508, 34 },
		{ 117, 509, 34 },
		{ 117, 510, 34 },
		{ 117, 511, 34 },
		{ 117, 512, 34 },
		{ 118, 480, 34 },
		{ 118, 481, 34 },
		{ 118, 482, 34 },
		{ 118, 483, 34 },
		{ 118, 484, 34 },
		{ 118, 485, 34 },
		{ 118, 486, 34 },
		{ 118, 487, 34 },
		{ 118, 488, 34 },
		{ 118, 489, 34 },
		{ 118, 490, 34 },
		{ 118, 491, 34 },
		{ 118, 492, 34 },
		{ 118, 493, 34 },
		{ 118, 494, 34 },
		{ 118, 495, 34 },
		{ 118, 496, 34 },
		{ 118, 497, 34 },
		{ 118, 498, 34 },
		{ 118, 499, 34 },
		{ 118, 500, 34 },
		{ 118, 501, 34 },
		{ 118, 502, 34 },
		{ 118, 503, 34 },
		{ 118, 504, 34 },
		{ 118, 505, 34 },
		{ 118, 506, 34 },
		{ 118, 507, 34 },
		{ 118, 508, 34 },
		{ 118, 509, 34 },
		{ 118, 510, 34 },
		{ 118, 511, 34 },
		{ 118, 512, 34 },
		{ 119, 478, 34 },
		{ 119, 479, 34 },
		{ 119, 480, 34 },
		{ 119, 481, 34 },
		{ 119, 482, 34 },
		{ 119, 483, 34 },
		{ 119, 484, 34 },
		{ 119, 485, 34 },
		{ 119, 486, 34 },
		{ 119, 487, 34 },
		{ 119, 488, 34 },
		{ 119, 489, 34 },
		{ 119, 490, 34 },
		{ 119, 491, 34 },
		{ 119, 492, 34 },
		{ 119, 493, 34 },
		{ 119, 494, 34 },
		{ 119, 495, 34 },
		{ 119, 496, 34 },
		{ 119, 497, 34 },
		{ 119, 498, 34 },
		{ 119, 499, 34 },
		{ 119, 500, 34 },
		{ 119, 501, 34 },
		{ 119, 502, 34 },
		{ 119, 503, 34 },
		{ 119, 504, 34 },
		{ 119, 505, 34 },
		{ 119, 506, 34 },
		{ 119, 507, 34 },
		{ 119, 508, 34 },
		{ 119, 509, 34 },
		{ 119, 510, 34 },
		{ 119, 511, 34 },
		{ 119, 512, 34 },
		{ 120, 477, 34 },
		{ 120, 478, 34 },
		{ 120, 479, 34 },
		{ 120, 480, 34 },
		{ 120, 481, 34 },
		{ 120, 482, 34 },
		{ 120, 483, 34 },
		{ 120, 484, 34 },
		{ 120, 485, 34 },
		{ 120, 486, 34 },
		{ 120, 487, 34 },
		{ 120, 488, 34 },
		{ 120, 489, 34 },
		{ 120, 490, 34 },
		{ 120, 491, 34 },
		{ 120, 492, 34 },
		{ 120, 493, 34 },
		{ 120, 494, 34 },
		{ 120, 495, 34 },
		{ 120, 496, 34 },
		{ 120, 497, 34 },
		{ 120, 498, 34 },
		{ 120, 499, 34 },
		{ 120, 500, 34 },
		{ 120, 501, 34 },
		{ 120, 502, 34 },
		{ 120, 503, 34 },
		{ 120, 504, 34 },
		{ 120, 505, 34 },
		{ 120, 506, 34 },
		{ 120, 507, 34 },
		{ 120, 508, 34 },
		{ 120, 509, 34 },
		{ 120, 510, 34 },
		{ 120, 511, 34 },
		{ 120, 512, 34 },
		{ 121, 475, 34 },
		{ 121, 476, 34 },
		{ 121, 477, 34 },
		{ 121, 478, 34 },
		{ 121, 479, 34 },
		{ 121, 480, 34 },
		{ 121, 481, 34 },
		{ 121, 482, 34 },
		{ 121, 483, 34 },
		{ 121, 484, 34 },
		{ 121, 485, 34 },
		{ 121, 486, 34 },
		{ 121, 487, 34 },
		{ 121, 488, 34 },
		{ 121, 489, 34 },
		{ 121, 490, 34 },
		{ 121, 491, 34 },
		{ 121, 492, 34 },
		{ 121, 493, 34 },
		{ 121, 494, 34 },
		{ 121, 495, 34 },
		{ 121, 496, 34 },
		{ 121, 497, 34 },
		{ 121, 498, 34 },
		{ 121, 499, 34 },
		{ 121, 500, 34 },
		{ 121, 501, 34 },
		{ 121, 502, 34 },
		{ 121, 503, 34 },
		{ 121, 504, 34 },
		{ 121, 505, 34 },
		{ 121, 506, 34 },
		{ 121, 507, 34 },
		{ 121, 508, 34 },
		{ 121, 509, 34 },
		{ 121, 510, 34 },
		{ 121, 511, 34 },
		{ 121, 512, 34 },
		{ 122, 474, 34 },
		{ 122, 475, 34 },
		{ 122, 476, 34 },
		{ 122, 477, 34 },
		{ 122, 478, 34 },
		{ 122, 479, 34 },
		{ 122, 480, 34 },
		{ 122, 481, 34 },
		{ 122, 482, 34 },
		{ 122, 483, 34 },
		{ 122, 484, 34 },
		{ 122, 485, 34 },
		{ 122, 486, 34 },
		{ 122, 487, 34 },
		{ 122, 488, 34 },
		{ 122, 489, 34 },
		{ 122, 490, 34 },
		{ 122, 491, 34 },
		{ 122, 492, 34 },
		{ 122, 493, 34 },
		{ 122, 494, 34 },
		{ 122, 495, 34 },
		{ 122, 496, 34 },
		{ 122, 497, 34 },
		{ 122, 498, 34 },
		{ 122, 499, 34 },
		{ 122, 500, 34 },
		{ 122, 501, 34 },
		{ 122, 502, 34 },
		{ 122, 503, 34 },
		{ 122, 504, 34 },
		{ 122, 505, 34 },
		{ 122, 506, 34 },
		{ 122, 507, 34 },
		{ 122, 508, 34 },
		{ 122, 509, 34 },
		{ 122, 510, 34 },
		{ 122, 511, 34 },
		{ 122, 512, 34 },
		{ 123, 472, 34 },
		{ 123, 473, 34 },
		{ 123, 474, 34 },
		{ 123, 475, 34 },
		{ 123, 476, 34 },
		{ 123, 477, 34 },
		{ 123, 478, 34 },
		{ 123, 479, 34 },
		{ 123, 480, 34 },
		{ 123, 481, 34 },
		{ 123, 482, 34 },
		{ 123, 483, 34 },
		{ 123, 484, 34 },
		{ 123, 485, 34 },
		{ 123, 486, 34 },
		{ 123, 487, 34 },
		{ 123, 488, 34 },
		{ 123, 489, 34 },
		{ 123, 490, 34 },
		{ 123, 491, 34 },
		{ 123, 492, 34 },
		{ 123, 493, 34 },
		{ 123, 494, 34 },
		{ 123, 495, 34 },
		{ 123, 496, 34 },
		{ 123, 497, 34 },
		{ 123, 498, 34 },
		{ 123, 499, 34 },
		{ 123, 500, 34 },
		{ 123, 501, 34 },
		{ 123, 502, 34 },
		{ 123, 503, 34 },
		{ 123, 504, 34 },
		{ 123, 505, 34 },
		{ 123, 506, 34 },
		{ 123, 507, 34 },
		{ 123, 508, 34 },
		{ 123, 509, 34 },
		{ 123, 510, 34 },
		{ 123, 511, 34 },
		{ 123, 512, 34 },
		{ 124, 470, 34 },
		{ 124, 471, 34 },
		{ 124, 472, 34 },
		{ 124, 473, 34 },
		{ 124, 474, 34 },
		{ 124, 475, 34 },
		{ 124, 476, 34 },
		{ 124, 477, 34 },
		{ 124, 478, 34 },
		{ 124, 479, 34 },
		{ 124, 480, 34 },
		{ 124, 481, 34 },
		{ 124, 482, 34 },
		{ 124, 483, 34 },
		{ 124, 484, 34 },
		{ 124, 485, 34 },
		{ 124, 486, 34 },
		{ 124, 487, 34 },
		{ 124, 488, 34 },
		{ 124, 489, 34 },
		{ 124, 490, 34 },
		{ 124, 491, 34 },
		{ 124, 492, 34 },
		{ 124, 493, 34 },
		{ 124, 494, 34 },
		{ 124, 495, 34 },
		{ 124, 496, 34 },
		{ 124, 497, 34 },
		{ 124, 498, 34 },
		{ 124, 499, 34 },
		{ 124, 500, 34 },
		{ 124, 501, 34 },
		{ 124, 502, 34 },
		{ 124, 503, 34 },
		{ 124, 504, 34 },
		{ 124, 505, 34 },
		{ 124, 506, 34 },
		{ 124, 507, 34 },
		{ 124, 508, 34 },
		{ 124, 509, 34 },
		{ 124, 510, 34 },
		{ 124, 511, 34 },
		{ 124, 512, 34 },
		{ 125, 469, 34 },
		{ 125, 470, 34 },
		{ 125, 471, 34 },
		{ 125, 472, 34 },
		{ 125, 473, 34 },
		{ 125, 474, 34 },
		{ 125, 475, 34 },
		{ 125, 476, 34 },
		{ 125, 477, 34 },
		{ 125, 478, 34 },
		{ 125, 479, 34 },
		{ 125, 480, 34 },
		{ 125, 481, 34 },
		{ 125, 482, 34 },
		{ 125, 483, 34 },
		{ 125, 484, 34 },
		{ 125, 485, 34 },
		{ 125, 486, 34 },
		{ 125, 487, 34 },
		{ 125, 488, 34 },
		{ 125, 489, 34 },
		{ 125, 490, 34 },
		{ 125, 491, 34 },
		{ 125, 492, 34 },
		{ 125, 493, 34 },
		{ 125, 494, 34 },
		{ 125, 495, 34 },
		{ 125, 496, 34 },
		{ 125, 497, 34 },
		{ 125, 498, 34 },
		{ 125, 499, 34 },
		{ 125, 500, 34 },
		{ 125, 501, 34 },
		{ 125, 502, 34 },
		{ 125, 503, 34 },
		{ 125, 504, 34 },
		{ 125, 505, 34 },
		{ 125, 506, 34 },
		{ 125, 507, 34 },
		{ 125, 508, 34 },
		{ 125, 509, 34 },
		{ 125, 510, 34 },
		{ 125, 511, 34 },
		{ 125, 512, 34 },
		{ 126, 467, 34 },
		{ 126, 468, 34 },
		{ 126, 469, 34 },
		{ 126, 470, 34 },
		{ 126, 471, 34 },
		{ 126, 472, 34 },
		{ 126, 473, 34 },
		{ 126, 474, 34 },
		{ 126, 475, 34 },
		{ 126, 476, 34 },
		{ 126, 477, 34 },
		{ 126, 478, 34 },
		{ 126, 479, 34 },
		{ 126, 480, 34 },
		{ 126, 481, 34 },
		{ 126, 482, 34 },
		{ 126, 483, 34 },
		{ 126, 484, 34 },
		{ 126, 485, 34 },
		{ 126, 486, 34 },
		{ 126, 487, 34 },
		{ 126, 488, 34 },
		{ 126, 489, 34 },
		{ 126, 490, 34 },
		{ 126, 491, 34 },
		{ 126, 492, 34 },
		{ 126, 493, 34 },
		{ 126, 494, 34 },
		{ 126, 495, 34 },
		{ 126, 496, 34 },
		{ 126, 497, 34 },
		{ 126, 498, 34 },
		{ 126, 499, 34 },
		{ 126, 500, 34 },
		{ 126, 501, 34 },
		{ 126, 502, 34 },
		{ 126, 503, 34 },
		{ 126, 504, 34 },
		{ 126, 505, 34 },
		{ 126, 506, 34 },
		{ 126, 507, 34 },
		{ 126, 508, 34 },
		{ 126, 509, 34 },
		{ 126, 510, 34 },
		{ 126, 511, 34 },
		{ 126, 512, 34 },
		{ 127, 466, 34 },
		{ 127, 467, 34 },
		{ 127, 468, 34 },
		{ 127, 469, 34 },
		{ 127, 470, 34 },
		{ 127, 471, 34 },
		{ 127, 472, 34 },
		{ 127, 473, 34 },
		{ 127, 474, 34 },
		{ 127, 475, 34 },
		{ 127, 476, 34 },
		{ 127, 477, 34 },
		{ 127, 478, 34 },
		{ 127, 479, 34 },
		{ 127, 480, 34 },
		{ 127, 481, 34 },
		{ 127, 482, 34 },
		{ 127, 483, 34 },
		{ 127, 484, 34 },
		{ 127, 485, 34 },
		{ 127, 486, 34 },
		{ 127, 487, 34 },
		{ 127, 488, 34 },
		{ 127, 489, 34 },
		{ 127, 490, 34 },
		{ 127, 491, 34 },
		{ 127, 492, 34 },
		{ 127, 493, 34 },
		{ 127, 494, 34 },
		{ 127, 495, 34 },
		{ 127, 496, 34 },
		{ 127, 497, 34 },
		{ 127, 498, 34 },
		{ 127, 499, 34 },
		{ 127, 500, 34 },
		{ 127, 501, 34 },
		{ 127, 502, 34 },
		{ 127, 503, 34 },
		{ 127, 504, 34 },
		{ 127, 505, 34 },
		{ 127, 506, 34 },
		{ 127, 507, 34 },
		{ 127, 508, 34 },
		{ 127, 509, 34 },
		{ 127, 510, 34 },
		{ 127, 511, 34 },
		{ 127, 512, 34 },
		{ 128, 464, 34 },
		{ 128, 465, 34 },
		{ 128, 466, 34 },
		{ 128, 467, 34 },
		{ 128, 468, 34 },
		{ 128, 469, 34 },
		{ 128, 470, 34 },
		{ 128, 471, 34 },
		{ 128, 472, 34 },
		{ 128, 473, 34 },
		{ 128, 474, 34 },
		{ 128, 475, 34 },
		{ 128, 476, 34 },
		{ 128, 477, 34 },
		{ 128, 478, 34 },
		{ 128, 479, 34 },
		{ 128, 480, 34 },
		{ 128, 481, 34 },
		{ 128, 482, 34 },
		{ 128, 483, 34 },
		{ 128, 484, 34 },
		{ 128, 485, 34 },
		{ 128, 486, 34 },
		{ 128, 487, 34 },
		{ 128, 488, 34 },
		{ 128, 489, 34 },
		{ 128, 490, 34 },
		{ 128, 491, 34 },
		{ 128, 492, 34 },
		{ 128, 493, 34 },
		{ 128, 494, 34 },
		{ 128, 495, 34 },
		{ 128, 496, 34 },
		{ 128, 497, 34 },
		{ 128, 498, 34 },
		{ 128, 499, 34 },
		{ 128, 500, 34 },
		{ 128, 501, 34 },
		{ 128, 502, 34 },
		{ 128, 503, 34 },
		{ 128, 504, 34 },
		{ 128, 505, 34 },
		{ 128, 506, 34 },
		{ 128, 507, 34 },
		{ 128, 508, 34 },
		{ 128, 509, 34 },
		{ 128, 510, 34 },
		{ 128, 511, 34 },
		{ 128, 512, 34 },
		{ 129, 462, 34 },
		{ 129, 463, 34 },
		{ 129, 464, 34 },
		{ 129, 465, 34 },
		{ 129, 466, 34 },
		{ 129, 467, 34 },
		{ 129, 468, 34 },
		{ 129, 469, 34 },
		{ 129, 470, 34 },
		{ 129, 471, 34 },
		{ 129, 472, 34 },
		{ 129, 473, 34 },
		{ 129, 474, 34 },
		{ 129, 475, 34 },
		{ 129, 476, 34 },
		{ 129, 477, 34 },
		{ 129, 478, 34 },
		{ 129, 479, 34 },
		{ 129, 480, 34 },
		{ 129, 481, 34 },
		{ 129, 482, 34 },
		{ 129, 483, 34 },
		{ 129, 484, 34 },
		{ 129, 485, 34 },
		{ 129, 486, 34 },
		{ 129, 487, 34 },
		{ 129, 488, 34 },
		{ 129, 489, 34 },
		{ 129, 490, 34 },
		{ 129, 491, 34 },
		{ 129, 492, 34 },
		{ 129, 493, 34 },
		{ 129, 494, 34 },
		{ 129, 495, 34 },
		{ 129, 496, 34 },
		{ 129, 497, 34 },
		{ 129, 498, 34 },
		{ 129, 499, 34 },
		{ 129, 500, 34 },
		{ 129, 501, 34 },
		{ 129, 502, 34 },
		{ 129, 503, 34 },
		{ 129, 504, 34 },
		{ 129, 505, 34 },
		{ 129, 506, 34 },
		{ 129, 507, 34 },
		{ 129, 508, 34 },
		{ 129, 509, 34 },
		{ 129, 510, 34 },
		{ 129, 511, 34 },
		{ 129, 512, 34 },
		{ 130, 461, 34 },
		{ 130, 462, 34 },
		{ 130, 463, 34 },
		{ 130, 464, 34 },
		{ 130, 465, 34 },
		{ 130, 466, 34 },
		{ 130, 467, 34 },
		{ 130, 468, 34 },
		{ 130, 469, 34 },
		{ 130, 470, 34 },
		{ 130, 471, 34 },
		{ 130, 472, 34 },
		{ 130, 473, 34 },
		{ 130, 474, 34 },
		{ 130, 475, 34 },
		{ 130, 476, 34 },
		{ 130, 477, 34 },
		{ 130, 478, 34 },
		{ 130, 479, 34 },
		{ 130, 480, 34 },
		{ 130, 481, 34 },
		{ 130, 482, 34 },
		{ 130, 483, 34 },
		{ 130, 484, 34 },
		{ 130, 485, 34 },
		{ 130, 486, 34 },
		{ 130, 487, 34 },
		{ 130, 488, 34 },
		{ 130, 489, 34 },
		{ 130, 490, 34 },
		{ 130, 491, 34 },
		{ 130, 492, 34 },
		{ 130, 493, 34 },
		{ 130, 494, 34 },
		{ 130, 495, 34 },
		{ 130, 496, 34 },
		{ 130, 497, 34 },
		{ 130, 498, 34 },
		{ 130, 499, 34 },
		{ 130, 500, 34 },
		{ 130, 501, 34 },
		{ 130, 502, 34 },
		{ 130, 503, 34 },
		{ 130, 504, 34 },
		{ 130, 505, 34 },
		{ 130, 506, 34 },
		{ 130, 507, 34 },
		{ 130, 508, 34 },
		{ 130, 509, 34 },
		{ 130, 510, 34 },
		{ 130, 511, 34 },
		{ 130, 512, 34 },
		{ 131, 459, 34 },
		{ 131, 460, 34 },
		{ 131, 461, 34 },
		{ 131, 462, 34 },
		{ 131, 463, 34 },
		{ 131, 464, 34 },
		{ 131, 465, 34 },
		{ 131, 466, 34 },
		{ 131, 467, 34 },
		{ 131, 468, 34 },
		{ 131, 469, 34 },
		{ 131, 470, 34 },
		{ 131, 471, 34 },
		{ 131, 472, 34 },
		{ 131, 473, 34 },
		{ 131, 474, 34 },
		{ 131, 475, 34 },
		{ 131, 476, 34 },
		{ 131, 477, 34 },
		{ 131, 478, 34 },
		{ 131, 479, 34 },
		{ 131, 480, 34 },
		{ 131, 481, 34 },
		{ 131, 482, 34 },
		{ 131, 483, 34 },
		{ 131, 484, 34 },
		{ 131, 485, 34 },
		{ 131, 486, 34 },
		{ 131, 487, 34 },
		{ 131, 488, 34 },
		{ 131, 489, 34 },
		{ 131, 490, 34 },
		{ 131, 491, 34 },
		{ 131, 492, 34 },
		{ 131, 493, 34 },
		{ 131, 494, 34 },
		{ 131, 495, 34 },
		{ 131, 496, 34 },
		{ 131, 497, 34 },
		{ 131, 498, 34 },
		{ 131, 499, 34 },
		{ 131, 500, 34 },
		{ 131, 501, 34 },
		{ 131, 502, 34 },
		{ 131, 503, 34 },
		{ 131, 504, 34 },
		{ 131, 505, 34 },
		{ 131, 506, 34 },
		{ 131, 507, 34 },
		{ 131, 508, 34 },
		{ 131, 509, 34 },
		{ 131, 510, 34 },
		{ 131, 511, 34 },
		{ 131, 512, 34 },
		{ 132, 458, 34 },
		{ 132, 459, 34 },
		{ 132, 460, 34 },
		{ 132, 461, 34 },
		{ 132, 462, 34 },
		{ 132, 463, 34 },
		{ 132, 464, 34 },
		{ 132, 465, 34 },
		{ 132, 466, 34 },
		{ 132, 467, 34 },
		{ 132, 468, 34 },
		{ 132, 469, 34 },
		{ 132, 470, 34 },
		{ 132, 471, 34 },
		{ 132, 472, 34 },
		{ 132, 473, 34 },
		{ 132, 474, 34 },
		{ 132, 475, 34 },
		{ 132, 476, 34 },
		{ 132, 477, 34 },
		{ 132, 478, 34 },
		{ 132, 479, 34 },
		{ 132, 480, 34 },
		{ 132, 481, 34 },
		{ 132, 482, 34 },
		{ 132, 483, 34 },
		{ 132, 484, 34 },
		{ 132, 485, 34 },
		{ 132, 486, 34 },
		{ 132, 487, 34 },
		{ 132, 488, 34 },
		{ 132, 489, 34 },
		{ 132, 490, 34 },
		{ 132, 491, 34 },
		{ 132, 492, 34 },
		{ 132, 493, 34 },
		{ 132, 494, 34 },
		{ 132, 495, 34 },
		{ 132, 496, 34 },
		{ 132, 497, 34 },
		{ 132, 498, 34 },
		{ 132, 499, 34 },
		{ 132, 500, 34 },
		{ 132, 501, 34 },
		{ 132, 502, 34 },
		{ 132, 503, 34 },
		{ 132, 504, 34 },
		{ 132, 505, 34 },
		{ 132, 506, 34 },
		{ 132, 507, 34 },
		{ 132, 508, 34 },
		{ 132, 509, 34 },
		{ 132, 510, 34 },
		{ 132, 511, 34 },
		{ 132, 512, 34 },
		{ 133, 456, 34 },
		{ 133, 457, 34 },
		{ 133, 458, 34 },
		{ 133, 459, 34 },
		{ 133, 460, 34 },
		{ 133, 461, 34 },
		{ 133, 462, 34 },
		{ 133, 463, 34 },
		{ 133, 464, 34 },
		{ 133, 465, 34 },
		{ 133, 466, 34 },
		{ 133, 467, 34 },
		{ 133, 468, 34 },
		{ 133, 469, 34 },
		{ 133, 470, 34 },
		{ 133, 471, 34 },
		{ 133, 472, 34 },
		{ 133, 473, 34 },
		{ 133, 474, 34 },
		{ 133, 475, 34 },
		{ 133, 476, 34 },
		{ 133, 477, 34 },
		{ 133, 478, 34 },
		{ 133, 479, 34 },
		{ 133, 480, 34 },
		{ 133, 481, 34 },
		{ 133, 482, 34 },
		{ 133, 483, 34 },
		{ 133, 484, 34 },
		{ 133, 485, 34 },
		{ 133, 486, 34 },
		{ 133, 487, 34 },
		{ 133, 488, 34 },
		{ 133, 489, 34 },
		{ 133, 490, 34 },
		{ 133, 491, 34 },
		{ 133, 492, 34 },
		{ 133, 493, 34 },
		{ 133, 494, 34 },
		{ 133, 495, 34 },
		{ 133, 496, 34 },
		{ 133, 497, 34 },
		{ 133, 498, 34 },
		{ 133, 499, 34 },
		{ 133, 500, 34 },
		{ 133, 501, 34 },
		{ 133, 502, 34 },
		{ 133, 503, 34 },
		{ 133, 504, 34 },
		{ 133, 505, 34 },
		{ 133, 506, 34 },
		{ 133, 507, 34 },
		{ 133, 508, 34 },
		{ 133, 509, 34 },
		{ 133, 510, 34 },
		{ 133, 511, 34 },
		{ 133, 512, 34 },
		{ 134, 454, 34 },
		{ 134, 455, 34 },
		{ 134, 456, 34 },
		{ 134, 457, 34 },
		{ 134, 458, 34 },
		{ 134, 459, 34 },
		{ 134, 460, 34 },
		{ 134, 461, 34 },
		{ 134, 462, 34 },
		{ 134, 463, 34 },
		{ 134, 464, 34 },
		{ 134, 465, 34 },
		{ 134, 466, 34 },
		{ 134, 467, 34 },
		{ 134, 468, 34 },
		{ 134, 469, 34 },
		{ 134, 470, 34 },
		{ 134, 471, 34 },
		{ 134, 472, 34 },
		{ 134, 473, 34 },
		{ 134, 474, 34 },
		{ 134, 475, 34 },
		{ 134, 476, 34 },
		{ 134, 477, 34 },
		{ 134, 478, 34 },
		{ 134, 479, 34 },
		{ 134, 480, 34 },
		{ 134, 481, 34 },
		{ 134, 482, 34 },
		{ 134, 483, 34 },
		{ 134, 484, 34 },
		{ 134, 485, 34 },
		{ 134, 486, 34 },
		{ 134, 487, 34 },
		{ 134, 488, 34 },
		{ 134, 489, 34 },
		{ 134, 490, 34 },
		{ 134, 491, 34 },
		{ 134, 492, 34 },
		{ 134, 493, 34 },
		{ 134, 494, 34 },
		{ 134, 495, 34 },
		{ 134, 496, 34 },
		{ 134, 497, 34 },
		{ 134, 498, 34 },
		{ 134, 499, 34 },
		{ 134, 500, 34 },
		{ 134, 501, 34 },
		{ 134, 502, 34 },
		{ 134, 503, 34 },
		{ 134, 504, 34 },
		{ 134, 505, 34 },
		{ 134, 506, 34 },
		{ 134, 507, 34 },
		{ 134, 508, 34 },
		{ 134, 509, 34 },
		{ 134, 510, 34 },
		{ 134, 511, 34 },
		{ 134, 512, 34 },
		{ 135, 453, 34 },
		{ 135, 454, 34 },
		{ 135, 455, 34 },
		{ 135, 456, 34 },
		{ 135, 457, 34 },
		{ 135, 458, 34 },
		{ 135, 459, 34 },
		{ 135, 460, 34 },
		{ 135, 461, 34 },
		{ 135, 462, 34 },
		{ 135, 463, 34 },
		{ 135, 464, 34 },
		{ 135, 465, 34 },
		{ 135, 466, 34 },
		{ 135, 467, 34 },
		{ 135, 468, 34 },
		{ 135, 469, 34 },
		{ 135, 470, 34 },
		{ 135, 471, 34 },
		{ 135, 472, 34 },
		{ 135, 473, 34 },
		{ 135, 474, 34 },
		{ 135, 475, 34 },
		{ 135, 476, 34 },
		{ 135, 477, 34 },
		{ 135, 478, 34 },
		{ 135, 479, 34 },
		{ 135, 480, 34 },
		{ 135, 481, 34 },
		{ 135, 482, 34 },
		{ 135, 483, 34 },
		{ 135, 484, 34 },
		{ 135, 485, 34 },
		{ 135, 486, 34 },
		{ 135, 487, 34 },
		{ 135, 488, 34 },
		{ 135, 489, 34 },
		{ 135, 490, 34 },
		{ 135, 491, 34 },
		{ 135, 492, 34 },
		{ 135, 493, 34 },
		{ 135, 494, 34 },
		{ 135, 495, 34 },
		{ 135, 496, 34 },
		{ 135, 497, 34 },
		{ 135, 498, 34 },
		{ 135, 499, 34 },
		{ 135, 500, 34 },
		{ 135, 501, 34 },
		{ 135, 502, 34 },
		{ 135, 503, 34 },
		{ 135, 504, 34 },
		{ 135, 505, 34 },
		{ 135, 506, 34 },
		{ 135, 507, 34 },
		{ 135, 508, 34 },
		{ 135, 509, 34 },
		{ 135, 510, 34 },
		{ 135, 511, 34 },
		{ 135, 512, 34 },
		{ 136, 451, 34 },
		{ 136, 452, 34 },
		{ 136, 453, 34 },
		{ 136, 454, 34 },
		{ 136, 455, 34 },
		{ 136, 456, 34 },
		{ 136, 457, 34 },
		{ 136, 458, 34 },
		{ 136, 459, 34 },
		{ 136, 460, 34 },
		{ 136, 461, 34 },
		{ 136, 462, 34 },
		{ 136, 463, 34 },
		{ 136, 464, 34 },
		{ 136, 465, 34 },
		{ 136, 466, 34 },
		{ 136, 467, 34 },
		{ 136, 468, 34 },
		{ 136, 469, 34 },
		{ 136, 470, 34 },
		{ 136, 471, 34 },
		{ 136, 472, 34 },
		{ 136, 473, 34 },
		{ 136, 474, 34 },
		{ 136, 475, 34 },
		{ 136, 476, 34 },
		{ 136, 477, 34 },
		{ 136, 478, 34 },
		{ 136, 479, 34 },
		{ 136, 480, 34 },
		{ 136, 481, 34 },
		{ 136, 482, 34 },
		{ 136, 483, 34 },
		{ 136, 484, 34 },
		{ 136, 485, 34 },
		{ 136, 486, 34 },
		{ 136, 487, 34 },
		{ 136, 488, 34 },
		{ 136, 489, 34 },
		{ 136, 490, 34 },
		{ 136, 491, 34 },
		{ 136, 492, 34 },
		{ 136, 493, 34 },
		{ 136, 494, 34 },
		{ 136, 495, 34 },
		{ 136, 496, 34 },
		{ 136, 497, 34 },
		{ 136, 498, 34 },
		{ 136, 499, 34 },
		{ 136, 500, 34 },
		{ 136, 501, 34 },
		{ 136, 502, 34 },
		{ 136, 503, 34 },
		{ 136, 504, 34 },
		{ 136, 505, 34 },
		{ 136, 506, 34 },
		{ 136, 507, 34 },
		{ 136, 508, 34 },
		{ 136, 509, 34 },
		{ 136, 510, 34 },
		{ 136, 511, 34 },
		{ 136, 512, 34 },
		{ 137, 450, 34 },
		{ 137, 451, 34 },
		{ 137, 452, 34 },
		{ 137, 453, 34 },
		{ 137, 454, 34 },
		{ 137, 455, 34 },
		{ 137, 456, 34 },
		{ 137, 457, 34 },
		{ 137, 458, 34 },
		{ 137, 459, 34 },
		{ 137, 460, 34 },
		{ 137, 461, 34 },
		{ 137, 462, 34 },
		{ 137, 463, 34 },
		{ 137, 464, 34 },
		{ 137, 465, 34 },
		{ 137, 466, 34 },
		{ 137, 467, 34 },
		{ 137, 468, 34 },
		{ 137, 469, 34 },
		{ 137, 470, 34 },
		{ 137, 471, 34 },
		{ 137, 472, 34 },
		{ 137, 473, 34 },
		{ 137, 474, 34 },
		{ 137, 475, 34 },
		{ 137, 476, 34 },
		{ 137, 477, 34 },
		{ 137, 478, 34 },
		{ 137, 479, 34 },
		{ 137, 480, 34 },
		{ 137, 481, 34 },
		{ 137, 482, 34 },
		{ 137, 483, 34 },
		{ 137, 484, 34 },
		{ 137, 485, 34 },
		{ 137, 486, 34 },
		{ 137, 487, 34 },
		{ 137, 488, 34 },
		{ 137, 489, 34 },
		{ 137, 490, 34 },
		{ 137, 491, 34 },
		{ 137, 492, 34 },
		{ 137, 493, 34 },
		{ 137, 494, 34 },
		{ 137, 495, 34 },
		{ 137, 496, 34 },
		{ 137, 497, 34 },
		{ 137, 498, 34 },
		{ 137, 499, 34 },
		{ 137, 500, 34 },
		{ 137, 501, 34 },
		{ 137, 502, 34 },
		{ 137, 503, 34 },
		{ 137, 504, 34 },
		{ 137, 505, 34 },
		{ 137, 506, 34 },
		{ 137, 507, 34 },
		{ 137, 508, 34 },
		{ 137, 509, 34 },
		{ 137, 510, 34 },
		{ 137, 511, 34 },
		{ 137, 512, 34 },
		{ 138, 448, 34 },
		{ 138, 449, 34 },
		{ 138, 450, 34 },
		{ 138, 451, 34 },
		{ 138, 452, 34 },
		{ 138, 453, 34 },
		{ 138, 454, 34 },
		{ 138, 455, 34 },
		{ 138, 456, 34 },
		{ 138, 457, 34 },
		{ 138, 458, 34 },
		{ 138, 459, 34 },
		{ 138, 460, 34 },
		{ 138, 461, 34 },
		{ 138, 462, 34 },
		{ 138, 463, 34 },
		{ 138, 464, 34 },
		{ 138, 465, 34 },
		{ 138, 466, 34 },
		{ 138, 467, 34 },
		{ 138, 468, 34 },
		{ 138, 469, 34 },
		{ 138, 470, 34 },
		{ 138, 471, 34 },
		{ 138, 472, 34 },
		{ 138, 473, 34 },
		{ 138, 474, 34 },
		{ 138, 475, 34 },
		{ 138, 476, 34 },
		{ 138, 477, 34 },
		{ 138, 478, 34 },
		{ 138, 479, 34 },
		{ 138, 480, 34 },
		{ 138, 481, 34 },
		{ 138, 482, 34 },
		{ 138, 483, 34 },
		{ 138, 484, 34 },
		{ 138, 485, 34 },
		{ 138, 486, 34 },
		{ 138, 487, 34 },
		{ 138, 488, 34 },
		{ 138, 489, 34 },
		{ 138, 490, 34 },
		{ 138, 491, 34 },
		{ 138, 492, 34 },
		{ 138, 493, 34 },
		{ 138, 494, 34 },
		{ 138, 495, 34 },
		{ 138, 496, 34 },
		{ 138, 497, 34 },
		{ 138, 498, 34 },
		{ 138, 499, 34 },
		{ 138, 500, 34 },
		{ 138, 501, 34 },
		{ 138, 502, 34 },
		{ 138, 503, 34 },
		{ 138, 504, 34 },
		{ 138, 505, 34 },
		{ 138, 506, 34 },
		{ 138, 507, 34 },
		{ 138, 508, 34 },
		{ 138, 509, 34 },
		{ 138, 510, 34 },
		{ 138, 511, 34 },
		{ 138, 512, 34 },
		{ 139, 446, 34 },
		{ 139, 447, 34 },
		{ 139, 448, 34 },
		{ 139, 449, 34 },
		{ 139, 450, 34 },
		{ 139, 451, 34 },
		{ 139, 452, 34 },
		{ 139, 453, 34 },
		{ 139, 454, 34 },
		{ 139, 455, 34 },
		{ 139, 456, 34 },
		{ 139, 457, 34 },
		{ 139, 458, 34 },
		{ 139, 459, 34 },
		{ 139, 460, 34 },
		{ 139, 461, 34 },
		{ 139, 462, 34 },
		{ 139, 463, 34 },
		{ 139, 464, 34 },
		{ 139, 465, 34 },
		{ 139, 466, 34 },
		{ 139, 467, 34 },
		{ 139, 468, 34 },
		{ 139, 469, 34 },
		{ 139, 470, 34 },
		{ 139, 471, 34 },
		{ 139, 472, 34 },
		{ 139, 473, 34 },
		{ 139, 474, 34 },
		{ 139, 475, 34 },
		{ 139, 476, 34 },
		{ 139, 477, 34 },
		{ 139, 478, 34 },
		{ 139, 479, 34 },
		{ 139, 480, 34 },
		{ 139, 481, 34 },
		{ 139, 482, 34 },
		{ 139, 483, 34 },
		{ 139, 484, 34 },
		{ 139, 485, 34 },
		{ 139, 486, 34 },
		{ 139, 487, 34 },
		{ 139, 488, 34 },
		{ 139, 489, 34 },
		{ 139, 490, 34 },
		{ 139, 491, 34 },
		{ 139, 492, 34 },
		{ 139, 493, 34 },
		{ 139, 494, 34 },
		{ 139, 495, 34 },
		{ 139, 496, 34 },
		{ 139, 497, 34 },
		{ 139, 498, 34 },
		{ 139, 499, 34 },
		{ 139, 500, 34 },
		{ 139, 501, 34 },
		{ 139, 502, 34 },
		{ 139, 503, 34 },
		{ 139, 504, 34 },
		{ 139, 505, 34 },
		{ 139, 506, 34 },
		{ 139, 507, 34 },
		{ 139, 508, 34 },
		{ 139, 509, 34 },
		{ 139, 510, 34 },
		{ 139, 511, 34 },
		{ 139, 512, 34 },
		{ 140, 445, 34 },
		{ 140, 446, 34 },
		{ 140, 447, 34 },
		{ 140, 448, 34 },
		{ 140, 449, 34 },
		{ 140, 450, 34 },
		{ 140, 451, 34 },
		{ 140, 452, 34 },
		{ 140, 453, 34 },
		{ 140, 454, 34 },
		{ 140, 455, 34 },
		{ 140, 456, 34 },
		{ 140, 457, 34 },
		{ 140, 458, 34 },
		{ 140, 459, 34 },
		{ 140, 460, 34 },
		{ 140, 461, 34 },
		{ 140, 462, 34 },
		{ 140, 463, 34 },
		{ 140, 464, 34 },
		{ 140, 465, 34 },
		{ 140, 466, 34 },
		{ 140, 467, 34 },
		{ 140, 468, 34 },
		{ 140, 469, 34 },
		{ 140, 470, 34 },
		{ 140, 471, 34 },
		{ 140, 472, 34 },
		{ 140, 473, 34 },
		{ 140, 474, 34 },
		{ 140, 475, 34 },
		{ 140, 476, 34 },
		{ 140, 477, 34 },
		{ 140, 478, 34 },
		{ 140, 479, 34 },
		{ 140, 480, 34 },
		{ 140, 481, 34 },
		{ 140, 482, 34 },
		{ 140, 483, 34 },
		{ 140, 484, 34 },
		{ 140, 485, 34 },
		{ 140, 486, 34 },
		{ 140, 487, 34 },
		{ 140, 488, 34 },
		{ 140, 489, 34 },
		{ 140, 490, 34 },
		{ 140, 491, 34 },
		{ 140, 492, 34 },
		{ 140, 493, 34 },
		{ 140, 494, 34 },
		{ 140, 495, 34 },
		{ 140, 496, 34 },
		{ 140, 497, 34 },
		{ 140, 498, 34 },
		{ 140, 499, 34 },
		{ 140, 500, 34 },
		{ 140, 501, 34 },
		{ 140, 502, 34 },
		{ 140, 503, 34 },
		{ 140, 504, 34 },
		{ 140, 505, 34 },
		{ 140, 506, 34 },
		{ 140, 507, 34 },
		{ 140, 508, 34 },
		{ 140, 509, 34 },
		{ 140, 510, 34 },
		{ 140, 511, 34 },
		{ 140, 512, 34 },
		{ 141, 443, 34 },
		{ 141, 444, 34 },
		{ 141, 445, 34 },
		{ 141, 446, 34 },
		{ 141, 447, 34 },
		{ 141, 448, 34 },
		{ 141, 449, 34 },
		{ 141, 450, 34 },
		{ 141, 451, 34 },
		{ 141, 452, 34 },
		{ 141, 453, 34 },
		{ 141, 454, 34 },
		{ 141, 455, 34 },
		{ 141, 456, 34 },
		{ 141, 457, 34 },
		{ 141, 458, 34 },
		{ 141, 459, 34 },
		{ 141, 460, 34 },
		{ 141, 461, 34 },
		{ 141, 462, 34 },
		{ 141, 463, 34 },
		{ 141, 464, 34 },
		{ 141, 465, 34 },
		{ 141, 466, 34 },
		{ 141, 467, 34 },
		{ 141, 468, 34 },
		{ 141, 469, 34 },
		{ 141, 470, 34 },
		{ 141, 471, 34 },
		{ 141, 472, 34 },
		{ 141, 473, 34 },
		{ 141, 474, 34 },
		{ 141, 475, 34 },
		{ 141, 476, 34 },
		{ 141, 477, 34 },
		{ 141, 478, 34 },
		{ 141, 479, 34 },
		{ 141, 480, 34 },
		{ 141, 481, 34 },
		{ 141, 482, 34 },
		{ 141, 483, 34 },
		{ 141, 484, 34 },
		{ 141, 485, 34 },
		{ 141, 486, 34 },
		{ 141, 487, 34 },
		{ 141, 488, 34 },
		{ 141, 489, 34 },
		{ 141, 490, 34 },
		{ 141, 491, 34 },
		{ 141, 492, 34 },
		{ 141, 493, 34 },
		{ 141, 494, 34 },
		{ 141, 495, 34 },
		{ 141, 496, 34 },
		{ 141, 497, 34 },
		{ 141, 498, 34 },
		{ 141, 499, 34 },
		{ 141, 500, 34 },
		{ 141, 501, 34 },
		{ 141, 502, 34 },
		{ 141, 503, 34 },
		{ 141, 504, 34 },
		{ 141, 505, 34 },
		{ 141, 506, 34 },
		{ 141, 507, 34 },
		{ 141, 508, 34 },
		{ 141, 509, 34 },
		{ 141, 510, 34 },
		{ 141, 511, 34 },
		{ 141, 512, 34 },
		{ 142, 442, 34 },
		{ 142, 443, 34 },
		{ 142, 444, 34 },
		{ 142, 445, 34 },
		{ 142, 446, 34 },
		{ 142, 447, 34 },
		{ 142, 448, 34 },
		{ 142, 449, 34 },
		{ 142, 450, 34 },
		{ 142, 451, 34 },
		{ 142, 452, 34 },
		{ 142, 453, 34 },
		{ 142, 454, 34 },
		{ 142, 455, 34 },
		{ 142, 456, 34 },
		{ 142, 457, 34 },
		{ 142, 458, 34 },
		{ 142, 459, 34 },
		{ 142, 460, 34 },
		{ 142, 461, 34 },
		{ 142, 462, 34 },
		{ 142, 463, 34 },
		{ 142, 464, 34 },
		{ 142, 465, 34 },
		{ 142, 466, 34 },
		{ 142, 467, 34 },
		{ 142, 468, 34 },
		{ 142, 469, 34 },
		{ 142, 470, 34 },
		{ 142, 471, 34 },
		{ 142, 472, 34 },
		{ 142, 473, 34 },
		{ 142, 474, 34 },
		{ 142, 475, 34 },
		{ 142, 476, 34 },
		{ 142, 477, 34 },
		{ 142, 478, 34 },
		{ 142, 479, 34 },
		{ 142, 480, 34 },
		{ 142, 481, 34 },
		{ 142, 482, 34 },
		{ 142, 483, 34 },
		{ 142, 484, 34 },
		{ 142, 485, 34 },
		{ 142, 486, 34 },
		{ 142, 487, 34 },
		{ 142, 488, 34 },
		{ 142, 489, 34 },
		{ 142, 490, 34 },
		{ 142, 491, 34 },
		{ 142, 492, 34 },
		{ 142, 493, 34 },
		{ 142, 494, 34 },
		{ 142, 495, 34 },
		{ 142, 496, 34 },
		{ 142, 497, 34 },
		{ 142, 498, 34 },
		{ 142, 499, 34 },
		{ 142, 500, 34 },
		{ 142, 501, 34 },
		{ 142, 502, 34 },
		{ 142, 503, 34 },
		{ 142, 504, 34 },
		{ 142, 505, 34 },
		{ 142, 506, 34 },
		{ 142, 507, 34 },
		{ 142, 508, 34 },
		{ 142, 509, 34 },
		{ 142, 510, 34 },
		{ 142, 511, 34 },
		{ 142, 512, 34 },
		{ 143, 440, 34 },
		{ 143, 441, 34 },
		{ 143, 442, 34 },
		{ 143, 443, 34 },
		{ 143, 444, 34 },
		{ 143, 445, 34 },
		{ 143, 446, 34 },
		{ 143, 447, 34 },
		{ 143, 448, 34 },
		{ 143, 449, 34 },
		{ 143, 450, 34 },
		{ 143, 451, 34 },
		{ 143, 452, 34 },
		{ 143, 453, 34 },
		{ 143, 454, 34 },
		{ 143, 455, 34 },
		{ 143, 456, 34 },
		{ 143, 457, 34 },
		{ 143, 458, 34 },
		{ 143, 459, 34 },
		{ 143, 460, 34 },
		{ 143, 461, 34 },
		{ 143, 462, 34 },
		{ 143, 463, 34 },
		{ 143, 464, 34 },
		{ 143, 465, 34 },
		{ 143, 466, 34 },
		{ 143, 467, 34 },
		{ 143, 468, 34 },
		{ 143, 469, 34 },
		{ 143, 470, 34 },
		{ 143, 471, 34 },
		{ 143, 472, 34 },
		{ 143, 473, 34 },
		{ 143, 474, 34 },
		{ 143, 475, 34 },
		{ 143, 476, 34 },
		{ 143, 477, 34 },
		{ 143, 478, 34 },
		{ 143, 479, 34 },
		{ 143, 480, 34 },
		{ 143, 481, 34 },
		{ 143, 482, 34 },
		{ 143, 483, 34 },
		{ 143, 484, 34 },
		{ 143, 485, 34 },
		{ 143, 486, 34 },
		{ 143, 487, 34 },
		{ 143, 488, 34 },
		{ 143, 489, 34 },
		{ 143, 490, 34 },
		{ 143, 491, 34 },
		{ 143, 492, 34 },
		{ 143, 493, 34 },
		{ 143, 494, 34 },
		{ 143, 495, 34 },
		{ 143, 496, 34 },
		{ 143, 497, 34 },
		{ 143, 498, 34 },
		{ 143, 499, 34 },
		{ 143, 500, 34 },
		{ 143, 501, 34 },
		{ 143, 502, 34 },
		{ 143, 503, 34 },
		{ 143, 504, 34 },
		{ 143, 505, 34 },
		{ 143, 506, 34 },
		{ 143, 507, 34 },
		{ 143, 508, 34 },
		{ 143, 509, 34 },
		{ 143, 510, 34 },
		{ 143, 511, 34 },
		{ 143, 512, 34 },
		{ 144, 438, 34 },
		{ 144, 439, 34 },
		{ 144, 440, 34 },
		{ 144, 441, 34 },
		{ 144, 442, 34 },
		{ 144, 443, 34 },
		{ 144, 444, 34 },
		{ 144, 445, 34 },
		{ 144, 446, 34 },
		{ 144, 447, 34 },
		{ 144, 448, 34 },
		{ 144, 449, 34 },
		{ 144, 450, 34 },
		{ 144, 451, 34 },
		{ 144, 452, 34 },
		{ 144, 453, 34 },
		{ 144, 454, 34 },
		{ 144, 455, 34 },
		{ 144, 456, 34 },
		{ 144, 457, 34 },
		{ 144, 458, 34 },
		{ 144, 459, 34 },
		{ 144, 460, 34 },
		{ 144, 461, 34 },
		{ 144, 462, 34 },
		{ 144, 463, 34 },
		{ 144, 464, 34 },
		{ 144, 465, 34 },
		{ 144, 466, 34 },
		{ 144, 467, 34 },
		{ 144, 468, 34 },
		{ 144, 469, 34 },
		{ 144, 470, 34 },
		{ 144, 471, 34 },
		{ 144, 472, 34 },
		{ 144, 473, 34 },
		{ 144, 474, 34 },
		{ 144, 475, 34 },
		{ 144, 476, 34 },
		{ 144, 477, 34 },
		{ 144, 478, 34 },
		{ 144, 479, 34 },
		{ 144, 480, 34 },
		{ 144, 481, 34 },
		{ 144, 482, 34 },
		{ 144, 483, 34 },
		{ 144, 484, 34 },
		{ 144, 485, 34 },
		{ 144, 486, 34 },
		{ 144, 487, 34 },
		{ 144, 488, 34 },
		{ 144, 489, 34 },
		{ 144, 490, 34 },
		{ 144, 491, 34 },
		{ 144, 492, 34 },
		{ 144, 493, 34 },
		{ 144, 494, 34 },
		{ 144, 495, 34 },
		{ 144, 496, 34 },
		{ 144, 497, 34 },
		{ 144, 498, 34 },
		{ 144, 499, 34 },
		{ 144, 500, 34 },
		{ 144, 501, 34 },
		{ 144, 502, 34 },
		{ 144, 503, 34 },
		{ 144, 504, 34 },
		{ 144, 505, 34 },
		{ 144, 506, 34 },
		{ 144, 507, 34 },
		{ 144, 508, 34 },
		{ 144, 509, 34 },
		{ 144, 510, 34 },
		{ 144, 511, 34 },
		{ 144, 512, 34 },
		{ 145, 437, 34 },
		{ 145, 438, 34 },
		{ 145, 439, 34 },
		{ 145, 440, 34 },
		{ 145, 441, 34 },
		{ 145, 442, 34 },
		{ 145, 443, 34 },
		{ 145, 444, 34 },
		{ 145, 445, 34 },
		{ 145, 446, 34 },
		{ 145, 447, 34 },
		{ 145, 448, 34 },
		{ 145, 449, 34 },
		{ 145, 450, 34 },
		{ 145, 451, 34 },
		{ 145, 452, 34 },
		{ 145, 453, 34 },
		{ 145, 454, 34 },
		{ 145, 455, 34 },
		{ 145, 456, 34 },
		{ 145, 457, 34 },
		{ 145, 458, 34 },
		{ 145, 459, 34 },
		{ 145, 460, 34 },
		{ 145, 461, 34 },
		{ 145, 462, 34 },
		{ 145, 463, 34 },
		{ 145, 464, 34 },
		{ 145, 465, 34 },
		{ 145, 466, 34 },
		{ 145, 467, 34 },
		{ 145, 468, 34 },
		{ 145, 469, 34 },
		{ 145, 470, 34 },
		{ 145, 471, 34 },
		{ 145, 472, 34 },
		{ 145, 473, 34 },
		{ 145, 474, 34 },
		{ 145, 475, 34 },
		{ 145, 476, 34 },
		{ 145, 477, 34 },
		{ 145, 478, 34 },
		{ 145, 479, 34 },
		{ 145, 480, 34 },
		{ 145, 481, 34 },
		{ 145, 482, 34 },
		{ 145, 483, 34 },
		{ 145, 484, 34 },
		{ 145, 485, 34 },
		{ 145, 486, 34 },
		{ 145, 487, 34 },
		{ 145, 488, 34 },
		{ 145, 489, 34 },
		{ 145, 490, 34 },
		{ 145, 491, 34 },
		{ 145, 492, 34 },
		{ 145, 493, 34 },
		{ 145, 494, 34 },
		{ 145, 495, 34 },
		{ 145, 496, 34 },
		{ 145, 497, 34 },
		{ 145, 498, 34 },
		{ 145, 499, 34 },
		{ 145, 500, 34 },
		{ 145, 501, 34 },
		{ 145, 502, 34 },
		{ 145, 503, 34 },
		{ 145, 504, 34 },
		{ 145, 505, 34 },
		{ 145, 506, 34 },
		{ 145, 507, 34 },
		{ 145, 508, 34 },
		{ 145, 509, 34 },
		{ 145, 510, 34 },
		{ 145, 511, 34 },
		{ 145, 512, 34 },
		{ 146, 435, 34 },
		{ 146, 436, 34 },
		{ 146, 437, 34 },
		{ 146, 438, 34 },
		{ 146, 439, 34 },
		{ 146, 440, 34 },
		{ 146, 441, 34 },
		{ 146, 442, 34 },
		{ 146, 443, 34 },
		{ 146, 444, 34 },
		{ 146, 445, 34 },
		{ 146, 446, 34 },
		{ 146, 447, 34 },
		{ 146, 448, 34 },
		{ 146, 449, 34 },
		{ 146, 450, 34 },
		{ 146, 451, 34 },
		{ 146, 452, 34 },
		{ 146, 453, 34 },
		{ 146, 454, 34 },
		{ 146, 455, 34 },
		{ 146, 456, 34 },
		{ 146, 457, 34 },
		{ 146, 458, 34 },
		{ 146, 459, 34 },
		{ 146, 460, 34 },
		{ 146, 461, 34 },
		{ 146, 462, 34 },
		{ 146, 463, 34 },
		{ 146, 464, 34 },
		{ 146, 465, 34 },
		{ 146, 466, 34 },
		{ 146, 467, 34 },
		{ 146, 468, 34 },
		{ 146, 469, 34 },
		{ 146, 470, 34 },
		{ 146, 471, 34 },
		{ 146, 472, 34 },
		{ 146, 473, 34 },
		{ 146, 474, 34 },
		{ 146, 475, 34 },
		{ 146, 476, 34 },
		{ 146, 477, 34 },
		{ 146, 478, 34 },
		{ 146, 479, 34 },
		{ 146, 480, 34 },
		{ 146, 481, 34 },
		{ 146, 482, 34 },
		{ 146, 483, 34 },
		{ 146, 484, 34 },
		{ 146, 485, 34 },
		{ 146, 486, 34 },
		{ 146, 487, 34 },
		{ 146, 488, 34 },
		{ 146, 489, 34 },
		{ 146, 490, 34 },
		{ 146, 491, 34 },
		{ 146, 492, 34 },
		{ 146, 493, 34 },
		{ 146, 494, 34 },
		{ 146, 495, 34 },
		{ 146, 496, 34 },
		{ 146, 497, 34 },
		{ 146, 498, 34 },
		{ 146, 499, 34 },
		{ 146, 500, 34 },
		{ 146, 501, 34 },
		{ 146, 502, 34 },
		{ 146, 503, 34 },
		{ 146, 504, 34 },
		{ 146, 505, 34 },
		{ 146, 506, 34 },
		{ 146, 507, 34 },
		{ 146, 508, 34 },
		{ 146, 509, 34 },
		{ 146, 510, 34 },
		{ 146, 511, 34 },
		{ 146, 512, 34 },
		{ 147, 434, 34 },
		{ 147, 435, 34 },
		{ 147, 436, 34 },
		{ 147, 437, 34 },
		{ 147, 438, 34 },
		{ 147, 439, 34 },
		{ 147, 440, 34 },
		{ 147, 441, 34 },
		{ 147, 442, 34 },
		{ 147, 443, 34 },
		{ 147, 444, 34 },
		{ 147, 445, 34 },
		{ 147, 446, 34 },
		{ 147, 447, 34 },
		{ 147, 448, 34 },
		{ 147, 449, 34 },
		{ 147, 450, 34 },
		{ 147, 451, 34 },
		{ 147, 452, 34 },
		{ 147, 453, 34 },
		{ 147, 454, 34 },
		{ 147, 455, 34 },
		{ 147, 456, 34 },
		{ 147, 457, 34 },
		{ 147, 458, 34 },
		{ 147, 459, 34 },
		{ 147, 460, 34 },
		{ 147, 461, 34 },
		{ 147, 462, 34 },
		{ 147, 463, 34 },
		{ 147, 464, 34 },
		{ 147, 465, 34 },
		{ 147, 466, 34 },
		{ 147, 467, 34 },
		{ 147, 468, 34 },
		{ 147, 469, 34 },
		{ 147, 470, 34 },
		{ 147, 471, 34 },
		{ 147, 472, 34 },
		{ 147, 473, 34 },
		{ 147, 474, 34 },
		{ 147, 475, 34 },
		{ 147, 476, 34 },
		{ 147, 477, 34 },
		{ 147, 478, 34 },
		{ 147, 479, 34 },
		{ 147, 480, 34 },
		{ 147, 481, 34 },
		{ 147, 482, 34 },
		{ 147, 483, 34 },
		{ 147, 484, 34 },
		{ 147, 485, 34 },
		{ 147, 486, 34 },
		{ 147, 487, 34 },
		{ 147, 488, 34 },
		{ 147, 489, 34 },
		{ 147, 490, 34 },
		{ 147, 491, 34 },
		{ 147, 492, 34 },
		{ 147, 493, 34 },
		{ 147, 494, 34 },
		{ 147, 495, 34 },
		{ 147, 496, 34 },
		{ 147, 497, 34 },
		{ 147, 498, 34 },
		{ 147, 499, 34 },
		{ 147, 500, 34 },
		{ 147, 501, 34 },
		{ 147, 502, 34 },
		{ 147, 503, 34 },
		{ 147, 504, 34 },
		{ 147, 505, 34 },
		{ 147, 506, 34 },
		{ 147, 507, 34 },
		{ 147, 508, 34 },
		{ 147, 509, 34 },
		{ 147, 510, 34 },
		{ 147, 511, 34 },
		{ 147, 512, 34 },
		{ 148, 432, 34 },
		{ 148, 433, 34 },
		{ 148, 434, 34 },
		{ 148, 435, 34 },
		{ 148, 436, 34 },
		{ 148, 437, 34 },
		{ 148, 438, 34 },
		{ 148, 439, 34 },
		{ 148, 440, 34 },
		{ 148, 441, 34 },
		{ 148, 442, 34 },
		{ 148, 443, 34 },
		{ 148, 444, 34 },
		{ 148, 445, 34 },
		{ 148, 446, 34 },
		{ 148, 447, 34 },
		{ 148, 448, 34 },
		{ 148, 449, 34 },
		{ 148, 450, 34 },
		{ 148, 451, 34 },
		{ 148, 452, 34 },
		{ 148, 453, 34 },
		{ 148, 454, 34 },
		{ 148, 455, 34 },
		{ 148, 456, 34 },
		{ 148, 457, 34 },
		{ 148, 458, 34 },
		{ 148, 459, 34 },
		{ 148, 460, 34 },
		{ 148, 461, 34 },
		{ 148, 462, 34 },
		{ 148, 463, 34 },
		{ 148, 464, 34 },
		{ 148, 465, 34 },
		{ 148, 466, 34 },
		{ 148, 467, 34 },
		{ 148, 468, 34 },
		{ 148, 469, 34 },
		{ 148, 470, 34 },
		{ 148, 471, 34 },
		{ 148, 472, 34 },
		{ 148, 473, 34 },
		{ 148, 474, 34 },
		{ 148, 475, 34 },
		{ 148, 476, 34 },
		{ 148, 477, 34 },
		{ 148, 478, 34 },
		{ 148, 479, 34 },
		{ 148, 480, 34 },
		{ 148, 481, 34 },
		{ 148, 482, 34 },
		{ 148, 483, 34 },
		{ 148, 484, 34 },
		{ 148, 485, 34 },
		{ 148, 486, 34 },
		{ 148, 487, 34 },
		{ 148, 488, 34 },
		{ 148, 489, 34 },
		{ 148, 490, 34 },
		{ 148, 491, 34 },
		{ 148, 492, 34 },
		{ 148, 493, 34 },
		{ 148, 494, 34 },
		{ 148, 495, 34 },
		{ 148, 496, 34 },
		{ 148, 497, 34 },
		{ 148, 498, 34 },
		{ 148, 499, 34 },
		{ 148, 500, 34 },
		{ 148, 501, 34 },
		{ 148, 502, 34 },
		{ 148, 503, 34 },
		{ 148, 504, 34 },
		{ 148, 505, 34 },
		{ 148, 506, 34 },
		{ 148, 507, 34 },
		{ 148, 508, 34 },
		{ 148, 509, 34 },
		{ 148, 510, 34 },
		{ 148, 511, 34 },
		{ 148, 512, 34 },
		{ 149, 430, 34 },
		{ 149, 431, 34 },
		{ 149, 432, 34 },
		{ 149, 433, 34 },
		{ 149, 434, 34 },
		{ 149, 435, 34 },
		{ 149, 436, 34 },
		{ 149, 437, 34 },
		{ 149, 438, 34 },
		{ 149, 439, 34 },
		{ 149, 440, 34 },
		{ 149, 441, 34 },
		{ 149, 442, 34 },
		{ 149, 443, 34 },
		{ 149, 444, 34 },
		{ 149, 445, 34 },
		{ 149, 446, 34 },
		{ 149, 447, 34 },
		{ 149, 448, 34 },
		{ 149, 449, 34 },
		{ 149, 450, 34 },
		{ 149, 451, 34 },
		{ 149, 452, 34 },
		{ 149, 453, 34 },
		{ 149, 454, 34 },
		{ 149, 455, 34 },
		{ 149, 456, 34 },
		{ 149, 457, 34 },
		{ 149, 458, 34 },
		{ 149, 459, 34 },
		{ 149, 460, 34 },
		{ 149, 461, 34 },
		{ 149, 462, 34 },
		{ 149, 463, 34 },
		{ 149, 464, 34 },
		{ 149, 465, 34 },
		{ 149, 466, 34 },
		{ 149, 467, 34 },
		{ 149, 468, 34 },
		{ 149, 469, 34 },
		{ 149, 470, 34 },
		{ 149, 471, 34 },
		{ 149, 472, 34 },
		{ 149, 473, 34 },
		{ 149, 474, 34 },
		{ 149, 475, 34 },
		{ 149, 476, 34 },
		{ 149, 477, 34 },
		{ 149, 478, 34 },
		{ 149, 479, 34 },
		{ 149, 480, 34 },
		{ 149, 481, 34 },
		{ 149, 482, 34 },
		{ 149, 483, 34 },
		{ 149, 484, 34 },
		{ 149, 485, 34 },
		{ 149, 486, 34 },
		{ 149, 487, 34 },
		{ 149, 488, 34 },
		{ 149, 489, 34 },
		{ 149, 490, 34 },
		{ 149, 491, 34 },
		{ 149, 492, 34 },
		{ 149, 493, 34 },
		{ 149, 494, 34 },
		{ 149, 495, 34 },
		{ 149, 496, 34 },
		{ 149, 497, 34 },
		{ 149, 498, 34 },
		{ 149, 499, 34 },
		{ 149, 500, 34 },
		{ 149, 501, 34 },
		{ 149, 502, 34 },
		{ 149, 503, 34 },
		{ 149, 504, 34 },
		{ 149, 505, 34 },
		{ 149, 506, 34 },
		{ 149, 507, 34 },
		{ 149, 508, 34 },
		{ 149, 509, 34 },
		{ 149, 510, 34 },
		{ 149, 511, 34 },
		{ 149, 512, 34 },
		{ 150, 429, 34 },
		{ 150, 430, 34 },
		{ 150, 431, 34 },
		{ 150, 432, 34 },
		{ 150, 433, 34 },
		{ 150, 434, 34 },
		{ 150, 435, 34 },
		{ 150, 436, 34 },
		{ 150, 437, 34 },
		{ 150, 438, 34 },
		{ 150, 439, 34 },
		{ 150, 440, 34 },
		{ 150, 441, 34 },
		{ 150, 442, 34 },
		{ 150, 443, 34 },
		{ 150, 444, 34 },
		{ 150, 445, 34 },
		{ 150, 446, 34 },
		{ 150, 447, 34 },
		{ 150, 448, 34 },
		{ 150, 449, 34 },
		{ 150, 450, 34 },
		{ 150, 451, 34 },
		{ 150, 452, 34 },
		{ 150, 453, 34 },
		{ 150, 454, 34 },
		{ 150, 455, 34 },
		{ 150, 456, 34 },
		{ 150, 457, 34 },
		{ 150, 458, 34 },
		{ 150, 459, 34 },
		{ 150, 460, 34 },
		{ 150, 461, 34 },
		{ 150, 462, 34 },
		{ 150, 463, 34 },
		{ 150, 464, 34 },
		{ 150, 465, 34 },
		{ 150, 466, 34 },
		{ 150, 467, 34 },
		{ 150, 468, 34 },
		{ 150, 469, 34 },
		{ 150, 470, 34 },
		{ 150, 471, 34 },
		{ 150, 472, 34 },
		{ 150, 473, 34 },
		{ 150, 474, 34 },
		{ 150, 475, 34 },
		{ 150, 476, 34 },
		{ 150, 477, 34 },
		{ 150, 478, 34 },
		{ 150, 479, 34 },
		{ 150, 480, 34 },
		{ 150, 481, 34 },
		{ 150, 482, 34 },
		{ 150, 483, 34 },
		{ 150, 484, 34 },
		{ 150, 485, 34 },
		{ 150, 486, 34 },
		{ 150, 487, 34 },
		{ 150, 488, 34 },
		{ 150, 489, 34 },
		{ 150, 490, 34 },
		{ 150, 491, 34 },
		{ 150, 492, 34 },
		{ 150, 493, 34 },
		{ 150, 494, 34 },
		{ 150, 495, 34 },
		{ 150, 496, 34 },
		{ 150, 497, 34 },
		{ 150, 498, 34 },
		{ 150, 499, 34 },
		{ 150, 500, 34 },
		{ 150, 501, 34 },
		{ 150, 502, 34 },
		{ 150, 503, 34 },
		{ 150, 504, 34 },
		{ 150, 505, 34 },
		{ 150, 506, 34 },
		{ 150, 507, 34 },
		{ 150, 508, 34 },
		{ 150, 509, 34 },
		{ 150, 510, 34 },
		{ 150, 511, 34 },
		{ 150, 512, 34 },
		{ 151, 427, 34 },
		{ 151, 428, 34 },
		{ 151, 429, 34 },
		{ 151, 430, 34 },
		{ 151, 431, 34 },
		{ 151, 432, 34 },
		{ 151, 433, 34 },
		{ 151, 434, 34 },
		{ 151, 435, 34 },
		{ 151, 436, 34 },
		{ 151, 437, 34 },
		{ 151, 438, 34 },
		{ 151, 439, 34 },
		{ 151, 440, 34 },
		{ 151, 441, 34 },
		{ 151, 442, 34 },
		{ 151, 443, 34 },
		{ 151, 444, 34 },
		{ 151, 445, 34 },
		{ 151, 446, 34 },
		{ 151, 447, 34 },
		{ 151, 448, 34 },
		{ 151, 449, 34 },
		{ 151, 450, 34 },
		{ 151, 451, 34 },
		{ 151, 452, 34 },
		{ 151, 453, 34 },
		{ 151, 454, 34 },
		{ 151, 455, 34 },
		{ 151, 456, 34 },
		{ 151, 457, 34 },
		{ 151, 458, 34 },
		{ 151, 459, 34 },
		{ 151, 460, 34 },
		{ 151, 461, 34 },
		{ 151, 462, 34 },
		{ 151, 463, 34 },
		{ 151, 464, 34 },
		{ 151, 465, 34 },
		{ 151, 466, 34 },
		{ 151, 467, 34 },
		{ 151, 468, 34 },
		{ 151, 469, 34 },
		{ 151, 470, 34 },
		{ 151, 471, 34 },
		{ 151, 472, 34 },
		{ 151, 473, 34 },
		{ 151, 474, 34 },
		{ 151, 475, 34 },
		{ 151, 476, 34 },
		{ 151, 477, 34 },
		{ 151, 478, 34 },
		{ 151, 479, 34 },
		{ 151, 480, 34 },
		{ 151, 481, 34 },
		{ 151, 482, 34 },
		{ 151, 483, 34 },
		{ 151, 484, 34 },
		{ 151, 485, 34 },
		{ 151, 486, 34 },
		{ 151, 487, 34 },
		{ 151, 488, 34 },
		{ 151, 489, 34 },
		{ 151, 490, 34 },
		{ 151, 491, 34 },
		{ 151, 492, 34 },
		{ 151, 493, 34 },
		{ 151, 494, 34 },
		{ 151, 495, 34 },
		{ 151, 496, 34 },
		{ 151, 497, 34 },
		{ 151, 498, 34 },
		{ 151, 499, 34 },
		{ 151, 500, 34 },
		{ 151, 501, 34 },
		{ 151, 502, 34 },
		{ 151, 503, 34 },
		{ 151, 504, 34 },
		{ 151, 505, 34 },
		{ 151, 506, 34 },
		{ 151, 507, 34 },
		{ 151, 508, 34 },
		{ 151, 509, 34 },
		{ 151, 510, 34 },
		{ 151, 511, 34 },
		{ 151, 512, 34 },
		{ 152, 426, 34 },
		{ 152, 427, 34 },
		{ 152, 428, 34 },
		{ 152, 429, 34 },
		{ 152, 430, 34 },
		{ 152, 431, 34 },
		{ 152, 432, 34 },
		{ 152, 433, 34 },
		{ 152, 434, 34 },
		{ 152, 435, 34 },
		{ 152, 436, 34 },
		{ 152, 437, 34 },
		{ 152, 438, 34 },
		{ 152, 439, 34 },
		{ 152, 440, 34 },
		{ 152, 441, 34 },
		{ 152, 442, 34 },
		{ 152, 443, 34 },
		{ 152, 444, 34 },
		{ 152, 445, 34 },
		{ 152, 446, 34 },
		{ 152, 447, 34 },
		{ 152, 448, 34 },
		{ 152, 449, 34 },
		{ 152, 450, 34 },
		{ 152, 451, 34 },
		{ 152, 452, 34 },
		{ 152, 453, 34 },
		{ 152, 454, 34 },
		{ 152, 455, 34 },
		{ 152, 456, 34 },
		{ 152, 457, 34 },
		{ 152, 458, 34 },
		{ 152, 459, 34 },
		{ 152, 460, 34 },
		{ 152, 461, 34 },
		{ 152, 462, 34 },
		{ 152, 463, 34 },
		{ 152, 464, 34 },
		{ 152, 465, 34 },
		{ 152, 466, 34 },
		{ 152, 467, 34 },
		{ 152, 468, 34 },
		{ 152, 469, 34 },
		{ 152, 470, 34 },
		{ 152, 471, 34 },
		{ 152, 472, 34 },
		{ 152, 473, 34 },
		{ 152, 474, 34 },
		{ 152, 475, 34 },
		{ 152, 476, 34 },
		{ 152, 477, 34 },
		{ 152, 478, 34 },
		{ 152, 479, 34 },
		{ 152, 480, 34 },
		{ 152, 481, 34 },
		{ 152, 482, 34 },
		{ 152, 483, 34 },
		{ 152, 484, 34 },
		{ 152, 485, 34 },
		{ 152, 486, 34 },
		{ 152, 487, 34 },
		{ 152, 488, 34 },
		{ 152, 489, 34 },
		{ 152, 490, 34 },
		{ 152, 491, 34 },
		{ 152, 492, 34 },
		{ 152, 493, 34 },
		{ 152, 494, 34 },
		{ 152, 495, 34 },
		{ 152, 496, 34 },
		{ 152, 497, 34 },
		{ 152, 498, 34 },
		{ 152, 499, 34 },
		{ 152, 500, 34 },
		{ 152, 501, 34 },
		{ 152, 502, 34 },
		{ 152, 503, 34 },
		{ 152, 504, 34 },
		{ 152, 505, 34 },
		{ 152, 506, 34 },
		{ 152, 507, 34 },
		{ 152, 508, 34 },
		{ 152, 509, 34 },
		{ 152, 510, 34 },
		{ 152, 511, 34 },
		{ 152, 512, 34 },
		{ 153, 424, 34 },
		{ 153, 425, 34 },
		{ 153, 426, 34 },
		{ 153, 427, 34 },
		{ 153, 428, 34 },
		{ 153, 429, 34 },
		{ 153, 430, 34 },
		{ 153, 431, 34 },
		{ 153, 432, 34 },
		{ 153, 433, 34 },
		{ 153, 434, 34 },
		{ 153, 435, 34 },
		{ 153, 436, 34 },
		{ 153, 437, 34 },
		{ 153, 438, 34 },
		{ 153, 439, 34 },
		{ 153, 440, 34 },
		{ 153, 441, 34 },
		{ 153, 442, 34 },
		{ 153, 443, 34 },
		{ 153, 444, 34 },
		{ 153, 445, 34 },
		{ 153, 446, 34 },
		{ 153, 447, 34 },
		{ 153, 448, 34 },
		{ 153, 449, 34 },
		{ 153, 450, 34 },
		{ 153, 451, 34 },
		{ 153, 452, 34 },
		{ 153, 453, 34 },
		{ 153, 454, 34 },
		{ 153, 455, 34 },
		{ 153, 456, 34 },
		{ 153, 457, 34 },
		{ 153, 458, 34 },
		{ 153, 459, 34 },
		{ 153, 460, 34 },
		{ 153, 461, 34 },
		{ 153, 462, 34 },
		{ 153, 463, 34 },
		{ 153, 464, 34 },
		{ 153, 465, 34 },
		{ 153, 466, 34 },
		{ 153, 467, 34 },
		{ 153, 468, 34 },
		{ 153, 469, 34 },
		{ 153, 470, 34 },
		{ 153, 471, 34 },
		{ 153, 472, 34 },
		{ 153, 473, 34 },
		{ 153, 474, 34 },
		{ 153, 475, 34 },
		{ 153, 476, 34 },
		{ 153, 477, 34 },
		{ 153, 478, 34 },
		{ 153, 479, 34 },
		{ 153, 480, 34 },
		{ 153, 481, 34 },
		{ 153, 482, 34 },
		{ 153, 483, 34 },
		{ 153, 484, 34 },
		{ 153, 485, 34 },
		{ 153, 486, 34 },
		{ 153, 487, 34 },
		{ 153, 488, 34 },
		{ 153, 489, 34 },
		{ 153, 490, 34 },
		{ 153, 491, 34 },
		{ 153, 492, 34 },
		{ 153, 493, 34 },
		{ 153, 494, 34 },
		{ 153, 495, 34 },
		{ 153, 496, 34 },
		{ 153, 497, 34 },
		{ 153, 498, 34 },
		{ 153, 499, 34 },
		{ 153, 500, 34 },
		{ 153, 501, 34 },
		{ 153, 502, 34 },
		{ 153, 503, 34 },
		{ 153, 504, 34 },
		{ 153, 505, 34 },
		{ 153, 506, 34 },
		{ 153, 507, 34 },
		{ 153, 508, 34 },
		{ 153, 509, 34 },
		{ 153, 510, 34 },
		{ 153, 511, 34 },
		{ 153, 512, 34 },
		{ 154, 422, 34 },
		{ 154, 423, 34 },
		{ 154, 424, 34 },
		{ 154, 425, 34 },
		{ 154, 426, 34 },
		{ 154, 427, 34 },
		{ 154, 428, 34 },
		{ 154, 429, 34 },
		{ 154, 430, 34 },
		{ 154, 431, 34 },
		{ 154, 432, 34 },
		{ 154, 433, 34 },
		{ 154, 434, 34 },
		{ 154, 435, 34 },
		{ 154, 436, 34 },
		{ 154, 437, 34 },
		{ 154, 438, 34 },
		{ 154, 439, 34 },
		{ 154, 440, 34 },
		{ 154, 441, 34 },
		{ 154, 442, 34 },
		{ 154, 443, 34 },
		{ 154, 444, 34 },
		{ 154, 445, 34 },
		{ 154, 446, 34 },
		{ 154, 447, 34 },
		{ 154, 448, 34 },
		{ 154, 449, 34 },
		{ 154, 450, 34 },
		{ 154, 451, 34 },
		{ 154, 452, 34 },
		{ 154, 453, 34 },
		{ 154, 454, 34 },
		{ 154, 455, 34 },
		{ 154, 456, 34 },
		{ 154, 457, 34 },
		{ 154, 458, 34 },
		{ 154, 459, 34 },
		{ 154, 460, 34 },
		{ 154, 461, 34 },
		{ 154, 462, 34 },
		{ 154, 463, 34 },
		{ 154, 464, 34 },
		{ 154, 465, 34 },
		{ 154, 466, 34 },
		{ 154, 467, 34 },
		{ 154, 468, 34 },
		{ 154, 469, 34 },
		{ 154, 470, 34 },
		{ 154, 471, 34 },
		{ 154, 472, 34 },
		{ 154, 473, 34 },
		{ 154, 474, 34 },
		{ 154, 475, 34 },
		{ 154, 476, 34 },
		{ 154, 477, 34 },
		{ 154, 478, 34 },
		{ 154, 479, 34 },
		{ 154, 480, 34 },
		{ 154, 481, 34 },
		{ 154, 482, 34 },
		{ 154, 483, 34 },
		{ 154, 484, 34 },
		{ 154, 485, 34 },
		{ 154, 486, 34 },
		{ 154, 487, 34 },
		{ 154, 488, 34 },
		{ 154, 489, 34 },
		{ 154, 490, 34 },
		{ 154, 491, 34 },
		{ 154, 492, 34 },
		{ 154, 493, 34 },
		{ 154, 494, 34 },
		{ 154, 495, 34 },
		{ 154, 496, 34 },
		{ 154, 497, 34 },
		{ 154, 498, 34 },
		{ 154, 499, 34 },
		{ 154, 500, 34 },
		{ 154, 501, 34 },
		{ 154, 502, 34 },
		{ 154, 503, 34 },
		{ 154, 504, 34 },
		{ 154, 505, 34 },
		{ 154, 506, 34 },
		{ 154, 507, 34 },
		{ 154, 508, 34 },
		{ 154, 509, 34 },
		{ 154, 510, 34 },
		{ 154, 511, 34 },
		{ 154, 512, 34 },
		{ 155, 421, 34 },
		{ 155, 422, 34 },
		{ 155, 423, 34 },
		{ 155, 424, 34 },
		{ 155, 425, 34 },
		{ 155, 426, 34 },
		{ 155, 427, 34 },
		{ 155, 428, 34 },
		{ 155, 429, 34 },
		{ 155, 430, 34 },
		{ 155, 431, 34 },
		{ 155, 432, 34 },
		{ 155, 433, 34 },
		{ 155, 434, 34 },
		{ 155, 435, 34 },
		{ 155, 436, 34 },
		{ 155, 437, 34 },
		{ 155, 438, 34 },
		{ 155, 439, 34 },
		{ 155, 440, 34 },
		{ 155, 441, 34 },
		{ 155, 442, 34 },
		{ 155, 443, 34 },
		{ 155, 444, 34 },
		{ 155, 445, 34 },
		{ 155, 446, 34 },
		{ 155, 447, 34 },
		{ 155, 448, 34 },
		{ 155, 449, 34 },
		{ 155, 450, 34 },
		{ 155, 451, 34 },
		{ 155, 452, 34 },
		{ 155, 453, 34 },
		{ 155, 454, 34 },
		{ 155, 455, 34 },
		{ 155, 456, 34 },
		{ 155, 457, 34 },
		{ 155, 458, 34 },
		{ 155, 459, 34 },
		{ 155, 460, 34 },
		{ 155, 461, 34 },
		{ 155, 462, 34 },
		{ 155, 463, 34 },
		{ 155, 464, 34 },
		{ 155, 465, 34 },
		{ 155, 466, 34 },
		{ 155, 467, 34 },
		{ 155, 468, 34 },
		{ 155, 469, 34 },
		{ 155, 470, 34 },
		{ 155, 471, 34 },
		{ 155, 472, 34 },
		{ 155, 473, 34 },
		{ 155, 474, 34 },
		{ 155, 475, 34 },
		{ 155, 476, 34 },
		{ 155, 477, 34 },
		{ 155, 478, 34 },
		{ 155, 479, 34 },
		{ 155, 480, 34 },
		{ 155, 481, 34 },
		{ 155, 482, 34 },
		{ 155, 483, 34 },
		{ 155, 484, 34 },
		{ 155, 485, 34 },
		{ 155, 486, 34 },
		{ 155, 487, 34 },
		{ 155, 488, 34 },
		{ 155, 489, 34 },
		{ 155, 490, 34 },
		{ 155, 491, 34 },
		{ 155, 492, 34 },
		{ 155, 493, 34 },
		{ 155, 494, 34 },
		{ 155, 495, 34 },
		{ 155, 496, 34 },
		{ 155, 497, 34 },
		{ 155, 498, 34 },
		{ 155, 499, 34 },
		{ 155, 500, 34 },
		{ 155, 501, 34 },
		{ 155, 502, 34 },
		{ 155, 503, 34 },
		{ 155, 504, 34 },
		{ 155, 505, 34 },
		{ 155, 506, 34 },
		{ 155, 507, 34 },
		{ 155, 508, 34 },
		{ 155, 509, 34 },
		{ 155, 510, 34 },
		{ 155, 511, 34 },
		{ 155, 512, 34 },
		{ 156, 419, 34 },
		{ 156, 420, 34 },
		{ 156, 421, 34 },
		{ 156, 422, 34 },
		{ 156, 423, 34 },
		{ 156, 424, 34 },
		{ 156, 425, 34 },
		{ 156, 426, 34 },
		{ 156, 427, 34 },
		{ 156, 428, 34 },
		{ 156, 429, 34 },
		{ 156, 430, 34 },
		{ 156, 431, 34 },
		{ 156, 432, 34 },
		{ 156, 433, 34 },
		{ 156, 434, 34 },
		{ 156, 435, 34 },
		{ 156, 436, 34 },
		{ 156, 437, 34 },
		{ 156, 438, 34 },
		{ 156, 439, 34 },
		{ 156, 440, 34 },
		{ 156, 441, 34 },
		{ 156, 442, 34 },
		{ 156, 443, 34 },
		{ 156, 444, 34 },
		{ 156, 445, 34 },
		{ 156, 446, 34 },
		{ 156, 447, 34 },
		{ 156, 448, 34 },
		{ 156, 449, 34 },
		{ 156, 450, 34 },
		{ 156, 451, 34 },
		{ 156, 452, 34 },
		{ 156, 453, 34 },
		{ 156, 454, 34 },
		{ 156, 455, 34 },
		{ 156, 456, 34 },
		{ 156, 457, 34 },
		{ 156, 458, 34 },
		{ 156, 459, 34 },
		{ 156, 460, 34 },
		{ 156, 461, 34 },
		{ 156, 462, 34 },
		{ 156, 463, 34 },
		{ 156, 464, 34 },
		{ 156, 465, 34 },
		{ 156, 466, 34 },
		{ 156, 467, 34 },
		{ 156, 468, 34 },
		{ 156, 469, 34 },
		{ 156, 470, 34 },
		{ 156, 471, 34 },
		{ 156, 472, 34 },
		{ 156, 473, 34 },
		{ 156, 474, 34 },
		{ 156, 475, 34 },
		{ 156, 476, 34 },
		{ 156, 477, 34 },
		{ 156, 478, 34 },
		{ 156, 479, 34 },
		{ 156, 480, 34 },
		{ 156, 481, 34 },
		{ 156, 482, 34 },
		{ 156, 483, 34 },
		{ 156, 484, 34 },
		{ 156, 485, 34 },
		{ 156, 486, 34 },
		{ 156, 487, 34 },
		{ 156, 488, 34 },
		{ 156, 489, 34 },
		{ 156, 490, 34 },
		{ 156, 491, 34 },
		{ 156, 492, 34 },
		{ 156, 493, 34 },
		{ 156, 494, 34 },
		{ 156, 495, 34 },
		{ 156, 496, 34 },
		{ 156, 497, 34 },
		{ 156, 498, 34 },
		{ 156, 499, 34 },
		{ 156, 500, 34 },
		{ 156, 501, 34 },
		{ 156, 502, 34 },
		{ 156, 503, 34 },
		{ 156, 504, 34 },
		{ 156, 505, 34 },
		{ 156, 506, 34 },
		{ 156, 507, 34 },
		{ 156, 508, 34 },
		{ 156, 509, 34 },
		{ 156, 510, 34 },
		{ 156, 511, 34 },
		{ 156, 512, 34 },
		{ 157, 418, 34 },
		{ 157, 419, 34 },
		{ 157, 420, 34 },
		{ 157, 421, 34 },
		{ 157, 422, 34 },
		{ 157, 423, 34 },
		{ 157, 424, 34 },
		{ 157, 425, 34 },
		{ 157, 426, 34 },
		{ 157, 427, 34 },
		{ 157, 428, 34 },
		{ 157, 429, 34 },
		{ 157, 430, 34 },
		{ 157, 431, 34 },
		{ 157, 432, 34 },
		{ 157, 433, 34 },
		{ 157, 434, 34 },
		{ 157, 435, 34 },
		{ 157, 436, 34 },
		{ 157, 437, 34 },
		{ 157, 438, 34 },
		{ 157, 439, 34 },
		{ 157, 440, 34 },
		{ 157, 441, 34 },
		{ 157, 442, 34 },
		{ 157, 443, 34 },
		{ 157, 444, 34 },
		{ 157, 445, 34 },
		{ 157, 446, 34 },
		{ 157, 447, 34 },
		{ 157, 448, 34 },
		{ 157, 449, 34 },
		{ 157, 450, 34 },
		{ 157, 451, 34 },
		{ 157, 452, 34 },
		{ 157, 453, 34 },
		{ 157, 454, 34 },
		{ 157, 455, 34 },
		{ 157, 456, 34 },
		{ 157, 457, 34 },
		{ 157, 458, 34 },
		{ 157, 459, 34 },
		{ 157, 460, 34 },
		{ 157, 461, 34 },
		{ 157, 462, 34 },
		{ 157, 463, 34 },
		{ 157, 464, 34 },
		{ 157, 465, 34 },
		{ 157, 466, 34 },
		{ 157, 467, 34 },
		{ 157, 468, 34 },
		{ 157, 469, 34 },
		{ 157, 470, 34 },
		{ 157, 471, 34 },
		{ 157, 472, 34 },
		{ 157, 473, 34 },
		{ 157, 474, 34 },
		{ 157, 475, 34 },
		{ 157, 476, 34 },
		{ 157, 477, 34 },
		{ 157, 478, 34 },
		{ 157, 479, 34 },
		{ 157, 480, 34 },
		{ 157, 481, 34 },
		{ 157, 482, 34 },
		{ 157, 483, 34 },
		{ 157, 484, 34 },
		{ 157, 485, 34 },
		{ 157, 486, 34 },
		{ 157, 487, 34 },
		{ 157, 488, 34 },
		{ 157, 489, 34 },
		{ 157, 490, 34 },
		{ 157, 491, 34 },
		{ 157, 492, 34 },
		{ 157, 493, 34 },
		{ 157, 494, 34 },
		{ 157, 495, 34 },
		{ 157, 496, 34 },
		{ 157, 497, 34 },
		{ 157, 498, 34 },
		{ 157, 499, 34 },
		{ 157, 500, 34 },
		{ 157, 501, 34 },
		{ 157, 502, 34 },
		{ 157, 503, 34 },
		{ 157, 504, 34 },
		{ 157, 505, 34 },
		{ 157, 506, 34 },
		{ 157, 507, 34 },
		{ 157, 508, 34 },
		{ 157, 509, 34 },
		{ 157, 510, 34 },
		{ 157, 511, 34 },
		{ 157, 512, 34 },
		{ 158, 416, 34 },
		{ 158, 417, 34 },
		{ 158, 418, 34 },
		{ 158, 419, 34 },
		{ 158, 420, 34 },
		{ 158, 421, 34 },
		{ 158, 422, 34 },
		{ 158, 423, 34 },
		{ 158, 424, 34 },
		{ 158, 425, 34 },
		{ 158, 426, 34 },
		{ 158, 427, 34 },
		{ 158, 428, 34 },
		{ 158, 429, 34 },
		{ 158, 430, 34 },
		{ 158, 431, 34 },
		{ 158, 432, 34 },
		{ 158, 433, 34 },
		{ 158, 434, 34 },
		{ 158, 435, 34 },
		{ 158, 436, 34 },
		{ 158, 437, 34 },
		{ 158, 438, 34 },
		{ 158, 439, 34 },
		{ 158, 440, 34 },
		{ 158, 441, 34 },
		{ 158, 442, 34 },
		{ 158, 443, 34 },
		{ 158, 444, 34 },
		{ 158, 445, 34 },
		{ 158, 446, 34 },
		{ 158, 447, 34 },
		{ 158, 448, 34 },
		{ 158, 449, 34 },
		{ 158, 450, 34 },
		{ 158, 451, 34 },
		{ 158, 452, 34 },
		{ 158, 453, 34 },
		{ 158, 454, 34 },
		{ 158, 455, 34 },
		{ 158, 456, 34 },
		{ 158, 457, 34 },
		{ 158, 458, 34 },
		{ 158, 459, 34 },
		{ 158, 460, 34 },
		{ 158, 461, 34 },
		{ 158, 462, 34 },
		{ 158, 463, 34 },
		{ 158, 464, 34 },
		{ 158, 465, 34 },
		{ 158, 466, 34 },
		{ 158, 467, 34 },
		{ 158, 468, 34 },
		{ 158, 469, 34 },
		{ 158, 470, 34 },
		{ 158, 471, 34 },
		{ 158, 472, 34 },
		{ 158, 473, 34 },
		{ 158, 474, 34 },
		{ 158, 475, 34 },
		{ 158, 476, 34 },
		{ 158, 477, 34 },
		{ 158, 478, 34 },
		{ 158, 479, 34 },
		{ 158, 480, 34 },
		{ 158, 481, 34 },
		{ 158, 482, 34 },
		{ 158, 483, 34 },
		{ 158, 484, 34 },
		{ 158, 485, 34 },
		{ 158, 486, 34 },
		{ 158, 487, 34 },
		{ 158, 488, 34 },
		{ 158, 489, 34 },
		{ 158, 490, 34 },
		{ 158, 491, 34 },
		{ 158, 492, 34 },
		{ 158, 493, 34 },
		{ 158, 494, 34 },
		{ 158, 495, 34 },
		{ 158, 496, 34 },
		{ 158, 497, 34 },
		{ 158, 498, 34 },
		{ 158, 499, 34 },
		{ 158, 500, 34 },
		{ 158, 501, 34 },
		{ 158, 502, 34 },
		{ 158, 503, 34 },
		{ 158, 504, 34 },
		{ 158, 505, 34 },
		{ 158, 506, 34 },
		{ 158, 507, 34 },
		{ 158, 508, 34 },
		{ 158, 509, 34 },
		{ 158, 510, 34 },
		{ 158, 511, 34 },
		{ 158, 512, 34 },
		{ 159, 414, 34 },
		{ 159, 415, 34 },
		{ 159, 416, 34 },
		{ 159, 417, 34 },
		{ 159, 418, 34 },
		{ 159, 419, 34 },
		{ 159, 420, 34 },
		{ 159, 421, 34 },
		{ 159, 422, 34 },
		{ 159, 423, 34 },
		{ 159, 424, 34 },
		{ 159, 425, 34 },
		{ 159, 426, 34 },
		{ 159, 427, 34 },
		{ 159, 428, 34 },
		{ 159, 429, 34 },
		{ 159, 430, 34 },
		{ 159, 431, 34 },
		{ 159, 432, 34 },
		{ 159, 433, 34 },
		{ 159, 434, 34 },
		{ 159, 435, 34 },
		{ 159, 436, 34 },
		{ 159, 437, 34 },
		{ 159, 438, 34 },
		{ 159, 439, 34 },
		{ 159, 440, 34 },
		{ 159, 441, 34 },
		{ 159, 442, 34 },
		{ 159, 443, 34 },
		{ 159, 444, 34 },
		{ 159, 445, 34 },
		{ 159, 446, 34 },
		{ 159, 447, 34 },
		{ 159, 448, 34 },
		{ 159, 449, 34 },
		{ 159, 450, 34 },
		{ 159, 451, 34 },
		{ 159, 452, 34 },
		{ 159, 453, 34 },
		{ 159, 454, 34 },
		{ 159, 455, 34 },
		{ 159, 456, 34 },
		{ 159, 457, 34 },
		{ 159, 458, 34 },
		{ 159, 459, 34 },
		{ 159, 460, 34 },
		{ 159, 461, 34 },
		{ 159, 462, 34 },
		{ 159, 463, 34 },
		{ 159, 464, 34 },
		{ 159, 465, 34 },
		{ 159, 466, 34 },
		{ 159, 467, 34 },
		{ 159, 468, 34 },
		{ 159, 469, 34 },
		{ 159, 470, 34 },
		{ 159, 471, 34 },
		{ 159, 472, 34 },
		{ 159, 473, 34 },
		{ 159, 474, 34 },
		{ 159, 475, 34 },
		{ 159, 476, 34 },
		{ 159, 477, 34 },
		{ 159, 478, 34 },
		{ 159, 479, 34 },
		{ 159, 480, 34 },
		{ 159, 481, 34 },
		{ 159, 482, 34 },
		{ 159, 483, 34 },
		{ 159, 484, 34 },
		{ 159, 485, 34 },
		{ 159, 486, 34 },
		{ 159, 487, 34 },
		{ 159, 488, 34 },
		{ 159, 489, 34 },
		{ 159, 490, 34 },
		{ 159, 491, 34 },
		{ 159, 492, 34 },
		{ 159, 493, 34 },
		{ 159, 494, 34 },
		{ 159, 495, 34 },
		{ 159, 496, 34 },
		{ 159, 497, 34 },
		{ 159, 498, 34 },
		{ 159, 499, 34 },
		{ 159, 500, 34 },
		{ 159, 501, 34 },
		{ 159, 502, 34 },
		{ 159, 503, 34 },
		{ 159, 504, 34 },
		{ 159, 505, 34 },
		{ 159, 506, 34 },
		{ 159, 507, 34 },
		{ 159, 508, 34 },
		{ 159, 509, 34 },
		{ 159, 510, 34 },
		{ 159, 511, 34 },
		{ 159, 512, 34 },
		{ 160, 413, 34 },
		{ 160, 414, 34 },
		{ 160, 415, 34 },
		{ 160, 416, 34 },
		{ 160, 417, 34 },
		{ 160, 418, 34 },
		{ 160, 419, 34 },
		{ 160, 420, 34 },
		{ 160, 421, 34 },
		{ 160, 422, 34 },
		{ 160, 423, 34 },
		{ 160, 424, 34 },
		{ 160, 425, 34 },
		{ 160, 426, 34 },
		{ 160, 427, 34 },
		{ 160, 428, 34 },
		{ 160, 429, 34 },
		{ 160, 430, 34 },
		{ 160, 431, 34 },
		{ 160, 432, 34 },
		{ 160, 433, 34 },
		{ 160, 434, 34 },
		{ 160, 435, 34 },
		{ 160, 436, 34 },
		{ 160, 437, 34 },
		{ 160, 438, 34 },
		{ 160, 439, 34 },
		{ 160, 440, 34 },
		{ 160, 441, 34 },
		{ 160, 442, 34 },
		{ 160, 443, 34 },
		{ 160, 444, 34 },
		{ 160, 445, 34 },
		{ 160, 446, 34 },
		{ 160, 447, 34 },
		{ 160, 448, 34 },
		{ 160, 449, 34 },
		{ 160, 450, 34 },
		{ 160, 451, 34 },
		{ 160, 452, 34 },
		{ 160, 453, 34 },
		{ 160, 454, 34 },
		{ 160, 455, 34 },
		{ 160, 456, 34 },
		{ 160, 457, 34 },
		{ 160, 458, 34 },
		{ 160, 459, 34 },
		{ 160, 460, 34 },
		{ 160, 461, 34 },
		{ 160, 462, 34 },
		{ 160, 463, 34 },
		{ 160, 464, 34 },
		{ 160, 465, 34 },
		{ 160, 466, 34 },
		{ 160, 467, 34 },
		{ 160, 468, 34 },
		{ 160, 469, 34 },
		{ 160, 470, 34 },
		{ 160, 471, 34 },
		{ 160, 472, 34 },
		{ 160, 473, 34 },
		{ 160, 474, 34 },
		{ 160, 475, 34 },
		{ 160, 476, 34 },
		{ 160, 477, 34 },
		{ 160, 478, 34 },
		{ 160, 479, 34 },
		{ 160, 480, 34 },
		{ 160, 481, 34 },
		{ 160, 482, 34 },
		{ 160, 483, 34 },
		{ 160, 484, 34 },
		{ 160, 485, 34 },
		{ 160, 486, 34 },
		{ 160, 487, 34 },
		{ 160, 488, 34 },
		{ 160, 489, 34 },
		{ 160, 490, 34 },
		{ 160, 491, 34 },
		{ 160, 492, 34 },
		{ 160, 493, 34 },
		{ 160, 494, 34 },
		{ 160, 495, 34 },
		{ 160, 496, 34 },
		{ 160, 497, 34 },
		{ 160, 498, 34 },
		{ 160, 499, 34 },
		{ 160, 500, 34 },
		{ 160, 501, 34 },
		{ 160, 502, 34 },
		{ 160, 503, 34 },
		{ 160, 504, 34 },
		{ 160, 505, 34 },
		{ 160, 506, 34 },
		{ 160, 507, 34 },
		{ 160, 508, 34 },
		{ 160, 509, 34 },
		{ 160, 510, 34 },
		{ 160, 511, 34 },
		{ 160, 512, 34 },
		{ 161, 411, 34 },
		{ 161, 412, 34 },
		{ 161, 413, 34 },
		{ 161, 414, 34 },
		{ 161, 415, 34 },
		{ 161, 416, 34 },
		{ 161, 417, 34 },
		{ 161, 418, 34 },
		{ 161, 419, 34 },
		{ 161, 420, 34 },
		{ 161, 421, 34 },
		{ 161, 422, 34 },
		{ 161, 423, 34 },
		{ 161, 424, 34 },
		{ 161, 425, 34 },
		{ 161, 426, 34 },
		{ 161, 427, 34 },
		{ 161, 428, 34 },
		{ 161, 429, 34 },
		{ 161, 430, 34 },
		{ 161, 431, 34 },
		{ 161, 432, 34 },
		{ 161, 433, 34 },
		{ 161, 434, 34 },
		{ 161, 435, 34 },
		{ 161, 436, 34 },
		{ 161, 437, 34 },
		{ 161, 438, 34 },
		{ 161, 439, 34 },
		{ 161, 440, 34 },
		{ 161, 441, 34 },
		{ 161, 442, 34 },
		{ 161, 443, 34 },
		{ 161, 444, 34 },
		{ 161, 445, 34 },
		{ 161, 446, 34 },
		{ 161, 447, 34 },
		{ 161, 448, 34 },
		{ 161, 449, 34 },
		{ 161, 450, 34 },
		{ 161, 451, 34 },
		{ 161, 452, 34 },
		{ 161, 453, 34 },
		{ 161, 454, 34 },
		{ 161, 455, 34 },
		{ 161, 456, 34 },
		{ 161, 457, 34 },
		{ 161, 458, 34 },
		{ 161, 459, 34 },
		{ 161, 460, 34 },
		{ 161, 461, 34 },
		{ 161, 462, 34 },
		{ 161, 463, 34 },
		{ 161, 464, 34 },
		{ 161, 465, 34 },
		{ 161, 466, 34 },
		{ 161, 467, 34 },
		{ 161, 468, 34 },
		{ 161, 469, 34 },
		{ 161, 470, 34 },
		{ 161, 471, 34 },
		{ 161, 472, 34 },
		{ 161, 473, 34 },
		{ 161, 474, 34 },
		{ 161, 475, 34 },
		{ 161, 476, 34 },
		{ 161, 477, 34 },
		{ 161, 478, 34 },
		{ 161, 479, 34 },
		{ 161, 480, 34 },
		{ 161, 481, 34 },
		{ 161, 482, 34 },
		{ 161, 483, 34 },
		{ 161, 484, 34 },
		{ 161, 485, 34 },
		{ 161, 486, 34 },
		{ 161, 487, 34 },
		{ 161, 488, 34 },
		{ 161, 489, 34 },
		{ 161, 490, 34 },
		{ 161, 491, 34 },
		{ 161, 492, 34 },
		{ 161, 493, 34 },
		{ 161, 494, 34 },
		{ 161, 495, 34 },
		{ 161, 496, 34 },
		{ 161, 497, 34 },
		{ 161, 498, 34 },
		{ 161, 499, 34 },
		{ 161, 500, 34 },
		{ 161, 501, 34 },
		{ 161, 502, 34 },
		{ 161, 503, 34 },
		{ 161, 504, 34 },
		{ 161, 505, 34 },
		{ 161, 506, 34 },
		{ 161, 507, 34 },
		{ 161, 508, 34 },
		{ 161, 509, 34 },
		{ 161, 510, 34 },
		{ 161, 511, 34 },
		{ 161, 512, 34 },
		{ 162, 410, 34 },
		{ 162, 411, 34 },
		{ 162, 412, 34 },
		{ 162, 413, 34 },
		{ 162, 414, 34 },
		{ 162, 415, 34 },
		{ 162, 416, 34 },
		{ 162, 417, 34 },
		{ 162, 418, 34 },
		{ 162, 419, 34 },
		{ 162, 420, 34 },
		{ 162, 421, 34 },
		{ 162, 422, 34 },
		{ 162, 423, 34 },
		{ 162, 424, 34 },
		{ 162, 425, 34 },
		{ 162, 426, 34 },
		{ 162, 427, 34 },
		{ 162, 428, 34 },
		{ 162, 429, 34 },
		{ 162, 430, 34 },
		{ 162, 431, 34 },
		{ 162, 432, 34 },
		{ 162, 433, 34 },
		{ 162, 434, 34 },
		{ 162, 435, 34 },
		{ 162, 436, 34 },
		{ 162, 437, 34 },
		{ 162, 438, 34 },
		{ 162, 439, 34 },
		{ 162, 440, 34 },
		{ 162, 441, 34 },
		{ 162, 442, 34 },
		{ 162, 443, 34 },
		{ 162, 444, 34 },
		{ 162, 445, 34 },
		{ 162, 446, 34 },
		{ 162, 447, 34 },
		{ 162, 448, 34 },
		{ 162, 449, 34 },
		{ 162, 450, 34 },
		{ 162, 451, 34 },
		{ 162, 452, 34 },
		{ 162, 453, 34 },
		{ 162, 454, 34 },
		{ 162, 455, 34 },
		{ 162, 456, 34 },
		{ 162, 457, 34 },
		{ 162, 458, 34 },
		{ 162, 459, 34 },
		{ 162, 460, 34 },
		{ 162, 461, 34 },
		{ 162, 462, 34 },
		{ 162, 463, 34 },
		{ 162, 464, 34 },
		{ 162, 465, 34 },
		{ 162, 466, 34 },
		{ 162, 467, 34 },
		{ 162, 468, 34 },
		{ 162, 469, 34 },
		{ 162, 470, 34 },
		{ 162, 471, 34 },
		{ 162, 472, 34 },
		{ 162, 473, 34 },
		{ 162, 474, 34 },
		{ 162, 475, 34 },
		{ 162, 476, 34 },
		{ 162, 477, 34 },
		{ 162, 478, 34 },
		{ 162, 479, 34 },
		{ 162, 480, 34 },
		{ 162, 481, 34 },
		{ 162, 482, 34 },
		{ 162, 483, 34 },
		{ 162, 484, 34 },
		{ 162, 485, 34 },
		{ 162, 486, 34 },
		{ 162, 487, 34 },
		{ 162, 488, 34 },
		{ 162, 489, 34 },
		{ 162, 490, 34 },
		{ 162, 491, 34 },
		{ 162, 492, 34 },
		{ 162, 493, 34 },
		{ 162, 494, 34 },
		{ 162, 495, 34 },
		{ 162, 496, 34 },
		{ 162, 497, 34 },
		{ 162, 498, 34 },
		{ 162, 499, 34 },
		{ 162, 500, 34 },
		{ 162, 501, 34 },
		{ 162, 502, 34 },
		{ 162, 503, 34 },
		{ 162, 504, 34 },
		{ 162, 505, 34 },
		{ 162, 506, 34 },
		{ 162, 507, 34 },
		{ 162, 508, 34 },
		{ 162, 509, 34 },
		{ 162, 510, 34 },
		{ 163, 408, 34 },
		{ 163, 409, 34 },
		{ 163, 410, 34 },
		{ 163, 411, 34 },
		{ 163, 412, 34 },
		{ 163, 413, 34 },
		{ 163, 414, 34 },
		{ 163, 415, 34 },
		{ 163, 416, 34 },
		{ 163, 417, 34 },
		{ 163, 418, 34 },
		{ 163, 419, 34 },
		{ 163, 420, 34 },
		{ 163, 421, 34 },
		{ 163, 422, 34 },
		{ 163, 423, 34 },
		{ 163, 424, 34 },
		{ 163, 425, 34 },
		{ 163, 426, 34 },
		{ 163, 427, 34 },
		{ 163, 428, 34 },
		{ 163, 429, 34 },
		{ 163, 430, 34 },
		{ 163, 431, 34 },
		{ 163, 432, 34 },
		{ 163, 433, 34 },
		{ 163, 434, 34 },
		{ 163, 435, 34 },
		{ 163, 436, 34 },
		{ 163, 437, 34 },
		{ 163, 438, 34 },
		{ 163, 439, 34 },
		{ 163, 440, 34 },
		{ 163, 441, 34 },
		{ 163, 442, 34 },
		{ 163, 443, 34 },
		{ 163, 444, 34 },
		{ 163, 445, 34 },
		{ 163, 446, 34 },
		{ 163, 447, 34 },
		{ 163, 448, 34 },
		{ 163, 449, 34 },
		{ 163, 450, 34 },
		{ 163, 451, 34 },
		{ 163, 452, 34 },
		{ 163, 453, 34 },
		{ 163, 454, 34 },
		{ 163, 455, 34 },
		{ 163, 456, 34 },
		{ 163, 457, 34 },
		{ 163, 458, 34 },
		{ 163, 459, 34 },
		{ 163, 460, 34 },
		{ 163, 461, 34 },
		{ 163, 462, 34 },
		{ 163, 463, 34 },
		{ 163, 464, 34 },
		{ 163, 465, 34 },
		{ 163, 466, 34 },
		{ 163, 467, 34 },
		{ 163, 468, 34 },
		{ 163, 469, 34 },
		{ 163, 470, 34 },
		{ 163, 471, 34 },
		{ 163, 472, 34 },
		{ 163, 473, 34 },
		{ 163, 474, 34 },
		{ 163, 475, 34 },
		{ 163, 476, 34 },
		{ 163, 477, 34 },
		{ 163, 478, 34 },
		{ 163, 479, 34 },
		{ 163, 480, 34 },
		{ 163, 481, 34 },
		{ 163, 482, 34 },
		{ 163, 483, 34 },
		{ 163, 484, 34 },
		{ 163, 485, 34 },
		{ 163, 486, 34 },
		{ 163, 487, 34 },
		{ 163, 488, 34 },
		{ 163, 489, 34 },
		{ 163, 490, 34 },
		{ 163, 491, 34 },
		{ 163, 492, 34 },
		{ 163, 493, 34 },
		{ 163, 494, 34 },
		{ 163, 495, 34 },
		{ 163, 496, 34 },
		{ 163, 497, 34 },
		{ 163, 498, 34 },
		{ 163, 499, 34 },
		{ 163, 500, 34 },
		{ 163, 501, 34 },
		{ 163, 502, 34 },
		{ 163, 503, 34 },
		{ 163, 504, 34 },
		{ 163, 505, 34 },
		{ 163, 506, 34 },
		{ 163, 507, 34 },
		{ 164, 406, 34 },
		{ 164, 407, 34 },
		{ 164, 408, 34 },
		{ 164, 409, 34 },
		{ 164, 410, 34 },
		{ 164, 411, 34 },
		{ 164, 412, 34 },
		{ 164, 413, 34 },
		{ 164, 414, 34 },
		{ 164, 415, 34 },
		{ 164, 416, 34 },
		{ 164, 417, 34 },
		{ 164, 418, 34 },
		{ 164, 419, 34 },
		{ 164, 420, 34 },
		{ 164, 421, 34 },
		{ 164, 422, 34 },
		{ 164, 423, 34 },
		{ 164, 424, 34 },
		{ 164, 425, 34 },
		{ 164, 426, 34 },
		{ 164, 427, 34 },
		{ 164, 428, 34 },
		{ 164, 429, 34 },
		{ 164, 430, 34 },
		{ 164, 431, 34 },
		{ 164, 432, 34 },
		{ 164, 433, 34 },
		{ 164, 434, 34 },
		{ 164, 435, 34 },
		{ 164, 436, 34 },
		{ 164, 437, 34 },
		{ 164, 438, 34 },
		{ 164, 439, 34 },
		{ 164, 440, 34 },
		{ 164, 441, 34 },
		{ 164, 442, 34 },
		{ 164, 443, 34 },
		{ 164, 444, 34 },
		{ 164, 445, 34 },
		{ 164, 446, 34 },
		{ 164, 447, 34 },
		{ 164, 448, 34 },
		{ 164, 449, 34 },
		{ 164, 450, 34 },
		{ 164, 451, 34 },
		{ 164, 452, 34 },
		{ 164, 453, 34 },
		{ 164, 454, 34 },
		{ 164, 455, 34 },
		{ 164, 456, 34 },
		{ 164, 457, 34 },
		{ 164, 458, 34 },
		{ 164, 459, 34 },
		{ 164, 460, 34 },
		{ 164, 461, 34 },
		{ 164, 462, 34 },
		{ 164, 463, 34 },
		{ 164, 464, 34 },
		{ 164, 465, 34 },
		{ 164, 466, 34 },
		{ 164, 467, 34 },
		{ 164, 468, 34 },
		{ 164, 469, 34 },
		{ 164, 470, 34 },
		{ 164, 471, 34 },
		{ 164, 472, 34 },
		{ 164, 473, 34 },
		{ 164, 474, 34 },
		{ 164, 475, 34 },
		{ 164, 476, 34 },
		{ 164, 477, 34 },
		{ 164, 478, 34 },
		{ 164, 479, 34 },
		{ 164, 480, 34 },
		{ 164, 481, 34 },
		{ 164, 482, 34 },
		{ 164, 483, 34 },
		{ 164, 484, 34 },
		{ 164, 485, 34 },
		{ 164, 486, 34 },
		{ 164, 487, 34 },
		{ 164, 488, 34 },
		{ 164, 489, 34 },
		{ 164, 490, 34 },
		{ 164, 491, 34 },
		{ 164, 492, 34 },
		{ 164, 493, 34 },
		{ 164, 494, 34 },
		{ 164, 495, 34 },
		{ 164, 496, 34 },
		{ 164, 497, 34 },
		{ 164, 498, 34 },
		{ 164, 499, 34 },
		{ 164, 500, 34 },
		{ 164, 501, 34 },
		{ 164, 502, 34 },
		{ 164, 503, 34 },
		{ 164, 504, 34 },
		{ 165, 405, 34 },
		{ 165, 406, 34 },
		{ 165, 407, 34 },
		{ 165, 408, 34 },
		{ 165, 409, 34 },
		{ 165, 410, 34 },
		{ 165, 411, 34 },
		{ 165, 412, 34 },
		{ 165, 413, 34 },
		{ 165, 414, 34 },
		{ 165, 415, 34 },
		{ 165, 416, 34 },
		{ 165, 417, 34 },
		{ 165, 418, 34 },
		{ 165, 419, 34 },
		{ 165, 420, 34 },
		{ 165, 421, 34 },
		{ 165, 422, 34 },
		{ 165, 423, 34 },
		{ 165, 424, 34 },
		{ 165, 425, 34 },
		{ 165, 426, 34 },
		{ 165, 427, 34 },
		{ 165, 428, 34 },
		{ 165, 429, 34 },
		{ 165, 430, 34 },
		{ 165, 431, 34 },
		{ 165, 432, 34 },
		{ 165, 433, 34 },
		{ 165, 434, 34 },
		{ 165, 435, 34 },
		{ 165, 436, 34 },
		{ 165, 437, 34 },
		{ 165, 438, 34 },
		{ 165, 439, 34 },
		{ 165, 440, 34 },
		{ 165, 441, 34 },
		{ 165, 442, 34 },
		{ 165, 443, 34 },
		{ 165, 444, 34 },
		{ 165, 445, 34 },
		{ 165, 446, 34 },
		{ 165, 447, 34 },
		{ 165, 448, 34 },
		{ 165, 449, 34 },
		{ 165, 450, 34 },
		{ 165, 451, 34 },
		{ 165, 452, 34 },
		{ 165, 453, 34 },
		{ 165, 454, 34 },
		{ 165, 455, 34 },
		{ 165, 456, 34 },
		{ 165, 457, 34 },
		{ 165, 458, 34 },
		{ 165, 459, 34 },
		{ 165, 460, 34 },
		{ 165, 461, 34 },
		{ 165, 462, 34 },
		{ 165, 463, 34 },
		{ 165, 464, 34 },
		{ 165, 465, 34 },
		{ 165, 466, 34 },
		{ 165, 467, 34 },
		{ 165, 468, 34 },
		{ 165, 469, 34 },
		{ 165, 470, 34 },
		{ 165, 471, 34 },
		{ 165, 472, 34 },
		{ 165, 473, 34 },
		{ 165, 474, 34 },
		{ 165, 475, 34 },
		{ 165, 476, 34 },
		{ 165, 477, 34 },
		{ 165, 478, 34 },
		{ 165, 479, 34 },
		{ 165, 480, 34 },
		{ 165, 481, 34 },
		{ 165, 482, 34 },
		{ 165, 483, 34 },
		{ 165, 484, 34 },
		{ 165, 485, 34 },
		{ 165, 486, 34 },
		{ 165, 487, 34 },
		{ 165, 488, 34 },
		{ 165, 489, 34 },
		{ 165, 490, 34 },
		{ 165, 491, 34 },
		{ 165, 492, 34 },
		{ 165, 493, 34 },
		{ 165, 494, 34 },
		{ 165, 495, 34 },
		{ 165, 496, 34 },
		{ 165, 497, 34 },
		{ 165, 498, 34 },
		{ 165, 499, 34 },
		{ 165, 500, 34 },
		{ 165, 501, 34 },
		{ 165, 502, 34 },
		{ 166, 403, 34 },
		{ 166, 404, 34 },
		{ 166, 405, 34 },
		{ 166, 406, 34 },
		{ 166, 407, 34 },
		{ 166, 408, 34 },
		{ 166, 409, 34 },
		{ 166, 410, 34 },
		{ 166, 411, 34 },
		{ 166, 412, 34 },
		{ 166, 413, 34 },
		{ 166, 414, 34 },
		{ 166, 415, 34 },
		{ 166, 416, 34 },
		{ 166, 417, 34 },
		{ 166, 418, 34 },
		{ 166, 419, 34 },
		{ 166, 420, 34 },
		{ 166, 421, 34 },
		{ 166, 422, 34 },
		{ 166, 423, 34 },
		{ 166, 424, 34 },
		{ 166, 425, 34 },
		{ 166, 426, 34 },
		{ 166, 427, 34 },
		{ 166, 428, 34 },
		{ 166, 429, 34 },
		{ 166, 430, 34 },
		{ 166, 431, 34 },
		{ 166, 432, 34 },
		{ 166, 433, 34 },
		{ 166, 434, 34 },
		{ 166, 435, 34 },
		{ 166, 436, 34 },
		{ 166, 437, 34 },
		{ 166, 438, 34 },
		{ 166, 439, 34 },
		{ 166, 440, 34 },
		{ 166, 441, 34 },
		{ 166, 442, 34 },
		{ 166, 443, 34 },
		{ 166, 444, 34 },
		{ 166, 445, 34 },
		{ 166, 446, 34 },
		{ 166, 447, 34 },
		{ 166, 448, 34 },
		{ 166, 449, 34 },
		{ 166, 450, 34 },
		{ 166, 451, 34 },
		{ 166, 452, 34 },
		{ 166, 453, 34 },
		{ 166, 454, 34 },
		{ 166, 455, 34 },
		{ 166, 456, 34 },
		{ 166, 457, 34 },
		{ 166, 458, 34 },
		{ 166, 459, 34 },
		{ 166, 460, 34 },
		{ 166, 461, 34 },
		{ 166, 462, 34 },
		{ 166, 463, 34 },
		{ 166, 464, 34 },
		{ 166, 465, 34 },
		{ 166, 466, 34 },
		{ 166, 467, 34 },
		{ 166, 468, 34 },
		{ 166, 469, 34 },
		{ 166, 470, 34 },
		{ 166, 471, 34 },
		{ 166, 472, 34 },
		{ 166, 473, 34 },
		{ 166, 474, 34 },
		{ 166, 475, 34 },
		{ 166, 476, 34 },
		{ 166, 477, 34 },
		{ 166, 478, 34 },
		{ 166, 479, 34 },
		{ 166, 480, 34 },
		{ 166, 481, 34 },
		{ 166, 482, 34 },
		{ 166, 483, 34 },
		{ 166, 484, 34 },
		{ 166, 485, 34 },
		{ 166, 486, 34 },
		{ 166, 487, 34 },
		{ 166, 488, 34 },
		{ 166, 489, 34 },
		{ 166, 490, 34 },
		{ 166, 491, 34 },
		{ 166, 492, 34 },
		{ 166, 493, 34 },
		{ 166, 494, 34 },
		{ 166, 495, 34 },
		{ 166, 496, 34 },
		{ 166, 497, 34 },
		{ 166, 498, 34 },
		{ 166, 499, 34 },
		{ 167, 402, 34 },
		{ 167, 403, 34 },
		{ 167, 404, 34 },
		{ 167, 405, 34 },
		{ 167, 406, 34 },
		{ 167, 407, 34 },
		{ 167, 408, 34 },
		{ 167, 409, 34 },
		{ 167, 410, 34 },
		{ 167, 411, 34 },
		{ 167, 412, 34 },
		{ 167, 413, 34 },
		{ 167, 414, 34 },
		{ 167, 415, 34 },
		{ 167, 416, 34 },
		{ 167, 417, 34 },
		{ 167, 418, 34 },
		{ 167, 419, 34 },
		{ 167, 420, 34 },
		{ 167, 421, 34 },
		{ 167, 422, 34 },
		{ 167, 423, 34 },
		{ 167, 424, 34 },
		{ 167, 425, 34 },
		{ 167, 426, 34 },
		{ 167, 427, 34 },
		{ 167, 428, 34 },
		{ 167, 429, 34 },
		{ 167, 430, 34 },
		{ 167, 431, 34 },
		{ 167, 432, 34 },
		{ 167, 433, 34 },
		{ 167, 434, 34 },
		{ 167, 435, 34 },
		{ 167, 436, 34 },
		{ 167, 437, 34 },
		{ 167, 438, 34 },
		{ 167, 439, 34 },
		{ 167, 440, 34 },
		{ 167, 441, 34 },
		{ 167, 442, 34 },
		{ 167, 443, 34 },
		{ 167, 444, 34 },
		{ 167, 445, 34 },
		{ 167, 446, 34 },
		{ 167, 447, 34 },
		{ 167, 448, 34 },
		{ 167, 449, 34 },
		{ 167, 450, 34 },
		{ 167, 451, 34 },
		{ 167, 452, 34 },
		{ 167, 453, 34 },
		{ 167, 454, 34 },
		{ 167, 455, 34 },
		{ 167, 456, 34 },
		{ 167, 457, 34 },
		{ 167, 458, 34 },
		{ 167, 459, 34 },
		{ 167, 460, 34 },
		{ 167, 461, 34 },
		{ 167, 462, 34 },
		{ 167, 463, 34 },
		{ 167, 464, 34 },
		{ 167, 465, 34 },
		{ 167, 466, 34 },
		{ 167, 467, 34 },
		{ 167, 468, 34 },
		{ 167, 469, 34 },
		{ 167, 470, 34 },
		{ 167, 471, 34 },
		{ 167, 472, 34 },
		{ 167, 473, 34 },
		{ 167, 474, 34 },
		{ 167, 475, 34 },
		{ 167, 476, 34 },
		{ 167, 477, 34 },
		{ 167, 478, 34 },
		{ 167, 479, 34 },
		{ 167, 480, 34 },
		{ 167, 481, 34 },
		{ 167, 482, 34 },
		{ 167, 483, 34 },
		{ 167, 484, 34 },
		{ 167, 485, 34 },
		{ 167, 486, 34 },
		{ 167, 487, 34 },
		{ 167, 488, 34 },
		{ 167, 489, 34 },
		{ 167, 490, 34 },
		{ 167, 491, 34 },
		{ 167, 492, 34 },
		{ 167, 493, 34 },
		{ 167, 494, 34 },
		{ 167, 495, 34 },
		{ 167, 496, 34 },
		{ 168, 400, 34 },
		{ 168, 401, 34 },
		{ 168, 402, 34 },
		{ 168, 403, 34 },
		{ 168, 404, 34 },
		{ 168, 405, 34 },
		{ 168, 406, 34 },
		{ 168, 407, 34 },
		{ 168, 408, 34 },
		{ 168, 409, 34 },
		{ 168, 410, 34 },
		{ 168, 411, 34 },
		{ 168, 412, 34 },
		{ 168, 413, 34 },
		{ 168, 414, 34 },
		{ 168, 415, 34 },
		{ 168, 416, 34 },
		{ 168, 417, 34 },
		{ 168, 418, 34 },
		{ 168, 419, 34 },
		{ 168, 420, 34 },
		{ 168, 421, 34 },
		{ 168, 422, 34 },
		{ 168, 423, 34 },
		{ 168, 424, 34 },
		{ 168, 425, 34 },
		{ 168, 426, 34 },
		{ 168, 427, 34 },
		{ 168, 428, 34 },
		{ 168, 429, 34 },
		{ 168, 430, 34 },
		{ 168, 431, 34 },
		{ 168, 432, 34 },
		{ 168, 433, 34 },
		{ 168, 434, 34 },
		{ 168, 435, 34 },
		{ 168, 436, 34 },
		{ 168, 437, 34 },
		{ 168, 438, 34 },
		{ 168, 439, 34 },
		{ 168, 440, 34 },
		{ 168, 441, 34 },
		{ 168, 442, 34 },
		{ 168, 443, 34 },
		{ 168, 444, 34 },
		{ 168, 445, 34 },
		{ 168, 446, 34 },
		{ 168, 447, 34 },
		{ 168, 448, 34 },
		{ 168, 449, 34 },
		{ 168, 450, 34 },
		{ 168, 451, 34 },
		{ 168, 452, 34 },
		{ 168, 453, 34 },
		{ 168, 454, 34 },
		{ 168, 455, 34 },
		{ 168, 456, 34 },
		{ 168, 457, 34 },
		{ 168, 458, 34 },
		{ 168, 459, 34 },
		{ 168, 460, 34 },
		{ 168, 461, 34 },
		{ 168, 462, 34 },
		{ 168, 463, 34 },
		{ 168, 464, 34 },
		{ 168, 465, 34 },
		{ 168, 466, 34 },
		{ 168, 467, 34 },
		{ 168, 468, 34 },
		{ 168, 469, 34 },
		{ 168, 470, 34 },
		{ 168, 471, 34 },
		{ 168, 472, 34 },
		{ 168, 473, 34 },
		{ 168, 474, 34 },
		{ 168, 475, 34 },
		{ 168, 476, 34 },
		{ 168, 477, 34 },
		{ 168, 478, 34 },
		{ 168, 479, 34 },
		{ 168, 480, 34 },
		{ 168, 481, 34 },
		{ 168, 482, 34 },
		{ 168, 483, 34 },
		{ 168, 484, 34 },
		{ 168, 485, 34 },
		{ 168, 486, 34 },
		{ 168, 487, 34 },
		{ 168, 488, 34 },
		{ 168, 489, 34 },
		{ 168, 490, 34 },
		{ 168, 491, 34 },
		{ 168, 492, 34 },
		{ 168, 493, 34 },
		{ 168, 494, 34 },
		{ 169, 398, 34 },
		{ 169, 399, 34 },
		{ 169, 400, 34 },
		{ 169, 401, 34 },
		{ 169, 402, 34 },
		{ 169, 403, 34 },
		{ 169, 404, 34 },
		{ 169, 405, 34 },
		{ 169, 406, 34 },
		{ 169, 407, 34 },
		{ 169, 408, 34 },
		{ 169, 409, 34 },
		{ 169, 410, 34 },
		{ 169, 411, 34 },
		{ 169, 412, 34 },
		{ 169, 413, 34 },
		{ 169, 414, 34 },
		{ 169, 415, 34 },
		{ 169, 416, 34 },
		{ 169, 417, 34 },
		{ 169, 418, 34 },
		{ 169, 419, 34 },
		{ 169, 420, 34 },
		{ 169, 421, 34 },
		{ 169, 422, 34 },
		{ 169, 423, 34 },
		{ 169, 424, 34 },
		{ 169, 425, 34 },
		{ 169, 426, 34 },
		{ 169, 427, 34 },
		{ 169, 428, 34 },
		{ 169, 429, 34 },
		{ 169, 430, 34 },
		{ 169, 431, 34 },
		{ 169, 432, 34 },
		{ 169, 433, 34 },
		{ 169, 434, 34 },
		{ 169, 435, 34 },
		{ 169, 436, 34 },
		{ 169, 437, 34 },
		{ 169, 438, 34 },
		{ 169, 439, 34 },
		{ 169, 440, 34 },
		{ 169, 441, 34 },
		{ 169, 442, 34 },
		{ 169, 443, 34 },
		{ 169, 444, 34 },
		{ 169, 445, 34 },
		{ 169, 446, 34 },
		{ 169, 447, 34 },
		{ 169, 448, 34 },
		{ 169, 449, 34 },
		{ 169, 450, 34 },
		{ 169, 451, 34 },
		{ 169, 452, 34 },
		{ 169, 453, 34 },
		{ 169, 454, 34 },
		{ 169, 455, 34 },
		{ 169, 456, 34 },
		{ 169, 457, 34 },
		{ 169, 458, 34 },
		{ 169, 459, 34 },
		{ 169, 460, 34 },
		{ 169, 461, 34 },
		{ 169, 462, 34 },
		{ 169, 463, 34 },
		{ 169, 464, 34 },
		{ 169, 465, 34 },
		{ 169, 466, 34 },
		{ 169, 467, 34 },
		{ 169, 468, 34 },
		{ 169, 469, 34 },
		{ 169, 470, 34 },
		{ 169, 471, 34 },
		{ 169, 472, 34 },
		{ 169, 473, 34 },
		{ 169, 474, 34 },
		{ 169, 475, 34 },
		{ 169, 476, 34 },
		{ 169, 477, 34 },
		{ 169, 478, 34 },
		{ 169, 479, 34 },
		{ 169, 480, 34 },
		{ 169, 481, 34 },
		{ 169, 482, 34 },
		{ 169, 483, 34 },
		{ 169, 484, 34 },
		{ 169, 485, 34 },
		{ 169, 486, 34 },
		{ 169, 487, 34 },
		{ 169, 488, 34 },
		{ 169, 489, 34 },
		{ 169, 490, 34 },
		{ 169, 491, 34 },
		{ 170, 397, 34 },
		{ 170, 398, 34 },
		{ 170, 399, 34 },
		{ 170, 400, 34 },
		{ 170, 401, 34 },
		{ 170, 402, 34 },
		{ 170, 403, 34 },
		{ 170, 404, 34 },
		{ 170, 405, 34 },
		{ 170, 406, 34 },
		{ 170, 407, 34 },
		{ 170, 408, 34 },
		{ 170, 409, 34 },
		{ 170, 410, 34 },
		{ 170, 411, 34 },
		{ 170, 412, 34 },
		{ 170, 413, 34 },
		{ 170, 414, 34 },
		{ 170, 415, 34 },
		{ 170, 416, 34 },
		{ 170, 417, 34 },
		{ 170, 418, 34 },
		{ 170, 419, 34 },
		{ 170, 420, 34 },
		{ 170, 421, 34 },
		{ 170, 422, 34 },
		{ 170, 423, 34 },
		{ 170, 424, 34 },
		{ 170, 425, 34 },
		{ 170, 426, 34 },
		{ 170, 427, 34 },
		{ 170, 428, 34 },
		{ 170, 429, 34 },
		{ 170, 430, 34 },
		{ 170, 431, 34 },
		{ 170, 432, 34 },
		{ 170, 433, 34 },
		{ 170, 434, 34 },
		{ 170, 435, 34 },
		{ 170, 436, 34 },
		{ 170, 437, 34 },
		{ 170, 438, 34 },
		{ 170, 439, 34 },
		{ 170, 440, 34 },
		{ 170, 441, 34 },
		{ 170, 442, 34 },
		{ 170, 443, 34 },
		{ 170, 444, 34 },
		{ 170, 445, 34 },
		{ 170, 446, 34 },
		{ 170, 447, 34 },
		{ 170, 448, 34 },
		{ 170, 449, 34 },
		{ 170, 450, 34 },
		{ 170, 451, 34 },
		{ 170, 452, 34 },
		{ 170, 453, 34 },
		{ 170, 454, 34 },
		{ 170, 455, 34 },
		{ 170, 456, 34 },
		{ 170, 457, 34 },
		{ 170, 458, 34 },
		{ 170, 459, 34 },
		{ 170, 460, 34 },
		{ 170, 461, 34 },
		{ 170, 462, 34 },
		{ 170, 463, 34 },
		{ 170, 464, 34 },
		{ 170, 465, 34 },
		{ 170, 466, 34 },
		{ 170, 467, 34 },
		{ 170, 468, 34 },
		{ 170, 469, 34 },
		{ 170, 470, 34 },
		{ 170, 471, 34 },
		{ 170, 472, 34 },
		{ 170, 473, 34 },
		{ 170, 474, 34 },
		{ 170, 475, 34 },
		{ 170, 476, 34 },
		{ 170, 477, 34 },
		{ 170, 478, 34 },
		{ 170, 479, 34 },
		{ 170, 480, 34 },
		{ 170, 481, 34 },
		{ 170, 482, 34 },
		{ 170, 483, 34 },
		{ 170, 484, 34 },
		{ 170, 485, 34 },
		{ 170, 486, 34 },
		{ 170, 487, 34 },
		{ 170, 488, 34 },
		{ 171, 395, 34 },
		{ 171, 396, 34 },
		{ 171, 397, 34 },
		{ 171, 398, 34 },
		{ 171, 399, 34 },
		{ 171, 400, 34 },
		{ 171, 401, 34 },
		{ 171, 402, 34 },
		{ 171, 403, 34 },
		{ 171, 404, 34 },
		{ 171, 405, 34 },
		{ 171, 406, 34 },
		{ 171, 407, 34 },
		{ 171, 408, 34 },
		{ 171, 409, 34 },
		{ 171, 410, 34 },
		{ 171, 411, 34 },
		{ 171, 412, 34 },
		{ 171, 413, 34 },
		{ 171, 414, 34 },
		{ 171, 415, 34 },
		{ 171, 416, 34 },
		{ 171, 417, 34 },
		{ 171, 418, 34 },
		{ 171, 419, 34 },
		{ 171, 420, 34 },
		{ 171, 421, 34 },
		{ 171, 422, 34 },
		{ 171, 423, 34 },
		{ 171, 424, 34 },
		{ 171, 425, 34 },
		{ 171, 426, 34 },
		{ 171, 427, 34 },
		{ 171, 428, 34 },
		{ 171, 429, 34 },
		{ 171, 430, 34 },
		{ 171, 431, 34 },
		{ 171, 432, 34 },
		{ 171, 433, 34 },
		{ 171, 434, 34 },
		{ 171, 435, 34 },
		{ 171, 436, 34 },
		{ 171, 437, 34 },
		{ 171, 438, 34 },
		{ 171, 439, 34 },
		{ 171, 440, 34 },
		{ 171, 441, 34 },
		{ 171, 442, 34 },
		{ 171, 443, 34 },
		{ 171, 444, 34 },
		{ 171, 445, 34 },
		{ 171, 446, 34 },
		{ 171, 447, 34 },
		{ 171, 448, 34 },
		{ 171, 449, 34 },
		{ 171, 450, 34 },
		{ 171, 451, 34 },
		{ 171, 452, 34 },
		{ 171, 453, 34 },
		{ 171, 454, 34 },
		{ 171, 455, 34 },
		{ 171, 456, 34 },
		{ 171, 457, 34 },
		{ 171, 458, 34 },
		{ 171, 459, 34 },
		{ 171, 460, 34 },
		{ 171, 461, 34 },
		{ 171, 462, 34 },
		{ 171, 463, 34 },
		{ 171, 464, 34 },
		{ 171, 465, 34 },
		{ 171, 466, 34 },
		{ 171, 467, 34 },
		{ 171, 468, 34 },
		{ 171, 469, 34 },
		{ 171, 470, 34 },
		{ 171, 471, 34 },
		{ 171, 472, 34 },
		{ 171, 473, 34 },
		{ 171, 474, 34 },
		{ 171, 475, 34 },
		{ 171, 476, 34 },
		{ 171, 477, 34 },
		{ 171, 478, 34 },
		{ 171, 479, 34 },
		{ 171, 480, 34 },
		{ 171, 481, 34 },
		{ 171, 482, 34 },
		{ 171, 483, 34 },
		{ 171, 484, 34 },
		{ 171, 485, 34 },
		{ 171, 486, 34 },
		{ 172, 394, 34 },
		{ 172, 395, 34 },
		{ 172, 396, 34 },
		{ 172, 397, 34 },
		{ 172, 398, 34 },
		{ 172, 399, 34 },
		{ 172, 400, 34 },
		{ 172, 401, 34 },
		{ 172, 402, 34 },
		{ 172, 403, 34 },
		{ 172, 404, 34 },
		{ 172, 405, 34 },
		{ 172, 406, 34 },
		{ 172, 407, 34 },
		{ 172, 408, 34 },
		{ 172, 409, 34 },
		{ 172, 410, 34 },
		{ 172, 411, 34 },
		{ 172, 412, 34 },
		{ 172, 413, 34 },
		{ 172, 414, 34 },
		{ 172, 415, 34 },
		{ 172, 416, 34 },
		{ 172, 417, 34 },
		{ 172, 418, 34 },
		{ 172, 419, 34 },
		{ 172, 420, 34 },
		{ 172, 421, 34 },
		{ 172, 422, 34 },
		{ 172, 423, 34 },
		{ 172, 424, 34 },
		{ 172, 425, 34 },
		{ 172, 426, 34 },
		{ 172, 427, 34 },
		{ 172, 428, 34 },
		{ 172, 429, 34 },
		{ 172, 430, 34 },
		{ 172, 431, 34 },
		{ 172, 432, 34 },
		{ 172, 433, 34 },
		{ 172, 434, 34 },
		{ 172, 435, 34 },
		{ 172, 436, 34 },
		{ 172, 437, 34 },
		{ 172, 438, 34 },
		{ 172, 439, 34 },
		{ 172, 440, 34 },
		{ 172, 441, 34 },
		{ 172, 442, 34 },
		{ 172, 443, 34 },
		{ 172, 444, 34 },
		{ 172, 445, 34 },
		{ 172, 446, 34 },
		{ 172, 447, 34 },
		{ 172, 448, 34 },
		{ 172, 449, 34 },
		{ 172, 450, 34 },
		{ 172, 451, 34 },
		{ 172, 452, 34 },
		{ 172, 453, 34 },
		{ 172, 454, 34 },
		{ 172, 455, 34 },
		{ 172, 456, 34 },
		{ 172, 457, 34 },
		{ 172, 458, 34 },
		{ 172, 459, 34 },
		{ 172, 460, 34 },
		{ 172, 461, 34 },
		{ 172, 462, 34 },
		{ 172, 463, 34 },
		{ 172, 464, 34 },
		{ 172, 465, 34 },
		{ 172, 466, 34 },
		{ 172, 467, 34 },
		{ 172, 468, 34 },
		{ 172, 469, 34 },
		{ 172, 470, 34 },
		{ 172, 471, 34 },
		{ 172, 472, 34 },
		{ 172, 473, 34 },
		{ 172, 474, 34 },
		{ 172, 475, 34 },
		{ 172, 476, 34 },
		{ 172, 477, 34 },
		{ 172, 478, 34 },
		{ 172, 479, 34 },
		{ 172, 480, 34 },
		{ 172, 481, 34 },
		{ 172, 482, 34 },
		{ 172, 483, 34 },
		{ 173, 392, 34 },
		{ 173, 393, 34 },
		{ 173, 394, 34 },
		{ 173, 395, 34 },
		{ 173, 396, 34 },
		{ 173, 397, 34 },
		{ 173, 398, 34 },
		{ 173, 399, 34 },
		{ 173, 400, 34 },
		{ 173, 401, 34 },
		{ 173, 402, 34 },
		{ 173, 403, 34 },
		{ 173, 404, 34 },
		{ 173, 405, 34 },
		{ 173, 406, 34 },
		{ 173, 407, 34 },
		{ 173, 408, 34 },
		{ 173, 409, 34 },
		{ 173, 410, 34 },
		{ 173, 411, 34 },
		{ 173, 412, 34 },
		{ 173, 413, 34 },
		{ 173, 414, 34 },
		{ 173, 415, 34 },
		{ 173, 416, 34 },
		{ 173, 417, 34 },
		{ 173, 418, 34 },
		{ 173, 419, 34 },
		{ 173, 420, 34 },
		{ 173, 421, 34 },
		{ 173, 422, 34 },
		{ 173, 423, 34 },
		{ 173, 424, 34 },
		{ 173, 425, 34 },
		{ 173, 426, 34 },
		{ 173, 427, 34 },
		{ 173, 428, 34 },
		{ 173, 429, 34 },
		{ 173, 430, 34 },
		{ 173, 431, 34 },
		{ 173, 432, 34 },
		{ 173, 433, 34 },
		{ 173, 434, 34 },
		{ 173, 435, 34 },
		{ 173, 436, 34 },
		{ 173, 437, 34 },
		{ 173, 438, 34 },
		{ 173, 439, 34 },
		{ 173, 440, 34 },
		{ 173, 441, 34 },
		{ 173, 442, 34 },
		{ 173, 443, 34 },
		{ 173, 444, 34 },
		{ 173, 445, 34 },
		{ 173, 446, 34 },
		{ 173, 447, 34 },
		{ 173, 448, 34 },
		{ 173, 449, 34 },
		{ 173, 450, 34 },
		{ 173, 451, 34 },
		{ 173, 452, 34 },
		{ 173, 453, 34 },
		{ 173, 454, 34 },
		{ 173, 455, 34 },
		{ 173, 456, 34 },
		{ 173, 457, 34 },
		{ 173, 458, 34 },
		{ 173, 459, 34 },
		{ 173, 460, 34 },
		{ 173, 461, 34 },
		{ 173, 462, 34 },
		{ 173, 463, 34 },
		{ 173, 464, 34 },
		{ 173, 465, 34 },
		{ 173, 466, 34 },
		{ 173, 467, 34 },
		{ 173, 468, 34 },
		{ 173, 469, 34 },
		{ 173, 470, 34 },
		{ 173, 471, 34 },
		{ 173, 472, 34 },
		{ 173, 473, 34 },
		{ 173, 474, 34 },
		{ 173, 475, 34 },
		{ 173, 476, 34 },
		{ 173, 477, 34 },
		{ 173, 478, 34 },
		{ 173, 479, 34 },
		{ 173, 480, 34 },
		{ 174, 390, 34 },
		{ 174, 391, 34 },
		{ 174, 392, 34 },
		{ 174, 393, 34 },
		{ 174, 394, 34 },
		{ 174, 395, 34 },
		{ 174, 396, 34 },
		{ 174, 397, 34 },
		{ 174, 398, 34 },
		{ 174, 399, 34 },
		{ 174, 400, 34 },
		{ 174, 401, 34 },
		{ 174, 402, 34 },
		{ 174, 403, 34 },
		{ 174, 404, 34 },
		{ 174, 405, 34 },
		{ 174, 406, 34 },
		{ 174, 407, 34 },
		{ 174, 408, 34 },
		{ 174, 409, 34 },
		{ 174, 410, 34 },
		{ 174, 411, 34 },
		{ 174, 412, 34 },
		{ 174, 413, 34 },
		{ 174, 414, 34 },
		{ 174, 415, 34 },
		{ 174, 416, 34 },
		{ 174, 417, 34 },
		{ 174, 418, 34 },
		{ 174, 419, 34 },
		{ 174, 420, 34 },
		{ 174, 421, 34 },
		{ 174, 422, 34 },
		{ 174, 423, 34 },
		{ 174, 424, 34 },
		{ 174, 425, 34 },
		{ 174, 426, 34 },
		{ 174, 427, 34 },
		{ 174, 428, 34 },
		{ 174, 429, 34 },
		{ 174, 430, 34 },
		{ 174, 431, 34 },
		{ 174, 432, 34 },
		{ 174, 433, 34 },
		{ 174, 434, 34 },
		{ 174, 435, 34 },
		{ 174, 436, 34 },
		{ 174, 437, 34 },
		{ 174, 438, 34 },
		{ 174, 439, 34 },
		{ 174, 440, 34 },
		{ 174, 441, 34 },
		{ 174, 442, 34 },
		{ 174, 443, 34 },
		{ 174, 444, 34 },
		{ 174, 445, 34 },
		{ 174, 446, 34 },
		{ 174, 447, 34 },
		{ 174, 448, 34 },
		{ 174, 449, 34 },
		{ 174, 450, 34 },
		{ 174, 451, 34 },
		{ 174, 452, 34 },
		{ 174, 453, 34 },
		{ 174, 454, 34 },
		{ 174, 455, 34 },
		{ 174, 456, 34 },
		{ 174, 457, 34 },
		{ 174, 458, 34 },
		{ 174, 459, 34 },
		{ 174, 460, 34 },
		{ 174, 461, 34 },
		{ 174, 462, 34 },
		{ 174, 463, 34 },
		{ 174, 464, 34 },
		{ 174, 465, 34 },
		{ 174, 466, 34 },
		{ 174, 467, 34 },
		{ 174, 468, 34 },
		{ 174, 469, 34 },
		{ 174, 470, 34 },
		{ 174, 471, 34 },
		{ 174, 472, 34 },
		{ 174, 473, 34 },
		{ 174, 474, 34 },
		{ 174, 475, 34 },
		{ 174, 476, 34 },
		{ 174, 477, 34 },
		{ 174, 478, 34 },
		{ 175, 389, 34 },
		{ 175, 390, 34 },
		{ 175, 391, 34 },
		{ 175, 392, 34 },
		{ 175, 393, 34 },
		{ 175, 394, 34 },
		{ 175, 395, 34 },
		{ 175, 396, 34 },
		{ 175, 397, 34 },
		{ 175, 398, 34 },
		{ 175, 399, 34 },
		{ 175, 400, 34 },
		{ 175, 401, 34 },
		{ 175, 402, 34 },
		{ 175, 403, 34 },
		{ 175, 404, 34 },
		{ 175, 405, 34 },
		{ 175, 406, 34 },
		{ 175, 407, 34 },
		{ 175, 408, 34 },
		{ 175, 409, 34 },
		{ 175, 410, 34 },
		{ 175, 411, 34 },
		{ 175, 412, 34 },
		{ 175, 413, 34 },
		{ 175, 414, 34 },
		{ 175, 415, 34 },
		{ 175, 416, 34 },
		{ 175, 417, 34 },
		{ 175, 418, 34 },
		{ 175, 419, 34 },
		{ 175, 420, 34 },
		{ 175, 421, 34 },
		{ 175, 422, 34 },
		{ 175, 423, 34 },
		{ 175, 424, 34 },
		{ 175, 425, 34 },
		{ 175, 426, 34 },
		{ 175, 427, 34 },
		{ 175, 428, 34 },
		{ 175, 429, 34 },
		{ 175, 430, 34 },
		{ 175, 431, 34 },
		{ 175, 432, 34 },
		{ 175, 433, 34 },
		{ 175, 434, 34 },
		{ 175, 435, 34 },
		{ 175, 436, 34 },
		{ 175, 437, 34 },
		{ 175, 438, 34 },
		{ 175, 439, 34 },
		{ 175, 440, 34 },
		{ 175, 441, 34 },
		{ 175, 442, 34 },
		{ 175, 443, 34 },
		{ 175, 444, 34 },
		{ 175, 445, 34 },
		{ 175, 446, 34 },
		{ 175, 447, 34 },
		{ 175, 448, 34 },
		{ 175, 449, 34 },
		{ 175, 450, 34 },
		{ 175, 451, 34 },
		{ 175, 452, 34 },
		{ 175, 453, 34 },
		{ 175, 454, 34 },
		{ 175, 455, 34 },
		{ 175, 456, 34 },
		{ 175, 457, 34 },
		{ 175, 458, 34 },
		{ 175, 459, 34 },
		{ 175, 460, 34 },
		{ 175, 461, 34 },
		{ 175, 462, 34 },
		{ 175, 463, 34 },
		{ 175, 464, 34 },
		{ 175, 465, 34 },
		{ 175, 466, 34 },
		{ 175, 467, 34 },
		{ 175, 468, 34 },
		{ 175, 469, 34 },
		{ 175, 470, 34 },
		{ 175, 471, 34 },
		{ 175, 472, 34 },
		{ 175, 473, 34 },
		{ 175, 474, 34 },
		{ 175, 475, 34 },
		{ 176, 387, 34 },
		{ 176, 388, 34 },
		{ 176, 389, 34 },
		{ 176, 390, 34 },
		{ 176, 391, 34 },
		{ 176, 392, 34 },
		{ 176, 393, 34 },
		{ 176, 394, 34 },
		{ 176, 395, 34 },
		{ 176, 396, 34 },
		{ 176, 397, 34 },
		{ 176, 398, 34 },
		{ 176, 399, 34 },
		{ 176, 400, 34 },
		{ 176, 401, 34 },
		{ 176, 402, 34 },
		{ 176, 403, 34 },
		{ 176, 404, 34 },
		{ 176, 405, 34 },
		{ 176, 406, 34 },
		{ 176, 407, 34 },
		{ 176, 408, 34 },
		{ 176, 409, 34 },
		{ 176, 410, 34 },
		{ 176, 411, 34 },
		{ 176, 412, 34 },
		{ 176, 413, 34 },
		{ 176, 414, 34 },
		{ 176, 415, 34 },
		{ 176, 416, 34 },
		{ 176, 417, 34 },
		{ 176, 418, 34 },
		{ 176, 419, 34 },
		{ 176, 420, 34 },
		{ 176, 421, 34 },
		{ 176, 422, 34 },
		{ 176, 423, 34 },
		{ 176, 424, 34 },
		{ 176, 425, 34 },
		{ 176, 426, 34 },
		{ 176, 427, 34 },
		{ 176, 428, 34 },
		{ 176, 429, 34 },
		{ 176, 430, 34 },
		{ 176, 431, 34 },
		{ 176, 432, 34 },
		{ 176, 433, 34 },
		{ 176, 434, 34 },
		{ 176, 435, 34 },
		{ 176, 436, 34 },
		{ 176, 437, 34 },
		{ 176, 438, 34 },
		{ 176, 439, 34 },
		{ 176, 440, 34 },
		{ 176, 441, 34 },
		{ 176, 442, 34 },
		{ 176, 443, 34 },
		{ 176, 444, 34 },
		{ 176, 445, 34 },
		{ 176, 446, 34 },
		{ 176, 447, 34 },
		{ 176, 448, 34 },
		{ 176, 449, 34 },
		{ 176, 450, 34 },
		{ 176, 451, 34 },
		{ 176, 452, 34 },
		{ 176, 453, 34 },
		{ 176, 454, 34 },
		{ 176, 455, 34 },
		{ 176, 456, 34 },
		{ 176, 457, 34 },
		{ 176, 458, 34 },
		{ 176, 459, 34 },
		{ 176, 460, 34 },
		{ 176, 461, 34 },
		{ 176, 462, 34 },
		{ 176, 463, 34 },
		{ 176, 464, 34 },
		{ 176, 465, 34 },
		{ 176, 466, 34 },
		{ 176, 467, 34 },
		{ 176, 468, 34 },
		{ 176, 469, 34 },
		{ 176, 470, 34 },
		{ 176, 471, 34 },
		{ 176, 472, 34 },
		{ 177, 386, 34 },
		{ 177, 387, 34 },
		{ 177, 388, 34 },
		{ 177, 389, 34 },
		{ 177, 390, 34 },
		{ 177, 391, 34 },
		{ 177, 392, 34 },
		{ 177, 393, 34 },
		{ 177, 394, 34 },
		{ 177, 395, 34 },
		{ 177, 396, 34 },
		{ 177, 397, 34 },
		{ 177, 398, 34 },
		{ 177, 399, 34 },
		{ 177, 400, 34 },
		{ 177, 401, 34 },
		{ 177, 402, 34 },
		{ 177, 403, 34 },
		{ 177, 404, 34 },
		{ 177, 405, 34 },
		{ 177, 406, 34 },
		{ 177, 407, 34 },
		{ 177, 408, 34 },
		{ 177, 409, 34 },
		{ 177, 410, 34 },
		{ 177, 411, 34 },
		{ 177, 412, 34 },
		{ 177, 413, 34 },
		{ 177, 414, 34 },
		{ 177, 415, 34 },
		{ 177, 416, 34 },
		{ 177, 417, 34 },
		{ 177, 418, 34 },
		{ 177, 419, 34 },
		{ 177, 420, 34 },
		{ 177, 421, 34 },
		{ 177, 422, 34 },
		{ 177, 423, 34 },
		{ 177, 424, 34 },
		{ 177, 425, 34 },
		{ 177, 426, 34 },
		{ 177, 427, 34 },
		{ 177, 428, 34 },
		{ 177, 429, 34 },
		{ 177, 430, 34 },
		{ 177, 431, 34 },
		{ 177, 432, 34 },
		{ 177, 433, 34 },
		{ 177, 434, 34 },
		{ 177, 435, 34 },
		{ 177, 436, 34 },
		{ 177, 437, 34 },
		{ 177, 438, 34 },
		{ 177, 439, 34 },
		{ 177, 440, 34 },
		{ 177, 441, 34 },
		{ 177, 442, 34 },
		{ 177, 443, 34 },
		{ 177, 444, 34 },
		{ 177, 445, 34 },
		{ 177, 446, 34 },
		{ 177, 447, 34 },
		{ 177, 448, 34 },
		{ 177, 449, 34 },
		{ 177, 450, 34 },
		{ 177, 451, 34 },
		{ 177, 452, 34 },
		{ 177, 453, 34 },
		{ 177, 454, 34 },
		{ 177, 455, 34 },
		{ 177, 456, 34 },
		{ 177, 457, 34 },
		{ 177, 458, 34 },
		{ 177, 459, 34 },
		{ 177, 460, 34 },
		{ 177, 461, 34 },
		{ 177, 462, 34 },
		{ 177, 463, 34 },
		{ 177, 464, 34 },
		{ 177, 465, 34 },
		{ 177, 466, 34 },
		{ 177, 467, 34 },
		{ 177, 468, 34 },
		{ 177, 469, 34 },
		{ 177, 470, 34 },
		{ 178, 384, 34 },
		{ 178, 385, 34 },
		{ 178, 386, 34 },
		{ 178, 387, 34 },
		{ 178, 388, 34 },
		{ 178, 389, 34 },
		{ 178, 390, 34 },
		{ 178, 391, 34 },
		{ 178, 392, 34 },
		{ 178, 393, 34 },
		{ 178, 394, 34 },
		{ 178, 395, 34 },
		{ 178, 396, 34 },
		{ 178, 397, 34 },
		{ 178, 398, 34 },
		{ 178, 399, 34 },
		{ 178, 400, 34 },
		{ 178, 401, 34 },
		{ 178, 402, 34 },
		{ 178, 403, 34 },
		{ 178, 404, 34 },
		{ 178, 405, 34 },
		{ 178, 406, 34 },
		{ 178, 407, 34 },
		{ 178, 408, 34 },
		{ 178, 409, 34 },
		{ 178, 410, 34 },
		{ 178, 411, 34 },
		{ 178, 412, 34 },
		{ 178, 413, 34 },
		{ 178, 414, 34 },
		{ 178, 415, 34 },
		{ 178, 416, 34 },
		{ 178, 417, 34 },
		{ 178, 418, 34 },
		{ 178, 419, 34 },
		{ 178, 420, 34 },
		{ 178, 421, 34 },
		{ 178, 422, 34 },
		{ 178, 423, 34 },
		{ 178, 424, 34 },
		{ 178, 425, 34 },
		{ 178, 426, 34 },
		{ 178, 427, 34 },
		{ 178, 428, 34 },
		{ 178, 429, 34 },
		{ 178, 430, 34 },
		{ 178, 431, 34 },
		{ 178, 432, 34 },
		{ 178, 433, 34 },
		{ 178, 434, 34 },
		{ 178, 435, 34 },
		{ 178, 436, 34 },
		{ 178, 437, 34 },
		{ 178, 438, 34 },
		{ 178, 439, 34 },
		{ 178, 440, 34 },
		{ 178, 441, 34 },
		{ 178, 442, 34 },
		{ 178, 443, 34 },
		{ 178, 444, 34 },
		{ 178, 445, 34 },
		{ 178, 446, 34 },
		{ 178, 447, 34 },
		{ 178, 448, 34 },
		{ 178, 449, 34 },
		{ 178, 450, 34 },
		{ 178, 451, 34 },
		{ 178, 452, 34 },
		{ 178, 453, 34 },
		{ 178, 454, 34 },
		{ 178, 455, 34 },
		{ 178, 456, 34 },
		{ 178, 457, 34 },
		{ 178, 458, 34 },
		{ 178, 459, 34 },
		{ 178, 460, 34 },
		{ 178, 461, 34 },
		{ 178, 462, 34 },
		{ 178, 463, 34 },
		{ 178, 464, 34 },
		{ 178, 465, 34 },
		{ 178, 466, 34 },
		{ 178, 467, 34 },
		{ 179, 382, 34 },
		{ 179, 383, 34 },
		{ 179, 384, 34 },
		{ 179, 385, 34 },
		{ 179, 386, 34 },
		{ 179, 387, 34 },
		{ 179, 388, 34 },
		{ 179, 389, 34 },
		{ 179, 390, 34 },
		{ 179, 391, 34 },
		{ 179, 392, 34 },
		{ 179, 393, 34 },
		{ 179, 394, 34 },
		{ 179, 395, 34 },
		{ 179, 396, 34 },
		{ 179, 397, 34 },
		{ 179, 398, 34 },
		{ 179, 399, 34 },
		{ 179, 400, 34 },
		{ 179, 401, 34 },
		{ 179, 402, 34 },
		{ 179, 403, 34 },
		{ 179, 404, 34 },
		{ 179, 405, 34 },
		{ 179, 406, 34 },
		{ 179, 407, 34 },
		{ 179, 408, 34 },
		{ 179, 409, 34 },
		{ 179, 410, 34 },
		{ 179, 411, 34 },
		{ 179, 412, 34 },
		{ 179, 413, 34 },
		{ 179, 414, 34 },
		{ 179, 415, 34 },
		{ 179, 416, 34 },
		{ 179, 417, 34 },
		{ 179, 418, 34 },
		{ 179, 419, 34 },
		{ 179, 420, 34 },
		{ 179, 421, 34 },
		{ 179, 422, 34 },
		{ 179, 423, 34 },
		{ 179, 424, 34 },
		{ 179, 425, 34 },
		{ 179, 426, 34 },
		{ 179, 427, 34 },
		{ 179, 428, 34 },
		{ 179, 429, 34 },
		{ 179, 430, 34 },
		{ 179, 431, 34 },
		{ 179, 432, 34 },
		{ 179, 433, 34 },
		{ 179, 434, 34 },
		{ 179, 435, 34 },
		{ 179, 436, 34 },
		{ 179, 437, 34 },
		{ 179, 438, 34 },
		{ 179, 439, 34 },
		{ 179, 440, 34 },
		{ 179, 441, 34 },
		{ 179, 442, 34 },
		{ 179, 443, 34 },
		{ 179, 444, 34 },
		{ 179, 445, 34 },
		{ 179, 446, 34 },
		{ 179, 447, 34 },
		{ 179, 448, 34 },
		{ 179, 449, 34 },
		{ 179, 450, 34 },
		{ 179, 451, 34 },
		{ 179, 452, 34 },
		{ 179, 453, 34 },
		{ 179, 454, 34 },
		{ 179, 455, 34 },
		{ 179, 456, 34 },
		{ 179, 457, 34 },
		{ 179, 458, 34 },
		{ 179, 459, 34 },
		{ 179, 460, 34 },
		{ 179, 461, 34 },
		{ 179, 462, 34 },
		{ 179, 463, 34 },
		{ 179, 464, 34 },
		{ 180, 381, 34 },
		{ 180, 382, 34 },
		{ 180, 383, 34 },
		{ 180, 384, 34 },
		{ 180, 385, 34 },
		{ 180, 386, 34 },
		{ 180, 387, 34 },
		{ 180, 388, 34 },
		{ 180, 389, 34 },
		{ 180, 390, 34 },
		{ 180, 391, 34 },
		{ 180, 392, 34 },
		{ 180, 393, 34 },
		{ 180, 394, 34 },
		{ 180, 395, 34 },
		{ 180, 396, 34 },
		{ 180, 397, 34 },
		{ 180, 398, 34 },
		{ 180, 399, 34 },
		{ 180, 400, 34 },
		{ 180, 401, 34 },
		{ 180, 402, 34 },
		{ 180, 403, 34 },
		{ 180, 404, 34 },
		{ 180, 405, 34 },
		{ 180, 406, 34 },
		{ 180, 407, 34 },
		{ 180, 408, 34 },
		{ 180, 409, 34 },
		{ 180, 410, 34 },
		{ 180, 411, 34 },
		{ 180, 412, 34 },
		{ 180, 413, 34 },
		{ 180, 414, 34 },
		{ 180, 415, 34 },
		{ 180, 416, 34 },
		{ 180, 417, 34 },
		{ 180, 418, 34 },
		{ 180, 419, 34 },
		{ 180, 420, 34 },
		{ 180, 421, 34 },
		{ 180, 422, 34 },
		{ 180, 423, 34 },
		{ 180, 424, 34 },
		{ 180, 425, 34 },
		{ 180, 426, 34 },
		{ 180, 427, 34 },
		{ 180, 428, 34 },
		{ 180, 429, 34 },
		{ 180, 430, 34 },
		{ 180, 431, 34 },
		{ 180, 432, 34 },
		{ 180, 433, 34 },
		{ 180, 434, 34 },
		{ 180, 435, 34 },
		{ 180, 436, 34 },
		{ 180, 437, 34 },
		{ 180, 438, 34 },
		{ 180, 439, 34 },
		{ 180, 440, 34 },
		{ 180, 441, 34 },
		{ 180, 442, 34 },
		{ 180, 443, 34 },
		{ 180, 444, 34 },
		{ 180, 445, 34 },
		{ 180, 446, 34 },
		{ 180, 447, 34 },
		{ 180, 448, 34 },
		{ 180, 449, 34 },
		{ 180, 450, 34 },
		{ 180, 451, 34 },
		{ 180, 452, 34 },
		{ 180, 453, 34 },
		{ 180, 454, 34 },
		{ 180, 455, 34 },
		{ 180, 456, 34 },
		{ 180, 457, 34 },
		{ 180, 458, 34 },
		{ 180, 459, 34 },
		{ 180, 460, 34 },
		{ 180, 461, 34 },
		{ 180, 462, 34 },
		{ 181, 379, 34 },
		{ 181, 380, 34 },
		{ 181, 381, 34 },
		{ 181, 382, 34 },
		{ 181, 383, 34 },
		{ 181, 384, 34 },
		{ 181, 385, 34 },
		{ 181, 386, 34 },
		{ 181, 387, 34 },
		{ 181, 388, 34 },
		{ 181, 389, 34 },
		{ 181, 390, 34 },
		{ 181, 391, 34 },
		{ 181, 392, 34 },
		{ 181, 393, 34 },
		{ 181, 394, 34 },
		{ 181, 395, 34 },
		{ 181, 396, 34 },
		{ 181, 397, 34 },
		{ 181, 398, 34 },
		{ 181, 399, 34 },
		{ 181, 400, 34 },
		{ 181, 401, 34 },
		{ 181, 402, 34 },
		{ 181, 403, 34 },
		{ 181, 404, 34 },
		{ 181, 405, 34 },
		{ 181, 406, 34 },
		{ 181, 407, 34 },
		{ 181, 408, 34 },
		{ 181, 409, 34 },
		{ 181, 410, 34 },
		{ 181, 411, 34 },
		{ 181, 412, 34 },
		{ 181, 413, 34 },
		{ 181, 414, 34 },
		{ 181, 415, 34 },
		{ 181, 416, 34 },
		{ 181, 417, 34 },
		{ 181, 418, 34 },
		{ 181, 419, 34 },
		{ 181, 420, 34 },
		{ 181, 421, 34 },
		{ 181, 422, 34 },
		{ 181, 423, 34 },
		{ 181, 424, 34 },
		{ 181, 425, 34 },
		{ 181, 426, 34 },
		{ 181, 427, 34 },
		{ 181, 428, 34 },
		{ 181, 429, 34 },
		{ 181, 430, 34 },
		{ 181, 431, 34 },
		{ 181, 432, 34 },
		{ 181, 433, 34 },
		{ 181, 434, 34 },
		{ 181, 435, 34 },
		{ 181, 436, 34 },
		{ 181, 437, 34 },
		{ 181, 438, 34 },
		{ 181, 439, 34 },
		{ 181, 440, 34 },
		{ 181, 441, 34 },
		{ 181, 442, 34 },
		{ 181, 443, 34 },
		{ 181, 444, 34 },
		{ 181, 445, 34 },
		{ 181, 446, 34 },
		{ 181, 447, 34 },
		{ 181, 448, 34 },
		{ 181, 449, 34 },
		{ 181, 450, 34 },
		{ 181, 451, 34 },
		{ 181, 452, 34 },
		{ 181, 453, 34 },
		{ 181, 454, 34 },
		{ 181, 455, 34 },
		{ 181, 456, 34 },
		{ 181, 457, 34 },
		{ 181, 458, 34 },
		{ 181, 459, 34 },
		{ 182, 378, 34 },
		{ 182, 379, 34 },
		{ 182, 380, 34 },
		{ 182, 381, 34 },
		{ 182, 382, 34 },
		{ 182, 383, 34 },
		{ 182, 384, 34 },
		{ 182, 385, 34 },
		{ 182, 386, 34 },
		{ 182, 387, 34 },
		{ 182, 388, 34 },
		{ 182, 389, 34 },
		{ 182, 390, 34 },
		{ 182, 391, 34 },
		{ 182, 392, 34 },
		{ 182, 393, 34 },
		{ 182, 394, 34 },
		{ 182, 395, 34 },
		{ 182, 396, 34 },
		{ 182, 397, 34 },
		{ 182, 398, 34 },
		{ 182, 399, 34 },
		{ 182, 400, 34 },
		{ 182, 401, 34 },
		{ 182, 402, 34 },
		{ 182, 403, 34 },
		{ 182, 404, 34 },
		{ 182, 405, 34 },
		{ 182, 406, 34 },
		{ 182, 407, 34 },
		{ 182, 408, 34 },
		{ 182, 409, 34 },
		{ 182, 410, 34 },
		{ 182, 411, 34 },
		{ 182, 412, 34 },
		{ 182, 413, 34 },
		{ 182, 414, 34 },
		{ 182, 415, 34 },
		{ 182, 416, 34 },
		{ 182, 417, 34 },
		{ 182, 418, 34 },
		{ 182, 419, 34 },
		{ 182, 420, 34 },
		{ 182, 421, 34 },
		{ 182, 422, 34 },
		{ 182, 423, 34 },
		{ 182, 424, 34 },
		{ 182, 425, 34 },
		{ 182, 426, 34 },
		{ 182, 427, 34 },
		{ 182, 428, 34 },
		{ 182, 429, 34 },
		{ 182, 430, 34 },
		{ 182, 431, 34 },
		{ 182, 432, 34 },
		{ 182, 433, 34 },
		{ 182, 434, 34 },
		{ 182, 435, 34 },
		{ 182, 436, 34 },
		{ 182, 437, 34 },
		{ 182, 438, 34 },
		{ 182, 439, 34 },
		{ 182, 440, 34 },
		{ 182, 441, 34 },
		{ 182, 442, 34 },
		{ 182, 443, 34 },
		{ 182, 444, 34 },
		{ 182, 445, 34 },
		{ 182, 446, 34 },
		{ 182, 447, 34 },
		{ 182, 448, 34 },
		{ 182, 449, 34 },
		{ 182, 450, 34 },
		{ 182, 451, 34 },
		{ 182, 452, 34 },
		{ 182, 453, 34 },
		{ 182, 454, 34 },
		{ 182, 455, 34 },
		{ 182, 456, 34 },
		{ 183, 376, 34 },
		{ 183, 377, 34 },
		{ 183, 378, 34 },
		{ 183, 379, 34 },
		{ 183, 380, 34 },
		{ 183, 381, 34 },
		{ 183, 382, 34 },
		{ 183, 383, 34 },
		{ 183, 384, 34 },
		{ 183, 385, 34 },
		{ 183, 386, 34 },
		{ 183, 387, 34 },
		{ 183, 388, 34 },
		{ 183, 389, 34 },
		{ 183, 390, 34 },
		{ 183, 391, 34 },
		{ 183, 392, 34 },
		{ 183, 393, 34 },
		{ 183, 394, 34 },
		{ 183, 395, 34 },
		{ 183, 396, 34 },
		{ 183, 397, 34 },
		{ 183, 398, 34 },
		{ 183, 399, 34 },
		{ 183, 400, 34 },
		{ 183, 401, 34 },
		{ 183, 402, 34 },
		{ 183, 403, 34 },
		{ 183, 404, 34 },
		{ 183, 405, 34 },
		{ 183, 406, 34 },
		{ 183, 407, 34 },
		{ 183, 408, 34 },
		{ 183, 409, 34 },
		{ 183, 410, 34 },
		{ 183, 411, 34 },
		{ 183, 412, 34 },
		{ 183, 413, 34 },
		{ 183, 414, 34 },
		{ 183, 415, 34 },
		{ 183, 416, 34 },
		{ 183, 417, 34 },
		{ 183, 418, 34 },
		{ 183, 419, 34 },
		{ 183, 420, 34 },
		{ 183, 421, 34 },
		{ 183, 422, 34 },
		{ 183, 423, 34 },
		{ 183, 424, 34 },
		{ 183, 425, 34 },
		{ 183, 426, 34 },
		{ 183, 427, 34 },
		{ 183, 428, 34 },
		{ 183, 429, 34 },
		{ 183, 430, 34 },
		{ 183, 431, 34 },
		{ 183, 432, 34 },
		{ 183, 433, 34 },
		{ 183, 434, 34 },
		{ 183, 435, 34 },
		{ 183, 436, 34 },
		{ 183, 437, 34 },
		{ 183, 438, 34 },
		{ 183, 439, 34 },
		{ 183, 440, 34 },
		{ 183, 441, 34 },
		{ 183, 442, 34 },
		{ 183, 443, 34 },
		{ 183, 444, 34 },
		{ 183, 445, 34 },
		{ 183, 446, 34 },
		{ 183, 447, 34 },
		{ 183, 448, 34 },
		{ 183, 449, 34 },
		{ 183, 450, 34 },
		{ 183, 451, 34 },
		{ 183, 452, 34 },
		{ 183, 453, 34 },
		{ 183, 454, 34 },
		{ 184, 374, 34 },
		{ 184, 375, 34 },
		{ 184, 376, 34 },
		{ 184, 377, 34 },
		{ 184, 378, 34 },
		{ 184, 379, 34 },
		{ 184, 380, 34 },
		{ 184, 381, 34 },
		{ 184, 382, 34 },
		{ 184, 383, 34 },
		{ 184, 384, 34 },
		{ 184, 385, 34 },
		{ 184, 386, 34 },
		{ 184, 387, 34 },
		{ 184, 388, 34 },
		{ 184, 389, 34 },
		{ 184, 390, 34 },
		{ 184, 391, 34 },
		{ 184, 392, 34 },
		{ 184, 393, 34 },
		{ 184, 394, 34 },
		{ 184, 395, 34 },
		{ 184, 396, 34 },
		{ 184, 397, 34 },
		{ 184, 398, 34 },
		{ 184, 399, 34 },
		{ 184, 400, 34 },
		{ 184, 401, 34 },
		{ 184, 402, 34 },
		{ 184, 403, 34 },
		{ 184, 404, 34 },
		{ 184, 405, 34 },
		{ 184, 406, 34 },
		{ 184, 407, 34 },
		{ 184, 408, 34 },
		{ 184, 409, 34 },
		{ 184, 410, 34 },
		{ 184, 411, 34 },
		{ 184, 412, 34 },
		{ 184, 413, 34 },
		{ 184, 414, 34 },
		{ 184, 415, 34 },
		{ 184, 416, 34 },
		{ 184, 417, 34 },
		{ 184, 418, 34 },
		{ 184, 419, 34 },
		{ 184, 420, 34 },
		{ 184, 421, 34 },
		{ 184, 422, 34 },
		{ 184, 423, 34 },
		{ 184, 424, 34 },
		{ 184, 425, 34 },
		{ 184, 426, 34 },
		{ 184, 427, 34 },
		{ 184, 428, 34 },
		{ 184, 429, 34 },
		{ 184, 430, 34 },
		{ 184, 431, 34 },
		{ 184, 432, 34 },
		{ 184, 433, 34 },
		{ 184, 434, 34 },
		{ 184, 435, 34 },
		{ 184, 436, 34 },
		{ 184, 437, 34 },
		{ 184, 438, 34 },
		{ 184, 439, 34 },
		{ 184, 440, 34 },
		{ 184, 441, 34 },
		{ 184, 442, 34 },
		{ 184, 443, 34 },
		{ 184, 444, 34 },
		{ 184, 445, 34 },
		{ 184, 446, 34 },
		{ 184, 447, 34 },
		{ 184, 448, 34 },
		{ 184, 449, 34 },
		{ 184, 450, 34 },
		{ 184, 451, 34 },
		{ 185, 373, 34 },
		{ 185, 374, 34 },
		{ 185, 375, 34 },
		{ 185, 376, 34 },
		{ 185, 377, 34 },
		{ 185, 378, 34 },
		{ 185, 379, 34 },
		{ 185, 380, 34 },
		{ 185, 381, 34 },
		{ 185, 382, 34 },
		{ 185, 383, 34 },
		{ 185, 384, 34 },
		{ 185, 385, 34 },
		{ 185, 386, 34 },
		{ 185, 387, 34 },
		{ 185, 388, 34 },
		{ 185, 389, 34 },
		{ 185, 390, 34 },
		{ 185, 391, 34 },
		{ 185, 392, 34 },
		{ 185, 393, 34 },
		{ 185, 394, 34 },
		{ 185, 395, 34 },
		{ 185, 396, 34 },
		{ 185, 397, 34 },
		{ 185, 398, 34 },
		{ 185, 399, 34 },
		{ 185, 400, 34 },
		{ 185, 401, 34 },
		{ 185, 402, 34 },
		{ 185, 403, 34 },
		{ 185, 404, 34 },
		{ 185, 405, 34 },
		{ 185, 406, 34 },
		{ 185, 407, 34 },
		{ 185, 408, 34 },
		{ 185, 409, 34 },
		{ 185, 410, 34 },
		{ 185, 411, 34 },
		{ 185, 412, 34 },
		{ 185, 413, 34 },
		{ 185, 414, 34 },
		{ 185, 415, 34 },
		{ 185, 416, 34 },
		{ 185, 417, 34 },
		{ 185, 418, 34 },
		{ 185, 419, 34 },
		{ 185, 420, 34 },
		{ 185, 421, 34 },
		{ 185, 422, 34 },
		{ 185, 423, 34 },
		{ 185, 424, 34 },
		{ 185, 425, 34 },
		{ 185, 426, 34 },
		{ 185, 427, 34 },
		{ 185, 428, 34 },
		{ 185, 429, 34 },
		{ 185, 430, 34 },
		{ 185, 431, 34 },
		{ 185, 432, 34 },
		{ 185, 433, 34 },
		{ 185, 434, 34 },
		{ 185, 435, 34 },
		{ 185, 436, 34 },
		{ 185, 437, 34 },
		{ 185, 438, 34 },
		{ 185, 439, 34 },
		{ 185, 440, 34 },
		{ 185, 441, 34 },
		{ 185, 442, 34 },
		{ 185, 443, 34 },
		{ 185, 444, 34 },
		{ 185, 445, 34 },
		{ 185, 446, 34 },
		{ 185, 447, 34 },
		{ 185, 448, 34 },
		{ 186, 371, 34 },
		{ 186, 372, 34 },
		{ 186, 373, 34 },
		{ 186, 374, 34 },
		{ 186, 375, 34 },
		{ 186, 376, 34 },
		{ 186, 377, 34 },
		{ 186, 378, 34 },
		{ 186, 379, 34 },
		{ 186, 380, 34 },
		{ 186, 381, 34 },
		{ 186, 382, 34 },
		{ 186, 383, 34 },
		{ 186, 384, 34 },
		{ 186, 385, 34 },
		{ 186, 386, 34 },
		{ 186, 387, 34 },
		{ 186, 388, 34 },
		{ 186, 389, 34 },
		{ 186, 390, 34 },
		{ 186, 391, 34 },
		{ 186, 392, 34 },
		{ 186, 393, 34 },
		{ 186, 394, 34 },
		{ 186, 395, 34 },
		{ 186, 396, 34 },
		{ 186, 397, 34 },
		{ 186, 398, 34 },
		{ 186, 399, 34 },
		{ 186, 400, 34 },
		{ 186, 401, 34 },
		{ 186, 402, 34 },
		{ 186, 403, 34 },
		{ 186, 404, 34 },
		{ 186, 405, 34 },
		{ 186, 406, 34 },
		{ 186, 407, 34 },
		{ 186, 408, 34 },
		{ 186, 409, 34 },
		{ 186, 410, 34 },
		{ 186, 411, 34 },
		{ 186, 412, 34 },
		{ 186, 413, 34 },
		{ 186, 414, 34 },
		{ 186, 415, 34 },
		{ 186, 416, 34 },
		{ 186, 417, 34 },
		{ 186, 418, 34 },
		{ 186, 419, 34 },
		{ 186, 420, 34 },
		{ 186, 421, 34 },
		{ 186, 422, 34 },
		{ 186, 423, 34 },
		{ 186, 424, 34 },
		{ 186, 425, 34 },
		{ 186, 426, 34 },
		{ 186, 427, 34 },
		{ 186, 428, 34 },
		{ 186, 429, 34 },
		{ 186, 430, 34 },
		{ 186, 431, 34 },
		{ 186, 432, 34 },
		{ 186, 433, 34 },
		{ 186, 434, 34 },
		{ 186, 435, 34 },
		{ 186, 436, 34 },
		{ 186, 437, 34 },
		{ 186, 438, 34 },
		{ 186, 439, 34 },
		{ 186, 440, 34 },
		{ 186, 441, 34 },
		{ 186, 442, 34 },
		{ 186, 443, 34 },
		{ 186, 444, 34 },
		{ 186, 445, 34 },
		{ 186, 446, 34 },
		{ 187, 370, 34 },
		{ 187, 371, 34 },
		{ 187, 372, 34 },
		{ 187, 373, 34 },
		{ 187, 374, 34 },
		{ 187, 375, 34 },
		{ 187, 376, 34 },
		{ 187, 377, 34 },
		{ 187, 378, 34 },
		{ 187, 379, 34 },
		{ 187, 380, 34 },
		{ 187, 381, 34 },
		{ 187, 382, 34 },
		{ 187, 383, 34 },
		{ 187, 384, 34 },
		{ 187, 385, 34 },
		{ 187, 386, 34 },
		{ 187, 387, 34 },
		{ 187, 388, 34 },
		{ 187, 389, 34 },
		{ 187, 390, 34 },
		{ 187, 391, 34 },
		{ 187, 392, 34 },
		{ 187, 393, 34 },
		{ 187, 394, 34 },
		{ 187, 395, 34 },
		{ 187, 396, 34 },
		{ 187, 397, 34 },
		{ 187, 398, 34 },
		{ 187, 399, 34 },
		{ 187, 400, 34 },
		{ 187, 401, 34 },
		{ 187, 402, 34 },
		{ 187, 403, 34 },
		{ 187, 404, 34 },
		{ 187, 405, 34 },
		{ 187, 406, 34 },
		{ 187, 407, 34 },
		{ 187, 408, 34 },
		{ 187, 409, 34 },
		{ 187, 410, 34 },
		{ 187, 411, 34 },
		{ 187, 412, 34 },
		{ 187, 413, 34 },
		{ 187, 414, 34 },
		{ 187, 415, 34 },
		{ 187, 416, 34 },
		{ 187, 417, 34 },
		{ 187, 418, 34 },
		{ 187, 419, 34 },
		{ 187, 420, 34 },
		{ 187, 421, 34 },
		{ 187, 422, 34 },
		{ 187, 423, 34 },
		{ 187, 424, 34 },
		{ 187, 425, 34 },
		{ 187, 426, 34 },
		{ 187, 427, 34 },
		{ 187, 428, 34 },
		{ 187, 429, 34 },
		{ 187, 430, 34 },
		{ 187, 431, 34 },
		{ 187, 432, 34 },
		{ 187, 433, 34 },
		{ 187, 434, 34 },
		{ 187, 435, 34 },
		{ 187, 436, 34 },
		{ 187, 437, 34 },
		{ 187, 438, 34 },
		{ 187, 439, 34 },
		{ 187, 440, 34 },
		{ 187, 441, 34 },
		{ 187, 442, 34 },
		{ 187, 443, 34 },
		{ 188, 368, 34 },
		{ 188, 369, 34 },
		{ 188, 370, 34 },
		{ 188, 371, 34 },
		{ 188, 372, 34 },
		{ 188, 373, 34 },
		{ 188, 374, 34 },
		{ 188, 375, 34 },
		{ 188, 376, 34 },
		{ 188, 377, 34 },
		{ 188, 378, 34 },
		{ 188, 379, 34 },
		{ 188, 380, 34 },
		{ 188, 381, 34 },
		{ 188, 382, 34 },
		{ 188, 383, 34 },
		{ 188, 384, 34 },
		{ 188, 385, 34 },
		{ 188, 386, 34 },
		{ 188, 387, 34 },
		{ 188, 388, 34 },
		{ 188, 389, 34 },
		{ 188, 390, 34 },
		{ 188, 391, 34 },
		{ 188, 392, 34 },
		{ 188, 393, 34 },
		{ 188, 394, 34 },
		{ 188, 395, 34 },
		{ 188, 396, 34 },
		{ 188, 397, 34 },
		{ 188, 398, 34 },
		{ 188, 399, 34 },
		{ 188, 400, 34 },
		{ 188, 401, 34 },
		{ 188, 402, 34 },
		{ 188, 403, 34 },
		{ 188, 404, 34 },
		{ 188, 405, 34 },
		{ 188, 406, 34 },
		{ 188, 407, 34 },
		{ 188, 408, 34 },
		{ 188, 409, 34 },
		{ 188, 410, 34 },
		{ 188, 411, 34 },
		{ 188, 412, 34 },
		{ 188, 413, 34 },
		{ 188, 414, 34 },
		{ 188, 415, 34 },
		{ 188, 416, 34 },
		{ 188, 417, 34 },
		{ 188, 418, 34 },
		{ 188, 419, 34 },
		{ 188, 420, 34 },
		{ 188, 421, 34 },
		{ 188, 422, 34 },
		{ 188, 423, 34 },
		{ 188, 424, 34 },
		{ 188, 425, 34 },
		{ 188, 426, 34 },
		{ 188, 427, 34 },
		{ 188, 428, 34 },
		{ 188, 429, 34 },
		{ 188, 430, 34 },
		{ 188, 431, 34 },
		{ 188, 432, 34 },
		{ 188, 433, 34 },
		{ 188, 434, 34 },
		{ 188, 435, 34 },
		{ 188, 436, 34 },
		{ 188, 437, 34 },
		{ 188, 438, 34 },
		{ 188, 439, 34 },
		{ 188, 440, 34 },
		{ 189, 366, 34 },
		{ 189, 367, 34 },
		{ 189, 368, 34 },
		{ 189, 369, 34 },
		{ 189, 370, 34 },
		{ 189, 371, 34 },
		{ 189, 372, 34 },
		{ 189, 373, 34 },
		{ 189, 374, 34 },
		{ 189, 375, 34 },
		{ 189, 376, 34 },
		{ 189, 377, 34 },
		{ 189, 378, 34 },
		{ 189, 379, 34 },
		{ 189, 380, 34 },
		{ 189, 381, 34 },
		{ 189, 382, 34 },
		{ 189, 383, 34 },
		{ 189, 384, 34 },
		{ 189, 385, 34 },
		{ 189, 386, 34 },
		{ 189, 387, 34 },
		{ 189, 388, 34 },
		{ 189, 389, 34 },
		{ 189, 390, 34 },
		{ 189, 391, 34 },
		{ 189, 392, 34 },
		{ 189, 393, 34 },
		{ 189, 394, 34 },
		{ 189, 395, 34 },
		{ 189, 396, 34 },
		{ 189, 397, 34 },
		{ 189, 398, 34 },
		{ 189, 399, 34 },
		{ 189, 400, 34 },
		{ 189, 401, 34 },
		{ 189, 402, 34 },
		{ 189, 403, 34 },
		{ 189, 404, 34 },
		{ 189, 405, 34 },
		{ 189, 406, 34 },
		{ 189, 407, 34 },
		{ 189, 408, 34 },
		{ 189, 409, 34 },
		{ 189, 410, 34 },
		{ 189, 411, 34 },
		{ 189, 412, 34 },
		{ 189, 413, 34 },
		{ 189, 414, 34 },
		{ 189, 415, 34 },
		{ 189, 416, 34 },
		{ 189, 417, 34 },
		{ 189, 418, 34 },
		{ 189, 419, 34 },
		{ 189, 420, 34 },
		{ 189, 421, 34 },
		{ 189, 422, 34 },
		{ 189, 423, 34 },
		{ 189, 424, 34 },
		{ 189, 425, 34 },
		{ 189, 426, 34 },
		{ 189, 427, 34 },
		{ 189, 428, 34 },
		{ 189, 429, 34 },
		{ 189, 430, 34 },
		{ 189, 431, 34 },
		{ 189, 432, 34 },
		{ 189, 433, 34 },
		{ 189, 434, 34 },
		{ 189, 435, 34 },
		{ 189, 436, 34 },
		{ 189, 437, 34 },
		{ 189, 438, 34 },
		{ 190, 365, 34 },
		{ 190, 366, 34 },
		{ 190, 367, 34 },
		{ 190, 368, 34 },
		{ 190, 369, 34 },
		{ 190, 370, 34 },
		{ 190, 371, 34 },
		{ 190, 372, 34 },
		{ 190, 373, 34 },
		{ 190, 374, 34 },
		{ 190, 375, 34 },
		{ 190, 376, 34 },
		{ 190, 377, 34 },
		{ 190, 378, 34 },
		{ 190, 379, 34 },
		{ 190, 380, 34 },
		{ 190, 381, 34 },
		{ 190, 382, 34 },
		{ 190, 383, 34 },
		{ 190, 384, 34 },
		{ 190, 385, 34 },
		{ 190, 386, 34 },
		{ 190, 387, 34 },
		{ 190, 388, 34 },
		{ 190, 389, 34 },
		{ 190, 390, 34 },
		{ 190, 391, 34 },
		{ 190, 392, 34 },
		{ 190, 393, 34 },
		{ 190, 394, 34 },
		{ 190, 395, 34 },
		{ 190, 396, 34 },
		{ 190, 397, 34 },
		{ 190, 398, 34 },
		{ 190, 399, 34 },
		{ 190, 400, 34 },
		{ 190, 401, 34 },
		{ 190, 402, 34 },
		{ 190, 403, 34 },
		{ 190, 404, 34 },
		{ 190, 405, 34 },
		{ 190, 406, 34 },
		{ 190, 407, 34 },
		{ 190, 408, 34 },
		{ 190, 409, 34 },
		{ 190, 410, 34 },
		{ 190, 411, 34 },
		{ 190, 412, 34 },
		{ 190, 413, 34 },
		{ 190, 414, 34 },
		{ 190, 415, 34 },
		{ 190, 416, 34 },
		{ 190, 417, 34 },
		{ 190, 418, 34 },
		{ 190, 419, 34 },
		{ 190, 420, 34 },
		{ 190, 421, 34 },
		{ 190, 422, 34 },
		{ 190, 423, 34 },
		{ 190, 424, 34 },
		{ 190, 425, 34 },
		{ 190, 426, 34 },
		{ 190, 427, 34 },
		{ 190, 428, 34 },
		{ 190, 429, 34 },
		{ 190, 430, 34 },
		{ 190, 431, 34 },
		{ 190, 432, 34 },
		{ 190, 433, 34 },
		{ 190, 434, 34 },
		{ 190, 435, 34 },
		{ 191, 363, 34 },
		{ 191, 364, 34 },
		{ 191, 365, 34 },
		{ 191, 366, 34 },
		{ 191, 367, 34 },
		{ 191, 368, 34 },
		{ 191, 369, 34 },
		{ 191, 370, 34 },
		{ 191, 371, 34 },
		{ 191, 372, 34 },
		{ 191, 373, 34 },
		{ 191, 374, 34 },
		{ 191, 375, 34 },
		{ 191, 376, 34 },
		{ 191, 377, 34 },
		{ 191, 378, 34 },
		{ 191, 379, 34 },
		{ 191, 380, 34 },
		{ 191, 381, 34 },
		{ 191, 382, 34 },
		{ 191, 383, 34 },
		{ 191, 384, 34 },
		{ 191, 385, 34 },
		{ 191, 386, 34 },
		{ 191, 387, 34 },
		{ 191, 388, 34 },
		{ 191, 389, 34 },
		{ 191, 390, 34 },
		{ 191, 391, 34 },
		{ 191, 392, 34 },
		{ 191, 393, 34 },
		{ 191, 394, 34 },
		{ 191, 395, 34 },
		{ 191, 396, 34 },
		{ 191, 397, 34 },
		{ 191, 398, 34 },
		{ 191, 399, 34 },
		{ 191, 400, 34 },
		{ 191, 401, 34 },
		{ 191, 402, 34 },
		{ 191, 403, 34 },
		{ 191, 404, 34 },
		{ 191, 405, 34 },
		{ 191, 406, 34 },
		{ 191, 407, 34 },
		{ 191, 408, 34 },
		{ 191, 409, 34 },
		{ 191, 410, 34 },
		{ 191, 411, 34 },
		{ 191, 412, 34 },
		{ 191, 413, 34 },
		{ 191, 414, 34 },
		{ 191, 415, 34 },
		{ 191, 416, 34 },
		{ 191, 417, 34 },
		{ 191, 418, 34 },
		{ 191, 419, 34 },
		{ 191, 420, 34 },
		{ 191, 421, 34 },
		{ 191, 422, 34 },
		{ 191, 423, 34 },
		{ 191, 424, 34 },
		{ 191, 425, 34 },
		{ 191, 426, 34 },
		{ 191, 427, 34 },
		{ 191, 428, 34 },
		{ 191, 429, 34 },
		{ 191, 430, 34 },
		{ 191, 431, 34 },
		{ 191, 432, 34 },
		{ 192, 362, 34 },
		{ 192, 363, 34 },
		{ 192, 364, 34 },
		{ 192, 365, 34 },
		{ 192, 366, 34 },
		{ 192, 367, 34 },
		{ 192, 368, 34 },
		{ 192, 369, 34 },
		{ 192, 370, 34 },
		{ 192, 371, 34 },
		{ 192, 372, 34 },
		{ 192, 373, 34 },
		{ 192, 374, 34 },
		{ 192, 375, 34 },
		{ 192, 376, 34 },
		{ 192, 377, 34 },
		{ 192, 378, 34 },
		{ 192, 379, 34 },
		{ 192, 380, 34 },
		{ 192, 381, 34 },
		{ 192, 382, 34 },
		{ 192, 383, 34 },
		{ 192, 384, 34 },
		{ 192, 385, 34 },
		{ 192, 386, 34 },
		{ 192, 387, 34 },
		{ 192, 388, 34 },
		{ 192, 389, 34 },
		{ 192, 390, 34 },
		{ 192, 391, 34 },
		{ 192, 392, 34 },
		{ 192, 393, 34 },
		{ 192, 394, 34 },
		{ 192, 395, 34 },
		{ 192, 396, 34 },
		{ 192, 397, 34 },
		{ 192, 398, 34 },
		{ 192, 399, 34 },
		{ 192, 400, 34 },
		{ 192, 401, 34 },
		{ 192, 402, 34 },
		{ 192, 403, 34 },
		{ 192, 404, 34 },
		{ 192, 405, 34 },
		{ 192, 406, 34 },
		{ 192, 407, 34 },
		{ 192, 408, 34 },
		{ 192, 409, 34 },
		{ 192, 410, 34 },
		{ 192, 411, 34 },
		{ 192, 412, 34 },
		{ 192, 413, 34 },
		{ 192, 414, 34 },
		{ 192, 415, 34 },
		{ 192, 416, 34 },
		{ 192, 417, 34 },
		{ 192, 418, 34 },
		{ 192, 419, 34 },
		{ 192, 420, 34 },
		{ 192, 421, 34 },
		{ 192, 422, 34 },
		{ 192, 423, 34 },
		{ 192, 424, 34 },
		{ 192, 425, 34 },
		{ 192, 426, 34 },
		{ 192, 427, 34 },
		{ 192, 428, 34 },
		{ 192, 429, 34 },
		{ 192, 430, 34 },
		{ 193, 360, 34 },
		{ 193, 361, 34 },
		{ 193, 362, 34 },
		{ 193, 363, 34 },
		{ 193, 364, 34 },
		{ 193, 365, 34 },
		{ 193, 366, 34 },
		{ 193, 367, 34 },
		{ 193, 368, 34 },
		{ 193, 369, 34 },
		{ 193, 370, 34 },
		{ 193, 371, 34 },
		{ 193, 372, 34 },
		{ 193, 373, 34 },
		{ 193, 374, 34 },
		{ 193, 375, 34 },
		{ 193, 376, 34 },
		{ 193, 377, 34 },
		{ 193, 378, 34 },
		{ 193, 379, 34 },
		{ 193, 380, 34 },
		{ 193, 381, 34 },
		{ 193, 382, 34 },
		{ 193, 383, 34 },
		{ 193, 384, 34 },
		{ 193, 385, 34 },
		{ 193, 386, 34 },
		{ 193, 387, 34 },
		{ 193, 388, 34 },
		{ 193, 389, 34 },
		{ 193, 390, 34 },
		{ 193, 391, 34 },
		{ 193, 392, 34 },
		{ 193, 393, 34 },
		{ 193, 394, 34 },
		{ 193, 395, 34 },
		{ 193, 396, 34 },
		{ 193, 397, 34 },
		{ 193, 398, 34 },
		{ 193, 399, 34 },
		{ 193, 400, 34 },
		{ 193, 401, 34 },
		{ 193, 402, 34 },
		{ 193, 403, 34 },
		{ 193, 404, 34 },
		{ 193, 405, 34 },
		{ 193, 406, 34 },
		{ 193, 407, 34 },
		{ 193, 408, 34 },
		{ 193, 409, 34 },
		{ 193, 410, 34 },
		{ 193, 411, 34 },
		{ 193, 412, 34 },
		{ 193, 413, 34 },
		{ 193, 414, 34 },
		{ 193, 415, 34 },
		{ 193, 416, 34 },
		{ 193, 417, 34 },
		{ 193, 418, 34 },
		{ 193, 419, 34 },
		{ 193, 420, 34 },
		{ 193, 421, 34 },
		{ 193, 422, 34 },
		{ 193, 423, 34 },
		{ 193, 424, 34 },
		{ 193, 425, 34 },
		{ 193, 426, 34 },
		{ 193, 427, 34 },
		{ 194, 358, 34 },
		{ 194, 359, 34 },
		{ 194, 360, 34 },
		{ 194, 361, 34 },
		{ 194, 362, 34 },
		{ 194, 363, 34 },
		{ 194, 364, 34 },
		{ 194, 365, 34 },
		{ 194, 366, 34 },
		{ 194, 367, 34 },
		{ 194, 368, 34 },
		{ 194, 369, 34 },
		{ 194, 370, 34 },
		{ 194, 371, 34 },
		{ 194, 372, 34 },
		{ 194, 373, 34 },
		{ 194, 374, 34 },
		{ 194, 375, 34 },
		{ 194, 376, 34 },
		{ 194, 377, 34 },
		{ 194, 378, 34 },
		{ 194, 379, 34 },
		{ 194, 380, 34 },
		{ 194, 381, 34 },
		{ 194, 382, 34 },
		{ 194, 383, 34 },
		{ 194, 384, 34 },
		{ 194, 385, 34 },
		{ 194, 386, 34 },
		{ 194, 387, 34 },
		{ 194, 388, 34 },
		{ 194, 389, 34 },
		{ 194, 390, 34 },
		{ 194, 391, 34 },
		{ 194, 392, 34 },
		{ 194, 393, 34 },
		{ 194, 394, 34 },
		{ 194, 395, 34 },
		{ 194, 396, 34 },
		{ 194, 397, 34 },
		{ 194, 398, 34 },
		{ 194, 399, 34 },
		{ 194, 400, 34 },
		{ 194, 401, 34 },
		{ 194, 402, 34 },
		{ 194, 403, 34 },
		{ 194, 404, 34 },
		{ 194, 405, 34 },
		{ 194, 406, 34 },
		{ 194, 407, 34 },
		{ 194, 408, 34 },
		{ 194, 409, 34 },
		{ 194, 410, 34 },
		{ 194, 411, 34 },
		{ 194, 412, 34 },
		{ 194, 413, 34 },
		{ 194, 414, 34 },
		{ 194, 415, 34 },
		{ 194, 416, 34 },
		{ 194, 417, 34 },
		{ 194, 418, 34 },
		{ 194, 419, 34 },
		{ 194, 420, 34 },
		{ 194, 421, 34 },
		{ 194, 422, 34 },
		{ 194, 423, 34 },
		{ 194, 424, 34 },
		{ 195, 357, 34 },
		{ 195, 358, 34 },
		{ 195, 359, 34 },
		{ 195, 360, 34 },
		{ 195, 361, 34 },
		{ 195, 362, 34 },
		{ 195, 363, 34 },
		{ 195, 364, 34 },
		{ 195, 365, 34 },
		{ 195, 366, 34 },
		{ 195, 367, 34 },
		{ 195, 368, 34 },
		{ 195, 369, 34 },
		{ 195, 370, 34 },
		{ 195, 371, 34 },
		{ 195, 372, 34 },
		{ 195, 373, 34 },
		{ 195, 374, 34 },
		{ 195, 375, 34 },
		{ 195, 376, 34 },
		{ 195, 377, 34 },
		{ 195, 378, 34 },
		{ 195, 379, 34 },
		{ 195, 380, 34 },
		{ 195, 381, 34 },
		{ 195, 382, 34 },
		{ 195, 383, 34 },
		{ 195, 384, 34 },
		{ 195, 385, 34 },
		{ 195, 386, 34 },
		{ 195, 387, 34 },
		{ 195, 388, 34 },
		{ 195, 389, 34 },
		{ 195, 390, 34 },
		{ 195, 391, 34 },
		{ 195, 392, 34 },
		{ 195, 393, 34 },
		{ 195, 394, 34 },
		{ 195, 395, 34 },
		{ 195, 396, 34 },
		{ 195, 397, 34 },
		{ 195, 398, 34 },
		{ 195, 399, 34 },
		{ 195, 400, 34 },
		{ 195, 401, 34 },
		{ 195, 402, 34 },
		{ 195, 403, 34 },
		{ 195, 404, 34 },
		{ 195, 405, 34 },
		{ 195, 406, 34 },
		{ 195, 407, 34 },
		{ 195, 408, 34 },
		{ 195, 409, 34 },
		{ 195, 410, 34 },
		{ 195, 411, 34 },
		{ 195, 412, 34 },
		{ 195, 413, 34 },
		{ 195, 414, 34 },
		{ 195, 415, 34 },
		{ 195, 416, 34 },
		{ 195, 417, 34 },
		{ 195, 418, 34 },
		{ 195, 419, 34 },
		{ 195, 420, 34 },
		{ 195, 421, 34 },
		{ 195, 422, 34 },
		{ 196, 355, 34 },
		{ 196, 356, 34 },
		{ 196, 357, 34 },
		{ 196, 358, 34 },
		{ 196, 359, 34 },
		{ 196, 360, 34 },
		{ 196, 361, 34 },
		{ 196, 362, 34 },
		{ 196, 363, 34 },
		{ 196, 364, 34 },
		{ 196, 365, 34 },
		{ 196, 366, 34 },
		{ 196, 367, 34 },
		{ 196, 368, 34 },
		{ 196, 369, 34 },
		{ 196, 370, 34 },
		{ 196, 371, 34 },
		{ 196, 372, 34 },
		{ 196, 373, 34 },
		{ 196, 374, 34 },
		{ 196, 375, 34 },
		{ 196, 376, 34 },
		{ 196, 377, 34 },
		{ 196, 378, 34 },
		{ 196, 379, 34 },
		{ 196, 380, 34 },
		{ 196, 381, 34 },
		{ 196, 382, 34 },
		{ 196, 383, 34 },
		{ 196, 384, 34 },
		{ 196, 385, 34 },
		{ 196, 386, 34 },
		{ 196, 387, 34 },
		{ 196, 388, 34 },
		{ 196, 389, 34 },
		{ 196, 390, 34 },
		{ 196, 391, 34 },
		{ 196, 392, 34 },
		{ 196, 393, 34 },
		{ 196, 394, 34 },
		{ 196, 395, 34 },
		{ 196, 396, 34 },
		{ 196, 397, 34 },
		{ 196, 398, 34 },
		{ 196, 399, 34 },
		{ 196, 400, 34 },
		{ 196, 401, 34 },
		{ 196, 402, 34 },
		{ 196, 403, 34 },
		{ 196, 404, 34 },
		{ 196, 405, 34 },
		{ 196, 406, 34 },
		{ 196, 407, 34 },
		{ 196, 408, 34 },
		{ 196, 409, 34 },
		{ 196, 410, 34 },
		{ 196, 411, 34 },
		{ 196, 412, 34 },
		{ 196, 413, 34 },
		{ 196, 414, 34 },
		{ 196, 415, 34 },
		{ 196, 416, 34 },
		{ 196, 417, 34 },
		{ 196, 418, 34 },
		{ 196, 419, 34 },
		{ 197, 354, 34 },
		{ 197, 355, 34 },
		{ 197, 356, 34 },
		{ 197, 357, 34 },
		{ 197, 358, 34 },
		{ 197, 359, 34 },
		{ 197, 360, 34 },
		{ 197, 361, 34 },
		{ 197, 362, 34 },
		{ 197, 363, 34 },
		{ 197, 364, 34 },
		{ 197, 365, 34 },
		{ 197, 366, 34 },
		{ 197, 367, 34 },
		{ 197, 368, 34 },
		{ 197, 369, 34 },
		{ 197, 370, 34 },
		{ 197, 371, 34 },
		{ 197, 372, 34 },
		{ 197, 373, 34 },
		{ 197, 374, 34 },
		{ 197, 375, 34 },
		{ 197, 376, 34 },
		{ 197, 377, 34 },
		{ 197, 378, 34 },
		{ 197, 379, 34 },
		{ 197, 380, 34 },
		{ 197, 381, 34 },
		{ 197, 382, 34 },
		{ 197, 383, 34 },
		{ 197, 384, 34 },
		{ 197, 385, 34 },
		{ 197, 386, 34 },
		{ 197, 387, 34 },
		{ 197, 388, 34 },
		{ 197, 389, 34 },
		{ 197, 390, 34 },
		{ 197, 391, 34 },
		{ 197, 392, 34 },
		{ 197, 393, 34 },
		{ 197, 394, 34 },
		{ 197, 395, 34 },
		{ 197, 396, 34 },
		{ 197, 397, 34 },
		{ 197, 398, 34 },
		{ 197, 399, 34 },
		{ 197, 400, 34 },
		{ 197, 401, 34 },
		{ 197, 402, 34 },
		{ 197, 403, 34 },
		{ 197, 404, 34 },
		{ 197, 405, 34 },
		{ 197, 406, 34 },
		{ 197, 407, 34 },
		{ 197, 408, 34 },
		{ 197, 409, 34 },
		{ 197, 410, 34 },
		{ 197, 411, 34 },
		{ 197, 412, 34 },
		{ 197, 413, 34 },
		{ 197, 414, 34 },
		{ 197, 415, 34 },
		{ 197, 416, 34 },
		{ 198, 352, 34 },
		{ 198, 353, 34 },
		{ 198, 354, 34 },
		{ 198, 355, 34 },
		{ 198, 356, 34 },
		{ 198, 357, 34 },
		{ 198, 358, 34 },
		{ 198, 359, 34 },
		{ 198, 360, 34 },
		{ 198, 361, 34 },
		{ 198, 362, 34 },
		{ 198, 363, 34 },
		{ 198, 364, 34 },
		{ 198, 365, 34 },
		{ 198, 366, 34 },
		{ 198, 367, 34 },
		{ 198, 368, 34 },
		{ 198, 369, 34 },
		{ 198, 370, 34 },
		{ 198, 371, 34 },
		{ 198, 372, 34 },
		{ 198, 373, 34 },
		{ 198, 374, 34 },
		{ 198, 375, 34 },
		{ 198, 376, 34 },
		{ 198, 377, 34 },
		{ 198, 378, 34 },
		{ 198, 379, 34 },
		{ 198, 380, 34 },
		{ 198, 381, 34 },
		{ 198, 382, 34 },
		{ 198, 383, 34 },
		{ 198, 384, 34 },
		{ 198, 385, 34 },
		{ 198, 386, 34 },
		{ 198, 387, 34 },
		{ 198, 388, 34 },
		{ 198, 389, 34 },
		{ 198, 390, 34 },
		{ 198, 391, 34 },
		{ 198, 392, 34 },
		{ 198, 393, 34 },
		{ 198, 394, 34 },
		{ 198, 395, 34 },
		{ 198, 396, 34 },
		{ 198, 397, 34 },
		{ 198, 398, 34 },
		{ 198, 399, 34 },
		{ 198, 400, 34 },
		{ 198, 401, 34 },
		{ 198, 402, 34 },
		{ 198, 403, 34 },
		{ 198, 404, 34 },
		{ 198, 405, 34 },
		{ 198, 406, 34 },
		{ 198, 407, 34 },
		{ 198, 408, 34 },
		{ 198, 409, 34 },
		{ 198, 410, 34 },
		{ 198, 411, 34 },
		{ 198, 412, 34 },
		{ 198, 413, 34 },
		{ 198, 414, 34 },
		{ 199, 350, 34 },
		{ 199, 351, 34 },
		{ 199, 352, 34 },
		{ 199, 353, 34 },
		{ 199, 354, 34 },
		{ 199, 355, 34 },
		{ 199, 356, 34 },
		{ 199, 357, 34 },
		{ 199, 358, 34 },
		{ 199, 359, 34 },
		{ 199, 360, 34 },
		{ 199, 361, 34 },
		{ 199, 362, 34 },
		{ 199, 363, 34 },
		{ 199, 364, 34 },
		{ 199, 365, 34 },
		{ 199, 366, 34 },
		{ 199, 367, 34 },
		{ 199, 368, 34 },
		{ 199, 369, 34 },
		{ 199, 370, 34 },
		{ 199, 371, 34 },
		{ 199, 372, 34 },
		{ 199, 373, 34 },
		{ 199, 374, 34 },
		{ 199, 375, 34 },
		{ 199, 376, 34 },
		{ 199, 377, 34 },
		{ 199, 378, 34 },
		{ 199, 379, 34 },
		{ 199, 380, 34 },
		{ 199, 381, 34 },
		{ 199, 382, 34 },
		{ 199, 383, 34 },
		{ 199, 384, 34 },
		{ 199, 385, 34 },
		{ 199, 386, 34 },
		{ 199, 387, 34 },
		{ 199, 388, 34 },
		{ 199, 389, 34 },
		{ 199, 390, 34 },
		{ 199, 391, 34 },
		{ 199, 392, 34 },
		{ 199, 393, 34 },
		{ 199, 394, 34 },
		{ 199, 395, 34 },
		{ 199, 396, 34 },
		{ 199, 397, 34 },
		{ 199, 398, 34 },
		{ 199, 399, 34 },
		{ 199, 400, 34 },
		{ 199, 401, 34 },
		{ 199, 402, 34 },
		{ 199, 403, 34 },
		{ 199, 404, 34 },
		{ 199, 405, 34 },
		{ 199, 406, 34 },
		{ 199, 407, 34 },
		{ 199, 408, 34 },
		{ 199, 409, 34 },
		{ 199, 410, 34 },
		{ 199, 411, 34 },
		{ 200, 349, 34 },
		{ 200, 350, 34 },
		{ 200, 351, 34 },
		{ 200, 352, 34 },
		{ 200, 353, 34 },
		{ 200, 354, 34 },
		{ 200, 355, 34 },
		{ 200, 356, 34 },
		{ 200, 357, 34 },
		{ 200, 358, 34 },
		{ 200, 359, 34 },
		{ 200, 360, 34 },
		{ 200, 361, 34 },
		{ 200, 362, 34 },
		{ 200, 363, 34 },
		{ 200, 364, 34 },
		{ 200, 365, 34 },
		{ 200, 366, 34 },
		{ 200, 367, 34 },
		{ 200, 368, 34 },
		{ 200, 369, 34 },
		{ 200, 370, 34 },
		{ 200, 371, 34 },
		{ 200, 372, 34 },
		{ 200, 373, 34 },
		{ 200, 374, 34 },
		{ 200, 375, 34 },
		{ 200, 376, 34 },
		{ 200, 377, 34 },
		{ 200, 378, 34 },
		{ 200, 379, 34 },
		{ 200, 380, 34 },
		{ 200, 381, 34 },
		{ 200, 382, 34 },
		{ 200, 383, 34 },
		{ 200, 384, 34 },
		{ 200, 385, 34 },
		{ 200, 386, 34 },
		{ 200, 387, 34 },
		{ 200, 388, 34 },
		{ 200, 389, 34 },
		{ 200, 390, 34 },
		{ 200, 391, 34 },
		{ 200, 392, 34 },
		{ 200, 393, 34 },
		{ 200, 394, 34 },
		{ 200, 395, 34 },
		{ 200, 396, 34 },
		{ 200, 397, 34 },
		{ 200, 398, 34 },
		{ 200, 399, 34 },
		{ 200, 400, 34 },
		{ 200, 401, 34 },
		{ 200, 402, 34 },
		{ 200, 403, 34 },
		{ 200, 404, 34 },
		{ 200, 405, 34 },
		{ 200, 406, 34 },
		{ 200, 407, 34 },
		{ 200, 408, 34 },
		{ 201, 347, 34 },
		{ 201, 348, 34 },
		{ 201, 349, 34 },
		{ 201, 350, 34 },
		{ 201, 351, 34 },
		{ 201, 352, 34 },
		{ 201, 353, 34 },
		{ 201, 354, 34 },
		{ 201, 355, 34 },
		{ 201, 356, 34 },
		{ 201, 357, 34 },
		{ 201, 358, 34 },
		{ 201, 359, 34 },
		{ 201, 360, 34 },
		{ 201, 361, 34 },
		{ 201, 362, 34 },
		{ 201, 363, 34 },
		{ 201, 364, 34 },
		{ 201, 365, 34 },
		{ 201, 366, 34 },
		{ 201, 367, 34 },
		{ 201, 368, 34 },
		{ 201, 369, 34 },
		{ 201, 370, 34 },
		{ 201, 371, 34 },
		{ 201, 372, 34 },
		{ 201, 373, 34 },
		{ 201, 374, 34 },
		{ 201, 375, 34 },
		{ 201, 376, 34 },
		{ 201, 377, 34 },
		{ 201, 378, 34 },
		{ 201, 379, 34 },
		{ 201, 380, 34 },
		{ 201, 381, 34 },
		{ 201, 382, 34 },
		{ 201, 383, 34 },
		{ 201, 384, 34 },
		{ 201, 385, 34 },
		{ 201, 386, 34 },
		{ 201, 387, 34 },
		{ 201, 388, 34 },
		{ 201, 389, 34 },
		{ 201, 390, 34 },
		{ 201, 391, 34 },
		{ 201, 392, 34 },
		{ 201, 393, 34 },
		{ 201, 394, 34 },
		{ 201, 395, 34 },
		{ 201, 396, 34 },
		{ 201, 397, 34 },
		{ 201, 398, 34 },
		{ 201, 399, 34 },
		{ 201, 400, 34 },
		{ 201, 401, 34 },
		{ 201, 402, 34 },
		{ 201, 403, 34 },
		{ 201, 404, 34 },
		{ 201, 405, 34 },
		{ 201, 406, 34 },
		{ 202, 346, 34 },
		{ 202, 347, 34 },
		{ 202, 348, 34 },
		{ 202, 349, 34 },
		{ 202, 350, 34 },
		{ 202, 351, 34 },
		{ 202, 352, 34 },
		{ 202, 353, 34 },
		{ 202, 354, 34 },
		{ 202, 355, 34 },
		{ 202, 356, 34 },
		{ 202, 357, 34 },
		{ 202, 358, 34 },
		{ 202, 359, 34 },
		{ 202, 360, 34 },
		{ 202, 361, 34 },
		{ 202, 362, 34 },
		{ 202, 363, 34 },
		{ 202, 364, 34 },
		{ 202, 365, 34 },
		{ 202, 366, 34 },
		{ 202, 367, 34 },
		{ 202, 368, 34 },
		{ 202, 369, 34 },
		{ 202, 370, 34 },
		{ 202, 371, 34 },
		{ 202, 372, 34 },
		{ 202, 373, 34 },
		{ 202, 374, 34 },
		{ 202, 375, 34 },
		{ 202, 376, 34 },
		{ 202, 377, 34 },
		{ 202, 378, 34 },
		{ 202, 379, 34 },
		{ 202, 380, 34 },
		{ 202, 381, 34 },
		{ 202, 382, 34 },
		{ 202, 383, 34 },
		{ 202, 384, 34 },
		{ 202, 385, 34 },
		{ 202, 386, 34 },
		{ 202, 387, 34 },
		{ 202, 388, 34 },
		{ 202, 389, 34 },
		{ 202, 390, 34 },
		{ 202, 391, 34 },
		{ 202, 392, 34 },
		{ 202, 393, 34 },
		{ 202, 394, 34 },
		{ 202, 395, 34 },
		{ 202, 396, 34 },
		{ 202, 397, 34 },
		{ 202, 398, 34 },
		{ 202, 399, 34 },
		{ 202, 400, 34 },
		{ 202, 401, 34 },
		{ 202, 402, 34 },
		{ 202, 403, 34 },
		{ 203, 344, 34 },
		{ 203, 345, 34 },
		{ 203, 346, 34 },
		{ 203, 347, 34 },
		{ 203, 348, 34 },
		{ 203, 349, 34 },
		{ 203, 350, 34 },
		{ 203, 351, 34 },
		{ 203, 352, 34 },
		{ 203, 353, 34 },
		{ 203, 354, 34 },
		{ 203, 355, 34 },
		{ 203, 356, 34 },
		{ 203, 357, 34 },
		{ 203, 358, 34 },
		{ 203, 359, 34 },
		{ 203, 360, 34 },
		{ 203, 361, 34 },
		{ 203, 362, 34 },
		{ 203, 363, 34 },
		{ 203, 364, 34 },
		{ 203, 365, 34 },
		{ 203, 366, 34 },
		{ 203, 367, 34 },
		{ 203, 368, 34 },
		{ 203, 369, 34 },
		{ 203, 370, 34 },
		{ 203, 371, 34 },
		{ 203, 372, 34 },
		{ 203, 373, 34 },
		{ 203, 374, 34 },
		{ 203, 375, 34 },
		{ 203, 376, 34 },
		{ 203, 377, 34 },
		{ 203, 378, 34 },
		{ 203, 379, 34 },
		{ 203, 380, 34 },
		{ 203, 381, 34 },
		{ 203, 382, 34 },
		{ 203, 383, 34 },
		{ 203, 384, 34 },
		{ 203, 385, 34 },
		{ 203, 386, 34 },
		{ 203, 387, 34 },
		{ 203, 388, 34 },
		{ 203, 389, 34 },
		{ 203, 390, 34 },
		{ 203, 391, 34 },
		{ 203, 392, 34 },
		{ 203, 393, 34 },
		{ 203, 394, 34 },
		{ 203, 395, 34 },
		{ 203, 396, 34 },
		{ 203, 397, 34 },
		{ 203, 398, 34 },
		{ 203, 399, 34 },
		{ 203, 400, 34 },
		{ 204, 342, 34 },
		{ 204, 343, 34 },
		{ 204, 344, 34 },
		{ 204, 345, 34 },
		{ 204, 346, 34 },
		{ 204, 347, 34 },
		{ 204, 348, 34 },
		{ 204, 349, 34 },
		{ 204, 350, 34 },
		{ 204, 351, 34 },
		{ 204, 352, 34 },
		{ 204, 353, 34 },
		{ 204, 354, 34 },
		{ 204, 355, 34 },
		{ 204, 356, 34 },
		{ 204, 357, 34 },
		{ 204, 358, 34 },
		{ 204, 359, 34 },
		{ 204, 360, 34 },
		{ 204, 361, 34 },
		{ 204, 362, 34 },
		{ 204, 363, 34 },
		{ 204, 364, 34 },
		{ 204, 365, 34 },
		{ 204, 366, 34 },
		{ 204, 367, 34 },
		{ 204, 368, 34 },
		{ 204, 369, 34 },
		{ 204, 370, 34 },
		{ 204, 371, 34 },
		{ 204, 372, 34 },
		{ 204, 373, 34 },
		{ 204, 374, 34 },
		{ 204, 375, 34 },
		{ 204, 376, 34 },
		{ 204, 377, 34 },
		{ 204, 378, 34 },
		{ 204, 379, 34 },
		{ 204, 380, 34 },
		{ 204, 381, 34 },
		{ 204, 382, 34 },
		{ 204, 383, 34 },
		{ 204, 384, 34 },
		{ 204, 385, 34 },
		{ 204, 386, 34 },
		{ 204, 387, 34 },
		{ 204, 388, 34 },
		{ 204, 389, 34 },
		{ 204, 390, 34 },
		{ 204, 391, 34 },
		{ 204, 392, 34 },
		{ 204, 393, 34 },
		{ 204, 394, 34 },
		{ 204, 395, 34 },
		{ 204, 396, 34 },
		{ 204, 397, 34 },
		{ 204, 398, 34 },
		{ 205, 341, 34 },
		{ 205, 342, 34 },
		{ 205, 343, 34 },
		{ 205, 344, 34 },
		{ 205, 345, 34 },
		{ 205, 346, 34 },
		{ 205, 347, 34 },
		{ 205, 348, 34 },
		{ 205, 349, 34 },
		{ 205, 350, 34 },
		{ 205, 351, 34 },
		{ 205, 352, 34 },
		{ 205, 353, 34 },
		{ 205, 354, 34 },
		{ 205, 355, 34 },
		{ 205, 356, 34 },
		{ 205, 357, 34 },
		{ 205, 358, 34 },
		{ 205, 359, 34 },
		{ 205, 360, 34 },
		{ 205, 361, 34 },
		{ 205, 362, 34 },
		{ 205, 363, 34 },
		{ 205, 364, 34 },
		{ 205, 365, 34 },
		{ 205, 366, 34 },
		{ 205, 367, 34 },
		{ 205, 368, 34 },
		{ 205, 369, 34 },
		{ 205, 370, 34 },
		{ 205, 371, 34 },
		{ 205, 372, 34 },
		{ 205, 373, 34 },
		{ 205, 374, 34 },
		{ 205, 375, 34 },
		{ 205, 376, 34 },
		{ 205, 377, 34 },
		{ 205, 378, 34 },
		{ 205, 379, 34 },
		{ 205, 380, 34 },
		{ 205, 381, 34 },
		{ 205, 382, 34 },
		{ 205, 383, 34 },
		{ 205, 384, 34 },
		{ 205, 385, 34 },
		{ 205, 386, 34 },
		{ 205, 387, 34 },
		{ 205, 388, 34 },
		{ 205, 389, 34 },
		{ 205, 390, 34 },
		{ 205, 391, 34 },
		{ 205, 392, 34 },
		{ 205, 393, 34 },
		{ 205, 394, 34 },
		{ 205, 395, 34 },
		{ 206, 339, 34 },
		{ 206, 340, 34 },
		{ 206, 341, 34 },
		{ 206, 342, 34 },
		{ 206, 343, 34 },
		{ 206, 344, 34 },
		{ 206, 345, 34 },
		{ 206, 346, 34 },
		{ 206, 347, 34 },
		{ 206, 348, 34 },
		{ 206, 349, 34 },
		{ 206, 350, 34 },
		{ 206, 351, 34 },
		{ 206, 352, 34 },
		{ 206, 353, 34 },
		{ 206, 354, 34 },
		{ 206, 355, 34 },
		{ 206, 356, 34 },
		{ 206, 357, 34 },
		{ 206, 358, 34 },
		{ 206, 359, 34 },
		{ 206, 360, 34 },
		{ 206, 361, 34 },
		{ 206, 362, 34 },
		{ 206, 363, 34 },
		{ 206, 364, 34 },
		{ 206, 365, 34 },
		{ 206, 366, 34 },
		{ 206, 367, 34 },
		{ 206, 368, 34 },
		{ 206, 369, 34 },
		{ 206, 370, 34 },
		{ 206, 371, 34 },
		{ 206, 372, 34 },
		{ 206, 373, 34 },
		{ 206, 374, 34 },
		{ 206, 375, 34 },
		{ 206, 376, 34 },
		{ 206, 377, 34 },
		{ 206, 378, 34 },
		{ 206, 379, 34 },
		{ 206, 380, 34 },
		{ 206, 381, 34 },
		{ 206, 382, 34 },
		{ 206, 383, 34 },
		{ 206, 384, 34 },
		{ 206, 385, 34 },
		{ 206, 386, 34 },
		{ 206, 387, 34 },
		{ 206, 388, 34 },
		{ 206, 389, 34 },
		{ 206, 390, 34 },
		{ 206, 391, 34 },
		{ 206, 392, 34 },
		{ 207, 338, 34 },
		{ 207, 339, 34 },
		{ 207, 340, 34 },
		{ 207, 341, 34 },
		{ 207, 342, 34 },
		{ 207, 343, 34 },
		{ 207, 344, 34 },
		{ 207, 345, 34 },
		{ 207, 346, 34 },
		{ 207, 347, 34 },
		{ 207, 348, 34 },
		{ 207, 349, 34 },
		{ 207, 350, 34 },
		{ 207, 351, 34 },
		{ 207, 352, 34 },
		{ 207, 353, 34 },
		{ 207, 354, 34 },
		{ 207, 355, 34 },
		{ 207, 356, 34 },
		{ 207, 357, 34 },
		{ 207, 358, 34 },
		{ 207, 359, 34 },
		{ 207, 360, 34 },
		{ 207, 361, 34 },
		{ 207, 362, 34 },
		{ 207, 363, 34 },
		{ 207, 364, 34 },
		{ 207, 365, 34 },
		{ 207, 366, 34 },
		{ 207, 367, 34 },
		{ 207, 368, 34 },
		{ 207, 369, 34 },
		{ 207, 370, 34 },
		{ 207, 371, 34 },
		{ 207, 372, 34 },
		{ 207, 373, 34 },
		{ 207, 374, 34 },
		{ 207, 375, 34 },
		{ 207, 376, 34 },
		{ 207, 377, 34 },
		{ 207, 378, 34 },
		{ 207, 379, 34 },
		{ 207, 380, 34 },
		{ 207, 381, 34 },
		{ 207, 382, 34 },
		{ 207, 383, 34 },
		{ 207, 384, 34 },
		{ 207, 385, 34 },
		{ 207, 386, 34 },
		{ 207, 387, 34 },
		{ 207, 388, 34 },
		{ 207, 389, 34 },
		{ 207, 390, 34 },
		{ 208, 336, 34 },
		{ 208, 337, 34 },
		{ 208, 338, 34 },
		{ 208, 339, 34 },
		{ 208, 340, 34 },
		{ 208, 341, 34 },
		{ 208, 342, 34 },
		{ 208, 343, 34 },
		{ 208, 344, 34 },
		{ 208, 345, 34 },
		{ 208, 346, 34 },
		{ 208, 347, 34 },
		{ 208, 348, 34 },
		{ 208, 349, 34 },
		{ 208, 350, 34 },
		{ 208, 351, 34 },
		{ 208, 352, 34 },
		{ 208, 353, 34 },
		{ 208, 354, 34 },
		{ 208, 355, 34 },
		{ 208, 356, 34 },
		{ 208, 357, 34 },
		{ 208, 358, 34 },
		{ 208, 359, 34 },
		{ 208, 360, 34 },
		{ 208, 361, 34 },
		{ 208, 362, 34 },
		{ 208, 363, 34 },
		{ 208, 364, 34 },
		{ 208, 365, 34 },
		{ 208, 366, 34 },
		{ 208, 367, 34 },
		{ 208, 368, 34 },
		{ 208, 369, 34 },
		{ 208, 370, 34 },
		{ 208, 371, 34 },
		{ 208, 372, 34 },
		{ 208, 373, 34 },
		{ 208, 374, 34 },
		{ 208, 375, 34 },
		{ 208, 376, 34 },
		{ 208, 377, 34 },
		{ 208, 378, 34 },
		{ 208, 379, 34 },
		{ 208, 380, 34 },
		{ 208, 381, 34 },
		{ 208, 382, 34 },
		{ 208, 383, 34 },
		{ 208, 384, 34 },
		{ 208, 385, 34 },
		{ 208, 386, 34 },
		{ 208, 387, 34 },
		{ 209, 334, 34 },
		{ 209, 335, 34 },
		{ 209, 336, 34 },
		{ 209, 337, 34 },
		{ 209, 338, 34 },
		{ 209, 339, 34 },
		{ 209, 340, 34 },
		{ 209, 341, 34 },
		{ 209, 342, 34 },
		{ 209, 343, 34 },
		{ 209, 344, 34 },
		{ 209, 345, 34 },
		{ 209, 346, 34 },
		{ 209, 347, 34 },
		{ 209, 348, 34 },
		{ 209, 349, 34 },
		{ 209, 350, 34 },
		{ 209, 351, 34 },
		{ 209, 352, 34 },
		{ 209, 353, 34 },
		{ 209, 354, 34 },
		{ 209, 355, 34 },
		{ 209, 356, 34 },
		{ 209, 357, 34 },
		{ 209, 358, 34 },
		{ 209, 359, 34 },
		{ 209, 360, 34 },
		{ 209, 361, 34 },
		{ 209, 362, 34 },
		{ 209, 363, 34 },
		{ 209, 364, 34 },
		{ 209, 365, 34 },
		{ 209, 366, 34 },
		{ 209, 367, 34 },
		{ 209, 368, 34 },
		{ 209, 369, 34 },
		{ 209, 370, 34 },
		{ 209, 371, 34 },
		{ 209, 372, 34 },
		{ 209, 373, 34 },
		{ 209, 374, 34 },
		{ 209, 375, 34 },
		{ 209, 376, 34 },
		{ 209, 377, 34 },
		{ 209, 378, 34 },
		{ 209, 379, 34 },
		{ 209, 380, 34 },
		{ 209, 381, 34 },
		{ 209, 382, 34 },
		{ 209, 383, 34 },
		{ 209, 384, 34 },
		{ 210, 333, 34 },
		{ 210, 334, 34 },
		{ 210, 335, 34 },
		{ 210, 336, 34 },
		{ 210, 337, 34 },
		{ 210, 338, 34 },
		{ 210, 339, 34 },
		{ 210, 340, 34 },
		{ 210, 341, 34 },
		{ 210, 342, 34 },
		{ 210, 343, 34 },
		{ 210, 344, 34 },
		{ 210, 345, 34 },
		{ 210, 346, 34 },
		{ 210, 347, 34 },
		{ 210, 348, 34 },
		{ 210, 349, 34 },
		{ 210, 350, 34 },
		{ 210, 351, 34 },
		{ 210, 352, 34 },
		{ 210, 353, 34 },
		{ 210, 354, 34 },
		{ 210, 355, 34 },
		{ 210, 356, 34 },
		{ 210, 357, 34 },
		{ 210, 358, 34 },
		{ 210, 359, 34 },
		{ 210, 360, 34 },
		{ 210, 361, 34 },
		{ 210, 362, 34 },
		{ 210, 363, 34 },
		{ 210, 364, 34 },
		{ 210, 365, 34 },
		{ 210, 366, 34 },
		{ 210, 367, 34 },
		{ 210, 368, 34 },
		{ 210, 369, 34 },
		{ 210, 370, 34 },
		{ 210, 371, 34 },
		{ 210, 372, 34 },
		{ 210, 373, 34 },
		{ 210, 374, 34 },
		{ 210, 375, 34 },
		{ 210, 376, 34 },
		{ 210, 377, 34 },
		{ 210, 378, 34 },
		{ 210, 379, 34 },
		{ 210, 380, 34 },
		{ 210, 381, 34 },
		{ 210, 382, 34 },
		{ 211, 331, 34 },
		{ 211, 332, 34 },
		{ 211, 333, 34 },
		{ 211, 334, 34 },
		{ 211, 335, 34 },
		{ 211, 336, 34 },
		{ 211, 337, 34 },
		{ 211, 338, 34 },
		{ 211, 339, 34 },
		{ 211, 340, 34 },
		{ 211, 341, 34 },
		{ 211, 342, 34 },
		{ 211, 343, 34 },
		{ 211, 344, 34 },
		{ 211, 345, 34 },
		{ 211, 346, 34 },
		{ 211, 347, 34 },
		{ 211, 348, 34 },
		{ 211, 349, 34 },
		{ 211, 350, 34 },
		{ 211, 351, 34 },
		{ 211, 352, 34 },
		{ 211, 353, 34 },
		{ 211, 354, 34 },
		{ 211, 355, 34 },
		{ 211, 356, 34 },
		{ 211, 357, 34 },
		{ 211, 358, 34 },
		{ 211, 359, 34 },
		{ 211, 360, 34 },
		{ 211, 361, 34 },
		{ 211, 362, 34 },
		{ 211, 363, 34 },
		{ 211, 364, 34 },
		{ 211, 365, 34 },
		{ 211, 366, 34 },
		{ 211, 367, 34 },
		{ 211, 368, 34 },
		{ 211, 369, 34 },
		{ 211, 370, 34 },
		{ 211, 371, 34 },
		{ 211, 372, 34 },
		{ 211, 373, 34 },
		{ 211, 374, 34 },
		{ 211, 375, 34 },
		{ 211, 376, 34 },
		{ 211, 377, 34 },
		{ 211, 378, 34 },
		{ 211, 379, 34 },
		{ 212, 330, 34 },
		{ 212, 331, 34 },
		{ 212, 332, 34 },
		{ 212, 333, 34 },
		{ 212, 334, 34 },
		{ 212, 335, 34 },
		{ 212, 336, 34 },
		{ 212, 337, 34 },
		{ 212, 338, 34 },
		{ 212, 339, 34 },
		{ 212, 340, 34 },
		{ 212, 341, 34 },
		{ 212, 342, 34 },
		{ 212, 343, 34 },
		{ 212, 344, 34 },
		{ 212, 345, 34 },
		{ 212, 346, 34 },
		{ 212, 347, 34 },
		{ 212, 348, 34 },
		{ 212, 349, 34 },
		{ 212, 350, 34 },
		{ 212, 351, 34 },
		{ 212, 352, 34 },
		{ 212, 353, 34 },
		{ 212, 354, 34 },
		{ 212, 355, 34 },
		{ 212, 356, 34 },
		{ 212, 357, 34 },
		{ 212, 358, 34 },
		{ 212, 359, 34 },
		{ 212, 360, 34 },
		{ 212, 361, 34 },
		{ 212, 362, 34 },
		{ 212, 363, 34 },
		{ 212, 364, 34 },
		{ 212, 365, 34 },
		{ 212, 366, 34 },
		{ 212, 367, 34 },
		{ 212, 368, 34 },
		{ 212, 369, 34 },
		{ 212, 370, 34 },
		{ 212, 371, 34 },
		{ 212, 372, 34 },
		{ 212, 373, 34 },
		{ 212, 374, 34 },
		{ 212, 375, 34 },
		{ 212, 376, 34 },
		{ 213, 328, 34 },
		{ 213, 329, 34 },
		{ 213, 330, 34 },
		{ 213, 331, 34 },
		{ 213, 332, 34 },
		{ 213, 333, 34 },
		{ 213, 334, 34 },
		{ 213, 335, 34 },
		{ 213, 336, 34 },
		{ 213, 337, 34 },
		{ 213, 338, 34 },
		{ 213, 339, 34 },
		{ 213, 340, 34 },
		{ 213, 341, 34 },
		{ 213, 342, 34 },
		{ 213, 343, 34 },
		{ 213, 344, 34 },
		{ 213, 345, 34 },
		{ 213, 346, 34 },
		{ 213, 347, 34 },
		{ 213, 348, 34 },
		{ 213, 349, 34 },
		{ 213, 350, 34 },
		{ 213, 351, 34 },
		{ 213, 352, 34 },
		{ 213, 353, 34 },
		{ 213, 354, 34 },
		{ 213, 355, 34 },
		{ 213, 356, 34 },
		{ 213, 357, 34 },
		{ 213, 358, 34 },
		{ 213, 359, 34 },
		{ 213, 360, 34 },
		{ 213, 361, 34 },
		{ 213, 362, 34 },
		{ 213, 363, 34 },
		{ 213, 364, 34 },
		{ 213, 365, 34 },
		{ 213, 366, 34 },
		{ 213, 367, 34 },
		{ 213, 368, 34 },
		{ 213, 369, 34 },
		{ 213, 370, 34 },
		{ 213, 371, 34 },
		{ 213, 372, 34 },
		{ 213, 373, 34 },
		{ 213, 374, 34 },
		{ 214, 327, 34 },
		{ 214, 328, 34 },
		{ 214, 329, 34 },
		{ 214, 330, 34 },
		{ 214, 331, 34 },
		{ 214, 332, 34 },
		{ 214, 333, 34 },
		{ 214, 334, 34 },
		{ 214, 335, 34 },
		{ 214, 336, 34 },
		{ 214, 337, 34 },
		{ 214, 338, 34 },
		{ 214, 339, 34 },
		{ 214, 340, 34 },
		{ 214, 341, 34 },
		{ 214, 342, 34 },
		{ 214, 343, 34 },
		{ 214, 344, 34 },
		{ 214, 345, 34 },
		{ 214, 346, 34 },
		{ 214, 347, 34 },
		{ 214, 348, 34 },
		{ 214, 349, 34 },
		{ 214, 350, 34 },
		{ 214, 351, 34 },
		{ 214, 352, 34 },
		{ 214, 353, 34 },
		{ 214, 354, 34 },
		{ 214, 355, 34 },
		{ 214, 356, 34 },
		{ 214, 357, 34 },
		{ 214, 358, 34 },
		{ 214, 359, 34 },
		{ 214, 360, 34 },
		{ 214, 361, 34 },
		{ 214, 362, 34 },
		{ 214, 363, 34 },
		{ 214, 364, 34 },
		{ 214, 365, 34 },
		{ 214, 366, 34 },
		{ 214, 367, 34 },
		{ 214, 368, 34 },
		{ 214, 369, 34 },
		{ 214, 370, 34 },
		{ 214, 371, 34 },
		{ 215, 325, 34 },
		{ 215, 326, 34 },
		{ 215, 327, 34 },
		{ 215, 328, 34 },
		{ 215, 329, 34 },
		{ 215, 330, 34 },
		{ 215, 331, 34 },
		{ 215, 332, 34 },
		{ 215, 333, 34 },
		{ 215, 334, 34 },
		{ 215, 335, 34 },
		{ 215, 336, 34 },
		{ 215, 337, 34 },
		{ 215, 338, 34 },
		{ 215, 339, 34 },
		{ 215, 340, 34 },
		{ 215, 341, 34 },
		{ 215, 342, 34 },
		{ 215, 343, 34 },
		{ 215, 344, 34 },
		{ 215, 345, 34 },
		{ 215, 346, 34 },
		{ 215, 347, 34 },
		{ 215, 348, 34 },
		{ 215, 349, 34 },
		{ 215, 350, 34 },
		{ 215, 351, 34 },
		{ 215, 352, 34 },
		{ 215, 353, 34 },
		{ 215, 354, 34 },
		{ 215, 355, 34 },
		{ 215, 356, 34 },
		{ 215, 357, 34 },
		{ 215, 358, 34 },
		{ 215, 359, 34 },
		{ 215, 360, 34 },
		{ 215, 361, 34 },
		{ 215, 362, 34 },
		{ 215, 363, 34 },
		{ 215, 364, 34 },
		{ 215, 365, 34 },
		{ 215, 366, 34 },
		{ 215, 367, 34 },
		{ 215, 368, 34 },
		{ 216, 324, 34 },
		{ 216, 325, 34 },
		{ 216, 326, 34 },
		{ 216, 327, 34 },
		{ 216, 328, 34 },
		{ 216, 329, 34 },
		{ 216, 330, 34 },
		{ 216, 331, 34 },
		{ 216, 332, 34 },
		{ 216, 333, 34 },
		{ 216, 334, 34 },
		{ 216, 335, 34 },
		{ 216, 336, 34 },
		{ 216, 337, 34 },
		{ 216, 338, 34 },
		{ 216, 339, 34 },
		{ 216, 340, 34 },
		{ 216, 341, 34 },
		{ 216, 342, 34 },
		{ 216, 343, 34 },
		{ 216, 344, 34 },
		{ 216, 345, 34 },
		{ 216, 346, 34 },
		{ 216, 347, 34 },
		{ 216, 348, 34 },
		{ 216, 349, 34 },
		{ 216, 350, 34 },
		{ 216, 351, 34 },
		{ 216, 352, 34 },
		{ 216, 353, 34 },
		{ 216, 354, 34 },
		{ 216, 355, 34 },
		{ 216, 356, 34 },
		{ 216, 357, 34 },
		{ 216, 358, 34 },
		{ 216, 359, 34 },
		{ 216, 360, 34 },
		{ 216, 361, 34 },
		{ 216, 362, 34 },
		{ 216, 363, 34 },
		{ 216, 364, 34 },
		{ 216, 365, 34 },
		{ 216, 366, 34 },
		{ 217, 324, 34 },
		{ 217, 325, 34 },
		{ 217, 326, 34 },
		{ 217, 327, 34 },
		{ 217, 328, 34 },
		{ 217, 329, 34 },
		{ 217, 330, 34 },
		{ 217, 331, 34 },
		{ 217, 332, 34 },
		{ 217, 333, 34 },
		{ 217, 334, 34 },
		{ 217, 335, 34 },
		{ 217, 336, 34 },
		{ 217, 337, 34 },
		{ 217, 338, 34 },
		{ 217, 339, 34 },
		{ 217, 340, 34 },
		{ 217, 341, 34 },
		{ 217, 342, 34 },
		{ 217, 343, 34 },
		{ 217, 344, 34 },
		{ 217, 345, 34 },
		{ 217, 346, 34 },
		{ 217, 347, 34 },
		{ 217, 348, 34 },
		{ 217, 349, 34 },
		{ 217, 350, 34 },
		{ 217, 351, 34 },
		{ 217, 352, 34 },
		{ 217, 353, 34 },
		{ 217, 354, 34 },
		{ 217, 355, 34 },
		{ 217, 356, 34 },
		{ 217, 357, 34 },
		{ 217, 358, 34 },
		{ 217, 359, 34 },
		{ 217, 360, 34 },
		{ 217, 361, 34 },
		{ 217, 362, 34 },
		{ 217, 363, 34 },
		{ 218, 323, 34 },
		{ 218, 324, 34 },
		{ 218, 325, 34 },
		{ 218, 326, 34 },
		{ 218, 327, 34 },
		{ 218, 328, 34 },
		{ 218, 329, 34 },
		{ 218, 330, 34 },
		{ 218, 331, 34 },
		{ 218, 332, 34 },
		{ 218, 333, 34 },
		{ 218, 334, 34 },
		{ 218, 335, 34 },
		{ 218, 336, 34 },
		{ 218, 337, 34 },
		{ 218, 338, 34 },
		{ 218, 339, 34 },
		{ 218, 340, 34 },
		{ 218, 341, 34 },
		{ 218, 342, 34 },
		{ 218, 343, 34 },
		{ 218, 344, 34 },
		{ 218, 345, 34 },
		{ 218, 346, 34 },
		{ 218, 347, 34 },
		{ 218, 348, 34 },
		{ 218, 349, 34 },
		{ 218, 350, 34 },
		{ 218, 351, 34 },
		{ 218, 352, 34 },
		{ 218, 353, 34 },
		{ 218, 354, 34 },
		{ 218, 355, 34 },
		{ 218, 356, 34 },
		{ 218, 357, 34 },
		{ 218, 358, 34 },
		{ 218, 359, 34 },
		{ 218, 360, 34 },
		{ 219, 323, 34 },
		{ 219, 324, 34 },
		{ 219, 325, 34 },
		{ 219, 326, 34 },
		{ 219, 327, 34 },
		{ 219, 328, 34 },
		{ 219, 329, 34 },
		{ 219, 330, 34 },
		{ 219, 331, 34 },
		{ 219, 332, 34 },
		{ 219, 333, 34 },
		{ 219, 334, 34 },
		{ 219, 335, 34 },
		{ 219, 336, 34 },
		{ 219, 337, 34 },
		{ 219, 338, 34 },
		{ 219, 339, 34 },
		{ 219, 340, 34 },
		{ 219, 341, 34 },
		{ 219, 342, 34 },
		{ 219, 343, 34 },
		{ 219, 344, 34 },
		{ 219, 345, 34 },
		{ 219, 346, 34 },
		{ 219, 347, 34 },
		{ 219, 348, 34 },
		{ 219, 349, 34 },
		{ 219, 350, 34 },
		{ 219, 351, 34 },
		{ 219, 352, 34 },
		{ 219, 353, 34 },
		{ 219, 354, 34 },
		{ 219, 355, 34 },
		{ 219, 356, 34 },
		{ 219, 357, 34 },
		{ 219, 358, 34 },
		{ 220, 323, 34 },
		{ 220, 324, 34 },
		{ 220, 325, 34 },
		{ 220, 326, 34 },
		{ 220, 327, 34 },
		{ 220, 328, 34 },
		{ 220, 329, 34 },
		{ 220, 330, 34 },
		{ 220, 331, 34 },
		{ 220, 332, 34 },
		{ 220, 333, 34 },
		{ 220, 334, 34 },
		{ 220, 335, 34 },
		{ 220, 336, 34 },
		{ 220, 337, 34 },
		{ 220, 338, 34 },
		{ 220, 339, 34 },
		{ 220, 340, 34 },
		{ 220, 341, 34 },
		{ 220, 342, 34 },
		{ 220, 343, 34 },
		{ 220, 344, 34 },
		{ 220, 345, 34 },
		{ 220, 346, 34 },
		{ 220, 347, 34 },
		{ 220, 348, 34 },
		{ 220, 349, 34 },
		{ 220, 350, 34 },
		{ 220, 351, 34 },
		{ 220, 352, 34 },
		{ 220, 353, 34 },
		{ 220, 354, 34 },
		{ 220, 355, 34 },
		{ 221, 323, 34 },
		{ 221, 324, 34 },
		{ 221, 325, 34 },
		{ 221, 326, 34 },
		{ 221, 327, 34 },
		{ 221, 328, 34 },
		{ 221, 329, 34 },
		{ 221, 330, 34 },
		{ 221, 331, 34 },
		{ 221, 332, 34 },
		{ 221, 333, 34 },
		{ 221, 334, 34 },
		{ 221, 335, 34 },
		{ 221, 336, 34 },
		{ 221, 337, 34 },
		{ 221, 338, 34 },
		{ 221, 339, 34 },
		{ 221, 340, 34 },
		{ 221, 341, 34 },
		{ 221, 342, 34 },
		{ 221, 343, 34 },
		{ 221, 344, 34 },
		{ 221, 345, 34 },
		{ 221, 346, 34 },
		{ 221, 347, 34 },
		{ 221, 348, 34 },
		{ 221, 349, 34 },
		{ 221, 350, 34 },
		{ 221, 351, 34 },
		{ 221, 352, 34 },
		{ 222, 323, 34 },
		{ 222, 324, 34 },
		{ 222, 325, 34 },
		{ 222, 326, 34 },
		{ 222, 327, 34 },
		{ 222, 328, 34 },
		{ 222, 329, 34 },
		{ 222, 330, 34 },
		{ 222, 331, 34 },
		{ 222, 332, 34 },
		{ 222, 333, 34 },
		{ 222, 334, 34 },
		{ 222, 335, 34 },
		{ 222, 336, 34 },
		{ 222, 337, 34 },
		{ 222, 338, 34 },
		{ 222, 339, 34 },
		{ 222, 340, 34 },
		{ 222, 341, 34 },
		{ 222, 342, 34 },
		{ 222, 343, 34 },
		{ 222, 344, 34 },
		{ 222, 345, 34 },
		{ 222, 346, 34 },
		{ 222, 347, 34 },
		{ 222, 348, 34 },
		{ 222, 349, 34 },
		{ 222, 350, 34 },
		{ 223, 323, 34 },
		{ 223, 324, 34 },
		{ 223, 325, 34 },
		{ 223, 326, 34 },
		{ 223, 327, 34 },
		{ 223, 328, 34 },
		{ 223, 329, 34 },
		{ 223, 330, 34 },
		{ 223, 331, 34 },
		{ 223, 332, 34 },
		{ 223, 333, 34 },
		{ 223, 334, 34 },
		{ 223, 335, 34 },
		{ 223, 336, 34 },
		{ 223, 337, 34 },
		{ 223, 338, 34 },
		{ 223, 339, 34 },
		{ 223, 340, 34 },
		{ 223, 341, 34 },
		{ 223, 342, 34 },
		{ 223, 343, 34 },
		{ 223, 344, 34 },
		{ 223, 345, 34 },
		{ 223, 346, 34 },
		{ 223, 347, 34 },
		{ 224, 323, 34 },
		{ 224, 324, 34 },
		{ 224, 325, 34 },
		{ 224, 326, 34 },
		{ 224, 327, 34 },
		{ 224, 328, 34 },
		{ 224, 329, 34 },
		{ 224, 330, 34 },
		{ 224, 331, 34 },
		{ 224, 332, 34 },
		{ 224, 333, 34 },
		{ 224, 334, 34 },
		{ 224, 335, 34 },
		{ 224, 336, 34 },
		{ 224, 337, 34 },
		{ 224, 338, 34 },
		{ 224, 339, 34 },
		{ 224, 340, 34 },
		{ 224, 341, 34 },
		{ 224, 342, 34 },
		{ 224, 343, 34 },
		{ 224, 344, 34 },
		{ 225, 323, 34 },
		{ 225, 324, 34 },
		{ 225, 325, 34 },
		{ 225, 326, 34 },
		{ 225, 327, 34 },
		{ 225, 328, 34 },
		{ 225, 329, 34 },
		{ 225, 330, 34 },
		{ 225, 331, 34 },
		{ 225, 332, 34 },
		{ 225, 333, 34 },
		{ 225, 334, 34 },
		{ 225, 335, 34 },
		{ 225, 336, 34 },
		{ 225, 337, 34 },
		{ 225, 338, 34 },
		{ 225, 339, 34 },
		{ 225, 340, 34 },
		{ 225, 341, 34 },
		{ 225, 342, 34 },
		{ 226, 323, 34 },
		{ 226, 324, 34 },
		{ 226, 325, 34 },
		{ 226, 326, 34 },
		{ 226, 327, 34 },
		{ 226, 328, 34 },
		{ 226, 329, 34 },
		{ 226, 330, 34 },
		{ 226, 331, 34 },
		{ 226, 332, 34 },
		{ 226, 333, 34 },
		{ 226, 334, 34 },
		{ 226, 335, 34 },
		{ 226, 336, 34 },
		{ 226, 337, 34 },
		{ 226, 338, 34 },
		{ 226, 339, 34 },
		{ 227, 323, 34 },
		{ 227, 324, 34 },
		{ 227, 325, 34 },
		{ 227, 326, 34 },
		{ 227, 327, 34 },
		{ 227, 328, 34 },
		{ 227, 329, 34 },
		{ 227, 330, 34 },
		{ 227, 331, 34 },
		{ 227, 332, 34 },
		{ 227, 333, 34 },
		{ 227, 334, 34 },
		{ 227, 335, 34 },
		{ 227, 336, 34 },
		{ 228, 323, 34 },
		{ 228, 324, 34 },
		{ 228, 325, 34 },
		{ 228, 326, 34 },
		{ 228, 327, 34 },
		{ 228, 328, 34 },
		{ 228, 329, 34 },
		{ 228, 330, 34 },
		{ 228, 331, 34 },
		{ 228, 332, 34 },
		{ 228, 333, 34 },
		{ 228, 334, 34 },
		{ 229, 323, 34 },
		{ 229, 324, 34 },
		{ 229, 325, 34 },
		{ 229, 326, 34 },
		{ 229, 327, 34 },
		{ 229, 328, 34 },
		{ 229, 329, 34 },
		{ 229, 330, 34 },
		{ 229, 331, 34 },
		{ 230, 323, 34 },
		{ 230, 324, 34 },
		{ 230, 325, 34 },
		{ 230, 326, 34 },
		{ 230, 327, 34 },
		{ 230, 328, 34 },
		{ 231, 324, 34 },
		{ 231, 325, 34 },
		{ 283, 189, 34 },
		{ 283, 190, 34 },
		{ 284, 186, 34 },
		{ 284, 187, 34 },
		{ 284, 188, 34 },
		{ 284, 189, 34 },
		{ 284, 190, 34 },
		{ 284, 191, 34 },
		{ 285, 183, 34 },
		{ 285, 184, 34 },
		{ 285, 185, 34 },
		{ 285, 186, 34 },
		{ 285, 187, 34 },
		{ 285, 188, 34 },
		{ 285, 189, 34 },
		{ 285, 190, 34 },
		{ 285, 191, 34 },
		{ 286, 180, 34 },
		{ 286, 181, 34 },
		{ 286, 182, 34 },
		{ 286, 183, 34 },
		{ 286, 184, 34 },
		{ 286, 185, 34 },
		{ 286, 186, 34 },
		{ 286, 187, 34 },
		{ 286, 188, 34 },
		{ 286, 189, 34 },
		{ 286, 190, 34 },
		{ 286, 191, 34 },
		{ 287, 178, 34 },
		{ 287, 179, 34 },
		{ 287, 180, 34 },
		{ 287, 181, 34 },
		{ 287, 182, 34 },
		{ 287, 183, 34 },
		{ 287, 184, 34 },
		{ 287, 185, 34 },
		{ 287, 186, 34 },
		{ 287, 187, 34 },
		{ 287, 188, 34 },
		{ 287, 189, 34 },
		{ 287, 190, 34 },
		{ 287, 191, 34 },
		{ 288, 175, 34 },
		{ 288, 176, 34 },
		{ 288, 177, 34 },
		{ 288, 178, 34 },
		{ 288, 179, 34 },
		{ 288, 180, 34 },
		{ 288, 181, 34 },
		{ 288, 182, 34 },
		{ 288, 183, 34 },
		{ 288, 184, 34 },
		{ 288, 185, 34 },
		{ 288, 186, 34 },
		{ 288, 187, 34 },
		{ 288, 188, 34 },
		{ 288, 189, 34 },
		{ 288, 190, 34 },
		{ 288, 191, 34 },
		{ 289, 172, 34 },
		{ 289, 173, 34 },
		{ 289, 174, 34 },
		{ 289, 175, 34 },
		{ 289, 176, 34 },
		{ 289, 177, 34 },
		{ 289, 178, 34 },
		{ 289, 179, 34 },
		{ 289, 180, 34 },
		{ 289, 181, 34 },
		{ 289, 182, 34 },
		{ 289, 183, 34 },
		{ 289, 184, 34 },
		{ 289, 185, 34 },
		{ 289, 186, 34 },
		{ 289, 187, 34 },
		{ 289, 188, 34 },
		{ 289, 189, 34 },
		{ 289, 190, 34 },
		{ 289, 191, 34 },
		{ 290, 170, 34 },
		{ 290, 171, 34 },
		{ 290, 172, 34 },
		{ 290, 173, 34 },
		{ 290, 174, 34 },
		{ 290, 175, 34 },
		{ 290, 176, 34 },
		{ 290, 177, 34 },
		{ 290, 178, 34 },
		{ 290, 179, 34 },
		{ 290, 180, 34 },
		{ 290, 181, 34 },
		{ 290, 182, 34 },
		{ 290, 183, 34 },
		{ 290, 184, 34 },
		{ 290, 185, 34 },
		{ 290, 186, 34 },
		{ 290, 187, 34 },
		{ 290, 188, 34 },
		{ 290, 189, 34 },
		{ 290, 190, 34 },
		{ 290, 191, 34 },
		{ 291, 167, 34 },
		{ 291, 168, 34 },
		{ 291, 169, 34 },
		{ 291, 170, 34 },
		{ 291, 171, 34 },
		{ 291, 172, 34 },
		{ 291, 173, 34 },
		{ 291, 174, 34 },
		{ 291, 175, 34 },
		{ 291, 176, 34 },
		{ 291, 177, 34 },
		{ 291, 178, 34 },
		{ 291, 179, 34 },
		{ 291, 180, 34 },
		{ 291, 181, 34 },
		{ 291, 182, 34 },
		{ 291, 183, 34 },
		{ 291, 184, 34 },
		{ 291, 185, 34 },
		{ 291, 186, 34 },
		{ 291, 187, 34 },
		{ 291, 188, 34 },
		{ 291, 189, 34 },
		{ 291, 190, 34 },
		{ 291, 191, 34 },
		{ 292, 164, 34 },
		{ 292, 165, 34 },
		{ 292, 166, 34 },
		{ 292, 167, 34 },
		{ 292, 168, 34 },
		{ 292, 169, 34 },
		{ 292, 170, 34 },
		{ 292, 171, 34 },
		{ 292, 172, 34 },
		{ 292, 173, 34 },
		{ 292, 174, 34 },
		{ 292, 175, 34 },
		{ 292, 176, 34 },
		{ 292, 177, 34 },
		{ 292, 178, 34 },
		{ 292, 179, 34 },
		{ 292, 180, 34 },
		{ 292, 181, 34 },
		{ 292, 182, 34 },
		{ 292, 183, 34 },
		{ 292, 184, 34 },
		{ 292, 185, 34 },
		{ 292, 186, 34 },
		{ 292, 187, 34 },
		{ 292, 188, 34 },
		{ 292, 189, 34 },
		{ 292, 190, 34 },
		{ 292, 191, 34 },
		{ 293, 162, 34 },
		{ 293, 163, 34 },
		{ 293, 164, 34 },
		{ 293, 165, 34 },
		{ 293, 166, 34 },
		{ 293, 167, 34 },
		{ 293, 168, 34 },
		{ 293, 169, 34 },
		{ 293, 170, 34 },
		{ 293, 171, 34 },
		{ 293, 172, 34 },
		{ 293, 173, 34 },
		{ 293, 174, 34 },
		{ 293, 175, 34 },
		{ 293, 176, 34 },
		{ 293, 177, 34 },
		{ 293, 178, 34 },
		{ 293, 179, 34 },
		{ 293, 180, 34 },
		{ 293, 181, 34 },
		{ 293, 182, 34 },
		{ 293, 183, 34 },
		{ 293, 184, 34 },
		{ 293, 185, 34 },
		{ 293, 186, 34 },
		{ 293, 187, 34 },
		{ 293, 188, 34 },
		{ 293, 189, 34 },
		{ 293, 190, 34 },
		{ 293, 191, 34 },
		{ 294, 159, 34 },
		{ 294, 160, 34 },
		{ 294, 161, 34 },
		{ 294, 162, 34 },
		{ 294, 163, 34 },
		{ 294, 164, 34 },
		{ 294, 165, 34 },
		{ 294, 166, 34 },
		{ 294, 167, 34 },
		{ 294, 168, 34 },
		{ 294, 169, 34 },
		{ 294, 170, 34 },
		{ 294, 171, 34 },
		{ 294, 172, 34 },
		{ 294, 173, 34 },
		{ 294, 174, 34 },
		{ 294, 175, 34 },
		{ 294, 176, 34 },
		{ 294, 177, 34 },
		{ 294, 178, 34 },
		{ 294, 179, 34 },
		{ 294, 180, 34 },
		{ 294, 181, 34 },
		{ 294, 182, 34 },
		{ 294, 183, 34 },
		{ 294, 184, 34 },
		{ 294, 185, 34 },
		{ 294, 186, 34 },
		{ 294, 187, 34 },
		{ 294, 188, 34 },
		{ 294, 189, 34 },
		{ 294, 190, 34 },
		{ 294, 191, 34 },
		{ 295, 156, 34 },
		{ 295, 157, 34 },
		{ 295, 158, 34 },
		{ 295, 159, 34 },
		{ 295, 160, 34 },
		{ 295, 161, 34 },
		{ 295, 162, 34 },
		{ 295, 163, 34 },
		{ 295, 164, 34 },
		{ 295, 165, 34 },
		{ 295, 166, 34 },
		{ 295, 167, 34 },
		{ 295, 168, 34 },
		{ 295, 169, 34 },
		{ 295, 170, 34 },
		{ 295, 171, 34 },
		{ 295, 172, 34 },
		{ 295, 173, 34 },
		{ 295, 174, 34 },
		{ 295, 175, 34 },
		{ 295, 176, 34 },
		{ 295, 177, 34 },
		{ 295, 178, 34 },
		{ 295, 179, 34 },
		{ 295, 180, 34 },
		{ 295, 181, 34 },
		{ 295, 182, 34 },
		{ 295, 183, 34 },
		{ 295, 184, 34 },
		{ 295, 185, 34 },
		{ 295, 186, 34 },
		{ 295, 187, 34 },
		{ 295, 188, 34 },
		{ 295, 189, 34 },
		{ 295, 190, 34 },
		{ 295, 191, 34 },
		{ 296, 154, 34 },
		{ 296, 155, 34 },
		{ 296, 156, 34 },
		{ 296, 157, 34 },
		{ 296, 158, 34 },
		{ 296, 159, 34 },
		{ 296, 160, 34 },
		{ 296, 161, 34 },
		{ 296, 162, 34 },
		{ 296, 163, 34 },
		{ 296, 164, 34 },
		{ 296, 165, 34 },
		{ 296, 166, 34 },
		{ 296, 167, 34 },
		{ 296, 168, 34 },
		{ 296, 169, 34 },
		{ 296, 170, 34 },
		{ 296, 171, 34 },
		{ 296, 172, 34 },
		{ 296, 173, 34 },
		{ 296, 174, 34 },
		{ 296, 175, 34 },
		{ 296, 176, 34 },
		{ 296, 177, 34 },
		{ 296, 178, 34 },
		{ 296, 179, 34 },
		{ 296, 180, 34 },
		{ 296, 181, 34 },
		{ 296, 182, 34 },
		{ 296, 183, 34 },
		{ 296, 184, 34 },
		{ 296, 185, 34 },
		{ 296, 186, 34 },
		{ 296, 187, 34 },
		{ 296, 188, 34 },
		{ 296, 189, 34 },
		{ 296, 190, 34 },
		{ 296, 191, 34 },
		{ 297, 151, 34 },
		{ 297, 152, 34 },
		{ 297, 153, 34 },
		{ 297, 154, 34 },
		{ 297, 155, 34 },
		{ 297, 156, 34 },
		{ 297, 157, 34 },
		{ 297, 158, 34 },
		{ 297, 159, 34 },
		{ 297, 160, 34 },
		{ 297, 161, 34 },
		{ 297, 162, 34 },
		{ 297, 163, 34 },
		{ 297, 164, 34 },
		{ 297, 165, 34 },
		{ 297, 166, 34 },
		{ 297, 167, 34 },
		{ 297, 168, 34 },
		{ 297, 169, 34 },
		{ 297, 170, 34 },
		{ 297, 171, 34 },
		{ 297, 172, 34 },
		{ 297, 173, 34 },
		{ 297, 174, 34 },
		{ 297, 175, 34 },
		{ 297, 176, 34 },
		{ 297, 177, 34 },
		{ 297, 178, 34 },
		{ 297, 179, 34 },
		{ 297, 180, 34 },
		{ 297, 181, 34 },
		{ 297, 182, 34 },
		{ 297, 183, 34 },
		{ 297, 184, 34 },
		{ 297, 185, 34 },
		{ 297, 186, 34 },
		{ 297, 187, 34 },
		{ 297, 188, 34 },
		{ 297, 189, 34 },
		{ 297, 190, 34 },
		{ 298, 148, 34 },
		{ 298, 149, 34 },
		{ 298, 150, 34 },
		{ 298, 151, 34 },
		{ 298, 152, 34 },
		{ 298, 153, 34 },
		{ 298, 154, 34 },
		{ 298, 155, 34 },
		{ 298, 156, 34 },
		{ 298, 157, 34 },
		{ 298, 158, 34 },
		{ 298, 159, 34 },
		{ 298, 160, 34 },
		{ 298, 161, 34 },
		{ 298, 162, 34 },
		{ 298, 163, 34 },
		{ 298, 164, 34 },
		{ 298, 165, 34 },
		{ 298, 166, 34 },
		{ 298, 167, 34 },
		{ 298, 168, 34 },
		{ 298, 169, 34 },
		{ 298, 170, 34 },
		{ 298, 171, 34 },
		{ 298, 172, 34 },
		{ 298, 173, 34 },
		{ 298, 174, 34 },
		{ 298, 175, 34 },
		{ 298, 176, 34 },
		{ 298, 177, 34 },
		{ 298, 178, 34 },
		{ 298, 179, 34 },
		{ 298, 180, 34 },
		{ 298, 181, 34 },
		{ 298, 182, 34 },
		{ 298, 183, 34 },
		{ 298, 184, 34 },
		{ 298, 185, 34 },
		{ 298, 186, 34 },
		{ 298, 187, 34 },
		{ 298, 188, 34 },
		{ 298, 189, 34 },
		{ 298, 190, 34 },
		{ 299, 146, 34 },
		{ 299, 147, 34 },
		{ 299, 148, 34 },
		{ 299, 149, 34 },
		{ 299, 150, 34 },
		{ 299, 151, 34 },
		{ 299, 152, 34 },
		{ 299, 153, 34 },
		{ 299, 154, 34 },
		{ 299, 155, 34 },
		{ 299, 156, 34 },
		{ 299, 157, 34 },
		{ 299, 158, 34 },
		{ 299, 159, 34 },
		{ 299, 160, 34 },
		{ 299, 161, 34 },
		{ 299, 162, 34 },
		{ 299, 163, 34 },
		{ 299, 164, 34 },
		{ 299, 165, 34 },
		{ 299, 166, 34 },
		{ 299, 167, 34 },
		{ 299, 168, 34 },
		{ 299, 169, 34 },
		{ 299, 170, 34 },
		{ 299, 171, 34 },
		{ 299, 172, 34 },
		{ 299, 173, 34 },
		{ 299, 174, 34 },
		{ 299, 175, 34 },
		{ 299, 176, 34 },
		{ 299, 177, 34 },
		{ 299, 178, 34 },
		{ 299, 179, 34 },
		{ 299, 180, 34 },
		{ 299, 181, 34 },
		{ 299, 182, 34 },
		{ 299, 183, 34 },
		{ 299, 184, 34 },
		{ 299, 185, 34 },
		{ 299, 186, 34 },
		{ 299, 187, 34 },
		{ 299, 188, 34 },
		{ 299, 189, 34 },
		{ 300, 143, 34 },
		{ 300, 144, 34 },
		{ 300, 145, 34 },
		{ 300, 146, 34 },
		{ 300, 147, 34 },
		{ 300, 148, 34 },
		{ 300, 149, 34 },
		{ 300, 150, 34 },
		{ 300, 151, 34 },
		{ 300, 152, 34 },
		{ 300, 153, 34 },
		{ 300, 154, 34 },
		{ 300, 155, 34 },
		{ 300, 156, 34 },
		{ 300, 157, 34 },
		{ 300, 158, 34 },
		{ 300, 159, 34 },
		{ 300, 160, 34 },
		{ 300, 161, 34 },
		{ 300, 162, 34 },
		{ 300, 163, 34 },
		{ 300, 164, 34 },
		{ 300, 165, 34 },
		{ 300, 166, 34 },
		{ 300, 167, 34 },
		{ 300, 168, 34 },
		{ 300, 169, 34 },
		{ 300, 170, 34 },
		{ 300, 171, 34 },
		{ 300, 172, 34 },
		{ 300, 173, 34 },
		{ 300, 174, 34 },
		{ 300, 175, 34 },
		{ 300, 176, 34 },
		{ 300, 177, 34 },
		{ 300, 178, 34 },
		{ 300, 179, 34 },
		{ 300, 180, 34 },
		{ 300, 181, 34 },
		{ 300, 182, 34 },
		{ 300, 183, 34 },
		{ 300, 184, 34 },
		{ 300, 185, 34 },
		{ 300, 186, 34 },
		{ 300, 187, 34 },
		{ 301, 140, 34 },
		{ 301, 141, 34 },
		{ 301, 142, 34 },
		{ 301, 143, 34 },
		{ 301, 144, 34 },
		{ 301, 145, 34 },
		{ 301, 146, 34 },
		{ 301, 147, 34 },
		{ 301, 148, 34 },
		{ 301, 149, 34 },
		{ 301, 150, 34 },
		{ 301, 151, 34 },
		{ 301, 152, 34 },
		{ 301, 153, 34 },
		{ 301, 154, 34 },
		{ 301, 155, 34 },
		{ 301, 156, 34 },
		{ 301, 157, 34 },
		{ 301, 158, 34 },
		{ 301, 159, 34 },
		{ 301, 160, 34 },
		{ 301, 161, 34 },
		{ 301, 162, 34 },
		{ 301, 163, 34 },
		{ 301, 164, 34 },
		{ 301, 165, 34 },
		{ 301, 166, 34 },
		{ 301, 167, 34 },
		{ 301, 168, 34 },
		{ 301, 169, 34 },
		{ 301, 170, 34 },
		{ 301, 171, 34 },
		{ 301, 172, 34 },
		{ 301, 173, 34 },
		{ 301, 174, 34 },
		{ 301, 175, 34 },
		{ 301, 176, 34 },
		{ 301, 177, 34 },
		{ 301, 178, 34 },
		{ 301, 179, 34 },
		{ 301, 180, 34 },
		{ 301, 181, 34 },
		{ 301, 182, 34 },
		{ 301, 183, 34 },
		{ 301, 184, 34 },
		{ 301, 185, 34 },
		{ 301, 186, 34 },
		{ 302, 138, 34 },
		{ 302, 139, 34 },
		{ 302, 140, 34 },
		{ 302, 141, 34 },
		{ 302, 142, 34 },
		{ 302, 143, 34 },
		{ 302, 144, 34 },
		{ 302, 145, 34 },
		{ 302, 146, 34 },
		{ 302, 147, 34 },
		{ 302, 148, 34 },
		{ 302, 149, 34 },
		{ 302, 150, 34 },
		{ 302, 151, 34 },
		{ 302, 152, 34 },
		{ 302, 153, 34 },
		{ 302, 154, 34 },
		{ 302, 155, 34 },
		{ 302, 156, 34 },
		{ 302, 157, 34 },
		{ 302, 158, 34 },
		{ 302, 159, 34 },
		{ 302, 160, 34 },
		{ 302, 161, 34 },
		{ 302, 162, 34 },
		{ 302, 163, 34 },
		{ 302, 164, 34 },
		{ 302, 165, 34 },
		{ 302, 166, 34 },
		{ 302, 167, 34 },
		{ 302, 168, 34 },
		{ 302, 169, 34 },
		{ 302, 170, 34 },
		{ 302, 171, 34 },
		{ 302, 172, 34 },
		{ 302, 173, 34 },
		{ 302, 174, 34 },
		{ 302, 175, 34 },
		{ 302, 176, 34 },
		{ 302, 177, 34 },
		{ 302, 178, 34 },
		{ 302, 179, 34 },
		{ 302, 180, 34 },
		{ 302, 181, 34 },
		{ 302, 182, 34 },
		{ 302, 183, 34 },
		{ 302, 184, 34 },
		{ 303, 135, 34 },
		{ 303, 136, 34 },
		{ 303, 137, 34 },
		{ 303, 138, 34 },
		{ 303, 139, 34 },
		{ 303, 140, 34 },
		{ 303, 141, 34 },
		{ 303, 142, 34 },
		{ 303, 143, 34 },
		{ 303, 144, 34 },
		{ 303, 145, 34 },
		{ 303, 146, 34 },
		{ 303, 147, 34 },
		{ 303, 148, 34 },
		{ 303, 149, 34 },
		{ 303, 150, 34 },
		{ 303, 151, 34 },
		{ 303, 152, 34 },
		{ 303, 153, 34 },
		{ 303, 154, 34 },
		{ 303, 155, 34 },
		{ 303, 156, 34 },
		{ 303, 157, 34 },
		{ 303, 158, 34 },
		{ 303, 159, 34 },
		{ 303, 160, 34 },
		{ 303, 161, 34 },
		{ 303, 162, 34 },
		{ 303, 163, 34 },
		{ 303, 164, 34 },
		{ 303, 165, 34 },
		{ 303, 166, 34 },
		{ 303, 167, 34 },
		{ 303, 168, 34 },
		{ 303, 169, 34 },
		{ 303, 170, 34 },
		{ 303, 171, 34 },
		{ 303, 172, 34 },
		{ 303, 173, 34 },
		{ 303, 174, 34 },
		{ 303, 175, 34 },
		{ 303, 176, 34 },
		{ 303, 177, 34 },
		{ 303, 178, 34 },
		{ 303, 179, 34 },
		{ 303, 180, 34 },
		{ 303, 181, 34 },
		{ 303, 182, 34 },
		{ 303, 183, 34 },
		{ 304, 132, 34 },
		{ 304, 133, 34 },
		{ 304, 134, 34 },
		{ 304, 135, 34 },
		{ 304, 136, 34 },
		{ 304, 137, 34 },
		{ 304, 138, 34 },
		{ 304, 139, 34 },
		{ 304, 140, 34 },
		{ 304, 141, 34 },
		{ 304, 142, 34 },
		{ 304, 143, 34 },
		{ 304, 144, 34 },
		{ 304, 145, 34 },
		{ 304, 146, 34 },
		{ 304, 147, 34 },
		{ 304, 148, 34 },
		{ 304, 149, 34 },
		{ 304, 150, 34 },
		{ 304, 151, 34 },
		{ 304, 152, 34 },
		{ 304, 153, 34 },
		{ 304, 154, 34 },
		{ 304, 155, 34 },
		{ 304, 156, 34 },
		{ 304, 157, 34 },
		{ 304, 158, 34 },
		{ 304, 159, 34 },
		{ 304, 160, 34 },
		{ 304, 161, 34 },
		{ 304, 162, 34 },
		{ 304, 163, 34 },
		{ 304, 164, 34 },
		{ 304, 165, 34 },
		{ 304, 166, 34 },
		{ 304, 167, 34 },
		{ 304, 168, 34 },
		{ 304, 169, 34 },
		{ 304, 170, 34 },
		{ 304, 171, 34 },
		{ 304, 172, 34 },
		{ 304, 173, 34 },
		{ 304, 174, 34 },
		{ 304, 175, 34 },
		{ 304, 176, 34 },
		{ 304, 177, 34 },
		{ 304, 178, 34 },
		{ 304, 179, 34 },
		{ 304, 180, 34 },
		{ 304, 181, 34 },
		{ 305, 130, 34 },
		{ 305, 131, 34 },
		{ 305, 132, 34 },
		{ 305, 133, 34 },
		{ 305, 134, 34 },
		{ 305, 135, 34 },
		{ 305, 136, 34 },
		{ 305, 137, 34 },
		{ 305, 138, 34 },
		{ 305, 139, 34 },
		{ 305, 140, 34 },
		{ 305, 141, 34 },
		{ 305, 142, 34 },
		{ 305, 143, 34 },
		{ 305, 144, 34 },
		{ 305, 145, 34 },
		{ 305, 146, 34 },
		{ 305, 147, 34 },
		{ 305, 148, 34 },
		{ 305, 149, 34 },
		{ 305, 150, 34 },
		{ 305, 151, 34 },
		{ 305, 152, 34 },
		{ 305, 153, 34 },
		{ 305, 154, 34 },
		{ 305, 155, 34 },
		{ 305, 156, 34 },
		{ 305, 157, 34 },
		{ 305, 158, 34 },
		{ 305, 159, 34 },
		{ 305, 160, 34 },
		{ 305, 161, 34 },
		{ 305, 162, 34 },
		{ 305, 163, 34 },
		{ 305, 164, 34 },
		{ 305, 165, 34 },
		{ 305, 166, 34 },
		{ 305, 167, 34 },
		{ 305, 168, 34 },
		{ 305, 169, 34 },
		{ 305, 170, 34 },
		{ 305, 171, 34 },
		{ 305, 172, 34 },
		{ 305, 173, 34 },
		{ 305, 174, 34 },
		{ 305, 175, 34 },
		{ 305, 176, 34 },
		{ 305, 177, 34 },
		{ 305, 178, 34 },
		{ 305, 179, 34 },
		{ 305, 180, 34 },
		{ 306, 127, 34 },
		{ 306, 128, 34 },
		{ 306, 129, 34 },
		{ 306, 130, 34 },
		{ 306, 131, 34 },
		{ 306, 132, 34 },
		{ 306, 133, 34 },
		{ 306, 134, 34 },
		{ 306, 135, 34 },
		{ 306, 136, 34 },
		{ 306, 137, 34 },
		{ 306, 138, 34 },
		{ 306, 139, 34 },
		{ 306, 140, 34 },
		{ 306, 141, 34 },
		{ 306, 142, 34 },
		{ 306, 143, 34 },
		{ 306, 144, 34 },
		{ 306, 145, 34 },
		{ 306, 146, 34 },
		{ 306, 147, 34 },
		{ 306, 148, 34 },
		{ 306, 149, 34 },
		{ 306, 150, 34 },
		{ 306, 151, 34 },
		{ 306, 152, 34 },
		{ 306, 153, 34 },
		{ 306, 154, 34 },
		{ 306, 155, 34 },
		{ 306, 156, 34 },
		{ 306, 157, 34 },
		{ 306, 158, 34 },
		{ 306, 159, 34 },
		{ 306, 160, 34 },
		{ 306, 161, 34 },
		{ 306, 162, 34 },
		{ 306, 163, 34 },
		{ 306, 164, 34 },
		{ 306, 165, 34 },
		{ 306, 166, 34 },
		{ 306, 167, 34 },
		{ 306, 168, 34 },
		{ 306, 169, 34 },
		{ 306, 170, 34 },
		{ 306, 171, 34 },
		{ 306, 172, 34 },
		{ 306, 173, 34 },
		{ 306, 174, 34 },
		{ 306, 175, 34 },
		{ 306, 176, 34 },
		{ 306, 177, 34 },
		{ 306, 178, 34 },
		{ 307, 124, 34 },
		{ 307, 125, 34 },
		{ 307, 126, 34 },
		{ 307, 127, 34 },
		{ 307, 128, 34 },
		{ 307, 129, 34 },
		{ 307, 130, 34 },
		{ 307, 131, 34 },
		{ 307, 132, 34 },
		{ 307, 133, 34 },
		{ 307, 134, 34 },
		{ 307, 135, 34 },
		{ 307, 136, 34 },
		{ 307, 137, 34 },
		{ 307, 138, 34 },
		{ 307, 139, 34 },
		{ 307, 140, 34 },
		{ 307, 141, 34 },
		{ 307, 142, 34 },
		{ 307, 143, 34 },
		{ 307, 144, 34 },
		{ 307, 145, 34 },
		{ 307, 146, 34 },
		{ 307, 147, 34 },
		{ 307, 148, 34 },
		{ 307, 149, 34 },
		{ 307, 150, 34 },
		{ 307, 151, 34 },
		{ 307, 152, 34 },
		{ 307, 153, 34 },
		{ 307, 154, 34 },
		{ 307, 155, 34 },
		{ 307, 156, 34 },
		{ 307, 157, 34 },
		{ 307, 158, 34 },
		{ 307, 159, 34 },
		{ 307, 160, 34 },
		{ 307, 161, 34 },
		{ 307, 162, 34 },
		{ 307, 163, 34 },
		{ 307, 164, 34 },
		{ 307, 165, 34 },
		{ 307, 166, 34 },
		{ 307, 167, 34 },
		{ 307, 168, 34 },
		{ 307, 169, 34 },
		{ 307, 170, 34 },
		{ 307, 171, 34 },
		{ 307, 172, 34 },
		{ 307, 173, 34 },
		{ 307, 174, 34 },
		{ 307, 175, 34 },
		{ 307, 176, 34 },
		{ 308, 122, 34 },
		{ 308, 123, 34 },
		{ 308, 124, 34 },
		{ 308, 125, 34 },
		{ 308, 126, 34 },
		{ 308, 127, 34 },
		{ 308, 128, 34 },
		{ 308, 129, 34 },
		{ 308, 130, 34 },
		{ 308, 131, 34 },
		{ 308, 132, 34 },
		{ 308, 133, 34 },
		{ 308, 134, 34 },
		{ 308, 135, 34 },
		{ 308, 136, 34 },
		{ 308, 137, 34 },
		{ 308, 138, 34 },
		{ 308, 139, 34 },
		{ 308, 140, 34 },
		{ 308, 141, 34 },
		{ 308, 142, 34 },
		{ 308, 143, 34 },
		{ 308, 144, 34 },
		{ 308, 145, 34 },
		{ 308, 146, 34 },
		{ 308, 147, 34 },
		{ 308, 148, 34 },
		{ 308, 149, 34 },
		{ 308, 150, 34 },
		{ 308, 151, 34 },
		{ 308, 152, 34 },
		{ 308, 153, 34 },
		{ 308, 154, 34 },
		{ 308, 155, 34 },
		{ 308, 156, 34 },
		{ 308, 157, 34 },
		{ 308, 158, 34 },
		{ 308, 159, 34 },
		{ 308, 160, 34 },
		{ 308, 161, 34 },
		{ 308, 162, 34 },
		{ 308, 163, 34 },
		{ 308, 164, 34 },
		{ 308, 165, 34 },
		{ 308, 166, 34 },
		{ 308, 167, 34 },
		{ 308, 168, 34 },
		{ 308, 169, 34 },
		{ 308, 170, 34 },
		{ 308, 171, 34 },
		{ 308, 172, 34 },
		{ 308, 173, 34 },
		{ 308, 174, 34 },
		{ 308, 175, 34 },
		{ 309, 119, 34 },
		{ 309, 120, 34 },
		{ 309, 121, 34 },
		{ 309, 122, 34 },
		{ 309, 123, 34 },
		{ 309, 124, 34 },
		{ 309, 125, 34 },
		{ 309, 126, 34 },
		{ 309, 127, 34 },
		{ 309, 128, 34 },
		{ 309, 129, 34 },
		{ 309, 130, 34 },
		{ 309, 131, 34 },
		{ 309, 132, 34 },
		{ 309, 133, 34 },
		{ 309, 134, 34 },
		{ 309, 135, 34 },
		{ 309, 136, 34 },
		{ 309, 137, 34 },
		{ 309, 138, 34 },
		{ 309, 139, 34 },
		{ 309, 140, 34 },
		{ 309, 141, 34 },
		{ 309, 142, 34 },
		{ 309, 143, 34 },
		{ 309, 144, 34 },
		{ 309, 145, 34 },
		{ 309, 146, 34 },
		{ 309, 147, 34 },
		{ 309, 148, 34 },
		{ 309, 149, 34 },
		{ 309, 150, 34 },
		{ 309, 151, 34 },
		{ 309, 152, 34 },
		{ 309, 153, 34 },
		{ 309, 154, 34 },
		{ 309, 155, 34 },
		{ 309, 156, 34 },
		{ 309, 157, 34 },
		{ 309, 158, 34 },
		{ 309, 159, 34 },
		{ 309, 160, 34 },
		{ 309, 161, 34 },
		{ 309, 162, 34 },
		{ 309, 163, 34 },
		{ 309, 164, 34 },
		{ 309, 165, 34 },
		{ 309, 166, 34 },
		{ 309, 167, 34 },
		{ 309, 168, 34 },
		{ 309, 169, 34 },
		{ 309, 170, 34 },
		{ 309, 171, 34 },
		{ 309, 172, 34 },
		{ 309, 173, 34 },
		{ 310, 116, 34 },
		{ 310, 117, 34 },
		{ 310, 118, 34 },
		{ 310, 119, 34 },
		{ 310, 120, 34 },
		{ 310, 121, 34 },
		{ 310, 122, 34 },
		{ 310, 123, 34 },
		{ 310, 124, 34 },
		{ 310, 125, 34 },
		{ 310, 126, 34 },
		{ 310, 127, 34 },
		{ 310, 128, 34 },
		{ 310, 129, 34 },
		{ 310, 130, 34 },
		{ 310, 131, 34 },
		{ 310, 132, 34 },
		{ 310, 133, 34 },
		{ 310, 134, 34 },
		{ 310, 135, 34 },
		{ 310, 136, 34 },
		{ 310, 137, 34 },
		{ 310, 138, 34 },
		{ 310, 139, 34 },
		{ 310, 140, 34 },
		{ 310, 141, 34 },
		{ 310, 142, 34 },
		{ 310, 143, 34 },
		{ 310, 144, 34 },
		{ 310, 145, 34 },
		{ 310, 146, 34 },
		{ 310, 147, 34 },
		{ 310, 148, 34 },
		{ 310, 149, 34 },
		{ 310, 150, 34 },
		{ 310, 151, 34 },
		{ 310, 152, 34 },
		{ 310, 153, 34 },
		{ 310, 154, 34 },
		{ 310, 155, 34 },
		{ 310, 156, 34 },
		{ 310, 157, 34 },
		{ 310, 158, 34 },
		{ 310, 159, 34 },
		{ 310, 160, 34 },
		{ 310, 161, 34 },
		{ 310, 162, 34 },
		{ 310, 163, 34 },
		{ 310, 164, 34 },
		{ 310, 165, 34 },
		{ 310, 166, 34 },
		{ 310, 167, 34 },
		{ 310, 168, 34 },
		{ 310, 169, 34 },
		{ 310, 170, 34 },
		{ 310, 171, 34 },
		{ 310, 172, 34 },
		{ 311, 114, 34 },
		{ 311, 115, 34 },
		{ 311, 116, 34 },
		{ 311, 117, 34 },
		{ 311, 118, 34 },
		{ 311, 119, 34 },
		{ 311, 120, 34 },
		{ 311, 121, 34 },
		{ 311, 122, 34 },
		{ 311, 123, 34 },
		{ 311, 124, 34 },
		{ 311, 125, 34 },
		{ 311, 126, 34 },
		{ 311, 127, 34 },
		{ 311, 128, 34 },
		{ 311, 129, 34 },
		{ 311, 130, 34 },
		{ 311, 131, 34 },
		{ 311, 132, 34 },
		{ 311, 133, 34 },
		{ 311, 134, 34 },
		{ 311, 135, 34 },
		{ 311, 136, 34 },
		{ 311, 137, 34 },
		{ 311, 138, 34 },
		{ 311, 139, 34 },
		{ 311, 140, 34 },
		{ 311, 141, 34 },
		{ 311, 142, 34 },
		{ 311, 143, 34 },
		{ 311, 144, 34 },
		{ 311, 145, 34 },
		{ 311, 146, 34 },
		{ 311, 147, 34 },
		{ 311, 148, 34 },
		{ 311, 149, 34 },
		{ 311, 150, 34 },
		{ 311, 151, 34 },
		{ 311, 152, 34 },
		{ 311, 153, 34 },
		{ 311, 154, 34 },
		{ 311, 155, 34 },
		{ 311, 156, 34 },
		{ 311, 157, 34 },
		{ 311, 158, 34 },
		{ 311, 159, 34 },
		{ 311, 160, 34 },
		{ 311, 161, 34 },
		{ 311, 162, 34 },
		{ 311, 163, 34 },
		{ 311, 164, 34 },
		{ 311, 165, 34 },
		{ 311, 166, 34 },
		{ 311, 167, 34 },
		{ 311, 168, 34 },
		{ 311, 169, 34 },
		{ 311, 170, 34 },
		{ 312, 111, 34 },
		{ 312, 112, 34 },
		{ 312, 113, 34 },
		{ 312, 114, 34 },
		{ 312, 115, 34 },
		{ 312, 116, 34 },
		{ 312, 117, 34 },
		{ 312, 118, 34 },
		{ 312, 119, 34 },
		{ 312, 120, 34 },
		{ 312, 121, 34 },
		{ 312, 122, 34 },
		{ 312, 123, 34 },
		{ 312, 124, 34 },
		{ 312, 125, 34 },
		{ 312, 126, 34 },
		{ 312, 127, 34 },
		{ 312, 128, 34 },
		{ 312, 129, 34 },
		{ 312, 130, 34 },
		{ 312, 131, 34 },
		{ 312, 132, 34 },
		{ 312, 133, 34 },
		{ 312, 134, 34 },
		{ 312, 135, 34 },
		{ 312, 136, 34 },
		{ 312, 137, 34 },
		{ 312, 138, 34 },
		{ 312, 139, 34 },
		{ 312, 140, 34 },
		{ 312, 141, 34 },
		{ 312, 142, 34 },
		{ 312, 143, 34 },
		{ 312, 144, 34 },
		{ 312, 145, 34 },
		{ 312, 146, 34 },
		{ 312, 147, 34 },
		{ 312, 148, 34 },
		{ 312, 149, 34 },
		{ 312, 150, 34 },
		{ 312, 151, 34 },
		{ 312, 152, 34 },
		{ 312, 153, 34 },
		{ 312, 154, 34 },
		{ 312, 155, 34 },
		{ 312, 156, 34 },
		{ 312, 157, 34 },
		{ 312, 158, 34 },
		{ 312, 159, 34 },
		{ 312, 160, 34 },
		{ 312, 161, 34 },
		{ 312, 162, 34 },
		{ 312, 163, 34 },
		{ 312, 164, 34 },
		{ 312, 165, 34 },
		{ 312, 166, 34 },
		{ 312, 167, 34 },
		{ 312, 168, 34 },
		{ 313, 108, 34 },
		{ 313, 109, 34 },
		{ 313, 110, 34 },
		{ 313, 111, 34 },
		{ 313, 112, 34 },
		{ 313, 113, 34 },
		{ 313, 114, 34 },
		{ 313, 115, 34 },
		{ 313, 116, 34 },
		{ 313, 117, 34 },
		{ 313, 118, 34 },
		{ 313, 119, 34 },
		{ 313, 120, 34 },
		{ 313, 121, 34 },
		{ 313, 122, 34 },
		{ 313, 123, 34 },
		{ 313, 124, 34 },
		{ 313, 125, 34 },
		{ 313, 126, 34 },
		{ 313, 127, 34 },
		{ 313, 128, 34 },
		{ 313, 129, 34 },
		{ 313, 130, 34 },
		{ 313, 131, 34 },
		{ 313, 132, 34 },
		{ 313, 133, 34 },
		{ 313, 134, 34 },
		{ 313, 135, 34 },
		{ 313, 136, 34 },
		{ 313, 137, 34 },
		{ 313, 138, 34 },
		{ 313, 139, 34 },
		{ 313, 140, 34 },
		{ 313, 141, 34 },
		{ 313, 142, 34 },
		{ 313, 143, 34 },
		{ 313, 144, 34 },
		{ 313, 145, 34 },
		{ 313, 146, 34 },
		{ 313, 147, 34 },
		{ 313, 148, 34 },
		{ 313, 149, 34 },
		{ 313, 150, 34 },
		{ 313, 151, 34 },
		{ 313, 152, 34 },
		{ 313, 153, 34 },
		{ 313, 154, 34 },
		{ 313, 155, 34 },
		{ 313, 156, 34 },
		{ 313, 157, 34 },
		{ 313, 158, 34 },
		{ 313, 159, 34 },
		{ 313, 160, 34 },
		{ 313, 161, 34 },
		{ 313, 162, 34 },
		{ 313, 163, 34 },
		{ 313, 164, 34 },
		{ 313, 165, 34 },
		{ 313, 166, 34 },
		{ 313, 167, 34 },
		{ 314, 106, 34 },
		{ 314, 107, 34 },
		{ 314, 108, 34 },
		{ 314, 109, 34 },
		{ 314, 110, 34 },
		{ 314, 111, 34 },
		{ 314, 112, 34 },
		{ 314, 113, 34 },
		{ 314, 114, 34 },
		{ 314, 115, 34 },
		{ 314, 116, 34 },
		{ 314, 117, 34 },
		{ 314, 118, 34 },
		{ 314, 119, 34 },
		{ 314, 120, 34 },
		{ 314, 121, 34 },
		{ 314, 122, 34 },
		{ 314, 123, 34 },
		{ 314, 124, 34 },
		{ 314, 125, 34 },
		{ 314, 126, 34 },
		{ 314, 127, 34 },
		{ 314, 128, 34 },
		{ 314, 129, 34 },
		{ 314, 130, 34 },
		{ 314, 131, 34 },
		{ 314, 132, 34 },
		{ 314, 133, 34 },
		{ 314, 134, 34 },
		{ 314, 135, 34 },
		{ 314, 136, 34 },
		{ 314, 137, 34 },
		{ 314, 138, 34 },
		{ 314, 139, 34 },
		{ 314, 140, 34 },
		{ 314, 141, 34 },
		{ 314, 142, 34 },
		{ 314, 143, 34 },
		{ 314, 144, 34 },
		{ 314, 145, 34 },
		{ 314, 146, 34 },
		{ 314, 147, 34 },
		{ 314, 148, 34 },
		{ 314, 149, 34 },
		{ 314, 150, 34 },
		{ 314, 151, 34 },
		{ 314, 152, 34 },
		{ 314, 153, 34 },
		{ 314, 154, 34 },
		{ 314, 155, 34 },
		{ 314, 156, 34 },
		{ 314, 157, 34 },
		{ 314, 158, 34 },
		{ 314, 159, 34 },
		{ 314, 160, 34 },
		{ 314, 161, 34 },
		{ 314, 162, 34 },
		{ 314, 163, 34 },
		{ 314, 164, 34 },
		{ 314, 165, 34 },
		{ 315, 103, 34 },
		{ 315, 104, 34 },
		{ 315, 105, 34 },
		{ 315, 106, 34 },
		{ 315, 107, 34 },
		{ 315, 108, 34 },
		{ 315, 109, 34 },
		{ 315, 110, 34 },
		{ 315, 111, 34 },
		{ 315, 112, 34 },
		{ 315, 113, 34 },
		{ 315, 114, 34 },
		{ 315, 115, 34 },
		{ 315, 116, 34 },
		{ 315, 117, 34 },
		{ 315, 118, 34 },
		{ 315, 119, 34 },
		{ 315, 120, 34 },
		{ 315, 121, 34 },
		{ 315, 122, 34 },
		{ 315, 123, 34 },
		{ 315, 124, 34 },
		{ 315, 125, 34 },
		{ 315, 126, 34 },
		{ 315, 127, 34 },
		{ 315, 128, 34 },
		{ 315, 129, 34 },
		{ 315, 130, 34 },
		{ 315, 131, 34 },
		{ 315, 132, 34 },
		{ 315, 133, 34 },
		{ 315, 134, 34 },
		{ 315, 135, 34 },
		{ 315, 136, 34 },
		{ 315, 137, 34 },
		{ 315, 138, 34 },
		{ 315, 139, 34 },
		{ 315, 140, 34 },
		{ 315, 141, 34 },
		{ 315, 142, 34 },
		{ 315, 143, 34 },
		{ 315, 144, 34 },
		{ 315, 145, 34 },
		{ 315, 146, 34 },
		{ 315, 147, 34 },
		{ 315, 148, 34 },
		{ 315, 149, 34 },
		{ 315, 150, 34 },
		{ 315, 151, 34 },
		{ 315, 152, 34 },
		{ 315, 153, 34 },
		{ 315, 154, 34 },
		{ 315, 155, 34 },
		{ 315, 156, 34 },
		{ 315, 157, 34 },
		{ 315, 158, 34 },
		{ 315, 159, 34 },
		{ 315, 160, 34 },
		{ 315, 161, 34 },
		{ 315, 162, 34 },
		{ 315, 163, 34 },
		{ 315, 164, 34 },
		{ 316, 100, 34 },
		{ 316, 101, 34 },
		{ 316, 102, 34 },
		{ 316, 103, 34 },
		{ 316, 104, 34 },
		{ 316, 105, 34 },
		{ 316, 106, 34 },
		{ 316, 107, 34 },
		{ 316, 108, 34 },
		{ 316, 109, 34 },
		{ 316, 110, 34 },
		{ 316, 111, 34 },
		{ 316, 112, 34 },
		{ 316, 113, 34 },
		{ 316, 114, 34 },
		{ 316, 115, 34 },
		{ 316, 116, 34 },
		{ 316, 117, 34 },
		{ 316, 118, 34 },
		{ 316, 119, 34 },
		{ 316, 120, 34 },
		{ 316, 121, 34 },
		{ 316, 122, 34 },
		{ 316, 123, 34 },
		{ 316, 124, 34 },
		{ 316, 125, 34 },
		{ 316, 126, 34 },
		{ 316, 127, 34 },
		{ 316, 128, 34 },
		{ 316, 129, 34 },
		{ 316, 130, 34 },
		{ 316, 131, 34 },
		{ 316, 132, 34 },
		{ 316, 133, 34 },
		{ 316, 134, 34 },
		{ 316, 135, 34 },
		{ 316, 136, 34 },
		{ 316, 137, 34 },
		{ 316, 138, 34 },
		{ 316, 139, 34 },
		{ 316, 140, 34 },
		{ 316, 141, 34 },
		{ 316, 142, 34 },
		{ 316, 143, 34 },
		{ 316, 144, 34 },
		{ 316, 145, 34 },
		{ 316, 146, 34 },
		{ 316, 147, 34 },
		{ 316, 148, 34 },
		{ 316, 149, 34 },
		{ 316, 150, 34 },
		{ 316, 151, 34 },
		{ 316, 152, 34 },
		{ 316, 153, 34 },
		{ 316, 154, 34 },
		{ 316, 155, 34 },
		{ 316, 156, 34 },
		{ 316, 157, 34 },
		{ 316, 158, 34 },
		{ 316, 159, 34 },
		{ 316, 160, 34 },
		{ 316, 161, 34 },
		{ 316, 162, 34 },
		{ 317, 98, 34 },
		{ 317, 99, 34 },
		{ 317, 100, 34 },
		{ 317, 101, 34 },
		{ 317, 102, 34 },
		{ 317, 103, 34 },
		{ 317, 104, 34 },
		{ 317, 105, 34 },
		{ 317, 106, 34 },
		{ 317, 107, 34 },
		{ 317, 108, 34 },
		{ 317, 109, 34 },
		{ 317, 110, 34 },
		{ 317, 111, 34 },
		{ 317, 112, 34 },
		{ 317, 113, 34 },
		{ 317, 114, 34 },
		{ 317, 115, 34 },
		{ 317, 116, 34 },
		{ 317, 117, 34 },
		{ 317, 118, 34 },
		{ 317, 119, 34 },
		{ 317, 120, 34 },
		{ 317, 121, 34 },
		{ 317, 122, 34 },
		{ 317, 123, 34 },
		{ 317, 124, 34 },
		{ 317, 125, 34 },
		{ 317, 126, 34 },
		{ 317, 127, 34 },
		{ 317, 128, 34 },
		{ 317, 129, 34 },
		{ 317, 130, 34 },
		{ 317, 131, 34 },
		{ 317, 132, 34 },
		{ 317, 133, 34 },
		{ 317, 134, 34 },
		{ 317, 135, 34 },
		{ 317, 136, 34 },
		{ 317, 137, 34 },
		{ 317, 138, 34 },
		{ 317, 139, 34 },
		{ 317, 140, 34 },
		{ 317, 141, 34 },
		{ 317, 142, 34 },
		{ 317, 143, 34 },
		{ 317, 144, 34 },
		{ 317, 145, 34 },
		{ 317, 146, 34 },
		{ 317, 147, 34 },
		{ 317, 148, 34 },
		{ 317, 149, 34 },
		{ 317, 150, 34 },
		{ 317, 151, 34 },
		{ 317, 152, 34 },
		{ 317, 153, 34 },
		{ 317, 154, 34 },
		{ 317, 155, 34 },
		{ 317, 156, 34 },
		{ 317, 157, 34 },
		{ 317, 158, 34 },
		{ 317, 159, 34 },
		{ 317, 160, 34 },
		{ 318, 95, 34 },
		{ 318, 96, 34 },
		{ 318, 97, 34 },
		{ 318, 98, 34 },
		{ 318, 99, 34 },
		{ 318, 100, 34 },
		{ 318, 101, 34 },
		{ 318, 102, 34 },
		{ 318, 103, 34 },
		{ 318, 104, 34 },
		{ 318, 105, 34 },
		{ 318, 106, 34 },
		{ 318, 107, 34 },
		{ 318, 108, 34 },
		{ 318, 109, 34 },
		{ 318, 110, 34 },
		{ 318, 111, 34 },
		{ 318, 112, 34 },
		{ 318, 113, 34 },
		{ 318, 114, 34 },
		{ 318, 115, 34 },
		{ 318, 116, 34 },
		{ 318, 117, 34 },
		{ 318, 118, 34 },
		{ 318, 119, 34 },
		{ 318, 120, 34 },
		{ 318, 121, 34 },
		{ 318, 122, 34 },
		{ 318, 123, 34 },
		{ 318, 124, 34 },
		{ 318, 125, 34 },
		{ 318, 126, 34 },
		{ 318, 127, 34 },
		{ 318, 128, 34 },
		{ 318, 129, 34 },
		{ 318, 130, 34 },
		{ 318, 131, 34 },
		{ 318, 132, 34 },
		{ 318, 133, 34 },
		{ 318, 134, 34 },
		{ 318, 135, 34 },
		{ 318, 136, 34 },
		{ 318, 137, 34 },
		{ 318, 138, 34 },
		{ 318, 139, 34 },
		{ 318, 140, 34 },
		{ 318, 141, 34 },
		{ 318, 142, 34 },
		{ 318, 143, 34 },
		{ 318, 144, 34 },
		{ 318, 145, 34 },
		{ 318, 146, 34 },
		{ 318, 147, 34 },
		{ 318, 148, 34 },
		{ 318, 149, 34 },
		{ 318, 150, 34 },
		{ 318, 151, 34 },
		{ 318, 152, 34 },
		{ 318, 153, 34 },
		{ 318, 154, 34 },
		{ 318, 155, 34 },
		{ 318, 156, 34 },
		{ 318, 157, 34 },
		{ 318, 158, 34 },
		{ 318, 159, 34 },
		{ 319, 92, 34 },
		{ 319, 93, 34 },
		{ 319, 94, 34 },
		{ 319, 95, 34 },
		{ 319, 96, 34 },
		{ 319, 97, 34 },
		{ 319, 98, 34 },
		{ 319, 99, 34 },
		{ 319, 100, 34 },
		{ 319, 101, 34 },
		{ 319, 102, 34 },
		{ 319, 103, 34 },
		{ 319, 104, 34 },
		{ 319, 105, 34 },
		{ 319, 106, 34 },
		{ 319, 107, 34 },
		{ 319, 108, 34 },
		{ 319, 109, 34 },
		{ 319, 110, 34 },
		{ 319, 111, 34 },
		{ 319, 112, 34 },
		{ 319, 113, 34 },
		{ 319, 114, 34 },
		{ 319, 115, 34 },
		{ 319, 116, 34 },
		{ 319, 117, 34 },
		{ 319, 118, 34 },
		{ 319, 119, 34 },
		{ 319, 120, 34 },
		{ 319, 121, 34 },
		{ 319, 122, 34 },
		{ 319, 123, 34 },
		{ 319, 124, 34 },
		{ 319, 125, 34 },
		{ 319, 126, 34 },
		{ 319, 127, 34 },
		{ 319, 128, 34 },
		{ 319, 129, 34 },
		{ 319, 130, 34 },
		{ 319, 131, 34 },
		{ 319, 132, 34 },
		{ 319, 133, 34 },
		{ 319, 134, 34 },
		{ 319, 135, 34 },
		{ 319, 136, 34 },
		{ 319, 137, 34 },
		{ 319, 138, 34 },
		{ 319, 139, 34 },
		{ 319, 140, 34 },
		{ 319, 141, 34 },
		{ 319, 142, 34 },
		{ 319, 143, 34 },
		{ 319, 144, 34 },
		{ 319, 145, 34 },
		{ 319, 146, 34 },
		{ 319, 147, 34 },
		{ 319, 148, 34 },
		{ 319, 149, 34 },
		{ 319, 150, 34 },
		{ 319, 151, 34 },
		{ 319, 152, 34 },
		{ 319, 153, 34 },
		{ 319, 154, 34 },
		{ 319, 155, 34 },
		{ 319, 156, 34 },
		{ 319, 157, 34 },
		{ 320, 90, 34 },
		{ 320, 91, 34 },
		{ 320, 92, 34 },
		{ 320, 93, 34 },
		{ 320, 94, 34 },
		{ 320, 95, 34 },
		{ 320, 96, 34 },
		{ 320, 97, 34 },
		{ 320, 98, 34 },
		{ 320, 99, 34 },
		{ 320, 100, 34 },
		{ 320, 101, 34 },
		{ 320, 102, 34 },
		{ 320, 103, 34 },
		{ 320, 104, 34 },
		{ 320, 105, 34 },
		{ 320, 106, 34 },
		{ 320, 107, 34 },
		{ 320, 108, 34 },
		{ 320, 109, 34 },
		{ 320, 110, 34 },
		{ 320, 111, 34 },
		{ 320, 112, 34 },
		{ 320, 113, 34 },
		{ 320, 114, 34 },
		{ 320, 115, 34 },
		{ 320, 116, 34 },
		{ 320, 117, 34 },
		{ 320, 118, 34 },
		{ 320, 119, 34 },
		{ 320, 120, 34 },
		{ 320, 121, 34 },
		{ 320, 122, 34 },
		{ 320, 123, 34 },
		{ 320, 124, 34 },
		{ 320, 125, 34 },
		{ 320, 126, 34 },
		{ 320, 127, 34 },
		{ 320, 128, 34 },
		{ 320, 129, 34 },
		{ 320, 130, 34 },
		{ 320, 131, 34 },
		{ 320, 132, 34 },
		{ 320, 133, 34 },
		{ 320, 134, 34 },
		{ 320, 135, 34 },
		{ 320, 136, 34 },
		{ 320, 137, 34 },
		{ 320, 138, 34 },
		{ 320, 139, 34 },
		{ 320, 140, 34 },
		{ 320, 141, 34 },
		{ 320, 142, 34 },
		{ 320, 143, 34 },
		{ 320, 144, 34 },
		{ 320, 145, 34 },
		{ 320, 146, 34 },
		{ 320, 147, 34 },
		{ 320, 148, 34 },
		{ 320, 149, 34 },
		{ 320, 150, 34 },
		{ 320, 151, 34 },
		{ 320, 152, 34 },
		{ 320, 153, 34 },
		{ 320, 154, 34 },
		{ 320, 155, 34 },
		{ 320, 156, 34 },
		{ 321, 87, 34 },
		{ 321, 88, 34 },
		{ 321, 89, 34 },
		{ 321, 90, 34 },
		{ 321, 91, 34 },
		{ 321, 92, 34 },
		{ 321, 93, 34 },
		{ 321, 94, 34 },
		{ 321, 95, 34 },
		{ 321, 96, 34 },
		{ 321, 97, 34 },
		{ 321, 98, 34 },
		{ 321, 99, 34 },
		{ 321, 100, 34 },
		{ 321, 101, 34 },
		{ 321, 102, 34 },
		{ 321, 103, 34 },
		{ 321, 104, 34 },
		{ 321, 105, 34 },
		{ 321, 106, 34 },
		{ 321, 107, 34 },
		{ 321, 108, 34 },
		{ 321, 109, 34 },
		{ 321, 110, 34 },
		{ 321, 111, 34 },
		{ 321, 112, 34 },
		{ 321, 113, 34 },
		{ 321, 114, 34 },
		{ 321, 115, 34 },
		{ 321, 116, 34 },
		{ 321, 117, 34 },
		{ 321, 118, 34 },
		{ 321, 119, 34 },
		{ 321, 120, 34 },
		{ 321, 121, 34 },
		{ 321, 122, 34 },
		{ 321, 123, 34 },
		{ 321, 124, 34 },
		{ 321, 125, 34 },
		{ 321, 126, 34 },
		{ 321, 127, 34 },
		{ 321, 128, 34 },
		{ 321, 129, 34 },
		{ 321, 130, 34 },
		{ 321, 131, 34 },
		{ 321, 132, 34 },
		{ 321, 133, 34 },
		{ 321, 134, 34 },
		{ 321, 135, 34 },
		{ 321, 136, 34 },
		{ 321, 137, 34 },
		{ 321, 138, 34 },
		{ 321, 139, 34 },
		{ 321, 140, 34 },
		{ 321, 141, 34 },
		{ 321, 142, 34 },
		{ 321, 143, 34 },
		{ 321, 144, 34 },
		{ 321, 145, 34 },
		{ 321, 146, 34 },
		{ 321, 147, 34 },
		{ 321, 148, 34 },
		{ 321, 149, 34 },
		{ 321, 150, 34 },
		{ 321, 151, 34 },
		{ 321, 152, 34 },
		{ 321, 153, 34 },
		{ 321, 154, 34 },
		{ 322, 84, 34 },
		{ 322, 85, 34 },
		{ 322, 86, 34 },
		{ 322, 87, 34 },
		{ 322, 88, 34 },
		{ 322, 89, 34 },
		{ 322, 90, 34 },
		{ 322, 91, 34 },
		{ 322, 92, 34 },
		{ 322, 93, 34 },
		{ 322, 94, 34 },
		{ 322, 95, 34 },
		{ 322, 96, 34 },
		{ 322, 97, 34 },
		{ 322, 98, 34 },
		{ 322, 99, 34 },
		{ 322, 100, 34 },
		{ 322, 101, 34 },
		{ 322, 102, 34 },
		{ 322, 103, 34 },
		{ 322, 104, 34 },
		{ 322, 105, 34 },
		{ 322, 106, 34 },
		{ 322, 107, 34 },
		{ 322, 108, 34 },
		{ 322, 109, 34 },
		{ 322, 110, 34 },
		{ 322, 111, 34 },
		{ 322, 112, 34 },
		{ 322, 113, 34 },
		{ 322, 114, 34 },
		{ 322, 115, 34 },
		{ 322, 116, 34 },
		{ 322, 117, 34 },
		{ 322, 118, 34 },
		{ 322, 119, 34 },
		{ 322, 120, 34 },
		{ 322, 121, 34 },
		{ 322, 122, 34 },
		{ 322, 123, 34 },
		{ 322, 124, 34 },
		{ 322, 125, 34 },
		{ 322, 126, 34 },
		{ 322, 127, 34 },
		{ 322, 128, 34 },
		{ 322, 129, 34 },
		{ 322, 130, 34 },
		{ 322, 131, 34 },
		{ 322, 132, 34 },
		{ 322, 133, 34 },
		{ 322, 134, 34 },
		{ 322, 135, 34 },
		{ 322, 136, 34 },
		{ 322, 137, 34 },
		{ 322, 138, 34 },
		{ 322, 139, 34 },
		{ 322, 140, 34 },
		{ 322, 141, 34 },
		{ 322, 142, 34 },
		{ 322, 143, 34 },
		{ 322, 144, 34 },
		{ 322, 145, 34 },
		{ 322, 146, 34 },
		{ 322, 147, 34 },
		{ 322, 148, 34 },
		{ 322, 149, 34 },
		{ 322, 150, 34 },
		{ 322, 151, 34 },
		{ 322, 152, 34 },
		{ 323, 82, 34 },
		{ 323, 83, 34 },
		{ 323, 84, 34 },
		{ 323, 85, 34 },
		{ 323, 86, 34 },
		{ 323, 87, 34 },
		{ 323, 88, 34 },
		{ 323, 89, 34 },
		{ 323, 90, 34 },
		{ 323, 91, 34 },
		{ 323, 92, 34 },
		{ 323, 93, 34 },
		{ 323, 94, 34 },
		{ 323, 95, 34 },
		{ 323, 96, 34 },
		{ 323, 97, 34 },
		{ 323, 98, 34 },
		{ 323, 99, 34 },
		{ 323, 100, 34 },
		{ 323, 101, 34 },
		{ 323, 102, 34 },
		{ 323, 103, 34 },
		{ 323, 104, 34 },
		{ 323, 105, 34 },
		{ 323, 106, 34 },
		{ 323, 107, 34 },
		{ 323, 108, 34 },
		{ 323, 109, 34 },
		{ 323, 110, 34 },
		{ 323, 111, 34 },
		{ 323, 112, 34 },
		{ 323, 113, 34 },
		{ 323, 114, 34 },
		{ 323, 115, 34 },
		{ 323, 116, 34 },
		{ 323, 117, 34 },
		{ 323, 118, 34 },
		{ 323, 119, 34 },
		{ 323, 120, 34 },
		{ 323, 121, 34 },
		{ 323, 122, 34 },
		{ 323, 123, 34 },
		{ 323, 124, 34 },
		{ 323, 125, 34 },
		{ 323, 126, 34 },
		{ 323, 127, 34 },
		{ 323, 128, 34 },
		{ 323, 129, 34 },
		{ 323, 130, 34 },
		{ 323, 131, 34 },
		{ 323, 132, 34 },
		{ 323, 133, 34 },
		{ 323, 134, 34 },
		{ 323, 135, 34 },
		{ 323, 136, 34 },
		{ 323, 137, 34 },
		{ 323, 138, 34 },
		{ 323, 139, 34 },
		{ 323, 140, 34 },
		{ 323, 141, 34 },
		{ 323, 142, 34 },
		{ 323, 143, 34 },
		{ 323, 144, 34 },
		{ 323, 145, 34 },
		{ 323, 146, 34 },
		{ 323, 147, 34 },
		{ 323, 148, 34 },
		{ 323, 149, 34 },
		{ 323, 150, 34 },
		{ 323, 151, 34 },
		{ 324, 79, 34 },
		{ 324, 80, 34 },
		{ 324, 81, 34 },
		{ 324, 82, 34 },
		{ 324, 83, 34 },
		{ 324, 84, 34 },
		{ 324, 85, 34 },
		{ 324, 86, 34 },
		{ 324, 87, 34 },
		{ 324, 88, 34 },
		{ 324, 89, 34 },
		{ 324, 90, 34 },
		{ 324, 91, 34 },
		{ 324, 92, 34 },
		{ 324, 93, 34 },
		{ 324, 94, 34 },
		{ 324, 95, 34 },
		{ 324, 96, 34 },
		{ 324, 97, 34 },
		{ 324, 98, 34 },
		{ 324, 99, 34 },
		{ 324, 100, 34 },
		{ 324, 101, 34 },
		{ 324, 102, 34 },
		{ 324, 103, 34 },
		{ 324, 104, 34 },
		{ 324, 105, 34 },
		{ 324, 106, 34 },
		{ 324, 107, 34 },
		{ 324, 108, 34 },
		{ 324, 109, 34 },
		{ 324, 110, 34 },
		{ 324, 111, 34 },
		{ 324, 112, 34 },
		{ 324, 113, 34 },
		{ 324, 114, 34 },
		{ 324, 115, 34 },
		{ 324, 116, 34 },
		{ 324, 117, 34 },
		{ 324, 118, 34 },
		{ 324, 119, 34 },
		{ 324, 120, 34 },
		{ 324, 121, 34 },
		{ 324, 122, 34 },
		{ 324, 123, 34 },
		{ 324, 124, 34 },
		{ 324, 125, 34 },
		{ 324, 126, 34 },
		{ 324, 127, 34 },
		{ 324, 128, 34 },
		{ 324, 129, 34 },
		{ 324, 130, 34 },
		{ 324, 131, 34 },
		{ 324, 132, 34 },
		{ 324, 133, 34 },
		{ 324, 134, 34 },
		{ 324, 135, 34 },
		{ 324, 136, 34 },
		{ 324, 137, 34 },
		{ 324, 138, 34 },
		{ 324, 139, 34 },
		{ 324, 140, 34 },
		{ 324, 141, 34 },
		{ 324, 142, 34 },
		{ 324, 143, 34 },
		{ 324, 144, 34 },
		{ 324, 145, 34 },
		{ 324, 146, 34 },
		{ 324, 147, 34 },
		{ 324, 148, 34 },
		{ 324, 149, 34 },
		{ 325, 76, 34 },
		{ 325, 77, 34 },
		{ 325, 78, 34 },
		{ 325, 79, 34 },
		{ 325, 80, 34 },
		{ 325, 81, 34 },
		{ 325, 82, 34 },
		{ 325, 83, 34 },
		{ 325, 84, 34 },
		{ 325, 85, 34 },
		{ 325, 86, 34 },
		{ 325, 87, 34 },
		{ 325, 88, 34 },
		{ 325, 89, 34 },
		{ 325, 90, 34 },
		{ 325, 91, 34 },
		{ 325, 92, 34 },
		{ 325, 93, 34 },
		{ 325, 94, 34 },
		{ 325, 95, 34 },
		{ 325, 96, 34 },
		{ 325, 97, 34 },
		{ 325, 98, 34 },
		{ 325, 99, 34 },
		{ 325, 100, 34 },
		{ 325, 101, 34 },
		{ 325, 102, 34 },
		{ 325, 103, 34 },
		{ 325, 104, 34 },
		{ 325, 105, 34 },
		{ 325, 106, 34 },
		{ 325, 107, 34 },
		{ 325, 108, 34 },
		{ 325, 109, 34 },
		{ 325, 110, 34 },
		{ 325, 111, 34 },
		{ 325, 112, 34 },
		{ 325, 113, 34 },
		{ 325, 114, 34 },
		{ 325, 115, 34 },
		{ 325, 116, 34 },
		{ 325, 117, 34 },
		{ 325, 118, 34 },
		{ 325, 119, 34 },
		{ 325, 120, 34 },
		{ 325, 121, 34 },
		{ 325, 122, 34 },
		{ 325, 123, 34 },
		{ 325, 124, 34 },
		{ 325, 125, 34 },
		{ 325, 126, 34 },
		{ 325, 127, 34 },
		{ 325, 128, 34 },
		{ 325, 129, 34 },
		{ 325, 130, 34 },
		{ 325, 131, 34 },
		{ 325, 132, 34 },
		{ 325, 133, 34 },
		{ 325, 134, 34 },
		{ 325, 135, 34 },
		{ 325, 136, 34 },
		{ 325, 137, 34 },
		{ 325, 138, 34 },
		{ 325, 139, 34 },
		{ 325, 140, 34 },
		{ 325, 141, 34 },
		{ 325, 142, 34 },
		{ 325, 143, 34 },
		{ 325, 144, 34 },
		{ 325, 145, 34 },
		{ 325, 146, 34 },
		{ 325, 147, 34 },
		{ 325, 148, 34 },
		{ 326, 74, 34 },
		{ 326, 75, 34 },
		{ 326, 76, 34 },
		{ 326, 77, 34 },
		{ 326, 78, 34 },
		{ 326, 79, 34 },
		{ 326, 80, 34 },
		{ 326, 81, 34 },
		{ 326, 82, 34 },
		{ 326, 83, 34 },
		{ 326, 84, 34 },
		{ 326, 85, 34 },
		{ 326, 86, 34 },
		{ 326, 87, 34 },
		{ 326, 88, 34 },
		{ 326, 89, 34 },
		{ 326, 90, 34 },
		{ 326, 91, 34 },
		{ 326, 92, 34 },
		{ 326, 93, 34 },
		{ 326, 94, 34 },
		{ 326, 95, 34 },
		{ 326, 96, 34 },
		{ 326, 97, 34 },
		{ 326, 98, 34 },
		{ 326, 99, 34 },
		{ 326, 100, 34 },
		{ 326, 101, 34 },
		{ 326, 102, 34 },
		{ 326, 103, 34 },
		{ 326, 104, 34 },
		{ 326, 105, 34 },
		{ 326, 106, 34 },
		{ 326, 107, 34 },
		{ 326, 108, 34 },
		{ 326, 109, 34 },
		{ 326, 110, 34 },
		{ 326, 111, 34 },
		{ 326, 112, 34 },
		{ 326, 113, 34 },
		{ 326, 114, 34 },
		{ 326, 115, 34 },
		{ 326, 116, 34 },
		{ 326, 117, 34 },
		{ 326, 118, 34 },
		{ 326, 119, 34 },
		{ 326, 120, 34 },
		{ 326, 121, 34 },
		{ 326, 122, 34 },
		{ 326, 123, 34 },
		{ 326, 124, 34 },
		{ 326, 125, 34 },
		{ 326, 126, 34 },
		{ 326, 127, 34 },
		{ 326, 128, 34 },
		{ 326, 129, 34 },
		{ 326, 130, 34 },
		{ 326, 131, 34 },
		{ 326, 132, 34 },
		{ 326, 133, 34 },
		{ 326, 134, 34 },
		{ 326, 135, 34 },
		{ 326, 136, 34 },
		{ 326, 137, 34 },
		{ 326, 138, 34 },
		{ 326, 139, 34 },
		{ 326, 140, 34 },
		{ 326, 141, 34 },
		{ 326, 142, 34 },
		{ 326, 143, 34 },
		{ 326, 144, 34 },
		{ 326, 145, 34 },
		{ 326, 146, 34 },
		{ 327, 71, 34 },
		{ 327, 72, 34 },
		{ 327, 73, 34 },
		{ 327, 74, 34 },
		{ 327, 75, 34 },
		{ 327, 76, 34 },
		{ 327, 77, 34 },
		{ 327, 78, 34 },
		{ 327, 79, 34 },
		{ 327, 80, 34 },
		{ 327, 81, 34 },
		{ 327, 82, 34 },
		{ 327, 83, 34 },
		{ 327, 84, 34 },
		{ 327, 85, 34 },
		{ 327, 86, 34 },
		{ 327, 87, 34 },
		{ 327, 88, 34 },
		{ 327, 89, 34 },
		{ 327, 90, 34 },
		{ 327, 91, 34 },
		{ 327, 92, 34 },
		{ 327, 93, 34 },
		{ 327, 94, 34 },
		{ 327, 95, 34 },
		{ 327, 96, 34 },
		{ 327, 97, 34 },
		{ 327, 98, 34 },
		{ 327, 99, 34 },
		{ 327, 100, 34 },
		{ 327, 101, 34 },
		{ 327, 102, 34 },
		{ 327, 103, 34 },
		{ 327, 104, 34 },
		{ 327, 105, 34 },
		{ 327, 106, 34 },
		{ 327, 107, 34 },
		{ 327, 108, 34 },
		{ 327, 109, 34 },
		{ 327, 110, 34 },
		{ 327, 111, 34 },
		{ 327, 112, 34 },
		{ 327, 113, 34 },
		{ 327, 114, 34 },
		{ 327, 115, 34 },
		{ 327, 116, 34 },
		{ 327, 117, 34 },
		{ 327, 118, 34 },
		{ 327, 119, 34 },
		{ 327, 120, 34 },
		{ 327, 121, 34 },
		{ 327, 122, 34 },
		{ 327, 123, 34 },
		{ 327, 124, 34 },
		{ 327, 125, 34 },
		{ 327, 126, 34 },
		{ 327, 127, 34 },
		{ 327, 128, 34 },
		{ 327, 129, 34 },
		{ 327, 130, 34 },
		{ 327, 131, 34 },
		{ 327, 132, 34 },
		{ 327, 133, 34 },
		{ 327, 134, 34 },
		{ 327, 135, 34 },
		{ 327, 136, 34 },
		{ 327, 137, 34 },
		{ 327, 138, 34 },
		{ 327, 139, 34 },
		{ 327, 140, 34 },
		{ 327, 141, 34 },
		{ 327, 142, 34 },
		{ 327, 143, 34 },
		{ 327, 144, 34 },
		{ 328, 68, 34 },
		{ 328, 69, 34 },
		{ 328, 70, 34 },
		{ 328, 71, 34 },
		{ 328, 72, 34 },
		{ 328, 73, 34 },
		{ 328, 74, 34 },
		{ 328, 75, 34 },
		{ 328, 76, 34 },
		{ 328, 77, 34 },
		{ 328, 78, 34 },
		{ 328, 79, 34 },
		{ 328, 80, 34 },
		{ 328, 81, 34 },
		{ 328, 82, 34 },
		{ 328, 83, 34 },
		{ 328, 84, 34 },
		{ 328, 85, 34 },
		{ 328, 86, 34 },
		{ 328, 87, 34 },
		{ 328, 88, 34 },
		{ 328, 89, 34 },
		{ 328, 90, 34 },
		{ 328, 91, 34 },
		{ 328, 92, 34 },
		{ 328, 93, 34 },
		{ 328, 94, 34 },
		{ 328, 95, 34 },
		{ 328, 96, 34 },
		{ 328, 97, 34 },
		{ 328, 98, 34 },
		{ 328, 99, 34 },
		{ 328, 100, 34 },
		{ 328, 101, 34 },
		{ 328, 102, 34 },
		{ 328, 103, 34 },
		{ 328, 104, 34 },
		{ 328, 105, 34 },
		{ 328, 106, 34 },
		{ 328, 107, 34 },
		{ 328, 108, 34 },
		{ 328, 109, 34 },
		{ 328, 110, 34 },
		{ 328, 111, 34 },
		{ 328, 112, 34 },
		{ 328, 113, 34 },
		{ 328, 114, 34 },
		{ 328, 115, 34 },
		{ 328, 116, 34 },
		{ 328, 117, 34 },
		{ 328, 118, 34 },
		{ 328, 119, 34 },
		{ 328, 120, 34 },
		{ 328, 121, 34 },
		{ 328, 122, 34 },
		{ 328, 123, 34 },
		{ 328, 124, 34 },
		{ 328, 125, 34 },
		{ 328, 126, 34 },
		{ 328, 127, 34 },
		{ 328, 128, 34 },
		{ 328, 129, 34 },
		{ 328, 130, 34 },
		{ 328, 131, 34 },
		{ 328, 132, 34 },
		{ 328, 133, 34 },
		{ 328, 134, 34 },
		{ 328, 135, 34 },
		{ 328, 136, 34 },
		{ 328, 137, 34 },
		{ 328, 138, 34 },
		{ 328, 139, 34 },
		{ 328, 140, 34 },
		{ 328, 141, 34 },
		{ 328, 142, 34 },
		{ 328, 143, 34 },
		{ 329, 66, 34 },
		{ 329, 67, 34 },
		{ 329, 68, 34 },
		{ 329, 69, 34 },
		{ 329, 70, 34 },
		{ 329, 71, 34 },
		{ 329, 72, 34 },
		{ 329, 73, 34 },
		{ 329, 74, 34 },
		{ 329, 75, 34 },
		{ 329, 76, 34 },
		{ 329, 77, 34 },
		{ 329, 78, 34 },
		{ 329, 79, 34 },
		{ 329, 80, 34 },
		{ 329, 81, 34 },
		{ 329, 82, 34 },
		{ 329, 83, 34 },
		{ 329, 84, 34 },
		{ 329, 85, 34 },
		{ 329, 86, 34 },
		{ 329, 87, 34 },
		{ 329, 88, 34 },
		{ 329, 89, 34 },
		{ 329, 90, 34 },
		{ 329, 91, 34 },
		{ 329, 92, 34 },
		{ 329, 93, 34 },
		{ 329, 94, 34 },
		{ 329, 95, 34 },
		{ 329, 96, 34 },
		{ 329, 97, 34 },
		{ 329, 98, 34 },
		{ 329, 99, 34 },
		{ 329, 100, 34 },
		{ 329, 101, 34 },
		{ 329, 102, 34 },
		{ 329, 103, 34 },
		{ 329, 104, 34 },
		{ 329, 105, 34 },
		{ 329, 106, 34 },
		{ 329, 107, 34 },
		{ 329, 108, 34 },
		{ 329, 109, 34 },
		{ 329, 110, 34 },
		{ 329, 111, 34 },
		{ 329, 112, 34 },
		{ 329, 113, 34 },
		{ 329, 114, 34 },
		{ 329, 115, 34 },
		{ 329, 116, 34 },
		{ 329, 117, 34 },
		{ 329, 118, 34 },
		{ 329, 119, 34 },
		{ 329, 120, 34 },
		{ 329, 121, 34 },
		{ 329, 122, 34 },
		{ 329, 123, 34 },
		{ 329, 124, 34 },
		{ 329, 125, 34 },
		{ 329, 126, 34 },
		{ 329, 127, 34 },
		{ 329, 128, 34 },
		{ 329, 129, 34 },
		{ 329, 130, 34 },
		{ 329, 131, 34 },
		{ 329, 132, 34 },
		{ 329, 133, 34 },
		{ 329, 134, 34 },
		{ 329, 135, 34 },
		{ 329, 136, 34 },
		{ 329, 137, 34 },
		{ 329, 138, 34 },
		{ 329, 139, 34 },
		{ 329, 140, 34 },
		{ 329, 141, 34 },
		{ 330, 63, 34 },
		{ 330, 64, 34 },
		{ 330, 65, 34 },
		{ 330, 66, 34 },
		{ 330, 67, 34 },
		{ 330, 68, 34 },
		{ 330, 69, 34 },
		{ 330, 70, 34 },
		{ 330, 71, 34 },
		{ 330, 72, 34 },
		{ 330, 73, 34 },
		{ 330, 74, 34 },
		{ 330, 75, 34 },
		{ 330, 76, 34 },
		{ 330, 77, 34 },
		{ 330, 78, 34 },
		{ 330, 79, 34 },
		{ 330, 80, 34 },
		{ 330, 81, 34 },
		{ 330, 82, 34 },
		{ 330, 83, 34 },
		{ 330, 84, 34 },
		{ 330, 85, 34 },
		{ 330, 86, 34 },
		{ 330, 87, 34 },
		{ 330, 88, 34 },
		{ 330, 89, 34 },
		{ 330, 90, 34 },
		{ 330, 91, 34 },
		{ 330, 92, 34 },
		{ 330, 93, 34 },
		{ 330, 94, 34 },
		{ 330, 95, 34 },
		{ 330, 96, 34 },
		{ 330, 97, 34 },
		{ 330, 98, 34 },
		{ 330, 99, 34 },
		{ 330, 100, 34 },
		{ 330, 101, 34 },
		{ 330, 102, 34 },
		{ 330, 103, 34 },
		{ 330, 104, 34 },
		{ 330, 105, 34 },
		{ 330, 106, 34 },
		{ 330, 107, 34 },
		{ 330, 108, 34 },
		{ 330, 109, 34 },
		{ 330, 110, 34 },
		{ 330, 111, 34 },
		{ 330, 112, 34 },
		{ 330, 113, 34 },
		{ 330, 114, 34 },
		{ 330, 115, 34 },
		{ 330, 116, 34 },
		{ 330, 117, 34 },
		{ 330, 118, 34 },
		{ 330, 119, 34 },
		{ 330, 120, 34 },
		{ 330, 121, 34 },
		{ 330, 122, 34 },
		{ 330, 123, 34 },
		{ 330, 124, 34 },
		{ 330, 125, 34 },
		{ 330, 126, 34 },
		{ 330, 127, 34 },
		{ 330, 128, 34 },
		{ 330, 129, 34 },
		{ 330, 130, 34 },
		{ 330, 131, 34 },
		{ 330, 132, 34 },
		{ 330, 133, 34 },
		{ 330, 134, 34 },
		{ 330, 135, 34 },
		{ 330, 136, 34 },
		{ 330, 137, 34 },
		{ 330, 138, 34 },
		{ 330, 139, 34 },
		{ 330, 140, 34 },
		{ 331, 60, 34 },
		{ 331, 61, 34 },
		{ 331, 62, 34 },
		{ 331, 63, 34 },
		{ 331, 64, 34 },
		{ 331, 65, 34 },
		{ 331, 66, 34 },
		{ 331, 67, 34 },
		{ 331, 68, 34 },
		{ 331, 69, 34 },
		{ 331, 70, 34 },
		{ 331, 71, 34 },
		{ 331, 72, 34 },
		{ 331, 73, 34 },
		{ 331, 74, 34 },
		{ 331, 75, 34 },
		{ 331, 76, 34 },
		{ 331, 77, 34 },
		{ 331, 78, 34 },
		{ 331, 79, 34 },
		{ 331, 80, 34 },
		{ 331, 81, 34 },
		{ 331, 82, 34 },
		{ 331, 83, 34 },
		{ 331, 84, 34 },
		{ 331, 85, 34 },
		{ 331, 86, 34 },
		{ 331, 87, 34 },
		{ 331, 88, 34 },
		{ 331, 89, 34 },
		{ 331, 90, 34 },
		{ 331, 91, 34 },
		{ 331, 92, 34 },
		{ 331, 93, 34 },
		{ 331, 94, 34 },
		{ 331, 95, 34 },
		{ 331, 96, 34 },
		{ 331, 97, 34 },
		{ 331, 98, 34 },
		{ 331, 99, 34 },
		{ 331, 100, 34 },
		{ 331, 101, 34 },
		{ 331, 102, 34 },
		{ 331, 103, 34 },
		{ 331, 104, 34 },
		{ 331, 105, 34 },
		{ 331, 106, 34 },
		{ 331, 107, 34 },
		{ 331, 108, 34 },
		{ 331, 109, 34 },
		{ 331, 110, 34 },
		{ 331, 111, 34 },
		{ 331, 112, 34 },
		{ 331, 113, 34 },
		{ 331, 114, 34 },
		{ 331, 115, 34 },
		{ 331, 116, 34 },
		{ 331, 117, 34 },
		{ 331, 118, 34 },
		{ 331, 119, 34 },
		{ 331, 120, 34 },
		{ 331, 121, 34 },
		{ 331, 122, 34 },
		{ 331, 123, 34 },
		{ 331, 124, 34 },
		{ 331, 125, 34 },
		{ 331, 126, 34 },
		{ 331, 127, 34 },
		{ 331, 128, 34 },
		{ 331, 129, 34 },
		{ 331, 130, 34 },
		{ 331, 131, 34 },
		{ 331, 132, 34 },
		{ 331, 133, 34 },
		{ 331, 134, 34 },
		{ 331, 135, 34 },
		{ 331, 136, 34 },
		{ 331, 137, 34 },
		{ 331, 138, 34 },
		{ 332, 58, 34 },
		{ 332, 59, 34 },
		{ 332, 60, 34 },
		{ 332, 61, 34 },
		{ 332, 62, 34 },
		{ 332, 63, 34 },
		{ 332, 64, 34 },
		{ 332, 65, 34 },
		{ 332, 66, 34 },
		{ 332, 67, 34 },
		{ 332, 68, 34 },
		{ 332, 69, 34 },
		{ 332, 70, 34 },
		{ 332, 71, 34 },
		{ 332, 72, 34 },
		{ 332, 73, 34 },
		{ 332, 74, 34 },
		{ 332, 75, 34 },
		{ 332, 76, 34 },
		{ 332, 77, 34 },
		{ 332, 78, 34 },
		{ 332, 79, 34 },
		{ 332, 80, 34 },
		{ 332, 81, 34 },
		{ 332, 82, 34 },
		{ 332, 83, 34 },
		{ 332, 84, 34 },
		{ 332, 85, 34 },
		{ 332, 86, 34 },
		{ 332, 87, 34 },
		{ 332, 88, 34 },
		{ 332, 89, 34 },
		{ 332, 90, 34 },
		{ 332, 91, 34 },
		{ 332, 92, 34 },
		{ 332, 93, 34 },
		{ 332, 94, 34 },
		{ 332, 95, 34 },
		{ 332, 96, 34 },
		{ 332, 97, 34 },
		{ 332, 98, 34 },
		{ 332, 99, 34 },
		{ 332, 100, 34 },
		{ 332, 101, 34 },
		{ 332, 102, 34 },
		{ 332, 103, 34 },
		{ 332, 104, 34 },
		{ 332, 105, 34 },
		{ 332, 106, 34 },
		{ 332, 107, 34 },
		{ 332, 108, 34 },
		{ 332, 109, 34 },
		{ 332, 110, 34 },
		{ 332, 111, 34 },
		{ 332, 112, 34 },
		{ 332, 113, 34 },
		{ 332, 114, 34 },
		{ 332, 115, 34 },
		{ 332, 116, 34 },
		{ 332, 117, 34 },
		{ 332, 118, 34 },
		{ 332, 119, 34 },
		{ 332, 120, 34 },
		{ 332, 121, 34 },
		{ 332, 122, 34 },
		{ 332, 123, 34 },
		{ 332, 124, 34 },
		{ 332, 125, 34 },
		{ 332, 126, 34 },
		{ 332, 127, 34 },
		{ 332, 128, 34 },
		{ 332, 129, 34 },
		{ 332, 130, 34 },
		{ 332, 131, 34 },
		{ 332, 132, 34 },
		{ 332, 133, 34 },
		{ 332, 134, 34 },
		{ 332, 135, 34 },
		{ 332, 136, 34 },
		{ 333, 55, 34 },
		{ 333, 56, 34 },
		{ 333, 57, 34 },
		{ 333, 58, 34 },
		{ 333, 59, 34 },
		{ 333, 60, 34 },
		{ 333, 61, 34 },
		{ 333, 62, 34 },
		{ 333, 63, 34 },
		{ 333, 64, 34 },
		{ 333, 65, 34 },
		{ 333, 66, 34 },
		{ 333, 67, 34 },
		{ 333, 68, 34 },
		{ 333, 69, 34 },
		{ 333, 70, 34 },
		{ 333, 71, 34 },
		{ 333, 72, 34 },
		{ 333, 73, 34 },
		{ 333, 74, 34 },
		{ 333, 75, 34 },
		{ 333, 76, 34 },
		{ 333, 77, 34 },
		{ 333, 78, 34 },
		{ 333, 79, 34 },
		{ 333, 80, 34 },
		{ 333, 81, 34 },
		{ 333, 82, 34 },
		{ 333, 83, 34 },
		{ 333, 84, 34 },
		{ 333, 85, 34 },
		{ 333, 86, 34 },
		{ 333, 87, 34 },
		{ 333, 88, 34 },
		{ 333, 89, 34 },
		{ 333, 90, 34 },
		{ 333, 91, 34 },
		{ 333, 92, 34 },
		{ 333, 93, 34 },
		{ 333, 94, 34 },
		{ 333, 95, 34 },
		{ 333, 96, 34 },
		{ 333, 97, 34 },
		{ 333, 98, 34 },
		{ 333, 99, 34 },
		{ 333, 100, 34 },
		{ 333, 101, 34 },
		{ 333, 102, 34 },
		{ 333, 103, 34 },
		{ 333, 104, 34 },
		{ 333, 105, 34 },
		{ 333, 106, 34 },
		{ 333, 107, 34 },
		{ 333, 108, 34 },
		{ 333, 109, 34 },
		{ 333, 110, 34 },
		{ 333, 111, 34 },
		{ 333, 112, 34 },
		{ 333, 113, 34 },
		{ 333, 114, 34 },
		{ 333, 115, 34 },
		{ 333, 116, 34 },
		{ 333, 117, 34 },
		{ 333, 118, 34 },
		{ 333, 119, 34 },
		{ 333, 120, 34 },
		{ 333, 121, 34 },
		{ 333, 122, 34 },
		{ 333, 123, 34 },
		{ 333, 124, 34 },
		{ 333, 125, 34 },
		{ 333, 126, 34 },
		{ 333, 127, 34 },
		{ 333, 128, 34 },
		{ 333, 129, 34 },
		{ 333, 130, 34 },
		{ 333, 131, 34 },
		{ 333, 132, 34 },
		{ 333, 133, 34 },
		{ 333, 134, 34 },
		{ 333, 135, 34 },
		{ 334, 52, 34 },
		{ 334, 53, 34 },
		{ 334, 54, 34 },
		{ 334, 55, 34 },
		{ 334, 56, 34 },
		{ 334, 57, 34 },
		{ 334, 58, 34 },
		{ 334, 59, 34 },
		{ 334, 60, 34 },
		{ 334, 61, 34 },
		{ 334, 62, 34 },
		{ 334, 63, 34 },
		{ 334, 64, 34 },
		{ 334, 65, 34 },
		{ 334, 66, 34 },
		{ 334, 67, 34 },
		{ 334, 68, 34 },
		{ 334, 69, 34 },
		{ 334, 70, 34 },
		{ 334, 71, 34 },
		{ 334, 72, 34 },
		{ 334, 73, 34 },
		{ 334, 74, 34 },
		{ 334, 75, 34 },
		{ 334, 76, 34 },
		{ 334, 77, 34 },
		{ 334, 78, 34 },
		{ 334, 79, 34 },
		{ 334, 80, 34 },
		{ 334, 81, 34 },
		{ 334, 82, 34 },
		{ 334, 83, 34 },
		{ 334, 84, 34 },
		{ 334, 85, 34 },
		{ 334, 86, 34 },
		{ 334, 87, 34 },
		{ 334, 88, 34 },
		{ 334, 89, 34 },
		{ 334, 90, 34 },
		{ 334, 91, 34 },
		{ 334, 92, 34 },
		{ 334, 93, 34 },
		{ 334, 94, 34 },
		{ 334, 95, 34 },
		{ 334, 96, 34 },
		{ 334, 97, 34 },
		{ 334, 98, 34 },
		{ 334, 99, 34 },
		{ 334, 100, 34 },
		{ 334, 101, 34 },
		{ 334, 102, 34 },
		{ 334, 103, 34 },
		{ 334, 104, 34 },
		{ 334, 105, 34 },
		{ 334, 106, 34 },
		{ 334, 107, 34 },
		{ 334, 108, 34 },
		{ 334, 109, 34 },
		{ 334, 110, 34 },
		{ 334, 111, 34 },
		{ 334, 112, 34 },
		{ 334, 113, 34 },
		{ 334, 114, 34 },
		{ 334, 115, 34 },
		{ 334, 116, 34 },
		{ 334, 117, 34 },
		{ 334, 118, 34 },
		{ 334, 119, 34 },
		{ 334, 120, 34 },
		{ 334, 121, 34 },
		{ 334, 122, 34 },
		{ 334, 123, 34 },
		{ 334, 124, 34 },
		{ 334, 125, 34 },
		{ 334, 126, 34 },
		{ 334, 127, 34 },
		{ 334, 128, 34 },
		{ 334, 129, 34 },
		{ 334, 130, 34 },
		{ 334, 131, 34 },
		{ 334, 132, 34 },
		{ 334, 133, 34 },
		{ 335, 50, 34 },
		{ 335, 51, 34 },
		{ 335, 52, 34 },
		{ 335, 53, 34 },
		{ 335, 54, 34 },
		{ 335, 55, 34 },
		{ 335, 56, 34 },
		{ 335, 57, 34 },
		{ 335, 58, 34 },
		{ 335, 59, 34 },
		{ 335, 60, 34 },
		{ 335, 61, 34 },
		{ 335, 62, 34 },
		{ 335, 63, 34 },
		{ 335, 64, 34 },
		{ 335, 65, 34 },
		{ 335, 66, 34 },
		{ 335, 67, 34 },
		{ 335, 68, 34 },
		{ 335, 69, 34 },
		{ 335, 70, 34 },
		{ 335, 71, 34 },
		{ 335, 72, 34 },
		{ 335, 73, 34 },
		{ 335, 74, 34 },
		{ 335, 75, 34 },
		{ 335, 76, 34 },
		{ 335, 77, 34 },
		{ 335, 78, 34 },
		{ 335, 79, 34 },
		{ 335, 80, 34 },
		{ 335, 81, 34 },
		{ 335, 82, 34 },
		{ 335, 83, 34 },
		{ 335, 84, 34 },
		{ 335, 85, 34 },
		{ 335, 86, 34 },
		{ 335, 87, 34 },
		{ 335, 88, 34 },
		{ 335, 89, 34 },
		{ 335, 90, 34 },
		{ 335, 91, 34 },
		{ 335, 92, 34 },
		{ 335, 93, 34 },
		{ 335, 94, 34 },
		{ 335, 95, 34 },
		{ 335, 96, 34 },
		{ 335, 97, 34 },
		{ 335, 98, 34 },
		{ 335, 99, 34 },
		{ 335, 100, 34 },
		{ 335, 101, 34 },
		{ 335, 102, 34 },
		{ 335, 103, 34 },
		{ 335, 104, 34 },
		{ 335, 105, 34 },
		{ 335, 106, 34 },
		{ 335, 107, 34 },
		{ 335, 108, 34 },
		{ 335, 109, 34 },
		{ 335, 110, 34 },
		{ 335, 111, 34 },
		{ 335, 112, 34 },
		{ 335, 113, 34 },
		{ 335, 114, 34 },
		{ 335, 115, 34 },
		{ 335, 116, 34 },
		{ 335, 117, 34 },
		{ 335, 118, 34 },
		{ 335, 119, 34 },
		{ 335, 120, 34 },
		{ 335, 121, 34 },
		{ 335, 122, 34 },
		{ 335, 123, 34 },
		{ 335, 124, 34 },
		{ 335, 125, 34 },
		{ 335, 126, 34 },
		{ 335, 127, 34 },
		{ 335, 128, 34 },
		{ 335, 129, 34 },
		{ 335, 130, 34 },
		{ 335, 131, 34 },
		{ 335, 132, 34 },
		{ 336, 47, 34 },
		{ 336, 48, 34 },
		{ 336, 49, 34 },
		{ 336, 50, 34 },
		{ 336, 51, 34 },
		{ 336, 52, 34 },
		{ 336, 53, 34 },
		{ 336, 54, 34 },
		{ 336, 55, 34 },
		{ 336, 56, 34 },
		{ 336, 57, 34 },
		{ 336, 58, 34 },
		{ 336, 59, 34 },
		{ 336, 60, 34 },
		{ 336, 61, 34 },
		{ 336, 62, 34 },
		{ 336, 63, 34 },
		{ 336, 64, 34 },
		{ 336, 65, 34 },
		{ 336, 66, 34 },
		{ 336, 67, 34 },
		{ 336, 68, 34 },
		{ 336, 69, 34 },
		{ 336, 70, 34 },
		{ 336, 71, 34 },
		{ 336, 72, 34 },
		{ 336, 73, 34 },
		{ 336, 74, 34 },
		{ 336, 75, 34 },
		{ 336, 76, 34 },
		{ 336, 77, 34 },
		{ 336, 78, 34 },
		{ 336, 79, 34 },
		{ 336, 80, 34 },
		{ 336, 81, 34 },
		{ 336, 82, 34 },
		{ 336, 83, 34 },
		{ 336, 84, 34 },
		{ 336, 85, 34 },
		{ 336, 86, 34 },
		{ 336, 87, 34 },
		{ 336, 88, 34 },
		{ 336, 89, 34 },
		{ 336, 90, 34 },
		{ 336, 91, 34 },
		{ 336, 92, 34 },
		{ 336, 93, 34 },
		{ 336, 94, 34 },
		{ 336, 95, 34 },
		{ 336, 96, 34 },
		{ 336, 97, 34 },
		{ 336, 98, 34 },
		{ 336, 99, 34 },
		{ 336, 100, 34 },
		{ 336, 101, 34 },
		{ 336, 102, 34 },
		{ 336, 103, 34 },
		{ 336, 104, 34 },
		{ 336, 105, 34 },
		{ 336, 106, 34 },
		{ 336, 107, 34 },
		{ 336, 108, 34 },
		{ 336, 109, 34 },
		{ 336, 110, 34 },
		{ 336, 111, 34 },
		{ 336, 112, 34 },
		{ 336, 113, 34 },
		{ 336, 114, 34 },
		{ 336, 115, 34 },
		{ 336, 116, 34 },
		{ 336, 117, 34 },
		{ 336, 118, 34 },
		{ 336, 119, 34 },
		{ 336, 120, 34 },
		{ 336, 121, 34 },
		{ 336, 122, 34 },
		{ 336, 123, 34 },
		{ 336, 124, 34 },
		{ 336, 125, 34 },
		{ 336, 126, 34 },
		{ 336, 127, 34 },
		{ 336, 128, 34 },
		{ 336, 129, 34 },
		{ 336, 130, 34 },
		{ 337, 44, 34 },
		{ 337, 45, 34 },
		{ 337, 46, 34 },
		{ 337, 47, 34 },
		{ 337, 48, 34 },
		{ 337, 49, 34 },
		{ 337, 50, 34 },
		{ 337, 51, 34 },
		{ 337, 52, 34 },
		{ 337, 53, 34 },
		{ 337, 54, 34 },
		{ 337, 55, 34 },
		{ 337, 56, 34 },
		{ 337, 57, 34 },
		{ 337, 58, 34 },
		{ 337, 59, 34 },
		{ 337, 60, 34 },
		{ 337, 61, 34 },
		{ 337, 62, 34 },
		{ 337, 63, 34 },
		{ 337, 64, 34 },
		{ 337, 65, 34 },
		{ 337, 66, 34 },
		{ 337, 67, 34 },
		{ 337, 68, 34 },
		{ 337, 69, 34 },
		{ 337, 70, 34 },
		{ 337, 71, 34 },
		{ 337, 72, 34 },
		{ 337, 73, 34 },
		{ 337, 74, 34 },
		{ 337, 75, 34 },
		{ 337, 76, 34 },
		{ 337, 77, 34 },
		{ 337, 78, 34 },
		{ 337, 79, 34 },
		{ 337, 80, 34 },
		{ 337, 81, 34 },
		{ 337, 82, 34 },
		{ 337, 83, 34 },
		{ 337, 84, 34 },
		{ 337, 85, 34 },
		{ 337, 86, 34 },
		{ 337, 87, 34 },
		{ 337, 88, 34 },
		{ 337, 89, 34 },
		{ 337, 90, 34 },
		{ 337, 91, 34 },
		{ 337, 92, 34 },
		{ 337, 93, 34 },
		{ 337, 94, 34 },
		{ 337, 95, 34 },
		{ 337, 96, 34 },
		{ 337, 97, 34 },
		{ 337, 98, 34 },
		{ 337, 99, 34 },
		{ 337, 100, 34 },
		{ 337, 101, 34 },
		{ 337, 102, 34 },
		{ 337, 103, 34 },
		{ 337, 104, 34 },
		{ 337, 105, 34 },
		{ 337, 106, 34 },
		{ 337, 107, 34 },
		{ 337, 108, 34 },
		{ 337, 109, 34 },
		{ 337, 110, 34 },
		{ 337, 111, 34 },
		{ 337, 112, 34 },
		{ 337, 113, 34 },
		{ 337, 114, 34 },
		{ 337, 115, 34 },
		{ 337, 116, 34 },
		{ 337, 117, 34 },
		{ 337, 118, 34 },
		{ 337, 119, 34 },
		{ 337, 120, 34 },
		{ 337, 121, 34 },
		{ 337, 122, 34 },
		{ 337, 123, 34 },
		{ 337, 124, 34 },
		{ 337, 125, 34 },
		{ 337, 126, 34 },
		{ 337, 127, 34 },
		{ 337, 128, 34 },
		{ 338, 42, 34 },
		{ 338, 43, 34 },
		{ 338, 44, 34 },
		{ 338, 45, 34 },
		{ 338, 46, 34 },
		{ 338, 47, 34 },
		{ 338, 48, 34 },
		{ 338, 49, 34 },
		{ 338, 50, 34 },
		{ 338, 51, 34 },
		{ 338, 52, 34 },
		{ 338, 53, 34 },
		{ 338, 54, 34 },
		{ 338, 55, 34 },
		{ 338, 56, 34 },
		{ 338, 57, 34 },
		{ 338, 58, 34 },
		{ 338, 59, 34 },
		{ 338, 60, 34 },
		{ 338, 61, 34 },
		{ 338, 62, 34 },
		{ 338, 63, 34 },
		{ 338, 64, 34 },
		{ 338, 65, 34 },
		{ 338, 66, 34 },
		{ 338, 67, 34 },
		{ 338, 68, 34 },
		{ 338, 69, 34 },
		{ 338, 70, 34 },
		{ 338, 71, 34 },
		{ 338, 72, 34 },
		{ 338, 73, 34 },
		{ 338, 74, 34 },
		{ 338, 75, 34 },
		{ 338, 76, 34 },
		{ 338, 77, 34 },
		{ 338, 78, 34 },
		{ 338, 79, 34 },
		{ 338, 80, 34 },
		{ 338, 81, 34 },
		{ 338, 82, 34 },
		{ 338, 83, 34 },
		{ 338, 84, 34 },
		{ 338, 85, 34 },
		{ 338, 86, 34 },
		{ 338, 87, 34 },
		{ 338, 88, 34 },
		{ 338, 89, 34 },
		{ 338, 90, 34 },
		{ 338, 91, 34 },
		{ 338, 92, 34 },
		{ 338, 93, 34 },
		{ 338, 94, 34 },
		{ 338, 95, 34 },
		{ 338, 96, 34 },
		{ 338, 97, 34 },
		{ 338, 98, 34 },
		{ 338, 99, 34 },
		{ 338, 100, 34 },
		{ 338, 101, 34 },
		{ 338, 102, 34 },
		{ 338, 103, 34 },
		{ 338, 104, 34 },
		{ 338, 105, 34 },
		{ 338, 106, 34 },
		{ 338, 107, 34 },
		{ 338, 108, 34 },
		{ 338, 109, 34 },
		{ 338, 110, 34 },
		{ 338, 111, 34 },
		{ 338, 112, 34 },
		{ 338, 113, 34 },
		{ 338, 114, 34 },
		{ 338, 115, 34 },
		{ 338, 116, 34 },
		{ 338, 117, 34 },
		{ 338, 118, 34 },
		{ 338, 119, 34 },
		{ 338, 120, 34 },
		{ 338, 121, 34 },
		{ 338, 122, 34 },
		{ 338, 123, 34 },
		{ 338, 124, 34 },
		{ 338, 125, 34 },
		{ 338, 126, 34 },
		{ 338, 127, 34 },
		{ 339, 39, 34 },
		{ 339, 40, 34 },
		{ 339, 41, 34 },
		{ 339, 42, 34 },
		{ 339, 43, 34 },
		{ 339, 44, 34 },
		{ 339, 45, 34 },
		{ 339, 46, 34 },
		{ 339, 47, 34 },
		{ 339, 48, 34 },
		{ 339, 49, 34 },
		{ 339, 50, 34 },
		{ 339, 51, 34 },
		{ 339, 52, 34 },
		{ 339, 53, 34 },
		{ 339, 54, 34 },
		{ 339, 55, 34 },
		{ 339, 56, 34 },
		{ 339, 57, 34 },
		{ 339, 58, 34 },
		{ 339, 59, 34 },
		{ 339, 60, 34 },
		{ 339, 61, 34 },
		{ 339, 62, 34 },
		{ 339, 63, 34 },
		{ 339, 64, 34 },
		{ 339, 65, 34 },
		{ 339, 66, 34 },
		{ 339, 67, 34 },
		{ 339, 68, 34 },
		{ 339, 69, 34 },
		{ 339, 70, 34 },
		{ 339, 71, 34 },
		{ 339, 72, 34 },
		{ 339, 73, 34 },
		{ 339, 74, 34 },
		{ 339, 75, 34 },
		{ 339, 76, 34 },
		{ 339, 77, 34 },
		{ 339, 78, 34 },
		{ 339, 79, 34 },
		{ 339, 80, 34 },
		{ 339, 81, 34 },
		{ 339, 82, 34 },
		{ 339, 83, 34 },
		{ 339, 84, 34 },
		{ 339, 85, 34 },
		{ 339, 86, 34 },
		{ 339, 87, 34 },
		{ 339, 88, 34 },
		{ 339, 89, 34 },
		{ 339, 90, 34 },
		{ 339, 91, 34 },
		{ 339, 92, 34 },
		{ 339, 93, 34 },
		{ 339, 94, 34 },
		{ 339, 95, 34 },
		{ 339, 96, 34 },
		{ 339, 97, 34 },
		{ 339, 98, 34 },
		{ 339, 99, 34 },
		{ 339, 100, 34 },
		{ 339, 101, 34 },
		{ 339, 102, 34 },
		{ 339, 103, 34 },
		{ 339, 104, 34 },
		{ 339, 105, 34 },
		{ 339, 106, 34 },
		{ 339, 107, 34 },
		{ 339, 108, 34 },
		{ 339, 109, 34 },
		{ 339, 110, 34 },
		{ 339, 111, 34 },
		{ 339, 112, 34 },
		{ 339, 113, 34 },
		{ 339, 114, 34 },
		{ 339, 115, 34 },
		{ 339, 116, 34 },
		{ 339, 117, 34 },
		{ 339, 118, 34 },
		{ 339, 119, 34 },
		{ 339, 120, 34 },
		{ 339, 121, 34 },
		{ 339, 122, 34 },
		{ 339, 123, 34 },
		{ 339, 124, 34 },
		{ 339, 125, 34 },
		{ 340, 36, 34 },
		{ 340, 37, 34 },
		{ 340, 38, 34 },
		{ 340, 39, 34 },
		{ 340, 40, 34 },
		{ 340, 41, 34 },
		{ 340, 42, 34 },
		{ 340, 43, 34 },
		{ 340, 44, 34 },
		{ 340, 45, 34 },
		{ 340, 46, 34 },
		{ 340, 47, 34 },
		{ 340, 48, 34 },
		{ 340, 49, 34 },
		{ 340, 50, 34 },
		{ 340, 51, 34 },
		{ 340, 52, 34 },
		{ 340, 53, 34 },
		{ 340, 54, 34 },
		{ 340, 55, 34 },
		{ 340, 56, 34 },
		{ 340, 57, 34 },
		{ 340, 58, 34 },
		{ 340, 59, 34 },
		{ 340, 60, 34 },
		{ 340, 61, 34 },
		{ 340, 62, 34 },
		{ 340, 63, 34 },
		{ 340, 64, 34 },
		{ 340, 65, 34 },
		{ 340, 66, 34 },
		{ 340, 67, 34 },
		{ 340, 68, 34 },
		{ 340, 69, 34 },
		{ 340, 70, 34 },
		{ 340, 71, 34 },
		{ 340, 72, 34 },
		{ 340, 73, 34 },
		{ 340, 74, 34 },
		{ 340, 75, 34 },
		{ 340, 76, 34 },
		{ 340, 77, 34 },
		{ 340, 78, 34 },
		{ 340, 79, 34 },
		{ 340, 80, 34 },
		{ 340, 81, 34 },
		{ 340, 82, 34 },
		{ 340, 83, 34 },
		{ 340, 84, 34 },
		{ 340, 85, 34 },
		{ 340, 86, 34 },
		{ 340, 87, 34 },
		{ 340, 88, 34 },
		{ 340, 89, 34 },
		{ 340, 90, 34 },
		{ 340, 91, 34 },
		{ 340, 92, 34 },
		{ 340, 93, 34 },
		{ 340, 94, 34 },
		{ 340, 95, 34 },
		{ 340, 96, 34 },
		{ 340, 97, 34 },
		{ 340, 98, 34 },
		{ 340, 99, 34 },
		{ 340, 100, 34 },
		{ 340, 101, 34 },
		{ 340, 102, 34 },
		{ 340, 103, 34 },
		{ 340, 104, 34 },
		{ 340, 105, 34 },
		{ 340, 106, 34 },
		{ 340, 107, 34 },
		{ 340, 108, 34 },
		{ 340, 109, 34 },
		{ 340, 110, 34 },
		{ 340, 111, 34 },
		{ 340, 112, 34 },
		{ 340, 113, 34 },
		{ 340, 114, 34 },
		{ 340, 115, 34 },
		{ 340, 116, 34 },
		{ 340, 117, 34 },
		{ 340, 118, 34 },
		{ 340, 119, 34 },
		{ 340, 120, 34 },
		{ 340, 121, 34 },
		{ 340, 122, 34 },
		{ 340, 123, 34 },
		{ 340, 124, 34 },
		{ 341, 34, 34 },
		{ 341, 35, 34 },
		{ 341, 36, 34 },
		{ 341, 37, 34 },
		{ 341, 38, 34 },
		{ 341, 39, 34 },
		{ 341, 40, 34 },
		{ 341, 41, 34 },
		{ 341, 42, 34 },
		{ 341, 43, 34 },
		{ 341, 44, 34 },
		{ 341, 45, 34 },
		{ 341, 46, 34 },
		{ 341, 47, 34 },
		{ 341, 48, 34 },
		{ 341, 49, 34 },
		{ 341, 50, 34 },
		{ 341, 51, 34 },
		{ 341, 52, 34 },
		{ 341, 53, 34 },
		{ 341, 54, 34 },
		{ 341, 55, 34 },
		{ 341, 56, 34 },
		{ 341, 57, 34 },
		{ 341, 58, 34 },
		{ 341, 59, 34 },
		{ 341, 60, 34 },
		{ 341, 61, 34 },
		{ 341, 62, 34 },
		{ 341, 63, 34 },
		{ 341, 64, 34 },
		{ 341, 65, 34 },
		{ 341, 66, 34 },
		{ 341, 67, 34 },
		{ 341, 68, 34 },
		{ 341, 69, 34 },
		{ 341, 70, 34 },
		{ 341, 71, 34 },
		{ 341, 72, 34 },
		{ 341, 73, 34 },
		{ 341, 74, 34 },
		{ 341, 75, 34 },
		{ 341, 76, 34 },
		{ 341, 77, 34 },
		{ 341, 78, 34 },
		{ 341, 79, 34 },
		{ 341, 80, 34 },
		{ 341, 81, 34 },
		{ 341, 82, 34 },
		{ 341, 83, 34 },
		{ 341, 84, 34 },
		{ 341, 85, 34 },
		{ 341, 86, 34 },
		{ 341, 87, 34 },
		{ 341, 88, 34 },
		{ 341, 89, 34 },
		{ 341, 90, 34 },
		{ 341, 91, 34 },
		{ 341, 92, 34 },
		{ 341, 93, 34 },
		{ 341, 94, 34 },
		{ 341, 95, 34 },
		{ 341, 96, 34 },
		{ 341, 97, 34 },
		{ 341, 98, 34 },
		{ 341, 99, 34 },
		{ 341, 100, 34 },
		{ 341, 101, 34 },
		{ 341, 102, 34 },
		{ 341, 103, 34 },
		{ 341, 104, 34 },
		{ 341, 105, 34 },
		{ 341, 106, 34 },
		{ 341, 107, 34 },
		{ 341, 108, 34 },
		{ 341, 109, 34 },
		{ 341, 110, 34 },
		{ 341, 111, 34 },
		{ 341, 112, 34 },
		{ 341, 113, 34 },
		{ 341, 114, 34 },
		{ 341, 115, 34 },
		{ 341, 116, 34 },
		{ 341, 117, 34 },
		{ 341, 118, 34 },
		{ 341, 119, 34 },
		{ 341, 120, 34 },
		{ 341, 121, 34 },
		{ 341, 122, 34 },
		{ 342, 31, 34 },
		{ 342, 32, 34 },
		{ 342, 33, 34 },
		{ 342, 34, 34 },
		{ 342, 35, 34 },
		{ 342, 36, 34 },
		{ 342, 37, 34 },
		{ 342, 38, 34 },
		{ 342, 39, 34 },
		{ 342, 40, 34 },
		{ 342, 41, 34 },
		{ 342, 42, 34 },
		{ 342, 43, 34 },
		{ 342, 44, 34 },
		{ 342, 45, 34 },
		{ 342, 46, 34 },
		{ 342, 47, 34 },
		{ 342, 48, 34 },
		{ 342, 49, 34 },
		{ 342, 50, 34 },
		{ 342, 51, 34 },
		{ 342, 52, 34 },
		{ 342, 53, 34 },
		{ 342, 54, 34 },
		{ 342, 55, 34 },
		{ 342, 56, 34 },
		{ 342, 57, 34 },
		{ 342, 58, 34 },
		{ 342, 59, 34 },
		{ 342, 60, 34 },
		{ 342, 61, 34 },
		{ 342, 62, 34 },
		{ 342, 63, 34 },
		{ 342, 64, 34 },
		{ 342, 65, 34 },
		{ 342, 66, 34 },
		{ 342, 67, 34 },
		{ 342, 68, 34 },
		{ 342, 69, 34 },
		{ 342, 70, 34 },
		{ 342, 71, 34 },
		{ 342, 72, 34 },
		{ 342, 73, 34 },
		{ 342, 74, 34 },
		{ 342, 75, 34 },
		{ 342, 76, 34 },
		{ 342, 77, 34 },
		{ 342, 78, 34 },
		{ 342, 79, 34 },
		{ 342, 80, 34 },
		{ 342, 81, 34 },
		{ 342, 82, 34 },
		{ 342, 83, 34 },
		{ 342, 84, 34 },
		{ 342, 85, 34 },
		{ 342, 86, 34 },
		{ 342, 87, 34 },
		{ 342, 88, 34 },
		{ 342, 89, 34 },
		{ 342, 90, 34 },
		{ 342, 91, 34 },
		{ 342, 92, 34 },
		{ 342, 93, 34 },
		{ 342, 94, 34 },
		{ 342, 95, 34 },
		{ 342, 96, 34 },
		{ 342, 97, 34 },
		{ 342, 98, 34 },
		{ 342, 99, 34 },
		{ 342, 100, 34 },
		{ 342, 101, 34 },
		{ 342, 102, 34 },
		{ 342, 103, 34 },
		{ 342, 104, 34 },
		{ 342, 105, 34 },
		{ 342, 106, 34 },
		{ 342, 107, 34 },
		{ 342, 108, 34 },
		{ 342, 109, 34 },
		{ 342, 110, 34 },
		{ 342, 111, 34 },
		{ 342, 112, 34 },
		{ 342, 113, 34 },
		{ 342, 114, 34 },
		{ 342, 115, 34 },
		{ 342, 116, 34 },
		{ 342, 117, 34 },
		{ 342, 118, 34 },
		{ 342, 119, 34 },
		{ 342, 120, 34 },
		{ 343, 28, 34 },
		{ 343, 29, 34 },
		{ 343, 30, 34 },
		{ 343, 31, 34 },
		{ 343, 32, 34 },
		{ 343, 33, 34 },
		{ 343, 34, 34 },
		{ 343, 35, 34 },
		{ 343, 36, 34 },
		{ 343, 37, 34 },
		{ 343, 38, 34 },
		{ 343, 39, 34 },
		{ 343, 40, 34 },
		{ 343, 41, 34 },
		{ 343, 42, 34 },
		{ 343, 43, 34 },
		{ 343, 44, 34 },
		{ 343, 45, 34 },
		{ 343, 46, 34 },
		{ 343, 47, 34 },
		{ 343, 48, 34 },
		{ 343, 49, 34 },
		{ 343, 50, 34 },
		{ 343, 51, 34 },
		{ 343, 52, 34 },
		{ 343, 53, 34 },
		{ 343, 54, 34 },
		{ 343, 55, 34 },
		{ 343, 56, 34 },
		{ 343, 57, 34 },
		{ 343, 58, 34 },
		{ 343, 59, 34 },
		{ 343, 60, 34 },
		{ 343, 61, 34 },
		{ 343, 62, 34 },
		{ 343, 63, 34 },
		{ 343, 64, 34 },
		{ 343, 65, 34 },
		{ 343, 66, 34 },
		{ 343, 67, 34 },
		{ 343, 68, 34 },
		{ 343, 69, 34 },
		{ 343, 70, 34 },
		{ 343, 71, 34 },
		{ 343, 72, 34 },
		{ 343, 73, 34 },
		{ 343, 74, 34 },
		{ 343, 75, 34 },
		{ 343, 76, 34 },
		{ 343, 77, 34 },
		{ 343, 78, 34 },
		{ 343, 79, 34 },
		{ 343, 80, 34 },
		{ 343, 81, 34 },
		{ 343, 82, 34 },
		{ 343, 83, 34 },
		{ 343, 84, 34 },
		{ 343, 85, 34 },
		{ 343, 86, 34 },
		{ 343, 87, 34 },
		{ 343, 88, 34 },
		{ 343, 89, 34 },
		{ 343, 90, 34 },
		{ 343, 91, 34 },
		{ 343, 92, 34 },
		{ 343, 93, 34 },
		{ 343, 94, 34 },
		{ 343, 95, 34 },
		{ 343, 96, 34 },
		{ 343, 97, 34 },
		{ 343, 98, 34 },
		{ 343, 99, 34 },
		{ 343, 100, 34 },
		{ 343, 101, 34 },
		{ 343, 102, 34 },
		{ 343, 103, 34 },
		{ 343, 104, 34 },
		{ 343, 105, 34 },
		{ 343, 106, 34 },
		{ 343, 107, 34 },
		{ 343, 108, 34 },
		{ 343, 109, 34 },
		{ 343, 110, 34 },
		{ 343, 111, 34 },
		{ 343, 112, 34 },
		{ 343, 113, 34 },
		{ 343, 114, 34 },
		{ 343, 115, 34 },
		{ 343, 116, 34 },
		{ 343, 117, 34 },
		{ 343, 118, 34 },
		{ 343, 119, 34 },
		{ 344, 26, 34 },
		{ 344, 27, 34 },
		{ 344, 28, 34 },
		{ 344, 29, 34 },
		{ 344, 30, 34 },
		{ 344, 31, 34 },
		{ 344, 32, 34 },
		{ 344, 33, 34 },
		{ 344, 34, 34 },
		{ 344, 35, 34 },
		{ 344, 36, 34 },
		{ 344, 37, 34 },
		{ 344, 38, 34 },
		{ 344, 39, 34 },
		{ 344, 40, 34 },
		{ 344, 41, 34 },
		{ 344, 42, 34 },
		{ 344, 43, 34 },
		{ 344, 44, 34 },
		{ 344, 45, 34 },
		{ 344, 46, 34 },
		{ 344, 47, 34 },
		{ 344, 48, 34 },
		{ 344, 49, 34 },
		{ 344, 50, 34 },
		{ 344, 51, 34 },
		{ 344, 52, 34 },
		{ 344, 53, 34 },
		{ 344, 54, 34 },
		{ 344, 55, 34 },
		{ 344, 56, 34 },
		{ 344, 57, 34 },
		{ 344, 58, 34 },
		{ 344, 59, 34 },
		{ 344, 60, 34 },
		{ 344, 61, 34 },
		{ 344, 62, 34 },
		{ 344, 63, 34 },
		{ 344, 64, 34 },
		{ 344, 65, 34 },
		{ 344, 66, 34 },
		{ 344, 67, 34 },
		{ 344, 68, 34 },
		{ 344, 69, 34 },
		{ 344, 70, 34 },
		{ 344, 71, 34 },
		{ 344, 72, 34 },
		{ 344, 73, 34 },
		{ 344, 74, 34 },
		{ 344, 75, 34 },
		{ 344, 76, 34 },
		{ 344, 77, 34 },
		{ 344, 78, 34 },
		{ 344, 79, 34 },
		{ 344, 80, 34 },
		{ 344, 81, 34 },
		{ 344, 82, 34 },
		{ 344, 83, 34 },
		{ 344, 84, 34 },
		{ 344, 85, 34 },
		{ 344, 86, 34 },
		{ 344, 87, 34 },
		{ 344, 88, 34 },
		{ 344, 89, 34 },
		{ 344, 90, 34 },
		{ 344, 91, 34 },
		{ 344, 92, 34 },
		{ 344, 93, 34 },
		{ 344, 94, 34 },
		{ 344, 95, 34 },
		{ 344, 96, 34 },
		{ 344, 97, 34 },
		{ 344, 98, 34 },
		{ 344, 99, 34 },
		{ 344, 100, 34 },
		{ 344, 101, 34 },
		{ 344, 102, 34 },
		{ 344, 103, 34 },
		{ 344, 104, 34 },
		{ 344, 105, 34 },
		{ 344, 106, 34 },
		{ 344, 107, 34 },
		{ 344, 108, 34 },
		{ 344, 109, 34 },
		{ 344, 110, 34 },
		{ 344, 111, 34 },
		{ 344, 112, 34 },
		{ 344, 113, 34 },
		{ 344, 114, 34 },
		{ 344, 115, 34 },
		{ 344, 116, 34 },
		{ 344, 117, 34 },
		{ 345, 23, 34 },
		{ 345, 24, 34 },
		{ 345, 25, 34 },
		{ 345, 26, 34 },
		{ 345, 27, 34 },
		{ 345, 28, 34 },
		{ 345, 29, 34 },
		{ 345, 30, 34 },
		{ 345, 31, 34 },
		{ 345, 32, 34 },
		{ 345, 33, 34 },
		{ 345, 34, 34 },
		{ 345, 35, 34 },
		{ 345, 36, 34 },
		{ 345, 37, 34 },
		{ 345, 38, 34 },
		{ 345, 39, 34 },
		{ 345, 40, 34 },
		{ 345, 41, 34 },
		{ 345, 42, 34 },
		{ 345, 43, 34 },
		{ 345, 44, 34 },
		{ 345, 45, 34 },
		{ 345, 46, 34 },
		{ 345, 47, 34 },
		{ 345, 48, 34 },
		{ 345, 49, 34 },
		{ 345, 50, 34 },
		{ 345, 51, 34 },
		{ 345, 52, 34 },
		{ 345, 53, 34 },
		{ 345, 54, 34 },
		{ 345, 55, 34 },
		{ 345, 56, 34 },
		{ 345, 57, 34 },
		{ 345, 58, 34 },
		{ 345, 59, 34 },
		{ 345, 60, 34 },
		{ 345, 61, 34 },
		{ 345, 62, 34 },
		{ 345, 63, 34 },
		{ 345, 64, 34 },
		{ 345, 65, 34 },
		{ 345, 66, 34 },
		{ 345, 67, 34 },
		{ 345, 68, 34 },
		{ 345, 69, 34 },
		{ 345, 70, 34 },
		{ 345, 71, 34 },
		{ 345, 72, 34 },
		{ 345, 73, 34 },
		{ 345, 74, 34 },
		{ 345, 75, 34 },
		{ 345, 76, 34 },
		{ 345, 77, 34 },
		{ 345, 78, 34 },
		{ 345, 79, 34 },
		{ 345, 80, 34 },
		{ 345, 81, 34 },
		{ 345, 82, 34 },
		{ 345, 83, 34 },
		{ 345, 84, 34 },
		{ 345, 85, 34 },
		{ 345, 86, 34 },
		{ 345, 87, 34 },
		{ 345, 88, 34 },
		{ 345, 89, 34 },
		{ 345, 90, 34 },
		{ 345, 91, 34 },
		{ 345, 92, 34 },
		{ 345, 93, 34 },
		{ 345, 94, 34 },
		{ 345, 95, 34 },
		{ 345, 96, 34 },
		{ 345, 97, 34 },
		{ 345, 98, 34 },
		{ 345, 99, 34 },
		{ 345, 100, 34 },
		{ 345, 101, 34 },
		{ 345, 102, 34 },
		{ 345, 103, 34 },
		{ 345, 104, 34 },
		{ 345, 105, 34 },
		{ 345, 106, 34 },
		{ 345, 107, 34 },
		{ 345, 108, 34 },
		{ 345, 109, 34 },
		{ 345, 110, 34 },
		{ 345, 111, 34 },
		{ 345, 112, 34 },
		{ 345, 113, 34 },
		{ 345, 114, 34 },
		{ 345, 115, 34 },
		{ 345, 116, 34 },
		{ 346, 20, 34 },
		{ 346, 21, 34 },
		{ 346, 22, 34 },
		{ 346, 23, 34 },
		{ 346, 24, 34 },
		{ 346, 25, 34 },
		{ 346, 26, 34 },
		{ 346, 27, 34 },
		{ 346, 28, 34 },
		{ 346, 29, 34 },
		{ 346, 30, 34 },
		{ 346, 31, 34 },
		{ 346, 32, 34 },
		{ 346, 33, 34 },
		{ 346, 34, 34 },
		{ 346, 35, 34 },
		{ 346, 36, 34 },
		{ 346, 37, 34 },
		{ 346, 38, 34 },
		{ 346, 39, 34 },
		{ 346, 40, 34 },
		{ 346, 41, 34 },
		{ 346, 42, 34 },
		{ 346, 43, 34 },
		{ 346, 44, 34 },
		{ 346, 45, 34 },
		{ 346, 46, 34 },
		{ 346, 47, 34 },
		{ 346, 48, 34 },
		{ 346, 49, 34 },
		{ 346, 50, 34 },
		{ 346, 51, 34 },
		{ 346, 52, 34 },
		{ 346, 53, 34 },
		{ 346, 54, 34 },
		{ 346, 55, 34 },
		{ 346, 56, 34 },
		{ 346, 57, 34 },
		{ 346, 58, 34 },
		{ 346, 59, 34 },
		{ 346, 60, 34 },
		{ 346, 61, 34 },
		{ 346, 62, 34 },
		{ 346, 63, 34 },
		{ 346, 64, 34 },
		{ 346, 65, 34 },
		{ 346, 66, 34 },
		{ 346, 67, 34 },
		{ 346, 68, 34 },
		{ 346, 69, 34 },
		{ 346, 70, 34 },
		{ 346, 71, 34 },
		{ 346, 72, 34 },
		{ 346, 73, 34 },
		{ 346, 74, 34 },
		{ 346, 75, 34 },
		{ 346, 76, 34 },
		{ 346, 77, 34 },
		{ 346, 78, 34 },
		{ 346, 79, 34 },
		{ 346, 80, 34 },
		{ 346, 81, 34 },
		{ 346, 82, 34 },
		{ 346, 83, 34 },
		{ 346, 84, 34 },
		{ 346, 85, 34 },
		{ 346, 86, 34 },
		{ 346, 87, 34 },
		{ 346, 88, 34 },
		{ 346, 89, 34 },
		{ 346, 90, 34 },
		{ 346, 91, 34 },
		{ 346, 92, 34 },
		{ 346, 93, 34 },
		{ 346, 94, 34 },
		{ 346, 95, 34 },
		{ 346, 96, 34 },
		{ 346, 97, 34 },
		{ 346, 98, 34 },
		{ 346, 99, 34 },
		{ 346, 100, 34 },
		{ 346, 101, 34 },
		{ 346, 102, 34 },
		{ 346, 103, 34 },
		{ 346, 104, 34 },
		{ 346, 105, 34 },
		{ 346, 106, 34 },
		{ 346, 107, 34 },
		{ 346, 108, 34 },
		{ 346, 109, 34 },
		{ 346, 110, 34 },
		{ 346, 111, 34 },
		{ 346, 112, 34 },
		{ 346, 113, 34 },
		{ 346, 114, 34 },
		{ 347, 18, 34 },
		{ 347, 19, 34 },
		{ 347, 20, 34 },
		{ 347, 21, 34 },
		{ 347, 22, 34 },
		{ 347, 23, 34 },
		{ 347, 24, 34 },
		{ 347, 25, 34 },
		{ 347, 26, 34 },
		{ 347, 27, 34 },
		{ 347, 28, 34 },
		{ 347, 29, 34 },
		{ 347, 30, 34 },
		{ 347, 31, 34 },
		{ 347, 32, 34 },
		{ 347, 33, 34 },
		{ 347, 34, 34 },
		{ 347, 35, 34 },
		{ 347, 36, 34 },
		{ 347, 37, 34 },
		{ 347, 38, 34 },
		{ 347, 39, 34 },
		{ 347, 40, 34 },
		{ 347, 41, 34 },
		{ 347, 42, 34 },
		{ 347, 43, 34 },
		{ 347, 44, 34 },
		{ 347, 45, 34 },
		{ 347, 46, 34 },
		{ 347, 47, 34 },
		{ 347, 48, 34 },
		{ 347, 49, 34 },
		{ 347, 50, 34 },
		{ 347, 51, 34 },
		{ 347, 52, 34 },
		{ 347, 53, 34 },
		{ 347, 54, 34 },
		{ 347, 55, 34 },
		{ 347, 56, 34 },
		{ 347, 57, 34 },
		{ 347, 58, 34 },
		{ 347, 59, 34 },
		{ 347, 60, 34 },
		{ 347, 61, 34 },
		{ 347, 62, 34 },
		{ 347, 63, 34 },
		{ 347, 64, 34 },
		{ 347, 65, 34 },
		{ 347, 66, 34 },
		{ 347, 67, 34 },
		{ 347, 68, 34 },
		{ 347, 69, 34 },
		{ 347, 70, 34 },
		{ 347, 71, 34 },
		{ 347, 72, 34 },
		{ 347, 73, 34 },
		{ 347, 74, 34 },
		{ 347, 75, 34 },
		{ 347, 76, 34 },
		{ 347, 77, 34 },
		{ 347, 78, 34 },
		{ 347, 79, 34 },
		{ 347, 80, 34 },
		{ 347, 81, 34 },
		{ 347, 82, 34 },
		{ 347, 83, 34 },
		{ 347, 84, 34 },
		{ 347, 85, 34 },
		{ 347, 86, 34 },
		{ 347, 87, 34 },
		{ 347, 88, 34 },
		{ 347, 89, 34 },
		{ 347, 90, 34 },
		{ 347, 91, 34 },
		{ 347, 92, 34 },
		{ 347, 93, 34 },
		{ 347, 94, 34 },
		{ 347, 95, 34 },
		{ 347, 96, 34 },
		{ 347, 97, 34 },
		{ 347, 98, 34 },
		{ 347, 99, 34 },
		{ 347, 100, 34 },
		{ 347, 101, 34 },
		{ 347, 102, 34 },
		{ 347, 103, 34 },
		{ 347, 104, 34 },
		{ 347, 105, 34 },
		{ 347, 106, 34 },
		{ 347, 107, 34 },
		{ 347, 108, 34 },
		{ 347, 109, 34 },
		{ 347, 110, 34 },
		{ 347, 111, 34 },
		{ 347, 112, 34 },
		{ 348, 15, 34 },
		{ 348, 16, 34 },
		{ 348, 17, 34 },
		{ 348, 18, 34 },
		{ 348, 19, 34 },
		{ 348, 20, 34 },
		{ 348, 21, 34 },
		{ 348, 22, 34 },
		{ 348, 23, 34 },
		{ 348, 24, 34 },
		{ 348, 25, 34 },
		{ 348, 26, 34 },
		{ 348, 27, 34 },
		{ 348, 28, 34 },
		{ 348, 29, 34 },
		{ 348, 30, 34 },
		{ 348, 31, 34 },
		{ 348, 32, 34 },
		{ 348, 33, 34 },
		{ 348, 34, 34 },
		{ 348, 35, 34 },
		{ 348, 36, 34 },
		{ 348, 37, 34 },
		{ 348, 38, 34 },
		{ 348, 39, 34 },
		{ 348, 40, 34 },
		{ 348, 41, 34 },
		{ 348, 42, 34 },
		{ 348, 43, 34 },
		{ 348, 44, 34 },
		{ 348, 45, 34 },
		{ 348, 46, 34 },
		{ 348, 47, 34 },
		{ 348, 48, 34 },
		{ 348, 49, 34 },
		{ 348, 50, 34 },
		{ 348, 51, 34 },
		{ 348, 52, 34 },
		{ 348, 53, 34 },
		{ 348, 54, 34 },
		{ 348, 55, 34 },
		{ 348, 56, 34 },
		{ 348, 57, 34 },
		{ 348, 58, 34 },
		{ 348, 59, 34 },
		{ 348, 60, 34 },
		{ 348, 61, 34 },
		{ 348, 62, 34 },
		{ 348, 63, 34 },
		{ 348, 64, 34 },
		{ 348, 65, 34 },
		{ 348, 66, 34 },
		{ 348, 67, 34 },
		{ 348, 68, 34 },
		{ 348, 69, 34 },
		{ 348, 70, 34 },
		{ 348, 71, 34 },
		{ 348, 72, 34 },
		{ 348, 73, 34 },
		{ 348, 74, 34 },
		{ 348, 75, 34 },
		{ 348, 76, 34 },
		{ 348, 77, 34 },
		{ 348, 78, 34 },
		{ 348, 79, 34 },
		{ 348, 80, 34 },
		{ 348, 81, 34 },
		{ 348, 82, 34 },
		{ 348, 83, 34 },
		{ 348, 84, 34 },
		{ 348, 85, 34 },
		{ 348, 86, 34 },
		{ 348, 87, 34 },
		{ 348, 88, 34 },
		{ 348, 89, 34 },
		{ 348, 90, 34 },
		{ 348, 91, 34 },
		{ 348, 92, 34 },
		{ 348, 93, 34 },
		{ 348, 94, 34 },
		{ 348, 95, 34 },
		{ 348, 96, 34 },
		{ 348, 97, 34 },
		{ 348, 98, 34 },
		{ 348, 99, 34 },
		{ 348, 100, 34 },
		{ 348, 101, 34 },
		{ 348, 102, 34 },
		{ 348, 103, 34 },
		{ 348, 104, 34 },
		{ 348, 105, 34 },
		{ 348, 106, 34 },
		{ 348, 107, 34 },
		{ 348, 108, 34 },
		{ 348, 109, 34 },
		{ 348, 110, 34 },
		{ 348, 111, 34 },
		{ 349, 12, 34 },
		{ 349, 13, 34 },
		{ 349, 14, 34 },
		{ 349, 15, 34 },
		{ 349, 16, 34 },
		{ 349, 17, 34 },
		{ 349, 18, 34 },
		{ 349, 19, 34 },
		{ 349, 20, 34 },
		{ 349, 21, 34 },
		{ 349, 22, 34 },
		{ 349, 23, 34 },
		{ 349, 24, 34 },
		{ 349, 25, 34 },
		{ 349, 26, 34 },
		{ 349, 27, 34 },
		{ 349, 28, 34 },
		{ 349, 29, 34 },
		{ 349, 30, 34 },
		{ 349, 31, 34 },
		{ 349, 32, 34 },
		{ 349, 33, 34 },
		{ 349, 34, 34 },
		{ 349, 35, 34 },
		{ 349, 36, 34 },
		{ 349, 37, 34 },
		{ 349, 38, 34 },
		{ 349, 39, 34 },
		{ 349, 40, 34 },
		{ 349, 41, 34 },
		{ 349, 42, 34 },
		{ 349, 43, 34 },
		{ 349, 44, 34 },
		{ 349, 45, 34 },
		{ 349, 46, 34 },
		{ 349, 47, 34 },
		{ 349, 48, 34 },
		{ 349, 49, 34 },
		{ 349, 50, 34 },
		{ 349, 51, 34 },
		{ 349, 52, 34 },
		{ 349, 53, 34 },
		{ 349, 54, 34 },
		{ 349, 55, 34 },
		{ 349, 56, 34 },
		{ 349, 57, 34 },
		{ 349, 58, 34 },
		{ 349, 59, 34 },
		{ 349, 60, 34 },
		{ 349, 61, 34 },
		{ 349, 62, 34 },
		{ 349, 63, 34 },
		{ 349, 64, 34 },
		{ 349, 65, 34 },
		{ 349, 66, 34 },
		{ 349, 67, 34 },
		{ 349, 68, 34 },
		{ 349, 69, 34 },
		{ 349, 70, 34 },
		{ 349, 71, 34 },
		{ 349, 72, 34 },
		{ 349, 73, 34 },
		{ 349, 74, 34 },
		{ 349, 75, 34 },
		{ 349, 76, 34 },
		{ 349, 77, 34 },
		{ 349, 78, 34 },
		{ 349, 79, 34 },
		{ 349, 80, 34 },
		{ 349, 81, 34 },
		{ 349, 82, 34 },
		{ 349, 83, 34 },
		{ 349, 84, 34 },
		{ 349, 85, 34 },
		{ 349, 86, 34 },
		{ 349, 87, 34 },
		{ 349, 88, 34 },
		{ 349, 89, 34 },
		{ 349, 90, 34 },
		{ 349, 91, 34 },
		{ 349, 92, 34 },
		{ 349, 93, 34 },
		{ 349, 94, 34 },
		{ 349, 95, 34 },
		{ 349, 96, 34 },
		{ 349, 97, 34 },
		{ 349, 98, 34 },
		{ 349, 99, 34 },
		{ 349, 100, 34 },
		{ 349, 101, 34 },
		{ 349, 102, 34 },
		{ 349, 103, 34 },
		{ 349, 104, 34 },
		{ 349, 105, 34 },
		{ 349, 106, 34 },
		{ 349, 107, 34 },
		{ 349, 108, 34 },
		{ 349, 109, 34 },
		{ 350, 10, 34 },
		{ 350, 11, 34 },
		{ 350, 12, 34 },
		{ 350, 13, 34 },
		{ 350, 14, 34 },
		{ 350, 15, 34 },
		{ 350, 16, 34 },
		{ 350, 17, 34 },
		{ 350, 18, 34 },
		{ 350, 19, 34 },
		{ 350, 20, 34 },
		{ 350, 21, 34 },
		{ 350, 22, 34 },
		{ 350, 23, 34 },
		{ 350, 24, 34 },
		{ 350, 25, 34 },
		{ 350, 26, 34 },
		{ 350, 27, 34 },
		{ 350, 28, 34 },
		{ 350, 29, 34 },
		{ 350, 30, 34 },
		{ 350, 31, 34 },
		{ 350, 32, 34 },
		{ 350, 33, 34 },
		{ 350, 34, 34 },
		{ 350, 35, 34 },
		{ 350, 36, 34 },
		{ 350, 37, 34 },
		{ 350, 38, 34 },
		{ 350, 39, 34 },
		{ 350, 40, 34 },
		{ 350, 41, 34 },
		{ 350, 42, 34 },
		{ 350, 43, 34 },
		{ 350, 44, 34 },
		{ 350, 45, 34 },
		{ 350, 46, 34 },
		{ 350, 47, 34 },
		{ 350, 48, 34 },
		{ 350, 49, 34 },
		{ 350, 50, 34 },
		{ 350, 51, 34 },
		{ 350, 52, 34 },
		{ 350, 53, 34 },
		{ 350, 54, 34 },
		{ 350, 55, 34 },
		{ 350, 56, 34 },
		{ 350, 57, 34 },
		{ 350, 58, 34 },
		{ 350, 59, 34 },
		{ 350, 60, 34 },
		{ 350, 61, 34 },
		{ 350, 62, 34 },
		{ 350, 63, 34 },
		{ 350, 64, 34 },
		{ 350, 65, 34 },
		{ 350, 66, 34 },
		{ 350, 67, 34 },
		{ 350, 68, 34 },
		{ 350, 69, 34 },
		{ 350, 70, 34 },
		{ 350, 71, 34 },
		{ 350, 72, 34 },
		{ 350, 73, 34 },
		{ 350, 74, 34 },
		{ 350, 75, 34 },
		{ 350, 76, 34 },
		{ 350, 77, 34 },
		{ 350, 78, 34 },
		{ 350, 79, 34 },
		{ 350, 80, 34 },
		{ 350, 81, 34 },
		{ 350, 82, 34 },
		{ 350, 83, 34 },
		{ 350, 84, 34 },
		{ 350, 85, 34 },
		{ 350, 86, 34 },
		{ 350, 87, 34 },
		{ 350, 88, 34 },
		{ 350, 89, 34 },
		{ 350, 90, 34 },
		{ 350, 91, 34 },
		{ 350, 92, 34 },
		{ 350, 93, 34 },
		{ 350, 94, 34 },
		{ 350, 95, 34 },
		{ 350, 96, 34 },
		{ 350, 97, 34 },
		{ 350, 98, 34 },
		{ 350, 99, 34 },
		{ 350, 100, 34 },
		{ 350, 101, 34 },
		{ 350, 102, 34 },
		{ 350, 103, 34 },
		{ 350, 104, 34 },
		{ 350, 105, 34 },
		{ 350, 106, 34 },
		{ 350, 107, 34 },
		{ 350, 108, 34 },
		{ 351, 7, 34 },
		{ 351, 8, 34 },
		{ 351, 9, 34 },
		{ 351, 10, 34 },
		{ 351, 11, 34 },
		{ 351, 12, 34 },
		{ 351, 13, 34 },
		{ 351, 14, 34 },
		{ 351, 15, 34 },
		{ 351, 16, 34 },
		{ 351, 17, 34 },
		{ 351, 18, 34 },
		{ 351, 19, 34 },
		{ 351, 20, 34 },
		{ 351, 21, 34 },
		{ 351, 22, 34 },
		{ 351, 23, 34 },
		{ 351, 24, 34 },
		{ 351, 25, 34 },
		{ 351, 26, 34 },
		{ 351, 27, 34 },
		{ 351, 28, 34 },
		{ 351, 29, 34 },
		{ 351, 30, 34 },
		{ 351, 31, 34 },
		{ 351, 32, 34 },
		{ 351, 33, 34 },
		{ 351, 34, 34 },
		{ 351, 35, 34 },
		{ 351, 36, 34 },
		{ 351, 37, 34 },
		{ 351, 38, 34 },
		{ 351, 39, 34 },
		{ 351, 40, 34 },
		{ 351, 41, 34 },
		{ 351, 42, 34 },
		{ 351, 43, 34 },
		{ 351, 44, 34 },
		{ 351, 45, 34 },
		{ 351, 46, 34 },
		{ 351, 47, 34 },
		{ 351, 48, 34 },
		{ 351, 49, 34 },
		{ 351, 50, 34 },
		{ 351, 51, 34 },
		{ 351, 52, 34 },
		{ 351, 53, 34 },
		{ 351, 54, 34 },
		{ 351, 55, 34 },
		{ 351, 56, 34 },
		{ 351, 57, 34 },
		{ 351, 58, 34 },
		{ 351, 59, 34 },
		{ 351, 60, 34 },
		{ 351, 61, 34 },
		{ 351, 62, 34 },
		{ 351, 63, 34 },
		{ 351, 64, 34 },
		{ 351, 65, 34 },
		{ 351, 66, 34 },
		{ 351, 67, 34 },
		{ 351, 68, 34 },
		{ 351, 69, 34 },
		{ 351, 70, 34 },
		{ 351, 71, 34 },
		{ 351, 72, 34 },
		{ 351, 73, 34 },
		{ 351, 74, 34 },
		{ 351, 75, 34 },
		{ 351, 76, 34 },
		{ 351, 77, 34 },
		{ 351, 78, 34 },
		{ 351, 79, 34 },
		{ 351, 80, 34 },
		{ 351, 81, 34 },
		{ 351, 82, 34 },
		{ 351, 83, 34 },
		{ 351, 84, 34 },
		{ 351, 85, 34 },
		{ 351, 86, 34 },
		{ 351, 87, 34 },
		{ 351, 88, 34 },
		{ 351, 89, 34 },
		{ 351, 90, 34 },
		{ 351, 91, 34 },
		{ 351, 92, 34 },
		{ 351, 93, 34 },
		{ 351, 94, 34 },
		{ 351, 95, 34 },
		{ 351, 96, 34 },
		{ 351, 97, 34 },
		{ 351, 98, 34 },
		{ 351, 99, 34 },
		{ 351, 100, 34 },
		{ 351, 101, 34 },
		{ 351, 102, 34 },
		{ 351, 103, 34 },
		{ 351, 104, 34 },
		{ 351, 105, 34 },
		{ 351, 106, 34 },
		{ 352, 4, 34 },
		{ 352, 5, 34 },
		{ 352, 6, 34 },
		{ 352, 7, 34 },
		{ 352, 8, 34 },
		{ 352, 9, 34 },
		{ 352, 10, 34 },
		{ 352, 11, 34 },
		{ 352, 12, 34 },
		{ 352, 13, 34 },
		{ 352, 14, 34 },
		{ 352, 15, 34 },
		{ 352, 16, 34 },
		{ 352, 17, 34 },
		{ 352, 18, 34 },
		{ 352, 19, 34 },
		{ 352, 20, 34 },
		{ 352, 21, 34 },
		{ 352, 22, 34 },
		{ 352, 23, 34 },
		{ 352, 24, 34 },
		{ 352, 25, 34 },
		{ 352, 26, 34 },
		{ 352, 27, 34 },
		{ 352, 28, 34 },
		{ 352, 29, 34 },
		{ 352, 30, 34 },
		{ 352, 31, 34 },
		{ 352, 32, 34 },
		{ 352, 33, 34 },
		{ 352, 34, 34 },
		{ 352, 35, 34 },
		{ 352, 36, 34 },
		{ 352, 37, 34 },
		{ 352, 38, 34 },
		{ 352, 39, 34 },
		{ 352, 40, 34 },
		{ 352, 41, 34 },
		{ 352, 42, 34 },
		{ 352, 43, 34 },
		{ 352, 44, 34 },
		{ 352, 45, 34 },
		{ 352, 46, 34 },
		{ 352, 47, 34 },
		{ 352, 48, 34 },
		{ 352, 49, 34 },
		{ 352, 50, 34 },
		{ 352, 51, 34 },
		{ 352, 52, 34 },
		{ 352, 53, 34 },
		{ 352, 54, 34 },
		{ 352, 55, 34 },
		{ 352, 56, 34 },
		{ 352, 57, 34 },
		{ 352, 58, 34 },
		{ 352, 59, 34 },
		{ 352, 60, 34 },
		{ 352, 61, 34 },
		{ 352, 62, 34 },
		{ 352, 63, 34 },
		{ 352, 64, 34 },
		{ 352, 65, 34 },
		{ 352, 66, 34 },
		{ 352, 67, 34 },
		{ 352, 68, 34 },
		{ 352, 69, 34 },
		{ 352, 70, 34 },
		{ 352, 71, 34 },
		{ 352, 72, 34 },
		{ 352, 73, 34 },
		{ 352, 74, 34 },
		{ 352, 75, 34 },
		{ 352, 76, 34 },
		{ 352, 77, 34 },
		{ 352, 78, 34 },
		{ 352, 79, 34 },
		{ 352, 80, 34 },
		{ 352, 81, 34 },
		{ 352, 82, 34 },
		{ 352, 83, 34 },
		{ 352, 84, 34 },
		{ 352, 85, 34 },
		{ 352, 86, 34 },
		{ 352, 87, 34 },
		{ 352, 88, 34 },
		{ 352, 89, 34 },
		{ 352, 90, 34 },
		{ 352, 91, 34 },
		{ 352, 92, 34 },
		{ 352, 93, 34 },
		{ 352, 94, 34 },
		{ 352, 95, 34 },
		{ 352, 96, 34 },
		{ 352, 97, 34 },
		{ 352, 98, 34 },
		{ 352, 99, 34 },
		{ 352, 100, 34 },
		{ 352, 101, 34 },
		{ 352, 102, 34 },
		{ 352, 103, 34 },
		{ 352, 104, 34 },
		{ 353, 2, 34 },
		{ 353, 3, 34 },
		{ 353, 4, 34 },
		{ 353, 5, 34 },
		{ 353, 6, 34 },
		{ 353, 7, 34 },
		{ 353, 8, 34 },
		{ 353, 9, 34 },
		{ 353, 10, 34 },
		{ 353, 11, 34 },
		{ 353, 12, 34 },
		{ 353, 13, 34 },
		{ 353, 14, 34 },
		{ 353, 15, 34 },
		{ 353, 16, 34 },
		{ 353, 17, 34 },
		{ 353, 18, 34 },
		{ 353, 19, 34 },
		{ 353, 20, 34 },
		{ 353, 21, 34 },
		{ 353, 22, 34 },
		{ 353, 23, 34 },
		{ 353, 24, 34 },
		{ 353, 25, 34 },
		{ 353, 26, 34 },
		{ 353, 27, 34 },
		{ 353, 28, 34 },
		{ 353, 29, 34 },
		{ 353, 30, 34 },
		{ 353, 31, 34 },
		{ 353, 32, 34 },
		{ 353, 33, 34 },
		{ 353, 34, 34 },
		{ 353, 35, 34 },
		{ 353, 36, 34 },
		{ 353, 37, 34 },
		{ 353, 38, 34 },
		{ 353, 39, 34 },
		{ 353, 40, 34 },
		{ 353, 41, 34 },
		{ 353, 42, 34 },
		{ 353, 43, 34 },
		{ 353, 44, 34 },
		{ 353, 45, 34 },
		{ 353, 46, 34 },
		{ 353, 47, 34 },
		{ 353, 48, 34 },
		{ 353, 49, 34 },
		{ 353, 50, 34 },
		{ 353, 51, 34 },
		{ 353, 52, 34 },
		{ 353, 53, 34 },
		{ 353, 54, 34 },
		{ 353, 55, 34 },
		{ 353, 56, 34 },
		{ 353, 57, 34 },
		{ 353, 58, 34 },
		{ 353, 59, 34 },
		{ 353, 60, 34 },
		{ 353, 61, 34 },
		{ 353, 62, 34 },
		{ 353, 63, 34 },
		{ 353, 64, 34 },
		{ 353, 65, 34 },
		{ 353, 66, 34 },
		{ 353, 67, 34 },
		{ 353, 68, 34 },
		{ 353, 69, 34 },
		{ 353, 70, 34 },
		{ 353, 71, 34 },
		{ 353, 72, 34 },
		{ 353, 73, 34 },
		{ 353, 74, 34 },
		{ 353, 75, 34 },
		{ 353, 76, 34 },
		{ 353, 77, 34 },
		{ 353, 78, 34 },
		{ 353, 79, 34 },
		{ 353, 80, 34 },
		{ 353, 81, 34 },
		{ 353, 82, 34 },
		{ 353, 83, 34 },
		{ 353, 84, 34 },
		{ 353, 85, 34 },
		{ 353, 86, 34 },
		{ 353, 87, 34 },
		{ 353, 88, 34 },
		{ 353, 89, 34 },
		{ 353, 90, 34 },
		{ 353, 91, 34 },
		{ 353, 92, 34 },
		{ 353, 93, 34 },
		{ 353, 94, 34 },
		{ 353, 95, 34 },
		{ 353, 96, 34 },
		{ 353, 97, 34 },
		{ 353, 98, 34 },
		{ 353, 99, 34 },
		{ 353, 100, 34 },
		{ 353, 101, 34 },
		{ 353, 102, 34 },
		{ 353, 103, 34 },
		{ 354, 1, 34 },
		{ 354, 2, 34 },
		{ 354, 3, 34 },
		{ 354, 4, 34 },
		{ 354, 5, 34 },
		{ 354, 6, 34 },
		{ 354, 7, 34 },
		{ 354, 8, 34 },
		{ 354, 9, 34 },
		{ 354, 10, 34 },
		{ 354, 11, 34 },
		{ 354, 12, 34 },
		{ 354, 13, 34 },
		{ 354, 14, 34 },
		{ 354, 15, 34 },
		{ 354, 16, 34 },
		{ 354, 17, 34 },
		{ 354, 18, 34 },
		{ 354, 19, 34 },
		{ 354, 20, 34 },
		{ 354, 21, 34 },
		{ 354, 22, 34 },
		{ 354, 23, 34 },
		{ 354, 24, 34 },
		{ 354, 25, 34 },
		{ 354, 26, 34 },
		{ 354, 27, 34 },
		{ 354, 28, 34 },
		{ 354, 29, 34 },
		{ 354, 30, 34 },
		{ 354, 31, 34 },
		{ 354, 32, 34 },
		{ 354, 33, 34 },
		{ 354, 34, 34 },
		{ 354, 35, 34 },
		{ 354, 36, 34 },
		{ 354, 37, 34 },
		{ 354, 38, 34 },
		{ 354, 39, 34 },
		{ 354, 40, 34 },
		{ 354, 41, 34 },
		{ 354, 42, 34 },
		{ 354, 43, 34 },
		{ 354, 44, 34 },
		{ 354, 45, 34 },
		{ 354, 46, 34 },
		{ 354, 47, 34 },
		{ 354, 48, 34 },
		{ 354, 49, 34 },
		{ 354, 50, 34 },
		{ 354, 51, 34 },
		{ 354, 52, 34 },
		{ 354, 53, 34 },
		{ 354, 54, 34 },
		{ 354, 55, 34 },
		{ 354, 56, 34 },
		{ 354, 57, 34 },
		{ 354, 58, 34 },
		{ 354, 59, 34 },
		{ 354, 60, 34 },
		{ 354, 61, 34 },
		{ 354, 62, 34 },
		{ 354, 63, 34 },
		{ 354, 64, 34 },
		{ 354, 65, 34 },
		{ 354, 66, 34 },
		{ 354, 67, 34 },
		{ 354, 68, 34 },
		{ 354, 69, 34 },
		{ 354, 70, 34 },
		{ 354, 71, 34 },
		{ 354, 72, 34 },
		{ 354, 73, 34 },
		{ 354, 74, 34 },
		{ 354, 75, 34 },
		{ 354, 76, 34 },
		{ 354, 77, 34 },
		{ 354, 78, 34 },
		{ 354, 79, 34 },
		{ 354, 80, 34 },
		{ 354, 81, 34 },
		{ 354, 82, 34 },
		{ 354, 83, 34 },
		{ 354, 84, 34 },
		{ 354, 85, 34 },
		{ 354, 86, 34 },
		{ 354, 87, 34 },
		{ 354, 88, 34 },
		{ 354, 89, 34 },
		{ 354, 90, 34 },
		{ 354, 91, 34 },
		{ 354, 92, 34 },
		{ 354, 93, 34 },
		{ 354, 94, 34 },
		{ 354, 95, 34 },
		{ 354, 96, 34 },
		{ 354, 97, 34 },
		{ 354, 98, 34 },
		{ 354, 99, 34 },
		{ 354, 100, 34 },
		{ 354, 101, 34 },
		{ 355, 1, 34 },
		{ 355, 2, 34 },
		{ 355, 3, 34 },
		{ 355, 4, 34 },
		{ 355, 5, 34 },
		{ 355, 6, 34 },
		{ 355, 7, 34 },
		{ 355, 8, 34 },
		{ 355, 9, 34 },
		{ 355, 10, 34 },
		{ 355, 11, 34 },
		{ 355, 12, 34 },
		{ 355, 13, 34 },
		{ 355, 14, 34 },
		{ 355, 15, 34 },
		{ 355, 16, 34 },
		{ 355, 17, 34 },
		{ 355, 18, 34 },
		{ 355, 19, 34 },
		{ 355, 20, 34 },
		{ 355, 21, 34 },
		{ 355, 22, 34 },
		{ 355, 23, 34 },
		{ 355, 24, 34 },
		{ 355, 25, 34 },
		{ 355, 26, 34 },
		{ 355, 27, 34 },
		{ 355, 28, 34 },
		{ 355, 29, 34 },
		{ 355, 30, 34 },
		{ 355, 31, 34 },
		{ 355, 32, 34 },
		{ 355, 33, 34 },
		{ 355, 34, 34 },
		{ 355, 35, 34 },
		{ 355, 36, 34 },
		{ 355, 37, 34 },
		{ 355, 38, 34 },
		{ 355, 39, 34 },
		{ 355, 40, 34 },
		{ 355, 41, 34 },
		{ 355, 42, 34 },
		{ 355, 43, 34 },
		{ 355, 44, 34 },
		{ 355, 45, 34 },
		{ 355, 46, 34 },
		{ 355, 47, 34 },
		{ 355, 48, 34 },
		{ 355, 49, 34 },
		{ 355, 50, 34 },
		{ 355, 51, 34 },
		{ 355, 52, 34 },
		{ 355, 53, 34 },
		{ 355, 54, 34 },
		{ 355, 55, 34 },
		{ 355, 56, 34 },
		{ 355, 57, 34 },
		{ 355, 58, 34 },
		{ 355, 59, 34 },
		{ 355, 60, 34 },
		{ 355, 61, 34 },
		{ 355, 62, 34 },
		{ 355, 63, 34 },
		{ 355, 64, 34 },
		{ 355, 65, 34 },
		{ 355, 66, 34 },
		{ 355, 67, 34 },
		{ 355, 68, 34 },
		{ 355, 69, 34 },
		{ 355, 70, 34 },
		{ 355, 71, 34 },
		{ 355, 72, 34 },
		{ 355, 73, 34 },
		{ 355, 74, 34 },
		{ 355, 75, 34 },
		{ 355, 76, 34 },
		{ 355, 77, 34 },
		{ 355, 78, 34 },
		{ 355, 79, 34 },
		{ 355, 80, 34 },
		{ 355, 81, 34 },
		{ 355, 82, 34 },
		{ 355, 83, 34 },
		{ 355, 84, 34 },
		{ 355, 85, 34 },
		{ 355, 86, 34 },
		{ 355, 87, 34 },
		{ 355, 88, 34 },
		{ 355, 89, 34 },
		{ 355, 90, 34 },
		{ 355, 91, 34 },
		{ 355, 92, 34 },
		{ 355, 93, 34 },
		{ 355, 94, 34 },
		{ 355, 95, 34 },
		{ 355, 96, 34 },
		{ 355, 97, 34 },
		{ 355, 98, 34 },
		{ 355, 99, 34 },
		{ 355, 100, 34 },
		{ 356, 1, 34 },
		{ 356, 2, 34 },
		{ 356, 3, 34 },
		{ 356, 4, 34 },
		{ 356, 5, 34 },
		{ 356, 6, 34 },
		{ 356, 7, 34 },
		{ 356, 8, 34 },
		{ 356, 9, 34 },
		{ 356, 10, 34 },
		{ 356, 11, 34 },
		{ 356, 12, 34 },
		{ 356, 13, 34 },
		{ 356, 14, 34 },
		{ 356, 15, 34 },
		{ 356, 16, 34 },
		{ 356, 17, 34 },
		{ 356, 18, 34 },
		{ 356, 19, 34 },
		{ 356, 20, 34 },
		{ 356, 21, 34 },
		{ 356, 22, 34 },
		{ 356, 23, 34 },
		{ 356, 24, 34 },
		{ 356, 25, 34 },
		{ 356, 26, 34 },
		{ 356, 27, 34 },
		{ 356, 28, 34 },
		{ 356, 29, 34 },
		{ 356, 30, 34 },
		{ 356, 31, 34 },
		{ 356, 32, 34 },
		{ 356, 33, 34 },
		{ 356, 34, 34 },
		{ 356, 35, 34 },
		{ 356, 36, 34 },
		{ 356, 37, 34 },
		{ 356, 38, 34 },
		{ 356, 39, 34 },
		{ 356, 40, 34 },
		{ 356, 41, 34 },
		{ 356, 42, 34 },
		{ 356, 43, 34 },
		{ 356, 44, 34 },
		{ 356, 45, 34 },
		{ 356, 46, 34 },
		{ 356, 47, 34 },
		{ 356, 48, 34 },
		{ 356, 49, 34 },
		{ 356, 50, 34 },
		{ 356, 51, 34 },
		{ 356, 52, 34 },
		{ 356, 53, 34 },
		{ 356, 54, 34 },
		{ 356, 55, 34 },
		{ 356, 56, 34 },
		{ 356, 57, 34 },
		{ 356, 58, 34 },
		{ 356, 59, 34 },
		{ 356, 60, 34 },
		{ 356, 61, 34 },
		{ 356, 62, 34 },
		{ 356, 63, 34 },
		{ 356, 64, 34 },
		{ 356, 65, 34 },
		{ 356, 66, 34 },
		{ 356, 67, 34 },
		{ 356, 68, 34 },
		{ 356, 69, 34 },
		{ 356, 70, 34 },
		{ 356, 71, 34 },
		{ 356, 72, 34 },
		{ 356, 73, 34 },
		{ 356, 74, 34 },
		{ 356, 75, 34 },
		{ 356, 76, 34 },
		{ 356, 77, 34 },
		{ 356, 78, 34 },
		{ 356, 79, 34 },
		{ 356, 80, 34 },
		{ 356, 81, 34 },
		{ 356, 82, 34 },
		{ 356, 83, 34 },
		{ 356, 84, 34 },
		{ 356, 85, 34 },
		{ 356, 86, 34 },
		{ 356, 87, 34 },
		{ 356, 88, 34 },
		{ 356, 89, 34 },
		{ 356, 90, 34 },
		{ 356, 91, 34 },
		{ 356, 92, 34 },
		{ 356, 93, 34 },
		{ 356, 94, 34 },
		{ 356, 95, 34 },
		{ 356, 96, 34 },
		{ 356, 97, 34 },
		{ 356, 98, 34 },
		{ 357, 1, 34 },
		{ 357, 2, 34 },
		{ 357, 3, 34 },
		{ 357, 4, 34 },
		{ 357, 5, 34 },
		{ 357, 6, 34 },
		{ 357, 7, 34 },
		{ 357, 8, 34 },
		{ 357, 9, 34 },
		{ 357, 10, 34 },
		{ 357, 11, 34 },
		{ 357, 12, 34 },
		{ 357, 13, 34 },
		{ 357, 14, 34 },
		{ 357, 15, 34 },
		{ 357, 16, 34 },
		{ 357, 17, 34 },
		{ 357, 18, 34 },
		{ 357, 19, 34 },
		{ 357, 20, 34 },
		{ 357, 21, 34 },
		{ 357, 22, 34 },
		{ 357, 23, 34 },
		{ 357, 24, 34 },
		{ 357, 25, 34 },
		{ 357, 26, 34 },
		{ 357, 27, 34 },
		{ 357, 28, 34 },
		{ 357, 29, 34 },
		{ 357, 30, 34 },
		{ 357, 31, 34 },
		{ 357, 32, 34 },
		{ 357, 33, 34 },
		{ 357, 34, 34 },
		{ 357, 35, 34 },
		{ 357, 36, 34 },
		{ 357, 37, 34 },
		{ 357, 38, 34 },
		{ 357, 39, 34 },
		{ 357, 40, 34 },
		{ 357, 41, 34 },
		{ 357, 42, 34 },
		{ 357, 43, 34 },
		{ 357, 44, 34 },
		{ 357, 45, 34 },
		{ 357, 46, 34 },
		{ 357, 47, 34 },
		{ 357, 48, 34 },
		{ 357, 49, 34 },
		{ 357, 50, 34 },
		{ 357, 51, 34 },
		{ 357, 52, 34 },
		{ 357, 53, 34 },
		{ 357, 54, 34 },
		{ 357, 55, 34 },
		{ 357, 56, 34 },
		{ 357, 57, 34 },
		{ 357, 58, 34 },
		{ 357, 59, 34 },
		{ 357, 60, 34 },
		{ 357, 61, 34 },
		{ 357, 62, 34 },
		{ 357, 63, 34 },
		{ 357, 64, 34 },
		{ 357, 65, 34 },
		{ 357, 66, 34 },
		{ 357, 67, 34 },
		{ 357, 68, 34 },
		{ 357, 69, 34 },
		{ 357, 70, 34 },
		{ 357, 71, 34 },
		{ 357, 72, 34 },
		{ 357, 73, 34 },
		{ 357, 74, 34 },
		{ 357, 75, 34 },
		{ 357, 76, 34 },
		{ 357, 77, 34 },
		{ 357, 78, 34 },
		{ 357, 79, 34 },
		{ 357, 80, 34 },
		{ 357, 81, 34 },
		{ 357, 82, 34 },
		{ 357, 83, 34 },
		{ 357, 84, 34 },
		{ 357, 85, 34 },
		{ 357, 86, 34 },
		{ 357, 87, 34 },
		{ 357, 88, 34 },
		{ 357, 89, 34 },
		{ 357, 90, 34 },
		{ 357, 91, 34 },
		{ 357, 92, 34 },
		{ 357, 93, 34 },
		{ 357, 94, 34 },
		{ 357, 95, 34 },
		{ 357, 96, 34 },
		{ 358, 1, 34 },
		{ 358, 2, 34 },
		{ 358, 3, 34 },
		{ 358, 4, 34 },
		{ 358, 5, 34 },
		{ 358, 6, 34 },
		{ 358, 7, 34 },
		{ 358, 8, 34 },
		{ 358, 9, 34 },
		{ 358, 10, 34 },
		{ 358, 11, 34 },
		{ 358, 12, 34 },
		{ 358, 13, 34 },
		{ 358, 14, 34 },
		{ 358, 15, 34 },
		{ 358, 16, 34 },
		{ 358, 17, 34 },
		{ 358, 18, 34 },
		{ 358, 19, 34 },
		{ 358, 20, 34 },
		{ 358, 21, 34 },
		{ 358, 22, 34 },
		{ 358, 23, 34 },
		{ 358, 24, 34 },
		{ 358, 25, 34 },
		{ 358, 26, 34 },
		{ 358, 27, 34 },
		{ 358, 28, 34 },
		{ 358, 29, 34 },
		{ 358, 30, 34 },
		{ 358, 31, 34 },
		{ 358, 32, 34 },
		{ 358, 33, 34 },
		{ 358, 34, 34 },
		{ 358, 35, 34 },
		{ 358, 36, 34 },
		{ 358, 37, 34 },
		{ 358, 38, 34 },
		{ 358, 39, 34 },
		{ 358, 40, 34 },
		{ 358, 41, 34 },
		{ 358, 42, 34 },
		{ 358, 43, 34 },
		{ 358, 44, 34 },
		{ 358, 45, 34 },
		{ 358, 46, 34 },
		{ 358, 47, 34 },
		{ 358, 48, 34 },
		{ 358, 49, 34 },
		{ 358, 50, 34 },
		{ 358, 51, 34 },
		{ 358, 52, 34 },
		{ 358, 53, 34 },
		{ 358, 54, 34 },
		{ 358, 55, 34 },
		{ 358, 56, 34 },
		{ 358, 57, 34 },
		{ 358, 58, 34 },
		{ 358, 59, 34 },
		{ 358, 60, 34 },
		{ 358, 61, 34 },
		{ 358, 62, 34 },
		{ 358, 63, 34 },
		{ 358, 64, 34 },
		{ 358, 65, 34 },
		{ 358, 66, 34 },
		{ 358, 67, 34 },
		{ 358, 68, 34 },
		{ 358, 69, 34 },
		{ 358, 70, 34 },
		{ 358, 71, 34 },
		{ 358, 72, 34 },
		{ 358, 73, 34 },
		{ 358, 74, 34 },
		{ 358, 75, 34 },
		{ 358, 76, 34 },
		{ 358, 77, 34 },
		{ 358, 78, 34 },
		{ 358, 79, 34 },
		{ 358, 80, 34 },
		{ 358, 81, 34 },
		{ 358, 82, 34 },
		{ 358, 83, 34 },
		{ 358, 84, 34 },
		{ 358, 85, 34 },
		{ 358, 86, 34 },
		{ 358, 87, 34 },
		{ 358, 88, 34 },
		{ 358, 89, 34 },
		{ 358, 90, 34 },
		{ 358, 91, 34 },
		{ 358, 92, 34 },
		{ 358, 93, 34 },
		{ 358, 94, 34 },
		{ 358, 95, 34 },
		{ 359, 1, 34 },
		{ 359, 2, 34 },
		{ 359, 3, 34 },
		{ 359, 4, 34 },
		{ 359, 5, 34 },
		{ 359, 6, 34 },
		{ 359, 7, 34 },
		{ 359, 8, 34 },
		{ 359, 9, 34 },
		{ 359, 10, 34 },
		{ 359, 11, 34 },
		{ 359, 12, 34 },
		{ 359, 13, 34 },
		{ 359, 14, 34 },
		{ 359, 15, 34 },
		{ 359, 16, 34 },
		{ 359, 17, 34 },
		{ 359, 18, 34 },
		{ 359, 19, 34 },
		{ 359, 20, 34 },
		{ 359, 21, 34 },
		{ 359, 22, 34 },
		{ 359, 23, 34 },
		{ 359, 24, 34 },
		{ 359, 25, 34 },
		{ 359, 26, 34 },
		{ 359, 27, 34 },
		{ 359, 28, 34 },
		{ 359, 29, 34 },
		{ 359, 30, 34 },
		{ 359, 31, 34 },
		{ 359, 32, 34 },
		{ 359, 33, 34 },
		{ 359, 34, 34 },
		{ 359, 35, 34 },
		{ 359, 36, 34 },
		{ 359, 37, 34 },
		{ 359, 38, 34 },
		{ 359, 39, 34 },
		{ 359, 40, 34 },
		{ 359, 41, 34 },
		{ 359, 42, 34 },
		{ 359, 43, 34 },
		{ 359, 44, 34 },
		{ 359, 45, 34 },
		{ 359, 46, 34 },
		{ 359, 47, 34 },
		{ 359, 48, 34 },
		{ 359, 49, 34 },
		{ 359, 50, 34 },
		{ 359, 51, 34 },
		{ 359, 52, 34 },
		{ 359, 53, 34 },
		{ 359, 54, 34 },
		{ 359, 55, 34 },
		{ 359, 56, 34 },
		{ 359, 57, 34 },
		{ 359, 58, 34 },
		{ 359, 59, 34 },
		{ 359, 60, 34 },
		{ 359, 61, 34 },
		{ 359, 62, 34 },
		{ 359, 63, 34 },
		{ 359, 64, 34 },
		{ 359, 65, 34 },
		{ 359, 66, 34 },
		{ 359, 67, 34 },
		{ 359, 68, 34 },
		{ 359, 69, 34 },
		{ 359, 70, 34 },
		{ 359, 71, 34 },
		{ 359, 72, 34 },
		{ 359, 73, 34 },
		{ 359, 74, 34 },
		{ 359, 75, 34 },
		{ 359, 76, 34 },
		{ 359, 77, 34 },
		{ 359, 78, 34 },
		{ 359, 79, 34 },
		{ 359, 80, 34 },
		{ 359, 81, 34 },
		{ 359, 82, 34 },
		{ 359, 83, 34 },
		{ 359, 84, 34 },
		{ 359, 85, 34 },
		{ 359, 86, 34 },
		{ 359, 87, 34 },
		{ 359, 88, 34 },
		{ 359, 89, 34 },
		{ 359, 90, 34 },
		{ 359, 91, 34 },
		{ 359, 92, 34 },
		{ 359, 93, 34 },
		{ 360, 1, 34 },
		{ 360, 2, 34 },
		{ 360, 3, 34 },
		{ 360, 4, 34 },
		{ 360, 5, 34 },
		{ 360, 6, 34 },
		{ 360, 7, 34 },
		{ 360, 8, 34 },
		{ 360, 9, 34 },
		{ 360, 10, 34 },
		{ 360, 11, 34 },
		{ 360, 12, 34 },
		{ 360, 13, 34 },
		{ 360, 14, 34 },
		{ 360, 15, 34 },
		{ 360, 16, 34 },
		{ 360, 17, 34 },
		{ 360, 18, 34 },
		{ 360, 19, 34 },
		{ 360, 20, 34 },
		{ 360, 21, 34 },
		{ 360, 22, 34 },
		{ 360, 23, 34 },
		{ 360, 24, 34 },
		{ 360, 25, 34 },
		{ 360, 26, 34 },
		{ 360, 27, 34 },
		{ 360, 28, 34 },
		{ 360, 29, 34 },
		{ 360, 30, 34 },
		{ 360, 31, 34 },
		{ 360, 32, 34 },
		{ 360, 33, 34 },
		{ 360, 34, 34 },
		{ 360, 35, 34 },
		{ 360, 36, 34 },
		{ 360, 37, 34 },
		{ 360, 38, 34 },
		{ 360, 39, 34 },
		{ 360, 40, 34 },
		{ 360, 41, 34 },
		{ 360, 42, 34 },
		{ 360, 43, 34 },
		{ 360, 44, 34 },
		{ 360, 45, 34 },
		{ 360, 46, 34 },
		{ 360, 47, 34 },
		{ 360, 48, 34 },
		{ 360, 49, 34 },
		{ 360, 50, 34 },
		{ 360, 51, 34 },
		{ 360, 52, 34 },
		{ 360, 53, 34 },
		{ 360, 54, 34 },
		{ 360, 55, 34 },
		{ 360, 56, 34 },
		{ 360, 57, 34 },
		{ 360, 58, 34 },
		{ 360, 59, 34 },
		{ 360, 60, 34 },
		{ 360, 61, 34 },
		{ 360, 62, 34 },
		{ 360, 63, 34 },
		{ 360, 64, 34 },
		{ 360, 65, 34 },
		{ 360, 66, 34 },
		{ 360, 67, 34 },
		{ 360, 68, 34 },
		{ 360, 69, 34 },
		{ 360, 70, 34 },
		{ 360, 71, 34 },
		{ 360, 72, 34 },
		{ 360, 73, 34 },
		{ 360, 74, 34 },
		{ 360, 75, 34 },
		{ 360, 76, 34 },
		{ 360, 77, 34 },
		{ 360, 78, 34 },
		{ 360, 79, 34 },
		{ 360, 80, 34 },
		{ 360, 81, 34 },
		{ 360, 82, 34 },
		{ 360, 83, 34 },
		{ 360, 84, 34 },
		{ 360, 85, 34 },
		{ 360, 86, 34 },
		{ 360, 87, 34 },
		{ 360, 88, 34 },
		{ 360, 89, 34 },
		{ 360, 90, 34 },
		{ 360, 91, 34 },
		{ 360, 92, 34 },
		{ 361, 1, 34 },
		{ 361, 2, 34 },
		{ 361, 3, 34 },
		{ 361, 4, 34 },
		{ 361, 5, 34 },
		{ 361, 6, 34 },
		{ 361, 7, 34 },
		{ 361, 8, 34 },
		{ 361, 9, 34 },
		{ 361, 10, 34 },
		{ 361, 11, 34 },
		{ 361, 12, 34 },
		{ 361, 13, 34 },
		{ 361, 14, 34 },
		{ 361, 15, 34 },
		{ 361, 16, 34 },
		{ 361, 17, 34 },
		{ 361, 18, 34 },
		{ 361, 19, 34 },
		{ 361, 20, 34 },
		{ 361, 21, 34 },
		{ 361, 22, 34 },
		{ 361, 23, 34 },
		{ 361, 24, 34 },
		{ 361, 25, 34 },
		{ 361, 26, 34 },
		{ 361, 27, 34 },
		{ 361, 28, 34 },
		{ 361, 29, 34 },
		{ 361, 30, 34 },
		{ 361, 31, 34 },
		{ 361, 32, 34 },
		{ 361, 33, 34 },
		{ 361, 34, 34 },
		{ 361, 35, 34 },
		{ 361, 36, 34 },
		{ 361, 37, 34 },
		{ 361, 38, 34 },
		{ 361, 39, 34 },
		{ 361, 40, 34 },
		{ 361, 41, 34 },
		{ 361, 42, 34 },
		{ 361, 43, 34 },
		{ 361, 44, 34 },
		{ 361, 45, 34 },
		{ 361, 46, 34 },
		{ 361, 47, 34 },
		{ 361, 48, 34 },
		{ 361, 49, 34 },
		{ 361, 50, 34 },
		{ 361, 51, 34 },
		{ 361, 52, 34 },
		{ 361, 53, 34 },
		{ 361, 54, 34 },
		{ 361, 55, 34 },
		{ 361, 56, 34 },
		{ 361, 57, 34 },
		{ 361, 58, 34 },
		{ 361, 59, 34 },
		{ 361, 60, 34 },
		{ 361, 61, 34 },
		{ 361, 62, 34 },
		{ 361, 63, 34 },
		{ 361, 64, 34 },
		{ 361, 65, 34 },
		{ 361, 66, 34 },
		{ 361, 67, 34 },
		{ 361, 68, 34 },
		{ 361, 69, 34 },
		{ 361, 70, 34 },
		{ 361, 71, 34 },
		{ 361, 72, 34 },
		{ 361, 73, 34 },
		{ 361, 74, 34 },
		{ 361, 75, 34 },
		{ 361, 76, 34 },
		{ 361, 77, 34 },
		{ 361, 78, 34 },
		{ 361, 79, 34 },
		{ 361, 80, 34 },
		{ 361, 81, 34 },
		{ 361, 82, 34 },
		{ 361, 83, 34 },
		{ 361, 84, 34 },
		{ 361, 85, 34 },
		{ 361, 86, 34 },
		{ 361, 87, 34 },
		{ 361, 88, 34 },
		{ 361, 89, 34 },
		{ 361, 90, 34 },
		{ 362, 1, 34 },
		{ 362, 2, 34 },
		{ 362, 3, 34 },
		{ 362, 4, 34 },
		{ 362, 5, 34 },
		{ 362, 6, 34 },
		{ 362, 7, 34 },
		{ 362, 8, 34 },
		{ 362, 9, 34 },
		{ 362, 10, 34 },
		{ 362, 11, 34 },
		{ 362, 12, 34 },
		{ 362, 13, 34 },
		{ 362, 14, 34 },
		{ 362, 15, 34 },
		{ 362, 16, 34 },
		{ 362, 17, 34 },
		{ 362, 18, 34 },
		{ 362, 19, 34 },
		{ 362, 20, 34 },
		{ 362, 21, 34 },
		{ 362, 22, 34 },
		{ 362, 23, 34 },
		{ 362, 24, 34 },
		{ 362, 25, 34 },
		{ 362, 26, 34 },
		{ 362, 27, 34 },
		{ 362, 28, 34 },
		{ 362, 29, 34 },
		{ 362, 30, 34 },
		{ 362, 31, 34 },
		{ 362, 32, 34 },
		{ 362, 33, 34 },
		{ 362, 34, 34 },
		{ 362, 35, 34 },
		{ 362, 36, 34 },
		{ 362, 37, 34 },
		{ 362, 38, 34 },
		{ 362, 39, 34 },
		{ 362, 40, 34 },
		{ 362, 41, 34 },
		{ 362, 42, 34 },
		{ 362, 43, 34 },
		{ 362, 44, 34 },
		{ 362, 45, 34 },
		{ 362, 46, 34 },
		{ 362, 47, 34 },
		{ 362, 48, 34 },
		{ 362, 49, 34 },
		{ 362, 50, 34 },
		{ 362, 51, 34 },
		{ 362, 52, 34 },
		{ 362, 53, 34 },
		{ 362, 54, 34 },
		{ 362, 55, 34 },
		{ 362, 56, 34 },
		{ 362, 57, 34 },
		{ 362, 58, 34 },
		{ 362, 59, 34 },
		{ 362, 60, 34 },
		{ 362, 61, 34 },
		{ 362, 62, 34 },
		{ 362, 63, 34 },
		{ 362, 64, 34 },
		{ 362, 65, 34 },
		{ 362, 66, 34 },
		{ 362, 67, 34 },
		{ 362, 68, 34 },
		{ 362, 69, 34 },
		{ 362, 70, 34 },
		{ 362, 71, 34 },
		{ 362, 72, 34 },
		{ 362, 73, 34 },
		{ 362, 74, 34 },
		{ 362, 75, 34 },
		{ 362, 76, 34 },
		{ 362, 77, 34 },
		{ 362, 78, 34 },
		{ 362, 79, 34 },
		{ 362, 80, 34 },
		{ 362, 81, 34 },
		{ 362, 82, 34 },
		{ 362, 83, 34 },
		{ 362, 84, 34 },
		{ 362, 85, 34 },
		{ 362, 86, 34 },
		{ 362, 87, 34 },
		{ 362, 88, 34 },
		{ 363, 1, 34 },
		{ 363, 2, 34 },
		{ 363, 3, 34 },
		{ 363, 4, 34 },
		{ 363, 5, 34 },
		{ 363, 6, 34 },
		{ 363, 7, 34 },
		{ 363, 8, 34 },
		{ 363, 9, 34 },
		{ 363, 10, 34 },
		{ 363, 11, 34 },
		{ 363, 12, 34 },
		{ 363, 13, 34 },
		{ 363, 14, 34 },
		{ 363, 15, 34 },
		{ 363, 16, 34 },
		{ 363, 17, 34 },
		{ 363, 18, 34 },
		{ 363, 19, 34 },
		{ 363, 20, 34 },
		{ 363, 21, 34 },
		{ 363, 22, 34 },
		{ 363, 23, 34 },
		{ 363, 24, 34 },
		{ 363, 25, 34 },
		{ 363, 26, 34 },
		{ 363, 27, 34 },
		{ 363, 28, 34 },
		{ 363, 29, 34 },
		{ 363, 30, 34 },
		{ 363, 31, 34 },
		{ 363, 32, 34 },
		{ 363, 33, 34 },
		{ 363, 34, 34 },
		{ 363, 35, 34 },
		{ 363, 36, 34 },
		{ 363, 37, 34 },
		{ 363, 38, 34 },
		{ 363, 39, 34 },
		{ 363, 40, 34 },
		{ 363, 41, 34 },
		{ 363, 42, 34 },
		{ 363, 43, 34 },
		{ 363, 44, 34 },
		{ 363, 45, 34 },
		{ 363, 46, 34 },
		{ 363, 47, 34 },
		{ 363, 48, 34 },
		{ 363, 49, 34 },
		{ 363, 50, 34 },
		{ 363, 51, 34 },
		{ 363, 52, 34 },
		{ 363, 53, 34 },
		{ 363, 54, 34 },
		{ 363, 55, 34 },
		{ 363, 56, 34 },
		{ 363, 57, 34 },
		{ 363, 58, 34 },
		{ 363, 59, 34 },
		{ 363, 60, 34 },
		{ 363, 61, 34 },
		{ 363, 62, 34 },
		{ 363, 63, 34 },
		{ 363, 64, 34 },
		{ 363, 65, 34 },
		{ 363, 66, 34 },
		{ 363, 67, 34 },
		{ 363, 68, 34 },
		{ 363, 69, 34 },
		{ 363, 70, 34 },
		{ 363, 71, 34 },
		{ 363, 72, 34 },
		{ 363, 73, 34 },
		{ 363, 74, 34 },
		{ 363, 75, 34 },
		{ 363, 76, 34 },
		{ 363, 77, 34 },
		{ 363, 78, 34 },
		{ 363, 79, 34 },
		{ 363, 80, 34 },
		{ 363, 81, 34 },
		{ 363, 82, 34 },
		{ 363, 83, 34 },
		{ 363, 84, 34 },
		{ 363, 85, 34 },
		{ 363, 86, 34 },
		{ 363, 87, 34 },
		{ 364, 1, 34 },
		{ 364, 2, 34 },
		{ 364, 3, 34 },
		{ 364, 4, 34 },
		{ 364, 5, 34 },
		{ 364, 6, 34 },
		{ 364, 7, 34 },
		{ 364, 8, 34 },
		{ 364, 9, 34 },
		{ 364, 10, 34 },
		{ 364, 11, 34 },
		{ 364, 12, 34 },
		{ 364, 13, 34 },
		{ 364, 14, 34 },
		{ 364, 15, 34 },
		{ 364, 16, 34 },
		{ 364, 17, 34 },
		{ 364, 18, 34 },
		{ 364, 19, 34 },
		{ 364, 20, 34 },
		{ 364, 21, 34 },
		{ 364, 22, 34 },
		{ 364, 23, 34 },
		{ 364, 24, 34 },
		{ 364, 25, 34 },
		{ 364, 26, 34 },
		{ 364, 27, 34 },
		{ 364, 28, 34 },
		{ 364, 29, 34 },
		{ 364, 30, 34 },
		{ 364, 31, 34 },
		{ 364, 32, 34 },
		{ 364, 33, 34 },
		{ 364, 34, 34 },
		{ 364, 35, 34 },
		{ 364, 36, 34 },
		{ 364, 37, 34 },
		{ 364, 38, 34 },
		{ 364, 39, 34 },
		{ 364, 40, 34 },
		{ 364, 41, 34 },
		{ 364, 42, 34 },
		{ 364, 43, 34 },
		{ 364, 44, 34 },
		{ 364, 45, 34 },
		{ 364, 46, 34 },
		{ 364, 47, 34 },
		{ 364, 48, 34 },
		{ 364, 49, 34 },
		{ 364, 50, 34 },
		{ 364, 51, 34 },
		{ 364, 52, 34 },
		{ 364, 53, 34 },
		{ 364, 54, 34 },
		{ 364, 55, 34 },
		{ 364, 56, 34 },
		{ 364, 57, 34 },
		{ 364, 58, 34 },
		{ 364, 59, 34 },
		{ 364, 60, 34 },
		{ 364, 61, 34 },
		{ 364, 62, 34 },
		{ 364, 63, 34 },
		{ 364, 64, 34 },
		{ 364, 65, 34 },
		{ 364, 66, 34 },
		{ 364, 67, 34 },
		{ 364, 68, 34 },
		{ 364, 69, 34 },
		{ 364, 70, 34 },
		{ 364, 71, 34 },
		{ 364, 72, 34 },
		{ 364, 73, 34 },
		{ 364, 74, 34 },
		{ 364, 75, 34 },
		{ 364, 76, 34 },
		{ 364, 77, 34 },
		{ 364, 78, 34 },
		{ 364, 79, 34 },
		{ 364, 80, 34 },
		{ 364, 81, 34 },
		{ 364, 82, 34 },
		{ 364, 83, 34 },
		{ 364, 84, 34 },
		{ 364, 85, 34 },
		{ 365, 1, 34 },
		{ 365, 2, 34 },
		{ 365, 3, 34 },
		{ 365, 4, 34 },
		{ 365, 5, 34 },
		{ 365, 6, 34 },
		{ 365, 7, 34 },
		{ 365, 8, 34 },
		{ 365, 9, 34 },
		{ 365, 10, 34 },
		{ 365, 11, 34 },
		{ 365, 12, 34 },
		{ 365, 13, 34 },
		{ 365, 14, 34 },
		{ 365, 15, 34 },
		{ 365, 16, 34 },
		{ 365, 17, 34 },
		{ 365, 18, 34 },
		{ 365, 19, 34 },
		{ 365, 20, 34 },
		{ 365, 21, 34 },
		{ 365, 22, 34 },
		{ 365, 23, 34 },
		{ 365, 24, 34 },
		{ 365, 25, 34 },
		{ 365, 26, 34 },
		{ 365, 27, 34 },
		{ 365, 28, 34 },
		{ 365, 29, 34 },
		{ 365, 30, 34 },
		{ 365, 31, 34 },
		{ 365, 32, 34 },
		{ 365, 33, 34 },
		{ 365, 34, 34 },
		{ 365, 35, 34 },
		{ 365, 36, 34 },
		{ 365, 37, 34 },
		{ 365, 38, 34 },
		{ 365, 39, 34 },
		{ 365, 40, 34 },
		{ 365, 41, 34 },
		{ 365, 42, 34 },
		{ 365, 43, 34 },
		{ 365, 44, 34 },
		{ 365, 45, 34 },
		{ 365, 46, 34 },
		{ 365, 47, 34 },
		{ 365, 48, 34 },
		{ 365, 49, 34 },
		{ 365, 50, 34 },
		{ 365, 51, 34 },
		{ 365, 52, 34 },
		{ 365, 53, 34 },
		{ 365, 54, 34 },
		{ 365, 55, 34 },
		{ 365, 56, 34 },
		{ 365, 57, 34 },
		{ 365, 58, 34 },
		{ 365, 59, 34 },
		{ 365, 60, 34 },
		{ 365, 61, 34 },
		{ 365, 62, 34 },
		{ 365, 63, 34 },
		{ 365, 64, 34 },
		{ 365, 65, 34 },
		{ 365, 66, 34 },
		{ 365, 67, 34 },
		{ 365, 68, 34 },
		{ 365, 69, 34 },
		{ 365, 70, 34 },
		{ 365, 71, 34 },
		{ 365, 72, 34 },
		{ 365, 73, 34 },
		{ 365, 74, 34 },
		{ 365, 75, 34 },
		{ 365, 76, 34 },
		{ 365, 77, 34 },
		{ 365, 78, 34 },
		{ 365, 79, 34 },
		{ 365, 80, 34 },
		{ 365, 81, 34 },
		{ 365, 82, 34 },
		{ 365, 83, 34 },
		{ 365, 84, 34 },
		{ 366, 1, 34 },
		{ 366, 2, 34 },
		{ 366, 3, 34 },
		{ 366, 4, 34 },
		{ 366, 5, 34 },
		{ 366, 6, 34 },
		{ 366, 7, 34 },
		{ 366, 8, 34 },
		{ 366, 9, 34 },
		{ 366, 10, 34 },
		{ 366, 11, 34 },
		{ 366, 12, 34 },
		{ 366, 13, 34 },
		{ 366, 14, 34 },
		{ 366, 15, 34 },
		{ 366, 16, 34 },
		{ 366, 17, 34 },
		{ 366, 18, 34 },
		{ 366, 19, 34 },
		{ 366, 20, 34 },
		{ 366, 21, 34 },
		{ 366, 22, 34 },
		{ 366, 23, 34 },
		{ 366, 24, 34 },
		{ 366, 25, 34 },
		{ 366, 26, 34 },
		{ 366, 27, 34 },
		{ 366, 28, 34 },
		{ 366, 29, 34 },
		{ 366, 30, 34 },
		{ 366, 31, 34 },
		{ 366, 32, 34 },
		{ 366, 33, 34 },
		{ 366, 34, 34 },
		{ 366, 35, 34 },
		{ 366, 36, 34 },
		{ 366, 37, 34 },
		{ 366, 38, 34 },
		{ 366, 39, 34 },
		{ 366, 40, 34 },
		{ 366, 41, 34 },
		{ 366, 42, 34 },
		{ 366, 43, 34 },
		{ 366, 44, 34 },
		{ 366, 45, 34 },
		{ 366, 46, 34 },
		{ 366, 47, 34 },
		{ 366, 48, 34 },
		{ 366, 49, 34 },
		{ 366, 50, 34 },
		{ 366, 51, 34 },
		{ 366, 52, 34 },
		{ 366, 53, 34 },
		{ 366, 54, 34 },
		{ 366, 55, 34 },
		{ 366, 56, 34 },
		{ 366, 57, 34 },
		{ 366, 58, 34 },
		{ 366, 59, 34 },
		{ 366, 60, 34 },
		{ 366, 61, 34 },
		{ 366, 62, 34 },
		{ 366, 63, 34 },
		{ 366, 64, 34 },
		{ 366, 65, 34 },
		{ 366, 66, 34 },
		{ 366, 67, 34 },
		{ 366, 68, 34 },
		{ 366, 69, 34 },
		{ 366, 70, 34 },
		{ 366, 71, 34 },
		{ 366, 72, 34 },
		{ 366, 73, 34 },
		{ 366, 74, 34 },
		{ 366, 75, 34 },
		{ 366, 76, 34 },
		{ 366, 77, 34 },
		{ 366, 78, 34 },
		{ 366, 79, 34 },
		{ 366, 80, 34 },
		{ 366, 81, 34 },
		{ 366, 82, 34 },
		{ 367, 1, 34 },
		{ 367, 2, 34 },
		{ 367, 3, 34 },
		{ 367, 4, 34 },
		{ 367, 5, 34 },
		{ 367, 6, 34 },
		{ 367, 7, 34 },
		{ 367, 8, 34 },
		{ 367, 9, 34 },
		{ 367, 10, 34 },
		{ 367, 11, 34 },
		{ 367, 12, 34 },
		{ 367, 13, 34 },
		{ 367, 14, 34 },
		{ 367, 15, 34 },
		{ 367, 16, 34 },
		{ 367, 17, 34 },
		{ 367, 18, 34 },
		{ 367, 19, 34 },
		{ 367, 20, 34 },
		{ 367, 21, 34 },
		{ 367, 22, 34 },
		{ 367, 23, 34 },
		{ 367, 24, 34 },
		{ 367, 25, 34 },
		{ 367, 26, 34 },
		{ 367, 27, 34 },
		{ 367, 28, 34 },
		{ 367, 29, 34 },
		{ 367, 30, 34 },
		{ 367, 31, 34 },
		{ 367, 32, 34 },
		{ 367, 33, 34 },
		{ 367, 34, 34 },
		{ 367, 35, 34 },
		{ 367, 36, 34 },
		{ 367, 37, 34 },
		{ 367, 38, 34 },
		{ 367, 39, 34 },
		{ 367, 40, 34 },
		{ 367, 41, 34 },
		{ 367, 42, 34 },
		{ 367, 43, 34 },
		{ 367, 44, 34 },
		{ 367, 45, 34 },
		{ 367, 46, 34 },
		{ 367, 47, 34 },
		{ 367, 48, 34 },
		{ 367, 49, 34 },
		{ 367, 50, 34 },
		{ 367, 51, 34 },
		{ 367, 52, 34 },
		{ 367, 53, 34 },
		{ 367, 54, 34 },
		{ 367, 55, 34 },
		{ 367, 56, 34 },
		{ 367, 57, 34 },
		{ 367, 58, 34 },
		{ 367, 59, 34 },
		{ 367, 60, 34 },
		{ 367, 61, 34 },
		{ 367, 62, 34 },
		{ 367, 63, 34 },
		{ 367, 64, 34 },
		{ 367, 65, 34 },
		{ 367, 66, 34 },
		{ 367, 67, 34 },
		{ 367, 68, 34 },
		{ 367, 69, 34 },
		{ 367, 70, 34 },
		{ 367, 71, 34 },
		{ 367, 72, 34 },
		{ 367, 73, 34 },
		{ 367, 74, 34 },
		{ 367, 75, 34 },
		{ 367, 76, 34 },
		{ 367, 77, 34 },
		{ 367, 78, 34 },
		{ 367, 79, 34 },
		{ 367, 80, 34 },
		{ 368, 1, 34 },
		{ 368, 2, 34 },
		{ 368, 3, 34 },
		{ 368, 4, 34 },
		{ 368, 5, 34 },
		{ 368, 6, 34 },
		{ 368, 7, 34 },
		{ 368, 8, 34 },
		{ 368, 9, 34 },
		{ 368, 10, 34 },
		{ 368, 11, 34 },
		{ 368, 12, 34 },
		{ 368, 13, 34 },
		{ 368, 14, 34 },
		{ 368, 15, 34 },
		{ 368, 16, 34 },
		{ 368, 17, 34 },
		{ 368, 18, 34 },
		{ 368, 19, 34 },
		{ 368, 20, 34 },
		{ 368, 21, 34 },
		{ 368, 22, 34 },
		{ 368, 23, 34 },
		{ 368, 24, 34 },
		{ 368, 25, 34 },
		{ 368, 26, 34 },
		{ 368, 27, 34 },
		{ 368, 28, 34 },
		{ 368, 29, 34 },
		{ 368, 30, 34 },
		{ 368, 31, 34 },
		{ 368, 32, 34 },
		{ 368, 33, 34 },
		{ 368, 34, 34 },
		{ 368, 35, 34 },
		{ 368, 36, 34 },
		{ 368, 37, 34 },
		{ 368, 38, 34 },
		{ 368, 39, 34 },
		{ 368, 40, 34 },
		{ 368, 41, 34 },
		{ 368, 42, 34 },
		{ 368, 43, 34 },
		{ 368, 44, 34 },
		{ 368, 45, 34 },
		{ 368, 46, 34 },
		{ 368, 47, 34 },
		{ 368, 48, 34 },
		{ 368, 49, 34 },
		{ 368, 50, 34 },
		{ 368, 51, 34 },
		{ 368, 52, 34 },
		{ 368, 53, 34 },
		{ 368, 54, 34 },
		{ 368, 55, 34 },
		{ 368, 56, 34 },
		{ 368, 57, 34 },
		{ 368, 58, 34 },
		{ 368, 59, 34 },
		{ 368, 60, 34 },
		{ 368, 61, 34 },
		{ 368, 62, 34 },
		{ 368, 63, 34 },
		{ 368, 64, 34 },
		{ 368, 65, 34 },
		{ 368, 66, 34 },
		{ 368, 67, 34 },
		{ 368, 68, 34 },
		{ 368, 69, 34 },
		{ 368, 70, 34 },
		{ 368, 71, 34 },
		{ 368, 72, 34 },
		{ 368, 73, 34 },
		{ 368, 74, 34 },
		{ 368, 75, 34 },
		{ 368, 76, 34 },
		{ 368, 77, 34 },
		{ 368, 78, 34 },
		{ 368, 79, 34 },
		{ 369, 1, 34 },
		{ 369, 2, 34 },
		{ 369, 3, 34 },
		{ 369, 4, 34 },
		{ 369, 5, 34 },
		{ 369, 6, 34 },
		{ 369, 7, 34 },
		{ 369, 8, 34 },
		{ 369, 9, 34 },
		{ 369, 10, 34 },
		{ 369, 11, 34 },
		{ 369, 12, 34 },
		{ 369, 13, 34 },
		{ 369, 14, 34 },
		{ 369, 15, 34 },
		{ 369, 16, 34 },
		{ 369, 17, 34 },
		{ 369, 18, 34 },
		{ 369, 19, 34 },
		{ 369, 20, 34 },
		{ 369, 21, 34 },
		{ 369, 22, 34 },
		{ 369, 23, 34 },
		{ 369, 24, 34 },
		{ 369, 25, 34 },
		{ 369, 26, 34 },
		{ 369, 27, 34 },
		{ 369, 28, 34 },
		{ 369, 29, 34 },
		{ 369, 30, 34 },
		{ 369, 31, 34 },
		{ 369, 32, 34 },
		{ 369, 33, 34 },
		{ 369, 34, 34 },
		{ 369, 35, 34 },
		{ 369, 36, 34 },
		{ 369, 37, 34 },
		{ 369, 38, 34 },
		{ 369, 39, 34 },
		{ 369, 40, 34 },
		{ 369, 41, 34 },
		{ 369, 42, 34 },
		{ 369, 43, 34 },
		{ 369, 44, 34 },
		{ 369, 45, 34 },
		{ 369, 46, 34 },
		{ 369, 47, 34 },
		{ 369, 48, 34 },
		{ 369, 49, 34 },
		{ 369, 50, 34 },
		{ 369, 51, 34 },
		{ 369, 52, 34 },
		{ 369, 53, 34 },
		{ 369, 54, 34 },
		{ 369, 55, 34 },
		{ 369, 56, 34 },
		{ 369, 57, 34 },
		{ 369, 58, 34 },
		{ 369, 59, 34 },
		{ 369, 60, 34 },
		{ 369, 61, 34 },
		{ 369, 62, 34 },
		{ 369, 63, 34 },
		{ 369, 64, 34 },
		{ 369, 65, 34 },
		{ 369, 66, 34 },
		{ 369, 67, 34 },
		{ 369, 68, 34 },
		{ 369, 69, 34 },
		{ 369, 70, 34 },
		{ 369, 71, 34 },
		{ 369, 72, 34 },
		{ 369, 73, 34 },
		{ 369, 74, 34 },
		{ 369, 75, 34 },
		{ 369, 76, 34 },
		{ 369, 77, 34 },
		{ 370, 1, 34 },
		{ 370, 2, 34 },
		{ 370, 3, 34 },
		{ 370, 4, 34 },
		{ 370, 5, 34 },
		{ 370, 6, 34 },
		{ 370, 7, 34 },
		{ 370, 8, 34 },
		{ 370, 9, 34 },
		{ 370, 10, 34 },
		{ 370, 11, 34 },
		{ 370, 12, 34 },
		{ 370, 13, 34 },
		{ 370, 14, 34 },
		{ 370, 15, 34 },
		{ 370, 16, 34 },
		{ 370, 17, 34 },
		{ 370, 18, 34 },
		{ 370, 19, 34 },
		{ 370, 20, 34 },
		{ 370, 21, 34 },
		{ 370, 22, 34 },
		{ 370, 23, 34 },
		{ 370, 24, 34 },
		{ 370, 25, 34 },
		{ 370, 26, 34 },
		{ 370, 27, 34 },
		{ 370, 28, 34 },
		{ 370, 29, 34 },
		{ 370, 30, 34 },
		{ 370, 31, 34 },
		{ 370, 32, 34 },
		{ 370, 33, 34 },
		{ 370, 34, 34 },
		{ 370, 35, 34 },
		{ 370, 36, 34 },
		{ 370, 37, 34 },
		{ 370, 38, 34 },
		{ 370, 39, 34 },
		{ 370, 40, 34 },
		{ 370, 41, 34 },
		{ 370, 42, 34 },
		{ 370, 43, 34 },
		{ 370, 44, 34 },
		{ 370, 45, 34 },
		{ 370, 46, 34 },
		{ 370, 47, 34 },
		{ 370, 48, 34 },
		{ 370, 49, 34 },
		{ 370, 50, 34 },
		{ 370, 51, 34 },
		{ 370, 52, 34 },
		{ 370, 53, 34 },
		{ 370, 54, 34 },
		{ 370, 55, 34 },
		{ 370, 56, 34 },
		{ 370, 57, 34 },
		{ 370, 58, 34 },
		{ 370, 59, 34 },
		{ 370, 60, 34 },
		{ 370, 61, 34 },
		{ 370, 62, 34 },
		{ 370, 63, 34 },
		{ 370, 64, 34 },
		{ 370, 65, 34 },
		{ 370, 66, 34 },
		{ 370, 67, 34 },
		{ 370, 68, 34 },
		{ 370, 69, 34 },
		{ 370, 70, 34 },
		{ 370, 71, 34 },
		{ 370, 72, 34 },
		{ 370, 73, 34 },
		{ 370, 74, 34 },
		{ 370, 75, 34 },
		{ 370, 76, 34 },
		{ 371, 1, 34 },
		{ 371, 2, 34 },
		{ 371, 3, 34 },
		{ 371, 4, 34 },
		{ 371, 5, 34 },
		{ 371, 6, 34 },
		{ 371, 7, 34 },
		{ 371, 8, 34 },
		{ 371, 9, 34 },
		{ 371, 10, 34 },
		{ 371, 11, 34 },
		{ 371, 12, 34 },
		{ 371, 13, 34 },
		{ 371, 14, 34 },
		{ 371, 15, 34 },
		{ 371, 16, 34 },
		{ 371, 17, 34 },
		{ 371, 18, 34 },
		{ 371, 19, 34 },
		{ 371, 20, 34 },
		{ 371, 21, 34 },
		{ 371, 22, 34 },
		{ 371, 23, 34 },
		{ 371, 24, 34 },
		{ 371, 25, 34 },
		{ 371, 26, 34 },
		{ 371, 27, 34 },
		{ 371, 28, 34 },
		{ 371, 29, 34 },
		{ 371, 30, 34 },
		{ 371, 31, 34 },
		{ 371, 32, 34 },
		{ 371, 33, 34 },
		{ 371, 34, 34 },
		{ 371, 35, 34 },
		{ 371, 36, 34 },
		{ 371, 37, 34 },
		{ 371, 38, 34 },
		{ 371, 39, 34 },
		{ 371, 40, 34 },
		{ 371, 41, 34 },
		{ 371, 42, 34 },
		{ 371, 43, 34 },
		{ 371, 44, 34 },
		{ 371, 45, 34 },
		{ 371, 46, 34 },
		{ 371, 47, 34 },
		{ 371, 48, 34 },
		{ 371, 49, 34 },
		{ 371, 50, 34 },
		{ 371, 51, 34 },
		{ 371, 52, 34 },
		{ 371, 53, 34 },
		{ 371, 54, 34 },
		{ 371, 55, 34 },
		{ 371, 56, 34 },
		{ 371, 57, 34 },
		{ 371, 58, 34 },
		{ 371, 59, 34 },
		{ 371, 60, 34 },
		{ 371, 61, 34 },
		{ 371, 62, 34 },
		{ 371, 63, 34 },
		{ 371, 64, 34 },
		{ 371, 65, 34 },
		{ 371, 66, 34 },
		{ 371, 67, 34 },
		{ 371, 68, 34 },
		{ 371, 69, 34 },
		{ 371, 70, 34 },
		{ 371, 71, 34 },
		{ 371, 72, 34 },
		{ 371, 73, 34 },
		{ 371, 74, 34 },
		{ 372, 1, 34 },
		{ 372, 2, 34 },
		{ 372, 3, 34 },
		{ 372, 4, 34 },
		{ 372, 5, 34 },
		{ 372, 6, 34 },
		{ 372, 7, 34 },
		{ 372, 8, 34 },
		{ 372, 9, 34 },
		{ 372, 10, 34 },
		{ 372, 11, 34 },
		{ 372, 12, 34 },
		{ 372, 13, 34 },
		{ 372, 14, 34 },
		{ 372, 15, 34 },
		{ 372, 16, 34 },
		{ 372, 17, 34 },
		{ 372, 18, 34 },
		{ 372, 19, 34 },
		{ 372, 20, 34 },
		{ 372, 21, 34 },
		{ 372, 22, 34 },
		{ 372, 23, 34 },
		{ 372, 24, 34 },
		{ 372, 25, 34 },
		{ 372, 26, 34 },
		{ 372, 27, 34 },
		{ 372, 28, 34 },
		{ 372, 29, 34 },
		{ 372, 30, 34 },
		{ 372, 31, 34 },
		{ 372, 32, 34 },
		{ 372, 33, 34 },
		{ 372, 34, 34 },
		{ 372, 35, 34 },
		{ 372, 36, 34 },
		{ 372, 37, 34 },
		{ 372, 38, 34 },
		{ 372, 39, 34 },
		{ 372, 40, 34 },
		{ 372, 41, 34 },
		{ 372, 42, 34 },
		{ 372, 43, 34 },
		{ 372, 44, 34 },
		{ 372, 45, 34 },
		{ 372, 46, 34 },
		{ 372, 47, 34 },
		{ 372, 48, 34 },
		{ 372, 49, 34 },
		{ 372, 50, 34 },
		{ 372, 51, 34 },
		{ 372, 52, 34 },
		{ 372, 53, 34 },
		{ 372, 54, 34 },
		{ 372, 55, 34 },
		{ 372, 56, 34 },
		{ 372, 57, 34 },
		{ 372, 58, 34 },
		{ 372, 59, 34 },
		{ 372, 60, 34 },
		{ 372, 61, 34 },
		{ 372, 62, 34 },
		{ 372, 63, 34 },
		{ 372, 64, 34 },
		{ 372, 65, 34 },
		{ 372, 66, 34 },
		{ 372, 67, 34 },
		{ 372, 68, 34 },
		{ 372, 69, 34 },
		{ 372, 70, 34 },
		{ 372, 71, 34 },
		{ 372, 72, 34 },
		{ 373, 1, 34 },
		{ 373, 2, 34 },
		{ 373, 3, 34 },
		{ 373, 4, 34 },
		{ 373, 5, 34 },
		{ 373, 6, 34 },
		{ 373, 7, 34 },
		{ 373, 8, 34 },
		{ 373, 9, 34 },
		{ 373, 10, 34 },
		{ 373, 11, 34 },
		{ 373, 12, 34 },
		{ 373, 13, 34 },
		{ 373, 14, 34 },
		{ 373, 15, 34 },
		{ 373, 16, 34 },
		{ 373, 17, 34 },
		{ 373, 18, 34 },
		{ 373, 19, 34 },
		{ 373, 20, 34 },
		{ 373, 21, 34 },
		{ 373, 22, 34 },
		{ 373, 23, 34 },
		{ 373, 24, 34 },
		{ 373, 25, 34 },
		{ 373, 26, 34 },
		{ 373, 27, 34 },
		{ 373, 28, 34 },
		{ 373, 29, 34 },
		{ 373, 30, 34 },
		{ 373, 31, 34 },
		{ 373, 32, 34 },
		{ 373, 33, 34 },
		{ 373, 34, 34 },
		{ 373, 35, 34 },
		{ 373, 36, 34 },
		{ 373, 37, 34 },
		{ 373, 38, 34 },
		{ 373, 39, 34 },
		{ 373, 40, 34 },
		{ 373, 41, 34 },
		{ 373, 42, 34 },
		{ 373, 43, 34 },
		{ 373, 44, 34 },
		{ 373, 45, 34 },
		{ 373, 46, 34 },
		{ 373, 47, 34 },
		{ 373, 48, 34 },
		{ 373, 49, 34 },
		{ 373, 50, 34 },
		{ 373, 51, 34 },
		{ 373, 52, 34 },
		{ 373, 53, 34 },
		{ 373, 54, 34 },
		{ 373, 55, 34 },
		{ 373, 56, 34 },
		{ 373, 57, 34 },
		{ 373, 58, 34 },
		{ 373, 59, 34 },
		{ 373, 60, 34 },
		{ 373, 61, 34 },
		{ 373, 62, 34 },
		{ 373, 63, 34 },
		{ 373, 64, 34 },
		{ 373, 65, 34 },
		{ 373, 66, 34 },
		{ 373, 67, 34 },
		{ 373, 68, 34 },
		{ 373, 69, 34 },
		{ 373, 70, 34 },
		{ 373, 71, 34 },
		{ 374, 1, 34 },
		{ 374, 2, 34 },
		{ 374, 3, 34 },
		{ 374, 4, 34 },
		{ 374, 5, 34 },
		{ 374, 6, 34 },
		{ 374, 7, 34 },
		{ 374, 8, 34 },
		{ 374, 9, 34 },
		{ 374, 10, 34 },
		{ 374, 11, 34 },
		{ 374, 12, 34 },
		{ 374, 13, 34 },
		{ 374, 14, 34 },
		{ 374, 15, 34 },
		{ 374, 16, 34 },
		{ 374, 17, 34 },
		{ 374, 18, 34 },
		{ 374, 19, 34 },
		{ 374, 20, 34 },
		{ 374, 21, 34 },
		{ 374, 22, 34 },
		{ 374, 23, 34 },
		{ 374, 24, 34 },
		{ 374, 25, 34 },
		{ 374, 26, 34 },
		{ 374, 27, 34 },
		{ 374, 28, 34 },
		{ 374, 29, 34 },
		{ 374, 30, 34 },
		{ 374, 31, 34 },
		{ 374, 32, 34 },
		{ 374, 33, 34 },
		{ 374, 34, 34 },
		{ 374, 35, 34 },
		{ 374, 36, 34 },
		{ 374, 37, 34 },
		{ 374, 38, 34 },
		{ 374, 39, 34 },
		{ 374, 40, 34 },
		{ 374, 41, 34 },
		{ 374, 42, 34 },
		{ 374, 43, 34 },
		{ 374, 44, 34 },
		{ 374, 45, 34 },
		{ 374, 46, 34 },
		{ 374, 47, 34 },
		{ 374, 48, 34 },
		{ 374, 49, 34 },
		{ 374, 50, 34 },
		{ 374, 51, 34 },
		{ 374, 52, 34 },
		{ 374, 53, 34 },
		{ 374, 54, 34 },
		{ 374, 55, 34 },
		{ 374, 56, 34 },
		{ 374, 57, 34 },
		{ 374, 58, 34 },
		{ 374, 59, 34 },
		{ 374, 60, 34 },
		{ 374, 61, 34 },
		{ 374, 62, 34 },
		{ 374, 63, 34 },
		{ 374, 64, 34 },
		{ 374, 65, 34 },
		{ 374, 66, 34 },
		{ 374, 67, 34 },
		{ 374, 68, 34 },
		{ 374, 69, 34 },
		{ 375, 1, 34 },
		{ 375, 2, 34 },
		{ 375, 3, 34 },
		{ 375, 4, 34 },
		{ 375, 5, 34 },
		{ 375, 6, 34 },
		{ 375, 7, 34 },
		{ 375, 8, 34 },
		{ 375, 9, 34 },
		{ 375, 10, 34 },
		{ 375, 11, 34 },
		{ 375, 12, 34 },
		{ 375, 13, 34 },
		{ 375, 14, 34 },
		{ 375, 15, 34 },
		{ 375, 16, 34 },
		{ 375, 17, 34 },
		{ 375, 18, 34 },
		{ 375, 19, 34 },
		{ 375, 20, 34 },
		{ 375, 21, 34 },
		{ 375, 22, 34 },
		{ 375, 23, 34 },
		{ 375, 24, 34 },
		{ 375, 25, 34 },
		{ 375, 26, 34 },
		{ 375, 27, 34 },
		{ 375, 28, 34 },
		{ 375, 29, 34 },
		{ 375, 30, 34 },
		{ 375, 31, 34 },
		{ 375, 32, 34 },
		{ 375, 33, 34 },
		{ 375, 34, 34 },
		{ 375, 35, 34 },
		{ 375, 36, 34 },
		{ 375, 37, 34 },
		{ 375, 38, 34 },
		{ 375, 39, 34 },
		{ 375, 40, 34 },
		{ 375, 41, 34 },
		{ 375, 42, 34 },
		{ 375, 43, 34 },
		{ 375, 44, 34 },
		{ 375, 45, 34 },
		{ 375, 46, 34 },
		{ 375, 47, 34 },
		{ 375, 48, 34 },
		{ 375, 49, 34 },
		{ 375, 50, 34 },
		{ 375, 51, 34 },
		{ 375, 52, 34 },
		{ 375, 53, 34 },
		{ 375, 54, 34 },
		{ 375, 55, 34 },
		{ 375, 56, 34 },
		{ 375, 57, 34 },
		{ 375, 58, 34 },
		{ 375, 59, 34 },
		{ 375, 60, 34 },
		{ 375, 61, 34 },
		{ 375, 62, 34 },
		{ 375, 63, 34 },
		{ 375, 64, 34 },
		{ 375, 65, 34 },
		{ 375, 66, 34 },
		{ 375, 67, 34 },
		{ 375, 68, 34 },
		{ 376, 1, 34 },
		{ 376, 2, 34 },
		{ 376, 3, 34 },
		{ 376, 4, 34 },
		{ 376, 5, 34 },
		{ 376, 6, 34 },
		{ 376, 7, 34 },
		{ 376, 8, 34 },
		{ 376, 9, 34 },
		{ 376, 10, 34 },
		{ 376, 11, 34 },
		{ 376, 12, 34 },
		{ 376, 13, 34 },
		{ 376, 14, 34 },
		{ 376, 15, 34 },
		{ 376, 16, 34 },
		{ 376, 17, 34 },
		{ 376, 18, 34 },
		{ 376, 19, 34 },
		{ 376, 20, 34 },
		{ 376, 21, 34 },
		{ 376, 22, 34 },
		{ 376, 23, 34 },
		{ 376, 24, 34 },
		{ 376, 25, 34 },
		{ 376, 26, 34 },
		{ 376, 27, 34 },
		{ 376, 28, 34 },
		{ 376, 29, 34 },
		{ 376, 30, 34 },
		{ 376, 31, 34 },
		{ 376, 32, 34 },
		{ 376, 33, 34 },
		{ 376, 34, 34 },
		{ 376, 35, 34 },
		{ 376, 36, 34 },
		{ 376, 37, 34 },
		{ 376, 38, 34 },
		{ 376, 39, 34 },
		{ 376, 40, 34 },
		{ 376, 41, 34 },
		{ 376, 42, 34 },
		{ 376, 43, 34 },
		{ 376, 44, 34 },
		{ 376, 45, 34 },
		{ 376, 46, 34 },
		{ 376, 47, 34 },
		{ 376, 48, 34 },
		{ 376, 49, 34 },
		{ 376, 50, 34 },
		{ 376, 51, 34 },
		{ 376, 52, 34 },
		{ 376, 53, 34 },
		{ 376, 54, 34 },
		{ 376, 55, 34 },
		{ 376, 56, 34 },
		{ 376, 57, 34 },
		{ 376, 58, 34 },
		{ 376, 59, 34 },
		{ 376, 60, 34 },
		{ 376, 61, 34 },
		{ 376, 62, 34 },
		{ 376, 63, 34 },
		{ 376, 64, 34 },
		{ 376, 65, 34 },
		{ 376, 66, 34 },
		{ 377, 1, 34 },
		{ 377, 2, 34 },
		{ 377, 3, 34 },
		{ 377, 4, 34 },
		{ 377, 5, 34 },
		{ 377, 6, 34 },
		{ 377, 7, 34 },
		{ 377, 8, 34 },
		{ 377, 9, 34 },
		{ 377, 10, 34 },
		{ 377, 11, 34 },
		{ 377, 12, 34 },
		{ 377, 13, 34 },
		{ 377, 14, 34 },
		{ 377, 15, 34 },
		{ 377, 16, 34 },
		{ 377, 17, 34 },
		{ 377, 18, 34 },
		{ 377, 19, 34 },
		{ 377, 20, 34 },
		{ 377, 21, 34 },
		{ 377, 22, 34 },
		{ 377, 23, 34 },
		{ 377, 24, 34 },
		{ 377, 25, 34 },
		{ 377, 26, 34 },
		{ 377, 27, 34 },
		{ 377, 28, 34 },
		{ 377, 29, 34 },
		{ 377, 30, 34 },
		{ 377, 31, 34 },
		{ 377, 32, 34 },
		{ 377, 33, 34 },
		{ 377, 34, 34 },
		{ 377, 35, 34 },
		{ 377, 36, 34 },
		{ 377, 37, 34 },
		{ 377, 38, 34 },
		{ 377, 39, 34 },
		{ 377, 40, 34 },
		{ 377, 41, 34 },
		{ 377, 42, 34 },
		{ 377, 43, 34 },
		{ 377, 44, 34 },
		{ 377, 45, 34 },
		{ 377, 46, 34 },
		{ 377, 47, 34 },
		{ 377, 48, 34 },
		{ 377, 49, 34 },
		{ 377, 50, 34 },
		{ 377, 51, 34 },
		{ 377, 52, 34 },
		{ 377, 53, 34 },
		{ 377, 54, 34 },
		{ 377, 55, 34 },
		{ 377, 56, 34 },
		{ 377, 57, 34 },
		{ 377, 58, 34 },
		{ 377, 59, 34 },
		{ 377, 60, 34 },
		{ 377, 61, 34 },
		{ 377, 62, 34 },
		{ 377, 63, 34 },
		{ 377, 64, 34 },
		{ 378, 1, 34 },
		{ 378, 2, 34 },
		{ 378, 3, 34 },
		{ 378, 4, 34 },
		{ 378, 5, 34 },
		{ 378, 6, 34 },
		{ 378, 7, 34 },
		{ 378, 8, 34 },
		{ 378, 9, 34 },
		{ 378, 10, 34 },
		{ 378, 11, 34 },
		{ 378, 12, 34 },
		{ 378, 13, 34 },
		{ 378, 14, 34 },
		{ 378, 15, 34 },
		{ 378, 16, 34 },
		{ 378, 17, 34 },
		{ 378, 18, 34 },
		{ 378, 19, 34 },
		{ 378, 20, 34 },
		{ 378, 21, 34 },
		{ 378, 22, 34 },
		{ 378, 23, 34 },
		{ 378, 24, 34 },
		{ 378, 25, 34 },
		{ 378, 26, 34 },
		{ 378, 27, 34 },
		{ 378, 28, 34 },
		{ 378, 29, 34 },
		{ 378, 30, 34 },
		{ 378, 31, 34 },
		{ 378, 32, 34 },
		{ 378, 33, 34 },
		{ 378, 34, 34 },
		{ 378, 35, 34 },
		{ 378, 36, 34 },
		{ 378, 37, 34 },
		{ 378, 38, 34 },
		{ 378, 39, 34 },
		{ 378, 40, 34 },
		{ 378, 41, 34 },
		{ 378, 42, 34 },
		{ 378, 43, 34 },
		{ 378, 44, 34 },
		{ 378, 45, 34 },
		{ 378, 46, 34 },
		{ 378, 47, 34 },
		{ 378, 48, 34 },
		{ 378, 49, 34 },
		{ 378, 50, 34 },
		{ 378, 51, 34 },
		{ 378, 52, 34 },
		{ 378, 53, 34 },
		{ 378, 54, 34 },
		{ 378, 55, 34 },
		{ 378, 56, 34 },
		{ 378, 57, 34 },
		{ 378, 58, 34 },
		{ 378, 59, 34 },
		{ 378, 60, 34 },
		{ 378, 61, 34 },
		{ 378, 62, 34 },
		{ 378, 63, 34 },
		{ 379, 1, 34 },
		{ 379, 2, 34 },
		{ 379, 3, 34 },
		{ 379, 4, 34 },
		{ 379, 5, 34 },
		{ 379, 6, 34 },
		{ 379, 7, 34 },
		{ 379, 8, 34 },
		{ 379, 9, 34 },
		{ 379, 10, 34 },
		{ 379, 11, 34 },
		{ 379, 12, 34 },
		{ 379, 13, 34 },
		{ 379, 14, 34 },
		{ 379, 15, 34 },
		{ 379, 16, 34 },
		{ 379, 17, 34 },
		{ 379, 18, 34 },
		{ 379, 19, 34 },
		{ 379, 20, 34 },
		{ 379, 21, 34 },
		{ 379, 22, 34 },
		{ 379, 23, 34 },
		{ 379, 24, 34 },
		{ 379, 25, 34 },
		{ 379, 26, 34 },
		{ 379, 27, 34 },
		{ 379, 28, 34 },
		{ 379, 29, 34 },
		{ 379, 30, 34 },
		{ 379, 31, 34 },
		{ 379, 32, 34 },
		{ 379, 33, 34 },
		{ 379, 34, 34 },
		{ 379, 35, 34 },
		{ 379, 36, 34 },
		{ 379, 37, 34 },
		{ 379, 38, 34 },
		{ 379, 39, 34 },
		{ 379, 40, 34 },
		{ 379, 41, 34 },
		{ 379, 42, 34 },
		{ 379, 43, 34 },
		{ 379, 44, 34 },
		{ 379, 45, 34 },
		{ 379, 46, 34 },
		{ 379, 47, 34 },
		{ 379, 48, 34 },
		{ 379, 49, 34 },
		{ 379, 50, 34 },
		{ 379, 51, 34 },
		{ 379, 52, 34 },
		{ 379, 53, 34 },
		{ 379, 54, 34 },
		{ 379, 55, 34 },
		{ 379, 56, 34 },
		{ 379, 57, 34 },
		{ 379, 58, 34 },
		{ 379, 59, 34 },
		{ 379, 60, 34 },
		{ 379, 61, 34 },
		{ 380, 1, 34 },
		{ 380, 2, 34 },
		{ 380, 3, 34 },
		{ 380, 4, 34 },
		{ 380, 5, 34 },
		{ 380, 6, 34 },
		{ 380, 7, 34 },
		{ 380, 8, 34 },
		{ 380, 9, 34 },
		{ 380, 10, 34 },
		{ 380, 11, 34 },
		{ 380, 12, 34 },
		{ 380, 13, 34 },
		{ 380, 14, 34 },
		{ 380, 15, 34 },
		{ 380, 16, 34 },
		{ 380, 17, 34 },
		{ 380, 18, 34 },
		{ 380, 19, 34 },
		{ 380, 20, 34 },
		{ 380, 21, 34 },
		{ 380, 22, 34 },
		{ 380, 23, 34 },
		{ 380, 24, 34 },
		{ 380, 25, 34 },
		{ 380, 26, 34 },
		{ 380, 27, 34 },
		{ 380, 28, 34 },
		{ 380, 29, 34 },
		{ 380, 30, 34 },
		{ 380, 31, 34 },
		{ 380, 32, 34 },
		{ 380, 33, 34 },
		{ 380, 34, 34 },
		{ 380, 35, 34 },
		{ 380, 36, 34 },
		{ 380, 37, 34 },
		{ 380, 38, 34 },
		{ 380, 39, 34 },
		{ 380, 40, 34 },
		{ 380, 41, 34 },
		{ 380, 42, 34 },
		{ 380, 43, 34 },
		{ 380, 44, 34 },
		{ 380, 45, 34 },
		{ 380, 46, 34 },
		{ 380, 47, 34 },
		{ 380, 48, 34 },
		{ 380, 49, 34 },
		{ 380, 50, 34 },
		{ 380, 51, 34 },
		{ 380, 52, 34 },
		{ 380, 53, 34 },
		{ 380, 54, 34 },
		{ 380, 55, 34 },
		{ 380, 56, 34 },
		{ 380, 57, 34 },
		{ 380, 58, 34 },
		{ 380, 59, 34 },
		{ 380, 60, 34 },
		{ 381, 1, 34 },
		{ 381, 2, 34 },
		{ 381, 3, 34 },
		{ 381, 4, 34 },
		{ 381, 5, 34 },
		{ 381, 6, 34 },
		{ 381, 7, 34 },
		{ 381, 8, 34 },
		{ 381, 9, 34 },
		{ 381, 10, 34 },
		{ 381, 11, 34 },
		{ 381, 12, 34 },
		{ 381, 13, 34 },
		{ 381, 14, 34 },
		{ 381, 15, 34 },
		{ 381, 16, 34 },
		{ 381, 17, 34 },
		{ 381, 18, 34 },
		{ 381, 19, 34 },
		{ 381, 20, 34 },
		{ 381, 21, 34 },
		{ 381, 22, 34 },
		{ 381, 23, 34 },
		{ 381, 24, 34 },
		{ 381, 25, 34 },
		{ 381, 26, 34 },
		{ 381, 27, 34 },
		{ 381, 28, 34 },
		{ 381, 29, 34 },
		{ 381, 30, 34 },
		{ 381, 31, 34 },
		{ 381, 32, 34 },
		{ 381, 33, 34 },
		{ 381, 34, 34 },
		{ 381, 35, 34 },
		{ 381, 36, 34 },
		{ 381, 37, 34 },
		{ 381, 38, 34 },
		{ 381, 39, 34 },
		{ 381, 40, 34 },
		{ 381, 41, 34 },
		{ 381, 42, 34 },
		{ 381, 43, 34 },
		{ 381, 44, 34 },
		{ 381, 45, 34 },
		{ 381, 46, 34 },
		{ 381, 47, 34 },
		{ 381, 48, 34 },
		{ 381, 49, 34 },
		{ 381, 50, 34 },
		{ 381, 51, 34 },
		{ 381, 52, 34 },
		{ 381, 53, 34 },
		{ 381, 54, 34 },
		{ 381, 55, 34 },
		{ 381, 56, 34 },
		{ 381, 57, 34 },
		{ 381, 58, 34 },
		{ 382, 1, 34 },
		{ 382, 2, 34 },
		{ 382, 3, 34 },
		{ 382, 4, 34 },
		{ 382, 5, 34 },
		{ 382, 6, 34 },
		{ 382, 7, 34 },
		{ 382, 8, 34 },
		{ 382, 9, 34 },
		{ 382, 10, 34 },
		{ 382, 11, 34 },
		{ 382, 12, 34 },
		{ 382, 13, 34 },
		{ 382, 14, 34 },
		{ 382, 15, 34 },
		{ 382, 16, 34 },
		{ 382, 17, 34 },
		{ 382, 18, 34 },
		{ 382, 19, 34 },
		{ 382, 20, 34 },
		{ 382, 21, 34 },
		{ 382, 22, 34 },
		{ 382, 23, 34 },
		{ 382, 24, 34 },
		{ 382, 25, 34 },
		{ 382, 26, 34 },
		{ 382, 27, 34 },
		{ 382, 28, 34 },
		{ 382, 29, 34 },
		{ 382, 30, 34 },
		{ 382, 31, 34 },
		{ 382, 32, 34 },
		{ 382, 33, 34 },
		{ 382, 34, 34 },
		{ 382, 35, 34 },
		{ 382, 36, 34 },
		{ 382, 37, 34 },
		{ 382, 38, 34 },
		{ 382, 39, 34 },
		{ 382, 40, 34 },
		{ 382, 41, 34 },
		{ 382, 42, 34 },
		{ 382, 43, 34 },
		{ 382, 44, 34 },
		{ 382, 45, 34 },
		{ 382, 46, 34 },
		{ 382, 47, 34 },
		{ 382, 48, 34 },
		{ 382, 49, 34 },
		{ 382, 50, 34 },
		{ 382, 51, 34 },
		{ 382, 52, 34 },
		{ 382, 53, 34 },
		{ 382, 54, 34 },
		{ 382, 55, 34 },
		{ 382, 56, 34 },
		{ 383, 1, 34 },
		{ 383, 2, 34 },
		{ 383, 3, 34 },
		{ 383, 4, 34 },
		{ 383, 5, 34 },
		{ 383, 6, 34 },
		{ 383, 7, 34 },
		{ 383, 8, 34 },
		{ 383, 9, 34 },
		{ 383, 10, 34 },
		{ 383, 11, 34 },
		{ 383, 12, 34 },
		{ 383, 13, 34 },
		{ 383, 14, 34 },
		{ 383, 15, 34 },
		{ 383, 16, 34 },
		{ 383, 17, 34 },
		{ 383, 18, 34 },
		{ 383, 19, 34 },
		{ 383, 20, 34 },
		{ 383, 21, 34 },
		{ 383, 22, 34 },
		{ 383, 23, 34 },
		{ 383, 24, 34 },
		{ 383, 25, 34 },
		{ 383, 26, 34 },
		{ 383, 27, 34 },
		{ 383, 28, 34 },
		{ 383, 29, 34 },
		{ 383, 30, 34 },
		{ 383, 31, 34 },
		{ 383, 32, 34 },
		{ 383, 33, 34 },
		{ 383, 34, 34 },
		{ 383, 35, 34 },
		{ 383, 36, 34 },
		{ 383, 37, 34 },
		{ 383, 38, 34 },
		{ 383, 39, 34 },
		{ 383, 40, 34 },
		{ 383, 41, 34 },
		{ 383, 42, 34 },
		{ 383, 43, 34 },
		{ 383, 44, 34 },
		{ 383, 45, 34 },
		{ 383, 46, 34 },
		{ 383, 47, 34 },
		{ 383, 48, 34 },
		{ 383, 49, 34 },
		{ 383, 50, 34 },
		{ 383, 51, 34 },
		{ 383, 52, 34 },
		{ 383, 53, 34 },
		{ 383, 54, 34 },
		{ 383, 55, 34 },
		{ 384, 1, 34 },
		{ 384, 2, 34 },
		{ 384, 3, 34 },
		{ 384, 4, 34 },
		{ 384, 5, 34 },
		{ 384, 6, 34 },
		{ 384, 7, 34 },
		{ 384, 8, 34 },
		{ 384, 9, 34 },
		{ 384, 10, 34 },
		{ 384, 11, 34 },
		{ 384, 12, 34 },
		{ 384, 13, 34 },
		{ 384, 14, 34 },
		{ 384, 15, 34 },
		{ 384, 16, 34 },
		{ 384, 17, 34 },
		{ 384, 18, 34 },
		{ 384, 19, 34 },
		{ 384, 20, 34 },
		{ 384, 21, 34 },
		{ 384, 22, 34 },
		{ 384, 23, 34 },
		{ 384, 24, 34 },
		{ 384, 25, 34 },
		{ 384, 26, 34 },
		{ 384, 27, 34 },
		{ 384, 28, 34 },
		{ 384, 29, 34 },
		{ 384, 30, 34 },
		{ 384, 31, 34 },
		{ 384, 32, 34 },
		{ 384, 33, 34 },
		{ 384, 34, 34 },
		{ 384, 35, 34 },
		{ 384, 36, 34 },
		{ 384, 37, 34 },
		{ 384, 38, 34 },
		{ 384, 39, 34 },
		{ 384, 40, 34 },
		{ 384, 41, 34 },
		{ 384, 42, 34 },
		{ 384, 43, 34 },
		{ 384, 44, 34 },
		{ 384, 45, 34 },
		{ 384, 46, 34 },
		{ 384, 47, 34 },
		{ 384, 48, 34 },
		{ 384, 49, 34 },
		{ 384, 50, 34 },
		{ 384, 51, 34 },
		{ 384, 52, 34 },
		{ 384, 53, 34 },
		{ 385, 1, 34 },
		{ 385, 2, 34 },
		{ 385, 3, 34 },
		{ 385, 4, 34 },
		{ 385, 5, 34 },
		{ 385, 6, 34 },
		{ 385, 7, 34 },
		{ 385, 8, 34 },
		{ 385, 9, 34 },
		{ 385, 10, 34 },
		{ 385, 11, 34 },
		{ 385, 12, 34 },
		{ 385, 13, 34 },
		{ 385, 14, 34 },
		{ 385, 15, 34 },
		{ 385, 16, 34 },
		{ 385, 17, 34 },
		{ 385, 18, 34 },
		{ 385, 19, 34 },
		{ 385, 20, 34 },
		{ 385, 21, 34 },
		{ 385, 22, 34 },
		{ 385, 23, 34 },
		{ 385, 24, 34 },
		{ 385, 25, 34 },
		{ 385, 26, 34 },
		{ 385, 27, 34 },
		{ 385, 28, 34 },
		{ 385, 29, 34 },
		{ 385, 30, 34 },
		{ 385, 31, 34 },
		{ 385, 32, 34 },
		{ 385, 33, 34 },
		{ 385, 34, 34 },
		{ 385, 35, 34 },
		{ 385, 36, 34 },
		{ 385, 37, 34 },
		{ 385, 38, 34 },
		{ 385, 39, 34 },
		{ 385, 40, 34 },
		{ 385, 41, 34 },
		{ 385, 42, 34 },
		{ 385, 43, 34 },
		{ 385, 44, 34 },
		{ 385, 45, 34 },
		{ 385, 46, 34 },
		{ 385, 47, 34 },
		{ 385, 48, 34 },
		{ 385, 49, 34 },
		{ 385, 50, 34 },
		{ 385, 51, 34 },
		{ 385, 52, 34 },
		{ 386, 1, 34 },
		{ 386, 2, 34 },
		{ 386, 3, 34 },
		{ 386, 4, 34 },
		{ 386, 5, 34 },
		{ 386, 6, 34 },
		{ 386, 7, 34 },
		{ 386, 8, 34 },
		{ 386, 9, 34 },
		{ 386, 10, 34 },
		{ 386, 11, 34 },
		{ 386, 12, 34 },
		{ 386, 13, 34 },
		{ 386, 14, 34 },
		{ 386, 15, 34 },
		{ 386, 16, 34 },
		{ 386, 17, 34 },
		{ 386, 18, 34 },
		{ 386, 19, 34 },
		{ 386, 20, 34 },
		{ 386, 21, 34 },
		{ 386, 22, 34 },
		{ 386, 23, 34 },
		{ 386, 24, 34 },
		{ 386, 25, 34 },
		{ 386, 26, 34 },
		{ 386, 27, 34 },
		{ 386, 28, 34 },
		{ 386, 29, 34 },
		{ 386, 30, 34 },
		{ 386, 31, 34 },
		{ 386, 32, 34 },
		{ 386, 33, 34 },
		{ 386, 34, 34 },
		{ 386, 35, 34 },
		{ 386, 36, 34 },
		{ 386, 37, 34 },
		{ 386, 38, 34 },
		{ 386, 39, 34 },
		{ 386, 40, 34 },
		{ 386, 41, 34 },
		{ 386, 42, 34 },
		{ 386, 43, 34 },
		{ 386, 44, 34 },
		{ 386, 45, 34 },
		{ 386, 46, 34 },
		{ 386, 47, 34 },
		{ 386, 48, 34 },
		{ 386, 49, 34 },
		{ 386, 50, 34 },
		{ 387, 1, 34 },
		{ 387, 2, 34 },
		{ 387, 3, 34 },
		{ 387, 4, 34 },
		{ 387, 5, 34 },
		{ 387, 6, 34 },
		{ 387, 7, 34 },
		{ 387, 8, 34 },
		{ 387, 9, 34 },
		{ 387, 10, 34 },
		{ 387, 11, 34 },
		{ 387, 12, 34 },
		{ 387, 13, 34 },
		{ 387, 14, 34 },
		{ 387, 15, 34 },
		{ 387, 16, 34 },
		{ 387, 17, 34 },
		{ 387, 18, 34 },
		{ 387, 19, 34 },
		{ 387, 20, 34 },
		{ 387, 21, 34 },
		{ 387, 22, 34 },
		{ 387, 23, 34 },
		{ 387, 24, 34 },
		{ 387, 25, 34 },
		{ 387, 26, 34 },
		{ 387, 27, 34 },
		{ 387, 28, 34 },
		{ 387, 29, 34 },
		{ 387, 30, 34 },
		{ 387, 31, 34 },
		{ 387, 32, 34 },
		{ 387, 33, 34 },
		{ 387, 34, 34 },
		{ 387, 35, 34 },
		{ 387, 36, 34 },
		{ 387, 37, 34 },
		{ 387, 38, 34 },
		{ 387, 39, 34 },
		{ 387, 40, 34 },
		{ 387, 41, 34 },
		{ 387, 42, 34 },
		{ 387, 43, 34 },
		{ 387, 44, 34 },
		{ 387, 45, 34 },
		{ 387, 46, 34 },
		{ 387, 47, 34 },
		{ 387, 48, 34 },
		{ 388, 1, 34 },
		{ 388, 2, 34 },
		{ 388, 3, 34 },
		{ 388, 4, 34 },
		{ 388, 5, 34 },
		{ 388, 6, 34 },
		{ 388, 7, 34 },
		{ 388, 8, 34 },
		{ 388, 9, 34 },
		{ 388, 10, 34 },
		{ 388, 11, 34 },
		{ 388, 12, 34 },
		{ 388, 13, 34 },
		{ 388, 14, 34 },
		{ 388, 15, 34 },
		{ 388, 16, 34 },
		{ 388, 17, 34 },
		{ 388, 18, 34 },
		{ 388, 19, 34 },
		{ 388, 20, 34 },
		{ 388, 21, 34 },
		{ 388, 22, 34 },
		{ 388, 23, 34 },
		{ 388, 24, 34 },
		{ 388, 25, 34 },
		{ 388, 26, 34 },
		{ 388, 27, 34 },
		{ 388, 28, 34 },
		{ 388, 29, 34 },
		{ 388, 30, 34 },
		{ 388, 31, 34 },
		{ 388, 32, 34 },
		{ 388, 33, 34 },
		{ 388, 34, 34 },
		{ 388, 35, 34 },
		{ 388, 36, 34 },
		{ 388, 37, 34 },
		{ 388, 38, 34 },
		{ 388, 39, 34 },
		{ 388, 40, 34 },
		{ 388, 41, 34 },
		{ 388, 42, 34 },
		{ 388, 43, 34 },
		{ 388, 44, 34 },
		{ 388, 45, 34 },
		{ 388, 46, 34 },
		{ 388, 47, 34 },
		{ 389, 1, 34 },
		{ 389, 2, 34 },
		{ 389, 3, 34 },
		{ 389, 4, 34 },
		{ 389, 5, 34 },
		{ 389, 6, 34 },
		{ 389, 7, 34 },
		{ 389, 8, 34 },
		{ 389, 9, 34 },
		{ 389, 10, 34 },
		{ 389, 11, 34 },
		{ 389, 12, 34 },
		{ 389, 13, 34 },
		{ 389, 14, 34 },
		{ 389, 15, 34 },
		{ 389, 16, 34 },
		{ 389, 17, 34 },
		{ 389, 18, 34 },
		{ 389, 19, 34 },
		{ 389, 20, 34 },
		{ 389, 21, 34 },
		{ 389, 22, 34 },
		{ 389, 23, 34 },
		{ 389, 24, 34 },
		{ 389, 25, 34 },
		{ 389, 26, 34 },
		{ 389, 27, 34 },
		{ 389, 28, 34 },
		{ 389, 29, 34 },
		{ 389, 30, 34 },
		{ 389, 31, 34 },
		{ 389, 32, 34 },
		{ 389, 33, 34 },
		{ 389, 34, 34 },
		{ 389, 35, 34 },
		{ 389, 36, 34 },
		{ 389, 37, 34 },
		{ 389, 38, 34 },
		{ 389, 39, 34 },
		{ 389, 40, 34 },
		{ 389, 41, 34 },
		{ 389, 42, 34 },
		{ 389, 43, 34 },
		{ 389, 44, 34 },
		{ 389, 45, 34 },
		{ 390, 1, 34 },
		{ 390, 2, 34 },
		{ 390, 3, 34 },
		{ 390, 4, 34 },
		{ 390, 5, 34 },
		{ 390, 6, 34 },
		{ 390, 7, 34 },
		{ 390, 8, 34 },
		{ 390, 9, 34 },
		{ 390, 10, 34 },
		{ 390, 11, 34 },
		{ 390, 12, 34 },
		{ 390, 13, 34 },
		{ 390, 14, 34 },
		{ 390, 15, 34 },
		{ 390, 16, 34 },
		{ 390, 17, 34 },
		{ 390, 18, 34 },
		{ 390, 19, 34 },
		{ 390, 20, 34 },
		{ 390, 21, 34 },
		{ 390, 22, 34 },
		{ 390, 23, 34 },
		{ 390, 24, 34 },
		{ 390, 25, 34 },
		{ 390, 26, 34 },
		{ 390, 27, 34 },
		{ 390, 28, 34 },
		{ 390, 29, 34 },
		{ 390, 30, 34 },
		{ 390, 31, 34 },
		{ 390, 32, 34 },
		{ 390, 33, 34 },
		{ 390, 34, 34 },
		{ 390, 35, 34 },
		{ 390, 36, 34 },
		{ 390, 37, 34 },
		{ 390, 38, 34 },
		{ 390, 39, 34 },
		{ 390, 40, 34 },
		{ 390, 41, 34 },
		{ 390, 42, 34 },
		{ 390, 43, 34 },
		{ 390, 44, 34 },
		{ 391, 1, 34 },
		{ 391, 2, 34 },
		{ 391, 3, 34 },
		{ 391, 4, 34 },
		{ 391, 5, 34 },
		{ 391, 6, 34 },
		{ 391, 7, 34 },
		{ 391, 8, 34 },
		{ 391, 9, 34 },
		{ 391, 10, 34 },
		{ 391, 11, 34 },
		{ 391, 12, 34 },
		{ 391, 13, 34 },
		{ 391, 14, 34 },
		{ 391, 15, 34 },
		{ 391, 16, 34 },
		{ 391, 17, 34 },
		{ 391, 18, 34 },
		{ 391, 19, 34 },
		{ 391, 20, 34 },
		{ 391, 21, 34 },
		{ 391, 22, 34 },
		{ 391, 23, 34 },
		{ 391, 24, 34 },
		{ 391, 25, 34 },
		{ 391, 26, 34 },
		{ 391, 27, 34 },
		{ 391, 28, 34 },
		{ 391, 29, 34 },
		{ 391, 30, 34 },
		{ 391, 31, 34 },
		{ 391, 32, 34 },
		{ 391, 33, 34 },
		{ 391, 34, 34 },
		{ 391, 35, 34 },
		{ 391, 36, 34 },
		{ 391, 37, 34 },
		{ 391, 38, 34 },
		{ 391, 39, 34 },
		{ 391, 40, 34 },
		{ 391, 41, 34 },
		{ 391, 42, 34 },
		{ 392, 1, 34 },
		{ 392, 2, 34 },
		{ 392, 3, 34 },
		{ 392, 4, 34 },
		{ 392, 5, 34 },
		{ 392, 6, 34 },
		{ 392, 7, 34 },
		{ 392, 8, 34 },
		{ 392, 9, 34 },
		{ 392, 10, 34 },
		{ 392, 11, 34 },
		{ 392, 12, 34 },
		{ 392, 13, 34 },
		{ 392, 14, 34 },
		{ 392, 15, 34 },
		{ 392, 16, 34 },
		{ 392, 17, 34 },
		{ 392, 18, 34 },
		{ 392, 19, 34 },
		{ 392, 20, 34 },
		{ 392, 21, 34 },
		{ 392, 22, 34 },
		{ 392, 23, 34 },
		{ 392, 24, 34 },
		{ 392, 25, 34 },
		{ 392, 26, 34 },
		{ 392, 27, 34 },
		{ 392, 28, 34 },
		{ 392, 29, 34 },
		{ 392, 30, 34 },
		{ 392, 31, 34 },
		{ 392, 32, 34 },
		{ 392, 33, 34 },
		{ 392, 34, 34 },
		{ 392, 35, 34 },
		{ 392, 36, 34 },
		{ 392, 37, 34 },
		{ 392, 38, 34 },
		{ 392, 39, 34 },
		{ 392, 40, 34 },
		{ 393, 1, 34 },
		{ 393, 2, 34 },
		{ 393, 3, 34 },
		{ 393, 4, 34 },
		{ 393, 5, 34 },
		{ 393, 6, 34 },
		{ 393, 7, 34 },
		{ 393, 8, 34 },
		{ 393, 9, 34 },
		{ 393, 10, 34 },
		{ 393, 11, 34 },
		{ 393, 12, 34 },
		{ 393, 13, 34 },
		{ 393, 14, 34 },
		{ 393, 15, 34 },
		{ 393, 16, 34 },
		{ 393, 17, 34 },
		{ 393, 18, 34 },
		{ 393, 19, 34 },
		{ 393, 20, 34 },
		{ 393, 21, 34 },
		{ 393, 22, 34 },
		{ 393, 23, 34 },
		{ 393, 24, 34 },
		{ 393, 25, 34 },
		{ 393, 26, 34 },
		{ 393, 27, 34 },
		{ 393, 28, 34 },
		{ 393, 29, 34 },
		{ 393, 30, 34 },
		{ 393, 31, 34 },
		{ 393, 32, 34 },
		{ 393, 33, 34 },
		{ 393, 34, 34 },
		{ 393, 35, 34 },
		{ 393, 36, 34 },
		{ 393, 37, 34 },
		{ 393, 38, 34 },
		{ 393, 39, 34 },
		{ 394, 1, 34 },
		{ 394, 2, 34 },
		{ 394, 3, 34 },
		{ 394, 4, 34 },
		{ 394, 5, 34 },
		{ 394, 6, 34 },
		{ 394, 7, 34 },
		{ 394, 8, 34 },
		{ 394, 9, 34 },
		{ 394, 10, 34 },
		{ 394, 11, 34 },
		{ 394, 12, 34 },
		{ 394, 13, 34 },
		{ 394, 14, 34 },
		{ 394, 15, 34 },
		{ 394, 16, 34 },
		{ 394, 17, 34 },
		{ 394, 18, 34 },
		{ 394, 19, 34 },
		{ 394, 20, 34 },
		{ 394, 21, 34 },
		{ 394, 22, 34 },
		{ 394, 23, 34 },
		{ 394, 24, 34 },
		{ 394, 25, 34 },
		{ 394, 26, 34 },
		{ 394, 27, 34 },
		{ 394, 28, 34 },
		{ 394, 29, 34 },
		{ 394, 30, 34 },
		{ 394, 31, 34 },
		{ 394, 32, 34 },
		{ 394, 33, 34 },
		{ 394, 34, 34 },
		{ 394, 35, 34 },
		{ 394, 36, 34 },
		{ 394, 37, 34 },
		{ 395, 1, 34 },
		{ 395, 2, 34 },
		{ 395, 3, 34 },
		{ 395, 4, 34 },
		{ 395, 5, 34 },
		{ 395, 6, 34 },
		{ 395, 7, 34 },
		{ 395, 8, 34 },
		{ 395, 9, 34 },
		{ 395, 10, 34 },
		{ 395, 11, 34 },
		{ 395, 12, 34 },
		{ 395, 13, 34 },
		{ 395, 14, 34 },
		{ 395, 15, 34 },
		{ 395, 16, 34 },
		{ 395, 17, 34 },
		{ 395, 18, 34 },
		{ 395, 19, 34 },
		{ 395, 20, 34 },
		{ 395, 21, 34 },
		{ 395, 22, 34 },
		{ 395, 23, 34 },
		{ 395, 24, 34 },
		{ 395, 25, 34 },
		{ 395, 26, 34 },
		{ 395, 27, 34 },
		{ 395, 28, 34 },
		{ 395, 29, 34 },
		{ 395, 30, 34 },
		{ 395, 31, 34 },
		{ 395, 32, 34 },
		{ 395, 33, 34 },
		{ 395, 34, 34 },
		{ 395, 35, 34 },
		{ 395, 36, 34 },
		{ 396, 1, 34 },
		{ 396, 2, 34 },
		{ 396, 3, 34 },
		{ 396, 4, 34 },
		{ 396, 5, 34 },
		{ 396, 6, 34 },
		{ 396, 7, 34 },
		{ 396, 8, 34 },
		{ 396, 9, 34 },
		{ 396, 10, 34 },
		{ 396, 11, 34 },
		{ 396, 12, 34 },
		{ 396, 13, 34 },
		{ 396, 14, 34 },
		{ 396, 15, 34 },
		{ 396, 16, 34 },
		{ 396, 17, 34 },
		{ 396, 18, 34 },
		{ 396, 19, 34 },
		{ 396, 20, 34 },
		{ 396, 21, 34 },
		{ 396, 22, 34 },
		{ 396, 23, 34 },
		{ 396, 24, 34 },
		{ 396, 25, 34 },
		{ 396, 26, 34 },
		{ 396, 27, 34 },
		{ 396, 28, 34 },
		{ 396, 29, 34 },
		{ 396, 30, 34 },
		{ 396, 31, 34 },
		{ 396, 32, 34 },
		{ 396, 33, 34 },
		{ 396, 34, 34 },
		{ 397, 1, 34 },
		{ 397, 2, 34 },
		{ 397, 3, 34 },
		{ 397, 4, 34 },
		{ 397, 5, 34 },
		{ 397, 6, 34 },
		{ 397, 7, 34 },
		{ 397, 8, 34 },
		{ 397, 9, 34 },
		{ 397, 10, 34 },
		{ 397, 11, 34 },
		{ 397, 12, 34 },
		{ 397, 13, 34 },
		{ 397, 14, 34 },
		{ 397, 15, 34 },
		{ 397, 16, 34 },
		{ 397, 17, 34 },
		{ 397, 18, 34 },
		{ 397, 19, 34 },
		{ 397, 20, 34 },
		{ 397, 21, 34 },
		{ 397, 22, 34 },
		{ 397, 23, 34 },
		{ 397, 24, 34 },
		{ 397, 25, 34 },
		{ 397, 26, 34 },
		{ 397, 27, 34 },
		{ 397, 28, 34 },
		{ 397, 29, 34 },
		{ 397, 30, 34 },
		{ 397, 31, 34 },
		{ 397, 32, 34 },
		{ 398, 1, 34 },
		{ 398, 2, 34 },
		{ 398, 3, 34 },
		{ 398, 4, 34 },
		{ 398, 5, 34 },
		{ 398, 6, 34 },
		{ 398, 7, 34 },
		{ 398, 8, 34 },
		{ 398, 9, 34 },
		{ 398, 10, 34 },
		{ 398, 11, 34 },
		{ 398, 12, 34 },
		{ 398, 13, 34 },
		{ 398, 14, 34 },
		{ 398, 15, 34 },
		{ 398, 16, 34 },
		{ 398, 17, 34 },
		{ 398, 18, 34 },
		{ 398, 19, 34 },
		{ 398, 20, 34 },
		{ 398, 21, 34 },
		{ 398, 22, 34 },
		{ 398, 23, 34 },
		{ 398, 24, 34 },
		{ 398, 25, 34 },
		{ 398, 26, 34 },
		{ 398, 27, 34 },
		{ 398, 28, 34 },
		{ 398, 29, 34 },
		{ 398, 30, 34 },
		{ 398, 31, 34 },
		{ 399, 1, 34 },
		{ 399, 2, 34 },
		{ 399, 3, 34 },
		{ 399, 4, 34 },
		{ 399, 5, 34 },
		{ 399, 6, 34 },
		{ 399, 7, 34 },
		{ 399, 8, 34 },
		{ 399, 9, 34 },
		{ 399, 10, 34 },
		{ 399, 11, 34 },
		{ 399, 12, 34 },
		{ 399, 13, 34 },
		{ 399, 14, 34 },
		{ 399, 15, 34 },
		{ 399, 16, 34 },
		{ 399, 17, 34 },
		{ 399, 18, 34 },
		{ 399, 19, 34 },
		{ 399, 20, 34 },
		{ 399, 21, 34 },
		{ 399, 22, 34 },
		{ 399, 23, 34 },
		{ 399, 24, 34 },
		{ 399, 25, 34 },
		{ 399, 26, 34 },
		{ 399, 27, 34 },
		{ 399, 28, 34 },
		{ 399, 29, 34 },
		{ 400, 1, 34 },
		{ 400, 2, 34 },
		{ 400, 3, 34 },
		{ 400, 4, 34 },
		{ 400, 5, 34 },
		{ 400, 6, 34 },
		{ 400, 7, 34 },
		{ 400, 8, 34 },
		{ 400, 9, 34 },
		{ 400, 10, 34 },
		{ 400, 11, 34 },
		{ 400, 12, 34 },
		{ 400, 13, 34 },
		{ 400, 14, 34 },
		{ 400, 15, 34 },
		{ 400, 16, 34 },
		{ 400, 17, 34 },
		{ 400, 18, 34 },
		{ 400, 19, 34 },
		{ 400, 20, 34 },
		{ 400, 21, 34 },
		{ 400, 22, 34 },
		{ 400, 23, 34 },
		{ 400, 24, 34 },
		{ 400, 25, 34 },
		{ 400, 26, 34 },
		{ 400, 27, 34 },
		{ 400, 28, 34 },
		{ 401, 1, 34 },
		{ 401, 2, 34 },
		{ 401, 3, 34 },
		{ 401, 4, 34 },
		{ 401, 5, 34 },
		{ 401, 6, 34 },
		{ 401, 7, 34 },
		{ 401, 8, 34 },
		{ 401, 9, 34 },
		{ 401, 10, 34 },
		{ 401, 11, 34 },
		{ 401, 12, 34 },
		{ 401, 13, 34 },
		{ 401, 14, 34 },
		{ 401, 15, 34 },
		{ 401, 16, 34 },
		{ 401, 17, 34 },
		{ 401, 18, 34 },
		{ 401, 19, 34 },
		{ 401, 20, 34 },
		{ 401, 21, 34 },
		{ 401, 22, 34 },
		{ 401, 23, 34 },
		{ 401, 24, 34 },
		{ 401, 25, 34 },
		{ 401, 26, 34 },
		{ 402, 1, 34 },
		{ 402, 2, 34 },
		{ 402, 3, 34 },
		{ 402, 4, 34 },
		{ 402, 5, 34 },
		{ 402, 6, 34 },
		{ 402, 7, 34 },
		{ 402, 8, 34 },
		{ 402, 9, 34 },
		{ 402, 10, 34 },
		{ 402, 11, 34 },
		{ 402, 12, 34 },
		{ 402, 13, 34 },
		{ 402, 14, 34 },
		{ 402, 15, 34 },
		{ 402, 16, 34 },
		{ 402, 17, 34 },
		{ 402, 18, 34 },
		{ 402, 19, 34 },
		{ 402, 20, 34 },
		{ 402, 21, 34 },
		{ 402, 22, 34 },
		{ 402, 23, 34 },
		{ 402, 24, 34 },
		{ 403, 1, 34 },
		{ 403, 2, 34 },
		{ 403, 3, 34 },
		{ 403, 4, 34 },
		{ 403, 5, 34 },
		{ 403, 6, 34 },
		{ 403, 7, 34 },
		{ 403, 8, 34 },
		{ 403, 9, 34 },
		{ 403, 10, 34 },
		{ 403, 11, 34 },
		{ 403, 12, 34 },
		{ 403, 13, 34 },
		{ 403, 14, 34 },
		{ 403, 15, 34 },
		{ 403, 16, 34 },
		{ 403, 17, 34 },
		{ 403, 18, 34 },
		{ 403, 19, 34 },
		{ 403, 20, 34 },
		{ 403, 21, 34 },
		{ 403, 22, 34 },
		{ 403, 23, 34 },
		{ 404, 1, 34 },
		{ 404, 2, 34 },
		{ 404, 3, 34 },
		{ 404, 4, 34 },
		{ 404, 5, 34 },
		{ 404, 6, 34 },
		{ 404, 7, 34 },
		{ 404, 8, 34 },
		{ 404, 9, 34 },
		{ 404, 10, 34 },
		{ 404, 11, 34 },
		{ 404, 12, 34 },
		{ 404, 13, 34 },
		{ 404, 14, 34 },
		{ 404, 15, 34 },
		{ 404, 16, 34 },
		{ 404, 17, 34 },
		{ 404, 18, 34 },
		{ 404, 19, 34 },
		{ 404, 20, 34 },
		{ 404, 21, 34 },
		{ 405, 1, 34 },
		{ 405, 2, 34 },
		{ 405, 3, 34 },
		{ 405, 4, 34 },
		{ 405, 5, 34 },
		{ 405, 6, 34 },
		{ 405, 7, 34 },
		{ 405, 8, 34 },
		{ 405, 9, 34 },
		{ 405, 10, 34 },
		{ 405, 11, 34 },
		{ 405, 12, 34 },
		{ 405, 13, 34 },
		{ 405, 14, 34 },
		{ 405, 15, 34 },
		{ 405, 16, 34 },
		{ 405, 17, 34 },
		{ 405, 18, 34 },
		{ 405, 19, 34 },
		{ 405, 20, 34 },
		{ 406, 1, 34 },
		{ 406, 2, 34 },
		{ 406, 3, 34 },
		{ 406, 4, 34 },
		{ 406, 5, 34 },
		{ 406, 6, 34 },
		{ 406, 7, 34 },
		{ 406, 8, 34 },
		{ 406, 9, 34 },
		{ 406, 10, 34 },
		{ 406, 11, 34 },
		{ 406, 12, 34 },
		{ 406, 13, 34 },
		{ 406, 14, 34 },
		{ 406, 15, 34 },
		{ 406, 16, 34 },
		{ 406, 17, 34 },
		{ 406, 18, 34 },
		{ 407, 1, 34 },
		{ 407, 2, 34 },
		{ 407, 3, 34 },
		{ 407, 4, 34 },
		{ 407, 5, 34 },
		{ 407, 6, 34 },
		{ 407, 7, 34 },
		{ 407, 8, 34 },
		{ 407, 9, 34 },
		{ 407, 10, 34 },
		{ 407, 11, 34 },
		{ 407, 12, 34 },
		{ 407, 13, 34 },
		{ 407, 14, 34 },
		{ 407, 15, 34 },
		{ 407, 16, 34 },
		{ 408, 1, 34 },
		{ 408, 2, 34 },
		{ 408, 3, 34 },
		{ 408, 4, 34 },
		{ 408, 5, 34 },
		{ 408, 6, 34 },
		{ 408, 7, 34 },
		{ 408, 8, 34 },
		{ 408, 9, 34 },
		{ 408, 10, 34 },
		{ 408, 11, 34 },
		{ 408, 12, 34 },
		{ 408, 13, 34 },
		{ 408, 14, 34 },
		{ 408, 15, 34 },
		{ 409, 1, 34 },
		{ 409, 2, 34 },
		{ 409, 3, 34 },
		{ 409, 4, 34 },
		{ 409, 5, 34 },
		{ 409, 6, 34 },
		{ 409, 7, 34 },
		{ 409, 8, 34 },
		{ 409, 9, 34 },
		{ 409, 10, 34 },
		{ 409, 11, 34 },
		{ 409, 12, 34 },
		{ 409, 13, 34 },
		{ 410, 1, 34 },
		{ 410, 2, 34 },
		{ 410, 3, 34 },
		{ 410, 4, 34 },
		{ 410, 5, 34 },
		{ 410, 6, 34 },
		{ 410, 7, 34 },
		{ 410, 8, 34 },
		{ 410, 9, 34 },
		{ 410, 10, 34 },
		{ 410, 11, 34 },
		{ 410, 12, 34 },
		{ 411, 1, 34 },
		{ 411, 2, 34 },
		{ 411, 3, 34 },
		{ 411, 4, 34 },
		{ 411, 5, 34 },
		{ 411, 6, 34 },
		{ 411, 7, 34 },
		{ 411, 8, 34 },
		{ 411, 9, 34 },
		{ 411, 10, 34 },
		{ 412, 1, 34 },
		{ 412, 2, 34 },
		{ 412, 3, 34 },
		{ 412, 4, 34 },
		{ 412, 5, 34 },
		{ 412, 6, 34 },
		{ 412, 7, 34 },
		{ 412, 8, 34 },
		{ 413, 1, 34 },
		{ 413, 2, 34 },
		{ 413, 3, 34 },
		{ 413, 4, 34 },
		{ 413, 5, 34 },
		{ 413, 6, 34 },
		{ 413, 7, 34 },
		{ 414, 1, 34 },
		{ 414, 2, 34 },
		{ 414, 3, 34 },
		{ 414, 4, 34 },
		{ 414, 5, 34 },
		{ 415, 1, 34 },
		{ 415, 2, 34 },
		{ 415, 3, 34 },
		{ 415, 4, 34 },
		{ 416, 1, 34 },
		{ 416, 2, 34 },
		{ 66, 512, 35 },
		{ 67, 511, 35 },
		{ 67, 512, 35 },
		{ 68, 510, 35 },
		{ 68, 511, 35 },
		{ 68, 512, 35 },
		{ 69, 508, 35 },
		{ 69, 509, 35 },
		{ 69, 510, 35 },
		{ 69, 511, 35 },
		{ 69, 512, 35 },
		{ 70, 507, 35 },
		{ 70, 508, 35 },
		{ 70, 509, 35 },
		{ 70, 510, 35 },
		{ 70, 511, 35 },
		{ 70, 512, 35 },
		{ 71, 506, 35 },
		{ 71, 507, 35 },
		{ 71, 508, 35 },
		{ 71, 509, 35 },
		{ 71, 510, 35 },
		{ 71, 511, 35 },
		{ 71, 512, 35 },
		{ 72, 504, 35 },
		{ 72, 505, 35 },
		{ 72, 506, 35 },
		{ 72, 507, 35 },
		{ 72, 508, 35 },
		{ 72, 509, 35 },
		{ 72, 510, 35 },
		{ 72, 511, 35 },
		{ 72, 512, 35 },
		{ 73, 503, 35 },
		{ 73, 504, 35 },
		{ 73, 505, 35 },
		{ 73, 506, 35 },
		{ 73, 507, 35 },
		{ 73, 508, 35 },
		{ 73, 509, 35 },
		{ 73, 510, 35 },
		{ 73, 511, 35 },
		{ 73, 512, 35 },
		{ 74, 502, 35 },
		{ 74, 503, 35 },
		{ 74, 504, 35 },
		{ 74, 505, 35 },
		{ 74, 506, 35 },
		{ 74, 507, 35 },
		{ 74, 508, 35 },
		{ 74, 509, 35 },
		{ 74, 510, 35 },
		{ 74, 511, 35 },
		{ 74, 512, 35 },
		{ 75, 500, 35 },
		{ 75, 501, 35 },
		{ 75, 502, 35 },
		{ 75, 503, 35 },
		{ 75, 504, 35 },
		{ 75, 505, 35 },
		{ 75, 506, 35 },
		{ 75, 507, 35 },
		{ 75, 508, 35 },
		{ 75, 509, 35 },
		{ 75, 510, 35 },
		{ 75, 511, 35 },
		{ 75, 512, 35 },
		{ 76, 499, 35 },
		{ 76, 500, 35 },
		{ 76, 501, 35 },
		{ 76, 502, 35 },
		{ 76, 503, 35 },
		{ 76, 504, 35 },
		{ 76, 505, 35 },
		{ 76, 506, 35 },
		{ 76, 507, 35 },
		{ 76, 508, 35 },
		{ 76, 509, 35 },
		{ 76, 510, 35 },
		{ 76, 511, 35 },
		{ 76, 512, 35 },
		{ 77, 498, 35 },
		{ 77, 499, 35 },
		{ 77, 500, 35 },
		{ 77, 501, 35 },
		{ 77, 502, 35 },
		{ 77, 503, 35 },
		{ 77, 504, 35 },
		{ 77, 505, 35 },
		{ 77, 506, 35 },
		{ 77, 507, 35 },
		{ 77, 508, 35 },
		{ 77, 509, 35 },
		{ 77, 510, 35 },
		{ 77, 511, 35 },
		{ 77, 512, 35 },
		{ 78, 496, 35 },
		{ 78, 497, 35 },
		{ 78, 498, 35 },
		{ 78, 499, 35 },
		{ 78, 500, 35 },
		{ 78, 501, 35 },
		{ 78, 502, 35 },
		{ 78, 503, 35 },
		{ 78, 504, 35 },
		{ 78, 505, 35 },
		{ 78, 506, 35 },
		{ 78, 507, 35 },
		{ 78, 508, 35 },
		{ 78, 509, 35 },
		{ 78, 510, 35 },
		{ 78, 511, 35 },
		{ 78, 512, 35 },
		{ 79, 495, 35 },
		{ 79, 496, 35 },
		{ 79, 497, 35 },
		{ 79, 498, 35 },
		{ 79, 499, 35 },
		{ 79, 500, 35 },
		{ 79, 501, 35 },
		{ 79, 502, 35 },
		{ 79, 503, 35 },
		{ 79, 504, 35 },
		{ 79, 505, 35 },
		{ 79, 506, 35 },
		{ 79, 507, 35 },
		{ 79, 508, 35 },
		{ 79, 509, 35 },
		{ 79, 510, 35 },
		{ 79, 511, 35 },
		{ 79, 512, 35 },
		{ 80, 494, 35 },
		{ 80, 495, 35 },
		{ 80, 496, 35 },
		{ 80, 497, 35 },
		{ 80, 498, 35 },
		{ 80, 499, 35 },
		{ 80, 500, 35 },
		{ 80, 501, 35 },
		{ 80, 502, 35 },
		{ 80, 503, 35 },
		{ 80, 504, 35 },
		{ 80, 505, 35 },
		{ 80, 506, 35 },
		{ 80, 507, 35 },
		{ 80, 508, 35 },
		{ 80, 509, 35 },
		{ 80, 510, 35 },
		{ 80, 511, 35 },
		{ 80, 512, 35 },
		{ 81, 492, 35 },
		{ 81, 493, 35 },
		{ 81, 494, 35 },
		{ 81, 495, 35 },
		{ 81, 496, 35 },
		{ 81, 497, 35 },
		{ 81, 498, 35 },
		{ 81, 499, 35 },
		{ 81, 500, 35 },
		{ 81, 501, 35 },
		{ 81, 502, 35 },
		{ 81, 503, 35 },
		{ 81, 504, 35 },
		{ 81, 505, 35 },
		{ 81, 506, 35 },
		{ 81, 507, 35 },
		{ 81, 508, 35 },
		{ 81, 509, 35 },
		{ 81, 510, 35 },
		{ 81, 511, 35 },
		{ 81, 512, 35 },
		{ 82, 491, 35 },
		{ 82, 492, 35 },
		{ 82, 493, 35 },
		{ 82, 494, 35 },
		{ 82, 495, 35 },
		{ 82, 496, 35 },
		{ 82, 497, 35 },
		{ 82, 498, 35 },
		{ 82, 499, 35 },
		{ 82, 500, 35 },
		{ 82, 501, 35 },
		{ 82, 502, 35 },
		{ 82, 503, 35 },
		{ 82, 504, 35 },
		{ 82, 505, 35 },
		{ 82, 506, 35 },
		{ 82, 507, 35 },
		{ 82, 508, 35 },
		{ 82, 509, 35 },
		{ 82, 510, 35 },
		{ 82, 511, 35 },
		{ 82, 512, 35 },
		{ 83, 490, 35 },
		{ 83, 491, 35 },
		{ 83, 492, 35 },
		{ 83, 493, 35 },
		{ 83, 494, 35 },
		{ 83, 495, 35 },
		{ 83, 496, 35 },
		{ 83, 497, 35 },
		{ 83, 498, 35 },
		{ 83, 499, 35 },
		{ 83, 500, 35 },
		{ 83, 501, 35 },
		{ 83, 502, 35 },
		{ 83, 503, 35 },
		{ 83, 504, 35 },
		{ 83, 505, 35 },
		{ 83, 506, 35 },
		{ 83, 507, 35 },
		{ 83, 508, 35 },
		{ 83, 509, 35 },
		{ 83, 510, 35 },
		{ 83, 511, 35 },
		{ 83, 512, 35 },
		{ 84, 488, 35 },
		{ 84, 489, 35 },
		{ 84, 490, 35 },
		{ 84, 491, 35 },
		{ 84, 492, 35 },
		{ 84, 493, 35 },
		{ 84, 494, 35 },
		{ 84, 495, 35 },
		{ 84, 496, 35 },
		{ 84, 497, 35 },
		{ 84, 498, 35 },
		{ 84, 499, 35 },
		{ 84, 500, 35 },
		{ 84, 501, 35 },
		{ 84, 502, 35 },
		{ 84, 503, 35 },
		{ 84, 504, 35 },
		{ 84, 505, 35 },
		{ 84, 506, 35 },
		{ 84, 507, 35 },
		{ 84, 508, 35 },
		{ 84, 509, 35 },
		{ 84, 510, 35 },
		{ 84, 511, 35 },
		{ 84, 512, 35 },
		{ 85, 487, 35 },
		{ 85, 488, 35 },
		{ 85, 489, 35 },
		{ 85, 490, 35 },
		{ 85, 491, 35 },
		{ 85, 492, 35 },
		{ 85, 493, 35 },
		{ 85, 494, 35 },
		{ 85, 495, 35 },
		{ 85, 496, 35 },
		{ 85, 497, 35 },
		{ 85, 498, 35 },
		{ 85, 499, 35 },
		{ 85, 500, 35 },
		{ 85, 501, 35 },
		{ 85, 502, 35 },
		{ 85, 503, 35 },
		{ 85, 504, 35 },
		{ 85, 505, 35 },
		{ 85, 506, 35 },
		{ 85, 507, 35 },
		{ 85, 508, 35 },
		{ 85, 509, 35 },
		{ 85, 510, 35 },
		{ 85, 511, 35 },
		{ 85, 512, 35 },
		{ 86, 486, 35 },
		{ 86, 487, 35 },
		{ 86, 488, 35 },
		{ 86, 489, 35 },
		{ 86, 490, 35 },
		{ 86, 491, 35 },
		{ 86, 492, 35 },
		{ 86, 493, 35 },
		{ 86, 494, 35 },
		{ 86, 495, 35 },
		{ 86, 496, 35 },
		{ 86, 497, 35 },
		{ 86, 498, 35 },
		{ 86, 499, 35 },
		{ 86, 500, 35 },
		{ 86, 501, 35 },
		{ 86, 502, 35 },
		{ 86, 503, 35 },
		{ 86, 504, 35 },
		{ 86, 505, 35 },
		{ 86, 506, 35 },
		{ 86, 507, 35 },
		{ 86, 508, 35 },
		{ 86, 509, 35 },
		{ 86, 510, 35 },
		{ 86, 511, 35 },
		{ 86, 512, 35 },
		{ 87, 484, 35 },
		{ 87, 485, 35 },
		{ 87, 486, 35 },
		{ 87, 487, 35 },
		{ 87, 488, 35 },
		{ 87, 489, 35 },
		{ 87, 490, 35 },
		{ 87, 491, 35 },
		{ 87, 492, 35 },
		{ 87, 493, 35 },
		{ 87, 494, 35 },
		{ 87, 495, 35 },
		{ 87, 496, 35 },
		{ 87, 497, 35 },
		{ 87, 498, 35 },
		{ 87, 499, 35 },
		{ 87, 500, 35 },
		{ 87, 501, 35 },
		{ 87, 502, 35 },
		{ 87, 503, 35 },
		{ 87, 504, 35 },
		{ 87, 505, 35 },
		{ 87, 506, 35 },
		{ 87, 507, 35 },
		{ 87, 508, 35 },
		{ 87, 509, 35 },
		{ 87, 510, 35 },
		{ 87, 511, 35 },
		{ 87, 512, 35 },
		{ 88, 483, 35 },
		{ 88, 484, 35 },
		{ 88, 485, 35 },
		{ 88, 486, 35 },
		{ 88, 487, 35 },
		{ 88, 488, 35 },
		{ 88, 489, 35 },
		{ 88, 490, 35 },
		{ 88, 491, 35 },
		{ 88, 492, 35 },
		{ 88, 493, 35 },
		{ 88, 494, 35 },
		{ 88, 495, 35 },
		{ 88, 496, 35 },
		{ 88, 497, 35 },
		{ 88, 498, 35 },
		{ 88, 499, 35 },
		{ 88, 500, 35 },
		{ 88, 501, 35 },
		{ 88, 502, 35 },
		{ 88, 503, 35 },
		{ 88, 504, 35 },
		{ 88, 505, 35 },
		{ 88, 506, 35 },
		{ 88, 507, 35 },
		{ 88, 508, 35 },
		{ 88, 509, 35 },
		{ 88, 510, 35 },
		{ 88, 511, 35 },
		{ 88, 512, 35 },
		{ 89, 482, 35 },
		{ 89, 483, 35 },
		{ 89, 484, 35 },
		{ 89, 485, 35 },
		{ 89, 486, 35 },
		{ 89, 487, 35 },
		{ 89, 488, 35 },
		{ 89, 489, 35 },
		{ 89, 490, 35 },
		{ 89, 491, 35 },
		{ 89, 492, 35 },
		{ 89, 493, 35 },
		{ 89, 494, 35 },
		{ 89, 495, 35 },
		{ 89, 496, 35 },
		{ 89, 497, 35 },
		{ 89, 498, 35 },
		{ 89, 499, 35 },
		{ 89, 500, 35 },
		{ 89, 501, 35 },
		{ 89, 502, 35 },
		{ 89, 503, 35 },
		{ 89, 504, 35 },
		{ 89, 505, 35 },
		{ 89, 506, 35 },
		{ 89, 507, 35 },
		{ 89, 508, 35 },
		{ 89, 509, 35 },
		{ 89, 510, 35 },
		{ 89, 511, 35 },
		{ 89, 512, 35 },
		{ 90, 480, 35 },
		{ 90, 481, 35 },
		{ 90, 482, 35 },
		{ 90, 483, 35 },
		{ 90, 484, 35 },
		{ 90, 485, 35 },
		{ 90, 486, 35 },
		{ 90, 487, 35 },
		{ 90, 488, 35 },
		{ 90, 489, 35 },
		{ 90, 490, 35 },
		{ 90, 491, 35 },
		{ 90, 492, 35 },
		{ 90, 493, 35 },
		{ 90, 494, 35 },
		{ 90, 495, 35 },
		{ 90, 496, 35 },
		{ 90, 497, 35 },
		{ 90, 498, 35 },
		{ 90, 499, 35 },
		{ 90, 500, 35 },
		{ 90, 501, 35 },
		{ 90, 502, 35 },
		{ 90, 503, 35 },
		{ 90, 504, 35 },
		{ 90, 505, 35 },
		{ 90, 506, 35 },
		{ 90, 507, 35 },
		{ 90, 508, 35 },
		{ 90, 509, 35 },
		{ 90, 510, 35 },
		{ 90, 511, 35 },
		{ 90, 512, 35 },
		{ 91, 479, 35 },
		{ 91, 480, 35 },
		{ 91, 481, 35 },
		{ 91, 482, 35 },
		{ 91, 483, 35 },
		{ 91, 484, 35 },
		{ 91, 485, 35 },
		{ 91, 486, 35 },
		{ 91, 487, 35 },
		{ 91, 488, 35 },
		{ 91, 489, 35 },
		{ 91, 490, 35 },
		{ 91, 491, 35 },
		{ 91, 492, 35 },
		{ 91, 493, 35 },
		{ 91, 494, 35 },
		{ 91, 495, 35 },
		{ 91, 496, 35 },
		{ 91, 497, 35 },
		{ 91, 498, 35 },
		{ 91, 499, 35 },
		{ 91, 500, 35 },
		{ 91, 501, 35 },
		{ 91, 502, 35 },
		{ 91, 503, 35 },
		{ 91, 504, 35 },
		{ 91, 505, 35 },
		{ 91, 506, 35 },
		{ 91, 507, 35 },
		{ 91, 508, 35 },
		{ 91, 509, 35 },
		{ 91, 510, 35 },
		{ 91, 511, 35 },
		{ 91, 512, 35 },
		{ 92, 478, 35 },
		{ 92, 479, 35 },
		{ 92, 480, 35 },
		{ 92, 481, 35 },
		{ 92, 482, 35 },
		{ 92, 483, 35 },
		{ 92, 484, 35 },
		{ 92, 485, 35 },
		{ 92, 486, 35 },
		{ 92, 487, 35 },
		{ 92, 488, 35 },
		{ 92, 489, 35 },
		{ 92, 490, 35 },
		{ 92, 491, 35 },
		{ 92, 492, 35 },
		{ 92, 493, 35 },
		{ 92, 494, 35 },
		{ 92, 495, 35 },
		{ 92, 496, 35 },
		{ 92, 497, 35 },
		{ 92, 498, 35 },
		{ 92, 499, 35 },
		{ 92, 500, 35 },
		{ 92, 501, 35 },
		{ 92, 502, 35 },
		{ 92, 503, 35 },
		{ 92, 504, 35 },
		{ 92, 505, 35 },
		{ 92, 506, 35 },
		{ 92, 507, 35 },
		{ 92, 508, 35 },
		{ 92, 509, 35 },
		{ 92, 510, 35 },
		{ 92, 511, 35 },
		{ 92, 512, 35 },
		{ 93, 476, 35 },
		{ 93, 477, 35 },
		{ 93, 478, 35 },
		{ 93, 479, 35 },
		{ 93, 480, 35 },
		{ 93, 481, 35 },
		{ 93, 482, 35 },
		{ 93, 483, 35 },
		{ 93, 484, 35 },
		{ 93, 485, 35 },
		{ 93, 486, 35 },
		{ 93, 487, 35 },
		{ 93, 488, 35 },
		{ 93, 489, 35 },
		{ 93, 490, 35 },
		{ 93, 491, 35 },
		{ 93, 492, 35 },
		{ 93, 493, 35 },
		{ 93, 494, 35 },
		{ 93, 495, 35 },
		{ 93, 496, 35 },
		{ 93, 497, 35 },
		{ 93, 498, 35 },
		{ 93, 499, 35 },
		{ 93, 500, 35 },
		{ 93, 501, 35 },
		{ 93, 502, 35 },
		{ 93, 503, 35 },
		{ 93, 504, 35 },
		{ 93, 505, 35 },
		{ 93, 506, 35 },
		{ 93, 507, 35 },
		{ 93, 508, 35 },
		{ 93, 509, 35 },
		{ 93, 510, 35 },
		{ 93, 511, 35 },
		{ 93, 512, 35 },
		{ 94, 475, 35 },
		{ 94, 476, 35 },
		{ 94, 477, 35 },
		{ 94, 478, 35 },
		{ 94, 479, 35 },
		{ 94, 480, 35 },
		{ 94, 481, 35 },
		{ 94, 482, 35 },
		{ 94, 483, 35 },
		{ 94, 484, 35 },
		{ 94, 485, 35 },
		{ 94, 486, 35 },
		{ 94, 487, 35 },
		{ 94, 488, 35 },
		{ 94, 489, 35 },
		{ 94, 490, 35 },
		{ 94, 491, 35 },
		{ 94, 492, 35 },
		{ 94, 493, 35 },
		{ 94, 494, 35 },
		{ 94, 495, 35 },
		{ 94, 496, 35 },
		{ 94, 497, 35 },
		{ 94, 498, 35 },
		{ 94, 499, 35 },
		{ 94, 500, 35 },
		{ 94, 501, 35 },
		{ 94, 502, 35 },
		{ 94, 503, 35 },
		{ 94, 504, 35 },
		{ 94, 505, 35 },
		{ 94, 506, 35 },
		{ 94, 507, 35 },
		{ 94, 508, 35 },
		{ 94, 509, 35 },
		{ 94, 510, 35 },
		{ 94, 511, 35 },
		{ 94, 512, 35 },
		{ 95, 474, 35 },
		{ 95, 475, 35 },
		{ 95, 476, 35 },
		{ 95, 477, 35 },
		{ 95, 478, 35 },
		{ 95, 479, 35 },
		{ 95, 480, 35 },
		{ 95, 481, 35 },
		{ 95, 482, 35 },
		{ 95, 483, 35 },
		{ 95, 484, 35 },
		{ 95, 485, 35 },
		{ 95, 486, 35 },
		{ 95, 487, 35 },
		{ 95, 488, 35 },
		{ 95, 489, 35 },
		{ 95, 490, 35 },
		{ 95, 491, 35 },
		{ 95, 492, 35 },
		{ 95, 493, 35 },
		{ 95, 494, 35 },
		{ 95, 495, 35 },
		{ 95, 496, 35 },
		{ 95, 497, 35 },
		{ 95, 498, 35 },
		{ 95, 499, 35 },
		{ 95, 500, 35 },
		{ 95, 501, 35 },
		{ 95, 502, 35 },
		{ 95, 503, 35 },
		{ 95, 504, 35 },
		{ 95, 505, 35 },
		{ 95, 506, 35 },
		{ 95, 507, 35 },
		{ 95, 508, 35 },
		{ 95, 509, 35 },
		{ 95, 510, 35 },
		{ 95, 511, 35 },
		{ 95, 512, 35 },
		{ 96, 472, 35 },
		{ 96, 473, 35 },
		{ 96, 474, 35 },
		{ 96, 475, 35 },
		{ 96, 476, 35 },
		{ 96, 477, 35 },
		{ 96, 478, 35 },
		{ 96, 479, 35 },
		{ 96, 480, 35 },
		{ 96, 481, 35 },
		{ 96, 482, 35 },
		{ 96, 483, 35 },
		{ 96, 484, 35 },
		{ 96, 485, 35 },
		{ 96, 486, 35 },
		{ 96, 487, 35 },
		{ 96, 488, 35 },
		{ 96, 489, 35 },
		{ 96, 490, 35 },
		{ 96, 491, 35 },
		{ 96, 492, 35 },
		{ 96, 493, 35 },
		{ 96, 494, 35 },
		{ 96, 495, 35 },
		{ 96, 496, 35 },
		{ 96, 497, 35 },
		{ 96, 498, 35 },
		{ 96, 499, 35 },
		{ 96, 500, 35 },
		{ 96, 501, 35 },
		{ 96, 502, 35 },
		{ 96, 503, 35 },
		{ 96, 504, 35 },
		{ 96, 505, 35 },
		{ 96, 506, 35 },
		{ 96, 507, 35 },
		{ 96, 508, 35 },
		{ 96, 509, 35 },
		{ 96, 510, 35 },
		{ 96, 511, 35 },
		{ 96, 512, 35 },
		{ 97, 471, 35 },
		{ 97, 472, 35 },
		{ 97, 473, 35 },
		{ 97, 474, 35 },
		{ 97, 475, 35 },
		{ 97, 476, 35 },
		{ 97, 477, 35 },
		{ 97, 478, 35 },
		{ 97, 479, 35 },
		{ 97, 480, 35 },
		{ 97, 481, 35 },
		{ 97, 482, 35 },
		{ 97, 483, 35 },
		{ 97, 484, 35 },
		{ 97, 485, 35 },
		{ 97, 486, 35 },
		{ 97, 487, 35 },
		{ 97, 488, 35 },
		{ 97, 489, 35 },
		{ 97, 490, 35 },
		{ 97, 491, 35 },
		{ 97, 492, 35 },
		{ 97, 493, 35 },
		{ 97, 494, 35 },
		{ 97, 495, 35 },
		{ 97, 496, 35 },
		{ 97, 497, 35 },
		{ 97, 498, 35 },
		{ 97, 499, 35 },
		{ 97, 500, 35 },
		{ 97, 501, 35 },
		{ 97, 502, 35 },
		{ 97, 503, 35 },
		{ 97, 504, 35 },
		{ 97, 505, 35 },
		{ 97, 506, 35 },
		{ 97, 507, 35 },
		{ 97, 508, 35 },
		{ 97, 509, 35 },
		{ 97, 510, 35 },
		{ 97, 511, 35 },
		{ 97, 512, 35 },
		{ 98, 470, 35 },
		{ 98, 471, 35 },
		{ 98, 472, 35 },
		{ 98, 473, 35 },
		{ 98, 474, 35 },
		{ 98, 475, 35 },
		{ 98, 476, 35 },
		{ 98, 477, 35 },
		{ 98, 478, 35 },
		{ 98, 479, 35 },
		{ 98, 480, 35 },
		{ 98, 481, 35 },
		{ 98, 482, 35 },
		{ 98, 483, 35 },
		{ 98, 484, 35 },
		{ 98, 485, 35 },
		{ 98, 486, 35 },
		{ 98, 487, 35 },
		{ 98, 488, 35 },
		{ 98, 489, 35 },
		{ 98, 490, 35 },
		{ 98, 491, 35 },
		{ 98, 492, 35 },
		{ 98, 493, 35 },
		{ 98, 494, 35 },
		{ 98, 495, 35 },
		{ 98, 496, 35 },
		{ 98, 497, 35 },
		{ 98, 498, 35 },
		{ 98, 499, 35 },
		{ 98, 500, 35 },
		{ 98, 501, 35 },
		{ 98, 502, 35 },
		{ 98, 503, 35 },
		{ 98, 504, 35 },
		{ 98, 505, 35 },
		{ 98, 506, 35 },
		{ 98, 507, 35 },
		{ 98, 508, 35 },
		{ 98, 509, 35 },
		{ 98, 510, 35 },
		{ 98, 511, 35 },
		{ 98, 512, 35 },
		{ 99, 468, 35 },
		{ 99, 469, 35 },
		{ 99, 470, 35 },
		{ 99, 471, 35 },
		{ 99, 472, 35 },
		{ 99, 473, 35 },
		{ 99, 474, 35 },
		{ 99, 475, 35 },
		{ 99, 476, 35 },
		{ 99, 477, 35 },
		{ 99, 478, 35 },
		{ 99, 479, 35 },
		{ 99, 480, 35 },
		{ 99, 481, 35 },
		{ 99, 482, 35 },
		{ 99, 483, 35 },
		{ 99, 484, 35 },
		{ 99, 485, 35 },
		{ 99, 486, 35 },
		{ 99, 487, 35 },
		{ 99, 488, 35 },
		{ 99, 489, 35 },
		{ 99, 490, 35 },
		{ 99, 491, 35 },
		{ 99, 492, 35 },
		{ 99, 493, 35 },
		{ 99, 494, 35 },
		{ 99, 495, 35 },
		{ 99, 496, 35 },
		{ 99, 497, 35 },
		{ 99, 498, 35 },
		{ 99, 499, 35 },
		{ 99, 500, 35 },
		{ 99, 501, 35 },
		{ 99, 502, 35 },
		{ 99, 503, 35 },
		{ 99, 504, 35 },
		{ 99, 505, 35 },
		{ 99, 506, 35 },
		{ 99, 507, 35 },
		{ 99, 508, 35 },
		{ 99, 509, 35 },
		{ 99, 510, 35 },
		{ 99, 511, 35 },
		{ 99, 512, 35 },
		{ 100, 467, 35 },
		{ 100, 468, 35 },
		{ 100, 469, 35 },
		{ 100, 470, 35 },
		{ 100, 471, 35 },
		{ 100, 472, 35 },
		{ 100, 473, 35 },
		{ 100, 474, 35 },
		{ 100, 475, 35 },
		{ 100, 476, 35 },
		{ 100, 477, 35 },
		{ 100, 478, 35 },
		{ 100, 479, 35 },
		{ 100, 480, 35 },
		{ 100, 481, 35 },
		{ 100, 482, 35 },
		{ 100, 483, 35 },
		{ 100, 484, 35 },
		{ 100, 485, 35 },
		{ 100, 486, 35 },
		{ 100, 487, 35 },
		{ 100, 488, 35 },
		{ 100, 489, 35 },
		{ 100, 490, 35 },
		{ 100, 491, 35 },
		{ 100, 492, 35 },
		{ 100, 493, 35 },
		{ 100, 494, 35 },
		{ 100, 495, 35 },
		{ 100, 496, 35 },
		{ 100, 497, 35 },
		{ 100, 498, 35 },
		{ 100, 499, 35 },
		{ 100, 500, 35 },
		{ 100, 501, 35 },
		{ 100, 502, 35 },
		{ 100, 503, 35 },
		{ 100, 504, 35 },
		{ 100, 505, 35 },
		{ 100, 506, 35 },
		{ 100, 507, 35 },
		{ 100, 508, 35 },
		{ 100, 509, 35 },
		{ 100, 510, 35 },
		{ 100, 511, 35 },
		{ 100, 512, 35 },
		{ 101, 466, 35 },
		{ 101, 467, 35 },
		{ 101, 468, 35 },
		{ 101, 469, 35 },
		{ 101, 470, 35 },
		{ 101, 471, 35 },
		{ 101, 472, 35 },
		{ 101, 473, 35 },
		{ 101, 474, 35 },
		{ 101, 475, 35 },
		{ 101, 476, 35 },
		{ 101, 477, 35 },
		{ 101, 478, 35 },
		{ 101, 479, 35 },
		{ 101, 480, 35 },
		{ 101, 481, 35 },
		{ 101, 482, 35 },
		{ 101, 483, 35 },
		{ 101, 484, 35 },
		{ 101, 485, 35 },
		{ 101, 486, 35 },
		{ 101, 487, 35 },
		{ 101, 488, 35 },
		{ 101, 489, 35 },
		{ 101, 490, 35 },
		{ 101, 491, 35 },
		{ 101, 492, 35 },
		{ 101, 493, 35 },
		{ 101, 494, 35 },
		{ 101, 495, 35 },
		{ 101, 496, 35 },
		{ 101, 497, 35 },
		{ 101, 498, 35 },
		{ 101, 499, 35 },
		{ 101, 500, 35 },
		{ 101, 501, 35 },
		{ 101, 502, 35 },
		{ 101, 503, 35 },
		{ 101, 504, 35 },
		{ 101, 505, 35 },
		{ 101, 506, 35 },
		{ 101, 507, 35 },
		{ 101, 508, 35 },
		{ 101, 509, 35 },
		{ 101, 510, 35 },
		{ 101, 511, 35 },
		{ 101, 512, 35 },
		{ 102, 464, 35 },
		{ 102, 465, 35 },
		{ 102, 466, 35 },
		{ 102, 467, 35 },
		{ 102, 468, 35 },
		{ 102, 469, 35 },
		{ 102, 470, 35 },
		{ 102, 471, 35 },
		{ 102, 472, 35 },
		{ 102, 473, 35 },
		{ 102, 474, 35 },
		{ 102, 475, 35 },
		{ 102, 476, 35 },
		{ 102, 477, 35 },
		{ 102, 478, 35 },
		{ 102, 479, 35 },
		{ 102, 480, 35 },
		{ 102, 481, 35 },
		{ 102, 482, 35 },
		{ 102, 483, 35 },
		{ 102, 484, 35 },
		{ 102, 485, 35 },
		{ 102, 486, 35 },
		{ 102, 487, 35 },
		{ 102, 488, 35 },
		{ 102, 489, 35 },
		{ 102, 490, 35 },
		{ 102, 491, 35 },
		{ 102, 492, 35 },
		{ 102, 493, 35 },
		{ 102, 494, 35 },
		{ 102, 495, 35 },
		{ 102, 496, 35 },
		{ 102, 497, 35 },
		{ 102, 498, 35 },
		{ 102, 499, 35 },
		{ 102, 500, 35 },
		{ 102, 501, 35 },
		{ 102, 502, 35 },
		{ 102, 503, 35 },
		{ 102, 504, 35 },
		{ 102, 505, 35 },
		{ 102, 506, 35 },
		{ 102, 507, 35 },
		{ 102, 508, 35 },
		{ 102, 509, 35 },
		{ 102, 510, 35 },
		{ 102, 511, 35 },
		{ 102, 512, 35 },
		{ 103, 463, 35 },
		{ 103, 464, 35 },
		{ 103, 465, 35 },
		{ 103, 466, 35 },
		{ 103, 467, 35 },
		{ 103, 468, 35 },
		{ 103, 469, 35 },
		{ 103, 470, 35 },
		{ 103, 471, 35 },
		{ 103, 472, 35 },
		{ 103, 473, 35 },
		{ 103, 474, 35 },
		{ 103, 475, 35 },
		{ 103, 476, 35 },
		{ 103, 477, 35 },
		{ 103, 478, 35 },
		{ 103, 479, 35 },
		{ 103, 480, 35 },
		{ 103, 481, 35 },
		{ 103, 482, 35 },
		{ 103, 483, 35 },
		{ 103, 484, 35 },
		{ 103, 485, 35 },
		{ 103, 486, 35 },
		{ 103, 487, 35 },
		{ 103, 488, 35 },
		{ 103, 489, 35 },
		{ 103, 490, 35 },
		{ 103, 491, 35 },
		{ 103, 492, 35 },
		{ 103, 493, 35 },
		{ 103, 494, 35 },
		{ 103, 495, 35 },
		{ 103, 496, 35 },
		{ 103, 497, 35 },
		{ 103, 498, 35 },
		{ 103, 499, 35 },
		{ 103, 500, 35 },
		{ 103, 501, 35 },
		{ 103, 502, 35 },
		{ 103, 503, 35 },
		{ 103, 504, 35 },
		{ 103, 505, 35 },
		{ 103, 506, 35 },
		{ 103, 507, 35 },
		{ 103, 508, 35 },
		{ 103, 509, 35 },
		{ 103, 510, 35 },
		{ 103, 511, 35 },
		{ 103, 512, 35 },
		{ 104, 462, 35 },
		{ 104, 463, 35 },
		{ 104, 464, 35 },
		{ 104, 465, 35 },
		{ 104, 466, 35 },
		{ 104, 467, 35 },
		{ 104, 468, 35 },
		{ 104, 469, 35 },
		{ 104, 470, 35 },
		{ 104, 471, 35 },
		{ 104, 472, 35 },
		{ 104, 473, 35 },
		{ 104, 474, 35 },
		{ 104, 475, 35 },
		{ 104, 476, 35 },
		{ 104, 477, 35 },
		{ 104, 478, 35 },
		{ 104, 479, 35 },
		{ 104, 480, 35 },
		{ 104, 481, 35 },
		{ 104, 482, 35 },
		{ 104, 483, 35 },
		{ 104, 484, 35 },
		{ 104, 485, 35 },
		{ 104, 486, 35 },
		{ 104, 487, 35 },
		{ 104, 488, 35 },
		{ 104, 489, 35 },
		{ 104, 490, 35 },
		{ 104, 491, 35 },
		{ 104, 492, 35 },
		{ 104, 493, 35 },
		{ 104, 494, 35 },
		{ 104, 495, 35 },
		{ 104, 496, 35 },
		{ 104, 497, 35 },
		{ 104, 498, 35 },
		{ 104, 499, 35 },
		{ 104, 500, 35 },
		{ 104, 501, 35 },
		{ 104, 502, 35 },
		{ 104, 503, 35 },
		{ 104, 504, 35 },
		{ 104, 505, 35 },
		{ 104, 506, 35 },
		{ 104, 507, 35 },
		{ 104, 508, 35 },
		{ 104, 509, 35 },
		{ 104, 510, 35 },
		{ 104, 511, 35 },
		{ 104, 512, 35 },
		{ 105, 460, 35 },
		{ 105, 461, 35 },
		{ 105, 462, 35 },
		{ 105, 463, 35 },
		{ 105, 464, 35 },
		{ 105, 465, 35 },
		{ 105, 466, 35 },
		{ 105, 467, 35 },
		{ 105, 468, 35 },
		{ 105, 469, 35 },
		{ 105, 470, 35 },
		{ 105, 471, 35 },
		{ 105, 472, 35 },
		{ 105, 473, 35 },
		{ 105, 474, 35 },
		{ 105, 475, 35 },
		{ 105, 476, 35 },
		{ 105, 477, 35 },
		{ 105, 478, 35 },
		{ 105, 479, 35 },
		{ 105, 480, 35 },
		{ 105, 481, 35 },
		{ 105, 482, 35 },
		{ 105, 483, 35 },
		{ 105, 484, 35 },
		{ 105, 485, 35 },
		{ 105, 486, 35 },
		{ 105, 487, 35 },
		{ 105, 488, 35 },
		{ 105, 489, 35 },
		{ 105, 490, 35 },
		{ 105, 491, 35 },
		{ 105, 492, 35 },
		{ 105, 493, 35 },
		{ 105, 494, 35 },
		{ 105, 495, 35 },
		{ 105, 496, 35 },
		{ 105, 497, 35 },
		{ 105, 498, 35 },
		{ 105, 499, 35 },
		{ 105, 500, 35 },
		{ 105, 501, 35 },
		{ 105, 502, 35 },
		{ 105, 503, 35 },
		{ 105, 504, 35 },
		{ 105, 505, 35 },
		{ 105, 506, 35 },
		{ 105, 507, 35 },
		{ 105, 508, 35 },
		{ 105, 509, 35 },
		{ 105, 510, 35 },
		{ 105, 511, 35 },
		{ 105, 512, 35 },
		{ 106, 459, 35 },
		{ 106, 460, 35 },
		{ 106, 461, 35 },
		{ 106, 462, 35 },
		{ 106, 463, 35 },
		{ 106, 464, 35 },
		{ 106, 465, 35 },
		{ 106, 466, 35 },
		{ 106, 467, 35 },
		{ 106, 468, 35 },
		{ 106, 469, 35 },
		{ 106, 470, 35 },
		{ 106, 471, 35 },
		{ 106, 472, 35 },
		{ 106, 473, 35 },
		{ 106, 474, 35 },
		{ 106, 475, 35 },
		{ 106, 476, 35 },
		{ 106, 477, 35 },
		{ 106, 478, 35 },
		{ 106, 479, 35 },
		{ 106, 480, 35 },
		{ 106, 481, 35 },
		{ 106, 482, 35 },
		{ 106, 483, 35 },
		{ 106, 484, 35 },
		{ 106, 485, 35 },
		{ 106, 486, 35 },
		{ 106, 487, 35 },
		{ 106, 488, 35 },
		{ 106, 489, 35 },
		{ 106, 490, 35 },
		{ 106, 491, 35 },
		{ 106, 492, 35 },
		{ 106, 493, 35 },
		{ 106, 494, 35 },
		{ 106, 495, 35 },
		{ 106, 496, 35 },
		{ 106, 497, 35 },
		{ 106, 498, 35 },
		{ 106, 499, 35 },
		{ 106, 500, 35 },
		{ 106, 501, 35 },
		{ 106, 502, 35 },
		{ 106, 503, 35 },
		{ 106, 504, 35 },
		{ 106, 505, 35 },
		{ 106, 506, 35 },
		{ 106, 507, 35 },
		{ 106, 508, 35 },
		{ 106, 509, 35 },
		{ 106, 510, 35 },
		{ 106, 511, 35 },
		{ 106, 512, 35 },
		{ 107, 458, 35 },
		{ 107, 459, 35 },
		{ 107, 460, 35 },
		{ 107, 461, 35 },
		{ 107, 462, 35 },
		{ 107, 463, 35 },
		{ 107, 464, 35 },
		{ 107, 465, 35 },
		{ 107, 466, 35 },
		{ 107, 467, 35 },
		{ 107, 468, 35 },
		{ 107, 469, 35 },
		{ 107, 470, 35 },
		{ 107, 471, 35 },
		{ 107, 472, 35 },
		{ 107, 473, 35 },
		{ 107, 474, 35 },
		{ 107, 475, 35 },
		{ 107, 476, 35 },
		{ 107, 477, 35 },
		{ 107, 478, 35 },
		{ 107, 479, 35 },
		{ 107, 480, 35 },
		{ 107, 481, 35 },
		{ 107, 482, 35 },
		{ 107, 483, 35 },
		{ 107, 484, 35 },
		{ 107, 485, 35 },
		{ 107, 486, 35 },
		{ 107, 487, 35 },
		{ 107, 488, 35 },
		{ 107, 489, 35 },
		{ 107, 490, 35 },
		{ 107, 491, 35 },
		{ 107, 492, 35 },
		{ 107, 493, 35 },
		{ 107, 494, 35 },
		{ 107, 495, 35 },
		{ 107, 496, 35 },
		{ 107, 497, 35 },
		{ 107, 498, 35 },
		{ 107, 499, 35 },
		{ 107, 500, 35 },
		{ 107, 501, 35 },
		{ 107, 502, 35 },
		{ 107, 503, 35 },
		{ 107, 504, 35 },
		{ 107, 505, 35 },
		{ 107, 506, 35 },
		{ 107, 507, 35 },
		{ 107, 508, 35 },
		{ 107, 509, 35 },
		{ 107, 510, 35 },
		{ 107, 511, 35 },
		{ 107, 512, 35 },
		{ 108, 456, 35 },
		{ 108, 457, 35 },
		{ 108, 458, 35 },
		{ 108, 459, 35 },
		{ 108, 460, 35 },
		{ 108, 461, 35 },
		{ 108, 462, 35 },
		{ 108, 463, 35 },
		{ 108, 464, 35 },
		{ 108, 465, 35 },
		{ 108, 466, 35 },
		{ 108, 467, 35 },
		{ 108, 468, 35 },
		{ 108, 469, 35 },
		{ 108, 470, 35 },
		{ 108, 471, 35 },
		{ 108, 472, 35 },
		{ 108, 473, 35 },
		{ 108, 474, 35 },
		{ 108, 475, 35 },
		{ 108, 476, 35 },
		{ 108, 477, 35 },
		{ 108, 478, 35 },
		{ 108, 479, 35 },
		{ 108, 480, 35 },
		{ 108, 481, 35 },
		{ 108, 482, 35 },
		{ 108, 483, 35 },
		{ 108, 484, 35 },
		{ 108, 485, 35 },
		{ 108, 486, 35 },
		{ 108, 487, 35 },
		{ 108, 488, 35 },
		{ 108, 489, 35 },
		{ 108, 490, 35 },
		{ 108, 491, 35 },
		{ 108, 492, 35 },
		{ 108, 493, 35 },
		{ 108, 494, 35 },
		{ 108, 495, 35 },
		{ 108, 496, 35 },
		{ 108, 497, 35 },
		{ 108, 498, 35 },
		{ 108, 499, 35 },
		{ 108, 500, 35 },
		{ 108, 501, 35 },
		{ 108, 502, 35 },
		{ 108, 503, 35 },
		{ 108, 504, 35 },
		{ 108, 505, 35 },
		{ 108, 506, 35 },
		{ 108, 507, 35 },
		{ 108, 508, 35 },
		{ 108, 509, 35 },
		{ 108, 510, 35 },
		{ 108, 511, 35 },
		{ 108, 512, 35 },
		{ 109, 455, 35 },
		{ 109, 456, 35 },
		{ 109, 457, 35 },
		{ 109, 458, 35 },
		{ 109, 459, 35 },
		{ 109, 460, 35 },
		{ 109, 461, 35 },
		{ 109, 462, 35 },
		{ 109, 463, 35 },
		{ 109, 464, 35 },
		{ 109, 465, 35 },
		{ 109, 466, 35 },
		{ 109, 467, 35 },
		{ 109, 468, 35 },
		{ 109, 469, 35 },
		{ 109, 470, 35 },
		{ 109, 471, 35 },
		{ 109, 472, 35 },
		{ 109, 473, 35 },
		{ 109, 474, 35 },
		{ 109, 475, 35 },
		{ 109, 476, 35 },
		{ 109, 477, 35 },
		{ 109, 478, 35 },
		{ 109, 479, 35 },
		{ 109, 480, 35 },
		{ 109, 481, 35 },
		{ 109, 482, 35 },
		{ 109, 483, 35 },
		{ 109, 484, 35 },
		{ 109, 485, 35 },
		{ 109, 486, 35 },
		{ 109, 487, 35 },
		{ 109, 488, 35 },
		{ 109, 489, 35 },
		{ 109, 490, 35 },
		{ 109, 491, 35 },
		{ 109, 492, 35 },
		{ 109, 493, 35 },
		{ 109, 494, 35 },
		{ 109, 495, 35 },
		{ 109, 496, 35 },
		{ 109, 497, 35 },
		{ 109, 498, 35 },
		{ 109, 499, 35 },
		{ 109, 500, 35 },
		{ 109, 501, 35 },
		{ 109, 502, 35 },
		{ 109, 503, 35 },
		{ 109, 504, 35 },
		{ 109, 505, 35 },
		{ 109, 506, 35 },
		{ 109, 507, 35 },
		{ 109, 508, 35 },
		{ 109, 509, 35 },
		{ 109, 510, 35 },
		{ 109, 511, 35 },
		{ 109, 512, 35 },
		{ 110, 454, 35 },
		{ 110, 455, 35 },
		{ 110, 456, 35 },
		{ 110, 457, 35 },
		{ 110, 458, 35 },
		{ 110, 459, 35 },
		{ 110, 460, 35 },
		{ 110, 461, 35 },
		{ 110, 462, 35 },
		{ 110, 463, 35 },
		{ 110, 464, 35 },
		{ 110, 465, 35 },
		{ 110, 466, 35 },
		{ 110, 467, 35 },
		{ 110, 468, 35 },
		{ 110, 469, 35 },
		{ 110, 470, 35 },
		{ 110, 471, 35 },
		{ 110, 472, 35 },
		{ 110, 473, 35 },
		{ 110, 474, 35 },
		{ 110, 475, 35 },
		{ 110, 476, 35 },
		{ 110, 477, 35 },
		{ 110, 478, 35 },
		{ 110, 479, 35 },
		{ 110, 480, 35 },
		{ 110, 481, 35 },
		{ 110, 482, 35 },
		{ 110, 483, 35 },
		{ 110, 484, 35 },
		{ 110, 485, 35 },
		{ 110, 486, 35 },
		{ 110, 487, 35 },
		{ 110, 488, 35 },
		{ 110, 489, 35 },
		{ 110, 490, 35 },
		{ 110, 491, 35 },
		{ 110, 492, 35 },
		{ 110, 493, 35 },
		{ 110, 494, 35 },
		{ 110, 495, 35 },
		{ 110, 496, 35 },
		{ 110, 497, 35 },
		{ 110, 498, 35 },
		{ 110, 499, 35 },
		{ 110, 500, 35 },
		{ 110, 501, 35 },
		{ 110, 502, 35 },
		{ 110, 503, 35 },
		{ 110, 504, 35 },
		{ 110, 505, 35 },
		{ 110, 506, 35 },
		{ 110, 507, 35 },
		{ 110, 508, 35 },
		{ 110, 509, 35 },
		{ 110, 510, 35 },
		{ 110, 511, 35 },
		{ 110, 512, 35 },
		{ 111, 452, 35 },
		{ 111, 453, 35 },
		{ 111, 454, 35 },
		{ 111, 455, 35 },
		{ 111, 456, 35 },
		{ 111, 457, 35 },
		{ 111, 458, 35 },
		{ 111, 459, 35 },
		{ 111, 460, 35 },
		{ 111, 461, 35 },
		{ 111, 462, 35 },
		{ 111, 463, 35 },
		{ 111, 464, 35 },
		{ 111, 465, 35 },
		{ 111, 466, 35 },
		{ 111, 467, 35 },
		{ 111, 468, 35 },
		{ 111, 469, 35 },
		{ 111, 470, 35 },
		{ 111, 471, 35 },
		{ 111, 472, 35 },
		{ 111, 473, 35 },
		{ 111, 474, 35 },
		{ 111, 475, 35 },
		{ 111, 476, 35 },
		{ 111, 477, 35 },
		{ 111, 478, 35 },
		{ 111, 479, 35 },
		{ 111, 480, 35 },
		{ 111, 481, 35 },
		{ 111, 482, 35 },
		{ 111, 483, 35 },
		{ 111, 484, 35 },
		{ 111, 485, 35 },
		{ 111, 486, 35 },
		{ 111, 487, 35 },
		{ 111, 488, 35 },
		{ 111, 489, 35 },
		{ 111, 490, 35 },
		{ 111, 491, 35 },
		{ 111, 492, 35 },
		{ 111, 493, 35 },
		{ 111, 494, 35 },
		{ 111, 495, 35 },
		{ 111, 496, 35 },
		{ 111, 497, 35 },
		{ 111, 498, 35 },
		{ 111, 499, 35 },
		{ 111, 500, 35 },
		{ 111, 501, 35 },
		{ 111, 502, 35 },
		{ 111, 503, 35 },
		{ 111, 504, 35 },
		{ 111, 505, 35 },
		{ 111, 506, 35 },
		{ 111, 507, 35 },
		{ 111, 508, 35 },
		{ 111, 509, 35 },
		{ 111, 510, 35 },
		{ 111, 511, 35 },
		{ 111, 512, 35 },
		{ 112, 451, 35 },
		{ 112, 452, 35 },
		{ 112, 453, 35 },
		{ 112, 454, 35 },
		{ 112, 455, 35 },
		{ 112, 456, 35 },
		{ 112, 457, 35 },
		{ 112, 458, 35 },
		{ 112, 459, 35 },
		{ 112, 460, 35 },
		{ 112, 461, 35 },
		{ 112, 462, 35 },
		{ 112, 463, 35 },
		{ 112, 464, 35 },
		{ 112, 465, 35 },
		{ 112, 466, 35 },
		{ 112, 467, 35 },
		{ 112, 468, 35 },
		{ 112, 469, 35 },
		{ 112, 470, 35 },
		{ 112, 471, 35 },
		{ 112, 472, 35 },
		{ 112, 473, 35 },
		{ 112, 474, 35 },
		{ 112, 475, 35 },
		{ 112, 476, 35 },
		{ 112, 477, 35 },
		{ 112, 478, 35 },
		{ 112, 479, 35 },
		{ 112, 480, 35 },
		{ 112, 481, 35 },
		{ 112, 482, 35 },
		{ 112, 483, 35 },
		{ 112, 484, 35 },
		{ 112, 485, 35 },
		{ 112, 486, 35 },
		{ 112, 487, 35 },
		{ 112, 488, 35 },
		{ 112, 489, 35 },
		{ 112, 490, 35 },
		{ 112, 491, 35 },
		{ 112, 492, 35 },
		{ 112, 493, 35 },
		{ 112, 494, 35 },
		{ 112, 495, 35 },
		{ 112, 496, 35 },
		{ 112, 497, 35 },
		{ 112, 498, 35 },
		{ 112, 499, 35 },
		{ 112, 500, 35 },
		{ 112, 501, 35 },
		{ 112, 502, 35 },
		{ 112, 503, 35 },
		{ 112, 504, 35 },
		{ 112, 505, 35 },
		{ 112, 506, 35 },
		{ 112, 507, 35 },
		{ 112, 508, 35 },
		{ 112, 509, 35 },
		{ 112, 510, 35 },
		{ 112, 511, 35 },
		{ 112, 512, 35 },
		{ 113, 450, 35 },
		{ 113, 451, 35 },
		{ 113, 452, 35 },
		{ 113, 453, 35 },
		{ 113, 454, 35 },
		{ 113, 455, 35 },
		{ 113, 456, 35 },
		{ 113, 457, 35 },
		{ 113, 458, 35 },
		{ 113, 459, 35 },
		{ 113, 460, 35 },
		{ 113, 461, 35 },
		{ 113, 462, 35 },
		{ 113, 463, 35 },
		{ 113, 464, 35 },
		{ 113, 465, 35 },
		{ 113, 466, 35 },
		{ 113, 467, 35 },
		{ 113, 468, 35 },
		{ 113, 469, 35 },
		{ 113, 470, 35 },
		{ 113, 471, 35 },
		{ 113, 472, 35 },
		{ 113, 473, 35 },
		{ 113, 474, 35 },
		{ 113, 475, 35 },
		{ 113, 476, 35 },
		{ 113, 477, 35 },
		{ 113, 478, 35 },
		{ 113, 479, 35 },
		{ 113, 480, 35 },
		{ 113, 481, 35 },
		{ 113, 482, 35 },
		{ 113, 483, 35 },
		{ 113, 484, 35 },
		{ 113, 485, 35 },
		{ 113, 486, 35 },
		{ 113, 487, 35 },
		{ 113, 488, 35 },
		{ 113, 489, 35 },
		{ 113, 490, 35 },
		{ 113, 491, 35 },
		{ 113, 492, 35 },
		{ 113, 493, 35 },
		{ 113, 494, 35 },
		{ 113, 495, 35 },
		{ 113, 496, 35 },
		{ 113, 497, 35 },
		{ 113, 498, 35 },
		{ 113, 499, 35 },
		{ 113, 500, 35 },
		{ 113, 501, 35 },
		{ 113, 502, 35 },
		{ 113, 503, 35 },
		{ 113, 504, 35 },
		{ 113, 505, 35 },
		{ 113, 506, 35 },
		{ 113, 507, 35 },
		{ 113, 508, 35 },
		{ 113, 509, 35 },
		{ 113, 510, 35 },
		{ 113, 511, 35 },
		{ 113, 512, 35 },
		{ 114, 448, 35 },
		{ 114, 449, 35 },
		{ 114, 450, 35 },
		{ 114, 451, 35 },
		{ 114, 452, 35 },
		{ 114, 453, 35 },
		{ 114, 454, 35 },
		{ 114, 455, 35 },
		{ 114, 456, 35 },
		{ 114, 457, 35 },
		{ 114, 458, 35 },
		{ 114, 459, 35 },
		{ 114, 460, 35 },
		{ 114, 461, 35 },
		{ 114, 462, 35 },
		{ 114, 463, 35 },
		{ 114, 464, 35 },
		{ 114, 465, 35 },
		{ 114, 466, 35 },
		{ 114, 467, 35 },
		{ 114, 468, 35 },
		{ 114, 469, 35 },
		{ 114, 470, 35 },
		{ 114, 471, 35 },
		{ 114, 472, 35 },
		{ 114, 473, 35 },
		{ 114, 474, 35 },
		{ 114, 475, 35 },
		{ 114, 476, 35 },
		{ 114, 477, 35 },
		{ 114, 478, 35 },
		{ 114, 479, 35 },
		{ 114, 480, 35 },
		{ 114, 481, 35 },
		{ 114, 482, 35 },
		{ 114, 483, 35 },
		{ 114, 484, 35 },
		{ 114, 485, 35 },
		{ 114, 486, 35 },
		{ 114, 487, 35 },
		{ 114, 488, 35 },
		{ 114, 489, 35 },
		{ 114, 490, 35 },
		{ 114, 491, 35 },
		{ 114, 492, 35 },
		{ 114, 493, 35 },
		{ 114, 494, 35 },
		{ 114, 495, 35 },
		{ 114, 496, 35 },
		{ 114, 497, 35 },
		{ 114, 498, 35 },
		{ 114, 499, 35 },
		{ 114, 500, 35 },
		{ 114, 501, 35 },
		{ 114, 502, 35 },
		{ 114, 503, 35 },
		{ 114, 504, 35 },
		{ 114, 505, 35 },
		{ 114, 506, 35 },
		{ 114, 507, 35 },
		{ 114, 508, 35 },
		{ 114, 509, 35 },
		{ 114, 510, 35 },
		{ 114, 511, 35 },
		{ 114, 512, 35 },
		{ 115, 447, 35 },
		{ 115, 448, 35 },
		{ 115, 449, 35 },
		{ 115, 450, 35 },
		{ 115, 451, 35 },
		{ 115, 452, 35 },
		{ 115, 453, 35 },
		{ 115, 454, 35 },
		{ 115, 455, 35 },
		{ 115, 456, 35 },
		{ 115, 457, 35 },
		{ 115, 458, 35 },
		{ 115, 459, 35 },
		{ 115, 460, 35 },
		{ 115, 461, 35 },
		{ 115, 462, 35 },
		{ 115, 463, 35 },
		{ 115, 464, 35 },
		{ 115, 465, 35 },
		{ 115, 466, 35 },
		{ 115, 467, 35 },
		{ 115, 468, 35 },
		{ 115, 469, 35 },
		{ 115, 470, 35 },
		{ 115, 471, 35 },
		{ 115, 472, 35 },
		{ 115, 473, 35 },
		{ 115, 474, 35 },
		{ 115, 475, 35 },
		{ 115, 476, 35 },
		{ 115, 477, 35 },
		{ 115, 478, 35 },
		{ 115, 479, 35 },
		{ 115, 480, 35 },
		{ 115, 481, 35 },
		{ 115, 482, 35 },
		{ 115, 483, 35 },
		{ 115, 484, 35 },
		{ 115, 485, 35 },
		{ 115, 486, 35 },
		{ 115, 487, 35 },
		{ 115, 488, 35 },
		{ 115, 489, 35 },
		{ 115, 490, 35 },
		{ 115, 491, 35 },
		{ 115, 492, 35 },
		{ 115, 493, 35 },
		{ 115, 494, 35 },
		{ 115, 495, 35 },
		{ 115, 496, 35 },
		{ 115, 497, 35 },
		{ 115, 498, 35 },
		{ 115, 499, 35 },
		{ 115, 500, 35 },
		{ 115, 501, 35 },
		{ 115, 502, 35 },
		{ 115, 503, 35 },
		{ 115, 504, 35 },
		{ 115, 505, 35 },
		{ 115, 506, 35 },
		{ 115, 507, 35 },
		{ 115, 508, 35 },
		{ 115, 509, 35 },
		{ 115, 510, 35 },
		{ 115, 511, 35 },
		{ 115, 512, 35 },
		{ 116, 446, 35 },
		{ 116, 447, 35 },
		{ 116, 448, 35 },
		{ 116, 449, 35 },
		{ 116, 450, 35 },
		{ 116, 451, 35 },
		{ 116, 452, 35 },
		{ 116, 453, 35 },
		{ 116, 454, 35 },
		{ 116, 455, 35 },
		{ 116, 456, 35 },
		{ 116, 457, 35 },
		{ 116, 458, 35 },
		{ 116, 459, 35 },
		{ 116, 460, 35 },
		{ 116, 461, 35 },
		{ 116, 462, 35 },
		{ 116, 463, 35 },
		{ 116, 464, 35 },
		{ 116, 465, 35 },
		{ 116, 466, 35 },
		{ 116, 467, 35 },
		{ 116, 468, 35 },
		{ 116, 469, 35 },
		{ 116, 470, 35 },
		{ 116, 471, 35 },
		{ 116, 472, 35 },
		{ 116, 473, 35 },
		{ 116, 474, 35 },
		{ 116, 475, 35 },
		{ 116, 476, 35 },
		{ 116, 477, 35 },
		{ 116, 478, 35 },
		{ 116, 479, 35 },
		{ 116, 480, 35 },
		{ 116, 481, 35 },
		{ 116, 482, 35 },
		{ 116, 483, 35 },
		{ 116, 484, 35 },
		{ 116, 485, 35 },
		{ 116, 486, 35 },
		{ 116, 487, 35 },
		{ 116, 488, 35 },
		{ 116, 489, 35 },
		{ 116, 490, 35 },
		{ 116, 491, 35 },
		{ 116, 492, 35 },
		{ 116, 493, 35 },
		{ 116, 494, 35 },
		{ 116, 495, 35 },
		{ 116, 496, 35 },
		{ 116, 497, 35 },
		{ 116, 498, 35 },
		{ 116, 499, 35 },
		{ 116, 500, 35 },
		{ 116, 501, 35 },
		{ 116, 502, 35 },
		{ 116, 503, 35 },
		{ 116, 504, 35 },
		{ 116, 505, 35 },
		{ 116, 506, 35 },
		{ 116, 507, 35 },
		{ 116, 508, 35 },
		{ 116, 509, 35 },
		{ 116, 510, 35 },
		{ 116, 511, 35 },
		{ 116, 512, 35 },
		{ 117, 444, 35 },
		{ 117, 445, 35 },
		{ 117, 446, 35 },
		{ 117, 447, 35 },
		{ 117, 448, 35 },
		{ 117, 449, 35 },
		{ 117, 450, 35 },
		{ 117, 451, 35 },
		{ 117, 452, 35 },
		{ 117, 453, 35 },
		{ 117, 454, 35 },
		{ 117, 455, 35 },
		{ 117, 456, 35 },
		{ 117, 457, 35 },
		{ 117, 458, 35 },
		{ 117, 459, 35 },
		{ 117, 460, 35 },
		{ 117, 461, 35 },
		{ 117, 462, 35 },
		{ 117, 463, 35 },
		{ 117, 464, 35 },
		{ 117, 465, 35 },
		{ 117, 466, 35 },
		{ 117, 467, 35 },
		{ 117, 468, 35 },
		{ 117, 469, 35 },
		{ 117, 470, 35 },
		{ 117, 471, 35 },
		{ 117, 472, 35 },
		{ 117, 473, 35 },
		{ 117, 474, 35 },
		{ 117, 475, 35 },
		{ 117, 476, 35 },
		{ 117, 477, 35 },
		{ 117, 478, 35 },
		{ 117, 479, 35 },
		{ 117, 480, 35 },
		{ 117, 481, 35 },
		{ 117, 482, 35 },
		{ 117, 483, 35 },
		{ 117, 484, 35 },
		{ 117, 485, 35 },
		{ 117, 486, 35 },
		{ 117, 487, 35 },
		{ 117, 488, 35 },
		{ 117, 489, 35 },
		{ 117, 490, 35 },
		{ 117, 491, 35 },
		{ 117, 492, 35 },
		{ 117, 493, 35 },
		{ 117, 494, 35 },
		{ 117, 495, 35 },
		{ 117, 496, 35 },
		{ 117, 497, 35 },
		{ 117, 498, 35 },
		{ 117, 499, 35 },
		{ 117, 500, 35 },
		{ 117, 501, 35 },
		{ 117, 502, 35 },
		{ 117, 503, 35 },
		{ 117, 504, 35 },
		{ 117, 505, 35 },
		{ 117, 506, 35 },
		{ 117, 507, 35 },
		{ 117, 508, 35 },
		{ 117, 509, 35 },
		{ 117, 510, 35 },
		{ 117, 511, 35 },
		{ 117, 512, 35 },
		{ 118, 443, 35 },
		{ 118, 444, 35 },
		{ 118, 445, 35 },
		{ 118, 446, 35 },
		{ 118, 447, 35 },
		{ 118, 448, 35 },
		{ 118, 449, 35 },
		{ 118, 450, 35 },
		{ 118, 451, 35 },
		{ 118, 452, 35 },
		{ 118, 453, 35 },
		{ 118, 454, 35 },
		{ 118, 455, 35 },
		{ 118, 456, 35 },
		{ 118, 457, 35 },
		{ 118, 458, 35 },
		{ 118, 459, 35 },
		{ 118, 460, 35 },
		{ 118, 461, 35 },
		{ 118, 462, 35 },
		{ 118, 463, 35 },
		{ 118, 464, 35 },
		{ 118, 465, 35 },
		{ 118, 466, 35 },
		{ 118, 467, 35 },
		{ 118, 468, 35 },
		{ 118, 469, 35 },
		{ 118, 470, 35 },
		{ 118, 471, 35 },
		{ 118, 472, 35 },
		{ 118, 473, 35 },
		{ 118, 474, 35 },
		{ 118, 475, 35 },
		{ 118, 476, 35 },
		{ 118, 477, 35 },
		{ 118, 478, 35 },
		{ 118, 479, 35 },
		{ 118, 480, 35 },
		{ 118, 481, 35 },
		{ 118, 482, 35 },
		{ 118, 483, 35 },
		{ 118, 484, 35 },
		{ 118, 485, 35 },
		{ 118, 486, 35 },
		{ 118, 487, 35 },
		{ 118, 488, 35 },
		{ 118, 489, 35 },
		{ 118, 490, 35 },
		{ 118, 491, 35 },
		{ 118, 492, 35 },
		{ 118, 493, 35 },
		{ 118, 494, 35 },
		{ 118, 495, 35 },
		{ 118, 496, 35 },
		{ 118, 497, 35 },
		{ 118, 498, 35 },
		{ 118, 499, 35 },
		{ 118, 500, 35 },
		{ 118, 501, 35 },
		{ 118, 502, 35 },
		{ 118, 503, 35 },
		{ 118, 504, 35 },
		{ 118, 505, 35 },
		{ 118, 506, 35 },
		{ 118, 507, 35 },
		{ 118, 508, 35 },
		{ 118, 509, 35 },
		{ 118, 510, 35 },
		{ 118, 511, 35 },
		{ 118, 512, 35 },
		{ 119, 442, 35 },
		{ 119, 443, 35 },
		{ 119, 444, 35 },
		{ 119, 445, 35 },
		{ 119, 446, 35 },
		{ 119, 447, 35 },
		{ 119, 448, 35 },
		{ 119, 449, 35 },
		{ 119, 450, 35 },
		{ 119, 451, 35 },
		{ 119, 452, 35 },
		{ 119, 453, 35 },
		{ 119, 454, 35 },
		{ 119, 455, 35 },
		{ 119, 456, 35 },
		{ 119, 457, 35 },
		{ 119, 458, 35 },
		{ 119, 459, 35 },
		{ 119, 460, 35 },
		{ 119, 461, 35 },
		{ 119, 462, 35 },
		{ 119, 463, 35 },
		{ 119, 464, 35 },
		{ 119, 465, 35 },
		{ 119, 466, 35 },
		{ 119, 467, 35 },
		{ 119, 468, 35 },
		{ 119, 469, 35 },
		{ 119, 470, 35 },
		{ 119, 471, 35 },
		{ 119, 472, 35 },
		{ 119, 473, 35 },
		{ 119, 474, 35 },
		{ 119, 475, 35 },
		{ 119, 476, 35 },
		{ 119, 477, 35 },
		{ 119, 478, 35 },
		{ 119, 479, 35 },
		{ 119, 480, 35 },
		{ 119, 481, 35 },
		{ 119, 482, 35 },
		{ 119, 483, 35 },
		{ 119, 484, 35 },
		{ 119, 485, 35 },
		{ 119, 486, 35 },
		{ 119, 487, 35 },
		{ 119, 488, 35 },
		{ 119, 489, 35 },
		{ 119, 490, 35 },
		{ 119, 491, 35 },
		{ 119, 492, 35 },
		{ 119, 493, 35 },
		{ 119, 494, 35 },
		{ 119, 495, 35 },
		{ 119, 496, 35 },
		{ 119, 497, 35 },
		{ 119, 498, 35 },
		{ 119, 499, 35 },
		{ 119, 500, 35 },
		{ 119, 501, 35 },
		{ 119, 502, 35 },
		{ 119, 503, 35 },
		{ 119, 504, 35 },
		{ 119, 505, 35 },
		{ 119, 506, 35 },
		{ 119, 507, 35 },
		{ 119, 508, 35 },
		{ 119, 509, 35 },
		{ 119, 510, 35 },
		{ 119, 511, 35 },
		{ 119, 512, 35 },
		{ 120, 440, 35 },
		{ 120, 441, 35 },
		{ 120, 442, 35 },
		{ 120, 443, 35 },
		{ 120, 444, 35 },
		{ 120, 445, 35 },
		{ 120, 446, 35 },
		{ 120, 447, 35 },
		{ 120, 448, 35 },
		{ 120, 449, 35 },
		{ 120, 450, 35 },
		{ 120, 451, 35 },
		{ 120, 452, 35 },
		{ 120, 453, 35 },
		{ 120, 454, 35 },
		{ 120, 455, 35 },
		{ 120, 456, 35 },
		{ 120, 457, 35 },
		{ 120, 458, 35 },
		{ 120, 459, 35 },
		{ 120, 460, 35 },
		{ 120, 461, 35 },
		{ 120, 462, 35 },
		{ 120, 463, 35 },
		{ 120, 464, 35 },
		{ 120, 465, 35 },
		{ 120, 466, 35 },
		{ 120, 467, 35 },
		{ 120, 468, 35 },
		{ 120, 469, 35 },
		{ 120, 470, 35 },
		{ 120, 471, 35 },
		{ 120, 472, 35 },
		{ 120, 473, 35 },
		{ 120, 474, 35 },
		{ 120, 475, 35 },
		{ 120, 476, 35 },
		{ 120, 477, 35 },
		{ 120, 478, 35 },
		{ 120, 479, 35 },
		{ 120, 480, 35 },
		{ 120, 481, 35 },
		{ 120, 482, 35 },
		{ 120, 483, 35 },
		{ 120, 484, 35 },
		{ 120, 485, 35 },
		{ 120, 486, 35 },
		{ 120, 487, 35 },
		{ 120, 488, 35 },
		{ 120, 489, 35 },
		{ 120, 490, 35 },
		{ 120, 491, 35 },
		{ 120, 492, 35 },
		{ 120, 493, 35 },
		{ 120, 494, 35 },
		{ 120, 495, 35 },
		{ 120, 496, 35 },
		{ 120, 497, 35 },
		{ 120, 498, 35 },
		{ 120, 499, 35 },
		{ 120, 500, 35 },
		{ 120, 501, 35 },
		{ 120, 502, 35 },
		{ 120, 503, 35 },
		{ 120, 504, 35 },
		{ 120, 505, 35 },
		{ 120, 506, 35 },
		{ 120, 507, 35 },
		{ 120, 508, 35 },
		{ 120, 509, 35 },
		{ 120, 510, 35 },
		{ 120, 511, 35 },
		{ 120, 512, 35 },
		{ 121, 439, 35 },
		{ 121, 440, 35 },
		{ 121, 441, 35 },
		{ 121, 442, 35 },
		{ 121, 443, 35 },
		{ 121, 444, 35 },
		{ 121, 445, 35 },
		{ 121, 446, 35 },
		{ 121, 447, 35 },
		{ 121, 448, 35 },
		{ 121, 449, 35 },
		{ 121, 450, 35 },
		{ 121, 451, 35 },
		{ 121, 452, 35 },
		{ 121, 453, 35 },
		{ 121, 454, 35 },
		{ 121, 455, 35 },
		{ 121, 456, 35 },
		{ 121, 457, 35 },
		{ 121, 458, 35 },
		{ 121, 459, 35 },
		{ 121, 460, 35 },
		{ 121, 461, 35 },
		{ 121, 462, 35 },
		{ 121, 463, 35 },
		{ 121, 464, 35 },
		{ 121, 465, 35 },
		{ 121, 466, 35 },
		{ 121, 467, 35 },
		{ 121, 468, 35 },
		{ 121, 469, 35 },
		{ 121, 470, 35 },
		{ 121, 471, 35 },
		{ 121, 472, 35 },
		{ 121, 473, 35 },
		{ 121, 474, 35 },
		{ 121, 475, 35 },
		{ 121, 476, 35 },
		{ 121, 477, 35 },
		{ 121, 478, 35 },
		{ 121, 479, 35 },
		{ 121, 480, 35 },
		{ 121, 481, 35 },
		{ 121, 482, 35 },
		{ 121, 483, 35 },
		{ 121, 484, 35 },
		{ 121, 485, 35 },
		{ 121, 486, 35 },
		{ 121, 487, 35 },
		{ 121, 488, 35 },
		{ 121, 489, 35 },
		{ 121, 490, 35 },
		{ 121, 491, 35 },
		{ 121, 492, 35 },
		{ 121, 493, 35 },
		{ 121, 494, 35 },
		{ 121, 495, 35 },
		{ 121, 496, 35 },
		{ 121, 497, 35 },
		{ 121, 498, 35 },
		{ 121, 499, 35 },
		{ 121, 500, 35 },
		{ 121, 501, 35 },
		{ 121, 502, 35 },
		{ 121, 503, 35 },
		{ 121, 504, 35 },
		{ 121, 505, 35 },
		{ 121, 506, 35 },
		{ 121, 507, 35 },
		{ 121, 508, 35 },
		{ 121, 509, 35 },
		{ 121, 510, 35 },
		{ 121, 511, 35 },
		{ 121, 512, 35 },
		{ 122, 438, 35 },
		{ 122, 439, 35 },
		{ 122, 440, 35 },
		{ 122, 441, 35 },
		{ 122, 442, 35 },
		{ 122, 443, 35 },
		{ 122, 444, 35 },
		{ 122, 445, 35 },
		{ 122, 446, 35 },
		{ 122, 447, 35 },
		{ 122, 448, 35 },
		{ 122, 449, 35 },
		{ 122, 450, 35 },
		{ 122, 451, 35 },
		{ 122, 452, 35 },
		{ 122, 453, 35 },
		{ 122, 454, 35 },
		{ 122, 455, 35 },
		{ 122, 456, 35 },
		{ 122, 457, 35 },
		{ 122, 458, 35 },
		{ 122, 459, 35 },
		{ 122, 460, 35 },
		{ 122, 461, 35 },
		{ 122, 462, 35 },
		{ 122, 463, 35 },
		{ 122, 464, 35 },
		{ 122, 465, 35 },
		{ 122, 466, 35 },
		{ 122, 467, 35 },
		{ 122, 468, 35 },
		{ 122, 469, 35 },
		{ 122, 470, 35 },
		{ 122, 471, 35 },
		{ 122, 472, 35 },
		{ 122, 473, 35 },
		{ 122, 474, 35 },
		{ 122, 475, 35 },
		{ 122, 476, 35 },
		{ 122, 477, 35 },
		{ 122, 478, 35 },
		{ 122, 479, 35 },
		{ 122, 480, 35 },
		{ 122, 481, 35 },
		{ 122, 482, 35 },
		{ 122, 483, 35 },
		{ 122, 484, 35 },
		{ 122, 485, 35 },
		{ 122, 486, 35 },
		{ 122, 487, 35 },
		{ 122, 488, 35 },
		{ 122, 489, 35 },
		{ 122, 490, 35 },
		{ 122, 491, 35 },
		{ 122, 492, 35 },
		{ 122, 493, 35 },
		{ 122, 494, 35 },
		{ 122, 495, 35 },
		{ 122, 496, 35 },
		{ 122, 497, 35 },
		{ 122, 498, 35 },
		{ 122, 499, 35 },
		{ 122, 500, 35 },
		{ 122, 501, 35 },
		{ 122, 502, 35 },
		{ 122, 503, 35 },
		{ 122, 504, 35 },
		{ 122, 505, 35 },
		{ 122, 506, 35 },
		{ 122, 507, 35 },
		{ 122, 508, 35 },
		{ 122, 509, 35 },
		{ 122, 510, 35 },
		{ 122, 511, 35 },
		{ 122, 512, 35 },
		{ 123, 436, 35 },
		{ 123, 437, 35 },
		{ 123, 438, 35 },
		{ 123, 439, 35 },
		{ 123, 440, 35 },
		{ 123, 441, 35 },
		{ 123, 442, 35 },
		{ 123, 443, 35 },
		{ 123, 444, 35 },
		{ 123, 445, 35 },
		{ 123, 446, 35 },
		{ 123, 447, 35 },
		{ 123, 448, 35 },
		{ 123, 449, 35 },
		{ 123, 450, 35 },
		{ 123, 451, 35 },
		{ 123, 452, 35 },
		{ 123, 453, 35 },
		{ 123, 454, 35 },
		{ 123, 455, 35 },
		{ 123, 456, 35 },
		{ 123, 457, 35 },
		{ 123, 458, 35 },
		{ 123, 459, 35 },
		{ 123, 460, 35 },
		{ 123, 461, 35 },
		{ 123, 462, 35 },
		{ 123, 463, 35 },
		{ 123, 464, 35 },
		{ 123, 465, 35 },
		{ 123, 466, 35 },
		{ 123, 467, 35 },
		{ 123, 468, 35 },
		{ 123, 469, 35 },
		{ 123, 470, 35 },
		{ 123, 471, 35 },
		{ 123, 472, 35 },
		{ 123, 473, 35 },
		{ 123, 474, 35 },
		{ 123, 475, 35 },
		{ 123, 476, 35 },
		{ 123, 477, 35 },
		{ 123, 478, 35 },
		{ 123, 479, 35 },
		{ 123, 480, 35 },
		{ 123, 481, 35 },
		{ 123, 482, 35 },
		{ 123, 483, 35 },
		{ 123, 484, 35 },
		{ 123, 485, 35 },
		{ 123, 486, 35 },
		{ 123, 487, 35 },
		{ 123, 488, 35 },
		{ 123, 489, 35 },
		{ 123, 490, 35 },
		{ 123, 491, 35 },
		{ 123, 492, 35 },
		{ 123, 493, 35 },
		{ 123, 494, 35 },
		{ 123, 495, 35 },
		{ 123, 496, 35 },
		{ 123, 497, 35 },
		{ 123, 498, 35 },
		{ 123, 499, 35 },
		{ 123, 500, 35 },
		{ 123, 501, 35 },
		{ 123, 502, 35 },
		{ 123, 503, 35 },
		{ 123, 504, 35 },
		{ 123, 505, 35 },
		{ 123, 506, 35 },
		{ 123, 507, 35 },
		{ 123, 508, 35 },
		{ 123, 509, 35 },
		{ 123, 510, 35 },
		{ 123, 511, 35 },
		{ 123, 512, 35 },
		{ 124, 435, 35 },
		{ 124, 436, 35 },
		{ 124, 437, 35 },
		{ 124, 438, 35 },
		{ 124, 439, 35 },
		{ 124, 440, 35 },
		{ 124, 441, 35 },
		{ 124, 442, 35 },
		{ 124, 443, 35 },
		{ 124, 444, 35 },
		{ 124, 445, 35 },
		{ 124, 446, 35 },
		{ 124, 447, 35 },
		{ 124, 448, 35 },
		{ 124, 449, 35 },
		{ 124, 450, 35 },
		{ 124, 451, 35 },
		{ 124, 452, 35 },
		{ 124, 453, 35 },
		{ 124, 454, 35 },
		{ 124, 455, 35 },
		{ 124, 456, 35 },
		{ 124, 457, 35 },
		{ 124, 458, 35 },
		{ 124, 459, 35 },
		{ 124, 460, 35 },
		{ 124, 461, 35 },
		{ 124, 462, 35 },
		{ 124, 463, 35 },
		{ 124, 464, 35 },
		{ 124, 465, 35 },
		{ 124, 466, 35 },
		{ 124, 467, 35 },
		{ 124, 468, 35 },
		{ 124, 469, 35 },
		{ 124, 470, 35 },
		{ 124, 471, 35 },
		{ 124, 472, 35 },
		{ 124, 473, 35 },
		{ 124, 474, 35 },
		{ 124, 475, 35 },
		{ 124, 476, 35 },
		{ 124, 477, 35 },
		{ 124, 478, 35 },
		{ 124, 479, 35 },
		{ 124, 480, 35 },
		{ 124, 481, 35 },
		{ 124, 482, 35 },
		{ 124, 483, 35 },
		{ 124, 484, 35 },
		{ 124, 485, 35 },
		{ 124, 486, 35 },
		{ 124, 487, 35 },
		{ 124, 488, 35 },
		{ 124, 489, 35 },
		{ 124, 490, 35 },
		{ 124, 491, 35 },
		{ 124, 492, 35 },
		{ 124, 493, 35 },
		{ 124, 494, 35 },
		{ 124, 495, 35 },
		{ 124, 496, 35 },
		{ 124, 497, 35 },
		{ 124, 498, 35 },
		{ 124, 499, 35 },
		{ 124, 500, 35 },
		{ 124, 501, 35 },
		{ 124, 502, 35 },
		{ 124, 503, 35 },
		{ 124, 504, 35 },
		{ 124, 505, 35 },
		{ 124, 506, 35 },
		{ 124, 507, 35 },
		{ 124, 508, 35 },
		{ 124, 509, 35 },
		{ 124, 510, 35 },
		{ 124, 511, 35 },
		{ 124, 512, 35 },
		{ 125, 434, 35 },
		{ 125, 435, 35 },
		{ 125, 436, 35 },
		{ 125, 437, 35 },
		{ 125, 438, 35 },
		{ 125, 439, 35 },
		{ 125, 440, 35 },
		{ 125, 441, 35 },
		{ 125, 442, 35 },
		{ 125, 443, 35 },
		{ 125, 444, 35 },
		{ 125, 445, 35 },
		{ 125, 446, 35 },
		{ 125, 447, 35 },
		{ 125, 448, 35 },
		{ 125, 449, 35 },
		{ 125, 450, 35 },
		{ 125, 451, 35 },
		{ 125, 452, 35 },
		{ 125, 453, 35 },
		{ 125, 454, 35 },
		{ 125, 455, 35 },
		{ 125, 456, 35 },
		{ 125, 457, 35 },
		{ 125, 458, 35 },
		{ 125, 459, 35 },
		{ 125, 460, 35 },
		{ 125, 461, 35 },
		{ 125, 462, 35 },
		{ 125, 463, 35 },
		{ 125, 464, 35 },
		{ 125, 465, 35 },
		{ 125, 466, 35 },
		{ 125, 467, 35 },
		{ 125, 468, 35 },
		{ 125, 469, 35 },
		{ 125, 470, 35 },
		{ 125, 471, 35 },
		{ 125, 472, 35 },
		{ 125, 473, 35 },
		{ 125, 474, 35 },
		{ 125, 475, 35 },
		{ 125, 476, 35 },
		{ 125, 477, 35 },
		{ 125, 478, 35 },
		{ 125, 479, 35 },
		{ 125, 480, 35 },
		{ 125, 481, 35 },
		{ 125, 482, 35 },
		{ 125, 483, 35 },
		{ 125, 484, 35 },
		{ 125, 485, 35 },
		{ 125, 486, 35 },
		{ 125, 487, 35 },
		{ 125, 488, 35 },
		{ 125, 489, 35 },
		{ 125, 490, 35 },
		{ 125, 491, 35 },
		{ 125, 492, 35 },
		{ 125, 493, 35 },
		{ 125, 494, 35 },
		{ 125, 495, 35 },
		{ 125, 496, 35 },
		{ 125, 497, 35 },
		{ 125, 498, 35 },
		{ 125, 499, 35 },
		{ 125, 500, 35 },
		{ 125, 501, 35 },
		{ 125, 502, 35 },
		{ 125, 503, 35 },
		{ 125, 504, 35 },
		{ 125, 505, 35 },
		{ 125, 506, 35 },
		{ 125, 507, 35 },
		{ 125, 508, 35 },
		{ 125, 509, 35 },
		{ 125, 510, 35 },
		{ 125, 511, 35 },
		{ 125, 512, 35 },
		{ 126, 432, 35 },
		{ 126, 433, 35 },
		{ 126, 434, 35 },
		{ 126, 435, 35 },
		{ 126, 436, 35 },
		{ 126, 437, 35 },
		{ 126, 438, 35 },
		{ 126, 439, 35 },
		{ 126, 440, 35 },
		{ 126, 441, 35 },
		{ 126, 442, 35 },
		{ 126, 443, 35 },
		{ 126, 444, 35 },
		{ 126, 445, 35 },
		{ 126, 446, 35 },
		{ 126, 447, 35 },
		{ 126, 448, 35 },
		{ 126, 449, 35 },
		{ 126, 450, 35 },
		{ 126, 451, 35 },
		{ 126, 452, 35 },
		{ 126, 453, 35 },
		{ 126, 454, 35 },
		{ 126, 455, 35 },
		{ 126, 456, 35 },
		{ 126, 457, 35 },
		{ 126, 458, 35 },
		{ 126, 459, 35 },
		{ 126, 460, 35 },
		{ 126, 461, 35 },
		{ 126, 462, 35 },
		{ 126, 463, 35 },
		{ 126, 464, 35 },
		{ 126, 465, 35 },
		{ 126, 466, 35 },
		{ 126, 467, 35 },
		{ 126, 468, 35 },
		{ 126, 469, 35 },
		{ 126, 470, 35 },
		{ 126, 471, 35 },
		{ 126, 472, 35 },
		{ 126, 473, 35 },
		{ 126, 474, 35 },
		{ 126, 475, 35 },
		{ 126, 476, 35 },
		{ 126, 477, 35 },
		{ 126, 478, 35 },
		{ 126, 479, 35 },
		{ 126, 480, 35 },
		{ 126, 481, 35 },
		{ 126, 482, 35 },
		{ 126, 483, 35 },
		{ 126, 484, 35 },
		{ 126, 485, 35 },
		{ 126, 486, 35 },
		{ 126, 487, 35 },
		{ 126, 488, 35 },
		{ 126, 489, 35 },
		{ 126, 490, 35 },
		{ 126, 491, 35 },
		{ 126, 492, 35 },
		{ 126, 493, 35 },
		{ 126, 494, 35 },
		{ 126, 495, 35 },
		{ 126, 496, 35 },
		{ 126, 497, 35 },
		{ 126, 498, 35 },
		{ 126, 499, 35 },
		{ 126, 500, 35 },
		{ 126, 501, 35 },
		{ 126, 502, 35 },
		{ 126, 503, 35 },
		{ 126, 504, 35 },
		{ 126, 505, 35 },
		{ 126, 506, 35 },
		{ 126, 507, 35 },
		{ 126, 508, 35 },
		{ 126, 509, 35 },
		{ 126, 510, 35 },
		{ 126, 511, 35 },
		{ 126, 512, 35 },
		{ 127, 431, 35 },
		{ 127, 432, 35 },
		{ 127, 433, 35 },
		{ 127, 434, 35 },
		{ 127, 435, 35 },
		{ 127, 436, 35 },
		{ 127, 437, 35 },
		{ 127, 438, 35 },
		{ 127, 439, 35 },
		{ 127, 440, 35 },
		{ 127, 441, 35 },
		{ 127, 442, 35 },
		{ 127, 443, 35 },
		{ 127, 444, 35 },
		{ 127, 445, 35 },
		{ 127, 446, 35 },
		{ 127, 447, 35 },
		{ 127, 448, 35 },
		{ 127, 449, 35 },
		{ 127, 450, 35 },
		{ 127, 451, 35 },
		{ 127, 452, 35 },
		{ 127, 453, 35 },
		{ 127, 454, 35 },
		{ 127, 455, 35 },
		{ 127, 456, 35 },
		{ 127, 457, 35 },
		{ 127, 458, 35 },
		{ 127, 459, 35 },
		{ 127, 460, 35 },
		{ 127, 461, 35 },
		{ 127, 462, 35 },
		{ 127, 463, 35 },
		{ 127, 464, 35 },
		{ 127, 465, 35 },
		{ 127, 466, 35 },
		{ 127, 467, 35 },
		{ 127, 468, 35 },
		{ 127, 469, 35 },
		{ 127, 470, 35 },
		{ 127, 471, 35 },
		{ 127, 472, 35 },
		{ 127, 473, 35 },
		{ 127, 474, 35 },
		{ 127, 475, 35 },
		{ 127, 476, 35 },
		{ 127, 477, 35 },
		{ 127, 478, 35 },
		{ 127, 479, 35 },
		{ 127, 480, 35 },
		{ 127, 481, 35 },
		{ 127, 482, 35 },
		{ 127, 483, 35 },
		{ 127, 484, 35 },
		{ 127, 485, 35 },
		{ 127, 486, 35 },
		{ 127, 487, 35 },
		{ 127, 488, 35 },
		{ 127, 489, 35 },
		{ 127, 490, 35 },
		{ 127, 491, 35 },
		{ 127, 492, 35 },
		{ 127, 493, 35 },
		{ 127, 494, 35 },
		{ 127, 495, 35 },
		{ 127, 496, 35 },
		{ 127, 497, 35 },
		{ 127, 498, 35 },
		{ 127, 499, 35 },
		{ 127, 500, 35 },
		{ 127, 501, 35 },
		{ 127, 502, 35 },
		{ 127, 503, 35 },
		{ 127, 504, 35 },
		{ 127, 505, 35 },
		{ 127, 506, 35 },
		{ 127, 507, 35 },
		{ 127, 508, 35 },
		{ 127, 509, 35 },
		{ 127, 510, 35 },
		{ 127, 511, 35 },
		{ 127, 512, 35 },
		{ 128, 430, 35 },
		{ 128, 431, 35 },
		{ 128, 432, 35 },
		{ 128, 433, 35 },
		{ 128, 434, 35 },
		{ 128, 435, 35 },
		{ 128, 436, 35 },
		{ 128, 437, 35 },
		{ 128, 438, 35 },
		{ 128, 439, 35 },
		{ 128, 440, 35 },
		{ 128, 441, 35 },
		{ 128, 442, 35 },
		{ 128, 443, 35 },
		{ 128, 444, 35 },
		{ 128, 445, 35 },
		{ 128, 446, 35 },
		{ 128, 447, 35 },
		{ 128, 448, 35 },
		{ 128, 449, 35 },
		{ 128, 450, 35 },
		{ 128, 451, 35 },
		{ 128, 452, 35 },
		{ 128, 453, 35 },
		{ 128, 454, 35 },
		{ 128, 455, 35 },
		{ 128, 456, 35 },
		{ 128, 457, 35 },
		{ 128, 458, 35 },
		{ 128, 459, 35 },
		{ 128, 460, 35 },
		{ 128, 461, 35 },
		{ 128, 462, 35 },
		{ 128, 463, 35 },
		{ 128, 464, 35 },
		{ 128, 465, 35 },
		{ 128, 466, 35 },
		{ 128, 467, 35 },
		{ 128, 468, 35 },
		{ 128, 469, 35 },
		{ 128, 470, 35 },
		{ 128, 471, 35 },
		{ 128, 472, 35 },
		{ 128, 473, 35 },
		{ 128, 474, 35 },
		{ 128, 475, 35 },
		{ 128, 476, 35 },
		{ 128, 477, 35 },
		{ 128, 478, 35 },
		{ 128, 479, 35 },
		{ 128, 480, 35 },
		{ 128, 481, 35 },
		{ 128, 482, 35 },
		{ 128, 483, 35 },
		{ 128, 484, 35 },
		{ 128, 485, 35 },
		{ 128, 486, 35 },
		{ 128, 487, 35 },
		{ 128, 488, 35 },
		{ 128, 489, 35 },
		{ 128, 490, 35 },
		{ 128, 491, 35 },
		{ 128, 492, 35 },
		{ 128, 493, 35 },
		{ 128, 494, 35 },
		{ 128, 495, 35 },
		{ 128, 496, 35 },
		{ 128, 497, 35 },
		{ 128, 498, 35 },
		{ 128, 499, 35 },
		{ 128, 500, 35 },
		{ 128, 501, 35 },
		{ 128, 502, 35 },
		{ 128, 503, 35 },
		{ 128, 504, 35 },
		{ 128, 505, 35 },
		{ 128, 506, 35 },
		{ 128, 507, 35 },
		{ 128, 508, 35 },
		{ 128, 509, 35 },
		{ 128, 510, 35 },
		{ 128, 511, 35 },
		{ 128, 512, 35 },
		{ 129, 428, 35 },
		{ 129, 429, 35 },
		{ 129, 430, 35 },
		{ 129, 431, 35 },
		{ 129, 432, 35 },
		{ 129, 433, 35 },
		{ 129, 434, 35 },
		{ 129, 435, 35 },
		{ 129, 436, 35 },
		{ 129, 437, 35 },
		{ 129, 438, 35 },
		{ 129, 439, 35 },
		{ 129, 440, 35 },
		{ 129, 441, 35 },
		{ 129, 442, 35 },
		{ 129, 443, 35 },
		{ 129, 444, 35 },
		{ 129, 445, 35 },
		{ 129, 446, 35 },
		{ 129, 447, 35 },
		{ 129, 448, 35 },
		{ 129, 449, 35 },
		{ 129, 450, 35 },
		{ 129, 451, 35 },
		{ 129, 452, 35 },
		{ 129, 453, 35 },
		{ 129, 454, 35 },
		{ 129, 455, 35 },
		{ 129, 456, 35 },
		{ 129, 457, 35 },
		{ 129, 458, 35 },
		{ 129, 459, 35 },
		{ 129, 460, 35 },
		{ 129, 461, 35 },
		{ 129, 462, 35 },
		{ 129, 463, 35 },
		{ 129, 464, 35 },
		{ 129, 465, 35 },
		{ 129, 466, 35 },
		{ 129, 467, 35 },
		{ 129, 468, 35 },
		{ 129, 469, 35 },
		{ 129, 470, 35 },
		{ 129, 471, 35 },
		{ 129, 472, 35 },
		{ 129, 473, 35 },
		{ 129, 474, 35 },
		{ 129, 475, 35 },
		{ 129, 476, 35 },
		{ 129, 477, 35 },
		{ 129, 478, 35 },
		{ 129, 479, 35 },
		{ 129, 480, 35 },
		{ 129, 481, 35 },
		{ 129, 482, 35 },
		{ 129, 483, 35 },
		{ 129, 484, 35 },
		{ 129, 485, 35 },
		{ 129, 486, 35 },
		{ 129, 487, 35 },
		{ 129, 488, 35 },
		{ 129, 489, 35 },
		{ 129, 490, 35 },
		{ 129, 491, 35 },
		{ 129, 492, 35 },
		{ 129, 493, 35 },
		{ 129, 494, 35 },
		{ 129, 495, 35 },
		{ 129, 496, 35 },
		{ 129, 497, 35 },
		{ 129, 498, 35 },
		{ 129, 499, 35 },
		{ 129, 500, 35 },
		{ 129, 501, 35 },
		{ 129, 502, 35 },
		{ 129, 503, 35 },
		{ 129, 504, 35 },
		{ 129, 505, 35 },
		{ 129, 506, 35 },
		{ 129, 507, 35 },
		{ 129, 508, 35 },
		{ 129, 509, 35 },
		{ 129, 510, 35 },
		{ 129, 511, 35 },
		{ 129, 512, 35 },
		{ 130, 427, 35 },
		{ 130, 428, 35 },
		{ 130, 429, 35 },
		{ 130, 430, 35 },
		{ 130, 431, 35 },
		{ 130, 432, 35 },
		{ 130, 433, 35 },
		{ 130, 434, 35 },
		{ 130, 435, 35 },
		{ 130, 436, 35 },
		{ 130, 437, 35 },
		{ 130, 438, 35 },
		{ 130, 439, 35 },
		{ 130, 440, 35 },
		{ 130, 441, 35 },
		{ 130, 442, 35 },
		{ 130, 443, 35 },
		{ 130, 444, 35 },
		{ 130, 445, 35 },
		{ 130, 446, 35 },
		{ 130, 447, 35 },
		{ 130, 448, 35 },
		{ 130, 449, 35 },
		{ 130, 450, 35 },
		{ 130, 451, 35 },
		{ 130, 452, 35 },
		{ 130, 453, 35 },
		{ 130, 454, 35 },
		{ 130, 455, 35 },
		{ 130, 456, 35 },
		{ 130, 457, 35 },
		{ 130, 458, 35 },
		{ 130, 459, 35 },
		{ 130, 460, 35 },
		{ 130, 461, 35 },
		{ 130, 462, 35 },
		{ 130, 463, 35 },
		{ 130, 464, 35 },
		{ 130, 465, 35 },
		{ 130, 466, 35 },
		{ 130, 467, 35 },
		{ 130, 468, 35 },
		{ 130, 469, 35 },
		{ 130, 470, 35 },
		{ 130, 471, 35 },
		{ 130, 472, 35 },
		{ 130, 473, 35 },
		{ 130, 474, 35 },
		{ 130, 475, 35 },
		{ 130, 476, 35 },
		{ 130, 477, 35 },
		{ 130, 478, 35 },
		{ 130, 479, 35 },
		{ 130, 480, 35 },
		{ 130, 481, 35 },
		{ 130, 482, 35 },
		{ 130, 483, 35 },
		{ 130, 484, 35 },
		{ 130, 485, 35 },
		{ 130, 486, 35 },
		{ 130, 487, 35 },
		{ 130, 488, 35 },
		{ 130, 489, 35 },
		{ 130, 490, 35 },
		{ 130, 491, 35 },
		{ 130, 492, 35 },
		{ 130, 493, 35 },
		{ 130, 494, 35 },
		{ 130, 495, 35 },
		{ 130, 496, 35 },
		{ 130, 497, 35 },
		{ 130, 498, 35 },
		{ 130, 499, 35 },
		{ 130, 500, 35 },
		{ 130, 501, 35 },
		{ 130, 502, 35 },
		{ 130, 503, 35 },
		{ 130, 504, 35 },
		{ 130, 505, 35 },
		{ 130, 506, 35 },
		{ 130, 507, 35 },
		{ 130, 508, 35 },
		{ 130, 509, 35 },
		{ 130, 510, 35 },
		{ 131, 426, 35 },
		{ 131, 427, 35 },
		{ 131, 428, 35 },
		{ 131, 429, 35 },
		{ 131, 430, 35 },
		{ 131, 431, 35 },
		{ 131, 432, 35 },
		{ 131, 433, 35 },
		{ 131, 434, 35 },
		{ 131, 435, 35 },
		{ 131, 436, 35 },
		{ 131, 437, 35 },
		{ 131, 438, 35 },
		{ 131, 439, 35 },
		{ 131, 440, 35 },
		{ 131, 441, 35 },
		{ 131, 442, 35 },
		{ 131, 443, 35 },
		{ 131, 444, 35 },
		{ 131, 445, 35 },
		{ 131, 446, 35 },
		{ 131, 447, 35 },
		{ 131, 448, 35 },
		{ 131, 449, 35 },
		{ 131, 450, 35 },
		{ 131, 451, 35 },
		{ 131, 452, 35 },
		{ 131, 453, 35 },
		{ 131, 454, 35 },
		{ 131, 455, 35 },
		{ 131, 456, 35 },
		{ 131, 457, 35 },
		{ 131, 458, 35 },
		{ 131, 459, 35 },
		{ 131, 460, 35 },
		{ 131, 461, 35 },
		{ 131, 462, 35 },
		{ 131, 463, 35 },
		{ 131, 464, 35 },
		{ 131, 465, 35 },
		{ 131, 466, 35 },
		{ 131, 467, 35 },
		{ 131, 468, 35 },
		{ 131, 469, 35 },
		{ 131, 470, 35 },
		{ 131, 471, 35 },
		{ 131, 472, 35 },
		{ 131, 473, 35 },
		{ 131, 474, 35 },
		{ 131, 475, 35 },
		{ 131, 476, 35 },
		{ 131, 477, 35 },
		{ 131, 478, 35 },
		{ 131, 479, 35 },
		{ 131, 480, 35 },
		{ 131, 481, 35 },
		{ 131, 482, 35 },
		{ 131, 483, 35 },
		{ 131, 484, 35 },
		{ 131, 485, 35 },
		{ 131, 486, 35 },
		{ 131, 487, 35 },
		{ 131, 488, 35 },
		{ 131, 489, 35 },
		{ 131, 490, 35 },
		{ 131, 491, 35 },
		{ 131, 492, 35 },
		{ 131, 493, 35 },
		{ 131, 494, 35 },
		{ 131, 495, 35 },
		{ 131, 496, 35 },
		{ 131, 497, 35 },
		{ 131, 498, 35 },
		{ 131, 499, 35 },
		{ 131, 500, 35 },
		{ 131, 501, 35 },
		{ 131, 502, 35 },
		{ 131, 503, 35 },
		{ 131, 504, 35 },
		{ 131, 505, 35 },
		{ 131, 506, 35 },
		{ 131, 507, 35 },
		{ 131, 508, 35 },
		{ 132, 424, 35 },
		{ 132, 425, 35 },
		{ 132, 426, 35 },
		{ 132, 427, 35 },
		{ 132, 428, 35 },
		{ 132, 429, 35 },
		{ 132, 430, 35 },
		{ 132, 431, 35 },
		{ 132, 432, 35 },
		{ 132, 433, 35 },
		{ 132, 434, 35 },
		{ 132, 435, 35 },
		{ 132, 436, 35 },
		{ 132, 437, 35 },
		{ 132, 438, 35 },
		{ 132, 439, 35 },
		{ 132, 440, 35 },
		{ 132, 441, 35 },
		{ 132, 442, 35 },
		{ 132, 443, 35 },
		{ 132, 444, 35 },
		{ 132, 445, 35 },
		{ 132, 446, 35 },
		{ 132, 447, 35 },
		{ 132, 448, 35 },
		{ 132, 449, 35 },
		{ 132, 450, 35 },
		{ 132, 451, 35 },
		{ 132, 452, 35 },
		{ 132, 453, 35 },
		{ 132, 454, 35 },
		{ 132, 455, 35 },
		{ 132, 456, 35 },
		{ 132, 457, 35 },
		{ 132, 458, 35 },
		{ 132, 459, 35 },
		{ 132, 460, 35 },
		{ 132, 461, 35 },
		{ 132, 462, 35 },
		{ 132, 463, 35 },
		{ 132, 464, 35 },
		{ 132, 465, 35 },
		{ 132, 466, 35 },
		{ 132, 467, 35 },
		{ 132, 468, 35 },
		{ 132, 469, 35 },
		{ 132, 470, 35 },
		{ 132, 471, 35 },
		{ 132, 472, 35 },
		{ 132, 473, 35 },
		{ 132, 474, 35 },
		{ 132, 475, 35 },
		{ 132, 476, 35 },
		{ 132, 477, 35 },
		{ 132, 478, 35 },
		{ 132, 479, 35 },
		{ 132, 480, 35 },
		{ 132, 481, 35 },
		{ 132, 482, 35 },
		{ 132, 483, 35 },
		{ 132, 484, 35 },
		{ 132, 485, 35 },
		{ 132, 486, 35 },
		{ 132, 487, 35 },
		{ 132, 488, 35 },
		{ 132, 489, 35 },
		{ 132, 490, 35 },
		{ 132, 491, 35 },
		{ 132, 492, 35 },
		{ 132, 493, 35 },
		{ 132, 494, 35 },
		{ 132, 495, 35 },
		{ 132, 496, 35 },
		{ 132, 497, 35 },
		{ 132, 498, 35 },
		{ 132, 499, 35 },
		{ 132, 500, 35 },
		{ 132, 501, 35 },
		{ 132, 502, 35 },
		{ 132, 503, 35 },
		{ 132, 504, 35 },
		{ 132, 505, 35 },
		{ 132, 506, 35 },
		{ 133, 423, 35 },
		{ 133, 424, 35 },
		{ 133, 425, 35 },
		{ 133, 426, 35 },
		{ 133, 427, 35 },
		{ 133, 428, 35 },
		{ 133, 429, 35 },
		{ 133, 430, 35 },
		{ 133, 431, 35 },
		{ 133, 432, 35 },
		{ 133, 433, 35 },
		{ 133, 434, 35 },
		{ 133, 435, 35 },
		{ 133, 436, 35 },
		{ 133, 437, 35 },
		{ 133, 438, 35 },
		{ 133, 439, 35 },
		{ 133, 440, 35 },
		{ 133, 441, 35 },
		{ 133, 442, 35 },
		{ 133, 443, 35 },
		{ 133, 444, 35 },
		{ 133, 445, 35 },
		{ 133, 446, 35 },
		{ 133, 447, 35 },
		{ 133, 448, 35 },
		{ 133, 449, 35 },
		{ 133, 450, 35 },
		{ 133, 451, 35 },
		{ 133, 452, 35 },
		{ 133, 453, 35 },
		{ 133, 454, 35 },
		{ 133, 455, 35 },
		{ 133, 456, 35 },
		{ 133, 457, 35 },
		{ 133, 458, 35 },
		{ 133, 459, 35 },
		{ 133, 460, 35 },
		{ 133, 461, 35 },
		{ 133, 462, 35 },
		{ 133, 463, 35 },
		{ 133, 464, 35 },
		{ 133, 465, 35 },
		{ 133, 466, 35 },
		{ 133, 467, 35 },
		{ 133, 468, 35 },
		{ 133, 469, 35 },
		{ 133, 470, 35 },
		{ 133, 471, 35 },
		{ 133, 472, 35 },
		{ 133, 473, 35 },
		{ 133, 474, 35 },
		{ 133, 475, 35 },
		{ 133, 476, 35 },
		{ 133, 477, 35 },
		{ 133, 478, 35 },
		{ 133, 479, 35 },
		{ 133, 480, 35 },
		{ 133, 481, 35 },
		{ 133, 482, 35 },
		{ 133, 483, 35 },
		{ 133, 484, 35 },
		{ 133, 485, 35 },
		{ 133, 486, 35 },
		{ 133, 487, 35 },
		{ 133, 488, 35 },
		{ 133, 489, 35 },
		{ 133, 490, 35 },
		{ 133, 491, 35 },
		{ 133, 492, 35 },
		{ 133, 493, 35 },
		{ 133, 494, 35 },
		{ 133, 495, 35 },
		{ 133, 496, 35 },
		{ 133, 497, 35 },
		{ 133, 498, 35 },
		{ 133, 499, 35 },
		{ 133, 500, 35 },
		{ 133, 501, 35 },
		{ 133, 502, 35 },
		{ 133, 503, 35 },
		{ 133, 504, 35 },
		{ 134, 422, 35 },
		{ 134, 423, 35 },
		{ 134, 424, 35 },
		{ 134, 425, 35 },
		{ 134, 426, 35 },
		{ 134, 427, 35 },
		{ 134, 428, 35 },
		{ 134, 429, 35 },
		{ 134, 430, 35 },
		{ 134, 431, 35 },
		{ 134, 432, 35 },
		{ 134, 433, 35 },
		{ 134, 434, 35 },
		{ 134, 435, 35 },
		{ 134, 436, 35 },
		{ 134, 437, 35 },
		{ 134, 438, 35 },
		{ 134, 439, 35 },
		{ 134, 440, 35 },
		{ 134, 441, 35 },
		{ 134, 442, 35 },
		{ 134, 443, 35 },
		{ 134, 444, 35 },
		{ 134, 445, 35 },
		{ 134, 446, 35 },
		{ 134, 447, 35 },
		{ 134, 448, 35 },
		{ 134, 449, 35 },
		{ 134, 450, 35 },
		{ 134, 451, 35 },
		{ 134, 452, 35 },
		{ 134, 453, 35 },
		{ 134, 454, 35 },
		{ 134, 455, 35 },
		{ 134, 456, 35 },
		{ 134, 457, 35 },
		{ 134, 458, 35 },
		{ 134, 459, 35 },
		{ 134, 460, 35 },
		{ 134, 461, 35 },
		{ 134, 462, 35 },
		{ 134, 463, 35 },
		{ 134, 464, 35 },
		{ 134, 465, 35 },
		{ 134, 466, 35 },
		{ 134, 467, 35 },
		{ 134, 468, 35 },
		{ 134, 469, 35 },
		{ 134, 470, 35 },
		{ 134, 471, 35 },
		{ 134, 472, 35 },
		{ 134, 473, 35 },
		{ 134, 474, 35 },
		{ 134, 475, 35 },
		{ 134, 476, 35 },
		{ 134, 477, 35 },
		{ 134, 478, 35 },
		{ 134, 479, 35 },
		{ 134, 480, 35 },
		{ 134, 481, 35 },
		{ 134, 482, 35 },
		{ 134, 483, 35 },
		{ 134, 484, 35 },
		{ 134, 485, 35 },
		{ 134, 486, 35 },
		{ 134, 487, 35 },
		{ 134, 488, 35 },
		{ 134, 489, 35 },
		{ 134, 490, 35 },
		{ 134, 491, 35 },
		{ 134, 492, 35 },
		{ 134, 493, 35 },
		{ 134, 494, 35 },
		{ 134, 495, 35 },
		{ 134, 496, 35 },
		{ 134, 497, 35 },
		{ 134, 498, 35 },
		{ 134, 499, 35 },
		{ 134, 500, 35 },
		{ 134, 501, 35 },
		{ 134, 502, 35 },
		{ 135, 420, 35 },
		{ 135, 421, 35 },
		{ 135, 422, 35 },
		{ 135, 423, 35 },
		{ 135, 424, 35 },
		{ 135, 425, 35 },
		{ 135, 426, 35 },
		{ 135, 427, 35 },
		{ 135, 428, 35 },
		{ 135, 429, 35 },
		{ 135, 430, 35 },
		{ 135, 431, 35 },
		{ 135, 432, 35 },
		{ 135, 433, 35 },
		{ 135, 434, 35 },
		{ 135, 435, 35 },
		{ 135, 436, 35 },
		{ 135, 437, 35 },
		{ 135, 438, 35 },
		{ 135, 439, 35 },
		{ 135, 440, 35 },
		{ 135, 441, 35 },
		{ 135, 442, 35 },
		{ 135, 443, 35 },
		{ 135, 444, 35 },
		{ 135, 445, 35 },
		{ 135, 446, 35 },
		{ 135, 447, 35 },
		{ 135, 448, 35 },
		{ 135, 449, 35 },
		{ 135, 450, 35 },
		{ 135, 451, 35 },
		{ 135, 452, 35 },
		{ 135, 453, 35 },
		{ 135, 454, 35 },
		{ 135, 455, 35 },
		{ 135, 456, 35 },
		{ 135, 457, 35 },
		{ 135, 458, 35 },
		{ 135, 459, 35 },
		{ 135, 460, 35 },
		{ 135, 461, 35 },
		{ 135, 462, 35 },
		{ 135, 463, 35 },
		{ 135, 464, 35 },
		{ 135, 465, 35 },
		{ 135, 466, 35 },
		{ 135, 467, 35 },
		{ 135, 468, 35 },
		{ 135, 469, 35 },
		{ 135, 470, 35 },
		{ 135, 471, 35 },
		{ 135, 472, 35 },
		{ 135, 473, 35 },
		{ 135, 474, 35 },
		{ 135, 475, 35 },
		{ 135, 476, 35 },
		{ 135, 477, 35 },
		{ 135, 478, 35 },
		{ 135, 479, 35 },
		{ 135, 480, 35 },
		{ 135, 481, 35 },
		{ 135, 482, 35 },
		{ 135, 483, 35 },
		{ 135, 484, 35 },
		{ 135, 485, 35 },
		{ 135, 486, 35 },
		{ 135, 487, 35 },
		{ 135, 488, 35 },
		{ 135, 489, 35 },
		{ 135, 490, 35 },
		{ 135, 491, 35 },
		{ 135, 492, 35 },
		{ 135, 493, 35 },
		{ 135, 494, 35 },
		{ 135, 495, 35 },
		{ 135, 496, 35 },
		{ 135, 497, 35 },
		{ 135, 498, 35 },
		{ 135, 499, 35 },
		{ 135, 500, 35 },
		{ 136, 419, 35 },
		{ 136, 420, 35 },
		{ 136, 421, 35 },
		{ 136, 422, 35 },
		{ 136, 423, 35 },
		{ 136, 424, 35 },
		{ 136, 425, 35 },
		{ 136, 426, 35 },
		{ 136, 427, 35 },
		{ 136, 428, 35 },
		{ 136, 429, 35 },
		{ 136, 430, 35 },
		{ 136, 431, 35 },
		{ 136, 432, 35 },
		{ 136, 433, 35 },
		{ 136, 434, 35 },
		{ 136, 435, 35 },
		{ 136, 436, 35 },
		{ 136, 437, 35 },
		{ 136, 438, 35 },
		{ 136, 439, 35 },
		{ 136, 440, 35 },
		{ 136, 441, 35 },
		{ 136, 442, 35 },
		{ 136, 443, 35 },
		{ 136, 444, 35 },
		{ 136, 445, 35 },
		{ 136, 446, 35 },
		{ 136, 447, 35 },
		{ 136, 448, 35 },
		{ 136, 449, 35 },
		{ 136, 450, 35 },
		{ 136, 451, 35 },
		{ 136, 452, 35 },
		{ 136, 453, 35 },
		{ 136, 454, 35 },
		{ 136, 455, 35 },
		{ 136, 456, 35 },
		{ 136, 457, 35 },
		{ 136, 458, 35 },
		{ 136, 459, 35 },
		{ 136, 460, 35 },
		{ 136, 461, 35 },
		{ 136, 462, 35 },
		{ 136, 463, 35 },
		{ 136, 464, 35 },
		{ 136, 465, 35 },
		{ 136, 466, 35 },
		{ 136, 467, 35 },
		{ 136, 468, 35 },
		{ 136, 469, 35 },
		{ 136, 470, 35 },
		{ 136, 471, 35 },
		{ 136, 472, 35 },
		{ 136, 473, 35 },
		{ 136, 474, 35 },
		{ 136, 475, 35 },
		{ 136, 476, 35 },
		{ 136, 477, 35 },
		{ 136, 478, 35 },
		{ 136, 479, 35 },
		{ 136, 480, 35 },
		{ 136, 481, 35 },
		{ 136, 482, 35 },
		{ 136, 483, 35 },
		{ 136, 484, 35 },
		{ 136, 485, 35 },
		{ 136, 486, 35 },
		{ 136, 487, 35 },
		{ 136, 488, 35 },
		{ 136, 489, 35 },
		{ 136, 490, 35 },
		{ 136, 491, 35 },
		{ 136, 492, 35 },
		{ 136, 493, 35 },
		{ 136, 494, 35 },
		{ 136, 495, 35 },
		{ 136, 496, 35 },
		{ 136, 497, 35 },
		{ 136, 498, 35 },
		{ 137, 418, 35 },
		{ 137, 419, 35 },
		{ 137, 420, 35 },
		{ 137, 421, 35 },
		{ 137, 422, 35 },
		{ 137, 423, 35 },
		{ 137, 424, 35 },
		{ 137, 425, 35 },
		{ 137, 426, 35 },
		{ 137, 427, 35 },
		{ 137, 428, 35 },
		{ 137, 429, 35 },
		{ 137, 430, 35 },
		{ 137, 431, 35 },
		{ 137, 432, 35 },
		{ 137, 433, 35 },
		{ 137, 434, 35 },
		{ 137, 435, 35 },
		{ 137, 436, 35 },
		{ 137, 437, 35 },
		{ 137, 438, 35 },
		{ 137, 439, 35 },
		{ 137, 440, 35 },
		{ 137, 441, 35 },
		{ 137, 442, 35 },
		{ 137, 443, 35 },
		{ 137, 444, 35 },
		{ 137, 445, 35 },
		{ 137, 446, 35 },
		{ 137, 447, 35 },
		{ 137, 448, 35 },
		{ 137, 449, 35 },
		{ 137, 450, 35 },
		{ 137, 451, 35 },
		{ 137, 452, 35 },
		{ 137, 453, 35 },
		{ 137, 454, 35 },
		{ 137, 455, 35 },
		{ 137, 456, 35 },
		{ 137, 457, 35 },
		{ 137, 458, 35 },
		{ 137, 459, 35 },
		{ 137, 460, 35 },
		{ 137, 461, 35 },
		{ 137, 462, 35 },
		{ 137, 463, 35 },
		{ 137, 464, 35 },
		{ 137, 465, 35 },
		{ 137, 466, 35 },
		{ 137, 467, 35 },
		{ 137, 468, 35 },
		{ 137, 469, 35 },
		{ 137, 470, 35 },
		{ 137, 471, 35 },
		{ 137, 472, 35 },
		{ 137, 473, 35 },
		{ 137, 474, 35 },
		{ 137, 475, 35 },
		{ 137, 476, 35 },
		{ 137, 477, 35 },
		{ 137, 478, 35 },
		{ 137, 479, 35 },
		{ 137, 480, 35 },
		{ 137, 481, 35 },
		{ 137, 482, 35 },
		{ 137, 483, 35 },
		{ 137, 484, 35 },
		{ 137, 485, 35 },
		{ 137, 486, 35 },
		{ 137, 487, 35 },
		{ 137, 488, 35 },
		{ 137, 489, 35 },
		{ 137, 490, 35 },
		{ 137, 491, 35 },
		{ 137, 492, 35 },
		{ 137, 493, 35 },
		{ 137, 494, 35 },
		{ 137, 495, 35 },
		{ 137, 496, 35 },
		{ 138, 416, 35 },
		{ 138, 417, 35 },
		{ 138, 418, 35 },
		{ 138, 419, 35 },
		{ 138, 420, 35 },
		{ 138, 421, 35 },
		{ 138, 422, 35 },
		{ 138, 423, 35 },
		{ 138, 424, 35 },
		{ 138, 425, 35 },
		{ 138, 426, 35 },
		{ 138, 427, 35 },
		{ 138, 428, 35 },
		{ 138, 429, 35 },
		{ 138, 430, 35 },
		{ 138, 431, 35 },
		{ 138, 432, 35 },
		{ 138, 433, 35 },
		{ 138, 434, 35 },
		{ 138, 435, 35 },
		{ 138, 436, 35 },
		{ 138, 437, 35 },
		{ 138, 438, 35 },
		{ 138, 439, 35 },
		{ 138, 440, 35 },
		{ 138, 441, 35 },
		{ 138, 442, 35 },
		{ 138, 443, 35 },
		{ 138, 444, 35 },
		{ 138, 445, 35 },
		{ 138, 446, 35 },
		{ 138, 447, 35 },
		{ 138, 448, 35 },
		{ 138, 449, 35 },
		{ 138, 450, 35 },
		{ 138, 451, 35 },
		{ 138, 452, 35 },
		{ 138, 453, 35 },
		{ 138, 454, 35 },
		{ 138, 455, 35 },
		{ 138, 456, 35 },
		{ 138, 457, 35 },
		{ 138, 458, 35 },
		{ 138, 459, 35 },
		{ 138, 460, 35 },
		{ 138, 461, 35 },
		{ 138, 462, 35 },
		{ 138, 463, 35 },
		{ 138, 464, 35 },
		{ 138, 465, 35 },
		{ 138, 466, 35 },
		{ 138, 467, 35 },
		{ 138, 468, 35 },
		{ 138, 469, 35 },
		{ 138, 470, 35 },
		{ 138, 471, 35 },
		{ 138, 472, 35 },
		{ 138, 473, 35 },
		{ 138, 474, 35 },
		{ 138, 475, 35 },
		{ 138, 476, 35 },
		{ 138, 477, 35 },
		{ 138, 478, 35 },
		{ 138, 479, 35 },
		{ 138, 480, 35 },
		{ 138, 481, 35 },
		{ 138, 482, 35 },
		{ 138, 483, 35 },
		{ 138, 484, 35 },
		{ 138, 485, 35 },
		{ 138, 486, 35 },
		{ 138, 487, 35 },
		{ 138, 488, 35 },
		{ 138, 489, 35 },
		{ 138, 490, 35 },
		{ 138, 491, 35 },
		{ 138, 492, 35 },
		{ 138, 493, 35 },
		{ 138, 494, 35 },
		{ 139, 415, 35 },
		{ 139, 416, 35 },
		{ 139, 417, 35 },
		{ 139, 418, 35 },
		{ 139, 419, 35 },
		{ 139, 420, 35 },
		{ 139, 421, 35 },
		{ 139, 422, 35 },
		{ 139, 423, 35 },
		{ 139, 424, 35 },
		{ 139, 425, 35 },
		{ 139, 426, 35 },
		{ 139, 427, 35 },
		{ 139, 428, 35 },
		{ 139, 429, 35 },
		{ 139, 430, 35 },
		{ 139, 431, 35 },
		{ 139, 432, 35 },
		{ 139, 433, 35 },
		{ 139, 434, 35 },
		{ 139, 435, 35 },
		{ 139, 436, 35 },
		{ 139, 437, 35 },
		{ 139, 438, 35 },
		{ 139, 439, 35 },
		{ 139, 440, 35 },
		{ 139, 441, 35 },
		{ 139, 442, 35 },
		{ 139, 443, 35 },
		{ 139, 444, 35 },
		{ 139, 445, 35 },
		{ 139, 446, 35 },
		{ 139, 447, 35 },
		{ 139, 448, 35 },
		{ 139, 449, 35 },
		{ 139, 450, 35 },
		{ 139, 451, 35 },
		{ 139, 452, 35 },
		{ 139, 453, 35 },
		{ 139, 454, 35 },
		{ 139, 455, 35 },
		{ 139, 456, 35 },
		{ 139, 457, 35 },
		{ 139, 458, 35 },
		{ 139, 459, 35 },
		{ 139, 460, 35 },
		{ 139, 461, 35 },
		{ 139, 462, 35 },
		{ 139, 463, 35 },
		{ 139, 464, 35 },
		{ 139, 465, 35 },
		{ 139, 466, 35 },
		{ 139, 467, 35 },
		{ 139, 468, 35 },
		{ 139, 469, 35 },
		{ 139, 470, 35 },
		{ 139, 471, 35 },
		{ 139, 472, 35 },
		{ 139, 473, 35 },
		{ 139, 474, 35 },
		{ 139, 475, 35 },
		{ 139, 476, 35 },
		{ 139, 477, 35 },
		{ 139, 478, 35 },
		{ 139, 479, 35 },
		{ 139, 480, 35 },
		{ 139, 481, 35 },
		{ 139, 482, 35 },
		{ 139, 483, 35 },
		{ 139, 484, 35 },
		{ 139, 485, 35 },
		{ 139, 486, 35 },
		{ 139, 487, 35 },
		{ 139, 488, 35 },
		{ 139, 489, 35 },
		{ 139, 490, 35 },
		{ 139, 491, 35 },
		{ 139, 492, 35 },
		{ 140, 414, 35 },
		{ 140, 415, 35 },
		{ 140, 416, 35 },
		{ 140, 417, 35 },
		{ 140, 418, 35 },
		{ 140, 419, 35 },
		{ 140, 420, 35 },
		{ 140, 421, 35 },
		{ 140, 422, 35 },
		{ 140, 423, 35 },
		{ 140, 424, 35 },
		{ 140, 425, 35 },
		{ 140, 426, 35 },
		{ 140, 427, 35 },
		{ 140, 428, 35 },
		{ 140, 429, 35 },
		{ 140, 430, 35 },
		{ 140, 431, 35 },
		{ 140, 432, 35 },
		{ 140, 433, 35 },
		{ 140, 434, 35 },
		{ 140, 435, 35 },
		{ 140, 436, 35 },
		{ 140, 437, 35 },
		{ 140, 438, 35 },
		{ 140, 439, 35 },
		{ 140, 440, 35 },
		{ 140, 441, 35 },
		{ 140, 442, 35 },
		{ 140, 443, 35 },
		{ 140, 444, 35 },
		{ 140, 445, 35 },
		{ 140, 446, 35 },
		{ 140, 447, 35 },
		{ 140, 448, 35 },
		{ 140, 449, 35 },
		{ 140, 450, 35 },
		{ 140, 451, 35 },
		{ 140, 452, 35 },
		{ 140, 453, 35 },
		{ 140, 454, 35 },
		{ 140, 455, 35 },
		{ 140, 456, 35 },
		{ 140, 457, 35 },
		{ 140, 458, 35 },
		{ 140, 459, 35 },
		{ 140, 460, 35 },
		{ 140, 461, 35 },
		{ 140, 462, 35 },
		{ 140, 463, 35 },
		{ 140, 464, 35 },
		{ 140, 465, 35 },
		{ 140, 466, 35 },
		{ 140, 467, 35 },
		{ 140, 468, 35 },
		{ 140, 469, 35 },
		{ 140, 470, 35 },
		{ 140, 471, 35 },
		{ 140, 472, 35 },
		{ 140, 473, 35 },
		{ 140, 474, 35 },
		{ 140, 475, 35 },
		{ 140, 476, 35 },
		{ 140, 477, 35 },
		{ 140, 478, 35 },
		{ 140, 479, 35 },
		{ 140, 480, 35 },
		{ 140, 481, 35 },
		{ 140, 482, 35 },
		{ 140, 483, 35 },
		{ 140, 484, 35 },
		{ 140, 485, 35 },
		{ 140, 486, 35 },
		{ 140, 487, 35 },
		{ 140, 488, 35 },
		{ 140, 489, 35 },
		{ 140, 490, 35 },
		{ 141, 412, 35 },
		{ 141, 413, 35 },
		{ 141, 414, 35 },
		{ 141, 415, 35 },
		{ 141, 416, 35 },
		{ 141, 417, 35 },
		{ 141, 418, 35 },
		{ 141, 419, 35 },
		{ 141, 420, 35 },
		{ 141, 421, 35 },
		{ 141, 422, 35 },
		{ 141, 423, 35 },
		{ 141, 424, 35 },
		{ 141, 425, 35 },
		{ 141, 426, 35 },
		{ 141, 427, 35 },
		{ 141, 428, 35 },
		{ 141, 429, 35 },
		{ 141, 430, 35 },
		{ 141, 431, 35 },
		{ 141, 432, 35 },
		{ 141, 433, 35 },
		{ 141, 434, 35 },
		{ 141, 435, 35 },
		{ 141, 436, 35 },
		{ 141, 437, 35 },
		{ 141, 438, 35 },
		{ 141, 439, 35 },
		{ 141, 440, 35 },
		{ 141, 441, 35 },
		{ 141, 442, 35 },
		{ 141, 443, 35 },
		{ 141, 444, 35 },
		{ 141, 445, 35 },
		{ 141, 446, 35 },
		{ 141, 447, 35 },
		{ 141, 448, 35 },
		{ 141, 449, 35 },
		{ 141, 450, 35 },
		{ 141, 451, 35 },
		{ 141, 452, 35 },
		{ 141, 453, 35 },
		{ 141, 454, 35 },
		{ 141, 455, 35 },
		{ 141, 456, 35 },
		{ 141, 457, 35 },
		{ 141, 458, 35 },
		{ 141, 459, 35 },
		{ 141, 460, 35 },
		{ 141, 461, 35 },
		{ 141, 462, 35 },
		{ 141, 463, 35 },
		{ 141, 464, 35 },
		{ 141, 465, 35 },
		{ 141, 466, 35 },
		{ 141, 467, 35 },
		{ 141, 468, 35 },
		{ 141, 469, 35 },
		{ 141, 470, 35 },
		{ 141, 471, 35 },
		{ 141, 472, 35 },
		{ 141, 473, 35 },
		{ 141, 474, 35 },
		{ 141, 475, 35 },
		{ 141, 476, 35 },
		{ 141, 477, 35 },
		{ 141, 478, 35 },
		{ 141, 479, 35 },
		{ 141, 480, 35 },
		{ 141, 481, 35 },
		{ 141, 482, 35 },
		{ 141, 483, 35 },
		{ 141, 484, 35 },
		{ 141, 485, 35 },
		{ 141, 486, 35 },
		{ 141, 487, 35 },
		{ 141, 488, 35 },
		{ 142, 411, 35 },
		{ 142, 412, 35 },
		{ 142, 413, 35 },
		{ 142, 414, 35 },
		{ 142, 415, 35 },
		{ 142, 416, 35 },
		{ 142, 417, 35 },
		{ 142, 418, 35 },
		{ 142, 419, 35 },
		{ 142, 420, 35 },
		{ 142, 421, 35 },
		{ 142, 422, 35 },
		{ 142, 423, 35 },
		{ 142, 424, 35 },
		{ 142, 425, 35 },
		{ 142, 426, 35 },
		{ 142, 427, 35 },
		{ 142, 428, 35 },
		{ 142, 429, 35 },
		{ 142, 430, 35 },
		{ 142, 431, 35 },
		{ 142, 432, 35 },
		{ 142, 433, 35 },
		{ 142, 434, 35 },
		{ 142, 435, 35 },
		{ 142, 436, 35 },
		{ 142, 437, 35 },
		{ 142, 438, 35 },
		{ 142, 439, 35 },
		{ 142, 440, 35 },
		{ 142, 441, 35 },
		{ 142, 442, 35 },
		{ 142, 443, 35 },
		{ 142, 444, 35 },
		{ 142, 445, 35 },
		{ 142, 446, 35 },
		{ 142, 447, 35 },
		{ 142, 448, 35 },
		{ 142, 449, 35 },
		{ 142, 450, 35 },
		{ 142, 451, 35 },
		{ 142, 452, 35 },
		{ 142, 453, 35 },
		{ 142, 454, 35 },
		{ 142, 455, 35 },
		{ 142, 456, 35 },
		{ 142, 457, 35 },
		{ 142, 458, 35 },
		{ 142, 459, 35 },
		{ 142, 460, 35 },
		{ 142, 461, 35 },
		{ 142, 462, 35 },
		{ 142, 463, 35 },
		{ 142, 464, 35 },
		{ 142, 465, 35 },
		{ 142, 466, 35 },
		{ 142, 467, 35 },
		{ 142, 468, 35 },
		{ 142, 469, 35 },
		{ 142, 470, 35 },
		{ 142, 471, 35 },
		{ 142, 472, 35 },
		{ 142, 473, 35 },
		{ 142, 474, 35 },
		{ 142, 475, 35 },
		{ 142, 476, 35 },
		{ 142, 477, 35 },
		{ 142, 478, 35 },
		{ 142, 479, 35 },
		{ 142, 480, 35 },
		{ 142, 481, 35 },
		{ 142, 482, 35 },
		{ 142, 483, 35 },
		{ 142, 484, 35 },
		{ 142, 485, 35 },
		{ 142, 486, 35 },
		{ 143, 410, 35 },
		{ 143, 411, 35 },
		{ 143, 412, 35 },
		{ 143, 413, 35 },
		{ 143, 414, 35 },
		{ 143, 415, 35 },
		{ 143, 416, 35 },
		{ 143, 417, 35 },
		{ 143, 418, 35 },
		{ 143, 419, 35 },
		{ 143, 420, 35 },
		{ 143, 421, 35 },
		{ 143, 422, 35 },
		{ 143, 423, 35 },
		{ 143, 424, 35 },
		{ 143, 425, 35 },
		{ 143, 426, 35 },
		{ 143, 427, 35 },
		{ 143, 428, 35 },
		{ 143, 429, 35 },
		{ 143, 430, 35 },
		{ 143, 431, 35 },
		{ 143, 432, 35 },
		{ 143, 433, 35 },
		{ 143, 434, 35 },
		{ 143, 435, 35 },
		{ 143, 436, 35 },
		{ 143, 437, 35 },
		{ 143, 438, 35 },
		{ 143, 439, 35 },
		{ 143, 440, 35 },
		{ 143, 441, 35 },
		{ 143, 442, 35 },
		{ 143, 443, 35 },
		{ 143, 444, 35 },
		{ 143, 445, 35 },
		{ 143, 446, 35 },
		{ 143, 447, 35 },
		{ 143, 448, 35 },
		{ 143, 449, 35 },
		{ 143, 450, 35 },
		{ 143, 451, 35 },
		{ 143, 452, 35 },
		{ 143, 453, 35 },
		{ 143, 454, 35 },
		{ 143, 455, 35 },
		{ 143, 456, 35 },
		{ 143, 457, 35 },
		{ 143, 458, 35 },
		{ 143, 459, 35 },
		{ 143, 460, 35 },
		{ 143, 461, 35 },
		{ 143, 462, 35 },
		{ 143, 463, 35 },
		{ 143, 464, 35 },
		{ 143, 465, 35 },
		{ 143, 466, 35 },
		{ 143, 467, 35 },
		{ 143, 468, 35 },
		{ 143, 469, 35 },
		{ 143, 470, 35 },
		{ 143, 471, 35 },
		{ 143, 472, 35 },
		{ 143, 473, 35 },
		{ 143, 474, 35 },
		{ 143, 475, 35 },
		{ 143, 476, 35 },
		{ 143, 477, 35 },
		{ 143, 478, 35 },
		{ 143, 479, 35 },
		{ 143, 480, 35 },
		{ 143, 481, 35 },
		{ 143, 482, 35 },
		{ 143, 483, 35 },
		{ 143, 484, 35 },
		{ 144, 408, 35 },
		{ 144, 409, 35 },
		{ 144, 410, 35 },
		{ 144, 411, 35 },
		{ 144, 412, 35 },
		{ 144, 413, 35 },
		{ 144, 414, 35 },
		{ 144, 415, 35 },
		{ 144, 416, 35 },
		{ 144, 417, 35 },
		{ 144, 418, 35 },
		{ 144, 419, 35 },
		{ 144, 420, 35 },
		{ 144, 421, 35 },
		{ 144, 422, 35 },
		{ 144, 423, 35 },
		{ 144, 424, 35 },
		{ 144, 425, 35 },
		{ 144, 426, 35 },
		{ 144, 427, 35 },
		{ 144, 428, 35 },
		{ 144, 429, 35 },
		{ 144, 430, 35 },
		{ 144, 431, 35 },
		{ 144, 432, 35 },
		{ 144, 433, 35 },
		{ 144, 434, 35 },
		{ 144, 435, 35 },
		{ 144, 436, 35 },
		{ 144, 437, 35 },
		{ 144, 438, 35 },
		{ 144, 439, 35 },
		{ 144, 440, 35 },
		{ 144, 441, 35 },
		{ 144, 442, 35 },
		{ 144, 443, 35 },
		{ 144, 444, 35 },
		{ 144, 445, 35 },
		{ 144, 446, 35 },
		{ 144, 447, 35 },
		{ 144, 448, 35 },
		{ 144, 449, 35 },
		{ 144, 450, 35 },
		{ 144, 451, 35 },
		{ 144, 452, 35 },
		{ 144, 453, 35 },
		{ 144, 454, 35 },
		{ 144, 455, 35 },
		{ 144, 456, 35 },
		{ 144, 457, 35 },
		{ 144, 458, 35 },
		{ 144, 459, 35 },
		{ 144, 460, 35 },
		{ 144, 461, 35 },
		{ 144, 462, 35 },
		{ 144, 463, 35 },
		{ 144, 464, 35 },
		{ 144, 465, 35 },
		{ 144, 466, 35 },
		{ 144, 467, 35 },
		{ 144, 468, 35 },
		{ 144, 469, 35 },
		{ 144, 470, 35 },
		{ 144, 471, 35 },
		{ 144, 472, 35 },
		{ 144, 473, 35 },
		{ 144, 474, 35 },
		{ 144, 475, 35 },
		{ 144, 476, 35 },
		{ 144, 477, 35 },
		{ 144, 478, 35 },
		{ 144, 479, 35 },
		{ 144, 480, 35 },
		{ 144, 481, 35 },
		{ 144, 482, 35 },
		{ 145, 407, 35 },
		{ 145, 408, 35 },
		{ 145, 409, 35 },
		{ 145, 410, 35 },
		{ 145, 411, 35 },
		{ 145, 412, 35 },
		{ 145, 413, 35 },
		{ 145, 414, 35 },
		{ 145, 415, 35 },
		{ 145, 416, 35 },
		{ 145, 417, 35 },
		{ 145, 418, 35 },
		{ 145, 419, 35 },
		{ 145, 420, 35 },
		{ 145, 421, 35 },
		{ 145, 422, 35 },
		{ 145, 423, 35 },
		{ 145, 424, 35 },
		{ 145, 425, 35 },
		{ 145, 426, 35 },
		{ 145, 427, 35 },
		{ 145, 428, 35 },
		{ 145, 429, 35 },
		{ 145, 430, 35 },
		{ 145, 431, 35 },
		{ 145, 432, 35 },
		{ 145, 433, 35 },
		{ 145, 434, 35 },
		{ 145, 435, 35 },
		{ 145, 436, 35 },
		{ 145, 437, 35 },
		{ 145, 438, 35 },
		{ 145, 439, 35 },
		{ 145, 440, 35 },
		{ 145, 441, 35 },
		{ 145, 442, 35 },
		{ 145, 443, 35 },
		{ 145, 444, 35 },
		{ 145, 445, 35 },
		{ 145, 446, 35 },
		{ 145, 447, 35 },
		{ 145, 448, 35 },
		{ 145, 449, 35 },
		{ 145, 450, 35 },
		{ 145, 451, 35 },
		{ 145, 452, 35 },
		{ 145, 453, 35 },
		{ 145, 454, 35 },
		{ 145, 455, 35 },
		{ 145, 456, 35 },
		{ 145, 457, 35 },
		{ 145, 458, 35 },
		{ 145, 459, 35 },
		{ 145, 460, 35 },
		{ 145, 461, 35 },
		{ 145, 462, 35 },
		{ 145, 463, 35 },
		{ 145, 464, 35 },
		{ 145, 465, 35 },
		{ 145, 466, 35 },
		{ 145, 467, 35 },
		{ 145, 468, 35 },
		{ 145, 469, 35 },
		{ 145, 470, 35 },
		{ 145, 471, 35 },
		{ 145, 472, 35 },
		{ 145, 473, 35 },
		{ 145, 474, 35 },
		{ 145, 475, 35 },
		{ 145, 476, 35 },
		{ 145, 477, 35 },
		{ 145, 478, 35 },
		{ 145, 479, 35 },
		{ 145, 480, 35 },
		{ 146, 406, 35 },
		{ 146, 407, 35 },
		{ 146, 408, 35 },
		{ 146, 409, 35 },
		{ 146, 410, 35 },
		{ 146, 411, 35 },
		{ 146, 412, 35 },
		{ 146, 413, 35 },
		{ 146, 414, 35 },
		{ 146, 415, 35 },
		{ 146, 416, 35 },
		{ 146, 417, 35 },
		{ 146, 418, 35 },
		{ 146, 419, 35 },
		{ 146, 420, 35 },
		{ 146, 421, 35 },
		{ 146, 422, 35 },
		{ 146, 423, 35 },
		{ 146, 424, 35 },
		{ 146, 425, 35 },
		{ 146, 426, 35 },
		{ 146, 427, 35 },
		{ 146, 428, 35 },
		{ 146, 429, 35 },
		{ 146, 430, 35 },
		{ 146, 431, 35 },
		{ 146, 432, 35 },
		{ 146, 433, 35 },
		{ 146, 434, 35 },
		{ 146, 435, 35 },
		{ 146, 436, 35 },
		{ 146, 437, 35 },
		{ 146, 438, 35 },
		{ 146, 439, 35 },
		{ 146, 440, 35 },
		{ 146, 441, 35 },
		{ 146, 442, 35 },
		{ 146, 443, 35 },
		{ 146, 444, 35 },
		{ 146, 445, 35 },
		{ 146, 446, 35 },
		{ 146, 447, 35 },
		{ 146, 448, 35 },
		{ 146, 449, 35 },
		{ 146, 450, 35 },
		{ 146, 451, 35 },
		{ 146, 452, 35 },
		{ 146, 453, 35 },
		{ 146, 454, 35 },
		{ 146, 455, 35 },
		{ 146, 456, 35 },
		{ 146, 457, 35 },
		{ 146, 458, 35 },
		{ 146, 459, 35 },
		{ 146, 460, 35 },
		{ 146, 461, 35 },
		{ 146, 462, 35 },
		{ 146, 463, 35 },
		{ 146, 464, 35 },
		{ 146, 465, 35 },
		{ 146, 466, 35 },
		{ 146, 467, 35 },
		{ 146, 468, 35 },
		{ 146, 469, 35 },
		{ 146, 470, 35 },
		{ 146, 471, 35 },
		{ 146, 472, 35 },
		{ 146, 473, 35 },
		{ 146, 474, 35 },
		{ 146, 475, 35 },
		{ 146, 476, 35 },
		{ 146, 477, 35 },
		{ 146, 478, 35 },
		{ 147, 404, 35 },
		{ 147, 405, 35 },
		{ 147, 406, 35 },
		{ 147, 407, 35 },
		{ 147, 408, 35 },
		{ 147, 409, 35 },
		{ 147, 410, 35 },
		{ 147, 411, 35 },
		{ 147, 412, 35 },
		{ 147, 413, 35 },
		{ 147, 414, 35 },
		{ 147, 415, 35 },
		{ 147, 416, 35 },
		{ 147, 417, 35 },
		{ 147, 418, 35 },
		{ 147, 419, 35 },
		{ 147, 420, 35 },
		{ 147, 421, 35 },
		{ 147, 422, 35 },
		{ 147, 423, 35 },
		{ 147, 424, 35 },
		{ 147, 425, 35 },
		{ 147, 426, 35 },
		{ 147, 427, 35 },
		{ 147, 428, 35 },
		{ 147, 429, 35 },
		{ 147, 430, 35 },
		{ 147, 431, 35 },
		{ 147, 432, 35 },
		{ 147, 433, 35 },
		{ 147, 434, 35 },
		{ 147, 435, 35 },
		{ 147, 436, 35 },
		{ 147, 437, 35 },
		{ 147, 438, 35 },
		{ 147, 439, 35 },
		{ 147, 440, 35 },
		{ 147, 441, 35 },
		{ 147, 442, 35 },
		{ 147, 443, 35 },
		{ 147, 444, 35 },
		{ 147, 445, 35 },
		{ 147, 446, 35 },
		{ 147, 447, 35 },
		{ 147, 448, 35 },
		{ 147, 449, 35 },
		{ 147, 450, 35 },
		{ 147, 451, 35 },
		{ 147, 452, 35 },
		{ 147, 453, 35 },
		{ 147, 454, 35 },
		{ 147, 455, 35 },
		{ 147, 456, 35 },
		{ 147, 457, 35 },
		{ 147, 458, 35 },
		{ 147, 459, 35 },
		{ 147, 460, 35 },
		{ 147, 461, 35 },
		{ 147, 462, 35 },
		{ 147, 463, 35 },
		{ 147, 464, 35 },
		{ 147, 465, 35 },
		{ 147, 466, 35 },
		{ 147, 467, 35 },
		{ 147, 468, 35 },
		{ 147, 469, 35 },
		{ 147, 470, 35 },
		{ 147, 471, 35 },
		{ 147, 472, 35 },
		{ 147, 473, 35 },
		{ 147, 474, 35 },
		{ 147, 475, 35 },
		{ 147, 476, 35 },
		{ 148, 403, 35 },
		{ 148, 404, 35 },
		{ 148, 405, 35 },
		{ 148, 406, 35 },
		{ 148, 407, 35 },
		{ 148, 408, 35 },
		{ 148, 409, 35 },
		{ 148, 410, 35 },
		{ 148, 411, 35 },
		{ 148, 412, 35 },
		{ 148, 413, 35 },
		{ 148, 414, 35 },
		{ 148, 415, 35 },
		{ 148, 416, 35 },
		{ 148, 417, 35 },
		{ 148, 418, 35 },
		{ 148, 419, 35 },
		{ 148, 420, 35 },
		{ 148, 421, 35 },
		{ 148, 422, 35 },
		{ 148, 423, 35 },
		{ 148, 424, 35 },
		{ 148, 425, 35 },
		{ 148, 426, 35 },
		{ 148, 427, 35 },
		{ 148, 428, 35 },
		{ 148, 429, 35 },
		{ 148, 430, 35 },
		{ 148, 431, 35 },
		{ 148, 432, 35 },
		{ 148, 433, 35 },
		{ 148, 434, 35 },
		{ 148, 435, 35 },
		{ 148, 436, 35 },
		{ 148, 437, 35 },
		{ 148, 438, 35 },
		{ 148, 439, 35 },
		{ 148, 440, 35 },
		{ 148, 441, 35 },
		{ 148, 442, 35 },
		{ 148, 443, 35 },
		{ 148, 444, 35 },
		{ 148, 445, 35 },
		{ 148, 446, 35 },
		{ 148, 447, 35 },
		{ 148, 448, 35 },
		{ 148, 449, 35 },
		{ 148, 450, 35 },
		{ 148, 451, 35 },
		{ 148, 452, 35 },
		{ 148, 453, 35 },
		{ 148, 454, 35 },
		{ 148, 455, 35 },
		{ 148, 456, 35 },
		{ 148, 457, 35 },
		{ 148, 458, 35 },
		{ 148, 459, 35 },
		{ 148, 460, 35 },
		{ 148, 461, 35 },
		{ 148, 462, 35 },
		{ 148, 463, 35 },
		{ 148, 464, 35 },
		{ 148, 465, 35 },
		{ 148, 466, 35 },
		{ 148, 467, 35 },
		{ 148, 468, 35 },
		{ 148, 469, 35 },
		{ 148, 470, 35 },
		{ 148, 471, 35 },
		{ 148, 472, 35 },
		{ 148, 473, 35 },
		{ 148, 474, 35 },
		{ 149, 402, 35 },
		{ 149, 403, 35 },
		{ 149, 404, 35 },
		{ 149, 405, 35 },
		{ 149, 406, 35 },
		{ 149, 407, 35 },
		{ 149, 408, 35 },
		{ 149, 409, 35 },
		{ 149, 410, 35 },
		{ 149, 411, 35 },
		{ 149, 412, 35 },
		{ 149, 413, 35 },
		{ 149, 414, 35 },
		{ 149, 415, 35 },
		{ 149, 416, 35 },
		{ 149, 417, 35 },
		{ 149, 418, 35 },
		{ 149, 419, 35 },
		{ 149, 420, 35 },
		{ 149, 421, 35 },
		{ 149, 422, 35 },
		{ 149, 423, 35 },
		{ 149, 424, 35 },
		{ 149, 425, 35 },
		{ 149, 426, 35 },
		{ 149, 427, 35 },
		{ 149, 428, 35 },
		{ 149, 429, 35 },
		{ 149, 430, 35 },
		{ 149, 431, 35 },
		{ 149, 432, 35 },
		{ 149, 433, 35 },
		{ 149, 434, 35 },
		{ 149, 435, 35 },
		{ 149, 436, 35 },
		{ 149, 437, 35 },
		{ 149, 438, 35 },
		{ 149, 439, 35 },
		{ 149, 440, 35 },
		{ 149, 441, 35 },
		{ 149, 442, 35 },
		{ 149, 443, 35 },
		{ 149, 444, 35 },
		{ 149, 445, 35 },
		{ 149, 446, 35 },
		{ 149, 447, 35 },
		{ 149, 448, 35 },
		{ 149, 449, 35 },
		{ 149, 450, 35 },
		{ 149, 451, 35 },
		{ 149, 452, 35 },
		{ 149, 453, 35 },
		{ 149, 454, 35 },
		{ 149, 455, 35 },
		{ 149, 456, 35 },
		{ 149, 457, 35 },
		{ 149, 458, 35 },
		{ 149, 459, 35 },
		{ 149, 460, 35 },
		{ 149, 461, 35 },
		{ 149, 462, 35 },
		{ 149, 463, 35 },
		{ 149, 464, 35 },
		{ 149, 465, 35 },
		{ 149, 466, 35 },
		{ 149, 467, 35 },
		{ 149, 468, 35 },
		{ 149, 469, 35 },
		{ 149, 470, 35 },
		{ 149, 471, 35 },
		{ 149, 472, 35 },
		{ 150, 400, 35 },
		{ 150, 401, 35 },
		{ 150, 402, 35 },
		{ 150, 403, 35 },
		{ 150, 404, 35 },
		{ 150, 405, 35 },
		{ 150, 406, 35 },
		{ 150, 407, 35 },
		{ 150, 408, 35 },
		{ 150, 409, 35 },
		{ 150, 410, 35 },
		{ 150, 411, 35 },
		{ 150, 412, 35 },
		{ 150, 413, 35 },
		{ 150, 414, 35 },
		{ 150, 415, 35 },
		{ 150, 416, 35 },
		{ 150, 417, 35 },
		{ 150, 418, 35 },
		{ 150, 419, 35 },
		{ 150, 420, 35 },
		{ 150, 421, 35 },
		{ 150, 422, 35 },
		{ 150, 423, 35 },
		{ 150, 424, 35 },
		{ 150, 425, 35 },
		{ 150, 426, 35 },
		{ 150, 427, 35 },
		{ 150, 428, 35 },
		{ 150, 429, 35 },
		{ 150, 430, 35 },
		{ 150, 431, 35 },
		{ 150, 432, 35 },
		{ 150, 433, 35 },
		{ 150, 434, 35 },
		{ 150, 435, 35 },
		{ 150, 436, 35 },
		{ 150, 437, 35 },
		{ 150, 438, 35 },
		{ 150, 439, 35 },
		{ 150, 440, 35 },
		{ 150, 441, 35 },
		{ 150, 442, 35 },
		{ 150, 443, 35 },
		{ 150, 444, 35 },
		{ 150, 445, 35 },
		{ 150, 446, 35 },
		{ 150, 447, 35 },
		{ 150, 448, 35 },
		{ 150, 449, 35 },
		{ 150, 450, 35 },
		{ 150, 451, 35 },
		{ 150, 452, 35 },
		{ 150, 453, 35 },
		{ 150, 454, 35 },
		{ 150, 455, 35 },
		{ 150, 456, 35 },
		{ 150, 457, 35 },
		{ 150, 458, 35 },
		{ 150, 459, 35 },
		{ 150, 460, 35 },
		{ 150, 461, 35 },
		{ 150, 462, 35 },
		{ 150, 463, 35 },
		{ 150, 464, 35 },
		{ 150, 465, 35 },
		{ 150, 466, 35 },
		{ 150, 467, 35 },
		{ 150, 468, 35 },
		{ 150, 469, 35 },
		{ 150, 470, 35 },
		{ 151, 399, 35 },
		{ 151, 400, 35 },
		{ 151, 401, 35 },
		{ 151, 402, 35 },
		{ 151, 403, 35 },
		{ 151, 404, 35 },
		{ 151, 405, 35 },
		{ 151, 406, 35 },
		{ 151, 407, 35 },
		{ 151, 408, 35 },
		{ 151, 409, 35 },
		{ 151, 410, 35 },
		{ 151, 411, 35 },
		{ 151, 412, 35 },
		{ 151, 413, 35 },
		{ 151, 414, 35 },
		{ 151, 415, 35 },
		{ 151, 416, 35 },
		{ 151, 417, 35 },
		{ 151, 418, 35 },
		{ 151, 419, 35 },
		{ 151, 420, 35 },
		{ 151, 421, 35 },
		{ 151, 422, 35 },
		{ 151, 423, 35 },
		{ 151, 424, 35 },
		{ 151, 425, 35 },
		{ 151, 426, 35 },
		{ 151, 427, 35 },
		{ 151, 428, 35 },
		{ 151, 429, 35 },
		{ 151, 430, 35 },
		{ 151, 431, 35 },
		{ 151, 432, 35 },
		{ 151, 433, 35 },
		{ 151, 434, 35 },
		{ 151, 435, 35 },
		{ 151, 436, 35 },
		{ 151, 437, 35 },
		{ 151, 438, 35 },
		{ 151, 439, 35 },
		{ 151, 440, 35 },
		{ 151, 441, 35 },
		{ 151, 442, 35 },
		{ 151, 443, 35 },
		{ 151, 444, 35 },
		{ 151, 445, 35 },
		{ 151, 446, 35 },
		{ 151, 447, 35 },
		{ 151, 448, 35 },
		{ 151, 449, 35 },
		{ 151, 450, 35 },
		{ 151, 451, 35 },
		{ 151, 452, 35 },
		{ 151, 453, 35 },
		{ 151, 454, 35 },
		{ 151, 455, 35 },
		{ 151, 456, 35 },
		{ 151, 457, 35 },
		{ 151, 458, 35 },
		{ 151, 459, 35 },
		{ 151, 460, 35 },
		{ 151, 461, 35 },
		{ 151, 462, 35 },
		{ 151, 463, 35 },
		{ 151, 464, 35 },
		{ 151, 465, 35 },
		{ 151, 466, 35 },
		{ 151, 467, 35 },
		{ 151, 468, 35 },
		{ 152, 398, 35 },
		{ 152, 399, 35 },
		{ 152, 400, 35 },
		{ 152, 401, 35 },
		{ 152, 402, 35 },
		{ 152, 403, 35 },
		{ 152, 404, 35 },
		{ 152, 405, 35 },
		{ 152, 406, 35 },
		{ 152, 407, 35 },
		{ 152, 408, 35 },
		{ 152, 409, 35 },
		{ 152, 410, 35 },
		{ 152, 411, 35 },
		{ 152, 412, 35 },
		{ 152, 413, 35 },
		{ 152, 414, 35 },
		{ 152, 415, 35 },
		{ 152, 416, 35 },
		{ 152, 417, 35 },
		{ 152, 418, 35 },
		{ 152, 419, 35 },
		{ 152, 420, 35 },
		{ 152, 421, 35 },
		{ 152, 422, 35 },
		{ 152, 423, 35 },
		{ 152, 424, 35 },
		{ 152, 425, 35 },
		{ 152, 426, 35 },
		{ 152, 427, 35 },
		{ 152, 428, 35 },
		{ 152, 429, 35 },
		{ 152, 430, 35 },
		{ 152, 431, 35 },
		{ 152, 432, 35 },
		{ 152, 433, 35 },
		{ 152, 434, 35 },
		{ 152, 435, 35 },
		{ 152, 436, 35 },
		{ 152, 437, 35 },
		{ 152, 438, 35 },
		{ 152, 439, 35 },
		{ 152, 440, 35 },
		{ 152, 441, 35 },
		{ 152, 442, 35 },
		{ 152, 443, 35 },
		{ 152, 444, 35 },
		{ 152, 445, 35 },
		{ 152, 446, 35 },
		{ 152, 447, 35 },
		{ 152, 448, 35 },
		{ 152, 449, 35 },
		{ 152, 450, 35 },
		{ 152, 451, 35 },
		{ 152, 452, 35 },
		{ 152, 453, 35 },
		{ 152, 454, 35 },
		{ 152, 455, 35 },
		{ 152, 456, 35 },
		{ 152, 457, 35 },
		{ 152, 458, 35 },
		{ 152, 459, 35 },
		{ 152, 460, 35 },
		{ 152, 461, 35 },
		{ 152, 462, 35 },
		{ 152, 463, 35 },
		{ 152, 464, 35 },
		{ 152, 465, 35 },
		{ 152, 466, 35 },
		{ 153, 396, 35 },
		{ 153, 397, 35 },
		{ 153, 398, 35 },
		{ 153, 399, 35 },
		{ 153, 400, 35 },
		{ 153, 401, 35 },
		{ 153, 402, 35 },
		{ 153, 403, 35 },
		{ 153, 404, 35 },
		{ 153, 405, 35 },
		{ 153, 406, 35 },
		{ 153, 407, 35 },
		{ 153, 408, 35 },
		{ 153, 409, 35 },
		{ 153, 410, 35 },
		{ 153, 411, 35 },
		{ 153, 412, 35 },
		{ 153, 413, 35 },
		{ 153, 414, 35 },
		{ 153, 415, 35 },
		{ 153, 416, 35 },
		{ 153, 417, 35 },
		{ 153, 418, 35 },
		{ 153, 419, 35 },
		{ 153, 420, 35 },
		{ 153, 421, 35 },
		{ 153, 422, 35 },
		{ 153, 423, 35 },
		{ 153, 424, 35 },
		{ 153, 425, 35 },
		{ 153, 426, 35 },
		{ 153, 427, 35 },
		{ 153, 428, 35 },
		{ 153, 429, 35 },
		{ 153, 430, 35 },
		{ 153, 431, 35 },
		{ 153, 432, 35 },
		{ 153, 433, 35 },
		{ 153, 434, 35 },
		{ 153, 435, 35 },
		{ 153, 436, 35 },
		{ 153, 437, 35 },
		{ 153, 438, 35 },
		{ 153, 439, 35 },
		{ 153, 440, 35 },
		{ 153, 441, 35 },
		{ 153, 442, 35 },
		{ 153, 443, 35 },
		{ 153, 444, 35 },
		{ 153, 445, 35 },
		{ 153, 446, 35 },
		{ 153, 447, 35 },
		{ 153, 448, 35 },
		{ 153, 449, 35 },
		{ 153, 450, 35 },
		{ 153, 451, 35 },
		{ 153, 452, 35 },
		{ 153, 453, 35 },
		{ 153, 454, 35 },
		{ 153, 455, 35 },
		{ 153, 456, 35 },
		{ 153, 457, 35 },
		{ 153, 458, 35 },
		{ 153, 459, 35 },
		{ 153, 460, 35 },
		{ 153, 461, 35 },
		{ 153, 462, 35 },
		{ 153, 463, 35 },
		{ 153, 464, 35 },
		{ 154, 395, 35 },
		{ 154, 396, 35 },
		{ 154, 397, 35 },
		{ 154, 398, 35 },
		{ 154, 399, 35 },
		{ 154, 400, 35 },
		{ 154, 401, 35 },
		{ 154, 402, 35 },
		{ 154, 403, 35 },
		{ 154, 404, 35 },
		{ 154, 405, 35 },
		{ 154, 406, 35 },
		{ 154, 407, 35 },
		{ 154, 408, 35 },
		{ 154, 409, 35 },
		{ 154, 410, 35 },
		{ 154, 411, 35 },
		{ 154, 412, 35 },
		{ 154, 413, 35 },
		{ 154, 414, 35 },
		{ 154, 415, 35 },
		{ 154, 416, 35 },
		{ 154, 417, 35 },
		{ 154, 418, 35 },
		{ 154, 419, 35 },
		{ 154, 420, 35 },
		{ 154, 421, 35 },
		{ 154, 422, 35 },
		{ 154, 423, 35 },
		{ 154, 424, 35 },
		{ 154, 425, 35 },
		{ 154, 426, 35 },
		{ 154, 427, 35 },
		{ 154, 428, 35 },
		{ 154, 429, 35 },
		{ 154, 430, 35 },
		{ 154, 431, 35 },
		{ 154, 432, 35 },
		{ 154, 433, 35 },
		{ 154, 434, 35 },
		{ 154, 435, 35 },
		{ 154, 436, 35 },
		{ 154, 437, 35 },
		{ 154, 438, 35 },
		{ 154, 439, 35 },
		{ 154, 440, 35 },
		{ 154, 441, 35 },
		{ 154, 442, 35 },
		{ 154, 443, 35 },
		{ 154, 444, 35 },
		{ 154, 445, 35 },
		{ 154, 446, 35 },
		{ 154, 447, 35 },
		{ 154, 448, 35 },
		{ 154, 449, 35 },
		{ 154, 450, 35 },
		{ 154, 451, 35 },
		{ 154, 452, 35 },
		{ 154, 453, 35 },
		{ 154, 454, 35 },
		{ 154, 455, 35 },
		{ 154, 456, 35 },
		{ 154, 457, 35 },
		{ 154, 458, 35 },
		{ 154, 459, 35 },
		{ 154, 460, 35 },
		{ 154, 461, 35 },
		{ 154, 462, 35 },
		{ 155, 394, 35 },
		{ 155, 395, 35 },
		{ 155, 396, 35 },
		{ 155, 397, 35 },
		{ 155, 398, 35 },
		{ 155, 399, 35 },
		{ 155, 400, 35 },
		{ 155, 401, 35 },
		{ 155, 402, 35 },
		{ 155, 403, 35 },
		{ 155, 404, 35 },
		{ 155, 405, 35 },
		{ 155, 406, 35 },
		{ 155, 407, 35 },
		{ 155, 408, 35 },
		{ 155, 409, 35 },
		{ 155, 410, 35 },
		{ 155, 411, 35 },
		{ 155, 412, 35 },
		{ 155, 413, 35 },
		{ 155, 414, 35 },
		{ 155, 415, 35 },
		{ 155, 416, 35 },
		{ 155, 417, 35 },
		{ 155, 418, 35 },
		{ 155, 419, 35 },
		{ 155, 420, 35 },
		{ 155, 421, 35 },
		{ 155, 422, 35 },
		{ 155, 423, 35 },
		{ 155, 424, 35 },
		{ 155, 425, 35 },
		{ 155, 426, 35 },
		{ 155, 427, 35 },
		{ 155, 428, 35 },
		{ 155, 429, 35 },
		{ 155, 430, 35 },
		{ 155, 431, 35 },
		{ 155, 432, 35 },
		{ 155, 433, 35 },
		{ 155, 434, 35 },
		{ 155, 435, 35 },
		{ 155, 436, 35 },
		{ 155, 437, 35 },
		{ 155, 438, 35 },
		{ 155, 439, 35 },
		{ 155, 440, 35 },
		{ 155, 441, 35 },
		{ 155, 442, 35 },
		{ 155, 443, 35 },
		{ 155, 444, 35 },
		{ 155, 445, 35 },
		{ 155, 446, 35 },
		{ 155, 447, 35 },
		{ 155, 448, 35 },
		{ 155, 449, 35 },
		{ 155, 450, 35 },
		{ 155, 451, 35 },
		{ 155, 452, 35 },
		{ 155, 453, 35 },
		{ 155, 454, 35 },
		{ 155, 455, 35 },
		{ 155, 456, 35 },
		{ 155, 457, 35 },
		{ 155, 458, 35 },
		{ 155, 459, 35 },
		{ 155, 460, 35 },
		{ 156, 392, 35 },
		{ 156, 393, 35 },
		{ 156, 394, 35 },
		{ 156, 395, 35 },
		{ 156, 396, 35 },
		{ 156, 397, 35 },
		{ 156, 398, 35 },
		{ 156, 399, 35 },
		{ 156, 400, 35 },
		{ 156, 401, 35 },
		{ 156, 402, 35 },
		{ 156, 403, 35 },
		{ 156, 404, 35 },
		{ 156, 405, 35 },
		{ 156, 406, 35 },
		{ 156, 407, 35 },
		{ 156, 408, 35 },
		{ 156, 409, 35 },
		{ 156, 410, 35 },
		{ 156, 411, 35 },
		{ 156, 412, 35 },
		{ 156, 413, 35 },
		{ 156, 414, 35 },
		{ 156, 415, 35 },
		{ 156, 416, 35 },
		{ 156, 417, 35 },
		{ 156, 418, 35 },
		{ 156, 419, 35 },
		{ 156, 420, 35 },
		{ 156, 421, 35 },
		{ 156, 422, 35 },
		{ 156, 423, 35 },
		{ 156, 424, 35 },
		{ 156, 425, 35 },
		{ 156, 426, 35 },
		{ 156, 427, 35 },
		{ 156, 428, 35 },
		{ 156, 429, 35 },
		{ 156, 430, 35 },
		{ 156, 431, 35 },
		{ 156, 432, 35 },
		{ 156, 433, 35 },
		{ 156, 434, 35 },
		{ 156, 435, 35 },
		{ 156, 436, 35 },
		{ 156, 437, 35 },
		{ 156, 438, 35 },
		{ 156, 439, 35 },
		{ 156, 440, 35 },
		{ 156, 441, 35 },
		{ 156, 442, 35 },
		{ 156, 443, 35 },
		{ 156, 444, 35 },
		{ 156, 445, 35 },
		{ 156, 446, 35 },
		{ 156, 447, 35 },
		{ 156, 448, 35 },
		{ 156, 449, 35 },
		{ 156, 450, 35 },
		{ 156, 451, 35 },
		{ 156, 452, 35 },
		{ 156, 453, 35 },
		{ 156, 454, 35 },
		{ 156, 455, 35 },
		{ 156, 456, 35 },
		{ 156, 457, 35 },
		{ 156, 458, 35 },
		{ 157, 391, 35 },
		{ 157, 392, 35 },
		{ 157, 393, 35 },
		{ 157, 394, 35 },
		{ 157, 395, 35 },
		{ 157, 396, 35 },
		{ 157, 397, 35 },
		{ 157, 398, 35 },
		{ 157, 399, 35 },
		{ 157, 400, 35 },
		{ 157, 401, 35 },
		{ 157, 402, 35 },
		{ 157, 403, 35 },
		{ 157, 404, 35 },
		{ 157, 405, 35 },
		{ 157, 406, 35 },
		{ 157, 407, 35 },
		{ 157, 408, 35 },
		{ 157, 409, 35 },
		{ 157, 410, 35 },
		{ 157, 411, 35 },
		{ 157, 412, 35 },
		{ 157, 413, 35 },
		{ 157, 414, 35 },
		{ 157, 415, 35 },
		{ 157, 416, 35 },
		{ 157, 417, 35 },
		{ 157, 418, 35 },
		{ 157, 419, 35 },
		{ 157, 420, 35 },
		{ 157, 421, 35 },
		{ 157, 422, 35 },
		{ 157, 423, 35 },
		{ 157, 424, 35 },
		{ 157, 425, 35 },
		{ 157, 426, 35 },
		{ 157, 427, 35 },
		{ 157, 428, 35 },
		{ 157, 429, 35 },
		{ 157, 430, 35 },
		{ 157, 431, 35 },
		{ 157, 432, 35 },
		{ 157, 433, 35 },
		{ 157, 434, 35 },
		{ 157, 435, 35 },
		{ 157, 436, 35 },
		{ 157, 437, 35 },
		{ 157, 438, 35 },
		{ 157, 439, 35 },
		{ 157, 440, 35 },
		{ 157, 441, 35 },
		{ 157, 442, 35 },
		{ 157, 443, 35 },
		{ 157, 444, 35 },
		{ 157, 445, 35 },
		{ 157, 446, 35 },
		{ 157, 447, 35 },
		{ 157, 448, 35 },
		{ 157, 449, 35 },
		{ 157, 450, 35 },
		{ 157, 451, 35 },
		{ 157, 452, 35 },
		{ 157, 453, 35 },
		{ 157, 454, 35 },
		{ 157, 455, 35 },
		{ 157, 456, 35 },
		{ 158, 390, 35 },
		{ 158, 391, 35 },
		{ 158, 392, 35 },
		{ 158, 393, 35 },
		{ 158, 394, 35 },
		{ 158, 395, 35 },
		{ 158, 396, 35 },
		{ 158, 397, 35 },
		{ 158, 398, 35 },
		{ 158, 399, 35 },
		{ 158, 400, 35 },
		{ 158, 401, 35 },
		{ 158, 402, 35 },
		{ 158, 403, 35 },
		{ 158, 404, 35 },
		{ 158, 405, 35 },
		{ 158, 406, 35 },
		{ 158, 407, 35 },
		{ 158, 408, 35 },
		{ 158, 409, 35 },
		{ 158, 410, 35 },
		{ 158, 411, 35 },
		{ 158, 412, 35 },
		{ 158, 413, 35 },
		{ 158, 414, 35 },
		{ 158, 415, 35 },
		{ 158, 416, 35 },
		{ 158, 417, 35 },
		{ 158, 418, 35 },
		{ 158, 419, 35 },
		{ 158, 420, 35 },
		{ 158, 421, 35 },
		{ 158, 422, 35 },
		{ 158, 423, 35 },
		{ 158, 424, 35 },
		{ 158, 425, 35 },
		{ 158, 426, 35 },
		{ 158, 427, 35 },
		{ 158, 428, 35 },
		{ 158, 429, 35 },
		{ 158, 430, 35 },
		{ 158, 431, 35 },
		{ 158, 432, 35 },
		{ 158, 433, 35 },
		{ 158, 434, 35 },
		{ 158, 435, 35 },
		{ 158, 436, 35 },
		{ 158, 437, 35 },
		{ 158, 438, 35 },
		{ 158, 439, 35 },
		{ 158, 440, 35 },
		{ 158, 441, 35 },
		{ 158, 442, 35 },
		{ 158, 443, 35 },
		{ 158, 444, 35 },
		{ 158, 445, 35 },
		{ 158, 446, 35 },
		{ 158, 447, 35 },
		{ 158, 448, 35 },
		{ 158, 449, 35 },
		{ 158, 450, 35 },
		{ 158, 451, 35 },
		{ 158, 452, 35 },
		{ 158, 453, 35 },
		{ 158, 454, 35 },
		{ 159, 388, 35 },
		{ 159, 389, 35 },
		{ 159, 390, 35 },
		{ 159, 391, 35 },
		{ 159, 392, 35 },
		{ 159, 393, 35 },
		{ 159, 394, 35 },
		{ 159, 395, 35 },
		{ 159, 396, 35 },
		{ 159, 397, 35 },
		{ 159, 398, 35 },
		{ 159, 399, 35 },
		{ 159, 400, 35 },
		{ 159, 401, 35 },
		{ 159, 402, 35 },
		{ 159, 403, 35 },
		{ 159, 404, 35 },
		{ 159, 405, 35 },
		{ 159, 406, 35 },
		{ 159, 407, 35 },
		{ 159, 408, 35 },
		{ 159, 409, 35 },
		{ 159, 410, 35 },
		{ 159, 411, 35 },
		{ 159, 412, 35 },
		{ 159, 413, 35 },
		{ 159, 414, 35 },
		{ 159, 415, 35 },
		{ 159, 416, 35 },
		{ 159, 417, 35 },
		{ 159, 418, 35 },
		{ 159, 419, 35 },
		{ 159, 420, 35 },
		{ 159, 421, 35 },
		{ 159, 422, 35 },
		{ 159, 423, 35 },
		{ 159, 424, 35 },
		{ 159, 425, 35 },
		{ 159, 426, 35 },
		{ 159, 427, 35 },
		{ 159, 428, 35 },
		{ 159, 429, 35 },
		{ 159, 430, 35 },
		{ 159, 431, 35 },
		{ 159, 432, 35 },
		{ 159, 433, 35 },
		{ 159, 434, 35 },
		{ 159, 435, 35 },
		{ 159, 436, 35 },
		{ 159, 437, 35 },
		{ 159, 438, 35 },
		{ 159, 439, 35 },
		{ 159, 440, 35 },
		{ 159, 441, 35 },
		{ 159, 442, 35 },
		{ 159, 443, 35 },
		{ 159, 444, 35 },
		{ 159, 445, 35 },
		{ 159, 446, 35 },
		{ 159, 447, 35 },
		{ 159, 448, 35 },
		{ 159, 449, 35 },
		{ 159, 450, 35 },
		{ 159, 451, 35 },
		{ 159, 452, 35 },
		{ 160, 387, 35 },
		{ 160, 388, 35 },
		{ 160, 389, 35 },
		{ 160, 390, 35 },
		{ 160, 391, 35 },
		{ 160, 392, 35 },
		{ 160, 393, 35 },
		{ 160, 394, 35 },
		{ 160, 395, 35 },
		{ 160, 396, 35 },
		{ 160, 397, 35 },
		{ 160, 398, 35 },
		{ 160, 399, 35 },
		{ 160, 400, 35 },
		{ 160, 401, 35 },
		{ 160, 402, 35 },
		{ 160, 403, 35 },
		{ 160, 404, 35 },
		{ 160, 405, 35 },
		{ 160, 406, 35 },
		{ 160, 407, 35 },
		{ 160, 408, 35 },
		{ 160, 409, 35 },
		{ 160, 410, 35 },
		{ 160, 411, 35 },
		{ 160, 412, 35 },
		{ 160, 413, 35 },
		{ 160, 414, 35 },
		{ 160, 415, 35 },
		{ 160, 416, 35 },
		{ 160, 417, 35 },
		{ 160, 418, 35 },
		{ 160, 419, 35 },
		{ 160, 420, 35 },
		{ 160, 421, 35 },
		{ 160, 422, 35 },
		{ 160, 423, 35 },
		{ 160, 424, 35 },
		{ 160, 425, 35 },
		{ 160, 426, 35 },
		{ 160, 427, 35 },
		{ 160, 428, 35 },
		{ 160, 429, 35 },
		{ 160, 430, 35 },
		{ 160, 431, 35 },
		{ 160, 432, 35 },
		{ 160, 433, 35 },
		{ 160, 434, 35 },
		{ 160, 435, 35 },
		{ 160, 436, 35 },
		{ 160, 437, 35 },
		{ 160, 438, 35 },
		{ 160, 439, 35 },
		{ 160, 440, 35 },
		{ 160, 441, 35 },
		{ 160, 442, 35 },
		{ 160, 443, 35 },
		{ 160, 444, 35 },
		{ 160, 445, 35 },
		{ 160, 446, 35 },
		{ 160, 447, 35 },
		{ 160, 448, 35 },
		{ 160, 449, 35 },
		{ 160, 450, 35 },
		{ 161, 386, 35 },
		{ 161, 387, 35 },
		{ 161, 388, 35 },
		{ 161, 389, 35 },
		{ 161, 390, 35 },
		{ 161, 391, 35 },
		{ 161, 392, 35 },
		{ 161, 393, 35 },
		{ 161, 394, 35 },
		{ 161, 395, 35 },
		{ 161, 396, 35 },
		{ 161, 397, 35 },
		{ 161, 398, 35 },
		{ 161, 399, 35 },
		{ 161, 400, 35 },
		{ 161, 401, 35 },
		{ 161, 402, 35 },
		{ 161, 403, 35 },
		{ 161, 404, 35 },
		{ 161, 405, 35 },
		{ 161, 406, 35 },
		{ 161, 407, 35 },
		{ 161, 408, 35 },
		{ 161, 409, 35 },
		{ 161, 410, 35 },
		{ 161, 411, 35 },
		{ 161, 412, 35 },
		{ 161, 413, 35 },
		{ 161, 414, 35 },
		{ 161, 415, 35 },
		{ 161, 416, 35 },
		{ 161, 417, 35 },
		{ 161, 418, 35 },
		{ 161, 419, 35 },
		{ 161, 420, 35 },
		{ 161, 421, 35 },
		{ 161, 422, 35 },
		{ 161, 423, 35 },
		{ 161, 424, 35 },
		{ 161, 425, 35 },
		{ 161, 426, 35 },
		{ 161, 427, 35 },
		{ 161, 428, 35 },
		{ 161, 429, 35 },
		{ 161, 430, 35 },
		{ 161, 431, 35 },
		{ 161, 432, 35 },
		{ 161, 433, 35 },
		{ 161, 434, 35 },
		{ 161, 435, 35 },
		{ 161, 436, 35 },
		{ 161, 437, 35 },
		{ 161, 438, 35 },
		{ 161, 439, 35 },
		{ 161, 440, 35 },
		{ 161, 441, 35 },
		{ 161, 442, 35 },
		{ 161, 443, 35 },
		{ 161, 444, 35 },
		{ 161, 445, 35 },
		{ 161, 446, 35 },
		{ 161, 447, 35 },
		{ 161, 448, 35 },
		{ 162, 384, 35 },
		{ 162, 385, 35 },
		{ 162, 386, 35 },
		{ 162, 387, 35 },
		{ 162, 388, 35 },
		{ 162, 389, 35 },
		{ 162, 390, 35 },
		{ 162, 391, 35 },
		{ 162, 392, 35 },
		{ 162, 393, 35 },
		{ 162, 394, 35 },
		{ 162, 395, 35 },
		{ 162, 396, 35 },
		{ 162, 397, 35 },
		{ 162, 398, 35 },
		{ 162, 399, 35 },
		{ 162, 400, 35 },
		{ 162, 401, 35 },
		{ 162, 402, 35 },
		{ 162, 403, 35 },
		{ 162, 404, 35 },
		{ 162, 405, 35 },
		{ 162, 406, 35 },
		{ 162, 407, 35 },
		{ 162, 408, 35 },
		{ 162, 409, 35 },
		{ 162, 410, 35 },
		{ 162, 411, 35 },
		{ 162, 412, 35 },
		{ 162, 413, 35 },
		{ 162, 414, 35 },
		{ 162, 415, 35 },
		{ 162, 416, 35 },
		{ 162, 417, 35 },
		{ 162, 418, 35 },
		{ 162, 419, 35 },
		{ 162, 420, 35 },
		{ 162, 421, 35 },
		{ 162, 422, 35 },
		{ 162, 423, 35 },
		{ 162, 424, 35 },
		{ 162, 425, 35 },
		{ 162, 426, 35 },
		{ 162, 427, 35 },
		{ 162, 428, 35 },
		{ 162, 429, 35 },
		{ 162, 430, 35 },
		{ 162, 431, 35 },
		{ 162, 432, 35 },
		{ 162, 433, 35 },
		{ 162, 434, 35 },
		{ 162, 435, 35 },
		{ 162, 436, 35 },
		{ 162, 437, 35 },
		{ 162, 438, 35 },
		{ 162, 439, 35 },
		{ 162, 440, 35 },
		{ 162, 441, 35 },
		{ 162, 442, 35 },
		{ 162, 443, 35 },
		{ 162, 444, 35 },
		{ 162, 445, 35 },
		{ 162, 446, 35 },
		{ 163, 383, 35 },
		{ 163, 384, 35 },
		{ 163, 385, 35 },
		{ 163, 386, 35 },
		{ 163, 387, 35 },
		{ 163, 388, 35 },
		{ 163, 389, 35 },
		{ 163, 390, 35 },
		{ 163, 391, 35 },
		{ 163, 392, 35 },
		{ 163, 393, 35 },
		{ 163, 394, 35 },
		{ 163, 395, 35 },
		{ 163, 396, 35 },
		{ 163, 397, 35 },
		{ 163, 398, 35 },
		{ 163, 399, 35 },
		{ 163, 400, 35 },
		{ 163, 401, 35 },
		{ 163, 402, 35 },
		{ 163, 403, 35 },
		{ 163, 404, 35 },
		{ 163, 405, 35 },
		{ 163, 406, 35 },
		{ 163, 407, 35 },
		{ 163, 408, 35 },
		{ 163, 409, 35 },
		{ 163, 410, 35 },
		{ 163, 411, 35 },
		{ 163, 412, 35 },
		{ 163, 413, 35 },
		{ 163, 414, 35 },
		{ 163, 415, 35 },
		{ 163, 416, 35 },
		{ 163, 417, 35 },
		{ 163, 418, 35 },
		{ 163, 419, 35 },
		{ 163, 420, 35 },
		{ 163, 421, 35 },
		{ 163, 422, 35 },
		{ 163, 423, 35 },
		{ 163, 424, 35 },
		{ 163, 425, 35 },
		{ 163, 426, 35 },
		{ 163, 427, 35 },
		{ 163, 428, 35 },
		{ 163, 429, 35 },
		{ 163, 430, 35 },
		{ 163, 431, 35 },
		{ 163, 432, 35 },
		{ 163, 433, 35 },
		{ 163, 434, 35 },
		{ 163, 435, 35 },
		{ 163, 436, 35 },
		{ 163, 437, 35 },
		{ 163, 438, 35 },
		{ 163, 439, 35 },
		{ 163, 440, 35 },
		{ 163, 441, 35 },
		{ 163, 442, 35 },
		{ 163, 443, 35 },
		{ 163, 444, 35 },
		{ 164, 382, 35 },
		{ 164, 383, 35 },
		{ 164, 384, 35 },
		{ 164, 385, 35 },
		{ 164, 386, 35 },
		{ 164, 387, 35 },
		{ 164, 388, 35 },
		{ 164, 389, 35 },
		{ 164, 390, 35 },
		{ 164, 391, 35 },
		{ 164, 392, 35 },
		{ 164, 393, 35 },
		{ 164, 394, 35 },
		{ 164, 395, 35 },
		{ 164, 396, 35 },
		{ 164, 397, 35 },
		{ 164, 398, 35 },
		{ 164, 399, 35 },
		{ 164, 400, 35 },
		{ 164, 401, 35 },
		{ 164, 402, 35 },
		{ 164, 403, 35 },
		{ 164, 404, 35 },
		{ 164, 405, 35 },
		{ 164, 406, 35 },
		{ 164, 407, 35 },
		{ 164, 408, 35 },
		{ 164, 409, 35 },
		{ 164, 410, 35 },
		{ 164, 411, 35 },
		{ 164, 412, 35 },
		{ 164, 413, 35 },
		{ 164, 414, 35 },
		{ 164, 415, 35 },
		{ 164, 416, 35 },
		{ 164, 417, 35 },
		{ 164, 418, 35 },
		{ 164, 419, 35 },
		{ 164, 420, 35 },
		{ 164, 421, 35 },
		{ 164, 422, 35 },
		{ 164, 423, 35 },
		{ 164, 424, 35 },
		{ 164, 425, 35 },
		{ 164, 426, 35 },
		{ 164, 427, 35 },
		{ 164, 428, 35 },
		{ 164, 429, 35 },
		{ 164, 430, 35 },
		{ 164, 431, 35 },
		{ 164, 432, 35 },
		{ 164, 433, 35 },
		{ 164, 434, 35 },
		{ 164, 435, 35 },
		{ 164, 436, 35 },
		{ 164, 437, 35 },
		{ 164, 438, 35 },
		{ 164, 439, 35 },
		{ 164, 440, 35 },
		{ 164, 441, 35 },
		{ 164, 442, 35 },
		{ 165, 380, 35 },
		{ 165, 381, 35 },
		{ 165, 382, 35 },
		{ 165, 383, 35 },
		{ 165, 384, 35 },
		{ 165, 385, 35 },
		{ 165, 386, 35 },
		{ 165, 387, 35 },
		{ 165, 388, 35 },
		{ 165, 389, 35 },
		{ 165, 390, 35 },
		{ 165, 391, 35 },
		{ 165, 392, 35 },
		{ 165, 393, 35 },
		{ 165, 394, 35 },
		{ 165, 395, 35 },
		{ 165, 396, 35 },
		{ 165, 397, 35 },
		{ 165, 398, 35 },
		{ 165, 399, 35 },
		{ 165, 400, 35 },
		{ 165, 401, 35 },
		{ 165, 402, 35 },
		{ 165, 403, 35 },
		{ 165, 404, 35 },
		{ 165, 405, 35 },
		{ 165, 406, 35 },
		{ 165, 407, 35 },
		{ 165, 408, 35 },
		{ 165, 409, 35 },
		{ 165, 410, 35 },
		{ 165, 411, 35 },
		{ 165, 412, 35 },
		{ 165, 413, 35 },
		{ 165, 414, 35 },
		{ 165, 415, 35 },
		{ 165, 416, 35 },
		{ 165, 417, 35 },
		{ 165, 418, 35 },
		{ 165, 419, 35 },
		{ 165, 420, 35 },
		{ 165, 421, 35 },
		{ 165, 422, 35 },
		{ 165, 423, 35 },
		{ 165, 424, 35 },
		{ 165, 425, 35 },
		{ 165, 426, 35 },
		{ 165, 427, 35 },
		{ 165, 428, 35 },
		{ 165, 429, 35 },
		{ 165, 430, 35 },
		{ 165, 431, 35 },
		{ 165, 432, 35 },
		{ 165, 433, 35 },
		{ 165, 434, 35 },
		{ 165, 435, 35 },
		{ 165, 436, 35 },
		{ 165, 437, 35 },
		{ 165, 438, 35 },
		{ 165, 439, 35 },
		{ 165, 440, 35 },
		{ 166, 379, 35 },
		{ 166, 380, 35 },
		{ 166, 381, 35 },
		{ 166, 382, 35 },
		{ 166, 383, 35 },
		{ 166, 384, 35 },
		{ 166, 385, 35 },
		{ 166, 386, 35 },
		{ 166, 387, 35 },
		{ 166, 388, 35 },
		{ 166, 389, 35 },
		{ 166, 390, 35 },
		{ 166, 391, 35 },
		{ 166, 392, 35 },
		{ 166, 393, 35 },
		{ 166, 394, 35 },
		{ 166, 395, 35 },
		{ 166, 396, 35 },
		{ 166, 397, 35 },
		{ 166, 398, 35 },
		{ 166, 399, 35 },
		{ 166, 400, 35 },
		{ 166, 401, 35 },
		{ 166, 402, 35 },
		{ 166, 403, 35 },
		{ 166, 404, 35 },
		{ 166, 405, 35 },
		{ 166, 406, 35 },
		{ 166, 407, 35 },
		{ 166, 408, 35 },
		{ 166, 409, 35 },
		{ 166, 410, 35 },
		{ 166, 411, 35 },
		{ 166, 412, 35 },
		{ 166, 413, 35 },
		{ 166, 414, 35 },
		{ 166, 415, 35 },
		{ 166, 416, 35 },
		{ 166, 417, 35 },
		{ 166, 418, 35 },
		{ 166, 419, 35 },
		{ 166, 420, 35 },
		{ 166, 421, 35 },
		{ 166, 422, 35 },
		{ 166, 423, 35 },
		{ 166, 424, 35 },
		{ 166, 425, 35 },
		{ 166, 426, 35 },
		{ 166, 427, 35 },
		{ 166, 428, 35 },
		{ 166, 429, 35 },
		{ 166, 430, 35 },
		{ 166, 431, 35 },
		{ 166, 432, 35 },
		{ 166, 433, 35 },
		{ 166, 434, 35 },
		{ 166, 435, 35 },
		{ 166, 436, 35 },
		{ 166, 437, 35 },
		{ 166, 438, 35 },
		{ 167, 378, 35 },
		{ 167, 379, 35 },
		{ 167, 380, 35 },
		{ 167, 381, 35 },
		{ 167, 382, 35 },
		{ 167, 383, 35 },
		{ 167, 384, 35 },
		{ 167, 385, 35 },
		{ 167, 386, 35 },
		{ 167, 387, 35 },
		{ 167, 388, 35 },
		{ 167, 389, 35 },
		{ 167, 390, 35 },
		{ 167, 391, 35 },
		{ 167, 392, 35 },
		{ 167, 393, 35 },
		{ 167, 394, 35 },
		{ 167, 395, 35 },
		{ 167, 396, 35 },
		{ 167, 397, 35 },
		{ 167, 398, 35 },
		{ 167, 399, 35 },
		{ 167, 400, 35 },
		{ 167, 401, 35 },
		{ 167, 402, 35 },
		{ 167, 403, 35 },
		{ 167, 404, 35 },
		{ 167, 405, 35 },
		{ 167, 406, 35 },
		{ 167, 407, 35 },
		{ 167, 408, 35 },
		{ 167, 409, 35 },
		{ 167, 410, 35 },
		{ 167, 411, 35 },
		{ 167, 412, 35 },
		{ 167, 413, 35 },
		{ 167, 414, 35 },
		{ 167, 415, 35 },
		{ 167, 416, 35 },
		{ 167, 417, 35 },
		{ 167, 418, 35 },
		{ 167, 419, 35 },
		{ 167, 420, 35 },
		{ 167, 421, 35 },
		{ 167, 422, 35 },
		{ 167, 423, 35 },
		{ 167, 424, 35 },
		{ 167, 425, 35 },
		{ 167, 426, 35 },
		{ 167, 427, 35 },
		{ 167, 428, 35 },
		{ 167, 429, 35 },
		{ 167, 430, 35 },
		{ 167, 431, 35 },
		{ 167, 432, 35 },
		{ 167, 433, 35 },
		{ 167, 434, 35 },
		{ 167, 435, 35 },
		{ 167, 436, 35 },
		{ 168, 376, 35 },
		{ 168, 377, 35 },
		{ 168, 378, 35 },
		{ 168, 379, 35 },
		{ 168, 380, 35 },
		{ 168, 381, 35 },
		{ 168, 382, 35 },
		{ 168, 383, 35 },
		{ 168, 384, 35 },
		{ 168, 385, 35 },
		{ 168, 386, 35 },
		{ 168, 387, 35 },
		{ 168, 388, 35 },
		{ 168, 389, 35 },
		{ 168, 390, 35 },
		{ 168, 391, 35 },
		{ 168, 392, 35 },
		{ 168, 393, 35 },
		{ 168, 394, 35 },
		{ 168, 395, 35 },
		{ 168, 396, 35 },
		{ 168, 397, 35 },
		{ 168, 398, 35 },
		{ 168, 399, 35 },
		{ 168, 400, 35 },
		{ 168, 401, 35 },
		{ 168, 402, 35 },
		{ 168, 403, 35 },
		{ 168, 404, 35 },
		{ 168, 405, 35 },
		{ 168, 406, 35 },
		{ 168, 407, 35 },
		{ 168, 408, 35 },
		{ 168, 409, 35 },
		{ 168, 410, 35 },
		{ 168, 411, 35 },
		{ 168, 412, 35 },
		{ 168, 413, 35 },
		{ 168, 414, 35 },
		{ 168, 415, 35 },
		{ 168, 416, 35 },
		{ 168, 417, 35 },
		{ 168, 418, 35 },
		{ 168, 419, 35 },
		{ 168, 420, 35 },
		{ 168, 421, 35 },
		{ 168, 422, 35 },
		{ 168, 423, 35 },
		{ 168, 424, 35 },
		{ 168, 425, 35 },
		{ 168, 426, 35 },
		{ 168, 427, 35 },
		{ 168, 428, 35 },
		{ 168, 429, 35 },
		{ 168, 430, 35 },
		{ 168, 431, 35 },
		{ 168, 432, 35 },
		{ 168, 433, 35 },
		{ 168, 434, 35 },
		{ 169, 375, 35 },
		{ 169, 376, 35 },
		{ 169, 377, 35 },
		{ 169, 378, 35 },
		{ 169, 379, 35 },
		{ 169, 380, 35 },
		{ 169, 381, 35 },
		{ 169, 382, 35 },
		{ 169, 383, 35 },
		{ 169, 384, 35 },
		{ 169, 385, 35 },
		{ 169, 386, 35 },
		{ 169, 387, 35 },
		{ 169, 388, 35 },
		{ 169, 389, 35 },
		{ 169, 390, 35 },
		{ 169, 391, 35 },
		{ 169, 392, 35 },
		{ 169, 393, 35 },
		{ 169, 394, 35 },
		{ 169, 395, 35 },
		{ 169, 396, 35 },
		{ 169, 397, 35 },
		{ 169, 398, 35 },
		{ 169, 399, 35 },
		{ 169, 400, 35 },
		{ 169, 401, 35 },
		{ 169, 402, 35 },
		{ 169, 403, 35 },
		{ 169, 404, 35 },
		{ 169, 405, 35 },
		{ 169, 406, 35 },
		{ 169, 407, 35 },
		{ 169, 408, 35 },
		{ 169, 409, 35 },
		{ 169, 410, 35 },
		{ 169, 411, 35 },
		{ 169, 412, 35 },
		{ 169, 413, 35 },
		{ 169, 414, 35 },
		{ 169, 415, 35 },
		{ 169, 416, 35 },
		{ 169, 417, 35 },
		{ 169, 418, 35 },
		{ 169, 419, 35 },
		{ 169, 420, 35 },
		{ 169, 421, 35 },
		{ 169, 422, 35 },
		{ 169, 423, 35 },
		{ 169, 424, 35 },
		{ 169, 425, 35 },
		{ 169, 426, 35 },
		{ 169, 427, 35 },
		{ 169, 428, 35 },
		{ 169, 429, 35 },
		{ 169, 430, 35 },
		{ 169, 431, 35 },
		{ 169, 432, 35 },
		{ 170, 374, 35 },
		{ 170, 375, 35 },
		{ 170, 376, 35 },
		{ 170, 377, 35 },
		{ 170, 378, 35 },
		{ 170, 379, 35 },
		{ 170, 380, 35 },
		{ 170, 381, 35 },
		{ 170, 382, 35 },
		{ 170, 383, 35 },
		{ 170, 384, 35 },
		{ 170, 385, 35 },
		{ 170, 386, 35 },
		{ 170, 387, 35 },
		{ 170, 388, 35 },
		{ 170, 389, 35 },
		{ 170, 390, 35 },
		{ 170, 391, 35 },
		{ 170, 392, 35 },
		{ 170, 393, 35 },
		{ 170, 394, 35 },
		{ 170, 395, 35 },
		{ 170, 396, 35 },
		{ 170, 397, 35 },
		{ 170, 398, 35 },
		{ 170, 399, 35 },
		{ 170, 400, 35 },
		{ 170, 401, 35 },
		{ 170, 402, 35 },
		{ 170, 403, 35 },
		{ 170, 404, 35 },
		{ 170, 405, 35 },
		{ 170, 406, 35 },
		{ 170, 407, 35 },
		{ 170, 408, 35 },
		{ 170, 409, 35 },
		{ 170, 410, 35 },
		{ 170, 411, 35 },
		{ 170, 412, 35 },
		{ 170, 413, 35 },
		{ 170, 414, 35 },
		{ 170, 415, 35 },
		{ 170, 416, 35 },
		{ 170, 417, 35 },
		{ 170, 418, 35 },
		{ 170, 419, 35 },
		{ 170, 420, 35 },
		{ 170, 421, 35 },
		{ 170, 422, 35 },
		{ 170, 423, 35 },
		{ 170, 424, 35 },
		{ 170, 425, 35 },
		{ 170, 426, 35 },
		{ 170, 427, 35 },
		{ 170, 428, 35 },
		{ 170, 429, 35 },
		{ 170, 430, 35 },
		{ 171, 372, 35 },
		{ 171, 373, 35 },
		{ 171, 374, 35 },
		{ 171, 375, 35 },
		{ 171, 376, 35 },
		{ 171, 377, 35 },
		{ 171, 378, 35 },
		{ 171, 379, 35 },
		{ 171, 380, 35 },
		{ 171, 381, 35 },
		{ 171, 382, 35 },
		{ 171, 383, 35 },
		{ 171, 384, 35 },
		{ 171, 385, 35 },
		{ 171, 386, 35 },
		{ 171, 387, 35 },
		{ 171, 388, 35 },
		{ 171, 389, 35 },
		{ 171, 390, 35 },
		{ 171, 391, 35 },
		{ 171, 392, 35 },
		{ 171, 393, 35 },
		{ 171, 394, 35 },
		{ 171, 395, 35 },
		{ 171, 396, 35 },
		{ 171, 397, 35 },
		{ 171, 398, 35 },
		{ 171, 399, 35 },
		{ 171, 400, 35 },
		{ 171, 401, 35 },
		{ 171, 402, 35 },
		{ 171, 403, 35 },
		{ 171, 404, 35 },
		{ 171, 405, 35 },
		{ 171, 406, 35 },
		{ 171, 407, 35 },
		{ 171, 408, 35 },
		{ 171, 409, 35 },
		{ 171, 410, 35 },
		{ 171, 411, 35 },
		{ 171, 412, 35 },
		{ 171, 413, 35 },
		{ 171, 414, 35 },
		{ 171, 415, 35 },
		{ 171, 416, 35 },
		{ 171, 417, 35 },
		{ 171, 418, 35 },
		{ 171, 419, 35 },
		{ 171, 420, 35 },
		{ 171, 421, 35 },
		{ 171, 422, 35 },
		{ 171, 423, 35 },
		{ 171, 424, 35 },
		{ 171, 425, 35 },
		{ 171, 426, 35 },
		{ 171, 427, 35 },
		{ 171, 428, 35 },
		{ 172, 371, 35 },
		{ 172, 372, 35 },
		{ 172, 373, 35 },
		{ 172, 374, 35 },
		{ 172, 375, 35 },
		{ 172, 376, 35 },
		{ 172, 377, 35 },
		{ 172, 378, 35 },
		{ 172, 379, 35 },
		{ 172, 380, 35 },
		{ 172, 381, 35 },
		{ 172, 382, 35 },
		{ 172, 383, 35 },
		{ 172, 384, 35 },
		{ 172, 385, 35 },
		{ 172, 386, 35 },
		{ 172, 387, 35 },
		{ 172, 388, 35 },
		{ 172, 389, 35 },
		{ 172, 390, 35 },
		{ 172, 391, 35 },
		{ 172, 392, 35 },
		{ 172, 393, 35 },
		{ 172, 394, 35 },
		{ 172, 395, 35 },
		{ 172, 396, 35 },
		{ 172, 397, 35 },
		{ 172, 398, 35 },
		{ 172, 399, 35 },
		{ 172, 400, 35 },
		{ 172, 401, 35 },
		{ 172, 402, 35 },
		{ 172, 403, 35 },
		{ 172, 404, 35 },
		{ 172, 405, 35 },
		{ 172, 406, 35 },
		{ 172, 407, 35 },
		{ 172, 408, 35 },
		{ 172, 409, 35 },
		{ 172, 410, 35 },
		{ 172, 411, 35 },
		{ 172, 412, 35 },
		{ 172, 413, 35 },
		{ 172, 414, 35 },
		{ 172, 415, 35 },
		{ 172, 416, 35 },
		{ 172, 417, 35 },
		{ 172, 418, 35 },
		{ 172, 419, 35 },
		{ 172, 420, 35 },
		{ 172, 421, 35 },
		{ 172, 422, 35 },
		{ 172, 423, 35 },
		{ 172, 424, 35 },
		{ 172, 425, 35 },
		{ 172, 426, 35 },
		{ 173, 370, 35 },
		{ 173, 371, 35 },
		{ 173, 372, 35 },
		{ 173, 373, 35 },
		{ 173, 374, 35 },
		{ 173, 375, 35 },
		{ 173, 376, 35 },
		{ 173, 377, 35 },
		{ 173, 378, 35 },
		{ 173, 379, 35 },
		{ 173, 380, 35 },
		{ 173, 381, 35 },
		{ 173, 382, 35 },
		{ 173, 383, 35 },
		{ 173, 384, 35 },
		{ 173, 385, 35 },
		{ 173, 386, 35 },
		{ 173, 387, 35 },
		{ 173, 388, 35 },
		{ 173, 389, 35 },
		{ 173, 390, 35 },
		{ 173, 391, 35 },
		{ 173, 392, 35 },
		{ 173, 393, 35 },
		{ 173, 394, 35 },
		{ 173, 395, 35 },
		{ 173, 396, 35 },
		{ 173, 397, 35 },
		{ 173, 398, 35 },
		{ 173, 399, 35 },
		{ 173, 400, 35 },
		{ 173, 401, 35 },
		{ 173, 402, 35 },
		{ 173, 403, 35 },
		{ 173, 404, 35 },
		{ 173, 405, 35 },
		{ 173, 406, 35 },
		{ 173, 407, 35 },
		{ 173, 408, 35 },
		{ 173, 409, 35 },
		{ 173, 410, 35 },
		{ 173, 411, 35 },
		{ 173, 412, 35 },
		{ 173, 413, 35 },
		{ 173, 414, 35 },
		{ 173, 415, 35 },
		{ 173, 416, 35 },
		{ 173, 417, 35 },
		{ 173, 418, 35 },
		{ 173, 419, 35 },
		{ 173, 420, 35 },
		{ 173, 421, 35 },
		{ 173, 422, 35 },
		{ 173, 423, 35 },
		{ 173, 424, 35 },
		{ 174, 368, 35 },
		{ 174, 369, 35 },
		{ 174, 370, 35 },
		{ 174, 371, 35 },
		{ 174, 372, 35 },
		{ 174, 373, 35 },
		{ 174, 374, 35 },
		{ 174, 375, 35 },
		{ 174, 376, 35 },
		{ 174, 377, 35 },
		{ 174, 378, 35 },
		{ 174, 379, 35 },
		{ 174, 380, 35 },
		{ 174, 381, 35 },
		{ 174, 382, 35 },
		{ 174, 383, 35 },
		{ 174, 384, 35 },
		{ 174, 385, 35 },
		{ 174, 386, 35 },
		{ 174, 387, 35 },
		{ 174, 388, 35 },
		{ 174, 389, 35 },
		{ 174, 390, 35 },
		{ 174, 391, 35 },
		{ 174, 392, 35 },
		{ 174, 393, 35 },
		{ 174, 394, 35 },
		{ 174, 395, 35 },
		{ 174, 396, 35 },
		{ 174, 397, 35 },
		{ 174, 398, 35 },
		{ 174, 399, 35 },
		{ 174, 400, 35 },
		{ 174, 401, 35 },
		{ 174, 402, 35 },
		{ 174, 403, 35 },
		{ 174, 404, 35 },
		{ 174, 405, 35 },
		{ 174, 406, 35 },
		{ 174, 407, 35 },
		{ 174, 408, 35 },
		{ 174, 409, 35 },
		{ 174, 410, 35 },
		{ 174, 411, 35 },
		{ 174, 412, 35 },
		{ 174, 413, 35 },
		{ 174, 414, 35 },
		{ 174, 415, 35 },
		{ 174, 416, 35 },
		{ 174, 417, 35 },
		{ 174, 418, 35 },
		{ 174, 419, 35 },
		{ 174, 420, 35 },
		{ 174, 421, 35 },
		{ 174, 422, 35 },
		{ 175, 367, 35 },
		{ 175, 368, 35 },
		{ 175, 369, 35 },
		{ 175, 370, 35 },
		{ 175, 371, 35 },
		{ 175, 372, 35 },
		{ 175, 373, 35 },
		{ 175, 374, 35 },
		{ 175, 375, 35 },
		{ 175, 376, 35 },
		{ 175, 377, 35 },
		{ 175, 378, 35 },
		{ 175, 379, 35 },
		{ 175, 380, 35 },
		{ 175, 381, 35 },
		{ 175, 382, 35 },
		{ 175, 383, 35 },
		{ 175, 384, 35 },
		{ 175, 385, 35 },
		{ 175, 386, 35 },
		{ 175, 387, 35 },
		{ 175, 388, 35 },
		{ 175, 389, 35 },
		{ 175, 390, 35 },
		{ 175, 391, 35 },
		{ 175, 392, 35 },
		{ 175, 393, 35 },
		{ 175, 394, 35 },
		{ 175, 395, 35 },
		{ 175, 396, 35 },
		{ 175, 397, 35 },
		{ 175, 398, 35 },
		{ 175, 399, 35 },
		{ 175, 400, 35 },
		{ 175, 401, 35 },
		{ 175, 402, 35 },
		{ 175, 403, 35 },
		{ 175, 404, 35 },
		{ 175, 405, 35 },
		{ 175, 406, 35 },
		{ 175, 407, 35 },
		{ 175, 408, 35 },
		{ 175, 409, 35 },
		{ 175, 410, 35 },
		{ 175, 411, 35 },
		{ 175, 412, 35 },
		{ 175, 413, 35 },
		{ 175, 414, 35 },
		{ 175, 415, 35 },
		{ 175, 416, 35 },
		{ 175, 417, 35 },
		{ 175, 418, 35 },
		{ 175, 419, 35 },
		{ 175, 420, 35 },
		{ 176, 366, 35 },
		{ 176, 367, 35 },
		{ 176, 368, 35 },
		{ 176, 369, 35 },
		{ 176, 370, 35 },
		{ 176, 371, 35 },
		{ 176, 372, 35 },
		{ 176, 373, 35 },
		{ 176, 374, 35 },
		{ 176, 375, 35 },
		{ 176, 376, 35 },
		{ 176, 377, 35 },
		{ 176, 378, 35 },
		{ 176, 379, 35 },
		{ 176, 380, 35 },
		{ 176, 381, 35 },
		{ 176, 382, 35 },
		{ 176, 383, 35 },
		{ 176, 384, 35 },
		{ 176, 385, 35 },
		{ 176, 386, 35 },
		{ 176, 387, 35 },
		{ 176, 388, 35 },
		{ 176, 389, 35 },
		{ 176, 390, 35 },
		{ 176, 391, 35 },
		{ 176, 392, 35 },
		{ 176, 393, 35 },
		{ 176, 394, 35 },
		{ 176, 395, 35 },
		{ 176, 396, 35 },
		{ 176, 397, 35 },
		{ 176, 398, 35 },
		{ 176, 399, 35 },
		{ 176, 400, 35 },
		{ 176, 401, 35 },
		{ 176, 402, 35 },
		{ 176, 403, 35 },
		{ 176, 404, 35 },
		{ 176, 405, 35 },
		{ 176, 406, 35 },
		{ 176, 407, 35 },
		{ 176, 408, 35 },
		{ 176, 409, 35 },
		{ 176, 410, 35 },
		{ 176, 411, 35 },
		{ 176, 412, 35 },
		{ 176, 413, 35 },
		{ 176, 414, 35 },
		{ 176, 415, 35 },
		{ 176, 416, 35 },
		{ 176, 417, 35 },
		{ 176, 418, 35 },
		{ 177, 364, 35 },
		{ 177, 365, 35 },
		{ 177, 366, 35 },
		{ 177, 367, 35 },
		{ 177, 368, 35 },
		{ 177, 369, 35 },
		{ 177, 370, 35 },
		{ 177, 371, 35 },
		{ 177, 372, 35 },
		{ 177, 373, 35 },
		{ 177, 374, 35 },
		{ 177, 375, 35 },
		{ 177, 376, 35 },
		{ 177, 377, 35 },
		{ 177, 378, 35 },
		{ 177, 379, 35 },
		{ 177, 380, 35 },
		{ 177, 381, 35 },
		{ 177, 382, 35 },
		{ 177, 383, 35 },
		{ 177, 384, 35 },
		{ 177, 385, 35 },
		{ 177, 386, 35 },
		{ 177, 387, 35 },
		{ 177, 388, 35 },
		{ 177, 389, 35 },
		{ 177, 390, 35 },
		{ 177, 391, 35 },
		{ 177, 392, 35 },
		{ 177, 393, 35 },
		{ 177, 394, 35 },
		{ 177, 395, 35 },
		{ 177, 396, 35 },
		{ 177, 397, 35 },
		{ 177, 398, 35 },
		{ 177, 399, 35 },
		{ 177, 400, 35 },
		{ 177, 401, 35 },
		{ 177, 402, 35 },
		{ 177, 403, 35 },
		{ 177, 404, 35 },
		{ 177, 405, 35 },
		{ 177, 406, 35 },
		{ 177, 407, 35 },
		{ 177, 408, 35 },
		{ 177, 409, 35 },
		{ 177, 410, 35 },
		{ 177, 411, 35 },
		{ 177, 412, 35 },
		{ 177, 413, 35 },
		{ 177, 414, 35 },
		{ 177, 415, 35 },
		{ 177, 416, 35 },
		{ 178, 363, 35 },
		{ 178, 364, 35 },
		{ 178, 365, 35 },
		{ 178, 366, 35 },
		{ 178, 367, 35 },
		{ 178, 368, 35 },
		{ 178, 369, 35 },
		{ 178, 370, 35 },
		{ 178, 371, 35 },
		{ 178, 372, 35 },
		{ 178, 373, 35 },
		{ 178, 374, 35 },
		{ 178, 375, 35 },
		{ 178, 376, 35 },
		{ 178, 377, 35 },
		{ 178, 378, 35 },
		{ 178, 379, 35 },
		{ 178, 380, 35 },
		{ 178, 381, 35 },
		{ 178, 382, 35 },
		{ 178, 383, 35 },
		{ 178, 384, 35 },
		{ 178, 385, 35 },
		{ 178, 386, 35 },
		{ 178, 387, 35 },
		{ 178, 388, 35 },
		{ 178, 389, 35 },
		{ 178, 390, 35 },
		{ 178, 391, 35 },
		{ 178, 392, 35 },
		{ 178, 393, 35 },
		{ 178, 394, 35 },
		{ 178, 395, 35 },
		{ 178, 396, 35 },
		{ 178, 397, 35 },
		{ 178, 398, 35 },
		{ 178, 399, 35 },
		{ 178, 400, 35 },
		{ 178, 401, 35 },
		{ 178, 402, 35 },
		{ 178, 403, 35 },
		{ 178, 404, 35 },
		{ 178, 405, 35 },
		{ 178, 406, 35 },
		{ 178, 407, 35 },
		{ 178, 408, 35 },
		{ 178, 409, 35 },
		{ 178, 410, 35 },
		{ 178, 411, 35 },
		{ 178, 412, 35 },
		{ 178, 413, 35 },
		{ 178, 414, 35 },
		{ 179, 362, 35 },
		{ 179, 363, 35 },
		{ 179, 364, 35 },
		{ 179, 365, 35 },
		{ 179, 366, 35 },
		{ 179, 367, 35 },
		{ 179, 368, 35 },
		{ 179, 369, 35 },
		{ 179, 370, 35 },
		{ 179, 371, 35 },
		{ 179, 372, 35 },
		{ 179, 373, 35 },
		{ 179, 374, 35 },
		{ 179, 375, 35 },
		{ 179, 376, 35 },
		{ 179, 377, 35 },
		{ 179, 378, 35 },
		{ 179, 379, 35 },
		{ 179, 380, 35 },
		{ 179, 381, 35 },
		{ 179, 382, 35 },
		{ 179, 383, 35 },
		{ 179, 384, 35 },
		{ 179, 385, 35 },
		{ 179, 386, 35 },
		{ 179, 387, 35 },
		{ 179, 388, 35 },
		{ 179, 389, 35 },
		{ 179, 390, 35 },
		{ 179, 391, 35 },
		{ 179, 392, 35 },
		{ 179, 393, 35 },
		{ 179, 394, 35 },
		{ 179, 395, 35 },
		{ 179, 396, 35 },
		{ 179, 397, 35 },
		{ 179, 398, 35 },
		{ 179, 399, 35 },
		{ 179, 400, 35 },
		{ 179, 401, 35 },
		{ 179, 402, 35 },
		{ 179, 403, 35 },
		{ 179, 404, 35 },
		{ 179, 405, 35 },
		{ 179, 406, 35 },
		{ 179, 407, 35 },
		{ 179, 408, 35 },
		{ 179, 409, 35 },
		{ 179, 410, 35 },
		{ 179, 411, 35 },
		{ 179, 412, 35 },
		{ 180, 360, 35 },
		{ 180, 361, 35 },
		{ 180, 362, 35 },
		{ 180, 363, 35 },
		{ 180, 364, 35 },
		{ 180, 365, 35 },
		{ 180, 366, 35 },
		{ 180, 367, 35 },
		{ 180, 368, 35 },
		{ 180, 369, 35 },
		{ 180, 370, 35 },
		{ 180, 371, 35 },
		{ 180, 372, 35 },
		{ 180, 373, 35 },
		{ 180, 374, 35 },
		{ 180, 375, 35 },
		{ 180, 376, 35 },
		{ 180, 377, 35 },
		{ 180, 378, 35 },
		{ 180, 379, 35 },
		{ 180, 380, 35 },
		{ 180, 381, 35 },
		{ 180, 382, 35 },
		{ 180, 383, 35 },
		{ 180, 384, 35 },
		{ 180, 385, 35 },
		{ 180, 386, 35 },
		{ 180, 387, 35 },
		{ 180, 388, 35 },
		{ 180, 389, 35 },
		{ 180, 390, 35 },
		{ 180, 391, 35 },
		{ 180, 392, 35 },
		{ 180, 393, 35 },
		{ 180, 394, 35 },
		{ 180, 395, 35 },
		{ 180, 396, 35 },
		{ 180, 397, 35 },
		{ 180, 398, 35 },
		{ 180, 399, 35 },
		{ 180, 400, 35 },
		{ 180, 401, 35 },
		{ 180, 402, 35 },
		{ 180, 403, 35 },
		{ 180, 404, 35 },
		{ 180, 405, 35 },
		{ 180, 406, 35 },
		{ 180, 407, 35 },
		{ 180, 408, 35 },
		{ 180, 409, 35 },
		{ 180, 410, 35 },
		{ 181, 359, 35 },
		{ 181, 360, 35 },
		{ 181, 361, 35 },
		{ 181, 362, 35 },
		{ 181, 363, 35 },
		{ 181, 364, 35 },
		{ 181, 365, 35 },
		{ 181, 366, 35 },
		{ 181, 367, 35 },
		{ 181, 368, 35 },
		{ 181, 369, 35 },
		{ 181, 370, 35 },
		{ 181, 371, 35 },
		{ 181, 372, 35 },
		{ 181, 373, 35 },
		{ 181, 374, 35 },
		{ 181, 375, 35 },
		{ 181, 376, 35 },
		{ 181, 377, 35 },
		{ 181, 378, 35 },
		{ 181, 379, 35 },
		{ 181, 380, 35 },
		{ 181, 381, 35 },
		{ 181, 382, 35 },
		{ 181, 383, 35 },
		{ 181, 384, 35 },
		{ 181, 385, 35 },
		{ 181, 386, 35 },
		{ 181, 387, 35 },
		{ 181, 388, 35 },
		{ 181, 389, 35 },
		{ 181, 390, 35 },
		{ 181, 391, 35 },
		{ 181, 392, 35 },
		{ 181, 393, 35 },
		{ 181, 394, 35 },
		{ 181, 395, 35 },
		{ 181, 396, 35 },
		{ 181, 397, 35 },
		{ 181, 398, 35 },
		{ 181, 399, 35 },
		{ 181, 400, 35 },
		{ 181, 401, 35 },
		{ 181, 402, 35 },
		{ 181, 403, 35 },
		{ 181, 404, 35 },
		{ 181, 405, 35 },
		{ 181, 406, 35 },
		{ 181, 407, 35 },
		{ 181, 408, 35 },
		{ 182, 358, 35 },
		{ 182, 359, 35 },
		{ 182, 360, 35 },
		{ 182, 361, 35 },
		{ 182, 362, 35 },
		{ 182, 363, 35 },
		{ 182, 364, 35 },
		{ 182, 365, 35 },
		{ 182, 366, 35 },
		{ 182, 367, 35 },
		{ 182, 368, 35 },
		{ 182, 369, 35 },
		{ 182, 370, 35 },
		{ 182, 371, 35 },
		{ 182, 372, 35 },
		{ 182, 373, 35 },
		{ 182, 374, 35 },
		{ 182, 375, 35 },
		{ 182, 376, 35 },
		{ 182, 377, 35 },
		{ 182, 378, 35 },
		{ 182, 379, 35 },
		{ 182, 380, 35 },
		{ 182, 381, 35 },
		{ 182, 382, 35 },
		{ 182, 383, 35 },
		{ 182, 384, 35 },
		{ 182, 385, 35 },
		{ 182, 386, 35 },
		{ 182, 387, 35 },
		{ 182, 388, 35 },
		{ 182, 389, 35 },
		{ 182, 390, 35 },
		{ 182, 391, 35 },
		{ 182, 392, 35 },
		{ 182, 393, 35 },
		{ 182, 394, 35 },
		{ 182, 395, 35 },
		{ 182, 396, 35 },
		{ 182, 397, 35 },
		{ 182, 398, 35 },
		{ 182, 399, 35 },
		{ 182, 400, 35 },
		{ 182, 401, 35 },
		{ 182, 402, 35 },
		{ 182, 403, 35 },
		{ 182, 404, 35 },
		{ 182, 405, 35 },
		{ 182, 406, 35 },
		{ 183, 356, 35 },
		{ 183, 357, 35 },
		{ 183, 358, 35 },
		{ 183, 359, 35 },
		{ 183, 360, 35 },
		{ 183, 361, 35 },
		{ 183, 362, 35 },
		{ 183, 363, 35 },
		{ 183, 364, 35 },
		{ 183, 365, 35 },
		{ 183, 366, 35 },
		{ 183, 367, 35 },
		{ 183, 368, 35 },
		{ 183, 369, 35 },
		{ 183, 370, 35 },
		{ 183, 371, 35 },
		{ 183, 372, 35 },
		{ 183, 373, 35 },
		{ 183, 374, 35 },
		{ 183, 375, 35 },
		{ 183, 376, 35 },
		{ 183, 377, 35 },
		{ 183, 378, 35 },
		{ 183, 379, 35 },
		{ 183, 380, 35 },
		{ 183, 381, 35 },
		{ 183, 382, 35 },
		{ 183, 383, 35 },
		{ 183, 384, 35 },
		{ 183, 385, 35 },
		{ 183, 386, 35 },
		{ 183, 387, 35 },
		{ 183, 388, 35 },
		{ 183, 389, 35 },
		{ 183, 390, 35 },
		{ 183, 391, 35 },
		{ 183, 392, 35 },
		{ 183, 393, 35 },
		{ 183, 394, 35 },
		{ 183, 395, 35 },
		{ 183, 396, 35 },
		{ 183, 397, 35 },
		{ 183, 398, 35 },
		{ 183, 399, 35 },
		{ 183, 400, 35 },
		{ 183, 401, 35 },
		{ 183, 402, 35 },
		{ 183, 403, 35 },
		{ 183, 404, 35 },
		{ 184, 355, 35 },
		{ 184, 356, 35 },
		{ 184, 357, 35 },
		{ 184, 358, 35 },
		{ 184, 359, 35 },
		{ 184, 360, 35 },
		{ 184, 361, 35 },
		{ 184, 362, 35 },
		{ 184, 363, 35 },
		{ 184, 364, 35 },
		{ 184, 365, 35 },
		{ 184, 366, 35 },
		{ 184, 367, 35 },
		{ 184, 368, 35 },
		{ 184, 369, 35 },
		{ 184, 370, 35 },
		{ 184, 371, 35 },
		{ 184, 372, 35 },
		{ 184, 373, 35 },
		{ 184, 374, 35 },
		{ 184, 375, 35 },
		{ 184, 376, 35 },
		{ 184, 377, 35 },
		{ 184, 378, 35 },
		{ 184, 379, 35 },
		{ 184, 380, 35 },
		{ 184, 381, 35 },
		{ 184, 382, 35 },
		{ 184, 383, 35 },
		{ 184, 384, 35 },
		{ 184, 385, 35 },
		{ 184, 386, 35 },
		{ 184, 387, 35 },
		{ 184, 388, 35 },
		{ 184, 389, 35 },
		{ 184, 390, 35 },
		{ 184, 391, 35 },
		{ 184, 392, 35 },
		{ 184, 393, 35 },
		{ 184, 394, 35 },
		{ 184, 395, 35 },
		{ 184, 396, 35 },
		{ 184, 397, 35 },
		{ 184, 398, 35 },
		{ 184, 399, 35 },
		{ 184, 400, 35 },
		{ 184, 401, 35 },
		{ 184, 402, 35 },
		{ 185, 354, 35 },
		{ 185, 355, 35 },
		{ 185, 356, 35 },
		{ 185, 357, 35 },
		{ 185, 358, 35 },
		{ 185, 359, 35 },
		{ 185, 360, 35 },
		{ 185, 361, 35 },
		{ 185, 362, 35 },
		{ 185, 363, 35 },
		{ 185, 364, 35 },
		{ 185, 365, 35 },
		{ 185, 366, 35 },
		{ 185, 367, 35 },
		{ 185, 368, 35 },
		{ 185, 369, 35 },
		{ 185, 370, 35 },
		{ 185, 371, 35 },
		{ 185, 372, 35 },
		{ 185, 373, 35 },
		{ 185, 374, 35 },
		{ 185, 375, 35 },
		{ 185, 376, 35 },
		{ 185, 377, 35 },
		{ 185, 378, 35 },
		{ 185, 379, 35 },
		{ 185, 380, 35 },
		{ 185, 381, 35 },
		{ 185, 382, 35 },
		{ 185, 383, 35 },
		{ 185, 384, 35 },
		{ 185, 385, 35 },
		{ 185, 386, 35 },
		{ 185, 387, 35 },
		{ 185, 388, 35 },
		{ 185, 389, 35 },
		{ 185, 390, 35 },
		{ 185, 391, 35 },
		{ 185, 392, 35 },
		{ 185, 393, 35 },
		{ 185, 394, 35 },
		{ 185, 395, 35 },
		{ 185, 396, 35 },
		{ 185, 397, 35 },
		{ 185, 398, 35 },
		{ 185, 399, 35 },
		{ 185, 400, 35 },
		{ 186, 352, 35 },
		{ 186, 353, 35 },
		{ 186, 354, 35 },
		{ 186, 355, 35 },
		{ 186, 356, 35 },
		{ 186, 357, 35 },
		{ 186, 358, 35 },
		{ 186, 359, 35 },
		{ 186, 360, 35 },
		{ 186, 361, 35 },
		{ 186, 362, 35 },
		{ 186, 363, 35 },
		{ 186, 364, 35 },
		{ 186, 365, 35 },
		{ 186, 366, 35 },
		{ 186, 367, 35 },
		{ 186, 368, 35 },
		{ 186, 369, 35 },
		{ 186, 370, 35 },
		{ 186, 371, 35 },
		{ 186, 372, 35 },
		{ 186, 373, 35 },
		{ 186, 374, 35 },
		{ 186, 375, 35 },
		{ 186, 376, 35 },
		{ 186, 377, 35 },
		{ 186, 378, 35 },
		{ 186, 379, 35 },
		{ 186, 380, 35 },
		{ 186, 381, 35 },
		{ 186, 382, 35 },
		{ 186, 383, 35 },
		{ 186, 384, 35 },
		{ 186, 385, 35 },
		{ 186, 386, 35 },
		{ 186, 387, 35 },
		{ 186, 388, 35 },
		{ 186, 389, 35 },
		{ 186, 390, 35 },
		{ 186, 391, 35 },
		{ 186, 392, 35 },
		{ 186, 393, 35 },
		{ 186, 394, 35 },
		{ 186, 395, 35 },
		{ 186, 396, 35 },
		{ 186, 397, 35 },
		{ 186, 398, 35 },
		{ 187, 351, 35 },
		{ 187, 352, 35 },
		{ 187, 353, 35 },
		{ 187, 354, 35 },
		{ 187, 355, 35 },
		{ 187, 356, 35 },
		{ 187, 357, 35 },
		{ 187, 358, 35 },
		{ 187, 359, 35 },
		{ 187, 360, 35 },
		{ 187, 361, 35 },
		{ 187, 362, 35 },
		{ 187, 363, 35 },
		{ 187, 364, 35 },
		{ 187, 365, 35 },
		{ 187, 366, 35 },
		{ 187, 367, 35 },
		{ 187, 368, 35 },
		{ 187, 369, 35 },
		{ 187, 370, 35 },
		{ 187, 371, 35 },
		{ 187, 372, 35 },
		{ 187, 373, 35 },
		{ 187, 374, 35 },
		{ 187, 375, 35 },
		{ 187, 376, 35 },
		{ 187, 377, 35 },
		{ 187, 378, 35 },
		{ 187, 379, 35 },
		{ 187, 380, 35 },
		{ 187, 381, 35 },
		{ 187, 382, 35 },
		{ 187, 383, 35 },
		{ 187, 384, 35 },
		{ 187, 385, 35 },
		{ 187, 386, 35 },
		{ 187, 387, 35 },
		{ 187, 388, 35 },
		{ 187, 389, 35 },
		{ 187, 390, 35 },
		{ 187, 391, 35 },
		{ 187, 392, 35 },
		{ 187, 393, 35 },
		{ 187, 394, 35 },
		{ 187, 395, 35 },
		{ 187, 396, 35 },
		{ 188, 350, 35 },
		{ 188, 351, 35 },
		{ 188, 352, 35 },
		{ 188, 353, 35 },
		{ 188, 354, 35 },
		{ 188, 355, 35 },
		{ 188, 356, 35 },
		{ 188, 357, 35 },
		{ 188, 358, 35 },
		{ 188, 359, 35 },
		{ 188, 360, 35 },
		{ 188, 361, 35 },
		{ 188, 362, 35 },
		{ 188, 363, 35 },
		{ 188, 364, 35 },
		{ 188, 365, 35 },
		{ 188, 366, 35 },
		{ 188, 367, 35 },
		{ 188, 368, 35 },
		{ 188, 369, 35 },
		{ 188, 370, 35 },
		{ 188, 371, 35 },
		{ 188, 372, 35 },
		{ 188, 373, 35 },
		{ 188, 374, 35 },
		{ 188, 375, 35 },
		{ 188, 376, 35 },
		{ 188, 377, 35 },
		{ 188, 378, 35 },
		{ 188, 379, 35 },
		{ 188, 380, 35 },
		{ 188, 381, 35 },
		{ 188, 382, 35 },
		{ 188, 383, 35 },
		{ 188, 384, 35 },
		{ 188, 385, 35 },
		{ 188, 386, 35 },
		{ 188, 387, 35 },
		{ 188, 388, 35 },
		{ 188, 389, 35 },
		{ 188, 390, 35 },
		{ 188, 391, 35 },
		{ 188, 392, 35 },
		{ 188, 393, 35 },
		{ 188, 394, 35 },
		{ 189, 348, 35 },
		{ 189, 349, 35 },
		{ 189, 350, 35 },
		{ 189, 351, 35 },
		{ 189, 352, 35 },
		{ 189, 353, 35 },
		{ 189, 354, 35 },
		{ 189, 355, 35 },
		{ 189, 356, 35 },
		{ 189, 357, 35 },
		{ 189, 358, 35 },
		{ 189, 359, 35 },
		{ 189, 360, 35 },
		{ 189, 361, 35 },
		{ 189, 362, 35 },
		{ 189, 363, 35 },
		{ 189, 364, 35 },
		{ 189, 365, 35 },
		{ 189, 366, 35 },
		{ 189, 367, 35 },
		{ 189, 368, 35 },
		{ 189, 369, 35 },
		{ 189, 370, 35 },
		{ 189, 371, 35 },
		{ 189, 372, 35 },
		{ 189, 373, 35 },
		{ 189, 374, 35 },
		{ 189, 375, 35 },
		{ 189, 376, 35 },
		{ 189, 377, 35 },
		{ 189, 378, 35 },
		{ 189, 379, 35 },
		{ 189, 380, 35 },
		{ 189, 381, 35 },
		{ 189, 382, 35 },
		{ 189, 383, 35 },
		{ 189, 384, 35 },
		{ 189, 385, 35 },
		{ 189, 386, 35 },
		{ 189, 387, 35 },
		{ 189, 388, 35 },
		{ 189, 389, 35 },
		{ 189, 390, 35 },
		{ 189, 391, 35 },
		{ 189, 392, 35 },
		{ 190, 347, 35 },
		{ 190, 348, 35 },
		{ 190, 349, 35 },
		{ 190, 350, 35 },
		{ 190, 351, 35 },
		{ 190, 352, 35 },
		{ 190, 353, 35 },
		{ 190, 354, 35 },
		{ 190, 355, 35 },
		{ 190, 356, 35 },
		{ 190, 357, 35 },
		{ 190, 358, 35 },
		{ 190, 359, 35 },
		{ 190, 360, 35 },
		{ 190, 361, 35 },
		{ 190, 362, 35 },
		{ 190, 363, 35 },
		{ 190, 364, 35 },
		{ 190, 365, 35 },
		{ 190, 366, 35 },
		{ 190, 367, 35 },
		{ 190, 368, 35 },
		{ 190, 369, 35 },
		{ 190, 370, 35 },
		{ 190, 371, 35 },
		{ 190, 372, 35 },
		{ 190, 373, 35 },
		{ 190, 374, 35 },
		{ 190, 375, 35 },
		{ 190, 376, 35 },
		{ 190, 377, 35 },
		{ 190, 378, 35 },
		{ 190, 379, 35 },
		{ 190, 380, 35 },
		{ 190, 381, 35 },
		{ 190, 382, 35 },
		{ 190, 383, 35 },
		{ 190, 384, 35 },
		{ 190, 385, 35 },
		{ 190, 386, 35 },
		{ 190, 387, 35 },
		{ 190, 388, 35 },
		{ 190, 389, 35 },
		{ 190, 390, 35 },
		{ 191, 346, 35 },
		{ 191, 347, 35 },
		{ 191, 348, 35 },
		{ 191, 349, 35 },
		{ 191, 350, 35 },
		{ 191, 351, 35 },
		{ 191, 352, 35 },
		{ 191, 353, 35 },
		{ 191, 354, 35 },
		{ 191, 355, 35 },
		{ 191, 356, 35 },
		{ 191, 357, 35 },
		{ 191, 358, 35 },
		{ 191, 359, 35 },
		{ 191, 360, 35 },
		{ 191, 361, 35 },
		{ 191, 362, 35 },
		{ 191, 363, 35 },
		{ 191, 364, 35 },
		{ 191, 365, 35 },
		{ 191, 366, 35 },
		{ 191, 367, 35 },
		{ 191, 368, 35 },
		{ 191, 369, 35 },
		{ 191, 370, 35 },
		{ 191, 371, 35 },
		{ 191, 372, 35 },
		{ 191, 373, 35 },
		{ 191, 374, 35 },
		{ 191, 375, 35 },
		{ 191, 376, 35 },
		{ 191, 377, 35 },
		{ 191, 378, 35 },
		{ 191, 379, 35 },
		{ 191, 380, 35 },
		{ 191, 381, 35 },
		{ 191, 382, 35 },
		{ 191, 383, 35 },
		{ 191, 384, 35 },
		{ 191, 385, 35 },
		{ 191, 386, 35 },
		{ 191, 387, 35 },
		{ 191, 388, 35 },
		{ 192, 344, 35 },
		{ 192, 345, 35 },
		{ 192, 346, 35 },
		{ 192, 347, 35 },
		{ 192, 348, 35 },
		{ 192, 349, 35 },
		{ 192, 350, 35 },
		{ 192, 351, 35 },
		{ 192, 352, 35 },
		{ 192, 353, 35 },
		{ 192, 354, 35 },
		{ 192, 355, 35 },
		{ 192, 356, 35 },
		{ 192, 357, 35 },
		{ 192, 358, 35 },
		{ 192, 359, 35 },
		{ 192, 360, 35 },
		{ 192, 361, 35 },
		{ 192, 362, 35 },
		{ 192, 363, 35 },
		{ 192, 364, 35 },
		{ 192, 365, 35 },
		{ 192, 366, 35 },
		{ 192, 367, 35 },
		{ 192, 368, 35 },
		{ 192, 369, 35 },
		{ 192, 370, 35 },
		{ 192, 371, 35 },
		{ 192, 372, 35 },
		{ 192, 373, 35 },
		{ 192, 374, 35 },
		{ 192, 375, 35 },
		{ 192, 376, 35 },
		{ 192, 377, 35 },
		{ 192, 378, 35 },
		{ 192, 379, 35 },
		{ 192, 380, 35 },
		{ 192, 381, 35 },
		{ 192, 382, 35 },
		{ 192, 383, 35 },
		{ 192, 384, 35 },
		{ 192, 385, 35 },
		{ 192, 386, 35 },
		{ 193, 343, 35 },
		{ 193, 344, 35 },
		{ 193, 345, 35 },
		{ 193, 346, 35 },
		{ 193, 347, 35 },
		{ 193, 348, 35 },
		{ 193, 349, 35 },
		{ 193, 350, 35 },
		{ 193, 351, 35 },
		{ 193, 352, 35 },
		{ 193, 353, 35 },
		{ 193, 354, 35 },
		{ 193, 355, 35 },
		{ 193, 356, 35 },
		{ 193, 357, 35 },
		{ 193, 358, 35 },
		{ 193, 359, 35 },
		{ 193, 360, 35 },
		{ 193, 361, 35 },
		{ 193, 362, 35 },
		{ 193, 363, 35 },
		{ 193, 364, 35 },
		{ 193, 365, 35 },
		{ 193, 366, 35 },
		{ 193, 367, 35 },
		{ 193, 368, 35 },
		{ 193, 369, 35 },
		{ 193, 370, 35 },
		{ 193, 371, 35 },
		{ 193, 372, 35 },
		{ 193, 373, 35 },
		{ 193, 374, 35 },
		{ 193, 375, 35 },
		{ 193, 376, 35 },
		{ 193, 377, 35 },
		{ 193, 378, 35 },
		{ 193, 379, 35 },
		{ 193, 380, 35 },
		{ 193, 381, 35 },
		{ 193, 382, 35 },
		{ 193, 383, 35 },
		{ 193, 384, 35 },
		{ 194, 342, 35 },
		{ 194, 343, 35 },
		{ 194, 344, 35 },
		{ 194, 345, 35 },
		{ 194, 346, 35 },
		{ 194, 347, 35 },
		{ 194, 348, 35 },
		{ 194, 349, 35 },
		{ 194, 350, 35 },
		{ 194, 351, 35 },
		{ 194, 352, 35 },
		{ 194, 353, 35 },
		{ 194, 354, 35 },
		{ 194, 355, 35 },
		{ 194, 356, 35 },
		{ 194, 357, 35 },
		{ 194, 358, 35 },
		{ 194, 359, 35 },
		{ 194, 360, 35 },
		{ 194, 361, 35 },
		{ 194, 362, 35 },
		{ 194, 363, 35 },
		{ 194, 364, 35 },
		{ 194, 365, 35 },
		{ 194, 366, 35 },
		{ 194, 367, 35 },
		{ 194, 368, 35 },
		{ 194, 369, 35 },
		{ 194, 370, 35 },
		{ 194, 371, 35 },
		{ 194, 372, 35 },
		{ 194, 373, 35 },
		{ 194, 374, 35 },
		{ 194, 375, 35 },
		{ 194, 376, 35 },
		{ 194, 377, 35 },
		{ 194, 378, 35 },
		{ 194, 379, 35 },
		{ 194, 380, 35 },
		{ 194, 381, 35 },
		{ 194, 382, 35 },
		{ 195, 340, 35 },
		{ 195, 341, 35 },
		{ 195, 342, 35 },
		{ 195, 343, 35 },
		{ 195, 344, 35 },
		{ 195, 345, 35 },
		{ 195, 346, 35 },
		{ 195, 347, 35 },
		{ 195, 348, 35 },
		{ 195, 349, 35 },
		{ 195, 350, 35 },
		{ 195, 351, 35 },
		{ 195, 352, 35 },
		{ 195, 353, 35 },
		{ 195, 354, 35 },
		{ 195, 355, 35 },
		{ 195, 356, 35 },
		{ 195, 357, 35 },
		{ 195, 358, 35 },
		{ 195, 359, 35 },
		{ 195, 360, 35 },
		{ 195, 361, 35 },
		{ 195, 362, 35 },
		{ 195, 363, 35 },
		{ 195, 364, 35 },
		{ 195, 365, 35 },
		{ 195, 366, 35 },
		{ 195, 367, 35 },
		{ 195, 368, 35 },
		{ 195, 369, 35 },
		{ 195, 370, 35 },
		{ 195, 371, 35 },
		{ 195, 372, 35 },
		{ 195, 373, 35 },
		{ 195, 374, 35 },
		{ 195, 375, 35 },
		{ 195, 376, 35 },
		{ 195, 377, 35 },
		{ 195, 378, 35 },
		{ 195, 379, 35 },
		{ 195, 380, 35 },
		{ 196, 339, 35 },
		{ 196, 340, 35 },
		{ 196, 341, 35 },
		{ 196, 342, 35 },
		{ 196, 343, 35 },
		{ 196, 344, 35 },
		{ 196, 345, 35 },
		{ 196, 346, 35 },
		{ 196, 347, 35 },
		{ 196, 348, 35 },
		{ 196, 349, 35 },
		{ 196, 350, 35 },
		{ 196, 351, 35 },
		{ 196, 352, 35 },
		{ 196, 353, 35 },
		{ 196, 354, 35 },
		{ 196, 355, 35 },
		{ 196, 356, 35 },
		{ 196, 357, 35 },
		{ 196, 358, 35 },
		{ 196, 359, 35 },
		{ 196, 360, 35 },
		{ 196, 361, 35 },
		{ 196, 362, 35 },
		{ 196, 363, 35 },
		{ 196, 364, 35 },
		{ 196, 365, 35 },
		{ 196, 366, 35 },
		{ 196, 367, 35 },
		{ 196, 368, 35 },
		{ 196, 369, 35 },
		{ 196, 370, 35 },
		{ 196, 371, 35 },
		{ 196, 372, 35 },
		{ 196, 373, 35 },
		{ 196, 374, 35 },
		{ 196, 375, 35 },
		{ 196, 376, 35 },
		{ 196, 377, 35 },
		{ 196, 378, 35 },
		{ 197, 338, 35 },
		{ 197, 339, 35 },
		{ 197, 340, 35 },
		{ 197, 341, 35 },
		{ 197, 342, 35 },
		{ 197, 343, 35 },
		{ 197, 344, 35 },
		{ 197, 345, 35 },
		{ 197, 346, 35 },
		{ 197, 347, 35 },
		{ 197, 348, 35 },
		{ 197, 349, 35 },
		{ 197, 350, 35 },
		{ 197, 351, 35 },
		{ 197, 352, 35 },
		{ 197, 353, 35 },
		{ 197, 354, 35 },
		{ 197, 355, 35 },
		{ 197, 356, 35 },
		{ 197, 357, 35 },
		{ 197, 358, 35 },
		{ 197, 359, 35 },
		{ 197, 360, 35 },
		{ 197, 361, 35 },
		{ 197, 362, 35 },
		{ 197, 363, 35 },
		{ 197, 364, 35 },
		{ 197, 365, 35 },
		{ 197, 366, 35 },
		{ 197, 367, 35 },
		{ 197, 368, 35 },
		{ 197, 369, 35 },
		{ 197, 370, 35 },
		{ 197, 371, 35 },
		{ 197, 372, 35 },
		{ 197, 373, 35 },
		{ 197, 374, 35 },
		{ 197, 375, 35 },
		{ 197, 376, 35 },
		{ 198, 336, 35 },
		{ 198, 337, 35 },
		{ 198, 338, 35 },
		{ 198, 339, 35 },
		{ 198, 340, 35 },
		{ 198, 341, 35 },
		{ 198, 342, 35 },
		{ 198, 343, 35 },
		{ 198, 344, 35 },
		{ 198, 345, 35 },
		{ 198, 346, 35 },
		{ 198, 347, 35 },
		{ 198, 348, 35 },
		{ 198, 349, 35 },
		{ 198, 350, 35 },
		{ 198, 351, 35 },
		{ 198, 352, 35 },
		{ 198, 353, 35 },
		{ 198, 354, 35 },
		{ 198, 355, 35 },
		{ 198, 356, 35 },
		{ 198, 357, 35 },
		{ 198, 358, 35 },
		{ 198, 359, 35 },
		{ 198, 360, 35 },
		{ 198, 361, 35 },
		{ 198, 362, 35 },
		{ 198, 363, 35 },
		{ 198, 364, 35 },
		{ 198, 365, 35 },
		{ 198, 366, 35 },
		{ 198, 367, 35 },
		{ 198, 368, 35 },
		{ 198, 369, 35 },
		{ 198, 370, 35 },
		{ 198, 371, 35 },
		{ 198, 372, 35 },
		{ 198, 373, 35 },
		{ 198, 374, 35 },
		{ 199, 335, 35 },
		{ 199, 336, 35 },
		{ 199, 337, 35 },
		{ 199, 338, 35 },
		{ 199, 339, 35 },
		{ 199, 340, 35 },
		{ 199, 341, 35 },
		{ 199, 342, 35 },
		{ 199, 343, 35 },
		{ 199, 344, 35 },
		{ 199, 345, 35 },
		{ 199, 346, 35 },
		{ 199, 347, 35 },
		{ 199, 348, 35 },
		{ 199, 349, 35 },
		{ 199, 350, 35 },
		{ 199, 351, 35 },
		{ 199, 352, 35 },
		{ 199, 353, 35 },
		{ 199, 354, 35 },
		{ 199, 355, 35 },
		{ 199, 356, 35 },
		{ 199, 357, 35 },
		{ 199, 358, 35 },
		{ 199, 359, 35 },
		{ 199, 360, 35 },
		{ 199, 361, 35 },
		{ 199, 362, 35 },
		{ 199, 363, 35 },
		{ 199, 364, 35 },
		{ 199, 365, 35 },
		{ 199, 366, 35 },
		{ 199, 367, 35 },
		{ 199, 368, 35 },
		{ 199, 369, 35 },
		{ 199, 370, 35 },
		{ 199, 371, 35 },
		{ 199, 372, 35 },
		{ 200, 334, 35 },
		{ 200, 335, 35 },
		{ 200, 336, 35 },
		{ 200, 337, 35 },
		{ 200, 338, 35 },
		{ 200, 339, 35 },
		{ 200, 340, 35 },
		{ 200, 341, 35 },
		{ 200, 342, 35 },
		{ 200, 343, 35 },
		{ 200, 344, 35 },
		{ 200, 345, 35 },
		{ 200, 346, 35 },
		{ 200, 347, 35 },
		{ 200, 348, 35 },
		{ 200, 349, 35 },
		{ 200, 350, 35 },
		{ 200, 351, 35 },
		{ 200, 352, 35 },
		{ 200, 353, 35 },
		{ 200, 354, 35 },
		{ 200, 355, 35 },
		{ 200, 356, 35 },
		{ 200, 357, 35 },
		{ 200, 358, 35 },
		{ 200, 359, 35 },
		{ 200, 360, 35 },
		{ 200, 361, 35 },
		{ 200, 362, 35 },
		{ 200, 363, 35 },
		{ 200, 364, 35 },
		{ 200, 365, 35 },
		{ 200, 366, 35 },
		{ 200, 367, 35 },
		{ 200, 368, 35 },
		{ 200, 369, 35 },
		{ 200, 370, 35 },
		{ 201, 332, 35 },
		{ 201, 333, 35 },
		{ 201, 334, 35 },
		{ 201, 335, 35 },
		{ 201, 336, 35 },
		{ 201, 337, 35 },
		{ 201, 338, 35 },
		{ 201, 339, 35 },
		{ 201, 340, 35 },
		{ 201, 341, 35 },
		{ 201, 342, 35 },
		{ 201, 343, 35 },
		{ 201, 344, 35 },
		{ 201, 345, 35 },
		{ 201, 346, 35 },
		{ 201, 347, 35 },
		{ 201, 348, 35 },
		{ 201, 349, 35 },
		{ 201, 350, 35 },
		{ 201, 351, 35 },
		{ 201, 352, 35 },
		{ 201, 353, 35 },
		{ 201, 354, 35 },
		{ 201, 355, 35 },
		{ 201, 356, 35 },
		{ 201, 357, 35 },
		{ 201, 358, 35 },
		{ 201, 359, 35 },
		{ 201, 360, 35 },
		{ 201, 361, 35 },
		{ 201, 362, 35 },
		{ 201, 363, 35 },
		{ 201, 364, 35 },
		{ 201, 365, 35 },
		{ 201, 366, 35 },
		{ 201, 367, 35 },
		{ 201, 368, 35 },
		{ 202, 331, 35 },
		{ 202, 332, 35 },
		{ 202, 333, 35 },
		{ 202, 334, 35 },
		{ 202, 335, 35 },
		{ 202, 336, 35 },
		{ 202, 337, 35 },
		{ 202, 338, 35 },
		{ 202, 339, 35 },
		{ 202, 340, 35 },
		{ 202, 341, 35 },
		{ 202, 342, 35 },
		{ 202, 343, 35 },
		{ 202, 344, 35 },
		{ 202, 345, 35 },
		{ 202, 346, 35 },
		{ 202, 347, 35 },
		{ 202, 348, 35 },
		{ 202, 349, 35 },
		{ 202, 350, 35 },
		{ 202, 351, 35 },
		{ 202, 352, 35 },
		{ 202, 353, 35 },
		{ 202, 354, 35 },
		{ 202, 355, 35 },
		{ 202, 356, 35 },
		{ 202, 357, 35 },
		{ 202, 358, 35 },
		{ 202, 359, 35 },
		{ 202, 360, 35 },
		{ 202, 361, 35 },
		{ 202, 362, 35 },
		{ 202, 363, 35 },
		{ 202, 364, 35 },
		{ 202, 365, 35 },
		{ 202, 366, 35 },
		{ 203, 330, 35 },
		{ 203, 331, 35 },
		{ 203, 332, 35 },
		{ 203, 333, 35 },
		{ 203, 334, 35 },
		{ 203, 335, 35 },
		{ 203, 336, 35 },
		{ 203, 337, 35 },
		{ 203, 338, 35 },
		{ 203, 339, 35 },
		{ 203, 340, 35 },
		{ 203, 341, 35 },
		{ 203, 342, 35 },
		{ 203, 343, 35 },
		{ 203, 344, 35 },
		{ 203, 345, 35 },
		{ 203, 346, 35 },
		{ 203, 347, 35 },
		{ 203, 348, 35 },
		{ 203, 349, 35 },
		{ 203, 350, 35 },
		{ 203, 351, 35 },
		{ 203, 352, 35 },
		{ 203, 353, 35 },
		{ 203, 354, 35 },
		{ 203, 355, 35 },
		{ 203, 356, 35 },
		{ 203, 357, 35 },
		{ 203, 358, 35 },
		{ 203, 359, 35 },
		{ 203, 360, 35 },
		{ 203, 361, 35 },
		{ 203, 362, 35 },
		{ 203, 363, 35 },
		{ 203, 364, 35 },
		{ 204, 328, 35 },
		{ 204, 329, 35 },
		{ 204, 330, 35 },
		{ 204, 331, 35 },
		{ 204, 332, 35 },
		{ 204, 333, 35 },
		{ 204, 334, 35 },
		{ 204, 335, 35 },
		{ 204, 336, 35 },
		{ 204, 337, 35 },
		{ 204, 338, 35 },
		{ 204, 339, 35 },
		{ 204, 340, 35 },
		{ 204, 341, 35 },
		{ 204, 342, 35 },
		{ 204, 343, 35 },
		{ 204, 344, 35 },
		{ 204, 345, 35 },
		{ 204, 346, 35 },
		{ 204, 347, 35 },
		{ 204, 348, 35 },
		{ 204, 349, 35 },
		{ 204, 350, 35 },
		{ 204, 351, 35 },
		{ 204, 352, 35 },
		{ 204, 353, 35 },
		{ 204, 354, 35 },
		{ 204, 355, 35 },
		{ 204, 356, 35 },
		{ 204, 357, 35 },
		{ 204, 358, 35 },
		{ 204, 359, 35 },
		{ 204, 360, 35 },
		{ 204, 361, 35 },
		{ 204, 362, 35 },
		{ 205, 327, 35 },
		{ 205, 328, 35 },
		{ 205, 329, 35 },
		{ 205, 330, 35 },
		{ 205, 331, 35 },
		{ 205, 332, 35 },
		{ 205, 333, 35 },
		{ 205, 334, 35 },
		{ 205, 335, 35 },
		{ 205, 336, 35 },
		{ 205, 337, 35 },
		{ 205, 338, 35 },
		{ 205, 339, 35 },
		{ 205, 340, 35 },
		{ 205, 341, 35 },
		{ 205, 342, 35 },
		{ 205, 343, 35 },
		{ 205, 344, 35 },
		{ 205, 345, 35 },
		{ 205, 346, 35 },
		{ 205, 347, 35 },
		{ 205, 348, 35 },
		{ 205, 349, 35 },
		{ 205, 350, 35 },
		{ 205, 351, 35 },
		{ 205, 352, 35 },
		{ 205, 353, 35 },
		{ 205, 354, 35 },
		{ 205, 355, 35 },
		{ 205, 356, 35 },
		{ 205, 357, 35 },
		{ 205, 358, 35 },
		{ 205, 359, 35 },
		{ 205, 360, 35 },
		{ 206, 326, 35 },
		{ 206, 327, 35 },
		{ 206, 328, 35 },
		{ 206, 329, 35 },
		{ 206, 330, 35 },
		{ 206, 331, 35 },
		{ 206, 332, 35 },
		{ 206, 333, 35 },
		{ 206, 334, 35 },
		{ 206, 335, 35 },
		{ 206, 336, 35 },
		{ 206, 337, 35 },
		{ 206, 338, 35 },
		{ 206, 339, 35 },
		{ 206, 340, 35 },
		{ 206, 341, 35 },
		{ 206, 342, 35 },
		{ 206, 343, 35 },
		{ 206, 344, 35 },
		{ 206, 345, 35 },
		{ 206, 346, 35 },
		{ 206, 347, 35 },
		{ 206, 348, 35 },
		{ 206, 349, 35 },
		{ 206, 350, 35 },
		{ 206, 351, 35 },
		{ 206, 352, 35 },
		{ 206, 353, 35 },
		{ 206, 354, 35 },
		{ 206, 355, 35 },
		{ 206, 356, 35 },
		{ 206, 357, 35 },
		{ 206, 358, 35 },
		{ 207, 325, 35 },
		{ 207, 326, 35 },
		{ 207, 327, 35 },
		{ 207, 328, 35 },
		{ 207, 329, 35 },
		{ 207, 330, 35 },
		{ 207, 331, 35 },
		{ 207, 332, 35 },
		{ 207, 333, 35 },
		{ 207, 334, 35 },
		{ 207, 335, 35 },
		{ 207, 336, 35 },
		{ 207, 337, 35 },
		{ 207, 338, 35 },
		{ 207, 339, 35 },
		{ 207, 340, 35 },
		{ 207, 341, 35 },
		{ 207, 342, 35 },
		{ 207, 343, 35 },
		{ 207, 344, 35 },
		{ 207, 345, 35 },
		{ 207, 346, 35 },
		{ 207, 347, 35 },
		{ 207, 348, 35 },
		{ 207, 349, 35 },
		{ 207, 350, 35 },
		{ 207, 351, 35 },
		{ 207, 352, 35 },
		{ 207, 353, 35 },
		{ 207, 354, 35 },
		{ 207, 355, 35 },
		{ 207, 356, 35 },
		{ 208, 324, 35 },
		{ 208, 325, 35 },
		{ 208, 326, 35 },
		{ 208, 327, 35 },
		{ 208, 328, 35 },
		{ 208, 329, 35 },
		{ 208, 330, 35 },
		{ 208, 331, 35 },
		{ 208, 332, 35 },
		{ 208, 333, 35 },
		{ 208, 334, 35 },
		{ 208, 335, 35 },
		{ 208, 336, 35 },
		{ 208, 337, 35 },
		{ 208, 338, 35 },
		{ 208, 339, 35 },
		{ 208, 340, 35 },
		{ 208, 341, 35 },
		{ 208, 342, 35 },
		{ 208, 343, 35 },
		{ 208, 344, 35 },
		{ 208, 345, 35 },
		{ 208, 346, 35 },
		{ 208, 347, 35 },
		{ 208, 348, 35 },
		{ 208, 349, 35 },
		{ 208, 350, 35 },
		{ 208, 351, 35 },
		{ 208, 352, 35 },
		{ 208, 353, 35 },
		{ 208, 354, 35 },
		{ 209, 324, 35 },
		{ 209, 325, 35 },
		{ 209, 326, 35 },
		{ 209, 327, 35 },
		{ 209, 328, 35 },
		{ 209, 329, 35 },
		{ 209, 330, 35 },
		{ 209, 331, 35 },
		{ 209, 332, 35 },
		{ 209, 333, 35 },
		{ 209, 334, 35 },
		{ 209, 335, 35 },
		{ 209, 336, 35 },
		{ 209, 337, 35 },
		{ 209, 338, 35 },
		{ 209, 339, 35 },
		{ 209, 340, 35 },
		{ 209, 341, 35 },
		{ 209, 342, 35 },
		{ 209, 343, 35 },
		{ 209, 344, 35 },
		{ 209, 345, 35 },
		{ 209, 346, 35 },
		{ 209, 347, 35 },
		{ 209, 348, 35 },
		{ 209, 349, 35 },
		{ 209, 350, 35 },
		{ 209, 351, 35 },
		{ 209, 352, 35 },
		{ 210, 323, 35 },
		{ 210, 324, 35 },
		{ 210, 325, 35 },
		{ 210, 326, 35 },
		{ 210, 327, 35 },
		{ 210, 328, 35 },
		{ 210, 329, 35 },
		{ 210, 330, 35 },
		{ 210, 331, 35 },
		{ 210, 332, 35 },
		{ 210, 333, 35 },
		{ 210, 334, 35 },
		{ 210, 335, 35 },
		{ 210, 336, 35 },
		{ 210, 337, 35 },
		{ 210, 338, 35 },
		{ 210, 339, 35 },
		{ 210, 340, 35 },
		{ 210, 341, 35 },
		{ 210, 342, 35 },
		{ 210, 343, 35 },
		{ 210, 344, 35 },
		{ 210, 345, 35 },
		{ 210, 346, 35 },
		{ 210, 347, 35 },
		{ 210, 348, 35 },
		{ 210, 349, 35 },
		{ 210, 350, 35 },
		{ 211, 323, 35 },
		{ 211, 324, 35 },
		{ 211, 325, 35 },
		{ 211, 326, 35 },
		{ 211, 327, 35 },
		{ 211, 328, 35 },
		{ 211, 329, 35 },
		{ 211, 330, 35 },
		{ 211, 331, 35 },
		{ 211, 332, 35 },
		{ 211, 333, 35 },
		{ 211, 334, 35 },
		{ 211, 335, 35 },
		{ 211, 336, 35 },
		{ 211, 337, 35 },
		{ 211, 338, 35 },
		{ 211, 339, 35 },
		{ 211, 340, 35 },
		{ 211, 341, 35 },
		{ 211, 342, 35 },
		{ 211, 343, 35 },
		{ 211, 344, 35 },
		{ 211, 345, 35 },
		{ 211, 346, 35 },
		{ 211, 347, 35 },
		{ 211, 348, 35 },
		{ 212, 323, 35 },
		{ 212, 324, 35 },
		{ 212, 325, 35 },
		{ 212, 326, 35 },
		{ 212, 327, 35 },
		{ 212, 328, 35 },
		{ 212, 329, 35 },
		{ 212, 330, 35 },
		{ 212, 331, 35 },
		{ 212, 332, 35 },
		{ 212, 333, 35 },
		{ 212, 334, 35 },
		{ 212, 335, 35 },
		{ 212, 336, 35 },
		{ 212, 337, 35 },
		{ 212, 338, 35 },
		{ 212, 339, 35 },
		{ 212, 340, 35 },
		{ 212, 341, 35 },
		{ 212, 342, 35 },
		{ 212, 343, 35 },
		{ 212, 344, 35 },
		{ 212, 345, 35 },
		{ 212, 346, 35 },
		{ 213, 323, 35 },
		{ 213, 324, 35 },
		{ 213, 325, 35 },
		{ 213, 326, 35 },
		{ 213, 327, 35 },
		{ 213, 328, 35 },
		{ 213, 329, 35 },
		{ 213, 330, 35 },
		{ 213, 331, 35 },
		{ 213, 332, 35 },
		{ 213, 333, 35 },
		{ 213, 334, 35 },
		{ 213, 335, 35 },
		{ 213, 336, 35 },
		{ 213, 337, 35 },
		{ 213, 338, 35 },
		{ 213, 339, 35 },
		{ 213, 340, 35 },
		{ 213, 341, 35 },
		{ 213, 342, 35 },
		{ 213, 343, 35 },
		{ 213, 344, 35 },
		{ 214, 323, 35 },
		{ 214, 324, 35 },
		{ 214, 325, 35 },
		{ 214, 326, 35 },
		{ 214, 327, 35 },
		{ 214, 328, 35 },
		{ 214, 329, 35 },
		{ 214, 330, 35 },
		{ 214, 331, 35 },
		{ 214, 332, 35 },
		{ 214, 333, 35 },
		{ 214, 334, 35 },
		{ 214, 335, 35 },
		{ 214, 336, 35 },
		{ 214, 337, 35 },
		{ 214, 338, 35 },
		{ 214, 339, 35 },
		{ 214, 340, 35 },
		{ 214, 341, 35 },
		{ 214, 342, 35 },
		{ 215, 323, 35 },
		{ 215, 324, 35 },
		{ 215, 325, 35 },
		{ 215, 326, 35 },
		{ 215, 327, 35 },
		{ 215, 328, 35 },
		{ 215, 329, 35 },
		{ 215, 330, 35 },
		{ 215, 331, 35 },
		{ 215, 332, 35 },
		{ 215, 333, 35 },
		{ 215, 334, 35 },
		{ 215, 335, 35 },
		{ 215, 336, 35 },
		{ 215, 337, 35 },
		{ 215, 338, 35 },
		{ 215, 339, 35 },
		{ 215, 340, 35 },
		{ 216, 323, 35 },
		{ 216, 324, 35 },
		{ 216, 325, 35 },
		{ 216, 326, 35 },
		{ 216, 327, 35 },
		{ 216, 328, 35 },
		{ 216, 329, 35 },
		{ 216, 330, 35 },
		{ 216, 331, 35 },
		{ 216, 332, 35 },
		{ 216, 333, 35 },
		{ 216, 334, 35 },
		{ 216, 335, 35 },
		{ 216, 336, 35 },
		{ 216, 337, 35 },
		{ 216, 338, 35 },
		{ 217, 323, 35 },
		{ 217, 324, 35 },
		{ 217, 325, 35 },
		{ 217, 326, 35 },
		{ 217, 327, 35 },
		{ 217, 328, 35 },
		{ 217, 329, 35 },
		{ 217, 330, 35 },
		{ 217, 331, 35 },
		{ 217, 332, 35 },
		{ 217, 333, 35 },
		{ 217, 334, 35 },
		{ 217, 335, 35 },
		{ 217, 336, 35 },
		{ 218, 323, 35 },
		{ 218, 324, 35 },
		{ 218, 325, 35 },
		{ 218, 326, 35 },
		{ 218, 327, 35 },
		{ 218, 328, 35 },
		{ 218, 329, 35 },
		{ 218, 330, 35 },
		{ 218, 331, 35 },
		{ 218, 332, 35 },
		{ 218, 333, 35 },
		{ 218, 334, 35 },
		{ 219, 323, 35 },
		{ 219, 324, 35 },
		{ 219, 325, 35 },
		{ 219, 326, 35 },
		{ 219, 327, 35 },
		{ 219, 328, 35 },
		{ 219, 329, 35 },
		{ 219, 330, 35 },
		{ 219, 331, 35 },
		{ 219, 332, 35 },
		{ 220, 323, 35 },
		{ 220, 324, 35 },
		{ 220, 325, 35 },
		{ 220, 326, 35 },
		{ 220, 327, 35 },
		{ 220, 328, 35 },
		{ 220, 329, 35 },
		{ 220, 330, 35 },
		{ 221, 323, 35 },
		{ 221, 324, 35 },
		{ 221, 325, 35 },
		{ 221, 326, 35 },
		{ 221, 327, 35 },
		{ 221, 328, 35 },
		{ 222, 323, 35 },
		{ 222, 324, 35 },
		{ 222, 325, 35 },
		{ 222, 326, 35 },
		{ 292, 188, 35 },
		{ 292, 189, 35 },
		{ 292, 190, 35 },
		{ 292, 191, 35 },
		{ 293, 186, 35 },
		{ 293, 187, 35 },
		{ 293, 188, 35 },
		{ 293, 189, 35 },
		{ 293, 190, 35 },
		{ 293, 191, 35 },
		{ 294, 184, 35 },
		{ 294, 185, 35 },
		{ 294, 186, 35 },
		{ 294, 187, 35 },
		{ 294, 188, 35 },
		{ 294, 189, 35 },
		{ 294, 190, 35 },
		{ 294, 191, 35 },
		{ 295, 182, 35 },
		{ 295, 183, 35 },
		{ 295, 184, 35 },
		{ 295, 185, 35 },
		{ 295, 186, 35 },
		{ 295, 187, 35 },
		{ 295, 188, 35 },
		{ 295, 189, 35 },
		{ 295, 190, 35 },
		{ 295, 191, 35 },
		{ 296, 180, 35 },
		{ 296, 181, 35 },
		{ 296, 182, 35 },
		{ 296, 183, 35 },
		{ 296, 184, 35 },
		{ 296, 185, 35 },
		{ 296, 186, 35 },
		{ 296, 187, 35 },
		{ 296, 188, 35 },
		{ 296, 189, 35 },
		{ 296, 190, 35 },
		{ 296, 191, 35 },
		{ 297, 178, 35 },
		{ 297, 179, 35 },
		{ 297, 180, 35 },
		{ 297, 181, 35 },
		{ 297, 182, 35 },
		{ 297, 183, 35 },
		{ 297, 184, 35 },
		{ 297, 185, 35 },
		{ 297, 186, 35 },
		{ 297, 187, 35 },
		{ 297, 188, 35 },
		{ 297, 189, 35 },
		{ 297, 190, 35 },
		{ 297, 191, 35 },
		{ 298, 176, 35 },
		{ 298, 177, 35 },
		{ 298, 178, 35 },
		{ 298, 179, 35 },
		{ 298, 180, 35 },
		{ 298, 181, 35 },
		{ 298, 182, 35 },
		{ 298, 183, 35 },
		{ 298, 184, 35 },
		{ 298, 185, 35 },
		{ 298, 186, 35 },
		{ 298, 187, 35 },
		{ 298, 188, 35 },
		{ 298, 189, 35 },
		{ 298, 190, 35 },
		{ 298, 191, 35 },
		{ 299, 174, 35 },
		{ 299, 175, 35 },
		{ 299, 176, 35 },
		{ 299, 177, 35 },
		{ 299, 178, 35 },
		{ 299, 179, 35 },
		{ 299, 180, 35 },
		{ 299, 181, 35 },
		{ 299, 182, 35 },
		{ 299, 183, 35 },
		{ 299, 184, 35 },
		{ 299, 185, 35 },
		{ 299, 186, 35 },
		{ 299, 187, 35 },
		{ 299, 188, 35 },
		{ 299, 189, 35 },
		{ 299, 190, 35 },
		{ 299, 191, 35 },
		{ 300, 172, 35 },
		{ 300, 173, 35 },
		{ 300, 174, 35 },
		{ 300, 175, 35 },
		{ 300, 176, 35 },
		{ 300, 177, 35 },
		{ 300, 178, 35 },
		{ 300, 179, 35 },
		{ 300, 180, 35 },
		{ 300, 181, 35 },
		{ 300, 182, 35 },
		{ 300, 183, 35 },
		{ 300, 184, 35 },
		{ 300, 185, 35 },
		{ 300, 186, 35 },
		{ 300, 187, 35 },
		{ 300, 188, 35 },
		{ 300, 189, 35 },
		{ 300, 190, 35 },
		{ 300, 191, 35 },
		{ 301, 170, 35 },
		{ 301, 171, 35 },
		{ 301, 172, 35 },
		{ 301, 173, 35 },
		{ 301, 174, 35 },
		{ 301, 175, 35 },
		{ 301, 176, 35 },
		{ 301, 177, 35 },
		{ 301, 178, 35 },
		{ 301, 179, 35 },
		{ 301, 180, 35 },
		{ 301, 181, 35 },
		{ 301, 182, 35 },
		{ 301, 183, 35 },
		{ 301, 184, 35 },
		{ 301, 185, 35 },
		{ 301, 186, 35 },
		{ 301, 187, 35 },
		{ 301, 188, 35 },
		{ 301, 189, 35 },
		{ 301, 190, 35 },
		{ 301, 191, 35 },
		{ 302, 168, 35 },
		{ 302, 169, 35 },
		{ 302, 170, 35 },
		{ 302, 171, 35 },
		{ 302, 172, 35 },
		{ 302, 173, 35 },
		{ 302, 174, 35 },
		{ 302, 175, 35 },
		{ 302, 176, 35 },
		{ 302, 177, 35 },
		{ 302, 178, 35 },
		{ 302, 179, 35 },
		{ 302, 180, 35 },
		{ 302, 181, 35 },
		{ 302, 182, 35 },
		{ 302, 183, 35 },
		{ 302, 184, 35 },
		{ 302, 185, 35 },
		{ 302, 186, 35 },
		{ 302, 187, 35 },
		{ 302, 188, 35 },
		{ 302, 189, 35 },
		{ 302, 190, 35 },
		{ 302, 191, 35 },
		{ 303, 166, 35 },
		{ 303, 167, 35 },
		{ 303, 168, 35 },
		{ 303, 169, 35 },
		{ 303, 170, 35 },
		{ 303, 171, 35 },
		{ 303, 172, 35 },
		{ 303, 173, 35 },
		{ 303, 174, 35 },
		{ 303, 175, 35 },
		{ 303, 176, 35 },
		{ 303, 177, 35 },
		{ 303, 178, 35 },
		{ 303, 179, 35 },
		{ 303, 180, 35 },
		{ 303, 181, 35 },
		{ 303, 182, 35 },
		{ 303, 183, 35 },
		{ 303, 184, 35 },
		{ 303, 185, 35 },
		{ 303, 186, 35 },
		{ 303, 187, 35 },
		{ 303, 188, 35 },
		{ 303, 189, 35 },
		{ 303, 190, 35 },
		{ 303, 191, 35 },
		{ 304, 164, 35 },
		{ 304, 165, 35 },
		{ 304, 166, 35 },
		{ 304, 167, 35 },
		{ 304, 168, 35 },
		{ 304, 169, 35 },
		{ 304, 170, 35 },
		{ 304, 171, 35 },
		{ 304, 172, 35 },
		{ 304, 173, 35 },
		{ 304, 174, 35 },
		{ 304, 175, 35 },
		{ 304, 176, 35 },
		{ 304, 177, 35 },
		{ 304, 178, 35 },
		{ 304, 179, 35 },
		{ 304, 180, 35 },
		{ 304, 181, 35 },
		{ 304, 182, 35 },
		{ 304, 183, 35 },
		{ 304, 184, 35 },
		{ 304, 185, 35 },
		{ 304, 186, 35 },
		{ 304, 187, 35 },
		{ 304, 188, 35 },
		{ 304, 189, 35 },
		{ 304, 190, 35 },
		{ 304, 191, 35 },
		{ 305, 162, 35 },
		{ 305, 163, 35 },
		{ 305, 164, 35 },
		{ 305, 165, 35 },
		{ 305, 166, 35 },
		{ 305, 167, 35 },
		{ 305, 168, 35 },
		{ 305, 169, 35 },
		{ 305, 170, 35 },
		{ 305, 171, 35 },
		{ 305, 172, 35 },
		{ 305, 173, 35 },
		{ 305, 174, 35 },
		{ 305, 175, 35 },
		{ 305, 176, 35 },
		{ 305, 177, 35 },
		{ 305, 178, 35 },
		{ 305, 179, 35 },
		{ 305, 180, 35 },
		{ 305, 181, 35 },
		{ 305, 182, 35 },
		{ 305, 183, 35 },
		{ 305, 184, 35 },
		{ 305, 185, 35 },
		{ 305, 186, 35 },
		{ 305, 187, 35 },
		{ 305, 188, 35 },
		{ 305, 189, 35 },
		{ 305, 190, 35 },
		{ 306, 160, 35 },
		{ 306, 161, 35 },
		{ 306, 162, 35 },
		{ 306, 163, 35 },
		{ 306, 164, 35 },
		{ 306, 165, 35 },
		{ 306, 166, 35 },
		{ 306, 167, 35 },
		{ 306, 168, 35 },
		{ 306, 169, 35 },
		{ 306, 170, 35 },
		{ 306, 171, 35 },
		{ 306, 172, 35 },
		{ 306, 173, 35 },
		{ 306, 174, 35 },
		{ 306, 175, 35 },
		{ 306, 176, 35 },
		{ 306, 177, 35 },
		{ 306, 178, 35 },
		{ 306, 179, 35 },
		{ 306, 180, 35 },
		{ 306, 181, 35 },
		{ 306, 182, 35 },
		{ 306, 183, 35 },
		{ 306, 184, 35 },
		{ 306, 185, 35 },
		{ 306, 186, 35 },
		{ 306, 187, 35 },
		{ 306, 188, 35 },
		{ 306, 189, 35 },
		{ 306, 190, 35 },
		{ 307, 158, 35 },
		{ 307, 159, 35 },
		{ 307, 160, 35 },
		{ 307, 161, 35 },
		{ 307, 162, 35 },
		{ 307, 163, 35 },
		{ 307, 164, 35 },
		{ 307, 165, 35 },
		{ 307, 166, 35 },
		{ 307, 167, 35 },
		{ 307, 168, 35 },
		{ 307, 169, 35 },
		{ 307, 170, 35 },
		{ 307, 171, 35 },
		{ 307, 172, 35 },
		{ 307, 173, 35 },
		{ 307, 174, 35 },
		{ 307, 175, 35 },
		{ 307, 176, 35 },
		{ 307, 177, 35 },
		{ 307, 178, 35 },
		{ 307, 179, 35 },
		{ 307, 180, 35 },
		{ 307, 181, 35 },
		{ 307, 182, 35 },
		{ 307, 183, 35 },
		{ 307, 184, 35 },
		{ 307, 185, 35 },
		{ 307, 186, 35 },
		{ 307, 187, 35 },
		{ 307, 188, 35 },
		{ 307, 189, 35 },
		{ 308, 156, 35 },
		{ 308, 157, 35 },
		{ 308, 158, 35 },
		{ 308, 159, 35 },
		{ 308, 160, 35 },
		{ 308, 161, 35 },
		{ 308, 162, 35 },
		{ 308, 163, 35 },
		{ 308, 164, 35 },
		{ 308, 165, 35 },
		{ 308, 166, 35 },
		{ 308, 167, 35 },
		{ 308, 168, 35 },
		{ 308, 169, 35 },
		{ 308, 170, 35 },
		{ 308, 171, 35 },
		{ 308, 172, 35 },
		{ 308, 173, 35 },
		{ 308, 174, 35 },
		{ 308, 175, 35 },
		{ 308, 176, 35 },
		{ 308, 177, 35 },
		{ 308, 178, 35 },
		{ 308, 179, 35 },
		{ 308, 180, 35 },
		{ 308, 181, 35 },
		{ 308, 182, 35 },
		{ 308, 183, 35 },
		{ 308, 184, 35 },
		{ 308, 185, 35 },
		{ 308, 186, 35 },
		{ 308, 187, 35 },
		{ 308, 188, 35 },
		{ 309, 154, 35 },
		{ 309, 155, 35 },
		{ 309, 156, 35 },
		{ 309, 157, 35 },
		{ 309, 158, 35 },
		{ 309, 159, 35 },
		{ 309, 160, 35 },
		{ 309, 161, 35 },
		{ 309, 162, 35 },
		{ 309, 163, 35 },
		{ 309, 164, 35 },
		{ 309, 165, 35 },
		{ 309, 166, 35 },
		{ 309, 167, 35 },
		{ 309, 168, 35 },
		{ 309, 169, 35 },
		{ 309, 170, 35 },
		{ 309, 171, 35 },
		{ 309, 172, 35 },
		{ 309, 173, 35 },
		{ 309, 174, 35 },
		{ 309, 175, 35 },
		{ 309, 176, 35 },
		{ 309, 177, 35 },
		{ 309, 178, 35 },
		{ 309, 179, 35 },
		{ 309, 180, 35 },
		{ 309, 181, 35 },
		{ 309, 182, 35 },
		{ 309, 183, 35 },
		{ 309, 184, 35 },
		{ 309, 185, 35 },
		{ 309, 186, 35 },
		{ 309, 187, 35 },
		{ 310, 152, 35 },
		{ 310, 153, 35 },
		{ 310, 154, 35 },
		{ 310, 155, 35 },
		{ 310, 156, 35 },
		{ 310, 157, 35 },
		{ 310, 158, 35 },
		{ 310, 159, 35 },
		{ 310, 160, 35 },
		{ 310, 161, 35 },
		{ 310, 162, 35 },
		{ 310, 163, 35 },
		{ 310, 164, 35 },
		{ 310, 165, 35 },
		{ 310, 166, 35 },
		{ 310, 167, 35 },
		{ 310, 168, 35 },
		{ 310, 169, 35 },
		{ 310, 170, 35 },
		{ 310, 171, 35 },
		{ 310, 172, 35 },
		{ 310, 173, 35 },
		{ 310, 174, 35 },
		{ 310, 175, 35 },
		{ 310, 176, 35 },
		{ 310, 177, 35 },
		{ 310, 178, 35 },
		{ 310, 179, 35 },
		{ 310, 180, 35 },
		{ 310, 181, 35 },
		{ 310, 182, 35 },
		{ 310, 183, 35 },
		{ 310, 184, 35 },
		{ 310, 185, 35 },
		{ 310, 186, 35 },
		{ 311, 150, 35 },
		{ 311, 151, 35 },
		{ 311, 152, 35 },
		{ 311, 153, 35 },
		{ 311, 154, 35 },
		{ 311, 155, 35 },
		{ 311, 156, 35 },
		{ 311, 157, 35 },
		{ 311, 158, 35 },
		{ 311, 159, 35 },
		{ 311, 160, 35 },
		{ 311, 161, 35 },
		{ 311, 162, 35 },
		{ 311, 163, 35 },
		{ 311, 164, 35 },
		{ 311, 165, 35 },
		{ 311, 166, 35 },
		{ 311, 167, 35 },
		{ 311, 168, 35 },
		{ 311, 169, 35 },
		{ 311, 170, 35 },
		{ 311, 171, 35 },
		{ 311, 172, 35 },
		{ 311, 173, 35 },
		{ 311, 174, 35 },
		{ 311, 175, 35 },
		{ 311, 176, 35 },
		{ 311, 177, 35 },
		{ 311, 178, 35 },
		{ 311, 179, 35 },
		{ 311, 180, 35 },
		{ 311, 181, 35 },
		{ 311, 182, 35 },
		{ 311, 183, 35 },
		{ 311, 184, 35 },
		{ 312, 148, 35 },
		{ 312, 149, 35 },
		{ 312, 150, 35 },
		{ 312, 151, 35 },
		{ 312, 152, 35 },
		{ 312, 153, 35 },
		{ 312, 154, 35 },
		{ 312, 155, 35 },
		{ 312, 156, 35 },
		{ 312, 157, 35 },
		{ 312, 158, 35 },
		{ 312, 159, 35 },
		{ 312, 160, 35 },
		{ 312, 161, 35 },
		{ 312, 162, 35 },
		{ 312, 163, 35 },
		{ 312, 164, 35 },
		{ 312, 165, 35 },
		{ 312, 166, 35 },
		{ 312, 167, 35 },
		{ 312, 168, 35 },
		{ 312, 169, 35 },
		{ 312, 170, 35 },
		{ 312, 171, 35 },
		{ 312, 172, 35 },
		{ 312, 173, 35 },
		{ 312, 174, 35 },
		{ 312, 175, 35 },
		{ 312, 176, 35 },
		{ 312, 177, 35 },
		{ 312, 178, 35 },
		{ 312, 179, 35 },
		{ 312, 180, 35 },
		{ 312, 181, 35 },
		{ 312, 182, 35 },
		{ 312, 183, 35 },
		{ 313, 146, 35 },
		{ 313, 147, 35 },
		{ 313, 148, 35 },
		{ 313, 149, 35 },
		{ 313, 150, 35 },
		{ 313, 151, 35 },
		{ 313, 152, 35 },
		{ 313, 153, 35 },
		{ 313, 154, 35 },
		{ 313, 155, 35 },
		{ 313, 156, 35 },
		{ 313, 157, 35 },
		{ 313, 158, 35 },
		{ 313, 159, 35 },
		{ 313, 160, 35 },
		{ 313, 161, 35 },
		{ 313, 162, 35 },
		{ 313, 163, 35 },
		{ 313, 164, 35 },
		{ 313, 165, 35 },
		{ 313, 166, 35 },
		{ 313, 167, 35 },
		{ 313, 168, 35 },
		{ 313, 169, 35 },
		{ 313, 170, 35 },
		{ 313, 171, 35 },
		{ 313, 172, 35 },
		{ 313, 173, 35 },
		{ 313, 174, 35 },
		{ 313, 175, 35 },
		{ 313, 176, 35 },
		{ 313, 177, 35 },
		{ 313, 178, 35 },
		{ 313, 179, 35 },
		{ 313, 180, 35 },
		{ 313, 181, 35 },
		{ 313, 182, 35 },
		{ 314, 144, 35 },
		{ 314, 145, 35 },
		{ 314, 146, 35 },
		{ 314, 147, 35 },
		{ 314, 148, 35 },
		{ 314, 149, 35 },
		{ 314, 150, 35 },
		{ 314, 151, 35 },
		{ 314, 152, 35 },
		{ 314, 153, 35 },
		{ 314, 154, 35 },
		{ 314, 155, 35 },
		{ 314, 156, 35 },
		{ 314, 157, 35 },
		{ 314, 158, 35 },
		{ 314, 159, 35 },
		{ 314, 160, 35 },
		{ 314, 161, 35 },
		{ 314, 162, 35 },
		{ 314, 163, 35 },
		{ 314, 164, 35 },
		{ 314, 165, 35 },
		{ 314, 166, 35 },
		{ 314, 167, 35 },
		{ 314, 168, 35 },
		{ 314, 169, 35 },
		{ 314, 170, 35 },
		{ 314, 171, 35 },
		{ 314, 172, 35 },
		{ 314, 173, 35 },
		{ 314, 174, 35 },
		{ 314, 175, 35 },
		{ 314, 176, 35 },
		{ 314, 177, 35 },
		{ 314, 178, 35 },
		{ 314, 179, 35 },
		{ 314, 180, 35 },
		{ 315, 142, 35 },
		{ 315, 143, 35 },
		{ 315, 144, 35 },
		{ 315, 145, 35 },
		{ 315, 146, 35 },
		{ 315, 147, 35 },
		{ 315, 148, 35 },
		{ 315, 149, 35 },
		{ 315, 150, 35 },
		{ 315, 151, 35 },
		{ 315, 152, 35 },
		{ 315, 153, 35 },
		{ 315, 154, 35 },
		{ 315, 155, 35 },
		{ 315, 156, 35 },
		{ 315, 157, 35 },
		{ 315, 158, 35 },
		{ 315, 159, 35 },
		{ 315, 160, 35 },
		{ 315, 161, 35 },
		{ 315, 162, 35 },
		{ 315, 163, 35 },
		{ 315, 164, 35 },
		{ 315, 165, 35 },
		{ 315, 166, 35 },
		{ 315, 167, 35 },
		{ 315, 168, 35 },
		{ 315, 169, 35 },
		{ 315, 170, 35 },
		{ 315, 171, 35 },
		{ 315, 172, 35 },
		{ 315, 173, 35 },
		{ 315, 174, 35 },
		{ 315, 175, 35 },
		{ 315, 176, 35 },
		{ 315, 177, 35 },
		{ 315, 178, 35 },
		{ 315, 179, 35 },
		{ 316, 140, 35 },
		{ 316, 141, 35 },
		{ 316, 142, 35 },
		{ 316, 143, 35 },
		{ 316, 144, 35 },
		{ 316, 145, 35 },
		{ 316, 146, 35 },
		{ 316, 147, 35 },
		{ 316, 148, 35 },
		{ 316, 149, 35 },
		{ 316, 150, 35 },
		{ 316, 151, 35 },
		{ 316, 152, 35 },
		{ 316, 153, 35 },
		{ 316, 154, 35 },
		{ 316, 155, 35 },
		{ 316, 156, 35 },
		{ 316, 157, 35 },
		{ 316, 158, 35 },
		{ 316, 159, 35 },
		{ 316, 160, 35 },
		{ 316, 161, 35 },
		{ 316, 162, 35 },
		{ 316, 163, 35 },
		{ 316, 164, 35 },
		{ 316, 165, 35 },
		{ 316, 166, 35 },
		{ 316, 167, 35 },
		{ 316, 168, 35 },
		{ 316, 169, 35 },
		{ 316, 170, 35 },
		{ 316, 171, 35 },
		{ 316, 172, 35 },
		{ 316, 173, 35 },
		{ 316, 174, 35 },
		{ 316, 175, 35 },
		{ 316, 176, 35 },
		{ 316, 177, 35 },
		{ 316, 178, 35 },
		{ 317, 138, 35 },
		{ 317, 139, 35 },
		{ 317, 140, 35 },
		{ 317, 141, 35 },
		{ 317, 142, 35 },
		{ 317, 143, 35 },
		{ 317, 144, 35 },
		{ 317, 145, 35 },
		{ 317, 146, 35 },
		{ 317, 147, 35 },
		{ 317, 148, 35 },
		{ 317, 149, 35 },
		{ 317, 150, 35 },
		{ 317, 151, 35 },
		{ 317, 152, 35 },
		{ 317, 153, 35 },
		{ 317, 154, 35 },
		{ 317, 155, 35 },
		{ 317, 156, 35 },
		{ 317, 157, 35 },
		{ 317, 158, 35 },
		{ 317, 159, 35 },
		{ 317, 160, 35 },
		{ 317, 161, 35 },
		{ 317, 162, 35 },
		{ 317, 163, 35 },
		{ 317, 164, 35 },
		{ 317, 165, 35 },
		{ 317, 166, 35 },
		{ 317, 167, 35 },
		{ 317, 168, 35 },
		{ 317, 169, 35 },
		{ 317, 170, 35 },
		{ 317, 171, 35 },
		{ 317, 172, 35 },
		{ 317, 173, 35 },
		{ 317, 174, 35 },
		{ 317, 175, 35 },
		{ 317, 176, 35 },
		{ 318, 136, 35 },
		{ 318, 137, 35 },
		{ 318, 138, 35 },
		{ 318, 139, 35 },
		{ 318, 140, 35 },
		{ 318, 141, 35 },
		{ 318, 142, 35 },
		{ 318, 143, 35 },
		{ 318, 144, 35 },
		{ 318, 145, 35 },
		{ 318, 146, 35 },
		{ 318, 147, 35 },
		{ 318, 148, 35 },
		{ 318, 149, 35 },
		{ 318, 150, 35 },
		{ 318, 151, 35 },
		{ 318, 152, 35 },
		{ 318, 153, 35 },
		{ 318, 154, 35 },
		{ 318, 155, 35 },
		{ 318, 156, 35 },
		{ 318, 157, 35 },
		{ 318, 158, 35 },
		{ 318, 159, 35 },
		{ 318, 160, 35 },
		{ 318, 161, 35 },
		{ 318, 162, 35 },
		{ 318, 163, 35 },
		{ 318, 164, 35 },
		{ 318, 165, 35 },
		{ 318, 166, 35 },
		{ 318, 167, 35 },
		{ 318, 168, 35 },
		{ 318, 169, 35 },
		{ 318, 170, 35 },
		{ 318, 171, 35 },
		{ 318, 172, 35 },
		{ 318, 173, 35 },
		{ 318, 174, 35 },
		{ 318, 175, 35 },
		{ 319, 134, 35 },
		{ 319, 135, 35 },
		{ 319, 136, 35 },
		{ 319, 137, 35 },
		{ 319, 138, 35 },
		{ 319, 139, 35 },
		{ 319, 140, 35 },
		{ 319, 141, 35 },
		{ 319, 142, 35 },
		{ 319, 143, 35 },
		{ 319, 144, 35 },
		{ 319, 145, 35 },
		{ 319, 146, 35 },
		{ 319, 147, 35 },
		{ 319, 148, 35 },
		{ 319, 149, 35 },
		{ 319, 150, 35 },
		{ 319, 151, 35 },
		{ 319, 152, 35 },
		{ 319, 153, 35 },
		{ 319, 154, 35 },
		{ 319, 155, 35 },
		{ 319, 156, 35 },
		{ 319, 157, 35 },
		{ 319, 158, 35 },
		{ 319, 159, 35 },
		{ 319, 160, 35 },
		{ 319, 161, 35 },
		{ 319, 162, 35 },
		{ 319, 163, 35 },
		{ 319, 164, 35 },
		{ 319, 165, 35 },
		{ 319, 166, 35 },
		{ 319, 167, 35 },
		{ 319, 168, 35 },
		{ 319, 169, 35 },
		{ 319, 170, 35 },
		{ 319, 171, 35 },
		{ 319, 172, 35 },
		{ 319, 173, 35 },
		{ 319, 174, 35 },
		{ 320, 132, 35 },
		{ 320, 133, 35 },
		{ 320, 134, 35 },
		{ 320, 135, 35 },
		{ 320, 136, 35 },
		{ 320, 137, 35 },
		{ 320, 138, 35 },
		{ 320, 139, 35 },
		{ 320, 140, 35 },
		{ 320, 141, 35 },
		{ 320, 142, 35 },
		{ 320, 143, 35 },
		{ 320, 144, 35 },
		{ 320, 145, 35 },
		{ 320, 146, 35 },
		{ 320, 147, 35 },
		{ 320, 148, 35 },
		{ 320, 149, 35 },
		{ 320, 150, 35 },
		{ 320, 151, 35 },
		{ 320, 152, 35 },
		{ 320, 153, 35 },
		{ 320, 154, 35 },
		{ 320, 155, 35 },
		{ 320, 156, 35 },
		{ 320, 157, 35 },
		{ 320, 158, 35 },
		{ 320, 159, 35 },
		{ 320, 160, 35 },
		{ 320, 161, 35 },
		{ 320, 162, 35 },
		{ 320, 163, 35 },
		{ 320, 164, 35 },
		{ 320, 165, 35 },
		{ 320, 166, 35 },
		{ 320, 167, 35 },
		{ 320, 168, 35 },
		{ 320, 169, 35 },
		{ 320, 170, 35 },
		{ 320, 171, 35 },
		{ 320, 172, 35 },
		{ 321, 130, 35 },
		{ 321, 131, 35 },
		{ 321, 132, 35 },
		{ 321, 133, 35 },
		{ 321, 134, 35 },
		{ 321, 135, 35 },
		{ 321, 136, 35 },
		{ 321, 137, 35 },
		{ 321, 138, 35 },
		{ 321, 139, 35 },
		{ 321, 140, 35 },
		{ 321, 141, 35 },
		{ 321, 142, 35 },
		{ 321, 143, 35 },
		{ 321, 144, 35 },
		{ 321, 145, 35 },
		{ 321, 146, 35 },
		{ 321, 147, 35 },
		{ 321, 148, 35 },
		{ 321, 149, 35 },
		{ 321, 150, 35 },
		{ 321, 151, 35 },
		{ 321, 152, 35 },
		{ 321, 153, 35 },
		{ 321, 154, 35 },
		{ 321, 155, 35 },
		{ 321, 156, 35 },
		{ 321, 157, 35 },
		{ 321, 158, 35 },
		{ 321, 159, 35 },
		{ 321, 160, 35 },
		{ 321, 161, 35 },
		{ 321, 162, 35 },
		{ 321, 163, 35 },
		{ 321, 164, 35 },
		{ 321, 165, 35 },
		{ 321, 166, 35 },
		{ 321, 167, 35 },
		{ 321, 168, 35 },
		{ 321, 169, 35 },
		{ 321, 170, 35 },
		{ 321, 171, 35 },
		{ 322, 128, 35 },
		{ 322, 129, 35 },
		{ 322, 130, 35 },
		{ 322, 131, 35 },
		{ 322, 132, 35 },
		{ 322, 133, 35 },
		{ 322, 134, 35 },
		{ 322, 135, 35 },
		{ 322, 136, 35 },
		{ 322, 137, 35 },
		{ 322, 138, 35 },
		{ 322, 139, 35 },
		{ 322, 140, 35 },
		{ 322, 141, 35 },
		{ 322, 142, 35 },
		{ 322, 143, 35 },
		{ 322, 144, 35 },
		{ 322, 145, 35 },
		{ 322, 146, 35 },
		{ 322, 147, 35 },
		{ 322, 148, 35 },
		{ 322, 149, 35 },
		{ 322, 150, 35 },
		{ 322, 151, 35 },
		{ 322, 152, 35 },
		{ 322, 153, 35 },
		{ 322, 154, 35 },
		{ 322, 155, 35 },
		{ 322, 156, 35 },
		{ 322, 157, 35 },
		{ 322, 158, 35 },
		{ 322, 159, 35 },
		{ 322, 160, 35 },
		{ 322, 161, 35 },
		{ 322, 162, 35 },
		{ 322, 163, 35 },
		{ 322, 164, 35 },
		{ 322, 165, 35 },
		{ 322, 166, 35 },
		{ 322, 167, 35 },
		{ 322, 168, 35 },
		{ 322, 169, 35 },
		{ 322, 170, 35 },
		{ 323, 126, 35 },
		{ 323, 127, 35 },
		{ 323, 128, 35 },
		{ 323, 129, 35 },
		{ 323, 130, 35 },
		{ 323, 131, 35 },
		{ 323, 132, 35 },
		{ 323, 133, 35 },
		{ 323, 134, 35 },
		{ 323, 135, 35 },
		{ 323, 136, 35 },
		{ 323, 137, 35 },
		{ 323, 138, 35 },
		{ 323, 139, 35 },
		{ 323, 140, 35 },
		{ 323, 141, 35 },
		{ 323, 142, 35 },
		{ 323, 143, 35 },
		{ 323, 144, 35 },
		{ 323, 145, 35 },
		{ 323, 146, 35 },
		{ 323, 147, 35 },
		{ 323, 148, 35 },
		{ 323, 149, 35 },
		{ 323, 150, 35 },
		{ 323, 151, 35 },
		{ 323, 152, 35 },
		{ 323, 153, 35 },
		{ 323, 154, 35 },
		{ 323, 155, 35 },
		{ 323, 156, 35 },
		{ 323, 157, 35 },
		{ 323, 158, 35 },
		{ 323, 159, 35 },
		{ 323, 160, 35 },
		{ 323, 161, 35 },
		{ 323, 162, 35 },
		{ 323, 163, 35 },
		{ 323, 164, 35 },
		{ 323, 165, 35 },
		{ 323, 166, 35 },
		{ 323, 167, 35 },
		{ 323, 168, 35 },
		{ 324, 124, 35 },
		{ 324, 125, 35 },
		{ 324, 126, 35 },
		{ 324, 127, 35 },
		{ 324, 128, 35 },
		{ 324, 129, 35 },
		{ 324, 130, 35 },
		{ 324, 131, 35 },
		{ 324, 132, 35 },
		{ 324, 133, 35 },
		{ 324, 134, 35 },
		{ 324, 135, 35 },
		{ 324, 136, 35 },
		{ 324, 137, 35 },
		{ 324, 138, 35 },
		{ 324, 139, 35 },
		{ 324, 140, 35 },
		{ 324, 141, 35 },
		{ 324, 142, 35 },
		{ 324, 143, 35 },
		{ 324, 144, 35 },
		{ 324, 145, 35 },
		{ 324, 146, 35 },
		{ 324, 147, 35 },
		{ 324, 148, 35 },
		{ 324, 149, 35 },
		{ 324, 150, 35 },
		{ 324, 151, 35 },
		{ 324, 152, 35 },
		{ 324, 153, 35 },
		{ 324, 154, 35 },
		{ 324, 155, 35 },
		{ 324, 156, 35 },
		{ 324, 157, 35 },
		{ 324, 158, 35 },
		{ 324, 159, 35 },
		{ 324, 160, 35 },
		{ 324, 161, 35 },
		{ 324, 162, 35 },
		{ 324, 163, 35 },
		{ 324, 164, 35 },
		{ 324, 165, 35 },
		{ 324, 166, 35 },
		{ 324, 167, 35 },
		{ 325, 122, 35 },
		{ 325, 123, 35 },
		{ 325, 124, 35 },
		{ 325, 125, 35 },
		{ 325, 126, 35 },
		{ 325, 127, 35 },
		{ 325, 128, 35 },
		{ 325, 129, 35 },
		{ 325, 130, 35 },
		{ 325, 131, 35 },
		{ 325, 132, 35 },
		{ 325, 133, 35 },
		{ 325, 134, 35 },
		{ 325, 135, 35 },
		{ 325, 136, 35 },
		{ 325, 137, 35 },
		{ 325, 138, 35 },
		{ 325, 139, 35 },
		{ 325, 140, 35 },
		{ 325, 141, 35 },
		{ 325, 142, 35 },
		{ 325, 143, 35 },
		{ 325, 144, 35 },
		{ 325, 145, 35 },
		{ 325, 146, 35 },
		{ 325, 147, 35 },
		{ 325, 148, 35 },
		{ 325, 149, 35 },
		{ 325, 150, 35 },
		{ 325, 151, 35 },
		{ 325, 152, 35 },
		{ 325, 153, 35 },
		{ 325, 154, 35 },
		{ 325, 155, 35 },
		{ 325, 156, 35 },
		{ 325, 157, 35 },
		{ 325, 158, 35 },
		{ 325, 159, 35 },
		{ 325, 160, 35 },
		{ 325, 161, 35 },
		{ 325, 162, 35 },
		{ 325, 163, 35 },
		{ 325, 164, 35 },
		{ 325, 165, 35 },
		{ 325, 166, 35 },
		{ 326, 120, 35 },
		{ 326, 121, 35 },
		{ 326, 122, 35 },
		{ 326, 123, 35 },
		{ 326, 124, 35 },
		{ 326, 125, 35 },
		{ 326, 126, 35 },
		{ 326, 127, 35 },
		{ 326, 128, 35 },
		{ 326, 129, 35 },
		{ 326, 130, 35 },
		{ 326, 131, 35 },
		{ 326, 132, 35 },
		{ 326, 133, 35 },
		{ 326, 134, 35 },
		{ 326, 135, 35 },
		{ 326, 136, 35 },
		{ 326, 137, 35 },
		{ 326, 138, 35 },
		{ 326, 139, 35 },
		{ 326, 140, 35 },
		{ 326, 141, 35 },
		{ 326, 142, 35 },
		{ 326, 143, 35 },
		{ 326, 144, 35 },
		{ 326, 145, 35 },
		{ 326, 146, 35 },
		{ 326, 147, 35 },
		{ 326, 148, 35 },
		{ 326, 149, 35 },
		{ 326, 150, 35 },
		{ 326, 151, 35 },
		{ 326, 152, 35 },
		{ 326, 153, 35 },
		{ 326, 154, 35 },
		{ 326, 155, 35 },
		{ 326, 156, 35 },
		{ 326, 157, 35 },
		{ 326, 158, 35 },
		{ 326, 159, 35 },
		{ 326, 160, 35 },
		{ 326, 161, 35 },
		{ 326, 162, 35 },
		{ 326, 163, 35 },
		{ 326, 164, 35 },
		{ 327, 118, 35 },
		{ 327, 119, 35 },
		{ 327, 120, 35 },
		{ 327, 121, 35 },
		{ 327, 122, 35 },
		{ 327, 123, 35 },
		{ 327, 124, 35 },
		{ 327, 125, 35 },
		{ 327, 126, 35 },
		{ 327, 127, 35 },
		{ 327, 128, 35 },
		{ 327, 129, 35 },
		{ 327, 130, 35 },
		{ 327, 131, 35 },
		{ 327, 132, 35 },
		{ 327, 133, 35 },
		{ 327, 134, 35 },
		{ 327, 135, 35 },
		{ 327, 136, 35 },
		{ 327, 137, 35 },
		{ 327, 138, 35 },
		{ 327, 139, 35 },
		{ 327, 140, 35 },
		{ 327, 141, 35 },
		{ 327, 142, 35 },
		{ 327, 143, 35 },
		{ 327, 144, 35 },
		{ 327, 145, 35 },
		{ 327, 146, 35 },
		{ 327, 147, 35 },
		{ 327, 148, 35 },
		{ 327, 149, 35 },
		{ 327, 150, 35 },
		{ 327, 151, 35 },
		{ 327, 152, 35 },
		{ 327, 153, 35 },
		{ 327, 154, 35 },
		{ 327, 155, 35 },
		{ 327, 156, 35 },
		{ 327, 157, 35 },
		{ 327, 158, 35 },
		{ 327, 159, 35 },
		{ 327, 160, 35 },
		{ 327, 161, 35 },
		{ 327, 162, 35 },
		{ 327, 163, 35 },
		{ 328, 116, 35 },
		{ 328, 117, 35 },
		{ 328, 118, 35 },
		{ 328, 119, 35 },
		{ 328, 120, 35 },
		{ 328, 121, 35 },
		{ 328, 122, 35 },
		{ 328, 123, 35 },
		{ 328, 124, 35 },
		{ 328, 125, 35 },
		{ 328, 126, 35 },
		{ 328, 127, 35 },
		{ 328, 128, 35 },
		{ 328, 129, 35 },
		{ 328, 130, 35 },
		{ 328, 131, 35 },
		{ 328, 132, 35 },
		{ 328, 133, 35 },
		{ 328, 134, 35 },
		{ 328, 135, 35 },
		{ 328, 136, 35 },
		{ 328, 137, 35 },
		{ 328, 138, 35 },
		{ 328, 139, 35 },
		{ 328, 140, 35 },
		{ 328, 141, 35 },
		{ 328, 142, 35 },
		{ 328, 143, 35 },
		{ 328, 144, 35 },
		{ 328, 145, 35 },
		{ 328, 146, 35 },
		{ 328, 147, 35 },
		{ 328, 148, 35 },
		{ 328, 149, 35 },
		{ 328, 150, 35 },
		{ 328, 151, 35 },
		{ 328, 152, 35 },
		{ 328, 153, 35 },
		{ 328, 154, 35 },
		{ 328, 155, 35 },
		{ 328, 156, 35 },
		{ 328, 157, 35 },
		{ 328, 158, 35 },
		{ 328, 159, 35 },
		{ 328, 160, 35 },
		{ 328, 161, 35 },
		{ 328, 162, 35 },
		{ 329, 114, 35 },
		{ 329, 115, 35 },
		{ 329, 116, 35 },
		{ 329, 117, 35 },
		{ 329, 118, 35 },
		{ 329, 119, 35 },
		{ 329, 120, 35 },
		{ 329, 121, 35 },
		{ 329, 122, 35 },
		{ 329, 123, 35 },
		{ 329, 124, 35 },
		{ 329, 125, 35 },
		{ 329, 126, 35 },
		{ 329, 127, 35 },
		{ 329, 128, 35 },
		{ 329, 129, 35 },
		{ 329, 130, 35 },
		{ 329, 131, 35 },
		{ 329, 132, 35 },
		{ 329, 133, 35 },
		{ 329, 134, 35 },
		{ 329, 135, 35 },
		{ 329, 136, 35 },
		{ 329, 137, 35 },
		{ 329, 138, 35 },
		{ 329, 139, 35 },
		{ 329, 140, 35 },
		{ 329, 141, 35 },
		{ 329, 142, 35 },
		{ 329, 143, 35 },
		{ 329, 144, 35 },
		{ 329, 145, 35 },
		{ 329, 146, 35 },
		{ 329, 147, 35 },
		{ 329, 148, 35 },
		{ 329, 149, 35 },
		{ 329, 150, 35 },
		{ 329, 151, 35 },
		{ 329, 152, 35 },
		{ 329, 153, 35 },
		{ 329, 154, 35 },
		{ 329, 155, 35 },
		{ 329, 156, 35 },
		{ 329, 157, 35 },
		{ 329, 158, 35 },
		{ 329, 159, 35 },
		{ 329, 160, 35 },
		{ 330, 112, 35 },
		{ 330, 113, 35 },
		{ 330, 114, 35 },
		{ 330, 115, 35 },
		{ 330, 116, 35 },
		{ 330, 117, 35 },
		{ 330, 118, 35 },
		{ 330, 119, 35 },
		{ 330, 120, 35 },
		{ 330, 121, 35 },
		{ 330, 122, 35 },
		{ 330, 123, 35 },
		{ 330, 124, 35 },
		{ 330, 125, 35 },
		{ 330, 126, 35 },
		{ 330, 127, 35 },
		{ 330, 128, 35 },
		{ 330, 129, 35 },
		{ 330, 130, 35 },
		{ 330, 131, 35 },
		{ 330, 132, 35 },
		{ 330, 133, 35 },
		{ 330, 134, 35 },
		{ 330, 135, 35 },
		{ 330, 136, 35 },
		{ 330, 137, 35 },
		{ 330, 138, 35 },
		{ 330, 139, 35 },
		{ 330, 140, 35 },
		{ 330, 141, 35 },
		{ 330, 142, 35 },
		{ 330, 143, 35 },
		{ 330, 144, 35 },
		{ 330, 145, 35 },
		{ 330, 146, 35 },
		{ 330, 147, 35 },
		{ 330, 148, 35 },
		{ 330, 149, 35 },
		{ 330, 150, 35 },
		{ 330, 151, 35 },
		{ 330, 152, 35 },
		{ 330, 153, 35 },
		{ 330, 154, 35 },
		{ 330, 155, 35 },
		{ 330, 156, 35 },
		{ 330, 157, 35 },
		{ 330, 158, 35 },
		{ 330, 159, 35 },
		{ 331, 110, 35 },
		{ 331, 111, 35 },
		{ 331, 112, 35 },
		{ 331, 113, 35 },
		{ 331, 114, 35 },
		{ 331, 115, 35 },
		{ 331, 116, 35 },
		{ 331, 117, 35 },
		{ 331, 118, 35 },
		{ 331, 119, 35 },
		{ 331, 120, 35 },
		{ 331, 121, 35 },
		{ 331, 122, 35 },
		{ 331, 123, 35 },
		{ 331, 124, 35 },
		{ 331, 125, 35 },
		{ 331, 126, 35 },
		{ 331, 127, 35 },
		{ 331, 128, 35 },
		{ 331, 129, 35 },
		{ 331, 130, 35 },
		{ 331, 131, 35 },
		{ 331, 132, 35 },
		{ 331, 133, 35 },
		{ 331, 134, 35 },
		{ 331, 135, 35 },
		{ 331, 136, 35 },
		{ 331, 137, 35 },
		{ 331, 138, 35 },
		{ 331, 139, 35 },
		{ 331, 140, 35 },
		{ 331, 141, 35 },
		{ 331, 142, 35 },
		{ 331, 143, 35 },
		{ 331, 144, 35 },
		{ 331, 145, 35 },
		{ 331, 146, 35 },
		{ 331, 147, 35 },
		{ 331, 148, 35 },
		{ 331, 149, 35 },
		{ 331, 150, 35 },
		{ 331, 151, 35 },
		{ 331, 152, 35 },
		{ 331, 153, 35 },
		{ 331, 154, 35 },
		{ 331, 155, 35 },
		{ 331, 156, 35 },
		{ 331, 157, 35 },
		{ 331, 158, 35 },
		{ 332, 108, 35 },
		{ 332, 109, 35 },
		{ 332, 110, 35 },
		{ 332, 111, 35 },
		{ 332, 112, 35 },
		{ 332, 113, 35 },
		{ 332, 114, 35 },
		{ 332, 115, 35 },
		{ 332, 116, 35 },
		{ 332, 117, 35 },
		{ 332, 118, 35 },
		{ 332, 119, 35 },
		{ 332, 120, 35 },
		{ 332, 121, 35 },
		{ 332, 122, 35 },
		{ 332, 123, 35 },
		{ 332, 124, 35 },
		{ 332, 125, 35 },
		{ 332, 126, 35 },
		{ 332, 127, 35 },
		{ 332, 128, 35 },
		{ 332, 129, 35 },
		{ 332, 130, 35 },
		{ 332, 131, 35 },
		{ 332, 132, 35 },
		{ 332, 133, 35 },
		{ 332, 134, 35 },
		{ 332, 135, 35 },
		{ 332, 136, 35 },
		{ 332, 137, 35 },
		{ 332, 138, 35 },
		{ 332, 139, 35 },
		{ 332, 140, 35 },
		{ 332, 141, 35 },
		{ 332, 142, 35 },
		{ 332, 143, 35 },
		{ 332, 144, 35 },
		{ 332, 145, 35 },
		{ 332, 146, 35 },
		{ 332, 147, 35 },
		{ 332, 148, 35 },
		{ 332, 149, 35 },
		{ 332, 150, 35 },
		{ 332, 151, 35 },
		{ 332, 152, 35 },
		{ 332, 153, 35 },
		{ 332, 154, 35 },
		{ 332, 155, 35 },
		{ 332, 156, 35 },
		{ 333, 106, 35 },
		{ 333, 107, 35 },
		{ 333, 108, 35 },
		{ 333, 109, 35 },
		{ 333, 110, 35 },
		{ 333, 111, 35 },
		{ 333, 112, 35 },
		{ 333, 113, 35 },
		{ 333, 114, 35 },
		{ 333, 115, 35 },
		{ 333, 116, 35 },
		{ 333, 117, 35 },
		{ 333, 118, 35 },
		{ 333, 119, 35 },
		{ 333, 120, 35 },
		{ 333, 121, 35 },
		{ 333, 122, 35 },
		{ 333, 123, 35 },
		{ 333, 124, 35 },
		{ 333, 125, 35 },
		{ 333, 126, 35 },
		{ 333, 127, 35 },
		{ 333, 128, 35 },
		{ 333, 129, 35 },
		{ 333, 130, 35 },
		{ 333, 131, 35 },
		{ 333, 132, 35 },
		{ 333, 133, 35 },
		{ 333, 134, 35 },
		{ 333, 135, 35 },
		{ 333, 136, 35 },
		{ 333, 137, 35 },
		{ 333, 138, 35 },
		{ 333, 139, 35 },
		{ 333, 140, 35 },
		{ 333, 141, 35 },
		{ 333, 142, 35 },
		{ 333, 143, 35 },
		{ 333, 144, 35 },
		{ 333, 145, 35 },
		{ 333, 146, 35 },
		{ 333, 147, 35 },
		{ 333, 148, 35 },
		{ 333, 149, 35 },
		{ 333, 150, 35 },
		{ 333, 151, 35 },
		{ 333, 152, 35 },
		{ 333, 153, 35 },
		{ 333, 154, 35 },
		{ 333, 155, 35 },
		{ 334, 104, 35 },
		{ 334, 105, 35 },
		{ 334, 106, 35 },
		{ 334, 107, 35 },
		{ 334, 108, 35 },
		{ 334, 109, 35 },
		{ 334, 110, 35 },
		{ 334, 111, 35 },
		{ 334, 112, 35 },
		{ 334, 113, 35 },
		{ 334, 114, 35 },
		{ 334, 115, 35 },
		{ 334, 116, 35 },
		{ 334, 117, 35 },
		{ 334, 118, 35 },
		{ 334, 119, 35 },
		{ 334, 120, 35 },
		{ 334, 121, 35 },
		{ 334, 122, 35 },
		{ 334, 123, 35 },
		{ 334, 124, 35 },
		{ 334, 125, 35 },
		{ 334, 126, 35 },
		{ 334, 127, 35 },
		{ 334, 128, 35 },
		{ 334, 129, 35 },
		{ 334, 130, 35 },
		{ 334, 131, 35 },
		{ 334, 132, 35 },
		{ 334, 133, 35 },
		{ 334, 134, 35 },
		{ 334, 135, 35 },
		{ 334, 136, 35 },
		{ 334, 137, 35 },
		{ 334, 138, 35 },
		{ 334, 139, 35 },
		{ 334, 140, 35 },
		{ 334, 141, 35 },
		{ 334, 142, 35 },
		{ 334, 143, 35 },
		{ 334, 144, 35 },
		{ 334, 145, 35 },
		{ 334, 146, 35 },
		{ 334, 147, 35 },
		{ 334, 148, 35 },
		{ 334, 149, 35 },
		{ 334, 150, 35 },
		{ 334, 151, 35 },
		{ 334, 152, 35 },
		{ 334, 153, 35 },
		{ 334, 154, 35 },
		{ 335, 102, 35 },
		{ 335, 103, 35 },
		{ 335, 104, 35 },
		{ 335, 105, 35 },
		{ 335, 106, 35 },
		{ 335, 107, 35 },
		{ 335, 108, 35 },
		{ 335, 109, 35 },
		{ 335, 110, 35 },
		{ 335, 111, 35 },
		{ 335, 112, 35 },
		{ 335, 113, 35 },
		{ 335, 114, 35 },
		{ 335, 115, 35 },
		{ 335, 116, 35 },
		{ 335, 117, 35 },
		{ 335, 118, 35 },
		{ 335, 119, 35 },
		{ 335, 120, 35 },
		{ 335, 121, 35 },
		{ 335, 122, 35 },
		{ 335, 123, 35 },
		{ 335, 124, 35 },
		{ 335, 125, 35 },
		{ 335, 126, 35 },
		{ 335, 127, 35 },
		{ 335, 128, 35 },
		{ 335, 129, 35 },
		{ 335, 130, 35 },
		{ 335, 131, 35 },
		{ 335, 132, 35 },
		{ 335, 133, 35 },
		{ 335, 134, 35 },
		{ 335, 135, 35 },
		{ 335, 136, 35 },
		{ 335, 137, 35 },
		{ 335, 138, 35 },
		{ 335, 139, 35 },
		{ 335, 140, 35 },
		{ 335, 141, 35 },
		{ 335, 142, 35 },
		{ 335, 143, 35 },
		{ 335, 144, 35 },
		{ 335, 145, 35 },
		{ 335, 146, 35 },
		{ 335, 147, 35 },
		{ 335, 148, 35 },
		{ 335, 149, 35 },
		{ 335, 150, 35 },
		{ 335, 151, 35 },
		{ 335, 152, 35 },
		{ 336, 100, 35 },
		{ 336, 101, 35 },
		{ 336, 102, 35 },
		{ 336, 103, 35 },
		{ 336, 104, 35 },
		{ 336, 105, 35 },
		{ 336, 106, 35 },
		{ 336, 107, 35 },
		{ 336, 108, 35 },
		{ 336, 109, 35 },
		{ 336, 110, 35 },
		{ 336, 111, 35 },
		{ 336, 112, 35 },
		{ 336, 113, 35 },
		{ 336, 114, 35 },
		{ 336, 115, 35 },
		{ 336, 116, 35 },
		{ 336, 117, 35 },
		{ 336, 118, 35 },
		{ 336, 119, 35 },
		{ 336, 120, 35 },
		{ 336, 121, 35 },
		{ 336, 122, 35 },
		{ 336, 123, 35 },
		{ 336, 124, 35 },
		{ 336, 125, 35 },
		{ 336, 126, 35 },
		{ 336, 127, 35 },
		{ 336, 128, 35 },
		{ 336, 129, 35 },
		{ 336, 130, 35 },
		{ 336, 131, 35 },
		{ 336, 132, 35 },
		{ 336, 133, 35 },
		{ 336, 134, 35 },
		{ 336, 135, 35 },
		{ 336, 136, 35 },
		{ 336, 137, 35 },
		{ 336, 138, 35 },
		{ 336, 139, 35 },
		{ 336, 140, 35 },
		{ 336, 141, 35 },
		{ 336, 142, 35 },
		{ 336, 143, 35 },
		{ 336, 144, 35 },
		{ 336, 145, 35 },
		{ 336, 146, 35 },
		{ 336, 147, 35 },
		{ 336, 148, 35 },
		{ 336, 149, 35 },
		{ 336, 150, 35 },
		{ 336, 151, 35 },
		{ 337, 98, 35 },
		{ 337, 99, 35 },
		{ 337, 100, 35 },
		{ 337, 101, 35 },
		{ 337, 102, 35 },
		{ 337, 103, 35 },
		{ 337, 104, 35 },
		{ 337, 105, 35 },
		{ 337, 106, 35 },
		{ 337, 107, 35 },
		{ 337, 108, 35 },
		{ 337, 109, 35 },
		{ 337, 110, 35 },
		{ 337, 111, 35 },
		{ 337, 112, 35 },
		{ 337, 113, 35 },
		{ 337, 114, 35 },
		{ 337, 115, 35 },
		{ 337, 116, 35 },
		{ 337, 117, 35 },
		{ 337, 118, 35 },
		{ 337, 119, 35 },
		{ 337, 120, 35 },
		{ 337, 121, 35 },
		{ 337, 122, 35 },
		{ 337, 123, 35 },
		{ 337, 124, 35 },
		{ 337, 125, 35 },
		{ 337, 126, 35 },
		{ 337, 127, 35 },
		{ 337, 128, 35 },
		{ 337, 129, 35 },
		{ 337, 130, 35 },
		{ 337, 131, 35 },
		{ 337, 132, 35 },
		{ 337, 133, 35 },
		{ 337, 134, 35 },
		{ 337, 135, 35 },
		{ 337, 136, 35 },
		{ 337, 137, 35 },
		{ 337, 138, 35 },
		{ 337, 139, 35 },
		{ 337, 140, 35 },
		{ 337, 141, 35 },
		{ 337, 142, 35 },
		{ 337, 143, 35 },
		{ 337, 144, 35 },
		{ 337, 145, 35 },
		{ 337, 146, 35 },
		{ 337, 147, 35 },
		{ 337, 148, 35 },
		{ 337, 149, 35 },
		{ 337, 150, 35 },
		{ 338, 96, 35 },
		{ 338, 97, 35 },
		{ 338, 98, 35 },
		{ 338, 99, 35 },
		{ 338, 100, 35 },
		{ 338, 101, 35 },
		{ 338, 102, 35 },
		{ 338, 103, 35 },
		{ 338, 104, 35 },
		{ 338, 105, 35 },
		{ 338, 106, 35 },
		{ 338, 107, 35 },
		{ 338, 108, 35 },
		{ 338, 109, 35 },
		{ 338, 110, 35 },
		{ 338, 111, 35 },
		{ 338, 112, 35 },
		{ 338, 113, 35 },
		{ 338, 114, 35 },
		{ 338, 115, 35 },
		{ 338, 116, 35 },
		{ 338, 117, 35 },
		{ 338, 118, 35 },
		{ 338, 119, 35 },
		{ 338, 120, 35 },
		{ 338, 121, 35 },
		{ 338, 122, 35 },
		{ 338, 123, 35 },
		{ 338, 124, 35 },
		{ 338, 125, 35 },
		{ 338, 126, 35 },
		{ 338, 127, 35 },
		{ 338, 128, 35 },
		{ 338, 129, 35 },
		{ 338, 130, 35 },
		{ 338, 131, 35 },
		{ 338, 132, 35 },
		{ 338, 133, 35 },
		{ 338, 134, 35 },
		{ 338, 135, 35 },
		{ 338, 136, 35 },
		{ 338, 137, 35 },
		{ 338, 138, 35 },
		{ 338, 139, 35 },
		{ 338, 140, 35 },
		{ 338, 141, 35 },
		{ 338, 142, 35 },
		{ 338, 143, 35 },
		{ 338, 144, 35 },
		{ 338, 145, 35 },
		{ 338, 146, 35 },
		{ 338, 147, 35 },
		{ 338, 148, 35 },
		{ 339, 94, 35 },
		{ 339, 95, 35 },
		{ 339, 96, 35 },
		{ 339, 97, 35 },
		{ 339, 98, 35 },
		{ 339, 99, 35 },
		{ 339, 100, 35 },
		{ 339, 101, 35 },
		{ 339, 102, 35 },
		{ 339, 103, 35 },
		{ 339, 104, 35 },
		{ 339, 105, 35 },
		{ 339, 106, 35 },
		{ 339, 107, 35 },
		{ 339, 108, 35 },
		{ 339, 109, 35 },
		{ 339, 110, 35 },
		{ 339, 111, 35 },
		{ 339, 112, 35 },
		{ 339, 113, 35 },
		{ 339, 114, 35 },
		{ 339, 115, 35 },
		{ 339, 116, 35 },
		{ 339, 117, 35 },
		{ 339, 118, 35 },
		{ 339, 119, 35 },
		{ 339, 120, 35 },
		{ 339, 121, 35 },
		{ 339, 122, 35 },
		{ 339, 123, 35 },
		{ 339, 124, 35 },
		{ 339, 125, 35 },
		{ 339, 126, 35 },
		{ 339, 127, 35 },
		{ 339, 128, 35 },
		{ 339, 129, 35 },
		{ 339, 130, 35 },
		{ 339, 131, 35 },
		{ 339, 132, 35 },
		{ 339, 133, 35 },
		{ 339, 134, 35 },
		{ 339, 135, 35 },
		{ 339, 136, 35 },
		{ 339, 137, 35 },
		{ 339, 138, 35 },
		{ 339, 139, 35 },
		{ 339, 140, 35 },
		{ 339, 141, 35 },
		{ 339, 142, 35 },
		{ 339, 143, 35 },
		{ 339, 144, 35 },
		{ 339, 145, 35 },
		{ 339, 146, 35 },
		{ 339, 147, 35 },
		{ 340, 92, 35 },
		{ 340, 93, 35 },
		{ 340, 94, 35 },
		{ 340, 95, 35 },
		{ 340, 96, 35 },
		{ 340, 97, 35 },
		{ 340, 98, 35 },
		{ 340, 99, 35 },
		{ 340, 100, 35 },
		{ 340, 101, 35 },
		{ 340, 102, 35 },
		{ 340, 103, 35 },
		{ 340, 104, 35 },
		{ 340, 105, 35 },
		{ 340, 106, 35 },
		{ 340, 107, 35 },
		{ 340, 108, 35 },
		{ 340, 109, 35 },
		{ 340, 110, 35 },
		{ 340, 111, 35 },
		{ 340, 112, 35 },
		{ 340, 113, 35 },
		{ 340, 114, 35 },
		{ 340, 115, 35 },
		{ 340, 116, 35 },
		{ 340, 117, 35 },
		{ 340, 118, 35 },
		{ 340, 119, 35 },
		{ 340, 120, 35 },
		{ 340, 121, 35 },
		{ 340, 122, 35 },
		{ 340, 123, 35 },
		{ 340, 124, 35 },
		{ 340, 125, 35 },
		{ 340, 126, 35 },
		{ 340, 127, 35 },
		{ 340, 128, 35 },
		{ 340, 129, 35 },
		{ 340, 130, 35 },
		{ 340, 131, 35 },
		{ 340, 132, 35 },
		{ 340, 133, 35 },
		{ 340, 134, 35 },
		{ 340, 135, 35 },
		{ 340, 136, 35 },
		{ 340, 137, 35 },
		{ 340, 138, 35 },
		{ 340, 139, 35 },
		{ 340, 140, 35 },
		{ 340, 141, 35 },
		{ 340, 142, 35 },
		{ 340, 143, 35 },
		{ 340, 144, 35 },
		{ 340, 145, 35 },
		{ 340, 146, 35 },
		{ 341, 90, 35 },
		{ 341, 91, 35 },
		{ 341, 92, 35 },
		{ 341, 93, 35 },
		{ 341, 94, 35 },
		{ 341, 95, 35 },
		{ 341, 96, 35 },
		{ 341, 97, 35 },
		{ 341, 98, 35 },
		{ 341, 99, 35 },
		{ 341, 100, 35 },
		{ 341, 101, 35 },
		{ 341, 102, 35 },
		{ 341, 103, 35 },
		{ 341, 104, 35 },
		{ 341, 105, 35 },
		{ 341, 106, 35 },
		{ 341, 107, 35 },
		{ 341, 108, 35 },
		{ 341, 109, 35 },
		{ 341, 110, 35 },
		{ 341, 111, 35 },
		{ 341, 112, 35 },
		{ 341, 113, 35 },
		{ 341, 114, 35 },
		{ 341, 115, 35 },
		{ 341, 116, 35 },
		{ 341, 117, 35 },
		{ 341, 118, 35 },
		{ 341, 119, 35 },
		{ 341, 120, 35 },
		{ 341, 121, 35 },
		{ 341, 122, 35 },
		{ 341, 123, 35 },
		{ 341, 124, 35 },
		{ 341, 125, 35 },
		{ 341, 126, 35 },
		{ 341, 127, 35 },
		{ 341, 128, 35 },
		{ 341, 129, 35 },
		{ 341, 130, 35 },
		{ 341, 131, 35 },
		{ 341, 132, 35 },
		{ 341, 133, 35 },
		{ 341, 134, 35 },
		{ 341, 135, 35 },
		{ 341, 136, 35 },
		{ 341, 137, 35 },
		{ 341, 138, 35 },
		{ 341, 139, 35 },
		{ 341, 140, 35 },
		{ 341, 141, 35 },
		{ 341, 142, 35 },
		{ 341, 143, 35 },
		{ 341, 144, 35 },
		{ 342, 88, 35 },
		{ 342, 89, 35 },
		{ 342, 90, 35 },
		{ 342, 91, 35 },
		{ 342, 92, 35 },
		{ 342, 93, 35 },
		{ 342, 94, 35 },
		{ 342, 95, 35 },
		{ 342, 96, 35 },
		{ 342, 97, 35 },
		{ 342, 98, 35 },
		{ 342, 99, 35 },
		{ 342, 100, 35 },
		{ 342, 101, 35 },
		{ 342, 102, 35 },
		{ 342, 103, 35 },
		{ 342, 104, 35 },
		{ 342, 105, 35 },
		{ 342, 106, 35 },
		{ 342, 107, 35 },
		{ 342, 108, 35 },
		{ 342, 109, 35 },
		{ 342, 110, 35 },
		{ 342, 111, 35 },
		{ 342, 112, 35 },
		{ 342, 113, 35 },
		{ 342, 114, 35 },
		{ 342, 115, 35 },
		{ 342, 116, 35 },
		{ 342, 117, 35 },
		{ 342, 118, 35 },
		{ 342, 119, 35 },
		{ 342, 120, 35 },
		{ 342, 121, 35 },
		{ 342, 122, 35 },
		{ 342, 123, 35 },
		{ 342, 124, 35 },
		{ 342, 125, 35 },
		{ 342, 126, 35 },
		{ 342, 127, 35 },
		{ 342, 128, 35 },
		{ 342, 129, 35 },
		{ 342, 130, 35 },
		{ 342, 131, 35 },
		{ 342, 132, 35 },
		{ 342, 133, 35 },
		{ 342, 134, 35 },
		{ 342, 135, 35 },
		{ 342, 136, 35 },
		{ 342, 137, 35 },
		{ 342, 138, 35 },
		{ 342, 139, 35 },
		{ 342, 140, 35 },
		{ 342, 141, 35 },
		{ 342, 142, 35 },
		{ 342, 143, 35 },
		{ 343, 86, 35 },
		{ 343, 87, 35 },
		{ 343, 88, 35 },
		{ 343, 89, 35 },
		{ 343, 90, 35 },
		{ 343, 91, 35 },
		{ 343, 92, 35 },
		{ 343, 93, 35 },
		{ 343, 94, 35 },
		{ 343, 95, 35 },
		{ 343, 96, 35 },
		{ 343, 97, 35 },
		{ 343, 98, 35 },
		{ 343, 99, 35 },
		{ 343, 100, 35 },
		{ 343, 101, 35 },
		{ 343, 102, 35 },
		{ 343, 103, 35 },
		{ 343, 104, 35 },
		{ 343, 105, 35 },
		{ 343, 106, 35 },
		{ 343, 107, 35 },
		{ 343, 108, 35 },
		{ 343, 109, 35 },
		{ 343, 110, 35 },
		{ 343, 111, 35 },
		{ 343, 112, 35 },
		{ 343, 113, 35 },
		{ 343, 114, 35 },
		{ 343, 115, 35 },
		{ 343, 116, 35 },
		{ 343, 117, 35 },
		{ 343, 118, 35 },
		{ 343, 119, 35 },
		{ 343, 120, 35 },
		{ 343, 121, 35 },
		{ 343, 122, 35 },
		{ 343, 123, 35 },
		{ 343, 124, 35 },
		{ 343, 125, 35 },
		{ 343, 126, 35 },
		{ 343, 127, 35 },
		{ 343, 128, 35 },
		{ 343, 129, 35 },
		{ 343, 130, 35 },
		{ 343, 131, 35 },
		{ 343, 132, 35 },
		{ 343, 133, 35 },
		{ 343, 134, 35 },
		{ 343, 135, 35 },
		{ 343, 136, 35 },
		{ 343, 137, 35 },
		{ 343, 138, 35 },
		{ 343, 139, 35 },
		{ 343, 140, 35 },
		{ 343, 141, 35 },
		{ 343, 142, 35 },
		{ 344, 84, 35 },
		{ 344, 85, 35 },
		{ 344, 86, 35 },
		{ 344, 87, 35 },
		{ 344, 88, 35 },
		{ 344, 89, 35 },
		{ 344, 90, 35 },
		{ 344, 91, 35 },
		{ 344, 92, 35 },
		{ 344, 93, 35 },
		{ 344, 94, 35 },
		{ 344, 95, 35 },
		{ 344, 96, 35 },
		{ 344, 97, 35 },
		{ 344, 98, 35 },
		{ 344, 99, 35 },
		{ 344, 100, 35 },
		{ 344, 101, 35 },
		{ 344, 102, 35 },
		{ 344, 103, 35 },
		{ 344, 104, 35 },
		{ 344, 105, 35 },
		{ 344, 106, 35 },
		{ 344, 107, 35 },
		{ 344, 108, 35 },
		{ 344, 109, 35 },
		{ 344, 110, 35 },
		{ 344, 111, 35 },
		{ 344, 112, 35 },
		{ 344, 113, 35 },
		{ 344, 114, 35 },
		{ 344, 115, 35 },
		{ 344, 116, 35 },
		{ 344, 117, 35 },
		{ 344, 118, 35 },
		{ 344, 119, 35 },
		{ 344, 120, 35 },
		{ 344, 121, 35 },
		{ 344, 122, 35 },
		{ 344, 123, 35 },
		{ 344, 124, 35 },
		{ 344, 125, 35 },
		{ 344, 126, 35 },
		{ 344, 127, 35 },
		{ 344, 128, 35 },
		{ 344, 129, 35 },
		{ 344, 130, 35 },
		{ 344, 131, 35 },
		{ 344, 132, 35 },
		{ 344, 133, 35 },
		{ 344, 134, 35 },
		{ 344, 135, 35 },
		{ 344, 136, 35 },
		{ 344, 137, 35 },
		{ 344, 138, 35 },
		{ 344, 139, 35 },
		{ 344, 140, 35 },
		{ 345, 82, 35 },
		{ 345, 83, 35 },
		{ 345, 84, 35 },
		{ 345, 85, 35 },
		{ 345, 86, 35 },
		{ 345, 87, 35 },
		{ 345, 88, 35 },
		{ 345, 89, 35 },
		{ 345, 90, 35 },
		{ 345, 91, 35 },
		{ 345, 92, 35 },
		{ 345, 93, 35 },
		{ 345, 94, 35 },
		{ 345, 95, 35 },
		{ 345, 96, 35 },
		{ 345, 97, 35 },
		{ 345, 98, 35 },
		{ 345, 99, 35 },
		{ 345, 100, 35 },
		{ 345, 101, 35 },
		{ 345, 102, 35 },
		{ 345, 103, 35 },
		{ 345, 104, 35 },
		{ 345, 105, 35 },
		{ 345, 106, 35 },
		{ 345, 107, 35 },
		{ 345, 108, 35 },
		{ 345, 109, 35 },
		{ 345, 110, 35 },
		{ 345, 111, 35 },
		{ 345, 112, 35 },
		{ 345, 113, 35 },
		{ 345, 114, 35 },
		{ 345, 115, 35 },
		{ 345, 116, 35 },
		{ 345, 117, 35 },
		{ 345, 118, 35 },
		{ 345, 119, 35 },
		{ 345, 120, 35 },
		{ 345, 121, 35 },
		{ 345, 122, 35 },
		{ 345, 123, 35 },
		{ 345, 124, 35 },
		{ 345, 125, 35 },
		{ 345, 126, 35 },
		{ 345, 127, 35 },
		{ 345, 128, 35 },
		{ 345, 129, 35 },
		{ 345, 130, 35 },
		{ 345, 131, 35 },
		{ 345, 132, 35 },
		{ 345, 133, 35 },
		{ 345, 134, 35 },
		{ 345, 135, 35 },
		{ 345, 136, 35 },
		{ 345, 137, 35 },
		{ 345, 138, 35 },
		{ 345, 139, 35 },
		{ 346, 80, 35 },
		{ 346, 81, 35 },
		{ 346, 82, 35 },
		{ 346, 83, 35 },
		{ 346, 84, 35 },
		{ 346, 85, 35 },
		{ 346, 86, 35 },
		{ 346, 87, 35 },
		{ 346, 88, 35 },
		{ 346, 89, 35 },
		{ 346, 90, 35 },
		{ 346, 91, 35 },
		{ 346, 92, 35 },
		{ 346, 93, 35 },
		{ 346, 94, 35 },
		{ 346, 95, 35 },
		{ 346, 96, 35 },
		{ 346, 97, 35 },
		{ 346, 98, 35 },
		{ 346, 99, 35 },
		{ 346, 100, 35 },
		{ 346, 101, 35 },
		{ 346, 102, 35 },
		{ 346, 103, 35 },
		{ 346, 104, 35 },
		{ 346, 105, 35 },
		{ 346, 106, 35 },
		{ 346, 107, 35 },
		{ 346, 108, 35 },
		{ 346, 109, 35 },
		{ 346, 110, 35 },
		{ 346, 111, 35 },
		{ 346, 112, 35 },
		{ 346, 113, 35 },
		{ 346, 114, 35 },
		{ 346, 115, 35 },
		{ 346, 116, 35 },
		{ 346, 117, 35 },
		{ 346, 118, 35 },
		{ 346, 119, 35 },
		{ 346, 120, 35 },
		{ 346, 121, 35 },
		{ 346, 122, 35 },
		{ 346, 123, 35 },
		{ 346, 124, 35 },
		{ 346, 125, 35 },
		{ 346, 126, 35 },
		{ 346, 127, 35 },
		{ 346, 128, 35 },
		{ 346, 129, 35 },
		{ 346, 130, 35 },
		{ 346, 131, 35 },
		{ 346, 132, 35 },
		{ 346, 133, 35 },
		{ 346, 134, 35 },
		{ 346, 135, 35 },
		{ 346, 136, 35 },
		{ 346, 137, 35 },
		{ 346, 138, 35 },
		{ 347, 78, 35 },
		{ 347, 79, 35 },
		{ 347, 80, 35 },
		{ 347, 81, 35 },
		{ 347, 82, 35 },
		{ 347, 83, 35 },
		{ 347, 84, 35 },
		{ 347, 85, 35 },
		{ 347, 86, 35 },
		{ 347, 87, 35 },
		{ 347, 88, 35 },
		{ 347, 89, 35 },
		{ 347, 90, 35 },
		{ 347, 91, 35 },
		{ 347, 92, 35 },
		{ 347, 93, 35 },
		{ 347, 94, 35 },
		{ 347, 95, 35 },
		{ 347, 96, 35 },
		{ 347, 97, 35 },
		{ 347, 98, 35 },
		{ 347, 99, 35 },
		{ 347, 100, 35 },
		{ 347, 101, 35 },
		{ 347, 102, 35 },
		{ 347, 103, 35 },
		{ 347, 104, 35 },
		{ 347, 105, 35 },
		{ 347, 106, 35 },
		{ 347, 107, 35 },
		{ 347, 108, 35 },
		{ 347, 109, 35 },
		{ 347, 110, 35 },
		{ 347, 111, 35 },
		{ 347, 112, 35 },
		{ 347, 113, 35 },
		{ 347, 114, 35 },
		{ 347, 115, 35 },
		{ 347, 116, 35 },
		{ 347, 117, 35 },
		{ 347, 118, 35 },
		{ 347, 119, 35 },
		{ 347, 120, 35 },
		{ 347, 121, 35 },
		{ 347, 122, 35 },
		{ 347, 123, 35 },
		{ 347, 124, 35 },
		{ 347, 125, 35 },
		{ 347, 126, 35 },
		{ 347, 127, 35 },
		{ 347, 128, 35 },
		{ 347, 129, 35 },
		{ 347, 130, 35 },
		{ 347, 131, 35 },
		{ 347, 132, 35 },
		{ 347, 133, 35 },
		{ 347, 134, 35 },
		{ 347, 135, 35 },
		{ 347, 136, 35 },
		{ 348, 76, 35 },
		{ 348, 77, 35 },
		{ 348, 78, 35 },
		{ 348, 79, 35 },
		{ 348, 80, 35 },
		{ 348, 81, 35 },
		{ 348, 82, 35 },
		{ 348, 83, 35 },
		{ 348, 84, 35 },
		{ 348, 85, 35 },
		{ 348, 86, 35 },
		{ 348, 87, 35 },
		{ 348, 88, 35 },
		{ 348, 89, 35 },
		{ 348, 90, 35 },
		{ 348, 91, 35 },
		{ 348, 92, 35 },
		{ 348, 93, 35 },
		{ 348, 94, 35 },
		{ 348, 95, 35 },
		{ 348, 96, 35 },
		{ 348, 97, 35 },
		{ 348, 98, 35 },
		{ 348, 99, 35 },
		{ 348, 100, 35 },
		{ 348, 101, 35 },
		{ 348, 102, 35 },
		{ 348, 103, 35 },
		{ 348, 104, 35 },
		{ 348, 105, 35 },
		{ 348, 106, 35 },
		{ 348, 107, 35 },
		{ 348, 108, 35 },
		{ 348, 109, 35 },
		{ 348, 110, 35 },
		{ 348, 111, 35 },
		{ 348, 112, 35 },
		{ 348, 113, 35 },
		{ 348, 114, 35 },
		{ 348, 115, 35 },
		{ 348, 116, 35 },
		{ 348, 117, 35 },
		{ 348, 118, 35 },
		{ 348, 119, 35 },
		{ 348, 120, 35 },
		{ 348, 121, 35 },
		{ 348, 122, 35 },
		{ 348, 123, 35 },
		{ 348, 124, 35 },
		{ 348, 125, 35 },
		{ 348, 126, 35 },
		{ 348, 127, 35 },
		{ 348, 128, 35 },
		{ 348, 129, 35 },
		{ 348, 130, 35 },
		{ 348, 131, 35 },
		{ 348, 132, 35 },
		{ 348, 133, 35 },
		{ 348, 134, 35 },
		{ 348, 135, 35 },
		{ 349, 74, 35 },
		{ 349, 75, 35 },
		{ 349, 76, 35 },
		{ 349, 77, 35 },
		{ 349, 78, 35 },
		{ 349, 79, 35 },
		{ 349, 80, 35 },
		{ 349, 81, 35 },
		{ 349, 82, 35 },
		{ 349, 83, 35 },
		{ 349, 84, 35 },
		{ 349, 85, 35 },
		{ 349, 86, 35 },
		{ 349, 87, 35 },
		{ 349, 88, 35 },
		{ 349, 89, 35 },
		{ 349, 90, 35 },
		{ 349, 91, 35 },
		{ 349, 92, 35 },
		{ 349, 93, 35 },
		{ 349, 94, 35 },
		{ 349, 95, 35 },
		{ 349, 96, 35 },
		{ 349, 97, 35 },
		{ 349, 98, 35 },
		{ 349, 99, 35 },
		{ 349, 100, 35 },
		{ 349, 101, 35 },
		{ 349, 102, 35 },
		{ 349, 103, 35 },
		{ 349, 104, 35 },
		{ 349, 105, 35 },
		{ 349, 106, 35 },
		{ 349, 107, 35 },
		{ 349, 108, 35 },
		{ 349, 109, 35 },
		{ 349, 110, 35 },
		{ 349, 111, 35 },
		{ 349, 112, 35 },
		{ 349, 113, 35 },
		{ 349, 114, 35 },
		{ 349, 115, 35 },
		{ 349, 116, 35 },
		{ 349, 117, 35 },
		{ 349, 118, 35 },
		{ 349, 119, 35 },
		{ 349, 120, 35 },
		{ 349, 121, 35 },
		{ 349, 122, 35 },
		{ 349, 123, 35 },
		{ 349, 124, 35 },
		{ 349, 125, 35 },
		{ 349, 126, 35 },
		{ 349, 127, 35 },
		{ 349, 128, 35 },
		{ 349, 129, 35 },
		{ 349, 130, 35 },
		{ 349, 131, 35 },
		{ 349, 132, 35 },
		{ 349, 133, 35 },
		{ 349, 134, 35 },
		{ 350, 72, 35 },
		{ 350, 73, 35 },
		{ 350, 74, 35 },
		{ 350, 75, 35 },
		{ 350, 76, 35 },
		{ 350, 77, 35 },
		{ 350, 78, 35 },
		{ 350, 79, 35 },
		{ 350, 80, 35 },
		{ 350, 81, 35 },
		{ 350, 82, 35 },
		{ 350, 83, 35 },
		{ 350, 84, 35 },
		{ 350, 85, 35 },
		{ 350, 86, 35 },
		{ 350, 87, 35 },
		{ 350, 88, 35 },
		{ 350, 89, 35 },
		{ 350, 90, 35 },
		{ 350, 91, 35 },
		{ 350, 92, 35 },
		{ 350, 93, 35 },
		{ 350, 94, 35 },
		{ 350, 95, 35 },
		{ 350, 96, 35 },
		{ 350, 97, 35 },
		{ 350, 98, 35 },
		{ 350, 99, 35 },
		{ 350, 100, 35 },
		{ 350, 101, 35 },
		{ 350, 102, 35 },
		{ 350, 103, 35 },
		{ 350, 104, 35 },
		{ 350, 105, 35 },
		{ 350, 106, 35 },
		{ 350, 107, 35 },
		{ 350, 108, 35 },
		{ 350, 109, 35 },
		{ 350, 110, 35 },
		{ 350, 111, 35 },
		{ 350, 112, 35 },
		{ 350, 113, 35 },
		{ 350, 114, 35 },
		{ 350, 115, 35 },
		{ 350, 116, 35 },
		{ 350, 117, 35 },
		{ 350, 118, 35 },
		{ 350, 119, 35 },
		{ 350, 120, 35 },
		{ 350, 121, 35 },
		{ 350, 122, 35 },
		{ 350, 123, 35 },
		{ 350, 124, 35 },
		{ 350, 125, 35 },
		{ 350, 126, 35 },
		{ 350, 127, 35 },
		{ 350, 128, 35 },
		{ 350, 129, 35 },
		{ 350, 130, 35 },
		{ 350, 131, 35 },
		{ 350, 132, 35 },
		{ 351, 70, 35 },
		{ 351, 71, 35 },
		{ 351, 72, 35 },
		{ 351, 73, 35 },
		{ 351, 74, 35 },
		{ 351, 75, 35 },
		{ 351, 76, 35 },
		{ 351, 77, 35 },
		{ 351, 78, 35 },
		{ 351, 79, 35 },
		{ 351, 80, 35 },
		{ 351, 81, 35 },
		{ 351, 82, 35 },
		{ 351, 83, 35 },
		{ 351, 84, 35 },
		{ 351, 85, 35 },
		{ 351, 86, 35 },
		{ 351, 87, 35 },
		{ 351, 88, 35 },
		{ 351, 89, 35 },
		{ 351, 90, 35 },
		{ 351, 91, 35 },
		{ 351, 92, 35 },
		{ 351, 93, 35 },
		{ 351, 94, 35 },
		{ 351, 95, 35 },
		{ 351, 96, 35 },
		{ 351, 97, 35 },
		{ 351, 98, 35 },
		{ 351, 99, 35 },
		{ 351, 100, 35 },
		{ 351, 101, 35 },
		{ 351, 102, 35 },
		{ 351, 103, 35 },
		{ 351, 104, 35 },
		{ 351, 105, 35 },
		{ 351, 106, 35 },
		{ 351, 107, 35 },
		{ 351, 108, 35 },
		{ 351, 109, 35 },
		{ 351, 110, 35 },
		{ 351, 111, 35 },
		{ 351, 112, 35 },
		{ 351, 113, 35 },
		{ 351, 114, 35 },
		{ 351, 115, 35 },
		{ 351, 116, 35 },
		{ 351, 117, 35 },
		{ 351, 118, 35 },
		{ 351, 119, 35 },
		{ 351, 120, 35 },
		{ 351, 121, 35 },
		{ 351, 122, 35 },
		{ 351, 123, 35 },
		{ 351, 124, 35 },
		{ 351, 125, 35 },
		{ 351, 126, 35 },
		{ 351, 127, 35 },
		{ 351, 128, 35 },
		{ 351, 129, 35 },
		{ 351, 130, 35 },
		{ 351, 131, 35 },
		{ 352, 68, 35 },
		{ 352, 69, 35 },
		{ 352, 70, 35 },
		{ 352, 71, 35 },
		{ 352, 72, 35 },
		{ 352, 73, 35 },
		{ 352, 74, 35 },
		{ 352, 75, 35 },
		{ 352, 76, 35 },
		{ 352, 77, 35 },
		{ 352, 78, 35 },
		{ 352, 79, 35 },
		{ 352, 80, 35 },
		{ 352, 81, 35 },
		{ 352, 82, 35 },
		{ 352, 83, 35 },
		{ 352, 84, 35 },
		{ 352, 85, 35 },
		{ 352, 86, 35 },
		{ 352, 87, 35 },
		{ 352, 88, 35 },
		{ 352, 89, 35 },
		{ 352, 90, 35 },
		{ 352, 91, 35 },
		{ 352, 92, 35 },
		{ 352, 93, 35 },
		{ 352, 94, 35 },
		{ 352, 95, 35 },
		{ 352, 96, 35 },
		{ 352, 97, 35 },
		{ 352, 98, 35 },
		{ 352, 99, 35 },
		{ 352, 100, 35 },
		{ 352, 101, 35 },
		{ 352, 102, 35 },
		{ 352, 103, 35 },
		{ 352, 104, 35 },
		{ 352, 105, 35 },
		{ 352, 106, 35 },
		{ 352, 107, 35 },
		{ 352, 108, 35 },
		{ 352, 109, 35 },
		{ 352, 110, 35 },
		{ 352, 111, 35 },
		{ 352, 112, 35 },
		{ 352, 113, 35 },
		{ 352, 114, 35 },
		{ 352, 115, 35 },
		{ 352, 116, 35 },
		{ 352, 117, 35 },
		{ 352, 118, 35 },
		{ 352, 119, 35 },
		{ 352, 120, 35 },
		{ 352, 121, 35 },
		{ 352, 122, 35 },
		{ 352, 123, 35 },
		{ 352, 124, 35 },
		{ 352, 125, 35 },
		{ 352, 126, 35 },
		{ 352, 127, 35 },
		{ 352, 128, 35 },
		{ 352, 129, 35 },
		{ 352, 130, 35 },
		{ 353, 66, 35 },
		{ 353, 67, 35 },
		{ 353, 68, 35 },
		{ 353, 69, 35 },
		{ 353, 70, 35 },
		{ 353, 71, 35 },
		{ 353, 72, 35 },
		{ 353, 73, 35 },
		{ 353, 74, 35 },
		{ 353, 75, 35 },
		{ 353, 76, 35 },
		{ 353, 77, 35 },
		{ 353, 78, 35 },
		{ 353, 79, 35 },
		{ 353, 80, 35 },
		{ 353, 81, 35 },
		{ 353, 82, 35 },
		{ 353, 83, 35 },
		{ 353, 84, 35 },
		{ 353, 85, 35 },
		{ 353, 86, 35 },
		{ 353, 87, 35 },
		{ 353, 88, 35 },
		{ 353, 89, 35 },
		{ 353, 90, 35 },
		{ 353, 91, 35 },
		{ 353, 92, 35 },
		{ 353, 93, 35 },
		{ 353, 94, 35 },
		{ 353, 95, 35 },
		{ 353, 96, 35 },
		{ 353, 97, 35 },
		{ 353, 98, 35 },
		{ 353, 99, 35 },
		{ 353, 100, 35 },
		{ 353, 101, 35 },
		{ 353, 102, 35 },
		{ 353, 103, 35 },
		{ 353, 104, 35 },
		{ 353, 105, 35 },
		{ 353, 106, 35 },
		{ 353, 107, 35 },
		{ 353, 108, 35 },
		{ 353, 109, 35 },
		{ 353, 110, 35 },
		{ 353, 111, 35 },
		{ 353, 112, 35 },
		{ 353, 113, 35 },
		{ 353, 114, 35 },
		{ 353, 115, 35 },
		{ 353, 116, 35 },
		{ 353, 117, 35 },
		{ 353, 118, 35 },
		{ 353, 119, 35 },
		{ 353, 120, 35 },
		{ 353, 121, 35 },
		{ 353, 122, 35 },
		{ 353, 123, 35 },
		{ 353, 124, 35 },
		{ 353, 125, 35 },
		{ 353, 126, 35 },
		{ 353, 127, 35 },
		{ 353, 128, 35 },
		{ 354, 64, 35 },
		{ 354, 65, 35 },
		{ 354, 66, 35 },
		{ 354, 67, 35 },
		{ 354, 68, 35 },
		{ 354, 69, 35 },
		{ 354, 70, 35 },
		{ 354, 71, 35 },
		{ 354, 72, 35 },
		{ 354, 73, 35 },
		{ 354, 74, 35 },
		{ 354, 75, 35 },
		{ 354, 76, 35 },
		{ 354, 77, 35 },
		{ 354, 78, 35 },
		{ 354, 79, 35 },
		{ 354, 80, 35 },
		{ 354, 81, 35 },
		{ 354, 82, 35 },
		{ 354, 83, 35 },
		{ 354, 84, 35 },
		{ 354, 85, 35 },
		{ 354, 86, 35 },
		{ 354, 87, 35 },
		{ 354, 88, 35 },
		{ 354, 89, 35 },
		{ 354, 90, 35 },
		{ 354, 91, 35 },
		{ 354, 92, 35 },
		{ 354, 93, 35 },
		{ 354, 94, 35 },
		{ 354, 95, 35 },
		{ 354, 96, 35 },
		{ 354, 97, 35 },
		{ 354, 98, 35 },
		{ 354, 99, 35 },
		{ 354, 100, 35 },
		{ 354, 101, 35 },
		{ 354, 102, 35 },
		{ 354, 103, 35 },
		{ 354, 104, 35 },
		{ 354, 105, 35 },
		{ 354, 106, 35 },
		{ 354, 107, 35 },
		{ 354, 108, 35 },
		{ 354, 109, 35 },
		{ 354, 110, 35 },
		{ 354, 111, 35 },
		{ 354, 112, 35 },
		{ 354, 113, 35 },
		{ 354, 114, 35 },
		{ 354, 115, 35 },
		{ 354, 116, 35 },
		{ 354, 117, 35 },
		{ 354, 118, 35 },
		{ 354, 119, 35 },
		{ 354, 120, 35 },
		{ 354, 121, 35 },
		{ 354, 122, 35 },
		{ 354, 123, 35 },
		{ 354, 124, 35 },
		{ 354, 125, 35 },
		{ 354, 126, 35 },
		{ 354, 127, 35 },
		{ 355, 62, 35 },
		{ 355, 63, 35 },
		{ 355, 64, 35 },
		{ 355, 65, 35 },
		{ 355, 66, 35 },
		{ 355, 67, 35 },
		{ 355, 68, 35 },
		{ 355, 69, 35 },
		{ 355, 70, 35 },
		{ 355, 71, 35 },
		{ 355, 72, 35 },
		{ 355, 73, 35 },
		{ 355, 74, 35 },
		{ 355, 75, 35 },
		{ 355, 76, 35 },
		{ 355, 77, 35 },
		{ 355, 78, 35 },
		{ 355, 79, 35 },
		{ 355, 80, 35 },
		{ 355, 81, 35 },
		{ 355, 82, 35 },
		{ 355, 83, 35 },
		{ 355, 84, 35 },
		{ 355, 85, 35 },
		{ 355, 86, 35 },
		{ 355, 87, 35 },
		{ 355, 88, 35 },
		{ 355, 89, 35 },
		{ 355, 90, 35 },
		{ 355, 91, 35 },
		{ 355, 92, 35 },
		{ 355, 93, 35 },
		{ 355, 94, 35 },
		{ 355, 95, 35 },
		{ 355, 96, 35 },
		{ 355, 97, 35 },
		{ 355, 98, 35 },
		{ 355, 99, 35 },
		{ 355, 100, 35 },
		{ 355, 101, 35 },
		{ 355, 102, 35 },
		{ 355, 103, 35 },
		{ 355, 104, 35 },
		{ 355, 105, 35 },
		{ 355, 106, 35 },
		{ 355, 107, 35 },
		{ 355, 108, 35 },
		{ 355, 109, 35 },
		{ 355, 110, 35 },
		{ 355, 111, 35 },
		{ 355, 112, 35 },
		{ 355, 113, 35 },
		{ 355, 114, 35 },
		{ 355, 115, 35 },
		{ 355, 116, 35 },
		{ 355, 117, 35 },
		{ 355, 118, 35 },
		{ 355, 119, 35 },
		{ 355, 120, 35 },
		{ 355, 121, 35 },
		{ 355, 122, 35 },
		{ 355, 123, 35 },
		{ 355, 124, 35 },
		{ 355, 125, 35 },
		{ 355, 126, 35 },
		{ 356, 60, 35 },
		{ 356, 61, 35 },
		{ 356, 62, 35 },
		{ 356, 63, 35 },
		{ 356, 64, 35 },
		{ 356, 65, 35 },
		{ 356, 66, 35 },
		{ 356, 67, 35 },
		{ 356, 68, 35 },
		{ 356, 69, 35 },
		{ 356, 70, 35 },
		{ 356, 71, 35 },
		{ 356, 72, 35 },
		{ 356, 73, 35 },
		{ 356, 74, 35 },
		{ 356, 75, 35 },
		{ 356, 76, 35 },
		{ 356, 77, 35 },
		{ 356, 78, 35 },
		{ 356, 79, 35 },
		{ 356, 80, 35 },
		{ 356, 81, 35 },
		{ 356, 82, 35 },
		{ 356, 83, 35 },
		{ 356, 84, 35 },
		{ 356, 85, 35 },
		{ 356, 86, 35 },
		{ 356, 87, 35 },
		{ 356, 88, 35 },
		{ 356, 89, 35 },
		{ 356, 90, 35 },
		{ 356, 91, 35 },
		{ 356, 92, 35 },
		{ 356, 93, 35 },
		{ 356, 94, 35 },
		{ 356, 95, 35 },
		{ 356, 96, 35 },
		{ 356, 97, 35 },
		{ 356, 98, 35 },
		{ 356, 99, 35 },
		{ 356, 100, 35 },
		{ 356, 101, 35 },
		{ 356, 102, 35 },
		{ 356, 103, 35 },
		{ 356, 104, 35 },
		{ 356, 105, 35 },
		{ 356, 106, 35 },
		{ 356, 107, 35 },
		{ 356, 108, 35 },
		{ 356, 109, 35 },
		{ 356, 110, 35 },
		{ 356, 111, 35 },
		{ 356, 112, 35 },
		{ 356, 113, 35 },
		{ 356, 114, 35 },
		{ 356, 115, 35 },
		{ 356, 116, 35 },
		{ 356, 117, 35 },
		{ 356, 118, 35 },
		{ 356, 119, 35 },
		{ 356, 120, 35 },
		{ 356, 121, 35 },
		{ 356, 122, 35 },
		{ 356, 123, 35 },
		{ 356, 124, 35 },
		{ 357, 58, 35 },
		{ 357, 59, 35 },
		{ 357, 60, 35 },
		{ 357, 61, 35 },
		{ 357, 62, 35 },
		{ 357, 63, 35 },
		{ 357, 64, 35 },
		{ 357, 65, 35 },
		{ 357, 66, 35 },
		{ 357, 67, 35 },
		{ 357, 68, 35 },
		{ 357, 69, 35 },
		{ 357, 70, 35 },
		{ 357, 71, 35 },
		{ 357, 72, 35 },
		{ 357, 73, 35 },
		{ 357, 74, 35 },
		{ 357, 75, 35 },
		{ 357, 76, 35 },
		{ 357, 77, 35 },
		{ 357, 78, 35 },
		{ 357, 79, 35 },
		{ 357, 80, 35 },
		{ 357, 81, 35 },
		{ 357, 82, 35 },
		{ 357, 83, 35 },
		{ 357, 84, 35 },
		{ 357, 85, 35 },
		{ 357, 86, 35 },
		{ 357, 87, 35 },
		{ 357, 88, 35 },
		{ 357, 89, 35 },
		{ 357, 90, 35 },
		{ 357, 91, 35 },
		{ 357, 92, 35 },
		{ 357, 93, 35 },
		{ 357, 94, 35 },
		{ 357, 95, 35 },
		{ 357, 96, 35 },
		{ 357, 97, 35 },
		{ 357, 98, 35 },
		{ 357, 99, 35 },
		{ 357, 100, 35 },
		{ 357, 101, 35 },
		{ 357, 102, 35 },
		{ 357, 103, 35 },
		{ 357, 104, 35 },
		{ 357, 105, 35 },
		{ 357, 106, 35 },
		{ 357, 107, 35 },
		{ 357, 108, 35 },
		{ 357, 109, 35 },
		{ 357, 110, 35 },
		{ 357, 111, 35 },
		{ 357, 112, 35 },
		{ 357, 113, 35 },
		{ 357, 114, 35 },
		{ 357, 115, 35 },
		{ 357, 116, 35 },
		{ 357, 117, 35 },
		{ 357, 118, 35 },
		{ 357, 119, 35 },
		{ 357, 120, 35 },
		{ 357, 121, 35 },
		{ 357, 122, 35 },
		{ 357, 123, 35 },
		{ 358, 56, 35 },
		{ 358, 57, 35 },
		{ 358, 58, 35 },
		{ 358, 59, 35 },
		{ 358, 60, 35 },
		{ 358, 61, 35 },
		{ 358, 62, 35 },
		{ 358, 63, 35 },
		{ 358, 64, 35 },
		{ 358, 65, 35 },
		{ 358, 66, 35 },
		{ 358, 67, 35 },
		{ 358, 68, 35 },
		{ 358, 69, 35 },
		{ 358, 70, 35 },
		{ 358, 71, 35 },
		{ 358, 72, 35 },
		{ 358, 73, 35 },
		{ 358, 74, 35 },
		{ 358, 75, 35 },
		{ 358, 76, 35 },
		{ 358, 77, 35 },
		{ 358, 78, 35 },
		{ 358, 79, 35 },
		{ 358, 80, 35 },
		{ 358, 81, 35 },
		{ 358, 82, 35 },
		{ 358, 83, 35 },
		{ 358, 84, 35 },
		{ 358, 85, 35 },
		{ 358, 86, 35 },
		{ 358, 87, 35 },
		{ 358, 88, 35 },
		{ 358, 89, 35 },
		{ 358, 90, 35 },
		{ 358, 91, 35 },
		{ 358, 92, 35 },
		{ 358, 93, 35 },
		{ 358, 94, 35 },
		{ 358, 95, 35 },
		{ 358, 96, 35 },
		{ 358, 97, 35 },
		{ 358, 98, 35 },
		{ 358, 99, 35 },
		{ 358, 100, 35 },
		{ 358, 101, 35 },
		{ 358, 102, 35 },
		{ 358, 103, 35 },
		{ 358, 104, 35 },
		{ 358, 105, 35 },
		{ 358, 106, 35 },
		{ 358, 107, 35 },
		{ 358, 108, 35 },
		{ 358, 109, 35 },
		{ 358, 110, 35 },
		{ 358, 111, 35 },
		{ 358, 112, 35 },
		{ 358, 113, 35 },
		{ 358, 114, 35 },
		{ 358, 115, 35 },
		{ 358, 116, 35 },
		{ 358, 117, 35 },
		{ 358, 118, 35 },
		{ 358, 119, 35 },
		{ 358, 120, 35 },
		{ 358, 121, 35 },
		{ 358, 122, 35 },
		{ 359, 54, 35 },
		{ 359, 55, 35 },
		{ 359, 56, 35 },
		{ 359, 57, 35 },
		{ 359, 58, 35 },
		{ 359, 59, 35 },
		{ 359, 60, 35 },
		{ 359, 61, 35 },
		{ 359, 62, 35 },
		{ 359, 63, 35 },
		{ 359, 64, 35 },
		{ 359, 65, 35 },
		{ 359, 66, 35 },
		{ 359, 67, 35 },
		{ 359, 68, 35 },
		{ 359, 69, 35 },
		{ 359, 70, 35 },
		{ 359, 71, 35 },
		{ 359, 72, 35 },
		{ 359, 73, 35 },
		{ 359, 74, 35 },
		{ 359, 75, 35 },
		{ 359, 76, 35 },
		{ 359, 77, 35 },
		{ 359, 78, 35 },
		{ 359, 79, 35 },
		{ 359, 80, 35 },
		{ 359, 81, 35 },
		{ 359, 82, 35 },
		{ 359, 83, 35 },
		{ 359, 84, 35 },
		{ 359, 85, 35 },
		{ 359, 86, 35 },
		{ 359, 87, 35 },
		{ 359, 88, 35 },
		{ 359, 89, 35 },
		{ 359, 90, 35 },
		{ 359, 91, 35 },
		{ 359, 92, 35 },
		{ 359, 93, 35 },
		{ 359, 94, 35 },
		{ 359, 95, 35 },
		{ 359, 96, 35 },
		{ 359, 97, 35 },
		{ 359, 98, 35 },
		{ 359, 99, 35 },
		{ 359, 100, 35 },
		{ 359, 101, 35 },
		{ 359, 102, 35 },
		{ 359, 103, 35 },
		{ 359, 104, 35 },
		{ 359, 105, 35 },
		{ 359, 106, 35 },
		{ 359, 107, 35 },
		{ 359, 108, 35 },
		{ 359, 109, 35 },
		{ 359, 110, 35 },
		{ 359, 111, 35 },
		{ 359, 112, 35 },
		{ 359, 113, 35 },
		{ 359, 114, 35 },
		{ 359, 115, 35 },
		{ 359, 116, 35 },
		{ 359, 117, 35 },
		{ 359, 118, 35 },
		{ 359, 119, 35 },
		{ 359, 120, 35 },
		{ 360, 52, 35 },
		{ 360, 53, 35 },
		{ 360, 54, 35 },
		{ 360, 55, 35 },
		{ 360, 56, 35 },
		{ 360, 57, 35 },
		{ 360, 58, 35 },
		{ 360, 59, 35 },
		{ 360, 60, 35 },
		{ 360, 61, 35 },
		{ 360, 62, 35 },
		{ 360, 63, 35 },
		{ 360, 64, 35 },
		{ 360, 65, 35 },
		{ 360, 66, 35 },
		{ 360, 67, 35 },
		{ 360, 68, 35 },
		{ 360, 69, 35 },
		{ 360, 70, 35 },
		{ 360, 71, 35 },
		{ 360, 72, 35 },
		{ 360, 73, 35 },
		{ 360, 74, 35 },
		{ 360, 75, 35 },
		{ 360, 76, 35 },
		{ 360, 77, 35 },
		{ 360, 78, 35 },
		{ 360, 79, 35 },
		{ 360, 80, 35 },
		{ 360, 81, 35 },
		{ 360, 82, 35 },
		{ 360, 83, 35 },
		{ 360, 84, 35 },
		{ 360, 85, 35 },
		{ 360, 86, 35 },
		{ 360, 87, 35 },
		{ 360, 88, 35 },
		{ 360, 89, 35 },
		{ 360, 90, 35 },
		{ 360, 91, 35 },
		{ 360, 92, 35 },
		{ 360, 93, 35 },
		{ 360, 94, 35 },
		{ 360, 95, 35 },
		{ 360, 96, 35 },
		{ 360, 97, 35 },
		{ 360, 98, 35 },
		{ 360, 99, 35 },
		{ 360, 100, 35 },
		{ 360, 101, 35 },
		{ 360, 102, 35 },
		{ 360, 103, 35 },
		{ 360, 104, 35 },
		{ 360, 105, 35 },
		{ 360, 106, 35 },
		{ 360, 107, 35 },
		{ 360, 108, 35 },
		{ 360, 109, 35 },
		{ 360, 110, 35 },
		{ 360, 111, 35 },
		{ 360, 112, 35 },
		{ 360, 113, 35 },
		{ 360, 114, 35 },
		{ 360, 115, 35 },
		{ 360, 116, 35 },
		{ 360, 117, 35 },
		{ 360, 118, 35 },
		{ 360, 119, 35 },
		{ 361, 50, 35 },
		{ 361, 51, 35 },
		{ 361, 52, 35 },
		{ 361, 53, 35 },
		{ 361, 54, 35 },
		{ 361, 55, 35 },
		{ 361, 56, 35 },
		{ 361, 57, 35 },
		{ 361, 58, 35 },
		{ 361, 59, 35 },
		{ 361, 60, 35 },
		{ 361, 61, 35 },
		{ 361, 62, 35 },
		{ 361, 63, 35 },
		{ 361, 64, 35 },
		{ 361, 65, 35 },
		{ 361, 66, 35 },
		{ 361, 67, 35 },
		{ 361, 68, 35 },
		{ 361, 69, 35 },
		{ 361, 70, 35 },
		{ 361, 71, 35 },
		{ 361, 72, 35 },
		{ 361, 73, 35 },
		{ 361, 74, 35 },
		{ 361, 75, 35 },
		{ 361, 76, 35 },
		{ 361, 77, 35 },
		{ 361, 78, 35 },
		{ 361, 79, 35 },
		{ 361, 80, 35 },
		{ 361, 81, 35 },
		{ 361, 82, 35 },
		{ 361, 83, 35 },
		{ 361, 84, 35 },
		{ 361, 85, 35 },
		{ 361, 86, 35 },
		{ 361, 87, 35 },
		{ 361, 88, 35 },
		{ 361, 89, 35 },
		{ 361, 90, 35 },
		{ 361, 91, 35 },
		{ 361, 92, 35 },
		{ 361, 93, 35 },
		{ 361, 94, 35 },
		{ 361, 95, 35 },
		{ 361, 96, 35 },
		{ 361, 97, 35 },
		{ 361, 98, 35 },
		{ 361, 99, 35 },
		{ 361, 100, 35 },
		{ 361, 101, 35 },
		{ 361, 102, 35 },
		{ 361, 103, 35 },
		{ 361, 104, 35 },
		{ 361, 105, 35 },
		{ 361, 106, 35 },
		{ 361, 107, 35 },
		{ 361, 108, 35 },
		{ 361, 109, 35 },
		{ 361, 110, 35 },
		{ 361, 111, 35 },
		{ 361, 112, 35 },
		{ 361, 113, 35 },
		{ 361, 114, 35 },
		{ 361, 115, 35 },
		{ 361, 116, 35 },
		{ 361, 117, 35 },
		{ 361, 118, 35 },
		{ 362, 48, 35 },
		{ 362, 49, 35 },
		{ 362, 50, 35 },
		{ 362, 51, 35 },
		{ 362, 52, 35 },
		{ 362, 53, 35 },
		{ 362, 54, 35 },
		{ 362, 55, 35 },
		{ 362, 56, 35 },
		{ 362, 57, 35 },
		{ 362, 58, 35 },
		{ 362, 59, 35 },
		{ 362, 60, 35 },
		{ 362, 61, 35 },
		{ 362, 62, 35 },
		{ 362, 63, 35 },
		{ 362, 64, 35 },
		{ 362, 65, 35 },
		{ 362, 66, 35 },
		{ 362, 67, 35 },
		{ 362, 68, 35 },
		{ 362, 69, 35 },
		{ 362, 70, 35 },
		{ 362, 71, 35 },
		{ 362, 72, 35 },
		{ 362, 73, 35 },
		{ 362, 74, 35 },
		{ 362, 75, 35 },
		{ 362, 76, 35 },
		{ 362, 77, 35 },
		{ 362, 78, 35 },
		{ 362, 79, 35 },
		{ 362, 80, 35 },
		{ 362, 81, 35 },
		{ 362, 82, 35 },
		{ 362, 83, 35 },
		{ 362, 84, 35 },
		{ 362, 85, 35 },
		{ 362, 86, 35 },
		{ 362, 87, 35 },
		{ 362, 88, 35 },
		{ 362, 89, 35 },
		{ 362, 90, 35 },
		{ 362, 91, 35 },
		{ 362, 92, 35 },
		{ 362, 93, 35 },
		{ 362, 94, 35 },
		{ 362, 95, 35 },
		{ 362, 96, 35 },
		{ 362, 97, 35 },
		{ 362, 98, 35 },
		{ 362, 99, 35 },
		{ 362, 100, 35 },
		{ 362, 101, 35 },
		{ 362, 102, 35 },
		{ 362, 103, 35 },
		{ 362, 104, 35 },
		{ 362, 105, 35 },
		{ 362, 106, 35 },
		{ 362, 107, 35 },
		{ 362, 108, 35 },
		{ 362, 109, 35 },
		{ 362, 110, 35 },
		{ 362, 111, 35 },
		{ 362, 112, 35 },
		{ 362, 113, 35 },
		{ 362, 114, 35 },
		{ 362, 115, 35 },
		{ 362, 116, 35 },
		{ 363, 46, 35 },
		{ 363, 47, 35 },
		{ 363, 48, 35 },
		{ 363, 49, 35 },
		{ 363, 50, 35 },
		{ 363, 51, 35 },
		{ 363, 52, 35 },
		{ 363, 53, 35 },
		{ 363, 54, 35 },
		{ 363, 55, 35 },
		{ 363, 56, 35 },
		{ 363, 57, 35 },
		{ 363, 58, 35 },
		{ 363, 59, 35 },
		{ 363, 60, 35 },
		{ 363, 61, 35 },
		{ 363, 62, 35 },
		{ 363, 63, 35 },
		{ 363, 64, 35 },
		{ 363, 65, 35 },
		{ 363, 66, 35 },
		{ 363, 67, 35 },
		{ 363, 68, 35 },
		{ 363, 69, 35 },
		{ 363, 70, 35 },
		{ 363, 71, 35 },
		{ 363, 72, 35 },
		{ 363, 73, 35 },
		{ 363, 74, 35 },
		{ 363, 75, 35 },
		{ 363, 76, 35 },
		{ 363, 77, 35 },
		{ 363, 78, 35 },
		{ 363, 79, 35 },
		{ 363, 80, 35 },
		{ 363, 81, 35 },
		{ 363, 82, 35 },
		{ 363, 83, 35 },
		{ 363, 84, 35 },
		{ 363, 85, 35 },
		{ 363, 86, 35 },
		{ 363, 87, 35 },
		{ 363, 88, 35 },
		{ 363, 89, 35 },
		{ 363, 90, 35 },
		{ 363, 91, 35 },
		{ 363, 92, 35 },
		{ 363, 93, 35 },
		{ 363, 94, 35 },
		{ 363, 95, 35 },
		{ 363, 96, 35 },
		{ 363, 97, 35 },
		{ 363, 98, 35 },
		{ 363, 99, 35 },
		{ 363, 100, 35 },
		{ 363, 101, 35 },
		{ 363, 102, 35 },
		{ 363, 103, 35 },
		{ 363, 104, 35 },
		{ 363, 105, 35 },
		{ 363, 106, 35 },
		{ 363, 107, 35 },
		{ 363, 108, 35 },
		{ 363, 109, 35 },
		{ 363, 110, 35 },
		{ 363, 111, 35 },
		{ 363, 112, 35 },
		{ 363, 113, 35 },
		{ 363, 114, 35 },
		{ 363, 115, 35 },
		{ 364, 44, 35 },
		{ 364, 45, 35 },
		{ 364, 46, 35 },
		{ 364, 47, 35 },
		{ 364, 48, 35 },
		{ 364, 49, 35 },
		{ 364, 50, 35 },
		{ 364, 51, 35 },
		{ 364, 52, 35 },
		{ 364, 53, 35 },
		{ 364, 54, 35 },
		{ 364, 55, 35 },
		{ 364, 56, 35 },
		{ 364, 57, 35 },
		{ 364, 58, 35 },
		{ 364, 59, 35 },
		{ 364, 60, 35 },
		{ 364, 61, 35 },
		{ 364, 62, 35 },
		{ 364, 63, 35 },
		{ 364, 64, 35 },
		{ 364, 65, 35 },
		{ 364, 66, 35 },
		{ 364, 67, 35 },
		{ 364, 68, 35 },
		{ 364, 69, 35 },
		{ 364, 70, 35 },
		{ 364, 71, 35 },
		{ 364, 72, 35 },
		{ 364, 73, 35 },
		{ 364, 74, 35 },
		{ 364, 75, 35 },
		{ 364, 76, 35 },
		{ 364, 77, 35 },
		{ 364, 78, 35 },
		{ 364, 79, 35 },
		{ 364, 80, 35 },
		{ 364, 81, 35 },
		{ 364, 82, 35 },
		{ 364, 83, 35 },
		{ 364, 84, 35 },
		{ 364, 85, 35 },
		{ 364, 86, 35 },
		{ 364, 87, 35 },
		{ 364, 88, 35 },
		{ 364, 89, 35 },
		{ 364, 90, 35 },
		{ 364, 91, 35 },
		{ 364, 92, 35 },
		{ 364, 93, 35 },
		{ 364, 94, 35 },
		{ 364, 95, 35 },
		{ 364, 96, 35 },
		{ 364, 97, 35 },
		{ 364, 98, 35 },
		{ 364, 99, 35 },
		{ 364, 100, 35 },
		{ 364, 101, 35 },
		{ 364, 102, 35 },
		{ 364, 103, 35 },
		{ 364, 104, 35 },
		{ 364, 105, 35 },
		{ 364, 106, 35 },
		{ 364, 107, 35 },
		{ 364, 108, 35 },
		{ 364, 109, 35 },
		{ 364, 110, 35 },
		{ 364, 111, 35 },
		{ 364, 112, 35 },
		{ 364, 113, 35 },
		{ 364, 114, 35 },
		{ 365, 42, 35 },
		{ 365, 43, 35 },
		{ 365, 44, 35 },
		{ 365, 45, 35 },
		{ 365, 46, 35 },
		{ 365, 47, 35 },
		{ 365, 48, 35 },
		{ 365, 49, 35 },
		{ 365, 50, 35 },
		{ 365, 51, 35 },
		{ 365, 52, 35 },
		{ 365, 53, 35 },
		{ 365, 54, 35 },
		{ 365, 55, 35 },
		{ 365, 56, 35 },
		{ 365, 57, 35 },
		{ 365, 58, 35 },
		{ 365, 59, 35 },
		{ 365, 60, 35 },
		{ 365, 61, 35 },
		{ 365, 62, 35 },
		{ 365, 63, 35 },
		{ 365, 64, 35 },
		{ 365, 65, 35 },
		{ 365, 66, 35 },
		{ 365, 67, 35 },
		{ 365, 68, 35 },
		{ 365, 69, 35 },
		{ 365, 70, 35 },
		{ 365, 71, 35 },
		{ 365, 72, 35 },
		{ 365, 73, 35 },
		{ 365, 74, 35 },
		{ 365, 75, 35 },
		{ 365, 76, 35 },
		{ 365, 77, 35 },
		{ 365, 78, 35 },
		{ 365, 79, 35 },
		{ 365, 80, 35 },
		{ 365, 81, 35 },
		{ 365, 82, 35 },
		{ 365, 83, 35 },
		{ 365, 84, 35 },
		{ 365, 85, 35 },
		{ 365, 86, 35 },
		{ 365, 87, 35 },
		{ 365, 88, 35 },
		{ 365, 89, 35 },
		{ 365, 90, 35 },
		{ 365, 91, 35 },
		{ 365, 92, 35 },
		{ 365, 93, 35 },
		{ 365, 94, 35 },
		{ 365, 95, 35 },
		{ 365, 96, 35 },
		{ 365, 97, 35 },
		{ 365, 98, 35 },
		{ 365, 99, 35 },
		{ 365, 100, 35 },
		{ 365, 101, 35 },
		{ 365, 102, 35 },
		{ 365, 103, 35 },
		{ 365, 104, 35 },
		{ 365, 105, 35 },
		{ 365, 106, 35 },
		{ 365, 107, 35 },
		{ 365, 108, 35 },
		{ 365, 109, 35 },
		{ 365, 110, 35 },
		{ 365, 111, 35 },
		{ 365, 112, 35 },
		{ 366, 40, 35 },
		{ 366, 41, 35 },
		{ 366, 42, 35 },
		{ 366, 43, 35 },
		{ 366, 44, 35 },
		{ 366, 45, 35 },
		{ 366, 46, 35 },
		{ 366, 47, 35 },
		{ 366, 48, 35 },
		{ 366, 49, 35 },
		{ 366, 50, 35 },
		{ 366, 51, 35 },
		{ 366, 52, 35 },
		{ 366, 53, 35 },
		{ 366, 54, 35 },
		{ 366, 55, 35 },
		{ 366, 56, 35 },
		{ 366, 57, 35 },
		{ 366, 58, 35 },
		{ 366, 59, 35 },
		{ 366, 60, 35 },
		{ 366, 61, 35 },
		{ 366, 62, 35 },
		{ 366, 63, 35 },
		{ 366, 64, 35 },
		{ 366, 65, 35 },
		{ 366, 66, 35 },
		{ 366, 67, 35 },
		{ 366, 68, 35 },
		{ 366, 69, 35 },
		{ 366, 70, 35 },
		{ 366, 71, 35 },
		{ 366, 72, 35 },
		{ 366, 73, 35 },
		{ 366, 74, 35 },
		{ 366, 75, 35 },
		{ 366, 76, 35 },
		{ 366, 77, 35 },
		{ 366, 78, 35 },
		{ 366, 79, 35 },
		{ 366, 80, 35 },
		{ 366, 81, 35 },
		{ 366, 82, 35 },
		{ 366, 83, 35 },
		{ 366, 84, 35 },
		{ 366, 85, 35 },
		{ 366, 86, 35 },
		{ 366, 87, 35 },
		{ 366, 88, 35 },
		{ 366, 89, 35 },
		{ 366, 90, 35 },
		{ 366, 91, 35 },
		{ 366, 92, 35 },
		{ 366, 93, 35 },
		{ 366, 94, 35 },
		{ 366, 95, 35 },
		{ 366, 96, 35 },
		{ 366, 97, 35 },
		{ 366, 98, 35 },
		{ 366, 99, 35 },
		{ 366, 100, 35 },
		{ 366, 101, 35 },
		{ 366, 102, 35 },
		{ 366, 103, 35 },
		{ 366, 104, 35 },
		{ 366, 105, 35 },
		{ 366, 106, 35 },
		{ 366, 107, 35 },
		{ 366, 108, 35 },
		{ 366, 109, 35 },
		{ 366, 110, 35 },
		{ 366, 111, 35 },
		{ 367, 38, 35 },
		{ 367, 39, 35 },
		{ 367, 40, 35 },
		{ 367, 41, 35 },
		{ 367, 42, 35 },
		{ 367, 43, 35 },
		{ 367, 44, 35 },
		{ 367, 45, 35 },
		{ 367, 46, 35 },
		{ 367, 47, 35 },
		{ 367, 48, 35 },
		{ 367, 49, 35 },
		{ 367, 50, 35 },
		{ 367, 51, 35 },
		{ 367, 52, 35 },
		{ 367, 53, 35 },
		{ 367, 54, 35 },
		{ 367, 55, 35 },
		{ 367, 56, 35 },
		{ 367, 57, 35 },
		{ 367, 58, 35 },
		{ 367, 59, 35 },
		{ 367, 60, 35 },
		{ 367, 61, 35 },
		{ 367, 62, 35 },
		{ 367, 63, 35 },
		{ 367, 64, 35 },
		{ 367, 65, 35 },
		{ 367, 66, 35 },
		{ 367, 67, 35 },
		{ 367, 68, 35 },
		{ 367, 69, 35 },
		{ 367, 70, 35 },
		{ 367, 71, 35 },
		{ 367, 72, 35 },
		{ 367, 73, 35 },
		{ 367, 74, 35 },
		{ 367, 75, 35 },
		{ 367, 76, 35 },
		{ 367, 77, 35 },
		{ 367, 78, 35 },
		{ 367, 79, 35 },
		{ 367, 80, 35 },
		{ 367, 81, 35 },
		{ 367, 82, 35 },
		{ 367, 83, 35 },
		{ 367, 84, 35 },
		{ 367, 85, 35 },
		{ 367, 86, 35 },
		{ 367, 87, 35 },
		{ 367, 88, 35 },
		{ 367, 89, 35 },
		{ 367, 90, 35 },
		{ 367, 91, 35 },
		{ 367, 92, 35 },
		{ 367, 93, 35 },
		{ 367, 94, 35 },
		{ 367, 95, 35 },
		{ 367, 96, 35 },
		{ 367, 97, 35 },
		{ 367, 98, 35 },
		{ 367, 99, 35 },
		{ 367, 100, 35 },
		{ 367, 101, 35 },
		{ 367, 102, 35 },
		{ 367, 103, 35 },
		{ 367, 104, 35 },
		{ 367, 105, 35 },
		{ 367, 106, 35 },
		{ 367, 107, 35 },
		{ 367, 108, 35 },
		{ 367, 109, 35 },
		{ 367, 110, 35 },
		{ 368, 36, 35 },
		{ 368, 37, 35 },
		{ 368, 38, 35 },
		{ 368, 39, 35 },
		{ 368, 40, 35 },
		{ 368, 41, 35 },
		{ 368, 42, 35 },
		{ 368, 43, 35 },
		{ 368, 44, 35 },
		{ 368, 45, 35 },
		{ 368, 46, 35 },
		{ 368, 47, 35 },
		{ 368, 48, 35 },
		{ 368, 49, 35 },
		{ 368, 50, 35 },
		{ 368, 51, 35 },
		{ 368, 52, 35 },
		{ 368, 53, 35 },
		{ 368, 54, 35 },
		{ 368, 55, 35 },
		{ 368, 56, 35 },
		{ 368, 57, 35 },
		{ 368, 58, 35 },
		{ 368, 59, 35 },
		{ 368, 60, 35 },
		{ 368, 61, 35 },
		{ 368, 62, 35 },
		{ 368, 63, 35 },
		{ 368, 64, 35 },
		{ 368, 65, 35 },
		{ 368, 66, 35 },
		{ 368, 67, 35 },
		{ 368, 68, 35 },
		{ 368, 69, 35 },
		{ 368, 70, 35 },
		{ 368, 71, 35 },
		{ 368, 72, 35 },
		{ 368, 73, 35 },
		{ 368, 74, 35 },
		{ 368, 75, 35 },
		{ 368, 76, 35 },
		{ 368, 77, 35 },
		{ 368, 78, 35 },
		{ 368, 79, 35 },
		{ 368, 80, 35 },
		{ 368, 81, 35 },
		{ 368, 82, 35 },
		{ 368, 83, 35 },
		{ 368, 84, 35 },
		{ 368, 85, 35 },
		{ 368, 86, 35 },
		{ 368, 87, 35 },
		{ 368, 88, 35 },
		{ 368, 89, 35 },
		{ 368, 90, 35 },
		{ 368, 91, 35 },
		{ 368, 92, 35 },
		{ 368, 93, 35 },
		{ 368, 94, 35 },
		{ 368, 95, 35 },
		{ 368, 96, 35 },
		{ 368, 97, 35 },
		{ 368, 98, 35 },
		{ 368, 99, 35 },
		{ 368, 100, 35 },
		{ 368, 101, 35 },
		{ 368, 102, 35 },
		{ 368, 103, 35 },
		{ 368, 104, 35 },
		{ 368, 105, 35 },
		{ 368, 106, 35 },
		{ 368, 107, 35 },
		{ 368, 108, 35 },
		{ 369, 34, 35 },
		{ 369, 35, 35 },
		{ 369, 36, 35 },
		{ 369, 37, 35 },
		{ 369, 38, 35 },
		{ 369, 39, 35 },
		{ 369, 40, 35 },
		{ 369, 41, 35 },
		{ 369, 42, 35 },
		{ 369, 43, 35 },
		{ 369, 44, 35 },
		{ 369, 45, 35 },
		{ 369, 46, 35 },
		{ 369, 47, 35 },
		{ 369, 48, 35 },
		{ 369, 49, 35 },
		{ 369, 50, 35 },
		{ 369, 51, 35 },
		{ 369, 52, 35 },
		{ 369, 53, 35 },
		{ 369, 54, 35 },
		{ 369, 55, 35 },
		{ 369, 56, 35 },
		{ 369, 57, 35 },
		{ 369, 58, 35 },
		{ 369, 59, 35 },
		{ 369, 60, 35 },
		{ 369, 61, 35 },
		{ 369, 62, 35 },
		{ 369, 63, 35 },
		{ 369, 64, 35 },
		{ 369, 65, 35 },
		{ 369, 66, 35 },
		{ 369, 67, 35 },
		{ 369, 68, 35 },
		{ 369, 69, 35 },
		{ 369, 70, 35 },
		{ 369, 71, 35 },
		{ 369, 72, 35 },
		{ 369, 73, 35 },
		{ 369, 74, 35 },
		{ 369, 75, 35 },
		{ 369, 76, 35 },
		{ 369, 77, 35 },
		{ 369, 78, 35 },
		{ 369, 79, 35 },
		{ 369, 80, 35 },
		{ 369, 81, 35 },
		{ 369, 82, 35 },
		{ 369, 83, 35 },
		{ 369, 84, 35 },
		{ 369, 85, 35 },
		{ 369, 86, 35 },
		{ 369, 87, 35 },
		{ 369, 88, 35 },
		{ 369, 89, 35 },
		{ 369, 90, 35 },
		{ 369, 91, 35 },
		{ 369, 92, 35 },
		{ 369, 93, 35 },
		{ 369, 94, 35 },
		{ 369, 95, 35 },
		{ 369, 96, 35 },
		{ 369, 97, 35 },
		{ 369, 98, 35 },
		{ 369, 99, 35 },
		{ 369, 100, 35 },
		{ 369, 101, 35 },
		{ 369, 102, 35 },
		{ 369, 103, 35 },
		{ 369, 104, 35 },
		{ 369, 105, 35 },
		{ 369, 106, 35 },
		{ 369, 107, 35 },
		{ 370, 32, 35 },
		{ 370, 33, 35 },
		{ 370, 34, 35 },
		{ 370, 35, 35 },
		{ 370, 36, 35 },
		{ 370, 37, 35 },
		{ 370, 38, 35 },
		{ 370, 39, 35 },
		{ 370, 40, 35 },
		{ 370, 41, 35 },
		{ 370, 42, 35 },
		{ 370, 43, 35 },
		{ 370, 44, 35 },
		{ 370, 45, 35 },
		{ 370, 46, 35 },
		{ 370, 47, 35 },
		{ 370, 48, 35 },
		{ 370, 49, 35 },
		{ 370, 50, 35 },
		{ 370, 51, 35 },
		{ 370, 52, 35 },
		{ 370, 53, 35 },
		{ 370, 54, 35 },
		{ 370, 55, 35 },
		{ 370, 56, 35 },
		{ 370, 57, 35 },
		{ 370, 58, 35 },
		{ 370, 59, 35 },
		{ 370, 60, 35 },
		{ 370, 61, 35 },
		{ 370, 62, 35 },
		{ 370, 63, 35 },
		{ 370, 64, 35 },
		{ 370, 65, 35 },
		{ 370, 66, 35 },
		{ 370, 67, 35 },
		{ 370, 68, 35 },
		{ 370, 69, 35 },
		{ 370, 70, 35 },
		{ 370, 71, 35 },
		{ 370, 72, 35 },
		{ 370, 73, 35 },
		{ 370, 74, 35 },
		{ 370, 75, 35 },
		{ 370, 76, 35 },
		{ 370, 77, 35 },
		{ 370, 78, 35 },
		{ 370, 79, 35 },
		{ 370, 80, 35 },
		{ 370, 81, 35 },
		{ 370, 82, 35 },
		{ 370, 83, 35 },
		{ 370, 84, 35 },
		{ 370, 85, 35 },
		{ 370, 86, 35 },
		{ 370, 87, 35 },
		{ 370, 88, 35 },
		{ 370, 89, 35 },
		{ 370, 90, 35 },
		{ 370, 91, 35 },
		{ 370, 92, 35 },
		{ 370, 93, 35 },
		{ 370, 94, 35 },
		{ 370, 95, 35 },
		{ 370, 96, 35 },
		{ 370, 97, 35 },
		{ 370, 98, 35 },
		{ 370, 99, 35 },
		{ 370, 100, 35 },
		{ 370, 101, 35 },
		{ 370, 102, 35 },
		{ 370, 103, 35 },
		{ 370, 104, 35 },
		{ 370, 105, 35 },
		{ 370, 106, 35 },
		{ 371, 30, 35 },
		{ 371, 31, 35 },
		{ 371, 32, 35 },
		{ 371, 33, 35 },
		{ 371, 34, 35 },
		{ 371, 35, 35 },
		{ 371, 36, 35 },
		{ 371, 37, 35 },
		{ 371, 38, 35 },
		{ 371, 39, 35 },
		{ 371, 40, 35 },
		{ 371, 41, 35 },
		{ 371, 42, 35 },
		{ 371, 43, 35 },
		{ 371, 44, 35 },
		{ 371, 45, 35 },
		{ 371, 46, 35 },
		{ 371, 47, 35 },
		{ 371, 48, 35 },
		{ 371, 49, 35 },
		{ 371, 50, 35 },
		{ 371, 51, 35 },
		{ 371, 52, 35 },
		{ 371, 53, 35 },
		{ 371, 54, 35 },
		{ 371, 55, 35 },
		{ 371, 56, 35 },
		{ 371, 57, 35 },
		{ 371, 58, 35 },
		{ 371, 59, 35 },
		{ 371, 60, 35 },
		{ 371, 61, 35 },
		{ 371, 62, 35 },
		{ 371, 63, 35 },
		{ 371, 64, 35 },
		{ 371, 65, 35 },
		{ 371, 66, 35 },
		{ 371, 67, 35 },
		{ 371, 68, 35 },
		{ 371, 69, 35 },
		{ 371, 70, 35 },
		{ 371, 71, 35 },
		{ 371, 72, 35 },
		{ 371, 73, 35 },
		{ 371, 74, 35 },
		{ 371, 75, 35 },
		{ 371, 76, 35 },
		{ 371, 77, 35 },
		{ 371, 78, 35 },
		{ 371, 79, 35 },
		{ 371, 80, 35 },
		{ 371, 81, 35 },
		{ 371, 82, 35 },
		{ 371, 83, 35 },
		{ 371, 84, 35 },
		{ 371, 85, 35 },
		{ 371, 86, 35 },
		{ 371, 87, 35 },
		{ 371, 88, 35 },
		{ 371, 89, 35 },
		{ 371, 90, 35 },
		{ 371, 91, 35 },
		{ 371, 92, 35 },
		{ 371, 93, 35 },
		{ 371, 94, 35 },
		{ 371, 95, 35 },
		{ 371, 96, 35 },
		{ 371, 97, 35 },
		{ 371, 98, 35 },
		{ 371, 99, 35 },
		{ 371, 100, 35 },
		{ 371, 101, 35 },
		{ 371, 102, 35 },
		{ 371, 103, 35 },
		{ 371, 104, 35 },
		{ 372, 28, 35 },
		{ 372, 29, 35 },
		{ 372, 30, 35 },
		{ 372, 31, 35 },
		{ 372, 32, 35 },
		{ 372, 33, 35 },
		{ 372, 34, 35 },
		{ 372, 35, 35 },
		{ 372, 36, 35 },
		{ 372, 37, 35 },
		{ 372, 38, 35 },
		{ 372, 39, 35 },
		{ 372, 40, 35 },
		{ 372, 41, 35 },
		{ 372, 42, 35 },
		{ 372, 43, 35 },
		{ 372, 44, 35 },
		{ 372, 45, 35 },
		{ 372, 46, 35 },
		{ 372, 47, 35 },
		{ 372, 48, 35 },
		{ 372, 49, 35 },
		{ 372, 50, 35 },
		{ 372, 51, 35 },
		{ 372, 52, 35 },
		{ 372, 53, 35 },
		{ 372, 54, 35 },
		{ 372, 55, 35 },
		{ 372, 56, 35 },
		{ 372, 57, 35 },
		{ 372, 58, 35 },
		{ 372, 59, 35 },
		{ 372, 60, 35 },
		{ 372, 61, 35 },
		{ 372, 62, 35 },
		{ 372, 63, 35 },
		{ 372, 64, 35 },
		{ 372, 65, 35 },
		{ 372, 66, 35 },
		{ 372, 67, 35 },
		{ 372, 68, 35 },
		{ 372, 69, 35 },
		{ 372, 70, 35 },
		{ 372, 71, 35 },
		{ 372, 72, 35 },
		{ 372, 73, 35 },
		{ 372, 74, 35 },
		{ 372, 75, 35 },
		{ 372, 76, 35 },
		{ 372, 77, 35 },
		{ 372, 78, 35 },
		{ 372, 79, 35 },
		{ 372, 80, 35 },
		{ 372, 81, 35 },
		{ 372, 82, 35 },
		{ 372, 83, 35 },
		{ 372, 84, 35 },
		{ 372, 85, 35 },
		{ 372, 86, 35 },
		{ 372, 87, 35 },
		{ 372, 88, 35 },
		{ 372, 89, 35 },
		{ 372, 90, 35 },
		{ 372, 91, 35 },
		{ 372, 92, 35 },
		{ 372, 93, 35 },
		{ 372, 94, 35 },
		{ 372, 95, 35 },
		{ 372, 96, 35 },
		{ 372, 97, 35 },
		{ 372, 98, 35 },
		{ 372, 99, 35 },
		{ 372, 100, 35 },
		{ 372, 101, 35 },
		{ 372, 102, 35 },
		{ 372, 103, 35 },
		{ 373, 26, 35 },
		{ 373, 27, 35 },
		{ 373, 28, 35 },
		{ 373, 29, 35 },
		{ 373, 30, 35 },
		{ 373, 31, 35 },
		{ 373, 32, 35 },
		{ 373, 33, 35 },
		{ 373, 34, 35 },
		{ 373, 35, 35 },
		{ 373, 36, 35 },
		{ 373, 37, 35 },
		{ 373, 38, 35 },
		{ 373, 39, 35 },
		{ 373, 40, 35 },
		{ 373, 41, 35 },
		{ 373, 42, 35 },
		{ 373, 43, 35 },
		{ 373, 44, 35 },
		{ 373, 45, 35 },
		{ 373, 46, 35 },
		{ 373, 47, 35 },
		{ 373, 48, 35 },
		{ 373, 49, 35 },
		{ 373, 50, 35 },
		{ 373, 51, 35 },
		{ 373, 52, 35 },
		{ 373, 53, 35 },
		{ 373, 54, 35 },
		{ 373, 55, 35 },
		{ 373, 56, 35 },
		{ 373, 57, 35 },
		{ 373, 58, 35 },
		{ 373, 59, 35 },
		{ 373, 60, 35 },
		{ 373, 61, 35 },
		{ 373, 62, 35 },
		{ 373, 63, 35 },
		{ 373, 64, 35 },
		{ 373, 65, 35 },
		{ 373, 66, 35 },
		{ 373, 67, 35 },
		{ 373, 68, 35 },
		{ 373, 69, 35 },
		{ 373, 70, 35 },
		{ 373, 71, 35 },
		{ 373, 72, 35 },
		{ 373, 73, 35 },
		{ 373, 74, 35 },
		{ 373, 75, 35 },
		{ 373, 76, 35 },
		{ 373, 77, 35 },
		{ 373, 78, 35 },
		{ 373, 79, 35 },
		{ 373, 80, 35 },
		{ 373, 81, 35 },
		{ 373, 82, 35 },
		{ 373, 83, 35 },
		{ 373, 84, 35 },
		{ 373, 85, 35 },
		{ 373, 86, 35 },
		{ 373, 87, 35 },
		{ 373, 88, 35 },
		{ 373, 89, 35 },
		{ 373, 90, 35 },
		{ 373, 91, 35 },
		{ 373, 92, 35 },
		{ 373, 93, 35 },
		{ 373, 94, 35 },
		{ 373, 95, 35 },
		{ 373, 96, 35 },
		{ 373, 97, 35 },
		{ 373, 98, 35 },
		{ 373, 99, 35 },
		{ 373, 100, 35 },
		{ 373, 101, 35 },
		{ 373, 102, 35 },
		{ 374, 24, 35 },
		{ 374, 25, 35 },
		{ 374, 26, 35 },
		{ 374, 27, 35 },
		{ 374, 28, 35 },
		{ 374, 29, 35 },
		{ 374, 30, 35 },
		{ 374, 31, 35 },
		{ 374, 32, 35 },
		{ 374, 33, 35 },
		{ 374, 34, 35 },
		{ 374, 35, 35 },
		{ 374, 36, 35 },
		{ 374, 37, 35 },
		{ 374, 38, 35 },
		{ 374, 39, 35 },
		{ 374, 40, 35 },
		{ 374, 41, 35 },
		{ 374, 42, 35 },
		{ 374, 43, 35 },
		{ 374, 44, 35 },
		{ 374, 45, 35 },
		{ 374, 46, 35 },
		{ 374, 47, 35 },
		{ 374, 48, 35 },
		{ 374, 49, 35 },
		{ 374, 50, 35 },
		{ 374, 51, 35 },
		{ 374, 52, 35 },
		{ 374, 53, 35 },
		{ 374, 54, 35 },
		{ 374, 55, 35 },
		{ 374, 56, 35 },
		{ 374, 57, 35 },
		{ 374, 58, 35 },
		{ 374, 59, 35 },
		{ 374, 60, 35 },
		{ 374, 61, 35 },
		{ 374, 62, 35 },
		{ 374, 63, 35 },
		{ 374, 64, 35 },
		{ 374, 65, 35 },
		{ 374, 66, 35 },
		{ 374, 67, 35 },
		{ 374, 68, 35 },
		{ 374, 69, 35 },
		{ 374, 70, 35 },
		{ 374, 71, 35 },
		{ 374, 72, 35 },
		{ 374, 73, 35 },
		{ 374, 74, 35 },
		{ 374, 75, 35 },
		{ 374, 76, 35 },
		{ 374, 77, 35 },
		{ 374, 78, 35 },
		{ 374, 79, 35 },
		{ 374, 80, 35 },
		{ 374, 81, 35 },
		{ 374, 82, 35 },
		{ 374, 83, 35 },
		{ 374, 84, 35 },
		{ 374, 85, 35 },
		{ 374, 86, 35 },
		{ 374, 87, 35 },
		{ 374, 88, 35 },
		{ 374, 89, 35 },
		{ 374, 90, 35 },
		{ 374, 91, 35 },
		{ 374, 92, 35 },
		{ 374, 93, 35 },
		{ 374, 94, 35 },
		{ 374, 95, 35 },
		{ 374, 96, 35 },
		{ 374, 97, 35 },
		{ 374, 98, 35 },
		{ 374, 99, 35 },
		{ 374, 100, 35 },
		{ 375, 22, 35 },
		{ 375, 23, 35 },
		{ 375, 24, 35 },
		{ 375, 25, 35 },
		{ 375, 26, 35 },
		{ 375, 27, 35 },
		{ 375, 28, 35 },
		{ 375, 29, 35 },
		{ 375, 30, 35 },
		{ 375, 31, 35 },
		{ 375, 32, 35 },
		{ 375, 33, 35 },
		{ 375, 34, 35 },
		{ 375, 35, 35 },
		{ 375, 36, 35 },
		{ 375, 37, 35 },
		{ 375, 38, 35 },
		{ 375, 39, 35 },
		{ 375, 40, 35 },
		{ 375, 41, 35 },
		{ 375, 42, 35 },
		{ 375, 43, 35 },
		{ 375, 44, 35 },
		{ 375, 45, 35 },
		{ 375, 46, 35 },
		{ 375, 47, 35 },
		{ 375, 48, 35 },
		{ 375, 49, 35 },
		{ 375, 50, 35 },
		{ 375, 51, 35 },
		{ 375, 52, 35 },
		{ 375, 53, 35 },
		{ 375, 54, 35 },
		{ 375, 55, 35 },
		{ 375, 56, 35 },
		{ 375, 57, 35 },
		{ 375, 58, 35 },
		{ 375, 59, 35 },
		{ 375, 60, 35 },
		{ 375, 61, 35 },
		{ 375, 62, 35 },
		{ 375, 63, 35 },
		{ 375, 64, 35 },
		{ 375, 65, 35 },
		{ 375, 66, 35 },
		{ 375, 67, 35 },
		{ 375, 68, 35 },
		{ 375, 69, 35 },
		{ 375, 70, 35 },
		{ 375, 71, 35 },
		{ 375, 72, 35 },
		{ 375, 73, 35 },
		{ 375, 74, 35 },
		{ 375, 75, 35 },
		{ 375, 76, 35 },
		{ 375, 77, 35 },
		{ 375, 78, 35 },
		{ 375, 79, 35 },
		{ 375, 80, 35 },
		{ 375, 81, 35 },
		{ 375, 82, 35 },
		{ 375, 83, 35 },
		{ 375, 84, 35 },
		{ 375, 85, 35 },
		{ 375, 86, 35 },
		{ 375, 87, 35 },
		{ 375, 88, 35 },
		{ 375, 89, 35 },
		{ 375, 90, 35 },
		{ 375, 91, 35 },
		{ 375, 92, 35 },
		{ 375, 93, 35 },
		{ 375, 94, 35 },
		{ 375, 95, 35 },
		{ 375, 96, 35 },
		{ 375, 97, 35 },
		{ 375, 98, 35 },
		{ 375, 99, 35 },
		{ 376, 20, 35 },
		{ 376, 21, 35 },
		{ 376, 22, 35 },
		{ 376, 23, 35 },
		{ 376, 24, 35 },
		{ 376, 25, 35 },
		{ 376, 26, 35 },
		{ 376, 27, 35 },
		{ 376, 28, 35 },
		{ 376, 29, 35 },
		{ 376, 30, 35 },
		{ 376, 31, 35 },
		{ 376, 32, 35 },
		{ 376, 33, 35 },
		{ 376, 34, 35 },
		{ 376, 35, 35 },
		{ 376, 36, 35 },
		{ 376, 37, 35 },
		{ 376, 38, 35 },
		{ 376, 39, 35 },
		{ 376, 40, 35 },
		{ 376, 41, 35 },
		{ 376, 42, 35 },
		{ 376, 43, 35 },
		{ 376, 44, 35 },
		{ 376, 45, 35 },
		{ 376, 46, 35 },
		{ 376, 47, 35 },
		{ 376, 48, 35 },
		{ 376, 49, 35 },
		{ 376, 50, 35 },
		{ 376, 51, 35 },
		{ 376, 52, 35 },
		{ 376, 53, 35 },
		{ 376, 54, 35 },
		{ 376, 55, 35 },
		{ 376, 56, 35 },
		{ 376, 57, 35 },
		{ 376, 58, 35 },
		{ 376, 59, 35 },
		{ 376, 60, 35 },
		{ 376, 61, 35 },
		{ 376, 62, 35 },
		{ 376, 63, 35 },
		{ 376, 64, 35 },
		{ 376, 65, 35 },
		{ 376, 66, 35 },
		{ 376, 67, 35 },
		{ 376, 68, 35 },
		{ 376, 69, 35 },
		{ 376, 70, 35 },
		{ 376, 71, 35 },
		{ 376, 72, 35 },
		{ 376, 73, 35 },
		{ 376, 74, 35 },
		{ 376, 75, 35 },
		{ 376, 76, 35 },
		{ 376, 77, 35 },
		{ 376, 78, 35 },
		{ 376, 79, 35 },
		{ 376, 80, 35 },
		{ 376, 81, 35 },
		{ 376, 82, 35 },
		{ 376, 83, 35 },
		{ 376, 84, 35 },
		{ 376, 85, 35 },
		{ 376, 86, 35 },
		{ 376, 87, 35 },
		{ 376, 88, 35 },
		{ 376, 89, 35 },
		{ 376, 90, 35 },
		{ 376, 91, 35 },
		{ 376, 92, 35 },
		{ 376, 93, 35 },
		{ 376, 94, 35 },
		{ 376, 95, 35 },
		{ 376, 96, 35 },
		{ 376, 97, 35 },
		{ 376, 98, 35 },
		{ 377, 18, 35 },
		{ 377, 19, 35 },
		{ 377, 20, 35 },
		{ 377, 21, 35 },
		{ 377, 22, 35 },
		{ 377, 23, 35 },
		{ 377, 24, 35 },
		{ 377, 25, 35 },
		{ 377, 26, 35 },
		{ 377, 27, 35 },
		{ 377, 28, 35 },
		{ 377, 29, 35 },
		{ 377, 30, 35 },
		{ 377, 31, 35 },
		{ 377, 32, 35 },
		{ 377, 33, 35 },
		{ 377, 34, 35 },
		{ 377, 35, 35 },
		{ 377, 36, 35 },
		{ 377, 37, 35 },
		{ 377, 38, 35 },
		{ 377, 39, 35 },
		{ 377, 40, 35 },
		{ 377, 41, 35 },
		{ 377, 42, 35 },
		{ 377, 43, 35 },
		{ 377, 44, 35 },
		{ 377, 45, 35 },
		{ 377, 46, 35 },
		{ 377, 47, 35 },
		{ 377, 48, 35 },
		{ 377, 49, 35 },
		{ 377, 50, 35 },
		{ 377, 51, 35 },
		{ 377, 52, 35 },
		{ 377, 53, 35 },
		{ 377, 54, 35 },
		{ 377, 55, 35 },
		{ 377, 56, 35 },
		{ 377, 57, 35 },
		{ 377, 58, 35 },
		{ 377, 59, 35 },
		{ 377, 60, 35 },
		{ 377, 61, 35 },
		{ 377, 62, 35 },
		{ 377, 63, 35 },
		{ 377, 64, 35 },
		{ 377, 65, 35 },
		{ 377, 66, 35 },
		{ 377, 67, 35 },
		{ 377, 68, 35 },
		{ 377, 69, 35 },
		{ 377, 70, 35 },
		{ 377, 71, 35 },
		{ 377, 72, 35 },
		{ 377, 73, 35 },
		{ 377, 74, 35 },
		{ 377, 75, 35 },
		{ 377, 76, 35 },
		{ 377, 77, 35 },
		{ 377, 78, 35 },
		{ 377, 79, 35 },
		{ 377, 80, 35 },
		{ 377, 81, 35 },
		{ 377, 82, 35 },
		{ 377, 83, 35 },
		{ 377, 84, 35 },
		{ 377, 85, 35 },
		{ 377, 86, 35 },
		{ 377, 87, 35 },
		{ 377, 88, 35 },
		{ 377, 89, 35 },
		{ 377, 90, 35 },
		{ 377, 91, 35 },
		{ 377, 92, 35 },
		{ 377, 93, 35 },
		{ 377, 94, 35 },
		{ 377, 95, 35 },
		{ 377, 96, 35 },
		{ 378, 16, 35 },
		{ 378, 17, 35 },
		{ 378, 18, 35 },
		{ 378, 19, 35 },
		{ 378, 20, 35 },
		{ 378, 21, 35 },
		{ 378, 22, 35 },
		{ 378, 23, 35 },
		{ 378, 24, 35 },
		{ 378, 25, 35 },
		{ 378, 26, 35 },
		{ 378, 27, 35 },
		{ 378, 28, 35 },
		{ 378, 29, 35 },
		{ 378, 30, 35 },
		{ 378, 31, 35 },
		{ 378, 32, 35 },
		{ 378, 33, 35 },
		{ 378, 34, 35 },
		{ 378, 35, 35 },
		{ 378, 36, 35 },
		{ 378, 37, 35 },
		{ 378, 38, 35 },
		{ 378, 39, 35 },
		{ 378, 40, 35 },
		{ 378, 41, 35 },
		{ 378, 42, 35 },
		{ 378, 43, 35 },
		{ 378, 44, 35 },
		{ 378, 45, 35 },
		{ 378, 46, 35 },
		{ 378, 47, 35 },
		{ 378, 48, 35 },
		{ 378, 49, 35 },
		{ 378, 50, 35 },
		{ 378, 51, 35 },
		{ 378, 52, 35 },
		{ 378, 53, 35 },
		{ 378, 54, 35 },
		{ 378, 55, 35 },
		{ 378, 56, 35 },
		{ 378, 57, 35 },
		{ 378, 58, 35 },
		{ 378, 59, 35 },
		{ 378, 60, 35 },
		{ 378, 61, 35 },
		{ 378, 62, 35 },
		{ 378, 63, 35 },
		{ 378, 64, 35 },
		{ 378, 65, 35 },
		{ 378, 66, 35 },
		{ 378, 67, 35 },
		{ 378, 68, 35 },
		{ 378, 69, 35 },
		{ 378, 70, 35 },
		{ 378, 71, 35 },
		{ 378, 72, 35 },
		{ 378, 73, 35 },
		{ 378, 74, 35 },
		{ 378, 75, 35 },
		{ 378, 76, 35 },
		{ 378, 77, 35 },
		{ 378, 78, 35 },
		{ 378, 79, 35 },
		{ 378, 80, 35 },
		{ 378, 81, 35 },
		{ 378, 82, 35 },
		{ 378, 83, 35 },
		{ 378, 84, 35 },
		{ 378, 85, 35 },
		{ 378, 86, 35 },
		{ 378, 87, 35 },
		{ 378, 88, 35 },
		{ 378, 89, 35 },
		{ 378, 90, 35 },
		{ 378, 91, 35 },
		{ 378, 92, 35 },
		{ 378, 93, 35 },
		{ 378, 94, 35 },
		{ 378, 95, 35 },
		{ 379, 14, 35 },
		{ 379, 15, 35 },
		{ 379, 16, 35 },
		{ 379, 17, 35 },
		{ 379, 18, 35 },
		{ 379, 19, 35 },
		{ 379, 20, 35 },
		{ 379, 21, 35 },
		{ 379, 22, 35 },
		{ 379, 23, 35 },
		{ 379, 24, 35 },
		{ 379, 25, 35 },
		{ 379, 26, 35 },
		{ 379, 27, 35 },
		{ 379, 28, 35 },
		{ 379, 29, 35 },
		{ 379, 30, 35 },
		{ 379, 31, 35 },
		{ 379, 32, 35 },
		{ 379, 33, 35 },
		{ 379, 34, 35 },
		{ 379, 35, 35 },
		{ 379, 36, 35 },
		{ 379, 37, 35 },
		{ 379, 38, 35 },
		{ 379, 39, 35 },
		{ 379, 40, 35 },
		{ 379, 41, 35 },
		{ 379, 42, 35 },
		{ 379, 43, 35 },
		{ 379, 44, 35 },
		{ 379, 45, 35 },
		{ 379, 46, 35 },
		{ 379, 47, 35 },
		{ 379, 48, 35 },
		{ 379, 49, 35 },
		{ 379, 50, 35 },
		{ 379, 51, 35 },
		{ 379, 52, 35 },
		{ 379, 53, 35 },
		{ 379, 54, 35 },
		{ 379, 55, 35 },
		{ 379, 56, 35 },
		{ 379, 57, 35 },
		{ 379, 58, 35 },
		{ 379, 59, 35 },
		{ 379, 60, 35 },
		{ 379, 61, 35 },
		{ 379, 62, 35 },
		{ 379, 63, 35 },
		{ 379, 64, 35 },
		{ 379, 65, 35 },
		{ 379, 66, 35 },
		{ 379, 67, 35 },
		{ 379, 68, 35 },
		{ 379, 69, 35 },
		{ 379, 70, 35 },
		{ 379, 71, 35 },
		{ 379, 72, 35 },
		{ 379, 73, 35 },
		{ 379, 74, 35 },
		{ 379, 75, 35 },
		{ 379, 76, 35 },
		{ 379, 77, 35 },
		{ 379, 78, 35 },
		{ 379, 79, 35 },
		{ 379, 80, 35 },
		{ 379, 81, 35 },
		{ 379, 82, 35 },
		{ 379, 83, 35 },
		{ 379, 84, 35 },
		{ 379, 85, 35 },
		{ 379, 86, 35 },
		{ 379, 87, 35 },
		{ 379, 88, 35 },
		{ 379, 89, 35 },
		{ 379, 90, 35 },
		{ 379, 91, 35 },
		{ 379, 92, 35 },
		{ 379, 93, 35 },
		{ 379, 94, 35 },
		{ 380, 12, 35 },
		{ 380, 13, 35 },
		{ 380, 14, 35 },
		{ 380, 15, 35 },
		{ 380, 16, 35 },
		{ 380, 17, 35 },
		{ 380, 18, 35 },
		{ 380, 19, 35 },
		{ 380, 20, 35 },
		{ 380, 21, 35 },
		{ 380, 22, 35 },
		{ 380, 23, 35 },
		{ 380, 24, 35 },
		{ 380, 25, 35 },
		{ 380, 26, 35 },
		{ 380, 27, 35 },
		{ 380, 28, 35 },
		{ 380, 29, 35 },
		{ 380, 30, 35 },
		{ 380, 31, 35 },
		{ 380, 32, 35 },
		{ 380, 33, 35 },
		{ 380, 34, 35 },
		{ 380, 35, 35 },
		{ 380, 36, 35 },
		{ 380, 37, 35 },
		{ 380, 38, 35 },
		{ 380, 39, 35 },
		{ 380, 40, 35 },
		{ 380, 41, 35 },
		{ 380, 42, 35 },
		{ 380, 43, 35 },
		{ 380, 44, 35 },
		{ 380, 45, 35 },
		{ 380, 46, 35 },
		{ 380, 47, 35 },
		{ 380, 48, 35 },
		{ 380, 49, 35 },
		{ 380, 50, 35 },
		{ 380, 51, 35 },
		{ 380, 52, 35 },
		{ 380, 53, 35 },
		{ 380, 54, 35 },
		{ 380, 55, 35 },
		{ 380, 56, 35 },
		{ 380, 57, 35 },
		{ 380, 58, 35 },
		{ 380, 59, 35 },
		{ 380, 60, 35 },
		{ 380, 61, 35 },
		{ 380, 62, 35 },
		{ 380, 63, 35 },
		{ 380, 64, 35 },
		{ 380, 65, 35 },
		{ 380, 66, 35 },
		{ 380, 67, 35 },
		{ 380, 68, 35 },
		{ 380, 69, 35 },
		{ 380, 70, 35 },
		{ 380, 71, 35 },
		{ 380, 72, 35 },
		{ 380, 73, 35 },
		{ 380, 74, 35 },
		{ 380, 75, 35 },
		{ 380, 76, 35 },
		{ 380, 77, 35 },
		{ 380, 78, 35 },
		{ 380, 79, 35 },
		{ 380, 80, 35 },
		{ 380, 81, 35 },
		{ 380, 82, 35 },
		{ 380, 83, 35 },
		{ 380, 84, 35 },
		{ 380, 85, 35 },
		{ 380, 86, 35 },
		{ 380, 87, 35 },
		{ 380, 88, 35 },
		{ 380, 89, 35 },
		{ 380, 90, 35 },
		{ 380, 91, 35 },
		{ 380, 92, 35 },
		{ 381, 10, 35 },
		{ 381, 11, 35 },
		{ 381, 12, 35 },
		{ 381, 13, 35 },
		{ 381, 14, 35 },
		{ 381, 15, 35 },
		{ 381, 16, 35 },
		{ 381, 17, 35 },
		{ 381, 18, 35 },
		{ 381, 19, 35 },
		{ 381, 20, 35 },
		{ 381, 21, 35 },
		{ 381, 22, 35 },
		{ 381, 23, 35 },
		{ 381, 24, 35 },
		{ 381, 25, 35 },
		{ 381, 26, 35 },
		{ 381, 27, 35 },
		{ 381, 28, 35 },
		{ 381, 29, 35 },
		{ 381, 30, 35 },
		{ 381, 31, 35 },
		{ 381, 32, 35 },
		{ 381, 33, 35 },
		{ 381, 34, 35 },
		{ 381, 35, 35 },
		{ 381, 36, 35 },
		{ 381, 37, 35 },
		{ 381, 38, 35 },
		{ 381, 39, 35 },
		{ 381, 40, 35 },
		{ 381, 41, 35 },
		{ 381, 42, 35 },
		{ 381, 43, 35 },
		{ 381, 44, 35 },
		{ 381, 45, 35 },
		{ 381, 46, 35 },
		{ 381, 47, 35 },
		{ 381, 48, 35 },
		{ 381, 49, 35 },
		{ 381, 50, 35 },
		{ 381, 51, 35 },
		{ 381, 52, 35 },
		{ 381, 53, 35 },
		{ 381, 54, 35 },
		{ 381, 55, 35 },
		{ 381, 56, 35 },
		{ 381, 57, 35 },
		{ 381, 58, 35 },
		{ 381, 59, 35 },
		{ 381, 60, 35 },
		{ 381, 61, 35 },
		{ 381, 62, 35 },
		{ 381, 63, 35 },
		{ 381, 64, 35 },
		{ 381, 65, 35 },
		{ 381, 66, 35 },
		{ 381, 67, 35 },
		{ 381, 68, 35 },
		{ 381, 69, 35 },
		{ 381, 70, 35 },
		{ 381, 71, 35 },
		{ 381, 72, 35 },
		{ 381, 73, 35 },
		{ 381, 74, 35 },
		{ 381, 75, 35 },
		{ 381, 76, 35 },
		{ 381, 77, 35 },
		{ 381, 78, 35 },
		{ 381, 79, 35 },
		{ 381, 80, 35 },
		{ 381, 81, 35 },
		{ 381, 82, 35 },
		{ 381, 83, 35 },
		{ 381, 84, 35 },
		{ 381, 85, 35 },
		{ 381, 86, 35 },
		{ 381, 87, 35 },
		{ 381, 88, 35 },
		{ 381, 89, 35 },
		{ 381, 90, 35 },
		{ 381, 91, 35 },
		{ 382, 8, 35 },
		{ 382, 9, 35 },
		{ 382, 10, 35 },
		{ 382, 11, 35 },
		{ 382, 12, 35 },
		{ 382, 13, 35 },
		{ 382, 14, 35 },
		{ 382, 15, 35 },
		{ 382, 16, 35 },
		{ 382, 17, 35 },
		{ 382, 18, 35 },
		{ 382, 19, 35 },
		{ 382, 20, 35 },
		{ 382, 21, 35 },
		{ 382, 22, 35 },
		{ 382, 23, 35 },
		{ 382, 24, 35 },
		{ 382, 25, 35 },
		{ 382, 26, 35 },
		{ 382, 27, 35 },
		{ 382, 28, 35 },
		{ 382, 29, 35 },
		{ 382, 30, 35 },
		{ 382, 31, 35 },
		{ 382, 32, 35 },
		{ 382, 33, 35 },
		{ 382, 34, 35 },
		{ 382, 35, 35 },
		{ 382, 36, 35 },
		{ 382, 37, 35 },
		{ 382, 38, 35 },
		{ 382, 39, 35 },
		{ 382, 40, 35 },
		{ 382, 41, 35 },
		{ 382, 42, 35 },
		{ 382, 43, 35 },
		{ 382, 44, 35 },
		{ 382, 45, 35 },
		{ 382, 46, 35 },
		{ 382, 47, 35 },
		{ 382, 48, 35 },
		{ 382, 49, 35 },
		{ 382, 50, 35 },
		{ 382, 51, 35 },
		{ 382, 52, 35 },
		{ 382, 53, 35 },
		{ 382, 54, 35 },
		{ 382, 55, 35 },
		{ 382, 56, 35 },
		{ 382, 57, 35 },
		{ 382, 58, 35 },
		{ 382, 59, 35 },
		{ 382, 60, 35 },
		{ 382, 61, 35 },
		{ 382, 62, 35 },
		{ 382, 63, 35 },
		{ 382, 64, 35 },
		{ 382, 65, 35 },
		{ 382, 66, 35 },
		{ 382, 67, 35 },
		{ 382, 68, 35 },
		{ 382, 69, 35 },
		{ 382, 70, 35 },
		{ 382, 71, 35 },
		{ 382, 72, 35 },
		{ 382, 73, 35 },
		{ 382, 74, 35 },
		{ 382, 75, 35 },
		{ 382, 76, 35 },
		{ 382, 77, 35 },
		{ 382, 78, 35 },
		{ 382, 79, 35 },
		{ 382, 80, 35 },
		{ 382, 81, 35 },
		{ 382, 82, 35 },
		{ 382, 83, 35 },
		{ 382, 84, 35 },
		{ 382, 85, 35 },
		{ 382, 86, 35 },
		{ 382, 87, 35 },
		{ 382, 88, 35 },
		{ 382, 89, 35 },
		{ 382, 90, 35 },
		{ 383, 6, 35 },
		{ 383, 7, 35 },
		{ 383, 8, 35 },
		{ 383, 9, 35 },
		{ 383, 10, 35 },
		{ 383, 11, 35 },
		{ 383, 12, 35 },
		{ 383, 13, 35 },
		{ 383, 14, 35 },
		{ 383, 15, 35 },
		{ 383, 16, 35 },
		{ 383, 17, 35 },
		{ 383, 18, 35 },
		{ 383, 19, 35 },
		{ 383, 20, 35 },
		{ 383, 21, 35 },
		{ 383, 22, 35 },
		{ 383, 23, 35 },
		{ 383, 24, 35 },
		{ 383, 25, 35 },
		{ 383, 26, 35 },
		{ 383, 27, 35 },
		{ 383, 28, 35 },
		{ 383, 29, 35 },
		{ 383, 30, 35 },
		{ 383, 31, 35 },
		{ 383, 32, 35 },
		{ 383, 33, 35 },
		{ 383, 34, 35 },
		{ 383, 35, 35 },
		{ 383, 36, 35 },
		{ 383, 37, 35 },
		{ 383, 38, 35 },
		{ 383, 39, 35 },
		{ 383, 40, 35 },
		{ 383, 41, 35 },
		{ 383, 42, 35 },
		{ 383, 43, 35 },
		{ 383, 44, 35 },
		{ 383, 45, 35 },
		{ 383, 46, 35 },
		{ 383, 47, 35 },
		{ 383, 48, 35 },
		{ 383, 49, 35 },
		{ 383, 50, 35 },
		{ 383, 51, 35 },
		{ 383, 52, 35 },
		{ 383, 53, 35 },
		{ 383, 54, 35 },
		{ 383, 55, 35 },
		{ 383, 56, 35 },
		{ 383, 57, 35 },
		{ 383, 58, 35 },
		{ 383, 59, 35 },
		{ 383, 60, 35 },
		{ 383, 61, 35 },
		{ 383, 62, 35 },
		{ 383, 63, 35 },
		{ 383, 64, 35 },
		{ 383, 65, 35 },
		{ 383, 66, 35 },
		{ 383, 67, 35 },
		{ 383, 68, 35 },
		{ 383, 69, 35 },
		{ 383, 70, 35 },
		{ 383, 71, 35 },
		{ 383, 72, 35 },
		{ 383, 73, 35 },
		{ 383, 74, 35 },
		{ 383, 75, 35 },
		{ 383, 76, 35 },
		{ 383, 77, 35 },
		{ 383, 78, 35 },
		{ 383, 79, 35 },
		{ 383, 80, 35 },
		{ 383, 81, 35 },
		{ 383, 82, 35 },
		{ 383, 83, 35 },
		{ 383, 84, 35 },
		{ 383, 85, 35 },
		{ 383, 86, 35 },
		{ 383, 87, 35 },
		{ 383, 88, 35 },
		{ 384, 4, 35 },
		{ 384, 5, 35 },
		{ 384, 6, 35 },
		{ 384, 7, 35 },
		{ 384, 8, 35 },
		{ 384, 9, 35 },
		{ 384, 10, 35 },
		{ 384, 11, 35 },
		{ 384, 12, 35 },
		{ 384, 13, 35 },
		{ 384, 14, 35 },
		{ 384, 15, 35 },
		{ 384, 16, 35 },
		{ 384, 17, 35 },
		{ 384, 18, 35 },
		{ 384, 19, 35 },
		{ 384, 20, 35 },
		{ 384, 21, 35 },
		{ 384, 22, 35 },
		{ 384, 23, 35 },
		{ 384, 24, 35 },
		{ 384, 25, 35 },
		{ 384, 26, 35 },
		{ 384, 27, 35 },
		{ 384, 28, 35 },
		{ 384, 29, 35 },
		{ 384, 30, 35 },
		{ 384, 31, 35 },
		{ 384, 32, 35 },
		{ 384, 33, 35 },
		{ 384, 34, 35 },
		{ 384, 35, 35 },
		{ 384, 36, 35 },
		{ 384, 37, 35 },
		{ 384, 38, 35 },
		{ 384, 39, 35 },
		{ 384, 40, 35 },
		{ 384, 41, 35 },
		{ 384, 42, 35 },
		{ 384, 43, 35 },
		{ 384, 44, 35 },
		{ 384, 45, 35 },
		{ 384, 46, 35 },
		{ 384, 47, 35 },
		{ 384, 48, 35 },
		{ 384, 49, 35 },
		{ 384, 50, 35 },
		{ 384, 51, 35 },
		{ 384, 52, 35 },
		{ 384, 53, 35 },
		{ 384, 54, 35 },
		{ 384, 55, 35 },
		{ 384, 56, 35 },
		{ 384, 57, 35 },
		{ 384, 58, 35 },
		{ 384, 59, 35 },
		{ 384, 60, 35 },
		{ 384, 61, 35 },
		{ 384, 62, 35 },
		{ 384, 63, 35 },
		{ 384, 64, 35 },
		{ 384, 65, 35 },
		{ 384, 66, 35 },
		{ 384, 67, 35 },
		{ 384, 68, 35 },
		{ 384, 69, 35 },
		{ 384, 70, 35 },
		{ 384, 71, 35 },
		{ 384, 72, 35 },
		{ 384, 73, 35 },
		{ 384, 74, 35 },
		{ 384, 75, 35 },
		{ 384, 76, 35 },
		{ 384, 77, 35 },
		{ 384, 78, 35 },
		{ 384, 79, 35 },
		{ 384, 80, 35 },
		{ 384, 81, 35 },
		{ 384, 82, 35 },
		{ 384, 83, 35 },
		{ 384, 84, 35 },
		{ 384, 85, 35 },
		{ 384, 86, 35 },
		{ 384, 87, 35 },
		{ 385, 2, 35 },
		{ 385, 3, 35 },
		{ 385, 4, 35 },
		{ 385, 5, 35 },
		{ 385, 6, 35 },
		{ 385, 7, 35 },
		{ 385, 8, 35 },
		{ 385, 9, 35 },
		{ 385, 10, 35 },
		{ 385, 11, 35 },
		{ 385, 12, 35 },
		{ 385, 13, 35 },
		{ 385, 14, 35 },
		{ 385, 15, 35 },
		{ 385, 16, 35 },
		{ 385, 17, 35 },
		{ 385, 18, 35 },
		{ 385, 19, 35 },
		{ 385, 20, 35 },
		{ 385, 21, 35 },
		{ 385, 22, 35 },
		{ 385, 23, 35 },
		{ 385, 24, 35 },
		{ 385, 25, 35 },
		{ 385, 26, 35 },
		{ 385, 27, 35 },
		{ 385, 28, 35 },
		{ 385, 29, 35 },
		{ 385, 30, 35 },
		{ 385, 31, 35 },
		{ 385, 32, 35 },
		{ 385, 33, 35 },
		{ 385, 34, 35 },
		{ 385, 35, 35 },
		{ 385, 36, 35 },
		{ 385, 37, 35 },
		{ 385, 38, 35 },
		{ 385, 39, 35 },
		{ 385, 40, 35 },
		{ 385, 41, 35 },
		{ 385, 42, 35 },
		{ 385, 43, 35 },
		{ 385, 44, 35 },
		{ 385, 45, 35 },
		{ 385, 46, 35 },
		{ 385, 47, 35 },
		{ 385, 48, 35 },
		{ 385, 49, 35 },
		{ 385, 50, 35 },
		{ 385, 51, 35 },
		{ 385, 52, 35 },
		{ 385, 53, 35 },
		{ 385, 54, 35 },
		{ 385, 55, 35 },
		{ 385, 56, 35 },
		{ 385, 57, 35 },
		{ 385, 58, 35 },
		{ 385, 59, 35 },
		{ 385, 60, 35 },
		{ 385, 61, 35 },
		{ 385, 62, 35 },
		{ 385, 63, 35 },
		{ 385, 64, 35 },
		{ 385, 65, 35 },
		{ 385, 66, 35 },
		{ 385, 67, 35 },
		{ 385, 68, 35 },
		{ 385, 69, 35 },
		{ 385, 70, 35 },
		{ 385, 71, 35 },
		{ 385, 72, 35 },
		{ 385, 73, 35 },
		{ 385, 74, 35 },
		{ 385, 75, 35 },
		{ 385, 76, 35 },
		{ 385, 77, 35 },
		{ 385, 78, 35 },
		{ 385, 79, 35 },
		{ 385, 80, 35 },
		{ 385, 81, 35 },
		{ 385, 82, 35 },
		{ 385, 83, 35 },
		{ 385, 84, 35 },
		{ 385, 85, 35 },
		{ 385, 86, 35 },
		{ 386, 1, 35 },
		{ 386, 2, 35 },
		{ 386, 3, 35 },
		{ 386, 4, 35 },
		{ 386, 5, 35 },
		{ 386, 6, 35 },
		{ 386, 7, 35 },
		{ 386, 8, 35 },
		{ 386, 9, 35 },
		{ 386, 10, 35 },
		{ 386, 11, 35 },
		{ 386, 12, 35 },
		{ 386, 13, 35 },
		{ 386, 14, 35 },
		{ 386, 15, 35 },
		{ 386, 16, 35 },
		{ 386, 17, 35 },
		{ 386, 18, 35 },
		{ 386, 19, 35 },
		{ 386, 20, 35 },
		{ 386, 21, 35 },
		{ 386, 22, 35 },
		{ 386, 23, 35 },
		{ 386, 24, 35 },
		{ 386, 25, 35 },
		{ 386, 26, 35 },
		{ 386, 27, 35 },
		{ 386, 28, 35 },
		{ 386, 29, 35 },
		{ 386, 30, 35 },
		{ 386, 31, 35 },
		{ 386, 32, 35 },
		{ 386, 33, 35 },
		{ 386, 34, 35 },
		{ 386, 35, 35 },
		{ 386, 36, 35 },
		{ 386, 37, 35 },
		{ 386, 38, 35 },
		{ 386, 39, 35 },
		{ 386, 40, 35 },
		{ 386, 41, 35 },
		{ 386, 42, 35 },
		{ 386, 43, 35 },
		{ 386, 44, 35 },
		{ 386, 45, 35 },
		{ 386, 46, 35 },
		{ 386, 47, 35 },
		{ 386, 48, 35 },
		{ 386, 49, 35 },
		{ 386, 50, 35 },
		{ 386, 51, 35 },
		{ 386, 52, 35 },
		{ 386, 53, 35 },
		{ 386, 54, 35 },
		{ 386, 55, 35 },
		{ 386, 56, 35 },
		{ 386, 57, 35 },
		{ 386, 58, 35 },
		{ 386, 59, 35 },
		{ 386, 60, 35 },
		{ 386, 61, 35 },
		{ 386, 62, 35 },
		{ 386, 63, 35 },
		{ 386, 64, 35 },
		{ 386, 65, 35 },
		{ 386, 66, 35 },
		{ 386, 67, 35 },
		{ 386, 68, 35 },
		{ 386, 69, 35 },
		{ 386, 70, 35 },
		{ 386, 71, 35 },
		{ 386, 72, 35 },
		{ 386, 73, 35 },
		{ 386, 74, 35 },
		{ 386, 75, 35 },
		{ 386, 76, 35 },
		{ 386, 77, 35 },
		{ 386, 78, 35 },
		{ 386, 79, 35 },
		{ 386, 80, 35 },
		{ 386, 81, 35 },
		{ 386, 82, 35 },
		{ 386, 83, 35 },
		{ 386, 84, 35 },
		{ 387, 1, 35 },
		{ 387, 2, 35 },
		{ 387, 3, 35 },
		{ 387, 4, 35 },
		{ 387, 5, 35 },
		{ 387, 6, 35 },
		{ 387, 7, 35 },
		{ 387, 8, 35 },
		{ 387, 9, 35 },
		{ 387, 10, 35 },
		{ 387, 11, 35 },
		{ 387, 12, 35 },
		{ 387, 13, 35 },
		{ 387, 14, 35 },
		{ 387, 15, 35 },
		{ 387, 16, 35 },
		{ 387, 17, 35 },
		{ 387, 18, 35 },
		{ 387, 19, 35 },
		{ 387, 20, 35 },
		{ 387, 21, 35 },
		{ 387, 22, 35 },
		{ 387, 23, 35 },
		{ 387, 24, 35 },
		{ 387, 25, 35 },
		{ 387, 26, 35 },
		{ 387, 27, 35 },
		{ 387, 28, 35 },
		{ 387, 29, 35 },
		{ 387, 30, 35 },
		{ 387, 31, 35 },
		{ 387, 32, 35 },
		{ 387, 33, 35 },
		{ 387, 34, 35 },
		{ 387, 35, 35 },
		{ 387, 36, 35 },
		{ 387, 37, 35 },
		{ 387, 38, 35 },
		{ 387, 39, 35 },
		{ 387, 40, 35 },
		{ 387, 41, 35 },
		{ 387, 42, 35 },
		{ 387, 43, 35 },
		{ 387, 44, 35 },
		{ 387, 45, 35 },
		{ 387, 46, 35 },
		{ 387, 47, 35 },
		{ 387, 48, 35 },
		{ 387, 49, 35 },
		{ 387, 50, 35 },
		{ 387, 51, 35 },
		{ 387, 52, 35 },
		{ 387, 53, 35 },
		{ 387, 54, 35 },
		{ 387, 55, 35 },
		{ 387, 56, 35 },
		{ 387, 57, 35 },
		{ 387, 58, 35 },
		{ 387, 59, 35 },
		{ 387, 60, 35 },
		{ 387, 61, 35 },
		{ 387, 62, 35 },
		{ 387, 63, 35 },
		{ 387, 64, 35 },
		{ 387, 65, 35 },
		{ 387, 66, 35 },
		{ 387, 67, 35 },
		{ 387, 68, 35 },
		{ 387, 69, 35 },
		{ 387, 70, 35 },
		{ 387, 71, 35 },
		{ 387, 72, 35 },
		{ 387, 73, 35 },
		{ 387, 74, 35 },
		{ 387, 75, 35 },
		{ 387, 76, 35 },
		{ 387, 77, 35 },
		{ 387, 78, 35 },
		{ 387, 79, 35 },
		{ 387, 80, 35 },
		{ 387, 81, 35 },
		{ 387, 82, 35 },
		{ 387, 83, 35 },
		{ 388, 1, 35 },
		{ 388, 2, 35 },
		{ 388, 3, 35 },
		{ 388, 4, 35 },
		{ 388, 5, 35 },
		{ 388, 6, 35 },
		{ 388, 7, 35 },
		{ 388, 8, 35 },
		{ 388, 9, 35 },
		{ 388, 10, 35 },
		{ 388, 11, 35 },
		{ 388, 12, 35 },
		{ 388, 13, 35 },
		{ 388, 14, 35 },
		{ 388, 15, 35 },
		{ 388, 16, 35 },
		{ 388, 17, 35 },
		{ 388, 18, 35 },
		{ 388, 19, 35 },
		{ 388, 20, 35 },
		{ 388, 21, 35 },
		{ 388, 22, 35 },
		{ 388, 23, 35 },
		{ 388, 24, 35 },
		{ 388, 25, 35 },
		{ 388, 26, 35 },
		{ 388, 27, 35 },
		{ 388, 28, 35 },
		{ 388, 29, 35 },
		{ 388, 30, 35 },
		{ 388, 31, 35 },
		{ 388, 32, 35 },
		{ 388, 33, 35 },
		{ 388, 34, 35 },
		{ 388, 35, 35 },
		{ 388, 36, 35 },
		{ 388, 37, 35 },
		{ 388, 38, 35 },
		{ 388, 39, 35 },
		{ 388, 40, 35 },
		{ 388, 41, 35 },
		{ 388, 42, 35 },
		{ 388, 43, 35 },
		{ 388, 44, 35 },
		{ 388, 45, 35 },
		{ 388, 46, 35 },
		{ 388, 47, 35 },
		{ 388, 48, 35 },
		{ 388, 49, 35 },
		{ 388, 50, 35 },
		{ 388, 51, 35 },
		{ 388, 52, 35 },
		{ 388, 53, 35 },
		{ 388, 54, 35 },
		{ 388, 55, 35 },
		{ 388, 56, 35 },
		{ 388, 57, 35 },
		{ 388, 58, 35 },
		{ 388, 59, 35 },
		{ 388, 60, 35 },
		{ 388, 61, 35 },
		{ 388, 62, 35 },
		{ 388, 63, 35 },
		{ 388, 64, 35 },
		{ 388, 65, 35 },
		{ 388, 66, 35 },
		{ 388, 67, 35 },
		{ 388, 68, 35 },
		{ 388, 69, 35 },
		{ 388, 70, 35 },
		{ 388, 71, 35 },
		{ 388, 72, 35 },
		{ 388, 73, 35 },
		{ 388, 74, 35 },
		{ 388, 75, 35 },
		{ 388, 76, 35 },
		{ 388, 77, 35 },
		{ 388, 78, 35 },
		{ 388, 79, 35 },
		{ 388, 80, 35 },
		{ 388, 81, 35 },
		{ 388, 82, 35 },
		{ 389, 1, 35 },
		{ 389, 2, 35 },
		{ 389, 3, 35 },
		{ 389, 4, 35 },
		{ 389, 5, 35 },
		{ 389, 6, 35 },
		{ 389, 7, 35 },
		{ 389, 8, 35 },
		{ 389, 9, 35 },
		{ 389, 10, 35 },
		{ 389, 11, 35 },
		{ 389, 12, 35 },
		{ 389, 13, 35 },
		{ 389, 14, 35 },
		{ 389, 15, 35 },
		{ 389, 16, 35 },
		{ 389, 17, 35 },
		{ 389, 18, 35 },
		{ 389, 19, 35 },
		{ 389, 20, 35 },
		{ 389, 21, 35 },
		{ 389, 22, 35 },
		{ 389, 23, 35 },
		{ 389, 24, 35 },
		{ 389, 25, 35 },
		{ 389, 26, 35 },
		{ 389, 27, 35 },
		{ 389, 28, 35 },
		{ 389, 29, 35 },
		{ 389, 30, 35 },
		{ 389, 31, 35 },
		{ 389, 32, 35 },
		{ 389, 33, 35 },
		{ 389, 34, 35 },
		{ 389, 35, 35 },
		{ 389, 36, 35 },
		{ 389, 37, 35 },
		{ 389, 38, 35 },
		{ 389, 39, 35 },
		{ 389, 40, 35 },
		{ 389, 41, 35 },
		{ 389, 42, 35 },
		{ 389, 43, 35 },
		{ 389, 44, 35 },
		{ 389, 45, 35 },
		{ 389, 46, 35 },
		{ 389, 47, 35 },
		{ 389, 48, 35 },
		{ 389, 49, 35 },
		{ 389, 50, 35 },
		{ 389, 51, 35 },
		{ 389, 52, 35 },
		{ 389, 53, 35 },
		{ 389, 54, 35 },
		{ 389, 55, 35 },
		{ 389, 56, 35 },
		{ 389, 57, 35 },
		{ 389, 58, 35 },
		{ 389, 59, 35 },
		{ 389, 60, 35 },
		{ 389, 61, 35 },
		{ 389, 62, 35 },
		{ 389, 63, 35 },
		{ 389, 64, 35 },
		{ 389, 65, 35 },
		{ 389, 66, 35 },
		{ 389, 67, 35 },
		{ 389, 68, 35 },
		{ 389, 69, 35 },
		{ 389, 70, 35 },
		{ 389, 71, 35 },
		{ 389, 72, 35 },
		{ 389, 73, 35 },
		{ 389, 74, 35 },
		{ 389, 75, 35 },
		{ 389, 76, 35 },
		{ 389, 77, 35 },
		{ 389, 78, 35 },
		{ 389, 79, 35 },
		{ 389, 80, 35 },
		{ 390, 1, 35 },
		{ 390, 2, 35 },
		{ 390, 3, 35 },
		{ 390, 4, 35 },
		{ 390, 5, 35 },
		{ 390, 6, 35 },
		{ 390, 7, 35 },
		{ 390, 8, 35 },
		{ 390, 9, 35 },
		{ 390, 10, 35 },
		{ 390, 11, 35 },
		{ 390, 12, 35 },
		{ 390, 13, 35 },
		{ 390, 14, 35 },
		{ 390, 15, 35 },
		{ 390, 16, 35 },
		{ 390, 17, 35 },
		{ 390, 18, 35 },
		{ 390, 19, 35 },
		{ 390, 20, 35 },
		{ 390, 21, 35 },
		{ 390, 22, 35 },
		{ 390, 23, 35 },
		{ 390, 24, 35 },
		{ 390, 25, 35 },
		{ 390, 26, 35 },
		{ 390, 27, 35 },
		{ 390, 28, 35 },
		{ 390, 29, 35 },
		{ 390, 30, 35 },
		{ 390, 31, 35 },
		{ 390, 32, 35 },
		{ 390, 33, 35 },
		{ 390, 34, 35 },
		{ 390, 35, 35 },
		{ 390, 36, 35 },
		{ 390, 37, 35 },
		{ 390, 38, 35 },
		{ 390, 39, 35 },
		{ 390, 40, 35 },
		{ 390, 41, 35 },
		{ 390, 42, 35 },
		{ 390, 43, 35 },
		{ 390, 44, 35 },
		{ 390, 45, 35 },
		{ 390, 46, 35 },
		{ 390, 47, 35 },
		{ 390, 48, 35 },
		{ 390, 49, 35 },
		{ 390, 50, 35 },
		{ 390, 51, 35 },
		{ 390, 52, 35 },
		{ 390, 53, 35 },
		{ 390, 54, 35 },
		{ 390, 55, 35 },
		{ 390, 56, 35 },
		{ 390, 57, 35 },
		{ 390, 58, 35 },
		{ 390, 59, 35 },
		{ 390, 60, 35 },
		{ 390, 61, 35 },
		{ 390, 62, 35 },
		{ 390, 63, 35 },
		{ 390, 64, 35 },
		{ 390, 65, 35 },
		{ 390, 66, 35 },
		{ 390, 67, 35 },
		{ 390, 68, 35 },
		{ 390, 69, 35 },
		{ 390, 70, 35 },
		{ 390, 71, 35 },
		{ 390, 72, 35 },
		{ 390, 73, 35 },
		{ 390, 74, 35 },
		{ 390, 75, 35 },
		{ 390, 76, 35 },
		{ 390, 77, 35 },
		{ 390, 78, 35 },
		{ 390, 79, 35 },
		{ 391, 1, 35 },
		{ 391, 2, 35 },
		{ 391, 3, 35 },
		{ 391, 4, 35 },
		{ 391, 5, 35 },
		{ 391, 6, 35 },
		{ 391, 7, 35 },
		{ 391, 8, 35 },
		{ 391, 9, 35 },
		{ 391, 10, 35 },
		{ 391, 11, 35 },
		{ 391, 12, 35 },
		{ 391, 13, 35 },
		{ 391, 14, 35 },
		{ 391, 15, 35 },
		{ 391, 16, 35 },
		{ 391, 17, 35 },
		{ 391, 18, 35 },
		{ 391, 19, 35 },
		{ 391, 20, 35 },
		{ 391, 21, 35 },
		{ 391, 22, 35 },
		{ 391, 23, 35 },
		{ 391, 24, 35 },
		{ 391, 25, 35 },
		{ 391, 26, 35 },
		{ 391, 27, 35 },
		{ 391, 28, 35 },
		{ 391, 29, 35 },
		{ 391, 30, 35 },
		{ 391, 31, 35 },
		{ 391, 32, 35 },
		{ 391, 33, 35 },
		{ 391, 34, 35 },
		{ 391, 35, 35 },
		{ 391, 36, 35 },
		{ 391, 37, 35 },
		{ 391, 38, 35 },
		{ 391, 39, 35 },
		{ 391, 40, 35 },
		{ 391, 41, 35 },
		{ 391, 42, 35 },
		{ 391, 43, 35 },
		{ 391, 44, 35 },
		{ 391, 45, 35 },
		{ 391, 46, 35 },
		{ 391, 47, 35 },
		{ 391, 48, 35 },
		{ 391, 49, 35 },
		{ 391, 50, 35 },
		{ 391, 51, 35 },
		{ 391, 52, 35 },
		{ 391, 53, 35 },
		{ 391, 54, 35 },
		{ 391, 55, 35 },
		{ 391, 56, 35 },
		{ 391, 57, 35 },
		{ 391, 58, 35 },
		{ 391, 59, 35 },
		{ 391, 60, 35 },
		{ 391, 61, 35 },
		{ 391, 62, 35 },
		{ 391, 63, 35 },
		{ 391, 64, 35 },
		{ 391, 65, 35 },
		{ 391, 66, 35 },
		{ 391, 67, 35 },
		{ 391, 68, 35 },
		{ 391, 69, 35 },
		{ 391, 70, 35 },
		{ 391, 71, 35 },
		{ 391, 72, 35 },
		{ 391, 73, 35 },
		{ 391, 74, 35 },
		{ 391, 75, 35 },
		{ 391, 76, 35 },
		{ 391, 77, 35 },
		{ 391, 78, 35 },
		{ 392, 1, 35 },
		{ 392, 2, 35 },
		{ 392, 3, 35 },
		{ 392, 4, 35 },
		{ 392, 5, 35 },
		{ 392, 6, 35 },
		{ 392, 7, 35 },
		{ 392, 8, 35 },
		{ 392, 9, 35 },
		{ 392, 10, 35 },
		{ 392, 11, 35 },
		{ 392, 12, 35 },
		{ 392, 13, 35 },
		{ 392, 14, 35 },
		{ 392, 15, 35 },
		{ 392, 16, 35 },
		{ 392, 17, 35 },
		{ 392, 18, 35 },
		{ 392, 19, 35 },
		{ 392, 20, 35 },
		{ 392, 21, 35 },
		{ 392, 22, 35 },
		{ 392, 23, 35 },
		{ 392, 24, 35 },
		{ 392, 25, 35 },
		{ 392, 26, 35 },
		{ 392, 27, 35 },
		{ 392, 28, 35 },
		{ 392, 29, 35 },
		{ 392, 30, 35 },
		{ 392, 31, 35 },
		{ 392, 32, 35 },
		{ 392, 33, 35 },
		{ 392, 34, 35 },
		{ 392, 35, 35 },
		{ 392, 36, 35 },
		{ 392, 37, 35 },
		{ 392, 38, 35 },
		{ 392, 39, 35 },
		{ 392, 40, 35 },
		{ 392, 41, 35 },
		{ 392, 42, 35 },
		{ 392, 43, 35 },
		{ 392, 44, 35 },
		{ 392, 45, 35 },
		{ 392, 46, 35 },
		{ 392, 47, 35 },
		{ 392, 48, 35 },
		{ 392, 49, 35 },
		{ 392, 50, 35 },
		{ 392, 51, 35 },
		{ 392, 52, 35 },
		{ 392, 53, 35 },
		{ 392, 54, 35 },
		{ 392, 55, 35 },
		{ 392, 56, 35 },
		{ 392, 57, 35 },
		{ 392, 58, 35 },
		{ 392, 59, 35 },
		{ 392, 60, 35 },
		{ 392, 61, 35 },
		{ 392, 62, 35 },
		{ 392, 63, 35 },
		{ 392, 64, 35 },
		{ 392, 65, 35 },
		{ 392, 66, 35 },
		{ 392, 67, 35 },
		{ 392, 68, 35 },
		{ 392, 69, 35 },
		{ 392, 70, 35 },
		{ 392, 71, 35 },
		{ 392, 72, 35 },
		{ 392, 73, 35 },
		{ 392, 74, 35 },
		{ 392, 75, 35 },
		{ 392, 76, 35 },
		{ 393, 1, 35 },
		{ 393, 2, 35 },
		{ 393, 3, 35 },
		{ 393, 4, 35 },
		{ 393, 5, 35 },
		{ 393, 6, 35 },
		{ 393, 7, 35 },
		{ 393, 8, 35 },
		{ 393, 9, 35 },
		{ 393, 10, 35 },
		{ 393, 11, 35 },
		{ 393, 12, 35 },
		{ 393, 13, 35 },
		{ 393, 14, 35 },
		{ 393, 15, 35 },
		{ 393, 16, 35 },
		{ 393, 17, 35 },
		{ 393, 18, 35 },
		{ 393, 19, 35 },
		{ 393, 20, 35 },
		{ 393, 21, 35 },
		{ 393, 22, 35 },
		{ 393, 23, 35 },
		{ 393, 24, 35 },
		{ 393, 25, 35 },
		{ 393, 26, 35 },
		{ 393, 27, 35 },
		{ 393, 28, 35 },
		{ 393, 29, 35 },
		{ 393, 30, 35 },
		{ 393, 31, 35 },
		{ 393, 32, 35 },
		{ 393, 33, 35 },
		{ 393, 34, 35 },
		{ 393, 35, 35 },
		{ 393, 36, 35 },
		{ 393, 37, 35 },
		{ 393, 38, 35 },
		{ 393, 39, 35 },
		{ 393, 40, 35 },
		{ 393, 41, 35 },
		{ 393, 42, 35 },
		{ 393, 43, 35 },
		{ 393, 44, 35 },
		{ 393, 45, 35 },
		{ 393, 46, 35 },
		{ 393, 47, 35 },
		{ 393, 48, 35 },
		{ 393, 49, 35 },
		{ 393, 50, 35 },
		{ 393, 51, 35 },
		{ 393, 52, 35 },
		{ 393, 53, 35 },
		{ 393, 54, 35 },
		{ 393, 55, 35 },
		{ 393, 56, 35 },
		{ 393, 57, 35 },
		{ 393, 58, 35 },
		{ 393, 59, 35 },
		{ 393, 60, 35 },
		{ 393, 61, 35 },
		{ 393, 62, 35 },
		{ 393, 63, 35 },
		{ 393, 64, 35 },
		{ 393, 65, 35 },
		{ 393, 66, 35 },
		{ 393, 67, 35 },
		{ 393, 68, 35 },
		{ 393, 69, 35 },
		{ 393, 70, 35 },
		{ 393, 71, 35 },
		{ 393, 72, 35 },
		{ 393, 73, 35 },
		{ 393, 74, 35 },
		{ 393, 75, 35 },
		{ 394, 1, 35 },
		{ 394, 2, 35 },
		{ 394, 3, 35 },
		{ 394, 4, 35 },
		{ 394, 5, 35 },
		{ 394, 6, 35 },
		{ 394, 7, 35 },
		{ 394, 8, 35 },
		{ 394, 9, 35 },
		{ 394, 10, 35 },
		{ 394, 11, 35 },
		{ 394, 12, 35 },
		{ 394, 13, 35 },
		{ 394, 14, 35 },
		{ 394, 15, 35 },
		{ 394, 16, 35 },
		{ 394, 17, 35 },
		{ 394, 18, 35 },
		{ 394, 19, 35 },
		{ 394, 20, 35 },
		{ 394, 21, 35 },
		{ 394, 22, 35 },
		{ 394, 23, 35 },
		{ 394, 24, 35 },
		{ 394, 25, 35 },
		{ 394, 26, 35 },
		{ 394, 27, 35 },
		{ 394, 28, 35 },
		{ 394, 29, 35 },
		{ 394, 30, 35 },
		{ 394, 31, 35 },
		{ 394, 32, 35 },
		{ 394, 33, 35 },
		{ 394, 34, 35 },
		{ 394, 35, 35 },
		{ 394, 36, 35 },
		{ 394, 37, 35 },
		{ 394, 38, 35 },
		{ 394, 39, 35 },
		{ 394, 40, 35 },
		{ 394, 41, 35 },
		{ 394, 42, 35 },
		{ 394, 43, 35 },
		{ 394, 44, 35 },
		{ 394, 45, 35 },
		{ 394, 46, 35 },
		{ 394, 47, 35 },
		{ 394, 48, 35 },
		{ 394, 49, 35 },
		{ 394, 50, 35 },
		{ 394, 51, 35 },
		{ 394, 52, 35 },
		{ 394, 53, 35 },
		{ 394, 54, 35 },
		{ 394, 55, 35 },
		{ 394, 56, 35 },
		{ 394, 57, 35 },
		{ 394, 58, 35 },
		{ 394, 59, 35 },
		{ 394, 60, 35 },
		{ 394, 61, 35 },
		{ 394, 62, 35 },
		{ 394, 63, 35 },
		{ 394, 64, 35 },
		{ 394, 65, 35 },
		{ 394, 66, 35 },
		{ 394, 67, 35 },
		{ 394, 68, 35 },
		{ 394, 69, 35 },
		{ 394, 70, 35 },
		{ 394, 71, 35 },
		{ 394, 72, 35 },
		{ 394, 73, 35 },
		{ 394, 74, 35 },
		{ 395, 1, 35 },
		{ 395, 2, 35 },
		{ 395, 3, 35 },
		{ 395, 4, 35 },
		{ 395, 5, 35 },
		{ 395, 6, 35 },
		{ 395, 7, 35 },
		{ 395, 8, 35 },
		{ 395, 9, 35 },
		{ 395, 10, 35 },
		{ 395, 11, 35 },
		{ 395, 12, 35 },
		{ 395, 13, 35 },
		{ 395, 14, 35 },
		{ 395, 15, 35 },
		{ 395, 16, 35 },
		{ 395, 17, 35 },
		{ 395, 18, 35 },
		{ 395, 19, 35 },
		{ 395, 20, 35 },
		{ 395, 21, 35 },
		{ 395, 22, 35 },
		{ 395, 23, 35 },
		{ 395, 24, 35 },
		{ 395, 25, 35 },
		{ 395, 26, 35 },
		{ 395, 27, 35 },
		{ 395, 28, 35 },
		{ 395, 29, 35 },
		{ 395, 30, 35 },
		{ 395, 31, 35 },
		{ 395, 32, 35 },
		{ 395, 33, 35 },
		{ 395, 34, 35 },
		{ 395, 35, 35 },
		{ 395, 36, 35 },
		{ 395, 37, 35 },
		{ 395, 38, 35 },
		{ 395, 39, 35 },
		{ 395, 40, 35 },
		{ 395, 41, 35 },
		{ 395, 42, 35 },
		{ 395, 43, 35 },
		{ 395, 44, 35 },
		{ 395, 45, 35 },
		{ 395, 46, 35 },
		{ 395, 47, 35 },
		{ 395, 48, 35 },
		{ 395, 49, 35 },
		{ 395, 50, 35 },
		{ 395, 51, 35 },
		{ 395, 52, 35 },
		{ 395, 53, 35 },
		{ 395, 54, 35 },
		{ 395, 55, 35 },
		{ 395, 56, 35 },
		{ 395, 57, 35 },
		{ 395, 58, 35 },
		{ 395, 59, 35 },
		{ 395, 60, 35 },
		{ 395, 61, 35 },
		{ 395, 62, 35 },
		{ 395, 63, 35 },
		{ 395, 64, 35 },
		{ 395, 65, 35 },
		{ 395, 66, 35 },
		{ 395, 67, 35 },
		{ 395, 68, 35 },
		{ 395, 69, 35 },
		{ 395, 70, 35 },
		{ 395, 71, 35 },
		{ 395, 72, 35 },
		{ 396, 1, 35 },
		{ 396, 2, 35 },
		{ 396, 3, 35 },
		{ 396, 4, 35 },
		{ 396, 5, 35 },
		{ 396, 6, 35 },
		{ 396, 7, 35 },
		{ 396, 8, 35 },
		{ 396, 9, 35 },
		{ 396, 10, 35 },
		{ 396, 11, 35 },
		{ 396, 12, 35 },
		{ 396, 13, 35 },
		{ 396, 14, 35 },
		{ 396, 15, 35 },
		{ 396, 16, 35 },
		{ 396, 17, 35 },
		{ 396, 18, 35 },
		{ 396, 19, 35 },
		{ 396, 20, 35 },
		{ 396, 21, 35 },
		{ 396, 22, 35 },
		{ 396, 23, 35 },
		{ 396, 24, 35 },
		{ 396, 25, 35 },
		{ 396, 26, 35 },
		{ 396, 27, 35 },
		{ 396, 28, 35 },
		{ 396, 29, 35 },
		{ 396, 30, 35 },
		{ 396, 31, 35 },
		{ 396, 32, 35 },
		{ 396, 33, 35 },
		{ 396, 34, 35 },
		{ 396, 35, 35 },
		{ 396, 36, 35 },
		{ 396, 37, 35 },
		{ 396, 38, 35 },
		{ 396, 39, 35 },
		{ 396, 40, 35 },
		{ 396, 41, 35 },
		{ 396, 42, 35 },
		{ 396, 43, 35 },
		{ 396, 44, 35 },
		{ 396, 45, 35 },
		{ 396, 46, 35 },
		{ 396, 47, 35 },
		{ 396, 48, 35 },
		{ 396, 49, 35 },
		{ 396, 50, 35 },
		{ 396, 51, 35 },
		{ 396, 52, 35 },
		{ 396, 53, 35 },
		{ 396, 54, 35 },
		{ 396, 55, 35 },
		{ 396, 56, 35 },
		{ 396, 57, 35 },
		{ 396, 58, 35 },
		{ 396, 59, 35 },
		{ 396, 60, 35 },
		{ 396, 61, 35 },
		{ 396, 62, 35 },
		{ 396, 63, 35 },
		{ 396, 64, 35 },
		{ 396, 65, 35 },
		{ 396, 66, 35 },
		{ 396, 67, 35 },
		{ 396, 68, 35 },
		{ 396, 69, 35 },
		{ 396, 70, 35 },
		{ 396, 71, 35 },
		{ 397, 1, 35 },
		{ 397, 2, 35 },
		{ 397, 3, 35 },
		{ 397, 4, 35 },
		{ 397, 5, 35 },
		{ 397, 6, 35 },
		{ 397, 7, 35 },
		{ 397, 8, 35 },
		{ 397, 9, 35 },
		{ 397, 10, 35 },
		{ 397, 11, 35 },
		{ 397, 12, 35 },
		{ 397, 13, 35 },
		{ 397, 14, 35 },
		{ 397, 15, 35 },
		{ 397, 16, 35 },
		{ 397, 17, 35 },
		{ 397, 18, 35 },
		{ 397, 19, 35 },
		{ 397, 20, 35 },
		{ 397, 21, 35 },
		{ 397, 22, 35 },
		{ 397, 23, 35 },
		{ 397, 24, 35 },
		{ 397, 25, 35 },
		{ 397, 26, 35 },
		{ 397, 27, 35 },
		{ 397, 28, 35 },
		{ 397, 29, 35 },
		{ 397, 30, 35 },
		{ 397, 31, 35 },
		{ 397, 32, 35 },
		{ 397, 33, 35 },
		{ 397, 34, 35 },
		{ 397, 35, 35 },
		{ 397, 36, 35 },
		{ 397, 37, 35 },
		{ 397, 38, 35 },
		{ 397, 39, 35 },
		{ 397, 40, 35 },
		{ 397, 41, 35 },
		{ 397, 42, 35 },
		{ 397, 43, 35 },
		{ 397, 44, 35 },
		{ 397, 45, 35 },
		{ 397, 46, 35 },
		{ 397, 47, 35 },
		{ 397, 48, 35 },
		{ 397, 49, 35 },
		{ 397, 50, 35 },
		{ 397, 51, 35 },
		{ 397, 52, 35 },
		{ 397, 53, 35 },
		{ 397, 54, 35 },
		{ 397, 55, 35 },
		{ 397, 56, 35 },
		{ 397, 57, 35 },
		{ 397, 58, 35 },
		{ 397, 59, 35 },
		{ 397, 60, 35 },
		{ 397, 61, 35 },
		{ 397, 62, 35 },
		{ 397, 63, 35 },
		{ 397, 64, 35 },
		{ 397, 65, 35 },
		{ 397, 66, 35 },
		{ 397, 67, 35 },
		{ 397, 68, 35 },
		{ 397, 69, 35 },
		{ 397, 70, 35 },
		{ 398, 1, 35 },
		{ 398, 2, 35 },
		{ 398, 3, 35 },
		{ 398, 4, 35 },
		{ 398, 5, 35 },
		{ 398, 6, 35 },
		{ 398, 7, 35 },
		{ 398, 8, 35 },
		{ 398, 9, 35 },
		{ 398, 10, 35 },
		{ 398, 11, 35 },
		{ 398, 12, 35 },
		{ 398, 13, 35 },
		{ 398, 14, 35 },
		{ 398, 15, 35 },
		{ 398, 16, 35 },
		{ 398, 17, 35 },
		{ 398, 18, 35 },
		{ 398, 19, 35 },
		{ 398, 20, 35 },
		{ 398, 21, 35 },
		{ 398, 22, 35 },
		{ 398, 23, 35 },
		{ 398, 24, 35 },
		{ 398, 25, 35 },
		{ 398, 26, 35 },
		{ 398, 27, 35 },
		{ 398, 28, 35 },
		{ 398, 29, 35 },
		{ 398, 30, 35 },
		{ 398, 31, 35 },
		{ 398, 32, 35 },
		{ 398, 33, 35 },
		{ 398, 34, 35 },
		{ 398, 35, 35 },
		{ 398, 36, 35 },
		{ 398, 37, 35 },
		{ 398, 38, 35 },
		{ 398, 39, 35 },
		{ 398, 40, 35 },
		{ 398, 41, 35 },
		{ 398, 42, 35 },
		{ 398, 43, 35 },
		{ 398, 44, 35 },
		{ 398, 45, 35 },
		{ 398, 46, 35 },
		{ 398, 47, 35 },
		{ 398, 48, 35 },
		{ 398, 49, 35 },
		{ 398, 50, 35 },
		{ 398, 51, 35 },
		{ 398, 52, 35 },
		{ 398, 53, 35 },
		{ 398, 54, 35 },
		{ 398, 55, 35 },
		{ 398, 56, 35 },
		{ 398, 57, 35 },
		{ 398, 58, 35 },
		{ 398, 59, 35 },
		{ 398, 60, 35 },
		{ 398, 61, 35 },
		{ 398, 62, 35 },
		{ 398, 63, 35 },
		{ 398, 64, 35 },
		{ 398, 65, 35 },
		{ 398, 66, 35 },
		{ 398, 67, 35 },
		{ 398, 68, 35 },
		{ 399, 1, 35 },
		{ 399, 2, 35 },
		{ 399, 3, 35 },
		{ 399, 4, 35 },
		{ 399, 5, 35 },
		{ 399, 6, 35 },
		{ 399, 7, 35 },
		{ 399, 8, 35 },
		{ 399, 9, 35 },
		{ 399, 10, 35 },
		{ 399, 11, 35 },
		{ 399, 12, 35 },
		{ 399, 13, 35 },
		{ 399, 14, 35 },
		{ 399, 15, 35 },
		{ 399, 16, 35 },
		{ 399, 17, 35 },
		{ 399, 18, 35 },
		{ 399, 19, 35 },
		{ 399, 20, 35 },
		{ 399, 21, 35 },
		{ 399, 22, 35 },
		{ 399, 23, 35 },
		{ 399, 24, 35 },
		{ 399, 25, 35 },
		{ 399, 26, 35 },
		{ 399, 27, 35 },
		{ 399, 28, 35 },
		{ 399, 29, 35 },
		{ 399, 30, 35 },
		{ 399, 31, 35 },
		{ 399, 32, 35 },
		{ 399, 33, 35 },
		{ 399, 34, 35 },
		{ 399, 35, 35 },
		{ 399, 36, 35 },
		{ 399, 37, 35 },
		{ 399, 38, 35 },
		{ 399, 39, 35 },
		{ 399, 40, 35 },
		{ 399, 41, 35 },
		{ 399, 42, 35 },
		{ 399, 43, 35 },
		{ 399, 44, 35 },
		{ 399, 45, 35 },
		{ 399, 46, 35 },
		{ 399, 47, 35 },
		{ 399, 48, 35 },
		{ 399, 49, 35 },
		{ 399, 50, 35 },
		{ 399, 51, 35 },
		{ 399, 52, 35 },
		{ 399, 53, 35 },
		{ 399, 54, 35 },
		{ 399, 55, 35 },
		{ 399, 56, 35 },
		{ 399, 57, 35 },
		{ 399, 58, 35 },
		{ 399, 59, 35 },
		{ 399, 60, 35 },
		{ 399, 61, 35 },
		{ 399, 62, 35 },
		{ 399, 63, 35 },
		{ 399, 64, 35 },
		{ 399, 65, 35 },
		{ 399, 66, 35 },
		{ 399, 67, 35 },
		{ 400, 1, 35 },
		{ 400, 2, 35 },
		{ 400, 3, 35 },
		{ 400, 4, 35 },
		{ 400, 5, 35 },
		{ 400, 6, 35 },
		{ 400, 7, 35 },
		{ 400, 8, 35 },
		{ 400, 9, 35 },
		{ 400, 10, 35 },
		{ 400, 11, 35 },
		{ 400, 12, 35 },
		{ 400, 13, 35 },
		{ 400, 14, 35 },
		{ 400, 15, 35 },
		{ 400, 16, 35 },
		{ 400, 17, 35 },
		{ 400, 18, 35 },
		{ 400, 19, 35 },
		{ 400, 20, 35 },
		{ 400, 21, 35 },
		{ 400, 22, 35 },
		{ 400, 23, 35 },
		{ 400, 24, 35 },
		{ 400, 25, 35 },
		{ 400, 26, 35 },
		{ 400, 27, 35 },
		{ 400, 28, 35 },
		{ 400, 29, 35 },
		{ 400, 30, 35 },
		{ 400, 31, 35 },
		{ 400, 32, 35 },
		{ 400, 33, 35 },
		{ 400, 34, 35 },
		{ 400, 35, 35 },
		{ 400, 36, 35 },
		{ 400, 37, 35 },
		{ 400, 38, 35 },
		{ 400, 39, 35 },
		{ 400, 40, 35 },
		{ 400, 41, 35 },
		{ 400, 42, 35 },
		{ 400, 43, 35 },
		{ 400, 44, 35 },
		{ 400, 45, 35 },
		{ 400, 46, 35 },
		{ 400, 47, 35 },
		{ 400, 48, 35 },
		{ 400, 49, 35 },
		{ 400, 50, 35 },
		{ 400, 51, 35 },
		{ 400, 52, 35 },
		{ 400, 53, 35 },
		{ 400, 54, 35 },
		{ 400, 55, 35 },
		{ 400, 56, 35 },
		{ 400, 57, 35 },
		{ 400, 58, 35 },
		{ 400, 59, 35 },
		{ 400, 60, 35 },
		{ 400, 61, 35 },
		{ 400, 62, 35 },
		{ 400, 63, 35 },
		{ 400, 64, 35 },
		{ 400, 65, 35 },
		{ 400, 66, 35 },
		{ 401, 1, 35 },
		{ 401, 2, 35 },
		{ 401, 3, 35 },
		{ 401, 4, 35 },
		{ 401, 5, 35 },
		{ 401, 6, 35 },
		{ 401, 7, 35 },
		{ 401, 8, 35 },
		{ 401, 9, 35 },
		{ 401, 10, 35 },
		{ 401, 11, 35 },
		{ 401, 12, 35 },
		{ 401, 13, 35 },
		{ 401, 14, 35 },
		{ 401, 15, 35 },
		{ 401, 16, 35 },
		{ 401, 17, 35 },
		{ 401, 18, 35 },
		{ 401, 19, 35 },
		{ 401, 20, 35 },
		{ 401, 21, 35 },
		{ 401, 22, 35 },
		{ 401, 23, 35 },
		{ 401, 24, 35 },
		{ 401, 25, 35 },
		{ 401, 26, 35 },
		{ 401, 27, 35 },
		{ 401, 28, 35 },
		{ 401, 29, 35 },
		{ 401, 30, 35 },
		{ 401, 31, 35 },
		{ 401, 32, 35 },
		{ 401, 33, 35 },
		{ 401, 34, 35 },
		{ 401, 35, 35 },
		{ 401, 36, 35 },
		{ 401, 37, 35 },
		{ 401, 38, 35 },
		{ 401, 39, 35 },
		{ 401, 40, 35 },
		{ 401, 41, 35 },
		{ 401, 42, 35 },
		{ 401, 43, 35 },
		{ 401, 44, 35 },
		{ 401, 45, 35 },
		{ 401, 46, 35 },
		{ 401, 47, 35 },
		{ 401, 48, 35 },
		{ 401, 49, 35 },
		{ 401, 50, 35 },
		{ 401, 51, 35 },
		{ 401, 52, 35 },
		{ 401, 53, 35 },
		{ 401, 54, 35 },
		{ 401, 55, 35 },
		{ 401, 56, 35 },
		{ 401, 57, 35 },
		{ 401, 58, 35 },
		{ 401, 59, 35 },
		{ 401, 60, 35 },
		{ 401, 61, 35 },
		{ 401, 62, 35 },
		{ 401, 63, 35 },
		{ 401, 64, 35 },
		{ 402, 1, 35 },
		{ 402, 2, 35 },
		{ 402, 3, 35 },
		{ 402, 4, 35 },
		{ 402, 5, 35 },
		{ 402, 6, 35 },
		{ 402, 7, 35 },
		{ 402, 8, 35 },
		{ 402, 9, 35 },
		{ 402, 10, 35 },
		{ 402, 11, 35 },
		{ 402, 12, 35 },
		{ 402, 13, 35 },
		{ 402, 14, 35 },
		{ 402, 15, 35 },
		{ 402, 16, 35 },
		{ 402, 17, 35 },
		{ 402, 18, 35 },
		{ 402, 19, 35 },
		{ 402, 20, 35 },
		{ 402, 21, 35 },
		{ 402, 22, 35 },
		{ 402, 23, 35 },
		{ 402, 24, 35 },
		{ 402, 25, 35 },
		{ 402, 26, 35 },
		{ 402, 27, 35 },
		{ 402, 28, 35 },
		{ 402, 29, 35 },
		{ 402, 30, 35 },
		{ 402, 31, 35 },
		{ 402, 32, 35 },
		{ 402, 33, 35 },
		{ 402, 34, 35 },
		{ 402, 35, 35 },
		{ 402, 36, 35 },
		{ 402, 37, 35 },
		{ 402, 38, 35 },
		{ 402, 39, 35 },
		{ 402, 40, 35 },
		{ 402, 41, 35 },
		{ 402, 42, 35 },
		{ 402, 43, 35 },
		{ 402, 44, 35 },
		{ 402, 45, 35 },
		{ 402, 46, 35 },
		{ 402, 47, 35 },
		{ 402, 48, 35 },
		{ 402, 49, 35 },
		{ 402, 50, 35 },
		{ 402, 51, 35 },
		{ 402, 52, 35 },
		{ 402, 53, 35 },
		{ 402, 54, 35 },
		{ 402, 55, 35 },
		{ 402, 56, 35 },
		{ 402, 57, 35 },
		{ 402, 58, 35 },
		{ 402, 59, 35 },
		{ 402, 60, 35 },
		{ 402, 61, 35 },
		{ 402, 62, 35 },
		{ 402, 63, 35 },
		{ 403, 1, 35 },
		{ 403, 2, 35 },
		{ 403, 3, 35 },
		{ 403, 4, 35 },
		{ 403, 5, 35 },
		{ 403, 6, 35 },
		{ 403, 7, 35 },
		{ 403, 8, 35 },
		{ 403, 9, 35 },
		{ 403, 10, 35 },
		{ 403, 11, 35 },
		{ 403, 12, 35 },
		{ 403, 13, 35 },
		{ 403, 14, 35 },
		{ 403, 15, 35 },
		{ 403, 16, 35 },
		{ 403, 17, 35 },
		{ 403, 18, 35 },
		{ 403, 19, 35 },
		{ 403, 20, 35 },
		{ 403, 21, 35 },
		{ 403, 22, 35 },
		{ 403, 23, 35 },
		{ 403, 24, 35 },
		{ 403, 25, 35 },
		{ 403, 26, 35 },
		{ 403, 27, 35 },
		{ 403, 28, 35 },
		{ 403, 29, 35 },
		{ 403, 30, 35 },
		{ 403, 31, 35 },
		{ 403, 32, 35 },
		{ 403, 33, 35 },
		{ 403, 34, 35 },
		{ 403, 35, 35 },
		{ 403, 36, 35 },
		{ 403, 37, 35 },
		{ 403, 38, 35 },
		{ 403, 39, 35 },
		{ 403, 40, 35 },
		{ 403, 41, 35 },
		{ 403, 42, 35 },
		{ 403, 43, 35 },
		{ 403, 44, 35 },
		{ 403, 45, 35 },
		{ 403, 46, 35 },
		{ 403, 47, 35 },
		{ 403, 48, 35 },
		{ 403, 49, 35 },
		{ 403, 50, 35 },
		{ 403, 51, 35 },
		{ 403, 52, 35 },
		{ 403, 53, 35 },
		{ 403, 54, 35 },
		{ 403, 55, 35 },
		{ 403, 56, 35 },
		{ 403, 57, 35 },
		{ 403, 58, 35 },
		{ 403, 59, 35 },
		{ 403, 60, 35 },
		{ 403, 61, 35 },
		{ 403, 62, 35 },
		{ 404, 1, 35 },
		{ 404, 2, 35 },
		{ 404, 3, 35 },
		{ 404, 4, 35 },
		{ 404, 5, 35 },
		{ 404, 6, 35 },
		{ 404, 7, 35 },
		{ 404, 8, 35 },
		{ 404, 9, 35 },
		{ 404, 10, 35 },
		{ 404, 11, 35 },
		{ 404, 12, 35 },
		{ 404, 13, 35 },
		{ 404, 14, 35 },
		{ 404, 15, 35 },
		{ 404, 16, 35 },
		{ 404, 17, 35 },
		{ 404, 18, 35 },
		{ 404, 19, 35 },
		{ 404, 20, 35 },
		{ 404, 21, 35 },
		{ 404, 22, 35 },
		{ 404, 23, 35 },
		{ 404, 24, 35 },
		{ 404, 25, 35 },
		{ 404, 26, 35 },
		{ 404, 27, 35 },
		{ 404, 28, 35 },
		{ 404, 29, 35 },
		{ 404, 30, 35 },
		{ 404, 31, 35 },
		{ 404, 32, 35 },
		{ 404, 33, 35 },
		{ 404, 34, 35 },
		{ 404, 35, 35 },
		{ 404, 36, 35 },
		{ 404, 37, 35 },
		{ 404, 38, 35 },
		{ 404, 39, 35 },
		{ 404, 40, 35 },
		{ 404, 41, 35 },
		{ 404, 42, 35 },
		{ 404, 43, 35 },
		{ 404, 44, 35 },
		{ 404, 45, 35 },
		{ 404, 46, 35 },
		{ 404, 47, 35 },
		{ 404, 48, 35 },
		{ 404, 49, 35 },
		{ 404, 50, 35 },
		{ 404, 51, 35 },
		{ 404, 52, 35 },
		{ 404, 53, 35 },
		{ 404, 54, 35 },
		{ 404, 55, 35 },
		{ 404, 56, 35 },
		{ 404, 57, 35 },
		{ 404, 58, 35 },
		{ 404, 59, 35 },
		{ 404, 60, 35 },
		{ 405, 1, 35 },
		{ 405, 2, 35 },
		{ 405, 3, 35 },
		{ 405, 4, 35 },
		{ 405, 5, 35 },
		{ 405, 6, 35 },
		{ 405, 7, 35 },
		{ 405, 8, 35 },
		{ 405, 9, 35 },
		{ 405, 10, 35 },
		{ 405, 11, 35 },
		{ 405, 12, 35 },
		{ 405, 13, 35 },
		{ 405, 14, 35 },
		{ 405, 15, 35 },
		{ 405, 16, 35 },
		{ 405, 17, 35 },
		{ 405, 18, 35 },
		{ 405, 19, 35 },
		{ 405, 20, 35 },
		{ 405, 21, 35 },
		{ 405, 22, 35 },
		{ 405, 23, 35 },
		{ 405, 24, 35 },
		{ 405, 25, 35 },
		{ 405, 26, 35 },
		{ 405, 27, 35 },
		{ 405, 28, 35 },
		{ 405, 29, 35 },
		{ 405, 30, 35 },
		{ 405, 31, 35 },
		{ 405, 32, 35 },
		{ 405, 33, 35 },
		{ 405, 34, 35 },
		{ 405, 35, 35 },
		{ 405, 36, 35 },
		{ 405, 37, 35 },
		{ 405, 38, 35 },
		{ 405, 39, 35 },
		{ 405, 40, 35 },
		{ 405, 41, 35 },
		{ 405, 42, 35 },
		{ 405, 43, 35 },
		{ 405, 44, 35 },
		{ 405, 45, 35 },
		{ 405, 46, 35 },
		{ 405, 47, 35 },
		{ 405, 48, 35 },
		{ 405, 49, 35 },
		{ 405, 50, 35 },
		{ 405, 51, 35 },
		{ 405, 52, 35 },
		{ 405, 53, 35 },
		{ 405, 54, 35 },
		{ 405, 55, 35 },
		{ 405, 56, 35 },
		{ 405, 57, 35 },
		{ 405, 58, 35 },
		{ 405, 59, 35 },
		{ 406, 1, 35 },
		{ 406, 2, 35 },
		{ 406, 3, 35 },
		{ 406, 4, 35 },
		{ 406, 5, 35 },
		{ 406, 6, 35 },
		{ 406, 7, 35 },
		{ 406, 8, 35 },
		{ 406, 9, 35 },
		{ 406, 10, 35 },
		{ 406, 11, 35 },
		{ 406, 12, 35 },
		{ 406, 13, 35 },
		{ 406, 14, 35 },
		{ 406, 15, 35 },
		{ 406, 16, 35 },
		{ 406, 17, 35 },
		{ 406, 18, 35 },
		{ 406, 19, 35 },
		{ 406, 20, 35 },
		{ 406, 21, 35 },
		{ 406, 22, 35 },
		{ 406, 23, 35 },
		{ 406, 24, 35 },
		{ 406, 25, 35 },
		{ 406, 26, 35 },
		{ 406, 27, 35 },
		{ 406, 28, 35 },
		{ 406, 29, 35 },
		{ 406, 30, 35 },
		{ 406, 31, 35 },
		{ 406, 32, 35 },
		{ 406, 33, 35 },
		{ 406, 34, 35 },
		{ 406, 35, 35 },
		{ 406, 36, 35 },
		{ 406, 37, 35 },
		{ 406, 38, 35 },
		{ 406, 39, 35 },
		{ 406, 40, 35 },
		{ 406, 41, 35 },
		{ 406, 42, 35 },
		{ 406, 43, 35 },
		{ 406, 44, 35 },
		{ 406, 45, 35 },
		{ 406, 46, 35 },
		{ 406, 47, 35 },
		{ 406, 48, 35 },
		{ 406, 49, 35 },
		{ 406, 50, 35 },
		{ 406, 51, 35 },
		{ 406, 52, 35 },
		{ 406, 53, 35 },
		{ 406, 54, 35 },
		{ 406, 55, 35 },
		{ 406, 56, 35 },
		{ 406, 57, 35 },
		{ 406, 58, 35 },
		{ 407, 1, 35 },
		{ 407, 2, 35 },
		{ 407, 3, 35 },
		{ 407, 4, 35 },
		{ 407, 5, 35 },
		{ 407, 6, 35 },
		{ 407, 7, 35 },
		{ 407, 8, 35 },
		{ 407, 9, 35 },
		{ 407, 10, 35 },
		{ 407, 11, 35 },
		{ 407, 12, 35 },
		{ 407, 13, 35 },
		{ 407, 14, 35 },
		{ 407, 15, 35 },
		{ 407, 16, 35 },
		{ 407, 17, 35 },
		{ 407, 18, 35 },
		{ 407, 19, 35 },
		{ 407, 20, 35 },
		{ 407, 21, 35 },
		{ 407, 22, 35 },
		{ 407, 23, 35 },
		{ 407, 24, 35 },
		{ 407, 25, 35 },
		{ 407, 26, 35 },
		{ 407, 27, 35 },
		{ 407, 28, 35 },
		{ 407, 29, 35 },
		{ 407, 30, 35 },
		{ 407, 31, 35 },
		{ 407, 32, 35 },
		{ 407, 33, 35 },
		{ 407, 34, 35 },
		{ 407, 35, 35 },
		{ 407, 36, 35 },
		{ 407, 37, 35 },
		{ 407, 38, 35 },
		{ 407, 39, 35 },
		{ 407, 40, 35 },
		{ 407, 41, 35 },
		{ 407, 42, 35 },
		{ 407, 43, 35 },
		{ 407, 44, 35 },
		{ 407, 45, 35 },
		{ 407, 46, 35 },
		{ 407, 47, 35 },
		{ 407, 48, 35 },
		{ 407, 49, 35 },
		{ 407, 50, 35 },
		{ 407, 51, 35 },
		{ 407, 52, 35 },
		{ 407, 53, 35 },
		{ 407, 54, 35 },
		{ 407, 55, 35 },
		{ 407, 56, 35 },
		{ 408, 1, 35 },
		{ 408, 2, 35 },
		{ 408, 3, 35 },
		{ 408, 4, 35 },
		{ 408, 5, 35 },
		{ 408, 6, 35 },
		{ 408, 7, 35 },
		{ 408, 8, 35 },
		{ 408, 9, 35 },
		{ 408, 10, 35 },
		{ 408, 11, 35 },
		{ 408, 12, 35 },
		{ 408, 13, 35 },
		{ 408, 14, 35 },
		{ 408, 15, 35 },
		{ 408, 16, 35 },
		{ 408, 17, 35 },
		{ 408, 18, 35 },
		{ 408, 19, 35 },
		{ 408, 20, 35 },
		{ 408, 21, 35 },
		{ 408, 22, 35 },
		{ 408, 23, 35 },
		{ 408, 24, 35 },
		{ 408, 25, 35 },
		{ 408, 26, 35 },
		{ 408, 27, 35 },
		{ 408, 28, 35 },
		{ 408, 29, 35 },
		{ 408, 30, 35 },
		{ 408, 31, 35 },
		{ 408, 32, 35 },
		{ 408, 33, 35 },
		{ 408, 34, 35 },
		{ 408, 35, 35 },
		{ 408, 36, 35 },
		{ 408, 37, 35 },
		{ 408, 38, 35 },
		{ 408, 39, 35 },
		{ 408, 40, 35 },
		{ 408, 41, 35 },
		{ 408, 42, 35 },
		{ 408, 43, 35 },
		{ 408, 44, 35 },
		{ 408, 45, 35 },
		{ 408, 46, 35 },
		{ 408, 47, 35 },
		{ 408, 48, 35 },
		{ 408, 49, 35 },
		{ 408, 50, 35 },
		{ 408, 51, 35 },
		{ 408, 52, 35 },
		{ 408, 53, 35 },
		{ 408, 54, 35 },
		{ 408, 55, 35 },
		{ 409, 1, 35 },
		{ 409, 2, 35 },
		{ 409, 3, 35 },
		{ 409, 4, 35 },
		{ 409, 5, 35 },
		{ 409, 6, 35 },
		{ 409, 7, 35 },
		{ 409, 8, 35 },
		{ 409, 9, 35 },
		{ 409, 10, 35 },
		{ 409, 11, 35 },
		{ 409, 12, 35 },
		{ 409, 13, 35 },
		{ 409, 14, 35 },
		{ 409, 15, 35 },
		{ 409, 16, 35 },
		{ 409, 17, 35 },
		{ 409, 18, 35 },
		{ 409, 19, 35 },
		{ 409, 20, 35 },
		{ 409, 21, 35 },
		{ 409, 22, 35 },
		{ 409, 23, 35 },
		{ 409, 24, 35 },
		{ 409, 25, 35 },
		{ 409, 26, 35 },
		{ 409, 27, 35 },
		{ 409, 28, 35 },
		{ 409, 29, 35 },
		{ 409, 30, 35 },
		{ 409, 31, 35 },
		{ 409, 32, 35 },
		{ 409, 33, 35 },
		{ 409, 34, 35 },
		{ 409, 35, 35 },
		{ 409, 36, 35 },
		{ 409, 37, 35 },
		{ 409, 38, 35 },
		{ 409, 39, 35 },
		{ 409, 40, 35 },
		{ 409, 41, 35 },
		{ 409, 42, 35 },
		{ 409, 43, 35 },
		{ 409, 44, 35 },
		{ 409, 45, 35 },
		{ 409, 46, 35 },
		{ 409, 47, 35 },
		{ 409, 48, 35 },
		{ 409, 49, 35 },
		{ 409, 50, 35 },
		{ 409, 51, 35 },
		{ 409, 52, 35 },
		{ 409, 53, 35 },
		{ 409, 54, 35 },
		{ 410, 1, 35 },
		{ 410, 2, 35 },
		{ 410, 3, 35 },
		{ 410, 4, 35 },
		{ 410, 5, 35 },
		{ 410, 6, 35 },
		{ 410, 7, 35 },
		{ 410, 8, 35 },
		{ 410, 9, 35 },
		{ 410, 10, 35 },
		{ 410, 11, 35 },
		{ 410, 12, 35 },
		{ 410, 13, 35 },
		{ 410, 14, 35 },
		{ 410, 15, 35 },
		{ 410, 16, 35 },
		{ 410, 17, 35 },
		{ 410, 18, 35 },
		{ 410, 19, 35 },
		{ 410, 20, 35 },
		{ 410, 21, 35 },
		{ 410, 22, 35 },
		{ 410, 23, 35 },
		{ 410, 24, 35 },
		{ 410, 25, 35 },
		{ 410, 26, 35 },
		{ 410, 27, 35 },
		{ 410, 28, 35 },
		{ 410, 29, 35 },
		{ 410, 30, 35 },
		{ 410, 31, 35 },
		{ 410, 32, 35 },
		{ 410, 33, 35 },
		{ 410, 34, 35 },
		{ 410, 35, 35 },
		{ 410, 36, 35 },
		{ 410, 37, 35 },
		{ 410, 38, 35 },
		{ 410, 39, 35 },
		{ 410, 40, 35 },
		{ 410, 41, 35 },
		{ 410, 42, 35 },
		{ 410, 43, 35 },
		{ 410, 44, 35 },
		{ 410, 45, 35 },
		{ 410, 46, 35 },
		{ 410, 47, 35 },
		{ 410, 48, 35 },
		{ 410, 49, 35 },
		{ 410, 50, 35 },
		{ 410, 51, 35 },
		{ 410, 52, 35 },
		{ 411, 1, 35 },
		{ 411, 2, 35 },
		{ 411, 3, 35 },
		{ 411, 4, 35 },
		{ 411, 5, 35 },
		{ 411, 6, 35 },
		{ 411, 7, 35 },
		{ 411, 8, 35 },
		{ 411, 9, 35 },
		{ 411, 10, 35 },
		{ 411, 11, 35 },
		{ 411, 12, 35 },
		{ 411, 13, 35 },
		{ 411, 14, 35 },
		{ 411, 15, 35 },
		{ 411, 16, 35 },
		{ 411, 17, 35 },
		{ 411, 18, 35 },
		{ 411, 19, 35 },
		{ 411, 20, 35 },
		{ 411, 21, 35 },
		{ 411, 22, 35 },
		{ 411, 23, 35 },
		{ 411, 24, 35 },
		{ 411, 25, 35 },
		{ 411, 26, 35 },
		{ 411, 27, 35 },
		{ 411, 28, 35 },
		{ 411, 29, 35 },
		{ 411, 30, 35 },
		{ 411, 31, 35 },
		{ 411, 32, 35 },
		{ 411, 33, 35 },
		{ 411, 34, 35 },
		{ 411, 35, 35 },
		{ 411, 36, 35 },
		{ 411, 37, 35 },
		{ 411, 38, 35 },
		{ 411, 39, 35 },
		{ 411, 40, 35 },
		{ 411, 41, 35 },
		{ 411, 42, 35 },
		{ 411, 43, 35 },
		{ 411, 44, 35 },
		{ 411, 45, 35 },
		{ 411, 46, 35 },
		{ 411, 47, 35 },
		{ 411, 48, 35 },
		{ 411, 49, 35 },
		{ 411, 50, 35 },
		{ 411, 51, 35 },
		{ 412, 1, 35 },
		{ 412, 2, 35 },
		{ 412, 3, 35 },
		{ 412, 4, 35 },
		{ 412, 5, 35 },
		{ 412, 6, 35 },
		{ 412, 7, 35 },
		{ 412, 8, 35 },
		{ 412, 9, 35 },
		{ 412, 10, 35 },
		{ 412, 11, 35 },
		{ 412, 12, 35 },
		{ 412, 13, 35 },
		{ 412, 14, 35 },
		{ 412, 15, 35 },
		{ 412, 16, 35 },
		{ 412, 17, 35 },
		{ 412, 18, 35 },
		{ 412, 19, 35 },
		{ 412, 20, 35 },
		{ 412, 21, 35 },
		{ 412, 22, 35 },
		{ 412, 23, 35 },
		{ 412, 24, 35 },
		{ 412, 25, 35 },
		{ 412, 26, 35 },
		{ 412, 27, 35 },
		{ 412, 28, 35 },
		{ 412, 29, 35 },
		{ 412, 30, 35 },
		{ 412, 31, 35 },
		{ 412, 32, 35 },
		{ 412, 33, 35 },
		{ 412, 34, 35 },
		{ 412, 35, 35 },
		{ 412, 36, 35 },
		{ 412, 37, 35 },
		{ 412, 38, 35 },
		{ 412, 39, 35 },
		{ 412, 40, 35 },
		{ 412, 41, 35 },
		{ 412, 42, 35 },
		{ 412, 43, 35 },
		{ 412, 44, 35 },
		{ 412, 45, 35 },
		{ 412, 46, 35 },
		{ 412, 47, 35 },
		{ 412, 48, 35 },
		{ 412, 49, 35 },
		{ 412, 50, 35 },
		{ 413, 1, 35 },
		{ 413, 2, 35 },
		{ 413, 3, 35 },
		{ 413, 4, 35 },
		{ 413, 5, 35 },
		{ 413, 6, 35 },
		{ 413, 7, 35 },
		{ 413, 8, 35 },
		{ 413, 9, 35 },
		{ 413, 10, 35 },
		{ 413, 11, 35 },
		{ 413, 12, 35 },
		{ 413, 13, 35 },
		{ 413, 14, 35 },
		{ 413, 15, 35 },
		{ 413, 16, 35 },
		{ 413, 17, 35 },
		{ 413, 18, 35 },
		{ 413, 19, 35 },
		{ 413, 20, 35 },
		{ 413, 21, 35 },
		{ 413, 22, 35 },
		{ 413, 23, 35 },
		{ 413, 24, 35 },
		{ 413, 25, 35 },
		{ 413, 26, 35 },
		{ 413, 27, 35 },
		{ 413, 28, 35 },
		{ 413, 29, 35 },
		{ 413, 30, 35 },
		{ 413, 31, 35 },
		{ 413, 32, 35 },
		{ 413, 33, 35 },
		{ 413, 34, 35 },
		{ 413, 35, 35 },
		{ 413, 36, 35 },
		{ 413, 37, 35 },
		{ 413, 38, 35 },
		{ 413, 39, 35 },
		{ 413, 40, 35 },
		{ 413, 41, 35 },
		{ 413, 42, 35 },
		{ 413, 43, 35 },
		{ 413, 44, 35 },
		{ 413, 45, 35 },
		{ 413, 46, 35 },
		{ 413, 47, 35 },
		{ 413, 48, 35 },
		{ 414, 1, 35 },
		{ 414, 2, 35 },
		{ 414, 3, 35 },
		{ 414, 4, 35 },
		{ 414, 5, 35 },
		{ 414, 6, 35 },
		{ 414, 7, 35 },
		{ 414, 8, 35 },
		{ 414, 9, 35 },
		{ 414, 10, 35 },
		{ 414, 11, 35 },
		{ 414, 12, 35 },
		{ 414, 13, 35 },
		{ 414, 14, 35 },
		{ 414, 15, 35 },
		{ 414, 16, 35 },
		{ 414, 17, 35 },
		{ 414, 18, 35 },
		{ 414, 19, 35 },
		{ 414, 20, 35 },
		{ 414, 21, 35 },
		{ 414, 22, 35 },
		{ 414, 23, 35 },
		{ 414, 24, 35 },
		{ 414, 25, 35 },
		{ 414, 26, 35 },
		{ 414, 27, 35 },
		{ 414, 28, 35 },
		{ 414, 29, 35 },
		{ 414, 30, 35 },
		{ 414, 31, 35 },
		{ 414, 32, 35 },
		{ 414, 33, 35 },
		{ 414, 34, 35 },
		{ 414, 35, 35 },
		{ 414, 36, 35 },
		{ 414, 37, 35 },
		{ 414, 38, 35 },
		{ 414, 39, 35 },
		{ 414, 40, 35 },
		{ 414, 41, 35 },
		{ 414, 42, 35 },
		{ 414, 43, 35 },
		{ 414, 44, 35 },
		{ 414, 45, 35 },
		{ 414, 46, 35 },
		{ 414, 47, 35 },
		{ 415, 1, 35 },
		{ 415, 2, 35 },
		{ 415, 3, 35 },
		{ 415, 4, 35 },
		{ 415, 5, 35 },
		{ 415, 6, 35 },
		{ 415, 7, 35 },
		{ 415, 8, 35 },
		{ 415, 9, 35 },
		{ 415, 10, 35 },
		{ 415, 11, 35 },
		{ 415, 12, 35 },
		{ 415, 13, 35 },
		{ 415, 14, 35 },
		{ 415, 15, 35 },
		{ 415, 16, 35 },
		{ 415, 17, 35 },
		{ 415, 18, 35 },
		{ 415, 19, 35 },
		{ 415, 20, 35 },
		{ 415, 21, 35 },
		{ 415, 22, 35 },
		{ 415, 23, 35 },
		{ 415, 24, 35 },
		{ 415, 25, 35 },
		{ 415, 26, 35 },
		{ 415, 27, 35 },
		{ 415, 28, 35 },
		{ 415, 29, 35 },
		{ 415, 30, 35 },
		{ 415, 31, 35 },
		{ 415, 32, 35 },
		{ 415, 33, 35 },
		{ 415, 34, 35 },
		{ 415, 35, 35 },
		{ 415, 36, 35 },
		{ 415, 37, 35 },
		{ 415, 38, 35 },
		{ 415, 39, 35 },
		{ 415, 40, 35 },
		{ 415, 41, 35 },
		{ 415, 42, 35 },
		{ 415, 43, 35 },
		{ 415, 44, 35 },
		{ 415, 45, 35 },
		{ 415, 46, 35 },
		{ 416, 1, 35 },
		{ 416, 2, 35 },
		{ 416, 3, 35 },
		{ 416, 4, 35 },
		{ 416, 5, 35 },
		{ 416, 6, 35 },
		{ 416, 7, 35 },
		{ 416, 8, 35 },
		{ 416, 9, 35 },
		{ 416, 10, 35 },
		{ 416, 11, 35 },
		{ 416, 12, 35 },
		{ 416, 13, 35 },
		{ 416, 14, 35 },
		{ 416, 15, 35 },
		{ 416, 16, 35 },
		{ 416, 17, 35 },
		{ 416, 18, 35 },
		{ 416, 19, 35 },
		{ 416, 20, 35 },
		{ 416, 21, 35 },
		{ 416, 22, 35 },
		{ 416, 23, 35 },
		{ 416, 24, 35 },
		{ 416, 25, 35 },
		{ 416, 26, 35 },
		{ 416, 27, 35 },
		{ 416, 28, 35 },
		{ 416, 29, 35 },
		{ 416, 30, 35 },
		{ 416, 31, 35 },
		{ 416, 32, 35 },
		{ 416, 33, 35 },
		{ 416, 34, 35 },
		{ 416, 35, 35 },
		{ 416, 36, 35 },
		{ 416, 37, 35 },
		{ 416, 38, 35 },
		{ 416, 39, 35 },
		{ 416, 40, 35 },
		{ 416, 41, 35 },
		{ 416, 42, 35 },
		{ 416, 43, 35 },
		{ 416, 44, 35 },
		{ 417, 1, 35 },
		{ 417, 2, 35 },
		{ 417, 3, 35 },
		{ 417, 4, 35 },
		{ 417, 5, 35 },
		{ 417, 6, 35 },
		{ 417, 7, 35 },
		{ 417, 8, 35 },
		{ 417, 9, 35 },
		{ 417, 10, 35 },
		{ 417, 11, 35 },
		{ 417, 12, 35 },
		{ 417, 13, 35 },
		{ 417, 14, 35 },
		{ 417, 15, 35 },
		{ 417, 16, 35 },
		{ 417, 17, 35 },
		{ 417, 18, 35 },
		{ 417, 19, 35 },
		{ 417, 20, 35 },
		{ 417, 21, 35 },
		{ 417, 22, 35 },
		{ 417, 23, 35 },
		{ 417, 24, 35 },
		{ 417, 25, 35 },
		{ 417, 26, 35 },
		{ 417, 27, 35 },
		{ 417, 28, 35 },
		{ 417, 29, 35 },
		{ 417, 30, 35 },
		{ 417, 31, 35 },
		{ 417, 32, 35 },
		{ 417, 33, 35 },
		{ 417, 34, 35 },
		{ 417, 35, 35 },
		{ 417, 36, 35 },
		{ 417, 37, 35 },
		{ 417, 38, 35 },
		{ 417, 39, 35 },
		{ 417, 40, 35 },
		{ 417, 41, 35 },
		{ 417, 42, 35 },
		{ 417, 43, 35 },
		{ 418, 1, 35 },
		{ 418, 2, 35 },
		{ 418, 3, 35 },
		{ 418, 4, 35 },
		{ 418, 5, 35 },
		{ 418, 6, 35 },
		{ 418, 7, 35 },
		{ 418, 8, 35 },
		{ 418, 9, 35 },
		{ 418, 10, 35 },
		{ 418, 11, 35 },
		{ 418, 12, 35 },
		{ 418, 13, 35 },
		{ 418, 14, 35 },
		{ 418, 15, 35 },
		{ 418, 16, 35 },
		{ 418, 17, 35 },
		{ 418, 18, 35 },
		{ 418, 19, 35 },
		{ 418, 20, 35 },
		{ 418, 21, 35 },
		{ 418, 22, 35 },
		{ 418, 23, 35 },
		{ 418, 24, 35 },
		{ 418, 25, 35 },
		{ 418, 26, 35 },
		{ 418, 27, 35 },
		{ 418, 28, 35 },
		{ 418, 29, 35 },
		{ 418, 30, 35 },
		{ 418, 31, 35 },
		{ 418, 32, 35 },
		{ 418, 33, 35 },
		{ 418, 34, 35 },
		{ 418, 35, 35 },
		{ 418, 36, 35 },
		{ 418, 37, 35 },
		{ 418, 38, 35 },
		{ 418, 39, 35 },
		{ 418, 40, 35 },
		{ 418, 41, 35 },
		{ 418, 42, 35 },
		{ 419, 1, 35 },
		{ 419, 2, 35 },
		{ 419, 3, 35 },
		{ 419, 4, 35 },
		{ 419, 5, 35 },
		{ 419, 6, 35 },
		{ 419, 7, 35 },
		{ 419, 8, 35 },
		{ 419, 9, 35 },
		{ 419, 10, 35 },
		{ 419, 11, 35 },
		{ 419, 12, 35 },
		{ 419, 13, 35 },
		{ 419, 14, 35 },
		{ 419, 15, 35 },
		{ 419, 16, 35 },
		{ 419, 17, 35 },
		{ 419, 18, 35 },
		{ 419, 19, 35 },
		{ 419, 20, 35 },
		{ 419, 21, 35 },
		{ 419, 22, 35 },
		{ 419, 23, 35 },
		{ 419, 24, 35 },
		{ 419, 25, 35 },
		{ 419, 26, 35 },
		{ 419, 27, 35 },
		{ 419, 28, 35 },
		{ 419, 29, 35 },
		{ 419, 30, 35 },
		{ 419, 31, 35 },
		{ 419, 32, 35 },
		{ 419, 33, 35 },
		{ 419, 34, 35 },
		{ 419, 35, 35 },
		{ 419, 36, 35 },
		{ 419, 37, 35 },
		{ 419, 38, 35 },
		{ 419, 39, 35 },
		{ 419, 40, 35 },
		{ 420, 1, 35 },
		{ 420, 2, 35 },
		{ 420, 3, 35 },
		{ 420, 4, 35 },
		{ 420, 5, 35 },
		{ 420, 6, 35 },
		{ 420, 7, 35 },
		{ 420, 8, 35 },
		{ 420, 9, 35 },
		{ 420, 10, 35 },
		{ 420, 11, 35 },
		{ 420, 12, 35 },
		{ 420, 13, 35 },
		{ 420, 14, 35 },
		{ 420, 15, 35 },
		{ 420, 16, 35 },
		{ 420, 17, 35 },
		{ 420, 18, 35 },
		{ 420, 19, 35 },
		{ 420, 20, 35 },
		{ 420, 21, 35 },
		{ 420, 22, 35 },
		{ 420, 23, 35 },
		{ 420, 24, 35 },
		{ 420, 25, 35 },
		{ 420, 26, 35 },
		{ 420, 27, 35 },
		{ 420, 28, 35 },
		{ 420, 29, 35 },
		{ 420, 30, 35 },
		{ 420, 31, 35 },
		{ 420, 32, 35 },
		{ 420, 33, 35 },
		{ 420, 34, 35 },
		{ 420, 35, 35 },
		{ 420, 36, 35 },
		{ 420, 37, 35 },
		{ 420, 38, 35 },
		{ 420, 39, 35 },
		{ 421, 1, 35 },
		{ 421, 2, 35 },
		{ 421, 3, 35 },
		{ 421, 4, 35 },
		{ 421, 5, 35 },
		{ 421, 6, 35 },
		{ 421, 7, 35 },
		{ 421, 8, 35 },
		{ 421, 9, 35 },
		{ 421, 10, 35 },
		{ 421, 11, 35 },
		{ 421, 12, 35 },
		{ 421, 13, 35 },
		{ 421, 14, 35 },
		{ 421, 15, 35 },
		{ 421, 16, 35 },
		{ 421, 17, 35 },
		{ 421, 18, 35 },
		{ 421, 19, 35 },
		{ 421, 20, 35 },
		{ 421, 21, 35 },
		{ 421, 22, 35 },
		{ 421, 23, 35 },
		{ 421, 24, 35 },
		{ 421, 25, 35 },
		{ 421, 26, 35 },
		{ 421, 27, 35 },
		{ 421, 28, 35 },
		{ 421, 29, 35 },
		{ 421, 30, 35 },
		{ 421, 31, 35 },
		{ 421, 32, 35 },
		{ 421, 33, 35 },
		{ 421, 34, 35 },
		{ 421, 35, 35 },
		{ 421, 36, 35 },
		{ 421, 37, 35 },
		{ 421, 38, 35 },
		{ 422, 1, 35 },
		{ 422, 2, 35 },
		{ 422, 3, 35 },
		{ 422, 4, 35 },
		{ 422, 5, 35 },
		{ 422, 6, 35 },
		{ 422, 7, 35 },
		{ 422, 8, 35 },
		{ 422, 9, 35 },
		{ 422, 10, 35 },
		{ 422, 11, 35 },
		{ 422, 12, 35 },
		{ 422, 13, 35 },
		{ 422, 14, 35 },
		{ 422, 15, 35 },
		{ 422, 16, 35 },
		{ 422, 17, 35 },
		{ 422, 18, 35 },
		{ 422, 19, 35 },
		{ 422, 20, 35 },
		{ 422, 21, 35 },
		{ 422, 22, 35 },
		{ 422, 23, 35 },
		{ 422, 24, 35 },
		{ 422, 25, 35 },
		{ 422, 26, 35 },
		{ 422, 27, 35 },
		{ 422, 28, 35 },
		{ 422, 29, 35 },
		{ 422, 30, 35 },
		{ 422, 31, 35 },
		{ 422, 32, 35 },
		{ 422, 33, 35 },
		{ 422, 34, 35 },
		{ 422, 35, 35 },
		{ 422, 36, 35 },
		{ 423, 1, 35 },
		{ 423, 2, 35 },
		{ 423, 3, 35 },
		{ 423, 4, 35 },
		{ 423, 5, 35 },
		{ 423, 6, 35 },
		{ 423, 7, 35 },
		{ 423, 8, 35 },
		{ 423, 9, 35 },
		{ 423, 10, 35 },
		{ 423, 11, 35 },
		{ 423, 12, 35 },
		{ 423, 13, 35 },
		{ 423, 14, 35 },
		{ 423, 15, 35 },
		{ 423, 16, 35 },
		{ 423, 17, 35 },
		{ 423, 18, 35 },
		{ 423, 19, 35 },
		{ 423, 20, 35 },
		{ 423, 21, 35 },
		{ 423, 22, 35 },
		{ 423, 23, 35 },
		{ 423, 24, 35 },
		{ 423, 25, 35 },
		{ 423, 26, 35 },
		{ 423, 27, 35 },
		{ 423, 28, 35 },
		{ 423, 29, 35 },
		{ 423, 30, 35 },
		{ 423, 31, 35 },
		{ 423, 32, 35 },
		{ 423, 33, 35 },
		{ 423, 34, 35 },
		{ 423, 35, 35 },
		{ 424, 1, 35 },
		{ 424, 2, 35 },
		{ 424, 3, 35 },
		{ 424, 4, 35 },
		{ 424, 5, 35 },
		{ 424, 6, 35 },
		{ 424, 7, 35 },
		{ 424, 8, 35 },
		{ 424, 9, 35 },
		{ 424, 10, 35 },
		{ 424, 11, 35 },
		{ 424, 12, 35 },
		{ 424, 13, 35 },
		{ 424, 14, 35 },
		{ 424, 15, 35 },
		{ 424, 16, 35 },
		{ 424, 17, 35 },
		{ 424, 18, 35 },
		{ 424, 19, 35 },
		{ 424, 20, 35 },
		{ 424, 21, 35 },
		{ 424, 22, 35 },
		{ 424, 23, 35 },
		{ 424, 24, 35 },
		{ 424, 25, 35 },
		{ 424, 26, 35 },
		{ 424, 27, 35 },
		{ 424, 28, 35 },
		{ 424, 29, 35 },
		{ 424, 30, 35 },
		{ 424, 31, 35 },
		{ 424, 32, 35 },
		{ 424, 33, 35 },
		{ 424, 34, 35 },
		{ 425, 1, 35 },
		{ 425, 2, 35 },
		{ 425, 3, 35 },
		{ 425, 4, 35 },
		{ 425, 5, 35 },
		{ 425, 6, 35 },
		{ 425, 7, 35 },
		{ 425, 8, 35 },
		{ 425, 9, 35 },
		{ 425, 10, 35 },
		{ 425, 11, 35 },
		{ 425, 12, 35 },
		{ 425, 13, 35 },
		{ 425, 14, 35 },
		{ 425, 15, 35 },
		{ 425, 16, 35 },
		{ 425, 17, 35 },
		{ 425, 18, 35 },
		{ 425, 19, 35 },
		{ 425, 20, 35 },
		{ 425, 21, 35 },
		{ 425, 22, 35 },
		{ 425, 23, 35 },
		{ 425, 24, 35 },
		{ 425, 25, 35 },
		{ 425, 26, 35 },
		{ 425, 27, 35 },
		{ 425, 28, 35 },
		{ 425, 29, 35 },
		{ 425, 30, 35 },
		{ 425, 31, 35 },
		{ 425, 32, 35 },
		{ 426, 1, 35 },
		{ 426, 2, 35 },
		{ 426, 3, 35 },
		{ 426, 4, 35 },
		{ 426, 5, 35 },
		{ 426, 6, 35 },
		{ 426, 7, 35 },
		{ 426, 8, 35 },
		{ 426, 9, 35 },
		{ 426, 10, 35 },
		{ 426, 11, 35 },
		{ 426, 12, 35 },
		{ 426, 13, 35 },
		{ 426, 14, 35 },
		{ 426, 15, 35 },
		{ 426, 16, 35 },
		{ 426, 17, 35 },
		{ 426, 18, 35 },
		{ 426, 19, 35 },
		{ 426, 20, 35 },
		{ 426, 21, 35 },
		{ 426, 22, 35 },
		{ 426, 23, 35 },
		{ 426, 24, 35 },
		{ 426, 25, 35 },
		{ 426, 26, 35 },
		{ 426, 27, 35 },
		{ 426, 28, 35 },
		{ 426, 29, 35 },
		{ 426, 30, 35 },
		{ 426, 31, 35 },
		{ 427, 1, 35 },
		{ 427, 2, 35 },
		{ 427, 3, 35 },
		{ 427, 4, 35 },
		{ 427, 5, 35 },
		{ 427, 6, 35 },
		{ 427, 7, 35 },
		{ 427, 8, 35 },
		{ 427, 9, 35 },
		{ 427, 10, 35 },
		{ 427, 11, 35 },
		{ 427, 12, 35 },
		{ 427, 13, 35 },
		{ 427, 14, 35 },
		{ 427, 15, 35 },
		{ 427, 16, 35 },
		{ 427, 17, 35 },
		{ 427, 18, 35 },
		{ 427, 19, 35 },
		{ 427, 20, 35 },
		{ 427, 21, 35 },
		{ 427, 22, 35 },
		{ 427, 23, 35 },
		{ 427, 24, 35 },
		{ 427, 25, 35 },
		{ 427, 26, 35 },
		{ 427, 27, 35 },
		{ 427, 28, 35 },
		{ 427, 29, 35 },
		{ 427, 30, 35 },
		{ 428, 1, 35 },
		{ 428, 2, 35 },
		{ 428, 3, 35 },
		{ 428, 4, 35 },
		{ 428, 5, 35 },
		{ 428, 6, 35 },
		{ 428, 7, 35 },
		{ 428, 8, 35 },
		{ 428, 9, 35 },
		{ 428, 10, 35 },
		{ 428, 11, 35 },
		{ 428, 12, 35 },
		{ 428, 13, 35 },
		{ 428, 14, 35 },
		{ 428, 15, 35 },
		{ 428, 16, 35 },
		{ 428, 17, 35 },
		{ 428, 18, 35 },
		{ 428, 19, 35 },
		{ 428, 20, 35 },
		{ 428, 21, 35 },
		{ 428, 22, 35 },
		{ 428, 23, 35 },
		{ 428, 24, 35 },
		{ 428, 25, 35 },
		{ 428, 26, 35 },
		{ 428, 27, 35 },
		{ 428, 28, 35 },
		{ 429, 1, 35 },
		{ 429, 2, 35 },
		{ 429, 3, 35 },
		{ 429, 4, 35 },
		{ 429, 5, 35 },
		{ 429, 6, 35 },
		{ 429, 7, 35 },
		{ 429, 8, 35 },
		{ 429, 9, 35 },
		{ 429, 10, 35 },
		{ 429, 11, 35 },
		{ 429, 12, 35 },
		{ 429, 13, 35 },
		{ 429, 14, 35 },
		{ 429, 15, 35 },
		{ 429, 16, 35 },
		{ 429, 17, 35 },
		{ 429, 18, 35 },
		{ 429, 19, 35 },
		{ 429, 20, 35 },
		{ 429, 21, 35 },
		{ 429, 22, 35 },
		{ 429, 23, 35 },
		{ 429, 24, 35 },
		{ 429, 25, 35 },
		{ 429, 26, 35 },
		{ 429, 27, 35 },
		{ 430, 1, 35 },
		{ 430, 2, 35 },
		{ 430, 3, 35 },
		{ 430, 4, 35 },
		{ 430, 5, 35 },
		{ 430, 6, 35 },
		{ 430, 7, 35 },
		{ 430, 8, 35 },
		{ 430, 9, 35 },
		{ 430, 10, 35 },
		{ 430, 11, 35 },
		{ 430, 12, 35 },
		{ 430, 13, 35 },
		{ 430, 14, 35 },
		{ 430, 15, 35 },
		{ 430, 16, 35 },
		{ 430, 17, 35 },
		{ 430, 18, 35 },
		{ 430, 19, 35 },
		{ 430, 20, 35 },
		{ 430, 21, 35 },
		{ 430, 22, 35 },
		{ 430, 23, 35 },
		{ 430, 24, 35 },
		{ 430, 25, 35 },
		{ 430, 26, 35 },
		{ 431, 1, 35 },
		{ 431, 2, 35 },
		{ 431, 3, 35 },
		{ 431, 4, 35 },
		{ 431, 5, 35 },
		{ 431, 6, 35 },
		{ 431, 7, 35 },
		{ 431, 8, 35 },
		{ 431, 9, 35 },
		{ 431, 10, 35 },
		{ 431, 11, 35 },
		{ 431, 12, 35 },
		{ 431, 13, 35 },
		{ 431, 14, 35 },
		{ 431, 15, 35 },
		{ 431, 16, 35 },
		{ 431, 17, 35 },
		{ 431, 18, 35 },
		{ 431, 19, 35 },
		{ 431, 20, 35 },
		{ 431, 21, 35 },
		{ 431, 22, 35 },
		{ 431, 23, 35 },
		{ 431, 24, 35 },
		{ 432, 1, 35 },
		{ 432, 2, 35 },
		{ 432, 3, 35 },
		{ 432, 4, 35 },
		{ 432, 5, 35 },
		{ 432, 6, 35 },
		{ 432, 7, 35 },
		{ 432, 8, 35 },
		{ 432, 9, 35 },
		{ 432, 10, 35 },
		{ 432, 11, 35 },
		{ 432, 12, 35 },
		{ 432, 13, 35 },
		{ 432, 14, 35 },
		{ 432, 15, 35 },
		{ 432, 16, 35 },
		{ 432, 17, 35 },
		{ 432, 18, 35 },
		{ 432, 19, 35 },
		{ 432, 20, 35 },
		{ 432, 21, 35 },
		{ 432, 22, 35 },
		{ 432, 23, 35 },
		{ 433, 1, 35 },
		{ 433, 2, 35 },
		{ 433, 3, 35 },
		{ 433, 4, 35 },
		{ 433, 5, 35 },
		{ 433, 6, 35 },
		{ 433, 7, 35 },
		{ 433, 8, 35 },
		{ 433, 9, 35 },
		{ 433, 10, 35 },
		{ 433, 11, 35 },
		{ 433, 12, 35 },
		{ 433, 13, 35 },
		{ 433, 14, 35 },
		{ 433, 15, 35 },
		{ 433, 16, 35 },
		{ 433, 17, 35 },
		{ 433, 18, 35 },
		{ 433, 19, 35 },
		{ 433, 20, 35 },
		{ 433, 21, 35 },
		{ 433, 22, 35 },
		{ 434, 1, 35 },
		{ 434, 2, 35 },
		{ 434, 3, 35 },
		{ 434, 4, 35 },
		{ 434, 5, 35 },
		{ 434, 6, 35 },
		{ 434, 7, 35 },
		{ 434, 8, 35 },
		{ 434, 9, 35 },
		{ 434, 10, 35 },
		{ 434, 11, 35 },
		{ 434, 12, 35 },
		{ 434, 13, 35 },
		{ 434, 14, 35 },
		{ 434, 15, 35 },
		{ 434, 16, 35 },
		{ 434, 17, 35 },
		{ 434, 18, 35 },
		{ 434, 19, 35 },
		{ 434, 20, 35 },
		{ 435, 1, 35 },
		{ 435, 2, 35 },
		{ 435, 3, 35 },
		{ 435, 4, 35 },
		{ 435, 5, 35 },
		{ 435, 6, 35 },
		{ 435, 7, 35 },
		{ 435, 8, 35 },
		{ 435, 9, 35 },
		{ 435, 10, 35 },
		{ 435, 11, 35 },
		{ 435, 12, 35 },
		{ 435, 13, 35 },
		{ 435, 14, 35 },
		{ 435, 15, 35 },
		{ 435, 16, 35 },
		{ 435, 17, 35 },
		{ 435, 18, 35 },
		{ 435, 19, 35 },
		{ 436, 1, 35 },
		{ 436, 2, 35 },
		{ 436, 3, 35 },
		{ 436, 4, 35 },
		{ 436, 5, 35 },
		{ 436, 6, 35 },
		{ 436, 7, 35 },
		{ 436, 8, 35 },
		{ 436, 9, 35 },
		{ 436, 10, 35 },
		{ 436, 11, 35 },
		{ 436, 12, 35 },
		{ 436, 13, 35 },
		{ 436, 14, 35 },
		{ 436, 15, 35 },
		{ 436, 16, 35 },
		{ 436, 17, 35 },
		{ 436, 18, 35 },
		{ 437, 1, 35 },
		{ 437, 2, 35 },
		{ 437, 3, 35 },
		{ 437, 4, 35 },
		{ 437, 5, 35 },
		{ 437, 6, 35 },
		{ 437, 7, 35 },
		{ 437, 8, 35 },
		{ 437, 9, 35 },
		{ 437, 10, 35 },
		{ 437, 11, 35 },
		{ 437, 12, 35 },
		{ 437, 13, 35 },
		{ 437, 14, 35 },
		{ 437, 15, 35 },
		{ 437, 16, 35 },
		{ 438, 1, 35 },
		{ 438, 2, 35 },
		{ 438, 3, 35 },
		{ 438, 4, 35 },
		{ 438, 5, 35 },
		{ 438, 6, 35 },
		{ 438, 7, 35 },
		{ 438, 8, 35 },
		{ 438, 9, 35 },
		{ 438, 10, 35 },
		{ 438, 11, 35 },
		{ 438, 12, 35 },
		{ 438, 13, 35 },
		{ 438, 14, 35 },
		{ 438, 15, 35 },
		{ 439, 1, 35 },
		{ 439, 2, 35 },
		{ 439, 3, 35 },
		{ 439, 4, 35 },
		{ 439, 5, 35 },
		{ 439, 6, 35 },
		{ 439, 7, 35 },
		{ 439, 8, 35 },
		{ 439, 9, 35 },
		{ 439, 10, 35 },
		{ 439, 11, 35 },
		{ 439, 12, 35 },
		{ 439, 13, 35 },
		{ 439, 14, 35 },
		{ 440, 1, 35 },
		{ 440, 2, 35 },
		{ 440, 3, 35 },
		{ 440, 4, 35 },
		{ 440, 5, 35 },
		{ 440, 6, 35 },
		{ 440, 7, 35 },
		{ 440, 8, 35 },
		{ 440, 9, 35 },
		{ 440, 10, 35 },
		{ 440, 11, 35 },
		{ 440, 12, 35 },
		{ 441, 1, 35 },
		{ 441, 2, 35 },
		{ 441, 3, 35 },
		{ 441, 4, 35 },
		{ 441, 5, 35 },
		{ 441, 6, 35 },
		{ 441, 7, 35 },
		{ 441, 8, 35 },
		{ 441, 9, 35 },
		{ 441, 10, 35 },
		{ 441, 11, 35 },
		{ 442, 1, 35 },
		{ 442, 2, 35 },
		{ 442, 3, 35 },
		{ 442, 4, 35 },
		{ 442, 5, 35 },
		{ 442, 6, 35 },
		{ 442, 7, 35 },
		{ 442, 8, 35 },
		{ 442, 9, 35 },
		{ 442, 10, 35 },
		{ 443, 1, 35 },
		{ 443, 2, 35 },
		{ 443, 3, 35 },
		{ 443, 4, 35 },
		{ 443, 5, 35 },
		{ 443, 6, 35 },
		{ 443, 7, 35 },
		{ 443, 8, 35 },
		{ 444, 1, 35 },
		{ 444, 2, 35 },
		{ 444, 3, 35 },
		{ 444, 4, 35 },
		{ 444, 5, 35 },
		{ 444, 6, 35 },
		{ 444, 7, 35 },
		{ 445, 1, 35 },
		{ 445, 2, 35 },
		{ 445, 3, 35 },
		{ 445, 4, 35 },
		{ 445, 5, 35 },
		{ 445, 6, 35 },
		{ 446, 1, 35 },
		{ 446, 2, 35 },
		{ 446, 3, 35 },
		{ 446, 4, 35 },
		{ 447, 1, 35 },
		{ 447, 2, 35 },
		{ 447, 3, 35 },
		{ 448, 1, 35 },
		{ 448, 2, 35 },
		{ 34, 512, 36 },
		{ 35, 511, 36 },
		{ 35, 512, 36 },
		{ 36, 510, 36 },
		{ 36, 511, 36 },
		{ 36, 512, 36 },
		{ 37, 509, 36 },
		{ 37, 510, 36 },
		{ 37, 511, 36 },
		{ 37, 512, 36 },
		{ 38, 508, 36 },
		{ 38, 509, 36 },
		{ 38, 510, 36 },
		{ 38, 511, 36 },
		{ 38, 512, 36 },
		{ 39, 507, 36 },
		{ 39, 508, 36 },
		{ 39, 509, 36 },
		{ 39, 510, 36 },
		{ 39, 511, 36 },
		{ 39, 512, 36 },
		{ 40, 506, 36 },
		{ 40, 507, 36 },
		{ 40, 508, 36 },
		{ 40, 509, 36 },
		{ 40, 510, 36 },
		{ 40, 511, 36 },
		{ 40, 512, 36 },
		{ 41, 504, 36 },
		{ 41, 505, 36 },
		{ 41, 506, 36 },
		{ 41, 507, 36 },
		{ 41, 508, 36 },
		{ 41, 509, 36 },
		{ 41, 510, 36 },
		{ 41, 511, 36 },
		{ 41, 512, 36 },
		{ 42, 503, 36 },
		{ 42, 504, 36 },
		{ 42, 505, 36 },
		{ 42, 506, 36 },
		{ 42, 507, 36 },
		{ 42, 508, 36 },
		{ 42, 509, 36 },
		{ 42, 510, 36 },
		{ 42, 511, 36 },
		{ 42, 512, 36 },
		{ 43, 502, 36 },
		{ 43, 503, 36 },
		{ 43, 504, 36 },
		{ 43, 505, 36 },
		{ 43, 506, 36 },
		{ 43, 507, 36 },
		{ 43, 508, 36 },
		{ 43, 509, 36 },
		{ 43, 510, 36 },
		{ 43, 511, 36 },
		{ 43, 512, 36 },
		{ 44, 501, 36 },
		{ 44, 502, 36 },
		{ 44, 503, 36 },
		{ 44, 504, 36 },
		{ 44, 505, 36 },
		{ 44, 506, 36 },
		{ 44, 507, 36 },
		{ 44, 508, 36 },
		{ 44, 509, 36 },
		{ 44, 510, 36 },
		{ 44, 511, 36 },
		{ 44, 512, 36 },
		{ 45, 500, 36 },
		{ 45, 501, 36 },
		{ 45, 502, 36 },
		{ 45, 503, 36 },
		{ 45, 504, 36 },
		{ 45, 505, 36 },
		{ 45, 506, 36 },
		{ 45, 507, 36 },
		{ 45, 508, 36 },
		{ 45, 509, 36 },
		{ 45, 510, 36 },
		{ 45, 511, 36 },
		{ 45, 512, 36 },
		{ 46, 499, 36 },
		{ 46, 500, 36 },
		{ 46, 501, 36 },
		{ 46, 502, 36 },
		{ 46, 503, 36 },
		{ 46, 504, 36 },
		{ 46, 505, 36 },
		{ 46, 506, 36 },
		{ 46, 507, 36 },
		{ 46, 508, 36 },
		{ 46, 509, 36 },
		{ 46, 510, 36 },
		{ 46, 511, 36 },
		{ 46, 512, 36 },
		{ 47, 498, 36 },
		{ 47, 499, 36 },
		{ 47, 500, 36 },
		{ 47, 501, 36 },
		{ 47, 502, 36 },
		{ 47, 503, 36 },
		{ 47, 504, 36 },
		{ 47, 505, 36 },
		{ 47, 506, 36 },
		{ 47, 507, 36 },
		{ 47, 508, 36 },
		{ 47, 509, 36 },
		{ 47, 510, 36 },
		{ 47, 511, 36 },
		{ 47, 512, 36 },
		{ 48, 496, 36 },
		{ 48, 497, 36 },
		{ 48, 498, 36 },
		{ 48, 499, 36 },
		{ 48, 500, 36 },
		{ 48, 501, 36 },
		{ 48, 502, 36 },
		{ 48, 503, 36 },
		{ 48, 504, 36 },
		{ 48, 505, 36 },
		{ 48, 506, 36 },
		{ 48, 507, 36 },
		{ 48, 508, 36 },
		{ 48, 509, 36 },
		{ 48, 510, 36 },
		{ 48, 511, 36 },
		{ 48, 512, 36 },
		{ 49, 495, 36 },
		{ 49, 496, 36 },
		{ 49, 497, 36 },
		{ 49, 498, 36 },
		{ 49, 499, 36 },
		{ 49, 500, 36 },
		{ 49, 501, 36 },
		{ 49, 502, 36 },
		{ 49, 503, 36 },
		{ 49, 504, 36 },
		{ 49, 505, 36 },
		{ 49, 506, 36 },
		{ 49, 507, 36 },
		{ 49, 508, 36 },
		{ 49, 509, 36 },
		{ 49, 510, 36 },
		{ 49, 511, 36 },
		{ 49, 512, 36 },
		{ 50, 494, 36 },
		{ 50, 495, 36 },
		{ 50, 496, 36 },
		{ 50, 497, 36 },
		{ 50, 498, 36 },
		{ 50, 499, 36 },
		{ 50, 500, 36 },
		{ 50, 501, 36 },
		{ 50, 502, 36 },
		{ 50, 503, 36 },
		{ 50, 504, 36 },
		{ 50, 505, 36 },
		{ 50, 506, 36 },
		{ 50, 507, 36 },
		{ 50, 508, 36 },
		{ 50, 509, 36 },
		{ 50, 510, 36 },
		{ 50, 511, 36 },
		{ 50, 512, 36 },
		{ 51, 493, 36 },
		{ 51, 494, 36 },
		{ 51, 495, 36 },
		{ 51, 496, 36 },
		{ 51, 497, 36 },
		{ 51, 498, 36 },
		{ 51, 499, 36 },
		{ 51, 500, 36 },
		{ 51, 501, 36 },
		{ 51, 502, 36 },
		{ 51, 503, 36 },
		{ 51, 504, 36 },
		{ 51, 505, 36 },
		{ 51, 506, 36 },
		{ 51, 507, 36 },
		{ 51, 508, 36 },
		{ 51, 509, 36 },
		{ 51, 510, 36 },
		{ 51, 511, 36 },
		{ 51, 512, 36 },
		{ 52, 492, 36 },
		{ 52, 493, 36 },
		{ 52, 494, 36 },
		{ 52, 495, 36 },
		{ 52, 496, 36 },
		{ 52, 497, 36 },
		{ 52, 498, 36 },
		{ 52, 499, 36 },
		{ 52, 500, 36 },
		{ 52, 501, 36 },
		{ 52, 502, 36 },
		{ 52, 503, 36 },
		{ 52, 504, 36 },
		{ 52, 505, 36 },
		{ 52, 506, 36 },
		{ 52, 507, 36 },
		{ 52, 508, 36 },
		{ 52, 509, 36 },
		{ 52, 510, 36 },
		{ 52, 511, 36 },
		{ 52, 512, 36 },
		{ 53, 491, 36 },
		{ 53, 492, 36 },
		{ 53, 493, 36 },
		{ 53, 494, 36 },
		{ 53, 495, 36 },
		{ 53, 496, 36 },
		{ 53, 497, 36 },
		{ 53, 498, 36 },
		{ 53, 499, 36 },
		{ 53, 500, 36 },
		{ 53, 501, 36 },
		{ 53, 502, 36 },
		{ 53, 503, 36 },
		{ 53, 504, 36 },
		{ 53, 505, 36 },
		{ 53, 506, 36 },
		{ 53, 507, 36 },
		{ 53, 508, 36 },
		{ 53, 509, 36 },
		{ 53, 510, 36 },
		{ 53, 511, 36 },
		{ 53, 512, 36 },
		{ 54, 490, 36 },
		{ 54, 491, 36 },
		{ 54, 492, 36 },
		{ 54, 493, 36 },
		{ 54, 494, 36 },
		{ 54, 495, 36 },
		{ 54, 496, 36 },
		{ 54, 497, 36 },
		{ 54, 498, 36 },
		{ 54, 499, 36 },
		{ 54, 500, 36 },
		{ 54, 501, 36 },
		{ 54, 502, 36 },
		{ 54, 503, 36 },
		{ 54, 504, 36 },
		{ 54, 505, 36 },
		{ 54, 506, 36 },
		{ 54, 507, 36 },
		{ 54, 508, 36 },
		{ 54, 509, 36 },
		{ 54, 510, 36 },
		{ 54, 511, 36 },
		{ 54, 512, 36 },
		{ 55, 488, 36 },
		{ 55, 489, 36 },
		{ 55, 490, 36 },
		{ 55, 491, 36 },
		{ 55, 492, 36 },
		{ 55, 493, 36 },
		{ 55, 494, 36 },
		{ 55, 495, 36 },
		{ 55, 496, 36 },
		{ 55, 497, 36 },
		{ 55, 498, 36 },
		{ 55, 499, 36 },
		{ 55, 500, 36 },
		{ 55, 501, 36 },
		{ 55, 502, 36 },
		{ 55, 503, 36 },
		{ 55, 504, 36 },
		{ 55, 505, 36 },
		{ 55, 506, 36 },
		{ 55, 507, 36 },
		{ 55, 508, 36 },
		{ 55, 509, 36 },
		{ 55, 510, 36 },
		{ 55, 511, 36 },
		{ 55, 512, 36 },
		{ 56, 487, 36 },
		{ 56, 488, 36 },
		{ 56, 489, 36 },
		{ 56, 490, 36 },
		{ 56, 491, 36 },
		{ 56, 492, 36 },
		{ 56, 493, 36 },
		{ 56, 494, 36 },
		{ 56, 495, 36 },
		{ 56, 496, 36 },
		{ 56, 497, 36 },
		{ 56, 498, 36 },
		{ 56, 499, 36 },
		{ 56, 500, 36 },
		{ 56, 501, 36 },
		{ 56, 502, 36 },
		{ 56, 503, 36 },
		{ 56, 504, 36 },
		{ 56, 505, 36 },
		{ 56, 506, 36 },
		{ 56, 507, 36 },
		{ 56, 508, 36 },
		{ 56, 509, 36 },
		{ 56, 510, 36 },
		{ 56, 511, 36 },
		{ 56, 512, 36 },
		{ 57, 486, 36 },
		{ 57, 487, 36 },
		{ 57, 488, 36 },
		{ 57, 489, 36 },
		{ 57, 490, 36 },
		{ 57, 491, 36 },
		{ 57, 492, 36 },
		{ 57, 493, 36 },
		{ 57, 494, 36 },
		{ 57, 495, 36 },
		{ 57, 496, 36 },
		{ 57, 497, 36 },
		{ 57, 498, 36 },
		{ 57, 499, 36 },
		{ 57, 500, 36 },
		{ 57, 501, 36 },
		{ 57, 502, 36 },
		{ 57, 503, 36 },
		{ 57, 504, 36 },
		{ 57, 505, 36 },
		{ 57, 506, 36 },
		{ 57, 507, 36 },
		{ 57, 508, 36 },
		{ 57, 509, 36 },
		{ 57, 510, 36 },
		{ 57, 511, 36 },
		{ 57, 512, 36 },
		{ 58, 485, 36 },
		{ 58, 486, 36 },
		{ 58, 487, 36 },
		{ 58, 488, 36 },
		{ 58, 489, 36 },
		{ 58, 490, 36 },
		{ 58, 491, 36 },
		{ 58, 492, 36 },
		{ 58, 493, 36 },
		{ 58, 494, 36 },
		{ 58, 495, 36 },
		{ 58, 496, 36 },
		{ 58, 497, 36 },
		{ 58, 498, 36 },
		{ 58, 499, 36 },
		{ 58, 500, 36 },
		{ 58, 501, 36 },
		{ 58, 502, 36 },
		{ 58, 503, 36 },
		{ 58, 504, 36 },
		{ 58, 505, 36 },
		{ 58, 506, 36 },
		{ 58, 507, 36 },
		{ 58, 508, 36 },
		{ 58, 509, 36 },
		{ 58, 510, 36 },
		{ 58, 511, 36 },
		{ 58, 512, 36 },
		{ 59, 484, 36 },
		{ 59, 485, 36 },
		{ 59, 486, 36 },
		{ 59, 487, 36 },
		{ 59, 488, 36 },
		{ 59, 489, 36 },
		{ 59, 490, 36 },
		{ 59, 491, 36 },
		{ 59, 492, 36 },
		{ 59, 493, 36 },
		{ 59, 494, 36 },
		{ 59, 495, 36 },
		{ 59, 496, 36 },
		{ 59, 497, 36 },
		{ 59, 498, 36 },
		{ 59, 499, 36 },
		{ 59, 500, 36 },
		{ 59, 501, 36 },
		{ 59, 502, 36 },
		{ 59, 503, 36 },
		{ 59, 504, 36 },
		{ 59, 505, 36 },
		{ 59, 506, 36 },
		{ 59, 507, 36 },
		{ 59, 508, 36 },
		{ 59, 509, 36 },
		{ 59, 510, 36 },
		{ 59, 511, 36 },
		{ 59, 512, 36 },
		{ 60, 483, 36 },
		{ 60, 484, 36 },
		{ 60, 485, 36 },
		{ 60, 486, 36 },
		{ 60, 487, 36 },
		{ 60, 488, 36 },
		{ 60, 489, 36 },
		{ 60, 490, 36 },
		{ 60, 491, 36 },
		{ 60, 492, 36 },
		{ 60, 493, 36 },
		{ 60, 494, 36 },
		{ 60, 495, 36 },
		{ 60, 496, 36 },
		{ 60, 497, 36 },
		{ 60, 498, 36 },
		{ 60, 499, 36 },
		{ 60, 500, 36 },
		{ 60, 501, 36 },
		{ 60, 502, 36 },
		{ 60, 503, 36 },
		{ 60, 504, 36 },
		{ 60, 505, 36 },
		{ 60, 506, 36 },
		{ 60, 507, 36 },
		{ 60, 508, 36 },
		{ 60, 509, 36 },
		{ 60, 510, 36 },
		{ 60, 511, 36 },
		{ 60, 512, 36 },
		{ 61, 482, 36 },
		{ 61, 483, 36 },
		{ 61, 484, 36 },
		{ 61, 485, 36 },
		{ 61, 486, 36 },
		{ 61, 487, 36 },
		{ 61, 488, 36 },
		{ 61, 489, 36 },
		{ 61, 490, 36 },
		{ 61, 491, 36 },
		{ 61, 492, 36 },
		{ 61, 493, 36 },
		{ 61, 494, 36 },
		{ 61, 495, 36 },
		{ 61, 496, 36 },
		{ 61, 497, 36 },
		{ 61, 498, 36 },
		{ 61, 499, 36 },
		{ 61, 500, 36 },
		{ 61, 501, 36 },
		{ 61, 502, 36 },
		{ 61, 503, 36 },
		{ 61, 504, 36 },
		{ 61, 505, 36 },
		{ 61, 506, 36 },
		{ 61, 507, 36 },
		{ 61, 508, 36 },
		{ 61, 509, 36 },
		{ 61, 510, 36 },
		{ 61, 511, 36 },
		{ 61, 512, 36 },
		{ 62, 480, 36 },
		{ 62, 481, 36 },
		{ 62, 482, 36 },
		{ 62, 483, 36 },
		{ 62, 484, 36 },
		{ 62, 485, 36 },
		{ 62, 486, 36 },
		{ 62, 487, 36 },
		{ 62, 488, 36 },
		{ 62, 489, 36 },
		{ 62, 490, 36 },
		{ 62, 491, 36 },
		{ 62, 492, 36 },
		{ 62, 493, 36 },
		{ 62, 494, 36 },
		{ 62, 495, 36 },
		{ 62, 496, 36 },
		{ 62, 497, 36 },
		{ 62, 498, 36 },
		{ 62, 499, 36 },
		{ 62, 500, 36 },
		{ 62, 501, 36 },
		{ 62, 502, 36 },
		{ 62, 503, 36 },
		{ 62, 504, 36 },
		{ 62, 505, 36 },
		{ 62, 506, 36 },
		{ 62, 507, 36 },
		{ 62, 508, 36 },
		{ 62, 509, 36 },
		{ 62, 510, 36 },
		{ 62, 511, 36 },
		{ 62, 512, 36 },
		{ 63, 479, 36 },
		{ 63, 480, 36 },
		{ 63, 481, 36 },
		{ 63, 482, 36 },
		{ 63, 483, 36 },
		{ 63, 484, 36 },
		{ 63, 485, 36 },
		{ 63, 486, 36 },
		{ 63, 487, 36 },
		{ 63, 488, 36 },
		{ 63, 489, 36 },
		{ 63, 490, 36 },
		{ 63, 491, 36 },
		{ 63, 492, 36 },
		{ 63, 493, 36 },
		{ 63, 494, 36 },
		{ 63, 495, 36 },
		{ 63, 496, 36 },
		{ 63, 497, 36 },
		{ 63, 498, 36 },
		{ 63, 499, 36 },
		{ 63, 500, 36 },
		{ 63, 501, 36 },
		{ 63, 502, 36 },
		{ 63, 503, 36 },
		{ 63, 504, 36 },
		{ 63, 505, 36 },
		{ 63, 506, 36 },
		{ 63, 507, 36 },
		{ 63, 508, 36 },
		{ 63, 509, 36 },
		{ 63, 510, 36 },
		{ 63, 511, 36 },
		{ 63, 512, 36 },
		{ 64, 478, 36 },
		{ 64, 479, 36 },
		{ 64, 480, 36 },
		{ 64, 481, 36 },
		{ 64, 482, 36 },
		{ 64, 483, 36 },
		{ 64, 484, 36 },
		{ 64, 485, 36 },
		{ 64, 486, 36 },
		{ 64, 487, 36 },
		{ 64, 488, 36 },
		{ 64, 489, 36 },
		{ 64, 490, 36 },
		{ 64, 491, 36 },
		{ 64, 492, 36 },
		{ 64, 493, 36 },
		{ 64, 494, 36 },
		{ 64, 495, 36 },
		{ 64, 496, 36 },
		{ 64, 497, 36 },
		{ 64, 498, 36 },
		{ 64, 499, 36 },
		{ 64, 500, 36 },
		{ 64, 501, 36 },
		{ 64, 502, 36 },
		{ 64, 503, 36 },
		{ 64, 504, 36 },
		{ 64, 505, 36 },
		{ 64, 506, 36 },
		{ 64, 507, 36 },
		{ 64, 508, 36 },
		{ 64, 509, 36 },
		{ 64, 510, 36 },
		{ 64, 511, 36 },
		{ 64, 512, 36 },
		{ 65, 477, 36 },
		{ 65, 478, 36 },
		{ 65, 479, 36 },
		{ 65, 480, 36 },
		{ 65, 481, 36 },
		{ 65, 482, 36 },
		{ 65, 483, 36 },
		{ 65, 484, 36 },
		{ 65, 485, 36 },
		{ 65, 486, 36 },
		{ 65, 487, 36 },
		{ 65, 488, 36 },
		{ 65, 489, 36 },
		{ 65, 490, 36 },
		{ 65, 491, 36 },
		{ 65, 492, 36 },
		{ 65, 493, 36 },
		{ 65, 494, 36 },
		{ 65, 495, 36 },
		{ 65, 496, 36 },
		{ 65, 497, 36 },
		{ 65, 498, 36 },
		{ 65, 499, 36 },
		{ 65, 500, 36 },
		{ 65, 501, 36 },
		{ 65, 502, 36 },
		{ 65, 503, 36 },
		{ 65, 504, 36 },
		{ 65, 505, 36 },
		{ 65, 506, 36 },
		{ 65, 507, 36 },
		{ 65, 508, 36 },
		{ 65, 509, 36 },
		{ 65, 510, 36 },
		{ 65, 511, 36 },
		{ 65, 512, 36 },
		{ 66, 476, 36 },
		{ 66, 477, 36 },
		{ 66, 478, 36 },
		{ 66, 479, 36 },
		{ 66, 480, 36 },
		{ 66, 481, 36 },
		{ 66, 482, 36 },
		{ 66, 483, 36 },
		{ 66, 484, 36 },
		{ 66, 485, 36 },
		{ 66, 486, 36 },
		{ 66, 487, 36 },
		{ 66, 488, 36 },
		{ 66, 489, 36 },
		{ 66, 490, 36 },
		{ 66, 491, 36 },
		{ 66, 492, 36 },
		{ 66, 493, 36 },
		{ 66, 494, 36 },
		{ 66, 495, 36 },
		{ 66, 496, 36 },
		{ 66, 497, 36 },
		{ 66, 498, 36 },
		{ 66, 499, 36 },
		{ 66, 500, 36 },
		{ 66, 501, 36 },
		{ 66, 502, 36 },
		{ 66, 503, 36 },
		{ 66, 504, 36 },
		{ 66, 505, 36 },
		{ 66, 506, 36 },
		{ 66, 507, 36 },
		{ 66, 508, 36 },
		{ 66, 509, 36 },
		{ 66, 510, 36 },
		{ 66, 511, 36 },
		{ 66, 512, 36 },
		{ 67, 475, 36 },
		{ 67, 476, 36 },
		{ 67, 477, 36 },
		{ 67, 478, 36 },
		{ 67, 479, 36 },
		{ 67, 480, 36 },
		{ 67, 481, 36 },
		{ 67, 482, 36 },
		{ 67, 483, 36 },
		{ 67, 484, 36 },
		{ 67, 485, 36 },
		{ 67, 486, 36 },
		{ 67, 487, 36 },
		{ 67, 488, 36 },
		{ 67, 489, 36 },
		{ 67, 490, 36 },
		{ 67, 491, 36 },
		{ 67, 492, 36 },
		{ 67, 493, 36 },
		{ 67, 494, 36 },
		{ 67, 495, 36 },
		{ 67, 496, 36 },
		{ 67, 497, 36 },
		{ 67, 498, 36 },
		{ 67, 499, 36 },
		{ 67, 500, 36 },
		{ 67, 501, 36 },
		{ 67, 502, 36 },
		{ 67, 503, 36 },
		{ 67, 504, 36 },
		{ 67, 505, 36 },
		{ 67, 506, 36 },
		{ 67, 507, 36 },
		{ 67, 508, 36 },
		{ 67, 509, 36 },
		{ 67, 510, 36 },
		{ 67, 511, 36 },
		{ 67, 512, 36 },
		{ 68, 474, 36 },
		{ 68, 475, 36 },
		{ 68, 476, 36 },
		{ 68, 477, 36 },
		{ 68, 478, 36 },
		{ 68, 479, 36 },
		{ 68, 480, 36 },
		{ 68, 481, 36 },
		{ 68, 482, 36 },
		{ 68, 483, 36 },
		{ 68, 484, 36 },
		{ 68, 485, 36 },
		{ 68, 486, 36 },
		{ 68, 487, 36 },
		{ 68, 488, 36 },
		{ 68, 489, 36 },
		{ 68, 490, 36 },
		{ 68, 491, 36 },
		{ 68, 492, 36 },
		{ 68, 493, 36 },
		{ 68, 494, 36 },
		{ 68, 495, 36 },
		{ 68, 496, 36 },
		{ 68, 497, 36 },
		{ 68, 498, 36 },
		{ 68, 499, 36 },
		{ 68, 500, 36 },
		{ 68, 501, 36 },
		{ 68, 502, 36 },
		{ 68, 503, 36 },
		{ 68, 504, 36 },
		{ 68, 505, 36 },
		{ 68, 506, 36 },
		{ 68, 507, 36 },
		{ 68, 508, 36 },
		{ 68, 509, 36 },
		{ 68, 510, 36 },
		{ 68, 511, 36 },
		{ 68, 512, 36 },
		{ 69, 472, 36 },
		{ 69, 473, 36 },
		{ 69, 474, 36 },
		{ 69, 475, 36 },
		{ 69, 476, 36 },
		{ 69, 477, 36 },
		{ 69, 478, 36 },
		{ 69, 479, 36 },
		{ 69, 480, 36 },
		{ 69, 481, 36 },
		{ 69, 482, 36 },
		{ 69, 483, 36 },
		{ 69, 484, 36 },
		{ 69, 485, 36 },
		{ 69, 486, 36 },
		{ 69, 487, 36 },
		{ 69, 488, 36 },
		{ 69, 489, 36 },
		{ 69, 490, 36 },
		{ 69, 491, 36 },
		{ 69, 492, 36 },
		{ 69, 493, 36 },
		{ 69, 494, 36 },
		{ 69, 495, 36 },
		{ 69, 496, 36 },
		{ 69, 497, 36 },
		{ 69, 498, 36 },
		{ 69, 499, 36 },
		{ 69, 500, 36 },
		{ 69, 501, 36 },
		{ 69, 502, 36 },
		{ 69, 503, 36 },
		{ 69, 504, 36 },
		{ 69, 505, 36 },
		{ 69, 506, 36 },
		{ 69, 507, 36 },
		{ 69, 508, 36 },
		{ 69, 509, 36 },
		{ 69, 510, 36 },
		{ 69, 511, 36 },
		{ 69, 512, 36 },
		{ 70, 471, 36 },
		{ 70, 472, 36 },
		{ 70, 473, 36 },
		{ 70, 474, 36 },
		{ 70, 475, 36 },
		{ 70, 476, 36 },
		{ 70, 477, 36 },
		{ 70, 478, 36 },
		{ 70, 479, 36 },
		{ 70, 480, 36 },
		{ 70, 481, 36 },
		{ 70, 482, 36 },
		{ 70, 483, 36 },
		{ 70, 484, 36 },
		{ 70, 485, 36 },
		{ 70, 486, 36 },
		{ 70, 487, 36 },
		{ 70, 488, 36 },
		{ 70, 489, 36 },
		{ 70, 490, 36 },
		{ 70, 491, 36 },
		{ 70, 492, 36 },
		{ 70, 493, 36 },
		{ 70, 494, 36 },
		{ 70, 495, 36 },
		{ 70, 496, 36 },
		{ 70, 497, 36 },
		{ 70, 498, 36 },
		{ 70, 499, 36 },
		{ 70, 500, 36 },
		{ 70, 501, 36 },
		{ 70, 502, 36 },
		{ 70, 503, 36 },
		{ 70, 504, 36 },
		{ 70, 505, 36 },
		{ 70, 506, 36 },
		{ 70, 507, 36 },
		{ 70, 508, 36 },
		{ 70, 509, 36 },
		{ 70, 510, 36 },
		{ 70, 511, 36 },
		{ 70, 512, 36 },
		{ 71, 470, 36 },
		{ 71, 471, 36 },
		{ 71, 472, 36 },
		{ 71, 473, 36 },
		{ 71, 474, 36 },
		{ 71, 475, 36 },
		{ 71, 476, 36 },
		{ 71, 477, 36 },
		{ 71, 478, 36 },
		{ 71, 479, 36 },
		{ 71, 480, 36 },
		{ 71, 481, 36 },
		{ 71, 482, 36 },
		{ 71, 483, 36 },
		{ 71, 484, 36 },
		{ 71, 485, 36 },
		{ 71, 486, 36 },
		{ 71, 487, 36 },
		{ 71, 488, 36 },
		{ 71, 489, 36 },
		{ 71, 490, 36 },
		{ 71, 491, 36 },
		{ 71, 492, 36 },
		{ 71, 493, 36 },
		{ 71, 494, 36 },
		{ 71, 495, 36 },
		{ 71, 496, 36 },
		{ 71, 497, 36 },
		{ 71, 498, 36 },
		{ 71, 499, 36 },
		{ 71, 500, 36 },
		{ 71, 501, 36 },
		{ 71, 502, 36 },
		{ 71, 503, 36 },
		{ 71, 504, 36 },
		{ 71, 505, 36 },
		{ 71, 506, 36 },
		{ 71, 507, 36 },
		{ 71, 508, 36 },
		{ 71, 509, 36 },
		{ 71, 510, 36 },
		{ 71, 511, 36 },
		{ 71, 512, 36 },
		{ 72, 469, 36 },
		{ 72, 470, 36 },
		{ 72, 471, 36 },
		{ 72, 472, 36 },
		{ 72, 473, 36 },
		{ 72, 474, 36 },
		{ 72, 475, 36 },
		{ 72, 476, 36 },
		{ 72, 477, 36 },
		{ 72, 478, 36 },
		{ 72, 479, 36 },
		{ 72, 480, 36 },
		{ 72, 481, 36 },
		{ 72, 482, 36 },
		{ 72, 483, 36 },
		{ 72, 484, 36 },
		{ 72, 485, 36 },
		{ 72, 486, 36 },
		{ 72, 487, 36 },
		{ 72, 488, 36 },
		{ 72, 489, 36 },
		{ 72, 490, 36 },
		{ 72, 491, 36 },
		{ 72, 492, 36 },
		{ 72, 493, 36 },
		{ 72, 494, 36 },
		{ 72, 495, 36 },
		{ 72, 496, 36 },
		{ 72, 497, 36 },
		{ 72, 498, 36 },
		{ 72, 499, 36 },
		{ 72, 500, 36 },
		{ 72, 501, 36 },
		{ 72, 502, 36 },
		{ 72, 503, 36 },
		{ 72, 504, 36 },
		{ 72, 505, 36 },
		{ 72, 506, 36 },
		{ 72, 507, 36 },
		{ 72, 508, 36 },
		{ 72, 509, 36 },
		{ 72, 510, 36 },
		{ 72, 511, 36 },
		{ 72, 512, 36 },
		{ 73, 468, 36 },
		{ 73, 469, 36 },
		{ 73, 470, 36 },
		{ 73, 471, 36 },
		{ 73, 472, 36 },
		{ 73, 473, 36 },
		{ 73, 474, 36 },
		{ 73, 475, 36 },
		{ 73, 476, 36 },
		{ 73, 477, 36 },
		{ 73, 478, 36 },
		{ 73, 479, 36 },
		{ 73, 480, 36 },
		{ 73, 481, 36 },
		{ 73, 482, 36 },
		{ 73, 483, 36 },
		{ 73, 484, 36 },
		{ 73, 485, 36 },
		{ 73, 486, 36 },
		{ 73, 487, 36 },
		{ 73, 488, 36 },
		{ 73, 489, 36 },
		{ 73, 490, 36 },
		{ 73, 491, 36 },
		{ 73, 492, 36 },
		{ 73, 493, 36 },
		{ 73, 494, 36 },
		{ 73, 495, 36 },
		{ 73, 496, 36 },
		{ 73, 497, 36 },
		{ 73, 498, 36 },
		{ 73, 499, 36 },
		{ 73, 500, 36 },
		{ 73, 501, 36 },
		{ 73, 502, 36 },
		{ 73, 503, 36 },
		{ 73, 504, 36 },
		{ 73, 505, 36 },
		{ 73, 506, 36 },
		{ 73, 507, 36 },
		{ 73, 508, 36 },
		{ 73, 509, 36 },
		{ 73, 510, 36 },
		{ 73, 511, 36 },
		{ 73, 512, 36 },
		{ 74, 467, 36 },
		{ 74, 468, 36 },
		{ 74, 469, 36 },
		{ 74, 470, 36 },
		{ 74, 471, 36 },
		{ 74, 472, 36 },
		{ 74, 473, 36 },
		{ 74, 474, 36 },
		{ 74, 475, 36 },
		{ 74, 476, 36 },
		{ 74, 477, 36 },
		{ 74, 478, 36 },
		{ 74, 479, 36 },
		{ 74, 480, 36 },
		{ 74, 481, 36 },
		{ 74, 482, 36 },
		{ 74, 483, 36 },
		{ 74, 484, 36 },
		{ 74, 485, 36 },
		{ 74, 486, 36 },
		{ 74, 487, 36 },
		{ 74, 488, 36 },
		{ 74, 489, 36 },
		{ 74, 490, 36 },
		{ 74, 491, 36 },
		{ 74, 492, 36 },
		{ 74, 493, 36 },
		{ 74, 494, 36 },
		{ 74, 495, 36 },
		{ 74, 496, 36 },
		{ 74, 497, 36 },
		{ 74, 498, 36 },
		{ 74, 499, 36 },
		{ 74, 500, 36 },
		{ 74, 501, 36 },
		{ 74, 502, 36 },
		{ 74, 503, 36 },
		{ 74, 504, 36 },
		{ 74, 505, 36 },
		{ 74, 506, 36 },
		{ 74, 507, 36 },
		{ 74, 508, 36 },
		{ 74, 509, 36 },
		{ 74, 510, 36 },
		{ 74, 511, 36 },
		{ 74, 512, 36 },
		{ 75, 466, 36 },
		{ 75, 467, 36 },
		{ 75, 468, 36 },
		{ 75, 469, 36 },
		{ 75, 470, 36 },
		{ 75, 471, 36 },
		{ 75, 472, 36 },
		{ 75, 473, 36 },
		{ 75, 474, 36 },
		{ 75, 475, 36 },
		{ 75, 476, 36 },
		{ 75, 477, 36 },
		{ 75, 478, 36 },
		{ 75, 479, 36 },
		{ 75, 480, 36 },
		{ 75, 481, 36 },
		{ 75, 482, 36 },
		{ 75, 483, 36 },
		{ 75, 484, 36 },
		{ 75, 485, 36 },
		{ 75, 486, 36 },
		{ 75, 487, 36 },
		{ 75, 488, 36 },
		{ 75, 489, 36 },
		{ 75, 490, 36 },
		{ 75, 491, 36 },
		{ 75, 492, 36 },
		{ 75, 493, 36 },
		{ 75, 494, 36 },
		{ 75, 495, 36 },
		{ 75, 496, 36 },
		{ 75, 497, 36 },
		{ 75, 498, 36 },
		{ 75, 499, 36 },
		{ 75, 500, 36 },
		{ 75, 501, 36 },
		{ 75, 502, 36 },
		{ 75, 503, 36 },
		{ 75, 504, 36 },
		{ 75, 505, 36 },
		{ 75, 506, 36 },
		{ 75, 507, 36 },
		{ 75, 508, 36 },
		{ 75, 509, 36 },
		{ 75, 510, 36 },
		{ 75, 511, 36 },
		{ 75, 512, 36 },
		{ 76, 464, 36 },
		{ 76, 465, 36 },
		{ 76, 466, 36 },
		{ 76, 467, 36 },
		{ 76, 468, 36 },
		{ 76, 469, 36 },
		{ 76, 470, 36 },
		{ 76, 471, 36 },
		{ 76, 472, 36 },
		{ 76, 473, 36 },
		{ 76, 474, 36 },
		{ 76, 475, 36 },
		{ 76, 476, 36 },
		{ 76, 477, 36 },
		{ 76, 478, 36 },
		{ 76, 479, 36 },
		{ 76, 480, 36 },
		{ 76, 481, 36 },
		{ 76, 482, 36 },
		{ 76, 483, 36 },
		{ 76, 484, 36 },
		{ 76, 485, 36 },
		{ 76, 486, 36 },
		{ 76, 487, 36 },
		{ 76, 488, 36 },
		{ 76, 489, 36 },
		{ 76, 490, 36 },
		{ 76, 491, 36 },
		{ 76, 492, 36 },
		{ 76, 493, 36 },
		{ 76, 494, 36 },
		{ 76, 495, 36 },
		{ 76, 496, 36 },
		{ 76, 497, 36 },
		{ 76, 498, 36 },
		{ 76, 499, 36 },
		{ 76, 500, 36 },
		{ 76, 501, 36 },
		{ 76, 502, 36 },
		{ 76, 503, 36 },
		{ 76, 504, 36 },
		{ 76, 505, 36 },
		{ 76, 506, 36 },
		{ 76, 507, 36 },
		{ 76, 508, 36 },
		{ 76, 509, 36 },
		{ 76, 510, 36 },
		{ 76, 511, 36 },
		{ 76, 512, 36 },
		{ 77, 463, 36 },
		{ 77, 464, 36 },
		{ 77, 465, 36 },
		{ 77, 466, 36 },
		{ 77, 467, 36 },
		{ 77, 468, 36 },
		{ 77, 469, 36 },
		{ 77, 470, 36 },
		{ 77, 471, 36 },
		{ 77, 472, 36 },
		{ 77, 473, 36 },
		{ 77, 474, 36 },
		{ 77, 475, 36 },
		{ 77, 476, 36 },
		{ 77, 477, 36 },
		{ 77, 478, 36 },
		{ 77, 479, 36 },
		{ 77, 480, 36 },
		{ 77, 481, 36 },
		{ 77, 482, 36 },
		{ 77, 483, 36 },
		{ 77, 484, 36 },
		{ 77, 485, 36 },
		{ 77, 486, 36 },
		{ 77, 487, 36 },
		{ 77, 488, 36 },
		{ 77, 489, 36 },
		{ 77, 490, 36 },
		{ 77, 491, 36 },
		{ 77, 492, 36 },
		{ 77, 493, 36 },
		{ 77, 494, 36 },
		{ 77, 495, 36 },
		{ 77, 496, 36 },
		{ 77, 497, 36 },
		{ 77, 498, 36 },
		{ 77, 499, 36 },
		{ 77, 500, 36 },
		{ 77, 501, 36 },
		{ 77, 502, 36 },
		{ 77, 503, 36 },
		{ 77, 504, 36 },
		{ 77, 505, 36 },
		{ 77, 506, 36 },
		{ 77, 507, 36 },
		{ 77, 508, 36 },
		{ 77, 509, 36 },
		{ 77, 510, 36 },
		{ 77, 511, 36 },
		{ 77, 512, 36 },
		{ 78, 462, 36 },
		{ 78, 463, 36 },
		{ 78, 464, 36 },
		{ 78, 465, 36 },
		{ 78, 466, 36 },
		{ 78, 467, 36 },
		{ 78, 468, 36 },
		{ 78, 469, 36 },
		{ 78, 470, 36 },
		{ 78, 471, 36 },
		{ 78, 472, 36 },
		{ 78, 473, 36 },
		{ 78, 474, 36 },
		{ 78, 475, 36 },
		{ 78, 476, 36 },
		{ 78, 477, 36 },
		{ 78, 478, 36 },
		{ 78, 479, 36 },
		{ 78, 480, 36 },
		{ 78, 481, 36 },
		{ 78, 482, 36 },
		{ 78, 483, 36 },
		{ 78, 484, 36 },
		{ 78, 485, 36 },
		{ 78, 486, 36 },
		{ 78, 487, 36 },
		{ 78, 488, 36 },
		{ 78, 489, 36 },
		{ 78, 490, 36 },
		{ 78, 491, 36 },
		{ 78, 492, 36 },
		{ 78, 493, 36 },
		{ 78, 494, 36 },
		{ 78, 495, 36 },
		{ 78, 496, 36 },
		{ 78, 497, 36 },
		{ 78, 498, 36 },
		{ 78, 499, 36 },
		{ 78, 500, 36 },
		{ 78, 501, 36 },
		{ 78, 502, 36 },
		{ 78, 503, 36 },
		{ 78, 504, 36 },
		{ 78, 505, 36 },
		{ 78, 506, 36 },
		{ 78, 507, 36 },
		{ 78, 508, 36 },
		{ 78, 509, 36 },
		{ 78, 510, 36 },
		{ 78, 511, 36 },
		{ 78, 512, 36 },
		{ 79, 461, 36 },
		{ 79, 462, 36 },
		{ 79, 463, 36 },
		{ 79, 464, 36 },
		{ 79, 465, 36 },
		{ 79, 466, 36 },
		{ 79, 467, 36 },
		{ 79, 468, 36 },
		{ 79, 469, 36 },
		{ 79, 470, 36 },
		{ 79, 471, 36 },
		{ 79, 472, 36 },
		{ 79, 473, 36 },
		{ 79, 474, 36 },
		{ 79, 475, 36 },
		{ 79, 476, 36 },
		{ 79, 477, 36 },
		{ 79, 478, 36 },
		{ 79, 479, 36 },
		{ 79, 480, 36 },
		{ 79, 481, 36 },
		{ 79, 482, 36 },
		{ 79, 483, 36 },
		{ 79, 484, 36 },
		{ 79, 485, 36 },
		{ 79, 486, 36 },
		{ 79, 487, 36 },
		{ 79, 488, 36 },
		{ 79, 489, 36 },
		{ 79, 490, 36 },
		{ 79, 491, 36 },
		{ 79, 492, 36 },
		{ 79, 493, 36 },
		{ 79, 494, 36 },
		{ 79, 495, 36 },
		{ 79, 496, 36 },
		{ 79, 497, 36 },
		{ 79, 498, 36 },
		{ 79, 499, 36 },
		{ 79, 500, 36 },
		{ 79, 501, 36 },
		{ 79, 502, 36 },
		{ 79, 503, 36 },
		{ 79, 504, 36 },
		{ 79, 505, 36 },
		{ 79, 506, 36 },
		{ 79, 507, 36 },
		{ 79, 508, 36 },
		{ 79, 509, 36 },
		{ 79, 510, 36 },
		{ 79, 511, 36 },
		{ 79, 512, 36 },
		{ 80, 460, 36 },
		{ 80, 461, 36 },
		{ 80, 462, 36 },
		{ 80, 463, 36 },
		{ 80, 464, 36 },
		{ 80, 465, 36 },
		{ 80, 466, 36 },
		{ 80, 467, 36 },
		{ 80, 468, 36 },
		{ 80, 469, 36 },
		{ 80, 470, 36 },
		{ 80, 471, 36 },
		{ 80, 472, 36 },
		{ 80, 473, 36 },
		{ 80, 474, 36 },
		{ 80, 475, 36 },
		{ 80, 476, 36 },
		{ 80, 477, 36 },
		{ 80, 478, 36 },
		{ 80, 479, 36 },
		{ 80, 480, 36 },
		{ 80, 481, 36 },
		{ 80, 482, 36 },
		{ 80, 483, 36 },
		{ 80, 484, 36 },
		{ 80, 485, 36 },
		{ 80, 486, 36 },
		{ 80, 487, 36 },
		{ 80, 488, 36 },
		{ 80, 489, 36 },
		{ 80, 490, 36 },
		{ 80, 491, 36 },
		{ 80, 492, 36 },
		{ 80, 493, 36 },
		{ 80, 494, 36 },
		{ 80, 495, 36 },
		{ 80, 496, 36 },
		{ 80, 497, 36 },
		{ 80, 498, 36 },
		{ 80, 499, 36 },
		{ 80, 500, 36 },
		{ 80, 501, 36 },
		{ 80, 502, 36 },
		{ 80, 503, 36 },
		{ 80, 504, 36 },
		{ 80, 505, 36 },
		{ 80, 506, 36 },
		{ 80, 507, 36 },
		{ 80, 508, 36 },
		{ 80, 509, 36 },
		{ 80, 510, 36 },
		{ 80, 511, 36 },
		{ 80, 512, 36 },
		{ 81, 459, 36 },
		{ 81, 460, 36 },
		{ 81, 461, 36 },
		{ 81, 462, 36 },
		{ 81, 463, 36 },
		{ 81, 464, 36 },
		{ 81, 465, 36 },
		{ 81, 466, 36 },
		{ 81, 467, 36 },
		{ 81, 468, 36 },
		{ 81, 469, 36 },
		{ 81, 470, 36 },
		{ 81, 471, 36 },
		{ 81, 472, 36 },
		{ 81, 473, 36 },
		{ 81, 474, 36 },
		{ 81, 475, 36 },
		{ 81, 476, 36 },
		{ 81, 477, 36 },
		{ 81, 478, 36 },
		{ 81, 479, 36 },
		{ 81, 480, 36 },
		{ 81, 481, 36 },
		{ 81, 482, 36 },
		{ 81, 483, 36 },
		{ 81, 484, 36 },
		{ 81, 485, 36 },
		{ 81, 486, 36 },
		{ 81, 487, 36 },
		{ 81, 488, 36 },
		{ 81, 489, 36 },
		{ 81, 490, 36 },
		{ 81, 491, 36 },
		{ 81, 492, 36 },
		{ 81, 493, 36 },
		{ 81, 494, 36 },
		{ 81, 495, 36 },
		{ 81, 496, 36 },
		{ 81, 497, 36 },
		{ 81, 498, 36 },
		{ 81, 499, 36 },
		{ 81, 500, 36 },
		{ 81, 501, 36 },
		{ 81, 502, 36 },
		{ 81, 503, 36 },
		{ 81, 504, 36 },
		{ 81, 505, 36 },
		{ 81, 506, 36 },
		{ 81, 507, 36 },
		{ 81, 508, 36 },
		{ 81, 509, 36 },
		{ 81, 510, 36 },
		{ 81, 511, 36 },
		{ 81, 512, 36 },
		{ 82, 458, 36 },
		{ 82, 459, 36 },
		{ 82, 460, 36 },
		{ 82, 461, 36 },
		{ 82, 462, 36 },
		{ 82, 463, 36 },
		{ 82, 464, 36 },
		{ 82, 465, 36 },
		{ 82, 466, 36 },
		{ 82, 467, 36 },
		{ 82, 468, 36 },
		{ 82, 469, 36 },
		{ 82, 470, 36 },
		{ 82, 471, 36 },
		{ 82, 472, 36 },
		{ 82, 473, 36 },
		{ 82, 474, 36 },
		{ 82, 475, 36 },
		{ 82, 476, 36 },
		{ 82, 477, 36 },
		{ 82, 478, 36 },
		{ 82, 479, 36 },
		{ 82, 480, 36 },
		{ 82, 481, 36 },
		{ 82, 482, 36 },
		{ 82, 483, 36 },
		{ 82, 484, 36 },
		{ 82, 485, 36 },
		{ 82, 486, 36 },
		{ 82, 487, 36 },
		{ 82, 488, 36 },
		{ 82, 489, 36 },
		{ 82, 490, 36 },
		{ 82, 491, 36 },
		{ 82, 492, 36 },
		{ 82, 493, 36 },
		{ 82, 494, 36 },
		{ 82, 495, 36 },
		{ 82, 496, 36 },
		{ 82, 497, 36 },
		{ 82, 498, 36 },
		{ 82, 499, 36 },
		{ 82, 500, 36 },
		{ 82, 501, 36 },
		{ 82, 502, 36 },
		{ 82, 503, 36 },
		{ 82, 504, 36 },
		{ 82, 505, 36 },
		{ 82, 506, 36 },
		{ 82, 507, 36 },
		{ 82, 508, 36 },
		{ 82, 509, 36 },
		{ 82, 510, 36 },
		{ 82, 511, 36 },
		{ 82, 512, 36 },
		{ 83, 456, 36 },
		{ 83, 457, 36 },
		{ 83, 458, 36 },
		{ 83, 459, 36 },
		{ 83, 460, 36 },
		{ 83, 461, 36 },
		{ 83, 462, 36 },
		{ 83, 463, 36 },
		{ 83, 464, 36 },
		{ 83, 465, 36 },
		{ 83, 466, 36 },
		{ 83, 467, 36 },
		{ 83, 468, 36 },
		{ 83, 469, 36 },
		{ 83, 470, 36 },
		{ 83, 471, 36 },
		{ 83, 472, 36 },
		{ 83, 473, 36 },
		{ 83, 474, 36 },
		{ 83, 475, 36 },
		{ 83, 476, 36 },
		{ 83, 477, 36 },
		{ 83, 478, 36 },
		{ 83, 479, 36 },
		{ 83, 480, 36 },
		{ 83, 481, 36 },
		{ 83, 482, 36 },
		{ 83, 483, 36 },
		{ 83, 484, 36 },
		{ 83, 485, 36 },
		{ 83, 486, 36 },
		{ 83, 487, 36 },
		{ 83, 488, 36 },
		{ 83, 489, 36 },
		{ 83, 490, 36 },
		{ 83, 491, 36 },
		{ 83, 492, 36 },
		{ 83, 493, 36 },
		{ 83, 494, 36 },
		{ 83, 495, 36 },
		{ 83, 496, 36 },
		{ 83, 497, 36 },
		{ 83, 498, 36 },
		{ 83, 499, 36 },
		{ 83, 500, 36 },
		{ 83, 501, 36 },
		{ 83, 502, 36 },
		{ 83, 503, 36 },
		{ 83, 504, 36 },
		{ 83, 505, 36 },
		{ 83, 506, 36 },
		{ 83, 507, 36 },
		{ 83, 508, 36 },
		{ 83, 509, 36 },
		{ 83, 510, 36 },
		{ 83, 511, 36 },
		{ 83, 512, 36 },
		{ 84, 455, 36 },
		{ 84, 456, 36 },
		{ 84, 457, 36 },
		{ 84, 458, 36 },
		{ 84, 459, 36 },
		{ 84, 460, 36 },
		{ 84, 461, 36 },
		{ 84, 462, 36 },
		{ 84, 463, 36 },
		{ 84, 464, 36 },
		{ 84, 465, 36 },
		{ 84, 466, 36 },
		{ 84, 467, 36 },
		{ 84, 468, 36 },
		{ 84, 469, 36 },
		{ 84, 470, 36 },
		{ 84, 471, 36 },
		{ 84, 472, 36 },
		{ 84, 473, 36 },
		{ 84, 474, 36 },
		{ 84, 475, 36 },
		{ 84, 476, 36 },
		{ 84, 477, 36 },
		{ 84, 478, 36 },
		{ 84, 479, 36 },
		{ 84, 480, 36 },
		{ 84, 481, 36 },
		{ 84, 482, 36 },
		{ 84, 483, 36 },
		{ 84, 484, 36 },
		{ 84, 485, 36 },
		{ 84, 486, 36 },
		{ 84, 487, 36 },
		{ 84, 488, 36 },
		{ 84, 489, 36 },
		{ 84, 490, 36 },
		{ 84, 491, 36 },
		{ 84, 492, 36 },
		{ 84, 493, 36 },
		{ 84, 494, 36 },
		{ 84, 495, 36 },
		{ 84, 496, 36 },
		{ 84, 497, 36 },
		{ 84, 498, 36 },
		{ 84, 499, 36 },
		{ 84, 500, 36 },
		{ 84, 501, 36 },
		{ 84, 502, 36 },
		{ 84, 503, 36 },
		{ 84, 504, 36 },
		{ 84, 505, 36 },
		{ 84, 506, 36 },
		{ 84, 507, 36 },
		{ 84, 508, 36 },
		{ 84, 509, 36 },
		{ 84, 510, 36 },
		{ 84, 511, 36 },
		{ 84, 512, 36 },
		{ 85, 454, 36 },
		{ 85, 455, 36 },
		{ 85, 456, 36 },
		{ 85, 457, 36 },
		{ 85, 458, 36 },
		{ 85, 459, 36 },
		{ 85, 460, 36 },
		{ 85, 461, 36 },
		{ 85, 462, 36 },
		{ 85, 463, 36 },
		{ 85, 464, 36 },
		{ 85, 465, 36 },
		{ 85, 466, 36 },
		{ 85, 467, 36 },
		{ 85, 468, 36 },
		{ 85, 469, 36 },
		{ 85, 470, 36 },
		{ 85, 471, 36 },
		{ 85, 472, 36 },
		{ 85, 473, 36 },
		{ 85, 474, 36 },
		{ 85, 475, 36 },
		{ 85, 476, 36 },
		{ 85, 477, 36 },
		{ 85, 478, 36 },
		{ 85, 479, 36 },
		{ 85, 480, 36 },
		{ 85, 481, 36 },
		{ 85, 482, 36 },
		{ 85, 483, 36 },
		{ 85, 484, 36 },
		{ 85, 485, 36 },
		{ 85, 486, 36 },
		{ 85, 487, 36 },
		{ 85, 488, 36 },
		{ 85, 489, 36 },
		{ 85, 490, 36 },
		{ 85, 491, 36 },
		{ 85, 492, 36 },
		{ 85, 493, 36 },
		{ 85, 494, 36 },
		{ 85, 495, 36 },
		{ 85, 496, 36 },
		{ 85, 497, 36 },
		{ 85, 498, 36 },
		{ 85, 499, 36 },
		{ 85, 500, 36 },
		{ 85, 501, 36 },
		{ 85, 502, 36 },
		{ 85, 503, 36 },
		{ 85, 504, 36 },
		{ 85, 505, 36 },
		{ 85, 506, 36 },
		{ 85, 507, 36 },
		{ 85, 508, 36 },
		{ 85, 509, 36 },
		{ 85, 510, 36 },
		{ 85, 511, 36 },
		{ 85, 512, 36 },
		{ 86, 453, 36 },
		{ 86, 454, 36 },
		{ 86, 455, 36 },
		{ 86, 456, 36 },
		{ 86, 457, 36 },
		{ 86, 458, 36 },
		{ 86, 459, 36 },
		{ 86, 460, 36 },
		{ 86, 461, 36 },
		{ 86, 462, 36 },
		{ 86, 463, 36 },
		{ 86, 464, 36 },
		{ 86, 465, 36 },
		{ 86, 466, 36 },
		{ 86, 467, 36 },
		{ 86, 468, 36 },
		{ 86, 469, 36 },
		{ 86, 470, 36 },
		{ 86, 471, 36 },
		{ 86, 472, 36 },
		{ 86, 473, 36 },
		{ 86, 474, 36 },
		{ 86, 475, 36 },
		{ 86, 476, 36 },
		{ 86, 477, 36 },
		{ 86, 478, 36 },
		{ 86, 479, 36 },
		{ 86, 480, 36 },
		{ 86, 481, 36 },
		{ 86, 482, 36 },
		{ 86, 483, 36 },
		{ 86, 484, 36 },
		{ 86, 485, 36 },
		{ 86, 486, 36 },
		{ 86, 487, 36 },
		{ 86, 488, 36 },
		{ 86, 489, 36 },
		{ 86, 490, 36 },
		{ 86, 491, 36 },
		{ 86, 492, 36 },
		{ 86, 493, 36 },
		{ 86, 494, 36 },
		{ 86, 495, 36 },
		{ 86, 496, 36 },
		{ 86, 497, 36 },
		{ 86, 498, 36 },
		{ 86, 499, 36 },
		{ 86, 500, 36 },
		{ 86, 501, 36 },
		{ 86, 502, 36 },
		{ 86, 503, 36 },
		{ 86, 504, 36 },
		{ 86, 505, 36 },
		{ 86, 506, 36 },
		{ 86, 507, 36 },
		{ 86, 508, 36 },
		{ 86, 509, 36 },
		{ 86, 510, 36 },
		{ 86, 511, 36 },
		{ 86, 512, 36 },
		{ 87, 452, 36 },
		{ 87, 453, 36 },
		{ 87, 454, 36 },
		{ 87, 455, 36 },
		{ 87, 456, 36 },
		{ 87, 457, 36 },
		{ 87, 458, 36 },
		{ 87, 459, 36 },
		{ 87, 460, 36 },
		{ 87, 461, 36 },
		{ 87, 462, 36 },
		{ 87, 463, 36 },
		{ 87, 464, 36 },
		{ 87, 465, 36 },
		{ 87, 466, 36 },
		{ 87, 467, 36 },
		{ 87, 468, 36 },
		{ 87, 469, 36 },
		{ 87, 470, 36 },
		{ 87, 471, 36 },
		{ 87, 472, 36 },
		{ 87, 473, 36 },
		{ 87, 474, 36 },
		{ 87, 475, 36 },
		{ 87, 476, 36 },
		{ 87, 477, 36 },
		{ 87, 478, 36 },
		{ 87, 479, 36 },
		{ 87, 480, 36 },
		{ 87, 481, 36 },
		{ 87, 482, 36 },
		{ 87, 483, 36 },
		{ 87, 484, 36 },
		{ 87, 485, 36 },
		{ 87, 486, 36 },
		{ 87, 487, 36 },
		{ 87, 488, 36 },
		{ 87, 489, 36 },
		{ 87, 490, 36 },
		{ 87, 491, 36 },
		{ 87, 492, 36 },
		{ 87, 493, 36 },
		{ 87, 494, 36 },
		{ 87, 495, 36 },
		{ 87, 496, 36 },
		{ 87, 497, 36 },
		{ 87, 498, 36 },
		{ 87, 499, 36 },
		{ 87, 500, 36 },
		{ 87, 501, 36 },
		{ 87, 502, 36 },
		{ 87, 503, 36 },
		{ 87, 504, 36 },
		{ 87, 505, 36 },
		{ 87, 506, 36 },
		{ 87, 507, 36 },
		{ 87, 508, 36 },
		{ 87, 509, 36 },
		{ 87, 510, 36 },
		{ 87, 511, 36 },
		{ 87, 512, 36 },
		{ 88, 451, 36 },
		{ 88, 452, 36 },
		{ 88, 453, 36 },
		{ 88, 454, 36 },
		{ 88, 455, 36 },
		{ 88, 456, 36 },
		{ 88, 457, 36 },
		{ 88, 458, 36 },
		{ 88, 459, 36 },
		{ 88, 460, 36 },
		{ 88, 461, 36 },
		{ 88, 462, 36 },
		{ 88, 463, 36 },
		{ 88, 464, 36 },
		{ 88, 465, 36 },
		{ 88, 466, 36 },
		{ 88, 467, 36 },
		{ 88, 468, 36 },
		{ 88, 469, 36 },
		{ 88, 470, 36 },
		{ 88, 471, 36 },
		{ 88, 472, 36 },
		{ 88, 473, 36 },
		{ 88, 474, 36 },
		{ 88, 475, 36 },
		{ 88, 476, 36 },
		{ 88, 477, 36 },
		{ 88, 478, 36 },
		{ 88, 479, 36 },
		{ 88, 480, 36 },
		{ 88, 481, 36 },
		{ 88, 482, 36 },
		{ 88, 483, 36 },
		{ 88, 484, 36 },
		{ 88, 485, 36 },
		{ 88, 486, 36 },
		{ 88, 487, 36 },
		{ 88, 488, 36 },
		{ 88, 489, 36 },
		{ 88, 490, 36 },
		{ 88, 491, 36 },
		{ 88, 492, 36 },
		{ 88, 493, 36 },
		{ 88, 494, 36 },
		{ 88, 495, 36 },
		{ 88, 496, 36 },
		{ 88, 497, 36 },
		{ 88, 498, 36 },
		{ 88, 499, 36 },
		{ 88, 500, 36 },
		{ 88, 501, 36 },
		{ 88, 502, 36 },
		{ 88, 503, 36 },
		{ 88, 504, 36 },
		{ 88, 505, 36 },
		{ 88, 506, 36 },
		{ 88, 507, 36 },
		{ 88, 508, 36 },
		{ 88, 509, 36 },
		{ 88, 510, 36 },
		{ 88, 511, 36 },
		{ 88, 512, 36 },
		{ 89, 450, 36 },
		{ 89, 451, 36 },
		{ 89, 452, 36 },
		{ 89, 453, 36 },
		{ 89, 454, 36 },
		{ 89, 455, 36 },
		{ 89, 456, 36 },
		{ 89, 457, 36 },
		{ 89, 458, 36 },
		{ 89, 459, 36 },
		{ 89, 460, 36 },
		{ 89, 461, 36 },
		{ 89, 462, 36 },
		{ 89, 463, 36 },
		{ 89, 464, 36 },
		{ 89, 465, 36 },
		{ 89, 466, 36 },
		{ 89, 467, 36 },
		{ 89, 468, 36 },
		{ 89, 469, 36 },
		{ 89, 470, 36 },
		{ 89, 471, 36 },
		{ 89, 472, 36 },
		{ 89, 473, 36 },
		{ 89, 474, 36 },
		{ 89, 475, 36 },
		{ 89, 476, 36 },
		{ 89, 477, 36 },
		{ 89, 478, 36 },
		{ 89, 479, 36 },
		{ 89, 480, 36 },
		{ 89, 481, 36 },
		{ 89, 482, 36 },
		{ 89, 483, 36 },
		{ 89, 484, 36 },
		{ 89, 485, 36 },
		{ 89, 486, 36 },
		{ 89, 487, 36 },
		{ 89, 488, 36 },
		{ 89, 489, 36 },
		{ 89, 490, 36 },
		{ 89, 491, 36 },
		{ 89, 492, 36 },
		{ 89, 493, 36 },
		{ 89, 494, 36 },
		{ 89, 495, 36 },
		{ 89, 496, 36 },
		{ 89, 497, 36 },
		{ 89, 498, 36 },
		{ 89, 499, 36 },
		{ 89, 500, 36 },
		{ 89, 501, 36 },
		{ 89, 502, 36 },
		{ 89, 503, 36 },
		{ 89, 504, 36 },
		{ 89, 505, 36 },
		{ 89, 506, 36 },
		{ 89, 507, 36 },
		{ 89, 508, 36 },
		{ 89, 509, 36 },
		{ 89, 510, 36 },
		{ 89, 511, 36 },
		{ 89, 512, 36 },
		{ 90, 448, 36 },
		{ 90, 449, 36 },
		{ 90, 450, 36 },
		{ 90, 451, 36 },
		{ 90, 452, 36 },
		{ 90, 453, 36 },
		{ 90, 454, 36 },
		{ 90, 455, 36 },
		{ 90, 456, 36 },
		{ 90, 457, 36 },
		{ 90, 458, 36 },
		{ 90, 459, 36 },
		{ 90, 460, 36 },
		{ 90, 461, 36 },
		{ 90, 462, 36 },
		{ 90, 463, 36 },
		{ 90, 464, 36 },
		{ 90, 465, 36 },
		{ 90, 466, 36 },
		{ 90, 467, 36 },
		{ 90, 468, 36 },
		{ 90, 469, 36 },
		{ 90, 470, 36 },
		{ 90, 471, 36 },
		{ 90, 472, 36 },
		{ 90, 473, 36 },
		{ 90, 474, 36 },
		{ 90, 475, 36 },
		{ 90, 476, 36 },
		{ 90, 477, 36 },
		{ 90, 478, 36 },
		{ 90, 479, 36 },
		{ 90, 480, 36 },
		{ 90, 481, 36 },
		{ 90, 482, 36 },
		{ 90, 483, 36 },
		{ 90, 484, 36 },
		{ 90, 485, 36 },
		{ 90, 486, 36 },
		{ 90, 487, 36 },
		{ 90, 488, 36 },
		{ 90, 489, 36 },
		{ 90, 490, 36 },
		{ 90, 491, 36 },
		{ 90, 492, 36 },
		{ 90, 493, 36 },
		{ 90, 494, 36 },
		{ 90, 495, 36 },
		{ 90, 496, 36 },
		{ 90, 497, 36 },
		{ 90, 498, 36 },
		{ 90, 499, 36 },
		{ 90, 500, 36 },
		{ 90, 501, 36 },
		{ 90, 502, 36 },
		{ 90, 503, 36 },
		{ 90, 504, 36 },
		{ 90, 505, 36 },
		{ 90, 506, 36 },
		{ 90, 507, 36 },
		{ 90, 508, 36 },
		{ 90, 509, 36 },
		{ 90, 510, 36 },
		{ 90, 511, 36 },
		{ 90, 512, 36 },
		{ 91, 447, 36 },
		{ 91, 448, 36 },
		{ 91, 449, 36 },
		{ 91, 450, 36 },
		{ 91, 451, 36 },
		{ 91, 452, 36 },
		{ 91, 453, 36 },
		{ 91, 454, 36 },
		{ 91, 455, 36 },
		{ 91, 456, 36 },
		{ 91, 457, 36 },
		{ 91, 458, 36 },
		{ 91, 459, 36 },
		{ 91, 460, 36 },
		{ 91, 461, 36 },
		{ 91, 462, 36 },
		{ 91, 463, 36 },
		{ 91, 464, 36 },
		{ 91, 465, 36 },
		{ 91, 466, 36 },
		{ 91, 467, 36 },
		{ 91, 468, 36 },
		{ 91, 469, 36 },
		{ 91, 470, 36 },
		{ 91, 471, 36 },
		{ 91, 472, 36 },
		{ 91, 473, 36 },
		{ 91, 474, 36 },
		{ 91, 475, 36 },
		{ 91, 476, 36 },
		{ 91, 477, 36 },
		{ 91, 478, 36 },
		{ 91, 479, 36 },
		{ 91, 480, 36 },
		{ 91, 481, 36 },
		{ 91, 482, 36 },
		{ 91, 483, 36 },
		{ 91, 484, 36 },
		{ 91, 485, 36 },
		{ 91, 486, 36 },
		{ 91, 487, 36 },
		{ 91, 488, 36 },
		{ 91, 489, 36 },
		{ 91, 490, 36 },
		{ 91, 491, 36 },
		{ 91, 492, 36 },
		{ 91, 493, 36 },
		{ 91, 494, 36 },
		{ 91, 495, 36 },
		{ 91, 496, 36 },
		{ 91, 497, 36 },
		{ 91, 498, 36 },
		{ 91, 499, 36 },
		{ 91, 500, 36 },
		{ 91, 501, 36 },
		{ 91, 502, 36 },
		{ 91, 503, 36 },
		{ 91, 504, 36 },
		{ 91, 505, 36 },
		{ 91, 506, 36 },
		{ 91, 507, 36 },
		{ 91, 508, 36 },
		{ 91, 509, 36 },
		{ 91, 510, 36 },
		{ 91, 511, 36 },
		{ 91, 512, 36 },
		{ 92, 446, 36 },
		{ 92, 447, 36 },
		{ 92, 448, 36 },
		{ 92, 449, 36 },
		{ 92, 450, 36 },
		{ 92, 451, 36 },
		{ 92, 452, 36 },
		{ 92, 453, 36 },
		{ 92, 454, 36 },
		{ 92, 455, 36 },
		{ 92, 456, 36 },
		{ 92, 457, 36 },
		{ 92, 458, 36 },
		{ 92, 459, 36 },
		{ 92, 460, 36 },
		{ 92, 461, 36 },
		{ 92, 462, 36 },
		{ 92, 463, 36 },
		{ 92, 464, 36 },
		{ 92, 465, 36 },
		{ 92, 466, 36 },
		{ 92, 467, 36 },
		{ 92, 468, 36 },
		{ 92, 469, 36 },
		{ 92, 470, 36 },
		{ 92, 471, 36 },
		{ 92, 472, 36 },
		{ 92, 473, 36 },
		{ 92, 474, 36 },
		{ 92, 475, 36 },
		{ 92, 476, 36 },
		{ 92, 477, 36 },
		{ 92, 478, 36 },
		{ 92, 479, 36 },
		{ 92, 480, 36 },
		{ 92, 481, 36 },
		{ 92, 482, 36 },
		{ 92, 483, 36 },
		{ 92, 484, 36 },
		{ 92, 485, 36 },
		{ 92, 486, 36 },
		{ 92, 487, 36 },
		{ 92, 488, 36 },
		{ 92, 489, 36 },
		{ 92, 490, 36 },
		{ 92, 491, 36 },
		{ 92, 492, 36 },
		{ 92, 493, 36 },
		{ 92, 494, 36 },
		{ 92, 495, 36 },
		{ 92, 496, 36 },
		{ 92, 497, 36 },
		{ 92, 498, 36 },
		{ 92, 499, 36 },
		{ 92, 500, 36 },
		{ 92, 501, 36 },
		{ 92, 502, 36 },
		{ 92, 503, 36 },
		{ 92, 504, 36 },
		{ 92, 505, 36 },
		{ 92, 506, 36 },
		{ 92, 507, 36 },
		{ 92, 508, 36 },
		{ 92, 509, 36 },
		{ 92, 510, 36 },
		{ 92, 511, 36 },
		{ 92, 512, 36 },
		{ 93, 445, 36 },
		{ 93, 446, 36 },
		{ 93, 447, 36 },
		{ 93, 448, 36 },
		{ 93, 449, 36 },
		{ 93, 450, 36 },
		{ 93, 451, 36 },
		{ 93, 452, 36 },
		{ 93, 453, 36 },
		{ 93, 454, 36 },
		{ 93, 455, 36 },
		{ 93, 456, 36 },
		{ 93, 457, 36 },
		{ 93, 458, 36 },
		{ 93, 459, 36 },
		{ 93, 460, 36 },
		{ 93, 461, 36 },
		{ 93, 462, 36 },
		{ 93, 463, 36 },
		{ 93, 464, 36 },
		{ 93, 465, 36 },
		{ 93, 466, 36 },
		{ 93, 467, 36 },
		{ 93, 468, 36 },
		{ 93, 469, 36 },
		{ 93, 470, 36 },
		{ 93, 471, 36 },
		{ 93, 472, 36 },
		{ 93, 473, 36 },
		{ 93, 474, 36 },
		{ 93, 475, 36 },
		{ 93, 476, 36 },
		{ 93, 477, 36 },
		{ 93, 478, 36 },
		{ 93, 479, 36 },
		{ 93, 480, 36 },
		{ 93, 481, 36 },
		{ 93, 482, 36 },
		{ 93, 483, 36 },
		{ 93, 484, 36 },
		{ 93, 485, 36 },
		{ 93, 486, 36 },
		{ 93, 487, 36 },
		{ 93, 488, 36 },
		{ 93, 489, 36 },
		{ 93, 490, 36 },
		{ 93, 491, 36 },
		{ 93, 492, 36 },
		{ 93, 493, 36 },
		{ 93, 494, 36 },
		{ 93, 495, 36 },
		{ 93, 496, 36 },
		{ 93, 497, 36 },
		{ 93, 498, 36 },
		{ 93, 499, 36 },
		{ 93, 500, 36 },
		{ 93, 501, 36 },
		{ 93, 502, 36 },
		{ 93, 503, 36 },
		{ 93, 504, 36 },
		{ 93, 505, 36 },
		{ 93, 506, 36 },
		{ 93, 507, 36 },
		{ 93, 508, 36 },
		{ 93, 509, 36 },
		{ 93, 510, 36 },
		{ 93, 511, 36 },
		{ 93, 512, 36 },
		{ 94, 444, 36 },
		{ 94, 445, 36 },
		{ 94, 446, 36 },
		{ 94, 447, 36 },
		{ 94, 448, 36 },
		{ 94, 449, 36 },
		{ 94, 450, 36 },
		{ 94, 451, 36 },
		{ 94, 452, 36 },
		{ 94, 453, 36 },
		{ 94, 454, 36 },
		{ 94, 455, 36 },
		{ 94, 456, 36 },
		{ 94, 457, 36 },
		{ 94, 458, 36 },
		{ 94, 459, 36 },
		{ 94, 460, 36 },
		{ 94, 461, 36 },
		{ 94, 462, 36 },
		{ 94, 463, 36 },
		{ 94, 464, 36 },
		{ 94, 465, 36 },
		{ 94, 466, 36 },
		{ 94, 467, 36 },
		{ 94, 468, 36 },
		{ 94, 469, 36 },
		{ 94, 470, 36 },
		{ 94, 471, 36 },
		{ 94, 472, 36 },
		{ 94, 473, 36 },
		{ 94, 474, 36 },
		{ 94, 475, 36 },
		{ 94, 476, 36 },
		{ 94, 477, 36 },
		{ 94, 478, 36 },
		{ 94, 479, 36 },
		{ 94, 480, 36 },
		{ 94, 481, 36 },
		{ 94, 482, 36 },
		{ 94, 483, 36 },
		{ 94, 484, 36 },
		{ 94, 485, 36 },
		{ 94, 486, 36 },
		{ 94, 487, 36 },
		{ 94, 488, 36 },
		{ 94, 489, 36 },
		{ 94, 490, 36 },
		{ 94, 491, 36 },
		{ 94, 492, 36 },
		{ 94, 493, 36 },
		{ 94, 494, 36 },
		{ 94, 495, 36 },
		{ 94, 496, 36 },
		{ 94, 497, 36 },
		{ 94, 498, 36 },
		{ 94, 499, 36 },
		{ 94, 500, 36 },
		{ 94, 501, 36 },
		{ 94, 502, 36 },
		{ 94, 503, 36 },
		{ 94, 504, 36 },
		{ 94, 505, 36 },
		{ 94, 506, 36 },
		{ 94, 507, 36 },
		{ 94, 508, 36 },
		{ 94, 509, 36 },
		{ 94, 510, 36 },
		{ 94, 511, 36 },
		{ 94, 512, 36 },
		{ 95, 443, 36 },
		{ 95, 444, 36 },
		{ 95, 445, 36 },
		{ 95, 446, 36 },
		{ 95, 447, 36 },
		{ 95, 448, 36 },
		{ 95, 449, 36 },
		{ 95, 450, 36 },
		{ 95, 451, 36 },
		{ 95, 452, 36 },
		{ 95, 453, 36 },
		{ 95, 454, 36 },
		{ 95, 455, 36 },
		{ 95, 456, 36 },
		{ 95, 457, 36 },
		{ 95, 458, 36 },
		{ 95, 459, 36 },
		{ 95, 460, 36 },
		{ 95, 461, 36 },
		{ 95, 462, 36 },
		{ 95, 463, 36 },
		{ 95, 464, 36 },
		{ 95, 465, 36 },
		{ 95, 466, 36 },
		{ 95, 467, 36 },
		{ 95, 468, 36 },
		{ 95, 469, 36 },
		{ 95, 470, 36 },
		{ 95, 471, 36 },
		{ 95, 472, 36 },
		{ 95, 473, 36 },
		{ 95, 474, 36 },
		{ 95, 475, 36 },
		{ 95, 476, 36 },
		{ 95, 477, 36 },
		{ 95, 478, 36 },
		{ 95, 479, 36 },
		{ 95, 480, 36 },
		{ 95, 481, 36 },
		{ 95, 482, 36 },
		{ 95, 483, 36 },
		{ 95, 484, 36 },
		{ 95, 485, 36 },
		{ 95, 486, 36 },
		{ 95, 487, 36 },
		{ 95, 488, 36 },
		{ 95, 489, 36 },
		{ 95, 490, 36 },
		{ 95, 491, 36 },
		{ 95, 492, 36 },
		{ 95, 493, 36 },
		{ 95, 494, 36 },
		{ 95, 495, 36 },
		{ 95, 496, 36 },
		{ 95, 497, 36 },
		{ 95, 498, 36 },
		{ 95, 499, 36 },
		{ 95, 500, 36 },
		{ 95, 501, 36 },
		{ 95, 502, 36 },
		{ 95, 503, 36 },
		{ 95, 504, 36 },
		{ 95, 505, 36 },
		{ 95, 506, 36 },
		{ 95, 507, 36 },
		{ 95, 508, 36 },
		{ 95, 509, 36 },
		{ 95, 510, 36 },
		{ 95, 511, 36 },
		{ 95, 512, 36 },
		{ 96, 442, 36 },
		{ 96, 443, 36 },
		{ 96, 444, 36 },
		{ 96, 445, 36 },
		{ 96, 446, 36 },
		{ 96, 447, 36 },
		{ 96, 448, 36 },
		{ 96, 449, 36 },
		{ 96, 450, 36 },
		{ 96, 451, 36 },
		{ 96, 452, 36 },
		{ 96, 453, 36 },
		{ 96, 454, 36 },
		{ 96, 455, 36 },
		{ 96, 456, 36 },
		{ 96, 457, 36 },
		{ 96, 458, 36 },
		{ 96, 459, 36 },
		{ 96, 460, 36 },
		{ 96, 461, 36 },
		{ 96, 462, 36 },
		{ 96, 463, 36 },
		{ 96, 464, 36 },
		{ 96, 465, 36 },
		{ 96, 466, 36 },
		{ 96, 467, 36 },
		{ 96, 468, 36 },
		{ 96, 469, 36 },
		{ 96, 470, 36 },
		{ 96, 471, 36 },
		{ 96, 472, 36 },
		{ 96, 473, 36 },
		{ 96, 474, 36 },
		{ 96, 475, 36 },
		{ 96, 476, 36 },
		{ 96, 477, 36 },
		{ 96, 478, 36 },
		{ 96, 479, 36 },
		{ 96, 480, 36 },
		{ 96, 481, 36 },
		{ 96, 482, 36 },
		{ 96, 483, 36 },
		{ 96, 484, 36 },
		{ 96, 485, 36 },
		{ 96, 486, 36 },
		{ 96, 487, 36 },
		{ 96, 488, 36 },
		{ 96, 489, 36 },
		{ 96, 490, 36 },
		{ 96, 491, 36 },
		{ 96, 492, 36 },
		{ 96, 493, 36 },
		{ 96, 494, 36 },
		{ 96, 495, 36 },
		{ 96, 496, 36 },
		{ 96, 497, 36 },
		{ 96, 498, 36 },
		{ 96, 499, 36 },
		{ 96, 500, 36 },
		{ 96, 501, 36 },
		{ 96, 502, 36 },
		{ 96, 503, 36 },
		{ 96, 504, 36 },
		{ 96, 505, 36 },
		{ 96, 506, 36 },
		{ 96, 507, 36 },
		{ 96, 508, 36 },
		{ 96, 509, 36 },
		{ 96, 510, 36 },
		{ 96, 511, 36 },
		{ 96, 512, 36 },
		{ 97, 440, 36 },
		{ 97, 441, 36 },
		{ 97, 442, 36 },
		{ 97, 443, 36 },
		{ 97, 444, 36 },
		{ 97, 445, 36 },
		{ 97, 446, 36 },
		{ 97, 447, 36 },
		{ 97, 448, 36 },
		{ 97, 449, 36 },
		{ 97, 450, 36 },
		{ 97, 451, 36 },
		{ 97, 452, 36 },
		{ 97, 453, 36 },
		{ 97, 454, 36 },
		{ 97, 455, 36 },
		{ 97, 456, 36 },
		{ 97, 457, 36 },
		{ 97, 458, 36 },
		{ 97, 459, 36 },
		{ 97, 460, 36 },
		{ 97, 461, 36 },
		{ 97, 462, 36 },
		{ 97, 463, 36 },
		{ 97, 464, 36 },
		{ 97, 465, 36 },
		{ 97, 466, 36 },
		{ 97, 467, 36 },
		{ 97, 468, 36 },
		{ 97, 469, 36 },
		{ 97, 470, 36 },
		{ 97, 471, 36 },
		{ 97, 472, 36 },
		{ 97, 473, 36 },
		{ 97, 474, 36 },
		{ 97, 475, 36 },
		{ 97, 476, 36 },
		{ 97, 477, 36 },
		{ 97, 478, 36 },
		{ 97, 479, 36 },
		{ 97, 480, 36 },
		{ 97, 481, 36 },
		{ 97, 482, 36 },
		{ 97, 483, 36 },
		{ 97, 484, 36 },
		{ 97, 485, 36 },
		{ 97, 486, 36 },
		{ 97, 487, 36 },
		{ 97, 488, 36 },
		{ 97, 489, 36 },
		{ 97, 490, 36 },
		{ 97, 491, 36 },
		{ 97, 492, 36 },
		{ 97, 493, 36 },
		{ 97, 494, 36 },
		{ 97, 495, 36 },
		{ 97, 496, 36 },
		{ 97, 497, 36 },
		{ 97, 498, 36 },
		{ 97, 499, 36 },
		{ 97, 500, 36 },
		{ 97, 501, 36 },
		{ 97, 502, 36 },
		{ 97, 503, 36 },
		{ 97, 504, 36 },
		{ 97, 505, 36 },
		{ 97, 506, 36 },
		{ 97, 507, 36 },
		{ 97, 508, 36 },
		{ 97, 509, 36 },
		{ 97, 510, 36 },
		{ 97, 511, 36 },
		{ 97, 512, 36 },
		{ 98, 439, 36 },
		{ 98, 440, 36 },
		{ 98, 441, 36 },
		{ 98, 442, 36 },
		{ 98, 443, 36 },
		{ 98, 444, 36 },
		{ 98, 445, 36 },
		{ 98, 446, 36 },
		{ 98, 447, 36 },
		{ 98, 448, 36 },
		{ 98, 449, 36 },
		{ 98, 450, 36 },
		{ 98, 451, 36 },
		{ 98, 452, 36 },
		{ 98, 453, 36 },
		{ 98, 454, 36 },
		{ 98, 455, 36 },
		{ 98, 456, 36 },
		{ 98, 457, 36 },
		{ 98, 458, 36 },
		{ 98, 459, 36 },
		{ 98, 460, 36 },
		{ 98, 461, 36 },
		{ 98, 462, 36 },
		{ 98, 463, 36 },
		{ 98, 464, 36 },
		{ 98, 465, 36 },
		{ 98, 466, 36 },
		{ 98, 467, 36 },
		{ 98, 468, 36 },
		{ 98, 469, 36 },
		{ 98, 470, 36 },
		{ 98, 471, 36 },
		{ 98, 472, 36 },
		{ 98, 473, 36 },
		{ 98, 474, 36 },
		{ 98, 475, 36 },
		{ 98, 476, 36 },
		{ 98, 477, 36 },
		{ 98, 478, 36 },
		{ 98, 479, 36 },
		{ 98, 480, 36 },
		{ 98, 481, 36 },
		{ 98, 482, 36 },
		{ 98, 483, 36 },
		{ 98, 484, 36 },
		{ 98, 485, 36 },
		{ 98, 486, 36 },
		{ 98, 487, 36 },
		{ 98, 488, 36 },
		{ 98, 489, 36 },
		{ 98, 490, 36 },
		{ 98, 491, 36 },
		{ 98, 492, 36 },
		{ 98, 493, 36 },
		{ 98, 494, 36 },
		{ 98, 495, 36 },
		{ 98, 496, 36 },
		{ 98, 497, 36 },
		{ 98, 498, 36 },
		{ 98, 499, 36 },
		{ 98, 500, 36 },
		{ 98, 501, 36 },
		{ 98, 502, 36 },
		{ 98, 503, 36 },
		{ 98, 504, 36 },
		{ 98, 505, 36 },
		{ 98, 506, 36 },
		{ 98, 507, 36 },
		{ 98, 508, 36 },
		{ 98, 509, 36 },
		{ 98, 510, 36 },
		{ 98, 511, 36 },
		{ 99, 438, 36 },
		{ 99, 439, 36 },
		{ 99, 440, 36 },
		{ 99, 441, 36 },
		{ 99, 442, 36 },
		{ 99, 443, 36 },
		{ 99, 444, 36 },
		{ 99, 445, 36 },
		{ 99, 446, 36 },
		{ 99, 447, 36 },
		{ 99, 448, 36 },
		{ 99, 449, 36 },
		{ 99, 450, 36 },
		{ 99, 451, 36 },
		{ 99, 452, 36 },
		{ 99, 453, 36 },
		{ 99, 454, 36 },
		{ 99, 455, 36 },
		{ 99, 456, 36 },
		{ 99, 457, 36 },
		{ 99, 458, 36 },
		{ 99, 459, 36 },
		{ 99, 460, 36 },
		{ 99, 461, 36 },
		{ 99, 462, 36 },
		{ 99, 463, 36 },
		{ 99, 464, 36 },
		{ 99, 465, 36 },
		{ 99, 466, 36 },
		{ 99, 467, 36 },
		{ 99, 468, 36 },
		{ 99, 469, 36 },
		{ 99, 470, 36 },
		{ 99, 471, 36 },
		{ 99, 472, 36 },
		{ 99, 473, 36 },
		{ 99, 474, 36 },
		{ 99, 475, 36 },
		{ 99, 476, 36 },
		{ 99, 477, 36 },
		{ 99, 478, 36 },
		{ 99, 479, 36 },
		{ 99, 480, 36 },
		{ 99, 481, 36 },
		{ 99, 482, 36 },
		{ 99, 483, 36 },
		{ 99, 484, 36 },
		{ 99, 485, 36 },
		{ 99, 486, 36 },
		{ 99, 487, 36 },
		{ 99, 488, 36 },
		{ 99, 489, 36 },
		{ 99, 490, 36 },
		{ 99, 491, 36 },
		{ 99, 492, 36 },
		{ 99, 493, 36 },
		{ 99, 494, 36 },
		{ 99, 495, 36 },
		{ 99, 496, 36 },
		{ 99, 497, 36 },
		{ 99, 498, 36 },
		{ 99, 499, 36 },
		{ 99, 500, 36 },
		{ 99, 501, 36 },
		{ 99, 502, 36 },
		{ 99, 503, 36 },
		{ 99, 504, 36 },
		{ 99, 505, 36 },
		{ 99, 506, 36 },
		{ 99, 507, 36 },
		{ 99, 508, 36 },
		{ 99, 509, 36 },
		{ 100, 437, 36 },
		{ 100, 438, 36 },
		{ 100, 439, 36 },
		{ 100, 440, 36 },
		{ 100, 441, 36 },
		{ 100, 442, 36 },
		{ 100, 443, 36 },
		{ 100, 444, 36 },
		{ 100, 445, 36 },
		{ 100, 446, 36 },
		{ 100, 447, 36 },
		{ 100, 448, 36 },
		{ 100, 449, 36 },
		{ 100, 450, 36 },
		{ 100, 451, 36 },
		{ 100, 452, 36 },
		{ 100, 453, 36 },
		{ 100, 454, 36 },
		{ 100, 455, 36 },
		{ 100, 456, 36 },
		{ 100, 457, 36 },
		{ 100, 458, 36 },
		{ 100, 459, 36 },
		{ 100, 460, 36 },
		{ 100, 461, 36 },
		{ 100, 462, 36 },
		{ 100, 463, 36 },
		{ 100, 464, 36 },
		{ 100, 465, 36 },
		{ 100, 466, 36 },
		{ 100, 467, 36 },
		{ 100, 468, 36 },
		{ 100, 469, 36 },
		{ 100, 470, 36 },
		{ 100, 471, 36 },
		{ 100, 472, 36 },
		{ 100, 473, 36 },
		{ 100, 474, 36 },
		{ 100, 475, 36 },
		{ 100, 476, 36 },
		{ 100, 477, 36 },
		{ 100, 478, 36 },
		{ 100, 479, 36 },
		{ 100, 480, 36 },
		{ 100, 481, 36 },
		{ 100, 482, 36 },
		{ 100, 483, 36 },
		{ 100, 484, 36 },
		{ 100, 485, 36 },
		{ 100, 486, 36 },
		{ 100, 487, 36 },
		{ 100, 488, 36 },
		{ 100, 489, 36 },
		{ 100, 490, 36 },
		{ 100, 491, 36 },
		{ 100, 492, 36 },
		{ 100, 493, 36 },
		{ 100, 494, 36 },
		{ 100, 495, 36 },
		{ 100, 496, 36 },
		{ 100, 497, 36 },
		{ 100, 498, 36 },
		{ 100, 499, 36 },
		{ 100, 500, 36 },
		{ 100, 501, 36 },
		{ 100, 502, 36 },
		{ 100, 503, 36 },
		{ 100, 504, 36 },
		{ 100, 505, 36 },
		{ 100, 506, 36 },
		{ 100, 507, 36 },
		{ 100, 508, 36 },
		{ 101, 436, 36 },
		{ 101, 437, 36 },
		{ 101, 438, 36 },
		{ 101, 439, 36 },
		{ 101, 440, 36 },
		{ 101, 441, 36 },
		{ 101, 442, 36 },
		{ 101, 443, 36 },
		{ 101, 444, 36 },
		{ 101, 445, 36 },
		{ 101, 446, 36 },
		{ 101, 447, 36 },
		{ 101, 448, 36 },
		{ 101, 449, 36 },
		{ 101, 450, 36 },
		{ 101, 451, 36 },
		{ 101, 452, 36 },
		{ 101, 453, 36 },
		{ 101, 454, 36 },
		{ 101, 455, 36 },
		{ 101, 456, 36 },
		{ 101, 457, 36 },
		{ 101, 458, 36 },
		{ 101, 459, 36 },
		{ 101, 460, 36 },
		{ 101, 461, 36 },
		{ 101, 462, 36 },
		{ 101, 463, 36 },
		{ 101, 464, 36 },
		{ 101, 465, 36 },
		{ 101, 466, 36 },
		{ 101, 467, 36 },
		{ 101, 468, 36 },
		{ 101, 469, 36 },
		{ 101, 470, 36 },
		{ 101, 471, 36 },
		{ 101, 472, 36 },
		{ 101, 473, 36 },
		{ 101, 474, 36 },
		{ 101, 475, 36 },
		{ 101, 476, 36 },
		{ 101, 477, 36 },
		{ 101, 478, 36 },
		{ 101, 479, 36 },
		{ 101, 480, 36 },
		{ 101, 481, 36 },
		{ 101, 482, 36 },
		{ 101, 483, 36 },
		{ 101, 484, 36 },
		{ 101, 485, 36 },
		{ 101, 486, 36 },
		{ 101, 487, 36 },
		{ 101, 488, 36 },
		{ 101, 489, 36 },
		{ 101, 490, 36 },
		{ 101, 491, 36 },
		{ 101, 492, 36 },
		{ 101, 493, 36 },
		{ 101, 494, 36 },
		{ 101, 495, 36 },
		{ 101, 496, 36 },
		{ 101, 497, 36 },
		{ 101, 498, 36 },
		{ 101, 499, 36 },
		{ 101, 500, 36 },
		{ 101, 501, 36 },
		{ 101, 502, 36 },
		{ 101, 503, 36 },
		{ 101, 504, 36 },
		{ 101, 505, 36 },
		{ 101, 506, 36 },
		{ 102, 435, 36 },
		{ 102, 436, 36 },
		{ 102, 437, 36 },
		{ 102, 438, 36 },
		{ 102, 439, 36 },
		{ 102, 440, 36 },
		{ 102, 441, 36 },
		{ 102, 442, 36 },
		{ 102, 443, 36 },
		{ 102, 444, 36 },
		{ 102, 445, 36 },
		{ 102, 446, 36 },
		{ 102, 447, 36 },
		{ 102, 448, 36 },
		{ 102, 449, 36 },
		{ 102, 450, 36 },
		{ 102, 451, 36 },
		{ 102, 452, 36 },
		{ 102, 453, 36 },
		{ 102, 454, 36 },
		{ 102, 455, 36 },
		{ 102, 456, 36 },
		{ 102, 457, 36 },
		{ 102, 458, 36 },
		{ 102, 459, 36 },
		{ 102, 460, 36 },
		{ 102, 461, 36 },
		{ 102, 462, 36 },
		{ 102, 463, 36 },
		{ 102, 464, 36 },
		{ 102, 465, 36 },
		{ 102, 466, 36 },
		{ 102, 467, 36 },
		{ 102, 468, 36 },
		{ 102, 469, 36 },
		{ 102, 470, 36 },
		{ 102, 471, 36 },
		{ 102, 472, 36 },
		{ 102, 473, 36 },
		{ 102, 474, 36 },
		{ 102, 475, 36 },
		{ 102, 476, 36 },
		{ 102, 477, 36 },
		{ 102, 478, 36 },
		{ 102, 479, 36 },
		{ 102, 480, 36 },
		{ 102, 481, 36 },
		{ 102, 482, 36 },
		{ 102, 483, 36 },
		{ 102, 484, 36 },
		{ 102, 485, 36 },
		{ 102, 486, 36 },
		{ 102, 487, 36 },
		{ 102, 488, 36 },
		{ 102, 489, 36 },
		{ 102, 490, 36 },
		{ 102, 491, 36 },
		{ 102, 492, 36 },
		{ 102, 493, 36 },
		{ 102, 494, 36 },
		{ 102, 495, 36 },
		{ 102, 496, 36 },
		{ 102, 497, 36 },
		{ 102, 498, 36 },
		{ 102, 499, 36 },
		{ 102, 500, 36 },
		{ 102, 501, 36 },
		{ 102, 502, 36 },
		{ 102, 503, 36 },
		{ 102, 504, 36 },
		{ 103, 434, 36 },
		{ 103, 435, 36 },
		{ 103, 436, 36 },
		{ 103, 437, 36 },
		{ 103, 438, 36 },
		{ 103, 439, 36 },
		{ 103, 440, 36 },
		{ 103, 441, 36 },
		{ 103, 442, 36 },
		{ 103, 443, 36 },
		{ 103, 444, 36 },
		{ 103, 445, 36 },
		{ 103, 446, 36 },
		{ 103, 447, 36 },
		{ 103, 448, 36 },
		{ 103, 449, 36 },
		{ 103, 450, 36 },
		{ 103, 451, 36 },
		{ 103, 452, 36 },
		{ 103, 453, 36 },
		{ 103, 454, 36 },
		{ 103, 455, 36 },
		{ 103, 456, 36 },
		{ 103, 457, 36 },
		{ 103, 458, 36 },
		{ 103, 459, 36 },
		{ 103, 460, 36 },
		{ 103, 461, 36 },
		{ 103, 462, 36 },
		{ 103, 463, 36 },
		{ 103, 464, 36 },
		{ 103, 465, 36 },
		{ 103, 466, 36 },
		{ 103, 467, 36 },
		{ 103, 468, 36 },
		{ 103, 469, 36 },
		{ 103, 470, 36 },
		{ 103, 471, 36 },
		{ 103, 472, 36 },
		{ 103, 473, 36 },
		{ 103, 474, 36 },
		{ 103, 475, 36 },
		{ 103, 476, 36 },
		{ 103, 477, 36 },
		{ 103, 478, 36 },
		{ 103, 479, 36 },
		{ 103, 480, 36 },
		{ 103, 481, 36 },
		{ 103, 482, 36 },
		{ 103, 483, 36 },
		{ 103, 484, 36 },
		{ 103, 485, 36 },
		{ 103, 486, 36 },
		{ 103, 487, 36 },
		{ 103, 488, 36 },
		{ 103, 489, 36 },
		{ 103, 490, 36 },
		{ 103, 491, 36 },
		{ 103, 492, 36 },
		{ 103, 493, 36 },
		{ 103, 494, 36 },
		{ 103, 495, 36 },
		{ 103, 496, 36 },
		{ 103, 497, 36 },
		{ 103, 498, 36 },
		{ 103, 499, 36 },
		{ 103, 500, 36 },
		{ 103, 501, 36 },
		{ 103, 502, 36 },
		{ 103, 503, 36 },
		{ 104, 432, 36 },
		{ 104, 433, 36 },
		{ 104, 434, 36 },
		{ 104, 435, 36 },
		{ 104, 436, 36 },
		{ 104, 437, 36 },
		{ 104, 438, 36 },
		{ 104, 439, 36 },
		{ 104, 440, 36 },
		{ 104, 441, 36 },
		{ 104, 442, 36 },
		{ 104, 443, 36 },
		{ 104, 444, 36 },
		{ 104, 445, 36 },
		{ 104, 446, 36 },
		{ 104, 447, 36 },
		{ 104, 448, 36 },
		{ 104, 449, 36 },
		{ 104, 450, 36 },
		{ 104, 451, 36 },
		{ 104, 452, 36 },
		{ 104, 453, 36 },
		{ 104, 454, 36 },
		{ 104, 455, 36 },
		{ 104, 456, 36 },
		{ 104, 457, 36 },
		{ 104, 458, 36 },
		{ 104, 459, 36 },
		{ 104, 460, 36 },
		{ 104, 461, 36 },
		{ 104, 462, 36 },
		{ 104, 463, 36 },
		{ 104, 464, 36 },
		{ 104, 465, 36 },
		{ 104, 466, 36 },
		{ 104, 467, 36 },
		{ 104, 468, 36 },
		{ 104, 469, 36 },
		{ 104, 470, 36 },
		{ 104, 471, 36 },
		{ 104, 472, 36 },
		{ 104, 473, 36 },
		{ 104, 474, 36 },
		{ 104, 475, 36 },
		{ 104, 476, 36 },
		{ 104, 477, 36 },
		{ 104, 478, 36 },
		{ 104, 479, 36 },
		{ 104, 480, 36 },
		{ 104, 481, 36 },
		{ 104, 482, 36 },
		{ 104, 483, 36 },
		{ 104, 484, 36 },
		{ 104, 485, 36 },
		{ 104, 486, 36 },
		{ 104, 487, 36 },
		{ 104, 488, 36 },
		{ 104, 489, 36 },
		{ 104, 490, 36 },
		{ 104, 491, 36 },
		{ 104, 492, 36 },
		{ 104, 493, 36 },
		{ 104, 494, 36 },
		{ 104, 495, 36 },
		{ 104, 496, 36 },
		{ 104, 497, 36 },
		{ 104, 498, 36 },
		{ 104, 499, 36 },
		{ 104, 500, 36 },
		{ 104, 501, 36 },
		{ 105, 431, 36 },
		{ 105, 432, 36 },
		{ 105, 433, 36 },
		{ 105, 434, 36 },
		{ 105, 435, 36 },
		{ 105, 436, 36 },
		{ 105, 437, 36 },
		{ 105, 438, 36 },
		{ 105, 439, 36 },
		{ 105, 440, 36 },
		{ 105, 441, 36 },
		{ 105, 442, 36 },
		{ 105, 443, 36 },
		{ 105, 444, 36 },
		{ 105, 445, 36 },
		{ 105, 446, 36 },
		{ 105, 447, 36 },
		{ 105, 448, 36 },
		{ 105, 449, 36 },
		{ 105, 450, 36 },
		{ 105, 451, 36 },
		{ 105, 452, 36 },
		{ 105, 453, 36 },
		{ 105, 454, 36 },
		{ 105, 455, 36 },
		{ 105, 456, 36 },
		{ 105, 457, 36 },
		{ 105, 458, 36 },
		{ 105, 459, 36 },
		{ 105, 460, 36 },
		{ 105, 461, 36 },
		{ 105, 462, 36 },
		{ 105, 463, 36 },
		{ 105, 464, 36 },
		{ 105, 465, 36 },
		{ 105, 466, 36 },
		{ 105, 467, 36 },
		{ 105, 468, 36 },
		{ 105, 469, 36 },
		{ 105, 470, 36 },
		{ 105, 471, 36 },
		{ 105, 472, 36 },
		{ 105, 473, 36 },
		{ 105, 474, 36 },
		{ 105, 475, 36 },
		{ 105, 476, 36 },
		{ 105, 477, 36 },
		{ 105, 478, 36 },
		{ 105, 479, 36 },
		{ 105, 480, 36 },
		{ 105, 481, 36 },
		{ 105, 482, 36 },
		{ 105, 483, 36 },
		{ 105, 484, 36 },
		{ 105, 485, 36 },
		{ 105, 486, 36 },
		{ 105, 487, 36 },
		{ 105, 488, 36 },
		{ 105, 489, 36 },
		{ 105, 490, 36 },
		{ 105, 491, 36 },
		{ 105, 492, 36 },
		{ 105, 493, 36 },
		{ 105, 494, 36 },
		{ 105, 495, 36 },
		{ 105, 496, 36 },
		{ 105, 497, 36 },
		{ 105, 498, 36 },
		{ 105, 499, 36 },
		{ 105, 500, 36 },
		{ 106, 430, 36 },
		{ 106, 431, 36 },
		{ 106, 432, 36 },
		{ 106, 433, 36 },
		{ 106, 434, 36 },
		{ 106, 435, 36 },
		{ 106, 436, 36 },
		{ 106, 437, 36 },
		{ 106, 438, 36 },
		{ 106, 439, 36 },
		{ 106, 440, 36 },
		{ 106, 441, 36 },
		{ 106, 442, 36 },
		{ 106, 443, 36 },
		{ 106, 444, 36 },
		{ 106, 445, 36 },
		{ 106, 446, 36 },
		{ 106, 447, 36 },
		{ 106, 448, 36 },
		{ 106, 449, 36 },
		{ 106, 450, 36 },
		{ 106, 451, 36 },
		{ 106, 452, 36 },
		{ 106, 453, 36 },
		{ 106, 454, 36 },
		{ 106, 455, 36 },
		{ 106, 456, 36 },
		{ 106, 457, 36 },
		{ 106, 458, 36 },
		{ 106, 459, 36 },
		{ 106, 460, 36 },
		{ 106, 461, 36 },
		{ 106, 462, 36 },
		{ 106, 463, 36 },
		{ 106, 464, 36 },
		{ 106, 465, 36 },
		{ 106, 466, 36 },
		{ 106, 467, 36 },
		{ 106, 468, 36 },
		{ 106, 469, 36 },
		{ 106, 470, 36 },
		{ 106, 471, 36 },
		{ 106, 472, 36 },
		{ 106, 473, 36 },
		{ 106, 474, 36 },
		{ 106, 475, 36 },
		{ 106, 476, 36 },
		{ 106, 477, 36 },
		{ 106, 478, 36 },
		{ 106, 479, 36 },
		{ 106, 480, 36 },
		{ 106, 481, 36 },
		{ 106, 482, 36 },
		{ 106, 483, 36 },
		{ 106, 484, 36 },
		{ 106, 485, 36 },
		{ 106, 486, 36 },
		{ 106, 487, 36 },
		{ 106, 488, 36 },
		{ 106, 489, 36 },
		{ 106, 490, 36 },
		{ 106, 491, 36 },
		{ 106, 492, 36 },
		{ 106, 493, 36 },
		{ 106, 494, 36 },
		{ 106, 495, 36 },
		{ 106, 496, 36 },
		{ 106, 497, 36 },
		{ 106, 498, 36 },
		{ 107, 429, 36 },
		{ 107, 430, 36 },
		{ 107, 431, 36 },
		{ 107, 432, 36 },
		{ 107, 433, 36 },
		{ 107, 434, 36 },
		{ 107, 435, 36 },
		{ 107, 436, 36 },
		{ 107, 437, 36 },
		{ 107, 438, 36 },
		{ 107, 439, 36 },
		{ 107, 440, 36 },
		{ 107, 441, 36 },
		{ 107, 442, 36 },
		{ 107, 443, 36 },
		{ 107, 444, 36 },
		{ 107, 445, 36 },
		{ 107, 446, 36 },
		{ 107, 447, 36 },
		{ 107, 448, 36 },
		{ 107, 449, 36 },
		{ 107, 450, 36 },
		{ 107, 451, 36 },
		{ 107, 452, 36 },
		{ 107, 453, 36 },
		{ 107, 454, 36 },
		{ 107, 455, 36 },
		{ 107, 456, 36 },
		{ 107, 457, 36 },
		{ 107, 458, 36 },
		{ 107, 459, 36 },
		{ 107, 460, 36 },
		{ 107, 461, 36 },
		{ 107, 462, 36 },
		{ 107, 463, 36 },
		{ 107, 464, 36 },
		{ 107, 465, 36 },
		{ 107, 466, 36 },
		{ 107, 467, 36 },
		{ 107, 468, 36 },
		{ 107, 469, 36 },
		{ 107, 470, 36 },
		{ 107, 471, 36 },
		{ 107, 472, 36 },
		{ 107, 473, 36 },
		{ 107, 474, 36 },
		{ 107, 475, 36 },
		{ 107, 476, 36 },
		{ 107, 477, 36 },
		{ 107, 478, 36 },
		{ 107, 479, 36 },
		{ 107, 480, 36 },
		{ 107, 481, 36 },
		{ 107, 482, 36 },
		{ 107, 483, 36 },
		{ 107, 484, 36 },
		{ 107, 485, 36 },
		{ 107, 486, 36 },
		{ 107, 487, 36 },
		{ 107, 488, 36 },
		{ 107, 489, 36 },
		{ 107, 490, 36 },
		{ 107, 491, 36 },
		{ 107, 492, 36 },
		{ 107, 493, 36 },
		{ 107, 494, 36 },
		{ 107, 495, 36 },
		{ 107, 496, 36 },
		{ 108, 428, 36 },
		{ 108, 429, 36 },
		{ 108, 430, 36 },
		{ 108, 431, 36 },
		{ 108, 432, 36 },
		{ 108, 433, 36 },
		{ 108, 434, 36 },
		{ 108, 435, 36 },
		{ 108, 436, 36 },
		{ 108, 437, 36 },
		{ 108, 438, 36 },
		{ 108, 439, 36 },
		{ 108, 440, 36 },
		{ 108, 441, 36 },
		{ 108, 442, 36 },
		{ 108, 443, 36 },
		{ 108, 444, 36 },
		{ 108, 445, 36 },
		{ 108, 446, 36 },
		{ 108, 447, 36 },
		{ 108, 448, 36 },
		{ 108, 449, 36 },
		{ 108, 450, 36 },
		{ 108, 451, 36 },
		{ 108, 452, 36 },
		{ 108, 453, 36 },
		{ 108, 454, 36 },
		{ 108, 455, 36 },
		{ 108, 456, 36 },
		{ 108, 457, 36 },
		{ 108, 458, 36 },
		{ 108, 459, 36 },
		{ 108, 460, 36 },
		{ 108, 461, 36 },
		{ 108, 462, 36 },
		{ 108, 463, 36 },
		{ 108, 464, 36 },
		{ 108, 465, 36 },
		{ 108, 466, 36 },
		{ 108, 467, 36 },
		{ 108, 468, 36 },
		{ 108, 469, 36 },
		{ 108, 470, 36 },
		{ 108, 471, 36 },
		{ 108, 472, 36 },
		{ 108, 473, 36 },
		{ 108, 474, 36 },
		{ 108, 475, 36 },
		{ 108, 476, 36 },
		{ 108, 477, 36 },
		{ 108, 478, 36 },
		{ 108, 479, 36 },
		{ 108, 480, 36 },
		{ 108, 481, 36 },
		{ 108, 482, 36 },
		{ 108, 483, 36 },
		{ 108, 484, 36 },
		{ 108, 485, 36 },
		{ 108, 486, 36 },
		{ 108, 487, 36 },
		{ 108, 488, 36 },
		{ 108, 489, 36 },
		{ 108, 490, 36 },
		{ 108, 491, 36 },
		{ 108, 492, 36 },
		{ 108, 493, 36 },
		{ 108, 494, 36 },
		{ 108, 495, 36 },
		{ 109, 427, 36 },
		{ 109, 428, 36 },
		{ 109, 429, 36 },
		{ 109, 430, 36 },
		{ 109, 431, 36 },
		{ 109, 432, 36 },
		{ 109, 433, 36 },
		{ 109, 434, 36 },
		{ 109, 435, 36 },
		{ 109, 436, 36 },
		{ 109, 437, 36 },
		{ 109, 438, 36 },
		{ 109, 439, 36 },
		{ 109, 440, 36 },
		{ 109, 441, 36 },
		{ 109, 442, 36 },
		{ 109, 443, 36 },
		{ 109, 444, 36 },
		{ 109, 445, 36 },
		{ 109, 446, 36 },
		{ 109, 447, 36 },
		{ 109, 448, 36 },
		{ 109, 449, 36 },
		{ 109, 450, 36 },
		{ 109, 451, 36 },
		{ 109, 452, 36 },
		{ 109, 453, 36 },
		{ 109, 454, 36 },
		{ 109, 455, 36 },
		{ 109, 456, 36 },
		{ 109, 457, 36 },
		{ 109, 458, 36 },
		{ 109, 459, 36 },
		{ 109, 460, 36 },
		{ 109, 461, 36 },
		{ 109, 462, 36 },
		{ 109, 463, 36 },
		{ 109, 464, 36 },
		{ 109, 465, 36 },
		{ 109, 466, 36 },
		{ 109, 467, 36 },
		{ 109, 468, 36 },
		{ 109, 469, 36 },
		{ 109, 470, 36 },
		{ 109, 471, 36 },
		{ 109, 472, 36 },
		{ 109, 473, 36 },
		{ 109, 474, 36 },
		{ 109, 475, 36 },
		{ 109, 476, 36 },
		{ 109, 477, 36 },
		{ 109, 478, 36 },
		{ 109, 479, 36 },
		{ 109, 480, 36 },
		{ 109, 481, 36 },
		{ 109, 482, 36 },
		{ 109, 483, 36 },
		{ 109, 484, 36 },
		{ 109, 485, 36 },
		{ 109, 486, 36 },
		{ 109, 487, 36 },
		{ 109, 488, 36 },
		{ 109, 489, 36 },
		{ 109, 490, 36 },
		{ 109, 491, 36 },
		{ 109, 492, 36 },
		{ 109, 493, 36 },
		{ 110, 426, 36 },
		{ 110, 427, 36 },
		{ 110, 428, 36 },
		{ 110, 429, 36 },
		{ 110, 430, 36 },
		{ 110, 431, 36 },
		{ 110, 432, 36 },
		{ 110, 433, 36 },
		{ 110, 434, 36 },
		{ 110, 435, 36 },
		{ 110, 436, 36 },
		{ 110, 437, 36 },
		{ 110, 438, 36 },
		{ 110, 439, 36 },
		{ 110, 440, 36 },
		{ 110, 441, 36 },
		{ 110, 442, 36 },
		{ 110, 443, 36 },
		{ 110, 444, 36 },
		{ 110, 445, 36 },
		{ 110, 446, 36 },
		{ 110, 447, 36 },
		{ 110, 448, 36 },
		{ 110, 449, 36 },
		{ 110, 450, 36 },
		{ 110, 451, 36 },
		{ 110, 452, 36 },
		{ 110, 453, 36 },
		{ 110, 454, 36 },
		{ 110, 455, 36 },
		{ 110, 456, 36 },
		{ 110, 457, 36 },
		{ 110, 458, 36 },
		{ 110, 459, 36 },
		{ 110, 460, 36 },
		{ 110, 461, 36 },
		{ 110, 462, 36 },
		{ 110, 463, 36 },
		{ 110, 464, 36 },
		{ 110, 465, 36 },
		{ 110, 466, 36 },
		{ 110, 467, 36 },
		{ 110, 468, 36 },
		{ 110, 469, 36 },
		{ 110, 470, 36 },
		{ 110, 471, 36 },
		{ 110, 472, 36 },
		{ 110, 473, 36 },
		{ 110, 474, 36 },
		{ 110, 475, 36 },
		{ 110, 476, 36 },
		{ 110, 477, 36 },
		{ 110, 478, 36 },
		{ 110, 479, 36 },
		{ 110, 480, 36 },
		{ 110, 481, 36 },
		{ 110, 482, 36 },
		{ 110, 483, 36 },
		{ 110, 484, 36 },
		{ 110, 485, 36 },
		{ 110, 486, 36 },
		{ 110, 487, 36 },
		{ 110, 488, 36 },
		{ 110, 489, 36 },
		{ 110, 490, 36 },
		{ 110, 491, 36 },
		{ 110, 492, 36 },
		{ 111, 424, 36 },
		{ 111, 425, 36 },
		{ 111, 426, 36 },
		{ 111, 427, 36 },
		{ 111, 428, 36 },
		{ 111, 429, 36 },
		{ 111, 430, 36 },
		{ 111, 431, 36 },
		{ 111, 432, 36 },
		{ 111, 433, 36 },
		{ 111, 434, 36 },
		{ 111, 435, 36 },
		{ 111, 436, 36 },
		{ 111, 437, 36 },
		{ 111, 438, 36 },
		{ 111, 439, 36 },
		{ 111, 440, 36 },
		{ 111, 441, 36 },
		{ 111, 442, 36 },
		{ 111, 443, 36 },
		{ 111, 444, 36 },
		{ 111, 445, 36 },
		{ 111, 446, 36 },
		{ 111, 447, 36 },
		{ 111, 448, 36 },
		{ 111, 449, 36 },
		{ 111, 450, 36 },
		{ 111, 451, 36 },
		{ 111, 452, 36 },
		{ 111, 453, 36 },
		{ 111, 454, 36 },
		{ 111, 455, 36 },
		{ 111, 456, 36 },
		{ 111, 457, 36 },
		{ 111, 458, 36 },
		{ 111, 459, 36 },
		{ 111, 460, 36 },
		{ 111, 461, 36 },
		{ 111, 462, 36 },
		{ 111, 463, 36 },
		{ 111, 464, 36 },
		{ 111, 465, 36 },
		{ 111, 466, 36 },
		{ 111, 467, 36 },
		{ 111, 468, 36 },
		{ 111, 469, 36 },
		{ 111, 470, 36 },
		{ 111, 471, 36 },
		{ 111, 472, 36 },
		{ 111, 473, 36 },
		{ 111, 474, 36 },
		{ 111, 475, 36 },
		{ 111, 476, 36 },
		{ 111, 477, 36 },
		{ 111, 478, 36 },
		{ 111, 479, 36 },
		{ 111, 480, 36 },
		{ 111, 481, 36 },
		{ 111, 482, 36 },
		{ 111, 483, 36 },
		{ 111, 484, 36 },
		{ 111, 485, 36 },
		{ 111, 486, 36 },
		{ 111, 487, 36 },
		{ 111, 488, 36 },
		{ 111, 489, 36 },
		{ 111, 490, 36 },
		{ 112, 423, 36 },
		{ 112, 424, 36 },
		{ 112, 425, 36 },
		{ 112, 426, 36 },
		{ 112, 427, 36 },
		{ 112, 428, 36 },
		{ 112, 429, 36 },
		{ 112, 430, 36 },
		{ 112, 431, 36 },
		{ 112, 432, 36 },
		{ 112, 433, 36 },
		{ 112, 434, 36 },
		{ 112, 435, 36 },
		{ 112, 436, 36 },
		{ 112, 437, 36 },
		{ 112, 438, 36 },
		{ 112, 439, 36 },
		{ 112, 440, 36 },
		{ 112, 441, 36 },
		{ 112, 442, 36 },
		{ 112, 443, 36 },
		{ 112, 444, 36 },
		{ 112, 445, 36 },
		{ 112, 446, 36 },
		{ 112, 447, 36 },
		{ 112, 448, 36 },
		{ 112, 449, 36 },
		{ 112, 450, 36 },
		{ 112, 451, 36 },
		{ 112, 452, 36 },
		{ 112, 453, 36 },
		{ 112, 454, 36 },
		{ 112, 455, 36 },
		{ 112, 456, 36 },
		{ 112, 457, 36 },
		{ 112, 458, 36 },
		{ 112, 459, 36 },
		{ 112, 460, 36 },
		{ 112, 461, 36 },
		{ 112, 462, 36 },
		{ 112, 463, 36 },
		{ 112, 464, 36 },
		{ 112, 465, 36 },
		{ 112, 466, 36 },
		{ 112, 467, 36 },
		{ 112, 468, 36 },
		{ 112, 469, 36 },
		{ 112, 470, 36 },
		{ 112, 471, 36 },
		{ 112, 472, 36 },
		{ 112, 473, 36 },
		{ 112, 474, 36 },
		{ 112, 475, 36 },
		{ 112, 476, 36 },
		{ 112, 477, 36 },
		{ 112, 478, 36 },
		{ 112, 479, 36 },
		{ 112, 480, 36 },
		{ 112, 481, 36 },
		{ 112, 482, 36 },
		{ 112, 483, 36 },
		{ 112, 484, 36 },
		{ 112, 485, 36 },
		{ 112, 486, 36 },
		{ 112, 487, 36 },
		{ 112, 488, 36 },
		{ 113, 422, 36 },
		{ 113, 423, 36 },
		{ 113, 424, 36 },
		{ 113, 425, 36 },
		{ 113, 426, 36 },
		{ 113, 427, 36 },
		{ 113, 428, 36 },
		{ 113, 429, 36 },
		{ 113, 430, 36 },
		{ 113, 431, 36 },
		{ 113, 432, 36 },
		{ 113, 433, 36 },
		{ 113, 434, 36 },
		{ 113, 435, 36 },
		{ 113, 436, 36 },
		{ 113, 437, 36 },
		{ 113, 438, 36 },
		{ 113, 439, 36 },
		{ 113, 440, 36 },
		{ 113, 441, 36 },
		{ 113, 442, 36 },
		{ 113, 443, 36 },
		{ 113, 444, 36 },
		{ 113, 445, 36 },
		{ 113, 446, 36 },
		{ 113, 447, 36 },
		{ 113, 448, 36 },
		{ 113, 449, 36 },
		{ 113, 450, 36 },
		{ 113, 451, 36 },
		{ 113, 452, 36 },
		{ 113, 453, 36 },
		{ 113, 454, 36 },
		{ 113, 455, 36 },
		{ 113, 456, 36 },
		{ 113, 457, 36 },
		{ 113, 458, 36 },
		{ 113, 459, 36 },
		{ 113, 460, 36 },
		{ 113, 461, 36 },
		{ 113, 462, 36 },
		{ 113, 463, 36 },
		{ 113, 464, 36 },
		{ 113, 465, 36 },
		{ 113, 466, 36 },
		{ 113, 467, 36 },
		{ 113, 468, 36 },
		{ 113, 469, 36 },
		{ 113, 470, 36 },
		{ 113, 471, 36 },
		{ 113, 472, 36 },
		{ 113, 473, 36 },
		{ 113, 474, 36 },
		{ 113, 475, 36 },
		{ 113, 476, 36 },
		{ 113, 477, 36 },
		{ 113, 478, 36 },
		{ 113, 479, 36 },
		{ 113, 480, 36 },
		{ 113, 481, 36 },
		{ 113, 482, 36 },
		{ 113, 483, 36 },
		{ 113, 484, 36 },
		{ 113, 485, 36 },
		{ 113, 486, 36 },
		{ 113, 487, 36 },
		{ 114, 421, 36 },
		{ 114, 422, 36 },
		{ 114, 423, 36 },
		{ 114, 424, 36 },
		{ 114, 425, 36 },
		{ 114, 426, 36 },
		{ 114, 427, 36 },
		{ 114, 428, 36 },
		{ 114, 429, 36 },
		{ 114, 430, 36 },
		{ 114, 431, 36 },
		{ 114, 432, 36 },
		{ 114, 433, 36 },
		{ 114, 434, 36 },
		{ 114, 435, 36 },
		{ 114, 436, 36 },
		{ 114, 437, 36 },
		{ 114, 438, 36 },
		{ 114, 439, 36 },
		{ 114, 440, 36 },
		{ 114, 441, 36 },
		{ 114, 442, 36 },
		{ 114, 443, 36 },
		{ 114, 444, 36 },
		{ 114, 445, 36 },
		{ 114, 446, 36 },
		{ 114, 447, 36 },
		{ 114, 448, 36 },
		{ 114, 449, 36 },
		{ 114, 450, 36 },
		{ 114, 451, 36 },
		{ 114, 452, 36 },
		{ 114, 453, 36 },
		{ 114, 454, 36 },
		{ 114, 455, 36 },
		{ 114, 456, 36 },
		{ 114, 457, 36 },
		{ 114, 458, 36 },
		{ 114, 459, 36 },
		{ 114, 460, 36 },
		{ 114, 461, 36 },
		{ 114, 462, 36 },
		{ 114, 463, 36 },
		{ 114, 464, 36 },
		{ 114, 465, 36 },
		{ 114, 466, 36 },
		{ 114, 467, 36 },
		{ 114, 468, 36 },
		{ 114, 469, 36 },
		{ 114, 470, 36 },
		{ 114, 471, 36 },
		{ 114, 472, 36 },
		{ 114, 473, 36 },
		{ 114, 474, 36 },
		{ 114, 475, 36 },
		{ 114, 476, 36 },
		{ 114, 477, 36 },
		{ 114, 478, 36 },
		{ 114, 479, 36 },
		{ 114, 480, 36 },
		{ 114, 481, 36 },
		{ 114, 482, 36 },
		{ 114, 483, 36 },
		{ 114, 484, 36 },
		{ 114, 485, 36 },
		{ 115, 420, 36 },
		{ 115, 421, 36 },
		{ 115, 422, 36 },
		{ 115, 423, 36 },
		{ 115, 424, 36 },
		{ 115, 425, 36 },
		{ 115, 426, 36 },
		{ 115, 427, 36 },
		{ 115, 428, 36 },
		{ 115, 429, 36 },
		{ 115, 430, 36 },
		{ 115, 431, 36 },
		{ 115, 432, 36 },
		{ 115, 433, 36 },
		{ 115, 434, 36 },
		{ 115, 435, 36 },
		{ 115, 436, 36 },
		{ 115, 437, 36 },
		{ 115, 438, 36 },
		{ 115, 439, 36 },
		{ 115, 440, 36 },
		{ 115, 441, 36 },
		{ 115, 442, 36 },
		{ 115, 443, 36 },
		{ 115, 444, 36 },
		{ 115, 445, 36 },
		{ 115, 446, 36 },
		{ 115, 447, 36 },
		{ 115, 448, 36 },
		{ 115, 449, 36 },
		{ 115, 450, 36 },
		{ 115, 451, 36 },
		{ 115, 452, 36 },
		{ 115, 453, 36 },
		{ 115, 454, 36 },
		{ 115, 455, 36 },
		{ 115, 456, 36 },
		{ 115, 457, 36 },
		{ 115, 458, 36 },
		{ 115, 459, 36 },
		{ 115, 460, 36 },
		{ 115, 461, 36 },
		{ 115, 462, 36 },
		{ 115, 463, 36 },
		{ 115, 464, 36 },
		{ 115, 465, 36 },
		{ 115, 466, 36 },
		{ 115, 467, 36 },
		{ 115, 468, 36 },
		{ 115, 469, 36 },
		{ 115, 470, 36 },
		{ 115, 471, 36 },
		{ 115, 472, 36 },
		{ 115, 473, 36 },
		{ 115, 474, 36 },
		{ 115, 475, 36 },
		{ 115, 476, 36 },
		{ 115, 477, 36 },
		{ 115, 478, 36 },
		{ 115, 479, 36 },
		{ 115, 480, 36 },
		{ 115, 481, 36 },
		{ 115, 482, 36 },
		{ 115, 483, 36 },
		{ 115, 484, 36 },
		{ 116, 419, 36 },
		{ 116, 420, 36 },
		{ 116, 421, 36 },
		{ 116, 422, 36 },
		{ 116, 423, 36 },
		{ 116, 424, 36 },
		{ 116, 425, 36 },
		{ 116, 426, 36 },
		{ 116, 427, 36 },
		{ 116, 428, 36 },
		{ 116, 429, 36 },
		{ 116, 430, 36 },
		{ 116, 431, 36 },
		{ 116, 432, 36 },
		{ 116, 433, 36 },
		{ 116, 434, 36 },
		{ 116, 435, 36 },
		{ 116, 436, 36 },
		{ 116, 437, 36 },
		{ 116, 438, 36 },
		{ 116, 439, 36 },
		{ 116, 440, 36 },
		{ 116, 441, 36 },
		{ 116, 442, 36 },
		{ 116, 443, 36 },
		{ 116, 444, 36 },
		{ 116, 445, 36 },
		{ 116, 446, 36 },
		{ 116, 447, 36 },
		{ 116, 448, 36 },
		{ 116, 449, 36 },
		{ 116, 450, 36 },
		{ 116, 451, 36 },
		{ 116, 452, 36 },
		{ 116, 453, 36 },
		{ 116, 454, 36 },
		{ 116, 455, 36 },
		{ 116, 456, 36 },
		{ 116, 457, 36 },
		{ 116, 458, 36 },
		{ 116, 459, 36 },
		{ 116, 460, 36 },
		{ 116, 461, 36 },
		{ 116, 462, 36 },
		{ 116, 463, 36 },
		{ 116, 464, 36 },
		{ 116, 465, 36 },
		{ 116, 466, 36 },
		{ 116, 467, 36 },
		{ 116, 468, 36 },
		{ 116, 469, 36 },
		{ 116, 470, 36 },
		{ 116, 471, 36 },
		{ 116, 472, 36 },
		{ 116, 473, 36 },
		{ 116, 474, 36 },
		{ 116, 475, 36 },
		{ 116, 476, 36 },
		{ 116, 477, 36 },
		{ 116, 478, 36 },
		{ 116, 479, 36 },
		{ 116, 480, 36 },
		{ 116, 481, 36 },
		{ 116, 482, 36 },
		{ 117, 418, 36 },
		{ 117, 419, 36 },
		{ 117, 420, 36 },
		{ 117, 421, 36 },
		{ 117, 422, 36 },
		{ 117, 423, 36 },
		{ 117, 424, 36 },
		{ 117, 425, 36 },
		{ 117, 426, 36 },
		{ 117, 427, 36 },
		{ 117, 428, 36 },
		{ 117, 429, 36 },
		{ 117, 430, 36 },
		{ 117, 431, 36 },
		{ 117, 432, 36 },
		{ 117, 433, 36 },
		{ 117, 434, 36 },
		{ 117, 435, 36 },
		{ 117, 436, 36 },
		{ 117, 437, 36 },
		{ 117, 438, 36 },
		{ 117, 439, 36 },
		{ 117, 440, 36 },
		{ 117, 441, 36 },
		{ 117, 442, 36 },
		{ 117, 443, 36 },
		{ 117, 444, 36 },
		{ 117, 445, 36 },
		{ 117, 446, 36 },
		{ 117, 447, 36 },
		{ 117, 448, 36 },
		{ 117, 449, 36 },
		{ 117, 450, 36 },
		{ 117, 451, 36 },
		{ 117, 452, 36 },
		{ 117, 453, 36 },
		{ 117, 454, 36 },
		{ 117, 455, 36 },
		{ 117, 456, 36 },
		{ 117, 457, 36 },
		{ 117, 458, 36 },
		{ 117, 459, 36 },
		{ 117, 460, 36 },
		{ 117, 461, 36 },
		{ 117, 462, 36 },
		{ 117, 463, 36 },
		{ 117, 464, 36 },
		{ 117, 465, 36 },
		{ 117, 466, 36 },
		{ 117, 467, 36 },
		{ 117, 468, 36 },
		{ 117, 469, 36 },
		{ 117, 470, 36 },
		{ 117, 471, 36 },
		{ 117, 472, 36 },
		{ 117, 473, 36 },
		{ 117, 474, 36 },
		{ 117, 475, 36 },
		{ 117, 476, 36 },
		{ 117, 477, 36 },
		{ 117, 478, 36 },
		{ 117, 479, 36 },
		{ 117, 480, 36 },
		{ 118, 416, 36 },
		{ 118, 417, 36 },
		{ 118, 418, 36 },
		{ 118, 419, 36 },
		{ 118, 420, 36 },
		{ 118, 421, 36 },
		{ 118, 422, 36 },
		{ 118, 423, 36 },
		{ 118, 424, 36 },
		{ 118, 425, 36 },
		{ 118, 426, 36 },
		{ 118, 427, 36 },
		{ 118, 428, 36 },
		{ 118, 429, 36 },
		{ 118, 430, 36 },
		{ 118, 431, 36 },
		{ 118, 432, 36 },
		{ 118, 433, 36 },
		{ 118, 434, 36 },
		{ 118, 435, 36 },
		{ 118, 436, 36 },
		{ 118, 437, 36 },
		{ 118, 438, 36 },
		{ 118, 439, 36 },
		{ 118, 440, 36 },
		{ 118, 441, 36 },
		{ 118, 442, 36 },
		{ 118, 443, 36 },
		{ 118, 444, 36 },
		{ 118, 445, 36 },
		{ 118, 446, 36 },
		{ 118, 447, 36 },
		{ 118, 448, 36 },
		{ 118, 449, 36 },
		{ 118, 450, 36 },
		{ 118, 451, 36 },
		{ 118, 452, 36 },
		{ 118, 453, 36 },
		{ 118, 454, 36 },
		{ 118, 455, 36 },
		{ 118, 456, 36 },
		{ 118, 457, 36 },
		{ 118, 458, 36 },
		{ 118, 459, 36 },
		{ 118, 460, 36 },
		{ 118, 461, 36 },
		{ 118, 462, 36 },
		{ 118, 463, 36 },
		{ 118, 464, 36 },
		{ 118, 465, 36 },
		{ 118, 466, 36 },
		{ 118, 467, 36 },
		{ 118, 468, 36 },
		{ 118, 469, 36 },
		{ 118, 470, 36 },
		{ 118, 471, 36 },
		{ 118, 472, 36 },
		{ 118, 473, 36 },
		{ 118, 474, 36 },
		{ 118, 475, 36 },
		{ 118, 476, 36 },
		{ 118, 477, 36 },
		{ 118, 478, 36 },
		{ 118, 479, 36 },
		{ 119, 415, 36 },
		{ 119, 416, 36 },
		{ 119, 417, 36 },
		{ 119, 418, 36 },
		{ 119, 419, 36 },
		{ 119, 420, 36 },
		{ 119, 421, 36 },
		{ 119, 422, 36 },
		{ 119, 423, 36 },
		{ 119, 424, 36 },
		{ 119, 425, 36 },
		{ 119, 426, 36 },
		{ 119, 427, 36 },
		{ 119, 428, 36 },
		{ 119, 429, 36 },
		{ 119, 430, 36 },
		{ 119, 431, 36 },
		{ 119, 432, 36 },
		{ 119, 433, 36 },
		{ 119, 434, 36 },
		{ 119, 435, 36 },
		{ 119, 436, 36 },
		{ 119, 437, 36 },
		{ 119, 438, 36 },
		{ 119, 439, 36 },
		{ 119, 440, 36 },
		{ 119, 441, 36 },
		{ 119, 442, 36 },
		{ 119, 443, 36 },
		{ 119, 444, 36 },
		{ 119, 445, 36 },
		{ 119, 446, 36 },
		{ 119, 447, 36 },
		{ 119, 448, 36 },
		{ 119, 449, 36 },
		{ 119, 450, 36 },
		{ 119, 451, 36 },
		{ 119, 452, 36 },
		{ 119, 453, 36 },
		{ 119, 454, 36 },
		{ 119, 455, 36 },
		{ 119, 456, 36 },
		{ 119, 457, 36 },
		{ 119, 458, 36 },
		{ 119, 459, 36 },
		{ 119, 460, 36 },
		{ 119, 461, 36 },
		{ 119, 462, 36 },
		{ 119, 463, 36 },
		{ 119, 464, 36 },
		{ 119, 465, 36 },
		{ 119, 466, 36 },
		{ 119, 467, 36 },
		{ 119, 468, 36 },
		{ 119, 469, 36 },
		{ 119, 470, 36 },
		{ 119, 471, 36 },
		{ 119, 472, 36 },
		{ 119, 473, 36 },
		{ 119, 474, 36 },
		{ 119, 475, 36 },
		{ 119, 476, 36 },
		{ 119, 477, 36 },
		{ 120, 414, 36 },
		{ 120, 415, 36 },
		{ 120, 416, 36 },
		{ 120, 417, 36 },
		{ 120, 418, 36 },
		{ 120, 419, 36 },
		{ 120, 420, 36 },
		{ 120, 421, 36 },
		{ 120, 422, 36 },
		{ 120, 423, 36 },
		{ 120, 424, 36 },
		{ 120, 425, 36 },
		{ 120, 426, 36 },
		{ 120, 427, 36 },
		{ 120, 428, 36 },
		{ 120, 429, 36 },
		{ 120, 430, 36 },
		{ 120, 431, 36 },
		{ 120, 432, 36 },
		{ 120, 433, 36 },
		{ 120, 434, 36 },
		{ 120, 435, 36 },
		{ 120, 436, 36 },
		{ 120, 437, 36 },
		{ 120, 438, 36 },
		{ 120, 439, 36 },
		{ 120, 440, 36 },
		{ 120, 441, 36 },
		{ 120, 442, 36 },
		{ 120, 443, 36 },
		{ 120, 444, 36 },
		{ 120, 445, 36 },
		{ 120, 446, 36 },
		{ 120, 447, 36 },
		{ 120, 448, 36 },
		{ 120, 449, 36 },
		{ 120, 450, 36 },
		{ 120, 451, 36 },
		{ 120, 452, 36 },
		{ 120, 453, 36 },
		{ 120, 454, 36 },
		{ 120, 455, 36 },
		{ 120, 456, 36 },
		{ 120, 457, 36 },
		{ 120, 458, 36 },
		{ 120, 459, 36 },
		{ 120, 460, 36 },
		{ 120, 461, 36 },
		{ 120, 462, 36 },
		{ 120, 463, 36 },
		{ 120, 464, 36 },
		{ 120, 465, 36 },
		{ 120, 466, 36 },
		{ 120, 467, 36 },
		{ 120, 468, 36 },
		{ 120, 469, 36 },
		{ 120, 470, 36 },
		{ 120, 471, 36 },
		{ 120, 472, 36 },
		{ 120, 473, 36 },
		{ 120, 474, 36 },
		{ 120, 475, 36 },
		{ 120, 476, 36 },
		{ 121, 413, 36 },
		{ 121, 414, 36 },
		{ 121, 415, 36 },
		{ 121, 416, 36 },
		{ 121, 417, 36 },
		{ 121, 418, 36 },
		{ 121, 419, 36 },
		{ 121, 420, 36 },
		{ 121, 421, 36 },
		{ 121, 422, 36 },
		{ 121, 423, 36 },
		{ 121, 424, 36 },
		{ 121, 425, 36 },
		{ 121, 426, 36 },
		{ 121, 427, 36 },
		{ 121, 428, 36 },
		{ 121, 429, 36 },
		{ 121, 430, 36 },
		{ 121, 431, 36 },
		{ 121, 432, 36 },
		{ 121, 433, 36 },
		{ 121, 434, 36 },
		{ 121, 435, 36 },
		{ 121, 436, 36 },
		{ 121, 437, 36 },
		{ 121, 438, 36 },
		{ 121, 439, 36 },
		{ 121, 440, 36 },
		{ 121, 441, 36 },
		{ 121, 442, 36 },
		{ 121, 443, 36 },
		{ 121, 444, 36 },
		{ 121, 445, 36 },
		{ 121, 446, 36 },
		{ 121, 447, 36 },
		{ 121, 448, 36 },
		{ 121, 449, 36 },
		{ 121, 450, 36 },
		{ 121, 451, 36 },
		{ 121, 452, 36 },
		{ 121, 453, 36 },
		{ 121, 454, 36 },
		{ 121, 455, 36 },
		{ 121, 456, 36 },
		{ 121, 457, 36 },
		{ 121, 458, 36 },
		{ 121, 459, 36 },
		{ 121, 460, 36 },
		{ 121, 461, 36 },
		{ 121, 462, 36 },
		{ 121, 463, 36 },
		{ 121, 464, 36 },
		{ 121, 465, 36 },
		{ 121, 466, 36 },
		{ 121, 467, 36 },
		{ 121, 468, 36 },
		{ 121, 469, 36 },
		{ 121, 470, 36 },
		{ 121, 471, 36 },
		{ 121, 472, 36 },
		{ 121, 473, 36 },
		{ 121, 474, 36 },
		{ 122, 412, 36 },
		{ 122, 413, 36 },
		{ 122, 414, 36 },
		{ 122, 415, 36 },
		{ 122, 416, 36 },
		{ 122, 417, 36 },
		{ 122, 418, 36 },
		{ 122, 419, 36 },
		{ 122, 420, 36 },
		{ 122, 421, 36 },
		{ 122, 422, 36 },
		{ 122, 423, 36 },
		{ 122, 424, 36 },
		{ 122, 425, 36 },
		{ 122, 426, 36 },
		{ 122, 427, 36 },
		{ 122, 428, 36 },
		{ 122, 429, 36 },
		{ 122, 430, 36 },
		{ 122, 431, 36 },
		{ 122, 432, 36 },
		{ 122, 433, 36 },
		{ 122, 434, 36 },
		{ 122, 435, 36 },
		{ 122, 436, 36 },
		{ 122, 437, 36 },
		{ 122, 438, 36 },
		{ 122, 439, 36 },
		{ 122, 440, 36 },
		{ 122, 441, 36 },
		{ 122, 442, 36 },
		{ 122, 443, 36 },
		{ 122, 444, 36 },
		{ 122, 445, 36 },
		{ 122, 446, 36 },
		{ 122, 447, 36 },
		{ 122, 448, 36 },
		{ 122, 449, 36 },
		{ 122, 450, 36 },
		{ 122, 451, 36 },
		{ 122, 452, 36 },
		{ 122, 453, 36 },
		{ 122, 454, 36 },
		{ 122, 455, 36 },
		{ 122, 456, 36 },
		{ 122, 457, 36 },
		{ 122, 458, 36 },
		{ 122, 459, 36 },
		{ 122, 460, 36 },
		{ 122, 461, 36 },
		{ 122, 462, 36 },
		{ 122, 463, 36 },
		{ 122, 464, 36 },
		{ 122, 465, 36 },
		{ 122, 466, 36 },
		{ 122, 467, 36 },
		{ 122, 468, 36 },
		{ 122, 469, 36 },
		{ 122, 470, 36 },
		{ 122, 471, 36 },
		{ 122, 472, 36 },
		{ 123, 411, 36 },
		{ 123, 412, 36 },
		{ 123, 413, 36 },
		{ 123, 414, 36 },
		{ 123, 415, 36 },
		{ 123, 416, 36 },
		{ 123, 417, 36 },
		{ 123, 418, 36 },
		{ 123, 419, 36 },
		{ 123, 420, 36 },
		{ 123, 421, 36 },
		{ 123, 422, 36 },
		{ 123, 423, 36 },
		{ 123, 424, 36 },
		{ 123, 425, 36 },
		{ 123, 426, 36 },
		{ 123, 427, 36 },
		{ 123, 428, 36 },
		{ 123, 429, 36 },
		{ 123, 430, 36 },
		{ 123, 431, 36 },
		{ 123, 432, 36 },
		{ 123, 433, 36 },
		{ 123, 434, 36 },
		{ 123, 435, 36 },
		{ 123, 436, 36 },
		{ 123, 437, 36 },
		{ 123, 438, 36 },
		{ 123, 439, 36 },
		{ 123, 440, 36 },
		{ 123, 441, 36 },
		{ 123, 442, 36 },
		{ 123, 443, 36 },
		{ 123, 444, 36 },
		{ 123, 445, 36 },
		{ 123, 446, 36 },
		{ 123, 447, 36 },
		{ 123, 448, 36 },
		{ 123, 449, 36 },
		{ 123, 450, 36 },
		{ 123, 451, 36 },
		{ 123, 452, 36 },
		{ 123, 453, 36 },
		{ 123, 454, 36 },
		{ 123, 455, 36 },
		{ 123, 456, 36 },
		{ 123, 457, 36 },
		{ 123, 458, 36 },
		{ 123, 459, 36 },
		{ 123, 460, 36 },
		{ 123, 461, 36 },
		{ 123, 462, 36 },
		{ 123, 463, 36 },
		{ 123, 464, 36 },
		{ 123, 465, 36 },
		{ 123, 466, 36 },
		{ 123, 467, 36 },
		{ 123, 468, 36 },
		{ 123, 469, 36 },
		{ 123, 470, 36 },
		{ 123, 471, 36 },
		{ 124, 410, 36 },
		{ 124, 411, 36 },
		{ 124, 412, 36 },
		{ 124, 413, 36 },
		{ 124, 414, 36 },
		{ 124, 415, 36 },
		{ 124, 416, 36 },
		{ 124, 417, 36 },
		{ 124, 418, 36 },
		{ 124, 419, 36 },
		{ 124, 420, 36 },
		{ 124, 421, 36 },
		{ 124, 422, 36 },
		{ 124, 423, 36 },
		{ 124, 424, 36 },
		{ 124, 425, 36 },
		{ 124, 426, 36 },
		{ 124, 427, 36 },
		{ 124, 428, 36 },
		{ 124, 429, 36 },
		{ 124, 430, 36 },
		{ 124, 431, 36 },
		{ 124, 432, 36 },
		{ 124, 433, 36 },
		{ 124, 434, 36 },
		{ 124, 435, 36 },
		{ 124, 436, 36 },
		{ 124, 437, 36 },
		{ 124, 438, 36 },
		{ 124, 439, 36 },
		{ 124, 440, 36 },
		{ 124, 441, 36 },
		{ 124, 442, 36 },
		{ 124, 443, 36 },
		{ 124, 444, 36 },
		{ 124, 445, 36 },
		{ 124, 446, 36 },
		{ 124, 447, 36 },
		{ 124, 448, 36 },
		{ 124, 449, 36 },
		{ 124, 450, 36 },
		{ 124, 451, 36 },
		{ 124, 452, 36 },
		{ 124, 453, 36 },
		{ 124, 454, 36 },
		{ 124, 455, 36 },
		{ 124, 456, 36 },
		{ 124, 457, 36 },
		{ 124, 458, 36 },
		{ 124, 459, 36 },
		{ 124, 460, 36 },
		{ 124, 461, 36 },
		{ 124, 462, 36 },
		{ 124, 463, 36 },
		{ 124, 464, 36 },
		{ 124, 465, 36 },
		{ 124, 466, 36 },
		{ 124, 467, 36 },
		{ 124, 468, 36 },
		{ 124, 469, 36 },
		{ 125, 408, 36 },
		{ 125, 409, 36 },
		{ 125, 410, 36 },
		{ 125, 411, 36 },
		{ 125, 412, 36 },
		{ 125, 413, 36 },
		{ 125, 414, 36 },
		{ 125, 415, 36 },
		{ 125, 416, 36 },
		{ 125, 417, 36 },
		{ 125, 418, 36 },
		{ 125, 419, 36 },
		{ 125, 420, 36 },
		{ 125, 421, 36 },
		{ 125, 422, 36 },
		{ 125, 423, 36 },
		{ 125, 424, 36 },
		{ 125, 425, 36 },
		{ 125, 426, 36 },
		{ 125, 427, 36 },
		{ 125, 428, 36 },
		{ 125, 429, 36 },
		{ 125, 430, 36 },
		{ 125, 431, 36 },
		{ 125, 432, 36 },
		{ 125, 433, 36 },
		{ 125, 434, 36 },
		{ 125, 435, 36 },
		{ 125, 436, 36 },
		{ 125, 437, 36 },
		{ 125, 438, 36 },
		{ 125, 439, 36 },
		{ 125, 440, 36 },
		{ 125, 441, 36 },
		{ 125, 442, 36 },
		{ 125, 443, 36 },
		{ 125, 444, 36 },
		{ 125, 445, 36 },
		{ 125, 446, 36 },
		{ 125, 447, 36 },
		{ 125, 448, 36 },
		{ 125, 449, 36 },
		{ 125, 450, 36 },
		{ 125, 451, 36 },
		{ 125, 452, 36 },
		{ 125, 453, 36 },
		{ 125, 454, 36 },
		{ 125, 455, 36 },
		{ 125, 456, 36 },
		{ 125, 457, 36 },
		{ 125, 458, 36 },
		{ 125, 459, 36 },
		{ 125, 460, 36 },
		{ 125, 461, 36 },
		{ 125, 462, 36 },
		{ 125, 463, 36 },
		{ 125, 464, 36 },
		{ 125, 465, 36 },
		{ 125, 466, 36 },
		{ 125, 467, 36 },
		{ 125, 468, 36 },
		{ 126, 407, 36 },
		{ 126, 408, 36 },
		{ 126, 409, 36 },
		{ 126, 410, 36 },
		{ 126, 411, 36 },
		{ 126, 412, 36 },
		{ 126, 413, 36 },
		{ 126, 414, 36 },
		{ 126, 415, 36 },
		{ 126, 416, 36 },
		{ 126, 417, 36 },
		{ 126, 418, 36 },
		{ 126, 419, 36 },
		{ 126, 420, 36 },
		{ 126, 421, 36 },
		{ 126, 422, 36 },
		{ 126, 423, 36 },
		{ 126, 424, 36 },
		{ 126, 425, 36 },
		{ 126, 426, 36 },
		{ 126, 427, 36 },
		{ 126, 428, 36 },
		{ 126, 429, 36 },
		{ 126, 430, 36 },
		{ 126, 431, 36 },
		{ 126, 432, 36 },
		{ 126, 433, 36 },
		{ 126, 434, 36 },
		{ 126, 435, 36 },
		{ 126, 436, 36 },
		{ 126, 437, 36 },
		{ 126, 438, 36 },
		{ 126, 439, 36 },
		{ 126, 440, 36 },
		{ 126, 441, 36 },
		{ 126, 442, 36 },
		{ 126, 443, 36 },
		{ 126, 444, 36 },
		{ 126, 445, 36 },
		{ 126, 446, 36 },
		{ 126, 447, 36 },
		{ 126, 448, 36 },
		{ 126, 449, 36 },
		{ 126, 450, 36 },
		{ 126, 451, 36 },
		{ 126, 452, 36 },
		{ 126, 453, 36 },
		{ 126, 454, 36 },
		{ 126, 455, 36 },
		{ 126, 456, 36 },
		{ 126, 457, 36 },
		{ 126, 458, 36 },
		{ 126, 459, 36 },
		{ 126, 460, 36 },
		{ 126, 461, 36 },
		{ 126, 462, 36 },
		{ 126, 463, 36 },
		{ 126, 464, 36 },
		{ 126, 465, 36 },
		{ 126, 466, 36 },
		{ 127, 406, 36 },
		{ 127, 407, 36 },
		{ 127, 408, 36 },
		{ 127, 409, 36 },
		{ 127, 410, 36 },
		{ 127, 411, 36 },
		{ 127, 412, 36 },
		{ 127, 413, 36 },
		{ 127, 414, 36 },
		{ 127, 415, 36 },
		{ 127, 416, 36 },
		{ 127, 417, 36 },
		{ 127, 418, 36 },
		{ 127, 419, 36 },
		{ 127, 420, 36 },
		{ 127, 421, 36 },
		{ 127, 422, 36 },
		{ 127, 423, 36 },
		{ 127, 424, 36 },
		{ 127, 425, 36 },
		{ 127, 426, 36 },
		{ 127, 427, 36 },
		{ 127, 428, 36 },
		{ 127, 429, 36 },
		{ 127, 430, 36 },
		{ 127, 431, 36 },
		{ 127, 432, 36 },
		{ 127, 433, 36 },
		{ 127, 434, 36 },
		{ 127, 435, 36 },
		{ 127, 436, 36 },
		{ 127, 437, 36 },
		{ 127, 438, 36 },
		{ 127, 439, 36 },
		{ 127, 440, 36 },
		{ 127, 441, 36 },
		{ 127, 442, 36 },
		{ 127, 443, 36 },
		{ 127, 444, 36 },
		{ 127, 445, 36 },
		{ 127, 446, 36 },
		{ 127, 447, 36 },
		{ 127, 448, 36 },
		{ 127, 449, 36 },
		{ 127, 450, 36 },
		{ 127, 451, 36 },
		{ 127, 452, 36 },
		{ 127, 453, 36 },
		{ 127, 454, 36 },
		{ 127, 455, 36 },
		{ 127, 456, 36 },
		{ 127, 457, 36 },
		{ 127, 458, 36 },
		{ 127, 459, 36 },
		{ 127, 460, 36 },
		{ 127, 461, 36 },
		{ 127, 462, 36 },
		{ 127, 463, 36 },
		{ 127, 464, 36 },
		{ 128, 405, 36 },
		{ 128, 406, 36 },
		{ 128, 407, 36 },
		{ 128, 408, 36 },
		{ 128, 409, 36 },
		{ 128, 410, 36 },
		{ 128, 411, 36 },
		{ 128, 412, 36 },
		{ 128, 413, 36 },
		{ 128, 414, 36 },
		{ 128, 415, 36 },
		{ 128, 416, 36 },
		{ 128, 417, 36 },
		{ 128, 418, 36 },
		{ 128, 419, 36 },
		{ 128, 420, 36 },
		{ 128, 421, 36 },
		{ 128, 422, 36 },
		{ 128, 423, 36 },
		{ 128, 424, 36 },
		{ 128, 425, 36 },
		{ 128, 426, 36 },
		{ 128, 427, 36 },
		{ 128, 428, 36 },
		{ 128, 429, 36 },
		{ 128, 430, 36 },
		{ 128, 431, 36 },
		{ 128, 432, 36 },
		{ 128, 433, 36 },
		{ 128, 434, 36 },
		{ 128, 435, 36 },
		{ 128, 436, 36 },
		{ 128, 437, 36 },
		{ 128, 438, 36 },
		{ 128, 439, 36 },
		{ 128, 440, 36 },
		{ 128, 441, 36 },
		{ 128, 442, 36 },
		{ 128, 443, 36 },
		{ 128, 444, 36 },
		{ 128, 445, 36 },
		{ 128, 446, 36 },
		{ 128, 447, 36 },
		{ 128, 448, 36 },
		{ 128, 449, 36 },
		{ 128, 450, 36 },
		{ 128, 451, 36 },
		{ 128, 452, 36 },
		{ 128, 453, 36 },
		{ 128, 454, 36 },
		{ 128, 455, 36 },
		{ 128, 456, 36 },
		{ 128, 457, 36 },
		{ 128, 458, 36 },
		{ 128, 459, 36 },
		{ 128, 460, 36 },
		{ 128, 461, 36 },
		{ 128, 462, 36 },
		{ 128, 463, 36 },
		{ 129, 404, 36 },
		{ 129, 405, 36 },
		{ 129, 406, 36 },
		{ 129, 407, 36 },
		{ 129, 408, 36 },
		{ 129, 409, 36 },
		{ 129, 410, 36 },
		{ 129, 411, 36 },
		{ 129, 412, 36 },
		{ 129, 413, 36 },
		{ 129, 414, 36 },
		{ 129, 415, 36 },
		{ 129, 416, 36 },
		{ 129, 417, 36 },
		{ 129, 418, 36 },
		{ 129, 419, 36 },
		{ 129, 420, 36 },
		{ 129, 421, 36 },
		{ 129, 422, 36 },
		{ 129, 423, 36 },
		{ 129, 424, 36 },
		{ 129, 425, 36 },
		{ 129, 426, 36 },
		{ 129, 427, 36 },
		{ 129, 428, 36 },
		{ 129, 429, 36 },
		{ 129, 430, 36 },
		{ 129, 431, 36 },
		{ 129, 432, 36 },
		{ 129, 433, 36 },
		{ 129, 434, 36 },
		{ 129, 435, 36 },
		{ 129, 436, 36 },
		{ 129, 437, 36 },
		{ 129, 438, 36 },
		{ 129, 439, 36 },
		{ 129, 440, 36 },
		{ 129, 441, 36 },
		{ 129, 442, 36 },
		{ 129, 443, 36 },
		{ 129, 444, 36 },
		{ 129, 445, 36 },
		{ 129, 446, 36 },
		{ 129, 447, 36 },
		{ 129, 448, 36 },
		{ 129, 449, 36 },
		{ 129, 450, 36 },
		{ 129, 451, 36 },
		{ 129, 452, 36 },
		{ 129, 453, 36 },
		{ 129, 454, 36 },
		{ 129, 455, 36 },
		{ 129, 456, 36 },
		{ 129, 457, 36 },
		{ 129, 458, 36 },
		{ 129, 459, 36 },
		{ 129, 460, 36 },
		{ 129, 461, 36 },
		{ 130, 403, 36 },
		{ 130, 404, 36 },
		{ 130, 405, 36 },
		{ 130, 406, 36 },
		{ 130, 407, 36 },
		{ 130, 408, 36 },
		{ 130, 409, 36 },
		{ 130, 410, 36 },
		{ 130, 411, 36 },
		{ 130, 412, 36 },
		{ 130, 413, 36 },
		{ 130, 414, 36 },
		{ 130, 415, 36 },
		{ 130, 416, 36 },
		{ 130, 417, 36 },
		{ 130, 418, 36 },
		{ 130, 419, 36 },
		{ 130, 420, 36 },
		{ 130, 421, 36 },
		{ 130, 422, 36 },
		{ 130, 423, 36 },
		{ 130, 424, 36 },
		{ 130, 425, 36 },
		{ 130, 426, 36 },
		{ 130, 427, 36 },
		{ 130, 428, 36 },
		{ 130, 429, 36 },
		{ 130, 430, 36 },
		{ 130, 431, 36 },
		{ 130, 432, 36 },
		{ 130, 433, 36 },
		{ 130, 434, 36 },
		{ 130, 435, 36 },
		{ 130, 436, 36 },
		{ 130, 437, 36 },
		{ 130, 438, 36 },
		{ 130, 439, 36 },
		{ 130, 440, 36 },
		{ 130, 441, 36 },
		{ 130, 442, 36 },
		{ 130, 443, 36 },
		{ 130, 444, 36 },
		{ 130, 445, 36 },
		{ 130, 446, 36 },
		{ 130, 447, 36 },
		{ 130, 448, 36 },
		{ 130, 449, 36 },
		{ 130, 450, 36 },
		{ 130, 451, 36 },
		{ 130, 452, 36 },
		{ 130, 453, 36 },
		{ 130, 454, 36 },
		{ 130, 455, 36 },
		{ 130, 456, 36 },
		{ 130, 457, 36 },
		{ 130, 458, 36 },
		{ 130, 459, 36 },
		{ 130, 460, 36 },
		{ 131, 402, 36 },
		{ 131, 403, 36 },
		{ 131, 404, 36 },
		{ 131, 405, 36 },
		{ 131, 406, 36 },
		{ 131, 407, 36 },
		{ 131, 408, 36 },
		{ 131, 409, 36 },
		{ 131, 410, 36 },
		{ 131, 411, 36 },
		{ 131, 412, 36 },
		{ 131, 413, 36 },
		{ 131, 414, 36 },
		{ 131, 415, 36 },
		{ 131, 416, 36 },
		{ 131, 417, 36 },
		{ 131, 418, 36 },
		{ 131, 419, 36 },
		{ 131, 420, 36 },
		{ 131, 421, 36 },
		{ 131, 422, 36 },
		{ 131, 423, 36 },
		{ 131, 424, 36 },
		{ 131, 425, 36 },
		{ 131, 426, 36 },
		{ 131, 427, 36 },
		{ 131, 428, 36 },
		{ 131, 429, 36 },
		{ 131, 430, 36 },
		{ 131, 431, 36 },
		{ 131, 432, 36 },
		{ 131, 433, 36 },
		{ 131, 434, 36 },
		{ 131, 435, 36 },
		{ 131, 436, 36 },
		{ 131, 437, 36 },
		{ 131, 438, 36 },
		{ 131, 439, 36 },
		{ 131, 440, 36 },
		{ 131, 441, 36 },
		{ 131, 442, 36 },
		{ 131, 443, 36 },
		{ 131, 444, 36 },
		{ 131, 445, 36 },
		{ 131, 446, 36 },
		{ 131, 447, 36 },
		{ 131, 448, 36 },
		{ 131, 449, 36 },
		{ 131, 450, 36 },
		{ 131, 451, 36 },
		{ 131, 452, 36 },
		{ 131, 453, 36 },
		{ 131, 454, 36 },
		{ 131, 455, 36 },
		{ 131, 456, 36 },
		{ 131, 457, 36 },
		{ 131, 458, 36 },
		{ 132, 400, 36 },
		{ 132, 401, 36 },
		{ 132, 402, 36 },
		{ 132, 403, 36 },
		{ 132, 404, 36 },
		{ 132, 405, 36 },
		{ 132, 406, 36 },
		{ 132, 407, 36 },
		{ 132, 408, 36 },
		{ 132, 409, 36 },
		{ 132, 410, 36 },
		{ 132, 411, 36 },
		{ 132, 412, 36 },
		{ 132, 413, 36 },
		{ 132, 414, 36 },
		{ 132, 415, 36 },
		{ 132, 416, 36 },
		{ 132, 417, 36 },
		{ 132, 418, 36 },
		{ 132, 419, 36 },
		{ 132, 420, 36 },
		{ 132, 421, 36 },
		{ 132, 422, 36 },
		{ 132, 423, 36 },
		{ 132, 424, 36 },
		{ 132, 425, 36 },
		{ 132, 426, 36 },
		{ 132, 427, 36 },
		{ 132, 428, 36 },
		{ 132, 429, 36 },
		{ 132, 430, 36 },
		{ 132, 431, 36 },
		{ 132, 432, 36 },
		{ 132, 433, 36 },
		{ 132, 434, 36 },
		{ 132, 435, 36 },
		{ 132, 436, 36 },
		{ 132, 437, 36 },
		{ 132, 438, 36 },
		{ 132, 439, 36 },
		{ 132, 440, 36 },
		{ 132, 441, 36 },
		{ 132, 442, 36 },
		{ 132, 443, 36 },
		{ 132, 444, 36 },
		{ 132, 445, 36 },
		{ 132, 446, 36 },
		{ 132, 447, 36 },
		{ 132, 448, 36 },
		{ 132, 449, 36 },
		{ 132, 450, 36 },
		{ 132, 451, 36 },
		{ 132, 452, 36 },
		{ 132, 453, 36 },
		{ 132, 454, 36 },
		{ 132, 455, 36 },
		{ 132, 456, 36 },
		{ 133, 399, 36 },
		{ 133, 400, 36 },
		{ 133, 401, 36 },
		{ 133, 402, 36 },
		{ 133, 403, 36 },
		{ 133, 404, 36 },
		{ 133, 405, 36 },
		{ 133, 406, 36 },
		{ 133, 407, 36 },
		{ 133, 408, 36 },
		{ 133, 409, 36 },
		{ 133, 410, 36 },
		{ 133, 411, 36 },
		{ 133, 412, 36 },
		{ 133, 413, 36 },
		{ 133, 414, 36 },
		{ 133, 415, 36 },
		{ 133, 416, 36 },
		{ 133, 417, 36 },
		{ 133, 418, 36 },
		{ 133, 419, 36 },
		{ 133, 420, 36 },
		{ 133, 421, 36 },
		{ 133, 422, 36 },
		{ 133, 423, 36 },
		{ 133, 424, 36 },
		{ 133, 425, 36 },
		{ 133, 426, 36 },
		{ 133, 427, 36 },
		{ 133, 428, 36 },
		{ 133, 429, 36 },
		{ 133, 430, 36 },
		{ 133, 431, 36 },
		{ 133, 432, 36 },
		{ 133, 433, 36 },
		{ 133, 434, 36 },
		{ 133, 435, 36 },
		{ 133, 436, 36 },
		{ 133, 437, 36 },
		{ 133, 438, 36 },
		{ 133, 439, 36 },
		{ 133, 440, 36 },
		{ 133, 441, 36 },
		{ 133, 442, 36 },
		{ 133, 443, 36 },
		{ 133, 444, 36 },
		{ 133, 445, 36 },
		{ 133, 446, 36 },
		{ 133, 447, 36 },
		{ 133, 448, 36 },
		{ 133, 449, 36 },
		{ 133, 450, 36 },
		{ 133, 451, 36 },
		{ 133, 452, 36 },
		{ 133, 453, 36 },
		{ 133, 454, 36 },
		{ 133, 455, 36 },
		{ 134, 398, 36 },
		{ 134, 399, 36 },
		{ 134, 400, 36 },
		{ 134, 401, 36 },
		{ 134, 402, 36 },
		{ 134, 403, 36 },
		{ 134, 404, 36 },
		{ 134, 405, 36 },
		{ 134, 406, 36 },
		{ 134, 407, 36 },
		{ 134, 408, 36 },
		{ 134, 409, 36 },
		{ 134, 410, 36 },
		{ 134, 411, 36 },
		{ 134, 412, 36 },
		{ 134, 413, 36 },
		{ 134, 414, 36 },
		{ 134, 415, 36 },
		{ 134, 416, 36 },
		{ 134, 417, 36 },
		{ 134, 418, 36 },
		{ 134, 419, 36 },
		{ 134, 420, 36 },
		{ 134, 421, 36 },
		{ 134, 422, 36 },
		{ 134, 423, 36 },
		{ 134, 424, 36 },
		{ 134, 425, 36 },
		{ 134, 426, 36 },
		{ 134, 427, 36 },
		{ 134, 428, 36 },
		{ 134, 429, 36 },
		{ 134, 430, 36 },
		{ 134, 431, 36 },
		{ 134, 432, 36 },
		{ 134, 433, 36 },
		{ 134, 434, 36 },
		{ 134, 435, 36 },
		{ 134, 436, 36 },
		{ 134, 437, 36 },
		{ 134, 438, 36 },
		{ 134, 439, 36 },
		{ 134, 440, 36 },
		{ 134, 441, 36 },
		{ 134, 442, 36 },
		{ 134, 443, 36 },
		{ 134, 444, 36 },
		{ 134, 445, 36 },
		{ 134, 446, 36 },
		{ 134, 447, 36 },
		{ 134, 448, 36 },
		{ 134, 449, 36 },
		{ 134, 450, 36 },
		{ 134, 451, 36 },
		{ 134, 452, 36 },
		{ 134, 453, 36 },
		{ 135, 397, 36 },
		{ 135, 398, 36 },
		{ 135, 399, 36 },
		{ 135, 400, 36 },
		{ 135, 401, 36 },
		{ 135, 402, 36 },
		{ 135, 403, 36 },
		{ 135, 404, 36 },
		{ 135, 405, 36 },
		{ 135, 406, 36 },
		{ 135, 407, 36 },
		{ 135, 408, 36 },
		{ 135, 409, 36 },
		{ 135, 410, 36 },
		{ 135, 411, 36 },
		{ 135, 412, 36 },
		{ 135, 413, 36 },
		{ 135, 414, 36 },
		{ 135, 415, 36 },
		{ 135, 416, 36 },
		{ 135, 417, 36 },
		{ 135, 418, 36 },
		{ 135, 419, 36 },
		{ 135, 420, 36 },
		{ 135, 421, 36 },
		{ 135, 422, 36 },
		{ 135, 423, 36 },
		{ 135, 424, 36 },
		{ 135, 425, 36 },
		{ 135, 426, 36 },
		{ 135, 427, 36 },
		{ 135, 428, 36 },
		{ 135, 429, 36 },
		{ 135, 430, 36 },
		{ 135, 431, 36 },
		{ 135, 432, 36 },
		{ 135, 433, 36 },
		{ 135, 434, 36 },
		{ 135, 435, 36 },
		{ 135, 436, 36 },
		{ 135, 437, 36 },
		{ 135, 438, 36 },
		{ 135, 439, 36 },
		{ 135, 440, 36 },
		{ 135, 441, 36 },
		{ 135, 442, 36 },
		{ 135, 443, 36 },
		{ 135, 444, 36 },
		{ 135, 445, 36 },
		{ 135, 446, 36 },
		{ 135, 447, 36 },
		{ 135, 448, 36 },
		{ 135, 449, 36 },
		{ 135, 450, 36 },
		{ 135, 451, 36 },
		{ 135, 452, 36 },
		{ 136, 396, 36 },
		{ 136, 397, 36 },
		{ 136, 398, 36 },
		{ 136, 399, 36 },
		{ 136, 400, 36 },
		{ 136, 401, 36 },
		{ 136, 402, 36 },
		{ 136, 403, 36 },
		{ 136, 404, 36 },
		{ 136, 405, 36 },
		{ 136, 406, 36 },
		{ 136, 407, 36 },
		{ 136, 408, 36 },
		{ 136, 409, 36 },
		{ 136, 410, 36 },
		{ 136, 411, 36 },
		{ 136, 412, 36 },
		{ 136, 413, 36 },
		{ 136, 414, 36 },
		{ 136, 415, 36 },
		{ 136, 416, 36 },
		{ 136, 417, 36 },
		{ 136, 418, 36 },
		{ 136, 419, 36 },
		{ 136, 420, 36 },
		{ 136, 421, 36 },
		{ 136, 422, 36 },
		{ 136, 423, 36 },
		{ 136, 424, 36 },
		{ 136, 425, 36 },
		{ 136, 426, 36 },
		{ 136, 427, 36 },
		{ 136, 428, 36 },
		{ 136, 429, 36 },
		{ 136, 430, 36 },
		{ 136, 431, 36 },
		{ 136, 432, 36 },
		{ 136, 433, 36 },
		{ 136, 434, 36 },
		{ 136, 435, 36 },
		{ 136, 436, 36 },
		{ 136, 437, 36 },
		{ 136, 438, 36 },
		{ 136, 439, 36 },
		{ 136, 440, 36 },
		{ 136, 441, 36 },
		{ 136, 442, 36 },
		{ 136, 443, 36 },
		{ 136, 444, 36 },
		{ 136, 445, 36 },
		{ 136, 446, 36 },
		{ 136, 447, 36 },
		{ 136, 448, 36 },
		{ 136, 449, 36 },
		{ 136, 450, 36 },
		{ 137, 395, 36 },
		{ 137, 396, 36 },
		{ 137, 397, 36 },
		{ 137, 398, 36 },
		{ 137, 399, 36 },
		{ 137, 400, 36 },
		{ 137, 401, 36 },
		{ 137, 402, 36 },
		{ 137, 403, 36 },
		{ 137, 404, 36 },
		{ 137, 405, 36 },
		{ 137, 406, 36 },
		{ 137, 407, 36 },
		{ 137, 408, 36 },
		{ 137, 409, 36 },
		{ 137, 410, 36 },
		{ 137, 411, 36 },
		{ 137, 412, 36 },
		{ 137, 413, 36 },
		{ 137, 414, 36 },
		{ 137, 415, 36 },
		{ 137, 416, 36 },
		{ 137, 417, 36 },
		{ 137, 418, 36 },
		{ 137, 419, 36 },
		{ 137, 420, 36 },
		{ 137, 421, 36 },
		{ 137, 422, 36 },
		{ 137, 423, 36 },
		{ 137, 424, 36 },
		{ 137, 425, 36 },
		{ 137, 426, 36 },
		{ 137, 427, 36 },
		{ 137, 428, 36 },
		{ 137, 429, 36 },
		{ 137, 430, 36 },
		{ 137, 431, 36 },
		{ 137, 432, 36 },
		{ 137, 433, 36 },
		{ 137, 434, 36 },
		{ 137, 435, 36 },
		{ 137, 436, 36 },
		{ 137, 437, 36 },
		{ 137, 438, 36 },
		{ 137, 439, 36 },
		{ 137, 440, 36 },
		{ 137, 441, 36 },
		{ 137, 442, 36 },
		{ 137, 443, 36 },
		{ 137, 444, 36 },
		{ 137, 445, 36 },
		{ 137, 446, 36 },
		{ 137, 447, 36 },
		{ 137, 448, 36 },
		{ 138, 394, 36 },
		{ 138, 395, 36 },
		{ 138, 396, 36 },
		{ 138, 397, 36 },
		{ 138, 398, 36 },
		{ 138, 399, 36 },
		{ 138, 400, 36 },
		{ 138, 401, 36 },
		{ 138, 402, 36 },
		{ 138, 403, 36 },
		{ 138, 404, 36 },
		{ 138, 405, 36 },
		{ 138, 406, 36 },
		{ 138, 407, 36 },
		{ 138, 408, 36 },
		{ 138, 409, 36 },
		{ 138, 410, 36 },
		{ 138, 411, 36 },
		{ 138, 412, 36 },
		{ 138, 413, 36 },
		{ 138, 414, 36 },
		{ 138, 415, 36 },
		{ 138, 416, 36 },
		{ 138, 417, 36 },
		{ 138, 418, 36 },
		{ 138, 419, 36 },
		{ 138, 420, 36 },
		{ 138, 421, 36 },
		{ 138, 422, 36 },
		{ 138, 423, 36 },
		{ 138, 424, 36 },
		{ 138, 425, 36 },
		{ 138, 426, 36 },
		{ 138, 427, 36 },
		{ 138, 428, 36 },
		{ 138, 429, 36 },
		{ 138, 430, 36 },
		{ 138, 431, 36 },
		{ 138, 432, 36 },
		{ 138, 433, 36 },
		{ 138, 434, 36 },
		{ 138, 435, 36 },
		{ 138, 436, 36 },
		{ 138, 437, 36 },
		{ 138, 438, 36 },
		{ 138, 439, 36 },
		{ 138, 440, 36 },
		{ 138, 441, 36 },
		{ 138, 442, 36 },
		{ 138, 443, 36 },
		{ 138, 444, 36 },
		{ 138, 445, 36 },
		{ 138, 446, 36 },
		{ 138, 447, 36 },
		{ 139, 392, 36 },
		{ 139, 393, 36 },
		{ 139, 394, 36 },
		{ 139, 395, 36 },
		{ 139, 396, 36 },
		{ 139, 397, 36 },
		{ 139, 398, 36 },
		{ 139, 399, 36 },
		{ 139, 400, 36 },
		{ 139, 401, 36 },
		{ 139, 402, 36 },
		{ 139, 403, 36 },
		{ 139, 404, 36 },
		{ 139, 405, 36 },
		{ 139, 406, 36 },
		{ 139, 407, 36 },
		{ 139, 408, 36 },
		{ 139, 409, 36 },
		{ 139, 410, 36 },
		{ 139, 411, 36 },
		{ 139, 412, 36 },
		{ 139, 413, 36 },
		{ 139, 414, 36 },
		{ 139, 415, 36 },
		{ 139, 416, 36 },
		{ 139, 417, 36 },
		{ 139, 418, 36 },
		{ 139, 419, 36 },
		{ 139, 420, 36 },
		{ 139, 421, 36 },
		{ 139, 422, 36 },
		{ 139, 423, 36 },
		{ 139, 424, 36 },
		{ 139, 425, 36 },
		{ 139, 426, 36 },
		{ 139, 427, 36 },
		{ 139, 428, 36 },
		{ 139, 429, 36 },
		{ 139, 430, 36 },
		{ 139, 431, 36 },
		{ 139, 432, 36 },
		{ 139, 433, 36 },
		{ 139, 434, 36 },
		{ 139, 435, 36 },
		{ 139, 436, 36 },
		{ 139, 437, 36 },
		{ 139, 438, 36 },
		{ 139, 439, 36 },
		{ 139, 440, 36 },
		{ 139, 441, 36 },
		{ 139, 442, 36 },
		{ 139, 443, 36 },
		{ 139, 444, 36 },
		{ 139, 445, 36 },
		{ 140, 391, 36 },
		{ 140, 392, 36 },
		{ 140, 393, 36 },
		{ 140, 394, 36 },
		{ 140, 395, 36 },
		{ 140, 396, 36 },
		{ 140, 397, 36 },
		{ 140, 398, 36 },
		{ 140, 399, 36 },
		{ 140, 400, 36 },
		{ 140, 401, 36 },
		{ 140, 402, 36 },
		{ 140, 403, 36 },
		{ 140, 404, 36 },
		{ 140, 405, 36 },
		{ 140, 406, 36 },
		{ 140, 407, 36 },
		{ 140, 408, 36 },
		{ 140, 409, 36 },
		{ 140, 410, 36 },
		{ 140, 411, 36 },
		{ 140, 412, 36 },
		{ 140, 413, 36 },
		{ 140, 414, 36 },
		{ 140, 415, 36 },
		{ 140, 416, 36 },
		{ 140, 417, 36 },
		{ 140, 418, 36 },
		{ 140, 419, 36 },
		{ 140, 420, 36 },
		{ 140, 421, 36 },
		{ 140, 422, 36 },
		{ 140, 423, 36 },
		{ 140, 424, 36 },
		{ 140, 425, 36 },
		{ 140, 426, 36 },
		{ 140, 427, 36 },
		{ 140, 428, 36 },
		{ 140, 429, 36 },
		{ 140, 430, 36 },
		{ 140, 431, 36 },
		{ 140, 432, 36 },
		{ 140, 433, 36 },
		{ 140, 434, 36 },
		{ 140, 435, 36 },
		{ 140, 436, 36 },
		{ 140, 437, 36 },
		{ 140, 438, 36 },
		{ 140, 439, 36 },
		{ 140, 440, 36 },
		{ 140, 441, 36 },
		{ 140, 442, 36 },
		{ 140, 443, 36 },
		{ 140, 444, 36 },
		{ 141, 390, 36 },
		{ 141, 391, 36 },
		{ 141, 392, 36 },
		{ 141, 393, 36 },
		{ 141, 394, 36 },
		{ 141, 395, 36 },
		{ 141, 396, 36 },
		{ 141, 397, 36 },
		{ 141, 398, 36 },
		{ 141, 399, 36 },
		{ 141, 400, 36 },
		{ 141, 401, 36 },
		{ 141, 402, 36 },
		{ 141, 403, 36 },
		{ 141, 404, 36 },
		{ 141, 405, 36 },
		{ 141, 406, 36 },
		{ 141, 407, 36 },
		{ 141, 408, 36 },
		{ 141, 409, 36 },
		{ 141, 410, 36 },
		{ 141, 411, 36 },
		{ 141, 412, 36 },
		{ 141, 413, 36 },
		{ 141, 414, 36 },
		{ 141, 415, 36 },
		{ 141, 416, 36 },
		{ 141, 417, 36 },
		{ 141, 418, 36 },
		{ 141, 419, 36 },
		{ 141, 420, 36 },
		{ 141, 421, 36 },
		{ 141, 422, 36 },
		{ 141, 423, 36 },
		{ 141, 424, 36 },
		{ 141, 425, 36 },
		{ 141, 426, 36 },
		{ 141, 427, 36 },
		{ 141, 428, 36 },
		{ 141, 429, 36 },
		{ 141, 430, 36 },
		{ 141, 431, 36 },
		{ 141, 432, 36 },
		{ 141, 433, 36 },
		{ 141, 434, 36 },
		{ 141, 435, 36 },
		{ 141, 436, 36 },
		{ 141, 437, 36 },
		{ 141, 438, 36 },
		{ 141, 439, 36 },
		{ 141, 440, 36 },
		{ 141, 441, 36 },
		{ 141, 442, 36 },
		{ 142, 389, 36 },
		{ 142, 390, 36 },
		{ 142, 391, 36 },
		{ 142, 392, 36 },
		{ 142, 393, 36 },
		{ 142, 394, 36 },
		{ 142, 395, 36 },
		{ 142, 396, 36 },
		{ 142, 397, 36 },
		{ 142, 398, 36 },
		{ 142, 399, 36 },
		{ 142, 400, 36 },
		{ 142, 401, 36 },
		{ 142, 402, 36 },
		{ 142, 403, 36 },
		{ 142, 404, 36 },
		{ 142, 405, 36 },
		{ 142, 406, 36 },
		{ 142, 407, 36 },
		{ 142, 408, 36 },
		{ 142, 409, 36 },
		{ 142, 410, 36 },
		{ 142, 411, 36 },
		{ 142, 412, 36 },
		{ 142, 413, 36 },
		{ 142, 414, 36 },
		{ 142, 415, 36 },
		{ 142, 416, 36 },
		{ 142, 417, 36 },
		{ 142, 418, 36 },
		{ 142, 419, 36 },
		{ 142, 420, 36 },
		{ 142, 421, 36 },
		{ 142, 422, 36 },
		{ 142, 423, 36 },
		{ 142, 424, 36 },
		{ 142, 425, 36 },
		{ 142, 426, 36 },
		{ 142, 427, 36 },
		{ 142, 428, 36 },
		{ 142, 429, 36 },
		{ 142, 430, 36 },
		{ 142, 431, 36 },
		{ 142, 432, 36 },
		{ 142, 433, 36 },
		{ 142, 434, 36 },
		{ 142, 435, 36 },
		{ 142, 436, 36 },
		{ 142, 437, 36 },
		{ 142, 438, 36 },
		{ 142, 439, 36 },
		{ 142, 440, 36 },
		{ 143, 388, 36 },
		{ 143, 389, 36 },
		{ 143, 390, 36 },
		{ 143, 391, 36 },
		{ 143, 392, 36 },
		{ 143, 393, 36 },
		{ 143, 394, 36 },
		{ 143, 395, 36 },
		{ 143, 396, 36 },
		{ 143, 397, 36 },
		{ 143, 398, 36 },
		{ 143, 399, 36 },
		{ 143, 400, 36 },
		{ 143, 401, 36 },
		{ 143, 402, 36 },
		{ 143, 403, 36 },
		{ 143, 404, 36 },
		{ 143, 405, 36 },
		{ 143, 406, 36 },
		{ 143, 407, 36 },
		{ 143, 408, 36 },
		{ 143, 409, 36 },
		{ 143, 410, 36 },
		{ 143, 411, 36 },
		{ 143, 412, 36 },
		{ 143, 413, 36 },
		{ 143, 414, 36 },
		{ 143, 415, 36 },
		{ 143, 416, 36 },
		{ 143, 417, 36 },
		{ 143, 418, 36 },
		{ 143, 419, 36 },
		{ 143, 420, 36 },
		{ 143, 421, 36 },
		{ 143, 422, 36 },
		{ 143, 423, 36 },
		{ 143, 424, 36 },
		{ 143, 425, 36 },
		{ 143, 426, 36 },
		{ 143, 427, 36 },
		{ 143, 428, 36 },
		{ 143, 429, 36 },
		{ 143, 430, 36 },
		{ 143, 431, 36 },
		{ 143, 432, 36 },
		{ 143, 433, 36 },
		{ 143, 434, 36 },
		{ 143, 435, 36 },
		{ 143, 436, 36 },
		{ 143, 437, 36 },
		{ 143, 438, 36 },
		{ 143, 439, 36 },
		{ 144, 387, 36 },
		{ 144, 388, 36 },
		{ 144, 389, 36 },
		{ 144, 390, 36 },
		{ 144, 391, 36 },
		{ 144, 392, 36 },
		{ 144, 393, 36 },
		{ 144, 394, 36 },
		{ 144, 395, 36 },
		{ 144, 396, 36 },
		{ 144, 397, 36 },
		{ 144, 398, 36 },
		{ 144, 399, 36 },
		{ 144, 400, 36 },
		{ 144, 401, 36 },
		{ 144, 402, 36 },
		{ 144, 403, 36 },
		{ 144, 404, 36 },
		{ 144, 405, 36 },
		{ 144, 406, 36 },
		{ 144, 407, 36 },
		{ 144, 408, 36 },
		{ 144, 409, 36 },
		{ 144, 410, 36 },
		{ 144, 411, 36 },
		{ 144, 412, 36 },
		{ 144, 413, 36 },
		{ 144, 414, 36 },
		{ 144, 415, 36 },
		{ 144, 416, 36 },
		{ 144, 417, 36 },
		{ 144, 418, 36 },
		{ 144, 419, 36 },
		{ 144, 420, 36 },
		{ 144, 421, 36 },
		{ 144, 422, 36 },
		{ 144, 423, 36 },
		{ 144, 424, 36 },
		{ 144, 425, 36 },
		{ 144, 426, 36 },
		{ 144, 427, 36 },
		{ 144, 428, 36 },
		{ 144, 429, 36 },
		{ 144, 430, 36 },
		{ 144, 431, 36 },
		{ 144, 432, 36 },
		{ 144, 433, 36 },
		{ 144, 434, 36 },
		{ 144, 435, 36 },
		{ 144, 436, 36 },
		{ 144, 437, 36 },
		{ 145, 386, 36 },
		{ 145, 387, 36 },
		{ 145, 388, 36 },
		{ 145, 389, 36 },
		{ 145, 390, 36 },
		{ 145, 391, 36 },
		{ 145, 392, 36 },
		{ 145, 393, 36 },
		{ 145, 394, 36 },
		{ 145, 395, 36 },
		{ 145, 396, 36 },
		{ 145, 397, 36 },
		{ 145, 398, 36 },
		{ 145, 399, 36 },
		{ 145, 400, 36 },
		{ 145, 401, 36 },
		{ 145, 402, 36 },
		{ 145, 403, 36 },
		{ 145, 404, 36 },
		{ 145, 405, 36 },
		{ 145, 406, 36 },
		{ 145, 407, 36 },
		{ 145, 408, 36 },
		{ 145, 409, 36 },
		{ 145, 410, 36 },
		{ 145, 411, 36 },
		{ 145, 412, 36 },
		{ 145, 413, 36 },
		{ 145, 414, 36 },
		{ 145, 415, 36 },
		{ 145, 416, 36 },
		{ 145, 417, 36 },
		{ 145, 418, 36 },
		{ 145, 419, 36 },
		{ 145, 420, 36 },
		{ 145, 421, 36 },
		{ 145, 422, 36 },
		{ 145, 423, 36 },
		{ 145, 424, 36 },
		{ 145, 425, 36 },
		{ 145, 426, 36 },
		{ 145, 427, 36 },
		{ 145, 428, 36 },
		{ 145, 429, 36 },
		{ 145, 430, 36 },
		{ 145, 431, 36 },
		{ 145, 432, 36 },
		{ 145, 433, 36 },
		{ 145, 434, 36 },
		{ 145, 435, 36 },
		{ 145, 436, 36 },
		{ 146, 384, 36 },
		{ 146, 385, 36 },
		{ 146, 386, 36 },
		{ 146, 387, 36 },
		{ 146, 388, 36 },
		{ 146, 389, 36 },
		{ 146, 390, 36 },
		{ 146, 391, 36 },
		{ 146, 392, 36 },
		{ 146, 393, 36 },
		{ 146, 394, 36 },
		{ 146, 395, 36 },
		{ 146, 396, 36 },
		{ 146, 397, 36 },
		{ 146, 398, 36 },
		{ 146, 399, 36 },
		{ 146, 400, 36 },
		{ 146, 401, 36 },
		{ 146, 402, 36 },
		{ 146, 403, 36 },
		{ 146, 404, 36 },
		{ 146, 405, 36 },
		{ 146, 406, 36 },
		{ 146, 407, 36 },
		{ 146, 408, 36 },
		{ 146, 409, 36 },
		{ 146, 410, 36 },
		{ 146, 411, 36 },
		{ 146, 412, 36 },
		{ 146, 413, 36 },
		{ 146, 414, 36 },
		{ 146, 415, 36 },
		{ 146, 416, 36 },
		{ 146, 417, 36 },
		{ 146, 418, 36 },
		{ 146, 419, 36 },
		{ 146, 420, 36 },
		{ 146, 421, 36 },
		{ 146, 422, 36 },
		{ 146, 423, 36 },
		{ 146, 424, 36 },
		{ 146, 425, 36 },
		{ 146, 426, 36 },
		{ 146, 427, 36 },
		{ 146, 428, 36 },
		{ 146, 429, 36 },
		{ 146, 430, 36 },
		{ 146, 431, 36 },
		{ 146, 432, 36 },
		{ 146, 433, 36 },
		{ 146, 434, 36 },
		{ 147, 383, 36 },
		{ 147, 384, 36 },
		{ 147, 385, 36 },
		{ 147, 386, 36 },
		{ 147, 387, 36 },
		{ 147, 388, 36 },
		{ 147, 389, 36 },
		{ 147, 390, 36 },
		{ 147, 391, 36 },
		{ 147, 392, 36 },
		{ 147, 393, 36 },
		{ 147, 394, 36 },
		{ 147, 395, 36 },
		{ 147, 396, 36 },
		{ 147, 397, 36 },
		{ 147, 398, 36 },
		{ 147, 399, 36 },
		{ 147, 400, 36 },
		{ 147, 401, 36 },
		{ 147, 402, 36 },
		{ 147, 403, 36 },
		{ 147, 404, 36 },
		{ 147, 405, 36 },
		{ 147, 406, 36 },
		{ 147, 407, 36 },
		{ 147, 408, 36 },
		{ 147, 409, 36 },
		{ 147, 410, 36 },
		{ 147, 411, 36 },
		{ 147, 412, 36 },
		{ 147, 413, 36 },
		{ 147, 414, 36 },
		{ 147, 415, 36 },
		{ 147, 416, 36 },
		{ 147, 417, 36 },
		{ 147, 418, 36 },
		{ 147, 419, 36 },
		{ 147, 420, 36 },
		{ 147, 421, 36 },
		{ 147, 422, 36 },
		{ 147, 423, 36 },
		{ 147, 424, 36 },
		{ 147, 425, 36 },
		{ 147, 426, 36 },
		{ 147, 427, 36 },
		{ 147, 428, 36 },
		{ 147, 429, 36 },
		{ 147, 430, 36 },
		{ 147, 431, 36 },
		{ 147, 432, 36 },
		{ 148, 382, 36 },
		{ 148, 383, 36 },
		{ 148, 384, 36 },
		{ 148, 385, 36 },
		{ 148, 386, 36 },
		{ 148, 387, 36 },
		{ 148, 388, 36 },
		{ 148, 389, 36 },
		{ 148, 390, 36 },
		{ 148, 391, 36 },
		{ 148, 392, 36 },
		{ 148, 393, 36 },
		{ 148, 394, 36 },
		{ 148, 395, 36 },
		{ 148, 396, 36 },
		{ 148, 397, 36 },
		{ 148, 398, 36 },
		{ 148, 399, 36 },
		{ 148, 400, 36 },
		{ 148, 401, 36 },
		{ 148, 402, 36 },
		{ 148, 403, 36 },
		{ 148, 404, 36 },
		{ 148, 405, 36 },
		{ 148, 406, 36 },
		{ 148, 407, 36 },
		{ 148, 408, 36 },
		{ 148, 409, 36 },
		{ 148, 410, 36 },
		{ 148, 411, 36 },
		{ 148, 412, 36 },
		{ 148, 413, 36 },
		{ 148, 414, 36 },
		{ 148, 415, 36 },
		{ 148, 416, 36 },
		{ 148, 417, 36 },
		{ 148, 418, 36 },
		{ 148, 419, 36 },
		{ 148, 420, 36 },
		{ 148, 421, 36 },
		{ 148, 422, 36 },
		{ 148, 423, 36 },
		{ 148, 424, 36 },
		{ 148, 425, 36 },
		{ 148, 426, 36 },
		{ 148, 427, 36 },
		{ 148, 428, 36 },
		{ 148, 429, 36 },
		{ 148, 430, 36 },
		{ 148, 431, 36 },
		{ 149, 381, 36 },
		{ 149, 382, 36 },
		{ 149, 383, 36 },
		{ 149, 384, 36 },
		{ 149, 385, 36 },
		{ 149, 386, 36 },
		{ 149, 387, 36 },
		{ 149, 388, 36 },
		{ 149, 389, 36 },
		{ 149, 390, 36 },
		{ 149, 391, 36 },
		{ 149, 392, 36 },
		{ 149, 393, 36 },
		{ 149, 394, 36 },
		{ 149, 395, 36 },
		{ 149, 396, 36 },
		{ 149, 397, 36 },
		{ 149, 398, 36 },
		{ 149, 399, 36 },
		{ 149, 400, 36 },
		{ 149, 401, 36 },
		{ 149, 402, 36 },
		{ 149, 403, 36 },
		{ 149, 404, 36 },
		{ 149, 405, 36 },
		{ 149, 406, 36 },
		{ 149, 407, 36 },
		{ 149, 408, 36 },
		{ 149, 409, 36 },
		{ 149, 410, 36 },
		{ 149, 411, 36 },
		{ 149, 412, 36 },
		{ 149, 413, 36 },
		{ 149, 414, 36 },
		{ 149, 415, 36 },
		{ 149, 416, 36 },
		{ 149, 417, 36 },
		{ 149, 418, 36 },
		{ 149, 419, 36 },
		{ 149, 420, 36 },
		{ 149, 421, 36 },
		{ 149, 422, 36 },
		{ 149, 423, 36 },
		{ 149, 424, 36 },
		{ 149, 425, 36 },
		{ 149, 426, 36 },
		{ 149, 427, 36 },
		{ 149, 428, 36 },
		{ 149, 429, 36 },
		{ 150, 380, 36 },
		{ 150, 381, 36 },
		{ 150, 382, 36 },
		{ 150, 383, 36 },
		{ 150, 384, 36 },
		{ 150, 385, 36 },
		{ 150, 386, 36 },
		{ 150, 387, 36 },
		{ 150, 388, 36 },
		{ 150, 389, 36 },
		{ 150, 390, 36 },
		{ 150, 391, 36 },
		{ 150, 392, 36 },
		{ 150, 393, 36 },
		{ 150, 394, 36 },
		{ 150, 395, 36 },
		{ 150, 396, 36 },
		{ 150, 397, 36 },
		{ 150, 398, 36 },
		{ 150, 399, 36 },
		{ 150, 400, 36 },
		{ 150, 401, 36 },
		{ 150, 402, 36 },
		{ 150, 403, 36 },
		{ 150, 404, 36 },
		{ 150, 405, 36 },
		{ 150, 406, 36 },
		{ 150, 407, 36 },
		{ 150, 408, 36 },
		{ 150, 409, 36 },
		{ 150, 410, 36 },
		{ 150, 411, 36 },
		{ 150, 412, 36 },
		{ 150, 413, 36 },
		{ 150, 414, 36 },
		{ 150, 415, 36 },
		{ 150, 416, 36 },
		{ 150, 417, 36 },
		{ 150, 418, 36 },
		{ 150, 419, 36 },
		{ 150, 420, 36 },
		{ 150, 421, 36 },
		{ 150, 422, 36 },
		{ 150, 423, 36 },
		{ 150, 424, 36 },
		{ 150, 425, 36 },
		{ 150, 426, 36 },
		{ 150, 427, 36 },
		{ 150, 428, 36 },
		{ 151, 379, 36 },
		{ 151, 380, 36 },
		{ 151, 381, 36 },
		{ 151, 382, 36 },
		{ 151, 383, 36 },
		{ 151, 384, 36 },
		{ 151, 385, 36 },
		{ 151, 386, 36 },
		{ 151, 387, 36 },
		{ 151, 388, 36 },
		{ 151, 389, 36 },
		{ 151, 390, 36 },
		{ 151, 391, 36 },
		{ 151, 392, 36 },
		{ 151, 393, 36 },
		{ 151, 394, 36 },
		{ 151, 395, 36 },
		{ 151, 396, 36 },
		{ 151, 397, 36 },
		{ 151, 398, 36 },
		{ 151, 399, 36 },
		{ 151, 400, 36 },
		{ 151, 401, 36 },
		{ 151, 402, 36 },
		{ 151, 403, 36 },
		{ 151, 404, 36 },
		{ 151, 405, 36 },
		{ 151, 406, 36 },
		{ 151, 407, 36 },
		{ 151, 408, 36 },
		{ 151, 409, 36 },
		{ 151, 410, 36 },
		{ 151, 411, 36 },
		{ 151, 412, 36 },
		{ 151, 413, 36 },
		{ 151, 414, 36 },
		{ 151, 415, 36 },
		{ 151, 416, 36 },
		{ 151, 417, 36 },
		{ 151, 418, 36 },
		{ 151, 419, 36 },
		{ 151, 420, 36 },
		{ 151, 421, 36 },
		{ 151, 422, 36 },
		{ 151, 423, 36 },
		{ 151, 424, 36 },
		{ 151, 425, 36 },
		{ 151, 426, 36 },
		{ 152, 378, 36 },
		{ 152, 379, 36 },
		{ 152, 380, 36 },
		{ 152, 381, 36 },
		{ 152, 382, 36 },
		{ 152, 383, 36 },
		{ 152, 384, 36 },
		{ 152, 385, 36 },
		{ 152, 386, 36 },
		{ 152, 387, 36 },
		{ 152, 388, 36 },
		{ 152, 389, 36 },
		{ 152, 390, 36 },
		{ 152, 391, 36 },
		{ 152, 392, 36 },
		{ 152, 393, 36 },
		{ 152, 394, 36 },
		{ 152, 395, 36 },
		{ 152, 396, 36 },
		{ 152, 397, 36 },
		{ 152, 398, 36 },
		{ 152, 399, 36 },
		{ 152, 400, 36 },
		{ 152, 401, 36 },
		{ 152, 402, 36 },
		{ 152, 403, 36 },
		{ 152, 404, 36 },
		{ 152, 405, 36 },
		{ 152, 406, 36 },
		{ 152, 407, 36 },
		{ 152, 408, 36 },
		{ 152, 409, 36 },
		{ 152, 410, 36 },
		{ 152, 411, 36 },
		{ 152, 412, 36 },
		{ 152, 413, 36 },
		{ 152, 414, 36 },
		{ 152, 415, 36 },
		{ 152, 416, 36 },
		{ 152, 417, 36 },
		{ 152, 418, 36 },
		{ 152, 419, 36 },
		{ 152, 420, 36 },
		{ 152, 421, 36 },
		{ 152, 422, 36 },
		{ 152, 423, 36 },
		{ 152, 424, 36 },
		{ 153, 376, 36 },
		{ 153, 377, 36 },
		{ 153, 378, 36 },
		{ 153, 379, 36 },
		{ 153, 380, 36 },
		{ 153, 381, 36 },
		{ 153, 382, 36 },
		{ 153, 383, 36 },
		{ 153, 384, 36 },
		{ 153, 385, 36 },
		{ 153, 386, 36 },
		{ 153, 387, 36 },
		{ 153, 388, 36 },
		{ 153, 389, 36 },
		{ 153, 390, 36 },
		{ 153, 391, 36 },
		{ 153, 392, 36 },
		{ 153, 393, 36 },
		{ 153, 394, 36 },
		{ 153, 395, 36 },
		{ 153, 396, 36 },
		{ 153, 397, 36 },
		{ 153, 398, 36 },
		{ 153, 399, 36 },
		{ 153, 400, 36 },
		{ 153, 401, 36 },
		{ 153, 402, 36 },
		{ 153, 403, 36 },
		{ 153, 404, 36 },
		{ 153, 405, 36 },
		{ 153, 406, 36 },
		{ 153, 407, 36 },
		{ 153, 408, 36 },
		{ 153, 409, 36 },
		{ 153, 410, 36 },
		{ 153, 411, 36 },
		{ 153, 412, 36 },
		{ 153, 413, 36 },
		{ 153, 414, 36 },
		{ 153, 415, 36 },
		{ 153, 416, 36 },
		{ 153, 417, 36 },
		{ 153, 418, 36 },
		{ 153, 419, 36 },
		{ 153, 420, 36 },
		{ 153, 421, 36 },
		{ 153, 422, 36 },
		{ 153, 423, 36 },
		{ 154, 375, 36 },
		{ 154, 376, 36 },
		{ 154, 377, 36 },
		{ 154, 378, 36 },
		{ 154, 379, 36 },
		{ 154, 380, 36 },
		{ 154, 381, 36 },
		{ 154, 382, 36 },
		{ 154, 383, 36 },
		{ 154, 384, 36 },
		{ 154, 385, 36 },
		{ 154, 386, 36 },
		{ 154, 387, 36 },
		{ 154, 388, 36 },
		{ 154, 389, 36 },
		{ 154, 390, 36 },
		{ 154, 391, 36 },
		{ 154, 392, 36 },
		{ 154, 393, 36 },
		{ 154, 394, 36 },
		{ 154, 395, 36 },
		{ 154, 396, 36 },
		{ 154, 397, 36 },
		{ 154, 398, 36 },
		{ 154, 399, 36 },
		{ 154, 400, 36 },
		{ 154, 401, 36 },
		{ 154, 402, 36 },
		{ 154, 403, 36 },
		{ 154, 404, 36 },
		{ 154, 405, 36 },
		{ 154, 406, 36 },
		{ 154, 407, 36 },
		{ 154, 408, 36 },
		{ 154, 409, 36 },
		{ 154, 410, 36 },
		{ 154, 411, 36 },
		{ 154, 412, 36 },
		{ 154, 413, 36 },
		{ 154, 414, 36 },
		{ 154, 415, 36 },
		{ 154, 416, 36 },
		{ 154, 417, 36 },
		{ 154, 418, 36 },
		{ 154, 419, 36 },
		{ 154, 420, 36 },
		{ 154, 421, 36 },
		{ 155, 374, 36 },
		{ 155, 375, 36 },
		{ 155, 376, 36 },
		{ 155, 377, 36 },
		{ 155, 378, 36 },
		{ 155, 379, 36 },
		{ 155, 380, 36 },
		{ 155, 381, 36 },
		{ 155, 382, 36 },
		{ 155, 383, 36 },
		{ 155, 384, 36 },
		{ 155, 385, 36 },
		{ 155, 386, 36 },
		{ 155, 387, 36 },
		{ 155, 388, 36 },
		{ 155, 389, 36 },
		{ 155, 390, 36 },
		{ 155, 391, 36 },
		{ 155, 392, 36 },
		{ 155, 393, 36 },
		{ 155, 394, 36 },
		{ 155, 395, 36 },
		{ 155, 396, 36 },
		{ 155, 397, 36 },
		{ 155, 398, 36 },
		{ 155, 399, 36 },
		{ 155, 400, 36 },
		{ 155, 401, 36 },
		{ 155, 402, 36 },
		{ 155, 403, 36 },
		{ 155, 404, 36 },
		{ 155, 405, 36 },
		{ 155, 406, 36 },
		{ 155, 407, 36 },
		{ 155, 408, 36 },
		{ 155, 409, 36 },
		{ 155, 410, 36 },
		{ 155, 411, 36 },
		{ 155, 412, 36 },
		{ 155, 413, 36 },
		{ 155, 414, 36 },
		{ 155, 415, 36 },
		{ 155, 416, 36 },
		{ 155, 417, 36 },
		{ 155, 418, 36 },
		{ 155, 419, 36 },
		{ 155, 420, 36 },
		{ 156, 373, 36 },
		{ 156, 374, 36 },
		{ 156, 375, 36 },
		{ 156, 376, 36 },
		{ 156, 377, 36 },
		{ 156, 378, 36 },
		{ 156, 379, 36 },
		{ 156, 380, 36 },
		{ 156, 381, 36 },
		{ 156, 382, 36 },
		{ 156, 383, 36 },
		{ 156, 384, 36 },
		{ 156, 385, 36 },
		{ 156, 386, 36 },
		{ 156, 387, 36 },
		{ 156, 388, 36 },
		{ 156, 389, 36 },
		{ 156, 390, 36 },
		{ 156, 391, 36 },
		{ 156, 392, 36 },
		{ 156, 393, 36 },
		{ 156, 394, 36 },
		{ 156, 395, 36 },
		{ 156, 396, 36 },
		{ 156, 397, 36 },
		{ 156, 398, 36 },
		{ 156, 399, 36 },
		{ 156, 400, 36 },
		{ 156, 401, 36 },
		{ 156, 402, 36 },
		{ 156, 403, 36 },
		{ 156, 404, 36 },
		{ 156, 405, 36 },
		{ 156, 406, 36 },
		{ 156, 407, 36 },
		{ 156, 408, 36 },
		{ 156, 409, 36 },
		{ 156, 410, 36 },
		{ 156, 411, 36 },
		{ 156, 412, 36 },
		{ 156, 413, 36 },
		{ 156, 414, 36 },
		{ 156, 415, 36 },
		{ 156, 416, 36 },
		{ 156, 417, 36 },
		{ 156, 418, 36 },
		{ 157, 372, 36 },
		{ 157, 373, 36 },
		{ 157, 374, 36 },
		{ 157, 375, 36 },
		{ 157, 376, 36 },
		{ 157, 377, 36 },
		{ 157, 378, 36 },
		{ 157, 379, 36 },
		{ 157, 380, 36 },
		{ 157, 381, 36 },
		{ 157, 382, 36 },
		{ 157, 383, 36 },
		{ 157, 384, 36 },
		{ 157, 385, 36 },
		{ 157, 386, 36 },
		{ 157, 387, 36 },
		{ 157, 388, 36 },
		{ 157, 389, 36 },
		{ 157, 390, 36 },
		{ 157, 391, 36 },
		{ 157, 392, 36 },
		{ 157, 393, 36 },
		{ 157, 394, 36 },
		{ 157, 395, 36 },
		{ 157, 396, 36 },
		{ 157, 397, 36 },
		{ 157, 398, 36 },
		{ 157, 399, 36 },
		{ 157, 400, 36 },
		{ 157, 401, 36 },
		{ 157, 402, 36 },
		{ 157, 403, 36 },
		{ 157, 404, 36 },
		{ 157, 405, 36 },
		{ 157, 406, 36 },
		{ 157, 407, 36 },
		{ 157, 408, 36 },
		{ 157, 409, 36 },
		{ 157, 410, 36 },
		{ 157, 411, 36 },
		{ 157, 412, 36 },
		{ 157, 413, 36 },
		{ 157, 414, 36 },
		{ 157, 415, 36 },
		{ 157, 416, 36 },
		{ 158, 371, 36 },
		{ 158, 372, 36 },
		{ 158, 373, 36 },
		{ 158, 374, 36 },
		{ 158, 375, 36 },
		{ 158, 376, 36 },
		{ 158, 377, 36 },
		{ 158, 378, 36 },
		{ 158, 379, 36 },
		{ 158, 380, 36 },
		{ 158, 381, 36 },
		{ 158, 382, 36 },
		{ 158, 383, 36 },
		{ 158, 384, 36 },
		{ 158, 385, 36 },
		{ 158, 386, 36 },
		{ 158, 387, 36 },
		{ 158, 388, 36 },
		{ 158, 389, 36 },
		{ 158, 390, 36 },
		{ 158, 391, 36 },
		{ 158, 392, 36 },
		{ 158, 393, 36 },
		{ 158, 394, 36 },
		{ 158, 395, 36 },
		{ 158, 396, 36 },
		{ 158, 397, 36 },
		{ 158, 398, 36 },
		{ 158, 399, 36 },
		{ 158, 400, 36 },
		{ 158, 401, 36 },
		{ 158, 402, 36 },
		{ 158, 403, 36 },
		{ 158, 404, 36 },
		{ 158, 405, 36 },
		{ 158, 406, 36 },
		{ 158, 407, 36 },
		{ 158, 408, 36 },
		{ 158, 409, 36 },
		{ 158, 410, 36 },
		{ 158, 411, 36 },
		{ 158, 412, 36 },
		{ 158, 413, 36 },
		{ 158, 414, 36 },
		{ 158, 415, 36 },
		{ 159, 370, 36 },
		{ 159, 371, 36 },
		{ 159, 372, 36 },
		{ 159, 373, 36 },
		{ 159, 374, 36 },
		{ 159, 375, 36 },
		{ 159, 376, 36 },
		{ 159, 377, 36 },
		{ 159, 378, 36 },
		{ 159, 379, 36 },
		{ 159, 380, 36 },
		{ 159, 381, 36 },
		{ 159, 382, 36 },
		{ 159, 383, 36 },
		{ 159, 384, 36 },
		{ 159, 385, 36 },
		{ 159, 386, 36 },
		{ 159, 387, 36 },
		{ 159, 388, 36 },
		{ 159, 389, 36 },
		{ 159, 390, 36 },
		{ 159, 391, 36 },
		{ 159, 392, 36 },
		{ 159, 393, 36 },
		{ 159, 394, 36 },
		{ 159, 395, 36 },
		{ 159, 396, 36 },
		{ 159, 397, 36 },
		{ 159, 398, 36 },
		{ 159, 399, 36 },
		{ 159, 400, 36 },
		{ 159, 401, 36 },
		{ 159, 402, 36 },
		{ 159, 403, 36 },
		{ 159, 404, 36 },
		{ 159, 405, 36 },
		{ 159, 406, 36 },
		{ 159, 407, 36 },
		{ 159, 408, 36 },
		{ 159, 409, 36 },
		{ 159, 410, 36 },
		{ 159, 411, 36 },
		{ 159, 412, 36 },
		{ 159, 413, 36 },
		{ 160, 368, 36 },
		{ 160, 369, 36 },
		{ 160, 370, 36 },
		{ 160, 371, 36 },
		{ 160, 372, 36 },
		{ 160, 373, 36 },
		{ 160, 374, 36 },
		{ 160, 375, 36 },
		{ 160, 376, 36 },
		{ 160, 377, 36 },
		{ 160, 378, 36 },
		{ 160, 379, 36 },
		{ 160, 380, 36 },
		{ 160, 381, 36 },
		{ 160, 382, 36 },
		{ 160, 383, 36 },
		{ 160, 384, 36 },
		{ 160, 385, 36 },
		{ 160, 386, 36 },
		{ 160, 387, 36 },
		{ 160, 388, 36 },
		{ 160, 389, 36 },
		{ 160, 390, 36 },
		{ 160, 391, 36 },
		{ 160, 392, 36 },
		{ 160, 393, 36 },
		{ 160, 394, 36 },
		{ 160, 395, 36 },
		{ 160, 396, 36 },
		{ 160, 397, 36 },
		{ 160, 398, 36 },
		{ 160, 399, 36 },
		{ 160, 400, 36 },
		{ 160, 401, 36 },
		{ 160, 402, 36 },
		{ 160, 403, 36 },
		{ 160, 404, 36 },
		{ 160, 405, 36 },
		{ 160, 406, 36 },
		{ 160, 407, 36 },
		{ 160, 408, 36 },
		{ 160, 409, 36 },
		{ 160, 410, 36 },
		{ 160, 411, 36 },
		{ 160, 412, 36 },
		{ 161, 367, 36 },
		{ 161, 368, 36 },
		{ 161, 369, 36 },
		{ 161, 370, 36 },
		{ 161, 371, 36 },
		{ 161, 372, 36 },
		{ 161, 373, 36 },
		{ 161, 374, 36 },
		{ 161, 375, 36 },
		{ 161, 376, 36 },
		{ 161, 377, 36 },
		{ 161, 378, 36 },
		{ 161, 379, 36 },
		{ 161, 380, 36 },
		{ 161, 381, 36 },
		{ 161, 382, 36 },
		{ 161, 383, 36 },
		{ 161, 384, 36 },
		{ 161, 385, 36 },
		{ 161, 386, 36 },
		{ 161, 387, 36 },
		{ 161, 388, 36 },
		{ 161, 389, 36 },
		{ 161, 390, 36 },
		{ 161, 391, 36 },
		{ 161, 392, 36 },
		{ 161, 393, 36 },
		{ 161, 394, 36 },
		{ 161, 395, 36 },
		{ 161, 396, 36 },
		{ 161, 397, 36 },
		{ 161, 398, 36 },
		{ 161, 399, 36 },
		{ 161, 400, 36 },
		{ 161, 401, 36 },
		{ 161, 402, 36 },
		{ 161, 403, 36 },
		{ 161, 404, 36 },
		{ 161, 405, 36 },
		{ 161, 406, 36 },
		{ 161, 407, 36 },
		{ 161, 408, 36 },
		{ 161, 409, 36 },
		{ 161, 410, 36 },
		{ 162, 366, 36 },
		{ 162, 367, 36 },
		{ 162, 368, 36 },
		{ 162, 369, 36 },
		{ 162, 370, 36 },
		{ 162, 371, 36 },
		{ 162, 372, 36 },
		{ 162, 373, 36 },
		{ 162, 374, 36 },
		{ 162, 375, 36 },
		{ 162, 376, 36 },
		{ 162, 377, 36 },
		{ 162, 378, 36 },
		{ 162, 379, 36 },
		{ 162, 380, 36 },
		{ 162, 381, 36 },
		{ 162, 382, 36 },
		{ 162, 383, 36 },
		{ 162, 384, 36 },
		{ 162, 385, 36 },
		{ 162, 386, 36 },
		{ 162, 387, 36 },
		{ 162, 388, 36 },
		{ 162, 389, 36 },
		{ 162, 390, 36 },
		{ 162, 391, 36 },
		{ 162, 392, 36 },
		{ 162, 393, 36 },
		{ 162, 394, 36 },
		{ 162, 395, 36 },
		{ 162, 396, 36 },
		{ 162, 397, 36 },
		{ 162, 398, 36 },
		{ 162, 399, 36 },
		{ 162, 400, 36 },
		{ 162, 401, 36 },
		{ 162, 402, 36 },
		{ 162, 403, 36 },
		{ 162, 404, 36 },
		{ 162, 405, 36 },
		{ 162, 406, 36 },
		{ 162, 407, 36 },
		{ 162, 408, 36 },
		{ 163, 365, 36 },
		{ 163, 366, 36 },
		{ 163, 367, 36 },
		{ 163, 368, 36 },
		{ 163, 369, 36 },
		{ 163, 370, 36 },
		{ 163, 371, 36 },
		{ 163, 372, 36 },
		{ 163, 373, 36 },
		{ 163, 374, 36 },
		{ 163, 375, 36 },
		{ 163, 376, 36 },
		{ 163, 377, 36 },
		{ 163, 378, 36 },
		{ 163, 379, 36 },
		{ 163, 380, 36 },
		{ 163, 381, 36 },
		{ 163, 382, 36 },
		{ 163, 383, 36 },
		{ 163, 384, 36 },
		{ 163, 385, 36 },
		{ 163, 386, 36 },
		{ 163, 387, 36 },
		{ 163, 388, 36 },
		{ 163, 389, 36 },
		{ 163, 390, 36 },
		{ 163, 391, 36 },
		{ 163, 392, 36 },
		{ 163, 393, 36 },
		{ 163, 394, 36 },
		{ 163, 395, 36 },
		{ 163, 396, 36 },
		{ 163, 397, 36 },
		{ 163, 398, 36 },
		{ 163, 399, 36 },
		{ 163, 400, 36 },
		{ 163, 401, 36 },
		{ 163, 402, 36 },
		{ 163, 403, 36 },
		{ 163, 404, 36 },
		{ 163, 405, 36 },
		{ 163, 406, 36 },
		{ 163, 407, 36 },
		{ 164, 364, 36 },
		{ 164, 365, 36 },
		{ 164, 366, 36 },
		{ 164, 367, 36 },
		{ 164, 368, 36 },
		{ 164, 369, 36 },
		{ 164, 370, 36 },
		{ 164, 371, 36 },
		{ 164, 372, 36 },
		{ 164, 373, 36 },
		{ 164, 374, 36 },
		{ 164, 375, 36 },
		{ 164, 376, 36 },
		{ 164, 377, 36 },
		{ 164, 378, 36 },
		{ 164, 379, 36 },
		{ 164, 380, 36 },
		{ 164, 381, 36 },
		{ 164, 382, 36 },
		{ 164, 383, 36 },
		{ 164, 384, 36 },
		{ 164, 385, 36 },
		{ 164, 386, 36 },
		{ 164, 387, 36 },
		{ 164, 388, 36 },
		{ 164, 389, 36 },
		{ 164, 390, 36 },
		{ 164, 391, 36 },
		{ 164, 392, 36 },
		{ 164, 393, 36 },
		{ 164, 394, 36 },
		{ 164, 395, 36 },
		{ 164, 396, 36 },
		{ 164, 397, 36 },
		{ 164, 398, 36 },
		{ 164, 399, 36 },
		{ 164, 400, 36 },
		{ 164, 401, 36 },
		{ 164, 402, 36 },
		{ 164, 403, 36 },
		{ 164, 404, 36 },
		{ 164, 405, 36 },
		{ 165, 363, 36 },
		{ 165, 364, 36 },
		{ 165, 365, 36 },
		{ 165, 366, 36 },
		{ 165, 367, 36 },
		{ 165, 368, 36 },
		{ 165, 369, 36 },
		{ 165, 370, 36 },
		{ 165, 371, 36 },
		{ 165, 372, 36 },
		{ 165, 373, 36 },
		{ 165, 374, 36 },
		{ 165, 375, 36 },
		{ 165, 376, 36 },
		{ 165, 377, 36 },
		{ 165, 378, 36 },
		{ 165, 379, 36 },
		{ 165, 380, 36 },
		{ 165, 381, 36 },
		{ 165, 382, 36 },
		{ 165, 383, 36 },
		{ 165, 384, 36 },
		{ 165, 385, 36 },
		{ 165, 386, 36 },
		{ 165, 387, 36 },
		{ 165, 388, 36 },
		{ 165, 389, 36 },
		{ 165, 390, 36 },
		{ 165, 391, 36 },
		{ 165, 392, 36 },
		{ 165, 393, 36 },
		{ 165, 394, 36 },
		{ 165, 395, 36 },
		{ 165, 396, 36 },
		{ 165, 397, 36 },
		{ 165, 398, 36 },
		{ 165, 399, 36 },
		{ 165, 400, 36 },
		{ 165, 401, 36 },
		{ 165, 402, 36 },
		{ 165, 403, 36 },
		{ 165, 404, 36 },
		{ 166, 362, 36 },
		{ 166, 363, 36 },
		{ 166, 364, 36 },
		{ 166, 365, 36 },
		{ 166, 366, 36 },
		{ 166, 367, 36 },
		{ 166, 368, 36 },
		{ 166, 369, 36 },
		{ 166, 370, 36 },
		{ 166, 371, 36 },
		{ 166, 372, 36 },
		{ 166, 373, 36 },
		{ 166, 374, 36 },
		{ 166, 375, 36 },
		{ 166, 376, 36 },
		{ 166, 377, 36 },
		{ 166, 378, 36 },
		{ 166, 379, 36 },
		{ 166, 380, 36 },
		{ 166, 381, 36 },
		{ 166, 382, 36 },
		{ 166, 383, 36 },
		{ 166, 384, 36 },
		{ 166, 385, 36 },
		{ 166, 386, 36 },
		{ 166, 387, 36 },
		{ 166, 388, 36 },
		{ 166, 389, 36 },
		{ 166, 390, 36 },
		{ 166, 391, 36 },
		{ 166, 392, 36 },
		{ 166, 393, 36 },
		{ 166, 394, 36 },
		{ 166, 395, 36 },
		{ 166, 396, 36 },
		{ 166, 397, 36 },
		{ 166, 398, 36 },
		{ 166, 399, 36 },
		{ 166, 400, 36 },
		{ 166, 401, 36 },
		{ 166, 402, 36 },
		{ 167, 360, 36 },
		{ 167, 361, 36 },
		{ 167, 362, 36 },
		{ 167, 363, 36 },
		{ 167, 364, 36 },
		{ 167, 365, 36 },
		{ 167, 366, 36 },
		{ 167, 367, 36 },
		{ 167, 368, 36 },
		{ 167, 369, 36 },
		{ 167, 370, 36 },
		{ 167, 371, 36 },
		{ 167, 372, 36 },
		{ 167, 373, 36 },
		{ 167, 374, 36 },
		{ 167, 375, 36 },
		{ 167, 376, 36 },
		{ 167, 377, 36 },
		{ 167, 378, 36 },
		{ 167, 379, 36 },
		{ 167, 380, 36 },
		{ 167, 381, 36 },
		{ 167, 382, 36 },
		{ 167, 383, 36 },
		{ 167, 384, 36 },
		{ 167, 385, 36 },
		{ 167, 386, 36 },
		{ 167, 387, 36 },
		{ 167, 388, 36 },
		{ 167, 389, 36 },
		{ 167, 390, 36 },
		{ 167, 391, 36 },
		{ 167, 392, 36 },
		{ 167, 393, 36 },
		{ 167, 394, 36 },
		{ 167, 395, 36 },
		{ 167, 396, 36 },
		{ 167, 397, 36 },
		{ 167, 398, 36 },
		{ 167, 399, 36 },
		{ 167, 400, 36 },
		{ 168, 359, 36 },
		{ 168, 360, 36 },
		{ 168, 361, 36 },
		{ 168, 362, 36 },
		{ 168, 363, 36 },
		{ 168, 364, 36 },
		{ 168, 365, 36 },
		{ 168, 366, 36 },
		{ 168, 367, 36 },
		{ 168, 368, 36 },
		{ 168, 369, 36 },
		{ 168, 370, 36 },
		{ 168, 371, 36 },
		{ 168, 372, 36 },
		{ 168, 373, 36 },
		{ 168, 374, 36 },
		{ 168, 375, 36 },
		{ 168, 376, 36 },
		{ 168, 377, 36 },
		{ 168, 378, 36 },
		{ 168, 379, 36 },
		{ 168, 380, 36 },
		{ 168, 381, 36 },
		{ 168, 382, 36 },
		{ 168, 383, 36 },
		{ 168, 384, 36 },
		{ 168, 385, 36 },
		{ 168, 386, 36 },
		{ 168, 387, 36 },
		{ 168, 388, 36 },
		{ 168, 389, 36 },
		{ 168, 390, 36 },
		{ 168, 391, 36 },
		{ 168, 392, 36 },
		{ 168, 393, 36 },
		{ 168, 394, 36 },
		{ 168, 395, 36 },
		{ 168, 396, 36 },
		{ 168, 397, 36 },
		{ 168, 398, 36 },
		{ 168, 399, 36 },
		{ 169, 358, 36 },
		{ 169, 359, 36 },
		{ 169, 360, 36 },
		{ 169, 361, 36 },
		{ 169, 362, 36 },
		{ 169, 363, 36 },
		{ 169, 364, 36 },
		{ 169, 365, 36 },
		{ 169, 366, 36 },
		{ 169, 367, 36 },
		{ 169, 368, 36 },
		{ 169, 369, 36 },
		{ 169, 370, 36 },
		{ 169, 371, 36 },
		{ 169, 372, 36 },
		{ 169, 373, 36 },
		{ 169, 374, 36 },
		{ 169, 375, 36 },
		{ 169, 376, 36 },
		{ 169, 377, 36 },
		{ 169, 378, 36 },
		{ 169, 379, 36 },
		{ 169, 380, 36 },
		{ 169, 381, 36 },
		{ 169, 382, 36 },
		{ 169, 383, 36 },
		{ 169, 384, 36 },
		{ 169, 385, 36 },
		{ 169, 386, 36 },
		{ 169, 387, 36 },
		{ 169, 388, 36 },
		{ 169, 389, 36 },
		{ 169, 390, 36 },
		{ 169, 391, 36 },
		{ 169, 392, 36 },
		{ 169, 393, 36 },
		{ 169, 394, 36 },
		{ 169, 395, 36 },
		{ 169, 396, 36 },
		{ 169, 397, 36 },
		{ 170, 357, 36 },
		{ 170, 358, 36 },
		{ 170, 359, 36 },
		{ 170, 360, 36 },
		{ 170, 361, 36 },
		{ 170, 362, 36 },
		{ 170, 363, 36 },
		{ 170, 364, 36 },
		{ 170, 365, 36 },
		{ 170, 366, 36 },
		{ 170, 367, 36 },
		{ 170, 368, 36 },
		{ 170, 369, 36 },
		{ 170, 370, 36 },
		{ 170, 371, 36 },
		{ 170, 372, 36 },
		{ 170, 373, 36 },
		{ 170, 374, 36 },
		{ 170, 375, 36 },
		{ 170, 376, 36 },
		{ 170, 377, 36 },
		{ 170, 378, 36 },
		{ 170, 379, 36 },
		{ 170, 380, 36 },
		{ 170, 381, 36 },
		{ 170, 382, 36 },
		{ 170, 383, 36 },
		{ 170, 384, 36 },
		{ 170, 385, 36 },
		{ 170, 386, 36 },
		{ 170, 387, 36 },
		{ 170, 388, 36 },
		{ 170, 389, 36 },
		{ 170, 390, 36 },
		{ 170, 391, 36 },
		{ 170, 392, 36 },
		{ 170, 393, 36 },
		{ 170, 394, 36 },
		{ 170, 395, 36 },
		{ 170, 396, 36 },
		{ 171, 356, 36 },
		{ 171, 357, 36 },
		{ 171, 358, 36 },
		{ 171, 359, 36 },
		{ 171, 360, 36 },
		{ 171, 361, 36 },
		{ 171, 362, 36 },
		{ 171, 363, 36 },
		{ 171, 364, 36 },
		{ 171, 365, 36 },
		{ 171, 366, 36 },
		{ 171, 367, 36 },
		{ 171, 368, 36 },
		{ 171, 369, 36 },
		{ 171, 370, 36 },
		{ 171, 371, 36 },
		{ 171, 372, 36 },
		{ 171, 373, 36 },
		{ 171, 374, 36 },
		{ 171, 375, 36 },
		{ 171, 376, 36 },
		{ 171, 377, 36 },
		{ 171, 378, 36 },
		{ 171, 379, 36 },
		{ 171, 380, 36 },
		{ 171, 381, 36 },
		{ 171, 382, 36 },
		{ 171, 383, 36 },
		{ 171, 384, 36 },
		{ 171, 385, 36 },
		{ 171, 386, 36 },
		{ 171, 387, 36 },
		{ 171, 388, 36 },
		{ 171, 389, 36 },
		{ 171, 390, 36 },
		{ 171, 391, 36 },
		{ 171, 392, 36 },
		{ 171, 393, 36 },
		{ 171, 394, 36 },
		{ 172, 355, 36 },
		{ 172, 356, 36 },
		{ 172, 357, 36 },
		{ 172, 358, 36 },
		{ 172, 359, 36 },
		{ 172, 360, 36 },
		{ 172, 361, 36 },
		{ 172, 362, 36 },
		{ 172, 363, 36 },
		{ 172, 364, 36 },
		{ 172, 365, 36 },
		{ 172, 366, 36 },
		{ 172, 367, 36 },
		{ 172, 368, 36 },
		{ 172, 369, 36 },
		{ 172, 370, 36 },
		{ 172, 371, 36 },
		{ 172, 372, 36 },
		{ 172, 373, 36 },
		{ 172, 374, 36 },
		{ 172, 375, 36 },
		{ 172, 376, 36 },
		{ 172, 377, 36 },
		{ 172, 378, 36 },
		{ 172, 379, 36 },
		{ 172, 380, 36 },
		{ 172, 381, 36 },
		{ 172, 382, 36 },
		{ 172, 383, 36 },
		{ 172, 384, 36 },
		{ 172, 385, 36 },
		{ 172, 386, 36 },
		{ 172, 387, 36 },
		{ 172, 388, 36 },
		{ 172, 389, 36 },
		{ 172, 390, 36 },
		{ 172, 391, 36 },
		{ 172, 392, 36 },
		{ 173, 354, 36 },
		{ 173, 355, 36 },
		{ 173, 356, 36 },
		{ 173, 357, 36 },
		{ 173, 358, 36 },
		{ 173, 359, 36 },
		{ 173, 360, 36 },
		{ 173, 361, 36 },
		{ 173, 362, 36 },
		{ 173, 363, 36 },
		{ 173, 364, 36 },
		{ 173, 365, 36 },
		{ 173, 366, 36 },
		{ 173, 367, 36 },
		{ 173, 368, 36 },
		{ 173, 369, 36 },
		{ 173, 370, 36 },
		{ 173, 371, 36 },
		{ 173, 372, 36 },
		{ 173, 373, 36 },
		{ 173, 374, 36 },
		{ 173, 375, 36 },
		{ 173, 376, 36 },
		{ 173, 377, 36 },
		{ 173, 378, 36 },
		{ 173, 379, 36 },
		{ 173, 380, 36 },
		{ 173, 381, 36 },
		{ 173, 382, 36 },
		{ 173, 383, 36 },
		{ 173, 384, 36 },
		{ 173, 385, 36 },
		{ 173, 386, 36 },
		{ 173, 387, 36 },
		{ 173, 388, 36 },
		{ 173, 389, 36 },
		{ 173, 390, 36 },
		{ 173, 391, 36 },
		{ 174, 352, 36 },
		{ 174, 353, 36 },
		{ 174, 354, 36 },
		{ 174, 355, 36 },
		{ 174, 356, 36 },
		{ 174, 357, 36 },
		{ 174, 358, 36 },
		{ 174, 359, 36 },
		{ 174, 360, 36 },
		{ 174, 361, 36 },
		{ 174, 362, 36 },
		{ 174, 363, 36 },
		{ 174, 364, 36 },
		{ 174, 365, 36 },
		{ 174, 366, 36 },
		{ 174, 367, 36 },
		{ 174, 368, 36 },
		{ 174, 369, 36 },
		{ 174, 370, 36 },
		{ 174, 371, 36 },
		{ 174, 372, 36 },
		{ 174, 373, 36 },
		{ 174, 374, 36 },
		{ 174, 375, 36 },
		{ 174, 376, 36 },
		{ 174, 377, 36 },
		{ 174, 378, 36 },
		{ 174, 379, 36 },
		{ 174, 380, 36 },
		{ 174, 381, 36 },
		{ 174, 382, 36 },
		{ 174, 383, 36 },
		{ 174, 384, 36 },
		{ 174, 385, 36 },
		{ 174, 386, 36 },
		{ 174, 387, 36 },
		{ 174, 388, 36 },
		{ 174, 389, 36 },
		{ 175, 351, 36 },
		{ 175, 352, 36 },
		{ 175, 353, 36 },
		{ 175, 354, 36 },
		{ 175, 355, 36 },
		{ 175, 356, 36 },
		{ 175, 357, 36 },
		{ 175, 358, 36 },
		{ 175, 359, 36 },
		{ 175, 360, 36 },
		{ 175, 361, 36 },
		{ 175, 362, 36 },
		{ 175, 363, 36 },
		{ 175, 364, 36 },
		{ 175, 365, 36 },
		{ 175, 366, 36 },
		{ 175, 367, 36 },
		{ 175, 368, 36 },
		{ 175, 369, 36 },
		{ 175, 370, 36 },
		{ 175, 371, 36 },
		{ 175, 372, 36 },
		{ 175, 373, 36 },
		{ 175, 374, 36 },
		{ 175, 375, 36 },
		{ 175, 376, 36 },
		{ 175, 377, 36 },
		{ 175, 378, 36 },
		{ 175, 379, 36 },
		{ 175, 380, 36 },
		{ 175, 381, 36 },
		{ 175, 382, 36 },
		{ 175, 383, 36 },
		{ 175, 384, 36 },
		{ 175, 385, 36 },
		{ 175, 386, 36 },
		{ 175, 387, 36 },
		{ 175, 388, 36 },
		{ 176, 350, 36 },
		{ 176, 351, 36 },
		{ 176, 352, 36 },
		{ 176, 353, 36 },
		{ 176, 354, 36 },
		{ 176, 355, 36 },
		{ 176, 356, 36 },
		{ 176, 357, 36 },
		{ 176, 358, 36 },
		{ 176, 359, 36 },
		{ 176, 360, 36 },
		{ 176, 361, 36 },
		{ 176, 362, 36 },
		{ 176, 363, 36 },
		{ 176, 364, 36 },
		{ 176, 365, 36 },
		{ 176, 366, 36 },
		{ 176, 367, 36 },
		{ 176, 368, 36 },
		{ 176, 369, 36 },
		{ 176, 370, 36 },
		{ 176, 371, 36 },
		{ 176, 372, 36 },
		{ 176, 373, 36 },
		{ 176, 374, 36 },
		{ 176, 375, 36 },
		{ 176, 376, 36 },
		{ 176, 377, 36 },
		{ 176, 378, 36 },
		{ 176, 379, 36 },
		{ 176, 380, 36 },
		{ 176, 381, 36 },
		{ 176, 382, 36 },
		{ 176, 383, 36 },
		{ 176, 384, 36 },
		{ 176, 385, 36 },
		{ 176, 386, 36 },
		{ 177, 349, 36 },
		{ 177, 350, 36 },
		{ 177, 351, 36 },
		{ 177, 352, 36 },
		{ 177, 353, 36 },
		{ 177, 354, 36 },
		{ 177, 355, 36 },
		{ 177, 356, 36 },
		{ 177, 357, 36 },
		{ 177, 358, 36 },
		{ 177, 359, 36 },
		{ 177, 360, 36 },
		{ 177, 361, 36 },
		{ 177, 362, 36 },
		{ 177, 363, 36 },
		{ 177, 364, 36 },
		{ 177, 365, 36 },
		{ 177, 366, 36 },
		{ 177, 367, 36 },
		{ 177, 368, 36 },
		{ 177, 369, 36 },
		{ 177, 370, 36 },
		{ 177, 371, 36 },
		{ 177, 372, 36 },
		{ 177, 373, 36 },
		{ 177, 374, 36 },
		{ 177, 375, 36 },
		{ 177, 376, 36 },
		{ 177, 377, 36 },
		{ 177, 378, 36 },
		{ 177, 379, 36 },
		{ 177, 380, 36 },
		{ 177, 381, 36 },
		{ 177, 382, 36 },
		{ 177, 383, 36 },
		{ 177, 384, 36 },
		{ 178, 348, 36 },
		{ 178, 349, 36 },
		{ 178, 350, 36 },
		{ 178, 351, 36 },
		{ 178, 352, 36 },
		{ 178, 353, 36 },
		{ 178, 354, 36 },
		{ 178, 355, 36 },
		{ 178, 356, 36 },
		{ 178, 357, 36 },
		{ 178, 358, 36 },
		{ 178, 359, 36 },
		{ 178, 360, 36 },
		{ 178, 361, 36 },
		{ 178, 362, 36 },
		{ 178, 363, 36 },
		{ 178, 364, 36 },
		{ 178, 365, 36 },
		{ 178, 366, 36 },
		{ 178, 367, 36 },
		{ 178, 368, 36 },
		{ 178, 369, 36 },
		{ 178, 370, 36 },
		{ 178, 371, 36 },
		{ 178, 372, 36 },
		{ 178, 373, 36 },
		{ 178, 374, 36 },
		{ 178, 375, 36 },
		{ 178, 376, 36 },
		{ 178, 377, 36 },
		{ 178, 378, 36 },
		{ 178, 379, 36 },
		{ 178, 380, 36 },
		{ 178, 381, 36 },
		{ 178, 382, 36 },
		{ 178, 383, 36 },
		{ 179, 347, 36 },
		{ 179, 348, 36 },
		{ 179, 349, 36 },
		{ 179, 350, 36 },
		{ 179, 351, 36 },
		{ 179, 352, 36 },
		{ 179, 353, 36 },
		{ 179, 354, 36 },
		{ 179, 355, 36 },
		{ 179, 356, 36 },
		{ 179, 357, 36 },
		{ 179, 358, 36 },
		{ 179, 359, 36 },
		{ 179, 360, 36 },
		{ 179, 361, 36 },
		{ 179, 362, 36 },
		{ 179, 363, 36 },
		{ 179, 364, 36 },
		{ 179, 365, 36 },
		{ 179, 366, 36 },
		{ 179, 367, 36 },
		{ 179, 368, 36 },
		{ 179, 369, 36 },
		{ 179, 370, 36 },
		{ 179, 371, 36 },
		{ 179, 372, 36 },
		{ 179, 373, 36 },
		{ 179, 374, 36 },
		{ 179, 375, 36 },
		{ 179, 376, 36 },
		{ 179, 377, 36 },
		{ 179, 378, 36 },
		{ 179, 379, 36 },
		{ 179, 380, 36 },
		{ 179, 381, 36 },
		{ 180, 346, 36 },
		{ 180, 347, 36 },
		{ 180, 348, 36 },
		{ 180, 349, 36 },
		{ 180, 350, 36 },
		{ 180, 351, 36 },
		{ 180, 352, 36 },
		{ 180, 353, 36 },
		{ 180, 354, 36 },
		{ 180, 355, 36 },
		{ 180, 356, 36 },
		{ 180, 357, 36 },
		{ 180, 358, 36 },
		{ 180, 359, 36 },
		{ 180, 360, 36 },
		{ 180, 361, 36 },
		{ 180, 362, 36 },
		{ 180, 363, 36 },
		{ 180, 364, 36 },
		{ 180, 365, 36 },
		{ 180, 366, 36 },
		{ 180, 367, 36 },
		{ 180, 368, 36 },
		{ 180, 369, 36 },
		{ 180, 370, 36 },
		{ 180, 371, 36 },
		{ 180, 372, 36 },
		{ 180, 373, 36 },
		{ 180, 374, 36 },
		{ 180, 375, 36 },
		{ 180, 376, 36 },
		{ 180, 377, 36 },
		{ 180, 378, 36 },
		{ 180, 379, 36 },
		{ 180, 380, 36 },
		{ 181, 344, 36 },
		{ 181, 345, 36 },
		{ 181, 346, 36 },
		{ 181, 347, 36 },
		{ 181, 348, 36 },
		{ 181, 349, 36 },
		{ 181, 350, 36 },
		{ 181, 351, 36 },
		{ 181, 352, 36 },
		{ 181, 353, 36 },
		{ 181, 354, 36 },
		{ 181, 355, 36 },
		{ 181, 356, 36 },
		{ 181, 357, 36 },
		{ 181, 358, 36 },
		{ 181, 359, 36 },
		{ 181, 360, 36 },
		{ 181, 361, 36 },
		{ 181, 362, 36 },
		{ 181, 363, 36 },
		{ 181, 364, 36 },
		{ 181, 365, 36 },
		{ 181, 366, 36 },
		{ 181, 367, 36 },
		{ 181, 368, 36 },
		{ 181, 369, 36 },
		{ 181, 370, 36 },
		{ 181, 371, 36 },
		{ 181, 372, 36 },
		{ 181, 373, 36 },
		{ 181, 374, 36 },
		{ 181, 375, 36 },
		{ 181, 376, 36 },
		{ 181, 377, 36 },
		{ 181, 378, 36 },
		{ 182, 343, 36 },
		{ 182, 344, 36 },
		{ 182, 345, 36 },
		{ 182, 346, 36 },
		{ 182, 347, 36 },
		{ 182, 348, 36 },
		{ 182, 349, 36 },
		{ 182, 350, 36 },
		{ 182, 351, 36 },
		{ 182, 352, 36 },
		{ 182, 353, 36 },
		{ 182, 354, 36 },
		{ 182, 355, 36 },
		{ 182, 356, 36 },
		{ 182, 357, 36 },
		{ 182, 358, 36 },
		{ 182, 359, 36 },
		{ 182, 360, 36 },
		{ 182, 361, 36 },
		{ 182, 362, 36 },
		{ 182, 363, 36 },
		{ 182, 364, 36 },
		{ 182, 365, 36 },
		{ 182, 366, 36 },
		{ 182, 367, 36 },
		{ 182, 368, 36 },
		{ 182, 369, 36 },
		{ 182, 370, 36 },
		{ 182, 371, 36 },
		{ 182, 372, 36 },
		{ 182, 373, 36 },
		{ 182, 374, 36 },
		{ 182, 375, 36 },
		{ 182, 376, 36 },
		{ 183, 342, 36 },
		{ 183, 343, 36 },
		{ 183, 344, 36 },
		{ 183, 345, 36 },
		{ 183, 346, 36 },
		{ 183, 347, 36 },
		{ 183, 348, 36 },
		{ 183, 349, 36 },
		{ 183, 350, 36 },
		{ 183, 351, 36 },
		{ 183, 352, 36 },
		{ 183, 353, 36 },
		{ 183, 354, 36 },
		{ 183, 355, 36 },
		{ 183, 356, 36 },
		{ 183, 357, 36 },
		{ 183, 358, 36 },
		{ 183, 359, 36 },
		{ 183, 360, 36 },
		{ 183, 361, 36 },
		{ 183, 362, 36 },
		{ 183, 363, 36 },
		{ 183, 364, 36 },
		{ 183, 365, 36 },
		{ 183, 366, 36 },
		{ 183, 367, 36 },
		{ 183, 368, 36 },
		{ 183, 369, 36 },
		{ 183, 370, 36 },
		{ 183, 371, 36 },
		{ 183, 372, 36 },
		{ 183, 373, 36 },
		{ 183, 374, 36 },
		{ 183, 375, 36 },
		{ 184, 341, 36 },
		{ 184, 342, 36 },
		{ 184, 343, 36 },
		{ 184, 344, 36 },
		{ 184, 345, 36 },
		{ 184, 346, 36 },
		{ 184, 347, 36 },
		{ 184, 348, 36 },
		{ 184, 349, 36 },
		{ 184, 350, 36 },
		{ 184, 351, 36 },
		{ 184, 352, 36 },
		{ 184, 353, 36 },
		{ 184, 354, 36 },
		{ 184, 355, 36 },
		{ 184, 356, 36 },
		{ 184, 357, 36 },
		{ 184, 358, 36 },
		{ 184, 359, 36 },
		{ 184, 360, 36 },
		{ 184, 361, 36 },
		{ 184, 362, 36 },
		{ 184, 363, 36 },
		{ 184, 364, 36 },
		{ 184, 365, 36 },
		{ 184, 366, 36 },
		{ 184, 367, 36 },
		{ 184, 368, 36 },
		{ 184, 369, 36 },
		{ 184, 370, 36 },
		{ 184, 371, 36 },
		{ 184, 372, 36 },
		{ 184, 373, 36 },
		{ 185, 340, 36 },
		{ 185, 341, 36 },
		{ 185, 342, 36 },
		{ 185, 343, 36 },
		{ 185, 344, 36 },
		{ 185, 345, 36 },
		{ 185, 346, 36 },
		{ 185, 347, 36 },
		{ 185, 348, 36 },
		{ 185, 349, 36 },
		{ 185, 350, 36 },
		{ 185, 351, 36 },
		{ 185, 352, 36 },
		{ 185, 353, 36 },
		{ 185, 354, 36 },
		{ 185, 355, 36 },
		{ 185, 356, 36 },
		{ 185, 357, 36 },
		{ 185, 358, 36 },
		{ 185, 359, 36 },
		{ 185, 360, 36 },
		{ 185, 361, 36 },
		{ 185, 362, 36 },
		{ 185, 363, 36 },
		{ 185, 364, 36 },
		{ 185, 365, 36 },
		{ 185, 366, 36 },
		{ 185, 367, 36 },
		{ 185, 368, 36 },
		{ 185, 369, 36 },
		{ 185, 370, 36 },
		{ 185, 371, 36 },
		{ 185, 372, 36 },
		{ 186, 339, 36 },
		{ 186, 340, 36 },
		{ 186, 341, 36 },
		{ 186, 342, 36 },
		{ 186, 343, 36 },
		{ 186, 344, 36 },
		{ 186, 345, 36 },
		{ 186, 346, 36 },
		{ 186, 347, 36 },
		{ 186, 348, 36 },
		{ 186, 349, 36 },
		{ 186, 350, 36 },
		{ 186, 351, 36 },
		{ 186, 352, 36 },
		{ 186, 353, 36 },
		{ 186, 354, 36 },
		{ 186, 355, 36 },
		{ 186, 356, 36 },
		{ 186, 357, 36 },
		{ 186, 358, 36 },
		{ 186, 359, 36 },
		{ 186, 360, 36 },
		{ 186, 361, 36 },
		{ 186, 362, 36 },
		{ 186, 363, 36 },
		{ 186, 364, 36 },
		{ 186, 365, 36 },
		{ 186, 366, 36 },
		{ 186, 367, 36 },
		{ 186, 368, 36 },
		{ 186, 369, 36 },
		{ 186, 370, 36 },
		{ 187, 338, 36 },
		{ 187, 339, 36 },
		{ 187, 340, 36 },
		{ 187, 341, 36 },
		{ 187, 342, 36 },
		{ 187, 343, 36 },
		{ 187, 344, 36 },
		{ 187, 345, 36 },
		{ 187, 346, 36 },
		{ 187, 347, 36 },
		{ 187, 348, 36 },
		{ 187, 349, 36 },
		{ 187, 350, 36 },
		{ 187, 351, 36 },
		{ 187, 352, 36 },
		{ 187, 353, 36 },
		{ 187, 354, 36 },
		{ 187, 355, 36 },
		{ 187, 356, 36 },
		{ 187, 357, 36 },
		{ 187, 358, 36 },
		{ 187, 359, 36 },
		{ 187, 360, 36 },
		{ 187, 361, 36 },
		{ 187, 362, 36 },
		{ 187, 363, 36 },
		{ 187, 364, 36 },
		{ 187, 365, 36 },
		{ 187, 366, 36 },
		{ 187, 367, 36 },
		{ 187, 368, 36 },
		{ 188, 336, 36 },
		{ 188, 337, 36 },
		{ 188, 338, 36 },
		{ 188, 339, 36 },
		{ 188, 340, 36 },
		{ 188, 341, 36 },
		{ 188, 342, 36 },
		{ 188, 343, 36 },
		{ 188, 344, 36 },
		{ 188, 345, 36 },
		{ 188, 346, 36 },
		{ 188, 347, 36 },
		{ 188, 348, 36 },
		{ 188, 349, 36 },
		{ 188, 350, 36 },
		{ 188, 351, 36 },
		{ 188, 352, 36 },
		{ 188, 353, 36 },
		{ 188, 354, 36 },
		{ 188, 355, 36 },
		{ 188, 356, 36 },
		{ 188, 357, 36 },
		{ 188, 358, 36 },
		{ 188, 359, 36 },
		{ 188, 360, 36 },
		{ 188, 361, 36 },
		{ 188, 362, 36 },
		{ 188, 363, 36 },
		{ 188, 364, 36 },
		{ 188, 365, 36 },
		{ 188, 366, 36 },
		{ 188, 367, 36 },
		{ 189, 335, 36 },
		{ 189, 336, 36 },
		{ 189, 337, 36 },
		{ 189, 338, 36 },
		{ 189, 339, 36 },
		{ 189, 340, 36 },
		{ 189, 341, 36 },
		{ 189, 342, 36 },
		{ 189, 343, 36 },
		{ 189, 344, 36 },
		{ 189, 345, 36 },
		{ 189, 346, 36 },
		{ 189, 347, 36 },
		{ 189, 348, 36 },
		{ 189, 349, 36 },
		{ 189, 350, 36 },
		{ 189, 351, 36 },
		{ 189, 352, 36 },
		{ 189, 353, 36 },
		{ 189, 354, 36 },
		{ 189, 355, 36 },
		{ 189, 356, 36 },
		{ 189, 357, 36 },
		{ 189, 358, 36 },
		{ 189, 359, 36 },
		{ 189, 360, 36 },
		{ 189, 361, 36 },
		{ 189, 362, 36 },
		{ 189, 363, 36 },
		{ 189, 364, 36 },
		{ 189, 365, 36 },
		{ 190, 334, 36 },
		{ 190, 335, 36 },
		{ 190, 336, 36 },
		{ 190, 337, 36 },
		{ 190, 338, 36 },
		{ 190, 339, 36 },
		{ 190, 340, 36 },
		{ 190, 341, 36 },
		{ 190, 342, 36 },
		{ 190, 343, 36 },
		{ 190, 344, 36 },
		{ 190, 345, 36 },
		{ 190, 346, 36 },
		{ 190, 347, 36 },
		{ 190, 348, 36 },
		{ 190, 349, 36 },
		{ 190, 350, 36 },
		{ 190, 351, 36 },
		{ 190, 352, 36 },
		{ 190, 353, 36 },
		{ 190, 354, 36 },
		{ 190, 355, 36 },
		{ 190, 356, 36 },
		{ 190, 357, 36 },
		{ 190, 358, 36 },
		{ 190, 359, 36 },
		{ 190, 360, 36 },
		{ 190, 361, 36 },
		{ 190, 362, 36 },
		{ 190, 363, 36 },
		{ 190, 364, 36 },
		{ 191, 333, 36 },
		{ 191, 334, 36 },
		{ 191, 335, 36 },
		{ 191, 336, 36 },
		{ 191, 337, 36 },
		{ 191, 338, 36 },
		{ 191, 339, 36 },
		{ 191, 340, 36 },
		{ 191, 341, 36 },
		{ 191, 342, 36 },
		{ 191, 343, 36 },
		{ 191, 344, 36 },
		{ 191, 345, 36 },
		{ 191, 346, 36 },
		{ 191, 347, 36 },
		{ 191, 348, 36 },
		{ 191, 349, 36 },
		{ 191, 350, 36 },
		{ 191, 351, 36 },
		{ 191, 352, 36 },
		{ 191, 353, 36 },
		{ 191, 354, 36 },
		{ 191, 355, 36 },
		{ 191, 356, 36 },
		{ 191, 357, 36 },
		{ 191, 358, 36 },
		{ 191, 359, 36 },
		{ 191, 360, 36 },
		{ 191, 361, 36 },
		{ 191, 362, 36 },
		{ 192, 332, 36 },
		{ 192, 333, 36 },
		{ 192, 334, 36 },
		{ 192, 335, 36 },
		{ 192, 336, 36 },
		{ 192, 337, 36 },
		{ 192, 338, 36 },
		{ 192, 339, 36 },
		{ 192, 340, 36 },
		{ 192, 341, 36 },
		{ 192, 342, 36 },
		{ 192, 343, 36 },
		{ 192, 344, 36 },
		{ 192, 345, 36 },
		{ 192, 346, 36 },
		{ 192, 347, 36 },
		{ 192, 348, 36 },
		{ 192, 349, 36 },
		{ 192, 350, 36 },
		{ 192, 351, 36 },
		{ 192, 352, 36 },
		{ 192, 353, 36 },
		{ 192, 354, 36 },
		{ 192, 355, 36 },
		{ 192, 356, 36 },
		{ 192, 357, 36 },
		{ 192, 358, 36 },
		{ 192, 359, 36 },
		{ 192, 360, 36 },
		{ 193, 331, 36 },
		{ 193, 332, 36 },
		{ 193, 333, 36 },
		{ 193, 334, 36 },
		{ 193, 335, 36 },
		{ 193, 336, 36 },
		{ 193, 337, 36 },
		{ 193, 338, 36 },
		{ 193, 339, 36 },
		{ 193, 340, 36 },
		{ 193, 341, 36 },
		{ 193, 342, 36 },
		{ 193, 343, 36 },
		{ 193, 344, 36 },
		{ 193, 345, 36 },
		{ 193, 346, 36 },
		{ 193, 347, 36 },
		{ 193, 348, 36 },
		{ 193, 349, 36 },
		{ 193, 350, 36 },
		{ 193, 351, 36 },
		{ 193, 352, 36 },
		{ 193, 353, 36 },
		{ 193, 354, 36 },
		{ 193, 355, 36 },
		{ 193, 356, 36 },
		{ 193, 357, 36 },
		{ 193, 358, 36 },
		{ 193, 359, 36 },
		{ 194, 330, 36 },
		{ 194, 331, 36 },
		{ 194, 332, 36 },
		{ 194, 333, 36 },
		{ 194, 334, 36 },
		{ 194, 335, 36 },
		{ 194, 336, 36 },
		{ 194, 337, 36 },
		{ 194, 338, 36 },
		{ 194, 339, 36 },
		{ 194, 340, 36 },
		{ 194, 341, 36 },
		{ 194, 342, 36 },
		{ 194, 343, 36 },
		{ 194, 344, 36 },
		{ 194, 345, 36 },
		{ 194, 346, 36 },
		{ 194, 347, 36 },
		{ 194, 348, 36 },
		{ 194, 349, 36 },
		{ 194, 350, 36 },
		{ 194, 351, 36 },
		{ 194, 352, 36 },
		{ 194, 353, 36 },
		{ 194, 354, 36 },
		{ 194, 355, 36 },
		{ 194, 356, 36 },
		{ 194, 357, 36 },
		{ 195, 329, 36 },
		{ 195, 330, 36 },
		{ 195, 331, 36 },
		{ 195, 332, 36 },
		{ 195, 333, 36 },
		{ 195, 334, 36 },
		{ 195, 335, 36 },
		{ 195, 336, 36 },
		{ 195, 337, 36 },
		{ 195, 338, 36 },
		{ 195, 339, 36 },
		{ 195, 340, 36 },
		{ 195, 341, 36 },
		{ 195, 342, 36 },
		{ 195, 343, 36 },
		{ 195, 344, 36 },
		{ 195, 345, 36 },
		{ 195, 346, 36 },
		{ 195, 347, 36 },
		{ 195, 348, 36 },
		{ 195, 349, 36 },
		{ 195, 350, 36 },
		{ 195, 351, 36 },
		{ 195, 352, 36 },
		{ 195, 353, 36 },
		{ 195, 354, 36 },
		{ 195, 355, 36 },
		{ 195, 356, 36 },
		{ 196, 327, 36 },
		{ 196, 328, 36 },
		{ 196, 329, 36 },
		{ 196, 330, 36 },
		{ 196, 331, 36 },
		{ 196, 332, 36 },
		{ 196, 333, 36 },
		{ 196, 334, 36 },
		{ 196, 335, 36 },
		{ 196, 336, 36 },
		{ 196, 337, 36 },
		{ 196, 338, 36 },
		{ 196, 339, 36 },
		{ 196, 340, 36 },
		{ 196, 341, 36 },
		{ 196, 342, 36 },
		{ 196, 343, 36 },
		{ 196, 344, 36 },
		{ 196, 345, 36 },
		{ 196, 346, 36 },
		{ 196, 347, 36 },
		{ 196, 348, 36 },
		{ 196, 349, 36 },
		{ 196, 350, 36 },
		{ 196, 351, 36 },
		{ 196, 352, 36 },
		{ 196, 353, 36 },
		{ 196, 354, 36 },
		{ 197, 326, 36 },
		{ 197, 327, 36 },
		{ 197, 328, 36 },
		{ 197, 329, 36 },
		{ 197, 330, 36 },
		{ 197, 331, 36 },
		{ 197, 332, 36 },
		{ 197, 333, 36 },
		{ 197, 334, 36 },
		{ 197, 335, 36 },
		{ 197, 336, 36 },
		{ 197, 337, 36 },
		{ 197, 338, 36 },
		{ 197, 339, 36 },
		{ 197, 340, 36 },
		{ 197, 341, 36 },
		{ 197, 342, 36 },
		{ 197, 343, 36 },
		{ 197, 344, 36 },
		{ 197, 345, 36 },
		{ 197, 346, 36 },
		{ 197, 347, 36 },
		{ 197, 348, 36 },
		{ 197, 349, 36 },
		{ 197, 350, 36 },
		{ 197, 351, 36 },
		{ 197, 352, 36 },
		{ 198, 325, 36 },
		{ 198, 326, 36 },
		{ 198, 327, 36 },
		{ 198, 328, 36 },
		{ 198, 329, 36 },
		{ 198, 330, 36 },
		{ 198, 331, 36 },
		{ 198, 332, 36 },
		{ 198, 333, 36 },
		{ 198, 334, 36 },
		{ 198, 335, 36 },
		{ 198, 336, 36 },
		{ 198, 337, 36 },
		{ 198, 338, 36 },
		{ 198, 339, 36 },
		{ 198, 340, 36 },
		{ 198, 341, 36 },
		{ 198, 342, 36 },
		{ 198, 343, 36 },
		{ 198, 344, 36 },
		{ 198, 345, 36 },
		{ 198, 346, 36 },
		{ 198, 347, 36 },
		{ 198, 348, 36 },
		{ 198, 349, 36 },
		{ 198, 350, 36 },
		{ 198, 351, 36 },
		{ 199, 325, 36 },
		{ 199, 326, 36 },
		{ 199, 327, 36 },
		{ 199, 328, 36 },
		{ 199, 329, 36 },
		{ 199, 330, 36 },
		{ 199, 331, 36 },
		{ 199, 332, 36 },
		{ 199, 333, 36 },
		{ 199, 334, 36 },
		{ 199, 335, 36 },
		{ 199, 336, 36 },
		{ 199, 337, 36 },
		{ 199, 338, 36 },
		{ 199, 339, 36 },
		{ 199, 340, 36 },
		{ 199, 341, 36 },
		{ 199, 342, 36 },
		{ 199, 343, 36 },
		{ 199, 344, 36 },
		{ 199, 345, 36 },
		{ 199, 346, 36 },
		{ 199, 347, 36 },
		{ 199, 348, 36 },
		{ 199, 349, 36 },
		{ 200, 324, 36 },
		{ 200, 325, 36 },
		{ 200, 326, 36 },
		{ 200, 327, 36 },
		{ 200, 328, 36 },
		{ 200, 329, 36 },
		{ 200, 330, 36 },
		{ 200, 331, 36 },
		{ 200, 332, 36 },
		{ 200, 333, 36 },
		{ 200, 334, 36 },
		{ 200, 335, 36 },
		{ 200, 336, 36 },
		{ 200, 337, 36 },
		{ 200, 338, 36 },
		{ 200, 339, 36 },
		{ 200, 340, 36 },
		{ 200, 341, 36 },
		{ 200, 342, 36 },
		{ 200, 343, 36 },
		{ 200, 344, 36 },
		{ 200, 345, 36 },
		{ 200, 346, 36 },
		{ 200, 347, 36 },
		{ 200, 348, 36 },
		{ 201, 324, 36 },
		{ 201, 325, 36 },
		{ 201, 326, 36 },
		{ 201, 327, 36 },
		{ 201, 328, 36 },
		{ 201, 329, 36 },
		{ 201, 330, 36 },
		{ 201, 331, 36 },
		{ 201, 332, 36 },
		{ 201, 333, 36 },
		{ 201, 334, 36 },
		{ 201, 335, 36 },
		{ 201, 336, 36 },
		{ 201, 337, 36 },
		{ 201, 338, 36 },
		{ 201, 339, 36 },
		{ 201, 340, 36 },
		{ 201, 341, 36 },
		{ 201, 342, 36 },
		{ 201, 343, 36 },
		{ 201, 344, 36 },
		{ 201, 345, 36 },
		{ 201, 346, 36 },
		{ 202, 323, 36 },
		{ 202, 324, 36 },
		{ 202, 325, 36 },
		{ 202, 326, 36 },
		{ 202, 327, 36 },
		{ 202, 328, 36 },
		{ 202, 329, 36 },
		{ 202, 330, 36 },
		{ 202, 331, 36 },
		{ 202, 332, 36 },
		{ 202, 333, 36 },
		{ 202, 334, 36 },
		{ 202, 335, 36 },
		{ 202, 336, 36 },
		{ 202, 337, 36 },
		{ 202, 338, 36 },
		{ 202, 339, 36 },
		{ 202, 340, 36 },
		{ 202, 341, 36 },
		{ 202, 342, 36 },
		{ 202, 343, 36 },
		{ 202, 344, 36 },
		{ 203, 323, 36 },
		{ 203, 324, 36 },
		{ 203, 325, 36 },
		{ 203, 326, 36 },
		{ 203, 327, 36 },
		{ 203, 328, 36 },
		{ 203, 329, 36 },
		{ 203, 330, 36 },
		{ 203, 331, 36 },
		{ 203, 332, 36 },
		{ 203, 333, 36 },
		{ 203, 334, 36 },
		{ 203, 335, 36 },
		{ 203, 336, 36 },
		{ 203, 337, 36 },
		{ 203, 338, 36 },
		{ 203, 339, 36 },
		{ 203, 340, 36 },
		{ 203, 341, 36 },
		{ 203, 342, 36 },
		{ 203, 343, 36 },
		{ 204, 323, 36 },
		{ 204, 324, 36 },
		{ 204, 325, 36 },
		{ 204, 326, 36 },
		{ 204, 327, 36 },
		{ 204, 328, 36 },
		{ 204, 329, 36 },
		{ 204, 330, 36 },
		{ 204, 331, 36 },
		{ 204, 332, 36 },
		{ 204, 333, 36 },
		{ 204, 334, 36 },
		{ 204, 335, 36 },
		{ 204, 336, 36 },
		{ 204, 337, 36 },
		{ 204, 338, 36 },
		{ 204, 339, 36 },
		{ 204, 340, 36 },
		{ 204, 341, 36 },
		{ 205, 323, 36 },
		{ 205, 324, 36 },
		{ 205, 325, 36 },
		{ 205, 326, 36 },
		{ 205, 327, 36 },
		{ 205, 328, 36 },
		{ 205, 329, 36 },
		{ 205, 330, 36 },
		{ 205, 331, 36 },
		{ 205, 332, 36 },
		{ 205, 333, 36 },
		{ 205, 334, 36 },
		{ 205, 335, 36 },
		{ 205, 336, 36 },
		{ 205, 337, 36 },
		{ 205, 338, 36 },
		{ 205, 339, 36 },
		{ 205, 340, 36 },
		{ 206, 323, 36 },
		{ 206, 324, 36 },
		{ 206, 325, 36 },
		{ 206, 326, 36 },
		{ 206, 327, 36 },
		{ 206, 328, 36 },
		{ 206, 329, 36 },
		{ 206, 330, 36 },
		{ 206, 331, 36 },
		{ 206, 332, 36 },
		{ 206, 333, 36 },
		{ 206, 334, 36 },
		{ 206, 335, 36 },
		{ 206, 336, 36 },
		{ 206, 337, 36 },
		{ 206, 338, 36 },
		{ 207, 323, 36 },
		{ 207, 324, 36 },
		{ 207, 325, 36 },
		{ 207, 326, 36 },
		{ 207, 327, 36 },
		{ 207, 328, 36 },
		{ 207, 329, 36 },
		{ 207, 330, 36 },
		{ 207, 331, 36 },
		{ 207, 332, 36 },
		{ 207, 333, 36 },
		{ 207, 334, 36 },
		{ 207, 335, 36 },
		{ 207, 336, 36 },
		{ 208, 323, 36 },
		{ 208, 324, 36 },
		{ 208, 325, 36 },
		{ 208, 326, 36 },
		{ 208, 327, 36 },
		{ 208, 328, 36 },
		{ 208, 329, 36 },
		{ 208, 330, 36 },
		{ 208, 331, 36 },
		{ 208, 332, 36 },
		{ 208, 333, 36 },
		{ 208, 334, 36 },
		{ 208, 335, 36 },
		{ 209, 323, 36 },
		{ 209, 324, 36 },
		{ 209, 325, 36 },
		{ 209, 326, 36 },
		{ 209, 327, 36 },
		{ 209, 328, 36 },
		{ 209, 329, 36 },
		{ 209, 330, 36 },
		{ 209, 331, 36 },
		{ 209, 332, 36 },
		{ 209, 333, 36 },
		{ 210, 323, 36 },
		{ 210, 324, 36 },
		{ 210, 325, 36 },
		{ 210, 326, 36 },
		{ 210, 327, 36 },
		{ 210, 328, 36 },
		{ 210, 329, 36 },
		{ 210, 330, 36 },
		{ 210, 331, 36 },
		{ 210, 332, 36 },
		{ 211, 323, 36 },
		{ 211, 324, 36 },
		{ 211, 325, 36 },
		{ 211, 326, 36 },
		{ 211, 327, 36 },
		{ 211, 328, 36 },
		{ 211, 329, 36 },
		{ 211, 330, 36 },
		{ 212, 323, 36 },
		{ 212, 324, 36 },
		{ 212, 325, 36 },
		{ 212, 326, 36 },
		{ 212, 327, 36 },
		{ 212, 328, 36 },
		{ 213, 323, 36 },
		{ 213, 324, 36 },
		{ 213, 325, 36 },
		{ 213, 326, 36 },
		{ 213, 327, 36 },
		{ 214, 324, 36 },
		{ 214, 325, 36 },
		{ 300, 189, 36 },
		{ 300, 190, 36 },
		{ 301, 187, 36 },
		{ 301, 188, 36 },
		{ 301, 189, 36 },
		{ 301, 190, 36 },
		{ 301, 191, 36 },
		{ 302, 186, 36 },
		{ 302, 187, 36 },
		{ 302, 188, 36 },
		{ 302, 189, 36 },
		{ 302, 190, 36 },
		{ 302, 191, 36 },
		{ 303, 184, 36 },
		{ 303, 185, 36 },
		{ 303, 186, 36 },
		{ 303, 187, 36 },
		{ 303, 188, 36 },
		{ 303, 189, 36 },
		{ 303, 190, 36 },
		{ 303, 191, 36 },
		{ 304, 182, 36 },
		{ 304, 183, 36 },
		{ 304, 184, 36 },
		{ 304, 185, 36 },
		{ 304, 186, 36 },
		{ 304, 187, 36 },
		{ 304, 188, 36 },
		{ 304, 189, 36 },
		{ 304, 190, 36 },
		{ 304, 191, 36 },
		{ 305, 181, 36 },
		{ 305, 182, 36 },
		{ 305, 183, 36 },
		{ 305, 184, 36 },
		{ 305, 185, 36 },
		{ 305, 186, 36 },
		{ 305, 187, 36 },
		{ 305, 188, 36 },
		{ 305, 189, 36 },
		{ 305, 190, 36 },
		{ 305, 191, 36 },
		{ 306, 179, 36 },
		{ 306, 180, 36 },
		{ 306, 181, 36 },
		{ 306, 182, 36 },
		{ 306, 183, 36 },
		{ 306, 184, 36 },
		{ 306, 185, 36 },
		{ 306, 186, 36 },
		{ 306, 187, 36 },
		{ 306, 188, 36 },
		{ 306, 189, 36 },
		{ 306, 190, 36 },
		{ 306, 191, 36 },
		{ 307, 178, 36 },
		{ 307, 179, 36 },
		{ 307, 180, 36 },
		{ 307, 181, 36 },
		{ 307, 182, 36 },
		{ 307, 183, 36 },
		{ 307, 184, 36 },
		{ 307, 185, 36 },
		{ 307, 186, 36 },
		{ 307, 187, 36 },
		{ 307, 188, 36 },
		{ 307, 189, 36 },
		{ 307, 190, 36 },
		{ 307, 191, 36 },
		{ 308, 176, 36 },
		{ 308, 177, 36 },
		{ 308, 178, 36 },
		{ 308, 179, 36 },
		{ 308, 180, 36 },
		{ 308, 181, 36 },
		{ 308, 182, 36 },
		{ 308, 183, 36 },
		{ 308, 184, 36 },
		{ 308, 185, 36 },
		{ 308, 186, 36 },
		{ 308, 187, 36 },
		{ 308, 188, 36 },
		{ 308, 189, 36 },
		{ 308, 190, 36 },
		{ 308, 191, 36 },
		{ 309, 174, 36 },
		{ 309, 175, 36 },
		{ 309, 176, 36 },
		{ 309, 177, 36 },
		{ 309, 178, 36 },
		{ 309, 179, 36 },
		{ 309, 180, 36 },
		{ 309, 181, 36 },
		{ 309, 182, 36 },
		{ 309, 183, 36 },
		{ 309, 184, 36 },
		{ 309, 185, 36 },
		{ 309, 186, 36 },
		{ 309, 187, 36 },
		{ 309, 188, 36 },
		{ 309, 189, 36 },
		{ 309, 190, 36 },
		{ 309, 191, 36 },
		{ 310, 173, 36 },
		{ 310, 174, 36 },
		{ 310, 175, 36 },
		{ 310, 176, 36 },
		{ 310, 177, 36 },
		{ 310, 178, 36 },
		{ 310, 179, 36 },
		{ 310, 180, 36 },
		{ 310, 181, 36 },
		{ 310, 182, 36 },
		{ 310, 183, 36 },
		{ 310, 184, 36 },
		{ 310, 185, 36 },
		{ 310, 186, 36 },
		{ 310, 187, 36 },
		{ 310, 188, 36 },
		{ 310, 189, 36 },
		{ 310, 190, 36 },
		{ 310, 191, 36 },
		{ 311, 171, 36 },
		{ 311, 172, 36 },
		{ 311, 173, 36 },
		{ 311, 174, 36 },
		{ 311, 175, 36 },
		{ 311, 176, 36 },
		{ 311, 177, 36 },
		{ 311, 178, 36 },
		{ 311, 179, 36 },
		{ 311, 180, 36 },
		{ 311, 181, 36 },
		{ 311, 182, 36 },
		{ 311, 183, 36 },
		{ 311, 184, 36 },
		{ 311, 185, 36 },
		{ 311, 186, 36 },
		{ 311, 187, 36 },
		{ 311, 188, 36 },
		{ 311, 189, 36 },
		{ 311, 190, 36 },
		{ 311, 191, 36 },
		{ 312, 170, 36 },
		{ 312, 171, 36 },
		{ 312, 172, 36 },
		{ 312, 173, 36 },
		{ 312, 174, 36 },
		{ 312, 175, 36 },
		{ 312, 176, 36 },
		{ 312, 177, 36 },
		{ 312, 178, 36 },
		{ 312, 179, 36 },
		{ 312, 180, 36 },
		{ 312, 181, 36 },
		{ 312, 182, 36 },
		{ 312, 183, 36 },
		{ 312, 184, 36 },
		{ 312, 185, 36 },
		{ 312, 186, 36 },
		{ 312, 187, 36 },
		{ 312, 188, 36 },
		{ 312, 189, 36 },
		{ 312, 190, 36 },
		{ 312, 191, 36 },
		{ 313, 168, 36 },
		{ 313, 169, 36 },
		{ 313, 170, 36 },
		{ 313, 171, 36 },
		{ 313, 172, 36 },
		{ 313, 173, 36 },
		{ 313, 174, 36 },
		{ 313, 175, 36 },
		{ 313, 176, 36 },
		{ 313, 177, 36 },
		{ 313, 178, 36 },
		{ 313, 179, 36 },
		{ 313, 180, 36 },
		{ 313, 181, 36 },
		{ 313, 182, 36 },
		{ 313, 183, 36 },
		{ 313, 184, 36 },
		{ 313, 185, 36 },
		{ 313, 186, 36 },
		{ 313, 187, 36 },
		{ 313, 188, 36 },
		{ 313, 189, 36 },
		{ 313, 190, 36 },
		{ 314, 166, 36 },
		{ 314, 167, 36 },
		{ 314, 168, 36 },
		{ 314, 169, 36 },
		{ 314, 170, 36 },
		{ 314, 171, 36 },
		{ 314, 172, 36 },
		{ 314, 173, 36 },
		{ 314, 174, 36 },
		{ 314, 175, 36 },
		{ 314, 176, 36 },
		{ 314, 177, 36 },
		{ 314, 178, 36 },
		{ 314, 179, 36 },
		{ 314, 180, 36 },
		{ 314, 181, 36 },
		{ 314, 182, 36 },
		{ 314, 183, 36 },
		{ 314, 184, 36 },
		{ 314, 185, 36 },
		{ 314, 186, 36 },
		{ 314, 187, 36 },
		{ 314, 188, 36 },
		{ 314, 189, 36 },
		{ 314, 190, 36 },
		{ 315, 165, 36 },
		{ 315, 166, 36 },
		{ 315, 167, 36 },
		{ 315, 168, 36 },
		{ 315, 169, 36 },
		{ 315, 170, 36 },
		{ 315, 171, 36 },
		{ 315, 172, 36 },
		{ 315, 173, 36 },
		{ 315, 174, 36 },
		{ 315, 175, 36 },
		{ 315, 176, 36 },
		{ 315, 177, 36 },
		{ 315, 178, 36 },
		{ 315, 179, 36 },
		{ 315, 180, 36 },
		{ 315, 181, 36 },
		{ 315, 182, 36 },
		{ 315, 183, 36 },
		{ 315, 184, 36 },
		{ 315, 185, 36 },
		{ 315, 186, 36 },
		{ 315, 187, 36 },
		{ 315, 188, 36 },
		{ 315, 189, 36 },
		{ 316, 163, 36 },
		{ 316, 164, 36 },
		{ 316, 165, 36 },
		{ 316, 166, 36 },
		{ 316, 167, 36 },
		{ 316, 168, 36 },
		{ 316, 169, 36 },
		{ 316, 170, 36 },
		{ 316, 171, 36 },
		{ 316, 172, 36 },
		{ 316, 173, 36 },
		{ 316, 174, 36 },
		{ 316, 175, 36 },
		{ 316, 176, 36 },
		{ 316, 177, 36 },
		{ 316, 178, 36 },
		{ 316, 179, 36 },
		{ 316, 180, 36 },
		{ 316, 181, 36 },
		{ 316, 182, 36 },
		{ 316, 183, 36 },
		{ 316, 184, 36 },
		{ 316, 185, 36 },
		{ 316, 186, 36 },
		{ 316, 187, 36 },
		{ 316, 188, 36 },
		{ 316, 189, 36 },
		{ 317, 162, 36 },
		{ 317, 163, 36 },
		{ 317, 164, 36 },
		{ 317, 165, 36 },
		{ 317, 166, 36 },
		{ 317, 167, 36 },
		{ 317, 168, 36 },
		{ 317, 169, 36 },
		{ 317, 170, 36 },
		{ 317, 171, 36 },
		{ 317, 172, 36 },
		{ 317, 173, 36 },
		{ 317, 174, 36 },
		{ 317, 175, 36 },
		{ 317, 176, 36 },
		{ 317, 177, 36 },
		{ 317, 178, 36 },
		{ 317, 179, 36 },
		{ 317, 180, 36 },
		{ 317, 181, 36 },
		{ 317, 182, 36 },
		{ 317, 183, 36 },
		{ 317, 184, 36 },
		{ 317, 185, 36 },
		{ 317, 186, 36 },
		{ 317, 187, 36 },
		{ 317, 188, 36 },
		{ 318, 160, 36 },
		{ 318, 161, 36 },
		{ 318, 162, 36 },
		{ 318, 163, 36 },
		{ 318, 164, 36 },
		{ 318, 165, 36 },
		{ 318, 166, 36 },
		{ 318, 167, 36 },
		{ 318, 168, 36 },
		{ 318, 169, 36 },
		{ 318, 170, 36 },
		{ 318, 171, 36 },
		{ 318, 172, 36 },
		{ 318, 173, 36 },
		{ 318, 174, 36 },
		{ 318, 175, 36 },
		{ 318, 176, 36 },
		{ 318, 177, 36 },
		{ 318, 178, 36 },
		{ 318, 179, 36 },
		{ 318, 180, 36 },
		{ 318, 181, 36 },
		{ 318, 182, 36 },
		{ 318, 183, 36 },
		{ 318, 184, 36 },
		{ 318, 185, 36 },
		{ 318, 186, 36 },
		{ 318, 187, 36 },
		{ 319, 158, 36 },
		{ 319, 159, 36 },
		{ 319, 160, 36 },
		{ 319, 161, 36 },
		{ 319, 162, 36 },
		{ 319, 163, 36 },
		{ 319, 164, 36 },
		{ 319, 165, 36 },
		{ 319, 166, 36 },
		{ 319, 167, 36 },
		{ 319, 168, 36 },
		{ 319, 169, 36 },
		{ 319, 170, 36 },
		{ 319, 171, 36 },
		{ 319, 172, 36 },
		{ 319, 173, 36 },
		{ 319, 174, 36 },
		{ 319, 175, 36 },
		{ 319, 176, 36 },
		{ 319, 177, 36 },
		{ 319, 178, 36 },
		{ 319, 179, 36 },
		{ 319, 180, 36 },
		{ 319, 181, 36 },
		{ 319, 182, 36 },
		{ 319, 183, 36 },
		{ 319, 184, 36 },
		{ 319, 185, 36 },
		{ 320, 157, 36 },
		{ 320, 158, 36 },
		{ 320, 159, 36 },
		{ 320, 160, 36 },
		{ 320, 161, 36 },
		{ 320, 162, 36 },
		{ 320, 163, 36 },
		{ 320, 164, 36 },
		{ 320, 165, 36 },
		{ 320, 166, 36 },
		{ 320, 167, 36 },
		{ 320, 168, 36 },
		{ 320, 169, 36 },
		{ 320, 170, 36 },
		{ 320, 171, 36 },
		{ 320, 172, 36 },
		{ 320, 173, 36 },
		{ 320, 174, 36 },
		{ 320, 175, 36 },
		{ 320, 176, 36 },
		{ 320, 177, 36 },
		{ 320, 178, 36 },
		{ 320, 179, 36 },
		{ 320, 180, 36 },
		{ 320, 181, 36 },
		{ 320, 182, 36 },
		{ 320, 183, 36 },
		{ 320, 184, 36 },
		{ 321, 155, 36 },
		{ 321, 156, 36 },
		{ 321, 157, 36 },
		{ 321, 158, 36 },
		{ 321, 159, 36 },
		{ 321, 160, 36 },
		{ 321, 161, 36 },
		{ 321, 162, 36 },
		{ 321, 163, 36 },
		{ 321, 164, 36 },
		{ 321, 165, 36 },
		{ 321, 166, 36 },
		{ 321, 167, 36 },
		{ 321, 168, 36 },
		{ 321, 169, 36 },
		{ 321, 170, 36 },
		{ 321, 171, 36 },
		{ 321, 172, 36 },
		{ 321, 173, 36 },
		{ 321, 174, 36 },
		{ 321, 175, 36 },
		{ 321, 176, 36 },
		{ 321, 177, 36 },
		{ 321, 178, 36 },
		{ 321, 179, 36 },
		{ 321, 180, 36 },
		{ 321, 181, 36 },
		{ 321, 182, 36 },
		{ 321, 183, 36 },
		{ 322, 154, 36 },
		{ 322, 155, 36 },
		{ 322, 156, 36 },
		{ 322, 157, 36 },
		{ 322, 158, 36 },
		{ 322, 159, 36 },
		{ 322, 160, 36 },
		{ 322, 161, 36 },
		{ 322, 162, 36 },
		{ 322, 163, 36 },
		{ 322, 164, 36 },
		{ 322, 165, 36 },
		{ 322, 166, 36 },
		{ 322, 167, 36 },
		{ 322, 168, 36 },
		{ 322, 169, 36 },
		{ 322, 170, 36 },
		{ 322, 171, 36 },
		{ 322, 172, 36 },
		{ 322, 173, 36 },
		{ 322, 174, 36 },
		{ 322, 175, 36 },
		{ 322, 176, 36 },
		{ 322, 177, 36 },
		{ 322, 178, 36 },
		{ 322, 179, 36 },
		{ 322, 180, 36 },
		{ 322, 181, 36 },
		{ 322, 182, 36 },
		{ 323, 152, 36 },
		{ 323, 153, 36 },
		{ 323, 154, 36 },
		{ 323, 155, 36 },
		{ 323, 156, 36 },
		{ 323, 157, 36 },
		{ 323, 158, 36 },
		{ 323, 159, 36 },
		{ 323, 160, 36 },
		{ 323, 161, 36 },
		{ 323, 162, 36 },
		{ 323, 163, 36 },
		{ 323, 164, 36 },
		{ 323, 165, 36 },
		{ 323, 166, 36 },
		{ 323, 167, 36 },
		{ 323, 168, 36 },
		{ 323, 169, 36 },
		{ 323, 170, 36 },
		{ 323, 171, 36 },
		{ 323, 172, 36 },
		{ 323, 173, 36 },
		{ 323, 174, 36 },
		{ 323, 175, 36 },
		{ 323, 176, 36 },
		{ 323, 177, 36 },
		{ 323, 178, 36 },
		{ 323, 179, 36 },
		{ 323, 180, 36 },
		{ 323, 181, 36 },
		{ 324, 150, 36 },
		{ 324, 151, 36 },
		{ 324, 152, 36 },
		{ 324, 153, 36 },
		{ 324, 154, 36 },
		{ 324, 155, 36 },
		{ 324, 156, 36 },
		{ 324, 157, 36 },
		{ 324, 158, 36 },
		{ 324, 159, 36 },
		{ 324, 160, 36 },
		{ 324, 161, 36 },
		{ 324, 162, 36 },
		{ 324, 163, 36 },
		{ 324, 164, 36 },
		{ 324, 165, 36 },
		{ 324, 166, 36 },
		{ 324, 167, 36 },
		{ 324, 168, 36 },
		{ 324, 169, 36 },
		{ 324, 170, 36 },
		{ 324, 171, 36 },
		{ 324, 172, 36 },
		{ 324, 173, 36 },
		{ 324, 174, 36 },
		{ 324, 175, 36 },
		{ 324, 176, 36 },
		{ 324, 177, 36 },
		{ 324, 178, 36 },
		{ 324, 179, 36 },
		{ 324, 180, 36 },
		{ 325, 149, 36 },
		{ 325, 150, 36 },
		{ 325, 151, 36 },
		{ 325, 152, 36 },
		{ 325, 153, 36 },
		{ 325, 154, 36 },
		{ 325, 155, 36 },
		{ 325, 156, 36 },
		{ 325, 157, 36 },
		{ 325, 158, 36 },
		{ 325, 159, 36 },
		{ 325, 160, 36 },
		{ 325, 161, 36 },
		{ 325, 162, 36 },
		{ 325, 163, 36 },
		{ 325, 164, 36 },
		{ 325, 165, 36 },
		{ 325, 166, 36 },
		{ 325, 167, 36 },
		{ 325, 168, 36 },
		{ 325, 169, 36 },
		{ 325, 170, 36 },
		{ 325, 171, 36 },
		{ 325, 172, 36 },
		{ 325, 173, 36 },
		{ 325, 174, 36 },
		{ 325, 175, 36 },
		{ 325, 176, 36 },
		{ 325, 177, 36 },
		{ 325, 178, 36 },
		{ 325, 179, 36 },
		{ 326, 147, 36 },
		{ 326, 148, 36 },
		{ 326, 149, 36 },
		{ 326, 150, 36 },
		{ 326, 151, 36 },
		{ 326, 152, 36 },
		{ 326, 153, 36 },
		{ 326, 154, 36 },
		{ 326, 155, 36 },
		{ 326, 156, 36 },
		{ 326, 157, 36 },
		{ 326, 158, 36 },
		{ 326, 159, 36 },
		{ 326, 160, 36 },
		{ 326, 161, 36 },
		{ 326, 162, 36 },
		{ 326, 163, 36 },
		{ 326, 164, 36 },
		{ 326, 165, 36 },
		{ 326, 166, 36 },
		{ 326, 167, 36 },
		{ 326, 168, 36 },
		{ 326, 169, 36 },
		{ 326, 170, 36 },
		{ 326, 171, 36 },
		{ 326, 172, 36 },
		{ 326, 173, 36 },
		{ 326, 174, 36 },
		{ 326, 175, 36 },
		{ 326, 176, 36 },
		{ 326, 177, 36 },
		{ 326, 178, 36 },
		{ 327, 146, 36 },
		{ 327, 147, 36 },
		{ 327, 148, 36 },
		{ 327, 149, 36 },
		{ 327, 150, 36 },
		{ 327, 151, 36 },
		{ 327, 152, 36 },
		{ 327, 153, 36 },
		{ 327, 154, 36 },
		{ 327, 155, 36 },
		{ 327, 156, 36 },
		{ 327, 157, 36 },
		{ 327, 158, 36 },
		{ 327, 159, 36 },
		{ 327, 160, 36 },
		{ 327, 161, 36 },
		{ 327, 162, 36 },
		{ 327, 163, 36 },
		{ 327, 164, 36 },
		{ 327, 165, 36 },
		{ 327, 166, 36 },
		{ 327, 167, 36 },
		{ 327, 168, 36 },
		{ 327, 169, 36 },
		{ 327, 170, 36 },
		{ 327, 171, 36 },
		{ 327, 172, 36 },
		{ 327, 173, 36 },
		{ 327, 174, 36 },
		{ 327, 175, 36 },
		{ 327, 176, 36 },
		{ 328, 144, 36 },
		{ 328, 145, 36 },
		{ 328, 146, 36 },
		{ 328, 147, 36 },
		{ 328, 148, 36 },
		{ 328, 149, 36 },
		{ 328, 150, 36 },
		{ 328, 151, 36 },
		{ 328, 152, 36 },
		{ 328, 153, 36 },
		{ 328, 154, 36 },
		{ 328, 155, 36 },
		{ 328, 156, 36 },
		{ 328, 157, 36 },
		{ 328, 158, 36 },
		{ 328, 159, 36 },
		{ 328, 160, 36 },
		{ 328, 161, 36 },
		{ 328, 162, 36 },
		{ 328, 163, 36 },
		{ 328, 164, 36 },
		{ 328, 165, 36 },
		{ 328, 166, 36 },
		{ 328, 167, 36 },
		{ 328, 168, 36 },
		{ 328, 169, 36 },
		{ 328, 170, 36 },
		{ 328, 171, 36 },
		{ 328, 172, 36 },
		{ 328, 173, 36 },
		{ 328, 174, 36 },
		{ 328, 175, 36 },
		{ 329, 142, 36 },
		{ 329, 143, 36 },
		{ 329, 144, 36 },
		{ 329, 145, 36 },
		{ 329, 146, 36 },
		{ 329, 147, 36 },
		{ 329, 148, 36 },
		{ 329, 149, 36 },
		{ 329, 150, 36 },
		{ 329, 151, 36 },
		{ 329, 152, 36 },
		{ 329, 153, 36 },
		{ 329, 154, 36 },
		{ 329, 155, 36 },
		{ 329, 156, 36 },
		{ 329, 157, 36 },
		{ 329, 158, 36 },
		{ 329, 159, 36 },
		{ 329, 160, 36 },
		{ 329, 161, 36 },
		{ 329, 162, 36 },
		{ 329, 163, 36 },
		{ 329, 164, 36 },
		{ 329, 165, 36 },
		{ 329, 166, 36 },
		{ 329, 167, 36 },
		{ 329, 168, 36 },
		{ 329, 169, 36 },
		{ 329, 170, 36 },
		{ 329, 171, 36 },
		{ 329, 172, 36 },
		{ 329, 173, 36 },
		{ 329, 174, 36 },
		{ 330, 141, 36 },
		{ 330, 142, 36 },
		{ 330, 143, 36 },
		{ 330, 144, 36 },
		{ 330, 145, 36 },
		{ 330, 146, 36 },
		{ 330, 147, 36 },
		{ 330, 148, 36 },
		{ 330, 149, 36 },
		{ 330, 150, 36 },
		{ 330, 151, 36 },
		{ 330, 152, 36 },
		{ 330, 153, 36 },
		{ 330, 154, 36 },
		{ 330, 155, 36 },
		{ 330, 156, 36 },
		{ 330, 157, 36 },
		{ 330, 158, 36 },
		{ 330, 159, 36 },
		{ 330, 160, 36 },
		{ 330, 161, 36 },
		{ 330, 162, 36 },
		{ 330, 163, 36 },
		{ 330, 164, 36 },
		{ 330, 165, 36 },
		{ 330, 166, 36 },
		{ 330, 167, 36 },
		{ 330, 168, 36 },
		{ 330, 169, 36 },
		{ 330, 170, 36 },
		{ 330, 171, 36 },
		{ 330, 172, 36 },
		{ 330, 173, 36 },
		{ 331, 139, 36 },
		{ 331, 140, 36 },
		{ 331, 141, 36 },
		{ 331, 142, 36 },
		{ 331, 143, 36 },
		{ 331, 144, 36 },
		{ 331, 145, 36 },
		{ 331, 146, 36 },
		{ 331, 147, 36 },
		{ 331, 148, 36 },
		{ 331, 149, 36 },
		{ 331, 150, 36 },
		{ 331, 151, 36 },
		{ 331, 152, 36 },
		{ 331, 153, 36 },
		{ 331, 154, 36 },
		{ 331, 155, 36 },
		{ 331, 156, 36 },
		{ 331, 157, 36 },
		{ 331, 158, 36 },
		{ 331, 159, 36 },
		{ 331, 160, 36 },
		{ 331, 161, 36 },
		{ 331, 162, 36 },
		{ 331, 163, 36 },
		{ 331, 164, 36 },
		{ 331, 165, 36 },
		{ 331, 166, 36 },
		{ 331, 167, 36 },
		{ 331, 168, 36 },
		{ 331, 169, 36 },
		{ 331, 170, 36 },
		{ 331, 171, 36 },
		{ 331, 172, 36 },
		{ 332, 138, 36 },
		{ 332, 139, 36 },
		{ 332, 140, 36 },
		{ 332, 141, 36 },
		{ 332, 142, 36 },
		{ 332, 143, 36 },
		{ 332, 144, 36 },
		{ 332, 145, 36 },
		{ 332, 146, 36 },
		{ 332, 147, 36 },
		{ 332, 148, 36 },
		{ 332, 149, 36 },
		{ 332, 150, 36 },
		{ 332, 151, 36 },
		{ 332, 152, 36 },
		{ 332, 153, 36 },
		{ 332, 154, 36 },
		{ 332, 155, 36 },
		{ 332, 156, 36 },
		{ 332, 157, 36 },
		{ 332, 158, 36 },
		{ 332, 159, 36 },
		{ 332, 160, 36 },
		{ 332, 161, 36 },
		{ 332, 162, 36 },
		{ 332, 163, 36 },
		{ 332, 164, 36 },
		{ 332, 165, 36 },
		{ 332, 166, 36 },
		{ 332, 167, 36 },
		{ 332, 168, 36 },
		{ 332, 169, 36 },
		{ 332, 170, 36 },
		{ 332, 171, 36 },
		{ 333, 136, 36 },
		{ 333, 137, 36 },
		{ 333, 138, 36 },
		{ 333, 139, 36 },
		{ 333, 140, 36 },
		{ 333, 141, 36 },
		{ 333, 142, 36 },
		{ 333, 143, 36 },
		{ 333, 144, 36 },
		{ 333, 145, 36 },
		{ 333, 146, 36 },
		{ 333, 147, 36 },
		{ 333, 148, 36 },
		{ 333, 149, 36 },
		{ 333, 150, 36 },
		{ 333, 151, 36 },
		{ 333, 152, 36 },
		{ 333, 153, 36 },
		{ 333, 154, 36 },
		{ 333, 155, 36 },
		{ 333, 156, 36 },
		{ 333, 157, 36 },
		{ 333, 158, 36 },
		{ 333, 159, 36 },
		{ 333, 160, 36 },
		{ 333, 161, 36 },
		{ 333, 162, 36 },
		{ 333, 163, 36 },
		{ 333, 164, 36 },
		{ 333, 165, 36 },
		{ 333, 166, 36 },
		{ 333, 167, 36 },
		{ 333, 168, 36 },
		{ 333, 169, 36 },
		{ 333, 170, 36 },
		{ 334, 134, 36 },
		{ 334, 135, 36 },
		{ 334, 136, 36 },
		{ 334, 137, 36 },
		{ 334, 138, 36 },
		{ 334, 139, 36 },
		{ 334, 140, 36 },
		{ 334, 141, 36 },
		{ 334, 142, 36 },
		{ 334, 143, 36 },
		{ 334, 144, 36 },
		{ 334, 145, 36 },
		{ 334, 146, 36 },
		{ 334, 147, 36 },
		{ 334, 148, 36 },
		{ 334, 149, 36 },
		{ 334, 150, 36 },
		{ 334, 151, 36 },
		{ 334, 152, 36 },
		{ 334, 153, 36 },
		{ 334, 154, 36 },
		{ 334, 155, 36 },
		{ 334, 156, 36 },
		{ 334, 157, 36 },
		{ 334, 158, 36 },
		{ 334, 159, 36 },
		{ 334, 160, 36 },
		{ 334, 161, 36 },
		{ 334, 162, 36 },
		{ 334, 163, 36 },
		{ 334, 164, 36 },
		{ 334, 165, 36 },
		{ 334, 166, 36 },
		{ 334, 167, 36 },
		{ 334, 168, 36 },
		{ 335, 133, 36 },
		{ 335, 134, 36 },
		{ 335, 135, 36 },
		{ 335, 136, 36 },
		{ 335, 137, 36 },
		{ 335, 138, 36 },
		{ 335, 139, 36 },
		{ 335, 140, 36 },
		{ 335, 141, 36 },
		{ 335, 142, 36 },
		{ 335, 143, 36 },
		{ 335, 144, 36 },
		{ 335, 145, 36 },
		{ 335, 146, 36 },
		{ 335, 147, 36 },
		{ 335, 148, 36 },
		{ 335, 149, 36 },
		{ 335, 150, 36 },
		{ 335, 151, 36 },
		{ 335, 152, 36 },
		{ 335, 153, 36 },
		{ 335, 154, 36 },
		{ 335, 155, 36 },
		{ 335, 156, 36 },
		{ 335, 157, 36 },
		{ 335, 158, 36 },
		{ 335, 159, 36 },
		{ 335, 160, 36 },
		{ 335, 161, 36 },
		{ 335, 162, 36 },
		{ 335, 163, 36 },
		{ 335, 164, 36 },
		{ 335, 165, 36 },
		{ 335, 166, 36 },
		{ 335, 167, 36 },
		{ 336, 131, 36 },
		{ 336, 132, 36 },
		{ 336, 133, 36 },
		{ 336, 134, 36 },
		{ 336, 135, 36 },
		{ 336, 136, 36 },
		{ 336, 137, 36 },
		{ 336, 138, 36 },
		{ 336, 139, 36 },
		{ 336, 140, 36 },
		{ 336, 141, 36 },
		{ 336, 142, 36 },
		{ 336, 143, 36 },
		{ 336, 144, 36 },
		{ 336, 145, 36 },
		{ 336, 146, 36 },
		{ 336, 147, 36 },
		{ 336, 148, 36 },
		{ 336, 149, 36 },
		{ 336, 150, 36 },
		{ 336, 151, 36 },
		{ 336, 152, 36 },
		{ 336, 153, 36 },
		{ 336, 154, 36 },
		{ 336, 155, 36 },
		{ 336, 156, 36 },
		{ 336, 157, 36 },
		{ 336, 158, 36 },
		{ 336, 159, 36 },
		{ 336, 160, 36 },
		{ 336, 161, 36 },
		{ 336, 162, 36 },
		{ 336, 163, 36 },
		{ 336, 164, 36 },
		{ 336, 165, 36 },
		{ 336, 166, 36 },
		{ 337, 130, 36 },
		{ 337, 131, 36 },
		{ 337, 132, 36 },
		{ 337, 133, 36 },
		{ 337, 134, 36 },
		{ 337, 135, 36 },
		{ 337, 136, 36 },
		{ 337, 137, 36 },
		{ 337, 138, 36 },
		{ 337, 139, 36 },
		{ 337, 140, 36 },
		{ 337, 141, 36 },
		{ 337, 142, 36 },
		{ 337, 143, 36 },
		{ 337, 144, 36 },
		{ 337, 145, 36 },
		{ 337, 146, 36 },
		{ 337, 147, 36 },
		{ 337, 148, 36 },
		{ 337, 149, 36 },
		{ 337, 150, 36 },
		{ 337, 151, 36 },
		{ 337, 152, 36 },
		{ 337, 153, 36 },
		{ 337, 154, 36 },
		{ 337, 155, 36 },
		{ 337, 156, 36 },
		{ 337, 157, 36 },
		{ 337, 158, 36 },
		{ 337, 159, 36 },
		{ 337, 160, 36 },
		{ 337, 161, 36 },
		{ 337, 162, 36 },
		{ 337, 163, 36 },
		{ 337, 164, 36 },
		{ 337, 165, 36 },
		{ 338, 128, 36 },
		{ 338, 129, 36 },
		{ 338, 130, 36 },
		{ 338, 131, 36 },
		{ 338, 132, 36 },
		{ 338, 133, 36 },
		{ 338, 134, 36 },
		{ 338, 135, 36 },
		{ 338, 136, 36 },
		{ 338, 137, 36 },
		{ 338, 138, 36 },
		{ 338, 139, 36 },
		{ 338, 140, 36 },
		{ 338, 141, 36 },
		{ 338, 142, 36 },
		{ 338, 143, 36 },
		{ 338, 144, 36 },
		{ 338, 145, 36 },
		{ 338, 146, 36 },
		{ 338, 147, 36 },
		{ 338, 148, 36 },
		{ 338, 149, 36 },
		{ 338, 150, 36 },
		{ 338, 151, 36 },
		{ 338, 152, 36 },
		{ 338, 153, 36 },
		{ 338, 154, 36 },
		{ 338, 155, 36 },
		{ 338, 156, 36 },
		{ 338, 157, 36 },
		{ 338, 158, 36 },
		{ 338, 159, 36 },
		{ 338, 160, 36 },
		{ 338, 161, 36 },
		{ 338, 162, 36 },
		{ 338, 163, 36 },
		{ 338, 164, 36 },
		{ 339, 126, 36 },
		{ 339, 127, 36 },
		{ 339, 128, 36 },
		{ 339, 129, 36 },
		{ 339, 130, 36 },
		{ 339, 131, 36 },
		{ 339, 132, 36 },
		{ 339, 133, 36 },
		{ 339, 134, 36 },
		{ 339, 135, 36 },
		{ 339, 136, 36 },
		{ 339, 137, 36 },
		{ 339, 138, 36 },
		{ 339, 139, 36 },
		{ 339, 140, 36 },
		{ 339, 141, 36 },
		{ 339, 142, 36 },
		{ 339, 143, 36 },
		{ 339, 144, 36 },
		{ 339, 145, 36 },
		{ 339, 146, 36 },
		{ 339, 147, 36 },
		{ 339, 148, 36 },
		{ 339, 149, 36 },
		{ 339, 150, 36 },
		{ 339, 151, 36 },
		{ 339, 152, 36 },
		{ 339, 153, 36 },
		{ 339, 154, 36 },
		{ 339, 155, 36 },
		{ 339, 156, 36 },
		{ 339, 157, 36 },
		{ 339, 158, 36 },
		{ 339, 159, 36 },
		{ 339, 160, 36 },
		{ 339, 161, 36 },
		{ 339, 162, 36 },
		{ 339, 163, 36 },
		{ 340, 125, 36 },
		{ 340, 126, 36 },
		{ 340, 127, 36 },
		{ 340, 128, 36 },
		{ 340, 129, 36 },
		{ 340, 130, 36 },
		{ 340, 131, 36 },
		{ 340, 132, 36 },
		{ 340, 133, 36 },
		{ 340, 134, 36 },
		{ 340, 135, 36 },
		{ 340, 136, 36 },
		{ 340, 137, 36 },
		{ 340, 138, 36 },
		{ 340, 139, 36 },
		{ 340, 140, 36 },
		{ 340, 141, 36 },
		{ 340, 142, 36 },
		{ 340, 143, 36 },
		{ 340, 144, 36 },
		{ 340, 145, 36 },
		{ 340, 146, 36 },
		{ 340, 147, 36 },
		{ 340, 148, 36 },
		{ 340, 149, 36 },
		{ 340, 150, 36 },
		{ 340, 151, 36 },
		{ 340, 152, 36 },
		{ 340, 153, 36 },
		{ 340, 154, 36 },
		{ 340, 155, 36 },
		{ 340, 156, 36 },
		{ 340, 157, 36 },
		{ 340, 158, 36 },
		{ 340, 159, 36 },
		{ 340, 160, 36 },
		{ 340, 161, 36 },
		{ 340, 162, 36 },
		{ 341, 123, 36 },
		{ 341, 124, 36 },
		{ 341, 125, 36 },
		{ 341, 126, 36 },
		{ 341, 127, 36 },
		{ 341, 128, 36 },
		{ 341, 129, 36 },
		{ 341, 130, 36 },
		{ 341, 131, 36 },
		{ 341, 132, 36 },
		{ 341, 133, 36 },
		{ 341, 134, 36 },
		{ 341, 135, 36 },
		{ 341, 136, 36 },
		{ 341, 137, 36 },
		{ 341, 138, 36 },
		{ 341, 139, 36 },
		{ 341, 140, 36 },
		{ 341, 141, 36 },
		{ 341, 142, 36 },
		{ 341, 143, 36 },
		{ 341, 144, 36 },
		{ 341, 145, 36 },
		{ 341, 146, 36 },
		{ 341, 147, 36 },
		{ 341, 148, 36 },
		{ 341, 149, 36 },
		{ 341, 150, 36 },
		{ 341, 151, 36 },
		{ 341, 152, 36 },
		{ 341, 153, 36 },
		{ 341, 154, 36 },
		{ 341, 155, 36 },
		{ 341, 156, 36 },
		{ 341, 157, 36 },
		{ 341, 158, 36 },
		{ 341, 159, 36 },
		{ 341, 160, 36 },
		{ 342, 122, 36 },
		{ 342, 123, 36 },
		{ 342, 124, 36 },
		{ 342, 125, 36 },
		{ 342, 126, 36 },
		{ 342, 127, 36 },
		{ 342, 128, 36 },
		{ 342, 129, 36 },
		{ 342, 130, 36 },
		{ 342, 131, 36 },
		{ 342, 132, 36 },
		{ 342, 133, 36 },
		{ 342, 134, 36 },
		{ 342, 135, 36 },
		{ 342, 136, 36 },
		{ 342, 137, 36 },
		{ 342, 138, 36 },
		{ 342, 139, 36 },
		{ 342, 140, 36 },
		{ 342, 141, 36 },
		{ 342, 142, 36 },
		{ 342, 143, 36 },
		{ 342, 144, 36 },
		{ 342, 145, 36 },
		{ 342, 146, 36 },
		{ 342, 147, 36 },
		{ 342, 148, 36 },
		{ 342, 149, 36 },
		{ 342, 150, 36 },
		{ 342, 151, 36 },
		{ 342, 152, 36 },
		{ 342, 153, 36 },
		{ 342, 154, 36 },
		{ 342, 155, 36 },
		{ 342, 156, 36 },
		{ 342, 157, 36 },
		{ 342, 158, 36 },
		{ 342, 159, 36 },
		{ 343, 120, 36 },
		{ 343, 121, 36 },
		{ 343, 122, 36 },
		{ 343, 123, 36 },
		{ 343, 124, 36 },
		{ 343, 125, 36 },
		{ 343, 126, 36 },
		{ 343, 127, 36 },
		{ 343, 128, 36 },
		{ 343, 129, 36 },
		{ 343, 130, 36 },
		{ 343, 131, 36 },
		{ 343, 132, 36 },
		{ 343, 133, 36 },
		{ 343, 134, 36 },
		{ 343, 135, 36 },
		{ 343, 136, 36 },
		{ 343, 137, 36 },
		{ 343, 138, 36 },
		{ 343, 139, 36 },
		{ 343, 140, 36 },
		{ 343, 141, 36 },
		{ 343, 142, 36 },
		{ 343, 143, 36 },
		{ 343, 144, 36 },
		{ 343, 145, 36 },
		{ 343, 146, 36 },
		{ 343, 147, 36 },
		{ 343, 148, 36 },
		{ 343, 149, 36 },
		{ 343, 150, 36 },
		{ 343, 151, 36 },
		{ 343, 152, 36 },
		{ 343, 153, 36 },
		{ 343, 154, 36 },
		{ 343, 155, 36 },
		{ 343, 156, 36 },
		{ 343, 157, 36 },
		{ 343, 158, 36 },
		{ 344, 118, 36 },
		{ 344, 119, 36 },
		{ 344, 120, 36 },
		{ 344, 121, 36 },
		{ 344, 122, 36 },
		{ 344, 123, 36 },
		{ 344, 124, 36 },
		{ 344, 125, 36 },
		{ 344, 126, 36 },
		{ 344, 127, 36 },
		{ 344, 128, 36 },
		{ 344, 129, 36 },
		{ 344, 130, 36 },
		{ 344, 131, 36 },
		{ 344, 132, 36 },
		{ 344, 133, 36 },
		{ 344, 134, 36 },
		{ 344, 135, 36 },
		{ 344, 136, 36 },
		{ 344, 137, 36 },
		{ 344, 138, 36 },
		{ 344, 139, 36 },
		{ 344, 140, 36 },
		{ 344, 141, 36 },
		{ 344, 142, 36 },
		{ 344, 143, 36 },
		{ 344, 144, 36 },
		{ 344, 145, 36 },
		{ 344, 146, 36 },
		{ 344, 147, 36 },
		{ 344, 148, 36 },
		{ 344, 149, 36 },
		{ 344, 150, 36 },
		{ 344, 151, 36 },
		{ 344, 152, 36 },
		{ 344, 153, 36 },
		{ 344, 154, 36 },
		{ 344, 155, 36 },
		{ 344, 156, 36 },
		{ 344, 157, 36 },
		{ 345, 117, 36 },
		{ 345, 118, 36 },
		{ 345, 119, 36 },
		{ 345, 120, 36 },
		{ 345, 121, 36 },
		{ 345, 122, 36 },
		{ 345, 123, 36 },
		{ 345, 124, 36 },
		{ 345, 125, 36 },
		{ 345, 126, 36 },
		{ 345, 127, 36 },
		{ 345, 128, 36 },
		{ 345, 129, 36 },
		{ 345, 130, 36 },
		{ 345, 131, 36 },
		{ 345, 132, 36 },
		{ 345, 133, 36 },
		{ 345, 134, 36 },
		{ 345, 135, 36 },
		{ 345, 136, 36 },
		{ 345, 137, 36 },
		{ 345, 138, 36 },
		{ 345, 139, 36 },
		{ 345, 140, 36 },
		{ 345, 141, 36 },
		{ 345, 142, 36 },
		{ 345, 143, 36 },
		{ 345, 144, 36 },
		{ 345, 145, 36 },
		{ 345, 146, 36 },
		{ 345, 147, 36 },
		{ 345, 148, 36 },
		{ 345, 149, 36 },
		{ 345, 150, 36 },
		{ 345, 151, 36 },
		{ 345, 152, 36 },
		{ 345, 153, 36 },
		{ 345, 154, 36 },
		{ 345, 155, 36 },
		{ 345, 156, 36 },
		{ 346, 115, 36 },
		{ 346, 116, 36 },
		{ 346, 117, 36 },
		{ 346, 118, 36 },
		{ 346, 119, 36 },
		{ 346, 120, 36 },
		{ 346, 121, 36 },
		{ 346, 122, 36 },
		{ 346, 123, 36 },
		{ 346, 124, 36 },
		{ 346, 125, 36 },
		{ 346, 126, 36 },
		{ 346, 127, 36 },
		{ 346, 128, 36 },
		{ 346, 129, 36 },
		{ 346, 130, 36 },
		{ 346, 131, 36 },
		{ 346, 132, 36 },
		{ 346, 133, 36 },
		{ 346, 134, 36 },
		{ 346, 135, 36 },
		{ 346, 136, 36 },
		{ 346, 137, 36 },
		{ 346, 138, 36 },
		{ 346, 139, 36 },
		{ 346, 140, 36 },
		{ 346, 141, 36 },
		{ 346, 142, 36 },
		{ 346, 143, 36 },
		{ 346, 144, 36 },
		{ 346, 145, 36 },
		{ 346, 146, 36 },
		{ 346, 147, 36 },
		{ 346, 148, 36 },
		{ 346, 149, 36 },
		{ 346, 150, 36 },
		{ 346, 151, 36 },
		{ 346, 152, 36 },
		{ 346, 153, 36 },
		{ 346, 154, 36 },
		{ 346, 155, 36 },
		{ 347, 114, 36 },
		{ 347, 115, 36 },
		{ 347, 116, 36 },
		{ 347, 117, 36 },
		{ 347, 118, 36 },
		{ 347, 119, 36 },
		{ 347, 120, 36 },
		{ 347, 121, 36 },
		{ 347, 122, 36 },
		{ 347, 123, 36 },
		{ 347, 124, 36 },
		{ 347, 125, 36 },
		{ 347, 126, 36 },
		{ 347, 127, 36 },
		{ 347, 128, 36 },
		{ 347, 129, 36 },
		{ 347, 130, 36 },
		{ 347, 131, 36 },
		{ 347, 132, 36 },
		{ 347, 133, 36 },
		{ 347, 134, 36 },
		{ 347, 135, 36 },
		{ 347, 136, 36 },
		{ 347, 137, 36 },
		{ 347, 138, 36 },
		{ 347, 139, 36 },
		{ 347, 140, 36 },
		{ 347, 141, 36 },
		{ 347, 142, 36 },
		{ 347, 143, 36 },
		{ 347, 144, 36 },
		{ 347, 145, 36 },
		{ 347, 146, 36 },
		{ 347, 147, 36 },
		{ 347, 148, 36 },
		{ 347, 149, 36 },
		{ 347, 150, 36 },
		{ 347, 151, 36 },
		{ 347, 152, 36 },
		{ 347, 153, 36 },
		{ 347, 154, 36 },
		{ 348, 112, 36 },
		{ 348, 113, 36 },
		{ 348, 114, 36 },
		{ 348, 115, 36 },
		{ 348, 116, 36 },
		{ 348, 117, 36 },
		{ 348, 118, 36 },
		{ 348, 119, 36 },
		{ 348, 120, 36 },
		{ 348, 121, 36 },
		{ 348, 122, 36 },
		{ 348, 123, 36 },
		{ 348, 124, 36 },
		{ 348, 125, 36 },
		{ 348, 126, 36 },
		{ 348, 127, 36 },
		{ 348, 128, 36 },
		{ 348, 129, 36 },
		{ 348, 130, 36 },
		{ 348, 131, 36 },
		{ 348, 132, 36 },
		{ 348, 133, 36 },
		{ 348, 134, 36 },
		{ 348, 135, 36 },
		{ 348, 136, 36 },
		{ 348, 137, 36 },
		{ 348, 138, 36 },
		{ 348, 139, 36 },
		{ 348, 140, 36 },
		{ 348, 141, 36 },
		{ 348, 142, 36 },
		{ 348, 143, 36 },
		{ 348, 144, 36 },
		{ 348, 145, 36 },
		{ 348, 146, 36 },
		{ 348, 147, 36 },
		{ 348, 148, 36 },
		{ 348, 149, 36 },
		{ 348, 150, 36 },
		{ 348, 151, 36 },
		{ 348, 152, 36 },
		{ 349, 110, 36 },
		{ 349, 111, 36 },
		{ 349, 112, 36 },
		{ 349, 113, 36 },
		{ 349, 114, 36 },
		{ 349, 115, 36 },
		{ 349, 116, 36 },
		{ 349, 117, 36 },
		{ 349, 118, 36 },
		{ 349, 119, 36 },
		{ 349, 120, 36 },
		{ 349, 121, 36 },
		{ 349, 122, 36 },
		{ 349, 123, 36 },
		{ 349, 124, 36 },
		{ 349, 125, 36 },
		{ 349, 126, 36 },
		{ 349, 127, 36 },
		{ 349, 128, 36 },
		{ 349, 129, 36 },
		{ 349, 130, 36 },
		{ 349, 131, 36 },
		{ 349, 132, 36 },
		{ 349, 133, 36 },
		{ 349, 134, 36 },
		{ 349, 135, 36 },
		{ 349, 136, 36 },
		{ 349, 137, 36 },
		{ 349, 138, 36 },
		{ 349, 139, 36 },
		{ 349, 140, 36 },
		{ 349, 141, 36 },
		{ 349, 142, 36 },
		{ 349, 143, 36 },
		{ 349, 144, 36 },
		{ 349, 145, 36 },
		{ 349, 146, 36 },
		{ 349, 147, 36 },
		{ 349, 148, 36 },
		{ 349, 149, 36 },
		{ 349, 150, 36 },
		{ 349, 151, 36 },
		{ 350, 109, 36 },
		{ 350, 110, 36 },
		{ 350, 111, 36 },
		{ 350, 112, 36 },
		{ 350, 113, 36 },
		{ 350, 114, 36 },
		{ 350, 115, 36 },
		{ 350, 116, 36 },
		{ 350, 117, 36 },
		{ 350, 118, 36 },
		{ 350, 119, 36 },
		{ 350, 120, 36 },
		{ 350, 121, 36 },
		{ 350, 122, 36 },
		{ 350, 123, 36 },
		{ 350, 124, 36 },
		{ 350, 125, 36 },
		{ 350, 126, 36 },
		{ 350, 127, 36 },
		{ 350, 128, 36 },
		{ 350, 129, 36 },
		{ 350, 130, 36 },
		{ 350, 131, 36 },
		{ 350, 132, 36 },
		{ 350, 133, 36 },
		{ 350, 134, 36 },
		{ 350, 135, 36 },
		{ 350, 136, 36 },
		{ 350, 137, 36 },
		{ 350, 138, 36 },
		{ 350, 139, 36 },
		{ 350, 140, 36 },
		{ 350, 141, 36 },
		{ 350, 142, 36 },
		{ 350, 143, 36 },
		{ 350, 144, 36 },
		{ 350, 145, 36 },
		{ 350, 146, 36 },
		{ 350, 147, 36 },
		{ 350, 148, 36 },
		{ 350, 149, 36 },
		{ 350, 150, 36 },
		{ 351, 107, 36 },
		{ 351, 108, 36 },
		{ 351, 109, 36 },
		{ 351, 110, 36 },
		{ 351, 111, 36 },
		{ 351, 112, 36 },
		{ 351, 113, 36 },
		{ 351, 114, 36 },
		{ 351, 115, 36 },
		{ 351, 116, 36 },
		{ 351, 117, 36 },
		{ 351, 118, 36 },
		{ 351, 119, 36 },
		{ 351, 120, 36 },
		{ 351, 121, 36 },
		{ 351, 122, 36 },
		{ 351, 123, 36 },
		{ 351, 124, 36 },
		{ 351, 125, 36 },
		{ 351, 126, 36 },
		{ 351, 127, 36 },
		{ 351, 128, 36 },
		{ 351, 129, 36 },
		{ 351, 130, 36 },
		{ 351, 131, 36 },
		{ 351, 132, 36 },
		{ 351, 133, 36 },
		{ 351, 134, 36 },
		{ 351, 135, 36 },
		{ 351, 136, 36 },
		{ 351, 137, 36 },
		{ 351, 138, 36 },
		{ 351, 139, 36 },
		{ 351, 140, 36 },
		{ 351, 141, 36 },
		{ 351, 142, 36 },
		{ 351, 143, 36 },
		{ 351, 144, 36 },
		{ 351, 145, 36 },
		{ 351, 146, 36 },
		{ 351, 147, 36 },
		{ 351, 148, 36 },
		{ 351, 149, 36 },
		{ 352, 106, 36 },
		{ 352, 107, 36 },
		{ 352, 108, 36 },
		{ 352, 109, 36 },
		{ 352, 110, 36 },
		{ 352, 111, 36 },
		{ 352, 112, 36 },
		{ 352, 113, 36 },
		{ 352, 114, 36 },
		{ 352, 115, 36 },
		{ 352, 116, 36 },
		{ 352, 117, 36 },
		{ 352, 118, 36 },
		{ 352, 119, 36 },
		{ 352, 120, 36 },
		{ 352, 121, 36 },
		{ 352, 122, 36 },
		{ 352, 123, 36 },
		{ 352, 124, 36 },
		{ 352, 125, 36 },
		{ 352, 126, 36 },
		{ 352, 127, 36 },
		{ 352, 128, 36 },
		{ 352, 129, 36 },
		{ 352, 130, 36 },
		{ 352, 131, 36 },
		{ 352, 132, 36 },
		{ 352, 133, 36 },
		{ 352, 134, 36 },
		{ 352, 135, 36 },
		{ 352, 136, 36 },
		{ 352, 137, 36 },
		{ 352, 138, 36 },
		{ 352, 139, 36 },
		{ 352, 140, 36 },
		{ 352, 141, 36 },
		{ 352, 142, 36 },
		{ 352, 143, 36 },
		{ 352, 144, 36 },
		{ 352, 145, 36 },
		{ 352, 146, 36 },
		{ 352, 147, 36 },
		{ 352, 148, 36 },
		{ 353, 104, 36 },
		{ 353, 105, 36 },
		{ 353, 106, 36 },
		{ 353, 107, 36 },
		{ 353, 108, 36 },
		{ 353, 109, 36 },
		{ 353, 110, 36 },
		{ 353, 111, 36 },
		{ 353, 112, 36 },
		{ 353, 113, 36 },
		{ 353, 114, 36 },
		{ 353, 115, 36 },
		{ 353, 116, 36 },
		{ 353, 117, 36 },
		{ 353, 118, 36 },
		{ 353, 119, 36 },
		{ 353, 120, 36 },
		{ 353, 121, 36 },
		{ 353, 122, 36 },
		{ 353, 123, 36 },
		{ 353, 124, 36 },
		{ 353, 125, 36 },
		{ 353, 126, 36 },
		{ 353, 127, 36 },
		{ 353, 128, 36 },
		{ 353, 129, 36 },
		{ 353, 130, 36 },
		{ 353, 131, 36 },
		{ 353, 132, 36 },
		{ 353, 133, 36 },
		{ 353, 134, 36 },
		{ 353, 135, 36 },
		{ 353, 136, 36 },
		{ 353, 137, 36 },
		{ 353, 138, 36 },
		{ 353, 139, 36 },
		{ 353, 140, 36 },
		{ 353, 141, 36 },
		{ 353, 142, 36 },
		{ 353, 143, 36 },
		{ 353, 144, 36 },
		{ 353, 145, 36 },
		{ 353, 146, 36 },
		{ 353, 147, 36 },
		{ 354, 102, 36 },
		{ 354, 103, 36 },
		{ 354, 104, 36 },
		{ 354, 105, 36 },
		{ 354, 106, 36 },
		{ 354, 107, 36 },
		{ 354, 108, 36 },
		{ 354, 109, 36 },
		{ 354, 110, 36 },
		{ 354, 111, 36 },
		{ 354, 112, 36 },
		{ 354, 113, 36 },
		{ 354, 114, 36 },
		{ 354, 115, 36 },
		{ 354, 116, 36 },
		{ 354, 117, 36 },
		{ 354, 118, 36 },
		{ 354, 119, 36 },
		{ 354, 120, 36 },
		{ 354, 121, 36 },
		{ 354, 122, 36 },
		{ 354, 123, 36 },
		{ 354, 124, 36 },
		{ 354, 125, 36 },
		{ 354, 126, 36 },
		{ 354, 127, 36 },
		{ 354, 128, 36 },
		{ 354, 129, 36 },
		{ 354, 130, 36 },
		{ 354, 131, 36 },
		{ 354, 132, 36 },
		{ 354, 133, 36 },
		{ 354, 134, 36 },
		{ 354, 135, 36 },
		{ 354, 136, 36 },
		{ 354, 137, 36 },
		{ 354, 138, 36 },
		{ 354, 139, 36 },
		{ 354, 140, 36 },
		{ 354, 141, 36 },
		{ 354, 142, 36 },
		{ 354, 143, 36 },
		{ 354, 144, 36 },
		{ 354, 145, 36 },
		{ 354, 146, 36 },
		{ 355, 101, 36 },
		{ 355, 102, 36 },
		{ 355, 103, 36 },
		{ 355, 104, 36 },
		{ 355, 105, 36 },
		{ 355, 106, 36 },
		{ 355, 107, 36 },
		{ 355, 108, 36 },
		{ 355, 109, 36 },
		{ 355, 110, 36 },
		{ 355, 111, 36 },
		{ 355, 112, 36 },
		{ 355, 113, 36 },
		{ 355, 114, 36 },
		{ 355, 115, 36 },
		{ 355, 116, 36 },
		{ 355, 117, 36 },
		{ 355, 118, 36 },
		{ 355, 119, 36 },
		{ 355, 120, 36 },
		{ 355, 121, 36 },
		{ 355, 122, 36 },
		{ 355, 123, 36 },
		{ 355, 124, 36 },
		{ 355, 125, 36 },
		{ 355, 126, 36 },
		{ 355, 127, 36 },
		{ 355, 128, 36 },
		{ 355, 129, 36 },
		{ 355, 130, 36 },
		{ 355, 131, 36 },
		{ 355, 132, 36 },
		{ 355, 133, 36 },
		{ 355, 134, 36 },
		{ 355, 135, 36 },
		{ 355, 136, 36 },
		{ 355, 137, 36 },
		{ 355, 138, 36 },
		{ 355, 139, 36 },
		{ 355, 140, 36 },
		{ 355, 141, 36 },
		{ 355, 142, 36 },
		{ 355, 143, 36 },
		{ 355, 144, 36 },
		{ 356, 99, 36 },
		{ 356, 100, 36 },
		{ 356, 101, 36 },
		{ 356, 102, 36 },
		{ 356, 103, 36 },
		{ 356, 104, 36 },
		{ 356, 105, 36 },
		{ 356, 106, 36 },
		{ 356, 107, 36 },
		{ 356, 108, 36 },
		{ 356, 109, 36 },
		{ 356, 110, 36 },
		{ 356, 111, 36 },
		{ 356, 112, 36 },
		{ 356, 113, 36 },
		{ 356, 114, 36 },
		{ 356, 115, 36 },
		{ 356, 116, 36 },
		{ 356, 117, 36 },
		{ 356, 118, 36 },
		{ 356, 119, 36 },
		{ 356, 120, 36 },
		{ 356, 121, 36 },
		{ 356, 122, 36 },
		{ 356, 123, 36 },
		{ 356, 124, 36 },
		{ 356, 125, 36 },
		{ 356, 126, 36 },
		{ 356, 127, 36 },
		{ 356, 128, 36 },
		{ 356, 129, 36 },
		{ 356, 130, 36 },
		{ 356, 131, 36 },
		{ 356, 132, 36 },
		{ 356, 133, 36 },
		{ 356, 134, 36 },
		{ 356, 135, 36 },
		{ 356, 136, 36 },
		{ 356, 137, 36 },
		{ 356, 138, 36 },
		{ 356, 139, 36 },
		{ 356, 140, 36 },
		{ 356, 141, 36 },
		{ 356, 142, 36 },
		{ 356, 143, 36 },
		{ 357, 98, 36 },
		{ 357, 99, 36 },
		{ 357, 100, 36 },
		{ 357, 101, 36 },
		{ 357, 102, 36 },
		{ 357, 103, 36 },
		{ 357, 104, 36 },
		{ 357, 105, 36 },
		{ 357, 106, 36 },
		{ 357, 107, 36 },
		{ 357, 108, 36 },
		{ 357, 109, 36 },
		{ 357, 110, 36 },
		{ 357, 111, 36 },
		{ 357, 112, 36 },
		{ 357, 113, 36 },
		{ 357, 114, 36 },
		{ 357, 115, 36 },
		{ 357, 116, 36 },
		{ 357, 117, 36 },
		{ 357, 118, 36 },
		{ 357, 119, 36 },
		{ 357, 120, 36 },
		{ 357, 121, 36 },
		{ 357, 122, 36 },
		{ 357, 123, 36 },
		{ 357, 124, 36 },
		{ 357, 125, 36 },
		{ 357, 126, 36 },
		{ 357, 127, 36 },
		{ 357, 128, 36 },
		{ 357, 129, 36 },
		{ 357, 130, 36 },
		{ 357, 131, 36 },
		{ 357, 132, 36 },
		{ 357, 133, 36 },
		{ 357, 134, 36 },
		{ 357, 135, 36 },
		{ 357, 136, 36 },
		{ 357, 137, 36 },
		{ 357, 138, 36 },
		{ 357, 139, 36 },
		{ 357, 140, 36 },
		{ 357, 141, 36 },
		{ 357, 142, 36 },
		{ 358, 96, 36 },
		{ 358, 97, 36 },
		{ 358, 98, 36 },
		{ 358, 99, 36 },
		{ 358, 100, 36 },
		{ 358, 101, 36 },
		{ 358, 102, 36 },
		{ 358, 103, 36 },
		{ 358, 104, 36 },
		{ 358, 105, 36 },
		{ 358, 106, 36 },
		{ 358, 107, 36 },
		{ 358, 108, 36 },
		{ 358, 109, 36 },
		{ 358, 110, 36 },
		{ 358, 111, 36 },
		{ 358, 112, 36 },
		{ 358, 113, 36 },
		{ 358, 114, 36 },
		{ 358, 115, 36 },
		{ 358, 116, 36 },
		{ 358, 117, 36 },
		{ 358, 118, 36 },
		{ 358, 119, 36 },
		{ 358, 120, 36 },
		{ 358, 121, 36 },
		{ 358, 122, 36 },
		{ 358, 123, 36 },
		{ 358, 124, 36 },
		{ 358, 125, 36 },
		{ 358, 126, 36 },
		{ 358, 127, 36 },
		{ 358, 128, 36 },
		{ 358, 129, 36 },
		{ 358, 130, 36 },
		{ 358, 131, 36 },
		{ 358, 132, 36 },
		{ 358, 133, 36 },
		{ 358, 134, 36 },
		{ 358, 135, 36 },
		{ 358, 136, 36 },
		{ 358, 137, 36 },
		{ 358, 138, 36 },
		{ 358, 139, 36 },
		{ 358, 140, 36 },
		{ 358, 141, 36 },
		{ 359, 94, 36 },
		{ 359, 95, 36 },
		{ 359, 96, 36 },
		{ 359, 97, 36 },
		{ 359, 98, 36 },
		{ 359, 99, 36 },
		{ 359, 100, 36 },
		{ 359, 101, 36 },
		{ 359, 102, 36 },
		{ 359, 103, 36 },
		{ 359, 104, 36 },
		{ 359, 105, 36 },
		{ 359, 106, 36 },
		{ 359, 107, 36 },
		{ 359, 108, 36 },
		{ 359, 109, 36 },
		{ 359, 110, 36 },
		{ 359, 111, 36 },
		{ 359, 112, 36 },
		{ 359, 113, 36 },
		{ 359, 114, 36 },
		{ 359, 115, 36 },
		{ 359, 116, 36 },
		{ 359, 117, 36 },
		{ 359, 118, 36 },
		{ 359, 119, 36 },
		{ 359, 120, 36 },
		{ 359, 121, 36 },
		{ 359, 122, 36 },
		{ 359, 123, 36 },
		{ 359, 124, 36 },
		{ 359, 125, 36 },
		{ 359, 126, 36 },
		{ 359, 127, 36 },
		{ 359, 128, 36 },
		{ 359, 129, 36 },
		{ 359, 130, 36 },
		{ 359, 131, 36 },
		{ 359, 132, 36 },
		{ 359, 133, 36 },
		{ 359, 134, 36 },
		{ 359, 135, 36 },
		{ 359, 136, 36 },
		{ 359, 137, 36 },
		{ 359, 138, 36 },
		{ 359, 139, 36 },
		{ 359, 140, 36 },
		{ 360, 93, 36 },
		{ 360, 94, 36 },
		{ 360, 95, 36 },
		{ 360, 96, 36 },
		{ 360, 97, 36 },
		{ 360, 98, 36 },
		{ 360, 99, 36 },
		{ 360, 100, 36 },
		{ 360, 101, 36 },
		{ 360, 102, 36 },
		{ 360, 103, 36 },
		{ 360, 104, 36 },
		{ 360, 105, 36 },
		{ 360, 106, 36 },
		{ 360, 107, 36 },
		{ 360, 108, 36 },
		{ 360, 109, 36 },
		{ 360, 110, 36 },
		{ 360, 111, 36 },
		{ 360, 112, 36 },
		{ 360, 113, 36 },
		{ 360, 114, 36 },
		{ 360, 115, 36 },
		{ 360, 116, 36 },
		{ 360, 117, 36 },
		{ 360, 118, 36 },
		{ 360, 119, 36 },
		{ 360, 120, 36 },
		{ 360, 121, 36 },
		{ 360, 122, 36 },
		{ 360, 123, 36 },
		{ 360, 124, 36 },
		{ 360, 125, 36 },
		{ 360, 126, 36 },
		{ 360, 127, 36 },
		{ 360, 128, 36 },
		{ 360, 129, 36 },
		{ 360, 130, 36 },
		{ 360, 131, 36 },
		{ 360, 132, 36 },
		{ 360, 133, 36 },
		{ 360, 134, 36 },
		{ 360, 135, 36 },
		{ 360, 136, 36 },
		{ 360, 137, 36 },
		{ 360, 138, 36 },
		{ 360, 139, 36 },
		{ 361, 91, 36 },
		{ 361, 92, 36 },
		{ 361, 93, 36 },
		{ 361, 94, 36 },
		{ 361, 95, 36 },
		{ 361, 96, 36 },
		{ 361, 97, 36 },
		{ 361, 98, 36 },
		{ 361, 99, 36 },
		{ 361, 100, 36 },
		{ 361, 101, 36 },
		{ 361, 102, 36 },
		{ 361, 103, 36 },
		{ 361, 104, 36 },
		{ 361, 105, 36 },
		{ 361, 106, 36 },
		{ 361, 107, 36 },
		{ 361, 108, 36 },
		{ 361, 109, 36 },
		{ 361, 110, 36 },
		{ 361, 111, 36 },
		{ 361, 112, 36 },
		{ 361, 113, 36 },
		{ 361, 114, 36 },
		{ 361, 115, 36 },
		{ 361, 116, 36 },
		{ 361, 117, 36 },
		{ 361, 118, 36 },
		{ 361, 119, 36 },
		{ 361, 120, 36 },
		{ 361, 121, 36 },
		{ 361, 122, 36 },
		{ 361, 123, 36 },
		{ 361, 124, 36 },
		{ 361, 125, 36 },
		{ 361, 126, 36 },
		{ 361, 127, 36 },
		{ 361, 128, 36 },
		{ 361, 129, 36 },
		{ 361, 130, 36 },
		{ 361, 131, 36 },
		{ 361, 132, 36 },
		{ 361, 133, 36 },
		{ 361, 134, 36 },
		{ 361, 135, 36 },
		{ 361, 136, 36 },
		{ 361, 137, 36 },
		{ 361, 138, 36 },
		{ 362, 90, 36 },
		{ 362, 91, 36 },
		{ 362, 92, 36 },
		{ 362, 93, 36 },
		{ 362, 94, 36 },
		{ 362, 95, 36 },
		{ 362, 96, 36 },
		{ 362, 97, 36 },
		{ 362, 98, 36 },
		{ 362, 99, 36 },
		{ 362, 100, 36 },
		{ 362, 101, 36 },
		{ 362, 102, 36 },
		{ 362, 103, 36 },
		{ 362, 104, 36 },
		{ 362, 105, 36 },
		{ 362, 106, 36 },
		{ 362, 107, 36 },
		{ 362, 108, 36 },
		{ 362, 109, 36 },
		{ 362, 110, 36 },
		{ 362, 111, 36 },
		{ 362, 112, 36 },
		{ 362, 113, 36 },
		{ 362, 114, 36 },
		{ 362, 115, 36 },
		{ 362, 116, 36 },
		{ 362, 117, 36 },
		{ 362, 118, 36 },
		{ 362, 119, 36 },
		{ 362, 120, 36 },
		{ 362, 121, 36 },
		{ 362, 122, 36 },
		{ 362, 123, 36 },
		{ 362, 124, 36 },
		{ 362, 125, 36 },
		{ 362, 126, 36 },
		{ 362, 127, 36 },
		{ 362, 128, 36 },
		{ 362, 129, 36 },
		{ 362, 130, 36 },
		{ 362, 131, 36 },
		{ 362, 132, 36 },
		{ 362, 133, 36 },
		{ 362, 134, 36 },
		{ 362, 135, 36 },
		{ 362, 136, 36 },
		{ 363, 88, 36 },
		{ 363, 89, 36 },
		{ 363, 90, 36 },
		{ 363, 91, 36 },
		{ 363, 92, 36 },
		{ 363, 93, 36 },
		{ 363, 94, 36 },
		{ 363, 95, 36 },
		{ 363, 96, 36 },
		{ 363, 97, 36 },
		{ 363, 98, 36 },
		{ 363, 99, 36 },
		{ 363, 100, 36 },
		{ 363, 101, 36 },
		{ 363, 102, 36 },
		{ 363, 103, 36 },
		{ 363, 104, 36 },
		{ 363, 105, 36 },
		{ 363, 106, 36 },
		{ 363, 107, 36 },
		{ 363, 108, 36 },
		{ 363, 109, 36 },
		{ 363, 110, 36 },
		{ 363, 111, 36 },
		{ 363, 112, 36 },
		{ 363, 113, 36 },
		{ 363, 114, 36 },
		{ 363, 115, 36 },
		{ 363, 116, 36 },
		{ 363, 117, 36 },
		{ 363, 118, 36 },
		{ 363, 119, 36 },
		{ 363, 120, 36 },
		{ 363, 121, 36 },
		{ 363, 122, 36 },
		{ 363, 123, 36 },
		{ 363, 124, 36 },
		{ 363, 125, 36 },
		{ 363, 126, 36 },
		{ 363, 127, 36 },
		{ 363, 128, 36 },
		{ 363, 129, 36 },
		{ 363, 130, 36 },
		{ 363, 131, 36 },
		{ 363, 132, 36 },
		{ 363, 133, 36 },
		{ 363, 134, 36 },
		{ 363, 135, 36 },
		{ 364, 86, 36 },
		{ 364, 87, 36 },
		{ 364, 88, 36 },
		{ 364, 89, 36 },
		{ 364, 90, 36 },
		{ 364, 91, 36 },
		{ 364, 92, 36 },
		{ 364, 93, 36 },
		{ 364, 94, 36 },
		{ 364, 95, 36 },
		{ 364, 96, 36 },
		{ 364, 97, 36 },
		{ 364, 98, 36 },
		{ 364, 99, 36 },
		{ 364, 100, 36 },
		{ 364, 101, 36 },
		{ 364, 102, 36 },
		{ 364, 103, 36 },
		{ 364, 104, 36 },
		{ 364, 105, 36 },
		{ 364, 106, 36 },
		{ 364, 107, 36 },
		{ 364, 108, 36 },
		{ 364, 109, 36 },
		{ 364, 110, 36 },
		{ 364, 111, 36 },
		{ 364, 112, 36 },
		{ 364, 113, 36 },
		{ 364, 114, 36 },
		{ 364, 115, 36 },
		{ 364, 116, 36 },
		{ 364, 117, 36 },
		{ 364, 118, 36 },
		{ 364, 119, 36 },
		{ 364, 120, 36 },
		{ 364, 121, 36 },
		{ 364, 122, 36 },
		{ 364, 123, 36 },
		{ 364, 124, 36 },
		{ 364, 125, 36 },
		{ 364, 126, 36 },
		{ 364, 127, 36 },
		{ 364, 128, 36 },
		{ 364, 129, 36 },
		{ 364, 130, 36 },
		{ 364, 131, 36 },
		{ 364, 132, 36 },
		{ 364, 133, 36 },
		{ 364, 134, 36 },
		{ 365, 85, 36 },
		{ 365, 86, 36 },
		{ 365, 87, 36 },
		{ 365, 88, 36 },
		{ 365, 89, 36 },
		{ 365, 90, 36 },
		{ 365, 91, 36 },
		{ 365, 92, 36 },
		{ 365, 93, 36 },
		{ 365, 94, 36 },
		{ 365, 95, 36 },
		{ 365, 96, 36 },
		{ 365, 97, 36 },
		{ 365, 98, 36 },
		{ 365, 99, 36 },
		{ 365, 100, 36 },
		{ 365, 101, 36 },
		{ 365, 102, 36 },
		{ 365, 103, 36 },
		{ 365, 104, 36 },
		{ 365, 105, 36 },
		{ 365, 106, 36 },
		{ 365, 107, 36 },
		{ 365, 108, 36 },
		{ 365, 109, 36 },
		{ 365, 110, 36 },
		{ 365, 111, 36 },
		{ 365, 112, 36 },
		{ 365, 113, 36 },
		{ 365, 114, 36 },
		{ 365, 115, 36 },
		{ 365, 116, 36 },
		{ 365, 117, 36 },
		{ 365, 118, 36 },
		{ 365, 119, 36 },
		{ 365, 120, 36 },
		{ 365, 121, 36 },
		{ 365, 122, 36 },
		{ 365, 123, 36 },
		{ 365, 124, 36 },
		{ 365, 125, 36 },
		{ 365, 126, 36 },
		{ 365, 127, 36 },
		{ 365, 128, 36 },
		{ 365, 129, 36 },
		{ 365, 130, 36 },
		{ 365, 131, 36 },
		{ 365, 132, 36 },
		{ 365, 133, 36 },
		{ 366, 83, 36 },
		{ 366, 84, 36 },
		{ 366, 85, 36 },
		{ 366, 86, 36 },
		{ 366, 87, 36 },
		{ 366, 88, 36 },
		{ 366, 89, 36 },
		{ 366, 90, 36 },
		{ 366, 91, 36 },
		{ 366, 92, 36 },
		{ 366, 93, 36 },
		{ 366, 94, 36 },
		{ 366, 95, 36 },
		{ 366, 96, 36 },
		{ 366, 97, 36 },
		{ 366, 98, 36 },
		{ 366, 99, 36 },
		{ 366, 100, 36 },
		{ 366, 101, 36 },
		{ 366, 102, 36 },
		{ 366, 103, 36 },
		{ 366, 104, 36 },
		{ 366, 105, 36 },
		{ 366, 106, 36 },
		{ 366, 107, 36 },
		{ 366, 108, 36 },
		{ 366, 109, 36 },
		{ 366, 110, 36 },
		{ 366, 111, 36 },
		{ 366, 112, 36 },
		{ 366, 113, 36 },
		{ 366, 114, 36 },
		{ 366, 115, 36 },
		{ 366, 116, 36 },
		{ 366, 117, 36 },
		{ 366, 118, 36 },
		{ 366, 119, 36 },
		{ 366, 120, 36 },
		{ 366, 121, 36 },
		{ 366, 122, 36 },
		{ 366, 123, 36 },
		{ 366, 124, 36 },
		{ 366, 125, 36 },
		{ 366, 126, 36 },
		{ 366, 127, 36 },
		{ 366, 128, 36 },
		{ 366, 129, 36 },
		{ 366, 130, 36 },
		{ 366, 131, 36 },
		{ 366, 132, 36 },
		{ 367, 82, 36 },
		{ 367, 83, 36 },
		{ 367, 84, 36 },
		{ 367, 85, 36 },
		{ 367, 86, 36 },
		{ 367, 87, 36 },
		{ 367, 88, 36 },
		{ 367, 89, 36 },
		{ 367, 90, 36 },
		{ 367, 91, 36 },
		{ 367, 92, 36 },
		{ 367, 93, 36 },
		{ 367, 94, 36 },
		{ 367, 95, 36 },
		{ 367, 96, 36 },
		{ 367, 97, 36 },
		{ 367, 98, 36 },
		{ 367, 99, 36 },
		{ 367, 100, 36 },
		{ 367, 101, 36 },
		{ 367, 102, 36 },
		{ 367, 103, 36 },
		{ 367, 104, 36 },
		{ 367, 105, 36 },
		{ 367, 106, 36 },
		{ 367, 107, 36 },
		{ 367, 108, 36 },
		{ 367, 109, 36 },
		{ 367, 110, 36 },
		{ 367, 111, 36 },
		{ 367, 112, 36 },
		{ 367, 113, 36 },
		{ 367, 114, 36 },
		{ 367, 115, 36 },
		{ 367, 116, 36 },
		{ 367, 117, 36 },
		{ 367, 118, 36 },
		{ 367, 119, 36 },
		{ 367, 120, 36 },
		{ 367, 121, 36 },
		{ 367, 122, 36 },
		{ 367, 123, 36 },
		{ 367, 124, 36 },
		{ 367, 125, 36 },
		{ 367, 126, 36 },
		{ 367, 127, 36 },
		{ 367, 128, 36 },
		{ 367, 129, 36 },
		{ 367, 130, 36 },
		{ 367, 131, 36 },
		{ 368, 80, 36 },
		{ 368, 81, 36 },
		{ 368, 82, 36 },
		{ 368, 83, 36 },
		{ 368, 84, 36 },
		{ 368, 85, 36 },
		{ 368, 86, 36 },
		{ 368, 87, 36 },
		{ 368, 88, 36 },
		{ 368, 89, 36 },
		{ 368, 90, 36 },
		{ 368, 91, 36 },
		{ 368, 92, 36 },
		{ 368, 93, 36 },
		{ 368, 94, 36 },
		{ 368, 95, 36 },
		{ 368, 96, 36 },
		{ 368, 97, 36 },
		{ 368, 98, 36 },
		{ 368, 99, 36 },
		{ 368, 100, 36 },
		{ 368, 101, 36 },
		{ 368, 102, 36 },
		{ 368, 103, 36 },
		{ 368, 104, 36 },
		{ 368, 105, 36 },
		{ 368, 106, 36 },
		{ 368, 107, 36 },
		{ 368, 108, 36 },
		{ 368, 109, 36 },
		{ 368, 110, 36 },
		{ 368, 111, 36 },
		{ 368, 112, 36 },
		{ 368, 113, 36 },
		{ 368, 114, 36 },
		{ 368, 115, 36 },
		{ 368, 116, 36 },
		{ 368, 117, 36 },
		{ 368, 118, 36 },
		{ 368, 119, 36 },
		{ 368, 120, 36 },
		{ 368, 121, 36 },
		{ 368, 122, 36 },
		{ 368, 123, 36 },
		{ 368, 124, 36 },
		{ 368, 125, 36 },
		{ 368, 126, 36 },
		{ 368, 127, 36 },
		{ 368, 128, 36 },
		{ 368, 129, 36 },
		{ 368, 130, 36 },
		{ 369, 78, 36 },
		{ 369, 79, 36 },
		{ 369, 80, 36 },
		{ 369, 81, 36 },
		{ 369, 82, 36 },
		{ 369, 83, 36 },
		{ 369, 84, 36 },
		{ 369, 85, 36 },
		{ 369, 86, 36 },
		{ 369, 87, 36 },
		{ 369, 88, 36 },
		{ 369, 89, 36 },
		{ 369, 90, 36 },
		{ 369, 91, 36 },
		{ 369, 92, 36 },
		{ 369, 93, 36 },
		{ 369, 94, 36 },
		{ 369, 95, 36 },
		{ 369, 96, 36 },
		{ 369, 97, 36 },
		{ 369, 98, 36 },
		{ 369, 99, 36 },
		{ 369, 100, 36 },
		{ 369, 101, 36 },
		{ 369, 102, 36 },
		{ 369, 103, 36 },
		{ 369, 104, 36 },
		{ 369, 105, 36 },
		{ 369, 106, 36 },
		{ 369, 107, 36 },
		{ 369, 108, 36 },
		{ 369, 109, 36 },
		{ 369, 110, 36 },
		{ 369, 111, 36 },
		{ 369, 112, 36 },
		{ 369, 113, 36 },
		{ 369, 114, 36 },
		{ 369, 115, 36 },
		{ 369, 116, 36 },
		{ 369, 117, 36 },
		{ 369, 118, 36 },
		{ 369, 119, 36 },
		{ 369, 120, 36 },
		{ 369, 121, 36 },
		{ 369, 122, 36 },
		{ 369, 123, 36 },
		{ 369, 124, 36 },
		{ 369, 125, 36 },
		{ 369, 126, 36 },
		{ 369, 127, 36 },
		{ 369, 128, 36 },
		{ 370, 77, 36 },
		{ 370, 78, 36 },
		{ 370, 79, 36 },
		{ 370, 80, 36 },
		{ 370, 81, 36 },
		{ 370, 82, 36 },
		{ 370, 83, 36 },
		{ 370, 84, 36 },
		{ 370, 85, 36 },
		{ 370, 86, 36 },
		{ 370, 87, 36 },
		{ 370, 88, 36 },
		{ 370, 89, 36 },
		{ 370, 90, 36 },
		{ 370, 91, 36 },
		{ 370, 92, 36 },
		{ 370, 93, 36 },
		{ 370, 94, 36 },
		{ 370, 95, 36 },
		{ 370, 96, 36 },
		{ 370, 97, 36 },
		{ 370, 98, 36 },
		{ 370, 99, 36 },
		{ 370, 100, 36 },
		{ 370, 101, 36 },
		{ 370, 102, 36 },
		{ 370, 103, 36 },
		{ 370, 104, 36 },
		{ 370, 105, 36 },
		{ 370, 106, 36 },
		{ 370, 107, 36 },
		{ 370, 108, 36 },
		{ 370, 109, 36 },
		{ 370, 110, 36 },
		{ 370, 111, 36 },
		{ 370, 112, 36 },
		{ 370, 113, 36 },
		{ 370, 114, 36 },
		{ 370, 115, 36 },
		{ 370, 116, 36 },
		{ 370, 117, 36 },
		{ 370, 118, 36 },
		{ 370, 119, 36 },
		{ 370, 120, 36 },
		{ 370, 121, 36 },
		{ 370, 122, 36 },
		{ 370, 123, 36 },
		{ 370, 124, 36 },
		{ 370, 125, 36 },
		{ 370, 126, 36 },
		{ 370, 127, 36 },
		{ 371, 75, 36 },
		{ 371, 76, 36 },
		{ 371, 77, 36 },
		{ 371, 78, 36 },
		{ 371, 79, 36 },
		{ 371, 80, 36 },
		{ 371, 81, 36 },
		{ 371, 82, 36 },
		{ 371, 83, 36 },
		{ 371, 84, 36 },
		{ 371, 85, 36 },
		{ 371, 86, 36 },
		{ 371, 87, 36 },
		{ 371, 88, 36 },
		{ 371, 89, 36 },
		{ 371, 90, 36 },
		{ 371, 91, 36 },
		{ 371, 92, 36 },
		{ 371, 93, 36 },
		{ 371, 94, 36 },
		{ 371, 95, 36 },
		{ 371, 96, 36 },
		{ 371, 97, 36 },
		{ 371, 98, 36 },
		{ 371, 99, 36 },
		{ 371, 100, 36 },
		{ 371, 101, 36 },
		{ 371, 102, 36 },
		{ 371, 103, 36 },
		{ 371, 104, 36 },
		{ 371, 105, 36 },
		{ 371, 106, 36 },
		{ 371, 107, 36 },
		{ 371, 108, 36 },
		{ 371, 109, 36 },
		{ 371, 110, 36 },
		{ 371, 111, 36 },
		{ 371, 112, 36 },
		{ 371, 113, 36 },
		{ 371, 114, 36 },
		{ 371, 115, 36 },
		{ 371, 116, 36 },
		{ 371, 117, 36 },
		{ 371, 118, 36 },
		{ 371, 119, 36 },
		{ 371, 120, 36 },
		{ 371, 121, 36 },
		{ 371, 122, 36 },
		{ 371, 123, 36 },
		{ 371, 124, 36 },
		{ 371, 125, 36 },
		{ 371, 126, 36 },
		{ 372, 74, 36 },
		{ 372, 75, 36 },
		{ 372, 76, 36 },
		{ 372, 77, 36 },
		{ 372, 78, 36 },
		{ 372, 79, 36 },
		{ 372, 80, 36 },
		{ 372, 81, 36 },
		{ 372, 82, 36 },
		{ 372, 83, 36 },
		{ 372, 84, 36 },
		{ 372, 85, 36 },
		{ 372, 86, 36 },
		{ 372, 87, 36 },
		{ 372, 88, 36 },
		{ 372, 89, 36 },
		{ 372, 90, 36 },
		{ 372, 91, 36 },
		{ 372, 92, 36 },
		{ 372, 93, 36 },
		{ 372, 94, 36 },
		{ 372, 95, 36 },
		{ 372, 96, 36 },
		{ 372, 97, 36 },
		{ 372, 98, 36 },
		{ 372, 99, 36 },
		{ 372, 100, 36 },
		{ 372, 101, 36 },
		{ 372, 102, 36 },
		{ 372, 103, 36 },
		{ 372, 104, 36 },
		{ 372, 105, 36 },
		{ 372, 106, 36 },
		{ 372, 107, 36 },
		{ 372, 108, 36 },
		{ 372, 109, 36 },
		{ 372, 110, 36 },
		{ 372, 111, 36 },
		{ 372, 112, 36 },
		{ 372, 113, 36 },
		{ 372, 114, 36 },
		{ 372, 115, 36 },
		{ 372, 116, 36 },
		{ 372, 117, 36 },
		{ 372, 118, 36 },
		{ 372, 119, 36 },
		{ 372, 120, 36 },
		{ 372, 121, 36 },
		{ 372, 122, 36 },
		{ 372, 123, 36 },
		{ 372, 124, 36 },
		{ 372, 125, 36 },
		{ 373, 72, 36 },
		{ 373, 73, 36 },
		{ 373, 74, 36 },
		{ 373, 75, 36 },
		{ 373, 76, 36 },
		{ 373, 77, 36 },
		{ 373, 78, 36 },
		{ 373, 79, 36 },
		{ 373, 80, 36 },
		{ 373, 81, 36 },
		{ 373, 82, 36 },
		{ 373, 83, 36 },
		{ 373, 84, 36 },
		{ 373, 85, 36 },
		{ 373, 86, 36 },
		{ 373, 87, 36 },
		{ 373, 88, 36 },
		{ 373, 89, 36 },
		{ 373, 90, 36 },
		{ 373, 91, 36 },
		{ 373, 92, 36 },
		{ 373, 93, 36 },
		{ 373, 94, 36 },
		{ 373, 95, 36 },
		{ 373, 96, 36 },
		{ 373, 97, 36 },
		{ 373, 98, 36 },
		{ 373, 99, 36 },
		{ 373, 100, 36 },
		{ 373, 101, 36 },
		{ 373, 102, 36 },
		{ 373, 103, 36 },
		{ 373, 104, 36 },
		{ 373, 105, 36 },
		{ 373, 106, 36 },
		{ 373, 107, 36 },
		{ 373, 108, 36 },
		{ 373, 109, 36 },
		{ 373, 110, 36 },
		{ 373, 111, 36 },
		{ 373, 112, 36 },
		{ 373, 113, 36 },
		{ 373, 114, 36 },
		{ 373, 115, 36 },
		{ 373, 116, 36 },
		{ 373, 117, 36 },
		{ 373, 118, 36 },
		{ 373, 119, 36 },
		{ 373, 120, 36 },
		{ 373, 121, 36 },
		{ 373, 122, 36 },
		{ 373, 123, 36 },
		{ 373, 124, 36 },
		{ 374, 70, 36 },
		{ 374, 71, 36 },
		{ 374, 72, 36 },
		{ 374, 73, 36 },
		{ 374, 74, 36 },
		{ 374, 75, 36 },
		{ 374, 76, 36 },
		{ 374, 77, 36 },
		{ 374, 78, 36 },
		{ 374, 79, 36 },
		{ 374, 80, 36 },
		{ 374, 81, 36 },
		{ 374, 82, 36 },
		{ 374, 83, 36 },
		{ 374, 84, 36 },
		{ 374, 85, 36 },
		{ 374, 86, 36 },
		{ 374, 87, 36 },
		{ 374, 88, 36 },
		{ 374, 89, 36 },
		{ 374, 90, 36 },
		{ 374, 91, 36 },
		{ 374, 92, 36 },
		{ 374, 93, 36 },
		{ 374, 94, 36 },
		{ 374, 95, 36 },
		{ 374, 96, 36 },
		{ 374, 97, 36 },
		{ 374, 98, 36 },
		{ 374, 99, 36 },
		{ 374, 100, 36 },
		{ 374, 101, 36 },
		{ 374, 102, 36 },
		{ 374, 103, 36 },
		{ 374, 104, 36 },
		{ 374, 105, 36 },
		{ 374, 106, 36 },
		{ 374, 107, 36 },
		{ 374, 108, 36 },
		{ 374, 109, 36 },
		{ 374, 110, 36 },
		{ 374, 111, 36 },
		{ 374, 112, 36 },
		{ 374, 113, 36 },
		{ 374, 114, 36 },
		{ 374, 115, 36 },
		{ 374, 116, 36 },
		{ 374, 117, 36 },
		{ 374, 118, 36 },
		{ 374, 119, 36 },
		{ 374, 120, 36 },
		{ 374, 121, 36 },
		{ 374, 122, 36 },
		{ 374, 123, 36 },
		{ 375, 69, 36 },
		{ 375, 70, 36 },
		{ 375, 71, 36 },
		{ 375, 72, 36 },
		{ 375, 73, 36 },
		{ 375, 74, 36 },
		{ 375, 75, 36 },
		{ 375, 76, 36 },
		{ 375, 77, 36 },
		{ 375, 78, 36 },
		{ 375, 79, 36 },
		{ 375, 80, 36 },
		{ 375, 81, 36 },
		{ 375, 82, 36 },
		{ 375, 83, 36 },
		{ 375, 84, 36 },
		{ 375, 85, 36 },
		{ 375, 86, 36 },
		{ 375, 87, 36 },
		{ 375, 88, 36 },
		{ 375, 89, 36 },
		{ 375, 90, 36 },
		{ 375, 91, 36 },
		{ 375, 92, 36 },
		{ 375, 93, 36 },
		{ 375, 94, 36 },
		{ 375, 95, 36 },
		{ 375, 96, 36 },
		{ 375, 97, 36 },
		{ 375, 98, 36 },
		{ 375, 99, 36 },
		{ 375, 100, 36 },
		{ 375, 101, 36 },
		{ 375, 102, 36 },
		{ 375, 103, 36 },
		{ 375, 104, 36 },
		{ 375, 105, 36 },
		{ 375, 106, 36 },
		{ 375, 107, 36 },
		{ 375, 108, 36 },
		{ 375, 109, 36 },
		{ 375, 110, 36 },
		{ 375, 111, 36 },
		{ 375, 112, 36 },
		{ 375, 113, 36 },
		{ 375, 114, 36 },
		{ 375, 115, 36 },
		{ 375, 116, 36 },
		{ 375, 117, 36 },
		{ 375, 118, 36 },
		{ 375, 119, 36 },
		{ 375, 120, 36 },
		{ 375, 121, 36 },
		{ 375, 122, 36 },
		{ 376, 67, 36 },
		{ 376, 68, 36 },
		{ 376, 69, 36 },
		{ 376, 70, 36 },
		{ 376, 71, 36 },
		{ 376, 72, 36 },
		{ 376, 73, 36 },
		{ 376, 74, 36 },
		{ 376, 75, 36 },
		{ 376, 76, 36 },
		{ 376, 77, 36 },
		{ 376, 78, 36 },
		{ 376, 79, 36 },
		{ 376, 80, 36 },
		{ 376, 81, 36 },
		{ 376, 82, 36 },
		{ 376, 83, 36 },
		{ 376, 84, 36 },
		{ 376, 85, 36 },
		{ 376, 86, 36 },
		{ 376, 87, 36 },
		{ 376, 88, 36 },
		{ 376, 89, 36 },
		{ 376, 90, 36 },
		{ 376, 91, 36 },
		{ 376, 92, 36 },
		{ 376, 93, 36 },
		{ 376, 94, 36 },
		{ 376, 95, 36 },
		{ 376, 96, 36 },
		{ 376, 97, 36 },
		{ 376, 98, 36 },
		{ 376, 99, 36 },
		{ 376, 100, 36 },
		{ 376, 101, 36 },
		{ 376, 102, 36 },
		{ 376, 103, 36 },
		{ 376, 104, 36 },
		{ 376, 105, 36 },
		{ 376, 106, 36 },
		{ 376, 107, 36 },
		{ 376, 108, 36 },
		{ 376, 109, 36 },
		{ 376, 110, 36 },
		{ 376, 111, 36 },
		{ 376, 112, 36 },
		{ 376, 113, 36 },
		{ 376, 114, 36 },
		{ 376, 115, 36 },
		{ 376, 116, 36 },
		{ 376, 117, 36 },
		{ 376, 118, 36 },
		{ 376, 119, 36 },
		{ 376, 120, 36 },
		{ 377, 66, 36 },
		{ 377, 67, 36 },
		{ 377, 68, 36 },
		{ 377, 69, 36 },
		{ 377, 70, 36 },
		{ 377, 71, 36 },
		{ 377, 72, 36 },
		{ 377, 73, 36 },
		{ 377, 74, 36 },
		{ 377, 75, 36 },
		{ 377, 76, 36 },
		{ 377, 77, 36 },
		{ 377, 78, 36 },
		{ 377, 79, 36 },
		{ 377, 80, 36 },
		{ 377, 81, 36 },
		{ 377, 82, 36 },
		{ 377, 83, 36 },
		{ 377, 84, 36 },
		{ 377, 85, 36 },
		{ 377, 86, 36 },
		{ 377, 87, 36 },
		{ 377, 88, 36 },
		{ 377, 89, 36 },
		{ 377, 90, 36 },
		{ 377, 91, 36 },
		{ 377, 92, 36 },
		{ 377, 93, 36 },
		{ 377, 94, 36 },
		{ 377, 95, 36 },
		{ 377, 96, 36 },
		{ 377, 97, 36 },
		{ 377, 98, 36 },
		{ 377, 99, 36 },
		{ 377, 100, 36 },
		{ 377, 101, 36 },
		{ 377, 102, 36 },
		{ 377, 103, 36 },
		{ 377, 104, 36 },
		{ 377, 105, 36 },
		{ 377, 106, 36 },
		{ 377, 107, 36 },
		{ 377, 108, 36 },
		{ 377, 109, 36 },
		{ 377, 110, 36 },
		{ 377, 111, 36 },
		{ 377, 112, 36 },
		{ 377, 113, 36 },
		{ 377, 114, 36 },
		{ 377, 115, 36 },
		{ 377, 116, 36 },
		{ 377, 117, 36 },
		{ 377, 118, 36 },
		{ 377, 119, 36 },
		{ 378, 64, 36 },
		{ 378, 65, 36 },
		{ 378, 66, 36 },
		{ 378, 67, 36 },
		{ 378, 68, 36 },
		{ 378, 69, 36 },
		{ 378, 70, 36 },
		{ 378, 71, 36 },
		{ 378, 72, 36 },
		{ 378, 73, 36 },
		{ 378, 74, 36 },
		{ 378, 75, 36 },
		{ 378, 76, 36 },
		{ 378, 77, 36 },
		{ 378, 78, 36 },
		{ 378, 79, 36 },
		{ 378, 80, 36 },
		{ 378, 81, 36 },
		{ 378, 82, 36 },
		{ 378, 83, 36 },
		{ 378, 84, 36 },
		{ 378, 85, 36 },
		{ 378, 86, 36 },
		{ 378, 87, 36 },
		{ 378, 88, 36 },
		{ 378, 89, 36 },
		{ 378, 90, 36 },
		{ 378, 91, 36 },
		{ 378, 92, 36 },
		{ 378, 93, 36 },
		{ 378, 94, 36 },
		{ 378, 95, 36 },
		{ 378, 96, 36 },
		{ 378, 97, 36 },
		{ 378, 98, 36 },
		{ 378, 99, 36 },
		{ 378, 100, 36 },
		{ 378, 101, 36 },
		{ 378, 102, 36 },
		{ 378, 103, 36 },
		{ 378, 104, 36 },
		{ 378, 105, 36 },
		{ 378, 106, 36 },
		{ 378, 107, 36 },
		{ 378, 108, 36 },
		{ 378, 109, 36 },
		{ 378, 110, 36 },
		{ 378, 111, 36 },
		{ 378, 112, 36 },
		{ 378, 113, 36 },
		{ 378, 114, 36 },
		{ 378, 115, 36 },
		{ 378, 116, 36 },
		{ 378, 117, 36 },
		{ 378, 118, 36 },
		{ 379, 62, 36 },
		{ 379, 63, 36 },
		{ 379, 64, 36 },
		{ 379, 65, 36 },
		{ 379, 66, 36 },
		{ 379, 67, 36 },
		{ 379, 68, 36 },
		{ 379, 69, 36 },
		{ 379, 70, 36 },
		{ 379, 71, 36 },
		{ 379, 72, 36 },
		{ 379, 73, 36 },
		{ 379, 74, 36 },
		{ 379, 75, 36 },
		{ 379, 76, 36 },
		{ 379, 77, 36 },
		{ 379, 78, 36 },
		{ 379, 79, 36 },
		{ 379, 80, 36 },
		{ 379, 81, 36 },
		{ 379, 82, 36 },
		{ 379, 83, 36 },
		{ 379, 84, 36 },
		{ 379, 85, 36 },
		{ 379, 86, 36 },
		{ 379, 87, 36 },
		{ 379, 88, 36 },
		{ 379, 89, 36 },
		{ 379, 90, 36 },
		{ 379, 91, 36 },
		{ 379, 92, 36 },
		{ 379, 93, 36 },
		{ 379, 94, 36 },
		{ 379, 95, 36 },
		{ 379, 96, 36 },
		{ 379, 97, 36 },
		{ 379, 98, 36 },
		{ 379, 99, 36 },
		{ 379, 100, 36 },
		{ 379, 101, 36 },
		{ 379, 102, 36 },
		{ 379, 103, 36 },
		{ 379, 104, 36 },
		{ 379, 105, 36 },
		{ 379, 106, 36 },
		{ 379, 107, 36 },
		{ 379, 108, 36 },
		{ 379, 109, 36 },
		{ 379, 110, 36 },
		{ 379, 111, 36 },
		{ 379, 112, 36 },
		{ 379, 113, 36 },
		{ 379, 114, 36 },
		{ 379, 115, 36 },
		{ 379, 116, 36 },
		{ 379, 117, 36 },
		{ 380, 61, 36 },
		{ 380, 62, 36 },
		{ 380, 63, 36 },
		{ 380, 64, 36 },
		{ 380, 65, 36 },
		{ 380, 66, 36 },
		{ 380, 67, 36 },
		{ 380, 68, 36 },
		{ 380, 69, 36 },
		{ 380, 70, 36 },
		{ 380, 71, 36 },
		{ 380, 72, 36 },
		{ 380, 73, 36 },
		{ 380, 74, 36 },
		{ 380, 75, 36 },
		{ 380, 76, 36 },
		{ 380, 77, 36 },
		{ 380, 78, 36 },
		{ 380, 79, 36 },
		{ 380, 80, 36 },
		{ 380, 81, 36 },
		{ 380, 82, 36 },
		{ 380, 83, 36 },
		{ 380, 84, 36 },
		{ 380, 85, 36 },
		{ 380, 86, 36 },
		{ 380, 87, 36 },
		{ 380, 88, 36 },
		{ 380, 89, 36 },
		{ 380, 90, 36 },
		{ 380, 91, 36 },
		{ 380, 92, 36 },
		{ 380, 93, 36 },
		{ 380, 94, 36 },
		{ 380, 95, 36 },
		{ 380, 96, 36 },
		{ 380, 97, 36 },
		{ 380, 98, 36 },
		{ 380, 99, 36 },
		{ 380, 100, 36 },
		{ 380, 101, 36 },
		{ 380, 102, 36 },
		{ 380, 103, 36 },
		{ 380, 104, 36 },
		{ 380, 105, 36 },
		{ 380, 106, 36 },
		{ 380, 107, 36 },
		{ 380, 108, 36 },
		{ 380, 109, 36 },
		{ 380, 110, 36 },
		{ 380, 111, 36 },
		{ 380, 112, 36 },
		{ 380, 113, 36 },
		{ 380, 114, 36 },
		{ 380, 115, 36 },
		{ 380, 116, 36 },
		{ 381, 59, 36 },
		{ 381, 60, 36 },
		{ 381, 61, 36 },
		{ 381, 62, 36 },
		{ 381, 63, 36 },
		{ 381, 64, 36 },
		{ 381, 65, 36 },
		{ 381, 66, 36 },
		{ 381, 67, 36 },
		{ 381, 68, 36 },
		{ 381, 69, 36 },
		{ 381, 70, 36 },
		{ 381, 71, 36 },
		{ 381, 72, 36 },
		{ 381, 73, 36 },
		{ 381, 74, 36 },
		{ 381, 75, 36 },
		{ 381, 76, 36 },
		{ 381, 77, 36 },
		{ 381, 78, 36 },
		{ 381, 79, 36 },
		{ 381, 80, 36 },
		{ 381, 81, 36 },
		{ 381, 82, 36 },
		{ 381, 83, 36 },
		{ 381, 84, 36 },
		{ 381, 85, 36 },
		{ 381, 86, 36 },
		{ 381, 87, 36 },
		{ 381, 88, 36 },
		{ 381, 89, 36 },
		{ 381, 90, 36 },
		{ 381, 91, 36 },
		{ 381, 92, 36 },
		{ 381, 93, 36 },
		{ 381, 94, 36 },
		{ 381, 95, 36 },
		{ 381, 96, 36 },
		{ 381, 97, 36 },
		{ 381, 98, 36 },
		{ 381, 99, 36 },
		{ 381, 100, 36 },
		{ 381, 101, 36 },
		{ 381, 102, 36 },
		{ 381, 103, 36 },
		{ 381, 104, 36 },
		{ 381, 105, 36 },
		{ 381, 106, 36 },
		{ 381, 107, 36 },
		{ 381, 108, 36 },
		{ 381, 109, 36 },
		{ 381, 110, 36 },
		{ 381, 111, 36 },
		{ 381, 112, 36 },
		{ 381, 113, 36 },
		{ 381, 114, 36 },
		{ 381, 115, 36 },
		{ 382, 58, 36 },
		{ 382, 59, 36 },
		{ 382, 60, 36 },
		{ 382, 61, 36 },
		{ 382, 62, 36 },
		{ 382, 63, 36 },
		{ 382, 64, 36 },
		{ 382, 65, 36 },
		{ 382, 66, 36 },
		{ 382, 67, 36 },
		{ 382, 68, 36 },
		{ 382, 69, 36 },
		{ 382, 70, 36 },
		{ 382, 71, 36 },
		{ 382, 72, 36 },
		{ 382, 73, 36 },
		{ 382, 74, 36 },
		{ 382, 75, 36 },
		{ 382, 76, 36 },
		{ 382, 77, 36 },
		{ 382, 78, 36 },
		{ 382, 79, 36 },
		{ 382, 80, 36 },
		{ 382, 81, 36 },
		{ 382, 82, 36 },
		{ 382, 83, 36 },
		{ 382, 84, 36 },
		{ 382, 85, 36 },
		{ 382, 86, 36 },
		{ 382, 87, 36 },
		{ 382, 88, 36 },
		{ 382, 89, 36 },
		{ 382, 90, 36 },
		{ 382, 91, 36 },
		{ 382, 92, 36 },
		{ 382, 93, 36 },
		{ 382, 94, 36 },
		{ 382, 95, 36 },
		{ 382, 96, 36 },
		{ 382, 97, 36 },
		{ 382, 98, 36 },
		{ 382, 99, 36 },
		{ 382, 100, 36 },
		{ 382, 101, 36 },
		{ 382, 102, 36 },
		{ 382, 103, 36 },
		{ 382, 104, 36 },
		{ 382, 105, 36 },
		{ 382, 106, 36 },
		{ 382, 107, 36 },
		{ 382, 108, 36 },
		{ 382, 109, 36 },
		{ 382, 110, 36 },
		{ 382, 111, 36 },
		{ 382, 112, 36 },
		{ 382, 113, 36 },
		{ 382, 114, 36 },
		{ 383, 56, 36 },
		{ 383, 57, 36 },
		{ 383, 58, 36 },
		{ 383, 59, 36 },
		{ 383, 60, 36 },
		{ 383, 61, 36 },
		{ 383, 62, 36 },
		{ 383, 63, 36 },
		{ 383, 64, 36 },
		{ 383, 65, 36 },
		{ 383, 66, 36 },
		{ 383, 67, 36 },
		{ 383, 68, 36 },
		{ 383, 69, 36 },
		{ 383, 70, 36 },
		{ 383, 71, 36 },
		{ 383, 72, 36 },
		{ 383, 73, 36 },
		{ 383, 74, 36 },
		{ 383, 75, 36 },
		{ 383, 76, 36 },
		{ 383, 77, 36 },
		{ 383, 78, 36 },
		{ 383, 79, 36 },
		{ 383, 80, 36 },
		{ 383, 81, 36 },
		{ 383, 82, 36 },
		{ 383, 83, 36 },
		{ 383, 84, 36 },
		{ 383, 85, 36 },
		{ 383, 86, 36 },
		{ 383, 87, 36 },
		{ 383, 88, 36 },
		{ 383, 89, 36 },
		{ 383, 90, 36 },
		{ 383, 91, 36 },
		{ 383, 92, 36 },
		{ 383, 93, 36 },
		{ 383, 94, 36 },
		{ 383, 95, 36 },
		{ 383, 96, 36 },
		{ 383, 97, 36 },
		{ 383, 98, 36 },
		{ 383, 99, 36 },
		{ 383, 100, 36 },
		{ 383, 101, 36 },
		{ 383, 102, 36 },
		{ 383, 103, 36 },
		{ 383, 104, 36 },
		{ 383, 105, 36 },
		{ 383, 106, 36 },
		{ 383, 107, 36 },
		{ 383, 108, 36 },
		{ 383, 109, 36 },
		{ 383, 110, 36 },
		{ 383, 111, 36 },
		{ 383, 112, 36 },
		{ 384, 54, 36 },
		{ 384, 55, 36 },
		{ 384, 56, 36 },
		{ 384, 57, 36 },
		{ 384, 58, 36 },
		{ 384, 59, 36 },
		{ 384, 60, 36 },
		{ 384, 61, 36 },
		{ 384, 62, 36 },
		{ 384, 63, 36 },
		{ 384, 64, 36 },
		{ 384, 65, 36 },
		{ 384, 66, 36 },
		{ 384, 67, 36 },
		{ 384, 68, 36 },
		{ 384, 69, 36 },
		{ 384, 70, 36 },
		{ 384, 71, 36 },
		{ 384, 72, 36 },
		{ 384, 73, 36 },
		{ 384, 74, 36 },
		{ 384, 75, 36 },
		{ 384, 76, 36 },
		{ 384, 77, 36 },
		{ 384, 78, 36 },
		{ 384, 79, 36 },
		{ 384, 80, 36 },
		{ 384, 81, 36 },
		{ 384, 82, 36 },
		{ 384, 83, 36 },
		{ 384, 84, 36 },
		{ 384, 85, 36 },
		{ 384, 86, 36 },
		{ 384, 87, 36 },
		{ 384, 88, 36 },
		{ 384, 89, 36 },
		{ 384, 90, 36 },
		{ 384, 91, 36 },
		{ 384, 92, 36 },
		{ 384, 93, 36 },
		{ 384, 94, 36 },
		{ 384, 95, 36 },
		{ 384, 96, 36 },
		{ 384, 97, 36 },
		{ 384, 98, 36 },
		{ 384, 99, 36 },
		{ 384, 100, 36 },
		{ 384, 101, 36 },
		{ 384, 102, 36 },
		{ 384, 103, 36 },
		{ 384, 104, 36 },
		{ 384, 105, 36 },
		{ 384, 106, 36 },
		{ 384, 107, 36 },
		{ 384, 108, 36 },
		{ 384, 109, 36 },
		{ 384, 110, 36 },
		{ 384, 111, 36 },
		{ 385, 53, 36 },
		{ 385, 54, 36 },
		{ 385, 55, 36 },
		{ 385, 56, 36 },
		{ 385, 57, 36 },
		{ 385, 58, 36 },
		{ 385, 59, 36 },
		{ 385, 60, 36 },
		{ 385, 61, 36 },
		{ 385, 62, 36 },
		{ 385, 63, 36 },
		{ 385, 64, 36 },
		{ 385, 65, 36 },
		{ 385, 66, 36 },
		{ 385, 67, 36 },
		{ 385, 68, 36 },
		{ 385, 69, 36 },
		{ 385, 70, 36 },
		{ 385, 71, 36 },
		{ 385, 72, 36 },
		{ 385, 73, 36 },
		{ 385, 74, 36 },
		{ 385, 75, 36 },
		{ 385, 76, 36 },
		{ 385, 77, 36 },
		{ 385, 78, 36 },
		{ 385, 79, 36 },
		{ 385, 80, 36 },
		{ 385, 81, 36 },
		{ 385, 82, 36 },
		{ 385, 83, 36 },
		{ 385, 84, 36 },
		{ 385, 85, 36 },
		{ 385, 86, 36 },
		{ 385, 87, 36 },
		{ 385, 88, 36 },
		{ 385, 89, 36 },
		{ 385, 90, 36 },
		{ 385, 91, 36 },
		{ 385, 92, 36 },
		{ 385, 93, 36 },
		{ 385, 94, 36 },
		{ 385, 95, 36 },
		{ 385, 96, 36 },
		{ 385, 97, 36 },
		{ 385, 98, 36 },
		{ 385, 99, 36 },
		{ 385, 100, 36 },
		{ 385, 101, 36 },
		{ 385, 102, 36 },
		{ 385, 103, 36 },
		{ 385, 104, 36 },
		{ 385, 105, 36 },
		{ 385, 106, 36 },
		{ 385, 107, 36 },
		{ 385, 108, 36 },
		{ 385, 109, 36 },
		{ 385, 110, 36 },
		{ 386, 51, 36 },
		{ 386, 52, 36 },
		{ 386, 53, 36 },
		{ 386, 54, 36 },
		{ 386, 55, 36 },
		{ 386, 56, 36 },
		{ 386, 57, 36 },
		{ 386, 58, 36 },
		{ 386, 59, 36 },
		{ 386, 60, 36 },
		{ 386, 61, 36 },
		{ 386, 62, 36 },
		{ 386, 63, 36 },
		{ 386, 64, 36 },
		{ 386, 65, 36 },
		{ 386, 66, 36 },
		{ 386, 67, 36 },
		{ 386, 68, 36 },
		{ 386, 69, 36 },
		{ 386, 70, 36 },
		{ 386, 71, 36 },
		{ 386, 72, 36 },
		{ 386, 73, 36 },
		{ 386, 74, 36 },
		{ 386, 75, 36 },
		{ 386, 76, 36 },
		{ 386, 77, 36 },
		{ 386, 78, 36 },
		{ 386, 79, 36 },
		{ 386, 80, 36 },
		{ 386, 81, 36 },
		{ 386, 82, 36 },
		{ 386, 83, 36 },
		{ 386, 84, 36 },
		{ 386, 85, 36 },
		{ 386, 86, 36 },
		{ 386, 87, 36 },
		{ 386, 88, 36 },
		{ 386, 89, 36 },
		{ 386, 90, 36 },
		{ 386, 91, 36 },
		{ 386, 92, 36 },
		{ 386, 93, 36 },
		{ 386, 94, 36 },
		{ 386, 95, 36 },
		{ 386, 96, 36 },
		{ 386, 97, 36 },
		{ 386, 98, 36 },
		{ 386, 99, 36 },
		{ 386, 100, 36 },
		{ 386, 101, 36 },
		{ 386, 102, 36 },
		{ 386, 103, 36 },
		{ 386, 104, 36 },
		{ 386, 105, 36 },
		{ 386, 106, 36 },
		{ 386, 107, 36 },
		{ 386, 108, 36 },
		{ 386, 109, 36 },
		{ 387, 50, 36 },
		{ 387, 51, 36 },
		{ 387, 52, 36 },
		{ 387, 53, 36 },
		{ 387, 54, 36 },
		{ 387, 55, 36 },
		{ 387, 56, 36 },
		{ 387, 57, 36 },
		{ 387, 58, 36 },
		{ 387, 59, 36 },
		{ 387, 60, 36 },
		{ 387, 61, 36 },
		{ 387, 62, 36 },
		{ 387, 63, 36 },
		{ 387, 64, 36 },
		{ 387, 65, 36 },
		{ 387, 66, 36 },
		{ 387, 67, 36 },
		{ 387, 68, 36 },
		{ 387, 69, 36 },
		{ 387, 70, 36 },
		{ 387, 71, 36 },
		{ 387, 72, 36 },
		{ 387, 73, 36 },
		{ 387, 74, 36 },
		{ 387, 75, 36 },
		{ 387, 76, 36 },
		{ 387, 77, 36 },
		{ 387, 78, 36 },
		{ 387, 79, 36 },
		{ 387, 80, 36 },
		{ 387, 81, 36 },
		{ 387, 82, 36 },
		{ 387, 83, 36 },
		{ 387, 84, 36 },
		{ 387, 85, 36 },
		{ 387, 86, 36 },
		{ 387, 87, 36 },
		{ 387, 88, 36 },
		{ 387, 89, 36 },
		{ 387, 90, 36 },
		{ 387, 91, 36 },
		{ 387, 92, 36 },
		{ 387, 93, 36 },
		{ 387, 94, 36 },
		{ 387, 95, 36 },
		{ 387, 96, 36 },
		{ 387, 97, 36 },
		{ 387, 98, 36 },
		{ 387, 99, 36 },
		{ 387, 100, 36 },
		{ 387, 101, 36 },
		{ 387, 102, 36 },
		{ 387, 103, 36 },
		{ 387, 104, 36 },
		{ 387, 105, 36 },
		{ 387, 106, 36 },
		{ 387, 107, 36 },
		{ 387, 108, 36 },
		{ 388, 48, 36 },
		{ 388, 49, 36 },
		{ 388, 50, 36 },
		{ 388, 51, 36 },
		{ 388, 52, 36 },
		{ 388, 53, 36 },
		{ 388, 54, 36 },
		{ 388, 55, 36 },
		{ 388, 56, 36 },
		{ 388, 57, 36 },
		{ 388, 58, 36 },
		{ 388, 59, 36 },
		{ 388, 60, 36 },
		{ 388, 61, 36 },
		{ 388, 62, 36 },
		{ 388, 63, 36 },
		{ 388, 64, 36 },
		{ 388, 65, 36 },
		{ 388, 66, 36 },
		{ 388, 67, 36 },
		{ 388, 68, 36 },
		{ 388, 69, 36 },
		{ 388, 70, 36 },
		{ 388, 71, 36 },
		{ 388, 72, 36 },
		{ 388, 73, 36 },
		{ 388, 74, 36 },
		{ 388, 75, 36 },
		{ 388, 76, 36 },
		{ 388, 77, 36 },
		{ 388, 78, 36 },
		{ 388, 79, 36 },
		{ 388, 80, 36 },
		{ 388, 81, 36 },
		{ 388, 82, 36 },
		{ 388, 83, 36 },
		{ 388, 84, 36 },
		{ 388, 85, 36 },
		{ 388, 86, 36 },
		{ 388, 87, 36 },
		{ 388, 88, 36 },
		{ 388, 89, 36 },
		{ 388, 90, 36 },
		{ 388, 91, 36 },
		{ 388, 92, 36 },
		{ 388, 93, 36 },
		{ 388, 94, 36 },
		{ 388, 95, 36 },
		{ 388, 96, 36 },
		{ 388, 97, 36 },
		{ 388, 98, 36 },
		{ 388, 99, 36 },
		{ 388, 100, 36 },
		{ 388, 101, 36 },
		{ 388, 102, 36 },
		{ 388, 103, 36 },
		{ 388, 104, 36 },
		{ 388, 105, 36 },
		{ 388, 106, 36 },
		{ 388, 107, 36 },
		{ 389, 46, 36 },
		{ 389, 47, 36 },
		{ 389, 48, 36 },
		{ 389, 49, 36 },
		{ 389, 50, 36 },
		{ 389, 51, 36 },
		{ 389, 52, 36 },
		{ 389, 53, 36 },
		{ 389, 54, 36 },
		{ 389, 55, 36 },
		{ 389, 56, 36 },
		{ 389, 57, 36 },
		{ 389, 58, 36 },
		{ 389, 59, 36 },
		{ 389, 60, 36 },
		{ 389, 61, 36 },
		{ 389, 62, 36 },
		{ 389, 63, 36 },
		{ 389, 64, 36 },
		{ 389, 65, 36 },
		{ 389, 66, 36 },
		{ 389, 67, 36 },
		{ 389, 68, 36 },
		{ 389, 69, 36 },
		{ 389, 70, 36 },
		{ 389, 71, 36 },
		{ 389, 72, 36 },
		{ 389, 73, 36 },
		{ 389, 74, 36 },
		{ 389, 75, 36 },
		{ 389, 76, 36 },
		{ 389, 77, 36 },
		{ 389, 78, 36 },
		{ 389, 79, 36 },
		{ 389, 80, 36 },
		{ 389, 81, 36 },
		{ 389, 82, 36 },
		{ 389, 83, 36 },
		{ 389, 84, 36 },
		{ 389, 85, 36 },
		{ 389, 86, 36 },
		{ 389, 87, 36 },
		{ 389, 88, 36 },
		{ 389, 89, 36 },
		{ 389, 90, 36 },
		{ 389, 91, 36 },
		{ 389, 92, 36 },
		{ 389, 93, 36 },
		{ 389, 94, 36 },
		{ 389, 95, 36 },
		{ 389, 96, 36 },
		{ 389, 97, 36 },
		{ 389, 98, 36 },
		{ 389, 99, 36 },
		{ 389, 100, 36 },
		{ 389, 101, 36 },
		{ 389, 102, 36 },
		{ 389, 103, 36 },
		{ 389, 104, 36 },
		{ 389, 105, 36 },
		{ 389, 106, 36 },
		{ 390, 45, 36 },
		{ 390, 46, 36 },
		{ 390, 47, 36 },
		{ 390, 48, 36 },
		{ 390, 49, 36 },
		{ 390, 50, 36 },
		{ 390, 51, 36 },
		{ 390, 52, 36 },
		{ 390, 53, 36 },
		{ 390, 54, 36 },
		{ 390, 55, 36 },
		{ 390, 56, 36 },
		{ 390, 57, 36 },
		{ 390, 58, 36 },
		{ 390, 59, 36 },
		{ 390, 60, 36 },
		{ 390, 61, 36 },
		{ 390, 62, 36 },
		{ 390, 63, 36 },
		{ 390, 64, 36 },
		{ 390, 65, 36 },
		{ 390, 66, 36 },
		{ 390, 67, 36 },
		{ 390, 68, 36 },
		{ 390, 69, 36 },
		{ 390, 70, 36 },
		{ 390, 71, 36 },
		{ 390, 72, 36 },
		{ 390, 73, 36 },
		{ 390, 74, 36 },
		{ 390, 75, 36 },
		{ 390, 76, 36 },
		{ 390, 77, 36 },
		{ 390, 78, 36 },
		{ 390, 79, 36 },
		{ 390, 80, 36 },
		{ 390, 81, 36 },
		{ 390, 82, 36 },
		{ 390, 83, 36 },
		{ 390, 84, 36 },
		{ 390, 85, 36 },
		{ 390, 86, 36 },
		{ 390, 87, 36 },
		{ 390, 88, 36 },
		{ 390, 89, 36 },
		{ 390, 90, 36 },
		{ 390, 91, 36 },
		{ 390, 92, 36 },
		{ 390, 93, 36 },
		{ 390, 94, 36 },
		{ 390, 95, 36 },
		{ 390, 96, 36 },
		{ 390, 97, 36 },
		{ 390, 98, 36 },
		{ 390, 99, 36 },
		{ 390, 100, 36 },
		{ 390, 101, 36 },
		{ 390, 102, 36 },
		{ 390, 103, 36 },
		{ 390, 104, 36 },
		{ 391, 43, 36 },
		{ 391, 44, 36 },
		{ 391, 45, 36 },
		{ 391, 46, 36 },
		{ 391, 47, 36 },
		{ 391, 48, 36 },
		{ 391, 49, 36 },
		{ 391, 50, 36 },
		{ 391, 51, 36 },
		{ 391, 52, 36 },
		{ 391, 53, 36 },
		{ 391, 54, 36 },
		{ 391, 55, 36 },
		{ 391, 56, 36 },
		{ 391, 57, 36 },
		{ 391, 58, 36 },
		{ 391, 59, 36 },
		{ 391, 60, 36 },
		{ 391, 61, 36 },
		{ 391, 62, 36 },
		{ 391, 63, 36 },
		{ 391, 64, 36 },
		{ 391, 65, 36 },
		{ 391, 66, 36 },
		{ 391, 67, 36 },
		{ 391, 68, 36 },
		{ 391, 69, 36 },
		{ 391, 70, 36 },
		{ 391, 71, 36 },
		{ 391, 72, 36 },
		{ 391, 73, 36 },
		{ 391, 74, 36 },
		{ 391, 75, 36 },
		{ 391, 76, 36 },
		{ 391, 77, 36 },
		{ 391, 78, 36 },
		{ 391, 79, 36 },
		{ 391, 80, 36 },
		{ 391, 81, 36 },
		{ 391, 82, 36 },
		{ 391, 83, 36 },
		{ 391, 84, 36 },
		{ 391, 85, 36 },
		{ 391, 86, 36 },
		{ 391, 87, 36 },
		{ 391, 88, 36 },
		{ 391, 89, 36 },
		{ 391, 90, 36 },
		{ 391, 91, 36 },
		{ 391, 92, 36 },
		{ 391, 93, 36 },
		{ 391, 94, 36 },
		{ 391, 95, 36 },
		{ 391, 96, 36 },
		{ 391, 97, 36 },
		{ 391, 98, 36 },
		{ 391, 99, 36 },
		{ 391, 100, 36 },
		{ 391, 101, 36 },
		{ 391, 102, 36 },
		{ 391, 103, 36 },
		{ 392, 42, 36 },
		{ 392, 43, 36 },
		{ 392, 44, 36 },
		{ 392, 45, 36 },
		{ 392, 46, 36 },
		{ 392, 47, 36 },
		{ 392, 48, 36 },
		{ 392, 49, 36 },
		{ 392, 50, 36 },
		{ 392, 51, 36 },
		{ 392, 52, 36 },
		{ 392, 53, 36 },
		{ 392, 54, 36 },
		{ 392, 55, 36 },
		{ 392, 56, 36 },
		{ 392, 57, 36 },
		{ 392, 58, 36 },
		{ 392, 59, 36 },
		{ 392, 60, 36 },
		{ 392, 61, 36 },
		{ 392, 62, 36 },
		{ 392, 63, 36 },
		{ 392, 64, 36 },
		{ 392, 65, 36 },
		{ 392, 66, 36 },
		{ 392, 67, 36 },
		{ 392, 68, 36 },
		{ 392, 69, 36 },
		{ 392, 70, 36 },
		{ 392, 71, 36 },
		{ 392, 72, 36 },
		{ 392, 73, 36 },
		{ 392, 74, 36 },
		{ 392, 75, 36 },
		{ 392, 76, 36 },
		{ 392, 77, 36 },
		{ 392, 78, 36 },
		{ 392, 79, 36 },
		{ 392, 80, 36 },
		{ 392, 81, 36 },
		{ 392, 82, 36 },
		{ 392, 83, 36 },
		{ 392, 84, 36 },
		{ 392, 85, 36 },
		{ 392, 86, 36 },
		{ 392, 87, 36 },
		{ 392, 88, 36 },
		{ 392, 89, 36 },
		{ 392, 90, 36 },
		{ 392, 91, 36 },
		{ 392, 92, 36 },
		{ 392, 93, 36 },
		{ 392, 94, 36 },
		{ 392, 95, 36 },
		{ 392, 96, 36 },
		{ 392, 97, 36 },
		{ 392, 98, 36 },
		{ 392, 99, 36 },
		{ 392, 100, 36 },
		{ 392, 101, 36 },
		{ 392, 102, 36 },
		{ 393, 40, 36 },
		{ 393, 41, 36 },
		{ 393, 42, 36 },
		{ 393, 43, 36 },
		{ 393, 44, 36 },
		{ 393, 45, 36 },
		{ 393, 46, 36 },
		{ 393, 47, 36 },
		{ 393, 48, 36 },
		{ 393, 49, 36 },
		{ 393, 50, 36 },
		{ 393, 51, 36 },
		{ 393, 52, 36 },
		{ 393, 53, 36 },
		{ 393, 54, 36 },
		{ 393, 55, 36 },
		{ 393, 56, 36 },
		{ 393, 57, 36 },
		{ 393, 58, 36 },
		{ 393, 59, 36 },
		{ 393, 60, 36 },
		{ 393, 61, 36 },
		{ 393, 62, 36 },
		{ 393, 63, 36 },
		{ 393, 64, 36 },
		{ 393, 65, 36 },
		{ 393, 66, 36 },
		{ 393, 67, 36 },
		{ 393, 68, 36 },
		{ 393, 69, 36 },
		{ 393, 70, 36 },
		{ 393, 71, 36 },
		{ 393, 72, 36 },
		{ 393, 73, 36 },
		{ 393, 74, 36 },
		{ 393, 75, 36 },
		{ 393, 76, 36 },
		{ 393, 77, 36 },
		{ 393, 78, 36 },
		{ 393, 79, 36 },
		{ 393, 80, 36 },
		{ 393, 81, 36 },
		{ 393, 82, 36 },
		{ 393, 83, 36 },
		{ 393, 84, 36 },
		{ 393, 85, 36 },
		{ 393, 86, 36 },
		{ 393, 87, 36 },
		{ 393, 88, 36 },
		{ 393, 89, 36 },
		{ 393, 90, 36 },
		{ 393, 91, 36 },
		{ 393, 92, 36 },
		{ 393, 93, 36 },
		{ 393, 94, 36 },
		{ 393, 95, 36 },
		{ 393, 96, 36 },
		{ 393, 97, 36 },
		{ 393, 98, 36 },
		{ 393, 99, 36 },
		{ 393, 100, 36 },
		{ 393, 101, 36 },
		{ 394, 38, 36 },
		{ 394, 39, 36 },
		{ 394, 40, 36 },
		{ 394, 41, 36 },
		{ 394, 42, 36 },
		{ 394, 43, 36 },
		{ 394, 44, 36 },
		{ 394, 45, 36 },
		{ 394, 46, 36 },
		{ 394, 47, 36 },
		{ 394, 48, 36 },
		{ 394, 49, 36 },
		{ 394, 50, 36 },
		{ 394, 51, 36 },
		{ 394, 52, 36 },
		{ 394, 53, 36 },
		{ 394, 54, 36 },
		{ 394, 55, 36 },
		{ 394, 56, 36 },
		{ 394, 57, 36 },
		{ 394, 58, 36 },
		{ 394, 59, 36 },
		{ 394, 60, 36 },
		{ 394, 61, 36 },
		{ 394, 62, 36 },
		{ 394, 63, 36 },
		{ 394, 64, 36 },
		{ 394, 65, 36 },
		{ 394, 66, 36 },
		{ 394, 67, 36 },
		{ 394, 68, 36 },
		{ 394, 69, 36 },
		{ 394, 70, 36 },
		{ 394, 71, 36 },
		{ 394, 72, 36 },
		{ 394, 73, 36 },
		{ 394, 74, 36 },
		{ 394, 75, 36 },
		{ 394, 76, 36 },
		{ 394, 77, 36 },
		{ 394, 78, 36 },
		{ 394, 79, 36 },
		{ 394, 80, 36 },
		{ 394, 81, 36 },
		{ 394, 82, 36 },
		{ 394, 83, 36 },
		{ 394, 84, 36 },
		{ 394, 85, 36 },
		{ 394, 86, 36 },
		{ 394, 87, 36 },
		{ 394, 88, 36 },
		{ 394, 89, 36 },
		{ 394, 90, 36 },
		{ 394, 91, 36 },
		{ 394, 92, 36 },
		{ 394, 93, 36 },
		{ 394, 94, 36 },
		{ 394, 95, 36 },
		{ 394, 96, 36 },
		{ 394, 97, 36 },
		{ 394, 98, 36 },
		{ 394, 99, 36 },
		{ 394, 100, 36 },
		{ 395, 37, 36 },
		{ 395, 38, 36 },
		{ 395, 39, 36 },
		{ 395, 40, 36 },
		{ 395, 41, 36 },
		{ 395, 42, 36 },
		{ 395, 43, 36 },
		{ 395, 44, 36 },
		{ 395, 45, 36 },
		{ 395, 46, 36 },
		{ 395, 47, 36 },
		{ 395, 48, 36 },
		{ 395, 49, 36 },
		{ 395, 50, 36 },
		{ 395, 51, 36 },
		{ 395, 52, 36 },
		{ 395, 53, 36 },
		{ 395, 54, 36 },
		{ 395, 55, 36 },
		{ 395, 56, 36 },
		{ 395, 57, 36 },
		{ 395, 58, 36 },
		{ 395, 59, 36 },
		{ 395, 60, 36 },
		{ 395, 61, 36 },
		{ 395, 62, 36 },
		{ 395, 63, 36 },
		{ 395, 64, 36 },
		{ 395, 65, 36 },
		{ 395, 66, 36 },
		{ 395, 67, 36 },
		{ 395, 68, 36 },
		{ 395, 69, 36 },
		{ 395, 70, 36 },
		{ 395, 71, 36 },
		{ 395, 72, 36 },
		{ 395, 73, 36 },
		{ 395, 74, 36 },
		{ 395, 75, 36 },
		{ 395, 76, 36 },
		{ 395, 77, 36 },
		{ 395, 78, 36 },
		{ 395, 79, 36 },
		{ 395, 80, 36 },
		{ 395, 81, 36 },
		{ 395, 82, 36 },
		{ 395, 83, 36 },
		{ 395, 84, 36 },
		{ 395, 85, 36 },
		{ 395, 86, 36 },
		{ 395, 87, 36 },
		{ 395, 88, 36 },
		{ 395, 89, 36 },
		{ 395, 90, 36 },
		{ 395, 91, 36 },
		{ 395, 92, 36 },
		{ 395, 93, 36 },
		{ 395, 94, 36 },
		{ 395, 95, 36 },
		{ 395, 96, 36 },
		{ 395, 97, 36 },
		{ 395, 98, 36 },
		{ 395, 99, 36 },
		{ 396, 35, 36 },
		{ 396, 36, 36 },
		{ 396, 37, 36 },
		{ 396, 38, 36 },
		{ 396, 39, 36 },
		{ 396, 40, 36 },
		{ 396, 41, 36 },
		{ 396, 42, 36 },
		{ 396, 43, 36 },
		{ 396, 44, 36 },
		{ 396, 45, 36 },
		{ 396, 46, 36 },
		{ 396, 47, 36 },
		{ 396, 48, 36 },
		{ 396, 49, 36 },
		{ 396, 50, 36 },
		{ 396, 51, 36 },
		{ 396, 52, 36 },
		{ 396, 53, 36 },
		{ 396, 54, 36 },
		{ 396, 55, 36 },
		{ 396, 56, 36 },
		{ 396, 57, 36 },
		{ 396, 58, 36 },
		{ 396, 59, 36 },
		{ 396, 60, 36 },
		{ 396, 61, 36 },
		{ 396, 62, 36 },
		{ 396, 63, 36 },
		{ 396, 64, 36 },
		{ 396, 65, 36 },
		{ 396, 66, 36 },
		{ 396, 67, 36 },
		{ 396, 68, 36 },
		{ 396, 69, 36 },
		{ 396, 70, 36 },
		{ 396, 71, 36 },
		{ 396, 72, 36 },
		{ 396, 73, 36 },
		{ 396, 74, 36 },
		{ 396, 75, 36 },
		{ 396, 76, 36 },
		{ 396, 77, 36 },
		{ 396, 78, 36 },
		{ 396, 79, 36 },
		{ 396, 80, 36 },
		{ 396, 81, 36 },
		{ 396, 82, 36 },
		{ 396, 83, 36 },
		{ 396, 84, 36 },
		{ 396, 85, 36 },
		{ 396, 86, 36 },
		{ 396, 87, 36 },
		{ 396, 88, 36 },
		{ 396, 89, 36 },
		{ 396, 90, 36 },
		{ 396, 91, 36 },
		{ 396, 92, 36 },
		{ 396, 93, 36 },
		{ 396, 94, 36 },
		{ 396, 95, 36 },
		{ 396, 96, 36 },
		{ 396, 97, 36 },
		{ 396, 98, 36 },
		{ 397, 34, 36 },
		{ 397, 35, 36 },
		{ 397, 36, 36 },
		{ 397, 37, 36 },
		{ 397, 38, 36 },
		{ 397, 39, 36 },
		{ 397, 40, 36 },
		{ 397, 41, 36 },
		{ 397, 42, 36 },
		{ 397, 43, 36 },
		{ 397, 44, 36 },
		{ 397, 45, 36 },
		{ 397, 46, 36 },
		{ 397, 47, 36 },
		{ 397, 48, 36 },
		{ 397, 49, 36 },
		{ 397, 50, 36 },
		{ 397, 51, 36 },
		{ 397, 52, 36 },
		{ 397, 53, 36 },
		{ 397, 54, 36 },
		{ 397, 55, 36 },
		{ 397, 56, 36 },
		{ 397, 57, 36 },
		{ 397, 58, 36 },
		{ 397, 59, 36 },
		{ 397, 60, 36 },
		{ 397, 61, 36 },
		{ 397, 62, 36 },
		{ 397, 63, 36 },
		{ 397, 64, 36 },
		{ 397, 65, 36 },
		{ 397, 66, 36 },
		{ 397, 67, 36 },
		{ 397, 68, 36 },
		{ 397, 69, 36 },
		{ 397, 70, 36 },
		{ 397, 71, 36 },
		{ 397, 72, 36 },
		{ 397, 73, 36 },
		{ 397, 74, 36 },
		{ 397, 75, 36 },
		{ 397, 76, 36 },
		{ 397, 77, 36 },
		{ 397, 78, 36 },
		{ 397, 79, 36 },
		{ 397, 80, 36 },
		{ 397, 81, 36 },
		{ 397, 82, 36 },
		{ 397, 83, 36 },
		{ 397, 84, 36 },
		{ 397, 85, 36 },
		{ 397, 86, 36 },
		{ 397, 87, 36 },
		{ 397, 88, 36 },
		{ 397, 89, 36 },
		{ 397, 90, 36 },
		{ 397, 91, 36 },
		{ 397, 92, 36 },
		{ 397, 93, 36 },
		{ 397, 94, 36 },
		{ 397, 95, 36 },
		{ 397, 96, 36 },
		{ 398, 32, 36 },
		{ 398, 33, 36 },
		{ 398, 34, 36 },
		{ 398, 35, 36 },
		{ 398, 36, 36 },
		{ 398, 37, 36 },
		{ 398, 38, 36 },
		{ 398, 39, 36 },
		{ 398, 40, 36 },
		{ 398, 41, 36 },
		{ 398, 42, 36 },
		{ 398, 43, 36 },
		{ 398, 44, 36 },
		{ 398, 45, 36 },
		{ 398, 46, 36 },
		{ 398, 47, 36 },
		{ 398, 48, 36 },
		{ 398, 49, 36 },
		{ 398, 50, 36 },
		{ 398, 51, 36 },
		{ 398, 52, 36 },
		{ 398, 53, 36 },
		{ 398, 54, 36 },
		{ 398, 55, 36 },
		{ 398, 56, 36 },
		{ 398, 57, 36 },
		{ 398, 58, 36 },
		{ 398, 59, 36 },
		{ 398, 60, 36 },
		{ 398, 61, 36 },
		{ 398, 62, 36 },
		{ 398, 63, 36 },
		{ 398, 64, 36 },
		{ 398, 65, 36 },
		{ 398, 66, 36 },
		{ 398, 67, 36 },
		{ 398, 68, 36 },
		{ 398, 69, 36 },
		{ 398, 70, 36 },
		{ 398, 71, 36 },
		{ 398, 72, 36 },
		{ 398, 73, 36 },
		{ 398, 74, 36 },
		{ 398, 75, 36 },
		{ 398, 76, 36 },
		{ 398, 77, 36 },
		{ 398, 78, 36 },
		{ 398, 79, 36 },
		{ 398, 80, 36 },
		{ 398, 81, 36 },
		{ 398, 82, 36 },
		{ 398, 83, 36 },
		{ 398, 84, 36 },
		{ 398, 85, 36 },
		{ 398, 86, 36 },
		{ 398, 87, 36 },
		{ 398, 88, 36 },
		{ 398, 89, 36 },
		{ 398, 90, 36 },
		{ 398, 91, 36 },
		{ 398, 92, 36 },
		{ 398, 93, 36 },
		{ 398, 94, 36 },
		{ 398, 95, 36 },
		{ 399, 30, 36 },
		{ 399, 31, 36 },
		{ 399, 32, 36 },
		{ 399, 33, 36 },
		{ 399, 34, 36 },
		{ 399, 35, 36 },
		{ 399, 36, 36 },
		{ 399, 37, 36 },
		{ 399, 38, 36 },
		{ 399, 39, 36 },
		{ 399, 40, 36 },
		{ 399, 41, 36 },
		{ 399, 42, 36 },
		{ 399, 43, 36 },
		{ 399, 44, 36 },
		{ 399, 45, 36 },
		{ 399, 46, 36 },
		{ 399, 47, 36 },
		{ 399, 48, 36 },
		{ 399, 49, 36 },
		{ 399, 50, 36 },
		{ 399, 51, 36 },
		{ 399, 52, 36 },
		{ 399, 53, 36 },
		{ 399, 54, 36 },
		{ 399, 55, 36 },
		{ 399, 56, 36 },
		{ 399, 57, 36 },
		{ 399, 58, 36 },
		{ 399, 59, 36 },
		{ 399, 60, 36 },
		{ 399, 61, 36 },
		{ 399, 62, 36 },
		{ 399, 63, 36 },
		{ 399, 64, 36 },
		{ 399, 65, 36 },
		{ 399, 66, 36 },
		{ 399, 67, 36 },
		{ 399, 68, 36 },
		{ 399, 69, 36 },
		{ 399, 70, 36 },
		{ 399, 71, 36 },
		{ 399, 72, 36 },
		{ 399, 73, 36 },
		{ 399, 74, 36 },
		{ 399, 75, 36 },
		{ 399, 76, 36 },
		{ 399, 77, 36 },
		{ 399, 78, 36 },
		{ 399, 79, 36 },
		{ 399, 80, 36 },
		{ 399, 81, 36 },
		{ 399, 82, 36 },
		{ 399, 83, 36 },
		{ 399, 84, 36 },
		{ 399, 85, 36 },
		{ 399, 86, 36 },
		{ 399, 87, 36 },
		{ 399, 88, 36 },
		{ 399, 89, 36 },
		{ 399, 90, 36 },
		{ 399, 91, 36 },
		{ 399, 92, 36 },
		{ 399, 93, 36 },
		{ 399, 94, 36 },
		{ 400, 29, 36 },
		{ 400, 30, 36 },
		{ 400, 31, 36 },
		{ 400, 32, 36 },
		{ 400, 33, 36 },
		{ 400, 34, 36 },
		{ 400, 35, 36 },
		{ 400, 36, 36 },
		{ 400, 37, 36 },
		{ 400, 38, 36 },
		{ 400, 39, 36 },
		{ 400, 40, 36 },
		{ 400, 41, 36 },
		{ 400, 42, 36 },
		{ 400, 43, 36 },
		{ 400, 44, 36 },
		{ 400, 45, 36 },
		{ 400, 46, 36 },
		{ 400, 47, 36 },
		{ 400, 48, 36 },
		{ 400, 49, 36 },
		{ 400, 50, 36 },
		{ 400, 51, 36 },
		{ 400, 52, 36 },
		{ 400, 53, 36 },
		{ 400, 54, 36 },
		{ 400, 55, 36 },
		{ 400, 56, 36 },
		{ 400, 57, 36 },
		{ 400, 58, 36 },
		{ 400, 59, 36 },
		{ 400, 60, 36 },
		{ 400, 61, 36 },
		{ 400, 62, 36 },
		{ 400, 63, 36 },
		{ 400, 64, 36 },
		{ 400, 65, 36 },
		{ 400, 66, 36 },
		{ 400, 67, 36 },
		{ 400, 68, 36 },
		{ 400, 69, 36 },
		{ 400, 70, 36 },
		{ 400, 71, 36 },
		{ 400, 72, 36 },
		{ 400, 73, 36 },
		{ 400, 74, 36 },
		{ 400, 75, 36 },
		{ 400, 76, 36 },
		{ 400, 77, 36 },
		{ 400, 78, 36 },
		{ 400, 79, 36 },
		{ 400, 80, 36 },
		{ 400, 81, 36 },
		{ 400, 82, 36 },
		{ 400, 83, 36 },
		{ 400, 84, 36 },
		{ 400, 85, 36 },
		{ 400, 86, 36 },
		{ 400, 87, 36 },
		{ 400, 88, 36 },
		{ 400, 89, 36 },
		{ 400, 90, 36 },
		{ 400, 91, 36 },
		{ 400, 92, 36 },
		{ 400, 93, 36 },
		{ 401, 27, 36 },
		{ 401, 28, 36 },
		{ 401, 29, 36 },
		{ 401, 30, 36 },
		{ 401, 31, 36 },
		{ 401, 32, 36 },
		{ 401, 33, 36 },
		{ 401, 34, 36 },
		{ 401, 35, 36 },
		{ 401, 36, 36 },
		{ 401, 37, 36 },
		{ 401, 38, 36 },
		{ 401, 39, 36 },
		{ 401, 40, 36 },
		{ 401, 41, 36 },
		{ 401, 42, 36 },
		{ 401, 43, 36 },
		{ 401, 44, 36 },
		{ 401, 45, 36 },
		{ 401, 46, 36 },
		{ 401, 47, 36 },
		{ 401, 48, 36 },
		{ 401, 49, 36 },
		{ 401, 50, 36 },
		{ 401, 51, 36 },
		{ 401, 52, 36 },
		{ 401, 53, 36 },
		{ 401, 54, 36 },
		{ 401, 55, 36 },
		{ 401, 56, 36 },
		{ 401, 57, 36 },
		{ 401, 58, 36 },
		{ 401, 59, 36 },
		{ 401, 60, 36 },
		{ 401, 61, 36 },
		{ 401, 62, 36 },
		{ 401, 63, 36 },
		{ 401, 64, 36 },
		{ 401, 65, 36 },
		{ 401, 66, 36 },
		{ 401, 67, 36 },
		{ 401, 68, 36 },
		{ 401, 69, 36 },
		{ 401, 70, 36 },
		{ 401, 71, 36 },
		{ 401, 72, 36 },
		{ 401, 73, 36 },
		{ 401, 74, 36 },
		{ 401, 75, 36 },
		{ 401, 76, 36 },
		{ 401, 77, 36 },
		{ 401, 78, 36 },
		{ 401, 79, 36 },
		{ 401, 80, 36 },
		{ 401, 81, 36 },
		{ 401, 82, 36 },
		{ 401, 83, 36 },
		{ 401, 84, 36 },
		{ 401, 85, 36 },
		{ 401, 86, 36 },
		{ 401, 87, 36 },
		{ 401, 88, 36 },
		{ 401, 89, 36 },
		{ 401, 90, 36 },
		{ 401, 91, 36 },
		{ 401, 92, 36 },
		{ 402, 26, 36 },
		{ 402, 27, 36 },
		{ 402, 28, 36 },
		{ 402, 29, 36 },
		{ 402, 30, 36 },
		{ 402, 31, 36 },
		{ 402, 32, 36 },
		{ 402, 33, 36 },
		{ 402, 34, 36 },
		{ 402, 35, 36 },
		{ 402, 36, 36 },
		{ 402, 37, 36 },
		{ 402, 38, 36 },
		{ 402, 39, 36 },
		{ 402, 40, 36 },
		{ 402, 41, 36 },
		{ 402, 42, 36 },
		{ 402, 43, 36 },
		{ 402, 44, 36 },
		{ 402, 45, 36 },
		{ 402, 46, 36 },
		{ 402, 47, 36 },
		{ 402, 48, 36 },
		{ 402, 49, 36 },
		{ 402, 50, 36 },
		{ 402, 51, 36 },
		{ 402, 52, 36 },
		{ 402, 53, 36 },
		{ 402, 54, 36 },
		{ 402, 55, 36 },
		{ 402, 56, 36 },
		{ 402, 57, 36 },
		{ 402, 58, 36 },
		{ 402, 59, 36 },
		{ 402, 60, 36 },
		{ 402, 61, 36 },
		{ 402, 62, 36 },
		{ 402, 63, 36 },
		{ 402, 64, 36 },
		{ 402, 65, 36 },
		{ 402, 66, 36 },
		{ 402, 67, 36 },
		{ 402, 68, 36 },
		{ 402, 69, 36 },
		{ 402, 70, 36 },
		{ 402, 71, 36 },
		{ 402, 72, 36 },
		{ 402, 73, 36 },
		{ 402, 74, 36 },
		{ 402, 75, 36 },
		{ 402, 76, 36 },
		{ 402, 77, 36 },
		{ 402, 78, 36 },
		{ 402, 79, 36 },
		{ 402, 80, 36 },
		{ 402, 81, 36 },
		{ 402, 82, 36 },
		{ 402, 83, 36 },
		{ 402, 84, 36 },
		{ 402, 85, 36 },
		{ 402, 86, 36 },
		{ 402, 87, 36 },
		{ 402, 88, 36 },
		{ 402, 89, 36 },
		{ 402, 90, 36 },
		{ 402, 91, 36 },
		{ 403, 24, 36 },
		{ 403, 25, 36 },
		{ 403, 26, 36 },
		{ 403, 27, 36 },
		{ 403, 28, 36 },
		{ 403, 29, 36 },
		{ 403, 30, 36 },
		{ 403, 31, 36 },
		{ 403, 32, 36 },
		{ 403, 33, 36 },
		{ 403, 34, 36 },
		{ 403, 35, 36 },
		{ 403, 36, 36 },
		{ 403, 37, 36 },
		{ 403, 38, 36 },
		{ 403, 39, 36 },
		{ 403, 40, 36 },
		{ 403, 41, 36 },
		{ 403, 42, 36 },
		{ 403, 43, 36 },
		{ 403, 44, 36 },
		{ 403, 45, 36 },
		{ 403, 46, 36 },
		{ 403, 47, 36 },
		{ 403, 48, 36 },
		{ 403, 49, 36 },
		{ 403, 50, 36 },
		{ 403, 51, 36 },
		{ 403, 52, 36 },
		{ 403, 53, 36 },
		{ 403, 54, 36 },
		{ 403, 55, 36 },
		{ 403, 56, 36 },
		{ 403, 57, 36 },
		{ 403, 58, 36 },
		{ 403, 59, 36 },
		{ 403, 60, 36 },
		{ 403, 61, 36 },
		{ 403, 62, 36 },
		{ 403, 63, 36 },
		{ 403, 64, 36 },
		{ 403, 65, 36 },
		{ 403, 66, 36 },
		{ 403, 67, 36 },
		{ 403, 68, 36 },
		{ 403, 69, 36 },
		{ 403, 70, 36 },
		{ 403, 71, 36 },
		{ 403, 72, 36 },
		{ 403, 73, 36 },
		{ 403, 74, 36 },
		{ 403, 75, 36 },
		{ 403, 76, 36 },
		{ 403, 77, 36 },
		{ 403, 78, 36 },
		{ 403, 79, 36 },
		{ 403, 80, 36 },
		{ 403, 81, 36 },
		{ 403, 82, 36 },
		{ 403, 83, 36 },
		{ 403, 84, 36 },
		{ 403, 85, 36 },
		{ 403, 86, 36 },
		{ 403, 87, 36 },
		{ 403, 88, 36 },
		{ 403, 89, 36 },
		{ 403, 90, 36 },
		{ 404, 22, 36 },
		{ 404, 23, 36 },
		{ 404, 24, 36 },
		{ 404, 25, 36 },
		{ 404, 26, 36 },
		{ 404, 27, 36 },
		{ 404, 28, 36 },
		{ 404, 29, 36 },
		{ 404, 30, 36 },
		{ 404, 31, 36 },
		{ 404, 32, 36 },
		{ 404, 33, 36 },
		{ 404, 34, 36 },
		{ 404, 35, 36 },
		{ 404, 36, 36 },
		{ 404, 37, 36 },
		{ 404, 38, 36 },
		{ 404, 39, 36 },
		{ 404, 40, 36 },
		{ 404, 41, 36 },
		{ 404, 42, 36 },
		{ 404, 43, 36 },
		{ 404, 44, 36 },
		{ 404, 45, 36 },
		{ 404, 46, 36 },
		{ 404, 47, 36 },
		{ 404, 48, 36 },
		{ 404, 49, 36 },
		{ 404, 50, 36 },
		{ 404, 51, 36 },
		{ 404, 52, 36 },
		{ 404, 53, 36 },
		{ 404, 54, 36 },
		{ 404, 55, 36 },
		{ 404, 56, 36 },
		{ 404, 57, 36 },
		{ 404, 58, 36 },
		{ 404, 59, 36 },
		{ 404, 60, 36 },
		{ 404, 61, 36 },
		{ 404, 62, 36 },
		{ 404, 63, 36 },
		{ 404, 64, 36 },
		{ 404, 65, 36 },
		{ 404, 66, 36 },
		{ 404, 67, 36 },
		{ 404, 68, 36 },
		{ 404, 69, 36 },
		{ 404, 70, 36 },
		{ 404, 71, 36 },
		{ 404, 72, 36 },
		{ 404, 73, 36 },
		{ 404, 74, 36 },
		{ 404, 75, 36 },
		{ 404, 76, 36 },
		{ 404, 77, 36 },
		{ 404, 78, 36 },
		{ 404, 79, 36 },
		{ 404, 80, 36 },
		{ 404, 81, 36 },
		{ 404, 82, 36 },
		{ 404, 83, 36 },
		{ 404, 84, 36 },
		{ 404, 85, 36 },
		{ 404, 86, 36 },
		{ 404, 87, 36 },
		{ 404, 88, 36 },
		{ 405, 21, 36 },
		{ 405, 22, 36 },
		{ 405, 23, 36 },
		{ 405, 24, 36 },
		{ 405, 25, 36 },
		{ 405, 26, 36 },
		{ 405, 27, 36 },
		{ 405, 28, 36 },
		{ 405, 29, 36 },
		{ 405, 30, 36 },
		{ 405, 31, 36 },
		{ 405, 32, 36 },
		{ 405, 33, 36 },
		{ 405, 34, 36 },
		{ 405, 35, 36 },
		{ 405, 36, 36 },
		{ 405, 37, 36 },
		{ 405, 38, 36 },
		{ 405, 39, 36 },
		{ 405, 40, 36 },
		{ 405, 41, 36 },
		{ 405, 42, 36 },
		{ 405, 43, 36 },
		{ 405, 44, 36 },
		{ 405, 45, 36 },
		{ 405, 46, 36 },
		{ 405, 47, 36 },
		{ 405, 48, 36 },
		{ 405, 49, 36 },
		{ 405, 50, 36 },
		{ 405, 51, 36 },
		{ 405, 52, 36 },
		{ 405, 53, 36 },
		{ 405, 54, 36 },
		{ 405, 55, 36 },
		{ 405, 56, 36 },
		{ 405, 57, 36 },
		{ 405, 58, 36 },
		{ 405, 59, 36 },
		{ 405, 60, 36 },
		{ 405, 61, 36 },
		{ 405, 62, 36 },
		{ 405, 63, 36 },
		{ 405, 64, 36 },
		{ 405, 65, 36 },
		{ 405, 66, 36 },
		{ 405, 67, 36 },
		{ 405, 68, 36 },
		{ 405, 69, 36 },
		{ 405, 70, 36 },
		{ 405, 71, 36 },
		{ 405, 72, 36 },
		{ 405, 73, 36 },
		{ 405, 74, 36 },
		{ 405, 75, 36 },
		{ 405, 76, 36 },
		{ 405, 77, 36 },
		{ 405, 78, 36 },
		{ 405, 79, 36 },
		{ 405, 80, 36 },
		{ 405, 81, 36 },
		{ 405, 82, 36 },
		{ 405, 83, 36 },
		{ 405, 84, 36 },
		{ 405, 85, 36 },
		{ 405, 86, 36 },
		{ 405, 87, 36 },
		{ 406, 19, 36 },
		{ 406, 20, 36 },
		{ 406, 21, 36 },
		{ 406, 22, 36 },
		{ 406, 23, 36 },
		{ 406, 24, 36 },
		{ 406, 25, 36 },
		{ 406, 26, 36 },
		{ 406, 27, 36 },
		{ 406, 28, 36 },
		{ 406, 29, 36 },
		{ 406, 30, 36 },
		{ 406, 31, 36 },
		{ 406, 32, 36 },
		{ 406, 33, 36 },
		{ 406, 34, 36 },
		{ 406, 35, 36 },
		{ 406, 36, 36 },
		{ 406, 37, 36 },
		{ 406, 38, 36 },
		{ 406, 39, 36 },
		{ 406, 40, 36 },
		{ 406, 41, 36 },
		{ 406, 42, 36 },
		{ 406, 43, 36 },
		{ 406, 44, 36 },
		{ 406, 45, 36 },
		{ 406, 46, 36 },
		{ 406, 47, 36 },
		{ 406, 48, 36 },
		{ 406, 49, 36 },
		{ 406, 50, 36 },
		{ 406, 51, 36 },
		{ 406, 52, 36 },
		{ 406, 53, 36 },
		{ 406, 54, 36 },
		{ 406, 55, 36 },
		{ 406, 56, 36 },
		{ 406, 57, 36 },
		{ 406, 58, 36 },
		{ 406, 59, 36 },
		{ 406, 60, 36 },
		{ 406, 61, 36 },
		{ 406, 62, 36 },
		{ 406, 63, 36 },
		{ 406, 64, 36 },
		{ 406, 65, 36 },
		{ 406, 66, 36 },
		{ 406, 67, 36 },
		{ 406, 68, 36 },
		{ 406, 69, 36 },
		{ 406, 70, 36 },
		{ 406, 71, 36 },
		{ 406, 72, 36 },
		{ 406, 73, 36 },
		{ 406, 74, 36 },
		{ 406, 75, 36 },
		{ 406, 76, 36 },
		{ 406, 77, 36 },
		{ 406, 78, 36 },
		{ 406, 79, 36 },
		{ 406, 80, 36 },
		{ 406, 81, 36 },
		{ 406, 82, 36 },
		{ 406, 83, 36 },
		{ 406, 84, 36 },
		{ 406, 85, 36 },
		{ 406, 86, 36 },
		{ 407, 18, 36 },
		{ 407, 19, 36 },
		{ 407, 20, 36 },
		{ 407, 21, 36 },
		{ 407, 22, 36 },
		{ 407, 23, 36 },
		{ 407, 24, 36 },
		{ 407, 25, 36 },
		{ 407, 26, 36 },
		{ 407, 27, 36 },
		{ 407, 28, 36 },
		{ 407, 29, 36 },
		{ 407, 30, 36 },
		{ 407, 31, 36 },
		{ 407, 32, 36 },
		{ 407, 33, 36 },
		{ 407, 34, 36 },
		{ 407, 35, 36 },
		{ 407, 36, 36 },
		{ 407, 37, 36 },
		{ 407, 38, 36 },
		{ 407, 39, 36 },
		{ 407, 40, 36 },
		{ 407, 41, 36 },
		{ 407, 42, 36 },
		{ 407, 43, 36 },
		{ 407, 44, 36 },
		{ 407, 45, 36 },
		{ 407, 46, 36 },
		{ 407, 47, 36 },
		{ 407, 48, 36 },
		{ 407, 49, 36 },
		{ 407, 50, 36 },
		{ 407, 51, 36 },
		{ 407, 52, 36 },
		{ 407, 53, 36 },
		{ 407, 54, 36 },
		{ 407, 55, 36 },
		{ 407, 56, 36 },
		{ 407, 57, 36 },
		{ 407, 58, 36 },
		{ 407, 59, 36 },
		{ 407, 60, 36 },
		{ 407, 61, 36 },
		{ 407, 62, 36 },
		{ 407, 63, 36 },
		{ 407, 64, 36 },
		{ 407, 65, 36 },
		{ 407, 66, 36 },
		{ 407, 67, 36 },
		{ 407, 68, 36 },
		{ 407, 69, 36 },
		{ 407, 70, 36 },
		{ 407, 71, 36 },
		{ 407, 72, 36 },
		{ 407, 73, 36 },
		{ 407, 74, 36 },
		{ 407, 75, 36 },
		{ 407, 76, 36 },
		{ 407, 77, 36 },
		{ 407, 78, 36 },
		{ 407, 79, 36 },
		{ 407, 80, 36 },
		{ 407, 81, 36 },
		{ 407, 82, 36 },
		{ 407, 83, 36 },
		{ 407, 84, 36 },
		{ 407, 85, 36 },
		{ 408, 16, 36 },
		{ 408, 17, 36 },
		{ 408, 18, 36 },
		{ 408, 19, 36 },
		{ 408, 20, 36 },
		{ 408, 21, 36 },
		{ 408, 22, 36 },
		{ 408, 23, 36 },
		{ 408, 24, 36 },
		{ 408, 25, 36 },
		{ 408, 26, 36 },
		{ 408, 27, 36 },
		{ 408, 28, 36 },
		{ 408, 29, 36 },
		{ 408, 30, 36 },
		{ 408, 31, 36 },
		{ 408, 32, 36 },
		{ 408, 33, 36 },
		{ 408, 34, 36 },
		{ 408, 35, 36 },
		{ 408, 36, 36 },
		{ 408, 37, 36 },
		{ 408, 38, 36 },
		{ 408, 39, 36 },
		{ 408, 40, 36 },
		{ 408, 41, 36 },
		{ 408, 42, 36 },
		{ 408, 43, 36 },
		{ 408, 44, 36 },
		{ 408, 45, 36 },
		{ 408, 46, 36 },
		{ 408, 47, 36 },
		{ 408, 48, 36 },
		{ 408, 49, 36 },
		{ 408, 50, 36 },
		{ 408, 51, 36 },
		{ 408, 52, 36 },
		{ 408, 53, 36 },
		{ 408, 54, 36 },
		{ 408, 55, 36 },
		{ 408, 56, 36 },
		{ 408, 57, 36 },
		{ 408, 58, 36 },
		{ 408, 59, 36 },
		{ 408, 60, 36 },
		{ 408, 61, 36 },
		{ 408, 62, 36 },
		{ 408, 63, 36 },
		{ 408, 64, 36 },
		{ 408, 65, 36 },
		{ 408, 66, 36 },
		{ 408, 67, 36 },
		{ 408, 68, 36 },
		{ 408, 69, 36 },
		{ 408, 70, 36 },
		{ 408, 71, 36 },
		{ 408, 72, 36 },
		{ 408, 73, 36 },
		{ 408, 74, 36 },
		{ 408, 75, 36 },
		{ 408, 76, 36 },
		{ 408, 77, 36 },
		{ 408, 78, 36 },
		{ 408, 79, 36 },
		{ 408, 80, 36 },
		{ 408, 81, 36 },
		{ 408, 82, 36 },
		{ 408, 83, 36 },
		{ 408, 84, 36 },
		{ 409, 14, 36 },
		{ 409, 15, 36 },
		{ 409, 16, 36 },
		{ 409, 17, 36 },
		{ 409, 18, 36 },
		{ 409, 19, 36 },
		{ 409, 20, 36 },
		{ 409, 21, 36 },
		{ 409, 22, 36 },
		{ 409, 23, 36 },
		{ 409, 24, 36 },
		{ 409, 25, 36 },
		{ 409, 26, 36 },
		{ 409, 27, 36 },
		{ 409, 28, 36 },
		{ 409, 29, 36 },
		{ 409, 30, 36 },
		{ 409, 31, 36 },
		{ 409, 32, 36 },
		{ 409, 33, 36 },
		{ 409, 34, 36 },
		{ 409, 35, 36 },
		{ 409, 36, 36 },
		{ 409, 37, 36 },
		{ 409, 38, 36 },
		{ 409, 39, 36 },
		{ 409, 40, 36 },
		{ 409, 41, 36 },
		{ 409, 42, 36 },
		{ 409, 43, 36 },
		{ 409, 44, 36 },
		{ 409, 45, 36 },
		{ 409, 46, 36 },
		{ 409, 47, 36 },
		{ 409, 48, 36 },
		{ 409, 49, 36 },
		{ 409, 50, 36 },
		{ 409, 51, 36 },
		{ 409, 52, 36 },
		{ 409, 53, 36 },
		{ 409, 54, 36 },
		{ 409, 55, 36 },
		{ 409, 56, 36 },
		{ 409, 57, 36 },
		{ 409, 58, 36 },
		{ 409, 59, 36 },
		{ 409, 60, 36 },
		{ 409, 61, 36 },
		{ 409, 62, 36 },
		{ 409, 63, 36 },
		{ 409, 64, 36 },
		{ 409, 65, 36 },
		{ 409, 66, 36 },
		{ 409, 67, 36 },
		{ 409, 68, 36 },
		{ 409, 69, 36 },
		{ 409, 70, 36 },
		{ 409, 71, 36 },
		{ 409, 72, 36 },
		{ 409, 73, 36 },
		{ 409, 74, 36 },
		{ 409, 75, 36 },
		{ 409, 76, 36 },
		{ 409, 77, 36 },
		{ 409, 78, 36 },
		{ 409, 79, 36 },
		{ 409, 80, 36 },
		{ 409, 81, 36 },
		{ 409, 82, 36 },
		{ 409, 83, 36 },
		{ 410, 13, 36 },
		{ 410, 14, 36 },
		{ 410, 15, 36 },
		{ 410, 16, 36 },
		{ 410, 17, 36 },
		{ 410, 18, 36 },
		{ 410, 19, 36 },
		{ 410, 20, 36 },
		{ 410, 21, 36 },
		{ 410, 22, 36 },
		{ 410, 23, 36 },
		{ 410, 24, 36 },
		{ 410, 25, 36 },
		{ 410, 26, 36 },
		{ 410, 27, 36 },
		{ 410, 28, 36 },
		{ 410, 29, 36 },
		{ 410, 30, 36 },
		{ 410, 31, 36 },
		{ 410, 32, 36 },
		{ 410, 33, 36 },
		{ 410, 34, 36 },
		{ 410, 35, 36 },
		{ 410, 36, 36 },
		{ 410, 37, 36 },
		{ 410, 38, 36 },
		{ 410, 39, 36 },
		{ 410, 40, 36 },
		{ 410, 41, 36 },
		{ 410, 42, 36 },
		{ 410, 43, 36 },
		{ 410, 44, 36 },
		{ 410, 45, 36 },
		{ 410, 46, 36 },
		{ 410, 47, 36 },
		{ 410, 48, 36 },
		{ 410, 49, 36 },
		{ 410, 50, 36 },
		{ 410, 51, 36 },
		{ 410, 52, 36 },
		{ 410, 53, 36 },
		{ 410, 54, 36 },
		{ 410, 55, 36 },
		{ 410, 56, 36 },
		{ 410, 57, 36 },
		{ 410, 58, 36 },
		{ 410, 59, 36 },
		{ 410, 60, 36 },
		{ 410, 61, 36 },
		{ 410, 62, 36 },
		{ 410, 63, 36 },
		{ 410, 64, 36 },
		{ 410, 65, 36 },
		{ 410, 66, 36 },
		{ 410, 67, 36 },
		{ 410, 68, 36 },
		{ 410, 69, 36 },
		{ 410, 70, 36 },
		{ 410, 71, 36 },
		{ 410, 72, 36 },
		{ 410, 73, 36 },
		{ 410, 74, 36 },
		{ 410, 75, 36 },
		{ 410, 76, 36 },
		{ 410, 77, 36 },
		{ 410, 78, 36 },
		{ 410, 79, 36 },
		{ 410, 80, 36 },
		{ 410, 81, 36 },
		{ 410, 82, 36 },
		{ 411, 11, 36 },
		{ 411, 12, 36 },
		{ 411, 13, 36 },
		{ 411, 14, 36 },
		{ 411, 15, 36 },
		{ 411, 16, 36 },
		{ 411, 17, 36 },
		{ 411, 18, 36 },
		{ 411, 19, 36 },
		{ 411, 20, 36 },
		{ 411, 21, 36 },
		{ 411, 22, 36 },
		{ 411, 23, 36 },
		{ 411, 24, 36 },
		{ 411, 25, 36 },
		{ 411, 26, 36 },
		{ 411, 27, 36 },
		{ 411, 28, 36 },
		{ 411, 29, 36 },
		{ 411, 30, 36 },
		{ 411, 31, 36 },
		{ 411, 32, 36 },
		{ 411, 33, 36 },
		{ 411, 34, 36 },
		{ 411, 35, 36 },
		{ 411, 36, 36 },
		{ 411, 37, 36 },
		{ 411, 38, 36 },
		{ 411, 39, 36 },
		{ 411, 40, 36 },
		{ 411, 41, 36 },
		{ 411, 42, 36 },
		{ 411, 43, 36 },
		{ 411, 44, 36 },
		{ 411, 45, 36 },
		{ 411, 46, 36 },
		{ 411, 47, 36 },
		{ 411, 48, 36 },
		{ 411, 49, 36 },
		{ 411, 50, 36 },
		{ 411, 51, 36 },
		{ 411, 52, 36 },
		{ 411, 53, 36 },
		{ 411, 54, 36 },
		{ 411, 55, 36 },
		{ 411, 56, 36 },
		{ 411, 57, 36 },
		{ 411, 58, 36 },
		{ 411, 59, 36 },
		{ 411, 60, 36 },
		{ 411, 61, 36 },
		{ 411, 62, 36 },
		{ 411, 63, 36 },
		{ 411, 64, 36 },
		{ 411, 65, 36 },
		{ 411, 66, 36 },
		{ 411, 67, 36 },
		{ 411, 68, 36 },
		{ 411, 69, 36 },
		{ 411, 70, 36 },
		{ 411, 71, 36 },
		{ 411, 72, 36 },
		{ 411, 73, 36 },
		{ 411, 74, 36 },
		{ 411, 75, 36 },
		{ 411, 76, 36 },
		{ 411, 77, 36 },
		{ 411, 78, 36 },
		{ 411, 79, 36 },
		{ 411, 80, 36 },
		{ 412, 10, 36 },
		{ 412, 11, 36 },
		{ 412, 12, 36 },
		{ 412, 13, 36 },
		{ 412, 14, 36 },
		{ 412, 15, 36 },
		{ 412, 16, 36 },
		{ 412, 17, 36 },
		{ 412, 18, 36 },
		{ 412, 19, 36 },
		{ 412, 20, 36 },
		{ 412, 21, 36 },
		{ 412, 22, 36 },
		{ 412, 23, 36 },
		{ 412, 24, 36 },
		{ 412, 25, 36 },
		{ 412, 26, 36 },
		{ 412, 27, 36 },
		{ 412, 28, 36 },
		{ 412, 29, 36 },
		{ 412, 30, 36 },
		{ 412, 31, 36 },
		{ 412, 32, 36 },
		{ 412, 33, 36 },
		{ 412, 34, 36 },
		{ 412, 35, 36 },
		{ 412, 36, 36 },
		{ 412, 37, 36 },
		{ 412, 38, 36 },
		{ 412, 39, 36 },
		{ 412, 40, 36 },
		{ 412, 41, 36 },
		{ 412, 42, 36 },
		{ 412, 43, 36 },
		{ 412, 44, 36 },
		{ 412, 45, 36 },
		{ 412, 46, 36 },
		{ 412, 47, 36 },
		{ 412, 48, 36 },
		{ 412, 49, 36 },
		{ 412, 50, 36 },
		{ 412, 51, 36 },
		{ 412, 52, 36 },
		{ 412, 53, 36 },
		{ 412, 54, 36 },
		{ 412, 55, 36 },
		{ 412, 56, 36 },
		{ 412, 57, 36 },
		{ 412, 58, 36 },
		{ 412, 59, 36 },
		{ 412, 60, 36 },
		{ 412, 61, 36 },
		{ 412, 62, 36 },
		{ 412, 63, 36 },
		{ 412, 64, 36 },
		{ 412, 65, 36 },
		{ 412, 66, 36 },
		{ 412, 67, 36 },
		{ 412, 68, 36 },
		{ 412, 69, 36 },
		{ 412, 70, 36 },
		{ 412, 71, 36 },
		{ 412, 72, 36 },
		{ 412, 73, 36 },
		{ 412, 74, 36 },
		{ 412, 75, 36 },
		{ 412, 76, 36 },
		{ 412, 77, 36 },
		{ 412, 78, 36 },
		{ 412, 79, 36 },
		{ 413, 8, 36 },
		{ 413, 9, 36 },
		{ 413, 10, 36 },
		{ 413, 11, 36 },
		{ 413, 12, 36 },
		{ 413, 13, 36 },
		{ 413, 14, 36 },
		{ 413, 15, 36 },
		{ 413, 16, 36 },
		{ 413, 17, 36 },
		{ 413, 18, 36 },
		{ 413, 19, 36 },
		{ 413, 20, 36 },
		{ 413, 21, 36 },
		{ 413, 22, 36 },
		{ 413, 23, 36 },
		{ 413, 24, 36 },
		{ 413, 25, 36 },
		{ 413, 26, 36 },
		{ 413, 27, 36 },
		{ 413, 28, 36 },
		{ 413, 29, 36 },
		{ 413, 30, 36 },
		{ 413, 31, 36 },
		{ 413, 32, 36 },
		{ 413, 33, 36 },
		{ 413, 34, 36 },
		{ 413, 35, 36 },
		{ 413, 36, 36 },
		{ 413, 37, 36 },
		{ 413, 38, 36 },
		{ 413, 39, 36 },
		{ 413, 40, 36 },
		{ 413, 41, 36 },
		{ 413, 42, 36 },
		{ 413, 43, 36 },
		{ 413, 44, 36 },
		{ 413, 45, 36 },
		{ 413, 46, 36 },
		{ 413, 47, 36 },
		{ 413, 48, 36 },
		{ 413, 49, 36 },
		{ 413, 50, 36 },
		{ 413, 51, 36 },
		{ 413, 52, 36 },
		{ 413, 53, 36 },
		{ 413, 54, 36 },
		{ 413, 55, 36 },
		{ 413, 56, 36 },
		{ 413, 57, 36 },
		{ 413, 58, 36 },
		{ 413, 59, 36 },
		{ 413, 60, 36 },
		{ 413, 61, 36 },
		{ 413, 62, 36 },
		{ 413, 63, 36 },
		{ 413, 64, 36 },
		{ 413, 65, 36 },
		{ 413, 66, 36 },
		{ 413, 67, 36 },
		{ 413, 68, 36 },
		{ 413, 69, 36 },
		{ 413, 70, 36 },
		{ 413, 71, 36 },
		{ 413, 72, 36 },
		{ 413, 73, 36 },
		{ 413, 74, 36 },
		{ 413, 75, 36 },
		{ 413, 76, 36 },
		{ 413, 77, 36 },
		{ 413, 78, 36 },
		{ 414, 6, 36 },
		{ 414, 7, 36 },
		{ 414, 8, 36 },
		{ 414, 9, 36 },
		{ 414, 10, 36 },
		{ 414, 11, 36 },
		{ 414, 12, 36 },
		{ 414, 13, 36 },
		{ 414, 14, 36 },
		{ 414, 15, 36 },
		{ 414, 16, 36 },
		{ 414, 17, 36 },
		{ 414, 18, 36 },
		{ 414, 19, 36 },
		{ 414, 20, 36 },
		{ 414, 21, 36 },
		{ 414, 22, 36 },
		{ 414, 23, 36 },
		{ 414, 24, 36 },
		{ 414, 25, 36 },
		{ 414, 26, 36 },
		{ 414, 27, 36 },
		{ 414, 28, 36 },
		{ 414, 29, 36 },
		{ 414, 30, 36 },
		{ 414, 31, 36 },
		{ 414, 32, 36 },
		{ 414, 33, 36 },
		{ 414, 34, 36 },
		{ 414, 35, 36 },
		{ 414, 36, 36 },
		{ 414, 37, 36 },
		{ 414, 38, 36 },
		{ 414, 39, 36 },
		{ 414, 40, 36 },
		{ 414, 41, 36 },
		{ 414, 42, 36 },
		{ 414, 43, 36 },
		{ 414, 44, 36 },
		{ 414, 45, 36 },
		{ 414, 46, 36 },
		{ 414, 47, 36 },
		{ 414, 48, 36 },
		{ 414, 49, 36 },
		{ 414, 50, 36 },
		{ 414, 51, 36 },
		{ 414, 52, 36 },
		{ 414, 53, 36 },
		{ 414, 54, 36 },
		{ 414, 55, 36 },
		{ 414, 56, 36 },
		{ 414, 57, 36 },
		{ 414, 58, 36 },
		{ 414, 59, 36 },
		{ 414, 60, 36 },
		{ 414, 61, 36 },
		{ 414, 62, 36 },
		{ 414, 63, 36 },
		{ 414, 64, 36 },
		{ 414, 65, 36 },
		{ 414, 66, 36 },
		{ 414, 67, 36 },
		{ 414, 68, 36 },
		{ 414, 69, 36 },
		{ 414, 70, 36 },
		{ 414, 71, 36 },
		{ 414, 72, 36 },
		{ 414, 73, 36 },
		{ 414, 74, 36 },
		{ 414, 75, 36 },
		{ 414, 76, 36 },
		{ 414, 77, 36 },
		{ 415, 5, 36 },
		{ 415, 6, 36 },
		{ 415, 7, 36 },
		{ 415, 8, 36 },
		{ 415, 9, 36 },
		{ 415, 10, 36 },
		{ 415, 11, 36 },
		{ 415, 12, 36 },
		{ 415, 13, 36 },
		{ 415, 14, 36 },
		{ 415, 15, 36 },
		{ 415, 16, 36 },
		{ 415, 17, 36 },
		{ 415, 18, 36 },
		{ 415, 19, 36 },
		{ 415, 20, 36 },
		{ 415, 21, 36 },
		{ 415, 22, 36 },
		{ 415, 23, 36 },
		{ 415, 24, 36 },
		{ 415, 25, 36 },
		{ 415, 26, 36 },
		{ 415, 27, 36 },
		{ 415, 28, 36 },
		{ 415, 29, 36 },
		{ 415, 30, 36 },
		{ 415, 31, 36 },
		{ 415, 32, 36 },
		{ 415, 33, 36 },
		{ 415, 34, 36 },
		{ 415, 35, 36 },
		{ 415, 36, 36 },
		{ 415, 37, 36 },
		{ 415, 38, 36 },
		{ 415, 39, 36 },
		{ 415, 40, 36 },
		{ 415, 41, 36 },
		{ 415, 42, 36 },
		{ 415, 43, 36 },
		{ 415, 44, 36 },
		{ 415, 45, 36 },
		{ 415, 46, 36 },
		{ 415, 47, 36 },
		{ 415, 48, 36 },
		{ 415, 49, 36 },
		{ 415, 50, 36 },
		{ 415, 51, 36 },
		{ 415, 52, 36 },
		{ 415, 53, 36 },
		{ 415, 54, 36 },
		{ 415, 55, 36 },
		{ 415, 56, 36 },
		{ 415, 57, 36 },
		{ 415, 58, 36 },
		{ 415, 59, 36 },
		{ 415, 60, 36 },
		{ 415, 61, 36 },
		{ 415, 62, 36 },
		{ 415, 63, 36 },
		{ 415, 64, 36 },
		{ 415, 65, 36 },
		{ 415, 66, 36 },
		{ 415, 67, 36 },
		{ 415, 68, 36 },
		{ 415, 69, 36 },
		{ 415, 70, 36 },
		{ 415, 71, 36 },
		{ 415, 72, 36 },
		{ 415, 73, 36 },
		{ 415, 74, 36 },
		{ 415, 75, 36 },
		{ 415, 76, 36 },
		{ 416, 3, 36 },
		{ 416, 4, 36 },
		{ 416, 5, 36 },
		{ 416, 6, 36 },
		{ 416, 7, 36 },
		{ 416, 8, 36 },
		{ 416, 9, 36 },
		{ 416, 10, 36 },
		{ 416, 11, 36 },
		{ 416, 12, 36 },
		{ 416, 13, 36 },
		{ 416, 14, 36 },
		{ 416, 15, 36 },
		{ 416, 16, 36 },
		{ 416, 17, 36 },
		{ 416, 18, 36 },
		{ 416, 19, 36 },
		{ 416, 20, 36 },
		{ 416, 21, 36 },
		{ 416, 22, 36 },
		{ 416, 23, 36 },
		{ 416, 24, 36 },
		{ 416, 25, 36 },
		{ 416, 26, 36 },
		{ 416, 27, 36 },
		{ 416, 28, 36 },
		{ 416, 29, 36 },
		{ 416, 30, 36 },
		{ 416, 31, 36 },
		{ 416, 32, 36 },
		{ 416, 33, 36 },
		{ 416, 34, 36 },
		{ 416, 35, 36 },
		{ 416, 36, 36 },
		{ 416, 37, 36 },
		{ 416, 38, 36 },
		{ 416, 39, 36 },
		{ 416, 40, 36 },
		{ 416, 41, 36 },
		{ 416, 42, 36 },
		{ 416, 43, 36 },
		{ 416, 44, 36 },
		{ 416, 45, 36 },
		{ 416, 46, 36 },
		{ 416, 47, 36 },
		{ 416, 48, 36 },
		{ 416, 49, 36 },
		{ 416, 50, 36 },
		{ 416, 51, 36 },
		{ 416, 52, 36 },
		{ 416, 53, 36 },
		{ 416, 54, 36 },
		{ 416, 55, 36 },
		{ 416, 56, 36 },
		{ 416, 57, 36 },
		{ 416, 58, 36 },
		{ 416, 59, 36 },
		{ 416, 60, 36 },
		{ 416, 61, 36 },
		{ 416, 62, 36 },
		{ 416, 63, 36 },
		{ 416, 64, 36 },
		{ 416, 65, 36 },
		{ 416, 66, 36 },
		{ 416, 67, 36 },
		{ 416, 68, 36 },
		{ 416, 69, 36 },
		{ 416, 70, 36 },
		{ 416, 71, 36 },
		{ 416, 72, 36 },
		{ 416, 73, 36 },
		{ 416, 74, 36 },
		{ 416, 75, 36 },
		{ 417, 2, 36 },
		{ 417, 3, 36 },
		{ 417, 4, 36 },
		{ 417, 5, 36 },
		{ 417, 6, 36 },
		{ 417, 7, 36 },
		{ 417, 8, 36 },
		{ 417, 9, 36 },
		{ 417, 10, 36 },
		{ 417, 11, 36 },
		{ 417, 12, 36 },
		{ 417, 13, 36 },
		{ 417, 14, 36 },
		{ 417, 15, 36 },
		{ 417, 16, 36 },
		{ 417, 17, 36 },
		{ 417, 18, 36 },
		{ 417, 19, 36 },
		{ 417, 20, 36 },
		{ 417, 21, 36 },
		{ 417, 22, 36 },
		{ 417, 23, 36 },
		{ 417, 24, 36 },
		{ 417, 25, 36 },
		{ 417, 26, 36 },
		{ 417, 27, 36 },
		{ 417, 28, 36 },
		{ 417, 29, 36 },
		{ 417, 30, 36 },
		{ 417, 31, 36 },
		{ 417, 32, 36 },
		{ 417, 33, 36 },
		{ 417, 34, 36 },
		{ 417, 35, 36 },
		{ 417, 36, 36 },
		{ 417, 37, 36 },
		{ 417, 38, 36 },
		{ 417, 39, 36 },
		{ 417, 40, 36 },
		{ 417, 41, 36 },
		{ 417, 42, 36 },
		{ 417, 43, 36 },
		{ 417, 44, 36 },
		{ 417, 45, 36 },
		{ 417, 46, 36 },
		{ 417, 47, 36 },
		{ 417, 48, 36 },
		{ 417, 49, 36 },
		{ 417, 50, 36 },
		{ 417, 51, 36 },
		{ 417, 52, 36 },
		{ 417, 53, 36 },
		{ 417, 54, 36 },
		{ 417, 55, 36 },
		{ 417, 56, 36 },
		{ 417, 57, 36 },
		{ 417, 58, 36 },
		{ 417, 59, 36 },
		{ 417, 60, 36 },
		{ 417, 61, 36 },
		{ 417, 62, 36 },
		{ 417, 63, 36 },
		{ 417, 64, 36 },
		{ 417, 65, 36 },
		{ 417, 66, 36 },
		{ 417, 67, 36 },
		{ 417, 68, 36 },
		{ 417, 69, 36 },
		{ 417, 70, 36 },
		{ 417, 71, 36 },
		{ 417, 72, 36 },
		{ 417, 73, 36 },
		{ 417, 74, 36 },
		{ 418, 1, 36 },
		{ 418, 2, 36 },
		{ 418, 3, 36 },
		{ 418, 4, 36 },
		{ 418, 5, 36 },
		{ 418, 6, 36 },
		{ 418, 7, 36 },
		{ 418, 8, 36 },
		{ 418, 9, 36 },
		{ 418, 10, 36 },
		{ 418, 11, 36 },
		{ 418, 12, 36 },
		{ 418, 13, 36 },
		{ 418, 14, 36 },
		{ 418, 15, 36 },
		{ 418, 16, 36 },
		{ 418, 17, 36 },
		{ 418, 18, 36 },
		{ 418, 19, 36 },
		{ 418, 20, 36 },
		{ 418, 21, 36 },
		{ 418, 22, 36 },
		{ 418, 23, 36 },
		{ 418, 24, 36 },
		{ 418, 25, 36 },
		{ 418, 26, 36 },
		{ 418, 27, 36 },
		{ 418, 28, 36 },
		{ 418, 29, 36 },
		{ 418, 30, 36 },
		{ 418, 31, 36 },
		{ 418, 32, 36 },
		{ 418, 33, 36 },
		{ 418, 34, 36 },
		{ 418, 35, 36 },
		{ 418, 36, 36 },
		{ 418, 37, 36 },
		{ 418, 38, 36 },
		{ 418, 39, 36 },
		{ 418, 40, 36 },
		{ 418, 41, 36 },
		{ 418, 42, 36 },
		{ 418, 43, 36 },
		{ 418, 44, 36 },
		{ 418, 45, 36 },
		{ 418, 46, 36 },
		{ 418, 47, 36 },
		{ 418, 48, 36 },
		{ 418, 49, 36 },
		{ 418, 50, 36 },
		{ 418, 51, 36 },
		{ 418, 52, 36 },
		{ 418, 53, 36 },
		{ 418, 54, 36 },
		{ 418, 55, 36 },
		{ 418, 56, 36 },
		{ 418, 57, 36 },
		{ 418, 58, 36 },
		{ 418, 59, 36 },
		{ 418, 60, 36 },
		{ 418, 61, 36 },
		{ 418, 62, 36 },
		{ 418, 63, 36 },
		{ 418, 64, 36 },
		{ 418, 65, 36 },
		{ 418, 66, 36 },
		{ 418, 67, 36 },
		{ 418, 68, 36 },
		{ 418, 69, 36 },
		{ 418, 70, 36 },
		{ 418, 71, 36 },
		{ 418, 72, 36 },
		{ 419, 1, 36 },
		{ 419, 2, 36 },
		{ 419, 3, 36 },
		{ 419, 4, 36 },
		{ 419, 5, 36 },
		{ 419, 6, 36 },
		{ 419, 7, 36 },
		{ 419, 8, 36 },
		{ 419, 9, 36 },
		{ 419, 10, 36 },
		{ 419, 11, 36 },
		{ 419, 12, 36 },
		{ 419, 13, 36 },
		{ 419, 14, 36 },
		{ 419, 15, 36 },
		{ 419, 16, 36 },
		{ 419, 17, 36 },
		{ 419, 18, 36 },
		{ 419, 19, 36 },
		{ 419, 20, 36 },
		{ 419, 21, 36 },
		{ 419, 22, 36 },
		{ 419, 23, 36 },
		{ 419, 24, 36 },
		{ 419, 25, 36 },
		{ 419, 26, 36 },
		{ 419, 27, 36 },
		{ 419, 28, 36 },
		{ 419, 29, 36 },
		{ 419, 30, 36 },
		{ 419, 31, 36 },
		{ 419, 32, 36 },
		{ 419, 33, 36 },
		{ 419, 34, 36 },
		{ 419, 35, 36 },
		{ 419, 36, 36 },
		{ 419, 37, 36 },
		{ 419, 38, 36 },
		{ 419, 39, 36 },
		{ 419, 40, 36 },
		{ 419, 41, 36 },
		{ 419, 42, 36 },
		{ 419, 43, 36 },
		{ 419, 44, 36 },
		{ 419, 45, 36 },
		{ 419, 46, 36 },
		{ 419, 47, 36 },
		{ 419, 48, 36 },
		{ 419, 49, 36 },
		{ 419, 50, 36 },
		{ 419, 51, 36 },
		{ 419, 52, 36 },
		{ 419, 53, 36 },
		{ 419, 54, 36 },
		{ 419, 55, 36 },
		{ 419, 56, 36 },
		{ 419, 57, 36 },
		{ 419, 58, 36 },
		{ 419, 59, 36 },
		{ 419, 60, 36 },
		{ 419, 61, 36 },
		{ 419, 62, 36 },
		{ 419, 63, 36 },
		{ 419, 64, 36 },
		{ 419, 65, 36 },
		{ 419, 66, 36 },
		{ 419, 67, 36 },
		{ 419, 68, 36 },
		{ 419, 69, 36 },
		{ 419, 70, 36 },
		{ 419, 71, 36 },
		{ 420, 1, 36 },
		{ 420, 2, 36 },
		{ 420, 3, 36 },
		{ 420, 4, 36 },
		{ 420, 5, 36 },
		{ 420, 6, 36 },
		{ 420, 7, 36 },
		{ 420, 8, 36 },
		{ 420, 9, 36 },
		{ 420, 10, 36 },
		{ 420, 11, 36 },
		{ 420, 12, 36 },
		{ 420, 13, 36 },
		{ 420, 14, 36 },
		{ 420, 15, 36 },
		{ 420, 16, 36 },
		{ 420, 17, 36 },
		{ 420, 18, 36 },
		{ 420, 19, 36 },
		{ 420, 20, 36 },
		{ 420, 21, 36 },
		{ 420, 22, 36 },
		{ 420, 23, 36 },
		{ 420, 24, 36 },
		{ 420, 25, 36 },
		{ 420, 26, 36 },
		{ 420, 27, 36 },
		{ 420, 28, 36 },
		{ 420, 29, 36 },
		{ 420, 30, 36 },
		{ 420, 31, 36 },
		{ 420, 32, 36 },
		{ 420, 33, 36 },
		{ 420, 34, 36 },
		{ 420, 35, 36 },
		{ 420, 36, 36 },
		{ 420, 37, 36 },
		{ 420, 38, 36 },
		{ 420, 39, 36 },
		{ 420, 40, 36 },
		{ 420, 41, 36 },
		{ 420, 42, 36 },
		{ 420, 43, 36 },
		{ 420, 44, 36 },
		{ 420, 45, 36 },
		{ 420, 46, 36 },
		{ 420, 47, 36 },
		{ 420, 48, 36 },
		{ 420, 49, 36 },
		{ 420, 50, 36 },
		{ 420, 51, 36 },
		{ 420, 52, 36 },
		{ 420, 53, 36 },
		{ 420, 54, 36 },
		{ 420, 55, 36 },
		{ 420, 56, 36 },
		{ 420, 57, 36 },
		{ 420, 58, 36 },
		{ 420, 59, 36 },
		{ 420, 60, 36 },
		{ 420, 61, 36 },
		{ 420, 62, 36 },
		{ 420, 63, 36 },
		{ 420, 64, 36 },
		{ 420, 65, 36 },
		{ 420, 66, 36 },
		{ 420, 67, 36 },
		{ 420, 68, 36 },
		{ 420, 69, 36 },
		{ 420, 70, 36 },
		{ 421, 1, 36 },
		{ 421, 2, 36 },
		{ 421, 3, 36 },
		{ 421, 4, 36 },
		{ 421, 5, 36 },
		{ 421, 6, 36 },
		{ 421, 7, 36 },
		{ 421, 8, 36 },
		{ 421, 9, 36 },
		{ 421, 10, 36 },
		{ 421, 11, 36 },
		{ 421, 12, 36 },
		{ 421, 13, 36 },
		{ 421, 14, 36 },
		{ 421, 15, 36 },
		{ 421, 16, 36 },
		{ 421, 17, 36 },
		{ 421, 18, 36 },
		{ 421, 19, 36 },
		{ 421, 20, 36 },
		{ 421, 21, 36 },
		{ 421, 22, 36 },
		{ 421, 23, 36 },
		{ 421, 24, 36 },
		{ 421, 25, 36 },
		{ 421, 26, 36 },
		{ 421, 27, 36 },
		{ 421, 28, 36 },
		{ 421, 29, 36 },
		{ 421, 30, 36 },
		{ 421, 31, 36 },
		{ 421, 32, 36 },
		{ 421, 33, 36 },
		{ 421, 34, 36 },
		{ 421, 35, 36 },
		{ 421, 36, 36 },
		{ 421, 37, 36 },
		{ 421, 38, 36 },
		{ 421, 39, 36 },
		{ 421, 40, 36 },
		{ 421, 41, 36 },
		{ 421, 42, 36 },
		{ 421, 43, 36 },
		{ 421, 44, 36 },
		{ 421, 45, 36 },
		{ 421, 46, 36 },
		{ 421, 47, 36 },
		{ 421, 48, 36 },
		{ 421, 49, 36 },
		{ 421, 50, 36 },
		{ 421, 51, 36 },
		{ 421, 52, 36 },
		{ 421, 53, 36 },
		{ 421, 54, 36 },
		{ 421, 55, 36 },
		{ 421, 56, 36 },
		{ 421, 57, 36 },
		{ 421, 58, 36 },
		{ 421, 59, 36 },
		{ 421, 60, 36 },
		{ 421, 61, 36 },
		{ 421, 62, 36 },
		{ 421, 63, 36 },
		{ 421, 64, 36 },
		{ 421, 65, 36 },
		{ 421, 66, 36 },
		{ 421, 67, 36 },
		{ 421, 68, 36 },
		{ 421, 69, 36 },
		{ 422, 1, 36 },
		{ 422, 2, 36 },
		{ 422, 3, 36 },
		{ 422, 4, 36 },
		{ 422, 5, 36 },
		{ 422, 6, 36 },
		{ 422, 7, 36 },
		{ 422, 8, 36 },
		{ 422, 9, 36 },
		{ 422, 10, 36 },
		{ 422, 11, 36 },
		{ 422, 12, 36 },
		{ 422, 13, 36 },
		{ 422, 14, 36 },
		{ 422, 15, 36 },
		{ 422, 16, 36 },
		{ 422, 17, 36 },
		{ 422, 18, 36 },
		{ 422, 19, 36 },
		{ 422, 20, 36 },
		{ 422, 21, 36 },
		{ 422, 22, 36 },
		{ 422, 23, 36 },
		{ 422, 24, 36 },
		{ 422, 25, 36 },
		{ 422, 26, 36 },
		{ 422, 27, 36 },
		{ 422, 28, 36 },
		{ 422, 29, 36 },
		{ 422, 30, 36 },
		{ 422, 31, 36 },
		{ 422, 32, 36 },
		{ 422, 33, 36 },
		{ 422, 34, 36 },
		{ 422, 35, 36 },
		{ 422, 36, 36 },
		{ 422, 37, 36 },
		{ 422, 38, 36 },
		{ 422, 39, 36 },
		{ 422, 40, 36 },
		{ 422, 41, 36 },
		{ 422, 42, 36 },
		{ 422, 43, 36 },
		{ 422, 44, 36 },
		{ 422, 45, 36 },
		{ 422, 46, 36 },
		{ 422, 47, 36 },
		{ 422, 48, 36 },
		{ 422, 49, 36 },
		{ 422, 50, 36 },
		{ 422, 51, 36 },
		{ 422, 52, 36 },
		{ 422, 53, 36 },
		{ 422, 54, 36 },
		{ 422, 55, 36 },
		{ 422, 56, 36 },
		{ 422, 57, 36 },
		{ 422, 58, 36 },
		{ 422, 59, 36 },
		{ 422, 60, 36 },
		{ 422, 61, 36 },
		{ 422, 62, 36 },
		{ 422, 63, 36 },
		{ 422, 64, 36 },
		{ 422, 65, 36 },
		{ 422, 66, 36 },
		{ 422, 67, 36 },
		{ 422, 68, 36 },
		{ 423, 1, 36 },
		{ 423, 2, 36 },
		{ 423, 3, 36 },
		{ 423, 4, 36 },
		{ 423, 5, 36 },
		{ 423, 6, 36 },
		{ 423, 7, 36 },
		{ 423, 8, 36 },
		{ 423, 9, 36 },
		{ 423, 10, 36 },
		{ 423, 11, 36 },
		{ 423, 12, 36 },
		{ 423, 13, 36 },
		{ 423, 14, 36 },
		{ 423, 15, 36 },
		{ 423, 16, 36 },
		{ 423, 17, 36 },
		{ 423, 18, 36 },
		{ 423, 19, 36 },
		{ 423, 20, 36 },
		{ 423, 21, 36 },
		{ 423, 22, 36 },
		{ 423, 23, 36 },
		{ 423, 24, 36 },
		{ 423, 25, 36 },
		{ 423, 26, 36 },
		{ 423, 27, 36 },
		{ 423, 28, 36 },
		{ 423, 29, 36 },
		{ 423, 30, 36 },
		{ 423, 31, 36 },
		{ 423, 32, 36 },
		{ 423, 33, 36 },
		{ 423, 34, 36 },
		{ 423, 35, 36 },
		{ 423, 36, 36 },
		{ 423, 37, 36 },
		{ 423, 38, 36 },
		{ 423, 39, 36 },
		{ 423, 40, 36 },
		{ 423, 41, 36 },
		{ 423, 42, 36 },
		{ 423, 43, 36 },
		{ 423, 44, 36 },
		{ 423, 45, 36 },
		{ 423, 46, 36 },
		{ 423, 47, 36 },
		{ 423, 48, 36 },
		{ 423, 49, 36 },
		{ 423, 50, 36 },
		{ 423, 51, 36 },
		{ 423, 52, 36 },
		{ 423, 53, 36 },
		{ 423, 54, 36 },
		{ 423, 55, 36 },
		{ 423, 56, 36 },
		{ 423, 57, 36 },
		{ 423, 58, 36 },
		{ 423, 59, 36 },
		{ 423, 60, 36 },
		{ 423, 61, 36 },
		{ 423, 62, 36 },
		{ 423, 63, 36 },
		{ 423, 64, 36 },
		{ 423, 65, 36 },
		{ 423, 66, 36 },
		{ 423, 67, 36 },
		{ 424, 1, 36 },
		{ 424, 2, 36 },
		{ 424, 3, 36 },
		{ 424, 4, 36 },
		{ 424, 5, 36 },
		{ 424, 6, 36 },
		{ 424, 7, 36 },
		{ 424, 8, 36 },
		{ 424, 9, 36 },
		{ 424, 10, 36 },
		{ 424, 11, 36 },
		{ 424, 12, 36 },
		{ 424, 13, 36 },
		{ 424, 14, 36 },
		{ 424, 15, 36 },
		{ 424, 16, 36 },
		{ 424, 17, 36 },
		{ 424, 18, 36 },
		{ 424, 19, 36 },
		{ 424, 20, 36 },
		{ 424, 21, 36 },
		{ 424, 22, 36 },
		{ 424, 23, 36 },
		{ 424, 24, 36 },
		{ 424, 25, 36 },
		{ 424, 26, 36 },
		{ 424, 27, 36 },
		{ 424, 28, 36 },
		{ 424, 29, 36 },
		{ 424, 30, 36 },
		{ 424, 31, 36 },
		{ 424, 32, 36 },
		{ 424, 33, 36 },
		{ 424, 34, 36 },
		{ 424, 35, 36 },
		{ 424, 36, 36 },
		{ 424, 37, 36 },
		{ 424, 38, 36 },
		{ 424, 39, 36 },
		{ 424, 40, 36 },
		{ 424, 41, 36 },
		{ 424, 42, 36 },
		{ 424, 43, 36 },
		{ 424, 44, 36 },
		{ 424, 45, 36 },
		{ 424, 46, 36 },
		{ 424, 47, 36 },
		{ 424, 48, 36 },
		{ 424, 49, 36 },
		{ 424, 50, 36 },
		{ 424, 51, 36 },
		{ 424, 52, 36 },
		{ 424, 53, 36 },
		{ 424, 54, 36 },
		{ 424, 55, 36 },
		{ 424, 56, 36 },
		{ 424, 57, 36 },
		{ 424, 58, 36 },
		{ 424, 59, 36 },
		{ 424, 60, 36 },
		{ 424, 61, 36 },
		{ 424, 62, 36 },
		{ 424, 63, 36 },
		{ 424, 64, 36 },
		{ 424, 65, 36 },
		{ 424, 66, 36 },
		{ 425, 1, 36 },
		{ 425, 2, 36 },
		{ 425, 3, 36 },
		{ 425, 4, 36 },
		{ 425, 5, 36 },
		{ 425, 6, 36 },
		{ 425, 7, 36 },
		{ 425, 8, 36 },
		{ 425, 9, 36 },
		{ 425, 10, 36 },
		{ 425, 11, 36 },
		{ 425, 12, 36 },
		{ 425, 13, 36 },
		{ 425, 14, 36 },
		{ 425, 15, 36 },
		{ 425, 16, 36 },
		{ 425, 17, 36 },
		{ 425, 18, 36 },
		{ 425, 19, 36 },
		{ 425, 20, 36 },
		{ 425, 21, 36 },
		{ 425, 22, 36 },
		{ 425, 23, 36 },
		{ 425, 24, 36 },
		{ 425, 25, 36 },
		{ 425, 26, 36 },
		{ 425, 27, 36 },
		{ 425, 28, 36 },
		{ 425, 29, 36 },
		{ 425, 30, 36 },
		{ 425, 31, 36 },
		{ 425, 32, 36 },
		{ 425, 33, 36 },
		{ 425, 34, 36 },
		{ 425, 35, 36 },
		{ 425, 36, 36 },
		{ 425, 37, 36 },
		{ 425, 38, 36 },
		{ 425, 39, 36 },
		{ 425, 40, 36 },
		{ 425, 41, 36 },
		{ 425, 42, 36 },
		{ 425, 43, 36 },
		{ 425, 44, 36 },
		{ 425, 45, 36 },
		{ 425, 46, 36 },
		{ 425, 47, 36 },
		{ 425, 48, 36 },
		{ 425, 49, 36 },
		{ 425, 50, 36 },
		{ 425, 51, 36 },
		{ 425, 52, 36 },
		{ 425, 53, 36 },
		{ 425, 54, 36 },
		{ 425, 55, 36 },
		{ 425, 56, 36 },
		{ 425, 57, 36 },
		{ 425, 58, 36 },
		{ 425, 59, 36 },
		{ 425, 60, 36 },
		{ 425, 61, 36 },
		{ 425, 62, 36 },
		{ 425, 63, 36 },
		{ 425, 64, 36 },
		{ 426, 1, 36 },
		{ 426, 2, 36 },
		{ 426, 3, 36 },
		{ 426, 4, 36 },
		{ 426, 5, 36 },
		{ 426, 6, 36 },
		{ 426, 7, 36 },
		{ 426, 8, 36 },
		{ 426, 9, 36 },
		{ 426, 10, 36 },
		{ 426, 11, 36 },
		{ 426, 12, 36 },
		{ 426, 13, 36 },
		{ 426, 14, 36 },
		{ 426, 15, 36 },
		{ 426, 16, 36 },
		{ 426, 17, 36 },
		{ 426, 18, 36 },
		{ 426, 19, 36 },
		{ 426, 20, 36 },
		{ 426, 21, 36 },
		{ 426, 22, 36 },
		{ 426, 23, 36 },
		{ 426, 24, 36 },
		{ 426, 25, 36 },
		{ 426, 26, 36 },
		{ 426, 27, 36 },
		{ 426, 28, 36 },
		{ 426, 29, 36 },
		{ 426, 30, 36 },
		{ 426, 31, 36 },
		{ 426, 32, 36 },
		{ 426, 33, 36 },
		{ 426, 34, 36 },
		{ 426, 35, 36 },
		{ 426, 36, 36 },
		{ 426, 37, 36 },
		{ 426, 38, 36 },
		{ 426, 39, 36 },
		{ 426, 40, 36 },
		{ 426, 41, 36 },
		{ 426, 42, 36 },
		{ 426, 43, 36 },
		{ 426, 44, 36 },
		{ 426, 45, 36 },
		{ 426, 46, 36 },
		{ 426, 47, 36 },
		{ 426, 48, 36 },
		{ 426, 49, 36 },
		{ 426, 50, 36 },
		{ 426, 51, 36 },
		{ 426, 52, 36 },
		{ 426, 53, 36 },
		{ 426, 54, 36 },
		{ 426, 55, 36 },
		{ 426, 56, 36 },
		{ 426, 57, 36 },
		{ 426, 58, 36 },
		{ 426, 59, 36 },
		{ 426, 60, 36 },
		{ 426, 61, 36 },
		{ 426, 62, 36 },
		{ 426, 63, 36 },
		{ 427, 1, 36 },
		{ 427, 2, 36 },
		{ 427, 3, 36 },
		{ 427, 4, 36 },
		{ 427, 5, 36 },
		{ 427, 6, 36 },
		{ 427, 7, 36 },
		{ 427, 8, 36 },
		{ 427, 9, 36 },
		{ 427, 10, 36 },
		{ 427, 11, 36 },
		{ 427, 12, 36 },
		{ 427, 13, 36 },
		{ 427, 14, 36 },
		{ 427, 15, 36 },
		{ 427, 16, 36 },
		{ 427, 17, 36 },
		{ 427, 18, 36 },
		{ 427, 19, 36 },
		{ 427, 20, 36 },
		{ 427, 21, 36 },
		{ 427, 22, 36 },
		{ 427, 23, 36 },
		{ 427, 24, 36 },
		{ 427, 25, 36 },
		{ 427, 26, 36 },
		{ 427, 27, 36 },
		{ 427, 28, 36 },
		{ 427, 29, 36 },
		{ 427, 30, 36 },
		{ 427, 31, 36 },
		{ 427, 32, 36 },
		{ 427, 33, 36 },
		{ 427, 34, 36 },
		{ 427, 35, 36 },
		{ 427, 36, 36 },
		{ 427, 37, 36 },
		{ 427, 38, 36 },
		{ 427, 39, 36 },
		{ 427, 40, 36 },
		{ 427, 41, 36 },
		{ 427, 42, 36 },
		{ 427, 43, 36 },
		{ 427, 44, 36 },
		{ 427, 45, 36 },
		{ 427, 46, 36 },
		{ 427, 47, 36 },
		{ 427, 48, 36 },
		{ 427, 49, 36 },
		{ 427, 50, 36 },
		{ 427, 51, 36 },
		{ 427, 52, 36 },
		{ 427, 53, 36 },
		{ 427, 54, 36 },
		{ 427, 55, 36 },
		{ 427, 56, 36 },
		{ 427, 57, 36 },
		{ 427, 58, 36 },
		{ 427, 59, 36 },
		{ 427, 60, 36 },
		{ 427, 61, 36 },
		{ 427, 62, 36 },
		{ 428, 1, 36 },
		{ 428, 2, 36 },
		{ 428, 3, 36 },
		{ 428, 4, 36 },
		{ 428, 5, 36 },
		{ 428, 6, 36 },
		{ 428, 7, 36 },
		{ 428, 8, 36 },
		{ 428, 9, 36 },
		{ 428, 10, 36 },
		{ 428, 11, 36 },
		{ 428, 12, 36 },
		{ 428, 13, 36 },
		{ 428, 14, 36 },
		{ 428, 15, 36 },
		{ 428, 16, 36 },
		{ 428, 17, 36 },
		{ 428, 18, 36 },
		{ 428, 19, 36 },
		{ 428, 20, 36 },
		{ 428, 21, 36 },
		{ 428, 22, 36 },
		{ 428, 23, 36 },
		{ 428, 24, 36 },
		{ 428, 25, 36 },
		{ 428, 26, 36 },
		{ 428, 27, 36 },
		{ 428, 28, 36 },
		{ 428, 29, 36 },
		{ 428, 30, 36 },
		{ 428, 31, 36 },
		{ 428, 32, 36 },
		{ 428, 33, 36 },
		{ 428, 34, 36 },
		{ 428, 35, 36 },
		{ 428, 36, 36 },
		{ 428, 37, 36 },
		{ 428, 38, 36 },
		{ 428, 39, 36 },
		{ 428, 40, 36 },
		{ 428, 41, 36 },
		{ 428, 42, 36 },
		{ 428, 43, 36 },
		{ 428, 44, 36 },
		{ 428, 45, 36 },
		{ 428, 46, 36 },
		{ 428, 47, 36 },
		{ 428, 48, 36 },
		{ 428, 49, 36 },
		{ 428, 50, 36 },
		{ 428, 51, 36 },
		{ 428, 52, 36 },
		{ 428, 53, 36 },
		{ 428, 54, 36 },
		{ 428, 55, 36 },
		{ 428, 56, 36 },
		{ 428, 57, 36 },
		{ 428, 58, 36 },
		{ 428, 59, 36 },
		{ 428, 60, 36 },
		{ 428, 61, 36 },
		{ 429, 1, 36 },
		{ 429, 2, 36 },
		{ 429, 3, 36 },
		{ 429, 4, 36 },
		{ 429, 5, 36 },
		{ 429, 6, 36 },
		{ 429, 7, 36 },
		{ 429, 8, 36 },
		{ 429, 9, 36 },
		{ 429, 10, 36 },
		{ 429, 11, 36 },
		{ 429, 12, 36 },
		{ 429, 13, 36 },
		{ 429, 14, 36 },
		{ 429, 15, 36 },
		{ 429, 16, 36 },
		{ 429, 17, 36 },
		{ 429, 18, 36 },
		{ 429, 19, 36 },
		{ 429, 20, 36 },
		{ 429, 21, 36 },
		{ 429, 22, 36 },
		{ 429, 23, 36 },
		{ 429, 24, 36 },
		{ 429, 25, 36 },
		{ 429, 26, 36 },
		{ 429, 27, 36 },
		{ 429, 28, 36 },
		{ 429, 29, 36 },
		{ 429, 30, 36 },
		{ 429, 31, 36 },
		{ 429, 32, 36 },
		{ 429, 33, 36 },
		{ 429, 34, 36 },
		{ 429, 35, 36 },
		{ 429, 36, 36 },
		{ 429, 37, 36 },
		{ 429, 38, 36 },
		{ 429, 39, 36 },
		{ 429, 40, 36 },
		{ 429, 41, 36 },
		{ 429, 42, 36 },
		{ 429, 43, 36 },
		{ 429, 44, 36 },
		{ 429, 45, 36 },
		{ 429, 46, 36 },
		{ 429, 47, 36 },
		{ 429, 48, 36 },
		{ 429, 49, 36 },
		{ 429, 50, 36 },
		{ 429, 51, 36 },
		{ 429, 52, 36 },
		{ 429, 53, 36 },
		{ 429, 54, 36 },
		{ 429, 55, 36 },
		{ 429, 56, 36 },
		{ 429, 57, 36 },
		{ 429, 58, 36 },
		{ 429, 59, 36 },
		{ 429, 60, 36 },
		{ 430, 1, 36 },
		{ 430, 2, 36 },
		{ 430, 3, 36 },
		{ 430, 4, 36 },
		{ 430, 5, 36 },
		{ 430, 6, 36 },
		{ 430, 7, 36 },
		{ 430, 8, 36 },
		{ 430, 9, 36 },
		{ 430, 10, 36 },
		{ 430, 11, 36 },
		{ 430, 12, 36 },
		{ 430, 13, 36 },
		{ 430, 14, 36 },
		{ 430, 15, 36 },
		{ 430, 16, 36 },
		{ 430, 17, 36 },
		{ 430, 18, 36 },
		{ 430, 19, 36 },
		{ 430, 20, 36 },
		{ 430, 21, 36 },
		{ 430, 22, 36 },
		{ 430, 23, 36 },
		{ 430, 24, 36 },
		{ 430, 25, 36 },
		{ 430, 26, 36 },
		{ 430, 27, 36 },
		{ 430, 28, 36 },
		{ 430, 29, 36 },
		{ 430, 30, 36 },
		{ 430, 31, 36 },
		{ 430, 32, 36 },
		{ 430, 33, 36 },
		{ 430, 34, 36 },
		{ 430, 35, 36 },
		{ 430, 36, 36 },
		{ 430, 37, 36 },
		{ 430, 38, 36 },
		{ 430, 39, 36 },
		{ 430, 40, 36 },
		{ 430, 41, 36 },
		{ 430, 42, 36 },
		{ 430, 43, 36 },
		{ 430, 44, 36 },
		{ 430, 45, 36 },
		{ 430, 46, 36 },
		{ 430, 47, 36 },
		{ 430, 48, 36 },
		{ 430, 49, 36 },
		{ 430, 50, 36 },
		{ 430, 51, 36 },
		{ 430, 52, 36 },
		{ 430, 53, 36 },
		{ 430, 54, 36 },
		{ 430, 55, 36 },
		{ 430, 56, 36 },
		{ 430, 57, 36 },
		{ 430, 58, 36 },
		{ 430, 59, 36 },
		{ 431, 1, 36 },
		{ 431, 2, 36 },
		{ 431, 3, 36 },
		{ 431, 4, 36 },
		{ 431, 5, 36 },
		{ 431, 6, 36 },
		{ 431, 7, 36 },
		{ 431, 8, 36 },
		{ 431, 9, 36 },
		{ 431, 10, 36 },
		{ 431, 11, 36 },
		{ 431, 12, 36 },
		{ 431, 13, 36 },
		{ 431, 14, 36 },
		{ 431, 15, 36 },
		{ 431, 16, 36 },
		{ 431, 17, 36 },
		{ 431, 18, 36 },
		{ 431, 19, 36 },
		{ 431, 20, 36 },
		{ 431, 21, 36 },
		{ 431, 22, 36 },
		{ 431, 23, 36 },
		{ 431, 24, 36 },
		{ 431, 25, 36 },
		{ 431, 26, 36 },
		{ 431, 27, 36 },
		{ 431, 28, 36 },
		{ 431, 29, 36 },
		{ 431, 30, 36 },
		{ 431, 31, 36 },
		{ 431, 32, 36 },
		{ 431, 33, 36 },
		{ 431, 34, 36 },
		{ 431, 35, 36 },
		{ 431, 36, 36 },
		{ 431, 37, 36 },
		{ 431, 38, 36 },
		{ 431, 39, 36 },
		{ 431, 40, 36 },
		{ 431, 41, 36 },
		{ 431, 42, 36 },
		{ 431, 43, 36 },
		{ 431, 44, 36 },
		{ 431, 45, 36 },
		{ 431, 46, 36 },
		{ 431, 47, 36 },
		{ 431, 48, 36 },
		{ 431, 49, 36 },
		{ 431, 50, 36 },
		{ 431, 51, 36 },
		{ 431, 52, 36 },
		{ 431, 53, 36 },
		{ 431, 54, 36 },
		{ 431, 55, 36 },
		{ 431, 56, 36 },
		{ 431, 57, 36 },
		{ 431, 58, 36 },
		{ 432, 1, 36 },
		{ 432, 2, 36 },
		{ 432, 3, 36 },
		{ 432, 4, 36 },
		{ 432, 5, 36 },
		{ 432, 6, 36 },
		{ 432, 7, 36 },
		{ 432, 8, 36 },
		{ 432, 9, 36 },
		{ 432, 10, 36 },
		{ 432, 11, 36 },
		{ 432, 12, 36 },
		{ 432, 13, 36 },
		{ 432, 14, 36 },
		{ 432, 15, 36 },
		{ 432, 16, 36 },
		{ 432, 17, 36 },
		{ 432, 18, 36 },
		{ 432, 19, 36 },
		{ 432, 20, 36 },
		{ 432, 21, 36 },
		{ 432, 22, 36 },
		{ 432, 23, 36 },
		{ 432, 24, 36 },
		{ 432, 25, 36 },
		{ 432, 26, 36 },
		{ 432, 27, 36 },
		{ 432, 28, 36 },
		{ 432, 29, 36 },
		{ 432, 30, 36 },
		{ 432, 31, 36 },
		{ 432, 32, 36 },
		{ 432, 33, 36 },
		{ 432, 34, 36 },
		{ 432, 35, 36 },
		{ 432, 36, 36 },
		{ 432, 37, 36 },
		{ 432, 38, 36 },
		{ 432, 39, 36 },
		{ 432, 40, 36 },
		{ 432, 41, 36 },
		{ 432, 42, 36 },
		{ 432, 43, 36 },
		{ 432, 44, 36 },
		{ 432, 45, 36 },
		{ 432, 46, 36 },
		{ 432, 47, 36 },
		{ 432, 48, 36 },
		{ 432, 49, 36 },
		{ 432, 50, 36 },
		{ 432, 51, 36 },
		{ 432, 52, 36 },
		{ 432, 53, 36 },
		{ 432, 54, 36 },
		{ 432, 55, 36 },
		{ 432, 56, 36 },
		{ 433, 1, 36 },
		{ 433, 2, 36 },
		{ 433, 3, 36 },
		{ 433, 4, 36 },
		{ 433, 5, 36 },
		{ 433, 6, 36 },
		{ 433, 7, 36 },
		{ 433, 8, 36 },
		{ 433, 9, 36 },
		{ 433, 10, 36 },
		{ 433, 11, 36 },
		{ 433, 12, 36 },
		{ 433, 13, 36 },
		{ 433, 14, 36 },
		{ 433, 15, 36 },
		{ 433, 16, 36 },
		{ 433, 17, 36 },
		{ 433, 18, 36 },
		{ 433, 19, 36 },
		{ 433, 20, 36 },
		{ 433, 21, 36 },
		{ 433, 22, 36 },
		{ 433, 23, 36 },
		{ 433, 24, 36 },
		{ 433, 25, 36 },
		{ 433, 26, 36 },
		{ 433, 27, 36 },
		{ 433, 28, 36 },
		{ 433, 29, 36 },
		{ 433, 30, 36 },
		{ 433, 31, 36 },
		{ 433, 32, 36 },
		{ 433, 33, 36 },
		{ 433, 34, 36 },
		{ 433, 35, 36 },
		{ 433, 36, 36 },
		{ 433, 37, 36 },
		{ 433, 38, 36 },
		{ 433, 39, 36 },
		{ 433, 40, 36 },
		{ 433, 41, 36 },
		{ 433, 42, 36 },
		{ 433, 43, 36 },
		{ 433, 44, 36 },
		{ 433, 45, 36 },
		{ 433, 46, 36 },
		{ 433, 47, 36 },
		{ 433, 48, 36 },
		{ 433, 49, 36 },
		{ 433, 50, 36 },
		{ 433, 51, 36 },
		{ 433, 52, 36 },
		{ 433, 53, 36 },
		{ 433, 54, 36 },
		{ 433, 55, 36 },
		{ 434, 1, 36 },
		{ 434, 2, 36 },
		{ 434, 3, 36 },
		{ 434, 4, 36 },
		{ 434, 5, 36 },
		{ 434, 6, 36 },
		{ 434, 7, 36 },
		{ 434, 8, 36 },
		{ 434, 9, 36 },
		{ 434, 10, 36 },
		{ 434, 11, 36 },
		{ 434, 12, 36 },
		{ 434, 13, 36 },
		{ 434, 14, 36 },
		{ 434, 15, 36 },
		{ 434, 16, 36 },
		{ 434, 17, 36 },
		{ 434, 18, 36 },
		{ 434, 19, 36 },
		{ 434, 20, 36 },
		{ 434, 21, 36 },
		{ 434, 22, 36 },
		{ 434, 23, 36 },
		{ 434, 24, 36 },
		{ 434, 25, 36 },
		{ 434, 26, 36 },
		{ 434, 27, 36 },
		{ 434, 28, 36 },
		{ 434, 29, 36 },
		{ 434, 30, 36 },
		{ 434, 31, 36 },
		{ 434, 32, 36 },
		{ 434, 33, 36 },
		{ 434, 34, 36 },
		{ 434, 35, 36 },
		{ 434, 36, 36 },
		{ 434, 37, 36 },
		{ 434, 38, 36 },
		{ 434, 39, 36 },
		{ 434, 40, 36 },
		{ 434, 41, 36 },
		{ 434, 42, 36 },
		{ 434, 43, 36 },
		{ 434, 44, 36 },
		{ 434, 45, 36 },
		{ 434, 46, 36 },
		{ 434, 47, 36 },
		{ 434, 48, 36 },
		{ 434, 49, 36 },
		{ 434, 50, 36 },
		{ 434, 51, 36 },
		{ 434, 52, 36 },
		{ 434, 53, 36 },
		{ 434, 54, 36 },
		{ 435, 1, 36 },
		{ 435, 2, 36 },
		{ 435, 3, 36 },
		{ 435, 4, 36 },
		{ 435, 5, 36 },
		{ 435, 6, 36 },
		{ 435, 7, 36 },
		{ 435, 8, 36 },
		{ 435, 9, 36 },
		{ 435, 10, 36 },
		{ 435, 11, 36 },
		{ 435, 12, 36 },
		{ 435, 13, 36 },
		{ 435, 14, 36 },
		{ 435, 15, 36 },
		{ 435, 16, 36 },
		{ 435, 17, 36 },
		{ 435, 18, 36 },
		{ 435, 19, 36 },
		{ 435, 20, 36 },
		{ 435, 21, 36 },
		{ 435, 22, 36 },
		{ 435, 23, 36 },
		{ 435, 24, 36 },
		{ 435, 25, 36 },
		{ 435, 26, 36 },
		{ 435, 27, 36 },
		{ 435, 28, 36 },
		{ 435, 29, 36 },
		{ 435, 30, 36 },
		{ 435, 31, 36 },
		{ 435, 32, 36 },
		{ 435, 33, 36 },
		{ 435, 34, 36 },
		{ 435, 35, 36 },
		{ 435, 36, 36 },
		{ 435, 37, 36 },
		{ 435, 38, 36 },
		{ 435, 39, 36 },
		{ 435, 40, 36 },
		{ 435, 41, 36 },
		{ 435, 42, 36 },
		{ 435, 43, 36 },
		{ 435, 44, 36 },
		{ 435, 45, 36 },
		{ 435, 46, 36 },
		{ 435, 47, 36 },
		{ 435, 48, 36 },
		{ 435, 49, 36 },
		{ 435, 50, 36 },
		{ 435, 51, 36 },
		{ 435, 52, 36 },
		{ 435, 53, 36 },
		{ 436, 1, 36 },
		{ 436, 2, 36 },
		{ 436, 3, 36 },
		{ 436, 4, 36 },
		{ 436, 5, 36 },
		{ 436, 6, 36 },
		{ 436, 7, 36 },
		{ 436, 8, 36 },
		{ 436, 9, 36 },
		{ 436, 10, 36 },
		{ 436, 11, 36 },
		{ 436, 12, 36 },
		{ 436, 13, 36 },
		{ 436, 14, 36 },
		{ 436, 15, 36 },
		{ 436, 16, 36 },
		{ 436, 17, 36 },
		{ 436, 18, 36 },
		{ 436, 19, 36 },
		{ 436, 20, 36 },
		{ 436, 21, 36 },
		{ 436, 22, 36 },
		{ 436, 23, 36 },
		{ 436, 24, 36 },
		{ 436, 25, 36 },
		{ 436, 26, 36 },
		{ 436, 27, 36 },
		{ 436, 28, 36 },
		{ 436, 29, 36 },
		{ 436, 30, 36 },
		{ 436, 31, 36 },
		{ 436, 32, 36 },
		{ 436, 33, 36 },
		{ 436, 34, 36 },
		{ 436, 35, 36 },
		{ 436, 36, 36 },
		{ 436, 37, 36 },
		{ 436, 38, 36 },
		{ 436, 39, 36 },
		{ 436, 40, 36 },
		{ 436, 41, 36 },
		{ 436, 42, 36 },
		{ 436, 43, 36 },
		{ 436, 44, 36 },
		{ 436, 45, 36 },
		{ 436, 46, 36 },
		{ 436, 47, 36 },
		{ 436, 48, 36 },
		{ 436, 49, 36 },
		{ 436, 50, 36 },
		{ 436, 51, 36 },
		{ 436, 52, 36 },
		{ 437, 1, 36 },
		{ 437, 2, 36 },
		{ 437, 3, 36 },
		{ 437, 4, 36 },
		{ 437, 5, 36 },
		{ 437, 6, 36 },
		{ 437, 7, 36 },
		{ 437, 8, 36 },
		{ 437, 9, 36 },
		{ 437, 10, 36 },
		{ 437, 11, 36 },
		{ 437, 12, 36 },
		{ 437, 13, 36 },
		{ 437, 14, 36 },
		{ 437, 15, 36 },
		{ 437, 16, 36 },
		{ 437, 17, 36 },
		{ 437, 18, 36 },
		{ 437, 19, 36 },
		{ 437, 20, 36 },
		{ 437, 21, 36 },
		{ 437, 22, 36 },
		{ 437, 23, 36 },
		{ 437, 24, 36 },
		{ 437, 25, 36 },
		{ 437, 26, 36 },
		{ 437, 27, 36 },
		{ 437, 28, 36 },
		{ 437, 29, 36 },
		{ 437, 30, 36 },
		{ 437, 31, 36 },
		{ 437, 32, 36 },
		{ 437, 33, 36 },
		{ 437, 34, 36 },
		{ 437, 35, 36 },
		{ 437, 36, 36 },
		{ 437, 37, 36 },
		{ 437, 38, 36 },
		{ 437, 39, 36 },
		{ 437, 40, 36 },
		{ 437, 41, 36 },
		{ 437, 42, 36 },
		{ 437, 43, 36 },
		{ 437, 44, 36 },
		{ 437, 45, 36 },
		{ 437, 46, 36 },
		{ 437, 47, 36 },
		{ 437, 48, 36 },
		{ 437, 49, 36 },
		{ 437, 50, 36 },
		{ 437, 51, 36 },
		{ 438, 1, 36 },
		{ 438, 2, 36 },
		{ 438, 3, 36 },
		{ 438, 4, 36 },
		{ 438, 5, 36 },
		{ 438, 6, 36 },
		{ 438, 7, 36 },
		{ 438, 8, 36 },
		{ 438, 9, 36 },
		{ 438, 10, 36 },
		{ 438, 11, 36 },
		{ 438, 12, 36 },
		{ 438, 13, 36 },
		{ 438, 14, 36 },
		{ 438, 15, 36 },
		{ 438, 16, 36 },
		{ 438, 17, 36 },
		{ 438, 18, 36 },
		{ 438, 19, 36 },
		{ 438, 20, 36 },
		{ 438, 21, 36 },
		{ 438, 22, 36 },
		{ 438, 23, 36 },
		{ 438, 24, 36 },
		{ 438, 25, 36 },
		{ 438, 26, 36 },
		{ 438, 27, 36 },
		{ 438, 28, 36 },
		{ 438, 29, 36 },
		{ 438, 30, 36 },
		{ 438, 31, 36 },
		{ 438, 32, 36 },
		{ 438, 33, 36 },
		{ 438, 34, 36 },
		{ 438, 35, 36 },
		{ 438, 36, 36 },
		{ 438, 37, 36 },
		{ 438, 38, 36 },
		{ 438, 39, 36 },
		{ 438, 40, 36 },
		{ 438, 41, 36 },
		{ 438, 42, 36 },
		{ 438, 43, 36 },
		{ 438, 44, 36 },
		{ 438, 45, 36 },
		{ 438, 46, 36 },
		{ 438, 47, 36 },
		{ 438, 48, 36 },
		{ 438, 49, 36 },
		{ 438, 50, 36 },
		{ 439, 1, 36 },
		{ 439, 2, 36 },
		{ 439, 3, 36 },
		{ 439, 4, 36 },
		{ 439, 5, 36 },
		{ 439, 6, 36 },
		{ 439, 7, 36 },
		{ 439, 8, 36 },
		{ 439, 9, 36 },
		{ 439, 10, 36 },
		{ 439, 11, 36 },
		{ 439, 12, 36 },
		{ 439, 13, 36 },
		{ 439, 14, 36 },
		{ 439, 15, 36 },
		{ 439, 16, 36 },
		{ 439, 17, 36 },
		{ 439, 18, 36 },
		{ 439, 19, 36 },
		{ 439, 20, 36 },
		{ 439, 21, 36 },
		{ 439, 22, 36 },
		{ 439, 23, 36 },
		{ 439, 24, 36 },
		{ 439, 25, 36 },
		{ 439, 26, 36 },
		{ 439, 27, 36 },
		{ 439, 28, 36 },
		{ 439, 29, 36 },
		{ 439, 30, 36 },
		{ 439, 31, 36 },
		{ 439, 32, 36 },
		{ 439, 33, 36 },
		{ 439, 34, 36 },
		{ 439, 35, 36 },
		{ 439, 36, 36 },
		{ 439, 37, 36 },
		{ 439, 38, 36 },
		{ 439, 39, 36 },
		{ 439, 40, 36 },
		{ 439, 41, 36 },
		{ 439, 42, 36 },
		{ 439, 43, 36 },
		{ 439, 44, 36 },
		{ 439, 45, 36 },
		{ 439, 46, 36 },
		{ 439, 47, 36 },
		{ 439, 48, 36 },
		{ 440, 1, 36 },
		{ 440, 2, 36 },
		{ 440, 3, 36 },
		{ 440, 4, 36 },
		{ 440, 5, 36 },
		{ 440, 6, 36 },
		{ 440, 7, 36 },
		{ 440, 8, 36 },
		{ 440, 9, 36 },
		{ 440, 10, 36 },
		{ 440, 11, 36 },
		{ 440, 12, 36 },
		{ 440, 13, 36 },
		{ 440, 14, 36 },
		{ 440, 15, 36 },
		{ 440, 16, 36 },
		{ 440, 17, 36 },
		{ 440, 18, 36 },
		{ 440, 19, 36 },
		{ 440, 20, 36 },
		{ 440, 21, 36 },
		{ 440, 22, 36 },
		{ 440, 23, 36 },
		{ 440, 24, 36 },
		{ 440, 25, 36 },
		{ 440, 26, 36 },
		{ 440, 27, 36 },
		{ 440, 28, 36 },
		{ 440, 29, 36 },
		{ 440, 30, 36 },
		{ 440, 31, 36 },
		{ 440, 32, 36 },
		{ 440, 33, 36 },
		{ 440, 34, 36 },
		{ 440, 35, 36 },
		{ 440, 36, 36 },
		{ 440, 37, 36 },
		{ 440, 38, 36 },
		{ 440, 39, 36 },
		{ 440, 40, 36 },
		{ 440, 41, 36 },
		{ 440, 42, 36 },
		{ 440, 43, 36 },
		{ 440, 44, 36 },
		{ 440, 45, 36 },
		{ 440, 46, 36 },
		{ 440, 47, 36 },
		{ 441, 1, 36 },
		{ 441, 2, 36 },
		{ 441, 3, 36 },
		{ 441, 4, 36 },
		{ 441, 5, 36 },
		{ 441, 6, 36 },
		{ 441, 7, 36 },
		{ 441, 8, 36 },
		{ 441, 9, 36 },
		{ 441, 10, 36 },
		{ 441, 11, 36 },
		{ 441, 12, 36 },
		{ 441, 13, 36 },
		{ 441, 14, 36 },
		{ 441, 15, 36 },
		{ 441, 16, 36 },
		{ 441, 17, 36 },
		{ 441, 18, 36 },
		{ 441, 19, 36 },
		{ 441, 20, 36 },
		{ 441, 21, 36 },
		{ 441, 22, 36 },
		{ 441, 23, 36 },
		{ 441, 24, 36 },
		{ 441, 25, 36 },
		{ 441, 26, 36 },
		{ 441, 27, 36 },
		{ 441, 28, 36 },
		{ 441, 29, 36 },
		{ 441, 30, 36 },
		{ 441, 31, 36 },
		{ 441, 32, 36 },
		{ 441, 33, 36 },
		{ 441, 34, 36 },
		{ 441, 35, 36 },
		{ 441, 36, 36 },
		{ 441, 37, 36 },
		{ 441, 38, 36 },
		{ 441, 39, 36 },
		{ 441, 40, 36 },
		{ 441, 41, 36 },
		{ 441, 42, 36 },
		{ 441, 43, 36 },
		{ 441, 44, 36 },
		{ 441, 45, 36 },
		{ 441, 46, 36 },
		{ 442, 1, 36 },
		{ 442, 2, 36 },
		{ 442, 3, 36 },
		{ 442, 4, 36 },
		{ 442, 5, 36 },
		{ 442, 6, 36 },
		{ 442, 7, 36 },
		{ 442, 8, 36 },
		{ 442, 9, 36 },
		{ 442, 10, 36 },
		{ 442, 11, 36 },
		{ 442, 12, 36 },
		{ 442, 13, 36 },
		{ 442, 14, 36 },
		{ 442, 15, 36 },
		{ 442, 16, 36 },
		{ 442, 17, 36 },
		{ 442, 18, 36 },
		{ 442, 19, 36 },
		{ 442, 20, 36 },
		{ 442, 21, 36 },
		{ 442, 22, 36 },
		{ 442, 23, 36 },
		{ 442, 24, 36 },
		{ 442, 25, 36 },
		{ 442, 26, 36 },
		{ 442, 27, 36 },
		{ 442, 28, 36 },
		{ 442, 29, 36 },
		{ 442, 30, 36 },
		{ 442, 31, 36 },
		{ 442, 32, 36 },
		{ 442, 33, 36 },
		{ 442, 34, 36 },
		{ 442, 35, 36 },
		{ 442, 36, 36 },
		{ 442, 37, 36 },
		{ 442, 38, 36 },
		{ 442, 39, 36 },
		{ 442, 40, 36 },
		{ 442, 41, 36 },
		{ 442, 42, 36 },
		{ 442, 43, 36 },
		{ 442, 44, 36 },
		{ 442, 45, 36 },
		{ 443, 1, 36 },
		{ 443, 2, 36 },
		{ 443, 3, 36 },
		{ 443, 4, 36 },
		{ 443, 5, 36 },
		{ 443, 6, 36 },
		{ 443, 7, 36 },
		{ 443, 8, 36 },
		{ 443, 9, 36 },
		{ 443, 10, 36 },
		{ 443, 11, 36 },
		{ 443, 12, 36 },
		{ 443, 13, 36 },
		{ 443, 14, 36 },
		{ 443, 15, 36 },
		{ 443, 16, 36 },
		{ 443, 17, 36 },
		{ 443, 18, 36 },
		{ 443, 19, 36 },
		{ 443, 20, 36 },
		{ 443, 21, 36 },
		{ 443, 22, 36 },
		{ 443, 23, 36 },
		{ 443, 24, 36 },
		{ 443, 25, 36 },
		{ 443, 26, 36 },
		{ 443, 27, 36 },
		{ 443, 28, 36 },
		{ 443, 29, 36 },
		{ 443, 30, 36 },
		{ 443, 31, 36 },
		{ 443, 32, 36 },
		{ 443, 33, 36 },
		{ 443, 34, 36 },
		{ 443, 35, 36 },
		{ 443, 36, 36 },
		{ 443, 37, 36 },
		{ 443, 38, 36 },
		{ 443, 39, 36 },
		{ 443, 40, 36 },
		{ 443, 41, 36 },
		{ 443, 42, 36 },
		{ 443, 43, 36 },
		{ 443, 44, 36 },
		{ 444, 1, 36 },
		{ 444, 2, 36 },
		{ 444, 3, 36 },
		{ 444, 4, 36 },
		{ 444, 5, 36 },
		{ 444, 6, 36 },
		{ 444, 7, 36 },
		{ 444, 8, 36 },
		{ 444, 9, 36 },
		{ 444, 10, 36 },
		{ 444, 11, 36 },
		{ 444, 12, 36 },
		{ 444, 13, 36 },
		{ 444, 14, 36 },
		{ 444, 15, 36 },
		{ 444, 16, 36 },
		{ 444, 17, 36 },
		{ 444, 18, 36 },
		{ 444, 19, 36 },
		{ 444, 20, 36 },
		{ 444, 21, 36 },
		{ 444, 22, 36 },
		{ 444, 23, 36 },
		{ 444, 24, 36 },
		{ 444, 25, 36 },
		{ 444, 26, 36 },
		{ 444, 27, 36 },
		{ 444, 28, 36 },
		{ 444, 29, 36 },
		{ 444, 30, 36 },
		{ 444, 31, 36 },
		{ 444, 32, 36 },
		{ 444, 33, 36 },
		{ 444, 34, 36 },
		{ 444, 35, 36 },
		{ 444, 36, 36 },
		{ 444, 37, 36 },
		{ 444, 38, 36 },
		{ 444, 39, 36 },
		{ 444, 40, 36 },
		{ 444, 41, 36 },
		{ 444, 42, 36 },
		{ 444, 43, 36 },
		{ 445, 1, 36 },
		{ 445, 2, 36 },
		{ 445, 3, 36 },
		{ 445, 4, 36 },
		{ 445, 5, 36 },
		{ 445, 6, 36 },
		{ 445, 7, 36 },
		{ 445, 8, 36 },
		{ 445, 9, 36 },
		{ 445, 10, 36 },
		{ 445, 11, 36 },
		{ 445, 12, 36 },
		{ 445, 13, 36 },
		{ 445, 14, 36 },
		{ 445, 15, 36 },
		{ 445, 16, 36 },
		{ 445, 17, 36 },
		{ 445, 18, 36 },
		{ 445, 19, 36 },
		{ 445, 20, 36 },
		{ 445, 21, 36 },
		{ 445, 22, 36 },
		{ 445, 23, 36 },
		{ 445, 24, 36 },
		{ 445, 25, 36 },
		{ 445, 26, 36 },
		{ 445, 27, 36 },
		{ 445, 28, 36 },
		{ 445, 29, 36 },
		{ 445, 30, 36 },
		{ 445, 31, 36 },
		{ 445, 32, 36 },
		{ 445, 33, 36 },
		{ 445, 34, 36 },
		{ 445, 35, 36 },
		{ 445, 36, 36 },
		{ 445, 37, 36 },
		{ 445, 38, 36 },
		{ 445, 39, 36 },
		{ 445, 40, 36 },
		{ 445, 41, 36 },
		{ 445, 42, 36 },
		{ 446, 1, 36 },
		{ 446, 2, 36 },
		{ 446, 3, 36 },
		{ 446, 4, 36 },
		{ 446, 5, 36 },
		{ 446, 6, 36 },
		{ 446, 7, 36 },
		{ 446, 8, 36 },
		{ 446, 9, 36 },
		{ 446, 10, 36 },
		{ 446, 11, 36 },
		{ 446, 12, 36 },
		{ 446, 13, 36 },
		{ 446, 14, 36 },
		{ 446, 15, 36 },
		{ 446, 16, 36 },
		{ 446, 17, 36 },
		{ 446, 18, 36 },
		{ 446, 19, 36 },
		{ 446, 20, 36 },
		{ 446, 21, 36 },
		{ 446, 22, 36 },
		{ 446, 23, 36 },
		{ 446, 24, 36 },
		{ 446, 25, 36 },
		{ 446, 26, 36 },
		{ 446, 27, 36 },
		{ 446, 28, 36 },
		{ 446, 29, 36 },
		{ 446, 30, 36 },
		{ 446, 31, 36 },
		{ 446, 32, 36 },
		{ 446, 33, 36 },
		{ 446, 34, 36 },
		{ 446, 35, 36 },
		{ 446, 36, 36 },
		{ 446, 37, 36 },
		{ 446, 38, 36 },
		{ 446, 39, 36 },
		{ 446, 40, 36 },
		{ 447, 1, 36 },
		{ 447, 2, 36 },
		{ 447, 3, 36 },
		{ 447, 4, 36 },
		{ 447, 5, 36 },
		{ 447, 6, 36 },
		{ 447, 7, 36 },
		{ 447, 8, 36 },
		{ 447, 9, 36 },
		{ 447, 10, 36 },
		{ 447, 11, 36 },
		{ 447, 12, 36 },
		{ 447, 13, 36 },
		{ 447, 14, 36 },
		{ 447, 15, 36 },
		{ 447, 16, 36 },
		{ 447, 17, 36 },
		{ 447, 18, 36 },
		{ 447, 19, 36 },
		{ 447, 20, 36 },
		{ 447, 21, 36 },
		{ 447, 22, 36 },
		{ 447, 23, 36 },
		{ 447, 24, 36 },
		{ 447, 25, 36 },
		{ 447, 26, 36 },
		{ 447, 27, 36 },
		{ 447, 28, 36 },
		{ 447, 29, 36 },
		{ 447, 30, 36 },
		{ 447, 31, 36 },
		{ 447, 32, 36 },
		{ 447, 33, 36 },
		{ 447, 34, 36 },
		{ 447, 35, 36 },
		{ 447, 36, 36 },
		{ 447, 37, 36 },
		{ 447, 38, 36 },
		{ 447, 39, 36 },
		{ 448, 1, 36 },
		{ 448, 2, 36 },
		{ 448, 3, 36 },
		{ 448, 4, 36 },
		{ 448, 5, 36 },
		{ 448, 6, 36 },
		{ 448, 7, 36 },
		{ 448, 8, 36 },
		{ 448, 9, 36 },
		{ 448, 10, 36 },
		{ 448, 11, 36 },
		{ 448, 12, 36 },
		{ 448, 13, 36 },
		{ 448, 14, 36 },
		{ 448, 15, 36 },
		{ 448, 16, 36 },
		{ 448, 17, 36 },
		{ 448, 18, 36 },
		{ 448, 19, 36 },
		{ 448, 20, 36 },
		{ 448, 21, 36 },
		{ 448, 22, 36 },
		{ 448, 23, 36 },
		{ 448, 24, 36 },
		{ 448, 25, 36 },
		{ 448, 26, 36 },
		{ 448, 27, 36 },
		{ 448, 28, 36 },
		{ 448, 29, 36 },
		{ 448, 30, 36 },
		{ 448, 31, 36 },
		{ 448, 32, 36 },
		{ 448, 33, 36 },
		{ 448, 34, 36 },
		{ 448, 35, 36 },
		{ 448, 36, 36 },
		{ 448, 37, 36 },
		{ 448, 38, 36 },
		{ 449, 1, 36 },
		{ 449, 2, 36 },
		{ 449, 3, 36 },
		{ 449, 4, 36 },
		{ 449, 5, 36 },
		{ 449, 6, 36 },
		{ 449, 7, 36 },
		{ 449, 8, 36 },
		{ 449, 9, 36 },
		{ 449, 10, 36 },
		{ 449, 11, 36 },
		{ 449, 12, 36 },
		{ 449, 13, 36 },
		{ 449, 14, 36 },
		{ 449, 15, 36 },
		{ 449, 16, 36 },
		{ 449, 17, 36 },
		{ 449, 18, 36 },
		{ 449, 19, 36 },
		{ 449, 20, 36 },
		{ 449, 21, 36 },
		{ 449, 22, 36 },
		{ 449, 23, 36 },
		{ 449, 24, 36 },
		{ 449, 25, 36 },
		{ 449, 26, 36 },
		{ 449, 27, 36 },
		{ 449, 28, 36 },
		{ 449, 29, 36 },
		{ 449, 30, 36 },
		{ 449, 31, 36 },
		{ 449, 32, 36 },
		{ 449, 33, 36 },
		{ 449, 34, 36 },
		{ 449, 35, 36 },
		{ 449, 36, 36 },
		{ 449, 37, 36 },
		{ 450, 1, 36 },
		{ 450, 2, 36 },
		{ 450, 3, 36 },
		{ 450, 4, 36 },
		{ 450, 5, 36 },
		{ 450, 6, 36 },
		{ 450, 7, 36 },
		{ 450, 8, 36 },
		{ 450, 9, 36 },
		{ 450, 10, 36 },
		{ 450, 11, 36 },
		{ 450, 12, 36 },
		{ 450, 13, 36 },
		{ 450, 14, 36 },
		{ 450, 15, 36 },
		{ 450, 16, 36 },
		{ 450, 17, 36 },
		{ 450, 18, 36 },
		{ 450, 19, 36 },
		{ 450, 20, 36 },
		{ 450, 21, 36 },
		{ 450, 22, 36 },
		{ 450, 23, 36 },
		{ 450, 24, 36 },
		{ 450, 25, 36 },
		{ 450, 26, 36 },
		{ 450, 27, 36 },
		{ 450, 28, 36 },
		{ 450, 29, 36 },
		{ 450, 30, 36 },
		{ 450, 31, 36 },
		{ 450, 32, 36 },
		{ 450, 33, 36 },
		{ 450, 34, 36 },
		{ 450, 35, 36 },
		{ 450, 36, 36 },
		{ 451, 1, 36 },
		{ 451, 2, 36 },
		{ 451, 3, 36 },
		{ 451, 4, 36 },
		{ 451, 5, 36 },
		{ 451, 6, 36 },
		{ 451, 7, 36 },
		{ 451, 8, 36 },
		{ 451, 9, 36 },
		{ 451, 10, 36 },
		{ 451, 11, 36 },
		{ 451, 12, 36 },
		{ 451, 13, 36 },
		{ 451, 14, 36 },
		{ 451, 15, 36 },
		{ 451, 16, 36 },
		{ 451, 17, 36 },
		{ 451, 18, 36 },
		{ 451, 19, 36 },
		{ 451, 20, 36 },
		{ 451, 21, 36 },
		{ 451, 22, 36 },
		{ 451, 23, 36 },
		{ 451, 24, 36 },
		{ 451, 25, 36 },
		{ 451, 26, 36 },
		{ 451, 27, 36 },
		{ 451, 28, 36 },
		{ 451, 29, 36 },
		{ 451, 30, 36 },
		{ 451, 31, 36 },
		{ 451, 32, 36 },
		{ 451, 33, 36 },
		{ 451, 34, 36 },
		{ 451, 35, 36 },
		{ 452, 1, 36 },
		{ 452, 2, 36 },
		{ 452, 3, 36 },
		{ 452, 4, 36 },
		{ 452, 5, 36 },
		{ 452, 6, 36 },
		{ 452, 7, 36 },
		{ 452, 8, 36 },
		{ 452, 9, 36 },
		{ 452, 10, 36 },
		{ 452, 11, 36 },
		{ 452, 12, 36 },
		{ 452, 13, 36 },
		{ 452, 14, 36 },
		{ 452, 15, 36 },
		{ 452, 16, 36 },
		{ 452, 17, 36 },
		{ 452, 18, 36 },
		{ 452, 19, 36 },
		{ 452, 20, 36 },
		{ 452, 21, 36 },
		{ 452, 22, 36 },
		{ 452, 23, 36 },
		{ 452, 24, 36 },
		{ 452, 25, 36 },
		{ 452, 26, 36 },
		{ 452, 27, 36 },
		{ 452, 28, 36 },
		{ 452, 29, 36 },
		{ 452, 30, 36 },
		{ 452, 31, 36 },
		{ 452, 32, 36 },
		{ 452, 33, 36 },
		{ 452, 34, 36 },
		{ 453, 1, 36 },
		{ 453, 2, 36 },
		{ 453, 3, 36 },
		{ 453, 4, 36 },
		{ 453, 5, 36 },
		{ 453, 6, 36 },
		{ 453, 7, 36 },
		{ 453, 8, 36 },
		{ 453, 9, 36 },
		{ 453, 10, 36 },
		{ 453, 11, 36 },
		{ 453, 12, 36 },
		{ 453, 13, 36 },
		{ 453, 14, 36 },
		{ 453, 15, 36 },
		{ 453, 16, 36 },
		{ 453, 17, 36 },
		{ 453, 18, 36 },
		{ 453, 19, 36 },
		{ 453, 20, 36 },
		{ 453, 21, 36 },
		{ 453, 22, 36 },
		{ 453, 23, 36 },
		{ 453, 24, 36 },
		{ 453, 25, 36 },
		{ 453, 26, 36 },
		{ 453, 27, 36 },
		{ 453, 28, 36 },
		{ 453, 29, 36 },
		{ 453, 30, 36 },
		{ 453, 31, 36 },
		{ 453, 32, 36 },
		{ 454, 1, 36 },
		{ 454, 2, 36 },
		{ 454, 3, 36 },
		{ 454, 4, 36 },
		{ 454, 5, 36 },
		{ 454, 6, 36 },
		{ 454, 7, 36 },
		{ 454, 8, 36 },
		{ 454, 9, 36 },
		{ 454, 10, 36 },
		{ 454, 11, 36 },
		{ 454, 12, 36 },
		{ 454, 13, 36 },
		{ 454, 14, 36 },
		{ 454, 15, 36 },
		{ 454, 16, 36 },
		{ 454, 17, 36 },
		{ 454, 18, 36 },
		{ 454, 19, 36 },
		{ 454, 20, 36 },
		{ 454, 21, 36 },
		{ 454, 22, 36 },
		{ 454, 23, 36 },
		{ 454, 24, 36 },
		{ 454, 25, 36 },
		{ 454, 26, 36 },
		{ 454, 27, 36 },
		{ 454, 28, 36 },
		{ 454, 29, 36 },
		{ 454, 30, 36 },
		{ 454, 31, 36 },
		{ 455, 1, 36 },
		{ 455, 2, 36 },
		{ 455, 3, 36 },
		{ 455, 4, 36 },
		{ 455, 5, 36 },
		{ 455, 6, 36 },
		{ 455, 7, 36 },
		{ 455, 8, 36 },
		{ 455, 9, 36 },
		{ 455, 10, 36 },
		{ 455, 11, 36 },
		{ 455, 12, 36 },
		{ 455, 13, 36 },
		{ 455, 14, 36 },
		{ 455, 15, 36 },
		{ 455, 16, 36 },
		{ 455, 17, 36 },
		{ 455, 18, 36 },
		{ 455, 19, 36 },
		{ 455, 20, 36 },
		{ 455, 21, 36 },
		{ 455, 22, 36 },
		{ 455, 23, 36 },
		{ 455, 24, 36 },
		{ 455, 25, 36 },
		{ 455, 26, 36 },
		{ 455, 27, 36 },
		{ 455, 28, 36 },
		{ 455, 29, 36 },
		{ 455, 30, 36 },
		{ 456, 1, 36 },
		{ 456, 2, 36 },
		{ 456, 3, 36 },
		{ 456, 4, 36 },
		{ 456, 5, 36 },
		{ 456, 6, 36 },
		{ 456, 7, 36 },
		{ 456, 8, 36 },
		{ 456, 9, 36 },
		{ 456, 10, 36 },
		{ 456, 11, 36 },
		{ 456, 12, 36 },
		{ 456, 13, 36 },
		{ 456, 14, 36 },
		{ 456, 15, 36 },
		{ 456, 16, 36 },
		{ 456, 17, 36 },
		{ 456, 18, 36 },
		{ 456, 19, 36 },
		{ 456, 20, 36 },
		{ 456, 21, 36 },
		{ 456, 22, 36 },
		{ 456, 23, 36 },
		{ 456, 24, 36 },
		{ 456, 25, 36 },
		{ 456, 26, 36 },
		{ 456, 27, 36 },
		{ 456, 28, 36 },
		{ 456, 29, 36 },
		{ 457, 1, 36 },
		{ 457, 2, 36 },
		{ 457, 3, 36 },
		{ 457, 4, 36 },
		{ 457, 5, 36 },
		{ 457, 6, 36 },
		{ 457, 7, 36 },
		{ 457, 8, 36 },
		{ 457, 9, 36 },
		{ 457, 10, 36 },
		{ 457, 11, 36 },
		{ 457, 12, 36 },
		{ 457, 13, 36 },
		{ 457, 14, 36 },
		{ 457, 15, 36 },
		{ 457, 16, 36 },
		{ 457, 17, 36 },
		{ 457, 18, 36 },
		{ 457, 19, 36 },
		{ 457, 20, 36 },
		{ 457, 21, 36 },
		{ 457, 22, 36 },
		{ 457, 23, 36 },
		{ 457, 24, 36 },
		{ 457, 25, 36 },
		{ 457, 26, 36 },
		{ 457, 27, 36 },
		{ 457, 28, 36 },
		{ 458, 1, 36 },
		{ 458, 2, 36 },
		{ 458, 3, 36 },
		{ 458, 4, 36 },
		{ 458, 5, 36 },
		{ 458, 6, 36 },
		{ 458, 7, 36 },
		{ 458, 8, 36 },
		{ 458, 9, 36 },
		{ 458, 10, 36 },
		{ 458, 11, 36 },
		{ 458, 12, 36 },
		{ 458, 13, 36 },
		{ 458, 14, 36 },
		{ 458, 15, 36 },
		{ 458, 16, 36 },
		{ 458, 17, 36 },
		{ 458, 18, 36 },
		{ 458, 19, 36 },
		{ 458, 20, 36 },
		{ 458, 21, 36 },
		{ 458, 22, 36 },
		{ 458, 23, 36 },
		{ 458, 24, 36 },
		{ 458, 25, 36 },
		{ 458, 26, 36 },
		{ 458, 27, 36 },
		{ 459, 1, 36 },
		{ 459, 2, 36 },
		{ 459, 3, 36 },
		{ 459, 4, 36 },
		{ 459, 5, 36 },
		{ 459, 6, 36 },
		{ 459, 7, 36 },
		{ 459, 8, 36 },
		{ 459, 9, 36 },
		{ 459, 10, 36 },
		{ 459, 11, 36 },
		{ 459, 12, 36 },
		{ 459, 13, 36 },
		{ 459, 14, 36 },
		{ 459, 15, 36 },
		{ 459, 16, 36 },
		{ 459, 17, 36 },
		{ 459, 18, 36 },
		{ 459, 19, 36 },
		{ 459, 20, 36 },
		{ 459, 21, 36 },
		{ 459, 22, 36 },
		{ 459, 23, 36 },
		{ 459, 24, 36 },
		{ 459, 25, 36 },
		{ 459, 26, 36 },
		{ 460, 1, 36 },
		{ 460, 2, 36 },
		{ 460, 3, 36 },
		{ 460, 4, 36 },
		{ 460, 5, 36 },
		{ 460, 6, 36 },
		{ 460, 7, 36 },
		{ 460, 8, 36 },
		{ 460, 9, 36 },
		{ 460, 10, 36 },
		{ 460, 11, 36 },
		{ 460, 12, 36 },
		{ 460, 13, 36 },
		{ 460, 14, 36 },
		{ 460, 15, 36 },
		{ 460, 16, 36 },
		{ 460, 17, 36 },
		{ 460, 18, 36 },
		{ 460, 19, 36 },
		{ 460, 20, 36 },
		{ 460, 21, 36 },
		{ 460, 22, 36 },
		{ 460, 23, 36 },
		{ 460, 24, 36 },
		{ 461, 1, 36 },
		{ 461, 2, 36 },
		{ 461, 3, 36 },
		{ 461, 4, 36 },
		{ 461, 5, 36 },
		{ 461, 6, 36 },
		{ 461, 7, 36 },
		{ 461, 8, 36 },
		{ 461, 9, 36 },
		{ 461, 10, 36 },
		{ 461, 11, 36 },
		{ 461, 12, 36 },
		{ 461, 13, 36 },
		{ 461, 14, 36 },
		{ 461, 15, 36 },
		{ 461, 16, 36 },
		{ 461, 17, 36 },
		{ 461, 18, 36 },
		{ 461, 19, 36 },
		{ 461, 20, 36 },
		{ 461, 21, 36 },
		{ 461, 22, 36 },
		{ 461, 23, 36 },
		{ 462, 1, 36 },
		{ 462, 2, 36 },
		{ 462, 3, 36 },
		{ 462, 4, 36 },
		{ 462, 5, 36 },
		{ 462, 6, 36 },
		{ 462, 7, 36 },
		{ 462, 8, 36 },
		{ 462, 9, 36 },
		{ 462, 10, 36 },
		{ 462, 11, 36 },
		{ 462, 12, 36 },
		{ 462, 13, 36 },
		{ 462, 14, 36 },
		{ 462, 15, 36 },
		{ 462, 16, 36 },
		{ 462, 17, 36 },
		{ 462, 18, 36 },
		{ 462, 19, 36 },
		{ 462, 20, 36 },
		{ 462, 21, 36 },
		{ 462, 22, 36 },
		{ 463, 1, 36 },
		{ 463, 2, 36 },
		{ 463, 3, 36 },
		{ 463, 4, 36 },
		{ 463, 5, 36 },
		{ 463, 6, 36 },
		{ 463, 7, 36 },
		{ 463, 8, 36 },
		{ 463, 9, 36 },
		{ 463, 10, 36 },
		{ 463, 11, 36 },
		{ 463, 12, 36 },
		{ 463, 13, 36 },
		{ 463, 14, 36 },
		{ 463, 15, 36 },
		{ 463, 16, 36 },
		{ 463, 17, 36 },
		{ 463, 18, 36 },
		{ 463, 19, 36 },
		{ 463, 20, 36 },
		{ 463, 21, 36 },
		{ 464, 1, 36 },
		{ 464, 2, 36 },
		{ 464, 3, 36 },
		{ 464, 4, 36 },
		{ 464, 5, 36 },
		{ 464, 6, 36 },
		{ 464, 7, 36 },
		{ 464, 8, 36 },
		{ 464, 9, 36 },
		{ 464, 10, 36 },
		{ 464, 11, 36 },
		{ 464, 12, 36 },
		{ 464, 13, 36 },
		{ 464, 14, 36 },
		{ 464, 15, 36 },
		{ 464, 16, 36 },
		{ 464, 17, 36 },
		{ 464, 18, 36 },
		{ 464, 19, 36 },
		{ 464, 20, 36 },
		{ 465, 1, 36 },
		{ 465, 2, 36 },
		{ 465, 3, 36 },
		{ 465, 4, 36 },
		{ 465, 5, 36 },
		{ 465, 6, 36 },
		{ 465, 7, 36 },
		{ 465, 8, 36 },
		{ 465, 9, 36 },
		{ 465, 10, 36 },
		{ 465, 11, 36 },
		{ 465, 12, 36 },
		{ 465, 13, 36 },
		{ 465, 14, 36 },
		{ 465, 15, 36 },
		{ 465, 16, 36 },
		{ 465, 17, 36 },
		{ 465, 18, 36 },
		{ 465, 19, 36 },
		{ 466, 1, 36 },
		{ 466, 2, 36 },
		{ 466, 3, 36 },
		{ 466, 4, 36 },
		{ 466, 5, 36 },
		{ 466, 6, 36 },
		{ 466, 7, 36 },
		{ 466, 8, 36 },
		{ 466, 9, 36 },
		{ 466, 10, 36 },
		{ 466, 11, 36 },
		{ 466, 12, 36 },
		{ 466, 13, 36 },
		{ 466, 14, 36 },
		{ 466, 15, 36 },
		{ 466, 16, 36 },
		{ 466, 17, 36 },
		{ 466, 18, 36 },
		{ 467, 1, 36 },
		{ 467, 2, 36 },
		{ 467, 3, 36 },
		{ 467, 4, 36 },
		{ 467, 5, 36 },
		{ 467, 6, 36 },
		{ 467, 7, 36 },
		{ 467, 8, 36 },
		{ 467, 9, 36 },
		{ 467, 10, 36 },
		{ 467, 11, 36 },
		{ 467, 12, 36 },
		{ 467, 13, 36 },
		{ 467, 14, 36 },
		{ 467, 15, 36 },
		{ 467, 16, 36 },
		{ 468, 1, 36 },
		{ 468, 2, 36 },
		{ 468, 3, 36 },
		{ 468, 4, 36 },
		{ 468, 5, 36 },
		{ 468, 6, 36 },
		{ 468, 7, 36 },
		{ 468, 8, 36 },
		{ 468, 9, 36 },
		{ 468, 10, 36 },
		{ 468, 11, 36 },
		{ 468, 12, 36 },
		{ 468, 13, 36 },
		{ 468, 14, 36 },
		{ 468, 15, 36 },
		{ 469, 1, 36 },
		{ 469, 2, 36 },
		{ 469, 3, 36 },
		{ 469, 4, 36 },
		{ 469, 5, 36 },
		{ 469, 6, 36 },
		{ 469, 7, 36 },
		{ 469, 8, 36 },
		{ 469, 9, 36 },
		{ 469, 10, 36 },
		{ 469, 11, 36 },
		{ 469, 12, 36 },
		{ 469, 13, 36 },
		{ 469, 14, 36 },
		{ 470, 1, 36 },
		{ 470, 2, 36 },
		{ 470, 3, 36 },
		{ 470, 4, 36 },
		{ 470, 5, 36 },
		{ 470, 6, 36 },
		{ 470, 7, 36 },
		{ 470, 8, 36 },
		{ 470, 9, 36 },
		{ 470, 10, 36 },
		{ 470, 11, 36 },
		{ 470, 12, 36 },
		{ 470, 13, 36 },
		{ 471, 1, 36 },
		{ 471, 2, 36 },
		{ 471, 3, 36 },
		{ 471, 4, 36 },
		{ 471, 5, 36 },
		{ 471, 6, 36 },
		{ 471, 7, 36 },
		{ 471, 8, 36 },
		{ 471, 9, 36 },
		{ 471, 10, 36 },
		{ 471, 11, 36 },
		{ 471, 12, 36 },
		{ 472, 1, 36 },
		{ 472, 2, 36 },
		{ 472, 3, 36 },
		{ 472, 4, 36 },
		{ 472, 5, 36 },
		{ 472, 6, 36 },
		{ 472, 7, 36 },
		{ 472, 8, 36 },
		{ 472, 9, 36 },
		{ 472, 10, 36 },
		{ 472, 11, 36 },
		{ 473, 1, 36 },
		{ 473, 2, 36 },
		{ 473, 3, 36 },
		{ 473, 4, 36 },
		{ 473, 5, 36 },
		{ 473, 6, 36 },
		{ 473, 7, 36 },
		{ 473, 8, 36 },
		{ 473, 9, 36 },
		{ 473, 10, 36 },
		{ 474, 1, 36 },
		{ 474, 2, 36 },
		{ 474, 3, 36 },
		{ 474, 4, 36 },
		{ 474, 5, 36 },
		{ 474, 6, 36 },
		{ 474, 7, 36 },
		{ 474, 8, 36 },
		{ 475, 1, 36 },
		{ 475, 2, 36 },
		{ 475, 3, 36 },
		{ 475, 4, 36 },
		{ 475, 5, 36 },
		{ 475, 6, 36 },
		{ 475, 7, 36 },
		{ 476, 1, 36 },
		{ 476, 2, 36 },
		{ 476, 3, 36 },
		{ 476, 4, 36 },
		{ 476, 5, 36 },
		{ 476, 6, 36 },
		{ 477, 1, 36 },
		{ 477, 2, 36 },
		{ 477, 3, 36 },
		{ 477, 4, 36 },
		{ 477, 5, 36 },
		{ 478, 1, 36 },
		{ 478, 2, 36 },
		{ 478, 3, 36 },
		{ 478, 4, 36 },
		{ 479, 1, 36 },
		{ 479, 2, 36 },
		{ 479, 3, 36 },
		{ 480, 1, 36 },
		{ 480, 2, 36 },
		{ 3, 512, 37 },
		{ 4, 511, 37 },
		{ 4, 512, 37 },
		{ 5, 510, 37 },
		{ 5, 511, 37 },
		{ 5, 512, 37 },
		{ 6, 509, 37 },
		{ 6, 510, 37 },
		{ 6, 511, 37 },
		{ 6, 512, 37 },
		{ 7, 508, 37 },
		{ 7, 509, 37 },
		{ 7, 510, 37 },
		{ 7, 511, 37 },
		{ 7, 512, 37 },
		{ 8, 507, 37 },
		{ 8, 508, 37 },
		{ 8, 509, 37 },
		{ 8, 510, 37 },
		{ 8, 511, 37 },
		{ 8, 512, 37 },
		{ 9, 506, 37 },
		{ 9, 507, 37 },
		{ 9, 508, 37 },
		{ 9, 509, 37 },
		{ 9, 510, 37 },
		{ 9, 511, 37 },
		{ 9, 512, 37 },
		{ 10, 505, 37 },
		{ 10, 506, 37 },
		{ 10, 507, 37 },
		{ 10, 508, 37 },
		{ 10, 509, 37 },
		{ 10, 510, 37 },
		{ 10, 511, 37 },
		{ 10, 512, 37 },
		{ 11, 504, 37 },
		{ 11, 505, 37 },
		{ 11, 506, 37 },
		{ 11, 507, 37 },
		{ 11, 508, 37 },
		{ 11, 509, 37 },
		{ 11, 510, 37 },
		{ 11, 511, 37 },
		{ 11, 512, 37 },
		{ 12, 503, 37 },
		{ 12, 504, 37 },
		{ 12, 505, 37 },
		{ 12, 506, 37 },
		{ 12, 507, 37 },
		{ 12, 508, 37 },
		{ 12, 509, 37 },
		{ 12, 510, 37 },
		{ 12, 511, 37 },
		{ 12, 512, 37 },
		{ 13, 502, 37 },
		{ 13, 503, 37 },
		{ 13, 504, 37 },
		{ 13, 505, 37 },
		{ 13, 506, 37 },
		{ 13, 507, 37 },
		{ 13, 508, 37 },
		{ 13, 509, 37 },
		{ 13, 510, 37 },
		{ 13, 511, 37 },
		{ 13, 512, 37 },
		{ 14, 501, 37 },
		{ 14, 502, 37 },
		{ 14, 503, 37 },
		{ 14, 504, 37 },
		{ 14, 505, 37 },
		{ 14, 506, 37 },
		{ 14, 507, 37 },
		{ 14, 508, 37 },
		{ 14, 509, 37 },
		{ 14, 510, 37 },
		{ 14, 511, 37 },
		{ 14, 512, 37 },
		{ 15, 500, 37 },
		{ 15, 501, 37 },
		{ 15, 502, 37 },
		{ 15, 503, 37 },
		{ 15, 504, 37 },
		{ 15, 505, 37 },
		{ 15, 506, 37 },
		{ 15, 507, 37 },
		{ 15, 508, 37 },
		{ 15, 509, 37 },
		{ 15, 510, 37 },
		{ 15, 511, 37 },
		{ 15, 512, 37 },
		{ 16, 499, 37 },
		{ 16, 500, 37 },
		{ 16, 501, 37 },
		{ 16, 502, 37 },
		{ 16, 503, 37 },
		{ 16, 504, 37 },
		{ 16, 505, 37 },
		{ 16, 506, 37 },
		{ 16, 507, 37 },
		{ 16, 508, 37 },
		{ 16, 509, 37 },
		{ 16, 510, 37 },
		{ 16, 511, 37 },
		{ 16, 512, 37 },
		{ 17, 498, 37 },
		{ 17, 499, 37 },
		{ 17, 500, 37 },
		{ 17, 501, 37 },
		{ 17, 502, 37 },
		{ 17, 503, 37 },
		{ 17, 504, 37 },
		{ 17, 505, 37 },
		{ 17, 506, 37 },
		{ 17, 507, 37 },
		{ 17, 508, 37 },
		{ 17, 509, 37 },
		{ 17, 510, 37 },
		{ 17, 511, 37 },
		{ 17, 512, 37 },
		{ 18, 497, 37 },
		{ 18, 498, 37 },
		{ 18, 499, 37 },
		{ 18, 500, 37 },
		{ 18, 501, 37 },
		{ 18, 502, 37 },
		{ 18, 503, 37 },
		{ 18, 504, 37 },
		{ 18, 505, 37 },
		{ 18, 506, 37 },
		{ 18, 507, 37 },
		{ 18, 508, 37 },
		{ 18, 509, 37 },
		{ 18, 510, 37 },
		{ 18, 511, 37 },
		{ 18, 512, 37 },
		{ 19, 496, 37 },
		{ 19, 497, 37 },
		{ 19, 498, 37 },
		{ 19, 499, 37 },
		{ 19, 500, 37 },
		{ 19, 501, 37 },
		{ 19, 502, 37 },
		{ 19, 503, 37 },
		{ 19, 504, 37 },
		{ 19, 505, 37 },
		{ 19, 506, 37 },
		{ 19, 507, 37 },
		{ 19, 508, 37 },
		{ 19, 509, 37 },
		{ 19, 510, 37 },
		{ 19, 511, 37 },
		{ 19, 512, 37 },
		{ 20, 495, 37 },
		{ 20, 496, 37 },
		{ 20, 497, 37 },
		{ 20, 498, 37 },
		{ 20, 499, 37 },
		{ 20, 500, 37 },
		{ 20, 501, 37 },
		{ 20, 502, 37 },
		{ 20, 503, 37 },
		{ 20, 504, 37 },
		{ 20, 505, 37 },
		{ 20, 506, 37 },
		{ 20, 507, 37 },
		{ 20, 508, 37 },
		{ 20, 509, 37 },
		{ 20, 510, 37 },
		{ 20, 511, 37 },
		{ 20, 512, 37 },
		{ 21, 494, 37 },
		{ 21, 495, 37 },
		{ 21, 496, 37 },
		{ 21, 497, 37 },
		{ 21, 498, 37 },
		{ 21, 499, 37 },
		{ 21, 500, 37 },
		{ 21, 501, 37 },
		{ 21, 502, 37 },
		{ 21, 503, 37 },
		{ 21, 504, 37 },
		{ 21, 505, 37 },
		{ 21, 506, 37 },
		{ 21, 507, 37 },
		{ 21, 508, 37 },
		{ 21, 509, 37 },
		{ 21, 510, 37 },
		{ 21, 511, 37 },
		{ 21, 512, 37 },
		{ 22, 493, 37 },
		{ 22, 494, 37 },
		{ 22, 495, 37 },
		{ 22, 496, 37 },
		{ 22, 497, 37 },
		{ 22, 498, 37 },
		{ 22, 499, 37 },
		{ 22, 500, 37 },
		{ 22, 501, 37 },
		{ 22, 502, 37 },
		{ 22, 503, 37 },
		{ 22, 504, 37 },
		{ 22, 505, 37 },
		{ 22, 506, 37 },
		{ 22, 507, 37 },
		{ 22, 508, 37 },
		{ 22, 509, 37 },
		{ 22, 510, 37 },
		{ 22, 511, 37 },
		{ 22, 512, 37 },
		{ 23, 492, 37 },
		{ 23, 493, 37 },
		{ 23, 494, 37 },
		{ 23, 495, 37 },
		{ 23, 496, 37 },
		{ 23, 497, 37 },
		{ 23, 498, 37 },
		{ 23, 499, 37 },
		{ 23, 500, 37 },
		{ 23, 501, 37 },
		{ 23, 502, 37 },
		{ 23, 503, 37 },
		{ 23, 504, 37 },
		{ 23, 505, 37 },
		{ 23, 506, 37 },
		{ 23, 507, 37 },
		{ 23, 508, 37 },
		{ 23, 509, 37 },
		{ 23, 510, 37 },
		{ 23, 511, 37 },
		{ 23, 512, 37 },
		{ 24, 491, 37 },
		{ 24, 492, 37 },
		{ 24, 493, 37 },
		{ 24, 494, 37 },
		{ 24, 495, 37 },
		{ 24, 496, 37 },
		{ 24, 497, 37 },
		{ 24, 498, 37 },
		{ 24, 499, 37 },
		{ 24, 500, 37 },
		{ 24, 501, 37 },
		{ 24, 502, 37 },
		{ 24, 503, 37 },
		{ 24, 504, 37 },
		{ 24, 505, 37 },
		{ 24, 506, 37 },
		{ 24, 507, 37 },
		{ 24, 508, 37 },
		{ 24, 509, 37 },
		{ 24, 510, 37 },
		{ 24, 511, 37 },
		{ 24, 512, 37 },
		{ 25, 490, 37 },
		{ 25, 491, 37 },
		{ 25, 492, 37 },
		{ 25, 493, 37 },
		{ 25, 494, 37 },
		{ 25, 495, 37 },
		{ 25, 496, 37 },
		{ 25, 497, 37 },
		{ 25, 498, 37 },
		{ 25, 499, 37 },
		{ 25, 500, 37 },
		{ 25, 501, 37 },
		{ 25, 502, 37 },
		{ 25, 503, 37 },
		{ 25, 504, 37 },
		{ 25, 505, 37 },
		{ 25, 506, 37 },
		{ 25, 507, 37 },
		{ 25, 508, 37 },
		{ 25, 509, 37 },
		{ 25, 510, 37 },
		{ 25, 511, 37 },
		{ 25, 512, 37 },
		{ 26, 489, 37 },
		{ 26, 490, 37 },
		{ 26, 491, 37 },
		{ 26, 492, 37 },
		{ 26, 493, 37 },
		{ 26, 494, 37 },
		{ 26, 495, 37 },
		{ 26, 496, 37 },
		{ 26, 497, 37 },
		{ 26, 498, 37 },
		{ 26, 499, 37 },
		{ 26, 500, 37 },
		{ 26, 501, 37 },
		{ 26, 502, 37 },
		{ 26, 503, 37 },
		{ 26, 504, 37 },
		{ 26, 505, 37 },
		{ 26, 506, 37 },
		{ 26, 507, 37 },
		{ 26, 508, 37 },
		{ 26, 509, 37 },
		{ 26, 510, 37 },
		{ 26, 511, 37 },
		{ 26, 512, 37 },
		{ 27, 488, 37 },
		{ 27, 489, 37 },
		{ 27, 490, 37 },
		{ 27, 491, 37 },
		{ 27, 492, 37 },
		{ 27, 493, 37 },
		{ 27, 494, 37 },
		{ 27, 495, 37 },
		{ 27, 496, 37 },
		{ 27, 497, 37 },
		{ 27, 498, 37 },
		{ 27, 499, 37 },
		{ 27, 500, 37 },
		{ 27, 501, 37 },
		{ 27, 502, 37 },
		{ 27, 503, 37 },
		{ 27, 504, 37 },
		{ 27, 505, 37 },
		{ 27, 506, 37 },
		{ 27, 507, 37 },
		{ 27, 508, 37 },
		{ 27, 509, 37 },
		{ 27, 510, 37 },
		{ 27, 511, 37 },
		{ 27, 512, 37 },
		{ 28, 487, 37 },
		{ 28, 488, 37 },
		{ 28, 489, 37 },
		{ 28, 490, 37 },
		{ 28, 491, 37 },
		{ 28, 492, 37 },
		{ 28, 493, 37 },
		{ 28, 494, 37 },
		{ 28, 495, 37 },
		{ 28, 496, 37 },
		{ 28, 497, 37 },
		{ 28, 498, 37 },
		{ 28, 499, 37 },
		{ 28, 500, 37 },
		{ 28, 501, 37 },
		{ 28, 502, 37 },
		{ 28, 503, 37 },
		{ 28, 504, 37 },
		{ 28, 505, 37 },
		{ 28, 506, 37 },
		{ 28, 507, 37 },
		{ 28, 508, 37 },
		{ 28, 509, 37 },
		{ 28, 510, 37 },
		{ 28, 511, 37 },
		{ 28, 512, 37 },
		{ 29, 486, 37 },
		{ 29, 487, 37 },
		{ 29, 488, 37 },
		{ 29, 489, 37 },
		{ 29, 490, 37 },
		{ 29, 491, 37 },
		{ 29, 492, 37 },
		{ 29, 493, 37 },
		{ 29, 494, 37 },
		{ 29, 495, 37 },
		{ 29, 496, 37 },
		{ 29, 497, 37 },
		{ 29, 498, 37 },
		{ 29, 499, 37 },
		{ 29, 500, 37 },
		{ 29, 501, 37 },
		{ 29, 502, 37 },
		{ 29, 503, 37 },
		{ 29, 504, 37 },
		{ 29, 505, 37 },
		{ 29, 506, 37 },
		{ 29, 507, 37 },
		{ 29, 508, 37 },
		{ 29, 509, 37 },
		{ 29, 510, 37 },
		{ 29, 511, 37 },
		{ 29, 512, 37 },
		{ 30, 485, 37 },
		{ 30, 486, 37 },
		{ 30, 487, 37 },
		{ 30, 488, 37 },
		{ 30, 489, 37 },
		{ 30, 490, 37 },
		{ 30, 491, 37 },
		{ 30, 492, 37 },
		{ 30, 493, 37 },
		{ 30, 494, 37 },
		{ 30, 495, 37 },
		{ 30, 496, 37 },
		{ 30, 497, 37 },
		{ 30, 498, 37 },
		{ 30, 499, 37 },
		{ 30, 500, 37 },
		{ 30, 501, 37 },
		{ 30, 502, 37 },
		{ 30, 503, 37 },
		{ 30, 504, 37 },
		{ 30, 505, 37 },
		{ 30, 506, 37 },
		{ 30, 507, 37 },
		{ 30, 508, 37 },
		{ 30, 509, 37 },
		{ 30, 510, 37 },
		{ 30, 511, 37 },
		{ 30, 512, 37 },
		{ 31, 484, 37 },
		{ 31, 485, 37 },
		{ 31, 486, 37 },
		{ 31, 487, 37 },
		{ 31, 488, 37 },
		{ 31, 489, 37 },
		{ 31, 490, 37 },
		{ 31, 491, 37 },
		{ 31, 492, 37 },
		{ 31, 493, 37 },
		{ 31, 494, 37 },
		{ 31, 495, 37 },
		{ 31, 496, 37 },
		{ 31, 497, 37 },
		{ 31, 498, 37 },
		{ 31, 499, 37 },
		{ 31, 500, 37 },
		{ 31, 501, 37 },
		{ 31, 502, 37 },
		{ 31, 503, 37 },
		{ 31, 504, 37 },
		{ 31, 505, 37 },
		{ 31, 506, 37 },
		{ 31, 507, 37 },
		{ 31, 508, 37 },
		{ 31, 509, 37 },
		{ 31, 510, 37 },
		{ 31, 511, 37 },
		{ 31, 512, 37 },
		{ 32, 483, 37 },
		{ 32, 484, 37 },
		{ 32, 485, 37 },
		{ 32, 486, 37 },
		{ 32, 487, 37 },
		{ 32, 488, 37 },
		{ 32, 489, 37 },
		{ 32, 490, 37 },
		{ 32, 491, 37 },
		{ 32, 492, 37 },
		{ 32, 493, 37 },
		{ 32, 494, 37 },
		{ 32, 495, 37 },
		{ 32, 496, 37 },
		{ 32, 497, 37 },
		{ 32, 498, 37 },
		{ 32, 499, 37 },
		{ 32, 500, 37 },
		{ 32, 501, 37 },
		{ 32, 502, 37 },
		{ 32, 503, 37 },
		{ 32, 504, 37 },
		{ 32, 505, 37 },
		{ 32, 506, 37 },
		{ 32, 507, 37 },
		{ 32, 508, 37 },
		{ 32, 509, 37 },
		{ 32, 510, 37 },
		{ 32, 511, 37 },
		{ 32, 512, 37 },
		{ 33, 482, 37 },
		{ 33, 483, 37 },
		{ 33, 484, 37 },
		{ 33, 485, 37 },
		{ 33, 486, 37 },
		{ 33, 487, 37 },
		{ 33, 488, 37 },
		{ 33, 489, 37 },
		{ 33, 490, 37 },
		{ 33, 491, 37 },
		{ 33, 492, 37 },
		{ 33, 493, 37 },
		{ 33, 494, 37 },
		{ 33, 495, 37 },
		{ 33, 496, 37 },
		{ 33, 497, 37 },
		{ 33, 498, 37 },
		{ 33, 499, 37 },
		{ 33, 500, 37 },
		{ 33, 501, 37 },
		{ 33, 502, 37 },
		{ 33, 503, 37 },
		{ 33, 504, 37 },
		{ 33, 505, 37 },
		{ 33, 506, 37 },
		{ 33, 507, 37 },
		{ 33, 508, 37 },
		{ 33, 509, 37 },
		{ 33, 510, 37 },
		{ 33, 511, 37 },
		{ 33, 512, 37 },
		{ 34, 481, 37 },
		{ 34, 482, 37 },
		{ 34, 483, 37 },
		{ 34, 484, 37 },
		{ 34, 485, 37 },
		{ 34, 486, 37 },
		{ 34, 487, 37 },
		{ 34, 488, 37 },
		{ 34, 489, 37 },
		{ 34, 490, 37 },
		{ 34, 491, 37 },
		{ 34, 492, 37 },
		{ 34, 493, 37 },
		{ 34, 494, 37 },
		{ 34, 495, 37 },
		{ 34, 496, 37 },
		{ 34, 497, 37 },
		{ 34, 498, 37 },
		{ 34, 499, 37 },
		{ 34, 500, 37 },
		{ 34, 501, 37 },
		{ 34, 502, 37 },
		{ 34, 503, 37 },
		{ 34, 504, 37 },
		{ 34, 505, 37 },
		{ 34, 506, 37 },
		{ 34, 507, 37 },
		{ 34, 508, 37 },
		{ 34, 509, 37 },
		{ 34, 510, 37 },
		{ 34, 511, 37 },
		{ 34, 512, 37 },
		{ 35, 480, 37 },
		{ 35, 481, 37 },
		{ 35, 482, 37 },
		{ 35, 483, 37 },
		{ 35, 484, 37 },
		{ 35, 485, 37 },
		{ 35, 486, 37 },
		{ 35, 487, 37 },
		{ 35, 488, 37 },
		{ 35, 489, 37 },
		{ 35, 490, 37 },
		{ 35, 491, 37 },
		{ 35, 492, 37 },
		{ 35, 493, 37 },
		{ 35, 494, 37 },
		{ 35, 495, 37 },
		{ 35, 496, 37 },
		{ 35, 497, 37 },
		{ 35, 498, 37 },
		{ 35, 499, 37 },
		{ 35, 500, 37 },
		{ 35, 501, 37 },
		{ 35, 502, 37 },
		{ 35, 503, 37 },
		{ 35, 504, 37 },
		{ 35, 505, 37 },
		{ 35, 506, 37 },
		{ 35, 507, 37 },
		{ 35, 508, 37 },
		{ 35, 509, 37 },
		{ 35, 510, 37 },
		{ 35, 511, 37 },
		{ 35, 512, 37 },
		{ 36, 479, 37 },
		{ 36, 480, 37 },
		{ 36, 481, 37 },
		{ 36, 482, 37 },
		{ 36, 483, 37 },
		{ 36, 484, 37 },
		{ 36, 485, 37 },
		{ 36, 486, 37 },
		{ 36, 487, 37 },
		{ 36, 488, 37 },
		{ 36, 489, 37 },
		{ 36, 490, 37 },
		{ 36, 491, 37 },
		{ 36, 492, 37 },
		{ 36, 493, 37 },
		{ 36, 494, 37 },
		{ 36, 495, 37 },
		{ 36, 496, 37 },
		{ 36, 497, 37 },
		{ 36, 498, 37 },
		{ 36, 499, 37 },
		{ 36, 500, 37 },
		{ 36, 501, 37 },
		{ 36, 502, 37 },
		{ 36, 503, 37 },
		{ 36, 504, 37 },
		{ 36, 505, 37 },
		{ 36, 506, 37 },
		{ 36, 507, 37 },
		{ 36, 508, 37 },
		{ 36, 509, 37 },
		{ 36, 510, 37 },
		{ 36, 511, 37 },
		{ 36, 512, 37 },
		{ 37, 478, 37 },
		{ 37, 479, 37 },
		{ 37, 480, 37 },
		{ 37, 481, 37 },
		{ 37, 482, 37 },
		{ 37, 483, 37 },
		{ 37, 484, 37 },
		{ 37, 485, 37 },
		{ 37, 486, 37 },
		{ 37, 487, 37 },
		{ 37, 488, 37 },
		{ 37, 489, 37 },
		{ 37, 490, 37 },
		{ 37, 491, 37 },
		{ 37, 492, 37 },
		{ 37, 493, 37 },
		{ 37, 494, 37 },
		{ 37, 495, 37 },
		{ 37, 496, 37 },
		{ 37, 497, 37 },
		{ 37, 498, 37 },
		{ 37, 499, 37 },
		{ 37, 500, 37 },
		{ 37, 501, 37 },
		{ 37, 502, 37 },
		{ 37, 503, 37 },
		{ 37, 504, 37 },
		{ 37, 505, 37 },
		{ 37, 506, 37 },
		{ 37, 507, 37 },
		{ 37, 508, 37 },
		{ 37, 509, 37 },
		{ 37, 510, 37 },
		{ 37, 511, 37 },
		{ 37, 512, 37 },
		{ 38, 477, 37 },
		{ 38, 478, 37 },
		{ 38, 479, 37 },
		{ 38, 480, 37 },
		{ 38, 481, 37 },
		{ 38, 482, 37 },
		{ 38, 483, 37 },
		{ 38, 484, 37 },
		{ 38, 485, 37 },
		{ 38, 486, 37 },
		{ 38, 487, 37 },
		{ 38, 488, 37 },
		{ 38, 489, 37 },
		{ 38, 490, 37 },
		{ 38, 491, 37 },
		{ 38, 492, 37 },
		{ 38, 493, 37 },
		{ 38, 494, 37 },
		{ 38, 495, 37 },
		{ 38, 496, 37 },
		{ 38, 497, 37 },
		{ 38, 498, 37 },
		{ 38, 499, 37 },
		{ 38, 500, 37 },
		{ 38, 501, 37 },
		{ 38, 502, 37 },
		{ 38, 503, 37 },
		{ 38, 504, 37 },
		{ 38, 505, 37 },
		{ 38, 506, 37 },
		{ 38, 507, 37 },
		{ 38, 508, 37 },
		{ 38, 509, 37 },
		{ 38, 510, 37 },
		{ 38, 511, 37 },
		{ 38, 512, 37 },
		{ 39, 476, 37 },
		{ 39, 477, 37 },
		{ 39, 478, 37 },
		{ 39, 479, 37 },
		{ 39, 480, 37 },
		{ 39, 481, 37 },
		{ 39, 482, 37 },
		{ 39, 483, 37 },
		{ 39, 484, 37 },
		{ 39, 485, 37 },
		{ 39, 486, 37 },
		{ 39, 487, 37 },
		{ 39, 488, 37 },
		{ 39, 489, 37 },
		{ 39, 490, 37 },
		{ 39, 491, 37 },
		{ 39, 492, 37 },
		{ 39, 493, 37 },
		{ 39, 494, 37 },
		{ 39, 495, 37 },
		{ 39, 496, 37 },
		{ 39, 497, 37 },
		{ 39, 498, 37 },
		{ 39, 499, 37 },
		{ 39, 500, 37 },
		{ 39, 501, 37 },
		{ 39, 502, 37 },
		{ 39, 503, 37 },
		{ 39, 504, 37 },
		{ 39, 505, 37 },
		{ 39, 506, 37 },
		{ 39, 507, 37 },
		{ 39, 508, 37 },
		{ 39, 509, 37 },
		{ 39, 510, 37 },
		{ 39, 511, 37 },
		{ 39, 512, 37 },
		{ 40, 475, 37 },
		{ 40, 476, 37 },
		{ 40, 477, 37 },
		{ 40, 478, 37 },
		{ 40, 479, 37 },
		{ 40, 480, 37 },
		{ 40, 481, 37 },
		{ 40, 482, 37 },
		{ 40, 483, 37 },
		{ 40, 484, 37 },
		{ 40, 485, 37 },
		{ 40, 486, 37 },
		{ 40, 487, 37 },
		{ 40, 488, 37 },
		{ 40, 489, 37 },
		{ 40, 490, 37 },
		{ 40, 491, 37 },
		{ 40, 492, 37 },
		{ 40, 493, 37 },
		{ 40, 494, 37 },
		{ 40, 495, 37 },
		{ 40, 496, 37 },
		{ 40, 497, 37 },
		{ 40, 498, 37 },
		{ 40, 499, 37 },
		{ 40, 500, 37 },
		{ 40, 501, 37 },
		{ 40, 502, 37 },
		{ 40, 503, 37 },
		{ 40, 504, 37 },
		{ 40, 505, 37 },
		{ 40, 506, 37 },
		{ 40, 507, 37 },
		{ 40, 508, 37 },
		{ 40, 509, 37 },
		{ 40, 510, 37 },
		{ 40, 511, 37 },
		{ 40, 512, 37 },
		{ 41, 474, 37 },
		{ 41, 475, 37 },
		{ 41, 476, 37 },
		{ 41, 477, 37 },
		{ 41, 478, 37 },
		{ 41, 479, 37 },
		{ 41, 480, 37 },
		{ 41, 481, 37 },
		{ 41, 482, 37 },
		{ 41, 483, 37 },
		{ 41, 484, 37 },
		{ 41, 485, 37 },
		{ 41, 486, 37 },
		{ 41, 487, 37 },
		{ 41, 488, 37 },
		{ 41, 489, 37 },
		{ 41, 490, 37 },
		{ 41, 491, 37 },
		{ 41, 492, 37 },
		{ 41, 493, 37 },
		{ 41, 494, 37 },
		{ 41, 495, 37 },
		{ 41, 496, 37 },
		{ 41, 497, 37 },
		{ 41, 498, 37 },
		{ 41, 499, 37 },
		{ 41, 500, 37 },
		{ 41, 501, 37 },
		{ 41, 502, 37 },
		{ 41, 503, 37 },
		{ 41, 504, 37 },
		{ 41, 505, 37 },
		{ 41, 506, 37 },
		{ 41, 507, 37 },
		{ 41, 508, 37 },
		{ 41, 509, 37 },
		{ 41, 510, 37 },
		{ 41, 511, 37 },
		{ 41, 512, 37 },
		{ 42, 473, 37 },
		{ 42, 474, 37 },
		{ 42, 475, 37 },
		{ 42, 476, 37 },
		{ 42, 477, 37 },
		{ 42, 478, 37 },
		{ 42, 479, 37 },
		{ 42, 480, 37 },
		{ 42, 481, 37 },
		{ 42, 482, 37 },
		{ 42, 483, 37 },
		{ 42, 484, 37 },
		{ 42, 485, 37 },
		{ 42, 486, 37 },
		{ 42, 487, 37 },
		{ 42, 488, 37 },
		{ 42, 489, 37 },
		{ 42, 490, 37 },
		{ 42, 491, 37 },
		{ 42, 492, 37 },
		{ 42, 493, 37 },
		{ 42, 494, 37 },
		{ 42, 495, 37 },
		{ 42, 496, 37 },
		{ 42, 497, 37 },
		{ 42, 498, 37 },
		{ 42, 499, 37 },
		{ 42, 500, 37 },
		{ 42, 501, 37 },
		{ 42, 502, 37 },
		{ 42, 503, 37 },
		{ 42, 504, 37 },
		{ 42, 505, 37 },
		{ 42, 506, 37 },
		{ 42, 507, 37 },
		{ 42, 508, 37 },
		{ 42, 509, 37 },
		{ 42, 510, 37 },
		{ 42, 511, 37 },
		{ 42, 512, 37 },
		{ 43, 472, 37 },
		{ 43, 473, 37 },
		{ 43, 474, 37 },
		{ 43, 475, 37 },
		{ 43, 476, 37 },
		{ 43, 477, 37 },
		{ 43, 478, 37 },
		{ 43, 479, 37 },
		{ 43, 480, 37 },
		{ 43, 481, 37 },
		{ 43, 482, 37 },
		{ 43, 483, 37 },
		{ 43, 484, 37 },
		{ 43, 485, 37 },
		{ 43, 486, 37 },
		{ 43, 487, 37 },
		{ 43, 488, 37 },
		{ 43, 489, 37 },
		{ 43, 490, 37 },
		{ 43, 491, 37 },
		{ 43, 492, 37 },
		{ 43, 493, 37 },
		{ 43, 494, 37 },
		{ 43, 495, 37 },
		{ 43, 496, 37 },
		{ 43, 497, 37 },
		{ 43, 498, 37 },
		{ 43, 499, 37 },
		{ 43, 500, 37 },
		{ 43, 501, 37 },
		{ 43, 502, 37 },
		{ 43, 503, 37 },
		{ 43, 504, 37 },
		{ 43, 505, 37 },
		{ 43, 506, 37 },
		{ 43, 507, 37 },
		{ 43, 508, 37 },
		{ 43, 509, 37 },
		{ 43, 510, 37 },
		{ 43, 511, 37 },
		{ 43, 512, 37 },
		{ 44, 471, 37 },
		{ 44, 472, 37 },
		{ 44, 473, 37 },
		{ 44, 474, 37 },
		{ 44, 475, 37 },
		{ 44, 476, 37 },
		{ 44, 477, 37 },
		{ 44, 478, 37 },
		{ 44, 479, 37 },
		{ 44, 480, 37 },
		{ 44, 481, 37 },
		{ 44, 482, 37 },
		{ 44, 483, 37 },
		{ 44, 484, 37 },
		{ 44, 485, 37 },
		{ 44, 486, 37 },
		{ 44, 487, 37 },
		{ 44, 488, 37 },
		{ 44, 489, 37 },
		{ 44, 490, 37 },
		{ 44, 491, 37 },
		{ 44, 492, 37 },
		{ 44, 493, 37 },
		{ 44, 494, 37 },
		{ 44, 495, 37 },
		{ 44, 496, 37 },
		{ 44, 497, 37 },
		{ 44, 498, 37 },
		{ 44, 499, 37 },
		{ 44, 500, 37 },
		{ 44, 501, 37 },
		{ 44, 502, 37 },
		{ 44, 503, 37 },
		{ 44, 504, 37 },
		{ 44, 505, 37 },
		{ 44, 506, 37 },
		{ 44, 507, 37 },
		{ 44, 508, 37 },
		{ 44, 509, 37 },
		{ 44, 510, 37 },
		{ 44, 511, 37 },
		{ 44, 512, 37 },
		{ 45, 470, 37 },
		{ 45, 471, 37 },
		{ 45, 472, 37 },
		{ 45, 473, 37 },
		{ 45, 474, 37 },
		{ 45, 475, 37 },
		{ 45, 476, 37 },
		{ 45, 477, 37 },
		{ 45, 478, 37 },
		{ 45, 479, 37 },
		{ 45, 480, 37 },
		{ 45, 481, 37 },
		{ 45, 482, 37 },
		{ 45, 483, 37 },
		{ 45, 484, 37 },
		{ 45, 485, 37 },
		{ 45, 486, 37 },
		{ 45, 487, 37 },
		{ 45, 488, 37 },
		{ 45, 489, 37 },
		{ 45, 490, 37 },
		{ 45, 491, 37 },
		{ 45, 492, 37 },
		{ 45, 493, 37 },
		{ 45, 494, 37 },
		{ 45, 495, 37 },
		{ 45, 496, 37 },
		{ 45, 497, 37 },
		{ 45, 498, 37 },
		{ 45, 499, 37 },
		{ 45, 500, 37 },
		{ 45, 501, 37 },
		{ 45, 502, 37 },
		{ 45, 503, 37 },
		{ 45, 504, 37 },
		{ 45, 505, 37 },
		{ 45, 506, 37 },
		{ 45, 507, 37 },
		{ 45, 508, 37 },
		{ 45, 509, 37 },
		{ 45, 510, 37 },
		{ 45, 511, 37 },
		{ 45, 512, 37 },
		{ 46, 469, 37 },
		{ 46, 470, 37 },
		{ 46, 471, 37 },
		{ 46, 472, 37 },
		{ 46, 473, 37 },
		{ 46, 474, 37 },
		{ 46, 475, 37 },
		{ 46, 476, 37 },
		{ 46, 477, 37 },
		{ 46, 478, 37 },
		{ 46, 479, 37 },
		{ 46, 480, 37 },
		{ 46, 481, 37 },
		{ 46, 482, 37 },
		{ 46, 483, 37 },
		{ 46, 484, 37 },
		{ 46, 485, 37 },
		{ 46, 486, 37 },
		{ 46, 487, 37 },
		{ 46, 488, 37 },
		{ 46, 489, 37 },
		{ 46, 490, 37 },
		{ 46, 491, 37 },
		{ 46, 492, 37 },
		{ 46, 493, 37 },
		{ 46, 494, 37 },
		{ 46, 495, 37 },
		{ 46, 496, 37 },
		{ 46, 497, 37 },
		{ 46, 498, 37 },
		{ 46, 499, 37 },
		{ 46, 500, 37 },
		{ 46, 501, 37 },
		{ 46, 502, 37 },
		{ 46, 503, 37 },
		{ 46, 504, 37 },
		{ 46, 505, 37 },
		{ 46, 506, 37 },
		{ 46, 507, 37 },
		{ 46, 508, 37 },
		{ 46, 509, 37 },
		{ 46, 510, 37 },
		{ 46, 511, 37 },
		{ 46, 512, 37 },
		{ 47, 468, 37 },
		{ 47, 469, 37 },
		{ 47, 470, 37 },
		{ 47, 471, 37 },
		{ 47, 472, 37 },
		{ 47, 473, 37 },
		{ 47, 474, 37 },
		{ 47, 475, 37 },
		{ 47, 476, 37 },
		{ 47, 477, 37 },
		{ 47, 478, 37 },
		{ 47, 479, 37 },
		{ 47, 480, 37 },
		{ 47, 481, 37 },
		{ 47, 482, 37 },
		{ 47, 483, 37 },
		{ 47, 484, 37 },
		{ 47, 485, 37 },
		{ 47, 486, 37 },
		{ 47, 487, 37 },
		{ 47, 488, 37 },
		{ 47, 489, 37 },
		{ 47, 490, 37 },
		{ 47, 491, 37 },
		{ 47, 492, 37 },
		{ 47, 493, 37 },
		{ 47, 494, 37 },
		{ 47, 495, 37 },
		{ 47, 496, 37 },
		{ 47, 497, 37 },
		{ 47, 498, 37 },
		{ 47, 499, 37 },
		{ 47, 500, 37 },
		{ 47, 501, 37 },
		{ 47, 502, 37 },
		{ 47, 503, 37 },
		{ 47, 504, 37 },
		{ 47, 505, 37 },
		{ 47, 506, 37 },
		{ 47, 507, 37 },
		{ 47, 508, 37 },
		{ 47, 509, 37 },
		{ 47, 510, 37 },
		{ 47, 511, 37 },
		{ 47, 512, 37 },
		{ 48, 467, 37 },
		{ 48, 468, 37 },
		{ 48, 469, 37 },
		{ 48, 470, 37 },
		{ 48, 471, 37 },
		{ 48, 472, 37 },
		{ 48, 473, 37 },
		{ 48, 474, 37 },
		{ 48, 475, 37 },
		{ 48, 476, 37 },
		{ 48, 477, 37 },
		{ 48, 478, 37 },
		{ 48, 479, 37 },
		{ 48, 480, 37 },
		{ 48, 481, 37 },
		{ 48, 482, 37 },
		{ 48, 483, 37 },
		{ 48, 484, 37 },
		{ 48, 485, 37 },
		{ 48, 486, 37 },
		{ 48, 487, 37 },
		{ 48, 488, 37 },
		{ 48, 489, 37 },
		{ 48, 490, 37 },
		{ 48, 491, 37 },
		{ 48, 492, 37 },
		{ 48, 493, 37 },
		{ 48, 494, 37 },
		{ 48, 495, 37 },
		{ 48, 496, 37 },
		{ 48, 497, 37 },
		{ 48, 498, 37 },
		{ 48, 499, 37 },
		{ 48, 500, 37 },
		{ 48, 501, 37 },
		{ 48, 502, 37 },
		{ 48, 503, 37 },
		{ 48, 504, 37 },
		{ 48, 505, 37 },
		{ 48, 506, 37 },
		{ 48, 507, 37 },
		{ 48, 508, 37 },
		{ 48, 509, 37 },
		{ 48, 510, 37 },
		{ 48, 511, 37 },
		{ 48, 512, 37 },
		{ 49, 466, 37 },
		{ 49, 467, 37 },
		{ 49, 468, 37 },
		{ 49, 469, 37 },
		{ 49, 470, 37 },
		{ 49, 471, 37 },
		{ 49, 472, 37 },
		{ 49, 473, 37 },
		{ 49, 474, 37 },
		{ 49, 475, 37 },
		{ 49, 476, 37 },
		{ 49, 477, 37 },
		{ 49, 478, 37 },
		{ 49, 479, 37 },
		{ 49, 480, 37 },
		{ 49, 481, 37 },
		{ 49, 482, 37 },
		{ 49, 483, 37 },
		{ 49, 484, 37 },
		{ 49, 485, 37 },
		{ 49, 486, 37 },
		{ 49, 487, 37 },
		{ 49, 488, 37 },
		{ 49, 489, 37 },
		{ 49, 490, 37 },
		{ 49, 491, 37 },
		{ 49, 492, 37 },
		{ 49, 493, 37 },
		{ 49, 494, 37 },
		{ 49, 495, 37 },
		{ 49, 496, 37 },
		{ 49, 497, 37 },
		{ 49, 498, 37 },
		{ 49, 499, 37 },
		{ 49, 500, 37 },
		{ 49, 501, 37 },
		{ 49, 502, 37 },
		{ 49, 503, 37 },
		{ 49, 504, 37 },
		{ 49, 505, 37 },
		{ 49, 506, 37 },
		{ 49, 507, 37 },
		{ 49, 508, 37 },
		{ 49, 509, 37 },
		{ 49, 510, 37 },
		{ 49, 511, 37 },
		{ 49, 512, 37 },
		{ 50, 465, 37 },
		{ 50, 466, 37 },
		{ 50, 467, 37 },
		{ 50, 468, 37 },
		{ 50, 469, 37 },
		{ 50, 470, 37 },
		{ 50, 471, 37 },
		{ 50, 472, 37 },
		{ 50, 473, 37 },
		{ 50, 474, 37 },
		{ 50, 475, 37 },
		{ 50, 476, 37 },
		{ 50, 477, 37 },
		{ 50, 478, 37 },
		{ 50, 479, 37 },
		{ 50, 480, 37 },
		{ 50, 481, 37 },
		{ 50, 482, 37 },
		{ 50, 483, 37 },
		{ 50, 484, 37 },
		{ 50, 485, 37 },
		{ 50, 486, 37 },
		{ 50, 487, 37 },
		{ 50, 488, 37 },
		{ 50, 489, 37 },
		{ 50, 490, 37 },
		{ 50, 491, 37 },
		{ 50, 492, 37 },
		{ 50, 493, 37 },
		{ 50, 494, 37 },
		{ 50, 495, 37 },
		{ 50, 496, 37 },
		{ 50, 497, 37 },
		{ 50, 498, 37 },
		{ 50, 499, 37 },
		{ 50, 500, 37 },
		{ 50, 501, 37 },
		{ 50, 502, 37 },
		{ 50, 503, 37 },
		{ 50, 504, 37 },
		{ 50, 505, 37 },
		{ 50, 506, 37 },
		{ 50, 507, 37 },
		{ 50, 508, 37 },
		{ 50, 509, 37 },
		{ 50, 510, 37 },
		{ 50, 511, 37 },
		{ 50, 512, 37 },
		{ 51, 464, 37 },
		{ 51, 465, 37 },
		{ 51, 466, 37 },
		{ 51, 467, 37 },
		{ 51, 468, 37 },
		{ 51, 469, 37 },
		{ 51, 470, 37 },
		{ 51, 471, 37 },
		{ 51, 472, 37 },
		{ 51, 473, 37 },
		{ 51, 474, 37 },
		{ 51, 475, 37 },
		{ 51, 476, 37 },
		{ 51, 477, 37 },
		{ 51, 478, 37 },
		{ 51, 479, 37 },
		{ 51, 480, 37 },
		{ 51, 481, 37 },
		{ 51, 482, 37 },
		{ 51, 483, 37 },
		{ 51, 484, 37 },
		{ 51, 485, 37 },
		{ 51, 486, 37 },
		{ 51, 487, 37 },
		{ 51, 488, 37 },
		{ 51, 489, 37 },
		{ 51, 490, 37 },
		{ 51, 491, 37 },
		{ 51, 492, 37 },
		{ 51, 493, 37 },
		{ 51, 494, 37 },
		{ 51, 495, 37 },
		{ 51, 496, 37 },
		{ 51, 497, 37 },
		{ 51, 498, 37 },
		{ 51, 499, 37 },
		{ 51, 500, 37 },
		{ 51, 501, 37 },
		{ 51, 502, 37 },
		{ 51, 503, 37 },
		{ 51, 504, 37 },
		{ 51, 505, 37 },
		{ 51, 506, 37 },
		{ 51, 507, 37 },
		{ 51, 508, 37 },
		{ 51, 509, 37 },
		{ 51, 510, 37 },
		{ 51, 511, 37 },
		{ 51, 512, 37 },
		{ 52, 463, 37 },
		{ 52, 464, 37 },
		{ 52, 465, 37 },
		{ 52, 466, 37 },
		{ 52, 467, 37 },
		{ 52, 468, 37 },
		{ 52, 469, 37 },
		{ 52, 470, 37 },
		{ 52, 471, 37 },
		{ 52, 472, 37 },
		{ 52, 473, 37 },
		{ 52, 474, 37 },
		{ 52, 475, 37 },
		{ 52, 476, 37 },
		{ 52, 477, 37 },
		{ 52, 478, 37 },
		{ 52, 479, 37 },
		{ 52, 480, 37 },
		{ 52, 481, 37 },
		{ 52, 482, 37 },
		{ 52, 483, 37 },
		{ 52, 484, 37 },
		{ 52, 485, 37 },
		{ 52, 486, 37 },
		{ 52, 487, 37 },
		{ 52, 488, 37 },
		{ 52, 489, 37 },
		{ 52, 490, 37 },
		{ 52, 491, 37 },
		{ 52, 492, 37 },
		{ 52, 493, 37 },
		{ 52, 494, 37 },
		{ 52, 495, 37 },
		{ 52, 496, 37 },
		{ 52, 497, 37 },
		{ 52, 498, 37 },
		{ 52, 499, 37 },
		{ 52, 500, 37 },
		{ 52, 501, 37 },
		{ 52, 502, 37 },
		{ 52, 503, 37 },
		{ 52, 504, 37 },
		{ 52, 505, 37 },
		{ 52, 506, 37 },
		{ 52, 507, 37 },
		{ 52, 508, 37 },
		{ 52, 509, 37 },
		{ 52, 510, 37 },
		{ 52, 511, 37 },
		{ 52, 512, 37 },
		{ 53, 462, 37 },
		{ 53, 463, 37 },
		{ 53, 464, 37 },
		{ 53, 465, 37 },
		{ 53, 466, 37 },
		{ 53, 467, 37 },
		{ 53, 468, 37 },
		{ 53, 469, 37 },
		{ 53, 470, 37 },
		{ 53, 471, 37 },
		{ 53, 472, 37 },
		{ 53, 473, 37 },
		{ 53, 474, 37 },
		{ 53, 475, 37 },
		{ 53, 476, 37 },
		{ 53, 477, 37 },
		{ 53, 478, 37 },
		{ 53, 479, 37 },
		{ 53, 480, 37 },
		{ 53, 481, 37 },
		{ 53, 482, 37 },
		{ 53, 483, 37 },
		{ 53, 484, 37 },
		{ 53, 485, 37 },
		{ 53, 486, 37 },
		{ 53, 487, 37 },
		{ 53, 488, 37 },
		{ 53, 489, 37 },
		{ 53, 490, 37 },
		{ 53, 491, 37 },
		{ 53, 492, 37 },
		{ 53, 493, 37 },
		{ 53, 494, 37 },
		{ 53, 495, 37 },
		{ 53, 496, 37 },
		{ 53, 497, 37 },
		{ 53, 498, 37 },
		{ 53, 499, 37 },
		{ 53, 500, 37 },
		{ 53, 501, 37 },
		{ 53, 502, 37 },
		{ 53, 503, 37 },
		{ 53, 504, 37 },
		{ 53, 505, 37 },
		{ 53, 506, 37 },
		{ 53, 507, 37 },
		{ 53, 508, 37 },
		{ 53, 509, 37 },
		{ 53, 510, 37 },
		{ 53, 511, 37 },
		{ 53, 512, 37 },
		{ 54, 461, 37 },
		{ 54, 462, 37 },
		{ 54, 463, 37 },
		{ 54, 464, 37 },
		{ 54, 465, 37 },
		{ 54, 466, 37 },
		{ 54, 467, 37 },
		{ 54, 468, 37 },
		{ 54, 469, 37 },
		{ 54, 470, 37 },
		{ 54, 471, 37 },
		{ 54, 472, 37 },
		{ 54, 473, 37 },
		{ 54, 474, 37 },
		{ 54, 475, 37 },
		{ 54, 476, 37 },
		{ 54, 477, 37 },
		{ 54, 478, 37 },
		{ 54, 479, 37 },
		{ 54, 480, 37 },
		{ 54, 481, 37 },
		{ 54, 482, 37 },
		{ 54, 483, 37 },
		{ 54, 484, 37 },
		{ 54, 485, 37 },
		{ 54, 486, 37 },
		{ 54, 487, 37 },
		{ 54, 488, 37 },
		{ 54, 489, 37 },
		{ 54, 490, 37 },
		{ 54, 491, 37 },
		{ 54, 492, 37 },
		{ 54, 493, 37 },
		{ 54, 494, 37 },
		{ 54, 495, 37 },
		{ 54, 496, 37 },
		{ 54, 497, 37 },
		{ 54, 498, 37 },
		{ 54, 499, 37 },
		{ 54, 500, 37 },
		{ 54, 501, 37 },
		{ 54, 502, 37 },
		{ 54, 503, 37 },
		{ 54, 504, 37 },
		{ 54, 505, 37 },
		{ 54, 506, 37 },
		{ 54, 507, 37 },
		{ 54, 508, 37 },
		{ 54, 509, 37 },
		{ 54, 510, 37 },
		{ 54, 511, 37 },
		{ 54, 512, 37 },
		{ 55, 460, 37 },
		{ 55, 461, 37 },
		{ 55, 462, 37 },
		{ 55, 463, 37 },
		{ 55, 464, 37 },
		{ 55, 465, 37 },
		{ 55, 466, 37 },
		{ 55, 467, 37 },
		{ 55, 468, 37 },
		{ 55, 469, 37 },
		{ 55, 470, 37 },
		{ 55, 471, 37 },
		{ 55, 472, 37 },
		{ 55, 473, 37 },
		{ 55, 474, 37 },
		{ 55, 475, 37 },
		{ 55, 476, 37 },
		{ 55, 477, 37 },
		{ 55, 478, 37 },
		{ 55, 479, 37 },
		{ 55, 480, 37 },
		{ 55, 481, 37 },
		{ 55, 482, 37 },
		{ 55, 483, 37 },
		{ 55, 484, 37 },
		{ 55, 485, 37 },
		{ 55, 486, 37 },
		{ 55, 487, 37 },
		{ 55, 488, 37 },
		{ 55, 489, 37 },
		{ 55, 490, 37 },
		{ 55, 491, 37 },
		{ 55, 492, 37 },
		{ 55, 493, 37 },
		{ 55, 494, 37 },
		{ 55, 495, 37 },
		{ 55, 496, 37 },
		{ 55, 497, 37 },
		{ 55, 498, 37 },
		{ 55, 499, 37 },
		{ 55, 500, 37 },
		{ 55, 501, 37 },
		{ 55, 502, 37 },
		{ 55, 503, 37 },
		{ 55, 504, 37 },
		{ 55, 505, 37 },
		{ 55, 506, 37 },
		{ 55, 507, 37 },
		{ 55, 508, 37 },
		{ 55, 509, 37 },
		{ 55, 510, 37 },
		{ 55, 511, 37 },
		{ 55, 512, 37 },
		{ 56, 459, 37 },
		{ 56, 460, 37 },
		{ 56, 461, 37 },
		{ 56, 462, 37 },
		{ 56, 463, 37 },
		{ 56, 464, 37 },
		{ 56, 465, 37 },
		{ 56, 466, 37 },
		{ 56, 467, 37 },
		{ 56, 468, 37 },
		{ 56, 469, 37 },
		{ 56, 470, 37 },
		{ 56, 471, 37 },
		{ 56, 472, 37 },
		{ 56, 473, 37 },
		{ 56, 474, 37 },
		{ 56, 475, 37 },
		{ 56, 476, 37 },
		{ 56, 477, 37 },
		{ 56, 478, 37 },
		{ 56, 479, 37 },
		{ 56, 480, 37 },
		{ 56, 481, 37 },
		{ 56, 482, 37 },
		{ 56, 483, 37 },
		{ 56, 484, 37 },
		{ 56, 485, 37 },
		{ 56, 486, 37 },
		{ 56, 487, 37 },
		{ 56, 488, 37 },
		{ 56, 489, 37 },
		{ 56, 490, 37 },
		{ 56, 491, 37 },
		{ 56, 492, 37 },
		{ 56, 493, 37 },
		{ 56, 494, 37 },
		{ 56, 495, 37 },
		{ 56, 496, 37 },
		{ 56, 497, 37 },
		{ 56, 498, 37 },
		{ 56, 499, 37 },
		{ 56, 500, 37 },
		{ 56, 501, 37 },
		{ 56, 502, 37 },
		{ 56, 503, 37 },
		{ 56, 504, 37 },
		{ 56, 505, 37 },
		{ 56, 506, 37 },
		{ 56, 507, 37 },
		{ 56, 508, 37 },
		{ 56, 509, 37 },
		{ 56, 510, 37 },
		{ 56, 511, 37 },
		{ 56, 512, 37 },
		{ 57, 458, 37 },
		{ 57, 459, 37 },
		{ 57, 460, 37 },
		{ 57, 461, 37 },
		{ 57, 462, 37 },
		{ 57, 463, 37 },
		{ 57, 464, 37 },
		{ 57, 465, 37 },
		{ 57, 466, 37 },
		{ 57, 467, 37 },
		{ 57, 468, 37 },
		{ 57, 469, 37 },
		{ 57, 470, 37 },
		{ 57, 471, 37 },
		{ 57, 472, 37 },
		{ 57, 473, 37 },
		{ 57, 474, 37 },
		{ 57, 475, 37 },
		{ 57, 476, 37 },
		{ 57, 477, 37 },
		{ 57, 478, 37 },
		{ 57, 479, 37 },
		{ 57, 480, 37 },
		{ 57, 481, 37 },
		{ 57, 482, 37 },
		{ 57, 483, 37 },
		{ 57, 484, 37 },
		{ 57, 485, 37 },
		{ 57, 486, 37 },
		{ 57, 487, 37 },
		{ 57, 488, 37 },
		{ 57, 489, 37 },
		{ 57, 490, 37 },
		{ 57, 491, 37 },
		{ 57, 492, 37 },
		{ 57, 493, 37 },
		{ 57, 494, 37 },
		{ 57, 495, 37 },
		{ 57, 496, 37 },
		{ 57, 497, 37 },
		{ 57, 498, 37 },
		{ 57, 499, 37 },
		{ 57, 500, 37 },
		{ 57, 501, 37 },
		{ 57, 502, 37 },
		{ 57, 503, 37 },
		{ 57, 504, 37 },
		{ 57, 505, 37 },
		{ 57, 506, 37 },
		{ 57, 507, 37 },
		{ 57, 508, 37 },
		{ 57, 509, 37 },
		{ 57, 510, 37 },
		{ 57, 511, 37 },
		{ 57, 512, 37 },
		{ 58, 457, 37 },
		{ 58, 458, 37 },
		{ 58, 459, 37 },
		{ 58, 460, 37 },
		{ 58, 461, 37 },
		{ 58, 462, 37 },
		{ 58, 463, 37 },
		{ 58, 464, 37 },
		{ 58, 465, 37 },
		{ 58, 466, 37 },
		{ 58, 467, 37 },
		{ 58, 468, 37 },
		{ 58, 469, 37 },
		{ 58, 470, 37 },
		{ 58, 471, 37 },
		{ 58, 472, 37 },
		{ 58, 473, 37 },
		{ 58, 474, 37 },
		{ 58, 475, 37 },
		{ 58, 476, 37 },
		{ 58, 477, 37 },
		{ 58, 478, 37 },
		{ 58, 479, 37 },
		{ 58, 480, 37 },
		{ 58, 481, 37 },
		{ 58, 482, 37 },
		{ 58, 483, 37 },
		{ 58, 484, 37 },
		{ 58, 485, 37 },
		{ 58, 486, 37 },
		{ 58, 487, 37 },
		{ 58, 488, 37 },
		{ 58, 489, 37 },
		{ 58, 490, 37 },
		{ 58, 491, 37 },
		{ 58, 492, 37 },
		{ 58, 493, 37 },
		{ 58, 494, 37 },
		{ 58, 495, 37 },
		{ 58, 496, 37 },
		{ 58, 497, 37 },
		{ 58, 498, 37 },
		{ 58, 499, 37 },
		{ 58, 500, 37 },
		{ 58, 501, 37 },
		{ 58, 502, 37 },
		{ 58, 503, 37 },
		{ 58, 504, 37 },
		{ 58, 505, 37 },
		{ 58, 506, 37 },
		{ 58, 507, 37 },
		{ 58, 508, 37 },
		{ 58, 509, 37 },
		{ 58, 510, 37 },
		{ 58, 511, 37 },
		{ 58, 512, 37 },
		{ 59, 456, 37 },
		{ 59, 457, 37 },
		{ 59, 458, 37 },
		{ 59, 459, 37 },
		{ 59, 460, 37 },
		{ 59, 461, 37 },
		{ 59, 462, 37 },
		{ 59, 463, 37 },
		{ 59, 464, 37 },
		{ 59, 465, 37 },
		{ 59, 466, 37 },
		{ 59, 467, 37 },
		{ 59, 468, 37 },
		{ 59, 469, 37 },
		{ 59, 470, 37 },
		{ 59, 471, 37 },
		{ 59, 472, 37 },
		{ 59, 473, 37 },
		{ 59, 474, 37 },
		{ 59, 475, 37 },
		{ 59, 476, 37 },
		{ 59, 477, 37 },
		{ 59, 478, 37 },
		{ 59, 479, 37 },
		{ 59, 480, 37 },
		{ 59, 481, 37 },
		{ 59, 482, 37 },
		{ 59, 483, 37 },
		{ 59, 484, 37 },
		{ 59, 485, 37 },
		{ 59, 486, 37 },
		{ 59, 487, 37 },
		{ 59, 488, 37 },
		{ 59, 489, 37 },
		{ 59, 490, 37 },
		{ 59, 491, 37 },
		{ 59, 492, 37 },
		{ 59, 493, 37 },
		{ 59, 494, 37 },
		{ 59, 495, 37 },
		{ 59, 496, 37 },
		{ 59, 497, 37 },
		{ 59, 498, 37 },
		{ 59, 499, 37 },
		{ 59, 500, 37 },
		{ 59, 501, 37 },
		{ 59, 502, 37 },
		{ 59, 503, 37 },
		{ 59, 504, 37 },
		{ 59, 505, 37 },
		{ 59, 506, 37 },
		{ 59, 507, 37 },
		{ 59, 508, 37 },
		{ 59, 509, 37 },
		{ 59, 510, 37 },
		{ 59, 511, 37 },
		{ 59, 512, 37 },
		{ 60, 455, 37 },
		{ 60, 456, 37 },
		{ 60, 457, 37 },
		{ 60, 458, 37 },
		{ 60, 459, 37 },
		{ 60, 460, 37 },
		{ 60, 461, 37 },
		{ 60, 462, 37 },
		{ 60, 463, 37 },
		{ 60, 464, 37 },
		{ 60, 465, 37 },
		{ 60, 466, 37 },
		{ 60, 467, 37 },
		{ 60, 468, 37 },
		{ 60, 469, 37 },
		{ 60, 470, 37 },
		{ 60, 471, 37 },
		{ 60, 472, 37 },
		{ 60, 473, 37 },
		{ 60, 474, 37 },
		{ 60, 475, 37 },
		{ 60, 476, 37 },
		{ 60, 477, 37 },
		{ 60, 478, 37 },
		{ 60, 479, 37 },
		{ 60, 480, 37 },
		{ 60, 481, 37 },
		{ 60, 482, 37 },
		{ 60, 483, 37 },
		{ 60, 484, 37 },
		{ 60, 485, 37 },
		{ 60, 486, 37 },
		{ 60, 487, 37 },
		{ 60, 488, 37 },
		{ 60, 489, 37 },
		{ 60, 490, 37 },
		{ 60, 491, 37 },
		{ 60, 492, 37 },
		{ 60, 493, 37 },
		{ 60, 494, 37 },
		{ 60, 495, 37 },
		{ 60, 496, 37 },
		{ 60, 497, 37 },
		{ 60, 498, 37 },
		{ 60, 499, 37 },
		{ 60, 500, 37 },
		{ 60, 501, 37 },
		{ 60, 502, 37 },
		{ 60, 503, 37 },
		{ 60, 504, 37 },
		{ 60, 505, 37 },
		{ 60, 506, 37 },
		{ 60, 507, 37 },
		{ 60, 508, 37 },
		{ 60, 509, 37 },
		{ 60, 510, 37 },
		{ 60, 511, 37 },
		{ 60, 512, 37 },
		{ 61, 454, 37 },
		{ 61, 455, 37 },
		{ 61, 456, 37 },
		{ 61, 457, 37 },
		{ 61, 458, 37 },
		{ 61, 459, 37 },
		{ 61, 460, 37 },
		{ 61, 461, 37 },
		{ 61, 462, 37 },
		{ 61, 463, 37 },
		{ 61, 464, 37 },
		{ 61, 465, 37 },
		{ 61, 466, 37 },
		{ 61, 467, 37 },
		{ 61, 468, 37 },
		{ 61, 469, 37 },
		{ 61, 470, 37 },
		{ 61, 471, 37 },
		{ 61, 472, 37 },
		{ 61, 473, 37 },
		{ 61, 474, 37 },
		{ 61, 475, 37 },
		{ 61, 476, 37 },
		{ 61, 477, 37 },
		{ 61, 478, 37 },
		{ 61, 479, 37 },
		{ 61, 480, 37 },
		{ 61, 481, 37 },
		{ 61, 482, 37 },
		{ 61, 483, 37 },
		{ 61, 484, 37 },
		{ 61, 485, 37 },
		{ 61, 486, 37 },
		{ 61, 487, 37 },
		{ 61, 488, 37 },
		{ 61, 489, 37 },
		{ 61, 490, 37 },
		{ 61, 491, 37 },
		{ 61, 492, 37 },
		{ 61, 493, 37 },
		{ 61, 494, 37 },
		{ 61, 495, 37 },
		{ 61, 496, 37 },
		{ 61, 497, 37 },
		{ 61, 498, 37 },
		{ 61, 499, 37 },
		{ 61, 500, 37 },
		{ 61, 501, 37 },
		{ 61, 502, 37 },
		{ 61, 503, 37 },
		{ 61, 504, 37 },
		{ 61, 505, 37 },
		{ 61, 506, 37 },
		{ 61, 507, 37 },
		{ 61, 508, 37 },
		{ 61, 509, 37 },
		{ 61, 510, 37 },
		{ 61, 511, 37 },
		{ 61, 512, 37 },
		{ 62, 453, 37 },
		{ 62, 454, 37 },
		{ 62, 455, 37 },
		{ 62, 456, 37 },
		{ 62, 457, 37 },
		{ 62, 458, 37 },
		{ 62, 459, 37 },
		{ 62, 460, 37 },
		{ 62, 461, 37 },
		{ 62, 462, 37 },
		{ 62, 463, 37 },
		{ 62, 464, 37 },
		{ 62, 465, 37 },
		{ 62, 466, 37 },
		{ 62, 467, 37 },
		{ 62, 468, 37 },
		{ 62, 469, 37 },
		{ 62, 470, 37 },
		{ 62, 471, 37 },
		{ 62, 472, 37 },
		{ 62, 473, 37 },
		{ 62, 474, 37 },
		{ 62, 475, 37 },
		{ 62, 476, 37 },
		{ 62, 477, 37 },
		{ 62, 478, 37 },
		{ 62, 479, 37 },
		{ 62, 480, 37 },
		{ 62, 481, 37 },
		{ 62, 482, 37 },
		{ 62, 483, 37 },
		{ 62, 484, 37 },
		{ 62, 485, 37 },
		{ 62, 486, 37 },
		{ 62, 487, 37 },
		{ 62, 488, 37 },
		{ 62, 489, 37 },
		{ 62, 490, 37 },
		{ 62, 491, 37 },
		{ 62, 492, 37 },
		{ 62, 493, 37 },
		{ 62, 494, 37 },
		{ 62, 495, 37 },
		{ 62, 496, 37 },
		{ 62, 497, 37 },
		{ 62, 498, 37 },
		{ 62, 499, 37 },
		{ 62, 500, 37 },
		{ 62, 501, 37 },
		{ 62, 502, 37 },
		{ 62, 503, 37 },
		{ 62, 504, 37 },
		{ 62, 505, 37 },
		{ 62, 506, 37 },
		{ 62, 507, 37 },
		{ 62, 508, 37 },
		{ 62, 509, 37 },
		{ 62, 510, 37 },
		{ 62, 511, 37 },
		{ 62, 512, 37 },
		{ 63, 452, 37 },
		{ 63, 453, 37 },
		{ 63, 454, 37 },
		{ 63, 455, 37 },
		{ 63, 456, 37 },
		{ 63, 457, 37 },
		{ 63, 458, 37 },
		{ 63, 459, 37 },
		{ 63, 460, 37 },
		{ 63, 461, 37 },
		{ 63, 462, 37 },
		{ 63, 463, 37 },
		{ 63, 464, 37 },
		{ 63, 465, 37 },
		{ 63, 466, 37 },
		{ 63, 467, 37 },
		{ 63, 468, 37 },
		{ 63, 469, 37 },
		{ 63, 470, 37 },
		{ 63, 471, 37 },
		{ 63, 472, 37 },
		{ 63, 473, 37 },
		{ 63, 474, 37 },
		{ 63, 475, 37 },
		{ 63, 476, 37 },
		{ 63, 477, 37 },
		{ 63, 478, 37 },
		{ 63, 479, 37 },
		{ 63, 480, 37 },
		{ 63, 481, 37 },
		{ 63, 482, 37 },
		{ 63, 483, 37 },
		{ 63, 484, 37 },
		{ 63, 485, 37 },
		{ 63, 486, 37 },
		{ 63, 487, 37 },
		{ 63, 488, 37 },
		{ 63, 489, 37 },
		{ 63, 490, 37 },
		{ 63, 491, 37 },
		{ 63, 492, 37 },
		{ 63, 493, 37 },
		{ 63, 494, 37 },
		{ 63, 495, 37 },
		{ 63, 496, 37 },
		{ 63, 497, 37 },
		{ 63, 498, 37 },
		{ 63, 499, 37 },
		{ 63, 500, 37 },
		{ 63, 501, 37 },
		{ 63, 502, 37 },
		{ 63, 503, 37 },
		{ 63, 504, 37 },
		{ 63, 505, 37 },
		{ 63, 506, 37 },
		{ 63, 507, 37 },
		{ 63, 508, 37 },
		{ 63, 509, 37 },
		{ 63, 510, 37 },
		{ 63, 511, 37 },
		{ 63, 512, 37 },
		{ 64, 451, 37 },
		{ 64, 452, 37 },
		{ 64, 453, 37 },
		{ 64, 454, 37 },
		{ 64, 455, 37 },
		{ 64, 456, 37 },
		{ 64, 457, 37 },
		{ 64, 458, 37 },
		{ 64, 459, 37 },
		{ 64, 460, 37 },
		{ 64, 461, 37 },
		{ 64, 462, 37 },
		{ 64, 463, 37 },
		{ 64, 464, 37 },
		{ 64, 465, 37 },
		{ 64, 466, 37 },
		{ 64, 467, 37 },
		{ 64, 468, 37 },
		{ 64, 469, 37 },
		{ 64, 470, 37 },
		{ 64, 471, 37 },
		{ 64, 472, 37 },
		{ 64, 473, 37 },
		{ 64, 474, 37 },
		{ 64, 475, 37 },
		{ 64, 476, 37 },
		{ 64, 477, 37 },
		{ 64, 478, 37 },
		{ 64, 479, 37 },
		{ 64, 480, 37 },
		{ 64, 481, 37 },
		{ 64, 482, 37 },
		{ 64, 483, 37 },
		{ 64, 484, 37 },
		{ 64, 485, 37 },
		{ 64, 486, 37 },
		{ 64, 487, 37 },
		{ 64, 488, 37 },
		{ 64, 489, 37 },
		{ 64, 490, 37 },
		{ 64, 491, 37 },
		{ 64, 492, 37 },
		{ 64, 493, 37 },
		{ 64, 494, 37 },
		{ 64, 495, 37 },
		{ 64, 496, 37 },
		{ 64, 497, 37 },
		{ 64, 498, 37 },
		{ 64, 499, 37 },
		{ 64, 500, 37 },
		{ 64, 501, 37 },
		{ 64, 502, 37 },
		{ 64, 503, 37 },
		{ 64, 504, 37 },
		{ 64, 505, 37 },
		{ 64, 506, 37 },
		{ 64, 507, 37 },
		{ 64, 508, 37 },
		{ 64, 509, 37 },
		{ 64, 510, 37 },
		{ 64, 511, 37 },
		{ 64, 512, 37 },
		{ 65, 450, 37 },
		{ 65, 451, 37 },
		{ 65, 452, 37 },
		{ 65, 453, 37 },
		{ 65, 454, 37 },
		{ 65, 455, 37 },
		{ 65, 456, 37 },
		{ 65, 457, 37 },
		{ 65, 458, 37 },
		{ 65, 459, 37 },
		{ 65, 460, 37 },
		{ 65, 461, 37 },
		{ 65, 462, 37 },
		{ 65, 463, 37 },
		{ 65, 464, 37 },
		{ 65, 465, 37 },
		{ 65, 466, 37 },
		{ 65, 467, 37 },
		{ 65, 468, 37 },
		{ 65, 469, 37 },
		{ 65, 470, 37 },
		{ 65, 471, 37 },
		{ 65, 472, 37 },
		{ 65, 473, 37 },
		{ 65, 474, 37 },
		{ 65, 475, 37 },
		{ 65, 476, 37 },
		{ 65, 477, 37 },
		{ 65, 478, 37 },
		{ 65, 479, 37 },
		{ 65, 480, 37 },
		{ 65, 481, 37 },
		{ 65, 482, 37 },
		{ 65, 483, 37 },
		{ 65, 484, 37 },
		{ 65, 485, 37 },
		{ 65, 486, 37 },
		{ 65, 487, 37 },
		{ 65, 488, 37 },
		{ 65, 489, 37 },
		{ 65, 490, 37 },
		{ 65, 491, 37 },
		{ 65, 492, 37 },
		{ 65, 493, 37 },
		{ 65, 494, 37 },
		{ 65, 495, 37 },
		{ 65, 496, 37 },
		{ 65, 497, 37 },
		{ 65, 498, 37 },
		{ 65, 499, 37 },
		{ 65, 500, 37 },
		{ 65, 501, 37 },
		{ 65, 502, 37 },
		{ 65, 503, 37 },
		{ 65, 504, 37 },
		{ 65, 505, 37 },
		{ 65, 506, 37 },
		{ 65, 507, 37 },
		{ 65, 508, 37 },
		{ 65, 509, 37 },
		{ 65, 510, 37 },
		{ 65, 511, 37 },
		{ 65, 512, 37 },
		{ 66, 449, 37 },
		{ 66, 450, 37 },
		{ 66, 451, 37 },
		{ 66, 452, 37 },
		{ 66, 453, 37 },
		{ 66, 454, 37 },
		{ 66, 455, 37 },
		{ 66, 456, 37 },
		{ 66, 457, 37 },
		{ 66, 458, 37 },
		{ 66, 459, 37 },
		{ 66, 460, 37 },
		{ 66, 461, 37 },
		{ 66, 462, 37 },
		{ 66, 463, 37 },
		{ 66, 464, 37 },
		{ 66, 465, 37 },
		{ 66, 466, 37 },
		{ 66, 467, 37 },
		{ 66, 468, 37 },
		{ 66, 469, 37 },
		{ 66, 470, 37 },
		{ 66, 471, 37 },
		{ 66, 472, 37 },
		{ 66, 473, 37 },
		{ 66, 474, 37 },
		{ 66, 475, 37 },
		{ 66, 476, 37 },
		{ 66, 477, 37 },
		{ 66, 478, 37 },
		{ 66, 479, 37 },
		{ 66, 480, 37 },
		{ 66, 481, 37 },
		{ 66, 482, 37 },
		{ 66, 483, 37 },
		{ 66, 484, 37 },
		{ 66, 485, 37 },
		{ 66, 486, 37 },
		{ 66, 487, 37 },
		{ 66, 488, 37 },
		{ 66, 489, 37 },
		{ 66, 490, 37 },
		{ 66, 491, 37 },
		{ 66, 492, 37 },
		{ 66, 493, 37 },
		{ 66, 494, 37 },
		{ 66, 495, 37 },
		{ 66, 496, 37 },
		{ 66, 497, 37 },
		{ 66, 498, 37 },
		{ 66, 499, 37 },
		{ 66, 500, 37 },
		{ 66, 501, 37 },
		{ 66, 502, 37 },
		{ 66, 503, 37 },
		{ 66, 504, 37 },
		{ 66, 505, 37 },
		{ 66, 506, 37 },
		{ 66, 507, 37 },
		{ 66, 508, 37 },
		{ 66, 509, 37 },
		{ 66, 510, 37 },
		{ 66, 511, 37 },
		{ 67, 448, 37 },
		{ 67, 449, 37 },
		{ 67, 450, 37 },
		{ 67, 451, 37 },
		{ 67, 452, 37 },
		{ 67, 453, 37 },
		{ 67, 454, 37 },
		{ 67, 455, 37 },
		{ 67, 456, 37 },
		{ 67, 457, 37 },
		{ 67, 458, 37 },
		{ 67, 459, 37 },
		{ 67, 460, 37 },
		{ 67, 461, 37 },
		{ 67, 462, 37 },
		{ 67, 463, 37 },
		{ 67, 464, 37 },
		{ 67, 465, 37 },
		{ 67, 466, 37 },
		{ 67, 467, 37 },
		{ 67, 468, 37 },
		{ 67, 469, 37 },
		{ 67, 470, 37 },
		{ 67, 471, 37 },
		{ 67, 472, 37 },
		{ 67, 473, 37 },
		{ 67, 474, 37 },
		{ 67, 475, 37 },
		{ 67, 476, 37 },
		{ 67, 477, 37 },
		{ 67, 478, 37 },
		{ 67, 479, 37 },
		{ 67, 480, 37 },
		{ 67, 481, 37 },
		{ 67, 482, 37 },
		{ 67, 483, 37 },
		{ 67, 484, 37 },
		{ 67, 485, 37 },
		{ 67, 486, 37 },
		{ 67, 487, 37 },
		{ 67, 488, 37 },
		{ 67, 489, 37 },
		{ 67, 490, 37 },
		{ 67, 491, 37 },
		{ 67, 492, 37 },
		{ 67, 493, 37 },
		{ 67, 494, 37 },
		{ 67, 495, 37 },
		{ 67, 496, 37 },
		{ 67, 497, 37 },
		{ 67, 498, 37 },
		{ 67, 499, 37 },
		{ 67, 500, 37 },
		{ 67, 501, 37 },
		{ 67, 502, 37 },
		{ 67, 503, 37 },
		{ 67, 504, 37 },
		{ 67, 505, 37 },
		{ 67, 506, 37 },
		{ 67, 507, 37 },
		{ 67, 508, 37 },
		{ 67, 509, 37 },
		{ 67, 510, 37 },
		{ 68, 447, 37 },
		{ 68, 448, 37 },
		{ 68, 449, 37 },
		{ 68, 450, 37 },
		{ 68, 451, 37 },
		{ 68, 452, 37 },
		{ 68, 453, 37 },
		{ 68, 454, 37 },
		{ 68, 455, 37 },
		{ 68, 456, 37 },
		{ 68, 457, 37 },
		{ 68, 458, 37 },
		{ 68, 459, 37 },
		{ 68, 460, 37 },
		{ 68, 461, 37 },
		{ 68, 462, 37 },
		{ 68, 463, 37 },
		{ 68, 464, 37 },
		{ 68, 465, 37 },
		{ 68, 466, 37 },
		{ 68, 467, 37 },
		{ 68, 468, 37 },
		{ 68, 469, 37 },
		{ 68, 470, 37 },
		{ 68, 471, 37 },
		{ 68, 472, 37 },
		{ 68, 473, 37 },
		{ 68, 474, 37 },
		{ 68, 475, 37 },
		{ 68, 476, 37 },
		{ 68, 477, 37 },
		{ 68, 478, 37 },
		{ 68, 479, 37 },
		{ 68, 480, 37 },
		{ 68, 481, 37 },
		{ 68, 482, 37 },
		{ 68, 483, 37 },
		{ 68, 484, 37 },
		{ 68, 485, 37 },
		{ 68, 486, 37 },
		{ 68, 487, 37 },
		{ 68, 488, 37 },
		{ 68, 489, 37 },
		{ 68, 490, 37 },
		{ 68, 491, 37 },
		{ 68, 492, 37 },
		{ 68, 493, 37 },
		{ 68, 494, 37 },
		{ 68, 495, 37 },
		{ 68, 496, 37 },
		{ 68, 497, 37 },
		{ 68, 498, 37 },
		{ 68, 499, 37 },
		{ 68, 500, 37 },
		{ 68, 501, 37 },
		{ 68, 502, 37 },
		{ 68, 503, 37 },
		{ 68, 504, 37 },
		{ 68, 505, 37 },
		{ 68, 506, 37 },
		{ 68, 507, 37 },
		{ 68, 508, 37 },
		{ 69, 446, 37 },
		{ 69, 447, 37 },
		{ 69, 448, 37 },
		{ 69, 449, 37 },
		{ 69, 450, 37 },
		{ 69, 451, 37 },
		{ 69, 452, 37 },
		{ 69, 453, 37 },
		{ 69, 454, 37 },
		{ 69, 455, 37 },
		{ 69, 456, 37 },
		{ 69, 457, 37 },
		{ 69, 458, 37 },
		{ 69, 459, 37 },
		{ 69, 460, 37 },
		{ 69, 461, 37 },
		{ 69, 462, 37 },
		{ 69, 463, 37 },
		{ 69, 464, 37 },
		{ 69, 465, 37 },
		{ 69, 466, 37 },
		{ 69, 467, 37 },
		{ 69, 468, 37 },
		{ 69, 469, 37 },
		{ 69, 470, 37 },
		{ 69, 471, 37 },
		{ 69, 472, 37 },
		{ 69, 473, 37 },
		{ 69, 474, 37 },
		{ 69, 475, 37 },
		{ 69, 476, 37 },
		{ 69, 477, 37 },
		{ 69, 478, 37 },
		{ 69, 479, 37 },
		{ 69, 480, 37 },
		{ 69, 481, 37 },
		{ 69, 482, 37 },
		{ 69, 483, 37 },
		{ 69, 484, 37 },
		{ 69, 485, 37 },
		{ 69, 486, 37 },
		{ 69, 487, 37 },
		{ 69, 488, 37 },
		{ 69, 489, 37 },
		{ 69, 490, 37 },
		{ 69, 491, 37 },
		{ 69, 492, 37 },
		{ 69, 493, 37 },
		{ 69, 494, 37 },
		{ 69, 495, 37 },
		{ 69, 496, 37 },
		{ 69, 497, 37 },
		{ 69, 498, 37 },
		{ 69, 499, 37 },
		{ 69, 500, 37 },
		{ 69, 501, 37 },
		{ 69, 502, 37 },
		{ 69, 503, 37 },
		{ 69, 504, 37 },
		{ 69, 505, 37 },
		{ 69, 506, 37 },
		{ 69, 507, 37 },
		{ 70, 445, 37 },
		{ 70, 446, 37 },
		{ 70, 447, 37 },
		{ 70, 448, 37 },
		{ 70, 449, 37 },
		{ 70, 450, 37 },
		{ 70, 451, 37 },
		{ 70, 452, 37 },
		{ 70, 453, 37 },
		{ 70, 454, 37 },
		{ 70, 455, 37 },
		{ 70, 456, 37 },
		{ 70, 457, 37 },
		{ 70, 458, 37 },
		{ 70, 459, 37 },
		{ 70, 460, 37 },
		{ 70, 461, 37 },
		{ 70, 462, 37 },
		{ 70, 463, 37 },
		{ 70, 464, 37 },
		{ 70, 465, 37 },
		{ 70, 466, 37 },
		{ 70, 467, 37 },
		{ 70, 468, 37 },
		{ 70, 469, 37 },
		{ 70, 470, 37 },
		{ 70, 471, 37 },
		{ 70, 472, 37 },
		{ 70, 473, 37 },
		{ 70, 474, 37 },
		{ 70, 475, 37 },
		{ 70, 476, 37 },
		{ 70, 477, 37 },
		{ 70, 478, 37 },
		{ 70, 479, 37 },
		{ 70, 480, 37 },
		{ 70, 481, 37 },
		{ 70, 482, 37 },
		{ 70, 483, 37 },
		{ 70, 484, 37 },
		{ 70, 485, 37 },
		{ 70, 486, 37 },
		{ 70, 487, 37 },
		{ 70, 488, 37 },
		{ 70, 489, 37 },
		{ 70, 490, 37 },
		{ 70, 491, 37 },
		{ 70, 492, 37 },
		{ 70, 493, 37 },
		{ 70, 494, 37 },
		{ 70, 495, 37 },
		{ 70, 496, 37 },
		{ 70, 497, 37 },
		{ 70, 498, 37 },
		{ 70, 499, 37 },
		{ 70, 500, 37 },
		{ 70, 501, 37 },
		{ 70, 502, 37 },
		{ 70, 503, 37 },
		{ 70, 504, 37 },
		{ 70, 505, 37 },
		{ 70, 506, 37 },
		{ 71, 444, 37 },
		{ 71, 445, 37 },
		{ 71, 446, 37 },
		{ 71, 447, 37 },
		{ 71, 448, 37 },
		{ 71, 449, 37 },
		{ 71, 450, 37 },
		{ 71, 451, 37 },
		{ 71, 452, 37 },
		{ 71, 453, 37 },
		{ 71, 454, 37 },
		{ 71, 455, 37 },
		{ 71, 456, 37 },
		{ 71, 457, 37 },
		{ 71, 458, 37 },
		{ 71, 459, 37 },
		{ 71, 460, 37 },
		{ 71, 461, 37 },
		{ 71, 462, 37 },
		{ 71, 463, 37 },
		{ 71, 464, 37 },
		{ 71, 465, 37 },
		{ 71, 466, 37 },
		{ 71, 467, 37 },
		{ 71, 468, 37 },
		{ 71, 469, 37 },
		{ 71, 470, 37 },
		{ 71, 471, 37 },
		{ 71, 472, 37 },
		{ 71, 473, 37 },
		{ 71, 474, 37 },
		{ 71, 475, 37 },
		{ 71, 476, 37 },
		{ 71, 477, 37 },
		{ 71, 478, 37 },
		{ 71, 479, 37 },
		{ 71, 480, 37 },
		{ 71, 481, 37 },
		{ 71, 482, 37 },
		{ 71, 483, 37 },
		{ 71, 484, 37 },
		{ 71, 485, 37 },
		{ 71, 486, 37 },
		{ 71, 487, 37 },
		{ 71, 488, 37 },
		{ 71, 489, 37 },
		{ 71, 490, 37 },
		{ 71, 491, 37 },
		{ 71, 492, 37 },
		{ 71, 493, 37 },
		{ 71, 494, 37 },
		{ 71, 495, 37 },
		{ 71, 496, 37 },
		{ 71, 497, 37 },
		{ 71, 498, 37 },
		{ 71, 499, 37 },
		{ 71, 500, 37 },
		{ 71, 501, 37 },
		{ 71, 502, 37 },
		{ 71, 503, 37 },
		{ 71, 504, 37 },
		{ 72, 443, 37 },
		{ 72, 444, 37 },
		{ 72, 445, 37 },
		{ 72, 446, 37 },
		{ 72, 447, 37 },
		{ 72, 448, 37 },
		{ 72, 449, 37 },
		{ 72, 450, 37 },
		{ 72, 451, 37 },
		{ 72, 452, 37 },
		{ 72, 453, 37 },
		{ 72, 454, 37 },
		{ 72, 455, 37 },
		{ 72, 456, 37 },
		{ 72, 457, 37 },
		{ 72, 458, 37 },
		{ 72, 459, 37 },
		{ 72, 460, 37 },
		{ 72, 461, 37 },
		{ 72, 462, 37 },
		{ 72, 463, 37 },
		{ 72, 464, 37 },
		{ 72, 465, 37 },
		{ 72, 466, 37 },
		{ 72, 467, 37 },
		{ 72, 468, 37 },
		{ 72, 469, 37 },
		{ 72, 470, 37 },
		{ 72, 471, 37 },
		{ 72, 472, 37 },
		{ 72, 473, 37 },
		{ 72, 474, 37 },
		{ 72, 475, 37 },
		{ 72, 476, 37 },
		{ 72, 477, 37 },
		{ 72, 478, 37 },
		{ 72, 479, 37 },
		{ 72, 480, 37 },
		{ 72, 481, 37 },
		{ 72, 482, 37 },
		{ 72, 483, 37 },
		{ 72, 484, 37 },
		{ 72, 485, 37 },
		{ 72, 486, 37 },
		{ 72, 487, 37 },
		{ 72, 488, 37 },
		{ 72, 489, 37 },
		{ 72, 490, 37 },
		{ 72, 491, 37 },
		{ 72, 492, 37 },
		{ 72, 493, 37 },
		{ 72, 494, 37 },
		{ 72, 495, 37 },
		{ 72, 496, 37 },
		{ 72, 497, 37 },
		{ 72, 498, 37 },
		{ 72, 499, 37 },
		{ 72, 500, 37 },
		{ 72, 501, 37 },
		{ 72, 502, 37 },
		{ 72, 503, 37 },
		{ 73, 442, 37 },
		{ 73, 443, 37 },
		{ 73, 444, 37 },
		{ 73, 445, 37 },
		{ 73, 446, 37 },
		{ 73, 447, 37 },
		{ 73, 448, 37 },
		{ 73, 449, 37 },
		{ 73, 450, 37 },
		{ 73, 451, 37 },
		{ 73, 452, 37 },
		{ 73, 453, 37 },
		{ 73, 454, 37 },
		{ 73, 455, 37 },
		{ 73, 456, 37 },
		{ 73, 457, 37 },
		{ 73, 458, 37 },
		{ 73, 459, 37 },
		{ 73, 460, 37 },
		{ 73, 461, 37 },
		{ 73, 462, 37 },
		{ 73, 463, 37 },
		{ 73, 464, 37 },
		{ 73, 465, 37 },
		{ 73, 466, 37 },
		{ 73, 467, 37 },
		{ 73, 468, 37 },
		{ 73, 469, 37 },
		{ 73, 470, 37 },
		{ 73, 471, 37 },
		{ 73, 472, 37 },
		{ 73, 473, 37 },
		{ 73, 474, 37 },
		{ 73, 475, 37 },
		{ 73, 476, 37 },
		{ 73, 477, 37 },
		{ 73, 478, 37 },
		{ 73, 479, 37 },
		{ 73, 480, 37 },
		{ 73, 481, 37 },
		{ 73, 482, 37 },
		{ 73, 483, 37 },
		{ 73, 484, 37 },
		{ 73, 485, 37 },
		{ 73, 486, 37 },
		{ 73, 487, 37 },
		{ 73, 488, 37 },
		{ 73, 489, 37 },
		{ 73, 490, 37 },
		{ 73, 491, 37 },
		{ 73, 492, 37 },
		{ 73, 493, 37 },
		{ 73, 494, 37 },
		{ 73, 495, 37 },
		{ 73, 496, 37 },
		{ 73, 497, 37 },
		{ 73, 498, 37 },
		{ 73, 499, 37 },
		{ 73, 500, 37 },
		{ 73, 501, 37 },
		{ 73, 502, 37 },
		{ 74, 441, 37 },
		{ 74, 442, 37 },
		{ 74, 443, 37 },
		{ 74, 444, 37 },
		{ 74, 445, 37 },
		{ 74, 446, 37 },
		{ 74, 447, 37 },
		{ 74, 448, 37 },
		{ 74, 449, 37 },
		{ 74, 450, 37 },
		{ 74, 451, 37 },
		{ 74, 452, 37 },
		{ 74, 453, 37 },
		{ 74, 454, 37 },
		{ 74, 455, 37 },
		{ 74, 456, 37 },
		{ 74, 457, 37 },
		{ 74, 458, 37 },
		{ 74, 459, 37 },
		{ 74, 460, 37 },
		{ 74, 461, 37 },
		{ 74, 462, 37 },
		{ 74, 463, 37 },
		{ 74, 464, 37 },
		{ 74, 465, 37 },
		{ 74, 466, 37 },
		{ 74, 467, 37 },
		{ 74, 468, 37 },
		{ 74, 469, 37 },
		{ 74, 470, 37 },
		{ 74, 471, 37 },
		{ 74, 472, 37 },
		{ 74, 473, 37 },
		{ 74, 474, 37 },
		{ 74, 475, 37 },
		{ 74, 476, 37 },
		{ 74, 477, 37 },
		{ 74, 478, 37 },
		{ 74, 479, 37 },
		{ 74, 480, 37 },
		{ 74, 481, 37 },
		{ 74, 482, 37 },
		{ 74, 483, 37 },
		{ 74, 484, 37 },
		{ 74, 485, 37 },
		{ 74, 486, 37 },
		{ 74, 487, 37 },
		{ 74, 488, 37 },
		{ 74, 489, 37 },
		{ 74, 490, 37 },
		{ 74, 491, 37 },
		{ 74, 492, 37 },
		{ 74, 493, 37 },
		{ 74, 494, 37 },
		{ 74, 495, 37 },
		{ 74, 496, 37 },
		{ 74, 497, 37 },
		{ 74, 498, 37 },
		{ 74, 499, 37 },
		{ 74, 500, 37 },
		{ 75, 440, 37 },
		{ 75, 441, 37 },
		{ 75, 442, 37 },
		{ 75, 443, 37 },
		{ 75, 444, 37 },
		{ 75, 445, 37 },
		{ 75, 446, 37 },
		{ 75, 447, 37 },
		{ 75, 448, 37 },
		{ 75, 449, 37 },
		{ 75, 450, 37 },
		{ 75, 451, 37 },
		{ 75, 452, 37 },
		{ 75, 453, 37 },
		{ 75, 454, 37 },
		{ 75, 455, 37 },
		{ 75, 456, 37 },
		{ 75, 457, 37 },
		{ 75, 458, 37 },
		{ 75, 459, 37 },
		{ 75, 460, 37 },
		{ 75, 461, 37 },
		{ 75, 462, 37 },
		{ 75, 463, 37 },
		{ 75, 464, 37 },
		{ 75, 465, 37 },
		{ 75, 466, 37 },
		{ 75, 467, 37 },
		{ 75, 468, 37 },
		{ 75, 469, 37 },
		{ 75, 470, 37 },
		{ 75, 471, 37 },
		{ 75, 472, 37 },
		{ 75, 473, 37 },
		{ 75, 474, 37 },
		{ 75, 475, 37 },
		{ 75, 476, 37 },
		{ 75, 477, 37 },
		{ 75, 478, 37 },
		{ 75, 479, 37 },
		{ 75, 480, 37 },
		{ 75, 481, 37 },
		{ 75, 482, 37 },
		{ 75, 483, 37 },
		{ 75, 484, 37 },
		{ 75, 485, 37 },
		{ 75, 486, 37 },
		{ 75, 487, 37 },
		{ 75, 488, 37 },
		{ 75, 489, 37 },
		{ 75, 490, 37 },
		{ 75, 491, 37 },
		{ 75, 492, 37 },
		{ 75, 493, 37 },
		{ 75, 494, 37 },
		{ 75, 495, 37 },
		{ 75, 496, 37 },
		{ 75, 497, 37 },
		{ 75, 498, 37 },
		{ 75, 499, 37 },
		{ 76, 439, 37 },
		{ 76, 440, 37 },
		{ 76, 441, 37 },
		{ 76, 442, 37 },
		{ 76, 443, 37 },
		{ 76, 444, 37 },
		{ 76, 445, 37 },
		{ 76, 446, 37 },
		{ 76, 447, 37 },
		{ 76, 448, 37 },
		{ 76, 449, 37 },
		{ 76, 450, 37 },
		{ 76, 451, 37 },
		{ 76, 452, 37 },
		{ 76, 453, 37 },
		{ 76, 454, 37 },
		{ 76, 455, 37 },
		{ 76, 456, 37 },
		{ 76, 457, 37 },
		{ 76, 458, 37 },
		{ 76, 459, 37 },
		{ 76, 460, 37 },
		{ 76, 461, 37 },
		{ 76, 462, 37 },
		{ 76, 463, 37 },
		{ 76, 464, 37 },
		{ 76, 465, 37 },
		{ 76, 466, 37 },
		{ 76, 467, 37 },
		{ 76, 468, 37 },
		{ 76, 469, 37 },
		{ 76, 470, 37 },
		{ 76, 471, 37 },
		{ 76, 472, 37 },
		{ 76, 473, 37 },
		{ 76, 474, 37 },
		{ 76, 475, 37 },
		{ 76, 476, 37 },
		{ 76, 477, 37 },
		{ 76, 478, 37 },
		{ 76, 479, 37 },
		{ 76, 480, 37 },
		{ 76, 481, 37 },
		{ 76, 482, 37 },
		{ 76, 483, 37 },
		{ 76, 484, 37 },
		{ 76, 485, 37 },
		{ 76, 486, 37 },
		{ 76, 487, 37 },
		{ 76, 488, 37 },
		{ 76, 489, 37 },
		{ 76, 490, 37 },
		{ 76, 491, 37 },
		{ 76, 492, 37 },
		{ 76, 493, 37 },
		{ 76, 494, 37 },
		{ 76, 495, 37 },
		{ 76, 496, 37 },
		{ 76, 497, 37 },
		{ 76, 498, 37 },
		{ 77, 438, 37 },
		{ 77, 439, 37 },
		{ 77, 440, 37 },
		{ 77, 441, 37 },
		{ 77, 442, 37 },
		{ 77, 443, 37 },
		{ 77, 444, 37 },
		{ 77, 445, 37 },
		{ 77, 446, 37 },
		{ 77, 447, 37 },
		{ 77, 448, 37 },
		{ 77, 449, 37 },
		{ 77, 450, 37 },
		{ 77, 451, 37 },
		{ 77, 452, 37 },
		{ 77, 453, 37 },
		{ 77, 454, 37 },
		{ 77, 455, 37 },
		{ 77, 456, 37 },
		{ 77, 457, 37 },
		{ 77, 458, 37 },
		{ 77, 459, 37 },
		{ 77, 460, 37 },
		{ 77, 461, 37 },
		{ 77, 462, 37 },
		{ 77, 463, 37 },
		{ 77, 464, 37 },
		{ 77, 465, 37 },
		{ 77, 466, 37 },
		{ 77, 467, 37 },
		{ 77, 468, 37 },
		{ 77, 469, 37 },
		{ 77, 470, 37 },
		{ 77, 471, 37 },
		{ 77, 472, 37 },
		{ 77, 473, 37 },
		{ 77, 474, 37 },
		{ 77, 475, 37 },
		{ 77, 476, 37 },
		{ 77, 477, 37 },
		{ 77, 478, 37 },
		{ 77, 479, 37 },
		{ 77, 480, 37 },
		{ 77, 481, 37 },
		{ 77, 482, 37 },
		{ 77, 483, 37 },
		{ 77, 484, 37 },
		{ 77, 485, 37 },
		{ 77, 486, 37 },
		{ 77, 487, 37 },
		{ 77, 488, 37 },
		{ 77, 489, 37 },
		{ 77, 490, 37 },
		{ 77, 491, 37 },
		{ 77, 492, 37 },
		{ 77, 493, 37 },
		{ 77, 494, 37 },
		{ 77, 495, 37 },
		{ 77, 496, 37 },
		{ 78, 437, 37 },
		{ 78, 438, 37 },
		{ 78, 439, 37 },
		{ 78, 440, 37 },
		{ 78, 441, 37 },
		{ 78, 442, 37 },
		{ 78, 443, 37 },
		{ 78, 444, 37 },
		{ 78, 445, 37 },
		{ 78, 446, 37 },
		{ 78, 447, 37 },
		{ 78, 448, 37 },
		{ 78, 449, 37 },
		{ 78, 450, 37 },
		{ 78, 451, 37 },
		{ 78, 452, 37 },
		{ 78, 453, 37 },
		{ 78, 454, 37 },
		{ 78, 455, 37 },
		{ 78, 456, 37 },
		{ 78, 457, 37 },
		{ 78, 458, 37 },
		{ 78, 459, 37 },
		{ 78, 460, 37 },
		{ 78, 461, 37 },
		{ 78, 462, 37 },
		{ 78, 463, 37 },
		{ 78, 464, 37 },
		{ 78, 465, 37 },
		{ 78, 466, 37 },
		{ 78, 467, 37 },
		{ 78, 468, 37 },
		{ 78, 469, 37 },
		{ 78, 470, 37 },
		{ 78, 471, 37 },
		{ 78, 472, 37 },
		{ 78, 473, 37 },
		{ 78, 474, 37 },
		{ 78, 475, 37 },
		{ 78, 476, 37 },
		{ 78, 477, 37 },
		{ 78, 478, 37 },
		{ 78, 479, 37 },
		{ 78, 480, 37 },
		{ 78, 481, 37 },
		{ 78, 482, 37 },
		{ 78, 483, 37 },
		{ 78, 484, 37 },
		{ 78, 485, 37 },
		{ 78, 486, 37 },
		{ 78, 487, 37 },
		{ 78, 488, 37 },
		{ 78, 489, 37 },
		{ 78, 490, 37 },
		{ 78, 491, 37 },
		{ 78, 492, 37 },
		{ 78, 493, 37 },
		{ 78, 494, 37 },
		{ 78, 495, 37 },
		{ 79, 436, 37 },
		{ 79, 437, 37 },
		{ 79, 438, 37 },
		{ 79, 439, 37 },
		{ 79, 440, 37 },
		{ 79, 441, 37 },
		{ 79, 442, 37 },
		{ 79, 443, 37 },
		{ 79, 444, 37 },
		{ 79, 445, 37 },
		{ 79, 446, 37 },
		{ 79, 447, 37 },
		{ 79, 448, 37 },
		{ 79, 449, 37 },
		{ 79, 450, 37 },
		{ 79, 451, 37 },
		{ 79, 452, 37 },
		{ 79, 453, 37 },
		{ 79, 454, 37 },
		{ 79, 455, 37 },
		{ 79, 456, 37 },
		{ 79, 457, 37 },
		{ 79, 458, 37 },
		{ 79, 459, 37 },
		{ 79, 460, 37 },
		{ 79, 461, 37 },
		{ 79, 462, 37 },
		{ 79, 463, 37 },
		{ 79, 464, 37 },
		{ 79, 465, 37 },
		{ 79, 466, 37 },
		{ 79, 467, 37 },
		{ 79, 468, 37 },
		{ 79, 469, 37 },
		{ 79, 470, 37 },
		{ 79, 471, 37 },
		{ 79, 472, 37 },
		{ 79, 473, 37 },
		{ 79, 474, 37 },
		{ 79, 475, 37 },
		{ 79, 476, 37 },
		{ 79, 477, 37 },
		{ 79, 478, 37 },
		{ 79, 479, 37 },
		{ 79, 480, 37 },
		{ 79, 481, 37 },
		{ 79, 482, 37 },
		{ 79, 483, 37 },
		{ 79, 484, 37 },
		{ 79, 485, 37 },
		{ 79, 486, 37 },
		{ 79, 487, 37 },
		{ 79, 488, 37 },
		{ 79, 489, 37 },
		{ 79, 490, 37 },
		{ 79, 491, 37 },
		{ 79, 492, 37 },
		{ 79, 493, 37 },
		{ 79, 494, 37 },
		{ 80, 435, 37 },
		{ 80, 436, 37 },
		{ 80, 437, 37 },
		{ 80, 438, 37 },
		{ 80, 439, 37 },
		{ 80, 440, 37 },
		{ 80, 441, 37 },
		{ 80, 442, 37 },
		{ 80, 443, 37 },
		{ 80, 444, 37 },
		{ 80, 445, 37 },
		{ 80, 446, 37 },
		{ 80, 447, 37 },
		{ 80, 448, 37 },
		{ 80, 449, 37 },
		{ 80, 450, 37 },
		{ 80, 451, 37 },
		{ 80, 452, 37 },
		{ 80, 453, 37 },
		{ 80, 454, 37 },
		{ 80, 455, 37 },
		{ 80, 456, 37 },
		{ 80, 457, 37 },
		{ 80, 458, 37 },
		{ 80, 459, 37 },
		{ 80, 460, 37 },
		{ 80, 461, 37 },
		{ 80, 462, 37 },
		{ 80, 463, 37 },
		{ 80, 464, 37 },
		{ 80, 465, 37 },
		{ 80, 466, 37 },
		{ 80, 467, 37 },
		{ 80, 468, 37 },
		{ 80, 469, 37 },
		{ 80, 470, 37 },
		{ 80, 471, 37 },
		{ 80, 472, 37 },
		{ 80, 473, 37 },
		{ 80, 474, 37 },
		{ 80, 475, 37 },
		{ 80, 476, 37 },
		{ 80, 477, 37 },
		{ 80, 478, 37 },
		{ 80, 479, 37 },
		{ 80, 480, 37 },
		{ 80, 481, 37 },
		{ 80, 482, 37 },
		{ 80, 483, 37 },
		{ 80, 484, 37 },
		{ 80, 485, 37 },
		{ 80, 486, 37 },
		{ 80, 487, 37 },
		{ 80, 488, 37 },
		{ 80, 489, 37 },
		{ 80, 490, 37 },
		{ 80, 491, 37 },
		{ 80, 492, 37 },
		{ 81, 434, 37 },
		{ 81, 435, 37 },
		{ 81, 436, 37 },
		{ 81, 437, 37 },
		{ 81, 438, 37 },
		{ 81, 439, 37 },
		{ 81, 440, 37 },
		{ 81, 441, 37 },
		{ 81, 442, 37 },
		{ 81, 443, 37 },
		{ 81, 444, 37 },
		{ 81, 445, 37 },
		{ 81, 446, 37 },
		{ 81, 447, 37 },
		{ 81, 448, 37 },
		{ 81, 449, 37 },
		{ 81, 450, 37 },
		{ 81, 451, 37 },
		{ 81, 452, 37 },
		{ 81, 453, 37 },
		{ 81, 454, 37 },
		{ 81, 455, 37 },
		{ 81, 456, 37 },
		{ 81, 457, 37 },
		{ 81, 458, 37 },
		{ 81, 459, 37 },
		{ 81, 460, 37 },
		{ 81, 461, 37 },
		{ 81, 462, 37 },
		{ 81, 463, 37 },
		{ 81, 464, 37 },
		{ 81, 465, 37 },
		{ 81, 466, 37 },
		{ 81, 467, 37 },
		{ 81, 468, 37 },
		{ 81, 469, 37 },
		{ 81, 470, 37 },
		{ 81, 471, 37 },
		{ 81, 472, 37 },
		{ 81, 473, 37 },
		{ 81, 474, 37 },
		{ 81, 475, 37 },
		{ 81, 476, 37 },
		{ 81, 477, 37 },
		{ 81, 478, 37 },
		{ 81, 479, 37 },
		{ 81, 480, 37 },
		{ 81, 481, 37 },
		{ 81, 482, 37 },
		{ 81, 483, 37 },
		{ 81, 484, 37 },
		{ 81, 485, 37 },
		{ 81, 486, 37 },
		{ 81, 487, 37 },
		{ 81, 488, 37 },
		{ 81, 489, 37 },
		{ 81, 490, 37 },
		{ 81, 491, 37 },
		{ 82, 433, 37 },
		{ 82, 434, 37 },
		{ 82, 435, 37 },
		{ 82, 436, 37 },
		{ 82, 437, 37 },
		{ 82, 438, 37 },
		{ 82, 439, 37 },
		{ 82, 440, 37 },
		{ 82, 441, 37 },
		{ 82, 442, 37 },
		{ 82, 443, 37 },
		{ 82, 444, 37 },
		{ 82, 445, 37 },
		{ 82, 446, 37 },
		{ 82, 447, 37 },
		{ 82, 448, 37 },
		{ 82, 449, 37 },
		{ 82, 450, 37 },
		{ 82, 451, 37 },
		{ 82, 452, 37 },
		{ 82, 453, 37 },
		{ 82, 454, 37 },
		{ 82, 455, 37 },
		{ 82, 456, 37 },
		{ 82, 457, 37 },
		{ 82, 458, 37 },
		{ 82, 459, 37 },
		{ 82, 460, 37 },
		{ 82, 461, 37 },
		{ 82, 462, 37 },
		{ 82, 463, 37 },
		{ 82, 464, 37 },
		{ 82, 465, 37 },
		{ 82, 466, 37 },
		{ 82, 467, 37 },
		{ 82, 468, 37 },
		{ 82, 469, 37 },
		{ 82, 470, 37 },
		{ 82, 471, 37 },
		{ 82, 472, 37 },
		{ 82, 473, 37 },
		{ 82, 474, 37 },
		{ 82, 475, 37 },
		{ 82, 476, 37 },
		{ 82, 477, 37 },
		{ 82, 478, 37 },
		{ 82, 479, 37 },
		{ 82, 480, 37 },
		{ 82, 481, 37 },
		{ 82, 482, 37 },
		{ 82, 483, 37 },
		{ 82, 484, 37 },
		{ 82, 485, 37 },
		{ 82, 486, 37 },
		{ 82, 487, 37 },
		{ 82, 488, 37 },
		{ 82, 489, 37 },
		{ 82, 490, 37 },
		{ 83, 432, 37 },
		{ 83, 433, 37 },
		{ 83, 434, 37 },
		{ 83, 435, 37 },
		{ 83, 436, 37 },
		{ 83, 437, 37 },
		{ 83, 438, 37 },
		{ 83, 439, 37 },
		{ 83, 440, 37 },
		{ 83, 441, 37 },
		{ 83, 442, 37 },
		{ 83, 443, 37 },
		{ 83, 444, 37 },
		{ 83, 445, 37 },
		{ 83, 446, 37 },
		{ 83, 447, 37 },
		{ 83, 448, 37 },
		{ 83, 449, 37 },
		{ 83, 450, 37 },
		{ 83, 451, 37 },
		{ 83, 452, 37 },
		{ 83, 453, 37 },
		{ 83, 454, 37 },
		{ 83, 455, 37 },
		{ 83, 456, 37 },
		{ 83, 457, 37 },
		{ 83, 458, 37 },
		{ 83, 459, 37 },
		{ 83, 460, 37 },
		{ 83, 461, 37 },
		{ 83, 462, 37 },
		{ 83, 463, 37 },
		{ 83, 464, 37 },
		{ 83, 465, 37 },
		{ 83, 466, 37 },
		{ 83, 467, 37 },
		{ 83, 468, 37 },
		{ 83, 469, 37 },
		{ 83, 470, 37 },
		{ 83, 471, 37 },
		{ 83, 472, 37 },
		{ 83, 473, 37 },
		{ 83, 474, 37 },
		{ 83, 475, 37 },
		{ 83, 476, 37 },
		{ 83, 477, 37 },
		{ 83, 478, 37 },
		{ 83, 479, 37 },
		{ 83, 480, 37 },
		{ 83, 481, 37 },
		{ 83, 482, 37 },
		{ 83, 483, 37 },
		{ 83, 484, 37 },
		{ 83, 485, 37 },
		{ 83, 486, 37 },
		{ 83, 487, 37 },
		{ 83, 488, 37 },
		{ 84, 431, 37 },
		{ 84, 432, 37 },
		{ 84, 433, 37 },
		{ 84, 434, 37 },
		{ 84, 435, 37 },
		{ 84, 436, 37 },
		{ 84, 437, 37 },
		{ 84, 438, 37 },
		{ 84, 439, 37 },
		{ 84, 440, 37 },
		{ 84, 441, 37 },
		{ 84, 442, 37 },
		{ 84, 443, 37 },
		{ 84, 444, 37 },
		{ 84, 445, 37 },
		{ 84, 446, 37 },
		{ 84, 447, 37 },
		{ 84, 448, 37 },
		{ 84, 449, 37 },
		{ 84, 450, 37 },
		{ 84, 451, 37 },
		{ 84, 452, 37 },
		{ 84, 453, 37 },
		{ 84, 454, 37 },
		{ 84, 455, 37 },
		{ 84, 456, 37 },
		{ 84, 457, 37 },
		{ 84, 458, 37 },
		{ 84, 459, 37 },
		{ 84, 460, 37 },
		{ 84, 461, 37 },
		{ 84, 462, 37 },
		{ 84, 463, 37 },
		{ 84, 464, 37 },
		{ 84, 465, 37 },
		{ 84, 466, 37 },
		{ 84, 467, 37 },
		{ 84, 468, 37 },
		{ 84, 469, 37 },
		{ 84, 470, 37 },
		{ 84, 471, 37 },
		{ 84, 472, 37 },
		{ 84, 473, 37 },
		{ 84, 474, 37 },
		{ 84, 475, 37 },
		{ 84, 476, 37 },
		{ 84, 477, 37 },
		{ 84, 478, 37 },
		{ 84, 479, 37 },
		{ 84, 480, 37 },
		{ 84, 481, 37 },
		{ 84, 482, 37 },
		{ 84, 483, 37 },
		{ 84, 484, 37 },
		{ 84, 485, 37 },
		{ 84, 486, 37 },
		{ 84, 487, 37 },
		{ 85, 430, 37 },
		{ 85, 431, 37 },
		{ 85, 432, 37 },
		{ 85, 433, 37 },
		{ 85, 434, 37 },
		{ 85, 435, 37 },
		{ 85, 436, 37 },
		{ 85, 437, 37 },
		{ 85, 438, 37 },
		{ 85, 439, 37 },
		{ 85, 440, 37 },
		{ 85, 441, 37 },
		{ 85, 442, 37 },
		{ 85, 443, 37 },
		{ 85, 444, 37 },
		{ 85, 445, 37 },
		{ 85, 446, 37 },
		{ 85, 447, 37 },
		{ 85, 448, 37 },
		{ 85, 449, 37 },
		{ 85, 450, 37 },
		{ 85, 451, 37 },
		{ 85, 452, 37 },
		{ 85, 453, 37 },
		{ 85, 454, 37 },
		{ 85, 455, 37 },
		{ 85, 456, 37 },
		{ 85, 457, 37 },
		{ 85, 458, 37 },
		{ 85, 459, 37 },
		{ 85, 460, 37 },
		{ 85, 461, 37 },
		{ 85, 462, 37 },
		{ 85, 463, 37 },
		{ 85, 464, 37 },
		{ 85, 465, 37 },
		{ 85, 466, 37 },
		{ 85, 467, 37 },
		{ 85, 468, 37 },
		{ 85, 469, 37 },
		{ 85, 470, 37 },
		{ 85, 471, 37 },
		{ 85, 472, 37 },
		{ 85, 473, 37 },
		{ 85, 474, 37 },
		{ 85, 475, 37 },
		{ 85, 476, 37 },
		{ 85, 477, 37 },
		{ 85, 478, 37 },
		{ 85, 479, 37 },
		{ 85, 480, 37 },
		{ 85, 481, 37 },
		{ 85, 482, 37 },
		{ 85, 483, 37 },
		{ 85, 484, 37 },
		{ 85, 485, 37 },
		{ 85, 486, 37 },
		{ 86, 429, 37 },
		{ 86, 430, 37 },
		{ 86, 431, 37 },
		{ 86, 432, 37 },
		{ 86, 433, 37 },
		{ 86, 434, 37 },
		{ 86, 435, 37 },
		{ 86, 436, 37 },
		{ 86, 437, 37 },
		{ 86, 438, 37 },
		{ 86, 439, 37 },
		{ 86, 440, 37 },
		{ 86, 441, 37 },
		{ 86, 442, 37 },
		{ 86, 443, 37 },
		{ 86, 444, 37 },
		{ 86, 445, 37 },
		{ 86, 446, 37 },
		{ 86, 447, 37 },
		{ 86, 448, 37 },
		{ 86, 449, 37 },
		{ 86, 450, 37 },
		{ 86, 451, 37 },
		{ 86, 452, 37 },
		{ 86, 453, 37 },
		{ 86, 454, 37 },
		{ 86, 455, 37 },
		{ 86, 456, 37 },
		{ 86, 457, 37 },
		{ 86, 458, 37 },
		{ 86, 459, 37 },
		{ 86, 460, 37 },
		{ 86, 461, 37 },
		{ 86, 462, 37 },
		{ 86, 463, 37 },
		{ 86, 464, 37 },
		{ 86, 465, 37 },
		{ 86, 466, 37 },
		{ 86, 467, 37 },
		{ 86, 468, 37 },
		{ 86, 469, 37 },
		{ 86, 470, 37 },
		{ 86, 471, 37 },
		{ 86, 472, 37 },
		{ 86, 473, 37 },
		{ 86, 474, 37 },
		{ 86, 475, 37 },
		{ 86, 476, 37 },
		{ 86, 477, 37 },
		{ 86, 478, 37 },
		{ 86, 479, 37 },
		{ 86, 480, 37 },
		{ 86, 481, 37 },
		{ 86, 482, 37 },
		{ 86, 483, 37 },
		{ 86, 484, 37 },
		{ 87, 428, 37 },
		{ 87, 429, 37 },
		{ 87, 430, 37 },
		{ 87, 431, 37 },
		{ 87, 432, 37 },
		{ 87, 433, 37 },
		{ 87, 434, 37 },
		{ 87, 435, 37 },
		{ 87, 436, 37 },
		{ 87, 437, 37 },
		{ 87, 438, 37 },
		{ 87, 439, 37 },
		{ 87, 440, 37 },
		{ 87, 441, 37 },
		{ 87, 442, 37 },
		{ 87, 443, 37 },
		{ 87, 444, 37 },
		{ 87, 445, 37 },
		{ 87, 446, 37 },
		{ 87, 447, 37 },
		{ 87, 448, 37 },
		{ 87, 449, 37 },
		{ 87, 450, 37 },
		{ 87, 451, 37 },
		{ 87, 452, 37 },
		{ 87, 453, 37 },
		{ 87, 454, 37 },
		{ 87, 455, 37 },
		{ 87, 456, 37 },
		{ 87, 457, 37 },
		{ 87, 458, 37 },
		{ 87, 459, 37 },
		{ 87, 460, 37 },
		{ 87, 461, 37 },
		{ 87, 462, 37 },
		{ 87, 463, 37 },
		{ 87, 464, 37 },
		{ 87, 465, 37 },
		{ 87, 466, 37 },
		{ 87, 467, 37 },
		{ 87, 468, 37 },
		{ 87, 469, 37 },
		{ 87, 470, 37 },
		{ 87, 471, 37 },
		{ 87, 472, 37 },
		{ 87, 473, 37 },
		{ 87, 474, 37 },
		{ 87, 475, 37 },
		{ 87, 476, 37 },
		{ 87, 477, 37 },
		{ 87, 478, 37 },
		{ 87, 479, 37 },
		{ 87, 480, 37 },
		{ 87, 481, 37 },
		{ 87, 482, 37 },
		{ 87, 483, 37 },
		{ 88, 427, 37 },
		{ 88, 428, 37 },
		{ 88, 429, 37 },
		{ 88, 430, 37 },
		{ 88, 431, 37 },
		{ 88, 432, 37 },
		{ 88, 433, 37 },
		{ 88, 434, 37 },
		{ 88, 435, 37 },
		{ 88, 436, 37 },
		{ 88, 437, 37 },
		{ 88, 438, 37 },
		{ 88, 439, 37 },
		{ 88, 440, 37 },
		{ 88, 441, 37 },
		{ 88, 442, 37 },
		{ 88, 443, 37 },
		{ 88, 444, 37 },
		{ 88, 445, 37 },
		{ 88, 446, 37 },
		{ 88, 447, 37 },
		{ 88, 448, 37 },
		{ 88, 449, 37 },
		{ 88, 450, 37 },
		{ 88, 451, 37 },
		{ 88, 452, 37 },
		{ 88, 453, 37 },
		{ 88, 454, 37 },
		{ 88, 455, 37 },
		{ 88, 456, 37 },
		{ 88, 457, 37 },
		{ 88, 458, 37 },
		{ 88, 459, 37 },
		{ 88, 460, 37 },
		{ 88, 461, 37 },
		{ 88, 462, 37 },
		{ 88, 463, 37 },
		{ 88, 464, 37 },
		{ 88, 465, 37 },
		{ 88, 466, 37 },
		{ 88, 467, 37 },
		{ 88, 468, 37 },
		{ 88, 469, 37 },
		{ 88, 470, 37 },
		{ 88, 471, 37 },
		{ 88, 472, 37 },
		{ 88, 473, 37 },
		{ 88, 474, 37 },
		{ 88, 475, 37 },
		{ 88, 476, 37 },
		{ 88, 477, 37 },
		{ 88, 478, 37 },
		{ 88, 479, 37 },
		{ 88, 480, 37 },
		{ 88, 481, 37 },
		{ 88, 482, 37 },
		{ 89, 426, 37 },
		{ 89, 427, 37 },
		{ 89, 428, 37 },
		{ 89, 429, 37 },
		{ 89, 430, 37 },
		{ 89, 431, 37 },
		{ 89, 432, 37 },
		{ 89, 433, 37 },
		{ 89, 434, 37 },
		{ 89, 435, 37 },
		{ 89, 436, 37 },
		{ 89, 437, 37 },
		{ 89, 438, 37 },
		{ 89, 439, 37 },
		{ 89, 440, 37 },
		{ 89, 441, 37 },
		{ 89, 442, 37 },
		{ 89, 443, 37 },
		{ 89, 444, 37 },
		{ 89, 445, 37 },
		{ 89, 446, 37 },
		{ 89, 447, 37 },
		{ 89, 448, 37 },
		{ 89, 449, 37 },
		{ 89, 450, 37 },
		{ 89, 451, 37 },
		{ 89, 452, 37 },
		{ 89, 453, 37 },
		{ 89, 454, 37 },
		{ 89, 455, 37 },
		{ 89, 456, 37 },
		{ 89, 457, 37 },
		{ 89, 458, 37 },
		{ 89, 459, 37 },
		{ 89, 460, 37 },
		{ 89, 461, 37 },
		{ 89, 462, 37 },
		{ 89, 463, 37 },
		{ 89, 464, 37 },
		{ 89, 465, 37 },
		{ 89, 466, 37 },
		{ 89, 467, 37 },
		{ 89, 468, 37 },
		{ 89, 469, 37 },
		{ 89, 470, 37 },
		{ 89, 471, 37 },
		{ 89, 472, 37 },
		{ 89, 473, 37 },
		{ 89, 474, 37 },
		{ 89, 475, 37 },
		{ 89, 476, 37 },
		{ 89, 477, 37 },
		{ 89, 478, 37 },
		{ 89, 479, 37 },
		{ 89, 480, 37 },
		{ 90, 425, 37 },
		{ 90, 426, 37 },
		{ 90, 427, 37 },
		{ 90, 428, 37 },
		{ 90, 429, 37 },
		{ 90, 430, 37 },
		{ 90, 431, 37 },
		{ 90, 432, 37 },
		{ 90, 433, 37 },
		{ 90, 434, 37 },
		{ 90, 435, 37 },
		{ 90, 436, 37 },
		{ 90, 437, 37 },
		{ 90, 438, 37 },
		{ 90, 439, 37 },
		{ 90, 440, 37 },
		{ 90, 441, 37 },
		{ 90, 442, 37 },
		{ 90, 443, 37 },
		{ 90, 444, 37 },
		{ 90, 445, 37 },
		{ 90, 446, 37 },
		{ 90, 447, 37 },
		{ 90, 448, 37 },
		{ 90, 449, 37 },
		{ 90, 450, 37 },
		{ 90, 451, 37 },
		{ 90, 452, 37 },
		{ 90, 453, 37 },
		{ 90, 454, 37 },
		{ 90, 455, 37 },
		{ 90, 456, 37 },
		{ 90, 457, 37 },
		{ 90, 458, 37 },
		{ 90, 459, 37 },
		{ 90, 460, 37 },
		{ 90, 461, 37 },
		{ 90, 462, 37 },
		{ 90, 463, 37 },
		{ 90, 464, 37 },
		{ 90, 465, 37 },
		{ 90, 466, 37 },
		{ 90, 467, 37 },
		{ 90, 468, 37 },
		{ 90, 469, 37 },
		{ 90, 470, 37 },
		{ 90, 471, 37 },
		{ 90, 472, 37 },
		{ 90, 473, 37 },
		{ 90, 474, 37 },
		{ 90, 475, 37 },
		{ 90, 476, 37 },
		{ 90, 477, 37 },
		{ 90, 478, 37 },
		{ 90, 479, 37 },
		{ 91, 424, 37 },
		{ 91, 425, 37 },
		{ 91, 426, 37 },
		{ 91, 427, 37 },
		{ 91, 428, 37 },
		{ 91, 429, 37 },
		{ 91, 430, 37 },
		{ 91, 431, 37 },
		{ 91, 432, 37 },
		{ 91, 433, 37 },
		{ 91, 434, 37 },
		{ 91, 435, 37 },
		{ 91, 436, 37 },
		{ 91, 437, 37 },
		{ 91, 438, 37 },
		{ 91, 439, 37 },
		{ 91, 440, 37 },
		{ 91, 441, 37 },
		{ 91, 442, 37 },
		{ 91, 443, 37 },
		{ 91, 444, 37 },
		{ 91, 445, 37 },
		{ 91, 446, 37 },
		{ 91, 447, 37 },
		{ 91, 448, 37 },
		{ 91, 449, 37 },
		{ 91, 450, 37 },
		{ 91, 451, 37 },
		{ 91, 452, 37 },
		{ 91, 453, 37 },
		{ 91, 454, 37 },
		{ 91, 455, 37 },
		{ 91, 456, 37 },
		{ 91, 457, 37 },
		{ 91, 458, 37 },
		{ 91, 459, 37 },
		{ 91, 460, 37 },
		{ 91, 461, 37 },
		{ 91, 462, 37 },
		{ 91, 463, 37 },
		{ 91, 464, 37 },
		{ 91, 465, 37 },
		{ 91, 466, 37 },
		{ 91, 467, 37 },
		{ 91, 468, 37 },
		{ 91, 469, 37 },
		{ 91, 470, 37 },
		{ 91, 471, 37 },
		{ 91, 472, 37 },
		{ 91, 473, 37 },
		{ 91, 474, 37 },
		{ 91, 475, 37 },
		{ 91, 476, 37 },
		{ 91, 477, 37 },
		{ 91, 478, 37 },
		{ 92, 423, 37 },
		{ 92, 424, 37 },
		{ 92, 425, 37 },
		{ 92, 426, 37 },
		{ 92, 427, 37 },
		{ 92, 428, 37 },
		{ 92, 429, 37 },
		{ 92, 430, 37 },
		{ 92, 431, 37 },
		{ 92, 432, 37 },
		{ 92, 433, 37 },
		{ 92, 434, 37 },
		{ 92, 435, 37 },
		{ 92, 436, 37 },
		{ 92, 437, 37 },
		{ 92, 438, 37 },
		{ 92, 439, 37 },
		{ 92, 440, 37 },
		{ 92, 441, 37 },
		{ 92, 442, 37 },
		{ 92, 443, 37 },
		{ 92, 444, 37 },
		{ 92, 445, 37 },
		{ 92, 446, 37 },
		{ 92, 447, 37 },
		{ 92, 448, 37 },
		{ 92, 449, 37 },
		{ 92, 450, 37 },
		{ 92, 451, 37 },
		{ 92, 452, 37 },
		{ 92, 453, 37 },
		{ 92, 454, 37 },
		{ 92, 455, 37 },
		{ 92, 456, 37 },
		{ 92, 457, 37 },
		{ 92, 458, 37 },
		{ 92, 459, 37 },
		{ 92, 460, 37 },
		{ 92, 461, 37 },
		{ 92, 462, 37 },
		{ 92, 463, 37 },
		{ 92, 464, 37 },
		{ 92, 465, 37 },
		{ 92, 466, 37 },
		{ 92, 467, 37 },
		{ 92, 468, 37 },
		{ 92, 469, 37 },
		{ 92, 470, 37 },
		{ 92, 471, 37 },
		{ 92, 472, 37 },
		{ 92, 473, 37 },
		{ 92, 474, 37 },
		{ 92, 475, 37 },
		{ 92, 476, 37 },
		{ 93, 422, 37 },
		{ 93, 423, 37 },
		{ 93, 424, 37 },
		{ 93, 425, 37 },
		{ 93, 426, 37 },
		{ 93, 427, 37 },
		{ 93, 428, 37 },
		{ 93, 429, 37 },
		{ 93, 430, 37 },
		{ 93, 431, 37 },
		{ 93, 432, 37 },
		{ 93, 433, 37 },
		{ 93, 434, 37 },
		{ 93, 435, 37 },
		{ 93, 436, 37 },
		{ 93, 437, 37 },
		{ 93, 438, 37 },
		{ 93, 439, 37 },
		{ 93, 440, 37 },
		{ 93, 441, 37 },
		{ 93, 442, 37 },
		{ 93, 443, 37 },
		{ 93, 444, 37 },
		{ 93, 445, 37 },
		{ 93, 446, 37 },
		{ 93, 447, 37 },
		{ 93, 448, 37 },
		{ 93, 449, 37 },
		{ 93, 450, 37 },
		{ 93, 451, 37 },
		{ 93, 452, 37 },
		{ 93, 453, 37 },
		{ 93, 454, 37 },
		{ 93, 455, 37 },
		{ 93, 456, 37 },
		{ 93, 457, 37 },
		{ 93, 458, 37 },
		{ 93, 459, 37 },
		{ 93, 460, 37 },
		{ 93, 461, 37 },
		{ 93, 462, 37 },
		{ 93, 463, 37 },
		{ 93, 464, 37 },
		{ 93, 465, 37 },
		{ 93, 466, 37 },
		{ 93, 467, 37 },
		{ 93, 468, 37 },
		{ 93, 469, 37 },
		{ 93, 470, 37 },
		{ 93, 471, 37 },
		{ 93, 472, 37 },
		{ 93, 473, 37 },
		{ 93, 474, 37 },
		{ 93, 475, 37 },
		{ 94, 421, 37 },
		{ 94, 422, 37 },
		{ 94, 423, 37 },
		{ 94, 424, 37 },
		{ 94, 425, 37 },
		{ 94, 426, 37 },
		{ 94, 427, 37 },
		{ 94, 428, 37 },
		{ 94, 429, 37 },
		{ 94, 430, 37 },
		{ 94, 431, 37 },
		{ 94, 432, 37 },
		{ 94, 433, 37 },
		{ 94, 434, 37 },
		{ 94, 435, 37 },
		{ 94, 436, 37 },
		{ 94, 437, 37 },
		{ 94, 438, 37 },
		{ 94, 439, 37 },
		{ 94, 440, 37 },
		{ 94, 441, 37 },
		{ 94, 442, 37 },
		{ 94, 443, 37 },
		{ 94, 444, 37 },
		{ 94, 445, 37 },
		{ 94, 446, 37 },
		{ 94, 447, 37 },
		{ 94, 448, 37 },
		{ 94, 449, 37 },
		{ 94, 450, 37 },
		{ 94, 451, 37 },
		{ 94, 452, 37 },
		{ 94, 453, 37 },
		{ 94, 454, 37 },
		{ 94, 455, 37 },
		{ 94, 456, 37 },
		{ 94, 457, 37 },
		{ 94, 458, 37 },
		{ 94, 459, 37 },
		{ 94, 460, 37 },
		{ 94, 461, 37 },
		{ 94, 462, 37 },
		{ 94, 463, 37 },
		{ 94, 464, 37 },
		{ 94, 465, 37 },
		{ 94, 466, 37 },
		{ 94, 467, 37 },
		{ 94, 468, 37 },
		{ 94, 469, 37 },
		{ 94, 470, 37 },
		{ 94, 471, 37 },
		{ 94, 472, 37 },
		{ 94, 473, 37 },
		{ 94, 474, 37 },
		{ 95, 420, 37 },
		{ 95, 421, 37 },
		{ 95, 422, 37 },
		{ 95, 423, 37 },
		{ 95, 424, 37 },
		{ 95, 425, 37 },
		{ 95, 426, 37 },
		{ 95, 427, 37 },
		{ 95, 428, 37 },
		{ 95, 429, 37 },
		{ 95, 430, 37 },
		{ 95, 431, 37 },
		{ 95, 432, 37 },
		{ 95, 433, 37 },
		{ 95, 434, 37 },
		{ 95, 435, 37 },
		{ 95, 436, 37 },
		{ 95, 437, 37 },
		{ 95, 438, 37 },
		{ 95, 439, 37 },
		{ 95, 440, 37 },
		{ 95, 441, 37 },
		{ 95, 442, 37 },
		{ 95, 443, 37 },
		{ 95, 444, 37 },
		{ 95, 445, 37 },
		{ 95, 446, 37 },
		{ 95, 447, 37 },
		{ 95, 448, 37 },
		{ 95, 449, 37 },
		{ 95, 450, 37 },
		{ 95, 451, 37 },
		{ 95, 452, 37 },
		{ 95, 453, 37 },
		{ 95, 454, 37 },
		{ 95, 455, 37 },
		{ 95, 456, 37 },
		{ 95, 457, 37 },
		{ 95, 458, 37 },
		{ 95, 459, 37 },
		{ 95, 460, 37 },
		{ 95, 461, 37 },
		{ 95, 462, 37 },
		{ 95, 463, 37 },
		{ 95, 464, 37 },
		{ 95, 465, 37 },
		{ 95, 466, 37 },
		{ 95, 467, 37 },
		{ 95, 468, 37 },
		{ 95, 469, 37 },
		{ 95, 470, 37 },
		{ 95, 471, 37 },
		{ 95, 472, 37 },
		{ 96, 419, 37 },
		{ 96, 420, 37 },
		{ 96, 421, 37 },
		{ 96, 422, 37 },
		{ 96, 423, 37 },
		{ 96, 424, 37 },
		{ 96, 425, 37 },
		{ 96, 426, 37 },
		{ 96, 427, 37 },
		{ 96, 428, 37 },
		{ 96, 429, 37 },
		{ 96, 430, 37 },
		{ 96, 431, 37 },
		{ 96, 432, 37 },
		{ 96, 433, 37 },
		{ 96, 434, 37 },
		{ 96, 435, 37 },
		{ 96, 436, 37 },
		{ 96, 437, 37 },
		{ 96, 438, 37 },
		{ 96, 439, 37 },
		{ 96, 440, 37 },
		{ 96, 441, 37 },
		{ 96, 442, 37 },
		{ 96, 443, 37 },
		{ 96, 444, 37 },
		{ 96, 445, 37 },
		{ 96, 446, 37 },
		{ 96, 447, 37 },
		{ 96, 448, 37 },
		{ 96, 449, 37 },
		{ 96, 450, 37 },
		{ 96, 451, 37 },
		{ 96, 452, 37 },
		{ 96, 453, 37 },
		{ 96, 454, 37 },
		{ 96, 455, 37 },
		{ 96, 456, 37 },
		{ 96, 457, 37 },
		{ 96, 458, 37 },
		{ 96, 459, 37 },
		{ 96, 460, 37 },
		{ 96, 461, 37 },
		{ 96, 462, 37 },
		{ 96, 463, 37 },
		{ 96, 464, 37 },
		{ 96, 465, 37 },
		{ 96, 466, 37 },
		{ 96, 467, 37 },
		{ 96, 468, 37 },
		{ 96, 469, 37 },
		{ 96, 470, 37 },
		{ 96, 471, 37 },
		{ 97, 418, 37 },
		{ 97, 419, 37 },
		{ 97, 420, 37 },
		{ 97, 421, 37 },
		{ 97, 422, 37 },
		{ 97, 423, 37 },
		{ 97, 424, 37 },
		{ 97, 425, 37 },
		{ 97, 426, 37 },
		{ 97, 427, 37 },
		{ 97, 428, 37 },
		{ 97, 429, 37 },
		{ 97, 430, 37 },
		{ 97, 431, 37 },
		{ 97, 432, 37 },
		{ 97, 433, 37 },
		{ 97, 434, 37 },
		{ 97, 435, 37 },
		{ 97, 436, 37 },
		{ 97, 437, 37 },
		{ 97, 438, 37 },
		{ 97, 439, 37 },
		{ 97, 440, 37 },
		{ 97, 441, 37 },
		{ 97, 442, 37 },
		{ 97, 443, 37 },
		{ 97, 444, 37 },
		{ 97, 445, 37 },
		{ 97, 446, 37 },
		{ 97, 447, 37 },
		{ 97, 448, 37 },
		{ 97, 449, 37 },
		{ 97, 450, 37 },
		{ 97, 451, 37 },
		{ 97, 452, 37 },
		{ 97, 453, 37 },
		{ 97, 454, 37 },
		{ 97, 455, 37 },
		{ 97, 456, 37 },
		{ 97, 457, 37 },
		{ 97, 458, 37 },
		{ 97, 459, 37 },
		{ 97, 460, 37 },
		{ 97, 461, 37 },
		{ 97, 462, 37 },
		{ 97, 463, 37 },
		{ 97, 464, 37 },
		{ 97, 465, 37 },
		{ 97, 466, 37 },
		{ 97, 467, 37 },
		{ 97, 468, 37 },
		{ 97, 469, 37 },
		{ 97, 470, 37 },
		{ 98, 417, 37 },
		{ 98, 418, 37 },
		{ 98, 419, 37 },
		{ 98, 420, 37 },
		{ 98, 421, 37 },
		{ 98, 422, 37 },
		{ 98, 423, 37 },
		{ 98, 424, 37 },
		{ 98, 425, 37 },
		{ 98, 426, 37 },
		{ 98, 427, 37 },
		{ 98, 428, 37 },
		{ 98, 429, 37 },
		{ 98, 430, 37 },
		{ 98, 431, 37 },
		{ 98, 432, 37 },
		{ 98, 433, 37 },
		{ 98, 434, 37 },
		{ 98, 435, 37 },
		{ 98, 436, 37 },
		{ 98, 437, 37 },
		{ 98, 438, 37 },
		{ 98, 439, 37 },
		{ 98, 440, 37 },
		{ 98, 441, 37 },
		{ 98, 442, 37 },
		{ 98, 443, 37 },
		{ 98, 444, 37 },
		{ 98, 445, 37 },
		{ 98, 446, 37 },
		{ 98, 447, 37 },
		{ 98, 448, 37 },
		{ 98, 449, 37 },
		{ 98, 450, 37 },
		{ 98, 451, 37 },
		{ 98, 452, 37 },
		{ 98, 453, 37 },
		{ 98, 454, 37 },
		{ 98, 455, 37 },
		{ 98, 456, 37 },
		{ 98, 457, 37 },
		{ 98, 458, 37 },
		{ 98, 459, 37 },
		{ 98, 460, 37 },
		{ 98, 461, 37 },
		{ 98, 462, 37 },
		{ 98, 463, 37 },
		{ 98, 464, 37 },
		{ 98, 465, 37 },
		{ 98, 466, 37 },
		{ 98, 467, 37 },
		{ 98, 468, 37 },
		{ 99, 416, 37 },
		{ 99, 417, 37 },
		{ 99, 418, 37 },
		{ 99, 419, 37 },
		{ 99, 420, 37 },
		{ 99, 421, 37 },
		{ 99, 422, 37 },
		{ 99, 423, 37 },
		{ 99, 424, 37 },
		{ 99, 425, 37 },
		{ 99, 426, 37 },
		{ 99, 427, 37 },
		{ 99, 428, 37 },
		{ 99, 429, 37 },
		{ 99, 430, 37 },
		{ 99, 431, 37 },
		{ 99, 432, 37 },
		{ 99, 433, 37 },
		{ 99, 434, 37 },
		{ 99, 435, 37 },
		{ 99, 436, 37 },
		{ 99, 437, 37 },
		{ 99, 438, 37 },
		{ 99, 439, 37 },
		{ 99, 440, 37 },
		{ 99, 441, 37 },
		{ 99, 442, 37 },
		{ 99, 443, 37 },
		{ 99, 444, 37 },
		{ 99, 445, 37 },
		{ 99, 446, 37 },
		{ 99, 447, 37 },
		{ 99, 448, 37 },
		{ 99, 449, 37 },
		{ 99, 450, 37 },
		{ 99, 451, 37 },
		{ 99, 452, 37 },
		{ 99, 453, 37 },
		{ 99, 454, 37 },
		{ 99, 455, 37 },
		{ 99, 456, 37 },
		{ 99, 457, 37 },
		{ 99, 458, 37 },
		{ 99, 459, 37 },
		{ 99, 460, 37 },
		{ 99, 461, 37 },
		{ 99, 462, 37 },
		{ 99, 463, 37 },
		{ 99, 464, 37 },
		{ 99, 465, 37 },
		{ 99, 466, 37 },
		{ 99, 467, 37 },
		{ 100, 415, 37 },
		{ 100, 416, 37 },
		{ 100, 417, 37 },
		{ 100, 418, 37 },
		{ 100, 419, 37 },
		{ 100, 420, 37 },
		{ 100, 421, 37 },
		{ 100, 422, 37 },
		{ 100, 423, 37 },
		{ 100, 424, 37 },
		{ 100, 425, 37 },
		{ 100, 426, 37 },
		{ 100, 427, 37 },
		{ 100, 428, 37 },
		{ 100, 429, 37 },
		{ 100, 430, 37 },
		{ 100, 431, 37 },
		{ 100, 432, 37 },
		{ 100, 433, 37 },
		{ 100, 434, 37 },
		{ 100, 435, 37 },
		{ 100, 436, 37 },
		{ 100, 437, 37 },
		{ 100, 438, 37 },
		{ 100, 439, 37 },
		{ 100, 440, 37 },
		{ 100, 441, 37 },
		{ 100, 442, 37 },
		{ 100, 443, 37 },
		{ 100, 444, 37 },
		{ 100, 445, 37 },
		{ 100, 446, 37 },
		{ 100, 447, 37 },
		{ 100, 448, 37 },
		{ 100, 449, 37 },
		{ 100, 450, 37 },
		{ 100, 451, 37 },
		{ 100, 452, 37 },
		{ 100, 453, 37 },
		{ 100, 454, 37 },
		{ 100, 455, 37 },
		{ 100, 456, 37 },
		{ 100, 457, 37 },
		{ 100, 458, 37 },
		{ 100, 459, 37 },
		{ 100, 460, 37 },
		{ 100, 461, 37 },
		{ 100, 462, 37 },
		{ 100, 463, 37 },
		{ 100, 464, 37 },
		{ 100, 465, 37 },
		{ 100, 466, 37 },
		{ 101, 414, 37 },
		{ 101, 415, 37 },
		{ 101, 416, 37 },
		{ 101, 417, 37 },
		{ 101, 418, 37 },
		{ 101, 419, 37 },
		{ 101, 420, 37 },
		{ 101, 421, 37 },
		{ 101, 422, 37 },
		{ 101, 423, 37 },
		{ 101, 424, 37 },
		{ 101, 425, 37 },
		{ 101, 426, 37 },
		{ 101, 427, 37 },
		{ 101, 428, 37 },
		{ 101, 429, 37 },
		{ 101, 430, 37 },
		{ 101, 431, 37 },
		{ 101, 432, 37 },
		{ 101, 433, 37 },
		{ 101, 434, 37 },
		{ 101, 435, 37 },
		{ 101, 436, 37 },
		{ 101, 437, 37 },
		{ 101, 438, 37 },
		{ 101, 439, 37 },
		{ 101, 440, 37 },
		{ 101, 441, 37 },
		{ 101, 442, 37 },
		{ 101, 443, 37 },
		{ 101, 444, 37 },
		{ 101, 445, 37 },
		{ 101, 446, 37 },
		{ 101, 447, 37 },
		{ 101, 448, 37 },
		{ 101, 449, 37 },
		{ 101, 450, 37 },
		{ 101, 451, 37 },
		{ 101, 452, 37 },
		{ 101, 453, 37 },
		{ 101, 454, 37 },
		{ 101, 455, 37 },
		{ 101, 456, 37 },
		{ 101, 457, 37 },
		{ 101, 458, 37 },
		{ 101, 459, 37 },
		{ 101, 460, 37 },
		{ 101, 461, 37 },
		{ 101, 462, 37 },
		{ 101, 463, 37 },
		{ 101, 464, 37 },
		{ 102, 413, 37 },
		{ 102, 414, 37 },
		{ 102, 415, 37 },
		{ 102, 416, 37 },
		{ 102, 417, 37 },
		{ 102, 418, 37 },
		{ 102, 419, 37 },
		{ 102, 420, 37 },
		{ 102, 421, 37 },
		{ 102, 422, 37 },
		{ 102, 423, 37 },
		{ 102, 424, 37 },
		{ 102, 425, 37 },
		{ 102, 426, 37 },
		{ 102, 427, 37 },
		{ 102, 428, 37 },
		{ 102, 429, 37 },
		{ 102, 430, 37 },
		{ 102, 431, 37 },
		{ 102, 432, 37 },
		{ 102, 433, 37 },
		{ 102, 434, 37 },
		{ 102, 435, 37 },
		{ 102, 436, 37 },
		{ 102, 437, 37 },
		{ 102, 438, 37 },
		{ 102, 439, 37 },
		{ 102, 440, 37 },
		{ 102, 441, 37 },
		{ 102, 442, 37 },
		{ 102, 443, 37 },
		{ 102, 444, 37 },
		{ 102, 445, 37 },
		{ 102, 446, 37 },
		{ 102, 447, 37 },
		{ 102, 448, 37 },
		{ 102, 449, 37 },
		{ 102, 450, 37 },
		{ 102, 451, 37 },
		{ 102, 452, 37 },
		{ 102, 453, 37 },
		{ 102, 454, 37 },
		{ 102, 455, 37 },
		{ 102, 456, 37 },
		{ 102, 457, 37 },
		{ 102, 458, 37 },
		{ 102, 459, 37 },
		{ 102, 460, 37 },
		{ 102, 461, 37 },
		{ 102, 462, 37 },
		{ 102, 463, 37 },
		{ 103, 412, 37 },
		{ 103, 413, 37 },
		{ 103, 414, 37 },
		{ 103, 415, 37 },
		{ 103, 416, 37 },
		{ 103, 417, 37 },
		{ 103, 418, 37 },
		{ 103, 419, 37 },
		{ 103, 420, 37 },
		{ 103, 421, 37 },
		{ 103, 422, 37 },
		{ 103, 423, 37 },
		{ 103, 424, 37 },
		{ 103, 425, 37 },
		{ 103, 426, 37 },
		{ 103, 427, 37 },
		{ 103, 428, 37 },
		{ 103, 429, 37 },
		{ 103, 430, 37 },
		{ 103, 431, 37 },
		{ 103, 432, 37 },
		{ 103, 433, 37 },
		{ 103, 434, 37 },
		{ 103, 435, 37 },
		{ 103, 436, 37 },
		{ 103, 437, 37 },
		{ 103, 438, 37 },
		{ 103, 439, 37 },
		{ 103, 440, 37 },
		{ 103, 441, 37 },
		{ 103, 442, 37 },
		{ 103, 443, 37 },
		{ 103, 444, 37 },
		{ 103, 445, 37 },
		{ 103, 446, 37 },
		{ 103, 447, 37 },
		{ 103, 448, 37 },
		{ 103, 449, 37 },
		{ 103, 450, 37 },
		{ 103, 451, 37 },
		{ 103, 452, 37 },
		{ 103, 453, 37 },
		{ 103, 454, 37 },
		{ 103, 455, 37 },
		{ 103, 456, 37 },
		{ 103, 457, 37 },
		{ 103, 458, 37 },
		{ 103, 459, 37 },
		{ 103, 460, 37 },
		{ 103, 461, 37 },
		{ 103, 462, 37 },
		{ 104, 411, 37 },
		{ 104, 412, 37 },
		{ 104, 413, 37 },
		{ 104, 414, 37 },
		{ 104, 415, 37 },
		{ 104, 416, 37 },
		{ 104, 417, 37 },
		{ 104, 418, 37 },
		{ 104, 419, 37 },
		{ 104, 420, 37 },
		{ 104, 421, 37 },
		{ 104, 422, 37 },
		{ 104, 423, 37 },
		{ 104, 424, 37 },
		{ 104, 425, 37 },
		{ 104, 426, 37 },
		{ 104, 427, 37 },
		{ 104, 428, 37 },
		{ 104, 429, 37 },
		{ 104, 430, 37 },
		{ 104, 431, 37 },
		{ 104, 432, 37 },
		{ 104, 433, 37 },
		{ 104, 434, 37 },
		{ 104, 435, 37 },
		{ 104, 436, 37 },
		{ 104, 437, 37 },
		{ 104, 438, 37 },
		{ 104, 439, 37 },
		{ 104, 440, 37 },
		{ 104, 441, 37 },
		{ 104, 442, 37 },
		{ 104, 443, 37 },
		{ 104, 444, 37 },
		{ 104, 445, 37 },
		{ 104, 446, 37 },
		{ 104, 447, 37 },
		{ 104, 448, 37 },
		{ 104, 449, 37 },
		{ 104, 450, 37 },
		{ 104, 451, 37 },
		{ 104, 452, 37 },
		{ 104, 453, 37 },
		{ 104, 454, 37 },
		{ 104, 455, 37 },
		{ 104, 456, 37 },
		{ 104, 457, 37 },
		{ 104, 458, 37 },
		{ 104, 459, 37 },
		{ 104, 460, 37 },
		{ 105, 410, 37 },
		{ 105, 411, 37 },
		{ 105, 412, 37 },
		{ 105, 413, 37 },
		{ 105, 414, 37 },
		{ 105, 415, 37 },
		{ 105, 416, 37 },
		{ 105, 417, 37 },
		{ 105, 418, 37 },
		{ 105, 419, 37 },
		{ 105, 420, 37 },
		{ 105, 421, 37 },
		{ 105, 422, 37 },
		{ 105, 423, 37 },
		{ 105, 424, 37 },
		{ 105, 425, 37 },
		{ 105, 426, 37 },
		{ 105, 427, 37 },
		{ 105, 428, 37 },
		{ 105, 429, 37 },
		{ 105, 430, 37 },
		{ 105, 431, 37 },
		{ 105, 432, 37 },
		{ 105, 433, 37 },
		{ 105, 434, 37 },
		{ 105, 435, 37 },
		{ 105, 436, 37 },
		{ 105, 437, 37 },
		{ 105, 438, 37 },
		{ 105, 439, 37 },
		{ 105, 440, 37 },
		{ 105, 441, 37 },
		{ 105, 442, 37 },
		{ 105, 443, 37 },
		{ 105, 444, 37 },
		{ 105, 445, 37 },
		{ 105, 446, 37 },
		{ 105, 447, 37 },
		{ 105, 448, 37 },
		{ 105, 449, 37 },
		{ 105, 450, 37 },
		{ 105, 451, 37 },
		{ 105, 452, 37 },
		{ 105, 453, 37 },
		{ 105, 454, 37 },
		{ 105, 455, 37 },
		{ 105, 456, 37 },
		{ 105, 457, 37 },
		{ 105, 458, 37 },
		{ 105, 459, 37 },
		{ 106, 409, 37 },
		{ 106, 410, 37 },
		{ 106, 411, 37 },
		{ 106, 412, 37 },
		{ 106, 413, 37 },
		{ 106, 414, 37 },
		{ 106, 415, 37 },
		{ 106, 416, 37 },
		{ 106, 417, 37 },
		{ 106, 418, 37 },
		{ 106, 419, 37 },
		{ 106, 420, 37 },
		{ 106, 421, 37 },
		{ 106, 422, 37 },
		{ 106, 423, 37 },
		{ 106, 424, 37 },
		{ 106, 425, 37 },
		{ 106, 426, 37 },
		{ 106, 427, 37 },
		{ 106, 428, 37 },
		{ 106, 429, 37 },
		{ 106, 430, 37 },
		{ 106, 431, 37 },
		{ 106, 432, 37 },
		{ 106, 433, 37 },
		{ 106, 434, 37 },
		{ 106, 435, 37 },
		{ 106, 436, 37 },
		{ 106, 437, 37 },
		{ 106, 438, 37 },
		{ 106, 439, 37 },
		{ 106, 440, 37 },
		{ 106, 441, 37 },
		{ 106, 442, 37 },
		{ 106, 443, 37 },
		{ 106, 444, 37 },
		{ 106, 445, 37 },
		{ 106, 446, 37 },
		{ 106, 447, 37 },
		{ 106, 448, 37 },
		{ 106, 449, 37 },
		{ 106, 450, 37 },
		{ 106, 451, 37 },
		{ 106, 452, 37 },
		{ 106, 453, 37 },
		{ 106, 454, 37 },
		{ 106, 455, 37 },
		{ 106, 456, 37 },
		{ 106, 457, 37 },
		{ 106, 458, 37 },
		{ 107, 408, 37 },
		{ 107, 409, 37 },
		{ 107, 410, 37 },
		{ 107, 411, 37 },
		{ 107, 412, 37 },
		{ 107, 413, 37 },
		{ 107, 414, 37 },
		{ 107, 415, 37 },
		{ 107, 416, 37 },
		{ 107, 417, 37 },
		{ 107, 418, 37 },
		{ 107, 419, 37 },
		{ 107, 420, 37 },
		{ 107, 421, 37 },
		{ 107, 422, 37 },
		{ 107, 423, 37 },
		{ 107, 424, 37 },
		{ 107, 425, 37 },
		{ 107, 426, 37 },
		{ 107, 427, 37 },
		{ 107, 428, 37 },
		{ 107, 429, 37 },
		{ 107, 430, 37 },
		{ 107, 431, 37 },
		{ 107, 432, 37 },
		{ 107, 433, 37 },
		{ 107, 434, 37 },
		{ 107, 435, 37 },
		{ 107, 436, 37 },
		{ 107, 437, 37 },
		{ 107, 438, 37 },
		{ 107, 439, 37 },
		{ 107, 440, 37 },
		{ 107, 441, 37 },
		{ 107, 442, 37 },
		{ 107, 443, 37 },
		{ 107, 444, 37 },
		{ 107, 445, 37 },
		{ 107, 446, 37 },
		{ 107, 447, 37 },
		{ 107, 448, 37 },
		{ 107, 449, 37 },
		{ 107, 450, 37 },
		{ 107, 451, 37 },
		{ 107, 452, 37 },
		{ 107, 453, 37 },
		{ 107, 454, 37 },
		{ 107, 455, 37 },
		{ 107, 456, 37 },
		{ 108, 407, 37 },
		{ 108, 408, 37 },
		{ 108, 409, 37 },
		{ 108, 410, 37 },
		{ 108, 411, 37 },
		{ 108, 412, 37 },
		{ 108, 413, 37 },
		{ 108, 414, 37 },
		{ 108, 415, 37 },
		{ 108, 416, 37 },
		{ 108, 417, 37 },
		{ 108, 418, 37 },
		{ 108, 419, 37 },
		{ 108, 420, 37 },
		{ 108, 421, 37 },
		{ 108, 422, 37 },
		{ 108, 423, 37 },
		{ 108, 424, 37 },
		{ 108, 425, 37 },
		{ 108, 426, 37 },
		{ 108, 427, 37 },
		{ 108, 428, 37 },
		{ 108, 429, 37 },
		{ 108, 430, 37 },
		{ 108, 431, 37 },
		{ 108, 432, 37 },
		{ 108, 433, 37 },
		{ 108, 434, 37 },
		{ 108, 435, 37 },
		{ 108, 436, 37 },
		{ 108, 437, 37 },
		{ 108, 438, 37 },
		{ 108, 439, 37 },
		{ 108, 440, 37 },
		{ 108, 441, 37 },
		{ 108, 442, 37 },
		{ 108, 443, 37 },
		{ 108, 444, 37 },
		{ 108, 445, 37 },
		{ 108, 446, 37 },
		{ 108, 447, 37 },
		{ 108, 448, 37 },
		{ 108, 449, 37 },
		{ 108, 450, 37 },
		{ 108, 451, 37 },
		{ 108, 452, 37 },
		{ 108, 453, 37 },
		{ 108, 454, 37 },
		{ 108, 455, 37 },
		{ 109, 406, 37 },
		{ 109, 407, 37 },
		{ 109, 408, 37 },
		{ 109, 409, 37 },
		{ 109, 410, 37 },
		{ 109, 411, 37 },
		{ 109, 412, 37 },
		{ 109, 413, 37 },
		{ 109, 414, 37 },
		{ 109, 415, 37 },
		{ 109, 416, 37 },
		{ 109, 417, 37 },
		{ 109, 418, 37 },
		{ 109, 419, 37 },
		{ 109, 420, 37 },
		{ 109, 421, 37 },
		{ 109, 422, 37 },
		{ 109, 423, 37 },
		{ 109, 424, 37 },
		{ 109, 425, 37 },
		{ 109, 426, 37 },
		{ 109, 427, 37 },
		{ 109, 428, 37 },
		{ 109, 429, 37 },
		{ 109, 430, 37 },
		{ 109, 431, 37 },
		{ 109, 432, 37 },
		{ 109, 433, 37 },
		{ 109, 434, 37 },
		{ 109, 435, 37 },
		{ 109, 436, 37 },
		{ 109, 437, 37 },
		{ 109, 438, 37 },
		{ 109, 439, 37 },
		{ 109, 440, 37 },
		{ 109, 441, 37 },
		{ 109, 442, 37 },
		{ 109, 443, 37 },
		{ 109, 444, 37 },
		{ 109, 445, 37 },
		{ 109, 446, 37 },
		{ 109, 447, 37 },
		{ 109, 448, 37 },
		{ 109, 449, 37 },
		{ 109, 450, 37 },
		{ 109, 451, 37 },
		{ 109, 452, 37 },
		{ 109, 453, 37 },
		{ 109, 454, 37 },
		{ 110, 405, 37 },
		{ 110, 406, 37 },
		{ 110, 407, 37 },
		{ 110, 408, 37 },
		{ 110, 409, 37 },
		{ 110, 410, 37 },
		{ 110, 411, 37 },
		{ 110, 412, 37 },
		{ 110, 413, 37 },
		{ 110, 414, 37 },
		{ 110, 415, 37 },
		{ 110, 416, 37 },
		{ 110, 417, 37 },
		{ 110, 418, 37 },
		{ 110, 419, 37 },
		{ 110, 420, 37 },
		{ 110, 421, 37 },
		{ 110, 422, 37 },
		{ 110, 423, 37 },
		{ 110, 424, 37 },
		{ 110, 425, 37 },
		{ 110, 426, 37 },
		{ 110, 427, 37 },
		{ 110, 428, 37 },
		{ 110, 429, 37 },
		{ 110, 430, 37 },
		{ 110, 431, 37 },
		{ 110, 432, 37 },
		{ 110, 433, 37 },
		{ 110, 434, 37 },
		{ 110, 435, 37 },
		{ 110, 436, 37 },
		{ 110, 437, 37 },
		{ 110, 438, 37 },
		{ 110, 439, 37 },
		{ 110, 440, 37 },
		{ 110, 441, 37 },
		{ 110, 442, 37 },
		{ 110, 443, 37 },
		{ 110, 444, 37 },
		{ 110, 445, 37 },
		{ 110, 446, 37 },
		{ 110, 447, 37 },
		{ 110, 448, 37 },
		{ 110, 449, 37 },
		{ 110, 450, 37 },
		{ 110, 451, 37 },
		{ 110, 452, 37 },
		{ 111, 404, 37 },
		{ 111, 405, 37 },
		{ 111, 406, 37 },
		{ 111, 407, 37 },
		{ 111, 408, 37 },
		{ 111, 409, 37 },
		{ 111, 410, 37 },
		{ 111, 411, 37 },
		{ 111, 412, 37 },
		{ 111, 413, 37 },
		{ 111, 414, 37 },
		{ 111, 415, 37 },
		{ 111, 416, 37 },
		{ 111, 417, 37 },
		{ 111, 418, 37 },
		{ 111, 419, 37 },
		{ 111, 420, 37 },
		{ 111, 421, 37 },
		{ 111, 422, 37 },
		{ 111, 423, 37 },
		{ 111, 424, 37 },
		{ 111, 425, 37 },
		{ 111, 426, 37 },
		{ 111, 427, 37 },
		{ 111, 428, 37 },
		{ 111, 429, 37 },
		{ 111, 430, 37 },
		{ 111, 431, 37 },
		{ 111, 432, 37 },
		{ 111, 433, 37 },
		{ 111, 434, 37 },
		{ 111, 435, 37 },
		{ 111, 436, 37 },
		{ 111, 437, 37 },
		{ 111, 438, 37 },
		{ 111, 439, 37 },
		{ 111, 440, 37 },
		{ 111, 441, 37 },
		{ 111, 442, 37 },
		{ 111, 443, 37 },
		{ 111, 444, 37 },
		{ 111, 445, 37 },
		{ 111, 446, 37 },
		{ 111, 447, 37 },
		{ 111, 448, 37 },
		{ 111, 449, 37 },
		{ 111, 450, 37 },
		{ 111, 451, 37 },
		{ 112, 403, 37 },
		{ 112, 404, 37 },
		{ 112, 405, 37 },
		{ 112, 406, 37 },
		{ 112, 407, 37 },
		{ 112, 408, 37 },
		{ 112, 409, 37 },
		{ 112, 410, 37 },
		{ 112, 411, 37 },
		{ 112, 412, 37 },
		{ 112, 413, 37 },
		{ 112, 414, 37 },
		{ 112, 415, 37 },
		{ 112, 416, 37 },
		{ 112, 417, 37 },
		{ 112, 418, 37 },
		{ 112, 419, 37 },
		{ 112, 420, 37 },
		{ 112, 421, 37 },
		{ 112, 422, 37 },
		{ 112, 423, 37 },
		{ 112, 424, 37 },
		{ 112, 425, 37 },
		{ 112, 426, 37 },
		{ 112, 427, 37 },
		{ 112, 428, 37 },
		{ 112, 429, 37 },
		{ 112, 430, 37 },
		{ 112, 431, 37 },
		{ 112, 432, 37 },
		{ 112, 433, 37 },
		{ 112, 434, 37 },
		{ 112, 435, 37 },
		{ 112, 436, 37 },
		{ 112, 437, 37 },
		{ 112, 438, 37 },
		{ 112, 439, 37 },
		{ 112, 440, 37 },
		{ 112, 441, 37 },
		{ 112, 442, 37 },
		{ 112, 443, 37 },
		{ 112, 444, 37 },
		{ 112, 445, 37 },
		{ 112, 446, 37 },
		{ 112, 447, 37 },
		{ 112, 448, 37 },
		{ 112, 449, 37 },
		{ 112, 450, 37 },
		{ 113, 402, 37 },
		{ 113, 403, 37 },
		{ 113, 404, 37 },
		{ 113, 405, 37 },
		{ 113, 406, 37 },
		{ 113, 407, 37 },
		{ 113, 408, 37 },
		{ 113, 409, 37 },
		{ 113, 410, 37 },
		{ 113, 411, 37 },
		{ 113, 412, 37 },
		{ 113, 413, 37 },
		{ 113, 414, 37 },
		{ 113, 415, 37 },
		{ 113, 416, 37 },
		{ 113, 417, 37 },
		{ 113, 418, 37 },
		{ 113, 419, 37 },
		{ 113, 420, 37 },
		{ 113, 421, 37 },
		{ 113, 422, 37 },
		{ 113, 423, 37 },
		{ 113, 424, 37 },
		{ 113, 425, 37 },
		{ 113, 426, 37 },
		{ 113, 427, 37 },
		{ 113, 428, 37 },
		{ 113, 429, 37 },
		{ 113, 430, 37 },
		{ 113, 431, 37 },
		{ 113, 432, 37 },
		{ 113, 433, 37 },
		{ 113, 434, 37 },
		{ 113, 435, 37 },
		{ 113, 436, 37 },
		{ 113, 437, 37 },
		{ 113, 438, 37 },
		{ 113, 439, 37 },
		{ 113, 440, 37 },
		{ 113, 441, 37 },
		{ 113, 442, 37 },
		{ 113, 443, 37 },
		{ 113, 444, 37 },
		{ 113, 445, 37 },
		{ 113, 446, 37 },
		{ 113, 447, 37 },
		{ 113, 448, 37 },
		{ 114, 401, 37 },
		{ 114, 402, 37 },
		{ 114, 403, 37 },
		{ 114, 404, 37 },
		{ 114, 405, 37 },
		{ 114, 406, 37 },
		{ 114, 407, 37 },
		{ 114, 408, 37 },
		{ 114, 409, 37 },
		{ 114, 410, 37 },
		{ 114, 411, 37 },
		{ 114, 412, 37 },
		{ 114, 413, 37 },
		{ 114, 414, 37 },
		{ 114, 415, 37 },
		{ 114, 416, 37 },
		{ 114, 417, 37 },
		{ 114, 418, 37 },
		{ 114, 419, 37 },
		{ 114, 420, 37 },
		{ 114, 421, 37 },
		{ 114, 422, 37 },
		{ 114, 423, 37 },
		{ 114, 424, 37 },
		{ 114, 425, 37 },
		{ 114, 426, 37 },
		{ 114, 427, 37 },
		{ 114, 428, 37 },
		{ 114, 429, 37 },
		{ 114, 430, 37 },
		{ 114, 431, 37 },
		{ 114, 432, 37 },
		{ 114, 433, 37 },
		{ 114, 434, 37 },
		{ 114, 435, 37 },
		{ 114, 436, 37 },
		{ 114, 437, 37 },
		{ 114, 438, 37 },
		{ 114, 439, 37 },
		{ 114, 440, 37 },
		{ 114, 441, 37 },
		{ 114, 442, 37 },
		{ 114, 443, 37 },
		{ 114, 444, 37 },
		{ 114, 445, 37 },
		{ 114, 446, 37 },
		{ 114, 447, 37 },
		{ 115, 400, 37 },
		{ 115, 401, 37 },
		{ 115, 402, 37 },
		{ 115, 403, 37 },
		{ 115, 404, 37 },
		{ 115, 405, 37 },
		{ 115, 406, 37 },
		{ 115, 407, 37 },
		{ 115, 408, 37 },
		{ 115, 409, 37 },
		{ 115, 410, 37 },
		{ 115, 411, 37 },
		{ 115, 412, 37 },
		{ 115, 413, 37 },
		{ 115, 414, 37 },
		{ 115, 415, 37 },
		{ 115, 416, 37 },
		{ 115, 417, 37 },
		{ 115, 418, 37 },
		{ 115, 419, 37 },
		{ 115, 420, 37 },
		{ 115, 421, 37 },
		{ 115, 422, 37 },
		{ 115, 423, 37 },
		{ 115, 424, 37 },
		{ 115, 425, 37 },
		{ 115, 426, 37 },
		{ 115, 427, 37 },
		{ 115, 428, 37 },
		{ 115, 429, 37 },
		{ 115, 430, 37 },
		{ 115, 431, 37 },
		{ 115, 432, 37 },
		{ 115, 433, 37 },
		{ 115, 434, 37 },
		{ 115, 435, 37 },
		{ 115, 436, 37 },
		{ 115, 437, 37 },
		{ 115, 438, 37 },
		{ 115, 439, 37 },
		{ 115, 440, 37 },
		{ 115, 441, 37 },
		{ 115, 442, 37 },
		{ 115, 443, 37 },
		{ 115, 444, 37 },
		{ 115, 445, 37 },
		{ 115, 446, 37 },
		{ 116, 399, 37 },
		{ 116, 400, 37 },
		{ 116, 401, 37 },
		{ 116, 402, 37 },
		{ 116, 403, 37 },
		{ 116, 404, 37 },
		{ 116, 405, 37 },
		{ 116, 406, 37 },
		{ 116, 407, 37 },
		{ 116, 408, 37 },
		{ 116, 409, 37 },
		{ 116, 410, 37 },
		{ 116, 411, 37 },
		{ 116, 412, 37 },
		{ 116, 413, 37 },
		{ 116, 414, 37 },
		{ 116, 415, 37 },
		{ 116, 416, 37 },
		{ 116, 417, 37 },
		{ 116, 418, 37 },
		{ 116, 419, 37 },
		{ 116, 420, 37 },
		{ 116, 421, 37 },
		{ 116, 422, 37 },
		{ 116, 423, 37 },
		{ 116, 424, 37 },
		{ 116, 425, 37 },
		{ 116, 426, 37 },
		{ 116, 427, 37 },
		{ 116, 428, 37 },
		{ 116, 429, 37 },
		{ 116, 430, 37 },
		{ 116, 431, 37 },
		{ 116, 432, 37 },
		{ 116, 433, 37 },
		{ 116, 434, 37 },
		{ 116, 435, 37 },
		{ 116, 436, 37 },
		{ 116, 437, 37 },
		{ 116, 438, 37 },
		{ 116, 439, 37 },
		{ 116, 440, 37 },
		{ 116, 441, 37 },
		{ 116, 442, 37 },
		{ 116, 443, 37 },
		{ 116, 444, 37 },
		{ 117, 398, 37 },
		{ 117, 399, 37 },
		{ 117, 400, 37 },
		{ 117, 401, 37 },
		{ 117, 402, 37 },
		{ 117, 403, 37 },
		{ 117, 404, 37 },
		{ 117, 405, 37 },
		{ 117, 406, 37 },
		{ 117, 407, 37 },
		{ 117, 408, 37 },
		{ 117, 409, 37 },
		{ 117, 410, 37 },
		{ 117, 411, 37 },
		{ 117, 412, 37 },
		{ 117, 413, 37 },
		{ 117, 414, 37 },
		{ 117, 415, 37 },
		{ 117, 416, 37 },
		{ 117, 417, 37 },
		{ 117, 418, 37 },
		{ 117, 419, 37 },
		{ 117, 420, 37 },
		{ 117, 421, 37 },
		{ 117, 422, 37 },
		{ 117, 423, 37 },
		{ 117, 424, 37 },
		{ 117, 425, 37 },
		{ 117, 426, 37 },
		{ 117, 427, 37 },
		{ 117, 428, 37 },
		{ 117, 429, 37 },
		{ 117, 430, 37 },
		{ 117, 431, 37 },
		{ 117, 432, 37 },
		{ 117, 433, 37 },
		{ 117, 434, 37 },
		{ 117, 435, 37 },
		{ 117, 436, 37 },
		{ 117, 437, 37 },
		{ 117, 438, 37 },
		{ 117, 439, 37 },
		{ 117, 440, 37 },
		{ 117, 441, 37 },
		{ 117, 442, 37 },
		{ 117, 443, 37 },
		{ 118, 397, 37 },
		{ 118, 398, 37 },
		{ 118, 399, 37 },
		{ 118, 400, 37 },
		{ 118, 401, 37 },
		{ 118, 402, 37 },
		{ 118, 403, 37 },
		{ 118, 404, 37 },
		{ 118, 405, 37 },
		{ 118, 406, 37 },
		{ 118, 407, 37 },
		{ 118, 408, 37 },
		{ 118, 409, 37 },
		{ 118, 410, 37 },
		{ 118, 411, 37 },
		{ 118, 412, 37 },
		{ 118, 413, 37 },
		{ 118, 414, 37 },
		{ 118, 415, 37 },
		{ 118, 416, 37 },
		{ 118, 417, 37 },
		{ 118, 418, 37 },
		{ 118, 419, 37 },
		{ 118, 420, 37 },
		{ 118, 421, 37 },
		{ 118, 422, 37 },
		{ 118, 423, 37 },
		{ 118, 424, 37 },
		{ 118, 425, 37 },
		{ 118, 426, 37 },
		{ 118, 427, 37 },
		{ 118, 428, 37 },
		{ 118, 429, 37 },
		{ 118, 430, 37 },
		{ 118, 431, 37 },
		{ 118, 432, 37 },
		{ 118, 433, 37 },
		{ 118, 434, 37 },
		{ 118, 435, 37 },
		{ 118, 436, 37 },
		{ 118, 437, 37 },
		{ 118, 438, 37 },
		{ 118, 439, 37 },
		{ 118, 440, 37 },
		{ 118, 441, 37 },
		{ 118, 442, 37 },
		{ 119, 396, 37 },
		{ 119, 397, 37 },
		{ 119, 398, 37 },
		{ 119, 399, 37 },
		{ 119, 400, 37 },
		{ 119, 401, 37 },
		{ 119, 402, 37 },
		{ 119, 403, 37 },
		{ 119, 404, 37 },
		{ 119, 405, 37 },
		{ 119, 406, 37 },
		{ 119, 407, 37 },
		{ 119, 408, 37 },
		{ 119, 409, 37 },
		{ 119, 410, 37 },
		{ 119, 411, 37 },
		{ 119, 412, 37 },
		{ 119, 413, 37 },
		{ 119, 414, 37 },
		{ 119, 415, 37 },
		{ 119, 416, 37 },
		{ 119, 417, 37 },
		{ 119, 418, 37 },
		{ 119, 419, 37 },
		{ 119, 420, 37 },
		{ 119, 421, 37 },
		{ 119, 422, 37 },
		{ 119, 423, 37 },
		{ 119, 424, 37 },
		{ 119, 425, 37 },
		{ 119, 426, 37 },
		{ 119, 427, 37 },
		{ 119, 428, 37 },
		{ 119, 429, 37 },
		{ 119, 430, 37 },
		{ 119, 431, 37 },
		{ 119, 432, 37 },
		{ 119, 433, 37 },
		{ 119, 434, 37 },
		{ 119, 435, 37 },
		{ 119, 436, 37 },
		{ 119, 437, 37 },
		{ 119, 438, 37 },
		{ 119, 439, 37 },
		{ 119, 440, 37 },
		{ 120, 395, 37 },
		{ 120, 396, 37 },
		{ 120, 397, 37 },
		{ 120, 398, 37 },
		{ 120, 399, 37 },
		{ 120, 400, 37 },
		{ 120, 401, 37 },
		{ 120, 402, 37 },
		{ 120, 403, 37 },
		{ 120, 404, 37 },
		{ 120, 405, 37 },
		{ 120, 406, 37 },
		{ 120, 407, 37 },
		{ 120, 408, 37 },
		{ 120, 409, 37 },
		{ 120, 410, 37 },
		{ 120, 411, 37 },
		{ 120, 412, 37 },
		{ 120, 413, 37 },
		{ 120, 414, 37 },
		{ 120, 415, 37 },
		{ 120, 416, 37 },
		{ 120, 417, 37 },
		{ 120, 418, 37 },
		{ 120, 419, 37 },
		{ 120, 420, 37 },
		{ 120, 421, 37 },
		{ 120, 422, 37 },
		{ 120, 423, 37 },
		{ 120, 424, 37 },
		{ 120, 425, 37 },
		{ 120, 426, 37 },
		{ 120, 427, 37 },
		{ 120, 428, 37 },
		{ 120, 429, 37 },
		{ 120, 430, 37 },
		{ 120, 431, 37 },
		{ 120, 432, 37 },
		{ 120, 433, 37 },
		{ 120, 434, 37 },
		{ 120, 435, 37 },
		{ 120, 436, 37 },
		{ 120, 437, 37 },
		{ 120, 438, 37 },
		{ 120, 439, 37 },
		{ 121, 394, 37 },
		{ 121, 395, 37 },
		{ 121, 396, 37 },
		{ 121, 397, 37 },
		{ 121, 398, 37 },
		{ 121, 399, 37 },
		{ 121, 400, 37 },
		{ 121, 401, 37 },
		{ 121, 402, 37 },
		{ 121, 403, 37 },
		{ 121, 404, 37 },
		{ 121, 405, 37 },
		{ 121, 406, 37 },
		{ 121, 407, 37 },
		{ 121, 408, 37 },
		{ 121, 409, 37 },
		{ 121, 410, 37 },
		{ 121, 411, 37 },
		{ 121, 412, 37 },
		{ 121, 413, 37 },
		{ 121, 414, 37 },
		{ 121, 415, 37 },
		{ 121, 416, 37 },
		{ 121, 417, 37 },
		{ 121, 418, 37 },
		{ 121, 419, 37 },
		{ 121, 420, 37 },
		{ 121, 421, 37 },
		{ 121, 422, 37 },
		{ 121, 423, 37 },
		{ 121, 424, 37 },
		{ 121, 425, 37 },
		{ 121, 426, 37 },
		{ 121, 427, 37 },
		{ 121, 428, 37 },
		{ 121, 429, 37 },
		{ 121, 430, 37 },
		{ 121, 431, 37 },
		{ 121, 432, 37 },
		{ 121, 433, 37 },
		{ 121, 434, 37 },
		{ 121, 435, 37 },
		{ 121, 436, 37 },
		{ 121, 437, 37 },
		{ 121, 438, 37 },
		{ 122, 393, 37 },
		{ 122, 394, 37 },
		{ 122, 395, 37 },
		{ 122, 396, 37 },
		{ 122, 397, 37 },
		{ 122, 398, 37 },
		{ 122, 399, 37 },
		{ 122, 400, 37 },
		{ 122, 401, 37 },
		{ 122, 402, 37 },
		{ 122, 403, 37 },
		{ 122, 404, 37 },
		{ 122, 405, 37 },
		{ 122, 406, 37 },
		{ 122, 407, 37 },
		{ 122, 408, 37 },
		{ 122, 409, 37 },
		{ 122, 410, 37 },
		{ 122, 411, 37 },
		{ 122, 412, 37 },
		{ 122, 413, 37 },
		{ 122, 414, 37 },
		{ 122, 415, 37 },
		{ 122, 416, 37 },
		{ 122, 417, 37 },
		{ 122, 418, 37 },
		{ 122, 419, 37 },
		{ 122, 420, 37 },
		{ 122, 421, 37 },
		{ 122, 422, 37 },
		{ 122, 423, 37 },
		{ 122, 424, 37 },
		{ 122, 425, 37 },
		{ 122, 426, 37 },
		{ 122, 427, 37 },
		{ 122, 428, 37 },
		{ 122, 429, 37 },
		{ 122, 430, 37 },
		{ 122, 431, 37 },
		{ 122, 432, 37 },
		{ 122, 433, 37 },
		{ 122, 434, 37 },
		{ 122, 435, 37 },
		{ 122, 436, 37 },
		{ 123, 392, 37 },
		{ 123, 393, 37 },
		{ 123, 394, 37 },
		{ 123, 395, 37 },
		{ 123, 396, 37 },
		{ 123, 397, 37 },
		{ 123, 398, 37 },
		{ 123, 399, 37 },
		{ 123, 400, 37 },
		{ 123, 401, 37 },
		{ 123, 402, 37 },
		{ 123, 403, 37 },
		{ 123, 404, 37 },
		{ 123, 405, 37 },
		{ 123, 406, 37 },
		{ 123, 407, 37 },
		{ 123, 408, 37 },
		{ 123, 409, 37 },
		{ 123, 410, 37 },
		{ 123, 411, 37 },
		{ 123, 412, 37 },
		{ 123, 413, 37 },
		{ 123, 414, 37 },
		{ 123, 415, 37 },
		{ 123, 416, 37 },
		{ 123, 417, 37 },
		{ 123, 418, 37 },
		{ 123, 419, 37 },
		{ 123, 420, 37 },
		{ 123, 421, 37 },
		{ 123, 422, 37 },
		{ 123, 423, 37 },
		{ 123, 424, 37 },
		{ 123, 425, 37 },
		{ 123, 426, 37 },
		{ 123, 427, 37 },
		{ 123, 428, 37 },
		{ 123, 429, 37 },
		{ 123, 430, 37 },
		{ 123, 431, 37 },
		{ 123, 432, 37 },
		{ 123, 433, 37 },
		{ 123, 434, 37 },
		{ 123, 435, 37 },
		{ 124, 391, 37 },
		{ 124, 392, 37 },
		{ 124, 393, 37 },
		{ 124, 394, 37 },
		{ 124, 395, 37 },
		{ 124, 396, 37 },
		{ 124, 397, 37 },
		{ 124, 398, 37 },
		{ 124, 399, 37 },
		{ 124, 400, 37 },
		{ 124, 401, 37 },
		{ 124, 402, 37 },
		{ 124, 403, 37 },
		{ 124, 404, 37 },
		{ 124, 405, 37 },
		{ 124, 406, 37 },
		{ 124, 407, 37 },
		{ 124, 408, 37 },
		{ 124, 409, 37 },
		{ 124, 410, 37 },
		{ 124, 411, 37 },
		{ 124, 412, 37 },
		{ 124, 413, 37 },
		{ 124, 414, 37 },
		{ 124, 415, 37 },
		{ 124, 416, 37 },
		{ 124, 417, 37 },
		{ 124, 418, 37 },
		{ 124, 419, 37 },
		{ 124, 420, 37 },
		{ 124, 421, 37 },
		{ 124, 422, 37 },
		{ 124, 423, 37 },
		{ 124, 424, 37 },
		{ 124, 425, 37 },
		{ 124, 426, 37 },
		{ 124, 427, 37 },
		{ 124, 428, 37 },
		{ 124, 429, 37 },
		{ 124, 430, 37 },
		{ 124, 431, 37 },
		{ 124, 432, 37 },
		{ 124, 433, 37 },
		{ 124, 434, 37 },
		{ 125, 390, 37 },
		{ 125, 391, 37 },
		{ 125, 392, 37 },
		{ 125, 393, 37 },
		{ 125, 394, 37 },
		{ 125, 395, 37 },
		{ 125, 396, 37 },
		{ 125, 397, 37 },
		{ 125, 398, 37 },
		{ 125, 399, 37 },
		{ 125, 400, 37 },
		{ 125, 401, 37 },
		{ 125, 402, 37 },
		{ 125, 403, 37 },
		{ 125, 404, 37 },
		{ 125, 405, 37 },
		{ 125, 406, 37 },
		{ 125, 407, 37 },
		{ 125, 408, 37 },
		{ 125, 409, 37 },
		{ 125, 410, 37 },
		{ 125, 411, 37 },
		{ 125, 412, 37 },
		{ 125, 413, 37 },
		{ 125, 414, 37 },
		{ 125, 415, 37 },
		{ 125, 416, 37 },
		{ 125, 417, 37 },
		{ 125, 418, 37 },
		{ 125, 419, 37 },
		{ 125, 420, 37 },
		{ 125, 421, 37 },
		{ 125, 422, 37 },
		{ 125, 423, 37 },
		{ 125, 424, 37 },
		{ 125, 425, 37 },
		{ 125, 426, 37 },
		{ 125, 427, 37 },
		{ 125, 428, 37 },
		{ 125, 429, 37 },
		{ 125, 430, 37 },
		{ 125, 431, 37 },
		{ 125, 432, 37 },
		{ 126, 389, 37 },
		{ 126, 390, 37 },
		{ 126, 391, 37 },
		{ 126, 392, 37 },
		{ 126, 393, 37 },
		{ 126, 394, 37 },
		{ 126, 395, 37 },
		{ 126, 396, 37 },
		{ 126, 397, 37 },
		{ 126, 398, 37 },
		{ 126, 399, 37 },
		{ 126, 400, 37 },
		{ 126, 401, 37 },
		{ 126, 402, 37 },
		{ 126, 403, 37 },
		{ 126, 404, 37 },
		{ 126, 405, 37 },
		{ 126, 406, 37 },
		{ 126, 407, 37 },
		{ 126, 408, 37 },
		{ 126, 409, 37 },
		{ 126, 410, 37 },
		{ 126, 411, 37 },
		{ 126, 412, 37 },
		{ 126, 413, 37 },
		{ 126, 414, 37 },
		{ 126, 415, 37 },
		{ 126, 416, 37 },
		{ 126, 417, 37 },
		{ 126, 418, 37 },
		{ 126, 419, 37 },
		{ 126, 420, 37 },
		{ 126, 421, 37 },
		{ 126, 422, 37 },
		{ 126, 423, 37 },
		{ 126, 424, 37 },
		{ 126, 425, 37 },
		{ 126, 426, 37 },
		{ 126, 427, 37 },
		{ 126, 428, 37 },
		{ 126, 429, 37 },
		{ 126, 430, 37 },
		{ 126, 431, 37 },
		{ 127, 388, 37 },
		{ 127, 389, 37 },
		{ 127, 390, 37 },
		{ 127, 391, 37 },
		{ 127, 392, 37 },
		{ 127, 393, 37 },
		{ 127, 394, 37 },
		{ 127, 395, 37 },
		{ 127, 396, 37 },
		{ 127, 397, 37 },
		{ 127, 398, 37 },
		{ 127, 399, 37 },
		{ 127, 400, 37 },
		{ 127, 401, 37 },
		{ 127, 402, 37 },
		{ 127, 403, 37 },
		{ 127, 404, 37 },
		{ 127, 405, 37 },
		{ 127, 406, 37 },
		{ 127, 407, 37 },
		{ 127, 408, 37 },
		{ 127, 409, 37 },
		{ 127, 410, 37 },
		{ 127, 411, 37 },
		{ 127, 412, 37 },
		{ 127, 413, 37 },
		{ 127, 414, 37 },
		{ 127, 415, 37 },
		{ 127, 416, 37 },
		{ 127, 417, 37 },
		{ 127, 418, 37 },
		{ 127, 419, 37 },
		{ 127, 420, 37 },
		{ 127, 421, 37 },
		{ 127, 422, 37 },
		{ 127, 423, 37 },
		{ 127, 424, 37 },
		{ 127, 425, 37 },
		{ 127, 426, 37 },
		{ 127, 427, 37 },
		{ 127, 428, 37 },
		{ 127, 429, 37 },
		{ 127, 430, 37 },
		{ 128, 387, 37 },
		{ 128, 388, 37 },
		{ 128, 389, 37 },
		{ 128, 390, 37 },
		{ 128, 391, 37 },
		{ 128, 392, 37 },
		{ 128, 393, 37 },
		{ 128, 394, 37 },
		{ 128, 395, 37 },
		{ 128, 396, 37 },
		{ 128, 397, 37 },
		{ 128, 398, 37 },
		{ 128, 399, 37 },
		{ 128, 400, 37 },
		{ 128, 401, 37 },
		{ 128, 402, 37 },
		{ 128, 403, 37 },
		{ 128, 404, 37 },
		{ 128, 405, 37 },
		{ 128, 406, 37 },
		{ 128, 407, 37 },
		{ 128, 408, 37 },
		{ 128, 409, 37 },
		{ 128, 410, 37 },
		{ 128, 411, 37 },
		{ 128, 412, 37 },
		{ 128, 413, 37 },
		{ 128, 414, 37 },
		{ 128, 415, 37 },
		{ 128, 416, 37 },
		{ 128, 417, 37 },
		{ 128, 418, 37 },
		{ 128, 419, 37 },
		{ 128, 420, 37 },
		{ 128, 421, 37 },
		{ 128, 422, 37 },
		{ 128, 423, 37 },
		{ 128, 424, 37 },
		{ 128, 425, 37 },
		{ 128, 426, 37 },
		{ 128, 427, 37 },
		{ 128, 428, 37 },
		{ 129, 386, 37 },
		{ 129, 387, 37 },
		{ 129, 388, 37 },
		{ 129, 389, 37 },
		{ 129, 390, 37 },
		{ 129, 391, 37 },
		{ 129, 392, 37 },
		{ 129, 393, 37 },
		{ 129, 394, 37 },
		{ 129, 395, 37 },
		{ 129, 396, 37 },
		{ 129, 397, 37 },
		{ 129, 398, 37 },
		{ 129, 399, 37 },
		{ 129, 400, 37 },
		{ 129, 401, 37 },
		{ 129, 402, 37 },
		{ 129, 403, 37 },
		{ 129, 404, 37 },
		{ 129, 405, 37 },
		{ 129, 406, 37 },
		{ 129, 407, 37 },
		{ 129, 408, 37 },
		{ 129, 409, 37 },
		{ 129, 410, 37 },
		{ 129, 411, 37 },
		{ 129, 412, 37 },
		{ 129, 413, 37 },
		{ 129, 414, 37 },
		{ 129, 415, 37 },
		{ 129, 416, 37 },
		{ 129, 417, 37 },
		{ 129, 418, 37 },
		{ 129, 419, 37 },
		{ 129, 420, 37 },
		{ 129, 421, 37 },
		{ 129, 422, 37 },
		{ 129, 423, 37 },
		{ 129, 424, 37 },
		{ 129, 425, 37 },
		{ 129, 426, 37 },
		{ 129, 427, 37 },
		{ 130, 385, 37 },
		{ 130, 386, 37 },
		{ 130, 387, 37 },
		{ 130, 388, 37 },
		{ 130, 389, 37 },
		{ 130, 390, 37 },
		{ 130, 391, 37 },
		{ 130, 392, 37 },
		{ 130, 393, 37 },
		{ 130, 394, 37 },
		{ 130, 395, 37 },
		{ 130, 396, 37 },
		{ 130, 397, 37 },
		{ 130, 398, 37 },
		{ 130, 399, 37 },
		{ 130, 400, 37 },
		{ 130, 401, 37 },
		{ 130, 402, 37 },
		{ 130, 403, 37 },
		{ 130, 404, 37 },
		{ 130, 405, 37 },
		{ 130, 406, 37 },
		{ 130, 407, 37 },
		{ 130, 408, 37 },
		{ 130, 409, 37 },
		{ 130, 410, 37 },
		{ 130, 411, 37 },
		{ 130, 412, 37 },
		{ 130, 413, 37 },
		{ 130, 414, 37 },
		{ 130, 415, 37 },
		{ 130, 416, 37 },
		{ 130, 417, 37 },
		{ 130, 418, 37 },
		{ 130, 419, 37 },
		{ 130, 420, 37 },
		{ 130, 421, 37 },
		{ 130, 422, 37 },
		{ 130, 423, 37 },
		{ 130, 424, 37 },
		{ 130, 425, 37 },
		{ 130, 426, 37 },
		{ 131, 384, 37 },
		{ 131, 385, 37 },
		{ 131, 386, 37 },
		{ 131, 387, 37 },
		{ 131, 388, 37 },
		{ 131, 389, 37 },
		{ 131, 390, 37 },
		{ 131, 391, 37 },
		{ 131, 392, 37 },
		{ 131, 393, 37 },
		{ 131, 394, 37 },
		{ 131, 395, 37 },
		{ 131, 396, 37 },
		{ 131, 397, 37 },
		{ 131, 398, 37 },
		{ 131, 399, 37 },
		{ 131, 400, 37 },
		{ 131, 401, 37 },
		{ 131, 402, 37 },
		{ 131, 403, 37 },
		{ 131, 404, 37 },
		{ 131, 405, 37 },
		{ 131, 406, 37 },
		{ 131, 407, 37 },
		{ 131, 408, 37 },
		{ 131, 409, 37 },
		{ 131, 410, 37 },
		{ 131, 411, 37 },
		{ 131, 412, 37 },
		{ 131, 413, 37 },
		{ 131, 414, 37 },
		{ 131, 415, 37 },
		{ 131, 416, 37 },
		{ 131, 417, 37 },
		{ 131, 418, 37 },
		{ 131, 419, 37 },
		{ 131, 420, 37 },
		{ 131, 421, 37 },
		{ 131, 422, 37 },
		{ 131, 423, 37 },
		{ 131, 424, 37 },
		{ 132, 383, 37 },
		{ 132, 384, 37 },
		{ 132, 385, 37 },
		{ 132, 386, 37 },
		{ 132, 387, 37 },
		{ 132, 388, 37 },
		{ 132, 389, 37 },
		{ 132, 390, 37 },
		{ 132, 391, 37 },
		{ 132, 392, 37 },
		{ 132, 393, 37 },
		{ 132, 394, 37 },
		{ 132, 395, 37 },
		{ 132, 396, 37 },
		{ 132, 397, 37 },
		{ 132, 398, 37 },
		{ 132, 399, 37 },
		{ 132, 400, 37 },
		{ 132, 401, 37 },
		{ 132, 402, 37 },
		{ 132, 403, 37 },
		{ 132, 404, 37 },
		{ 132, 405, 37 },
		{ 132, 406, 37 },
		{ 132, 407, 37 },
		{ 132, 408, 37 },
		{ 132, 409, 37 },
		{ 132, 410, 37 },
		{ 132, 411, 37 },
		{ 132, 412, 37 },
		{ 132, 413, 37 },
		{ 132, 414, 37 },
		{ 132, 415, 37 },
		{ 132, 416, 37 },
		{ 132, 417, 37 },
		{ 132, 418, 37 },
		{ 132, 419, 37 },
		{ 132, 420, 37 },
		{ 132, 421, 37 },
		{ 132, 422, 37 },
		{ 132, 423, 37 },
		{ 133, 382, 37 },
		{ 133, 383, 37 },
		{ 133, 384, 37 },
		{ 133, 385, 37 },
		{ 133, 386, 37 },
		{ 133, 387, 37 },
		{ 133, 388, 37 },
		{ 133, 389, 37 },
		{ 133, 390, 37 },
		{ 133, 391, 37 },
		{ 133, 392, 37 },
		{ 133, 393, 37 },
		{ 133, 394, 37 },
		{ 133, 395, 37 },
		{ 133, 396, 37 },
		{ 133, 397, 37 },
		{ 133, 398, 37 },
		{ 133, 399, 37 },
		{ 133, 400, 37 },
		{ 133, 401, 37 },
		{ 133, 402, 37 },
		{ 133, 403, 37 },
		{ 133, 404, 37 },
		{ 133, 405, 37 },
		{ 133, 406, 37 },
		{ 133, 407, 37 },
		{ 133, 408, 37 },
		{ 133, 409, 37 },
		{ 133, 410, 37 },
		{ 133, 411, 37 },
		{ 133, 412, 37 },
		{ 133, 413, 37 },
		{ 133, 414, 37 },
		{ 133, 415, 37 },
		{ 133, 416, 37 },
		{ 133, 417, 37 },
		{ 133, 418, 37 },
		{ 133, 419, 37 },
		{ 133, 420, 37 },
		{ 133, 421, 37 },
		{ 133, 422, 37 },
		{ 134, 381, 37 },
		{ 134, 382, 37 },
		{ 134, 383, 37 },
		{ 134, 384, 37 },
		{ 134, 385, 37 },
		{ 134, 386, 37 },
		{ 134, 387, 37 },
		{ 134, 388, 37 },
		{ 134, 389, 37 },
		{ 134, 390, 37 },
		{ 134, 391, 37 },
		{ 134, 392, 37 },
		{ 134, 393, 37 },
		{ 134, 394, 37 },
		{ 134, 395, 37 },
		{ 134, 396, 37 },
		{ 134, 397, 37 },
		{ 134, 398, 37 },
		{ 134, 399, 37 },
		{ 134, 400, 37 },
		{ 134, 401, 37 },
		{ 134, 402, 37 },
		{ 134, 403, 37 },
		{ 134, 404, 37 },
		{ 134, 405, 37 },
		{ 134, 406, 37 },
		{ 134, 407, 37 },
		{ 134, 408, 37 },
		{ 134, 409, 37 },
		{ 134, 410, 37 },
		{ 134, 411, 37 },
		{ 134, 412, 37 },
		{ 134, 413, 37 },
		{ 134, 414, 37 },
		{ 134, 415, 37 },
		{ 134, 416, 37 },
		{ 134, 417, 37 },
		{ 134, 418, 37 },
		{ 134, 419, 37 },
		{ 134, 420, 37 },
		{ 135, 380, 37 },
		{ 135, 381, 37 },
		{ 135, 382, 37 },
		{ 135, 383, 37 },
		{ 135, 384, 37 },
		{ 135, 385, 37 },
		{ 135, 386, 37 },
		{ 135, 387, 37 },
		{ 135, 388, 37 },
		{ 135, 389, 37 },
		{ 135, 390, 37 },
		{ 135, 391, 37 },
		{ 135, 392, 37 },
		{ 135, 393, 37 },
		{ 135, 394, 37 },
		{ 135, 395, 37 },
		{ 135, 396, 37 },
		{ 135, 397, 37 },
		{ 135, 398, 37 },
		{ 135, 399, 37 },
		{ 135, 400, 37 },
		{ 135, 401, 37 },
		{ 135, 402, 37 },
		{ 135, 403, 37 },
		{ 135, 404, 37 },
		{ 135, 405, 37 },
		{ 135, 406, 37 },
		{ 135, 407, 37 },
		{ 135, 408, 37 },
		{ 135, 409, 37 },
		{ 135, 410, 37 },
		{ 135, 411, 37 },
		{ 135, 412, 37 },
		{ 135, 413, 37 },
		{ 135, 414, 37 },
		{ 135, 415, 37 },
		{ 135, 416, 37 },
		{ 135, 417, 37 },
		{ 135, 418, 37 },
		{ 135, 419, 37 },
		{ 136, 379, 37 },
		{ 136, 380, 37 },
		{ 136, 381, 37 },
		{ 136, 382, 37 },
		{ 136, 383, 37 },
		{ 136, 384, 37 },
		{ 136, 385, 37 },
		{ 136, 386, 37 },
		{ 136, 387, 37 },
		{ 136, 388, 37 },
		{ 136, 389, 37 },
		{ 136, 390, 37 },
		{ 136, 391, 37 },
		{ 136, 392, 37 },
		{ 136, 393, 37 },
		{ 136, 394, 37 },
		{ 136, 395, 37 },
		{ 136, 396, 37 },
		{ 136, 397, 37 },
		{ 136, 398, 37 },
		{ 136, 399, 37 },
		{ 136, 400, 37 },
		{ 136, 401, 37 },
		{ 136, 402, 37 },
		{ 136, 403, 37 },
		{ 136, 404, 37 },
		{ 136, 405, 37 },
		{ 136, 406, 37 },
		{ 136, 407, 37 },
		{ 136, 408, 37 },
		{ 136, 409, 37 },
		{ 136, 410, 37 },
		{ 136, 411, 37 },
		{ 136, 412, 37 },
		{ 136, 413, 37 },
		{ 136, 414, 37 },
		{ 136, 415, 37 },
		{ 136, 416, 37 },
		{ 136, 417, 37 },
		{ 136, 418, 37 },
		{ 137, 378, 37 },
		{ 137, 379, 37 },
		{ 137, 380, 37 },
		{ 137, 381, 37 },
		{ 137, 382, 37 },
		{ 137, 383, 37 },
		{ 137, 384, 37 },
		{ 137, 385, 37 },
		{ 137, 386, 37 },
		{ 137, 387, 37 },
		{ 137, 388, 37 },
		{ 137, 389, 37 },
		{ 137, 390, 37 },
		{ 137, 391, 37 },
		{ 137, 392, 37 },
		{ 137, 393, 37 },
		{ 137, 394, 37 },
		{ 137, 395, 37 },
		{ 137, 396, 37 },
		{ 137, 397, 37 },
		{ 137, 398, 37 },
		{ 137, 399, 37 },
		{ 137, 400, 37 },
		{ 137, 401, 37 },
		{ 137, 402, 37 },
		{ 137, 403, 37 },
		{ 137, 404, 37 },
		{ 137, 405, 37 },
		{ 137, 406, 37 },
		{ 137, 407, 37 },
		{ 137, 408, 37 },
		{ 137, 409, 37 },
		{ 137, 410, 37 },
		{ 137, 411, 37 },
		{ 137, 412, 37 },
		{ 137, 413, 37 },
		{ 137, 414, 37 },
		{ 137, 415, 37 },
		{ 137, 416, 37 },
		{ 138, 377, 37 },
		{ 138, 378, 37 },
		{ 138, 379, 37 },
		{ 138, 380, 37 },
		{ 138, 381, 37 },
		{ 138, 382, 37 },
		{ 138, 383, 37 },
		{ 138, 384, 37 },
		{ 138, 385, 37 },
		{ 138, 386, 37 },
		{ 138, 387, 37 },
		{ 138, 388, 37 },
		{ 138, 389, 37 },
		{ 138, 390, 37 },
		{ 138, 391, 37 },
		{ 138, 392, 37 },
		{ 138, 393, 37 },
		{ 138, 394, 37 },
		{ 138, 395, 37 },
		{ 138, 396, 37 },
		{ 138, 397, 37 },
		{ 138, 398, 37 },
		{ 138, 399, 37 },
		{ 138, 400, 37 },
		{ 138, 401, 37 },
		{ 138, 402, 37 },
		{ 138, 403, 37 },
		{ 138, 404, 37 },
		{ 138, 405, 37 },
		{ 138, 406, 37 },
		{ 138, 407, 37 },
		{ 138, 408, 37 },
		{ 138, 409, 37 },
		{ 138, 410, 37 },
		{ 138, 411, 37 },
		{ 138, 412, 37 },
		{ 138, 413, 37 },
		{ 138, 414, 37 },
		{ 138, 415, 37 },
		{ 139, 376, 37 },
		{ 139, 377, 37 },
		{ 139, 378, 37 },
		{ 139, 379, 37 },
		{ 139, 380, 37 },
		{ 139, 381, 37 },
		{ 139, 382, 37 },
		{ 139, 383, 37 },
		{ 139, 384, 37 },
		{ 139, 385, 37 },
		{ 139, 386, 37 },
		{ 139, 387, 37 },
		{ 139, 388, 37 },
		{ 139, 389, 37 },
		{ 139, 390, 37 },
		{ 139, 391, 37 },
		{ 139, 392, 37 },
		{ 139, 393, 37 },
		{ 139, 394, 37 },
		{ 139, 395, 37 },
		{ 139, 396, 37 },
		{ 139, 397, 37 },
		{ 139, 398, 37 },
		{ 139, 399, 37 },
		{ 139, 400, 37 },
		{ 139, 401, 37 },
		{ 139, 402, 37 },
		{ 139, 403, 37 },
		{ 139, 404, 37 },
		{ 139, 405, 37 },
		{ 139, 406, 37 },
		{ 139, 407, 37 },
		{ 139, 408, 37 },
		{ 139, 409, 37 },
		{ 139, 410, 37 },
		{ 139, 411, 37 },
		{ 139, 412, 37 },
		{ 139, 413, 37 },
		{ 139, 414, 37 },
		{ 140, 375, 37 },
		{ 140, 376, 37 },
		{ 140, 377, 37 },
		{ 140, 378, 37 },
		{ 140, 379, 37 },
		{ 140, 380, 37 },
		{ 140, 381, 37 },
		{ 140, 382, 37 },
		{ 140, 383, 37 },
		{ 140, 384, 37 },
		{ 140, 385, 37 },
		{ 140, 386, 37 },
		{ 140, 387, 37 },
		{ 140, 388, 37 },
		{ 140, 389, 37 },
		{ 140, 390, 37 },
		{ 140, 391, 37 },
		{ 140, 392, 37 },
		{ 140, 393, 37 },
		{ 140, 394, 37 },
		{ 140, 395, 37 },
		{ 140, 396, 37 },
		{ 140, 397, 37 },
		{ 140, 398, 37 },
		{ 140, 399, 37 },
		{ 140, 400, 37 },
		{ 140, 401, 37 },
		{ 140, 402, 37 },
		{ 140, 403, 37 },
		{ 140, 404, 37 },
		{ 140, 405, 37 },
		{ 140, 406, 37 },
		{ 140, 407, 37 },
		{ 140, 408, 37 },
		{ 140, 409, 37 },
		{ 140, 410, 37 },
		{ 140, 411, 37 },
		{ 140, 412, 37 },
		{ 141, 374, 37 },
		{ 141, 375, 37 },
		{ 141, 376, 37 },
		{ 141, 377, 37 },
		{ 141, 378, 37 },
		{ 141, 379, 37 },
		{ 141, 380, 37 },
		{ 141, 381, 37 },
		{ 141, 382, 37 },
		{ 141, 383, 37 },
		{ 141, 384, 37 },
		{ 141, 385, 37 },
		{ 141, 386, 37 },
		{ 141, 387, 37 },
		{ 141, 388, 37 },
		{ 141, 389, 37 },
		{ 141, 390, 37 },
		{ 141, 391, 37 },
		{ 141, 392, 37 },
		{ 141, 393, 37 },
		{ 141, 394, 37 },
		{ 141, 395, 37 },
		{ 141, 396, 37 },
		{ 141, 397, 37 },
		{ 141, 398, 37 },
		{ 141, 399, 37 },
		{ 141, 400, 37 },
		{ 141, 401, 37 },
		{ 141, 402, 37 },
		{ 141, 403, 37 },
		{ 141, 404, 37 },
		{ 141, 405, 37 },
		{ 141, 406, 37 },
		{ 141, 407, 37 },
		{ 141, 408, 37 },
		{ 141, 409, 37 },
		{ 141, 410, 37 },
		{ 141, 411, 37 },
		{ 142, 373, 37 },
		{ 142, 374, 37 },
		{ 142, 375, 37 },
		{ 142, 376, 37 },
		{ 142, 377, 37 },
		{ 142, 378, 37 },
		{ 142, 379, 37 },
		{ 142, 380, 37 },
		{ 142, 381, 37 },
		{ 142, 382, 37 },
		{ 142, 383, 37 },
		{ 142, 384, 37 },
		{ 142, 385, 37 },
		{ 142, 386, 37 },
		{ 142, 387, 37 },
		{ 142, 388, 37 },
		{ 142, 389, 37 },
		{ 142, 390, 37 },
		{ 142, 391, 37 },
		{ 142, 392, 37 },
		{ 142, 393, 37 },
		{ 142, 394, 37 },
		{ 142, 395, 37 },
		{ 142, 396, 37 },
		{ 142, 397, 37 },
		{ 142, 398, 37 },
		{ 142, 399, 37 },
		{ 142, 400, 37 },
		{ 142, 401, 37 },
		{ 142, 402, 37 },
		{ 142, 403, 37 },
		{ 142, 404, 37 },
		{ 142, 405, 37 },
		{ 142, 406, 37 },
		{ 142, 407, 37 },
		{ 142, 408, 37 },
		{ 142, 409, 37 },
		{ 142, 410, 37 },
		{ 143, 372, 37 },
		{ 143, 373, 37 },
		{ 143, 374, 37 },
		{ 143, 375, 37 },
		{ 143, 376, 37 },
		{ 143, 377, 37 },
		{ 143, 378, 37 },
		{ 143, 379, 37 },
		{ 143, 380, 37 },
		{ 143, 381, 37 },
		{ 143, 382, 37 },
		{ 143, 383, 37 },
		{ 143, 384, 37 },
		{ 143, 385, 37 },
		{ 143, 386, 37 },
		{ 143, 387, 37 },
		{ 143, 388, 37 },
		{ 143, 389, 37 },
		{ 143, 390, 37 },
		{ 143, 391, 37 },
		{ 143, 392, 37 },
		{ 143, 393, 37 },
		{ 143, 394, 37 },
		{ 143, 395, 37 },
		{ 143, 396, 37 },
		{ 143, 397, 37 },
		{ 143, 398, 37 },
		{ 143, 399, 37 },
		{ 143, 400, 37 },
		{ 143, 401, 37 },
		{ 143, 402, 37 },
		{ 143, 403, 37 },
		{ 143, 404, 37 },
		{ 143, 405, 37 },
		{ 143, 406, 37 },
		{ 143, 407, 37 },
		{ 143, 408, 37 },
		{ 144, 371, 37 },
		{ 144, 372, 37 },
		{ 144, 373, 37 },
		{ 144, 374, 37 },
		{ 144, 375, 37 },
		{ 144, 376, 37 },
		{ 144, 377, 37 },
		{ 144, 378, 37 },
		{ 144, 379, 37 },
		{ 144, 380, 37 },
		{ 144, 381, 37 },
		{ 144, 382, 37 },
		{ 144, 383, 37 },
		{ 144, 384, 37 },
		{ 144, 385, 37 },
		{ 144, 386, 37 },
		{ 144, 387, 37 },
		{ 144, 388, 37 },
		{ 144, 389, 37 },
		{ 144, 390, 37 },
		{ 144, 391, 37 },
		{ 144, 392, 37 },
		{ 144, 393, 37 },
		{ 144, 394, 37 },
		{ 144, 395, 37 },
		{ 144, 396, 37 },
		{ 144, 397, 37 },
		{ 144, 398, 37 },
		{ 144, 399, 37 },
		{ 144, 400, 37 },
		{ 144, 401, 37 },
		{ 144, 402, 37 },
		{ 144, 403, 37 },
		{ 144, 404, 37 },
		{ 144, 405, 37 },
		{ 144, 406, 37 },
		{ 144, 407, 37 },
		{ 145, 370, 37 },
		{ 145, 371, 37 },
		{ 145, 372, 37 },
		{ 145, 373, 37 },
		{ 145, 374, 37 },
		{ 145, 375, 37 },
		{ 145, 376, 37 },
		{ 145, 377, 37 },
		{ 145, 378, 37 },
		{ 145, 379, 37 },
		{ 145, 380, 37 },
		{ 145, 381, 37 },
		{ 145, 382, 37 },
		{ 145, 383, 37 },
		{ 145, 384, 37 },
		{ 145, 385, 37 },
		{ 145, 386, 37 },
		{ 145, 387, 37 },
		{ 145, 388, 37 },
		{ 145, 389, 37 },
		{ 145, 390, 37 },
		{ 145, 391, 37 },
		{ 145, 392, 37 },
		{ 145, 393, 37 },
		{ 145, 394, 37 },
		{ 145, 395, 37 },
		{ 145, 396, 37 },
		{ 145, 397, 37 },
		{ 145, 398, 37 },
		{ 145, 399, 37 },
		{ 145, 400, 37 },
		{ 145, 401, 37 },
		{ 145, 402, 37 },
		{ 145, 403, 37 },
		{ 145, 404, 37 },
		{ 145, 405, 37 },
		{ 145, 406, 37 },
		{ 146, 369, 37 },
		{ 146, 370, 37 },
		{ 146, 371, 37 },
		{ 146, 372, 37 },
		{ 146, 373, 37 },
		{ 146, 374, 37 },
		{ 146, 375, 37 },
		{ 146, 376, 37 },
		{ 146, 377, 37 },
		{ 146, 378, 37 },
		{ 146, 379, 37 },
		{ 146, 380, 37 },
		{ 146, 381, 37 },
		{ 146, 382, 37 },
		{ 146, 383, 37 },
		{ 146, 384, 37 },
		{ 146, 385, 37 },
		{ 146, 386, 37 },
		{ 146, 387, 37 },
		{ 146, 388, 37 },
		{ 146, 389, 37 },
		{ 146, 390, 37 },
		{ 146, 391, 37 },
		{ 146, 392, 37 },
		{ 146, 393, 37 },
		{ 146, 394, 37 },
		{ 146, 395, 37 },
		{ 146, 396, 37 },
		{ 146, 397, 37 },
		{ 146, 398, 37 },
		{ 146, 399, 37 },
		{ 146, 400, 37 },
		{ 146, 401, 37 },
		{ 146, 402, 37 },
		{ 146, 403, 37 },
		{ 146, 404, 37 },
		{ 147, 368, 37 },
		{ 147, 369, 37 },
		{ 147, 370, 37 },
		{ 147, 371, 37 },
		{ 147, 372, 37 },
		{ 147, 373, 37 },
		{ 147, 374, 37 },
		{ 147, 375, 37 },
		{ 147, 376, 37 },
		{ 147, 377, 37 },
		{ 147, 378, 37 },
		{ 147, 379, 37 },
		{ 147, 380, 37 },
		{ 147, 381, 37 },
		{ 147, 382, 37 },
		{ 147, 383, 37 },
		{ 147, 384, 37 },
		{ 147, 385, 37 },
		{ 147, 386, 37 },
		{ 147, 387, 37 },
		{ 147, 388, 37 },
		{ 147, 389, 37 },
		{ 147, 390, 37 },
		{ 147, 391, 37 },
		{ 147, 392, 37 },
		{ 147, 393, 37 },
		{ 147, 394, 37 },
		{ 147, 395, 37 },
		{ 147, 396, 37 },
		{ 147, 397, 37 },
		{ 147, 398, 37 },
		{ 147, 399, 37 },
		{ 147, 400, 37 },
		{ 147, 401, 37 },
		{ 147, 402, 37 },
		{ 147, 403, 37 },
		{ 148, 367, 37 },
		{ 148, 368, 37 },
		{ 148, 369, 37 },
		{ 148, 370, 37 },
		{ 148, 371, 37 },
		{ 148, 372, 37 },
		{ 148, 373, 37 },
		{ 148, 374, 37 },
		{ 148, 375, 37 },
		{ 148, 376, 37 },
		{ 148, 377, 37 },
		{ 148, 378, 37 },
		{ 148, 379, 37 },
		{ 148, 380, 37 },
		{ 148, 381, 37 },
		{ 148, 382, 37 },
		{ 148, 383, 37 },
		{ 148, 384, 37 },
		{ 148, 385, 37 },
		{ 148, 386, 37 },
		{ 148, 387, 37 },
		{ 148, 388, 37 },
		{ 148, 389, 37 },
		{ 148, 390, 37 },
		{ 148, 391, 37 },
		{ 148, 392, 37 },
		{ 148, 393, 37 },
		{ 148, 394, 37 },
		{ 148, 395, 37 },
		{ 148, 396, 37 },
		{ 148, 397, 37 },
		{ 148, 398, 37 },
		{ 148, 399, 37 },
		{ 148, 400, 37 },
		{ 148, 401, 37 },
		{ 148, 402, 37 },
		{ 149, 366, 37 },
		{ 149, 367, 37 },
		{ 149, 368, 37 },
		{ 149, 369, 37 },
		{ 149, 370, 37 },
		{ 149, 371, 37 },
		{ 149, 372, 37 },
		{ 149, 373, 37 },
		{ 149, 374, 37 },
		{ 149, 375, 37 },
		{ 149, 376, 37 },
		{ 149, 377, 37 },
		{ 149, 378, 37 },
		{ 149, 379, 37 },
		{ 149, 380, 37 },
		{ 149, 381, 37 },
		{ 149, 382, 37 },
		{ 149, 383, 37 },
		{ 149, 384, 37 },
		{ 149, 385, 37 },
		{ 149, 386, 37 },
		{ 149, 387, 37 },
		{ 149, 388, 37 },
		{ 149, 389, 37 },
		{ 149, 390, 37 },
		{ 149, 391, 37 },
		{ 149, 392, 37 },
		{ 149, 393, 37 },
		{ 149, 394, 37 },
		{ 149, 395, 37 },
		{ 149, 396, 37 },
		{ 149, 397, 37 },
		{ 149, 398, 37 },
		{ 149, 399, 37 },
		{ 149, 400, 37 },
		{ 150, 365, 37 },
		{ 150, 366, 37 },
		{ 150, 367, 37 },
		{ 150, 368, 37 },
		{ 150, 369, 37 },
		{ 150, 370, 37 },
		{ 150, 371, 37 },
		{ 150, 372, 37 },
		{ 150, 373, 37 },
		{ 150, 374, 37 },
		{ 150, 375, 37 },
		{ 150, 376, 37 },
		{ 150, 377, 37 },
		{ 150, 378, 37 },
		{ 150, 379, 37 },
		{ 150, 380, 37 },
		{ 150, 381, 37 },
		{ 150, 382, 37 },
		{ 150, 383, 37 },
		{ 150, 384, 37 },
		{ 150, 385, 37 },
		{ 150, 386, 37 },
		{ 150, 387, 37 },
		{ 150, 388, 37 },
		{ 150, 389, 37 },
		{ 150, 390, 37 },
		{ 150, 391, 37 },
		{ 150, 392, 37 },
		{ 150, 393, 37 },
		{ 150, 394, 37 },
		{ 150, 395, 37 },
		{ 150, 396, 37 },
		{ 150, 397, 37 },
		{ 150, 398, 37 },
		{ 150, 399, 37 },
		{ 151, 364, 37 },
		{ 151, 365, 37 },
		{ 151, 366, 37 },
		{ 151, 367, 37 },
		{ 151, 368, 37 },
		{ 151, 369, 37 },
		{ 151, 370, 37 },
		{ 151, 371, 37 },
		{ 151, 372, 37 },
		{ 151, 373, 37 },
		{ 151, 374, 37 },
		{ 151, 375, 37 },
		{ 151, 376, 37 },
		{ 151, 377, 37 },
		{ 151, 378, 37 },
		{ 151, 379, 37 },
		{ 151, 380, 37 },
		{ 151, 381, 37 },
		{ 151, 382, 37 },
		{ 151, 383, 37 },
		{ 151, 384, 37 },
		{ 151, 385, 37 },
		{ 151, 386, 37 },
		{ 151, 387, 37 },
		{ 151, 388, 37 },
		{ 151, 389, 37 },
		{ 151, 390, 37 },
		{ 151, 391, 37 },
		{ 151, 392, 37 },
		{ 151, 393, 37 },
		{ 151, 394, 37 },
		{ 151, 395, 37 },
		{ 151, 396, 37 },
		{ 151, 397, 37 },
		{ 151, 398, 37 },
		{ 152, 363, 37 },
		{ 152, 364, 37 },
		{ 152, 365, 37 },
		{ 152, 366, 37 },
		{ 152, 367, 37 },
		{ 152, 368, 37 },
		{ 152, 369, 37 },
		{ 152, 370, 37 },
		{ 152, 371, 37 },
		{ 152, 372, 37 },
		{ 152, 373, 37 },
		{ 152, 374, 37 },
		{ 152, 375, 37 },
		{ 152, 376, 37 },
		{ 152, 377, 37 },
		{ 152, 378, 37 },
		{ 152, 379, 37 },
		{ 152, 380, 37 },
		{ 152, 381, 37 },
		{ 152, 382, 37 },
		{ 152, 383, 37 },
		{ 152, 384, 37 },
		{ 152, 385, 37 },
		{ 152, 386, 37 },
		{ 152, 387, 37 },
		{ 152, 388, 37 },
		{ 152, 389, 37 },
		{ 152, 390, 37 },
		{ 152, 391, 37 },
		{ 152, 392, 37 },
		{ 152, 393, 37 },
		{ 152, 394, 37 },
		{ 152, 395, 37 },
		{ 152, 396, 37 },
		{ 153, 362, 37 },
		{ 153, 363, 37 },
		{ 153, 364, 37 },
		{ 153, 365, 37 },
		{ 153, 366, 37 },
		{ 153, 367, 37 },
		{ 153, 368, 37 },
		{ 153, 369, 37 },
		{ 153, 370, 37 },
		{ 153, 371, 37 },
		{ 153, 372, 37 },
		{ 153, 373, 37 },
		{ 153, 374, 37 },
		{ 153, 375, 37 },
		{ 153, 376, 37 },
		{ 153, 377, 37 },
		{ 153, 378, 37 },
		{ 153, 379, 37 },
		{ 153, 380, 37 },
		{ 153, 381, 37 },
		{ 153, 382, 37 },
		{ 153, 383, 37 },
		{ 153, 384, 37 },
		{ 153, 385, 37 },
		{ 153, 386, 37 },
		{ 153, 387, 37 },
		{ 153, 388, 37 },
		{ 153, 389, 37 },
		{ 153, 390, 37 },
		{ 153, 391, 37 },
		{ 153, 392, 37 },
		{ 153, 393, 37 },
		{ 153, 394, 37 },
		{ 153, 395, 37 },
		{ 154, 361, 37 },
		{ 154, 362, 37 },
		{ 154, 363, 37 },
		{ 154, 364, 37 },
		{ 154, 365, 37 },
		{ 154, 366, 37 },
		{ 154, 367, 37 },
		{ 154, 368, 37 },
		{ 154, 369, 37 },
		{ 154, 370, 37 },
		{ 154, 371, 37 },
		{ 154, 372, 37 },
		{ 154, 373, 37 },
		{ 154, 374, 37 },
		{ 154, 375, 37 },
		{ 154, 376, 37 },
		{ 154, 377, 37 },
		{ 154, 378, 37 },
		{ 154, 379, 37 },
		{ 154, 380, 37 },
		{ 154, 381, 37 },
		{ 154, 382, 37 },
		{ 154, 383, 37 },
		{ 154, 384, 37 },
		{ 154, 385, 37 },
		{ 154, 386, 37 },
		{ 154, 387, 37 },
		{ 154, 388, 37 },
		{ 154, 389, 37 },
		{ 154, 390, 37 },
		{ 154, 391, 37 },
		{ 154, 392, 37 },
		{ 154, 393, 37 },
		{ 154, 394, 37 },
		{ 155, 360, 37 },
		{ 155, 361, 37 },
		{ 155, 362, 37 },
		{ 155, 363, 37 },
		{ 155, 364, 37 },
		{ 155, 365, 37 },
		{ 155, 366, 37 },
		{ 155, 367, 37 },
		{ 155, 368, 37 },
		{ 155, 369, 37 },
		{ 155, 370, 37 },
		{ 155, 371, 37 },
		{ 155, 372, 37 },
		{ 155, 373, 37 },
		{ 155, 374, 37 },
		{ 155, 375, 37 },
		{ 155, 376, 37 },
		{ 155, 377, 37 },
		{ 155, 378, 37 },
		{ 155, 379, 37 },
		{ 155, 380, 37 },
		{ 155, 381, 37 },
		{ 155, 382, 37 },
		{ 155, 383, 37 },
		{ 155, 384, 37 },
		{ 155, 385, 37 },
		{ 155, 386, 37 },
		{ 155, 387, 37 },
		{ 155, 388, 37 },
		{ 155, 389, 37 },
		{ 155, 390, 37 },
		{ 155, 391, 37 },
		{ 155, 392, 37 },
		{ 156, 359, 37 },
		{ 156, 360, 37 },
		{ 156, 361, 37 },
		{ 156, 362, 37 },
		{ 156, 363, 37 },
		{ 156, 364, 37 },
		{ 156, 365, 37 },
		{ 156, 366, 37 },
		{ 156, 367, 37 },
		{ 156, 368, 37 },
		{ 156, 369, 37 },
		{ 156, 370, 37 },
		{ 156, 371, 37 },
		{ 156, 372, 37 },
		{ 156, 373, 37 },
		{ 156, 374, 37 },
		{ 156, 375, 37 },
		{ 156, 376, 37 },
		{ 156, 377, 37 },
		{ 156, 378, 37 },
		{ 156, 379, 37 },
		{ 156, 380, 37 },
		{ 156, 381, 37 },
		{ 156, 382, 37 },
		{ 156, 383, 37 },
		{ 156, 384, 37 },
		{ 156, 385, 37 },
		{ 156, 386, 37 },
		{ 156, 387, 37 },
		{ 156, 388, 37 },
		{ 156, 389, 37 },
		{ 156, 390, 37 },
		{ 156, 391, 37 },
		{ 157, 358, 37 },
		{ 157, 359, 37 },
		{ 157, 360, 37 },
		{ 157, 361, 37 },
		{ 157, 362, 37 },
		{ 157, 363, 37 },
		{ 157, 364, 37 },
		{ 157, 365, 37 },
		{ 157, 366, 37 },
		{ 157, 367, 37 },
		{ 157, 368, 37 },
		{ 157, 369, 37 },
		{ 157, 370, 37 },
		{ 157, 371, 37 },
		{ 157, 372, 37 },
		{ 157, 373, 37 },
		{ 157, 374, 37 },
		{ 157, 375, 37 },
		{ 157, 376, 37 },
		{ 157, 377, 37 },
		{ 157, 378, 37 },
		{ 157, 379, 37 },
		{ 157, 380, 37 },
		{ 157, 381, 37 },
		{ 157, 382, 37 },
		{ 157, 383, 37 },
		{ 157, 384, 37 },
		{ 157, 385, 37 },
		{ 157, 386, 37 },
		{ 157, 387, 37 },
		{ 157, 388, 37 },
		{ 157, 389, 37 },
		{ 157, 390, 37 },
		{ 158, 357, 37 },
		{ 158, 358, 37 },
		{ 158, 359, 37 },
		{ 158, 360, 37 },
		{ 158, 361, 37 },
		{ 158, 362, 37 },
		{ 158, 363, 37 },
		{ 158, 364, 37 },
		{ 158, 365, 37 },
		{ 158, 366, 37 },
		{ 158, 367, 37 },
		{ 158, 368, 37 },
		{ 158, 369, 37 },
		{ 158, 370, 37 },
		{ 158, 371, 37 },
		{ 158, 372, 37 },
		{ 158, 373, 37 },
		{ 158, 374, 37 },
		{ 158, 375, 37 },
		{ 158, 376, 37 },
		{ 158, 377, 37 },
		{ 158, 378, 37 },
		{ 158, 379, 37 },
		{ 158, 380, 37 },
		{ 158, 381, 37 },
		{ 158, 382, 37 },
		{ 158, 383, 37 },
		{ 158, 384, 37 },
		{ 158, 385, 37 },
		{ 158, 386, 37 },
		{ 158, 387, 37 },
		{ 158, 388, 37 },
		{ 159, 356, 37 },
		{ 159, 357, 37 },
		{ 159, 358, 37 },
		{ 159, 359, 37 },
		{ 159, 360, 37 },
		{ 159, 361, 37 },
		{ 159, 362, 37 },
		{ 159, 363, 37 },
		{ 159, 364, 37 },
		{ 159, 365, 37 },
		{ 159, 366, 37 },
		{ 159, 367, 37 },
		{ 159, 368, 37 },
		{ 159, 369, 37 },
		{ 159, 370, 37 },
		{ 159, 371, 37 },
		{ 159, 372, 37 },
		{ 159, 373, 37 },
		{ 159, 374, 37 },
		{ 159, 375, 37 },
		{ 159, 376, 37 },
		{ 159, 377, 37 },
		{ 159, 378, 37 },
		{ 159, 379, 37 },
		{ 159, 380, 37 },
		{ 159, 381, 37 },
		{ 159, 382, 37 },
		{ 159, 383, 37 },
		{ 159, 384, 37 },
		{ 159, 385, 37 },
		{ 159, 386, 37 },
		{ 159, 387, 37 },
		{ 160, 355, 37 },
		{ 160, 356, 37 },
		{ 160, 357, 37 },
		{ 160, 358, 37 },
		{ 160, 359, 37 },
		{ 160, 360, 37 },
		{ 160, 361, 37 },
		{ 160, 362, 37 },
		{ 160, 363, 37 },
		{ 160, 364, 37 },
		{ 160, 365, 37 },
		{ 160, 366, 37 },
		{ 160, 367, 37 },
		{ 160, 368, 37 },
		{ 160, 369, 37 },
		{ 160, 370, 37 },
		{ 160, 371, 37 },
		{ 160, 372, 37 },
		{ 160, 373, 37 },
		{ 160, 374, 37 },
		{ 160, 375, 37 },
		{ 160, 376, 37 },
		{ 160, 377, 37 },
		{ 160, 378, 37 },
		{ 160, 379, 37 },
		{ 160, 380, 37 },
		{ 160, 381, 37 },
		{ 160, 382, 37 },
		{ 160, 383, 37 },
		{ 160, 384, 37 },
		{ 160, 385, 37 },
		{ 160, 386, 37 },
		{ 161, 354, 37 },
		{ 161, 355, 37 },
		{ 161, 356, 37 },
		{ 161, 357, 37 },
		{ 161, 358, 37 },
		{ 161, 359, 37 },
		{ 161, 360, 37 },
		{ 161, 361, 37 },
		{ 161, 362, 37 },
		{ 161, 363, 37 },
		{ 161, 364, 37 },
		{ 161, 365, 37 },
		{ 161, 366, 37 },
		{ 161, 367, 37 },
		{ 161, 368, 37 },
		{ 161, 369, 37 },
		{ 161, 370, 37 },
		{ 161, 371, 37 },
		{ 161, 372, 37 },
		{ 161, 373, 37 },
		{ 161, 374, 37 },
		{ 161, 375, 37 },
		{ 161, 376, 37 },
		{ 161, 377, 37 },
		{ 161, 378, 37 },
		{ 161, 379, 37 },
		{ 161, 380, 37 },
		{ 161, 381, 37 },
		{ 161, 382, 37 },
		{ 161, 383, 37 },
		{ 161, 384, 37 },
		{ 162, 353, 37 },
		{ 162, 354, 37 },
		{ 162, 355, 37 },
		{ 162, 356, 37 },
		{ 162, 357, 37 },
		{ 162, 358, 37 },
		{ 162, 359, 37 },
		{ 162, 360, 37 },
		{ 162, 361, 37 },
		{ 162, 362, 37 },
		{ 162, 363, 37 },
		{ 162, 364, 37 },
		{ 162, 365, 37 },
		{ 162, 366, 37 },
		{ 162, 367, 37 },
		{ 162, 368, 37 },
		{ 162, 369, 37 },
		{ 162, 370, 37 },
		{ 162, 371, 37 },
		{ 162, 372, 37 },
		{ 162, 373, 37 },
		{ 162, 374, 37 },
		{ 162, 375, 37 },
		{ 162, 376, 37 },
		{ 162, 377, 37 },
		{ 162, 378, 37 },
		{ 162, 379, 37 },
		{ 162, 380, 37 },
		{ 162, 381, 37 },
		{ 162, 382, 37 },
		{ 162, 383, 37 },
		{ 163, 352, 37 },
		{ 163, 353, 37 },
		{ 163, 354, 37 },
		{ 163, 355, 37 },
		{ 163, 356, 37 },
		{ 163, 357, 37 },
		{ 163, 358, 37 },
		{ 163, 359, 37 },
		{ 163, 360, 37 },
		{ 163, 361, 37 },
		{ 163, 362, 37 },
		{ 163, 363, 37 },
		{ 163, 364, 37 },
		{ 163, 365, 37 },
		{ 163, 366, 37 },
		{ 163, 367, 37 },
		{ 163, 368, 37 },
		{ 163, 369, 37 },
		{ 163, 370, 37 },
		{ 163, 371, 37 },
		{ 163, 372, 37 },
		{ 163, 373, 37 },
		{ 163, 374, 37 },
		{ 163, 375, 37 },
		{ 163, 376, 37 },
		{ 163, 377, 37 },
		{ 163, 378, 37 },
		{ 163, 379, 37 },
		{ 163, 380, 37 },
		{ 163, 381, 37 },
		{ 163, 382, 37 },
		{ 164, 351, 37 },
		{ 164, 352, 37 },
		{ 164, 353, 37 },
		{ 164, 354, 37 },
		{ 164, 355, 37 },
		{ 164, 356, 37 },
		{ 164, 357, 37 },
		{ 164, 358, 37 },
		{ 164, 359, 37 },
		{ 164, 360, 37 },
		{ 164, 361, 37 },
		{ 164, 362, 37 },
		{ 164, 363, 37 },
		{ 164, 364, 37 },
		{ 164, 365, 37 },
		{ 164, 366, 37 },
		{ 164, 367, 37 },
		{ 164, 368, 37 },
		{ 164, 369, 37 },
		{ 164, 370, 37 },
		{ 164, 371, 37 },
		{ 164, 372, 37 },
		{ 164, 373, 37 },
		{ 164, 374, 37 },
		{ 164, 375, 37 },
		{ 164, 376, 37 },
		{ 164, 377, 37 },
		{ 164, 378, 37 },
		{ 164, 379, 37 },
		{ 164, 380, 37 },
		{ 165, 350, 37 },
		{ 165, 351, 37 },
		{ 165, 352, 37 },
		{ 165, 353, 37 },
		{ 165, 354, 37 },
		{ 165, 355, 37 },
		{ 165, 356, 37 },
		{ 165, 357, 37 },
		{ 165, 358, 37 },
		{ 165, 359, 37 },
		{ 165, 360, 37 },
		{ 165, 361, 37 },
		{ 165, 362, 37 },
		{ 165, 363, 37 },
		{ 165, 364, 37 },
		{ 165, 365, 37 },
		{ 165, 366, 37 },
		{ 165, 367, 37 },
		{ 165, 368, 37 },
		{ 165, 369, 37 },
		{ 165, 370, 37 },
		{ 165, 371, 37 },
		{ 165, 372, 37 },
		{ 165, 373, 37 },
		{ 165, 374, 37 },
		{ 165, 375, 37 },
		{ 165, 376, 37 },
		{ 165, 377, 37 },
		{ 165, 378, 37 },
		{ 165, 379, 37 },
		{ 166, 349, 37 },
		{ 166, 350, 37 },
		{ 166, 351, 37 },
		{ 166, 352, 37 },
		{ 166, 353, 37 },
		{ 166, 354, 37 },
		{ 166, 355, 37 },
		{ 166, 356, 37 },
		{ 166, 357, 37 },
		{ 166, 358, 37 },
		{ 166, 359, 37 },
		{ 166, 360, 37 },
		{ 166, 361, 37 },
		{ 166, 362, 37 },
		{ 166, 363, 37 },
		{ 166, 364, 37 },
		{ 166, 365, 37 },
		{ 166, 366, 37 },
		{ 166, 367, 37 },
		{ 166, 368, 37 },
		{ 166, 369, 37 },
		{ 166, 370, 37 },
		{ 166, 371, 37 },
		{ 166, 372, 37 },
		{ 166, 373, 37 },
		{ 166, 374, 37 },
		{ 166, 375, 37 },
		{ 166, 376, 37 },
		{ 166, 377, 37 },
		{ 166, 378, 37 },
		{ 167, 348, 37 },
		{ 167, 349, 37 },
		{ 167, 350, 37 },
		{ 167, 351, 37 },
		{ 167, 352, 37 },
		{ 167, 353, 37 },
		{ 167, 354, 37 },
		{ 167, 355, 37 },
		{ 167, 356, 37 },
		{ 167, 357, 37 },
		{ 167, 358, 37 },
		{ 167, 359, 37 },
		{ 167, 360, 37 },
		{ 167, 361, 37 },
		{ 167, 362, 37 },
		{ 167, 363, 37 },
		{ 167, 364, 37 },
		{ 167, 365, 37 },
		{ 167, 366, 37 },
		{ 167, 367, 37 },
		{ 167, 368, 37 },
		{ 167, 369, 37 },
		{ 167, 370, 37 },
		{ 167, 371, 37 },
		{ 167, 372, 37 },
		{ 167, 373, 37 },
		{ 167, 374, 37 },
		{ 167, 375, 37 },
		{ 167, 376, 37 },
		{ 168, 347, 37 },
		{ 168, 348, 37 },
		{ 168, 349, 37 },
		{ 168, 350, 37 },
		{ 168, 351, 37 },
		{ 168, 352, 37 },
		{ 168, 353, 37 },
		{ 168, 354, 37 },
		{ 168, 355, 37 },
		{ 168, 356, 37 },
		{ 168, 357, 37 },
		{ 168, 358, 37 },
		{ 168, 359, 37 },
		{ 168, 360, 37 },
		{ 168, 361, 37 },
		{ 168, 362, 37 },
		{ 168, 363, 37 },
		{ 168, 364, 37 },
		{ 168, 365, 37 },
		{ 168, 366, 37 },
		{ 168, 367, 37 },
		{ 168, 368, 37 },
		{ 168, 369, 37 },
		{ 168, 370, 37 },
		{ 168, 371, 37 },
		{ 168, 372, 37 },
		{ 168, 373, 37 },
		{ 168, 374, 37 },
		{ 168, 375, 37 },
		{ 169, 346, 37 },
		{ 169, 347, 37 },
		{ 169, 348, 37 },
		{ 169, 349, 37 },
		{ 169, 350, 37 },
		{ 169, 351, 37 },
		{ 169, 352, 37 },
		{ 169, 353, 37 },
		{ 169, 354, 37 },
		{ 169, 355, 37 },
		{ 169, 356, 37 },
		{ 169, 357, 37 },
		{ 169, 358, 37 },
		{ 169, 359, 37 },
		{ 169, 360, 37 },
		{ 169, 361, 37 },
		{ 169, 362, 37 },
		{ 169, 363, 37 },
		{ 169, 364, 37 },
		{ 169, 365, 37 },
		{ 169, 366, 37 },
		{ 169, 367, 37 },
		{ 169, 368, 37 },
		{ 169, 369, 37 },
		{ 169, 370, 37 },
		{ 169, 371, 37 },
		{ 169, 372, 37 },
		{ 169, 373, 37 },
		{ 169, 374, 37 },
		{ 170, 345, 37 },
		{ 170, 346, 37 },
		{ 170, 347, 37 },
		{ 170, 348, 37 },
		{ 170, 349, 37 },
		{ 170, 350, 37 },
		{ 170, 351, 37 },
		{ 170, 352, 37 },
		{ 170, 353, 37 },
		{ 170, 354, 37 },
		{ 170, 355, 37 },
		{ 170, 356, 37 },
		{ 170, 357, 37 },
		{ 170, 358, 37 },
		{ 170, 359, 37 },
		{ 170, 360, 37 },
		{ 170, 361, 37 },
		{ 170, 362, 37 },
		{ 170, 363, 37 },
		{ 170, 364, 37 },
		{ 170, 365, 37 },
		{ 170, 366, 37 },
		{ 170, 367, 37 },
		{ 170, 368, 37 },
		{ 170, 369, 37 },
		{ 170, 370, 37 },
		{ 170, 371, 37 },
		{ 170, 372, 37 },
		{ 171, 344, 37 },
		{ 171, 345, 37 },
		{ 171, 346, 37 },
		{ 171, 347, 37 },
		{ 171, 348, 37 },
		{ 171, 349, 37 },
		{ 171, 350, 37 },
		{ 171, 351, 37 },
		{ 171, 352, 37 },
		{ 171, 353, 37 },
		{ 171, 354, 37 },
		{ 171, 355, 37 },
		{ 171, 356, 37 },
		{ 171, 357, 37 },
		{ 171, 358, 37 },
		{ 171, 359, 37 },
		{ 171, 360, 37 },
		{ 171, 361, 37 },
		{ 171, 362, 37 },
		{ 171, 363, 37 },
		{ 171, 364, 37 },
		{ 171, 365, 37 },
		{ 171, 366, 37 },
		{ 171, 367, 37 },
		{ 171, 368, 37 },
		{ 171, 369, 37 },
		{ 171, 370, 37 },
		{ 171, 371, 37 },
		{ 172, 343, 37 },
		{ 172, 344, 37 },
		{ 172, 345, 37 },
		{ 172, 346, 37 },
		{ 172, 347, 37 },
		{ 172, 348, 37 },
		{ 172, 349, 37 },
		{ 172, 350, 37 },
		{ 172, 351, 37 },
		{ 172, 352, 37 },
		{ 172, 353, 37 },
		{ 172, 354, 37 },
		{ 172, 355, 37 },
		{ 172, 356, 37 },
		{ 172, 357, 37 },
		{ 172, 358, 37 },
		{ 172, 359, 37 },
		{ 172, 360, 37 },
		{ 172, 361, 37 },
		{ 172, 362, 37 },
		{ 172, 363, 37 },
		{ 172, 364, 37 },
		{ 172, 365, 37 },
		{ 172, 366, 37 },
		{ 172, 367, 37 },
		{ 172, 368, 37 },
		{ 172, 369, 37 },
		{ 172, 370, 37 },
		{ 173, 342, 37 },
		{ 173, 343, 37 },
		{ 173, 344, 37 },
		{ 173, 345, 37 },
		{ 173, 346, 37 },
		{ 173, 347, 37 },
		{ 173, 348, 37 },
		{ 173, 349, 37 },
		{ 173, 350, 37 },
		{ 173, 351, 37 },
		{ 173, 352, 37 },
		{ 173, 353, 37 },
		{ 173, 354, 37 },
		{ 173, 355, 37 },
		{ 173, 356, 37 },
		{ 173, 357, 37 },
		{ 173, 358, 37 },
		{ 173, 359, 37 },
		{ 173, 360, 37 },
		{ 173, 361, 37 },
		{ 173, 362, 37 },
		{ 173, 363, 37 },
		{ 173, 364, 37 },
		{ 173, 365, 37 },
		{ 173, 366, 37 },
		{ 173, 367, 37 },
		{ 173, 368, 37 },
		{ 174, 341, 37 },
		{ 174, 342, 37 },
		{ 174, 343, 37 },
		{ 174, 344, 37 },
		{ 174, 345, 37 },
		{ 174, 346, 37 },
		{ 174, 347, 37 },
		{ 174, 348, 37 },
		{ 174, 349, 37 },
		{ 174, 350, 37 },
		{ 174, 351, 37 },
		{ 174, 352, 37 },
		{ 174, 353, 37 },
		{ 174, 354, 37 },
		{ 174, 355, 37 },
		{ 174, 356, 37 },
		{ 174, 357, 37 },
		{ 174, 358, 37 },
		{ 174, 359, 37 },
		{ 174, 360, 37 },
		{ 174, 361, 37 },
		{ 174, 362, 37 },
		{ 174, 363, 37 },
		{ 174, 364, 37 },
		{ 174, 365, 37 },
		{ 174, 366, 37 },
		{ 174, 367, 37 },
		{ 175, 340, 37 },
		{ 175, 341, 37 },
		{ 175, 342, 37 },
		{ 175, 343, 37 },
		{ 175, 344, 37 },
		{ 175, 345, 37 },
		{ 175, 346, 37 },
		{ 175, 347, 37 },
		{ 175, 348, 37 },
		{ 175, 349, 37 },
		{ 175, 350, 37 },
		{ 175, 351, 37 },
		{ 175, 352, 37 },
		{ 175, 353, 37 },
		{ 175, 354, 37 },
		{ 175, 355, 37 },
		{ 175, 356, 37 },
		{ 175, 357, 37 },
		{ 175, 358, 37 },
		{ 175, 359, 37 },
		{ 175, 360, 37 },
		{ 175, 361, 37 },
		{ 175, 362, 37 },
		{ 175, 363, 37 },
		{ 175, 364, 37 },
		{ 175, 365, 37 },
		{ 175, 366, 37 },
		{ 176, 339, 37 },
		{ 176, 340, 37 },
		{ 176, 341, 37 },
		{ 176, 342, 37 },
		{ 176, 343, 37 },
		{ 176, 344, 37 },
		{ 176, 345, 37 },
		{ 176, 346, 37 },
		{ 176, 347, 37 },
		{ 176, 348, 37 },
		{ 176, 349, 37 },
		{ 176, 350, 37 },
		{ 176, 351, 37 },
		{ 176, 352, 37 },
		{ 176, 353, 37 },
		{ 176, 354, 37 },
		{ 176, 355, 37 },
		{ 176, 356, 37 },
		{ 176, 357, 37 },
		{ 176, 358, 37 },
		{ 176, 359, 37 },
		{ 176, 360, 37 },
		{ 176, 361, 37 },
		{ 176, 362, 37 },
		{ 176, 363, 37 },
		{ 176, 364, 37 },
		{ 177, 338, 37 },
		{ 177, 339, 37 },
		{ 177, 340, 37 },
		{ 177, 341, 37 },
		{ 177, 342, 37 },
		{ 177, 343, 37 },
		{ 177, 344, 37 },
		{ 177, 345, 37 },
		{ 177, 346, 37 },
		{ 177, 347, 37 },
		{ 177, 348, 37 },
		{ 177, 349, 37 },
		{ 177, 350, 37 },
		{ 177, 351, 37 },
		{ 177, 352, 37 },
		{ 177, 353, 37 },
		{ 177, 354, 37 },
		{ 177, 355, 37 },
		{ 177, 356, 37 },
		{ 177, 357, 37 },
		{ 177, 358, 37 },
		{ 177, 359, 37 },
		{ 177, 360, 37 },
		{ 177, 361, 37 },
		{ 177, 362, 37 },
		{ 177, 363, 37 },
		{ 178, 337, 37 },
		{ 178, 338, 37 },
		{ 178, 339, 37 },
		{ 178, 340, 37 },
		{ 178, 341, 37 },
		{ 178, 342, 37 },
		{ 178, 343, 37 },
		{ 178, 344, 37 },
		{ 178, 345, 37 },
		{ 178, 346, 37 },
		{ 178, 347, 37 },
		{ 178, 348, 37 },
		{ 178, 349, 37 },
		{ 178, 350, 37 },
		{ 178, 351, 37 },
		{ 178, 352, 37 },
		{ 178, 353, 37 },
		{ 178, 354, 37 },
		{ 178, 355, 37 },
		{ 178, 356, 37 },
		{ 178, 357, 37 },
		{ 178, 358, 37 },
		{ 178, 359, 37 },
		{ 178, 360, 37 },
		{ 178, 361, 37 },
		{ 178, 362, 37 },
		{ 179, 336, 37 },
		{ 179, 337, 37 },
		{ 179, 338, 37 },
		{ 179, 339, 37 },
		{ 179, 340, 37 },
		{ 179, 341, 37 },
		{ 179, 342, 37 },
		{ 179, 343, 37 },
		{ 179, 344, 37 },
		{ 179, 345, 37 },
		{ 179, 346, 37 },
		{ 179, 347, 37 },
		{ 179, 348, 37 },
		{ 179, 349, 37 },
		{ 179, 350, 37 },
		{ 179, 351, 37 },
		{ 179, 352, 37 },
		{ 179, 353, 37 },
		{ 179, 354, 37 },
		{ 179, 355, 37 },
		{ 179, 356, 37 },
		{ 179, 357, 37 },
		{ 179, 358, 37 },
		{ 179, 359, 37 },
		{ 179, 360, 37 },
		{ 180, 335, 37 },
		{ 180, 336, 37 },
		{ 180, 337, 37 },
		{ 180, 338, 37 },
		{ 180, 339, 37 },
		{ 180, 340, 37 },
		{ 180, 341, 37 },
		{ 180, 342, 37 },
		{ 180, 343, 37 },
		{ 180, 344, 37 },
		{ 180, 345, 37 },
		{ 180, 346, 37 },
		{ 180, 347, 37 },
		{ 180, 348, 37 },
		{ 180, 349, 37 },
		{ 180, 350, 37 },
		{ 180, 351, 37 },
		{ 180, 352, 37 },
		{ 180, 353, 37 },
		{ 180, 354, 37 },
		{ 180, 355, 37 },
		{ 180, 356, 37 },
		{ 180, 357, 37 },
		{ 180, 358, 37 },
		{ 180, 359, 37 },
		{ 181, 334, 37 },
		{ 181, 335, 37 },
		{ 181, 336, 37 },
		{ 181, 337, 37 },
		{ 181, 338, 37 },
		{ 181, 339, 37 },
		{ 181, 340, 37 },
		{ 181, 341, 37 },
		{ 181, 342, 37 },
		{ 181, 343, 37 },
		{ 181, 344, 37 },
		{ 181, 345, 37 },
		{ 181, 346, 37 },
		{ 181, 347, 37 },
		{ 181, 348, 37 },
		{ 181, 349, 37 },
		{ 181, 350, 37 },
		{ 181, 351, 37 },
		{ 181, 352, 37 },
		{ 181, 353, 37 },
		{ 181, 354, 37 },
		{ 181, 355, 37 },
		{ 181, 356, 37 },
		{ 181, 357, 37 },
		{ 181, 358, 37 },
		{ 182, 333, 37 },
		{ 182, 334, 37 },
		{ 182, 335, 37 },
		{ 182, 336, 37 },
		{ 182, 337, 37 },
		{ 182, 338, 37 },
		{ 182, 339, 37 },
		{ 182, 340, 37 },
		{ 182, 341, 37 },
		{ 182, 342, 37 },
		{ 182, 343, 37 },
		{ 182, 344, 37 },
		{ 182, 345, 37 },
		{ 182, 346, 37 },
		{ 182, 347, 37 },
		{ 182, 348, 37 },
		{ 182, 349, 37 },
		{ 182, 350, 37 },
		{ 182, 351, 37 },
		{ 182, 352, 37 },
		{ 182, 353, 37 },
		{ 182, 354, 37 },
		{ 182, 355, 37 },
		{ 182, 356, 37 },
		{ 183, 332, 37 },
		{ 183, 333, 37 },
		{ 183, 334, 37 },
		{ 183, 335, 37 },
		{ 183, 336, 37 },
		{ 183, 337, 37 },
		{ 183, 338, 37 },
		{ 183, 339, 37 },
		{ 183, 340, 37 },
		{ 183, 341, 37 },
		{ 183, 342, 37 },
		{ 183, 343, 37 },
		{ 183, 344, 37 },
		{ 183, 345, 37 },
		{ 183, 346, 37 },
		{ 183, 347, 37 },
		{ 183, 348, 37 },
		{ 183, 349, 37 },
		{ 183, 350, 37 },
		{ 183, 351, 37 },
		{ 183, 352, 37 },
		{ 183, 353, 37 },
		{ 183, 354, 37 },
		{ 183, 355, 37 },
		{ 184, 331, 37 },
		{ 184, 332, 37 },
		{ 184, 333, 37 },
		{ 184, 334, 37 },
		{ 184, 335, 37 },
		{ 184, 336, 37 },
		{ 184, 337, 37 },
		{ 184, 338, 37 },
		{ 184, 339, 37 },
		{ 184, 340, 37 },
		{ 184, 341, 37 },
		{ 184, 342, 37 },
		{ 184, 343, 37 },
		{ 184, 344, 37 },
		{ 184, 345, 37 },
		{ 184, 346, 37 },
		{ 184, 347, 37 },
		{ 184, 348, 37 },
		{ 184, 349, 37 },
		{ 184, 350, 37 },
		{ 184, 351, 37 },
		{ 184, 352, 37 },
		{ 184, 353, 37 },
		{ 184, 354, 37 },
		{ 185, 330, 37 },
		{ 185, 331, 37 },
		{ 185, 332, 37 },
		{ 185, 333, 37 },
		{ 185, 334, 37 },
		{ 185, 335, 37 },
		{ 185, 336, 37 },
		{ 185, 337, 37 },
		{ 185, 338, 37 },
		{ 185, 339, 37 },
		{ 185, 340, 37 },
		{ 185, 341, 37 },
		{ 185, 342, 37 },
		{ 185, 343, 37 },
		{ 185, 344, 37 },
		{ 185, 345, 37 },
		{ 185, 346, 37 },
		{ 185, 347, 37 },
		{ 185, 348, 37 },
		{ 185, 349, 37 },
		{ 185, 350, 37 },
		{ 185, 351, 37 },
		{ 185, 352, 37 },
		{ 186, 329, 37 },
		{ 186, 330, 37 },
		{ 186, 331, 37 },
		{ 186, 332, 37 },
		{ 186, 333, 37 },
		{ 186, 334, 37 },
		{ 186, 335, 37 },
		{ 186, 336, 37 },
		{ 186, 337, 37 },
		{ 186, 338, 37 },
		{ 186, 339, 37 },
		{ 186, 340, 37 },
		{ 186, 341, 37 },
		{ 186, 342, 37 },
		{ 186, 343, 37 },
		{ 186, 344, 37 },
		{ 186, 345, 37 },
		{ 186, 346, 37 },
		{ 186, 347, 37 },
		{ 186, 348, 37 },
		{ 186, 349, 37 },
		{ 186, 350, 37 },
		{ 186, 351, 37 },
		{ 187, 328, 37 },
		{ 187, 329, 37 },
		{ 187, 330, 37 },
		{ 187, 331, 37 },
		{ 187, 332, 37 },
		{ 187, 333, 37 },
		{ 187, 334, 37 },
		{ 187, 335, 37 },
		{ 187, 336, 37 },
		{ 187, 337, 37 },
		{ 187, 338, 37 },
		{ 187, 339, 37 },
		{ 187, 340, 37 },
		{ 187, 341, 37 },
		{ 187, 342, 37 },
		{ 187, 343, 37 },
		{ 187, 344, 37 },
		{ 187, 345, 37 },
		{ 187, 346, 37 },
		{ 187, 347, 37 },
		{ 187, 348, 37 },
		{ 187, 349, 37 },
		{ 187, 350, 37 },
		{ 188, 327, 37 },
		{ 188, 328, 37 },
		{ 188, 329, 37 },
		{ 188, 330, 37 },
		{ 188, 331, 37 },
		{ 188, 332, 37 },
		{ 188, 333, 37 },
		{ 188, 334, 37 },
		{ 188, 335, 37 },
		{ 188, 336, 37 },
		{ 188, 337, 37 },
		{ 188, 338, 37 },
		{ 188, 339, 37 },
		{ 188, 340, 37 },
		{ 188, 341, 37 },
		{ 188, 342, 37 },
		{ 188, 343, 37 },
		{ 188, 344, 37 },
		{ 188, 345, 37 },
		{ 188, 346, 37 },
		{ 188, 347, 37 },
		{ 188, 348, 37 },
		{ 189, 326, 37 },
		{ 189, 327, 37 },
		{ 189, 328, 37 },
		{ 189, 329, 37 },
		{ 189, 330, 37 },
		{ 189, 331, 37 },
		{ 189, 332, 37 },
		{ 189, 333, 37 },
		{ 189, 334, 37 },
		{ 189, 335, 37 },
		{ 189, 336, 37 },
		{ 189, 337, 37 },
		{ 189, 338, 37 },
		{ 189, 339, 37 },
		{ 189, 340, 37 },
		{ 189, 341, 37 },
		{ 189, 342, 37 },
		{ 189, 343, 37 },
		{ 189, 344, 37 },
		{ 189, 345, 37 },
		{ 189, 346, 37 },
		{ 189, 347, 37 },
		{ 190, 325, 37 },
		{ 190, 326, 37 },
		{ 190, 327, 37 },
		{ 190, 328, 37 },
		{ 190, 329, 37 },
		{ 190, 330, 37 },
		{ 190, 331, 37 },
		{ 190, 332, 37 },
		{ 190, 333, 37 },
		{ 190, 334, 37 },
		{ 190, 335, 37 },
		{ 190, 336, 37 },
		{ 190, 337, 37 },
		{ 190, 338, 37 },
		{ 190, 339, 37 },
		{ 190, 340, 37 },
		{ 190, 341, 37 },
		{ 190, 342, 37 },
		{ 190, 343, 37 },
		{ 190, 344, 37 },
		{ 190, 345, 37 },
		{ 190, 346, 37 },
		{ 191, 324, 37 },
		{ 191, 325, 37 },
		{ 191, 326, 37 },
		{ 191, 327, 37 },
		{ 191, 328, 37 },
		{ 191, 329, 37 },
		{ 191, 330, 37 },
		{ 191, 331, 37 },
		{ 191, 332, 37 },
		{ 191, 333, 37 },
		{ 191, 334, 37 },
		{ 191, 335, 37 },
		{ 191, 336, 37 },
		{ 191, 337, 37 },
		{ 191, 338, 37 },
		{ 191, 339, 37 },
		{ 191, 340, 37 },
		{ 191, 341, 37 },
		{ 191, 342, 37 },
		{ 191, 343, 37 },
		{ 191, 344, 37 },
		{ 192, 324, 37 },
		{ 192, 325, 37 },
		{ 192, 326, 37 },
		{ 192, 327, 37 },
		{ 192, 328, 37 },
		{ 192, 329, 37 },
		{ 192, 330, 37 },
		{ 192, 331, 37 },
		{ 192, 332, 37 },
		{ 192, 333, 37 },
		{ 192, 334, 37 },
		{ 192, 335, 37 },
		{ 192, 336, 37 },
		{ 192, 337, 37 },
		{ 192, 338, 37 },
		{ 192, 339, 37 },
		{ 192, 340, 37 },
		{ 192, 341, 37 },
		{ 192, 342, 37 },
		{ 192, 343, 37 },
		{ 193, 323, 37 },
		{ 193, 324, 37 },
		{ 193, 325, 37 },
		{ 193, 326, 37 },
		{ 193, 327, 37 },
		{ 193, 328, 37 },
		{ 193, 329, 37 },
		{ 193, 330, 37 },
		{ 193, 331, 37 },
		{ 193, 332, 37 },
		{ 193, 333, 37 },
		{ 193, 334, 37 },
		{ 193, 335, 37 },
		{ 193, 336, 37 },
		{ 193, 337, 37 },
		{ 193, 338, 37 },
		{ 193, 339, 37 },
		{ 193, 340, 37 },
		{ 193, 341, 37 },
		{ 193, 342, 37 },
		{ 194, 323, 37 },
		{ 194, 324, 37 },
		{ 194, 325, 37 },
		{ 194, 326, 37 },
		{ 194, 327, 37 },
		{ 194, 328, 37 },
		{ 194, 329, 37 },
		{ 194, 330, 37 },
		{ 194, 331, 37 },
		{ 194, 332, 37 },
		{ 194, 333, 37 },
		{ 194, 334, 37 },
		{ 194, 335, 37 },
		{ 194, 336, 37 },
		{ 194, 337, 37 },
		{ 194, 338, 37 },
		{ 194, 339, 37 },
		{ 194, 340, 37 },
		{ 195, 323, 37 },
		{ 195, 324, 37 },
		{ 195, 325, 37 },
		{ 195, 326, 37 },
		{ 195, 327, 37 },
		{ 195, 328, 37 },
		{ 195, 329, 37 },
		{ 195, 330, 37 },
		{ 195, 331, 37 },
		{ 195, 332, 37 },
		{ 195, 333, 37 },
		{ 195, 334, 37 },
		{ 195, 335, 37 },
		{ 195, 336, 37 },
		{ 195, 337, 37 },
		{ 195, 338, 37 },
		{ 195, 339, 37 },
		{ 196, 323, 37 },
		{ 196, 324, 37 },
		{ 196, 325, 37 },
		{ 196, 326, 37 },
		{ 196, 327, 37 },
		{ 196, 328, 37 },
		{ 196, 329, 37 },
		{ 196, 330, 37 },
		{ 196, 331, 37 },
		{ 196, 332, 37 },
		{ 196, 333, 37 },
		{ 196, 334, 37 },
		{ 196, 335, 37 },
		{ 196, 336, 37 },
		{ 196, 337, 37 },
		{ 196, 338, 37 },
		{ 197, 323, 37 },
		{ 197, 324, 37 },
		{ 197, 325, 37 },
		{ 197, 326, 37 },
		{ 197, 327, 37 },
		{ 197, 328, 37 },
		{ 197, 329, 37 },
		{ 197, 330, 37 },
		{ 197, 331, 37 },
		{ 197, 332, 37 },
		{ 197, 333, 37 },
		{ 197, 334, 37 },
		{ 197, 335, 37 },
		{ 197, 336, 37 },
		{ 198, 323, 37 },
		{ 198, 324, 37 },
		{ 198, 325, 37 },
		{ 198, 326, 37 },
		{ 198, 327, 37 },
		{ 198, 328, 37 },
		{ 198, 329, 37 },
		{ 198, 330, 37 },
		{ 198, 331, 37 },
		{ 198, 332, 37 },
		{ 198, 333, 37 },
		{ 198, 334, 37 },
		{ 198, 335, 37 },
		{ 199, 323, 37 },
		{ 199, 324, 37 },
		{ 199, 325, 37 },
		{ 199, 326, 37 },
		{ 199, 327, 37 },
		{ 199, 328, 37 },
		{ 199, 329, 37 },
		{ 199, 330, 37 },
		{ 199, 331, 37 },
		{ 199, 332, 37 },
		{ 199, 333, 37 },
		{ 199, 334, 37 },
		{ 200, 323, 37 },
		{ 200, 324, 37 },
		{ 200, 325, 37 },
		{ 200, 326, 37 },
		{ 200, 327, 37 },
		{ 200, 328, 37 },
		{ 200, 329, 37 },
		{ 200, 330, 37 },
		{ 200, 331, 37 },
		{ 200, 332, 37 },
		{ 201, 323, 37 },
		{ 201, 324, 37 },
		{ 201, 325, 37 },
		{ 201, 326, 37 },
		{ 201, 327, 37 },
		{ 201, 328, 37 },
		{ 201, 329, 37 },
		{ 201, 330, 37 },
		{ 201, 331, 37 },
		{ 202, 323, 37 },
		{ 202, 324, 37 },
		{ 202, 325, 37 },
		{ 202, 326, 37 },
		{ 202, 327, 37 },
		{ 202, 328, 37 },
		{ 202, 329, 37 },
		{ 202, 330, 37 },
		{ 203, 323, 37 },
		{ 203, 324, 37 },
		{ 203, 325, 37 },
		{ 203, 326, 37 },
		{ 203, 327, 37 },
		{ 203, 328, 37 },
		{ 204, 323, 37 },
		{ 204, 324, 37 },
		{ 204, 325, 37 },
		{ 204, 326, 37 },
		{ 204, 327, 37 },
		{ 205, 323, 37 },
		{ 205, 324, 37 },
		{ 205, 325, 37 },
		{ 206, 324, 37 },
		{ 308, 190, 37 },
		{ 309, 189, 37 },
		{ 309, 190, 37 },
		{ 309, 191, 37 },
		{ 310, 187, 37 },
		{ 310, 188, 37 },
		{ 310, 189, 37 },
		{ 310, 190, 37 },
		{ 310, 191, 37 },
		{ 311, 186, 37 },
		{ 311, 187, 37 },
		{ 311, 188, 37 },
		{ 311, 189, 37 },
		{ 311, 190, 37 },
		{ 311, 191, 37 },
		{ 312, 184, 37 },
		{ 312, 185, 37 },
		{ 312, 186, 37 },
		{ 312, 187, 37 },
		{ 312, 188, 37 },
		{ 312, 189, 37 },
		{ 312, 190, 37 },
		{ 312, 191, 37 },
		{ 313, 183, 37 },
		{ 313, 184, 37 },
		{ 313, 185, 37 },
		{ 313, 186, 37 },
		{ 313, 187, 37 },
		{ 313, 188, 37 },
		{ 313, 189, 37 },
		{ 313, 190, 37 },
		{ 313, 191, 37 },
		{ 314, 182, 37 },
		{ 314, 183, 37 },
		{ 314, 184, 37 },
		{ 314, 185, 37 },
		{ 314, 186, 37 },
		{ 314, 187, 37 },
		{ 314, 188, 37 },
		{ 314, 189, 37 },
		{ 314, 190, 37 },
		{ 314, 191, 37 },
		{ 315, 180, 37 },
		{ 315, 181, 37 },
		{ 315, 182, 37 },
		{ 315, 183, 37 },
		{ 315, 184, 37 },
		{ 315, 185, 37 },
		{ 315, 186, 37 },
		{ 315, 187, 37 },
		{ 315, 188, 37 },
		{ 315, 189, 37 },
		{ 315, 190, 37 },
		{ 315, 191, 37 },
		{ 316, 179, 37 },
		{ 316, 180, 37 },
		{ 316, 181, 37 },
		{ 316, 182, 37 },
		{ 316, 183, 37 },
		{ 316, 184, 37 },
		{ 316, 185, 37 },
		{ 316, 186, 37 },
		{ 316, 187, 37 },
		{ 316, 188, 37 },
		{ 316, 189, 37 },
		{ 316, 190, 37 },
		{ 316, 191, 37 },
		{ 317, 178, 37 },
		{ 317, 179, 37 },
		{ 317, 180, 37 },
		{ 317, 181, 37 },
		{ 317, 182, 37 },
		{ 317, 183, 37 },
		{ 317, 184, 37 },
		{ 317, 185, 37 },
		{ 317, 186, 37 },
		{ 317, 187, 37 },
		{ 317, 188, 37 },
		{ 317, 189, 37 },
		{ 317, 190, 37 },
		{ 317, 191, 37 },
		{ 318, 176, 37 },
		{ 318, 177, 37 },
		{ 318, 178, 37 },
		{ 318, 179, 37 },
		{ 318, 180, 37 },
		{ 318, 181, 37 },
		{ 318, 182, 37 },
		{ 318, 183, 37 },
		{ 318, 184, 37 },
		{ 318, 185, 37 },
		{ 318, 186, 37 },
		{ 318, 187, 37 },
		{ 318, 188, 37 },
		{ 318, 189, 37 },
		{ 318, 190, 37 },
		{ 318, 191, 37 },
		{ 319, 175, 37 },
		{ 319, 176, 37 },
		{ 319, 177, 37 },
		{ 319, 178, 37 },
		{ 319, 179, 37 },
		{ 319, 180, 37 },
		{ 319, 181, 37 },
		{ 319, 182, 37 },
		{ 319, 183, 37 },
		{ 319, 184, 37 },
		{ 319, 185, 37 },
		{ 319, 186, 37 },
		{ 319, 187, 37 },
		{ 319, 188, 37 },
		{ 319, 189, 37 },
		{ 319, 190, 37 },
		{ 319, 191, 37 },
		{ 320, 174, 37 },
		{ 320, 175, 37 },
		{ 320, 176, 37 },
		{ 320, 177, 37 },
		{ 320, 178, 37 },
		{ 320, 179, 37 },
		{ 320, 180, 37 },
		{ 320, 181, 37 },
		{ 320, 182, 37 },
		{ 320, 183, 37 },
		{ 320, 184, 37 },
		{ 320, 185, 37 },
		{ 320, 186, 37 },
		{ 320, 187, 37 },
		{ 320, 188, 37 },
		{ 320, 189, 37 },
		{ 320, 190, 37 },
		{ 320, 191, 37 },
		{ 321, 172, 37 },
		{ 321, 173, 37 },
		{ 321, 174, 37 },
		{ 321, 175, 37 },
		{ 321, 176, 37 },
		{ 321, 177, 37 },
		{ 321, 178, 37 },
		{ 321, 179, 37 },
		{ 321, 180, 37 },
		{ 321, 181, 37 },
		{ 321, 182, 37 },
		{ 321, 183, 37 },
		{ 321, 184, 37 },
		{ 321, 185, 37 },
		{ 321, 186, 37 },
		{ 321, 187, 37 },
		{ 321, 188, 37 },
		{ 321, 189, 37 },
		{ 321, 190, 37 },
		{ 321, 191, 37 },
		{ 322, 171, 37 },
		{ 322, 172, 37 },
		{ 322, 173, 37 },
		{ 322, 174, 37 },
		{ 322, 175, 37 },
		{ 322, 176, 37 },
		{ 322, 177, 37 },
		{ 322, 178, 37 },
		{ 322, 179, 37 },
		{ 322, 180, 37 },
		{ 322, 181, 37 },
		{ 322, 182, 37 },
		{ 322, 183, 37 },
		{ 322, 184, 37 },
		{ 322, 185, 37 },
		{ 322, 186, 37 },
		{ 322, 187, 37 },
		{ 322, 188, 37 },
		{ 322, 189, 37 },
		{ 322, 190, 37 },
		{ 323, 170, 37 },
		{ 323, 171, 37 },
		{ 323, 172, 37 },
		{ 323, 173, 37 },
		{ 323, 174, 37 },
		{ 323, 175, 37 },
		{ 323, 176, 37 },
		{ 323, 177, 37 },
		{ 323, 178, 37 },
		{ 323, 179, 37 },
		{ 323, 180, 37 },
		{ 323, 181, 37 },
		{ 323, 182, 37 },
		{ 323, 183, 37 },
		{ 323, 184, 37 },
		{ 323, 185, 37 },
		{ 323, 186, 37 },
		{ 323, 187, 37 },
		{ 323, 188, 37 },
		{ 323, 189, 37 },
		{ 323, 190, 37 },
		{ 324, 168, 37 },
		{ 324, 169, 37 },
		{ 324, 170, 37 },
		{ 324, 171, 37 },
		{ 324, 172, 37 },
		{ 324, 173, 37 },
		{ 324, 174, 37 },
		{ 324, 175, 37 },
		{ 324, 176, 37 },
		{ 324, 177, 37 },
		{ 324, 178, 37 },
		{ 324, 179, 37 },
		{ 324, 180, 37 },
		{ 324, 181, 37 },
		{ 324, 182, 37 },
		{ 324, 183, 37 },
		{ 324, 184, 37 },
		{ 324, 185, 37 },
		{ 324, 186, 37 },
		{ 324, 187, 37 },
		{ 324, 188, 37 },
		{ 324, 189, 37 },
		{ 325, 167, 37 },
		{ 325, 168, 37 },
		{ 325, 169, 37 },
		{ 325, 170, 37 },
		{ 325, 171, 37 },
		{ 325, 172, 37 },
		{ 325, 173, 37 },
		{ 325, 174, 37 },
		{ 325, 175, 37 },
		{ 325, 176, 37 },
		{ 325, 177, 37 },
		{ 325, 178, 37 },
		{ 325, 179, 37 },
		{ 325, 180, 37 },
		{ 325, 181, 37 },
		{ 325, 182, 37 },
		{ 325, 183, 37 },
		{ 325, 184, 37 },
		{ 325, 185, 37 },
		{ 325, 186, 37 },
		{ 325, 187, 37 },
		{ 325, 188, 37 },
		{ 326, 166, 37 },
		{ 326, 167, 37 },
		{ 326, 168, 37 },
		{ 326, 169, 37 },
		{ 326, 170, 37 },
		{ 326, 171, 37 },
		{ 326, 172, 37 },
		{ 326, 173, 37 },
		{ 326, 174, 37 },
		{ 326, 175, 37 },
		{ 326, 176, 37 },
		{ 326, 177, 37 },
		{ 326, 178, 37 },
		{ 326, 179, 37 },
		{ 326, 180, 37 },
		{ 326, 181, 37 },
		{ 326, 182, 37 },
		{ 326, 183, 37 },
		{ 326, 184, 37 },
		{ 326, 185, 37 },
		{ 326, 186, 37 },
		{ 326, 187, 37 },
		{ 327, 164, 37 },
		{ 327, 165, 37 },
		{ 327, 166, 37 },
		{ 327, 167, 37 },
		{ 327, 168, 37 },
		{ 327, 169, 37 },
		{ 327, 170, 37 },
		{ 327, 171, 37 },
		{ 327, 172, 37 },
		{ 327, 173, 37 },
		{ 327, 174, 37 },
		{ 327, 175, 37 },
		{ 327, 176, 37 },
		{ 327, 177, 37 },
		{ 327, 178, 37 },
		{ 327, 179, 37 },
		{ 327, 180, 37 },
		{ 327, 181, 37 },
		{ 327, 182, 37 },
		{ 327, 183, 37 },
		{ 327, 184, 37 },
		{ 327, 185, 37 },
		{ 327, 186, 37 },
		{ 328, 163, 37 },
		{ 328, 164, 37 },
		{ 328, 165, 37 },
		{ 328, 166, 37 },
		{ 328, 167, 37 },
		{ 328, 168, 37 },
		{ 328, 169, 37 },
		{ 328, 170, 37 },
		{ 328, 171, 37 },
		{ 328, 172, 37 },
		{ 328, 173, 37 },
		{ 328, 174, 37 },
		{ 328, 175, 37 },
		{ 328, 176, 37 },
		{ 328, 177, 37 },
		{ 328, 178, 37 },
		{ 328, 179, 37 },
		{ 328, 180, 37 },
		{ 328, 181, 37 },
		{ 328, 182, 37 },
		{ 328, 183, 37 },
		{ 328, 184, 37 },
		{ 328, 185, 37 },
		{ 329, 162, 37 },
		{ 329, 163, 37 },
		{ 329, 164, 37 },
		{ 329, 165, 37 },
		{ 329, 166, 37 },
		{ 329, 167, 37 },
		{ 329, 168, 37 },
		{ 329, 169, 37 },
		{ 329, 170, 37 },
		{ 329, 171, 37 },
		{ 329, 172, 37 },
		{ 329, 173, 37 },
		{ 329, 174, 37 },
		{ 329, 175, 37 },
		{ 329, 176, 37 },
		{ 329, 177, 37 },
		{ 329, 178, 37 },
		{ 329, 179, 37 },
		{ 329, 180, 37 },
		{ 329, 181, 37 },
		{ 329, 182, 37 },
		{ 329, 183, 37 },
		{ 329, 184, 37 },
		{ 330, 160, 37 },
		{ 330, 161, 37 },
		{ 330, 162, 37 },
		{ 330, 163, 37 },
		{ 330, 164, 37 },
		{ 330, 165, 37 },
		{ 330, 166, 37 },
		{ 330, 167, 37 },
		{ 330, 168, 37 },
		{ 330, 169, 37 },
		{ 330, 170, 37 },
		{ 330, 171, 37 },
		{ 330, 172, 37 },
		{ 330, 173, 37 },
		{ 330, 174, 37 },
		{ 330, 175, 37 },
		{ 330, 176, 37 },
		{ 330, 177, 37 },
		{ 330, 178, 37 },
		{ 330, 179, 37 },
		{ 330, 180, 37 },
		{ 330, 181, 37 },
		{ 330, 182, 37 },
		{ 330, 183, 37 },
		{ 331, 159, 37 },
		{ 331, 160, 37 },
		{ 331, 161, 37 },
		{ 331, 162, 37 },
		{ 331, 163, 37 },
		{ 331, 164, 37 },
		{ 331, 165, 37 },
		{ 331, 166, 37 },
		{ 331, 167, 37 },
		{ 331, 168, 37 },
		{ 331, 169, 37 },
		{ 331, 170, 37 },
		{ 331, 171, 37 },
		{ 331, 172, 37 },
		{ 331, 173, 37 },
		{ 331, 174, 37 },
		{ 331, 175, 37 },
		{ 331, 176, 37 },
		{ 331, 177, 37 },
		{ 331, 178, 37 },
		{ 331, 179, 37 },
		{ 331, 180, 37 },
		{ 331, 181, 37 },
		{ 331, 182, 37 },
		{ 332, 158, 37 },
		{ 332, 159, 37 },
		{ 332, 160, 37 },
		{ 332, 161, 37 },
		{ 332, 162, 37 },
		{ 332, 163, 37 },
		{ 332, 164, 37 },
		{ 332, 165, 37 },
		{ 332, 166, 37 },
		{ 332, 167, 37 },
		{ 332, 168, 37 },
		{ 332, 169, 37 },
		{ 332, 170, 37 },
		{ 332, 171, 37 },
		{ 332, 172, 37 },
		{ 332, 173, 37 },
		{ 332, 174, 37 },
		{ 332, 175, 37 },
		{ 332, 176, 37 },
		{ 332, 177, 37 },
		{ 332, 178, 37 },
		{ 332, 179, 37 },
		{ 332, 180, 37 },
		{ 332, 181, 37 },
		{ 333, 156, 37 },
		{ 333, 157, 37 },
		{ 333, 158, 37 },
		{ 333, 159, 37 },
		{ 333, 160, 37 },
		{ 333, 161, 37 },
		{ 333, 162, 37 },
		{ 333, 163, 37 },
		{ 333, 164, 37 },
		{ 333, 165, 37 },
		{ 333, 166, 37 },
		{ 333, 167, 37 },
		{ 333, 168, 37 },
		{ 333, 169, 37 },
		{ 333, 170, 37 },
		{ 333, 171, 37 },
		{ 333, 172, 37 },
		{ 333, 173, 37 },
		{ 333, 174, 37 },
		{ 333, 175, 37 },
		{ 333, 176, 37 },
		{ 333, 177, 37 },
		{ 333, 178, 37 },
		{ 333, 179, 37 },
		{ 333, 180, 37 },
		{ 334, 155, 37 },
		{ 334, 156, 37 },
		{ 334, 157, 37 },
		{ 334, 158, 37 },
		{ 334, 159, 37 },
		{ 334, 160, 37 },
		{ 334, 161, 37 },
		{ 334, 162, 37 },
		{ 334, 163, 37 },
		{ 334, 164, 37 },
		{ 334, 165, 37 },
		{ 334, 166, 37 },
		{ 334, 167, 37 },
		{ 334, 168, 37 },
		{ 334, 169, 37 },
		{ 334, 170, 37 },
		{ 334, 171, 37 },
		{ 334, 172, 37 },
		{ 334, 173, 37 },
		{ 334, 174, 37 },
		{ 334, 175, 37 },
		{ 334, 176, 37 },
		{ 334, 177, 37 },
		{ 334, 178, 37 },
		{ 334, 179, 37 },
		{ 335, 154, 37 },
		{ 335, 155, 37 },
		{ 335, 156, 37 },
		{ 335, 157, 37 },
		{ 335, 158, 37 },
		{ 335, 159, 37 },
		{ 335, 160, 37 },
		{ 335, 161, 37 },
		{ 335, 162, 37 },
		{ 335, 163, 37 },
		{ 335, 164, 37 },
		{ 335, 165, 37 },
		{ 335, 166, 37 },
		{ 335, 167, 37 },
		{ 335, 168, 37 },
		{ 335, 169, 37 },
		{ 335, 170, 37 },
		{ 335, 171, 37 },
		{ 335, 172, 37 },
		{ 335, 173, 37 },
		{ 335, 174, 37 },
		{ 335, 175, 37 },
		{ 335, 176, 37 },
		{ 335, 177, 37 },
		{ 335, 178, 37 },
		{ 336, 152, 37 },
		{ 336, 153, 37 },
		{ 336, 154, 37 },
		{ 336, 155, 37 },
		{ 336, 156, 37 },
		{ 336, 157, 37 },
		{ 336, 158, 37 },
		{ 336, 159, 37 },
		{ 336, 160, 37 },
		{ 336, 161, 37 },
		{ 336, 162, 37 },
		{ 336, 163, 37 },
		{ 336, 164, 37 },
		{ 336, 165, 37 },
		{ 336, 166, 37 },
		{ 336, 167, 37 },
		{ 336, 168, 37 },
		{ 336, 169, 37 },
		{ 336, 170, 37 },
		{ 336, 171, 37 },
		{ 336, 172, 37 },
		{ 336, 173, 37 },
		{ 336, 174, 37 },
		{ 336, 175, 37 },
		{ 336, 176, 37 },
		{ 336, 177, 37 },
		{ 337, 151, 37 },
		{ 337, 152, 37 },
		{ 337, 153, 37 },
		{ 337, 154, 37 },
		{ 337, 155, 37 },
		{ 337, 156, 37 },
		{ 337, 157, 37 },
		{ 337, 158, 37 },
		{ 337, 159, 37 },
		{ 337, 160, 37 },
		{ 337, 161, 37 },
		{ 337, 162, 37 },
		{ 337, 163, 37 },
		{ 337, 164, 37 },
		{ 337, 165, 37 },
		{ 337, 166, 37 },
		{ 337, 167, 37 },
		{ 337, 168, 37 },
		{ 337, 169, 37 },
		{ 337, 170, 37 },
		{ 337, 171, 37 },
		{ 337, 172, 37 },
		{ 337, 173, 37 },
		{ 337, 174, 37 },
		{ 337, 175, 37 },
		{ 337, 176, 37 },
		{ 338, 150, 37 },
		{ 338, 151, 37 },
		{ 338, 152, 37 },
		{ 338, 153, 37 },
		{ 338, 154, 37 },
		{ 338, 155, 37 },
		{ 338, 156, 37 },
		{ 338, 157, 37 },
		{ 338, 158, 37 },
		{ 338, 159, 37 },
		{ 338, 160, 37 },
		{ 338, 161, 37 },
		{ 338, 162, 37 },
		{ 338, 163, 37 },
		{ 338, 164, 37 },
		{ 338, 165, 37 },
		{ 338, 166, 37 },
		{ 338, 167, 37 },
		{ 338, 168, 37 },
		{ 338, 169, 37 },
		{ 338, 170, 37 },
		{ 338, 171, 37 },
		{ 338, 172, 37 },
		{ 338, 173, 37 },
		{ 338, 174, 37 },
		{ 338, 175, 37 },
		{ 339, 148, 37 },
		{ 339, 149, 37 },
		{ 339, 150, 37 },
		{ 339, 151, 37 },
		{ 339, 152, 37 },
		{ 339, 153, 37 },
		{ 339, 154, 37 },
		{ 339, 155, 37 },
		{ 339, 156, 37 },
		{ 339, 157, 37 },
		{ 339, 158, 37 },
		{ 339, 159, 37 },
		{ 339, 160, 37 },
		{ 339, 161, 37 },
		{ 339, 162, 37 },
		{ 339, 163, 37 },
		{ 339, 164, 37 },
		{ 339, 165, 37 },
		{ 339, 166, 37 },
		{ 339, 167, 37 },
		{ 339, 168, 37 },
		{ 339, 169, 37 },
		{ 339, 170, 37 },
		{ 339, 171, 37 },
		{ 339, 172, 37 },
		{ 339, 173, 37 },
		{ 339, 174, 37 },
		{ 340, 147, 37 },
		{ 340, 148, 37 },
		{ 340, 149, 37 },
		{ 340, 150, 37 },
		{ 340, 151, 37 },
		{ 340, 152, 37 },
		{ 340, 153, 37 },
		{ 340, 154, 37 },
		{ 340, 155, 37 },
		{ 340, 156, 37 },
		{ 340, 157, 37 },
		{ 340, 158, 37 },
		{ 340, 159, 37 },
		{ 340, 160, 37 },
		{ 340, 161, 37 },
		{ 340, 162, 37 },
		{ 340, 163, 37 },
		{ 340, 164, 37 },
		{ 340, 165, 37 },
		{ 340, 166, 37 },
		{ 340, 167, 37 },
		{ 340, 168, 37 },
		{ 340, 169, 37 },
		{ 340, 170, 37 },
		{ 340, 171, 37 },
		{ 340, 172, 37 },
		{ 340, 173, 37 },
		{ 341, 146, 37 },
		{ 341, 147, 37 },
		{ 341, 148, 37 },
		{ 341, 149, 37 },
		{ 341, 150, 37 },
		{ 341, 151, 37 },
		{ 341, 152, 37 },
		{ 341, 153, 37 },
		{ 341, 154, 37 },
		{ 341, 155, 37 },
		{ 341, 156, 37 },
		{ 341, 157, 37 },
		{ 341, 158, 37 },
		{ 341, 159, 37 },
		{ 341, 160, 37 },
		{ 341, 161, 37 },
		{ 341, 162, 37 },
		{ 341, 163, 37 },
		{ 341, 164, 37 },
		{ 341, 165, 37 },
		{ 341, 166, 37 },
		{ 341, 167, 37 },
		{ 341, 168, 37 },
		{ 341, 169, 37 },
		{ 341, 170, 37 },
		{ 341, 171, 37 },
		{ 341, 172, 37 },
		{ 342, 144, 37 },
		{ 342, 145, 37 },
		{ 342, 146, 37 },
		{ 342, 147, 37 },
		{ 342, 148, 37 },
		{ 342, 149, 37 },
		{ 342, 150, 37 },
		{ 342, 151, 37 },
		{ 342, 152, 37 },
		{ 342, 153, 37 },
		{ 342, 154, 37 },
		{ 342, 155, 37 },
		{ 342, 156, 37 },
		{ 342, 157, 37 },
		{ 342, 158, 37 },
		{ 342, 159, 37 },
		{ 342, 160, 37 },
		{ 342, 161, 37 },
		{ 342, 162, 37 },
		{ 342, 163, 37 },
		{ 342, 164, 37 },
		{ 342, 165, 37 },
		{ 342, 166, 37 },
		{ 342, 167, 37 },
		{ 342, 168, 37 },
		{ 342, 169, 37 },
		{ 342, 170, 37 },
		{ 342, 171, 37 },
		{ 343, 143, 37 },
		{ 343, 144, 37 },
		{ 343, 145, 37 },
		{ 343, 146, 37 },
		{ 343, 147, 37 },
		{ 343, 148, 37 },
		{ 343, 149, 37 },
		{ 343, 150, 37 },
		{ 343, 151, 37 },
		{ 343, 152, 37 },
		{ 343, 153, 37 },
		{ 343, 154, 37 },
		{ 343, 155, 37 },
		{ 343, 156, 37 },
		{ 343, 157, 37 },
		{ 343, 158, 37 },
		{ 343, 159, 37 },
		{ 343, 160, 37 },
		{ 343, 161, 37 },
		{ 343, 162, 37 },
		{ 343, 163, 37 },
		{ 343, 164, 37 },
		{ 343, 165, 37 },
		{ 343, 166, 37 },
		{ 343, 167, 37 },
		{ 343, 168, 37 },
		{ 343, 169, 37 },
		{ 343, 170, 37 },
		{ 344, 142, 37 },
		{ 344, 143, 37 },
		{ 344, 144, 37 },
		{ 344, 145, 37 },
		{ 344, 146, 37 },
		{ 344, 147, 37 },
		{ 344, 148, 37 },
		{ 344, 149, 37 },
		{ 344, 150, 37 },
		{ 344, 151, 37 },
		{ 344, 152, 37 },
		{ 344, 153, 37 },
		{ 344, 154, 37 },
		{ 344, 155, 37 },
		{ 344, 156, 37 },
		{ 344, 157, 37 },
		{ 344, 158, 37 },
		{ 344, 159, 37 },
		{ 344, 160, 37 },
		{ 344, 161, 37 },
		{ 344, 162, 37 },
		{ 344, 163, 37 },
		{ 344, 164, 37 },
		{ 344, 165, 37 },
		{ 344, 166, 37 },
		{ 344, 167, 37 },
		{ 344, 168, 37 },
		{ 344, 169, 37 },
		{ 345, 140, 37 },
		{ 345, 141, 37 },
		{ 345, 142, 37 },
		{ 345, 143, 37 },
		{ 345, 144, 37 },
		{ 345, 145, 37 },
		{ 345, 146, 37 },
		{ 345, 147, 37 },
		{ 345, 148, 37 },
		{ 345, 149, 37 },
		{ 345, 150, 37 },
		{ 345, 151, 37 },
		{ 345, 152, 37 },
		{ 345, 153, 37 },
		{ 345, 154, 37 },
		{ 345, 155, 37 },
		{ 345, 156, 37 },
		{ 345, 157, 37 },
		{ 345, 158, 37 },
		{ 345, 159, 37 },
		{ 345, 160, 37 },
		{ 345, 161, 37 },
		{ 345, 162, 37 },
		{ 345, 163, 37 },
		{ 345, 164, 37 },
		{ 345, 165, 37 },
		{ 345, 166, 37 },
		{ 345, 167, 37 },
		{ 345, 168, 37 },
		{ 346, 139, 37 },
		{ 346, 140, 37 },
		{ 346, 141, 37 },
		{ 346, 142, 37 },
		{ 346, 143, 37 },
		{ 346, 144, 37 },
		{ 346, 145, 37 },
		{ 346, 146, 37 },
		{ 346, 147, 37 },
		{ 346, 148, 37 },
		{ 346, 149, 37 },
		{ 346, 150, 37 },
		{ 346, 151, 37 },
		{ 346, 152, 37 },
		{ 346, 153, 37 },
		{ 346, 154, 37 },
		{ 346, 155, 37 },
		{ 346, 156, 37 },
		{ 346, 157, 37 },
		{ 346, 158, 37 },
		{ 346, 159, 37 },
		{ 346, 160, 37 },
		{ 346, 161, 37 },
		{ 346, 162, 37 },
		{ 346, 163, 37 },
		{ 346, 164, 37 },
		{ 346, 165, 37 },
		{ 346, 166, 37 },
		{ 346, 167, 37 },
		{ 347, 138, 37 },
		{ 347, 139, 37 },
		{ 347, 140, 37 },
		{ 347, 141, 37 },
		{ 347, 142, 37 },
		{ 347, 143, 37 },
		{ 347, 144, 37 },
		{ 347, 145, 37 },
		{ 347, 146, 37 },
		{ 347, 147, 37 },
		{ 347, 148, 37 },
		{ 347, 149, 37 },
		{ 347, 150, 37 },
		{ 347, 151, 37 },
		{ 347, 152, 37 },
		{ 347, 153, 37 },
		{ 347, 154, 37 },
		{ 347, 155, 37 },
		{ 347, 156, 37 },
		{ 347, 157, 37 },
		{ 347, 158, 37 },
		{ 347, 159, 37 },
		{ 347, 160, 37 },
		{ 347, 161, 37 },
		{ 347, 162, 37 },
		{ 347, 163, 37 },
		{ 347, 164, 37 },
		{ 347, 165, 37 },
		{ 347, 166, 37 },
		{ 348, 136, 37 },
		{ 348, 137, 37 },
		{ 348, 138, 37 },
		{ 348, 139, 37 },
		{ 348, 140, 37 },
		{ 348, 141, 37 },
		{ 348, 142, 37 },
		{ 348, 143, 37 },
		{ 348, 144, 37 },
		{ 348, 145, 37 },
		{ 348, 146, 37 },
		{ 348, 147, 37 },
		{ 348, 148, 37 },
		{ 348, 149, 37 },
		{ 348, 150, 37 },
		{ 348, 151, 37 },
		{ 348, 152, 37 },
		{ 348, 153, 37 },
		{ 348, 154, 37 },
		{ 348, 155, 37 },
		{ 348, 156, 37 },
		{ 348, 157, 37 },
		{ 348, 158, 37 },
		{ 348, 159, 37 },
		{ 348, 160, 37 },
		{ 348, 161, 37 },
		{ 348, 162, 37 },
		{ 348, 163, 37 },
		{ 348, 164, 37 },
		{ 348, 165, 37 },
		{ 349, 135, 37 },
		{ 349, 136, 37 },
		{ 349, 137, 37 },
		{ 349, 138, 37 },
		{ 349, 139, 37 },
		{ 349, 140, 37 },
		{ 349, 141, 37 },
		{ 349, 142, 37 },
		{ 349, 143, 37 },
		{ 349, 144, 37 },
		{ 349, 145, 37 },
		{ 349, 146, 37 },
		{ 349, 147, 37 },
		{ 349, 148, 37 },
		{ 349, 149, 37 },
		{ 349, 150, 37 },
		{ 349, 151, 37 },
		{ 349, 152, 37 },
		{ 349, 153, 37 },
		{ 349, 154, 37 },
		{ 349, 155, 37 },
		{ 349, 156, 37 },
		{ 349, 157, 37 },
		{ 349, 158, 37 },
		{ 349, 159, 37 },
		{ 349, 160, 37 },
		{ 349, 161, 37 },
		{ 349, 162, 37 },
		{ 349, 163, 37 },
		{ 349, 164, 37 },
		{ 350, 134, 37 },
		{ 350, 135, 37 },
		{ 350, 136, 37 },
		{ 350, 137, 37 },
		{ 350, 138, 37 },
		{ 350, 139, 37 },
		{ 350, 140, 37 },
		{ 350, 141, 37 },
		{ 350, 142, 37 },
		{ 350, 143, 37 },
		{ 350, 144, 37 },
		{ 350, 145, 37 },
		{ 350, 146, 37 },
		{ 350, 147, 37 },
		{ 350, 148, 37 },
		{ 350, 149, 37 },
		{ 350, 150, 37 },
		{ 350, 151, 37 },
		{ 350, 152, 37 },
		{ 350, 153, 37 },
		{ 350, 154, 37 },
		{ 350, 155, 37 },
		{ 350, 156, 37 },
		{ 350, 157, 37 },
		{ 350, 158, 37 },
		{ 350, 159, 37 },
		{ 350, 160, 37 },
		{ 350, 161, 37 },
		{ 350, 162, 37 },
		{ 350, 163, 37 },
		{ 351, 132, 37 },
		{ 351, 133, 37 },
		{ 351, 134, 37 },
		{ 351, 135, 37 },
		{ 351, 136, 37 },
		{ 351, 137, 37 },
		{ 351, 138, 37 },
		{ 351, 139, 37 },
		{ 351, 140, 37 },
		{ 351, 141, 37 },
		{ 351, 142, 37 },
		{ 351, 143, 37 },
		{ 351, 144, 37 },
		{ 351, 145, 37 },
		{ 351, 146, 37 },
		{ 351, 147, 37 },
		{ 351, 148, 37 },
		{ 351, 149, 37 },
		{ 351, 150, 37 },
		{ 351, 151, 37 },
		{ 351, 152, 37 },
		{ 351, 153, 37 },
		{ 351, 154, 37 },
		{ 351, 155, 37 },
		{ 351, 156, 37 },
		{ 351, 157, 37 },
		{ 351, 158, 37 },
		{ 351, 159, 37 },
		{ 351, 160, 37 },
		{ 351, 161, 37 },
		{ 351, 162, 37 },
		{ 352, 131, 37 },
		{ 352, 132, 37 },
		{ 352, 133, 37 },
		{ 352, 134, 37 },
		{ 352, 135, 37 },
		{ 352, 136, 37 },
		{ 352, 137, 37 },
		{ 352, 138, 37 },
		{ 352, 139, 37 },
		{ 352, 140, 37 },
		{ 352, 141, 37 },
		{ 352, 142, 37 },
		{ 352, 143, 37 },
		{ 352, 144, 37 },
		{ 352, 145, 37 },
		{ 352, 146, 37 },
		{ 352, 147, 37 },
		{ 352, 148, 37 },
		{ 352, 149, 37 },
		{ 352, 150, 37 },
		{ 352, 151, 37 },
		{ 352, 152, 37 },
		{ 352, 153, 37 },
		{ 352, 154, 37 },
		{ 352, 155, 37 },
		{ 352, 156, 37 },
		{ 352, 157, 37 },
		{ 352, 158, 37 },
		{ 352, 159, 37 },
		{ 352, 160, 37 },
		{ 352, 161, 37 },
		{ 353, 130, 37 },
		{ 353, 131, 37 },
		{ 353, 132, 37 },
		{ 353, 133, 37 },
		{ 353, 134, 37 },
		{ 353, 135, 37 },
		{ 353, 136, 37 },
		{ 353, 137, 37 },
		{ 353, 138, 37 },
		{ 353, 139, 37 },
		{ 353, 140, 37 },
		{ 353, 141, 37 },
		{ 353, 142, 37 },
		{ 353, 143, 37 },
		{ 353, 144, 37 },
		{ 353, 145, 37 },
		{ 353, 146, 37 },
		{ 353, 147, 37 },
		{ 353, 148, 37 },
		{ 353, 149, 37 },
		{ 353, 150, 37 },
		{ 353, 151, 37 },
		{ 353, 152, 37 },
		{ 353, 153, 37 },
		{ 353, 154, 37 },
		{ 353, 155, 37 },
		{ 353, 156, 37 },
		{ 353, 157, 37 },
		{ 353, 158, 37 },
		{ 353, 159, 37 },
		{ 353, 160, 37 },
		{ 354, 128, 37 },
		{ 354, 129, 37 },
		{ 354, 130, 37 },
		{ 354, 131, 37 },
		{ 354, 132, 37 },
		{ 354, 133, 37 },
		{ 354, 134, 37 },
		{ 354, 135, 37 },
		{ 354, 136, 37 },
		{ 354, 137, 37 },
		{ 354, 138, 37 },
		{ 354, 139, 37 },
		{ 354, 140, 37 },
		{ 354, 141, 37 },
		{ 354, 142, 37 },
		{ 354, 143, 37 },
		{ 354, 144, 37 },
		{ 354, 145, 37 },
		{ 354, 146, 37 },
		{ 354, 147, 37 },
		{ 354, 148, 37 },
		{ 354, 149, 37 },
		{ 354, 150, 37 },
		{ 354, 151, 37 },
		{ 354, 152, 37 },
		{ 354, 153, 37 },
		{ 354, 154, 37 },
		{ 354, 155, 37 },
		{ 354, 156, 37 },
		{ 354, 157, 37 },
		{ 354, 158, 37 },
		{ 354, 159, 37 },
		{ 355, 127, 37 },
		{ 355, 128, 37 },
		{ 355, 129, 37 },
		{ 355, 130, 37 },
		{ 355, 131, 37 },
		{ 355, 132, 37 },
		{ 355, 133, 37 },
		{ 355, 134, 37 },
		{ 355, 135, 37 },
		{ 355, 136, 37 },
		{ 355, 137, 37 },
		{ 355, 138, 37 },
		{ 355, 139, 37 },
		{ 355, 140, 37 },
		{ 355, 141, 37 },
		{ 355, 142, 37 },
		{ 355, 143, 37 },
		{ 355, 144, 37 },
		{ 355, 145, 37 },
		{ 355, 146, 37 },
		{ 355, 147, 37 },
		{ 355, 148, 37 },
		{ 355, 149, 37 },
		{ 355, 150, 37 },
		{ 355, 151, 37 },
		{ 355, 152, 37 },
		{ 355, 153, 37 },
		{ 355, 154, 37 },
		{ 355, 155, 37 },
		{ 355, 156, 37 },
		{ 355, 157, 37 },
		{ 355, 158, 37 },
		{ 356, 126, 37 },
		{ 356, 127, 37 },
		{ 356, 128, 37 },
		{ 356, 129, 37 },
		{ 356, 130, 37 },
		{ 356, 131, 37 },
		{ 356, 132, 37 },
		{ 356, 133, 37 },
		{ 356, 134, 37 },
		{ 356, 135, 37 },
		{ 356, 136, 37 },
		{ 356, 137, 37 },
		{ 356, 138, 37 },
		{ 356, 139, 37 },
		{ 356, 140, 37 },
		{ 356, 141, 37 },
		{ 356, 142, 37 },
		{ 356, 143, 37 },
		{ 356, 144, 37 },
		{ 356, 145, 37 },
		{ 356, 146, 37 },
		{ 356, 147, 37 },
		{ 356, 148, 37 },
		{ 356, 149, 37 },
		{ 356, 150, 37 },
		{ 356, 151, 37 },
		{ 356, 152, 37 },
		{ 356, 153, 37 },
		{ 356, 154, 37 },
		{ 356, 155, 37 },
		{ 356, 156, 37 },
		{ 356, 157, 37 },
		{ 357, 124, 37 },
		{ 357, 125, 37 },
		{ 357, 126, 37 },
		{ 357, 127, 37 },
		{ 357, 128, 37 },
		{ 357, 129, 37 },
		{ 357, 130, 37 },
		{ 357, 131, 37 },
		{ 357, 132, 37 },
		{ 357, 133, 37 },
		{ 357, 134, 37 },
		{ 357, 135, 37 },
		{ 357, 136, 37 },
		{ 357, 137, 37 },
		{ 357, 138, 37 },
		{ 357, 139, 37 },
		{ 357, 140, 37 },
		{ 357, 141, 37 },
		{ 357, 142, 37 },
		{ 357, 143, 37 },
		{ 357, 144, 37 },
		{ 357, 145, 37 },
		{ 357, 146, 37 },
		{ 357, 147, 37 },
		{ 357, 148, 37 },
		{ 357, 149, 37 },
		{ 357, 150, 37 },
		{ 357, 151, 37 },
		{ 357, 152, 37 },
		{ 357, 153, 37 },
		{ 357, 154, 37 },
		{ 357, 155, 37 },
		{ 357, 156, 37 },
		{ 358, 123, 37 },
		{ 358, 124, 37 },
		{ 358, 125, 37 },
		{ 358, 126, 37 },
		{ 358, 127, 37 },
		{ 358, 128, 37 },
		{ 358, 129, 37 },
		{ 358, 130, 37 },
		{ 358, 131, 37 },
		{ 358, 132, 37 },
		{ 358, 133, 37 },
		{ 358, 134, 37 },
		{ 358, 135, 37 },
		{ 358, 136, 37 },
		{ 358, 137, 37 },
		{ 358, 138, 37 },
		{ 358, 139, 37 },
		{ 358, 140, 37 },
		{ 358, 141, 37 },
		{ 358, 142, 37 },
		{ 358, 143, 37 },
		{ 358, 144, 37 },
		{ 358, 145, 37 },
		{ 358, 146, 37 },
		{ 358, 147, 37 },
		{ 358, 148, 37 },
		{ 358, 149, 37 },
		{ 358, 150, 37 },
		{ 358, 151, 37 },
		{ 358, 152, 37 },
		{ 358, 153, 37 },
		{ 358, 154, 37 },
		{ 358, 155, 37 },
		{ 359, 122, 37 },
		{ 359, 123, 37 },
		{ 359, 124, 37 },
		{ 359, 125, 37 },
		{ 359, 126, 37 },
		{ 359, 127, 37 },
		{ 359, 128, 37 },
		{ 359, 129, 37 },
		{ 359, 130, 37 },
		{ 359, 131, 37 },
		{ 359, 132, 37 },
		{ 359, 133, 37 },
		{ 359, 134, 37 },
		{ 359, 135, 37 },
		{ 359, 136, 37 },
		{ 359, 137, 37 },
		{ 359, 138, 37 },
		{ 359, 139, 37 },
		{ 359, 140, 37 },
		{ 359, 141, 37 },
		{ 359, 142, 37 },
		{ 359, 143, 37 },
		{ 359, 144, 37 },
		{ 359, 145, 37 },
		{ 359, 146, 37 },
		{ 359, 147, 37 },
		{ 359, 148, 37 },
		{ 359, 149, 37 },
		{ 359, 150, 37 },
		{ 359, 151, 37 },
		{ 359, 152, 37 },
		{ 359, 153, 37 },
		{ 359, 154, 37 },
		{ 360, 120, 37 },
		{ 360, 121, 37 },
		{ 360, 122, 37 },
		{ 360, 123, 37 },
		{ 360, 124, 37 },
		{ 360, 125, 37 },
		{ 360, 126, 37 },
		{ 360, 127, 37 },
		{ 360, 128, 37 },
		{ 360, 129, 37 },
		{ 360, 130, 37 },
		{ 360, 131, 37 },
		{ 360, 132, 37 },
		{ 360, 133, 37 },
		{ 360, 134, 37 },
		{ 360, 135, 37 },
		{ 360, 136, 37 },
		{ 360, 137, 37 },
		{ 360, 138, 37 },
		{ 360, 139, 37 },
		{ 360, 140, 37 },
		{ 360, 141, 37 },
		{ 360, 142, 37 },
		{ 360, 143, 37 },
		{ 360, 144, 37 },
		{ 360, 145, 37 },
		{ 360, 146, 37 },
		{ 360, 147, 37 },
		{ 360, 148, 37 },
		{ 360, 149, 37 },
		{ 360, 150, 37 },
		{ 360, 151, 37 },
		{ 360, 152, 37 },
		{ 360, 153, 37 },
		{ 361, 119, 37 },
		{ 361, 120, 37 },
		{ 361, 121, 37 },
		{ 361, 122, 37 },
		{ 361, 123, 37 },
		{ 361, 124, 37 },
		{ 361, 125, 37 },
		{ 361, 126, 37 },
		{ 361, 127, 37 },
		{ 361, 128, 37 },
		{ 361, 129, 37 },
		{ 361, 130, 37 },
		{ 361, 131, 37 },
		{ 361, 132, 37 },
		{ 361, 133, 37 },
		{ 361, 134, 37 },
		{ 361, 135, 37 },
		{ 361, 136, 37 },
		{ 361, 137, 37 },
		{ 361, 138, 37 },
		{ 361, 139, 37 },
		{ 361, 140, 37 },
		{ 361, 141, 37 },
		{ 361, 142, 37 },
		{ 361, 143, 37 },
		{ 361, 144, 37 },
		{ 361, 145, 37 },
		{ 361, 146, 37 },
		{ 361, 147, 37 },
		{ 361, 148, 37 },
		{ 361, 149, 37 },
		{ 361, 150, 37 },
		{ 361, 151, 37 },
		{ 361, 152, 37 },
		{ 362, 118, 37 },
		{ 362, 119, 37 },
		{ 362, 120, 37 },
		{ 362, 121, 37 },
		{ 362, 122, 37 },
		{ 362, 123, 37 },
		{ 362, 124, 37 },
		{ 362, 125, 37 },
		{ 362, 126, 37 },
		{ 362, 127, 37 },
		{ 362, 128, 37 },
		{ 362, 129, 37 },
		{ 362, 130, 37 },
		{ 362, 131, 37 },
		{ 362, 132, 37 },
		{ 362, 133, 37 },
		{ 362, 134, 37 },
		{ 362, 135, 37 },
		{ 362, 136, 37 },
		{ 362, 137, 37 },
		{ 362, 138, 37 },
		{ 362, 139, 37 },
		{ 362, 140, 37 },
		{ 362, 141, 37 },
		{ 362, 142, 37 },
		{ 362, 143, 37 },
		{ 362, 144, 37 },
		{ 362, 145, 37 },
		{ 362, 146, 37 },
		{ 362, 147, 37 },
		{ 362, 148, 37 },
		{ 362, 149, 37 },
		{ 362, 150, 37 },
		{ 362, 151, 37 },
		{ 363, 116, 37 },
		{ 363, 117, 37 },
		{ 363, 118, 37 },
		{ 363, 119, 37 },
		{ 363, 120, 37 },
		{ 363, 121, 37 },
		{ 363, 122, 37 },
		{ 363, 123, 37 },
		{ 363, 124, 37 },
		{ 363, 125, 37 },
		{ 363, 126, 37 },
		{ 363, 127, 37 },
		{ 363, 128, 37 },
		{ 363, 129, 37 },
		{ 363, 130, 37 },
		{ 363, 131, 37 },
		{ 363, 132, 37 },
		{ 363, 133, 37 },
		{ 363, 134, 37 },
		{ 363, 135, 37 },
		{ 363, 136, 37 },
		{ 363, 137, 37 },
		{ 363, 138, 37 },
		{ 363, 139, 37 },
		{ 363, 140, 37 },
		{ 363, 141, 37 },
		{ 363, 142, 37 },
		{ 363, 143, 37 },
		{ 363, 144, 37 },
		{ 363, 145, 37 },
		{ 363, 146, 37 },
		{ 363, 147, 37 },
		{ 363, 148, 37 },
		{ 363, 149, 37 },
		{ 363, 150, 37 },
		{ 364, 115, 37 },
		{ 364, 116, 37 },
		{ 364, 117, 37 },
		{ 364, 118, 37 },
		{ 364, 119, 37 },
		{ 364, 120, 37 },
		{ 364, 121, 37 },
		{ 364, 122, 37 },
		{ 364, 123, 37 },
		{ 364, 124, 37 },
		{ 364, 125, 37 },
		{ 364, 126, 37 },
		{ 364, 127, 37 },
		{ 364, 128, 37 },
		{ 364, 129, 37 },
		{ 364, 130, 37 },
		{ 364, 131, 37 },
		{ 364, 132, 37 },
		{ 364, 133, 37 },
		{ 364, 134, 37 },
		{ 364, 135, 37 },
		{ 364, 136, 37 },
		{ 364, 137, 37 },
		{ 364, 138, 37 },
		{ 364, 139, 37 },
		{ 364, 140, 37 },
		{ 364, 141, 37 },
		{ 364, 142, 37 },
		{ 364, 143, 37 },
		{ 364, 144, 37 },
		{ 364, 145, 37 },
		{ 364, 146, 37 },
		{ 364, 147, 37 },
		{ 364, 148, 37 },
		{ 364, 149, 37 },
		{ 365, 114, 37 },
		{ 365, 115, 37 },
		{ 365, 116, 37 },
		{ 365, 117, 37 },
		{ 365, 118, 37 },
		{ 365, 119, 37 },
		{ 365, 120, 37 },
		{ 365, 121, 37 },
		{ 365, 122, 37 },
		{ 365, 123, 37 },
		{ 365, 124, 37 },
		{ 365, 125, 37 },
		{ 365, 126, 37 },
		{ 365, 127, 37 },
		{ 365, 128, 37 },
		{ 365, 129, 37 },
		{ 365, 130, 37 },
		{ 365, 131, 37 },
		{ 365, 132, 37 },
		{ 365, 133, 37 },
		{ 365, 134, 37 },
		{ 365, 135, 37 },
		{ 365, 136, 37 },
		{ 365, 137, 37 },
		{ 365, 138, 37 },
		{ 365, 139, 37 },
		{ 365, 140, 37 },
		{ 365, 141, 37 },
		{ 365, 142, 37 },
		{ 365, 143, 37 },
		{ 365, 144, 37 },
		{ 365, 145, 37 },
		{ 365, 146, 37 },
		{ 365, 147, 37 },
		{ 365, 148, 37 },
		{ 366, 112, 37 },
		{ 366, 113, 37 },
		{ 366, 114, 37 },
		{ 366, 115, 37 },
		{ 366, 116, 37 },
		{ 366, 117, 37 },
		{ 366, 118, 37 },
		{ 366, 119, 37 },
		{ 366, 120, 37 },
		{ 366, 121, 37 },
		{ 366, 122, 37 },
		{ 366, 123, 37 },
		{ 366, 124, 37 },
		{ 366, 125, 37 },
		{ 366, 126, 37 },
		{ 366, 127, 37 },
		{ 366, 128, 37 },
		{ 366, 129, 37 },
		{ 366, 130, 37 },
		{ 366, 131, 37 },
		{ 366, 132, 37 },
		{ 366, 133, 37 },
		{ 366, 134, 37 },
		{ 366, 135, 37 },
		{ 366, 136, 37 },
		{ 366, 137, 37 },
		{ 366, 138, 37 },
		{ 366, 139, 37 },
		{ 366, 140, 37 },
		{ 366, 141, 37 },
		{ 366, 142, 37 },
		{ 366, 143, 37 },
		{ 366, 144, 37 },
		{ 366, 145, 37 },
		{ 366, 146, 37 },
		{ 366, 147, 37 },
		{ 367, 111, 37 },
		{ 367, 112, 37 },
		{ 367, 113, 37 },
		{ 367, 114, 37 },
		{ 367, 115, 37 },
		{ 367, 116, 37 },
		{ 367, 117, 37 },
		{ 367, 118, 37 },
		{ 367, 119, 37 },
		{ 367, 120, 37 },
		{ 367, 121, 37 },
		{ 367, 122, 37 },
		{ 367, 123, 37 },
		{ 367, 124, 37 },
		{ 367, 125, 37 },
		{ 367, 126, 37 },
		{ 367, 127, 37 },
		{ 367, 128, 37 },
		{ 367, 129, 37 },
		{ 367, 130, 37 },
		{ 367, 131, 37 },
		{ 367, 132, 37 },
		{ 367, 133, 37 },
		{ 367, 134, 37 },
		{ 367, 135, 37 },
		{ 367, 136, 37 },
		{ 367, 137, 37 },
		{ 367, 138, 37 },
		{ 367, 139, 37 },
		{ 367, 140, 37 },
		{ 367, 141, 37 },
		{ 367, 142, 37 },
		{ 367, 143, 37 },
		{ 367, 144, 37 },
		{ 367, 145, 37 },
		{ 367, 146, 37 },
		{ 368, 110, 37 },
		{ 368, 111, 37 },
		{ 368, 112, 37 },
		{ 368, 113, 37 },
		{ 368, 114, 37 },
		{ 368, 115, 37 },
		{ 368, 116, 37 },
		{ 368, 117, 37 },
		{ 368, 118, 37 },
		{ 368, 119, 37 },
		{ 368, 120, 37 },
		{ 368, 121, 37 },
		{ 368, 122, 37 },
		{ 368, 123, 37 },
		{ 368, 124, 37 },
		{ 368, 125, 37 },
		{ 368, 126, 37 },
		{ 368, 127, 37 },
		{ 368, 128, 37 },
		{ 368, 129, 37 },
		{ 368, 130, 37 },
		{ 368, 131, 37 },
		{ 368, 132, 37 },
		{ 368, 133, 37 },
		{ 368, 134, 37 },
		{ 368, 135, 37 },
		{ 368, 136, 37 },
		{ 368, 137, 37 },
		{ 368, 138, 37 },
		{ 368, 139, 37 },
		{ 368, 140, 37 },
		{ 368, 141, 37 },
		{ 368, 142, 37 },
		{ 368, 143, 37 },
		{ 368, 144, 37 },
		{ 368, 145, 37 },
		{ 369, 108, 37 },
		{ 369, 109, 37 },
		{ 369, 110, 37 },
		{ 369, 111, 37 },
		{ 369, 112, 37 },
		{ 369, 113, 37 },
		{ 369, 114, 37 },
		{ 369, 115, 37 },
		{ 369, 116, 37 },
		{ 369, 117, 37 },
		{ 369, 118, 37 },
		{ 369, 119, 37 },
		{ 369, 120, 37 },
		{ 369, 121, 37 },
		{ 369, 122, 37 },
		{ 369, 123, 37 },
		{ 369, 124, 37 },
		{ 369, 125, 37 },
		{ 369, 126, 37 },
		{ 369, 127, 37 },
		{ 369, 128, 37 },
		{ 369, 129, 37 },
		{ 369, 130, 37 },
		{ 369, 131, 37 },
		{ 369, 132, 37 },
		{ 369, 133, 37 },
		{ 369, 134, 37 },
		{ 369, 135, 37 },
		{ 369, 136, 37 },
		{ 369, 137, 37 },
		{ 369, 138, 37 },
		{ 369, 139, 37 },
		{ 369, 140, 37 },
		{ 369, 141, 37 },
		{ 369, 142, 37 },
		{ 369, 143, 37 },
		{ 369, 144, 37 },
		{ 370, 107, 37 },
		{ 370, 108, 37 },
		{ 370, 109, 37 },
		{ 370, 110, 37 },
		{ 370, 111, 37 },
		{ 370, 112, 37 },
		{ 370, 113, 37 },
		{ 370, 114, 37 },
		{ 370, 115, 37 },
		{ 370, 116, 37 },
		{ 370, 117, 37 },
		{ 370, 118, 37 },
		{ 370, 119, 37 },
		{ 370, 120, 37 },
		{ 370, 121, 37 },
		{ 370, 122, 37 },
		{ 370, 123, 37 },
		{ 370, 124, 37 },
		{ 370, 125, 37 },
		{ 370, 126, 37 },
		{ 370, 127, 37 },
		{ 370, 128, 37 },
		{ 370, 129, 37 },
		{ 370, 130, 37 },
		{ 370, 131, 37 },
		{ 370, 132, 37 },
		{ 370, 133, 37 },
		{ 370, 134, 37 },
		{ 370, 135, 37 },
		{ 370, 136, 37 },
		{ 370, 137, 37 },
		{ 370, 138, 37 },
		{ 370, 139, 37 },
		{ 370, 140, 37 },
		{ 370, 141, 37 },
		{ 370, 142, 37 },
		{ 370, 143, 37 },
		{ 371, 106, 37 },
		{ 371, 107, 37 },
		{ 371, 108, 37 },
		{ 371, 109, 37 },
		{ 371, 110, 37 },
		{ 371, 111, 37 },
		{ 371, 112, 37 },
		{ 371, 113, 37 },
		{ 371, 114, 37 },
		{ 371, 115, 37 },
		{ 371, 116, 37 },
		{ 371, 117, 37 },
		{ 371, 118, 37 },
		{ 371, 119, 37 },
		{ 371, 120, 37 },
		{ 371, 121, 37 },
		{ 371, 122, 37 },
		{ 371, 123, 37 },
		{ 371, 124, 37 },
		{ 371, 125, 37 },
		{ 371, 126, 37 },
		{ 371, 127, 37 },
		{ 371, 128, 37 },
		{ 371, 129, 37 },
		{ 371, 130, 37 },
		{ 371, 131, 37 },
		{ 371, 132, 37 },
		{ 371, 133, 37 },
		{ 371, 134, 37 },
		{ 371, 135, 37 },
		{ 371, 136, 37 },
		{ 371, 137, 37 },
		{ 371, 138, 37 },
		{ 371, 139, 37 },
		{ 371, 140, 37 },
		{ 371, 141, 37 },
		{ 371, 142, 37 },
		{ 372, 104, 37 },
		{ 372, 105, 37 },
		{ 372, 106, 37 },
		{ 372, 107, 37 },
		{ 372, 108, 37 },
		{ 372, 109, 37 },
		{ 372, 110, 37 },
		{ 372, 111, 37 },
		{ 372, 112, 37 },
		{ 372, 113, 37 },
		{ 372, 114, 37 },
		{ 372, 115, 37 },
		{ 372, 116, 37 },
		{ 372, 117, 37 },
		{ 372, 118, 37 },
		{ 372, 119, 37 },
		{ 372, 120, 37 },
		{ 372, 121, 37 },
		{ 372, 122, 37 },
		{ 372, 123, 37 },
		{ 372, 124, 37 },
		{ 372, 125, 37 },
		{ 372, 126, 37 },
		{ 372, 127, 37 },
		{ 372, 128, 37 },
		{ 372, 129, 37 },
		{ 372, 130, 37 },
		{ 372, 131, 37 },
		{ 372, 132, 37 },
		{ 372, 133, 37 },
		{ 372, 134, 37 },
		{ 372, 135, 37 },
		{ 372, 136, 37 },
		{ 372, 137, 37 },
		{ 372, 138, 37 },
		{ 372, 139, 37 },
		{ 372, 140, 37 },
		{ 372, 141, 37 },
		{ 373, 103, 37 },
		{ 373, 104, 37 },
		{ 373, 105, 37 },
		{ 373, 106, 37 },
		{ 373, 107, 37 },
		{ 373, 108, 37 },
		{ 373, 109, 37 },
		{ 373, 110, 37 },
		{ 373, 111, 37 },
		{ 373, 112, 37 },
		{ 373, 113, 37 },
		{ 373, 114, 37 },
		{ 373, 115, 37 },
		{ 373, 116, 37 },
		{ 373, 117, 37 },
		{ 373, 118, 37 },
		{ 373, 119, 37 },
		{ 373, 120, 37 },
		{ 373, 121, 37 },
		{ 373, 122, 37 },
		{ 373, 123, 37 },
		{ 373, 124, 37 },
		{ 373, 125, 37 },
		{ 373, 126, 37 },
		{ 373, 127, 37 },
		{ 373, 128, 37 },
		{ 373, 129, 37 },
		{ 373, 130, 37 },
		{ 373, 131, 37 },
		{ 373, 132, 37 },
		{ 373, 133, 37 },
		{ 373, 134, 37 },
		{ 373, 135, 37 },
		{ 373, 136, 37 },
		{ 373, 137, 37 },
		{ 373, 138, 37 },
		{ 373, 139, 37 },
		{ 373, 140, 37 },
		{ 374, 102, 37 },
		{ 374, 103, 37 },
		{ 374, 104, 37 },
		{ 374, 105, 37 },
		{ 374, 106, 37 },
		{ 374, 107, 37 },
		{ 374, 108, 37 },
		{ 374, 109, 37 },
		{ 374, 110, 37 },
		{ 374, 111, 37 },
		{ 374, 112, 37 },
		{ 374, 113, 37 },
		{ 374, 114, 37 },
		{ 374, 115, 37 },
		{ 374, 116, 37 },
		{ 374, 117, 37 },
		{ 374, 118, 37 },
		{ 374, 119, 37 },
		{ 374, 120, 37 },
		{ 374, 121, 37 },
		{ 374, 122, 37 },
		{ 374, 123, 37 },
		{ 374, 124, 37 },
		{ 374, 125, 37 },
		{ 374, 126, 37 },
		{ 374, 127, 37 },
		{ 374, 128, 37 },
		{ 374, 129, 37 },
		{ 374, 130, 37 },
		{ 374, 131, 37 },
		{ 374, 132, 37 },
		{ 374, 133, 37 },
		{ 374, 134, 37 },
		{ 374, 135, 37 },
		{ 374, 136, 37 },
		{ 374, 137, 37 },
		{ 374, 138, 37 },
		{ 374, 139, 37 },
		{ 375, 100, 37 },
		{ 375, 101, 37 },
		{ 375, 102, 37 },
		{ 375, 103, 37 },
		{ 375, 104, 37 },
		{ 375, 105, 37 },
		{ 375, 106, 37 },
		{ 375, 107, 37 },
		{ 375, 108, 37 },
		{ 375, 109, 37 },
		{ 375, 110, 37 },
		{ 375, 111, 37 },
		{ 375, 112, 37 },
		{ 375, 113, 37 },
		{ 375, 114, 37 },
		{ 375, 115, 37 },
		{ 375, 116, 37 },
		{ 375, 117, 37 },
		{ 375, 118, 37 },
		{ 375, 119, 37 },
		{ 375, 120, 37 },
		{ 375, 121, 37 },
		{ 375, 122, 37 },
		{ 375, 123, 37 },
		{ 375, 124, 37 },
		{ 375, 125, 37 },
		{ 375, 126, 37 },
		{ 375, 127, 37 },
		{ 375, 128, 37 },
		{ 375, 129, 37 },
		{ 375, 130, 37 },
		{ 375, 131, 37 },
		{ 375, 132, 37 },
		{ 375, 133, 37 },
		{ 375, 134, 37 },
		{ 375, 135, 37 },
		{ 375, 136, 37 },
		{ 375, 137, 37 },
		{ 375, 138, 37 },
		{ 376, 99, 37 },
		{ 376, 100, 37 },
		{ 376, 101, 37 },
		{ 376, 102, 37 },
		{ 376, 103, 37 },
		{ 376, 104, 37 },
		{ 376, 105, 37 },
		{ 376, 106, 37 },
		{ 376, 107, 37 },
		{ 376, 108, 37 },
		{ 376, 109, 37 },
		{ 376, 110, 37 },
		{ 376, 111, 37 },
		{ 376, 112, 37 },
		{ 376, 113, 37 },
		{ 376, 114, 37 },
		{ 376, 115, 37 },
		{ 376, 116, 37 },
		{ 376, 117, 37 },
		{ 376, 118, 37 },
		{ 376, 119, 37 },
		{ 376, 120, 37 },
		{ 376, 121, 37 },
		{ 376, 122, 37 },
		{ 376, 123, 37 },
		{ 376, 124, 37 },
		{ 376, 125, 37 },
		{ 376, 126, 37 },
		{ 376, 127, 37 },
		{ 376, 128, 37 },
		{ 376, 129, 37 },
		{ 376, 130, 37 },
		{ 376, 131, 37 },
		{ 376, 132, 37 },
		{ 376, 133, 37 },
		{ 376, 134, 37 },
		{ 376, 135, 37 },
		{ 376, 136, 37 },
		{ 376, 137, 37 },
		{ 377, 98, 37 },
		{ 377, 99, 37 },
		{ 377, 100, 37 },
		{ 377, 101, 37 },
		{ 377, 102, 37 },
		{ 377, 103, 37 },
		{ 377, 104, 37 },
		{ 377, 105, 37 },
		{ 377, 106, 37 },
		{ 377, 107, 37 },
		{ 377, 108, 37 },
		{ 377, 109, 37 },
		{ 377, 110, 37 },
		{ 377, 111, 37 },
		{ 377, 112, 37 },
		{ 377, 113, 37 },
		{ 377, 114, 37 },
		{ 377, 115, 37 },
		{ 377, 116, 37 },
		{ 377, 117, 37 },
		{ 377, 118, 37 },
		{ 377, 119, 37 },
		{ 377, 120, 37 },
		{ 377, 121, 37 },
		{ 377, 122, 37 },
		{ 377, 123, 37 },
		{ 377, 124, 37 },
		{ 377, 125, 37 },
		{ 377, 126, 37 },
		{ 377, 127, 37 },
		{ 377, 128, 37 },
		{ 377, 129, 37 },
		{ 377, 130, 37 },
		{ 377, 131, 37 },
		{ 377, 132, 37 },
		{ 377, 133, 37 },
		{ 377, 134, 37 },
		{ 377, 135, 37 },
		{ 377, 136, 37 },
		{ 378, 96, 37 },
		{ 378, 97, 37 },
		{ 378, 98, 37 },
		{ 378, 99, 37 },
		{ 378, 100, 37 },
		{ 378, 101, 37 },
		{ 378, 102, 37 },
		{ 378, 103, 37 },
		{ 378, 104, 37 },
		{ 378, 105, 37 },
		{ 378, 106, 37 },
		{ 378, 107, 37 },
		{ 378, 108, 37 },
		{ 378, 109, 37 },
		{ 378, 110, 37 },
		{ 378, 111, 37 },
		{ 378, 112, 37 },
		{ 378, 113, 37 },
		{ 378, 114, 37 },
		{ 378, 115, 37 },
		{ 378, 116, 37 },
		{ 378, 117, 37 },
		{ 378, 118, 37 },
		{ 378, 119, 37 },
		{ 378, 120, 37 },
		{ 378, 121, 37 },
		{ 378, 122, 37 },
		{ 378, 123, 37 },
		{ 378, 124, 37 },
		{ 378, 125, 37 },
		{ 378, 126, 37 },
		{ 378, 127, 37 },
		{ 378, 128, 37 },
		{ 378, 129, 37 },
		{ 378, 130, 37 },
		{ 378, 131, 37 },
		{ 378, 132, 37 },
		{ 378, 133, 37 },
		{ 378, 134, 37 },
		{ 378, 135, 37 },
		{ 379, 95, 37 },
		{ 379, 96, 37 },
		{ 379, 97, 37 },
		{ 379, 98, 37 },
		{ 379, 99, 37 },
		{ 379, 100, 37 },
		{ 379, 101, 37 },
		{ 379, 102, 37 },
		{ 379, 103, 37 },
		{ 379, 104, 37 },
		{ 379, 105, 37 },
		{ 379, 106, 37 },
		{ 379, 107, 37 },
		{ 379, 108, 37 },
		{ 379, 109, 37 },
		{ 379, 110, 37 },
		{ 379, 111, 37 },
		{ 379, 112, 37 },
		{ 379, 113, 37 },
		{ 379, 114, 37 },
		{ 379, 115, 37 },
		{ 379, 116, 37 },
		{ 379, 117, 37 },
		{ 379, 118, 37 },
		{ 379, 119, 37 },
		{ 379, 120, 37 },
		{ 379, 121, 37 },
		{ 379, 122, 37 },
		{ 379, 123, 37 },
		{ 379, 124, 37 },
		{ 379, 125, 37 },
		{ 379, 126, 37 },
		{ 379, 127, 37 },
		{ 379, 128, 37 },
		{ 379, 129, 37 },
		{ 379, 130, 37 },
		{ 379, 131, 37 },
		{ 379, 132, 37 },
		{ 379, 133, 37 },
		{ 379, 134, 37 },
		{ 380, 94, 37 },
		{ 380, 95, 37 },
		{ 380, 96, 37 },
		{ 380, 97, 37 },
		{ 380, 98, 37 },
		{ 380, 99, 37 },
		{ 380, 100, 37 },
		{ 380, 101, 37 },
		{ 380, 102, 37 },
		{ 380, 103, 37 },
		{ 380, 104, 37 },
		{ 380, 105, 37 },
		{ 380, 106, 37 },
		{ 380, 107, 37 },
		{ 380, 108, 37 },
		{ 380, 109, 37 },
		{ 380, 110, 37 },
		{ 380, 111, 37 },
		{ 380, 112, 37 },
		{ 380, 113, 37 },
		{ 380, 114, 37 },
		{ 380, 115, 37 },
		{ 380, 116, 37 },
		{ 380, 117, 37 },
		{ 380, 118, 37 },
		{ 380, 119, 37 },
		{ 380, 120, 37 },
		{ 380, 121, 37 },
		{ 380, 122, 37 },
		{ 380, 123, 37 },
		{ 380, 124, 37 },
		{ 380, 125, 37 },
		{ 380, 126, 37 },
		{ 380, 127, 37 },
		{ 380, 128, 37 },
		{ 380, 129, 37 },
		{ 380, 130, 37 },
		{ 380, 131, 37 },
		{ 380, 132, 37 },
		{ 380, 133, 37 },
		{ 381, 92, 37 },
		{ 381, 93, 37 },
		{ 381, 94, 37 },
		{ 381, 95, 37 },
		{ 381, 96, 37 },
		{ 381, 97, 37 },
		{ 381, 98, 37 },
		{ 381, 99, 37 },
		{ 381, 100, 37 },
		{ 381, 101, 37 },
		{ 381, 102, 37 },
		{ 381, 103, 37 },
		{ 381, 104, 37 },
		{ 381, 105, 37 },
		{ 381, 106, 37 },
		{ 381, 107, 37 },
		{ 381, 108, 37 },
		{ 381, 109, 37 },
		{ 381, 110, 37 },
		{ 381, 111, 37 },
		{ 381, 112, 37 },
		{ 381, 113, 37 },
		{ 381, 114, 37 },
		{ 381, 115, 37 },
		{ 381, 116, 37 },
		{ 381, 117, 37 },
		{ 381, 118, 37 },
		{ 381, 119, 37 },
		{ 381, 120, 37 },
		{ 381, 121, 37 },
		{ 381, 122, 37 },
		{ 381, 123, 37 },
		{ 381, 124, 37 },
		{ 381, 125, 37 },
		{ 381, 126, 37 },
		{ 381, 127, 37 },
		{ 381, 128, 37 },
		{ 381, 129, 37 },
		{ 381, 130, 37 },
		{ 381, 131, 37 },
		{ 381, 132, 37 },
		{ 382, 91, 37 },
		{ 382, 92, 37 },
		{ 382, 93, 37 },
		{ 382, 94, 37 },
		{ 382, 95, 37 },
		{ 382, 96, 37 },
		{ 382, 97, 37 },
		{ 382, 98, 37 },
		{ 382, 99, 37 },
		{ 382, 100, 37 },
		{ 382, 101, 37 },
		{ 382, 102, 37 },
		{ 382, 103, 37 },
		{ 382, 104, 37 },
		{ 382, 105, 37 },
		{ 382, 106, 37 },
		{ 382, 107, 37 },
		{ 382, 108, 37 },
		{ 382, 109, 37 },
		{ 382, 110, 37 },
		{ 382, 111, 37 },
		{ 382, 112, 37 },
		{ 382, 113, 37 },
		{ 382, 114, 37 },
		{ 382, 115, 37 },
		{ 382, 116, 37 },
		{ 382, 117, 37 },
		{ 382, 118, 37 },
		{ 382, 119, 37 },
		{ 382, 120, 37 },
		{ 382, 121, 37 },
		{ 382, 122, 37 },
		{ 382, 123, 37 },
		{ 382, 124, 37 },
		{ 382, 125, 37 },
		{ 382, 126, 37 },
		{ 382, 127, 37 },
		{ 382, 128, 37 },
		{ 382, 129, 37 },
		{ 382, 130, 37 },
		{ 382, 131, 37 },
		{ 383, 90, 37 },
		{ 383, 91, 37 },
		{ 383, 92, 37 },
		{ 383, 93, 37 },
		{ 383, 94, 37 },
		{ 383, 95, 37 },
		{ 383, 96, 37 },
		{ 383, 97, 37 },
		{ 383, 98, 37 },
		{ 383, 99, 37 },
		{ 383, 100, 37 },
		{ 383, 101, 37 },
		{ 383, 102, 37 },
		{ 383, 103, 37 },
		{ 383, 104, 37 },
		{ 383, 105, 37 },
		{ 383, 106, 37 },
		{ 383, 107, 37 },
		{ 383, 108, 37 },
		{ 383, 109, 37 },
		{ 383, 110, 37 },
		{ 383, 111, 37 },
		{ 383, 112, 37 },
		{ 383, 113, 37 },
		{ 383, 114, 37 },
		{ 383, 115, 37 },
		{ 383, 116, 37 },
		{ 383, 117, 37 },
		{ 383, 118, 37 },
		{ 383, 119, 37 },
		{ 383, 120, 37 },
		{ 383, 121, 37 },
		{ 383, 122, 37 },
		{ 383, 123, 37 },
		{ 383, 124, 37 },
		{ 383, 125, 37 },
		{ 383, 126, 37 },
		{ 383, 127, 37 },
		{ 383, 128, 37 },
		{ 383, 129, 37 },
		{ 383, 130, 37 },
		{ 384, 88, 37 },
		{ 384, 89, 37 },
		{ 384, 90, 37 },
		{ 384, 91, 37 },
		{ 384, 92, 37 },
		{ 384, 93, 37 },
		{ 384, 94, 37 },
		{ 384, 95, 37 },
		{ 384, 96, 37 },
		{ 384, 97, 37 },
		{ 384, 98, 37 },
		{ 384, 99, 37 },
		{ 384, 100, 37 },
		{ 384, 101, 37 },
		{ 384, 102, 37 },
		{ 384, 103, 37 },
		{ 384, 104, 37 },
		{ 384, 105, 37 },
		{ 384, 106, 37 },
		{ 384, 107, 37 },
		{ 384, 108, 37 },
		{ 384, 109, 37 },
		{ 384, 110, 37 },
		{ 384, 111, 37 },
		{ 384, 112, 37 },
		{ 384, 113, 37 },
		{ 384, 114, 37 },
		{ 384, 115, 37 },
		{ 384, 116, 37 },
		{ 384, 117, 37 },
		{ 384, 118, 37 },
		{ 384, 119, 37 },
		{ 384, 120, 37 },
		{ 384, 121, 37 },
		{ 384, 122, 37 },
		{ 384, 123, 37 },
		{ 384, 124, 37 },
		{ 384, 125, 37 },
		{ 384, 126, 37 },
		{ 384, 127, 37 },
		{ 384, 128, 37 },
		{ 384, 129, 37 },
		{ 385, 87, 37 },
		{ 385, 88, 37 },
		{ 385, 89, 37 },
		{ 385, 90, 37 },
		{ 385, 91, 37 },
		{ 385, 92, 37 },
		{ 385, 93, 37 },
		{ 385, 94, 37 },
		{ 385, 95, 37 },
		{ 385, 96, 37 },
		{ 385, 97, 37 },
		{ 385, 98, 37 },
		{ 385, 99, 37 },
		{ 385, 100, 37 },
		{ 385, 101, 37 },
		{ 385, 102, 37 },
		{ 385, 103, 37 },
		{ 385, 104, 37 },
		{ 385, 105, 37 },
		{ 385, 106, 37 },
		{ 385, 107, 37 },
		{ 385, 108, 37 },
		{ 385, 109, 37 },
		{ 385, 110, 37 },
		{ 385, 111, 37 },
		{ 385, 112, 37 },
		{ 385, 113, 37 },
		{ 385, 114, 37 },
		{ 385, 115, 37 },
		{ 385, 116, 37 },
		{ 385, 117, 37 },
		{ 385, 118, 37 },
		{ 385, 119, 37 },
		{ 385, 120, 37 },
		{ 385, 121, 37 },
		{ 385, 122, 37 },
		{ 385, 123, 37 },
		{ 385, 124, 37 },
		{ 385, 125, 37 },
		{ 385, 126, 37 },
		{ 385, 127, 37 },
		{ 385, 128, 37 },
		{ 386, 86, 37 },
		{ 386, 87, 37 },
		{ 386, 88, 37 },
		{ 386, 89, 37 },
		{ 386, 90, 37 },
		{ 386, 91, 37 },
		{ 386, 92, 37 },
		{ 386, 93, 37 },
		{ 386, 94, 37 },
		{ 386, 95, 37 },
		{ 386, 96, 37 },
		{ 386, 97, 37 },
		{ 386, 98, 37 },
		{ 386, 99, 37 },
		{ 386, 100, 37 },
		{ 386, 101, 37 },
		{ 386, 102, 37 },
		{ 386, 103, 37 },
		{ 386, 104, 37 },
		{ 386, 105, 37 },
		{ 386, 106, 37 },
		{ 386, 107, 37 },
		{ 386, 108, 37 },
		{ 386, 109, 37 },
		{ 386, 110, 37 },
		{ 386, 111, 37 },
		{ 386, 112, 37 },
		{ 386, 113, 37 },
		{ 386, 114, 37 },
		{ 386, 115, 37 },
		{ 386, 116, 37 },
		{ 386, 117, 37 },
		{ 386, 118, 37 },
		{ 386, 119, 37 },
		{ 386, 120, 37 },
		{ 386, 121, 37 },
		{ 386, 122, 37 },
		{ 386, 123, 37 },
		{ 386, 124, 37 },
		{ 386, 125, 37 },
		{ 386, 126, 37 },
		{ 386, 127, 37 },
		{ 387, 84, 37 },
		{ 387, 85, 37 },
		{ 387, 86, 37 },
		{ 387, 87, 37 },
		{ 387, 88, 37 },
		{ 387, 89, 37 },
		{ 387, 90, 37 },
		{ 387, 91, 37 },
		{ 387, 92, 37 },
		{ 387, 93, 37 },
		{ 387, 94, 37 },
		{ 387, 95, 37 },
		{ 387, 96, 37 },
		{ 387, 97, 37 },
		{ 387, 98, 37 },
		{ 387, 99, 37 },
		{ 387, 100, 37 },
		{ 387, 101, 37 },
		{ 387, 102, 37 },
		{ 387, 103, 37 },
		{ 387, 104, 37 },
		{ 387, 105, 37 },
		{ 387, 106, 37 },
		{ 387, 107, 37 },
		{ 387, 108, 37 },
		{ 387, 109, 37 },
		{ 387, 110, 37 },
		{ 387, 111, 37 },
		{ 387, 112, 37 },
		{ 387, 113, 37 },
		{ 387, 114, 37 },
		{ 387, 115, 37 },
		{ 387, 116, 37 },
		{ 387, 117, 37 },
		{ 387, 118, 37 },
		{ 387, 119, 37 },
		{ 387, 120, 37 },
		{ 387, 121, 37 },
		{ 387, 122, 37 },
		{ 387, 123, 37 },
		{ 387, 124, 37 },
		{ 387, 125, 37 },
		{ 387, 126, 37 },
		{ 388, 83, 37 },
		{ 388, 84, 37 },
		{ 388, 85, 37 },
		{ 388, 86, 37 },
		{ 388, 87, 37 },
		{ 388, 88, 37 },
		{ 388, 89, 37 },
		{ 388, 90, 37 },
		{ 388, 91, 37 },
		{ 388, 92, 37 },
		{ 388, 93, 37 },
		{ 388, 94, 37 },
		{ 388, 95, 37 },
		{ 388, 96, 37 },
		{ 388, 97, 37 },
		{ 388, 98, 37 },
		{ 388, 99, 37 },
		{ 388, 100, 37 },
		{ 388, 101, 37 },
		{ 388, 102, 37 },
		{ 388, 103, 37 },
		{ 388, 104, 37 },
		{ 388, 105, 37 },
		{ 388, 106, 37 },
		{ 388, 107, 37 },
		{ 388, 108, 37 },
		{ 388, 109, 37 },
		{ 388, 110, 37 },
		{ 388, 111, 37 },
		{ 388, 112, 37 },
		{ 388, 113, 37 },
		{ 388, 114, 37 },
		{ 388, 115, 37 },
		{ 388, 116, 37 },
		{ 388, 117, 37 },
		{ 388, 118, 37 },
		{ 388, 119, 37 },
		{ 388, 120, 37 },
		{ 388, 121, 37 },
		{ 388, 122, 37 },
		{ 388, 123, 37 },
		{ 388, 124, 37 },
		{ 388, 125, 37 },
		{ 389, 82, 37 },
		{ 389, 83, 37 },
		{ 389, 84, 37 },
		{ 389, 85, 37 },
		{ 389, 86, 37 },
		{ 389, 87, 37 },
		{ 389, 88, 37 },
		{ 389, 89, 37 },
		{ 389, 90, 37 },
		{ 389, 91, 37 },
		{ 389, 92, 37 },
		{ 389, 93, 37 },
		{ 389, 94, 37 },
		{ 389, 95, 37 },
		{ 389, 96, 37 },
		{ 389, 97, 37 },
		{ 389, 98, 37 },
		{ 389, 99, 37 },
		{ 389, 100, 37 },
		{ 389, 101, 37 },
		{ 389, 102, 37 },
		{ 389, 103, 37 },
		{ 389, 104, 37 },
		{ 389, 105, 37 },
		{ 389, 106, 37 },
		{ 389, 107, 37 },
		{ 389, 108, 37 },
		{ 389, 109, 37 },
		{ 389, 110, 37 },
		{ 389, 111, 37 },
		{ 389, 112, 37 },
		{ 389, 113, 37 },
		{ 389, 114, 37 },
		{ 389, 115, 37 },
		{ 389, 116, 37 },
		{ 389, 117, 37 },
		{ 389, 118, 37 },
		{ 389, 119, 37 },
		{ 389, 120, 37 },
		{ 389, 121, 37 },
		{ 389, 122, 37 },
		{ 389, 123, 37 },
		{ 389, 124, 37 },
		{ 390, 80, 37 },
		{ 390, 81, 37 },
		{ 390, 82, 37 },
		{ 390, 83, 37 },
		{ 390, 84, 37 },
		{ 390, 85, 37 },
		{ 390, 86, 37 },
		{ 390, 87, 37 },
		{ 390, 88, 37 },
		{ 390, 89, 37 },
		{ 390, 90, 37 },
		{ 390, 91, 37 },
		{ 390, 92, 37 },
		{ 390, 93, 37 },
		{ 390, 94, 37 },
		{ 390, 95, 37 },
		{ 390, 96, 37 },
		{ 390, 97, 37 },
		{ 390, 98, 37 },
		{ 390, 99, 37 },
		{ 390, 100, 37 },
		{ 390, 101, 37 },
		{ 390, 102, 37 },
		{ 390, 103, 37 },
		{ 390, 104, 37 },
		{ 390, 105, 37 },
		{ 390, 106, 37 },
		{ 390, 107, 37 },
		{ 390, 108, 37 },
		{ 390, 109, 37 },
		{ 390, 110, 37 },
		{ 390, 111, 37 },
		{ 390, 112, 37 },
		{ 390, 113, 37 },
		{ 390, 114, 37 },
		{ 390, 115, 37 },
		{ 390, 116, 37 },
		{ 390, 117, 37 },
		{ 390, 118, 37 },
		{ 390, 119, 37 },
		{ 390, 120, 37 },
		{ 390, 121, 37 },
		{ 390, 122, 37 },
		{ 390, 123, 37 },
		{ 391, 79, 37 },
		{ 391, 80, 37 },
		{ 391, 81, 37 },
		{ 391, 82, 37 },
		{ 391, 83, 37 },
		{ 391, 84, 37 },
		{ 391, 85, 37 },
		{ 391, 86, 37 },
		{ 391, 87, 37 },
		{ 391, 88, 37 },
		{ 391, 89, 37 },
		{ 391, 90, 37 },
		{ 391, 91, 37 },
		{ 391, 92, 37 },
		{ 391, 93, 37 },
		{ 391, 94, 37 },
		{ 391, 95, 37 },
		{ 391, 96, 37 },
		{ 391, 97, 37 },
		{ 391, 98, 37 },
		{ 391, 99, 37 },
		{ 391, 100, 37 },
		{ 391, 101, 37 },
		{ 391, 102, 37 },
		{ 391, 103, 37 },
		{ 391, 104, 37 },
		{ 391, 105, 37 },
		{ 391, 106, 37 },
		{ 391, 107, 37 },
		{ 391, 108, 37 },
		{ 391, 109, 37 },
		{ 391, 110, 37 },
		{ 391, 111, 37 },
		{ 391, 112, 37 },
		{ 391, 113, 37 },
		{ 391, 114, 37 },
		{ 391, 115, 37 },
		{ 391, 116, 37 },
		{ 391, 117, 37 },
		{ 391, 118, 37 },
		{ 391, 119, 37 },
		{ 391, 120, 37 },
		{ 391, 121, 37 },
		{ 391, 122, 37 },
		{ 392, 78, 37 },
		{ 392, 79, 37 },
		{ 392, 80, 37 },
		{ 392, 81, 37 },
		{ 392, 82, 37 },
		{ 392, 83, 37 },
		{ 392, 84, 37 },
		{ 392, 85, 37 },
		{ 392, 86, 37 },
		{ 392, 87, 37 },
		{ 392, 88, 37 },
		{ 392, 89, 37 },
		{ 392, 90, 37 },
		{ 392, 91, 37 },
		{ 392, 92, 37 },
		{ 392, 93, 37 },
		{ 392, 94, 37 },
		{ 392, 95, 37 },
		{ 392, 96, 37 },
		{ 392, 97, 37 },
		{ 392, 98, 37 },
		{ 392, 99, 37 },
		{ 392, 100, 37 },
		{ 392, 101, 37 },
		{ 392, 102, 37 },
		{ 392, 103, 37 },
		{ 392, 104, 37 },
		{ 392, 105, 37 },
		{ 392, 106, 37 },
		{ 392, 107, 37 },
		{ 392, 108, 37 },
		{ 392, 109, 37 },
		{ 392, 110, 37 },
		{ 392, 111, 37 },
		{ 392, 112, 37 },
		{ 392, 113, 37 },
		{ 392, 114, 37 },
		{ 392, 115, 37 },
		{ 392, 116, 37 },
		{ 392, 117, 37 },
		{ 392, 118, 37 },
		{ 392, 119, 37 },
		{ 392, 120, 37 },
		{ 392, 121, 37 },
		{ 393, 76, 37 },
		{ 393, 77, 37 },
		{ 393, 78, 37 },
		{ 393, 79, 37 },
		{ 393, 80, 37 },
		{ 393, 81, 37 },
		{ 393, 82, 37 },
		{ 393, 83, 37 },
		{ 393, 84, 37 },
		{ 393, 85, 37 },
		{ 393, 86, 37 },
		{ 393, 87, 37 },
		{ 393, 88, 37 },
		{ 393, 89, 37 },
		{ 393, 90, 37 },
		{ 393, 91, 37 },
		{ 393, 92, 37 },
		{ 393, 93, 37 },
		{ 393, 94, 37 },
		{ 393, 95, 37 },
		{ 393, 96, 37 },
		{ 393, 97, 37 },
		{ 393, 98, 37 },
		{ 393, 99, 37 },
		{ 393, 100, 37 },
		{ 393, 101, 37 },
		{ 393, 102, 37 },
		{ 393, 103, 37 },
		{ 393, 104, 37 },
		{ 393, 105, 37 },
		{ 393, 106, 37 },
		{ 393, 107, 37 },
		{ 393, 108, 37 },
		{ 393, 109, 37 },
		{ 393, 110, 37 },
		{ 393, 111, 37 },
		{ 393, 112, 37 },
		{ 393, 113, 37 },
		{ 393, 114, 37 },
		{ 393, 115, 37 },
		{ 393, 116, 37 },
		{ 393, 117, 37 },
		{ 393, 118, 37 },
		{ 393, 119, 37 },
		{ 393, 120, 37 },
		{ 394, 75, 37 },
		{ 394, 76, 37 },
		{ 394, 77, 37 },
		{ 394, 78, 37 },
		{ 394, 79, 37 },
		{ 394, 80, 37 },
		{ 394, 81, 37 },
		{ 394, 82, 37 },
		{ 394, 83, 37 },
		{ 394, 84, 37 },
		{ 394, 85, 37 },
		{ 394, 86, 37 },
		{ 394, 87, 37 },
		{ 394, 88, 37 },
		{ 394, 89, 37 },
		{ 394, 90, 37 },
		{ 394, 91, 37 },
		{ 394, 92, 37 },
		{ 394, 93, 37 },
		{ 394, 94, 37 },
		{ 394, 95, 37 },
		{ 394, 96, 37 },
		{ 394, 97, 37 },
		{ 394, 98, 37 },
		{ 394, 99, 37 },
		{ 394, 100, 37 },
		{ 394, 101, 37 },
		{ 394, 102, 37 },
		{ 394, 103, 37 },
		{ 394, 104, 37 },
		{ 394, 105, 37 },
		{ 394, 106, 37 },
		{ 394, 107, 37 },
		{ 394, 108, 37 },
		{ 394, 109, 37 },
		{ 394, 110, 37 },
		{ 394, 111, 37 },
		{ 394, 112, 37 },
		{ 394, 113, 37 },
		{ 394, 114, 37 },
		{ 394, 115, 37 },
		{ 394, 116, 37 },
		{ 394, 117, 37 },
		{ 394, 118, 37 },
		{ 394, 119, 37 },
		{ 395, 74, 37 },
		{ 395, 75, 37 },
		{ 395, 76, 37 },
		{ 395, 77, 37 },
		{ 395, 78, 37 },
		{ 395, 79, 37 },
		{ 395, 80, 37 },
		{ 395, 81, 37 },
		{ 395, 82, 37 },
		{ 395, 83, 37 },
		{ 395, 84, 37 },
		{ 395, 85, 37 },
		{ 395, 86, 37 },
		{ 395, 87, 37 },
		{ 395, 88, 37 },
		{ 395, 89, 37 },
		{ 395, 90, 37 },
		{ 395, 91, 37 },
		{ 395, 92, 37 },
		{ 395, 93, 37 },
		{ 395, 94, 37 },
		{ 395, 95, 37 },
		{ 395, 96, 37 },
		{ 395, 97, 37 },
		{ 395, 98, 37 },
		{ 395, 99, 37 },
		{ 395, 100, 37 },
		{ 395, 101, 37 },
		{ 395, 102, 37 },
		{ 395, 103, 37 },
		{ 395, 104, 37 },
		{ 395, 105, 37 },
		{ 395, 106, 37 },
		{ 395, 107, 37 },
		{ 395, 108, 37 },
		{ 395, 109, 37 },
		{ 395, 110, 37 },
		{ 395, 111, 37 },
		{ 395, 112, 37 },
		{ 395, 113, 37 },
		{ 395, 114, 37 },
		{ 395, 115, 37 },
		{ 395, 116, 37 },
		{ 395, 117, 37 },
		{ 395, 118, 37 },
		{ 396, 72, 37 },
		{ 396, 73, 37 },
		{ 396, 74, 37 },
		{ 396, 75, 37 },
		{ 396, 76, 37 },
		{ 396, 77, 37 },
		{ 396, 78, 37 },
		{ 396, 79, 37 },
		{ 396, 80, 37 },
		{ 396, 81, 37 },
		{ 396, 82, 37 },
		{ 396, 83, 37 },
		{ 396, 84, 37 },
		{ 396, 85, 37 },
		{ 396, 86, 37 },
		{ 396, 87, 37 },
		{ 396, 88, 37 },
		{ 396, 89, 37 },
		{ 396, 90, 37 },
		{ 396, 91, 37 },
		{ 396, 92, 37 },
		{ 396, 93, 37 },
		{ 396, 94, 37 },
		{ 396, 95, 37 },
		{ 396, 96, 37 },
		{ 396, 97, 37 },
		{ 396, 98, 37 },
		{ 396, 99, 37 },
		{ 396, 100, 37 },
		{ 396, 101, 37 },
		{ 396, 102, 37 },
		{ 396, 103, 37 },
		{ 396, 104, 37 },
		{ 396, 105, 37 },
		{ 396, 106, 37 },
		{ 396, 107, 37 },
		{ 396, 108, 37 },
		{ 396, 109, 37 },
		{ 396, 110, 37 },
		{ 396, 111, 37 },
		{ 396, 112, 37 },
		{ 396, 113, 37 },
		{ 396, 114, 37 },
		{ 396, 115, 37 },
		{ 396, 116, 37 },
		{ 396, 117, 37 },
		{ 397, 71, 37 },
		{ 397, 72, 37 },
		{ 397, 73, 37 },
		{ 397, 74, 37 },
		{ 397, 75, 37 },
		{ 397, 76, 37 },
		{ 397, 77, 37 },
		{ 397, 78, 37 },
		{ 397, 79, 37 },
		{ 397, 80, 37 },
		{ 397, 81, 37 },
		{ 397, 82, 37 },
		{ 397, 83, 37 },
		{ 397, 84, 37 },
		{ 397, 85, 37 },
		{ 397, 86, 37 },
		{ 397, 87, 37 },
		{ 397, 88, 37 },
		{ 397, 89, 37 },
		{ 397, 90, 37 },
		{ 397, 91, 37 },
		{ 397, 92, 37 },
		{ 397, 93, 37 },
		{ 397, 94, 37 },
		{ 397, 95, 37 },
		{ 397, 96, 37 },
		{ 397, 97, 37 },
		{ 397, 98, 37 },
		{ 397, 99, 37 },
		{ 397, 100, 37 },
		{ 397, 101, 37 },
		{ 397, 102, 37 },
		{ 397, 103, 37 },
		{ 397, 104, 37 },
		{ 397, 105, 37 },
		{ 397, 106, 37 },
		{ 397, 107, 37 },
		{ 397, 108, 37 },
		{ 397, 109, 37 },
		{ 397, 110, 37 },
		{ 397, 111, 37 },
		{ 397, 112, 37 },
		{ 397, 113, 37 },
		{ 397, 114, 37 },
		{ 397, 115, 37 },
		{ 397, 116, 37 },
		{ 398, 70, 37 },
		{ 398, 71, 37 },
		{ 398, 72, 37 },
		{ 398, 73, 37 },
		{ 398, 74, 37 },
		{ 398, 75, 37 },
		{ 398, 76, 37 },
		{ 398, 77, 37 },
		{ 398, 78, 37 },
		{ 398, 79, 37 },
		{ 398, 80, 37 },
		{ 398, 81, 37 },
		{ 398, 82, 37 },
		{ 398, 83, 37 },
		{ 398, 84, 37 },
		{ 398, 85, 37 },
		{ 398, 86, 37 },
		{ 398, 87, 37 },
		{ 398, 88, 37 },
		{ 398, 89, 37 },
		{ 398, 90, 37 },
		{ 398, 91, 37 },
		{ 398, 92, 37 },
		{ 398, 93, 37 },
		{ 398, 94, 37 },
		{ 398, 95, 37 },
		{ 398, 96, 37 },
		{ 398, 97, 37 },
		{ 398, 98, 37 },
		{ 398, 99, 37 },
		{ 398, 100, 37 },
		{ 398, 101, 37 },
		{ 398, 102, 37 },
		{ 398, 103, 37 },
		{ 398, 104, 37 },
		{ 398, 105, 37 },
		{ 398, 106, 37 },
		{ 398, 107, 37 },
		{ 398, 108, 37 },
		{ 398, 109, 37 },
		{ 398, 110, 37 },
		{ 398, 111, 37 },
		{ 398, 112, 37 },
		{ 398, 113, 37 },
		{ 398, 114, 37 },
		{ 398, 115, 37 },
		{ 399, 68, 37 },
		{ 399, 69, 37 },
		{ 399, 70, 37 },
		{ 399, 71, 37 },
		{ 399, 72, 37 },
		{ 399, 73, 37 },
		{ 399, 74, 37 },
		{ 399, 75, 37 },
		{ 399, 76, 37 },
		{ 399, 77, 37 },
		{ 399, 78, 37 },
		{ 399, 79, 37 },
		{ 399, 80, 37 },
		{ 399, 81, 37 },
		{ 399, 82, 37 },
		{ 399, 83, 37 },
		{ 399, 84, 37 },
		{ 399, 85, 37 },
		{ 399, 86, 37 },
		{ 399, 87, 37 },
		{ 399, 88, 37 },
		{ 399, 89, 37 },
		{ 399, 90, 37 },
		{ 399, 91, 37 },
		{ 399, 92, 37 },
		{ 399, 93, 37 },
		{ 399, 94, 37 },
		{ 399, 95, 37 },
		{ 399, 96, 37 },
		{ 399, 97, 37 },
		{ 399, 98, 37 },
		{ 399, 99, 37 },
		{ 399, 100, 37 },
		{ 399, 101, 37 },
		{ 399, 102, 37 },
		{ 399, 103, 37 },
		{ 399, 104, 37 },
		{ 399, 105, 37 },
		{ 399, 106, 37 },
		{ 399, 107, 37 },
		{ 399, 108, 37 },
		{ 399, 109, 37 },
		{ 399, 110, 37 },
		{ 399, 111, 37 },
		{ 399, 112, 37 },
		{ 399, 113, 37 },
		{ 399, 114, 37 },
		{ 400, 67, 37 },
		{ 400, 68, 37 },
		{ 400, 69, 37 },
		{ 400, 70, 37 },
		{ 400, 71, 37 },
		{ 400, 72, 37 },
		{ 400, 73, 37 },
		{ 400, 74, 37 },
		{ 400, 75, 37 },
		{ 400, 76, 37 },
		{ 400, 77, 37 },
		{ 400, 78, 37 },
		{ 400, 79, 37 },
		{ 400, 80, 37 },
		{ 400, 81, 37 },
		{ 400, 82, 37 },
		{ 400, 83, 37 },
		{ 400, 84, 37 },
		{ 400, 85, 37 },
		{ 400, 86, 37 },
		{ 400, 87, 37 },
		{ 400, 88, 37 },
		{ 400, 89, 37 },
		{ 400, 90, 37 },
		{ 400, 91, 37 },
		{ 400, 92, 37 },
		{ 400, 93, 37 },
		{ 400, 94, 37 },
		{ 400, 95, 37 },
		{ 400, 96, 37 },
		{ 400, 97, 37 },
		{ 400, 98, 37 },
		{ 400, 99, 37 },
		{ 400, 100, 37 },
		{ 400, 101, 37 },
		{ 400, 102, 37 },
		{ 400, 103, 37 },
		{ 400, 104, 37 },
		{ 400, 105, 37 },
		{ 400, 106, 37 },
		{ 400, 107, 37 },
		{ 400, 108, 37 },
		{ 400, 109, 37 },
		{ 400, 110, 37 },
		{ 400, 111, 37 },
		{ 400, 112, 37 },
		{ 400, 113, 37 },
		{ 401, 66, 37 },
		{ 401, 67, 37 },
		{ 401, 68, 37 },
		{ 401, 69, 37 },
		{ 401, 70, 37 },
		{ 401, 71, 37 },
		{ 401, 72, 37 },
		{ 401, 73, 37 },
		{ 401, 74, 37 },
		{ 401, 75, 37 },
		{ 401, 76, 37 },
		{ 401, 77, 37 },
		{ 401, 78, 37 },
		{ 401, 79, 37 },
		{ 401, 80, 37 },
		{ 401, 81, 37 },
		{ 401, 82, 37 },
		{ 401, 83, 37 },
		{ 401, 84, 37 },
		{ 401, 85, 37 },
		{ 401, 86, 37 },
		{ 401, 87, 37 },
		{ 401, 88, 37 },
		{ 401, 89, 37 },
		{ 401, 90, 37 },
		{ 401, 91, 37 },
		{ 401, 92, 37 },
		{ 401, 93, 37 },
		{ 401, 94, 37 },
		{ 401, 95, 37 },
		{ 401, 96, 37 },
		{ 401, 97, 37 },
		{ 401, 98, 37 },
		{ 401, 99, 37 },
		{ 401, 100, 37 },
		{ 401, 101, 37 },
		{ 401, 102, 37 },
		{ 401, 103, 37 },
		{ 401, 104, 37 },
		{ 401, 105, 37 },
		{ 401, 106, 37 },
		{ 401, 107, 37 },
		{ 401, 108, 37 },
		{ 401, 109, 37 },
		{ 401, 110, 37 },
		{ 401, 111, 37 },
		{ 401, 112, 37 },
		{ 402, 64, 37 },
		{ 402, 65, 37 },
		{ 402, 66, 37 },
		{ 402, 67, 37 },
		{ 402, 68, 37 },
		{ 402, 69, 37 },
		{ 402, 70, 37 },
		{ 402, 71, 37 },
		{ 402, 72, 37 },
		{ 402, 73, 37 },
		{ 402, 74, 37 },
		{ 402, 75, 37 },
		{ 402, 76, 37 },
		{ 402, 77, 37 },
		{ 402, 78, 37 },
		{ 402, 79, 37 },
		{ 402, 80, 37 },
		{ 402, 81, 37 },
		{ 402, 82, 37 },
		{ 402, 83, 37 },
		{ 402, 84, 37 },
		{ 402, 85, 37 },
		{ 402, 86, 37 },
		{ 402, 87, 37 },
		{ 402, 88, 37 },
		{ 402, 89, 37 },
		{ 402, 90, 37 },
		{ 402, 91, 37 },
		{ 402, 92, 37 },
		{ 402, 93, 37 },
		{ 402, 94, 37 },
		{ 402, 95, 37 },
		{ 402, 96, 37 },
		{ 402, 97, 37 },
		{ 402, 98, 37 },
		{ 402, 99, 37 },
		{ 402, 100, 37 },
		{ 402, 101, 37 },
		{ 402, 102, 37 },
		{ 402, 103, 37 },
		{ 402, 104, 37 },
		{ 402, 105, 37 },
		{ 402, 106, 37 },
		{ 402, 107, 37 },
		{ 402, 108, 37 },
		{ 402, 109, 37 },
		{ 402, 110, 37 },
		{ 402, 111, 37 },
		{ 403, 63, 37 },
		{ 403, 64, 37 },
		{ 403, 65, 37 },
		{ 403, 66, 37 },
		{ 403, 67, 37 },
		{ 403, 68, 37 },
		{ 403, 69, 37 },
		{ 403, 70, 37 },
		{ 403, 71, 37 },
		{ 403, 72, 37 },
		{ 403, 73, 37 },
		{ 403, 74, 37 },
		{ 403, 75, 37 },
		{ 403, 76, 37 },
		{ 403, 77, 37 },
		{ 403, 78, 37 },
		{ 403, 79, 37 },
		{ 403, 80, 37 },
		{ 403, 81, 37 },
		{ 403, 82, 37 },
		{ 403, 83, 37 },
		{ 403, 84, 37 },
		{ 403, 85, 37 },
		{ 403, 86, 37 },
		{ 403, 87, 37 },
		{ 403, 88, 37 },
		{ 403, 89, 37 },
		{ 403, 90, 37 },
		{ 403, 91, 37 },
		{ 403, 92, 37 },
		{ 403, 93, 37 },
		{ 403, 94, 37 },
		{ 403, 95, 37 },
		{ 403, 96, 37 },
		{ 403, 97, 37 },
		{ 403, 98, 37 },
		{ 403, 99, 37 },
		{ 403, 100, 37 },
		{ 403, 101, 37 },
		{ 403, 102, 37 },
		{ 403, 103, 37 },
		{ 403, 104, 37 },
		{ 403, 105, 37 },
		{ 403, 106, 37 },
		{ 403, 107, 37 },
		{ 403, 108, 37 },
		{ 403, 109, 37 },
		{ 403, 110, 37 },
		{ 404, 62, 37 },
		{ 404, 63, 37 },
		{ 404, 64, 37 },
		{ 404, 65, 37 },
		{ 404, 66, 37 },
		{ 404, 67, 37 },
		{ 404, 68, 37 },
		{ 404, 69, 37 },
		{ 404, 70, 37 },
		{ 404, 71, 37 },
		{ 404, 72, 37 },
		{ 404, 73, 37 },
		{ 404, 74, 37 },
		{ 404, 75, 37 },
		{ 404, 76, 37 },
		{ 404, 77, 37 },
		{ 404, 78, 37 },
		{ 404, 79, 37 },
		{ 404, 80, 37 },
		{ 404, 81, 37 },
		{ 404, 82, 37 },
		{ 404, 83, 37 },
		{ 404, 84, 37 },
		{ 404, 85, 37 },
		{ 404, 86, 37 },
		{ 404, 87, 37 },
		{ 404, 88, 37 },
		{ 404, 89, 37 },
		{ 404, 90, 37 },
		{ 404, 91, 37 },
		{ 404, 92, 37 },
		{ 404, 93, 37 },
		{ 404, 94, 37 },
		{ 404, 95, 37 },
		{ 404, 96, 37 },
		{ 404, 97, 37 },
		{ 404, 98, 37 },
		{ 404, 99, 37 },
		{ 404, 100, 37 },
		{ 404, 101, 37 },
		{ 404, 102, 37 },
		{ 404, 103, 37 },
		{ 404, 104, 37 },
		{ 404, 105, 37 },
		{ 404, 106, 37 },
		{ 404, 107, 37 },
		{ 404, 108, 37 },
		{ 404, 109, 37 },
		{ 405, 60, 37 },
		{ 405, 61, 37 },
		{ 405, 62, 37 },
		{ 405, 63, 37 },
		{ 405, 64, 37 },
		{ 405, 65, 37 },
		{ 405, 66, 37 },
		{ 405, 67, 37 },
		{ 405, 68, 37 },
		{ 405, 69, 37 },
		{ 405, 70, 37 },
		{ 405, 71, 37 },
		{ 405, 72, 37 },
		{ 405, 73, 37 },
		{ 405, 74, 37 },
		{ 405, 75, 37 },
		{ 405, 76, 37 },
		{ 405, 77, 37 },
		{ 405, 78, 37 },
		{ 405, 79, 37 },
		{ 405, 80, 37 },
		{ 405, 81, 37 },
		{ 405, 82, 37 },
		{ 405, 83, 37 },
		{ 405, 84, 37 },
		{ 405, 85, 37 },
		{ 405, 86, 37 },
		{ 405, 87, 37 },
		{ 405, 88, 37 },
		{ 405, 89, 37 },
		{ 405, 90, 37 },
		{ 405, 91, 37 },
		{ 405, 92, 37 },
		{ 405, 93, 37 },
		{ 405, 94, 37 },
		{ 405, 95, 37 },
		{ 405, 96, 37 },
		{ 405, 97, 37 },
		{ 405, 98, 37 },
		{ 405, 99, 37 },
		{ 405, 100, 37 },
		{ 405, 101, 37 },
		{ 405, 102, 37 },
		{ 405, 103, 37 },
		{ 405, 104, 37 },
		{ 405, 105, 37 },
		{ 405, 106, 37 },
		{ 405, 107, 37 },
		{ 405, 108, 37 },
		{ 406, 59, 37 },
		{ 406, 60, 37 },
		{ 406, 61, 37 },
		{ 406, 62, 37 },
		{ 406, 63, 37 },
		{ 406, 64, 37 },
		{ 406, 65, 37 },
		{ 406, 66, 37 },
		{ 406, 67, 37 },
		{ 406, 68, 37 },
		{ 406, 69, 37 },
		{ 406, 70, 37 },
		{ 406, 71, 37 },
		{ 406, 72, 37 },
		{ 406, 73, 37 },
		{ 406, 74, 37 },
		{ 406, 75, 37 },
		{ 406, 76, 37 },
		{ 406, 77, 37 },
		{ 406, 78, 37 },
		{ 406, 79, 37 },
		{ 406, 80, 37 },
		{ 406, 81, 37 },
		{ 406, 82, 37 },
		{ 406, 83, 37 },
		{ 406, 84, 37 },
		{ 406, 85, 37 },
		{ 406, 86, 37 },
		{ 406, 87, 37 },
		{ 406, 88, 37 },
		{ 406, 89, 37 },
		{ 406, 90, 37 },
		{ 406, 91, 37 },
		{ 406, 92, 37 },
		{ 406, 93, 37 },
		{ 406, 94, 37 },
		{ 406, 95, 37 },
		{ 406, 96, 37 },
		{ 406, 97, 37 },
		{ 406, 98, 37 },
		{ 406, 99, 37 },
		{ 406, 100, 37 },
		{ 406, 101, 37 },
		{ 406, 102, 37 },
		{ 406, 103, 37 },
		{ 406, 104, 37 },
		{ 406, 105, 37 },
		{ 406, 106, 37 },
		{ 406, 107, 37 },
		{ 407, 58, 37 },
		{ 407, 59, 37 },
		{ 407, 60, 37 },
		{ 407, 61, 37 },
		{ 407, 62, 37 },
		{ 407, 63, 37 },
		{ 407, 64, 37 },
		{ 407, 65, 37 },
		{ 407, 66, 37 },
		{ 407, 67, 37 },
		{ 407, 68, 37 },
		{ 407, 69, 37 },
		{ 407, 70, 37 },
		{ 407, 71, 37 },
		{ 407, 72, 37 },
		{ 407, 73, 37 },
		{ 407, 74, 37 },
		{ 407, 75, 37 },
		{ 407, 76, 37 },
		{ 407, 77, 37 },
		{ 407, 78, 37 },
		{ 407, 79, 37 },
		{ 407, 80, 37 },
		{ 407, 81, 37 },
		{ 407, 82, 37 },
		{ 407, 83, 37 },
		{ 407, 84, 37 },
		{ 407, 85, 37 },
		{ 407, 86, 37 },
		{ 407, 87, 37 },
		{ 407, 88, 37 },
		{ 407, 89, 37 },
		{ 407, 90, 37 },
		{ 407, 91, 37 },
		{ 407, 92, 37 },
		{ 407, 93, 37 },
		{ 407, 94, 37 },
		{ 407, 95, 37 },
		{ 407, 96, 37 },
		{ 407, 97, 37 },
		{ 407, 98, 37 },
		{ 407, 99, 37 },
		{ 407, 100, 37 },
		{ 407, 101, 37 },
		{ 407, 102, 37 },
		{ 407, 103, 37 },
		{ 407, 104, 37 },
		{ 407, 105, 37 },
		{ 407, 106, 37 },
		{ 408, 56, 37 },
		{ 408, 57, 37 },
		{ 408, 58, 37 },
		{ 408, 59, 37 },
		{ 408, 60, 37 },
		{ 408, 61, 37 },
		{ 408, 62, 37 },
		{ 408, 63, 37 },
		{ 408, 64, 37 },
		{ 408, 65, 37 },
		{ 408, 66, 37 },
		{ 408, 67, 37 },
		{ 408, 68, 37 },
		{ 408, 69, 37 },
		{ 408, 70, 37 },
		{ 408, 71, 37 },
		{ 408, 72, 37 },
		{ 408, 73, 37 },
		{ 408, 74, 37 },
		{ 408, 75, 37 },
		{ 408, 76, 37 },
		{ 408, 77, 37 },
		{ 408, 78, 37 },
		{ 408, 79, 37 },
		{ 408, 80, 37 },
		{ 408, 81, 37 },
		{ 408, 82, 37 },
		{ 408, 83, 37 },
		{ 408, 84, 37 },
		{ 408, 85, 37 },
		{ 408, 86, 37 },
		{ 408, 87, 37 },
		{ 408, 88, 37 },
		{ 408, 89, 37 },
		{ 408, 90, 37 },
		{ 408, 91, 37 },
		{ 408, 92, 37 },
		{ 408, 93, 37 },
		{ 408, 94, 37 },
		{ 408, 95, 37 },
		{ 408, 96, 37 },
		{ 408, 97, 37 },
		{ 408, 98, 37 },
		{ 408, 99, 37 },
		{ 408, 100, 37 },
		{ 408, 101, 37 },
		{ 408, 102, 37 },
		{ 408, 103, 37 },
		{ 408, 104, 37 },
		{ 408, 105, 37 },
		{ 409, 55, 37 },
		{ 409, 56, 37 },
		{ 409, 57, 37 },
		{ 409, 58, 37 },
		{ 409, 59, 37 },
		{ 409, 60, 37 },
		{ 409, 61, 37 },
		{ 409, 62, 37 },
		{ 409, 63, 37 },
		{ 409, 64, 37 },
		{ 409, 65, 37 },
		{ 409, 66, 37 },
		{ 409, 67, 37 },
		{ 409, 68, 37 },
		{ 409, 69, 37 },
		{ 409, 70, 37 },
		{ 409, 71, 37 },
		{ 409, 72, 37 },
		{ 409, 73, 37 },
		{ 409, 74, 37 },
		{ 409, 75, 37 },
		{ 409, 76, 37 },
		{ 409, 77, 37 },
		{ 409, 78, 37 },
		{ 409, 79, 37 },
		{ 409, 80, 37 },
		{ 409, 81, 37 },
		{ 409, 82, 37 },
		{ 409, 83, 37 },
		{ 409, 84, 37 },
		{ 409, 85, 37 },
		{ 409, 86, 37 },
		{ 409, 87, 37 },
		{ 409, 88, 37 },
		{ 409, 89, 37 },
		{ 409, 90, 37 },
		{ 409, 91, 37 },
		{ 409, 92, 37 },
		{ 409, 93, 37 },
		{ 409, 94, 37 },
		{ 409, 95, 37 },
		{ 409, 96, 37 },
		{ 409, 97, 37 },
		{ 409, 98, 37 },
		{ 409, 99, 37 },
		{ 409, 100, 37 },
		{ 409, 101, 37 },
		{ 409, 102, 37 },
		{ 409, 103, 37 },
		{ 409, 104, 37 },
		{ 410, 54, 37 },
		{ 410, 55, 37 },
		{ 410, 56, 37 },
		{ 410, 57, 37 },
		{ 410, 58, 37 },
		{ 410, 59, 37 },
		{ 410, 60, 37 },
		{ 410, 61, 37 },
		{ 410, 62, 37 },
		{ 410, 63, 37 },
		{ 410, 64, 37 },
		{ 410, 65, 37 },
		{ 410, 66, 37 },
		{ 410, 67, 37 },
		{ 410, 68, 37 },
		{ 410, 69, 37 },
		{ 410, 70, 37 },
		{ 410, 71, 37 },
		{ 410, 72, 37 },
		{ 410, 73, 37 },
		{ 410, 74, 37 },
		{ 410, 75, 37 },
		{ 410, 76, 37 },
		{ 410, 77, 37 },
		{ 410, 78, 37 },
		{ 410, 79, 37 },
		{ 410, 80, 37 },
		{ 410, 81, 37 },
		{ 410, 82, 37 },
		{ 410, 83, 37 },
		{ 410, 84, 37 },
		{ 410, 85, 37 },
		{ 410, 86, 37 },
		{ 410, 87, 37 },
		{ 410, 88, 37 },
		{ 410, 89, 37 },
		{ 410, 90, 37 },
		{ 410, 91, 37 },
		{ 410, 92, 37 },
		{ 410, 93, 37 },
		{ 410, 94, 37 },
		{ 410, 95, 37 },
		{ 410, 96, 37 },
		{ 410, 97, 37 },
		{ 410, 98, 37 },
		{ 410, 99, 37 },
		{ 410, 100, 37 },
		{ 410, 101, 37 },
		{ 410, 102, 37 },
		{ 410, 103, 37 },
		{ 411, 52, 37 },
		{ 411, 53, 37 },
		{ 411, 54, 37 },
		{ 411, 55, 37 },
		{ 411, 56, 37 },
		{ 411, 57, 37 },
		{ 411, 58, 37 },
		{ 411, 59, 37 },
		{ 411, 60, 37 },
		{ 411, 61, 37 },
		{ 411, 62, 37 },
		{ 411, 63, 37 },
		{ 411, 64, 37 },
		{ 411, 65, 37 },
		{ 411, 66, 37 },
		{ 411, 67, 37 },
		{ 411, 68, 37 },
		{ 411, 69, 37 },
		{ 411, 70, 37 },
		{ 411, 71, 37 },
		{ 411, 72, 37 },
		{ 411, 73, 37 },
		{ 411, 74, 37 },
		{ 411, 75, 37 },
		{ 411, 76, 37 },
		{ 411, 77, 37 },
		{ 411, 78, 37 },
		{ 411, 79, 37 },
		{ 411, 80, 37 },
		{ 411, 81, 37 },
		{ 411, 82, 37 },
		{ 411, 83, 37 },
		{ 411, 84, 37 },
		{ 411, 85, 37 },
		{ 411, 86, 37 },
		{ 411, 87, 37 },
		{ 411, 88, 37 },
		{ 411, 89, 37 },
		{ 411, 90, 37 },
		{ 411, 91, 37 },
		{ 411, 92, 37 },
		{ 411, 93, 37 },
		{ 411, 94, 37 },
		{ 411, 95, 37 },
		{ 411, 96, 37 },
		{ 411, 97, 37 },
		{ 411, 98, 37 },
		{ 411, 99, 37 },
		{ 411, 100, 37 },
		{ 411, 101, 37 },
		{ 411, 102, 37 },
		{ 412, 51, 37 },
		{ 412, 52, 37 },
		{ 412, 53, 37 },
		{ 412, 54, 37 },
		{ 412, 55, 37 },
		{ 412, 56, 37 },
		{ 412, 57, 37 },
		{ 412, 58, 37 },
		{ 412, 59, 37 },
		{ 412, 60, 37 },
		{ 412, 61, 37 },
		{ 412, 62, 37 },
		{ 412, 63, 37 },
		{ 412, 64, 37 },
		{ 412, 65, 37 },
		{ 412, 66, 37 },
		{ 412, 67, 37 },
		{ 412, 68, 37 },
		{ 412, 69, 37 },
		{ 412, 70, 37 },
		{ 412, 71, 37 },
		{ 412, 72, 37 },
		{ 412, 73, 37 },
		{ 412, 74, 37 },
		{ 412, 75, 37 },
		{ 412, 76, 37 },
		{ 412, 77, 37 },
		{ 412, 78, 37 },
		{ 412, 79, 37 },
		{ 412, 80, 37 },
		{ 412, 81, 37 },
		{ 412, 82, 37 },
		{ 412, 83, 37 },
		{ 412, 84, 37 },
		{ 412, 85, 37 },
		{ 412, 86, 37 },
		{ 412, 87, 37 },
		{ 412, 88, 37 },
		{ 412, 89, 37 },
		{ 412, 90, 37 },
		{ 412, 91, 37 },
		{ 412, 92, 37 },
		{ 412, 93, 37 },
		{ 412, 94, 37 },
		{ 412, 95, 37 },
		{ 412, 96, 37 },
		{ 412, 97, 37 },
		{ 412, 98, 37 },
		{ 412, 99, 37 },
		{ 412, 100, 37 },
		{ 412, 101, 37 },
		{ 413, 50, 37 },
		{ 413, 51, 37 },
		{ 413, 52, 37 },
		{ 413, 53, 37 },
		{ 413, 54, 37 },
		{ 413, 55, 37 },
		{ 413, 56, 37 },
		{ 413, 57, 37 },
		{ 413, 58, 37 },
		{ 413, 59, 37 },
		{ 413, 60, 37 },
		{ 413, 61, 37 },
		{ 413, 62, 37 },
		{ 413, 63, 37 },
		{ 413, 64, 37 },
		{ 413, 65, 37 },
		{ 413, 66, 37 },
		{ 413, 67, 37 },
		{ 413, 68, 37 },
		{ 413, 69, 37 },
		{ 413, 70, 37 },
		{ 413, 71, 37 },
		{ 413, 72, 37 },
		{ 413, 73, 37 },
		{ 413, 74, 37 },
		{ 413, 75, 37 },
		{ 413, 76, 37 },
		{ 413, 77, 37 },
		{ 413, 78, 37 },
		{ 413, 79, 37 },
		{ 413, 80, 37 },
		{ 413, 81, 37 },
		{ 413, 82, 37 },
		{ 413, 83, 37 },
		{ 413, 84, 37 },
		{ 413, 85, 37 },
		{ 413, 86, 37 },
		{ 413, 87, 37 },
		{ 413, 88, 37 },
		{ 413, 89, 37 },
		{ 413, 90, 37 },
		{ 413, 91, 37 },
		{ 413, 92, 37 },
		{ 413, 93, 37 },
		{ 413, 94, 37 },
		{ 413, 95, 37 },
		{ 413, 96, 37 },
		{ 413, 97, 37 },
		{ 413, 98, 37 },
		{ 413, 99, 37 },
		{ 413, 100, 37 },
		{ 414, 48, 37 },
		{ 414, 49, 37 },
		{ 414, 50, 37 },
		{ 414, 51, 37 },
		{ 414, 52, 37 },
		{ 414, 53, 37 },
		{ 414, 54, 37 },
		{ 414, 55, 37 },
		{ 414, 56, 37 },
		{ 414, 57, 37 },
		{ 414, 58, 37 },
		{ 414, 59, 37 },
		{ 414, 60, 37 },
		{ 414, 61, 37 },
		{ 414, 62, 37 },
		{ 414, 63, 37 },
		{ 414, 64, 37 },
		{ 414, 65, 37 },
		{ 414, 66, 37 },
		{ 414, 67, 37 },
		{ 414, 68, 37 },
		{ 414, 69, 37 },
		{ 414, 70, 37 },
		{ 414, 71, 37 },
		{ 414, 72, 37 },
		{ 414, 73, 37 },
		{ 414, 74, 37 },
		{ 414, 75, 37 },
		{ 414, 76, 37 },
		{ 414, 77, 37 },
		{ 414, 78, 37 },
		{ 414, 79, 37 },
		{ 414, 80, 37 },
		{ 414, 81, 37 },
		{ 414, 82, 37 },
		{ 414, 83, 37 },
		{ 414, 84, 37 },
		{ 414, 85, 37 },
		{ 414, 86, 37 },
		{ 414, 87, 37 },
		{ 414, 88, 37 },
		{ 414, 89, 37 },
		{ 414, 90, 37 },
		{ 414, 91, 37 },
		{ 414, 92, 37 },
		{ 414, 93, 37 },
		{ 414, 94, 37 },
		{ 414, 95, 37 },
		{ 414, 96, 37 },
		{ 414, 97, 37 },
		{ 414, 98, 37 },
		{ 414, 99, 37 },
		{ 415, 47, 37 },
		{ 415, 48, 37 },
		{ 415, 49, 37 },
		{ 415, 50, 37 },
		{ 415, 51, 37 },
		{ 415, 52, 37 },
		{ 415, 53, 37 },
		{ 415, 54, 37 },
		{ 415, 55, 37 },
		{ 415, 56, 37 },
		{ 415, 57, 37 },
		{ 415, 58, 37 },
		{ 415, 59, 37 },
		{ 415, 60, 37 },
		{ 415, 61, 37 },
		{ 415, 62, 37 },
		{ 415, 63, 37 },
		{ 415, 64, 37 },
		{ 415, 65, 37 },
		{ 415, 66, 37 },
		{ 415, 67, 37 },
		{ 415, 68, 37 },
		{ 415, 69, 37 },
		{ 415, 70, 37 },
		{ 415, 71, 37 },
		{ 415, 72, 37 },
		{ 415, 73, 37 },
		{ 415, 74, 37 },
		{ 415, 75, 37 },
		{ 415, 76, 37 },
		{ 415, 77, 37 },
		{ 415, 78, 37 },
		{ 415, 79, 37 },
		{ 415, 80, 37 },
		{ 415, 81, 37 },
		{ 415, 82, 37 },
		{ 415, 83, 37 },
		{ 415, 84, 37 },
		{ 415, 85, 37 },
		{ 415, 86, 37 },
		{ 415, 87, 37 },
		{ 415, 88, 37 },
		{ 415, 89, 37 },
		{ 415, 90, 37 },
		{ 415, 91, 37 },
		{ 415, 92, 37 },
		{ 415, 93, 37 },
		{ 415, 94, 37 },
		{ 415, 95, 37 },
		{ 415, 96, 37 },
		{ 415, 97, 37 },
		{ 415, 98, 37 },
		{ 416, 46, 37 },
		{ 416, 47, 37 },
		{ 416, 48, 37 },
		{ 416, 49, 37 },
		{ 416, 50, 37 },
		{ 416, 51, 37 },
		{ 416, 52, 37 },
		{ 416, 53, 37 },
		{ 416, 54, 37 },
		{ 416, 55, 37 },
		{ 416, 56, 37 },
		{ 416, 57, 37 },
		{ 416, 58, 37 },
		{ 416, 59, 37 },
		{ 416, 60, 37 },
		{ 416, 61, 37 },
		{ 416, 62, 37 },
		{ 416, 63, 37 },
		{ 416, 64, 37 },
		{ 416, 65, 37 },
		{ 416, 66, 37 },
		{ 416, 67, 37 },
		{ 416, 68, 37 },
		{ 416, 69, 37 },
		{ 416, 70, 37 },
		{ 416, 71, 37 },
		{ 416, 72, 37 },
		{ 416, 73, 37 },
		{ 416, 74, 37 },
		{ 416, 75, 37 },
		{ 416, 76, 37 },
		{ 416, 77, 37 },
		{ 416, 78, 37 },
		{ 416, 79, 37 },
		{ 416, 80, 37 },
		{ 416, 81, 37 },
		{ 416, 82, 37 },
		{ 416, 83, 37 },
		{ 416, 84, 37 },
		{ 416, 85, 37 },
		{ 416, 86, 37 },
		{ 416, 87, 37 },
		{ 416, 88, 37 },
		{ 416, 89, 37 },
		{ 416, 90, 37 },
		{ 416, 91, 37 },
		{ 416, 92, 37 },
		{ 416, 93, 37 },
		{ 416, 94, 37 },
		{ 416, 95, 37 },
		{ 416, 96, 37 },
		{ 416, 97, 37 },
		{ 417, 44, 37 },
		{ 417, 45, 37 },
		{ 417, 46, 37 },
		{ 417, 47, 37 },
		{ 417, 48, 37 },
		{ 417, 49, 37 },
		{ 417, 50, 37 },
		{ 417, 51, 37 },
		{ 417, 52, 37 },
		{ 417, 53, 37 },
		{ 417, 54, 37 },
		{ 417, 55, 37 },
		{ 417, 56, 37 },
		{ 417, 57, 37 },
		{ 417, 58, 37 },
		{ 417, 59, 37 },
		{ 417, 60, 37 },
		{ 417, 61, 37 },
		{ 417, 62, 37 },
		{ 417, 63, 37 },
		{ 417, 64, 37 },
		{ 417, 65, 37 },
		{ 417, 66, 37 },
		{ 417, 67, 37 },
		{ 417, 68, 37 },
		{ 417, 69, 37 },
		{ 417, 70, 37 },
		{ 417, 71, 37 },
		{ 417, 72, 37 },
		{ 417, 73, 37 },
		{ 417, 74, 37 },
		{ 417, 75, 37 },
		{ 417, 76, 37 },
		{ 417, 77, 37 },
		{ 417, 78, 37 },
		{ 417, 79, 37 },
		{ 417, 80, 37 },
		{ 417, 81, 37 },
		{ 417, 82, 37 },
		{ 417, 83, 37 },
		{ 417, 84, 37 },
		{ 417, 85, 37 },
		{ 417, 86, 37 },
		{ 417, 87, 37 },
		{ 417, 88, 37 },
		{ 417, 89, 37 },
		{ 417, 90, 37 },
		{ 417, 91, 37 },
		{ 417, 92, 37 },
		{ 417, 93, 37 },
		{ 417, 94, 37 },
		{ 417, 95, 37 },
		{ 417, 96, 37 },
		{ 418, 43, 37 },
		{ 418, 44, 37 },
		{ 418, 45, 37 },
		{ 418, 46, 37 },
		{ 418, 47, 37 },
		{ 418, 48, 37 },
		{ 418, 49, 37 },
		{ 418, 50, 37 },
		{ 418, 51, 37 },
		{ 418, 52, 37 },
		{ 418, 53, 37 },
		{ 418, 54, 37 },
		{ 418, 55, 37 },
		{ 418, 56, 37 },
		{ 418, 57, 37 },
		{ 418, 58, 37 },
		{ 418, 59, 37 },
		{ 418, 60, 37 },
		{ 418, 61, 37 },
		{ 418, 62, 37 },
		{ 418, 63, 37 },
		{ 418, 64, 37 },
		{ 418, 65, 37 },
		{ 418, 66, 37 },
		{ 418, 67, 37 },
		{ 418, 68, 37 },
		{ 418, 69, 37 },
		{ 418, 70, 37 },
		{ 418, 71, 37 },
		{ 418, 72, 37 },
		{ 418, 73, 37 },
		{ 418, 74, 37 },
		{ 418, 75, 37 },
		{ 418, 76, 37 },
		{ 418, 77, 37 },
		{ 418, 78, 37 },
		{ 418, 79, 37 },
		{ 418, 80, 37 },
		{ 418, 81, 37 },
		{ 418, 82, 37 },
		{ 418, 83, 37 },
		{ 418, 84, 37 },
		{ 418, 85, 37 },
		{ 418, 86, 37 },
		{ 418, 87, 37 },
		{ 418, 88, 37 },
		{ 418, 89, 37 },
		{ 418, 90, 37 },
		{ 418, 91, 37 },
		{ 418, 92, 37 },
		{ 418, 93, 37 },
		{ 418, 94, 37 },
		{ 418, 95, 37 },
		{ 419, 42, 37 },
		{ 419, 43, 37 },
		{ 419, 44, 37 },
		{ 419, 45, 37 },
		{ 419, 46, 37 },
		{ 419, 47, 37 },
		{ 419, 48, 37 },
		{ 419, 49, 37 },
		{ 419, 50, 37 },
		{ 419, 51, 37 },
		{ 419, 52, 37 },
		{ 419, 53, 37 },
		{ 419, 54, 37 },
		{ 419, 55, 37 },
		{ 419, 56, 37 },
		{ 419, 57, 37 },
		{ 419, 58, 37 },
		{ 419, 59, 37 },
		{ 419, 60, 37 },
		{ 419, 61, 37 },
		{ 419, 62, 37 },
		{ 419, 63, 37 },
		{ 419, 64, 37 },
		{ 419, 65, 37 },
		{ 419, 66, 37 },
		{ 419, 67, 37 },
		{ 419, 68, 37 },
		{ 419, 69, 37 },
		{ 419, 70, 37 },
		{ 419, 71, 37 },
		{ 419, 72, 37 },
		{ 419, 73, 37 },
		{ 419, 74, 37 },
		{ 419, 75, 37 },
		{ 419, 76, 37 },
		{ 419, 77, 37 },
		{ 419, 78, 37 },
		{ 419, 79, 37 },
		{ 419, 80, 37 },
		{ 419, 81, 37 },
		{ 419, 82, 37 },
		{ 419, 83, 37 },
		{ 419, 84, 37 },
		{ 419, 85, 37 },
		{ 419, 86, 37 },
		{ 419, 87, 37 },
		{ 419, 88, 37 },
		{ 419, 89, 37 },
		{ 419, 90, 37 },
		{ 419, 91, 37 },
		{ 419, 92, 37 },
		{ 419, 93, 37 },
		{ 419, 94, 37 },
		{ 420, 40, 37 },
		{ 420, 41, 37 },
		{ 420, 42, 37 },
		{ 420, 43, 37 },
		{ 420, 44, 37 },
		{ 420, 45, 37 },
		{ 420, 46, 37 },
		{ 420, 47, 37 },
		{ 420, 48, 37 },
		{ 420, 49, 37 },
		{ 420, 50, 37 },
		{ 420, 51, 37 },
		{ 420, 52, 37 },
		{ 420, 53, 37 },
		{ 420, 54, 37 },
		{ 420, 55, 37 },
		{ 420, 56, 37 },
		{ 420, 57, 37 },
		{ 420, 58, 37 },
		{ 420, 59, 37 },
		{ 420, 60, 37 },
		{ 420, 61, 37 },
		{ 420, 62, 37 },
		{ 420, 63, 37 },
		{ 420, 64, 37 },
		{ 420, 65, 37 },
		{ 420, 66, 37 },
		{ 420, 67, 37 },
		{ 420, 68, 37 },
		{ 420, 69, 37 },
		{ 420, 70, 37 },
		{ 420, 71, 37 },
		{ 420, 72, 37 },
		{ 420, 73, 37 },
		{ 420, 74, 37 },
		{ 420, 75, 37 },
		{ 420, 76, 37 },
		{ 420, 77, 37 },
		{ 420, 78, 37 },
		{ 420, 79, 37 },
		{ 420, 80, 37 },
		{ 420, 81, 37 },
		{ 420, 82, 37 },
		{ 420, 83, 37 },
		{ 420, 84, 37 },
		{ 420, 85, 37 },
		{ 420, 86, 37 },
		{ 420, 87, 37 },
		{ 420, 88, 37 },
		{ 420, 89, 37 },
		{ 420, 90, 37 },
		{ 420, 91, 37 },
		{ 420, 92, 37 },
		{ 420, 93, 37 },
		{ 421, 39, 37 },
		{ 421, 40, 37 },
		{ 421, 41, 37 },
		{ 421, 42, 37 },
		{ 421, 43, 37 },
		{ 421, 44, 37 },
		{ 421, 45, 37 },
		{ 421, 46, 37 },
		{ 421, 47, 37 },
		{ 421, 48, 37 },
		{ 421, 49, 37 },
		{ 421, 50, 37 },
		{ 421, 51, 37 },
		{ 421, 52, 37 },
		{ 421, 53, 37 },
		{ 421, 54, 37 },
		{ 421, 55, 37 },
		{ 421, 56, 37 },
		{ 421, 57, 37 },
		{ 421, 58, 37 },
		{ 421, 59, 37 },
		{ 421, 60, 37 },
		{ 421, 61, 37 },
		{ 421, 62, 37 },
		{ 421, 63, 37 },
		{ 421, 64, 37 },
		{ 421, 65, 37 },
		{ 421, 66, 37 },
		{ 421, 67, 37 },
		{ 421, 68, 37 },
		{ 421, 69, 37 },
		{ 421, 70, 37 },
		{ 421, 71, 37 },
		{ 421, 72, 37 },
		{ 421, 73, 37 },
		{ 421, 74, 37 },
		{ 421, 75, 37 },
		{ 421, 76, 37 },
		{ 421, 77, 37 },
		{ 421, 78, 37 },
		{ 421, 79, 37 },
		{ 421, 80, 37 },
		{ 421, 81, 37 },
		{ 421, 82, 37 },
		{ 421, 83, 37 },
		{ 421, 84, 37 },
		{ 421, 85, 37 },
		{ 421, 86, 37 },
		{ 421, 87, 37 },
		{ 421, 88, 37 },
		{ 421, 89, 37 },
		{ 421, 90, 37 },
		{ 421, 91, 37 },
		{ 421, 92, 37 },
		{ 422, 38, 37 },
		{ 422, 39, 37 },
		{ 422, 40, 37 },
		{ 422, 41, 37 },
		{ 422, 42, 37 },
		{ 422, 43, 37 },
		{ 422, 44, 37 },
		{ 422, 45, 37 },
		{ 422, 46, 37 },
		{ 422, 47, 37 },
		{ 422, 48, 37 },
		{ 422, 49, 37 },
		{ 422, 50, 37 },
		{ 422, 51, 37 },
		{ 422, 52, 37 },
		{ 422, 53, 37 },
		{ 422, 54, 37 },
		{ 422, 55, 37 },
		{ 422, 56, 37 },
		{ 422, 57, 37 },
		{ 422, 58, 37 },
		{ 422, 59, 37 },
		{ 422, 60, 37 },
		{ 422, 61, 37 },
		{ 422, 62, 37 },
		{ 422, 63, 37 },
		{ 422, 64, 37 },
		{ 422, 65, 37 },
		{ 422, 66, 37 },
		{ 422, 67, 37 },
		{ 422, 68, 37 },
		{ 422, 69, 37 },
		{ 422, 70, 37 },
		{ 422, 71, 37 },
		{ 422, 72, 37 },
		{ 422, 73, 37 },
		{ 422, 74, 37 },
		{ 422, 75, 37 },
		{ 422, 76, 37 },
		{ 422, 77, 37 },
		{ 422, 78, 37 },
		{ 422, 79, 37 },
		{ 422, 80, 37 },
		{ 422, 81, 37 },
		{ 422, 82, 37 },
		{ 422, 83, 37 },
		{ 422, 84, 37 },
		{ 422, 85, 37 },
		{ 422, 86, 37 },
		{ 422, 87, 37 },
		{ 422, 88, 37 },
		{ 422, 89, 37 },
		{ 422, 90, 37 },
		{ 422, 91, 37 },
		{ 423, 36, 37 },
		{ 423, 37, 37 },
		{ 423, 38, 37 },
		{ 423, 39, 37 },
		{ 423, 40, 37 },
		{ 423, 41, 37 },
		{ 423, 42, 37 },
		{ 423, 43, 37 },
		{ 423, 44, 37 },
		{ 423, 45, 37 },
		{ 423, 46, 37 },
		{ 423, 47, 37 },
		{ 423, 48, 37 },
		{ 423, 49, 37 },
		{ 423, 50, 37 },
		{ 423, 51, 37 },
		{ 423, 52, 37 },
		{ 423, 53, 37 },
		{ 423, 54, 37 },
		{ 423, 55, 37 },
		{ 423, 56, 37 },
		{ 423, 57, 37 },
		{ 423, 58, 37 },
		{ 423, 59, 37 },
		{ 423, 60, 37 },
		{ 423, 61, 37 },
		{ 423, 62, 37 },
		{ 423, 63, 37 },
		{ 423, 64, 37 },
		{ 423, 65, 37 },
		{ 423, 66, 37 },
		{ 423, 67, 37 },
		{ 423, 68, 37 },
		{ 423, 69, 37 },
		{ 423, 70, 37 },
		{ 423, 71, 37 },
		{ 423, 72, 37 },
		{ 423, 73, 37 },
		{ 423, 74, 37 },
		{ 423, 75, 37 },
		{ 423, 76, 37 },
		{ 423, 77, 37 },
		{ 423, 78, 37 },
		{ 423, 79, 37 },
		{ 423, 80, 37 },
		{ 423, 81, 37 },
		{ 423, 82, 37 },
		{ 423, 83, 37 },
		{ 423, 84, 37 },
		{ 423, 85, 37 },
		{ 423, 86, 37 },
		{ 423, 87, 37 },
		{ 423, 88, 37 },
		{ 423, 89, 37 },
		{ 423, 90, 37 },
		{ 424, 35, 37 },
		{ 424, 36, 37 },
		{ 424, 37, 37 },
		{ 424, 38, 37 },
		{ 424, 39, 37 },
		{ 424, 40, 37 },
		{ 424, 41, 37 },
		{ 424, 42, 37 },
		{ 424, 43, 37 },
		{ 424, 44, 37 },
		{ 424, 45, 37 },
		{ 424, 46, 37 },
		{ 424, 47, 37 },
		{ 424, 48, 37 },
		{ 424, 49, 37 },
		{ 424, 50, 37 },
		{ 424, 51, 37 },
		{ 424, 52, 37 },
		{ 424, 53, 37 },
		{ 424, 54, 37 },
		{ 424, 55, 37 },
		{ 424, 56, 37 },
		{ 424, 57, 37 },
		{ 424, 58, 37 },
		{ 424, 59, 37 },
		{ 424, 60, 37 },
		{ 424, 61, 37 },
		{ 424, 62, 37 },
		{ 424, 63, 37 },
		{ 424, 64, 37 },
		{ 424, 65, 37 },
		{ 424, 66, 37 },
		{ 424, 67, 37 },
		{ 424, 68, 37 },
		{ 424, 69, 37 },
		{ 424, 70, 37 },
		{ 424, 71, 37 },
		{ 424, 72, 37 },
		{ 424, 73, 37 },
		{ 424, 74, 37 },
		{ 424, 75, 37 },
		{ 424, 76, 37 },
		{ 424, 77, 37 },
		{ 424, 78, 37 },
		{ 424, 79, 37 },
		{ 424, 80, 37 },
		{ 424, 81, 37 },
		{ 424, 82, 37 },
		{ 424, 83, 37 },
		{ 424, 84, 37 },
		{ 424, 85, 37 },
		{ 424, 86, 37 },
		{ 424, 87, 37 },
		{ 424, 88, 37 },
		{ 424, 89, 37 },
		{ 425, 34, 37 },
		{ 425, 35, 37 },
		{ 425, 36, 37 },
		{ 425, 37, 37 },
		{ 425, 38, 37 },
		{ 425, 39, 37 },
		{ 425, 40, 37 },
		{ 425, 41, 37 },
		{ 425, 42, 37 },
		{ 425, 43, 37 },
		{ 425, 44, 37 },
		{ 425, 45, 37 },
		{ 425, 46, 37 },
		{ 425, 47, 37 },
		{ 425, 48, 37 },
		{ 425, 49, 37 },
		{ 425, 50, 37 },
		{ 425, 51, 37 },
		{ 425, 52, 37 },
		{ 425, 53, 37 },
		{ 425, 54, 37 },
		{ 425, 55, 37 },
		{ 425, 56, 37 },
		{ 425, 57, 37 },
		{ 425, 58, 37 },
		{ 425, 59, 37 },
		{ 425, 60, 37 },
		{ 425, 61, 37 },
		{ 425, 62, 37 },
		{ 425, 63, 37 },
		{ 425, 64, 37 },
		{ 425, 65, 37 },
		{ 425, 66, 37 },
		{ 425, 67, 37 },
		{ 425, 68, 37 },
		{ 425, 69, 37 },
		{ 425, 70, 37 },
		{ 425, 71, 37 },
		{ 425, 72, 37 },
		{ 425, 73, 37 },
		{ 425, 74, 37 },
		{ 425, 75, 37 },
		{ 425, 76, 37 },
		{ 425, 77, 37 },
		{ 425, 78, 37 },
		{ 425, 79, 37 },
		{ 425, 80, 37 },
		{ 425, 81, 37 },
		{ 425, 82, 37 },
		{ 425, 83, 37 },
		{ 425, 84, 37 },
		{ 425, 85, 37 },
		{ 425, 86, 37 },
		{ 425, 87, 37 },
		{ 425, 88, 37 },
		{ 426, 32, 37 },
		{ 426, 33, 37 },
		{ 426, 34, 37 },
		{ 426, 35, 37 },
		{ 426, 36, 37 },
		{ 426, 37, 37 },
		{ 426, 38, 37 },
		{ 426, 39, 37 },
		{ 426, 40, 37 },
		{ 426, 41, 37 },
		{ 426, 42, 37 },
		{ 426, 43, 37 },
		{ 426, 44, 37 },
		{ 426, 45, 37 },
		{ 426, 46, 37 },
		{ 426, 47, 37 },
		{ 426, 48, 37 },
		{ 426, 49, 37 },
		{ 426, 50, 37 },
		{ 426, 51, 37 },
		{ 426, 52, 37 },
		{ 426, 53, 37 },
		{ 426, 54, 37 },
		{ 426, 55, 37 },
		{ 426, 56, 37 },
		{ 426, 57, 37 },
		{ 426, 58, 37 },
		{ 426, 59, 37 },
		{ 426, 60, 37 },
		{ 426, 61, 37 },
		{ 426, 62, 37 },
		{ 426, 63, 37 },
		{ 426, 64, 37 },
		{ 426, 65, 37 },
		{ 426, 66, 37 },
		{ 426, 67, 37 },
		{ 426, 68, 37 },
		{ 426, 69, 37 },
		{ 426, 70, 37 },
		{ 426, 71, 37 },
		{ 426, 72, 37 },
		{ 426, 73, 37 },
		{ 426, 74, 37 },
		{ 426, 75, 37 },
		{ 426, 76, 37 },
		{ 426, 77, 37 },
		{ 426, 78, 37 },
		{ 426, 79, 37 },
		{ 426, 80, 37 },
		{ 426, 81, 37 },
		{ 426, 82, 37 },
		{ 426, 83, 37 },
		{ 426, 84, 37 },
		{ 426, 85, 37 },
		{ 426, 86, 37 },
		{ 426, 87, 37 },
		{ 427, 31, 37 },
		{ 427, 32, 37 },
		{ 427, 33, 37 },
		{ 427, 34, 37 },
		{ 427, 35, 37 },
		{ 427, 36, 37 },
		{ 427, 37, 37 },
		{ 427, 38, 37 },
		{ 427, 39, 37 },
		{ 427, 40, 37 },
		{ 427, 41, 37 },
		{ 427, 42, 37 },
		{ 427, 43, 37 },
		{ 427, 44, 37 },
		{ 427, 45, 37 },
		{ 427, 46, 37 },
		{ 427, 47, 37 },
		{ 427, 48, 37 },
		{ 427, 49, 37 },
		{ 427, 50, 37 },
		{ 427, 51, 37 },
		{ 427, 52, 37 },
		{ 427, 53, 37 },
		{ 427, 54, 37 },
		{ 427, 55, 37 },
		{ 427, 56, 37 },
		{ 427, 57, 37 },
		{ 427, 58, 37 },
		{ 427, 59, 37 },
		{ 427, 60, 37 },
		{ 427, 61, 37 },
		{ 427, 62, 37 },
		{ 427, 63, 37 },
		{ 427, 64, 37 },
		{ 427, 65, 37 },
		{ 427, 66, 37 },
		{ 427, 67, 37 },
		{ 427, 68, 37 },
		{ 427, 69, 37 },
		{ 427, 70, 37 },
		{ 427, 71, 37 },
		{ 427, 72, 37 },
		{ 427, 73, 37 },
		{ 427, 74, 37 },
		{ 427, 75, 37 },
		{ 427, 76, 37 },
		{ 427, 77, 37 },
		{ 427, 78, 37 },
		{ 427, 79, 37 },
		{ 427, 80, 37 },
		{ 427, 81, 37 },
		{ 427, 82, 37 },
		{ 427, 83, 37 },
		{ 427, 84, 37 },
		{ 427, 85, 37 },
		{ 427, 86, 37 },
		{ 428, 30, 37 },
		{ 428, 31, 37 },
		{ 428, 32, 37 },
		{ 428, 33, 37 },
		{ 428, 34, 37 },
		{ 428, 35, 37 },
		{ 428, 36, 37 },
		{ 428, 37, 37 },
		{ 428, 38, 37 },
		{ 428, 39, 37 },
		{ 428, 40, 37 },
		{ 428, 41, 37 },
		{ 428, 42, 37 },
		{ 428, 43, 37 },
		{ 428, 44, 37 },
		{ 428, 45, 37 },
		{ 428, 46, 37 },
		{ 428, 47, 37 },
		{ 428, 48, 37 },
		{ 428, 49, 37 },
		{ 428, 50, 37 },
		{ 428, 51, 37 },
		{ 428, 52, 37 },
		{ 428, 53, 37 },
		{ 428, 54, 37 },
		{ 428, 55, 37 },
		{ 428, 56, 37 },
		{ 428, 57, 37 },
		{ 428, 58, 37 },
		{ 428, 59, 37 },
		{ 428, 60, 37 },
		{ 428, 61, 37 },
		{ 428, 62, 37 },
		{ 428, 63, 37 },
		{ 428, 64, 37 },
		{ 428, 65, 37 },
		{ 428, 66, 37 },
		{ 428, 67, 37 },
		{ 428, 68, 37 },
		{ 428, 69, 37 },
		{ 428, 70, 37 },
		{ 428, 71, 37 },
		{ 428, 72, 37 },
		{ 428, 73, 37 },
		{ 428, 74, 37 },
		{ 428, 75, 37 },
		{ 428, 76, 37 },
		{ 428, 77, 37 },
		{ 428, 78, 37 },
		{ 428, 79, 37 },
		{ 428, 80, 37 },
		{ 428, 81, 37 },
		{ 428, 82, 37 },
		{ 428, 83, 37 },
		{ 428, 84, 37 },
		{ 428, 85, 37 },
		{ 429, 28, 37 },
		{ 429, 29, 37 },
		{ 429, 30, 37 },
		{ 429, 31, 37 },
		{ 429, 32, 37 },
		{ 429, 33, 37 },
		{ 429, 34, 37 },
		{ 429, 35, 37 },
		{ 429, 36, 37 },
		{ 429, 37, 37 },
		{ 429, 38, 37 },
		{ 429, 39, 37 },
		{ 429, 40, 37 },
		{ 429, 41, 37 },
		{ 429, 42, 37 },
		{ 429, 43, 37 },
		{ 429, 44, 37 },
		{ 429, 45, 37 },
		{ 429, 46, 37 },
		{ 429, 47, 37 },
		{ 429, 48, 37 },
		{ 429, 49, 37 },
		{ 429, 50, 37 },
		{ 429, 51, 37 },
		{ 429, 52, 37 },
		{ 429, 53, 37 },
		{ 429, 54, 37 },
		{ 429, 55, 37 },
		{ 429, 56, 37 },
		{ 429, 57, 37 },
		{ 429, 58, 37 },
		{ 429, 59, 37 },
		{ 429, 60, 37 },
		{ 429, 61, 37 },
		{ 429, 62, 37 },
		{ 429, 63, 37 },
		{ 429, 64, 37 },
		{ 429, 65, 37 },
		{ 429, 66, 37 },
		{ 429, 67, 37 },
		{ 429, 68, 37 },
		{ 429, 69, 37 },
		{ 429, 70, 37 },
		{ 429, 71, 37 },
		{ 429, 72, 37 },
		{ 429, 73, 37 },
		{ 429, 74, 37 },
		{ 429, 75, 37 },
		{ 429, 76, 37 },
		{ 429, 77, 37 },
		{ 429, 78, 37 },
		{ 429, 79, 37 },
		{ 429, 80, 37 },
		{ 429, 81, 37 },
		{ 429, 82, 37 },
		{ 429, 83, 37 },
		{ 429, 84, 37 },
		{ 430, 27, 37 },
		{ 430, 28, 37 },
		{ 430, 29, 37 },
		{ 430, 30, 37 },
		{ 430, 31, 37 },
		{ 430, 32, 37 },
		{ 430, 33, 37 },
		{ 430, 34, 37 },
		{ 430, 35, 37 },
		{ 430, 36, 37 },
		{ 430, 37, 37 },
		{ 430, 38, 37 },
		{ 430, 39, 37 },
		{ 430, 40, 37 },
		{ 430, 41, 37 },
		{ 430, 42, 37 },
		{ 430, 43, 37 },
		{ 430, 44, 37 },
		{ 430, 45, 37 },
		{ 430, 46, 37 },
		{ 430, 47, 37 },
		{ 430, 48, 37 },
		{ 430, 49, 37 },
		{ 430, 50, 37 },
		{ 430, 51, 37 },
		{ 430, 52, 37 },
		{ 430, 53, 37 },
		{ 430, 54, 37 },
		{ 430, 55, 37 },
		{ 430, 56, 37 },
		{ 430, 57, 37 },
		{ 430, 58, 37 },
		{ 430, 59, 37 },
		{ 430, 60, 37 },
		{ 430, 61, 37 },
		{ 430, 62, 37 },
		{ 430, 63, 37 },
		{ 430, 64, 37 },
		{ 430, 65, 37 },
		{ 430, 66, 37 },
		{ 430, 67, 37 },
		{ 430, 68, 37 },
		{ 430, 69, 37 },
		{ 430, 70, 37 },
		{ 430, 71, 37 },
		{ 430, 72, 37 },
		{ 430, 73, 37 },
		{ 430, 74, 37 },
		{ 430, 75, 37 },
		{ 430, 76, 37 },
		{ 430, 77, 37 },
		{ 430, 78, 37 },
		{ 430, 79, 37 },
		{ 430, 80, 37 },
		{ 430, 81, 37 },
		{ 430, 82, 37 },
		{ 430, 83, 37 },
		{ 431, 26, 37 },
		{ 431, 27, 37 },
		{ 431, 28, 37 },
		{ 431, 29, 37 },
		{ 431, 30, 37 },
		{ 431, 31, 37 },
		{ 431, 32, 37 },
		{ 431, 33, 37 },
		{ 431, 34, 37 },
		{ 431, 35, 37 },
		{ 431, 36, 37 },
		{ 431, 37, 37 },
		{ 431, 38, 37 },
		{ 431, 39, 37 },
		{ 431, 40, 37 },
		{ 431, 41, 37 },
		{ 431, 42, 37 },
		{ 431, 43, 37 },
		{ 431, 44, 37 },
		{ 431, 45, 37 },
		{ 431, 46, 37 },
		{ 431, 47, 37 },
		{ 431, 48, 37 },
		{ 431, 49, 37 },
		{ 431, 50, 37 },
		{ 431, 51, 37 },
		{ 431, 52, 37 },
		{ 431, 53, 37 },
		{ 431, 54, 37 },
		{ 431, 55, 37 },
		{ 431, 56, 37 },
		{ 431, 57, 37 },
		{ 431, 58, 37 },
		{ 431, 59, 37 },
		{ 431, 60, 37 },
		{ 431, 61, 37 },
		{ 431, 62, 37 },
		{ 431, 63, 37 },
		{ 431, 64, 37 },
		{ 431, 65, 37 },
		{ 431, 66, 37 },
		{ 431, 67, 37 },
		{ 431, 68, 37 },
		{ 431, 69, 37 },
		{ 431, 70, 37 },
		{ 431, 71, 37 },
		{ 431, 72, 37 },
		{ 431, 73, 37 },
		{ 431, 74, 37 },
		{ 431, 75, 37 },
		{ 431, 76, 37 },
		{ 431, 77, 37 },
		{ 431, 78, 37 },
		{ 431, 79, 37 },
		{ 431, 80, 37 },
		{ 431, 81, 37 },
		{ 431, 82, 37 },
		{ 432, 24, 37 },
		{ 432, 25, 37 },
		{ 432, 26, 37 },
		{ 432, 27, 37 },
		{ 432, 28, 37 },
		{ 432, 29, 37 },
		{ 432, 30, 37 },
		{ 432, 31, 37 },
		{ 432, 32, 37 },
		{ 432, 33, 37 },
		{ 432, 34, 37 },
		{ 432, 35, 37 },
		{ 432, 36, 37 },
		{ 432, 37, 37 },
		{ 432, 38, 37 },
		{ 432, 39, 37 },
		{ 432, 40, 37 },
		{ 432, 41, 37 },
		{ 432, 42, 37 },
		{ 432, 43, 37 },
		{ 432, 44, 37 },
		{ 432, 45, 37 },
		{ 432, 46, 37 },
		{ 432, 47, 37 },
		{ 432, 48, 37 },
		{ 432, 49, 37 },
		{ 432, 50, 37 },
		{ 432, 51, 37 },
		{ 432, 52, 37 },
		{ 432, 53, 37 },
		{ 432, 54, 37 },
		{ 432, 55, 37 },
		{ 432, 56, 37 },
		{ 432, 57, 37 },
		{ 432, 58, 37 },
		{ 432, 59, 37 },
		{ 432, 60, 37 },
		{ 432, 61, 37 },
		{ 432, 62, 37 },
		{ 432, 63, 37 },
		{ 432, 64, 37 },
		{ 432, 65, 37 },
		{ 432, 66, 37 },
		{ 432, 67, 37 },
		{ 432, 68, 37 },
		{ 432, 69, 37 },
		{ 432, 70, 37 },
		{ 432, 71, 37 },
		{ 432, 72, 37 },
		{ 432, 73, 37 },
		{ 432, 74, 37 },
		{ 432, 75, 37 },
		{ 432, 76, 37 },
		{ 432, 77, 37 },
		{ 432, 78, 37 },
		{ 432, 79, 37 },
		{ 432, 80, 37 },
		{ 432, 81, 37 },
		{ 433, 23, 37 },
		{ 433, 24, 37 },
		{ 433, 25, 37 },
		{ 433, 26, 37 },
		{ 433, 27, 37 },
		{ 433, 28, 37 },
		{ 433, 29, 37 },
		{ 433, 30, 37 },
		{ 433, 31, 37 },
		{ 433, 32, 37 },
		{ 433, 33, 37 },
		{ 433, 34, 37 },
		{ 433, 35, 37 },
		{ 433, 36, 37 },
		{ 433, 37, 37 },
		{ 433, 38, 37 },
		{ 433, 39, 37 },
		{ 433, 40, 37 },
		{ 433, 41, 37 },
		{ 433, 42, 37 },
		{ 433, 43, 37 },
		{ 433, 44, 37 },
		{ 433, 45, 37 },
		{ 433, 46, 37 },
		{ 433, 47, 37 },
		{ 433, 48, 37 },
		{ 433, 49, 37 },
		{ 433, 50, 37 },
		{ 433, 51, 37 },
		{ 433, 52, 37 },
		{ 433, 53, 37 },
		{ 433, 54, 37 },
		{ 433, 55, 37 },
		{ 433, 56, 37 },
		{ 433, 57, 37 },
		{ 433, 58, 37 },
		{ 433, 59, 37 },
		{ 433, 60, 37 },
		{ 433, 61, 37 },
		{ 433, 62, 37 },
		{ 433, 63, 37 },
		{ 433, 64, 37 },
		{ 433, 65, 37 },
		{ 433, 66, 37 },
		{ 433, 67, 37 },
		{ 433, 68, 37 },
		{ 433, 69, 37 },
		{ 433, 70, 37 },
		{ 433, 71, 37 },
		{ 433, 72, 37 },
		{ 433, 73, 37 },
		{ 433, 74, 37 },
		{ 433, 75, 37 },
		{ 433, 76, 37 },
		{ 433, 77, 37 },
		{ 433, 78, 37 },
		{ 433, 79, 37 },
		{ 433, 80, 37 },
		{ 434, 22, 37 },
		{ 434, 23, 37 },
		{ 434, 24, 37 },
		{ 434, 25, 37 },
		{ 434, 26, 37 },
		{ 434, 27, 37 },
		{ 434, 28, 37 },
		{ 434, 29, 37 },
		{ 434, 30, 37 },
		{ 434, 31, 37 },
		{ 434, 32, 37 },
		{ 434, 33, 37 },
		{ 434, 34, 37 },
		{ 434, 35, 37 },
		{ 434, 36, 37 },
		{ 434, 37, 37 },
		{ 434, 38, 37 },
		{ 434, 39, 37 },
		{ 434, 40, 37 },
		{ 434, 41, 37 },
		{ 434, 42, 37 },
		{ 434, 43, 37 },
		{ 434, 44, 37 },
		{ 434, 45, 37 },
		{ 434, 46, 37 },
		{ 434, 47, 37 },
		{ 434, 48, 37 },
		{ 434, 49, 37 },
		{ 434, 50, 37 },
		{ 434, 51, 37 },
		{ 434, 52, 37 },
		{ 434, 53, 37 },
		{ 434, 54, 37 },
		{ 434, 55, 37 },
		{ 434, 56, 37 },
		{ 434, 57, 37 },
		{ 434, 58, 37 },
		{ 434, 59, 37 },
		{ 434, 60, 37 },
		{ 434, 61, 37 },
		{ 434, 62, 37 },
		{ 434, 63, 37 },
		{ 434, 64, 37 },
		{ 434, 65, 37 },
		{ 434, 66, 37 },
		{ 434, 67, 37 },
		{ 434, 68, 37 },
		{ 434, 69, 37 },
		{ 434, 70, 37 },
		{ 434, 71, 37 },
		{ 434, 72, 37 },
		{ 434, 73, 37 },
		{ 434, 74, 37 },
		{ 434, 75, 37 },
		{ 434, 76, 37 },
		{ 434, 77, 37 },
		{ 434, 78, 37 },
		{ 434, 79, 37 },
		{ 435, 20, 37 },
		{ 435, 21, 37 },
		{ 435, 22, 37 },
		{ 435, 23, 37 },
		{ 435, 24, 37 },
		{ 435, 25, 37 },
		{ 435, 26, 37 },
		{ 435, 27, 37 },
		{ 435, 28, 37 },
		{ 435, 29, 37 },
		{ 435, 30, 37 },
		{ 435, 31, 37 },
		{ 435, 32, 37 },
		{ 435, 33, 37 },
		{ 435, 34, 37 },
		{ 435, 35, 37 },
		{ 435, 36, 37 },
		{ 435, 37, 37 },
		{ 435, 38, 37 },
		{ 435, 39, 37 },
		{ 435, 40, 37 },
		{ 435, 41, 37 },
		{ 435, 42, 37 },
		{ 435, 43, 37 },
		{ 435, 44, 37 },
		{ 435, 45, 37 },
		{ 435, 46, 37 },
		{ 435, 47, 37 },
		{ 435, 48, 37 },
		{ 435, 49, 37 },
		{ 435, 50, 37 },
		{ 435, 51, 37 },
		{ 435, 52, 37 },
		{ 435, 53, 37 },
		{ 435, 54, 37 },
		{ 435, 55, 37 },
		{ 435, 56, 37 },
		{ 435, 57, 37 },
		{ 435, 58, 37 },
		{ 435, 59, 37 },
		{ 435, 60, 37 },
		{ 435, 61, 37 },
		{ 435, 62, 37 },
		{ 435, 63, 37 },
		{ 435, 64, 37 },
		{ 435, 65, 37 },
		{ 435, 66, 37 },
		{ 435, 67, 37 },
		{ 435, 68, 37 },
		{ 435, 69, 37 },
		{ 435, 70, 37 },
		{ 435, 71, 37 },
		{ 435, 72, 37 },
		{ 435, 73, 37 },
		{ 435, 74, 37 },
		{ 435, 75, 37 },
		{ 435, 76, 37 },
		{ 435, 77, 37 },
		{ 435, 78, 37 },
		{ 436, 19, 37 },
		{ 436, 20, 37 },
		{ 436, 21, 37 },
		{ 436, 22, 37 },
		{ 436, 23, 37 },
		{ 436, 24, 37 },
		{ 436, 25, 37 },
		{ 436, 26, 37 },
		{ 436, 27, 37 },
		{ 436, 28, 37 },
		{ 436, 29, 37 },
		{ 436, 30, 37 },
		{ 436, 31, 37 },
		{ 436, 32, 37 },
		{ 436, 33, 37 },
		{ 436, 34, 37 },
		{ 436, 35, 37 },
		{ 436, 36, 37 },
		{ 436, 37, 37 },
		{ 436, 38, 37 },
		{ 436, 39, 37 },
		{ 436, 40, 37 },
		{ 436, 41, 37 },
		{ 436, 42, 37 },
		{ 436, 43, 37 },
		{ 436, 44, 37 },
		{ 436, 45, 37 },
		{ 436, 46, 37 },
		{ 436, 47, 37 },
		{ 436, 48, 37 },
		{ 436, 49, 37 },
		{ 436, 50, 37 },
		{ 436, 51, 37 },
		{ 436, 52, 37 },
		{ 436, 53, 37 },
		{ 436, 54, 37 },
		{ 436, 55, 37 },
		{ 436, 56, 37 },
		{ 436, 57, 37 },
		{ 436, 58, 37 },
		{ 436, 59, 37 },
		{ 436, 60, 37 },
		{ 436, 61, 37 },
		{ 436, 62, 37 },
		{ 436, 63, 37 },
		{ 436, 64, 37 },
		{ 436, 65, 37 },
		{ 436, 66, 37 },
		{ 436, 67, 37 },
		{ 436, 68, 37 },
		{ 436, 69, 37 },
		{ 436, 70, 37 },
		{ 436, 71, 37 },
		{ 436, 72, 37 },
		{ 436, 73, 37 },
		{ 436, 74, 37 },
		{ 436, 75, 37 },
		{ 436, 76, 37 },
		{ 436, 77, 37 },
		{ 437, 18, 37 },
		{ 437, 19, 37 },
		{ 437, 20, 37 },
		{ 437, 21, 37 },
		{ 437, 22, 37 },
		{ 437, 23, 37 },
		{ 437, 24, 37 },
		{ 437, 25, 37 },
		{ 437, 26, 37 },
		{ 437, 27, 37 },
		{ 437, 28, 37 },
		{ 437, 29, 37 },
		{ 437, 30, 37 },
		{ 437, 31, 37 },
		{ 437, 32, 37 },
		{ 437, 33, 37 },
		{ 437, 34, 37 },
		{ 437, 35, 37 },
		{ 437, 36, 37 },
		{ 437, 37, 37 },
		{ 437, 38, 37 },
		{ 437, 39, 37 },
		{ 437, 40, 37 },
		{ 437, 41, 37 },
		{ 437, 42, 37 },
		{ 437, 43, 37 },
		{ 437, 44, 37 },
		{ 437, 45, 37 },
		{ 437, 46, 37 },
		{ 437, 47, 37 },
		{ 437, 48, 37 },
		{ 437, 49, 37 },
		{ 437, 50, 37 },
		{ 437, 51, 37 },
		{ 437, 52, 37 },
		{ 437, 53, 37 },
		{ 437, 54, 37 },
		{ 437, 55, 37 },
		{ 437, 56, 37 },
		{ 437, 57, 37 },
		{ 437, 58, 37 },
		{ 437, 59, 37 },
		{ 437, 60, 37 },
		{ 437, 61, 37 },
		{ 437, 62, 37 },
		{ 437, 63, 37 },
		{ 437, 64, 37 },
		{ 437, 65, 37 },
		{ 437, 66, 37 },
		{ 437, 67, 37 },
		{ 437, 68, 37 },
		{ 437, 69, 37 },
		{ 437, 70, 37 },
		{ 437, 71, 37 },
		{ 437, 72, 37 },
		{ 437, 73, 37 },
		{ 437, 74, 37 },
		{ 437, 75, 37 },
		{ 437, 76, 37 },
		{ 438, 16, 37 },
		{ 438, 17, 37 },
		{ 438, 18, 37 },
		{ 438, 19, 37 },
		{ 438, 20, 37 },
		{ 438, 21, 37 },
		{ 438, 22, 37 },
		{ 438, 23, 37 },
		{ 438, 24, 37 },
		{ 438, 25, 37 },
		{ 438, 26, 37 },
		{ 438, 27, 37 },
		{ 438, 28, 37 },
		{ 438, 29, 37 },
		{ 438, 30, 37 },
		{ 438, 31, 37 },
		{ 438, 32, 37 },
		{ 438, 33, 37 },
		{ 438, 34, 37 },
		{ 438, 35, 37 },
		{ 438, 36, 37 },
		{ 438, 37, 37 },
		{ 438, 38, 37 },
		{ 438, 39, 37 },
		{ 438, 40, 37 },
		{ 438, 41, 37 },
		{ 438, 42, 37 },
		{ 438, 43, 37 },
		{ 438, 44, 37 },
		{ 438, 45, 37 },
		{ 438, 46, 37 },
		{ 438, 47, 37 },
		{ 438, 48, 37 },
		{ 438, 49, 37 },
		{ 438, 50, 37 },
		{ 438, 51, 37 },
		{ 438, 52, 37 },
		{ 438, 53, 37 },
		{ 438, 54, 37 },
		{ 438, 55, 37 },
		{ 438, 56, 37 },
		{ 438, 57, 37 },
		{ 438, 58, 37 },
		{ 438, 59, 37 },
		{ 438, 60, 37 },
		{ 438, 61, 37 },
		{ 438, 62, 37 },
		{ 438, 63, 37 },
		{ 438, 64, 37 },
		{ 438, 65, 37 },
		{ 438, 66, 37 },
		{ 438, 67, 37 },
		{ 438, 68, 37 },
		{ 438, 69, 37 },
		{ 438, 70, 37 },
		{ 438, 71, 37 },
		{ 438, 72, 37 },
		{ 438, 73, 37 },
		{ 438, 74, 37 },
		{ 438, 75, 37 },
		{ 439, 15, 37 },
		{ 439, 16, 37 },
		{ 439, 17, 37 },
		{ 439, 18, 37 },
		{ 439, 19, 37 },
		{ 439, 20, 37 },
		{ 439, 21, 37 },
		{ 439, 22, 37 },
		{ 439, 23, 37 },
		{ 439, 24, 37 },
		{ 439, 25, 37 },
		{ 439, 26, 37 },
		{ 439, 27, 37 },
		{ 439, 28, 37 },
		{ 439, 29, 37 },
		{ 439, 30, 37 },
		{ 439, 31, 37 },
		{ 439, 32, 37 },
		{ 439, 33, 37 },
		{ 439, 34, 37 },
		{ 439, 35, 37 },
		{ 439, 36, 37 },
		{ 439, 37, 37 },
		{ 439, 38, 37 },
		{ 439, 39, 37 },
		{ 439, 40, 37 },
		{ 439, 41, 37 },
		{ 439, 42, 37 },
		{ 439, 43, 37 },
		{ 439, 44, 37 },
		{ 439, 45, 37 },
		{ 439, 46, 37 },
		{ 439, 47, 37 },
		{ 439, 48, 37 },
		{ 439, 49, 37 },
		{ 439, 50, 37 },
		{ 439, 51, 37 },
		{ 439, 52, 37 },
		{ 439, 53, 37 },
		{ 439, 54, 37 },
		{ 439, 55, 37 },
		{ 439, 56, 37 },
		{ 439, 57, 37 },
		{ 439, 58, 37 },
		{ 439, 59, 37 },
		{ 439, 60, 37 },
		{ 439, 61, 37 },
		{ 439, 62, 37 },
		{ 439, 63, 37 },
		{ 439, 64, 37 },
		{ 439, 65, 37 },
		{ 439, 66, 37 },
		{ 439, 67, 37 },
		{ 439, 68, 37 },
		{ 439, 69, 37 },
		{ 439, 70, 37 },
		{ 439, 71, 37 },
		{ 439, 72, 37 },
		{ 439, 73, 37 },
		{ 439, 74, 37 },
		{ 440, 14, 37 },
		{ 440, 15, 37 },
		{ 440, 16, 37 },
		{ 440, 17, 37 },
		{ 440, 18, 37 },
		{ 440, 19, 37 },
		{ 440, 20, 37 },
		{ 440, 21, 37 },
		{ 440, 22, 37 },
		{ 440, 23, 37 },
		{ 440, 24, 37 },
		{ 440, 25, 37 },
		{ 440, 26, 37 },
		{ 440, 27, 37 },
		{ 440, 28, 37 },
		{ 440, 29, 37 },
		{ 440, 30, 37 },
		{ 440, 31, 37 },
		{ 440, 32, 37 },
		{ 440, 33, 37 },
		{ 440, 34, 37 },
		{ 440, 35, 37 },
		{ 440, 36, 37 },
		{ 440, 37, 37 },
		{ 440, 38, 37 },
		{ 440, 39, 37 },
		{ 440, 40, 37 },
		{ 440, 41, 37 },
		{ 440, 42, 37 },
		{ 440, 43, 37 },
		{ 440, 44, 37 },
		{ 440, 45, 37 },
		{ 440, 46, 37 },
		{ 440, 47, 37 },
		{ 440, 48, 37 },
		{ 440, 49, 37 },
		{ 440, 50, 37 },
		{ 440, 51, 37 },
		{ 440, 52, 37 },
		{ 440, 53, 37 },
		{ 440, 54, 37 },
		{ 440, 55, 37 },
		{ 440, 56, 37 },
		{ 440, 57, 37 },
		{ 440, 58, 37 },
		{ 440, 59, 37 },
		{ 440, 60, 37 },
		{ 440, 61, 37 },
		{ 440, 62, 37 },
		{ 440, 63, 37 },
		{ 440, 64, 37 },
		{ 440, 65, 37 },
		{ 440, 66, 37 },
		{ 440, 67, 37 },
		{ 440, 68, 37 },
		{ 440, 69, 37 },
		{ 440, 70, 37 },
		{ 440, 71, 37 },
		{ 440, 72, 37 },
		{ 440, 73, 37 },
		{ 441, 12, 37 },
		{ 441, 13, 37 },
		{ 441, 14, 37 },
		{ 441, 15, 37 },
		{ 441, 16, 37 },
		{ 441, 17, 37 },
		{ 441, 18, 37 },
		{ 441, 19, 37 },
		{ 441, 20, 37 },
		{ 441, 21, 37 },
		{ 441, 22, 37 },
		{ 441, 23, 37 },
		{ 441, 24, 37 },
		{ 441, 25, 37 },
		{ 441, 26, 37 },
		{ 441, 27, 37 },
		{ 441, 28, 37 },
		{ 441, 29, 37 },
		{ 441, 30, 37 },
		{ 441, 31, 37 },
		{ 441, 32, 37 },
		{ 441, 33, 37 },
		{ 441, 34, 37 },
		{ 441, 35, 37 },
		{ 441, 36, 37 },
		{ 441, 37, 37 },
		{ 441, 38, 37 },
		{ 441, 39, 37 },
		{ 441, 40, 37 },
		{ 441, 41, 37 },
		{ 441, 42, 37 },
		{ 441, 43, 37 },
		{ 441, 44, 37 },
		{ 441, 45, 37 },
		{ 441, 46, 37 },
		{ 441, 47, 37 },
		{ 441, 48, 37 },
		{ 441, 49, 37 },
		{ 441, 50, 37 },
		{ 441, 51, 37 },
		{ 441, 52, 37 },
		{ 441, 53, 37 },
		{ 441, 54, 37 },
		{ 441, 55, 37 },
		{ 441, 56, 37 },
		{ 441, 57, 37 },
		{ 441, 58, 37 },
		{ 441, 59, 37 },
		{ 441, 60, 37 },
		{ 441, 61, 37 },
		{ 441, 62, 37 },
		{ 441, 63, 37 },
		{ 441, 64, 37 },
		{ 441, 65, 37 },
		{ 441, 66, 37 },
		{ 441, 67, 37 },
		{ 441, 68, 37 },
		{ 441, 69, 37 },
		{ 441, 70, 37 },
		{ 441, 71, 37 },
		{ 441, 72, 37 },
		{ 442, 11, 37 },
		{ 442, 12, 37 },
		{ 442, 13, 37 },
		{ 442, 14, 37 },
		{ 442, 15, 37 },
		{ 442, 16, 37 },
		{ 442, 17, 37 },
		{ 442, 18, 37 },
		{ 442, 19, 37 },
		{ 442, 20, 37 },
		{ 442, 21, 37 },
		{ 442, 22, 37 },
		{ 442, 23, 37 },
		{ 442, 24, 37 },
		{ 442, 25, 37 },
		{ 442, 26, 37 },
		{ 442, 27, 37 },
		{ 442, 28, 37 },
		{ 442, 29, 37 },
		{ 442, 30, 37 },
		{ 442, 31, 37 },
		{ 442, 32, 37 },
		{ 442, 33, 37 },
		{ 442, 34, 37 },
		{ 442, 35, 37 },
		{ 442, 36, 37 },
		{ 442, 37, 37 },
		{ 442, 38, 37 },
		{ 442, 39, 37 },
		{ 442, 40, 37 },
		{ 442, 41, 37 },
		{ 442, 42, 37 },
		{ 442, 43, 37 },
		{ 442, 44, 37 },
		{ 442, 45, 37 },
		{ 442, 46, 37 },
		{ 442, 47, 37 },
		{ 442, 48, 37 },
		{ 442, 49, 37 },
		{ 442, 50, 37 },
		{ 442, 51, 37 },
		{ 442, 52, 37 },
		{ 442, 53, 37 },
		{ 442, 54, 37 },
		{ 442, 55, 37 },
		{ 442, 56, 37 },
		{ 442, 57, 37 },
		{ 442, 58, 37 },
		{ 442, 59, 37 },
		{ 442, 60, 37 },
		{ 442, 61, 37 },
		{ 442, 62, 37 },
		{ 442, 63, 37 },
		{ 442, 64, 37 },
		{ 442, 65, 37 },
		{ 442, 66, 37 },
		{ 442, 67, 37 },
		{ 442, 68, 37 },
		{ 442, 69, 37 },
		{ 442, 70, 37 },
		{ 442, 71, 37 },
		{ 443, 10, 37 },
		{ 443, 11, 37 },
		{ 443, 12, 37 },
		{ 443, 13, 37 },
		{ 443, 14, 37 },
		{ 443, 15, 37 },
		{ 443, 16, 37 },
		{ 443, 17, 37 },
		{ 443, 18, 37 },
		{ 443, 19, 37 },
		{ 443, 20, 37 },
		{ 443, 21, 37 },
		{ 443, 22, 37 },
		{ 443, 23, 37 },
		{ 443, 24, 37 },
		{ 443, 25, 37 },
		{ 443, 26, 37 },
		{ 443, 27, 37 },
		{ 443, 28, 37 },
		{ 443, 29, 37 },
		{ 443, 30, 37 },
		{ 443, 31, 37 },
		{ 443, 32, 37 },
		{ 443, 33, 37 },
		{ 443, 34, 37 },
		{ 443, 35, 37 },
		{ 443, 36, 37 },
		{ 443, 37, 37 },
		{ 443, 38, 37 },
		{ 443, 39, 37 },
		{ 443, 40, 37 },
		{ 443, 41, 37 },
		{ 443, 42, 37 },
		{ 443, 43, 37 },
		{ 443, 44, 37 },
		{ 443, 45, 37 },
		{ 443, 46, 37 },
		{ 443, 47, 37 },
		{ 443, 48, 37 },
		{ 443, 49, 37 },
		{ 443, 50, 37 },
		{ 443, 51, 37 },
		{ 443, 52, 37 },
		{ 443, 53, 37 },
		{ 443, 54, 37 },
		{ 443, 55, 37 },
		{ 443, 56, 37 },
		{ 443, 57, 37 },
		{ 443, 58, 37 },
		{ 443, 59, 37 },
		{ 443, 60, 37 },
		{ 443, 61, 37 },
		{ 443, 62, 37 },
		{ 443, 63, 37 },
		{ 443, 64, 37 },
		{ 443, 65, 37 },
		{ 443, 66, 37 },
		{ 443, 67, 37 },
		{ 443, 68, 37 },
		{ 443, 69, 37 },
		{ 443, 70, 37 },
		{ 444, 8, 37 },
		{ 444, 9, 37 },
		{ 444, 10, 37 },
		{ 444, 11, 37 },
		{ 444, 12, 37 },
		{ 444, 13, 37 },
		{ 444, 14, 37 },
		{ 444, 15, 37 },
		{ 444, 16, 37 },
		{ 444, 17, 37 },
		{ 444, 18, 37 },
		{ 444, 19, 37 },
		{ 444, 20, 37 },
		{ 444, 21, 37 },
		{ 444, 22, 37 },
		{ 444, 23, 37 },
		{ 444, 24, 37 },
		{ 444, 25, 37 },
		{ 444, 26, 37 },
		{ 444, 27, 37 },
		{ 444, 28, 37 },
		{ 444, 29, 37 },
		{ 444, 30, 37 },
		{ 444, 31, 37 },
		{ 444, 32, 37 },
		{ 444, 33, 37 },
		{ 444, 34, 37 },
		{ 444, 35, 37 },
		{ 444, 36, 37 },
		{ 444, 37, 37 },
		{ 444, 38, 37 },
		{ 444, 39, 37 },
		{ 444, 40, 37 },
		{ 444, 41, 37 },
		{ 444, 42, 37 },
		{ 444, 43, 37 },
		{ 444, 44, 37 },
		{ 444, 45, 37 },
		{ 444, 46, 37 },
		{ 444, 47, 37 },
		{ 444, 48, 37 },
		{ 444, 49, 37 },
		{ 444, 50, 37 },
		{ 444, 51, 37 },
		{ 444, 52, 37 },
		{ 444, 53, 37 },
		{ 444, 54, 37 },
		{ 444, 55, 37 },
		{ 444, 56, 37 },
		{ 444, 57, 37 },
		{ 444, 58, 37 },
		{ 444, 59, 37 },
		{ 444, 60, 37 },
		{ 444, 61, 37 },
		{ 444, 62, 37 },
		{ 444, 63, 37 },
		{ 444, 64, 37 },
		{ 444, 65, 37 },
		{ 444, 66, 37 },
		{ 444, 67, 37 },
		{ 444, 68, 37 },
		{ 444, 69, 37 },
		{ 445, 7, 37 },
		{ 445, 8, 37 },
		{ 445, 9, 37 },
		{ 445, 10, 37 },
		{ 445, 11, 37 },
		{ 445, 12, 37 },
		{ 445, 13, 37 },
		{ 445, 14, 37 },
		{ 445, 15, 37 },
		{ 445, 16, 37 },
		{ 445, 17, 37 },
		{ 445, 18, 37 },
		{ 445, 19, 37 },
		{ 445, 20, 37 },
		{ 445, 21, 37 },
		{ 445, 22, 37 },
		{ 445, 23, 37 },
		{ 445, 24, 37 },
		{ 445, 25, 37 },
		{ 445, 26, 37 },
		{ 445, 27, 37 },
		{ 445, 28, 37 },
		{ 445, 29, 37 },
		{ 445, 30, 37 },
		{ 445, 31, 37 },
		{ 445, 32, 37 },
		{ 445, 33, 37 },
		{ 445, 34, 37 },
		{ 445, 35, 37 },
		{ 445, 36, 37 },
		{ 445, 37, 37 },
		{ 445, 38, 37 },
		{ 445, 39, 37 },
		{ 445, 40, 37 },
		{ 445, 41, 37 },
		{ 445, 42, 37 },
		{ 445, 43, 37 },
		{ 445, 44, 37 },
		{ 445, 45, 37 },
		{ 445, 46, 37 },
		{ 445, 47, 37 },
		{ 445, 48, 37 },
		{ 445, 49, 37 },
		{ 445, 50, 37 },
		{ 445, 51, 37 },
		{ 445, 52, 37 },
		{ 445, 53, 37 },
		{ 445, 54, 37 },
		{ 445, 55, 37 },
		{ 445, 56, 37 },
		{ 445, 57, 37 },
		{ 445, 58, 37 },
		{ 445, 59, 37 },
		{ 445, 60, 37 },
		{ 445, 61, 37 },
		{ 445, 62, 37 },
		{ 445, 63, 37 },
		{ 445, 64, 37 },
		{ 445, 65, 37 },
		{ 445, 66, 37 },
		{ 445, 67, 37 },
		{ 445, 68, 37 },
		{ 446, 6, 37 },
		{ 446, 7, 37 },
		{ 446, 8, 37 },
		{ 446, 9, 37 },
		{ 446, 10, 37 },
		{ 446, 11, 37 },
		{ 446, 12, 37 },
		{ 446, 13, 37 },
		{ 446, 14, 37 },
		{ 446, 15, 37 },
		{ 446, 16, 37 },
		{ 446, 17, 37 },
		{ 446, 18, 37 },
		{ 446, 19, 37 },
		{ 446, 20, 37 },
		{ 446, 21, 37 },
		{ 446, 22, 37 },
		{ 446, 23, 37 },
		{ 446, 24, 37 },
		{ 446, 25, 37 },
		{ 446, 26, 37 },
		{ 446, 27, 37 },
		{ 446, 28, 37 },
		{ 446, 29, 37 },
		{ 446, 30, 37 },
		{ 446, 31, 37 },
		{ 446, 32, 37 },
		{ 446, 33, 37 },
		{ 446, 34, 37 },
		{ 446, 35, 37 },
		{ 446, 36, 37 },
		{ 446, 37, 37 },
		{ 446, 38, 37 },
		{ 446, 39, 37 },
		{ 446, 40, 37 },
		{ 446, 41, 37 },
		{ 446, 42, 37 },
		{ 446, 43, 37 },
		{ 446, 44, 37 },
		{ 446, 45, 37 },
		{ 446, 46, 37 },
		{ 446, 47, 37 },
		{ 446, 48, 37 },
		{ 446, 49, 37 },
		{ 446, 50, 37 },
		{ 446, 51, 37 },
		{ 446, 52, 37 },
		{ 446, 53, 37 },
		{ 446, 54, 37 },
		{ 446, 55, 37 },
		{ 446, 56, 37 },
		{ 446, 57, 37 },
		{ 446, 58, 37 },
		{ 446, 59, 37 },
		{ 446, 60, 37 },
		{ 446, 61, 37 },
		{ 446, 62, 37 },
		{ 446, 63, 37 },
		{ 446, 64, 37 },
		{ 446, 65, 37 },
		{ 446, 66, 37 },
		{ 446, 67, 37 },
		{ 447, 4, 37 },
		{ 447, 5, 37 },
		{ 447, 6, 37 },
		{ 447, 7, 37 },
		{ 447, 8, 37 },
		{ 447, 9, 37 },
		{ 447, 10, 37 },
		{ 447, 11, 37 },
		{ 447, 12, 37 },
		{ 447, 13, 37 },
		{ 447, 14, 37 },
		{ 447, 15, 37 },
		{ 447, 16, 37 },
		{ 447, 17, 37 },
		{ 447, 18, 37 },
		{ 447, 19, 37 },
		{ 447, 20, 37 },
		{ 447, 21, 37 },
		{ 447, 22, 37 },
		{ 447, 23, 37 },
		{ 447, 24, 37 },
		{ 447, 25, 37 },
		{ 447, 26, 37 },
		{ 447, 27, 37 },
		{ 447, 28, 37 },
		{ 447, 29, 37 },
		{ 447, 30, 37 },
		{ 447, 31, 37 },
		{ 447, 32, 37 },
		{ 447, 33, 37 },
		{ 447, 34, 37 },
		{ 447, 35, 37 },
		{ 447, 36, 37 },
		{ 447, 37, 37 },
		{ 447, 38, 37 },
		{ 447, 39, 37 },
		{ 447, 40, 37 },
		{ 447, 41, 37 },
		{ 447, 42, 37 },
		{ 447, 43, 37 },
		{ 447, 44, 37 },
		{ 447, 45, 37 },
		{ 447, 46, 37 },
		{ 447, 47, 37 },
		{ 447, 48, 37 },
		{ 447, 49, 37 },
		{ 447, 50, 37 },
		{ 447, 51, 37 },
		{ 447, 52, 37 },
		{ 447, 53, 37 },
		{ 447, 54, 37 },
		{ 447, 55, 37 },
		{ 447, 56, 37 },
		{ 447, 57, 37 },
		{ 447, 58, 37 },
		{ 447, 59, 37 },
		{ 447, 60, 37 },
		{ 447, 61, 37 },
		{ 447, 62, 37 },
		{ 447, 63, 37 },
		{ 447, 64, 37 },
		{ 447, 65, 37 },
		{ 447, 66, 37 },
		{ 448, 3, 37 },
		{ 448, 4, 37 },
		{ 448, 5, 37 },
		{ 448, 6, 37 },
		{ 448, 7, 37 },
		{ 448, 8, 37 },
		{ 448, 9, 37 },
		{ 448, 10, 37 },
		{ 448, 11, 37 },
		{ 448, 12, 37 },
		{ 448, 13, 37 },
		{ 448, 14, 37 },
		{ 448, 15, 37 },
		{ 448, 16, 37 },
		{ 448, 17, 37 },
		{ 448, 18, 37 },
		{ 448, 19, 37 },
		{ 448, 20, 37 },
		{ 448, 21, 37 },
		{ 448, 22, 37 },
		{ 448, 23, 37 },
		{ 448, 24, 37 },
		{ 448, 25, 37 },
		{ 448, 26, 37 },
		{ 448, 27, 37 },
		{ 448, 28, 37 },
		{ 448, 29, 37 },
		{ 448, 30, 37 },
		{ 448, 31, 37 },
		{ 448, 32, 37 },
		{ 448, 33, 37 },
		{ 448, 34, 37 },
		{ 448, 35, 37 },
		{ 448, 36, 37 },
		{ 448, 37, 37 },
		{ 448, 38, 37 },
		{ 448, 39, 37 },
		{ 448, 40, 37 },
		{ 448, 41, 37 },
		{ 448, 42, 37 },
		{ 448, 43, 37 },
		{ 448, 44, 37 },
		{ 448, 45, 37 },
		{ 448, 46, 37 },
		{ 448, 47, 37 },
		{ 448, 48, 37 },
		{ 448, 49, 37 },
		{ 448, 50, 37 },
		{ 448, 51, 37 },
		{ 448, 52, 37 },
		{ 448, 53, 37 },
		{ 448, 54, 37 },
		{ 448, 55, 37 },
		{ 448, 56, 37 },
		{ 448, 57, 37 },
		{ 448, 58, 37 },
		{ 448, 59, 37 },
		{ 448, 60, 37 },
		{ 448, 61, 37 },
		{ 448, 62, 37 },
		{ 448, 63, 37 },
		{ 448, 64, 37 },
		{ 448, 65, 37 },
		{ 449, 2, 37 },
		{ 449, 3, 37 },
		{ 449, 4, 37 },
		{ 449, 5, 37 },
		{ 449, 6, 37 },
		{ 449, 7, 37 },
		{ 449, 8, 37 },
		{ 449, 9, 37 },
		{ 449, 10, 37 },
		{ 449, 11, 37 },
		{ 449, 12, 37 },
		{ 449, 13, 37 },
		{ 449, 14, 37 },
		{ 449, 15, 37 },
		{ 449, 16, 37 },
		{ 449, 17, 37 },
		{ 449, 18, 37 },
		{ 449, 19, 37 },
		{ 449, 20, 37 },
		{ 449, 21, 37 },
		{ 449, 22, 37 },
		{ 449, 23, 37 },
		{ 449, 24, 37 },
		{ 449, 25, 37 },
		{ 449, 26, 37 },
		{ 449, 27, 37 },
		{ 449, 28, 37 },
		{ 449, 29, 37 },
		{ 449, 30, 37 },
		{ 449, 31, 37 },
		{ 449, 32, 37 },
		{ 449, 33, 37 },
		{ 449, 34, 37 },
		{ 449, 35, 37 },
		{ 449, 36, 37 },
		{ 449, 37, 37 },
		{ 449, 38, 37 },
		{ 449, 39, 37 },
		{ 449, 40, 37 },
		{ 449, 41, 37 },
		{ 449, 42, 37 },
		{ 449, 43, 37 },
		{ 449, 44, 37 },
		{ 449, 45, 37 },
		{ 449, 46, 37 },
		{ 449, 47, 37 },
		{ 449, 48, 37 },
		{ 449, 49, 37 },
		{ 449, 50, 37 },
		{ 449, 51, 37 },
		{ 449, 52, 37 },
		{ 449, 53, 37 },
		{ 449, 54, 37 },
		{ 449, 55, 37 },
		{ 449, 56, 37 },
		{ 449, 57, 37 },
		{ 449, 58, 37 },
		{ 449, 59, 37 },
		{ 449, 60, 37 },
		{ 449, 61, 37 },
		{ 449, 62, 37 },
		{ 449, 63, 37 },
		{ 449, 64, 37 },
		{ 450, 1, 37 },
		{ 450, 2, 37 },
		{ 450, 3, 37 },
		{ 450, 4, 37 },
		{ 450, 5, 37 },
		{ 450, 6, 37 },
		{ 450, 7, 37 },
		{ 450, 8, 37 },
		{ 450, 9, 37 },
		{ 450, 10, 37 },
		{ 450, 11, 37 },
		{ 450, 12, 37 },
		{ 450, 13, 37 },
		{ 450, 14, 37 },
		{ 450, 15, 37 },
		{ 450, 16, 37 },
		{ 450, 17, 37 },
		{ 450, 18, 37 },
		{ 450, 19, 37 },
		{ 450, 20, 37 },
		{ 450, 21, 37 },
		{ 450, 22, 37 },
		{ 450, 23, 37 },
		{ 450, 24, 37 },
		{ 450, 25, 37 },
		{ 450, 26, 37 },
		{ 450, 27, 37 },
		{ 450, 28, 37 },
		{ 450, 29, 37 },
		{ 450, 30, 37 },
		{ 450, 31, 37 },
		{ 450, 32, 37 },
		{ 450, 33, 37 },
		{ 450, 34, 37 },
		{ 450, 35, 37 },
		{ 450, 36, 37 },
		{ 450, 37, 37 },
		{ 450, 38, 37 },
		{ 450, 39, 37 },
		{ 450, 40, 37 },
		{ 450, 41, 37 },
		{ 450, 42, 37 },
		{ 450, 43, 37 },
		{ 450, 44, 37 },
		{ 450, 45, 37 },
		{ 450, 46, 37 },
		{ 450, 47, 37 },
		{ 450, 48, 37 },
		{ 450, 49, 37 },
		{ 450, 50, 37 },
		{ 450, 51, 37 },
		{ 450, 52, 37 },
		{ 450, 53, 37 },
		{ 450, 54, 37 },
		{ 450, 55, 37 },
		{ 450, 56, 37 },
		{ 450, 57, 37 },
		{ 450, 58, 37 },
		{ 450, 59, 37 },
		{ 450, 60, 37 },
		{ 450, 61, 37 },
		{ 450, 62, 37 },
		{ 450, 63, 37 },
		{ 451, 1, 37 },
		{ 451, 2, 37 },
		{ 451, 3, 37 },
		{ 451, 4, 37 },
		{ 451, 5, 37 },
		{ 451, 6, 37 },
		{ 451, 7, 37 },
		{ 451, 8, 37 },
		{ 451, 9, 37 },
		{ 451, 10, 37 },
		{ 451, 11, 37 },
		{ 451, 12, 37 },
		{ 451, 13, 37 },
		{ 451, 14, 37 },
		{ 451, 15, 37 },
		{ 451, 16, 37 },
		{ 451, 17, 37 },
		{ 451, 18, 37 },
		{ 451, 19, 37 },
		{ 451, 20, 37 },
		{ 451, 21, 37 },
		{ 451, 22, 37 },
		{ 451, 23, 37 },
		{ 451, 24, 37 },
		{ 451, 25, 37 },
		{ 451, 26, 37 },
		{ 451, 27, 37 },
		{ 451, 28, 37 },
		{ 451, 29, 37 },
		{ 451, 30, 37 },
		{ 451, 31, 37 },
		{ 451, 32, 37 },
		{ 451, 33, 37 },
		{ 451, 34, 37 },
		{ 451, 35, 37 },
		{ 451, 36, 37 },
		{ 451, 37, 37 },
		{ 451, 38, 37 },
		{ 451, 39, 37 },
		{ 451, 40, 37 },
		{ 451, 41, 37 },
		{ 451, 42, 37 },
		{ 451, 43, 37 },
		{ 451, 44, 37 },
		{ 451, 45, 37 },
		{ 451, 46, 37 },
		{ 451, 47, 37 },
		{ 451, 48, 37 },
		{ 451, 49, 37 },
		{ 451, 50, 37 },
		{ 451, 51, 37 },
		{ 451, 52, 37 },
		{ 451, 53, 37 },
		{ 451, 54, 37 },
		{ 451, 55, 37 },
		{ 451, 56, 37 },
		{ 451, 57, 37 },
		{ 451, 58, 37 },
		{ 451, 59, 37 },
		{ 451, 60, 37 },
		{ 451, 61, 37 },
		{ 451, 62, 37 },
		{ 452, 1, 37 },
		{ 452, 2, 37 },
		{ 452, 3, 37 },
		{ 452, 4, 37 },
		{ 452, 5, 37 },
		{ 452, 6, 37 },
		{ 452, 7, 37 },
		{ 452, 8, 37 },
		{ 452, 9, 37 },
		{ 452, 10, 37 },
		{ 452, 11, 37 },
		{ 452, 12, 37 },
		{ 452, 13, 37 },
		{ 452, 14, 37 },
		{ 452, 15, 37 },
		{ 452, 16, 37 },
		{ 452, 17, 37 },
		{ 452, 18, 37 },
		{ 452, 19, 37 },
		{ 452, 20, 37 },
		{ 452, 21, 37 },
		{ 452, 22, 37 },
		{ 452, 23, 37 },
		{ 452, 24, 37 },
		{ 452, 25, 37 },
		{ 452, 26, 37 },
		{ 452, 27, 37 },
		{ 452, 28, 37 },
		{ 452, 29, 37 },
		{ 452, 30, 37 },
		{ 452, 31, 37 },
		{ 452, 32, 37 },
		{ 452, 33, 37 },
		{ 452, 34, 37 },
		{ 452, 35, 37 },
		{ 452, 36, 37 },
		{ 452, 37, 37 },
		{ 452, 38, 37 },
		{ 452, 39, 37 },
		{ 452, 40, 37 },
		{ 452, 41, 37 },
		{ 452, 42, 37 },
		{ 452, 43, 37 },
		{ 452, 44, 37 },
		{ 452, 45, 37 },
		{ 452, 46, 37 },
		{ 452, 47, 37 },
		{ 452, 48, 37 },
		{ 452, 49, 37 },
		{ 452, 50, 37 },
		{ 452, 51, 37 },
		{ 452, 52, 37 },
		{ 452, 53, 37 },
		{ 452, 54, 37 },
		{ 452, 55, 37 },
		{ 452, 56, 37 },
		{ 452, 57, 37 },
		{ 452, 58, 37 },
		{ 452, 59, 37 },
		{ 452, 60, 37 },
		{ 452, 61, 37 },
		{ 453, 1, 37 },
		{ 453, 2, 37 },
		{ 453, 3, 37 },
		{ 453, 4, 37 },
		{ 453, 5, 37 },
		{ 453, 6, 37 },
		{ 453, 7, 37 },
		{ 453, 8, 37 },
		{ 453, 9, 37 },
		{ 453, 10, 37 },
		{ 453, 11, 37 },
		{ 453, 12, 37 },
		{ 453, 13, 37 },
		{ 453, 14, 37 },
		{ 453, 15, 37 },
		{ 453, 16, 37 },
		{ 453, 17, 37 },
		{ 453, 18, 37 },
		{ 453, 19, 37 },
		{ 453, 20, 37 },
		{ 453, 21, 37 },
		{ 453, 22, 37 },
		{ 453, 23, 37 },
		{ 453, 24, 37 },
		{ 453, 25, 37 },
		{ 453, 26, 37 },
		{ 453, 27, 37 },
		{ 453, 28, 37 },
		{ 453, 29, 37 },
		{ 453, 30, 37 },
		{ 453, 31, 37 },
		{ 453, 32, 37 },
		{ 453, 33, 37 },
		{ 453, 34, 37 },
		{ 453, 35, 37 },
		{ 453, 36, 37 },
		{ 453, 37, 37 },
		{ 453, 38, 37 },
		{ 453, 39, 37 },
		{ 453, 40, 37 },
		{ 453, 41, 37 },
		{ 453, 42, 37 },
		{ 453, 43, 37 },
		{ 453, 44, 37 },
		{ 453, 45, 37 },
		{ 453, 46, 37 },
		{ 453, 47, 37 },
		{ 453, 48, 37 },
		{ 453, 49, 37 },
		{ 453, 50, 37 },
		{ 453, 51, 37 },
		{ 453, 52, 37 },
		{ 453, 53, 37 },
		{ 453, 54, 37 },
		{ 453, 55, 37 },
		{ 453, 56, 37 },
		{ 453, 57, 37 },
		{ 453, 58, 37 },
		{ 453, 59, 37 },
		{ 453, 60, 37 },
		{ 454, 1, 37 },
		{ 454, 2, 37 },
		{ 454, 3, 37 },
		{ 454, 4, 37 },
		{ 454, 5, 37 },
		{ 454, 6, 37 },
		{ 454, 7, 37 },
		{ 454, 8, 37 },
		{ 454, 9, 37 },
		{ 454, 10, 37 },
		{ 454, 11, 37 },
		{ 454, 12, 37 },
		{ 454, 13, 37 },
		{ 454, 14, 37 },
		{ 454, 15, 37 },
		{ 454, 16, 37 },
		{ 454, 17, 37 },
		{ 454, 18, 37 },
		{ 454, 19, 37 },
		{ 454, 20, 37 },
		{ 454, 21, 37 },
		{ 454, 22, 37 },
		{ 454, 23, 37 },
		{ 454, 24, 37 },
		{ 454, 25, 37 },
		{ 454, 26, 37 },
		{ 454, 27, 37 },
		{ 454, 28, 37 },
		{ 454, 29, 37 },
		{ 454, 30, 37 },
		{ 454, 31, 37 },
		{ 454, 32, 37 },
		{ 454, 33, 37 },
		{ 454, 34, 37 },
		{ 454, 35, 37 },
		{ 454, 36, 37 },
		{ 454, 37, 37 },
		{ 454, 38, 37 },
		{ 454, 39, 37 },
		{ 454, 40, 37 },
		{ 454, 41, 37 },
		{ 454, 42, 37 },
		{ 454, 43, 37 },
		{ 454, 44, 37 },
		{ 454, 45, 37 },
		{ 454, 46, 37 },
		{ 454, 47, 37 },
		{ 454, 48, 37 },
		{ 454, 49, 37 },
		{ 454, 50, 37 },
		{ 454, 51, 37 },
		{ 454, 52, 37 },
		{ 454, 53, 37 },
		{ 454, 54, 37 },
		{ 454, 55, 37 },
		{ 454, 56, 37 },
		{ 454, 57, 37 },
		{ 454, 58, 37 },
		{ 454, 59, 37 },
		{ 455, 1, 37 },
		{ 455, 2, 37 },
		{ 455, 3, 37 },
		{ 455, 4, 37 },
		{ 455, 5, 37 },
		{ 455, 6, 37 },
		{ 455, 7, 37 },
		{ 455, 8, 37 },
		{ 455, 9, 37 },
		{ 455, 10, 37 },
		{ 455, 11, 37 },
		{ 455, 12, 37 },
		{ 455, 13, 37 },
		{ 455, 14, 37 },
		{ 455, 15, 37 },
		{ 455, 16, 37 },
		{ 455, 17, 37 },
		{ 455, 18, 37 },
		{ 455, 19, 37 },
		{ 455, 20, 37 },
		{ 455, 21, 37 },
		{ 455, 22, 37 },
		{ 455, 23, 37 },
		{ 455, 24, 37 },
		{ 455, 25, 37 },
		{ 455, 26, 37 },
		{ 455, 27, 37 },
		{ 455, 28, 37 },
		{ 455, 29, 37 },
		{ 455, 30, 37 },
		{ 455, 31, 37 },
		{ 455, 32, 37 },
		{ 455, 33, 37 },
		{ 455, 34, 37 },
		{ 455, 35, 37 },
		{ 455, 36, 37 },
		{ 455, 37, 37 },
		{ 455, 38, 37 },
		{ 455, 39, 37 },
		{ 455, 40, 37 },
		{ 455, 41, 37 },
		{ 455, 42, 37 },
		{ 455, 43, 37 },
		{ 455, 44, 37 },
		{ 455, 45, 37 },
		{ 455, 46, 37 },
		{ 455, 47, 37 },
		{ 455, 48, 37 },
		{ 455, 49, 37 },
		{ 455, 50, 37 },
		{ 455, 51, 37 },
		{ 455, 52, 37 },
		{ 455, 53, 37 },
		{ 455, 54, 37 },
		{ 455, 55, 37 },
		{ 455, 56, 37 },
		{ 455, 57, 37 },
		{ 455, 58, 37 },
		{ 456, 1, 37 },
		{ 456, 2, 37 },
		{ 456, 3, 37 },
		{ 456, 4, 37 },
		{ 456, 5, 37 },
		{ 456, 6, 37 },
		{ 456, 7, 37 },
		{ 456, 8, 37 },
		{ 456, 9, 37 },
		{ 456, 10, 37 },
		{ 456, 11, 37 },
		{ 456, 12, 37 },
		{ 456, 13, 37 },
		{ 456, 14, 37 },
		{ 456, 15, 37 },
		{ 456, 16, 37 },
		{ 456, 17, 37 },
		{ 456, 18, 37 },
		{ 456, 19, 37 },
		{ 456, 20, 37 },
		{ 456, 21, 37 },
		{ 456, 22, 37 },
		{ 456, 23, 37 },
		{ 456, 24, 37 },
		{ 456, 25, 37 },
		{ 456, 26, 37 },
		{ 456, 27, 37 },
		{ 456, 28, 37 },
		{ 456, 29, 37 },
		{ 456, 30, 37 },
		{ 456, 31, 37 },
		{ 456, 32, 37 },
		{ 456, 33, 37 },
		{ 456, 34, 37 },
		{ 456, 35, 37 },
		{ 456, 36, 37 },
		{ 456, 37, 37 },
		{ 456, 38, 37 },
		{ 456, 39, 37 },
		{ 456, 40, 37 },
		{ 456, 41, 37 },
		{ 456, 42, 37 },
		{ 456, 43, 37 },
		{ 456, 44, 37 },
		{ 456, 45, 37 },
		{ 456, 46, 37 },
		{ 456, 47, 37 },
		{ 456, 48, 37 },
		{ 456, 49, 37 },
		{ 456, 50, 37 },
		{ 456, 51, 37 },
		{ 456, 52, 37 },
		{ 456, 53, 37 },
		{ 456, 54, 37 },
		{ 456, 55, 37 },
		{ 456, 56, 37 },
		{ 456, 57, 37 },
		{ 457, 1, 37 },
		{ 457, 2, 37 },
		{ 457, 3, 37 },
		{ 457, 4, 37 },
		{ 457, 5, 37 },
		{ 457, 6, 37 },
		{ 457, 7, 37 },
		{ 457, 8, 37 },
		{ 457, 9, 37 },
		{ 457, 10, 37 },
		{ 457, 11, 37 },
		{ 457, 12, 37 },
		{ 457, 13, 37 },
		{ 457, 14, 37 },
		{ 457, 15, 37 },
		{ 457, 16, 37 },
		{ 457, 17, 37 },
		{ 457, 18, 37 },
		{ 457, 19, 37 },
		{ 457, 20, 37 },
		{ 457, 21, 37 },
		{ 457, 22, 37 },
		{ 457, 23, 37 },
		{ 457, 24, 37 },
		{ 457, 25, 37 },
		{ 457, 26, 37 },
		{ 457, 27, 37 },
		{ 457, 28, 37 },
		{ 457, 29, 37 },
		{ 457, 30, 37 },
		{ 457, 31, 37 },
		{ 457, 32, 37 },
		{ 457, 33, 37 },
		{ 457, 34, 37 },
		{ 457, 35, 37 },
		{ 457, 36, 37 },
		{ 457, 37, 37 },
		{ 457, 38, 37 },
		{ 457, 39, 37 },
		{ 457, 40, 37 },
		{ 457, 41, 37 },
		{ 457, 42, 37 },
		{ 457, 43, 37 },
		{ 457, 44, 37 },
		{ 457, 45, 37 },
		{ 457, 46, 37 },
		{ 457, 47, 37 },
		{ 457, 48, 37 },
		{ 457, 49, 37 },
		{ 457, 50, 37 },
		{ 457, 51, 37 },
		{ 457, 52, 37 },
		{ 457, 53, 37 },
		{ 457, 54, 37 },
		{ 457, 55, 37 },
		{ 457, 56, 37 },
		{ 458, 1, 37 },
		{ 458, 2, 37 },
		{ 458, 3, 37 },
		{ 458, 4, 37 },
		{ 458, 5, 37 },
		{ 458, 6, 37 },
		{ 458, 7, 37 },
		{ 458, 8, 37 },
		{ 458, 9, 37 },
		{ 458, 10, 37 },
		{ 458, 11, 37 },
		{ 458, 12, 37 },
		{ 458, 13, 37 },
		{ 458, 14, 37 },
		{ 458, 15, 37 },
		{ 458, 16, 37 },
		{ 458, 17, 37 },
		{ 458, 18, 37 },
		{ 458, 19, 37 },
		{ 458, 20, 37 },
		{ 458, 21, 37 },
		{ 458, 22, 37 },
		{ 458, 23, 37 },
		{ 458, 24, 37 },
		{ 458, 25, 37 },
		{ 458, 26, 37 },
		{ 458, 27, 37 },
		{ 458, 28, 37 },
		{ 458, 29, 37 },
		{ 458, 30, 37 },
		{ 458, 31, 37 },
		{ 458, 32, 37 },
		{ 458, 33, 37 },
		{ 458, 34, 37 },
		{ 458, 35, 37 },
		{ 458, 36, 37 },
		{ 458, 37, 37 },
		{ 458, 38, 37 },
		{ 458, 39, 37 },
		{ 458, 40, 37 },
		{ 458, 41, 37 },
		{ 458, 42, 37 },
		{ 458, 43, 37 },
		{ 458, 44, 37 },
		{ 458, 45, 37 },
		{ 458, 46, 37 },
		{ 458, 47, 37 },
		{ 458, 48, 37 },
		{ 458, 49, 37 },
		{ 458, 50, 37 },
		{ 458, 51, 37 },
		{ 458, 52, 37 },
		{ 458, 53, 37 },
		{ 458, 54, 37 },
		{ 458, 55, 37 },
		{ 459, 1, 37 },
		{ 459, 2, 37 },
		{ 459, 3, 37 },
		{ 459, 4, 37 },
		{ 459, 5, 37 },
		{ 459, 6, 37 },
		{ 459, 7, 37 },
		{ 459, 8, 37 },
		{ 459, 9, 37 },
		{ 459, 10, 37 },
		{ 459, 11, 37 },
		{ 459, 12, 37 },
		{ 459, 13, 37 },
		{ 459, 14, 37 },
		{ 459, 15, 37 },
		{ 459, 16, 37 },
		{ 459, 17, 37 },
		{ 459, 18, 37 },
		{ 459, 19, 37 },
		{ 459, 20, 37 },
		{ 459, 21, 37 },
		{ 459, 22, 37 },
		{ 459, 23, 37 },
		{ 459, 24, 37 },
		{ 459, 25, 37 },
		{ 459, 26, 37 },
		{ 459, 27, 37 },
		{ 459, 28, 37 },
		{ 459, 29, 37 },
		{ 459, 30, 37 },
		{ 459, 31, 37 },
		{ 459, 32, 37 },
		{ 459, 33, 37 },
		{ 459, 34, 37 },
		{ 459, 35, 37 },
		{ 459, 36, 37 },
		{ 459, 37, 37 },
		{ 459, 38, 37 },
		{ 459, 39, 37 },
		{ 459, 40, 37 },
		{ 459, 41, 37 },
		{ 459, 42, 37 },
		{ 459, 43, 37 },
		{ 459, 44, 37 },
		{ 459, 45, 37 },
		{ 459, 46, 37 },
		{ 459, 47, 37 },
		{ 459, 48, 37 },
		{ 459, 49, 37 },
		{ 459, 50, 37 },
		{ 459, 51, 37 },
		{ 459, 52, 37 },
		{ 459, 53, 37 },
		{ 459, 54, 37 },
		{ 460, 1, 37 },
		{ 460, 2, 37 },
		{ 460, 3, 37 },
		{ 460, 4, 37 },
		{ 460, 5, 37 },
		{ 460, 6, 37 },
		{ 460, 7, 37 },
		{ 460, 8, 37 },
		{ 460, 9, 37 },
		{ 460, 10, 37 },
		{ 460, 11, 37 },
		{ 460, 12, 37 },
		{ 460, 13, 37 },
		{ 460, 14, 37 },
		{ 460, 15, 37 },
		{ 460, 16, 37 },
		{ 460, 17, 37 },
		{ 460, 18, 37 },
		{ 460, 19, 37 },
		{ 460, 20, 37 },
		{ 460, 21, 37 },
		{ 460, 22, 37 },
		{ 460, 23, 37 },
		{ 460, 24, 37 },
		{ 460, 25, 37 },
		{ 460, 26, 37 },
		{ 460, 27, 37 },
		{ 460, 28, 37 },
		{ 460, 29, 37 },
		{ 460, 30, 37 },
		{ 460, 31, 37 },
		{ 460, 32, 37 },
		{ 460, 33, 37 },
		{ 460, 34, 37 },
		{ 460, 35, 37 },
		{ 460, 36, 37 },
		{ 460, 37, 37 },
		{ 460, 38, 37 },
		{ 460, 39, 37 },
		{ 460, 40, 37 },
		{ 460, 41, 37 },
		{ 460, 42, 37 },
		{ 460, 43, 37 },
		{ 460, 44, 37 },
		{ 460, 45, 37 },
		{ 460, 46, 37 },
		{ 460, 47, 37 },
		{ 460, 48, 37 },
		{ 460, 49, 37 },
		{ 460, 50, 37 },
		{ 460, 51, 37 },
		{ 460, 52, 37 },
		{ 460, 53, 37 },
		{ 461, 1, 37 },
		{ 461, 2, 37 },
		{ 461, 3, 37 },
		{ 461, 4, 37 },
		{ 461, 5, 37 },
		{ 461, 6, 37 },
		{ 461, 7, 37 },
		{ 461, 8, 37 },
		{ 461, 9, 37 },
		{ 461, 10, 37 },
		{ 461, 11, 37 },
		{ 461, 12, 37 },
		{ 461, 13, 37 },
		{ 461, 14, 37 },
		{ 461, 15, 37 },
		{ 461, 16, 37 },
		{ 461, 17, 37 },
		{ 461, 18, 37 },
		{ 461, 19, 37 },
		{ 461, 20, 37 },
		{ 461, 21, 37 },
		{ 461, 22, 37 },
		{ 461, 23, 37 },
		{ 461, 24, 37 },
		{ 461, 25, 37 },
		{ 461, 26, 37 },
		{ 461, 27, 37 },
		{ 461, 28, 37 },
		{ 461, 29, 37 },
		{ 461, 30, 37 },
		{ 461, 31, 37 },
		{ 461, 32, 37 },
		{ 461, 33, 37 },
		{ 461, 34, 37 },
		{ 461, 35, 37 },
		{ 461, 36, 37 },
		{ 461, 37, 37 },
		{ 461, 38, 37 },
		{ 461, 39, 37 },
		{ 461, 40, 37 },
		{ 461, 41, 37 },
		{ 461, 42, 37 },
		{ 461, 43, 37 },
		{ 461, 44, 37 },
		{ 461, 45, 37 },
		{ 461, 46, 37 },
		{ 461, 47, 37 },
		{ 461, 48, 37 },
		{ 461, 49, 37 },
		{ 461, 50, 37 },
		{ 461, 51, 37 },
		{ 461, 52, 37 },
		{ 462, 1, 37 },
		{ 462, 2, 37 },
		{ 462, 3, 37 },
		{ 462, 4, 37 },
		{ 462, 5, 37 },
		{ 462, 6, 37 },
		{ 462, 7, 37 },
		{ 462, 8, 37 },
		{ 462, 9, 37 },
		{ 462, 10, 37 },
		{ 462, 11, 37 },
		{ 462, 12, 37 },
		{ 462, 13, 37 },
		{ 462, 14, 37 },
		{ 462, 15, 37 },
		{ 462, 16, 37 },
		{ 462, 17, 37 },
		{ 462, 18, 37 },
		{ 462, 19, 37 },
		{ 462, 20, 37 },
		{ 462, 21, 37 },
		{ 462, 22, 37 },
		{ 462, 23, 37 },
		{ 462, 24, 37 },
		{ 462, 25, 37 },
		{ 462, 26, 37 },
		{ 462, 27, 37 },
		{ 462, 28, 37 },
		{ 462, 29, 37 },
		{ 462, 30, 37 },
		{ 462, 31, 37 },
		{ 462, 32, 37 },
		{ 462, 33, 37 },
		{ 462, 34, 37 },
		{ 462, 35, 37 },
		{ 462, 36, 37 },
		{ 462, 37, 37 },
		{ 462, 38, 37 },
		{ 462, 39, 37 },
		{ 462, 40, 37 },
		{ 462, 41, 37 },
		{ 462, 42, 37 },
		{ 462, 43, 37 },
		{ 462, 44, 37 },
		{ 462, 45, 37 },
		{ 462, 46, 37 },
		{ 462, 47, 37 },
		{ 462, 48, 37 },
		{ 462, 49, 37 },
		{ 462, 50, 37 },
		{ 462, 51, 37 },
		{ 463, 1, 37 },
		{ 463, 2, 37 },
		{ 463, 3, 37 },
		{ 463, 4, 37 },
		{ 463, 5, 37 },
		{ 463, 6, 37 },
		{ 463, 7, 37 },
		{ 463, 8, 37 },
		{ 463, 9, 37 },
		{ 463, 10, 37 },
		{ 463, 11, 37 },
		{ 463, 12, 37 },
		{ 463, 13, 37 },
		{ 463, 14, 37 },
		{ 463, 15, 37 },
		{ 463, 16, 37 },
		{ 463, 17, 37 },
		{ 463, 18, 37 },
		{ 463, 19, 37 },
		{ 463, 20, 37 },
		{ 463, 21, 37 },
		{ 463, 22, 37 },
		{ 463, 23, 37 },
		{ 463, 24, 37 },
		{ 463, 25, 37 },
		{ 463, 26, 37 },
		{ 463, 27, 37 },
		{ 463, 28, 37 },
		{ 463, 29, 37 },
		{ 463, 30, 37 },
		{ 463, 31, 37 },
		{ 463, 32, 37 },
		{ 463, 33, 37 },
		{ 463, 34, 37 },
		{ 463, 35, 37 },
		{ 463, 36, 37 },
		{ 463, 37, 37 },
		{ 463, 38, 37 },
		{ 463, 39, 37 },
		{ 463, 40, 37 },
		{ 463, 41, 37 },
		{ 463, 42, 37 },
		{ 463, 43, 37 },
		{ 463, 44, 37 },
		{ 463, 45, 37 },
		{ 463, 46, 37 },
		{ 463, 47, 37 },
		{ 463, 48, 37 },
		{ 463, 49, 37 },
		{ 463, 50, 37 },
		{ 464, 1, 37 },
		{ 464, 2, 37 },
		{ 464, 3, 37 },
		{ 464, 4, 37 },
		{ 464, 5, 37 },
		{ 464, 6, 37 },
		{ 464, 7, 37 },
		{ 464, 8, 37 },
		{ 464, 9, 37 },
		{ 464, 10, 37 },
		{ 464, 11, 37 },
		{ 464, 12, 37 },
		{ 464, 13, 37 },
		{ 464, 14, 37 },
		{ 464, 15, 37 },
		{ 464, 16, 37 },
		{ 464, 17, 37 },
		{ 464, 18, 37 },
		{ 464, 19, 37 },
		{ 464, 20, 37 },
		{ 464, 21, 37 },
		{ 464, 22, 37 },
		{ 464, 23, 37 },
		{ 464, 24, 37 },
		{ 464, 25, 37 },
		{ 464, 26, 37 },
		{ 464, 27, 37 },
		{ 464, 28, 37 },
		{ 464, 29, 37 },
		{ 464, 30, 37 },
		{ 464, 31, 37 },
		{ 464, 32, 37 },
		{ 464, 33, 37 },
		{ 464, 34, 37 },
		{ 464, 35, 37 },
		{ 464, 36, 37 },
		{ 464, 37, 37 },
		{ 464, 38, 37 },
		{ 464, 39, 37 },
		{ 464, 40, 37 },
		{ 464, 41, 37 },
		{ 464, 42, 37 },
		{ 464, 43, 37 },
		{ 464, 44, 37 },
		{ 464, 45, 37 },
		{ 464, 46, 37 },
		{ 464, 47, 37 },
		{ 464, 48, 37 },
		{ 464, 49, 37 },
		{ 465, 1, 37 },
		{ 465, 2, 37 },
		{ 465, 3, 37 },
		{ 465, 4, 37 },
		{ 465, 5, 37 },
		{ 465, 6, 37 },
		{ 465, 7, 37 },
		{ 465, 8, 37 },
		{ 465, 9, 37 },
		{ 465, 10, 37 },
		{ 465, 11, 37 },
		{ 465, 12, 37 },
		{ 465, 13, 37 },
		{ 465, 14, 37 },
		{ 465, 15, 37 },
		{ 465, 16, 37 },
		{ 465, 17, 37 },
		{ 465, 18, 37 },
		{ 465, 19, 37 },
		{ 465, 20, 37 },
		{ 465, 21, 37 },
		{ 465, 22, 37 },
		{ 465, 23, 37 },
		{ 465, 24, 37 },
		{ 465, 25, 37 },
		{ 465, 26, 37 },
		{ 465, 27, 37 },
		{ 465, 28, 37 },
		{ 465, 29, 37 },
		{ 465, 30, 37 },
		{ 465, 31, 37 },
		{ 465, 32, 37 },
		{ 465, 33, 37 },
		{ 465, 34, 37 },
		{ 465, 35, 37 },
		{ 465, 36, 37 },
		{ 465, 37, 37 },
		{ 465, 38, 37 },
		{ 465, 39, 37 },
		{ 465, 40, 37 },
		{ 465, 41, 37 },
		{ 465, 42, 37 },
		{ 465, 43, 37 },
		{ 465, 44, 37 },
		{ 465, 45, 37 },
		{ 465, 46, 37 },
		{ 465, 47, 37 },
		{ 465, 48, 37 },
		{ 466, 1, 37 },
		{ 466, 2, 37 },
		{ 466, 3, 37 },
		{ 466, 4, 37 },
		{ 466, 5, 37 },
		{ 466, 6, 37 },
		{ 466, 7, 37 },
		{ 466, 8, 37 },
		{ 466, 9, 37 },
		{ 466, 10, 37 },
		{ 466, 11, 37 },
		{ 466, 12, 37 },
		{ 466, 13, 37 },
		{ 466, 14, 37 },
		{ 466, 15, 37 },
		{ 466, 16, 37 },
		{ 466, 17, 37 },
		{ 466, 18, 37 },
		{ 466, 19, 37 },
		{ 466, 20, 37 },
		{ 466, 21, 37 },
		{ 466, 22, 37 },
		{ 466, 23, 37 },
		{ 466, 24, 37 },
		{ 466, 25, 37 },
		{ 466, 26, 37 },
		{ 466, 27, 37 },
		{ 466, 28, 37 },
		{ 466, 29, 37 },
		{ 466, 30, 37 },
		{ 466, 31, 37 },
		{ 466, 32, 37 },
		{ 466, 33, 37 },
		{ 466, 34, 37 },
		{ 466, 35, 37 },
		{ 466, 36, 37 },
		{ 466, 37, 37 },
		{ 466, 38, 37 },
		{ 466, 39, 37 },
		{ 466, 40, 37 },
		{ 466, 41, 37 },
		{ 466, 42, 37 },
		{ 466, 43, 37 },
		{ 466, 44, 37 },
		{ 466, 45, 37 },
		{ 466, 46, 37 },
		{ 466, 47, 37 },
		{ 467, 1, 37 },
		{ 467, 2, 37 },
		{ 467, 3, 37 },
		{ 467, 4, 37 },
		{ 467, 5, 37 },
		{ 467, 6, 37 },
		{ 467, 7, 37 },
		{ 467, 8, 37 },
		{ 467, 9, 37 },
		{ 467, 10, 37 },
		{ 467, 11, 37 },
		{ 467, 12, 37 },
		{ 467, 13, 37 },
		{ 467, 14, 37 },
		{ 467, 15, 37 },
		{ 467, 16, 37 },
		{ 467, 17, 37 },
		{ 467, 18, 37 },
		{ 467, 19, 37 },
		{ 467, 20, 37 },
		{ 467, 21, 37 },
		{ 467, 22, 37 },
		{ 467, 23, 37 },
		{ 467, 24, 37 },
		{ 467, 25, 37 },
		{ 467, 26, 37 },
		{ 467, 27, 37 },
		{ 467, 28, 37 },
		{ 467, 29, 37 },
		{ 467, 30, 37 },
		{ 467, 31, 37 },
		{ 467, 32, 37 },
		{ 467, 33, 37 },
		{ 467, 34, 37 },
		{ 467, 35, 37 },
		{ 467, 36, 37 },
		{ 467, 37, 37 },
		{ 467, 38, 37 },
		{ 467, 39, 37 },
		{ 467, 40, 37 },
		{ 467, 41, 37 },
		{ 467, 42, 37 },
		{ 467, 43, 37 },
		{ 467, 44, 37 },
		{ 467, 45, 37 },
		{ 467, 46, 37 },
		{ 468, 1, 37 },
		{ 468, 2, 37 },
		{ 468, 3, 37 },
		{ 468, 4, 37 },
		{ 468, 5, 37 },
		{ 468, 6, 37 },
		{ 468, 7, 37 },
		{ 468, 8, 37 },
		{ 468, 9, 37 },
		{ 468, 10, 37 },
		{ 468, 11, 37 },
		{ 468, 12, 37 },
		{ 468, 13, 37 },
		{ 468, 14, 37 },
		{ 468, 15, 37 },
		{ 468, 16, 37 },
		{ 468, 17, 37 },
		{ 468, 18, 37 },
		{ 468, 19, 37 },
		{ 468, 20, 37 },
		{ 468, 21, 37 },
		{ 468, 22, 37 },
		{ 468, 23, 37 },
		{ 468, 24, 37 },
		{ 468, 25, 37 },
		{ 468, 26, 37 },
		{ 468, 27, 37 },
		{ 468, 28, 37 },
		{ 468, 29, 37 },
		{ 468, 30, 37 },
		{ 468, 31, 37 },
		{ 468, 32, 37 },
		{ 468, 33, 37 },
		{ 468, 34, 37 },
		{ 468, 35, 37 },
		{ 468, 36, 37 },
		{ 468, 37, 37 },
		{ 468, 38, 37 },
		{ 468, 39, 37 },
		{ 468, 40, 37 },
		{ 468, 41, 37 },
		{ 468, 42, 37 },
		{ 468, 43, 37 },
		{ 468, 44, 37 },
		{ 468, 45, 37 },
		{ 469, 1, 37 },
		{ 469, 2, 37 },
		{ 469, 3, 37 },
		{ 469, 4, 37 },
		{ 469, 5, 37 },
		{ 469, 6, 37 },
		{ 469, 7, 37 },
		{ 469, 8, 37 },
		{ 469, 9, 37 },
		{ 469, 10, 37 },
		{ 469, 11, 37 },
		{ 469, 12, 37 },
		{ 469, 13, 37 },
		{ 469, 14, 37 },
		{ 469, 15, 37 },
		{ 469, 16, 37 },
		{ 469, 17, 37 },
		{ 469, 18, 37 },
		{ 469, 19, 37 },
		{ 469, 20, 37 },
		{ 469, 21, 37 },
		{ 469, 22, 37 },
		{ 469, 23, 37 },
		{ 469, 24, 37 },
		{ 469, 25, 37 },
		{ 469, 26, 37 },
		{ 469, 27, 37 },
		{ 469, 28, 37 },
		{ 469, 29, 37 },
		{ 469, 30, 37 },
		{ 469, 31, 37 },
		{ 469, 32, 37 },
		{ 469, 33, 37 },
		{ 469, 34, 37 },
		{ 469, 35, 37 },
		{ 469, 36, 37 },
		{ 469, 37, 37 },
		{ 469, 38, 37 },
		{ 469, 39, 37 },
		{ 469, 40, 37 },
		{ 469, 41, 37 },
		{ 469, 42, 37 },
		{ 469, 43, 37 },
		{ 469, 44, 37 },
		{ 470, 1, 37 },
		{ 470, 2, 37 },
		{ 470, 3, 37 },
		{ 470, 4, 37 },
		{ 470, 5, 37 },
		{ 470, 6, 37 },
		{ 470, 7, 37 },
		{ 470, 8, 37 },
		{ 470, 9, 37 },
		{ 470, 10, 37 },
		{ 470, 11, 37 },
		{ 470, 12, 37 },
		{ 470, 13, 37 },
		{ 470, 14, 37 },
		{ 470, 15, 37 },
		{ 470, 16, 37 },
		{ 470, 17, 37 },
		{ 470, 18, 37 },
		{ 470, 19, 37 },
		{ 470, 20, 37 },
		{ 470, 21, 37 },
		{ 470, 22, 37 },
		{ 470, 23, 37 },
		{ 470, 24, 37 },
		{ 470, 25, 37 },
		{ 470, 26, 37 },
		{ 470, 27, 37 },
		{ 470, 28, 37 },
		{ 470, 29, 37 },
		{ 470, 30, 37 },
		{ 470, 31, 37 },
		{ 470, 32, 37 },
		{ 470, 33, 37 },
		{ 470, 34, 37 },
		{ 470, 35, 37 },
		{ 470, 36, 37 },
		{ 470, 37, 37 },
		{ 470, 38, 37 },
		{ 470, 39, 37 },
		{ 470, 40, 37 },
		{ 470, 41, 37 },
		{ 470, 42, 37 },
		{ 470, 43, 37 },
		{ 471, 1, 37 },
		{ 471, 2, 37 },
		{ 471, 3, 37 },
		{ 471, 4, 37 },
		{ 471, 5, 37 },
		{ 471, 6, 37 },
		{ 471, 7, 37 },
		{ 471, 8, 37 },
		{ 471, 9, 37 },
		{ 471, 10, 37 },
		{ 471, 11, 37 },
		{ 471, 12, 37 },
		{ 471, 13, 37 },
		{ 471, 14, 37 },
		{ 471, 15, 37 },
		{ 471, 16, 37 },
		{ 471, 17, 37 },
		{ 471, 18, 37 },
		{ 471, 19, 37 },
		{ 471, 20, 37 },
		{ 471, 21, 37 },
		{ 471, 22, 37 },
		{ 471, 23, 37 },
		{ 471, 24, 37 },
		{ 471, 25, 37 },
		{ 471, 26, 37 },
		{ 471, 27, 37 },
		{ 471, 28, 37 },
		{ 471, 29, 37 },
		{ 471, 30, 37 },
		{ 471, 31, 37 },
		{ 471, 32, 37 },
		{ 471, 33, 37 },
		{ 471, 34, 37 },
		{ 471, 35, 37 },
		{ 471, 36, 37 },
		{ 471, 37, 37 },
		{ 471, 38, 37 },
		{ 471, 39, 37 },
		{ 471, 40, 37 },
		{ 471, 41, 37 },
		{ 471, 42, 37 },
		{ 472, 1, 37 },
		{ 472, 2, 37 },
		{ 472, 3, 37 },
		{ 472, 4, 37 },
		{ 472, 5, 37 },
		{ 472, 6, 37 },
		{ 472, 7, 37 },
		{ 472, 8, 37 },
		{ 472, 9, 37 },
		{ 472, 10, 37 },
		{ 472, 11, 37 },
		{ 472, 12, 37 },
		{ 472, 13, 37 },
		{ 472, 14, 37 },
		{ 472, 15, 37 },
		{ 472, 16, 37 },
		{ 472, 17, 37 },
		{ 472, 18, 37 },
		{ 472, 19, 37 },
		{ 472, 20, 37 },
		{ 472, 21, 37 },
		{ 472, 22, 37 },
		{ 472, 23, 37 },
		{ 472, 24, 37 },
		{ 472, 25, 37 },
		{ 472, 26, 37 },
		{ 472, 27, 37 },
		{ 472, 28, 37 },
		{ 472, 29, 37 },
		{ 472, 30, 37 },
		{ 472, 31, 37 },
		{ 472, 32, 37 },
		{ 472, 33, 37 },
		{ 472, 34, 37 },
		{ 472, 35, 37 },
		{ 472, 36, 37 },
		{ 472, 37, 37 },
		{ 472, 38, 37 },
		{ 472, 39, 37 },
		{ 472, 40, 37 },
		{ 472, 41, 37 },
		{ 473, 1, 37 },
		{ 473, 2, 37 },
		{ 473, 3, 37 },
		{ 473, 4, 37 },
		{ 473, 5, 37 },
		{ 473, 6, 37 },
		{ 473, 7, 37 },
		{ 473, 8, 37 },
		{ 473, 9, 37 },
		{ 473, 10, 37 },
		{ 473, 11, 37 },
		{ 473, 12, 37 },
		{ 473, 13, 37 },
		{ 473, 14, 37 },
		{ 473, 15, 37 },
		{ 473, 16, 37 },
		{ 473, 17, 37 },
		{ 473, 18, 37 },
		{ 473, 19, 37 },
		{ 473, 20, 37 },
		{ 473, 21, 37 },
		{ 473, 22, 37 },
		{ 473, 23, 37 },
		{ 473, 24, 37 },
		{ 473, 25, 37 },
		{ 473, 26, 37 },
		{ 473, 27, 37 },
		{ 473, 28, 37 },
		{ 473, 29, 37 },
		{ 473, 30, 37 },
		{ 473, 31, 37 },
		{ 473, 32, 37 },
		{ 473, 33, 37 },
		{ 473, 34, 37 },
		{ 473, 35, 37 },
		{ 473, 36, 37 },
		{ 473, 37, 37 },
		{ 473, 38, 37 },
		{ 473, 39, 37 },
		{ 473, 40, 37 },
		{ 474, 1, 37 },
		{ 474, 2, 37 },
		{ 474, 3, 37 },
		{ 474, 4, 37 },
		{ 474, 5, 37 },
		{ 474, 6, 37 },
		{ 474, 7, 37 },
		{ 474, 8, 37 },
		{ 474, 9, 37 },
		{ 474, 10, 37 },
		{ 474, 11, 37 },
		{ 474, 12, 37 },
		{ 474, 13, 37 },
		{ 474, 14, 37 },
		{ 474, 15, 37 },
		{ 474, 16, 37 },
		{ 474, 17, 37 },
		{ 474, 18, 37 },
		{ 474, 19, 37 },
		{ 474, 20, 37 },
		{ 474, 21, 37 },
		{ 474, 22, 37 },
		{ 474, 23, 37 },
		{ 474, 24, 37 },
		{ 474, 25, 37 },
		{ 474, 26, 37 },
		{ 474, 27, 37 },
		{ 474, 28, 37 },
		{ 474, 29, 37 },
		{ 474, 30, 37 },
		{ 474, 31, 37 },
		{ 474, 32, 37 },
		{ 474, 33, 37 },
		{ 474, 34, 37 },
		{ 474, 35, 37 },
		{ 474, 36, 37 },
		{ 474, 37, 37 },
		{ 474, 38, 37 },
		{ 474, 39, 37 },
		{ 475, 1, 37 },
		{ 475, 2, 37 },
		{ 475, 3, 37 },
		{ 475, 4, 37 },
		{ 475, 5, 37 },
		{ 475, 6, 37 },
		{ 475, 7, 37 },
		{ 475, 8, 37 },
		{ 475, 9, 37 },
		{ 475, 10, 37 },
		{ 475, 11, 37 },
		{ 475, 12, 37 },
		{ 475, 13, 37 },
		{ 475, 14, 37 },
		{ 475, 15, 37 },
		{ 475, 16, 37 },
		{ 475, 17, 37 },
		{ 475, 18, 37 },
		{ 475, 19, 37 },
		{ 475, 20, 37 },
		{ 475, 21, 37 },
		{ 475, 22, 37 },
		{ 475, 23, 37 },
		{ 475, 24, 37 },
		{ 475, 25, 37 },
		{ 475, 26, 37 },
		{ 475, 27, 37 },
		{ 475, 28, 37 },
		{ 475, 29, 37 },
		{ 475, 30, 37 },
		{ 475, 31, 37 },
		{ 475, 32, 37 },
		{ 475, 33, 37 },
		{ 475, 34, 37 },
		{ 475, 35, 37 },
		{ 475, 36, 37 },
		{ 475, 37, 37 },
		{ 475, 38, 37 },
		{ 476, 1, 37 },
		{ 476, 2, 37 },
		{ 476, 3, 37 },
		{ 476, 4, 37 },
		{ 476, 5, 37 },
		{ 476, 6, 37 },
		{ 476, 7, 37 },
		{ 476, 8, 37 },
		{ 476, 9, 37 },
		{ 476, 10, 37 },
		{ 476, 11, 37 },
		{ 476, 12, 37 },
		{ 476, 13, 37 },
		{ 476, 14, 37 },
		{ 476, 15, 37 },
		{ 476, 16, 37 },
		{ 476, 17, 37 },
		{ 476, 18, 37 },
		{ 476, 19, 37 },
		{ 476, 20, 37 },
		{ 476, 21, 37 },
		{ 476, 22, 37 },
		{ 476, 23, 37 },
		{ 476, 24, 37 },
		{ 476, 25, 37 },
		{ 476, 26, 37 },
		{ 476, 27, 37 },
		{ 476, 28, 37 },
		{ 476, 29, 37 },
		{ 476, 30, 37 },
		{ 476, 31, 37 },
		{ 476, 32, 37 },
		{ 476, 33, 37 },
		{ 476, 34, 37 },
		{ 476, 35, 37 },
		{ 476, 36, 37 },
		{ 476, 37, 37 },
		{ 477, 1, 37 },
		{ 477, 2, 37 },
		{ 477, 3, 37 },
		{ 477, 4, 37 },
		{ 477, 5, 37 },
		{ 477, 6, 37 },
		{ 477, 7, 37 },
		{ 477, 8, 37 },
		{ 477, 9, 37 },
		{ 477, 10, 37 },
		{ 477, 11, 37 },
		{ 477, 12, 37 },
		{ 477, 13, 37 },
		{ 477, 14, 37 },
		{ 477, 15, 37 },
		{ 477, 16, 37 },
		{ 477, 17, 37 },
		{ 477, 18, 37 },
		{ 477, 19, 37 },
		{ 477, 20, 37 },
		{ 477, 21, 37 },
		{ 477, 22, 37 },
		{ 477, 23, 37 },
		{ 477, 24, 37 },
		{ 477, 25, 37 },
		{ 477, 26, 37 },
		{ 477, 27, 37 },
		{ 477, 28, 37 },
		{ 477, 29, 37 },
		{ 477, 30, 37 },
		{ 477, 31, 37 },
		{ 477, 32, 37 },
		{ 477, 33, 37 },
		{ 477, 34, 37 },
		{ 477, 35, 37 },
		{ 477, 36, 37 },
		{ 478, 1, 37 },
		{ 478, 2, 37 },
		{ 478, 3, 37 },
		{ 478, 4, 37 },
		{ 478, 5, 37 },
		{ 478, 6, 37 },
		{ 478, 7, 37 },
		{ 478, 8, 37 },
		{ 478, 9, 37 },
		{ 478, 10, 37 },
		{ 478, 11, 37 },
		{ 478, 12, 37 },
		{ 478, 13, 37 },
		{ 478, 14, 37 },
		{ 478, 15, 37 },
		{ 478, 16, 37 },
		{ 478, 17, 37 },
		{ 478, 18, 37 },
		{ 478, 19, 37 },
		{ 478, 20, 37 },
		{ 478, 21, 37 },
		{ 478, 22, 37 },
		{ 478, 23, 37 },
		{ 478, 24, 37 },
		{ 478, 25, 37 },
		{ 478, 26, 37 },
		{ 478, 27, 37 },
		{ 478, 28, 37 },
		{ 478, 29, 37 },
		{ 478, 30, 37 },
		{ 478, 31, 37 },
		{ 478, 32, 37 },
		{ 478, 33, 37 },
		{ 478, 34, 37 },
		{ 478, 35, 37 },
		{ 479, 1, 37 },
		{ 479, 2, 37 },
		{ 479, 3, 37 },
		{ 479, 4, 37 },
		{ 479, 5, 37 },
		{ 479, 6, 37 },
		{ 479, 7, 37 },
		{ 479, 8, 37 },
		{ 479, 9, 37 },
		{ 479, 10, 37 },
		{ 479, 11, 37 },
		{ 479, 12, 37 },
		{ 479, 13, 37 },
		{ 479, 14, 37 },
		{ 479, 15, 37 },
		{ 479, 16, 37 },
		{ 479, 17, 37 },
		{ 479, 18, 37 },
		{ 479, 19, 37 },
		{ 479, 20, 37 },
		{ 479, 21, 37 },
		{ 479, 22, 37 },
		{ 479, 23, 37 },
		{ 479, 24, 37 },
		{ 479, 25, 37 },
		{ 479, 26, 37 },
		{ 479, 27, 37 },
		{ 479, 28, 37 },
		{ 479, 29, 37 },
		{ 479, 30, 37 },
		{ 479, 31, 37 },
		{ 479, 32, 37 },
		{ 479, 33, 37 },
		{ 479, 34, 37 },
		{ 480, 1, 37 },
		{ 480, 2, 37 },
		{ 480, 3, 37 },
		{ 480, 4, 37 },
		{ 480, 5, 37 },
		{ 480, 6, 37 },
		{ 480, 7, 37 },
		{ 480, 8, 37 },
		{ 480, 9, 37 },
		{ 480, 10, 37 },
		{ 480, 11, 37 },
		{ 480, 12, 37 },
		{ 480, 13, 37 },
		{ 480, 14, 37 },
		{ 480, 15, 37 },
		{ 480, 16, 37 },
		{ 480, 17, 37 },
		{ 480, 18, 37 },
		{ 480, 19, 37 },
		{ 480, 20, 37 },
		{ 480, 21, 37 },
		{ 480, 22, 37 },
		{ 480, 23, 37 },
		{ 480, 24, 37 },
		{ 480, 25, 37 },
		{ 480, 26, 37 },
		{ 480, 27, 37 },
		{ 480, 28, 37 },
		{ 480, 29, 37 },
		{ 480, 30, 37 },
		{ 480, 31, 37 },
		{ 480, 32, 37 },
		{ 480, 33, 37 },
		{ 481, 1, 37 },
		{ 481, 2, 37 },
		{ 481, 3, 37 },
		{ 481, 4, 37 },
		{ 481, 5, 37 },
		{ 481, 6, 37 },
		{ 481, 7, 37 },
		{ 481, 8, 37 },
		{ 481, 9, 37 },
		{ 481, 10, 37 },
		{ 481, 11, 37 },
		{ 481, 12, 37 },
		{ 481, 13, 37 },
		{ 481, 14, 37 },
		{ 481, 15, 37 },
		{ 481, 16, 37 },
		{ 481, 17, 37 },
		{ 481, 18, 37 },
		{ 481, 19, 37 },
		{ 481, 20, 37 },
		{ 481, 21, 37 },
		{ 481, 22, 37 },
		{ 481, 23, 37 },
		{ 481, 24, 37 },
		{ 481, 25, 37 },
		{ 481, 26, 37 },
		{ 481, 27, 37 },
		{ 481, 28, 37 },
		{ 481, 29, 37 },
		{ 481, 30, 37 },
		{ 481, 31, 37 },
		{ 481, 32, 37 },
		{ 482, 1, 37 },
		{ 482, 2, 37 },
		{ 482, 3, 37 },
		{ 482, 4, 37 },
		{ 482, 5, 37 },
		{ 482, 6, 37 },
		{ 482, 7, 37 },
		{ 482, 8, 37 },
		{ 482, 9, 37 },
		{ 482, 10, 37 },
		{ 482, 11, 37 },
		{ 482, 12, 37 },
		{ 482, 13, 37 },
		{ 482, 14, 37 },
		{ 482, 15, 37 },
		{ 482, 16, 37 },
		{ 482, 17, 37 },
		{ 482, 18, 37 },
		{ 482, 19, 37 },
		{ 482, 20, 37 },
		{ 482, 21, 37 },
		{ 482, 22, 37 },
		{ 482, 23, 37 },
		{ 482, 24, 37 },
		{ 482, 25, 37 },
		{ 482, 26, 37 },
		{ 482, 27, 37 },
		{ 482, 28, 37 },
		{ 482, 29, 37 },
		{ 482, 30, 37 },
		{ 482, 31, 37 },
		{ 483, 1, 37 },
		{ 483, 2, 37 },
		{ 483, 3, 37 },
		{ 483, 4, 37 },
		{ 483, 5, 37 },
		{ 483, 6, 37 },
		{ 483, 7, 37 },
		{ 483, 8, 37 },
		{ 483, 9, 37 },
		{ 483, 10, 37 },
		{ 483, 11, 37 },
		{ 483, 12, 37 },
		{ 483, 13, 37 },
		{ 483, 14, 37 },
		{ 483, 15, 37 },
		{ 483, 16, 37 },
		{ 483, 17, 37 },
		{ 483, 18, 37 },
		{ 483, 19, 37 },
		{ 483, 20, 37 },
		{ 483, 21, 37 },
		{ 483, 22, 37 },
		{ 483, 23, 37 },
		{ 483, 24, 37 },
		{ 483, 25, 37 },
		{ 483, 26, 37 },
		{ 483, 27, 37 },
		{ 483, 28, 37 },
		{ 483, 29, 37 },
		{ 483, 30, 37 },
		{ 484, 1, 37 },
		{ 484, 2, 37 },
		{ 484, 3, 37 },
		{ 484, 4, 37 },
		{ 484, 5, 37 },
		{ 484, 6, 37 },
		{ 484, 7, 37 },
		{ 484, 8, 37 },
		{ 484, 9, 37 },
		{ 484, 10, 37 },
		{ 484, 11, 37 },
		{ 484, 12, 37 },
		{ 484, 13, 37 },
		{ 484, 14, 37 },
		{ 484, 15, 37 },
		{ 484, 16, 37 },
		{ 484, 17, 37 },
		{ 484, 18, 37 },
		{ 484, 19, 37 },
		{ 484, 20, 37 },
		{ 484, 21, 37 },
		{ 484, 22, 37 },
		{ 484, 23, 37 },
		{ 484, 24, 37 },
		{ 484, 25, 37 },
		{ 484, 26, 37 },
		{ 484, 27, 37 },
		{ 484, 28, 37 },
		{ 484, 29, 37 },
		{ 485, 1, 37 },
		{ 485, 2, 37 },
		{ 485, 3, 37 },
		{ 485, 4, 37 },
		{ 485, 5, 37 },
		{ 485, 6, 37 },
		{ 485, 7, 37 },
		{ 485, 8, 37 },
		{ 485, 9, 37 },
		{ 485, 10, 37 },
		{ 485, 11, 37 },
		{ 485, 12, 37 },
		{ 485, 13, 37 },
		{ 485, 14, 37 },
		{ 485, 15, 37 },
		{ 485, 16, 37 },
		{ 485, 17, 37 },
		{ 485, 18, 37 },
		{ 485, 19, 37 },
		{ 485, 20, 37 },
		{ 485, 21, 37 },
		{ 485, 22, 37 },
		{ 485, 23, 37 },
		{ 485, 24, 37 },
		{ 485, 25, 37 },
		{ 485, 26, 37 },
		{ 485, 27, 37 },
		{ 485, 28, 37 },
		{ 486, 1, 37 },
		{ 486, 2, 37 },
		{ 486, 3, 37 },
		{ 486, 4, 37 },
		{ 486, 5, 37 },
		{ 486, 6, 37 },
		{ 486, 7, 37 },
		{ 486, 8, 37 },
		{ 486, 9, 37 },
		{ 486, 10, 37 },
		{ 486, 11, 37 },
		{ 486, 12, 37 },
		{ 486, 13, 37 },
		{ 486, 14, 37 },
		{ 486, 15, 37 },
		{ 486, 16, 37 },
		{ 486, 17, 37 },
		{ 486, 18, 37 },
		{ 486, 19, 37 },
		{ 486, 20, 37 },
		{ 486, 21, 37 },
		{ 486, 22, 37 },
		{ 486, 23, 37 },
		{ 486, 24, 37 },
		{ 486, 25, 37 },
		{ 486, 26, 37 },
		{ 486, 27, 37 },
		{ 487, 1, 37 },
		{ 487, 2, 37 },
		{ 487, 3, 37 },
		{ 487, 4, 37 },
		{ 487, 5, 37 },
		{ 487, 6, 37 },
		{ 487, 7, 37 },
		{ 487, 8, 37 },
		{ 487, 9, 37 },
		{ 487, 10, 37 },
		{ 487, 11, 37 },
		{ 487, 12, 37 },
		{ 487, 13, 37 },
		{ 487, 14, 37 },
		{ 487, 15, 37 },
		{ 487, 16, 37 },
		{ 487, 17, 37 },
		{ 487, 18, 37 },
		{ 487, 19, 37 },
		{ 487, 20, 37 },
		{ 487, 21, 37 },
		{ 487, 22, 37 },
		{ 487, 23, 37 },
		{ 487, 24, 37 },
		{ 487, 25, 37 },
		{ 487, 26, 37 },
		{ 488, 1, 37 },
		{ 488, 2, 37 },
		{ 488, 3, 37 },
		{ 488, 4, 37 },
		{ 488, 5, 37 },
		{ 488, 6, 37 },
		{ 488, 7, 37 },
		{ 488, 8, 37 },
		{ 488, 9, 37 },
		{ 488, 10, 37 },
		{ 488, 11, 37 },
		{ 488, 12, 37 },
		{ 488, 13, 37 },
		{ 488, 14, 37 },
		{ 488, 15, 37 },
		{ 488, 16, 37 },
		{ 488, 17, 37 },
		{ 488, 18, 37 },
		{ 488, 19, 37 },
		{ 488, 20, 37 },
		{ 488, 21, 37 },
		{ 488, 22, 37 },
		{ 488, 23, 37 },
		{ 488, 24, 37 },
		{ 488, 25, 37 },
		{ 489, 1, 37 },
		{ 489, 2, 37 },
		{ 489, 3, 37 },
		{ 489, 4, 37 },
		{ 489, 5, 37 },
		{ 489, 6, 37 },
		{ 489, 7, 37 },
		{ 489, 8, 37 },
		{ 489, 9, 37 },
		{ 489, 10, 37 },
		{ 489, 11, 37 },
		{ 489, 12, 37 },
		{ 489, 13, 37 },
		{ 489, 14, 37 },
		{ 489, 15, 37 },
		{ 489, 16, 37 },
		{ 489, 17, 37 },
		{ 489, 18, 37 },
		{ 489, 19, 37 },
		{ 489, 20, 37 },
		{ 489, 21, 37 },
		{ 489, 22, 37 },
		{ 489, 23, 37 },
		{ 489, 24, 37 },
		{ 490, 1, 37 },
		{ 490, 2, 37 },
		{ 490, 3, 37 },
		{ 490, 4, 37 },
		{ 490, 5, 37 },
		{ 490, 6, 37 },
		{ 490, 7, 37 },
		{ 490, 8, 37 },
		{ 490, 9, 37 },
		{ 490, 10, 37 },
		{ 490, 11, 37 },
		{ 490, 12, 37 },
		{ 490, 13, 37 },
		{ 490, 14, 37 },
		{ 490, 15, 37 },
		{ 490, 16, 37 },
		{ 490, 17, 37 },
		{ 490, 18, 37 },
		{ 490, 19, 37 },
		{ 490, 20, 37 },
		{ 490, 21, 37 },
		{ 490, 22, 37 },
		{ 490, 23, 37 },
		{ 491, 1, 37 },
		{ 491, 2, 37 },
		{ 491, 3, 37 },
		{ 491, 4, 37 },
		{ 491, 5, 37 },
		{ 491, 6, 37 },
		{ 491, 7, 37 },
		{ 491, 8, 37 },
		{ 491, 9, 37 },
		{ 491, 10, 37 },
		{ 491, 11, 37 },
		{ 491, 12, 37 },
		{ 491, 13, 37 },
		{ 491, 14, 37 },
		{ 491, 15, 37 },
		{ 491, 16, 37 },
		{ 491, 17, 37 },
		{ 491, 18, 37 },
		{ 491, 19, 37 },
		{ 491, 20, 37 },
		{ 491, 21, 37 },
		{ 491, 22, 37 },
		{ 492, 1, 37 },
		{ 492, 2, 37 },
		{ 492, 3, 37 },
		{ 492, 4, 37 },
		{ 492, 5, 37 },
		{ 492, 6, 37 },
		{ 492, 7, 37 },
		{ 492, 8, 37 },
		{ 492, 9, 37 },
		{ 492, 10, 37 },
		{ 492, 11, 37 },
		{ 492, 12, 37 },
		{ 492, 13, 37 },
		{ 492, 14, 37 },
		{ 492, 15, 37 },
		{ 492, 16, 37 },
		{ 492, 17, 37 },
		{ 492, 18, 37 },
		{ 492, 19, 37 },
		{ 492, 20, 37 },
		{ 492, 21, 37 },
		{ 493, 1, 37 },
		{ 493, 2, 37 },
		{ 493, 3, 37 },
		{ 493, 4, 37 },
		{ 493, 5, 37 },
		{ 493, 6, 37 },
		{ 493, 7, 37 },
		{ 493, 8, 37 },
		{ 493, 9, 37 },
		{ 493, 10, 37 },
		{ 493, 11, 37 },
		{ 493, 12, 37 },
		{ 493, 13, 37 },
		{ 493, 14, 37 },
		{ 493, 15, 37 },
		{ 493, 16, 37 },
		{ 493, 17, 37 },
		{ 493, 18, 37 },
		{ 493, 19, 37 },
		{ 493, 20, 37 },
		{ 494, 1, 37 },
		{ 494, 2, 37 },
		{ 494, 3, 37 },
		{ 494, 4, 37 },
		{ 494, 5, 37 },
		{ 494, 6, 37 },
		{ 494, 7, 37 },
		{ 494, 8, 37 },
		{ 494, 9, 37 },
		{ 494, 10, 37 },
		{ 494, 11, 37 },
		{ 494, 12, 37 },
		{ 494, 13, 37 },
		{ 494, 14, 37 },
		{ 494, 15, 37 },
		{ 494, 16, 37 },
		{ 494, 17, 37 },
		{ 494, 18, 37 },
		{ 494, 19, 37 },
		{ 495, 1, 37 },
		{ 495, 2, 37 },
		{ 495, 3, 37 },
		{ 495, 4, 37 },
		{ 495, 5, 37 },
		{ 495, 6, 37 },
		{ 495, 7, 37 },
		{ 495, 8, 37 },
		{ 495, 9, 37 },
		{ 495, 10, 37 },
		{ 495, 11, 37 },
		{ 495, 12, 37 },
		{ 495, 13, 37 },
		{ 495, 14, 37 },
		{ 495, 15, 37 },
		{ 495, 16, 37 },
		{ 495, 17, 37 },
		{ 495, 18, 37 },
		{ 496, 1, 37 },
		{ 496, 2, 37 },
		{ 496, 3, 37 },
		{ 496, 4, 37 },
		{ 496, 5, 37 },
		{ 496, 6, 37 },
		{ 496, 7, 37 },
		{ 496, 8, 37 },
		{ 496, 9, 37 },
		{ 496, 10, 37 },
		{ 496, 11, 37 },
		{ 496, 12, 37 },
		{ 496, 13, 37 },
		{ 496, 14, 37 },
		{ 496, 15, 37 },
		{ 496, 16, 37 },
		{ 496, 17, 37 },
		{ 497, 1, 37 },
		{ 497, 2, 37 },
		{ 497, 3, 37 },
		{ 497, 4, 37 },
		{ 497, 5, 37 },
		{ 497, 6, 37 },
		{ 497, 7, 37 },
		{ 497, 8, 37 },
		{ 497, 9, 37 },
		{ 497, 10, 37 },
		{ 497, 11, 37 },
		{ 497, 12, 37 },
		{ 497, 13, 37 },
		{ 497, 14, 37 },
		{ 497, 15, 37 },
		{ 497, 16, 37 },
		{ 498, 1, 37 },
		{ 498, 2, 37 },
		{ 498, 3, 37 },
		{ 498, 4, 37 },
		{ 498, 5, 37 },
		{ 498, 6, 37 },
		{ 498, 7, 37 },
		{ 498, 8, 37 },
		{ 498, 9, 37 },
		{ 498, 10, 37 },
		{ 498, 11, 37 },
		{ 498, 12, 37 },
		{ 498, 13, 37 },
		{ 498, 14, 37 },
		{ 498, 15, 37 },
		{ 499, 1, 37 },
		{ 499, 2, 37 },
		{ 499, 3, 37 },
		{ 499, 4, 37 },
		{ 499, 5, 37 },
		{ 499, 6, 37 },
		{ 499, 7, 37 },
		{ 499, 8, 37 },
		{ 499, 9, 37 },
		{ 499, 10, 37 },
		{ 499, 11, 37 },
		{ 499, 12, 37 },
		{ 499, 13, 37 },
		{ 499, 14, 37 },
		{ 500, 1, 37 },
		{ 500, 2, 37 },
		{ 500, 3, 37 },
		{ 500, 4, 37 },
		{ 500, 5, 37 },
		{ 500, 6, 37 },
		{ 500, 7, 37 },
		{ 500, 8, 37 },
		{ 500, 9, 37 },
		{ 500, 10, 37 },
		{ 500, 11, 37 },
		{ 500, 12, 37 },
		{ 500, 13, 37 },
		{ 501, 1, 37 },
		{ 501, 2, 37 },
		{ 501, 3, 37 },
		{ 501, 4, 37 },
		{ 501, 5, 37 },
		{ 501, 6, 37 },
		{ 501, 7, 37 },
		{ 501, 8, 37 },
		{ 501, 9, 37 },
		{ 501, 10, 37 },
		{ 501, 11, 37 },
		{ 501, 12, 37 },
		{ 502, 1, 37 },
		{ 502, 2, 37 },
		{ 502, 3, 37 },
		{ 502, 4, 37 },
		{ 502, 5, 37 },
		{ 502, 6, 37 },
		{ 502, 7, 37 },
		{ 502, 8, 37 },
		{ 502, 9, 37 },
		{ 502, 10, 37 },
		{ 502, 11, 37 },
		{ 503, 1, 37 },
		{ 503, 2, 37 },
		{ 503, 3, 37 },
		{ 503, 4, 37 },
		{ 503, 5, 37 },
		{ 503, 6, 37 },
		{ 503, 7, 37 },
		{ 503, 8, 37 },
		{ 503, 9, 37 },
		{ 503, 10, 37 },
		{ 504, 1, 37 },
		{ 504, 2, 37 },
		{ 504, 3, 37 },
		{ 504, 4, 37 },
		{ 504, 5, 37 },
		{ 504, 6, 37 },
		{ 504, 7, 37 },
		{ 504, 8, 37 },
		{ 504, 9, 37 },
		{ 505, 1, 37 },
		{ 505, 2, 37 },
		{ 505, 3, 37 },
		{ 505, 4, 37 },
		{ 505, 5, 37 },
		{ 505, 6, 37 },
		{ 505, 7, 37 },
		{ 505, 8, 37 },
		{ 506, 1, 37 },
		{ 506, 2, 37 },
		{ 506, 3, 37 },
		{ 506, 4, 37 },
		{ 506, 5, 37 },
		{ 506, 6, 37 },
		{ 506, 7, 37 },
		{ 507, 1, 37 },
		{ 507, 2, 37 },
		{ 507, 3, 37 },
		{ 507, 4, 37 },
		{ 507, 5, 37 },
		{ 507, 6, 37 },
		{ 508, 1, 37 },
		{ 508, 2, 37 },
		{ 508, 3, 37 },
		{ 508, 4, 37 },
		{ 508, 5, 37 },
		{ 509, 1, 37 },
		{ 509, 2, 37 },
		{ 509, 3, 37 },
		{ 509, 4, 37 },
		{ 510, 1, 37 },
		{ 510, 2, 37 },
		{ 510, 3, 37 },
		{ 511, 1, 37 },
		{ 511, 2, 37 },
		{ 512, 1, 37 },
		{ 1, 482, 38 },
		{ 1, 483, 38 },
		{ 1, 484, 38 },
		{ 1, 485, 38 },
		{ 1, 486, 38 },
		{ 1, 487, 38 },
		{ 1, 488, 38 },
		{ 1, 489, 38 },
		{ 1, 490, 38 },
		{ 1, 491, 38 },
		{ 1, 492, 38 },
		{ 1, 493, 38 },
		{ 1, 494, 38 },
		{ 1, 495, 38 },
		{ 1, 496, 38 },
		{ 1, 497, 38 },
		{ 1, 498, 38 },
		{ 1, 499, 38 },
		{ 1, 500, 38 },
		{ 1, 501, 38 },
		{ 1, 502, 38 },
		{ 1, 503, 38 },
		{ 1, 504, 38 },
		{ 1, 505, 38 },
		{ 1, 506, 38 },
		{ 1, 507, 38 },
		{ 1, 508, 38 },
		{ 1, 509, 38 },
		{ 1, 510, 38 },
		{ 1, 511, 38 },
		{ 1, 512, 38 },
		{ 2, 481, 38 },
		{ 2, 482, 38 },
		{ 2, 483, 38 },
		{ 2, 484, 38 },
		{ 2, 485, 38 },
		{ 2, 486, 38 },
		{ 2, 487, 38 },
		{ 2, 488, 38 },
		{ 2, 489, 38 },
		{ 2, 490, 38 },
		{ 2, 491, 38 },
		{ 2, 492, 38 },
		{ 2, 493, 38 },
		{ 2, 494, 38 },
		{ 2, 495, 38 },
		{ 2, 496, 38 },
		{ 2, 497, 38 },
		{ 2, 498, 38 },
		{ 2, 499, 38 },
		{ 2, 500, 38 },
		{ 2, 501, 38 },
		{ 2, 502, 38 },
		{ 2, 503, 38 },
		{ 2, 504, 38 },
		{ 2, 505, 38 },
		{ 2, 506, 38 },
		{ 2, 507, 38 },
		{ 2, 508, 38 },
		{ 2, 509, 38 },
		{ 2, 510, 38 },
		{ 2, 511, 38 },
		{ 2, 512, 38 },
		{ 3, 480, 38 },
		{ 3, 481, 38 },
		{ 3, 482, 38 },
		{ 3, 483, 38 },
		{ 3, 484, 38 },
		{ 3, 485, 38 },
		{ 3, 486, 38 },
		{ 3, 487, 38 },
		{ 3, 488, 38 },
		{ 3, 489, 38 },
		{ 3, 490, 38 },
		{ 3, 491, 38 },
		{ 3, 492, 38 },
		{ 3, 493, 38 },
		{ 3, 494, 38 },
		{ 3, 495, 38 },
		{ 3, 496, 38 },
		{ 3, 497, 38 },
		{ 3, 498, 38 },
		{ 3, 499, 38 },
		{ 3, 500, 38 },
		{ 3, 501, 38 },
		{ 3, 502, 38 },
		{ 3, 503, 38 },
		{ 3, 504, 38 },
		{ 3, 505, 38 },
		{ 3, 506, 38 },
		{ 3, 507, 38 },
		{ 3, 508, 38 },
		{ 3, 509, 38 },
		{ 3, 510, 38 },
		{ 3, 511, 38 },
		{ 3, 512, 38 },
		{ 4, 479, 38 },
		{ 4, 480, 38 },
		{ 4, 481, 38 },
		{ 4, 482, 38 },
		{ 4, 483, 38 },
		{ 4, 484, 38 },
		{ 4, 485, 38 },
		{ 4, 486, 38 },
		{ 4, 487, 38 },
		{ 4, 488, 38 },
		{ 4, 489, 38 },
		{ 4, 490, 38 },
		{ 4, 491, 38 },
		{ 4, 492, 38 },
		{ 4, 493, 38 },
		{ 4, 494, 38 },
		{ 4, 495, 38 },
		{ 4, 496, 38 },
		{ 4, 497, 38 },
		{ 4, 498, 38 },
		{ 4, 499, 38 },
		{ 4, 500, 38 },
		{ 4, 501, 38 },
		{ 4, 502, 38 },
		{ 4, 503, 38 },
		{ 4, 504, 38 },
		{ 4, 505, 38 },
		{ 4, 506, 38 },
		{ 4, 507, 38 },
		{ 4, 508, 38 },
		{ 4, 509, 38 },
		{ 4, 510, 38 },
		{ 4, 511, 38 },
		{ 4, 512, 38 },
		{ 5, 478, 38 },
		{ 5, 479, 38 },
		{ 5, 480, 38 },
		{ 5, 481, 38 },
		{ 5, 482, 38 },
		{ 5, 483, 38 },
		{ 5, 484, 38 },
		{ 5, 485, 38 },
		{ 5, 486, 38 },
		{ 5, 487, 38 },
		{ 5, 488, 38 },
		{ 5, 489, 38 },
		{ 5, 490, 38 },
		{ 5, 491, 38 },
		{ 5, 492, 38 },
		{ 5, 493, 38 },
		{ 5, 494, 38 },
		{ 5, 495, 38 },
		{ 5, 496, 38 },
		{ 5, 497, 38 },
		{ 5, 498, 38 },
		{ 5, 499, 38 },
		{ 5, 500, 38 },
		{ 5, 501, 38 },
		{ 5, 502, 38 },
		{ 5, 503, 38 },
		{ 5, 504, 38 },
		{ 5, 505, 38 },
		{ 5, 506, 38 },
		{ 5, 507, 38 },
		{ 5, 508, 38 },
		{ 5, 509, 38 },
		{ 5, 510, 38 },
		{ 5, 511, 38 },
		{ 5, 512, 38 },
		{ 6, 477, 38 },
		{ 6, 478, 38 },
		{ 6, 479, 38 },
		{ 6, 480, 38 },
		{ 6, 481, 38 },
		{ 6, 482, 38 },
		{ 6, 483, 38 },
		{ 6, 484, 38 },
		{ 6, 485, 38 },
		{ 6, 486, 38 },
		{ 6, 487, 38 },
		{ 6, 488, 38 },
		{ 6, 489, 38 },
		{ 6, 490, 38 },
		{ 6, 491, 38 },
		{ 6, 492, 38 },
		{ 6, 493, 38 },
		{ 6, 494, 38 },
		{ 6, 495, 38 },
		{ 6, 496, 38 },
		{ 6, 497, 38 },
		{ 6, 498, 38 },
		{ 6, 499, 38 },
		{ 6, 500, 38 },
		{ 6, 501, 38 },
		{ 6, 502, 38 },
		{ 6, 503, 38 },
		{ 6, 504, 38 },
		{ 6, 505, 38 },
		{ 6, 506, 38 },
		{ 6, 507, 38 },
		{ 6, 508, 38 },
		{ 6, 509, 38 },
		{ 6, 510, 38 },
		{ 6, 511, 38 },
		{ 6, 512, 38 },
		{ 7, 476, 38 },
		{ 7, 477, 38 },
		{ 7, 478, 38 },
		{ 7, 479, 38 },
		{ 7, 480, 38 },
		{ 7, 481, 38 },
		{ 7, 482, 38 },
		{ 7, 483, 38 },
		{ 7, 484, 38 },
		{ 7, 485, 38 },
		{ 7, 486, 38 },
		{ 7, 487, 38 },
		{ 7, 488, 38 },
		{ 7, 489, 38 },
		{ 7, 490, 38 },
		{ 7, 491, 38 },
		{ 7, 492, 38 },
		{ 7, 493, 38 },
		{ 7, 494, 38 },
		{ 7, 495, 38 },
		{ 7, 496, 38 },
		{ 7, 497, 38 },
		{ 7, 498, 38 },
		{ 7, 499, 38 },
		{ 7, 500, 38 },
		{ 7, 501, 38 },
		{ 7, 502, 38 },
		{ 7, 503, 38 },
		{ 7, 504, 38 },
		{ 7, 505, 38 },
		{ 7, 506, 38 },
		{ 7, 507, 38 },
		{ 7, 508, 38 },
		{ 7, 509, 38 },
		{ 7, 510, 38 },
		{ 7, 511, 38 },
		{ 7, 512, 38 },
		{ 8, 475, 38 },
		{ 8, 476, 38 },
		{ 8, 477, 38 },
		{ 8, 478, 38 },
		{ 8, 479, 38 },
		{ 8, 480, 38 },
		{ 8, 481, 38 },
		{ 8, 482, 38 },
		{ 8, 483, 38 },
		{ 8, 484, 38 },
		{ 8, 485, 38 },
		{ 8, 486, 38 },
		{ 8, 487, 38 },
		{ 8, 488, 38 },
		{ 8, 489, 38 },
		{ 8, 490, 38 },
		{ 8, 491, 38 },
		{ 8, 492, 38 },
		{ 8, 493, 38 },
		{ 8, 494, 38 },
		{ 8, 495, 38 },
		{ 8, 496, 38 },
		{ 8, 497, 38 },
		{ 8, 498, 38 },
		{ 8, 499, 38 },
		{ 8, 500, 38 },
		{ 8, 501, 38 },
		{ 8, 502, 38 },
		{ 8, 503, 38 },
		{ 8, 504, 38 },
		{ 8, 505, 38 },
		{ 8, 506, 38 },
		{ 8, 507, 38 },
		{ 8, 508, 38 },
		{ 8, 509, 38 },
		{ 8, 510, 38 },
		{ 8, 511, 38 },
		{ 8, 512, 38 },
		{ 9, 475, 38 },
		{ 9, 476, 38 },
		{ 9, 477, 38 },
		{ 9, 478, 38 },
		{ 9, 479, 38 },
		{ 9, 480, 38 },
		{ 9, 481, 38 },
		{ 9, 482, 38 },
		{ 9, 483, 38 },
		{ 9, 484, 38 },
		{ 9, 485, 38 },
		{ 9, 486, 38 },
		{ 9, 487, 38 },
		{ 9, 488, 38 },
		{ 9, 489, 38 },
		{ 9, 490, 38 },
		{ 9, 491, 38 },
		{ 9, 492, 38 },
		{ 9, 493, 38 },
		{ 9, 494, 38 },
		{ 9, 495, 38 },
		{ 9, 496, 38 },
		{ 9, 497, 38 },
		{ 9, 498, 38 },
		{ 9, 499, 38 },
		{ 9, 500, 38 },
		{ 9, 501, 38 },
		{ 9, 502, 38 },
		{ 9, 503, 38 },
		{ 9, 504, 38 },
		{ 9, 505, 38 },
		{ 9, 506, 38 },
		{ 9, 507, 38 },
		{ 9, 508, 38 },
		{ 9, 509, 38 },
		{ 9, 510, 38 },
		{ 9, 511, 38 },
		{ 9, 512, 38 },
		{ 10, 474, 38 },
		{ 10, 475, 38 },
		{ 10, 476, 38 },
		{ 10, 477, 38 },
		{ 10, 478, 38 },
		{ 10, 479, 38 },
		{ 10, 480, 38 },
		{ 10, 481, 38 },
		{ 10, 482, 38 },
		{ 10, 483, 38 },
		{ 10, 484, 38 },
		{ 10, 485, 38 },
		{ 10, 486, 38 },
		{ 10, 487, 38 },
		{ 10, 488, 38 },
		{ 10, 489, 38 },
		{ 10, 490, 38 },
		{ 10, 491, 38 },
		{ 10, 492, 38 },
		{ 10, 493, 38 },
		{ 10, 494, 38 },
		{ 10, 495, 38 },
		{ 10, 496, 38 },
		{ 10, 497, 38 },
		{ 10, 498, 38 },
		{ 10, 499, 38 },
		{ 10, 500, 38 },
		{ 10, 501, 38 },
		{ 10, 502, 38 },
		{ 10, 503, 38 },
		{ 10, 504, 38 },
		{ 10, 505, 38 },
		{ 10, 506, 38 },
		{ 10, 507, 38 },
		{ 10, 508, 38 },
		{ 10, 509, 38 },
		{ 10, 510, 38 },
		{ 10, 511, 38 },
		{ 10, 512, 38 },
		{ 11, 473, 38 },
		{ 11, 474, 38 },
		{ 11, 475, 38 },
		{ 11, 476, 38 },
		{ 11, 477, 38 },
		{ 11, 478, 38 },
		{ 11, 479, 38 },
		{ 11, 480, 38 },
		{ 11, 481, 38 },
		{ 11, 482, 38 },
		{ 11, 483, 38 },
		{ 11, 484, 38 },
		{ 11, 485, 38 },
		{ 11, 486, 38 },
		{ 11, 487, 38 },
		{ 11, 488, 38 },
		{ 11, 489, 38 },
		{ 11, 490, 38 },
		{ 11, 491, 38 },
		{ 11, 492, 38 },
		{ 11, 493, 38 },
		{ 11, 494, 38 },
		{ 11, 495, 38 },
		{ 11, 496, 38 },
		{ 11, 497, 38 },
		{ 11, 498, 38 },
		{ 11, 499, 38 },
		{ 11, 500, 38 },
		{ 11, 501, 38 },
		{ 11, 502, 38 },
		{ 11, 503, 38 },
		{ 11, 504, 38 },
		{ 11, 505, 38 },
		{ 11, 506, 38 },
		{ 11, 507, 38 },
		{ 11, 508, 38 },
		{ 11, 509, 38 },
		{ 11, 510, 38 },
		{ 11, 511, 38 },
		{ 11, 512, 38 },
		{ 12, 472, 38 },
		{ 12, 473, 38 },
		{ 12, 474, 38 },
		{ 12, 475, 38 },
		{ 12, 476, 38 },
		{ 12, 477, 38 },
		{ 12, 478, 38 },
		{ 12, 479, 38 },
		{ 12, 480, 38 },
		{ 12, 481, 38 },
		{ 12, 482, 38 },
		{ 12, 483, 38 },
		{ 12, 484, 38 },
		{ 12, 485, 38 },
		{ 12, 486, 38 },
		{ 12, 487, 38 },
		{ 12, 488, 38 },
		{ 12, 489, 38 },
		{ 12, 490, 38 },
		{ 12, 491, 38 },
		{ 12, 492, 38 },
		{ 12, 493, 38 },
		{ 12, 494, 38 },
		{ 12, 495, 38 },
		{ 12, 496, 38 },
		{ 12, 497, 38 },
		{ 12, 498, 38 },
		{ 12, 499, 38 },
		{ 12, 500, 38 },
		{ 12, 501, 38 },
		{ 12, 502, 38 },
		{ 12, 503, 38 },
		{ 12, 504, 38 },
		{ 12, 505, 38 },
		{ 12, 506, 38 },
		{ 12, 507, 38 },
		{ 12, 508, 38 },
		{ 12, 509, 38 },
		{ 12, 510, 38 },
		{ 12, 511, 38 },
		{ 12, 512, 38 },
		{ 13, 471, 38 },
		{ 13, 472, 38 },
		{ 13, 473, 38 },
		{ 13, 474, 38 },
		{ 13, 475, 38 },
		{ 13, 476, 38 },
		{ 13, 477, 38 },
		{ 13, 478, 38 },
		{ 13, 479, 38 },
		{ 13, 480, 38 },
		{ 13, 481, 38 },
		{ 13, 482, 38 },
		{ 13, 483, 38 },
		{ 13, 484, 38 },
		{ 13, 485, 38 },
		{ 13, 486, 38 },
		{ 13, 487, 38 },
		{ 13, 488, 38 },
		{ 13, 489, 38 },
		{ 13, 490, 38 },
		{ 13, 491, 38 },
		{ 13, 492, 38 },
		{ 13, 493, 38 },
		{ 13, 494, 38 },
		{ 13, 495, 38 },
		{ 13, 496, 38 },
		{ 13, 497, 38 },
		{ 13, 498, 38 },
		{ 13, 499, 38 },
		{ 13, 500, 38 },
		{ 13, 501, 38 },
		{ 13, 502, 38 },
		{ 13, 503, 38 },
		{ 13, 504, 38 },
		{ 13, 505, 38 },
		{ 13, 506, 38 },
		{ 13, 507, 38 },
		{ 13, 508, 38 },
		{ 13, 509, 38 },
		{ 13, 510, 38 },
		{ 13, 511, 38 },
		{ 13, 512, 38 },
		{ 14, 470, 38 },
		{ 14, 471, 38 },
		{ 14, 472, 38 },
		{ 14, 473, 38 },
		{ 14, 474, 38 },
		{ 14, 475, 38 },
		{ 14, 476, 38 },
		{ 14, 477, 38 },
		{ 14, 478, 38 },
		{ 14, 479, 38 },
		{ 14, 480, 38 },
		{ 14, 481, 38 },
		{ 14, 482, 38 },
		{ 14, 483, 38 },
		{ 14, 484, 38 },
		{ 14, 485, 38 },
		{ 14, 486, 38 },
		{ 14, 487, 38 },
		{ 14, 488, 38 },
		{ 14, 489, 38 },
		{ 14, 490, 38 },
		{ 14, 491, 38 },
		{ 14, 492, 38 },
		{ 14, 493, 38 },
		{ 14, 494, 38 },
		{ 14, 495, 38 },
		{ 14, 496, 38 },
		{ 14, 497, 38 },
		{ 14, 498, 38 },
		{ 14, 499, 38 },
		{ 14, 500, 38 },
		{ 14, 501, 38 },
		{ 14, 502, 38 },
		{ 14, 503, 38 },
		{ 14, 504, 38 },
		{ 14, 505, 38 },
		{ 14, 506, 38 },
		{ 14, 507, 38 },
		{ 14, 508, 38 },
		{ 14, 509, 38 },
		{ 14, 510, 38 },
		{ 14, 511, 38 },
		{ 14, 512, 38 },
		{ 15, 469, 38 },
		{ 15, 470, 38 },
		{ 15, 471, 38 },
		{ 15, 472, 38 },
		{ 15, 473, 38 },
		{ 15, 474, 38 },
		{ 15, 475, 38 },
		{ 15, 476, 38 },
		{ 15, 477, 38 },
		{ 15, 478, 38 },
		{ 15, 479, 38 },
		{ 15, 480, 38 },
		{ 15, 481, 38 },
		{ 15, 482, 38 },
		{ 15, 483, 38 },
		{ 15, 484, 38 },
		{ 15, 485, 38 },
		{ 15, 486, 38 },
		{ 15, 487, 38 },
		{ 15, 488, 38 },
		{ 15, 489, 38 },
		{ 15, 490, 38 },
		{ 15, 491, 38 },
		{ 15, 492, 38 },
		{ 15, 493, 38 },
		{ 15, 494, 38 },
		{ 15, 495, 38 },
		{ 15, 496, 38 },
		{ 15, 497, 38 },
		{ 15, 498, 38 },
		{ 15, 499, 38 },
		{ 15, 500, 38 },
		{ 15, 501, 38 },
		{ 15, 502, 38 },
		{ 15, 503, 38 },
		{ 15, 504, 38 },
		{ 15, 505, 38 },
		{ 15, 506, 38 },
		{ 15, 507, 38 },
		{ 15, 508, 38 },
		{ 15, 509, 38 },
		{ 15, 510, 38 },
		{ 15, 511, 38 },
		{ 15, 512, 38 },
		{ 16, 468, 38 },
		{ 16, 469, 38 },
		{ 16, 470, 38 },
		{ 16, 471, 38 },
		{ 16, 472, 38 },
		{ 16, 473, 38 },
		{ 16, 474, 38 },
		{ 16, 475, 38 },
		{ 16, 476, 38 },
		{ 16, 477, 38 },
		{ 16, 478, 38 },
		{ 16, 479, 38 },
		{ 16, 480, 38 },
		{ 16, 481, 38 },
		{ 16, 482, 38 },
		{ 16, 483, 38 },
		{ 16, 484, 38 },
		{ 16, 485, 38 },
		{ 16, 486, 38 },
		{ 16, 487, 38 },
		{ 16, 488, 38 },
		{ 16, 489, 38 },
		{ 16, 490, 38 },
		{ 16, 491, 38 },
		{ 16, 492, 38 },
		{ 16, 493, 38 },
		{ 16, 494, 38 },
		{ 16, 495, 38 },
		{ 16, 496, 38 },
		{ 16, 497, 38 },
		{ 16, 498, 38 },
		{ 16, 499, 38 },
		{ 16, 500, 38 },
		{ 16, 501, 38 },
		{ 16, 502, 38 },
		{ 16, 503, 38 },
		{ 16, 504, 38 },
		{ 16, 505, 38 },
		{ 16, 506, 38 },
		{ 16, 507, 38 },
		{ 16, 508, 38 },
		{ 16, 509, 38 },
		{ 16, 510, 38 },
		{ 16, 511, 38 },
		{ 16, 512, 38 },
		{ 17, 468, 38 },
		{ 17, 469, 38 },
		{ 17, 470, 38 },
		{ 17, 471, 38 },
		{ 17, 472, 38 },
		{ 17, 473, 38 },
		{ 17, 474, 38 },
		{ 17, 475, 38 },
		{ 17, 476, 38 },
		{ 17, 477, 38 },
		{ 17, 478, 38 },
		{ 17, 479, 38 },
		{ 17, 480, 38 },
		{ 17, 481, 38 },
		{ 17, 482, 38 },
		{ 17, 483, 38 },
		{ 17, 484, 38 },
		{ 17, 485, 38 },
		{ 17, 486, 38 },
		{ 17, 487, 38 },
		{ 17, 488, 38 },
		{ 17, 489, 38 },
		{ 17, 490, 38 },
		{ 17, 491, 38 },
		{ 17, 492, 38 },
		{ 17, 493, 38 },
		{ 17, 494, 38 },
		{ 17, 495, 38 },
		{ 17, 496, 38 },
		{ 17, 497, 38 },
		{ 17, 498, 38 },
		{ 17, 499, 38 },
		{ 17, 500, 38 },
		{ 17, 501, 38 },
		{ 17, 502, 38 },
		{ 17, 503, 38 },
		{ 17, 504, 38 },
		{ 17, 505, 38 },
		{ 17, 506, 38 },
		{ 17, 507, 38 },
		{ 17, 508, 38 },
		{ 17, 509, 38 },
		{ 17, 510, 38 },
		{ 17, 511, 38 },
		{ 17, 512, 38 },
		{ 18, 467, 38 },
		{ 18, 468, 38 },
		{ 18, 469, 38 },
		{ 18, 470, 38 },
		{ 18, 471, 38 },
		{ 18, 472, 38 },
		{ 18, 473, 38 },
		{ 18, 474, 38 },
		{ 18, 475, 38 },
		{ 18, 476, 38 },
		{ 18, 477, 38 },
		{ 18, 478, 38 },
		{ 18, 479, 38 },
		{ 18, 480, 38 },
		{ 18, 481, 38 },
		{ 18, 482, 38 },
		{ 18, 483, 38 },
		{ 18, 484, 38 },
		{ 18, 485, 38 },
		{ 18, 486, 38 },
		{ 18, 487, 38 },
		{ 18, 488, 38 },
		{ 18, 489, 38 },
		{ 18, 490, 38 },
		{ 18, 491, 38 },
		{ 18, 492, 38 },
		{ 18, 493, 38 },
		{ 18, 494, 38 },
		{ 18, 495, 38 },
		{ 18, 496, 38 },
		{ 18, 497, 38 },
		{ 18, 498, 38 },
		{ 18, 499, 38 },
		{ 18, 500, 38 },
		{ 18, 501, 38 },
		{ 18, 502, 38 },
		{ 18, 503, 38 },
		{ 18, 504, 38 },
		{ 18, 505, 38 },
		{ 18, 506, 38 },
		{ 18, 507, 38 },
		{ 18, 508, 38 },
		{ 18, 509, 38 },
		{ 18, 510, 38 },
		{ 18, 511, 38 },
		{ 18, 512, 38 },
		{ 19, 466, 38 },
		{ 19, 467, 38 },
		{ 19, 468, 38 },
		{ 19, 469, 38 },
		{ 19, 470, 38 },
		{ 19, 471, 38 },
		{ 19, 472, 38 },
		{ 19, 473, 38 },
		{ 19, 474, 38 },
		{ 19, 475, 38 },
		{ 19, 476, 38 },
		{ 19, 477, 38 },
		{ 19, 478, 38 },
		{ 19, 479, 38 },
		{ 19, 480, 38 },
		{ 19, 481, 38 },
		{ 19, 482, 38 },
		{ 19, 483, 38 },
		{ 19, 484, 38 },
		{ 19, 485, 38 },
		{ 19, 486, 38 },
		{ 19, 487, 38 },
		{ 19, 488, 38 },
		{ 19, 489, 38 },
		{ 19, 490, 38 },
		{ 19, 491, 38 },
		{ 19, 492, 38 },
		{ 19, 493, 38 },
		{ 19, 494, 38 },
		{ 19, 495, 38 },
		{ 19, 496, 38 },
		{ 19, 497, 38 },
		{ 19, 498, 38 },
		{ 19, 499, 38 },
		{ 19, 500, 38 },
		{ 19, 501, 38 },
		{ 19, 502, 38 },
		{ 19, 503, 38 },
		{ 19, 504, 38 },
		{ 19, 505, 38 },
		{ 19, 506, 38 },
		{ 19, 507, 38 },
		{ 19, 508, 38 },
		{ 19, 509, 38 },
		{ 19, 510, 38 },
		{ 19, 511, 38 },
		{ 19, 512, 38 },
		{ 20, 465, 38 },
		{ 20, 466, 38 },
		{ 20, 467, 38 },
		{ 20, 468, 38 },
		{ 20, 469, 38 },
		{ 20, 470, 38 },
		{ 20, 471, 38 },
		{ 20, 472, 38 },
		{ 20, 473, 38 },
		{ 20, 474, 38 },
		{ 20, 475, 38 },
		{ 20, 476, 38 },
		{ 20, 477, 38 },
		{ 20, 478, 38 },
		{ 20, 479, 38 },
		{ 20, 480, 38 },
		{ 20, 481, 38 },
		{ 20, 482, 38 },
		{ 20, 483, 38 },
		{ 20, 484, 38 },
		{ 20, 485, 38 },
		{ 20, 486, 38 },
		{ 20, 487, 38 },
		{ 20, 488, 38 },
		{ 20, 489, 38 },
		{ 20, 490, 38 },
		{ 20, 491, 38 },
		{ 20, 492, 38 },
		{ 20, 493, 38 },
		{ 20, 494, 38 },
		{ 20, 495, 38 },
		{ 20, 496, 38 },
		{ 20, 497, 38 },
		{ 20, 498, 38 },
		{ 20, 499, 38 },
		{ 20, 500, 38 },
		{ 20, 501, 38 },
		{ 20, 502, 38 },
		{ 20, 503, 38 },
		{ 20, 504, 38 },
		{ 20, 505, 38 },
		{ 20, 506, 38 },
		{ 20, 507, 38 },
		{ 20, 508, 38 },
		{ 20, 509, 38 },
		{ 20, 510, 38 },
		{ 20, 511, 38 },
		{ 20, 512, 38 },
		{ 21, 464, 38 },
		{ 21, 465, 38 },
		{ 21, 466, 38 },
		{ 21, 467, 38 },
		{ 21, 468, 38 },
		{ 21, 469, 38 },
		{ 21, 470, 38 },
		{ 21, 471, 38 },
		{ 21, 472, 38 },
		{ 21, 473, 38 },
		{ 21, 474, 38 },
		{ 21, 475, 38 },
		{ 21, 476, 38 },
		{ 21, 477, 38 },
		{ 21, 478, 38 },
		{ 21, 479, 38 },
		{ 21, 480, 38 },
		{ 21, 481, 38 },
		{ 21, 482, 38 },
		{ 21, 483, 38 },
		{ 21, 484, 38 },
		{ 21, 485, 38 },
		{ 21, 486, 38 },
		{ 21, 487, 38 },
		{ 21, 488, 38 },
		{ 21, 489, 38 },
		{ 21, 490, 38 },
		{ 21, 491, 38 },
		{ 21, 492, 38 },
		{ 21, 493, 38 },
		{ 21, 494, 38 },
		{ 21, 495, 38 },
		{ 21, 496, 38 },
		{ 21, 497, 38 },
		{ 21, 498, 38 },
		{ 21, 499, 38 },
		{ 21, 500, 38 },
		{ 21, 501, 38 },
		{ 21, 502, 38 },
		{ 21, 503, 38 },
		{ 21, 504, 38 },
		{ 21, 505, 38 },
		{ 21, 506, 38 },
		{ 21, 507, 38 },
		{ 21, 508, 38 },
		{ 21, 509, 38 },
		{ 21, 510, 38 },
		{ 21, 511, 38 },
		{ 21, 512, 38 },
		{ 22, 463, 38 },
		{ 22, 464, 38 },
		{ 22, 465, 38 },
		{ 22, 466, 38 },
		{ 22, 467, 38 },
		{ 22, 468, 38 },
		{ 22, 469, 38 },
		{ 22, 470, 38 },
		{ 22, 471, 38 },
		{ 22, 472, 38 },
		{ 22, 473, 38 },
		{ 22, 474, 38 },
		{ 22, 475, 38 },
		{ 22, 476, 38 },
		{ 22, 477, 38 },
		{ 22, 478, 38 },
		{ 22, 479, 38 },
		{ 22, 480, 38 },
		{ 22, 481, 38 },
		{ 22, 482, 38 },
		{ 22, 483, 38 },
		{ 22, 484, 38 },
		{ 22, 485, 38 },
		{ 22, 486, 38 },
		{ 22, 487, 38 },
		{ 22, 488, 38 },
		{ 22, 489, 38 },
		{ 22, 490, 38 },
		{ 22, 491, 38 },
		{ 22, 492, 38 },
		{ 22, 493, 38 },
		{ 22, 494, 38 },
		{ 22, 495, 38 },
		{ 22, 496, 38 },
		{ 22, 497, 38 },
		{ 22, 498, 38 },
		{ 22, 499, 38 },
		{ 22, 500, 38 },
		{ 22, 501, 38 },
		{ 22, 502, 38 },
		{ 22, 503, 38 },
		{ 22, 504, 38 },
		{ 22, 505, 38 },
		{ 22, 506, 38 },
		{ 22, 507, 38 },
		{ 22, 508, 38 },
		{ 22, 509, 38 },
		{ 22, 510, 38 },
		{ 22, 511, 38 },
		{ 22, 512, 38 },
		{ 23, 462, 38 },
		{ 23, 463, 38 },
		{ 23, 464, 38 },
		{ 23, 465, 38 },
		{ 23, 466, 38 },
		{ 23, 467, 38 },
		{ 23, 468, 38 },
		{ 23, 469, 38 },
		{ 23, 470, 38 },
		{ 23, 471, 38 },
		{ 23, 472, 38 },
		{ 23, 473, 38 },
		{ 23, 474, 38 },
		{ 23, 475, 38 },
		{ 23, 476, 38 },
		{ 23, 477, 38 },
		{ 23, 478, 38 },
		{ 23, 479, 38 },
		{ 23, 480, 38 },
		{ 23, 481, 38 },
		{ 23, 482, 38 },
		{ 23, 483, 38 },
		{ 23, 484, 38 },
		{ 23, 485, 38 },
		{ 23, 486, 38 },
		{ 23, 487, 38 },
		{ 23, 488, 38 },
		{ 23, 489, 38 },
		{ 23, 490, 38 },
		{ 23, 491, 38 },
		{ 23, 492, 38 },
		{ 23, 493, 38 },
		{ 23, 494, 38 },
		{ 23, 495, 38 },
		{ 23, 496, 38 },
		{ 23, 497, 38 },
		{ 23, 498, 38 },
		{ 23, 499, 38 },
		{ 23, 500, 38 },
		{ 23, 501, 38 },
		{ 23, 502, 38 },
		{ 23, 503, 38 },
		{ 23, 504, 38 },
		{ 23, 505, 38 },
		{ 23, 506, 38 },
		{ 23, 507, 38 },
		{ 23, 508, 38 },
		{ 23, 509, 38 },
		{ 23, 510, 38 },
		{ 23, 511, 38 },
		{ 23, 512, 38 },
		{ 24, 461, 38 },
		{ 24, 462, 38 },
		{ 24, 463, 38 },
		{ 24, 464, 38 },
		{ 24, 465, 38 },
		{ 24, 466, 38 },
		{ 24, 467, 38 },
		{ 24, 468, 38 },
		{ 24, 469, 38 },
		{ 24, 470, 38 },
		{ 24, 471, 38 },
		{ 24, 472, 38 },
		{ 24, 473, 38 },
		{ 24, 474, 38 },
		{ 24, 475, 38 },
		{ 24, 476, 38 },
		{ 24, 477, 38 },
		{ 24, 478, 38 },
		{ 24, 479, 38 },
		{ 24, 480, 38 },
		{ 24, 481, 38 },
		{ 24, 482, 38 },
		{ 24, 483, 38 },
		{ 24, 484, 38 },
		{ 24, 485, 38 },
		{ 24, 486, 38 },
		{ 24, 487, 38 },
		{ 24, 488, 38 },
		{ 24, 489, 38 },
		{ 24, 490, 38 },
		{ 24, 491, 38 },
		{ 24, 492, 38 },
		{ 24, 493, 38 },
		{ 24, 494, 38 },
		{ 24, 495, 38 },
		{ 24, 496, 38 },
		{ 24, 497, 38 },
		{ 24, 498, 38 },
		{ 24, 499, 38 },
		{ 24, 500, 38 },
		{ 24, 501, 38 },
		{ 24, 502, 38 },
		{ 24, 503, 38 },
		{ 24, 504, 38 },
		{ 24, 505, 38 },
		{ 24, 506, 38 },
		{ 24, 507, 38 },
		{ 24, 508, 38 },
		{ 24, 509, 38 },
		{ 24, 510, 38 },
		{ 24, 511, 38 },
		{ 24, 512, 38 },
		{ 25, 461, 38 },
		{ 25, 462, 38 },
		{ 25, 463, 38 },
		{ 25, 464, 38 },
		{ 25, 465, 38 },
		{ 25, 466, 38 },
		{ 25, 467, 38 },
		{ 25, 468, 38 },
		{ 25, 469, 38 },
		{ 25, 470, 38 },
		{ 25, 471, 38 },
		{ 25, 472, 38 },
		{ 25, 473, 38 },
		{ 25, 474, 38 },
		{ 25, 475, 38 },
		{ 25, 476, 38 },
		{ 25, 477, 38 },
		{ 25, 478, 38 },
		{ 25, 479, 38 },
		{ 25, 480, 38 },
		{ 25, 481, 38 },
		{ 25, 482, 38 },
		{ 25, 483, 38 },
		{ 25, 484, 38 },
		{ 25, 485, 38 },
		{ 25, 486, 38 },
		{ 25, 487, 38 },
		{ 25, 488, 38 },
		{ 25, 489, 38 },
		{ 25, 490, 38 },
		{ 25, 491, 38 },
		{ 25, 492, 38 },
		{ 25, 493, 38 },
		{ 25, 494, 38 },
		{ 25, 495, 38 },
		{ 25, 496, 38 },
		{ 25, 497, 38 },
		{ 25, 498, 38 },
		{ 25, 499, 38 },
		{ 25, 500, 38 },
		{ 25, 501, 38 },
		{ 25, 502, 38 },
		{ 25, 503, 38 },
		{ 25, 504, 38 },
		{ 25, 505, 38 },
		{ 25, 506, 38 },
		{ 25, 507, 38 },
		{ 25, 508, 38 },
		{ 25, 509, 38 },
		{ 25, 510, 38 },
		{ 25, 511, 38 },
		{ 25, 512, 38 },
		{ 26, 460, 38 },
		{ 26, 461, 38 },
		{ 26, 462, 38 },
		{ 26, 463, 38 },
		{ 26, 464, 38 },
		{ 26, 465, 38 },
		{ 26, 466, 38 },
		{ 26, 467, 38 },
		{ 26, 468, 38 },
		{ 26, 469, 38 },
		{ 26, 470, 38 },
		{ 26, 471, 38 },
		{ 26, 472, 38 },
		{ 26, 473, 38 },
		{ 26, 474, 38 },
		{ 26, 475, 38 },
		{ 26, 476, 38 },
		{ 26, 477, 38 },
		{ 26, 478, 38 },
		{ 26, 479, 38 },
		{ 26, 480, 38 },
		{ 26, 481, 38 },
		{ 26, 482, 38 },
		{ 26, 483, 38 },
		{ 26, 484, 38 },
		{ 26, 485, 38 },
		{ 26, 486, 38 },
		{ 26, 487, 38 },
		{ 26, 488, 38 },
		{ 26, 489, 38 },
		{ 26, 490, 38 },
		{ 26, 491, 38 },
		{ 26, 492, 38 },
		{ 26, 493, 38 },
		{ 26, 494, 38 },
		{ 26, 495, 38 },
		{ 26, 496, 38 },
		{ 26, 497, 38 },
		{ 26, 498, 38 },
		{ 26, 499, 38 },
		{ 26, 500, 38 },
		{ 26, 501, 38 },
		{ 26, 502, 38 },
		{ 26, 503, 38 },
		{ 26, 504, 38 },
		{ 26, 505, 38 },
		{ 26, 506, 38 },
		{ 26, 507, 38 },
		{ 26, 508, 38 },
		{ 26, 509, 38 },
		{ 26, 510, 38 },
		{ 26, 511, 38 },
		{ 26, 512, 38 },
		{ 27, 459, 38 },
		{ 27, 460, 38 },
		{ 27, 461, 38 },
		{ 27, 462, 38 },
		{ 27, 463, 38 },
		{ 27, 464, 38 },
		{ 27, 465, 38 },
		{ 27, 466, 38 },
		{ 27, 467, 38 },
		{ 27, 468, 38 },
		{ 27, 469, 38 },
		{ 27, 470, 38 },
		{ 27, 471, 38 },
		{ 27, 472, 38 },
		{ 27, 473, 38 },
		{ 27, 474, 38 },
		{ 27, 475, 38 },
		{ 27, 476, 38 },
		{ 27, 477, 38 },
		{ 27, 478, 38 },
		{ 27, 479, 38 },
		{ 27, 480, 38 },
		{ 27, 481, 38 },
		{ 27, 482, 38 },
		{ 27, 483, 38 },
		{ 27, 484, 38 },
		{ 27, 485, 38 },
		{ 27, 486, 38 },
		{ 27, 487, 38 },
		{ 27, 488, 38 },
		{ 27, 489, 38 },
		{ 27, 490, 38 },
		{ 27, 491, 38 },
		{ 27, 492, 38 },
		{ 27, 493, 38 },
		{ 27, 494, 38 },
		{ 27, 495, 38 },
		{ 27, 496, 38 },
		{ 27, 497, 38 },
		{ 27, 498, 38 },
		{ 27, 499, 38 },
		{ 27, 500, 38 },
		{ 27, 501, 38 },
		{ 27, 502, 38 },
		{ 27, 503, 38 },
		{ 27, 504, 38 },
		{ 27, 505, 38 },
		{ 27, 506, 38 },
		{ 27, 507, 38 },
		{ 27, 508, 38 },
		{ 27, 509, 38 },
		{ 27, 510, 38 },
		{ 27, 511, 38 },
		{ 27, 512, 38 },
		{ 28, 458, 38 },
		{ 28, 459, 38 },
		{ 28, 460, 38 },
		{ 28, 461, 38 },
		{ 28, 462, 38 },
		{ 28, 463, 38 },
		{ 28, 464, 38 },
		{ 28, 465, 38 },
		{ 28, 466, 38 },
		{ 28, 467, 38 },
		{ 28, 468, 38 },
		{ 28, 469, 38 },
		{ 28, 470, 38 },
		{ 28, 471, 38 },
		{ 28, 472, 38 },
		{ 28, 473, 38 },
		{ 28, 474, 38 },
		{ 28, 475, 38 },
		{ 28, 476, 38 },
		{ 28, 477, 38 },
		{ 28, 478, 38 },
		{ 28, 479, 38 },
		{ 28, 480, 38 },
		{ 28, 481, 38 },
		{ 28, 482, 38 },
		{ 28, 483, 38 },
		{ 28, 484, 38 },
		{ 28, 485, 38 },
		{ 28, 486, 38 },
		{ 28, 487, 38 },
		{ 28, 488, 38 },
		{ 28, 489, 38 },
		{ 28, 490, 38 },
		{ 28, 491, 38 },
		{ 28, 492, 38 },
		{ 28, 493, 38 },
		{ 28, 494, 38 },
		{ 28, 495, 38 },
		{ 28, 496, 38 },
		{ 28, 497, 38 },
		{ 28, 498, 38 },
		{ 28, 499, 38 },
		{ 28, 500, 38 },
		{ 28, 501, 38 },
		{ 28, 502, 38 },
		{ 28, 503, 38 },
		{ 28, 504, 38 },
		{ 28, 505, 38 },
		{ 28, 506, 38 },
		{ 28, 507, 38 },
		{ 28, 508, 38 },
		{ 28, 509, 38 },
		{ 28, 510, 38 },
		{ 28, 511, 38 },
		{ 28, 512, 38 },
		{ 29, 457, 38 },
		{ 29, 458, 38 },
		{ 29, 459, 38 },
		{ 29, 460, 38 },
		{ 29, 461, 38 },
		{ 29, 462, 38 },
		{ 29, 463, 38 },
		{ 29, 464, 38 },
		{ 29, 465, 38 },
		{ 29, 466, 38 },
		{ 29, 467, 38 },
		{ 29, 468, 38 },
		{ 29, 469, 38 },
		{ 29, 470, 38 },
		{ 29, 471, 38 },
		{ 29, 472, 38 },
		{ 29, 473, 38 },
		{ 29, 474, 38 },
		{ 29, 475, 38 },
		{ 29, 476, 38 },
		{ 29, 477, 38 },
		{ 29, 478, 38 },
		{ 29, 479, 38 },
		{ 29, 480, 38 },
		{ 29, 481, 38 },
		{ 29, 482, 38 },
		{ 29, 483, 38 },
		{ 29, 484, 38 },
		{ 29, 485, 38 },
		{ 29, 486, 38 },
		{ 29, 487, 38 },
		{ 29, 488, 38 },
		{ 29, 489, 38 },
		{ 29, 490, 38 },
		{ 29, 491, 38 },
		{ 29, 492, 38 },
		{ 29, 493, 38 },
		{ 29, 494, 38 },
		{ 29, 495, 38 },
		{ 29, 496, 38 },
		{ 29, 497, 38 },
		{ 29, 498, 38 },
		{ 29, 499, 38 },
		{ 29, 500, 38 },
		{ 29, 501, 38 },
		{ 29, 502, 38 },
		{ 29, 503, 38 },
		{ 29, 504, 38 },
		{ 29, 505, 38 },
		{ 29, 506, 38 },
		{ 29, 507, 38 },
		{ 29, 508, 38 },
		{ 29, 509, 38 },
		{ 29, 510, 38 },
		{ 29, 511, 38 },
		{ 29, 512, 38 },
		{ 30, 456, 38 },
		{ 30, 457, 38 },
		{ 30, 458, 38 },
		{ 30, 459, 38 },
		{ 30, 460, 38 },
		{ 30, 461, 38 },
		{ 30, 462, 38 },
		{ 30, 463, 38 },
		{ 30, 464, 38 },
		{ 30, 465, 38 },
		{ 30, 466, 38 },
		{ 30, 467, 38 },
		{ 30, 468, 38 },
		{ 30, 469, 38 },
		{ 30, 470, 38 },
		{ 30, 471, 38 },
		{ 30, 472, 38 },
		{ 30, 473, 38 },
		{ 30, 474, 38 },
		{ 30, 475, 38 },
		{ 30, 476, 38 },
		{ 30, 477, 38 },
		{ 30, 478, 38 },
		{ 30, 479, 38 },
		{ 30, 480, 38 },
		{ 30, 481, 38 },
		{ 30, 482, 38 },
		{ 30, 483, 38 },
		{ 30, 484, 38 },
		{ 30, 485, 38 },
		{ 30, 486, 38 },
		{ 30, 487, 38 },
		{ 30, 488, 38 },
		{ 30, 489, 38 },
		{ 30, 490, 38 },
		{ 30, 491, 38 },
		{ 30, 492, 38 },
		{ 30, 493, 38 },
		{ 30, 494, 38 },
		{ 30, 495, 38 },
		{ 30, 496, 38 },
		{ 30, 497, 38 },
		{ 30, 498, 38 },
		{ 30, 499, 38 },
		{ 30, 500, 38 },
		{ 30, 501, 38 },
		{ 30, 502, 38 },
		{ 30, 503, 38 },
		{ 30, 504, 38 },
		{ 30, 505, 38 },
		{ 30, 506, 38 },
		{ 30, 507, 38 },
		{ 30, 508, 38 },
		{ 30, 509, 38 },
		{ 30, 510, 38 },
		{ 30, 511, 38 },
		{ 30, 512, 38 },
		{ 31, 455, 38 },
		{ 31, 456, 38 },
		{ 31, 457, 38 },
		{ 31, 458, 38 },
		{ 31, 459, 38 },
		{ 31, 460, 38 },
		{ 31, 461, 38 },
		{ 31, 462, 38 },
		{ 31, 463, 38 },
		{ 31, 464, 38 },
		{ 31, 465, 38 },
		{ 31, 466, 38 },
		{ 31, 467, 38 },
		{ 31, 468, 38 },
		{ 31, 469, 38 },
		{ 31, 470, 38 },
		{ 31, 471, 38 },
		{ 31, 472, 38 },
		{ 31, 473, 38 },
		{ 31, 474, 38 },
		{ 31, 475, 38 },
		{ 31, 476, 38 },
		{ 31, 477, 38 },
		{ 31, 478, 38 },
		{ 31, 479, 38 },
		{ 31, 480, 38 },
		{ 31, 481, 38 },
		{ 31, 482, 38 },
		{ 31, 483, 38 },
		{ 31, 484, 38 },
		{ 31, 485, 38 },
		{ 31, 486, 38 },
		{ 31, 487, 38 },
		{ 31, 488, 38 },
		{ 31, 489, 38 },
		{ 31, 490, 38 },
		{ 31, 491, 38 },
		{ 31, 492, 38 },
		{ 31, 493, 38 },
		{ 31, 494, 38 },
		{ 31, 495, 38 },
		{ 31, 496, 38 },
		{ 31, 497, 38 },
		{ 31, 498, 38 },
		{ 31, 499, 38 },
		{ 31, 500, 38 },
		{ 31, 501, 38 },
		{ 31, 502, 38 },
		{ 31, 503, 38 },
		{ 31, 504, 38 },
		{ 31, 505, 38 },
		{ 31, 506, 38 },
		{ 31, 507, 38 },
		{ 31, 508, 38 },
		{ 31, 509, 38 },
		{ 31, 510, 38 },
		{ 31, 511, 38 },
		{ 31, 512, 38 },
		{ 32, 454, 38 },
		{ 32, 455, 38 },
		{ 32, 456, 38 },
		{ 32, 457, 38 },
		{ 32, 458, 38 },
		{ 32, 459, 38 },
		{ 32, 460, 38 },
		{ 32, 461, 38 },
		{ 32, 462, 38 },
		{ 32, 463, 38 },
		{ 32, 464, 38 },
		{ 32, 465, 38 },
		{ 32, 466, 38 },
		{ 32, 467, 38 },
		{ 32, 468, 38 },
		{ 32, 469, 38 },
		{ 32, 470, 38 },
		{ 32, 471, 38 },
		{ 32, 472, 38 },
		{ 32, 473, 38 },
		{ 32, 474, 38 },
		{ 32, 475, 38 },
		{ 32, 476, 38 },
		{ 32, 477, 38 },
		{ 32, 478, 38 },
		{ 32, 479, 38 },
		{ 32, 480, 38 },
		{ 32, 481, 38 },
		{ 32, 482, 38 },
		{ 32, 483, 38 },
		{ 32, 484, 38 },
		{ 32, 485, 38 },
		{ 32, 486, 38 },
		{ 32, 487, 38 },
		{ 32, 488, 38 },
		{ 32, 489, 38 },
		{ 32, 490, 38 },
		{ 32, 491, 38 },
		{ 32, 492, 38 },
		{ 32, 493, 38 },
		{ 32, 494, 38 },
		{ 32, 495, 38 },
		{ 32, 496, 38 },
		{ 32, 497, 38 },
		{ 32, 498, 38 },
		{ 32, 499, 38 },
		{ 32, 500, 38 },
		{ 32, 501, 38 },
		{ 32, 502, 38 },
		{ 32, 503, 38 },
		{ 32, 504, 38 },
		{ 32, 505, 38 },
		{ 32, 506, 38 },
		{ 32, 507, 38 },
		{ 32, 508, 38 },
		{ 32, 509, 38 },
		{ 32, 510, 38 },
		{ 32, 511, 38 },
		{ 32, 512, 38 },
		{ 33, 454, 38 },
		{ 33, 455, 38 },
		{ 33, 456, 38 },
		{ 33, 457, 38 },
		{ 33, 458, 38 },
		{ 33, 459, 38 },
		{ 33, 460, 38 },
		{ 33, 461, 38 },
		{ 33, 462, 38 },
		{ 33, 463, 38 },
		{ 33, 464, 38 },
		{ 33, 465, 38 },
		{ 33, 466, 38 },
		{ 33, 467, 38 },
		{ 33, 468, 38 },
		{ 33, 469, 38 },
		{ 33, 470, 38 },
		{ 33, 471, 38 },
		{ 33, 472, 38 },
		{ 33, 473, 38 },
		{ 33, 474, 38 },
		{ 33, 475, 38 },
		{ 33, 476, 38 },
		{ 33, 477, 38 },
		{ 33, 478, 38 },
		{ 33, 479, 38 },
		{ 33, 480, 38 },
		{ 33, 481, 38 },
		{ 33, 482, 38 },
		{ 33, 483, 38 },
		{ 33, 484, 38 },
		{ 33, 485, 38 },
		{ 33, 486, 38 },
		{ 33, 487, 38 },
		{ 33, 488, 38 },
		{ 33, 489, 38 },
		{ 33, 490, 38 },
		{ 33, 491, 38 },
		{ 33, 492, 38 },
		{ 33, 493, 38 },
		{ 33, 494, 38 },
		{ 33, 495, 38 },
		{ 33, 496, 38 },
		{ 33, 497, 38 },
		{ 33, 498, 38 },
		{ 33, 499, 38 },
		{ 33, 500, 38 },
		{ 33, 501, 38 },
		{ 33, 502, 38 },
		{ 33, 503, 38 },
		{ 33, 504, 38 },
		{ 33, 505, 38 },
		{ 33, 506, 38 },
		{ 33, 507, 38 },
		{ 33, 508, 38 },
		{ 33, 509, 38 },
		{ 33, 510, 38 },
		{ 33, 511, 38 },
		{ 33, 512, 38 },
		{ 34, 453, 38 },
		{ 34, 454, 38 },
		{ 34, 455, 38 },
		{ 34, 456, 38 },
		{ 34, 457, 38 },
		{ 34, 458, 38 },
		{ 34, 459, 38 },
		{ 34, 460, 38 },
		{ 34, 461, 38 },
		{ 34, 462, 38 },
		{ 34, 463, 38 },
		{ 34, 464, 38 },
		{ 34, 465, 38 },
		{ 34, 466, 38 },
		{ 34, 467, 38 },
		{ 34, 468, 38 },
		{ 34, 469, 38 },
		{ 34, 470, 38 },
		{ 34, 471, 38 },
		{ 34, 472, 38 },
		{ 34, 473, 38 },
		{ 34, 474, 38 },
		{ 34, 475, 38 },
		{ 34, 476, 38 },
		{ 34, 477, 38 },
		{ 34, 478, 38 },
		{ 34, 479, 38 },
		{ 34, 480, 38 },
		{ 34, 481, 38 },
		{ 34, 482, 38 },
		{ 34, 483, 38 },
		{ 34, 484, 38 },
		{ 34, 485, 38 },
		{ 34, 486, 38 },
		{ 34, 487, 38 },
		{ 34, 488, 38 },
		{ 34, 489, 38 },
		{ 34, 490, 38 },
		{ 34, 491, 38 },
		{ 34, 492, 38 },
		{ 34, 493, 38 },
		{ 34, 494, 38 },
		{ 34, 495, 38 },
		{ 34, 496, 38 },
		{ 34, 497, 38 },
		{ 34, 498, 38 },
		{ 34, 499, 38 },
		{ 34, 500, 38 },
		{ 34, 501, 38 },
		{ 34, 502, 38 },
		{ 34, 503, 38 },
		{ 34, 504, 38 },
		{ 34, 505, 38 },
		{ 34, 506, 38 },
		{ 34, 507, 38 },
		{ 34, 508, 38 },
		{ 34, 509, 38 },
		{ 34, 510, 38 },
		{ 34, 511, 38 },
		{ 35, 452, 38 },
		{ 35, 453, 38 },
		{ 35, 454, 38 },
		{ 35, 455, 38 },
		{ 35, 456, 38 },
		{ 35, 457, 38 },
		{ 35, 458, 38 },
		{ 35, 459, 38 },
		{ 35, 460, 38 },
		{ 35, 461, 38 },
		{ 35, 462, 38 },
		{ 35, 463, 38 },
		{ 35, 464, 38 },
		{ 35, 465, 38 },
		{ 35, 466, 38 },
		{ 35, 467, 38 },
		{ 35, 468, 38 },
		{ 35, 469, 38 },
		{ 35, 470, 38 },
		{ 35, 471, 38 },
		{ 35, 472, 38 },
		{ 35, 473, 38 },
		{ 35, 474, 38 },
		{ 35, 475, 38 },
		{ 35, 476, 38 },
		{ 35, 477, 38 },
		{ 35, 478, 38 },
		{ 35, 479, 38 },
		{ 35, 480, 38 },
		{ 35, 481, 38 },
		{ 35, 482, 38 },
		{ 35, 483, 38 },
		{ 35, 484, 38 },
		{ 35, 485, 38 },
		{ 35, 486, 38 },
		{ 35, 487, 38 },
		{ 35, 488, 38 },
		{ 35, 489, 38 },
		{ 35, 490, 38 },
		{ 35, 491, 38 },
		{ 35, 492, 38 },
		{ 35, 493, 38 },
		{ 35, 494, 38 },
		{ 35, 495, 38 },
		{ 35, 496, 38 },
		{ 35, 497, 38 },
		{ 35, 498, 38 },
		{ 35, 499, 38 },
		{ 35, 500, 38 },
		{ 35, 501, 38 },
		{ 35, 502, 38 },
		{ 35, 503, 38 },
		{ 35, 504, 38 },
		{ 35, 505, 38 },
		{ 35, 506, 38 },
		{ 35, 507, 38 },
		{ 35, 508, 38 },
		{ 35, 509, 38 },
		{ 35, 510, 38 },
		{ 36, 451, 38 },
		{ 36, 452, 38 },
		{ 36, 453, 38 },
		{ 36, 454, 38 },
		{ 36, 455, 38 },
		{ 36, 456, 38 },
		{ 36, 457, 38 },
		{ 36, 458, 38 },
		{ 36, 459, 38 },
		{ 36, 460, 38 },
		{ 36, 461, 38 },
		{ 36, 462, 38 },
		{ 36, 463, 38 },
		{ 36, 464, 38 },
		{ 36, 465, 38 },
		{ 36, 466, 38 },
		{ 36, 467, 38 },
		{ 36, 468, 38 },
		{ 36, 469, 38 },
		{ 36, 470, 38 },
		{ 36, 471, 38 },
		{ 36, 472, 38 },
		{ 36, 473, 38 },
		{ 36, 474, 38 },
		{ 36, 475, 38 },
		{ 36, 476, 38 },
		{ 36, 477, 38 },
		{ 36, 478, 38 },
		{ 36, 479, 38 },
		{ 36, 480, 38 },
		{ 36, 481, 38 },
		{ 36, 482, 38 },
		{ 36, 483, 38 },
		{ 36, 484, 38 },
		{ 36, 485, 38 },
		{ 36, 486, 38 },
		{ 36, 487, 38 },
		{ 36, 488, 38 },
		{ 36, 489, 38 },
		{ 36, 490, 38 },
		{ 36, 491, 38 },
		{ 36, 492, 38 },
		{ 36, 493, 38 },
		{ 36, 494, 38 },
		{ 36, 495, 38 },
		{ 36, 496, 38 },
		{ 36, 497, 38 },
		{ 36, 498, 38 },
		{ 36, 499, 38 },
		{ 36, 500, 38 },
		{ 36, 501, 38 },
		{ 36, 502, 38 },
		{ 36, 503, 38 },
		{ 36, 504, 38 },
		{ 36, 505, 38 },
		{ 36, 506, 38 },
		{ 36, 507, 38 },
		{ 36, 508, 38 },
		{ 36, 509, 38 },
		{ 37, 450, 38 },
		{ 37, 451, 38 },
		{ 37, 452, 38 },
		{ 37, 453, 38 },
		{ 37, 454, 38 },
		{ 37, 455, 38 },
		{ 37, 456, 38 },
		{ 37, 457, 38 },
		{ 37, 458, 38 },
		{ 37, 459, 38 },
		{ 37, 460, 38 },
		{ 37, 461, 38 },
		{ 37, 462, 38 },
		{ 37, 463, 38 },
		{ 37, 464, 38 },
		{ 37, 465, 38 },
		{ 37, 466, 38 },
		{ 37, 467, 38 },
		{ 37, 468, 38 },
		{ 37, 469, 38 },
		{ 37, 470, 38 },
		{ 37, 471, 38 },
		{ 37, 472, 38 },
		{ 37, 473, 38 },
		{ 37, 474, 38 },
		{ 37, 475, 38 },
		{ 37, 476, 38 },
		{ 37, 477, 38 },
		{ 37, 478, 38 },
		{ 37, 479, 38 },
		{ 37, 480, 38 },
		{ 37, 481, 38 },
		{ 37, 482, 38 },
		{ 37, 483, 38 },
		{ 37, 484, 38 },
		{ 37, 485, 38 },
		{ 37, 486, 38 },
		{ 37, 487, 38 },
		{ 37, 488, 38 },
		{ 37, 489, 38 },
		{ 37, 490, 38 },
		{ 37, 491, 38 },
		{ 37, 492, 38 },
		{ 37, 493, 38 },
		{ 37, 494, 38 },
		{ 37, 495, 38 },
		{ 37, 496, 38 },
		{ 37, 497, 38 },
		{ 37, 498, 38 },
		{ 37, 499, 38 },
		{ 37, 500, 38 },
		{ 37, 501, 38 },
		{ 37, 502, 38 },
		{ 37, 503, 38 },
		{ 37, 504, 38 },
		{ 37, 505, 38 },
		{ 37, 506, 38 },
		{ 37, 507, 38 },
		{ 37, 508, 38 },
		{ 38, 449, 38 },
		{ 38, 450, 38 },
		{ 38, 451, 38 },
		{ 38, 452, 38 },
		{ 38, 453, 38 },
		{ 38, 454, 38 },
		{ 38, 455, 38 },
		{ 38, 456, 38 },
		{ 38, 457, 38 },
		{ 38, 458, 38 },
		{ 38, 459, 38 },
		{ 38, 460, 38 },
		{ 38, 461, 38 },
		{ 38, 462, 38 },
		{ 38, 463, 38 },
		{ 38, 464, 38 },
		{ 38, 465, 38 },
		{ 38, 466, 38 },
		{ 38, 467, 38 },
		{ 38, 468, 38 },
		{ 38, 469, 38 },
		{ 38, 470, 38 },
		{ 38, 471, 38 },
		{ 38, 472, 38 },
		{ 38, 473, 38 },
		{ 38, 474, 38 },
		{ 38, 475, 38 },
		{ 38, 476, 38 },
		{ 38, 477, 38 },
		{ 38, 478, 38 },
		{ 38, 479, 38 },
		{ 38, 480, 38 },
		{ 38, 481, 38 },
		{ 38, 482, 38 },
		{ 38, 483, 38 },
		{ 38, 484, 38 },
		{ 38, 485, 38 },
		{ 38, 486, 38 },
		{ 38, 487, 38 },
		{ 38, 488, 38 },
		{ 38, 489, 38 },
		{ 38, 490, 38 },
		{ 38, 491, 38 },
		{ 38, 492, 38 },
		{ 38, 493, 38 },
		{ 38, 494, 38 },
		{ 38, 495, 38 },
		{ 38, 496, 38 },
		{ 38, 497, 38 },
		{ 38, 498, 38 },
		{ 38, 499, 38 },
		{ 38, 500, 38 },
		{ 38, 501, 38 },
		{ 38, 502, 38 },
		{ 38, 503, 38 },
		{ 38, 504, 38 },
		{ 38, 505, 38 },
		{ 38, 506, 38 },
		{ 38, 507, 38 },
		{ 39, 448, 38 },
		{ 39, 449, 38 },
		{ 39, 450, 38 },
		{ 39, 451, 38 },
		{ 39, 452, 38 },
		{ 39, 453, 38 },
		{ 39, 454, 38 },
		{ 39, 455, 38 },
		{ 39, 456, 38 },
		{ 39, 457, 38 },
		{ 39, 458, 38 },
		{ 39, 459, 38 },
		{ 39, 460, 38 },
		{ 39, 461, 38 },
		{ 39, 462, 38 },
		{ 39, 463, 38 },
		{ 39, 464, 38 },
		{ 39, 465, 38 },
		{ 39, 466, 38 },
		{ 39, 467, 38 },
		{ 39, 468, 38 },
		{ 39, 469, 38 },
		{ 39, 470, 38 },
		{ 39, 471, 38 },
		{ 39, 472, 38 },
		{ 39, 473, 38 },
		{ 39, 474, 38 },
		{ 39, 475, 38 },
		{ 39, 476, 38 },
		{ 39, 477, 38 },
		{ 39, 478, 38 },
		{ 39, 479, 38 },
		{ 39, 480, 38 },
		{ 39, 481, 38 },
		{ 39, 482, 38 },
		{ 39, 483, 38 },
		{ 39, 484, 38 },
		{ 39, 485, 38 },
		{ 39, 486, 38 },
		{ 39, 487, 38 },
		{ 39, 488, 38 },
		{ 39, 489, 38 },
		{ 39, 490, 38 },
		{ 39, 491, 38 },
		{ 39, 492, 38 },
		{ 39, 493, 38 },
		{ 39, 494, 38 },
		{ 39, 495, 38 },
		{ 39, 496, 38 },
		{ 39, 497, 38 },
		{ 39, 498, 38 },
		{ 39, 499, 38 },
		{ 39, 500, 38 },
		{ 39, 501, 38 },
		{ 39, 502, 38 },
		{ 39, 503, 38 },
		{ 39, 504, 38 },
		{ 39, 505, 38 },
		{ 39, 506, 38 },
		{ 40, 447, 38 },
		{ 40, 448, 38 },
		{ 40, 449, 38 },
		{ 40, 450, 38 },
		{ 40, 451, 38 },
		{ 40, 452, 38 },
		{ 40, 453, 38 },
		{ 40, 454, 38 },
		{ 40, 455, 38 },
		{ 40, 456, 38 },
		{ 40, 457, 38 },
		{ 40, 458, 38 },
		{ 40, 459, 38 },
		{ 40, 460, 38 },
		{ 40, 461, 38 },
		{ 40, 462, 38 },
		{ 40, 463, 38 },
		{ 40, 464, 38 },
		{ 40, 465, 38 },
		{ 40, 466, 38 },
		{ 40, 467, 38 },
		{ 40, 468, 38 },
		{ 40, 469, 38 },
		{ 40, 470, 38 },
		{ 40, 471, 38 },
		{ 40, 472, 38 },
		{ 40, 473, 38 },
		{ 40, 474, 38 },
		{ 40, 475, 38 },
		{ 40, 476, 38 },
		{ 40, 477, 38 },
		{ 40, 478, 38 },
		{ 40, 479, 38 },
		{ 40, 480, 38 },
		{ 40, 481, 38 },
		{ 40, 482, 38 },
		{ 40, 483, 38 },
		{ 40, 484, 38 },
		{ 40, 485, 38 },
		{ 40, 486, 38 },
		{ 40, 487, 38 },
		{ 40, 488, 38 },
		{ 40, 489, 38 },
		{ 40, 490, 38 },
		{ 40, 491, 38 },
		{ 40, 492, 38 },
		{ 40, 493, 38 },
		{ 40, 494, 38 },
		{ 40, 495, 38 },
		{ 40, 496, 38 },
		{ 40, 497, 38 },
		{ 40, 498, 38 },
		{ 40, 499, 38 },
		{ 40, 500, 38 },
		{ 40, 501, 38 },
		{ 40, 502, 38 },
		{ 40, 503, 38 },
		{ 40, 504, 38 },
		{ 41, 447, 38 },
		{ 41, 448, 38 },
		{ 41, 449, 38 },
		{ 41, 450, 38 },
		{ 41, 451, 38 },
		{ 41, 452, 38 },
		{ 41, 453, 38 },
		{ 41, 454, 38 },
		{ 41, 455, 38 },
		{ 41, 456, 38 },
		{ 41, 457, 38 },
		{ 41, 458, 38 },
		{ 41, 459, 38 },
		{ 41, 460, 38 },
		{ 41, 461, 38 },
		{ 41, 462, 38 },
		{ 41, 463, 38 },
		{ 41, 464, 38 },
		{ 41, 465, 38 },
		{ 41, 466, 38 },
		{ 41, 467, 38 },
		{ 41, 468, 38 },
		{ 41, 469, 38 },
		{ 41, 470, 38 },
		{ 41, 471, 38 },
		{ 41, 472, 38 },
		{ 41, 473, 38 },
		{ 41, 474, 38 },
		{ 41, 475, 38 },
		{ 41, 476, 38 },
		{ 41, 477, 38 },
		{ 41, 478, 38 },
		{ 41, 479, 38 },
		{ 41, 480, 38 },
		{ 41, 481, 38 },
		{ 41, 482, 38 },
		{ 41, 483, 38 },
		{ 41, 484, 38 },
		{ 41, 485, 38 },
		{ 41, 486, 38 },
		{ 41, 487, 38 },
		{ 41, 488, 38 },
		{ 41, 489, 38 },
		{ 41, 490, 38 },
		{ 41, 491, 38 },
		{ 41, 492, 38 },
		{ 41, 493, 38 },
		{ 41, 494, 38 },
		{ 41, 495, 38 },
		{ 41, 496, 38 },
		{ 41, 497, 38 },
		{ 41, 498, 38 },
		{ 41, 499, 38 },
		{ 41, 500, 38 },
		{ 41, 501, 38 },
		{ 41, 502, 38 },
		{ 41, 503, 38 },
		{ 42, 446, 38 },
		{ 42, 447, 38 },
		{ 42, 448, 38 },
		{ 42, 449, 38 },
		{ 42, 450, 38 },
		{ 42, 451, 38 },
		{ 42, 452, 38 },
		{ 42, 453, 38 },
		{ 42, 454, 38 },
		{ 42, 455, 38 },
		{ 42, 456, 38 },
		{ 42, 457, 38 },
		{ 42, 458, 38 },
		{ 42, 459, 38 },
		{ 42, 460, 38 },
		{ 42, 461, 38 },
		{ 42, 462, 38 },
		{ 42, 463, 38 },
		{ 42, 464, 38 },
		{ 42, 465, 38 },
		{ 42, 466, 38 },
		{ 42, 467, 38 },
		{ 42, 468, 38 },
		{ 42, 469, 38 },
		{ 42, 470, 38 },
		{ 42, 471, 38 },
		{ 42, 472, 38 },
		{ 42, 473, 38 },
		{ 42, 474, 38 },
		{ 42, 475, 38 },
		{ 42, 476, 38 },
		{ 42, 477, 38 },
		{ 42, 478, 38 },
		{ 42, 479, 38 },
		{ 42, 480, 38 },
		{ 42, 481, 38 },
		{ 42, 482, 38 },
		{ 42, 483, 38 },
		{ 42, 484, 38 },
		{ 42, 485, 38 },
		{ 42, 486, 38 },
		{ 42, 487, 38 },
		{ 42, 488, 38 },
		{ 42, 489, 38 },
		{ 42, 490, 38 },
		{ 42, 491, 38 },
		{ 42, 492, 38 },
		{ 42, 493, 38 },
		{ 42, 494, 38 },
		{ 42, 495, 38 },
		{ 42, 496, 38 },
		{ 42, 497, 38 },
		{ 42, 498, 38 },
		{ 42, 499, 38 },
		{ 42, 500, 38 },
		{ 42, 501, 38 },
		{ 42, 502, 38 },
		{ 43, 445, 38 },
		{ 43, 446, 38 },
		{ 43, 447, 38 },
		{ 43, 448, 38 },
		{ 43, 449, 38 },
		{ 43, 450, 38 },
		{ 43, 451, 38 },
		{ 43, 452, 38 },
		{ 43, 453, 38 },
		{ 43, 454, 38 },
		{ 43, 455, 38 },
		{ 43, 456, 38 },
		{ 43, 457, 38 },
		{ 43, 458, 38 },
		{ 43, 459, 38 },
		{ 43, 460, 38 },
		{ 43, 461, 38 },
		{ 43, 462, 38 },
		{ 43, 463, 38 },
		{ 43, 464, 38 },
		{ 43, 465, 38 },
		{ 43, 466, 38 },
		{ 43, 467, 38 },
		{ 43, 468, 38 },
		{ 43, 469, 38 },
		{ 43, 470, 38 },
		{ 43, 471, 38 },
		{ 43, 472, 38 },
		{ 43, 473, 38 },
		{ 43, 474, 38 },
		{ 43, 475, 38 },
		{ 43, 476, 38 },
		{ 43, 477, 38 },
		{ 43, 478, 38 },
		{ 43, 479, 38 },
		{ 43, 480, 38 },
		{ 43, 481, 38 },
		{ 43, 482, 38 },
		{ 43, 483, 38 },
		{ 43, 484, 38 },
		{ 43, 485, 38 },
		{ 43, 486, 38 },
		{ 43, 487, 38 },
		{ 43, 488, 38 },
		{ 43, 489, 38 },
		{ 43, 490, 38 },
		{ 43, 491, 38 },
		{ 43, 492, 38 },
		{ 43, 493, 38 },
		{ 43, 494, 38 },
		{ 43, 495, 38 },
		{ 43, 496, 38 },
		{ 43, 497, 38 },
		{ 43, 498, 38 },
		{ 43, 499, 38 },
		{ 43, 500, 38 },
		{ 43, 501, 38 },
		{ 44, 444, 38 },
		{ 44, 445, 38 },
		{ 44, 446, 38 },
		{ 44, 447, 38 },
		{ 44, 448, 38 },
		{ 44, 449, 38 },
		{ 44, 450, 38 },
		{ 44, 451, 38 },
		{ 44, 452, 38 },
		{ 44, 453, 38 },
		{ 44, 454, 38 },
		{ 44, 455, 38 },
		{ 44, 456, 38 },
		{ 44, 457, 38 },
		{ 44, 458, 38 },
		{ 44, 459, 38 },
		{ 44, 460, 38 },
		{ 44, 461, 38 },
		{ 44, 462, 38 },
		{ 44, 463, 38 },
		{ 44, 464, 38 },
		{ 44, 465, 38 },
		{ 44, 466, 38 },
		{ 44, 467, 38 },
		{ 44, 468, 38 },
		{ 44, 469, 38 },
		{ 44, 470, 38 },
		{ 44, 471, 38 },
		{ 44, 472, 38 },
		{ 44, 473, 38 },
		{ 44, 474, 38 },
		{ 44, 475, 38 },
		{ 44, 476, 38 },
		{ 44, 477, 38 },
		{ 44, 478, 38 },
		{ 44, 479, 38 },
		{ 44, 480, 38 },
		{ 44, 481, 38 },
		{ 44, 482, 38 },
		{ 44, 483, 38 },
		{ 44, 484, 38 },
		{ 44, 485, 38 },
		{ 44, 486, 38 },
		{ 44, 487, 38 },
		{ 44, 488, 38 },
		{ 44, 489, 38 },
		{ 44, 490, 38 },
		{ 44, 491, 38 },
		{ 44, 492, 38 },
		{ 44, 493, 38 },
		{ 44, 494, 38 },
		{ 44, 495, 38 },
		{ 44, 496, 38 },
		{ 44, 497, 38 },
		{ 44, 498, 38 },
		{ 44, 499, 38 },
		{ 44, 500, 38 },
		{ 45, 443, 38 },
		{ 45, 444, 38 },
		{ 45, 445, 38 },
		{ 45, 446, 38 },
		{ 45, 447, 38 },
		{ 45, 448, 38 },
		{ 45, 449, 38 },
		{ 45, 450, 38 },
		{ 45, 451, 38 },
		{ 45, 452, 38 },
		{ 45, 453, 38 },
		{ 45, 454, 38 },
		{ 45, 455, 38 },
		{ 45, 456, 38 },
		{ 45, 457, 38 },
		{ 45, 458, 38 },
		{ 45, 459, 38 },
		{ 45, 460, 38 },
		{ 45, 461, 38 },
		{ 45, 462, 38 },
		{ 45, 463, 38 },
		{ 45, 464, 38 },
		{ 45, 465, 38 },
		{ 45, 466, 38 },
		{ 45, 467, 38 },
		{ 45, 468, 38 },
		{ 45, 469, 38 },
		{ 45, 470, 38 },
		{ 45, 471, 38 },
		{ 45, 472, 38 },
		{ 45, 473, 38 },
		{ 45, 474, 38 },
		{ 45, 475, 38 },
		{ 45, 476, 38 },
		{ 45, 477, 38 },
		{ 45, 478, 38 },
		{ 45, 479, 38 },
		{ 45, 480, 38 },
		{ 45, 481, 38 },
		{ 45, 482, 38 },
		{ 45, 483, 38 },
		{ 45, 484, 38 },
		{ 45, 485, 38 },
		{ 45, 486, 38 },
		{ 45, 487, 38 },
		{ 45, 488, 38 },
		{ 45, 489, 38 },
		{ 45, 490, 38 },
		{ 45, 491, 38 },
		{ 45, 492, 38 },
		{ 45, 493, 38 },
		{ 45, 494, 38 },
		{ 45, 495, 38 },
		{ 45, 496, 38 },
		{ 45, 497, 38 },
		{ 45, 498, 38 },
		{ 45, 499, 38 },
		{ 46, 442, 38 },
		{ 46, 443, 38 },
		{ 46, 444, 38 },
		{ 46, 445, 38 },
		{ 46, 446, 38 },
		{ 46, 447, 38 },
		{ 46, 448, 38 },
		{ 46, 449, 38 },
		{ 46, 450, 38 },
		{ 46, 451, 38 },
		{ 46, 452, 38 },
		{ 46, 453, 38 },
		{ 46, 454, 38 },
		{ 46, 455, 38 },
		{ 46, 456, 38 },
		{ 46, 457, 38 },
		{ 46, 458, 38 },
		{ 46, 459, 38 },
		{ 46, 460, 38 },
		{ 46, 461, 38 },
		{ 46, 462, 38 },
		{ 46, 463, 38 },
		{ 46, 464, 38 },
		{ 46, 465, 38 },
		{ 46, 466, 38 },
		{ 46, 467, 38 },
		{ 46, 468, 38 },
		{ 46, 469, 38 },
		{ 46, 470, 38 },
		{ 46, 471, 38 },
		{ 46, 472, 38 },
		{ 46, 473, 38 },
		{ 46, 474, 38 },
		{ 46, 475, 38 },
		{ 46, 476, 38 },
		{ 46, 477, 38 },
		{ 46, 478, 38 },
		{ 46, 479, 38 },
		{ 46, 480, 38 },
		{ 46, 481, 38 },
		{ 46, 482, 38 },
		{ 46, 483, 38 },
		{ 46, 484, 38 },
		{ 46, 485, 38 },
		{ 46, 486, 38 },
		{ 46, 487, 38 },
		{ 46, 488, 38 },
		{ 46, 489, 38 },
		{ 46, 490, 38 },
		{ 46, 491, 38 },
		{ 46, 492, 38 },
		{ 46, 493, 38 },
		{ 46, 494, 38 },
		{ 46, 495, 38 },
		{ 46, 496, 38 },
		{ 46, 497, 38 },
		{ 46, 498, 38 },
		{ 47, 441, 38 },
		{ 47, 442, 38 },
		{ 47, 443, 38 },
		{ 47, 444, 38 },
		{ 47, 445, 38 },
		{ 47, 446, 38 },
		{ 47, 447, 38 },
		{ 47, 448, 38 },
		{ 47, 449, 38 },
		{ 47, 450, 38 },
		{ 47, 451, 38 },
		{ 47, 452, 38 },
		{ 47, 453, 38 },
		{ 47, 454, 38 },
		{ 47, 455, 38 },
		{ 47, 456, 38 },
		{ 47, 457, 38 },
		{ 47, 458, 38 },
		{ 47, 459, 38 },
		{ 47, 460, 38 },
		{ 47, 461, 38 },
		{ 47, 462, 38 },
		{ 47, 463, 38 },
		{ 47, 464, 38 },
		{ 47, 465, 38 },
		{ 47, 466, 38 },
		{ 47, 467, 38 },
		{ 47, 468, 38 },
		{ 47, 469, 38 },
		{ 47, 470, 38 },
		{ 47, 471, 38 },
		{ 47, 472, 38 },
		{ 47, 473, 38 },
		{ 47, 474, 38 },
		{ 47, 475, 38 },
		{ 47, 476, 38 },
		{ 47, 477, 38 },
		{ 47, 478, 38 },
		{ 47, 479, 38 },
		{ 47, 480, 38 },
		{ 47, 481, 38 },
		{ 47, 482, 38 },
		{ 47, 483, 38 },
		{ 47, 484, 38 },
		{ 47, 485, 38 },
		{ 47, 486, 38 },
		{ 47, 487, 38 },
		{ 47, 488, 38 },
		{ 47, 489, 38 },
		{ 47, 490, 38 },
		{ 47, 491, 38 },
		{ 47, 492, 38 },
		{ 47, 493, 38 },
		{ 47, 494, 38 },
		{ 47, 495, 38 },
		{ 47, 496, 38 },
		{ 48, 440, 38 },
		{ 48, 441, 38 },
		{ 48, 442, 38 },
		{ 48, 443, 38 },
		{ 48, 444, 38 },
		{ 48, 445, 38 },
		{ 48, 446, 38 },
		{ 48, 447, 38 },
		{ 48, 448, 38 },
		{ 48, 449, 38 },
		{ 48, 450, 38 },
		{ 48, 451, 38 },
		{ 48, 452, 38 },
		{ 48, 453, 38 },
		{ 48, 454, 38 },
		{ 48, 455, 38 },
		{ 48, 456, 38 },
		{ 48, 457, 38 },
		{ 48, 458, 38 },
		{ 48, 459, 38 },
		{ 48, 460, 38 },
		{ 48, 461, 38 },
		{ 48, 462, 38 },
		{ 48, 463, 38 },
		{ 48, 464, 38 },
		{ 48, 465, 38 },
		{ 48, 466, 38 },
		{ 48, 467, 38 },
		{ 48, 468, 38 },
		{ 48, 469, 38 },
		{ 48, 470, 38 },
		{ 48, 471, 38 },
		{ 48, 472, 38 },
		{ 48, 473, 38 },
		{ 48, 474, 38 },
		{ 48, 475, 38 },
		{ 48, 476, 38 },
		{ 48, 477, 38 },
		{ 48, 478, 38 },
		{ 48, 479, 38 },
		{ 48, 480, 38 },
		{ 48, 481, 38 },
		{ 48, 482, 38 },
		{ 48, 483, 38 },
		{ 48, 484, 38 },
		{ 48, 485, 38 },
		{ 48, 486, 38 },
		{ 48, 487, 38 },
		{ 48, 488, 38 },
		{ 48, 489, 38 },
		{ 48, 490, 38 },
		{ 48, 491, 38 },
		{ 48, 492, 38 },
		{ 48, 493, 38 },
		{ 48, 494, 38 },
		{ 48, 495, 38 },
		{ 49, 440, 38 },
		{ 49, 441, 38 },
		{ 49, 442, 38 },
		{ 49, 443, 38 },
		{ 49, 444, 38 },
		{ 49, 445, 38 },
		{ 49, 446, 38 },
		{ 49, 447, 38 },
		{ 49, 448, 38 },
		{ 49, 449, 38 },
		{ 49, 450, 38 },
		{ 49, 451, 38 },
		{ 49, 452, 38 },
		{ 49, 453, 38 },
		{ 49, 454, 38 },
		{ 49, 455, 38 },
		{ 49, 456, 38 },
		{ 49, 457, 38 },
		{ 49, 458, 38 },
		{ 49, 459, 38 },
		{ 49, 460, 38 },
		{ 49, 461, 38 },
		{ 49, 462, 38 },
		{ 49, 463, 38 },
		{ 49, 464, 38 },
		{ 49, 465, 38 },
		{ 49, 466, 38 },
		{ 49, 467, 38 },
		{ 49, 468, 38 },
		{ 49, 469, 38 },
		{ 49, 470, 38 },
		{ 49, 471, 38 },
		{ 49, 472, 38 },
		{ 49, 473, 38 },
		{ 49, 474, 38 },
		{ 49, 475, 38 },
		{ 49, 476, 38 },
		{ 49, 477, 38 },
		{ 49, 478, 38 },
		{ 49, 479, 38 },
		{ 49, 480, 38 },
		{ 49, 481, 38 },
		{ 49, 482, 38 },
		{ 49, 483, 38 },
		{ 49, 484, 38 },
		{ 49, 485, 38 },
		{ 49, 486, 38 },
		{ 49, 487, 38 },
		{ 49, 488, 38 },
		{ 49, 489, 38 },
		{ 49, 490, 38 },
		{ 49, 491, 38 },
		{ 49, 492, 38 },
		{ 49, 493, 38 },
		{ 49, 494, 38 },
		{ 50, 439, 38 },
		{ 50, 440, 38 },
		{ 50, 441, 38 },
		{ 50, 442, 38 },
		{ 50, 443, 38 },
		{ 50, 444, 38 },
		{ 50, 445, 38 },
		{ 50, 446, 38 },
		{ 50, 447, 38 },
		{ 50, 448, 38 },
		{ 50, 449, 38 },
		{ 50, 450, 38 },
		{ 50, 451, 38 },
		{ 50, 452, 38 },
		{ 50, 453, 38 },
		{ 50, 454, 38 },
		{ 50, 455, 38 },
		{ 50, 456, 38 },
		{ 50, 457, 38 },
		{ 50, 458, 38 },
		{ 50, 459, 38 },
		{ 50, 460, 38 },
		{ 50, 461, 38 },
		{ 50, 462, 38 },
		{ 50, 463, 38 },
		{ 50, 464, 38 },
		{ 50, 465, 38 },
		{ 50, 466, 38 },
		{ 50, 467, 38 },
		{ 50, 468, 38 },
		{ 50, 469, 38 },
		{ 50, 470, 38 },
		{ 50, 471, 38 },
		{ 50, 472, 38 },
		{ 50, 473, 38 },
		{ 50, 474, 38 },
		{ 50, 475, 38 },
		{ 50, 476, 38 },
		{ 50, 477, 38 },
		{ 50, 478, 38 },
		{ 50, 479, 38 },
		{ 50, 480, 38 },
		{ 50, 481, 38 },
		{ 50, 482, 38 },
		{ 50, 483, 38 },
		{ 50, 484, 38 },
		{ 50, 485, 38 },
		{ 50, 486, 38 },
		{ 50, 487, 38 },
		{ 50, 488, 38 },
		{ 50, 489, 38 },
		{ 50, 490, 38 },
		{ 50, 491, 38 },
		{ 50, 492, 38 },
		{ 50, 493, 38 },
		{ 51, 438, 38 },
		{ 51, 439, 38 },
		{ 51, 440, 38 },
		{ 51, 441, 38 },
		{ 51, 442, 38 },
		{ 51, 443, 38 },
		{ 51, 444, 38 },
		{ 51, 445, 38 },
		{ 51, 446, 38 },
		{ 51, 447, 38 },
		{ 51, 448, 38 },
		{ 51, 449, 38 },
		{ 51, 450, 38 },
		{ 51, 451, 38 },
		{ 51, 452, 38 },
		{ 51, 453, 38 },
		{ 51, 454, 38 },
		{ 51, 455, 38 },
		{ 51, 456, 38 },
		{ 51, 457, 38 },
		{ 51, 458, 38 },
		{ 51, 459, 38 },
		{ 51, 460, 38 },
		{ 51, 461, 38 },
		{ 51, 462, 38 },
		{ 51, 463, 38 },
		{ 51, 464, 38 },
		{ 51, 465, 38 },
		{ 51, 466, 38 },
		{ 51, 467, 38 },
		{ 51, 468, 38 },
		{ 51, 469, 38 },
		{ 51, 470, 38 },
		{ 51, 471, 38 },
		{ 51, 472, 38 },
		{ 51, 473, 38 },
		{ 51, 474, 38 },
		{ 51, 475, 38 },
		{ 51, 476, 38 },
		{ 51, 477, 38 },
		{ 51, 478, 38 },
		{ 51, 479, 38 },
		{ 51, 480, 38 },
		{ 51, 481, 38 },
		{ 51, 482, 38 },
		{ 51, 483, 38 },
		{ 51, 484, 38 },
		{ 51, 485, 38 },
		{ 51, 486, 38 },
		{ 51, 487, 38 },
		{ 51, 488, 38 },
		{ 51, 489, 38 },
		{ 51, 490, 38 },
		{ 51, 491, 38 },
		{ 51, 492, 38 },
		{ 52, 437, 38 },
		{ 52, 438, 38 },
		{ 52, 439, 38 },
		{ 52, 440, 38 },
		{ 52, 441, 38 },
		{ 52, 442, 38 },
		{ 52, 443, 38 },
		{ 52, 444, 38 },
		{ 52, 445, 38 },
		{ 52, 446, 38 },
		{ 52, 447, 38 },
		{ 52, 448, 38 },
		{ 52, 449, 38 },
		{ 52, 450, 38 },
		{ 52, 451, 38 },
		{ 52, 452, 38 },
		{ 52, 453, 38 },
		{ 52, 454, 38 },
		{ 52, 455, 38 },
		{ 52, 456, 38 },
		{ 52, 457, 38 },
		{ 52, 458, 38 },
		{ 52, 459, 38 },
		{ 52, 460, 38 },
		{ 52, 461, 38 },
		{ 52, 462, 38 },
		{ 52, 463, 38 },
		{ 52, 464, 38 },
		{ 52, 465, 38 },
		{ 52, 466, 38 },
		{ 52, 467, 38 },
		{ 52, 468, 38 },
		{ 52, 469, 38 },
		{ 52, 470, 38 },
		{ 52, 471, 38 },
		{ 52, 472, 38 },
		{ 52, 473, 38 },
		{ 52, 474, 38 },
		{ 52, 475, 38 },
		{ 52, 476, 38 },
		{ 52, 477, 38 },
		{ 52, 478, 38 },
		{ 52, 479, 38 },
		{ 52, 480, 38 },
		{ 52, 481, 38 },
		{ 52, 482, 38 },
		{ 52, 483, 38 },
		{ 52, 484, 38 },
		{ 52, 485, 38 },
		{ 52, 486, 38 },
		{ 52, 487, 38 },
		{ 52, 488, 38 },
		{ 52, 489, 38 },
		{ 52, 490, 38 },
		{ 52, 491, 38 },
		{ 53, 436, 38 },
		{ 53, 437, 38 },
		{ 53, 438, 38 },
		{ 53, 439, 38 },
		{ 53, 440, 38 },
		{ 53, 441, 38 },
		{ 53, 442, 38 },
		{ 53, 443, 38 },
		{ 53, 444, 38 },
		{ 53, 445, 38 },
		{ 53, 446, 38 },
		{ 53, 447, 38 },
		{ 53, 448, 38 },
		{ 53, 449, 38 },
		{ 53, 450, 38 },
		{ 53, 451, 38 },
		{ 53, 452, 38 },
		{ 53, 453, 38 },
		{ 53, 454, 38 },
		{ 53, 455, 38 },
		{ 53, 456, 38 },
		{ 53, 457, 38 },
		{ 53, 458, 38 },
		{ 53, 459, 38 },
		{ 53, 460, 38 },
		{ 53, 461, 38 },
		{ 53, 462, 38 },
		{ 53, 463, 38 },
		{ 53, 464, 38 },
		{ 53, 465, 38 },
		{ 53, 466, 38 },
		{ 53, 467, 38 },
		{ 53, 468, 38 },
		{ 53, 469, 38 },
		{ 53, 470, 38 },
		{ 53, 471, 38 },
		{ 53, 472, 38 },
		{ 53, 473, 38 },
		{ 53, 474, 38 },
		{ 53, 475, 38 },
		{ 53, 476, 38 },
		{ 53, 477, 38 },
		{ 53, 478, 38 },
		{ 53, 479, 38 },
		{ 53, 480, 38 },
		{ 53, 481, 38 },
		{ 53, 482, 38 },
		{ 53, 483, 38 },
		{ 53, 484, 38 },
		{ 53, 485, 38 },
		{ 53, 486, 38 },
		{ 53, 487, 38 },
		{ 53, 488, 38 },
		{ 53, 489, 38 },
		{ 53, 490, 38 },
		{ 54, 435, 38 },
		{ 54, 436, 38 },
		{ 54, 437, 38 },
		{ 54, 438, 38 },
		{ 54, 439, 38 },
		{ 54, 440, 38 },
		{ 54, 441, 38 },
		{ 54, 442, 38 },
		{ 54, 443, 38 },
		{ 54, 444, 38 },
		{ 54, 445, 38 },
		{ 54, 446, 38 },
		{ 54, 447, 38 },
		{ 54, 448, 38 },
		{ 54, 449, 38 },
		{ 54, 450, 38 },
		{ 54, 451, 38 },
		{ 54, 452, 38 },
		{ 54, 453, 38 },
		{ 54, 454, 38 },
		{ 54, 455, 38 },
		{ 54, 456, 38 },
		{ 54, 457, 38 },
		{ 54, 458, 38 },
		{ 54, 459, 38 },
		{ 54, 460, 38 },
		{ 54, 461, 38 },
		{ 54, 462, 38 },
		{ 54, 463, 38 },
		{ 54, 464, 38 },
		{ 54, 465, 38 },
		{ 54, 466, 38 },
		{ 54, 467, 38 },
		{ 54, 468, 38 },
		{ 54, 469, 38 },
		{ 54, 470, 38 },
		{ 54, 471, 38 },
		{ 54, 472, 38 },
		{ 54, 473, 38 },
		{ 54, 474, 38 },
		{ 54, 475, 38 },
		{ 54, 476, 38 },
		{ 54, 477, 38 },
		{ 54, 478, 38 },
		{ 54, 479, 38 },
		{ 54, 480, 38 },
		{ 54, 481, 38 },
		{ 54, 482, 38 },
		{ 54, 483, 38 },
		{ 54, 484, 38 },
		{ 54, 485, 38 },
		{ 54, 486, 38 },
		{ 54, 487, 38 },
		{ 54, 488, 38 },
		{ 55, 434, 38 },
		{ 55, 435, 38 },
		{ 55, 436, 38 },
		{ 55, 437, 38 },
		{ 55, 438, 38 },
		{ 55, 439, 38 },
		{ 55, 440, 38 },
		{ 55, 441, 38 },
		{ 55, 442, 38 },
		{ 55, 443, 38 },
		{ 55, 444, 38 },
		{ 55, 445, 38 },
		{ 55, 446, 38 },
		{ 55, 447, 38 },
		{ 55, 448, 38 },
		{ 55, 449, 38 },
		{ 55, 450, 38 },
		{ 55, 451, 38 },
		{ 55, 452, 38 },
		{ 55, 453, 38 },
		{ 55, 454, 38 },
		{ 55, 455, 38 },
		{ 55, 456, 38 },
		{ 55, 457, 38 },
		{ 55, 458, 38 },
		{ 55, 459, 38 },
		{ 55, 460, 38 },
		{ 55, 461, 38 },
		{ 55, 462, 38 },
		{ 55, 463, 38 },
		{ 55, 464, 38 },
		{ 55, 465, 38 },
		{ 55, 466, 38 },
		{ 55, 467, 38 },
		{ 55, 468, 38 },
		{ 55, 469, 38 },
		{ 55, 470, 38 },
		{ 55, 471, 38 },
		{ 55, 472, 38 },
		{ 55, 473, 38 },
		{ 55, 474, 38 },
		{ 55, 475, 38 },
		{ 55, 476, 38 },
		{ 55, 477, 38 },
		{ 55, 478, 38 },
		{ 55, 479, 38 },
		{ 55, 480, 38 },
		{ 55, 481, 38 },
		{ 55, 482, 38 },
		{ 55, 483, 38 },
		{ 55, 484, 38 },
		{ 55, 485, 38 },
		{ 55, 486, 38 },
		{ 55, 487, 38 },
		{ 56, 433, 38 },
		{ 56, 434, 38 },
		{ 56, 435, 38 },
		{ 56, 436, 38 },
		{ 56, 437, 38 },
		{ 56, 438, 38 },
		{ 56, 439, 38 },
		{ 56, 440, 38 },
		{ 56, 441, 38 },
		{ 56, 442, 38 },
		{ 56, 443, 38 },
		{ 56, 444, 38 },
		{ 56, 445, 38 },
		{ 56, 446, 38 },
		{ 56, 447, 38 },
		{ 56, 448, 38 },
		{ 56, 449, 38 },
		{ 56, 450, 38 },
		{ 56, 451, 38 },
		{ 56, 452, 38 },
		{ 56, 453, 38 },
		{ 56, 454, 38 },
		{ 56, 455, 38 },
		{ 56, 456, 38 },
		{ 56, 457, 38 },
		{ 56, 458, 38 },
		{ 56, 459, 38 },
		{ 56, 460, 38 },
		{ 56, 461, 38 },
		{ 56, 462, 38 },
		{ 56, 463, 38 },
		{ 56, 464, 38 },
		{ 56, 465, 38 },
		{ 56, 466, 38 },
		{ 56, 467, 38 },
		{ 56, 468, 38 },
		{ 56, 469, 38 },
		{ 56, 470, 38 },
		{ 56, 471, 38 },
		{ 56, 472, 38 },
		{ 56, 473, 38 },
		{ 56, 474, 38 },
		{ 56, 475, 38 },
		{ 56, 476, 38 },
		{ 56, 477, 38 },
		{ 56, 478, 38 },
		{ 56, 479, 38 },
		{ 56, 480, 38 },
		{ 56, 481, 38 },
		{ 56, 482, 38 },
		{ 56, 483, 38 },
		{ 56, 484, 38 },
		{ 56, 485, 38 },
		{ 56, 486, 38 },
		{ 57, 433, 38 },
		{ 57, 434, 38 },
		{ 57, 435, 38 },
		{ 57, 436, 38 },
		{ 57, 437, 38 },
		{ 57, 438, 38 },
		{ 57, 439, 38 },
		{ 57, 440, 38 },
		{ 57, 441, 38 },
		{ 57, 442, 38 },
		{ 57, 443, 38 },
		{ 57, 444, 38 },
		{ 57, 445, 38 },
		{ 57, 446, 38 },
		{ 57, 447, 38 },
		{ 57, 448, 38 },
		{ 57, 449, 38 },
		{ 57, 450, 38 },
		{ 57, 451, 38 },
		{ 57, 452, 38 },
		{ 57, 453, 38 },
		{ 57, 454, 38 },
		{ 57, 455, 38 },
		{ 57, 456, 38 },
		{ 57, 457, 38 },
		{ 57, 458, 38 },
		{ 57, 459, 38 },
		{ 57, 460, 38 },
		{ 57, 461, 38 },
		{ 57, 462, 38 },
		{ 57, 463, 38 },
		{ 57, 464, 38 },
		{ 57, 465, 38 },
		{ 57, 466, 38 },
		{ 57, 467, 38 },
		{ 57, 468, 38 },
		{ 57, 469, 38 },
		{ 57, 470, 38 },
		{ 57, 471, 38 },
		{ 57, 472, 38 },
		{ 57, 473, 38 },
		{ 57, 474, 38 },
		{ 57, 475, 38 },
		{ 57, 476, 38 },
		{ 57, 477, 38 },
		{ 57, 478, 38 },
		{ 57, 479, 38 },
		{ 57, 480, 38 },
		{ 57, 481, 38 },
		{ 57, 482, 38 },
		{ 57, 483, 38 },
		{ 57, 484, 38 },
		{ 57, 485, 38 },
		{ 58, 432, 38 },
		{ 58, 433, 38 },
		{ 58, 434, 38 },
		{ 58, 435, 38 },
		{ 58, 436, 38 },
		{ 58, 437, 38 },
		{ 58, 438, 38 },
		{ 58, 439, 38 },
		{ 58, 440, 38 },
		{ 58, 441, 38 },
		{ 58, 442, 38 },
		{ 58, 443, 38 },
		{ 58, 444, 38 },
		{ 58, 445, 38 },
		{ 58, 446, 38 },
		{ 58, 447, 38 },
		{ 58, 448, 38 },
		{ 58, 449, 38 },
		{ 58, 450, 38 },
		{ 58, 451, 38 },
		{ 58, 452, 38 },
		{ 58, 453, 38 },
		{ 58, 454, 38 },
		{ 58, 455, 38 },
		{ 58, 456, 38 },
		{ 58, 457, 38 },
		{ 58, 458, 38 },
		{ 58, 459, 38 },
		{ 58, 460, 38 },
		{ 58, 461, 38 },
		{ 58, 462, 38 },
		{ 58, 463, 38 },
		{ 58, 464, 38 },
		{ 58, 465, 38 },
		{ 58, 466, 38 },
		{ 58, 467, 38 },
		{ 58, 468, 38 },
		{ 58, 469, 38 },
		{ 58, 470, 38 },
		{ 58, 471, 38 },
		{ 58, 472, 38 },
		{ 58, 473, 38 },
		{ 58, 474, 38 },
		{ 58, 475, 38 },
		{ 58, 476, 38 },
		{ 58, 477, 38 },
		{ 58, 478, 38 },
		{ 58, 479, 38 },
		{ 58, 480, 38 },
		{ 58, 481, 38 },
		{ 58, 482, 38 },
		{ 58, 483, 38 },
		{ 58, 484, 38 },
		{ 59, 431, 38 },
		{ 59, 432, 38 },
		{ 59, 433, 38 },
		{ 59, 434, 38 },
		{ 59, 435, 38 },
		{ 59, 436, 38 },
		{ 59, 437, 38 },
		{ 59, 438, 38 },
		{ 59, 439, 38 },
		{ 59, 440, 38 },
		{ 59, 441, 38 },
		{ 59, 442, 38 },
		{ 59, 443, 38 },
		{ 59, 444, 38 },
		{ 59, 445, 38 },
		{ 59, 446, 38 },
		{ 59, 447, 38 },
		{ 59, 448, 38 },
		{ 59, 449, 38 },
		{ 59, 450, 38 },
		{ 59, 451, 38 },
		{ 59, 452, 38 },
		{ 59, 453, 38 },
		{ 59, 454, 38 },
		{ 59, 455, 38 },
		{ 59, 456, 38 },
		{ 59, 457, 38 },
		{ 59, 458, 38 },
		{ 59, 459, 38 },
		{ 59, 460, 38 },
		{ 59, 461, 38 },
		{ 59, 462, 38 },
		{ 59, 463, 38 },
		{ 59, 464, 38 },
		{ 59, 465, 38 },
		{ 59, 466, 38 },
		{ 59, 467, 38 },
		{ 59, 468, 38 },
		{ 59, 469, 38 },
		{ 59, 470, 38 },
		{ 59, 471, 38 },
		{ 59, 472, 38 },
		{ 59, 473, 38 },
		{ 59, 474, 38 },
		{ 59, 475, 38 },
		{ 59, 476, 38 },
		{ 59, 477, 38 },
		{ 59, 478, 38 },
		{ 59, 479, 38 },
		{ 59, 480, 38 },
		{ 59, 481, 38 },
		{ 59, 482, 38 },
		{ 59, 483, 38 },
		{ 60, 430, 38 },
		{ 60, 431, 38 },
		{ 60, 432, 38 },
		{ 60, 433, 38 },
		{ 60, 434, 38 },
		{ 60, 435, 38 },
		{ 60, 436, 38 },
		{ 60, 437, 38 },
		{ 60, 438, 38 },
		{ 60, 439, 38 },
		{ 60, 440, 38 },
		{ 60, 441, 38 },
		{ 60, 442, 38 },
		{ 60, 443, 38 },
		{ 60, 444, 38 },
		{ 60, 445, 38 },
		{ 60, 446, 38 },
		{ 60, 447, 38 },
		{ 60, 448, 38 },
		{ 60, 449, 38 },
		{ 60, 450, 38 },
		{ 60, 451, 38 },
		{ 60, 452, 38 },
		{ 60, 453, 38 },
		{ 60, 454, 38 },
		{ 60, 455, 38 },
		{ 60, 456, 38 },
		{ 60, 457, 38 },
		{ 60, 458, 38 },
		{ 60, 459, 38 },
		{ 60, 460, 38 },
		{ 60, 461, 38 },
		{ 60, 462, 38 },
		{ 60, 463, 38 },
		{ 60, 464, 38 },
		{ 60, 465, 38 },
		{ 60, 466, 38 },
		{ 60, 467, 38 },
		{ 60, 468, 38 },
		{ 60, 469, 38 },
		{ 60, 470, 38 },
		{ 60, 471, 38 },
		{ 60, 472, 38 },
		{ 60, 473, 38 },
		{ 60, 474, 38 },
		{ 60, 475, 38 },
		{ 60, 476, 38 },
		{ 60, 477, 38 },
		{ 60, 478, 38 },
		{ 60, 479, 38 },
		{ 60, 480, 38 },
		{ 60, 481, 38 },
		{ 60, 482, 38 },
		{ 61, 429, 38 },
		{ 61, 430, 38 },
		{ 61, 431, 38 },
		{ 61, 432, 38 },
		{ 61, 433, 38 },
		{ 61, 434, 38 },
		{ 61, 435, 38 },
		{ 61, 436, 38 },
		{ 61, 437, 38 },
		{ 61, 438, 38 },
		{ 61, 439, 38 },
		{ 61, 440, 38 },
		{ 61, 441, 38 },
		{ 61, 442, 38 },
		{ 61, 443, 38 },
		{ 61, 444, 38 },
		{ 61, 445, 38 },
		{ 61, 446, 38 },
		{ 61, 447, 38 },
		{ 61, 448, 38 },
		{ 61, 449, 38 },
		{ 61, 450, 38 },
		{ 61, 451, 38 },
		{ 61, 452, 38 },
		{ 61, 453, 38 },
		{ 61, 454, 38 },
		{ 61, 455, 38 },
		{ 61, 456, 38 },
		{ 61, 457, 38 },
		{ 61, 458, 38 },
		{ 61, 459, 38 },
		{ 61, 460, 38 },
		{ 61, 461, 38 },
		{ 61, 462, 38 },
		{ 61, 463, 38 },
		{ 61, 464, 38 },
		{ 61, 465, 38 },
		{ 61, 466, 38 },
		{ 61, 467, 38 },
		{ 61, 468, 38 },
		{ 61, 469, 38 },
		{ 61, 470, 38 },
		{ 61, 471, 38 },
		{ 61, 472, 38 },
		{ 61, 473, 38 },
		{ 61, 474, 38 },
		{ 61, 475, 38 },
		{ 61, 476, 38 },
		{ 61, 477, 38 },
		{ 61, 478, 38 },
		{ 61, 479, 38 },
		{ 61, 480, 38 },
		{ 62, 428, 38 },
		{ 62, 429, 38 },
		{ 62, 430, 38 },
		{ 62, 431, 38 },
		{ 62, 432, 38 },
		{ 62, 433, 38 },
		{ 62, 434, 38 },
		{ 62, 435, 38 },
		{ 62, 436, 38 },
		{ 62, 437, 38 },
		{ 62, 438, 38 },
		{ 62, 439, 38 },
		{ 62, 440, 38 },
		{ 62, 441, 38 },
		{ 62, 442, 38 },
		{ 62, 443, 38 },
		{ 62, 444, 38 },
		{ 62, 445, 38 },
		{ 62, 446, 38 },
		{ 62, 447, 38 },
		{ 62, 448, 38 },
		{ 62, 449, 38 },
		{ 62, 450, 38 },
		{ 62, 451, 38 },
		{ 62, 452, 38 },
		{ 62, 453, 38 },
		{ 62, 454, 38 },
		{ 62, 455, 38 },
		{ 62, 456, 38 },
		{ 62, 457, 38 },
		{ 62, 458, 38 },
		{ 62, 459, 38 },
		{ 62, 460, 38 },
		{ 62, 461, 38 },
		{ 62, 462, 38 },
		{ 62, 463, 38 },
		{ 62, 464, 38 },
		{ 62, 465, 38 },
		{ 62, 466, 38 },
		{ 62, 467, 38 },
		{ 62, 468, 38 },
		{ 62, 469, 38 },
		{ 62, 470, 38 },
		{ 62, 471, 38 },
		{ 62, 472, 38 },
		{ 62, 473, 38 },
		{ 62, 474, 38 },
		{ 62, 475, 38 },
		{ 62, 476, 38 },
		{ 62, 477, 38 },
		{ 62, 478, 38 },
		{ 62, 479, 38 },
		{ 63, 427, 38 },
		{ 63, 428, 38 },
		{ 63, 429, 38 },
		{ 63, 430, 38 },
		{ 63, 431, 38 },
		{ 63, 432, 38 },
		{ 63, 433, 38 },
		{ 63, 434, 38 },
		{ 63, 435, 38 },
		{ 63, 436, 38 },
		{ 63, 437, 38 },
		{ 63, 438, 38 },
		{ 63, 439, 38 },
		{ 63, 440, 38 },
		{ 63, 441, 38 },
		{ 63, 442, 38 },
		{ 63, 443, 38 },
		{ 63, 444, 38 },
		{ 63, 445, 38 },
		{ 63, 446, 38 },
		{ 63, 447, 38 },
		{ 63, 448, 38 },
		{ 63, 449, 38 },
		{ 63, 450, 38 },
		{ 63, 451, 38 },
		{ 63, 452, 38 },
		{ 63, 453, 38 },
		{ 63, 454, 38 },
		{ 63, 455, 38 },
		{ 63, 456, 38 },
		{ 63, 457, 38 },
		{ 63, 458, 38 },
		{ 63, 459, 38 },
		{ 63, 460, 38 },
		{ 63, 461, 38 },
		{ 63, 462, 38 },
		{ 63, 463, 38 },
		{ 63, 464, 38 },
		{ 63, 465, 38 },
		{ 63, 466, 38 },
		{ 63, 467, 38 },
		{ 63, 468, 38 },
		{ 63, 469, 38 },
		{ 63, 470, 38 },
		{ 63, 471, 38 },
		{ 63, 472, 38 },
		{ 63, 473, 38 },
		{ 63, 474, 38 },
		{ 63, 475, 38 },
		{ 63, 476, 38 },
		{ 63, 477, 38 },
		{ 63, 478, 38 },
		{ 64, 426, 38 },
		{ 64, 427, 38 },
		{ 64, 428, 38 },
		{ 64, 429, 38 },
		{ 64, 430, 38 },
		{ 64, 431, 38 },
		{ 64, 432, 38 },
		{ 64, 433, 38 },
		{ 64, 434, 38 },
		{ 64, 435, 38 },
		{ 64, 436, 38 },
		{ 64, 437, 38 },
		{ 64, 438, 38 },
		{ 64, 439, 38 },
		{ 64, 440, 38 },
		{ 64, 441, 38 },
		{ 64, 442, 38 },
		{ 64, 443, 38 },
		{ 64, 444, 38 },
		{ 64, 445, 38 },
		{ 64, 446, 38 },
		{ 64, 447, 38 },
		{ 64, 448, 38 },
		{ 64, 449, 38 },
		{ 64, 450, 38 },
		{ 64, 451, 38 },
		{ 64, 452, 38 },
		{ 64, 453, 38 },
		{ 64, 454, 38 },
		{ 64, 455, 38 },
		{ 64, 456, 38 },
		{ 64, 457, 38 },
		{ 64, 458, 38 },
		{ 64, 459, 38 },
		{ 64, 460, 38 },
		{ 64, 461, 38 },
		{ 64, 462, 38 },
		{ 64, 463, 38 },
		{ 64, 464, 38 },
		{ 64, 465, 38 },
		{ 64, 466, 38 },
		{ 64, 467, 38 },
		{ 64, 468, 38 },
		{ 64, 469, 38 },
		{ 64, 470, 38 },
		{ 64, 471, 38 },
		{ 64, 472, 38 },
		{ 64, 473, 38 },
		{ 64, 474, 38 },
		{ 64, 475, 38 },
		{ 64, 476, 38 },
		{ 64, 477, 38 },
		{ 65, 426, 38 },
		{ 65, 427, 38 },
		{ 65, 428, 38 },
		{ 65, 429, 38 },
		{ 65, 430, 38 },
		{ 65, 431, 38 },
		{ 65, 432, 38 },
		{ 65, 433, 38 },
		{ 65, 434, 38 },
		{ 65, 435, 38 },
		{ 65, 436, 38 },
		{ 65, 437, 38 },
		{ 65, 438, 38 },
		{ 65, 439, 38 },
		{ 65, 440, 38 },
		{ 65, 441, 38 },
		{ 65, 442, 38 },
		{ 65, 443, 38 },
		{ 65, 444, 38 },
		{ 65, 445, 38 },
		{ 65, 446, 38 },
		{ 65, 447, 38 },
		{ 65, 448, 38 },
		{ 65, 449, 38 },
		{ 65, 450, 38 },
		{ 65, 451, 38 },
		{ 65, 452, 38 },
		{ 65, 453, 38 },
		{ 65, 454, 38 },
		{ 65, 455, 38 },
		{ 65, 456, 38 },
		{ 65, 457, 38 },
		{ 65, 458, 38 },
		{ 65, 459, 38 },
		{ 65, 460, 38 },
		{ 65, 461, 38 },
		{ 65, 462, 38 },
		{ 65, 463, 38 },
		{ 65, 464, 38 },
		{ 65, 465, 38 },
		{ 65, 466, 38 },
		{ 65, 467, 38 },
		{ 65, 468, 38 },
		{ 65, 469, 38 },
		{ 65, 470, 38 },
		{ 65, 471, 38 },
		{ 65, 472, 38 },
		{ 65, 473, 38 },
		{ 65, 474, 38 },
		{ 65, 475, 38 },
		{ 65, 476, 38 },
		{ 66, 425, 38 },
		{ 66, 426, 38 },
		{ 66, 427, 38 },
		{ 66, 428, 38 },
		{ 66, 429, 38 },
		{ 66, 430, 38 },
		{ 66, 431, 38 },
		{ 66, 432, 38 },
		{ 66, 433, 38 },
		{ 66, 434, 38 },
		{ 66, 435, 38 },
		{ 66, 436, 38 },
		{ 66, 437, 38 },
		{ 66, 438, 38 },
		{ 66, 439, 38 },
		{ 66, 440, 38 },
		{ 66, 441, 38 },
		{ 66, 442, 38 },
		{ 66, 443, 38 },
		{ 66, 444, 38 },
		{ 66, 445, 38 },
		{ 66, 446, 38 },
		{ 66, 447, 38 },
		{ 66, 448, 38 },
		{ 66, 449, 38 },
		{ 66, 450, 38 },
		{ 66, 451, 38 },
		{ 66, 452, 38 },
		{ 66, 453, 38 },
		{ 66, 454, 38 },
		{ 66, 455, 38 },
		{ 66, 456, 38 },
		{ 66, 457, 38 },
		{ 66, 458, 38 },
		{ 66, 459, 38 },
		{ 66, 460, 38 },
		{ 66, 461, 38 },
		{ 66, 462, 38 },
		{ 66, 463, 38 },
		{ 66, 464, 38 },
		{ 66, 465, 38 },
		{ 66, 466, 38 },
		{ 66, 467, 38 },
		{ 66, 468, 38 },
		{ 66, 469, 38 },
		{ 66, 470, 38 },
		{ 66, 471, 38 },
		{ 66, 472, 38 },
		{ 66, 473, 38 },
		{ 66, 474, 38 },
		{ 66, 475, 38 },
		{ 67, 424, 38 },
		{ 67, 425, 38 },
		{ 67, 426, 38 },
		{ 67, 427, 38 },
		{ 67, 428, 38 },
		{ 67, 429, 38 },
		{ 67, 430, 38 },
		{ 67, 431, 38 },
		{ 67, 432, 38 },
		{ 67, 433, 38 },
		{ 67, 434, 38 },
		{ 67, 435, 38 },
		{ 67, 436, 38 },
		{ 67, 437, 38 },
		{ 67, 438, 38 },
		{ 67, 439, 38 },
		{ 67, 440, 38 },
		{ 67, 441, 38 },
		{ 67, 442, 38 },
		{ 67, 443, 38 },
		{ 67, 444, 38 },
		{ 67, 445, 38 },
		{ 67, 446, 38 },
		{ 67, 447, 38 },
		{ 67, 448, 38 },
		{ 67, 449, 38 },
		{ 67, 450, 38 },
		{ 67, 451, 38 },
		{ 67, 452, 38 },
		{ 67, 453, 38 },
		{ 67, 454, 38 },
		{ 67, 455, 38 },
		{ 67, 456, 38 },
		{ 67, 457, 38 },
		{ 67, 458, 38 },
		{ 67, 459, 38 },
		{ 67, 460, 38 },
		{ 67, 461, 38 },
		{ 67, 462, 38 },
		{ 67, 463, 38 },
		{ 67, 464, 38 },
		{ 67, 465, 38 },
		{ 67, 466, 38 },
		{ 67, 467, 38 },
		{ 67, 468, 38 },
		{ 67, 469, 38 },
		{ 67, 470, 38 },
		{ 67, 471, 38 },
		{ 67, 472, 38 },
		{ 67, 473, 38 },
		{ 67, 474, 38 },
		{ 68, 423, 38 },
		{ 68, 424, 38 },
		{ 68, 425, 38 },
		{ 68, 426, 38 },
		{ 68, 427, 38 },
		{ 68, 428, 38 },
		{ 68, 429, 38 },
		{ 68, 430, 38 },
		{ 68, 431, 38 },
		{ 68, 432, 38 },
		{ 68, 433, 38 },
		{ 68, 434, 38 },
		{ 68, 435, 38 },
		{ 68, 436, 38 },
		{ 68, 437, 38 },
		{ 68, 438, 38 },
		{ 68, 439, 38 },
		{ 68, 440, 38 },
		{ 68, 441, 38 },
		{ 68, 442, 38 },
		{ 68, 443, 38 },
		{ 68, 444, 38 },
		{ 68, 445, 38 },
		{ 68, 446, 38 },
		{ 68, 447, 38 },
		{ 68, 448, 38 },
		{ 68, 449, 38 },
		{ 68, 450, 38 },
		{ 68, 451, 38 },
		{ 68, 452, 38 },
		{ 68, 453, 38 },
		{ 68, 454, 38 },
		{ 68, 455, 38 },
		{ 68, 456, 38 },
		{ 68, 457, 38 },
		{ 68, 458, 38 },
		{ 68, 459, 38 },
		{ 68, 460, 38 },
		{ 68, 461, 38 },
		{ 68, 462, 38 },
		{ 68, 463, 38 },
		{ 68, 464, 38 },
		{ 68, 465, 38 },
		{ 68, 466, 38 },
		{ 68, 467, 38 },
		{ 68, 468, 38 },
		{ 68, 469, 38 },
		{ 68, 470, 38 },
		{ 68, 471, 38 },
		{ 68, 472, 38 },
		{ 69, 422, 38 },
		{ 69, 423, 38 },
		{ 69, 424, 38 },
		{ 69, 425, 38 },
		{ 69, 426, 38 },
		{ 69, 427, 38 },
		{ 69, 428, 38 },
		{ 69, 429, 38 },
		{ 69, 430, 38 },
		{ 69, 431, 38 },
		{ 69, 432, 38 },
		{ 69, 433, 38 },
		{ 69, 434, 38 },
		{ 69, 435, 38 },
		{ 69, 436, 38 },
		{ 69, 437, 38 },
		{ 69, 438, 38 },
		{ 69, 439, 38 },
		{ 69, 440, 38 },
		{ 69, 441, 38 },
		{ 69, 442, 38 },
		{ 69, 443, 38 },
		{ 69, 444, 38 },
		{ 69, 445, 38 },
		{ 69, 446, 38 },
		{ 69, 447, 38 },
		{ 69, 448, 38 },
		{ 69, 449, 38 },
		{ 69, 450, 38 },
		{ 69, 451, 38 },
		{ 69, 452, 38 },
		{ 69, 453, 38 },
		{ 69, 454, 38 },
		{ 69, 455, 38 },
		{ 69, 456, 38 },
		{ 69, 457, 38 },
		{ 69, 458, 38 },
		{ 69, 459, 38 },
		{ 69, 460, 38 },
		{ 69, 461, 38 },
		{ 69, 462, 38 },
		{ 69, 463, 38 },
		{ 69, 464, 38 },
		{ 69, 465, 38 },
		{ 69, 466, 38 },
		{ 69, 467, 38 },
		{ 69, 468, 38 },
		{ 69, 469, 38 },
		{ 69, 470, 38 },
		{ 69, 471, 38 },
		{ 70, 421, 38 },
		{ 70, 422, 38 },
		{ 70, 423, 38 },
		{ 70, 424, 38 },
		{ 70, 425, 38 },
		{ 70, 426, 38 },
		{ 70, 427, 38 },
		{ 70, 428, 38 },
		{ 70, 429, 38 },
		{ 70, 430, 38 },
		{ 70, 431, 38 },
		{ 70, 432, 38 },
		{ 70, 433, 38 },
		{ 70, 434, 38 },
		{ 70, 435, 38 },
		{ 70, 436, 38 },
		{ 70, 437, 38 },
		{ 70, 438, 38 },
		{ 70, 439, 38 },
		{ 70, 440, 38 },
		{ 70, 441, 38 },
		{ 70, 442, 38 },
		{ 70, 443, 38 },
		{ 70, 444, 38 },
		{ 70, 445, 38 },
		{ 70, 446, 38 },
		{ 70, 447, 38 },
		{ 70, 448, 38 },
		{ 70, 449, 38 },
		{ 70, 450, 38 },
		{ 70, 451, 38 },
		{ 70, 452, 38 },
		{ 70, 453, 38 },
		{ 70, 454, 38 },
		{ 70, 455, 38 },
		{ 70, 456, 38 },
		{ 70, 457, 38 },
		{ 70, 458, 38 },
		{ 70, 459, 38 },
		{ 70, 460, 38 },
		{ 70, 461, 38 },
		{ 70, 462, 38 },
		{ 70, 463, 38 },
		{ 70, 464, 38 },
		{ 70, 465, 38 },
		{ 70, 466, 38 },
		{ 70, 467, 38 },
		{ 70, 468, 38 },
		{ 70, 469, 38 },
		{ 70, 470, 38 },
		{ 71, 420, 38 },
		{ 71, 421, 38 },
		{ 71, 422, 38 },
		{ 71, 423, 38 },
		{ 71, 424, 38 },
		{ 71, 425, 38 },
		{ 71, 426, 38 },
		{ 71, 427, 38 },
		{ 71, 428, 38 },
		{ 71, 429, 38 },
		{ 71, 430, 38 },
		{ 71, 431, 38 },
		{ 71, 432, 38 },
		{ 71, 433, 38 },
		{ 71, 434, 38 },
		{ 71, 435, 38 },
		{ 71, 436, 38 },
		{ 71, 437, 38 },
		{ 71, 438, 38 },
		{ 71, 439, 38 },
		{ 71, 440, 38 },
		{ 71, 441, 38 },
		{ 71, 442, 38 },
		{ 71, 443, 38 },
		{ 71, 444, 38 },
		{ 71, 445, 38 },
		{ 71, 446, 38 },
		{ 71, 447, 38 },
		{ 71, 448, 38 },
		{ 71, 449, 38 },
		{ 71, 450, 38 },
		{ 71, 451, 38 },
		{ 71, 452, 38 },
		{ 71, 453, 38 },
		{ 71, 454, 38 },
		{ 71, 455, 38 },
		{ 71, 456, 38 },
		{ 71, 457, 38 },
		{ 71, 458, 38 },
		{ 71, 459, 38 },
		{ 71, 460, 38 },
		{ 71, 461, 38 },
		{ 71, 462, 38 },
		{ 71, 463, 38 },
		{ 71, 464, 38 },
		{ 71, 465, 38 },
		{ 71, 466, 38 },
		{ 71, 467, 38 },
		{ 71, 468, 38 },
		{ 71, 469, 38 },
		{ 72, 419, 38 },
		{ 72, 420, 38 },
		{ 72, 421, 38 },
		{ 72, 422, 38 },
		{ 72, 423, 38 },
		{ 72, 424, 38 },
		{ 72, 425, 38 },
		{ 72, 426, 38 },
		{ 72, 427, 38 },
		{ 72, 428, 38 },
		{ 72, 429, 38 },
		{ 72, 430, 38 },
		{ 72, 431, 38 },
		{ 72, 432, 38 },
		{ 72, 433, 38 },
		{ 72, 434, 38 },
		{ 72, 435, 38 },
		{ 72, 436, 38 },
		{ 72, 437, 38 },
		{ 72, 438, 38 },
		{ 72, 439, 38 },
		{ 72, 440, 38 },
		{ 72, 441, 38 },
		{ 72, 442, 38 },
		{ 72, 443, 38 },
		{ 72, 444, 38 },
		{ 72, 445, 38 },
		{ 72, 446, 38 },
		{ 72, 447, 38 },
		{ 72, 448, 38 },
		{ 72, 449, 38 },
		{ 72, 450, 38 },
		{ 72, 451, 38 },
		{ 72, 452, 38 },
		{ 72, 453, 38 },
		{ 72, 454, 38 },
		{ 72, 455, 38 },
		{ 72, 456, 38 },
		{ 72, 457, 38 },
		{ 72, 458, 38 },
		{ 72, 459, 38 },
		{ 72, 460, 38 },
		{ 72, 461, 38 },
		{ 72, 462, 38 },
		{ 72, 463, 38 },
		{ 72, 464, 38 },
		{ 72, 465, 38 },
		{ 72, 466, 38 },
		{ 72, 467, 38 },
		{ 72, 468, 38 },
		{ 73, 419, 38 },
		{ 73, 420, 38 },
		{ 73, 421, 38 },
		{ 73, 422, 38 },
		{ 73, 423, 38 },
		{ 73, 424, 38 },
		{ 73, 425, 38 },
		{ 73, 426, 38 },
		{ 73, 427, 38 },
		{ 73, 428, 38 },
		{ 73, 429, 38 },
		{ 73, 430, 38 },
		{ 73, 431, 38 },
		{ 73, 432, 38 },
		{ 73, 433, 38 },
		{ 73, 434, 38 },
		{ 73, 435, 38 },
		{ 73, 436, 38 },
		{ 73, 437, 38 },
		{ 73, 438, 38 },
		{ 73, 439, 38 },
		{ 73, 440, 38 },
		{ 73, 441, 38 },
		{ 73, 442, 38 },
		{ 73, 443, 38 },
		{ 73, 444, 38 },
		{ 73, 445, 38 },
		{ 73, 446, 38 },
		{ 73, 447, 38 },
		{ 73, 448, 38 },
		{ 73, 449, 38 },
		{ 73, 450, 38 },
		{ 73, 451, 38 },
		{ 73, 452, 38 },
		{ 73, 453, 38 },
		{ 73, 454, 38 },
		{ 73, 455, 38 },
		{ 73, 456, 38 },
		{ 73, 457, 38 },
		{ 73, 458, 38 },
		{ 73, 459, 38 },
		{ 73, 460, 38 },
		{ 73, 461, 38 },
		{ 73, 462, 38 },
		{ 73, 463, 38 },
		{ 73, 464, 38 },
		{ 73, 465, 38 },
		{ 73, 466, 38 },
		{ 73, 467, 38 },
		{ 74, 418, 38 },
		{ 74, 419, 38 },
		{ 74, 420, 38 },
		{ 74, 421, 38 },
		{ 74, 422, 38 },
		{ 74, 423, 38 },
		{ 74, 424, 38 },
		{ 74, 425, 38 },
		{ 74, 426, 38 },
		{ 74, 427, 38 },
		{ 74, 428, 38 },
		{ 74, 429, 38 },
		{ 74, 430, 38 },
		{ 74, 431, 38 },
		{ 74, 432, 38 },
		{ 74, 433, 38 },
		{ 74, 434, 38 },
		{ 74, 435, 38 },
		{ 74, 436, 38 },
		{ 74, 437, 38 },
		{ 74, 438, 38 },
		{ 74, 439, 38 },
		{ 74, 440, 38 },
		{ 74, 441, 38 },
		{ 74, 442, 38 },
		{ 74, 443, 38 },
		{ 74, 444, 38 },
		{ 74, 445, 38 },
		{ 74, 446, 38 },
		{ 74, 447, 38 },
		{ 74, 448, 38 },
		{ 74, 449, 38 },
		{ 74, 450, 38 },
		{ 74, 451, 38 },
		{ 74, 452, 38 },
		{ 74, 453, 38 },
		{ 74, 454, 38 },
		{ 74, 455, 38 },
		{ 74, 456, 38 },
		{ 74, 457, 38 },
		{ 74, 458, 38 },
		{ 74, 459, 38 },
		{ 74, 460, 38 },
		{ 74, 461, 38 },
		{ 74, 462, 38 },
		{ 74, 463, 38 },
		{ 74, 464, 38 },
		{ 74, 465, 38 },
		{ 74, 466, 38 },
		{ 75, 417, 38 },
		{ 75, 418, 38 },
		{ 75, 419, 38 },
		{ 75, 420, 38 },
		{ 75, 421, 38 },
		{ 75, 422, 38 },
		{ 75, 423, 38 },
		{ 75, 424, 38 },
		{ 75, 425, 38 },
		{ 75, 426, 38 },
		{ 75, 427, 38 },
		{ 75, 428, 38 },
		{ 75, 429, 38 },
		{ 75, 430, 38 },
		{ 75, 431, 38 },
		{ 75, 432, 38 },
		{ 75, 433, 38 },
		{ 75, 434, 38 },
		{ 75, 435, 38 },
		{ 75, 436, 38 },
		{ 75, 437, 38 },
		{ 75, 438, 38 },
		{ 75, 439, 38 },
		{ 75, 440, 38 },
		{ 75, 441, 38 },
		{ 75, 442, 38 },
		{ 75, 443, 38 },
		{ 75, 444, 38 },
		{ 75, 445, 38 },
		{ 75, 446, 38 },
		{ 75, 447, 38 },
		{ 75, 448, 38 },
		{ 75, 449, 38 },
		{ 75, 450, 38 },
		{ 75, 451, 38 },
		{ 75, 452, 38 },
		{ 75, 453, 38 },
		{ 75, 454, 38 },
		{ 75, 455, 38 },
		{ 75, 456, 38 },
		{ 75, 457, 38 },
		{ 75, 458, 38 },
		{ 75, 459, 38 },
		{ 75, 460, 38 },
		{ 75, 461, 38 },
		{ 75, 462, 38 },
		{ 75, 463, 38 },
		{ 75, 464, 38 },
		{ 76, 416, 38 },
		{ 76, 417, 38 },
		{ 76, 418, 38 },
		{ 76, 419, 38 },
		{ 76, 420, 38 },
		{ 76, 421, 38 },
		{ 76, 422, 38 },
		{ 76, 423, 38 },
		{ 76, 424, 38 },
		{ 76, 425, 38 },
		{ 76, 426, 38 },
		{ 76, 427, 38 },
		{ 76, 428, 38 },
		{ 76, 429, 38 },
		{ 76, 430, 38 },
		{ 76, 431, 38 },
		{ 76, 432, 38 },
		{ 76, 433, 38 },
		{ 76, 434, 38 },
		{ 76, 435, 38 },
		{ 76, 436, 38 },
		{ 76, 437, 38 },
		{ 76, 438, 38 },
		{ 76, 439, 38 },
		{ 76, 440, 38 },
		{ 76, 441, 38 },
		{ 76, 442, 38 },
		{ 76, 443, 38 },
		{ 76, 444, 38 },
		{ 76, 445, 38 },
		{ 76, 446, 38 },
		{ 76, 447, 38 },
		{ 76, 448, 38 },
		{ 76, 449, 38 },
		{ 76, 450, 38 },
		{ 76, 451, 38 },
		{ 76, 452, 38 },
		{ 76, 453, 38 },
		{ 76, 454, 38 },
		{ 76, 455, 38 },
		{ 76, 456, 38 },
		{ 76, 457, 38 },
		{ 76, 458, 38 },
		{ 76, 459, 38 },
		{ 76, 460, 38 },
		{ 76, 461, 38 },
		{ 76, 462, 38 },
		{ 76, 463, 38 },
		{ 77, 415, 38 },
		{ 77, 416, 38 },
		{ 77, 417, 38 },
		{ 77, 418, 38 },
		{ 77, 419, 38 },
		{ 77, 420, 38 },
		{ 77, 421, 38 },
		{ 77, 422, 38 },
		{ 77, 423, 38 },
		{ 77, 424, 38 },
		{ 77, 425, 38 },
		{ 77, 426, 38 },
		{ 77, 427, 38 },
		{ 77, 428, 38 },
		{ 77, 429, 38 },
		{ 77, 430, 38 },
		{ 77, 431, 38 },
		{ 77, 432, 38 },
		{ 77, 433, 38 },
		{ 77, 434, 38 },
		{ 77, 435, 38 },
		{ 77, 436, 38 },
		{ 77, 437, 38 },
		{ 77, 438, 38 },
		{ 77, 439, 38 },
		{ 77, 440, 38 },
		{ 77, 441, 38 },
		{ 77, 442, 38 },
		{ 77, 443, 38 },
		{ 77, 444, 38 },
		{ 77, 445, 38 },
		{ 77, 446, 38 },
		{ 77, 447, 38 },
		{ 77, 448, 38 },
		{ 77, 449, 38 },
		{ 77, 450, 38 },
		{ 77, 451, 38 },
		{ 77, 452, 38 },
		{ 77, 453, 38 },
		{ 77, 454, 38 },
		{ 77, 455, 38 },
		{ 77, 456, 38 },
		{ 77, 457, 38 },
		{ 77, 458, 38 },
		{ 77, 459, 38 },
		{ 77, 460, 38 },
		{ 77, 461, 38 },
		{ 77, 462, 38 },
		{ 78, 414, 38 },
		{ 78, 415, 38 },
		{ 78, 416, 38 },
		{ 78, 417, 38 },
		{ 78, 418, 38 },
		{ 78, 419, 38 },
		{ 78, 420, 38 },
		{ 78, 421, 38 },
		{ 78, 422, 38 },
		{ 78, 423, 38 },
		{ 78, 424, 38 },
		{ 78, 425, 38 },
		{ 78, 426, 38 },
		{ 78, 427, 38 },
		{ 78, 428, 38 },
		{ 78, 429, 38 },
		{ 78, 430, 38 },
		{ 78, 431, 38 },
		{ 78, 432, 38 },
		{ 78, 433, 38 },
		{ 78, 434, 38 },
		{ 78, 435, 38 },
		{ 78, 436, 38 },
		{ 78, 437, 38 },
		{ 78, 438, 38 },
		{ 78, 439, 38 },
		{ 78, 440, 38 },
		{ 78, 441, 38 },
		{ 78, 442, 38 },
		{ 78, 443, 38 },
		{ 78, 444, 38 },
		{ 78, 445, 38 },
		{ 78, 446, 38 },
		{ 78, 447, 38 },
		{ 78, 448, 38 },
		{ 78, 449, 38 },
		{ 78, 450, 38 },
		{ 78, 451, 38 },
		{ 78, 452, 38 },
		{ 78, 453, 38 },
		{ 78, 454, 38 },
		{ 78, 455, 38 },
		{ 78, 456, 38 },
		{ 78, 457, 38 },
		{ 78, 458, 38 },
		{ 78, 459, 38 },
		{ 78, 460, 38 },
		{ 78, 461, 38 },
		{ 79, 413, 38 },
		{ 79, 414, 38 },
		{ 79, 415, 38 },
		{ 79, 416, 38 },
		{ 79, 417, 38 },
		{ 79, 418, 38 },
		{ 79, 419, 38 },
		{ 79, 420, 38 },
		{ 79, 421, 38 },
		{ 79, 422, 38 },
		{ 79, 423, 38 },
		{ 79, 424, 38 },
		{ 79, 425, 38 },
		{ 79, 426, 38 },
		{ 79, 427, 38 },
		{ 79, 428, 38 },
		{ 79, 429, 38 },
		{ 79, 430, 38 },
		{ 79, 431, 38 },
		{ 79, 432, 38 },
		{ 79, 433, 38 },
		{ 79, 434, 38 },
		{ 79, 435, 38 },
		{ 79, 436, 38 },
		{ 79, 437, 38 },
		{ 79, 438, 38 },
		{ 79, 439, 38 },
		{ 79, 440, 38 },
		{ 79, 441, 38 },
		{ 79, 442, 38 },
		{ 79, 443, 38 },
		{ 79, 444, 38 },
		{ 79, 445, 38 },
		{ 79, 446, 38 },
		{ 79, 447, 38 },
		{ 79, 448, 38 },
		{ 79, 449, 38 },
		{ 79, 450, 38 },
		{ 79, 451, 38 },
		{ 79, 452, 38 },
		{ 79, 453, 38 },
		{ 79, 454, 38 },
		{ 79, 455, 38 },
		{ 79, 456, 38 },
		{ 79, 457, 38 },
		{ 79, 458, 38 },
		{ 79, 459, 38 },
		{ 79, 460, 38 },
		{ 80, 412, 38 },
		{ 80, 413, 38 },
		{ 80, 414, 38 },
		{ 80, 415, 38 },
		{ 80, 416, 38 },
		{ 80, 417, 38 },
		{ 80, 418, 38 },
		{ 80, 419, 38 },
		{ 80, 420, 38 },
		{ 80, 421, 38 },
		{ 80, 422, 38 },
		{ 80, 423, 38 },
		{ 80, 424, 38 },
		{ 80, 425, 38 },
		{ 80, 426, 38 },
		{ 80, 427, 38 },
		{ 80, 428, 38 },
		{ 80, 429, 38 },
		{ 80, 430, 38 },
		{ 80, 431, 38 },
		{ 80, 432, 38 },
		{ 80, 433, 38 },
		{ 80, 434, 38 },
		{ 80, 435, 38 },
		{ 80, 436, 38 },
		{ 80, 437, 38 },
		{ 80, 438, 38 },
		{ 80, 439, 38 },
		{ 80, 440, 38 },
		{ 80, 441, 38 },
		{ 80, 442, 38 },
		{ 80, 443, 38 },
		{ 80, 444, 38 },
		{ 80, 445, 38 },
		{ 80, 446, 38 },
		{ 80, 447, 38 },
		{ 80, 448, 38 },
		{ 80, 449, 38 },
		{ 80, 450, 38 },
		{ 80, 451, 38 },
		{ 80, 452, 38 },
		{ 80, 453, 38 },
		{ 80, 454, 38 },
		{ 80, 455, 38 },
		{ 80, 456, 38 },
		{ 80, 457, 38 },
		{ 80, 458, 38 },
		{ 80, 459, 38 },
		{ 81, 412, 38 },
		{ 81, 413, 38 },
		{ 81, 414, 38 },
		{ 81, 415, 38 },
		{ 81, 416, 38 },
		{ 81, 417, 38 },
		{ 81, 418, 38 },
		{ 81, 419, 38 },
		{ 81, 420, 38 },
		{ 81, 421, 38 },
		{ 81, 422, 38 },
		{ 81, 423, 38 },
		{ 81, 424, 38 },
		{ 81, 425, 38 },
		{ 81, 426, 38 },
		{ 81, 427, 38 },
		{ 81, 428, 38 },
		{ 81, 429, 38 },
		{ 81, 430, 38 },
		{ 81, 431, 38 },
		{ 81, 432, 38 },
		{ 81, 433, 38 },
		{ 81, 434, 38 },
		{ 81, 435, 38 },
		{ 81, 436, 38 },
		{ 81, 437, 38 },
		{ 81, 438, 38 },
		{ 81, 439, 38 },
		{ 81, 440, 38 },
		{ 81, 441, 38 },
		{ 81, 442, 38 },
		{ 81, 443, 38 },
		{ 81, 444, 38 },
		{ 81, 445, 38 },
		{ 81, 446, 38 },
		{ 81, 447, 38 },
		{ 81, 448, 38 },
		{ 81, 449, 38 },
		{ 81, 450, 38 },
		{ 81, 451, 38 },
		{ 81, 452, 38 },
		{ 81, 453, 38 },
		{ 81, 454, 38 },
		{ 81, 455, 38 },
		{ 81, 456, 38 },
		{ 81, 457, 38 },
		{ 81, 458, 38 },
		{ 82, 411, 38 },
		{ 82, 412, 38 },
		{ 82, 413, 38 },
		{ 82, 414, 38 },
		{ 82, 415, 38 },
		{ 82, 416, 38 },
		{ 82, 417, 38 },
		{ 82, 418, 38 },
		{ 82, 419, 38 },
		{ 82, 420, 38 },
		{ 82, 421, 38 },
		{ 82, 422, 38 },
		{ 82, 423, 38 },
		{ 82, 424, 38 },
		{ 82, 425, 38 },
		{ 82, 426, 38 },
		{ 82, 427, 38 },
		{ 82, 428, 38 },
		{ 82, 429, 38 },
		{ 82, 430, 38 },
		{ 82, 431, 38 },
		{ 82, 432, 38 },
		{ 82, 433, 38 },
		{ 82, 434, 38 },
		{ 82, 435, 38 },
		{ 82, 436, 38 },
		{ 82, 437, 38 },
		{ 82, 438, 38 },
		{ 82, 439, 38 },
		{ 82, 440, 38 },
		{ 82, 441, 38 },
		{ 82, 442, 38 },
		{ 82, 443, 38 },
		{ 82, 444, 38 },
		{ 82, 445, 38 },
		{ 82, 446, 38 },
		{ 82, 447, 38 },
		{ 82, 448, 38 },
		{ 82, 449, 38 },
		{ 82, 450, 38 },
		{ 82, 451, 38 },
		{ 82, 452, 38 },
		{ 82, 453, 38 },
		{ 82, 454, 38 },
		{ 82, 455, 38 },
		{ 82, 456, 38 },
		{ 83, 410, 38 },
		{ 83, 411, 38 },
		{ 83, 412, 38 },
		{ 83, 413, 38 },
		{ 83, 414, 38 },
		{ 83, 415, 38 },
		{ 83, 416, 38 },
		{ 83, 417, 38 },
		{ 83, 418, 38 },
		{ 83, 419, 38 },
		{ 83, 420, 38 },
		{ 83, 421, 38 },
		{ 83, 422, 38 },
		{ 83, 423, 38 },
		{ 83, 424, 38 },
		{ 83, 425, 38 },
		{ 83, 426, 38 },
		{ 83, 427, 38 },
		{ 83, 428, 38 },
		{ 83, 429, 38 },
		{ 83, 430, 38 },
		{ 83, 431, 38 },
		{ 83, 432, 38 },
		{ 83, 433, 38 },
		{ 83, 434, 38 },
		{ 83, 435, 38 },
		{ 83, 436, 38 },
		{ 83, 437, 38 },
		{ 83, 438, 38 },
		{ 83, 439, 38 },
		{ 83, 440, 38 },
		{ 83, 441, 38 },
		{ 83, 442, 38 },
		{ 83, 443, 38 },
		{ 83, 444, 38 },
		{ 83, 445, 38 },
		{ 83, 446, 38 },
		{ 83, 447, 38 },
		{ 83, 448, 38 },
		{ 83, 449, 38 },
		{ 83, 450, 38 },
		{ 83, 451, 38 },
		{ 83, 452, 38 },
		{ 83, 453, 38 },
		{ 83, 454, 38 },
		{ 83, 455, 38 },
		{ 84, 409, 38 },
		{ 84, 410, 38 },
		{ 84, 411, 38 },
		{ 84, 412, 38 },
		{ 84, 413, 38 },
		{ 84, 414, 38 },
		{ 84, 415, 38 },
		{ 84, 416, 38 },
		{ 84, 417, 38 },
		{ 84, 418, 38 },
		{ 84, 419, 38 },
		{ 84, 420, 38 },
		{ 84, 421, 38 },
		{ 84, 422, 38 },
		{ 84, 423, 38 },
		{ 84, 424, 38 },
		{ 84, 425, 38 },
		{ 84, 426, 38 },
		{ 84, 427, 38 },
		{ 84, 428, 38 },
		{ 84, 429, 38 },
		{ 84, 430, 38 },
		{ 84, 431, 38 },
		{ 84, 432, 38 },
		{ 84, 433, 38 },
		{ 84, 434, 38 },
		{ 84, 435, 38 },
		{ 84, 436, 38 },
		{ 84, 437, 38 },
		{ 84, 438, 38 },
		{ 84, 439, 38 },
		{ 84, 440, 38 },
		{ 84, 441, 38 },
		{ 84, 442, 38 },
		{ 84, 443, 38 },
		{ 84, 444, 38 },
		{ 84, 445, 38 },
		{ 84, 446, 38 },
		{ 84, 447, 38 },
		{ 84, 448, 38 },
		{ 84, 449, 38 },
		{ 84, 450, 38 },
		{ 84, 451, 38 },
		{ 84, 452, 38 },
		{ 84, 453, 38 },
		{ 84, 454, 38 },
		{ 85, 408, 38 },
		{ 85, 409, 38 },
		{ 85, 410, 38 },
		{ 85, 411, 38 },
		{ 85, 412, 38 },
		{ 85, 413, 38 },
		{ 85, 414, 38 },
		{ 85, 415, 38 },
		{ 85, 416, 38 },
		{ 85, 417, 38 },
		{ 85, 418, 38 },
		{ 85, 419, 38 },
		{ 85, 420, 38 },
		{ 85, 421, 38 },
		{ 85, 422, 38 },
		{ 85, 423, 38 },
		{ 85, 424, 38 },
		{ 85, 425, 38 },
		{ 85, 426, 38 },
		{ 85, 427, 38 },
		{ 85, 428, 38 },
		{ 85, 429, 38 },
		{ 85, 430, 38 },
		{ 85, 431, 38 },
		{ 85, 432, 38 },
		{ 85, 433, 38 },
		{ 85, 434, 38 },
		{ 85, 435, 38 },
		{ 85, 436, 38 },
		{ 85, 437, 38 },
		{ 85, 438, 38 },
		{ 85, 439, 38 },
		{ 85, 440, 38 },
		{ 85, 441, 38 },
		{ 85, 442, 38 },
		{ 85, 443, 38 },
		{ 85, 444, 38 },
		{ 85, 445, 38 },
		{ 85, 446, 38 },
		{ 85, 447, 38 },
		{ 85, 448, 38 },
		{ 85, 449, 38 },
		{ 85, 450, 38 },
		{ 85, 451, 38 },
		{ 85, 452, 38 },
		{ 85, 453, 38 },
		{ 86, 407, 38 },
		{ 86, 408, 38 },
		{ 86, 409, 38 },
		{ 86, 410, 38 },
		{ 86, 411, 38 },
		{ 86, 412, 38 },
		{ 86, 413, 38 },
		{ 86, 414, 38 },
		{ 86, 415, 38 },
		{ 86, 416, 38 },
		{ 86, 417, 38 },
		{ 86, 418, 38 },
		{ 86, 419, 38 },
		{ 86, 420, 38 },
		{ 86, 421, 38 },
		{ 86, 422, 38 },
		{ 86, 423, 38 },
		{ 86, 424, 38 },
		{ 86, 425, 38 },
		{ 86, 426, 38 },
		{ 86, 427, 38 },
		{ 86, 428, 38 },
		{ 86, 429, 38 },
		{ 86, 430, 38 },
		{ 86, 431, 38 },
		{ 86, 432, 38 },
		{ 86, 433, 38 },
		{ 86, 434, 38 },
		{ 86, 435, 38 },
		{ 86, 436, 38 },
		{ 86, 437, 38 },
		{ 86, 438, 38 },
		{ 86, 439, 38 },
		{ 86, 440, 38 },
		{ 86, 441, 38 },
		{ 86, 442, 38 },
		{ 86, 443, 38 },
		{ 86, 444, 38 },
		{ 86, 445, 38 },
		{ 86, 446, 38 },
		{ 86, 447, 38 },
		{ 86, 448, 38 },
		{ 86, 449, 38 },
		{ 86, 450, 38 },
		{ 86, 451, 38 },
		{ 86, 452, 38 },
		{ 87, 406, 38 },
		{ 87, 407, 38 },
		{ 87, 408, 38 },
		{ 87, 409, 38 },
		{ 87, 410, 38 },
		{ 87, 411, 38 },
		{ 87, 412, 38 },
		{ 87, 413, 38 },
		{ 87, 414, 38 },
		{ 87, 415, 38 },
		{ 87, 416, 38 },
		{ 87, 417, 38 },
		{ 87, 418, 38 },
		{ 87, 419, 38 },
		{ 87, 420, 38 },
		{ 87, 421, 38 },
		{ 87, 422, 38 },
		{ 87, 423, 38 },
		{ 87, 424, 38 },
		{ 87, 425, 38 },
		{ 87, 426, 38 },
		{ 87, 427, 38 },
		{ 87, 428, 38 },
		{ 87, 429, 38 },
		{ 87, 430, 38 },
		{ 87, 431, 38 },
		{ 87, 432, 38 },
		{ 87, 433, 38 },
		{ 87, 434, 38 },
		{ 87, 435, 38 },
		{ 87, 436, 38 },
		{ 87, 437, 38 },
		{ 87, 438, 38 },
		{ 87, 439, 38 },
		{ 87, 440, 38 },
		{ 87, 441, 38 },
		{ 87, 442, 38 },
		{ 87, 443, 38 },
		{ 87, 444, 38 },
		{ 87, 445, 38 },
		{ 87, 446, 38 },
		{ 87, 447, 38 },
		{ 87, 448, 38 },
		{ 87, 449, 38 },
		{ 87, 450, 38 },
		{ 87, 451, 38 },
		{ 88, 405, 38 },
		{ 88, 406, 38 },
		{ 88, 407, 38 },
		{ 88, 408, 38 },
		{ 88, 409, 38 },
		{ 88, 410, 38 },
		{ 88, 411, 38 },
		{ 88, 412, 38 },
		{ 88, 413, 38 },
		{ 88, 414, 38 },
		{ 88, 415, 38 },
		{ 88, 416, 38 },
		{ 88, 417, 38 },
		{ 88, 418, 38 },
		{ 88, 419, 38 },
		{ 88, 420, 38 },
		{ 88, 421, 38 },
		{ 88, 422, 38 },
		{ 88, 423, 38 },
		{ 88, 424, 38 },
		{ 88, 425, 38 },
		{ 88, 426, 38 },
		{ 88, 427, 38 },
		{ 88, 428, 38 },
		{ 88, 429, 38 },
		{ 88, 430, 38 },
		{ 88, 431, 38 },
		{ 88, 432, 38 },
		{ 88, 433, 38 },
		{ 88, 434, 38 },
		{ 88, 435, 38 },
		{ 88, 436, 38 },
		{ 88, 437, 38 },
		{ 88, 438, 38 },
		{ 88, 439, 38 },
		{ 88, 440, 38 },
		{ 88, 441, 38 },
		{ 88, 442, 38 },
		{ 88, 443, 38 },
		{ 88, 444, 38 },
		{ 88, 445, 38 },
		{ 88, 446, 38 },
		{ 88, 447, 38 },
		{ 88, 448, 38 },
		{ 88, 449, 38 },
		{ 88, 450, 38 },
		{ 89, 405, 38 },
		{ 89, 406, 38 },
		{ 89, 407, 38 },
		{ 89, 408, 38 },
		{ 89, 409, 38 },
		{ 89, 410, 38 },
		{ 89, 411, 38 },
		{ 89, 412, 38 },
		{ 89, 413, 38 },
		{ 89, 414, 38 },
		{ 89, 415, 38 },
		{ 89, 416, 38 },
		{ 89, 417, 38 },
		{ 89, 418, 38 },
		{ 89, 419, 38 },
		{ 89, 420, 38 },
		{ 89, 421, 38 },
		{ 89, 422, 38 },
		{ 89, 423, 38 },
		{ 89, 424, 38 },
		{ 89, 425, 38 },
		{ 89, 426, 38 },
		{ 89, 427, 38 },
		{ 89, 428, 38 },
		{ 89, 429, 38 },
		{ 89, 430, 38 },
		{ 89, 431, 38 },
		{ 89, 432, 38 },
		{ 89, 433, 38 },
		{ 89, 434, 38 },
		{ 89, 435, 38 },
		{ 89, 436, 38 },
		{ 89, 437, 38 },
		{ 89, 438, 38 },
		{ 89, 439, 38 },
		{ 89, 440, 38 },
		{ 89, 441, 38 },
		{ 89, 442, 38 },
		{ 89, 443, 38 },
		{ 89, 444, 38 },
		{ 89, 445, 38 },
		{ 89, 446, 38 },
		{ 89, 447, 38 },
		{ 89, 448, 38 },
		{ 90, 404, 38 },
		{ 90, 405, 38 },
		{ 90, 406, 38 },
		{ 90, 407, 38 },
		{ 90, 408, 38 },
		{ 90, 409, 38 },
		{ 90, 410, 38 },
		{ 90, 411, 38 },
		{ 90, 412, 38 },
		{ 90, 413, 38 },
		{ 90, 414, 38 },
		{ 90, 415, 38 },
		{ 90, 416, 38 },
		{ 90, 417, 38 },
		{ 90, 418, 38 },
		{ 90, 419, 38 },
		{ 90, 420, 38 },
		{ 90, 421, 38 },
		{ 90, 422, 38 },
		{ 90, 423, 38 },
		{ 90, 424, 38 },
		{ 90, 425, 38 },
		{ 90, 426, 38 },
		{ 90, 427, 38 },
		{ 90, 428, 38 },
		{ 90, 429, 38 },
		{ 90, 430, 38 },
		{ 90, 431, 38 },
		{ 90, 432, 38 },
		{ 90, 433, 38 },
		{ 90, 434, 38 },
		{ 90, 435, 38 },
		{ 90, 436, 38 },
		{ 90, 437, 38 },
		{ 90, 438, 38 },
		{ 90, 439, 38 },
		{ 90, 440, 38 },
		{ 90, 441, 38 },
		{ 90, 442, 38 },
		{ 90, 443, 38 },
		{ 90, 444, 38 },
		{ 90, 445, 38 },
		{ 90, 446, 38 },
		{ 90, 447, 38 },
		{ 91, 403, 38 },
		{ 91, 404, 38 },
		{ 91, 405, 38 },
		{ 91, 406, 38 },
		{ 91, 407, 38 },
		{ 91, 408, 38 },
		{ 91, 409, 38 },
		{ 91, 410, 38 },
		{ 91, 411, 38 },
		{ 91, 412, 38 },
		{ 91, 413, 38 },
		{ 91, 414, 38 },
		{ 91, 415, 38 },
		{ 91, 416, 38 },
		{ 91, 417, 38 },
		{ 91, 418, 38 },
		{ 91, 419, 38 },
		{ 91, 420, 38 },
		{ 91, 421, 38 },
		{ 91, 422, 38 },
		{ 91, 423, 38 },
		{ 91, 424, 38 },
		{ 91, 425, 38 },
		{ 91, 426, 38 },
		{ 91, 427, 38 },
		{ 91, 428, 38 },
		{ 91, 429, 38 },
		{ 91, 430, 38 },
		{ 91, 431, 38 },
		{ 91, 432, 38 },
		{ 91, 433, 38 },
		{ 91, 434, 38 },
		{ 91, 435, 38 },
		{ 91, 436, 38 },
		{ 91, 437, 38 },
		{ 91, 438, 38 },
		{ 91, 439, 38 },
		{ 91, 440, 38 },
		{ 91, 441, 38 },
		{ 91, 442, 38 },
		{ 91, 443, 38 },
		{ 91, 444, 38 },
		{ 91, 445, 38 },
		{ 91, 446, 38 },
		{ 92, 402, 38 },
		{ 92, 403, 38 },
		{ 92, 404, 38 },
		{ 92, 405, 38 },
		{ 92, 406, 38 },
		{ 92, 407, 38 },
		{ 92, 408, 38 },
		{ 92, 409, 38 },
		{ 92, 410, 38 },
		{ 92, 411, 38 },
		{ 92, 412, 38 },
		{ 92, 413, 38 },
		{ 92, 414, 38 },
		{ 92, 415, 38 },
		{ 92, 416, 38 },
		{ 92, 417, 38 },
		{ 92, 418, 38 },
		{ 92, 419, 38 },
		{ 92, 420, 38 },
		{ 92, 421, 38 },
		{ 92, 422, 38 },
		{ 92, 423, 38 },
		{ 92, 424, 38 },
		{ 92, 425, 38 },
		{ 92, 426, 38 },
		{ 92, 427, 38 },
		{ 92, 428, 38 },
		{ 92, 429, 38 },
		{ 92, 430, 38 },
		{ 92, 431, 38 },
		{ 92, 432, 38 },
		{ 92, 433, 38 },
		{ 92, 434, 38 },
		{ 92, 435, 38 },
		{ 92, 436, 38 },
		{ 92, 437, 38 },
		{ 92, 438, 38 },
		{ 92, 439, 38 },
		{ 92, 440, 38 },
		{ 92, 441, 38 },
		{ 92, 442, 38 },
		{ 92, 443, 38 },
		{ 92, 444, 38 },
		{ 92, 445, 38 },
		{ 93, 401, 38 },
		{ 93, 402, 38 },
		{ 93, 403, 38 },
		{ 93, 404, 38 },
		{ 93, 405, 38 },
		{ 93, 406, 38 },
		{ 93, 407, 38 },
		{ 93, 408, 38 },
		{ 93, 409, 38 },
		{ 93, 410, 38 },
		{ 93, 411, 38 },
		{ 93, 412, 38 },
		{ 93, 413, 38 },
		{ 93, 414, 38 },
		{ 93, 415, 38 },
		{ 93, 416, 38 },
		{ 93, 417, 38 },
		{ 93, 418, 38 },
		{ 93, 419, 38 },
		{ 93, 420, 38 },
		{ 93, 421, 38 },
		{ 93, 422, 38 },
		{ 93, 423, 38 },
		{ 93, 424, 38 },
		{ 93, 425, 38 },
		{ 93, 426, 38 },
		{ 93, 427, 38 },
		{ 93, 428, 38 },
		{ 93, 429, 38 },
		{ 93, 430, 38 },
		{ 93, 431, 38 },
		{ 93, 432, 38 },
		{ 93, 433, 38 },
		{ 93, 434, 38 },
		{ 93, 435, 38 },
		{ 93, 436, 38 },
		{ 93, 437, 38 },
		{ 93, 438, 38 },
		{ 93, 439, 38 },
		{ 93, 440, 38 },
		{ 93, 441, 38 },
		{ 93, 442, 38 },
		{ 93, 443, 38 },
		{ 93, 444, 38 },
		{ 94, 400, 38 },
		{ 94, 401, 38 },
		{ 94, 402, 38 },
		{ 94, 403, 38 },
		{ 94, 404, 38 },
		{ 94, 405, 38 },
		{ 94, 406, 38 },
		{ 94, 407, 38 },
		{ 94, 408, 38 },
		{ 94, 409, 38 },
		{ 94, 410, 38 },
		{ 94, 411, 38 },
		{ 94, 412, 38 },
		{ 94, 413, 38 },
		{ 94, 414, 38 },
		{ 94, 415, 38 },
		{ 94, 416, 38 },
		{ 94, 417, 38 },
		{ 94, 418, 38 },
		{ 94, 419, 38 },
		{ 94, 420, 38 },
		{ 94, 421, 38 },
		{ 94, 422, 38 },
		{ 94, 423, 38 },
		{ 94, 424, 38 },
		{ 94, 425, 38 },
		{ 94, 426, 38 },
		{ 94, 427, 38 },
		{ 94, 428, 38 },
		{ 94, 429, 38 },
		{ 94, 430, 38 },
		{ 94, 431, 38 },
		{ 94, 432, 38 },
		{ 94, 433, 38 },
		{ 94, 434, 38 },
		{ 94, 435, 38 },
		{ 94, 436, 38 },
		{ 94, 437, 38 },
		{ 94, 438, 38 },
		{ 94, 439, 38 },
		{ 94, 440, 38 },
		{ 94, 441, 38 },
		{ 94, 442, 38 },
		{ 94, 443, 38 },
		{ 95, 399, 38 },
		{ 95, 400, 38 },
		{ 95, 401, 38 },
		{ 95, 402, 38 },
		{ 95, 403, 38 },
		{ 95, 404, 38 },
		{ 95, 405, 38 },
		{ 95, 406, 38 },
		{ 95, 407, 38 },
		{ 95, 408, 38 },
		{ 95, 409, 38 },
		{ 95, 410, 38 },
		{ 95, 411, 38 },
		{ 95, 412, 38 },
		{ 95, 413, 38 },
		{ 95, 414, 38 },
		{ 95, 415, 38 },
		{ 95, 416, 38 },
		{ 95, 417, 38 },
		{ 95, 418, 38 },
		{ 95, 419, 38 },
		{ 95, 420, 38 },
		{ 95, 421, 38 },
		{ 95, 422, 38 },
		{ 95, 423, 38 },
		{ 95, 424, 38 },
		{ 95, 425, 38 },
		{ 95, 426, 38 },
		{ 95, 427, 38 },
		{ 95, 428, 38 },
		{ 95, 429, 38 },
		{ 95, 430, 38 },
		{ 95, 431, 38 },
		{ 95, 432, 38 },
		{ 95, 433, 38 },
		{ 95, 434, 38 },
		{ 95, 435, 38 },
		{ 95, 436, 38 },
		{ 95, 437, 38 },
		{ 95, 438, 38 },
		{ 95, 439, 38 },
		{ 95, 440, 38 },
		{ 95, 441, 38 },
		{ 95, 442, 38 },
		{ 96, 398, 38 },
		{ 96, 399, 38 },
		{ 96, 400, 38 },
		{ 96, 401, 38 },
		{ 96, 402, 38 },
		{ 96, 403, 38 },
		{ 96, 404, 38 },
		{ 96, 405, 38 },
		{ 96, 406, 38 },
		{ 96, 407, 38 },
		{ 96, 408, 38 },
		{ 96, 409, 38 },
		{ 96, 410, 38 },
		{ 96, 411, 38 },
		{ 96, 412, 38 },
		{ 96, 413, 38 },
		{ 96, 414, 38 },
		{ 96, 415, 38 },
		{ 96, 416, 38 },
		{ 96, 417, 38 },
		{ 96, 418, 38 },
		{ 96, 419, 38 },
		{ 96, 420, 38 },
		{ 96, 421, 38 },
		{ 96, 422, 38 },
		{ 96, 423, 38 },
		{ 96, 424, 38 },
		{ 96, 425, 38 },
		{ 96, 426, 38 },
		{ 96, 427, 38 },
		{ 96, 428, 38 },
		{ 96, 429, 38 },
		{ 96, 430, 38 },
		{ 96, 431, 38 },
		{ 96, 432, 38 },
		{ 96, 433, 38 },
		{ 96, 434, 38 },
		{ 96, 435, 38 },
		{ 96, 436, 38 },
		{ 96, 437, 38 },
		{ 96, 438, 38 },
		{ 96, 439, 38 },
		{ 96, 440, 38 },
		{ 97, 398, 38 },
		{ 97, 399, 38 },
		{ 97, 400, 38 },
		{ 97, 401, 38 },
		{ 97, 402, 38 },
		{ 97, 403, 38 },
		{ 97, 404, 38 },
		{ 97, 405, 38 },
		{ 97, 406, 38 },
		{ 97, 407, 38 },
		{ 97, 408, 38 },
		{ 97, 409, 38 },
		{ 97, 410, 38 },
		{ 97, 411, 38 },
		{ 97, 412, 38 },
		{ 97, 413, 38 },
		{ 97, 414, 38 },
		{ 97, 415, 38 },
		{ 97, 416, 38 },
		{ 97, 417, 38 },
		{ 97, 418, 38 },
		{ 97, 419, 38 },
		{ 97, 420, 38 },
		{ 97, 421, 38 },
		{ 97, 422, 38 },
		{ 97, 423, 38 },
		{ 97, 424, 38 },
		{ 97, 425, 38 },
		{ 97, 426, 38 },
		{ 97, 427, 38 },
		{ 97, 428, 38 },
		{ 97, 429, 38 },
		{ 97, 430, 38 },
		{ 97, 431, 38 },
		{ 97, 432, 38 },
		{ 97, 433, 38 },
		{ 97, 434, 38 },
		{ 97, 435, 38 },
		{ 97, 436, 38 },
		{ 97, 437, 38 },
		{ 97, 438, 38 },
		{ 97, 439, 38 },
		{ 98, 397, 38 },
		{ 98, 398, 38 },
		{ 98, 399, 38 },
		{ 98, 400, 38 },
		{ 98, 401, 38 },
		{ 98, 402, 38 },
		{ 98, 403, 38 },
		{ 98, 404, 38 },
		{ 98, 405, 38 },
		{ 98, 406, 38 },
		{ 98, 407, 38 },
		{ 98, 408, 38 },
		{ 98, 409, 38 },
		{ 98, 410, 38 },
		{ 98, 411, 38 },
		{ 98, 412, 38 },
		{ 98, 413, 38 },
		{ 98, 414, 38 },
		{ 98, 415, 38 },
		{ 98, 416, 38 },
		{ 98, 417, 38 },
		{ 98, 418, 38 },
		{ 98, 419, 38 },
		{ 98, 420, 38 },
		{ 98, 421, 38 },
		{ 98, 422, 38 },
		{ 98, 423, 38 },
		{ 98, 424, 38 },
		{ 98, 425, 38 },
		{ 98, 426, 38 },
		{ 98, 427, 38 },
		{ 98, 428, 38 },
		{ 98, 429, 38 },
		{ 98, 430, 38 },
		{ 98, 431, 38 },
		{ 98, 432, 38 },
		{ 98, 433, 38 },
		{ 98, 434, 38 },
		{ 98, 435, 38 },
		{ 98, 436, 38 },
		{ 98, 437, 38 },
		{ 98, 438, 38 },
		{ 99, 396, 38 },
		{ 99, 397, 38 },
		{ 99, 398, 38 },
		{ 99, 399, 38 },
		{ 99, 400, 38 },
		{ 99, 401, 38 },
		{ 99, 402, 38 },
		{ 99, 403, 38 },
		{ 99, 404, 38 },
		{ 99, 405, 38 },
		{ 99, 406, 38 },
		{ 99, 407, 38 },
		{ 99, 408, 38 },
		{ 99, 409, 38 },
		{ 99, 410, 38 },
		{ 99, 411, 38 },
		{ 99, 412, 38 },
		{ 99, 413, 38 },
		{ 99, 414, 38 },
		{ 99, 415, 38 },
		{ 99, 416, 38 },
		{ 99, 417, 38 },
		{ 99, 418, 38 },
		{ 99, 419, 38 },
		{ 99, 420, 38 },
		{ 99, 421, 38 },
		{ 99, 422, 38 },
		{ 99, 423, 38 },
		{ 99, 424, 38 },
		{ 99, 425, 38 },
		{ 99, 426, 38 },
		{ 99, 427, 38 },
		{ 99, 428, 38 },
		{ 99, 429, 38 },
		{ 99, 430, 38 },
		{ 99, 431, 38 },
		{ 99, 432, 38 },
		{ 99, 433, 38 },
		{ 99, 434, 38 },
		{ 99, 435, 38 },
		{ 99, 436, 38 },
		{ 99, 437, 38 },
		{ 100, 395, 38 },
		{ 100, 396, 38 },
		{ 100, 397, 38 },
		{ 100, 398, 38 },
		{ 100, 399, 38 },
		{ 100, 400, 38 },
		{ 100, 401, 38 },
		{ 100, 402, 38 },
		{ 100, 403, 38 },
		{ 100, 404, 38 },
		{ 100, 405, 38 },
		{ 100, 406, 38 },
		{ 100, 407, 38 },
		{ 100, 408, 38 },
		{ 100, 409, 38 },
		{ 100, 410, 38 },
		{ 100, 411, 38 },
		{ 100, 412, 38 },
		{ 100, 413, 38 },
		{ 100, 414, 38 },
		{ 100, 415, 38 },
		{ 100, 416, 38 },
		{ 100, 417, 38 },
		{ 100, 418, 38 },
		{ 100, 419, 38 },
		{ 100, 420, 38 },
		{ 100, 421, 38 },
		{ 100, 422, 38 },
		{ 100, 423, 38 },
		{ 100, 424, 38 },
		{ 100, 425, 38 },
		{ 100, 426, 38 },
		{ 100, 427, 38 },
		{ 100, 428, 38 },
		{ 100, 429, 38 },
		{ 100, 430, 38 },
		{ 100, 431, 38 },
		{ 100, 432, 38 },
		{ 100, 433, 38 },
		{ 100, 434, 38 },
		{ 100, 435, 38 },
		{ 100, 436, 38 },
		{ 101, 394, 38 },
		{ 101, 395, 38 },
		{ 101, 396, 38 },
		{ 101, 397, 38 },
		{ 101, 398, 38 },
		{ 101, 399, 38 },
		{ 101, 400, 38 },
		{ 101, 401, 38 },
		{ 101, 402, 38 },
		{ 101, 403, 38 },
		{ 101, 404, 38 },
		{ 101, 405, 38 },
		{ 101, 406, 38 },
		{ 101, 407, 38 },
		{ 101, 408, 38 },
		{ 101, 409, 38 },
		{ 101, 410, 38 },
		{ 101, 411, 38 },
		{ 101, 412, 38 },
		{ 101, 413, 38 },
		{ 101, 414, 38 },
		{ 101, 415, 38 },
		{ 101, 416, 38 },
		{ 101, 417, 38 },
		{ 101, 418, 38 },
		{ 101, 419, 38 },
		{ 101, 420, 38 },
		{ 101, 421, 38 },
		{ 101, 422, 38 },
		{ 101, 423, 38 },
		{ 101, 424, 38 },
		{ 101, 425, 38 },
		{ 101, 426, 38 },
		{ 101, 427, 38 },
		{ 101, 428, 38 },
		{ 101, 429, 38 },
		{ 101, 430, 38 },
		{ 101, 431, 38 },
		{ 101, 432, 38 },
		{ 101, 433, 38 },
		{ 101, 434, 38 },
		{ 101, 435, 38 },
		{ 102, 393, 38 },
		{ 102, 394, 38 },
		{ 102, 395, 38 },
		{ 102, 396, 38 },
		{ 102, 397, 38 },
		{ 102, 398, 38 },
		{ 102, 399, 38 },
		{ 102, 400, 38 },
		{ 102, 401, 38 },
		{ 102, 402, 38 },
		{ 102, 403, 38 },
		{ 102, 404, 38 },
		{ 102, 405, 38 },
		{ 102, 406, 38 },
		{ 102, 407, 38 },
		{ 102, 408, 38 },
		{ 102, 409, 38 },
		{ 102, 410, 38 },
		{ 102, 411, 38 },
		{ 102, 412, 38 },
		{ 102, 413, 38 },
		{ 102, 414, 38 },
		{ 102, 415, 38 },
		{ 102, 416, 38 },
		{ 102, 417, 38 },
		{ 102, 418, 38 },
		{ 102, 419, 38 },
		{ 102, 420, 38 },
		{ 102, 421, 38 },
		{ 102, 422, 38 },
		{ 102, 423, 38 },
		{ 102, 424, 38 },
		{ 102, 425, 38 },
		{ 102, 426, 38 },
		{ 102, 427, 38 },
		{ 102, 428, 38 },
		{ 102, 429, 38 },
		{ 102, 430, 38 },
		{ 102, 431, 38 },
		{ 102, 432, 38 },
		{ 102, 433, 38 },
		{ 102, 434, 38 },
		{ 103, 392, 38 },
		{ 103, 393, 38 },
		{ 103, 394, 38 },
		{ 103, 395, 38 },
		{ 103, 396, 38 },
		{ 103, 397, 38 },
		{ 103, 398, 38 },
		{ 103, 399, 38 },
		{ 103, 400, 38 },
		{ 103, 401, 38 },
		{ 103, 402, 38 },
		{ 103, 403, 38 },
		{ 103, 404, 38 },
		{ 103, 405, 38 },
		{ 103, 406, 38 },
		{ 103, 407, 38 },
		{ 103, 408, 38 },
		{ 103, 409, 38 },
		{ 103, 410, 38 },
		{ 103, 411, 38 },
		{ 103, 412, 38 },
		{ 103, 413, 38 },
		{ 103, 414, 38 },
		{ 103, 415, 38 },
		{ 103, 416, 38 },
		{ 103, 417, 38 },
		{ 103, 418, 38 },
		{ 103, 419, 38 },
		{ 103, 420, 38 },
		{ 103, 421, 38 },
		{ 103, 422, 38 },
		{ 103, 423, 38 },
		{ 103, 424, 38 },
		{ 103, 425, 38 },
		{ 103, 426, 38 },
		{ 103, 427, 38 },
		{ 103, 428, 38 },
		{ 103, 429, 38 },
		{ 103, 430, 38 },
		{ 103, 431, 38 },
		{ 103, 432, 38 },
		{ 104, 391, 38 },
		{ 104, 392, 38 },
		{ 104, 393, 38 },
		{ 104, 394, 38 },
		{ 104, 395, 38 },
		{ 104, 396, 38 },
		{ 104, 397, 38 },
		{ 104, 398, 38 },
		{ 104, 399, 38 },
		{ 104, 400, 38 },
		{ 104, 401, 38 },
		{ 104, 402, 38 },
		{ 104, 403, 38 },
		{ 104, 404, 38 },
		{ 104, 405, 38 },
		{ 104, 406, 38 },
		{ 104, 407, 38 },
		{ 104, 408, 38 },
		{ 104, 409, 38 },
		{ 104, 410, 38 },
		{ 104, 411, 38 },
		{ 104, 412, 38 },
		{ 104, 413, 38 },
		{ 104, 414, 38 },
		{ 104, 415, 38 },
		{ 104, 416, 38 },
		{ 104, 417, 38 },
		{ 104, 418, 38 },
		{ 104, 419, 38 },
		{ 104, 420, 38 },
		{ 104, 421, 38 },
		{ 104, 422, 38 },
		{ 104, 423, 38 },
		{ 104, 424, 38 },
		{ 104, 425, 38 },
		{ 104, 426, 38 },
		{ 104, 427, 38 },
		{ 104, 428, 38 },
		{ 104, 429, 38 },
		{ 104, 430, 38 },
		{ 104, 431, 38 },
		{ 105, 391, 38 },
		{ 105, 392, 38 },
		{ 105, 393, 38 },
		{ 105, 394, 38 },
		{ 105, 395, 38 },
		{ 105, 396, 38 },
		{ 105, 397, 38 },
		{ 105, 398, 38 },
		{ 105, 399, 38 },
		{ 105, 400, 38 },
		{ 105, 401, 38 },
		{ 105, 402, 38 },
		{ 105, 403, 38 },
		{ 105, 404, 38 },
		{ 105, 405, 38 },
		{ 105, 406, 38 },
		{ 105, 407, 38 },
		{ 105, 408, 38 },
		{ 105, 409, 38 },
		{ 105, 410, 38 },
		{ 105, 411, 38 },
		{ 105, 412, 38 },
		{ 105, 413, 38 },
		{ 105, 414, 38 },
		{ 105, 415, 38 },
		{ 105, 416, 38 },
		{ 105, 417, 38 },
		{ 105, 418, 38 },
		{ 105, 419, 38 },
		{ 105, 420, 38 },
		{ 105, 421, 38 },
		{ 105, 422, 38 },
		{ 105, 423, 38 },
		{ 105, 424, 38 },
		{ 105, 425, 38 },
		{ 105, 426, 38 },
		{ 105, 427, 38 },
		{ 105, 428, 38 },
		{ 105, 429, 38 },
		{ 105, 430, 38 },
		{ 106, 390, 38 },
		{ 106, 391, 38 },
		{ 106, 392, 38 },
		{ 106, 393, 38 },
		{ 106, 394, 38 },
		{ 106, 395, 38 },
		{ 106, 396, 38 },
		{ 106, 397, 38 },
		{ 106, 398, 38 },
		{ 106, 399, 38 },
		{ 106, 400, 38 },
		{ 106, 401, 38 },
		{ 106, 402, 38 },
		{ 106, 403, 38 },
		{ 106, 404, 38 },
		{ 106, 405, 38 },
		{ 106, 406, 38 },
		{ 106, 407, 38 },
		{ 106, 408, 38 },
		{ 106, 409, 38 },
		{ 106, 410, 38 },
		{ 106, 411, 38 },
		{ 106, 412, 38 },
		{ 106, 413, 38 },
		{ 106, 414, 38 },
		{ 106, 415, 38 },
		{ 106, 416, 38 },
		{ 106, 417, 38 },
		{ 106, 418, 38 },
		{ 106, 419, 38 },
		{ 106, 420, 38 },
		{ 106, 421, 38 },
		{ 106, 422, 38 },
		{ 106, 423, 38 },
		{ 106, 424, 38 },
		{ 106, 425, 38 },
		{ 106, 426, 38 },
		{ 106, 427, 38 },
		{ 106, 428, 38 },
		{ 106, 429, 38 },
		{ 107, 389, 38 },
		{ 107, 390, 38 },
		{ 107, 391, 38 },
		{ 107, 392, 38 },
		{ 107, 393, 38 },
		{ 107, 394, 38 },
		{ 107, 395, 38 },
		{ 107, 396, 38 },
		{ 107, 397, 38 },
		{ 107, 398, 38 },
		{ 107, 399, 38 },
		{ 107, 400, 38 },
		{ 107, 401, 38 },
		{ 107, 402, 38 },
		{ 107, 403, 38 },
		{ 107, 404, 38 },
		{ 107, 405, 38 },
		{ 107, 406, 38 },
		{ 107, 407, 38 },
		{ 107, 408, 38 },
		{ 107, 409, 38 },
		{ 107, 410, 38 },
		{ 107, 411, 38 },
		{ 107, 412, 38 },
		{ 107, 413, 38 },
		{ 107, 414, 38 },
		{ 107, 415, 38 },
		{ 107, 416, 38 },
		{ 107, 417, 38 },
		{ 107, 418, 38 },
		{ 107, 419, 38 },
		{ 107, 420, 38 },
		{ 107, 421, 38 },
		{ 107, 422, 38 },
		{ 107, 423, 38 },
		{ 107, 424, 38 },
		{ 107, 425, 38 },
		{ 107, 426, 38 },
		{ 107, 427, 38 },
		{ 107, 428, 38 },
		{ 108, 388, 38 },
		{ 108, 389, 38 },
		{ 108, 390, 38 },
		{ 108, 391, 38 },
		{ 108, 392, 38 },
		{ 108, 393, 38 },
		{ 108, 394, 38 },
		{ 108, 395, 38 },
		{ 108, 396, 38 },
		{ 108, 397, 38 },
		{ 108, 398, 38 },
		{ 108, 399, 38 },
		{ 108, 400, 38 },
		{ 108, 401, 38 },
		{ 108, 402, 38 },
		{ 108, 403, 38 },
		{ 108, 404, 38 },
		{ 108, 405, 38 },
		{ 108, 406, 38 },
		{ 108, 407, 38 },
		{ 108, 408, 38 },
		{ 108, 409, 38 },
		{ 108, 410, 38 },
		{ 108, 411, 38 },
		{ 108, 412, 38 },
		{ 108, 413, 38 },
		{ 108, 414, 38 },
		{ 108, 415, 38 },
		{ 108, 416, 38 },
		{ 108, 417, 38 },
		{ 108, 418, 38 },
		{ 108, 419, 38 },
		{ 108, 420, 38 },
		{ 108, 421, 38 },
		{ 108, 422, 38 },
		{ 108, 423, 38 },
		{ 108, 424, 38 },
		{ 108, 425, 38 },
		{ 108, 426, 38 },
		{ 108, 427, 38 },
		{ 109, 387, 38 },
		{ 109, 388, 38 },
		{ 109, 389, 38 },
		{ 109, 390, 38 },
		{ 109, 391, 38 },
		{ 109, 392, 38 },
		{ 109, 393, 38 },
		{ 109, 394, 38 },
		{ 109, 395, 38 },
		{ 109, 396, 38 },
		{ 109, 397, 38 },
		{ 109, 398, 38 },
		{ 109, 399, 38 },
		{ 109, 400, 38 },
		{ 109, 401, 38 },
		{ 109, 402, 38 },
		{ 109, 403, 38 },
		{ 109, 404, 38 },
		{ 109, 405, 38 },
		{ 109, 406, 38 },
		{ 109, 407, 38 },
		{ 109, 408, 38 },
		{ 109, 409, 38 },
		{ 109, 410, 38 },
		{ 109, 411, 38 },
		{ 109, 412, 38 },
		{ 109, 413, 38 },
		{ 109, 414, 38 },
		{ 109, 415, 38 },
		{ 109, 416, 38 },
		{ 109, 417, 38 },
		{ 109, 418, 38 },
		{ 109, 419, 38 },
		{ 109, 420, 38 },
		{ 109, 421, 38 },
		{ 109, 422, 38 },
		{ 109, 423, 38 },
		{ 109, 424, 38 },
		{ 109, 425, 38 },
		{ 109, 426, 38 },
		{ 110, 386, 38 },
		{ 110, 387, 38 },
		{ 110, 388, 38 },
		{ 110, 389, 38 },
		{ 110, 390, 38 },
		{ 110, 391, 38 },
		{ 110, 392, 38 },
		{ 110, 393, 38 },
		{ 110, 394, 38 },
		{ 110, 395, 38 },
		{ 110, 396, 38 },
		{ 110, 397, 38 },
		{ 110, 398, 38 },
		{ 110, 399, 38 },
		{ 110, 400, 38 },
		{ 110, 401, 38 },
		{ 110, 402, 38 },
		{ 110, 403, 38 },
		{ 110, 404, 38 },
		{ 110, 405, 38 },
		{ 110, 406, 38 },
		{ 110, 407, 38 },
		{ 110, 408, 38 },
		{ 110, 409, 38 },
		{ 110, 410, 38 },
		{ 110, 411, 38 },
		{ 110, 412, 38 },
		{ 110, 413, 38 },
		{ 110, 414, 38 },
		{ 110, 415, 38 },
		{ 110, 416, 38 },
		{ 110, 417, 38 },
		{ 110, 418, 38 },
		{ 110, 419, 38 },
		{ 110, 420, 38 },
		{ 110, 421, 38 },
		{ 110, 422, 38 },
		{ 110, 423, 38 },
		{ 110, 424, 38 },
		{ 111, 385, 38 },
		{ 111, 386, 38 },
		{ 111, 387, 38 },
		{ 111, 388, 38 },
		{ 111, 389, 38 },
		{ 111, 390, 38 },
		{ 111, 391, 38 },
		{ 111, 392, 38 },
		{ 111, 393, 38 },
		{ 111, 394, 38 },
		{ 111, 395, 38 },
		{ 111, 396, 38 },
		{ 111, 397, 38 },
		{ 111, 398, 38 },
		{ 111, 399, 38 },
		{ 111, 400, 38 },
		{ 111, 401, 38 },
		{ 111, 402, 38 },
		{ 111, 403, 38 },
		{ 111, 404, 38 },
		{ 111, 405, 38 },
		{ 111, 406, 38 },
		{ 111, 407, 38 },
		{ 111, 408, 38 },
		{ 111, 409, 38 },
		{ 111, 410, 38 },
		{ 111, 411, 38 },
		{ 111, 412, 38 },
		{ 111, 413, 38 },
		{ 111, 414, 38 },
		{ 111, 415, 38 },
		{ 111, 416, 38 },
		{ 111, 417, 38 },
		{ 111, 418, 38 },
		{ 111, 419, 38 },
		{ 111, 420, 38 },
		{ 111, 421, 38 },
		{ 111, 422, 38 },
		{ 111, 423, 38 },
		{ 112, 384, 38 },
		{ 112, 385, 38 },
		{ 112, 386, 38 },
		{ 112, 387, 38 },
		{ 112, 388, 38 },
		{ 112, 389, 38 },
		{ 112, 390, 38 },
		{ 112, 391, 38 },
		{ 112, 392, 38 },
		{ 112, 393, 38 },
		{ 112, 394, 38 },
		{ 112, 395, 38 },
		{ 112, 396, 38 },
		{ 112, 397, 38 },
		{ 112, 398, 38 },
		{ 112, 399, 38 },
		{ 112, 400, 38 },
		{ 112, 401, 38 },
		{ 112, 402, 38 },
		{ 112, 403, 38 },
		{ 112, 404, 38 },
		{ 112, 405, 38 },
		{ 112, 406, 38 },
		{ 112, 407, 38 },
		{ 112, 408, 38 },
		{ 112, 409, 38 },
		{ 112, 410, 38 },
		{ 112, 411, 38 },
		{ 112, 412, 38 },
		{ 112, 413, 38 },
		{ 112, 414, 38 },
		{ 112, 415, 38 },
		{ 112, 416, 38 },
		{ 112, 417, 38 },
		{ 112, 418, 38 },
		{ 112, 419, 38 },
		{ 112, 420, 38 },
		{ 112, 421, 38 },
		{ 112, 422, 38 },
		{ 113, 384, 38 },
		{ 113, 385, 38 },
		{ 113, 386, 38 },
		{ 113, 387, 38 },
		{ 113, 388, 38 },
		{ 113, 389, 38 },
		{ 113, 390, 38 },
		{ 113, 391, 38 },
		{ 113, 392, 38 },
		{ 113, 393, 38 },
		{ 113, 394, 38 },
		{ 113, 395, 38 },
		{ 113, 396, 38 },
		{ 113, 397, 38 },
		{ 113, 398, 38 },
		{ 113, 399, 38 },
		{ 113, 400, 38 },
		{ 113, 401, 38 },
		{ 113, 402, 38 },
		{ 113, 403, 38 },
		{ 113, 404, 38 },
		{ 113, 405, 38 },
		{ 113, 406, 38 },
		{ 113, 407, 38 },
		{ 113, 408, 38 },
		{ 113, 409, 38 },
		{ 113, 410, 38 },
		{ 113, 411, 38 },
		{ 113, 412, 38 },
		{ 113, 413, 38 },
		{ 113, 414, 38 },
		{ 113, 415, 38 },
		{ 113, 416, 38 },
		{ 113, 417, 38 },
		{ 113, 418, 38 },
		{ 113, 419, 38 },
		{ 113, 420, 38 },
		{ 113, 421, 38 },
		{ 114, 383, 38 },
		{ 114, 384, 38 },
		{ 114, 385, 38 },
		{ 114, 386, 38 },
		{ 114, 387, 38 },
		{ 114, 388, 38 },
		{ 114, 389, 38 },
		{ 114, 390, 38 },
		{ 114, 391, 38 },
		{ 114, 392, 38 },
		{ 114, 393, 38 },
		{ 114, 394, 38 },
		{ 114, 395, 38 },
		{ 114, 396, 38 },
		{ 114, 397, 38 },
		{ 114, 398, 38 },
		{ 114, 399, 38 },
		{ 114, 400, 38 },
		{ 114, 401, 38 },
		{ 114, 402, 38 },
		{ 114, 403, 38 },
		{ 114, 404, 38 },
		{ 114, 405, 38 },
		{ 114, 406, 38 },
		{ 114, 407, 38 },
		{ 114, 408, 38 },
		{ 114, 409, 38 },
		{ 114, 410, 38 },
		{ 114, 411, 38 },
		{ 114, 412, 38 },
		{ 114, 413, 38 },
		{ 114, 414, 38 },
		{ 114, 415, 38 },
		{ 114, 416, 38 },
		{ 114, 417, 38 },
		{ 114, 418, 38 },
		{ 114, 419, 38 },
		{ 114, 420, 38 },
		{ 115, 382, 38 },
		{ 115, 383, 38 },
		{ 115, 384, 38 },
		{ 115, 385, 38 },
		{ 115, 386, 38 },
		{ 115, 387, 38 },
		{ 115, 388, 38 },
		{ 115, 389, 38 },
		{ 115, 390, 38 },
		{ 115, 391, 38 },
		{ 115, 392, 38 },
		{ 115, 393, 38 },
		{ 115, 394, 38 },
		{ 115, 395, 38 },
		{ 115, 396, 38 },
		{ 115, 397, 38 },
		{ 115, 398, 38 },
		{ 115, 399, 38 },
		{ 115, 400, 38 },
		{ 115, 401, 38 },
		{ 115, 402, 38 },
		{ 115, 403, 38 },
		{ 115, 404, 38 },
		{ 115, 405, 38 },
		{ 115, 406, 38 },
		{ 115, 407, 38 },
		{ 115, 408, 38 },
		{ 115, 409, 38 },
		{ 115, 410, 38 },
		{ 115, 411, 38 },
		{ 115, 412, 38 },
		{ 115, 413, 38 },
		{ 115, 414, 38 },
		{ 115, 415, 38 },
		{ 115, 416, 38 },
		{ 115, 417, 38 },
		{ 115, 418, 38 },
		{ 115, 419, 38 },
		{ 116, 381, 38 },
		{ 116, 382, 38 },
		{ 116, 383, 38 },
		{ 116, 384, 38 },
		{ 116, 385, 38 },
		{ 116, 386, 38 },
		{ 116, 387, 38 },
		{ 116, 388, 38 },
		{ 116, 389, 38 },
		{ 116, 390, 38 },
		{ 116, 391, 38 },
		{ 116, 392, 38 },
		{ 116, 393, 38 },
		{ 116, 394, 38 },
		{ 116, 395, 38 },
		{ 116, 396, 38 },
		{ 116, 397, 38 },
		{ 116, 398, 38 },
		{ 116, 399, 38 },
		{ 116, 400, 38 },
		{ 116, 401, 38 },
		{ 116, 402, 38 },
		{ 116, 403, 38 },
		{ 116, 404, 38 },
		{ 116, 405, 38 },
		{ 116, 406, 38 },
		{ 116, 407, 38 },
		{ 116, 408, 38 },
		{ 116, 409, 38 },
		{ 116, 410, 38 },
		{ 116, 411, 38 },
		{ 116, 412, 38 },
		{ 116, 413, 38 },
		{ 116, 414, 38 },
		{ 116, 415, 38 },
		{ 116, 416, 38 },
		{ 116, 417, 38 },
		{ 116, 418, 38 },
		{ 117, 380, 38 },
		{ 117, 381, 38 },
		{ 117, 382, 38 },
		{ 117, 383, 38 },
		{ 117, 384, 38 },
		{ 117, 385, 38 },
		{ 117, 386, 38 },
		{ 117, 387, 38 },
		{ 117, 388, 38 },
		{ 117, 389, 38 },
		{ 117, 390, 38 },
		{ 117, 391, 38 },
		{ 117, 392, 38 },
		{ 117, 393, 38 },
		{ 117, 394, 38 },
		{ 117, 395, 38 },
		{ 117, 396, 38 },
		{ 117, 397, 38 },
		{ 117, 398, 38 },
		{ 117, 399, 38 },
		{ 117, 400, 38 },
		{ 117, 401, 38 },
		{ 117, 402, 38 },
		{ 117, 403, 38 },
		{ 117, 404, 38 },
		{ 117, 405, 38 },
		{ 117, 406, 38 },
		{ 117, 407, 38 },
		{ 117, 408, 38 },
		{ 117, 409, 38 },
		{ 117, 410, 38 },
		{ 117, 411, 38 },
		{ 117, 412, 38 },
		{ 117, 413, 38 },
		{ 117, 414, 38 },
		{ 117, 415, 38 },
		{ 117, 416, 38 },
		{ 118, 379, 38 },
		{ 118, 380, 38 },
		{ 118, 381, 38 },
		{ 118, 382, 38 },
		{ 118, 383, 38 },
		{ 118, 384, 38 },
		{ 118, 385, 38 },
		{ 118, 386, 38 },
		{ 118, 387, 38 },
		{ 118, 388, 38 },
		{ 118, 389, 38 },
		{ 118, 390, 38 },
		{ 118, 391, 38 },
		{ 118, 392, 38 },
		{ 118, 393, 38 },
		{ 118, 394, 38 },
		{ 118, 395, 38 },
		{ 118, 396, 38 },
		{ 118, 397, 38 },
		{ 118, 398, 38 },
		{ 118, 399, 38 },
		{ 118, 400, 38 },
		{ 118, 401, 38 },
		{ 118, 402, 38 },
		{ 118, 403, 38 },
		{ 118, 404, 38 },
		{ 118, 405, 38 },
		{ 118, 406, 38 },
		{ 118, 407, 38 },
		{ 118, 408, 38 },
		{ 118, 409, 38 },
		{ 118, 410, 38 },
		{ 118, 411, 38 },
		{ 118, 412, 38 },
		{ 118, 413, 38 },
		{ 118, 414, 38 },
		{ 118, 415, 38 },
		{ 119, 378, 38 },
		{ 119, 379, 38 },
		{ 119, 380, 38 },
		{ 119, 381, 38 },
		{ 119, 382, 38 },
		{ 119, 383, 38 },
		{ 119, 384, 38 },
		{ 119, 385, 38 },
		{ 119, 386, 38 },
		{ 119, 387, 38 },
		{ 119, 388, 38 },
		{ 119, 389, 38 },
		{ 119, 390, 38 },
		{ 119, 391, 38 },
		{ 119, 392, 38 },
		{ 119, 393, 38 },
		{ 119, 394, 38 },
		{ 119, 395, 38 },
		{ 119, 396, 38 },
		{ 119, 397, 38 },
		{ 119, 398, 38 },
		{ 119, 399, 38 },
		{ 119, 400, 38 },
		{ 119, 401, 38 },
		{ 119, 402, 38 },
		{ 119, 403, 38 },
		{ 119, 404, 38 },
		{ 119, 405, 38 },
		{ 119, 406, 38 },
		{ 119, 407, 38 },
		{ 119, 408, 38 },
		{ 119, 409, 38 },
		{ 119, 410, 38 },
		{ 119, 411, 38 },
		{ 119, 412, 38 },
		{ 119, 413, 38 },
		{ 119, 414, 38 },
		{ 120, 377, 38 },
		{ 120, 378, 38 },
		{ 120, 379, 38 },
		{ 120, 380, 38 },
		{ 120, 381, 38 },
		{ 120, 382, 38 },
		{ 120, 383, 38 },
		{ 120, 384, 38 },
		{ 120, 385, 38 },
		{ 120, 386, 38 },
		{ 120, 387, 38 },
		{ 120, 388, 38 },
		{ 120, 389, 38 },
		{ 120, 390, 38 },
		{ 120, 391, 38 },
		{ 120, 392, 38 },
		{ 120, 393, 38 },
		{ 120, 394, 38 },
		{ 120, 395, 38 },
		{ 120, 396, 38 },
		{ 120, 397, 38 },
		{ 120, 398, 38 },
		{ 120, 399, 38 },
		{ 120, 400, 38 },
		{ 120, 401, 38 },
		{ 120, 402, 38 },
		{ 120, 403, 38 },
		{ 120, 404, 38 },
		{ 120, 405, 38 },
		{ 120, 406, 38 },
		{ 120, 407, 38 },
		{ 120, 408, 38 },
		{ 120, 409, 38 },
		{ 120, 410, 38 },
		{ 120, 411, 38 },
		{ 120, 412, 38 },
		{ 120, 413, 38 },
		{ 121, 377, 38 },
		{ 121, 378, 38 },
		{ 121, 379, 38 },
		{ 121, 380, 38 },
		{ 121, 381, 38 },
		{ 121, 382, 38 },
		{ 121, 383, 38 },
		{ 121, 384, 38 },
		{ 121, 385, 38 },
		{ 121, 386, 38 },
		{ 121, 387, 38 },
		{ 121, 388, 38 },
		{ 121, 389, 38 },
		{ 121, 390, 38 },
		{ 121, 391, 38 },
		{ 121, 392, 38 },
		{ 121, 393, 38 },
		{ 121, 394, 38 },
		{ 121, 395, 38 },
		{ 121, 396, 38 },
		{ 121, 397, 38 },
		{ 121, 398, 38 },
		{ 121, 399, 38 },
		{ 121, 400, 38 },
		{ 121, 401, 38 },
		{ 121, 402, 38 },
		{ 121, 403, 38 },
		{ 121, 404, 38 },
		{ 121, 405, 38 },
		{ 121, 406, 38 },
		{ 121, 407, 38 },
		{ 121, 408, 38 },
		{ 121, 409, 38 },
		{ 121, 410, 38 },
		{ 121, 411, 38 },
		{ 121, 412, 38 },
		{ 122, 376, 38 },
		{ 122, 377, 38 },
		{ 122, 378, 38 },
		{ 122, 379, 38 },
		{ 122, 380, 38 },
		{ 122, 381, 38 },
		{ 122, 382, 38 },
		{ 122, 383, 38 },
		{ 122, 384, 38 },
		{ 122, 385, 38 },
		{ 122, 386, 38 },
		{ 122, 387, 38 },
		{ 122, 388, 38 },
		{ 122, 389, 38 },
		{ 122, 390, 38 },
		{ 122, 391, 38 },
		{ 122, 392, 38 },
		{ 122, 393, 38 },
		{ 122, 394, 38 },
		{ 122, 395, 38 },
		{ 122, 396, 38 },
		{ 122, 397, 38 },
		{ 122, 398, 38 },
		{ 122, 399, 38 },
		{ 122, 400, 38 },
		{ 122, 401, 38 },
		{ 122, 402, 38 },
		{ 122, 403, 38 },
		{ 122, 404, 38 },
		{ 122, 405, 38 },
		{ 122, 406, 38 },
		{ 122, 407, 38 },
		{ 122, 408, 38 },
		{ 122, 409, 38 },
		{ 122, 410, 38 },
		{ 122, 411, 38 },
		{ 123, 375, 38 },
		{ 123, 376, 38 },
		{ 123, 377, 38 },
		{ 123, 378, 38 },
		{ 123, 379, 38 },
		{ 123, 380, 38 },
		{ 123, 381, 38 },
		{ 123, 382, 38 },
		{ 123, 383, 38 },
		{ 123, 384, 38 },
		{ 123, 385, 38 },
		{ 123, 386, 38 },
		{ 123, 387, 38 },
		{ 123, 388, 38 },
		{ 123, 389, 38 },
		{ 123, 390, 38 },
		{ 123, 391, 38 },
		{ 123, 392, 38 },
		{ 123, 393, 38 },
		{ 123, 394, 38 },
		{ 123, 395, 38 },
		{ 123, 396, 38 },
		{ 123, 397, 38 },
		{ 123, 398, 38 },
		{ 123, 399, 38 },
		{ 123, 400, 38 },
		{ 123, 401, 38 },
		{ 123, 402, 38 },
		{ 123, 403, 38 },
		{ 123, 404, 38 },
		{ 123, 405, 38 },
		{ 123, 406, 38 },
		{ 123, 407, 38 },
		{ 123, 408, 38 },
		{ 123, 409, 38 },
		{ 123, 410, 38 },
		{ 124, 374, 38 },
		{ 124, 375, 38 },
		{ 124, 376, 38 },
		{ 124, 377, 38 },
		{ 124, 378, 38 },
		{ 124, 379, 38 },
		{ 124, 380, 38 },
		{ 124, 381, 38 },
		{ 124, 382, 38 },
		{ 124, 383, 38 },
		{ 124, 384, 38 },
		{ 124, 385, 38 },
		{ 124, 386, 38 },
		{ 124, 387, 38 },
		{ 124, 388, 38 },
		{ 124, 389, 38 },
		{ 124, 390, 38 },
		{ 124, 391, 38 },
		{ 124, 392, 38 },
		{ 124, 393, 38 },
		{ 124, 394, 38 },
		{ 124, 395, 38 },
		{ 124, 396, 38 },
		{ 124, 397, 38 },
		{ 124, 398, 38 },
		{ 124, 399, 38 },
		{ 124, 400, 38 },
		{ 124, 401, 38 },
		{ 124, 402, 38 },
		{ 124, 403, 38 },
		{ 124, 404, 38 },
		{ 124, 405, 38 },
		{ 124, 406, 38 },
		{ 124, 407, 38 },
		{ 124, 408, 38 },
		{ 125, 373, 38 },
		{ 125, 374, 38 },
		{ 125, 375, 38 },
		{ 125, 376, 38 },
		{ 125, 377, 38 },
		{ 125, 378, 38 },
		{ 125, 379, 38 },
		{ 125, 380, 38 },
		{ 125, 381, 38 },
		{ 125, 382, 38 },
		{ 125, 383, 38 },
		{ 125, 384, 38 },
		{ 125, 385, 38 },
		{ 125, 386, 38 },
		{ 125, 387, 38 },
		{ 125, 388, 38 },
		{ 125, 389, 38 },
		{ 125, 390, 38 },
		{ 125, 391, 38 },
		{ 125, 392, 38 },
		{ 125, 393, 38 },
		{ 125, 394, 38 },
		{ 125, 395, 38 },
		{ 125, 396, 38 },
		{ 125, 397, 38 },
		{ 125, 398, 38 },
		{ 125, 399, 38 },
		{ 125, 400, 38 },
		{ 125, 401, 38 },
		{ 125, 402, 38 },
		{ 125, 403, 38 },
		{ 125, 404, 38 },
		{ 125, 405, 38 },
		{ 125, 406, 38 },
		{ 125, 407, 38 },
		{ 126, 372, 38 },
		{ 126, 373, 38 },
		{ 126, 374, 38 },
		{ 126, 375, 38 },
		{ 126, 376, 38 },
		{ 126, 377, 38 },
		{ 126, 378, 38 },
		{ 126, 379, 38 },
		{ 126, 380, 38 },
		{ 126, 381, 38 },
		{ 126, 382, 38 },
		{ 126, 383, 38 },
		{ 126, 384, 38 },
		{ 126, 385, 38 },
		{ 126, 386, 38 },
		{ 126, 387, 38 },
		{ 126, 388, 38 },
		{ 126, 389, 38 },
		{ 126, 390, 38 },
		{ 126, 391, 38 },
		{ 126, 392, 38 },
		{ 126, 393, 38 },
		{ 126, 394, 38 },
		{ 126, 395, 38 },
		{ 126, 396, 38 },
		{ 126, 397, 38 },
		{ 126, 398, 38 },
		{ 126, 399, 38 },
		{ 126, 400, 38 },
		{ 126, 401, 38 },
		{ 126, 402, 38 },
		{ 126, 403, 38 },
		{ 126, 404, 38 },
		{ 126, 405, 38 },
		{ 126, 406, 38 },
		{ 127, 371, 38 },
		{ 127, 372, 38 },
		{ 127, 373, 38 },
		{ 127, 374, 38 },
		{ 127, 375, 38 },
		{ 127, 376, 38 },
		{ 127, 377, 38 },
		{ 127, 378, 38 },
		{ 127, 379, 38 },
		{ 127, 380, 38 },
		{ 127, 381, 38 },
		{ 127, 382, 38 },
		{ 127, 383, 38 },
		{ 127, 384, 38 },
		{ 127, 385, 38 },
		{ 127, 386, 38 },
		{ 127, 387, 38 },
		{ 127, 388, 38 },
		{ 127, 389, 38 },
		{ 127, 390, 38 },
		{ 127, 391, 38 },
		{ 127, 392, 38 },
		{ 127, 393, 38 },
		{ 127, 394, 38 },
		{ 127, 395, 38 },
		{ 127, 396, 38 },
		{ 127, 397, 38 },
		{ 127, 398, 38 },
		{ 127, 399, 38 },
		{ 127, 400, 38 },
		{ 127, 401, 38 },
		{ 127, 402, 38 },
		{ 127, 403, 38 },
		{ 127, 404, 38 },
		{ 127, 405, 38 },
		{ 128, 370, 38 },
		{ 128, 371, 38 },
		{ 128, 372, 38 },
		{ 128, 373, 38 },
		{ 128, 374, 38 },
		{ 128, 375, 38 },
		{ 128, 376, 38 },
		{ 128, 377, 38 },
		{ 128, 378, 38 },
		{ 128, 379, 38 },
		{ 128, 380, 38 },
		{ 128, 381, 38 },
		{ 128, 382, 38 },
		{ 128, 383, 38 },
		{ 128, 384, 38 },
		{ 128, 385, 38 },
		{ 128, 386, 38 },
		{ 128, 387, 38 },
		{ 128, 388, 38 },
		{ 128, 389, 38 },
		{ 128, 390, 38 },
		{ 128, 391, 38 },
		{ 128, 392, 38 },
		{ 128, 393, 38 },
		{ 128, 394, 38 },
		{ 128, 395, 38 },
		{ 128, 396, 38 },
		{ 128, 397, 38 },
		{ 128, 398, 38 },
		{ 128, 399, 38 },
		{ 128, 400, 38 },
		{ 128, 401, 38 },
		{ 128, 402, 38 },
		{ 128, 403, 38 },
		{ 128, 404, 38 },
		{ 129, 370, 38 },
		{ 129, 371, 38 },
		{ 129, 372, 38 },
		{ 129, 373, 38 },
		{ 129, 374, 38 },
		{ 129, 375, 38 },
		{ 129, 376, 38 },
		{ 129, 377, 38 },
		{ 129, 378, 38 },
		{ 129, 379, 38 },
		{ 129, 380, 38 },
		{ 129, 381, 38 },
		{ 129, 382, 38 },
		{ 129, 383, 38 },
		{ 129, 384, 38 },
		{ 129, 385, 38 },
		{ 129, 386, 38 },
		{ 129, 387, 38 },
		{ 129, 388, 38 },
		{ 129, 389, 38 },
		{ 129, 390, 38 },
		{ 129, 391, 38 },
		{ 129, 392, 38 },
		{ 129, 393, 38 },
		{ 129, 394, 38 },
		{ 129, 395, 38 },
		{ 129, 396, 38 },
		{ 129, 397, 38 },
		{ 129, 398, 38 },
		{ 129, 399, 38 },
		{ 129, 400, 38 },
		{ 129, 401, 38 },
		{ 129, 402, 38 },
		{ 129, 403, 38 },
		{ 130, 369, 38 },
		{ 130, 370, 38 },
		{ 130, 371, 38 },
		{ 130, 372, 38 },
		{ 130, 373, 38 },
		{ 130, 374, 38 },
		{ 130, 375, 38 },
		{ 130, 376, 38 },
		{ 130, 377, 38 },
		{ 130, 378, 38 },
		{ 130, 379, 38 },
		{ 130, 380, 38 },
		{ 130, 381, 38 },
		{ 130, 382, 38 },
		{ 130, 383, 38 },
		{ 130, 384, 38 },
		{ 130, 385, 38 },
		{ 130, 386, 38 },
		{ 130, 387, 38 },
		{ 130, 388, 38 },
		{ 130, 389, 38 },
		{ 130, 390, 38 },
		{ 130, 391, 38 },
		{ 130, 392, 38 },
		{ 130, 393, 38 },
		{ 130, 394, 38 },
		{ 130, 395, 38 },
		{ 130, 396, 38 },
		{ 130, 397, 38 },
		{ 130, 398, 38 },
		{ 130, 399, 38 },
		{ 130, 400, 38 },
		{ 130, 401, 38 },
		{ 130, 402, 38 },
		{ 131, 368, 38 },
		{ 131, 369, 38 },
		{ 131, 370, 38 },
		{ 131, 371, 38 },
		{ 131, 372, 38 },
		{ 131, 373, 38 },
		{ 131, 374, 38 },
		{ 131, 375, 38 },
		{ 131, 376, 38 },
		{ 131, 377, 38 },
		{ 131, 378, 38 },
		{ 131, 379, 38 },
		{ 131, 380, 38 },
		{ 131, 381, 38 },
		{ 131, 382, 38 },
		{ 131, 383, 38 },
		{ 131, 384, 38 },
		{ 131, 385, 38 },
		{ 131, 386, 38 },
		{ 131, 387, 38 },
		{ 131, 388, 38 },
		{ 131, 389, 38 },
		{ 131, 390, 38 },
		{ 131, 391, 38 },
		{ 131, 392, 38 },
		{ 131, 393, 38 },
		{ 131, 394, 38 },
		{ 131, 395, 38 },
		{ 131, 396, 38 },
		{ 131, 397, 38 },
		{ 131, 398, 38 },
		{ 131, 399, 38 },
		{ 131, 400, 38 },
		{ 132, 367, 38 },
		{ 132, 368, 38 },
		{ 132, 369, 38 },
		{ 132, 370, 38 },
		{ 132, 371, 38 },
		{ 132, 372, 38 },
		{ 132, 373, 38 },
		{ 132, 374, 38 },
		{ 132, 375, 38 },
		{ 132, 376, 38 },
		{ 132, 377, 38 },
		{ 132, 378, 38 },
		{ 132, 379, 38 },
		{ 132, 380, 38 },
		{ 132, 381, 38 },
		{ 132, 382, 38 },
		{ 132, 383, 38 },
		{ 132, 384, 38 },
		{ 132, 385, 38 },
		{ 132, 386, 38 },
		{ 132, 387, 38 },
		{ 132, 388, 38 },
		{ 132, 389, 38 },
		{ 132, 390, 38 },
		{ 132, 391, 38 },
		{ 132, 392, 38 },
		{ 132, 393, 38 },
		{ 132, 394, 38 },
		{ 132, 395, 38 },
		{ 132, 396, 38 },
		{ 132, 397, 38 },
		{ 132, 398, 38 },
		{ 132, 399, 38 },
		{ 133, 366, 38 },
		{ 133, 367, 38 },
		{ 133, 368, 38 },
		{ 133, 369, 38 },
		{ 133, 370, 38 },
		{ 133, 371, 38 },
		{ 133, 372, 38 },
		{ 133, 373, 38 },
		{ 133, 374, 38 },
		{ 133, 375, 38 },
		{ 133, 376, 38 },
		{ 133, 377, 38 },
		{ 133, 378, 38 },
		{ 133, 379, 38 },
		{ 133, 380, 38 },
		{ 133, 381, 38 },
		{ 133, 382, 38 },
		{ 133, 383, 38 },
		{ 133, 384, 38 },
		{ 133, 385, 38 },
		{ 133, 386, 38 },
		{ 133, 387, 38 },
		{ 133, 388, 38 },
		{ 133, 389, 38 },
		{ 133, 390, 38 },
		{ 133, 391, 38 },
		{ 133, 392, 38 },
		{ 133, 393, 38 },
		{ 133, 394, 38 },
		{ 133, 395, 38 },
		{ 133, 396, 38 },
		{ 133, 397, 38 },
		{ 133, 398, 38 },
		{ 134, 365, 38 },
		{ 134, 366, 38 },
		{ 134, 367, 38 },
		{ 134, 368, 38 },
		{ 134, 369, 38 },
		{ 134, 370, 38 },
		{ 134, 371, 38 },
		{ 134, 372, 38 },
		{ 134, 373, 38 },
		{ 134, 374, 38 },
		{ 134, 375, 38 },
		{ 134, 376, 38 },
		{ 134, 377, 38 },
		{ 134, 378, 38 },
		{ 134, 379, 38 },
		{ 134, 380, 38 },
		{ 134, 381, 38 },
		{ 134, 382, 38 },
		{ 134, 383, 38 },
		{ 134, 384, 38 },
		{ 134, 385, 38 },
		{ 134, 386, 38 },
		{ 134, 387, 38 },
		{ 134, 388, 38 },
		{ 134, 389, 38 },
		{ 134, 390, 38 },
		{ 134, 391, 38 },
		{ 134, 392, 38 },
		{ 134, 393, 38 },
		{ 134, 394, 38 },
		{ 134, 395, 38 },
		{ 134, 396, 38 },
		{ 134, 397, 38 },
		{ 135, 364, 38 },
		{ 135, 365, 38 },
		{ 135, 366, 38 },
		{ 135, 367, 38 },
		{ 135, 368, 38 },
		{ 135, 369, 38 },
		{ 135, 370, 38 },
		{ 135, 371, 38 },
		{ 135, 372, 38 },
		{ 135, 373, 38 },
		{ 135, 374, 38 },
		{ 135, 375, 38 },
		{ 135, 376, 38 },
		{ 135, 377, 38 },
		{ 135, 378, 38 },
		{ 135, 379, 38 },
		{ 135, 380, 38 },
		{ 135, 381, 38 },
		{ 135, 382, 38 },
		{ 135, 383, 38 },
		{ 135, 384, 38 },
		{ 135, 385, 38 },
		{ 135, 386, 38 },
		{ 135, 387, 38 },
		{ 135, 388, 38 },
		{ 135, 389, 38 },
		{ 135, 390, 38 },
		{ 135, 391, 38 },
		{ 135, 392, 38 },
		{ 135, 393, 38 },
		{ 135, 394, 38 },
		{ 135, 395, 38 },
		{ 135, 396, 38 },
		{ 136, 363, 38 },
		{ 136, 364, 38 },
		{ 136, 365, 38 },
		{ 136, 366, 38 },
		{ 136, 367, 38 },
		{ 136, 368, 38 },
		{ 136, 369, 38 },
		{ 136, 370, 38 },
		{ 136, 371, 38 },
		{ 136, 372, 38 },
		{ 136, 373, 38 },
		{ 136, 374, 38 },
		{ 136, 375, 38 },
		{ 136, 376, 38 },
		{ 136, 377, 38 },
		{ 136, 378, 38 },
		{ 136, 379, 38 },
		{ 136, 380, 38 },
		{ 136, 381, 38 },
		{ 136, 382, 38 },
		{ 136, 383, 38 },
		{ 136, 384, 38 },
		{ 136, 385, 38 },
		{ 136, 386, 38 },
		{ 136, 387, 38 },
		{ 136, 388, 38 },
		{ 136, 389, 38 },
		{ 136, 390, 38 },
		{ 136, 391, 38 },
		{ 136, 392, 38 },
		{ 136, 393, 38 },
		{ 136, 394, 38 },
		{ 136, 395, 38 },
		{ 137, 363, 38 },
		{ 137, 364, 38 },
		{ 137, 365, 38 },
		{ 137, 366, 38 },
		{ 137, 367, 38 },
		{ 137, 368, 38 },
		{ 137, 369, 38 },
		{ 137, 370, 38 },
		{ 137, 371, 38 },
		{ 137, 372, 38 },
		{ 137, 373, 38 },
		{ 137, 374, 38 },
		{ 137, 375, 38 },
		{ 137, 376, 38 },
		{ 137, 377, 38 },
		{ 137, 378, 38 },
		{ 137, 379, 38 },
		{ 137, 380, 38 },
		{ 137, 381, 38 },
		{ 137, 382, 38 },
		{ 137, 383, 38 },
		{ 137, 384, 38 },
		{ 137, 385, 38 },
		{ 137, 386, 38 },
		{ 137, 387, 38 },
		{ 137, 388, 38 },
		{ 137, 389, 38 },
		{ 137, 390, 38 },
		{ 137, 391, 38 },
		{ 137, 392, 38 },
		{ 137, 393, 38 },
		{ 137, 394, 38 },
		{ 138, 362, 38 },
		{ 138, 363, 38 },
		{ 138, 364, 38 },
		{ 138, 365, 38 },
		{ 138, 366, 38 },
		{ 138, 367, 38 },
		{ 138, 368, 38 },
		{ 138, 369, 38 },
		{ 138, 370, 38 },
		{ 138, 371, 38 },
		{ 138, 372, 38 },
		{ 138, 373, 38 },
		{ 138, 374, 38 },
		{ 138, 375, 38 },
		{ 138, 376, 38 },
		{ 138, 377, 38 },
		{ 138, 378, 38 },
		{ 138, 379, 38 },
		{ 138, 380, 38 },
		{ 138, 381, 38 },
		{ 138, 382, 38 },
		{ 138, 383, 38 },
		{ 138, 384, 38 },
		{ 138, 385, 38 },
		{ 138, 386, 38 },
		{ 138, 387, 38 },
		{ 138, 388, 38 },
		{ 138, 389, 38 },
		{ 138, 390, 38 },
		{ 138, 391, 38 },
		{ 138, 392, 38 },
		{ 139, 361, 38 },
		{ 139, 362, 38 },
		{ 139, 363, 38 },
		{ 139, 364, 38 },
		{ 139, 365, 38 },
		{ 139, 366, 38 },
		{ 139, 367, 38 },
		{ 139, 368, 38 },
		{ 139, 369, 38 },
		{ 139, 370, 38 },
		{ 139, 371, 38 },
		{ 139, 372, 38 },
		{ 139, 373, 38 },
		{ 139, 374, 38 },
		{ 139, 375, 38 },
		{ 139, 376, 38 },
		{ 139, 377, 38 },
		{ 139, 378, 38 },
		{ 139, 379, 38 },
		{ 139, 380, 38 },
		{ 139, 381, 38 },
		{ 139, 382, 38 },
		{ 139, 383, 38 },
		{ 139, 384, 38 },
		{ 139, 385, 38 },
		{ 139, 386, 38 },
		{ 139, 387, 38 },
		{ 139, 388, 38 },
		{ 139, 389, 38 },
		{ 139, 390, 38 },
		{ 139, 391, 38 },
		{ 140, 360, 38 },
		{ 140, 361, 38 },
		{ 140, 362, 38 },
		{ 140, 363, 38 },
		{ 140, 364, 38 },
		{ 140, 365, 38 },
		{ 140, 366, 38 },
		{ 140, 367, 38 },
		{ 140, 368, 38 },
		{ 140, 369, 38 },
		{ 140, 370, 38 },
		{ 140, 371, 38 },
		{ 140, 372, 38 },
		{ 140, 373, 38 },
		{ 140, 374, 38 },
		{ 140, 375, 38 },
		{ 140, 376, 38 },
		{ 140, 377, 38 },
		{ 140, 378, 38 },
		{ 140, 379, 38 },
		{ 140, 380, 38 },
		{ 140, 381, 38 },
		{ 140, 382, 38 },
		{ 140, 383, 38 },
		{ 140, 384, 38 },
		{ 140, 385, 38 },
		{ 140, 386, 38 },
		{ 140, 387, 38 },
		{ 140, 388, 38 },
		{ 140, 389, 38 },
		{ 140, 390, 38 },
		{ 141, 359, 38 },
		{ 141, 360, 38 },
		{ 141, 361, 38 },
		{ 141, 362, 38 },
		{ 141, 363, 38 },
		{ 141, 364, 38 },
		{ 141, 365, 38 },
		{ 141, 366, 38 },
		{ 141, 367, 38 },
		{ 141, 368, 38 },
		{ 141, 369, 38 },
		{ 141, 370, 38 },
		{ 141, 371, 38 },
		{ 141, 372, 38 },
		{ 141, 373, 38 },
		{ 141, 374, 38 },
		{ 141, 375, 38 },
		{ 141, 376, 38 },
		{ 141, 377, 38 },
		{ 141, 378, 38 },
		{ 141, 379, 38 },
		{ 141, 380, 38 },
		{ 141, 381, 38 },
		{ 141, 382, 38 },
		{ 141, 383, 38 },
		{ 141, 384, 38 },
		{ 141, 385, 38 },
		{ 141, 386, 38 },
		{ 141, 387, 38 },
		{ 141, 388, 38 },
		{ 141, 389, 38 },
		{ 142, 358, 38 },
		{ 142, 359, 38 },
		{ 142, 360, 38 },
		{ 142, 361, 38 },
		{ 142, 362, 38 },
		{ 142, 363, 38 },
		{ 142, 364, 38 },
		{ 142, 365, 38 },
		{ 142, 366, 38 },
		{ 142, 367, 38 },
		{ 142, 368, 38 },
		{ 142, 369, 38 },
		{ 142, 370, 38 },
		{ 142, 371, 38 },
		{ 142, 372, 38 },
		{ 142, 373, 38 },
		{ 142, 374, 38 },
		{ 142, 375, 38 },
		{ 142, 376, 38 },
		{ 142, 377, 38 },
		{ 142, 378, 38 },
		{ 142, 379, 38 },
		{ 142, 380, 38 },
		{ 142, 381, 38 },
		{ 142, 382, 38 },
		{ 142, 383, 38 },
		{ 142, 384, 38 },
		{ 142, 385, 38 },
		{ 142, 386, 38 },
		{ 142, 387, 38 },
		{ 142, 388, 38 },
		{ 143, 357, 38 },
		{ 143, 358, 38 },
		{ 143, 359, 38 },
		{ 143, 360, 38 },
		{ 143, 361, 38 },
		{ 143, 362, 38 },
		{ 143, 363, 38 },
		{ 143, 364, 38 },
		{ 143, 365, 38 },
		{ 143, 366, 38 },
		{ 143, 367, 38 },
		{ 143, 368, 38 },
		{ 143, 369, 38 },
		{ 143, 370, 38 },
		{ 143, 371, 38 },
		{ 143, 372, 38 },
		{ 143, 373, 38 },
		{ 143, 374, 38 },
		{ 143, 375, 38 },
		{ 143, 376, 38 },
		{ 143, 377, 38 },
		{ 143, 378, 38 },
		{ 143, 379, 38 },
		{ 143, 380, 38 },
		{ 143, 381, 38 },
		{ 143, 382, 38 },
		{ 143, 383, 38 },
		{ 143, 384, 38 },
		{ 143, 385, 38 },
		{ 143, 386, 38 },
		{ 143, 387, 38 },
		{ 144, 356, 38 },
		{ 144, 357, 38 },
		{ 144, 358, 38 },
		{ 144, 359, 38 },
		{ 144, 360, 38 },
		{ 144, 361, 38 },
		{ 144, 362, 38 },
		{ 144, 363, 38 },
		{ 144, 364, 38 },
		{ 144, 365, 38 },
		{ 144, 366, 38 },
		{ 144, 367, 38 },
		{ 144, 368, 38 },
		{ 144, 369, 38 },
		{ 144, 370, 38 },
		{ 144, 371, 38 },
		{ 144, 372, 38 },
		{ 144, 373, 38 },
		{ 144, 374, 38 },
		{ 144, 375, 38 },
		{ 144, 376, 38 },
		{ 144, 377, 38 },
		{ 144, 378, 38 },
		{ 144, 379, 38 },
		{ 144, 380, 38 },
		{ 144, 381, 38 },
		{ 144, 382, 38 },
		{ 144, 383, 38 },
		{ 144, 384, 38 },
		{ 144, 385, 38 },
		{ 144, 386, 38 },
		{ 145, 356, 38 },
		{ 145, 357, 38 },
		{ 145, 358, 38 },
		{ 145, 359, 38 },
		{ 145, 360, 38 },
		{ 145, 361, 38 },
		{ 145, 362, 38 },
		{ 145, 363, 38 },
		{ 145, 364, 38 },
		{ 145, 365, 38 },
		{ 145, 366, 38 },
		{ 145, 367, 38 },
		{ 145, 368, 38 },
		{ 145, 369, 38 },
		{ 145, 370, 38 },
		{ 145, 371, 38 },
		{ 145, 372, 38 },
		{ 145, 373, 38 },
		{ 145, 374, 38 },
		{ 145, 375, 38 },
		{ 145, 376, 38 },
		{ 145, 377, 38 },
		{ 145, 378, 38 },
		{ 145, 379, 38 },
		{ 145, 380, 38 },
		{ 145, 381, 38 },
		{ 145, 382, 38 },
		{ 145, 383, 38 },
		{ 145, 384, 38 },
		{ 146, 355, 38 },
		{ 146, 356, 38 },
		{ 146, 357, 38 },
		{ 146, 358, 38 },
		{ 146, 359, 38 },
		{ 146, 360, 38 },
		{ 146, 361, 38 },
		{ 146, 362, 38 },
		{ 146, 363, 38 },
		{ 146, 364, 38 },
		{ 146, 365, 38 },
		{ 146, 366, 38 },
		{ 146, 367, 38 },
		{ 146, 368, 38 },
		{ 146, 369, 38 },
		{ 146, 370, 38 },
		{ 146, 371, 38 },
		{ 146, 372, 38 },
		{ 146, 373, 38 },
		{ 146, 374, 38 },
		{ 146, 375, 38 },
		{ 146, 376, 38 },
		{ 146, 377, 38 },
		{ 146, 378, 38 },
		{ 146, 379, 38 },
		{ 146, 380, 38 },
		{ 146, 381, 38 },
		{ 146, 382, 38 },
		{ 146, 383, 38 },
		{ 147, 354, 38 },
		{ 147, 355, 38 },
		{ 147, 356, 38 },
		{ 147, 357, 38 },
		{ 147, 358, 38 },
		{ 147, 359, 38 },
		{ 147, 360, 38 },
		{ 147, 361, 38 },
		{ 147, 362, 38 },
		{ 147, 363, 38 },
		{ 147, 364, 38 },
		{ 147, 365, 38 },
		{ 147, 366, 38 },
		{ 147, 367, 38 },
		{ 147, 368, 38 },
		{ 147, 369, 38 },
		{ 147, 370, 38 },
		{ 147, 371, 38 },
		{ 147, 372, 38 },
		{ 147, 373, 38 },
		{ 147, 374, 38 },
		{ 147, 375, 38 },
		{ 147, 376, 38 },
		{ 147, 377, 38 },
		{ 147, 378, 38 },
		{ 147, 379, 38 },
		{ 147, 380, 38 },
		{ 147, 381, 38 },
		{ 147, 382, 38 },
		{ 148, 353, 38 },
		{ 148, 354, 38 },
		{ 148, 355, 38 },
		{ 148, 356, 38 },
		{ 148, 357, 38 },
		{ 148, 358, 38 },
		{ 148, 359, 38 },
		{ 148, 360, 38 },
		{ 148, 361, 38 },
		{ 148, 362, 38 },
		{ 148, 363, 38 },
		{ 148, 364, 38 },
		{ 148, 365, 38 },
		{ 148, 366, 38 },
		{ 148, 367, 38 },
		{ 148, 368, 38 },
		{ 148, 369, 38 },
		{ 148, 370, 38 },
		{ 148, 371, 38 },
		{ 148, 372, 38 },
		{ 148, 373, 38 },
		{ 148, 374, 38 },
		{ 148, 375, 38 },
		{ 148, 376, 38 },
		{ 148, 377, 38 },
		{ 148, 378, 38 },
		{ 148, 379, 38 },
		{ 148, 380, 38 },
		{ 148, 381, 38 },
		{ 149, 352, 38 },
		{ 149, 353, 38 },
		{ 149, 354, 38 },
		{ 149, 355, 38 },
		{ 149, 356, 38 },
		{ 149, 357, 38 },
		{ 149, 358, 38 },
		{ 149, 359, 38 },
		{ 149, 360, 38 },
		{ 149, 361, 38 },
		{ 149, 362, 38 },
		{ 149, 363, 38 },
		{ 149, 364, 38 },
		{ 149, 365, 38 },
		{ 149, 366, 38 },
		{ 149, 367, 38 },
		{ 149, 368, 38 },
		{ 149, 369, 38 },
		{ 149, 370, 38 },
		{ 149, 371, 38 },
		{ 149, 372, 38 },
		{ 149, 373, 38 },
		{ 149, 374, 38 },
		{ 149, 375, 38 },
		{ 149, 376, 38 },
		{ 149, 377, 38 },
		{ 149, 378, 38 },
		{ 149, 379, 38 },
		{ 149, 380, 38 },
		{ 150, 351, 38 },
		{ 150, 352, 38 },
		{ 150, 353, 38 },
		{ 150, 354, 38 },
		{ 150, 355, 38 },
		{ 150, 356, 38 },
		{ 150, 357, 38 },
		{ 150, 358, 38 },
		{ 150, 359, 38 },
		{ 150, 360, 38 },
		{ 150, 361, 38 },
		{ 150, 362, 38 },
		{ 150, 363, 38 },
		{ 150, 364, 38 },
		{ 150, 365, 38 },
		{ 150, 366, 38 },
		{ 150, 367, 38 },
		{ 150, 368, 38 },
		{ 150, 369, 38 },
		{ 150, 370, 38 },
		{ 150, 371, 38 },
		{ 150, 372, 38 },
		{ 150, 373, 38 },
		{ 150, 374, 38 },
		{ 150, 375, 38 },
		{ 150, 376, 38 },
		{ 150, 377, 38 },
		{ 150, 378, 38 },
		{ 150, 379, 38 },
		{ 151, 350, 38 },
		{ 151, 351, 38 },
		{ 151, 352, 38 },
		{ 151, 353, 38 },
		{ 151, 354, 38 },
		{ 151, 355, 38 },
		{ 151, 356, 38 },
		{ 151, 357, 38 },
		{ 151, 358, 38 },
		{ 151, 359, 38 },
		{ 151, 360, 38 },
		{ 151, 361, 38 },
		{ 151, 362, 38 },
		{ 151, 363, 38 },
		{ 151, 364, 38 },
		{ 151, 365, 38 },
		{ 151, 366, 38 },
		{ 151, 367, 38 },
		{ 151, 368, 38 },
		{ 151, 369, 38 },
		{ 151, 370, 38 },
		{ 151, 371, 38 },
		{ 151, 372, 38 },
		{ 151, 373, 38 },
		{ 151, 374, 38 },
		{ 151, 375, 38 },
		{ 151, 376, 38 },
		{ 151, 377, 38 },
		{ 151, 378, 38 },
		{ 152, 349, 38 },
		{ 152, 350, 38 },
		{ 152, 351, 38 },
		{ 152, 352, 38 },
		{ 152, 353, 38 },
		{ 152, 354, 38 },
		{ 152, 355, 38 },
		{ 152, 356, 38 },
		{ 152, 357, 38 },
		{ 152, 358, 38 },
		{ 152, 359, 38 },
		{ 152, 360, 38 },
		{ 152, 361, 38 },
		{ 152, 362, 38 },
		{ 152, 363, 38 },
		{ 152, 364, 38 },
		{ 152, 365, 38 },
		{ 152, 366, 38 },
		{ 152, 367, 38 },
		{ 152, 368, 38 },
		{ 152, 369, 38 },
		{ 152, 370, 38 },
		{ 152, 371, 38 },
		{ 152, 372, 38 },
		{ 152, 373, 38 },
		{ 152, 374, 38 },
		{ 152, 375, 38 },
		{ 152, 376, 38 },
		{ 153, 349, 38 },
		{ 153, 350, 38 },
		{ 153, 351, 38 },
		{ 153, 352, 38 },
		{ 153, 353, 38 },
		{ 153, 354, 38 },
		{ 153, 355, 38 },
		{ 153, 356, 38 },
		{ 153, 357, 38 },
		{ 153, 358, 38 },
		{ 153, 359, 38 },
		{ 153, 360, 38 },
		{ 153, 361, 38 },
		{ 153, 362, 38 },
		{ 153, 363, 38 },
		{ 153, 364, 38 },
		{ 153, 365, 38 },
		{ 153, 366, 38 },
		{ 153, 367, 38 },
		{ 153, 368, 38 },
		{ 153, 369, 38 },
		{ 153, 370, 38 },
		{ 153, 371, 38 },
		{ 153, 372, 38 },
		{ 153, 373, 38 },
		{ 153, 374, 38 },
		{ 153, 375, 38 },
		{ 154, 348, 38 },
		{ 154, 349, 38 },
		{ 154, 350, 38 },
		{ 154, 351, 38 },
		{ 154, 352, 38 },
		{ 154, 353, 38 },
		{ 154, 354, 38 },
		{ 154, 355, 38 },
		{ 154, 356, 38 },
		{ 154, 357, 38 },
		{ 154, 358, 38 },
		{ 154, 359, 38 },
		{ 154, 360, 38 },
		{ 154, 361, 38 },
		{ 154, 362, 38 },
		{ 154, 363, 38 },
		{ 154, 364, 38 },
		{ 154, 365, 38 },
		{ 154, 366, 38 },
		{ 154, 367, 38 },
		{ 154, 368, 38 },
		{ 154, 369, 38 },
		{ 154, 370, 38 },
		{ 154, 371, 38 },
		{ 154, 372, 38 },
		{ 154, 373, 38 },
		{ 154, 374, 38 },
		{ 155, 347, 38 },
		{ 155, 348, 38 },
		{ 155, 349, 38 },
		{ 155, 350, 38 },
		{ 155, 351, 38 },
		{ 155, 352, 38 },
		{ 155, 353, 38 },
		{ 155, 354, 38 },
		{ 155, 355, 38 },
		{ 155, 356, 38 },
		{ 155, 357, 38 },
		{ 155, 358, 38 },
		{ 155, 359, 38 },
		{ 155, 360, 38 },
		{ 155, 361, 38 },
		{ 155, 362, 38 },
		{ 155, 363, 38 },
		{ 155, 364, 38 },
		{ 155, 365, 38 },
		{ 155, 366, 38 },
		{ 155, 367, 38 },
		{ 155, 368, 38 },
		{ 155, 369, 38 },
		{ 155, 370, 38 },
		{ 155, 371, 38 },
		{ 155, 372, 38 },
		{ 155, 373, 38 },
		{ 156, 346, 38 },
		{ 156, 347, 38 },
		{ 156, 348, 38 },
		{ 156, 349, 38 },
		{ 156, 350, 38 },
		{ 156, 351, 38 },
		{ 156, 352, 38 },
		{ 156, 353, 38 },
		{ 156, 354, 38 },
		{ 156, 355, 38 },
		{ 156, 356, 38 },
		{ 156, 357, 38 },
		{ 156, 358, 38 },
		{ 156, 359, 38 },
		{ 156, 360, 38 },
		{ 156, 361, 38 },
		{ 156, 362, 38 },
		{ 156, 363, 38 },
		{ 156, 364, 38 },
		{ 156, 365, 38 },
		{ 156, 366, 38 },
		{ 156, 367, 38 },
		{ 156, 368, 38 },
		{ 156, 369, 38 },
		{ 156, 370, 38 },
		{ 156, 371, 38 },
		{ 156, 372, 38 },
		{ 157, 345, 38 },
		{ 157, 346, 38 },
		{ 157, 347, 38 },
		{ 157, 348, 38 },
		{ 157, 349, 38 },
		{ 157, 350, 38 },
		{ 157, 351, 38 },
		{ 157, 352, 38 },
		{ 157, 353, 38 },
		{ 157, 354, 38 },
		{ 157, 355, 38 },
		{ 157, 356, 38 },
		{ 157, 357, 38 },
		{ 157, 358, 38 },
		{ 157, 359, 38 },
		{ 157, 360, 38 },
		{ 157, 361, 38 },
		{ 157, 362, 38 },
		{ 157, 363, 38 },
		{ 157, 364, 38 },
		{ 157, 365, 38 },
		{ 157, 366, 38 },
		{ 157, 367, 38 },
		{ 157, 368, 38 },
		{ 157, 369, 38 },
		{ 157, 370, 38 },
		{ 157, 371, 38 },
		{ 158, 344, 38 },
		{ 158, 345, 38 },
		{ 158, 346, 38 },
		{ 158, 347, 38 },
		{ 158, 348, 38 },
		{ 158, 349, 38 },
		{ 158, 350, 38 },
		{ 158, 351, 38 },
		{ 158, 352, 38 },
		{ 158, 353, 38 },
		{ 158, 354, 38 },
		{ 158, 355, 38 },
		{ 158, 356, 38 },
		{ 158, 357, 38 },
		{ 158, 358, 38 },
		{ 158, 359, 38 },
		{ 158, 360, 38 },
		{ 158, 361, 38 },
		{ 158, 362, 38 },
		{ 158, 363, 38 },
		{ 158, 364, 38 },
		{ 158, 365, 38 },
		{ 158, 366, 38 },
		{ 158, 367, 38 },
		{ 158, 368, 38 },
		{ 158, 369, 38 },
		{ 158, 370, 38 },
		{ 159, 343, 38 },
		{ 159, 344, 38 },
		{ 159, 345, 38 },
		{ 159, 346, 38 },
		{ 159, 347, 38 },
		{ 159, 348, 38 },
		{ 159, 349, 38 },
		{ 159, 350, 38 },
		{ 159, 351, 38 },
		{ 159, 352, 38 },
		{ 159, 353, 38 },
		{ 159, 354, 38 },
		{ 159, 355, 38 },
		{ 159, 356, 38 },
		{ 159, 357, 38 },
		{ 159, 358, 38 },
		{ 159, 359, 38 },
		{ 159, 360, 38 },
		{ 159, 361, 38 },
		{ 159, 362, 38 },
		{ 159, 363, 38 },
		{ 159, 364, 38 },
		{ 159, 365, 38 },
		{ 159, 366, 38 },
		{ 159, 367, 38 },
		{ 159, 368, 38 },
		{ 160, 342, 38 },
		{ 160, 343, 38 },
		{ 160, 344, 38 },
		{ 160, 345, 38 },
		{ 160, 346, 38 },
		{ 160, 347, 38 },
		{ 160, 348, 38 },
		{ 160, 349, 38 },
		{ 160, 350, 38 },
		{ 160, 351, 38 },
		{ 160, 352, 38 },
		{ 160, 353, 38 },
		{ 160, 354, 38 },
		{ 160, 355, 38 },
		{ 160, 356, 38 },
		{ 160, 357, 38 },
		{ 160, 358, 38 },
		{ 160, 359, 38 },
		{ 160, 360, 38 },
		{ 160, 361, 38 },
		{ 160, 362, 38 },
		{ 160, 363, 38 },
		{ 160, 364, 38 },
		{ 160, 365, 38 },
		{ 160, 366, 38 },
		{ 160, 367, 38 },
		{ 161, 342, 38 },
		{ 161, 343, 38 },
		{ 161, 344, 38 },
		{ 161, 345, 38 },
		{ 161, 346, 38 },
		{ 161, 347, 38 },
		{ 161, 348, 38 },
		{ 161, 349, 38 },
		{ 161, 350, 38 },
		{ 161, 351, 38 },
		{ 161, 352, 38 },
		{ 161, 353, 38 },
		{ 161, 354, 38 },
		{ 161, 355, 38 },
		{ 161, 356, 38 },
		{ 161, 357, 38 },
		{ 161, 358, 38 },
		{ 161, 359, 38 },
		{ 161, 360, 38 },
		{ 161, 361, 38 },
		{ 161, 362, 38 },
		{ 161, 363, 38 },
		{ 161, 364, 38 },
		{ 161, 365, 38 },
		{ 161, 366, 38 },
		{ 162, 341, 38 },
		{ 162, 342, 38 },
		{ 162, 343, 38 },
		{ 162, 344, 38 },
		{ 162, 345, 38 },
		{ 162, 346, 38 },
		{ 162, 347, 38 },
		{ 162, 348, 38 },
		{ 162, 349, 38 },
		{ 162, 350, 38 },
		{ 162, 351, 38 },
		{ 162, 352, 38 },
		{ 162, 353, 38 },
		{ 162, 354, 38 },
		{ 162, 355, 38 },
		{ 162, 356, 38 },
		{ 162, 357, 38 },
		{ 162, 358, 38 },
		{ 162, 359, 38 },
		{ 162, 360, 38 },
		{ 162, 361, 38 },
		{ 162, 362, 38 },
		{ 162, 363, 38 },
		{ 162, 364, 38 },
		{ 162, 365, 38 },
		{ 163, 340, 38 },
		{ 163, 341, 38 },
		{ 163, 342, 38 },
		{ 163, 343, 38 },
		{ 163, 344, 38 },
		{ 163, 345, 38 },
		{ 163, 346, 38 },
		{ 163, 347, 38 },
		{ 163, 348, 38 },
		{ 163, 349, 38 },
		{ 163, 350, 38 },
		{ 163, 351, 38 },
		{ 163, 352, 38 },
		{ 163, 353, 38 },
		{ 163, 354, 38 },
		{ 163, 355, 38 },
		{ 163, 356, 38 },
		{ 163, 357, 38 },
		{ 163, 358, 38 },
		{ 163, 359, 38 },
		{ 163, 360, 38 },
		{ 163, 361, 38 },
		{ 163, 362, 38 },
		{ 163, 363, 38 },
		{ 163, 364, 38 },
		{ 164, 339, 38 },
		{ 164, 340, 38 },
		{ 164, 341, 38 },
		{ 164, 342, 38 },
		{ 164, 343, 38 },
		{ 164, 344, 38 },
		{ 164, 345, 38 },
		{ 164, 346, 38 },
		{ 164, 347, 38 },
		{ 164, 348, 38 },
		{ 164, 349, 38 },
		{ 164, 350, 38 },
		{ 164, 351, 38 },
		{ 164, 352, 38 },
		{ 164, 353, 38 },
		{ 164, 354, 38 },
		{ 164, 355, 38 },
		{ 164, 356, 38 },
		{ 164, 357, 38 },
		{ 164, 358, 38 },
		{ 164, 359, 38 },
		{ 164, 360, 38 },
		{ 164, 361, 38 },
		{ 164, 362, 38 },
		{ 164, 363, 38 },
		{ 165, 338, 38 },
		{ 165, 339, 38 },
		{ 165, 340, 38 },
		{ 165, 341, 38 },
		{ 165, 342, 38 },
		{ 165, 343, 38 },
		{ 165, 344, 38 },
		{ 165, 345, 38 },
		{ 165, 346, 38 },
		{ 165, 347, 38 },
		{ 165, 348, 38 },
		{ 165, 349, 38 },
		{ 165, 350, 38 },
		{ 165, 351, 38 },
		{ 165, 352, 38 },
		{ 165, 353, 38 },
		{ 165, 354, 38 },
		{ 165, 355, 38 },
		{ 165, 356, 38 },
		{ 165, 357, 38 },
		{ 165, 358, 38 },
		{ 165, 359, 38 },
		{ 165, 360, 38 },
		{ 165, 361, 38 },
		{ 165, 362, 38 },
		{ 166, 337, 38 },
		{ 166, 338, 38 },
		{ 166, 339, 38 },
		{ 166, 340, 38 },
		{ 166, 341, 38 },
		{ 166, 342, 38 },
		{ 166, 343, 38 },
		{ 166, 344, 38 },
		{ 166, 345, 38 },
		{ 166, 346, 38 },
		{ 166, 347, 38 },
		{ 166, 348, 38 },
		{ 166, 349, 38 },
		{ 166, 350, 38 },
		{ 166, 351, 38 },
		{ 166, 352, 38 },
		{ 166, 353, 38 },
		{ 166, 354, 38 },
		{ 166, 355, 38 },
		{ 166, 356, 38 },
		{ 166, 357, 38 },
		{ 166, 358, 38 },
		{ 166, 359, 38 },
		{ 166, 360, 38 },
		{ 167, 336, 38 },
		{ 167, 337, 38 },
		{ 167, 338, 38 },
		{ 167, 339, 38 },
		{ 167, 340, 38 },
		{ 167, 341, 38 },
		{ 167, 342, 38 },
		{ 167, 343, 38 },
		{ 167, 344, 38 },
		{ 167, 345, 38 },
		{ 167, 346, 38 },
		{ 167, 347, 38 },
		{ 167, 348, 38 },
		{ 167, 349, 38 },
		{ 167, 350, 38 },
		{ 167, 351, 38 },
		{ 167, 352, 38 },
		{ 167, 353, 38 },
		{ 167, 354, 38 },
		{ 167, 355, 38 },
		{ 167, 356, 38 },
		{ 167, 357, 38 },
		{ 167, 358, 38 },
		{ 167, 359, 38 },
		{ 168, 335, 38 },
		{ 168, 336, 38 },
		{ 168, 337, 38 },
		{ 168, 338, 38 },
		{ 168, 339, 38 },
		{ 168, 340, 38 },
		{ 168, 341, 38 },
		{ 168, 342, 38 },
		{ 168, 343, 38 },
		{ 168, 344, 38 },
		{ 168, 345, 38 },
		{ 168, 346, 38 },
		{ 168, 347, 38 },
		{ 168, 348, 38 },
		{ 168, 349, 38 },
		{ 168, 350, 38 },
		{ 168, 351, 38 },
		{ 168, 352, 38 },
		{ 168, 353, 38 },
		{ 168, 354, 38 },
		{ 168, 355, 38 },
		{ 168, 356, 38 },
		{ 168, 357, 38 },
		{ 168, 358, 38 },
		{ 169, 335, 38 },
		{ 169, 336, 38 },
		{ 169, 337, 38 },
		{ 169, 338, 38 },
		{ 169, 339, 38 },
		{ 169, 340, 38 },
		{ 169, 341, 38 },
		{ 169, 342, 38 },
		{ 169, 343, 38 },
		{ 169, 344, 38 },
		{ 169, 345, 38 },
		{ 169, 346, 38 },
		{ 169, 347, 38 },
		{ 169, 348, 38 },
		{ 169, 349, 38 },
		{ 169, 350, 38 },
		{ 169, 351, 38 },
		{ 169, 352, 38 },
		{ 169, 353, 38 },
		{ 169, 354, 38 },
		{ 169, 355, 38 },
		{ 169, 356, 38 },
		{ 169, 357, 38 },
		{ 170, 334, 38 },
		{ 170, 335, 38 },
		{ 170, 336, 38 },
		{ 170, 337, 38 },
		{ 170, 338, 38 },
		{ 170, 339, 38 },
		{ 170, 340, 38 },
		{ 170, 341, 38 },
		{ 170, 342, 38 },
		{ 170, 343, 38 },
		{ 170, 344, 38 },
		{ 170, 345, 38 },
		{ 170, 346, 38 },
		{ 170, 347, 38 },
		{ 170, 348, 38 },
		{ 170, 349, 38 },
		{ 170, 350, 38 },
		{ 170, 351, 38 },
		{ 170, 352, 38 },
		{ 170, 353, 38 },
		{ 170, 354, 38 },
		{ 170, 355, 38 },
		{ 170, 356, 38 },
		{ 171, 333, 38 },
		{ 171, 334, 38 },
		{ 171, 335, 38 },
		{ 171, 336, 38 },
		{ 171, 337, 38 },
		{ 171, 338, 38 },
		{ 171, 339, 38 },
		{ 171, 340, 38 },
		{ 171, 341, 38 },
		{ 171, 342, 38 },
		{ 171, 343, 38 },
		{ 171, 344, 38 },
		{ 171, 345, 38 },
		{ 171, 346, 38 },
		{ 171, 347, 38 },
		{ 171, 348, 38 },
		{ 171, 349, 38 },
		{ 171, 350, 38 },
		{ 171, 351, 38 },
		{ 171, 352, 38 },
		{ 171, 353, 38 },
		{ 171, 354, 38 },
		{ 171, 355, 38 },
		{ 172, 332, 38 },
		{ 172, 333, 38 },
		{ 172, 334, 38 },
		{ 172, 335, 38 },
		{ 172, 336, 38 },
		{ 172, 337, 38 },
		{ 172, 338, 38 },
		{ 172, 339, 38 },
		{ 172, 340, 38 },
		{ 172, 341, 38 },
		{ 172, 342, 38 },
		{ 172, 343, 38 },
		{ 172, 344, 38 },
		{ 172, 345, 38 },
		{ 172, 346, 38 },
		{ 172, 347, 38 },
		{ 172, 348, 38 },
		{ 172, 349, 38 },
		{ 172, 350, 38 },
		{ 172, 351, 38 },
		{ 172, 352, 38 },
		{ 172, 353, 38 },
		{ 172, 354, 38 },
		{ 173, 331, 38 },
		{ 173, 332, 38 },
		{ 173, 333, 38 },
		{ 173, 334, 38 },
		{ 173, 335, 38 },
		{ 173, 336, 38 },
		{ 173, 337, 38 },
		{ 173, 338, 38 },
		{ 173, 339, 38 },
		{ 173, 340, 38 },
		{ 173, 341, 38 },
		{ 173, 342, 38 },
		{ 173, 343, 38 },
		{ 173, 344, 38 },
		{ 173, 345, 38 },
		{ 173, 346, 38 },
		{ 173, 347, 38 },
		{ 173, 348, 38 },
		{ 173, 349, 38 },
		{ 173, 350, 38 },
		{ 173, 351, 38 },
		{ 173, 352, 38 },
		{ 174, 330, 38 },
		{ 174, 331, 38 },
		{ 174, 332, 38 },
		{ 174, 333, 38 },
		{ 174, 334, 38 },
		{ 174, 335, 38 },
		{ 174, 336, 38 },
		{ 174, 337, 38 },
		{ 174, 338, 38 },
		{ 174, 339, 38 },
		{ 174, 340, 38 },
		{ 174, 341, 38 },
		{ 174, 342, 38 },
		{ 174, 343, 38 },
		{ 174, 344, 38 },
		{ 174, 345, 38 },
		{ 174, 346, 38 },
		{ 174, 347, 38 },
		{ 174, 348, 38 },
		{ 174, 349, 38 },
		{ 174, 350, 38 },
		{ 174, 351, 38 },
		{ 175, 329, 38 },
		{ 175, 330, 38 },
		{ 175, 331, 38 },
		{ 175, 332, 38 },
		{ 175, 333, 38 },
		{ 175, 334, 38 },
		{ 175, 335, 38 },
		{ 175, 336, 38 },
		{ 175, 337, 38 },
		{ 175, 338, 38 },
		{ 175, 339, 38 },
		{ 175, 340, 38 },
		{ 175, 341, 38 },
		{ 175, 342, 38 },
		{ 175, 343, 38 },
		{ 175, 344, 38 },
		{ 175, 345, 38 },
		{ 175, 346, 38 },
		{ 175, 347, 38 },
		{ 175, 348, 38 },
		{ 175, 349, 38 },
		{ 175, 350, 38 },
		{ 176, 328, 38 },
		{ 176, 329, 38 },
		{ 176, 330, 38 },
		{ 176, 331, 38 },
		{ 176, 332, 38 },
		{ 176, 333, 38 },
		{ 176, 334, 38 },
		{ 176, 335, 38 },
		{ 176, 336, 38 },
		{ 176, 337, 38 },
		{ 176, 338, 38 },
		{ 176, 339, 38 },
		{ 176, 340, 38 },
		{ 176, 341, 38 },
		{ 176, 342, 38 },
		{ 176, 343, 38 },
		{ 176, 344, 38 },
		{ 176, 345, 38 },
		{ 176, 346, 38 },
		{ 176, 347, 38 },
		{ 176, 348, 38 },
		{ 176, 349, 38 },
		{ 177, 328, 38 },
		{ 177, 329, 38 },
		{ 177, 330, 38 },
		{ 177, 331, 38 },
		{ 177, 332, 38 },
		{ 177, 333, 38 },
		{ 177, 334, 38 },
		{ 177, 335, 38 },
		{ 177, 336, 38 },
		{ 177, 337, 38 },
		{ 177, 338, 38 },
		{ 177, 339, 38 },
		{ 177, 340, 38 },
		{ 177, 341, 38 },
		{ 177, 342, 38 },
		{ 177, 343, 38 },
		{ 177, 344, 38 },
		{ 177, 345, 38 },
		{ 177, 346, 38 },
		{ 177, 347, 38 },
		{ 177, 348, 38 },
		{ 178, 327, 38 },
		{ 178, 328, 38 },
		{ 178, 329, 38 },
		{ 178, 330, 38 },
		{ 178, 331, 38 },
		{ 178, 332, 38 },
		{ 178, 333, 38 },
		{ 178, 334, 38 },
		{ 178, 335, 38 },
		{ 178, 336, 38 },
		{ 178, 337, 38 },
		{ 178, 338, 38 },
		{ 178, 339, 38 },
		{ 178, 340, 38 },
		{ 178, 341, 38 },
		{ 178, 342, 38 },
		{ 178, 343, 38 },
		{ 178, 344, 38 },
		{ 178, 345, 38 },
		{ 178, 346, 38 },
		{ 178, 347, 38 },
		{ 179, 326, 38 },
		{ 179, 327, 38 },
		{ 179, 328, 38 },
		{ 179, 329, 38 },
		{ 179, 330, 38 },
		{ 179, 331, 38 },
		{ 179, 332, 38 },
		{ 179, 333, 38 },
		{ 179, 334, 38 },
		{ 179, 335, 38 },
		{ 179, 336, 38 },
		{ 179, 337, 38 },
		{ 179, 338, 38 },
		{ 179, 339, 38 },
		{ 179, 340, 38 },
		{ 179, 341, 38 },
		{ 179, 342, 38 },
		{ 179, 343, 38 },
		{ 179, 344, 38 },
		{ 179, 345, 38 },
		{ 179, 346, 38 },
		{ 180, 325, 38 },
		{ 180, 326, 38 },
		{ 180, 327, 38 },
		{ 180, 328, 38 },
		{ 180, 329, 38 },
		{ 180, 330, 38 },
		{ 180, 331, 38 },
		{ 180, 332, 38 },
		{ 180, 333, 38 },
		{ 180, 334, 38 },
		{ 180, 335, 38 },
		{ 180, 336, 38 },
		{ 180, 337, 38 },
		{ 180, 338, 38 },
		{ 180, 339, 38 },
		{ 180, 340, 38 },
		{ 180, 341, 38 },
		{ 180, 342, 38 },
		{ 180, 343, 38 },
		{ 180, 344, 38 },
		{ 181, 324, 38 },
		{ 181, 325, 38 },
		{ 181, 326, 38 },
		{ 181, 327, 38 },
		{ 181, 328, 38 },
		{ 181, 329, 38 },
		{ 181, 330, 38 },
		{ 181, 331, 38 },
		{ 181, 332, 38 },
		{ 181, 333, 38 },
		{ 181, 334, 38 },
		{ 181, 335, 38 },
		{ 181, 336, 38 },
		{ 181, 337, 38 },
		{ 181, 338, 38 },
		{ 181, 339, 38 },
		{ 181, 340, 38 },
		{ 181, 341, 38 },
		{ 181, 342, 38 },
		{ 181, 343, 38 },
		{ 182, 323, 38 },
		{ 182, 324, 38 },
		{ 182, 325, 38 },
		{ 182, 326, 38 },
		{ 182, 327, 38 },
		{ 182, 328, 38 },
		{ 182, 329, 38 },
		{ 182, 330, 38 },
		{ 182, 331, 38 },
		{ 182, 332, 38 },
		{ 182, 333, 38 },
		{ 182, 334, 38 },
		{ 182, 335, 38 },
		{ 182, 336, 38 },
		{ 182, 337, 38 },
		{ 182, 338, 38 },
		{ 182, 339, 38 },
		{ 182, 340, 38 },
		{ 182, 341, 38 },
		{ 182, 342, 38 },
		{ 183, 322, 38 },
		{ 183, 323, 38 },
		{ 183, 324, 38 },
		{ 183, 325, 38 },
		{ 183, 326, 38 },
		{ 183, 327, 38 },
		{ 183, 328, 38 },
		{ 183, 329, 38 },
		{ 183, 330, 38 },
		{ 183, 331, 38 },
		{ 183, 332, 38 },
		{ 183, 333, 38 },
		{ 183, 334, 38 },
		{ 183, 335, 38 },
		{ 183, 336, 38 },
		{ 183, 337, 38 },
		{ 183, 338, 38 },
		{ 183, 339, 38 },
		{ 183, 340, 38 },
		{ 183, 341, 38 },
		{ 184, 321, 38 },
		{ 184, 322, 38 },
		{ 184, 323, 38 },
		{ 184, 324, 38 },
		{ 184, 325, 38 },
		{ 184, 326, 38 },
		{ 184, 327, 38 },
		{ 184, 328, 38 },
		{ 184, 329, 38 },
		{ 184, 330, 38 },
		{ 184, 331, 38 },
		{ 184, 332, 38 },
		{ 184, 333, 38 },
		{ 184, 334, 38 },
		{ 184, 335, 38 },
		{ 184, 336, 38 },
		{ 184, 337, 38 },
		{ 184, 338, 38 },
		{ 184, 339, 38 },
		{ 184, 340, 38 },
		{ 185, 321, 38 },
		{ 185, 322, 38 },
		{ 185, 323, 38 },
		{ 185, 324, 38 },
		{ 185, 325, 38 },
		{ 185, 326, 38 },
		{ 185, 327, 38 },
		{ 185, 328, 38 },
		{ 185, 329, 38 },
		{ 185, 330, 38 },
		{ 185, 331, 38 },
		{ 185, 332, 38 },
		{ 185, 333, 38 },
		{ 185, 334, 38 },
		{ 185, 335, 38 },
		{ 185, 336, 38 },
		{ 185, 337, 38 },
		{ 185, 338, 38 },
		{ 185, 339, 38 },
		{ 186, 320, 38 },
		{ 186, 321, 38 },
		{ 186, 322, 38 },
		{ 186, 323, 38 },
		{ 186, 324, 38 },
		{ 186, 325, 38 },
		{ 186, 326, 38 },
		{ 186, 327, 38 },
		{ 186, 328, 38 },
		{ 186, 329, 38 },
		{ 186, 330, 38 },
		{ 186, 331, 38 },
		{ 186, 332, 38 },
		{ 186, 333, 38 },
		{ 186, 334, 38 },
		{ 186, 335, 38 },
		{ 186, 336, 38 },
		{ 186, 337, 38 },
		{ 186, 338, 38 },
		{ 187, 319, 38 },
		{ 187, 320, 38 },
		{ 187, 321, 38 },
		{ 187, 322, 38 },
		{ 187, 323, 38 },
		{ 187, 324, 38 },
		{ 187, 325, 38 },
		{ 187, 326, 38 },
		{ 187, 327, 38 },
		{ 187, 328, 38 },
		{ 187, 329, 38 },
		{ 187, 330, 38 },
		{ 187, 331, 38 },
		{ 187, 332, 38 },
		{ 187, 333, 38 },
		{ 187, 334, 38 },
		{ 187, 335, 38 },
		{ 187, 336, 38 },
		{ 188, 318, 38 },
		{ 188, 319, 38 },
		{ 188, 320, 38 },
		{ 188, 321, 38 },
		{ 188, 322, 38 },
		{ 188, 323, 38 },
		{ 188, 324, 38 },
		{ 188, 325, 38 },
		{ 188, 326, 38 },
		{ 188, 327, 38 },
		{ 188, 328, 38 },
		{ 188, 329, 38 },
		{ 188, 330, 38 },
		{ 188, 331, 38 },
		{ 188, 332, 38 },
		{ 188, 333, 38 },
		{ 188, 334, 38 },
		{ 188, 335, 38 },
		{ 189, 317, 38 },
		{ 189, 318, 38 },
		{ 189, 319, 38 },
		{ 189, 320, 38 },
		{ 189, 321, 38 },
		{ 189, 322, 38 },
		{ 189, 323, 38 },
		{ 189, 324, 38 },
		{ 189, 325, 38 },
		{ 189, 326, 38 },
		{ 189, 327, 38 },
		{ 189, 328, 38 },
		{ 189, 329, 38 },
		{ 189, 330, 38 },
		{ 189, 331, 38 },
		{ 189, 332, 38 },
		{ 189, 333, 38 },
		{ 189, 334, 38 },
		{ 190, 317, 38 },
		{ 190, 318, 38 },
		{ 190, 319, 38 },
		{ 190, 320, 38 },
		{ 190, 321, 38 },
		{ 190, 322, 38 },
		{ 190, 323, 38 },
		{ 190, 324, 38 },
		{ 190, 325, 38 },
		{ 190, 326, 38 },
		{ 190, 327, 38 },
		{ 190, 328, 38 },
		{ 190, 329, 38 },
		{ 190, 330, 38 },
		{ 190, 331, 38 },
		{ 190, 332, 38 },
		{ 190, 333, 38 },
		{ 191, 317, 38 },
		{ 191, 318, 38 },
		{ 191, 319, 38 },
		{ 191, 320, 38 },
		{ 191, 321, 38 },
		{ 191, 322, 38 },
		{ 191, 323, 38 },
		{ 191, 324, 38 },
		{ 191, 325, 38 },
		{ 191, 326, 38 },
		{ 191, 327, 38 },
		{ 191, 328, 38 },
		{ 191, 329, 38 },
		{ 191, 330, 38 },
		{ 191, 331, 38 },
		{ 191, 332, 38 },
		{ 192, 323, 38 },
		{ 192, 324, 38 },
		{ 192, 325, 38 },
		{ 192, 326, 38 },
		{ 192, 327, 38 },
		{ 192, 328, 38 },
		{ 192, 329, 38 },
		{ 192, 330, 38 },
		{ 192, 331, 38 },
		{ 193, 323, 38 },
		{ 193, 324, 38 },
		{ 193, 325, 38 },
		{ 193, 326, 38 },
		{ 193, 327, 38 },
		{ 193, 328, 38 },
		{ 193, 329, 38 },
		{ 193, 330, 38 },
		{ 194, 323, 38 },
		{ 194, 324, 38 },
		{ 194, 325, 38 },
		{ 194, 326, 38 },
		{ 194, 327, 38 },
		{ 194, 328, 38 },
		{ 195, 323, 38 },
		{ 195, 324, 38 },
		{ 195, 325, 38 },
		{ 195, 326, 38 },
		{ 195, 327, 38 },
		{ 196, 323, 38 },
		{ 196, 324, 38 },
		{ 196, 325, 38 },
		{ 196, 326, 38 },
		{ 197, 323, 38 },
		{ 197, 324, 38 },
		{ 197, 325, 38 },
		{ 317, 189, 38 },
		{ 317, 190, 38 },
		{ 317, 191, 38 },
		{ 318, 188, 38 },
		{ 318, 189, 38 },
		{ 318, 190, 38 },
		{ 318, 191, 38 },
		{ 319, 187, 38 },
		{ 319, 188, 38 },
		{ 319, 189, 38 },
		{ 319, 190, 38 },
		{ 319, 191, 38 },
		{ 320, 186, 38 },
		{ 320, 187, 38 },
		{ 320, 188, 38 },
		{ 320, 189, 38 },
		{ 320, 190, 38 },
		{ 320, 191, 38 },
		{ 321, 184, 38 },
		{ 321, 185, 38 },
		{ 321, 186, 38 },
		{ 321, 187, 38 },
		{ 321, 188, 38 },
		{ 321, 189, 38 },
		{ 321, 190, 38 },
		{ 321, 191, 38 },
		{ 322, 183, 38 },
		{ 322, 184, 38 },
		{ 322, 185, 38 },
		{ 322, 186, 38 },
		{ 322, 187, 38 },
		{ 322, 188, 38 },
		{ 322, 189, 38 },
		{ 322, 190, 38 },
		{ 322, 191, 38 },
		{ 323, 182, 38 },
		{ 323, 183, 38 },
		{ 323, 184, 38 },
		{ 323, 185, 38 },
		{ 323, 186, 38 },
		{ 323, 187, 38 },
		{ 323, 188, 38 },
		{ 323, 189, 38 },
		{ 323, 190, 38 },
		{ 323, 191, 38 },
		{ 323, 192, 38 },
		{ 323, 193, 38 },
		{ 323, 194, 38 },
		{ 323, 195, 38 },
		{ 323, 196, 38 },
		{ 323, 197, 38 },
		{ 324, 181, 38 },
		{ 324, 182, 38 },
		{ 324, 183, 38 },
		{ 324, 184, 38 },
		{ 324, 185, 38 },
		{ 324, 186, 38 },
		{ 324, 187, 38 },
		{ 324, 188, 38 },
		{ 324, 189, 38 },
		{ 324, 190, 38 },
		{ 324, 191, 38 },
		{ 324, 192, 38 },
		{ 324, 193, 38 },
		{ 324, 194, 38 },
		{ 324, 195, 38 },
		{ 324, 196, 38 },
		{ 324, 197, 38 },
		{ 325, 180, 38 },
		{ 325, 181, 38 },
		{ 325, 182, 38 },
		{ 325, 183, 38 },
		{ 325, 184, 38 },
		{ 325, 185, 38 },
		{ 325, 186, 38 },
		{ 325, 187, 38 },
		{ 325, 188, 38 },
		{ 325, 189, 38 },
		{ 325, 190, 38 },
		{ 325, 191, 38 },
		{ 325, 192, 38 },
		{ 325, 193, 38 },
		{ 325, 194, 38 },
		{ 325, 195, 38 },
		{ 325, 196, 38 },
		{ 325, 197, 38 },
		{ 326, 179, 38 },
		{ 326, 180, 38 },
		{ 326, 181, 38 },
		{ 326, 182, 38 },
		{ 326, 183, 38 },
		{ 326, 184, 38 },
		{ 326, 185, 38 },
		{ 326, 186, 38 },
		{ 326, 187, 38 },
		{ 326, 188, 38 },
		{ 326, 189, 38 },
		{ 326, 190, 38 },
		{ 326, 191, 38 },
		{ 326, 192, 38 },
		{ 326, 193, 38 },
		{ 326, 194, 38 },
		{ 326, 195, 38 },
		{ 326, 196, 38 },
		{ 327, 178, 38 },
		{ 327, 179, 38 },
		{ 327, 180, 38 },
		{ 327, 181, 38 },
		{ 327, 182, 38 },
		{ 327, 183, 38 },
		{ 327, 184, 38 },
		{ 327, 185, 38 },
		{ 327, 186, 38 },
		{ 327, 187, 38 },
		{ 327, 188, 38 },
		{ 327, 189, 38 },
		{ 327, 190, 38 },
		{ 327, 191, 38 },
		{ 327, 192, 38 },
		{ 327, 193, 38 },
		{ 327, 194, 38 },
		{ 327, 195, 38 },
		{ 328, 176, 38 },
		{ 328, 177, 38 },
		{ 328, 178, 38 },
		{ 328, 179, 38 },
		{ 328, 180, 38 },
		{ 328, 181, 38 },
		{ 328, 182, 38 },
		{ 328, 183, 38 },
		{ 328, 184, 38 },
		{ 328, 185, 38 },
		{ 328, 186, 38 },
		{ 328, 187, 38 },
		{ 328, 188, 38 },
		{ 328, 189, 38 },
		{ 328, 190, 38 },
		{ 328, 191, 38 },
		{ 328, 192, 38 },
		{ 328, 193, 38 },
		{ 328, 194, 38 },
		{ 329, 175, 38 },
		{ 329, 176, 38 },
		{ 329, 177, 38 },
		{ 329, 178, 38 },
		{ 329, 179, 38 },
		{ 329, 180, 38 },
		{ 329, 181, 38 },
		{ 329, 182, 38 },
		{ 329, 183, 38 },
		{ 329, 184, 38 },
		{ 329, 185, 38 },
		{ 329, 186, 38 },
		{ 329, 187, 38 },
		{ 329, 188, 38 },
		{ 329, 189, 38 },
		{ 329, 190, 38 },
		{ 329, 191, 38 },
		{ 329, 192, 38 },
		{ 329, 193, 38 },
		{ 330, 174, 38 },
		{ 330, 175, 38 },
		{ 330, 176, 38 },
		{ 330, 177, 38 },
		{ 330, 178, 38 },
		{ 330, 179, 38 },
		{ 330, 180, 38 },
		{ 330, 181, 38 },
		{ 330, 182, 38 },
		{ 330, 183, 38 },
		{ 330, 184, 38 },
		{ 330, 185, 38 },
		{ 330, 186, 38 },
		{ 330, 187, 38 },
		{ 330, 188, 38 },
		{ 330, 189, 38 },
		{ 330, 190, 38 },
		{ 330, 191, 38 },
		{ 330, 192, 38 },
		{ 330, 193, 38 },
		{ 331, 173, 38 },
		{ 331, 174, 38 },
		{ 331, 175, 38 },
		{ 331, 176, 38 },
		{ 331, 177, 38 },
		{ 331, 178, 38 },
		{ 331, 179, 38 },
		{ 331, 180, 38 },
		{ 331, 181, 38 },
		{ 331, 182, 38 },
		{ 331, 183, 38 },
		{ 331, 184, 38 },
		{ 331, 185, 38 },
		{ 331, 186, 38 },
		{ 331, 187, 38 },
		{ 331, 188, 38 },
		{ 331, 189, 38 },
		{ 331, 190, 38 },
		{ 331, 191, 38 },
		{ 331, 192, 38 },
		{ 332, 172, 38 },
		{ 332, 173, 38 },
		{ 332, 174, 38 },
		{ 332, 175, 38 },
		{ 332, 176, 38 },
		{ 332, 177, 38 },
		{ 332, 178, 38 },
		{ 332, 179, 38 },
		{ 332, 180, 38 },
		{ 332, 181, 38 },
		{ 332, 182, 38 },
		{ 332, 183, 38 },
		{ 332, 184, 38 },
		{ 332, 185, 38 },
		{ 332, 186, 38 },
		{ 332, 187, 38 },
		{ 332, 188, 38 },
		{ 332, 189, 38 },
		{ 332, 190, 38 },
		{ 332, 191, 38 },
		{ 333, 171, 38 },
		{ 333, 172, 38 },
		{ 333, 173, 38 },
		{ 333, 174, 38 },
		{ 333, 175, 38 },
		{ 333, 176, 38 },
		{ 333, 177, 38 },
		{ 333, 178, 38 },
		{ 333, 179, 38 },
		{ 333, 180, 38 },
		{ 333, 181, 38 },
		{ 333, 182, 38 },
		{ 333, 183, 38 },
		{ 333, 184, 38 },
		{ 333, 185, 38 },
		{ 333, 186, 38 },
		{ 333, 187, 38 },
		{ 333, 188, 38 },
		{ 333, 189, 38 },
		{ 333, 190, 38 },
		{ 334, 170, 38 },
		{ 334, 171, 38 },
		{ 334, 172, 38 },
		{ 334, 173, 38 },
		{ 334, 174, 38 },
		{ 334, 175, 38 },
		{ 334, 176, 38 },
		{ 334, 177, 38 },
		{ 334, 178, 38 },
		{ 334, 179, 38 },
		{ 334, 180, 38 },
		{ 334, 181, 38 },
		{ 334, 182, 38 },
		{ 334, 183, 38 },
		{ 334, 184, 38 },
		{ 334, 185, 38 },
		{ 334, 186, 38 },
		{ 334, 187, 38 },
		{ 334, 188, 38 },
		{ 334, 189, 38 },
		{ 335, 168, 38 },
		{ 335, 169, 38 },
		{ 335, 170, 38 },
		{ 335, 171, 38 },
		{ 335, 172, 38 },
		{ 335, 173, 38 },
		{ 335, 174, 38 },
		{ 335, 175, 38 },
		{ 335, 176, 38 },
		{ 335, 177, 38 },
		{ 335, 178, 38 },
		{ 335, 179, 38 },
		{ 335, 180, 38 },
		{ 335, 181, 38 },
		{ 335, 182, 38 },
		{ 335, 183, 38 },
		{ 335, 184, 38 },
		{ 335, 185, 38 },
		{ 335, 186, 38 },
		{ 335, 187, 38 },
		{ 335, 188, 38 },
		{ 336, 167, 38 },
		{ 336, 168, 38 },
		{ 336, 169, 38 },
		{ 336, 170, 38 },
		{ 336, 171, 38 },
		{ 336, 172, 38 },
		{ 336, 173, 38 },
		{ 336, 174, 38 },
		{ 336, 175, 38 },
		{ 336, 176, 38 },
		{ 336, 177, 38 },
		{ 336, 178, 38 },
		{ 336, 179, 38 },
		{ 336, 180, 38 },
		{ 336, 181, 38 },
		{ 336, 182, 38 },
		{ 336, 183, 38 },
		{ 336, 184, 38 },
		{ 336, 185, 38 },
		{ 336, 186, 38 },
		{ 336, 187, 38 },
		{ 337, 166, 38 },
		{ 337, 167, 38 },
		{ 337, 168, 38 },
		{ 337, 169, 38 },
		{ 337, 170, 38 },
		{ 337, 171, 38 },
		{ 337, 172, 38 },
		{ 337, 173, 38 },
		{ 337, 174, 38 },
		{ 337, 175, 38 },
		{ 337, 176, 38 },
		{ 337, 177, 38 },
		{ 337, 178, 38 },
		{ 337, 179, 38 },
		{ 337, 180, 38 },
		{ 337, 181, 38 },
		{ 337, 182, 38 },
		{ 337, 183, 38 },
		{ 337, 184, 38 },
		{ 337, 185, 38 },
		{ 337, 186, 38 },
		{ 338, 165, 38 },
		{ 338, 166, 38 },
		{ 338, 167, 38 },
		{ 338, 168, 38 },
		{ 338, 169, 38 },
		{ 338, 170, 38 },
		{ 338, 171, 38 },
		{ 338, 172, 38 },
		{ 338, 173, 38 },
		{ 338, 174, 38 },
		{ 338, 175, 38 },
		{ 338, 176, 38 },
		{ 338, 177, 38 },
		{ 338, 178, 38 },
		{ 338, 179, 38 },
		{ 338, 180, 38 },
		{ 338, 181, 38 },
		{ 338, 182, 38 },
		{ 338, 183, 38 },
		{ 338, 184, 38 },
		{ 338, 185, 38 },
		{ 338, 186, 38 },
		{ 339, 164, 38 },
		{ 339, 165, 38 },
		{ 339, 166, 38 },
		{ 339, 167, 38 },
		{ 339, 168, 38 },
		{ 339, 169, 38 },
		{ 339, 170, 38 },
		{ 339, 171, 38 },
		{ 339, 172, 38 },
		{ 339, 173, 38 },
		{ 339, 174, 38 },
		{ 339, 175, 38 },
		{ 339, 176, 38 },
		{ 339, 177, 38 },
		{ 339, 178, 38 },
		{ 339, 179, 38 },
		{ 339, 180, 38 },
		{ 339, 181, 38 },
		{ 339, 182, 38 },
		{ 339, 183, 38 },
		{ 339, 184, 38 },
		{ 339, 185, 38 },
		{ 340, 163, 38 },
		{ 340, 164, 38 },
		{ 340, 165, 38 },
		{ 340, 166, 38 },
		{ 340, 167, 38 },
		{ 340, 168, 38 },
		{ 340, 169, 38 },
		{ 340, 170, 38 },
		{ 340, 171, 38 },
		{ 340, 172, 38 },
		{ 340, 173, 38 },
		{ 340, 174, 38 },
		{ 340, 175, 38 },
		{ 340, 176, 38 },
		{ 340, 177, 38 },
		{ 340, 178, 38 },
		{ 340, 179, 38 },
		{ 340, 180, 38 },
		{ 340, 181, 38 },
		{ 340, 182, 38 },
		{ 340, 183, 38 },
		{ 340, 184, 38 },
		{ 341, 162, 38 },
		{ 341, 163, 38 },
		{ 341, 164, 38 },
		{ 341, 165, 38 },
		{ 341, 166, 38 },
		{ 341, 167, 38 },
		{ 341, 168, 38 },
		{ 341, 169, 38 },
		{ 341, 170, 38 },
		{ 341, 171, 38 },
		{ 341, 172, 38 },
		{ 341, 173, 38 },
		{ 341, 174, 38 },
		{ 341, 175, 38 },
		{ 341, 176, 38 },
		{ 341, 177, 38 },
		{ 341, 178, 38 },
		{ 341, 179, 38 },
		{ 341, 180, 38 },
		{ 341, 181, 38 },
		{ 341, 182, 38 },
		{ 341, 183, 38 },
		{ 342, 160, 38 },
		{ 342, 161, 38 },
		{ 342, 162, 38 },
		{ 342, 163, 38 },
		{ 342, 164, 38 },
		{ 342, 165, 38 },
		{ 342, 166, 38 },
		{ 342, 167, 38 },
		{ 342, 168, 38 },
		{ 342, 169, 38 },
		{ 342, 170, 38 },
		{ 342, 171, 38 },
		{ 342, 172, 38 },
		{ 342, 173, 38 },
		{ 342, 174, 38 },
		{ 342, 175, 38 },
		{ 342, 176, 38 },
		{ 342, 177, 38 },
		{ 342, 178, 38 },
		{ 342, 179, 38 },
		{ 342, 180, 38 },
		{ 342, 181, 38 },
		{ 342, 182, 38 },
		{ 343, 159, 38 },
		{ 343, 160, 38 },
		{ 343, 161, 38 },
		{ 343, 162, 38 },
		{ 343, 163, 38 },
		{ 343, 164, 38 },
		{ 343, 165, 38 },
		{ 343, 166, 38 },
		{ 343, 167, 38 },
		{ 343, 168, 38 },
		{ 343, 169, 38 },
		{ 343, 170, 38 },
		{ 343, 171, 38 },
		{ 343, 172, 38 },
		{ 343, 173, 38 },
		{ 343, 174, 38 },
		{ 343, 175, 38 },
		{ 343, 176, 38 },
		{ 343, 177, 38 },
		{ 343, 178, 38 },
		{ 343, 179, 38 },
		{ 343, 180, 38 },
		{ 343, 181, 38 },
		{ 344, 158, 38 },
		{ 344, 159, 38 },
		{ 344, 160, 38 },
		{ 344, 161, 38 },
		{ 344, 162, 38 },
		{ 344, 163, 38 },
		{ 344, 164, 38 },
		{ 344, 165, 38 },
		{ 344, 166, 38 },
		{ 344, 167, 38 },
		{ 344, 168, 38 },
		{ 344, 169, 38 },
		{ 344, 170, 38 },
		{ 344, 171, 38 },
		{ 344, 172, 38 },
		{ 344, 173, 38 },
		{ 344, 174, 38 },
		{ 344, 175, 38 },
		{ 344, 176, 38 },
		{ 344, 177, 38 },
		{ 344, 178, 38 },
		{ 344, 179, 38 },
		{ 344, 180, 38 },
		{ 345, 157, 38 },
		{ 345, 158, 38 },
		{ 345, 159, 38 },
		{ 345, 160, 38 },
		{ 345, 161, 38 },
		{ 345, 162, 38 },
		{ 345, 163, 38 },
		{ 345, 164, 38 },
		{ 345, 165, 38 },
		{ 345, 166, 38 },
		{ 345, 167, 38 },
		{ 345, 168, 38 },
		{ 345, 169, 38 },
		{ 345, 170, 38 },
		{ 345, 171, 38 },
		{ 345, 172, 38 },
		{ 345, 173, 38 },
		{ 345, 174, 38 },
		{ 345, 175, 38 },
		{ 345, 176, 38 },
		{ 345, 177, 38 },
		{ 345, 178, 38 },
		{ 345, 179, 38 },
		{ 346, 156, 38 },
		{ 346, 157, 38 },
		{ 346, 158, 38 },
		{ 346, 159, 38 },
		{ 346, 160, 38 },
		{ 346, 161, 38 },
		{ 346, 162, 38 },
		{ 346, 163, 38 },
		{ 346, 164, 38 },
		{ 346, 165, 38 },
		{ 346, 166, 38 },
		{ 346, 167, 38 },
		{ 346, 168, 38 },
		{ 346, 169, 38 },
		{ 346, 170, 38 },
		{ 346, 171, 38 },
		{ 346, 172, 38 },
		{ 346, 173, 38 },
		{ 346, 174, 38 },
		{ 346, 175, 38 },
		{ 346, 176, 38 },
		{ 346, 177, 38 },
		{ 346, 178, 38 },
		{ 346, 179, 38 },
		{ 347, 155, 38 },
		{ 347, 156, 38 },
		{ 347, 157, 38 },
		{ 347, 158, 38 },
		{ 347, 159, 38 },
		{ 347, 160, 38 },
		{ 347, 161, 38 },
		{ 347, 162, 38 },
		{ 347, 163, 38 },
		{ 347, 164, 38 },
		{ 347, 165, 38 },
		{ 347, 166, 38 },
		{ 347, 167, 38 },
		{ 347, 168, 38 },
		{ 347, 169, 38 },
		{ 347, 170, 38 },
		{ 347, 171, 38 },
		{ 347, 172, 38 },
		{ 347, 173, 38 },
		{ 347, 174, 38 },
		{ 347, 175, 38 },
		{ 347, 176, 38 },
		{ 347, 177, 38 },
		{ 347, 178, 38 },
		{ 348, 154, 38 },
		{ 348, 155, 38 },
		{ 348, 156, 38 },
		{ 348, 157, 38 },
		{ 348, 158, 38 },
		{ 348, 159, 38 },
		{ 348, 160, 38 },
		{ 348, 161, 38 },
		{ 348, 162, 38 },
		{ 348, 163, 38 },
		{ 348, 164, 38 },
		{ 348, 165, 38 },
		{ 348, 166, 38 },
		{ 348, 167, 38 },
		{ 348, 168, 38 },
		{ 348, 169, 38 },
		{ 348, 170, 38 },
		{ 348, 171, 38 },
		{ 348, 172, 38 },
		{ 348, 173, 38 },
		{ 348, 174, 38 },
		{ 348, 175, 38 },
		{ 348, 176, 38 },
		{ 348, 177, 38 },
		{ 349, 152, 38 },
		{ 349, 153, 38 },
		{ 349, 154, 38 },
		{ 349, 155, 38 },
		{ 349, 156, 38 },
		{ 349, 157, 38 },
		{ 349, 158, 38 },
		{ 349, 159, 38 },
		{ 349, 160, 38 },
		{ 349, 161, 38 },
		{ 349, 162, 38 },
		{ 349, 163, 38 },
		{ 349, 164, 38 },
		{ 349, 165, 38 },
		{ 349, 166, 38 },
		{ 349, 167, 38 },
		{ 349, 168, 38 },
		{ 349, 169, 38 },
		{ 349, 170, 38 },
		{ 349, 171, 38 },
		{ 349, 172, 38 },
		{ 349, 173, 38 },
		{ 349, 174, 38 },
		{ 349, 175, 38 },
		{ 349, 176, 38 },
		{ 350, 151, 38 },
		{ 350, 152, 38 },
		{ 350, 153, 38 },
		{ 350, 154, 38 },
		{ 350, 155, 38 },
		{ 350, 156, 38 },
		{ 350, 157, 38 },
		{ 350, 158, 38 },
		{ 350, 159, 38 },
		{ 350, 160, 38 },
		{ 350, 161, 38 },
		{ 350, 162, 38 },
		{ 350, 163, 38 },
		{ 350, 164, 38 },
		{ 350, 165, 38 },
		{ 350, 166, 38 },
		{ 350, 167, 38 },
		{ 350, 168, 38 },
		{ 350, 169, 38 },
		{ 350, 170, 38 },
		{ 350, 171, 38 },
		{ 350, 172, 38 },
		{ 350, 173, 38 },
		{ 350, 174, 38 },
		{ 350, 175, 38 },
		{ 351, 150, 38 },
		{ 351, 151, 38 },
		{ 351, 152, 38 },
		{ 351, 153, 38 },
		{ 351, 154, 38 },
		{ 351, 155, 38 },
		{ 351, 156, 38 },
		{ 351, 157, 38 },
		{ 351, 158, 38 },
		{ 351, 159, 38 },
		{ 351, 160, 38 },
		{ 351, 161, 38 },
		{ 351, 162, 38 },
		{ 351, 163, 38 },
		{ 351, 164, 38 },
		{ 351, 165, 38 },
		{ 351, 166, 38 },
		{ 351, 167, 38 },
		{ 351, 168, 38 },
		{ 351, 169, 38 },
		{ 351, 170, 38 },
		{ 351, 171, 38 },
		{ 351, 172, 38 },
		{ 351, 173, 38 },
		{ 351, 174, 38 },
		{ 352, 149, 38 },
		{ 352, 150, 38 },
		{ 352, 151, 38 },
		{ 352, 152, 38 },
		{ 352, 153, 38 },
		{ 352, 154, 38 },
		{ 352, 155, 38 },
		{ 352, 156, 38 },
		{ 352, 157, 38 },
		{ 352, 158, 38 },
		{ 352, 159, 38 },
		{ 352, 160, 38 },
		{ 352, 161, 38 },
		{ 352, 162, 38 },
		{ 352, 163, 38 },
		{ 352, 164, 38 },
		{ 352, 165, 38 },
		{ 352, 166, 38 },
		{ 352, 167, 38 },
		{ 352, 168, 38 },
		{ 352, 169, 38 },
		{ 352, 170, 38 },
		{ 352, 171, 38 },
		{ 352, 172, 38 },
		{ 352, 173, 38 },
		{ 353, 148, 38 },
		{ 353, 149, 38 },
		{ 353, 150, 38 },
		{ 353, 151, 38 },
		{ 353, 152, 38 },
		{ 353, 153, 38 },
		{ 353, 154, 38 },
		{ 353, 155, 38 },
		{ 353, 156, 38 },
		{ 353, 157, 38 },
		{ 353, 158, 38 },
		{ 353, 159, 38 },
		{ 353, 160, 38 },
		{ 353, 161, 38 },
		{ 353, 162, 38 },
		{ 353, 163, 38 },
		{ 353, 164, 38 },
		{ 353, 165, 38 },
		{ 353, 166, 38 },
		{ 353, 167, 38 },
		{ 353, 168, 38 },
		{ 353, 169, 38 },
		{ 353, 170, 38 },
		{ 353, 171, 38 },
		{ 353, 172, 38 },
		{ 354, 147, 38 },
		{ 354, 148, 38 },
		{ 354, 149, 38 },
		{ 354, 150, 38 },
		{ 354, 151, 38 },
		{ 354, 152, 38 },
		{ 354, 153, 38 },
		{ 354, 154, 38 },
		{ 354, 155, 38 },
		{ 354, 156, 38 },
		{ 354, 157, 38 },
		{ 354, 158, 38 },
		{ 354, 159, 38 },
		{ 354, 160, 38 },
		{ 354, 161, 38 },
		{ 354, 162, 38 },
		{ 354, 163, 38 },
		{ 354, 164, 38 },
		{ 354, 165, 38 },
		{ 354, 166, 38 },
		{ 354, 167, 38 },
		{ 354, 168, 38 },
		{ 354, 169, 38 },
		{ 354, 170, 38 },
		{ 354, 171, 38 },
		{ 354, 172, 38 },
		{ 355, 146, 38 },
		{ 355, 147, 38 },
		{ 355, 148, 38 },
		{ 355, 149, 38 },
		{ 355, 150, 38 },
		{ 355, 151, 38 },
		{ 355, 152, 38 },
		{ 355, 153, 38 },
		{ 355, 154, 38 },
		{ 355, 155, 38 },
		{ 355, 156, 38 },
		{ 355, 157, 38 },
		{ 355, 158, 38 },
		{ 355, 159, 38 },
		{ 355, 160, 38 },
		{ 355, 161, 38 },
		{ 355, 162, 38 },
		{ 355, 163, 38 },
		{ 355, 164, 38 },
		{ 355, 165, 38 },
		{ 355, 166, 38 },
		{ 355, 167, 38 },
		{ 355, 168, 38 },
		{ 355, 169, 38 },
		{ 355, 170, 38 },
		{ 355, 171, 38 },
		{ 356, 144, 38 },
		{ 356, 145, 38 },
		{ 356, 146, 38 },
		{ 356, 147, 38 },
		{ 356, 148, 38 },
		{ 356, 149, 38 },
		{ 356, 150, 38 },
		{ 356, 151, 38 },
		{ 356, 152, 38 },
		{ 356, 153, 38 },
		{ 356, 154, 38 },
		{ 356, 155, 38 },
		{ 356, 156, 38 },
		{ 356, 157, 38 },
		{ 356, 158, 38 },
		{ 356, 159, 38 },
		{ 356, 160, 38 },
		{ 356, 161, 38 },
		{ 356, 162, 38 },
		{ 356, 163, 38 },
		{ 356, 164, 38 },
		{ 356, 165, 38 },
		{ 356, 166, 38 },
		{ 356, 167, 38 },
		{ 356, 168, 38 },
		{ 356, 169, 38 },
		{ 356, 170, 38 },
		{ 357, 143, 38 },
		{ 357, 144, 38 },
		{ 357, 145, 38 },
		{ 357, 146, 38 },
		{ 357, 147, 38 },
		{ 357, 148, 38 },
		{ 357, 149, 38 },
		{ 357, 150, 38 },
		{ 357, 151, 38 },
		{ 357, 152, 38 },
		{ 357, 153, 38 },
		{ 357, 154, 38 },
		{ 357, 155, 38 },
		{ 357, 156, 38 },
		{ 357, 157, 38 },
		{ 357, 158, 38 },
		{ 357, 159, 38 },
		{ 357, 160, 38 },
		{ 357, 161, 38 },
		{ 357, 162, 38 },
		{ 357, 163, 38 },
		{ 357, 164, 38 },
		{ 357, 165, 38 },
		{ 357, 166, 38 },
		{ 357, 167, 38 },
		{ 357, 168, 38 },
		{ 357, 169, 38 },
		{ 358, 142, 38 },
		{ 358, 143, 38 },
		{ 358, 144, 38 },
		{ 358, 145, 38 },
		{ 358, 146, 38 },
		{ 358, 147, 38 },
		{ 358, 148, 38 },
		{ 358, 149, 38 },
		{ 358, 150, 38 },
		{ 358, 151, 38 },
		{ 358, 152, 38 },
		{ 358, 153, 38 },
		{ 358, 154, 38 },
		{ 358, 155, 38 },
		{ 358, 156, 38 },
		{ 358, 157, 38 },
		{ 358, 158, 38 },
		{ 358, 159, 38 },
		{ 358, 160, 38 },
		{ 358, 161, 38 },
		{ 358, 162, 38 },
		{ 358, 163, 38 },
		{ 358, 164, 38 },
		{ 358, 165, 38 },
		{ 358, 166, 38 },
		{ 358, 167, 38 },
		{ 358, 168, 38 },
		{ 359, 141, 38 },
		{ 359, 142, 38 },
		{ 359, 143, 38 },
		{ 359, 144, 38 },
		{ 359, 145, 38 },
		{ 359, 146, 38 },
		{ 359, 147, 38 },
		{ 359, 148, 38 },
		{ 359, 149, 38 },
		{ 359, 150, 38 },
		{ 359, 151, 38 },
		{ 359, 152, 38 },
		{ 359, 153, 38 },
		{ 359, 154, 38 },
		{ 359, 155, 38 },
		{ 359, 156, 38 },
		{ 359, 157, 38 },
		{ 359, 158, 38 },
		{ 359, 159, 38 },
		{ 359, 160, 38 },
		{ 359, 161, 38 },
		{ 359, 162, 38 },
		{ 359, 163, 38 },
		{ 359, 164, 38 },
		{ 359, 165, 38 },
		{ 359, 166, 38 },
		{ 359, 167, 38 },
		{ 360, 140, 38 },
		{ 360, 141, 38 },
		{ 360, 142, 38 },
		{ 360, 143, 38 },
		{ 360, 144, 38 },
		{ 360, 145, 38 },
		{ 360, 146, 38 },
		{ 360, 147, 38 },
		{ 360, 148, 38 },
		{ 360, 149, 38 },
		{ 360, 150, 38 },
		{ 360, 151, 38 },
		{ 360, 152, 38 },
		{ 360, 153, 38 },
		{ 360, 154, 38 },
		{ 360, 155, 38 },
		{ 360, 156, 38 },
		{ 360, 157, 38 },
		{ 360, 158, 38 },
		{ 360, 159, 38 },
		{ 360, 160, 38 },
		{ 360, 161, 38 },
		{ 360, 162, 38 },
		{ 360, 163, 38 },
		{ 360, 164, 38 },
		{ 360, 165, 38 },
		{ 360, 166, 38 },
		{ 361, 139, 38 },
		{ 361, 140, 38 },
		{ 361, 141, 38 },
		{ 361, 142, 38 },
		{ 361, 143, 38 },
		{ 361, 144, 38 },
		{ 361, 145, 38 },
		{ 361, 146, 38 },
		{ 361, 147, 38 },
		{ 361, 148, 38 },
		{ 361, 149, 38 },
		{ 361, 150, 38 },
		{ 361, 151, 38 },
		{ 361, 152, 38 },
		{ 361, 153, 38 },
		{ 361, 154, 38 },
		{ 361, 155, 38 },
		{ 361, 156, 38 },
		{ 361, 157, 38 },
		{ 361, 158, 38 },
		{ 361, 159, 38 },
		{ 361, 160, 38 },
		{ 361, 161, 38 },
		{ 361, 162, 38 },
		{ 361, 163, 38 },
		{ 361, 164, 38 },
		{ 361, 165, 38 },
		{ 362, 138, 38 },
		{ 362, 139, 38 },
		{ 362, 140, 38 },
		{ 362, 141, 38 },
		{ 362, 142, 38 },
		{ 362, 143, 38 },
		{ 362, 144, 38 },
		{ 362, 145, 38 },
		{ 362, 146, 38 },
		{ 362, 147, 38 },
		{ 362, 148, 38 },
		{ 362, 149, 38 },
		{ 362, 150, 38 },
		{ 362, 151, 38 },
		{ 362, 152, 38 },
		{ 362, 153, 38 },
		{ 362, 154, 38 },
		{ 362, 155, 38 },
		{ 362, 156, 38 },
		{ 362, 157, 38 },
		{ 362, 158, 38 },
		{ 362, 159, 38 },
		{ 362, 160, 38 },
		{ 362, 161, 38 },
		{ 362, 162, 38 },
		{ 362, 163, 38 },
		{ 362, 164, 38 },
		{ 362, 165, 38 },
		{ 363, 136, 38 },
		{ 363, 137, 38 },
		{ 363, 138, 38 },
		{ 363, 139, 38 },
		{ 363, 140, 38 },
		{ 363, 141, 38 },
		{ 363, 142, 38 },
		{ 363, 143, 38 },
		{ 363, 144, 38 },
		{ 363, 145, 38 },
		{ 363, 146, 38 },
		{ 363, 147, 38 },
		{ 363, 148, 38 },
		{ 363, 149, 38 },
		{ 363, 150, 38 },
		{ 363, 151, 38 },
		{ 363, 152, 38 },
		{ 363, 153, 38 },
		{ 363, 154, 38 },
		{ 363, 155, 38 },
		{ 363, 156, 38 },
		{ 363, 157, 38 },
		{ 363, 158, 38 },
		{ 363, 159, 38 },
		{ 363, 160, 38 },
		{ 363, 161, 38 },
		{ 363, 162, 38 },
		{ 363, 163, 38 },
		{ 363, 164, 38 },
		{ 364, 135, 38 },
		{ 364, 136, 38 },
		{ 364, 137, 38 },
		{ 364, 138, 38 },
		{ 364, 139, 38 },
		{ 364, 140, 38 },
		{ 364, 141, 38 },
		{ 364, 142, 38 },
		{ 364, 143, 38 },
		{ 364, 144, 38 },
		{ 364, 145, 38 },
		{ 364, 146, 38 },
		{ 364, 147, 38 },
		{ 364, 148, 38 },
		{ 364, 149, 38 },
		{ 364, 150, 38 },
		{ 364, 151, 38 },
		{ 364, 152, 38 },
		{ 364, 153, 38 },
		{ 364, 154, 38 },
		{ 364, 155, 38 },
		{ 364, 156, 38 },
		{ 364, 157, 38 },
		{ 364, 158, 38 },
		{ 364, 159, 38 },
		{ 364, 160, 38 },
		{ 364, 161, 38 },
		{ 364, 162, 38 },
		{ 364, 163, 38 },
		{ 365, 134, 38 },
		{ 365, 135, 38 },
		{ 365, 136, 38 },
		{ 365, 137, 38 },
		{ 365, 138, 38 },
		{ 365, 139, 38 },
		{ 365, 140, 38 },
		{ 365, 141, 38 },
		{ 365, 142, 38 },
		{ 365, 143, 38 },
		{ 365, 144, 38 },
		{ 365, 145, 38 },
		{ 365, 146, 38 },
		{ 365, 147, 38 },
		{ 365, 148, 38 },
		{ 365, 149, 38 },
		{ 365, 150, 38 },
		{ 365, 151, 38 },
		{ 365, 152, 38 },
		{ 365, 153, 38 },
		{ 365, 154, 38 },
		{ 365, 155, 38 },
		{ 365, 156, 38 },
		{ 365, 157, 38 },
		{ 365, 158, 38 },
		{ 365, 159, 38 },
		{ 365, 160, 38 },
		{ 365, 161, 38 },
		{ 365, 162, 38 },
		{ 366, 133, 38 },
		{ 366, 134, 38 },
		{ 366, 135, 38 },
		{ 366, 136, 38 },
		{ 366, 137, 38 },
		{ 366, 138, 38 },
		{ 366, 139, 38 },
		{ 366, 140, 38 },
		{ 366, 141, 38 },
		{ 366, 142, 38 },
		{ 366, 143, 38 },
		{ 366, 144, 38 },
		{ 366, 145, 38 },
		{ 366, 146, 38 },
		{ 366, 147, 38 },
		{ 366, 148, 38 },
		{ 366, 149, 38 },
		{ 366, 150, 38 },
		{ 366, 151, 38 },
		{ 366, 152, 38 },
		{ 366, 153, 38 },
		{ 366, 154, 38 },
		{ 366, 155, 38 },
		{ 366, 156, 38 },
		{ 366, 157, 38 },
		{ 366, 158, 38 },
		{ 366, 159, 38 },
		{ 366, 160, 38 },
		{ 366, 161, 38 },
		{ 367, 132, 38 },
		{ 367, 133, 38 },
		{ 367, 134, 38 },
		{ 367, 135, 38 },
		{ 367, 136, 38 },
		{ 367, 137, 38 },
		{ 367, 138, 38 },
		{ 367, 139, 38 },
		{ 367, 140, 38 },
		{ 367, 141, 38 },
		{ 367, 142, 38 },
		{ 367, 143, 38 },
		{ 367, 144, 38 },
		{ 367, 145, 38 },
		{ 367, 146, 38 },
		{ 367, 147, 38 },
		{ 367, 148, 38 },
		{ 367, 149, 38 },
		{ 367, 150, 38 },
		{ 367, 151, 38 },
		{ 367, 152, 38 },
		{ 367, 153, 38 },
		{ 367, 154, 38 },
		{ 367, 155, 38 },
		{ 367, 156, 38 },
		{ 367, 157, 38 },
		{ 367, 158, 38 },
		{ 367, 159, 38 },
		{ 367, 160, 38 },
		{ 368, 131, 38 },
		{ 368, 132, 38 },
		{ 368, 133, 38 },
		{ 368, 134, 38 },
		{ 368, 135, 38 },
		{ 368, 136, 38 },
		{ 368, 137, 38 },
		{ 368, 138, 38 },
		{ 368, 139, 38 },
		{ 368, 140, 38 },
		{ 368, 141, 38 },
		{ 368, 142, 38 },
		{ 368, 143, 38 },
		{ 368, 144, 38 },
		{ 368, 145, 38 },
		{ 368, 146, 38 },
		{ 368, 147, 38 },
		{ 368, 148, 38 },
		{ 368, 149, 38 },
		{ 368, 150, 38 },
		{ 368, 151, 38 },
		{ 368, 152, 38 },
		{ 368, 153, 38 },
		{ 368, 154, 38 },
		{ 368, 155, 38 },
		{ 368, 156, 38 },
		{ 368, 157, 38 },
		{ 368, 158, 38 },
		{ 368, 159, 38 },
		{ 369, 130, 38 },
		{ 369, 131, 38 },
		{ 369, 132, 38 },
		{ 369, 133, 38 },
		{ 369, 134, 38 },
		{ 369, 135, 38 },
		{ 369, 136, 38 },
		{ 369, 137, 38 },
		{ 369, 138, 38 },
		{ 369, 139, 38 },
		{ 369, 140, 38 },
		{ 369, 141, 38 },
		{ 369, 142, 38 },
		{ 369, 143, 38 },
		{ 369, 144, 38 },
		{ 369, 145, 38 },
		{ 369, 146, 38 },
		{ 369, 147, 38 },
		{ 369, 148, 38 },
		{ 369, 149, 38 },
		{ 369, 150, 38 },
		{ 369, 151, 38 },
		{ 369, 152, 38 },
		{ 369, 153, 38 },
		{ 369, 154, 38 },
		{ 369, 155, 38 },
		{ 369, 156, 38 },
		{ 369, 157, 38 },
		{ 369, 158, 38 },
		{ 370, 128, 38 },
		{ 370, 129, 38 },
		{ 370, 130, 38 },
		{ 370, 131, 38 },
		{ 370, 132, 38 },
		{ 370, 133, 38 },
		{ 370, 134, 38 },
		{ 370, 135, 38 },
		{ 370, 136, 38 },
		{ 370, 137, 38 },
		{ 370, 138, 38 },
		{ 370, 139, 38 },
		{ 370, 140, 38 },
		{ 370, 141, 38 },
		{ 370, 142, 38 },
		{ 370, 143, 38 },
		{ 370, 144, 38 },
		{ 370, 145, 38 },
		{ 370, 146, 38 },
		{ 370, 147, 38 },
		{ 370, 148, 38 },
		{ 370, 149, 38 },
		{ 370, 150, 38 },
		{ 370, 151, 38 },
		{ 370, 152, 38 },
		{ 370, 153, 38 },
		{ 370, 154, 38 },
		{ 370, 155, 38 },
		{ 370, 156, 38 },
		{ 370, 157, 38 },
		{ 370, 158, 38 },
		{ 371, 127, 38 },
		{ 371, 128, 38 },
		{ 371, 129, 38 },
		{ 371, 130, 38 },
		{ 371, 131, 38 },
		{ 371, 132, 38 },
		{ 371, 133, 38 },
		{ 371, 134, 38 },
		{ 371, 135, 38 },
		{ 371, 136, 38 },
		{ 371, 137, 38 },
		{ 371, 138, 38 },
		{ 371, 139, 38 },
		{ 371, 140, 38 },
		{ 371, 141, 38 },
		{ 371, 142, 38 },
		{ 371, 143, 38 },
		{ 371, 144, 38 },
		{ 371, 145, 38 },
		{ 371, 146, 38 },
		{ 371, 147, 38 },
		{ 371, 148, 38 },
		{ 371, 149, 38 },
		{ 371, 150, 38 },
		{ 371, 151, 38 },
		{ 371, 152, 38 },
		{ 371, 153, 38 },
		{ 371, 154, 38 },
		{ 371, 155, 38 },
		{ 371, 156, 38 },
		{ 371, 157, 38 },
		{ 372, 126, 38 },
		{ 372, 127, 38 },
		{ 372, 128, 38 },
		{ 372, 129, 38 },
		{ 372, 130, 38 },
		{ 372, 131, 38 },
		{ 372, 132, 38 },
		{ 372, 133, 38 },
		{ 372, 134, 38 },
		{ 372, 135, 38 },
		{ 372, 136, 38 },
		{ 372, 137, 38 },
		{ 372, 138, 38 },
		{ 372, 139, 38 },
		{ 372, 140, 38 },
		{ 372, 141, 38 },
		{ 372, 142, 38 },
		{ 372, 143, 38 },
		{ 372, 144, 38 },
		{ 372, 145, 38 },
		{ 372, 146, 38 },
		{ 372, 147, 38 },
		{ 372, 148, 38 },
		{ 372, 149, 38 },
		{ 372, 150, 38 },
		{ 372, 151, 38 },
		{ 372, 152, 38 },
		{ 372, 153, 38 },
		{ 372, 154, 38 },
		{ 372, 155, 38 },
		{ 372, 156, 38 },
		{ 373, 125, 38 },
		{ 373, 126, 38 },
		{ 373, 127, 38 },
		{ 373, 128, 38 },
		{ 373, 129, 38 },
		{ 373, 130, 38 },
		{ 373, 131, 38 },
		{ 373, 132, 38 },
		{ 373, 133, 38 },
		{ 373, 134, 38 },
		{ 373, 135, 38 },
		{ 373, 136, 38 },
		{ 373, 137, 38 },
		{ 373, 138, 38 },
		{ 373, 139, 38 },
		{ 373, 140, 38 },
		{ 373, 141, 38 },
		{ 373, 142, 38 },
		{ 373, 143, 38 },
		{ 373, 144, 38 },
		{ 373, 145, 38 },
		{ 373, 146, 38 },
		{ 373, 147, 38 },
		{ 373, 148, 38 },
		{ 373, 149, 38 },
		{ 373, 150, 38 },
		{ 373, 151, 38 },
		{ 373, 152, 38 },
		{ 373, 153, 38 },
		{ 373, 154, 38 },
		{ 373, 155, 38 },
		{ 374, 124, 38 },
		{ 374, 125, 38 },
		{ 374, 126, 38 },
		{ 374, 127, 38 },
		{ 374, 128, 38 },
		{ 374, 129, 38 },
		{ 374, 130, 38 },
		{ 374, 131, 38 },
		{ 374, 132, 38 },
		{ 374, 133, 38 },
		{ 374, 134, 38 },
		{ 374, 135, 38 },
		{ 374, 136, 38 },
		{ 374, 137, 38 },
		{ 374, 138, 38 },
		{ 374, 139, 38 },
		{ 374, 140, 38 },
		{ 374, 141, 38 },
		{ 374, 142, 38 },
		{ 374, 143, 38 },
		{ 374, 144, 38 },
		{ 374, 145, 38 },
		{ 374, 146, 38 },
		{ 374, 147, 38 },
		{ 374, 148, 38 },
		{ 374, 149, 38 },
		{ 374, 150, 38 },
		{ 374, 151, 38 },
		{ 374, 152, 38 },
		{ 374, 153, 38 },
		{ 374, 154, 38 },
		{ 375, 123, 38 },
		{ 375, 124, 38 },
		{ 375, 125, 38 },
		{ 375, 126, 38 },
		{ 375, 127, 38 },
		{ 375, 128, 38 },
		{ 375, 129, 38 },
		{ 375, 130, 38 },
		{ 375, 131, 38 },
		{ 375, 132, 38 },
		{ 375, 133, 38 },
		{ 375, 134, 38 },
		{ 375, 135, 38 },
		{ 375, 136, 38 },
		{ 375, 137, 38 },
		{ 375, 138, 38 },
		{ 375, 139, 38 },
		{ 375, 140, 38 },
		{ 375, 141, 38 },
		{ 375, 142, 38 },
		{ 375, 143, 38 },
		{ 375, 144, 38 },
		{ 375, 145, 38 },
		{ 375, 146, 38 },
		{ 375, 147, 38 },
		{ 375, 148, 38 },
		{ 375, 149, 38 },
		{ 375, 150, 38 },
		{ 375, 151, 38 },
		{ 375, 152, 38 },
		{ 375, 153, 38 },
		{ 376, 122, 38 },
		{ 376, 123, 38 },
		{ 376, 124, 38 },
		{ 376, 125, 38 },
		{ 376, 126, 38 },
		{ 376, 127, 38 },
		{ 376, 128, 38 },
		{ 376, 129, 38 },
		{ 376, 130, 38 },
		{ 376, 131, 38 },
		{ 376, 132, 38 },
		{ 376, 133, 38 },
		{ 376, 134, 38 },
		{ 376, 135, 38 },
		{ 376, 136, 38 },
		{ 376, 137, 38 },
		{ 376, 138, 38 },
		{ 376, 139, 38 },
		{ 376, 140, 38 },
		{ 376, 141, 38 },
		{ 376, 142, 38 },
		{ 376, 143, 38 },
		{ 376, 144, 38 },
		{ 376, 145, 38 },
		{ 376, 146, 38 },
		{ 376, 147, 38 },
		{ 376, 148, 38 },
		{ 376, 149, 38 },
		{ 376, 150, 38 },
		{ 376, 151, 38 },
		{ 376, 152, 38 },
		{ 377, 120, 38 },
		{ 377, 121, 38 },
		{ 377, 122, 38 },
		{ 377, 123, 38 },
		{ 377, 124, 38 },
		{ 377, 125, 38 },
		{ 377, 126, 38 },
		{ 377, 127, 38 },
		{ 377, 128, 38 },
		{ 377, 129, 38 },
		{ 377, 130, 38 },
		{ 377, 131, 38 },
		{ 377, 132, 38 },
		{ 377, 133, 38 },
		{ 377, 134, 38 },
		{ 377, 135, 38 },
		{ 377, 136, 38 },
		{ 377, 137, 38 },
		{ 377, 138, 38 },
		{ 377, 139, 38 },
		{ 377, 140, 38 },
		{ 377, 141, 38 },
		{ 377, 142, 38 },
		{ 377, 143, 38 },
		{ 377, 144, 38 },
		{ 377, 145, 38 },
		{ 377, 146, 38 },
		{ 377, 147, 38 },
		{ 377, 148, 38 },
		{ 377, 149, 38 },
		{ 377, 150, 38 },
		{ 377, 151, 38 },
		{ 378, 119, 38 },
		{ 378, 120, 38 },
		{ 378, 121, 38 },
		{ 378, 122, 38 },
		{ 378, 123, 38 },
		{ 378, 124, 38 },
		{ 378, 125, 38 },
		{ 378, 126, 38 },
		{ 378, 127, 38 },
		{ 378, 128, 38 },
		{ 378, 129, 38 },
		{ 378, 130, 38 },
		{ 378, 131, 38 },
		{ 378, 132, 38 },
		{ 378, 133, 38 },
		{ 378, 134, 38 },
		{ 378, 135, 38 },
		{ 378, 136, 38 },
		{ 378, 137, 38 },
		{ 378, 138, 38 },
		{ 378, 139, 38 },
		{ 378, 140, 38 },
		{ 378, 141, 38 },
		{ 378, 142, 38 },
		{ 378, 143, 38 },
		{ 378, 144, 38 },
		{ 378, 145, 38 },
		{ 378, 146, 38 },
		{ 378, 147, 38 },
		{ 378, 148, 38 },
		{ 378, 149, 38 },
		{ 378, 150, 38 },
		{ 378, 151, 38 },
		{ 379, 118, 38 },
		{ 379, 119, 38 },
		{ 379, 120, 38 },
		{ 379, 121, 38 },
		{ 379, 122, 38 },
		{ 379, 123, 38 },
		{ 379, 124, 38 },
		{ 379, 125, 38 },
		{ 379, 126, 38 },
		{ 379, 127, 38 },
		{ 379, 128, 38 },
		{ 379, 129, 38 },
		{ 379, 130, 38 },
		{ 379, 131, 38 },
		{ 379, 132, 38 },
		{ 379, 133, 38 },
		{ 379, 134, 38 },
		{ 379, 135, 38 },
		{ 379, 136, 38 },
		{ 379, 137, 38 },
		{ 379, 138, 38 },
		{ 379, 139, 38 },
		{ 379, 140, 38 },
		{ 379, 141, 38 },
		{ 379, 142, 38 },
		{ 379, 143, 38 },
		{ 379, 144, 38 },
		{ 379, 145, 38 },
		{ 379, 146, 38 },
		{ 379, 147, 38 },
		{ 379, 148, 38 },
		{ 379, 149, 38 },
		{ 379, 150, 38 },
		{ 380, 117, 38 },
		{ 380, 118, 38 },
		{ 380, 119, 38 },
		{ 380, 120, 38 },
		{ 380, 121, 38 },
		{ 380, 122, 38 },
		{ 380, 123, 38 },
		{ 380, 124, 38 },
		{ 380, 125, 38 },
		{ 380, 126, 38 },
		{ 380, 127, 38 },
		{ 380, 128, 38 },
		{ 380, 129, 38 },
		{ 380, 130, 38 },
		{ 380, 131, 38 },
		{ 380, 132, 38 },
		{ 380, 133, 38 },
		{ 380, 134, 38 },
		{ 380, 135, 38 },
		{ 380, 136, 38 },
		{ 380, 137, 38 },
		{ 380, 138, 38 },
		{ 380, 139, 38 },
		{ 380, 140, 38 },
		{ 380, 141, 38 },
		{ 380, 142, 38 },
		{ 380, 143, 38 },
		{ 380, 144, 38 },
		{ 380, 145, 38 },
		{ 380, 146, 38 },
		{ 380, 147, 38 },
		{ 380, 148, 38 },
		{ 380, 149, 38 },
		{ 381, 116, 38 },
		{ 381, 117, 38 },
		{ 381, 118, 38 },
		{ 381, 119, 38 },
		{ 381, 120, 38 },
		{ 381, 121, 38 },
		{ 381, 122, 38 },
		{ 381, 123, 38 },
		{ 381, 124, 38 },
		{ 381, 125, 38 },
		{ 381, 126, 38 },
		{ 381, 127, 38 },
		{ 381, 128, 38 },
		{ 381, 129, 38 },
		{ 381, 130, 38 },
		{ 381, 131, 38 },
		{ 381, 132, 38 },
		{ 381, 133, 38 },
		{ 381, 134, 38 },
		{ 381, 135, 38 },
		{ 381, 136, 38 },
		{ 381, 137, 38 },
		{ 381, 138, 38 },
		{ 381, 139, 38 },
		{ 381, 140, 38 },
		{ 381, 141, 38 },
		{ 381, 142, 38 },
		{ 381, 143, 38 },
		{ 381, 144, 38 },
		{ 381, 145, 38 },
		{ 381, 146, 38 },
		{ 381, 147, 38 },
		{ 381, 148, 38 },
		{ 382, 115, 38 },
		{ 382, 116, 38 },
		{ 382, 117, 38 },
		{ 382, 118, 38 },
		{ 382, 119, 38 },
		{ 382, 120, 38 },
		{ 382, 121, 38 },
		{ 382, 122, 38 },
		{ 382, 123, 38 },
		{ 382, 124, 38 },
		{ 382, 125, 38 },
		{ 382, 126, 38 },
		{ 382, 127, 38 },
		{ 382, 128, 38 },
		{ 382, 129, 38 },
		{ 382, 130, 38 },
		{ 382, 131, 38 },
		{ 382, 132, 38 },
		{ 382, 133, 38 },
		{ 382, 134, 38 },
		{ 382, 135, 38 },
		{ 382, 136, 38 },
		{ 382, 137, 38 },
		{ 382, 138, 38 },
		{ 382, 139, 38 },
		{ 382, 140, 38 },
		{ 382, 141, 38 },
		{ 382, 142, 38 },
		{ 382, 143, 38 },
		{ 382, 144, 38 },
		{ 382, 145, 38 },
		{ 382, 146, 38 },
		{ 382, 147, 38 },
		{ 383, 114, 38 },
		{ 383, 115, 38 },
		{ 383, 116, 38 },
		{ 383, 117, 38 },
		{ 383, 118, 38 },
		{ 383, 119, 38 },
		{ 383, 120, 38 },
		{ 383, 121, 38 },
		{ 383, 122, 38 },
		{ 383, 123, 38 },
		{ 383, 124, 38 },
		{ 383, 125, 38 },
		{ 383, 126, 38 },
		{ 383, 127, 38 },
		{ 383, 128, 38 },
		{ 383, 129, 38 },
		{ 383, 130, 38 },
		{ 383, 131, 38 },
		{ 383, 132, 38 },
		{ 383, 133, 38 },
		{ 383, 134, 38 },
		{ 383, 135, 38 },
		{ 383, 136, 38 },
		{ 383, 137, 38 },
		{ 383, 138, 38 },
		{ 383, 139, 38 },
		{ 383, 140, 38 },
		{ 383, 141, 38 },
		{ 383, 142, 38 },
		{ 383, 143, 38 },
		{ 383, 144, 38 },
		{ 383, 145, 38 },
		{ 383, 146, 38 },
		{ 384, 112, 38 },
		{ 384, 113, 38 },
		{ 384, 114, 38 },
		{ 384, 115, 38 },
		{ 384, 116, 38 },
		{ 384, 117, 38 },
		{ 384, 118, 38 },
		{ 384, 119, 38 },
		{ 384, 120, 38 },
		{ 384, 121, 38 },
		{ 384, 122, 38 },
		{ 384, 123, 38 },
		{ 384, 124, 38 },
		{ 384, 125, 38 },
		{ 384, 126, 38 },
		{ 384, 127, 38 },
		{ 384, 128, 38 },
		{ 384, 129, 38 },
		{ 384, 130, 38 },
		{ 384, 131, 38 },
		{ 384, 132, 38 },
		{ 384, 133, 38 },
		{ 384, 134, 38 },
		{ 384, 135, 38 },
		{ 384, 136, 38 },
		{ 384, 137, 38 },
		{ 384, 138, 38 },
		{ 384, 139, 38 },
		{ 384, 140, 38 },
		{ 384, 141, 38 },
		{ 384, 142, 38 },
		{ 384, 143, 38 },
		{ 384, 144, 38 },
		{ 384, 145, 38 },
		{ 385, 111, 38 },
		{ 385, 112, 38 },
		{ 385, 113, 38 },
		{ 385, 114, 38 },
		{ 385, 115, 38 },
		{ 385, 116, 38 },
		{ 385, 117, 38 },
		{ 385, 118, 38 },
		{ 385, 119, 38 },
		{ 385, 120, 38 },
		{ 385, 121, 38 },
		{ 385, 122, 38 },
		{ 385, 123, 38 },
		{ 385, 124, 38 },
		{ 385, 125, 38 },
		{ 385, 126, 38 },
		{ 385, 127, 38 },
		{ 385, 128, 38 },
		{ 385, 129, 38 },
		{ 385, 130, 38 },
		{ 385, 131, 38 },
		{ 385, 132, 38 },
		{ 385, 133, 38 },
		{ 385, 134, 38 },
		{ 385, 135, 38 },
		{ 385, 136, 38 },
		{ 385, 137, 38 },
		{ 385, 138, 38 },
		{ 385, 139, 38 },
		{ 385, 140, 38 },
		{ 385, 141, 38 },
		{ 385, 142, 38 },
		{ 385, 143, 38 },
		{ 385, 144, 38 },
		{ 386, 110, 38 },
		{ 386, 111, 38 },
		{ 386, 112, 38 },
		{ 386, 113, 38 },
		{ 386, 114, 38 },
		{ 386, 115, 38 },
		{ 386, 116, 38 },
		{ 386, 117, 38 },
		{ 386, 118, 38 },
		{ 386, 119, 38 },
		{ 386, 120, 38 },
		{ 386, 121, 38 },
		{ 386, 122, 38 },
		{ 386, 123, 38 },
		{ 386, 124, 38 },
		{ 386, 125, 38 },
		{ 386, 126, 38 },
		{ 386, 127, 38 },
		{ 386, 128, 38 },
		{ 386, 129, 38 },
		{ 386, 130, 38 },
		{ 386, 131, 38 },
		{ 386, 132, 38 },
		{ 386, 133, 38 },
		{ 386, 134, 38 },
		{ 386, 135, 38 },
		{ 386, 136, 38 },
		{ 386, 137, 38 },
		{ 386, 138, 38 },
		{ 386, 139, 38 },
		{ 386, 140, 38 },
		{ 386, 141, 38 },
		{ 386, 142, 38 },
		{ 386, 143, 38 },
		{ 386, 144, 38 },
		{ 387, 109, 38 },
		{ 387, 110, 38 },
		{ 387, 111, 38 },
		{ 387, 112, 38 },
		{ 387, 113, 38 },
		{ 387, 114, 38 },
		{ 387, 115, 38 },
		{ 387, 116, 38 },
		{ 387, 117, 38 },
		{ 387, 118, 38 },
		{ 387, 119, 38 },
		{ 387, 120, 38 },
		{ 387, 121, 38 },
		{ 387, 122, 38 },
		{ 387, 123, 38 },
		{ 387, 124, 38 },
		{ 387, 125, 38 },
		{ 387, 126, 38 },
		{ 387, 127, 38 },
		{ 387, 128, 38 },
		{ 387, 129, 38 },
		{ 387, 130, 38 },
		{ 387, 131, 38 },
		{ 387, 132, 38 },
		{ 387, 133, 38 },
		{ 387, 134, 38 },
		{ 387, 135, 38 },
		{ 387, 136, 38 },
		{ 387, 137, 38 },
		{ 387, 138, 38 },
		{ 387, 139, 38 },
		{ 387, 140, 38 },
		{ 387, 141, 38 },
		{ 387, 142, 38 },
		{ 387, 143, 38 },
		{ 388, 108, 38 },
		{ 388, 109, 38 },
		{ 388, 110, 38 },
		{ 388, 111, 38 },
		{ 388, 112, 38 },
		{ 388, 113, 38 },
		{ 388, 114, 38 },
		{ 388, 115, 38 },
		{ 388, 116, 38 },
		{ 388, 117, 38 },
		{ 388, 118, 38 },
		{ 388, 119, 38 },
		{ 388, 120, 38 },
		{ 388, 121, 38 },
		{ 388, 122, 38 },
		{ 388, 123, 38 },
		{ 388, 124, 38 },
		{ 388, 125, 38 },
		{ 388, 126, 38 },
		{ 388, 127, 38 },
		{ 388, 128, 38 },
		{ 388, 129, 38 },
		{ 388, 130, 38 },
		{ 388, 131, 38 },
		{ 388, 132, 38 },
		{ 388, 133, 38 },
		{ 388, 134, 38 },
		{ 388, 135, 38 },
		{ 388, 136, 38 },
		{ 388, 137, 38 },
		{ 388, 138, 38 },
		{ 388, 139, 38 },
		{ 388, 140, 38 },
		{ 388, 141, 38 },
		{ 388, 142, 38 },
		{ 389, 107, 38 },
		{ 389, 108, 38 },
		{ 389, 109, 38 },
		{ 389, 110, 38 },
		{ 389, 111, 38 },
		{ 389, 112, 38 },
		{ 389, 113, 38 },
		{ 389, 114, 38 },
		{ 389, 115, 38 },
		{ 389, 116, 38 },
		{ 389, 117, 38 },
		{ 389, 118, 38 },
		{ 389, 119, 38 },
		{ 389, 120, 38 },
		{ 389, 121, 38 },
		{ 389, 122, 38 },
		{ 389, 123, 38 },
		{ 389, 124, 38 },
		{ 389, 125, 38 },
		{ 389, 126, 38 },
		{ 389, 127, 38 },
		{ 389, 128, 38 },
		{ 389, 129, 38 },
		{ 389, 130, 38 },
		{ 389, 131, 38 },
		{ 389, 132, 38 },
		{ 389, 133, 38 },
		{ 389, 134, 38 },
		{ 389, 135, 38 },
		{ 389, 136, 38 },
		{ 389, 137, 38 },
		{ 389, 138, 38 },
		{ 389, 139, 38 },
		{ 389, 140, 38 },
		{ 389, 141, 38 },
		{ 390, 106, 38 },
		{ 390, 107, 38 },
		{ 390, 108, 38 },
		{ 390, 109, 38 },
		{ 390, 110, 38 },
		{ 390, 111, 38 },
		{ 390, 112, 38 },
		{ 390, 113, 38 },
		{ 390, 114, 38 },
		{ 390, 115, 38 },
		{ 390, 116, 38 },
		{ 390, 117, 38 },
		{ 390, 118, 38 },
		{ 390, 119, 38 },
		{ 390, 120, 38 },
		{ 390, 121, 38 },
		{ 390, 122, 38 },
		{ 390, 123, 38 },
		{ 390, 124, 38 },
		{ 390, 125, 38 },
		{ 390, 126, 38 },
		{ 390, 127, 38 },
		{ 390, 128, 38 },
		{ 390, 129, 38 },
		{ 390, 130, 38 },
		{ 390, 131, 38 },
		{ 390, 132, 38 },
		{ 390, 133, 38 },
		{ 390, 134, 38 },
		{ 390, 135, 38 },
		{ 390, 136, 38 },
		{ 390, 137, 38 },
		{ 390, 138, 38 },
		{ 390, 139, 38 },
		{ 390, 140, 38 },
		{ 391, 104, 38 },
		{ 391, 105, 38 },
		{ 391, 106, 38 },
		{ 391, 107, 38 },
		{ 391, 108, 38 },
		{ 391, 109, 38 },
		{ 391, 110, 38 },
		{ 391, 111, 38 },
		{ 391, 112, 38 },
		{ 391, 113, 38 },
		{ 391, 114, 38 },
		{ 391, 115, 38 },
		{ 391, 116, 38 },
		{ 391, 117, 38 },
		{ 391, 118, 38 },
		{ 391, 119, 38 },
		{ 391, 120, 38 },
		{ 391, 121, 38 },
		{ 391, 122, 38 },
		{ 391, 123, 38 },
		{ 391, 124, 38 },
		{ 391, 125, 38 },
		{ 391, 126, 38 },
		{ 391, 127, 38 },
		{ 391, 128, 38 },
		{ 391, 129, 38 },
		{ 391, 130, 38 },
		{ 391, 131, 38 },
		{ 391, 132, 38 },
		{ 391, 133, 38 },
		{ 391, 134, 38 },
		{ 391, 135, 38 },
		{ 391, 136, 38 },
		{ 391, 137, 38 },
		{ 391, 138, 38 },
		{ 391, 139, 38 },
		{ 392, 103, 38 },
		{ 392, 104, 38 },
		{ 392, 105, 38 },
		{ 392, 106, 38 },
		{ 392, 107, 38 },
		{ 392, 108, 38 },
		{ 392, 109, 38 },
		{ 392, 110, 38 },
		{ 392, 111, 38 },
		{ 392, 112, 38 },
		{ 392, 113, 38 },
		{ 392, 114, 38 },
		{ 392, 115, 38 },
		{ 392, 116, 38 },
		{ 392, 117, 38 },
		{ 392, 118, 38 },
		{ 392, 119, 38 },
		{ 392, 120, 38 },
		{ 392, 121, 38 },
		{ 392, 122, 38 },
		{ 392, 123, 38 },
		{ 392, 124, 38 },
		{ 392, 125, 38 },
		{ 392, 126, 38 },
		{ 392, 127, 38 },
		{ 392, 128, 38 },
		{ 392, 129, 38 },
		{ 392, 130, 38 },
		{ 392, 131, 38 },
		{ 392, 132, 38 },
		{ 392, 133, 38 },
		{ 392, 134, 38 },
		{ 392, 135, 38 },
		{ 392, 136, 38 },
		{ 392, 137, 38 },
		{ 392, 138, 38 },
		{ 393, 102, 38 },
		{ 393, 103, 38 },
		{ 393, 104, 38 },
		{ 393, 105, 38 },
		{ 393, 106, 38 },
		{ 393, 107, 38 },
		{ 393, 108, 38 },
		{ 393, 109, 38 },
		{ 393, 110, 38 },
		{ 393, 111, 38 },
		{ 393, 112, 38 },
		{ 393, 113, 38 },
		{ 393, 114, 38 },
		{ 393, 115, 38 },
		{ 393, 116, 38 },
		{ 393, 117, 38 },
		{ 393, 118, 38 },
		{ 393, 119, 38 },
		{ 393, 120, 38 },
		{ 393, 121, 38 },
		{ 393, 122, 38 },
		{ 393, 123, 38 },
		{ 393, 124, 38 },
		{ 393, 125, 38 },
		{ 393, 126, 38 },
		{ 393, 127, 38 },
		{ 393, 128, 38 },
		{ 393, 129, 38 },
		{ 393, 130, 38 },
		{ 393, 131, 38 },
		{ 393, 132, 38 },
		{ 393, 133, 38 },
		{ 393, 134, 38 },
		{ 393, 135, 38 },
		{ 393, 136, 38 },
		{ 393, 137, 38 },
		{ 394, 101, 38 },
		{ 394, 102, 38 },
		{ 394, 103, 38 },
		{ 394, 104, 38 },
		{ 394, 105, 38 },
		{ 394, 106, 38 },
		{ 394, 107, 38 },
		{ 394, 108, 38 },
		{ 394, 109, 38 },
		{ 394, 110, 38 },
		{ 394, 111, 38 },
		{ 394, 112, 38 },
		{ 394, 113, 38 },
		{ 394, 114, 38 },
		{ 394, 115, 38 },
		{ 394, 116, 38 },
		{ 394, 117, 38 },
		{ 394, 118, 38 },
		{ 394, 119, 38 },
		{ 394, 120, 38 },
		{ 394, 121, 38 },
		{ 394, 122, 38 },
		{ 394, 123, 38 },
		{ 394, 124, 38 },
		{ 394, 125, 38 },
		{ 394, 126, 38 },
		{ 394, 127, 38 },
		{ 394, 128, 38 },
		{ 394, 129, 38 },
		{ 394, 130, 38 },
		{ 394, 131, 38 },
		{ 394, 132, 38 },
		{ 394, 133, 38 },
		{ 394, 134, 38 },
		{ 394, 135, 38 },
		{ 394, 136, 38 },
		{ 394, 137, 38 },
		{ 395, 100, 38 },
		{ 395, 101, 38 },
		{ 395, 102, 38 },
		{ 395, 103, 38 },
		{ 395, 104, 38 },
		{ 395, 105, 38 },
		{ 395, 106, 38 },
		{ 395, 107, 38 },
		{ 395, 108, 38 },
		{ 395, 109, 38 },
		{ 395, 110, 38 },
		{ 395, 111, 38 },
		{ 395, 112, 38 },
		{ 395, 113, 38 },
		{ 395, 114, 38 },
		{ 395, 115, 38 },
		{ 395, 116, 38 },
		{ 395, 117, 38 },
		{ 395, 118, 38 },
		{ 395, 119, 38 },
		{ 395, 120, 38 },
		{ 395, 121, 38 },
		{ 395, 122, 38 },
		{ 395, 123, 38 },
		{ 395, 124, 38 },
		{ 395, 125, 38 },
		{ 395, 126, 38 },
		{ 395, 127, 38 },
		{ 395, 128, 38 },
		{ 395, 129, 38 },
		{ 395, 130, 38 },
		{ 395, 131, 38 },
		{ 395, 132, 38 },
		{ 395, 133, 38 },
		{ 395, 134, 38 },
		{ 395, 135, 38 },
		{ 395, 136, 38 },
		{ 396, 99, 38 },
		{ 396, 100, 38 },
		{ 396, 101, 38 },
		{ 396, 102, 38 },
		{ 396, 103, 38 },
		{ 396, 104, 38 },
		{ 396, 105, 38 },
		{ 396, 106, 38 },
		{ 396, 107, 38 },
		{ 396, 108, 38 },
		{ 396, 109, 38 },
		{ 396, 110, 38 },
		{ 396, 111, 38 },
		{ 396, 112, 38 },
		{ 396, 113, 38 },
		{ 396, 114, 38 },
		{ 396, 115, 38 },
		{ 396, 116, 38 },
		{ 396, 117, 38 },
		{ 396, 118, 38 },
		{ 396, 119, 38 },
		{ 396, 120, 38 },
		{ 396, 121, 38 },
		{ 396, 122, 38 },
		{ 396, 123, 38 },
		{ 396, 124, 38 },
		{ 396, 125, 38 },
		{ 396, 126, 38 },
		{ 396, 127, 38 },
		{ 396, 128, 38 },
		{ 396, 129, 38 },
		{ 396, 130, 38 },
		{ 396, 131, 38 },
		{ 396, 132, 38 },
		{ 396, 133, 38 },
		{ 396, 134, 38 },
		{ 396, 135, 38 },
		{ 397, 98, 38 },
		{ 397, 99, 38 },
		{ 397, 100, 38 },
		{ 397, 101, 38 },
		{ 397, 102, 38 },
		{ 397, 103, 38 },
		{ 397, 104, 38 },
		{ 397, 105, 38 },
		{ 397, 106, 38 },
		{ 397, 107, 38 },
		{ 397, 108, 38 },
		{ 397, 109, 38 },
		{ 397, 110, 38 },
		{ 397, 111, 38 },
		{ 397, 112, 38 },
		{ 397, 113, 38 },
		{ 397, 114, 38 },
		{ 397, 115, 38 },
		{ 397, 116, 38 },
		{ 397, 117, 38 },
		{ 397, 118, 38 },
		{ 397, 119, 38 },
		{ 397, 120, 38 },
		{ 397, 121, 38 },
		{ 397, 122, 38 },
		{ 397, 123, 38 },
		{ 397, 124, 38 },
		{ 397, 125, 38 },
		{ 397, 126, 38 },
		{ 397, 127, 38 },
		{ 397, 128, 38 },
		{ 397, 129, 38 },
		{ 397, 130, 38 },
		{ 397, 131, 38 },
		{ 397, 132, 38 },
		{ 397, 133, 38 },
		{ 397, 134, 38 },
		{ 398, 96, 38 },
		{ 398, 97, 38 },
		{ 398, 98, 38 },
		{ 398, 99, 38 },
		{ 398, 100, 38 },
		{ 398, 101, 38 },
		{ 398, 102, 38 },
		{ 398, 103, 38 },
		{ 398, 104, 38 },
		{ 398, 105, 38 },
		{ 398, 106, 38 },
		{ 398, 107, 38 },
		{ 398, 108, 38 },
		{ 398, 109, 38 },
		{ 398, 110, 38 },
		{ 398, 111, 38 },
		{ 398, 112, 38 },
		{ 398, 113, 38 },
		{ 398, 114, 38 },
		{ 398, 115, 38 },
		{ 398, 116, 38 },
		{ 398, 117, 38 },
		{ 398, 118, 38 },
		{ 398, 119, 38 },
		{ 398, 120, 38 },
		{ 398, 121, 38 },
		{ 398, 122, 38 },
		{ 398, 123, 38 },
		{ 398, 124, 38 },
		{ 398, 125, 38 },
		{ 398, 126, 38 },
		{ 398, 127, 38 },
		{ 398, 128, 38 },
		{ 398, 129, 38 },
		{ 398, 130, 38 },
		{ 398, 131, 38 },
		{ 398, 132, 38 },
		{ 398, 133, 38 },
		{ 399, 95, 38 },
		{ 399, 96, 38 },
		{ 399, 97, 38 },
		{ 399, 98, 38 },
		{ 399, 99, 38 },
		{ 399, 100, 38 },
		{ 399, 101, 38 },
		{ 399, 102, 38 },
		{ 399, 103, 38 },
		{ 399, 104, 38 },
		{ 399, 105, 38 },
		{ 399, 106, 38 },
		{ 399, 107, 38 },
		{ 399, 108, 38 },
		{ 399, 109, 38 },
		{ 399, 110, 38 },
		{ 399, 111, 38 },
		{ 399, 112, 38 },
		{ 399, 113, 38 },
		{ 399, 114, 38 },
		{ 399, 115, 38 },
		{ 399, 116, 38 },
		{ 399, 117, 38 },
		{ 399, 118, 38 },
		{ 399, 119, 38 },
		{ 399, 120, 38 },
		{ 399, 121, 38 },
		{ 399, 122, 38 },
		{ 399, 123, 38 },
		{ 399, 124, 38 },
		{ 399, 125, 38 },
		{ 399, 126, 38 },
		{ 399, 127, 38 },
		{ 399, 128, 38 },
		{ 399, 129, 38 },
		{ 399, 130, 38 },
		{ 399, 131, 38 },
		{ 399, 132, 38 },
		{ 400, 94, 38 },
		{ 400, 95, 38 },
		{ 400, 96, 38 },
		{ 400, 97, 38 },
		{ 400, 98, 38 },
		{ 400, 99, 38 },
		{ 400, 100, 38 },
		{ 400, 101, 38 },
		{ 400, 102, 38 },
		{ 400, 103, 38 },
		{ 400, 104, 38 },
		{ 400, 105, 38 },
		{ 400, 106, 38 },
		{ 400, 107, 38 },
		{ 400, 108, 38 },
		{ 400, 109, 38 },
		{ 400, 110, 38 },
		{ 400, 111, 38 },
		{ 400, 112, 38 },
		{ 400, 113, 38 },
		{ 400, 114, 38 },
		{ 400, 115, 38 },
		{ 400, 116, 38 },
		{ 400, 117, 38 },
		{ 400, 118, 38 },
		{ 400, 119, 38 },
		{ 400, 120, 38 },
		{ 400, 121, 38 },
		{ 400, 122, 38 },
		{ 400, 123, 38 },
		{ 400, 124, 38 },
		{ 400, 125, 38 },
		{ 400, 126, 38 },
		{ 400, 127, 38 },
		{ 400, 128, 38 },
		{ 400, 129, 38 },
		{ 400, 130, 38 },
		{ 400, 131, 38 },
		{ 401, 93, 38 },
		{ 401, 94, 38 },
		{ 401, 95, 38 },
		{ 401, 96, 38 },
		{ 401, 97, 38 },
		{ 401, 98, 38 },
		{ 401, 99, 38 },
		{ 401, 100, 38 },
		{ 401, 101, 38 },
		{ 401, 102, 38 },
		{ 401, 103, 38 },
		{ 401, 104, 38 },
		{ 401, 105, 38 },
		{ 401, 106, 38 },
		{ 401, 107, 38 },
		{ 401, 108, 38 },
		{ 401, 109, 38 },
		{ 401, 110, 38 },
		{ 401, 111, 38 },
		{ 401, 112, 38 },
		{ 401, 113, 38 },
		{ 401, 114, 38 },
		{ 401, 115, 38 },
		{ 401, 116, 38 },
		{ 401, 117, 38 },
		{ 401, 118, 38 },
		{ 401, 119, 38 },
		{ 401, 120, 38 },
		{ 401, 121, 38 },
		{ 401, 122, 38 },
		{ 401, 123, 38 },
		{ 401, 124, 38 },
		{ 401, 125, 38 },
		{ 401, 126, 38 },
		{ 401, 127, 38 },
		{ 401, 128, 38 },
		{ 401, 129, 38 },
		{ 401, 130, 38 },
		{ 402, 92, 38 },
		{ 402, 93, 38 },
		{ 402, 94, 38 },
		{ 402, 95, 38 },
		{ 402, 96, 38 },
		{ 402, 97, 38 },
		{ 402, 98, 38 },
		{ 402, 99, 38 },
		{ 402, 100, 38 },
		{ 402, 101, 38 },
		{ 402, 102, 38 },
		{ 402, 103, 38 },
		{ 402, 104, 38 },
		{ 402, 105, 38 },
		{ 402, 106, 38 },
		{ 402, 107, 38 },
		{ 402, 108, 38 },
		{ 402, 109, 38 },
		{ 402, 110, 38 },
		{ 402, 111, 38 },
		{ 402, 112, 38 },
		{ 402, 113, 38 },
		{ 402, 114, 38 },
		{ 402, 115, 38 },
		{ 402, 116, 38 },
		{ 402, 117, 38 },
		{ 402, 118, 38 },
		{ 402, 119, 38 },
		{ 402, 120, 38 },
		{ 402, 121, 38 },
		{ 402, 122, 38 },
		{ 402, 123, 38 },
		{ 402, 124, 38 },
		{ 402, 125, 38 },
		{ 402, 126, 38 },
		{ 402, 127, 38 },
		{ 402, 128, 38 },
		{ 402, 129, 38 },
		{ 402, 130, 38 },
		{ 403, 91, 38 },
		{ 403, 92, 38 },
		{ 403, 93, 38 },
		{ 403, 94, 38 },
		{ 403, 95, 38 },
		{ 403, 96, 38 },
		{ 403, 97, 38 },
		{ 403, 98, 38 },
		{ 403, 99, 38 },
		{ 403, 100, 38 },
		{ 403, 101, 38 },
		{ 403, 102, 38 },
		{ 403, 103, 38 },
		{ 403, 104, 38 },
		{ 403, 105, 38 },
		{ 403, 106, 38 },
		{ 403, 107, 38 },
		{ 403, 108, 38 },
		{ 403, 109, 38 },
		{ 403, 110, 38 },
		{ 403, 111, 38 },
		{ 403, 112, 38 },
		{ 403, 113, 38 },
		{ 403, 114, 38 },
		{ 403, 115, 38 },
		{ 403, 116, 38 },
		{ 403, 117, 38 },
		{ 403, 118, 38 },
		{ 403, 119, 38 },
		{ 403, 120, 38 },
		{ 403, 121, 38 },
		{ 403, 122, 38 },
		{ 403, 123, 38 },
		{ 403, 124, 38 },
		{ 403, 125, 38 },
		{ 403, 126, 38 },
		{ 403, 127, 38 },
		{ 403, 128, 38 },
		{ 403, 129, 38 },
		{ 404, 90, 38 },
		{ 404, 91, 38 },
		{ 404, 92, 38 },
		{ 404, 93, 38 },
		{ 404, 94, 38 },
		{ 404, 95, 38 },
		{ 404, 96, 38 },
		{ 404, 97, 38 },
		{ 404, 98, 38 },
		{ 404, 99, 38 },
		{ 404, 100, 38 },
		{ 404, 101, 38 },
		{ 404, 102, 38 },
		{ 404, 103, 38 },
		{ 404, 104, 38 },
		{ 404, 105, 38 },
		{ 404, 106, 38 },
		{ 404, 107, 38 },
		{ 404, 108, 38 },
		{ 404, 109, 38 },
		{ 404, 110, 38 },
		{ 404, 111, 38 },
		{ 404, 112, 38 },
		{ 404, 113, 38 },
		{ 404, 114, 38 },
		{ 404, 115, 38 },
		{ 404, 116, 38 },
		{ 404, 117, 38 },
		{ 404, 118, 38 },
		{ 404, 119, 38 },
		{ 404, 120, 38 },
		{ 404, 121, 38 },
		{ 404, 122, 38 },
		{ 404, 123, 38 },
		{ 404, 124, 38 },
		{ 404, 125, 38 },
		{ 404, 126, 38 },
		{ 404, 127, 38 },
		{ 404, 128, 38 },
		{ 405, 88, 38 },
		{ 405, 89, 38 },
		{ 405, 90, 38 },
		{ 405, 91, 38 },
		{ 405, 92, 38 },
		{ 405, 93, 38 },
		{ 405, 94, 38 },
		{ 405, 95, 38 },
		{ 405, 96, 38 },
		{ 405, 97, 38 },
		{ 405, 98, 38 },
		{ 405, 99, 38 },
		{ 405, 100, 38 },
		{ 405, 101, 38 },
		{ 405, 102, 38 },
		{ 405, 103, 38 },
		{ 405, 104, 38 },
		{ 405, 105, 38 },
		{ 405, 106, 38 },
		{ 405, 107, 38 },
		{ 405, 108, 38 },
		{ 405, 109, 38 },
		{ 405, 110, 38 },
		{ 405, 111, 38 },
		{ 405, 112, 38 },
		{ 405, 113, 38 },
		{ 405, 114, 38 },
		{ 405, 115, 38 },
		{ 405, 116, 38 },
		{ 405, 117, 38 },
		{ 405, 118, 38 },
		{ 405, 119, 38 },
		{ 405, 120, 38 },
		{ 405, 121, 38 },
		{ 405, 122, 38 },
		{ 405, 123, 38 },
		{ 405, 124, 38 },
		{ 405, 125, 38 },
		{ 405, 126, 38 },
		{ 405, 127, 38 },
		{ 406, 87, 38 },
		{ 406, 88, 38 },
		{ 406, 89, 38 },
		{ 406, 90, 38 },
		{ 406, 91, 38 },
		{ 406, 92, 38 },
		{ 406, 93, 38 },
		{ 406, 94, 38 },
		{ 406, 95, 38 },
		{ 406, 96, 38 },
		{ 406, 97, 38 },
		{ 406, 98, 38 },
		{ 406, 99, 38 },
		{ 406, 100, 38 },
		{ 406, 101, 38 },
		{ 406, 102, 38 },
		{ 406, 103, 38 },
		{ 406, 104, 38 },
		{ 406, 105, 38 },
		{ 406, 106, 38 },
		{ 406, 107, 38 },
		{ 406, 108, 38 },
		{ 406, 109, 38 },
		{ 406, 110, 38 },
		{ 406, 111, 38 },
		{ 406, 112, 38 },
		{ 406, 113, 38 },
		{ 406, 114, 38 },
		{ 406, 115, 38 },
		{ 406, 116, 38 },
		{ 406, 117, 38 },
		{ 406, 118, 38 },
		{ 406, 119, 38 },
		{ 406, 120, 38 },
		{ 406, 121, 38 },
		{ 406, 122, 38 },
		{ 406, 123, 38 },
		{ 406, 124, 38 },
		{ 406, 125, 38 },
		{ 406, 126, 38 },
		{ 407, 86, 38 },
		{ 407, 87, 38 },
		{ 407, 88, 38 },
		{ 407, 89, 38 },
		{ 407, 90, 38 },
		{ 407, 91, 38 },
		{ 407, 92, 38 },
		{ 407, 93, 38 },
		{ 407, 94, 38 },
		{ 407, 95, 38 },
		{ 407, 96, 38 },
		{ 407, 97, 38 },
		{ 407, 98, 38 },
		{ 407, 99, 38 },
		{ 407, 100, 38 },
		{ 407, 101, 38 },
		{ 407, 102, 38 },
		{ 407, 103, 38 },
		{ 407, 104, 38 },
		{ 407, 105, 38 },
		{ 407, 106, 38 },
		{ 407, 107, 38 },
		{ 407, 108, 38 },
		{ 407, 109, 38 },
		{ 407, 110, 38 },
		{ 407, 111, 38 },
		{ 407, 112, 38 },
		{ 407, 113, 38 },
		{ 407, 114, 38 },
		{ 407, 115, 38 },
		{ 407, 116, 38 },
		{ 407, 117, 38 },
		{ 407, 118, 38 },
		{ 407, 119, 38 },
		{ 407, 120, 38 },
		{ 407, 121, 38 },
		{ 407, 122, 38 },
		{ 407, 123, 38 },
		{ 407, 124, 38 },
		{ 407, 125, 38 },
		{ 408, 85, 38 },
		{ 408, 86, 38 },
		{ 408, 87, 38 },
		{ 408, 88, 38 },
		{ 408, 89, 38 },
		{ 408, 90, 38 },
		{ 408, 91, 38 },
		{ 408, 92, 38 },
		{ 408, 93, 38 },
		{ 408, 94, 38 },
		{ 408, 95, 38 },
		{ 408, 96, 38 },
		{ 408, 97, 38 },
		{ 408, 98, 38 },
		{ 408, 99, 38 },
		{ 408, 100, 38 },
		{ 408, 101, 38 },
		{ 408, 102, 38 },
		{ 408, 103, 38 },
		{ 408, 104, 38 },
		{ 408, 105, 38 },
		{ 408, 106, 38 },
		{ 408, 107, 38 },
		{ 408, 108, 38 },
		{ 408, 109, 38 },
		{ 408, 110, 38 },
		{ 408, 111, 38 },
		{ 408, 112, 38 },
		{ 408, 113, 38 },
		{ 408, 114, 38 },
		{ 408, 115, 38 },
		{ 408, 116, 38 },
		{ 408, 117, 38 },
		{ 408, 118, 38 },
		{ 408, 119, 38 },
		{ 408, 120, 38 },
		{ 408, 121, 38 },
		{ 408, 122, 38 },
		{ 408, 123, 38 },
		{ 408, 124, 38 },
		{ 409, 84, 38 },
		{ 409, 85, 38 },
		{ 409, 86, 38 },
		{ 409, 87, 38 },
		{ 409, 88, 38 },
		{ 409, 89, 38 },
		{ 409, 90, 38 },
		{ 409, 91, 38 },
		{ 409, 92, 38 },
		{ 409, 93, 38 },
		{ 409, 94, 38 },
		{ 409, 95, 38 },
		{ 409, 96, 38 },
		{ 409, 97, 38 },
		{ 409, 98, 38 },
		{ 409, 99, 38 },
		{ 409, 100, 38 },
		{ 409, 101, 38 },
		{ 409, 102, 38 },
		{ 409, 103, 38 },
		{ 409, 104, 38 },
		{ 409, 105, 38 },
		{ 409, 106, 38 },
		{ 409, 107, 38 },
		{ 409, 108, 38 },
		{ 409, 109, 38 },
		{ 409, 110, 38 },
		{ 409, 111, 38 },
		{ 409, 112, 38 },
		{ 409, 113, 38 },
		{ 409, 114, 38 },
		{ 409, 115, 38 },
		{ 409, 116, 38 },
		{ 409, 117, 38 },
		{ 409, 118, 38 },
		{ 409, 119, 38 },
		{ 409, 120, 38 },
		{ 409, 121, 38 },
		{ 409, 122, 38 },
		{ 409, 123, 38 },
		{ 410, 83, 38 },
		{ 410, 84, 38 },
		{ 410, 85, 38 },
		{ 410, 86, 38 },
		{ 410, 87, 38 },
		{ 410, 88, 38 },
		{ 410, 89, 38 },
		{ 410, 90, 38 },
		{ 410, 91, 38 },
		{ 410, 92, 38 },
		{ 410, 93, 38 },
		{ 410, 94, 38 },
		{ 410, 95, 38 },
		{ 410, 96, 38 },
		{ 410, 97, 38 },
		{ 410, 98, 38 },
		{ 410, 99, 38 },
		{ 410, 100, 38 },
		{ 410, 101, 38 },
		{ 410, 102, 38 },
		{ 410, 103, 38 },
		{ 410, 104, 38 },
		{ 410, 105, 38 },
		{ 410, 106, 38 },
		{ 410, 107, 38 },
		{ 410, 108, 38 },
		{ 410, 109, 38 },
		{ 410, 110, 38 },
		{ 410, 111, 38 },
		{ 410, 112, 38 },
		{ 410, 113, 38 },
		{ 410, 114, 38 },
		{ 410, 115, 38 },
		{ 410, 116, 38 },
		{ 410, 117, 38 },
		{ 410, 118, 38 },
		{ 410, 119, 38 },
		{ 410, 120, 38 },
		{ 410, 121, 38 },
		{ 410, 122, 38 },
		{ 410, 123, 38 },
		{ 411, 82, 38 },
		{ 411, 83, 38 },
		{ 411, 84, 38 },
		{ 411, 85, 38 },
		{ 411, 86, 38 },
		{ 411, 87, 38 },
		{ 411, 88, 38 },
		{ 411, 89, 38 },
		{ 411, 90, 38 },
		{ 411, 91, 38 },
		{ 411, 92, 38 },
		{ 411, 93, 38 },
		{ 411, 94, 38 },
		{ 411, 95, 38 },
		{ 411, 96, 38 },
		{ 411, 97, 38 },
		{ 411, 98, 38 },
		{ 411, 99, 38 },
		{ 411, 100, 38 },
		{ 411, 101, 38 },
		{ 411, 102, 38 },
		{ 411, 103, 38 },
		{ 411, 104, 38 },
		{ 411, 105, 38 },
		{ 411, 106, 38 },
		{ 411, 107, 38 },
		{ 411, 108, 38 },
		{ 411, 109, 38 },
		{ 411, 110, 38 },
		{ 411, 111, 38 },
		{ 411, 112, 38 },
		{ 411, 113, 38 },
		{ 411, 114, 38 },
		{ 411, 115, 38 },
		{ 411, 116, 38 },
		{ 411, 117, 38 },
		{ 411, 118, 38 },
		{ 411, 119, 38 },
		{ 411, 120, 38 },
		{ 411, 121, 38 },
		{ 411, 122, 38 },
		{ 412, 80, 38 },
		{ 412, 81, 38 },
		{ 412, 82, 38 },
		{ 412, 83, 38 },
		{ 412, 84, 38 },
		{ 412, 85, 38 },
		{ 412, 86, 38 },
		{ 412, 87, 38 },
		{ 412, 88, 38 },
		{ 412, 89, 38 },
		{ 412, 90, 38 },
		{ 412, 91, 38 },
		{ 412, 92, 38 },
		{ 412, 93, 38 },
		{ 412, 94, 38 },
		{ 412, 95, 38 },
		{ 412, 96, 38 },
		{ 412, 97, 38 },
		{ 412, 98, 38 },
		{ 412, 99, 38 },
		{ 412, 100, 38 },
		{ 412, 101, 38 },
		{ 412, 102, 38 },
		{ 412, 103, 38 },
		{ 412, 104, 38 },
		{ 412, 105, 38 },
		{ 412, 106, 38 },
		{ 412, 107, 38 },
		{ 412, 108, 38 },
		{ 412, 109, 38 },
		{ 412, 110, 38 },
		{ 412, 111, 38 },
		{ 412, 112, 38 },
		{ 412, 113, 38 },
		{ 412, 114, 38 },
		{ 412, 115, 38 },
		{ 412, 116, 38 },
		{ 412, 117, 38 },
		{ 412, 118, 38 },
		{ 412, 119, 38 },
		{ 412, 120, 38 },
		{ 412, 121, 38 },
		{ 413, 79, 38 },
		{ 413, 80, 38 },
		{ 413, 81, 38 },
		{ 413, 82, 38 },
		{ 413, 83, 38 },
		{ 413, 84, 38 },
		{ 413, 85, 38 },
		{ 413, 86, 38 },
		{ 413, 87, 38 },
		{ 413, 88, 38 },
		{ 413, 89, 38 },
		{ 413, 90, 38 },
		{ 413, 91, 38 },
		{ 413, 92, 38 },
		{ 413, 93, 38 },
		{ 413, 94, 38 },
		{ 413, 95, 38 },
		{ 413, 96, 38 },
		{ 413, 97, 38 },
		{ 413, 98, 38 },
		{ 413, 99, 38 },
		{ 413, 100, 38 },
		{ 413, 101, 38 },
		{ 413, 102, 38 },
		{ 413, 103, 38 },
		{ 413, 104, 38 },
		{ 413, 105, 38 },
		{ 413, 106, 38 },
		{ 413, 107, 38 },
		{ 413, 108, 38 },
		{ 413, 109, 38 },
		{ 413, 110, 38 },
		{ 413, 111, 38 },
		{ 413, 112, 38 },
		{ 413, 113, 38 },
		{ 413, 114, 38 },
		{ 413, 115, 38 },
		{ 413, 116, 38 },
		{ 413, 117, 38 },
		{ 413, 118, 38 },
		{ 413, 119, 38 },
		{ 413, 120, 38 },
		{ 414, 78, 38 },
		{ 414, 79, 38 },
		{ 414, 80, 38 },
		{ 414, 81, 38 },
		{ 414, 82, 38 },
		{ 414, 83, 38 },
		{ 414, 84, 38 },
		{ 414, 85, 38 },
		{ 414, 86, 38 },
		{ 414, 87, 38 },
		{ 414, 88, 38 },
		{ 414, 89, 38 },
		{ 414, 90, 38 },
		{ 414, 91, 38 },
		{ 414, 92, 38 },
		{ 414, 93, 38 },
		{ 414, 94, 38 },
		{ 414, 95, 38 },
		{ 414, 96, 38 },
		{ 414, 97, 38 },
		{ 414, 98, 38 },
		{ 414, 99, 38 },
		{ 414, 100, 38 },
		{ 414, 101, 38 },
		{ 414, 102, 38 },
		{ 414, 103, 38 },
		{ 414, 104, 38 },
		{ 414, 105, 38 },
		{ 414, 106, 38 },
		{ 414, 107, 38 },
		{ 414, 108, 38 },
		{ 414, 109, 38 },
		{ 414, 110, 38 },
		{ 414, 111, 38 },
		{ 414, 112, 38 },
		{ 414, 113, 38 },
		{ 414, 114, 38 },
		{ 414, 115, 38 },
		{ 414, 116, 38 },
		{ 414, 117, 38 },
		{ 414, 118, 38 },
		{ 414, 119, 38 },
		{ 415, 77, 38 },
		{ 415, 78, 38 },
		{ 415, 79, 38 },
		{ 415, 80, 38 },
		{ 415, 81, 38 },
		{ 415, 82, 38 },
		{ 415, 83, 38 },
		{ 415, 84, 38 },
		{ 415, 85, 38 },
		{ 415, 86, 38 },
		{ 415, 87, 38 },
		{ 415, 88, 38 },
		{ 415, 89, 38 },
		{ 415, 90, 38 },
		{ 415, 91, 38 },
		{ 415, 92, 38 },
		{ 415, 93, 38 },
		{ 415, 94, 38 },
		{ 415, 95, 38 },
		{ 415, 96, 38 },
		{ 415, 97, 38 },
		{ 415, 98, 38 },
		{ 415, 99, 38 },
		{ 415, 100, 38 },
		{ 415, 101, 38 },
		{ 415, 102, 38 },
		{ 415, 103, 38 },
		{ 415, 104, 38 },
		{ 415, 105, 38 },
		{ 415, 106, 38 },
		{ 415, 107, 38 },
		{ 415, 108, 38 },
		{ 415, 109, 38 },
		{ 415, 110, 38 },
		{ 415, 111, 38 },
		{ 415, 112, 38 },
		{ 415, 113, 38 },
		{ 415, 114, 38 },
		{ 415, 115, 38 },
		{ 415, 116, 38 },
		{ 415, 117, 38 },
		{ 415, 118, 38 },
		{ 416, 76, 38 },
		{ 416, 77, 38 },
		{ 416, 78, 38 },
		{ 416, 79, 38 },
		{ 416, 80, 38 },
		{ 416, 81, 38 },
		{ 416, 82, 38 },
		{ 416, 83, 38 },
		{ 416, 84, 38 },
		{ 416, 85, 38 },
		{ 416, 86, 38 },
		{ 416, 87, 38 },
		{ 416, 88, 38 },
		{ 416, 89, 38 },
		{ 416, 90, 38 },
		{ 416, 91, 38 },
		{ 416, 92, 38 },
		{ 416, 93, 38 },
		{ 416, 94, 38 },
		{ 416, 95, 38 },
		{ 416, 96, 38 },
		{ 416, 97, 38 },
		{ 416, 98, 38 },
		{ 416, 99, 38 },
		{ 416, 100, 38 },
		{ 416, 101, 38 },
		{ 416, 102, 38 },
		{ 416, 103, 38 },
		{ 416, 104, 38 },
		{ 416, 105, 38 },
		{ 416, 106, 38 },
		{ 416, 107, 38 },
		{ 416, 108, 38 },
		{ 416, 109, 38 },
		{ 416, 110, 38 },
		{ 416, 111, 38 },
		{ 416, 112, 38 },
		{ 416, 113, 38 },
		{ 416, 114, 38 },
		{ 416, 115, 38 },
		{ 416, 116, 38 },
		{ 416, 117, 38 },
		{ 417, 75, 38 },
		{ 417, 76, 38 },
		{ 417, 77, 38 },
		{ 417, 78, 38 },
		{ 417, 79, 38 },
		{ 417, 80, 38 },
		{ 417, 81, 38 },
		{ 417, 82, 38 },
		{ 417, 83, 38 },
		{ 417, 84, 38 },
		{ 417, 85, 38 },
		{ 417, 86, 38 },
		{ 417, 87, 38 },
		{ 417, 88, 38 },
		{ 417, 89, 38 },
		{ 417, 90, 38 },
		{ 417, 91, 38 },
		{ 417, 92, 38 },
		{ 417, 93, 38 },
		{ 417, 94, 38 },
		{ 417, 95, 38 },
		{ 417, 96, 38 },
		{ 417, 97, 38 },
		{ 417, 98, 38 },
		{ 417, 99, 38 },
		{ 417, 100, 38 },
		{ 417, 101, 38 },
		{ 417, 102, 38 },
		{ 417, 103, 38 },
		{ 417, 104, 38 },
		{ 417, 105, 38 },
		{ 417, 106, 38 },
		{ 417, 107, 38 },
		{ 417, 108, 38 },
		{ 417, 109, 38 },
		{ 417, 110, 38 },
		{ 417, 111, 38 },
		{ 417, 112, 38 },
		{ 417, 113, 38 },
		{ 417, 114, 38 },
		{ 417, 115, 38 },
		{ 417, 116, 38 },
		{ 418, 74, 38 },
		{ 418, 75, 38 },
		{ 418, 76, 38 },
		{ 418, 77, 38 },
		{ 418, 78, 38 },
		{ 418, 79, 38 },
		{ 418, 80, 38 },
		{ 418, 81, 38 },
		{ 418, 82, 38 },
		{ 418, 83, 38 },
		{ 418, 84, 38 },
		{ 418, 85, 38 },
		{ 418, 86, 38 },
		{ 418, 87, 38 },
		{ 418, 88, 38 },
		{ 418, 89, 38 },
		{ 418, 90, 38 },
		{ 418, 91, 38 },
		{ 418, 92, 38 },
		{ 418, 93, 38 },
		{ 418, 94, 38 },
		{ 418, 95, 38 },
		{ 418, 96, 38 },
		{ 418, 97, 38 },
		{ 418, 98, 38 },
		{ 418, 99, 38 },
		{ 418, 100, 38 },
		{ 418, 101, 38 },
		{ 418, 102, 38 },
		{ 418, 103, 38 },
		{ 418, 104, 38 },
		{ 418, 105, 38 },
		{ 418, 106, 38 },
		{ 418, 107, 38 },
		{ 418, 108, 38 },
		{ 418, 109, 38 },
		{ 418, 110, 38 },
		{ 418, 111, 38 },
		{ 418, 112, 38 },
		{ 418, 113, 38 },
		{ 418, 114, 38 },
		{ 418, 115, 38 },
		{ 418, 116, 38 },
		{ 419, 72, 38 },
		{ 419, 73, 38 },
		{ 419, 74, 38 },
		{ 419, 75, 38 },
		{ 419, 76, 38 },
		{ 419, 77, 38 },
		{ 419, 78, 38 },
		{ 419, 79, 38 },
		{ 419, 80, 38 },
		{ 419, 81, 38 },
		{ 419, 82, 38 },
		{ 419, 83, 38 },
		{ 419, 84, 38 },
		{ 419, 85, 38 },
		{ 419, 86, 38 },
		{ 419, 87, 38 },
		{ 419, 88, 38 },
		{ 419, 89, 38 },
		{ 419, 90, 38 },
		{ 419, 91, 38 },
		{ 419, 92, 38 },
		{ 419, 93, 38 },
		{ 419, 94, 38 },
		{ 419, 95, 38 },
		{ 419, 96, 38 },
		{ 419, 97, 38 },
		{ 419, 98, 38 },
		{ 419, 99, 38 },
		{ 419, 100, 38 },
		{ 419, 101, 38 },
		{ 419, 102, 38 },
		{ 419, 103, 38 },
		{ 419, 104, 38 },
		{ 419, 105, 38 },
		{ 419, 106, 38 },
		{ 419, 107, 38 },
		{ 419, 108, 38 },
		{ 419, 109, 38 },
		{ 419, 110, 38 },
		{ 419, 111, 38 },
		{ 419, 112, 38 },
		{ 419, 113, 38 },
		{ 419, 114, 38 },
		{ 419, 115, 38 },
		{ 420, 71, 38 },
		{ 420, 72, 38 },
		{ 420, 73, 38 },
		{ 420, 74, 38 },
		{ 420, 75, 38 },
		{ 420, 76, 38 },
		{ 420, 77, 38 },
		{ 420, 78, 38 },
		{ 420, 79, 38 },
		{ 420, 80, 38 },
		{ 420, 81, 38 },
		{ 420, 82, 38 },
		{ 420, 83, 38 },
		{ 420, 84, 38 },
		{ 420, 85, 38 },
		{ 420, 86, 38 },
		{ 420, 87, 38 },
		{ 420, 88, 38 },
		{ 420, 89, 38 },
		{ 420, 90, 38 },
		{ 420, 91, 38 },
		{ 420, 92, 38 },
		{ 420, 93, 38 },
		{ 420, 94, 38 },
		{ 420, 95, 38 },
		{ 420, 96, 38 },
		{ 420, 97, 38 },
		{ 420, 98, 38 },
		{ 420, 99, 38 },
		{ 420, 100, 38 },
		{ 420, 101, 38 },
		{ 420, 102, 38 },
		{ 420, 103, 38 },
		{ 420, 104, 38 },
		{ 420, 105, 38 },
		{ 420, 106, 38 },
		{ 420, 107, 38 },
		{ 420, 108, 38 },
		{ 420, 109, 38 },
		{ 420, 110, 38 },
		{ 420, 111, 38 },
		{ 420, 112, 38 },
		{ 420, 113, 38 },
		{ 420, 114, 38 },
		{ 421, 70, 38 },
		{ 421, 71, 38 },
		{ 421, 72, 38 },
		{ 421, 73, 38 },
		{ 421, 74, 38 },
		{ 421, 75, 38 },
		{ 421, 76, 38 },
		{ 421, 77, 38 },
		{ 421, 78, 38 },
		{ 421, 79, 38 },
		{ 421, 80, 38 },
		{ 421, 81, 38 },
		{ 421, 82, 38 },
		{ 421, 83, 38 },
		{ 421, 84, 38 },
		{ 421, 85, 38 },
		{ 421, 86, 38 },
		{ 421, 87, 38 },
		{ 421, 88, 38 },
		{ 421, 89, 38 },
		{ 421, 90, 38 },
		{ 421, 91, 38 },
		{ 421, 92, 38 },
		{ 421, 93, 38 },
		{ 421, 94, 38 },
		{ 421, 95, 38 },
		{ 421, 96, 38 },
		{ 421, 97, 38 },
		{ 421, 98, 38 },
		{ 421, 99, 38 },
		{ 421, 100, 38 },
		{ 421, 101, 38 },
		{ 421, 102, 38 },
		{ 421, 103, 38 },
		{ 421, 104, 38 },
		{ 421, 105, 38 },
		{ 421, 106, 38 },
		{ 421, 107, 38 },
		{ 421, 108, 38 },
		{ 421, 109, 38 },
		{ 421, 110, 38 },
		{ 421, 111, 38 },
		{ 421, 112, 38 },
		{ 421, 113, 38 },
		{ 422, 69, 38 },
		{ 422, 70, 38 },
		{ 422, 71, 38 },
		{ 422, 72, 38 },
		{ 422, 73, 38 },
		{ 422, 74, 38 },
		{ 422, 75, 38 },
		{ 422, 76, 38 },
		{ 422, 77, 38 },
		{ 422, 78, 38 },
		{ 422, 79, 38 },
		{ 422, 80, 38 },
		{ 422, 81, 38 },
		{ 422, 82, 38 },
		{ 422, 83, 38 },
		{ 422, 84, 38 },
		{ 422, 85, 38 },
		{ 422, 86, 38 },
		{ 422, 87, 38 },
		{ 422, 88, 38 },
		{ 422, 89, 38 },
		{ 422, 90, 38 },
		{ 422, 91, 38 },
		{ 422, 92, 38 },
		{ 422, 93, 38 },
		{ 422, 94, 38 },
		{ 422, 95, 38 },
		{ 422, 96, 38 },
		{ 422, 97, 38 },
		{ 422, 98, 38 },
		{ 422, 99, 38 },
		{ 422, 100, 38 },
		{ 422, 101, 38 },
		{ 422, 102, 38 },
		{ 422, 103, 38 },
		{ 422, 104, 38 },
		{ 422, 105, 38 },
		{ 422, 106, 38 },
		{ 422, 107, 38 },
		{ 422, 108, 38 },
		{ 422, 109, 38 },
		{ 422, 110, 38 },
		{ 422, 111, 38 },
		{ 422, 112, 38 },
		{ 423, 68, 38 },
		{ 423, 69, 38 },
		{ 423, 70, 38 },
		{ 423, 71, 38 },
		{ 423, 72, 38 },
		{ 423, 73, 38 },
		{ 423, 74, 38 },
		{ 423, 75, 38 },
		{ 423, 76, 38 },
		{ 423, 77, 38 },
		{ 423, 78, 38 },
		{ 423, 79, 38 },
		{ 423, 80, 38 },
		{ 423, 81, 38 },
		{ 423, 82, 38 },
		{ 423, 83, 38 },
		{ 423, 84, 38 },
		{ 423, 85, 38 },
		{ 423, 86, 38 },
		{ 423, 87, 38 },
		{ 423, 88, 38 },
		{ 423, 89, 38 },
		{ 423, 90, 38 },
		{ 423, 91, 38 },
		{ 423, 92, 38 },
		{ 423, 93, 38 },
		{ 423, 94, 38 },
		{ 423, 95, 38 },
		{ 423, 96, 38 },
		{ 423, 97, 38 },
		{ 423, 98, 38 },
		{ 423, 99, 38 },
		{ 423, 100, 38 },
		{ 423, 101, 38 },
		{ 423, 102, 38 },
		{ 423, 103, 38 },
		{ 423, 104, 38 },
		{ 423, 105, 38 },
		{ 423, 106, 38 },
		{ 423, 107, 38 },
		{ 423, 108, 38 },
		{ 423, 109, 38 },
		{ 423, 110, 38 },
		{ 423, 111, 38 },
		{ 424, 67, 38 },
		{ 424, 68, 38 },
		{ 424, 69, 38 },
		{ 424, 70, 38 },
		{ 424, 71, 38 },
		{ 424, 72, 38 },
		{ 424, 73, 38 },
		{ 424, 74, 38 },
		{ 424, 75, 38 },
		{ 424, 76, 38 },
		{ 424, 77, 38 },
		{ 424, 78, 38 },
		{ 424, 79, 38 },
		{ 424, 80, 38 },
		{ 424, 81, 38 },
		{ 424, 82, 38 },
		{ 424, 83, 38 },
		{ 424, 84, 38 },
		{ 424, 85, 38 },
		{ 424, 86, 38 },
		{ 424, 87, 38 },
		{ 424, 88, 38 },
		{ 424, 89, 38 },
		{ 424, 90, 38 },
		{ 424, 91, 38 },
		{ 424, 92, 38 },
		{ 424, 93, 38 },
		{ 424, 94, 38 },
		{ 424, 95, 38 },
		{ 424, 96, 38 },
		{ 424, 97, 38 },
		{ 424, 98, 38 },
		{ 424, 99, 38 },
		{ 424, 100, 38 },
		{ 424, 101, 38 },
		{ 424, 102, 38 },
		{ 424, 103, 38 },
		{ 424, 104, 38 },
		{ 424, 105, 38 },
		{ 424, 106, 38 },
		{ 424, 107, 38 },
		{ 424, 108, 38 },
		{ 424, 109, 38 },
		{ 424, 110, 38 },
		{ 425, 66, 38 },
		{ 425, 67, 38 },
		{ 425, 68, 38 },
		{ 425, 69, 38 },
		{ 425, 70, 38 },
		{ 425, 71, 38 },
		{ 425, 72, 38 },
		{ 425, 73, 38 },
		{ 425, 74, 38 },
		{ 425, 75, 38 },
		{ 425, 76, 38 },
		{ 425, 77, 38 },
		{ 425, 78, 38 },
		{ 425, 79, 38 },
		{ 425, 80, 38 },
		{ 425, 81, 38 },
		{ 425, 82, 38 },
		{ 425, 83, 38 },
		{ 425, 84, 38 },
		{ 425, 85, 38 },
		{ 425, 86, 38 },
		{ 425, 87, 38 },
		{ 425, 88, 38 },
		{ 425, 89, 38 },
		{ 425, 90, 38 },
		{ 425, 91, 38 },
		{ 425, 92, 38 },
		{ 425, 93, 38 },
		{ 425, 94, 38 },
		{ 425, 95, 38 },
		{ 425, 96, 38 },
		{ 425, 97, 38 },
		{ 425, 98, 38 },
		{ 425, 99, 38 },
		{ 425, 100, 38 },
		{ 425, 101, 38 },
		{ 425, 102, 38 },
		{ 425, 103, 38 },
		{ 425, 104, 38 },
		{ 425, 105, 38 },
		{ 425, 106, 38 },
		{ 425, 107, 38 },
		{ 425, 108, 38 },
		{ 425, 109, 38 },
		{ 426, 64, 38 },
		{ 426, 65, 38 },
		{ 426, 66, 38 },
		{ 426, 67, 38 },
		{ 426, 68, 38 },
		{ 426, 69, 38 },
		{ 426, 70, 38 },
		{ 426, 71, 38 },
		{ 426, 72, 38 },
		{ 426, 73, 38 },
		{ 426, 74, 38 },
		{ 426, 75, 38 },
		{ 426, 76, 38 },
		{ 426, 77, 38 },
		{ 426, 78, 38 },
		{ 426, 79, 38 },
		{ 426, 80, 38 },
		{ 426, 81, 38 },
		{ 426, 82, 38 },
		{ 426, 83, 38 },
		{ 426, 84, 38 },
		{ 426, 85, 38 },
		{ 426, 86, 38 },
		{ 426, 87, 38 },
		{ 426, 88, 38 },
		{ 426, 89, 38 },
		{ 426, 90, 38 },
		{ 426, 91, 38 },
		{ 426, 92, 38 },
		{ 426, 93, 38 },
		{ 426, 94, 38 },
		{ 426, 95, 38 },
		{ 426, 96, 38 },
		{ 426, 97, 38 },
		{ 426, 98, 38 },
		{ 426, 99, 38 },
		{ 426, 100, 38 },
		{ 426, 101, 38 },
		{ 426, 102, 38 },
		{ 426, 103, 38 },
		{ 426, 104, 38 },
		{ 426, 105, 38 },
		{ 426, 106, 38 },
		{ 426, 107, 38 },
		{ 426, 108, 38 },
		{ 426, 109, 38 },
		{ 427, 63, 38 },
		{ 427, 64, 38 },
		{ 427, 65, 38 },
		{ 427, 66, 38 },
		{ 427, 67, 38 },
		{ 427, 68, 38 },
		{ 427, 69, 38 },
		{ 427, 70, 38 },
		{ 427, 71, 38 },
		{ 427, 72, 38 },
		{ 427, 73, 38 },
		{ 427, 74, 38 },
		{ 427, 75, 38 },
		{ 427, 76, 38 },
		{ 427, 77, 38 },
		{ 427, 78, 38 },
		{ 427, 79, 38 },
		{ 427, 80, 38 },
		{ 427, 81, 38 },
		{ 427, 82, 38 },
		{ 427, 83, 38 },
		{ 427, 84, 38 },
		{ 427, 85, 38 },
		{ 427, 86, 38 },
		{ 427, 87, 38 },
		{ 427, 88, 38 },
		{ 427, 89, 38 },
		{ 427, 90, 38 },
		{ 427, 91, 38 },
		{ 427, 92, 38 },
		{ 427, 93, 38 },
		{ 427, 94, 38 },
		{ 427, 95, 38 },
		{ 427, 96, 38 },
		{ 427, 97, 38 },
		{ 427, 98, 38 },
		{ 427, 99, 38 },
		{ 427, 100, 38 },
		{ 427, 101, 38 },
		{ 427, 102, 38 },
		{ 427, 103, 38 },
		{ 427, 104, 38 },
		{ 427, 105, 38 },
		{ 427, 106, 38 },
		{ 427, 107, 38 },
		{ 427, 108, 38 },
		{ 428, 62, 38 },
		{ 428, 63, 38 },
		{ 428, 64, 38 },
		{ 428, 65, 38 },
		{ 428, 66, 38 },
		{ 428, 67, 38 },
		{ 428, 68, 38 },
		{ 428, 69, 38 },
		{ 428, 70, 38 },
		{ 428, 71, 38 },
		{ 428, 72, 38 },
		{ 428, 73, 38 },
		{ 428, 74, 38 },
		{ 428, 75, 38 },
		{ 428, 76, 38 },
		{ 428, 77, 38 },
		{ 428, 78, 38 },
		{ 428, 79, 38 },
		{ 428, 80, 38 },
		{ 428, 81, 38 },
		{ 428, 82, 38 },
		{ 428, 83, 38 },
		{ 428, 84, 38 },
		{ 428, 85, 38 },
		{ 428, 86, 38 },
		{ 428, 87, 38 },
		{ 428, 88, 38 },
		{ 428, 89, 38 },
		{ 428, 90, 38 },
		{ 428, 91, 38 },
		{ 428, 92, 38 },
		{ 428, 93, 38 },
		{ 428, 94, 38 },
		{ 428, 95, 38 },
		{ 428, 96, 38 },
		{ 428, 97, 38 },
		{ 428, 98, 38 },
		{ 428, 99, 38 },
		{ 428, 100, 38 },
		{ 428, 101, 38 },
		{ 428, 102, 38 },
		{ 428, 103, 38 },
		{ 428, 104, 38 },
		{ 428, 105, 38 },
		{ 428, 106, 38 },
		{ 428, 107, 38 },
		{ 429, 61, 38 },
		{ 429, 62, 38 },
		{ 429, 63, 38 },
		{ 429, 64, 38 },
		{ 429, 65, 38 },
		{ 429, 66, 38 },
		{ 429, 67, 38 },
		{ 429, 68, 38 },
		{ 429, 69, 38 },
		{ 429, 70, 38 },
		{ 429, 71, 38 },
		{ 429, 72, 38 },
		{ 429, 73, 38 },
		{ 429, 74, 38 },
		{ 429, 75, 38 },
		{ 429, 76, 38 },
		{ 429, 77, 38 },
		{ 429, 78, 38 },
		{ 429, 79, 38 },
		{ 429, 80, 38 },
		{ 429, 81, 38 },
		{ 429, 82, 38 },
		{ 429, 83, 38 },
		{ 429, 84, 38 },
		{ 429, 85, 38 },
		{ 429, 86, 38 },
		{ 429, 87, 38 },
		{ 429, 88, 38 },
		{ 429, 89, 38 },
		{ 429, 90, 38 },
		{ 429, 91, 38 },
		{ 429, 92, 38 },
		{ 429, 93, 38 },
		{ 429, 94, 38 },
		{ 429, 95, 38 },
		{ 429, 96, 38 },
		{ 429, 97, 38 },
		{ 429, 98, 38 },
		{ 429, 99, 38 },
		{ 429, 100, 38 },
		{ 429, 101, 38 },
		{ 429, 102, 38 },
		{ 429, 103, 38 },
		{ 429, 104, 38 },
		{ 429, 105, 38 },
		{ 429, 106, 38 },
		{ 430, 60, 38 },
		{ 430, 61, 38 },
		{ 430, 62, 38 },
		{ 430, 63, 38 },
		{ 430, 64, 38 },
		{ 430, 65, 38 },
		{ 430, 66, 38 },
		{ 430, 67, 38 },
		{ 430, 68, 38 },
		{ 430, 69, 38 },
		{ 430, 70, 38 },
		{ 430, 71, 38 },
		{ 430, 72, 38 },
		{ 430, 73, 38 },
		{ 430, 74, 38 },
		{ 430, 75, 38 },
		{ 430, 76, 38 },
		{ 430, 77, 38 },
		{ 430, 78, 38 },
		{ 430, 79, 38 },
		{ 430, 80, 38 },
		{ 430, 81, 38 },
		{ 430, 82, 38 },
		{ 430, 83, 38 },
		{ 430, 84, 38 },
		{ 430, 85, 38 },
		{ 430, 86, 38 },
		{ 430, 87, 38 },
		{ 430, 88, 38 },
		{ 430, 89, 38 },
		{ 430, 90, 38 },
		{ 430, 91, 38 },
		{ 430, 92, 38 },
		{ 430, 93, 38 },
		{ 430, 94, 38 },
		{ 430, 95, 38 },
		{ 430, 96, 38 },
		{ 430, 97, 38 },
		{ 430, 98, 38 },
		{ 430, 99, 38 },
		{ 430, 100, 38 },
		{ 430, 101, 38 },
		{ 430, 102, 38 },
		{ 430, 103, 38 },
		{ 430, 104, 38 },
		{ 430, 105, 38 },
		{ 431, 59, 38 },
		{ 431, 60, 38 },
		{ 431, 61, 38 },
		{ 431, 62, 38 },
		{ 431, 63, 38 },
		{ 431, 64, 38 },
		{ 431, 65, 38 },
		{ 431, 66, 38 },
		{ 431, 67, 38 },
		{ 431, 68, 38 },
		{ 431, 69, 38 },
		{ 431, 70, 38 },
		{ 431, 71, 38 },
		{ 431, 72, 38 },
		{ 431, 73, 38 },
		{ 431, 74, 38 },
		{ 431, 75, 38 },
		{ 431, 76, 38 },
		{ 431, 77, 38 },
		{ 431, 78, 38 },
		{ 431, 79, 38 },
		{ 431, 80, 38 },
		{ 431, 81, 38 },
		{ 431, 82, 38 },
		{ 431, 83, 38 },
		{ 431, 84, 38 },
		{ 431, 85, 38 },
		{ 431, 86, 38 },
		{ 431, 87, 38 },
		{ 431, 88, 38 },
		{ 431, 89, 38 },
		{ 431, 90, 38 },
		{ 431, 91, 38 },
		{ 431, 92, 38 },
		{ 431, 93, 38 },
		{ 431, 94, 38 },
		{ 431, 95, 38 },
		{ 431, 96, 38 },
		{ 431, 97, 38 },
		{ 431, 98, 38 },
		{ 431, 99, 38 },
		{ 431, 100, 38 },
		{ 431, 101, 38 },
		{ 431, 102, 38 },
		{ 431, 103, 38 },
		{ 431, 104, 38 },
		{ 432, 58, 38 },
		{ 432, 59, 38 },
		{ 432, 60, 38 },
		{ 432, 61, 38 },
		{ 432, 62, 38 },
		{ 432, 63, 38 },
		{ 432, 64, 38 },
		{ 432, 65, 38 },
		{ 432, 66, 38 },
		{ 432, 67, 38 },
		{ 432, 68, 38 },
		{ 432, 69, 38 },
		{ 432, 70, 38 },
		{ 432, 71, 38 },
		{ 432, 72, 38 },
		{ 432, 73, 38 },
		{ 432, 74, 38 },
		{ 432, 75, 38 },
		{ 432, 76, 38 },
		{ 432, 77, 38 },
		{ 432, 78, 38 },
		{ 432, 79, 38 },
		{ 432, 80, 38 },
		{ 432, 81, 38 },
		{ 432, 82, 38 },
		{ 432, 83, 38 },
		{ 432, 84, 38 },
		{ 432, 85, 38 },
		{ 432, 86, 38 },
		{ 432, 87, 38 },
		{ 432, 88, 38 },
		{ 432, 89, 38 },
		{ 432, 90, 38 },
		{ 432, 91, 38 },
		{ 432, 92, 38 },
		{ 432, 93, 38 },
		{ 432, 94, 38 },
		{ 432, 95, 38 },
		{ 432, 96, 38 },
		{ 432, 97, 38 },
		{ 432, 98, 38 },
		{ 432, 99, 38 },
		{ 432, 100, 38 },
		{ 432, 101, 38 },
		{ 432, 102, 38 },
		{ 432, 103, 38 },
		{ 433, 56, 38 },
		{ 433, 57, 38 },
		{ 433, 58, 38 },
		{ 433, 59, 38 },
		{ 433, 60, 38 },
		{ 433, 61, 38 },
		{ 433, 62, 38 },
		{ 433, 63, 38 },
		{ 433, 64, 38 },
		{ 433, 65, 38 },
		{ 433, 66, 38 },
		{ 433, 67, 38 },
		{ 433, 68, 38 },
		{ 433, 69, 38 },
		{ 433, 70, 38 },
		{ 433, 71, 38 },
		{ 433, 72, 38 },
		{ 433, 73, 38 },
		{ 433, 74, 38 },
		{ 433, 75, 38 },
		{ 433, 76, 38 },
		{ 433, 77, 38 },
		{ 433, 78, 38 },
		{ 433, 79, 38 },
		{ 433, 80, 38 },
		{ 433, 81, 38 },
		{ 433, 82, 38 },
		{ 433, 83, 38 },
		{ 433, 84, 38 },
		{ 433, 85, 38 },
		{ 433, 86, 38 },
		{ 433, 87, 38 },
		{ 433, 88, 38 },
		{ 433, 89, 38 },
		{ 433, 90, 38 },
		{ 433, 91, 38 },
		{ 433, 92, 38 },
		{ 433, 93, 38 },
		{ 433, 94, 38 },
		{ 433, 95, 38 },
		{ 433, 96, 38 },
		{ 433, 97, 38 },
		{ 433, 98, 38 },
		{ 433, 99, 38 },
		{ 433, 100, 38 },
		{ 433, 101, 38 },
		{ 433, 102, 38 },
		{ 434, 55, 38 },
		{ 434, 56, 38 },
		{ 434, 57, 38 },
		{ 434, 58, 38 },
		{ 434, 59, 38 },
		{ 434, 60, 38 },
		{ 434, 61, 38 },
		{ 434, 62, 38 },
		{ 434, 63, 38 },
		{ 434, 64, 38 },
		{ 434, 65, 38 },
		{ 434, 66, 38 },
		{ 434, 67, 38 },
		{ 434, 68, 38 },
		{ 434, 69, 38 },
		{ 434, 70, 38 },
		{ 434, 71, 38 },
		{ 434, 72, 38 },
		{ 434, 73, 38 },
		{ 434, 74, 38 },
		{ 434, 75, 38 },
		{ 434, 76, 38 },
		{ 434, 77, 38 },
		{ 434, 78, 38 },
		{ 434, 79, 38 },
		{ 434, 80, 38 },
		{ 434, 81, 38 },
		{ 434, 82, 38 },
		{ 434, 83, 38 },
		{ 434, 84, 38 },
		{ 434, 85, 38 },
		{ 434, 86, 38 },
		{ 434, 87, 38 },
		{ 434, 88, 38 },
		{ 434, 89, 38 },
		{ 434, 90, 38 },
		{ 434, 91, 38 },
		{ 434, 92, 38 },
		{ 434, 93, 38 },
		{ 434, 94, 38 },
		{ 434, 95, 38 },
		{ 434, 96, 38 },
		{ 434, 97, 38 },
		{ 434, 98, 38 },
		{ 434, 99, 38 },
		{ 434, 100, 38 },
		{ 434, 101, 38 },
		{ 434, 102, 38 },
		{ 435, 54, 38 },
		{ 435, 55, 38 },
		{ 435, 56, 38 },
		{ 435, 57, 38 },
		{ 435, 58, 38 },
		{ 435, 59, 38 },
		{ 435, 60, 38 },
		{ 435, 61, 38 },
		{ 435, 62, 38 },
		{ 435, 63, 38 },
		{ 435, 64, 38 },
		{ 435, 65, 38 },
		{ 435, 66, 38 },
		{ 435, 67, 38 },
		{ 435, 68, 38 },
		{ 435, 69, 38 },
		{ 435, 70, 38 },
		{ 435, 71, 38 },
		{ 435, 72, 38 },
		{ 435, 73, 38 },
		{ 435, 74, 38 },
		{ 435, 75, 38 },
		{ 435, 76, 38 },
		{ 435, 77, 38 },
		{ 435, 78, 38 },
		{ 435, 79, 38 },
		{ 435, 80, 38 },
		{ 435, 81, 38 },
		{ 435, 82, 38 },
		{ 435, 83, 38 },
		{ 435, 84, 38 },
		{ 435, 85, 38 },
		{ 435, 86, 38 },
		{ 435, 87, 38 },
		{ 435, 88, 38 },
		{ 435, 89, 38 },
		{ 435, 90, 38 },
		{ 435, 91, 38 },
		{ 435, 92, 38 },
		{ 435, 93, 38 },
		{ 435, 94, 38 },
		{ 435, 95, 38 },
		{ 435, 96, 38 },
		{ 435, 97, 38 },
		{ 435, 98, 38 },
		{ 435, 99, 38 },
		{ 435, 100, 38 },
		{ 435, 101, 38 },
		{ 436, 53, 38 },
		{ 436, 54, 38 },
		{ 436, 55, 38 },
		{ 436, 56, 38 },
		{ 436, 57, 38 },
		{ 436, 58, 38 },
		{ 436, 59, 38 },
		{ 436, 60, 38 },
		{ 436, 61, 38 },
		{ 436, 62, 38 },
		{ 436, 63, 38 },
		{ 436, 64, 38 },
		{ 436, 65, 38 },
		{ 436, 66, 38 },
		{ 436, 67, 38 },
		{ 436, 68, 38 },
		{ 436, 69, 38 },
		{ 436, 70, 38 },
		{ 436, 71, 38 },
		{ 436, 72, 38 },
		{ 436, 73, 38 },
		{ 436, 74, 38 },
		{ 436, 75, 38 },
		{ 436, 76, 38 },
		{ 436, 77, 38 },
		{ 436, 78, 38 },
		{ 436, 79, 38 },
		{ 436, 80, 38 },
		{ 436, 81, 38 },
		{ 436, 82, 38 },
		{ 436, 83, 38 },
		{ 436, 84, 38 },
		{ 436, 85, 38 },
		{ 436, 86, 38 },
		{ 436, 87, 38 },
		{ 436, 88, 38 },
		{ 436, 89, 38 },
		{ 436, 90, 38 },
		{ 436, 91, 38 },
		{ 436, 92, 38 },
		{ 436, 93, 38 },
		{ 436, 94, 38 },
		{ 436, 95, 38 },
		{ 436, 96, 38 },
		{ 436, 97, 38 },
		{ 436, 98, 38 },
		{ 436, 99, 38 },
		{ 436, 100, 38 },
		{ 437, 52, 38 },
		{ 437, 53, 38 },
		{ 437, 54, 38 },
		{ 437, 55, 38 },
		{ 437, 56, 38 },
		{ 437, 57, 38 },
		{ 437, 58, 38 },
		{ 437, 59, 38 },
		{ 437, 60, 38 },
		{ 437, 61, 38 },
		{ 437, 62, 38 },
		{ 437, 63, 38 },
		{ 437, 64, 38 },
		{ 437, 65, 38 },
		{ 437, 66, 38 },
		{ 437, 67, 38 },
		{ 437, 68, 38 },
		{ 437, 69, 38 },
		{ 437, 70, 38 },
		{ 437, 71, 38 },
		{ 437, 72, 38 },
		{ 437, 73, 38 },
		{ 437, 74, 38 },
		{ 437, 75, 38 },
		{ 437, 76, 38 },
		{ 437, 77, 38 },
		{ 437, 78, 38 },
		{ 437, 79, 38 },
		{ 437, 80, 38 },
		{ 437, 81, 38 },
		{ 437, 82, 38 },
		{ 437, 83, 38 },
		{ 437, 84, 38 },
		{ 437, 85, 38 },
		{ 437, 86, 38 },
		{ 437, 87, 38 },
		{ 437, 88, 38 },
		{ 437, 89, 38 },
		{ 437, 90, 38 },
		{ 437, 91, 38 },
		{ 437, 92, 38 },
		{ 437, 93, 38 },
		{ 437, 94, 38 },
		{ 437, 95, 38 },
		{ 437, 96, 38 },
		{ 437, 97, 38 },
		{ 437, 98, 38 },
		{ 437, 99, 38 },
		{ 438, 51, 38 },
		{ 438, 52, 38 },
		{ 438, 53, 38 },
		{ 438, 54, 38 },
		{ 438, 55, 38 },
		{ 438, 56, 38 },
		{ 438, 57, 38 },
		{ 438, 58, 38 },
		{ 438, 59, 38 },
		{ 438, 60, 38 },
		{ 438, 61, 38 },
		{ 438, 62, 38 },
		{ 438, 63, 38 },
		{ 438, 64, 38 },
		{ 438, 65, 38 },
		{ 438, 66, 38 },
		{ 438, 67, 38 },
		{ 438, 68, 38 },
		{ 438, 69, 38 },
		{ 438, 70, 38 },
		{ 438, 71, 38 },
		{ 438, 72, 38 },
		{ 438, 73, 38 },
		{ 438, 74, 38 },
		{ 438, 75, 38 },
		{ 438, 76, 38 },
		{ 438, 77, 38 },
		{ 438, 78, 38 },
		{ 438, 79, 38 },
		{ 438, 80, 38 },
		{ 438, 81, 38 },
		{ 438, 82, 38 },
		{ 438, 83, 38 },
		{ 438, 84, 38 },
		{ 438, 85, 38 },
		{ 438, 86, 38 },
		{ 438, 87, 38 },
		{ 438, 88, 38 },
		{ 438, 89, 38 },
		{ 438, 90, 38 },
		{ 438, 91, 38 },
		{ 438, 92, 38 },
		{ 438, 93, 38 },
		{ 438, 94, 38 },
		{ 438, 95, 38 },
		{ 438, 96, 38 },
		{ 438, 97, 38 },
		{ 438, 98, 38 },
		{ 439, 50, 38 },
		{ 439, 51, 38 },
		{ 439, 52, 38 },
		{ 439, 53, 38 },
		{ 439, 54, 38 },
		{ 439, 55, 38 },
		{ 439, 56, 38 },
		{ 439, 57, 38 },
		{ 439, 58, 38 },
		{ 439, 59, 38 },
		{ 439, 60, 38 },
		{ 439, 61, 38 },
		{ 439, 62, 38 },
		{ 439, 63, 38 },
		{ 439, 64, 38 },
		{ 439, 65, 38 },
		{ 439, 66, 38 },
		{ 439, 67, 38 },
		{ 439, 68, 38 },
		{ 439, 69, 38 },
		{ 439, 70, 38 },
		{ 439, 71, 38 },
		{ 439, 72, 38 },
		{ 439, 73, 38 },
		{ 439, 74, 38 },
		{ 439, 75, 38 },
		{ 439, 76, 38 },
		{ 439, 77, 38 },
		{ 439, 78, 38 },
		{ 439, 79, 38 },
		{ 439, 80, 38 },
		{ 439, 81, 38 },
		{ 439, 82, 38 },
		{ 439, 83, 38 },
		{ 439, 84, 38 },
		{ 439, 85, 38 },
		{ 439, 86, 38 },
		{ 439, 87, 38 },
		{ 439, 88, 38 },
		{ 439, 89, 38 },
		{ 439, 90, 38 },
		{ 439, 91, 38 },
		{ 439, 92, 38 },
		{ 439, 93, 38 },
		{ 439, 94, 38 },
		{ 439, 95, 38 },
		{ 439, 96, 38 },
		{ 439, 97, 38 },
		{ 440, 48, 38 },
		{ 440, 49, 38 },
		{ 440, 50, 38 },
		{ 440, 51, 38 },
		{ 440, 52, 38 },
		{ 440, 53, 38 },
		{ 440, 54, 38 },
		{ 440, 55, 38 },
		{ 440, 56, 38 },
		{ 440, 57, 38 },
		{ 440, 58, 38 },
		{ 440, 59, 38 },
		{ 440, 60, 38 },
		{ 440, 61, 38 },
		{ 440, 62, 38 },
		{ 440, 63, 38 },
		{ 440, 64, 38 },
		{ 440, 65, 38 },
		{ 440, 66, 38 },
		{ 440, 67, 38 },
		{ 440, 68, 38 },
		{ 440, 69, 38 },
		{ 440, 70, 38 },
		{ 440, 71, 38 },
		{ 440, 72, 38 },
		{ 440, 73, 38 },
		{ 440, 74, 38 },
		{ 440, 75, 38 },
		{ 440, 76, 38 },
		{ 440, 77, 38 },
		{ 440, 78, 38 },
		{ 440, 79, 38 },
		{ 440, 80, 38 },
		{ 440, 81, 38 },
		{ 440, 82, 38 },
		{ 440, 83, 38 },
		{ 440, 84, 38 },
		{ 440, 85, 38 },
		{ 440, 86, 38 },
		{ 440, 87, 38 },
		{ 440, 88, 38 },
		{ 440, 89, 38 },
		{ 440, 90, 38 },
		{ 440, 91, 38 },
		{ 440, 92, 38 },
		{ 440, 93, 38 },
		{ 440, 94, 38 },
		{ 440, 95, 38 },
		{ 440, 96, 38 },
		{ 441, 47, 38 },
		{ 441, 48, 38 },
		{ 441, 49, 38 },
		{ 441, 50, 38 },
		{ 441, 51, 38 },
		{ 441, 52, 38 },
		{ 441, 53, 38 },
		{ 441, 54, 38 },
		{ 441, 55, 38 },
		{ 441, 56, 38 },
		{ 441, 57, 38 },
		{ 441, 58, 38 },
		{ 441, 59, 38 },
		{ 441, 60, 38 },
		{ 441, 61, 38 },
		{ 441, 62, 38 },
		{ 441, 63, 38 },
		{ 441, 64, 38 },
		{ 441, 65, 38 },
		{ 441, 66, 38 },
		{ 441, 67, 38 },
		{ 441, 68, 38 },
		{ 441, 69, 38 },
		{ 441, 70, 38 },
		{ 441, 71, 38 },
		{ 441, 72, 38 },
		{ 441, 73, 38 },
		{ 441, 74, 38 },
		{ 441, 75, 38 },
		{ 441, 76, 38 },
		{ 441, 77, 38 },
		{ 441, 78, 38 },
		{ 441, 79, 38 },
		{ 441, 80, 38 },
		{ 441, 81, 38 },
		{ 441, 82, 38 },
		{ 441, 83, 38 },
		{ 441, 84, 38 },
		{ 441, 85, 38 },
		{ 441, 86, 38 },
		{ 441, 87, 38 },
		{ 441, 88, 38 },
		{ 441, 89, 38 },
		{ 441, 90, 38 },
		{ 441, 91, 38 },
		{ 441, 92, 38 },
		{ 441, 93, 38 },
		{ 441, 94, 38 },
		{ 441, 95, 38 },
		{ 442, 46, 38 },
		{ 442, 47, 38 },
		{ 442, 48, 38 },
		{ 442, 49, 38 },
		{ 442, 50, 38 },
		{ 442, 51, 38 },
		{ 442, 52, 38 },
		{ 442, 53, 38 },
		{ 442, 54, 38 },
		{ 442, 55, 38 },
		{ 442, 56, 38 },
		{ 442, 57, 38 },
		{ 442, 58, 38 },
		{ 442, 59, 38 },
		{ 442, 60, 38 },
		{ 442, 61, 38 },
		{ 442, 62, 38 },
		{ 442, 63, 38 },
		{ 442, 64, 38 },
		{ 442, 65, 38 },
		{ 442, 66, 38 },
		{ 442, 67, 38 },
		{ 442, 68, 38 },
		{ 442, 69, 38 },
		{ 442, 70, 38 },
		{ 442, 71, 38 },
		{ 442, 72, 38 },
		{ 442, 73, 38 },
		{ 442, 74, 38 },
		{ 442, 75, 38 },
		{ 442, 76, 38 },
		{ 442, 77, 38 },
		{ 442, 78, 38 },
		{ 442, 79, 38 },
		{ 442, 80, 38 },
		{ 442, 81, 38 },
		{ 442, 82, 38 },
		{ 442, 83, 38 },
		{ 442, 84, 38 },
		{ 442, 85, 38 },
		{ 442, 86, 38 },
		{ 442, 87, 38 },
		{ 442, 88, 38 },
		{ 442, 89, 38 },
		{ 442, 90, 38 },
		{ 442, 91, 38 },
		{ 442, 92, 38 },
		{ 442, 93, 38 },
		{ 442, 94, 38 },
		{ 442, 95, 38 },
		{ 443, 45, 38 },
		{ 443, 46, 38 },
		{ 443, 47, 38 },
		{ 443, 48, 38 },
		{ 443, 49, 38 },
		{ 443, 50, 38 },
		{ 443, 51, 38 },
		{ 443, 52, 38 },
		{ 443, 53, 38 },
		{ 443, 54, 38 },
		{ 443, 55, 38 },
		{ 443, 56, 38 },
		{ 443, 57, 38 },
		{ 443, 58, 38 },
		{ 443, 59, 38 },
		{ 443, 60, 38 },
		{ 443, 61, 38 },
		{ 443, 62, 38 },
		{ 443, 63, 38 },
		{ 443, 64, 38 },
		{ 443, 65, 38 },
		{ 443, 66, 38 },
		{ 443, 67, 38 },
		{ 443, 68, 38 },
		{ 443, 69, 38 },
		{ 443, 70, 38 },
		{ 443, 71, 38 },
		{ 443, 72, 38 },
		{ 443, 73, 38 },
		{ 443, 74, 38 },
		{ 443, 75, 38 },
		{ 443, 76, 38 },
		{ 443, 77, 38 },
		{ 443, 78, 38 },
		{ 443, 79, 38 },
		{ 443, 80, 38 },
		{ 443, 81, 38 },
		{ 443, 82, 38 },
		{ 443, 83, 38 },
		{ 443, 84, 38 },
		{ 443, 85, 38 },
		{ 443, 86, 38 },
		{ 443, 87, 38 },
		{ 443, 88, 38 },
		{ 443, 89, 38 },
		{ 443, 90, 38 },
		{ 443, 91, 38 },
		{ 443, 92, 38 },
		{ 443, 93, 38 },
		{ 443, 94, 38 },
		{ 444, 44, 38 },
		{ 444, 45, 38 },
		{ 444, 46, 38 },
		{ 444, 47, 38 },
		{ 444, 48, 38 },
		{ 444, 49, 38 },
		{ 444, 50, 38 },
		{ 444, 51, 38 },
		{ 444, 52, 38 },
		{ 444, 53, 38 },
		{ 444, 54, 38 },
		{ 444, 55, 38 },
		{ 444, 56, 38 },
		{ 444, 57, 38 },
		{ 444, 58, 38 },
		{ 444, 59, 38 },
		{ 444, 60, 38 },
		{ 444, 61, 38 },
		{ 444, 62, 38 },
		{ 444, 63, 38 },
		{ 444, 64, 38 },
		{ 444, 65, 38 },
		{ 444, 66, 38 },
		{ 444, 67, 38 },
		{ 444, 68, 38 },
		{ 444, 69, 38 },
		{ 444, 70, 38 },
		{ 444, 71, 38 },
		{ 444, 72, 38 },
		{ 444, 73, 38 },
		{ 444, 74, 38 },
		{ 444, 75, 38 },
		{ 444, 76, 38 },
		{ 444, 77, 38 },
		{ 444, 78, 38 },
		{ 444, 79, 38 },
		{ 444, 80, 38 },
		{ 444, 81, 38 },
		{ 444, 82, 38 },
		{ 444, 83, 38 },
		{ 444, 84, 38 },
		{ 444, 85, 38 },
		{ 444, 86, 38 },
		{ 444, 87, 38 },
		{ 444, 88, 38 },
		{ 444, 89, 38 },
		{ 444, 90, 38 },
		{ 444, 91, 38 },
		{ 444, 92, 38 },
		{ 444, 93, 38 },
		{ 445, 43, 38 },
		{ 445, 44, 38 },
		{ 445, 45, 38 },
		{ 445, 46, 38 },
		{ 445, 47, 38 },
		{ 445, 48, 38 },
		{ 445, 49, 38 },
		{ 445, 50, 38 },
		{ 445, 51, 38 },
		{ 445, 52, 38 },
		{ 445, 53, 38 },
		{ 445, 54, 38 },
		{ 445, 55, 38 },
		{ 445, 56, 38 },
		{ 445, 57, 38 },
		{ 445, 58, 38 },
		{ 445, 59, 38 },
		{ 445, 60, 38 },
		{ 445, 61, 38 },
		{ 445, 62, 38 },
		{ 445, 63, 38 },
		{ 445, 64, 38 },
		{ 445, 65, 38 },
		{ 445, 66, 38 },
		{ 445, 67, 38 },
		{ 445, 68, 38 },
		{ 445, 69, 38 },
		{ 445, 70, 38 },
		{ 445, 71, 38 },
		{ 445, 72, 38 },
		{ 445, 73, 38 },
		{ 445, 74, 38 },
		{ 445, 75, 38 },
		{ 445, 76, 38 },
		{ 445, 77, 38 },
		{ 445, 78, 38 },
		{ 445, 79, 38 },
		{ 445, 80, 38 },
		{ 445, 81, 38 },
		{ 445, 82, 38 },
		{ 445, 83, 38 },
		{ 445, 84, 38 },
		{ 445, 85, 38 },
		{ 445, 86, 38 },
		{ 445, 87, 38 },
		{ 445, 88, 38 },
		{ 445, 89, 38 },
		{ 445, 90, 38 },
		{ 445, 91, 38 },
		{ 445, 92, 38 },
		{ 446, 42, 38 },
		{ 446, 43, 38 },
		{ 446, 44, 38 },
		{ 446, 45, 38 },
		{ 446, 46, 38 },
		{ 446, 47, 38 },
		{ 446, 48, 38 },
		{ 446, 49, 38 },
		{ 446, 50, 38 },
		{ 446, 51, 38 },
		{ 446, 52, 38 },
		{ 446, 53, 38 },
		{ 446, 54, 38 },
		{ 446, 55, 38 },
		{ 446, 56, 38 },
		{ 446, 57, 38 },
		{ 446, 58, 38 },
		{ 446, 59, 38 },
		{ 446, 60, 38 },
		{ 446, 61, 38 },
		{ 446, 62, 38 },
		{ 446, 63, 38 },
		{ 446, 64, 38 },
		{ 446, 65, 38 },
		{ 446, 66, 38 },
		{ 446, 67, 38 },
		{ 446, 68, 38 },
		{ 446, 69, 38 },
		{ 446, 70, 38 },
		{ 446, 71, 38 },
		{ 446, 72, 38 },
		{ 446, 73, 38 },
		{ 446, 74, 38 },
		{ 446, 75, 38 },
		{ 446, 76, 38 },
		{ 446, 77, 38 },
		{ 446, 78, 38 },
		{ 446, 79, 38 },
		{ 446, 80, 38 },
		{ 446, 81, 38 },
		{ 446, 82, 38 },
		{ 446, 83, 38 },
		{ 446, 84, 38 },
		{ 446, 85, 38 },
		{ 446, 86, 38 },
		{ 446, 87, 38 },
		{ 446, 88, 38 },
		{ 446, 89, 38 },
		{ 446, 90, 38 },
		{ 446, 91, 38 },
		{ 447, 40, 38 },
		{ 447, 41, 38 },
		{ 447, 42, 38 },
		{ 447, 43, 38 },
		{ 447, 44, 38 },
		{ 447, 45, 38 },
		{ 447, 46, 38 },
		{ 447, 47, 38 },
		{ 447, 48, 38 },
		{ 447, 49, 38 },
		{ 447, 50, 38 },
		{ 447, 51, 38 },
		{ 447, 52, 38 },
		{ 447, 53, 38 },
		{ 447, 54, 38 },
		{ 447, 55, 38 },
		{ 447, 56, 38 },
		{ 447, 57, 38 },
		{ 447, 58, 38 },
		{ 447, 59, 38 },
		{ 447, 60, 38 },
		{ 447, 61, 38 },
		{ 447, 62, 38 },
		{ 447, 63, 38 },
		{ 447, 64, 38 },
		{ 447, 65, 38 },
		{ 447, 66, 38 },
		{ 447, 67, 38 },
		{ 447, 68, 38 },
		{ 447, 69, 38 },
		{ 447, 70, 38 },
		{ 447, 71, 38 },
		{ 447, 72, 38 },
		{ 447, 73, 38 },
		{ 447, 74, 38 },
		{ 447, 75, 38 },
		{ 447, 76, 38 },
		{ 447, 77, 38 },
		{ 447, 78, 38 },
		{ 447, 79, 38 },
		{ 447, 80, 38 },
		{ 447, 81, 38 },
		{ 447, 82, 38 },
		{ 447, 83, 38 },
		{ 447, 84, 38 },
		{ 447, 85, 38 },
		{ 447, 86, 38 },
		{ 447, 87, 38 },
		{ 447, 88, 38 },
		{ 447, 89, 38 },
		{ 447, 90, 38 },
		{ 448, 39, 38 },
		{ 448, 40, 38 },
		{ 448, 41, 38 },
		{ 448, 42, 38 },
		{ 448, 43, 38 },
		{ 448, 44, 38 },
		{ 448, 45, 38 },
		{ 448, 46, 38 },
		{ 448, 47, 38 },
		{ 448, 48, 38 },
		{ 448, 49, 38 },
		{ 448, 50, 38 },
		{ 448, 51, 38 },
		{ 448, 52, 38 },
		{ 448, 53, 38 },
		{ 448, 54, 38 },
		{ 448, 55, 38 },
		{ 448, 56, 38 },
		{ 448, 57, 38 },
		{ 448, 58, 38 },
		{ 448, 59, 38 },
		{ 448, 60, 38 },
		{ 448, 61, 38 },
		{ 448, 62, 38 },
		{ 448, 63, 38 },
		{ 448, 64, 38 },
		{ 448, 65, 38 },
		{ 448, 66, 38 },
		{ 448, 67, 38 },
		{ 448, 68, 38 },
		{ 448, 69, 38 },
		{ 448, 70, 38 },
		{ 448, 71, 38 },
		{ 448, 72, 38 },
		{ 448, 73, 38 },
		{ 448, 74, 38 },
		{ 448, 75, 38 },
		{ 448, 76, 38 },
		{ 448, 77, 38 },
		{ 448, 78, 38 },
		{ 448, 79, 38 },
		{ 448, 80, 38 },
		{ 448, 81, 38 },
		{ 448, 82, 38 },
		{ 448, 83, 38 },
		{ 448, 84, 38 },
		{ 448, 85, 38 },
		{ 448, 86, 38 },
		{ 448, 87, 38 },
		{ 448, 88, 38 },
		{ 448, 89, 38 },
		{ 449, 38, 38 },
		{ 449, 39, 38 },
		{ 449, 40, 38 },
		{ 449, 41, 38 },
		{ 449, 42, 38 },
		{ 449, 43, 38 },
		{ 449, 44, 38 },
		{ 449, 45, 38 },
		{ 449, 46, 38 },
		{ 449, 47, 38 },
		{ 449, 48, 38 },
		{ 449, 49, 38 },
		{ 449, 50, 38 },
		{ 449, 51, 38 },
		{ 449, 52, 38 },
		{ 449, 53, 38 },
		{ 449, 54, 38 },
		{ 449, 55, 38 },
		{ 449, 56, 38 },
		{ 449, 57, 38 },
		{ 449, 58, 38 },
		{ 449, 59, 38 },
		{ 449, 60, 38 },
		{ 449, 61, 38 },
		{ 449, 62, 38 },
		{ 449, 63, 38 },
		{ 449, 64, 38 },
		{ 449, 65, 38 },
		{ 449, 66, 38 },
		{ 449, 67, 38 },
		{ 449, 68, 38 },
		{ 449, 69, 38 },
		{ 449, 70, 38 },
		{ 449, 71, 38 },
		{ 449, 72, 38 },
		{ 449, 73, 38 },
		{ 449, 74, 38 },
		{ 449, 75, 38 },
		{ 449, 76, 38 },
		{ 449, 77, 38 },
		{ 449, 78, 38 },
		{ 449, 79, 38 },
		{ 449, 80, 38 },
		{ 449, 81, 38 },
		{ 449, 82, 38 },
		{ 449, 83, 38 },
		{ 449, 84, 38 },
		{ 449, 85, 38 },
		{ 449, 86, 38 },
		{ 449, 87, 38 },
		{ 449, 88, 38 },
		{ 450, 37, 38 },
		{ 450, 38, 38 },
		{ 450, 39, 38 },
		{ 450, 40, 38 },
		{ 450, 41, 38 },
		{ 450, 42, 38 },
		{ 450, 43, 38 },
		{ 450, 44, 38 },
		{ 450, 45, 38 },
		{ 450, 46, 38 },
		{ 450, 47, 38 },
		{ 450, 48, 38 },
		{ 450, 49, 38 },
		{ 450, 50, 38 },
		{ 450, 51, 38 },
		{ 450, 52, 38 },
		{ 450, 53, 38 },
		{ 450, 54, 38 },
		{ 450, 55, 38 },
		{ 450, 56, 38 },
		{ 450, 57, 38 },
		{ 450, 58, 38 },
		{ 450, 59, 38 },
		{ 450, 60, 38 },
		{ 450, 61, 38 },
		{ 450, 62, 38 },
		{ 450, 63, 38 },
		{ 450, 64, 38 },
		{ 450, 65, 38 },
		{ 450, 66, 38 },
		{ 450, 67, 38 },
		{ 450, 68, 38 },
		{ 450, 69, 38 },
		{ 450, 70, 38 },
		{ 450, 71, 38 },
		{ 450, 72, 38 },
		{ 450, 73, 38 },
		{ 450, 74, 38 },
		{ 450, 75, 38 },
		{ 450, 76, 38 },
		{ 450, 77, 38 },
		{ 450, 78, 38 },
		{ 450, 79, 38 },
		{ 450, 80, 38 },
		{ 450, 81, 38 },
		{ 450, 82, 38 },
		{ 450, 83, 38 },
		{ 450, 84, 38 },
		{ 450, 85, 38 },
		{ 450, 86, 38 },
		{ 450, 87, 38 },
		{ 450, 88, 38 },
		{ 451, 36, 38 },
		{ 451, 37, 38 },
		{ 451, 38, 38 },
		{ 451, 39, 38 },
		{ 451, 40, 38 },
		{ 451, 41, 38 },
		{ 451, 42, 38 },
		{ 451, 43, 38 },
		{ 451, 44, 38 },
		{ 451, 45, 38 },
		{ 451, 46, 38 },
		{ 451, 47, 38 },
		{ 451, 48, 38 },
		{ 451, 49, 38 },
		{ 451, 50, 38 },
		{ 451, 51, 38 },
		{ 451, 52, 38 },
		{ 451, 53, 38 },
		{ 451, 54, 38 },
		{ 451, 55, 38 },
		{ 451, 56, 38 },
		{ 451, 57, 38 },
		{ 451, 58, 38 },
		{ 451, 59, 38 },
		{ 451, 60, 38 },
		{ 451, 61, 38 },
		{ 451, 62, 38 },
		{ 451, 63, 38 },
		{ 451, 64, 38 },
		{ 451, 65, 38 },
		{ 451, 66, 38 },
		{ 451, 67, 38 },
		{ 451, 68, 38 },
		{ 451, 69, 38 },
		{ 451, 70, 38 },
		{ 451, 71, 38 },
		{ 451, 72, 38 },
		{ 451, 73, 38 },
		{ 451, 74, 38 },
		{ 451, 75, 38 },
		{ 451, 76, 38 },
		{ 451, 77, 38 },
		{ 451, 78, 38 },
		{ 451, 79, 38 },
		{ 451, 80, 38 },
		{ 451, 81, 38 },
		{ 451, 82, 38 },
		{ 451, 83, 38 },
		{ 451, 84, 38 },
		{ 451, 85, 38 },
		{ 451, 86, 38 },
		{ 451, 87, 38 },
		{ 452, 35, 38 },
		{ 452, 36, 38 },
		{ 452, 37, 38 },
		{ 452, 38, 38 },
		{ 452, 39, 38 },
		{ 452, 40, 38 },
		{ 452, 41, 38 },
		{ 452, 42, 38 },
		{ 452, 43, 38 },
		{ 452, 44, 38 },
		{ 452, 45, 38 },
		{ 452, 46, 38 },
		{ 452, 47, 38 },
		{ 452, 48, 38 },
		{ 452, 49, 38 },
		{ 452, 50, 38 },
		{ 452, 51, 38 },
		{ 452, 52, 38 },
		{ 452, 53, 38 },
		{ 452, 54, 38 },
		{ 452, 55, 38 },
		{ 452, 56, 38 },
		{ 452, 57, 38 },
		{ 452, 58, 38 },
		{ 452, 59, 38 },
		{ 452, 60, 38 },
		{ 452, 61, 38 },
		{ 452, 62, 38 },
		{ 452, 63, 38 },
		{ 452, 64, 38 },
		{ 452, 65, 38 },
		{ 452, 66, 38 },
		{ 452, 67, 38 },
		{ 452, 68, 38 },
		{ 452, 69, 38 },
		{ 452, 70, 38 },
		{ 452, 71, 38 },
		{ 452, 72, 38 },
		{ 452, 73, 38 },
		{ 452, 74, 38 },
		{ 452, 75, 38 },
		{ 452, 76, 38 },
		{ 452, 77, 38 },
		{ 452, 78, 38 },
		{ 452, 79, 38 },
		{ 452, 80, 38 },
		{ 452, 81, 38 },
		{ 452, 82, 38 },
		{ 452, 83, 38 },
		{ 452, 84, 38 },
		{ 452, 85, 38 },
		{ 452, 86, 38 },
		{ 453, 34, 38 },
		{ 453, 35, 38 },
		{ 453, 36, 38 },
		{ 453, 37, 38 },
		{ 453, 38, 38 },
		{ 453, 39, 38 },
		{ 453, 40, 38 },
		{ 453, 41, 38 },
		{ 453, 42, 38 },
		{ 453, 43, 38 },
		{ 453, 44, 38 },
		{ 453, 45, 38 },
		{ 453, 46, 38 },
		{ 453, 47, 38 },
		{ 453, 48, 38 },
		{ 453, 49, 38 },
		{ 453, 50, 38 },
		{ 453, 51, 38 },
		{ 453, 52, 38 },
		{ 453, 53, 38 },
		{ 453, 54, 38 },
		{ 453, 55, 38 },
		{ 453, 56, 38 },
		{ 453, 57, 38 },
		{ 453, 58, 38 },
		{ 453, 59, 38 },
		{ 453, 60, 38 },
		{ 453, 61, 38 },
		{ 453, 62, 38 },
		{ 453, 63, 38 },
		{ 453, 64, 38 },
		{ 453, 65, 38 },
		{ 453, 66, 38 },
		{ 453, 67, 38 },
		{ 453, 68, 38 },
		{ 453, 69, 38 },
		{ 453, 70, 38 },
		{ 453, 71, 38 },
		{ 453, 72, 38 },
		{ 453, 73, 38 },
		{ 453, 74, 38 },
		{ 453, 75, 38 },
		{ 453, 76, 38 },
		{ 453, 77, 38 },
		{ 453, 78, 38 },
		{ 453, 79, 38 },
		{ 453, 80, 38 },
		{ 453, 81, 38 },
		{ 453, 82, 38 },
		{ 453, 83, 38 },
		{ 453, 84, 38 },
		{ 453, 85, 38 },
		{ 454, 32, 38 },
		{ 454, 33, 38 },
		{ 454, 34, 38 },
		{ 454, 35, 38 },
		{ 454, 36, 38 },
		{ 454, 37, 38 },
		{ 454, 38, 38 },
		{ 454, 39, 38 },
		{ 454, 40, 38 },
		{ 454, 41, 38 },
		{ 454, 42, 38 },
		{ 454, 43, 38 },
		{ 454, 44, 38 },
		{ 454, 45, 38 },
		{ 454, 46, 38 },
		{ 454, 47, 38 },
		{ 454, 48, 38 },
		{ 454, 49, 38 },
		{ 454, 50, 38 },
		{ 454, 51, 38 },
		{ 454, 52, 38 },
		{ 454, 53, 38 },
		{ 454, 54, 38 },
		{ 454, 55, 38 },
		{ 454, 56, 38 },
		{ 454, 57, 38 },
		{ 454, 58, 38 },
		{ 454, 59, 38 },
		{ 454, 60, 38 },
		{ 454, 61, 38 },
		{ 454, 62, 38 },
		{ 454, 63, 38 },
		{ 454, 64, 38 },
		{ 454, 65, 38 },
		{ 454, 66, 38 },
		{ 454, 67, 38 },
		{ 454, 68, 38 },
		{ 454, 69, 38 },
		{ 454, 70, 38 },
		{ 454, 71, 38 },
		{ 454, 72, 38 },
		{ 454, 73, 38 },
		{ 454, 74, 38 },
		{ 454, 75, 38 },
		{ 454, 76, 38 },
		{ 454, 77, 38 },
		{ 454, 78, 38 },
		{ 454, 79, 38 },
		{ 454, 80, 38 },
		{ 454, 81, 38 },
		{ 454, 82, 38 },
		{ 454, 83, 38 },
		{ 454, 84, 38 },
		{ 455, 31, 38 },
		{ 455, 32, 38 },
		{ 455, 33, 38 },
		{ 455, 34, 38 },
		{ 455, 35, 38 },
		{ 455, 36, 38 },
		{ 455, 37, 38 },
		{ 455, 38, 38 },
		{ 455, 39, 38 },
		{ 455, 40, 38 },
		{ 455, 41, 38 },
		{ 455, 42, 38 },
		{ 455, 43, 38 },
		{ 455, 44, 38 },
		{ 455, 45, 38 },
		{ 455, 46, 38 },
		{ 455, 47, 38 },
		{ 455, 48, 38 },
		{ 455, 49, 38 },
		{ 455, 50, 38 },
		{ 455, 51, 38 },
		{ 455, 52, 38 },
		{ 455, 53, 38 },
		{ 455, 54, 38 },
		{ 455, 55, 38 },
		{ 455, 56, 38 },
		{ 455, 57, 38 },
		{ 455, 58, 38 },
		{ 455, 59, 38 },
		{ 455, 60, 38 },
		{ 455, 61, 38 },
		{ 455, 62, 38 },
		{ 455, 63, 38 },
		{ 455, 64, 38 },
		{ 455, 65, 38 },
		{ 455, 66, 38 },
		{ 455, 67, 38 },
		{ 455, 68, 38 },
		{ 455, 69, 38 },
		{ 455, 70, 38 },
		{ 455, 71, 38 },
		{ 455, 72, 38 },
		{ 455, 73, 38 },
		{ 455, 74, 38 },
		{ 455, 75, 38 },
		{ 455, 76, 38 },
		{ 455, 77, 38 },
		{ 455, 78, 38 },
		{ 455, 79, 38 },
		{ 455, 80, 38 },
		{ 455, 81, 38 },
		{ 455, 82, 38 },
		{ 455, 83, 38 },
		{ 456, 30, 38 },
		{ 456, 31, 38 },
		{ 456, 32, 38 },
		{ 456, 33, 38 },
		{ 456, 34, 38 },
		{ 456, 35, 38 },
		{ 456, 36, 38 },
		{ 456, 37, 38 },
		{ 456, 38, 38 },
		{ 456, 39, 38 },
		{ 456, 40, 38 },
		{ 456, 41, 38 },
		{ 456, 42, 38 },
		{ 456, 43, 38 },
		{ 456, 44, 38 },
		{ 456, 45, 38 },
		{ 456, 46, 38 },
		{ 456, 47, 38 },
		{ 456, 48, 38 },
		{ 456, 49, 38 },
		{ 456, 50, 38 },
		{ 456, 51, 38 },
		{ 456, 52, 38 },
		{ 456, 53, 38 },
		{ 456, 54, 38 },
		{ 456, 55, 38 },
		{ 456, 56, 38 },
		{ 456, 57, 38 },
		{ 456, 58, 38 },
		{ 456, 59, 38 },
		{ 456, 60, 38 },
		{ 456, 61, 38 },
		{ 456, 62, 38 },
		{ 456, 63, 38 },
		{ 456, 64, 38 },
		{ 456, 65, 38 },
		{ 456, 66, 38 },
		{ 456, 67, 38 },
		{ 456, 68, 38 },
		{ 456, 69, 38 },
		{ 456, 70, 38 },
		{ 456, 71, 38 },
		{ 456, 72, 38 },
		{ 456, 73, 38 },
		{ 456, 74, 38 },
		{ 456, 75, 38 },
		{ 456, 76, 38 },
		{ 456, 77, 38 },
		{ 456, 78, 38 },
		{ 456, 79, 38 },
		{ 456, 80, 38 },
		{ 456, 81, 38 },
		{ 456, 82, 38 },
		{ 457, 29, 38 },
		{ 457, 30, 38 },
		{ 457, 31, 38 },
		{ 457, 32, 38 },
		{ 457, 33, 38 },
		{ 457, 34, 38 },
		{ 457, 35, 38 },
		{ 457, 36, 38 },
		{ 457, 37, 38 },
		{ 457, 38, 38 },
		{ 457, 39, 38 },
		{ 457, 40, 38 },
		{ 457, 41, 38 },
		{ 457, 42, 38 },
		{ 457, 43, 38 },
		{ 457, 44, 38 },
		{ 457, 45, 38 },
		{ 457, 46, 38 },
		{ 457, 47, 38 },
		{ 457, 48, 38 },
		{ 457, 49, 38 },
		{ 457, 50, 38 },
		{ 457, 51, 38 },
		{ 457, 52, 38 },
		{ 457, 53, 38 },
		{ 457, 54, 38 },
		{ 457, 55, 38 },
		{ 457, 56, 38 },
		{ 457, 57, 38 },
		{ 457, 58, 38 },
		{ 457, 59, 38 },
		{ 457, 60, 38 },
		{ 457, 61, 38 },
		{ 457, 62, 38 },
		{ 457, 63, 38 },
		{ 457, 64, 38 },
		{ 457, 65, 38 },
		{ 457, 66, 38 },
		{ 457, 67, 38 },
		{ 457, 68, 38 },
		{ 457, 69, 38 },
		{ 457, 70, 38 },
		{ 457, 71, 38 },
		{ 457, 72, 38 },
		{ 457, 73, 38 },
		{ 457, 74, 38 },
		{ 457, 75, 38 },
		{ 457, 76, 38 },
		{ 457, 77, 38 },
		{ 457, 78, 38 },
		{ 457, 79, 38 },
		{ 457, 80, 38 },
		{ 457, 81, 38 },
		{ 458, 28, 38 },
		{ 458, 29, 38 },
		{ 458, 30, 38 },
		{ 458, 31, 38 },
		{ 458, 32, 38 },
		{ 458, 33, 38 },
		{ 458, 34, 38 },
		{ 458, 35, 38 },
		{ 458, 36, 38 },
		{ 458, 37, 38 },
		{ 458, 38, 38 },
		{ 458, 39, 38 },
		{ 458, 40, 38 },
		{ 458, 41, 38 },
		{ 458, 42, 38 },
		{ 458, 43, 38 },
		{ 458, 44, 38 },
		{ 458, 45, 38 },
		{ 458, 46, 38 },
		{ 458, 47, 38 },
		{ 458, 48, 38 },
		{ 458, 49, 38 },
		{ 458, 50, 38 },
		{ 458, 51, 38 },
		{ 458, 52, 38 },
		{ 458, 53, 38 },
		{ 458, 54, 38 },
		{ 458, 55, 38 },
		{ 458, 56, 38 },
		{ 458, 57, 38 },
		{ 458, 58, 38 },
		{ 458, 59, 38 },
		{ 458, 60, 38 },
		{ 458, 61, 38 },
		{ 458, 62, 38 },
		{ 458, 63, 38 },
		{ 458, 64, 38 },
		{ 458, 65, 38 },
		{ 458, 66, 38 },
		{ 458, 67, 38 },
		{ 458, 68, 38 },
		{ 458, 69, 38 },
		{ 458, 70, 38 },
		{ 458, 71, 38 },
		{ 458, 72, 38 },
		{ 458, 73, 38 },
		{ 458, 74, 38 },
		{ 458, 75, 38 },
		{ 458, 76, 38 },
		{ 458, 77, 38 },
		{ 458, 78, 38 },
		{ 458, 79, 38 },
		{ 458, 80, 38 },
		{ 458, 81, 38 },
		{ 459, 27, 38 },
		{ 459, 28, 38 },
		{ 459, 29, 38 },
		{ 459, 30, 38 },
		{ 459, 31, 38 },
		{ 459, 32, 38 },
		{ 459, 33, 38 },
		{ 459, 34, 38 },
		{ 459, 35, 38 },
		{ 459, 36, 38 },
		{ 459, 37, 38 },
		{ 459, 38, 38 },
		{ 459, 39, 38 },
		{ 459, 40, 38 },
		{ 459, 41, 38 },
		{ 459, 42, 38 },
		{ 459, 43, 38 },
		{ 459, 44, 38 },
		{ 459, 45, 38 },
		{ 459, 46, 38 },
		{ 459, 47, 38 },
		{ 459, 48, 38 },
		{ 459, 49, 38 },
		{ 459, 50, 38 },
		{ 459, 51, 38 },
		{ 459, 52, 38 },
		{ 459, 53, 38 },
		{ 459, 54, 38 },
		{ 459, 55, 38 },
		{ 459, 56, 38 },
		{ 459, 57, 38 },
		{ 459, 58, 38 },
		{ 459, 59, 38 },
		{ 459, 60, 38 },
		{ 459, 61, 38 },
		{ 459, 62, 38 },
		{ 459, 63, 38 },
		{ 459, 64, 38 },
		{ 459, 65, 38 },
		{ 459, 66, 38 },
		{ 459, 67, 38 },
		{ 459, 68, 38 },
		{ 459, 69, 38 },
		{ 459, 70, 38 },
		{ 459, 71, 38 },
		{ 459, 72, 38 },
		{ 459, 73, 38 },
		{ 459, 74, 38 },
		{ 459, 75, 38 },
		{ 459, 76, 38 },
		{ 459, 77, 38 },
		{ 459, 78, 38 },
		{ 459, 79, 38 },
		{ 459, 80, 38 },
		{ 460, 26, 38 },
		{ 460, 27, 38 },
		{ 460, 28, 38 },
		{ 460, 29, 38 },
		{ 460, 30, 38 },
		{ 460, 31, 38 },
		{ 460, 32, 38 },
		{ 460, 33, 38 },
		{ 460, 34, 38 },
		{ 460, 35, 38 },
		{ 460, 36, 38 },
		{ 460, 37, 38 },
		{ 460, 38, 38 },
		{ 460, 39, 38 },
		{ 460, 40, 38 },
		{ 460, 41, 38 },
		{ 460, 42, 38 },
		{ 460, 43, 38 },
		{ 460, 44, 38 },
		{ 460, 45, 38 },
		{ 460, 46, 38 },
		{ 460, 47, 38 },
		{ 460, 48, 38 },
		{ 460, 49, 38 },
		{ 460, 50, 38 },
		{ 460, 51, 38 },
		{ 460, 52, 38 },
		{ 460, 53, 38 },
		{ 460, 54, 38 },
		{ 460, 55, 38 },
		{ 460, 56, 38 },
		{ 460, 57, 38 },
		{ 460, 58, 38 },
		{ 460, 59, 38 },
		{ 460, 60, 38 },
		{ 460, 61, 38 },
		{ 460, 62, 38 },
		{ 460, 63, 38 },
		{ 460, 64, 38 },
		{ 460, 65, 38 },
		{ 460, 66, 38 },
		{ 460, 67, 38 },
		{ 460, 68, 38 },
		{ 460, 69, 38 },
		{ 460, 70, 38 },
		{ 460, 71, 38 },
		{ 460, 72, 38 },
		{ 460, 73, 38 },
		{ 460, 74, 38 },
		{ 460, 75, 38 },
		{ 460, 76, 38 },
		{ 460, 77, 38 },
		{ 460, 78, 38 },
		{ 460, 79, 38 },
		{ 461, 24, 38 },
		{ 461, 25, 38 },
		{ 461, 26, 38 },
		{ 461, 27, 38 },
		{ 461, 28, 38 },
		{ 461, 29, 38 },
		{ 461, 30, 38 },
		{ 461, 31, 38 },
		{ 461, 32, 38 },
		{ 461, 33, 38 },
		{ 461, 34, 38 },
		{ 461, 35, 38 },
		{ 461, 36, 38 },
		{ 461, 37, 38 },
		{ 461, 38, 38 },
		{ 461, 39, 38 },
		{ 461, 40, 38 },
		{ 461, 41, 38 },
		{ 461, 42, 38 },
		{ 461, 43, 38 },
		{ 461, 44, 38 },
		{ 461, 45, 38 },
		{ 461, 46, 38 },
		{ 461, 47, 38 },
		{ 461, 48, 38 },
		{ 461, 49, 38 },
		{ 461, 50, 38 },
		{ 461, 51, 38 },
		{ 461, 52, 38 },
		{ 461, 53, 38 },
		{ 461, 54, 38 },
		{ 461, 55, 38 },
		{ 461, 56, 38 },
		{ 461, 57, 38 },
		{ 461, 58, 38 },
		{ 461, 59, 38 },
		{ 461, 60, 38 },
		{ 461, 61, 38 },
		{ 461, 62, 38 },
		{ 461, 63, 38 },
		{ 461, 64, 38 },
		{ 461, 65, 38 },
		{ 461, 66, 38 },
		{ 461, 67, 38 },
		{ 461, 68, 38 },
		{ 461, 69, 38 },
		{ 461, 70, 38 },
		{ 461, 71, 38 },
		{ 461, 72, 38 },
		{ 461, 73, 38 },
		{ 461, 74, 38 },
		{ 461, 75, 38 },
		{ 461, 76, 38 },
		{ 461, 77, 38 },
		{ 461, 78, 38 },
		{ 462, 23, 38 },
		{ 462, 24, 38 },
		{ 462, 25, 38 },
		{ 462, 26, 38 },
		{ 462, 27, 38 },
		{ 462, 28, 38 },
		{ 462, 29, 38 },
		{ 462, 30, 38 },
		{ 462, 31, 38 },
		{ 462, 32, 38 },
		{ 462, 33, 38 },
		{ 462, 34, 38 },
		{ 462, 35, 38 },
		{ 462, 36, 38 },
		{ 462, 37, 38 },
		{ 462, 38, 38 },
		{ 462, 39, 38 },
		{ 462, 40, 38 },
		{ 462, 41, 38 },
		{ 462, 42, 38 },
		{ 462, 43, 38 },
		{ 462, 44, 38 },
		{ 462, 45, 38 },
		{ 462, 46, 38 },
		{ 462, 47, 38 },
		{ 462, 48, 38 },
		{ 462, 49, 38 },
		{ 462, 50, 38 },
		{ 462, 51, 38 },
		{ 462, 52, 38 },
		{ 462, 53, 38 },
		{ 462, 54, 38 },
		{ 462, 55, 38 },
		{ 462, 56, 38 },
		{ 462, 57, 38 },
		{ 462, 58, 38 },
		{ 462, 59, 38 },
		{ 462, 60, 38 },
		{ 462, 61, 38 },
		{ 462, 62, 38 },
		{ 462, 63, 38 },
		{ 462, 64, 38 },
		{ 462, 65, 38 },
		{ 462, 66, 38 },
		{ 462, 67, 38 },
		{ 462, 68, 38 },
		{ 462, 69, 38 },
		{ 462, 70, 38 },
		{ 462, 71, 38 },
		{ 462, 72, 38 },
		{ 462, 73, 38 },
		{ 462, 74, 38 },
		{ 462, 75, 38 },
		{ 462, 76, 38 },
		{ 462, 77, 38 },
		{ 463, 22, 38 },
		{ 463, 23, 38 },
		{ 463, 24, 38 },
		{ 463, 25, 38 },
		{ 463, 26, 38 },
		{ 463, 27, 38 },
		{ 463, 28, 38 },
		{ 463, 29, 38 },
		{ 463, 30, 38 },
		{ 463, 31, 38 },
		{ 463, 32, 38 },
		{ 463, 33, 38 },
		{ 463, 34, 38 },
		{ 463, 35, 38 },
		{ 463, 36, 38 },
		{ 463, 37, 38 },
		{ 463, 38, 38 },
		{ 463, 39, 38 },
		{ 463, 40, 38 },
		{ 463, 41, 38 },
		{ 463, 42, 38 },
		{ 463, 43, 38 },
		{ 463, 44, 38 },
		{ 463, 45, 38 },
		{ 463, 46, 38 },
		{ 463, 47, 38 },
		{ 463, 48, 38 },
		{ 463, 49, 38 },
		{ 463, 50, 38 },
		{ 463, 51, 38 },
		{ 463, 52, 38 },
		{ 463, 53, 38 },
		{ 463, 54, 38 },
		{ 463, 55, 38 },
		{ 463, 56, 38 },
		{ 463, 57, 38 },
		{ 463, 58, 38 },
		{ 463, 59, 38 },
		{ 463, 60, 38 },
		{ 463, 61, 38 },
		{ 463, 62, 38 },
		{ 463, 63, 38 },
		{ 463, 64, 38 },
		{ 463, 65, 38 },
		{ 463, 66, 38 },
		{ 463, 67, 38 },
		{ 463, 68, 38 },
		{ 463, 69, 38 },
		{ 463, 70, 38 },
		{ 463, 71, 38 },
		{ 463, 72, 38 },
		{ 463, 73, 38 },
		{ 463, 74, 38 },
		{ 463, 75, 38 },
		{ 463, 76, 38 },
		{ 464, 21, 38 },
		{ 464, 22, 38 },
		{ 464, 23, 38 },
		{ 464, 24, 38 },
		{ 464, 25, 38 },
		{ 464, 26, 38 },
		{ 464, 27, 38 },
		{ 464, 28, 38 },
		{ 464, 29, 38 },
		{ 464, 30, 38 },
		{ 464, 31, 38 },
		{ 464, 32, 38 },
		{ 464, 33, 38 },
		{ 464, 34, 38 },
		{ 464, 35, 38 },
		{ 464, 36, 38 },
		{ 464, 37, 38 },
		{ 464, 38, 38 },
		{ 464, 39, 38 },
		{ 464, 40, 38 },
		{ 464, 41, 38 },
		{ 464, 42, 38 },
		{ 464, 43, 38 },
		{ 464, 44, 38 },
		{ 464, 45, 38 },
		{ 464, 46, 38 },
		{ 464, 47, 38 },
		{ 464, 48, 38 },
		{ 464, 49, 38 },
		{ 464, 50, 38 },
		{ 464, 51, 38 },
		{ 464, 52, 38 },
		{ 464, 53, 38 },
		{ 464, 54, 38 },
		{ 464, 55, 38 },
		{ 464, 56, 38 },
		{ 464, 57, 38 },
		{ 464, 58, 38 },
		{ 464, 59, 38 },
		{ 464, 60, 38 },
		{ 464, 61, 38 },
		{ 464, 62, 38 },
		{ 464, 63, 38 },
		{ 464, 64, 38 },
		{ 464, 65, 38 },
		{ 464, 66, 38 },
		{ 464, 67, 38 },
		{ 464, 68, 38 },
		{ 464, 69, 38 },
		{ 464, 70, 38 },
		{ 464, 71, 38 },
		{ 464, 72, 38 },
		{ 464, 73, 38 },
		{ 464, 74, 38 },
		{ 464, 75, 38 },
		{ 465, 20, 38 },
		{ 465, 21, 38 },
		{ 465, 22, 38 },
		{ 465, 23, 38 },
		{ 465, 24, 38 },
		{ 465, 25, 38 },
		{ 465, 26, 38 },
		{ 465, 27, 38 },
		{ 465, 28, 38 },
		{ 465, 29, 38 },
		{ 465, 30, 38 },
		{ 465, 31, 38 },
		{ 465, 32, 38 },
		{ 465, 33, 38 },
		{ 465, 34, 38 },
		{ 465, 35, 38 },
		{ 465, 36, 38 },
		{ 465, 37, 38 },
		{ 465, 38, 38 },
		{ 465, 39, 38 },
		{ 465, 40, 38 },
		{ 465, 41, 38 },
		{ 465, 42, 38 },
		{ 465, 43, 38 },
		{ 465, 44, 38 },
		{ 465, 45, 38 },
		{ 465, 46, 38 },
		{ 465, 47, 38 },
		{ 465, 48, 38 },
		{ 465, 49, 38 },
		{ 465, 50, 38 },
		{ 465, 51, 38 },
		{ 465, 52, 38 },
		{ 465, 53, 38 },
		{ 465, 54, 38 },
		{ 465, 55, 38 },
		{ 465, 56, 38 },
		{ 465, 57, 38 },
		{ 465, 58, 38 },
		{ 465, 59, 38 },
		{ 465, 60, 38 },
		{ 465, 61, 38 },
		{ 465, 62, 38 },
		{ 465, 63, 38 },
		{ 465, 64, 38 },
		{ 465, 65, 38 },
		{ 465, 66, 38 },
		{ 465, 67, 38 },
		{ 465, 68, 38 },
		{ 465, 69, 38 },
		{ 465, 70, 38 },
		{ 465, 71, 38 },
		{ 465, 72, 38 },
		{ 465, 73, 38 },
		{ 465, 74, 38 },
		{ 466, 19, 38 },
		{ 466, 20, 38 },
		{ 466, 21, 38 },
		{ 466, 22, 38 },
		{ 466, 23, 38 },
		{ 466, 24, 38 },
		{ 466, 25, 38 },
		{ 466, 26, 38 },
		{ 466, 27, 38 },
		{ 466, 28, 38 },
		{ 466, 29, 38 },
		{ 466, 30, 38 },
		{ 466, 31, 38 },
		{ 466, 32, 38 },
		{ 466, 33, 38 },
		{ 466, 34, 38 },
		{ 466, 35, 38 },
		{ 466, 36, 38 },
		{ 466, 37, 38 },
		{ 466, 38, 38 },
		{ 466, 39, 38 },
		{ 466, 40, 38 },
		{ 466, 41, 38 },
		{ 466, 42, 38 },
		{ 466, 43, 38 },
		{ 466, 44, 38 },
		{ 466, 45, 38 },
		{ 466, 46, 38 },
		{ 466, 47, 38 },
		{ 466, 48, 38 },
		{ 466, 49, 38 },
		{ 466, 50, 38 },
		{ 466, 51, 38 },
		{ 466, 52, 38 },
		{ 466, 53, 38 },
		{ 466, 54, 38 },
		{ 466, 55, 38 },
		{ 466, 56, 38 },
		{ 466, 57, 38 },
		{ 466, 58, 38 },
		{ 466, 59, 38 },
		{ 466, 60, 38 },
		{ 466, 61, 38 },
		{ 466, 62, 38 },
		{ 466, 63, 38 },
		{ 466, 64, 38 },
		{ 466, 65, 38 },
		{ 466, 66, 38 },
		{ 466, 67, 38 },
		{ 466, 68, 38 },
		{ 466, 69, 38 },
		{ 466, 70, 38 },
		{ 466, 71, 38 },
		{ 466, 72, 38 },
		{ 466, 73, 38 },
		{ 466, 74, 38 },
		{ 467, 18, 38 },
		{ 467, 19, 38 },
		{ 467, 20, 38 },
		{ 467, 21, 38 },
		{ 467, 22, 38 },
		{ 467, 23, 38 },
		{ 467, 24, 38 },
		{ 467, 25, 38 },
		{ 467, 26, 38 },
		{ 467, 27, 38 },
		{ 467, 28, 38 },
		{ 467, 29, 38 },
		{ 467, 30, 38 },
		{ 467, 31, 38 },
		{ 467, 32, 38 },
		{ 467, 33, 38 },
		{ 467, 34, 38 },
		{ 467, 35, 38 },
		{ 467, 36, 38 },
		{ 467, 37, 38 },
		{ 467, 38, 38 },
		{ 467, 39, 38 },
		{ 467, 40, 38 },
		{ 467, 41, 38 },
		{ 467, 42, 38 },
		{ 467, 43, 38 },
		{ 467, 44, 38 },
		{ 467, 45, 38 },
		{ 467, 46, 38 },
		{ 467, 47, 38 },
		{ 467, 48, 38 },
		{ 467, 49, 38 },
		{ 467, 50, 38 },
		{ 467, 51, 38 },
		{ 467, 52, 38 },
		{ 467, 53, 38 },
		{ 467, 54, 38 },
		{ 467, 55, 38 },
		{ 467, 56, 38 },
		{ 467, 57, 38 },
		{ 467, 58, 38 },
		{ 467, 59, 38 },
		{ 467, 60, 38 },
		{ 467, 61, 38 },
		{ 467, 62, 38 },
		{ 467, 63, 38 },
		{ 467, 64, 38 },
		{ 467, 65, 38 },
		{ 467, 66, 38 },
		{ 467, 67, 38 },
		{ 467, 68, 38 },
		{ 467, 69, 38 },
		{ 467, 70, 38 },
		{ 467, 71, 38 },
		{ 467, 72, 38 },
		{ 467, 73, 38 },
		{ 468, 16, 38 },
		{ 468, 17, 38 },
		{ 468, 18, 38 },
		{ 468, 19, 38 },
		{ 468, 20, 38 },
		{ 468, 21, 38 },
		{ 468, 22, 38 },
		{ 468, 23, 38 },
		{ 468, 24, 38 },
		{ 468, 25, 38 },
		{ 468, 26, 38 },
		{ 468, 27, 38 },
		{ 468, 28, 38 },
		{ 468, 29, 38 },
		{ 468, 30, 38 },
		{ 468, 31, 38 },
		{ 468, 32, 38 },
		{ 468, 33, 38 },
		{ 468, 34, 38 },
		{ 468, 35, 38 },
		{ 468, 36, 38 },
		{ 468, 37, 38 },
		{ 468, 38, 38 },
		{ 468, 39, 38 },
		{ 468, 40, 38 },
		{ 468, 41, 38 },
		{ 468, 42, 38 },
		{ 468, 43, 38 },
		{ 468, 44, 38 },
		{ 468, 45, 38 },
		{ 468, 46, 38 },
		{ 468, 47, 38 },
		{ 468, 48, 38 },
		{ 468, 49, 38 },
		{ 468, 50, 38 },
		{ 468, 51, 38 },
		{ 468, 52, 38 },
		{ 468, 53, 38 },
		{ 468, 54, 38 },
		{ 468, 55, 38 },
		{ 468, 56, 38 },
		{ 468, 57, 38 },
		{ 468, 58, 38 },
		{ 468, 59, 38 },
		{ 468, 60, 38 },
		{ 468, 61, 38 },
		{ 468, 62, 38 },
		{ 468, 63, 38 },
		{ 468, 64, 38 },
		{ 468, 65, 38 },
		{ 468, 66, 38 },
		{ 468, 67, 38 },
		{ 468, 68, 38 },
		{ 468, 69, 38 },
		{ 468, 70, 38 },
		{ 468, 71, 38 },
		{ 468, 72, 38 },
		{ 469, 15, 38 },
		{ 469, 16, 38 },
		{ 469, 17, 38 },
		{ 469, 18, 38 },
		{ 469, 19, 38 },
		{ 469, 20, 38 },
		{ 469, 21, 38 },
		{ 469, 22, 38 },
		{ 469, 23, 38 },
		{ 469, 24, 38 },
		{ 469, 25, 38 },
		{ 469, 26, 38 },
		{ 469, 27, 38 },
		{ 469, 28, 38 },
		{ 469, 29, 38 },
		{ 469, 30, 38 },
		{ 469, 31, 38 },
		{ 469, 32, 38 },
		{ 469, 33, 38 },
		{ 469, 34, 38 },
		{ 469, 35, 38 },
		{ 469, 36, 38 },
		{ 469, 37, 38 },
		{ 469, 38, 38 },
		{ 469, 39, 38 },
		{ 469, 40, 38 },
		{ 469, 41, 38 },
		{ 469, 42, 38 },
		{ 469, 43, 38 },
		{ 469, 44, 38 },
		{ 469, 45, 38 },
		{ 469, 46, 38 },
		{ 469, 47, 38 },
		{ 469, 48, 38 },
		{ 469, 49, 38 },
		{ 469, 50, 38 },
		{ 469, 51, 38 },
		{ 469, 52, 38 },
		{ 469, 53, 38 },
		{ 469, 54, 38 },
		{ 469, 55, 38 },
		{ 469, 56, 38 },
		{ 469, 57, 38 },
		{ 469, 58, 38 },
		{ 469, 59, 38 },
		{ 469, 60, 38 },
		{ 469, 61, 38 },
		{ 469, 62, 38 },
		{ 469, 63, 38 },
		{ 469, 64, 38 },
		{ 469, 65, 38 },
		{ 469, 66, 38 },
		{ 469, 67, 38 },
		{ 469, 68, 38 },
		{ 469, 69, 38 },
		{ 469, 70, 38 },
		{ 469, 71, 38 },
		{ 470, 14, 38 },
		{ 470, 15, 38 },
		{ 470, 16, 38 },
		{ 470, 17, 38 },
		{ 470, 18, 38 },
		{ 470, 19, 38 },
		{ 470, 20, 38 },
		{ 470, 21, 38 },
		{ 470, 22, 38 },
		{ 470, 23, 38 },
		{ 470, 24, 38 },
		{ 470, 25, 38 },
		{ 470, 26, 38 },
		{ 470, 27, 38 },
		{ 470, 28, 38 },
		{ 470, 29, 38 },
		{ 470, 30, 38 },
		{ 470, 31, 38 },
		{ 470, 32, 38 },
		{ 470, 33, 38 },
		{ 470, 34, 38 },
		{ 470, 35, 38 },
		{ 470, 36, 38 },
		{ 470, 37, 38 },
		{ 470, 38, 38 },
		{ 470, 39, 38 },
		{ 470, 40, 38 },
		{ 470, 41, 38 },
		{ 470, 42, 38 },
		{ 470, 43, 38 },
		{ 470, 44, 38 },
		{ 470, 45, 38 },
		{ 470, 46, 38 },
		{ 470, 47, 38 },
		{ 470, 48, 38 },
		{ 470, 49, 38 },
		{ 470, 50, 38 },
		{ 470, 51, 38 },
		{ 470, 52, 38 },
		{ 470, 53, 38 },
		{ 470, 54, 38 },
		{ 470, 55, 38 },
		{ 470, 56, 38 },
		{ 470, 57, 38 },
		{ 470, 58, 38 },
		{ 470, 59, 38 },
		{ 470, 60, 38 },
		{ 470, 61, 38 },
		{ 470, 62, 38 },
		{ 470, 63, 38 },
		{ 470, 64, 38 },
		{ 470, 65, 38 },
		{ 470, 66, 38 },
		{ 470, 67, 38 },
		{ 470, 68, 38 },
		{ 470, 69, 38 },
		{ 470, 70, 38 },
		{ 471, 13, 38 },
		{ 471, 14, 38 },
		{ 471, 15, 38 },
		{ 471, 16, 38 },
		{ 471, 17, 38 },
		{ 471, 18, 38 },
		{ 471, 19, 38 },
		{ 471, 20, 38 },
		{ 471, 21, 38 },
		{ 471, 22, 38 },
		{ 471, 23, 38 },
		{ 471, 24, 38 },
		{ 471, 25, 38 },
		{ 471, 26, 38 },
		{ 471, 27, 38 },
		{ 471, 28, 38 },
		{ 471, 29, 38 },
		{ 471, 30, 38 },
		{ 471, 31, 38 },
		{ 471, 32, 38 },
		{ 471, 33, 38 },
		{ 471, 34, 38 },
		{ 471, 35, 38 },
		{ 471, 36, 38 },
		{ 471, 37, 38 },
		{ 471, 38, 38 },
		{ 471, 39, 38 },
		{ 471, 40, 38 },
		{ 471, 41, 38 },
		{ 471, 42, 38 },
		{ 471, 43, 38 },
		{ 471, 44, 38 },
		{ 471, 45, 38 },
		{ 471, 46, 38 },
		{ 471, 47, 38 },
		{ 471, 48, 38 },
		{ 471, 49, 38 },
		{ 471, 50, 38 },
		{ 471, 51, 38 },
		{ 471, 52, 38 },
		{ 471, 53, 38 },
		{ 471, 54, 38 },
		{ 471, 55, 38 },
		{ 471, 56, 38 },
		{ 471, 57, 38 },
		{ 471, 58, 38 },
		{ 471, 59, 38 },
		{ 471, 60, 38 },
		{ 471, 61, 38 },
		{ 471, 62, 38 },
		{ 471, 63, 38 },
		{ 471, 64, 38 },
		{ 471, 65, 38 },
		{ 471, 66, 38 },
		{ 471, 67, 38 },
		{ 471, 68, 38 },
		{ 471, 69, 38 },
		{ 472, 12, 38 },
		{ 472, 13, 38 },
		{ 472, 14, 38 },
		{ 472, 15, 38 },
		{ 472, 16, 38 },
		{ 472, 17, 38 },
		{ 472, 18, 38 },
		{ 472, 19, 38 },
		{ 472, 20, 38 },
		{ 472, 21, 38 },
		{ 472, 22, 38 },
		{ 472, 23, 38 },
		{ 472, 24, 38 },
		{ 472, 25, 38 },
		{ 472, 26, 38 },
		{ 472, 27, 38 },
		{ 472, 28, 38 },
		{ 472, 29, 38 },
		{ 472, 30, 38 },
		{ 472, 31, 38 },
		{ 472, 32, 38 },
		{ 472, 33, 38 },
		{ 472, 34, 38 },
		{ 472, 35, 38 },
		{ 472, 36, 38 },
		{ 472, 37, 38 },
		{ 472, 38, 38 },
		{ 472, 39, 38 },
		{ 472, 40, 38 },
		{ 472, 41, 38 },
		{ 472, 42, 38 },
		{ 472, 43, 38 },
		{ 472, 44, 38 },
		{ 472, 45, 38 },
		{ 472, 46, 38 },
		{ 472, 47, 38 },
		{ 472, 48, 38 },
		{ 472, 49, 38 },
		{ 472, 50, 38 },
		{ 472, 51, 38 },
		{ 472, 52, 38 },
		{ 472, 53, 38 },
		{ 472, 54, 38 },
		{ 472, 55, 38 },
		{ 472, 56, 38 },
		{ 472, 57, 38 },
		{ 472, 58, 38 },
		{ 472, 59, 38 },
		{ 472, 60, 38 },
		{ 472, 61, 38 },
		{ 472, 62, 38 },
		{ 472, 63, 38 },
		{ 472, 64, 38 },
		{ 472, 65, 38 },
		{ 472, 66, 38 },
		{ 472, 67, 38 },
		{ 472, 68, 38 },
		{ 473, 11, 38 },
		{ 473, 12, 38 },
		{ 473, 13, 38 },
		{ 473, 14, 38 },
		{ 473, 15, 38 },
		{ 473, 16, 38 },
		{ 473, 17, 38 },
		{ 473, 18, 38 },
		{ 473, 19, 38 },
		{ 473, 20, 38 },
		{ 473, 21, 38 },
		{ 473, 22, 38 },
		{ 473, 23, 38 },
		{ 473, 24, 38 },
		{ 473, 25, 38 },
		{ 473, 26, 38 },
		{ 473, 27, 38 },
		{ 473, 28, 38 },
		{ 473, 29, 38 },
		{ 473, 30, 38 },
		{ 473, 31, 38 },
		{ 473, 32, 38 },
		{ 473, 33, 38 },
		{ 473, 34, 38 },
		{ 473, 35, 38 },
		{ 473, 36, 38 },
		{ 473, 37, 38 },
		{ 473, 38, 38 },
		{ 473, 39, 38 },
		{ 473, 40, 38 },
		{ 473, 41, 38 },
		{ 473, 42, 38 },
		{ 473, 43, 38 },
		{ 473, 44, 38 },
		{ 473, 45, 38 },
		{ 473, 46, 38 },
		{ 473, 47, 38 },
		{ 473, 48, 38 },
		{ 473, 49, 38 },
		{ 473, 50, 38 },
		{ 473, 51, 38 },
		{ 473, 52, 38 },
		{ 473, 53, 38 },
		{ 473, 54, 38 },
		{ 473, 55, 38 },
		{ 473, 56, 38 },
		{ 473, 57, 38 },
		{ 473, 58, 38 },
		{ 473, 59, 38 },
		{ 473, 60, 38 },
		{ 473, 61, 38 },
		{ 473, 62, 38 },
		{ 473, 63, 38 },
		{ 473, 64, 38 },
		{ 473, 65, 38 },
		{ 473, 66, 38 },
		{ 473, 67, 38 },
		{ 474, 10, 38 },
		{ 474, 11, 38 },
		{ 474, 12, 38 },
		{ 474, 13, 38 },
		{ 474, 14, 38 },
		{ 474, 15, 38 },
		{ 474, 16, 38 },
		{ 474, 17, 38 },
		{ 474, 18, 38 },
		{ 474, 19, 38 },
		{ 474, 20, 38 },
		{ 474, 21, 38 },
		{ 474, 22, 38 },
		{ 474, 23, 38 },
		{ 474, 24, 38 },
		{ 474, 25, 38 },
		{ 474, 26, 38 },
		{ 474, 27, 38 },
		{ 474, 28, 38 },
		{ 474, 29, 38 },
		{ 474, 30, 38 },
		{ 474, 31, 38 },
		{ 474, 32, 38 },
		{ 474, 33, 38 },
		{ 474, 34, 38 },
		{ 474, 35, 38 },
		{ 474, 36, 38 },
		{ 474, 37, 38 },
		{ 474, 38, 38 },
		{ 474, 39, 38 },
		{ 474, 40, 38 },
		{ 474, 41, 38 },
		{ 474, 42, 38 },
		{ 474, 43, 38 },
		{ 474, 44, 38 },
		{ 474, 45, 38 },
		{ 474, 46, 38 },
		{ 474, 47, 38 },
		{ 474, 48, 38 },
		{ 474, 49, 38 },
		{ 474, 50, 38 },
		{ 474, 51, 38 },
		{ 474, 52, 38 },
		{ 474, 53, 38 },
		{ 474, 54, 38 },
		{ 474, 55, 38 },
		{ 474, 56, 38 },
		{ 474, 57, 38 },
		{ 474, 58, 38 },
		{ 474, 59, 38 },
		{ 474, 60, 38 },
		{ 474, 61, 38 },
		{ 474, 62, 38 },
		{ 474, 63, 38 },
		{ 474, 64, 38 },
		{ 474, 65, 38 },
		{ 474, 66, 38 },
		{ 474, 67, 38 },
		{ 475, 8, 38 },
		{ 475, 9, 38 },
		{ 475, 10, 38 },
		{ 475, 11, 38 },
		{ 475, 12, 38 },
		{ 475, 13, 38 },
		{ 475, 14, 38 },
		{ 475, 15, 38 },
		{ 475, 16, 38 },
		{ 475, 17, 38 },
		{ 475, 18, 38 },
		{ 475, 19, 38 },
		{ 475, 20, 38 },
		{ 475, 21, 38 },
		{ 475, 22, 38 },
		{ 475, 23, 38 },
		{ 475, 24, 38 },
		{ 475, 25, 38 },
		{ 475, 26, 38 },
		{ 475, 27, 38 },
		{ 475, 28, 38 },
		{ 475, 29, 38 },
		{ 475, 30, 38 },
		{ 475, 31, 38 },
		{ 475, 32, 38 },
		{ 475, 33, 38 },
		{ 475, 34, 38 },
		{ 475, 35, 38 },
		{ 475, 36, 38 },
		{ 475, 37, 38 },
		{ 475, 38, 38 },
		{ 475, 39, 38 },
		{ 475, 40, 38 },
		{ 475, 41, 38 },
		{ 475, 42, 38 },
		{ 475, 43, 38 },
		{ 475, 44, 38 },
		{ 475, 45, 38 },
		{ 475, 46, 38 },
		{ 475, 47, 38 },
		{ 475, 48, 38 },
		{ 475, 49, 38 },
		{ 475, 50, 38 },
		{ 475, 51, 38 },
		{ 475, 52, 38 },
		{ 475, 53, 38 },
		{ 475, 54, 38 },
		{ 475, 55, 38 },
		{ 475, 56, 38 },
		{ 475, 57, 38 },
		{ 475, 58, 38 },
		{ 475, 59, 38 },
		{ 475, 60, 38 },
		{ 475, 61, 38 },
		{ 475, 62, 38 },
		{ 475, 63, 38 },
		{ 475, 64, 38 },
		{ 475, 65, 38 },
		{ 475, 66, 38 },
		{ 476, 7, 38 },
		{ 476, 8, 38 },
		{ 476, 9, 38 },
		{ 476, 10, 38 },
		{ 476, 11, 38 },
		{ 476, 12, 38 },
		{ 476, 13, 38 },
		{ 476, 14, 38 },
		{ 476, 15, 38 },
		{ 476, 16, 38 },
		{ 476, 17, 38 },
		{ 476, 18, 38 },
		{ 476, 19, 38 },
		{ 476, 20, 38 },
		{ 476, 21, 38 },
		{ 476, 22, 38 },
		{ 476, 23, 38 },
		{ 476, 24, 38 },
		{ 476, 25, 38 },
		{ 476, 26, 38 },
		{ 476, 27, 38 },
		{ 476, 28, 38 },
		{ 476, 29, 38 },
		{ 476, 30, 38 },
		{ 476, 31, 38 },
		{ 476, 32, 38 },
		{ 476, 33, 38 },
		{ 476, 34, 38 },
		{ 476, 35, 38 },
		{ 476, 36, 38 },
		{ 476, 37, 38 },
		{ 476, 38, 38 },
		{ 476, 39, 38 },
		{ 476, 40, 38 },
		{ 476, 41, 38 },
		{ 476, 42, 38 },
		{ 476, 43, 38 },
		{ 476, 44, 38 },
		{ 476, 45, 38 },
		{ 476, 46, 38 },
		{ 476, 47, 38 },
		{ 476, 48, 38 },
		{ 476, 49, 38 },
		{ 476, 50, 38 },
		{ 476, 51, 38 },
		{ 476, 52, 38 },
		{ 476, 53, 38 },
		{ 476, 54, 38 },
		{ 476, 55, 38 },
		{ 476, 56, 38 },
		{ 476, 57, 38 },
		{ 476, 58, 38 },
		{ 476, 59, 38 },
		{ 476, 60, 38 },
		{ 476, 61, 38 },
		{ 476, 62, 38 },
		{ 476, 63, 38 },
		{ 476, 64, 38 },
		{ 476, 65, 38 },
		{ 477, 6, 38 },
		{ 477, 7, 38 },
		{ 477, 8, 38 },
		{ 477, 9, 38 },
		{ 477, 10, 38 },
		{ 477, 11, 38 },
		{ 477, 12, 38 },
		{ 477, 13, 38 },
		{ 477, 14, 38 },
		{ 477, 15, 38 },
		{ 477, 16, 38 },
		{ 477, 17, 38 },
		{ 477, 18, 38 },
		{ 477, 19, 38 },
		{ 477, 20, 38 },
		{ 477, 21, 38 },
		{ 477, 22, 38 },
		{ 477, 23, 38 },
		{ 477, 24, 38 },
		{ 477, 25, 38 },
		{ 477, 26, 38 },
		{ 477, 27, 38 },
		{ 477, 28, 38 },
		{ 477, 29, 38 },
		{ 477, 30, 38 },
		{ 477, 31, 38 },
		{ 477, 32, 38 },
		{ 477, 33, 38 },
		{ 477, 34, 38 },
		{ 477, 35, 38 },
		{ 477, 36, 38 },
		{ 477, 37, 38 },
		{ 477, 38, 38 },
		{ 477, 39, 38 },
		{ 477, 40, 38 },
		{ 477, 41, 38 },
		{ 477, 42, 38 },
		{ 477, 43, 38 },
		{ 477, 44, 38 },
		{ 477, 45, 38 },
		{ 477, 46, 38 },
		{ 477, 47, 38 },
		{ 477, 48, 38 },
		{ 477, 49, 38 },
		{ 477, 50, 38 },
		{ 477, 51, 38 },
		{ 477, 52, 38 },
		{ 477, 53, 38 },
		{ 477, 54, 38 },
		{ 477, 55, 38 },
		{ 477, 56, 38 },
		{ 477, 57, 38 },
		{ 477, 58, 38 },
		{ 477, 59, 38 },
		{ 477, 60, 38 },
		{ 477, 61, 38 },
		{ 477, 62, 38 },
		{ 477, 63, 38 },
		{ 477, 64, 38 },
		{ 478, 5, 38 },
		{ 478, 6, 38 },
		{ 478, 7, 38 },
		{ 478, 8, 38 },
		{ 478, 9, 38 },
		{ 478, 10, 38 },
		{ 478, 11, 38 },
		{ 478, 12, 38 },
		{ 478, 13, 38 },
		{ 478, 14, 38 },
		{ 478, 15, 38 },
		{ 478, 16, 38 },
		{ 478, 17, 38 },
		{ 478, 18, 38 },
		{ 478, 19, 38 },
		{ 478, 20, 38 },
		{ 478, 21, 38 },
		{ 478, 22, 38 },
		{ 478, 23, 38 },
		{ 478, 24, 38 },
		{ 478, 25, 38 },
		{ 478, 26, 38 },
		{ 478, 27, 38 },
		{ 478, 28, 38 },
		{ 478, 29, 38 },
		{ 478, 30, 38 },
		{ 478, 31, 38 },
		{ 478, 32, 38 },
		{ 478, 33, 38 },
		{ 478, 34, 38 },
		{ 478, 35, 38 },
		{ 478, 36, 38 },
		{ 478, 37, 38 },
		{ 478, 38, 38 },
		{ 478, 39, 38 },
		{ 478, 40, 38 },
		{ 478, 41, 38 },
		{ 478, 42, 38 },
		{ 478, 43, 38 },
		{ 478, 44, 38 },
		{ 478, 45, 38 },
		{ 478, 46, 38 },
		{ 478, 47, 38 },
		{ 478, 48, 38 },
		{ 478, 49, 38 },
		{ 478, 50, 38 },
		{ 478, 51, 38 },
		{ 478, 52, 38 },
		{ 478, 53, 38 },
		{ 478, 54, 38 },
		{ 478, 55, 38 },
		{ 478, 56, 38 },
		{ 478, 57, 38 },
		{ 478, 58, 38 },
		{ 478, 59, 38 },
		{ 478, 60, 38 },
		{ 478, 61, 38 },
		{ 478, 62, 38 },
		{ 478, 63, 38 },
		{ 479, 4, 38 },
		{ 479, 5, 38 },
		{ 479, 6, 38 },
		{ 479, 7, 38 },
		{ 479, 8, 38 },
		{ 479, 9, 38 },
		{ 479, 10, 38 },
		{ 479, 11, 38 },
		{ 479, 12, 38 },
		{ 479, 13, 38 },
		{ 479, 14, 38 },
		{ 479, 15, 38 },
		{ 479, 16, 38 },
		{ 479, 17, 38 },
		{ 479, 18, 38 },
		{ 479, 19, 38 },
		{ 479, 20, 38 },
		{ 479, 21, 38 },
		{ 479, 22, 38 },
		{ 479, 23, 38 },
		{ 479, 24, 38 },
		{ 479, 25, 38 },
		{ 479, 26, 38 },
		{ 479, 27, 38 },
		{ 479, 28, 38 },
		{ 479, 29, 38 },
		{ 479, 30, 38 },
		{ 479, 31, 38 },
		{ 479, 32, 38 },
		{ 479, 33, 38 },
		{ 479, 34, 38 },
		{ 479, 35, 38 },
		{ 479, 36, 38 },
		{ 479, 37, 38 },
		{ 479, 38, 38 },
		{ 479, 39, 38 },
		{ 479, 40, 38 },
		{ 479, 41, 38 },
		{ 479, 42, 38 },
		{ 479, 43, 38 },
		{ 479, 44, 38 },
		{ 479, 45, 38 },
		{ 479, 46, 38 },
		{ 479, 47, 38 },
		{ 479, 48, 38 },
		{ 479, 49, 38 },
		{ 479, 50, 38 },
		{ 479, 51, 38 },
		{ 479, 52, 38 },
		{ 479, 53, 38 },
		{ 479, 54, 38 },
		{ 479, 55, 38 },
		{ 479, 56, 38 },
		{ 479, 57, 38 },
		{ 479, 58, 38 },
		{ 479, 59, 38 },
		{ 479, 60, 38 },
		{ 479, 61, 38 },
		{ 479, 62, 38 },
		{ 480, 3, 38 },
		{ 480, 4, 38 },
		{ 480, 5, 38 },
		{ 480, 6, 38 },
		{ 480, 7, 38 },
		{ 480, 8, 38 },
		{ 480, 9, 38 },
		{ 480, 10, 38 },
		{ 480, 11, 38 },
		{ 480, 12, 38 },
		{ 480, 13, 38 },
		{ 480, 14, 38 },
		{ 480, 15, 38 },
		{ 480, 16, 38 },
		{ 480, 17, 38 },
		{ 480, 18, 38 },
		{ 480, 19, 38 },
		{ 480, 20, 38 },
		{ 480, 21, 38 },
		{ 480, 22, 38 },
		{ 480, 23, 38 },
		{ 480, 24, 38 },
		{ 480, 25, 38 },
		{ 480, 26, 38 },
		{ 480, 27, 38 },
		{ 480, 28, 38 },
		{ 480, 29, 38 },
		{ 480, 30, 38 },
		{ 480, 31, 38 },
		{ 480, 32, 38 },
		{ 480, 33, 38 },
		{ 480, 34, 38 },
		{ 480, 35, 38 },
		{ 480, 36, 38 },
		{ 480, 37, 38 },
		{ 480, 38, 38 },
		{ 480, 39, 38 },
		{ 480, 40, 38 },
		{ 480, 41, 38 },
		{ 480, 42, 38 },
		{ 480, 43, 38 },
		{ 480, 44, 38 },
		{ 480, 45, 38 },
		{ 480, 46, 38 },
		{ 480, 47, 38 },
		{ 480, 48, 38 },
		{ 480, 49, 38 },
		{ 480, 50, 38 },
		{ 480, 51, 38 },
		{ 480, 52, 38 },
		{ 480, 53, 38 },
		{ 480, 54, 38 },
		{ 480, 55, 38 },
		{ 480, 56, 38 },
		{ 480, 57, 38 },
		{ 480, 58, 38 },
		{ 480, 59, 38 },
		{ 480, 60, 38 },
		{ 480, 61, 38 },
		{ 481, 2, 38 },
		{ 481, 3, 38 },
		{ 481, 4, 38 },
		{ 481, 5, 38 },
		{ 481, 6, 38 },
		{ 481, 7, 38 },
		{ 481, 8, 38 },
		{ 481, 9, 38 },
		{ 481, 10, 38 },
		{ 481, 11, 38 },
		{ 481, 12, 38 },
		{ 481, 13, 38 },
		{ 481, 14, 38 },
		{ 481, 15, 38 },
		{ 481, 16, 38 },
		{ 481, 17, 38 },
		{ 481, 18, 38 },
		{ 481, 19, 38 },
		{ 481, 20, 38 },
		{ 481, 21, 38 },
		{ 481, 22, 38 },
		{ 481, 23, 38 },
		{ 481, 24, 38 },
		{ 481, 25, 38 },
		{ 481, 26, 38 },
		{ 481, 27, 38 },
		{ 481, 28, 38 },
		{ 481, 29, 38 },
		{ 481, 30, 38 },
		{ 481, 31, 38 },
		{ 481, 32, 38 },
		{ 481, 33, 38 },
		{ 481, 34, 38 },
		{ 481, 35, 38 },
		{ 481, 36, 38 },
		{ 481, 37, 38 },
		{ 481, 38, 38 },
		{ 481, 39, 38 },
		{ 481, 40, 38 },
		{ 481, 41, 38 },
		{ 481, 42, 38 },
		{ 481, 43, 38 },
		{ 481, 44, 38 },
		{ 481, 45, 38 },
		{ 481, 46, 38 },
		{ 481, 47, 38 },
		{ 481, 48, 38 },
		{ 481, 49, 38 },
		{ 481, 50, 38 },
		{ 481, 51, 38 },
		{ 481, 52, 38 },
		{ 481, 53, 38 },
		{ 481, 54, 38 },
		{ 481, 55, 38 },
		{ 481, 56, 38 },
		{ 481, 57, 38 },
		{ 481, 58, 38 },
		{ 481, 59, 38 },
		{ 481, 60, 38 },
		{ 482, 1, 38 },
		{ 482, 2, 38 },
		{ 482, 3, 38 },
		{ 482, 4, 38 },
		{ 482, 5, 38 },
		{ 482, 6, 38 },
		{ 482, 7, 38 },
		{ 482, 8, 38 },
		{ 482, 9, 38 },
		{ 482, 10, 38 },
		{ 482, 11, 38 },
		{ 482, 12, 38 },
		{ 482, 13, 38 },
		{ 482, 14, 38 },
		{ 482, 15, 38 },
		{ 482, 16, 38 },
		{ 482, 17, 38 },
		{ 482, 18, 38 },
		{ 482, 19, 38 },
		{ 482, 20, 38 },
		{ 482, 21, 38 },
		{ 482, 22, 38 },
		{ 482, 23, 38 },
		{ 482, 24, 38 },
		{ 482, 25, 38 },
		{ 482, 26, 38 },
		{ 482, 27, 38 },
		{ 482, 28, 38 },
		{ 482, 29, 38 },
		{ 482, 30, 38 },
		{ 482, 31, 38 },
		{ 482, 32, 38 },
		{ 482, 33, 38 },
		{ 482, 34, 38 },
		{ 482, 35, 38 },
		{ 482, 36, 38 },
		{ 482, 37, 38 },
		{ 482, 38, 38 },
		{ 482, 39, 38 },
		{ 482, 40, 38 },
		{ 482, 41, 38 },
		{ 482, 42, 38 },
		{ 482, 43, 38 },
		{ 482, 44, 38 },
		{ 482, 45, 38 },
		{ 482, 46, 38 },
		{ 482, 47, 38 },
		{ 482, 48, 38 },
		{ 482, 49, 38 },
		{ 482, 50, 38 },
		{ 482, 51, 38 },
		{ 482, 52, 38 },
		{ 482, 53, 38 },
		{ 482, 54, 38 },
		{ 482, 55, 38 },
		{ 482, 56, 38 },
		{ 482, 57, 38 },
		{ 482, 58, 38 },
		{ 482, 59, 38 },
		{ 482, 60, 38 },
		{ 483, 1, 38 },
		{ 483, 2, 38 },
		{ 483, 3, 38 },
		{ 483, 4, 38 },
		{ 483, 5, 38 },
		{ 483, 6, 38 },
		{ 483, 7, 38 },
		{ 483, 8, 38 },
		{ 483, 9, 38 },
		{ 483, 10, 38 },
		{ 483, 11, 38 },
		{ 483, 12, 38 },
		{ 483, 13, 38 },
		{ 483, 14, 38 },
		{ 483, 15, 38 },
		{ 483, 16, 38 },
		{ 483, 17, 38 },
		{ 483, 18, 38 },
		{ 483, 19, 38 },
		{ 483, 20, 38 },
		{ 483, 21, 38 },
		{ 483, 22, 38 },
		{ 483, 23, 38 },
		{ 483, 24, 38 },
		{ 483, 25, 38 },
		{ 483, 26, 38 },
		{ 483, 27, 38 },
		{ 483, 28, 38 },
		{ 483, 29, 38 },
		{ 483, 30, 38 },
		{ 483, 31, 38 },
		{ 483, 32, 38 },
		{ 483, 33, 38 },
		{ 483, 34, 38 },
		{ 483, 35, 38 },
		{ 483, 36, 38 },
		{ 483, 37, 38 },
		{ 483, 38, 38 },
		{ 483, 39, 38 },
		{ 483, 40, 38 },
		{ 483, 41, 38 },
		{ 483, 42, 38 },
		{ 483, 43, 38 },
		{ 483, 44, 38 },
		{ 483, 45, 38 },
		{ 483, 46, 38 },
		{ 483, 47, 38 },
		{ 483, 48, 38 },
		{ 483, 49, 38 },
		{ 483, 50, 38 },
		{ 483, 51, 38 },
		{ 483, 52, 38 },
		{ 483, 53, 38 },
		{ 483, 54, 38 },
		{ 483, 55, 38 },
		{ 483, 56, 38 },
		{ 483, 57, 38 },
		{ 483, 58, 38 },
		{ 483, 59, 38 },
		{ 484, 1, 38 },
		{ 484, 2, 38 },
		{ 484, 3, 38 },
		{ 484, 4, 38 },
		{ 484, 5, 38 },
		{ 484, 6, 38 },
		{ 484, 7, 38 },
		{ 484, 8, 38 },
		{ 484, 9, 38 },
		{ 484, 10, 38 },
		{ 484, 11, 38 },
		{ 484, 12, 38 },
		{ 484, 13, 38 },
		{ 484, 14, 38 },
		{ 484, 15, 38 },
		{ 484, 16, 38 },
		{ 484, 17, 38 },
		{ 484, 18, 38 },
		{ 484, 19, 38 },
		{ 484, 20, 38 },
		{ 484, 21, 38 },
		{ 484, 22, 38 },
		{ 484, 23, 38 },
		{ 484, 24, 38 },
		{ 484, 25, 38 },
		{ 484, 26, 38 },
		{ 484, 27, 38 },
		{ 484, 28, 38 },
		{ 484, 29, 38 },
		{ 484, 30, 38 },
		{ 484, 31, 38 },
		{ 484, 32, 38 },
		{ 484, 33, 38 },
		{ 484, 34, 38 },
		{ 484, 35, 38 },
		{ 484, 36, 38 },
		{ 484, 37, 38 },
		{ 484, 38, 38 },
		{ 484, 39, 38 },
		{ 484, 40, 38 },
		{ 484, 41, 38 },
		{ 484, 42, 38 },
		{ 484, 43, 38 },
		{ 484, 44, 38 },
		{ 484, 45, 38 },
		{ 484, 46, 38 },
		{ 484, 47, 38 },
		{ 484, 48, 38 },
		{ 484, 49, 38 },
		{ 484, 50, 38 },
		{ 484, 51, 38 },
		{ 484, 52, 38 },
		{ 484, 53, 38 },
		{ 484, 54, 38 },
		{ 484, 55, 38 },
		{ 484, 56, 38 },
		{ 484, 57, 38 },
		{ 484, 58, 38 },
		{ 485, 1, 38 },
		{ 485, 2, 38 },
		{ 485, 3, 38 },
		{ 485, 4, 38 },
		{ 485, 5, 38 },
		{ 485, 6, 38 },
		{ 485, 7, 38 },
		{ 485, 8, 38 },
		{ 485, 9, 38 },
		{ 485, 10, 38 },
		{ 485, 11, 38 },
		{ 485, 12, 38 },
		{ 485, 13, 38 },
		{ 485, 14, 38 },
		{ 485, 15, 38 },
		{ 485, 16, 38 },
		{ 485, 17, 38 },
		{ 485, 18, 38 },
		{ 485, 19, 38 },
		{ 485, 20, 38 },
		{ 485, 21, 38 },
		{ 485, 22, 38 },
		{ 485, 23, 38 },
		{ 485, 24, 38 },
		{ 485, 25, 38 },
		{ 485, 26, 38 },
		{ 485, 27, 38 },
		{ 485, 28, 38 },
		{ 485, 29, 38 },
		{ 485, 30, 38 },
		{ 485, 31, 38 },
		{ 485, 32, 38 },
		{ 485, 33, 38 },
		{ 485, 34, 38 },
		{ 485, 35, 38 },
		{ 485, 36, 38 },
		{ 485, 37, 38 },
		{ 485, 38, 38 },
		{ 485, 39, 38 },
		{ 485, 40, 38 },
		{ 485, 41, 38 },
		{ 485, 42, 38 },
		{ 485, 43, 38 },
		{ 485, 44, 38 },
		{ 485, 45, 38 },
		{ 485, 46, 38 },
		{ 485, 47, 38 },
		{ 485, 48, 38 },
		{ 485, 49, 38 },
		{ 485, 50, 38 },
		{ 485, 51, 38 },
		{ 485, 52, 38 },
		{ 485, 53, 38 },
		{ 485, 54, 38 },
		{ 485, 55, 38 },
		{ 485, 56, 38 },
		{ 485, 57, 38 },
		{ 486, 1, 38 },
		{ 486, 2, 38 },
		{ 486, 3, 38 },
		{ 486, 4, 38 },
		{ 486, 5, 38 },
		{ 486, 6, 38 },
		{ 486, 7, 38 },
		{ 486, 8, 38 },
		{ 486, 9, 38 },
		{ 486, 10, 38 },
		{ 486, 11, 38 },
		{ 486, 12, 38 },
		{ 486, 13, 38 },
		{ 486, 14, 38 },
		{ 486, 15, 38 },
		{ 486, 16, 38 },
		{ 486, 17, 38 },
		{ 486, 18, 38 },
		{ 486, 19, 38 },
		{ 486, 20, 38 },
		{ 486, 21, 38 },
		{ 486, 22, 38 },
		{ 486, 23, 38 },
		{ 486, 24, 38 },
		{ 486, 25, 38 },
		{ 486, 26, 38 },
		{ 486, 27, 38 },
		{ 486, 28, 38 },
		{ 486, 29, 38 },
		{ 486, 30, 38 },
		{ 486, 31, 38 },
		{ 486, 32, 38 },
		{ 486, 33, 38 },
		{ 486, 34, 38 },
		{ 486, 35, 38 },
		{ 486, 36, 38 },
		{ 486, 37, 38 },
		{ 486, 38, 38 },
		{ 486, 39, 38 },
		{ 486, 40, 38 },
		{ 486, 41, 38 },
		{ 486, 42, 38 },
		{ 486, 43, 38 },
		{ 486, 44, 38 },
		{ 486, 45, 38 },
		{ 486, 46, 38 },
		{ 486, 47, 38 },
		{ 486, 48, 38 },
		{ 486, 49, 38 },
		{ 486, 50, 38 },
		{ 486, 51, 38 },
		{ 486, 52, 38 },
		{ 486, 53, 38 },
		{ 486, 54, 38 },
		{ 486, 55, 38 },
		{ 486, 56, 38 },
		{ 487, 1, 38 },
		{ 487, 2, 38 },
		{ 487, 3, 38 },
		{ 487, 4, 38 },
		{ 487, 5, 38 },
		{ 487, 6, 38 },
		{ 487, 7, 38 },
		{ 487, 8, 38 },
		{ 487, 9, 38 },
		{ 487, 10, 38 },
		{ 487, 11, 38 },
		{ 487, 12, 38 },
		{ 487, 13, 38 },
		{ 487, 14, 38 },
		{ 487, 15, 38 },
		{ 487, 16, 38 },
		{ 487, 17, 38 },
		{ 487, 18, 38 },
		{ 487, 19, 38 },
		{ 487, 20, 38 },
		{ 487, 21, 38 },
		{ 487, 22, 38 },
		{ 487, 23, 38 },
		{ 487, 24, 38 },
		{ 487, 25, 38 },
		{ 487, 26, 38 },
		{ 487, 27, 38 },
		{ 487, 28, 38 },
		{ 487, 29, 38 },
		{ 487, 30, 38 },
		{ 487, 31, 38 },
		{ 487, 32, 38 },
		{ 487, 33, 38 },
		{ 487, 34, 38 },
		{ 487, 35, 38 },
		{ 487, 36, 38 },
		{ 487, 37, 38 },
		{ 487, 38, 38 },
		{ 487, 39, 38 },
		{ 487, 40, 38 },
		{ 487, 41, 38 },
		{ 487, 42, 38 },
		{ 487, 43, 38 },
		{ 487, 44, 38 },
		{ 487, 45, 38 },
		{ 487, 46, 38 },
		{ 487, 47, 38 },
		{ 487, 48, 38 },
		{ 487, 49, 38 },
		{ 487, 50, 38 },
		{ 487, 51, 38 },
		{ 487, 52, 38 },
		{ 487, 53, 38 },
		{ 487, 54, 38 },
		{ 487, 55, 38 },
		{ 488, 1, 38 },
		{ 488, 2, 38 },
		{ 488, 3, 38 },
		{ 488, 4, 38 },
		{ 488, 5, 38 },
		{ 488, 6, 38 },
		{ 488, 7, 38 },
		{ 488, 8, 38 },
		{ 488, 9, 38 },
		{ 488, 10, 38 },
		{ 488, 11, 38 },
		{ 488, 12, 38 },
		{ 488, 13, 38 },
		{ 488, 14, 38 },
		{ 488, 15, 38 },
		{ 488, 16, 38 },
		{ 488, 17, 38 },
		{ 488, 18, 38 },
		{ 488, 19, 38 },
		{ 488, 20, 38 },
		{ 488, 21, 38 },
		{ 488, 22, 38 },
		{ 488, 23, 38 },
		{ 488, 24, 38 },
		{ 488, 25, 38 },
		{ 488, 26, 38 },
		{ 488, 27, 38 },
		{ 488, 28, 38 },
		{ 488, 29, 38 },
		{ 488, 30, 38 },
		{ 488, 31, 38 },
		{ 488, 32, 38 },
		{ 488, 33, 38 },
		{ 488, 34, 38 },
		{ 488, 35, 38 },
		{ 488, 36, 38 },
		{ 488, 37, 38 },
		{ 488, 38, 38 },
		{ 488, 39, 38 },
		{ 488, 40, 38 },
		{ 488, 41, 38 },
		{ 488, 42, 38 },
		{ 488, 43, 38 },
		{ 488, 44, 38 },
		{ 488, 45, 38 },
		{ 488, 46, 38 },
		{ 488, 47, 38 },
		{ 488, 48, 38 },
		{ 488, 49, 38 },
		{ 488, 50, 38 },
		{ 488, 51, 38 },
		{ 488, 52, 38 },
		{ 488, 53, 38 },
		{ 488, 54, 38 },
		{ 489, 1, 38 },
		{ 489, 2, 38 },
		{ 489, 3, 38 },
		{ 489, 4, 38 },
		{ 489, 5, 38 },
		{ 489, 6, 38 },
		{ 489, 7, 38 },
		{ 489, 8, 38 },
		{ 489, 9, 38 },
		{ 489, 10, 38 },
		{ 489, 11, 38 },
		{ 489, 12, 38 },
		{ 489, 13, 38 },
		{ 489, 14, 38 },
		{ 489, 15, 38 },
		{ 489, 16, 38 },
		{ 489, 17, 38 },
		{ 489, 18, 38 },
		{ 489, 19, 38 },
		{ 489, 20, 38 },
		{ 489, 21, 38 },
		{ 489, 22, 38 },
		{ 489, 23, 38 },
		{ 489, 24, 38 },
		{ 489, 25, 38 },
		{ 489, 26, 38 },
		{ 489, 27, 38 },
		{ 489, 28, 38 },
		{ 489, 29, 38 },
		{ 489, 30, 38 },
		{ 489, 31, 38 },
		{ 489, 32, 38 },
		{ 489, 33, 38 },
		{ 489, 34, 38 },
		{ 489, 35, 38 },
		{ 489, 36, 38 },
		{ 489, 37, 38 },
		{ 489, 38, 38 },
		{ 489, 39, 38 },
		{ 489, 40, 38 },
		{ 489, 41, 38 },
		{ 489, 42, 38 },
		{ 489, 43, 38 },
		{ 489, 44, 38 },
		{ 489, 45, 38 },
		{ 489, 46, 38 },
		{ 489, 47, 38 },
		{ 489, 48, 38 },
		{ 489, 49, 38 },
		{ 489, 50, 38 },
		{ 489, 51, 38 },
		{ 489, 52, 38 },
		{ 489, 53, 38 },
		{ 490, 1, 38 },
		{ 490, 2, 38 },
		{ 490, 3, 38 },
		{ 490, 4, 38 },
		{ 490, 5, 38 },
		{ 490, 6, 38 },
		{ 490, 7, 38 },
		{ 490, 8, 38 },
		{ 490, 9, 38 },
		{ 490, 10, 38 },
		{ 490, 11, 38 },
		{ 490, 12, 38 },
		{ 490, 13, 38 },
		{ 490, 14, 38 },
		{ 490, 15, 38 },
		{ 490, 16, 38 },
		{ 490, 17, 38 },
		{ 490, 18, 38 },
		{ 490, 19, 38 },
		{ 490, 20, 38 },
		{ 490, 21, 38 },
		{ 490, 22, 38 },
		{ 490, 23, 38 },
		{ 490, 24, 38 },
		{ 490, 25, 38 },
		{ 490, 26, 38 },
		{ 490, 27, 38 },
		{ 490, 28, 38 },
		{ 490, 29, 38 },
		{ 490, 30, 38 },
		{ 490, 31, 38 },
		{ 490, 32, 38 },
		{ 490, 33, 38 },
		{ 490, 34, 38 },
		{ 490, 35, 38 },
		{ 490, 36, 38 },
		{ 490, 37, 38 },
		{ 490, 38, 38 },
		{ 490, 39, 38 },
		{ 490, 40, 38 },
		{ 490, 41, 38 },
		{ 490, 42, 38 },
		{ 490, 43, 38 },
		{ 490, 44, 38 },
		{ 490, 45, 38 },
		{ 490, 46, 38 },
		{ 490, 47, 38 },
		{ 490, 48, 38 },
		{ 490, 49, 38 },
		{ 490, 50, 38 },
		{ 490, 51, 38 },
		{ 490, 52, 38 },
		{ 490, 53, 38 },
		{ 491, 1, 38 },
		{ 491, 2, 38 },
		{ 491, 3, 38 },
		{ 491, 4, 38 },
		{ 491, 5, 38 },
		{ 491, 6, 38 },
		{ 491, 7, 38 },
		{ 491, 8, 38 },
		{ 491, 9, 38 },
		{ 491, 10, 38 },
		{ 491, 11, 38 },
		{ 491, 12, 38 },
		{ 491, 13, 38 },
		{ 491, 14, 38 },
		{ 491, 15, 38 },
		{ 491, 16, 38 },
		{ 491, 17, 38 },
		{ 491, 18, 38 },
		{ 491, 19, 38 },
		{ 491, 20, 38 },
		{ 491, 21, 38 },
		{ 491, 22, 38 },
		{ 491, 23, 38 },
		{ 491, 24, 38 },
		{ 491, 25, 38 },
		{ 491, 26, 38 },
		{ 491, 27, 38 },
		{ 491, 28, 38 },
		{ 491, 29, 38 },
		{ 491, 30, 38 },
		{ 491, 31, 38 },
		{ 491, 32, 38 },
		{ 491, 33, 38 },
		{ 491, 34, 38 },
		{ 491, 35, 38 },
		{ 491, 36, 38 },
		{ 491, 37, 38 },
		{ 491, 38, 38 },
		{ 491, 39, 38 },
		{ 491, 40, 38 },
		{ 491, 41, 38 },
		{ 491, 42, 38 },
		{ 491, 43, 38 },
		{ 491, 44, 38 },
		{ 491, 45, 38 },
		{ 491, 46, 38 },
		{ 491, 47, 38 },
		{ 491, 48, 38 },
		{ 491, 49, 38 },
		{ 491, 50, 38 },
		{ 491, 51, 38 },
		{ 491, 52, 38 },
		{ 492, 1, 38 },
		{ 492, 2, 38 },
		{ 492, 3, 38 },
		{ 492, 4, 38 },
		{ 492, 5, 38 },
		{ 492, 6, 38 },
		{ 492, 7, 38 },
		{ 492, 8, 38 },
		{ 492, 9, 38 },
		{ 492, 10, 38 },
		{ 492, 11, 38 },
		{ 492, 12, 38 },
		{ 492, 13, 38 },
		{ 492, 14, 38 },
		{ 492, 15, 38 },
		{ 492, 16, 38 },
		{ 492, 17, 38 },
		{ 492, 18, 38 },
		{ 492, 19, 38 },
		{ 492, 20, 38 },
		{ 492, 21, 38 },
		{ 492, 22, 38 },
		{ 492, 23, 38 },
		{ 492, 24, 38 },
		{ 492, 25, 38 },
		{ 492, 26, 38 },
		{ 492, 27, 38 },
		{ 492, 28, 38 },
		{ 492, 29, 38 },
		{ 492, 30, 38 },
		{ 492, 31, 38 },
		{ 492, 32, 38 },
		{ 492, 33, 38 },
		{ 492, 34, 38 },
		{ 492, 35, 38 },
		{ 492, 36, 38 },
		{ 492, 37, 38 },
		{ 492, 38, 38 },
		{ 492, 39, 38 },
		{ 492, 40, 38 },
		{ 492, 41, 38 },
		{ 492, 42, 38 },
		{ 492, 43, 38 },
		{ 492, 44, 38 },
		{ 492, 45, 38 },
		{ 492, 46, 38 },
		{ 492, 47, 38 },
		{ 492, 48, 38 },
		{ 492, 49, 38 },
		{ 492, 50, 38 },
		{ 492, 51, 38 },
		{ 493, 1, 38 },
		{ 493, 2, 38 },
		{ 493, 3, 38 },
		{ 493, 4, 38 },
		{ 493, 5, 38 },
		{ 493, 6, 38 },
		{ 493, 7, 38 },
		{ 493, 8, 38 },
		{ 493, 9, 38 },
		{ 493, 10, 38 },
		{ 493, 11, 38 },
		{ 493, 12, 38 },
		{ 493, 13, 38 },
		{ 493, 14, 38 },
		{ 493, 15, 38 },
		{ 493, 16, 38 },
		{ 493, 17, 38 },
		{ 493, 18, 38 },
		{ 493, 19, 38 },
		{ 493, 20, 38 },
		{ 493, 21, 38 },
		{ 493, 22, 38 },
		{ 493, 23, 38 },
		{ 493, 24, 38 },
		{ 493, 25, 38 },
		{ 493, 26, 38 },
		{ 493, 27, 38 },
		{ 493, 28, 38 },
		{ 493, 29, 38 },
		{ 493, 30, 38 },
		{ 493, 31, 38 },
		{ 493, 32, 38 },
		{ 493, 33, 38 },
		{ 493, 34, 38 },
		{ 493, 35, 38 },
		{ 493, 36, 38 },
		{ 493, 37, 38 },
		{ 493, 38, 38 },
		{ 493, 39, 38 },
		{ 493, 40, 38 },
		{ 493, 41, 38 },
		{ 493, 42, 38 },
		{ 493, 43, 38 },
		{ 493, 44, 38 },
		{ 493, 45, 38 },
		{ 493, 46, 38 },
		{ 493, 47, 38 },
		{ 493, 48, 38 },
		{ 493, 49, 38 },
		{ 493, 50, 38 },
		{ 494, 1, 38 },
		{ 494, 2, 38 },
		{ 494, 3, 38 },
		{ 494, 4, 38 },
		{ 494, 5, 38 },
		{ 494, 6, 38 },
		{ 494, 7, 38 },
		{ 494, 8, 38 },
		{ 494, 9, 38 },
		{ 494, 10, 38 },
		{ 494, 11, 38 },
		{ 494, 12, 38 },
		{ 494, 13, 38 },
		{ 494, 14, 38 },
		{ 494, 15, 38 },
		{ 494, 16, 38 },
		{ 494, 17, 38 },
		{ 494, 18, 38 },
		{ 494, 19, 38 },
		{ 494, 20, 38 },
		{ 494, 21, 38 },
		{ 494, 22, 38 },
		{ 494, 23, 38 },
		{ 494, 24, 38 },
		{ 494, 25, 38 },
		{ 494, 26, 38 },
		{ 494, 27, 38 },
		{ 494, 28, 38 },
		{ 494, 29, 38 },
		{ 494, 30, 38 },
		{ 494, 31, 38 },
		{ 494, 32, 38 },
		{ 494, 33, 38 },
		{ 494, 34, 38 },
		{ 494, 35, 38 },
		{ 494, 36, 38 },
		{ 494, 37, 38 },
		{ 494, 38, 38 },
		{ 494, 39, 38 },
		{ 494, 40, 38 },
		{ 494, 41, 38 },
		{ 494, 42, 38 },
		{ 494, 43, 38 },
		{ 494, 44, 38 },
		{ 494, 45, 38 },
		{ 494, 46, 38 },
		{ 494, 47, 38 },
		{ 494, 48, 38 },
		{ 494, 49, 38 },
		{ 495, 1, 38 },
		{ 495, 2, 38 },
		{ 495, 3, 38 },
		{ 495, 4, 38 },
		{ 495, 5, 38 },
		{ 495, 6, 38 },
		{ 495, 7, 38 },
		{ 495, 8, 38 },
		{ 495, 9, 38 },
		{ 495, 10, 38 },
		{ 495, 11, 38 },
		{ 495, 12, 38 },
		{ 495, 13, 38 },
		{ 495, 14, 38 },
		{ 495, 15, 38 },
		{ 495, 16, 38 },
		{ 495, 17, 38 },
		{ 495, 18, 38 },
		{ 495, 19, 38 },
		{ 495, 20, 38 },
		{ 495, 21, 38 },
		{ 495, 22, 38 },
		{ 495, 23, 38 },
		{ 495, 24, 38 },
		{ 495, 25, 38 },
		{ 495, 26, 38 },
		{ 495, 27, 38 },
		{ 495, 28, 38 },
		{ 495, 29, 38 },
		{ 495, 30, 38 },
		{ 495, 31, 38 },
		{ 495, 32, 38 },
		{ 495, 33, 38 },
		{ 495, 34, 38 },
		{ 495, 35, 38 },
		{ 495, 36, 38 },
		{ 495, 37, 38 },
		{ 495, 38, 38 },
		{ 495, 39, 38 },
		{ 495, 40, 38 },
		{ 495, 41, 38 },
		{ 495, 42, 38 },
		{ 495, 43, 38 },
		{ 495, 44, 38 },
		{ 495, 45, 38 },
		{ 495, 46, 38 },
		{ 495, 47, 38 },
		{ 495, 48, 38 },
		{ 496, 1, 38 },
		{ 496, 2, 38 },
		{ 496, 3, 38 },
		{ 496, 4, 38 },
		{ 496, 5, 38 },
		{ 496, 6, 38 },
		{ 496, 7, 38 },
		{ 496, 8, 38 },
		{ 496, 9, 38 },
		{ 496, 10, 38 },
		{ 496, 11, 38 },
		{ 496, 12, 38 },
		{ 496, 13, 38 },
		{ 496, 14, 38 },
		{ 496, 15, 38 },
		{ 496, 16, 38 },
		{ 496, 17, 38 },
		{ 496, 18, 38 },
		{ 496, 19, 38 },
		{ 496, 20, 38 },
		{ 496, 21, 38 },
		{ 496, 22, 38 },
		{ 496, 23, 38 },
		{ 496, 24, 38 },
		{ 496, 25, 38 },
		{ 496, 26, 38 },
		{ 496, 27, 38 },
		{ 496, 28, 38 },
		{ 496, 29, 38 },
		{ 496, 30, 38 },
		{ 496, 31, 38 },
		{ 496, 32, 38 },
		{ 496, 33, 38 },
		{ 496, 34, 38 },
		{ 496, 35, 38 },
		{ 496, 36, 38 },
		{ 496, 37, 38 },
		{ 496, 38, 38 },
		{ 496, 39, 38 },
		{ 496, 40, 38 },
		{ 496, 41, 38 },
		{ 496, 42, 38 },
		{ 496, 43, 38 },
		{ 496, 44, 38 },
		{ 496, 45, 38 },
		{ 496, 46, 38 },
		{ 496, 47, 38 },
		{ 497, 1, 38 },
		{ 497, 2, 38 },
		{ 497, 3, 38 },
		{ 497, 4, 38 },
		{ 497, 5, 38 },
		{ 497, 6, 38 },
		{ 497, 7, 38 },
		{ 497, 8, 38 },
		{ 497, 9, 38 },
		{ 497, 10, 38 },
		{ 497, 11, 38 },
		{ 497, 12, 38 },
		{ 497, 13, 38 },
		{ 497, 14, 38 },
		{ 497, 15, 38 },
		{ 497, 16, 38 },
		{ 497, 17, 38 },
		{ 497, 18, 38 },
		{ 497, 19, 38 },
		{ 497, 20, 38 },
		{ 497, 21, 38 },
		{ 497, 22, 38 },
		{ 497, 23, 38 },
		{ 497, 24, 38 },
		{ 497, 25, 38 },
		{ 497, 26, 38 },
		{ 497, 27, 38 },
		{ 497, 28, 38 },
		{ 497, 29, 38 },
		{ 497, 30, 38 },
		{ 497, 31, 38 },
		{ 497, 32, 38 },
		{ 497, 33, 38 },
		{ 497, 34, 38 },
		{ 497, 35, 38 },
		{ 497, 36, 38 },
		{ 497, 37, 38 },
		{ 497, 38, 38 },
		{ 497, 39, 38 },
		{ 497, 40, 38 },
		{ 497, 41, 38 },
		{ 497, 42, 38 },
		{ 497, 43, 38 },
		{ 497, 44, 38 },
		{ 497, 45, 38 },
		{ 497, 46, 38 },
		{ 498, 1, 38 },
		{ 498, 2, 38 },
		{ 498, 3, 38 },
		{ 498, 4, 38 },
		{ 498, 5, 38 },
		{ 498, 6, 38 },
		{ 498, 7, 38 },
		{ 498, 8, 38 },
		{ 498, 9, 38 },
		{ 498, 10, 38 },
		{ 498, 11, 38 },
		{ 498, 12, 38 },
		{ 498, 13, 38 },
		{ 498, 14, 38 },
		{ 498, 15, 38 },
		{ 498, 16, 38 },
		{ 498, 17, 38 },
		{ 498, 18, 38 },
		{ 498, 19, 38 },
		{ 498, 20, 38 },
		{ 498, 21, 38 },
		{ 498, 22, 38 },
		{ 498, 23, 38 },
		{ 498, 24, 38 },
		{ 498, 25, 38 },
		{ 498, 26, 38 },
		{ 498, 27, 38 },
		{ 498, 28, 38 },
		{ 498, 29, 38 },
		{ 498, 30, 38 },
		{ 498, 31, 38 },
		{ 498, 32, 38 },
		{ 498, 33, 38 },
		{ 498, 34, 38 },
		{ 498, 35, 38 },
		{ 498, 36, 38 },
		{ 498, 37, 38 },
		{ 498, 38, 38 },
		{ 498, 39, 38 },
		{ 498, 40, 38 },
		{ 498, 41, 38 },
		{ 498, 42, 38 },
		{ 498, 43, 38 },
		{ 498, 44, 38 },
		{ 498, 45, 38 },
		{ 498, 46, 38 },
		{ 499, 1, 38 },
		{ 499, 2, 38 },
		{ 499, 3, 38 },
		{ 499, 4, 38 },
		{ 499, 5, 38 },
		{ 499, 6, 38 },
		{ 499, 7, 38 },
		{ 499, 8, 38 },
		{ 499, 9, 38 },
		{ 499, 10, 38 },
		{ 499, 11, 38 },
		{ 499, 12, 38 },
		{ 499, 13, 38 },
		{ 499, 14, 38 },
		{ 499, 15, 38 },
		{ 499, 16, 38 },
		{ 499, 17, 38 },
		{ 499, 18, 38 },
		{ 499, 19, 38 },
		{ 499, 20, 38 },
		{ 499, 21, 38 },
		{ 499, 22, 38 },
		{ 499, 23, 38 },
		{ 499, 24, 38 },
		{ 499, 25, 38 },
		{ 499, 26, 38 },
		{ 499, 27, 38 },
		{ 499, 28, 38 },
		{ 499, 29, 38 },
		{ 499, 30, 38 },
		{ 499, 31, 38 },
		{ 499, 32, 38 },
		{ 499, 33, 38 },
		{ 499, 34, 38 },
		{ 499, 35, 38 },
		{ 499, 36, 38 },
		{ 499, 37, 38 },
		{ 499, 38, 38 },
		{ 499, 39, 38 },
		{ 499, 40, 38 },
		{ 499, 41, 38 },
		{ 499, 42, 38 },
		{ 499, 43, 38 },
		{ 499, 44, 38 },
		{ 499, 45, 38 },
		{ 500, 1, 38 },
		{ 500, 2, 38 },
		{ 500, 3, 38 },
		{ 500, 4, 38 },
		{ 500, 5, 38 },
		{ 500, 6, 38 },
		{ 500, 7, 38 },
		{ 500, 8, 38 },
		{ 500, 9, 38 },
		{ 500, 10, 38 },
		{ 500, 11, 38 },
		{ 500, 12, 38 },
		{ 500, 13, 38 },
		{ 500, 14, 38 },
		{ 500, 15, 38 },
		{ 500, 16, 38 },
		{ 500, 17, 38 },
		{ 500, 18, 38 },
		{ 500, 19, 38 },
		{ 500, 20, 38 },
		{ 500, 21, 38 },
		{ 500, 22, 38 },
		{ 500, 23, 38 },
		{ 500, 24, 38 },
		{ 500, 25, 38 },
		{ 500, 26, 38 },
		{ 500, 27, 38 },
		{ 500, 28, 38 },
		{ 500, 29, 38 },
		{ 500, 30, 38 },
		{ 500, 31, 38 },
		{ 500, 32, 38 },
		{ 500, 33, 38 },
		{ 500, 34, 38 },
		{ 500, 35, 38 },
		{ 500, 36, 38 },
		{ 500, 37, 38 },
		{ 500, 38, 38 },
		{ 500, 39, 38 },
		{ 500, 40, 38 },
		{ 500, 41, 38 },
		{ 500, 42, 38 },
		{ 500, 43, 38 },
		{ 500, 44, 38 },
		{ 501, 1, 38 },
		{ 501, 2, 38 },
		{ 501, 3, 38 },
		{ 501, 4, 38 },
		{ 501, 5, 38 },
		{ 501, 6, 38 },
		{ 501, 7, 38 },
		{ 501, 8, 38 },
		{ 501, 9, 38 },
		{ 501, 10, 38 },
		{ 501, 11, 38 },
		{ 501, 12, 38 },
		{ 501, 13, 38 },
		{ 501, 14, 38 },
		{ 501, 15, 38 },
		{ 501, 16, 38 },
		{ 501, 17, 38 },
		{ 501, 18, 38 },
		{ 501, 19, 38 },
		{ 501, 20, 38 },
		{ 501, 21, 38 },
		{ 501, 22, 38 },
		{ 501, 23, 38 },
		{ 501, 24, 38 },
		{ 501, 25, 38 },
		{ 501, 26, 38 },
		{ 501, 27, 38 },
		{ 501, 28, 38 },
		{ 501, 29, 38 },
		{ 501, 30, 38 },
		{ 501, 31, 38 },
		{ 501, 32, 38 },
		{ 501, 33, 38 },
		{ 501, 34, 38 },
		{ 501, 35, 38 },
		{ 501, 36, 38 },
		{ 501, 37, 38 },
		{ 501, 38, 38 },
		{ 501, 39, 38 },
		{ 501, 40, 38 },
		{ 501, 41, 38 },
		{ 501, 42, 38 },
		{ 501, 43, 38 },
		{ 502, 1, 38 },
		{ 502, 2, 38 },
		{ 502, 3, 38 },
		{ 502, 4, 38 },
		{ 502, 5, 38 },
		{ 502, 6, 38 },
		{ 502, 7, 38 },
		{ 502, 8, 38 },
		{ 502, 9, 38 },
		{ 502, 10, 38 },
		{ 502, 11, 38 },
		{ 502, 12, 38 },
		{ 502, 13, 38 },
		{ 502, 14, 38 },
		{ 502, 15, 38 },
		{ 502, 16, 38 },
		{ 502, 17, 38 },
		{ 502, 18, 38 },
		{ 502, 19, 38 },
		{ 502, 20, 38 },
		{ 502, 21, 38 },
		{ 502, 22, 38 },
		{ 502, 23, 38 },
		{ 502, 24, 38 },
		{ 502, 25, 38 },
		{ 502, 26, 38 },
		{ 502, 27, 38 },
		{ 502, 28, 38 },
		{ 502, 29, 38 },
		{ 502, 30, 38 },
		{ 502, 31, 38 },
		{ 502, 32, 38 },
		{ 502, 33, 38 },
		{ 502, 34, 38 },
		{ 502, 35, 38 },
		{ 502, 36, 38 },
		{ 502, 37, 38 },
		{ 502, 38, 38 },
		{ 502, 39, 38 },
		{ 502, 40, 38 },
		{ 502, 41, 38 },
		{ 502, 42, 38 },
		{ 503, 1, 38 },
		{ 503, 2, 38 },
		{ 503, 3, 38 },
		{ 503, 4, 38 },
		{ 503, 5, 38 },
		{ 503, 6, 38 },
		{ 503, 7, 38 },
		{ 503, 8, 38 },
		{ 503, 9, 38 },
		{ 503, 10, 38 },
		{ 503, 11, 38 },
		{ 503, 12, 38 },
		{ 503, 13, 38 },
		{ 503, 14, 38 },
		{ 503, 15, 38 },
		{ 503, 16, 38 },
		{ 503, 17, 38 },
		{ 503, 18, 38 },
		{ 503, 19, 38 },
		{ 503, 20, 38 },
		{ 503, 21, 38 },
		{ 503, 22, 38 },
		{ 503, 23, 38 },
		{ 503, 24, 38 },
		{ 503, 25, 38 },
		{ 503, 26, 38 },
		{ 503, 27, 38 },
		{ 503, 28, 38 },
		{ 503, 29, 38 },
		{ 503, 30, 38 },
		{ 503, 31, 38 },
		{ 503, 32, 38 },
		{ 503, 33, 38 },
		{ 503, 34, 38 },
		{ 503, 35, 38 },
		{ 503, 36, 38 },
		{ 503, 37, 38 },
		{ 503, 38, 38 },
		{ 503, 39, 38 },
		{ 503, 40, 38 },
		{ 503, 41, 38 },
		{ 504, 1, 38 },
		{ 504, 2, 38 },
		{ 504, 3, 38 },
		{ 504, 4, 38 },
		{ 504, 5, 38 },
		{ 504, 6, 38 },
		{ 504, 7, 38 },
		{ 504, 8, 38 },
		{ 504, 9, 38 },
		{ 504, 10, 38 },
		{ 504, 11, 38 },
		{ 504, 12, 38 },
		{ 504, 13, 38 },
		{ 504, 14, 38 },
		{ 504, 15, 38 },
		{ 504, 16, 38 },
		{ 504, 17, 38 },
		{ 504, 18, 38 },
		{ 504, 19, 38 },
		{ 504, 20, 38 },
		{ 504, 21, 38 },
		{ 504, 22, 38 },
		{ 504, 23, 38 },
		{ 504, 24, 38 },
		{ 504, 25, 38 },
		{ 504, 26, 38 },
		{ 504, 27, 38 },
		{ 504, 28, 38 },
		{ 504, 29, 38 },
		{ 504, 30, 38 },
		{ 504, 31, 38 },
		{ 504, 32, 38 },
		{ 504, 33, 38 },
		{ 504, 34, 38 },
		{ 504, 35, 38 },
		{ 504, 36, 38 },
		{ 504, 37, 38 },
		{ 504, 38, 38 },
		{ 504, 39, 38 },
		{ 504, 40, 38 },
		{ 505, 1, 38 },
		{ 505, 2, 38 },
		{ 505, 3, 38 },
		{ 505, 4, 38 },
		{ 505, 5, 38 },
		{ 505, 6, 38 },
		{ 505, 7, 38 },
		{ 505, 8, 38 },
		{ 505, 9, 38 },
		{ 505, 10, 38 },
		{ 505, 11, 38 },
		{ 505, 12, 38 },
		{ 505, 13, 38 },
		{ 505, 14, 38 },
		{ 505, 15, 38 },
		{ 505, 16, 38 },
		{ 505, 17, 38 },
		{ 505, 18, 38 },
		{ 505, 19, 38 },
		{ 505, 20, 38 },
		{ 505, 21, 38 },
		{ 505, 22, 38 },
		{ 505, 23, 38 },
		{ 505, 24, 38 },
		{ 505, 25, 38 },
		{ 505, 26, 38 },
		{ 505, 27, 38 },
		{ 505, 28, 38 },
		{ 505, 29, 38 },
		{ 505, 30, 38 },
		{ 505, 31, 38 },
		{ 505, 32, 38 },
		{ 505, 33, 38 },
		{ 505, 34, 38 },
		{ 505, 35, 38 },
		{ 505, 36, 38 },
		{ 505, 37, 38 },
		{ 505, 38, 38 },
		{ 505, 39, 38 },
		{ 506, 1, 38 },
		{ 506, 2, 38 },
		{ 506, 3, 38 },
		{ 506, 4, 38 },
		{ 506, 5, 38 },
		{ 506, 6, 38 },
		{ 506, 7, 38 },
		{ 506, 8, 38 },
		{ 506, 9, 38 },
		{ 506, 10, 38 },
		{ 506, 11, 38 },
		{ 506, 12, 38 },
		{ 506, 13, 38 },
		{ 506, 14, 38 },
		{ 506, 15, 38 },
		{ 506, 16, 38 },
		{ 506, 17, 38 },
		{ 506, 18, 38 },
		{ 506, 19, 38 },
		{ 506, 20, 38 },
		{ 506, 21, 38 },
		{ 506, 22, 38 },
		{ 506, 23, 38 },
		{ 506, 24, 38 },
		{ 506, 25, 38 },
		{ 506, 26, 38 },
		{ 506, 27, 38 },
		{ 506, 28, 38 },
		{ 506, 29, 38 },
		{ 506, 30, 38 },
		{ 506, 31, 38 },
		{ 506, 32, 38 },
		{ 506, 33, 38 },
		{ 506, 34, 38 },
		{ 506, 35, 38 },
		{ 506, 36, 38 },
		{ 506, 37, 38 },
		{ 506, 38, 38 },
		{ 506, 39, 38 },
		{ 507, 1, 38 },
		{ 507, 2, 38 },
		{ 507, 3, 38 },
		{ 507, 4, 38 },
		{ 507, 5, 38 },
		{ 507, 6, 38 },
		{ 507, 7, 38 },
		{ 507, 8, 38 },
		{ 507, 9, 38 },
		{ 507, 10, 38 },
		{ 507, 11, 38 },
		{ 507, 12, 38 },
		{ 507, 13, 38 },
		{ 507, 14, 38 },
		{ 507, 15, 38 },
		{ 507, 16, 38 },
		{ 507, 17, 38 },
		{ 507, 18, 38 },
		{ 507, 19, 38 },
		{ 507, 20, 38 },
		{ 507, 21, 38 },
		{ 507, 22, 38 },
		{ 507, 23, 38 },
		{ 507, 24, 38 },
		{ 507, 25, 38 },
		{ 507, 26, 38 },
		{ 507, 27, 38 },
		{ 507, 28, 38 },
		{ 507, 29, 38 },
		{ 507, 30, 38 },
		{ 507, 31, 38 },
		{ 507, 32, 38 },
		{ 507, 33, 38 },
		{ 507, 34, 38 },
		{ 507, 35, 38 },
		{ 507, 36, 38 },
		{ 507, 37, 38 },
		{ 507, 38, 38 },
		{ 508, 1, 38 },
		{ 508, 2, 38 },
		{ 508, 3, 38 },
		{ 508, 4, 38 },
		{ 508, 5, 38 },
		{ 508, 6, 38 },
		{ 508, 7, 38 },
		{ 508, 8, 38 },
		{ 508, 9, 38 },
		{ 508, 10, 38 },
		{ 508, 11, 38 },
		{ 508, 12, 38 },
		{ 508, 13, 38 },
		{ 508, 14, 38 },
		{ 508, 15, 38 },
		{ 508, 16, 38 },
		{ 508, 17, 38 },
		{ 508, 18, 38 },
		{ 508, 19, 38 },
		{ 508, 20, 38 },
		{ 508, 21, 38 },
		{ 508, 22, 38 },
		{ 508, 23, 38 },
		{ 508, 24, 38 },
		{ 508, 25, 38 },
		{ 508, 26, 38 },
		{ 508, 27, 38 },
		{ 508, 28, 38 },
		{ 508, 29, 38 },
		{ 508, 30, 38 },
		{ 508, 31, 38 },
		{ 508, 32, 38 },
		{ 508, 33, 38 },
		{ 508, 34, 38 },
		{ 508, 35, 38 },
		{ 508, 36, 38 },
		{ 508, 37, 38 },
		{ 509, 1, 38 },
		{ 509, 2, 38 },
		{ 509, 3, 38 },
		{ 509, 4, 38 },
		{ 509, 5, 38 },
		{ 509, 6, 38 },
		{ 509, 7, 38 },
		{ 509, 8, 38 },
		{ 509, 9, 38 },
		{ 509, 10, 38 },
		{ 509, 11, 38 },
		{ 509, 12, 38 },
		{ 509, 13, 38 },
		{ 509, 14, 38 },
		{ 509, 15, 38 },
		{ 509, 16, 38 },
		{ 509, 17, 38 },
		{ 509, 18, 38 },
		{ 509, 19, 38 },
		{ 509, 20, 38 },
		{ 509, 21, 38 },
		{ 509, 22, 38 },
		{ 509, 23, 38 },
		{ 509, 24, 38 },
		{ 509, 25, 38 },
		{ 509, 26, 38 },
		{ 509, 27, 38 },
		{ 509, 28, 38 },
		{ 509, 29, 38 },
		{ 509, 30, 38 },
		{ 509, 31, 38 },
		{ 509, 32, 38 },
		{ 509, 33, 38 },
		{ 509, 34, 38 },
		{ 509, 35, 38 },
		{ 509, 36, 38 },
		{ 510, 1, 38 },
		{ 510, 2, 38 },
		{ 510, 3, 38 },
		{ 510, 4, 38 },
		{ 510, 5, 38 },
		{ 510, 6, 38 },
		{ 510, 7, 38 },
		{ 510, 8, 38 },
		{ 510, 9, 38 },
		{ 510, 10, 38 },
		{ 510, 11, 38 },
		{ 510, 12, 38 },
		{ 510, 13, 38 },
		{ 510, 14, 38 },
		{ 510, 15, 38 },
		{ 510, 16, 38 },
		{ 510, 17, 38 },
		{ 510, 18, 38 },
		{ 510, 19, 38 },
		{ 510, 20, 38 },
		{ 510, 21, 38 },
		{ 510, 22, 38 },
		{ 510, 23, 38 },
		{ 510, 24, 38 },
		{ 510, 25, 38 },
		{ 510, 26, 38 },
		{ 510, 27, 38 },
		{ 510, 28, 38 },
		{ 510, 29, 38 },
		{ 510, 30, 38 },
		{ 510, 31, 38 },
		{ 510, 32, 38 },
		{ 510, 33, 38 },
		{ 510, 34, 38 },
		{ 510, 35, 38 },
		{ 511, 1, 38 },
		{ 511, 2, 38 },
		{ 511, 3, 38 },
		{ 511, 4, 38 },
		{ 511, 5, 38 },
		{ 511, 6, 38 },
		{ 511, 7, 38 },
		{ 511, 8, 38 },
		{ 511, 9, 38 },
		{ 511, 10, 38 },
		{ 511, 11, 38 },
		{ 511, 12, 38 },
		{ 511, 13, 38 },
		{ 511, 14, 38 },
		{ 511, 15, 38 },
		{ 511, 16, 38 },
		{ 511, 17, 38 },
		{ 511, 18, 38 },
		{ 511, 19, 38 },
		{ 511, 20, 38 },
		{ 511, 21, 38 },
		{ 511, 22, 38 },
		{ 511, 23, 38 },
		{ 511, 24, 38 },
		{ 511, 25, 38 },
		{ 511, 26, 38 },
		{ 511, 27, 38 },
		{ 511, 28, 38 },
		{ 511, 29, 38 },
		{ 511, 30, 38 },
		{ 511, 31, 38 },
		{ 511, 32, 38 },
		{ 511, 33, 38 },
		{ 511, 34, 38 },
		{ 512, 1, 38 },
		{ 512, 2, 38 },
		{ 512, 3, 38 },
		{ 512, 4, 38 },
		{ 512, 5, 38 },
		{ 512, 6, 38 },
		{ 512, 7, 38 },
		{ 512, 8, 38 },
		{ 512, 9, 38 },
		{ 512, 10, 38 },
		{ 512, 11, 38 },
		{ 512, 12, 38 },
		{ 512, 13, 38 },
		{ 512, 14, 38 },
		{ 512, 15, 38 },
		{ 512, 16, 38 },
		{ 512, 17, 38 },
		{ 512, 18, 38 },
		{ 512, 19, 38 },
		{ 512, 20, 38 },
		{ 512, 21, 38 },
		{ 512, 22, 38 },
		{ 512, 23, 38 },
		{ 512, 24, 38 },
		{ 512, 25, 38 },
		{ 512, 26, 38 },
		{ 512, 27, 38 },
		{ 512, 28, 38 },
		{ 512, 29, 38 },
		{ 512, 30, 38 },
		{ 512, 31, 38 },
		{ 512, 32, 38 },
		{ 512, 33, 38 },
		{ 1, 450, 39 },
		{ 1, 451, 39 },
		{ 1, 452, 39 },
		{ 1, 453, 39 },
		{ 1, 454, 39 },
		{ 1, 455, 39 },
		{ 1, 456, 39 },
		{ 1, 457, 39 },
		{ 1, 458, 39 },
		{ 1, 459, 39 },
		{ 1, 460, 39 },
		{ 1, 461, 39 },
		{ 1, 462, 39 },
		{ 1, 463, 39 },
		{ 1, 464, 39 },
		{ 1, 465, 39 },
		{ 1, 466, 39 },
		{ 1, 467, 39 },
		{ 1, 468, 39 },
		{ 1, 469, 39 },
		{ 1, 470, 39 },
		{ 1, 471, 39 },
		{ 1, 472, 39 },
		{ 1, 473, 39 },
		{ 1, 474, 39 },
		{ 1, 475, 39 },
		{ 1, 476, 39 },
		{ 1, 477, 39 },
		{ 1, 478, 39 },
		{ 1, 479, 39 },
		{ 1, 480, 39 },
		{ 1, 481, 39 },
		{ 1, 482, 39 },
		{ 1, 483, 39 },
		{ 1, 484, 39 },
		{ 1, 485, 39 },
		{ 1, 486, 39 },
		{ 1, 487, 39 },
		{ 1, 488, 39 },
		{ 1, 489, 39 },
		{ 1, 490, 39 },
		{ 1, 491, 39 },
		{ 1, 492, 39 },
		{ 1, 493, 39 },
		{ 1, 494, 39 },
		{ 1, 495, 39 },
		{ 1, 496, 39 },
		{ 1, 497, 39 },
		{ 1, 498, 39 },
		{ 1, 499, 39 },
		{ 1, 500, 39 },
		{ 1, 501, 39 },
		{ 1, 502, 39 },
		{ 1, 503, 39 },
		{ 1, 504, 39 },
		{ 1, 505, 39 },
		{ 1, 506, 39 },
		{ 1, 507, 39 },
		{ 1, 508, 39 },
		{ 1, 509, 39 },
		{ 1, 510, 39 },
		{ 1, 511, 39 },
		{ 1, 512, 39 },
		{ 2, 449, 39 },
		{ 2, 450, 39 },
		{ 2, 451, 39 },
		{ 2, 452, 39 },
		{ 2, 453, 39 },
		{ 2, 454, 39 },
		{ 2, 455, 39 },
		{ 2, 456, 39 },
		{ 2, 457, 39 },
		{ 2, 458, 39 },
		{ 2, 459, 39 },
		{ 2, 460, 39 },
		{ 2, 461, 39 },
		{ 2, 462, 39 },
		{ 2, 463, 39 },
		{ 2, 464, 39 },
		{ 2, 465, 39 },
		{ 2, 466, 39 },
		{ 2, 467, 39 },
		{ 2, 468, 39 },
		{ 2, 469, 39 },
		{ 2, 470, 39 },
		{ 2, 471, 39 },
		{ 2, 472, 39 },
		{ 2, 473, 39 },
		{ 2, 474, 39 },
		{ 2, 475, 39 },
		{ 2, 476, 39 },
		{ 2, 477, 39 },
		{ 2, 478, 39 },
		{ 2, 479, 39 },
		{ 2, 480, 39 },
		{ 2, 481, 39 },
		{ 2, 482, 39 },
		{ 2, 483, 39 },
		{ 2, 484, 39 },
		{ 2, 485, 39 },
		{ 2, 486, 39 },
		{ 2, 487, 39 },
		{ 2, 488, 39 },
		{ 2, 489, 39 },
		{ 2, 490, 39 },
		{ 2, 491, 39 },
		{ 2, 492, 39 },
		{ 2, 493, 39 },
		{ 2, 494, 39 },
		{ 2, 495, 39 },
		{ 2, 496, 39 },
		{ 2, 497, 39 },
		{ 2, 498, 39 },
		{ 2, 499, 39 },
		{ 2, 500, 39 },
		{ 2, 501, 39 },
		{ 2, 502, 39 },
		{ 2, 503, 39 },
		{ 2, 504, 39 },
		{ 2, 505, 39 },
		{ 2, 506, 39 },
		{ 2, 507, 39 },
		{ 2, 508, 39 },
		{ 2, 509, 39 },
		{ 2, 510, 39 },
		{ 2, 511, 39 },
		{ 3, 448, 39 },
		{ 3, 449, 39 },
		{ 3, 450, 39 },
		{ 3, 451, 39 },
		{ 3, 452, 39 },
		{ 3, 453, 39 },
		{ 3, 454, 39 },
		{ 3, 455, 39 },
		{ 3, 456, 39 },
		{ 3, 457, 39 },
		{ 3, 458, 39 },
		{ 3, 459, 39 },
		{ 3, 460, 39 },
		{ 3, 461, 39 },
		{ 3, 462, 39 },
		{ 3, 463, 39 },
		{ 3, 464, 39 },
		{ 3, 465, 39 },
		{ 3, 466, 39 },
		{ 3, 467, 39 },
		{ 3, 468, 39 },
		{ 3, 469, 39 },
		{ 3, 470, 39 },
		{ 3, 471, 39 },
		{ 3, 472, 39 },
		{ 3, 473, 39 },
		{ 3, 474, 39 },
		{ 3, 475, 39 },
		{ 3, 476, 39 },
		{ 3, 477, 39 },
		{ 3, 478, 39 },
		{ 3, 479, 39 },
		{ 3, 480, 39 },
		{ 3, 481, 39 },
		{ 3, 482, 39 },
		{ 3, 483, 39 },
		{ 3, 484, 39 },
		{ 3, 485, 39 },
		{ 3, 486, 39 },
		{ 3, 487, 39 },
		{ 3, 488, 39 },
		{ 3, 489, 39 },
		{ 3, 490, 39 },
		{ 3, 491, 39 },
		{ 3, 492, 39 },
		{ 3, 493, 39 },
		{ 3, 494, 39 },
		{ 3, 495, 39 },
		{ 3, 496, 39 },
		{ 3, 497, 39 },
		{ 3, 498, 39 },
		{ 3, 499, 39 },
		{ 3, 500, 39 },
		{ 3, 501, 39 },
		{ 3, 502, 39 },
		{ 3, 503, 39 },
		{ 3, 504, 39 },
		{ 3, 505, 39 },
		{ 3, 506, 39 },
		{ 3, 507, 39 },
		{ 3, 508, 39 },
		{ 3, 509, 39 },
		{ 3, 510, 39 },
		{ 4, 447, 39 },
		{ 4, 448, 39 },
		{ 4, 449, 39 },
		{ 4, 450, 39 },
		{ 4, 451, 39 },
		{ 4, 452, 39 },
		{ 4, 453, 39 },
		{ 4, 454, 39 },
		{ 4, 455, 39 },
		{ 4, 456, 39 },
		{ 4, 457, 39 },
		{ 4, 458, 39 },
		{ 4, 459, 39 },
		{ 4, 460, 39 },
		{ 4, 461, 39 },
		{ 4, 462, 39 },
		{ 4, 463, 39 },
		{ 4, 464, 39 },
		{ 4, 465, 39 },
		{ 4, 466, 39 },
		{ 4, 467, 39 },
		{ 4, 468, 39 },
		{ 4, 469, 39 },
		{ 4, 470, 39 },
		{ 4, 471, 39 },
		{ 4, 472, 39 },
		{ 4, 473, 39 },
		{ 4, 474, 39 },
		{ 4, 475, 39 },
		{ 4, 476, 39 },
		{ 4, 477, 39 },
		{ 4, 478, 39 },
		{ 4, 479, 39 },
		{ 4, 480, 39 },
		{ 4, 481, 39 },
		{ 4, 482, 39 },
		{ 4, 483, 39 },
		{ 4, 484, 39 },
		{ 4, 485, 39 },
		{ 4, 486, 39 },
		{ 4, 487, 39 },
		{ 4, 488, 39 },
		{ 4, 489, 39 },
		{ 4, 490, 39 },
		{ 4, 491, 39 },
		{ 4, 492, 39 },
		{ 4, 493, 39 },
		{ 4, 494, 39 },
		{ 4, 495, 39 },
		{ 4, 496, 39 },
		{ 4, 497, 39 },
		{ 4, 498, 39 },
		{ 4, 499, 39 },
		{ 4, 500, 39 },
		{ 4, 501, 39 },
		{ 4, 502, 39 },
		{ 4, 503, 39 },
		{ 4, 504, 39 },
		{ 4, 505, 39 },
		{ 4, 506, 39 },
		{ 4, 507, 39 },
		{ 4, 508, 39 },
		{ 4, 509, 39 },
		{ 5, 447, 39 },
		{ 5, 448, 39 },
		{ 5, 449, 39 },
		{ 5, 450, 39 },
		{ 5, 451, 39 },
		{ 5, 452, 39 },
		{ 5, 453, 39 },
		{ 5, 454, 39 },
		{ 5, 455, 39 },
		{ 5, 456, 39 },
		{ 5, 457, 39 },
		{ 5, 458, 39 },
		{ 5, 459, 39 },
		{ 5, 460, 39 },
		{ 5, 461, 39 },
		{ 5, 462, 39 },
		{ 5, 463, 39 },
		{ 5, 464, 39 },
		{ 5, 465, 39 },
		{ 5, 466, 39 },
		{ 5, 467, 39 },
		{ 5, 468, 39 },
		{ 5, 469, 39 },
		{ 5, 470, 39 },
		{ 5, 471, 39 },
		{ 5, 472, 39 },
		{ 5, 473, 39 },
		{ 5, 474, 39 },
		{ 5, 475, 39 },
		{ 5, 476, 39 },
		{ 5, 477, 39 },
		{ 5, 478, 39 },
		{ 5, 479, 39 },
		{ 5, 480, 39 },
		{ 5, 481, 39 },
		{ 5, 482, 39 },
		{ 5, 483, 39 },
		{ 5, 484, 39 },
		{ 5, 485, 39 },
		{ 5, 486, 39 },
		{ 5, 487, 39 },
		{ 5, 488, 39 },
		{ 5, 489, 39 },
		{ 5, 490, 39 },
		{ 5, 491, 39 },
		{ 5, 492, 39 },
		{ 5, 493, 39 },
		{ 5, 494, 39 },
		{ 5, 495, 39 },
		{ 5, 496, 39 },
		{ 5, 497, 39 },
		{ 5, 498, 39 },
		{ 5, 499, 39 },
		{ 5, 500, 39 },
		{ 5, 501, 39 },
		{ 5, 502, 39 },
		{ 5, 503, 39 },
		{ 5, 504, 39 },
		{ 5, 505, 39 },
		{ 5, 506, 39 },
		{ 5, 507, 39 },
		{ 5, 508, 39 },
		{ 6, 446, 39 },
		{ 6, 447, 39 },
		{ 6, 448, 39 },
		{ 6, 449, 39 },
		{ 6, 450, 39 },
		{ 6, 451, 39 },
		{ 6, 452, 39 },
		{ 6, 453, 39 },
		{ 6, 454, 39 },
		{ 6, 455, 39 },
		{ 6, 456, 39 },
		{ 6, 457, 39 },
		{ 6, 458, 39 },
		{ 6, 459, 39 },
		{ 6, 460, 39 },
		{ 6, 461, 39 },
		{ 6, 462, 39 },
		{ 6, 463, 39 },
		{ 6, 464, 39 },
		{ 6, 465, 39 },
		{ 6, 466, 39 },
		{ 6, 467, 39 },
		{ 6, 468, 39 },
		{ 6, 469, 39 },
		{ 6, 470, 39 },
		{ 6, 471, 39 },
		{ 6, 472, 39 },
		{ 6, 473, 39 },
		{ 6, 474, 39 },
		{ 6, 475, 39 },
		{ 6, 476, 39 },
		{ 6, 477, 39 },
		{ 6, 478, 39 },
		{ 6, 479, 39 },
		{ 6, 480, 39 },
		{ 6, 481, 39 },
		{ 6, 482, 39 },
		{ 6, 483, 39 },
		{ 6, 484, 39 },
		{ 6, 485, 39 },
		{ 6, 486, 39 },
		{ 6, 487, 39 },
		{ 6, 488, 39 },
		{ 6, 489, 39 },
		{ 6, 490, 39 },
		{ 6, 491, 39 },
		{ 6, 492, 39 },
		{ 6, 493, 39 },
		{ 6, 494, 39 },
		{ 6, 495, 39 },
		{ 6, 496, 39 },
		{ 6, 497, 39 },
		{ 6, 498, 39 },
		{ 6, 499, 39 },
		{ 6, 500, 39 },
		{ 6, 501, 39 },
		{ 6, 502, 39 },
		{ 6, 503, 39 },
		{ 6, 504, 39 },
		{ 6, 505, 39 },
		{ 6, 506, 39 },
		{ 6, 507, 39 },
		{ 7, 445, 39 },
		{ 7, 446, 39 },
		{ 7, 447, 39 },
		{ 7, 448, 39 },
		{ 7, 449, 39 },
		{ 7, 450, 39 },
		{ 7, 451, 39 },
		{ 7, 452, 39 },
		{ 7, 453, 39 },
		{ 7, 454, 39 },
		{ 7, 455, 39 },
		{ 7, 456, 39 },
		{ 7, 457, 39 },
		{ 7, 458, 39 },
		{ 7, 459, 39 },
		{ 7, 460, 39 },
		{ 7, 461, 39 },
		{ 7, 462, 39 },
		{ 7, 463, 39 },
		{ 7, 464, 39 },
		{ 7, 465, 39 },
		{ 7, 466, 39 },
		{ 7, 467, 39 },
		{ 7, 468, 39 },
		{ 7, 469, 39 },
		{ 7, 470, 39 },
		{ 7, 471, 39 },
		{ 7, 472, 39 },
		{ 7, 473, 39 },
		{ 7, 474, 39 },
		{ 7, 475, 39 },
		{ 7, 476, 39 },
		{ 7, 477, 39 },
		{ 7, 478, 39 },
		{ 7, 479, 39 },
		{ 7, 480, 39 },
		{ 7, 481, 39 },
		{ 7, 482, 39 },
		{ 7, 483, 39 },
		{ 7, 484, 39 },
		{ 7, 485, 39 },
		{ 7, 486, 39 },
		{ 7, 487, 39 },
		{ 7, 488, 39 },
		{ 7, 489, 39 },
		{ 7, 490, 39 },
		{ 7, 491, 39 },
		{ 7, 492, 39 },
		{ 7, 493, 39 },
		{ 7, 494, 39 },
		{ 7, 495, 39 },
		{ 7, 496, 39 },
		{ 7, 497, 39 },
		{ 7, 498, 39 },
		{ 7, 499, 39 },
		{ 7, 500, 39 },
		{ 7, 501, 39 },
		{ 7, 502, 39 },
		{ 7, 503, 39 },
		{ 7, 504, 39 },
		{ 7, 505, 39 },
		{ 7, 506, 39 },
		{ 8, 444, 39 },
		{ 8, 445, 39 },
		{ 8, 446, 39 },
		{ 8, 447, 39 },
		{ 8, 448, 39 },
		{ 8, 449, 39 },
		{ 8, 450, 39 },
		{ 8, 451, 39 },
		{ 8, 452, 39 },
		{ 8, 453, 39 },
		{ 8, 454, 39 },
		{ 8, 455, 39 },
		{ 8, 456, 39 },
		{ 8, 457, 39 },
		{ 8, 458, 39 },
		{ 8, 459, 39 },
		{ 8, 460, 39 },
		{ 8, 461, 39 },
		{ 8, 462, 39 },
		{ 8, 463, 39 },
		{ 8, 464, 39 },
		{ 8, 465, 39 },
		{ 8, 466, 39 },
		{ 8, 467, 39 },
		{ 8, 468, 39 },
		{ 8, 469, 39 },
		{ 8, 470, 39 },
		{ 8, 471, 39 },
		{ 8, 472, 39 },
		{ 8, 473, 39 },
		{ 8, 474, 39 },
		{ 8, 475, 39 },
		{ 8, 476, 39 },
		{ 8, 477, 39 },
		{ 8, 478, 39 },
		{ 8, 479, 39 },
		{ 8, 480, 39 },
		{ 8, 481, 39 },
		{ 8, 482, 39 },
		{ 8, 483, 39 },
		{ 8, 484, 39 },
		{ 8, 485, 39 },
		{ 8, 486, 39 },
		{ 8, 487, 39 },
		{ 8, 488, 39 },
		{ 8, 489, 39 },
		{ 8, 490, 39 },
		{ 8, 491, 39 },
		{ 8, 492, 39 },
		{ 8, 493, 39 },
		{ 8, 494, 39 },
		{ 8, 495, 39 },
		{ 8, 496, 39 },
		{ 8, 497, 39 },
		{ 8, 498, 39 },
		{ 8, 499, 39 },
		{ 8, 500, 39 },
		{ 8, 501, 39 },
		{ 8, 502, 39 },
		{ 8, 503, 39 },
		{ 8, 504, 39 },
		{ 8, 505, 39 },
		{ 9, 444, 39 },
		{ 9, 445, 39 },
		{ 9, 446, 39 },
		{ 9, 447, 39 },
		{ 9, 448, 39 },
		{ 9, 449, 39 },
		{ 9, 450, 39 },
		{ 9, 451, 39 },
		{ 9, 452, 39 },
		{ 9, 453, 39 },
		{ 9, 454, 39 },
		{ 9, 455, 39 },
		{ 9, 456, 39 },
		{ 9, 457, 39 },
		{ 9, 458, 39 },
		{ 9, 459, 39 },
		{ 9, 460, 39 },
		{ 9, 461, 39 },
		{ 9, 462, 39 },
		{ 9, 463, 39 },
		{ 9, 464, 39 },
		{ 9, 465, 39 },
		{ 9, 466, 39 },
		{ 9, 467, 39 },
		{ 9, 468, 39 },
		{ 9, 469, 39 },
		{ 9, 470, 39 },
		{ 9, 471, 39 },
		{ 9, 472, 39 },
		{ 9, 473, 39 },
		{ 9, 474, 39 },
		{ 9, 475, 39 },
		{ 9, 476, 39 },
		{ 9, 477, 39 },
		{ 9, 478, 39 },
		{ 9, 479, 39 },
		{ 9, 480, 39 },
		{ 9, 481, 39 },
		{ 9, 482, 39 },
		{ 9, 483, 39 },
		{ 9, 484, 39 },
		{ 9, 485, 39 },
		{ 9, 486, 39 },
		{ 9, 487, 39 },
		{ 9, 488, 39 },
		{ 9, 489, 39 },
		{ 9, 490, 39 },
		{ 9, 491, 39 },
		{ 9, 492, 39 },
		{ 9, 493, 39 },
		{ 9, 494, 39 },
		{ 9, 495, 39 },
		{ 9, 496, 39 },
		{ 9, 497, 39 },
		{ 9, 498, 39 },
		{ 9, 499, 39 },
		{ 9, 500, 39 },
		{ 9, 501, 39 },
		{ 9, 502, 39 },
		{ 9, 503, 39 },
		{ 9, 504, 39 },
		{ 10, 443, 39 },
		{ 10, 444, 39 },
		{ 10, 445, 39 },
		{ 10, 446, 39 },
		{ 10, 447, 39 },
		{ 10, 448, 39 },
		{ 10, 449, 39 },
		{ 10, 450, 39 },
		{ 10, 451, 39 },
		{ 10, 452, 39 },
		{ 10, 453, 39 },
		{ 10, 454, 39 },
		{ 10, 455, 39 },
		{ 10, 456, 39 },
		{ 10, 457, 39 },
		{ 10, 458, 39 },
		{ 10, 459, 39 },
		{ 10, 460, 39 },
		{ 10, 461, 39 },
		{ 10, 462, 39 },
		{ 10, 463, 39 },
		{ 10, 464, 39 },
		{ 10, 465, 39 },
		{ 10, 466, 39 },
		{ 10, 467, 39 },
		{ 10, 468, 39 },
		{ 10, 469, 39 },
		{ 10, 470, 39 },
		{ 10, 471, 39 },
		{ 10, 472, 39 },
		{ 10, 473, 39 },
		{ 10, 474, 39 },
		{ 10, 475, 39 },
		{ 10, 476, 39 },
		{ 10, 477, 39 },
		{ 10, 478, 39 },
		{ 10, 479, 39 },
		{ 10, 480, 39 },
		{ 10, 481, 39 },
		{ 10, 482, 39 },
		{ 10, 483, 39 },
		{ 10, 484, 39 },
		{ 10, 485, 39 },
		{ 10, 486, 39 },
		{ 10, 487, 39 },
		{ 10, 488, 39 },
		{ 10, 489, 39 },
		{ 10, 490, 39 },
		{ 10, 491, 39 },
		{ 10, 492, 39 },
		{ 10, 493, 39 },
		{ 10, 494, 39 },
		{ 10, 495, 39 },
		{ 10, 496, 39 },
		{ 10, 497, 39 },
		{ 10, 498, 39 },
		{ 10, 499, 39 },
		{ 10, 500, 39 },
		{ 10, 501, 39 },
		{ 10, 502, 39 },
		{ 10, 503, 39 },
		{ 11, 442, 39 },
		{ 11, 443, 39 },
		{ 11, 444, 39 },
		{ 11, 445, 39 },
		{ 11, 446, 39 },
		{ 11, 447, 39 },
		{ 11, 448, 39 },
		{ 11, 449, 39 },
		{ 11, 450, 39 },
		{ 11, 451, 39 },
		{ 11, 452, 39 },
		{ 11, 453, 39 },
		{ 11, 454, 39 },
		{ 11, 455, 39 },
		{ 11, 456, 39 },
		{ 11, 457, 39 },
		{ 11, 458, 39 },
		{ 11, 459, 39 },
		{ 11, 460, 39 },
		{ 11, 461, 39 },
		{ 11, 462, 39 },
		{ 11, 463, 39 },
		{ 11, 464, 39 },
		{ 11, 465, 39 },
		{ 11, 466, 39 },
		{ 11, 467, 39 },
		{ 11, 468, 39 },
		{ 11, 469, 39 },
		{ 11, 470, 39 },
		{ 11, 471, 39 },
		{ 11, 472, 39 },
		{ 11, 473, 39 },
		{ 11, 474, 39 },
		{ 11, 475, 39 },
		{ 11, 476, 39 },
		{ 11, 477, 39 },
		{ 11, 478, 39 },
		{ 11, 479, 39 },
		{ 11, 480, 39 },
		{ 11, 481, 39 },
		{ 11, 482, 39 },
		{ 11, 483, 39 },
		{ 11, 484, 39 },
		{ 11, 485, 39 },
		{ 11, 486, 39 },
		{ 11, 487, 39 },
		{ 11, 488, 39 },
		{ 11, 489, 39 },
		{ 11, 490, 39 },
		{ 11, 491, 39 },
		{ 11, 492, 39 },
		{ 11, 493, 39 },
		{ 11, 494, 39 },
		{ 11, 495, 39 },
		{ 11, 496, 39 },
		{ 11, 497, 39 },
		{ 11, 498, 39 },
		{ 11, 499, 39 },
		{ 11, 500, 39 },
		{ 11, 501, 39 },
		{ 11, 502, 39 },
		{ 12, 441, 39 },
		{ 12, 442, 39 },
		{ 12, 443, 39 },
		{ 12, 444, 39 },
		{ 12, 445, 39 },
		{ 12, 446, 39 },
		{ 12, 447, 39 },
		{ 12, 448, 39 },
		{ 12, 449, 39 },
		{ 12, 450, 39 },
		{ 12, 451, 39 },
		{ 12, 452, 39 },
		{ 12, 453, 39 },
		{ 12, 454, 39 },
		{ 12, 455, 39 },
		{ 12, 456, 39 },
		{ 12, 457, 39 },
		{ 12, 458, 39 },
		{ 12, 459, 39 },
		{ 12, 460, 39 },
		{ 12, 461, 39 },
		{ 12, 462, 39 },
		{ 12, 463, 39 },
		{ 12, 464, 39 },
		{ 12, 465, 39 },
		{ 12, 466, 39 },
		{ 12, 467, 39 },
		{ 12, 468, 39 },
		{ 12, 469, 39 },
		{ 12, 470, 39 },
		{ 12, 471, 39 },
		{ 12, 472, 39 },
		{ 12, 473, 39 },
		{ 12, 474, 39 },
		{ 12, 475, 39 },
		{ 12, 476, 39 },
		{ 12, 477, 39 },
		{ 12, 478, 39 },
		{ 12, 479, 39 },
		{ 12, 480, 39 },
		{ 12, 481, 39 },
		{ 12, 482, 39 },
		{ 12, 483, 39 },
		{ 12, 484, 39 },
		{ 12, 485, 39 },
		{ 12, 486, 39 },
		{ 12, 487, 39 },
		{ 12, 488, 39 },
		{ 12, 489, 39 },
		{ 12, 490, 39 },
		{ 12, 491, 39 },
		{ 12, 492, 39 },
		{ 12, 493, 39 },
		{ 12, 494, 39 },
		{ 12, 495, 39 },
		{ 12, 496, 39 },
		{ 12, 497, 39 },
		{ 12, 498, 39 },
		{ 12, 499, 39 },
		{ 12, 500, 39 },
		{ 12, 501, 39 },
		{ 13, 441, 39 },
		{ 13, 442, 39 },
		{ 13, 443, 39 },
		{ 13, 444, 39 },
		{ 13, 445, 39 },
		{ 13, 446, 39 },
		{ 13, 447, 39 },
		{ 13, 448, 39 },
		{ 13, 449, 39 },
		{ 13, 450, 39 },
		{ 13, 451, 39 },
		{ 13, 452, 39 },
		{ 13, 453, 39 },
		{ 13, 454, 39 },
		{ 13, 455, 39 },
		{ 13, 456, 39 },
		{ 13, 457, 39 },
		{ 13, 458, 39 },
		{ 13, 459, 39 },
		{ 13, 460, 39 },
		{ 13, 461, 39 },
		{ 13, 462, 39 },
		{ 13, 463, 39 },
		{ 13, 464, 39 },
		{ 13, 465, 39 },
		{ 13, 466, 39 },
		{ 13, 467, 39 },
		{ 13, 468, 39 },
		{ 13, 469, 39 },
		{ 13, 470, 39 },
		{ 13, 471, 39 },
		{ 13, 472, 39 },
		{ 13, 473, 39 },
		{ 13, 474, 39 },
		{ 13, 475, 39 },
		{ 13, 476, 39 },
		{ 13, 477, 39 },
		{ 13, 478, 39 },
		{ 13, 479, 39 },
		{ 13, 480, 39 },
		{ 13, 481, 39 },
		{ 13, 482, 39 },
		{ 13, 483, 39 },
		{ 13, 484, 39 },
		{ 13, 485, 39 },
		{ 13, 486, 39 },
		{ 13, 487, 39 },
		{ 13, 488, 39 },
		{ 13, 489, 39 },
		{ 13, 490, 39 },
		{ 13, 491, 39 },
		{ 13, 492, 39 },
		{ 13, 493, 39 },
		{ 13, 494, 39 },
		{ 13, 495, 39 },
		{ 13, 496, 39 },
		{ 13, 497, 39 },
		{ 13, 498, 39 },
		{ 13, 499, 39 },
		{ 13, 500, 39 },
		{ 14, 440, 39 },
		{ 14, 441, 39 },
		{ 14, 442, 39 },
		{ 14, 443, 39 },
		{ 14, 444, 39 },
		{ 14, 445, 39 },
		{ 14, 446, 39 },
		{ 14, 447, 39 },
		{ 14, 448, 39 },
		{ 14, 449, 39 },
		{ 14, 450, 39 },
		{ 14, 451, 39 },
		{ 14, 452, 39 },
		{ 14, 453, 39 },
		{ 14, 454, 39 },
		{ 14, 455, 39 },
		{ 14, 456, 39 },
		{ 14, 457, 39 },
		{ 14, 458, 39 },
		{ 14, 459, 39 },
		{ 14, 460, 39 },
		{ 14, 461, 39 },
		{ 14, 462, 39 },
		{ 14, 463, 39 },
		{ 14, 464, 39 },
		{ 14, 465, 39 },
		{ 14, 466, 39 },
		{ 14, 467, 39 },
		{ 14, 468, 39 },
		{ 14, 469, 39 },
		{ 14, 470, 39 },
		{ 14, 471, 39 },
		{ 14, 472, 39 },
		{ 14, 473, 39 },
		{ 14, 474, 39 },
		{ 14, 475, 39 },
		{ 14, 476, 39 },
		{ 14, 477, 39 },
		{ 14, 478, 39 },
		{ 14, 479, 39 },
		{ 14, 480, 39 },
		{ 14, 481, 39 },
		{ 14, 482, 39 },
		{ 14, 483, 39 },
		{ 14, 484, 39 },
		{ 14, 485, 39 },
		{ 14, 486, 39 },
		{ 14, 487, 39 },
		{ 14, 488, 39 },
		{ 14, 489, 39 },
		{ 14, 490, 39 },
		{ 14, 491, 39 },
		{ 14, 492, 39 },
		{ 14, 493, 39 },
		{ 14, 494, 39 },
		{ 14, 495, 39 },
		{ 14, 496, 39 },
		{ 14, 497, 39 },
		{ 14, 498, 39 },
		{ 14, 499, 39 },
		{ 15, 439, 39 },
		{ 15, 440, 39 },
		{ 15, 441, 39 },
		{ 15, 442, 39 },
		{ 15, 443, 39 },
		{ 15, 444, 39 },
		{ 15, 445, 39 },
		{ 15, 446, 39 },
		{ 15, 447, 39 },
		{ 15, 448, 39 },
		{ 15, 449, 39 },
		{ 15, 450, 39 },
		{ 15, 451, 39 },
		{ 15, 452, 39 },
		{ 15, 453, 39 },
		{ 15, 454, 39 },
		{ 15, 455, 39 },
		{ 15, 456, 39 },
		{ 15, 457, 39 },
		{ 15, 458, 39 },
		{ 15, 459, 39 },
		{ 15, 460, 39 },
		{ 15, 461, 39 },
		{ 15, 462, 39 },
		{ 15, 463, 39 },
		{ 15, 464, 39 },
		{ 15, 465, 39 },
		{ 15, 466, 39 },
		{ 15, 467, 39 },
		{ 15, 468, 39 },
		{ 15, 469, 39 },
		{ 15, 470, 39 },
		{ 15, 471, 39 },
		{ 15, 472, 39 },
		{ 15, 473, 39 },
		{ 15, 474, 39 },
		{ 15, 475, 39 },
		{ 15, 476, 39 },
		{ 15, 477, 39 },
		{ 15, 478, 39 },
		{ 15, 479, 39 },
		{ 15, 480, 39 },
		{ 15, 481, 39 },
		{ 15, 482, 39 },
		{ 15, 483, 39 },
		{ 15, 484, 39 },
		{ 15, 485, 39 },
		{ 15, 486, 39 },
		{ 15, 487, 39 },
		{ 15, 488, 39 },
		{ 15, 489, 39 },
		{ 15, 490, 39 },
		{ 15, 491, 39 },
		{ 15, 492, 39 },
		{ 15, 493, 39 },
		{ 15, 494, 39 },
		{ 15, 495, 39 },
		{ 15, 496, 39 },
		{ 15, 497, 39 },
		{ 15, 498, 39 },
		{ 16, 438, 39 },
		{ 16, 439, 39 },
		{ 16, 440, 39 },
		{ 16, 441, 39 },
		{ 16, 442, 39 },
		{ 16, 443, 39 },
		{ 16, 444, 39 },
		{ 16, 445, 39 },
		{ 16, 446, 39 },
		{ 16, 447, 39 },
		{ 16, 448, 39 },
		{ 16, 449, 39 },
		{ 16, 450, 39 },
		{ 16, 451, 39 },
		{ 16, 452, 39 },
		{ 16, 453, 39 },
		{ 16, 454, 39 },
		{ 16, 455, 39 },
		{ 16, 456, 39 },
		{ 16, 457, 39 },
		{ 16, 458, 39 },
		{ 16, 459, 39 },
		{ 16, 460, 39 },
		{ 16, 461, 39 },
		{ 16, 462, 39 },
		{ 16, 463, 39 },
		{ 16, 464, 39 },
		{ 16, 465, 39 },
		{ 16, 466, 39 },
		{ 16, 467, 39 },
		{ 16, 468, 39 },
		{ 16, 469, 39 },
		{ 16, 470, 39 },
		{ 16, 471, 39 },
		{ 16, 472, 39 },
		{ 16, 473, 39 },
		{ 16, 474, 39 },
		{ 16, 475, 39 },
		{ 16, 476, 39 },
		{ 16, 477, 39 },
		{ 16, 478, 39 },
		{ 16, 479, 39 },
		{ 16, 480, 39 },
		{ 16, 481, 39 },
		{ 16, 482, 39 },
		{ 16, 483, 39 },
		{ 16, 484, 39 },
		{ 16, 485, 39 },
		{ 16, 486, 39 },
		{ 16, 487, 39 },
		{ 16, 488, 39 },
		{ 16, 489, 39 },
		{ 16, 490, 39 },
		{ 16, 491, 39 },
		{ 16, 492, 39 },
		{ 16, 493, 39 },
		{ 16, 494, 39 },
		{ 16, 495, 39 },
		{ 16, 496, 39 },
		{ 16, 497, 39 },
		{ 17, 438, 39 },
		{ 17, 439, 39 },
		{ 17, 440, 39 },
		{ 17, 441, 39 },
		{ 17, 442, 39 },
		{ 17, 443, 39 },
		{ 17, 444, 39 },
		{ 17, 445, 39 },
		{ 17, 446, 39 },
		{ 17, 447, 39 },
		{ 17, 448, 39 },
		{ 17, 449, 39 },
		{ 17, 450, 39 },
		{ 17, 451, 39 },
		{ 17, 452, 39 },
		{ 17, 453, 39 },
		{ 17, 454, 39 },
		{ 17, 455, 39 },
		{ 17, 456, 39 },
		{ 17, 457, 39 },
		{ 17, 458, 39 },
		{ 17, 459, 39 },
		{ 17, 460, 39 },
		{ 17, 461, 39 },
		{ 17, 462, 39 },
		{ 17, 463, 39 },
		{ 17, 464, 39 },
		{ 17, 465, 39 },
		{ 17, 466, 39 },
		{ 17, 467, 39 },
		{ 17, 468, 39 },
		{ 17, 469, 39 },
		{ 17, 470, 39 },
		{ 17, 471, 39 },
		{ 17, 472, 39 },
		{ 17, 473, 39 },
		{ 17, 474, 39 },
		{ 17, 475, 39 },
		{ 17, 476, 39 },
		{ 17, 477, 39 },
		{ 17, 478, 39 },
		{ 17, 479, 39 },
		{ 17, 480, 39 },
		{ 17, 481, 39 },
		{ 17, 482, 39 },
		{ 17, 483, 39 },
		{ 17, 484, 39 },
		{ 17, 485, 39 },
		{ 17, 486, 39 },
		{ 17, 487, 39 },
		{ 17, 488, 39 },
		{ 17, 489, 39 },
		{ 17, 490, 39 },
		{ 17, 491, 39 },
		{ 17, 492, 39 },
		{ 17, 493, 39 },
		{ 17, 494, 39 },
		{ 17, 495, 39 },
		{ 17, 496, 39 },
		{ 18, 437, 39 },
		{ 18, 438, 39 },
		{ 18, 439, 39 },
		{ 18, 440, 39 },
		{ 18, 441, 39 },
		{ 18, 442, 39 },
		{ 18, 443, 39 },
		{ 18, 444, 39 },
		{ 18, 445, 39 },
		{ 18, 446, 39 },
		{ 18, 447, 39 },
		{ 18, 448, 39 },
		{ 18, 449, 39 },
		{ 18, 450, 39 },
		{ 18, 451, 39 },
		{ 18, 452, 39 },
		{ 18, 453, 39 },
		{ 18, 454, 39 },
		{ 18, 455, 39 },
		{ 18, 456, 39 },
		{ 18, 457, 39 },
		{ 18, 458, 39 },
		{ 18, 459, 39 },
		{ 18, 460, 39 },
		{ 18, 461, 39 },
		{ 18, 462, 39 },
		{ 18, 463, 39 },
		{ 18, 464, 39 },
		{ 18, 465, 39 },
		{ 18, 466, 39 },
		{ 18, 467, 39 },
		{ 18, 468, 39 },
		{ 18, 469, 39 },
		{ 18, 470, 39 },
		{ 18, 471, 39 },
		{ 18, 472, 39 },
		{ 18, 473, 39 },
		{ 18, 474, 39 },
		{ 18, 475, 39 },
		{ 18, 476, 39 },
		{ 18, 477, 39 },
		{ 18, 478, 39 },
		{ 18, 479, 39 },
		{ 18, 480, 39 },
		{ 18, 481, 39 },
		{ 18, 482, 39 },
		{ 18, 483, 39 },
		{ 18, 484, 39 },
		{ 18, 485, 39 },
		{ 18, 486, 39 },
		{ 18, 487, 39 },
		{ 18, 488, 39 },
		{ 18, 489, 39 },
		{ 18, 490, 39 },
		{ 18, 491, 39 },
		{ 18, 492, 39 },
		{ 18, 493, 39 },
		{ 18, 494, 39 },
		{ 18, 495, 39 },
		{ 19, 436, 39 },
		{ 19, 437, 39 },
		{ 19, 438, 39 },
		{ 19, 439, 39 },
		{ 19, 440, 39 },
		{ 19, 441, 39 },
		{ 19, 442, 39 },
		{ 19, 443, 39 },
		{ 19, 444, 39 },
		{ 19, 445, 39 },
		{ 19, 446, 39 },
		{ 19, 447, 39 },
		{ 19, 448, 39 },
		{ 19, 449, 39 },
		{ 19, 450, 39 },
		{ 19, 451, 39 },
		{ 19, 452, 39 },
		{ 19, 453, 39 },
		{ 19, 454, 39 },
		{ 19, 455, 39 },
		{ 19, 456, 39 },
		{ 19, 457, 39 },
		{ 19, 458, 39 },
		{ 19, 459, 39 },
		{ 19, 460, 39 },
		{ 19, 461, 39 },
		{ 19, 462, 39 },
		{ 19, 463, 39 },
		{ 19, 464, 39 },
		{ 19, 465, 39 },
		{ 19, 466, 39 },
		{ 19, 467, 39 },
		{ 19, 468, 39 },
		{ 19, 469, 39 },
		{ 19, 470, 39 },
		{ 19, 471, 39 },
		{ 19, 472, 39 },
		{ 19, 473, 39 },
		{ 19, 474, 39 },
		{ 19, 475, 39 },
		{ 19, 476, 39 },
		{ 19, 477, 39 },
		{ 19, 478, 39 },
		{ 19, 479, 39 },
		{ 19, 480, 39 },
		{ 19, 481, 39 },
		{ 19, 482, 39 },
		{ 19, 483, 39 },
		{ 19, 484, 39 },
		{ 19, 485, 39 },
		{ 19, 486, 39 },
		{ 19, 487, 39 },
		{ 19, 488, 39 },
		{ 19, 489, 39 },
		{ 19, 490, 39 },
		{ 19, 491, 39 },
		{ 19, 492, 39 },
		{ 19, 493, 39 },
		{ 19, 494, 39 },
		{ 20, 435, 39 },
		{ 20, 436, 39 },
		{ 20, 437, 39 },
		{ 20, 438, 39 },
		{ 20, 439, 39 },
		{ 20, 440, 39 },
		{ 20, 441, 39 },
		{ 20, 442, 39 },
		{ 20, 443, 39 },
		{ 20, 444, 39 },
		{ 20, 445, 39 },
		{ 20, 446, 39 },
		{ 20, 447, 39 },
		{ 20, 448, 39 },
		{ 20, 449, 39 },
		{ 20, 450, 39 },
		{ 20, 451, 39 },
		{ 20, 452, 39 },
		{ 20, 453, 39 },
		{ 20, 454, 39 },
		{ 20, 455, 39 },
		{ 20, 456, 39 },
		{ 20, 457, 39 },
		{ 20, 458, 39 },
		{ 20, 459, 39 },
		{ 20, 460, 39 },
		{ 20, 461, 39 },
		{ 20, 462, 39 },
		{ 20, 463, 39 },
		{ 20, 464, 39 },
		{ 20, 465, 39 },
		{ 20, 466, 39 },
		{ 20, 467, 39 },
		{ 20, 468, 39 },
		{ 20, 469, 39 },
		{ 20, 470, 39 },
		{ 20, 471, 39 },
		{ 20, 472, 39 },
		{ 20, 473, 39 },
		{ 20, 474, 39 },
		{ 20, 475, 39 },
		{ 20, 476, 39 },
		{ 20, 477, 39 },
		{ 20, 478, 39 },
		{ 20, 479, 39 },
		{ 20, 480, 39 },
		{ 20, 481, 39 },
		{ 20, 482, 39 },
		{ 20, 483, 39 },
		{ 20, 484, 39 },
		{ 20, 485, 39 },
		{ 20, 486, 39 },
		{ 20, 487, 39 },
		{ 20, 488, 39 },
		{ 20, 489, 39 },
		{ 20, 490, 39 },
		{ 20, 491, 39 },
		{ 20, 492, 39 },
		{ 20, 493, 39 },
		{ 21, 435, 39 },
		{ 21, 436, 39 },
		{ 21, 437, 39 },
		{ 21, 438, 39 },
		{ 21, 439, 39 },
		{ 21, 440, 39 },
		{ 21, 441, 39 },
		{ 21, 442, 39 },
		{ 21, 443, 39 },
		{ 21, 444, 39 },
		{ 21, 445, 39 },
		{ 21, 446, 39 },
		{ 21, 447, 39 },
		{ 21, 448, 39 },
		{ 21, 449, 39 },
		{ 21, 450, 39 },
		{ 21, 451, 39 },
		{ 21, 452, 39 },
		{ 21, 453, 39 },
		{ 21, 454, 39 },
		{ 21, 455, 39 },
		{ 21, 456, 39 },
		{ 21, 457, 39 },
		{ 21, 458, 39 },
		{ 21, 459, 39 },
		{ 21, 460, 39 },
		{ 21, 461, 39 },
		{ 21, 462, 39 },
		{ 21, 463, 39 },
		{ 21, 464, 39 },
		{ 21, 465, 39 },
		{ 21, 466, 39 },
		{ 21, 467, 39 },
		{ 21, 468, 39 },
		{ 21, 469, 39 },
		{ 21, 470, 39 },
		{ 21, 471, 39 },
		{ 21, 472, 39 },
		{ 21, 473, 39 },
		{ 21, 474, 39 },
		{ 21, 475, 39 },
		{ 21, 476, 39 },
		{ 21, 477, 39 },
		{ 21, 478, 39 },
		{ 21, 479, 39 },
		{ 21, 480, 39 },
		{ 21, 481, 39 },
		{ 21, 482, 39 },
		{ 21, 483, 39 },
		{ 21, 484, 39 },
		{ 21, 485, 39 },
		{ 21, 486, 39 },
		{ 21, 487, 39 },
		{ 21, 488, 39 },
		{ 21, 489, 39 },
		{ 21, 490, 39 },
		{ 21, 491, 39 },
		{ 21, 492, 39 },
		{ 22, 434, 39 },
		{ 22, 435, 39 },
		{ 22, 436, 39 },
		{ 22, 437, 39 },
		{ 22, 438, 39 },
		{ 22, 439, 39 },
		{ 22, 440, 39 },
		{ 22, 441, 39 },
		{ 22, 442, 39 },
		{ 22, 443, 39 },
		{ 22, 444, 39 },
		{ 22, 445, 39 },
		{ 22, 446, 39 },
		{ 22, 447, 39 },
		{ 22, 448, 39 },
		{ 22, 449, 39 },
		{ 22, 450, 39 },
		{ 22, 451, 39 },
		{ 22, 452, 39 },
		{ 22, 453, 39 },
		{ 22, 454, 39 },
		{ 22, 455, 39 },
		{ 22, 456, 39 },
		{ 22, 457, 39 },
		{ 22, 458, 39 },
		{ 22, 459, 39 },
		{ 22, 460, 39 },
		{ 22, 461, 39 },
		{ 22, 462, 39 },
		{ 22, 463, 39 },
		{ 22, 464, 39 },
		{ 22, 465, 39 },
		{ 22, 466, 39 },
		{ 22, 467, 39 },
		{ 22, 468, 39 },
		{ 22, 469, 39 },
		{ 22, 470, 39 },
		{ 22, 471, 39 },
		{ 22, 472, 39 },
		{ 22, 473, 39 },
		{ 22, 474, 39 },
		{ 22, 475, 39 },
		{ 22, 476, 39 },
		{ 22, 477, 39 },
		{ 22, 478, 39 },
		{ 22, 479, 39 },
		{ 22, 480, 39 },
		{ 22, 481, 39 },
		{ 22, 482, 39 },
		{ 22, 483, 39 },
		{ 22, 484, 39 },
		{ 22, 485, 39 },
		{ 22, 486, 39 },
		{ 22, 487, 39 },
		{ 22, 488, 39 },
		{ 22, 489, 39 },
		{ 22, 490, 39 },
		{ 22, 491, 39 },
		{ 23, 433, 39 },
		{ 23, 434, 39 },
		{ 23, 435, 39 },
		{ 23, 436, 39 },
		{ 23, 437, 39 },
		{ 23, 438, 39 },
		{ 23, 439, 39 },
		{ 23, 440, 39 },
		{ 23, 441, 39 },
		{ 23, 442, 39 },
		{ 23, 443, 39 },
		{ 23, 444, 39 },
		{ 23, 445, 39 },
		{ 23, 446, 39 },
		{ 23, 447, 39 },
		{ 23, 448, 39 },
		{ 23, 449, 39 },
		{ 23, 450, 39 },
		{ 23, 451, 39 },
		{ 23, 452, 39 },
		{ 23, 453, 39 },
		{ 23, 454, 39 },
		{ 23, 455, 39 },
		{ 23, 456, 39 },
		{ 23, 457, 39 },
		{ 23, 458, 39 },
		{ 23, 459, 39 },
		{ 23, 460, 39 },
		{ 23, 461, 39 },
		{ 23, 462, 39 },
		{ 23, 463, 39 },
		{ 23, 464, 39 },
		{ 23, 465, 39 },
		{ 23, 466, 39 },
		{ 23, 467, 39 },
		{ 23, 468, 39 },
		{ 23, 469, 39 },
		{ 23, 470, 39 },
		{ 23, 471, 39 },
		{ 23, 472, 39 },
		{ 23, 473, 39 },
		{ 23, 474, 39 },
		{ 23, 475, 39 },
		{ 23, 476, 39 },
		{ 23, 477, 39 },
		{ 23, 478, 39 },
		{ 23, 479, 39 },
		{ 23, 480, 39 },
		{ 23, 481, 39 },
		{ 23, 482, 39 },
		{ 23, 483, 39 },
		{ 23, 484, 39 },
		{ 23, 485, 39 },
		{ 23, 486, 39 },
		{ 23, 487, 39 },
		{ 23, 488, 39 },
		{ 23, 489, 39 },
		{ 23, 490, 39 },
		{ 24, 432, 39 },
		{ 24, 433, 39 },
		{ 24, 434, 39 },
		{ 24, 435, 39 },
		{ 24, 436, 39 },
		{ 24, 437, 39 },
		{ 24, 438, 39 },
		{ 24, 439, 39 },
		{ 24, 440, 39 },
		{ 24, 441, 39 },
		{ 24, 442, 39 },
		{ 24, 443, 39 },
		{ 24, 444, 39 },
		{ 24, 445, 39 },
		{ 24, 446, 39 },
		{ 24, 447, 39 },
		{ 24, 448, 39 },
		{ 24, 449, 39 },
		{ 24, 450, 39 },
		{ 24, 451, 39 },
		{ 24, 452, 39 },
		{ 24, 453, 39 },
		{ 24, 454, 39 },
		{ 24, 455, 39 },
		{ 24, 456, 39 },
		{ 24, 457, 39 },
		{ 24, 458, 39 },
		{ 24, 459, 39 },
		{ 24, 460, 39 },
		{ 24, 461, 39 },
		{ 24, 462, 39 },
		{ 24, 463, 39 },
		{ 24, 464, 39 },
		{ 24, 465, 39 },
		{ 24, 466, 39 },
		{ 24, 467, 39 },
		{ 24, 468, 39 },
		{ 24, 469, 39 },
		{ 24, 470, 39 },
		{ 24, 471, 39 },
		{ 24, 472, 39 },
		{ 24, 473, 39 },
		{ 24, 474, 39 },
		{ 24, 475, 39 },
		{ 24, 476, 39 },
		{ 24, 477, 39 },
		{ 24, 478, 39 },
		{ 24, 479, 39 },
		{ 24, 480, 39 },
		{ 24, 481, 39 },
		{ 24, 482, 39 },
		{ 24, 483, 39 },
		{ 24, 484, 39 },
		{ 24, 485, 39 },
		{ 24, 486, 39 },
		{ 24, 487, 39 },
		{ 24, 488, 39 },
		{ 24, 489, 39 },
		{ 25, 432, 39 },
		{ 25, 433, 39 },
		{ 25, 434, 39 },
		{ 25, 435, 39 },
		{ 25, 436, 39 },
		{ 25, 437, 39 },
		{ 25, 438, 39 },
		{ 25, 439, 39 },
		{ 25, 440, 39 },
		{ 25, 441, 39 },
		{ 25, 442, 39 },
		{ 25, 443, 39 },
		{ 25, 444, 39 },
		{ 25, 445, 39 },
		{ 25, 446, 39 },
		{ 25, 447, 39 },
		{ 25, 448, 39 },
		{ 25, 449, 39 },
		{ 25, 450, 39 },
		{ 25, 451, 39 },
		{ 25, 452, 39 },
		{ 25, 453, 39 },
		{ 25, 454, 39 },
		{ 25, 455, 39 },
		{ 25, 456, 39 },
		{ 25, 457, 39 },
		{ 25, 458, 39 },
		{ 25, 459, 39 },
		{ 25, 460, 39 },
		{ 25, 461, 39 },
		{ 25, 462, 39 },
		{ 25, 463, 39 },
		{ 25, 464, 39 },
		{ 25, 465, 39 },
		{ 25, 466, 39 },
		{ 25, 467, 39 },
		{ 25, 468, 39 },
		{ 25, 469, 39 },
		{ 25, 470, 39 },
		{ 25, 471, 39 },
		{ 25, 472, 39 },
		{ 25, 473, 39 },
		{ 25, 474, 39 },
		{ 25, 475, 39 },
		{ 25, 476, 39 },
		{ 25, 477, 39 },
		{ 25, 478, 39 },
		{ 25, 479, 39 },
		{ 25, 480, 39 },
		{ 25, 481, 39 },
		{ 25, 482, 39 },
		{ 25, 483, 39 },
		{ 25, 484, 39 },
		{ 25, 485, 39 },
		{ 25, 486, 39 },
		{ 25, 487, 39 },
		{ 25, 488, 39 },
		{ 26, 431, 39 },
		{ 26, 432, 39 },
		{ 26, 433, 39 },
		{ 26, 434, 39 },
		{ 26, 435, 39 },
		{ 26, 436, 39 },
		{ 26, 437, 39 },
		{ 26, 438, 39 },
		{ 26, 439, 39 },
		{ 26, 440, 39 },
		{ 26, 441, 39 },
		{ 26, 442, 39 },
		{ 26, 443, 39 },
		{ 26, 444, 39 },
		{ 26, 445, 39 },
		{ 26, 446, 39 },
		{ 26, 447, 39 },
		{ 26, 448, 39 },
		{ 26, 449, 39 },
		{ 26, 450, 39 },
		{ 26, 451, 39 },
		{ 26, 452, 39 },
		{ 26, 453, 39 },
		{ 26, 454, 39 },
		{ 26, 455, 39 },
		{ 26, 456, 39 },
		{ 26, 457, 39 },
		{ 26, 458, 39 },
		{ 26, 459, 39 },
		{ 26, 460, 39 },
		{ 26, 461, 39 },
		{ 26, 462, 39 },
		{ 26, 463, 39 },
		{ 26, 464, 39 },
		{ 26, 465, 39 },
		{ 26, 466, 39 },
		{ 26, 467, 39 },
		{ 26, 468, 39 },
		{ 26, 469, 39 },
		{ 26, 470, 39 },
		{ 26, 471, 39 },
		{ 26, 472, 39 },
		{ 26, 473, 39 },
		{ 26, 474, 39 },
		{ 26, 475, 39 },
		{ 26, 476, 39 },
		{ 26, 477, 39 },
		{ 26, 478, 39 },
		{ 26, 479, 39 },
		{ 26, 480, 39 },
		{ 26, 481, 39 },
		{ 26, 482, 39 },
		{ 26, 483, 39 },
		{ 26, 484, 39 },
		{ 26, 485, 39 },
		{ 26, 486, 39 },
		{ 26, 487, 39 },
		{ 27, 430, 39 },
		{ 27, 431, 39 },
		{ 27, 432, 39 },
		{ 27, 433, 39 },
		{ 27, 434, 39 },
		{ 27, 435, 39 },
		{ 27, 436, 39 },
		{ 27, 437, 39 },
		{ 27, 438, 39 },
		{ 27, 439, 39 },
		{ 27, 440, 39 },
		{ 27, 441, 39 },
		{ 27, 442, 39 },
		{ 27, 443, 39 },
		{ 27, 444, 39 },
		{ 27, 445, 39 },
		{ 27, 446, 39 },
		{ 27, 447, 39 },
		{ 27, 448, 39 },
		{ 27, 449, 39 },
		{ 27, 450, 39 },
		{ 27, 451, 39 },
		{ 27, 452, 39 },
		{ 27, 453, 39 },
		{ 27, 454, 39 },
		{ 27, 455, 39 },
		{ 27, 456, 39 },
		{ 27, 457, 39 },
		{ 27, 458, 39 },
		{ 27, 459, 39 },
		{ 27, 460, 39 },
		{ 27, 461, 39 },
		{ 27, 462, 39 },
		{ 27, 463, 39 },
		{ 27, 464, 39 },
		{ 27, 465, 39 },
		{ 27, 466, 39 },
		{ 27, 467, 39 },
		{ 27, 468, 39 },
		{ 27, 469, 39 },
		{ 27, 470, 39 },
		{ 27, 471, 39 },
		{ 27, 472, 39 },
		{ 27, 473, 39 },
		{ 27, 474, 39 },
		{ 27, 475, 39 },
		{ 27, 476, 39 },
		{ 27, 477, 39 },
		{ 27, 478, 39 },
		{ 27, 479, 39 },
		{ 27, 480, 39 },
		{ 27, 481, 39 },
		{ 27, 482, 39 },
		{ 27, 483, 39 },
		{ 27, 484, 39 },
		{ 27, 485, 39 },
		{ 27, 486, 39 },
		{ 28, 429, 39 },
		{ 28, 430, 39 },
		{ 28, 431, 39 },
		{ 28, 432, 39 },
		{ 28, 433, 39 },
		{ 28, 434, 39 },
		{ 28, 435, 39 },
		{ 28, 436, 39 },
		{ 28, 437, 39 },
		{ 28, 438, 39 },
		{ 28, 439, 39 },
		{ 28, 440, 39 },
		{ 28, 441, 39 },
		{ 28, 442, 39 },
		{ 28, 443, 39 },
		{ 28, 444, 39 },
		{ 28, 445, 39 },
		{ 28, 446, 39 },
		{ 28, 447, 39 },
		{ 28, 448, 39 },
		{ 28, 449, 39 },
		{ 28, 450, 39 },
		{ 28, 451, 39 },
		{ 28, 452, 39 },
		{ 28, 453, 39 },
		{ 28, 454, 39 },
		{ 28, 455, 39 },
		{ 28, 456, 39 },
		{ 28, 457, 39 },
		{ 28, 458, 39 },
		{ 28, 459, 39 },
		{ 28, 460, 39 },
		{ 28, 461, 39 },
		{ 28, 462, 39 },
		{ 28, 463, 39 },
		{ 28, 464, 39 },
		{ 28, 465, 39 },
		{ 28, 466, 39 },
		{ 28, 467, 39 },
		{ 28, 468, 39 },
		{ 28, 469, 39 },
		{ 28, 470, 39 },
		{ 28, 471, 39 },
		{ 28, 472, 39 },
		{ 28, 473, 39 },
		{ 28, 474, 39 },
		{ 28, 475, 39 },
		{ 28, 476, 39 },
		{ 28, 477, 39 },
		{ 28, 478, 39 },
		{ 28, 479, 39 },
		{ 28, 480, 39 },
		{ 28, 481, 39 },
		{ 28, 482, 39 },
		{ 28, 483, 39 },
		{ 28, 484, 39 },
		{ 28, 485, 39 },
		{ 29, 429, 39 },
		{ 29, 430, 39 },
		{ 29, 431, 39 },
		{ 29, 432, 39 },
		{ 29, 433, 39 },
		{ 29, 434, 39 },
		{ 29, 435, 39 },
		{ 29, 436, 39 },
		{ 29, 437, 39 },
		{ 29, 438, 39 },
		{ 29, 439, 39 },
		{ 29, 440, 39 },
		{ 29, 441, 39 },
		{ 29, 442, 39 },
		{ 29, 443, 39 },
		{ 29, 444, 39 },
		{ 29, 445, 39 },
		{ 29, 446, 39 },
		{ 29, 447, 39 },
		{ 29, 448, 39 },
		{ 29, 449, 39 },
		{ 29, 450, 39 },
		{ 29, 451, 39 },
		{ 29, 452, 39 },
		{ 29, 453, 39 },
		{ 29, 454, 39 },
		{ 29, 455, 39 },
		{ 29, 456, 39 },
		{ 29, 457, 39 },
		{ 29, 458, 39 },
		{ 29, 459, 39 },
		{ 29, 460, 39 },
		{ 29, 461, 39 },
		{ 29, 462, 39 },
		{ 29, 463, 39 },
		{ 29, 464, 39 },
		{ 29, 465, 39 },
		{ 29, 466, 39 },
		{ 29, 467, 39 },
		{ 29, 468, 39 },
		{ 29, 469, 39 },
		{ 29, 470, 39 },
		{ 29, 471, 39 },
		{ 29, 472, 39 },
		{ 29, 473, 39 },
		{ 29, 474, 39 },
		{ 29, 475, 39 },
		{ 29, 476, 39 },
		{ 29, 477, 39 },
		{ 29, 478, 39 },
		{ 29, 479, 39 },
		{ 29, 480, 39 },
		{ 29, 481, 39 },
		{ 29, 482, 39 },
		{ 29, 483, 39 },
		{ 29, 484, 39 },
		{ 30, 428, 39 },
		{ 30, 429, 39 },
		{ 30, 430, 39 },
		{ 30, 431, 39 },
		{ 30, 432, 39 },
		{ 30, 433, 39 },
		{ 30, 434, 39 },
		{ 30, 435, 39 },
		{ 30, 436, 39 },
		{ 30, 437, 39 },
		{ 30, 438, 39 },
		{ 30, 439, 39 },
		{ 30, 440, 39 },
		{ 30, 441, 39 },
		{ 30, 442, 39 },
		{ 30, 443, 39 },
		{ 30, 444, 39 },
		{ 30, 445, 39 },
		{ 30, 446, 39 },
		{ 30, 447, 39 },
		{ 30, 448, 39 },
		{ 30, 449, 39 },
		{ 30, 450, 39 },
		{ 30, 451, 39 },
		{ 30, 452, 39 },
		{ 30, 453, 39 },
		{ 30, 454, 39 },
		{ 30, 455, 39 },
		{ 30, 456, 39 },
		{ 30, 457, 39 },
		{ 30, 458, 39 },
		{ 30, 459, 39 },
		{ 30, 460, 39 },
		{ 30, 461, 39 },
		{ 30, 462, 39 },
		{ 30, 463, 39 },
		{ 30, 464, 39 },
		{ 30, 465, 39 },
		{ 30, 466, 39 },
		{ 30, 467, 39 },
		{ 30, 468, 39 },
		{ 30, 469, 39 },
		{ 30, 470, 39 },
		{ 30, 471, 39 },
		{ 30, 472, 39 },
		{ 30, 473, 39 },
		{ 30, 474, 39 },
		{ 30, 475, 39 },
		{ 30, 476, 39 },
		{ 30, 477, 39 },
		{ 30, 478, 39 },
		{ 30, 479, 39 },
		{ 30, 480, 39 },
		{ 30, 481, 39 },
		{ 30, 482, 39 },
		{ 30, 483, 39 },
		{ 31, 427, 39 },
		{ 31, 428, 39 },
		{ 31, 429, 39 },
		{ 31, 430, 39 },
		{ 31, 431, 39 },
		{ 31, 432, 39 },
		{ 31, 433, 39 },
		{ 31, 434, 39 },
		{ 31, 435, 39 },
		{ 31, 436, 39 },
		{ 31, 437, 39 },
		{ 31, 438, 39 },
		{ 31, 439, 39 },
		{ 31, 440, 39 },
		{ 31, 441, 39 },
		{ 31, 442, 39 },
		{ 31, 443, 39 },
		{ 31, 444, 39 },
		{ 31, 445, 39 },
		{ 31, 446, 39 },
		{ 31, 447, 39 },
		{ 31, 448, 39 },
		{ 31, 449, 39 },
		{ 31, 450, 39 },
		{ 31, 451, 39 },
		{ 31, 452, 39 },
		{ 31, 453, 39 },
		{ 31, 454, 39 },
		{ 31, 455, 39 },
		{ 31, 456, 39 },
		{ 31, 457, 39 },
		{ 31, 458, 39 },
		{ 31, 459, 39 },
		{ 31, 460, 39 },
		{ 31, 461, 39 },
		{ 31, 462, 39 },
		{ 31, 463, 39 },
		{ 31, 464, 39 },
		{ 31, 465, 39 },
		{ 31, 466, 39 },
		{ 31, 467, 39 },
		{ 31, 468, 39 },
		{ 31, 469, 39 },
		{ 31, 470, 39 },
		{ 31, 471, 39 },
		{ 31, 472, 39 },
		{ 31, 473, 39 },
		{ 31, 474, 39 },
		{ 31, 475, 39 },
		{ 31, 476, 39 },
		{ 31, 477, 39 },
		{ 31, 478, 39 },
		{ 31, 479, 39 },
		{ 31, 480, 39 },
		{ 31, 481, 39 },
		{ 31, 482, 39 },
		{ 32, 426, 39 },
		{ 32, 427, 39 },
		{ 32, 428, 39 },
		{ 32, 429, 39 },
		{ 32, 430, 39 },
		{ 32, 431, 39 },
		{ 32, 432, 39 },
		{ 32, 433, 39 },
		{ 32, 434, 39 },
		{ 32, 435, 39 },
		{ 32, 436, 39 },
		{ 32, 437, 39 },
		{ 32, 438, 39 },
		{ 32, 439, 39 },
		{ 32, 440, 39 },
		{ 32, 441, 39 },
		{ 32, 442, 39 },
		{ 32, 443, 39 },
		{ 32, 444, 39 },
		{ 32, 445, 39 },
		{ 32, 446, 39 },
		{ 32, 447, 39 },
		{ 32, 448, 39 },
		{ 32, 449, 39 },
		{ 32, 450, 39 },
		{ 32, 451, 39 },
		{ 32, 452, 39 },
		{ 32, 453, 39 },
		{ 32, 454, 39 },
		{ 32, 455, 39 },
		{ 32, 456, 39 },
		{ 32, 457, 39 },
		{ 32, 458, 39 },
		{ 32, 459, 39 },
		{ 32, 460, 39 },
		{ 32, 461, 39 },
		{ 32, 462, 39 },
		{ 32, 463, 39 },
		{ 32, 464, 39 },
		{ 32, 465, 39 },
		{ 32, 466, 39 },
		{ 32, 467, 39 },
		{ 32, 468, 39 },
		{ 32, 469, 39 },
		{ 32, 470, 39 },
		{ 32, 471, 39 },
		{ 32, 472, 39 },
		{ 32, 473, 39 },
		{ 32, 474, 39 },
		{ 32, 475, 39 },
		{ 32, 476, 39 },
		{ 32, 477, 39 },
		{ 32, 478, 39 },
		{ 32, 479, 39 },
		{ 32, 480, 39 },
		{ 32, 481, 39 },
		{ 33, 426, 39 },
		{ 33, 427, 39 },
		{ 33, 428, 39 },
		{ 33, 429, 39 },
		{ 33, 430, 39 },
		{ 33, 431, 39 },
		{ 33, 432, 39 },
		{ 33, 433, 39 },
		{ 33, 434, 39 },
		{ 33, 435, 39 },
		{ 33, 436, 39 },
		{ 33, 437, 39 },
		{ 33, 438, 39 },
		{ 33, 439, 39 },
		{ 33, 440, 39 },
		{ 33, 441, 39 },
		{ 33, 442, 39 },
		{ 33, 443, 39 },
		{ 33, 444, 39 },
		{ 33, 445, 39 },
		{ 33, 446, 39 },
		{ 33, 447, 39 },
		{ 33, 448, 39 },
		{ 33, 449, 39 },
		{ 33, 450, 39 },
		{ 33, 451, 39 },
		{ 33, 452, 39 },
		{ 33, 453, 39 },
		{ 33, 454, 39 },
		{ 33, 455, 39 },
		{ 33, 456, 39 },
		{ 33, 457, 39 },
		{ 33, 458, 39 },
		{ 33, 459, 39 },
		{ 33, 460, 39 },
		{ 33, 461, 39 },
		{ 33, 462, 39 },
		{ 33, 463, 39 },
		{ 33, 464, 39 },
		{ 33, 465, 39 },
		{ 33, 466, 39 },
		{ 33, 467, 39 },
		{ 33, 468, 39 },
		{ 33, 469, 39 },
		{ 33, 470, 39 },
		{ 33, 471, 39 },
		{ 33, 472, 39 },
		{ 33, 473, 39 },
		{ 33, 474, 39 },
		{ 33, 475, 39 },
		{ 33, 476, 39 },
		{ 33, 477, 39 },
		{ 33, 478, 39 },
		{ 33, 479, 39 },
		{ 33, 480, 39 },
		{ 34, 425, 39 },
		{ 34, 426, 39 },
		{ 34, 427, 39 },
		{ 34, 428, 39 },
		{ 34, 429, 39 },
		{ 34, 430, 39 },
		{ 34, 431, 39 },
		{ 34, 432, 39 },
		{ 34, 433, 39 },
		{ 34, 434, 39 },
		{ 34, 435, 39 },
		{ 34, 436, 39 },
		{ 34, 437, 39 },
		{ 34, 438, 39 },
		{ 34, 439, 39 },
		{ 34, 440, 39 },
		{ 34, 441, 39 },
		{ 34, 442, 39 },
		{ 34, 443, 39 },
		{ 34, 444, 39 },
		{ 34, 445, 39 },
		{ 34, 446, 39 },
		{ 34, 447, 39 },
		{ 34, 448, 39 },
		{ 34, 449, 39 },
		{ 34, 450, 39 },
		{ 34, 451, 39 },
		{ 34, 452, 39 },
		{ 34, 453, 39 },
		{ 34, 454, 39 },
		{ 34, 455, 39 },
		{ 34, 456, 39 },
		{ 34, 457, 39 },
		{ 34, 458, 39 },
		{ 34, 459, 39 },
		{ 34, 460, 39 },
		{ 34, 461, 39 },
		{ 34, 462, 39 },
		{ 34, 463, 39 },
		{ 34, 464, 39 },
		{ 34, 465, 39 },
		{ 34, 466, 39 },
		{ 34, 467, 39 },
		{ 34, 468, 39 },
		{ 34, 469, 39 },
		{ 34, 470, 39 },
		{ 34, 471, 39 },
		{ 34, 472, 39 },
		{ 34, 473, 39 },
		{ 34, 474, 39 },
		{ 34, 475, 39 },
		{ 34, 476, 39 },
		{ 34, 477, 39 },
		{ 34, 478, 39 },
		{ 34, 479, 39 },
		{ 35, 424, 39 },
		{ 35, 425, 39 },
		{ 35, 426, 39 },
		{ 35, 427, 39 },
		{ 35, 428, 39 },
		{ 35, 429, 39 },
		{ 35, 430, 39 },
		{ 35, 431, 39 },
		{ 35, 432, 39 },
		{ 35, 433, 39 },
		{ 35, 434, 39 },
		{ 35, 435, 39 },
		{ 35, 436, 39 },
		{ 35, 437, 39 },
		{ 35, 438, 39 },
		{ 35, 439, 39 },
		{ 35, 440, 39 },
		{ 35, 441, 39 },
		{ 35, 442, 39 },
		{ 35, 443, 39 },
		{ 35, 444, 39 },
		{ 35, 445, 39 },
		{ 35, 446, 39 },
		{ 35, 447, 39 },
		{ 35, 448, 39 },
		{ 35, 449, 39 },
		{ 35, 450, 39 },
		{ 35, 451, 39 },
		{ 35, 452, 39 },
		{ 35, 453, 39 },
		{ 35, 454, 39 },
		{ 35, 455, 39 },
		{ 35, 456, 39 },
		{ 35, 457, 39 },
		{ 35, 458, 39 },
		{ 35, 459, 39 },
		{ 35, 460, 39 },
		{ 35, 461, 39 },
		{ 35, 462, 39 },
		{ 35, 463, 39 },
		{ 35, 464, 39 },
		{ 35, 465, 39 },
		{ 35, 466, 39 },
		{ 35, 467, 39 },
		{ 35, 468, 39 },
		{ 35, 469, 39 },
		{ 35, 470, 39 },
		{ 35, 471, 39 },
		{ 35, 472, 39 },
		{ 35, 473, 39 },
		{ 35, 474, 39 },
		{ 35, 475, 39 },
		{ 35, 476, 39 },
		{ 35, 477, 39 },
		{ 35, 478, 39 },
		{ 36, 423, 39 },
		{ 36, 424, 39 },
		{ 36, 425, 39 },
		{ 36, 426, 39 },
		{ 36, 427, 39 },
		{ 36, 428, 39 },
		{ 36, 429, 39 },
		{ 36, 430, 39 },
		{ 36, 431, 39 },
		{ 36, 432, 39 },
		{ 36, 433, 39 },
		{ 36, 434, 39 },
		{ 36, 435, 39 },
		{ 36, 436, 39 },
		{ 36, 437, 39 },
		{ 36, 438, 39 },
		{ 36, 439, 39 },
		{ 36, 440, 39 },
		{ 36, 441, 39 },
		{ 36, 442, 39 },
		{ 36, 443, 39 },
		{ 36, 444, 39 },
		{ 36, 445, 39 },
		{ 36, 446, 39 },
		{ 36, 447, 39 },
		{ 36, 448, 39 },
		{ 36, 449, 39 },
		{ 36, 450, 39 },
		{ 36, 451, 39 },
		{ 36, 452, 39 },
		{ 36, 453, 39 },
		{ 36, 454, 39 },
		{ 36, 455, 39 },
		{ 36, 456, 39 },
		{ 36, 457, 39 },
		{ 36, 458, 39 },
		{ 36, 459, 39 },
		{ 36, 460, 39 },
		{ 36, 461, 39 },
		{ 36, 462, 39 },
		{ 36, 463, 39 },
		{ 36, 464, 39 },
		{ 36, 465, 39 },
		{ 36, 466, 39 },
		{ 36, 467, 39 },
		{ 36, 468, 39 },
		{ 36, 469, 39 },
		{ 36, 470, 39 },
		{ 36, 471, 39 },
		{ 36, 472, 39 },
		{ 36, 473, 39 },
		{ 36, 474, 39 },
		{ 36, 475, 39 },
		{ 36, 476, 39 },
		{ 36, 477, 39 },
		{ 37, 423, 39 },
		{ 37, 424, 39 },
		{ 37, 425, 39 },
		{ 37, 426, 39 },
		{ 37, 427, 39 },
		{ 37, 428, 39 },
		{ 37, 429, 39 },
		{ 37, 430, 39 },
		{ 37, 431, 39 },
		{ 37, 432, 39 },
		{ 37, 433, 39 },
		{ 37, 434, 39 },
		{ 37, 435, 39 },
		{ 37, 436, 39 },
		{ 37, 437, 39 },
		{ 37, 438, 39 },
		{ 37, 439, 39 },
		{ 37, 440, 39 },
		{ 37, 441, 39 },
		{ 37, 442, 39 },
		{ 37, 443, 39 },
		{ 37, 444, 39 },
		{ 37, 445, 39 },
		{ 37, 446, 39 },
		{ 37, 447, 39 },
		{ 37, 448, 39 },
		{ 37, 449, 39 },
		{ 37, 450, 39 },
		{ 37, 451, 39 },
		{ 37, 452, 39 },
		{ 37, 453, 39 },
		{ 37, 454, 39 },
		{ 37, 455, 39 },
		{ 37, 456, 39 },
		{ 37, 457, 39 },
		{ 37, 458, 39 },
		{ 37, 459, 39 },
		{ 37, 460, 39 },
		{ 37, 461, 39 },
		{ 37, 462, 39 },
		{ 37, 463, 39 },
		{ 37, 464, 39 },
		{ 37, 465, 39 },
		{ 37, 466, 39 },
		{ 37, 467, 39 },
		{ 37, 468, 39 },
		{ 37, 469, 39 },
		{ 37, 470, 39 },
		{ 37, 471, 39 },
		{ 37, 472, 39 },
		{ 37, 473, 39 },
		{ 37, 474, 39 },
		{ 37, 475, 39 },
		{ 37, 476, 39 },
		{ 38, 422, 39 },
		{ 38, 423, 39 },
		{ 38, 424, 39 },
		{ 38, 425, 39 },
		{ 38, 426, 39 },
		{ 38, 427, 39 },
		{ 38, 428, 39 },
		{ 38, 429, 39 },
		{ 38, 430, 39 },
		{ 38, 431, 39 },
		{ 38, 432, 39 },
		{ 38, 433, 39 },
		{ 38, 434, 39 },
		{ 38, 435, 39 },
		{ 38, 436, 39 },
		{ 38, 437, 39 },
		{ 38, 438, 39 },
		{ 38, 439, 39 },
		{ 38, 440, 39 },
		{ 38, 441, 39 },
		{ 38, 442, 39 },
		{ 38, 443, 39 },
		{ 38, 444, 39 },
		{ 38, 445, 39 },
		{ 38, 446, 39 },
		{ 38, 447, 39 },
		{ 38, 448, 39 },
		{ 38, 449, 39 },
		{ 38, 450, 39 },
		{ 38, 451, 39 },
		{ 38, 452, 39 },
		{ 38, 453, 39 },
		{ 38, 454, 39 },
		{ 38, 455, 39 },
		{ 38, 456, 39 },
		{ 38, 457, 39 },
		{ 38, 458, 39 },
		{ 38, 459, 39 },
		{ 38, 460, 39 },
		{ 38, 461, 39 },
		{ 38, 462, 39 },
		{ 38, 463, 39 },
		{ 38, 464, 39 },
		{ 38, 465, 39 },
		{ 38, 466, 39 },
		{ 38, 467, 39 },
		{ 38, 468, 39 },
		{ 38, 469, 39 },
		{ 38, 470, 39 },
		{ 38, 471, 39 },
		{ 38, 472, 39 },
		{ 38, 473, 39 },
		{ 38, 474, 39 },
		{ 38, 475, 39 },
		{ 39, 421, 39 },
		{ 39, 422, 39 },
		{ 39, 423, 39 },
		{ 39, 424, 39 },
		{ 39, 425, 39 },
		{ 39, 426, 39 },
		{ 39, 427, 39 },
		{ 39, 428, 39 },
		{ 39, 429, 39 },
		{ 39, 430, 39 },
		{ 39, 431, 39 },
		{ 39, 432, 39 },
		{ 39, 433, 39 },
		{ 39, 434, 39 },
		{ 39, 435, 39 },
		{ 39, 436, 39 },
		{ 39, 437, 39 },
		{ 39, 438, 39 },
		{ 39, 439, 39 },
		{ 39, 440, 39 },
		{ 39, 441, 39 },
		{ 39, 442, 39 },
		{ 39, 443, 39 },
		{ 39, 444, 39 },
		{ 39, 445, 39 },
		{ 39, 446, 39 },
		{ 39, 447, 39 },
		{ 39, 448, 39 },
		{ 39, 449, 39 },
		{ 39, 450, 39 },
		{ 39, 451, 39 },
		{ 39, 452, 39 },
		{ 39, 453, 39 },
		{ 39, 454, 39 },
		{ 39, 455, 39 },
		{ 39, 456, 39 },
		{ 39, 457, 39 },
		{ 39, 458, 39 },
		{ 39, 459, 39 },
		{ 39, 460, 39 },
		{ 39, 461, 39 },
		{ 39, 462, 39 },
		{ 39, 463, 39 },
		{ 39, 464, 39 },
		{ 39, 465, 39 },
		{ 39, 466, 39 },
		{ 39, 467, 39 },
		{ 39, 468, 39 },
		{ 39, 469, 39 },
		{ 39, 470, 39 },
		{ 39, 471, 39 },
		{ 39, 472, 39 },
		{ 39, 473, 39 },
		{ 39, 474, 39 },
		{ 40, 420, 39 },
		{ 40, 421, 39 },
		{ 40, 422, 39 },
		{ 40, 423, 39 },
		{ 40, 424, 39 },
		{ 40, 425, 39 },
		{ 40, 426, 39 },
		{ 40, 427, 39 },
		{ 40, 428, 39 },
		{ 40, 429, 39 },
		{ 40, 430, 39 },
		{ 40, 431, 39 },
		{ 40, 432, 39 },
		{ 40, 433, 39 },
		{ 40, 434, 39 },
		{ 40, 435, 39 },
		{ 40, 436, 39 },
		{ 40, 437, 39 },
		{ 40, 438, 39 },
		{ 40, 439, 39 },
		{ 40, 440, 39 },
		{ 40, 441, 39 },
		{ 40, 442, 39 },
		{ 40, 443, 39 },
		{ 40, 444, 39 },
		{ 40, 445, 39 },
		{ 40, 446, 39 },
		{ 40, 447, 39 },
		{ 40, 448, 39 },
		{ 40, 449, 39 },
		{ 40, 450, 39 },
		{ 40, 451, 39 },
		{ 40, 452, 39 },
		{ 40, 453, 39 },
		{ 40, 454, 39 },
		{ 40, 455, 39 },
		{ 40, 456, 39 },
		{ 40, 457, 39 },
		{ 40, 458, 39 },
		{ 40, 459, 39 },
		{ 40, 460, 39 },
		{ 40, 461, 39 },
		{ 40, 462, 39 },
		{ 40, 463, 39 },
		{ 40, 464, 39 },
		{ 40, 465, 39 },
		{ 40, 466, 39 },
		{ 40, 467, 39 },
		{ 40, 468, 39 },
		{ 40, 469, 39 },
		{ 40, 470, 39 },
		{ 40, 471, 39 },
		{ 40, 472, 39 },
		{ 40, 473, 39 },
		{ 41, 420, 39 },
		{ 41, 421, 39 },
		{ 41, 422, 39 },
		{ 41, 423, 39 },
		{ 41, 424, 39 },
		{ 41, 425, 39 },
		{ 41, 426, 39 },
		{ 41, 427, 39 },
		{ 41, 428, 39 },
		{ 41, 429, 39 },
		{ 41, 430, 39 },
		{ 41, 431, 39 },
		{ 41, 432, 39 },
		{ 41, 433, 39 },
		{ 41, 434, 39 },
		{ 41, 435, 39 },
		{ 41, 436, 39 },
		{ 41, 437, 39 },
		{ 41, 438, 39 },
		{ 41, 439, 39 },
		{ 41, 440, 39 },
		{ 41, 441, 39 },
		{ 41, 442, 39 },
		{ 41, 443, 39 },
		{ 41, 444, 39 },
		{ 41, 445, 39 },
		{ 41, 446, 39 },
		{ 41, 447, 39 },
		{ 41, 448, 39 },
		{ 41, 449, 39 },
		{ 41, 450, 39 },
		{ 41, 451, 39 },
		{ 41, 452, 39 },
		{ 41, 453, 39 },
		{ 41, 454, 39 },
		{ 41, 455, 39 },
		{ 41, 456, 39 },
		{ 41, 457, 39 },
		{ 41, 458, 39 },
		{ 41, 459, 39 },
		{ 41, 460, 39 },
		{ 41, 461, 39 },
		{ 41, 462, 39 },
		{ 41, 463, 39 },
		{ 41, 464, 39 },
		{ 41, 465, 39 },
		{ 41, 466, 39 },
		{ 41, 467, 39 },
		{ 41, 468, 39 },
		{ 41, 469, 39 },
		{ 41, 470, 39 },
		{ 41, 471, 39 },
		{ 41, 472, 39 },
		{ 42, 419, 39 },
		{ 42, 420, 39 },
		{ 42, 421, 39 },
		{ 42, 422, 39 },
		{ 42, 423, 39 },
		{ 42, 424, 39 },
		{ 42, 425, 39 },
		{ 42, 426, 39 },
		{ 42, 427, 39 },
		{ 42, 428, 39 },
		{ 42, 429, 39 },
		{ 42, 430, 39 },
		{ 42, 431, 39 },
		{ 42, 432, 39 },
		{ 42, 433, 39 },
		{ 42, 434, 39 },
		{ 42, 435, 39 },
		{ 42, 436, 39 },
		{ 42, 437, 39 },
		{ 42, 438, 39 },
		{ 42, 439, 39 },
		{ 42, 440, 39 },
		{ 42, 441, 39 },
		{ 42, 442, 39 },
		{ 42, 443, 39 },
		{ 42, 444, 39 },
		{ 42, 445, 39 },
		{ 42, 446, 39 },
		{ 42, 447, 39 },
		{ 42, 448, 39 },
		{ 42, 449, 39 },
		{ 42, 450, 39 },
		{ 42, 451, 39 },
		{ 42, 452, 39 },
		{ 42, 453, 39 },
		{ 42, 454, 39 },
		{ 42, 455, 39 },
		{ 42, 456, 39 },
		{ 42, 457, 39 },
		{ 42, 458, 39 },
		{ 42, 459, 39 },
		{ 42, 460, 39 },
		{ 42, 461, 39 },
		{ 42, 462, 39 },
		{ 42, 463, 39 },
		{ 42, 464, 39 },
		{ 42, 465, 39 },
		{ 42, 466, 39 },
		{ 42, 467, 39 },
		{ 42, 468, 39 },
		{ 42, 469, 39 },
		{ 42, 470, 39 },
		{ 42, 471, 39 },
		{ 43, 418, 39 },
		{ 43, 419, 39 },
		{ 43, 420, 39 },
		{ 43, 421, 39 },
		{ 43, 422, 39 },
		{ 43, 423, 39 },
		{ 43, 424, 39 },
		{ 43, 425, 39 },
		{ 43, 426, 39 },
		{ 43, 427, 39 },
		{ 43, 428, 39 },
		{ 43, 429, 39 },
		{ 43, 430, 39 },
		{ 43, 431, 39 },
		{ 43, 432, 39 },
		{ 43, 433, 39 },
		{ 43, 434, 39 },
		{ 43, 435, 39 },
		{ 43, 436, 39 },
		{ 43, 437, 39 },
		{ 43, 438, 39 },
		{ 43, 439, 39 },
		{ 43, 440, 39 },
		{ 43, 441, 39 },
		{ 43, 442, 39 },
		{ 43, 443, 39 },
		{ 43, 444, 39 },
		{ 43, 445, 39 },
		{ 43, 446, 39 },
		{ 43, 447, 39 },
		{ 43, 448, 39 },
		{ 43, 449, 39 },
		{ 43, 450, 39 },
		{ 43, 451, 39 },
		{ 43, 452, 39 },
		{ 43, 453, 39 },
		{ 43, 454, 39 },
		{ 43, 455, 39 },
		{ 43, 456, 39 },
		{ 43, 457, 39 },
		{ 43, 458, 39 },
		{ 43, 459, 39 },
		{ 43, 460, 39 },
		{ 43, 461, 39 },
		{ 43, 462, 39 },
		{ 43, 463, 39 },
		{ 43, 464, 39 },
		{ 43, 465, 39 },
		{ 43, 466, 39 },
		{ 43, 467, 39 },
		{ 43, 468, 39 },
		{ 43, 469, 39 },
		{ 43, 470, 39 },
		{ 44, 417, 39 },
		{ 44, 418, 39 },
		{ 44, 419, 39 },
		{ 44, 420, 39 },
		{ 44, 421, 39 },
		{ 44, 422, 39 },
		{ 44, 423, 39 },
		{ 44, 424, 39 },
		{ 44, 425, 39 },
		{ 44, 426, 39 },
		{ 44, 427, 39 },
		{ 44, 428, 39 },
		{ 44, 429, 39 },
		{ 44, 430, 39 },
		{ 44, 431, 39 },
		{ 44, 432, 39 },
		{ 44, 433, 39 },
		{ 44, 434, 39 },
		{ 44, 435, 39 },
		{ 44, 436, 39 },
		{ 44, 437, 39 },
		{ 44, 438, 39 },
		{ 44, 439, 39 },
		{ 44, 440, 39 },
		{ 44, 441, 39 },
		{ 44, 442, 39 },
		{ 44, 443, 39 },
		{ 44, 444, 39 },
		{ 44, 445, 39 },
		{ 44, 446, 39 },
		{ 44, 447, 39 },
		{ 44, 448, 39 },
		{ 44, 449, 39 },
		{ 44, 450, 39 },
		{ 44, 451, 39 },
		{ 44, 452, 39 },
		{ 44, 453, 39 },
		{ 44, 454, 39 },
		{ 44, 455, 39 },
		{ 44, 456, 39 },
		{ 44, 457, 39 },
		{ 44, 458, 39 },
		{ 44, 459, 39 },
		{ 44, 460, 39 },
		{ 44, 461, 39 },
		{ 44, 462, 39 },
		{ 44, 463, 39 },
		{ 44, 464, 39 },
		{ 44, 465, 39 },
		{ 44, 466, 39 },
		{ 44, 467, 39 },
		{ 44, 468, 39 },
		{ 44, 469, 39 },
		{ 45, 417, 39 },
		{ 45, 418, 39 },
		{ 45, 419, 39 },
		{ 45, 420, 39 },
		{ 45, 421, 39 },
		{ 45, 422, 39 },
		{ 45, 423, 39 },
		{ 45, 424, 39 },
		{ 45, 425, 39 },
		{ 45, 426, 39 },
		{ 45, 427, 39 },
		{ 45, 428, 39 },
		{ 45, 429, 39 },
		{ 45, 430, 39 },
		{ 45, 431, 39 },
		{ 45, 432, 39 },
		{ 45, 433, 39 },
		{ 45, 434, 39 },
		{ 45, 435, 39 },
		{ 45, 436, 39 },
		{ 45, 437, 39 },
		{ 45, 438, 39 },
		{ 45, 439, 39 },
		{ 45, 440, 39 },
		{ 45, 441, 39 },
		{ 45, 442, 39 },
		{ 45, 443, 39 },
		{ 45, 444, 39 },
		{ 45, 445, 39 },
		{ 45, 446, 39 },
		{ 45, 447, 39 },
		{ 45, 448, 39 },
		{ 45, 449, 39 },
		{ 45, 450, 39 },
		{ 45, 451, 39 },
		{ 45, 452, 39 },
		{ 45, 453, 39 },
		{ 45, 454, 39 },
		{ 45, 455, 39 },
		{ 45, 456, 39 },
		{ 45, 457, 39 },
		{ 45, 458, 39 },
		{ 45, 459, 39 },
		{ 45, 460, 39 },
		{ 45, 461, 39 },
		{ 45, 462, 39 },
		{ 45, 463, 39 },
		{ 45, 464, 39 },
		{ 45, 465, 39 },
		{ 45, 466, 39 },
		{ 45, 467, 39 },
		{ 45, 468, 39 },
		{ 46, 416, 39 },
		{ 46, 417, 39 },
		{ 46, 418, 39 },
		{ 46, 419, 39 },
		{ 46, 420, 39 },
		{ 46, 421, 39 },
		{ 46, 422, 39 },
		{ 46, 423, 39 },
		{ 46, 424, 39 },
		{ 46, 425, 39 },
		{ 46, 426, 39 },
		{ 46, 427, 39 },
		{ 46, 428, 39 },
		{ 46, 429, 39 },
		{ 46, 430, 39 },
		{ 46, 431, 39 },
		{ 46, 432, 39 },
		{ 46, 433, 39 },
		{ 46, 434, 39 },
		{ 46, 435, 39 },
		{ 46, 436, 39 },
		{ 46, 437, 39 },
		{ 46, 438, 39 },
		{ 46, 439, 39 },
		{ 46, 440, 39 },
		{ 46, 441, 39 },
		{ 46, 442, 39 },
		{ 46, 443, 39 },
		{ 46, 444, 39 },
		{ 46, 445, 39 },
		{ 46, 446, 39 },
		{ 46, 447, 39 },
		{ 46, 448, 39 },
		{ 46, 449, 39 },
		{ 46, 450, 39 },
		{ 46, 451, 39 },
		{ 46, 452, 39 },
		{ 46, 453, 39 },
		{ 46, 454, 39 },
		{ 46, 455, 39 },
		{ 46, 456, 39 },
		{ 46, 457, 39 },
		{ 46, 458, 39 },
		{ 46, 459, 39 },
		{ 46, 460, 39 },
		{ 46, 461, 39 },
		{ 46, 462, 39 },
		{ 46, 463, 39 },
		{ 46, 464, 39 },
		{ 46, 465, 39 },
		{ 46, 466, 39 },
		{ 46, 467, 39 },
		{ 47, 415, 39 },
		{ 47, 416, 39 },
		{ 47, 417, 39 },
		{ 47, 418, 39 },
		{ 47, 419, 39 },
		{ 47, 420, 39 },
		{ 47, 421, 39 },
		{ 47, 422, 39 },
		{ 47, 423, 39 },
		{ 47, 424, 39 },
		{ 47, 425, 39 },
		{ 47, 426, 39 },
		{ 47, 427, 39 },
		{ 47, 428, 39 },
		{ 47, 429, 39 },
		{ 47, 430, 39 },
		{ 47, 431, 39 },
		{ 47, 432, 39 },
		{ 47, 433, 39 },
		{ 47, 434, 39 },
		{ 47, 435, 39 },
		{ 47, 436, 39 },
		{ 47, 437, 39 },
		{ 47, 438, 39 },
		{ 47, 439, 39 },
		{ 47, 440, 39 },
		{ 47, 441, 39 },
		{ 47, 442, 39 },
		{ 47, 443, 39 },
		{ 47, 444, 39 },
		{ 47, 445, 39 },
		{ 47, 446, 39 },
		{ 47, 447, 39 },
		{ 47, 448, 39 },
		{ 47, 449, 39 },
		{ 47, 450, 39 },
		{ 47, 451, 39 },
		{ 47, 452, 39 },
		{ 47, 453, 39 },
		{ 47, 454, 39 },
		{ 47, 455, 39 },
		{ 47, 456, 39 },
		{ 47, 457, 39 },
		{ 47, 458, 39 },
		{ 47, 459, 39 },
		{ 47, 460, 39 },
		{ 47, 461, 39 },
		{ 47, 462, 39 },
		{ 47, 463, 39 },
		{ 47, 464, 39 },
		{ 47, 465, 39 },
		{ 47, 466, 39 },
		{ 48, 414, 39 },
		{ 48, 415, 39 },
		{ 48, 416, 39 },
		{ 48, 417, 39 },
		{ 48, 418, 39 },
		{ 48, 419, 39 },
		{ 48, 420, 39 },
		{ 48, 421, 39 },
		{ 48, 422, 39 },
		{ 48, 423, 39 },
		{ 48, 424, 39 },
		{ 48, 425, 39 },
		{ 48, 426, 39 },
		{ 48, 427, 39 },
		{ 48, 428, 39 },
		{ 48, 429, 39 },
		{ 48, 430, 39 },
		{ 48, 431, 39 },
		{ 48, 432, 39 },
		{ 48, 433, 39 },
		{ 48, 434, 39 },
		{ 48, 435, 39 },
		{ 48, 436, 39 },
		{ 48, 437, 39 },
		{ 48, 438, 39 },
		{ 48, 439, 39 },
		{ 48, 440, 39 },
		{ 48, 441, 39 },
		{ 48, 442, 39 },
		{ 48, 443, 39 },
		{ 48, 444, 39 },
		{ 48, 445, 39 },
		{ 48, 446, 39 },
		{ 48, 447, 39 },
		{ 48, 448, 39 },
		{ 48, 449, 39 },
		{ 48, 450, 39 },
		{ 48, 451, 39 },
		{ 48, 452, 39 },
		{ 48, 453, 39 },
		{ 48, 454, 39 },
		{ 48, 455, 39 },
		{ 48, 456, 39 },
		{ 48, 457, 39 },
		{ 48, 458, 39 },
		{ 48, 459, 39 },
		{ 48, 460, 39 },
		{ 48, 461, 39 },
		{ 48, 462, 39 },
		{ 48, 463, 39 },
		{ 48, 464, 39 },
		{ 48, 465, 39 },
		{ 49, 414, 39 },
		{ 49, 415, 39 },
		{ 49, 416, 39 },
		{ 49, 417, 39 },
		{ 49, 418, 39 },
		{ 49, 419, 39 },
		{ 49, 420, 39 },
		{ 49, 421, 39 },
		{ 49, 422, 39 },
		{ 49, 423, 39 },
		{ 49, 424, 39 },
		{ 49, 425, 39 },
		{ 49, 426, 39 },
		{ 49, 427, 39 },
		{ 49, 428, 39 },
		{ 49, 429, 39 },
		{ 49, 430, 39 },
		{ 49, 431, 39 },
		{ 49, 432, 39 },
		{ 49, 433, 39 },
		{ 49, 434, 39 },
		{ 49, 435, 39 },
		{ 49, 436, 39 },
		{ 49, 437, 39 },
		{ 49, 438, 39 },
		{ 49, 439, 39 },
		{ 49, 440, 39 },
		{ 49, 441, 39 },
		{ 49, 442, 39 },
		{ 49, 443, 39 },
		{ 49, 444, 39 },
		{ 49, 445, 39 },
		{ 49, 446, 39 },
		{ 49, 447, 39 },
		{ 49, 448, 39 },
		{ 49, 449, 39 },
		{ 49, 450, 39 },
		{ 49, 451, 39 },
		{ 49, 452, 39 },
		{ 49, 453, 39 },
		{ 49, 454, 39 },
		{ 49, 455, 39 },
		{ 49, 456, 39 },
		{ 49, 457, 39 },
		{ 49, 458, 39 },
		{ 49, 459, 39 },
		{ 49, 460, 39 },
		{ 49, 461, 39 },
		{ 49, 462, 39 },
		{ 49, 463, 39 },
		{ 49, 464, 39 },
		{ 50, 413, 39 },
		{ 50, 414, 39 },
		{ 50, 415, 39 },
		{ 50, 416, 39 },
		{ 50, 417, 39 },
		{ 50, 418, 39 },
		{ 50, 419, 39 },
		{ 50, 420, 39 },
		{ 50, 421, 39 },
		{ 50, 422, 39 },
		{ 50, 423, 39 },
		{ 50, 424, 39 },
		{ 50, 425, 39 },
		{ 50, 426, 39 },
		{ 50, 427, 39 },
		{ 50, 428, 39 },
		{ 50, 429, 39 },
		{ 50, 430, 39 },
		{ 50, 431, 39 },
		{ 50, 432, 39 },
		{ 50, 433, 39 },
		{ 50, 434, 39 },
		{ 50, 435, 39 },
		{ 50, 436, 39 },
		{ 50, 437, 39 },
		{ 50, 438, 39 },
		{ 50, 439, 39 },
		{ 50, 440, 39 },
		{ 50, 441, 39 },
		{ 50, 442, 39 },
		{ 50, 443, 39 },
		{ 50, 444, 39 },
		{ 50, 445, 39 },
		{ 50, 446, 39 },
		{ 50, 447, 39 },
		{ 50, 448, 39 },
		{ 50, 449, 39 },
		{ 50, 450, 39 },
		{ 50, 451, 39 },
		{ 50, 452, 39 },
		{ 50, 453, 39 },
		{ 50, 454, 39 },
		{ 50, 455, 39 },
		{ 50, 456, 39 },
		{ 50, 457, 39 },
		{ 50, 458, 39 },
		{ 50, 459, 39 },
		{ 50, 460, 39 },
		{ 50, 461, 39 },
		{ 50, 462, 39 },
		{ 50, 463, 39 },
		{ 51, 412, 39 },
		{ 51, 413, 39 },
		{ 51, 414, 39 },
		{ 51, 415, 39 },
		{ 51, 416, 39 },
		{ 51, 417, 39 },
		{ 51, 418, 39 },
		{ 51, 419, 39 },
		{ 51, 420, 39 },
		{ 51, 421, 39 },
		{ 51, 422, 39 },
		{ 51, 423, 39 },
		{ 51, 424, 39 },
		{ 51, 425, 39 },
		{ 51, 426, 39 },
		{ 51, 427, 39 },
		{ 51, 428, 39 },
		{ 51, 429, 39 },
		{ 51, 430, 39 },
		{ 51, 431, 39 },
		{ 51, 432, 39 },
		{ 51, 433, 39 },
		{ 51, 434, 39 },
		{ 51, 435, 39 },
		{ 51, 436, 39 },
		{ 51, 437, 39 },
		{ 51, 438, 39 },
		{ 51, 439, 39 },
		{ 51, 440, 39 },
		{ 51, 441, 39 },
		{ 51, 442, 39 },
		{ 51, 443, 39 },
		{ 51, 444, 39 },
		{ 51, 445, 39 },
		{ 51, 446, 39 },
		{ 51, 447, 39 },
		{ 51, 448, 39 },
		{ 51, 449, 39 },
		{ 51, 450, 39 },
		{ 51, 451, 39 },
		{ 51, 452, 39 },
		{ 51, 453, 39 },
		{ 51, 454, 39 },
		{ 51, 455, 39 },
		{ 51, 456, 39 },
		{ 51, 457, 39 },
		{ 51, 458, 39 },
		{ 51, 459, 39 },
		{ 51, 460, 39 },
		{ 51, 461, 39 },
		{ 51, 462, 39 },
		{ 52, 411, 39 },
		{ 52, 412, 39 },
		{ 52, 413, 39 },
		{ 52, 414, 39 },
		{ 52, 415, 39 },
		{ 52, 416, 39 },
		{ 52, 417, 39 },
		{ 52, 418, 39 },
		{ 52, 419, 39 },
		{ 52, 420, 39 },
		{ 52, 421, 39 },
		{ 52, 422, 39 },
		{ 52, 423, 39 },
		{ 52, 424, 39 },
		{ 52, 425, 39 },
		{ 52, 426, 39 },
		{ 52, 427, 39 },
		{ 52, 428, 39 },
		{ 52, 429, 39 },
		{ 52, 430, 39 },
		{ 52, 431, 39 },
		{ 52, 432, 39 },
		{ 52, 433, 39 },
		{ 52, 434, 39 },
		{ 52, 435, 39 },
		{ 52, 436, 39 },
		{ 52, 437, 39 },
		{ 52, 438, 39 },
		{ 52, 439, 39 },
		{ 52, 440, 39 },
		{ 52, 441, 39 },
		{ 52, 442, 39 },
		{ 52, 443, 39 },
		{ 52, 444, 39 },
		{ 52, 445, 39 },
		{ 52, 446, 39 },
		{ 52, 447, 39 },
		{ 52, 448, 39 },
		{ 52, 449, 39 },
		{ 52, 450, 39 },
		{ 52, 451, 39 },
		{ 52, 452, 39 },
		{ 52, 453, 39 },
		{ 52, 454, 39 },
		{ 52, 455, 39 },
		{ 52, 456, 39 },
		{ 52, 457, 39 },
		{ 52, 458, 39 },
		{ 52, 459, 39 },
		{ 52, 460, 39 },
		{ 52, 461, 39 },
		{ 53, 411, 39 },
		{ 53, 412, 39 },
		{ 53, 413, 39 },
		{ 53, 414, 39 },
		{ 53, 415, 39 },
		{ 53, 416, 39 },
		{ 53, 417, 39 },
		{ 53, 418, 39 },
		{ 53, 419, 39 },
		{ 53, 420, 39 },
		{ 53, 421, 39 },
		{ 53, 422, 39 },
		{ 53, 423, 39 },
		{ 53, 424, 39 },
		{ 53, 425, 39 },
		{ 53, 426, 39 },
		{ 53, 427, 39 },
		{ 53, 428, 39 },
		{ 53, 429, 39 },
		{ 53, 430, 39 },
		{ 53, 431, 39 },
		{ 53, 432, 39 },
		{ 53, 433, 39 },
		{ 53, 434, 39 },
		{ 53, 435, 39 },
		{ 53, 436, 39 },
		{ 53, 437, 39 },
		{ 53, 438, 39 },
		{ 53, 439, 39 },
		{ 53, 440, 39 },
		{ 53, 441, 39 },
		{ 53, 442, 39 },
		{ 53, 443, 39 },
		{ 53, 444, 39 },
		{ 53, 445, 39 },
		{ 53, 446, 39 },
		{ 53, 447, 39 },
		{ 53, 448, 39 },
		{ 53, 449, 39 },
		{ 53, 450, 39 },
		{ 53, 451, 39 },
		{ 53, 452, 39 },
		{ 53, 453, 39 },
		{ 53, 454, 39 },
		{ 53, 455, 39 },
		{ 53, 456, 39 },
		{ 53, 457, 39 },
		{ 53, 458, 39 },
		{ 53, 459, 39 },
		{ 53, 460, 39 },
		{ 54, 410, 39 },
		{ 54, 411, 39 },
		{ 54, 412, 39 },
		{ 54, 413, 39 },
		{ 54, 414, 39 },
		{ 54, 415, 39 },
		{ 54, 416, 39 },
		{ 54, 417, 39 },
		{ 54, 418, 39 },
		{ 54, 419, 39 },
		{ 54, 420, 39 },
		{ 54, 421, 39 },
		{ 54, 422, 39 },
		{ 54, 423, 39 },
		{ 54, 424, 39 },
		{ 54, 425, 39 },
		{ 54, 426, 39 },
		{ 54, 427, 39 },
		{ 54, 428, 39 },
		{ 54, 429, 39 },
		{ 54, 430, 39 },
		{ 54, 431, 39 },
		{ 54, 432, 39 },
		{ 54, 433, 39 },
		{ 54, 434, 39 },
		{ 54, 435, 39 },
		{ 54, 436, 39 },
		{ 54, 437, 39 },
		{ 54, 438, 39 },
		{ 54, 439, 39 },
		{ 54, 440, 39 },
		{ 54, 441, 39 },
		{ 54, 442, 39 },
		{ 54, 443, 39 },
		{ 54, 444, 39 },
		{ 54, 445, 39 },
		{ 54, 446, 39 },
		{ 54, 447, 39 },
		{ 54, 448, 39 },
		{ 54, 449, 39 },
		{ 54, 450, 39 },
		{ 54, 451, 39 },
		{ 54, 452, 39 },
		{ 54, 453, 39 },
		{ 54, 454, 39 },
		{ 54, 455, 39 },
		{ 54, 456, 39 },
		{ 54, 457, 39 },
		{ 54, 458, 39 },
		{ 54, 459, 39 },
		{ 55, 409, 39 },
		{ 55, 410, 39 },
		{ 55, 411, 39 },
		{ 55, 412, 39 },
		{ 55, 413, 39 },
		{ 55, 414, 39 },
		{ 55, 415, 39 },
		{ 55, 416, 39 },
		{ 55, 417, 39 },
		{ 55, 418, 39 },
		{ 55, 419, 39 },
		{ 55, 420, 39 },
		{ 55, 421, 39 },
		{ 55, 422, 39 },
		{ 55, 423, 39 },
		{ 55, 424, 39 },
		{ 55, 425, 39 },
		{ 55, 426, 39 },
		{ 55, 427, 39 },
		{ 55, 428, 39 },
		{ 55, 429, 39 },
		{ 55, 430, 39 },
		{ 55, 431, 39 },
		{ 55, 432, 39 },
		{ 55, 433, 39 },
		{ 55, 434, 39 },
		{ 55, 435, 39 },
		{ 55, 436, 39 },
		{ 55, 437, 39 },
		{ 55, 438, 39 },
		{ 55, 439, 39 },
		{ 55, 440, 39 },
		{ 55, 441, 39 },
		{ 55, 442, 39 },
		{ 55, 443, 39 },
		{ 55, 444, 39 },
		{ 55, 445, 39 },
		{ 55, 446, 39 },
		{ 55, 447, 39 },
		{ 55, 448, 39 },
		{ 55, 449, 39 },
		{ 55, 450, 39 },
		{ 55, 451, 39 },
		{ 55, 452, 39 },
		{ 55, 453, 39 },
		{ 55, 454, 39 },
		{ 55, 455, 39 },
		{ 55, 456, 39 },
		{ 55, 457, 39 },
		{ 55, 458, 39 },
		{ 56, 408, 39 },
		{ 56, 409, 39 },
		{ 56, 410, 39 },
		{ 56, 411, 39 },
		{ 56, 412, 39 },
		{ 56, 413, 39 },
		{ 56, 414, 39 },
		{ 56, 415, 39 },
		{ 56, 416, 39 },
		{ 56, 417, 39 },
		{ 56, 418, 39 },
		{ 56, 419, 39 },
		{ 56, 420, 39 },
		{ 56, 421, 39 },
		{ 56, 422, 39 },
		{ 56, 423, 39 },
		{ 56, 424, 39 },
		{ 56, 425, 39 },
		{ 56, 426, 39 },
		{ 56, 427, 39 },
		{ 56, 428, 39 },
		{ 56, 429, 39 },
		{ 56, 430, 39 },
		{ 56, 431, 39 },
		{ 56, 432, 39 },
		{ 56, 433, 39 },
		{ 56, 434, 39 },
		{ 56, 435, 39 },
		{ 56, 436, 39 },
		{ 56, 437, 39 },
		{ 56, 438, 39 },
		{ 56, 439, 39 },
		{ 56, 440, 39 },
		{ 56, 441, 39 },
		{ 56, 442, 39 },
		{ 56, 443, 39 },
		{ 56, 444, 39 },
		{ 56, 445, 39 },
		{ 56, 446, 39 },
		{ 56, 447, 39 },
		{ 56, 448, 39 },
		{ 56, 449, 39 },
		{ 56, 450, 39 },
		{ 56, 451, 39 },
		{ 56, 452, 39 },
		{ 56, 453, 39 },
		{ 56, 454, 39 },
		{ 56, 455, 39 },
		{ 56, 456, 39 },
		{ 56, 457, 39 },
		{ 57, 408, 39 },
		{ 57, 409, 39 },
		{ 57, 410, 39 },
		{ 57, 411, 39 },
		{ 57, 412, 39 },
		{ 57, 413, 39 },
		{ 57, 414, 39 },
		{ 57, 415, 39 },
		{ 57, 416, 39 },
		{ 57, 417, 39 },
		{ 57, 418, 39 },
		{ 57, 419, 39 },
		{ 57, 420, 39 },
		{ 57, 421, 39 },
		{ 57, 422, 39 },
		{ 57, 423, 39 },
		{ 57, 424, 39 },
		{ 57, 425, 39 },
		{ 57, 426, 39 },
		{ 57, 427, 39 },
		{ 57, 428, 39 },
		{ 57, 429, 39 },
		{ 57, 430, 39 },
		{ 57, 431, 39 },
		{ 57, 432, 39 },
		{ 57, 433, 39 },
		{ 57, 434, 39 },
		{ 57, 435, 39 },
		{ 57, 436, 39 },
		{ 57, 437, 39 },
		{ 57, 438, 39 },
		{ 57, 439, 39 },
		{ 57, 440, 39 },
		{ 57, 441, 39 },
		{ 57, 442, 39 },
		{ 57, 443, 39 },
		{ 57, 444, 39 },
		{ 57, 445, 39 },
		{ 57, 446, 39 },
		{ 57, 447, 39 },
		{ 57, 448, 39 },
		{ 57, 449, 39 },
		{ 57, 450, 39 },
		{ 57, 451, 39 },
		{ 57, 452, 39 },
		{ 57, 453, 39 },
		{ 57, 454, 39 },
		{ 57, 455, 39 },
		{ 57, 456, 39 },
		{ 58, 407, 39 },
		{ 58, 408, 39 },
		{ 58, 409, 39 },
		{ 58, 410, 39 },
		{ 58, 411, 39 },
		{ 58, 412, 39 },
		{ 58, 413, 39 },
		{ 58, 414, 39 },
		{ 58, 415, 39 },
		{ 58, 416, 39 },
		{ 58, 417, 39 },
		{ 58, 418, 39 },
		{ 58, 419, 39 },
		{ 58, 420, 39 },
		{ 58, 421, 39 },
		{ 58, 422, 39 },
		{ 58, 423, 39 },
		{ 58, 424, 39 },
		{ 58, 425, 39 },
		{ 58, 426, 39 },
		{ 58, 427, 39 },
		{ 58, 428, 39 },
		{ 58, 429, 39 },
		{ 58, 430, 39 },
		{ 58, 431, 39 },
		{ 58, 432, 39 },
		{ 58, 433, 39 },
		{ 58, 434, 39 },
		{ 58, 435, 39 },
		{ 58, 436, 39 },
		{ 58, 437, 39 },
		{ 58, 438, 39 },
		{ 58, 439, 39 },
		{ 58, 440, 39 },
		{ 58, 441, 39 },
		{ 58, 442, 39 },
		{ 58, 443, 39 },
		{ 58, 444, 39 },
		{ 58, 445, 39 },
		{ 58, 446, 39 },
		{ 58, 447, 39 },
		{ 58, 448, 39 },
		{ 58, 449, 39 },
		{ 58, 450, 39 },
		{ 58, 451, 39 },
		{ 58, 452, 39 },
		{ 58, 453, 39 },
		{ 58, 454, 39 },
		{ 58, 455, 39 },
		{ 59, 406, 39 },
		{ 59, 407, 39 },
		{ 59, 408, 39 },
		{ 59, 409, 39 },
		{ 59, 410, 39 },
		{ 59, 411, 39 },
		{ 59, 412, 39 },
		{ 59, 413, 39 },
		{ 59, 414, 39 },
		{ 59, 415, 39 },
		{ 59, 416, 39 },
		{ 59, 417, 39 },
		{ 59, 418, 39 },
		{ 59, 419, 39 },
		{ 59, 420, 39 },
		{ 59, 421, 39 },
		{ 59, 422, 39 },
		{ 59, 423, 39 },
		{ 59, 424, 39 },
		{ 59, 425, 39 },
		{ 59, 426, 39 },
		{ 59, 427, 39 },
		{ 59, 428, 39 },
		{ 59, 429, 39 },
		{ 59, 430, 39 },
		{ 59, 431, 39 },
		{ 59, 432, 39 },
		{ 59, 433, 39 },
		{ 59, 434, 39 },
		{ 59, 435, 39 },
		{ 59, 436, 39 },
		{ 59, 437, 39 },
		{ 59, 438, 39 },
		{ 59, 439, 39 },
		{ 59, 440, 39 },
		{ 59, 441, 39 },
		{ 59, 442, 39 },
		{ 59, 443, 39 },
		{ 59, 444, 39 },
		{ 59, 445, 39 },
		{ 59, 446, 39 },
		{ 59, 447, 39 },
		{ 59, 448, 39 },
		{ 59, 449, 39 },
		{ 59, 450, 39 },
		{ 59, 451, 39 },
		{ 59, 452, 39 },
		{ 59, 453, 39 },
		{ 59, 454, 39 },
		{ 60, 405, 39 },
		{ 60, 406, 39 },
		{ 60, 407, 39 },
		{ 60, 408, 39 },
		{ 60, 409, 39 },
		{ 60, 410, 39 },
		{ 60, 411, 39 },
		{ 60, 412, 39 },
		{ 60, 413, 39 },
		{ 60, 414, 39 },
		{ 60, 415, 39 },
		{ 60, 416, 39 },
		{ 60, 417, 39 },
		{ 60, 418, 39 },
		{ 60, 419, 39 },
		{ 60, 420, 39 },
		{ 60, 421, 39 },
		{ 60, 422, 39 },
		{ 60, 423, 39 },
		{ 60, 424, 39 },
		{ 60, 425, 39 },
		{ 60, 426, 39 },
		{ 60, 427, 39 },
		{ 60, 428, 39 },
		{ 60, 429, 39 },
		{ 60, 430, 39 },
		{ 60, 431, 39 },
		{ 60, 432, 39 },
		{ 60, 433, 39 },
		{ 60, 434, 39 },
		{ 60, 435, 39 },
		{ 60, 436, 39 },
		{ 60, 437, 39 },
		{ 60, 438, 39 },
		{ 60, 439, 39 },
		{ 60, 440, 39 },
		{ 60, 441, 39 },
		{ 60, 442, 39 },
		{ 60, 443, 39 },
		{ 60, 444, 39 },
		{ 60, 445, 39 },
		{ 60, 446, 39 },
		{ 60, 447, 39 },
		{ 60, 448, 39 },
		{ 60, 449, 39 },
		{ 60, 450, 39 },
		{ 60, 451, 39 },
		{ 60, 452, 39 },
		{ 60, 453, 39 },
		{ 61, 405, 39 },
		{ 61, 406, 39 },
		{ 61, 407, 39 },
		{ 61, 408, 39 },
		{ 61, 409, 39 },
		{ 61, 410, 39 },
		{ 61, 411, 39 },
		{ 61, 412, 39 },
		{ 61, 413, 39 },
		{ 61, 414, 39 },
		{ 61, 415, 39 },
		{ 61, 416, 39 },
		{ 61, 417, 39 },
		{ 61, 418, 39 },
		{ 61, 419, 39 },
		{ 61, 420, 39 },
		{ 61, 421, 39 },
		{ 61, 422, 39 },
		{ 61, 423, 39 },
		{ 61, 424, 39 },
		{ 61, 425, 39 },
		{ 61, 426, 39 },
		{ 61, 427, 39 },
		{ 61, 428, 39 },
		{ 61, 429, 39 },
		{ 61, 430, 39 },
		{ 61, 431, 39 },
		{ 61, 432, 39 },
		{ 61, 433, 39 },
		{ 61, 434, 39 },
		{ 61, 435, 39 },
		{ 61, 436, 39 },
		{ 61, 437, 39 },
		{ 61, 438, 39 },
		{ 61, 439, 39 },
		{ 61, 440, 39 },
		{ 61, 441, 39 },
		{ 61, 442, 39 },
		{ 61, 443, 39 },
		{ 61, 444, 39 },
		{ 61, 445, 39 },
		{ 61, 446, 39 },
		{ 61, 447, 39 },
		{ 61, 448, 39 },
		{ 61, 449, 39 },
		{ 61, 450, 39 },
		{ 61, 451, 39 },
		{ 61, 452, 39 },
		{ 62, 404, 39 },
		{ 62, 405, 39 },
		{ 62, 406, 39 },
		{ 62, 407, 39 },
		{ 62, 408, 39 },
		{ 62, 409, 39 },
		{ 62, 410, 39 },
		{ 62, 411, 39 },
		{ 62, 412, 39 },
		{ 62, 413, 39 },
		{ 62, 414, 39 },
		{ 62, 415, 39 },
		{ 62, 416, 39 },
		{ 62, 417, 39 },
		{ 62, 418, 39 },
		{ 62, 419, 39 },
		{ 62, 420, 39 },
		{ 62, 421, 39 },
		{ 62, 422, 39 },
		{ 62, 423, 39 },
		{ 62, 424, 39 },
		{ 62, 425, 39 },
		{ 62, 426, 39 },
		{ 62, 427, 39 },
		{ 62, 428, 39 },
		{ 62, 429, 39 },
		{ 62, 430, 39 },
		{ 62, 431, 39 },
		{ 62, 432, 39 },
		{ 62, 433, 39 },
		{ 62, 434, 39 },
		{ 62, 435, 39 },
		{ 62, 436, 39 },
		{ 62, 437, 39 },
		{ 62, 438, 39 },
		{ 62, 439, 39 },
		{ 62, 440, 39 },
		{ 62, 441, 39 },
		{ 62, 442, 39 },
		{ 62, 443, 39 },
		{ 62, 444, 39 },
		{ 62, 445, 39 },
		{ 62, 446, 39 },
		{ 62, 447, 39 },
		{ 62, 448, 39 },
		{ 62, 449, 39 },
		{ 62, 450, 39 },
		{ 62, 451, 39 },
		{ 63, 403, 39 },
		{ 63, 404, 39 },
		{ 63, 405, 39 },
		{ 63, 406, 39 },
		{ 63, 407, 39 },
		{ 63, 408, 39 },
		{ 63, 409, 39 },
		{ 63, 410, 39 },
		{ 63, 411, 39 },
		{ 63, 412, 39 },
		{ 63, 413, 39 },
		{ 63, 414, 39 },
		{ 63, 415, 39 },
		{ 63, 416, 39 },
		{ 63, 417, 39 },
		{ 63, 418, 39 },
		{ 63, 419, 39 },
		{ 63, 420, 39 },
		{ 63, 421, 39 },
		{ 63, 422, 39 },
		{ 63, 423, 39 },
		{ 63, 424, 39 },
		{ 63, 425, 39 },
		{ 63, 426, 39 },
		{ 63, 427, 39 },
		{ 63, 428, 39 },
		{ 63, 429, 39 },
		{ 63, 430, 39 },
		{ 63, 431, 39 },
		{ 63, 432, 39 },
		{ 63, 433, 39 },
		{ 63, 434, 39 },
		{ 63, 435, 39 },
		{ 63, 436, 39 },
		{ 63, 437, 39 },
		{ 63, 438, 39 },
		{ 63, 439, 39 },
		{ 63, 440, 39 },
		{ 63, 441, 39 },
		{ 63, 442, 39 },
		{ 63, 443, 39 },
		{ 63, 444, 39 },
		{ 63, 445, 39 },
		{ 63, 446, 39 },
		{ 63, 447, 39 },
		{ 63, 448, 39 },
		{ 63, 449, 39 },
		{ 63, 450, 39 },
		{ 64, 402, 39 },
		{ 64, 403, 39 },
		{ 64, 404, 39 },
		{ 64, 405, 39 },
		{ 64, 406, 39 },
		{ 64, 407, 39 },
		{ 64, 408, 39 },
		{ 64, 409, 39 },
		{ 64, 410, 39 },
		{ 64, 411, 39 },
		{ 64, 412, 39 },
		{ 64, 413, 39 },
		{ 64, 414, 39 },
		{ 64, 415, 39 },
		{ 64, 416, 39 },
		{ 64, 417, 39 },
		{ 64, 418, 39 },
		{ 64, 419, 39 },
		{ 64, 420, 39 },
		{ 64, 421, 39 },
		{ 64, 422, 39 },
		{ 64, 423, 39 },
		{ 64, 424, 39 },
		{ 64, 425, 39 },
		{ 64, 426, 39 },
		{ 64, 427, 39 },
		{ 64, 428, 39 },
		{ 64, 429, 39 },
		{ 64, 430, 39 },
		{ 64, 431, 39 },
		{ 64, 432, 39 },
		{ 64, 433, 39 },
		{ 64, 434, 39 },
		{ 64, 435, 39 },
		{ 64, 436, 39 },
		{ 64, 437, 39 },
		{ 64, 438, 39 },
		{ 64, 439, 39 },
		{ 64, 440, 39 },
		{ 64, 441, 39 },
		{ 64, 442, 39 },
		{ 64, 443, 39 },
		{ 64, 444, 39 },
		{ 64, 445, 39 },
		{ 64, 446, 39 },
		{ 64, 447, 39 },
		{ 64, 448, 39 },
		{ 64, 449, 39 },
		{ 65, 402, 39 },
		{ 65, 403, 39 },
		{ 65, 404, 39 },
		{ 65, 405, 39 },
		{ 65, 406, 39 },
		{ 65, 407, 39 },
		{ 65, 408, 39 },
		{ 65, 409, 39 },
		{ 65, 410, 39 },
		{ 65, 411, 39 },
		{ 65, 412, 39 },
		{ 65, 413, 39 },
		{ 65, 414, 39 },
		{ 65, 415, 39 },
		{ 65, 416, 39 },
		{ 65, 417, 39 },
		{ 65, 418, 39 },
		{ 65, 419, 39 },
		{ 65, 420, 39 },
		{ 65, 421, 39 },
		{ 65, 422, 39 },
		{ 65, 423, 39 },
		{ 65, 424, 39 },
		{ 65, 425, 39 },
		{ 65, 426, 39 },
		{ 65, 427, 39 },
		{ 65, 428, 39 },
		{ 65, 429, 39 },
		{ 65, 430, 39 },
		{ 65, 431, 39 },
		{ 65, 432, 39 },
		{ 65, 433, 39 },
		{ 65, 434, 39 },
		{ 65, 435, 39 },
		{ 65, 436, 39 },
		{ 65, 437, 39 },
		{ 65, 438, 39 },
		{ 65, 439, 39 },
		{ 65, 440, 39 },
		{ 65, 441, 39 },
		{ 65, 442, 39 },
		{ 65, 443, 39 },
		{ 65, 444, 39 },
		{ 65, 445, 39 },
		{ 65, 446, 39 },
		{ 65, 447, 39 },
		{ 65, 448, 39 },
		{ 66, 401, 39 },
		{ 66, 402, 39 },
		{ 66, 403, 39 },
		{ 66, 404, 39 },
		{ 66, 405, 39 },
		{ 66, 406, 39 },
		{ 66, 407, 39 },
		{ 66, 408, 39 },
		{ 66, 409, 39 },
		{ 66, 410, 39 },
		{ 66, 411, 39 },
		{ 66, 412, 39 },
		{ 66, 413, 39 },
		{ 66, 414, 39 },
		{ 66, 415, 39 },
		{ 66, 416, 39 },
		{ 66, 417, 39 },
		{ 66, 418, 39 },
		{ 66, 419, 39 },
		{ 66, 420, 39 },
		{ 66, 421, 39 },
		{ 66, 422, 39 },
		{ 66, 423, 39 },
		{ 66, 424, 39 },
		{ 66, 425, 39 },
		{ 66, 426, 39 },
		{ 66, 427, 39 },
		{ 66, 428, 39 },
		{ 66, 429, 39 },
		{ 66, 430, 39 },
		{ 66, 431, 39 },
		{ 66, 432, 39 },
		{ 66, 433, 39 },
		{ 66, 434, 39 },
		{ 66, 435, 39 },
		{ 66, 436, 39 },
		{ 66, 437, 39 },
		{ 66, 438, 39 },
		{ 66, 439, 39 },
		{ 66, 440, 39 },
		{ 66, 441, 39 },
		{ 66, 442, 39 },
		{ 66, 443, 39 },
		{ 66, 444, 39 },
		{ 66, 445, 39 },
		{ 66, 446, 39 },
		{ 66, 447, 39 },
		{ 67, 400, 39 },
		{ 67, 401, 39 },
		{ 67, 402, 39 },
		{ 67, 403, 39 },
		{ 67, 404, 39 },
		{ 67, 405, 39 },
		{ 67, 406, 39 },
		{ 67, 407, 39 },
		{ 67, 408, 39 },
		{ 67, 409, 39 },
		{ 67, 410, 39 },
		{ 67, 411, 39 },
		{ 67, 412, 39 },
		{ 67, 413, 39 },
		{ 67, 414, 39 },
		{ 67, 415, 39 },
		{ 67, 416, 39 },
		{ 67, 417, 39 },
		{ 67, 418, 39 },
		{ 67, 419, 39 },
		{ 67, 420, 39 },
		{ 67, 421, 39 },
		{ 67, 422, 39 },
		{ 67, 423, 39 },
		{ 67, 424, 39 },
		{ 67, 425, 39 },
		{ 67, 426, 39 },
		{ 67, 427, 39 },
		{ 67, 428, 39 },
		{ 67, 429, 39 },
		{ 67, 430, 39 },
		{ 67, 431, 39 },
		{ 67, 432, 39 },
		{ 67, 433, 39 },
		{ 67, 434, 39 },
		{ 67, 435, 39 },
		{ 67, 436, 39 },
		{ 67, 437, 39 },
		{ 67, 438, 39 },
		{ 67, 439, 39 },
		{ 67, 440, 39 },
		{ 67, 441, 39 },
		{ 67, 442, 39 },
		{ 67, 443, 39 },
		{ 67, 444, 39 },
		{ 67, 445, 39 },
		{ 67, 446, 39 },
		{ 68, 399, 39 },
		{ 68, 400, 39 },
		{ 68, 401, 39 },
		{ 68, 402, 39 },
		{ 68, 403, 39 },
		{ 68, 404, 39 },
		{ 68, 405, 39 },
		{ 68, 406, 39 },
		{ 68, 407, 39 },
		{ 68, 408, 39 },
		{ 68, 409, 39 },
		{ 68, 410, 39 },
		{ 68, 411, 39 },
		{ 68, 412, 39 },
		{ 68, 413, 39 },
		{ 68, 414, 39 },
		{ 68, 415, 39 },
		{ 68, 416, 39 },
		{ 68, 417, 39 },
		{ 68, 418, 39 },
		{ 68, 419, 39 },
		{ 68, 420, 39 },
		{ 68, 421, 39 },
		{ 68, 422, 39 },
		{ 68, 423, 39 },
		{ 68, 424, 39 },
		{ 68, 425, 39 },
		{ 68, 426, 39 },
		{ 68, 427, 39 },
		{ 68, 428, 39 },
		{ 68, 429, 39 },
		{ 68, 430, 39 },
		{ 68, 431, 39 },
		{ 68, 432, 39 },
		{ 68, 433, 39 },
		{ 68, 434, 39 },
		{ 68, 435, 39 },
		{ 68, 436, 39 },
		{ 68, 437, 39 },
		{ 68, 438, 39 },
		{ 68, 439, 39 },
		{ 68, 440, 39 },
		{ 68, 441, 39 },
		{ 68, 442, 39 },
		{ 68, 443, 39 },
		{ 68, 444, 39 },
		{ 68, 445, 39 },
		{ 69, 399, 39 },
		{ 69, 400, 39 },
		{ 69, 401, 39 },
		{ 69, 402, 39 },
		{ 69, 403, 39 },
		{ 69, 404, 39 },
		{ 69, 405, 39 },
		{ 69, 406, 39 },
		{ 69, 407, 39 },
		{ 69, 408, 39 },
		{ 69, 409, 39 },
		{ 69, 410, 39 },
		{ 69, 411, 39 },
		{ 69, 412, 39 },
		{ 69, 413, 39 },
		{ 69, 414, 39 },
		{ 69, 415, 39 },
		{ 69, 416, 39 },
		{ 69, 417, 39 },
		{ 69, 418, 39 },
		{ 69, 419, 39 },
		{ 69, 420, 39 },
		{ 69, 421, 39 },
		{ 69, 422, 39 },
		{ 69, 423, 39 },
		{ 69, 424, 39 },
		{ 69, 425, 39 },
		{ 69, 426, 39 },
		{ 69, 427, 39 },
		{ 69, 428, 39 },
		{ 69, 429, 39 },
		{ 69, 430, 39 },
		{ 69, 431, 39 },
		{ 69, 432, 39 },
		{ 69, 433, 39 },
		{ 69, 434, 39 },
		{ 69, 435, 39 },
		{ 69, 436, 39 },
		{ 69, 437, 39 },
		{ 69, 438, 39 },
		{ 69, 439, 39 },
		{ 69, 440, 39 },
		{ 69, 441, 39 },
		{ 69, 442, 39 },
		{ 69, 443, 39 },
		{ 69, 444, 39 },
		{ 70, 398, 39 },
		{ 70, 399, 39 },
		{ 70, 400, 39 },
		{ 70, 401, 39 },
		{ 70, 402, 39 },
		{ 70, 403, 39 },
		{ 70, 404, 39 },
		{ 70, 405, 39 },
		{ 70, 406, 39 },
		{ 70, 407, 39 },
		{ 70, 408, 39 },
		{ 70, 409, 39 },
		{ 70, 410, 39 },
		{ 70, 411, 39 },
		{ 70, 412, 39 },
		{ 70, 413, 39 },
		{ 70, 414, 39 },
		{ 70, 415, 39 },
		{ 70, 416, 39 },
		{ 70, 417, 39 },
		{ 70, 418, 39 },
		{ 70, 419, 39 },
		{ 70, 420, 39 },
		{ 70, 421, 39 },
		{ 70, 422, 39 },
		{ 70, 423, 39 },
		{ 70, 424, 39 },
		{ 70, 425, 39 },
		{ 70, 426, 39 },
		{ 70, 427, 39 },
		{ 70, 428, 39 },
		{ 70, 429, 39 },
		{ 70, 430, 39 },
		{ 70, 431, 39 },
		{ 70, 432, 39 },
		{ 70, 433, 39 },
		{ 70, 434, 39 },
		{ 70, 435, 39 },
		{ 70, 436, 39 },
		{ 70, 437, 39 },
		{ 70, 438, 39 },
		{ 70, 439, 39 },
		{ 70, 440, 39 },
		{ 70, 441, 39 },
		{ 70, 442, 39 },
		{ 70, 443, 39 },
		{ 71, 397, 39 },
		{ 71, 398, 39 },
		{ 71, 399, 39 },
		{ 71, 400, 39 },
		{ 71, 401, 39 },
		{ 71, 402, 39 },
		{ 71, 403, 39 },
		{ 71, 404, 39 },
		{ 71, 405, 39 },
		{ 71, 406, 39 },
		{ 71, 407, 39 },
		{ 71, 408, 39 },
		{ 71, 409, 39 },
		{ 71, 410, 39 },
		{ 71, 411, 39 },
		{ 71, 412, 39 },
		{ 71, 413, 39 },
		{ 71, 414, 39 },
		{ 71, 415, 39 },
		{ 71, 416, 39 },
		{ 71, 417, 39 },
		{ 71, 418, 39 },
		{ 71, 419, 39 },
		{ 71, 420, 39 },
		{ 71, 421, 39 },
		{ 71, 422, 39 },
		{ 71, 423, 39 },
		{ 71, 424, 39 },
		{ 71, 425, 39 },
		{ 71, 426, 39 },
		{ 71, 427, 39 },
		{ 71, 428, 39 },
		{ 71, 429, 39 },
		{ 71, 430, 39 },
		{ 71, 431, 39 },
		{ 71, 432, 39 },
		{ 71, 433, 39 },
		{ 71, 434, 39 },
		{ 71, 435, 39 },
		{ 71, 436, 39 },
		{ 71, 437, 39 },
		{ 71, 438, 39 },
		{ 71, 439, 39 },
		{ 71, 440, 39 },
		{ 71, 441, 39 },
		{ 71, 442, 39 },
		{ 72, 396, 39 },
		{ 72, 397, 39 },
		{ 72, 398, 39 },
		{ 72, 399, 39 },
		{ 72, 400, 39 },
		{ 72, 401, 39 },
		{ 72, 402, 39 },
		{ 72, 403, 39 },
		{ 72, 404, 39 },
		{ 72, 405, 39 },
		{ 72, 406, 39 },
		{ 72, 407, 39 },
		{ 72, 408, 39 },
		{ 72, 409, 39 },
		{ 72, 410, 39 },
		{ 72, 411, 39 },
		{ 72, 412, 39 },
		{ 72, 413, 39 },
		{ 72, 414, 39 },
		{ 72, 415, 39 },
		{ 72, 416, 39 },
		{ 72, 417, 39 },
		{ 72, 418, 39 },
		{ 72, 419, 39 },
		{ 72, 420, 39 },
		{ 72, 421, 39 },
		{ 72, 422, 39 },
		{ 72, 423, 39 },
		{ 72, 424, 39 },
		{ 72, 425, 39 },
		{ 72, 426, 39 },
		{ 72, 427, 39 },
		{ 72, 428, 39 },
		{ 72, 429, 39 },
		{ 72, 430, 39 },
		{ 72, 431, 39 },
		{ 72, 432, 39 },
		{ 72, 433, 39 },
		{ 72, 434, 39 },
		{ 72, 435, 39 },
		{ 72, 436, 39 },
		{ 72, 437, 39 },
		{ 72, 438, 39 },
		{ 72, 439, 39 },
		{ 72, 440, 39 },
		{ 72, 441, 39 },
		{ 73, 396, 39 },
		{ 73, 397, 39 },
		{ 73, 398, 39 },
		{ 73, 399, 39 },
		{ 73, 400, 39 },
		{ 73, 401, 39 },
		{ 73, 402, 39 },
		{ 73, 403, 39 },
		{ 73, 404, 39 },
		{ 73, 405, 39 },
		{ 73, 406, 39 },
		{ 73, 407, 39 },
		{ 73, 408, 39 },
		{ 73, 409, 39 },
		{ 73, 410, 39 },
		{ 73, 411, 39 },
		{ 73, 412, 39 },
		{ 73, 413, 39 },
		{ 73, 414, 39 },
		{ 73, 415, 39 },
		{ 73, 416, 39 },
		{ 73, 417, 39 },
		{ 73, 418, 39 },
		{ 73, 419, 39 },
		{ 73, 420, 39 },
		{ 73, 421, 39 },
		{ 73, 422, 39 },
		{ 73, 423, 39 },
		{ 73, 424, 39 },
		{ 73, 425, 39 },
		{ 73, 426, 39 },
		{ 73, 427, 39 },
		{ 73, 428, 39 },
		{ 73, 429, 39 },
		{ 73, 430, 39 },
		{ 73, 431, 39 },
		{ 73, 432, 39 },
		{ 73, 433, 39 },
		{ 73, 434, 39 },
		{ 73, 435, 39 },
		{ 73, 436, 39 },
		{ 73, 437, 39 },
		{ 73, 438, 39 },
		{ 73, 439, 39 },
		{ 73, 440, 39 },
		{ 74, 395, 39 },
		{ 74, 396, 39 },
		{ 74, 397, 39 },
		{ 74, 398, 39 },
		{ 74, 399, 39 },
		{ 74, 400, 39 },
		{ 74, 401, 39 },
		{ 74, 402, 39 },
		{ 74, 403, 39 },
		{ 74, 404, 39 },
		{ 74, 405, 39 },
		{ 74, 406, 39 },
		{ 74, 407, 39 },
		{ 74, 408, 39 },
		{ 74, 409, 39 },
		{ 74, 410, 39 },
		{ 74, 411, 39 },
		{ 74, 412, 39 },
		{ 74, 413, 39 },
		{ 74, 414, 39 },
		{ 74, 415, 39 },
		{ 74, 416, 39 },
		{ 74, 417, 39 },
		{ 74, 418, 39 },
		{ 74, 419, 39 },
		{ 74, 420, 39 },
		{ 74, 421, 39 },
		{ 74, 422, 39 },
		{ 74, 423, 39 },
		{ 74, 424, 39 },
		{ 74, 425, 39 },
		{ 74, 426, 39 },
		{ 74, 427, 39 },
		{ 74, 428, 39 },
		{ 74, 429, 39 },
		{ 74, 430, 39 },
		{ 74, 431, 39 },
		{ 74, 432, 39 },
		{ 74, 433, 39 },
		{ 74, 434, 39 },
		{ 74, 435, 39 },
		{ 74, 436, 39 },
		{ 74, 437, 39 },
		{ 74, 438, 39 },
		{ 74, 439, 39 },
		{ 75, 394, 39 },
		{ 75, 395, 39 },
		{ 75, 396, 39 },
		{ 75, 397, 39 },
		{ 75, 398, 39 },
		{ 75, 399, 39 },
		{ 75, 400, 39 },
		{ 75, 401, 39 },
		{ 75, 402, 39 },
		{ 75, 403, 39 },
		{ 75, 404, 39 },
		{ 75, 405, 39 },
		{ 75, 406, 39 },
		{ 75, 407, 39 },
		{ 75, 408, 39 },
		{ 75, 409, 39 },
		{ 75, 410, 39 },
		{ 75, 411, 39 },
		{ 75, 412, 39 },
		{ 75, 413, 39 },
		{ 75, 414, 39 },
		{ 75, 415, 39 },
		{ 75, 416, 39 },
		{ 75, 417, 39 },
		{ 75, 418, 39 },
		{ 75, 419, 39 },
		{ 75, 420, 39 },
		{ 75, 421, 39 },
		{ 75, 422, 39 },
		{ 75, 423, 39 },
		{ 75, 424, 39 },
		{ 75, 425, 39 },
		{ 75, 426, 39 },
		{ 75, 427, 39 },
		{ 75, 428, 39 },
		{ 75, 429, 39 },
		{ 75, 430, 39 },
		{ 75, 431, 39 },
		{ 75, 432, 39 },
		{ 75, 433, 39 },
		{ 75, 434, 39 },
		{ 75, 435, 39 },
		{ 75, 436, 39 },
		{ 75, 437, 39 },
		{ 75, 438, 39 },
		{ 76, 393, 39 },
		{ 76, 394, 39 },
		{ 76, 395, 39 },
		{ 76, 396, 39 },
		{ 76, 397, 39 },
		{ 76, 398, 39 },
		{ 76, 399, 39 },
		{ 76, 400, 39 },
		{ 76, 401, 39 },
		{ 76, 402, 39 },
		{ 76, 403, 39 },
		{ 76, 404, 39 },
		{ 76, 405, 39 },
		{ 76, 406, 39 },
		{ 76, 407, 39 },
		{ 76, 408, 39 },
		{ 76, 409, 39 },
		{ 76, 410, 39 },
		{ 76, 411, 39 },
		{ 76, 412, 39 },
		{ 76, 413, 39 },
		{ 76, 414, 39 },
		{ 76, 415, 39 },
		{ 76, 416, 39 },
		{ 76, 417, 39 },
		{ 76, 418, 39 },
		{ 76, 419, 39 },
		{ 76, 420, 39 },
		{ 76, 421, 39 },
		{ 76, 422, 39 },
		{ 76, 423, 39 },
		{ 76, 424, 39 },
		{ 76, 425, 39 },
		{ 76, 426, 39 },
		{ 76, 427, 39 },
		{ 76, 428, 39 },
		{ 76, 429, 39 },
		{ 76, 430, 39 },
		{ 76, 431, 39 },
		{ 76, 432, 39 },
		{ 76, 433, 39 },
		{ 76, 434, 39 },
		{ 76, 435, 39 },
		{ 76, 436, 39 },
		{ 76, 437, 39 },
		{ 77, 393, 39 },
		{ 77, 394, 39 },
		{ 77, 395, 39 },
		{ 77, 396, 39 },
		{ 77, 397, 39 },
		{ 77, 398, 39 },
		{ 77, 399, 39 },
		{ 77, 400, 39 },
		{ 77, 401, 39 },
		{ 77, 402, 39 },
		{ 77, 403, 39 },
		{ 77, 404, 39 },
		{ 77, 405, 39 },
		{ 77, 406, 39 },
		{ 77, 407, 39 },
		{ 77, 408, 39 },
		{ 77, 409, 39 },
		{ 77, 410, 39 },
		{ 77, 411, 39 },
		{ 77, 412, 39 },
		{ 77, 413, 39 },
		{ 77, 414, 39 },
		{ 77, 415, 39 },
		{ 77, 416, 39 },
		{ 77, 417, 39 },
		{ 77, 418, 39 },
		{ 77, 419, 39 },
		{ 77, 420, 39 },
		{ 77, 421, 39 },
		{ 77, 422, 39 },
		{ 77, 423, 39 },
		{ 77, 424, 39 },
		{ 77, 425, 39 },
		{ 77, 426, 39 },
		{ 77, 427, 39 },
		{ 77, 428, 39 },
		{ 77, 429, 39 },
		{ 77, 430, 39 },
		{ 77, 431, 39 },
		{ 77, 432, 39 },
		{ 77, 433, 39 },
		{ 77, 434, 39 },
		{ 77, 435, 39 },
		{ 77, 436, 39 },
		{ 78, 392, 39 },
		{ 78, 393, 39 },
		{ 78, 394, 39 },
		{ 78, 395, 39 },
		{ 78, 396, 39 },
		{ 78, 397, 39 },
		{ 78, 398, 39 },
		{ 78, 399, 39 },
		{ 78, 400, 39 },
		{ 78, 401, 39 },
		{ 78, 402, 39 },
		{ 78, 403, 39 },
		{ 78, 404, 39 },
		{ 78, 405, 39 },
		{ 78, 406, 39 },
		{ 78, 407, 39 },
		{ 78, 408, 39 },
		{ 78, 409, 39 },
		{ 78, 410, 39 },
		{ 78, 411, 39 },
		{ 78, 412, 39 },
		{ 78, 413, 39 },
		{ 78, 414, 39 },
		{ 78, 415, 39 },
		{ 78, 416, 39 },
		{ 78, 417, 39 },
		{ 78, 418, 39 },
		{ 78, 419, 39 },
		{ 78, 420, 39 },
		{ 78, 421, 39 },
		{ 78, 422, 39 },
		{ 78, 423, 39 },
		{ 78, 424, 39 },
		{ 78, 425, 39 },
		{ 78, 426, 39 },
		{ 78, 427, 39 },
		{ 78, 428, 39 },
		{ 78, 429, 39 },
		{ 78, 430, 39 },
		{ 78, 431, 39 },
		{ 78, 432, 39 },
		{ 78, 433, 39 },
		{ 78, 434, 39 },
		{ 78, 435, 39 },
		{ 79, 391, 39 },
		{ 79, 392, 39 },
		{ 79, 393, 39 },
		{ 79, 394, 39 },
		{ 79, 395, 39 },
		{ 79, 396, 39 },
		{ 79, 397, 39 },
		{ 79, 398, 39 },
		{ 79, 399, 39 },
		{ 79, 400, 39 },
		{ 79, 401, 39 },
		{ 79, 402, 39 },
		{ 79, 403, 39 },
		{ 79, 404, 39 },
		{ 79, 405, 39 },
		{ 79, 406, 39 },
		{ 79, 407, 39 },
		{ 79, 408, 39 },
		{ 79, 409, 39 },
		{ 79, 410, 39 },
		{ 79, 411, 39 },
		{ 79, 412, 39 },
		{ 79, 413, 39 },
		{ 79, 414, 39 },
		{ 79, 415, 39 },
		{ 79, 416, 39 },
		{ 79, 417, 39 },
		{ 79, 418, 39 },
		{ 79, 419, 39 },
		{ 79, 420, 39 },
		{ 79, 421, 39 },
		{ 79, 422, 39 },
		{ 79, 423, 39 },
		{ 79, 424, 39 },
		{ 79, 425, 39 },
		{ 79, 426, 39 },
		{ 79, 427, 39 },
		{ 79, 428, 39 },
		{ 79, 429, 39 },
		{ 79, 430, 39 },
		{ 79, 431, 39 },
		{ 79, 432, 39 },
		{ 79, 433, 39 },
		{ 79, 434, 39 },
		{ 80, 390, 39 },
		{ 80, 391, 39 },
		{ 80, 392, 39 },
		{ 80, 393, 39 },
		{ 80, 394, 39 },
		{ 80, 395, 39 },
		{ 80, 396, 39 },
		{ 80, 397, 39 },
		{ 80, 398, 39 },
		{ 80, 399, 39 },
		{ 80, 400, 39 },
		{ 80, 401, 39 },
		{ 80, 402, 39 },
		{ 80, 403, 39 },
		{ 80, 404, 39 },
		{ 80, 405, 39 },
		{ 80, 406, 39 },
		{ 80, 407, 39 },
		{ 80, 408, 39 },
		{ 80, 409, 39 },
		{ 80, 410, 39 },
		{ 80, 411, 39 },
		{ 80, 412, 39 },
		{ 80, 413, 39 },
		{ 80, 414, 39 },
		{ 80, 415, 39 },
		{ 80, 416, 39 },
		{ 80, 417, 39 },
		{ 80, 418, 39 },
		{ 80, 419, 39 },
		{ 80, 420, 39 },
		{ 80, 421, 39 },
		{ 80, 422, 39 },
		{ 80, 423, 39 },
		{ 80, 424, 39 },
		{ 80, 425, 39 },
		{ 80, 426, 39 },
		{ 80, 427, 39 },
		{ 80, 428, 39 },
		{ 80, 429, 39 },
		{ 80, 430, 39 },
		{ 80, 431, 39 },
		{ 80, 432, 39 },
		{ 80, 433, 39 },
		{ 81, 390, 39 },
		{ 81, 391, 39 },
		{ 81, 392, 39 },
		{ 81, 393, 39 },
		{ 81, 394, 39 },
		{ 81, 395, 39 },
		{ 81, 396, 39 },
		{ 81, 397, 39 },
		{ 81, 398, 39 },
		{ 81, 399, 39 },
		{ 81, 400, 39 },
		{ 81, 401, 39 },
		{ 81, 402, 39 },
		{ 81, 403, 39 },
		{ 81, 404, 39 },
		{ 81, 405, 39 },
		{ 81, 406, 39 },
		{ 81, 407, 39 },
		{ 81, 408, 39 },
		{ 81, 409, 39 },
		{ 81, 410, 39 },
		{ 81, 411, 39 },
		{ 81, 412, 39 },
		{ 81, 413, 39 },
		{ 81, 414, 39 },
		{ 81, 415, 39 },
		{ 81, 416, 39 },
		{ 81, 417, 39 },
		{ 81, 418, 39 },
		{ 81, 419, 39 },
		{ 81, 420, 39 },
		{ 81, 421, 39 },
		{ 81, 422, 39 },
		{ 81, 423, 39 },
		{ 81, 424, 39 },
		{ 81, 425, 39 },
		{ 81, 426, 39 },
		{ 81, 427, 39 },
		{ 81, 428, 39 },
		{ 81, 429, 39 },
		{ 81, 430, 39 },
		{ 81, 431, 39 },
		{ 81, 432, 39 },
		{ 82, 389, 39 },
		{ 82, 390, 39 },
		{ 82, 391, 39 },
		{ 82, 392, 39 },
		{ 82, 393, 39 },
		{ 82, 394, 39 },
		{ 82, 395, 39 },
		{ 82, 396, 39 },
		{ 82, 397, 39 },
		{ 82, 398, 39 },
		{ 82, 399, 39 },
		{ 82, 400, 39 },
		{ 82, 401, 39 },
		{ 82, 402, 39 },
		{ 82, 403, 39 },
		{ 82, 404, 39 },
		{ 82, 405, 39 },
		{ 82, 406, 39 },
		{ 82, 407, 39 },
		{ 82, 408, 39 },
		{ 82, 409, 39 },
		{ 82, 410, 39 },
		{ 82, 411, 39 },
		{ 82, 412, 39 },
		{ 82, 413, 39 },
		{ 82, 414, 39 },
		{ 82, 415, 39 },
		{ 82, 416, 39 },
		{ 82, 417, 39 },
		{ 82, 418, 39 },
		{ 82, 419, 39 },
		{ 82, 420, 39 },
		{ 82, 421, 39 },
		{ 82, 422, 39 },
		{ 82, 423, 39 },
		{ 82, 424, 39 },
		{ 82, 425, 39 },
		{ 82, 426, 39 },
		{ 82, 427, 39 },
		{ 82, 428, 39 },
		{ 82, 429, 39 },
		{ 82, 430, 39 },
		{ 82, 431, 39 },
		{ 83, 388, 39 },
		{ 83, 389, 39 },
		{ 83, 390, 39 },
		{ 83, 391, 39 },
		{ 83, 392, 39 },
		{ 83, 393, 39 },
		{ 83, 394, 39 },
		{ 83, 395, 39 },
		{ 83, 396, 39 },
		{ 83, 397, 39 },
		{ 83, 398, 39 },
		{ 83, 399, 39 },
		{ 83, 400, 39 },
		{ 83, 401, 39 },
		{ 83, 402, 39 },
		{ 83, 403, 39 },
		{ 83, 404, 39 },
		{ 83, 405, 39 },
		{ 83, 406, 39 },
		{ 83, 407, 39 },
		{ 83, 408, 39 },
		{ 83, 409, 39 },
		{ 83, 410, 39 },
		{ 83, 411, 39 },
		{ 83, 412, 39 },
		{ 83, 413, 39 },
		{ 83, 414, 39 },
		{ 83, 415, 39 },
		{ 83, 416, 39 },
		{ 83, 417, 39 },
		{ 83, 418, 39 },
		{ 83, 419, 39 },
		{ 83, 420, 39 },
		{ 83, 421, 39 },
		{ 83, 422, 39 },
		{ 83, 423, 39 },
		{ 83, 424, 39 },
		{ 83, 425, 39 },
		{ 83, 426, 39 },
		{ 83, 427, 39 },
		{ 83, 428, 39 },
		{ 83, 429, 39 },
		{ 83, 430, 39 },
		{ 84, 387, 39 },
		{ 84, 388, 39 },
		{ 84, 389, 39 },
		{ 84, 390, 39 },
		{ 84, 391, 39 },
		{ 84, 392, 39 },
		{ 84, 393, 39 },
		{ 84, 394, 39 },
		{ 84, 395, 39 },
		{ 84, 396, 39 },
		{ 84, 397, 39 },
		{ 84, 398, 39 },
		{ 84, 399, 39 },
		{ 84, 400, 39 },
		{ 84, 401, 39 },
		{ 84, 402, 39 },
		{ 84, 403, 39 },
		{ 84, 404, 39 },
		{ 84, 405, 39 },
		{ 84, 406, 39 },
		{ 84, 407, 39 },
		{ 84, 408, 39 },
		{ 84, 409, 39 },
		{ 84, 410, 39 },
		{ 84, 411, 39 },
		{ 84, 412, 39 },
		{ 84, 413, 39 },
		{ 84, 414, 39 },
		{ 84, 415, 39 },
		{ 84, 416, 39 },
		{ 84, 417, 39 },
		{ 84, 418, 39 },
		{ 84, 419, 39 },
		{ 84, 420, 39 },
		{ 84, 421, 39 },
		{ 84, 422, 39 },
		{ 84, 423, 39 },
		{ 84, 424, 39 },
		{ 84, 425, 39 },
		{ 84, 426, 39 },
		{ 84, 427, 39 },
		{ 84, 428, 39 },
		{ 84, 429, 39 },
		{ 85, 387, 39 },
		{ 85, 388, 39 },
		{ 85, 389, 39 },
		{ 85, 390, 39 },
		{ 85, 391, 39 },
		{ 85, 392, 39 },
		{ 85, 393, 39 },
		{ 85, 394, 39 },
		{ 85, 395, 39 },
		{ 85, 396, 39 },
		{ 85, 397, 39 },
		{ 85, 398, 39 },
		{ 85, 399, 39 },
		{ 85, 400, 39 },
		{ 85, 401, 39 },
		{ 85, 402, 39 },
		{ 85, 403, 39 },
		{ 85, 404, 39 },
		{ 85, 405, 39 },
		{ 85, 406, 39 },
		{ 85, 407, 39 },
		{ 85, 408, 39 },
		{ 85, 409, 39 },
		{ 85, 410, 39 },
		{ 85, 411, 39 },
		{ 85, 412, 39 },
		{ 85, 413, 39 },
		{ 85, 414, 39 },
		{ 85, 415, 39 },
		{ 85, 416, 39 },
		{ 85, 417, 39 },
		{ 85, 418, 39 },
		{ 85, 419, 39 },
		{ 85, 420, 39 },
		{ 85, 421, 39 },
		{ 85, 422, 39 },
		{ 85, 423, 39 },
		{ 85, 424, 39 },
		{ 85, 425, 39 },
		{ 85, 426, 39 },
		{ 85, 427, 39 },
		{ 85, 428, 39 },
		{ 86, 386, 39 },
		{ 86, 387, 39 },
		{ 86, 388, 39 },
		{ 86, 389, 39 },
		{ 86, 390, 39 },
		{ 86, 391, 39 },
		{ 86, 392, 39 },
		{ 86, 393, 39 },
		{ 86, 394, 39 },
		{ 86, 395, 39 },
		{ 86, 396, 39 },
		{ 86, 397, 39 },
		{ 86, 398, 39 },
		{ 86, 399, 39 },
		{ 86, 400, 39 },
		{ 86, 401, 39 },
		{ 86, 402, 39 },
		{ 86, 403, 39 },
		{ 86, 404, 39 },
		{ 86, 405, 39 },
		{ 86, 406, 39 },
		{ 86, 407, 39 },
		{ 86, 408, 39 },
		{ 86, 409, 39 },
		{ 86, 410, 39 },
		{ 86, 411, 39 },
		{ 86, 412, 39 },
		{ 86, 413, 39 },
		{ 86, 414, 39 },
		{ 86, 415, 39 },
		{ 86, 416, 39 },
		{ 86, 417, 39 },
		{ 86, 418, 39 },
		{ 86, 419, 39 },
		{ 86, 420, 39 },
		{ 86, 421, 39 },
		{ 86, 422, 39 },
		{ 86, 423, 39 },
		{ 86, 424, 39 },
		{ 86, 425, 39 },
		{ 86, 426, 39 },
		{ 86, 427, 39 },
		{ 87, 385, 39 },
		{ 87, 386, 39 },
		{ 87, 387, 39 },
		{ 87, 388, 39 },
		{ 87, 389, 39 },
		{ 87, 390, 39 },
		{ 87, 391, 39 },
		{ 87, 392, 39 },
		{ 87, 393, 39 },
		{ 87, 394, 39 },
		{ 87, 395, 39 },
		{ 87, 396, 39 },
		{ 87, 397, 39 },
		{ 87, 398, 39 },
		{ 87, 399, 39 },
		{ 87, 400, 39 },
		{ 87, 401, 39 },
		{ 87, 402, 39 },
		{ 87, 403, 39 },
		{ 87, 404, 39 },
		{ 87, 405, 39 },
		{ 87, 406, 39 },
		{ 87, 407, 39 },
		{ 87, 408, 39 },
		{ 87, 409, 39 },
		{ 87, 410, 39 },
		{ 87, 411, 39 },
		{ 87, 412, 39 },
		{ 87, 413, 39 },
		{ 87, 414, 39 },
		{ 87, 415, 39 },
		{ 87, 416, 39 },
		{ 87, 417, 39 },
		{ 87, 418, 39 },
		{ 87, 419, 39 },
		{ 87, 420, 39 },
		{ 87, 421, 39 },
		{ 87, 422, 39 },
		{ 87, 423, 39 },
		{ 87, 424, 39 },
		{ 87, 425, 39 },
		{ 87, 426, 39 },
		{ 88, 384, 39 },
		{ 88, 385, 39 },
		{ 88, 386, 39 },
		{ 88, 387, 39 },
		{ 88, 388, 39 },
		{ 88, 389, 39 },
		{ 88, 390, 39 },
		{ 88, 391, 39 },
		{ 88, 392, 39 },
		{ 88, 393, 39 },
		{ 88, 394, 39 },
		{ 88, 395, 39 },
		{ 88, 396, 39 },
		{ 88, 397, 39 },
		{ 88, 398, 39 },
		{ 88, 399, 39 },
		{ 88, 400, 39 },
		{ 88, 401, 39 },
		{ 88, 402, 39 },
		{ 88, 403, 39 },
		{ 88, 404, 39 },
		{ 88, 405, 39 },
		{ 88, 406, 39 },
		{ 88, 407, 39 },
		{ 88, 408, 39 },
		{ 88, 409, 39 },
		{ 88, 410, 39 },
		{ 88, 411, 39 },
		{ 88, 412, 39 },
		{ 88, 413, 39 },
		{ 88, 414, 39 },
		{ 88, 415, 39 },
		{ 88, 416, 39 },
		{ 88, 417, 39 },
		{ 88, 418, 39 },
		{ 88, 419, 39 },
		{ 88, 420, 39 },
		{ 88, 421, 39 },
		{ 88, 422, 39 },
		{ 88, 423, 39 },
		{ 88, 424, 39 },
		{ 88, 425, 39 },
		{ 89, 384, 39 },
		{ 89, 385, 39 },
		{ 89, 386, 39 },
		{ 89, 387, 39 },
		{ 89, 388, 39 },
		{ 89, 389, 39 },
		{ 89, 390, 39 },
		{ 89, 391, 39 },
		{ 89, 392, 39 },
		{ 89, 393, 39 },
		{ 89, 394, 39 },
		{ 89, 395, 39 },
		{ 89, 396, 39 },
		{ 89, 397, 39 },
		{ 89, 398, 39 },
		{ 89, 399, 39 },
		{ 89, 400, 39 },
		{ 89, 401, 39 },
		{ 89, 402, 39 },
		{ 89, 403, 39 },
		{ 89, 404, 39 },
		{ 89, 405, 39 },
		{ 89, 406, 39 },
		{ 89, 407, 39 },
		{ 89, 408, 39 },
		{ 89, 409, 39 },
		{ 89, 410, 39 },
		{ 89, 411, 39 },
		{ 89, 412, 39 },
		{ 89, 413, 39 },
		{ 89, 414, 39 },
		{ 89, 415, 39 },
		{ 89, 416, 39 },
		{ 89, 417, 39 },
		{ 89, 418, 39 },
		{ 89, 419, 39 },
		{ 89, 420, 39 },
		{ 89, 421, 39 },
		{ 89, 422, 39 },
		{ 89, 423, 39 },
		{ 89, 424, 39 },
		{ 90, 383, 39 },
		{ 90, 384, 39 },
		{ 90, 385, 39 },
		{ 90, 386, 39 },
		{ 90, 387, 39 },
		{ 90, 388, 39 },
		{ 90, 389, 39 },
		{ 90, 390, 39 },
		{ 90, 391, 39 },
		{ 90, 392, 39 },
		{ 90, 393, 39 },
		{ 90, 394, 39 },
		{ 90, 395, 39 },
		{ 90, 396, 39 },
		{ 90, 397, 39 },
		{ 90, 398, 39 },
		{ 90, 399, 39 },
		{ 90, 400, 39 },
		{ 90, 401, 39 },
		{ 90, 402, 39 },
		{ 90, 403, 39 },
		{ 90, 404, 39 },
		{ 90, 405, 39 },
		{ 90, 406, 39 },
		{ 90, 407, 39 },
		{ 90, 408, 39 },
		{ 90, 409, 39 },
		{ 90, 410, 39 },
		{ 90, 411, 39 },
		{ 90, 412, 39 },
		{ 90, 413, 39 },
		{ 90, 414, 39 },
		{ 90, 415, 39 },
		{ 90, 416, 39 },
		{ 90, 417, 39 },
		{ 90, 418, 39 },
		{ 90, 419, 39 },
		{ 90, 420, 39 },
		{ 90, 421, 39 },
		{ 90, 422, 39 },
		{ 90, 423, 39 },
		{ 91, 382, 39 },
		{ 91, 383, 39 },
		{ 91, 384, 39 },
		{ 91, 385, 39 },
		{ 91, 386, 39 },
		{ 91, 387, 39 },
		{ 91, 388, 39 },
		{ 91, 389, 39 },
		{ 91, 390, 39 },
		{ 91, 391, 39 },
		{ 91, 392, 39 },
		{ 91, 393, 39 },
		{ 91, 394, 39 },
		{ 91, 395, 39 },
		{ 91, 396, 39 },
		{ 91, 397, 39 },
		{ 91, 398, 39 },
		{ 91, 399, 39 },
		{ 91, 400, 39 },
		{ 91, 401, 39 },
		{ 91, 402, 39 },
		{ 91, 403, 39 },
		{ 91, 404, 39 },
		{ 91, 405, 39 },
		{ 91, 406, 39 },
		{ 91, 407, 39 },
		{ 91, 408, 39 },
		{ 91, 409, 39 },
		{ 91, 410, 39 },
		{ 91, 411, 39 },
		{ 91, 412, 39 },
		{ 91, 413, 39 },
		{ 91, 414, 39 },
		{ 91, 415, 39 },
		{ 91, 416, 39 },
		{ 91, 417, 39 },
		{ 91, 418, 39 },
		{ 91, 419, 39 },
		{ 91, 420, 39 },
		{ 91, 421, 39 },
		{ 91, 422, 39 },
		{ 92, 381, 39 },
		{ 92, 382, 39 },
		{ 92, 383, 39 },
		{ 92, 384, 39 },
		{ 92, 385, 39 },
		{ 92, 386, 39 },
		{ 92, 387, 39 },
		{ 92, 388, 39 },
		{ 92, 389, 39 },
		{ 92, 390, 39 },
		{ 92, 391, 39 },
		{ 92, 392, 39 },
		{ 92, 393, 39 },
		{ 92, 394, 39 },
		{ 92, 395, 39 },
		{ 92, 396, 39 },
		{ 92, 397, 39 },
		{ 92, 398, 39 },
		{ 92, 399, 39 },
		{ 92, 400, 39 },
		{ 92, 401, 39 },
		{ 92, 402, 39 },
		{ 92, 403, 39 },
		{ 92, 404, 39 },
		{ 92, 405, 39 },
		{ 92, 406, 39 },
		{ 92, 407, 39 },
		{ 92, 408, 39 },
		{ 92, 409, 39 },
		{ 92, 410, 39 },
		{ 92, 411, 39 },
		{ 92, 412, 39 },
		{ 92, 413, 39 },
		{ 92, 414, 39 },
		{ 92, 415, 39 },
		{ 92, 416, 39 },
		{ 92, 417, 39 },
		{ 92, 418, 39 },
		{ 92, 419, 39 },
		{ 92, 420, 39 },
		{ 92, 421, 39 },
		{ 93, 381, 39 },
		{ 93, 382, 39 },
		{ 93, 383, 39 },
		{ 93, 384, 39 },
		{ 93, 385, 39 },
		{ 93, 386, 39 },
		{ 93, 387, 39 },
		{ 93, 388, 39 },
		{ 93, 389, 39 },
		{ 93, 390, 39 },
		{ 93, 391, 39 },
		{ 93, 392, 39 },
		{ 93, 393, 39 },
		{ 93, 394, 39 },
		{ 93, 395, 39 },
		{ 93, 396, 39 },
		{ 93, 397, 39 },
		{ 93, 398, 39 },
		{ 93, 399, 39 },
		{ 93, 400, 39 },
		{ 93, 401, 39 },
		{ 93, 402, 39 },
		{ 93, 403, 39 },
		{ 93, 404, 39 },
		{ 93, 405, 39 },
		{ 93, 406, 39 },
		{ 93, 407, 39 },
		{ 93, 408, 39 },
		{ 93, 409, 39 },
		{ 93, 410, 39 },
		{ 93, 411, 39 },
		{ 93, 412, 39 },
		{ 93, 413, 39 },
		{ 93, 414, 39 },
		{ 93, 415, 39 },
		{ 93, 416, 39 },
		{ 93, 417, 39 },
		{ 93, 418, 39 },
		{ 93, 419, 39 },
		{ 93, 420, 39 },
		{ 94, 380, 39 },
		{ 94, 381, 39 },
		{ 94, 382, 39 },
		{ 94, 383, 39 },
		{ 94, 384, 39 },
		{ 94, 385, 39 },
		{ 94, 386, 39 },
		{ 94, 387, 39 },
		{ 94, 388, 39 },
		{ 94, 389, 39 },
		{ 94, 390, 39 },
		{ 94, 391, 39 },
		{ 94, 392, 39 },
		{ 94, 393, 39 },
		{ 94, 394, 39 },
		{ 94, 395, 39 },
		{ 94, 396, 39 },
		{ 94, 397, 39 },
		{ 94, 398, 39 },
		{ 94, 399, 39 },
		{ 94, 400, 39 },
		{ 94, 401, 39 },
		{ 94, 402, 39 },
		{ 94, 403, 39 },
		{ 94, 404, 39 },
		{ 94, 405, 39 },
		{ 94, 406, 39 },
		{ 94, 407, 39 },
		{ 94, 408, 39 },
		{ 94, 409, 39 },
		{ 94, 410, 39 },
		{ 94, 411, 39 },
		{ 94, 412, 39 },
		{ 94, 413, 39 },
		{ 94, 414, 39 },
		{ 94, 415, 39 },
		{ 94, 416, 39 },
		{ 94, 417, 39 },
		{ 94, 418, 39 },
		{ 94, 419, 39 },
		{ 95, 379, 39 },
		{ 95, 380, 39 },
		{ 95, 381, 39 },
		{ 95, 382, 39 },
		{ 95, 383, 39 },
		{ 95, 384, 39 },
		{ 95, 385, 39 },
		{ 95, 386, 39 },
		{ 95, 387, 39 },
		{ 95, 388, 39 },
		{ 95, 389, 39 },
		{ 95, 390, 39 },
		{ 95, 391, 39 },
		{ 95, 392, 39 },
		{ 95, 393, 39 },
		{ 95, 394, 39 },
		{ 95, 395, 39 },
		{ 95, 396, 39 },
		{ 95, 397, 39 },
		{ 95, 398, 39 },
		{ 95, 399, 39 },
		{ 95, 400, 39 },
		{ 95, 401, 39 },
		{ 95, 402, 39 },
		{ 95, 403, 39 },
		{ 95, 404, 39 },
		{ 95, 405, 39 },
		{ 95, 406, 39 },
		{ 95, 407, 39 },
		{ 95, 408, 39 },
		{ 95, 409, 39 },
		{ 95, 410, 39 },
		{ 95, 411, 39 },
		{ 95, 412, 39 },
		{ 95, 413, 39 },
		{ 95, 414, 39 },
		{ 95, 415, 39 },
		{ 95, 416, 39 },
		{ 95, 417, 39 },
		{ 95, 418, 39 },
		{ 96, 378, 39 },
		{ 96, 379, 39 },
		{ 96, 380, 39 },
		{ 96, 381, 39 },
		{ 96, 382, 39 },
		{ 96, 383, 39 },
		{ 96, 384, 39 },
		{ 96, 385, 39 },
		{ 96, 386, 39 },
		{ 96, 387, 39 },
		{ 96, 388, 39 },
		{ 96, 389, 39 },
		{ 96, 390, 39 },
		{ 96, 391, 39 },
		{ 96, 392, 39 },
		{ 96, 393, 39 },
		{ 96, 394, 39 },
		{ 96, 395, 39 },
		{ 96, 396, 39 },
		{ 96, 397, 39 },
		{ 96, 398, 39 },
		{ 96, 399, 39 },
		{ 96, 400, 39 },
		{ 96, 401, 39 },
		{ 96, 402, 39 },
		{ 96, 403, 39 },
		{ 96, 404, 39 },
		{ 96, 405, 39 },
		{ 96, 406, 39 },
		{ 96, 407, 39 },
		{ 96, 408, 39 },
		{ 96, 409, 39 },
		{ 96, 410, 39 },
		{ 96, 411, 39 },
		{ 96, 412, 39 },
		{ 96, 413, 39 },
		{ 96, 414, 39 },
		{ 96, 415, 39 },
		{ 96, 416, 39 },
		{ 96, 417, 39 },
		{ 97, 378, 39 },
		{ 97, 379, 39 },
		{ 97, 380, 39 },
		{ 97, 381, 39 },
		{ 97, 382, 39 },
		{ 97, 383, 39 },
		{ 97, 384, 39 },
		{ 97, 385, 39 },
		{ 97, 386, 39 },
		{ 97, 387, 39 },
		{ 97, 388, 39 },
		{ 97, 389, 39 },
		{ 97, 390, 39 },
		{ 97, 391, 39 },
		{ 97, 392, 39 },
		{ 97, 393, 39 },
		{ 97, 394, 39 },
		{ 97, 395, 39 },
		{ 97, 396, 39 },
		{ 97, 397, 39 },
		{ 97, 398, 39 },
		{ 97, 399, 39 },
		{ 97, 400, 39 },
		{ 97, 401, 39 },
		{ 97, 402, 39 },
		{ 97, 403, 39 },
		{ 97, 404, 39 },
		{ 97, 405, 39 },
		{ 97, 406, 39 },
		{ 97, 407, 39 },
		{ 97, 408, 39 },
		{ 97, 409, 39 },
		{ 97, 410, 39 },
		{ 97, 411, 39 },
		{ 97, 412, 39 },
		{ 97, 413, 39 },
		{ 97, 414, 39 },
		{ 97, 415, 39 },
		{ 97, 416, 39 },
		{ 98, 377, 39 },
		{ 98, 378, 39 },
		{ 98, 379, 39 },
		{ 98, 380, 39 },
		{ 98, 381, 39 },
		{ 98, 382, 39 },
		{ 98, 383, 39 },
		{ 98, 384, 39 },
		{ 98, 385, 39 },
		{ 98, 386, 39 },
		{ 98, 387, 39 },
		{ 98, 388, 39 },
		{ 98, 389, 39 },
		{ 98, 390, 39 },
		{ 98, 391, 39 },
		{ 98, 392, 39 },
		{ 98, 393, 39 },
		{ 98, 394, 39 },
		{ 98, 395, 39 },
		{ 98, 396, 39 },
		{ 98, 397, 39 },
		{ 98, 398, 39 },
		{ 98, 399, 39 },
		{ 98, 400, 39 },
		{ 98, 401, 39 },
		{ 98, 402, 39 },
		{ 98, 403, 39 },
		{ 98, 404, 39 },
		{ 98, 405, 39 },
		{ 98, 406, 39 },
		{ 98, 407, 39 },
		{ 98, 408, 39 },
		{ 98, 409, 39 },
		{ 98, 410, 39 },
		{ 98, 411, 39 },
		{ 98, 412, 39 },
		{ 98, 413, 39 },
		{ 98, 414, 39 },
		{ 98, 415, 39 },
		{ 99, 376, 39 },
		{ 99, 377, 39 },
		{ 99, 378, 39 },
		{ 99, 379, 39 },
		{ 99, 380, 39 },
		{ 99, 381, 39 },
		{ 99, 382, 39 },
		{ 99, 383, 39 },
		{ 99, 384, 39 },
		{ 99, 385, 39 },
		{ 99, 386, 39 },
		{ 99, 387, 39 },
		{ 99, 388, 39 },
		{ 99, 389, 39 },
		{ 99, 390, 39 },
		{ 99, 391, 39 },
		{ 99, 392, 39 },
		{ 99, 393, 39 },
		{ 99, 394, 39 },
		{ 99, 395, 39 },
		{ 99, 396, 39 },
		{ 99, 397, 39 },
		{ 99, 398, 39 },
		{ 99, 399, 39 },
		{ 99, 400, 39 },
		{ 99, 401, 39 },
		{ 99, 402, 39 },
		{ 99, 403, 39 },
		{ 99, 404, 39 },
		{ 99, 405, 39 },
		{ 99, 406, 39 },
		{ 99, 407, 39 },
		{ 99, 408, 39 },
		{ 99, 409, 39 },
		{ 99, 410, 39 },
		{ 99, 411, 39 },
		{ 99, 412, 39 },
		{ 99, 413, 39 },
		{ 99, 414, 39 },
		{ 100, 375, 39 },
		{ 100, 376, 39 },
		{ 100, 377, 39 },
		{ 100, 378, 39 },
		{ 100, 379, 39 },
		{ 100, 380, 39 },
		{ 100, 381, 39 },
		{ 100, 382, 39 },
		{ 100, 383, 39 },
		{ 100, 384, 39 },
		{ 100, 385, 39 },
		{ 100, 386, 39 },
		{ 100, 387, 39 },
		{ 100, 388, 39 },
		{ 100, 389, 39 },
		{ 100, 390, 39 },
		{ 100, 391, 39 },
		{ 100, 392, 39 },
		{ 100, 393, 39 },
		{ 100, 394, 39 },
		{ 100, 395, 39 },
		{ 100, 396, 39 },
		{ 100, 397, 39 },
		{ 100, 398, 39 },
		{ 100, 399, 39 },
		{ 100, 400, 39 },
		{ 100, 401, 39 },
		{ 100, 402, 39 },
		{ 100, 403, 39 },
		{ 100, 404, 39 },
		{ 100, 405, 39 },
		{ 100, 406, 39 },
		{ 100, 407, 39 },
		{ 100, 408, 39 },
		{ 100, 409, 39 },
		{ 100, 410, 39 },
		{ 100, 411, 39 },
		{ 100, 412, 39 },
		{ 100, 413, 39 },
		{ 101, 375, 39 },
		{ 101, 376, 39 },
		{ 101, 377, 39 },
		{ 101, 378, 39 },
		{ 101, 379, 39 },
		{ 101, 380, 39 },
		{ 101, 381, 39 },
		{ 101, 382, 39 },
		{ 101, 383, 39 },
		{ 101, 384, 39 },
		{ 101, 385, 39 },
		{ 101, 386, 39 },
		{ 101, 387, 39 },
		{ 101, 388, 39 },
		{ 101, 389, 39 },
		{ 101, 390, 39 },
		{ 101, 391, 39 },
		{ 101, 392, 39 },
		{ 101, 393, 39 },
		{ 101, 394, 39 },
		{ 101, 395, 39 },
		{ 101, 396, 39 },
		{ 101, 397, 39 },
		{ 101, 398, 39 },
		{ 101, 399, 39 },
		{ 101, 400, 39 },
		{ 101, 401, 39 },
		{ 101, 402, 39 },
		{ 101, 403, 39 },
		{ 101, 404, 39 },
		{ 101, 405, 39 },
		{ 101, 406, 39 },
		{ 101, 407, 39 },
		{ 101, 408, 39 },
		{ 101, 409, 39 },
		{ 101, 410, 39 },
		{ 101, 411, 39 },
		{ 101, 412, 39 },
		{ 102, 374, 39 },
		{ 102, 375, 39 },
		{ 102, 376, 39 },
		{ 102, 377, 39 },
		{ 102, 378, 39 },
		{ 102, 379, 39 },
		{ 102, 380, 39 },
		{ 102, 381, 39 },
		{ 102, 382, 39 },
		{ 102, 383, 39 },
		{ 102, 384, 39 },
		{ 102, 385, 39 },
		{ 102, 386, 39 },
		{ 102, 387, 39 },
		{ 102, 388, 39 },
		{ 102, 389, 39 },
		{ 102, 390, 39 },
		{ 102, 391, 39 },
		{ 102, 392, 39 },
		{ 102, 393, 39 },
		{ 102, 394, 39 },
		{ 102, 395, 39 },
		{ 102, 396, 39 },
		{ 102, 397, 39 },
		{ 102, 398, 39 },
		{ 102, 399, 39 },
		{ 102, 400, 39 },
		{ 102, 401, 39 },
		{ 102, 402, 39 },
		{ 102, 403, 39 },
		{ 102, 404, 39 },
		{ 102, 405, 39 },
		{ 102, 406, 39 },
		{ 102, 407, 39 },
		{ 102, 408, 39 },
		{ 102, 409, 39 },
		{ 102, 410, 39 },
		{ 102, 411, 39 },
		{ 103, 373, 39 },
		{ 103, 374, 39 },
		{ 103, 375, 39 },
		{ 103, 376, 39 },
		{ 103, 377, 39 },
		{ 103, 378, 39 },
		{ 103, 379, 39 },
		{ 103, 380, 39 },
		{ 103, 381, 39 },
		{ 103, 382, 39 },
		{ 103, 383, 39 },
		{ 103, 384, 39 },
		{ 103, 385, 39 },
		{ 103, 386, 39 },
		{ 103, 387, 39 },
		{ 103, 388, 39 },
		{ 103, 389, 39 },
		{ 103, 390, 39 },
		{ 103, 391, 39 },
		{ 103, 392, 39 },
		{ 103, 393, 39 },
		{ 103, 394, 39 },
		{ 103, 395, 39 },
		{ 103, 396, 39 },
		{ 103, 397, 39 },
		{ 103, 398, 39 },
		{ 103, 399, 39 },
		{ 103, 400, 39 },
		{ 103, 401, 39 },
		{ 103, 402, 39 },
		{ 103, 403, 39 },
		{ 103, 404, 39 },
		{ 103, 405, 39 },
		{ 103, 406, 39 },
		{ 103, 407, 39 },
		{ 103, 408, 39 },
		{ 103, 409, 39 },
		{ 103, 410, 39 },
		{ 104, 372, 39 },
		{ 104, 373, 39 },
		{ 104, 374, 39 },
		{ 104, 375, 39 },
		{ 104, 376, 39 },
		{ 104, 377, 39 },
		{ 104, 378, 39 },
		{ 104, 379, 39 },
		{ 104, 380, 39 },
		{ 104, 381, 39 },
		{ 104, 382, 39 },
		{ 104, 383, 39 },
		{ 104, 384, 39 },
		{ 104, 385, 39 },
		{ 104, 386, 39 },
		{ 104, 387, 39 },
		{ 104, 388, 39 },
		{ 104, 389, 39 },
		{ 104, 390, 39 },
		{ 104, 391, 39 },
		{ 104, 392, 39 },
		{ 104, 393, 39 },
		{ 104, 394, 39 },
		{ 104, 395, 39 },
		{ 104, 396, 39 },
		{ 104, 397, 39 },
		{ 104, 398, 39 },
		{ 104, 399, 39 },
		{ 104, 400, 39 },
		{ 104, 401, 39 },
		{ 104, 402, 39 },
		{ 104, 403, 39 },
		{ 104, 404, 39 },
		{ 104, 405, 39 },
		{ 104, 406, 39 },
		{ 104, 407, 39 },
		{ 104, 408, 39 },
		{ 104, 409, 39 },
		{ 105, 372, 39 },
		{ 105, 373, 39 },
		{ 105, 374, 39 },
		{ 105, 375, 39 },
		{ 105, 376, 39 },
		{ 105, 377, 39 },
		{ 105, 378, 39 },
		{ 105, 379, 39 },
		{ 105, 380, 39 },
		{ 105, 381, 39 },
		{ 105, 382, 39 },
		{ 105, 383, 39 },
		{ 105, 384, 39 },
		{ 105, 385, 39 },
		{ 105, 386, 39 },
		{ 105, 387, 39 },
		{ 105, 388, 39 },
		{ 105, 389, 39 },
		{ 105, 390, 39 },
		{ 105, 391, 39 },
		{ 105, 392, 39 },
		{ 105, 393, 39 },
		{ 105, 394, 39 },
		{ 105, 395, 39 },
		{ 105, 396, 39 },
		{ 105, 397, 39 },
		{ 105, 398, 39 },
		{ 105, 399, 39 },
		{ 105, 400, 39 },
		{ 105, 401, 39 },
		{ 105, 402, 39 },
		{ 105, 403, 39 },
		{ 105, 404, 39 },
		{ 105, 405, 39 },
		{ 105, 406, 39 },
		{ 105, 407, 39 },
		{ 105, 408, 39 },
		{ 106, 371, 39 },
		{ 106, 372, 39 },
		{ 106, 373, 39 },
		{ 106, 374, 39 },
		{ 106, 375, 39 },
		{ 106, 376, 39 },
		{ 106, 377, 39 },
		{ 106, 378, 39 },
		{ 106, 379, 39 },
		{ 106, 380, 39 },
		{ 106, 381, 39 },
		{ 106, 382, 39 },
		{ 106, 383, 39 },
		{ 106, 384, 39 },
		{ 106, 385, 39 },
		{ 106, 386, 39 },
		{ 106, 387, 39 },
		{ 106, 388, 39 },
		{ 106, 389, 39 },
		{ 106, 390, 39 },
		{ 106, 391, 39 },
		{ 106, 392, 39 },
		{ 106, 393, 39 },
		{ 106, 394, 39 },
		{ 106, 395, 39 },
		{ 106, 396, 39 },
		{ 106, 397, 39 },
		{ 106, 398, 39 },
		{ 106, 399, 39 },
		{ 106, 400, 39 },
		{ 106, 401, 39 },
		{ 106, 402, 39 },
		{ 106, 403, 39 },
		{ 106, 404, 39 },
		{ 106, 405, 39 },
		{ 106, 406, 39 },
		{ 106, 407, 39 },
		{ 107, 370, 39 },
		{ 107, 371, 39 },
		{ 107, 372, 39 },
		{ 107, 373, 39 },
		{ 107, 374, 39 },
		{ 107, 375, 39 },
		{ 107, 376, 39 },
		{ 107, 377, 39 },
		{ 107, 378, 39 },
		{ 107, 379, 39 },
		{ 107, 380, 39 },
		{ 107, 381, 39 },
		{ 107, 382, 39 },
		{ 107, 383, 39 },
		{ 107, 384, 39 },
		{ 107, 385, 39 },
		{ 107, 386, 39 },
		{ 107, 387, 39 },
		{ 107, 388, 39 },
		{ 107, 389, 39 },
		{ 107, 390, 39 },
		{ 107, 391, 39 },
		{ 107, 392, 39 },
		{ 107, 393, 39 },
		{ 107, 394, 39 },
		{ 107, 395, 39 },
		{ 107, 396, 39 },
		{ 107, 397, 39 },
		{ 107, 398, 39 },
		{ 107, 399, 39 },
		{ 107, 400, 39 },
		{ 107, 401, 39 },
		{ 107, 402, 39 },
		{ 107, 403, 39 },
		{ 107, 404, 39 },
		{ 107, 405, 39 },
		{ 107, 406, 39 },
		{ 108, 369, 39 },
		{ 108, 370, 39 },
		{ 108, 371, 39 },
		{ 108, 372, 39 },
		{ 108, 373, 39 },
		{ 108, 374, 39 },
		{ 108, 375, 39 },
		{ 108, 376, 39 },
		{ 108, 377, 39 },
		{ 108, 378, 39 },
		{ 108, 379, 39 },
		{ 108, 380, 39 },
		{ 108, 381, 39 },
		{ 108, 382, 39 },
		{ 108, 383, 39 },
		{ 108, 384, 39 },
		{ 108, 385, 39 },
		{ 108, 386, 39 },
		{ 108, 387, 39 },
		{ 108, 388, 39 },
		{ 108, 389, 39 },
		{ 108, 390, 39 },
		{ 108, 391, 39 },
		{ 108, 392, 39 },
		{ 108, 393, 39 },
		{ 108, 394, 39 },
		{ 108, 395, 39 },
		{ 108, 396, 39 },
		{ 108, 397, 39 },
		{ 108, 398, 39 },
		{ 108, 399, 39 },
		{ 108, 400, 39 },
		{ 108, 401, 39 },
		{ 108, 402, 39 },
		{ 108, 403, 39 },
		{ 108, 404, 39 },
		{ 108, 405, 39 },
		{ 109, 369, 39 },
		{ 109, 370, 39 },
		{ 109, 371, 39 },
		{ 109, 372, 39 },
		{ 109, 373, 39 },
		{ 109, 374, 39 },
		{ 109, 375, 39 },
		{ 109, 376, 39 },
		{ 109, 377, 39 },
		{ 109, 378, 39 },
		{ 109, 379, 39 },
		{ 109, 380, 39 },
		{ 109, 381, 39 },
		{ 109, 382, 39 },
		{ 109, 383, 39 },
		{ 109, 384, 39 },
		{ 109, 385, 39 },
		{ 109, 386, 39 },
		{ 109, 387, 39 },
		{ 109, 388, 39 },
		{ 109, 389, 39 },
		{ 109, 390, 39 },
		{ 109, 391, 39 },
		{ 109, 392, 39 },
		{ 109, 393, 39 },
		{ 109, 394, 39 },
		{ 109, 395, 39 },
		{ 109, 396, 39 },
		{ 109, 397, 39 },
		{ 109, 398, 39 },
		{ 109, 399, 39 },
		{ 109, 400, 39 },
		{ 109, 401, 39 },
		{ 109, 402, 39 },
		{ 109, 403, 39 },
		{ 109, 404, 39 },
		{ 110, 368, 39 },
		{ 110, 369, 39 },
		{ 110, 370, 39 },
		{ 110, 371, 39 },
		{ 110, 372, 39 },
		{ 110, 373, 39 },
		{ 110, 374, 39 },
		{ 110, 375, 39 },
		{ 110, 376, 39 },
		{ 110, 377, 39 },
		{ 110, 378, 39 },
		{ 110, 379, 39 },
		{ 110, 380, 39 },
		{ 110, 381, 39 },
		{ 110, 382, 39 },
		{ 110, 383, 39 },
		{ 110, 384, 39 },
		{ 110, 385, 39 },
		{ 110, 386, 39 },
		{ 110, 387, 39 },
		{ 110, 388, 39 },
		{ 110, 389, 39 },
		{ 110, 390, 39 },
		{ 110, 391, 39 },
		{ 110, 392, 39 },
		{ 110, 393, 39 },
		{ 110, 394, 39 },
		{ 110, 395, 39 },
		{ 110, 396, 39 },
		{ 110, 397, 39 },
		{ 110, 398, 39 },
		{ 110, 399, 39 },
		{ 110, 400, 39 },
		{ 110, 401, 39 },
		{ 110, 402, 39 },
		{ 110, 403, 39 },
		{ 111, 367, 39 },
		{ 111, 368, 39 },
		{ 111, 369, 39 },
		{ 111, 370, 39 },
		{ 111, 371, 39 },
		{ 111, 372, 39 },
		{ 111, 373, 39 },
		{ 111, 374, 39 },
		{ 111, 375, 39 },
		{ 111, 376, 39 },
		{ 111, 377, 39 },
		{ 111, 378, 39 },
		{ 111, 379, 39 },
		{ 111, 380, 39 },
		{ 111, 381, 39 },
		{ 111, 382, 39 },
		{ 111, 383, 39 },
		{ 111, 384, 39 },
		{ 111, 385, 39 },
		{ 111, 386, 39 },
		{ 111, 387, 39 },
		{ 111, 388, 39 },
		{ 111, 389, 39 },
		{ 111, 390, 39 },
		{ 111, 391, 39 },
		{ 111, 392, 39 },
		{ 111, 393, 39 },
		{ 111, 394, 39 },
		{ 111, 395, 39 },
		{ 111, 396, 39 },
		{ 111, 397, 39 },
		{ 111, 398, 39 },
		{ 111, 399, 39 },
		{ 111, 400, 39 },
		{ 111, 401, 39 },
		{ 111, 402, 39 },
		{ 112, 366, 39 },
		{ 112, 367, 39 },
		{ 112, 368, 39 },
		{ 112, 369, 39 },
		{ 112, 370, 39 },
		{ 112, 371, 39 },
		{ 112, 372, 39 },
		{ 112, 373, 39 },
		{ 112, 374, 39 },
		{ 112, 375, 39 },
		{ 112, 376, 39 },
		{ 112, 377, 39 },
		{ 112, 378, 39 },
		{ 112, 379, 39 },
		{ 112, 380, 39 },
		{ 112, 381, 39 },
		{ 112, 382, 39 },
		{ 112, 383, 39 },
		{ 112, 384, 39 },
		{ 112, 385, 39 },
		{ 112, 386, 39 },
		{ 112, 387, 39 },
		{ 112, 388, 39 },
		{ 112, 389, 39 },
		{ 112, 390, 39 },
		{ 112, 391, 39 },
		{ 112, 392, 39 },
		{ 112, 393, 39 },
		{ 112, 394, 39 },
		{ 112, 395, 39 },
		{ 112, 396, 39 },
		{ 112, 397, 39 },
		{ 112, 398, 39 },
		{ 112, 399, 39 },
		{ 112, 400, 39 },
		{ 112, 401, 39 },
		{ 113, 366, 39 },
		{ 113, 367, 39 },
		{ 113, 368, 39 },
		{ 113, 369, 39 },
		{ 113, 370, 39 },
		{ 113, 371, 39 },
		{ 113, 372, 39 },
		{ 113, 373, 39 },
		{ 113, 374, 39 },
		{ 113, 375, 39 },
		{ 113, 376, 39 },
		{ 113, 377, 39 },
		{ 113, 378, 39 },
		{ 113, 379, 39 },
		{ 113, 380, 39 },
		{ 113, 381, 39 },
		{ 113, 382, 39 },
		{ 113, 383, 39 },
		{ 113, 384, 39 },
		{ 113, 385, 39 },
		{ 113, 386, 39 },
		{ 113, 387, 39 },
		{ 113, 388, 39 },
		{ 113, 389, 39 },
		{ 113, 390, 39 },
		{ 113, 391, 39 },
		{ 113, 392, 39 },
		{ 113, 393, 39 },
		{ 113, 394, 39 },
		{ 113, 395, 39 },
		{ 113, 396, 39 },
		{ 113, 397, 39 },
		{ 113, 398, 39 },
		{ 113, 399, 39 },
		{ 113, 400, 39 },
		{ 114, 365, 39 },
		{ 114, 366, 39 },
		{ 114, 367, 39 },
		{ 114, 368, 39 },
		{ 114, 369, 39 },
		{ 114, 370, 39 },
		{ 114, 371, 39 },
		{ 114, 372, 39 },
		{ 114, 373, 39 },
		{ 114, 374, 39 },
		{ 114, 375, 39 },
		{ 114, 376, 39 },
		{ 114, 377, 39 },
		{ 114, 378, 39 },
		{ 114, 379, 39 },
		{ 114, 380, 39 },
		{ 114, 381, 39 },
		{ 114, 382, 39 },
		{ 114, 383, 39 },
		{ 114, 384, 39 },
		{ 114, 385, 39 },
		{ 114, 386, 39 },
		{ 114, 387, 39 },
		{ 114, 388, 39 },
		{ 114, 389, 39 },
		{ 114, 390, 39 },
		{ 114, 391, 39 },
		{ 114, 392, 39 },
		{ 114, 393, 39 },
		{ 114, 394, 39 },
		{ 114, 395, 39 },
		{ 114, 396, 39 },
		{ 114, 397, 39 },
		{ 114, 398, 39 },
		{ 114, 399, 39 },
		{ 115, 364, 39 },
		{ 115, 365, 39 },
		{ 115, 366, 39 },
		{ 115, 367, 39 },
		{ 115, 368, 39 },
		{ 115, 369, 39 },
		{ 115, 370, 39 },
		{ 115, 371, 39 },
		{ 115, 372, 39 },
		{ 115, 373, 39 },
		{ 115, 374, 39 },
		{ 115, 375, 39 },
		{ 115, 376, 39 },
		{ 115, 377, 39 },
		{ 115, 378, 39 },
		{ 115, 379, 39 },
		{ 115, 380, 39 },
		{ 115, 381, 39 },
		{ 115, 382, 39 },
		{ 115, 383, 39 },
		{ 115, 384, 39 },
		{ 115, 385, 39 },
		{ 115, 386, 39 },
		{ 115, 387, 39 },
		{ 115, 388, 39 },
		{ 115, 389, 39 },
		{ 115, 390, 39 },
		{ 115, 391, 39 },
		{ 115, 392, 39 },
		{ 115, 393, 39 },
		{ 115, 394, 39 },
		{ 115, 395, 39 },
		{ 115, 396, 39 },
		{ 115, 397, 39 },
		{ 115, 398, 39 },
		{ 116, 363, 39 },
		{ 116, 364, 39 },
		{ 116, 365, 39 },
		{ 116, 366, 39 },
		{ 116, 367, 39 },
		{ 116, 368, 39 },
		{ 116, 369, 39 },
		{ 116, 370, 39 },
		{ 116, 371, 39 },
		{ 116, 372, 39 },
		{ 116, 373, 39 },
		{ 116, 374, 39 },
		{ 116, 375, 39 },
		{ 116, 376, 39 },
		{ 116, 377, 39 },
		{ 116, 378, 39 },
		{ 116, 379, 39 },
		{ 116, 380, 39 },
		{ 116, 381, 39 },
		{ 116, 382, 39 },
		{ 116, 383, 39 },
		{ 116, 384, 39 },
		{ 116, 385, 39 },
		{ 116, 386, 39 },
		{ 116, 387, 39 },
		{ 116, 388, 39 },
		{ 116, 389, 39 },
		{ 116, 390, 39 },
		{ 116, 391, 39 },
		{ 116, 392, 39 },
		{ 116, 393, 39 },
		{ 116, 394, 39 },
		{ 116, 395, 39 },
		{ 116, 396, 39 },
		{ 116, 397, 39 },
		{ 117, 363, 39 },
		{ 117, 364, 39 },
		{ 117, 365, 39 },
		{ 117, 366, 39 },
		{ 117, 367, 39 },
		{ 117, 368, 39 },
		{ 117, 369, 39 },
		{ 117, 370, 39 },
		{ 117, 371, 39 },
		{ 117, 372, 39 },
		{ 117, 373, 39 },
		{ 117, 374, 39 },
		{ 117, 375, 39 },
		{ 117, 376, 39 },
		{ 117, 377, 39 },
		{ 117, 378, 39 },
		{ 117, 379, 39 },
		{ 117, 380, 39 },
		{ 117, 381, 39 },
		{ 117, 382, 39 },
		{ 117, 383, 39 },
		{ 117, 384, 39 },
		{ 117, 385, 39 },
		{ 117, 386, 39 },
		{ 117, 387, 39 },
		{ 117, 388, 39 },
		{ 117, 389, 39 },
		{ 117, 390, 39 },
		{ 117, 391, 39 },
		{ 117, 392, 39 },
		{ 117, 393, 39 },
		{ 117, 394, 39 },
		{ 117, 395, 39 },
		{ 117, 396, 39 },
		{ 118, 362, 39 },
		{ 118, 363, 39 },
		{ 118, 364, 39 },
		{ 118, 365, 39 },
		{ 118, 366, 39 },
		{ 118, 367, 39 },
		{ 118, 368, 39 },
		{ 118, 369, 39 },
		{ 118, 370, 39 },
		{ 118, 371, 39 },
		{ 118, 372, 39 },
		{ 118, 373, 39 },
		{ 118, 374, 39 },
		{ 118, 375, 39 },
		{ 118, 376, 39 },
		{ 118, 377, 39 },
		{ 118, 378, 39 },
		{ 118, 379, 39 },
		{ 118, 380, 39 },
		{ 118, 381, 39 },
		{ 118, 382, 39 },
		{ 118, 383, 39 },
		{ 118, 384, 39 },
		{ 118, 385, 39 },
		{ 118, 386, 39 },
		{ 118, 387, 39 },
		{ 118, 388, 39 },
		{ 118, 389, 39 },
		{ 118, 390, 39 },
		{ 118, 391, 39 },
		{ 118, 392, 39 },
		{ 118, 393, 39 },
		{ 118, 394, 39 },
		{ 118, 395, 39 },
		{ 119, 361, 39 },
		{ 119, 362, 39 },
		{ 119, 363, 39 },
		{ 119, 364, 39 },
		{ 119, 365, 39 },
		{ 119, 366, 39 },
		{ 119, 367, 39 },
		{ 119, 368, 39 },
		{ 119, 369, 39 },
		{ 119, 370, 39 },
		{ 119, 371, 39 },
		{ 119, 372, 39 },
		{ 119, 373, 39 },
		{ 119, 374, 39 },
		{ 119, 375, 39 },
		{ 119, 376, 39 },
		{ 119, 377, 39 },
		{ 119, 378, 39 },
		{ 119, 379, 39 },
		{ 119, 380, 39 },
		{ 119, 381, 39 },
		{ 119, 382, 39 },
		{ 119, 383, 39 },
		{ 119, 384, 39 },
		{ 119, 385, 39 },
		{ 119, 386, 39 },
		{ 119, 387, 39 },
		{ 119, 388, 39 },
		{ 119, 389, 39 },
		{ 119, 390, 39 },
		{ 119, 391, 39 },
		{ 119, 392, 39 },
		{ 119, 393, 39 },
		{ 119, 394, 39 },
		{ 120, 360, 39 },
		{ 120, 361, 39 },
		{ 120, 362, 39 },
		{ 120, 363, 39 },
		{ 120, 364, 39 },
		{ 120, 365, 39 },
		{ 120, 366, 39 },
		{ 120, 367, 39 },
		{ 120, 368, 39 },
		{ 120, 369, 39 },
		{ 120, 370, 39 },
		{ 120, 371, 39 },
		{ 120, 372, 39 },
		{ 120, 373, 39 },
		{ 120, 374, 39 },
		{ 120, 375, 39 },
		{ 120, 376, 39 },
		{ 120, 377, 39 },
		{ 120, 378, 39 },
		{ 120, 379, 39 },
		{ 120, 380, 39 },
		{ 120, 381, 39 },
		{ 120, 382, 39 },
		{ 120, 383, 39 },
		{ 120, 384, 39 },
		{ 120, 385, 39 },
		{ 120, 386, 39 },
		{ 120, 387, 39 },
		{ 120, 388, 39 },
		{ 120, 389, 39 },
		{ 120, 390, 39 },
		{ 120, 391, 39 },
		{ 120, 392, 39 },
		{ 120, 393, 39 },
		{ 121, 360, 39 },
		{ 121, 361, 39 },
		{ 121, 362, 39 },
		{ 121, 363, 39 },
		{ 121, 364, 39 },
		{ 121, 365, 39 },
		{ 121, 366, 39 },
		{ 121, 367, 39 },
		{ 121, 368, 39 },
		{ 121, 369, 39 },
		{ 121, 370, 39 },
		{ 121, 371, 39 },
		{ 121, 372, 39 },
		{ 121, 373, 39 },
		{ 121, 374, 39 },
		{ 121, 375, 39 },
		{ 121, 376, 39 },
		{ 121, 377, 39 },
		{ 121, 378, 39 },
		{ 121, 379, 39 },
		{ 121, 380, 39 },
		{ 121, 381, 39 },
		{ 121, 382, 39 },
		{ 121, 383, 39 },
		{ 121, 384, 39 },
		{ 121, 385, 39 },
		{ 121, 386, 39 },
		{ 121, 387, 39 },
		{ 121, 388, 39 },
		{ 121, 389, 39 },
		{ 121, 390, 39 },
		{ 121, 391, 39 },
		{ 121, 392, 39 },
		{ 122, 359, 39 },
		{ 122, 360, 39 },
		{ 122, 361, 39 },
		{ 122, 362, 39 },
		{ 122, 363, 39 },
		{ 122, 364, 39 },
		{ 122, 365, 39 },
		{ 122, 366, 39 },
		{ 122, 367, 39 },
		{ 122, 368, 39 },
		{ 122, 369, 39 },
		{ 122, 370, 39 },
		{ 122, 371, 39 },
		{ 122, 372, 39 },
		{ 122, 373, 39 },
		{ 122, 374, 39 },
		{ 122, 375, 39 },
		{ 122, 376, 39 },
		{ 122, 377, 39 },
		{ 122, 378, 39 },
		{ 122, 379, 39 },
		{ 122, 380, 39 },
		{ 122, 381, 39 },
		{ 122, 382, 39 },
		{ 122, 383, 39 },
		{ 122, 384, 39 },
		{ 122, 385, 39 },
		{ 122, 386, 39 },
		{ 122, 387, 39 },
		{ 122, 388, 39 },
		{ 122, 389, 39 },
		{ 122, 390, 39 },
		{ 122, 391, 39 },
		{ 123, 358, 39 },
		{ 123, 359, 39 },
		{ 123, 360, 39 },
		{ 123, 361, 39 },
		{ 123, 362, 39 },
		{ 123, 363, 39 },
		{ 123, 364, 39 },
		{ 123, 365, 39 },
		{ 123, 366, 39 },
		{ 123, 367, 39 },
		{ 123, 368, 39 },
		{ 123, 369, 39 },
		{ 123, 370, 39 },
		{ 123, 371, 39 },
		{ 123, 372, 39 },
		{ 123, 373, 39 },
		{ 123, 374, 39 },
		{ 123, 375, 39 },
		{ 123, 376, 39 },
		{ 123, 377, 39 },
		{ 123, 378, 39 },
		{ 123, 379, 39 },
		{ 123, 380, 39 },
		{ 123, 381, 39 },
		{ 123, 382, 39 },
		{ 123, 383, 39 },
		{ 123, 384, 39 },
		{ 123, 385, 39 },
		{ 123, 386, 39 },
		{ 123, 387, 39 },
		{ 123, 388, 39 },
		{ 123, 389, 39 },
		{ 123, 390, 39 },
		{ 124, 357, 39 },
		{ 124, 358, 39 },
		{ 124, 359, 39 },
		{ 124, 360, 39 },
		{ 124, 361, 39 },
		{ 124, 362, 39 },
		{ 124, 363, 39 },
		{ 124, 364, 39 },
		{ 124, 365, 39 },
		{ 124, 366, 39 },
		{ 124, 367, 39 },
		{ 124, 368, 39 },
		{ 124, 369, 39 },
		{ 124, 370, 39 },
		{ 124, 371, 39 },
		{ 124, 372, 39 },
		{ 124, 373, 39 },
		{ 124, 374, 39 },
		{ 124, 375, 39 },
		{ 124, 376, 39 },
		{ 124, 377, 39 },
		{ 124, 378, 39 },
		{ 124, 379, 39 },
		{ 124, 380, 39 },
		{ 124, 381, 39 },
		{ 124, 382, 39 },
		{ 124, 383, 39 },
		{ 124, 384, 39 },
		{ 124, 385, 39 },
		{ 124, 386, 39 },
		{ 124, 387, 39 },
		{ 124, 388, 39 },
		{ 124, 389, 39 },
		{ 125, 357, 39 },
		{ 125, 358, 39 },
		{ 125, 359, 39 },
		{ 125, 360, 39 },
		{ 125, 361, 39 },
		{ 125, 362, 39 },
		{ 125, 363, 39 },
		{ 125, 364, 39 },
		{ 125, 365, 39 },
		{ 125, 366, 39 },
		{ 125, 367, 39 },
		{ 125, 368, 39 },
		{ 125, 369, 39 },
		{ 125, 370, 39 },
		{ 125, 371, 39 },
		{ 125, 372, 39 },
		{ 125, 373, 39 },
		{ 125, 374, 39 },
		{ 125, 375, 39 },
		{ 125, 376, 39 },
		{ 125, 377, 39 },
		{ 125, 378, 39 },
		{ 125, 379, 39 },
		{ 125, 380, 39 },
		{ 125, 381, 39 },
		{ 125, 382, 39 },
		{ 125, 383, 39 },
		{ 125, 384, 39 },
		{ 125, 385, 39 },
		{ 125, 386, 39 },
		{ 125, 387, 39 },
		{ 125, 388, 39 },
		{ 126, 356, 39 },
		{ 126, 357, 39 },
		{ 126, 358, 39 },
		{ 126, 359, 39 },
		{ 126, 360, 39 },
		{ 126, 361, 39 },
		{ 126, 362, 39 },
		{ 126, 363, 39 },
		{ 126, 364, 39 },
		{ 126, 365, 39 },
		{ 126, 366, 39 },
		{ 126, 367, 39 },
		{ 126, 368, 39 },
		{ 126, 369, 39 },
		{ 126, 370, 39 },
		{ 126, 371, 39 },
		{ 126, 372, 39 },
		{ 126, 373, 39 },
		{ 126, 374, 39 },
		{ 126, 375, 39 },
		{ 126, 376, 39 },
		{ 126, 377, 39 },
		{ 126, 378, 39 },
		{ 126, 379, 39 },
		{ 126, 380, 39 },
		{ 126, 381, 39 },
		{ 126, 382, 39 },
		{ 126, 383, 39 },
		{ 126, 384, 39 },
		{ 126, 385, 39 },
		{ 126, 386, 39 },
		{ 126, 387, 39 },
		{ 127, 355, 39 },
		{ 127, 356, 39 },
		{ 127, 357, 39 },
		{ 127, 358, 39 },
		{ 127, 359, 39 },
		{ 127, 360, 39 },
		{ 127, 361, 39 },
		{ 127, 362, 39 },
		{ 127, 363, 39 },
		{ 127, 364, 39 },
		{ 127, 365, 39 },
		{ 127, 366, 39 },
		{ 127, 367, 39 },
		{ 127, 368, 39 },
		{ 127, 369, 39 },
		{ 127, 370, 39 },
		{ 127, 371, 39 },
		{ 127, 372, 39 },
		{ 127, 373, 39 },
		{ 127, 374, 39 },
		{ 127, 375, 39 },
		{ 127, 376, 39 },
		{ 127, 377, 39 },
		{ 127, 378, 39 },
		{ 127, 379, 39 },
		{ 127, 380, 39 },
		{ 127, 381, 39 },
		{ 127, 382, 39 },
		{ 127, 383, 39 },
		{ 127, 384, 39 },
		{ 127, 385, 39 },
		{ 127, 386, 39 },
		{ 128, 354, 39 },
		{ 128, 355, 39 },
		{ 128, 356, 39 },
		{ 128, 357, 39 },
		{ 128, 358, 39 },
		{ 128, 359, 39 },
		{ 128, 360, 39 },
		{ 128, 361, 39 },
		{ 128, 362, 39 },
		{ 128, 363, 39 },
		{ 128, 364, 39 },
		{ 128, 365, 39 },
		{ 128, 366, 39 },
		{ 128, 367, 39 },
		{ 128, 368, 39 },
		{ 128, 369, 39 },
		{ 128, 370, 39 },
		{ 128, 371, 39 },
		{ 128, 372, 39 },
		{ 128, 373, 39 },
		{ 128, 374, 39 },
		{ 128, 375, 39 },
		{ 128, 376, 39 },
		{ 128, 377, 39 },
		{ 128, 378, 39 },
		{ 128, 379, 39 },
		{ 128, 380, 39 },
		{ 128, 381, 39 },
		{ 128, 382, 39 },
		{ 128, 383, 39 },
		{ 128, 384, 39 },
		{ 128, 385, 39 },
		{ 129, 354, 39 },
		{ 129, 355, 39 },
		{ 129, 356, 39 },
		{ 129, 357, 39 },
		{ 129, 358, 39 },
		{ 129, 359, 39 },
		{ 129, 360, 39 },
		{ 129, 361, 39 },
		{ 129, 362, 39 },
		{ 129, 363, 39 },
		{ 129, 364, 39 },
		{ 129, 365, 39 },
		{ 129, 366, 39 },
		{ 129, 367, 39 },
		{ 129, 368, 39 },
		{ 129, 369, 39 },
		{ 129, 370, 39 },
		{ 129, 371, 39 },
		{ 129, 372, 39 },
		{ 129, 373, 39 },
		{ 129, 374, 39 },
		{ 129, 375, 39 },
		{ 129, 376, 39 },
		{ 129, 377, 39 },
		{ 129, 378, 39 },
		{ 129, 379, 39 },
		{ 129, 380, 39 },
		{ 129, 381, 39 },
		{ 129, 382, 39 },
		{ 129, 383, 39 },
		{ 129, 384, 39 },
		{ 130, 353, 39 },
		{ 130, 354, 39 },
		{ 130, 355, 39 },
		{ 130, 356, 39 },
		{ 130, 357, 39 },
		{ 130, 358, 39 },
		{ 130, 359, 39 },
		{ 130, 360, 39 },
		{ 130, 361, 39 },
		{ 130, 362, 39 },
		{ 130, 363, 39 },
		{ 130, 364, 39 },
		{ 130, 365, 39 },
		{ 130, 366, 39 },
		{ 130, 367, 39 },
		{ 130, 368, 39 },
		{ 130, 369, 39 },
		{ 130, 370, 39 },
		{ 130, 371, 39 },
		{ 130, 372, 39 },
		{ 130, 373, 39 },
		{ 130, 374, 39 },
		{ 130, 375, 39 },
		{ 130, 376, 39 },
		{ 130, 377, 39 },
		{ 130, 378, 39 },
		{ 130, 379, 39 },
		{ 130, 380, 39 },
		{ 130, 381, 39 },
		{ 130, 382, 39 },
		{ 130, 383, 39 },
		{ 131, 352, 39 },
		{ 131, 353, 39 },
		{ 131, 354, 39 },
		{ 131, 355, 39 },
		{ 131, 356, 39 },
		{ 131, 357, 39 },
		{ 131, 358, 39 },
		{ 131, 359, 39 },
		{ 131, 360, 39 },
		{ 131, 361, 39 },
		{ 131, 362, 39 },
		{ 131, 363, 39 },
		{ 131, 364, 39 },
		{ 131, 365, 39 },
		{ 131, 366, 39 },
		{ 131, 367, 39 },
		{ 131, 368, 39 },
		{ 131, 369, 39 },
		{ 131, 370, 39 },
		{ 131, 371, 39 },
		{ 131, 372, 39 },
		{ 131, 373, 39 },
		{ 131, 374, 39 },
		{ 131, 375, 39 },
		{ 131, 376, 39 },
		{ 131, 377, 39 },
		{ 131, 378, 39 },
		{ 131, 379, 39 },
		{ 131, 380, 39 },
		{ 131, 381, 39 },
		{ 131, 382, 39 },
		{ 132, 351, 39 },
		{ 132, 352, 39 },
		{ 132, 353, 39 },
		{ 132, 354, 39 },
		{ 132, 355, 39 },
		{ 132, 356, 39 },
		{ 132, 357, 39 },
		{ 132, 358, 39 },
		{ 132, 359, 39 },
		{ 132, 360, 39 },
		{ 132, 361, 39 },
		{ 132, 362, 39 },
		{ 132, 363, 39 },
		{ 132, 364, 39 },
		{ 132, 365, 39 },
		{ 132, 366, 39 },
		{ 132, 367, 39 },
		{ 132, 368, 39 },
		{ 132, 369, 39 },
		{ 132, 370, 39 },
		{ 132, 371, 39 },
		{ 132, 372, 39 },
		{ 132, 373, 39 },
		{ 132, 374, 39 },
		{ 132, 375, 39 },
		{ 132, 376, 39 },
		{ 132, 377, 39 },
		{ 132, 378, 39 },
		{ 132, 379, 39 },
		{ 132, 380, 39 },
		{ 132, 381, 39 },
		{ 133, 351, 39 },
		{ 133, 352, 39 },
		{ 133, 353, 39 },
		{ 133, 354, 39 },
		{ 133, 355, 39 },
		{ 133, 356, 39 },
		{ 133, 357, 39 },
		{ 133, 358, 39 },
		{ 133, 359, 39 },
		{ 133, 360, 39 },
		{ 133, 361, 39 },
		{ 133, 362, 39 },
		{ 133, 363, 39 },
		{ 133, 364, 39 },
		{ 133, 365, 39 },
		{ 133, 366, 39 },
		{ 133, 367, 39 },
		{ 133, 368, 39 },
		{ 133, 369, 39 },
		{ 133, 370, 39 },
		{ 133, 371, 39 },
		{ 133, 372, 39 },
		{ 133, 373, 39 },
		{ 133, 374, 39 },
		{ 133, 375, 39 },
		{ 133, 376, 39 },
		{ 133, 377, 39 },
		{ 133, 378, 39 },
		{ 133, 379, 39 },
		{ 133, 380, 39 },
		{ 134, 350, 39 },
		{ 134, 351, 39 },
		{ 134, 352, 39 },
		{ 134, 353, 39 },
		{ 134, 354, 39 },
		{ 134, 355, 39 },
		{ 134, 356, 39 },
		{ 134, 357, 39 },
		{ 134, 358, 39 },
		{ 134, 359, 39 },
		{ 134, 360, 39 },
		{ 134, 361, 39 },
		{ 134, 362, 39 },
		{ 134, 363, 39 },
		{ 134, 364, 39 },
		{ 134, 365, 39 },
		{ 134, 366, 39 },
		{ 134, 367, 39 },
		{ 134, 368, 39 },
		{ 134, 369, 39 },
		{ 134, 370, 39 },
		{ 134, 371, 39 },
		{ 134, 372, 39 },
		{ 134, 373, 39 },
		{ 134, 374, 39 },
		{ 134, 375, 39 },
		{ 134, 376, 39 },
		{ 134, 377, 39 },
		{ 134, 378, 39 },
		{ 134, 379, 39 },
		{ 135, 349, 39 },
		{ 135, 350, 39 },
		{ 135, 351, 39 },
		{ 135, 352, 39 },
		{ 135, 353, 39 },
		{ 135, 354, 39 },
		{ 135, 355, 39 },
		{ 135, 356, 39 },
		{ 135, 357, 39 },
		{ 135, 358, 39 },
		{ 135, 359, 39 },
		{ 135, 360, 39 },
		{ 135, 361, 39 },
		{ 135, 362, 39 },
		{ 135, 363, 39 },
		{ 135, 364, 39 },
		{ 135, 365, 39 },
		{ 135, 366, 39 },
		{ 135, 367, 39 },
		{ 135, 368, 39 },
		{ 135, 369, 39 },
		{ 135, 370, 39 },
		{ 135, 371, 39 },
		{ 135, 372, 39 },
		{ 135, 373, 39 },
		{ 135, 374, 39 },
		{ 135, 375, 39 },
		{ 135, 376, 39 },
		{ 135, 377, 39 },
		{ 135, 378, 39 },
		{ 136, 348, 39 },
		{ 136, 349, 39 },
		{ 136, 350, 39 },
		{ 136, 351, 39 },
		{ 136, 352, 39 },
		{ 136, 353, 39 },
		{ 136, 354, 39 },
		{ 136, 355, 39 },
		{ 136, 356, 39 },
		{ 136, 357, 39 },
		{ 136, 358, 39 },
		{ 136, 359, 39 },
		{ 136, 360, 39 },
		{ 136, 361, 39 },
		{ 136, 362, 39 },
		{ 136, 363, 39 },
		{ 136, 364, 39 },
		{ 136, 365, 39 },
		{ 136, 366, 39 },
		{ 136, 367, 39 },
		{ 136, 368, 39 },
		{ 136, 369, 39 },
		{ 136, 370, 39 },
		{ 136, 371, 39 },
		{ 136, 372, 39 },
		{ 136, 373, 39 },
		{ 136, 374, 39 },
		{ 136, 375, 39 },
		{ 136, 376, 39 },
		{ 136, 377, 39 },
		{ 137, 348, 39 },
		{ 137, 349, 39 },
		{ 137, 350, 39 },
		{ 137, 351, 39 },
		{ 137, 352, 39 },
		{ 137, 353, 39 },
		{ 137, 354, 39 },
		{ 137, 355, 39 },
		{ 137, 356, 39 },
		{ 137, 357, 39 },
		{ 137, 358, 39 },
		{ 137, 359, 39 },
		{ 137, 360, 39 },
		{ 137, 361, 39 },
		{ 137, 362, 39 },
		{ 137, 363, 39 },
		{ 137, 364, 39 },
		{ 137, 365, 39 },
		{ 137, 366, 39 },
		{ 137, 367, 39 },
		{ 137, 368, 39 },
		{ 137, 369, 39 },
		{ 137, 370, 39 },
		{ 137, 371, 39 },
		{ 137, 372, 39 },
		{ 137, 373, 39 },
		{ 137, 374, 39 },
		{ 137, 375, 39 },
		{ 137, 376, 39 },
		{ 138, 347, 39 },
		{ 138, 348, 39 },
		{ 138, 349, 39 },
		{ 138, 350, 39 },
		{ 138, 351, 39 },
		{ 138, 352, 39 },
		{ 138, 353, 39 },
		{ 138, 354, 39 },
		{ 138, 355, 39 },
		{ 138, 356, 39 },
		{ 138, 357, 39 },
		{ 138, 358, 39 },
		{ 138, 359, 39 },
		{ 138, 360, 39 },
		{ 138, 361, 39 },
		{ 138, 362, 39 },
		{ 138, 363, 39 },
		{ 138, 364, 39 },
		{ 138, 365, 39 },
		{ 138, 366, 39 },
		{ 138, 367, 39 },
		{ 138, 368, 39 },
		{ 138, 369, 39 },
		{ 138, 370, 39 },
		{ 138, 371, 39 },
		{ 138, 372, 39 },
		{ 138, 373, 39 },
		{ 138, 374, 39 },
		{ 138, 375, 39 },
		{ 139, 346, 39 },
		{ 139, 347, 39 },
		{ 139, 348, 39 },
		{ 139, 349, 39 },
		{ 139, 350, 39 },
		{ 139, 351, 39 },
		{ 139, 352, 39 },
		{ 139, 353, 39 },
		{ 139, 354, 39 },
		{ 139, 355, 39 },
		{ 139, 356, 39 },
		{ 139, 357, 39 },
		{ 139, 358, 39 },
		{ 139, 359, 39 },
		{ 139, 360, 39 },
		{ 139, 361, 39 },
		{ 139, 362, 39 },
		{ 139, 363, 39 },
		{ 139, 364, 39 },
		{ 139, 365, 39 },
		{ 139, 366, 39 },
		{ 139, 367, 39 },
		{ 139, 368, 39 },
		{ 139, 369, 39 },
		{ 139, 370, 39 },
		{ 139, 371, 39 },
		{ 139, 372, 39 },
		{ 139, 373, 39 },
		{ 139, 374, 39 },
		{ 140, 345, 39 },
		{ 140, 346, 39 },
		{ 140, 347, 39 },
		{ 140, 348, 39 },
		{ 140, 349, 39 },
		{ 140, 350, 39 },
		{ 140, 351, 39 },
		{ 140, 352, 39 },
		{ 140, 353, 39 },
		{ 140, 354, 39 },
		{ 140, 355, 39 },
		{ 140, 356, 39 },
		{ 140, 357, 39 },
		{ 140, 358, 39 },
		{ 140, 359, 39 },
		{ 140, 360, 39 },
		{ 140, 361, 39 },
		{ 140, 362, 39 },
		{ 140, 363, 39 },
		{ 140, 364, 39 },
		{ 140, 365, 39 },
		{ 140, 366, 39 },
		{ 140, 367, 39 },
		{ 140, 368, 39 },
		{ 140, 369, 39 },
		{ 140, 370, 39 },
		{ 140, 371, 39 },
		{ 140, 372, 39 },
		{ 140, 373, 39 },
		{ 141, 345, 39 },
		{ 141, 346, 39 },
		{ 141, 347, 39 },
		{ 141, 348, 39 },
		{ 141, 349, 39 },
		{ 141, 350, 39 },
		{ 141, 351, 39 },
		{ 141, 352, 39 },
		{ 141, 353, 39 },
		{ 141, 354, 39 },
		{ 141, 355, 39 },
		{ 141, 356, 39 },
		{ 141, 357, 39 },
		{ 141, 358, 39 },
		{ 141, 359, 39 },
		{ 141, 360, 39 },
		{ 141, 361, 39 },
		{ 141, 362, 39 },
		{ 141, 363, 39 },
		{ 141, 364, 39 },
		{ 141, 365, 39 },
		{ 141, 366, 39 },
		{ 141, 367, 39 },
		{ 141, 368, 39 },
		{ 141, 369, 39 },
		{ 141, 370, 39 },
		{ 141, 371, 39 },
		{ 141, 372, 39 },
		{ 142, 344, 39 },
		{ 142, 345, 39 },
		{ 142, 346, 39 },
		{ 142, 347, 39 },
		{ 142, 348, 39 },
		{ 142, 349, 39 },
		{ 142, 350, 39 },
		{ 142, 351, 39 },
		{ 142, 352, 39 },
		{ 142, 353, 39 },
		{ 142, 354, 39 },
		{ 142, 355, 39 },
		{ 142, 356, 39 },
		{ 142, 357, 39 },
		{ 142, 358, 39 },
		{ 142, 359, 39 },
		{ 142, 360, 39 },
		{ 142, 361, 39 },
		{ 142, 362, 39 },
		{ 142, 363, 39 },
		{ 142, 364, 39 },
		{ 142, 365, 39 },
		{ 142, 366, 39 },
		{ 142, 367, 39 },
		{ 142, 368, 39 },
		{ 142, 369, 39 },
		{ 142, 370, 39 },
		{ 142, 371, 39 },
		{ 143, 343, 39 },
		{ 143, 344, 39 },
		{ 143, 345, 39 },
		{ 143, 346, 39 },
		{ 143, 347, 39 },
		{ 143, 348, 39 },
		{ 143, 349, 39 },
		{ 143, 350, 39 },
		{ 143, 351, 39 },
		{ 143, 352, 39 },
		{ 143, 353, 39 },
		{ 143, 354, 39 },
		{ 143, 355, 39 },
		{ 143, 356, 39 },
		{ 143, 357, 39 },
		{ 143, 358, 39 },
		{ 143, 359, 39 },
		{ 143, 360, 39 },
		{ 143, 361, 39 },
		{ 143, 362, 39 },
		{ 143, 363, 39 },
		{ 143, 364, 39 },
		{ 143, 365, 39 },
		{ 143, 366, 39 },
		{ 143, 367, 39 },
		{ 143, 368, 39 },
		{ 143, 369, 39 },
		{ 143, 370, 39 },
		{ 144, 342, 39 },
		{ 144, 343, 39 },
		{ 144, 344, 39 },
		{ 144, 345, 39 },
		{ 144, 346, 39 },
		{ 144, 347, 39 },
		{ 144, 348, 39 },
		{ 144, 349, 39 },
		{ 144, 350, 39 },
		{ 144, 351, 39 },
		{ 144, 352, 39 },
		{ 144, 353, 39 },
		{ 144, 354, 39 },
		{ 144, 355, 39 },
		{ 144, 356, 39 },
		{ 144, 357, 39 },
		{ 144, 358, 39 },
		{ 144, 359, 39 },
		{ 144, 360, 39 },
		{ 144, 361, 39 },
		{ 144, 362, 39 },
		{ 144, 363, 39 },
		{ 144, 364, 39 },
		{ 144, 365, 39 },
		{ 144, 366, 39 },
		{ 144, 367, 39 },
		{ 144, 368, 39 },
		{ 144, 369, 39 },
		{ 145, 342, 39 },
		{ 145, 343, 39 },
		{ 145, 344, 39 },
		{ 145, 345, 39 },
		{ 145, 346, 39 },
		{ 145, 347, 39 },
		{ 145, 348, 39 },
		{ 145, 349, 39 },
		{ 145, 350, 39 },
		{ 145, 351, 39 },
		{ 145, 352, 39 },
		{ 145, 353, 39 },
		{ 145, 354, 39 },
		{ 145, 355, 39 },
		{ 145, 356, 39 },
		{ 145, 357, 39 },
		{ 145, 358, 39 },
		{ 145, 359, 39 },
		{ 145, 360, 39 },
		{ 145, 361, 39 },
		{ 145, 362, 39 },
		{ 145, 363, 39 },
		{ 145, 364, 39 },
		{ 145, 365, 39 },
		{ 145, 366, 39 },
		{ 145, 367, 39 },
		{ 145, 368, 39 },
		{ 146, 341, 39 },
		{ 146, 342, 39 },
		{ 146, 343, 39 },
		{ 146, 344, 39 },
		{ 146, 345, 39 },
		{ 146, 346, 39 },
		{ 146, 347, 39 },
		{ 146, 348, 39 },
		{ 146, 349, 39 },
		{ 146, 350, 39 },
		{ 146, 351, 39 },
		{ 146, 352, 39 },
		{ 146, 353, 39 },
		{ 146, 354, 39 },
		{ 146, 355, 39 },
		{ 146, 356, 39 },
		{ 146, 357, 39 },
		{ 146, 358, 39 },
		{ 146, 359, 39 },
		{ 146, 360, 39 },
		{ 146, 361, 39 },
		{ 146, 362, 39 },
		{ 146, 363, 39 },
		{ 146, 364, 39 },
		{ 146, 365, 39 },
		{ 146, 366, 39 },
		{ 146, 367, 39 },
		{ 147, 340, 39 },
		{ 147, 341, 39 },
		{ 147, 342, 39 },
		{ 147, 343, 39 },
		{ 147, 344, 39 },
		{ 147, 345, 39 },
		{ 147, 346, 39 },
		{ 147, 347, 39 },
		{ 147, 348, 39 },
		{ 147, 349, 39 },
		{ 147, 350, 39 },
		{ 147, 351, 39 },
		{ 147, 352, 39 },
		{ 147, 353, 39 },
		{ 147, 354, 39 },
		{ 147, 355, 39 },
		{ 147, 356, 39 },
		{ 147, 357, 39 },
		{ 147, 358, 39 },
		{ 147, 359, 39 },
		{ 147, 360, 39 },
		{ 147, 361, 39 },
		{ 147, 362, 39 },
		{ 147, 363, 39 },
		{ 147, 364, 39 },
		{ 147, 365, 39 },
		{ 147, 366, 39 },
		{ 148, 339, 39 },
		{ 148, 340, 39 },
		{ 148, 341, 39 },
		{ 148, 342, 39 },
		{ 148, 343, 39 },
		{ 148, 344, 39 },
		{ 148, 345, 39 },
		{ 148, 346, 39 },
		{ 148, 347, 39 },
		{ 148, 348, 39 },
		{ 148, 349, 39 },
		{ 148, 350, 39 },
		{ 148, 351, 39 },
		{ 148, 352, 39 },
		{ 148, 353, 39 },
		{ 148, 354, 39 },
		{ 148, 355, 39 },
		{ 148, 356, 39 },
		{ 148, 357, 39 },
		{ 148, 358, 39 },
		{ 148, 359, 39 },
		{ 148, 360, 39 },
		{ 148, 361, 39 },
		{ 148, 362, 39 },
		{ 148, 363, 39 },
		{ 148, 364, 39 },
		{ 148, 365, 39 },
		{ 149, 339, 39 },
		{ 149, 340, 39 },
		{ 149, 341, 39 },
		{ 149, 342, 39 },
		{ 149, 343, 39 },
		{ 149, 344, 39 },
		{ 149, 345, 39 },
		{ 149, 346, 39 },
		{ 149, 347, 39 },
		{ 149, 348, 39 },
		{ 149, 349, 39 },
		{ 149, 350, 39 },
		{ 149, 351, 39 },
		{ 149, 352, 39 },
		{ 149, 353, 39 },
		{ 149, 354, 39 },
		{ 149, 355, 39 },
		{ 149, 356, 39 },
		{ 149, 357, 39 },
		{ 149, 358, 39 },
		{ 149, 359, 39 },
		{ 149, 360, 39 },
		{ 149, 361, 39 },
		{ 149, 362, 39 },
		{ 149, 363, 39 },
		{ 149, 364, 39 },
		{ 150, 338, 39 },
		{ 150, 339, 39 },
		{ 150, 340, 39 },
		{ 150, 341, 39 },
		{ 150, 342, 39 },
		{ 150, 343, 39 },
		{ 150, 344, 39 },
		{ 150, 345, 39 },
		{ 150, 346, 39 },
		{ 150, 347, 39 },
		{ 150, 348, 39 },
		{ 150, 349, 39 },
		{ 150, 350, 39 },
		{ 150, 351, 39 },
		{ 150, 352, 39 },
		{ 150, 353, 39 },
		{ 150, 354, 39 },
		{ 150, 355, 39 },
		{ 150, 356, 39 },
		{ 150, 357, 39 },
		{ 150, 358, 39 },
		{ 150, 359, 39 },
		{ 150, 360, 39 },
		{ 150, 361, 39 },
		{ 150, 362, 39 },
		{ 150, 363, 39 },
		{ 151, 337, 39 },
		{ 151, 338, 39 },
		{ 151, 339, 39 },
		{ 151, 340, 39 },
		{ 151, 341, 39 },
		{ 151, 342, 39 },
		{ 151, 343, 39 },
		{ 151, 344, 39 },
		{ 151, 345, 39 },
		{ 151, 346, 39 },
		{ 151, 347, 39 },
		{ 151, 348, 39 },
		{ 151, 349, 39 },
		{ 151, 350, 39 },
		{ 151, 351, 39 },
		{ 151, 352, 39 },
		{ 151, 353, 39 },
		{ 151, 354, 39 },
		{ 151, 355, 39 },
		{ 151, 356, 39 },
		{ 151, 357, 39 },
		{ 151, 358, 39 },
		{ 151, 359, 39 },
		{ 151, 360, 39 },
		{ 151, 361, 39 },
		{ 151, 362, 39 },
		{ 152, 336, 39 },
		{ 152, 337, 39 },
		{ 152, 338, 39 },
		{ 152, 339, 39 },
		{ 152, 340, 39 },
		{ 152, 341, 39 },
		{ 152, 342, 39 },
		{ 152, 343, 39 },
		{ 152, 344, 39 },
		{ 152, 345, 39 },
		{ 152, 346, 39 },
		{ 152, 347, 39 },
		{ 152, 348, 39 },
		{ 152, 349, 39 },
		{ 152, 350, 39 },
		{ 152, 351, 39 },
		{ 152, 352, 39 },
		{ 152, 353, 39 },
		{ 152, 354, 39 },
		{ 152, 355, 39 },
		{ 152, 356, 39 },
		{ 152, 357, 39 },
		{ 152, 358, 39 },
		{ 152, 359, 39 },
		{ 152, 360, 39 },
		{ 152, 361, 39 },
		{ 153, 336, 39 },
		{ 153, 337, 39 },
		{ 153, 338, 39 },
		{ 153, 339, 39 },
		{ 153, 340, 39 },
		{ 153, 341, 39 },
		{ 153, 342, 39 },
		{ 153, 343, 39 },
		{ 153, 344, 39 },
		{ 153, 345, 39 },
		{ 153, 346, 39 },
		{ 153, 347, 39 },
		{ 153, 348, 39 },
		{ 153, 349, 39 },
		{ 153, 350, 39 },
		{ 153, 351, 39 },
		{ 153, 352, 39 },
		{ 153, 353, 39 },
		{ 153, 354, 39 },
		{ 153, 355, 39 },
		{ 153, 356, 39 },
		{ 153, 357, 39 },
		{ 153, 358, 39 },
		{ 153, 359, 39 },
		{ 153, 360, 39 },
		{ 154, 335, 39 },
		{ 154, 336, 39 },
		{ 154, 337, 39 },
		{ 154, 338, 39 },
		{ 154, 339, 39 },
		{ 154, 340, 39 },
		{ 154, 341, 39 },
		{ 154, 342, 39 },
		{ 154, 343, 39 },
		{ 154, 344, 39 },
		{ 154, 345, 39 },
		{ 154, 346, 39 },
		{ 154, 347, 39 },
		{ 154, 348, 39 },
		{ 154, 349, 39 },
		{ 154, 350, 39 },
		{ 154, 351, 39 },
		{ 154, 352, 39 },
		{ 154, 353, 39 },
		{ 154, 354, 39 },
		{ 154, 355, 39 },
		{ 154, 356, 39 },
		{ 154, 357, 39 },
		{ 154, 358, 39 },
		{ 154, 359, 39 },
		{ 155, 334, 39 },
		{ 155, 335, 39 },
		{ 155, 336, 39 },
		{ 155, 337, 39 },
		{ 155, 338, 39 },
		{ 155, 339, 39 },
		{ 155, 340, 39 },
		{ 155, 341, 39 },
		{ 155, 342, 39 },
		{ 155, 343, 39 },
		{ 155, 344, 39 },
		{ 155, 345, 39 },
		{ 155, 346, 39 },
		{ 155, 347, 39 },
		{ 155, 348, 39 },
		{ 155, 349, 39 },
		{ 155, 350, 39 },
		{ 155, 351, 39 },
		{ 155, 352, 39 },
		{ 155, 353, 39 },
		{ 155, 354, 39 },
		{ 155, 355, 39 },
		{ 155, 356, 39 },
		{ 155, 357, 39 },
		{ 155, 358, 39 },
		{ 156, 333, 39 },
		{ 156, 334, 39 },
		{ 156, 335, 39 },
		{ 156, 336, 39 },
		{ 156, 337, 39 },
		{ 156, 338, 39 },
		{ 156, 339, 39 },
		{ 156, 340, 39 },
		{ 156, 341, 39 },
		{ 156, 342, 39 },
		{ 156, 343, 39 },
		{ 156, 344, 39 },
		{ 156, 345, 39 },
		{ 156, 346, 39 },
		{ 156, 347, 39 },
		{ 156, 348, 39 },
		{ 156, 349, 39 },
		{ 156, 350, 39 },
		{ 156, 351, 39 },
		{ 156, 352, 39 },
		{ 156, 353, 39 },
		{ 156, 354, 39 },
		{ 156, 355, 39 },
		{ 156, 356, 39 },
		{ 156, 357, 39 },
		{ 157, 333, 39 },
		{ 157, 334, 39 },
		{ 157, 335, 39 },
		{ 157, 336, 39 },
		{ 157, 337, 39 },
		{ 157, 338, 39 },
		{ 157, 339, 39 },
		{ 157, 340, 39 },
		{ 157, 341, 39 },
		{ 157, 342, 39 },
		{ 157, 343, 39 },
		{ 157, 344, 39 },
		{ 157, 345, 39 },
		{ 157, 346, 39 },
		{ 157, 347, 39 },
		{ 157, 348, 39 },
		{ 157, 349, 39 },
		{ 157, 350, 39 },
		{ 157, 351, 39 },
		{ 157, 352, 39 },
		{ 157, 353, 39 },
		{ 157, 354, 39 },
		{ 157, 355, 39 },
		{ 157, 356, 39 },
		{ 158, 332, 39 },
		{ 158, 333, 39 },
		{ 158, 334, 39 },
		{ 158, 335, 39 },
		{ 158, 336, 39 },
		{ 158, 337, 39 },
		{ 158, 338, 39 },
		{ 158, 339, 39 },
		{ 158, 340, 39 },
		{ 158, 341, 39 },
		{ 158, 342, 39 },
		{ 158, 343, 39 },
		{ 158, 344, 39 },
		{ 158, 345, 39 },
		{ 158, 346, 39 },
		{ 158, 347, 39 },
		{ 158, 348, 39 },
		{ 158, 349, 39 },
		{ 158, 350, 39 },
		{ 158, 351, 39 },
		{ 158, 352, 39 },
		{ 158, 353, 39 },
		{ 158, 354, 39 },
		{ 158, 355, 39 },
		{ 159, 331, 39 },
		{ 159, 332, 39 },
		{ 159, 333, 39 },
		{ 159, 334, 39 },
		{ 159, 335, 39 },
		{ 159, 336, 39 },
		{ 159, 337, 39 },
		{ 159, 338, 39 },
		{ 159, 339, 39 },
		{ 159, 340, 39 },
		{ 159, 341, 39 },
		{ 159, 342, 39 },
		{ 159, 343, 39 },
		{ 159, 344, 39 },
		{ 159, 345, 39 },
		{ 159, 346, 39 },
		{ 159, 347, 39 },
		{ 159, 348, 39 },
		{ 159, 349, 39 },
		{ 159, 350, 39 },
		{ 159, 351, 39 },
		{ 159, 352, 39 },
		{ 159, 353, 39 },
		{ 159, 354, 39 },
		{ 160, 330, 39 },
		{ 160, 331, 39 },
		{ 160, 332, 39 },
		{ 160, 333, 39 },
		{ 160, 334, 39 },
		{ 160, 335, 39 },
		{ 160, 336, 39 },
		{ 160, 337, 39 },
		{ 160, 338, 39 },
		{ 160, 339, 39 },
		{ 160, 340, 39 },
		{ 160, 341, 39 },
		{ 160, 342, 39 },
		{ 160, 343, 39 },
		{ 160, 344, 39 },
		{ 160, 345, 39 },
		{ 160, 346, 39 },
		{ 160, 347, 39 },
		{ 160, 348, 39 },
		{ 160, 349, 39 },
		{ 160, 350, 39 },
		{ 160, 351, 39 },
		{ 160, 352, 39 },
		{ 160, 353, 39 },
		{ 161, 330, 39 },
		{ 161, 331, 39 },
		{ 161, 332, 39 },
		{ 161, 333, 39 },
		{ 161, 334, 39 },
		{ 161, 335, 39 },
		{ 161, 336, 39 },
		{ 161, 337, 39 },
		{ 161, 338, 39 },
		{ 161, 339, 39 },
		{ 161, 340, 39 },
		{ 161, 341, 39 },
		{ 161, 342, 39 },
		{ 161, 343, 39 },
		{ 161, 344, 39 },
		{ 161, 345, 39 },
		{ 161, 346, 39 },
		{ 161, 347, 39 },
		{ 161, 348, 39 },
		{ 161, 349, 39 },
		{ 161, 350, 39 },
		{ 161, 351, 39 },
		{ 161, 352, 39 },
		{ 162, 329, 39 },
		{ 162, 330, 39 },
		{ 162, 331, 39 },
		{ 162, 332, 39 },
		{ 162, 333, 39 },
		{ 162, 334, 39 },
		{ 162, 335, 39 },
		{ 162, 336, 39 },
		{ 162, 337, 39 },
		{ 162, 338, 39 },
		{ 162, 339, 39 },
		{ 162, 340, 39 },
		{ 162, 341, 39 },
		{ 162, 342, 39 },
		{ 162, 343, 39 },
		{ 162, 344, 39 },
		{ 162, 345, 39 },
		{ 162, 346, 39 },
		{ 162, 347, 39 },
		{ 162, 348, 39 },
		{ 162, 349, 39 },
		{ 162, 350, 39 },
		{ 162, 351, 39 },
		{ 163, 328, 39 },
		{ 163, 329, 39 },
		{ 163, 330, 39 },
		{ 163, 331, 39 },
		{ 163, 332, 39 },
		{ 163, 333, 39 },
		{ 163, 334, 39 },
		{ 163, 335, 39 },
		{ 163, 336, 39 },
		{ 163, 337, 39 },
		{ 163, 338, 39 },
		{ 163, 339, 39 },
		{ 163, 340, 39 },
		{ 163, 341, 39 },
		{ 163, 342, 39 },
		{ 163, 343, 39 },
		{ 163, 344, 39 },
		{ 163, 345, 39 },
		{ 163, 346, 39 },
		{ 163, 347, 39 },
		{ 163, 348, 39 },
		{ 163, 349, 39 },
		{ 163, 350, 39 },
		{ 164, 327, 39 },
		{ 164, 328, 39 },
		{ 164, 329, 39 },
		{ 164, 330, 39 },
		{ 164, 331, 39 },
		{ 164, 332, 39 },
		{ 164, 333, 39 },
		{ 164, 334, 39 },
		{ 164, 335, 39 },
		{ 164, 336, 39 },
		{ 164, 337, 39 },
		{ 164, 338, 39 },
		{ 164, 339, 39 },
		{ 164, 340, 39 },
		{ 164, 341, 39 },
		{ 164, 342, 39 },
		{ 164, 343, 39 },
		{ 164, 344, 39 },
		{ 164, 345, 39 },
		{ 164, 346, 39 },
		{ 164, 347, 39 },
		{ 164, 348, 39 },
		{ 164, 349, 39 },
		{ 165, 327, 39 },
		{ 165, 328, 39 },
		{ 165, 329, 39 },
		{ 165, 330, 39 },
		{ 165, 331, 39 },
		{ 165, 332, 39 },
		{ 165, 333, 39 },
		{ 165, 334, 39 },
		{ 165, 335, 39 },
		{ 165, 336, 39 },
		{ 165, 337, 39 },
		{ 165, 338, 39 },
		{ 165, 339, 39 },
		{ 165, 340, 39 },
		{ 165, 341, 39 },
		{ 165, 342, 39 },
		{ 165, 343, 39 },
		{ 165, 344, 39 },
		{ 165, 345, 39 },
		{ 165, 346, 39 },
		{ 165, 347, 39 },
		{ 165, 348, 39 },
		{ 166, 326, 39 },
		{ 166, 327, 39 },
		{ 166, 328, 39 },
		{ 166, 329, 39 },
		{ 166, 330, 39 },
		{ 166, 331, 39 },
		{ 166, 332, 39 },
		{ 166, 333, 39 },
		{ 166, 334, 39 },
		{ 166, 335, 39 },
		{ 166, 336, 39 },
		{ 166, 337, 39 },
		{ 166, 338, 39 },
		{ 166, 339, 39 },
		{ 166, 340, 39 },
		{ 166, 341, 39 },
		{ 166, 342, 39 },
		{ 166, 343, 39 },
		{ 166, 344, 39 },
		{ 166, 345, 39 },
		{ 166, 346, 39 },
		{ 166, 347, 39 },
		{ 167, 325, 39 },
		{ 167, 326, 39 },
		{ 167, 327, 39 },
		{ 167, 328, 39 },
		{ 167, 329, 39 },
		{ 167, 330, 39 },
		{ 167, 331, 39 },
		{ 167, 332, 39 },
		{ 167, 333, 39 },
		{ 167, 334, 39 },
		{ 167, 335, 39 },
		{ 167, 336, 39 },
		{ 167, 337, 39 },
		{ 167, 338, 39 },
		{ 167, 339, 39 },
		{ 167, 340, 39 },
		{ 167, 341, 39 },
		{ 167, 342, 39 },
		{ 167, 343, 39 },
		{ 167, 344, 39 },
		{ 167, 345, 39 },
		{ 167, 346, 39 },
		{ 168, 324, 39 },
		{ 168, 325, 39 },
		{ 168, 326, 39 },
		{ 168, 327, 39 },
		{ 168, 328, 39 },
		{ 168, 329, 39 },
		{ 168, 330, 39 },
		{ 168, 331, 39 },
		{ 168, 332, 39 },
		{ 168, 333, 39 },
		{ 168, 334, 39 },
		{ 168, 335, 39 },
		{ 168, 336, 39 },
		{ 168, 337, 39 },
		{ 168, 338, 39 },
		{ 168, 339, 39 },
		{ 168, 340, 39 },
		{ 168, 341, 39 },
		{ 168, 342, 39 },
		{ 168, 343, 39 },
		{ 168, 344, 39 },
		{ 168, 345, 39 },
		{ 169, 324, 39 },
		{ 169, 325, 39 },
		{ 169, 326, 39 },
		{ 169, 327, 39 },
		{ 169, 328, 39 },
		{ 169, 329, 39 },
		{ 169, 330, 39 },
		{ 169, 331, 39 },
		{ 169, 332, 39 },
		{ 169, 333, 39 },
		{ 169, 334, 39 },
		{ 169, 335, 39 },
		{ 169, 336, 39 },
		{ 169, 337, 39 },
		{ 169, 338, 39 },
		{ 169, 339, 39 },
		{ 169, 340, 39 },
		{ 169, 341, 39 },
		{ 169, 342, 39 },
		{ 169, 343, 39 },
		{ 169, 344, 39 },
		{ 170, 323, 39 },
		{ 170, 324, 39 },
		{ 170, 325, 39 },
		{ 170, 326, 39 },
		{ 170, 327, 39 },
		{ 170, 328, 39 },
		{ 170, 329, 39 },
		{ 170, 330, 39 },
		{ 170, 331, 39 },
		{ 170, 332, 39 },
		{ 170, 333, 39 },
		{ 170, 334, 39 },
		{ 170, 335, 39 },
		{ 170, 336, 39 },
		{ 170, 337, 39 },
		{ 170, 338, 39 },
		{ 170, 339, 39 },
		{ 170, 340, 39 },
		{ 170, 341, 39 },
		{ 170, 342, 39 },
		{ 170, 343, 39 },
		{ 171, 322, 39 },
		{ 171, 323, 39 },
		{ 171, 324, 39 },
		{ 171, 325, 39 },
		{ 171, 326, 39 },
		{ 171, 327, 39 },
		{ 171, 328, 39 },
		{ 171, 329, 39 },
		{ 171, 330, 39 },
		{ 171, 331, 39 },
		{ 171, 332, 39 },
		{ 171, 333, 39 },
		{ 171, 334, 39 },
		{ 171, 335, 39 },
		{ 171, 336, 39 },
		{ 171, 337, 39 },
		{ 171, 338, 39 },
		{ 171, 339, 39 },
		{ 171, 340, 39 },
		{ 171, 341, 39 },
		{ 171, 342, 39 },
		{ 172, 321, 39 },
		{ 172, 322, 39 },
		{ 172, 323, 39 },
		{ 172, 324, 39 },
		{ 172, 325, 39 },
		{ 172, 326, 39 },
		{ 172, 327, 39 },
		{ 172, 328, 39 },
		{ 172, 329, 39 },
		{ 172, 330, 39 },
		{ 172, 331, 39 },
		{ 172, 332, 39 },
		{ 172, 333, 39 },
		{ 172, 334, 39 },
		{ 172, 335, 39 },
		{ 172, 336, 39 },
		{ 172, 337, 39 },
		{ 172, 338, 39 },
		{ 172, 339, 39 },
		{ 172, 340, 39 },
		{ 172, 341, 39 },
		{ 173, 321, 39 },
		{ 173, 322, 39 },
		{ 173, 323, 39 },
		{ 173, 324, 39 },
		{ 173, 325, 39 },
		{ 173, 326, 39 },
		{ 173, 327, 39 },
		{ 173, 328, 39 },
		{ 173, 329, 39 },
		{ 173, 330, 39 },
		{ 173, 331, 39 },
		{ 173, 332, 39 },
		{ 173, 333, 39 },
		{ 173, 334, 39 },
		{ 173, 335, 39 },
		{ 173, 336, 39 },
		{ 173, 337, 39 },
		{ 173, 338, 39 },
		{ 173, 339, 39 },
		{ 173, 340, 39 },
		{ 174, 320, 39 },
		{ 174, 321, 39 },
		{ 174, 322, 39 },
		{ 174, 323, 39 },
		{ 174, 324, 39 },
		{ 174, 325, 39 },
		{ 174, 326, 39 },
		{ 174, 327, 39 },
		{ 174, 328, 39 },
		{ 174, 329, 39 },
		{ 174, 330, 39 },
		{ 174, 331, 39 },
		{ 174, 332, 39 },
		{ 174, 333, 39 },
		{ 174, 334, 39 },
		{ 174, 335, 39 },
		{ 174, 336, 39 },
		{ 174, 337, 39 },
		{ 174, 338, 39 },
		{ 174, 339, 39 },
		{ 175, 319, 39 },
		{ 175, 320, 39 },
		{ 175, 321, 39 },
		{ 175, 322, 39 },
		{ 175, 323, 39 },
		{ 175, 324, 39 },
		{ 175, 325, 39 },
		{ 175, 326, 39 },
		{ 175, 327, 39 },
		{ 175, 328, 39 },
		{ 175, 329, 39 },
		{ 175, 330, 39 },
		{ 175, 331, 39 },
		{ 175, 332, 39 },
		{ 175, 333, 39 },
		{ 175, 334, 39 },
		{ 175, 335, 39 },
		{ 175, 336, 39 },
		{ 175, 337, 39 },
		{ 175, 338, 39 },
		{ 176, 318, 39 },
		{ 176, 319, 39 },
		{ 176, 320, 39 },
		{ 176, 321, 39 },
		{ 176, 322, 39 },
		{ 176, 323, 39 },
		{ 176, 324, 39 },
		{ 176, 325, 39 },
		{ 176, 326, 39 },
		{ 176, 327, 39 },
		{ 176, 328, 39 },
		{ 176, 329, 39 },
		{ 176, 330, 39 },
		{ 176, 331, 39 },
		{ 176, 332, 39 },
		{ 176, 333, 39 },
		{ 176, 334, 39 },
		{ 176, 335, 39 },
		{ 176, 336, 39 },
		{ 176, 337, 39 },
		{ 177, 318, 39 },
		{ 177, 319, 39 },
		{ 177, 320, 39 },
		{ 177, 321, 39 },
		{ 177, 322, 39 },
		{ 177, 323, 39 },
		{ 177, 324, 39 },
		{ 177, 325, 39 },
		{ 177, 326, 39 },
		{ 177, 327, 39 },
		{ 177, 328, 39 },
		{ 177, 329, 39 },
		{ 177, 330, 39 },
		{ 177, 331, 39 },
		{ 177, 332, 39 },
		{ 177, 333, 39 },
		{ 177, 334, 39 },
		{ 177, 335, 39 },
		{ 177, 336, 39 },
		{ 178, 317, 39 },
		{ 178, 318, 39 },
		{ 178, 319, 39 },
		{ 178, 320, 39 },
		{ 178, 321, 39 },
		{ 178, 322, 39 },
		{ 178, 323, 39 },
		{ 178, 324, 39 },
		{ 178, 325, 39 },
		{ 178, 326, 39 },
		{ 178, 327, 39 },
		{ 178, 328, 39 },
		{ 178, 329, 39 },
		{ 178, 330, 39 },
		{ 178, 331, 39 },
		{ 178, 332, 39 },
		{ 178, 333, 39 },
		{ 178, 334, 39 },
		{ 178, 335, 39 },
		{ 179, 316, 39 },
		{ 179, 317, 39 },
		{ 179, 318, 39 },
		{ 179, 319, 39 },
		{ 179, 320, 39 },
		{ 179, 321, 39 },
		{ 179, 322, 39 },
		{ 179, 323, 39 },
		{ 179, 324, 39 },
		{ 179, 325, 39 },
		{ 179, 326, 39 },
		{ 179, 327, 39 },
		{ 179, 328, 39 },
		{ 179, 329, 39 },
		{ 179, 330, 39 },
		{ 179, 331, 39 },
		{ 179, 332, 39 },
		{ 179, 333, 39 },
		{ 179, 334, 39 },
		{ 180, 315, 39 },
		{ 180, 316, 39 },
		{ 180, 317, 39 },
		{ 180, 318, 39 },
		{ 180, 319, 39 },
		{ 180, 320, 39 },
		{ 180, 321, 39 },
		{ 180, 322, 39 },
		{ 180, 323, 39 },
		{ 180, 324, 39 },
		{ 180, 325, 39 },
		{ 180, 326, 39 },
		{ 180, 327, 39 },
		{ 180, 328, 39 },
		{ 180, 329, 39 },
		{ 180, 330, 39 },
		{ 180, 331, 39 },
		{ 180, 332, 39 },
		{ 180, 333, 39 },
		{ 181, 315, 39 },
		{ 181, 316, 39 },
		{ 181, 317, 39 },
		{ 181, 318, 39 },
		{ 181, 319, 39 },
		{ 181, 320, 39 },
		{ 181, 321, 39 },
		{ 181, 322, 39 },
		{ 181, 323, 39 },
		{ 181, 324, 39 },
		{ 181, 325, 39 },
		{ 181, 326, 39 },
		{ 181, 327, 39 },
		{ 181, 328, 39 },
		{ 181, 329, 39 },
		{ 181, 330, 39 },
		{ 181, 331, 39 },
		{ 181, 332, 39 },
		{ 182, 314, 39 },
		{ 182, 315, 39 },
		{ 182, 316, 39 },
		{ 182, 317, 39 },
		{ 182, 318, 39 },
		{ 182, 319, 39 },
		{ 182, 320, 39 },
		{ 182, 321, 39 },
		{ 182, 322, 39 },
		{ 182, 323, 39 },
		{ 182, 324, 39 },
		{ 182, 325, 39 },
		{ 182, 326, 39 },
		{ 182, 327, 39 },
		{ 182, 328, 39 },
		{ 182, 329, 39 },
		{ 182, 330, 39 },
		{ 182, 331, 39 },
		{ 183, 313, 39 },
		{ 183, 314, 39 },
		{ 183, 315, 39 },
		{ 183, 316, 39 },
		{ 183, 317, 39 },
		{ 183, 318, 39 },
		{ 183, 319, 39 },
		{ 183, 320, 39 },
		{ 183, 321, 39 },
		{ 183, 322, 39 },
		{ 183, 323, 39 },
		{ 183, 324, 39 },
		{ 183, 325, 39 },
		{ 183, 326, 39 },
		{ 183, 327, 39 },
		{ 183, 328, 39 },
		{ 183, 329, 39 },
		{ 183, 330, 39 },
		{ 184, 312, 39 },
		{ 184, 313, 39 },
		{ 184, 314, 39 },
		{ 184, 315, 39 },
		{ 184, 316, 39 },
		{ 184, 317, 39 },
		{ 184, 318, 39 },
		{ 184, 319, 39 },
		{ 184, 320, 39 },
		{ 184, 321, 39 },
		{ 184, 322, 39 },
		{ 184, 323, 39 },
		{ 184, 324, 39 },
		{ 184, 325, 39 },
		{ 184, 326, 39 },
		{ 184, 327, 39 },
		{ 184, 328, 39 },
		{ 184, 329, 39 },
		{ 185, 312, 39 },
		{ 185, 313, 39 },
		{ 185, 314, 39 },
		{ 185, 315, 39 },
		{ 185, 316, 39 },
		{ 185, 317, 39 },
		{ 185, 318, 39 },
		{ 185, 319, 39 },
		{ 185, 320, 39 },
		{ 185, 321, 39 },
		{ 185, 322, 39 },
		{ 185, 323, 39 },
		{ 185, 324, 39 },
		{ 185, 325, 39 },
		{ 185, 326, 39 },
		{ 185, 327, 39 },
		{ 185, 328, 39 },
		{ 186, 311, 39 },
		{ 186, 312, 39 },
		{ 186, 313, 39 },
		{ 186, 314, 39 },
		{ 186, 315, 39 },
		{ 186, 316, 39 },
		{ 186, 317, 39 },
		{ 186, 318, 39 },
		{ 186, 319, 39 },
		{ 186, 320, 39 },
		{ 186, 321, 39 },
		{ 186, 322, 39 },
		{ 186, 323, 39 },
		{ 186, 324, 39 },
		{ 186, 325, 39 },
		{ 186, 326, 39 },
		{ 186, 327, 39 },
		{ 187, 310, 39 },
		{ 187, 311, 39 },
		{ 187, 312, 39 },
		{ 187, 313, 39 },
		{ 187, 314, 39 },
		{ 187, 315, 39 },
		{ 187, 316, 39 },
		{ 187, 317, 39 },
		{ 187, 318, 39 },
		{ 187, 319, 39 },
		{ 187, 320, 39 },
		{ 187, 321, 39 },
		{ 187, 322, 39 },
		{ 187, 323, 39 },
		{ 187, 324, 39 },
		{ 187, 325, 39 },
		{ 187, 326, 39 },
		{ 188, 310, 39 },
		{ 188, 311, 39 },
		{ 188, 312, 39 },
		{ 188, 313, 39 },
		{ 188, 314, 39 },
		{ 188, 315, 39 },
		{ 188, 316, 39 },
		{ 188, 317, 39 },
		{ 188, 318, 39 },
		{ 188, 319, 39 },
		{ 188, 320, 39 },
		{ 188, 321, 39 },
		{ 188, 322, 39 },
		{ 188, 323, 39 },
		{ 188, 324, 39 },
		{ 188, 325, 39 },
		{ 189, 309, 39 },
		{ 189, 310, 39 },
		{ 189, 311, 39 },
		{ 189, 312, 39 },
		{ 189, 313, 39 },
		{ 189, 314, 39 },
		{ 189, 315, 39 },
		{ 189, 316, 39 },
		{ 189, 317, 39 },
		{ 189, 318, 39 },
		{ 189, 319, 39 },
		{ 189, 320, 39 },
		{ 189, 321, 39 },
		{ 189, 322, 39 },
		{ 189, 323, 39 },
		{ 189, 324, 39 },
		{ 190, 308, 39 },
		{ 190, 309, 39 },
		{ 190, 310, 39 },
		{ 190, 311, 39 },
		{ 190, 312, 39 },
		{ 190, 313, 39 },
		{ 190, 314, 39 },
		{ 190, 315, 39 },
		{ 190, 316, 39 },
		{ 190, 317, 39 },
		{ 190, 318, 39 },
		{ 190, 319, 39 },
		{ 190, 320, 39 },
		{ 190, 321, 39 },
		{ 190, 322, 39 },
		{ 190, 323, 39 },
		{ 191, 309, 39 },
		{ 191, 310, 39 },
		{ 191, 311, 39 },
		{ 191, 312, 39 },
		{ 191, 313, 39 },
		{ 191, 314, 39 },
		{ 191, 315, 39 },
		{ 191, 316, 39 },
		{ 191, 317, 39 },
		{ 191, 318, 39 },
		{ 191, 319, 39 },
		{ 191, 320, 39 },
		{ 191, 321, 39 },
		{ 323, 193, 39 },
		{ 323, 194, 39 },
		{ 323, 195, 39 },
		{ 323, 196, 39 },
		{ 323, 197, 39 },
		{ 323, 198, 39 },
		{ 323, 199, 39 },
		{ 323, 200, 39 },
		{ 323, 201, 39 },
		{ 323, 202, 39 },
		{ 323, 203, 39 },
		{ 323, 204, 39 },
		{ 323, 205, 39 },
		{ 324, 191, 39 },
		{ 324, 192, 39 },
		{ 324, 193, 39 },
		{ 324, 194, 39 },
		{ 324, 195, 39 },
		{ 324, 196, 39 },
		{ 324, 197, 39 },
		{ 324, 198, 39 },
		{ 324, 199, 39 },
		{ 324, 200, 39 },
		{ 324, 201, 39 },
		{ 324, 202, 39 },
		{ 324, 203, 39 },
		{ 324, 204, 39 },
		{ 324, 205, 39 },
		{ 324, 206, 39 },
		{ 325, 190, 39 },
		{ 325, 191, 39 },
		{ 325, 192, 39 },
		{ 325, 193, 39 },
		{ 325, 194, 39 },
		{ 325, 195, 39 },
		{ 325, 196, 39 },
		{ 325, 197, 39 },
		{ 325, 198, 39 },
		{ 325, 199, 39 },
		{ 325, 200, 39 },
		{ 325, 201, 39 },
		{ 325, 202, 39 },
		{ 325, 203, 39 },
		{ 325, 204, 39 },
		{ 325, 205, 39 },
		{ 326, 189, 39 },
		{ 326, 190, 39 },
		{ 326, 191, 39 },
		{ 326, 192, 39 },
		{ 326, 193, 39 },
		{ 326, 194, 39 },
		{ 326, 195, 39 },
		{ 326, 196, 39 },
		{ 326, 197, 39 },
		{ 326, 198, 39 },
		{ 326, 199, 39 },
		{ 326, 200, 39 },
		{ 326, 201, 39 },
		{ 326, 202, 39 },
		{ 326, 203, 39 },
		{ 326, 204, 39 },
		{ 327, 188, 39 },
		{ 327, 189, 39 },
		{ 327, 190, 39 },
		{ 327, 191, 39 },
		{ 327, 192, 39 },
		{ 327, 193, 39 },
		{ 327, 194, 39 },
		{ 327, 195, 39 },
		{ 327, 196, 39 },
		{ 327, 197, 39 },
		{ 327, 198, 39 },
		{ 327, 199, 39 },
		{ 327, 200, 39 },
		{ 327, 201, 39 },
		{ 327, 202, 39 },
		{ 327, 203, 39 },
		{ 327, 204, 39 },
		{ 328, 187, 39 },
		{ 328, 188, 39 },
		{ 328, 189, 39 },
		{ 328, 190, 39 },
		{ 328, 191, 39 },
		{ 328, 192, 39 },
		{ 328, 193, 39 },
		{ 328, 194, 39 },
		{ 328, 195, 39 },
		{ 328, 196, 39 },
		{ 328, 197, 39 },
		{ 328, 198, 39 },
		{ 328, 199, 39 },
		{ 328, 200, 39 },
		{ 328, 201, 39 },
		{ 328, 202, 39 },
		{ 328, 203, 39 },
		{ 329, 186, 39 },
		{ 329, 187, 39 },
		{ 329, 188, 39 },
		{ 329, 189, 39 },
		{ 329, 190, 39 },
		{ 329, 191, 39 },
		{ 329, 192, 39 },
		{ 329, 193, 39 },
		{ 329, 194, 39 },
		{ 329, 195, 39 },
		{ 329, 196, 39 },
		{ 329, 197, 39 },
		{ 329, 198, 39 },
		{ 329, 199, 39 },
		{ 329, 200, 39 },
		{ 329, 201, 39 },
		{ 329, 202, 39 },
		{ 330, 185, 39 },
		{ 330, 186, 39 },
		{ 330, 187, 39 },
		{ 330, 188, 39 },
		{ 330, 189, 39 },
		{ 330, 190, 39 },
		{ 330, 191, 39 },
		{ 330, 192, 39 },
		{ 330, 193, 39 },
		{ 330, 194, 39 },
		{ 330, 195, 39 },
		{ 330, 196, 39 },
		{ 330, 197, 39 },
		{ 330, 198, 39 },
		{ 330, 199, 39 },
		{ 330, 200, 39 },
		{ 330, 201, 39 },
		{ 330, 202, 39 },
		{ 331, 184, 39 },
		{ 331, 185, 39 },
		{ 331, 186, 39 },
		{ 331, 187, 39 },
		{ 331, 188, 39 },
		{ 331, 189, 39 },
		{ 331, 190, 39 },
		{ 331, 191, 39 },
		{ 331, 192, 39 },
		{ 331, 193, 39 },
		{ 331, 194, 39 },
		{ 331, 195, 39 },
		{ 331, 196, 39 },
		{ 331, 197, 39 },
		{ 331, 198, 39 },
		{ 331, 199, 39 },
		{ 331, 200, 39 },
		{ 331, 201, 39 },
		{ 332, 183, 39 },
		{ 332, 184, 39 },
		{ 332, 185, 39 },
		{ 332, 186, 39 },
		{ 332, 187, 39 },
		{ 332, 188, 39 },
		{ 332, 189, 39 },
		{ 332, 190, 39 },
		{ 332, 191, 39 },
		{ 332, 192, 39 },
		{ 332, 193, 39 },
		{ 332, 194, 39 },
		{ 332, 195, 39 },
		{ 332, 196, 39 },
		{ 332, 197, 39 },
		{ 332, 198, 39 },
		{ 332, 199, 39 },
		{ 332, 200, 39 },
		{ 333, 182, 39 },
		{ 333, 183, 39 },
		{ 333, 184, 39 },
		{ 333, 185, 39 },
		{ 333, 186, 39 },
		{ 333, 187, 39 },
		{ 333, 188, 39 },
		{ 333, 189, 39 },
		{ 333, 190, 39 },
		{ 333, 191, 39 },
		{ 333, 192, 39 },
		{ 333, 193, 39 },
		{ 333, 194, 39 },
		{ 333, 195, 39 },
		{ 333, 196, 39 },
		{ 333, 197, 39 },
		{ 333, 198, 39 },
		{ 333, 199, 39 },
		{ 334, 181, 39 },
		{ 334, 182, 39 },
		{ 334, 183, 39 },
		{ 334, 184, 39 },
		{ 334, 185, 39 },
		{ 334, 186, 39 },
		{ 334, 187, 39 },
		{ 334, 188, 39 },
		{ 334, 189, 39 },
		{ 334, 190, 39 },
		{ 334, 191, 39 },
		{ 334, 192, 39 },
		{ 334, 193, 39 },
		{ 334, 194, 39 },
		{ 334, 195, 39 },
		{ 334, 196, 39 },
		{ 334, 197, 39 },
		{ 334, 198, 39 },
		{ 334, 199, 39 },
		{ 335, 180, 39 },
		{ 335, 181, 39 },
		{ 335, 182, 39 },
		{ 335, 183, 39 },
		{ 335, 184, 39 },
		{ 335, 185, 39 },
		{ 335, 186, 39 },
		{ 335, 187, 39 },
		{ 335, 188, 39 },
		{ 335, 189, 39 },
		{ 335, 190, 39 },
		{ 335, 191, 39 },
		{ 335, 192, 39 },
		{ 335, 193, 39 },
		{ 335, 194, 39 },
		{ 335, 195, 39 },
		{ 335, 196, 39 },
		{ 335, 197, 39 },
		{ 335, 198, 39 },
		{ 336, 179, 39 },
		{ 336, 180, 39 },
		{ 336, 181, 39 },
		{ 336, 182, 39 },
		{ 336, 183, 39 },
		{ 336, 184, 39 },
		{ 336, 185, 39 },
		{ 336, 186, 39 },
		{ 336, 187, 39 },
		{ 336, 188, 39 },
		{ 336, 189, 39 },
		{ 336, 190, 39 },
		{ 336, 191, 39 },
		{ 336, 192, 39 },
		{ 336, 193, 39 },
		{ 336, 194, 39 },
		{ 336, 195, 39 },
		{ 336, 196, 39 },
		{ 336, 197, 39 },
		{ 337, 178, 39 },
		{ 337, 179, 39 },
		{ 337, 180, 39 },
		{ 337, 181, 39 },
		{ 337, 182, 39 },
		{ 337, 183, 39 },
		{ 337, 184, 39 },
		{ 337, 185, 39 },
		{ 337, 186, 39 },
		{ 337, 187, 39 },
		{ 337, 188, 39 },
		{ 337, 189, 39 },
		{ 337, 190, 39 },
		{ 337, 191, 39 },
		{ 337, 192, 39 },
		{ 337, 193, 39 },
		{ 337, 194, 39 },
		{ 337, 195, 39 },
		{ 337, 196, 39 },
		{ 338, 177, 39 },
		{ 338, 178, 39 },
		{ 338, 179, 39 },
		{ 338, 180, 39 },
		{ 338, 181, 39 },
		{ 338, 182, 39 },
		{ 338, 183, 39 },
		{ 338, 184, 39 },
		{ 338, 185, 39 },
		{ 338, 186, 39 },
		{ 338, 187, 39 },
		{ 338, 188, 39 },
		{ 338, 189, 39 },
		{ 338, 190, 39 },
		{ 338, 191, 39 },
		{ 338, 192, 39 },
		{ 338, 193, 39 },
		{ 338, 194, 39 },
		{ 338, 195, 39 },
		{ 338, 196, 39 },
		{ 339, 176, 39 },
		{ 339, 177, 39 },
		{ 339, 178, 39 },
		{ 339, 179, 39 },
		{ 339, 180, 39 },
		{ 339, 181, 39 },
		{ 339, 182, 39 },
		{ 339, 183, 39 },
		{ 339, 184, 39 },
		{ 339, 185, 39 },
		{ 339, 186, 39 },
		{ 339, 187, 39 },
		{ 339, 188, 39 },
		{ 339, 189, 39 },
		{ 339, 190, 39 },
		{ 339, 191, 39 },
		{ 339, 192, 39 },
		{ 339, 193, 39 },
		{ 339, 194, 39 },
		{ 339, 195, 39 },
		{ 340, 175, 39 },
		{ 340, 176, 39 },
		{ 340, 177, 39 },
		{ 340, 178, 39 },
		{ 340, 179, 39 },
		{ 340, 180, 39 },
		{ 340, 181, 39 },
		{ 340, 182, 39 },
		{ 340, 183, 39 },
		{ 340, 184, 39 },
		{ 340, 185, 39 },
		{ 340, 186, 39 },
		{ 340, 187, 39 },
		{ 340, 188, 39 },
		{ 340, 189, 39 },
		{ 340, 190, 39 },
		{ 340, 191, 39 },
		{ 340, 192, 39 },
		{ 340, 193, 39 },
		{ 340, 194, 39 },
		{ 341, 174, 39 },
		{ 341, 175, 39 },
		{ 341, 176, 39 },
		{ 341, 177, 39 },
		{ 341, 178, 39 },
		{ 341, 179, 39 },
		{ 341, 180, 39 },
		{ 341, 181, 39 },
		{ 341, 182, 39 },
		{ 341, 183, 39 },
		{ 341, 184, 39 },
		{ 341, 185, 39 },
		{ 341, 186, 39 },
		{ 341, 187, 39 },
		{ 341, 188, 39 },
		{ 341, 189, 39 },
		{ 341, 190, 39 },
		{ 341, 191, 39 },
		{ 341, 192, 39 },
		{ 341, 193, 39 },
		{ 342, 173, 39 },
		{ 342, 174, 39 },
		{ 342, 175, 39 },
		{ 342, 176, 39 },
		{ 342, 177, 39 },
		{ 342, 178, 39 },
		{ 342, 179, 39 },
		{ 342, 180, 39 },
		{ 342, 181, 39 },
		{ 342, 182, 39 },
		{ 342, 183, 39 },
		{ 342, 184, 39 },
		{ 342, 185, 39 },
		{ 342, 186, 39 },
		{ 342, 187, 39 },
		{ 342, 188, 39 },
		{ 342, 189, 39 },
		{ 342, 190, 39 },
		{ 342, 191, 39 },
		{ 342, 192, 39 },
		{ 342, 193, 39 },
		{ 343, 172, 39 },
		{ 343, 173, 39 },
		{ 343, 174, 39 },
		{ 343, 175, 39 },
		{ 343, 176, 39 },
		{ 343, 177, 39 },
		{ 343, 178, 39 },
		{ 343, 179, 39 },
		{ 343, 180, 39 },
		{ 343, 181, 39 },
		{ 343, 182, 39 },
		{ 343, 183, 39 },
		{ 343, 184, 39 },
		{ 343, 185, 39 },
		{ 343, 186, 39 },
		{ 343, 187, 39 },
		{ 343, 188, 39 },
		{ 343, 189, 39 },
		{ 343, 190, 39 },
		{ 343, 191, 39 },
		{ 343, 192, 39 },
		{ 344, 171, 39 },
		{ 344, 172, 39 },
		{ 344, 173, 39 },
		{ 344, 174, 39 },
		{ 344, 175, 39 },
		{ 344, 176, 39 },
		{ 344, 177, 39 },
		{ 344, 178, 39 },
		{ 344, 179, 39 },
		{ 344, 180, 39 },
		{ 344, 181, 39 },
		{ 344, 182, 39 },
		{ 344, 183, 39 },
		{ 344, 184, 39 },
		{ 344, 185, 39 },
		{ 344, 186, 39 },
		{ 344, 187, 39 },
		{ 344, 188, 39 },
		{ 344, 189, 39 },
		{ 344, 190, 39 },
		{ 344, 191, 39 },
		{ 345, 170, 39 },
		{ 345, 171, 39 },
		{ 345, 172, 39 },
		{ 345, 173, 39 },
		{ 345, 174, 39 },
		{ 345, 175, 39 },
		{ 345, 176, 39 },
		{ 345, 177, 39 },
		{ 345, 178, 39 },
		{ 345, 179, 39 },
		{ 345, 180, 39 },
		{ 345, 181, 39 },
		{ 345, 182, 39 },
		{ 345, 183, 39 },
		{ 345, 184, 39 },
		{ 345, 185, 39 },
		{ 345, 186, 39 },
		{ 345, 187, 39 },
		{ 345, 188, 39 },
		{ 345, 189, 39 },
		{ 345, 190, 39 },
		{ 346, 169, 39 },
		{ 346, 170, 39 },
		{ 346, 171, 39 },
		{ 346, 172, 39 },
		{ 346, 173, 39 },
		{ 346, 174, 39 },
		{ 346, 175, 39 },
		{ 346, 176, 39 },
		{ 346, 177, 39 },
		{ 346, 178, 39 },
		{ 346, 179, 39 },
		{ 346, 180, 39 },
		{ 346, 181, 39 },
		{ 346, 182, 39 },
		{ 346, 183, 39 },
		{ 346, 184, 39 },
		{ 346, 185, 39 },
		{ 346, 186, 39 },
		{ 346, 187, 39 },
		{ 346, 188, 39 },
		{ 346, 189, 39 },
		{ 346, 190, 39 },
		{ 347, 168, 39 },
		{ 347, 169, 39 },
		{ 347, 170, 39 },
		{ 347, 171, 39 },
		{ 347, 172, 39 },
		{ 347, 173, 39 },
		{ 347, 174, 39 },
		{ 347, 175, 39 },
		{ 347, 176, 39 },
		{ 347, 177, 39 },
		{ 347, 178, 39 },
		{ 347, 179, 39 },
		{ 347, 180, 39 },
		{ 347, 181, 39 },
		{ 347, 182, 39 },
		{ 347, 183, 39 },
		{ 347, 184, 39 },
		{ 347, 185, 39 },
		{ 347, 186, 39 },
		{ 347, 187, 39 },
		{ 347, 188, 39 },
		{ 347, 189, 39 },
		{ 348, 167, 39 },
		{ 348, 168, 39 },
		{ 348, 169, 39 },
		{ 348, 170, 39 },
		{ 348, 171, 39 },
		{ 348, 172, 39 },
		{ 348, 173, 39 },
		{ 348, 174, 39 },
		{ 348, 175, 39 },
		{ 348, 176, 39 },
		{ 348, 177, 39 },
		{ 348, 178, 39 },
		{ 348, 179, 39 },
		{ 348, 180, 39 },
		{ 348, 181, 39 },
		{ 348, 182, 39 },
		{ 348, 183, 39 },
		{ 348, 184, 39 },
		{ 348, 185, 39 },
		{ 348, 186, 39 },
		{ 348, 187, 39 },
		{ 348, 188, 39 },
		{ 349, 166, 39 },
		{ 349, 167, 39 },
		{ 349, 168, 39 },
		{ 349, 169, 39 },
		{ 349, 170, 39 },
		{ 349, 171, 39 },
		{ 349, 172, 39 },
		{ 349, 173, 39 },
		{ 349, 174, 39 },
		{ 349, 175, 39 },
		{ 349, 176, 39 },
		{ 349, 177, 39 },
		{ 349, 178, 39 },
		{ 349, 179, 39 },
		{ 349, 180, 39 },
		{ 349, 181, 39 },
		{ 349, 182, 39 },
		{ 349, 183, 39 },
		{ 349, 184, 39 },
		{ 349, 185, 39 },
		{ 349, 186, 39 },
		{ 349, 187, 39 },
		{ 350, 165, 39 },
		{ 350, 166, 39 },
		{ 350, 167, 39 },
		{ 350, 168, 39 },
		{ 350, 169, 39 },
		{ 350, 170, 39 },
		{ 350, 171, 39 },
		{ 350, 172, 39 },
		{ 350, 173, 39 },
		{ 350, 174, 39 },
		{ 350, 175, 39 },
		{ 350, 176, 39 },
		{ 350, 177, 39 },
		{ 350, 178, 39 },
		{ 350, 179, 39 },
		{ 350, 180, 39 },
		{ 350, 181, 39 },
		{ 350, 182, 39 },
		{ 350, 183, 39 },
		{ 350, 184, 39 },
		{ 350, 185, 39 },
		{ 350, 186, 39 },
		{ 350, 187, 39 },
		{ 351, 164, 39 },
		{ 351, 165, 39 },
		{ 351, 166, 39 },
		{ 351, 167, 39 },
		{ 351, 168, 39 },
		{ 351, 169, 39 },
		{ 351, 170, 39 },
		{ 351, 171, 39 },
		{ 351, 172, 39 },
		{ 351, 173, 39 },
		{ 351, 174, 39 },
		{ 351, 175, 39 },
		{ 351, 176, 39 },
		{ 351, 177, 39 },
		{ 351, 178, 39 },
		{ 351, 179, 39 },
		{ 351, 180, 39 },
		{ 351, 181, 39 },
		{ 351, 182, 39 },
		{ 351, 183, 39 },
		{ 351, 184, 39 },
		{ 351, 185, 39 },
		{ 351, 186, 39 },
		{ 352, 163, 39 },
		{ 352, 164, 39 },
		{ 352, 165, 39 },
		{ 352, 166, 39 },
		{ 352, 167, 39 },
		{ 352, 168, 39 },
		{ 352, 169, 39 },
		{ 352, 170, 39 },
		{ 352, 171, 39 },
		{ 352, 172, 39 },
		{ 352, 173, 39 },
		{ 352, 174, 39 },
		{ 352, 175, 39 },
		{ 352, 176, 39 },
		{ 352, 177, 39 },
		{ 352, 178, 39 },
		{ 352, 179, 39 },
		{ 352, 180, 39 },
		{ 352, 181, 39 },
		{ 352, 182, 39 },
		{ 352, 183, 39 },
		{ 352, 184, 39 },
		{ 352, 185, 39 },
		{ 353, 162, 39 },
		{ 353, 163, 39 },
		{ 353, 164, 39 },
		{ 353, 165, 39 },
		{ 353, 166, 39 },
		{ 353, 167, 39 },
		{ 353, 168, 39 },
		{ 353, 169, 39 },
		{ 353, 170, 39 },
		{ 353, 171, 39 },
		{ 353, 172, 39 },
		{ 353, 173, 39 },
		{ 353, 174, 39 },
		{ 353, 175, 39 },
		{ 353, 176, 39 },
		{ 353, 177, 39 },
		{ 353, 178, 39 },
		{ 353, 179, 39 },
		{ 353, 180, 39 },
		{ 353, 181, 39 },
		{ 353, 182, 39 },
		{ 353, 183, 39 },
		{ 353, 184, 39 },
		{ 354, 161, 39 },
		{ 354, 162, 39 },
		{ 354, 163, 39 },
		{ 354, 164, 39 },
		{ 354, 165, 39 },
		{ 354, 166, 39 },
		{ 354, 167, 39 },
		{ 354, 168, 39 },
		{ 354, 169, 39 },
		{ 354, 170, 39 },
		{ 354, 171, 39 },
		{ 354, 172, 39 },
		{ 354, 173, 39 },
		{ 354, 174, 39 },
		{ 354, 175, 39 },
		{ 354, 176, 39 },
		{ 354, 177, 39 },
		{ 354, 178, 39 },
		{ 354, 179, 39 },
		{ 354, 180, 39 },
		{ 354, 181, 39 },
		{ 354, 182, 39 },
		{ 354, 183, 39 },
		{ 354, 184, 39 },
		{ 355, 160, 39 },
		{ 355, 161, 39 },
		{ 355, 162, 39 },
		{ 355, 163, 39 },
		{ 355, 164, 39 },
		{ 355, 165, 39 },
		{ 355, 166, 39 },
		{ 355, 167, 39 },
		{ 355, 168, 39 },
		{ 355, 169, 39 },
		{ 355, 170, 39 },
		{ 355, 171, 39 },
		{ 355, 172, 39 },
		{ 355, 173, 39 },
		{ 355, 174, 39 },
		{ 355, 175, 39 },
		{ 355, 176, 39 },
		{ 355, 177, 39 },
		{ 355, 178, 39 },
		{ 355, 179, 39 },
		{ 355, 180, 39 },
		{ 355, 181, 39 },
		{ 355, 182, 39 },
		{ 355, 183, 39 },
		{ 356, 159, 39 },
		{ 356, 160, 39 },
		{ 356, 161, 39 },
		{ 356, 162, 39 },
		{ 356, 163, 39 },
		{ 356, 164, 39 },
		{ 356, 165, 39 },
		{ 356, 166, 39 },
		{ 356, 167, 39 },
		{ 356, 168, 39 },
		{ 356, 169, 39 },
		{ 356, 170, 39 },
		{ 356, 171, 39 },
		{ 356, 172, 39 },
		{ 356, 173, 39 },
		{ 356, 174, 39 },
		{ 356, 175, 39 },
		{ 356, 176, 39 },
		{ 356, 177, 39 },
		{ 356, 178, 39 },
		{ 356, 179, 39 },
		{ 356, 180, 39 },
		{ 356, 181, 39 },
		{ 356, 182, 39 },
		{ 357, 158, 39 },
		{ 357, 159, 39 },
		{ 357, 160, 39 },
		{ 357, 161, 39 },
		{ 357, 162, 39 },
		{ 357, 163, 39 },
		{ 357, 164, 39 },
		{ 357, 165, 39 },
		{ 357, 166, 39 },
		{ 357, 167, 39 },
		{ 357, 168, 39 },
		{ 357, 169, 39 },
		{ 357, 170, 39 },
		{ 357, 171, 39 },
		{ 357, 172, 39 },
		{ 357, 173, 39 },
		{ 357, 174, 39 },
		{ 357, 175, 39 },
		{ 357, 176, 39 },
		{ 357, 177, 39 },
		{ 357, 178, 39 },
		{ 357, 179, 39 },
		{ 357, 180, 39 },
		{ 357, 181, 39 },
		{ 358, 157, 39 },
		{ 358, 158, 39 },
		{ 358, 159, 39 },
		{ 358, 160, 39 },
		{ 358, 161, 39 },
		{ 358, 162, 39 },
		{ 358, 163, 39 },
		{ 358, 164, 39 },
		{ 358, 165, 39 },
		{ 358, 166, 39 },
		{ 358, 167, 39 },
		{ 358, 168, 39 },
		{ 358, 169, 39 },
		{ 358, 170, 39 },
		{ 358, 171, 39 },
		{ 358, 172, 39 },
		{ 358, 173, 39 },
		{ 358, 174, 39 },
		{ 358, 175, 39 },
		{ 358, 176, 39 },
		{ 358, 177, 39 },
		{ 358, 178, 39 },
		{ 358, 179, 39 },
		{ 358, 180, 39 },
		{ 358, 181, 39 },
		{ 359, 156, 39 },
		{ 359, 157, 39 },
		{ 359, 158, 39 },
		{ 359, 159, 39 },
		{ 359, 160, 39 },
		{ 359, 161, 39 },
		{ 359, 162, 39 },
		{ 359, 163, 39 },
		{ 359, 164, 39 },
		{ 359, 165, 39 },
		{ 359, 166, 39 },
		{ 359, 167, 39 },
		{ 359, 168, 39 },
		{ 359, 169, 39 },
		{ 359, 170, 39 },
		{ 359, 171, 39 },
		{ 359, 172, 39 },
		{ 359, 173, 39 },
		{ 359, 174, 39 },
		{ 359, 175, 39 },
		{ 359, 176, 39 },
		{ 359, 177, 39 },
		{ 359, 178, 39 },
		{ 359, 179, 39 },
		{ 359, 180, 39 },
		{ 360, 155, 39 },
		{ 360, 156, 39 },
		{ 360, 157, 39 },
		{ 360, 158, 39 },
		{ 360, 159, 39 },
		{ 360, 160, 39 },
		{ 360, 161, 39 },
		{ 360, 162, 39 },
		{ 360, 163, 39 },
		{ 360, 164, 39 },
		{ 360, 165, 39 },
		{ 360, 166, 39 },
		{ 360, 167, 39 },
		{ 360, 168, 39 },
		{ 360, 169, 39 },
		{ 360, 170, 39 },
		{ 360, 171, 39 },
		{ 360, 172, 39 },
		{ 360, 173, 39 },
		{ 360, 174, 39 },
		{ 360, 175, 39 },
		{ 360, 176, 39 },
		{ 360, 177, 39 },
		{ 360, 178, 39 },
		{ 360, 179, 39 },
		{ 361, 154, 39 },
		{ 361, 155, 39 },
		{ 361, 156, 39 },
		{ 361, 157, 39 },
		{ 361, 158, 39 },
		{ 361, 159, 39 },
		{ 361, 160, 39 },
		{ 361, 161, 39 },
		{ 361, 162, 39 },
		{ 361, 163, 39 },
		{ 361, 164, 39 },
		{ 361, 165, 39 },
		{ 361, 166, 39 },
		{ 361, 167, 39 },
		{ 361, 168, 39 },
		{ 361, 169, 39 },
		{ 361, 170, 39 },
		{ 361, 171, 39 },
		{ 361, 172, 39 },
		{ 361, 173, 39 },
		{ 361, 174, 39 },
		{ 361, 175, 39 },
		{ 361, 176, 39 },
		{ 361, 177, 39 },
		{ 361, 178, 39 },
		{ 362, 153, 39 },
		{ 362, 154, 39 },
		{ 362, 155, 39 },
		{ 362, 156, 39 },
		{ 362, 157, 39 },
		{ 362, 158, 39 },
		{ 362, 159, 39 },
		{ 362, 160, 39 },
		{ 362, 161, 39 },
		{ 362, 162, 39 },
		{ 362, 163, 39 },
		{ 362, 164, 39 },
		{ 362, 165, 39 },
		{ 362, 166, 39 },
		{ 362, 167, 39 },
		{ 362, 168, 39 },
		{ 362, 169, 39 },
		{ 362, 170, 39 },
		{ 362, 171, 39 },
		{ 362, 172, 39 },
		{ 362, 173, 39 },
		{ 362, 174, 39 },
		{ 362, 175, 39 },
		{ 362, 176, 39 },
		{ 362, 177, 39 },
		{ 362, 178, 39 },
		{ 363, 152, 39 },
		{ 363, 153, 39 },
		{ 363, 154, 39 },
		{ 363, 155, 39 },
		{ 363, 156, 39 },
		{ 363, 157, 39 },
		{ 363, 158, 39 },
		{ 363, 159, 39 },
		{ 363, 160, 39 },
		{ 363, 161, 39 },
		{ 363, 162, 39 },
		{ 363, 163, 39 },
		{ 363, 164, 39 },
		{ 363, 165, 39 },
		{ 363, 166, 39 },
		{ 363, 167, 39 },
		{ 363, 168, 39 },
		{ 363, 169, 39 },
		{ 363, 170, 39 },
		{ 363, 171, 39 },
		{ 363, 172, 39 },
		{ 363, 173, 39 },
		{ 363, 174, 39 },
		{ 363, 175, 39 },
		{ 363, 176, 39 },
		{ 363, 177, 39 },
		{ 364, 151, 39 },
		{ 364, 152, 39 },
		{ 364, 153, 39 },
		{ 364, 154, 39 },
		{ 364, 155, 39 },
		{ 364, 156, 39 },
		{ 364, 157, 39 },
		{ 364, 158, 39 },
		{ 364, 159, 39 },
		{ 364, 160, 39 },
		{ 364, 161, 39 },
		{ 364, 162, 39 },
		{ 364, 163, 39 },
		{ 364, 164, 39 },
		{ 364, 165, 39 },
		{ 364, 166, 39 },
		{ 364, 167, 39 },
		{ 364, 168, 39 },
		{ 364, 169, 39 },
		{ 364, 170, 39 },
		{ 364, 171, 39 },
		{ 364, 172, 39 },
		{ 364, 173, 39 },
		{ 364, 174, 39 },
		{ 364, 175, 39 },
		{ 364, 176, 39 },
		{ 365, 150, 39 },
		{ 365, 151, 39 },
		{ 365, 152, 39 },
		{ 365, 153, 39 },
		{ 365, 154, 39 },
		{ 365, 155, 39 },
		{ 365, 156, 39 },
		{ 365, 157, 39 },
		{ 365, 158, 39 },
		{ 365, 159, 39 },
		{ 365, 160, 39 },
		{ 365, 161, 39 },
		{ 365, 162, 39 },
		{ 365, 163, 39 },
		{ 365, 164, 39 },
		{ 365, 165, 39 },
		{ 365, 166, 39 },
		{ 365, 167, 39 },
		{ 365, 168, 39 },
		{ 365, 169, 39 },
		{ 365, 170, 39 },
		{ 365, 171, 39 },
		{ 365, 172, 39 },
		{ 365, 173, 39 },
		{ 365, 174, 39 },
		{ 365, 175, 39 },
		{ 366, 149, 39 },
		{ 366, 150, 39 },
		{ 366, 151, 39 },
		{ 366, 152, 39 },
		{ 366, 153, 39 },
		{ 366, 154, 39 },
		{ 366, 155, 39 },
		{ 366, 156, 39 },
		{ 366, 157, 39 },
		{ 366, 158, 39 },
		{ 366, 159, 39 },
		{ 366, 160, 39 },
		{ 366, 161, 39 },
		{ 366, 162, 39 },
		{ 366, 163, 39 },
		{ 366, 164, 39 },
		{ 366, 165, 39 },
		{ 366, 166, 39 },
		{ 366, 167, 39 },
		{ 366, 168, 39 },
		{ 366, 169, 39 },
		{ 366, 170, 39 },
		{ 366, 171, 39 },
		{ 366, 172, 39 },
		{ 366, 173, 39 },
		{ 366, 174, 39 },
		{ 366, 175, 39 },
		{ 367, 148, 39 },
		{ 367, 149, 39 },
		{ 367, 150, 39 },
		{ 367, 151, 39 },
		{ 367, 152, 39 },
		{ 367, 153, 39 },
		{ 367, 154, 39 },
		{ 367, 155, 39 },
		{ 367, 156, 39 },
		{ 367, 157, 39 },
		{ 367, 158, 39 },
		{ 367, 159, 39 },
		{ 367, 160, 39 },
		{ 367, 161, 39 },
		{ 367, 162, 39 },
		{ 367, 163, 39 },
		{ 367, 164, 39 },
		{ 367, 165, 39 },
		{ 367, 166, 39 },
		{ 367, 167, 39 },
		{ 367, 168, 39 },
		{ 367, 169, 39 },
		{ 367, 170, 39 },
		{ 367, 171, 39 },
		{ 367, 172, 39 },
		{ 367, 173, 39 },
		{ 367, 174, 39 },
		{ 368, 147, 39 },
		{ 368, 148, 39 },
		{ 368, 149, 39 },
		{ 368, 150, 39 },
		{ 368, 151, 39 },
		{ 368, 152, 39 },
		{ 368, 153, 39 },
		{ 368, 154, 39 },
		{ 368, 155, 39 },
		{ 368, 156, 39 },
		{ 368, 157, 39 },
		{ 368, 158, 39 },
		{ 368, 159, 39 },
		{ 368, 160, 39 },
		{ 368, 161, 39 },
		{ 368, 162, 39 },
		{ 368, 163, 39 },
		{ 368, 164, 39 },
		{ 368, 165, 39 },
		{ 368, 166, 39 },
		{ 368, 167, 39 },
		{ 368, 168, 39 },
		{ 368, 169, 39 },
		{ 368, 170, 39 },
		{ 368, 171, 39 },
		{ 368, 172, 39 },
		{ 368, 173, 39 },
		{ 369, 146, 39 },
		{ 369, 147, 39 },
		{ 369, 148, 39 },
		{ 369, 149, 39 },
		{ 369, 150, 39 },
		{ 369, 151, 39 },
		{ 369, 152, 39 },
		{ 369, 153, 39 },
		{ 369, 154, 39 },
		{ 369, 155, 39 },
		{ 369, 156, 39 },
		{ 369, 157, 39 },
		{ 369, 158, 39 },
		{ 369, 159, 39 },
		{ 369, 160, 39 },
		{ 369, 161, 39 },
		{ 369, 162, 39 },
		{ 369, 163, 39 },
		{ 369, 164, 39 },
		{ 369, 165, 39 },
		{ 369, 166, 39 },
		{ 369, 167, 39 },
		{ 369, 168, 39 },
		{ 369, 169, 39 },
		{ 369, 170, 39 },
		{ 369, 171, 39 },
		{ 369, 172, 39 },
		{ 370, 145, 39 },
		{ 370, 146, 39 },
		{ 370, 147, 39 },
		{ 370, 148, 39 },
		{ 370, 149, 39 },
		{ 370, 150, 39 },
		{ 370, 151, 39 },
		{ 370, 152, 39 },
		{ 370, 153, 39 },
		{ 370, 154, 39 },
		{ 370, 155, 39 },
		{ 370, 156, 39 },
		{ 370, 157, 39 },
		{ 370, 158, 39 },
		{ 370, 159, 39 },
		{ 370, 160, 39 },
		{ 370, 161, 39 },
		{ 370, 162, 39 },
		{ 370, 163, 39 },
		{ 370, 164, 39 },
		{ 370, 165, 39 },
		{ 370, 166, 39 },
		{ 370, 167, 39 },
		{ 370, 168, 39 },
		{ 370, 169, 39 },
		{ 370, 170, 39 },
		{ 370, 171, 39 },
		{ 370, 172, 39 },
		{ 371, 144, 39 },
		{ 371, 145, 39 },
		{ 371, 146, 39 },
		{ 371, 147, 39 },
		{ 371, 148, 39 },
		{ 371, 149, 39 },
		{ 371, 150, 39 },
		{ 371, 151, 39 },
		{ 371, 152, 39 },
		{ 371, 153, 39 },
		{ 371, 154, 39 },
		{ 371, 155, 39 },
		{ 371, 156, 39 },
		{ 371, 157, 39 },
		{ 371, 158, 39 },
		{ 371, 159, 39 },
		{ 371, 160, 39 },
		{ 371, 161, 39 },
		{ 371, 162, 39 },
		{ 371, 163, 39 },
		{ 371, 164, 39 },
		{ 371, 165, 39 },
		{ 371, 166, 39 },
		{ 371, 167, 39 },
		{ 371, 168, 39 },
		{ 371, 169, 39 },
		{ 371, 170, 39 },
		{ 371, 171, 39 },
		{ 372, 143, 39 },
		{ 372, 144, 39 },
		{ 372, 145, 39 },
		{ 372, 146, 39 },
		{ 372, 147, 39 },
		{ 372, 148, 39 },
		{ 372, 149, 39 },
		{ 372, 150, 39 },
		{ 372, 151, 39 },
		{ 372, 152, 39 },
		{ 372, 153, 39 },
		{ 372, 154, 39 },
		{ 372, 155, 39 },
		{ 372, 156, 39 },
		{ 372, 157, 39 },
		{ 372, 158, 39 },
		{ 372, 159, 39 },
		{ 372, 160, 39 },
		{ 372, 161, 39 },
		{ 372, 162, 39 },
		{ 372, 163, 39 },
		{ 372, 164, 39 },
		{ 372, 165, 39 },
		{ 372, 166, 39 },
		{ 372, 167, 39 },
		{ 372, 168, 39 },
		{ 372, 169, 39 },
		{ 372, 170, 39 },
		{ 373, 142, 39 },
		{ 373, 143, 39 },
		{ 373, 144, 39 },
		{ 373, 145, 39 },
		{ 373, 146, 39 },
		{ 373, 147, 39 },
		{ 373, 148, 39 },
		{ 373, 149, 39 },
		{ 373, 150, 39 },
		{ 373, 151, 39 },
		{ 373, 152, 39 },
		{ 373, 153, 39 },
		{ 373, 154, 39 },
		{ 373, 155, 39 },
		{ 373, 156, 39 },
		{ 373, 157, 39 },
		{ 373, 158, 39 },
		{ 373, 159, 39 },
		{ 373, 160, 39 },
		{ 373, 161, 39 },
		{ 373, 162, 39 },
		{ 373, 163, 39 },
		{ 373, 164, 39 },
		{ 373, 165, 39 },
		{ 373, 166, 39 },
		{ 373, 167, 39 },
		{ 373, 168, 39 },
		{ 373, 169, 39 },
		{ 374, 141, 39 },
		{ 374, 142, 39 },
		{ 374, 143, 39 },
		{ 374, 144, 39 },
		{ 374, 145, 39 },
		{ 374, 146, 39 },
		{ 374, 147, 39 },
		{ 374, 148, 39 },
		{ 374, 149, 39 },
		{ 374, 150, 39 },
		{ 374, 151, 39 },
		{ 374, 152, 39 },
		{ 374, 153, 39 },
		{ 374, 154, 39 },
		{ 374, 155, 39 },
		{ 374, 156, 39 },
		{ 374, 157, 39 },
		{ 374, 158, 39 },
		{ 374, 159, 39 },
		{ 374, 160, 39 },
		{ 374, 161, 39 },
		{ 374, 162, 39 },
		{ 374, 163, 39 },
		{ 374, 164, 39 },
		{ 374, 165, 39 },
		{ 374, 166, 39 },
		{ 374, 167, 39 },
		{ 374, 168, 39 },
		{ 374, 169, 39 },
		{ 375, 140, 39 },
		{ 375, 141, 39 },
		{ 375, 142, 39 },
		{ 375, 143, 39 },
		{ 375, 144, 39 },
		{ 375, 145, 39 },
		{ 375, 146, 39 },
		{ 375, 147, 39 },
		{ 375, 148, 39 },
		{ 375, 149, 39 },
		{ 375, 150, 39 },
		{ 375, 151, 39 },
		{ 375, 152, 39 },
		{ 375, 153, 39 },
		{ 375, 154, 39 },
		{ 375, 155, 39 },
		{ 375, 156, 39 },
		{ 375, 157, 39 },
		{ 375, 158, 39 },
		{ 375, 159, 39 },
		{ 375, 160, 39 },
		{ 375, 161, 39 },
		{ 375, 162, 39 },
		{ 375, 163, 39 },
		{ 375, 164, 39 },
		{ 375, 165, 39 },
		{ 375, 166, 39 },
		{ 375, 167, 39 },
		{ 375, 168, 39 },
		{ 376, 139, 39 },
		{ 376, 140, 39 },
		{ 376, 141, 39 },
		{ 376, 142, 39 },
		{ 376, 143, 39 },
		{ 376, 144, 39 },
		{ 376, 145, 39 },
		{ 376, 146, 39 },
		{ 376, 147, 39 },
		{ 376, 148, 39 },
		{ 376, 149, 39 },
		{ 376, 150, 39 },
		{ 376, 151, 39 },
		{ 376, 152, 39 },
		{ 376, 153, 39 },
		{ 376, 154, 39 },
		{ 376, 155, 39 },
		{ 376, 156, 39 },
		{ 376, 157, 39 },
		{ 376, 158, 39 },
		{ 376, 159, 39 },
		{ 376, 160, 39 },
		{ 376, 161, 39 },
		{ 376, 162, 39 },
		{ 376, 163, 39 },
		{ 376, 164, 39 },
		{ 376, 165, 39 },
		{ 376, 166, 39 },
		{ 376, 167, 39 },
		{ 377, 138, 39 },
		{ 377, 139, 39 },
		{ 377, 140, 39 },
		{ 377, 141, 39 },
		{ 377, 142, 39 },
		{ 377, 143, 39 },
		{ 377, 144, 39 },
		{ 377, 145, 39 },
		{ 377, 146, 39 },
		{ 377, 147, 39 },
		{ 377, 148, 39 },
		{ 377, 149, 39 },
		{ 377, 150, 39 },
		{ 377, 151, 39 },
		{ 377, 152, 39 },
		{ 377, 153, 39 },
		{ 377, 154, 39 },
		{ 377, 155, 39 },
		{ 377, 156, 39 },
		{ 377, 157, 39 },
		{ 377, 158, 39 },
		{ 377, 159, 39 },
		{ 377, 160, 39 },
		{ 377, 161, 39 },
		{ 377, 162, 39 },
		{ 377, 163, 39 },
		{ 377, 164, 39 },
		{ 377, 165, 39 },
		{ 377, 166, 39 },
		{ 378, 137, 39 },
		{ 378, 138, 39 },
		{ 378, 139, 39 },
		{ 378, 140, 39 },
		{ 378, 141, 39 },
		{ 378, 142, 39 },
		{ 378, 143, 39 },
		{ 378, 144, 39 },
		{ 378, 145, 39 },
		{ 378, 146, 39 },
		{ 378, 147, 39 },
		{ 378, 148, 39 },
		{ 378, 149, 39 },
		{ 378, 150, 39 },
		{ 378, 151, 39 },
		{ 378, 152, 39 },
		{ 378, 153, 39 },
		{ 378, 154, 39 },
		{ 378, 155, 39 },
		{ 378, 156, 39 },
		{ 378, 157, 39 },
		{ 378, 158, 39 },
		{ 378, 159, 39 },
		{ 378, 160, 39 },
		{ 378, 161, 39 },
		{ 378, 162, 39 },
		{ 378, 163, 39 },
		{ 378, 164, 39 },
		{ 378, 165, 39 },
		{ 378, 166, 39 },
		{ 379, 136, 39 },
		{ 379, 137, 39 },
		{ 379, 138, 39 },
		{ 379, 139, 39 },
		{ 379, 140, 39 },
		{ 379, 141, 39 },
		{ 379, 142, 39 },
		{ 379, 143, 39 },
		{ 379, 144, 39 },
		{ 379, 145, 39 },
		{ 379, 146, 39 },
		{ 379, 147, 39 },
		{ 379, 148, 39 },
		{ 379, 149, 39 },
		{ 379, 150, 39 },
		{ 379, 151, 39 },
		{ 379, 152, 39 },
		{ 379, 153, 39 },
		{ 379, 154, 39 },
		{ 379, 155, 39 },
		{ 379, 156, 39 },
		{ 379, 157, 39 },
		{ 379, 158, 39 },
		{ 379, 159, 39 },
		{ 379, 160, 39 },
		{ 379, 161, 39 },
		{ 379, 162, 39 },
		{ 379, 163, 39 },
		{ 379, 164, 39 },
		{ 379, 165, 39 },
		{ 380, 135, 39 },
		{ 380, 136, 39 },
		{ 380, 137, 39 },
		{ 380, 138, 39 },
		{ 380, 139, 39 },
		{ 380, 140, 39 },
		{ 380, 141, 39 },
		{ 380, 142, 39 },
		{ 380, 143, 39 },
		{ 380, 144, 39 },
		{ 380, 145, 39 },
		{ 380, 146, 39 },
		{ 380, 147, 39 },
		{ 380, 148, 39 },
		{ 380, 149, 39 },
		{ 380, 150, 39 },
		{ 380, 151, 39 },
		{ 380, 152, 39 },
		{ 380, 153, 39 },
		{ 380, 154, 39 },
		{ 380, 155, 39 },
		{ 380, 156, 39 },
		{ 380, 157, 39 },
		{ 380, 158, 39 },
		{ 380, 159, 39 },
		{ 380, 160, 39 },
		{ 380, 161, 39 },
		{ 380, 162, 39 },
		{ 380, 163, 39 },
		{ 380, 164, 39 },
		{ 381, 134, 39 },
		{ 381, 135, 39 },
		{ 381, 136, 39 },
		{ 381, 137, 39 },
		{ 381, 138, 39 },
		{ 381, 139, 39 },
		{ 381, 140, 39 },
		{ 381, 141, 39 },
		{ 381, 142, 39 },
		{ 381, 143, 39 },
		{ 381, 144, 39 },
		{ 381, 145, 39 },
		{ 381, 146, 39 },
		{ 381, 147, 39 },
		{ 381, 148, 39 },
		{ 381, 149, 39 },
		{ 381, 150, 39 },
		{ 381, 151, 39 },
		{ 381, 152, 39 },
		{ 381, 153, 39 },
		{ 381, 154, 39 },
		{ 381, 155, 39 },
		{ 381, 156, 39 },
		{ 381, 157, 39 },
		{ 381, 158, 39 },
		{ 381, 159, 39 },
		{ 381, 160, 39 },
		{ 381, 161, 39 },
		{ 381, 162, 39 },
		{ 381, 163, 39 },
		{ 382, 133, 39 },
		{ 382, 134, 39 },
		{ 382, 135, 39 },
		{ 382, 136, 39 },
		{ 382, 137, 39 },
		{ 382, 138, 39 },
		{ 382, 139, 39 },
		{ 382, 140, 39 },
		{ 382, 141, 39 },
		{ 382, 142, 39 },
		{ 382, 143, 39 },
		{ 382, 144, 39 },
		{ 382, 145, 39 },
		{ 382, 146, 39 },
		{ 382, 147, 39 },
		{ 382, 148, 39 },
		{ 382, 149, 39 },
		{ 382, 150, 39 },
		{ 382, 151, 39 },
		{ 382, 152, 39 },
		{ 382, 153, 39 },
		{ 382, 154, 39 },
		{ 382, 155, 39 },
		{ 382, 156, 39 },
		{ 382, 157, 39 },
		{ 382, 158, 39 },
		{ 382, 159, 39 },
		{ 382, 160, 39 },
		{ 382, 161, 39 },
		{ 382, 162, 39 },
		{ 382, 163, 39 },
		{ 383, 132, 39 },
		{ 383, 133, 39 },
		{ 383, 134, 39 },
		{ 383, 135, 39 },
		{ 383, 136, 39 },
		{ 383, 137, 39 },
		{ 383, 138, 39 },
		{ 383, 139, 39 },
		{ 383, 140, 39 },
		{ 383, 141, 39 },
		{ 383, 142, 39 },
		{ 383, 143, 39 },
		{ 383, 144, 39 },
		{ 383, 145, 39 },
		{ 383, 146, 39 },
		{ 383, 147, 39 },
		{ 383, 148, 39 },
		{ 383, 149, 39 },
		{ 383, 150, 39 },
		{ 383, 151, 39 },
		{ 383, 152, 39 },
		{ 383, 153, 39 },
		{ 383, 154, 39 },
		{ 383, 155, 39 },
		{ 383, 156, 39 },
		{ 383, 157, 39 },
		{ 383, 158, 39 },
		{ 383, 159, 39 },
		{ 383, 160, 39 },
		{ 383, 161, 39 },
		{ 383, 162, 39 },
		{ 384, 131, 39 },
		{ 384, 132, 39 },
		{ 384, 133, 39 },
		{ 384, 134, 39 },
		{ 384, 135, 39 },
		{ 384, 136, 39 },
		{ 384, 137, 39 },
		{ 384, 138, 39 },
		{ 384, 139, 39 },
		{ 384, 140, 39 },
		{ 384, 141, 39 },
		{ 384, 142, 39 },
		{ 384, 143, 39 },
		{ 384, 144, 39 },
		{ 384, 145, 39 },
		{ 384, 146, 39 },
		{ 384, 147, 39 },
		{ 384, 148, 39 },
		{ 384, 149, 39 },
		{ 384, 150, 39 },
		{ 384, 151, 39 },
		{ 384, 152, 39 },
		{ 384, 153, 39 },
		{ 384, 154, 39 },
		{ 384, 155, 39 },
		{ 384, 156, 39 },
		{ 384, 157, 39 },
		{ 384, 158, 39 },
		{ 384, 159, 39 },
		{ 384, 160, 39 },
		{ 384, 161, 39 },
		{ 385, 130, 39 },
		{ 385, 131, 39 },
		{ 385, 132, 39 },
		{ 385, 133, 39 },
		{ 385, 134, 39 },
		{ 385, 135, 39 },
		{ 385, 136, 39 },
		{ 385, 137, 39 },
		{ 385, 138, 39 },
		{ 385, 139, 39 },
		{ 385, 140, 39 },
		{ 385, 141, 39 },
		{ 385, 142, 39 },
		{ 385, 143, 39 },
		{ 385, 144, 39 },
		{ 385, 145, 39 },
		{ 385, 146, 39 },
		{ 385, 147, 39 },
		{ 385, 148, 39 },
		{ 385, 149, 39 },
		{ 385, 150, 39 },
		{ 385, 151, 39 },
		{ 385, 152, 39 },
		{ 385, 153, 39 },
		{ 385, 154, 39 },
		{ 385, 155, 39 },
		{ 385, 156, 39 },
		{ 385, 157, 39 },
		{ 385, 158, 39 },
		{ 385, 159, 39 },
		{ 385, 160, 39 },
		{ 386, 129, 39 },
		{ 386, 130, 39 },
		{ 386, 131, 39 },
		{ 386, 132, 39 },
		{ 386, 133, 39 },
		{ 386, 134, 39 },
		{ 386, 135, 39 },
		{ 386, 136, 39 },
		{ 386, 137, 39 },
		{ 386, 138, 39 },
		{ 386, 139, 39 },
		{ 386, 140, 39 },
		{ 386, 141, 39 },
		{ 386, 142, 39 },
		{ 386, 143, 39 },
		{ 386, 144, 39 },
		{ 386, 145, 39 },
		{ 386, 146, 39 },
		{ 386, 147, 39 },
		{ 386, 148, 39 },
		{ 386, 149, 39 },
		{ 386, 150, 39 },
		{ 386, 151, 39 },
		{ 386, 152, 39 },
		{ 386, 153, 39 },
		{ 386, 154, 39 },
		{ 386, 155, 39 },
		{ 386, 156, 39 },
		{ 386, 157, 39 },
		{ 386, 158, 39 },
		{ 386, 159, 39 },
		{ 386, 160, 39 },
		{ 387, 128, 39 },
		{ 387, 129, 39 },
		{ 387, 130, 39 },
		{ 387, 131, 39 },
		{ 387, 132, 39 },
		{ 387, 133, 39 },
		{ 387, 134, 39 },
		{ 387, 135, 39 },
		{ 387, 136, 39 },
		{ 387, 137, 39 },
		{ 387, 138, 39 },
		{ 387, 139, 39 },
		{ 387, 140, 39 },
		{ 387, 141, 39 },
		{ 387, 142, 39 },
		{ 387, 143, 39 },
		{ 387, 144, 39 },
		{ 387, 145, 39 },
		{ 387, 146, 39 },
		{ 387, 147, 39 },
		{ 387, 148, 39 },
		{ 387, 149, 39 },
		{ 387, 150, 39 },
		{ 387, 151, 39 },
		{ 387, 152, 39 },
		{ 387, 153, 39 },
		{ 387, 154, 39 },
		{ 387, 155, 39 },
		{ 387, 156, 39 },
		{ 387, 157, 39 },
		{ 387, 158, 39 },
		{ 387, 159, 39 },
		{ 388, 127, 39 },
		{ 388, 128, 39 },
		{ 388, 129, 39 },
		{ 388, 130, 39 },
		{ 388, 131, 39 },
		{ 388, 132, 39 },
		{ 388, 133, 39 },
		{ 388, 134, 39 },
		{ 388, 135, 39 },
		{ 388, 136, 39 },
		{ 388, 137, 39 },
		{ 388, 138, 39 },
		{ 388, 139, 39 },
		{ 388, 140, 39 },
		{ 388, 141, 39 },
		{ 388, 142, 39 },
		{ 388, 143, 39 },
		{ 388, 144, 39 },
		{ 388, 145, 39 },
		{ 388, 146, 39 },
		{ 388, 147, 39 },
		{ 388, 148, 39 },
		{ 388, 149, 39 },
		{ 388, 150, 39 },
		{ 388, 151, 39 },
		{ 388, 152, 39 },
		{ 388, 153, 39 },
		{ 388, 154, 39 },
		{ 388, 155, 39 },
		{ 388, 156, 39 },
		{ 388, 157, 39 },
		{ 388, 158, 39 },
		{ 389, 126, 39 },
		{ 389, 127, 39 },
		{ 389, 128, 39 },
		{ 389, 129, 39 },
		{ 389, 130, 39 },
		{ 389, 131, 39 },
		{ 389, 132, 39 },
		{ 389, 133, 39 },
		{ 389, 134, 39 },
		{ 389, 135, 39 },
		{ 389, 136, 39 },
		{ 389, 137, 39 },
		{ 389, 138, 39 },
		{ 389, 139, 39 },
		{ 389, 140, 39 },
		{ 389, 141, 39 },
		{ 389, 142, 39 },
		{ 389, 143, 39 },
		{ 389, 144, 39 },
		{ 389, 145, 39 },
		{ 389, 146, 39 },
		{ 389, 147, 39 },
		{ 389, 148, 39 },
		{ 389, 149, 39 },
		{ 389, 150, 39 },
		{ 389, 151, 39 },
		{ 389, 152, 39 },
		{ 389, 153, 39 },
		{ 389, 154, 39 },
		{ 389, 155, 39 },
		{ 389, 156, 39 },
		{ 389, 157, 39 },
		{ 390, 125, 39 },
		{ 390, 126, 39 },
		{ 390, 127, 39 },
		{ 390, 128, 39 },
		{ 390, 129, 39 },
		{ 390, 130, 39 },
		{ 390, 131, 39 },
		{ 390, 132, 39 },
		{ 390, 133, 39 },
		{ 390, 134, 39 },
		{ 390, 135, 39 },
		{ 390, 136, 39 },
		{ 390, 137, 39 },
		{ 390, 138, 39 },
		{ 390, 139, 39 },
		{ 390, 140, 39 },
		{ 390, 141, 39 },
		{ 390, 142, 39 },
		{ 390, 143, 39 },
		{ 390, 144, 39 },
		{ 390, 145, 39 },
		{ 390, 146, 39 },
		{ 390, 147, 39 },
		{ 390, 148, 39 },
		{ 390, 149, 39 },
		{ 390, 150, 39 },
		{ 390, 151, 39 },
		{ 390, 152, 39 },
		{ 390, 153, 39 },
		{ 390, 154, 39 },
		{ 390, 155, 39 },
		{ 390, 156, 39 },
		{ 390, 157, 39 },
		{ 391, 124, 39 },
		{ 391, 125, 39 },
		{ 391, 126, 39 },
		{ 391, 127, 39 },
		{ 391, 128, 39 },
		{ 391, 129, 39 },
		{ 391, 130, 39 },
		{ 391, 131, 39 },
		{ 391, 132, 39 },
		{ 391, 133, 39 },
		{ 391, 134, 39 },
		{ 391, 135, 39 },
		{ 391, 136, 39 },
		{ 391, 137, 39 },
		{ 391, 138, 39 },
		{ 391, 139, 39 },
		{ 391, 140, 39 },
		{ 391, 141, 39 },
		{ 391, 142, 39 },
		{ 391, 143, 39 },
		{ 391, 144, 39 },
		{ 391, 145, 39 },
		{ 391, 146, 39 },
		{ 391, 147, 39 },
		{ 391, 148, 39 },
		{ 391, 149, 39 },
		{ 391, 150, 39 },
		{ 391, 151, 39 },
		{ 391, 152, 39 },
		{ 391, 153, 39 },
		{ 391, 154, 39 },
		{ 391, 155, 39 },
		{ 391, 156, 39 },
		{ 392, 123, 39 },
		{ 392, 124, 39 },
		{ 392, 125, 39 },
		{ 392, 126, 39 },
		{ 392, 127, 39 },
		{ 392, 128, 39 },
		{ 392, 129, 39 },
		{ 392, 130, 39 },
		{ 392, 131, 39 },
		{ 392, 132, 39 },
		{ 392, 133, 39 },
		{ 392, 134, 39 },
		{ 392, 135, 39 },
		{ 392, 136, 39 },
		{ 392, 137, 39 },
		{ 392, 138, 39 },
		{ 392, 139, 39 },
		{ 392, 140, 39 },
		{ 392, 141, 39 },
		{ 392, 142, 39 },
		{ 392, 143, 39 },
		{ 392, 144, 39 },
		{ 392, 145, 39 },
		{ 392, 146, 39 },
		{ 392, 147, 39 },
		{ 392, 148, 39 },
		{ 392, 149, 39 },
		{ 392, 150, 39 },
		{ 392, 151, 39 },
		{ 392, 152, 39 },
		{ 392, 153, 39 },
		{ 392, 154, 39 },
		{ 392, 155, 39 },
		{ 393, 122, 39 },
		{ 393, 123, 39 },
		{ 393, 124, 39 },
		{ 393, 125, 39 },
		{ 393, 126, 39 },
		{ 393, 127, 39 },
		{ 393, 128, 39 },
		{ 393, 129, 39 },
		{ 393, 130, 39 },
		{ 393, 131, 39 },
		{ 393, 132, 39 },
		{ 393, 133, 39 },
		{ 393, 134, 39 },
		{ 393, 135, 39 },
		{ 393, 136, 39 },
		{ 393, 137, 39 },
		{ 393, 138, 39 },
		{ 393, 139, 39 },
		{ 393, 140, 39 },
		{ 393, 141, 39 },
		{ 393, 142, 39 },
		{ 393, 143, 39 },
		{ 393, 144, 39 },
		{ 393, 145, 39 },
		{ 393, 146, 39 },
		{ 393, 147, 39 },
		{ 393, 148, 39 },
		{ 393, 149, 39 },
		{ 393, 150, 39 },
		{ 393, 151, 39 },
		{ 393, 152, 39 },
		{ 393, 153, 39 },
		{ 393, 154, 39 },
		{ 394, 121, 39 },
		{ 394, 122, 39 },
		{ 394, 123, 39 },
		{ 394, 124, 39 },
		{ 394, 125, 39 },
		{ 394, 126, 39 },
		{ 394, 127, 39 },
		{ 394, 128, 39 },
		{ 394, 129, 39 },
		{ 394, 130, 39 },
		{ 394, 131, 39 },
		{ 394, 132, 39 },
		{ 394, 133, 39 },
		{ 394, 134, 39 },
		{ 394, 135, 39 },
		{ 394, 136, 39 },
		{ 394, 137, 39 },
		{ 394, 138, 39 },
		{ 394, 139, 39 },
		{ 394, 140, 39 },
		{ 394, 141, 39 },
		{ 394, 142, 39 },
		{ 394, 143, 39 },
		{ 394, 144, 39 },
		{ 394, 145, 39 },
		{ 394, 146, 39 },
		{ 394, 147, 39 },
		{ 394, 148, 39 },
		{ 394, 149, 39 },
		{ 394, 150, 39 },
		{ 394, 151, 39 },
		{ 394, 152, 39 },
		{ 394, 153, 39 },
		{ 394, 154, 39 },
		{ 395, 120, 39 },
		{ 395, 121, 39 },
		{ 395, 122, 39 },
		{ 395, 123, 39 },
		{ 395, 124, 39 },
		{ 395, 125, 39 },
		{ 395, 126, 39 },
		{ 395, 127, 39 },
		{ 395, 128, 39 },
		{ 395, 129, 39 },
		{ 395, 130, 39 },
		{ 395, 131, 39 },
		{ 395, 132, 39 },
		{ 395, 133, 39 },
		{ 395, 134, 39 },
		{ 395, 135, 39 },
		{ 395, 136, 39 },
		{ 395, 137, 39 },
		{ 395, 138, 39 },
		{ 395, 139, 39 },
		{ 395, 140, 39 },
		{ 395, 141, 39 },
		{ 395, 142, 39 },
		{ 395, 143, 39 },
		{ 395, 144, 39 },
		{ 395, 145, 39 },
		{ 395, 146, 39 },
		{ 395, 147, 39 },
		{ 395, 148, 39 },
		{ 395, 149, 39 },
		{ 395, 150, 39 },
		{ 395, 151, 39 },
		{ 395, 152, 39 },
		{ 395, 153, 39 },
		{ 396, 119, 39 },
		{ 396, 120, 39 },
		{ 396, 121, 39 },
		{ 396, 122, 39 },
		{ 396, 123, 39 },
		{ 396, 124, 39 },
		{ 396, 125, 39 },
		{ 396, 126, 39 },
		{ 396, 127, 39 },
		{ 396, 128, 39 },
		{ 396, 129, 39 },
		{ 396, 130, 39 },
		{ 396, 131, 39 },
		{ 396, 132, 39 },
		{ 396, 133, 39 },
		{ 396, 134, 39 },
		{ 396, 135, 39 },
		{ 396, 136, 39 },
		{ 396, 137, 39 },
		{ 396, 138, 39 },
		{ 396, 139, 39 },
		{ 396, 140, 39 },
		{ 396, 141, 39 },
		{ 396, 142, 39 },
		{ 396, 143, 39 },
		{ 396, 144, 39 },
		{ 396, 145, 39 },
		{ 396, 146, 39 },
		{ 396, 147, 39 },
		{ 396, 148, 39 },
		{ 396, 149, 39 },
		{ 396, 150, 39 },
		{ 396, 151, 39 },
		{ 396, 152, 39 },
		{ 397, 118, 39 },
		{ 397, 119, 39 },
		{ 397, 120, 39 },
		{ 397, 121, 39 },
		{ 397, 122, 39 },
		{ 397, 123, 39 },
		{ 397, 124, 39 },
		{ 397, 125, 39 },
		{ 397, 126, 39 },
		{ 397, 127, 39 },
		{ 397, 128, 39 },
		{ 397, 129, 39 },
		{ 397, 130, 39 },
		{ 397, 131, 39 },
		{ 397, 132, 39 },
		{ 397, 133, 39 },
		{ 397, 134, 39 },
		{ 397, 135, 39 },
		{ 397, 136, 39 },
		{ 397, 137, 39 },
		{ 397, 138, 39 },
		{ 397, 139, 39 },
		{ 397, 140, 39 },
		{ 397, 141, 39 },
		{ 397, 142, 39 },
		{ 397, 143, 39 },
		{ 397, 144, 39 },
		{ 397, 145, 39 },
		{ 397, 146, 39 },
		{ 397, 147, 39 },
		{ 397, 148, 39 },
		{ 397, 149, 39 },
		{ 397, 150, 39 },
		{ 397, 151, 39 },
		{ 398, 117, 39 },
		{ 398, 118, 39 },
		{ 398, 119, 39 },
		{ 398, 120, 39 },
		{ 398, 121, 39 },
		{ 398, 122, 39 },
		{ 398, 123, 39 },
		{ 398, 124, 39 },
		{ 398, 125, 39 },
		{ 398, 126, 39 },
		{ 398, 127, 39 },
		{ 398, 128, 39 },
		{ 398, 129, 39 },
		{ 398, 130, 39 },
		{ 398, 131, 39 },
		{ 398, 132, 39 },
		{ 398, 133, 39 },
		{ 398, 134, 39 },
		{ 398, 135, 39 },
		{ 398, 136, 39 },
		{ 398, 137, 39 },
		{ 398, 138, 39 },
		{ 398, 139, 39 },
		{ 398, 140, 39 },
		{ 398, 141, 39 },
		{ 398, 142, 39 },
		{ 398, 143, 39 },
		{ 398, 144, 39 },
		{ 398, 145, 39 },
		{ 398, 146, 39 },
		{ 398, 147, 39 },
		{ 398, 148, 39 },
		{ 398, 149, 39 },
		{ 398, 150, 39 },
		{ 398, 151, 39 },
		{ 399, 116, 39 },
		{ 399, 117, 39 },
		{ 399, 118, 39 },
		{ 399, 119, 39 },
		{ 399, 120, 39 },
		{ 399, 121, 39 },
		{ 399, 122, 39 },
		{ 399, 123, 39 },
		{ 399, 124, 39 },
		{ 399, 125, 39 },
		{ 399, 126, 39 },
		{ 399, 127, 39 },
		{ 399, 128, 39 },
		{ 399, 129, 39 },
		{ 399, 130, 39 },
		{ 399, 131, 39 },
		{ 399, 132, 39 },
		{ 399, 133, 39 },
		{ 399, 134, 39 },
		{ 399, 135, 39 },
		{ 399, 136, 39 },
		{ 399, 137, 39 },
		{ 399, 138, 39 },
		{ 399, 139, 39 },
		{ 399, 140, 39 },
		{ 399, 141, 39 },
		{ 399, 142, 39 },
		{ 399, 143, 39 },
		{ 399, 144, 39 },
		{ 399, 145, 39 },
		{ 399, 146, 39 },
		{ 399, 147, 39 },
		{ 399, 148, 39 },
		{ 399, 149, 39 },
		{ 399, 150, 39 },
		{ 400, 115, 39 },
		{ 400, 116, 39 },
		{ 400, 117, 39 },
		{ 400, 118, 39 },
		{ 400, 119, 39 },
		{ 400, 120, 39 },
		{ 400, 121, 39 },
		{ 400, 122, 39 },
		{ 400, 123, 39 },
		{ 400, 124, 39 },
		{ 400, 125, 39 },
		{ 400, 126, 39 },
		{ 400, 127, 39 },
		{ 400, 128, 39 },
		{ 400, 129, 39 },
		{ 400, 130, 39 },
		{ 400, 131, 39 },
		{ 400, 132, 39 },
		{ 400, 133, 39 },
		{ 400, 134, 39 },
		{ 400, 135, 39 },
		{ 400, 136, 39 },
		{ 400, 137, 39 },
		{ 400, 138, 39 },
		{ 400, 139, 39 },
		{ 400, 140, 39 },
		{ 400, 141, 39 },
		{ 400, 142, 39 },
		{ 400, 143, 39 },
		{ 400, 144, 39 },
		{ 400, 145, 39 },
		{ 400, 146, 39 },
		{ 400, 147, 39 },
		{ 400, 148, 39 },
		{ 400, 149, 39 },
		{ 401, 114, 39 },
		{ 401, 115, 39 },
		{ 401, 116, 39 },
		{ 401, 117, 39 },
		{ 401, 118, 39 },
		{ 401, 119, 39 },
		{ 401, 120, 39 },
		{ 401, 121, 39 },
		{ 401, 122, 39 },
		{ 401, 123, 39 },
		{ 401, 124, 39 },
		{ 401, 125, 39 },
		{ 401, 126, 39 },
		{ 401, 127, 39 },
		{ 401, 128, 39 },
		{ 401, 129, 39 },
		{ 401, 130, 39 },
		{ 401, 131, 39 },
		{ 401, 132, 39 },
		{ 401, 133, 39 },
		{ 401, 134, 39 },
		{ 401, 135, 39 },
		{ 401, 136, 39 },
		{ 401, 137, 39 },
		{ 401, 138, 39 },
		{ 401, 139, 39 },
		{ 401, 140, 39 },
		{ 401, 141, 39 },
		{ 401, 142, 39 },
		{ 401, 143, 39 },
		{ 401, 144, 39 },
		{ 401, 145, 39 },
		{ 401, 146, 39 },
		{ 401, 147, 39 },
		{ 401, 148, 39 },
		{ 402, 113, 39 },
		{ 402, 114, 39 },
		{ 402, 115, 39 },
		{ 402, 116, 39 },
		{ 402, 117, 39 },
		{ 402, 118, 39 },
		{ 402, 119, 39 },
		{ 402, 120, 39 },
		{ 402, 121, 39 },
		{ 402, 122, 39 },
		{ 402, 123, 39 },
		{ 402, 124, 39 },
		{ 402, 125, 39 },
		{ 402, 126, 39 },
		{ 402, 127, 39 },
		{ 402, 128, 39 },
		{ 402, 129, 39 },
		{ 402, 130, 39 },
		{ 402, 131, 39 },
		{ 402, 132, 39 },
		{ 402, 133, 39 },
		{ 402, 134, 39 },
		{ 402, 135, 39 },
		{ 402, 136, 39 },
		{ 402, 137, 39 },
		{ 402, 138, 39 },
		{ 402, 139, 39 },
		{ 402, 140, 39 },
		{ 402, 141, 39 },
		{ 402, 142, 39 },
		{ 402, 143, 39 },
		{ 402, 144, 39 },
		{ 402, 145, 39 },
		{ 402, 146, 39 },
		{ 402, 147, 39 },
		{ 402, 148, 39 },
		{ 403, 112, 39 },
		{ 403, 113, 39 },
		{ 403, 114, 39 },
		{ 403, 115, 39 },
		{ 403, 116, 39 },
		{ 403, 117, 39 },
		{ 403, 118, 39 },
		{ 403, 119, 39 },
		{ 403, 120, 39 },
		{ 403, 121, 39 },
		{ 403, 122, 39 },
		{ 403, 123, 39 },
		{ 403, 124, 39 },
		{ 403, 125, 39 },
		{ 403, 126, 39 },
		{ 403, 127, 39 },
		{ 403, 128, 39 },
		{ 403, 129, 39 },
		{ 403, 130, 39 },
		{ 403, 131, 39 },
		{ 403, 132, 39 },
		{ 403, 133, 39 },
		{ 403, 134, 39 },
		{ 403, 135, 39 },
		{ 403, 136, 39 },
		{ 403, 137, 39 },
		{ 403, 138, 39 },
		{ 403, 139, 39 },
		{ 403, 140, 39 },
		{ 403, 141, 39 },
		{ 403, 142, 39 },
		{ 403, 143, 39 },
		{ 403, 144, 39 },
		{ 403, 145, 39 },
		{ 403, 146, 39 },
		{ 403, 147, 39 },
		{ 404, 111, 39 },
		{ 404, 112, 39 },
		{ 404, 113, 39 },
		{ 404, 114, 39 },
		{ 404, 115, 39 },
		{ 404, 116, 39 },
		{ 404, 117, 39 },
		{ 404, 118, 39 },
		{ 404, 119, 39 },
		{ 404, 120, 39 },
		{ 404, 121, 39 },
		{ 404, 122, 39 },
		{ 404, 123, 39 },
		{ 404, 124, 39 },
		{ 404, 125, 39 },
		{ 404, 126, 39 },
		{ 404, 127, 39 },
		{ 404, 128, 39 },
		{ 404, 129, 39 },
		{ 404, 130, 39 },
		{ 404, 131, 39 },
		{ 404, 132, 39 },
		{ 404, 133, 39 },
		{ 404, 134, 39 },
		{ 404, 135, 39 },
		{ 404, 136, 39 },
		{ 404, 137, 39 },
		{ 404, 138, 39 },
		{ 404, 139, 39 },
		{ 404, 140, 39 },
		{ 404, 141, 39 },
		{ 404, 142, 39 },
		{ 404, 143, 39 },
		{ 404, 144, 39 },
		{ 404, 145, 39 },
		{ 404, 146, 39 },
		{ 405, 110, 39 },
		{ 405, 111, 39 },
		{ 405, 112, 39 },
		{ 405, 113, 39 },
		{ 405, 114, 39 },
		{ 405, 115, 39 },
		{ 405, 116, 39 },
		{ 405, 117, 39 },
		{ 405, 118, 39 },
		{ 405, 119, 39 },
		{ 405, 120, 39 },
		{ 405, 121, 39 },
		{ 405, 122, 39 },
		{ 405, 123, 39 },
		{ 405, 124, 39 },
		{ 405, 125, 39 },
		{ 405, 126, 39 },
		{ 405, 127, 39 },
		{ 405, 128, 39 },
		{ 405, 129, 39 },
		{ 405, 130, 39 },
		{ 405, 131, 39 },
		{ 405, 132, 39 },
		{ 405, 133, 39 },
		{ 405, 134, 39 },
		{ 405, 135, 39 },
		{ 405, 136, 39 },
		{ 405, 137, 39 },
		{ 405, 138, 39 },
		{ 405, 139, 39 },
		{ 405, 140, 39 },
		{ 405, 141, 39 },
		{ 405, 142, 39 },
		{ 405, 143, 39 },
		{ 405, 144, 39 },
		{ 405, 145, 39 },
		{ 406, 109, 39 },
		{ 406, 110, 39 },
		{ 406, 111, 39 },
		{ 406, 112, 39 },
		{ 406, 113, 39 },
		{ 406, 114, 39 },
		{ 406, 115, 39 },
		{ 406, 116, 39 },
		{ 406, 117, 39 },
		{ 406, 118, 39 },
		{ 406, 119, 39 },
		{ 406, 120, 39 },
		{ 406, 121, 39 },
		{ 406, 122, 39 },
		{ 406, 123, 39 },
		{ 406, 124, 39 },
		{ 406, 125, 39 },
		{ 406, 126, 39 },
		{ 406, 127, 39 },
		{ 406, 128, 39 },
		{ 406, 129, 39 },
		{ 406, 130, 39 },
		{ 406, 131, 39 },
		{ 406, 132, 39 },
		{ 406, 133, 39 },
		{ 406, 134, 39 },
		{ 406, 135, 39 },
		{ 406, 136, 39 },
		{ 406, 137, 39 },
		{ 406, 138, 39 },
		{ 406, 139, 39 },
		{ 406, 140, 39 },
		{ 406, 141, 39 },
		{ 406, 142, 39 },
		{ 406, 143, 39 },
		{ 406, 144, 39 },
		{ 406, 145, 39 },
		{ 407, 108, 39 },
		{ 407, 109, 39 },
		{ 407, 110, 39 },
		{ 407, 111, 39 },
		{ 407, 112, 39 },
		{ 407, 113, 39 },
		{ 407, 114, 39 },
		{ 407, 115, 39 },
		{ 407, 116, 39 },
		{ 407, 117, 39 },
		{ 407, 118, 39 },
		{ 407, 119, 39 },
		{ 407, 120, 39 },
		{ 407, 121, 39 },
		{ 407, 122, 39 },
		{ 407, 123, 39 },
		{ 407, 124, 39 },
		{ 407, 125, 39 },
		{ 407, 126, 39 },
		{ 407, 127, 39 },
		{ 407, 128, 39 },
		{ 407, 129, 39 },
		{ 407, 130, 39 },
		{ 407, 131, 39 },
		{ 407, 132, 39 },
		{ 407, 133, 39 },
		{ 407, 134, 39 },
		{ 407, 135, 39 },
		{ 407, 136, 39 },
		{ 407, 137, 39 },
		{ 407, 138, 39 },
		{ 407, 139, 39 },
		{ 407, 140, 39 },
		{ 407, 141, 39 },
		{ 407, 142, 39 },
		{ 407, 143, 39 },
		{ 407, 144, 39 },
		{ 408, 107, 39 },
		{ 408, 108, 39 },
		{ 408, 109, 39 },
		{ 408, 110, 39 },
		{ 408, 111, 39 },
		{ 408, 112, 39 },
		{ 408, 113, 39 },
		{ 408, 114, 39 },
		{ 408, 115, 39 },
		{ 408, 116, 39 },
		{ 408, 117, 39 },
		{ 408, 118, 39 },
		{ 408, 119, 39 },
		{ 408, 120, 39 },
		{ 408, 121, 39 },
		{ 408, 122, 39 },
		{ 408, 123, 39 },
		{ 408, 124, 39 },
		{ 408, 125, 39 },
		{ 408, 126, 39 },
		{ 408, 127, 39 },
		{ 408, 128, 39 },
		{ 408, 129, 39 },
		{ 408, 130, 39 },
		{ 408, 131, 39 },
		{ 408, 132, 39 },
		{ 408, 133, 39 },
		{ 408, 134, 39 },
		{ 408, 135, 39 },
		{ 408, 136, 39 },
		{ 408, 137, 39 },
		{ 408, 138, 39 },
		{ 408, 139, 39 },
		{ 408, 140, 39 },
		{ 408, 141, 39 },
		{ 408, 142, 39 },
		{ 408, 143, 39 },
		{ 409, 106, 39 },
		{ 409, 107, 39 },
		{ 409, 108, 39 },
		{ 409, 109, 39 },
		{ 409, 110, 39 },
		{ 409, 111, 39 },
		{ 409, 112, 39 },
		{ 409, 113, 39 },
		{ 409, 114, 39 },
		{ 409, 115, 39 },
		{ 409, 116, 39 },
		{ 409, 117, 39 },
		{ 409, 118, 39 },
		{ 409, 119, 39 },
		{ 409, 120, 39 },
		{ 409, 121, 39 },
		{ 409, 122, 39 },
		{ 409, 123, 39 },
		{ 409, 124, 39 },
		{ 409, 125, 39 },
		{ 409, 126, 39 },
		{ 409, 127, 39 },
		{ 409, 128, 39 },
		{ 409, 129, 39 },
		{ 409, 130, 39 },
		{ 409, 131, 39 },
		{ 409, 132, 39 },
		{ 409, 133, 39 },
		{ 409, 134, 39 },
		{ 409, 135, 39 },
		{ 409, 136, 39 },
		{ 409, 137, 39 },
		{ 409, 138, 39 },
		{ 409, 139, 39 },
		{ 409, 140, 39 },
		{ 409, 141, 39 },
		{ 409, 142, 39 },
		{ 410, 105, 39 },
		{ 410, 106, 39 },
		{ 410, 107, 39 },
		{ 410, 108, 39 },
		{ 410, 109, 39 },
		{ 410, 110, 39 },
		{ 410, 111, 39 },
		{ 410, 112, 39 },
		{ 410, 113, 39 },
		{ 410, 114, 39 },
		{ 410, 115, 39 },
		{ 410, 116, 39 },
		{ 410, 117, 39 },
		{ 410, 118, 39 },
		{ 410, 119, 39 },
		{ 410, 120, 39 },
		{ 410, 121, 39 },
		{ 410, 122, 39 },
		{ 410, 123, 39 },
		{ 410, 124, 39 },
		{ 410, 125, 39 },
		{ 410, 126, 39 },
		{ 410, 127, 39 },
		{ 410, 128, 39 },
		{ 410, 129, 39 },
		{ 410, 130, 39 },
		{ 410, 131, 39 },
		{ 410, 132, 39 },
		{ 410, 133, 39 },
		{ 410, 134, 39 },
		{ 410, 135, 39 },
		{ 410, 136, 39 },
		{ 410, 137, 39 },
		{ 410, 138, 39 },
		{ 410, 139, 39 },
		{ 410, 140, 39 },
		{ 410, 141, 39 },
		{ 410, 142, 39 },
		{ 411, 104, 39 },
		{ 411, 105, 39 },
		{ 411, 106, 39 },
		{ 411, 107, 39 },
		{ 411, 108, 39 },
		{ 411, 109, 39 },
		{ 411, 110, 39 },
		{ 411, 111, 39 },
		{ 411, 112, 39 },
		{ 411, 113, 39 },
		{ 411, 114, 39 },
		{ 411, 115, 39 },
		{ 411, 116, 39 },
		{ 411, 117, 39 },
		{ 411, 118, 39 },
		{ 411, 119, 39 },
		{ 411, 120, 39 },
		{ 411, 121, 39 },
		{ 411, 122, 39 },
		{ 411, 123, 39 },
		{ 411, 124, 39 },
		{ 411, 125, 39 },
		{ 411, 126, 39 },
		{ 411, 127, 39 },
		{ 411, 128, 39 },
		{ 411, 129, 39 },
		{ 411, 130, 39 },
		{ 411, 131, 39 },
		{ 411, 132, 39 },
		{ 411, 133, 39 },
		{ 411, 134, 39 },
		{ 411, 135, 39 },
		{ 411, 136, 39 },
		{ 411, 137, 39 },
		{ 411, 138, 39 },
		{ 411, 139, 39 },
		{ 411, 140, 39 },
		{ 411, 141, 39 },
		{ 412, 103, 39 },
		{ 412, 104, 39 },
		{ 412, 105, 39 },
		{ 412, 106, 39 },
		{ 412, 107, 39 },
		{ 412, 108, 39 },
		{ 412, 109, 39 },
		{ 412, 110, 39 },
		{ 412, 111, 39 },
		{ 412, 112, 39 },
		{ 412, 113, 39 },
		{ 412, 114, 39 },
		{ 412, 115, 39 },
		{ 412, 116, 39 },
		{ 412, 117, 39 },
		{ 412, 118, 39 },
		{ 412, 119, 39 },
		{ 412, 120, 39 },
		{ 412, 121, 39 },
		{ 412, 122, 39 },
		{ 412, 123, 39 },
		{ 412, 124, 39 },
		{ 412, 125, 39 },
		{ 412, 126, 39 },
		{ 412, 127, 39 },
		{ 412, 128, 39 },
		{ 412, 129, 39 },
		{ 412, 130, 39 },
		{ 412, 131, 39 },
		{ 412, 132, 39 },
		{ 412, 133, 39 },
		{ 412, 134, 39 },
		{ 412, 135, 39 },
		{ 412, 136, 39 },
		{ 412, 137, 39 },
		{ 412, 138, 39 },
		{ 412, 139, 39 },
		{ 412, 140, 39 },
		{ 413, 102, 39 },
		{ 413, 103, 39 },
		{ 413, 104, 39 },
		{ 413, 105, 39 },
		{ 413, 106, 39 },
		{ 413, 107, 39 },
		{ 413, 108, 39 },
		{ 413, 109, 39 },
		{ 413, 110, 39 },
		{ 413, 111, 39 },
		{ 413, 112, 39 },
		{ 413, 113, 39 },
		{ 413, 114, 39 },
		{ 413, 115, 39 },
		{ 413, 116, 39 },
		{ 413, 117, 39 },
		{ 413, 118, 39 },
		{ 413, 119, 39 },
		{ 413, 120, 39 },
		{ 413, 121, 39 },
		{ 413, 122, 39 },
		{ 413, 123, 39 },
		{ 413, 124, 39 },
		{ 413, 125, 39 },
		{ 413, 126, 39 },
		{ 413, 127, 39 },
		{ 413, 128, 39 },
		{ 413, 129, 39 },
		{ 413, 130, 39 },
		{ 413, 131, 39 },
		{ 413, 132, 39 },
		{ 413, 133, 39 },
		{ 413, 134, 39 },
		{ 413, 135, 39 },
		{ 413, 136, 39 },
		{ 413, 137, 39 },
		{ 413, 138, 39 },
		{ 413, 139, 39 },
		{ 414, 101, 39 },
		{ 414, 102, 39 },
		{ 414, 103, 39 },
		{ 414, 104, 39 },
		{ 414, 105, 39 },
		{ 414, 106, 39 },
		{ 414, 107, 39 },
		{ 414, 108, 39 },
		{ 414, 109, 39 },
		{ 414, 110, 39 },
		{ 414, 111, 39 },
		{ 414, 112, 39 },
		{ 414, 113, 39 },
		{ 414, 114, 39 },
		{ 414, 115, 39 },
		{ 414, 116, 39 },
		{ 414, 117, 39 },
		{ 414, 118, 39 },
		{ 414, 119, 39 },
		{ 414, 120, 39 },
		{ 414, 121, 39 },
		{ 414, 122, 39 },
		{ 414, 123, 39 },
		{ 414, 124, 39 },
		{ 414, 125, 39 },
		{ 414, 126, 39 },
		{ 414, 127, 39 },
		{ 414, 128, 39 },
		{ 414, 129, 39 },
		{ 414, 130, 39 },
		{ 414, 131, 39 },
		{ 414, 132, 39 },
		{ 414, 133, 39 },
		{ 414, 134, 39 },
		{ 414, 135, 39 },
		{ 414, 136, 39 },
		{ 414, 137, 39 },
		{ 414, 138, 39 },
		{ 414, 139, 39 },
		{ 415, 100, 39 },
		{ 415, 101, 39 },
		{ 415, 102, 39 },
		{ 415, 103, 39 },
		{ 415, 104, 39 },
		{ 415, 105, 39 },
		{ 415, 106, 39 },
		{ 415, 107, 39 },
		{ 415, 108, 39 },
		{ 415, 109, 39 },
		{ 415, 110, 39 },
		{ 415, 111, 39 },
		{ 415, 112, 39 },
		{ 415, 113, 39 },
		{ 415, 114, 39 },
		{ 415, 115, 39 },
		{ 415, 116, 39 },
		{ 415, 117, 39 },
		{ 415, 118, 39 },
		{ 415, 119, 39 },
		{ 415, 120, 39 },
		{ 415, 121, 39 },
		{ 415, 122, 39 },
		{ 415, 123, 39 },
		{ 415, 124, 39 },
		{ 415, 125, 39 },
		{ 415, 126, 39 },
		{ 415, 127, 39 },
		{ 415, 128, 39 },
		{ 415, 129, 39 },
		{ 415, 130, 39 },
		{ 415, 131, 39 },
		{ 415, 132, 39 },
		{ 415, 133, 39 },
		{ 415, 134, 39 },
		{ 415, 135, 39 },
		{ 415, 136, 39 },
		{ 415, 137, 39 },
		{ 415, 138, 39 },
		{ 416, 99, 39 },
		{ 416, 100, 39 },
		{ 416, 101, 39 },
		{ 416, 102, 39 },
		{ 416, 103, 39 },
		{ 416, 104, 39 },
		{ 416, 105, 39 },
		{ 416, 106, 39 },
		{ 416, 107, 39 },
		{ 416, 108, 39 },
		{ 416, 109, 39 },
		{ 416, 110, 39 },
		{ 416, 111, 39 },
		{ 416, 112, 39 },
		{ 416, 113, 39 },
		{ 416, 114, 39 },
		{ 416, 115, 39 },
		{ 416, 116, 39 },
		{ 416, 117, 39 },
		{ 416, 118, 39 },
		{ 416, 119, 39 },
		{ 416, 120, 39 },
		{ 416, 121, 39 },
		{ 416, 122, 39 },
		{ 416, 123, 39 },
		{ 416, 124, 39 },
		{ 416, 125, 39 },
		{ 416, 126, 39 },
		{ 416, 127, 39 },
		{ 416, 128, 39 },
		{ 416, 129, 39 },
		{ 416, 130, 39 },
		{ 416, 131, 39 },
		{ 416, 132, 39 },
		{ 416, 133, 39 },
		{ 416, 134, 39 },
		{ 416, 135, 39 },
		{ 416, 136, 39 },
		{ 416, 137, 39 },
		{ 417, 98, 39 },
		{ 417, 99, 39 },
		{ 417, 100, 39 },
		{ 417, 101, 39 },
		{ 417, 102, 39 },
		{ 417, 103, 39 },
		{ 417, 104, 39 },
		{ 417, 105, 39 },
		{ 417, 106, 39 },
		{ 417, 107, 39 },
		{ 417, 108, 39 },
		{ 417, 109, 39 },
		{ 417, 110, 39 },
		{ 417, 111, 39 },
		{ 417, 112, 39 },
		{ 417, 113, 39 },
		{ 417, 114, 39 },
		{ 417, 115, 39 },
		{ 417, 116, 39 },
		{ 417, 117, 39 },
		{ 417, 118, 39 },
		{ 417, 119, 39 },
		{ 417, 120, 39 },
		{ 417, 121, 39 },
		{ 417, 122, 39 },
		{ 417, 123, 39 },
		{ 417, 124, 39 },
		{ 417, 125, 39 },
		{ 417, 126, 39 },
		{ 417, 127, 39 },
		{ 417, 128, 39 },
		{ 417, 129, 39 },
		{ 417, 130, 39 },
		{ 417, 131, 39 },
		{ 417, 132, 39 },
		{ 417, 133, 39 },
		{ 417, 134, 39 },
		{ 417, 135, 39 },
		{ 417, 136, 39 },
		{ 418, 97, 39 },
		{ 418, 98, 39 },
		{ 418, 99, 39 },
		{ 418, 100, 39 },
		{ 418, 101, 39 },
		{ 418, 102, 39 },
		{ 418, 103, 39 },
		{ 418, 104, 39 },
		{ 418, 105, 39 },
		{ 418, 106, 39 },
		{ 418, 107, 39 },
		{ 418, 108, 39 },
		{ 418, 109, 39 },
		{ 418, 110, 39 },
		{ 418, 111, 39 },
		{ 418, 112, 39 },
		{ 418, 113, 39 },
		{ 418, 114, 39 },
		{ 418, 115, 39 },
		{ 418, 116, 39 },
		{ 418, 117, 39 },
		{ 418, 118, 39 },
		{ 418, 119, 39 },
		{ 418, 120, 39 },
		{ 418, 121, 39 },
		{ 418, 122, 39 },
		{ 418, 123, 39 },
		{ 418, 124, 39 },
		{ 418, 125, 39 },
		{ 418, 126, 39 },
		{ 418, 127, 39 },
		{ 418, 128, 39 },
		{ 418, 129, 39 },
		{ 418, 130, 39 },
		{ 418, 131, 39 },
		{ 418, 132, 39 },
		{ 418, 133, 39 },
		{ 418, 134, 39 },
		{ 418, 135, 39 },
		{ 418, 136, 39 },
		{ 419, 96, 39 },
		{ 419, 97, 39 },
		{ 419, 98, 39 },
		{ 419, 99, 39 },
		{ 419, 100, 39 },
		{ 419, 101, 39 },
		{ 419, 102, 39 },
		{ 419, 103, 39 },
		{ 419, 104, 39 },
		{ 419, 105, 39 },
		{ 419, 106, 39 },
		{ 419, 107, 39 },
		{ 419, 108, 39 },
		{ 419, 109, 39 },
		{ 419, 110, 39 },
		{ 419, 111, 39 },
		{ 419, 112, 39 },
		{ 419, 113, 39 },
		{ 419, 114, 39 },
		{ 419, 115, 39 },
		{ 419, 116, 39 },
		{ 419, 117, 39 },
		{ 419, 118, 39 },
		{ 419, 119, 39 },
		{ 419, 120, 39 },
		{ 419, 121, 39 },
		{ 419, 122, 39 },
		{ 419, 123, 39 },
		{ 419, 124, 39 },
		{ 419, 125, 39 },
		{ 419, 126, 39 },
		{ 419, 127, 39 },
		{ 419, 128, 39 },
		{ 419, 129, 39 },
		{ 419, 130, 39 },
		{ 419, 131, 39 },
		{ 419, 132, 39 },
		{ 419, 133, 39 },
		{ 419, 134, 39 },
		{ 419, 135, 39 },
		{ 420, 95, 39 },
		{ 420, 96, 39 },
		{ 420, 97, 39 },
		{ 420, 98, 39 },
		{ 420, 99, 39 },
		{ 420, 100, 39 },
		{ 420, 101, 39 },
		{ 420, 102, 39 },
		{ 420, 103, 39 },
		{ 420, 104, 39 },
		{ 420, 105, 39 },
		{ 420, 106, 39 },
		{ 420, 107, 39 },
		{ 420, 108, 39 },
		{ 420, 109, 39 },
		{ 420, 110, 39 },
		{ 420, 111, 39 },
		{ 420, 112, 39 },
		{ 420, 113, 39 },
		{ 420, 114, 39 },
		{ 420, 115, 39 },
		{ 420, 116, 39 },
		{ 420, 117, 39 },
		{ 420, 118, 39 },
		{ 420, 119, 39 },
		{ 420, 120, 39 },
		{ 420, 121, 39 },
		{ 420, 122, 39 },
		{ 420, 123, 39 },
		{ 420, 124, 39 },
		{ 420, 125, 39 },
		{ 420, 126, 39 },
		{ 420, 127, 39 },
		{ 420, 128, 39 },
		{ 420, 129, 39 },
		{ 420, 130, 39 },
		{ 420, 131, 39 },
		{ 420, 132, 39 },
		{ 420, 133, 39 },
		{ 420, 134, 39 },
		{ 421, 94, 39 },
		{ 421, 95, 39 },
		{ 421, 96, 39 },
		{ 421, 97, 39 },
		{ 421, 98, 39 },
		{ 421, 99, 39 },
		{ 421, 100, 39 },
		{ 421, 101, 39 },
		{ 421, 102, 39 },
		{ 421, 103, 39 },
		{ 421, 104, 39 },
		{ 421, 105, 39 },
		{ 421, 106, 39 },
		{ 421, 107, 39 },
		{ 421, 108, 39 },
		{ 421, 109, 39 },
		{ 421, 110, 39 },
		{ 421, 111, 39 },
		{ 421, 112, 39 },
		{ 421, 113, 39 },
		{ 421, 114, 39 },
		{ 421, 115, 39 },
		{ 421, 116, 39 },
		{ 421, 117, 39 },
		{ 421, 118, 39 },
		{ 421, 119, 39 },
		{ 421, 120, 39 },
		{ 421, 121, 39 },
		{ 421, 122, 39 },
		{ 421, 123, 39 },
		{ 421, 124, 39 },
		{ 421, 125, 39 },
		{ 421, 126, 39 },
		{ 421, 127, 39 },
		{ 421, 128, 39 },
		{ 421, 129, 39 },
		{ 421, 130, 39 },
		{ 421, 131, 39 },
		{ 421, 132, 39 },
		{ 421, 133, 39 },
		{ 422, 93, 39 },
		{ 422, 94, 39 },
		{ 422, 95, 39 },
		{ 422, 96, 39 },
		{ 422, 97, 39 },
		{ 422, 98, 39 },
		{ 422, 99, 39 },
		{ 422, 100, 39 },
		{ 422, 101, 39 },
		{ 422, 102, 39 },
		{ 422, 103, 39 },
		{ 422, 104, 39 },
		{ 422, 105, 39 },
		{ 422, 106, 39 },
		{ 422, 107, 39 },
		{ 422, 108, 39 },
		{ 422, 109, 39 },
		{ 422, 110, 39 },
		{ 422, 111, 39 },
		{ 422, 112, 39 },
		{ 422, 113, 39 },
		{ 422, 114, 39 },
		{ 422, 115, 39 },
		{ 422, 116, 39 },
		{ 422, 117, 39 },
		{ 422, 118, 39 },
		{ 422, 119, 39 },
		{ 422, 120, 39 },
		{ 422, 121, 39 },
		{ 422, 122, 39 },
		{ 422, 123, 39 },
		{ 422, 124, 39 },
		{ 422, 125, 39 },
		{ 422, 126, 39 },
		{ 422, 127, 39 },
		{ 422, 128, 39 },
		{ 422, 129, 39 },
		{ 422, 130, 39 },
		{ 422, 131, 39 },
		{ 422, 132, 39 },
		{ 422, 133, 39 },
		{ 423, 92, 39 },
		{ 423, 93, 39 },
		{ 423, 94, 39 },
		{ 423, 95, 39 },
		{ 423, 96, 39 },
		{ 423, 97, 39 },
		{ 423, 98, 39 },
		{ 423, 99, 39 },
		{ 423, 100, 39 },
		{ 423, 101, 39 },
		{ 423, 102, 39 },
		{ 423, 103, 39 },
		{ 423, 104, 39 },
		{ 423, 105, 39 },
		{ 423, 106, 39 },
		{ 423, 107, 39 },
		{ 423, 108, 39 },
		{ 423, 109, 39 },
		{ 423, 110, 39 },
		{ 423, 111, 39 },
		{ 423, 112, 39 },
		{ 423, 113, 39 },
		{ 423, 114, 39 },
		{ 423, 115, 39 },
		{ 423, 116, 39 },
		{ 423, 117, 39 },
		{ 423, 118, 39 },
		{ 423, 119, 39 },
		{ 423, 120, 39 },
		{ 423, 121, 39 },
		{ 423, 122, 39 },
		{ 423, 123, 39 },
		{ 423, 124, 39 },
		{ 423, 125, 39 },
		{ 423, 126, 39 },
		{ 423, 127, 39 },
		{ 423, 128, 39 },
		{ 423, 129, 39 },
		{ 423, 130, 39 },
		{ 423, 131, 39 },
		{ 423, 132, 39 },
		{ 424, 91, 39 },
		{ 424, 92, 39 },
		{ 424, 93, 39 },
		{ 424, 94, 39 },
		{ 424, 95, 39 },
		{ 424, 96, 39 },
		{ 424, 97, 39 },
		{ 424, 98, 39 },
		{ 424, 99, 39 },
		{ 424, 100, 39 },
		{ 424, 101, 39 },
		{ 424, 102, 39 },
		{ 424, 103, 39 },
		{ 424, 104, 39 },
		{ 424, 105, 39 },
		{ 424, 106, 39 },
		{ 424, 107, 39 },
		{ 424, 108, 39 },
		{ 424, 109, 39 },
		{ 424, 110, 39 },
		{ 424, 111, 39 },
		{ 424, 112, 39 },
		{ 424, 113, 39 },
		{ 424, 114, 39 },
		{ 424, 115, 39 },
		{ 424, 116, 39 },
		{ 424, 117, 39 },
		{ 424, 118, 39 },
		{ 424, 119, 39 },
		{ 424, 120, 39 },
		{ 424, 121, 39 },
		{ 424, 122, 39 },
		{ 424, 123, 39 },
		{ 424, 124, 39 },
		{ 424, 125, 39 },
		{ 424, 126, 39 },
		{ 424, 127, 39 },
		{ 424, 128, 39 },
		{ 424, 129, 39 },
		{ 424, 130, 39 },
		{ 424, 131, 39 },
		{ 425, 90, 39 },
		{ 425, 91, 39 },
		{ 425, 92, 39 },
		{ 425, 93, 39 },
		{ 425, 94, 39 },
		{ 425, 95, 39 },
		{ 425, 96, 39 },
		{ 425, 97, 39 },
		{ 425, 98, 39 },
		{ 425, 99, 39 },
		{ 425, 100, 39 },
		{ 425, 101, 39 },
		{ 425, 102, 39 },
		{ 425, 103, 39 },
		{ 425, 104, 39 },
		{ 425, 105, 39 },
		{ 425, 106, 39 },
		{ 425, 107, 39 },
		{ 425, 108, 39 },
		{ 425, 109, 39 },
		{ 425, 110, 39 },
		{ 425, 111, 39 },
		{ 425, 112, 39 },
		{ 425, 113, 39 },
		{ 425, 114, 39 },
		{ 425, 115, 39 },
		{ 425, 116, 39 },
		{ 425, 117, 39 },
		{ 425, 118, 39 },
		{ 425, 119, 39 },
		{ 425, 120, 39 },
		{ 425, 121, 39 },
		{ 425, 122, 39 },
		{ 425, 123, 39 },
		{ 425, 124, 39 },
		{ 425, 125, 39 },
		{ 425, 126, 39 },
		{ 425, 127, 39 },
		{ 425, 128, 39 },
		{ 425, 129, 39 },
		{ 425, 130, 39 },
		{ 426, 89, 39 },
		{ 426, 90, 39 },
		{ 426, 91, 39 },
		{ 426, 92, 39 },
		{ 426, 93, 39 },
		{ 426, 94, 39 },
		{ 426, 95, 39 },
		{ 426, 96, 39 },
		{ 426, 97, 39 },
		{ 426, 98, 39 },
		{ 426, 99, 39 },
		{ 426, 100, 39 },
		{ 426, 101, 39 },
		{ 426, 102, 39 },
		{ 426, 103, 39 },
		{ 426, 104, 39 },
		{ 426, 105, 39 },
		{ 426, 106, 39 },
		{ 426, 107, 39 },
		{ 426, 108, 39 },
		{ 426, 109, 39 },
		{ 426, 110, 39 },
		{ 426, 111, 39 },
		{ 426, 112, 39 },
		{ 426, 113, 39 },
		{ 426, 114, 39 },
		{ 426, 115, 39 },
		{ 426, 116, 39 },
		{ 426, 117, 39 },
		{ 426, 118, 39 },
		{ 426, 119, 39 },
		{ 426, 120, 39 },
		{ 426, 121, 39 },
		{ 426, 122, 39 },
		{ 426, 123, 39 },
		{ 426, 124, 39 },
		{ 426, 125, 39 },
		{ 426, 126, 39 },
		{ 426, 127, 39 },
		{ 426, 128, 39 },
		{ 426, 129, 39 },
		{ 426, 130, 39 },
		{ 427, 88, 39 },
		{ 427, 89, 39 },
		{ 427, 90, 39 },
		{ 427, 91, 39 },
		{ 427, 92, 39 },
		{ 427, 93, 39 },
		{ 427, 94, 39 },
		{ 427, 95, 39 },
		{ 427, 96, 39 },
		{ 427, 97, 39 },
		{ 427, 98, 39 },
		{ 427, 99, 39 },
		{ 427, 100, 39 },
		{ 427, 101, 39 },
		{ 427, 102, 39 },
		{ 427, 103, 39 },
		{ 427, 104, 39 },
		{ 427, 105, 39 },
		{ 427, 106, 39 },
		{ 427, 107, 39 },
		{ 427, 108, 39 },
		{ 427, 109, 39 },
		{ 427, 110, 39 },
		{ 427, 111, 39 },
		{ 427, 112, 39 },
		{ 427, 113, 39 },
		{ 427, 114, 39 },
		{ 427, 115, 39 },
		{ 427, 116, 39 },
		{ 427, 117, 39 },
		{ 427, 118, 39 },
		{ 427, 119, 39 },
		{ 427, 120, 39 },
		{ 427, 121, 39 },
		{ 427, 122, 39 },
		{ 427, 123, 39 },
		{ 427, 124, 39 },
		{ 427, 125, 39 },
		{ 427, 126, 39 },
		{ 427, 127, 39 },
		{ 427, 128, 39 },
		{ 427, 129, 39 },
		{ 428, 87, 39 },
		{ 428, 88, 39 },
		{ 428, 89, 39 },
		{ 428, 90, 39 },
		{ 428, 91, 39 },
		{ 428, 92, 39 },
		{ 428, 93, 39 },
		{ 428, 94, 39 },
		{ 428, 95, 39 },
		{ 428, 96, 39 },
		{ 428, 97, 39 },
		{ 428, 98, 39 },
		{ 428, 99, 39 },
		{ 428, 100, 39 },
		{ 428, 101, 39 },
		{ 428, 102, 39 },
		{ 428, 103, 39 },
		{ 428, 104, 39 },
		{ 428, 105, 39 },
		{ 428, 106, 39 },
		{ 428, 107, 39 },
		{ 428, 108, 39 },
		{ 428, 109, 39 },
		{ 428, 110, 39 },
		{ 428, 111, 39 },
		{ 428, 112, 39 },
		{ 428, 113, 39 },
		{ 428, 114, 39 },
		{ 428, 115, 39 },
		{ 428, 116, 39 },
		{ 428, 117, 39 },
		{ 428, 118, 39 },
		{ 428, 119, 39 },
		{ 428, 120, 39 },
		{ 428, 121, 39 },
		{ 428, 122, 39 },
		{ 428, 123, 39 },
		{ 428, 124, 39 },
		{ 428, 125, 39 },
		{ 428, 126, 39 },
		{ 428, 127, 39 },
		{ 428, 128, 39 },
		{ 429, 86, 39 },
		{ 429, 87, 39 },
		{ 429, 88, 39 },
		{ 429, 89, 39 },
		{ 429, 90, 39 },
		{ 429, 91, 39 },
		{ 429, 92, 39 },
		{ 429, 93, 39 },
		{ 429, 94, 39 },
		{ 429, 95, 39 },
		{ 429, 96, 39 },
		{ 429, 97, 39 },
		{ 429, 98, 39 },
		{ 429, 99, 39 },
		{ 429, 100, 39 },
		{ 429, 101, 39 },
		{ 429, 102, 39 },
		{ 429, 103, 39 },
		{ 429, 104, 39 },
		{ 429, 105, 39 },
		{ 429, 106, 39 },
		{ 429, 107, 39 },
		{ 429, 108, 39 },
		{ 429, 109, 39 },
		{ 429, 110, 39 },
		{ 429, 111, 39 },
		{ 429, 112, 39 },
		{ 429, 113, 39 },
		{ 429, 114, 39 },
		{ 429, 115, 39 },
		{ 429, 116, 39 },
		{ 429, 117, 39 },
		{ 429, 118, 39 },
		{ 429, 119, 39 },
		{ 429, 120, 39 },
		{ 429, 121, 39 },
		{ 429, 122, 39 },
		{ 429, 123, 39 },
		{ 429, 124, 39 },
		{ 429, 125, 39 },
		{ 429, 126, 39 },
		{ 429, 127, 39 },
		{ 430, 85, 39 },
		{ 430, 86, 39 },
		{ 430, 87, 39 },
		{ 430, 88, 39 },
		{ 430, 89, 39 },
		{ 430, 90, 39 },
		{ 430, 91, 39 },
		{ 430, 92, 39 },
		{ 430, 93, 39 },
		{ 430, 94, 39 },
		{ 430, 95, 39 },
		{ 430, 96, 39 },
		{ 430, 97, 39 },
		{ 430, 98, 39 },
		{ 430, 99, 39 },
		{ 430, 100, 39 },
		{ 430, 101, 39 },
		{ 430, 102, 39 },
		{ 430, 103, 39 },
		{ 430, 104, 39 },
		{ 430, 105, 39 },
		{ 430, 106, 39 },
		{ 430, 107, 39 },
		{ 430, 108, 39 },
		{ 430, 109, 39 },
		{ 430, 110, 39 },
		{ 430, 111, 39 },
		{ 430, 112, 39 },
		{ 430, 113, 39 },
		{ 430, 114, 39 },
		{ 430, 115, 39 },
		{ 430, 116, 39 },
		{ 430, 117, 39 },
		{ 430, 118, 39 },
		{ 430, 119, 39 },
		{ 430, 120, 39 },
		{ 430, 121, 39 },
		{ 430, 122, 39 },
		{ 430, 123, 39 },
		{ 430, 124, 39 },
		{ 430, 125, 39 },
		{ 430, 126, 39 },
		{ 430, 127, 39 },
		{ 431, 84, 39 },
		{ 431, 85, 39 },
		{ 431, 86, 39 },
		{ 431, 87, 39 },
		{ 431, 88, 39 },
		{ 431, 89, 39 },
		{ 431, 90, 39 },
		{ 431, 91, 39 },
		{ 431, 92, 39 },
		{ 431, 93, 39 },
		{ 431, 94, 39 },
		{ 431, 95, 39 },
		{ 431, 96, 39 },
		{ 431, 97, 39 },
		{ 431, 98, 39 },
		{ 431, 99, 39 },
		{ 431, 100, 39 },
		{ 431, 101, 39 },
		{ 431, 102, 39 },
		{ 431, 103, 39 },
		{ 431, 104, 39 },
		{ 431, 105, 39 },
		{ 431, 106, 39 },
		{ 431, 107, 39 },
		{ 431, 108, 39 },
		{ 431, 109, 39 },
		{ 431, 110, 39 },
		{ 431, 111, 39 },
		{ 431, 112, 39 },
		{ 431, 113, 39 },
		{ 431, 114, 39 },
		{ 431, 115, 39 },
		{ 431, 116, 39 },
		{ 431, 117, 39 },
		{ 431, 118, 39 },
		{ 431, 119, 39 },
		{ 431, 120, 39 },
		{ 431, 121, 39 },
		{ 431, 122, 39 },
		{ 431, 123, 39 },
		{ 431, 124, 39 },
		{ 431, 125, 39 },
		{ 431, 126, 39 },
		{ 432, 83, 39 },
		{ 432, 84, 39 },
		{ 432, 85, 39 },
		{ 432, 86, 39 },
		{ 432, 87, 39 },
		{ 432, 88, 39 },
		{ 432, 89, 39 },
		{ 432, 90, 39 },
		{ 432, 91, 39 },
		{ 432, 92, 39 },
		{ 432, 93, 39 },
		{ 432, 94, 39 },
		{ 432, 95, 39 },
		{ 432, 96, 39 },
		{ 432, 97, 39 },
		{ 432, 98, 39 },
		{ 432, 99, 39 },
		{ 432, 100, 39 },
		{ 432, 101, 39 },
		{ 432, 102, 39 },
		{ 432, 103, 39 },
		{ 432, 104, 39 },
		{ 432, 105, 39 },
		{ 432, 106, 39 },
		{ 432, 107, 39 },
		{ 432, 108, 39 },
		{ 432, 109, 39 },
		{ 432, 110, 39 },
		{ 432, 111, 39 },
		{ 432, 112, 39 },
		{ 432, 113, 39 },
		{ 432, 114, 39 },
		{ 432, 115, 39 },
		{ 432, 116, 39 },
		{ 432, 117, 39 },
		{ 432, 118, 39 },
		{ 432, 119, 39 },
		{ 432, 120, 39 },
		{ 432, 121, 39 },
		{ 432, 122, 39 },
		{ 432, 123, 39 },
		{ 432, 124, 39 },
		{ 432, 125, 39 },
		{ 433, 82, 39 },
		{ 433, 83, 39 },
		{ 433, 84, 39 },
		{ 433, 85, 39 },
		{ 433, 86, 39 },
		{ 433, 87, 39 },
		{ 433, 88, 39 },
		{ 433, 89, 39 },
		{ 433, 90, 39 },
		{ 433, 91, 39 },
		{ 433, 92, 39 },
		{ 433, 93, 39 },
		{ 433, 94, 39 },
		{ 433, 95, 39 },
		{ 433, 96, 39 },
		{ 433, 97, 39 },
		{ 433, 98, 39 },
		{ 433, 99, 39 },
		{ 433, 100, 39 },
		{ 433, 101, 39 },
		{ 433, 102, 39 },
		{ 433, 103, 39 },
		{ 433, 104, 39 },
		{ 433, 105, 39 },
		{ 433, 106, 39 },
		{ 433, 107, 39 },
		{ 433, 108, 39 },
		{ 433, 109, 39 },
		{ 433, 110, 39 },
		{ 433, 111, 39 },
		{ 433, 112, 39 },
		{ 433, 113, 39 },
		{ 433, 114, 39 },
		{ 433, 115, 39 },
		{ 433, 116, 39 },
		{ 433, 117, 39 },
		{ 433, 118, 39 },
		{ 433, 119, 39 },
		{ 433, 120, 39 },
		{ 433, 121, 39 },
		{ 433, 122, 39 },
		{ 433, 123, 39 },
		{ 433, 124, 39 },
		{ 434, 81, 39 },
		{ 434, 82, 39 },
		{ 434, 83, 39 },
		{ 434, 84, 39 },
		{ 434, 85, 39 },
		{ 434, 86, 39 },
		{ 434, 87, 39 },
		{ 434, 88, 39 },
		{ 434, 89, 39 },
		{ 434, 90, 39 },
		{ 434, 91, 39 },
		{ 434, 92, 39 },
		{ 434, 93, 39 },
		{ 434, 94, 39 },
		{ 434, 95, 39 },
		{ 434, 96, 39 },
		{ 434, 97, 39 },
		{ 434, 98, 39 },
		{ 434, 99, 39 },
		{ 434, 100, 39 },
		{ 434, 101, 39 },
		{ 434, 102, 39 },
		{ 434, 103, 39 },
		{ 434, 104, 39 },
		{ 434, 105, 39 },
		{ 434, 106, 39 },
		{ 434, 107, 39 },
		{ 434, 108, 39 },
		{ 434, 109, 39 },
		{ 434, 110, 39 },
		{ 434, 111, 39 },
		{ 434, 112, 39 },
		{ 434, 113, 39 },
		{ 434, 114, 39 },
		{ 434, 115, 39 },
		{ 434, 116, 39 },
		{ 434, 117, 39 },
		{ 434, 118, 39 },
		{ 434, 119, 39 },
		{ 434, 120, 39 },
		{ 434, 121, 39 },
		{ 434, 122, 39 },
		{ 434, 123, 39 },
		{ 434, 124, 39 },
		{ 435, 80, 39 },
		{ 435, 81, 39 },
		{ 435, 82, 39 },
		{ 435, 83, 39 },
		{ 435, 84, 39 },
		{ 435, 85, 39 },
		{ 435, 86, 39 },
		{ 435, 87, 39 },
		{ 435, 88, 39 },
		{ 435, 89, 39 },
		{ 435, 90, 39 },
		{ 435, 91, 39 },
		{ 435, 92, 39 },
		{ 435, 93, 39 },
		{ 435, 94, 39 },
		{ 435, 95, 39 },
		{ 435, 96, 39 },
		{ 435, 97, 39 },
		{ 435, 98, 39 },
		{ 435, 99, 39 },
		{ 435, 100, 39 },
		{ 435, 101, 39 },
		{ 435, 102, 39 },
		{ 435, 103, 39 },
		{ 435, 104, 39 },
		{ 435, 105, 39 },
		{ 435, 106, 39 },
		{ 435, 107, 39 },
		{ 435, 108, 39 },
		{ 435, 109, 39 },
		{ 435, 110, 39 },
		{ 435, 111, 39 },
		{ 435, 112, 39 },
		{ 435, 113, 39 },
		{ 435, 114, 39 },
		{ 435, 115, 39 },
		{ 435, 116, 39 },
		{ 435, 117, 39 },
		{ 435, 118, 39 },
		{ 435, 119, 39 },
		{ 435, 120, 39 },
		{ 435, 121, 39 },
		{ 435, 122, 39 },
		{ 435, 123, 39 },
		{ 436, 79, 39 },
		{ 436, 80, 39 },
		{ 436, 81, 39 },
		{ 436, 82, 39 },
		{ 436, 83, 39 },
		{ 436, 84, 39 },
		{ 436, 85, 39 },
		{ 436, 86, 39 },
		{ 436, 87, 39 },
		{ 436, 88, 39 },
		{ 436, 89, 39 },
		{ 436, 90, 39 },
		{ 436, 91, 39 },
		{ 436, 92, 39 },
		{ 436, 93, 39 },
		{ 436, 94, 39 },
		{ 436, 95, 39 },
		{ 436, 96, 39 },
		{ 436, 97, 39 },
		{ 436, 98, 39 },
		{ 436, 99, 39 },
		{ 436, 100, 39 },
		{ 436, 101, 39 },
		{ 436, 102, 39 },
		{ 436, 103, 39 },
		{ 436, 104, 39 },
		{ 436, 105, 39 },
		{ 436, 106, 39 },
		{ 436, 107, 39 },
		{ 436, 108, 39 },
		{ 436, 109, 39 },
		{ 436, 110, 39 },
		{ 436, 111, 39 },
		{ 436, 112, 39 },
		{ 436, 113, 39 },
		{ 436, 114, 39 },
		{ 436, 115, 39 },
		{ 436, 116, 39 },
		{ 436, 117, 39 },
		{ 436, 118, 39 },
		{ 436, 119, 39 },
		{ 436, 120, 39 },
		{ 436, 121, 39 },
		{ 436, 122, 39 },
		{ 437, 78, 39 },
		{ 437, 79, 39 },
		{ 437, 80, 39 },
		{ 437, 81, 39 },
		{ 437, 82, 39 },
		{ 437, 83, 39 },
		{ 437, 84, 39 },
		{ 437, 85, 39 },
		{ 437, 86, 39 },
		{ 437, 87, 39 },
		{ 437, 88, 39 },
		{ 437, 89, 39 },
		{ 437, 90, 39 },
		{ 437, 91, 39 },
		{ 437, 92, 39 },
		{ 437, 93, 39 },
		{ 437, 94, 39 },
		{ 437, 95, 39 },
		{ 437, 96, 39 },
		{ 437, 97, 39 },
		{ 437, 98, 39 },
		{ 437, 99, 39 },
		{ 437, 100, 39 },
		{ 437, 101, 39 },
		{ 437, 102, 39 },
		{ 437, 103, 39 },
		{ 437, 104, 39 },
		{ 437, 105, 39 },
		{ 437, 106, 39 },
		{ 437, 107, 39 },
		{ 437, 108, 39 },
		{ 437, 109, 39 },
		{ 437, 110, 39 },
		{ 437, 111, 39 },
		{ 437, 112, 39 },
		{ 437, 113, 39 },
		{ 437, 114, 39 },
		{ 437, 115, 39 },
		{ 437, 116, 39 },
		{ 437, 117, 39 },
		{ 437, 118, 39 },
		{ 437, 119, 39 },
		{ 437, 120, 39 },
		{ 437, 121, 39 },
		{ 438, 77, 39 },
		{ 438, 78, 39 },
		{ 438, 79, 39 },
		{ 438, 80, 39 },
		{ 438, 81, 39 },
		{ 438, 82, 39 },
		{ 438, 83, 39 },
		{ 438, 84, 39 },
		{ 438, 85, 39 },
		{ 438, 86, 39 },
		{ 438, 87, 39 },
		{ 438, 88, 39 },
		{ 438, 89, 39 },
		{ 438, 90, 39 },
		{ 438, 91, 39 },
		{ 438, 92, 39 },
		{ 438, 93, 39 },
		{ 438, 94, 39 },
		{ 438, 95, 39 },
		{ 438, 96, 39 },
		{ 438, 97, 39 },
		{ 438, 98, 39 },
		{ 438, 99, 39 },
		{ 438, 100, 39 },
		{ 438, 101, 39 },
		{ 438, 102, 39 },
		{ 438, 103, 39 },
		{ 438, 104, 39 },
		{ 438, 105, 39 },
		{ 438, 106, 39 },
		{ 438, 107, 39 },
		{ 438, 108, 39 },
		{ 438, 109, 39 },
		{ 438, 110, 39 },
		{ 438, 111, 39 },
		{ 438, 112, 39 },
		{ 438, 113, 39 },
		{ 438, 114, 39 },
		{ 438, 115, 39 },
		{ 438, 116, 39 },
		{ 438, 117, 39 },
		{ 438, 118, 39 },
		{ 438, 119, 39 },
		{ 438, 120, 39 },
		{ 438, 121, 39 },
		{ 439, 76, 39 },
		{ 439, 77, 39 },
		{ 439, 78, 39 },
		{ 439, 79, 39 },
		{ 439, 80, 39 },
		{ 439, 81, 39 },
		{ 439, 82, 39 },
		{ 439, 83, 39 },
		{ 439, 84, 39 },
		{ 439, 85, 39 },
		{ 439, 86, 39 },
		{ 439, 87, 39 },
		{ 439, 88, 39 },
		{ 439, 89, 39 },
		{ 439, 90, 39 },
		{ 439, 91, 39 },
		{ 439, 92, 39 },
		{ 439, 93, 39 },
		{ 439, 94, 39 },
		{ 439, 95, 39 },
		{ 439, 96, 39 },
		{ 439, 97, 39 },
		{ 439, 98, 39 },
		{ 439, 99, 39 },
		{ 439, 100, 39 },
		{ 439, 101, 39 },
		{ 439, 102, 39 },
		{ 439, 103, 39 },
		{ 439, 104, 39 },
		{ 439, 105, 39 },
		{ 439, 106, 39 },
		{ 439, 107, 39 },
		{ 439, 108, 39 },
		{ 439, 109, 39 },
		{ 439, 110, 39 },
		{ 439, 111, 39 },
		{ 439, 112, 39 },
		{ 439, 113, 39 },
		{ 439, 114, 39 },
		{ 439, 115, 39 },
		{ 439, 116, 39 },
		{ 439, 117, 39 },
		{ 439, 118, 39 },
		{ 439, 119, 39 },
		{ 439, 120, 39 },
		{ 440, 75, 39 },
		{ 440, 76, 39 },
		{ 440, 77, 39 },
		{ 440, 78, 39 },
		{ 440, 79, 39 },
		{ 440, 80, 39 },
		{ 440, 81, 39 },
		{ 440, 82, 39 },
		{ 440, 83, 39 },
		{ 440, 84, 39 },
		{ 440, 85, 39 },
		{ 440, 86, 39 },
		{ 440, 87, 39 },
		{ 440, 88, 39 },
		{ 440, 89, 39 },
		{ 440, 90, 39 },
		{ 440, 91, 39 },
		{ 440, 92, 39 },
		{ 440, 93, 39 },
		{ 440, 94, 39 },
		{ 440, 95, 39 },
		{ 440, 96, 39 },
		{ 440, 97, 39 },
		{ 440, 98, 39 },
		{ 440, 99, 39 },
		{ 440, 100, 39 },
		{ 440, 101, 39 },
		{ 440, 102, 39 },
		{ 440, 103, 39 },
		{ 440, 104, 39 },
		{ 440, 105, 39 },
		{ 440, 106, 39 },
		{ 440, 107, 39 },
		{ 440, 108, 39 },
		{ 440, 109, 39 },
		{ 440, 110, 39 },
		{ 440, 111, 39 },
		{ 440, 112, 39 },
		{ 440, 113, 39 },
		{ 440, 114, 39 },
		{ 440, 115, 39 },
		{ 440, 116, 39 },
		{ 440, 117, 39 },
		{ 440, 118, 39 },
		{ 440, 119, 39 },
		{ 441, 74, 39 },
		{ 441, 75, 39 },
		{ 441, 76, 39 },
		{ 441, 77, 39 },
		{ 441, 78, 39 },
		{ 441, 79, 39 },
		{ 441, 80, 39 },
		{ 441, 81, 39 },
		{ 441, 82, 39 },
		{ 441, 83, 39 },
		{ 441, 84, 39 },
		{ 441, 85, 39 },
		{ 441, 86, 39 },
		{ 441, 87, 39 },
		{ 441, 88, 39 },
		{ 441, 89, 39 },
		{ 441, 90, 39 },
		{ 441, 91, 39 },
		{ 441, 92, 39 },
		{ 441, 93, 39 },
		{ 441, 94, 39 },
		{ 441, 95, 39 },
		{ 441, 96, 39 },
		{ 441, 97, 39 },
		{ 441, 98, 39 },
		{ 441, 99, 39 },
		{ 441, 100, 39 },
		{ 441, 101, 39 },
		{ 441, 102, 39 },
		{ 441, 103, 39 },
		{ 441, 104, 39 },
		{ 441, 105, 39 },
		{ 441, 106, 39 },
		{ 441, 107, 39 },
		{ 441, 108, 39 },
		{ 441, 109, 39 },
		{ 441, 110, 39 },
		{ 441, 111, 39 },
		{ 441, 112, 39 },
		{ 441, 113, 39 },
		{ 441, 114, 39 },
		{ 441, 115, 39 },
		{ 441, 116, 39 },
		{ 441, 117, 39 },
		{ 441, 118, 39 },
		{ 442, 73, 39 },
		{ 442, 74, 39 },
		{ 442, 75, 39 },
		{ 442, 76, 39 },
		{ 442, 77, 39 },
		{ 442, 78, 39 },
		{ 442, 79, 39 },
		{ 442, 80, 39 },
		{ 442, 81, 39 },
		{ 442, 82, 39 },
		{ 442, 83, 39 },
		{ 442, 84, 39 },
		{ 442, 85, 39 },
		{ 442, 86, 39 },
		{ 442, 87, 39 },
		{ 442, 88, 39 },
		{ 442, 89, 39 },
		{ 442, 90, 39 },
		{ 442, 91, 39 },
		{ 442, 92, 39 },
		{ 442, 93, 39 },
		{ 442, 94, 39 },
		{ 442, 95, 39 },
		{ 442, 96, 39 },
		{ 442, 97, 39 },
		{ 442, 98, 39 },
		{ 442, 99, 39 },
		{ 442, 100, 39 },
		{ 442, 101, 39 },
		{ 442, 102, 39 },
		{ 442, 103, 39 },
		{ 442, 104, 39 },
		{ 442, 105, 39 },
		{ 442, 106, 39 },
		{ 442, 107, 39 },
		{ 442, 108, 39 },
		{ 442, 109, 39 },
		{ 442, 110, 39 },
		{ 442, 111, 39 },
		{ 442, 112, 39 },
		{ 442, 113, 39 },
		{ 442, 114, 39 },
		{ 442, 115, 39 },
		{ 442, 116, 39 },
		{ 442, 117, 39 },
		{ 442, 118, 39 },
		{ 443, 72, 39 },
		{ 443, 73, 39 },
		{ 443, 74, 39 },
		{ 443, 75, 39 },
		{ 443, 76, 39 },
		{ 443, 77, 39 },
		{ 443, 78, 39 },
		{ 443, 79, 39 },
		{ 443, 80, 39 },
		{ 443, 81, 39 },
		{ 443, 82, 39 },
		{ 443, 83, 39 },
		{ 443, 84, 39 },
		{ 443, 85, 39 },
		{ 443, 86, 39 },
		{ 443, 87, 39 },
		{ 443, 88, 39 },
		{ 443, 89, 39 },
		{ 443, 90, 39 },
		{ 443, 91, 39 },
		{ 443, 92, 39 },
		{ 443, 93, 39 },
		{ 443, 94, 39 },
		{ 443, 95, 39 },
		{ 443, 96, 39 },
		{ 443, 97, 39 },
		{ 443, 98, 39 },
		{ 443, 99, 39 },
		{ 443, 100, 39 },
		{ 443, 101, 39 },
		{ 443, 102, 39 },
		{ 443, 103, 39 },
		{ 443, 104, 39 },
		{ 443, 105, 39 },
		{ 443, 106, 39 },
		{ 443, 107, 39 },
		{ 443, 108, 39 },
		{ 443, 109, 39 },
		{ 443, 110, 39 },
		{ 443, 111, 39 },
		{ 443, 112, 39 },
		{ 443, 113, 39 },
		{ 443, 114, 39 },
		{ 443, 115, 39 },
		{ 443, 116, 39 },
		{ 443, 117, 39 },
		{ 444, 71, 39 },
		{ 444, 72, 39 },
		{ 444, 73, 39 },
		{ 444, 74, 39 },
		{ 444, 75, 39 },
		{ 444, 76, 39 },
		{ 444, 77, 39 },
		{ 444, 78, 39 },
		{ 444, 79, 39 },
		{ 444, 80, 39 },
		{ 444, 81, 39 },
		{ 444, 82, 39 },
		{ 444, 83, 39 },
		{ 444, 84, 39 },
		{ 444, 85, 39 },
		{ 444, 86, 39 },
		{ 444, 87, 39 },
		{ 444, 88, 39 },
		{ 444, 89, 39 },
		{ 444, 90, 39 },
		{ 444, 91, 39 },
		{ 444, 92, 39 },
		{ 444, 93, 39 },
		{ 444, 94, 39 },
		{ 444, 95, 39 },
		{ 444, 96, 39 },
		{ 444, 97, 39 },
		{ 444, 98, 39 },
		{ 444, 99, 39 },
		{ 444, 100, 39 },
		{ 444, 101, 39 },
		{ 444, 102, 39 },
		{ 444, 103, 39 },
		{ 444, 104, 39 },
		{ 444, 105, 39 },
		{ 444, 106, 39 },
		{ 444, 107, 39 },
		{ 444, 108, 39 },
		{ 444, 109, 39 },
		{ 444, 110, 39 },
		{ 444, 111, 39 },
		{ 444, 112, 39 },
		{ 444, 113, 39 },
		{ 444, 114, 39 },
		{ 444, 115, 39 },
		{ 444, 116, 39 },
		{ 445, 70, 39 },
		{ 445, 71, 39 },
		{ 445, 72, 39 },
		{ 445, 73, 39 },
		{ 445, 74, 39 },
		{ 445, 75, 39 },
		{ 445, 76, 39 },
		{ 445, 77, 39 },
		{ 445, 78, 39 },
		{ 445, 79, 39 },
		{ 445, 80, 39 },
		{ 445, 81, 39 },
		{ 445, 82, 39 },
		{ 445, 83, 39 },
		{ 445, 84, 39 },
		{ 445, 85, 39 },
		{ 445, 86, 39 },
		{ 445, 87, 39 },
		{ 445, 88, 39 },
		{ 445, 89, 39 },
		{ 445, 90, 39 },
		{ 445, 91, 39 },
		{ 445, 92, 39 },
		{ 445, 93, 39 },
		{ 445, 94, 39 },
		{ 445, 95, 39 },
		{ 445, 96, 39 },
		{ 445, 97, 39 },
		{ 445, 98, 39 },
		{ 445, 99, 39 },
		{ 445, 100, 39 },
		{ 445, 101, 39 },
		{ 445, 102, 39 },
		{ 445, 103, 39 },
		{ 445, 104, 39 },
		{ 445, 105, 39 },
		{ 445, 106, 39 },
		{ 445, 107, 39 },
		{ 445, 108, 39 },
		{ 445, 109, 39 },
		{ 445, 110, 39 },
		{ 445, 111, 39 },
		{ 445, 112, 39 },
		{ 445, 113, 39 },
		{ 445, 114, 39 },
		{ 445, 115, 39 },
		{ 446, 69, 39 },
		{ 446, 70, 39 },
		{ 446, 71, 39 },
		{ 446, 72, 39 },
		{ 446, 73, 39 },
		{ 446, 74, 39 },
		{ 446, 75, 39 },
		{ 446, 76, 39 },
		{ 446, 77, 39 },
		{ 446, 78, 39 },
		{ 446, 79, 39 },
		{ 446, 80, 39 },
		{ 446, 81, 39 },
		{ 446, 82, 39 },
		{ 446, 83, 39 },
		{ 446, 84, 39 },
		{ 446, 85, 39 },
		{ 446, 86, 39 },
		{ 446, 87, 39 },
		{ 446, 88, 39 },
		{ 446, 89, 39 },
		{ 446, 90, 39 },
		{ 446, 91, 39 },
		{ 446, 92, 39 },
		{ 446, 93, 39 },
		{ 446, 94, 39 },
		{ 446, 95, 39 },
		{ 446, 96, 39 },
		{ 446, 97, 39 },
		{ 446, 98, 39 },
		{ 446, 99, 39 },
		{ 446, 100, 39 },
		{ 446, 101, 39 },
		{ 446, 102, 39 },
		{ 446, 103, 39 },
		{ 446, 104, 39 },
		{ 446, 105, 39 },
		{ 446, 106, 39 },
		{ 446, 107, 39 },
		{ 446, 108, 39 },
		{ 446, 109, 39 },
		{ 446, 110, 39 },
		{ 446, 111, 39 },
		{ 446, 112, 39 },
		{ 446, 113, 39 },
		{ 446, 114, 39 },
		{ 446, 115, 39 },
		{ 447, 68, 39 },
		{ 447, 69, 39 },
		{ 447, 70, 39 },
		{ 447, 71, 39 },
		{ 447, 72, 39 },
		{ 447, 73, 39 },
		{ 447, 74, 39 },
		{ 447, 75, 39 },
		{ 447, 76, 39 },
		{ 447, 77, 39 },
		{ 447, 78, 39 },
		{ 447, 79, 39 },
		{ 447, 80, 39 },
		{ 447, 81, 39 },
		{ 447, 82, 39 },
		{ 447, 83, 39 },
		{ 447, 84, 39 },
		{ 447, 85, 39 },
		{ 447, 86, 39 },
		{ 447, 87, 39 },
		{ 447, 88, 39 },
		{ 447, 89, 39 },
		{ 447, 90, 39 },
		{ 447, 91, 39 },
		{ 447, 92, 39 },
		{ 447, 93, 39 },
		{ 447, 94, 39 },
		{ 447, 95, 39 },
		{ 447, 96, 39 },
		{ 447, 97, 39 },
		{ 447, 98, 39 },
		{ 447, 99, 39 },
		{ 447, 100, 39 },
		{ 447, 101, 39 },
		{ 447, 102, 39 },
		{ 447, 103, 39 },
		{ 447, 104, 39 },
		{ 447, 105, 39 },
		{ 447, 106, 39 },
		{ 447, 107, 39 },
		{ 447, 108, 39 },
		{ 447, 109, 39 },
		{ 447, 110, 39 },
		{ 447, 111, 39 },
		{ 447, 112, 39 },
		{ 447, 113, 39 },
		{ 447, 114, 39 },
		{ 448, 67, 39 },
		{ 448, 68, 39 },
		{ 448, 69, 39 },
		{ 448, 70, 39 },
		{ 448, 71, 39 },
		{ 448, 72, 39 },
		{ 448, 73, 39 },
		{ 448, 74, 39 },
		{ 448, 75, 39 },
		{ 448, 76, 39 },
		{ 448, 77, 39 },
		{ 448, 78, 39 },
		{ 448, 79, 39 },
		{ 448, 80, 39 },
		{ 448, 81, 39 },
		{ 448, 82, 39 },
		{ 448, 83, 39 },
		{ 448, 84, 39 },
		{ 448, 85, 39 },
		{ 448, 86, 39 },
		{ 448, 87, 39 },
		{ 448, 88, 39 },
		{ 448, 89, 39 },
		{ 448, 90, 39 },
		{ 448, 91, 39 },
		{ 448, 92, 39 },
		{ 448, 93, 39 },
		{ 448, 94, 39 },
		{ 448, 95, 39 },
		{ 448, 96, 39 },
		{ 448, 97, 39 },
		{ 448, 98, 39 },
		{ 448, 99, 39 },
		{ 448, 100, 39 },
		{ 448, 101, 39 },
		{ 448, 102, 39 },
		{ 448, 103, 39 },
		{ 448, 104, 39 },
		{ 448, 105, 39 },
		{ 448, 106, 39 },
		{ 448, 107, 39 },
		{ 448, 108, 39 },
		{ 448, 109, 39 },
		{ 448, 110, 39 },
		{ 448, 111, 39 },
		{ 448, 112, 39 },
		{ 448, 113, 39 },
		{ 449, 66, 39 },
		{ 449, 67, 39 },
		{ 449, 68, 39 },
		{ 449, 69, 39 },
		{ 449, 70, 39 },
		{ 449, 71, 39 },
		{ 449, 72, 39 },
		{ 449, 73, 39 },
		{ 449, 74, 39 },
		{ 449, 75, 39 },
		{ 449, 76, 39 },
		{ 449, 77, 39 },
		{ 449, 78, 39 },
		{ 449, 79, 39 },
		{ 449, 80, 39 },
		{ 449, 81, 39 },
		{ 449, 82, 39 },
		{ 449, 83, 39 },
		{ 449, 84, 39 },
		{ 449, 85, 39 },
		{ 449, 86, 39 },
		{ 449, 87, 39 },
		{ 449, 88, 39 },
		{ 449, 89, 39 },
		{ 449, 90, 39 },
		{ 449, 91, 39 },
		{ 449, 92, 39 },
		{ 449, 93, 39 },
		{ 449, 94, 39 },
		{ 449, 95, 39 },
		{ 449, 96, 39 },
		{ 449, 97, 39 },
		{ 449, 98, 39 },
		{ 449, 99, 39 },
		{ 449, 100, 39 },
		{ 449, 101, 39 },
		{ 449, 102, 39 },
		{ 449, 103, 39 },
		{ 449, 104, 39 },
		{ 449, 105, 39 },
		{ 449, 106, 39 },
		{ 449, 107, 39 },
		{ 449, 108, 39 },
		{ 449, 109, 39 },
		{ 449, 110, 39 },
		{ 449, 111, 39 },
		{ 449, 112, 39 },
		{ 450, 65, 39 },
		{ 450, 66, 39 },
		{ 450, 67, 39 },
		{ 450, 68, 39 },
		{ 450, 69, 39 },
		{ 450, 70, 39 },
		{ 450, 71, 39 },
		{ 450, 72, 39 },
		{ 450, 73, 39 },
		{ 450, 74, 39 },
		{ 450, 75, 39 },
		{ 450, 76, 39 },
		{ 450, 77, 39 },
		{ 450, 78, 39 },
		{ 450, 79, 39 },
		{ 450, 80, 39 },
		{ 450, 81, 39 },
		{ 450, 82, 39 },
		{ 450, 83, 39 },
		{ 450, 84, 39 },
		{ 450, 85, 39 },
		{ 450, 86, 39 },
		{ 450, 87, 39 },
		{ 450, 88, 39 },
		{ 450, 89, 39 },
		{ 450, 90, 39 },
		{ 450, 91, 39 },
		{ 450, 92, 39 },
		{ 450, 93, 39 },
		{ 450, 94, 39 },
		{ 450, 95, 39 },
		{ 450, 96, 39 },
		{ 450, 97, 39 },
		{ 450, 98, 39 },
		{ 450, 99, 39 },
		{ 450, 100, 39 },
		{ 450, 101, 39 },
		{ 450, 102, 39 },
		{ 450, 103, 39 },
		{ 450, 104, 39 },
		{ 450, 105, 39 },
		{ 450, 106, 39 },
		{ 450, 107, 39 },
		{ 450, 108, 39 },
		{ 450, 109, 39 },
		{ 450, 110, 39 },
		{ 450, 111, 39 },
		{ 450, 112, 39 },
		{ 451, 64, 39 },
		{ 451, 65, 39 },
		{ 451, 66, 39 },
		{ 451, 67, 39 },
		{ 451, 68, 39 },
		{ 451, 69, 39 },
		{ 451, 70, 39 },
		{ 451, 71, 39 },
		{ 451, 72, 39 },
		{ 451, 73, 39 },
		{ 451, 74, 39 },
		{ 451, 75, 39 },
		{ 451, 76, 39 },
		{ 451, 77, 39 },
		{ 451, 78, 39 },
		{ 451, 79, 39 },
		{ 451, 80, 39 },
		{ 451, 81, 39 },
		{ 451, 82, 39 },
		{ 451, 83, 39 },
		{ 451, 84, 39 },
		{ 451, 85, 39 },
		{ 451, 86, 39 },
		{ 451, 87, 39 },
		{ 451, 88, 39 },
		{ 451, 89, 39 },
		{ 451, 90, 39 },
		{ 451, 91, 39 },
		{ 451, 92, 39 },
		{ 451, 93, 39 },
		{ 451, 94, 39 },
		{ 451, 95, 39 },
		{ 451, 96, 39 },
		{ 451, 97, 39 },
		{ 451, 98, 39 },
		{ 451, 99, 39 },
		{ 451, 100, 39 },
		{ 451, 101, 39 },
		{ 451, 102, 39 },
		{ 451, 103, 39 },
		{ 451, 104, 39 },
		{ 451, 105, 39 },
		{ 451, 106, 39 },
		{ 451, 107, 39 },
		{ 451, 108, 39 },
		{ 451, 109, 39 },
		{ 451, 110, 39 },
		{ 451, 111, 39 },
		{ 452, 63, 39 },
		{ 452, 64, 39 },
		{ 452, 65, 39 },
		{ 452, 66, 39 },
		{ 452, 67, 39 },
		{ 452, 68, 39 },
		{ 452, 69, 39 },
		{ 452, 70, 39 },
		{ 452, 71, 39 },
		{ 452, 72, 39 },
		{ 452, 73, 39 },
		{ 452, 74, 39 },
		{ 452, 75, 39 },
		{ 452, 76, 39 },
		{ 452, 77, 39 },
		{ 452, 78, 39 },
		{ 452, 79, 39 },
		{ 452, 80, 39 },
		{ 452, 81, 39 },
		{ 452, 82, 39 },
		{ 452, 83, 39 },
		{ 452, 84, 39 },
		{ 452, 85, 39 },
		{ 452, 86, 39 },
		{ 452, 87, 39 },
		{ 452, 88, 39 },
		{ 452, 89, 39 },
		{ 452, 90, 39 },
		{ 452, 91, 39 },
		{ 452, 92, 39 },
		{ 452, 93, 39 },
		{ 452, 94, 39 },
		{ 452, 95, 39 },
		{ 452, 96, 39 },
		{ 452, 97, 39 },
		{ 452, 98, 39 },
		{ 452, 99, 39 },
		{ 452, 100, 39 },
		{ 452, 101, 39 },
		{ 452, 102, 39 },
		{ 452, 103, 39 },
		{ 452, 104, 39 },
		{ 452, 105, 39 },
		{ 452, 106, 39 },
		{ 452, 107, 39 },
		{ 452, 108, 39 },
		{ 452, 109, 39 },
		{ 452, 110, 39 },
		{ 453, 62, 39 },
		{ 453, 63, 39 },
		{ 453, 64, 39 },
		{ 453, 65, 39 },
		{ 453, 66, 39 },
		{ 453, 67, 39 },
		{ 453, 68, 39 },
		{ 453, 69, 39 },
		{ 453, 70, 39 },
		{ 453, 71, 39 },
		{ 453, 72, 39 },
		{ 453, 73, 39 },
		{ 453, 74, 39 },
		{ 453, 75, 39 },
		{ 453, 76, 39 },
		{ 453, 77, 39 },
		{ 453, 78, 39 },
		{ 453, 79, 39 },
		{ 453, 80, 39 },
		{ 453, 81, 39 },
		{ 453, 82, 39 },
		{ 453, 83, 39 },
		{ 453, 84, 39 },
		{ 453, 85, 39 },
		{ 453, 86, 39 },
		{ 453, 87, 39 },
		{ 453, 88, 39 },
		{ 453, 89, 39 },
		{ 453, 90, 39 },
		{ 453, 91, 39 },
		{ 453, 92, 39 },
		{ 453, 93, 39 },
		{ 453, 94, 39 },
		{ 453, 95, 39 },
		{ 453, 96, 39 },
		{ 453, 97, 39 },
		{ 453, 98, 39 },
		{ 453, 99, 39 },
		{ 453, 100, 39 },
		{ 453, 101, 39 },
		{ 453, 102, 39 },
		{ 453, 103, 39 },
		{ 453, 104, 39 },
		{ 453, 105, 39 },
		{ 453, 106, 39 },
		{ 453, 107, 39 },
		{ 453, 108, 39 },
		{ 453, 109, 39 },
		{ 454, 61, 39 },
		{ 454, 62, 39 },
		{ 454, 63, 39 },
		{ 454, 64, 39 },
		{ 454, 65, 39 },
		{ 454, 66, 39 },
		{ 454, 67, 39 },
		{ 454, 68, 39 },
		{ 454, 69, 39 },
		{ 454, 70, 39 },
		{ 454, 71, 39 },
		{ 454, 72, 39 },
		{ 454, 73, 39 },
		{ 454, 74, 39 },
		{ 454, 75, 39 },
		{ 454, 76, 39 },
		{ 454, 77, 39 },
		{ 454, 78, 39 },
		{ 454, 79, 39 },
		{ 454, 80, 39 },
		{ 454, 81, 39 },
		{ 454, 82, 39 },
		{ 454, 83, 39 },
		{ 454, 84, 39 },
		{ 454, 85, 39 },
		{ 454, 86, 39 },
		{ 454, 87, 39 },
		{ 454, 88, 39 },
		{ 454, 89, 39 },
		{ 454, 90, 39 },
		{ 454, 91, 39 },
		{ 454, 92, 39 },
		{ 454, 93, 39 },
		{ 454, 94, 39 },
		{ 454, 95, 39 },
		{ 454, 96, 39 },
		{ 454, 97, 39 },
		{ 454, 98, 39 },
		{ 454, 99, 39 },
		{ 454, 100, 39 },
		{ 454, 101, 39 },
		{ 454, 102, 39 },
		{ 454, 103, 39 },
		{ 454, 104, 39 },
		{ 454, 105, 39 },
		{ 454, 106, 39 },
		{ 454, 107, 39 },
		{ 454, 108, 39 },
		{ 454, 109, 39 },
		{ 455, 60, 39 },
		{ 455, 61, 39 },
		{ 455, 62, 39 },
		{ 455, 63, 39 },
		{ 455, 64, 39 },
		{ 455, 65, 39 },
		{ 455, 66, 39 },
		{ 455, 67, 39 },
		{ 455, 68, 39 },
		{ 455, 69, 39 },
		{ 455, 70, 39 },
		{ 455, 71, 39 },
		{ 455, 72, 39 },
		{ 455, 73, 39 },
		{ 455, 74, 39 },
		{ 455, 75, 39 },
		{ 455, 76, 39 },
		{ 455, 77, 39 },
		{ 455, 78, 39 },
		{ 455, 79, 39 },
		{ 455, 80, 39 },
		{ 455, 81, 39 },
		{ 455, 82, 39 },
		{ 455, 83, 39 },
		{ 455, 84, 39 },
		{ 455, 85, 39 },
		{ 455, 86, 39 },
		{ 455, 87, 39 },
		{ 455, 88, 39 },
		{ 455, 89, 39 },
		{ 455, 90, 39 },
		{ 455, 91, 39 },
		{ 455, 92, 39 },
		{ 455, 93, 39 },
		{ 455, 94, 39 },
		{ 455, 95, 39 },
		{ 455, 96, 39 },
		{ 455, 97, 39 },
		{ 455, 98, 39 },
		{ 455, 99, 39 },
		{ 455, 100, 39 },
		{ 455, 101, 39 },
		{ 455, 102, 39 },
		{ 455, 103, 39 },
		{ 455, 104, 39 },
		{ 455, 105, 39 },
		{ 455, 106, 39 },
		{ 455, 107, 39 },
		{ 455, 108, 39 },
		{ 456, 59, 39 },
		{ 456, 60, 39 },
		{ 456, 61, 39 },
		{ 456, 62, 39 },
		{ 456, 63, 39 },
		{ 456, 64, 39 },
		{ 456, 65, 39 },
		{ 456, 66, 39 },
		{ 456, 67, 39 },
		{ 456, 68, 39 },
		{ 456, 69, 39 },
		{ 456, 70, 39 },
		{ 456, 71, 39 },
		{ 456, 72, 39 },
		{ 456, 73, 39 },
		{ 456, 74, 39 },
		{ 456, 75, 39 },
		{ 456, 76, 39 },
		{ 456, 77, 39 },
		{ 456, 78, 39 },
		{ 456, 79, 39 },
		{ 456, 80, 39 },
		{ 456, 81, 39 },
		{ 456, 82, 39 },
		{ 456, 83, 39 },
		{ 456, 84, 39 },
		{ 456, 85, 39 },
		{ 456, 86, 39 },
		{ 456, 87, 39 },
		{ 456, 88, 39 },
		{ 456, 89, 39 },
		{ 456, 90, 39 },
		{ 456, 91, 39 },
		{ 456, 92, 39 },
		{ 456, 93, 39 },
		{ 456, 94, 39 },
		{ 456, 95, 39 },
		{ 456, 96, 39 },
		{ 456, 97, 39 },
		{ 456, 98, 39 },
		{ 456, 99, 39 },
		{ 456, 100, 39 },
		{ 456, 101, 39 },
		{ 456, 102, 39 },
		{ 456, 103, 39 },
		{ 456, 104, 39 },
		{ 456, 105, 39 },
		{ 456, 106, 39 },
		{ 456, 107, 39 },
		{ 457, 58, 39 },
		{ 457, 59, 39 },
		{ 457, 60, 39 },
		{ 457, 61, 39 },
		{ 457, 62, 39 },
		{ 457, 63, 39 },
		{ 457, 64, 39 },
		{ 457, 65, 39 },
		{ 457, 66, 39 },
		{ 457, 67, 39 },
		{ 457, 68, 39 },
		{ 457, 69, 39 },
		{ 457, 70, 39 },
		{ 457, 71, 39 },
		{ 457, 72, 39 },
		{ 457, 73, 39 },
		{ 457, 74, 39 },
		{ 457, 75, 39 },
		{ 457, 76, 39 },
		{ 457, 77, 39 },
		{ 457, 78, 39 },
		{ 457, 79, 39 },
		{ 457, 80, 39 },
		{ 457, 81, 39 },
		{ 457, 82, 39 },
		{ 457, 83, 39 },
		{ 457, 84, 39 },
		{ 457, 85, 39 },
		{ 457, 86, 39 },
		{ 457, 87, 39 },
		{ 457, 88, 39 },
		{ 457, 89, 39 },
		{ 457, 90, 39 },
		{ 457, 91, 39 },
		{ 457, 92, 39 },
		{ 457, 93, 39 },
		{ 457, 94, 39 },
		{ 457, 95, 39 },
		{ 457, 96, 39 },
		{ 457, 97, 39 },
		{ 457, 98, 39 },
		{ 457, 99, 39 },
		{ 457, 100, 39 },
		{ 457, 101, 39 },
		{ 457, 102, 39 },
		{ 457, 103, 39 },
		{ 457, 104, 39 },
		{ 457, 105, 39 },
		{ 457, 106, 39 },
		{ 458, 57, 39 },
		{ 458, 58, 39 },
		{ 458, 59, 39 },
		{ 458, 60, 39 },
		{ 458, 61, 39 },
		{ 458, 62, 39 },
		{ 458, 63, 39 },
		{ 458, 64, 39 },
		{ 458, 65, 39 },
		{ 458, 66, 39 },
		{ 458, 67, 39 },
		{ 458, 68, 39 },
		{ 458, 69, 39 },
		{ 458, 70, 39 },
		{ 458, 71, 39 },
		{ 458, 72, 39 },
		{ 458, 73, 39 },
		{ 458, 74, 39 },
		{ 458, 75, 39 },
		{ 458, 76, 39 },
		{ 458, 77, 39 },
		{ 458, 78, 39 },
		{ 458, 79, 39 },
		{ 458, 80, 39 },
		{ 458, 81, 39 },
		{ 458, 82, 39 },
		{ 458, 83, 39 },
		{ 458, 84, 39 },
		{ 458, 85, 39 },
		{ 458, 86, 39 },
		{ 458, 87, 39 },
		{ 458, 88, 39 },
		{ 458, 89, 39 },
		{ 458, 90, 39 },
		{ 458, 91, 39 },
		{ 458, 92, 39 },
		{ 458, 93, 39 },
		{ 458, 94, 39 },
		{ 458, 95, 39 },
		{ 458, 96, 39 },
		{ 458, 97, 39 },
		{ 458, 98, 39 },
		{ 458, 99, 39 },
		{ 458, 100, 39 },
		{ 458, 101, 39 },
		{ 458, 102, 39 },
		{ 458, 103, 39 },
		{ 458, 104, 39 },
		{ 458, 105, 39 },
		{ 458, 106, 39 },
		{ 459, 56, 39 },
		{ 459, 57, 39 },
		{ 459, 58, 39 },
		{ 459, 59, 39 },
		{ 459, 60, 39 },
		{ 459, 61, 39 },
		{ 459, 62, 39 },
		{ 459, 63, 39 },
		{ 459, 64, 39 },
		{ 459, 65, 39 },
		{ 459, 66, 39 },
		{ 459, 67, 39 },
		{ 459, 68, 39 },
		{ 459, 69, 39 },
		{ 459, 70, 39 },
		{ 459, 71, 39 },
		{ 459, 72, 39 },
		{ 459, 73, 39 },
		{ 459, 74, 39 },
		{ 459, 75, 39 },
		{ 459, 76, 39 },
		{ 459, 77, 39 },
		{ 459, 78, 39 },
		{ 459, 79, 39 },
		{ 459, 80, 39 },
		{ 459, 81, 39 },
		{ 459, 82, 39 },
		{ 459, 83, 39 },
		{ 459, 84, 39 },
		{ 459, 85, 39 },
		{ 459, 86, 39 },
		{ 459, 87, 39 },
		{ 459, 88, 39 },
		{ 459, 89, 39 },
		{ 459, 90, 39 },
		{ 459, 91, 39 },
		{ 459, 92, 39 },
		{ 459, 93, 39 },
		{ 459, 94, 39 },
		{ 459, 95, 39 },
		{ 459, 96, 39 },
		{ 459, 97, 39 },
		{ 459, 98, 39 },
		{ 459, 99, 39 },
		{ 459, 100, 39 },
		{ 459, 101, 39 },
		{ 459, 102, 39 },
		{ 459, 103, 39 },
		{ 459, 104, 39 },
		{ 459, 105, 39 },
		{ 460, 55, 39 },
		{ 460, 56, 39 },
		{ 460, 57, 39 },
		{ 460, 58, 39 },
		{ 460, 59, 39 },
		{ 460, 60, 39 },
		{ 460, 61, 39 },
		{ 460, 62, 39 },
		{ 460, 63, 39 },
		{ 460, 64, 39 },
		{ 460, 65, 39 },
		{ 460, 66, 39 },
		{ 460, 67, 39 },
		{ 460, 68, 39 },
		{ 460, 69, 39 },
		{ 460, 70, 39 },
		{ 460, 71, 39 },
		{ 460, 72, 39 },
		{ 460, 73, 39 },
		{ 460, 74, 39 },
		{ 460, 75, 39 },
		{ 460, 76, 39 },
		{ 460, 77, 39 },
		{ 460, 78, 39 },
		{ 460, 79, 39 },
		{ 460, 80, 39 },
		{ 460, 81, 39 },
		{ 460, 82, 39 },
		{ 460, 83, 39 },
		{ 460, 84, 39 },
		{ 460, 85, 39 },
		{ 460, 86, 39 },
		{ 460, 87, 39 },
		{ 460, 88, 39 },
		{ 460, 89, 39 },
		{ 460, 90, 39 },
		{ 460, 91, 39 },
		{ 460, 92, 39 },
		{ 460, 93, 39 },
		{ 460, 94, 39 },
		{ 460, 95, 39 },
		{ 460, 96, 39 },
		{ 460, 97, 39 },
		{ 460, 98, 39 },
		{ 460, 99, 39 },
		{ 460, 100, 39 },
		{ 460, 101, 39 },
		{ 460, 102, 39 },
		{ 460, 103, 39 },
		{ 460, 104, 39 },
		{ 461, 54, 39 },
		{ 461, 55, 39 },
		{ 461, 56, 39 },
		{ 461, 57, 39 },
		{ 461, 58, 39 },
		{ 461, 59, 39 },
		{ 461, 60, 39 },
		{ 461, 61, 39 },
		{ 461, 62, 39 },
		{ 461, 63, 39 },
		{ 461, 64, 39 },
		{ 461, 65, 39 },
		{ 461, 66, 39 },
		{ 461, 67, 39 },
		{ 461, 68, 39 },
		{ 461, 69, 39 },
		{ 461, 70, 39 },
		{ 461, 71, 39 },
		{ 461, 72, 39 },
		{ 461, 73, 39 },
		{ 461, 74, 39 },
		{ 461, 75, 39 },
		{ 461, 76, 39 },
		{ 461, 77, 39 },
		{ 461, 78, 39 },
		{ 461, 79, 39 },
		{ 461, 80, 39 },
		{ 461, 81, 39 },
		{ 461, 82, 39 },
		{ 461, 83, 39 },
		{ 461, 84, 39 },
		{ 461, 85, 39 },
		{ 461, 86, 39 },
		{ 461, 87, 39 },
		{ 461, 88, 39 },
		{ 461, 89, 39 },
		{ 461, 90, 39 },
		{ 461, 91, 39 },
		{ 461, 92, 39 },
		{ 461, 93, 39 },
		{ 461, 94, 39 },
		{ 461, 95, 39 },
		{ 461, 96, 39 },
		{ 461, 97, 39 },
		{ 461, 98, 39 },
		{ 461, 99, 39 },
		{ 461, 100, 39 },
		{ 461, 101, 39 },
		{ 461, 102, 39 },
		{ 461, 103, 39 },
		{ 462, 53, 39 },
		{ 462, 54, 39 },
		{ 462, 55, 39 },
		{ 462, 56, 39 },
		{ 462, 57, 39 },
		{ 462, 58, 39 },
		{ 462, 59, 39 },
		{ 462, 60, 39 },
		{ 462, 61, 39 },
		{ 462, 62, 39 },
		{ 462, 63, 39 },
		{ 462, 64, 39 },
		{ 462, 65, 39 },
		{ 462, 66, 39 },
		{ 462, 67, 39 },
		{ 462, 68, 39 },
		{ 462, 69, 39 },
		{ 462, 70, 39 },
		{ 462, 71, 39 },
		{ 462, 72, 39 },
		{ 462, 73, 39 },
		{ 462, 74, 39 },
		{ 462, 75, 39 },
		{ 462, 76, 39 },
		{ 462, 77, 39 },
		{ 462, 78, 39 },
		{ 462, 79, 39 },
		{ 462, 80, 39 },
		{ 462, 81, 39 },
		{ 462, 82, 39 },
		{ 462, 83, 39 },
		{ 462, 84, 39 },
		{ 462, 85, 39 },
		{ 462, 86, 39 },
		{ 462, 87, 39 },
		{ 462, 88, 39 },
		{ 462, 89, 39 },
		{ 462, 90, 39 },
		{ 462, 91, 39 },
		{ 462, 92, 39 },
		{ 462, 93, 39 },
		{ 462, 94, 39 },
		{ 462, 95, 39 },
		{ 462, 96, 39 },
		{ 462, 97, 39 },
		{ 462, 98, 39 },
		{ 462, 99, 39 },
		{ 462, 100, 39 },
		{ 462, 101, 39 },
		{ 462, 102, 39 },
		{ 462, 103, 39 },
		{ 463, 52, 39 },
		{ 463, 53, 39 },
		{ 463, 54, 39 },
		{ 463, 55, 39 },
		{ 463, 56, 39 },
		{ 463, 57, 39 },
		{ 463, 58, 39 },
		{ 463, 59, 39 },
		{ 463, 60, 39 },
		{ 463, 61, 39 },
		{ 463, 62, 39 },
		{ 463, 63, 39 },
		{ 463, 64, 39 },
		{ 463, 65, 39 },
		{ 463, 66, 39 },
		{ 463, 67, 39 },
		{ 463, 68, 39 },
		{ 463, 69, 39 },
		{ 463, 70, 39 },
		{ 463, 71, 39 },
		{ 463, 72, 39 },
		{ 463, 73, 39 },
		{ 463, 74, 39 },
		{ 463, 75, 39 },
		{ 463, 76, 39 },
		{ 463, 77, 39 },
		{ 463, 78, 39 },
		{ 463, 79, 39 },
		{ 463, 80, 39 },
		{ 463, 81, 39 },
		{ 463, 82, 39 },
		{ 463, 83, 39 },
		{ 463, 84, 39 },
		{ 463, 85, 39 },
		{ 463, 86, 39 },
		{ 463, 87, 39 },
		{ 463, 88, 39 },
		{ 463, 89, 39 },
		{ 463, 90, 39 },
		{ 463, 91, 39 },
		{ 463, 92, 39 },
		{ 463, 93, 39 },
		{ 463, 94, 39 },
		{ 463, 95, 39 },
		{ 463, 96, 39 },
		{ 463, 97, 39 },
		{ 463, 98, 39 },
		{ 463, 99, 39 },
		{ 463, 100, 39 },
		{ 463, 101, 39 },
		{ 463, 102, 39 },
		{ 464, 51, 39 },
		{ 464, 52, 39 },
		{ 464, 53, 39 },
		{ 464, 54, 39 },
		{ 464, 55, 39 },
		{ 464, 56, 39 },
		{ 464, 57, 39 },
		{ 464, 58, 39 },
		{ 464, 59, 39 },
		{ 464, 60, 39 },
		{ 464, 61, 39 },
		{ 464, 62, 39 },
		{ 464, 63, 39 },
		{ 464, 64, 39 },
		{ 464, 65, 39 },
		{ 464, 66, 39 },
		{ 464, 67, 39 },
		{ 464, 68, 39 },
		{ 464, 69, 39 },
		{ 464, 70, 39 },
		{ 464, 71, 39 },
		{ 464, 72, 39 },
		{ 464, 73, 39 },
		{ 464, 74, 39 },
		{ 464, 75, 39 },
		{ 464, 76, 39 },
		{ 464, 77, 39 },
		{ 464, 78, 39 },
		{ 464, 79, 39 },
		{ 464, 80, 39 },
		{ 464, 81, 39 },
		{ 464, 82, 39 },
		{ 464, 83, 39 },
		{ 464, 84, 39 },
		{ 464, 85, 39 },
		{ 464, 86, 39 },
		{ 464, 87, 39 },
		{ 464, 88, 39 },
		{ 464, 89, 39 },
		{ 464, 90, 39 },
		{ 464, 91, 39 },
		{ 464, 92, 39 },
		{ 464, 93, 39 },
		{ 464, 94, 39 },
		{ 464, 95, 39 },
		{ 464, 96, 39 },
		{ 464, 97, 39 },
		{ 464, 98, 39 },
		{ 464, 99, 39 },
		{ 464, 100, 39 },
		{ 464, 101, 39 },
		{ 465, 50, 39 },
		{ 465, 51, 39 },
		{ 465, 52, 39 },
		{ 465, 53, 39 },
		{ 465, 54, 39 },
		{ 465, 55, 39 },
		{ 465, 56, 39 },
		{ 465, 57, 39 },
		{ 465, 58, 39 },
		{ 465, 59, 39 },
		{ 465, 60, 39 },
		{ 465, 61, 39 },
		{ 465, 62, 39 },
		{ 465, 63, 39 },
		{ 465, 64, 39 },
		{ 465, 65, 39 },
		{ 465, 66, 39 },
		{ 465, 67, 39 },
		{ 465, 68, 39 },
		{ 465, 69, 39 },
		{ 465, 70, 39 },
		{ 465, 71, 39 },
		{ 465, 72, 39 },
		{ 465, 73, 39 },
		{ 465, 74, 39 },
		{ 465, 75, 39 },
		{ 465, 76, 39 },
		{ 465, 77, 39 },
		{ 465, 78, 39 },
		{ 465, 79, 39 },
		{ 465, 80, 39 },
		{ 465, 81, 39 },
		{ 465, 82, 39 },
		{ 465, 83, 39 },
		{ 465, 84, 39 },
		{ 465, 85, 39 },
		{ 465, 86, 39 },
		{ 465, 87, 39 },
		{ 465, 88, 39 },
		{ 465, 89, 39 },
		{ 465, 90, 39 },
		{ 465, 91, 39 },
		{ 465, 92, 39 },
		{ 465, 93, 39 },
		{ 465, 94, 39 },
		{ 465, 95, 39 },
		{ 465, 96, 39 },
		{ 465, 97, 39 },
		{ 465, 98, 39 },
		{ 465, 99, 39 },
		{ 465, 100, 39 },
		{ 466, 49, 39 },
		{ 466, 50, 39 },
		{ 466, 51, 39 },
		{ 466, 52, 39 },
		{ 466, 53, 39 },
		{ 466, 54, 39 },
		{ 466, 55, 39 },
		{ 466, 56, 39 },
		{ 466, 57, 39 },
		{ 466, 58, 39 },
		{ 466, 59, 39 },
		{ 466, 60, 39 },
		{ 466, 61, 39 },
		{ 466, 62, 39 },
		{ 466, 63, 39 },
		{ 466, 64, 39 },
		{ 466, 65, 39 },
		{ 466, 66, 39 },
		{ 466, 67, 39 },
		{ 466, 68, 39 },
		{ 466, 69, 39 },
		{ 466, 70, 39 },
		{ 466, 71, 39 },
		{ 466, 72, 39 },
		{ 466, 73, 39 },
		{ 466, 74, 39 },
		{ 466, 75, 39 },
		{ 466, 76, 39 },
		{ 466, 77, 39 },
		{ 466, 78, 39 },
		{ 466, 79, 39 },
		{ 466, 80, 39 },
		{ 466, 81, 39 },
		{ 466, 82, 39 },
		{ 466, 83, 39 },
		{ 466, 84, 39 },
		{ 466, 85, 39 },
		{ 466, 86, 39 },
		{ 466, 87, 39 },
		{ 466, 88, 39 },
		{ 466, 89, 39 },
		{ 466, 90, 39 },
		{ 466, 91, 39 },
		{ 466, 92, 39 },
		{ 466, 93, 39 },
		{ 466, 94, 39 },
		{ 466, 95, 39 },
		{ 466, 96, 39 },
		{ 466, 97, 39 },
		{ 466, 98, 39 },
		{ 466, 99, 39 },
		{ 466, 100, 39 },
		{ 467, 48, 39 },
		{ 467, 49, 39 },
		{ 467, 50, 39 },
		{ 467, 51, 39 },
		{ 467, 52, 39 },
		{ 467, 53, 39 },
		{ 467, 54, 39 },
		{ 467, 55, 39 },
		{ 467, 56, 39 },
		{ 467, 57, 39 },
		{ 467, 58, 39 },
		{ 467, 59, 39 },
		{ 467, 60, 39 },
		{ 467, 61, 39 },
		{ 467, 62, 39 },
		{ 467, 63, 39 },
		{ 467, 64, 39 },
		{ 467, 65, 39 },
		{ 467, 66, 39 },
		{ 467, 67, 39 },
		{ 467, 68, 39 },
		{ 467, 69, 39 },
		{ 467, 70, 39 },
		{ 467, 71, 39 },
		{ 467, 72, 39 },
		{ 467, 73, 39 },
		{ 467, 74, 39 },
		{ 467, 75, 39 },
		{ 467, 76, 39 },
		{ 467, 77, 39 },
		{ 467, 78, 39 },
		{ 467, 79, 39 },
		{ 467, 80, 39 },
		{ 467, 81, 39 },
		{ 467, 82, 39 },
		{ 467, 83, 39 },
		{ 467, 84, 39 },
		{ 467, 85, 39 },
		{ 467, 86, 39 },
		{ 467, 87, 39 },
		{ 467, 88, 39 },
		{ 467, 89, 39 },
		{ 467, 90, 39 },
		{ 467, 91, 39 },
		{ 467, 92, 39 },
		{ 467, 93, 39 },
		{ 467, 94, 39 },
		{ 467, 95, 39 },
		{ 467, 96, 39 },
		{ 467, 97, 39 },
		{ 467, 98, 39 },
		{ 467, 99, 39 },
		{ 468, 47, 39 },
		{ 468, 48, 39 },
		{ 468, 49, 39 },
		{ 468, 50, 39 },
		{ 468, 51, 39 },
		{ 468, 52, 39 },
		{ 468, 53, 39 },
		{ 468, 54, 39 },
		{ 468, 55, 39 },
		{ 468, 56, 39 },
		{ 468, 57, 39 },
		{ 468, 58, 39 },
		{ 468, 59, 39 },
		{ 468, 60, 39 },
		{ 468, 61, 39 },
		{ 468, 62, 39 },
		{ 468, 63, 39 },
		{ 468, 64, 39 },
		{ 468, 65, 39 },
		{ 468, 66, 39 },
		{ 468, 67, 39 },
		{ 468, 68, 39 },
		{ 468, 69, 39 },
		{ 468, 70, 39 },
		{ 468, 71, 39 },
		{ 468, 72, 39 },
		{ 468, 73, 39 },
		{ 468, 74, 39 },
		{ 468, 75, 39 },
		{ 468, 76, 39 },
		{ 468, 77, 39 },
		{ 468, 78, 39 },
		{ 468, 79, 39 },
		{ 468, 80, 39 },
		{ 468, 81, 39 },
		{ 468, 82, 39 },
		{ 468, 83, 39 },
		{ 468, 84, 39 },
		{ 468, 85, 39 },
		{ 468, 86, 39 },
		{ 468, 87, 39 },
		{ 468, 88, 39 },
		{ 468, 89, 39 },
		{ 468, 90, 39 },
		{ 468, 91, 39 },
		{ 468, 92, 39 },
		{ 468, 93, 39 },
		{ 468, 94, 39 },
		{ 468, 95, 39 },
		{ 468, 96, 39 },
		{ 468, 97, 39 },
		{ 468, 98, 39 },
		{ 469, 46, 39 },
		{ 469, 47, 39 },
		{ 469, 48, 39 },
		{ 469, 49, 39 },
		{ 469, 50, 39 },
		{ 469, 51, 39 },
		{ 469, 52, 39 },
		{ 469, 53, 39 },
		{ 469, 54, 39 },
		{ 469, 55, 39 },
		{ 469, 56, 39 },
		{ 469, 57, 39 },
		{ 469, 58, 39 },
		{ 469, 59, 39 },
		{ 469, 60, 39 },
		{ 469, 61, 39 },
		{ 469, 62, 39 },
		{ 469, 63, 39 },
		{ 469, 64, 39 },
		{ 469, 65, 39 },
		{ 469, 66, 39 },
		{ 469, 67, 39 },
		{ 469, 68, 39 },
		{ 469, 69, 39 },
		{ 469, 70, 39 },
		{ 469, 71, 39 },
		{ 469, 72, 39 },
		{ 469, 73, 39 },
		{ 469, 74, 39 },
		{ 469, 75, 39 },
		{ 469, 76, 39 },
		{ 469, 77, 39 },
		{ 469, 78, 39 },
		{ 469, 79, 39 },
		{ 469, 80, 39 },
		{ 469, 81, 39 },
		{ 469, 82, 39 },
		{ 469, 83, 39 },
		{ 469, 84, 39 },
		{ 469, 85, 39 },
		{ 469, 86, 39 },
		{ 469, 87, 39 },
		{ 469, 88, 39 },
		{ 469, 89, 39 },
		{ 469, 90, 39 },
		{ 469, 91, 39 },
		{ 469, 92, 39 },
		{ 469, 93, 39 },
		{ 469, 94, 39 },
		{ 469, 95, 39 },
		{ 469, 96, 39 },
		{ 469, 97, 39 },
		{ 470, 45, 39 },
		{ 470, 46, 39 },
		{ 470, 47, 39 },
		{ 470, 48, 39 },
		{ 470, 49, 39 },
		{ 470, 50, 39 },
		{ 470, 51, 39 },
		{ 470, 52, 39 },
		{ 470, 53, 39 },
		{ 470, 54, 39 },
		{ 470, 55, 39 },
		{ 470, 56, 39 },
		{ 470, 57, 39 },
		{ 470, 58, 39 },
		{ 470, 59, 39 },
		{ 470, 60, 39 },
		{ 470, 61, 39 },
		{ 470, 62, 39 },
		{ 470, 63, 39 },
		{ 470, 64, 39 },
		{ 470, 65, 39 },
		{ 470, 66, 39 },
		{ 470, 67, 39 },
		{ 470, 68, 39 },
		{ 470, 69, 39 },
		{ 470, 70, 39 },
		{ 470, 71, 39 },
		{ 470, 72, 39 },
		{ 470, 73, 39 },
		{ 470, 74, 39 },
		{ 470, 75, 39 },
		{ 470, 76, 39 },
		{ 470, 77, 39 },
		{ 470, 78, 39 },
		{ 470, 79, 39 },
		{ 470, 80, 39 },
		{ 470, 81, 39 },
		{ 470, 82, 39 },
		{ 470, 83, 39 },
		{ 470, 84, 39 },
		{ 470, 85, 39 },
		{ 470, 86, 39 },
		{ 470, 87, 39 },
		{ 470, 88, 39 },
		{ 470, 89, 39 },
		{ 470, 90, 39 },
		{ 470, 91, 39 },
		{ 470, 92, 39 },
		{ 470, 93, 39 },
		{ 470, 94, 39 },
		{ 470, 95, 39 },
		{ 470, 96, 39 },
		{ 470, 97, 39 },
		{ 471, 44, 39 },
		{ 471, 45, 39 },
		{ 471, 46, 39 },
		{ 471, 47, 39 },
		{ 471, 48, 39 },
		{ 471, 49, 39 },
		{ 471, 50, 39 },
		{ 471, 51, 39 },
		{ 471, 52, 39 },
		{ 471, 53, 39 },
		{ 471, 54, 39 },
		{ 471, 55, 39 },
		{ 471, 56, 39 },
		{ 471, 57, 39 },
		{ 471, 58, 39 },
		{ 471, 59, 39 },
		{ 471, 60, 39 },
		{ 471, 61, 39 },
		{ 471, 62, 39 },
		{ 471, 63, 39 },
		{ 471, 64, 39 },
		{ 471, 65, 39 },
		{ 471, 66, 39 },
		{ 471, 67, 39 },
		{ 471, 68, 39 },
		{ 471, 69, 39 },
		{ 471, 70, 39 },
		{ 471, 71, 39 },
		{ 471, 72, 39 },
		{ 471, 73, 39 },
		{ 471, 74, 39 },
		{ 471, 75, 39 },
		{ 471, 76, 39 },
		{ 471, 77, 39 },
		{ 471, 78, 39 },
		{ 471, 79, 39 },
		{ 471, 80, 39 },
		{ 471, 81, 39 },
		{ 471, 82, 39 },
		{ 471, 83, 39 },
		{ 471, 84, 39 },
		{ 471, 85, 39 },
		{ 471, 86, 39 },
		{ 471, 87, 39 },
		{ 471, 88, 39 },
		{ 471, 89, 39 },
		{ 471, 90, 39 },
		{ 471, 91, 39 },
		{ 471, 92, 39 },
		{ 471, 93, 39 },
		{ 471, 94, 39 },
		{ 471, 95, 39 },
		{ 471, 96, 39 },
		{ 472, 43, 39 },
		{ 472, 44, 39 },
		{ 472, 45, 39 },
		{ 472, 46, 39 },
		{ 472, 47, 39 },
		{ 472, 48, 39 },
		{ 472, 49, 39 },
		{ 472, 50, 39 },
		{ 472, 51, 39 },
		{ 472, 52, 39 },
		{ 472, 53, 39 },
		{ 472, 54, 39 },
		{ 472, 55, 39 },
		{ 472, 56, 39 },
		{ 472, 57, 39 },
		{ 472, 58, 39 },
		{ 472, 59, 39 },
		{ 472, 60, 39 },
		{ 472, 61, 39 },
		{ 472, 62, 39 },
		{ 472, 63, 39 },
		{ 472, 64, 39 },
		{ 472, 65, 39 },
		{ 472, 66, 39 },
		{ 472, 67, 39 },
		{ 472, 68, 39 },
		{ 472, 69, 39 },
		{ 472, 70, 39 },
		{ 472, 71, 39 },
		{ 472, 72, 39 },
		{ 472, 73, 39 },
		{ 472, 74, 39 },
		{ 472, 75, 39 },
		{ 472, 76, 39 },
		{ 472, 77, 39 },
		{ 472, 78, 39 },
		{ 472, 79, 39 },
		{ 472, 80, 39 },
		{ 472, 81, 39 },
		{ 472, 82, 39 },
		{ 472, 83, 39 },
		{ 472, 84, 39 },
		{ 472, 85, 39 },
		{ 472, 86, 39 },
		{ 472, 87, 39 },
		{ 472, 88, 39 },
		{ 472, 89, 39 },
		{ 472, 90, 39 },
		{ 472, 91, 39 },
		{ 472, 92, 39 },
		{ 472, 93, 39 },
		{ 472, 94, 39 },
		{ 472, 95, 39 },
		{ 473, 42, 39 },
		{ 473, 43, 39 },
		{ 473, 44, 39 },
		{ 473, 45, 39 },
		{ 473, 46, 39 },
		{ 473, 47, 39 },
		{ 473, 48, 39 },
		{ 473, 49, 39 },
		{ 473, 50, 39 },
		{ 473, 51, 39 },
		{ 473, 52, 39 },
		{ 473, 53, 39 },
		{ 473, 54, 39 },
		{ 473, 55, 39 },
		{ 473, 56, 39 },
		{ 473, 57, 39 },
		{ 473, 58, 39 },
		{ 473, 59, 39 },
		{ 473, 60, 39 },
		{ 473, 61, 39 },
		{ 473, 62, 39 },
		{ 473, 63, 39 },
		{ 473, 64, 39 },
		{ 473, 65, 39 },
		{ 473, 66, 39 },
		{ 473, 67, 39 },
		{ 473, 68, 39 },
		{ 473, 69, 39 },
		{ 473, 70, 39 },
		{ 473, 71, 39 },
		{ 473, 72, 39 },
		{ 473, 73, 39 },
		{ 473, 74, 39 },
		{ 473, 75, 39 },
		{ 473, 76, 39 },
		{ 473, 77, 39 },
		{ 473, 78, 39 },
		{ 473, 79, 39 },
		{ 473, 80, 39 },
		{ 473, 81, 39 },
		{ 473, 82, 39 },
		{ 473, 83, 39 },
		{ 473, 84, 39 },
		{ 473, 85, 39 },
		{ 473, 86, 39 },
		{ 473, 87, 39 },
		{ 473, 88, 39 },
		{ 473, 89, 39 },
		{ 473, 90, 39 },
		{ 473, 91, 39 },
		{ 473, 92, 39 },
		{ 473, 93, 39 },
		{ 473, 94, 39 },
		{ 474, 41, 39 },
		{ 474, 42, 39 },
		{ 474, 43, 39 },
		{ 474, 44, 39 },
		{ 474, 45, 39 },
		{ 474, 46, 39 },
		{ 474, 47, 39 },
		{ 474, 48, 39 },
		{ 474, 49, 39 },
		{ 474, 50, 39 },
		{ 474, 51, 39 },
		{ 474, 52, 39 },
		{ 474, 53, 39 },
		{ 474, 54, 39 },
		{ 474, 55, 39 },
		{ 474, 56, 39 },
		{ 474, 57, 39 },
		{ 474, 58, 39 },
		{ 474, 59, 39 },
		{ 474, 60, 39 },
		{ 474, 61, 39 },
		{ 474, 62, 39 },
		{ 474, 63, 39 },
		{ 474, 64, 39 },
		{ 474, 65, 39 },
		{ 474, 66, 39 },
		{ 474, 67, 39 },
		{ 474, 68, 39 },
		{ 474, 69, 39 },
		{ 474, 70, 39 },
		{ 474, 71, 39 },
		{ 474, 72, 39 },
		{ 474, 73, 39 },
		{ 474, 74, 39 },
		{ 474, 75, 39 },
		{ 474, 76, 39 },
		{ 474, 77, 39 },
		{ 474, 78, 39 },
		{ 474, 79, 39 },
		{ 474, 80, 39 },
		{ 474, 81, 39 },
		{ 474, 82, 39 },
		{ 474, 83, 39 },
		{ 474, 84, 39 },
		{ 474, 85, 39 },
		{ 474, 86, 39 },
		{ 474, 87, 39 },
		{ 474, 88, 39 },
		{ 474, 89, 39 },
		{ 474, 90, 39 },
		{ 474, 91, 39 },
		{ 474, 92, 39 },
		{ 474, 93, 39 },
		{ 474, 94, 39 },
		{ 475, 40, 39 },
		{ 475, 41, 39 },
		{ 475, 42, 39 },
		{ 475, 43, 39 },
		{ 475, 44, 39 },
		{ 475, 45, 39 },
		{ 475, 46, 39 },
		{ 475, 47, 39 },
		{ 475, 48, 39 },
		{ 475, 49, 39 },
		{ 475, 50, 39 },
		{ 475, 51, 39 },
		{ 475, 52, 39 },
		{ 475, 53, 39 },
		{ 475, 54, 39 },
		{ 475, 55, 39 },
		{ 475, 56, 39 },
		{ 475, 57, 39 },
		{ 475, 58, 39 },
		{ 475, 59, 39 },
		{ 475, 60, 39 },
		{ 475, 61, 39 },
		{ 475, 62, 39 },
		{ 475, 63, 39 },
		{ 475, 64, 39 },
		{ 475, 65, 39 },
		{ 475, 66, 39 },
		{ 475, 67, 39 },
		{ 475, 68, 39 },
		{ 475, 69, 39 },
		{ 475, 70, 39 },
		{ 475, 71, 39 },
		{ 475, 72, 39 },
		{ 475, 73, 39 },
		{ 475, 74, 39 },
		{ 475, 75, 39 },
		{ 475, 76, 39 },
		{ 475, 77, 39 },
		{ 475, 78, 39 },
		{ 475, 79, 39 },
		{ 475, 80, 39 },
		{ 475, 81, 39 },
		{ 475, 82, 39 },
		{ 475, 83, 39 },
		{ 475, 84, 39 },
		{ 475, 85, 39 },
		{ 475, 86, 39 },
		{ 475, 87, 39 },
		{ 475, 88, 39 },
		{ 475, 89, 39 },
		{ 475, 90, 39 },
		{ 475, 91, 39 },
		{ 475, 92, 39 },
		{ 475, 93, 39 },
		{ 476, 39, 39 },
		{ 476, 40, 39 },
		{ 476, 41, 39 },
		{ 476, 42, 39 },
		{ 476, 43, 39 },
		{ 476, 44, 39 },
		{ 476, 45, 39 },
		{ 476, 46, 39 },
		{ 476, 47, 39 },
		{ 476, 48, 39 },
		{ 476, 49, 39 },
		{ 476, 50, 39 },
		{ 476, 51, 39 },
		{ 476, 52, 39 },
		{ 476, 53, 39 },
		{ 476, 54, 39 },
		{ 476, 55, 39 },
		{ 476, 56, 39 },
		{ 476, 57, 39 },
		{ 476, 58, 39 },
		{ 476, 59, 39 },
		{ 476, 60, 39 },
		{ 476, 61, 39 },
		{ 476, 62, 39 },
		{ 476, 63, 39 },
		{ 476, 64, 39 },
		{ 476, 65, 39 },
		{ 476, 66, 39 },
		{ 476, 67, 39 },
		{ 476, 68, 39 },
		{ 476, 69, 39 },
		{ 476, 70, 39 },
		{ 476, 71, 39 },
		{ 476, 72, 39 },
		{ 476, 73, 39 },
		{ 476, 74, 39 },
		{ 476, 75, 39 },
		{ 476, 76, 39 },
		{ 476, 77, 39 },
		{ 476, 78, 39 },
		{ 476, 79, 39 },
		{ 476, 80, 39 },
		{ 476, 81, 39 },
		{ 476, 82, 39 },
		{ 476, 83, 39 },
		{ 476, 84, 39 },
		{ 476, 85, 39 },
		{ 476, 86, 39 },
		{ 476, 87, 39 },
		{ 476, 88, 39 },
		{ 476, 89, 39 },
		{ 476, 90, 39 },
		{ 476, 91, 39 },
		{ 476, 92, 39 },
		{ 477, 38, 39 },
		{ 477, 39, 39 },
		{ 477, 40, 39 },
		{ 477, 41, 39 },
		{ 477, 42, 39 },
		{ 477, 43, 39 },
		{ 477, 44, 39 },
		{ 477, 45, 39 },
		{ 477, 46, 39 },
		{ 477, 47, 39 },
		{ 477, 48, 39 },
		{ 477, 49, 39 },
		{ 477, 50, 39 },
		{ 477, 51, 39 },
		{ 477, 52, 39 },
		{ 477, 53, 39 },
		{ 477, 54, 39 },
		{ 477, 55, 39 },
		{ 477, 56, 39 },
		{ 477, 57, 39 },
		{ 477, 58, 39 },
		{ 477, 59, 39 },
		{ 477, 60, 39 },
		{ 477, 61, 39 },
		{ 477, 62, 39 },
		{ 477, 63, 39 },
		{ 477, 64, 39 },
		{ 477, 65, 39 },
		{ 477, 66, 39 },
		{ 477, 67, 39 },
		{ 477, 68, 39 },
		{ 477, 69, 39 },
		{ 477, 70, 39 },
		{ 477, 71, 39 },
		{ 477, 72, 39 },
		{ 477, 73, 39 },
		{ 477, 74, 39 },
		{ 477, 75, 39 },
		{ 477, 76, 39 },
		{ 477, 77, 39 },
		{ 477, 78, 39 },
		{ 477, 79, 39 },
		{ 477, 80, 39 },
		{ 477, 81, 39 },
		{ 477, 82, 39 },
		{ 477, 83, 39 },
		{ 477, 84, 39 },
		{ 477, 85, 39 },
		{ 477, 86, 39 },
		{ 477, 87, 39 },
		{ 477, 88, 39 },
		{ 477, 89, 39 },
		{ 477, 90, 39 },
		{ 477, 91, 39 },
		{ 478, 37, 39 },
		{ 478, 38, 39 },
		{ 478, 39, 39 },
		{ 478, 40, 39 },
		{ 478, 41, 39 },
		{ 478, 42, 39 },
		{ 478, 43, 39 },
		{ 478, 44, 39 },
		{ 478, 45, 39 },
		{ 478, 46, 39 },
		{ 478, 47, 39 },
		{ 478, 48, 39 },
		{ 478, 49, 39 },
		{ 478, 50, 39 },
		{ 478, 51, 39 },
		{ 478, 52, 39 },
		{ 478, 53, 39 },
		{ 478, 54, 39 },
		{ 478, 55, 39 },
		{ 478, 56, 39 },
		{ 478, 57, 39 },
		{ 478, 58, 39 },
		{ 478, 59, 39 },
		{ 478, 60, 39 },
		{ 478, 61, 39 },
		{ 478, 62, 39 },
		{ 478, 63, 39 },
		{ 478, 64, 39 },
		{ 478, 65, 39 },
		{ 478, 66, 39 },
		{ 478, 67, 39 },
		{ 478, 68, 39 },
		{ 478, 69, 39 },
		{ 478, 70, 39 },
		{ 478, 71, 39 },
		{ 478, 72, 39 },
		{ 478, 73, 39 },
		{ 478, 74, 39 },
		{ 478, 75, 39 },
		{ 478, 76, 39 },
		{ 478, 77, 39 },
		{ 478, 78, 39 },
		{ 478, 79, 39 },
		{ 478, 80, 39 },
		{ 478, 81, 39 },
		{ 478, 82, 39 },
		{ 478, 83, 39 },
		{ 478, 84, 39 },
		{ 478, 85, 39 },
		{ 478, 86, 39 },
		{ 478, 87, 39 },
		{ 478, 88, 39 },
		{ 478, 89, 39 },
		{ 478, 90, 39 },
		{ 478, 91, 39 },
		{ 479, 36, 39 },
		{ 479, 37, 39 },
		{ 479, 38, 39 },
		{ 479, 39, 39 },
		{ 479, 40, 39 },
		{ 479, 41, 39 },
		{ 479, 42, 39 },
		{ 479, 43, 39 },
		{ 479, 44, 39 },
		{ 479, 45, 39 },
		{ 479, 46, 39 },
		{ 479, 47, 39 },
		{ 479, 48, 39 },
		{ 479, 49, 39 },
		{ 479, 50, 39 },
		{ 479, 51, 39 },
		{ 479, 52, 39 },
		{ 479, 53, 39 },
		{ 479, 54, 39 },
		{ 479, 55, 39 },
		{ 479, 56, 39 },
		{ 479, 57, 39 },
		{ 479, 58, 39 },
		{ 479, 59, 39 },
		{ 479, 60, 39 },
		{ 479, 61, 39 },
		{ 479, 62, 39 },
		{ 479, 63, 39 },
		{ 479, 64, 39 },
		{ 479, 65, 39 },
		{ 479, 66, 39 },
		{ 479, 67, 39 },
		{ 479, 68, 39 },
		{ 479, 69, 39 },
		{ 479, 70, 39 },
		{ 479, 71, 39 },
		{ 479, 72, 39 },
		{ 479, 73, 39 },
		{ 479, 74, 39 },
		{ 479, 75, 39 },
		{ 479, 76, 39 },
		{ 479, 77, 39 },
		{ 479, 78, 39 },
		{ 479, 79, 39 },
		{ 479, 80, 39 },
		{ 479, 81, 39 },
		{ 479, 82, 39 },
		{ 479, 83, 39 },
		{ 479, 84, 39 },
		{ 479, 85, 39 },
		{ 479, 86, 39 },
		{ 479, 87, 39 },
		{ 479, 88, 39 },
		{ 479, 89, 39 },
		{ 479, 90, 39 },
		{ 480, 35, 39 },
		{ 480, 36, 39 },
		{ 480, 37, 39 },
		{ 480, 38, 39 },
		{ 480, 39, 39 },
		{ 480, 40, 39 },
		{ 480, 41, 39 },
		{ 480, 42, 39 },
		{ 480, 43, 39 },
		{ 480, 44, 39 },
		{ 480, 45, 39 },
		{ 480, 46, 39 },
		{ 480, 47, 39 },
		{ 480, 48, 39 },
		{ 480, 49, 39 },
		{ 480, 50, 39 },
		{ 480, 51, 39 },
		{ 480, 52, 39 },
		{ 480, 53, 39 },
		{ 480, 54, 39 },
		{ 480, 55, 39 },
		{ 480, 56, 39 },
		{ 480, 57, 39 },
		{ 480, 58, 39 },
		{ 480, 59, 39 },
		{ 480, 60, 39 },
		{ 480, 61, 39 },
		{ 480, 62, 39 },
		{ 480, 63, 39 },
		{ 480, 64, 39 },
		{ 480, 65, 39 },
		{ 480, 66, 39 },
		{ 480, 67, 39 },
		{ 480, 68, 39 },
		{ 480, 69, 39 },
		{ 480, 70, 39 },
		{ 480, 71, 39 },
		{ 480, 72, 39 },
		{ 480, 73, 39 },
		{ 480, 74, 39 },
		{ 480, 75, 39 },
		{ 480, 76, 39 },
		{ 480, 77, 39 },
		{ 480, 78, 39 },
		{ 480, 79, 39 },
		{ 480, 80, 39 },
		{ 480, 81, 39 },
		{ 480, 82, 39 },
		{ 480, 83, 39 },
		{ 480, 84, 39 },
		{ 480, 85, 39 },
		{ 480, 86, 39 },
		{ 480, 87, 39 },
		{ 480, 88, 39 },
		{ 480, 89, 39 },
		{ 481, 34, 39 },
		{ 481, 35, 39 },
		{ 481, 36, 39 },
		{ 481, 37, 39 },
		{ 481, 38, 39 },
		{ 481, 39, 39 },
		{ 481, 40, 39 },
		{ 481, 41, 39 },
		{ 481, 42, 39 },
		{ 481, 43, 39 },
		{ 481, 44, 39 },
		{ 481, 45, 39 },
		{ 481, 46, 39 },
		{ 481, 47, 39 },
		{ 481, 48, 39 },
		{ 481, 49, 39 },
		{ 481, 50, 39 },
		{ 481, 51, 39 },
		{ 481, 52, 39 },
		{ 481, 53, 39 },
		{ 481, 54, 39 },
		{ 481, 55, 39 },
		{ 481, 56, 39 },
		{ 481, 57, 39 },
		{ 481, 58, 39 },
		{ 481, 59, 39 },
		{ 481, 60, 39 },
		{ 481, 61, 39 },
		{ 481, 62, 39 },
		{ 481, 63, 39 },
		{ 481, 64, 39 },
		{ 481, 65, 39 },
		{ 481, 66, 39 },
		{ 481, 67, 39 },
		{ 481, 68, 39 },
		{ 481, 69, 39 },
		{ 481, 70, 39 },
		{ 481, 71, 39 },
		{ 481, 72, 39 },
		{ 481, 73, 39 },
		{ 481, 74, 39 },
		{ 481, 75, 39 },
		{ 481, 76, 39 },
		{ 481, 77, 39 },
		{ 481, 78, 39 },
		{ 481, 79, 39 },
		{ 481, 80, 39 },
		{ 481, 81, 39 },
		{ 481, 82, 39 },
		{ 481, 83, 39 },
		{ 481, 84, 39 },
		{ 481, 85, 39 },
		{ 481, 86, 39 },
		{ 481, 87, 39 },
		{ 481, 88, 39 },
		{ 482, 33, 39 },
		{ 482, 34, 39 },
		{ 482, 35, 39 },
		{ 482, 36, 39 },
		{ 482, 37, 39 },
		{ 482, 38, 39 },
		{ 482, 39, 39 },
		{ 482, 40, 39 },
		{ 482, 41, 39 },
		{ 482, 42, 39 },
		{ 482, 43, 39 },
		{ 482, 44, 39 },
		{ 482, 45, 39 },
		{ 482, 46, 39 },
		{ 482, 47, 39 },
		{ 482, 48, 39 },
		{ 482, 49, 39 },
		{ 482, 50, 39 },
		{ 482, 51, 39 },
		{ 482, 52, 39 },
		{ 482, 53, 39 },
		{ 482, 54, 39 },
		{ 482, 55, 39 },
		{ 482, 56, 39 },
		{ 482, 57, 39 },
		{ 482, 58, 39 },
		{ 482, 59, 39 },
		{ 482, 60, 39 },
		{ 482, 61, 39 },
		{ 482, 62, 39 },
		{ 482, 63, 39 },
		{ 482, 64, 39 },
		{ 482, 65, 39 },
		{ 482, 66, 39 },
		{ 482, 67, 39 },
		{ 482, 68, 39 },
		{ 482, 69, 39 },
		{ 482, 70, 39 },
		{ 482, 71, 39 },
		{ 482, 72, 39 },
		{ 482, 73, 39 },
		{ 482, 74, 39 },
		{ 482, 75, 39 },
		{ 482, 76, 39 },
		{ 482, 77, 39 },
		{ 482, 78, 39 },
		{ 482, 79, 39 },
		{ 482, 80, 39 },
		{ 482, 81, 39 },
		{ 482, 82, 39 },
		{ 482, 83, 39 },
		{ 482, 84, 39 },
		{ 482, 85, 39 },
		{ 482, 86, 39 },
		{ 482, 87, 39 },
		{ 482, 88, 39 },
		{ 483, 32, 39 },
		{ 483, 33, 39 },
		{ 483, 34, 39 },
		{ 483, 35, 39 },
		{ 483, 36, 39 },
		{ 483, 37, 39 },
		{ 483, 38, 39 },
		{ 483, 39, 39 },
		{ 483, 40, 39 },
		{ 483, 41, 39 },
		{ 483, 42, 39 },
		{ 483, 43, 39 },
		{ 483, 44, 39 },
		{ 483, 45, 39 },
		{ 483, 46, 39 },
		{ 483, 47, 39 },
		{ 483, 48, 39 },
		{ 483, 49, 39 },
		{ 483, 50, 39 },
		{ 483, 51, 39 },
		{ 483, 52, 39 },
		{ 483, 53, 39 },
		{ 483, 54, 39 },
		{ 483, 55, 39 },
		{ 483, 56, 39 },
		{ 483, 57, 39 },
		{ 483, 58, 39 },
		{ 483, 59, 39 },
		{ 483, 60, 39 },
		{ 483, 61, 39 },
		{ 483, 62, 39 },
		{ 483, 63, 39 },
		{ 483, 64, 39 },
		{ 483, 65, 39 },
		{ 483, 66, 39 },
		{ 483, 67, 39 },
		{ 483, 68, 39 },
		{ 483, 69, 39 },
		{ 483, 70, 39 },
		{ 483, 71, 39 },
		{ 483, 72, 39 },
		{ 483, 73, 39 },
		{ 483, 74, 39 },
		{ 483, 75, 39 },
		{ 483, 76, 39 },
		{ 483, 77, 39 },
		{ 483, 78, 39 },
		{ 483, 79, 39 },
		{ 483, 80, 39 },
		{ 483, 81, 39 },
		{ 483, 82, 39 },
		{ 483, 83, 39 },
		{ 483, 84, 39 },
		{ 483, 85, 39 },
		{ 483, 86, 39 },
		{ 483, 87, 39 },
		{ 484, 31, 39 },
		{ 484, 32, 39 },
		{ 484, 33, 39 },
		{ 484, 34, 39 },
		{ 484, 35, 39 },
		{ 484, 36, 39 },
		{ 484, 37, 39 },
		{ 484, 38, 39 },
		{ 484, 39, 39 },
		{ 484, 40, 39 },
		{ 484, 41, 39 },
		{ 484, 42, 39 },
		{ 484, 43, 39 },
		{ 484, 44, 39 },
		{ 484, 45, 39 },
		{ 484, 46, 39 },
		{ 484, 47, 39 },
		{ 484, 48, 39 },
		{ 484, 49, 39 },
		{ 484, 50, 39 },
		{ 484, 51, 39 },
		{ 484, 52, 39 },
		{ 484, 53, 39 },
		{ 484, 54, 39 },
		{ 484, 55, 39 },
		{ 484, 56, 39 },
		{ 484, 57, 39 },
		{ 484, 58, 39 },
		{ 484, 59, 39 },
		{ 484, 60, 39 },
		{ 484, 61, 39 },
		{ 484, 62, 39 },
		{ 484, 63, 39 },
		{ 484, 64, 39 },
		{ 484, 65, 39 },
		{ 484, 66, 39 },
		{ 484, 67, 39 },
		{ 484, 68, 39 },
		{ 484, 69, 39 },
		{ 484, 70, 39 },
		{ 484, 71, 39 },
		{ 484, 72, 39 },
		{ 484, 73, 39 },
		{ 484, 74, 39 },
		{ 484, 75, 39 },
		{ 484, 76, 39 },
		{ 484, 77, 39 },
		{ 484, 78, 39 },
		{ 484, 79, 39 },
		{ 484, 80, 39 },
		{ 484, 81, 39 },
		{ 484, 82, 39 },
		{ 484, 83, 39 },
		{ 484, 84, 39 },
		{ 484, 85, 39 },
		{ 484, 86, 39 },
		{ 485, 30, 39 },
		{ 485, 31, 39 },
		{ 485, 32, 39 },
		{ 485, 33, 39 },
		{ 485, 34, 39 },
		{ 485, 35, 39 },
		{ 485, 36, 39 },
		{ 485, 37, 39 },
		{ 485, 38, 39 },
		{ 485, 39, 39 },
		{ 485, 40, 39 },
		{ 485, 41, 39 },
		{ 485, 42, 39 },
		{ 485, 43, 39 },
		{ 485, 44, 39 },
		{ 485, 45, 39 },
		{ 485, 46, 39 },
		{ 485, 47, 39 },
		{ 485, 48, 39 },
		{ 485, 49, 39 },
		{ 485, 50, 39 },
		{ 485, 51, 39 },
		{ 485, 52, 39 },
		{ 485, 53, 39 },
		{ 485, 54, 39 },
		{ 485, 55, 39 },
		{ 485, 56, 39 },
		{ 485, 57, 39 },
		{ 485, 58, 39 },
		{ 485, 59, 39 },
		{ 485, 60, 39 },
		{ 485, 61, 39 },
		{ 485, 62, 39 },
		{ 485, 63, 39 },
		{ 485, 64, 39 },
		{ 485, 65, 39 },
		{ 485, 66, 39 },
		{ 485, 67, 39 },
		{ 485, 68, 39 },
		{ 485, 69, 39 },
		{ 485, 70, 39 },
		{ 485, 71, 39 },
		{ 485, 72, 39 },
		{ 485, 73, 39 },
		{ 485, 74, 39 },
		{ 485, 75, 39 },
		{ 485, 76, 39 },
		{ 485, 77, 39 },
		{ 485, 78, 39 },
		{ 485, 79, 39 },
		{ 485, 80, 39 },
		{ 485, 81, 39 },
		{ 485, 82, 39 },
		{ 485, 83, 39 },
		{ 485, 84, 39 },
		{ 485, 85, 39 },
		{ 486, 29, 39 },
		{ 486, 30, 39 },
		{ 486, 31, 39 },
		{ 486, 32, 39 },
		{ 486, 33, 39 },
		{ 486, 34, 39 },
		{ 486, 35, 39 },
		{ 486, 36, 39 },
		{ 486, 37, 39 },
		{ 486, 38, 39 },
		{ 486, 39, 39 },
		{ 486, 40, 39 },
		{ 486, 41, 39 },
		{ 486, 42, 39 },
		{ 486, 43, 39 },
		{ 486, 44, 39 },
		{ 486, 45, 39 },
		{ 486, 46, 39 },
		{ 486, 47, 39 },
		{ 486, 48, 39 },
		{ 486, 49, 39 },
		{ 486, 50, 39 },
		{ 486, 51, 39 },
		{ 486, 52, 39 },
		{ 486, 53, 39 },
		{ 486, 54, 39 },
		{ 486, 55, 39 },
		{ 486, 56, 39 },
		{ 486, 57, 39 },
		{ 486, 58, 39 },
		{ 486, 59, 39 },
		{ 486, 60, 39 },
		{ 486, 61, 39 },
		{ 486, 62, 39 },
		{ 486, 63, 39 },
		{ 486, 64, 39 },
		{ 486, 65, 39 },
		{ 486, 66, 39 },
		{ 486, 67, 39 },
		{ 486, 68, 39 },
		{ 486, 69, 39 },
		{ 486, 70, 39 },
		{ 486, 71, 39 },
		{ 486, 72, 39 },
		{ 486, 73, 39 },
		{ 486, 74, 39 },
		{ 486, 75, 39 },
		{ 486, 76, 39 },
		{ 486, 77, 39 },
		{ 486, 78, 39 },
		{ 486, 79, 39 },
		{ 486, 80, 39 },
		{ 486, 81, 39 },
		{ 486, 82, 39 },
		{ 486, 83, 39 },
		{ 486, 84, 39 },
		{ 486, 85, 39 },
		{ 487, 28, 39 },
		{ 487, 29, 39 },
		{ 487, 30, 39 },
		{ 487, 31, 39 },
		{ 487, 32, 39 },
		{ 487, 33, 39 },
		{ 487, 34, 39 },
		{ 487, 35, 39 },
		{ 487, 36, 39 },
		{ 487, 37, 39 },
		{ 487, 38, 39 },
		{ 487, 39, 39 },
		{ 487, 40, 39 },
		{ 487, 41, 39 },
		{ 487, 42, 39 },
		{ 487, 43, 39 },
		{ 487, 44, 39 },
		{ 487, 45, 39 },
		{ 487, 46, 39 },
		{ 487, 47, 39 },
		{ 487, 48, 39 },
		{ 487, 49, 39 },
		{ 487, 50, 39 },
		{ 487, 51, 39 },
		{ 487, 52, 39 },
		{ 487, 53, 39 },
		{ 487, 54, 39 },
		{ 487, 55, 39 },
		{ 487, 56, 39 },
		{ 487, 57, 39 },
		{ 487, 58, 39 },
		{ 487, 59, 39 },
		{ 487, 60, 39 },
		{ 487, 61, 39 },
		{ 487, 62, 39 },
		{ 487, 63, 39 },
		{ 487, 64, 39 },
		{ 487, 65, 39 },
		{ 487, 66, 39 },
		{ 487, 67, 39 },
		{ 487, 68, 39 },
		{ 487, 69, 39 },
		{ 487, 70, 39 },
		{ 487, 71, 39 },
		{ 487, 72, 39 },
		{ 487, 73, 39 },
		{ 487, 74, 39 },
		{ 487, 75, 39 },
		{ 487, 76, 39 },
		{ 487, 77, 39 },
		{ 487, 78, 39 },
		{ 487, 79, 39 },
		{ 487, 80, 39 },
		{ 487, 81, 39 },
		{ 487, 82, 39 },
		{ 487, 83, 39 },
		{ 487, 84, 39 },
		{ 488, 27, 39 },
		{ 488, 28, 39 },
		{ 488, 29, 39 },
		{ 488, 30, 39 },
		{ 488, 31, 39 },
		{ 488, 32, 39 },
		{ 488, 33, 39 },
		{ 488, 34, 39 },
		{ 488, 35, 39 },
		{ 488, 36, 39 },
		{ 488, 37, 39 },
		{ 488, 38, 39 },
		{ 488, 39, 39 },
		{ 488, 40, 39 },
		{ 488, 41, 39 },
		{ 488, 42, 39 },
		{ 488, 43, 39 },
		{ 488, 44, 39 },
		{ 488, 45, 39 },
		{ 488, 46, 39 },
		{ 488, 47, 39 },
		{ 488, 48, 39 },
		{ 488, 49, 39 },
		{ 488, 50, 39 },
		{ 488, 51, 39 },
		{ 488, 52, 39 },
		{ 488, 53, 39 },
		{ 488, 54, 39 },
		{ 488, 55, 39 },
		{ 488, 56, 39 },
		{ 488, 57, 39 },
		{ 488, 58, 39 },
		{ 488, 59, 39 },
		{ 488, 60, 39 },
		{ 488, 61, 39 },
		{ 488, 62, 39 },
		{ 488, 63, 39 },
		{ 488, 64, 39 },
		{ 488, 65, 39 },
		{ 488, 66, 39 },
		{ 488, 67, 39 },
		{ 488, 68, 39 },
		{ 488, 69, 39 },
		{ 488, 70, 39 },
		{ 488, 71, 39 },
		{ 488, 72, 39 },
		{ 488, 73, 39 },
		{ 488, 74, 39 },
		{ 488, 75, 39 },
		{ 488, 76, 39 },
		{ 488, 77, 39 },
		{ 488, 78, 39 },
		{ 488, 79, 39 },
		{ 488, 80, 39 },
		{ 488, 81, 39 },
		{ 488, 82, 39 },
		{ 488, 83, 39 },
		{ 489, 26, 39 },
		{ 489, 27, 39 },
		{ 489, 28, 39 },
		{ 489, 29, 39 },
		{ 489, 30, 39 },
		{ 489, 31, 39 },
		{ 489, 32, 39 },
		{ 489, 33, 39 },
		{ 489, 34, 39 },
		{ 489, 35, 39 },
		{ 489, 36, 39 },
		{ 489, 37, 39 },
		{ 489, 38, 39 },
		{ 489, 39, 39 },
		{ 489, 40, 39 },
		{ 489, 41, 39 },
		{ 489, 42, 39 },
		{ 489, 43, 39 },
		{ 489, 44, 39 },
		{ 489, 45, 39 },
		{ 489, 46, 39 },
		{ 489, 47, 39 },
		{ 489, 48, 39 },
		{ 489, 49, 39 },
		{ 489, 50, 39 },
		{ 489, 51, 39 },
		{ 489, 52, 39 },
		{ 489, 53, 39 },
		{ 489, 54, 39 },
		{ 489, 55, 39 },
		{ 489, 56, 39 },
		{ 489, 57, 39 },
		{ 489, 58, 39 },
		{ 489, 59, 39 },
		{ 489, 60, 39 },
		{ 489, 61, 39 },
		{ 489, 62, 39 },
		{ 489, 63, 39 },
		{ 489, 64, 39 },
		{ 489, 65, 39 },
		{ 489, 66, 39 },
		{ 489, 67, 39 },
		{ 489, 68, 39 },
		{ 489, 69, 39 },
		{ 489, 70, 39 },
		{ 489, 71, 39 },
		{ 489, 72, 39 },
		{ 489, 73, 39 },
		{ 489, 74, 39 },
		{ 489, 75, 39 },
		{ 489, 76, 39 },
		{ 489, 77, 39 },
		{ 489, 78, 39 },
		{ 489, 79, 39 },
		{ 489, 80, 39 },
		{ 489, 81, 39 },
		{ 489, 82, 39 },
		{ 490, 25, 39 },
		{ 490, 26, 39 },
		{ 490, 27, 39 },
		{ 490, 28, 39 },
		{ 490, 29, 39 },
		{ 490, 30, 39 },
		{ 490, 31, 39 },
		{ 490, 32, 39 },
		{ 490, 33, 39 },
		{ 490, 34, 39 },
		{ 490, 35, 39 },
		{ 490, 36, 39 },
		{ 490, 37, 39 },
		{ 490, 38, 39 },
		{ 490, 39, 39 },
		{ 490, 40, 39 },
		{ 490, 41, 39 },
		{ 490, 42, 39 },
		{ 490, 43, 39 },
		{ 490, 44, 39 },
		{ 490, 45, 39 },
		{ 490, 46, 39 },
		{ 490, 47, 39 },
		{ 490, 48, 39 },
		{ 490, 49, 39 },
		{ 490, 50, 39 },
		{ 490, 51, 39 },
		{ 490, 52, 39 },
		{ 490, 53, 39 },
		{ 490, 54, 39 },
		{ 490, 55, 39 },
		{ 490, 56, 39 },
		{ 490, 57, 39 },
		{ 490, 58, 39 },
		{ 490, 59, 39 },
		{ 490, 60, 39 },
		{ 490, 61, 39 },
		{ 490, 62, 39 },
		{ 490, 63, 39 },
		{ 490, 64, 39 },
		{ 490, 65, 39 },
		{ 490, 66, 39 },
		{ 490, 67, 39 },
		{ 490, 68, 39 },
		{ 490, 69, 39 },
		{ 490, 70, 39 },
		{ 490, 71, 39 },
		{ 490, 72, 39 },
		{ 490, 73, 39 },
		{ 490, 74, 39 },
		{ 490, 75, 39 },
		{ 490, 76, 39 },
		{ 490, 77, 39 },
		{ 490, 78, 39 },
		{ 490, 79, 39 },
		{ 490, 80, 39 },
		{ 490, 81, 39 },
		{ 490, 82, 39 },
		{ 491, 24, 39 },
		{ 491, 25, 39 },
		{ 491, 26, 39 },
		{ 491, 27, 39 },
		{ 491, 28, 39 },
		{ 491, 29, 39 },
		{ 491, 30, 39 },
		{ 491, 31, 39 },
		{ 491, 32, 39 },
		{ 491, 33, 39 },
		{ 491, 34, 39 },
		{ 491, 35, 39 },
		{ 491, 36, 39 },
		{ 491, 37, 39 },
		{ 491, 38, 39 },
		{ 491, 39, 39 },
		{ 491, 40, 39 },
		{ 491, 41, 39 },
		{ 491, 42, 39 },
		{ 491, 43, 39 },
		{ 491, 44, 39 },
		{ 491, 45, 39 },
		{ 491, 46, 39 },
		{ 491, 47, 39 },
		{ 491, 48, 39 },
		{ 491, 49, 39 },
		{ 491, 50, 39 },
		{ 491, 51, 39 },
		{ 491, 52, 39 },
		{ 491, 53, 39 },
		{ 491, 54, 39 },
		{ 491, 55, 39 },
		{ 491, 56, 39 },
		{ 491, 57, 39 },
		{ 491, 58, 39 },
		{ 491, 59, 39 },
		{ 491, 60, 39 },
		{ 491, 61, 39 },
		{ 491, 62, 39 },
		{ 491, 63, 39 },
		{ 491, 64, 39 },
		{ 491, 65, 39 },
		{ 491, 66, 39 },
		{ 491, 67, 39 },
		{ 491, 68, 39 },
		{ 491, 69, 39 },
		{ 491, 70, 39 },
		{ 491, 71, 39 },
		{ 491, 72, 39 },
		{ 491, 73, 39 },
		{ 491, 74, 39 },
		{ 491, 75, 39 },
		{ 491, 76, 39 },
		{ 491, 77, 39 },
		{ 491, 78, 39 },
		{ 491, 79, 39 },
		{ 491, 80, 39 },
		{ 491, 81, 39 },
		{ 492, 23, 39 },
		{ 492, 24, 39 },
		{ 492, 25, 39 },
		{ 492, 26, 39 },
		{ 492, 27, 39 },
		{ 492, 28, 39 },
		{ 492, 29, 39 },
		{ 492, 30, 39 },
		{ 492, 31, 39 },
		{ 492, 32, 39 },
		{ 492, 33, 39 },
		{ 492, 34, 39 },
		{ 492, 35, 39 },
		{ 492, 36, 39 },
		{ 492, 37, 39 },
		{ 492, 38, 39 },
		{ 492, 39, 39 },
		{ 492, 40, 39 },
		{ 492, 41, 39 },
		{ 492, 42, 39 },
		{ 492, 43, 39 },
		{ 492, 44, 39 },
		{ 492, 45, 39 },
		{ 492, 46, 39 },
		{ 492, 47, 39 },
		{ 492, 48, 39 },
		{ 492, 49, 39 },
		{ 492, 50, 39 },
		{ 492, 51, 39 },
		{ 492, 52, 39 },
		{ 492, 53, 39 },
		{ 492, 54, 39 },
		{ 492, 55, 39 },
		{ 492, 56, 39 },
		{ 492, 57, 39 },
		{ 492, 58, 39 },
		{ 492, 59, 39 },
		{ 492, 60, 39 },
		{ 492, 61, 39 },
		{ 492, 62, 39 },
		{ 492, 63, 39 },
		{ 492, 64, 39 },
		{ 492, 65, 39 },
		{ 492, 66, 39 },
		{ 492, 67, 39 },
		{ 492, 68, 39 },
		{ 492, 69, 39 },
		{ 492, 70, 39 },
		{ 492, 71, 39 },
		{ 492, 72, 39 },
		{ 492, 73, 39 },
		{ 492, 74, 39 },
		{ 492, 75, 39 },
		{ 492, 76, 39 },
		{ 492, 77, 39 },
		{ 492, 78, 39 },
		{ 492, 79, 39 },
		{ 492, 80, 39 },
		{ 493, 22, 39 },
		{ 493, 23, 39 },
		{ 493, 24, 39 },
		{ 493, 25, 39 },
		{ 493, 26, 39 },
		{ 493, 27, 39 },
		{ 493, 28, 39 },
		{ 493, 29, 39 },
		{ 493, 30, 39 },
		{ 493, 31, 39 },
		{ 493, 32, 39 },
		{ 493, 33, 39 },
		{ 493, 34, 39 },
		{ 493, 35, 39 },
		{ 493, 36, 39 },
		{ 493, 37, 39 },
		{ 493, 38, 39 },
		{ 493, 39, 39 },
		{ 493, 40, 39 },
		{ 493, 41, 39 },
		{ 493, 42, 39 },
		{ 493, 43, 39 },
		{ 493, 44, 39 },
		{ 493, 45, 39 },
		{ 493, 46, 39 },
		{ 493, 47, 39 },
		{ 493, 48, 39 },
		{ 493, 49, 39 },
		{ 493, 50, 39 },
		{ 493, 51, 39 },
		{ 493, 52, 39 },
		{ 493, 53, 39 },
		{ 493, 54, 39 },
		{ 493, 55, 39 },
		{ 493, 56, 39 },
		{ 493, 57, 39 },
		{ 493, 58, 39 },
		{ 493, 59, 39 },
		{ 493, 60, 39 },
		{ 493, 61, 39 },
		{ 493, 62, 39 },
		{ 493, 63, 39 },
		{ 493, 64, 39 },
		{ 493, 65, 39 },
		{ 493, 66, 39 },
		{ 493, 67, 39 },
		{ 493, 68, 39 },
		{ 493, 69, 39 },
		{ 493, 70, 39 },
		{ 493, 71, 39 },
		{ 493, 72, 39 },
		{ 493, 73, 39 },
		{ 493, 74, 39 },
		{ 493, 75, 39 },
		{ 493, 76, 39 },
		{ 493, 77, 39 },
		{ 493, 78, 39 },
		{ 493, 79, 39 },
		{ 494, 21, 39 },
		{ 494, 22, 39 },
		{ 494, 23, 39 },
		{ 494, 24, 39 },
		{ 494, 25, 39 },
		{ 494, 26, 39 },
		{ 494, 27, 39 },
		{ 494, 28, 39 },
		{ 494, 29, 39 },
		{ 494, 30, 39 },
		{ 494, 31, 39 },
		{ 494, 32, 39 },
		{ 494, 33, 39 },
		{ 494, 34, 39 },
		{ 494, 35, 39 },
		{ 494, 36, 39 },
		{ 494, 37, 39 },
		{ 494, 38, 39 },
		{ 494, 39, 39 },
		{ 494, 40, 39 },
		{ 494, 41, 39 },
		{ 494, 42, 39 },
		{ 494, 43, 39 },
		{ 494, 44, 39 },
		{ 494, 45, 39 },
		{ 494, 46, 39 },
		{ 494, 47, 39 },
		{ 494, 48, 39 },
		{ 494, 49, 39 },
		{ 494, 50, 39 },
		{ 494, 51, 39 },
		{ 494, 52, 39 },
		{ 494, 53, 39 },
		{ 494, 54, 39 },
		{ 494, 55, 39 },
		{ 494, 56, 39 },
		{ 494, 57, 39 },
		{ 494, 58, 39 },
		{ 494, 59, 39 },
		{ 494, 60, 39 },
		{ 494, 61, 39 },
		{ 494, 62, 39 },
		{ 494, 63, 39 },
		{ 494, 64, 39 },
		{ 494, 65, 39 },
		{ 494, 66, 39 },
		{ 494, 67, 39 },
		{ 494, 68, 39 },
		{ 494, 69, 39 },
		{ 494, 70, 39 },
		{ 494, 71, 39 },
		{ 494, 72, 39 },
		{ 494, 73, 39 },
		{ 494, 74, 39 },
		{ 494, 75, 39 },
		{ 494, 76, 39 },
		{ 494, 77, 39 },
		{ 494, 78, 39 },
		{ 494, 79, 39 },
		{ 495, 20, 39 },
		{ 495, 21, 39 },
		{ 495, 22, 39 },
		{ 495, 23, 39 },
		{ 495, 24, 39 },
		{ 495, 25, 39 },
		{ 495, 26, 39 },
		{ 495, 27, 39 },
		{ 495, 28, 39 },
		{ 495, 29, 39 },
		{ 495, 30, 39 },
		{ 495, 31, 39 },
		{ 495, 32, 39 },
		{ 495, 33, 39 },
		{ 495, 34, 39 },
		{ 495, 35, 39 },
		{ 495, 36, 39 },
		{ 495, 37, 39 },
		{ 495, 38, 39 },
		{ 495, 39, 39 },
		{ 495, 40, 39 },
		{ 495, 41, 39 },
		{ 495, 42, 39 },
		{ 495, 43, 39 },
		{ 495, 44, 39 },
		{ 495, 45, 39 },
		{ 495, 46, 39 },
		{ 495, 47, 39 },
		{ 495, 48, 39 },
		{ 495, 49, 39 },
		{ 495, 50, 39 },
		{ 495, 51, 39 },
		{ 495, 52, 39 },
		{ 495, 53, 39 },
		{ 495, 54, 39 },
		{ 495, 55, 39 },
		{ 495, 56, 39 },
		{ 495, 57, 39 },
		{ 495, 58, 39 },
		{ 495, 59, 39 },
		{ 495, 60, 39 },
		{ 495, 61, 39 },
		{ 495, 62, 39 },
		{ 495, 63, 39 },
		{ 495, 64, 39 },
		{ 495, 65, 39 },
		{ 495, 66, 39 },
		{ 495, 67, 39 },
		{ 495, 68, 39 },
		{ 495, 69, 39 },
		{ 495, 70, 39 },
		{ 495, 71, 39 },
		{ 495, 72, 39 },
		{ 495, 73, 39 },
		{ 495, 74, 39 },
		{ 495, 75, 39 },
		{ 495, 76, 39 },
		{ 495, 77, 39 },
		{ 495, 78, 39 },
		{ 496, 19, 39 },
		{ 496, 20, 39 },
		{ 496, 21, 39 },
		{ 496, 22, 39 },
		{ 496, 23, 39 },
		{ 496, 24, 39 },
		{ 496, 25, 39 },
		{ 496, 26, 39 },
		{ 496, 27, 39 },
		{ 496, 28, 39 },
		{ 496, 29, 39 },
		{ 496, 30, 39 },
		{ 496, 31, 39 },
		{ 496, 32, 39 },
		{ 496, 33, 39 },
		{ 496, 34, 39 },
		{ 496, 35, 39 },
		{ 496, 36, 39 },
		{ 496, 37, 39 },
		{ 496, 38, 39 },
		{ 496, 39, 39 },
		{ 496, 40, 39 },
		{ 496, 41, 39 },
		{ 496, 42, 39 },
		{ 496, 43, 39 },
		{ 496, 44, 39 },
		{ 496, 45, 39 },
		{ 496, 46, 39 },
		{ 496, 47, 39 },
		{ 496, 48, 39 },
		{ 496, 49, 39 },
		{ 496, 50, 39 },
		{ 496, 51, 39 },
		{ 496, 52, 39 },
		{ 496, 53, 39 },
		{ 496, 54, 39 },
		{ 496, 55, 39 },
		{ 496, 56, 39 },
		{ 496, 57, 39 },
		{ 496, 58, 39 },
		{ 496, 59, 39 },
		{ 496, 60, 39 },
		{ 496, 61, 39 },
		{ 496, 62, 39 },
		{ 496, 63, 39 },
		{ 496, 64, 39 },
		{ 496, 65, 39 },
		{ 496, 66, 39 },
		{ 496, 67, 39 },
		{ 496, 68, 39 },
		{ 496, 69, 39 },
		{ 496, 70, 39 },
		{ 496, 71, 39 },
		{ 496, 72, 39 },
		{ 496, 73, 39 },
		{ 496, 74, 39 },
		{ 496, 75, 39 },
		{ 496, 76, 39 },
		{ 496, 77, 39 },
		{ 497, 18, 39 },
		{ 497, 19, 39 },
		{ 497, 20, 39 },
		{ 497, 21, 39 },
		{ 497, 22, 39 },
		{ 497, 23, 39 },
		{ 497, 24, 39 },
		{ 497, 25, 39 },
		{ 497, 26, 39 },
		{ 497, 27, 39 },
		{ 497, 28, 39 },
		{ 497, 29, 39 },
		{ 497, 30, 39 },
		{ 497, 31, 39 },
		{ 497, 32, 39 },
		{ 497, 33, 39 },
		{ 497, 34, 39 },
		{ 497, 35, 39 },
		{ 497, 36, 39 },
		{ 497, 37, 39 },
		{ 497, 38, 39 },
		{ 497, 39, 39 },
		{ 497, 40, 39 },
		{ 497, 41, 39 },
		{ 497, 42, 39 },
		{ 497, 43, 39 },
		{ 497, 44, 39 },
		{ 497, 45, 39 },
		{ 497, 46, 39 },
		{ 497, 47, 39 },
		{ 497, 48, 39 },
		{ 497, 49, 39 },
		{ 497, 50, 39 },
		{ 497, 51, 39 },
		{ 497, 52, 39 },
		{ 497, 53, 39 },
		{ 497, 54, 39 },
		{ 497, 55, 39 },
		{ 497, 56, 39 },
		{ 497, 57, 39 },
		{ 497, 58, 39 },
		{ 497, 59, 39 },
		{ 497, 60, 39 },
		{ 497, 61, 39 },
		{ 497, 62, 39 },
		{ 497, 63, 39 },
		{ 497, 64, 39 },
		{ 497, 65, 39 },
		{ 497, 66, 39 },
		{ 497, 67, 39 },
		{ 497, 68, 39 },
		{ 497, 69, 39 },
		{ 497, 70, 39 },
		{ 497, 71, 39 },
		{ 497, 72, 39 },
		{ 497, 73, 39 },
		{ 497, 74, 39 },
		{ 497, 75, 39 },
		{ 497, 76, 39 },
		{ 498, 17, 39 },
		{ 498, 18, 39 },
		{ 498, 19, 39 },
		{ 498, 20, 39 },
		{ 498, 21, 39 },
		{ 498, 22, 39 },
		{ 498, 23, 39 },
		{ 498, 24, 39 },
		{ 498, 25, 39 },
		{ 498, 26, 39 },
		{ 498, 27, 39 },
		{ 498, 28, 39 },
		{ 498, 29, 39 },
		{ 498, 30, 39 },
		{ 498, 31, 39 },
		{ 498, 32, 39 },
		{ 498, 33, 39 },
		{ 498, 34, 39 },
		{ 498, 35, 39 },
		{ 498, 36, 39 },
		{ 498, 37, 39 },
		{ 498, 38, 39 },
		{ 498, 39, 39 },
		{ 498, 40, 39 },
		{ 498, 41, 39 },
		{ 498, 42, 39 },
		{ 498, 43, 39 },
		{ 498, 44, 39 },
		{ 498, 45, 39 },
		{ 498, 46, 39 },
		{ 498, 47, 39 },
		{ 498, 48, 39 },
		{ 498, 49, 39 },
		{ 498, 50, 39 },
		{ 498, 51, 39 },
		{ 498, 52, 39 },
		{ 498, 53, 39 },
		{ 498, 54, 39 },
		{ 498, 55, 39 },
		{ 498, 56, 39 },
		{ 498, 57, 39 },
		{ 498, 58, 39 },
		{ 498, 59, 39 },
		{ 498, 60, 39 },
		{ 498, 61, 39 },
		{ 498, 62, 39 },
		{ 498, 63, 39 },
		{ 498, 64, 39 },
		{ 498, 65, 39 },
		{ 498, 66, 39 },
		{ 498, 67, 39 },
		{ 498, 68, 39 },
		{ 498, 69, 39 },
		{ 498, 70, 39 },
		{ 498, 71, 39 },
		{ 498, 72, 39 },
		{ 498, 73, 39 },
		{ 498, 74, 39 },
		{ 498, 75, 39 },
		{ 498, 76, 39 },
		{ 499, 16, 39 },
		{ 499, 17, 39 },
		{ 499, 18, 39 },
		{ 499, 19, 39 },
		{ 499, 20, 39 },
		{ 499, 21, 39 },
		{ 499, 22, 39 },
		{ 499, 23, 39 },
		{ 499, 24, 39 },
		{ 499, 25, 39 },
		{ 499, 26, 39 },
		{ 499, 27, 39 },
		{ 499, 28, 39 },
		{ 499, 29, 39 },
		{ 499, 30, 39 },
		{ 499, 31, 39 },
		{ 499, 32, 39 },
		{ 499, 33, 39 },
		{ 499, 34, 39 },
		{ 499, 35, 39 },
		{ 499, 36, 39 },
		{ 499, 37, 39 },
		{ 499, 38, 39 },
		{ 499, 39, 39 },
		{ 499, 40, 39 },
		{ 499, 41, 39 },
		{ 499, 42, 39 },
		{ 499, 43, 39 },
		{ 499, 44, 39 },
		{ 499, 45, 39 },
		{ 499, 46, 39 },
		{ 499, 47, 39 },
		{ 499, 48, 39 },
		{ 499, 49, 39 },
		{ 499, 50, 39 },
		{ 499, 51, 39 },
		{ 499, 52, 39 },
		{ 499, 53, 39 },
		{ 499, 54, 39 },
		{ 499, 55, 39 },
		{ 499, 56, 39 },
		{ 499, 57, 39 },
		{ 499, 58, 39 },
		{ 499, 59, 39 },
		{ 499, 60, 39 },
		{ 499, 61, 39 },
		{ 499, 62, 39 },
		{ 499, 63, 39 },
		{ 499, 64, 39 },
		{ 499, 65, 39 },
		{ 499, 66, 39 },
		{ 499, 67, 39 },
		{ 499, 68, 39 },
		{ 499, 69, 39 },
		{ 499, 70, 39 },
		{ 499, 71, 39 },
		{ 499, 72, 39 },
		{ 499, 73, 39 },
		{ 499, 74, 39 },
		{ 499, 75, 39 },
		{ 500, 15, 39 },
		{ 500, 16, 39 },
		{ 500, 17, 39 },
		{ 500, 18, 39 },
		{ 500, 19, 39 },
		{ 500, 20, 39 },
		{ 500, 21, 39 },
		{ 500, 22, 39 },
		{ 500, 23, 39 },
		{ 500, 24, 39 },
		{ 500, 25, 39 },
		{ 500, 26, 39 },
		{ 500, 27, 39 },
		{ 500, 28, 39 },
		{ 500, 29, 39 },
		{ 500, 30, 39 },
		{ 500, 31, 39 },
		{ 500, 32, 39 },
		{ 500, 33, 39 },
		{ 500, 34, 39 },
		{ 500, 35, 39 },
		{ 500, 36, 39 },
		{ 500, 37, 39 },
		{ 500, 38, 39 },
		{ 500, 39, 39 },
		{ 500, 40, 39 },
		{ 500, 41, 39 },
		{ 500, 42, 39 },
		{ 500, 43, 39 },
		{ 500, 44, 39 },
		{ 500, 45, 39 },
		{ 500, 46, 39 },
		{ 500, 47, 39 },
		{ 500, 48, 39 },
		{ 500, 49, 39 },
		{ 500, 50, 39 },
		{ 500, 51, 39 },
		{ 500, 52, 39 },
		{ 500, 53, 39 },
		{ 500, 54, 39 },
		{ 500, 55, 39 },
		{ 500, 56, 39 },
		{ 500, 57, 39 },
		{ 500, 58, 39 },
		{ 500, 59, 39 },
		{ 500, 60, 39 },
		{ 500, 61, 39 },
		{ 500, 62, 39 },
		{ 500, 63, 39 },
		{ 500, 64, 39 },
		{ 500, 65, 39 },
		{ 500, 66, 39 },
		{ 500, 67, 39 },
		{ 500, 68, 39 },
		{ 500, 69, 39 },
		{ 500, 70, 39 },
		{ 500, 71, 39 },
		{ 500, 72, 39 },
		{ 500, 73, 39 },
		{ 500, 74, 39 },
		{ 501, 14, 39 },
		{ 501, 15, 39 },
		{ 501, 16, 39 },
		{ 501, 17, 39 },
		{ 501, 18, 39 },
		{ 501, 19, 39 },
		{ 501, 20, 39 },
		{ 501, 21, 39 },
		{ 501, 22, 39 },
		{ 501, 23, 39 },
		{ 501, 24, 39 },
		{ 501, 25, 39 },
		{ 501, 26, 39 },
		{ 501, 27, 39 },
		{ 501, 28, 39 },
		{ 501, 29, 39 },
		{ 501, 30, 39 },
		{ 501, 31, 39 },
		{ 501, 32, 39 },
		{ 501, 33, 39 },
		{ 501, 34, 39 },
		{ 501, 35, 39 },
		{ 501, 36, 39 },
		{ 501, 37, 39 },
		{ 501, 38, 39 },
		{ 501, 39, 39 },
		{ 501, 40, 39 },
		{ 501, 41, 39 },
		{ 501, 42, 39 },
		{ 501, 43, 39 },
		{ 501, 44, 39 },
		{ 501, 45, 39 },
		{ 501, 46, 39 },
		{ 501, 47, 39 },
		{ 501, 48, 39 },
		{ 501, 49, 39 },
		{ 501, 50, 39 },
		{ 501, 51, 39 },
		{ 501, 52, 39 },
		{ 501, 53, 39 },
		{ 501, 54, 39 },
		{ 501, 55, 39 },
		{ 501, 56, 39 },
		{ 501, 57, 39 },
		{ 501, 58, 39 },
		{ 501, 59, 39 },
		{ 501, 60, 39 },
		{ 501, 61, 39 },
		{ 501, 62, 39 },
		{ 501, 63, 39 },
		{ 501, 64, 39 },
		{ 501, 65, 39 },
		{ 501, 66, 39 },
		{ 501, 67, 39 },
		{ 501, 68, 39 },
		{ 501, 69, 39 },
		{ 501, 70, 39 },
		{ 501, 71, 39 },
		{ 501, 72, 39 },
		{ 501, 73, 39 },
		{ 502, 13, 39 },
		{ 502, 14, 39 },
		{ 502, 15, 39 },
		{ 502, 16, 39 },
		{ 502, 17, 39 },
		{ 502, 18, 39 },
		{ 502, 19, 39 },
		{ 502, 20, 39 },
		{ 502, 21, 39 },
		{ 502, 22, 39 },
		{ 502, 23, 39 },
		{ 502, 24, 39 },
		{ 502, 25, 39 },
		{ 502, 26, 39 },
		{ 502, 27, 39 },
		{ 502, 28, 39 },
		{ 502, 29, 39 },
		{ 502, 30, 39 },
		{ 502, 31, 39 },
		{ 502, 32, 39 },
		{ 502, 33, 39 },
		{ 502, 34, 39 },
		{ 502, 35, 39 },
		{ 502, 36, 39 },
		{ 502, 37, 39 },
		{ 502, 38, 39 },
		{ 502, 39, 39 },
		{ 502, 40, 39 },
		{ 502, 41, 39 },
		{ 502, 42, 39 },
		{ 502, 43, 39 },
		{ 502, 44, 39 },
		{ 502, 45, 39 },
		{ 502, 46, 39 },
		{ 502, 47, 39 },
		{ 502, 48, 39 },
		{ 502, 49, 39 },
		{ 502, 50, 39 },
		{ 502, 51, 39 },
		{ 502, 52, 39 },
		{ 502, 53, 39 },
		{ 502, 54, 39 },
		{ 502, 55, 39 },
		{ 502, 56, 39 },
		{ 502, 57, 39 },
		{ 502, 58, 39 },
		{ 502, 59, 39 },
		{ 502, 60, 39 },
		{ 502, 61, 39 },
		{ 502, 62, 39 },
		{ 502, 63, 39 },
		{ 502, 64, 39 },
		{ 502, 65, 39 },
		{ 502, 66, 39 },
		{ 502, 67, 39 },
		{ 502, 68, 39 },
		{ 502, 69, 39 },
		{ 502, 70, 39 },
		{ 502, 71, 39 },
		{ 502, 72, 39 },
		{ 502, 73, 39 },
		{ 503, 12, 39 },
		{ 503, 13, 39 },
		{ 503, 14, 39 },
		{ 503, 15, 39 },
		{ 503, 16, 39 },
		{ 503, 17, 39 },
		{ 503, 18, 39 },
		{ 503, 19, 39 },
		{ 503, 20, 39 },
		{ 503, 21, 39 },
		{ 503, 22, 39 },
		{ 503, 23, 39 },
		{ 503, 24, 39 },
		{ 503, 25, 39 },
		{ 503, 26, 39 },
		{ 503, 27, 39 },
		{ 503, 28, 39 },
		{ 503, 29, 39 },
		{ 503, 30, 39 },
		{ 503, 31, 39 },
		{ 503, 32, 39 },
		{ 503, 33, 39 },
		{ 503, 34, 39 },
		{ 503, 35, 39 },
		{ 503, 36, 39 },
		{ 503, 37, 39 },
		{ 503, 38, 39 },
		{ 503, 39, 39 },
		{ 503, 40, 39 },
		{ 503, 41, 39 },
		{ 503, 42, 39 },
		{ 503, 43, 39 },
		{ 503, 44, 39 },
		{ 503, 45, 39 },
		{ 503, 46, 39 },
		{ 503, 47, 39 },
		{ 503, 48, 39 },
		{ 503, 49, 39 },
		{ 503, 50, 39 },
		{ 503, 51, 39 },
		{ 503, 52, 39 },
		{ 503, 53, 39 },
		{ 503, 54, 39 },
		{ 503, 55, 39 },
		{ 503, 56, 39 },
		{ 503, 57, 39 },
		{ 503, 58, 39 },
		{ 503, 59, 39 },
		{ 503, 60, 39 },
		{ 503, 61, 39 },
		{ 503, 62, 39 },
		{ 503, 63, 39 },
		{ 503, 64, 39 },
		{ 503, 65, 39 },
		{ 503, 66, 39 },
		{ 503, 67, 39 },
		{ 503, 68, 39 },
		{ 503, 69, 39 },
		{ 503, 70, 39 },
		{ 503, 71, 39 },
		{ 503, 72, 39 },
		{ 504, 11, 39 },
		{ 504, 12, 39 },
		{ 504, 13, 39 },
		{ 504, 14, 39 },
		{ 504, 15, 39 },
		{ 504, 16, 39 },
		{ 504, 17, 39 },
		{ 504, 18, 39 },
		{ 504, 19, 39 },
		{ 504, 20, 39 },
		{ 504, 21, 39 },
		{ 504, 22, 39 },
		{ 504, 23, 39 },
		{ 504, 24, 39 },
		{ 504, 25, 39 },
		{ 504, 26, 39 },
		{ 504, 27, 39 },
		{ 504, 28, 39 },
		{ 504, 29, 39 },
		{ 504, 30, 39 },
		{ 504, 31, 39 },
		{ 504, 32, 39 },
		{ 504, 33, 39 },
		{ 504, 34, 39 },
		{ 504, 35, 39 },
		{ 504, 36, 39 },
		{ 504, 37, 39 },
		{ 504, 38, 39 },
		{ 504, 39, 39 },
		{ 504, 40, 39 },
		{ 504, 41, 39 },
		{ 504, 42, 39 },
		{ 504, 43, 39 },
		{ 504, 44, 39 },
		{ 504, 45, 39 },
		{ 504, 46, 39 },
		{ 504, 47, 39 },
		{ 504, 48, 39 },
		{ 504, 49, 39 },
		{ 504, 50, 39 },
		{ 504, 51, 39 },
		{ 504, 52, 39 },
		{ 504, 53, 39 },
		{ 504, 54, 39 },
		{ 504, 55, 39 },
		{ 504, 56, 39 },
		{ 504, 57, 39 },
		{ 504, 58, 39 },
		{ 504, 59, 39 },
		{ 504, 60, 39 },
		{ 504, 61, 39 },
		{ 504, 62, 39 },
		{ 504, 63, 39 },
		{ 504, 64, 39 },
		{ 504, 65, 39 },
		{ 504, 66, 39 },
		{ 504, 67, 39 },
		{ 504, 68, 39 },
		{ 504, 69, 39 },
		{ 504, 70, 39 },
		{ 504, 71, 39 },
		{ 505, 10, 39 },
		{ 505, 11, 39 },
		{ 505, 12, 39 },
		{ 505, 13, 39 },
		{ 505, 14, 39 },
		{ 505, 15, 39 },
		{ 505, 16, 39 },
		{ 505, 17, 39 },
		{ 505, 18, 39 },
		{ 505, 19, 39 },
		{ 505, 20, 39 },
		{ 505, 21, 39 },
		{ 505, 22, 39 },
		{ 505, 23, 39 },
		{ 505, 24, 39 },
		{ 505, 25, 39 },
		{ 505, 26, 39 },
		{ 505, 27, 39 },
		{ 505, 28, 39 },
		{ 505, 29, 39 },
		{ 505, 30, 39 },
		{ 505, 31, 39 },
		{ 505, 32, 39 },
		{ 505, 33, 39 },
		{ 505, 34, 39 },
		{ 505, 35, 39 },
		{ 505, 36, 39 },
		{ 505, 37, 39 },
		{ 505, 38, 39 },
		{ 505, 39, 39 },
		{ 505, 40, 39 },
		{ 505, 41, 39 },
		{ 505, 42, 39 },
		{ 505, 43, 39 },
		{ 505, 44, 39 },
		{ 505, 45, 39 },
		{ 505, 46, 39 },
		{ 505, 47, 39 },
		{ 505, 48, 39 },
		{ 505, 49, 39 },
		{ 505, 50, 39 },
		{ 505, 51, 39 },
		{ 505, 52, 39 },
		{ 505, 53, 39 },
		{ 505, 54, 39 },
		{ 505, 55, 39 },
		{ 505, 56, 39 },
		{ 505, 57, 39 },
		{ 505, 58, 39 },
		{ 505, 59, 39 },
		{ 505, 60, 39 },
		{ 505, 61, 39 },
		{ 505, 62, 39 },
		{ 505, 63, 39 },
		{ 505, 64, 39 },
		{ 505, 65, 39 },
		{ 505, 66, 39 },
		{ 505, 67, 39 },
		{ 505, 68, 39 },
		{ 505, 69, 39 },
		{ 505, 70, 39 },
		{ 506, 9, 39 },
		{ 506, 10, 39 },
		{ 506, 11, 39 },
		{ 506, 12, 39 },
		{ 506, 13, 39 },
		{ 506, 14, 39 },
		{ 506, 15, 39 },
		{ 506, 16, 39 },
		{ 506, 17, 39 },
		{ 506, 18, 39 },
		{ 506, 19, 39 },
		{ 506, 20, 39 },
		{ 506, 21, 39 },
		{ 506, 22, 39 },
		{ 506, 23, 39 },
		{ 506, 24, 39 },
		{ 506, 25, 39 },
		{ 506, 26, 39 },
		{ 506, 27, 39 },
		{ 506, 28, 39 },
		{ 506, 29, 39 },
		{ 506, 30, 39 },
		{ 506, 31, 39 },
		{ 506, 32, 39 },
		{ 506, 33, 39 },
		{ 506, 34, 39 },
		{ 506, 35, 39 },
		{ 506, 36, 39 },
		{ 506, 37, 39 },
		{ 506, 38, 39 },
		{ 506, 39, 39 },
		{ 506, 40, 39 },
		{ 506, 41, 39 },
		{ 506, 42, 39 },
		{ 506, 43, 39 },
		{ 506, 44, 39 },
		{ 506, 45, 39 },
		{ 506, 46, 39 },
		{ 506, 47, 39 },
		{ 506, 48, 39 },
		{ 506, 49, 39 },
		{ 506, 50, 39 },
		{ 506, 51, 39 },
		{ 506, 52, 39 },
		{ 506, 53, 39 },
		{ 506, 54, 39 },
		{ 506, 55, 39 },
		{ 506, 56, 39 },
		{ 506, 57, 39 },
		{ 506, 58, 39 },
		{ 506, 59, 39 },
		{ 506, 60, 39 },
		{ 506, 61, 39 },
		{ 506, 62, 39 },
		{ 506, 63, 39 },
		{ 506, 64, 39 },
		{ 506, 65, 39 },
		{ 506, 66, 39 },
		{ 506, 67, 39 },
		{ 506, 68, 39 },
		{ 506, 69, 39 },
		{ 506, 70, 39 },
		{ 507, 8, 39 },
		{ 507, 9, 39 },
		{ 507, 10, 39 },
		{ 507, 11, 39 },
		{ 507, 12, 39 },
		{ 507, 13, 39 },
		{ 507, 14, 39 },
		{ 507, 15, 39 },
		{ 507, 16, 39 },
		{ 507, 17, 39 },
		{ 507, 18, 39 },
		{ 507, 19, 39 },
		{ 507, 20, 39 },
		{ 507, 21, 39 },
		{ 507, 22, 39 },
		{ 507, 23, 39 },
		{ 507, 24, 39 },
		{ 507, 25, 39 },
		{ 507, 26, 39 },
		{ 507, 27, 39 },
		{ 507, 28, 39 },
		{ 507, 29, 39 },
		{ 507, 30, 39 },
		{ 507, 31, 39 },
		{ 507, 32, 39 },
		{ 507, 33, 39 },
		{ 507, 34, 39 },
		{ 507, 35, 39 },
		{ 507, 36, 39 },
		{ 507, 37, 39 },
		{ 507, 38, 39 },
		{ 507, 39, 39 },
		{ 507, 40, 39 },
		{ 507, 41, 39 },
		{ 507, 42, 39 },
		{ 507, 43, 39 },
		{ 507, 44, 39 },
		{ 507, 45, 39 },
		{ 507, 46, 39 },
		{ 507, 47, 39 },
		{ 507, 48, 39 },
		{ 507, 49, 39 },
		{ 507, 50, 39 },
		{ 507, 51, 39 },
		{ 507, 52, 39 },
		{ 507, 53, 39 },
		{ 507, 54, 39 },
		{ 507, 55, 39 },
		{ 507, 56, 39 },
		{ 507, 57, 39 },
		{ 507, 58, 39 },
		{ 507, 59, 39 },
		{ 507, 60, 39 },
		{ 507, 61, 39 },
		{ 507, 62, 39 },
		{ 507, 63, 39 },
		{ 507, 64, 39 },
		{ 507, 65, 39 },
		{ 507, 66, 39 },
		{ 507, 67, 39 },
		{ 507, 68, 39 },
		{ 507, 69, 39 },
		{ 508, 7, 39 },
		{ 508, 8, 39 },
		{ 508, 9, 39 },
		{ 508, 10, 39 },
		{ 508, 11, 39 },
		{ 508, 12, 39 },
		{ 508, 13, 39 },
		{ 508, 14, 39 },
		{ 508, 15, 39 },
		{ 508, 16, 39 },
		{ 508, 17, 39 },
		{ 508, 18, 39 },
		{ 508, 19, 39 },
		{ 508, 20, 39 },
		{ 508, 21, 39 },
		{ 508, 22, 39 },
		{ 508, 23, 39 },
		{ 508, 24, 39 },
		{ 508, 25, 39 },
		{ 508, 26, 39 },
		{ 508, 27, 39 },
		{ 508, 28, 39 },
		{ 508, 29, 39 },
		{ 508, 30, 39 },
		{ 508, 31, 39 },
		{ 508, 32, 39 },
		{ 508, 33, 39 },
		{ 508, 34, 39 },
		{ 508, 35, 39 },
		{ 508, 36, 39 },
		{ 508, 37, 39 },
		{ 508, 38, 39 },
		{ 508, 39, 39 },
		{ 508, 40, 39 },
		{ 508, 41, 39 },
		{ 508, 42, 39 },
		{ 508, 43, 39 },
		{ 508, 44, 39 },
		{ 508, 45, 39 },
		{ 508, 46, 39 },
		{ 508, 47, 39 },
		{ 508, 48, 39 },
		{ 508, 49, 39 },
		{ 508, 50, 39 },
		{ 508, 51, 39 },
		{ 508, 52, 39 },
		{ 508, 53, 39 },
		{ 508, 54, 39 },
		{ 508, 55, 39 },
		{ 508, 56, 39 },
		{ 508, 57, 39 },
		{ 508, 58, 39 },
		{ 508, 59, 39 },
		{ 508, 60, 39 },
		{ 508, 61, 39 },
		{ 508, 62, 39 },
		{ 508, 63, 39 },
		{ 508, 64, 39 },
		{ 508, 65, 39 },
		{ 508, 66, 39 },
		{ 508, 67, 39 },
		{ 508, 68, 39 },
		{ 509, 6, 39 },
		{ 509, 7, 39 },
		{ 509, 8, 39 },
		{ 509, 9, 39 },
		{ 509, 10, 39 },
		{ 509, 11, 39 },
		{ 509, 12, 39 },
		{ 509, 13, 39 },
		{ 509, 14, 39 },
		{ 509, 15, 39 },
		{ 509, 16, 39 },
		{ 509, 17, 39 },
		{ 509, 18, 39 },
		{ 509, 19, 39 },
		{ 509, 20, 39 },
		{ 509, 21, 39 },
		{ 509, 22, 39 },
		{ 509, 23, 39 },
		{ 509, 24, 39 },
		{ 509, 25, 39 },
		{ 509, 26, 39 },
		{ 509, 27, 39 },
		{ 509, 28, 39 },
		{ 509, 29, 39 },
		{ 509, 30, 39 },
		{ 509, 31, 39 },
		{ 509, 32, 39 },
		{ 509, 33, 39 },
		{ 509, 34, 39 },
		{ 509, 35, 39 },
		{ 509, 36, 39 },
		{ 509, 37, 39 },
		{ 509, 38, 39 },
		{ 509, 39, 39 },
		{ 509, 40, 39 },
		{ 509, 41, 39 },
		{ 509, 42, 39 },
		{ 509, 43, 39 },
		{ 509, 44, 39 },
		{ 509, 45, 39 },
		{ 509, 46, 39 },
		{ 509, 47, 39 },
		{ 509, 48, 39 },
		{ 509, 49, 39 },
		{ 509, 50, 39 },
		{ 509, 51, 39 },
		{ 509, 52, 39 },
		{ 509, 53, 39 },
		{ 509, 54, 39 },
		{ 509, 55, 39 },
		{ 509, 56, 39 },
		{ 509, 57, 39 },
		{ 509, 58, 39 },
		{ 509, 59, 39 },
		{ 509, 60, 39 },
		{ 509, 61, 39 },
		{ 509, 62, 39 },
		{ 509, 63, 39 },
		{ 509, 64, 39 },
		{ 509, 65, 39 },
		{ 509, 66, 39 },
		{ 509, 67, 39 },
		{ 510, 5, 39 },
		{ 510, 6, 39 },
		{ 510, 7, 39 },
		{ 510, 8, 39 },
		{ 510, 9, 39 },
		{ 510, 10, 39 },
		{ 510, 11, 39 },
		{ 510, 12, 39 },
		{ 510, 13, 39 },
		{ 510, 14, 39 },
		{ 510, 15, 39 },
		{ 510, 16, 39 },
		{ 510, 17, 39 },
		{ 510, 18, 39 },
		{ 510, 19, 39 },
		{ 510, 20, 39 },
		{ 510, 21, 39 },
		{ 510, 22, 39 },
		{ 510, 23, 39 },
		{ 510, 24, 39 },
		{ 510, 25, 39 },
		{ 510, 26, 39 },
		{ 510, 27, 39 },
		{ 510, 28, 39 },
		{ 510, 29, 39 },
		{ 510, 30, 39 },
		{ 510, 31, 39 },
		{ 510, 32, 39 },
		{ 510, 33, 39 },
		{ 510, 34, 39 },
		{ 510, 35, 39 },
		{ 510, 36, 39 },
		{ 510, 37, 39 },
		{ 510, 38, 39 },
		{ 510, 39, 39 },
		{ 510, 40, 39 },
		{ 510, 41, 39 },
		{ 510, 42, 39 },
		{ 510, 43, 39 },
		{ 510, 44, 39 },
		{ 510, 45, 39 },
		{ 510, 46, 39 },
		{ 510, 47, 39 },
		{ 510, 48, 39 },
		{ 510, 49, 39 },
		{ 510, 50, 39 },
		{ 510, 51, 39 },
		{ 510, 52, 39 },
		{ 510, 53, 39 },
		{ 510, 54, 39 },
		{ 510, 55, 39 },
		{ 510, 56, 39 },
		{ 510, 57, 39 },
		{ 510, 58, 39 },
		{ 510, 59, 39 },
		{ 510, 60, 39 },
		{ 510, 61, 39 },
		{ 510, 62, 39 },
		{ 510, 63, 39 },
		{ 510, 64, 39 },
		{ 510, 65, 39 },
		{ 510, 66, 39 },
		{ 510, 67, 39 },
		{ 511, 4, 39 },
		{ 511, 5, 39 },
		{ 511, 6, 39 },
		{ 511, 7, 39 },
		{ 511, 8, 39 },
		{ 511, 9, 39 },
		{ 511, 10, 39 },
		{ 511, 11, 39 },
		{ 511, 12, 39 },
		{ 511, 13, 39 },
		{ 511, 14, 39 },
		{ 511, 15, 39 },
		{ 511, 16, 39 },
		{ 511, 17, 39 },
		{ 511, 18, 39 },
		{ 511, 19, 39 },
		{ 511, 20, 39 },
		{ 511, 21, 39 },
		{ 511, 22, 39 },
		{ 511, 23, 39 },
		{ 511, 24, 39 },
		{ 511, 25, 39 },
		{ 511, 26, 39 },
		{ 511, 27, 39 },
		{ 511, 28, 39 },
		{ 511, 29, 39 },
		{ 511, 30, 39 },
		{ 511, 31, 39 },
		{ 511, 32, 39 },
		{ 511, 33, 39 },
		{ 511, 34, 39 },
		{ 511, 35, 39 },
		{ 511, 36, 39 },
		{ 511, 37, 39 },
		{ 511, 38, 39 },
		{ 511, 39, 39 },
		{ 511, 40, 39 },
		{ 511, 41, 39 },
		{ 511, 42, 39 },
		{ 511, 43, 39 },
		{ 511, 44, 39 },
		{ 511, 45, 39 },
		{ 511, 46, 39 },
		{ 511, 47, 39 },
		{ 511, 48, 39 },
		{ 511, 49, 39 },
		{ 511, 50, 39 },
		{ 511, 51, 39 },
		{ 511, 52, 39 },
		{ 511, 53, 39 },
		{ 511, 54, 39 },
		{ 511, 55, 39 },
		{ 511, 56, 39 },
		{ 511, 57, 39 },
		{ 511, 58, 39 },
		{ 511, 59, 39 },
		{ 511, 60, 39 },
		{ 511, 61, 39 },
		{ 511, 62, 39 },
		{ 511, 63, 39 },
		{ 511, 64, 39 },
		{ 511, 65, 39 },
		{ 511, 66, 39 },
		{ 512, 3, 39 },
		{ 512, 4, 39 },
		{ 512, 5, 39 },
		{ 512, 6, 39 },
		{ 512, 7, 39 },
		{ 512, 8, 39 },
		{ 512, 9, 39 },
		{ 512, 10, 39 },
		{ 512, 11, 39 },
		{ 512, 12, 39 },
		{ 512, 13, 39 },
		{ 512, 14, 39 },
		{ 512, 15, 39 },
		{ 512, 16, 39 },
		{ 512, 17, 39 },
		{ 512, 18, 39 },
		{ 512, 19, 39 },
		{ 512, 20, 39 },
		{ 512, 21, 39 },
		{ 512, 22, 39 },
		{ 512, 23, 39 },
		{ 512, 24, 39 },
		{ 512, 25, 39 },
		{ 512, 26, 39 },
		{ 512, 27, 39 },
		{ 512, 28, 39 },
		{ 512, 29, 39 },
		{ 512, 30, 39 },
		{ 512, 31, 39 },
		{ 512, 32, 39 },
		{ 512, 33, 39 },
		{ 512, 34, 39 },
		{ 512, 35, 39 },
		{ 512, 36, 39 },
		{ 512, 37, 39 },
		{ 512, 38, 39 },
		{ 512, 39, 39 },
		{ 512, 40, 39 },
		{ 512, 41, 39 },
		{ 512, 42, 39 },
		{ 512, 43, 39 },
		{ 512, 44, 39 },
		{ 512, 45, 39 },
		{ 512, 46, 39 },
		{ 512, 47, 39 },
		{ 512, 48, 39 },
		{ 512, 49, 39 },
		{ 512, 50, 39 },
		{ 512, 51, 39 },
		{ 512, 52, 39 },
		{ 512, 53, 39 },
		{ 512, 54, 39 },
		{ 512, 55, 39 },
		{ 512, 56, 39 },
		{ 512, 57, 39 },
		{ 512, 58, 39 },
		{ 512, 59, 39 },
		{ 512, 60, 39 },
		{ 512, 61, 39 },
		{ 512, 62, 39 },
		{ 512, 63, 39 },
		{ 512, 64, 39 },
		{ 512, 65, 39 },
		{ 1, 418, 40 },
		{ 1, 419, 40 },
		{ 1, 420, 40 },
		{ 1, 421, 40 },
		{ 1, 422, 40 },
		{ 1, 423, 40 },
		{ 1, 424, 40 },
		{ 1, 425, 40 },
		{ 1, 426, 40 },
		{ 1, 427, 40 },
		{ 1, 428, 40 },
		{ 1, 429, 40 },
		{ 1, 430, 40 },
		{ 1, 431, 40 },
		{ 1, 432, 40 },
		{ 1, 433, 40 },
		{ 1, 434, 40 },
		{ 1, 435, 40 },
		{ 1, 436, 40 },
		{ 1, 437, 40 },
		{ 1, 438, 40 },
		{ 1, 439, 40 },
		{ 1, 440, 40 },
		{ 1, 441, 40 },
		{ 1, 442, 40 },
		{ 1, 443, 40 },
		{ 1, 444, 40 },
		{ 1, 445, 40 },
		{ 1, 446, 40 },
		{ 1, 447, 40 },
		{ 1, 448, 40 },
		{ 1, 449, 40 },
		{ 1, 450, 40 },
		{ 1, 451, 40 },
		{ 1, 452, 40 },
		{ 1, 453, 40 },
		{ 1, 454, 40 },
		{ 1, 455, 40 },
		{ 1, 456, 40 },
		{ 1, 457, 40 },
		{ 1, 458, 40 },
		{ 1, 459, 40 },
		{ 1, 460, 40 },
		{ 1, 461, 40 },
		{ 1, 462, 40 },
		{ 1, 463, 40 },
		{ 1, 464, 40 },
		{ 1, 465, 40 },
		{ 1, 466, 40 },
		{ 1, 467, 40 },
		{ 1, 468, 40 },
		{ 1, 469, 40 },
		{ 1, 470, 40 },
		{ 1, 471, 40 },
		{ 1, 472, 40 },
		{ 1, 473, 40 },
		{ 1, 474, 40 },
		{ 1, 475, 40 },
		{ 1, 476, 40 },
		{ 1, 477, 40 },
		{ 1, 478, 40 },
		{ 1, 479, 40 },
		{ 1, 480, 40 },
		{ 2, 417, 40 },
		{ 2, 418, 40 },
		{ 2, 419, 40 },
		{ 2, 420, 40 },
		{ 2, 421, 40 },
		{ 2, 422, 40 },
		{ 2, 423, 40 },
		{ 2, 424, 40 },
		{ 2, 425, 40 },
		{ 2, 426, 40 },
		{ 2, 427, 40 },
		{ 2, 428, 40 },
		{ 2, 429, 40 },
		{ 2, 430, 40 },
		{ 2, 431, 40 },
		{ 2, 432, 40 },
		{ 2, 433, 40 },
		{ 2, 434, 40 },
		{ 2, 435, 40 },
		{ 2, 436, 40 },
		{ 2, 437, 40 },
		{ 2, 438, 40 },
		{ 2, 439, 40 },
		{ 2, 440, 40 },
		{ 2, 441, 40 },
		{ 2, 442, 40 },
		{ 2, 443, 40 },
		{ 2, 444, 40 },
		{ 2, 445, 40 },
		{ 2, 446, 40 },
		{ 2, 447, 40 },
		{ 2, 448, 40 },
		{ 2, 449, 40 },
		{ 2, 450, 40 },
		{ 2, 451, 40 },
		{ 2, 452, 40 },
		{ 2, 453, 40 },
		{ 2, 454, 40 },
		{ 2, 455, 40 },
		{ 2, 456, 40 },
		{ 2, 457, 40 },
		{ 2, 458, 40 },
		{ 2, 459, 40 },
		{ 2, 460, 40 },
		{ 2, 461, 40 },
		{ 2, 462, 40 },
		{ 2, 463, 40 },
		{ 2, 464, 40 },
		{ 2, 465, 40 },
		{ 2, 466, 40 },
		{ 2, 467, 40 },
		{ 2, 468, 40 },
		{ 2, 469, 40 },
		{ 2, 470, 40 },
		{ 2, 471, 40 },
		{ 2, 472, 40 },
		{ 2, 473, 40 },
		{ 2, 474, 40 },
		{ 2, 475, 40 },
		{ 2, 476, 40 },
		{ 2, 477, 40 },
		{ 2, 478, 40 },
		{ 2, 479, 40 },
		{ 2, 480, 40 },
		{ 3, 416, 40 },
		{ 3, 417, 40 },
		{ 3, 418, 40 },
		{ 3, 419, 40 },
		{ 3, 420, 40 },
		{ 3, 421, 40 },
		{ 3, 422, 40 },
		{ 3, 423, 40 },
		{ 3, 424, 40 },
		{ 3, 425, 40 },
		{ 3, 426, 40 },
		{ 3, 427, 40 },
		{ 3, 428, 40 },
		{ 3, 429, 40 },
		{ 3, 430, 40 },
		{ 3, 431, 40 },
		{ 3, 432, 40 },
		{ 3, 433, 40 },
		{ 3, 434, 40 },
		{ 3, 435, 40 },
		{ 3, 436, 40 },
		{ 3, 437, 40 },
		{ 3, 438, 40 },
		{ 3, 439, 40 },
		{ 3, 440, 40 },
		{ 3, 441, 40 },
		{ 3, 442, 40 },
		{ 3, 443, 40 },
		{ 3, 444, 40 },
		{ 3, 445, 40 },
		{ 3, 446, 40 },
		{ 3, 447, 40 },
		{ 3, 448, 40 },
		{ 3, 449, 40 },
		{ 3, 450, 40 },
		{ 3, 451, 40 },
		{ 3, 452, 40 },
		{ 3, 453, 40 },
		{ 3, 454, 40 },
		{ 3, 455, 40 },
		{ 3, 456, 40 },
		{ 3, 457, 40 },
		{ 3, 458, 40 },
		{ 3, 459, 40 },
		{ 3, 460, 40 },
		{ 3, 461, 40 },
		{ 3, 462, 40 },
		{ 3, 463, 40 },
		{ 3, 464, 40 },
		{ 3, 465, 40 },
		{ 3, 466, 40 },
		{ 3, 467, 40 },
		{ 3, 468, 40 },
		{ 3, 469, 40 },
		{ 3, 470, 40 },
		{ 3, 471, 40 },
		{ 3, 472, 40 },
		{ 3, 473, 40 },
		{ 3, 474, 40 },
		{ 3, 475, 40 },
		{ 3, 476, 40 },
		{ 3, 477, 40 },
		{ 3, 478, 40 },
		{ 3, 479, 40 },
		{ 4, 416, 40 },
		{ 4, 417, 40 },
		{ 4, 418, 40 },
		{ 4, 419, 40 },
		{ 4, 420, 40 },
		{ 4, 421, 40 },
		{ 4, 422, 40 },
		{ 4, 423, 40 },
		{ 4, 424, 40 },
		{ 4, 425, 40 },
		{ 4, 426, 40 },
		{ 4, 427, 40 },
		{ 4, 428, 40 },
		{ 4, 429, 40 },
		{ 4, 430, 40 },
		{ 4, 431, 40 },
		{ 4, 432, 40 },
		{ 4, 433, 40 },
		{ 4, 434, 40 },
		{ 4, 435, 40 },
		{ 4, 436, 40 },
		{ 4, 437, 40 },
		{ 4, 438, 40 },
		{ 4, 439, 40 },
		{ 4, 440, 40 },
		{ 4, 441, 40 },
		{ 4, 442, 40 },
		{ 4, 443, 40 },
		{ 4, 444, 40 },
		{ 4, 445, 40 },
		{ 4, 446, 40 },
		{ 4, 447, 40 },
		{ 4, 448, 40 },
		{ 4, 449, 40 },
		{ 4, 450, 40 },
		{ 4, 451, 40 },
		{ 4, 452, 40 },
		{ 4, 453, 40 },
		{ 4, 454, 40 },
		{ 4, 455, 40 },
		{ 4, 456, 40 },
		{ 4, 457, 40 },
		{ 4, 458, 40 },
		{ 4, 459, 40 },
		{ 4, 460, 40 },
		{ 4, 461, 40 },
		{ 4, 462, 40 },
		{ 4, 463, 40 },
		{ 4, 464, 40 },
		{ 4, 465, 40 },
		{ 4, 466, 40 },
		{ 4, 467, 40 },
		{ 4, 468, 40 },
		{ 4, 469, 40 },
		{ 4, 470, 40 },
		{ 4, 471, 40 },
		{ 4, 472, 40 },
		{ 4, 473, 40 },
		{ 4, 474, 40 },
		{ 4, 475, 40 },
		{ 4, 476, 40 },
		{ 4, 477, 40 },
		{ 4, 478, 40 },
		{ 5, 415, 40 },
		{ 5, 416, 40 },
		{ 5, 417, 40 },
		{ 5, 418, 40 },
		{ 5, 419, 40 },
		{ 5, 420, 40 },
		{ 5, 421, 40 },
		{ 5, 422, 40 },
		{ 5, 423, 40 },
		{ 5, 424, 40 },
		{ 5, 425, 40 },
		{ 5, 426, 40 },
		{ 5, 427, 40 },
		{ 5, 428, 40 },
		{ 5, 429, 40 },
		{ 5, 430, 40 },
		{ 5, 431, 40 },
		{ 5, 432, 40 },
		{ 5, 433, 40 },
		{ 5, 434, 40 },
		{ 5, 435, 40 },
		{ 5, 436, 40 },
		{ 5, 437, 40 },
		{ 5, 438, 40 },
		{ 5, 439, 40 },
		{ 5, 440, 40 },
		{ 5, 441, 40 },
		{ 5, 442, 40 },
		{ 5, 443, 40 },
		{ 5, 444, 40 },
		{ 5, 445, 40 },
		{ 5, 446, 40 },
		{ 5, 447, 40 },
		{ 5, 448, 40 },
		{ 5, 449, 40 },
		{ 5, 450, 40 },
		{ 5, 451, 40 },
		{ 5, 452, 40 },
		{ 5, 453, 40 },
		{ 5, 454, 40 },
		{ 5, 455, 40 },
		{ 5, 456, 40 },
		{ 5, 457, 40 },
		{ 5, 458, 40 },
		{ 5, 459, 40 },
		{ 5, 460, 40 },
		{ 5, 461, 40 },
		{ 5, 462, 40 },
		{ 5, 463, 40 },
		{ 5, 464, 40 },
		{ 5, 465, 40 },
		{ 5, 466, 40 },
		{ 5, 467, 40 },
		{ 5, 468, 40 },
		{ 5, 469, 40 },
		{ 5, 470, 40 },
		{ 5, 471, 40 },
		{ 5, 472, 40 },
		{ 5, 473, 40 },
		{ 5, 474, 40 },
		{ 5, 475, 40 },
		{ 5, 476, 40 },
		{ 5, 477, 40 },
		{ 6, 414, 40 },
		{ 6, 415, 40 },
		{ 6, 416, 40 },
		{ 6, 417, 40 },
		{ 6, 418, 40 },
		{ 6, 419, 40 },
		{ 6, 420, 40 },
		{ 6, 421, 40 },
		{ 6, 422, 40 },
		{ 6, 423, 40 },
		{ 6, 424, 40 },
		{ 6, 425, 40 },
		{ 6, 426, 40 },
		{ 6, 427, 40 },
		{ 6, 428, 40 },
		{ 6, 429, 40 },
		{ 6, 430, 40 },
		{ 6, 431, 40 },
		{ 6, 432, 40 },
		{ 6, 433, 40 },
		{ 6, 434, 40 },
		{ 6, 435, 40 },
		{ 6, 436, 40 },
		{ 6, 437, 40 },
		{ 6, 438, 40 },
		{ 6, 439, 40 },
		{ 6, 440, 40 },
		{ 6, 441, 40 },
		{ 6, 442, 40 },
		{ 6, 443, 40 },
		{ 6, 444, 40 },
		{ 6, 445, 40 },
		{ 6, 446, 40 },
		{ 6, 447, 40 },
		{ 6, 448, 40 },
		{ 6, 449, 40 },
		{ 6, 450, 40 },
		{ 6, 451, 40 },
		{ 6, 452, 40 },
		{ 6, 453, 40 },
		{ 6, 454, 40 },
		{ 6, 455, 40 },
		{ 6, 456, 40 },
		{ 6, 457, 40 },
		{ 6, 458, 40 },
		{ 6, 459, 40 },
		{ 6, 460, 40 },
		{ 6, 461, 40 },
		{ 6, 462, 40 },
		{ 6, 463, 40 },
		{ 6, 464, 40 },
		{ 6, 465, 40 },
		{ 6, 466, 40 },
		{ 6, 467, 40 },
		{ 6, 468, 40 },
		{ 6, 469, 40 },
		{ 6, 470, 40 },
		{ 6, 471, 40 },
		{ 6, 472, 40 },
		{ 6, 473, 40 },
		{ 6, 474, 40 },
		{ 6, 475, 40 },
		{ 6, 476, 40 },
		{ 7, 414, 40 },
		{ 7, 415, 40 },
		{ 7, 416, 40 },
		{ 7, 417, 40 },
		{ 7, 418, 40 },
		{ 7, 419, 40 },
		{ 7, 420, 40 },
		{ 7, 421, 40 },
		{ 7, 422, 40 },
		{ 7, 423, 40 },
		{ 7, 424, 40 },
		{ 7, 425, 40 },
		{ 7, 426, 40 },
		{ 7, 427, 40 },
		{ 7, 428, 40 },
		{ 7, 429, 40 },
		{ 7, 430, 40 },
		{ 7, 431, 40 },
		{ 7, 432, 40 },
		{ 7, 433, 40 },
		{ 7, 434, 40 },
		{ 7, 435, 40 },
		{ 7, 436, 40 },
		{ 7, 437, 40 },
		{ 7, 438, 40 },
		{ 7, 439, 40 },
		{ 7, 440, 40 },
		{ 7, 441, 40 },
		{ 7, 442, 40 },
		{ 7, 443, 40 },
		{ 7, 444, 40 },
		{ 7, 445, 40 },
		{ 7, 446, 40 },
		{ 7, 447, 40 },
		{ 7, 448, 40 },
		{ 7, 449, 40 },
		{ 7, 450, 40 },
		{ 7, 451, 40 },
		{ 7, 452, 40 },
		{ 7, 453, 40 },
		{ 7, 454, 40 },
		{ 7, 455, 40 },
		{ 7, 456, 40 },
		{ 7, 457, 40 },
		{ 7, 458, 40 },
		{ 7, 459, 40 },
		{ 7, 460, 40 },
		{ 7, 461, 40 },
		{ 7, 462, 40 },
		{ 7, 463, 40 },
		{ 7, 464, 40 },
		{ 7, 465, 40 },
		{ 7, 466, 40 },
		{ 7, 467, 40 },
		{ 7, 468, 40 },
		{ 7, 469, 40 },
		{ 7, 470, 40 },
		{ 7, 471, 40 },
		{ 7, 472, 40 },
		{ 7, 473, 40 },
		{ 7, 474, 40 },
		{ 7, 475, 40 },
		{ 8, 413, 40 },
		{ 8, 414, 40 },
		{ 8, 415, 40 },
		{ 8, 416, 40 },
		{ 8, 417, 40 },
		{ 8, 418, 40 },
		{ 8, 419, 40 },
		{ 8, 420, 40 },
		{ 8, 421, 40 },
		{ 8, 422, 40 },
		{ 8, 423, 40 },
		{ 8, 424, 40 },
		{ 8, 425, 40 },
		{ 8, 426, 40 },
		{ 8, 427, 40 },
		{ 8, 428, 40 },
		{ 8, 429, 40 },
		{ 8, 430, 40 },
		{ 8, 431, 40 },
		{ 8, 432, 40 },
		{ 8, 433, 40 },
		{ 8, 434, 40 },
		{ 8, 435, 40 },
		{ 8, 436, 40 },
		{ 8, 437, 40 },
		{ 8, 438, 40 },
		{ 8, 439, 40 },
		{ 8, 440, 40 },
		{ 8, 441, 40 },
		{ 8, 442, 40 },
		{ 8, 443, 40 },
		{ 8, 444, 40 },
		{ 8, 445, 40 },
		{ 8, 446, 40 },
		{ 8, 447, 40 },
		{ 8, 448, 40 },
		{ 8, 449, 40 },
		{ 8, 450, 40 },
		{ 8, 451, 40 },
		{ 8, 452, 40 },
		{ 8, 453, 40 },
		{ 8, 454, 40 },
		{ 8, 455, 40 },
		{ 8, 456, 40 },
		{ 8, 457, 40 },
		{ 8, 458, 40 },
		{ 8, 459, 40 },
		{ 8, 460, 40 },
		{ 8, 461, 40 },
		{ 8, 462, 40 },
		{ 8, 463, 40 },
		{ 8, 464, 40 },
		{ 8, 465, 40 },
		{ 8, 466, 40 },
		{ 8, 467, 40 },
		{ 8, 468, 40 },
		{ 8, 469, 40 },
		{ 8, 470, 40 },
		{ 8, 471, 40 },
		{ 8, 472, 40 },
		{ 8, 473, 40 },
		{ 8, 474, 40 },
		{ 9, 413, 40 },
		{ 9, 414, 40 },
		{ 9, 415, 40 },
		{ 9, 416, 40 },
		{ 9, 417, 40 },
		{ 9, 418, 40 },
		{ 9, 419, 40 },
		{ 9, 420, 40 },
		{ 9, 421, 40 },
		{ 9, 422, 40 },
		{ 9, 423, 40 },
		{ 9, 424, 40 },
		{ 9, 425, 40 },
		{ 9, 426, 40 },
		{ 9, 427, 40 },
		{ 9, 428, 40 },
		{ 9, 429, 40 },
		{ 9, 430, 40 },
		{ 9, 431, 40 },
		{ 9, 432, 40 },
		{ 9, 433, 40 },
		{ 9, 434, 40 },
		{ 9, 435, 40 },
		{ 9, 436, 40 },
		{ 9, 437, 40 },
		{ 9, 438, 40 },
		{ 9, 439, 40 },
		{ 9, 440, 40 },
		{ 9, 441, 40 },
		{ 9, 442, 40 },
		{ 9, 443, 40 },
		{ 9, 444, 40 },
		{ 9, 445, 40 },
		{ 9, 446, 40 },
		{ 9, 447, 40 },
		{ 9, 448, 40 },
		{ 9, 449, 40 },
		{ 9, 450, 40 },
		{ 9, 451, 40 },
		{ 9, 452, 40 },
		{ 9, 453, 40 },
		{ 9, 454, 40 },
		{ 9, 455, 40 },
		{ 9, 456, 40 },
		{ 9, 457, 40 },
		{ 9, 458, 40 },
		{ 9, 459, 40 },
		{ 9, 460, 40 },
		{ 9, 461, 40 },
		{ 9, 462, 40 },
		{ 9, 463, 40 },
		{ 9, 464, 40 },
		{ 9, 465, 40 },
		{ 9, 466, 40 },
		{ 9, 467, 40 },
		{ 9, 468, 40 },
		{ 9, 469, 40 },
		{ 9, 470, 40 },
		{ 9, 471, 40 },
		{ 9, 472, 40 },
		{ 9, 473, 40 },
		{ 10, 412, 40 },
		{ 10, 413, 40 },
		{ 10, 414, 40 },
		{ 10, 415, 40 },
		{ 10, 416, 40 },
		{ 10, 417, 40 },
		{ 10, 418, 40 },
		{ 10, 419, 40 },
		{ 10, 420, 40 },
		{ 10, 421, 40 },
		{ 10, 422, 40 },
		{ 10, 423, 40 },
		{ 10, 424, 40 },
		{ 10, 425, 40 },
		{ 10, 426, 40 },
		{ 10, 427, 40 },
		{ 10, 428, 40 },
		{ 10, 429, 40 },
		{ 10, 430, 40 },
		{ 10, 431, 40 },
		{ 10, 432, 40 },
		{ 10, 433, 40 },
		{ 10, 434, 40 },
		{ 10, 435, 40 },
		{ 10, 436, 40 },
		{ 10, 437, 40 },
		{ 10, 438, 40 },
		{ 10, 439, 40 },
		{ 10, 440, 40 },
		{ 10, 441, 40 },
		{ 10, 442, 40 },
		{ 10, 443, 40 },
		{ 10, 444, 40 },
		{ 10, 445, 40 },
		{ 10, 446, 40 },
		{ 10, 447, 40 },
		{ 10, 448, 40 },
		{ 10, 449, 40 },
		{ 10, 450, 40 },
		{ 10, 451, 40 },
		{ 10, 452, 40 },
		{ 10, 453, 40 },
		{ 10, 454, 40 },
		{ 10, 455, 40 },
		{ 10, 456, 40 },
		{ 10, 457, 40 },
		{ 10, 458, 40 },
		{ 10, 459, 40 },
		{ 10, 460, 40 },
		{ 10, 461, 40 },
		{ 10, 462, 40 },
		{ 10, 463, 40 },
		{ 10, 464, 40 },
		{ 10, 465, 40 },
		{ 10, 466, 40 },
		{ 10, 467, 40 },
		{ 10, 468, 40 },
		{ 10, 469, 40 },
		{ 10, 470, 40 },
		{ 10, 471, 40 },
		{ 10, 472, 40 },
		{ 10, 473, 40 },
		{ 11, 411, 40 },
		{ 11, 412, 40 },
		{ 11, 413, 40 },
		{ 11, 414, 40 },
		{ 11, 415, 40 },
		{ 11, 416, 40 },
		{ 11, 417, 40 },
		{ 11, 418, 40 },
		{ 11, 419, 40 },
		{ 11, 420, 40 },
		{ 11, 421, 40 },
		{ 11, 422, 40 },
		{ 11, 423, 40 },
		{ 11, 424, 40 },
		{ 11, 425, 40 },
		{ 11, 426, 40 },
		{ 11, 427, 40 },
		{ 11, 428, 40 },
		{ 11, 429, 40 },
		{ 11, 430, 40 },
		{ 11, 431, 40 },
		{ 11, 432, 40 },
		{ 11, 433, 40 },
		{ 11, 434, 40 },
		{ 11, 435, 40 },
		{ 11, 436, 40 },
		{ 11, 437, 40 },
		{ 11, 438, 40 },
		{ 11, 439, 40 },
		{ 11, 440, 40 },
		{ 11, 441, 40 },
		{ 11, 442, 40 },
		{ 11, 443, 40 },
		{ 11, 444, 40 },
		{ 11, 445, 40 },
		{ 11, 446, 40 },
		{ 11, 447, 40 },
		{ 11, 448, 40 },
		{ 11, 449, 40 },
		{ 11, 450, 40 },
		{ 11, 451, 40 },
		{ 11, 452, 40 },
		{ 11, 453, 40 },
		{ 11, 454, 40 },
		{ 11, 455, 40 },
		{ 11, 456, 40 },
		{ 11, 457, 40 },
		{ 11, 458, 40 },
		{ 11, 459, 40 },
		{ 11, 460, 40 },
		{ 11, 461, 40 },
		{ 11, 462, 40 },
		{ 11, 463, 40 },
		{ 11, 464, 40 },
		{ 11, 465, 40 },
		{ 11, 466, 40 },
		{ 11, 467, 40 },
		{ 11, 468, 40 },
		{ 11, 469, 40 },
		{ 11, 470, 40 },
		{ 11, 471, 40 },
		{ 11, 472, 40 },
		{ 12, 411, 40 },
		{ 12, 412, 40 },
		{ 12, 413, 40 },
		{ 12, 414, 40 },
		{ 12, 415, 40 },
		{ 12, 416, 40 },
		{ 12, 417, 40 },
		{ 12, 418, 40 },
		{ 12, 419, 40 },
		{ 12, 420, 40 },
		{ 12, 421, 40 },
		{ 12, 422, 40 },
		{ 12, 423, 40 },
		{ 12, 424, 40 },
		{ 12, 425, 40 },
		{ 12, 426, 40 },
		{ 12, 427, 40 },
		{ 12, 428, 40 },
		{ 12, 429, 40 },
		{ 12, 430, 40 },
		{ 12, 431, 40 },
		{ 12, 432, 40 },
		{ 12, 433, 40 },
		{ 12, 434, 40 },
		{ 12, 435, 40 },
		{ 12, 436, 40 },
		{ 12, 437, 40 },
		{ 12, 438, 40 },
		{ 12, 439, 40 },
		{ 12, 440, 40 },
		{ 12, 441, 40 },
		{ 12, 442, 40 },
		{ 12, 443, 40 },
		{ 12, 444, 40 },
		{ 12, 445, 40 },
		{ 12, 446, 40 },
		{ 12, 447, 40 },
		{ 12, 448, 40 },
		{ 12, 449, 40 },
		{ 12, 450, 40 },
		{ 12, 451, 40 },
		{ 12, 452, 40 },
		{ 12, 453, 40 },
		{ 12, 454, 40 },
		{ 12, 455, 40 },
		{ 12, 456, 40 },
		{ 12, 457, 40 },
		{ 12, 458, 40 },
		{ 12, 459, 40 },
		{ 12, 460, 40 },
		{ 12, 461, 40 },
		{ 12, 462, 40 },
		{ 12, 463, 40 },
		{ 12, 464, 40 },
		{ 12, 465, 40 },
		{ 12, 466, 40 },
		{ 12, 467, 40 },
		{ 12, 468, 40 },
		{ 12, 469, 40 },
		{ 12, 470, 40 },
		{ 12, 471, 40 },
		{ 13, 410, 40 },
		{ 13, 411, 40 },
		{ 13, 412, 40 },
		{ 13, 413, 40 },
		{ 13, 414, 40 },
		{ 13, 415, 40 },
		{ 13, 416, 40 },
		{ 13, 417, 40 },
		{ 13, 418, 40 },
		{ 13, 419, 40 },
		{ 13, 420, 40 },
		{ 13, 421, 40 },
		{ 13, 422, 40 },
		{ 13, 423, 40 },
		{ 13, 424, 40 },
		{ 13, 425, 40 },
		{ 13, 426, 40 },
		{ 13, 427, 40 },
		{ 13, 428, 40 },
		{ 13, 429, 40 },
		{ 13, 430, 40 },
		{ 13, 431, 40 },
		{ 13, 432, 40 },
		{ 13, 433, 40 },
		{ 13, 434, 40 },
		{ 13, 435, 40 },
		{ 13, 436, 40 },
		{ 13, 437, 40 },
		{ 13, 438, 40 },
		{ 13, 439, 40 },
		{ 13, 440, 40 },
		{ 13, 441, 40 },
		{ 13, 442, 40 },
		{ 13, 443, 40 },
		{ 13, 444, 40 },
		{ 13, 445, 40 },
		{ 13, 446, 40 },
		{ 13, 447, 40 },
		{ 13, 448, 40 },
		{ 13, 449, 40 },
		{ 13, 450, 40 },
		{ 13, 451, 40 },
		{ 13, 452, 40 },
		{ 13, 453, 40 },
		{ 13, 454, 40 },
		{ 13, 455, 40 },
		{ 13, 456, 40 },
		{ 13, 457, 40 },
		{ 13, 458, 40 },
		{ 13, 459, 40 },
		{ 13, 460, 40 },
		{ 13, 461, 40 },
		{ 13, 462, 40 },
		{ 13, 463, 40 },
		{ 13, 464, 40 },
		{ 13, 465, 40 },
		{ 13, 466, 40 },
		{ 13, 467, 40 },
		{ 13, 468, 40 },
		{ 13, 469, 40 },
		{ 13, 470, 40 },
		{ 14, 409, 40 },
		{ 14, 410, 40 },
		{ 14, 411, 40 },
		{ 14, 412, 40 },
		{ 14, 413, 40 },
		{ 14, 414, 40 },
		{ 14, 415, 40 },
		{ 14, 416, 40 },
		{ 14, 417, 40 },
		{ 14, 418, 40 },
		{ 14, 419, 40 },
		{ 14, 420, 40 },
		{ 14, 421, 40 },
		{ 14, 422, 40 },
		{ 14, 423, 40 },
		{ 14, 424, 40 },
		{ 14, 425, 40 },
		{ 14, 426, 40 },
		{ 14, 427, 40 },
		{ 14, 428, 40 },
		{ 14, 429, 40 },
		{ 14, 430, 40 },
		{ 14, 431, 40 },
		{ 14, 432, 40 },
		{ 14, 433, 40 },
		{ 14, 434, 40 },
		{ 14, 435, 40 },
		{ 14, 436, 40 },
		{ 14, 437, 40 },
		{ 14, 438, 40 },
		{ 14, 439, 40 },
		{ 14, 440, 40 },
		{ 14, 441, 40 },
		{ 14, 442, 40 },
		{ 14, 443, 40 },
		{ 14, 444, 40 },
		{ 14, 445, 40 },
		{ 14, 446, 40 },
		{ 14, 447, 40 },
		{ 14, 448, 40 },
		{ 14, 449, 40 },
		{ 14, 450, 40 },
		{ 14, 451, 40 },
		{ 14, 452, 40 },
		{ 14, 453, 40 },
		{ 14, 454, 40 },
		{ 14, 455, 40 },
		{ 14, 456, 40 },
		{ 14, 457, 40 },
		{ 14, 458, 40 },
		{ 14, 459, 40 },
		{ 14, 460, 40 },
		{ 14, 461, 40 },
		{ 14, 462, 40 },
		{ 14, 463, 40 },
		{ 14, 464, 40 },
		{ 14, 465, 40 },
		{ 14, 466, 40 },
		{ 14, 467, 40 },
		{ 14, 468, 40 },
		{ 14, 469, 40 },
		{ 15, 409, 40 },
		{ 15, 410, 40 },
		{ 15, 411, 40 },
		{ 15, 412, 40 },
		{ 15, 413, 40 },
		{ 15, 414, 40 },
		{ 15, 415, 40 },
		{ 15, 416, 40 },
		{ 15, 417, 40 },
		{ 15, 418, 40 },
		{ 15, 419, 40 },
		{ 15, 420, 40 },
		{ 15, 421, 40 },
		{ 15, 422, 40 },
		{ 15, 423, 40 },
		{ 15, 424, 40 },
		{ 15, 425, 40 },
		{ 15, 426, 40 },
		{ 15, 427, 40 },
		{ 15, 428, 40 },
		{ 15, 429, 40 },
		{ 15, 430, 40 },
		{ 15, 431, 40 },
		{ 15, 432, 40 },
		{ 15, 433, 40 },
		{ 15, 434, 40 },
		{ 15, 435, 40 },
		{ 15, 436, 40 },
		{ 15, 437, 40 },
		{ 15, 438, 40 },
		{ 15, 439, 40 },
		{ 15, 440, 40 },
		{ 15, 441, 40 },
		{ 15, 442, 40 },
		{ 15, 443, 40 },
		{ 15, 444, 40 },
		{ 15, 445, 40 },
		{ 15, 446, 40 },
		{ 15, 447, 40 },
		{ 15, 448, 40 },
		{ 15, 449, 40 },
		{ 15, 450, 40 },
		{ 15, 451, 40 },
		{ 15, 452, 40 },
		{ 15, 453, 40 },
		{ 15, 454, 40 },
		{ 15, 455, 40 },
		{ 15, 456, 40 },
		{ 15, 457, 40 },
		{ 15, 458, 40 },
		{ 15, 459, 40 },
		{ 15, 460, 40 },
		{ 15, 461, 40 },
		{ 15, 462, 40 },
		{ 15, 463, 40 },
		{ 15, 464, 40 },
		{ 15, 465, 40 },
		{ 15, 466, 40 },
		{ 15, 467, 40 },
		{ 15, 468, 40 },
		{ 16, 408, 40 },
		{ 16, 409, 40 },
		{ 16, 410, 40 },
		{ 16, 411, 40 },
		{ 16, 412, 40 },
		{ 16, 413, 40 },
		{ 16, 414, 40 },
		{ 16, 415, 40 },
		{ 16, 416, 40 },
		{ 16, 417, 40 },
		{ 16, 418, 40 },
		{ 16, 419, 40 },
		{ 16, 420, 40 },
		{ 16, 421, 40 },
		{ 16, 422, 40 },
		{ 16, 423, 40 },
		{ 16, 424, 40 },
		{ 16, 425, 40 },
		{ 16, 426, 40 },
		{ 16, 427, 40 },
		{ 16, 428, 40 },
		{ 16, 429, 40 },
		{ 16, 430, 40 },
		{ 16, 431, 40 },
		{ 16, 432, 40 },
		{ 16, 433, 40 },
		{ 16, 434, 40 },
		{ 16, 435, 40 },
		{ 16, 436, 40 },
		{ 16, 437, 40 },
		{ 16, 438, 40 },
		{ 16, 439, 40 },
		{ 16, 440, 40 },
		{ 16, 441, 40 },
		{ 16, 442, 40 },
		{ 16, 443, 40 },
		{ 16, 444, 40 },
		{ 16, 445, 40 },
		{ 16, 446, 40 },
		{ 16, 447, 40 },
		{ 16, 448, 40 },
		{ 16, 449, 40 },
		{ 16, 450, 40 },
		{ 16, 451, 40 },
		{ 16, 452, 40 },
		{ 16, 453, 40 },
		{ 16, 454, 40 },
		{ 16, 455, 40 },
		{ 16, 456, 40 },
		{ 16, 457, 40 },
		{ 16, 458, 40 },
		{ 16, 459, 40 },
		{ 16, 460, 40 },
		{ 16, 461, 40 },
		{ 16, 462, 40 },
		{ 16, 463, 40 },
		{ 16, 464, 40 },
		{ 16, 465, 40 },
		{ 16, 466, 40 },
		{ 16, 467, 40 },
		{ 17, 408, 40 },
		{ 17, 409, 40 },
		{ 17, 410, 40 },
		{ 17, 411, 40 },
		{ 17, 412, 40 },
		{ 17, 413, 40 },
		{ 17, 414, 40 },
		{ 17, 415, 40 },
		{ 17, 416, 40 },
		{ 17, 417, 40 },
		{ 17, 418, 40 },
		{ 17, 419, 40 },
		{ 17, 420, 40 },
		{ 17, 421, 40 },
		{ 17, 422, 40 },
		{ 17, 423, 40 },
		{ 17, 424, 40 },
		{ 17, 425, 40 },
		{ 17, 426, 40 },
		{ 17, 427, 40 },
		{ 17, 428, 40 },
		{ 17, 429, 40 },
		{ 17, 430, 40 },
		{ 17, 431, 40 },
		{ 17, 432, 40 },
		{ 17, 433, 40 },
		{ 17, 434, 40 },
		{ 17, 435, 40 },
		{ 17, 436, 40 },
		{ 17, 437, 40 },
		{ 17, 438, 40 },
		{ 17, 439, 40 },
		{ 17, 440, 40 },
		{ 17, 441, 40 },
		{ 17, 442, 40 },
		{ 17, 443, 40 },
		{ 17, 444, 40 },
		{ 17, 445, 40 },
		{ 17, 446, 40 },
		{ 17, 447, 40 },
		{ 17, 448, 40 },
		{ 17, 449, 40 },
		{ 17, 450, 40 },
		{ 17, 451, 40 },
		{ 17, 452, 40 },
		{ 17, 453, 40 },
		{ 17, 454, 40 },
		{ 17, 455, 40 },
		{ 17, 456, 40 },
		{ 17, 457, 40 },
		{ 17, 458, 40 },
		{ 17, 459, 40 },
		{ 17, 460, 40 },
		{ 17, 461, 40 },
		{ 17, 462, 40 },
		{ 17, 463, 40 },
		{ 17, 464, 40 },
		{ 17, 465, 40 },
		{ 17, 466, 40 },
		{ 18, 407, 40 },
		{ 18, 408, 40 },
		{ 18, 409, 40 },
		{ 18, 410, 40 },
		{ 18, 411, 40 },
		{ 18, 412, 40 },
		{ 18, 413, 40 },
		{ 18, 414, 40 },
		{ 18, 415, 40 },
		{ 18, 416, 40 },
		{ 18, 417, 40 },
		{ 18, 418, 40 },
		{ 18, 419, 40 },
		{ 18, 420, 40 },
		{ 18, 421, 40 },
		{ 18, 422, 40 },
		{ 18, 423, 40 },
		{ 18, 424, 40 },
		{ 18, 425, 40 },
		{ 18, 426, 40 },
		{ 18, 427, 40 },
		{ 18, 428, 40 },
		{ 18, 429, 40 },
		{ 18, 430, 40 },
		{ 18, 431, 40 },
		{ 18, 432, 40 },
		{ 18, 433, 40 },
		{ 18, 434, 40 },
		{ 18, 435, 40 },
		{ 18, 436, 40 },
		{ 18, 437, 40 },
		{ 18, 438, 40 },
		{ 18, 439, 40 },
		{ 18, 440, 40 },
		{ 18, 441, 40 },
		{ 18, 442, 40 },
		{ 18, 443, 40 },
		{ 18, 444, 40 },
		{ 18, 445, 40 },
		{ 18, 446, 40 },
		{ 18, 447, 40 },
		{ 18, 448, 40 },
		{ 18, 449, 40 },
		{ 18, 450, 40 },
		{ 18, 451, 40 },
		{ 18, 452, 40 },
		{ 18, 453, 40 },
		{ 18, 454, 40 },
		{ 18, 455, 40 },
		{ 18, 456, 40 },
		{ 18, 457, 40 },
		{ 18, 458, 40 },
		{ 18, 459, 40 },
		{ 18, 460, 40 },
		{ 18, 461, 40 },
		{ 18, 462, 40 },
		{ 18, 463, 40 },
		{ 18, 464, 40 },
		{ 18, 465, 40 },
		{ 18, 466, 40 },
		{ 19, 406, 40 },
		{ 19, 407, 40 },
		{ 19, 408, 40 },
		{ 19, 409, 40 },
		{ 19, 410, 40 },
		{ 19, 411, 40 },
		{ 19, 412, 40 },
		{ 19, 413, 40 },
		{ 19, 414, 40 },
		{ 19, 415, 40 },
		{ 19, 416, 40 },
		{ 19, 417, 40 },
		{ 19, 418, 40 },
		{ 19, 419, 40 },
		{ 19, 420, 40 },
		{ 19, 421, 40 },
		{ 19, 422, 40 },
		{ 19, 423, 40 },
		{ 19, 424, 40 },
		{ 19, 425, 40 },
		{ 19, 426, 40 },
		{ 19, 427, 40 },
		{ 19, 428, 40 },
		{ 19, 429, 40 },
		{ 19, 430, 40 },
		{ 19, 431, 40 },
		{ 19, 432, 40 },
		{ 19, 433, 40 },
		{ 19, 434, 40 },
		{ 19, 435, 40 },
		{ 19, 436, 40 },
		{ 19, 437, 40 },
		{ 19, 438, 40 },
		{ 19, 439, 40 },
		{ 19, 440, 40 },
		{ 19, 441, 40 },
		{ 19, 442, 40 },
		{ 19, 443, 40 },
		{ 19, 444, 40 },
		{ 19, 445, 40 },
		{ 19, 446, 40 },
		{ 19, 447, 40 },
		{ 19, 448, 40 },
		{ 19, 449, 40 },
		{ 19, 450, 40 },
		{ 19, 451, 40 },
		{ 19, 452, 40 },
		{ 19, 453, 40 },
		{ 19, 454, 40 },
		{ 19, 455, 40 },
		{ 19, 456, 40 },
		{ 19, 457, 40 },
		{ 19, 458, 40 },
		{ 19, 459, 40 },
		{ 19, 460, 40 },
		{ 19, 461, 40 },
		{ 19, 462, 40 },
		{ 19, 463, 40 },
		{ 19, 464, 40 },
		{ 19, 465, 40 },
		{ 20, 406, 40 },
		{ 20, 407, 40 },
		{ 20, 408, 40 },
		{ 20, 409, 40 },
		{ 20, 410, 40 },
		{ 20, 411, 40 },
		{ 20, 412, 40 },
		{ 20, 413, 40 },
		{ 20, 414, 40 },
		{ 20, 415, 40 },
		{ 20, 416, 40 },
		{ 20, 417, 40 },
		{ 20, 418, 40 },
		{ 20, 419, 40 },
		{ 20, 420, 40 },
		{ 20, 421, 40 },
		{ 20, 422, 40 },
		{ 20, 423, 40 },
		{ 20, 424, 40 },
		{ 20, 425, 40 },
		{ 20, 426, 40 },
		{ 20, 427, 40 },
		{ 20, 428, 40 },
		{ 20, 429, 40 },
		{ 20, 430, 40 },
		{ 20, 431, 40 },
		{ 20, 432, 40 },
		{ 20, 433, 40 },
		{ 20, 434, 40 },
		{ 20, 435, 40 },
		{ 20, 436, 40 },
		{ 20, 437, 40 },
		{ 20, 438, 40 },
		{ 20, 439, 40 },
		{ 20, 440, 40 },
		{ 20, 441, 40 },
		{ 20, 442, 40 },
		{ 20, 443, 40 },
		{ 20, 444, 40 },
		{ 20, 445, 40 },
		{ 20, 446, 40 },
		{ 20, 447, 40 },
		{ 20, 448, 40 },
		{ 20, 449, 40 },
		{ 20, 450, 40 },
		{ 20, 451, 40 },
		{ 20, 452, 40 },
		{ 20, 453, 40 },
		{ 20, 454, 40 },
		{ 20, 455, 40 },
		{ 20, 456, 40 },
		{ 20, 457, 40 },
		{ 20, 458, 40 },
		{ 20, 459, 40 },
		{ 20, 460, 40 },
		{ 20, 461, 40 },
		{ 20, 462, 40 },
		{ 20, 463, 40 },
		{ 20, 464, 40 },
		{ 21, 405, 40 },
		{ 21, 406, 40 },
		{ 21, 407, 40 },
		{ 21, 408, 40 },
		{ 21, 409, 40 },
		{ 21, 410, 40 },
		{ 21, 411, 40 },
		{ 21, 412, 40 },
		{ 21, 413, 40 },
		{ 21, 414, 40 },
		{ 21, 415, 40 },
		{ 21, 416, 40 },
		{ 21, 417, 40 },
		{ 21, 418, 40 },
		{ 21, 419, 40 },
		{ 21, 420, 40 },
		{ 21, 421, 40 },
		{ 21, 422, 40 },
		{ 21, 423, 40 },
		{ 21, 424, 40 },
		{ 21, 425, 40 },
		{ 21, 426, 40 },
		{ 21, 427, 40 },
		{ 21, 428, 40 },
		{ 21, 429, 40 },
		{ 21, 430, 40 },
		{ 21, 431, 40 },
		{ 21, 432, 40 },
		{ 21, 433, 40 },
		{ 21, 434, 40 },
		{ 21, 435, 40 },
		{ 21, 436, 40 },
		{ 21, 437, 40 },
		{ 21, 438, 40 },
		{ 21, 439, 40 },
		{ 21, 440, 40 },
		{ 21, 441, 40 },
		{ 21, 442, 40 },
		{ 21, 443, 40 },
		{ 21, 444, 40 },
		{ 21, 445, 40 },
		{ 21, 446, 40 },
		{ 21, 447, 40 },
		{ 21, 448, 40 },
		{ 21, 449, 40 },
		{ 21, 450, 40 },
		{ 21, 451, 40 },
		{ 21, 452, 40 },
		{ 21, 453, 40 },
		{ 21, 454, 40 },
		{ 21, 455, 40 },
		{ 21, 456, 40 },
		{ 21, 457, 40 },
		{ 21, 458, 40 },
		{ 21, 459, 40 },
		{ 21, 460, 40 },
		{ 21, 461, 40 },
		{ 21, 462, 40 },
		{ 21, 463, 40 },
		{ 22, 404, 40 },
		{ 22, 405, 40 },
		{ 22, 406, 40 },
		{ 22, 407, 40 },
		{ 22, 408, 40 },
		{ 22, 409, 40 },
		{ 22, 410, 40 },
		{ 22, 411, 40 },
		{ 22, 412, 40 },
		{ 22, 413, 40 },
		{ 22, 414, 40 },
		{ 22, 415, 40 },
		{ 22, 416, 40 },
		{ 22, 417, 40 },
		{ 22, 418, 40 },
		{ 22, 419, 40 },
		{ 22, 420, 40 },
		{ 22, 421, 40 },
		{ 22, 422, 40 },
		{ 22, 423, 40 },
		{ 22, 424, 40 },
		{ 22, 425, 40 },
		{ 22, 426, 40 },
		{ 22, 427, 40 },
		{ 22, 428, 40 },
		{ 22, 429, 40 },
		{ 22, 430, 40 },
		{ 22, 431, 40 },
		{ 22, 432, 40 },
		{ 22, 433, 40 },
		{ 22, 434, 40 },
		{ 22, 435, 40 },
		{ 22, 436, 40 },
		{ 22, 437, 40 },
		{ 22, 438, 40 },
		{ 22, 439, 40 },
		{ 22, 440, 40 },
		{ 22, 441, 40 },
		{ 22, 442, 40 },
		{ 22, 443, 40 },
		{ 22, 444, 40 },
		{ 22, 445, 40 },
		{ 22, 446, 40 },
		{ 22, 447, 40 },
		{ 22, 448, 40 },
		{ 22, 449, 40 },
		{ 22, 450, 40 },
		{ 22, 451, 40 },
		{ 22, 452, 40 },
		{ 22, 453, 40 },
		{ 22, 454, 40 },
		{ 22, 455, 40 },
		{ 22, 456, 40 },
		{ 22, 457, 40 },
		{ 22, 458, 40 },
		{ 22, 459, 40 },
		{ 22, 460, 40 },
		{ 22, 461, 40 },
		{ 22, 462, 40 },
		{ 23, 404, 40 },
		{ 23, 405, 40 },
		{ 23, 406, 40 },
		{ 23, 407, 40 },
		{ 23, 408, 40 },
		{ 23, 409, 40 },
		{ 23, 410, 40 },
		{ 23, 411, 40 },
		{ 23, 412, 40 },
		{ 23, 413, 40 },
		{ 23, 414, 40 },
		{ 23, 415, 40 },
		{ 23, 416, 40 },
		{ 23, 417, 40 },
		{ 23, 418, 40 },
		{ 23, 419, 40 },
		{ 23, 420, 40 },
		{ 23, 421, 40 },
		{ 23, 422, 40 },
		{ 23, 423, 40 },
		{ 23, 424, 40 },
		{ 23, 425, 40 },
		{ 23, 426, 40 },
		{ 23, 427, 40 },
		{ 23, 428, 40 },
		{ 23, 429, 40 },
		{ 23, 430, 40 },
		{ 23, 431, 40 },
		{ 23, 432, 40 },
		{ 23, 433, 40 },
		{ 23, 434, 40 },
		{ 23, 435, 40 },
		{ 23, 436, 40 },
		{ 23, 437, 40 },
		{ 23, 438, 40 },
		{ 23, 439, 40 },
		{ 23, 440, 40 },
		{ 23, 441, 40 },
		{ 23, 442, 40 },
		{ 23, 443, 40 },
		{ 23, 444, 40 },
		{ 23, 445, 40 },
		{ 23, 446, 40 },
		{ 23, 447, 40 },
		{ 23, 448, 40 },
		{ 23, 449, 40 },
		{ 23, 450, 40 },
		{ 23, 451, 40 },
		{ 23, 452, 40 },
		{ 23, 453, 40 },
		{ 23, 454, 40 },
		{ 23, 455, 40 },
		{ 23, 456, 40 },
		{ 23, 457, 40 },
		{ 23, 458, 40 },
		{ 23, 459, 40 },
		{ 23, 460, 40 },
		{ 23, 461, 40 },
		{ 24, 403, 40 },
		{ 24, 404, 40 },
		{ 24, 405, 40 },
		{ 24, 406, 40 },
		{ 24, 407, 40 },
		{ 24, 408, 40 },
		{ 24, 409, 40 },
		{ 24, 410, 40 },
		{ 24, 411, 40 },
		{ 24, 412, 40 },
		{ 24, 413, 40 },
		{ 24, 414, 40 },
		{ 24, 415, 40 },
		{ 24, 416, 40 },
		{ 24, 417, 40 },
		{ 24, 418, 40 },
		{ 24, 419, 40 },
		{ 24, 420, 40 },
		{ 24, 421, 40 },
		{ 24, 422, 40 },
		{ 24, 423, 40 },
		{ 24, 424, 40 },
		{ 24, 425, 40 },
		{ 24, 426, 40 },
		{ 24, 427, 40 },
		{ 24, 428, 40 },
		{ 24, 429, 40 },
		{ 24, 430, 40 },
		{ 24, 431, 40 },
		{ 24, 432, 40 },
		{ 24, 433, 40 },
		{ 24, 434, 40 },
		{ 24, 435, 40 },
		{ 24, 436, 40 },
		{ 24, 437, 40 },
		{ 24, 438, 40 },
		{ 24, 439, 40 },
		{ 24, 440, 40 },
		{ 24, 441, 40 },
		{ 24, 442, 40 },
		{ 24, 443, 40 },
		{ 24, 444, 40 },
		{ 24, 445, 40 },
		{ 24, 446, 40 },
		{ 24, 447, 40 },
		{ 24, 448, 40 },
		{ 24, 449, 40 },
		{ 24, 450, 40 },
		{ 24, 451, 40 },
		{ 24, 452, 40 },
		{ 24, 453, 40 },
		{ 24, 454, 40 },
		{ 24, 455, 40 },
		{ 24, 456, 40 },
		{ 24, 457, 40 },
		{ 24, 458, 40 },
		{ 24, 459, 40 },
		{ 24, 460, 40 },
		{ 25, 403, 40 },
		{ 25, 404, 40 },
		{ 25, 405, 40 },
		{ 25, 406, 40 },
		{ 25, 407, 40 },
		{ 25, 408, 40 },
		{ 25, 409, 40 },
		{ 25, 410, 40 },
		{ 25, 411, 40 },
		{ 25, 412, 40 },
		{ 25, 413, 40 },
		{ 25, 414, 40 },
		{ 25, 415, 40 },
		{ 25, 416, 40 },
		{ 25, 417, 40 },
		{ 25, 418, 40 },
		{ 25, 419, 40 },
		{ 25, 420, 40 },
		{ 25, 421, 40 },
		{ 25, 422, 40 },
		{ 25, 423, 40 },
		{ 25, 424, 40 },
		{ 25, 425, 40 },
		{ 25, 426, 40 },
		{ 25, 427, 40 },
		{ 25, 428, 40 },
		{ 25, 429, 40 },
		{ 25, 430, 40 },
		{ 25, 431, 40 },
		{ 25, 432, 40 },
		{ 25, 433, 40 },
		{ 25, 434, 40 },
		{ 25, 435, 40 },
		{ 25, 436, 40 },
		{ 25, 437, 40 },
		{ 25, 438, 40 },
		{ 25, 439, 40 },
		{ 25, 440, 40 },
		{ 25, 441, 40 },
		{ 25, 442, 40 },
		{ 25, 443, 40 },
		{ 25, 444, 40 },
		{ 25, 445, 40 },
		{ 25, 446, 40 },
		{ 25, 447, 40 },
		{ 25, 448, 40 },
		{ 25, 449, 40 },
		{ 25, 450, 40 },
		{ 25, 451, 40 },
		{ 25, 452, 40 },
		{ 25, 453, 40 },
		{ 25, 454, 40 },
		{ 25, 455, 40 },
		{ 25, 456, 40 },
		{ 25, 457, 40 },
		{ 25, 458, 40 },
		{ 25, 459, 40 },
		{ 26, 402, 40 },
		{ 26, 403, 40 },
		{ 26, 404, 40 },
		{ 26, 405, 40 },
		{ 26, 406, 40 },
		{ 26, 407, 40 },
		{ 26, 408, 40 },
		{ 26, 409, 40 },
		{ 26, 410, 40 },
		{ 26, 411, 40 },
		{ 26, 412, 40 },
		{ 26, 413, 40 },
		{ 26, 414, 40 },
		{ 26, 415, 40 },
		{ 26, 416, 40 },
		{ 26, 417, 40 },
		{ 26, 418, 40 },
		{ 26, 419, 40 },
		{ 26, 420, 40 },
		{ 26, 421, 40 },
		{ 26, 422, 40 },
		{ 26, 423, 40 },
		{ 26, 424, 40 },
		{ 26, 425, 40 },
		{ 26, 426, 40 },
		{ 26, 427, 40 },
		{ 26, 428, 40 },
		{ 26, 429, 40 },
		{ 26, 430, 40 },
		{ 26, 431, 40 },
		{ 26, 432, 40 },
		{ 26, 433, 40 },
		{ 26, 434, 40 },
		{ 26, 435, 40 },
		{ 26, 436, 40 },
		{ 26, 437, 40 },
		{ 26, 438, 40 },
		{ 26, 439, 40 },
		{ 26, 440, 40 },
		{ 26, 441, 40 },
		{ 26, 442, 40 },
		{ 26, 443, 40 },
		{ 26, 444, 40 },
		{ 26, 445, 40 },
		{ 26, 446, 40 },
		{ 26, 447, 40 },
		{ 26, 448, 40 },
		{ 26, 449, 40 },
		{ 26, 450, 40 },
		{ 26, 451, 40 },
		{ 26, 452, 40 },
		{ 26, 453, 40 },
		{ 26, 454, 40 },
		{ 26, 455, 40 },
		{ 26, 456, 40 },
		{ 26, 457, 40 },
		{ 26, 458, 40 },
		{ 26, 459, 40 },
		{ 27, 401, 40 },
		{ 27, 402, 40 },
		{ 27, 403, 40 },
		{ 27, 404, 40 },
		{ 27, 405, 40 },
		{ 27, 406, 40 },
		{ 27, 407, 40 },
		{ 27, 408, 40 },
		{ 27, 409, 40 },
		{ 27, 410, 40 },
		{ 27, 411, 40 },
		{ 27, 412, 40 },
		{ 27, 413, 40 },
		{ 27, 414, 40 },
		{ 27, 415, 40 },
		{ 27, 416, 40 },
		{ 27, 417, 40 },
		{ 27, 418, 40 },
		{ 27, 419, 40 },
		{ 27, 420, 40 },
		{ 27, 421, 40 },
		{ 27, 422, 40 },
		{ 27, 423, 40 },
		{ 27, 424, 40 },
		{ 27, 425, 40 },
		{ 27, 426, 40 },
		{ 27, 427, 40 },
		{ 27, 428, 40 },
		{ 27, 429, 40 },
		{ 27, 430, 40 },
		{ 27, 431, 40 },
		{ 27, 432, 40 },
		{ 27, 433, 40 },
		{ 27, 434, 40 },
		{ 27, 435, 40 },
		{ 27, 436, 40 },
		{ 27, 437, 40 },
		{ 27, 438, 40 },
		{ 27, 439, 40 },
		{ 27, 440, 40 },
		{ 27, 441, 40 },
		{ 27, 442, 40 },
		{ 27, 443, 40 },
		{ 27, 444, 40 },
		{ 27, 445, 40 },
		{ 27, 446, 40 },
		{ 27, 447, 40 },
		{ 27, 448, 40 },
		{ 27, 449, 40 },
		{ 27, 450, 40 },
		{ 27, 451, 40 },
		{ 27, 452, 40 },
		{ 27, 453, 40 },
		{ 27, 454, 40 },
		{ 27, 455, 40 },
		{ 27, 456, 40 },
		{ 27, 457, 40 },
		{ 27, 458, 40 },
		{ 28, 401, 40 },
		{ 28, 402, 40 },
		{ 28, 403, 40 },
		{ 28, 404, 40 },
		{ 28, 405, 40 },
		{ 28, 406, 40 },
		{ 28, 407, 40 },
		{ 28, 408, 40 },
		{ 28, 409, 40 },
		{ 28, 410, 40 },
		{ 28, 411, 40 },
		{ 28, 412, 40 },
		{ 28, 413, 40 },
		{ 28, 414, 40 },
		{ 28, 415, 40 },
		{ 28, 416, 40 },
		{ 28, 417, 40 },
		{ 28, 418, 40 },
		{ 28, 419, 40 },
		{ 28, 420, 40 },
		{ 28, 421, 40 },
		{ 28, 422, 40 },
		{ 28, 423, 40 },
		{ 28, 424, 40 },
		{ 28, 425, 40 },
		{ 28, 426, 40 },
		{ 28, 427, 40 },
		{ 28, 428, 40 },
		{ 28, 429, 40 },
		{ 28, 430, 40 },
		{ 28, 431, 40 },
		{ 28, 432, 40 },
		{ 28, 433, 40 },
		{ 28, 434, 40 },
		{ 28, 435, 40 },
		{ 28, 436, 40 },
		{ 28, 437, 40 },
		{ 28, 438, 40 },
		{ 28, 439, 40 },
		{ 28, 440, 40 },
		{ 28, 441, 40 },
		{ 28, 442, 40 },
		{ 28, 443, 40 },
		{ 28, 444, 40 },
		{ 28, 445, 40 },
		{ 28, 446, 40 },
		{ 28, 447, 40 },
		{ 28, 448, 40 },
		{ 28, 449, 40 },
		{ 28, 450, 40 },
		{ 28, 451, 40 },
		{ 28, 452, 40 },
		{ 28, 453, 40 },
		{ 28, 454, 40 },
		{ 28, 455, 40 },
		{ 28, 456, 40 },
		{ 28, 457, 40 },
		{ 29, 400, 40 },
		{ 29, 401, 40 },
		{ 29, 402, 40 },
		{ 29, 403, 40 },
		{ 29, 404, 40 },
		{ 29, 405, 40 },
		{ 29, 406, 40 },
		{ 29, 407, 40 },
		{ 29, 408, 40 },
		{ 29, 409, 40 },
		{ 29, 410, 40 },
		{ 29, 411, 40 },
		{ 29, 412, 40 },
		{ 29, 413, 40 },
		{ 29, 414, 40 },
		{ 29, 415, 40 },
		{ 29, 416, 40 },
		{ 29, 417, 40 },
		{ 29, 418, 40 },
		{ 29, 419, 40 },
		{ 29, 420, 40 },
		{ 29, 421, 40 },
		{ 29, 422, 40 },
		{ 29, 423, 40 },
		{ 29, 424, 40 },
		{ 29, 425, 40 },
		{ 29, 426, 40 },
		{ 29, 427, 40 },
		{ 29, 428, 40 },
		{ 29, 429, 40 },
		{ 29, 430, 40 },
		{ 29, 431, 40 },
		{ 29, 432, 40 },
		{ 29, 433, 40 },
		{ 29, 434, 40 },
		{ 29, 435, 40 },
		{ 29, 436, 40 },
		{ 29, 437, 40 },
		{ 29, 438, 40 },
		{ 29, 439, 40 },
		{ 29, 440, 40 },
		{ 29, 441, 40 },
		{ 29, 442, 40 },
		{ 29, 443, 40 },
		{ 29, 444, 40 },
		{ 29, 445, 40 },
		{ 29, 446, 40 },
		{ 29, 447, 40 },
		{ 29, 448, 40 },
		{ 29, 449, 40 },
		{ 29, 450, 40 },
		{ 29, 451, 40 },
		{ 29, 452, 40 },
		{ 29, 453, 40 },
		{ 29, 454, 40 },
		{ 29, 455, 40 },
		{ 29, 456, 40 },
		{ 30, 399, 40 },
		{ 30, 400, 40 },
		{ 30, 401, 40 },
		{ 30, 402, 40 },
		{ 30, 403, 40 },
		{ 30, 404, 40 },
		{ 30, 405, 40 },
		{ 30, 406, 40 },
		{ 30, 407, 40 },
		{ 30, 408, 40 },
		{ 30, 409, 40 },
		{ 30, 410, 40 },
		{ 30, 411, 40 },
		{ 30, 412, 40 },
		{ 30, 413, 40 },
		{ 30, 414, 40 },
		{ 30, 415, 40 },
		{ 30, 416, 40 },
		{ 30, 417, 40 },
		{ 30, 418, 40 },
		{ 30, 419, 40 },
		{ 30, 420, 40 },
		{ 30, 421, 40 },
		{ 30, 422, 40 },
		{ 30, 423, 40 },
		{ 30, 424, 40 },
		{ 30, 425, 40 },
		{ 30, 426, 40 },
		{ 30, 427, 40 },
		{ 30, 428, 40 },
		{ 30, 429, 40 },
		{ 30, 430, 40 },
		{ 30, 431, 40 },
		{ 30, 432, 40 },
		{ 30, 433, 40 },
		{ 30, 434, 40 },
		{ 30, 435, 40 },
		{ 30, 436, 40 },
		{ 30, 437, 40 },
		{ 30, 438, 40 },
		{ 30, 439, 40 },
		{ 30, 440, 40 },
		{ 30, 441, 40 },
		{ 30, 442, 40 },
		{ 30, 443, 40 },
		{ 30, 444, 40 },
		{ 30, 445, 40 },
		{ 30, 446, 40 },
		{ 30, 447, 40 },
		{ 30, 448, 40 },
		{ 30, 449, 40 },
		{ 30, 450, 40 },
		{ 30, 451, 40 },
		{ 30, 452, 40 },
		{ 30, 453, 40 },
		{ 30, 454, 40 },
		{ 30, 455, 40 },
		{ 31, 399, 40 },
		{ 31, 400, 40 },
		{ 31, 401, 40 },
		{ 31, 402, 40 },
		{ 31, 403, 40 },
		{ 31, 404, 40 },
		{ 31, 405, 40 },
		{ 31, 406, 40 },
		{ 31, 407, 40 },
		{ 31, 408, 40 },
		{ 31, 409, 40 },
		{ 31, 410, 40 },
		{ 31, 411, 40 },
		{ 31, 412, 40 },
		{ 31, 413, 40 },
		{ 31, 414, 40 },
		{ 31, 415, 40 },
		{ 31, 416, 40 },
		{ 31, 417, 40 },
		{ 31, 418, 40 },
		{ 31, 419, 40 },
		{ 31, 420, 40 },
		{ 31, 421, 40 },
		{ 31, 422, 40 },
		{ 31, 423, 40 },
		{ 31, 424, 40 },
		{ 31, 425, 40 },
		{ 31, 426, 40 },
		{ 31, 427, 40 },
		{ 31, 428, 40 },
		{ 31, 429, 40 },
		{ 31, 430, 40 },
		{ 31, 431, 40 },
		{ 31, 432, 40 },
		{ 31, 433, 40 },
		{ 31, 434, 40 },
		{ 31, 435, 40 },
		{ 31, 436, 40 },
		{ 31, 437, 40 },
		{ 31, 438, 40 },
		{ 31, 439, 40 },
		{ 31, 440, 40 },
		{ 31, 441, 40 },
		{ 31, 442, 40 },
		{ 31, 443, 40 },
		{ 31, 444, 40 },
		{ 31, 445, 40 },
		{ 31, 446, 40 },
		{ 31, 447, 40 },
		{ 31, 448, 40 },
		{ 31, 449, 40 },
		{ 31, 450, 40 },
		{ 31, 451, 40 },
		{ 31, 452, 40 },
		{ 31, 453, 40 },
		{ 31, 454, 40 },
		{ 32, 398, 40 },
		{ 32, 399, 40 },
		{ 32, 400, 40 },
		{ 32, 401, 40 },
		{ 32, 402, 40 },
		{ 32, 403, 40 },
		{ 32, 404, 40 },
		{ 32, 405, 40 },
		{ 32, 406, 40 },
		{ 32, 407, 40 },
		{ 32, 408, 40 },
		{ 32, 409, 40 },
		{ 32, 410, 40 },
		{ 32, 411, 40 },
		{ 32, 412, 40 },
		{ 32, 413, 40 },
		{ 32, 414, 40 },
		{ 32, 415, 40 },
		{ 32, 416, 40 },
		{ 32, 417, 40 },
		{ 32, 418, 40 },
		{ 32, 419, 40 },
		{ 32, 420, 40 },
		{ 32, 421, 40 },
		{ 32, 422, 40 },
		{ 32, 423, 40 },
		{ 32, 424, 40 },
		{ 32, 425, 40 },
		{ 32, 426, 40 },
		{ 32, 427, 40 },
		{ 32, 428, 40 },
		{ 32, 429, 40 },
		{ 32, 430, 40 },
		{ 32, 431, 40 },
		{ 32, 432, 40 },
		{ 32, 433, 40 },
		{ 32, 434, 40 },
		{ 32, 435, 40 },
		{ 32, 436, 40 },
		{ 32, 437, 40 },
		{ 32, 438, 40 },
		{ 32, 439, 40 },
		{ 32, 440, 40 },
		{ 32, 441, 40 },
		{ 32, 442, 40 },
		{ 32, 443, 40 },
		{ 32, 444, 40 },
		{ 32, 445, 40 },
		{ 32, 446, 40 },
		{ 32, 447, 40 },
		{ 32, 448, 40 },
		{ 32, 449, 40 },
		{ 32, 450, 40 },
		{ 32, 451, 40 },
		{ 32, 452, 40 },
		{ 32, 453, 40 },
		{ 33, 398, 40 },
		{ 33, 399, 40 },
		{ 33, 400, 40 },
		{ 33, 401, 40 },
		{ 33, 402, 40 },
		{ 33, 403, 40 },
		{ 33, 404, 40 },
		{ 33, 405, 40 },
		{ 33, 406, 40 },
		{ 33, 407, 40 },
		{ 33, 408, 40 },
		{ 33, 409, 40 },
		{ 33, 410, 40 },
		{ 33, 411, 40 },
		{ 33, 412, 40 },
		{ 33, 413, 40 },
		{ 33, 414, 40 },
		{ 33, 415, 40 },
		{ 33, 416, 40 },
		{ 33, 417, 40 },
		{ 33, 418, 40 },
		{ 33, 419, 40 },
		{ 33, 420, 40 },
		{ 33, 421, 40 },
		{ 33, 422, 40 },
		{ 33, 423, 40 },
		{ 33, 424, 40 },
		{ 33, 425, 40 },
		{ 33, 426, 40 },
		{ 33, 427, 40 },
		{ 33, 428, 40 },
		{ 33, 429, 40 },
		{ 33, 430, 40 },
		{ 33, 431, 40 },
		{ 33, 432, 40 },
		{ 33, 433, 40 },
		{ 33, 434, 40 },
		{ 33, 435, 40 },
		{ 33, 436, 40 },
		{ 33, 437, 40 },
		{ 33, 438, 40 },
		{ 33, 439, 40 },
		{ 33, 440, 40 },
		{ 33, 441, 40 },
		{ 33, 442, 40 },
		{ 33, 443, 40 },
		{ 33, 444, 40 },
		{ 33, 445, 40 },
		{ 33, 446, 40 },
		{ 33, 447, 40 },
		{ 33, 448, 40 },
		{ 33, 449, 40 },
		{ 33, 450, 40 },
		{ 33, 451, 40 },
		{ 33, 452, 40 },
		{ 34, 397, 40 },
		{ 34, 398, 40 },
		{ 34, 399, 40 },
		{ 34, 400, 40 },
		{ 34, 401, 40 },
		{ 34, 402, 40 },
		{ 34, 403, 40 },
		{ 34, 404, 40 },
		{ 34, 405, 40 },
		{ 34, 406, 40 },
		{ 34, 407, 40 },
		{ 34, 408, 40 },
		{ 34, 409, 40 },
		{ 34, 410, 40 },
		{ 34, 411, 40 },
		{ 34, 412, 40 },
		{ 34, 413, 40 },
		{ 34, 414, 40 },
		{ 34, 415, 40 },
		{ 34, 416, 40 },
		{ 34, 417, 40 },
		{ 34, 418, 40 },
		{ 34, 419, 40 },
		{ 34, 420, 40 },
		{ 34, 421, 40 },
		{ 34, 422, 40 },
		{ 34, 423, 40 },
		{ 34, 424, 40 },
		{ 34, 425, 40 },
		{ 34, 426, 40 },
		{ 34, 427, 40 },
		{ 34, 428, 40 },
		{ 34, 429, 40 },
		{ 34, 430, 40 },
		{ 34, 431, 40 },
		{ 34, 432, 40 },
		{ 34, 433, 40 },
		{ 34, 434, 40 },
		{ 34, 435, 40 },
		{ 34, 436, 40 },
		{ 34, 437, 40 },
		{ 34, 438, 40 },
		{ 34, 439, 40 },
		{ 34, 440, 40 },
		{ 34, 441, 40 },
		{ 34, 442, 40 },
		{ 34, 443, 40 },
		{ 34, 444, 40 },
		{ 34, 445, 40 },
		{ 34, 446, 40 },
		{ 34, 447, 40 },
		{ 34, 448, 40 },
		{ 34, 449, 40 },
		{ 34, 450, 40 },
		{ 34, 451, 40 },
		{ 34, 452, 40 },
		{ 35, 396, 40 },
		{ 35, 397, 40 },
		{ 35, 398, 40 },
		{ 35, 399, 40 },
		{ 35, 400, 40 },
		{ 35, 401, 40 },
		{ 35, 402, 40 },
		{ 35, 403, 40 },
		{ 35, 404, 40 },
		{ 35, 405, 40 },
		{ 35, 406, 40 },
		{ 35, 407, 40 },
		{ 35, 408, 40 },
		{ 35, 409, 40 },
		{ 35, 410, 40 },
		{ 35, 411, 40 },
		{ 35, 412, 40 },
		{ 35, 413, 40 },
		{ 35, 414, 40 },
		{ 35, 415, 40 },
		{ 35, 416, 40 },
		{ 35, 417, 40 },
		{ 35, 418, 40 },
		{ 35, 419, 40 },
		{ 35, 420, 40 },
		{ 35, 421, 40 },
		{ 35, 422, 40 },
		{ 35, 423, 40 },
		{ 35, 424, 40 },
		{ 35, 425, 40 },
		{ 35, 426, 40 },
		{ 35, 427, 40 },
		{ 35, 428, 40 },
		{ 35, 429, 40 },
		{ 35, 430, 40 },
		{ 35, 431, 40 },
		{ 35, 432, 40 },
		{ 35, 433, 40 },
		{ 35, 434, 40 },
		{ 35, 435, 40 },
		{ 35, 436, 40 },
		{ 35, 437, 40 },
		{ 35, 438, 40 },
		{ 35, 439, 40 },
		{ 35, 440, 40 },
		{ 35, 441, 40 },
		{ 35, 442, 40 },
		{ 35, 443, 40 },
		{ 35, 444, 40 },
		{ 35, 445, 40 },
		{ 35, 446, 40 },
		{ 35, 447, 40 },
		{ 35, 448, 40 },
		{ 35, 449, 40 },
		{ 35, 450, 40 },
		{ 35, 451, 40 },
		{ 36, 396, 40 },
		{ 36, 397, 40 },
		{ 36, 398, 40 },
		{ 36, 399, 40 },
		{ 36, 400, 40 },
		{ 36, 401, 40 },
		{ 36, 402, 40 },
		{ 36, 403, 40 },
		{ 36, 404, 40 },
		{ 36, 405, 40 },
		{ 36, 406, 40 },
		{ 36, 407, 40 },
		{ 36, 408, 40 },
		{ 36, 409, 40 },
		{ 36, 410, 40 },
		{ 36, 411, 40 },
		{ 36, 412, 40 },
		{ 36, 413, 40 },
		{ 36, 414, 40 },
		{ 36, 415, 40 },
		{ 36, 416, 40 },
		{ 36, 417, 40 },
		{ 36, 418, 40 },
		{ 36, 419, 40 },
		{ 36, 420, 40 },
		{ 36, 421, 40 },
		{ 36, 422, 40 },
		{ 36, 423, 40 },
		{ 36, 424, 40 },
		{ 36, 425, 40 },
		{ 36, 426, 40 },
		{ 36, 427, 40 },
		{ 36, 428, 40 },
		{ 36, 429, 40 },
		{ 36, 430, 40 },
		{ 36, 431, 40 },
		{ 36, 432, 40 },
		{ 36, 433, 40 },
		{ 36, 434, 40 },
		{ 36, 435, 40 },
		{ 36, 436, 40 },
		{ 36, 437, 40 },
		{ 36, 438, 40 },
		{ 36, 439, 40 },
		{ 36, 440, 40 },
		{ 36, 441, 40 },
		{ 36, 442, 40 },
		{ 36, 443, 40 },
		{ 36, 444, 40 },
		{ 36, 445, 40 },
		{ 36, 446, 40 },
		{ 36, 447, 40 },
		{ 36, 448, 40 },
		{ 36, 449, 40 },
		{ 36, 450, 40 },
		{ 37, 395, 40 },
		{ 37, 396, 40 },
		{ 37, 397, 40 },
		{ 37, 398, 40 },
		{ 37, 399, 40 },
		{ 37, 400, 40 },
		{ 37, 401, 40 },
		{ 37, 402, 40 },
		{ 37, 403, 40 },
		{ 37, 404, 40 },
		{ 37, 405, 40 },
		{ 37, 406, 40 },
		{ 37, 407, 40 },
		{ 37, 408, 40 },
		{ 37, 409, 40 },
		{ 37, 410, 40 },
		{ 37, 411, 40 },
		{ 37, 412, 40 },
		{ 37, 413, 40 },
		{ 37, 414, 40 },
		{ 37, 415, 40 },
		{ 37, 416, 40 },
		{ 37, 417, 40 },
		{ 37, 418, 40 },
		{ 37, 419, 40 },
		{ 37, 420, 40 },
		{ 37, 421, 40 },
		{ 37, 422, 40 },
		{ 37, 423, 40 },
		{ 37, 424, 40 },
		{ 37, 425, 40 },
		{ 37, 426, 40 },
		{ 37, 427, 40 },
		{ 37, 428, 40 },
		{ 37, 429, 40 },
		{ 37, 430, 40 },
		{ 37, 431, 40 },
		{ 37, 432, 40 },
		{ 37, 433, 40 },
		{ 37, 434, 40 },
		{ 37, 435, 40 },
		{ 37, 436, 40 },
		{ 37, 437, 40 },
		{ 37, 438, 40 },
		{ 37, 439, 40 },
		{ 37, 440, 40 },
		{ 37, 441, 40 },
		{ 37, 442, 40 },
		{ 37, 443, 40 },
		{ 37, 444, 40 },
		{ 37, 445, 40 },
		{ 37, 446, 40 },
		{ 37, 447, 40 },
		{ 37, 448, 40 },
		{ 37, 449, 40 },
		{ 38, 394, 40 },
		{ 38, 395, 40 },
		{ 38, 396, 40 },
		{ 38, 397, 40 },
		{ 38, 398, 40 },
		{ 38, 399, 40 },
		{ 38, 400, 40 },
		{ 38, 401, 40 },
		{ 38, 402, 40 },
		{ 38, 403, 40 },
		{ 38, 404, 40 },
		{ 38, 405, 40 },
		{ 38, 406, 40 },
		{ 38, 407, 40 },
		{ 38, 408, 40 },
		{ 38, 409, 40 },
		{ 38, 410, 40 },
		{ 38, 411, 40 },
		{ 38, 412, 40 },
		{ 38, 413, 40 },
		{ 38, 414, 40 },
		{ 38, 415, 40 },
		{ 38, 416, 40 },
		{ 38, 417, 40 },
		{ 38, 418, 40 },
		{ 38, 419, 40 },
		{ 38, 420, 40 },
		{ 38, 421, 40 },
		{ 38, 422, 40 },
		{ 38, 423, 40 },
		{ 38, 424, 40 },
		{ 38, 425, 40 },
		{ 38, 426, 40 },
		{ 38, 427, 40 },
		{ 38, 428, 40 },
		{ 38, 429, 40 },
		{ 38, 430, 40 },
		{ 38, 431, 40 },
		{ 38, 432, 40 },
		{ 38, 433, 40 },
		{ 38, 434, 40 },
		{ 38, 435, 40 },
		{ 38, 436, 40 },
		{ 38, 437, 40 },
		{ 38, 438, 40 },
		{ 38, 439, 40 },
		{ 38, 440, 40 },
		{ 38, 441, 40 },
		{ 38, 442, 40 },
		{ 38, 443, 40 },
		{ 38, 444, 40 },
		{ 38, 445, 40 },
		{ 38, 446, 40 },
		{ 38, 447, 40 },
		{ 38, 448, 40 },
		{ 39, 394, 40 },
		{ 39, 395, 40 },
		{ 39, 396, 40 },
		{ 39, 397, 40 },
		{ 39, 398, 40 },
		{ 39, 399, 40 },
		{ 39, 400, 40 },
		{ 39, 401, 40 },
		{ 39, 402, 40 },
		{ 39, 403, 40 },
		{ 39, 404, 40 },
		{ 39, 405, 40 },
		{ 39, 406, 40 },
		{ 39, 407, 40 },
		{ 39, 408, 40 },
		{ 39, 409, 40 },
		{ 39, 410, 40 },
		{ 39, 411, 40 },
		{ 39, 412, 40 },
		{ 39, 413, 40 },
		{ 39, 414, 40 },
		{ 39, 415, 40 },
		{ 39, 416, 40 },
		{ 39, 417, 40 },
		{ 39, 418, 40 },
		{ 39, 419, 40 },
		{ 39, 420, 40 },
		{ 39, 421, 40 },
		{ 39, 422, 40 },
		{ 39, 423, 40 },
		{ 39, 424, 40 },
		{ 39, 425, 40 },
		{ 39, 426, 40 },
		{ 39, 427, 40 },
		{ 39, 428, 40 },
		{ 39, 429, 40 },
		{ 39, 430, 40 },
		{ 39, 431, 40 },
		{ 39, 432, 40 },
		{ 39, 433, 40 },
		{ 39, 434, 40 },
		{ 39, 435, 40 },
		{ 39, 436, 40 },
		{ 39, 437, 40 },
		{ 39, 438, 40 },
		{ 39, 439, 40 },
		{ 39, 440, 40 },
		{ 39, 441, 40 },
		{ 39, 442, 40 },
		{ 39, 443, 40 },
		{ 39, 444, 40 },
		{ 39, 445, 40 },
		{ 39, 446, 40 },
		{ 39, 447, 40 },
		{ 40, 393, 40 },
		{ 40, 394, 40 },
		{ 40, 395, 40 },
		{ 40, 396, 40 },
		{ 40, 397, 40 },
		{ 40, 398, 40 },
		{ 40, 399, 40 },
		{ 40, 400, 40 },
		{ 40, 401, 40 },
		{ 40, 402, 40 },
		{ 40, 403, 40 },
		{ 40, 404, 40 },
		{ 40, 405, 40 },
		{ 40, 406, 40 },
		{ 40, 407, 40 },
		{ 40, 408, 40 },
		{ 40, 409, 40 },
		{ 40, 410, 40 },
		{ 40, 411, 40 },
		{ 40, 412, 40 },
		{ 40, 413, 40 },
		{ 40, 414, 40 },
		{ 40, 415, 40 },
		{ 40, 416, 40 },
		{ 40, 417, 40 },
		{ 40, 418, 40 },
		{ 40, 419, 40 },
		{ 40, 420, 40 },
		{ 40, 421, 40 },
		{ 40, 422, 40 },
		{ 40, 423, 40 },
		{ 40, 424, 40 },
		{ 40, 425, 40 },
		{ 40, 426, 40 },
		{ 40, 427, 40 },
		{ 40, 428, 40 },
		{ 40, 429, 40 },
		{ 40, 430, 40 },
		{ 40, 431, 40 },
		{ 40, 432, 40 },
		{ 40, 433, 40 },
		{ 40, 434, 40 },
		{ 40, 435, 40 },
		{ 40, 436, 40 },
		{ 40, 437, 40 },
		{ 40, 438, 40 },
		{ 40, 439, 40 },
		{ 40, 440, 40 },
		{ 40, 441, 40 },
		{ 40, 442, 40 },
		{ 40, 443, 40 },
		{ 40, 444, 40 },
		{ 40, 445, 40 },
		{ 40, 446, 40 },
		{ 41, 393, 40 },
		{ 41, 394, 40 },
		{ 41, 395, 40 },
		{ 41, 396, 40 },
		{ 41, 397, 40 },
		{ 41, 398, 40 },
		{ 41, 399, 40 },
		{ 41, 400, 40 },
		{ 41, 401, 40 },
		{ 41, 402, 40 },
		{ 41, 403, 40 },
		{ 41, 404, 40 },
		{ 41, 405, 40 },
		{ 41, 406, 40 },
		{ 41, 407, 40 },
		{ 41, 408, 40 },
		{ 41, 409, 40 },
		{ 41, 410, 40 },
		{ 41, 411, 40 },
		{ 41, 412, 40 },
		{ 41, 413, 40 },
		{ 41, 414, 40 },
		{ 41, 415, 40 },
		{ 41, 416, 40 },
		{ 41, 417, 40 },
		{ 41, 418, 40 },
		{ 41, 419, 40 },
		{ 41, 420, 40 },
		{ 41, 421, 40 },
		{ 41, 422, 40 },
		{ 41, 423, 40 },
		{ 41, 424, 40 },
		{ 41, 425, 40 },
		{ 41, 426, 40 },
		{ 41, 427, 40 },
		{ 41, 428, 40 },
		{ 41, 429, 40 },
		{ 41, 430, 40 },
		{ 41, 431, 40 },
		{ 41, 432, 40 },
		{ 41, 433, 40 },
		{ 41, 434, 40 },
		{ 41, 435, 40 },
		{ 41, 436, 40 },
		{ 41, 437, 40 },
		{ 41, 438, 40 },
		{ 41, 439, 40 },
		{ 41, 440, 40 },
		{ 41, 441, 40 },
		{ 41, 442, 40 },
		{ 41, 443, 40 },
		{ 41, 444, 40 },
		{ 41, 445, 40 },
		{ 42, 392, 40 },
		{ 42, 393, 40 },
		{ 42, 394, 40 },
		{ 42, 395, 40 },
		{ 42, 396, 40 },
		{ 42, 397, 40 },
		{ 42, 398, 40 },
		{ 42, 399, 40 },
		{ 42, 400, 40 },
		{ 42, 401, 40 },
		{ 42, 402, 40 },
		{ 42, 403, 40 },
		{ 42, 404, 40 },
		{ 42, 405, 40 },
		{ 42, 406, 40 },
		{ 42, 407, 40 },
		{ 42, 408, 40 },
		{ 42, 409, 40 },
		{ 42, 410, 40 },
		{ 42, 411, 40 },
		{ 42, 412, 40 },
		{ 42, 413, 40 },
		{ 42, 414, 40 },
		{ 42, 415, 40 },
		{ 42, 416, 40 },
		{ 42, 417, 40 },
		{ 42, 418, 40 },
		{ 42, 419, 40 },
		{ 42, 420, 40 },
		{ 42, 421, 40 },
		{ 42, 422, 40 },
		{ 42, 423, 40 },
		{ 42, 424, 40 },
		{ 42, 425, 40 },
		{ 42, 426, 40 },
		{ 42, 427, 40 },
		{ 42, 428, 40 },
		{ 42, 429, 40 },
		{ 42, 430, 40 },
		{ 42, 431, 40 },
		{ 42, 432, 40 },
		{ 42, 433, 40 },
		{ 42, 434, 40 },
		{ 42, 435, 40 },
		{ 42, 436, 40 },
		{ 42, 437, 40 },
		{ 42, 438, 40 },
		{ 42, 439, 40 },
		{ 42, 440, 40 },
		{ 42, 441, 40 },
		{ 42, 442, 40 },
		{ 42, 443, 40 },
		{ 42, 444, 40 },
		{ 42, 445, 40 },
		{ 43, 391, 40 },
		{ 43, 392, 40 },
		{ 43, 393, 40 },
		{ 43, 394, 40 },
		{ 43, 395, 40 },
		{ 43, 396, 40 },
		{ 43, 397, 40 },
		{ 43, 398, 40 },
		{ 43, 399, 40 },
		{ 43, 400, 40 },
		{ 43, 401, 40 },
		{ 43, 402, 40 },
		{ 43, 403, 40 },
		{ 43, 404, 40 },
		{ 43, 405, 40 },
		{ 43, 406, 40 },
		{ 43, 407, 40 },
		{ 43, 408, 40 },
		{ 43, 409, 40 },
		{ 43, 410, 40 },
		{ 43, 411, 40 },
		{ 43, 412, 40 },
		{ 43, 413, 40 },
		{ 43, 414, 40 },
		{ 43, 415, 40 },
		{ 43, 416, 40 },
		{ 43, 417, 40 },
		{ 43, 418, 40 },
		{ 43, 419, 40 },
		{ 43, 420, 40 },
		{ 43, 421, 40 },
		{ 43, 422, 40 },
		{ 43, 423, 40 },
		{ 43, 424, 40 },
		{ 43, 425, 40 },
		{ 43, 426, 40 },
		{ 43, 427, 40 },
		{ 43, 428, 40 },
		{ 43, 429, 40 },
		{ 43, 430, 40 },
		{ 43, 431, 40 },
		{ 43, 432, 40 },
		{ 43, 433, 40 },
		{ 43, 434, 40 },
		{ 43, 435, 40 },
		{ 43, 436, 40 },
		{ 43, 437, 40 },
		{ 43, 438, 40 },
		{ 43, 439, 40 },
		{ 43, 440, 40 },
		{ 43, 441, 40 },
		{ 43, 442, 40 },
		{ 43, 443, 40 },
		{ 43, 444, 40 },
		{ 44, 391, 40 },
		{ 44, 392, 40 },
		{ 44, 393, 40 },
		{ 44, 394, 40 },
		{ 44, 395, 40 },
		{ 44, 396, 40 },
		{ 44, 397, 40 },
		{ 44, 398, 40 },
		{ 44, 399, 40 },
		{ 44, 400, 40 },
		{ 44, 401, 40 },
		{ 44, 402, 40 },
		{ 44, 403, 40 },
		{ 44, 404, 40 },
		{ 44, 405, 40 },
		{ 44, 406, 40 },
		{ 44, 407, 40 },
		{ 44, 408, 40 },
		{ 44, 409, 40 },
		{ 44, 410, 40 },
		{ 44, 411, 40 },
		{ 44, 412, 40 },
		{ 44, 413, 40 },
		{ 44, 414, 40 },
		{ 44, 415, 40 },
		{ 44, 416, 40 },
		{ 44, 417, 40 },
		{ 44, 418, 40 },
		{ 44, 419, 40 },
		{ 44, 420, 40 },
		{ 44, 421, 40 },
		{ 44, 422, 40 },
		{ 44, 423, 40 },
		{ 44, 424, 40 },
		{ 44, 425, 40 },
		{ 44, 426, 40 },
		{ 44, 427, 40 },
		{ 44, 428, 40 },
		{ 44, 429, 40 },
		{ 44, 430, 40 },
		{ 44, 431, 40 },
		{ 44, 432, 40 },
		{ 44, 433, 40 },
		{ 44, 434, 40 },
		{ 44, 435, 40 },
		{ 44, 436, 40 },
		{ 44, 437, 40 },
		{ 44, 438, 40 },
		{ 44, 439, 40 },
		{ 44, 440, 40 },
		{ 44, 441, 40 },
		{ 44, 442, 40 },
		{ 44, 443, 40 },
		{ 45, 390, 40 },
		{ 45, 391, 40 },
		{ 45, 392, 40 },
		{ 45, 393, 40 },
		{ 45, 394, 40 },
		{ 45, 395, 40 },
		{ 45, 396, 40 },
		{ 45, 397, 40 },
		{ 45, 398, 40 },
		{ 45, 399, 40 },
		{ 45, 400, 40 },
		{ 45, 401, 40 },
		{ 45, 402, 40 },
		{ 45, 403, 40 },
		{ 45, 404, 40 },
		{ 45, 405, 40 },
		{ 45, 406, 40 },
		{ 45, 407, 40 },
		{ 45, 408, 40 },
		{ 45, 409, 40 },
		{ 45, 410, 40 },
		{ 45, 411, 40 },
		{ 45, 412, 40 },
		{ 45, 413, 40 },
		{ 45, 414, 40 },
		{ 45, 415, 40 },
		{ 45, 416, 40 },
		{ 45, 417, 40 },
		{ 45, 418, 40 },
		{ 45, 419, 40 },
		{ 45, 420, 40 },
		{ 45, 421, 40 },
		{ 45, 422, 40 },
		{ 45, 423, 40 },
		{ 45, 424, 40 },
		{ 45, 425, 40 },
		{ 45, 426, 40 },
		{ 45, 427, 40 },
		{ 45, 428, 40 },
		{ 45, 429, 40 },
		{ 45, 430, 40 },
		{ 45, 431, 40 },
		{ 45, 432, 40 },
		{ 45, 433, 40 },
		{ 45, 434, 40 },
		{ 45, 435, 40 },
		{ 45, 436, 40 },
		{ 45, 437, 40 },
		{ 45, 438, 40 },
		{ 45, 439, 40 },
		{ 45, 440, 40 },
		{ 45, 441, 40 },
		{ 45, 442, 40 },
		{ 46, 389, 40 },
		{ 46, 390, 40 },
		{ 46, 391, 40 },
		{ 46, 392, 40 },
		{ 46, 393, 40 },
		{ 46, 394, 40 },
		{ 46, 395, 40 },
		{ 46, 396, 40 },
		{ 46, 397, 40 },
		{ 46, 398, 40 },
		{ 46, 399, 40 },
		{ 46, 400, 40 },
		{ 46, 401, 40 },
		{ 46, 402, 40 },
		{ 46, 403, 40 },
		{ 46, 404, 40 },
		{ 46, 405, 40 },
		{ 46, 406, 40 },
		{ 46, 407, 40 },
		{ 46, 408, 40 },
		{ 46, 409, 40 },
		{ 46, 410, 40 },
		{ 46, 411, 40 },
		{ 46, 412, 40 },
		{ 46, 413, 40 },
		{ 46, 414, 40 },
		{ 46, 415, 40 },
		{ 46, 416, 40 },
		{ 46, 417, 40 },
		{ 46, 418, 40 },
		{ 46, 419, 40 },
		{ 46, 420, 40 },
		{ 46, 421, 40 },
		{ 46, 422, 40 },
		{ 46, 423, 40 },
		{ 46, 424, 40 },
		{ 46, 425, 40 },
		{ 46, 426, 40 },
		{ 46, 427, 40 },
		{ 46, 428, 40 },
		{ 46, 429, 40 },
		{ 46, 430, 40 },
		{ 46, 431, 40 },
		{ 46, 432, 40 },
		{ 46, 433, 40 },
		{ 46, 434, 40 },
		{ 46, 435, 40 },
		{ 46, 436, 40 },
		{ 46, 437, 40 },
		{ 46, 438, 40 },
		{ 46, 439, 40 },
		{ 46, 440, 40 },
		{ 46, 441, 40 },
		{ 47, 389, 40 },
		{ 47, 390, 40 },
		{ 47, 391, 40 },
		{ 47, 392, 40 },
		{ 47, 393, 40 },
		{ 47, 394, 40 },
		{ 47, 395, 40 },
		{ 47, 396, 40 },
		{ 47, 397, 40 },
		{ 47, 398, 40 },
		{ 47, 399, 40 },
		{ 47, 400, 40 },
		{ 47, 401, 40 },
		{ 47, 402, 40 },
		{ 47, 403, 40 },
		{ 47, 404, 40 },
		{ 47, 405, 40 },
		{ 47, 406, 40 },
		{ 47, 407, 40 },
		{ 47, 408, 40 },
		{ 47, 409, 40 },
		{ 47, 410, 40 },
		{ 47, 411, 40 },
		{ 47, 412, 40 },
		{ 47, 413, 40 },
		{ 47, 414, 40 },
		{ 47, 415, 40 },
		{ 47, 416, 40 },
		{ 47, 417, 40 },
		{ 47, 418, 40 },
		{ 47, 419, 40 },
		{ 47, 420, 40 },
		{ 47, 421, 40 },
		{ 47, 422, 40 },
		{ 47, 423, 40 },
		{ 47, 424, 40 },
		{ 47, 425, 40 },
		{ 47, 426, 40 },
		{ 47, 427, 40 },
		{ 47, 428, 40 },
		{ 47, 429, 40 },
		{ 47, 430, 40 },
		{ 47, 431, 40 },
		{ 47, 432, 40 },
		{ 47, 433, 40 },
		{ 47, 434, 40 },
		{ 47, 435, 40 },
		{ 47, 436, 40 },
		{ 47, 437, 40 },
		{ 47, 438, 40 },
		{ 47, 439, 40 },
		{ 47, 440, 40 },
		{ 48, 388, 40 },
		{ 48, 389, 40 },
		{ 48, 390, 40 },
		{ 48, 391, 40 },
		{ 48, 392, 40 },
		{ 48, 393, 40 },
		{ 48, 394, 40 },
		{ 48, 395, 40 },
		{ 48, 396, 40 },
		{ 48, 397, 40 },
		{ 48, 398, 40 },
		{ 48, 399, 40 },
		{ 48, 400, 40 },
		{ 48, 401, 40 },
		{ 48, 402, 40 },
		{ 48, 403, 40 },
		{ 48, 404, 40 },
		{ 48, 405, 40 },
		{ 48, 406, 40 },
		{ 48, 407, 40 },
		{ 48, 408, 40 },
		{ 48, 409, 40 },
		{ 48, 410, 40 },
		{ 48, 411, 40 },
		{ 48, 412, 40 },
		{ 48, 413, 40 },
		{ 48, 414, 40 },
		{ 48, 415, 40 },
		{ 48, 416, 40 },
		{ 48, 417, 40 },
		{ 48, 418, 40 },
		{ 48, 419, 40 },
		{ 48, 420, 40 },
		{ 48, 421, 40 },
		{ 48, 422, 40 },
		{ 48, 423, 40 },
		{ 48, 424, 40 },
		{ 48, 425, 40 },
		{ 48, 426, 40 },
		{ 48, 427, 40 },
		{ 48, 428, 40 },
		{ 48, 429, 40 },
		{ 48, 430, 40 },
		{ 48, 431, 40 },
		{ 48, 432, 40 },
		{ 48, 433, 40 },
		{ 48, 434, 40 },
		{ 48, 435, 40 },
		{ 48, 436, 40 },
		{ 48, 437, 40 },
		{ 48, 438, 40 },
		{ 48, 439, 40 },
		{ 49, 388, 40 },
		{ 49, 389, 40 },
		{ 49, 390, 40 },
		{ 49, 391, 40 },
		{ 49, 392, 40 },
		{ 49, 393, 40 },
		{ 49, 394, 40 },
		{ 49, 395, 40 },
		{ 49, 396, 40 },
		{ 49, 397, 40 },
		{ 49, 398, 40 },
		{ 49, 399, 40 },
		{ 49, 400, 40 },
		{ 49, 401, 40 },
		{ 49, 402, 40 },
		{ 49, 403, 40 },
		{ 49, 404, 40 },
		{ 49, 405, 40 },
		{ 49, 406, 40 },
		{ 49, 407, 40 },
		{ 49, 408, 40 },
		{ 49, 409, 40 },
		{ 49, 410, 40 },
		{ 49, 411, 40 },
		{ 49, 412, 40 },
		{ 49, 413, 40 },
		{ 49, 414, 40 },
		{ 49, 415, 40 },
		{ 49, 416, 40 },
		{ 49, 417, 40 },
		{ 49, 418, 40 },
		{ 49, 419, 40 },
		{ 49, 420, 40 },
		{ 49, 421, 40 },
		{ 49, 422, 40 },
		{ 49, 423, 40 },
		{ 49, 424, 40 },
		{ 49, 425, 40 },
		{ 49, 426, 40 },
		{ 49, 427, 40 },
		{ 49, 428, 40 },
		{ 49, 429, 40 },
		{ 49, 430, 40 },
		{ 49, 431, 40 },
		{ 49, 432, 40 },
		{ 49, 433, 40 },
		{ 49, 434, 40 },
		{ 49, 435, 40 },
		{ 49, 436, 40 },
		{ 49, 437, 40 },
		{ 49, 438, 40 },
		{ 50, 387, 40 },
		{ 50, 388, 40 },
		{ 50, 389, 40 },
		{ 50, 390, 40 },
		{ 50, 391, 40 },
		{ 50, 392, 40 },
		{ 50, 393, 40 },
		{ 50, 394, 40 },
		{ 50, 395, 40 },
		{ 50, 396, 40 },
		{ 50, 397, 40 },
		{ 50, 398, 40 },
		{ 50, 399, 40 },
		{ 50, 400, 40 },
		{ 50, 401, 40 },
		{ 50, 402, 40 },
		{ 50, 403, 40 },
		{ 50, 404, 40 },
		{ 50, 405, 40 },
		{ 50, 406, 40 },
		{ 50, 407, 40 },
		{ 50, 408, 40 },
		{ 50, 409, 40 },
		{ 50, 410, 40 },
		{ 50, 411, 40 },
		{ 50, 412, 40 },
		{ 50, 413, 40 },
		{ 50, 414, 40 },
		{ 50, 415, 40 },
		{ 50, 416, 40 },
		{ 50, 417, 40 },
		{ 50, 418, 40 },
		{ 50, 419, 40 },
		{ 50, 420, 40 },
		{ 50, 421, 40 },
		{ 50, 422, 40 },
		{ 50, 423, 40 },
		{ 50, 424, 40 },
		{ 50, 425, 40 },
		{ 50, 426, 40 },
		{ 50, 427, 40 },
		{ 50, 428, 40 },
		{ 50, 429, 40 },
		{ 50, 430, 40 },
		{ 50, 431, 40 },
		{ 50, 432, 40 },
		{ 50, 433, 40 },
		{ 50, 434, 40 },
		{ 50, 435, 40 },
		{ 50, 436, 40 },
		{ 50, 437, 40 },
		{ 50, 438, 40 },
		{ 51, 386, 40 },
		{ 51, 387, 40 },
		{ 51, 388, 40 },
		{ 51, 389, 40 },
		{ 51, 390, 40 },
		{ 51, 391, 40 },
		{ 51, 392, 40 },
		{ 51, 393, 40 },
		{ 51, 394, 40 },
		{ 51, 395, 40 },
		{ 51, 396, 40 },
		{ 51, 397, 40 },
		{ 51, 398, 40 },
		{ 51, 399, 40 },
		{ 51, 400, 40 },
		{ 51, 401, 40 },
		{ 51, 402, 40 },
		{ 51, 403, 40 },
		{ 51, 404, 40 },
		{ 51, 405, 40 },
		{ 51, 406, 40 },
		{ 51, 407, 40 },
		{ 51, 408, 40 },
		{ 51, 409, 40 },
		{ 51, 410, 40 },
		{ 51, 411, 40 },
		{ 51, 412, 40 },
		{ 51, 413, 40 },
		{ 51, 414, 40 },
		{ 51, 415, 40 },
		{ 51, 416, 40 },
		{ 51, 417, 40 },
		{ 51, 418, 40 },
		{ 51, 419, 40 },
		{ 51, 420, 40 },
		{ 51, 421, 40 },
		{ 51, 422, 40 },
		{ 51, 423, 40 },
		{ 51, 424, 40 },
		{ 51, 425, 40 },
		{ 51, 426, 40 },
		{ 51, 427, 40 },
		{ 51, 428, 40 },
		{ 51, 429, 40 },
		{ 51, 430, 40 },
		{ 51, 431, 40 },
		{ 51, 432, 40 },
		{ 51, 433, 40 },
		{ 51, 434, 40 },
		{ 51, 435, 40 },
		{ 51, 436, 40 },
		{ 51, 437, 40 },
		{ 52, 386, 40 },
		{ 52, 387, 40 },
		{ 52, 388, 40 },
		{ 52, 389, 40 },
		{ 52, 390, 40 },
		{ 52, 391, 40 },
		{ 52, 392, 40 },
		{ 52, 393, 40 },
		{ 52, 394, 40 },
		{ 52, 395, 40 },
		{ 52, 396, 40 },
		{ 52, 397, 40 },
		{ 52, 398, 40 },
		{ 52, 399, 40 },
		{ 52, 400, 40 },
		{ 52, 401, 40 },
		{ 52, 402, 40 },
		{ 52, 403, 40 },
		{ 52, 404, 40 },
		{ 52, 405, 40 },
		{ 52, 406, 40 },
		{ 52, 407, 40 },
		{ 52, 408, 40 },
		{ 52, 409, 40 },
		{ 52, 410, 40 },
		{ 52, 411, 40 },
		{ 52, 412, 40 },
		{ 52, 413, 40 },
		{ 52, 414, 40 },
		{ 52, 415, 40 },
		{ 52, 416, 40 },
		{ 52, 417, 40 },
		{ 52, 418, 40 },
		{ 52, 419, 40 },
		{ 52, 420, 40 },
		{ 52, 421, 40 },
		{ 52, 422, 40 },
		{ 52, 423, 40 },
		{ 52, 424, 40 },
		{ 52, 425, 40 },
		{ 52, 426, 40 },
		{ 52, 427, 40 },
		{ 52, 428, 40 },
		{ 52, 429, 40 },
		{ 52, 430, 40 },
		{ 52, 431, 40 },
		{ 52, 432, 40 },
		{ 52, 433, 40 },
		{ 52, 434, 40 },
		{ 52, 435, 40 },
		{ 52, 436, 40 },
		{ 53, 385, 40 },
		{ 53, 386, 40 },
		{ 53, 387, 40 },
		{ 53, 388, 40 },
		{ 53, 389, 40 },
		{ 53, 390, 40 },
		{ 53, 391, 40 },
		{ 53, 392, 40 },
		{ 53, 393, 40 },
		{ 53, 394, 40 },
		{ 53, 395, 40 },
		{ 53, 396, 40 },
		{ 53, 397, 40 },
		{ 53, 398, 40 },
		{ 53, 399, 40 },
		{ 53, 400, 40 },
		{ 53, 401, 40 },
		{ 53, 402, 40 },
		{ 53, 403, 40 },
		{ 53, 404, 40 },
		{ 53, 405, 40 },
		{ 53, 406, 40 },
		{ 53, 407, 40 },
		{ 53, 408, 40 },
		{ 53, 409, 40 },
		{ 53, 410, 40 },
		{ 53, 411, 40 },
		{ 53, 412, 40 },
		{ 53, 413, 40 },
		{ 53, 414, 40 },
		{ 53, 415, 40 },
		{ 53, 416, 40 },
		{ 53, 417, 40 },
		{ 53, 418, 40 },
		{ 53, 419, 40 },
		{ 53, 420, 40 },
		{ 53, 421, 40 },
		{ 53, 422, 40 },
		{ 53, 423, 40 },
		{ 53, 424, 40 },
		{ 53, 425, 40 },
		{ 53, 426, 40 },
		{ 53, 427, 40 },
		{ 53, 428, 40 },
		{ 53, 429, 40 },
		{ 53, 430, 40 },
		{ 53, 431, 40 },
		{ 53, 432, 40 },
		{ 53, 433, 40 },
		{ 53, 434, 40 },
		{ 53, 435, 40 },
		{ 54, 384, 40 },
		{ 54, 385, 40 },
		{ 54, 386, 40 },
		{ 54, 387, 40 },
		{ 54, 388, 40 },
		{ 54, 389, 40 },
		{ 54, 390, 40 },
		{ 54, 391, 40 },
		{ 54, 392, 40 },
		{ 54, 393, 40 },
		{ 54, 394, 40 },
		{ 54, 395, 40 },
		{ 54, 396, 40 },
		{ 54, 397, 40 },
		{ 54, 398, 40 },
		{ 54, 399, 40 },
		{ 54, 400, 40 },
		{ 54, 401, 40 },
		{ 54, 402, 40 },
		{ 54, 403, 40 },
		{ 54, 404, 40 },
		{ 54, 405, 40 },
		{ 54, 406, 40 },
		{ 54, 407, 40 },
		{ 54, 408, 40 },
		{ 54, 409, 40 },
		{ 54, 410, 40 },
		{ 54, 411, 40 },
		{ 54, 412, 40 },
		{ 54, 413, 40 },
		{ 54, 414, 40 },
		{ 54, 415, 40 },
		{ 54, 416, 40 },
		{ 54, 417, 40 },
		{ 54, 418, 40 },
		{ 54, 419, 40 },
		{ 54, 420, 40 },
		{ 54, 421, 40 },
		{ 54, 422, 40 },
		{ 54, 423, 40 },
		{ 54, 424, 40 },
		{ 54, 425, 40 },
		{ 54, 426, 40 },
		{ 54, 427, 40 },
		{ 54, 428, 40 },
		{ 54, 429, 40 },
		{ 54, 430, 40 },
		{ 54, 431, 40 },
		{ 54, 432, 40 },
		{ 54, 433, 40 },
		{ 54, 434, 40 },
		{ 55, 384, 40 },
		{ 55, 385, 40 },
		{ 55, 386, 40 },
		{ 55, 387, 40 },
		{ 55, 388, 40 },
		{ 55, 389, 40 },
		{ 55, 390, 40 },
		{ 55, 391, 40 },
		{ 55, 392, 40 },
		{ 55, 393, 40 },
		{ 55, 394, 40 },
		{ 55, 395, 40 },
		{ 55, 396, 40 },
		{ 55, 397, 40 },
		{ 55, 398, 40 },
		{ 55, 399, 40 },
		{ 55, 400, 40 },
		{ 55, 401, 40 },
		{ 55, 402, 40 },
		{ 55, 403, 40 },
		{ 55, 404, 40 },
		{ 55, 405, 40 },
		{ 55, 406, 40 },
		{ 55, 407, 40 },
		{ 55, 408, 40 },
		{ 55, 409, 40 },
		{ 55, 410, 40 },
		{ 55, 411, 40 },
		{ 55, 412, 40 },
		{ 55, 413, 40 },
		{ 55, 414, 40 },
		{ 55, 415, 40 },
		{ 55, 416, 40 },
		{ 55, 417, 40 },
		{ 55, 418, 40 },
		{ 55, 419, 40 },
		{ 55, 420, 40 },
		{ 55, 421, 40 },
		{ 55, 422, 40 },
		{ 55, 423, 40 },
		{ 55, 424, 40 },
		{ 55, 425, 40 },
		{ 55, 426, 40 },
		{ 55, 427, 40 },
		{ 55, 428, 40 },
		{ 55, 429, 40 },
		{ 55, 430, 40 },
		{ 55, 431, 40 },
		{ 55, 432, 40 },
		{ 55, 433, 40 },
		{ 56, 383, 40 },
		{ 56, 384, 40 },
		{ 56, 385, 40 },
		{ 56, 386, 40 },
		{ 56, 387, 40 },
		{ 56, 388, 40 },
		{ 56, 389, 40 },
		{ 56, 390, 40 },
		{ 56, 391, 40 },
		{ 56, 392, 40 },
		{ 56, 393, 40 },
		{ 56, 394, 40 },
		{ 56, 395, 40 },
		{ 56, 396, 40 },
		{ 56, 397, 40 },
		{ 56, 398, 40 },
		{ 56, 399, 40 },
		{ 56, 400, 40 },
		{ 56, 401, 40 },
		{ 56, 402, 40 },
		{ 56, 403, 40 },
		{ 56, 404, 40 },
		{ 56, 405, 40 },
		{ 56, 406, 40 },
		{ 56, 407, 40 },
		{ 56, 408, 40 },
		{ 56, 409, 40 },
		{ 56, 410, 40 },
		{ 56, 411, 40 },
		{ 56, 412, 40 },
		{ 56, 413, 40 },
		{ 56, 414, 40 },
		{ 56, 415, 40 },
		{ 56, 416, 40 },
		{ 56, 417, 40 },
		{ 56, 418, 40 },
		{ 56, 419, 40 },
		{ 56, 420, 40 },
		{ 56, 421, 40 },
		{ 56, 422, 40 },
		{ 56, 423, 40 },
		{ 56, 424, 40 },
		{ 56, 425, 40 },
		{ 56, 426, 40 },
		{ 56, 427, 40 },
		{ 56, 428, 40 },
		{ 56, 429, 40 },
		{ 56, 430, 40 },
		{ 56, 431, 40 },
		{ 56, 432, 40 },
		{ 57, 383, 40 },
		{ 57, 384, 40 },
		{ 57, 385, 40 },
		{ 57, 386, 40 },
		{ 57, 387, 40 },
		{ 57, 388, 40 },
		{ 57, 389, 40 },
		{ 57, 390, 40 },
		{ 57, 391, 40 },
		{ 57, 392, 40 },
		{ 57, 393, 40 },
		{ 57, 394, 40 },
		{ 57, 395, 40 },
		{ 57, 396, 40 },
		{ 57, 397, 40 },
		{ 57, 398, 40 },
		{ 57, 399, 40 },
		{ 57, 400, 40 },
		{ 57, 401, 40 },
		{ 57, 402, 40 },
		{ 57, 403, 40 },
		{ 57, 404, 40 },
		{ 57, 405, 40 },
		{ 57, 406, 40 },
		{ 57, 407, 40 },
		{ 57, 408, 40 },
		{ 57, 409, 40 },
		{ 57, 410, 40 },
		{ 57, 411, 40 },
		{ 57, 412, 40 },
		{ 57, 413, 40 },
		{ 57, 414, 40 },
		{ 57, 415, 40 },
		{ 57, 416, 40 },
		{ 57, 417, 40 },
		{ 57, 418, 40 },
		{ 57, 419, 40 },
		{ 57, 420, 40 },
		{ 57, 421, 40 },
		{ 57, 422, 40 },
		{ 57, 423, 40 },
		{ 57, 424, 40 },
		{ 57, 425, 40 },
		{ 57, 426, 40 },
		{ 57, 427, 40 },
		{ 57, 428, 40 },
		{ 57, 429, 40 },
		{ 57, 430, 40 },
		{ 57, 431, 40 },
		{ 58, 382, 40 },
		{ 58, 383, 40 },
		{ 58, 384, 40 },
		{ 58, 385, 40 },
		{ 58, 386, 40 },
		{ 58, 387, 40 },
		{ 58, 388, 40 },
		{ 58, 389, 40 },
		{ 58, 390, 40 },
		{ 58, 391, 40 },
		{ 58, 392, 40 },
		{ 58, 393, 40 },
		{ 58, 394, 40 },
		{ 58, 395, 40 },
		{ 58, 396, 40 },
		{ 58, 397, 40 },
		{ 58, 398, 40 },
		{ 58, 399, 40 },
		{ 58, 400, 40 },
		{ 58, 401, 40 },
		{ 58, 402, 40 },
		{ 58, 403, 40 },
		{ 58, 404, 40 },
		{ 58, 405, 40 },
		{ 58, 406, 40 },
		{ 58, 407, 40 },
		{ 58, 408, 40 },
		{ 58, 409, 40 },
		{ 58, 410, 40 },
		{ 58, 411, 40 },
		{ 58, 412, 40 },
		{ 58, 413, 40 },
		{ 58, 414, 40 },
		{ 58, 415, 40 },
		{ 58, 416, 40 },
		{ 58, 417, 40 },
		{ 58, 418, 40 },
		{ 58, 419, 40 },
		{ 58, 420, 40 },
		{ 58, 421, 40 },
		{ 58, 422, 40 },
		{ 58, 423, 40 },
		{ 58, 424, 40 },
		{ 58, 425, 40 },
		{ 58, 426, 40 },
		{ 58, 427, 40 },
		{ 58, 428, 40 },
		{ 58, 429, 40 },
		{ 58, 430, 40 },
		{ 58, 431, 40 },
		{ 59, 381, 40 },
		{ 59, 382, 40 },
		{ 59, 383, 40 },
		{ 59, 384, 40 },
		{ 59, 385, 40 },
		{ 59, 386, 40 },
		{ 59, 387, 40 },
		{ 59, 388, 40 },
		{ 59, 389, 40 },
		{ 59, 390, 40 },
		{ 59, 391, 40 },
		{ 59, 392, 40 },
		{ 59, 393, 40 },
		{ 59, 394, 40 },
		{ 59, 395, 40 },
		{ 59, 396, 40 },
		{ 59, 397, 40 },
		{ 59, 398, 40 },
		{ 59, 399, 40 },
		{ 59, 400, 40 },
		{ 59, 401, 40 },
		{ 59, 402, 40 },
		{ 59, 403, 40 },
		{ 59, 404, 40 },
		{ 59, 405, 40 },
		{ 59, 406, 40 },
		{ 59, 407, 40 },
		{ 59, 408, 40 },
		{ 59, 409, 40 },
		{ 59, 410, 40 },
		{ 59, 411, 40 },
		{ 59, 412, 40 },
		{ 59, 413, 40 },
		{ 59, 414, 40 },
		{ 59, 415, 40 },
		{ 59, 416, 40 },
		{ 59, 417, 40 },
		{ 59, 418, 40 },
		{ 59, 419, 40 },
		{ 59, 420, 40 },
		{ 59, 421, 40 },
		{ 59, 422, 40 },
		{ 59, 423, 40 },
		{ 59, 424, 40 },
		{ 59, 425, 40 },
		{ 59, 426, 40 },
		{ 59, 427, 40 },
		{ 59, 428, 40 },
		{ 59, 429, 40 },
		{ 59, 430, 40 },
		{ 60, 381, 40 },
		{ 60, 382, 40 },
		{ 60, 383, 40 },
		{ 60, 384, 40 },
		{ 60, 385, 40 },
		{ 60, 386, 40 },
		{ 60, 387, 40 },
		{ 60, 388, 40 },
		{ 60, 389, 40 },
		{ 60, 390, 40 },
		{ 60, 391, 40 },
		{ 60, 392, 40 },
		{ 60, 393, 40 },
		{ 60, 394, 40 },
		{ 60, 395, 40 },
		{ 60, 396, 40 },
		{ 60, 397, 40 },
		{ 60, 398, 40 },
		{ 60, 399, 40 },
		{ 60, 400, 40 },
		{ 60, 401, 40 },
		{ 60, 402, 40 },
		{ 60, 403, 40 },
		{ 60, 404, 40 },
		{ 60, 405, 40 },
		{ 60, 406, 40 },
		{ 60, 407, 40 },
		{ 60, 408, 40 },
		{ 60, 409, 40 },
		{ 60, 410, 40 },
		{ 60, 411, 40 },
		{ 60, 412, 40 },
		{ 60, 413, 40 },
		{ 60, 414, 40 },
		{ 60, 415, 40 },
		{ 60, 416, 40 },
		{ 60, 417, 40 },
		{ 60, 418, 40 },
		{ 60, 419, 40 },
		{ 60, 420, 40 },
		{ 60, 421, 40 },
		{ 60, 422, 40 },
		{ 60, 423, 40 },
		{ 60, 424, 40 },
		{ 60, 425, 40 },
		{ 60, 426, 40 },
		{ 60, 427, 40 },
		{ 60, 428, 40 },
		{ 60, 429, 40 },
		{ 61, 380, 40 },
		{ 61, 381, 40 },
		{ 61, 382, 40 },
		{ 61, 383, 40 },
		{ 61, 384, 40 },
		{ 61, 385, 40 },
		{ 61, 386, 40 },
		{ 61, 387, 40 },
		{ 61, 388, 40 },
		{ 61, 389, 40 },
		{ 61, 390, 40 },
		{ 61, 391, 40 },
		{ 61, 392, 40 },
		{ 61, 393, 40 },
		{ 61, 394, 40 },
		{ 61, 395, 40 },
		{ 61, 396, 40 },
		{ 61, 397, 40 },
		{ 61, 398, 40 },
		{ 61, 399, 40 },
		{ 61, 400, 40 },
		{ 61, 401, 40 },
		{ 61, 402, 40 },
		{ 61, 403, 40 },
		{ 61, 404, 40 },
		{ 61, 405, 40 },
		{ 61, 406, 40 },
		{ 61, 407, 40 },
		{ 61, 408, 40 },
		{ 61, 409, 40 },
		{ 61, 410, 40 },
		{ 61, 411, 40 },
		{ 61, 412, 40 },
		{ 61, 413, 40 },
		{ 61, 414, 40 },
		{ 61, 415, 40 },
		{ 61, 416, 40 },
		{ 61, 417, 40 },
		{ 61, 418, 40 },
		{ 61, 419, 40 },
		{ 61, 420, 40 },
		{ 61, 421, 40 },
		{ 61, 422, 40 },
		{ 61, 423, 40 },
		{ 61, 424, 40 },
		{ 61, 425, 40 },
		{ 61, 426, 40 },
		{ 61, 427, 40 },
		{ 61, 428, 40 },
		{ 62, 379, 40 },
		{ 62, 380, 40 },
		{ 62, 381, 40 },
		{ 62, 382, 40 },
		{ 62, 383, 40 },
		{ 62, 384, 40 },
		{ 62, 385, 40 },
		{ 62, 386, 40 },
		{ 62, 387, 40 },
		{ 62, 388, 40 },
		{ 62, 389, 40 },
		{ 62, 390, 40 },
		{ 62, 391, 40 },
		{ 62, 392, 40 },
		{ 62, 393, 40 },
		{ 62, 394, 40 },
		{ 62, 395, 40 },
		{ 62, 396, 40 },
		{ 62, 397, 40 },
		{ 62, 398, 40 },
		{ 62, 399, 40 },
		{ 62, 400, 40 },
		{ 62, 401, 40 },
		{ 62, 402, 40 },
		{ 62, 403, 40 },
		{ 62, 404, 40 },
		{ 62, 405, 40 },
		{ 62, 406, 40 },
		{ 62, 407, 40 },
		{ 62, 408, 40 },
		{ 62, 409, 40 },
		{ 62, 410, 40 },
		{ 62, 411, 40 },
		{ 62, 412, 40 },
		{ 62, 413, 40 },
		{ 62, 414, 40 },
		{ 62, 415, 40 },
		{ 62, 416, 40 },
		{ 62, 417, 40 },
		{ 62, 418, 40 },
		{ 62, 419, 40 },
		{ 62, 420, 40 },
		{ 62, 421, 40 },
		{ 62, 422, 40 },
		{ 62, 423, 40 },
		{ 62, 424, 40 },
		{ 62, 425, 40 },
		{ 62, 426, 40 },
		{ 62, 427, 40 },
		{ 63, 379, 40 },
		{ 63, 380, 40 },
		{ 63, 381, 40 },
		{ 63, 382, 40 },
		{ 63, 383, 40 },
		{ 63, 384, 40 },
		{ 63, 385, 40 },
		{ 63, 386, 40 },
		{ 63, 387, 40 },
		{ 63, 388, 40 },
		{ 63, 389, 40 },
		{ 63, 390, 40 },
		{ 63, 391, 40 },
		{ 63, 392, 40 },
		{ 63, 393, 40 },
		{ 63, 394, 40 },
		{ 63, 395, 40 },
		{ 63, 396, 40 },
		{ 63, 397, 40 },
		{ 63, 398, 40 },
		{ 63, 399, 40 },
		{ 63, 400, 40 },
		{ 63, 401, 40 },
		{ 63, 402, 40 },
		{ 63, 403, 40 },
		{ 63, 404, 40 },
		{ 63, 405, 40 },
		{ 63, 406, 40 },
		{ 63, 407, 40 },
		{ 63, 408, 40 },
		{ 63, 409, 40 },
		{ 63, 410, 40 },
		{ 63, 411, 40 },
		{ 63, 412, 40 },
		{ 63, 413, 40 },
		{ 63, 414, 40 },
		{ 63, 415, 40 },
		{ 63, 416, 40 },
		{ 63, 417, 40 },
		{ 63, 418, 40 },
		{ 63, 419, 40 },
		{ 63, 420, 40 },
		{ 63, 421, 40 },
		{ 63, 422, 40 },
		{ 63, 423, 40 },
		{ 63, 424, 40 },
		{ 63, 425, 40 },
		{ 63, 426, 40 },
		{ 64, 378, 40 },
		{ 64, 379, 40 },
		{ 64, 380, 40 },
		{ 64, 381, 40 },
		{ 64, 382, 40 },
		{ 64, 383, 40 },
		{ 64, 384, 40 },
		{ 64, 385, 40 },
		{ 64, 386, 40 },
		{ 64, 387, 40 },
		{ 64, 388, 40 },
		{ 64, 389, 40 },
		{ 64, 390, 40 },
		{ 64, 391, 40 },
		{ 64, 392, 40 },
		{ 64, 393, 40 },
		{ 64, 394, 40 },
		{ 64, 395, 40 },
		{ 64, 396, 40 },
		{ 64, 397, 40 },
		{ 64, 398, 40 },
		{ 64, 399, 40 },
		{ 64, 400, 40 },
		{ 64, 401, 40 },
		{ 64, 402, 40 },
		{ 64, 403, 40 },
		{ 64, 404, 40 },
		{ 64, 405, 40 },
		{ 64, 406, 40 },
		{ 64, 407, 40 },
		{ 64, 408, 40 },
		{ 64, 409, 40 },
		{ 64, 410, 40 },
		{ 64, 411, 40 },
		{ 64, 412, 40 },
		{ 64, 413, 40 },
		{ 64, 414, 40 },
		{ 64, 415, 40 },
		{ 64, 416, 40 },
		{ 64, 417, 40 },
		{ 64, 418, 40 },
		{ 64, 419, 40 },
		{ 64, 420, 40 },
		{ 64, 421, 40 },
		{ 64, 422, 40 },
		{ 64, 423, 40 },
		{ 64, 424, 40 },
		{ 64, 425, 40 },
		{ 65, 378, 40 },
		{ 65, 379, 40 },
		{ 65, 380, 40 },
		{ 65, 381, 40 },
		{ 65, 382, 40 },
		{ 65, 383, 40 },
		{ 65, 384, 40 },
		{ 65, 385, 40 },
		{ 65, 386, 40 },
		{ 65, 387, 40 },
		{ 65, 388, 40 },
		{ 65, 389, 40 },
		{ 65, 390, 40 },
		{ 65, 391, 40 },
		{ 65, 392, 40 },
		{ 65, 393, 40 },
		{ 65, 394, 40 },
		{ 65, 395, 40 },
		{ 65, 396, 40 },
		{ 65, 397, 40 },
		{ 65, 398, 40 },
		{ 65, 399, 40 },
		{ 65, 400, 40 },
		{ 65, 401, 40 },
		{ 65, 402, 40 },
		{ 65, 403, 40 },
		{ 65, 404, 40 },
		{ 65, 405, 40 },
		{ 65, 406, 40 },
		{ 65, 407, 40 },
		{ 65, 408, 40 },
		{ 65, 409, 40 },
		{ 65, 410, 40 },
		{ 65, 411, 40 },
		{ 65, 412, 40 },
		{ 65, 413, 40 },
		{ 65, 414, 40 },
		{ 65, 415, 40 },
		{ 65, 416, 40 },
		{ 65, 417, 40 },
		{ 65, 418, 40 },
		{ 65, 419, 40 },
		{ 65, 420, 40 },
		{ 65, 421, 40 },
		{ 65, 422, 40 },
		{ 65, 423, 40 },
		{ 65, 424, 40 },
		{ 66, 377, 40 },
		{ 66, 378, 40 },
		{ 66, 379, 40 },
		{ 66, 380, 40 },
		{ 66, 381, 40 },
		{ 66, 382, 40 },
		{ 66, 383, 40 },
		{ 66, 384, 40 },
		{ 66, 385, 40 },
		{ 66, 386, 40 },
		{ 66, 387, 40 },
		{ 66, 388, 40 },
		{ 66, 389, 40 },
		{ 66, 390, 40 },
		{ 66, 391, 40 },
		{ 66, 392, 40 },
		{ 66, 393, 40 },
		{ 66, 394, 40 },
		{ 66, 395, 40 },
		{ 66, 396, 40 },
		{ 66, 397, 40 },
		{ 66, 398, 40 },
		{ 66, 399, 40 },
		{ 66, 400, 40 },
		{ 66, 401, 40 },
		{ 66, 402, 40 },
		{ 66, 403, 40 },
		{ 66, 404, 40 },
		{ 66, 405, 40 },
		{ 66, 406, 40 },
		{ 66, 407, 40 },
		{ 66, 408, 40 },
		{ 66, 409, 40 },
		{ 66, 410, 40 },
		{ 66, 411, 40 },
		{ 66, 412, 40 },
		{ 66, 413, 40 },
		{ 66, 414, 40 },
		{ 66, 415, 40 },
		{ 66, 416, 40 },
		{ 66, 417, 40 },
		{ 66, 418, 40 },
		{ 66, 419, 40 },
		{ 66, 420, 40 },
		{ 66, 421, 40 },
		{ 66, 422, 40 },
		{ 66, 423, 40 },
		{ 66, 424, 40 },
		{ 67, 376, 40 },
		{ 67, 377, 40 },
		{ 67, 378, 40 },
		{ 67, 379, 40 },
		{ 67, 380, 40 },
		{ 67, 381, 40 },
		{ 67, 382, 40 },
		{ 67, 383, 40 },
		{ 67, 384, 40 },
		{ 67, 385, 40 },
		{ 67, 386, 40 },
		{ 67, 387, 40 },
		{ 67, 388, 40 },
		{ 67, 389, 40 },
		{ 67, 390, 40 },
		{ 67, 391, 40 },
		{ 67, 392, 40 },
		{ 67, 393, 40 },
		{ 67, 394, 40 },
		{ 67, 395, 40 },
		{ 67, 396, 40 },
		{ 67, 397, 40 },
		{ 67, 398, 40 },
		{ 67, 399, 40 },
		{ 67, 400, 40 },
		{ 67, 401, 40 },
		{ 67, 402, 40 },
		{ 67, 403, 40 },
		{ 67, 404, 40 },
		{ 67, 405, 40 },
		{ 67, 406, 40 },
		{ 67, 407, 40 },
		{ 67, 408, 40 },
		{ 67, 409, 40 },
		{ 67, 410, 40 },
		{ 67, 411, 40 },
		{ 67, 412, 40 },
		{ 67, 413, 40 },
		{ 67, 414, 40 },
		{ 67, 415, 40 },
		{ 67, 416, 40 },
		{ 67, 417, 40 },
		{ 67, 418, 40 },
		{ 67, 419, 40 },
		{ 67, 420, 40 },
		{ 67, 421, 40 },
		{ 67, 422, 40 },
		{ 67, 423, 40 },
		{ 68, 376, 40 },
		{ 68, 377, 40 },
		{ 68, 378, 40 },
		{ 68, 379, 40 },
		{ 68, 380, 40 },
		{ 68, 381, 40 },
		{ 68, 382, 40 },
		{ 68, 383, 40 },
		{ 68, 384, 40 },
		{ 68, 385, 40 },
		{ 68, 386, 40 },
		{ 68, 387, 40 },
		{ 68, 388, 40 },
		{ 68, 389, 40 },
		{ 68, 390, 40 },
		{ 68, 391, 40 },
		{ 68, 392, 40 },
		{ 68, 393, 40 },
		{ 68, 394, 40 },
		{ 68, 395, 40 },
		{ 68, 396, 40 },
		{ 68, 397, 40 },
		{ 68, 398, 40 },
		{ 68, 399, 40 },
		{ 68, 400, 40 },
		{ 68, 401, 40 },
		{ 68, 402, 40 },
		{ 68, 403, 40 },
		{ 68, 404, 40 },
		{ 68, 405, 40 },
		{ 68, 406, 40 },
		{ 68, 407, 40 },
		{ 68, 408, 40 },
		{ 68, 409, 40 },
		{ 68, 410, 40 },
		{ 68, 411, 40 },
		{ 68, 412, 40 },
		{ 68, 413, 40 },
		{ 68, 414, 40 },
		{ 68, 415, 40 },
		{ 68, 416, 40 },
		{ 68, 417, 40 },
		{ 68, 418, 40 },
		{ 68, 419, 40 },
		{ 68, 420, 40 },
		{ 68, 421, 40 },
		{ 68, 422, 40 },
		{ 69, 375, 40 },
		{ 69, 376, 40 },
		{ 69, 377, 40 },
		{ 69, 378, 40 },
		{ 69, 379, 40 },
		{ 69, 380, 40 },
		{ 69, 381, 40 },
		{ 69, 382, 40 },
		{ 69, 383, 40 },
		{ 69, 384, 40 },
		{ 69, 385, 40 },
		{ 69, 386, 40 },
		{ 69, 387, 40 },
		{ 69, 388, 40 },
		{ 69, 389, 40 },
		{ 69, 390, 40 },
		{ 69, 391, 40 },
		{ 69, 392, 40 },
		{ 69, 393, 40 },
		{ 69, 394, 40 },
		{ 69, 395, 40 },
		{ 69, 396, 40 },
		{ 69, 397, 40 },
		{ 69, 398, 40 },
		{ 69, 399, 40 },
		{ 69, 400, 40 },
		{ 69, 401, 40 },
		{ 69, 402, 40 },
		{ 69, 403, 40 },
		{ 69, 404, 40 },
		{ 69, 405, 40 },
		{ 69, 406, 40 },
		{ 69, 407, 40 },
		{ 69, 408, 40 },
		{ 69, 409, 40 },
		{ 69, 410, 40 },
		{ 69, 411, 40 },
		{ 69, 412, 40 },
		{ 69, 413, 40 },
		{ 69, 414, 40 },
		{ 69, 415, 40 },
		{ 69, 416, 40 },
		{ 69, 417, 40 },
		{ 69, 418, 40 },
		{ 69, 419, 40 },
		{ 69, 420, 40 },
		{ 69, 421, 40 },
		{ 70, 374, 40 },
		{ 70, 375, 40 },
		{ 70, 376, 40 },
		{ 70, 377, 40 },
		{ 70, 378, 40 },
		{ 70, 379, 40 },
		{ 70, 380, 40 },
		{ 70, 381, 40 },
		{ 70, 382, 40 },
		{ 70, 383, 40 },
		{ 70, 384, 40 },
		{ 70, 385, 40 },
		{ 70, 386, 40 },
		{ 70, 387, 40 },
		{ 70, 388, 40 },
		{ 70, 389, 40 },
		{ 70, 390, 40 },
		{ 70, 391, 40 },
		{ 70, 392, 40 },
		{ 70, 393, 40 },
		{ 70, 394, 40 },
		{ 70, 395, 40 },
		{ 70, 396, 40 },
		{ 70, 397, 40 },
		{ 70, 398, 40 },
		{ 70, 399, 40 },
		{ 70, 400, 40 },
		{ 70, 401, 40 },
		{ 70, 402, 40 },
		{ 70, 403, 40 },
		{ 70, 404, 40 },
		{ 70, 405, 40 },
		{ 70, 406, 40 },
		{ 70, 407, 40 },
		{ 70, 408, 40 },
		{ 70, 409, 40 },
		{ 70, 410, 40 },
		{ 70, 411, 40 },
		{ 70, 412, 40 },
		{ 70, 413, 40 },
		{ 70, 414, 40 },
		{ 70, 415, 40 },
		{ 70, 416, 40 },
		{ 70, 417, 40 },
		{ 70, 418, 40 },
		{ 70, 419, 40 },
		{ 70, 420, 40 },
		{ 71, 374, 40 },
		{ 71, 375, 40 },
		{ 71, 376, 40 },
		{ 71, 377, 40 },
		{ 71, 378, 40 },
		{ 71, 379, 40 },
		{ 71, 380, 40 },
		{ 71, 381, 40 },
		{ 71, 382, 40 },
		{ 71, 383, 40 },
		{ 71, 384, 40 },
		{ 71, 385, 40 },
		{ 71, 386, 40 },
		{ 71, 387, 40 },
		{ 71, 388, 40 },
		{ 71, 389, 40 },
		{ 71, 390, 40 },
		{ 71, 391, 40 },
		{ 71, 392, 40 },
		{ 71, 393, 40 },
		{ 71, 394, 40 },
		{ 71, 395, 40 },
		{ 71, 396, 40 },
		{ 71, 397, 40 },
		{ 71, 398, 40 },
		{ 71, 399, 40 },
		{ 71, 400, 40 },
		{ 71, 401, 40 },
		{ 71, 402, 40 },
		{ 71, 403, 40 },
		{ 71, 404, 40 },
		{ 71, 405, 40 },
		{ 71, 406, 40 },
		{ 71, 407, 40 },
		{ 71, 408, 40 },
		{ 71, 409, 40 },
		{ 71, 410, 40 },
		{ 71, 411, 40 },
		{ 71, 412, 40 },
		{ 71, 413, 40 },
		{ 71, 414, 40 },
		{ 71, 415, 40 },
		{ 71, 416, 40 },
		{ 71, 417, 40 },
		{ 71, 418, 40 },
		{ 71, 419, 40 },
		{ 72, 373, 40 },
		{ 72, 374, 40 },
		{ 72, 375, 40 },
		{ 72, 376, 40 },
		{ 72, 377, 40 },
		{ 72, 378, 40 },
		{ 72, 379, 40 },
		{ 72, 380, 40 },
		{ 72, 381, 40 },
		{ 72, 382, 40 },
		{ 72, 383, 40 },
		{ 72, 384, 40 },
		{ 72, 385, 40 },
		{ 72, 386, 40 },
		{ 72, 387, 40 },
		{ 72, 388, 40 },
		{ 72, 389, 40 },
		{ 72, 390, 40 },
		{ 72, 391, 40 },
		{ 72, 392, 40 },
		{ 72, 393, 40 },
		{ 72, 394, 40 },
		{ 72, 395, 40 },
		{ 72, 396, 40 },
		{ 72, 397, 40 },
		{ 72, 398, 40 },
		{ 72, 399, 40 },
		{ 72, 400, 40 },
		{ 72, 401, 40 },
		{ 72, 402, 40 },
		{ 72, 403, 40 },
		{ 72, 404, 40 },
		{ 72, 405, 40 },
		{ 72, 406, 40 },
		{ 72, 407, 40 },
		{ 72, 408, 40 },
		{ 72, 409, 40 },
		{ 72, 410, 40 },
		{ 72, 411, 40 },
		{ 72, 412, 40 },
		{ 72, 413, 40 },
		{ 72, 414, 40 },
		{ 72, 415, 40 },
		{ 72, 416, 40 },
		{ 72, 417, 40 },
		{ 72, 418, 40 },
		{ 73, 373, 40 },
		{ 73, 374, 40 },
		{ 73, 375, 40 },
		{ 73, 376, 40 },
		{ 73, 377, 40 },
		{ 73, 378, 40 },
		{ 73, 379, 40 },
		{ 73, 380, 40 },
		{ 73, 381, 40 },
		{ 73, 382, 40 },
		{ 73, 383, 40 },
		{ 73, 384, 40 },
		{ 73, 385, 40 },
		{ 73, 386, 40 },
		{ 73, 387, 40 },
		{ 73, 388, 40 },
		{ 73, 389, 40 },
		{ 73, 390, 40 },
		{ 73, 391, 40 },
		{ 73, 392, 40 },
		{ 73, 393, 40 },
		{ 73, 394, 40 },
		{ 73, 395, 40 },
		{ 73, 396, 40 },
		{ 73, 397, 40 },
		{ 73, 398, 40 },
		{ 73, 399, 40 },
		{ 73, 400, 40 },
		{ 73, 401, 40 },
		{ 73, 402, 40 },
		{ 73, 403, 40 },
		{ 73, 404, 40 },
		{ 73, 405, 40 },
		{ 73, 406, 40 },
		{ 73, 407, 40 },
		{ 73, 408, 40 },
		{ 73, 409, 40 },
		{ 73, 410, 40 },
		{ 73, 411, 40 },
		{ 73, 412, 40 },
		{ 73, 413, 40 },
		{ 73, 414, 40 },
		{ 73, 415, 40 },
		{ 73, 416, 40 },
		{ 73, 417, 40 },
		{ 74, 372, 40 },
		{ 74, 373, 40 },
		{ 74, 374, 40 },
		{ 74, 375, 40 },
		{ 74, 376, 40 },
		{ 74, 377, 40 },
		{ 74, 378, 40 },
		{ 74, 379, 40 },
		{ 74, 380, 40 },
		{ 74, 381, 40 },
		{ 74, 382, 40 },
		{ 74, 383, 40 },
		{ 74, 384, 40 },
		{ 74, 385, 40 },
		{ 74, 386, 40 },
		{ 74, 387, 40 },
		{ 74, 388, 40 },
		{ 74, 389, 40 },
		{ 74, 390, 40 },
		{ 74, 391, 40 },
		{ 74, 392, 40 },
		{ 74, 393, 40 },
		{ 74, 394, 40 },
		{ 74, 395, 40 },
		{ 74, 396, 40 },
		{ 74, 397, 40 },
		{ 74, 398, 40 },
		{ 74, 399, 40 },
		{ 74, 400, 40 },
		{ 74, 401, 40 },
		{ 74, 402, 40 },
		{ 74, 403, 40 },
		{ 74, 404, 40 },
		{ 74, 405, 40 },
		{ 74, 406, 40 },
		{ 74, 407, 40 },
		{ 74, 408, 40 },
		{ 74, 409, 40 },
		{ 74, 410, 40 },
		{ 74, 411, 40 },
		{ 74, 412, 40 },
		{ 74, 413, 40 },
		{ 74, 414, 40 },
		{ 74, 415, 40 },
		{ 74, 416, 40 },
		{ 74, 417, 40 },
		{ 75, 371, 40 },
		{ 75, 372, 40 },
		{ 75, 373, 40 },
		{ 75, 374, 40 },
		{ 75, 375, 40 },
		{ 75, 376, 40 },
		{ 75, 377, 40 },
		{ 75, 378, 40 },
		{ 75, 379, 40 },
		{ 75, 380, 40 },
		{ 75, 381, 40 },
		{ 75, 382, 40 },
		{ 75, 383, 40 },
		{ 75, 384, 40 },
		{ 75, 385, 40 },
		{ 75, 386, 40 },
		{ 75, 387, 40 },
		{ 75, 388, 40 },
		{ 75, 389, 40 },
		{ 75, 390, 40 },
		{ 75, 391, 40 },
		{ 75, 392, 40 },
		{ 75, 393, 40 },
		{ 75, 394, 40 },
		{ 75, 395, 40 },
		{ 75, 396, 40 },
		{ 75, 397, 40 },
		{ 75, 398, 40 },
		{ 75, 399, 40 },
		{ 75, 400, 40 },
		{ 75, 401, 40 },
		{ 75, 402, 40 },
		{ 75, 403, 40 },
		{ 75, 404, 40 },
		{ 75, 405, 40 },
		{ 75, 406, 40 },
		{ 75, 407, 40 },
		{ 75, 408, 40 },
		{ 75, 409, 40 },
		{ 75, 410, 40 },
		{ 75, 411, 40 },
		{ 75, 412, 40 },
		{ 75, 413, 40 },
		{ 75, 414, 40 },
		{ 75, 415, 40 },
		{ 75, 416, 40 },
		{ 76, 371, 40 },
		{ 76, 372, 40 },
		{ 76, 373, 40 },
		{ 76, 374, 40 },
		{ 76, 375, 40 },
		{ 76, 376, 40 },
		{ 76, 377, 40 },
		{ 76, 378, 40 },
		{ 76, 379, 40 },
		{ 76, 380, 40 },
		{ 76, 381, 40 },
		{ 76, 382, 40 },
		{ 76, 383, 40 },
		{ 76, 384, 40 },
		{ 76, 385, 40 },
		{ 76, 386, 40 },
		{ 76, 387, 40 },
		{ 76, 388, 40 },
		{ 76, 389, 40 },
		{ 76, 390, 40 },
		{ 76, 391, 40 },
		{ 76, 392, 40 },
		{ 76, 393, 40 },
		{ 76, 394, 40 },
		{ 76, 395, 40 },
		{ 76, 396, 40 },
		{ 76, 397, 40 },
		{ 76, 398, 40 },
		{ 76, 399, 40 },
		{ 76, 400, 40 },
		{ 76, 401, 40 },
		{ 76, 402, 40 },
		{ 76, 403, 40 },
		{ 76, 404, 40 },
		{ 76, 405, 40 },
		{ 76, 406, 40 },
		{ 76, 407, 40 },
		{ 76, 408, 40 },
		{ 76, 409, 40 },
		{ 76, 410, 40 },
		{ 76, 411, 40 },
		{ 76, 412, 40 },
		{ 76, 413, 40 },
		{ 76, 414, 40 },
		{ 76, 415, 40 },
		{ 77, 370, 40 },
		{ 77, 371, 40 },
		{ 77, 372, 40 },
		{ 77, 373, 40 },
		{ 77, 374, 40 },
		{ 77, 375, 40 },
		{ 77, 376, 40 },
		{ 77, 377, 40 },
		{ 77, 378, 40 },
		{ 77, 379, 40 },
		{ 77, 380, 40 },
		{ 77, 381, 40 },
		{ 77, 382, 40 },
		{ 77, 383, 40 },
		{ 77, 384, 40 },
		{ 77, 385, 40 },
		{ 77, 386, 40 },
		{ 77, 387, 40 },
		{ 77, 388, 40 },
		{ 77, 389, 40 },
		{ 77, 390, 40 },
		{ 77, 391, 40 },
		{ 77, 392, 40 },
		{ 77, 393, 40 },
		{ 77, 394, 40 },
		{ 77, 395, 40 },
		{ 77, 396, 40 },
		{ 77, 397, 40 },
		{ 77, 398, 40 },
		{ 77, 399, 40 },
		{ 77, 400, 40 },
		{ 77, 401, 40 },
		{ 77, 402, 40 },
		{ 77, 403, 40 },
		{ 77, 404, 40 },
		{ 77, 405, 40 },
		{ 77, 406, 40 },
		{ 77, 407, 40 },
		{ 77, 408, 40 },
		{ 77, 409, 40 },
		{ 77, 410, 40 },
		{ 77, 411, 40 },
		{ 77, 412, 40 },
		{ 77, 413, 40 },
		{ 77, 414, 40 },
		{ 78, 369, 40 },
		{ 78, 370, 40 },
		{ 78, 371, 40 },
		{ 78, 372, 40 },
		{ 78, 373, 40 },
		{ 78, 374, 40 },
		{ 78, 375, 40 },
		{ 78, 376, 40 },
		{ 78, 377, 40 },
		{ 78, 378, 40 },
		{ 78, 379, 40 },
		{ 78, 380, 40 },
		{ 78, 381, 40 },
		{ 78, 382, 40 },
		{ 78, 383, 40 },
		{ 78, 384, 40 },
		{ 78, 385, 40 },
		{ 78, 386, 40 },
		{ 78, 387, 40 },
		{ 78, 388, 40 },
		{ 78, 389, 40 },
		{ 78, 390, 40 },
		{ 78, 391, 40 },
		{ 78, 392, 40 },
		{ 78, 393, 40 },
		{ 78, 394, 40 },
		{ 78, 395, 40 },
		{ 78, 396, 40 },
		{ 78, 397, 40 },
		{ 78, 398, 40 },
		{ 78, 399, 40 },
		{ 78, 400, 40 },
		{ 78, 401, 40 },
		{ 78, 402, 40 },
		{ 78, 403, 40 },
		{ 78, 404, 40 },
		{ 78, 405, 40 },
		{ 78, 406, 40 },
		{ 78, 407, 40 },
		{ 78, 408, 40 },
		{ 78, 409, 40 },
		{ 78, 410, 40 },
		{ 78, 411, 40 },
		{ 78, 412, 40 },
		{ 78, 413, 40 },
		{ 79, 369, 40 },
		{ 79, 370, 40 },
		{ 79, 371, 40 },
		{ 79, 372, 40 },
		{ 79, 373, 40 },
		{ 79, 374, 40 },
		{ 79, 375, 40 },
		{ 79, 376, 40 },
		{ 79, 377, 40 },
		{ 79, 378, 40 },
		{ 79, 379, 40 },
		{ 79, 380, 40 },
		{ 79, 381, 40 },
		{ 79, 382, 40 },
		{ 79, 383, 40 },
		{ 79, 384, 40 },
		{ 79, 385, 40 },
		{ 79, 386, 40 },
		{ 79, 387, 40 },
		{ 79, 388, 40 },
		{ 79, 389, 40 },
		{ 79, 390, 40 },
		{ 79, 391, 40 },
		{ 79, 392, 40 },
		{ 79, 393, 40 },
		{ 79, 394, 40 },
		{ 79, 395, 40 },
		{ 79, 396, 40 },
		{ 79, 397, 40 },
		{ 79, 398, 40 },
		{ 79, 399, 40 },
		{ 79, 400, 40 },
		{ 79, 401, 40 },
		{ 79, 402, 40 },
		{ 79, 403, 40 },
		{ 79, 404, 40 },
		{ 79, 405, 40 },
		{ 79, 406, 40 },
		{ 79, 407, 40 },
		{ 79, 408, 40 },
		{ 79, 409, 40 },
		{ 79, 410, 40 },
		{ 79, 411, 40 },
		{ 79, 412, 40 },
		{ 80, 368, 40 },
		{ 80, 369, 40 },
		{ 80, 370, 40 },
		{ 80, 371, 40 },
		{ 80, 372, 40 },
		{ 80, 373, 40 },
		{ 80, 374, 40 },
		{ 80, 375, 40 },
		{ 80, 376, 40 },
		{ 80, 377, 40 },
		{ 80, 378, 40 },
		{ 80, 379, 40 },
		{ 80, 380, 40 },
		{ 80, 381, 40 },
		{ 80, 382, 40 },
		{ 80, 383, 40 },
		{ 80, 384, 40 },
		{ 80, 385, 40 },
		{ 80, 386, 40 },
		{ 80, 387, 40 },
		{ 80, 388, 40 },
		{ 80, 389, 40 },
		{ 80, 390, 40 },
		{ 80, 391, 40 },
		{ 80, 392, 40 },
		{ 80, 393, 40 },
		{ 80, 394, 40 },
		{ 80, 395, 40 },
		{ 80, 396, 40 },
		{ 80, 397, 40 },
		{ 80, 398, 40 },
		{ 80, 399, 40 },
		{ 80, 400, 40 },
		{ 80, 401, 40 },
		{ 80, 402, 40 },
		{ 80, 403, 40 },
		{ 80, 404, 40 },
		{ 80, 405, 40 },
		{ 80, 406, 40 },
		{ 80, 407, 40 },
		{ 80, 408, 40 },
		{ 80, 409, 40 },
		{ 80, 410, 40 },
		{ 80, 411, 40 },
		{ 81, 368, 40 },
		{ 81, 369, 40 },
		{ 81, 370, 40 },
		{ 81, 371, 40 },
		{ 81, 372, 40 },
		{ 81, 373, 40 },
		{ 81, 374, 40 },
		{ 81, 375, 40 },
		{ 81, 376, 40 },
		{ 81, 377, 40 },
		{ 81, 378, 40 },
		{ 81, 379, 40 },
		{ 81, 380, 40 },
		{ 81, 381, 40 },
		{ 81, 382, 40 },
		{ 81, 383, 40 },
		{ 81, 384, 40 },
		{ 81, 385, 40 },
		{ 81, 386, 40 },
		{ 81, 387, 40 },
		{ 81, 388, 40 },
		{ 81, 389, 40 },
		{ 81, 390, 40 },
		{ 81, 391, 40 },
		{ 81, 392, 40 },
		{ 81, 393, 40 },
		{ 81, 394, 40 },
		{ 81, 395, 40 },
		{ 81, 396, 40 },
		{ 81, 397, 40 },
		{ 81, 398, 40 },
		{ 81, 399, 40 },
		{ 81, 400, 40 },
		{ 81, 401, 40 },
		{ 81, 402, 40 },
		{ 81, 403, 40 },
		{ 81, 404, 40 },
		{ 81, 405, 40 },
		{ 81, 406, 40 },
		{ 81, 407, 40 },
		{ 81, 408, 40 },
		{ 81, 409, 40 },
		{ 81, 410, 40 },
		{ 82, 367, 40 },
		{ 82, 368, 40 },
		{ 82, 369, 40 },
		{ 82, 370, 40 },
		{ 82, 371, 40 },
		{ 82, 372, 40 },
		{ 82, 373, 40 },
		{ 82, 374, 40 },
		{ 82, 375, 40 },
		{ 82, 376, 40 },
		{ 82, 377, 40 },
		{ 82, 378, 40 },
		{ 82, 379, 40 },
		{ 82, 380, 40 },
		{ 82, 381, 40 },
		{ 82, 382, 40 },
		{ 82, 383, 40 },
		{ 82, 384, 40 },
		{ 82, 385, 40 },
		{ 82, 386, 40 },
		{ 82, 387, 40 },
		{ 82, 388, 40 },
		{ 82, 389, 40 },
		{ 82, 390, 40 },
		{ 82, 391, 40 },
		{ 82, 392, 40 },
		{ 82, 393, 40 },
		{ 82, 394, 40 },
		{ 82, 395, 40 },
		{ 82, 396, 40 },
		{ 82, 397, 40 },
		{ 82, 398, 40 },
		{ 82, 399, 40 },
		{ 82, 400, 40 },
		{ 82, 401, 40 },
		{ 82, 402, 40 },
		{ 82, 403, 40 },
		{ 82, 404, 40 },
		{ 82, 405, 40 },
		{ 82, 406, 40 },
		{ 82, 407, 40 },
		{ 82, 408, 40 },
		{ 82, 409, 40 },
		{ 82, 410, 40 },
		{ 83, 366, 40 },
		{ 83, 367, 40 },
		{ 83, 368, 40 },
		{ 83, 369, 40 },
		{ 83, 370, 40 },
		{ 83, 371, 40 },
		{ 83, 372, 40 },
		{ 83, 373, 40 },
		{ 83, 374, 40 },
		{ 83, 375, 40 },
		{ 83, 376, 40 },
		{ 83, 377, 40 },
		{ 83, 378, 40 },
		{ 83, 379, 40 },
		{ 83, 380, 40 },
		{ 83, 381, 40 },
		{ 83, 382, 40 },
		{ 83, 383, 40 },
		{ 83, 384, 40 },
		{ 83, 385, 40 },
		{ 83, 386, 40 },
		{ 83, 387, 40 },
		{ 83, 388, 40 },
		{ 83, 389, 40 },
		{ 83, 390, 40 },
		{ 83, 391, 40 },
		{ 83, 392, 40 },
		{ 83, 393, 40 },
		{ 83, 394, 40 },
		{ 83, 395, 40 },
		{ 83, 396, 40 },
		{ 83, 397, 40 },
		{ 83, 398, 40 },
		{ 83, 399, 40 },
		{ 83, 400, 40 },
		{ 83, 401, 40 },
		{ 83, 402, 40 },
		{ 83, 403, 40 },
		{ 83, 404, 40 },
		{ 83, 405, 40 },
		{ 83, 406, 40 },
		{ 83, 407, 40 },
		{ 83, 408, 40 },
		{ 83, 409, 40 },
		{ 84, 366, 40 },
		{ 84, 367, 40 },
		{ 84, 368, 40 },
		{ 84, 369, 40 },
		{ 84, 370, 40 },
		{ 84, 371, 40 },
		{ 84, 372, 40 },
		{ 84, 373, 40 },
		{ 84, 374, 40 },
		{ 84, 375, 40 },
		{ 84, 376, 40 },
		{ 84, 377, 40 },
		{ 84, 378, 40 },
		{ 84, 379, 40 },
		{ 84, 380, 40 },
		{ 84, 381, 40 },
		{ 84, 382, 40 },
		{ 84, 383, 40 },
		{ 84, 384, 40 },
		{ 84, 385, 40 },
		{ 84, 386, 40 },
		{ 84, 387, 40 },
		{ 84, 388, 40 },
		{ 84, 389, 40 },
		{ 84, 390, 40 },
		{ 84, 391, 40 },
		{ 84, 392, 40 },
		{ 84, 393, 40 },
		{ 84, 394, 40 },
		{ 84, 395, 40 },
		{ 84, 396, 40 },
		{ 84, 397, 40 },
		{ 84, 398, 40 },
		{ 84, 399, 40 },
		{ 84, 400, 40 },
		{ 84, 401, 40 },
		{ 84, 402, 40 },
		{ 84, 403, 40 },
		{ 84, 404, 40 },
		{ 84, 405, 40 },
		{ 84, 406, 40 },
		{ 84, 407, 40 },
		{ 84, 408, 40 },
		{ 85, 365, 40 },
		{ 85, 366, 40 },
		{ 85, 367, 40 },
		{ 85, 368, 40 },
		{ 85, 369, 40 },
		{ 85, 370, 40 },
		{ 85, 371, 40 },
		{ 85, 372, 40 },
		{ 85, 373, 40 },
		{ 85, 374, 40 },
		{ 85, 375, 40 },
		{ 85, 376, 40 },
		{ 85, 377, 40 },
		{ 85, 378, 40 },
		{ 85, 379, 40 },
		{ 85, 380, 40 },
		{ 85, 381, 40 },
		{ 85, 382, 40 },
		{ 85, 383, 40 },
		{ 85, 384, 40 },
		{ 85, 385, 40 },
		{ 85, 386, 40 },
		{ 85, 387, 40 },
		{ 85, 388, 40 },
		{ 85, 389, 40 },
		{ 85, 390, 40 },
		{ 85, 391, 40 },
		{ 85, 392, 40 },
		{ 85, 393, 40 },
		{ 85, 394, 40 },
		{ 85, 395, 40 },
		{ 85, 396, 40 },
		{ 85, 397, 40 },
		{ 85, 398, 40 },
		{ 85, 399, 40 },
		{ 85, 400, 40 },
		{ 85, 401, 40 },
		{ 85, 402, 40 },
		{ 85, 403, 40 },
		{ 85, 404, 40 },
		{ 85, 405, 40 },
		{ 85, 406, 40 },
		{ 85, 407, 40 },
		{ 86, 364, 40 },
		{ 86, 365, 40 },
		{ 86, 366, 40 },
		{ 86, 367, 40 },
		{ 86, 368, 40 },
		{ 86, 369, 40 },
		{ 86, 370, 40 },
		{ 86, 371, 40 },
		{ 86, 372, 40 },
		{ 86, 373, 40 },
		{ 86, 374, 40 },
		{ 86, 375, 40 },
		{ 86, 376, 40 },
		{ 86, 377, 40 },
		{ 86, 378, 40 },
		{ 86, 379, 40 },
		{ 86, 380, 40 },
		{ 86, 381, 40 },
		{ 86, 382, 40 },
		{ 86, 383, 40 },
		{ 86, 384, 40 },
		{ 86, 385, 40 },
		{ 86, 386, 40 },
		{ 86, 387, 40 },
		{ 86, 388, 40 },
		{ 86, 389, 40 },
		{ 86, 390, 40 },
		{ 86, 391, 40 },
		{ 86, 392, 40 },
		{ 86, 393, 40 },
		{ 86, 394, 40 },
		{ 86, 395, 40 },
		{ 86, 396, 40 },
		{ 86, 397, 40 },
		{ 86, 398, 40 },
		{ 86, 399, 40 },
		{ 86, 400, 40 },
		{ 86, 401, 40 },
		{ 86, 402, 40 },
		{ 86, 403, 40 },
		{ 86, 404, 40 },
		{ 86, 405, 40 },
		{ 86, 406, 40 },
		{ 87, 364, 40 },
		{ 87, 365, 40 },
		{ 87, 366, 40 },
		{ 87, 367, 40 },
		{ 87, 368, 40 },
		{ 87, 369, 40 },
		{ 87, 370, 40 },
		{ 87, 371, 40 },
		{ 87, 372, 40 },
		{ 87, 373, 40 },
		{ 87, 374, 40 },
		{ 87, 375, 40 },
		{ 87, 376, 40 },
		{ 87, 377, 40 },
		{ 87, 378, 40 },
		{ 87, 379, 40 },
		{ 87, 380, 40 },
		{ 87, 381, 40 },
		{ 87, 382, 40 },
		{ 87, 383, 40 },
		{ 87, 384, 40 },
		{ 87, 385, 40 },
		{ 87, 386, 40 },
		{ 87, 387, 40 },
		{ 87, 388, 40 },
		{ 87, 389, 40 },
		{ 87, 390, 40 },
		{ 87, 391, 40 },
		{ 87, 392, 40 },
		{ 87, 393, 40 },
		{ 87, 394, 40 },
		{ 87, 395, 40 },
		{ 87, 396, 40 },
		{ 87, 397, 40 },
		{ 87, 398, 40 },
		{ 87, 399, 40 },
		{ 87, 400, 40 },
		{ 87, 401, 40 },
		{ 87, 402, 40 },
		{ 87, 403, 40 },
		{ 87, 404, 40 },
		{ 87, 405, 40 },
		{ 88, 363, 40 },
		{ 88, 364, 40 },
		{ 88, 365, 40 },
		{ 88, 366, 40 },
		{ 88, 367, 40 },
		{ 88, 368, 40 },
		{ 88, 369, 40 },
		{ 88, 370, 40 },
		{ 88, 371, 40 },
		{ 88, 372, 40 },
		{ 88, 373, 40 },
		{ 88, 374, 40 },
		{ 88, 375, 40 },
		{ 88, 376, 40 },
		{ 88, 377, 40 },
		{ 88, 378, 40 },
		{ 88, 379, 40 },
		{ 88, 380, 40 },
		{ 88, 381, 40 },
		{ 88, 382, 40 },
		{ 88, 383, 40 },
		{ 88, 384, 40 },
		{ 88, 385, 40 },
		{ 88, 386, 40 },
		{ 88, 387, 40 },
		{ 88, 388, 40 },
		{ 88, 389, 40 },
		{ 88, 390, 40 },
		{ 88, 391, 40 },
		{ 88, 392, 40 },
		{ 88, 393, 40 },
		{ 88, 394, 40 },
		{ 88, 395, 40 },
		{ 88, 396, 40 },
		{ 88, 397, 40 },
		{ 88, 398, 40 },
		{ 88, 399, 40 },
		{ 88, 400, 40 },
		{ 88, 401, 40 },
		{ 88, 402, 40 },
		{ 88, 403, 40 },
		{ 88, 404, 40 },
		{ 89, 363, 40 },
		{ 89, 364, 40 },
		{ 89, 365, 40 },
		{ 89, 366, 40 },
		{ 89, 367, 40 },
		{ 89, 368, 40 },
		{ 89, 369, 40 },
		{ 89, 370, 40 },
		{ 89, 371, 40 },
		{ 89, 372, 40 },
		{ 89, 373, 40 },
		{ 89, 374, 40 },
		{ 89, 375, 40 },
		{ 89, 376, 40 },
		{ 89, 377, 40 },
		{ 89, 378, 40 },
		{ 89, 379, 40 },
		{ 89, 380, 40 },
		{ 89, 381, 40 },
		{ 89, 382, 40 },
		{ 89, 383, 40 },
		{ 89, 384, 40 },
		{ 89, 385, 40 },
		{ 89, 386, 40 },
		{ 89, 387, 40 },
		{ 89, 388, 40 },
		{ 89, 389, 40 },
		{ 89, 390, 40 },
		{ 89, 391, 40 },
		{ 89, 392, 40 },
		{ 89, 393, 40 },
		{ 89, 394, 40 },
		{ 89, 395, 40 },
		{ 89, 396, 40 },
		{ 89, 397, 40 },
		{ 89, 398, 40 },
		{ 89, 399, 40 },
		{ 89, 400, 40 },
		{ 89, 401, 40 },
		{ 89, 402, 40 },
		{ 89, 403, 40 },
		{ 90, 362, 40 },
		{ 90, 363, 40 },
		{ 90, 364, 40 },
		{ 90, 365, 40 },
		{ 90, 366, 40 },
		{ 90, 367, 40 },
		{ 90, 368, 40 },
		{ 90, 369, 40 },
		{ 90, 370, 40 },
		{ 90, 371, 40 },
		{ 90, 372, 40 },
		{ 90, 373, 40 },
		{ 90, 374, 40 },
		{ 90, 375, 40 },
		{ 90, 376, 40 },
		{ 90, 377, 40 },
		{ 90, 378, 40 },
		{ 90, 379, 40 },
		{ 90, 380, 40 },
		{ 90, 381, 40 },
		{ 90, 382, 40 },
		{ 90, 383, 40 },
		{ 90, 384, 40 },
		{ 90, 385, 40 },
		{ 90, 386, 40 },
		{ 90, 387, 40 },
		{ 90, 388, 40 },
		{ 90, 389, 40 },
		{ 90, 390, 40 },
		{ 90, 391, 40 },
		{ 90, 392, 40 },
		{ 90, 393, 40 },
		{ 90, 394, 40 },
		{ 90, 395, 40 },
		{ 90, 396, 40 },
		{ 90, 397, 40 },
		{ 90, 398, 40 },
		{ 90, 399, 40 },
		{ 90, 400, 40 },
		{ 90, 401, 40 },
		{ 90, 402, 40 },
		{ 90, 403, 40 },
		{ 91, 361, 40 },
		{ 91, 362, 40 },
		{ 91, 363, 40 },
		{ 91, 364, 40 },
		{ 91, 365, 40 },
		{ 91, 366, 40 },
		{ 91, 367, 40 },
		{ 91, 368, 40 },
		{ 91, 369, 40 },
		{ 91, 370, 40 },
		{ 91, 371, 40 },
		{ 91, 372, 40 },
		{ 91, 373, 40 },
		{ 91, 374, 40 },
		{ 91, 375, 40 },
		{ 91, 376, 40 },
		{ 91, 377, 40 },
		{ 91, 378, 40 },
		{ 91, 379, 40 },
		{ 91, 380, 40 },
		{ 91, 381, 40 },
		{ 91, 382, 40 },
		{ 91, 383, 40 },
		{ 91, 384, 40 },
		{ 91, 385, 40 },
		{ 91, 386, 40 },
		{ 91, 387, 40 },
		{ 91, 388, 40 },
		{ 91, 389, 40 },
		{ 91, 390, 40 },
		{ 91, 391, 40 },
		{ 91, 392, 40 },
		{ 91, 393, 40 },
		{ 91, 394, 40 },
		{ 91, 395, 40 },
		{ 91, 396, 40 },
		{ 91, 397, 40 },
		{ 91, 398, 40 },
		{ 91, 399, 40 },
		{ 91, 400, 40 },
		{ 91, 401, 40 },
		{ 91, 402, 40 },
		{ 92, 361, 40 },
		{ 92, 362, 40 },
		{ 92, 363, 40 },
		{ 92, 364, 40 },
		{ 92, 365, 40 },
		{ 92, 366, 40 },
		{ 92, 367, 40 },
		{ 92, 368, 40 },
		{ 92, 369, 40 },
		{ 92, 370, 40 },
		{ 92, 371, 40 },
		{ 92, 372, 40 },
		{ 92, 373, 40 },
		{ 92, 374, 40 },
		{ 92, 375, 40 },
		{ 92, 376, 40 },
		{ 92, 377, 40 },
		{ 92, 378, 40 },
		{ 92, 379, 40 },
		{ 92, 380, 40 },
		{ 92, 381, 40 },
		{ 92, 382, 40 },
		{ 92, 383, 40 },
		{ 92, 384, 40 },
		{ 92, 385, 40 },
		{ 92, 386, 40 },
		{ 92, 387, 40 },
		{ 92, 388, 40 },
		{ 92, 389, 40 },
		{ 92, 390, 40 },
		{ 92, 391, 40 },
		{ 92, 392, 40 },
		{ 92, 393, 40 },
		{ 92, 394, 40 },
		{ 92, 395, 40 },
		{ 92, 396, 40 },
		{ 92, 397, 40 },
		{ 92, 398, 40 },
		{ 92, 399, 40 },
		{ 92, 400, 40 },
		{ 92, 401, 40 },
		{ 93, 360, 40 },
		{ 93, 361, 40 },
		{ 93, 362, 40 },
		{ 93, 363, 40 },
		{ 93, 364, 40 },
		{ 93, 365, 40 },
		{ 93, 366, 40 },
		{ 93, 367, 40 },
		{ 93, 368, 40 },
		{ 93, 369, 40 },
		{ 93, 370, 40 },
		{ 93, 371, 40 },
		{ 93, 372, 40 },
		{ 93, 373, 40 },
		{ 93, 374, 40 },
		{ 93, 375, 40 },
		{ 93, 376, 40 },
		{ 93, 377, 40 },
		{ 93, 378, 40 },
		{ 93, 379, 40 },
		{ 93, 380, 40 },
		{ 93, 381, 40 },
		{ 93, 382, 40 },
		{ 93, 383, 40 },
		{ 93, 384, 40 },
		{ 93, 385, 40 },
		{ 93, 386, 40 },
		{ 93, 387, 40 },
		{ 93, 388, 40 },
		{ 93, 389, 40 },
		{ 93, 390, 40 },
		{ 93, 391, 40 },
		{ 93, 392, 40 },
		{ 93, 393, 40 },
		{ 93, 394, 40 },
		{ 93, 395, 40 },
		{ 93, 396, 40 },
		{ 93, 397, 40 },
		{ 93, 398, 40 },
		{ 93, 399, 40 },
		{ 93, 400, 40 },
		{ 94, 359, 40 },
		{ 94, 360, 40 },
		{ 94, 361, 40 },
		{ 94, 362, 40 },
		{ 94, 363, 40 },
		{ 94, 364, 40 },
		{ 94, 365, 40 },
		{ 94, 366, 40 },
		{ 94, 367, 40 },
		{ 94, 368, 40 },
		{ 94, 369, 40 },
		{ 94, 370, 40 },
		{ 94, 371, 40 },
		{ 94, 372, 40 },
		{ 94, 373, 40 },
		{ 94, 374, 40 },
		{ 94, 375, 40 },
		{ 94, 376, 40 },
		{ 94, 377, 40 },
		{ 94, 378, 40 },
		{ 94, 379, 40 },
		{ 94, 380, 40 },
		{ 94, 381, 40 },
		{ 94, 382, 40 },
		{ 94, 383, 40 },
		{ 94, 384, 40 },
		{ 94, 385, 40 },
		{ 94, 386, 40 },
		{ 94, 387, 40 },
		{ 94, 388, 40 },
		{ 94, 389, 40 },
		{ 94, 390, 40 },
		{ 94, 391, 40 },
		{ 94, 392, 40 },
		{ 94, 393, 40 },
		{ 94, 394, 40 },
		{ 94, 395, 40 },
		{ 94, 396, 40 },
		{ 94, 397, 40 },
		{ 94, 398, 40 },
		{ 94, 399, 40 },
		{ 95, 359, 40 },
		{ 95, 360, 40 },
		{ 95, 361, 40 },
		{ 95, 362, 40 },
		{ 95, 363, 40 },
		{ 95, 364, 40 },
		{ 95, 365, 40 },
		{ 95, 366, 40 },
		{ 95, 367, 40 },
		{ 95, 368, 40 },
		{ 95, 369, 40 },
		{ 95, 370, 40 },
		{ 95, 371, 40 },
		{ 95, 372, 40 },
		{ 95, 373, 40 },
		{ 95, 374, 40 },
		{ 95, 375, 40 },
		{ 95, 376, 40 },
		{ 95, 377, 40 },
		{ 95, 378, 40 },
		{ 95, 379, 40 },
		{ 95, 380, 40 },
		{ 95, 381, 40 },
		{ 95, 382, 40 },
		{ 95, 383, 40 },
		{ 95, 384, 40 },
		{ 95, 385, 40 },
		{ 95, 386, 40 },
		{ 95, 387, 40 },
		{ 95, 388, 40 },
		{ 95, 389, 40 },
		{ 95, 390, 40 },
		{ 95, 391, 40 },
		{ 95, 392, 40 },
		{ 95, 393, 40 },
		{ 95, 394, 40 },
		{ 95, 395, 40 },
		{ 95, 396, 40 },
		{ 95, 397, 40 },
		{ 95, 398, 40 },
		{ 96, 358, 40 },
		{ 96, 359, 40 },
		{ 96, 360, 40 },
		{ 96, 361, 40 },
		{ 96, 362, 40 },
		{ 96, 363, 40 },
		{ 96, 364, 40 },
		{ 96, 365, 40 },
		{ 96, 366, 40 },
		{ 96, 367, 40 },
		{ 96, 368, 40 },
		{ 96, 369, 40 },
		{ 96, 370, 40 },
		{ 96, 371, 40 },
		{ 96, 372, 40 },
		{ 96, 373, 40 },
		{ 96, 374, 40 },
		{ 96, 375, 40 },
		{ 96, 376, 40 },
		{ 96, 377, 40 },
		{ 96, 378, 40 },
		{ 96, 379, 40 },
		{ 96, 380, 40 },
		{ 96, 381, 40 },
		{ 96, 382, 40 },
		{ 96, 383, 40 },
		{ 96, 384, 40 },
		{ 96, 385, 40 },
		{ 96, 386, 40 },
		{ 96, 387, 40 },
		{ 96, 388, 40 },
		{ 96, 389, 40 },
		{ 96, 390, 40 },
		{ 96, 391, 40 },
		{ 96, 392, 40 },
		{ 96, 393, 40 },
		{ 96, 394, 40 },
		{ 96, 395, 40 },
		{ 96, 396, 40 },
		{ 96, 397, 40 },
		{ 97, 358, 40 },
		{ 97, 359, 40 },
		{ 97, 360, 40 },
		{ 97, 361, 40 },
		{ 97, 362, 40 },
		{ 97, 363, 40 },
		{ 97, 364, 40 },
		{ 97, 365, 40 },
		{ 97, 366, 40 },
		{ 97, 367, 40 },
		{ 97, 368, 40 },
		{ 97, 369, 40 },
		{ 97, 370, 40 },
		{ 97, 371, 40 },
		{ 97, 372, 40 },
		{ 97, 373, 40 },
		{ 97, 374, 40 },
		{ 97, 375, 40 },
		{ 97, 376, 40 },
		{ 97, 377, 40 },
		{ 97, 378, 40 },
		{ 97, 379, 40 },
		{ 97, 380, 40 },
		{ 97, 381, 40 },
		{ 97, 382, 40 },
		{ 97, 383, 40 },
		{ 97, 384, 40 },
		{ 97, 385, 40 },
		{ 97, 386, 40 },
		{ 97, 387, 40 },
		{ 97, 388, 40 },
		{ 97, 389, 40 },
		{ 97, 390, 40 },
		{ 97, 391, 40 },
		{ 97, 392, 40 },
		{ 97, 393, 40 },
		{ 97, 394, 40 },
		{ 97, 395, 40 },
		{ 97, 396, 40 },
		{ 98, 357, 40 },
		{ 98, 358, 40 },
		{ 98, 359, 40 },
		{ 98, 360, 40 },
		{ 98, 361, 40 },
		{ 98, 362, 40 },
		{ 98, 363, 40 },
		{ 98, 364, 40 },
		{ 98, 365, 40 },
		{ 98, 366, 40 },
		{ 98, 367, 40 },
		{ 98, 368, 40 },
		{ 98, 369, 40 },
		{ 98, 370, 40 },
		{ 98, 371, 40 },
		{ 98, 372, 40 },
		{ 98, 373, 40 },
		{ 98, 374, 40 },
		{ 98, 375, 40 },
		{ 98, 376, 40 },
		{ 98, 377, 40 },
		{ 98, 378, 40 },
		{ 98, 379, 40 },
		{ 98, 380, 40 },
		{ 98, 381, 40 },
		{ 98, 382, 40 },
		{ 98, 383, 40 },
		{ 98, 384, 40 },
		{ 98, 385, 40 },
		{ 98, 386, 40 },
		{ 98, 387, 40 },
		{ 98, 388, 40 },
		{ 98, 389, 40 },
		{ 98, 390, 40 },
		{ 98, 391, 40 },
		{ 98, 392, 40 },
		{ 98, 393, 40 },
		{ 98, 394, 40 },
		{ 98, 395, 40 },
		{ 98, 396, 40 },
		{ 99, 356, 40 },
		{ 99, 357, 40 },
		{ 99, 358, 40 },
		{ 99, 359, 40 },
		{ 99, 360, 40 },
		{ 99, 361, 40 },
		{ 99, 362, 40 },
		{ 99, 363, 40 },
		{ 99, 364, 40 },
		{ 99, 365, 40 },
		{ 99, 366, 40 },
		{ 99, 367, 40 },
		{ 99, 368, 40 },
		{ 99, 369, 40 },
		{ 99, 370, 40 },
		{ 99, 371, 40 },
		{ 99, 372, 40 },
		{ 99, 373, 40 },
		{ 99, 374, 40 },
		{ 99, 375, 40 },
		{ 99, 376, 40 },
		{ 99, 377, 40 },
		{ 99, 378, 40 },
		{ 99, 379, 40 },
		{ 99, 380, 40 },
		{ 99, 381, 40 },
		{ 99, 382, 40 },
		{ 99, 383, 40 },
		{ 99, 384, 40 },
		{ 99, 385, 40 },
		{ 99, 386, 40 },
		{ 99, 387, 40 },
		{ 99, 388, 40 },
		{ 99, 389, 40 },
		{ 99, 390, 40 },
		{ 99, 391, 40 },
		{ 99, 392, 40 },
		{ 99, 393, 40 },
		{ 99, 394, 40 },
		{ 99, 395, 40 },
		{ 100, 356, 40 },
		{ 100, 357, 40 },
		{ 100, 358, 40 },
		{ 100, 359, 40 },
		{ 100, 360, 40 },
		{ 100, 361, 40 },
		{ 100, 362, 40 },
		{ 100, 363, 40 },
		{ 100, 364, 40 },
		{ 100, 365, 40 },
		{ 100, 366, 40 },
		{ 100, 367, 40 },
		{ 100, 368, 40 },
		{ 100, 369, 40 },
		{ 100, 370, 40 },
		{ 100, 371, 40 },
		{ 100, 372, 40 },
		{ 100, 373, 40 },
		{ 100, 374, 40 },
		{ 100, 375, 40 },
		{ 100, 376, 40 },
		{ 100, 377, 40 },
		{ 100, 378, 40 },
		{ 100, 379, 40 },
		{ 100, 380, 40 },
		{ 100, 381, 40 },
		{ 100, 382, 40 },
		{ 100, 383, 40 },
		{ 100, 384, 40 },
		{ 100, 385, 40 },
		{ 100, 386, 40 },
		{ 100, 387, 40 },
		{ 100, 388, 40 },
		{ 100, 389, 40 },
		{ 100, 390, 40 },
		{ 100, 391, 40 },
		{ 100, 392, 40 },
		{ 100, 393, 40 },
		{ 100, 394, 40 },
		{ 101, 355, 40 },
		{ 101, 356, 40 },
		{ 101, 357, 40 },
		{ 101, 358, 40 },
		{ 101, 359, 40 },
		{ 101, 360, 40 },
		{ 101, 361, 40 },
		{ 101, 362, 40 },
		{ 101, 363, 40 },
		{ 101, 364, 40 },
		{ 101, 365, 40 },
		{ 101, 366, 40 },
		{ 101, 367, 40 },
		{ 101, 368, 40 },
		{ 101, 369, 40 },
		{ 101, 370, 40 },
		{ 101, 371, 40 },
		{ 101, 372, 40 },
		{ 101, 373, 40 },
		{ 101, 374, 40 },
		{ 101, 375, 40 },
		{ 101, 376, 40 },
		{ 101, 377, 40 },
		{ 101, 378, 40 },
		{ 101, 379, 40 },
		{ 101, 380, 40 },
		{ 101, 381, 40 },
		{ 101, 382, 40 },
		{ 101, 383, 40 },
		{ 101, 384, 40 },
		{ 101, 385, 40 },
		{ 101, 386, 40 },
		{ 101, 387, 40 },
		{ 101, 388, 40 },
		{ 101, 389, 40 },
		{ 101, 390, 40 },
		{ 101, 391, 40 },
		{ 101, 392, 40 },
		{ 101, 393, 40 },
		{ 102, 354, 40 },
		{ 102, 355, 40 },
		{ 102, 356, 40 },
		{ 102, 357, 40 },
		{ 102, 358, 40 },
		{ 102, 359, 40 },
		{ 102, 360, 40 },
		{ 102, 361, 40 },
		{ 102, 362, 40 },
		{ 102, 363, 40 },
		{ 102, 364, 40 },
		{ 102, 365, 40 },
		{ 102, 366, 40 },
		{ 102, 367, 40 },
		{ 102, 368, 40 },
		{ 102, 369, 40 },
		{ 102, 370, 40 },
		{ 102, 371, 40 },
		{ 102, 372, 40 },
		{ 102, 373, 40 },
		{ 102, 374, 40 },
		{ 102, 375, 40 },
		{ 102, 376, 40 },
		{ 102, 377, 40 },
		{ 102, 378, 40 },
		{ 102, 379, 40 },
		{ 102, 380, 40 },
		{ 102, 381, 40 },
		{ 102, 382, 40 },
		{ 102, 383, 40 },
		{ 102, 384, 40 },
		{ 102, 385, 40 },
		{ 102, 386, 40 },
		{ 102, 387, 40 },
		{ 102, 388, 40 },
		{ 102, 389, 40 },
		{ 102, 390, 40 },
		{ 102, 391, 40 },
		{ 102, 392, 40 },
		{ 103, 354, 40 },
		{ 103, 355, 40 },
		{ 103, 356, 40 },
		{ 103, 357, 40 },
		{ 103, 358, 40 },
		{ 103, 359, 40 },
		{ 103, 360, 40 },
		{ 103, 361, 40 },
		{ 103, 362, 40 },
		{ 103, 363, 40 },
		{ 103, 364, 40 },
		{ 103, 365, 40 },
		{ 103, 366, 40 },
		{ 103, 367, 40 },
		{ 103, 368, 40 },
		{ 103, 369, 40 },
		{ 103, 370, 40 },
		{ 103, 371, 40 },
		{ 103, 372, 40 },
		{ 103, 373, 40 },
		{ 103, 374, 40 },
		{ 103, 375, 40 },
		{ 103, 376, 40 },
		{ 103, 377, 40 },
		{ 103, 378, 40 },
		{ 103, 379, 40 },
		{ 103, 380, 40 },
		{ 103, 381, 40 },
		{ 103, 382, 40 },
		{ 103, 383, 40 },
		{ 103, 384, 40 },
		{ 103, 385, 40 },
		{ 103, 386, 40 },
		{ 103, 387, 40 },
		{ 103, 388, 40 },
		{ 103, 389, 40 },
		{ 103, 390, 40 },
		{ 103, 391, 40 },
		{ 104, 353, 40 },
		{ 104, 354, 40 },
		{ 104, 355, 40 },
		{ 104, 356, 40 },
		{ 104, 357, 40 },
		{ 104, 358, 40 },
		{ 104, 359, 40 },
		{ 104, 360, 40 },
		{ 104, 361, 40 },
		{ 104, 362, 40 },
		{ 104, 363, 40 },
		{ 104, 364, 40 },
		{ 104, 365, 40 },
		{ 104, 366, 40 },
		{ 104, 367, 40 },
		{ 104, 368, 40 },
		{ 104, 369, 40 },
		{ 104, 370, 40 },
		{ 104, 371, 40 },
		{ 104, 372, 40 },
		{ 104, 373, 40 },
		{ 104, 374, 40 },
		{ 104, 375, 40 },
		{ 104, 376, 40 },
		{ 104, 377, 40 },
		{ 104, 378, 40 },
		{ 104, 379, 40 },
		{ 104, 380, 40 },
		{ 104, 381, 40 },
		{ 104, 382, 40 },
		{ 104, 383, 40 },
		{ 104, 384, 40 },
		{ 104, 385, 40 },
		{ 104, 386, 40 },
		{ 104, 387, 40 },
		{ 104, 388, 40 },
		{ 104, 389, 40 },
		{ 104, 390, 40 },
		{ 105, 353, 40 },
		{ 105, 354, 40 },
		{ 105, 355, 40 },
		{ 105, 356, 40 },
		{ 105, 357, 40 },
		{ 105, 358, 40 },
		{ 105, 359, 40 },
		{ 105, 360, 40 },
		{ 105, 361, 40 },
		{ 105, 362, 40 },
		{ 105, 363, 40 },
		{ 105, 364, 40 },
		{ 105, 365, 40 },
		{ 105, 366, 40 },
		{ 105, 367, 40 },
		{ 105, 368, 40 },
		{ 105, 369, 40 },
		{ 105, 370, 40 },
		{ 105, 371, 40 },
		{ 105, 372, 40 },
		{ 105, 373, 40 },
		{ 105, 374, 40 },
		{ 105, 375, 40 },
		{ 105, 376, 40 },
		{ 105, 377, 40 },
		{ 105, 378, 40 },
		{ 105, 379, 40 },
		{ 105, 380, 40 },
		{ 105, 381, 40 },
		{ 105, 382, 40 },
		{ 105, 383, 40 },
		{ 105, 384, 40 },
		{ 105, 385, 40 },
		{ 105, 386, 40 },
		{ 105, 387, 40 },
		{ 105, 388, 40 },
		{ 105, 389, 40 },
		{ 106, 352, 40 },
		{ 106, 353, 40 },
		{ 106, 354, 40 },
		{ 106, 355, 40 },
		{ 106, 356, 40 },
		{ 106, 357, 40 },
		{ 106, 358, 40 },
		{ 106, 359, 40 },
		{ 106, 360, 40 },
		{ 106, 361, 40 },
		{ 106, 362, 40 },
		{ 106, 363, 40 },
		{ 106, 364, 40 },
		{ 106, 365, 40 },
		{ 106, 366, 40 },
		{ 106, 367, 40 },
		{ 106, 368, 40 },
		{ 106, 369, 40 },
		{ 106, 370, 40 },
		{ 106, 371, 40 },
		{ 106, 372, 40 },
		{ 106, 373, 40 },
		{ 106, 374, 40 },
		{ 106, 375, 40 },
		{ 106, 376, 40 },
		{ 106, 377, 40 },
		{ 106, 378, 40 },
		{ 106, 379, 40 },
		{ 106, 380, 40 },
		{ 106, 381, 40 },
		{ 106, 382, 40 },
		{ 106, 383, 40 },
		{ 106, 384, 40 },
		{ 106, 385, 40 },
		{ 106, 386, 40 },
		{ 106, 387, 40 },
		{ 106, 388, 40 },
		{ 106, 389, 40 },
		{ 107, 351, 40 },
		{ 107, 352, 40 },
		{ 107, 353, 40 },
		{ 107, 354, 40 },
		{ 107, 355, 40 },
		{ 107, 356, 40 },
		{ 107, 357, 40 },
		{ 107, 358, 40 },
		{ 107, 359, 40 },
		{ 107, 360, 40 },
		{ 107, 361, 40 },
		{ 107, 362, 40 },
		{ 107, 363, 40 },
		{ 107, 364, 40 },
		{ 107, 365, 40 },
		{ 107, 366, 40 },
		{ 107, 367, 40 },
		{ 107, 368, 40 },
		{ 107, 369, 40 },
		{ 107, 370, 40 },
		{ 107, 371, 40 },
		{ 107, 372, 40 },
		{ 107, 373, 40 },
		{ 107, 374, 40 },
		{ 107, 375, 40 },
		{ 107, 376, 40 },
		{ 107, 377, 40 },
		{ 107, 378, 40 },
		{ 107, 379, 40 },
		{ 107, 380, 40 },
		{ 107, 381, 40 },
		{ 107, 382, 40 },
		{ 107, 383, 40 },
		{ 107, 384, 40 },
		{ 107, 385, 40 },
		{ 107, 386, 40 },
		{ 107, 387, 40 },
		{ 107, 388, 40 },
		{ 108, 351, 40 },
		{ 108, 352, 40 },
		{ 108, 353, 40 },
		{ 108, 354, 40 },
		{ 108, 355, 40 },
		{ 108, 356, 40 },
		{ 108, 357, 40 },
		{ 108, 358, 40 },
		{ 108, 359, 40 },
		{ 108, 360, 40 },
		{ 108, 361, 40 },
		{ 108, 362, 40 },
		{ 108, 363, 40 },
		{ 108, 364, 40 },
		{ 108, 365, 40 },
		{ 108, 366, 40 },
		{ 108, 367, 40 },
		{ 108, 368, 40 },
		{ 108, 369, 40 },
		{ 108, 370, 40 },
		{ 108, 371, 40 },
		{ 108, 372, 40 },
		{ 108, 373, 40 },
		{ 108, 374, 40 },
		{ 108, 375, 40 },
		{ 108, 376, 40 },
		{ 108, 377, 40 },
		{ 108, 378, 40 },
		{ 108, 379, 40 },
		{ 108, 380, 40 },
		{ 108, 381, 40 },
		{ 108, 382, 40 },
		{ 108, 383, 40 },
		{ 108, 384, 40 },
		{ 108, 385, 40 },
		{ 108, 386, 40 },
		{ 108, 387, 40 },
		{ 109, 350, 40 },
		{ 109, 351, 40 },
		{ 109, 352, 40 },
		{ 109, 353, 40 },
		{ 109, 354, 40 },
		{ 109, 355, 40 },
		{ 109, 356, 40 },
		{ 109, 357, 40 },
		{ 109, 358, 40 },
		{ 109, 359, 40 },
		{ 109, 360, 40 },
		{ 109, 361, 40 },
		{ 109, 362, 40 },
		{ 109, 363, 40 },
		{ 109, 364, 40 },
		{ 109, 365, 40 },
		{ 109, 366, 40 },
		{ 109, 367, 40 },
		{ 109, 368, 40 },
		{ 109, 369, 40 },
		{ 109, 370, 40 },
		{ 109, 371, 40 },
		{ 109, 372, 40 },
		{ 109, 373, 40 },
		{ 109, 374, 40 },
		{ 109, 375, 40 },
		{ 109, 376, 40 },
		{ 109, 377, 40 },
		{ 109, 378, 40 },
		{ 109, 379, 40 },
		{ 109, 380, 40 },
		{ 109, 381, 40 },
		{ 109, 382, 40 },
		{ 109, 383, 40 },
		{ 109, 384, 40 },
		{ 109, 385, 40 },
		{ 109, 386, 40 },
		{ 110, 349, 40 },
		{ 110, 350, 40 },
		{ 110, 351, 40 },
		{ 110, 352, 40 },
		{ 110, 353, 40 },
		{ 110, 354, 40 },
		{ 110, 355, 40 },
		{ 110, 356, 40 },
		{ 110, 357, 40 },
		{ 110, 358, 40 },
		{ 110, 359, 40 },
		{ 110, 360, 40 },
		{ 110, 361, 40 },
		{ 110, 362, 40 },
		{ 110, 363, 40 },
		{ 110, 364, 40 },
		{ 110, 365, 40 },
		{ 110, 366, 40 },
		{ 110, 367, 40 },
		{ 110, 368, 40 },
		{ 110, 369, 40 },
		{ 110, 370, 40 },
		{ 110, 371, 40 },
		{ 110, 372, 40 },
		{ 110, 373, 40 },
		{ 110, 374, 40 },
		{ 110, 375, 40 },
		{ 110, 376, 40 },
		{ 110, 377, 40 },
		{ 110, 378, 40 },
		{ 110, 379, 40 },
		{ 110, 380, 40 },
		{ 110, 381, 40 },
		{ 110, 382, 40 },
		{ 110, 383, 40 },
		{ 110, 384, 40 },
		{ 110, 385, 40 },
		{ 111, 349, 40 },
		{ 111, 350, 40 },
		{ 111, 351, 40 },
		{ 111, 352, 40 },
		{ 111, 353, 40 },
		{ 111, 354, 40 },
		{ 111, 355, 40 },
		{ 111, 356, 40 },
		{ 111, 357, 40 },
		{ 111, 358, 40 },
		{ 111, 359, 40 },
		{ 111, 360, 40 },
		{ 111, 361, 40 },
		{ 111, 362, 40 },
		{ 111, 363, 40 },
		{ 111, 364, 40 },
		{ 111, 365, 40 },
		{ 111, 366, 40 },
		{ 111, 367, 40 },
		{ 111, 368, 40 },
		{ 111, 369, 40 },
		{ 111, 370, 40 },
		{ 111, 371, 40 },
		{ 111, 372, 40 },
		{ 111, 373, 40 },
		{ 111, 374, 40 },
		{ 111, 375, 40 },
		{ 111, 376, 40 },
		{ 111, 377, 40 },
		{ 111, 378, 40 },
		{ 111, 379, 40 },
		{ 111, 380, 40 },
		{ 111, 381, 40 },
		{ 111, 382, 40 },
		{ 111, 383, 40 },
		{ 111, 384, 40 },
		{ 112, 348, 40 },
		{ 112, 349, 40 },
		{ 112, 350, 40 },
		{ 112, 351, 40 },
		{ 112, 352, 40 },
		{ 112, 353, 40 },
		{ 112, 354, 40 },
		{ 112, 355, 40 },
		{ 112, 356, 40 },
		{ 112, 357, 40 },
		{ 112, 358, 40 },
		{ 112, 359, 40 },
		{ 112, 360, 40 },
		{ 112, 361, 40 },
		{ 112, 362, 40 },
		{ 112, 363, 40 },
		{ 112, 364, 40 },
		{ 112, 365, 40 },
		{ 112, 366, 40 },
		{ 112, 367, 40 },
		{ 112, 368, 40 },
		{ 112, 369, 40 },
		{ 112, 370, 40 },
		{ 112, 371, 40 },
		{ 112, 372, 40 },
		{ 112, 373, 40 },
		{ 112, 374, 40 },
		{ 112, 375, 40 },
		{ 112, 376, 40 },
		{ 112, 377, 40 },
		{ 112, 378, 40 },
		{ 112, 379, 40 },
		{ 112, 380, 40 },
		{ 112, 381, 40 },
		{ 112, 382, 40 },
		{ 112, 383, 40 },
		{ 113, 348, 40 },
		{ 113, 349, 40 },
		{ 113, 350, 40 },
		{ 113, 351, 40 },
		{ 113, 352, 40 },
		{ 113, 353, 40 },
		{ 113, 354, 40 },
		{ 113, 355, 40 },
		{ 113, 356, 40 },
		{ 113, 357, 40 },
		{ 113, 358, 40 },
		{ 113, 359, 40 },
		{ 113, 360, 40 },
		{ 113, 361, 40 },
		{ 113, 362, 40 },
		{ 113, 363, 40 },
		{ 113, 364, 40 },
		{ 113, 365, 40 },
		{ 113, 366, 40 },
		{ 113, 367, 40 },
		{ 113, 368, 40 },
		{ 113, 369, 40 },
		{ 113, 370, 40 },
		{ 113, 371, 40 },
		{ 113, 372, 40 },
		{ 113, 373, 40 },
		{ 113, 374, 40 },
		{ 113, 375, 40 },
		{ 113, 376, 40 },
		{ 113, 377, 40 },
		{ 113, 378, 40 },
		{ 113, 379, 40 },
		{ 113, 380, 40 },
		{ 113, 381, 40 },
		{ 113, 382, 40 },
		{ 114, 347, 40 },
		{ 114, 348, 40 },
		{ 114, 349, 40 },
		{ 114, 350, 40 },
		{ 114, 351, 40 },
		{ 114, 352, 40 },
		{ 114, 353, 40 },
		{ 114, 354, 40 },
		{ 114, 355, 40 },
		{ 114, 356, 40 },
		{ 114, 357, 40 },
		{ 114, 358, 40 },
		{ 114, 359, 40 },
		{ 114, 360, 40 },
		{ 114, 361, 40 },
		{ 114, 362, 40 },
		{ 114, 363, 40 },
		{ 114, 364, 40 },
		{ 114, 365, 40 },
		{ 114, 366, 40 },
		{ 114, 367, 40 },
		{ 114, 368, 40 },
		{ 114, 369, 40 },
		{ 114, 370, 40 },
		{ 114, 371, 40 },
		{ 114, 372, 40 },
		{ 114, 373, 40 },
		{ 114, 374, 40 },
		{ 114, 375, 40 },
		{ 114, 376, 40 },
		{ 114, 377, 40 },
		{ 114, 378, 40 },
		{ 114, 379, 40 },
		{ 114, 380, 40 },
		{ 114, 381, 40 },
		{ 114, 382, 40 },
		{ 115, 346, 40 },
		{ 115, 347, 40 },
		{ 115, 348, 40 },
		{ 115, 349, 40 },
		{ 115, 350, 40 },
		{ 115, 351, 40 },
		{ 115, 352, 40 },
		{ 115, 353, 40 },
		{ 115, 354, 40 },
		{ 115, 355, 40 },
		{ 115, 356, 40 },
		{ 115, 357, 40 },
		{ 115, 358, 40 },
		{ 115, 359, 40 },
		{ 115, 360, 40 },
		{ 115, 361, 40 },
		{ 115, 362, 40 },
		{ 115, 363, 40 },
		{ 115, 364, 40 },
		{ 115, 365, 40 },
		{ 115, 366, 40 },
		{ 115, 367, 40 },
		{ 115, 368, 40 },
		{ 115, 369, 40 },
		{ 115, 370, 40 },
		{ 115, 371, 40 },
		{ 115, 372, 40 },
		{ 115, 373, 40 },
		{ 115, 374, 40 },
		{ 115, 375, 40 },
		{ 115, 376, 40 },
		{ 115, 377, 40 },
		{ 115, 378, 40 },
		{ 115, 379, 40 },
		{ 115, 380, 40 },
		{ 115, 381, 40 },
		{ 116, 346, 40 },
		{ 116, 347, 40 },
		{ 116, 348, 40 },
		{ 116, 349, 40 },
		{ 116, 350, 40 },
		{ 116, 351, 40 },
		{ 116, 352, 40 },
		{ 116, 353, 40 },
		{ 116, 354, 40 },
		{ 116, 355, 40 },
		{ 116, 356, 40 },
		{ 116, 357, 40 },
		{ 116, 358, 40 },
		{ 116, 359, 40 },
		{ 116, 360, 40 },
		{ 116, 361, 40 },
		{ 116, 362, 40 },
		{ 116, 363, 40 },
		{ 116, 364, 40 },
		{ 116, 365, 40 },
		{ 116, 366, 40 },
		{ 116, 367, 40 },
		{ 116, 368, 40 },
		{ 116, 369, 40 },
		{ 116, 370, 40 },
		{ 116, 371, 40 },
		{ 116, 372, 40 },
		{ 116, 373, 40 },
		{ 116, 374, 40 },
		{ 116, 375, 40 },
		{ 116, 376, 40 },
		{ 116, 377, 40 },
		{ 116, 378, 40 },
		{ 116, 379, 40 },
		{ 116, 380, 40 },
		{ 117, 345, 40 },
		{ 117, 346, 40 },
		{ 117, 347, 40 },
		{ 117, 348, 40 },
		{ 117, 349, 40 },
		{ 117, 350, 40 },
		{ 117, 351, 40 },
		{ 117, 352, 40 },
		{ 117, 353, 40 },
		{ 117, 354, 40 },
		{ 117, 355, 40 },
		{ 117, 356, 40 },
		{ 117, 357, 40 },
		{ 117, 358, 40 },
		{ 117, 359, 40 },
		{ 117, 360, 40 },
		{ 117, 361, 40 },
		{ 117, 362, 40 },
		{ 117, 363, 40 },
		{ 117, 364, 40 },
		{ 117, 365, 40 },
		{ 117, 366, 40 },
		{ 117, 367, 40 },
		{ 117, 368, 40 },
		{ 117, 369, 40 },
		{ 117, 370, 40 },
		{ 117, 371, 40 },
		{ 117, 372, 40 },
		{ 117, 373, 40 },
		{ 117, 374, 40 },
		{ 117, 375, 40 },
		{ 117, 376, 40 },
		{ 117, 377, 40 },
		{ 117, 378, 40 },
		{ 117, 379, 40 },
		{ 118, 344, 40 },
		{ 118, 345, 40 },
		{ 118, 346, 40 },
		{ 118, 347, 40 },
		{ 118, 348, 40 },
		{ 118, 349, 40 },
		{ 118, 350, 40 },
		{ 118, 351, 40 },
		{ 118, 352, 40 },
		{ 118, 353, 40 },
		{ 118, 354, 40 },
		{ 118, 355, 40 },
		{ 118, 356, 40 },
		{ 118, 357, 40 },
		{ 118, 358, 40 },
		{ 118, 359, 40 },
		{ 118, 360, 40 },
		{ 118, 361, 40 },
		{ 118, 362, 40 },
		{ 118, 363, 40 },
		{ 118, 364, 40 },
		{ 118, 365, 40 },
		{ 118, 366, 40 },
		{ 118, 367, 40 },
		{ 118, 368, 40 },
		{ 118, 369, 40 },
		{ 118, 370, 40 },
		{ 118, 371, 40 },
		{ 118, 372, 40 },
		{ 118, 373, 40 },
		{ 118, 374, 40 },
		{ 118, 375, 40 },
		{ 118, 376, 40 },
		{ 118, 377, 40 },
		{ 118, 378, 40 },
		{ 119, 344, 40 },
		{ 119, 345, 40 },
		{ 119, 346, 40 },
		{ 119, 347, 40 },
		{ 119, 348, 40 },
		{ 119, 349, 40 },
		{ 119, 350, 40 },
		{ 119, 351, 40 },
		{ 119, 352, 40 },
		{ 119, 353, 40 },
		{ 119, 354, 40 },
		{ 119, 355, 40 },
		{ 119, 356, 40 },
		{ 119, 357, 40 },
		{ 119, 358, 40 },
		{ 119, 359, 40 },
		{ 119, 360, 40 },
		{ 119, 361, 40 },
		{ 119, 362, 40 },
		{ 119, 363, 40 },
		{ 119, 364, 40 },
		{ 119, 365, 40 },
		{ 119, 366, 40 },
		{ 119, 367, 40 },
		{ 119, 368, 40 },
		{ 119, 369, 40 },
		{ 119, 370, 40 },
		{ 119, 371, 40 },
		{ 119, 372, 40 },
		{ 119, 373, 40 },
		{ 119, 374, 40 },
		{ 119, 375, 40 },
		{ 119, 376, 40 },
		{ 119, 377, 40 },
		{ 120, 343, 40 },
		{ 120, 344, 40 },
		{ 120, 345, 40 },
		{ 120, 346, 40 },
		{ 120, 347, 40 },
		{ 120, 348, 40 },
		{ 120, 349, 40 },
		{ 120, 350, 40 },
		{ 120, 351, 40 },
		{ 120, 352, 40 },
		{ 120, 353, 40 },
		{ 120, 354, 40 },
		{ 120, 355, 40 },
		{ 120, 356, 40 },
		{ 120, 357, 40 },
		{ 120, 358, 40 },
		{ 120, 359, 40 },
		{ 120, 360, 40 },
		{ 120, 361, 40 },
		{ 120, 362, 40 },
		{ 120, 363, 40 },
		{ 120, 364, 40 },
		{ 120, 365, 40 },
		{ 120, 366, 40 },
		{ 120, 367, 40 },
		{ 120, 368, 40 },
		{ 120, 369, 40 },
		{ 120, 370, 40 },
		{ 120, 371, 40 },
		{ 120, 372, 40 },
		{ 120, 373, 40 },
		{ 120, 374, 40 },
		{ 120, 375, 40 },
		{ 120, 376, 40 },
		{ 121, 343, 40 },
		{ 121, 344, 40 },
		{ 121, 345, 40 },
		{ 121, 346, 40 },
		{ 121, 347, 40 },
		{ 121, 348, 40 },
		{ 121, 349, 40 },
		{ 121, 350, 40 },
		{ 121, 351, 40 },
		{ 121, 352, 40 },
		{ 121, 353, 40 },
		{ 121, 354, 40 },
		{ 121, 355, 40 },
		{ 121, 356, 40 },
		{ 121, 357, 40 },
		{ 121, 358, 40 },
		{ 121, 359, 40 },
		{ 121, 360, 40 },
		{ 121, 361, 40 },
		{ 121, 362, 40 },
		{ 121, 363, 40 },
		{ 121, 364, 40 },
		{ 121, 365, 40 },
		{ 121, 366, 40 },
		{ 121, 367, 40 },
		{ 121, 368, 40 },
		{ 121, 369, 40 },
		{ 121, 370, 40 },
		{ 121, 371, 40 },
		{ 121, 372, 40 },
		{ 121, 373, 40 },
		{ 121, 374, 40 },
		{ 121, 375, 40 },
		{ 122, 342, 40 },
		{ 122, 343, 40 },
		{ 122, 344, 40 },
		{ 122, 345, 40 },
		{ 122, 346, 40 },
		{ 122, 347, 40 },
		{ 122, 348, 40 },
		{ 122, 349, 40 },
		{ 122, 350, 40 },
		{ 122, 351, 40 },
		{ 122, 352, 40 },
		{ 122, 353, 40 },
		{ 122, 354, 40 },
		{ 122, 355, 40 },
		{ 122, 356, 40 },
		{ 122, 357, 40 },
		{ 122, 358, 40 },
		{ 122, 359, 40 },
		{ 122, 360, 40 },
		{ 122, 361, 40 },
		{ 122, 362, 40 },
		{ 122, 363, 40 },
		{ 122, 364, 40 },
		{ 122, 365, 40 },
		{ 122, 366, 40 },
		{ 122, 367, 40 },
		{ 122, 368, 40 },
		{ 122, 369, 40 },
		{ 122, 370, 40 },
		{ 122, 371, 40 },
		{ 122, 372, 40 },
		{ 122, 373, 40 },
		{ 122, 374, 40 },
		{ 122, 375, 40 },
		{ 123, 341, 40 },
		{ 123, 342, 40 },
		{ 123, 343, 40 },
		{ 123, 344, 40 },
		{ 123, 345, 40 },
		{ 123, 346, 40 },
		{ 123, 347, 40 },
		{ 123, 348, 40 },
		{ 123, 349, 40 },
		{ 123, 350, 40 },
		{ 123, 351, 40 },
		{ 123, 352, 40 },
		{ 123, 353, 40 },
		{ 123, 354, 40 },
		{ 123, 355, 40 },
		{ 123, 356, 40 },
		{ 123, 357, 40 },
		{ 123, 358, 40 },
		{ 123, 359, 40 },
		{ 123, 360, 40 },
		{ 123, 361, 40 },
		{ 123, 362, 40 },
		{ 123, 363, 40 },
		{ 123, 364, 40 },
		{ 123, 365, 40 },
		{ 123, 366, 40 },
		{ 123, 367, 40 },
		{ 123, 368, 40 },
		{ 123, 369, 40 },
		{ 123, 370, 40 },
		{ 123, 371, 40 },
		{ 123, 372, 40 },
		{ 123, 373, 40 },
		{ 123, 374, 40 },
		{ 124, 341, 40 },
		{ 124, 342, 40 },
		{ 124, 343, 40 },
		{ 124, 344, 40 },
		{ 124, 345, 40 },
		{ 124, 346, 40 },
		{ 124, 347, 40 },
		{ 124, 348, 40 },
		{ 124, 349, 40 },
		{ 124, 350, 40 },
		{ 124, 351, 40 },
		{ 124, 352, 40 },
		{ 124, 353, 40 },
		{ 124, 354, 40 },
		{ 124, 355, 40 },
		{ 124, 356, 40 },
		{ 124, 357, 40 },
		{ 124, 358, 40 },
		{ 124, 359, 40 },
		{ 124, 360, 40 },
		{ 124, 361, 40 },
		{ 124, 362, 40 },
		{ 124, 363, 40 },
		{ 124, 364, 40 },
		{ 124, 365, 40 },
		{ 124, 366, 40 },
		{ 124, 367, 40 },
		{ 124, 368, 40 },
		{ 124, 369, 40 },
		{ 124, 370, 40 },
		{ 124, 371, 40 },
		{ 124, 372, 40 },
		{ 124, 373, 40 },
		{ 125, 340, 40 },
		{ 125, 341, 40 },
		{ 125, 342, 40 },
		{ 125, 343, 40 },
		{ 125, 344, 40 },
		{ 125, 345, 40 },
		{ 125, 346, 40 },
		{ 125, 347, 40 },
		{ 125, 348, 40 },
		{ 125, 349, 40 },
		{ 125, 350, 40 },
		{ 125, 351, 40 },
		{ 125, 352, 40 },
		{ 125, 353, 40 },
		{ 125, 354, 40 },
		{ 125, 355, 40 },
		{ 125, 356, 40 },
		{ 125, 357, 40 },
		{ 125, 358, 40 },
		{ 125, 359, 40 },
		{ 125, 360, 40 },
		{ 125, 361, 40 },
		{ 125, 362, 40 },
		{ 125, 363, 40 },
		{ 125, 364, 40 },
		{ 125, 365, 40 },
		{ 125, 366, 40 },
		{ 125, 367, 40 },
		{ 125, 368, 40 },
		{ 125, 369, 40 },
		{ 125, 370, 40 },
		{ 125, 371, 40 },
		{ 125, 372, 40 },
		{ 126, 339, 40 },
		{ 126, 340, 40 },
		{ 126, 341, 40 },
		{ 126, 342, 40 },
		{ 126, 343, 40 },
		{ 126, 344, 40 },
		{ 126, 345, 40 },
		{ 126, 346, 40 },
		{ 126, 347, 40 },
		{ 126, 348, 40 },
		{ 126, 349, 40 },
		{ 126, 350, 40 },
		{ 126, 351, 40 },
		{ 126, 352, 40 },
		{ 126, 353, 40 },
		{ 126, 354, 40 },
		{ 126, 355, 40 },
		{ 126, 356, 40 },
		{ 126, 357, 40 },
		{ 126, 358, 40 },
		{ 126, 359, 40 },
		{ 126, 360, 40 },
		{ 126, 361, 40 },
		{ 126, 362, 40 },
		{ 126, 363, 40 },
		{ 126, 364, 40 },
		{ 126, 365, 40 },
		{ 126, 366, 40 },
		{ 126, 367, 40 },
		{ 126, 368, 40 },
		{ 126, 369, 40 },
		{ 126, 370, 40 },
		{ 126, 371, 40 },
		{ 127, 339, 40 },
		{ 127, 340, 40 },
		{ 127, 341, 40 },
		{ 127, 342, 40 },
		{ 127, 343, 40 },
		{ 127, 344, 40 },
		{ 127, 345, 40 },
		{ 127, 346, 40 },
		{ 127, 347, 40 },
		{ 127, 348, 40 },
		{ 127, 349, 40 },
		{ 127, 350, 40 },
		{ 127, 351, 40 },
		{ 127, 352, 40 },
		{ 127, 353, 40 },
		{ 127, 354, 40 },
		{ 127, 355, 40 },
		{ 127, 356, 40 },
		{ 127, 357, 40 },
		{ 127, 358, 40 },
		{ 127, 359, 40 },
		{ 127, 360, 40 },
		{ 127, 361, 40 },
		{ 127, 362, 40 },
		{ 127, 363, 40 },
		{ 127, 364, 40 },
		{ 127, 365, 40 },
		{ 127, 366, 40 },
		{ 127, 367, 40 },
		{ 127, 368, 40 },
		{ 127, 369, 40 },
		{ 127, 370, 40 },
		{ 128, 338, 40 },
		{ 128, 339, 40 },
		{ 128, 340, 40 },
		{ 128, 341, 40 },
		{ 128, 342, 40 },
		{ 128, 343, 40 },
		{ 128, 344, 40 },
		{ 128, 345, 40 },
		{ 128, 346, 40 },
		{ 128, 347, 40 },
		{ 128, 348, 40 },
		{ 128, 349, 40 },
		{ 128, 350, 40 },
		{ 128, 351, 40 },
		{ 128, 352, 40 },
		{ 128, 353, 40 },
		{ 128, 354, 40 },
		{ 128, 355, 40 },
		{ 128, 356, 40 },
		{ 128, 357, 40 },
		{ 128, 358, 40 },
		{ 128, 359, 40 },
		{ 128, 360, 40 },
		{ 128, 361, 40 },
		{ 128, 362, 40 },
		{ 128, 363, 40 },
		{ 128, 364, 40 },
		{ 128, 365, 40 },
		{ 128, 366, 40 },
		{ 128, 367, 40 },
		{ 128, 368, 40 },
		{ 128, 369, 40 },
		{ 129, 338, 40 },
		{ 129, 339, 40 },
		{ 129, 340, 40 },
		{ 129, 341, 40 },
		{ 129, 342, 40 },
		{ 129, 343, 40 },
		{ 129, 344, 40 },
		{ 129, 345, 40 },
		{ 129, 346, 40 },
		{ 129, 347, 40 },
		{ 129, 348, 40 },
		{ 129, 349, 40 },
		{ 129, 350, 40 },
		{ 129, 351, 40 },
		{ 129, 352, 40 },
		{ 129, 353, 40 },
		{ 129, 354, 40 },
		{ 129, 355, 40 },
		{ 129, 356, 40 },
		{ 129, 357, 40 },
		{ 129, 358, 40 },
		{ 129, 359, 40 },
		{ 129, 360, 40 },
		{ 129, 361, 40 },
		{ 129, 362, 40 },
		{ 129, 363, 40 },
		{ 129, 364, 40 },
		{ 129, 365, 40 },
		{ 129, 366, 40 },
		{ 129, 367, 40 },
		{ 129, 368, 40 },
		{ 130, 337, 40 },
		{ 130, 338, 40 },
		{ 130, 339, 40 },
		{ 130, 340, 40 },
		{ 130, 341, 40 },
		{ 130, 342, 40 },
		{ 130, 343, 40 },
		{ 130, 344, 40 },
		{ 130, 345, 40 },
		{ 130, 346, 40 },
		{ 130, 347, 40 },
		{ 130, 348, 40 },
		{ 130, 349, 40 },
		{ 130, 350, 40 },
		{ 130, 351, 40 },
		{ 130, 352, 40 },
		{ 130, 353, 40 },
		{ 130, 354, 40 },
		{ 130, 355, 40 },
		{ 130, 356, 40 },
		{ 130, 357, 40 },
		{ 130, 358, 40 },
		{ 130, 359, 40 },
		{ 130, 360, 40 },
		{ 130, 361, 40 },
		{ 130, 362, 40 },
		{ 130, 363, 40 },
		{ 130, 364, 40 },
		{ 130, 365, 40 },
		{ 130, 366, 40 },
		{ 130, 367, 40 },
		{ 130, 368, 40 },
		{ 131, 336, 40 },
		{ 131, 337, 40 },
		{ 131, 338, 40 },
		{ 131, 339, 40 },
		{ 131, 340, 40 },
		{ 131, 341, 40 },
		{ 131, 342, 40 },
		{ 131, 343, 40 },
		{ 131, 344, 40 },
		{ 131, 345, 40 },
		{ 131, 346, 40 },
		{ 131, 347, 40 },
		{ 131, 348, 40 },
		{ 131, 349, 40 },
		{ 131, 350, 40 },
		{ 131, 351, 40 },
		{ 131, 352, 40 },
		{ 131, 353, 40 },
		{ 131, 354, 40 },
		{ 131, 355, 40 },
		{ 131, 356, 40 },
		{ 131, 357, 40 },
		{ 131, 358, 40 },
		{ 131, 359, 40 },
		{ 131, 360, 40 },
		{ 131, 361, 40 },
		{ 131, 362, 40 },
		{ 131, 363, 40 },
		{ 131, 364, 40 },
		{ 131, 365, 40 },
		{ 131, 366, 40 },
		{ 131, 367, 40 },
		{ 132, 336, 40 },
		{ 132, 337, 40 },
		{ 132, 338, 40 },
		{ 132, 339, 40 },
		{ 132, 340, 40 },
		{ 132, 341, 40 },
		{ 132, 342, 40 },
		{ 132, 343, 40 },
		{ 132, 344, 40 },
		{ 132, 345, 40 },
		{ 132, 346, 40 },
		{ 132, 347, 40 },
		{ 132, 348, 40 },
		{ 132, 349, 40 },
		{ 132, 350, 40 },
		{ 132, 351, 40 },
		{ 132, 352, 40 },
		{ 132, 353, 40 },
		{ 132, 354, 40 },
		{ 132, 355, 40 },
		{ 132, 356, 40 },
		{ 132, 357, 40 },
		{ 132, 358, 40 },
		{ 132, 359, 40 },
		{ 132, 360, 40 },
		{ 132, 361, 40 },
		{ 132, 362, 40 },
		{ 132, 363, 40 },
		{ 132, 364, 40 },
		{ 132, 365, 40 },
		{ 132, 366, 40 },
		{ 133, 335, 40 },
		{ 133, 336, 40 },
		{ 133, 337, 40 },
		{ 133, 338, 40 },
		{ 133, 339, 40 },
		{ 133, 340, 40 },
		{ 133, 341, 40 },
		{ 133, 342, 40 },
		{ 133, 343, 40 },
		{ 133, 344, 40 },
		{ 133, 345, 40 },
		{ 133, 346, 40 },
		{ 133, 347, 40 },
		{ 133, 348, 40 },
		{ 133, 349, 40 },
		{ 133, 350, 40 },
		{ 133, 351, 40 },
		{ 133, 352, 40 },
		{ 133, 353, 40 },
		{ 133, 354, 40 },
		{ 133, 355, 40 },
		{ 133, 356, 40 },
		{ 133, 357, 40 },
		{ 133, 358, 40 },
		{ 133, 359, 40 },
		{ 133, 360, 40 },
		{ 133, 361, 40 },
		{ 133, 362, 40 },
		{ 133, 363, 40 },
		{ 133, 364, 40 },
		{ 133, 365, 40 },
		{ 134, 334, 40 },
		{ 134, 335, 40 },
		{ 134, 336, 40 },
		{ 134, 337, 40 },
		{ 134, 338, 40 },
		{ 134, 339, 40 },
		{ 134, 340, 40 },
		{ 134, 341, 40 },
		{ 134, 342, 40 },
		{ 134, 343, 40 },
		{ 134, 344, 40 },
		{ 134, 345, 40 },
		{ 134, 346, 40 },
		{ 134, 347, 40 },
		{ 134, 348, 40 },
		{ 134, 349, 40 },
		{ 134, 350, 40 },
		{ 134, 351, 40 },
		{ 134, 352, 40 },
		{ 134, 353, 40 },
		{ 134, 354, 40 },
		{ 134, 355, 40 },
		{ 134, 356, 40 },
		{ 134, 357, 40 },
		{ 134, 358, 40 },
		{ 134, 359, 40 },
		{ 134, 360, 40 },
		{ 134, 361, 40 },
		{ 134, 362, 40 },
		{ 134, 363, 40 },
		{ 134, 364, 40 },
		{ 135, 334, 40 },
		{ 135, 335, 40 },
		{ 135, 336, 40 },
		{ 135, 337, 40 },
		{ 135, 338, 40 },
		{ 135, 339, 40 },
		{ 135, 340, 40 },
		{ 135, 341, 40 },
		{ 135, 342, 40 },
		{ 135, 343, 40 },
		{ 135, 344, 40 },
		{ 135, 345, 40 },
		{ 135, 346, 40 },
		{ 135, 347, 40 },
		{ 135, 348, 40 },
		{ 135, 349, 40 },
		{ 135, 350, 40 },
		{ 135, 351, 40 },
		{ 135, 352, 40 },
		{ 135, 353, 40 },
		{ 135, 354, 40 },
		{ 135, 355, 40 },
		{ 135, 356, 40 },
		{ 135, 357, 40 },
		{ 135, 358, 40 },
		{ 135, 359, 40 },
		{ 135, 360, 40 },
		{ 135, 361, 40 },
		{ 135, 362, 40 },
		{ 135, 363, 40 },
		{ 136, 333, 40 },
		{ 136, 334, 40 },
		{ 136, 335, 40 },
		{ 136, 336, 40 },
		{ 136, 337, 40 },
		{ 136, 338, 40 },
		{ 136, 339, 40 },
		{ 136, 340, 40 },
		{ 136, 341, 40 },
		{ 136, 342, 40 },
		{ 136, 343, 40 },
		{ 136, 344, 40 },
		{ 136, 345, 40 },
		{ 136, 346, 40 },
		{ 136, 347, 40 },
		{ 136, 348, 40 },
		{ 136, 349, 40 },
		{ 136, 350, 40 },
		{ 136, 351, 40 },
		{ 136, 352, 40 },
		{ 136, 353, 40 },
		{ 136, 354, 40 },
		{ 136, 355, 40 },
		{ 136, 356, 40 },
		{ 136, 357, 40 },
		{ 136, 358, 40 },
		{ 136, 359, 40 },
		{ 136, 360, 40 },
		{ 136, 361, 40 },
		{ 136, 362, 40 },
		{ 137, 333, 40 },
		{ 137, 334, 40 },
		{ 137, 335, 40 },
		{ 137, 336, 40 },
		{ 137, 337, 40 },
		{ 137, 338, 40 },
		{ 137, 339, 40 },
		{ 137, 340, 40 },
		{ 137, 341, 40 },
		{ 137, 342, 40 },
		{ 137, 343, 40 },
		{ 137, 344, 40 },
		{ 137, 345, 40 },
		{ 137, 346, 40 },
		{ 137, 347, 40 },
		{ 137, 348, 40 },
		{ 137, 349, 40 },
		{ 137, 350, 40 },
		{ 137, 351, 40 },
		{ 137, 352, 40 },
		{ 137, 353, 40 },
		{ 137, 354, 40 },
		{ 137, 355, 40 },
		{ 137, 356, 40 },
		{ 137, 357, 40 },
		{ 137, 358, 40 },
		{ 137, 359, 40 },
		{ 137, 360, 40 },
		{ 137, 361, 40 },
		{ 138, 332, 40 },
		{ 138, 333, 40 },
		{ 138, 334, 40 },
		{ 138, 335, 40 },
		{ 138, 336, 40 },
		{ 138, 337, 40 },
		{ 138, 338, 40 },
		{ 138, 339, 40 },
		{ 138, 340, 40 },
		{ 138, 341, 40 },
		{ 138, 342, 40 },
		{ 138, 343, 40 },
		{ 138, 344, 40 },
		{ 138, 345, 40 },
		{ 138, 346, 40 },
		{ 138, 347, 40 },
		{ 138, 348, 40 },
		{ 138, 349, 40 },
		{ 138, 350, 40 },
		{ 138, 351, 40 },
		{ 138, 352, 40 },
		{ 138, 353, 40 },
		{ 138, 354, 40 },
		{ 138, 355, 40 },
		{ 138, 356, 40 },
		{ 138, 357, 40 },
		{ 138, 358, 40 },
		{ 138, 359, 40 },
		{ 138, 360, 40 },
		{ 138, 361, 40 },
		{ 139, 331, 40 },
		{ 139, 332, 40 },
		{ 139, 333, 40 },
		{ 139, 334, 40 },
		{ 139, 335, 40 },
		{ 139, 336, 40 },
		{ 139, 337, 40 },
		{ 139, 338, 40 },
		{ 139, 339, 40 },
		{ 139, 340, 40 },
		{ 139, 341, 40 },
		{ 139, 342, 40 },
		{ 139, 343, 40 },
		{ 139, 344, 40 },
		{ 139, 345, 40 },
		{ 139, 346, 40 },
		{ 139, 347, 40 },
		{ 139, 348, 40 },
		{ 139, 349, 40 },
		{ 139, 350, 40 },
		{ 139, 351, 40 },
		{ 139, 352, 40 },
		{ 139, 353, 40 },
		{ 139, 354, 40 },
		{ 139, 355, 40 },
		{ 139, 356, 40 },
		{ 139, 357, 40 },
		{ 139, 358, 40 },
		{ 139, 359, 40 },
		{ 139, 360, 40 },
		{ 140, 331, 40 },
		{ 140, 332, 40 },
		{ 140, 333, 40 },
		{ 140, 334, 40 },
		{ 140, 335, 40 },
		{ 140, 336, 40 },
		{ 140, 337, 40 },
		{ 140, 338, 40 },
		{ 140, 339, 40 },
		{ 140, 340, 40 },
		{ 140, 341, 40 },
		{ 140, 342, 40 },
		{ 140, 343, 40 },
		{ 140, 344, 40 },
		{ 140, 345, 40 },
		{ 140, 346, 40 },
		{ 140, 347, 40 },
		{ 140, 348, 40 },
		{ 140, 349, 40 },
		{ 140, 350, 40 },
		{ 140, 351, 40 },
		{ 140, 352, 40 },
		{ 140, 353, 40 },
		{ 140, 354, 40 },
		{ 140, 355, 40 },
		{ 140, 356, 40 },
		{ 140, 357, 40 },
		{ 140, 358, 40 },
		{ 140, 359, 40 },
		{ 141, 330, 40 },
		{ 141, 331, 40 },
		{ 141, 332, 40 },
		{ 141, 333, 40 },
		{ 141, 334, 40 },
		{ 141, 335, 40 },
		{ 141, 336, 40 },
		{ 141, 337, 40 },
		{ 141, 338, 40 },
		{ 141, 339, 40 },
		{ 141, 340, 40 },
		{ 141, 341, 40 },
		{ 141, 342, 40 },
		{ 141, 343, 40 },
		{ 141, 344, 40 },
		{ 141, 345, 40 },
		{ 141, 346, 40 },
		{ 141, 347, 40 },
		{ 141, 348, 40 },
		{ 141, 349, 40 },
		{ 141, 350, 40 },
		{ 141, 351, 40 },
		{ 141, 352, 40 },
		{ 141, 353, 40 },
		{ 141, 354, 40 },
		{ 141, 355, 40 },
		{ 141, 356, 40 },
		{ 141, 357, 40 },
		{ 141, 358, 40 },
		{ 142, 329, 40 },
		{ 142, 330, 40 },
		{ 142, 331, 40 },
		{ 142, 332, 40 },
		{ 142, 333, 40 },
		{ 142, 334, 40 },
		{ 142, 335, 40 },
		{ 142, 336, 40 },
		{ 142, 337, 40 },
		{ 142, 338, 40 },
		{ 142, 339, 40 },
		{ 142, 340, 40 },
		{ 142, 341, 40 },
		{ 142, 342, 40 },
		{ 142, 343, 40 },
		{ 142, 344, 40 },
		{ 142, 345, 40 },
		{ 142, 346, 40 },
		{ 142, 347, 40 },
		{ 142, 348, 40 },
		{ 142, 349, 40 },
		{ 142, 350, 40 },
		{ 142, 351, 40 },
		{ 142, 352, 40 },
		{ 142, 353, 40 },
		{ 142, 354, 40 },
		{ 142, 355, 40 },
		{ 142, 356, 40 },
		{ 142, 357, 40 },
		{ 143, 329, 40 },
		{ 143, 330, 40 },
		{ 143, 331, 40 },
		{ 143, 332, 40 },
		{ 143, 333, 40 },
		{ 143, 334, 40 },
		{ 143, 335, 40 },
		{ 143, 336, 40 },
		{ 143, 337, 40 },
		{ 143, 338, 40 },
		{ 143, 339, 40 },
		{ 143, 340, 40 },
		{ 143, 341, 40 },
		{ 143, 342, 40 },
		{ 143, 343, 40 },
		{ 143, 344, 40 },
		{ 143, 345, 40 },
		{ 143, 346, 40 },
		{ 143, 347, 40 },
		{ 143, 348, 40 },
		{ 143, 349, 40 },
		{ 143, 350, 40 },
		{ 143, 351, 40 },
		{ 143, 352, 40 },
		{ 143, 353, 40 },
		{ 143, 354, 40 },
		{ 143, 355, 40 },
		{ 143, 356, 40 },
		{ 144, 328, 40 },
		{ 144, 329, 40 },
		{ 144, 330, 40 },
		{ 144, 331, 40 },
		{ 144, 332, 40 },
		{ 144, 333, 40 },
		{ 144, 334, 40 },
		{ 144, 335, 40 },
		{ 144, 336, 40 },
		{ 144, 337, 40 },
		{ 144, 338, 40 },
		{ 144, 339, 40 },
		{ 144, 340, 40 },
		{ 144, 341, 40 },
		{ 144, 342, 40 },
		{ 144, 343, 40 },
		{ 144, 344, 40 },
		{ 144, 345, 40 },
		{ 144, 346, 40 },
		{ 144, 347, 40 },
		{ 144, 348, 40 },
		{ 144, 349, 40 },
		{ 144, 350, 40 },
		{ 144, 351, 40 },
		{ 144, 352, 40 },
		{ 144, 353, 40 },
		{ 144, 354, 40 },
		{ 144, 355, 40 },
		{ 145, 328, 40 },
		{ 145, 329, 40 },
		{ 145, 330, 40 },
		{ 145, 331, 40 },
		{ 145, 332, 40 },
		{ 145, 333, 40 },
		{ 145, 334, 40 },
		{ 145, 335, 40 },
		{ 145, 336, 40 },
		{ 145, 337, 40 },
		{ 145, 338, 40 },
		{ 145, 339, 40 },
		{ 145, 340, 40 },
		{ 145, 341, 40 },
		{ 145, 342, 40 },
		{ 145, 343, 40 },
		{ 145, 344, 40 },
		{ 145, 345, 40 },
		{ 145, 346, 40 },
		{ 145, 347, 40 },
		{ 145, 348, 40 },
		{ 145, 349, 40 },
		{ 145, 350, 40 },
		{ 145, 351, 40 },
		{ 145, 352, 40 },
		{ 145, 353, 40 },
		{ 145, 354, 40 },
		{ 146, 327, 40 },
		{ 146, 328, 40 },
		{ 146, 329, 40 },
		{ 146, 330, 40 },
		{ 146, 331, 40 },
		{ 146, 332, 40 },
		{ 146, 333, 40 },
		{ 146, 334, 40 },
		{ 146, 335, 40 },
		{ 146, 336, 40 },
		{ 146, 337, 40 },
		{ 146, 338, 40 },
		{ 146, 339, 40 },
		{ 146, 340, 40 },
		{ 146, 341, 40 },
		{ 146, 342, 40 },
		{ 146, 343, 40 },
		{ 146, 344, 40 },
		{ 146, 345, 40 },
		{ 146, 346, 40 },
		{ 146, 347, 40 },
		{ 146, 348, 40 },
		{ 146, 349, 40 },
		{ 146, 350, 40 },
		{ 146, 351, 40 },
		{ 146, 352, 40 },
		{ 146, 353, 40 },
		{ 146, 354, 40 },
		{ 147, 326, 40 },
		{ 147, 327, 40 },
		{ 147, 328, 40 },
		{ 147, 329, 40 },
		{ 147, 330, 40 },
		{ 147, 331, 40 },
		{ 147, 332, 40 },
		{ 147, 333, 40 },
		{ 147, 334, 40 },
		{ 147, 335, 40 },
		{ 147, 336, 40 },
		{ 147, 337, 40 },
		{ 147, 338, 40 },
		{ 147, 339, 40 },
		{ 147, 340, 40 },
		{ 147, 341, 40 },
		{ 147, 342, 40 },
		{ 147, 343, 40 },
		{ 147, 344, 40 },
		{ 147, 345, 40 },
		{ 147, 346, 40 },
		{ 147, 347, 40 },
		{ 147, 348, 40 },
		{ 147, 349, 40 },
		{ 147, 350, 40 },
		{ 147, 351, 40 },
		{ 147, 352, 40 },
		{ 147, 353, 40 },
		{ 148, 326, 40 },
		{ 148, 327, 40 },
		{ 148, 328, 40 },
		{ 148, 329, 40 },
		{ 148, 330, 40 },
		{ 148, 331, 40 },
		{ 148, 332, 40 },
		{ 148, 333, 40 },
		{ 148, 334, 40 },
		{ 148, 335, 40 },
		{ 148, 336, 40 },
		{ 148, 337, 40 },
		{ 148, 338, 40 },
		{ 148, 339, 40 },
		{ 148, 340, 40 },
		{ 148, 341, 40 },
		{ 148, 342, 40 },
		{ 148, 343, 40 },
		{ 148, 344, 40 },
		{ 148, 345, 40 },
		{ 148, 346, 40 },
		{ 148, 347, 40 },
		{ 148, 348, 40 },
		{ 148, 349, 40 },
		{ 148, 350, 40 },
		{ 148, 351, 40 },
		{ 148, 352, 40 },
		{ 149, 325, 40 },
		{ 149, 326, 40 },
		{ 149, 327, 40 },
		{ 149, 328, 40 },
		{ 149, 329, 40 },
		{ 149, 330, 40 },
		{ 149, 331, 40 },
		{ 149, 332, 40 },
		{ 149, 333, 40 },
		{ 149, 334, 40 },
		{ 149, 335, 40 },
		{ 149, 336, 40 },
		{ 149, 337, 40 },
		{ 149, 338, 40 },
		{ 149, 339, 40 },
		{ 149, 340, 40 },
		{ 149, 341, 40 },
		{ 149, 342, 40 },
		{ 149, 343, 40 },
		{ 149, 344, 40 },
		{ 149, 345, 40 },
		{ 149, 346, 40 },
		{ 149, 347, 40 },
		{ 149, 348, 40 },
		{ 149, 349, 40 },
		{ 149, 350, 40 },
		{ 149, 351, 40 },
		{ 150, 324, 40 },
		{ 150, 325, 40 },
		{ 150, 326, 40 },
		{ 150, 327, 40 },
		{ 150, 328, 40 },
		{ 150, 329, 40 },
		{ 150, 330, 40 },
		{ 150, 331, 40 },
		{ 150, 332, 40 },
		{ 150, 333, 40 },
		{ 150, 334, 40 },
		{ 150, 335, 40 },
		{ 150, 336, 40 },
		{ 150, 337, 40 },
		{ 150, 338, 40 },
		{ 150, 339, 40 },
		{ 150, 340, 40 },
		{ 150, 341, 40 },
		{ 150, 342, 40 },
		{ 150, 343, 40 },
		{ 150, 344, 40 },
		{ 150, 345, 40 },
		{ 150, 346, 40 },
		{ 150, 347, 40 },
		{ 150, 348, 40 },
		{ 150, 349, 40 },
		{ 150, 350, 40 },
		{ 151, 324, 40 },
		{ 151, 325, 40 },
		{ 151, 326, 40 },
		{ 151, 327, 40 },
		{ 151, 328, 40 },
		{ 151, 329, 40 },
		{ 151, 330, 40 },
		{ 151, 331, 40 },
		{ 151, 332, 40 },
		{ 151, 333, 40 },
		{ 151, 334, 40 },
		{ 151, 335, 40 },
		{ 151, 336, 40 },
		{ 151, 337, 40 },
		{ 151, 338, 40 },
		{ 151, 339, 40 },
		{ 151, 340, 40 },
		{ 151, 341, 40 },
		{ 151, 342, 40 },
		{ 151, 343, 40 },
		{ 151, 344, 40 },
		{ 151, 345, 40 },
		{ 151, 346, 40 },
		{ 151, 347, 40 },
		{ 151, 348, 40 },
		{ 151, 349, 40 },
		{ 152, 323, 40 },
		{ 152, 324, 40 },
		{ 152, 325, 40 },
		{ 152, 326, 40 },
		{ 152, 327, 40 },
		{ 152, 328, 40 },
		{ 152, 329, 40 },
		{ 152, 330, 40 },
		{ 152, 331, 40 },
		{ 152, 332, 40 },
		{ 152, 333, 40 },
		{ 152, 334, 40 },
		{ 152, 335, 40 },
		{ 152, 336, 40 },
		{ 152, 337, 40 },
		{ 152, 338, 40 },
		{ 152, 339, 40 },
		{ 152, 340, 40 },
		{ 152, 341, 40 },
		{ 152, 342, 40 },
		{ 152, 343, 40 },
		{ 152, 344, 40 },
		{ 152, 345, 40 },
		{ 152, 346, 40 },
		{ 152, 347, 40 },
		{ 152, 348, 40 },
		{ 153, 323, 40 },
		{ 153, 324, 40 },
		{ 153, 325, 40 },
		{ 153, 326, 40 },
		{ 153, 327, 40 },
		{ 153, 328, 40 },
		{ 153, 329, 40 },
		{ 153, 330, 40 },
		{ 153, 331, 40 },
		{ 153, 332, 40 },
		{ 153, 333, 40 },
		{ 153, 334, 40 },
		{ 153, 335, 40 },
		{ 153, 336, 40 },
		{ 153, 337, 40 },
		{ 153, 338, 40 },
		{ 153, 339, 40 },
		{ 153, 340, 40 },
		{ 153, 341, 40 },
		{ 153, 342, 40 },
		{ 153, 343, 40 },
		{ 153, 344, 40 },
		{ 153, 345, 40 },
		{ 153, 346, 40 },
		{ 153, 347, 40 },
		{ 154, 322, 40 },
		{ 154, 323, 40 },
		{ 154, 324, 40 },
		{ 154, 325, 40 },
		{ 154, 326, 40 },
		{ 154, 327, 40 },
		{ 154, 328, 40 },
		{ 154, 329, 40 },
		{ 154, 330, 40 },
		{ 154, 331, 40 },
		{ 154, 332, 40 },
		{ 154, 333, 40 },
		{ 154, 334, 40 },
		{ 154, 335, 40 },
		{ 154, 336, 40 },
		{ 154, 337, 40 },
		{ 154, 338, 40 },
		{ 154, 339, 40 },
		{ 154, 340, 40 },
		{ 154, 341, 40 },
		{ 154, 342, 40 },
		{ 154, 343, 40 },
		{ 154, 344, 40 },
		{ 154, 345, 40 },
		{ 154, 346, 40 },
		{ 154, 347, 40 },
		{ 155, 321, 40 },
		{ 155, 322, 40 },
		{ 155, 323, 40 },
		{ 155, 324, 40 },
		{ 155, 325, 40 },
		{ 155, 326, 40 },
		{ 155, 327, 40 },
		{ 155, 328, 40 },
		{ 155, 329, 40 },
		{ 155, 330, 40 },
		{ 155, 331, 40 },
		{ 155, 332, 40 },
		{ 155, 333, 40 },
		{ 155, 334, 40 },
		{ 155, 335, 40 },
		{ 155, 336, 40 },
		{ 155, 337, 40 },
		{ 155, 338, 40 },
		{ 155, 339, 40 },
		{ 155, 340, 40 },
		{ 155, 341, 40 },
		{ 155, 342, 40 },
		{ 155, 343, 40 },
		{ 155, 344, 40 },
		{ 155, 345, 40 },
		{ 155, 346, 40 },
		{ 156, 321, 40 },
		{ 156, 322, 40 },
		{ 156, 323, 40 },
		{ 156, 324, 40 },
		{ 156, 325, 40 },
		{ 156, 326, 40 },
		{ 156, 327, 40 },
		{ 156, 328, 40 },
		{ 156, 329, 40 },
		{ 156, 330, 40 },
		{ 156, 331, 40 },
		{ 156, 332, 40 },
		{ 156, 333, 40 },
		{ 156, 334, 40 },
		{ 156, 335, 40 },
		{ 156, 336, 40 },
		{ 156, 337, 40 },
		{ 156, 338, 40 },
		{ 156, 339, 40 },
		{ 156, 340, 40 },
		{ 156, 341, 40 },
		{ 156, 342, 40 },
		{ 156, 343, 40 },
		{ 156, 344, 40 },
		{ 156, 345, 40 },
		{ 157, 320, 40 },
		{ 157, 321, 40 },
		{ 157, 322, 40 },
		{ 157, 323, 40 },
		{ 157, 324, 40 },
		{ 157, 325, 40 },
		{ 157, 326, 40 },
		{ 157, 327, 40 },
		{ 157, 328, 40 },
		{ 157, 329, 40 },
		{ 157, 330, 40 },
		{ 157, 331, 40 },
		{ 157, 332, 40 },
		{ 157, 333, 40 },
		{ 157, 334, 40 },
		{ 157, 335, 40 },
		{ 157, 336, 40 },
		{ 157, 337, 40 },
		{ 157, 338, 40 },
		{ 157, 339, 40 },
		{ 157, 340, 40 },
		{ 157, 341, 40 },
		{ 157, 342, 40 },
		{ 157, 343, 40 },
		{ 157, 344, 40 },
		{ 158, 319, 40 },
		{ 158, 320, 40 },
		{ 158, 321, 40 },
		{ 158, 322, 40 },
		{ 158, 323, 40 },
		{ 158, 324, 40 },
		{ 158, 325, 40 },
		{ 158, 326, 40 },
		{ 158, 327, 40 },
		{ 158, 328, 40 },
		{ 158, 329, 40 },
		{ 158, 330, 40 },
		{ 158, 331, 40 },
		{ 158, 332, 40 },
		{ 158, 333, 40 },
		{ 158, 334, 40 },
		{ 158, 335, 40 },
		{ 158, 336, 40 },
		{ 158, 337, 40 },
		{ 158, 338, 40 },
		{ 158, 339, 40 },
		{ 158, 340, 40 },
		{ 158, 341, 40 },
		{ 158, 342, 40 },
		{ 158, 343, 40 },
		{ 159, 319, 40 },
		{ 159, 320, 40 },
		{ 159, 321, 40 },
		{ 159, 322, 40 },
		{ 159, 323, 40 },
		{ 159, 324, 40 },
		{ 159, 325, 40 },
		{ 159, 326, 40 },
		{ 159, 327, 40 },
		{ 159, 328, 40 },
		{ 159, 329, 40 },
		{ 159, 330, 40 },
		{ 159, 331, 40 },
		{ 159, 332, 40 },
		{ 159, 333, 40 },
		{ 159, 334, 40 },
		{ 159, 335, 40 },
		{ 159, 336, 40 },
		{ 159, 337, 40 },
		{ 159, 338, 40 },
		{ 159, 339, 40 },
		{ 159, 340, 40 },
		{ 159, 341, 40 },
		{ 159, 342, 40 },
		{ 160, 318, 40 },
		{ 160, 319, 40 },
		{ 160, 320, 40 },
		{ 160, 321, 40 },
		{ 160, 322, 40 },
		{ 160, 323, 40 },
		{ 160, 324, 40 },
		{ 160, 325, 40 },
		{ 160, 326, 40 },
		{ 160, 327, 40 },
		{ 160, 328, 40 },
		{ 160, 329, 40 },
		{ 160, 330, 40 },
		{ 160, 331, 40 },
		{ 160, 332, 40 },
		{ 160, 333, 40 },
		{ 160, 334, 40 },
		{ 160, 335, 40 },
		{ 160, 336, 40 },
		{ 160, 337, 40 },
		{ 160, 338, 40 },
		{ 160, 339, 40 },
		{ 160, 340, 40 },
		{ 160, 341, 40 },
		{ 161, 318, 40 },
		{ 161, 319, 40 },
		{ 161, 320, 40 },
		{ 161, 321, 40 },
		{ 161, 322, 40 },
		{ 161, 323, 40 },
		{ 161, 324, 40 },
		{ 161, 325, 40 },
		{ 161, 326, 40 },
		{ 161, 327, 40 },
		{ 161, 328, 40 },
		{ 161, 329, 40 },
		{ 161, 330, 40 },
		{ 161, 331, 40 },
		{ 161, 332, 40 },
		{ 161, 333, 40 },
		{ 161, 334, 40 },
		{ 161, 335, 40 },
		{ 161, 336, 40 },
		{ 161, 337, 40 },
		{ 161, 338, 40 },
		{ 161, 339, 40 },
		{ 161, 340, 40 },
		{ 162, 317, 40 },
		{ 162, 318, 40 },
		{ 162, 319, 40 },
		{ 162, 320, 40 },
		{ 162, 321, 40 },
		{ 162, 322, 40 },
		{ 162, 323, 40 },
		{ 162, 324, 40 },
		{ 162, 325, 40 },
		{ 162, 326, 40 },
		{ 162, 327, 40 },
		{ 162, 328, 40 },
		{ 162, 329, 40 },
		{ 162, 330, 40 },
		{ 162, 331, 40 },
		{ 162, 332, 40 },
		{ 162, 333, 40 },
		{ 162, 334, 40 },
		{ 162, 335, 40 },
		{ 162, 336, 40 },
		{ 162, 337, 40 },
		{ 162, 338, 40 },
		{ 162, 339, 40 },
		{ 162, 340, 40 },
		{ 163, 316, 40 },
		{ 163, 317, 40 },
		{ 163, 318, 40 },
		{ 163, 319, 40 },
		{ 163, 320, 40 },
		{ 163, 321, 40 },
		{ 163, 322, 40 },
		{ 163, 323, 40 },
		{ 163, 324, 40 },
		{ 163, 325, 40 },
		{ 163, 326, 40 },
		{ 163, 327, 40 },
		{ 163, 328, 40 },
		{ 163, 329, 40 },
		{ 163, 330, 40 },
		{ 163, 331, 40 },
		{ 163, 332, 40 },
		{ 163, 333, 40 },
		{ 163, 334, 40 },
		{ 163, 335, 40 },
		{ 163, 336, 40 },
		{ 163, 337, 40 },
		{ 163, 338, 40 },
		{ 163, 339, 40 },
		{ 164, 316, 40 },
		{ 164, 317, 40 },
		{ 164, 318, 40 },
		{ 164, 319, 40 },
		{ 164, 320, 40 },
		{ 164, 321, 40 },
		{ 164, 322, 40 },
		{ 164, 323, 40 },
		{ 164, 324, 40 },
		{ 164, 325, 40 },
		{ 164, 326, 40 },
		{ 164, 327, 40 },
		{ 164, 328, 40 },
		{ 164, 329, 40 },
		{ 164, 330, 40 },
		{ 164, 331, 40 },
		{ 164, 332, 40 },
		{ 164, 333, 40 },
		{ 164, 334, 40 },
		{ 164, 335, 40 },
		{ 164, 336, 40 },
		{ 164, 337, 40 },
		{ 164, 338, 40 },
		{ 165, 315, 40 },
		{ 165, 316, 40 },
		{ 165, 317, 40 },
		{ 165, 318, 40 },
		{ 165, 319, 40 },
		{ 165, 320, 40 },
		{ 165, 321, 40 },
		{ 165, 322, 40 },
		{ 165, 323, 40 },
		{ 165, 324, 40 },
		{ 165, 325, 40 },
		{ 165, 326, 40 },
		{ 165, 327, 40 },
		{ 165, 328, 40 },
		{ 165, 329, 40 },
		{ 165, 330, 40 },
		{ 165, 331, 40 },
		{ 165, 332, 40 },
		{ 165, 333, 40 },
		{ 165, 334, 40 },
		{ 165, 335, 40 },
		{ 165, 336, 40 },
		{ 165, 337, 40 },
		{ 166, 314, 40 },
		{ 166, 315, 40 },
		{ 166, 316, 40 },
		{ 166, 317, 40 },
		{ 166, 318, 40 },
		{ 166, 319, 40 },
		{ 166, 320, 40 },
		{ 166, 321, 40 },
		{ 166, 322, 40 },
		{ 166, 323, 40 },
		{ 166, 324, 40 },
		{ 166, 325, 40 },
		{ 166, 326, 40 },
		{ 166, 327, 40 },
		{ 166, 328, 40 },
		{ 166, 329, 40 },
		{ 166, 330, 40 },
		{ 166, 331, 40 },
		{ 166, 332, 40 },
		{ 166, 333, 40 },
		{ 166, 334, 40 },
		{ 166, 335, 40 },
		{ 166, 336, 40 },
		{ 167, 314, 40 },
		{ 167, 315, 40 },
		{ 167, 316, 40 },
		{ 167, 317, 40 },
		{ 167, 318, 40 },
		{ 167, 319, 40 },
		{ 167, 320, 40 },
		{ 167, 321, 40 },
		{ 167, 322, 40 },
		{ 167, 323, 40 },
		{ 167, 324, 40 },
		{ 167, 325, 40 },
		{ 167, 326, 40 },
		{ 167, 327, 40 },
		{ 167, 328, 40 },
		{ 167, 329, 40 },
		{ 167, 330, 40 },
		{ 167, 331, 40 },
		{ 167, 332, 40 },
		{ 167, 333, 40 },
		{ 167, 334, 40 },
		{ 167, 335, 40 },
		{ 168, 313, 40 },
		{ 168, 314, 40 },
		{ 168, 315, 40 },
		{ 168, 316, 40 },
		{ 168, 317, 40 },
		{ 168, 318, 40 },
		{ 168, 319, 40 },
		{ 168, 320, 40 },
		{ 168, 321, 40 },
		{ 168, 322, 40 },
		{ 168, 323, 40 },
		{ 168, 324, 40 },
		{ 168, 325, 40 },
		{ 168, 326, 40 },
		{ 168, 327, 40 },
		{ 168, 328, 40 },
		{ 168, 329, 40 },
		{ 168, 330, 40 },
		{ 168, 331, 40 },
		{ 168, 332, 40 },
		{ 168, 333, 40 },
		{ 168, 334, 40 },
		{ 169, 313, 40 },
		{ 169, 314, 40 },
		{ 169, 315, 40 },
		{ 169, 316, 40 },
		{ 169, 317, 40 },
		{ 169, 318, 40 },
		{ 169, 319, 40 },
		{ 169, 320, 40 },
		{ 169, 321, 40 },
		{ 169, 322, 40 },
		{ 169, 323, 40 },
		{ 169, 324, 40 },
		{ 169, 325, 40 },
		{ 169, 326, 40 },
		{ 169, 327, 40 },
		{ 169, 328, 40 },
		{ 169, 329, 40 },
		{ 169, 330, 40 },
		{ 169, 331, 40 },
		{ 169, 332, 40 },
		{ 169, 333, 40 },
		{ 170, 312, 40 },
		{ 170, 313, 40 },
		{ 170, 314, 40 },
		{ 170, 315, 40 },
		{ 170, 316, 40 },
		{ 170, 317, 40 },
		{ 170, 318, 40 },
		{ 170, 319, 40 },
		{ 170, 320, 40 },
		{ 170, 321, 40 },
		{ 170, 322, 40 },
		{ 170, 323, 40 },
		{ 170, 324, 40 },
		{ 170, 325, 40 },
		{ 170, 326, 40 },
		{ 170, 327, 40 },
		{ 170, 328, 40 },
		{ 170, 329, 40 },
		{ 170, 330, 40 },
		{ 170, 331, 40 },
		{ 170, 332, 40 },
		{ 170, 333, 40 },
		{ 171, 311, 40 },
		{ 171, 312, 40 },
		{ 171, 313, 40 },
		{ 171, 314, 40 },
		{ 171, 315, 40 },
		{ 171, 316, 40 },
		{ 171, 317, 40 },
		{ 171, 318, 40 },
		{ 171, 319, 40 },
		{ 171, 320, 40 },
		{ 171, 321, 40 },
		{ 171, 322, 40 },
		{ 171, 323, 40 },
		{ 171, 324, 40 },
		{ 171, 325, 40 },
		{ 171, 326, 40 },
		{ 171, 327, 40 },
		{ 171, 328, 40 },
		{ 171, 329, 40 },
		{ 171, 330, 40 },
		{ 171, 331, 40 },
		{ 171, 332, 40 },
		{ 172, 311, 40 },
		{ 172, 312, 40 },
		{ 172, 313, 40 },
		{ 172, 314, 40 },
		{ 172, 315, 40 },
		{ 172, 316, 40 },
		{ 172, 317, 40 },
		{ 172, 318, 40 },
		{ 172, 319, 40 },
		{ 172, 320, 40 },
		{ 172, 321, 40 },
		{ 172, 322, 40 },
		{ 172, 323, 40 },
		{ 172, 324, 40 },
		{ 172, 325, 40 },
		{ 172, 326, 40 },
		{ 172, 327, 40 },
		{ 172, 328, 40 },
		{ 172, 329, 40 },
		{ 172, 330, 40 },
		{ 172, 331, 40 },
		{ 173, 310, 40 },
		{ 173, 311, 40 },
		{ 173, 312, 40 },
		{ 173, 313, 40 },
		{ 173, 314, 40 },
		{ 173, 315, 40 },
		{ 173, 316, 40 },
		{ 173, 317, 40 },
		{ 173, 318, 40 },
		{ 173, 319, 40 },
		{ 173, 320, 40 },
		{ 173, 321, 40 },
		{ 173, 322, 40 },
		{ 173, 323, 40 },
		{ 173, 324, 40 },
		{ 173, 325, 40 },
		{ 173, 326, 40 },
		{ 173, 327, 40 },
		{ 173, 328, 40 },
		{ 173, 329, 40 },
		{ 173, 330, 40 },
		{ 174, 309, 40 },
		{ 174, 310, 40 },
		{ 174, 311, 40 },
		{ 174, 312, 40 },
		{ 174, 313, 40 },
		{ 174, 314, 40 },
		{ 174, 315, 40 },
		{ 174, 316, 40 },
		{ 174, 317, 40 },
		{ 174, 318, 40 },
		{ 174, 319, 40 },
		{ 174, 320, 40 },
		{ 174, 321, 40 },
		{ 174, 322, 40 },
		{ 174, 323, 40 },
		{ 174, 324, 40 },
		{ 174, 325, 40 },
		{ 174, 326, 40 },
		{ 174, 327, 40 },
		{ 174, 328, 40 },
		{ 174, 329, 40 },
		{ 175, 309, 40 },
		{ 175, 310, 40 },
		{ 175, 311, 40 },
		{ 175, 312, 40 },
		{ 175, 313, 40 },
		{ 175, 314, 40 },
		{ 175, 315, 40 },
		{ 175, 316, 40 },
		{ 175, 317, 40 },
		{ 175, 318, 40 },
		{ 175, 319, 40 },
		{ 175, 320, 40 },
		{ 175, 321, 40 },
		{ 175, 322, 40 },
		{ 175, 323, 40 },
		{ 175, 324, 40 },
		{ 175, 325, 40 },
		{ 175, 326, 40 },
		{ 175, 327, 40 },
		{ 175, 328, 40 },
		{ 176, 308, 40 },
		{ 176, 309, 40 },
		{ 176, 310, 40 },
		{ 176, 311, 40 },
		{ 176, 312, 40 },
		{ 176, 313, 40 },
		{ 176, 314, 40 },
		{ 176, 315, 40 },
		{ 176, 316, 40 },
		{ 176, 317, 40 },
		{ 176, 318, 40 },
		{ 176, 319, 40 },
		{ 176, 320, 40 },
		{ 176, 321, 40 },
		{ 176, 322, 40 },
		{ 176, 323, 40 },
		{ 176, 324, 40 },
		{ 176, 325, 40 },
		{ 176, 326, 40 },
		{ 176, 327, 40 },
		{ 177, 308, 40 },
		{ 177, 309, 40 },
		{ 177, 310, 40 },
		{ 177, 311, 40 },
		{ 177, 312, 40 },
		{ 177, 313, 40 },
		{ 177, 314, 40 },
		{ 177, 315, 40 },
		{ 177, 316, 40 },
		{ 177, 317, 40 },
		{ 177, 318, 40 },
		{ 177, 319, 40 },
		{ 177, 320, 40 },
		{ 177, 321, 40 },
		{ 177, 322, 40 },
		{ 177, 323, 40 },
		{ 177, 324, 40 },
		{ 177, 325, 40 },
		{ 177, 326, 40 },
		{ 178, 307, 40 },
		{ 178, 308, 40 },
		{ 178, 309, 40 },
		{ 178, 310, 40 },
		{ 178, 311, 40 },
		{ 178, 312, 40 },
		{ 178, 313, 40 },
		{ 178, 314, 40 },
		{ 178, 315, 40 },
		{ 178, 316, 40 },
		{ 178, 317, 40 },
		{ 178, 318, 40 },
		{ 178, 319, 40 },
		{ 178, 320, 40 },
		{ 178, 321, 40 },
		{ 178, 322, 40 },
		{ 178, 323, 40 },
		{ 178, 324, 40 },
		{ 178, 325, 40 },
		{ 178, 326, 40 },
		{ 179, 306, 40 },
		{ 179, 307, 40 },
		{ 179, 308, 40 },
		{ 179, 309, 40 },
		{ 179, 310, 40 },
		{ 179, 311, 40 },
		{ 179, 312, 40 },
		{ 179, 313, 40 },
		{ 179, 314, 40 },
		{ 179, 315, 40 },
		{ 179, 316, 40 },
		{ 179, 317, 40 },
		{ 179, 318, 40 },
		{ 179, 319, 40 },
		{ 179, 320, 40 },
		{ 179, 321, 40 },
		{ 179, 322, 40 },
		{ 179, 323, 40 },
		{ 179, 324, 40 },
		{ 179, 325, 40 },
		{ 180, 306, 40 },
		{ 180, 307, 40 },
		{ 180, 308, 40 },
		{ 180, 309, 40 },
		{ 180, 310, 40 },
		{ 180, 311, 40 },
		{ 180, 312, 40 },
		{ 180, 313, 40 },
		{ 180, 314, 40 },
		{ 180, 315, 40 },
		{ 180, 316, 40 },
		{ 180, 317, 40 },
		{ 180, 318, 40 },
		{ 180, 319, 40 },
		{ 180, 320, 40 },
		{ 180, 321, 40 },
		{ 180, 322, 40 },
		{ 180, 323, 40 },
		{ 180, 324, 40 },
		{ 181, 305, 40 },
		{ 181, 306, 40 },
		{ 181, 307, 40 },
		{ 181, 308, 40 },
		{ 181, 309, 40 },
		{ 181, 310, 40 },
		{ 181, 311, 40 },
		{ 181, 312, 40 },
		{ 181, 313, 40 },
		{ 181, 314, 40 },
		{ 181, 315, 40 },
		{ 181, 316, 40 },
		{ 181, 317, 40 },
		{ 181, 318, 40 },
		{ 181, 319, 40 },
		{ 181, 320, 40 },
		{ 181, 321, 40 },
		{ 181, 322, 40 },
		{ 181, 323, 40 },
		{ 182, 304, 40 },
		{ 182, 305, 40 },
		{ 182, 306, 40 },
		{ 182, 307, 40 },
		{ 182, 308, 40 },
		{ 182, 309, 40 },
		{ 182, 310, 40 },
		{ 182, 311, 40 },
		{ 182, 312, 40 },
		{ 182, 313, 40 },
		{ 182, 314, 40 },
		{ 182, 315, 40 },
		{ 182, 316, 40 },
		{ 182, 317, 40 },
		{ 182, 318, 40 },
		{ 182, 319, 40 },
		{ 182, 320, 40 },
		{ 182, 321, 40 },
		{ 182, 322, 40 },
		{ 183, 304, 40 },
		{ 183, 305, 40 },
		{ 183, 306, 40 },
		{ 183, 307, 40 },
		{ 183, 308, 40 },
		{ 183, 309, 40 },
		{ 183, 310, 40 },
		{ 183, 311, 40 },
		{ 183, 312, 40 },
		{ 183, 313, 40 },
		{ 183, 314, 40 },
		{ 183, 315, 40 },
		{ 183, 316, 40 },
		{ 183, 317, 40 },
		{ 183, 318, 40 },
		{ 183, 319, 40 },
		{ 183, 320, 40 },
		{ 183, 321, 40 },
		{ 184, 303, 40 },
		{ 184, 304, 40 },
		{ 184, 305, 40 },
		{ 184, 306, 40 },
		{ 184, 307, 40 },
		{ 184, 308, 40 },
		{ 184, 309, 40 },
		{ 184, 310, 40 },
		{ 184, 311, 40 },
		{ 184, 312, 40 },
		{ 184, 313, 40 },
		{ 184, 314, 40 },
		{ 184, 315, 40 },
		{ 184, 316, 40 },
		{ 184, 317, 40 },
		{ 184, 318, 40 },
		{ 184, 319, 40 },
		{ 184, 320, 40 },
		{ 185, 303, 40 },
		{ 185, 304, 40 },
		{ 185, 305, 40 },
		{ 185, 306, 40 },
		{ 185, 307, 40 },
		{ 185, 308, 40 },
		{ 185, 309, 40 },
		{ 185, 310, 40 },
		{ 185, 311, 40 },
		{ 185, 312, 40 },
		{ 185, 313, 40 },
		{ 185, 314, 40 },
		{ 185, 315, 40 },
		{ 185, 316, 40 },
		{ 185, 317, 40 },
		{ 185, 318, 40 },
		{ 185, 319, 40 },
		{ 186, 302, 40 },
		{ 186, 303, 40 },
		{ 186, 304, 40 },
		{ 186, 305, 40 },
		{ 186, 306, 40 },
		{ 186, 307, 40 },
		{ 186, 308, 40 },
		{ 186, 309, 40 },
		{ 186, 310, 40 },
		{ 186, 311, 40 },
		{ 186, 312, 40 },
		{ 186, 313, 40 },
		{ 186, 314, 40 },
		{ 186, 315, 40 },
		{ 186, 316, 40 },
		{ 186, 317, 40 },
		{ 186, 318, 40 },
		{ 187, 301, 40 },
		{ 187, 302, 40 },
		{ 187, 303, 40 },
		{ 187, 304, 40 },
		{ 187, 305, 40 },
		{ 187, 306, 40 },
		{ 187, 307, 40 },
		{ 187, 308, 40 },
		{ 187, 309, 40 },
		{ 187, 310, 40 },
		{ 187, 311, 40 },
		{ 187, 312, 40 },
		{ 187, 313, 40 },
		{ 187, 314, 40 },
		{ 187, 315, 40 },
		{ 187, 316, 40 },
		{ 187, 317, 40 },
		{ 187, 318, 40 },
		{ 188, 301, 40 },
		{ 188, 302, 40 },
		{ 188, 303, 40 },
		{ 188, 304, 40 },
		{ 188, 305, 40 },
		{ 188, 306, 40 },
		{ 188, 307, 40 },
		{ 188, 308, 40 },
		{ 188, 309, 40 },
		{ 188, 310, 40 },
		{ 188, 311, 40 },
		{ 188, 312, 40 },
		{ 188, 313, 40 },
		{ 188, 314, 40 },
		{ 188, 315, 40 },
		{ 188, 316, 40 },
		{ 188, 317, 40 },
		{ 189, 300, 40 },
		{ 189, 301, 40 },
		{ 189, 302, 40 },
		{ 189, 303, 40 },
		{ 189, 304, 40 },
		{ 189, 305, 40 },
		{ 189, 306, 40 },
		{ 189, 307, 40 },
		{ 189, 308, 40 },
		{ 189, 309, 40 },
		{ 189, 310, 40 },
		{ 189, 311, 40 },
		{ 189, 312, 40 },
		{ 189, 313, 40 },
		{ 189, 314, 40 },
		{ 189, 315, 40 },
		{ 189, 316, 40 },
		{ 190, 300, 40 },
		{ 190, 301, 40 },
		{ 190, 302, 40 },
		{ 190, 303, 40 },
		{ 190, 304, 40 },
		{ 190, 305, 40 },
		{ 190, 306, 40 },
		{ 190, 307, 40 },
		{ 190, 308, 40 },
		{ 190, 309, 40 },
		{ 190, 310, 40 },
		{ 190, 311, 40 },
		{ 190, 312, 40 },
		{ 190, 313, 40 },
		{ 190, 314, 40 },
		{ 191, 301, 40 },
		{ 191, 302, 40 },
		{ 191, 303, 40 },
		{ 191, 304, 40 },
		{ 191, 305, 40 },
		{ 191, 306, 40 },
		{ 191, 307, 40 },
		{ 191, 308, 40 },
		{ 191, 309, 40 },
		{ 191, 310, 40 },
		{ 191, 311, 40 },
		{ 191, 312, 40 },
		{ 323, 202, 40 },
		{ 323, 203, 40 },
		{ 323, 204, 40 },
		{ 323, 205, 40 },
		{ 323, 206, 40 },
		{ 323, 207, 40 },
		{ 323, 208, 40 },
		{ 323, 209, 40 },
		{ 323, 210, 40 },
		{ 323, 211, 40 },
		{ 323, 212, 40 },
		{ 323, 213, 40 },
		{ 324, 200, 40 },
		{ 324, 201, 40 },
		{ 324, 202, 40 },
		{ 324, 203, 40 },
		{ 324, 204, 40 },
		{ 324, 205, 40 },
		{ 324, 206, 40 },
		{ 324, 207, 40 },
		{ 324, 208, 40 },
		{ 324, 209, 40 },
		{ 324, 210, 40 },
		{ 324, 211, 40 },
		{ 324, 212, 40 },
		{ 324, 213, 40 },
		{ 324, 214, 40 },
		{ 325, 198, 40 },
		{ 325, 199, 40 },
		{ 325, 200, 40 },
		{ 325, 201, 40 },
		{ 325, 202, 40 },
		{ 325, 203, 40 },
		{ 325, 204, 40 },
		{ 325, 205, 40 },
		{ 325, 206, 40 },
		{ 325, 207, 40 },
		{ 325, 208, 40 },
		{ 325, 209, 40 },
		{ 325, 210, 40 },
		{ 325, 211, 40 },
		{ 325, 212, 40 },
		{ 325, 213, 40 },
		{ 325, 214, 40 },
		{ 326, 197, 40 },
		{ 326, 198, 40 },
		{ 326, 199, 40 },
		{ 326, 200, 40 },
		{ 326, 201, 40 },
		{ 326, 202, 40 },
		{ 326, 203, 40 },
		{ 326, 204, 40 },
		{ 326, 205, 40 },
		{ 326, 206, 40 },
		{ 326, 207, 40 },
		{ 326, 208, 40 },
		{ 326, 209, 40 },
		{ 326, 210, 40 },
		{ 326, 211, 40 },
		{ 326, 212, 40 },
		{ 326, 213, 40 },
		{ 327, 196, 40 },
		{ 327, 197, 40 },
		{ 327, 198, 40 },
		{ 327, 199, 40 },
		{ 327, 200, 40 },
		{ 327, 201, 40 },
		{ 327, 202, 40 },
		{ 327, 203, 40 },
		{ 327, 204, 40 },
		{ 327, 205, 40 },
		{ 327, 206, 40 },
		{ 327, 207, 40 },
		{ 327, 208, 40 },
		{ 327, 209, 40 },
		{ 327, 210, 40 },
		{ 327, 211, 40 },
		{ 327, 212, 40 },
		{ 327, 213, 40 },
		{ 328, 196, 40 },
		{ 328, 197, 40 },
		{ 328, 198, 40 },
		{ 328, 199, 40 },
		{ 328, 200, 40 },
		{ 328, 201, 40 },
		{ 328, 202, 40 },
		{ 328, 203, 40 },
		{ 328, 204, 40 },
		{ 328, 205, 40 },
		{ 328, 206, 40 },
		{ 328, 207, 40 },
		{ 328, 208, 40 },
		{ 328, 209, 40 },
		{ 328, 210, 40 },
		{ 328, 211, 40 },
		{ 328, 212, 40 },
		{ 329, 195, 40 },
		{ 329, 196, 40 },
		{ 329, 197, 40 },
		{ 329, 198, 40 },
		{ 329, 199, 40 },
		{ 329, 200, 40 },
		{ 329, 201, 40 },
		{ 329, 202, 40 },
		{ 329, 203, 40 },
		{ 329, 204, 40 },
		{ 329, 205, 40 },
		{ 329, 206, 40 },
		{ 329, 207, 40 },
		{ 329, 208, 40 },
		{ 329, 209, 40 },
		{ 329, 210, 40 },
		{ 329, 211, 40 },
		{ 330, 194, 40 },
		{ 330, 195, 40 },
		{ 330, 196, 40 },
		{ 330, 197, 40 },
		{ 330, 198, 40 },
		{ 330, 199, 40 },
		{ 330, 200, 40 },
		{ 330, 201, 40 },
		{ 330, 202, 40 },
		{ 330, 203, 40 },
		{ 330, 204, 40 },
		{ 330, 205, 40 },
		{ 330, 206, 40 },
		{ 330, 207, 40 },
		{ 330, 208, 40 },
		{ 330, 209, 40 },
		{ 330, 210, 40 },
		{ 330, 211, 40 },
		{ 331, 193, 40 },
		{ 331, 194, 40 },
		{ 331, 195, 40 },
		{ 331, 196, 40 },
		{ 331, 197, 40 },
		{ 331, 198, 40 },
		{ 331, 199, 40 },
		{ 331, 200, 40 },
		{ 331, 201, 40 },
		{ 331, 202, 40 },
		{ 331, 203, 40 },
		{ 331, 204, 40 },
		{ 331, 205, 40 },
		{ 331, 206, 40 },
		{ 331, 207, 40 },
		{ 331, 208, 40 },
		{ 331, 209, 40 },
		{ 331, 210, 40 },
		{ 332, 192, 40 },
		{ 332, 193, 40 },
		{ 332, 194, 40 },
		{ 332, 195, 40 },
		{ 332, 196, 40 },
		{ 332, 197, 40 },
		{ 332, 198, 40 },
		{ 332, 199, 40 },
		{ 332, 200, 40 },
		{ 332, 201, 40 },
		{ 332, 202, 40 },
		{ 332, 203, 40 },
		{ 332, 204, 40 },
		{ 332, 205, 40 },
		{ 332, 206, 40 },
		{ 332, 207, 40 },
		{ 332, 208, 40 },
		{ 332, 209, 40 },
		{ 332, 210, 40 },
		{ 333, 191, 40 },
		{ 333, 192, 40 },
		{ 333, 193, 40 },
		{ 333, 194, 40 },
		{ 333, 195, 40 },
		{ 333, 196, 40 },
		{ 333, 197, 40 },
		{ 333, 198, 40 },
		{ 333, 199, 40 },
		{ 333, 200, 40 },
		{ 333, 201, 40 },
		{ 333, 202, 40 },
		{ 333, 203, 40 },
		{ 333, 204, 40 },
		{ 333, 205, 40 },
		{ 333, 206, 40 },
		{ 333, 207, 40 },
		{ 333, 208, 40 },
		{ 333, 209, 40 },
		{ 334, 190, 40 },
		{ 334, 191, 40 },
		{ 334, 192, 40 },
		{ 334, 193, 40 },
		{ 334, 194, 40 },
		{ 334, 195, 40 },
		{ 334, 196, 40 },
		{ 334, 197, 40 },
		{ 334, 198, 40 },
		{ 334, 199, 40 },
		{ 334, 200, 40 },
		{ 334, 201, 40 },
		{ 334, 202, 40 },
		{ 334, 203, 40 },
		{ 334, 204, 40 },
		{ 334, 205, 40 },
		{ 334, 206, 40 },
		{ 334, 207, 40 },
		{ 334, 208, 40 },
		{ 335, 189, 40 },
		{ 335, 190, 40 },
		{ 335, 191, 40 },
		{ 335, 192, 40 },
		{ 335, 193, 40 },
		{ 335, 194, 40 },
		{ 335, 195, 40 },
		{ 335, 196, 40 },
		{ 335, 197, 40 },
		{ 335, 198, 40 },
		{ 335, 199, 40 },
		{ 335, 200, 40 },
		{ 335, 201, 40 },
		{ 335, 202, 40 },
		{ 335, 203, 40 },
		{ 335, 204, 40 },
		{ 335, 205, 40 },
		{ 335, 206, 40 },
		{ 335, 207, 40 },
		{ 335, 208, 40 },
		{ 336, 188, 40 },
		{ 336, 189, 40 },
		{ 336, 190, 40 },
		{ 336, 191, 40 },
		{ 336, 192, 40 },
		{ 336, 193, 40 },
		{ 336, 194, 40 },
		{ 336, 195, 40 },
		{ 336, 196, 40 },
		{ 336, 197, 40 },
		{ 336, 198, 40 },
		{ 336, 199, 40 },
		{ 336, 200, 40 },
		{ 336, 201, 40 },
		{ 336, 202, 40 },
		{ 336, 203, 40 },
		{ 336, 204, 40 },
		{ 336, 205, 40 },
		{ 336, 206, 40 },
		{ 336, 207, 40 },
		{ 337, 188, 40 },
		{ 337, 189, 40 },
		{ 337, 190, 40 },
		{ 337, 191, 40 },
		{ 337, 192, 40 },
		{ 337, 193, 40 },
		{ 337, 194, 40 },
		{ 337, 195, 40 },
		{ 337, 196, 40 },
		{ 337, 197, 40 },
		{ 337, 198, 40 },
		{ 337, 199, 40 },
		{ 337, 200, 40 },
		{ 337, 201, 40 },
		{ 337, 202, 40 },
		{ 337, 203, 40 },
		{ 337, 204, 40 },
		{ 337, 205, 40 },
		{ 337, 206, 40 },
		{ 338, 187, 40 },
		{ 338, 188, 40 },
		{ 338, 189, 40 },
		{ 338, 190, 40 },
		{ 338, 191, 40 },
		{ 338, 192, 40 },
		{ 338, 193, 40 },
		{ 338, 194, 40 },
		{ 338, 195, 40 },
		{ 338, 196, 40 },
		{ 338, 197, 40 },
		{ 338, 198, 40 },
		{ 338, 199, 40 },
		{ 338, 200, 40 },
		{ 338, 201, 40 },
		{ 338, 202, 40 },
		{ 338, 203, 40 },
		{ 338, 204, 40 },
		{ 338, 205, 40 },
		{ 338, 206, 40 },
		{ 339, 186, 40 },
		{ 339, 187, 40 },
		{ 339, 188, 40 },
		{ 339, 189, 40 },
		{ 339, 190, 40 },
		{ 339, 191, 40 },
		{ 339, 192, 40 },
		{ 339, 193, 40 },
		{ 339, 194, 40 },
		{ 339, 195, 40 },
		{ 339, 196, 40 },
		{ 339, 197, 40 },
		{ 339, 198, 40 },
		{ 339, 199, 40 },
		{ 339, 200, 40 },
		{ 339, 201, 40 },
		{ 339, 202, 40 },
		{ 339, 203, 40 },
		{ 339, 204, 40 },
		{ 339, 205, 40 },
		{ 340, 185, 40 },
		{ 340, 186, 40 },
		{ 340, 187, 40 },
		{ 340, 188, 40 },
		{ 340, 189, 40 },
		{ 340, 190, 40 },
		{ 340, 191, 40 },
		{ 340, 192, 40 },
		{ 340, 193, 40 },
		{ 340, 194, 40 },
		{ 340, 195, 40 },
		{ 340, 196, 40 },
		{ 340, 197, 40 },
		{ 340, 198, 40 },
		{ 340, 199, 40 },
		{ 340, 200, 40 },
		{ 340, 201, 40 },
		{ 340, 202, 40 },
		{ 340, 203, 40 },
		{ 340, 204, 40 },
		{ 340, 205, 40 },
		{ 341, 184, 40 },
		{ 341, 185, 40 },
		{ 341, 186, 40 },
		{ 341, 187, 40 },
		{ 341, 188, 40 },
		{ 341, 189, 40 },
		{ 341, 190, 40 },
		{ 341, 191, 40 },
		{ 341, 192, 40 },
		{ 341, 193, 40 },
		{ 341, 194, 40 },
		{ 341, 195, 40 },
		{ 341, 196, 40 },
		{ 341, 197, 40 },
		{ 341, 198, 40 },
		{ 341, 199, 40 },
		{ 341, 200, 40 },
		{ 341, 201, 40 },
		{ 341, 202, 40 },
		{ 341, 203, 40 },
		{ 341, 204, 40 },
		{ 342, 183, 40 },
		{ 342, 184, 40 },
		{ 342, 185, 40 },
		{ 342, 186, 40 },
		{ 342, 187, 40 },
		{ 342, 188, 40 },
		{ 342, 189, 40 },
		{ 342, 190, 40 },
		{ 342, 191, 40 },
		{ 342, 192, 40 },
		{ 342, 193, 40 },
		{ 342, 194, 40 },
		{ 342, 195, 40 },
		{ 342, 196, 40 },
		{ 342, 197, 40 },
		{ 342, 198, 40 },
		{ 342, 199, 40 },
		{ 342, 200, 40 },
		{ 342, 201, 40 },
		{ 342, 202, 40 },
		{ 342, 203, 40 },
		{ 343, 182, 40 },
		{ 343, 183, 40 },
		{ 343, 184, 40 },
		{ 343, 185, 40 },
		{ 343, 186, 40 },
		{ 343, 187, 40 },
		{ 343, 188, 40 },
		{ 343, 189, 40 },
		{ 343, 190, 40 },
		{ 343, 191, 40 },
		{ 343, 192, 40 },
		{ 343, 193, 40 },
		{ 343, 194, 40 },
		{ 343, 195, 40 },
		{ 343, 196, 40 },
		{ 343, 197, 40 },
		{ 343, 198, 40 },
		{ 343, 199, 40 },
		{ 343, 200, 40 },
		{ 343, 201, 40 },
		{ 343, 202, 40 },
		{ 343, 203, 40 },
		{ 344, 181, 40 },
		{ 344, 182, 40 },
		{ 344, 183, 40 },
		{ 344, 184, 40 },
		{ 344, 185, 40 },
		{ 344, 186, 40 },
		{ 344, 187, 40 },
		{ 344, 188, 40 },
		{ 344, 189, 40 },
		{ 344, 190, 40 },
		{ 344, 191, 40 },
		{ 344, 192, 40 },
		{ 344, 193, 40 },
		{ 344, 194, 40 },
		{ 344, 195, 40 },
		{ 344, 196, 40 },
		{ 344, 197, 40 },
		{ 344, 198, 40 },
		{ 344, 199, 40 },
		{ 344, 200, 40 },
		{ 344, 201, 40 },
		{ 344, 202, 40 },
		{ 345, 181, 40 },
		{ 345, 182, 40 },
		{ 345, 183, 40 },
		{ 345, 184, 40 },
		{ 345, 185, 40 },
		{ 345, 186, 40 },
		{ 345, 187, 40 },
		{ 345, 188, 40 },
		{ 345, 189, 40 },
		{ 345, 190, 40 },
		{ 345, 191, 40 },
		{ 345, 192, 40 },
		{ 345, 193, 40 },
		{ 345, 194, 40 },
		{ 345, 195, 40 },
		{ 345, 196, 40 },
		{ 345, 197, 40 },
		{ 345, 198, 40 },
		{ 345, 199, 40 },
		{ 345, 200, 40 },
		{ 345, 201, 40 },
		{ 346, 180, 40 },
		{ 346, 181, 40 },
		{ 346, 182, 40 },
		{ 346, 183, 40 },
		{ 346, 184, 40 },
		{ 346, 185, 40 },
		{ 346, 186, 40 },
		{ 346, 187, 40 },
		{ 346, 188, 40 },
		{ 346, 189, 40 },
		{ 346, 190, 40 },
		{ 346, 191, 40 },
		{ 346, 192, 40 },
		{ 346, 193, 40 },
		{ 346, 194, 40 },
		{ 346, 195, 40 },
		{ 346, 196, 40 },
		{ 346, 197, 40 },
		{ 346, 198, 40 },
		{ 346, 199, 40 },
		{ 346, 200, 40 },
		{ 346, 201, 40 },
		{ 347, 179, 40 },
		{ 347, 180, 40 },
		{ 347, 181, 40 },
		{ 347, 182, 40 },
		{ 347, 183, 40 },
		{ 347, 184, 40 },
		{ 347, 185, 40 },
		{ 347, 186, 40 },
		{ 347, 187, 40 },
		{ 347, 188, 40 },
		{ 347, 189, 40 },
		{ 347, 190, 40 },
		{ 347, 191, 40 },
		{ 347, 192, 40 },
		{ 347, 193, 40 },
		{ 347, 194, 40 },
		{ 347, 195, 40 },
		{ 347, 196, 40 },
		{ 347, 197, 40 },
		{ 347, 198, 40 },
		{ 347, 199, 40 },
		{ 347, 200, 40 },
		{ 348, 178, 40 },
		{ 348, 179, 40 },
		{ 348, 180, 40 },
		{ 348, 181, 40 },
		{ 348, 182, 40 },
		{ 348, 183, 40 },
		{ 348, 184, 40 },
		{ 348, 185, 40 },
		{ 348, 186, 40 },
		{ 348, 187, 40 },
		{ 348, 188, 40 },
		{ 348, 189, 40 },
		{ 348, 190, 40 },
		{ 348, 191, 40 },
		{ 348, 192, 40 },
		{ 348, 193, 40 },
		{ 348, 194, 40 },
		{ 348, 195, 40 },
		{ 348, 196, 40 },
		{ 348, 197, 40 },
		{ 348, 198, 40 },
		{ 348, 199, 40 },
		{ 348, 200, 40 },
		{ 349, 177, 40 },
		{ 349, 178, 40 },
		{ 349, 179, 40 },
		{ 349, 180, 40 },
		{ 349, 181, 40 },
		{ 349, 182, 40 },
		{ 349, 183, 40 },
		{ 349, 184, 40 },
		{ 349, 185, 40 },
		{ 349, 186, 40 },
		{ 349, 187, 40 },
		{ 349, 188, 40 },
		{ 349, 189, 40 },
		{ 349, 190, 40 },
		{ 349, 191, 40 },
		{ 349, 192, 40 },
		{ 349, 193, 40 },
		{ 349, 194, 40 },
		{ 349, 195, 40 },
		{ 349, 196, 40 },
		{ 349, 197, 40 },
		{ 349, 198, 40 },
		{ 349, 199, 40 },
		{ 350, 176, 40 },
		{ 350, 177, 40 },
		{ 350, 178, 40 },
		{ 350, 179, 40 },
		{ 350, 180, 40 },
		{ 350, 181, 40 },
		{ 350, 182, 40 },
		{ 350, 183, 40 },
		{ 350, 184, 40 },
		{ 350, 185, 40 },
		{ 350, 186, 40 },
		{ 350, 187, 40 },
		{ 350, 188, 40 },
		{ 350, 189, 40 },
		{ 350, 190, 40 },
		{ 350, 191, 40 },
		{ 350, 192, 40 },
		{ 350, 193, 40 },
		{ 350, 194, 40 },
		{ 350, 195, 40 },
		{ 350, 196, 40 },
		{ 350, 197, 40 },
		{ 350, 198, 40 },
		{ 351, 175, 40 },
		{ 351, 176, 40 },
		{ 351, 177, 40 },
		{ 351, 178, 40 },
		{ 351, 179, 40 },
		{ 351, 180, 40 },
		{ 351, 181, 40 },
		{ 351, 182, 40 },
		{ 351, 183, 40 },
		{ 351, 184, 40 },
		{ 351, 185, 40 },
		{ 351, 186, 40 },
		{ 351, 187, 40 },
		{ 351, 188, 40 },
		{ 351, 189, 40 },
		{ 351, 190, 40 },
		{ 351, 191, 40 },
		{ 351, 192, 40 },
		{ 351, 193, 40 },
		{ 351, 194, 40 },
		{ 351, 195, 40 },
		{ 351, 196, 40 },
		{ 351, 197, 40 },
		{ 351, 198, 40 },
		{ 352, 174, 40 },
		{ 352, 175, 40 },
		{ 352, 176, 40 },
		{ 352, 177, 40 },
		{ 352, 178, 40 },
		{ 352, 179, 40 },
		{ 352, 180, 40 },
		{ 352, 181, 40 },
		{ 352, 182, 40 },
		{ 352, 183, 40 },
		{ 352, 184, 40 },
		{ 352, 185, 40 },
		{ 352, 186, 40 },
		{ 352, 187, 40 },
		{ 352, 188, 40 },
		{ 352, 189, 40 },
		{ 352, 190, 40 },
		{ 352, 191, 40 },
		{ 352, 192, 40 },
		{ 352, 193, 40 },
		{ 352, 194, 40 },
		{ 352, 195, 40 },
		{ 352, 196, 40 },
		{ 352, 197, 40 },
		{ 353, 174, 40 },
		{ 353, 175, 40 },
		{ 353, 176, 40 },
		{ 353, 177, 40 },
		{ 353, 178, 40 },
		{ 353, 179, 40 },
		{ 353, 180, 40 },
		{ 353, 181, 40 },
		{ 353, 182, 40 },
		{ 353, 183, 40 },
		{ 353, 184, 40 },
		{ 353, 185, 40 },
		{ 353, 186, 40 },
		{ 353, 187, 40 },
		{ 353, 188, 40 },
		{ 353, 189, 40 },
		{ 353, 190, 40 },
		{ 353, 191, 40 },
		{ 353, 192, 40 },
		{ 353, 193, 40 },
		{ 353, 194, 40 },
		{ 353, 195, 40 },
		{ 353, 196, 40 },
		{ 354, 173, 40 },
		{ 354, 174, 40 },
		{ 354, 175, 40 },
		{ 354, 176, 40 },
		{ 354, 177, 40 },
		{ 354, 178, 40 },
		{ 354, 179, 40 },
		{ 354, 180, 40 },
		{ 354, 181, 40 },
		{ 354, 182, 40 },
		{ 354, 183, 40 },
		{ 354, 184, 40 },
		{ 354, 185, 40 },
		{ 354, 186, 40 },
		{ 354, 187, 40 },
		{ 354, 188, 40 },
		{ 354, 189, 40 },
		{ 354, 190, 40 },
		{ 354, 191, 40 },
		{ 354, 192, 40 },
		{ 354, 193, 40 },
		{ 354, 194, 40 },
		{ 354, 195, 40 },
		{ 354, 196, 40 },
		{ 355, 172, 40 },
		{ 355, 173, 40 },
		{ 355, 174, 40 },
		{ 355, 175, 40 },
		{ 355, 176, 40 },
		{ 355, 177, 40 },
		{ 355, 178, 40 },
		{ 355, 179, 40 },
		{ 355, 180, 40 },
		{ 355, 181, 40 },
		{ 355, 182, 40 },
		{ 355, 183, 40 },
		{ 355, 184, 40 },
		{ 355, 185, 40 },
		{ 355, 186, 40 },
		{ 355, 187, 40 },
		{ 355, 188, 40 },
		{ 355, 189, 40 },
		{ 355, 190, 40 },
		{ 355, 191, 40 },
		{ 355, 192, 40 },
		{ 355, 193, 40 },
		{ 355, 194, 40 },
		{ 355, 195, 40 },
		{ 356, 171, 40 },
		{ 356, 172, 40 },
		{ 356, 173, 40 },
		{ 356, 174, 40 },
		{ 356, 175, 40 },
		{ 356, 176, 40 },
		{ 356, 177, 40 },
		{ 356, 178, 40 },
		{ 356, 179, 40 },
		{ 356, 180, 40 },
		{ 356, 181, 40 },
		{ 356, 182, 40 },
		{ 356, 183, 40 },
		{ 356, 184, 40 },
		{ 356, 185, 40 },
		{ 356, 186, 40 },
		{ 356, 187, 40 },
		{ 356, 188, 40 },
		{ 356, 189, 40 },
		{ 356, 190, 40 },
		{ 356, 191, 40 },
		{ 356, 192, 40 },
		{ 356, 193, 40 },
		{ 356, 194, 40 },
		{ 356, 195, 40 },
		{ 357, 170, 40 },
		{ 357, 171, 40 },
		{ 357, 172, 40 },
		{ 357, 173, 40 },
		{ 357, 174, 40 },
		{ 357, 175, 40 },
		{ 357, 176, 40 },
		{ 357, 177, 40 },
		{ 357, 178, 40 },
		{ 357, 179, 40 },
		{ 357, 180, 40 },
		{ 357, 181, 40 },
		{ 357, 182, 40 },
		{ 357, 183, 40 },
		{ 357, 184, 40 },
		{ 357, 185, 40 },
		{ 357, 186, 40 },
		{ 357, 187, 40 },
		{ 357, 188, 40 },
		{ 357, 189, 40 },
		{ 357, 190, 40 },
		{ 357, 191, 40 },
		{ 357, 192, 40 },
		{ 357, 193, 40 },
		{ 357, 194, 40 },
		{ 358, 169, 40 },
		{ 358, 170, 40 },
		{ 358, 171, 40 },
		{ 358, 172, 40 },
		{ 358, 173, 40 },
		{ 358, 174, 40 },
		{ 358, 175, 40 },
		{ 358, 176, 40 },
		{ 358, 177, 40 },
		{ 358, 178, 40 },
		{ 358, 179, 40 },
		{ 358, 180, 40 },
		{ 358, 181, 40 },
		{ 358, 182, 40 },
		{ 358, 183, 40 },
		{ 358, 184, 40 },
		{ 358, 185, 40 },
		{ 358, 186, 40 },
		{ 358, 187, 40 },
		{ 358, 188, 40 },
		{ 358, 189, 40 },
		{ 358, 190, 40 },
		{ 358, 191, 40 },
		{ 358, 192, 40 },
		{ 358, 193, 40 },
		{ 359, 168, 40 },
		{ 359, 169, 40 },
		{ 359, 170, 40 },
		{ 359, 171, 40 },
		{ 359, 172, 40 },
		{ 359, 173, 40 },
		{ 359, 174, 40 },
		{ 359, 175, 40 },
		{ 359, 176, 40 },
		{ 359, 177, 40 },
		{ 359, 178, 40 },
		{ 359, 179, 40 },
		{ 359, 180, 40 },
		{ 359, 181, 40 },
		{ 359, 182, 40 },
		{ 359, 183, 40 },
		{ 359, 184, 40 },
		{ 359, 185, 40 },
		{ 359, 186, 40 },
		{ 359, 187, 40 },
		{ 359, 188, 40 },
		{ 359, 189, 40 },
		{ 359, 190, 40 },
		{ 359, 191, 40 },
		{ 359, 192, 40 },
		{ 359, 193, 40 },
		{ 360, 167, 40 },
		{ 360, 168, 40 },
		{ 360, 169, 40 },
		{ 360, 170, 40 },
		{ 360, 171, 40 },
		{ 360, 172, 40 },
		{ 360, 173, 40 },
		{ 360, 174, 40 },
		{ 360, 175, 40 },
		{ 360, 176, 40 },
		{ 360, 177, 40 },
		{ 360, 178, 40 },
		{ 360, 179, 40 },
		{ 360, 180, 40 },
		{ 360, 181, 40 },
		{ 360, 182, 40 },
		{ 360, 183, 40 },
		{ 360, 184, 40 },
		{ 360, 185, 40 },
		{ 360, 186, 40 },
		{ 360, 187, 40 },
		{ 360, 188, 40 },
		{ 360, 189, 40 },
		{ 360, 190, 40 },
		{ 360, 191, 40 },
		{ 360, 192, 40 },
		{ 361, 167, 40 },
		{ 361, 168, 40 },
		{ 361, 169, 40 },
		{ 361, 170, 40 },
		{ 361, 171, 40 },
		{ 361, 172, 40 },
		{ 361, 173, 40 },
		{ 361, 174, 40 },
		{ 361, 175, 40 },
		{ 361, 176, 40 },
		{ 361, 177, 40 },
		{ 361, 178, 40 },
		{ 361, 179, 40 },
		{ 361, 180, 40 },
		{ 361, 181, 40 },
		{ 361, 182, 40 },
		{ 361, 183, 40 },
		{ 361, 184, 40 },
		{ 361, 185, 40 },
		{ 361, 186, 40 },
		{ 361, 187, 40 },
		{ 361, 188, 40 },
		{ 361, 189, 40 },
		{ 361, 190, 40 },
		{ 361, 191, 40 },
		{ 362, 166, 40 },
		{ 362, 167, 40 },
		{ 362, 168, 40 },
		{ 362, 169, 40 },
		{ 362, 170, 40 },
		{ 362, 171, 40 },
		{ 362, 172, 40 },
		{ 362, 173, 40 },
		{ 362, 174, 40 },
		{ 362, 175, 40 },
		{ 362, 176, 40 },
		{ 362, 177, 40 },
		{ 362, 178, 40 },
		{ 362, 179, 40 },
		{ 362, 180, 40 },
		{ 362, 181, 40 },
		{ 362, 182, 40 },
		{ 362, 183, 40 },
		{ 362, 184, 40 },
		{ 362, 185, 40 },
		{ 362, 186, 40 },
		{ 362, 187, 40 },
		{ 362, 188, 40 },
		{ 362, 189, 40 },
		{ 362, 190, 40 },
		{ 362, 191, 40 },
		{ 363, 165, 40 },
		{ 363, 166, 40 },
		{ 363, 167, 40 },
		{ 363, 168, 40 },
		{ 363, 169, 40 },
		{ 363, 170, 40 },
		{ 363, 171, 40 },
		{ 363, 172, 40 },
		{ 363, 173, 40 },
		{ 363, 174, 40 },
		{ 363, 175, 40 },
		{ 363, 176, 40 },
		{ 363, 177, 40 },
		{ 363, 178, 40 },
		{ 363, 179, 40 },
		{ 363, 180, 40 },
		{ 363, 181, 40 },
		{ 363, 182, 40 },
		{ 363, 183, 40 },
		{ 363, 184, 40 },
		{ 363, 185, 40 },
		{ 363, 186, 40 },
		{ 363, 187, 40 },
		{ 363, 188, 40 },
		{ 363, 189, 40 },
		{ 363, 190, 40 },
		{ 364, 164, 40 },
		{ 364, 165, 40 },
		{ 364, 166, 40 },
		{ 364, 167, 40 },
		{ 364, 168, 40 },
		{ 364, 169, 40 },
		{ 364, 170, 40 },
		{ 364, 171, 40 },
		{ 364, 172, 40 },
		{ 364, 173, 40 },
		{ 364, 174, 40 },
		{ 364, 175, 40 },
		{ 364, 176, 40 },
		{ 364, 177, 40 },
		{ 364, 178, 40 },
		{ 364, 179, 40 },
		{ 364, 180, 40 },
		{ 364, 181, 40 },
		{ 364, 182, 40 },
		{ 364, 183, 40 },
		{ 364, 184, 40 },
		{ 364, 185, 40 },
		{ 364, 186, 40 },
		{ 364, 187, 40 },
		{ 364, 188, 40 },
		{ 364, 189, 40 },
		{ 364, 190, 40 },
		{ 365, 163, 40 },
		{ 365, 164, 40 },
		{ 365, 165, 40 },
		{ 365, 166, 40 },
		{ 365, 167, 40 },
		{ 365, 168, 40 },
		{ 365, 169, 40 },
		{ 365, 170, 40 },
		{ 365, 171, 40 },
		{ 365, 172, 40 },
		{ 365, 173, 40 },
		{ 365, 174, 40 },
		{ 365, 175, 40 },
		{ 365, 176, 40 },
		{ 365, 177, 40 },
		{ 365, 178, 40 },
		{ 365, 179, 40 },
		{ 365, 180, 40 },
		{ 365, 181, 40 },
		{ 365, 182, 40 },
		{ 365, 183, 40 },
		{ 365, 184, 40 },
		{ 365, 185, 40 },
		{ 365, 186, 40 },
		{ 365, 187, 40 },
		{ 365, 188, 40 },
		{ 365, 189, 40 },
		{ 366, 162, 40 },
		{ 366, 163, 40 },
		{ 366, 164, 40 },
		{ 366, 165, 40 },
		{ 366, 166, 40 },
		{ 366, 167, 40 },
		{ 366, 168, 40 },
		{ 366, 169, 40 },
		{ 366, 170, 40 },
		{ 366, 171, 40 },
		{ 366, 172, 40 },
		{ 366, 173, 40 },
		{ 366, 174, 40 },
		{ 366, 175, 40 },
		{ 366, 176, 40 },
		{ 366, 177, 40 },
		{ 366, 178, 40 },
		{ 366, 179, 40 },
		{ 366, 180, 40 },
		{ 366, 181, 40 },
		{ 366, 182, 40 },
		{ 366, 183, 40 },
		{ 366, 184, 40 },
		{ 366, 185, 40 },
		{ 366, 186, 40 },
		{ 366, 187, 40 },
		{ 366, 188, 40 },
		{ 367, 161, 40 },
		{ 367, 162, 40 },
		{ 367, 163, 40 },
		{ 367, 164, 40 },
		{ 367, 165, 40 },
		{ 367, 166, 40 },
		{ 367, 167, 40 },
		{ 367, 168, 40 },
		{ 367, 169, 40 },
		{ 367, 170, 40 },
		{ 367, 171, 40 },
		{ 367, 172, 40 },
		{ 367, 173, 40 },
		{ 367, 174, 40 },
		{ 367, 175, 40 },
		{ 367, 176, 40 },
		{ 367, 177, 40 },
		{ 367, 178, 40 },
		{ 367, 179, 40 },
		{ 367, 180, 40 },
		{ 367, 181, 40 },
		{ 367, 182, 40 },
		{ 367, 183, 40 },
		{ 367, 184, 40 },
		{ 367, 185, 40 },
		{ 367, 186, 40 },
		{ 367, 187, 40 },
		{ 367, 188, 40 },
		{ 368, 160, 40 },
		{ 368, 161, 40 },
		{ 368, 162, 40 },
		{ 368, 163, 40 },
		{ 368, 164, 40 },
		{ 368, 165, 40 },
		{ 368, 166, 40 },
		{ 368, 167, 40 },
		{ 368, 168, 40 },
		{ 368, 169, 40 },
		{ 368, 170, 40 },
		{ 368, 171, 40 },
		{ 368, 172, 40 },
		{ 368, 173, 40 },
		{ 368, 174, 40 },
		{ 368, 175, 40 },
		{ 368, 176, 40 },
		{ 368, 177, 40 },
		{ 368, 178, 40 },
		{ 368, 179, 40 },
		{ 368, 180, 40 },
		{ 368, 181, 40 },
		{ 368, 182, 40 },
		{ 368, 183, 40 },
		{ 368, 184, 40 },
		{ 368, 185, 40 },
		{ 368, 186, 40 },
		{ 368, 187, 40 },
		{ 369, 160, 40 },
		{ 369, 161, 40 },
		{ 369, 162, 40 },
		{ 369, 163, 40 },
		{ 369, 164, 40 },
		{ 369, 165, 40 },
		{ 369, 166, 40 },
		{ 369, 167, 40 },
		{ 369, 168, 40 },
		{ 369, 169, 40 },
		{ 369, 170, 40 },
		{ 369, 171, 40 },
		{ 369, 172, 40 },
		{ 369, 173, 40 },
		{ 369, 174, 40 },
		{ 369, 175, 40 },
		{ 369, 176, 40 },
		{ 369, 177, 40 },
		{ 369, 178, 40 },
		{ 369, 179, 40 },
		{ 369, 180, 40 },
		{ 369, 181, 40 },
		{ 369, 182, 40 },
		{ 369, 183, 40 },
		{ 369, 184, 40 },
		{ 369, 185, 40 },
		{ 369, 186, 40 },
		{ 370, 159, 40 },
		{ 370, 160, 40 },
		{ 370, 161, 40 },
		{ 370, 162, 40 },
		{ 370, 163, 40 },
		{ 370, 164, 40 },
		{ 370, 165, 40 },
		{ 370, 166, 40 },
		{ 370, 167, 40 },
		{ 370, 168, 40 },
		{ 370, 169, 40 },
		{ 370, 170, 40 },
		{ 370, 171, 40 },
		{ 370, 172, 40 },
		{ 370, 173, 40 },
		{ 370, 174, 40 },
		{ 370, 175, 40 },
		{ 370, 176, 40 },
		{ 370, 177, 40 },
		{ 370, 178, 40 },
		{ 370, 179, 40 },
		{ 370, 180, 40 },
		{ 370, 181, 40 },
		{ 370, 182, 40 },
		{ 370, 183, 40 },
		{ 370, 184, 40 },
		{ 370, 185, 40 },
		{ 370, 186, 40 },
		{ 371, 158, 40 },
		{ 371, 159, 40 },
		{ 371, 160, 40 },
		{ 371, 161, 40 },
		{ 371, 162, 40 },
		{ 371, 163, 40 },
		{ 371, 164, 40 },
		{ 371, 165, 40 },
		{ 371, 166, 40 },
		{ 371, 167, 40 },
		{ 371, 168, 40 },
		{ 371, 169, 40 },
		{ 371, 170, 40 },
		{ 371, 171, 40 },
		{ 371, 172, 40 },
		{ 371, 173, 40 },
		{ 371, 174, 40 },
		{ 371, 175, 40 },
		{ 371, 176, 40 },
		{ 371, 177, 40 },
		{ 371, 178, 40 },
		{ 371, 179, 40 },
		{ 371, 180, 40 },
		{ 371, 181, 40 },
		{ 371, 182, 40 },
		{ 371, 183, 40 },
		{ 371, 184, 40 },
		{ 371, 185, 40 },
		{ 372, 157, 40 },
		{ 372, 158, 40 },
		{ 372, 159, 40 },
		{ 372, 160, 40 },
		{ 372, 161, 40 },
		{ 372, 162, 40 },
		{ 372, 163, 40 },
		{ 372, 164, 40 },
		{ 372, 165, 40 },
		{ 372, 166, 40 },
		{ 372, 167, 40 },
		{ 372, 168, 40 },
		{ 372, 169, 40 },
		{ 372, 170, 40 },
		{ 372, 171, 40 },
		{ 372, 172, 40 },
		{ 372, 173, 40 },
		{ 372, 174, 40 },
		{ 372, 175, 40 },
		{ 372, 176, 40 },
		{ 372, 177, 40 },
		{ 372, 178, 40 },
		{ 372, 179, 40 },
		{ 372, 180, 40 },
		{ 372, 181, 40 },
		{ 372, 182, 40 },
		{ 372, 183, 40 },
		{ 372, 184, 40 },
		{ 372, 185, 40 },
		{ 373, 156, 40 },
		{ 373, 157, 40 },
		{ 373, 158, 40 },
		{ 373, 159, 40 },
		{ 373, 160, 40 },
		{ 373, 161, 40 },
		{ 373, 162, 40 },
		{ 373, 163, 40 },
		{ 373, 164, 40 },
		{ 373, 165, 40 },
		{ 373, 166, 40 },
		{ 373, 167, 40 },
		{ 373, 168, 40 },
		{ 373, 169, 40 },
		{ 373, 170, 40 },
		{ 373, 171, 40 },
		{ 373, 172, 40 },
		{ 373, 173, 40 },
		{ 373, 174, 40 },
		{ 373, 175, 40 },
		{ 373, 176, 40 },
		{ 373, 177, 40 },
		{ 373, 178, 40 },
		{ 373, 179, 40 },
		{ 373, 180, 40 },
		{ 373, 181, 40 },
		{ 373, 182, 40 },
		{ 373, 183, 40 },
		{ 373, 184, 40 },
		{ 374, 155, 40 },
		{ 374, 156, 40 },
		{ 374, 157, 40 },
		{ 374, 158, 40 },
		{ 374, 159, 40 },
		{ 374, 160, 40 },
		{ 374, 161, 40 },
		{ 374, 162, 40 },
		{ 374, 163, 40 },
		{ 374, 164, 40 },
		{ 374, 165, 40 },
		{ 374, 166, 40 },
		{ 374, 167, 40 },
		{ 374, 168, 40 },
		{ 374, 169, 40 },
		{ 374, 170, 40 },
		{ 374, 171, 40 },
		{ 374, 172, 40 },
		{ 374, 173, 40 },
		{ 374, 174, 40 },
		{ 374, 175, 40 },
		{ 374, 176, 40 },
		{ 374, 177, 40 },
		{ 374, 178, 40 },
		{ 374, 179, 40 },
		{ 374, 180, 40 },
		{ 374, 181, 40 },
		{ 374, 182, 40 },
		{ 374, 183, 40 },
		{ 375, 154, 40 },
		{ 375, 155, 40 },
		{ 375, 156, 40 },
		{ 375, 157, 40 },
		{ 375, 158, 40 },
		{ 375, 159, 40 },
		{ 375, 160, 40 },
		{ 375, 161, 40 },
		{ 375, 162, 40 },
		{ 375, 163, 40 },
		{ 375, 164, 40 },
		{ 375, 165, 40 },
		{ 375, 166, 40 },
		{ 375, 167, 40 },
		{ 375, 168, 40 },
		{ 375, 169, 40 },
		{ 375, 170, 40 },
		{ 375, 171, 40 },
		{ 375, 172, 40 },
		{ 375, 173, 40 },
		{ 375, 174, 40 },
		{ 375, 175, 40 },
		{ 375, 176, 40 },
		{ 375, 177, 40 },
		{ 375, 178, 40 },
		{ 375, 179, 40 },
		{ 375, 180, 40 },
		{ 375, 181, 40 },
		{ 375, 182, 40 },
		{ 375, 183, 40 },
		{ 376, 153, 40 },
		{ 376, 154, 40 },
		{ 376, 155, 40 },
		{ 376, 156, 40 },
		{ 376, 157, 40 },
		{ 376, 158, 40 },
		{ 376, 159, 40 },
		{ 376, 160, 40 },
		{ 376, 161, 40 },
		{ 376, 162, 40 },
		{ 376, 163, 40 },
		{ 376, 164, 40 },
		{ 376, 165, 40 },
		{ 376, 166, 40 },
		{ 376, 167, 40 },
		{ 376, 168, 40 },
		{ 376, 169, 40 },
		{ 376, 170, 40 },
		{ 376, 171, 40 },
		{ 376, 172, 40 },
		{ 376, 173, 40 },
		{ 376, 174, 40 },
		{ 376, 175, 40 },
		{ 376, 176, 40 },
		{ 376, 177, 40 },
		{ 376, 178, 40 },
		{ 376, 179, 40 },
		{ 376, 180, 40 },
		{ 376, 181, 40 },
		{ 376, 182, 40 },
		{ 377, 153, 40 },
		{ 377, 154, 40 },
		{ 377, 155, 40 },
		{ 377, 156, 40 },
		{ 377, 157, 40 },
		{ 377, 158, 40 },
		{ 377, 159, 40 },
		{ 377, 160, 40 },
		{ 377, 161, 40 },
		{ 377, 162, 40 },
		{ 377, 163, 40 },
		{ 377, 164, 40 },
		{ 377, 165, 40 },
		{ 377, 166, 40 },
		{ 377, 167, 40 },
		{ 377, 168, 40 },
		{ 377, 169, 40 },
		{ 377, 170, 40 },
		{ 377, 171, 40 },
		{ 377, 172, 40 },
		{ 377, 173, 40 },
		{ 377, 174, 40 },
		{ 377, 175, 40 },
		{ 377, 176, 40 },
		{ 377, 177, 40 },
		{ 377, 178, 40 },
		{ 377, 179, 40 },
		{ 377, 180, 40 },
		{ 377, 181, 40 },
		{ 378, 152, 40 },
		{ 378, 153, 40 },
		{ 378, 154, 40 },
		{ 378, 155, 40 },
		{ 378, 156, 40 },
		{ 378, 157, 40 },
		{ 378, 158, 40 },
		{ 378, 159, 40 },
		{ 378, 160, 40 },
		{ 378, 161, 40 },
		{ 378, 162, 40 },
		{ 378, 163, 40 },
		{ 378, 164, 40 },
		{ 378, 165, 40 },
		{ 378, 166, 40 },
		{ 378, 167, 40 },
		{ 378, 168, 40 },
		{ 378, 169, 40 },
		{ 378, 170, 40 },
		{ 378, 171, 40 },
		{ 378, 172, 40 },
		{ 378, 173, 40 },
		{ 378, 174, 40 },
		{ 378, 175, 40 },
		{ 378, 176, 40 },
		{ 378, 177, 40 },
		{ 378, 178, 40 },
		{ 378, 179, 40 },
		{ 378, 180, 40 },
		{ 378, 181, 40 },
		{ 379, 151, 40 },
		{ 379, 152, 40 },
		{ 379, 153, 40 },
		{ 379, 154, 40 },
		{ 379, 155, 40 },
		{ 379, 156, 40 },
		{ 379, 157, 40 },
		{ 379, 158, 40 },
		{ 379, 159, 40 },
		{ 379, 160, 40 },
		{ 379, 161, 40 },
		{ 379, 162, 40 },
		{ 379, 163, 40 },
		{ 379, 164, 40 },
		{ 379, 165, 40 },
		{ 379, 166, 40 },
		{ 379, 167, 40 },
		{ 379, 168, 40 },
		{ 379, 169, 40 },
		{ 379, 170, 40 },
		{ 379, 171, 40 },
		{ 379, 172, 40 },
		{ 379, 173, 40 },
		{ 379, 174, 40 },
		{ 379, 175, 40 },
		{ 379, 176, 40 },
		{ 379, 177, 40 },
		{ 379, 178, 40 },
		{ 379, 179, 40 },
		{ 379, 180, 40 },
		{ 380, 150, 40 },
		{ 380, 151, 40 },
		{ 380, 152, 40 },
		{ 380, 153, 40 },
		{ 380, 154, 40 },
		{ 380, 155, 40 },
		{ 380, 156, 40 },
		{ 380, 157, 40 },
		{ 380, 158, 40 },
		{ 380, 159, 40 },
		{ 380, 160, 40 },
		{ 380, 161, 40 },
		{ 380, 162, 40 },
		{ 380, 163, 40 },
		{ 380, 164, 40 },
		{ 380, 165, 40 },
		{ 380, 166, 40 },
		{ 380, 167, 40 },
		{ 380, 168, 40 },
		{ 380, 169, 40 },
		{ 380, 170, 40 },
		{ 380, 171, 40 },
		{ 380, 172, 40 },
		{ 380, 173, 40 },
		{ 380, 174, 40 },
		{ 380, 175, 40 },
		{ 380, 176, 40 },
		{ 380, 177, 40 },
		{ 380, 178, 40 },
		{ 380, 179, 40 },
		{ 380, 180, 40 },
		{ 381, 149, 40 },
		{ 381, 150, 40 },
		{ 381, 151, 40 },
		{ 381, 152, 40 },
		{ 381, 153, 40 },
		{ 381, 154, 40 },
		{ 381, 155, 40 },
		{ 381, 156, 40 },
		{ 381, 157, 40 },
		{ 381, 158, 40 },
		{ 381, 159, 40 },
		{ 381, 160, 40 },
		{ 381, 161, 40 },
		{ 381, 162, 40 },
		{ 381, 163, 40 },
		{ 381, 164, 40 },
		{ 381, 165, 40 },
		{ 381, 166, 40 },
		{ 381, 167, 40 },
		{ 381, 168, 40 },
		{ 381, 169, 40 },
		{ 381, 170, 40 },
		{ 381, 171, 40 },
		{ 381, 172, 40 },
		{ 381, 173, 40 },
		{ 381, 174, 40 },
		{ 381, 175, 40 },
		{ 381, 176, 40 },
		{ 381, 177, 40 },
		{ 381, 178, 40 },
		{ 381, 179, 40 },
		{ 382, 148, 40 },
		{ 382, 149, 40 },
		{ 382, 150, 40 },
		{ 382, 151, 40 },
		{ 382, 152, 40 },
		{ 382, 153, 40 },
		{ 382, 154, 40 },
		{ 382, 155, 40 },
		{ 382, 156, 40 },
		{ 382, 157, 40 },
		{ 382, 158, 40 },
		{ 382, 159, 40 },
		{ 382, 160, 40 },
		{ 382, 161, 40 },
		{ 382, 162, 40 },
		{ 382, 163, 40 },
		{ 382, 164, 40 },
		{ 382, 165, 40 },
		{ 382, 166, 40 },
		{ 382, 167, 40 },
		{ 382, 168, 40 },
		{ 382, 169, 40 },
		{ 382, 170, 40 },
		{ 382, 171, 40 },
		{ 382, 172, 40 },
		{ 382, 173, 40 },
		{ 382, 174, 40 },
		{ 382, 175, 40 },
		{ 382, 176, 40 },
		{ 382, 177, 40 },
		{ 382, 178, 40 },
		{ 383, 147, 40 },
		{ 383, 148, 40 },
		{ 383, 149, 40 },
		{ 383, 150, 40 },
		{ 383, 151, 40 },
		{ 383, 152, 40 },
		{ 383, 153, 40 },
		{ 383, 154, 40 },
		{ 383, 155, 40 },
		{ 383, 156, 40 },
		{ 383, 157, 40 },
		{ 383, 158, 40 },
		{ 383, 159, 40 },
		{ 383, 160, 40 },
		{ 383, 161, 40 },
		{ 383, 162, 40 },
		{ 383, 163, 40 },
		{ 383, 164, 40 },
		{ 383, 165, 40 },
		{ 383, 166, 40 },
		{ 383, 167, 40 },
		{ 383, 168, 40 },
		{ 383, 169, 40 },
		{ 383, 170, 40 },
		{ 383, 171, 40 },
		{ 383, 172, 40 },
		{ 383, 173, 40 },
		{ 383, 174, 40 },
		{ 383, 175, 40 },
		{ 383, 176, 40 },
		{ 383, 177, 40 },
		{ 383, 178, 40 },
		{ 384, 146, 40 },
		{ 384, 147, 40 },
		{ 384, 148, 40 },
		{ 384, 149, 40 },
		{ 384, 150, 40 },
		{ 384, 151, 40 },
		{ 384, 152, 40 },
		{ 384, 153, 40 },
		{ 384, 154, 40 },
		{ 384, 155, 40 },
		{ 384, 156, 40 },
		{ 384, 157, 40 },
		{ 384, 158, 40 },
		{ 384, 159, 40 },
		{ 384, 160, 40 },
		{ 384, 161, 40 },
		{ 384, 162, 40 },
		{ 384, 163, 40 },
		{ 384, 164, 40 },
		{ 384, 165, 40 },
		{ 384, 166, 40 },
		{ 384, 167, 40 },
		{ 384, 168, 40 },
		{ 384, 169, 40 },
		{ 384, 170, 40 },
		{ 384, 171, 40 },
		{ 384, 172, 40 },
		{ 384, 173, 40 },
		{ 384, 174, 40 },
		{ 384, 175, 40 },
		{ 384, 176, 40 },
		{ 384, 177, 40 },
		{ 385, 146, 40 },
		{ 385, 147, 40 },
		{ 385, 148, 40 },
		{ 385, 149, 40 },
		{ 385, 150, 40 },
		{ 385, 151, 40 },
		{ 385, 152, 40 },
		{ 385, 153, 40 },
		{ 385, 154, 40 },
		{ 385, 155, 40 },
		{ 385, 156, 40 },
		{ 385, 157, 40 },
		{ 385, 158, 40 },
		{ 385, 159, 40 },
		{ 385, 160, 40 },
		{ 385, 161, 40 },
		{ 385, 162, 40 },
		{ 385, 163, 40 },
		{ 385, 164, 40 },
		{ 385, 165, 40 },
		{ 385, 166, 40 },
		{ 385, 167, 40 },
		{ 385, 168, 40 },
		{ 385, 169, 40 },
		{ 385, 170, 40 },
		{ 385, 171, 40 },
		{ 385, 172, 40 },
		{ 385, 173, 40 },
		{ 385, 174, 40 },
		{ 385, 175, 40 },
		{ 385, 176, 40 },
		{ 386, 145, 40 },
		{ 386, 146, 40 },
		{ 386, 147, 40 },
		{ 386, 148, 40 },
		{ 386, 149, 40 },
		{ 386, 150, 40 },
		{ 386, 151, 40 },
		{ 386, 152, 40 },
		{ 386, 153, 40 },
		{ 386, 154, 40 },
		{ 386, 155, 40 },
		{ 386, 156, 40 },
		{ 386, 157, 40 },
		{ 386, 158, 40 },
		{ 386, 159, 40 },
		{ 386, 160, 40 },
		{ 386, 161, 40 },
		{ 386, 162, 40 },
		{ 386, 163, 40 },
		{ 386, 164, 40 },
		{ 386, 165, 40 },
		{ 386, 166, 40 },
		{ 386, 167, 40 },
		{ 386, 168, 40 },
		{ 386, 169, 40 },
		{ 386, 170, 40 },
		{ 386, 171, 40 },
		{ 386, 172, 40 },
		{ 386, 173, 40 },
		{ 386, 174, 40 },
		{ 386, 175, 40 },
		{ 386, 176, 40 },
		{ 387, 144, 40 },
		{ 387, 145, 40 },
		{ 387, 146, 40 },
		{ 387, 147, 40 },
		{ 387, 148, 40 },
		{ 387, 149, 40 },
		{ 387, 150, 40 },
		{ 387, 151, 40 },
		{ 387, 152, 40 },
		{ 387, 153, 40 },
		{ 387, 154, 40 },
		{ 387, 155, 40 },
		{ 387, 156, 40 },
		{ 387, 157, 40 },
		{ 387, 158, 40 },
		{ 387, 159, 40 },
		{ 387, 160, 40 },
		{ 387, 161, 40 },
		{ 387, 162, 40 },
		{ 387, 163, 40 },
		{ 387, 164, 40 },
		{ 387, 165, 40 },
		{ 387, 166, 40 },
		{ 387, 167, 40 },
		{ 387, 168, 40 },
		{ 387, 169, 40 },
		{ 387, 170, 40 },
		{ 387, 171, 40 },
		{ 387, 172, 40 },
		{ 387, 173, 40 },
		{ 387, 174, 40 },
		{ 387, 175, 40 },
		{ 388, 143, 40 },
		{ 388, 144, 40 },
		{ 388, 145, 40 },
		{ 388, 146, 40 },
		{ 388, 147, 40 },
		{ 388, 148, 40 },
		{ 388, 149, 40 },
		{ 388, 150, 40 },
		{ 388, 151, 40 },
		{ 388, 152, 40 },
		{ 388, 153, 40 },
		{ 388, 154, 40 },
		{ 388, 155, 40 },
		{ 388, 156, 40 },
		{ 388, 157, 40 },
		{ 388, 158, 40 },
		{ 388, 159, 40 },
		{ 388, 160, 40 },
		{ 388, 161, 40 },
		{ 388, 162, 40 },
		{ 388, 163, 40 },
		{ 388, 164, 40 },
		{ 388, 165, 40 },
		{ 388, 166, 40 },
		{ 388, 167, 40 },
		{ 388, 168, 40 },
		{ 388, 169, 40 },
		{ 388, 170, 40 },
		{ 388, 171, 40 },
		{ 388, 172, 40 },
		{ 388, 173, 40 },
		{ 388, 174, 40 },
		{ 388, 175, 40 },
		{ 389, 142, 40 },
		{ 389, 143, 40 },
		{ 389, 144, 40 },
		{ 389, 145, 40 },
		{ 389, 146, 40 },
		{ 389, 147, 40 },
		{ 389, 148, 40 },
		{ 389, 149, 40 },
		{ 389, 150, 40 },
		{ 389, 151, 40 },
		{ 389, 152, 40 },
		{ 389, 153, 40 },
		{ 389, 154, 40 },
		{ 389, 155, 40 },
		{ 389, 156, 40 },
		{ 389, 157, 40 },
		{ 389, 158, 40 },
		{ 389, 159, 40 },
		{ 389, 160, 40 },
		{ 389, 161, 40 },
		{ 389, 162, 40 },
		{ 389, 163, 40 },
		{ 389, 164, 40 },
		{ 389, 165, 40 },
		{ 389, 166, 40 },
		{ 389, 167, 40 },
		{ 389, 168, 40 },
		{ 389, 169, 40 },
		{ 389, 170, 40 },
		{ 389, 171, 40 },
		{ 389, 172, 40 },
		{ 389, 173, 40 },
		{ 389, 174, 40 },
		{ 390, 141, 40 },
		{ 390, 142, 40 },
		{ 390, 143, 40 },
		{ 390, 144, 40 },
		{ 390, 145, 40 },
		{ 390, 146, 40 },
		{ 390, 147, 40 },
		{ 390, 148, 40 },
		{ 390, 149, 40 },
		{ 390, 150, 40 },
		{ 390, 151, 40 },
		{ 390, 152, 40 },
		{ 390, 153, 40 },
		{ 390, 154, 40 },
		{ 390, 155, 40 },
		{ 390, 156, 40 },
		{ 390, 157, 40 },
		{ 390, 158, 40 },
		{ 390, 159, 40 },
		{ 390, 160, 40 },
		{ 390, 161, 40 },
		{ 390, 162, 40 },
		{ 390, 163, 40 },
		{ 390, 164, 40 },
		{ 390, 165, 40 },
		{ 390, 166, 40 },
		{ 390, 167, 40 },
		{ 390, 168, 40 },
		{ 390, 169, 40 },
		{ 390, 170, 40 },
		{ 390, 171, 40 },
		{ 390, 172, 40 },
		{ 390, 173, 40 },
		{ 391, 140, 40 },
		{ 391, 141, 40 },
		{ 391, 142, 40 },
		{ 391, 143, 40 },
		{ 391, 144, 40 },
		{ 391, 145, 40 },
		{ 391, 146, 40 },
		{ 391, 147, 40 },
		{ 391, 148, 40 },
		{ 391, 149, 40 },
		{ 391, 150, 40 },
		{ 391, 151, 40 },
		{ 391, 152, 40 },
		{ 391, 153, 40 },
		{ 391, 154, 40 },
		{ 391, 155, 40 },
		{ 391, 156, 40 },
		{ 391, 157, 40 },
		{ 391, 158, 40 },
		{ 391, 159, 40 },
		{ 391, 160, 40 },
		{ 391, 161, 40 },
		{ 391, 162, 40 },
		{ 391, 163, 40 },
		{ 391, 164, 40 },
		{ 391, 165, 40 },
		{ 391, 166, 40 },
		{ 391, 167, 40 },
		{ 391, 168, 40 },
		{ 391, 169, 40 },
		{ 391, 170, 40 },
		{ 391, 171, 40 },
		{ 391, 172, 40 },
		{ 391, 173, 40 },
		{ 392, 139, 40 },
		{ 392, 140, 40 },
		{ 392, 141, 40 },
		{ 392, 142, 40 },
		{ 392, 143, 40 },
		{ 392, 144, 40 },
		{ 392, 145, 40 },
		{ 392, 146, 40 },
		{ 392, 147, 40 },
		{ 392, 148, 40 },
		{ 392, 149, 40 },
		{ 392, 150, 40 },
		{ 392, 151, 40 },
		{ 392, 152, 40 },
		{ 392, 153, 40 },
		{ 392, 154, 40 },
		{ 392, 155, 40 },
		{ 392, 156, 40 },
		{ 392, 157, 40 },
		{ 392, 158, 40 },
		{ 392, 159, 40 },
		{ 392, 160, 40 },
		{ 392, 161, 40 },
		{ 392, 162, 40 },
		{ 392, 163, 40 },
		{ 392, 164, 40 },
		{ 392, 165, 40 },
		{ 392, 166, 40 },
		{ 392, 167, 40 },
		{ 392, 168, 40 },
		{ 392, 169, 40 },
		{ 392, 170, 40 },
		{ 392, 171, 40 },
		{ 392, 172, 40 },
		{ 393, 139, 40 },
		{ 393, 140, 40 },
		{ 393, 141, 40 },
		{ 393, 142, 40 },
		{ 393, 143, 40 },
		{ 393, 144, 40 },
		{ 393, 145, 40 },
		{ 393, 146, 40 },
		{ 393, 147, 40 },
		{ 393, 148, 40 },
		{ 393, 149, 40 },
		{ 393, 150, 40 },
		{ 393, 151, 40 },
		{ 393, 152, 40 },
		{ 393, 153, 40 },
		{ 393, 154, 40 },
		{ 393, 155, 40 },
		{ 393, 156, 40 },
		{ 393, 157, 40 },
		{ 393, 158, 40 },
		{ 393, 159, 40 },
		{ 393, 160, 40 },
		{ 393, 161, 40 },
		{ 393, 162, 40 },
		{ 393, 163, 40 },
		{ 393, 164, 40 },
		{ 393, 165, 40 },
		{ 393, 166, 40 },
		{ 393, 167, 40 },
		{ 393, 168, 40 },
		{ 393, 169, 40 },
		{ 393, 170, 40 },
		{ 393, 171, 40 },
		{ 394, 138, 40 },
		{ 394, 139, 40 },
		{ 394, 140, 40 },
		{ 394, 141, 40 },
		{ 394, 142, 40 },
		{ 394, 143, 40 },
		{ 394, 144, 40 },
		{ 394, 145, 40 },
		{ 394, 146, 40 },
		{ 394, 147, 40 },
		{ 394, 148, 40 },
		{ 394, 149, 40 },
		{ 394, 150, 40 },
		{ 394, 151, 40 },
		{ 394, 152, 40 },
		{ 394, 153, 40 },
		{ 394, 154, 40 },
		{ 394, 155, 40 },
		{ 394, 156, 40 },
		{ 394, 157, 40 },
		{ 394, 158, 40 },
		{ 394, 159, 40 },
		{ 394, 160, 40 },
		{ 394, 161, 40 },
		{ 394, 162, 40 },
		{ 394, 163, 40 },
		{ 394, 164, 40 },
		{ 394, 165, 40 },
		{ 394, 166, 40 },
		{ 394, 167, 40 },
		{ 394, 168, 40 },
		{ 394, 169, 40 },
		{ 394, 170, 40 },
		{ 394, 171, 40 },
		{ 395, 137, 40 },
		{ 395, 138, 40 },
		{ 395, 139, 40 },
		{ 395, 140, 40 },
		{ 395, 141, 40 },
		{ 395, 142, 40 },
		{ 395, 143, 40 },
		{ 395, 144, 40 },
		{ 395, 145, 40 },
		{ 395, 146, 40 },
		{ 395, 147, 40 },
		{ 395, 148, 40 },
		{ 395, 149, 40 },
		{ 395, 150, 40 },
		{ 395, 151, 40 },
		{ 395, 152, 40 },
		{ 395, 153, 40 },
		{ 395, 154, 40 },
		{ 395, 155, 40 },
		{ 395, 156, 40 },
		{ 395, 157, 40 },
		{ 395, 158, 40 },
		{ 395, 159, 40 },
		{ 395, 160, 40 },
		{ 395, 161, 40 },
		{ 395, 162, 40 },
		{ 395, 163, 40 },
		{ 395, 164, 40 },
		{ 395, 165, 40 },
		{ 395, 166, 40 },
		{ 395, 167, 40 },
		{ 395, 168, 40 },
		{ 395, 169, 40 },
		{ 395, 170, 40 },
		{ 396, 136, 40 },
		{ 396, 137, 40 },
		{ 396, 138, 40 },
		{ 396, 139, 40 },
		{ 396, 140, 40 },
		{ 396, 141, 40 },
		{ 396, 142, 40 },
		{ 396, 143, 40 },
		{ 396, 144, 40 },
		{ 396, 145, 40 },
		{ 396, 146, 40 },
		{ 396, 147, 40 },
		{ 396, 148, 40 },
		{ 396, 149, 40 },
		{ 396, 150, 40 },
		{ 396, 151, 40 },
		{ 396, 152, 40 },
		{ 396, 153, 40 },
		{ 396, 154, 40 },
		{ 396, 155, 40 },
		{ 396, 156, 40 },
		{ 396, 157, 40 },
		{ 396, 158, 40 },
		{ 396, 159, 40 },
		{ 396, 160, 40 },
		{ 396, 161, 40 },
		{ 396, 162, 40 },
		{ 396, 163, 40 },
		{ 396, 164, 40 },
		{ 396, 165, 40 },
		{ 396, 166, 40 },
		{ 396, 167, 40 },
		{ 396, 168, 40 },
		{ 396, 169, 40 },
		{ 396, 170, 40 },
		{ 397, 135, 40 },
		{ 397, 136, 40 },
		{ 397, 137, 40 },
		{ 397, 138, 40 },
		{ 397, 139, 40 },
		{ 397, 140, 40 },
		{ 397, 141, 40 },
		{ 397, 142, 40 },
		{ 397, 143, 40 },
		{ 397, 144, 40 },
		{ 397, 145, 40 },
		{ 397, 146, 40 },
		{ 397, 147, 40 },
		{ 397, 148, 40 },
		{ 397, 149, 40 },
		{ 397, 150, 40 },
		{ 397, 151, 40 },
		{ 397, 152, 40 },
		{ 397, 153, 40 },
		{ 397, 154, 40 },
		{ 397, 155, 40 },
		{ 397, 156, 40 },
		{ 397, 157, 40 },
		{ 397, 158, 40 },
		{ 397, 159, 40 },
		{ 397, 160, 40 },
		{ 397, 161, 40 },
		{ 397, 162, 40 },
		{ 397, 163, 40 },
		{ 397, 164, 40 },
		{ 397, 165, 40 },
		{ 397, 166, 40 },
		{ 397, 167, 40 },
		{ 397, 168, 40 },
		{ 397, 169, 40 },
		{ 398, 134, 40 },
		{ 398, 135, 40 },
		{ 398, 136, 40 },
		{ 398, 137, 40 },
		{ 398, 138, 40 },
		{ 398, 139, 40 },
		{ 398, 140, 40 },
		{ 398, 141, 40 },
		{ 398, 142, 40 },
		{ 398, 143, 40 },
		{ 398, 144, 40 },
		{ 398, 145, 40 },
		{ 398, 146, 40 },
		{ 398, 147, 40 },
		{ 398, 148, 40 },
		{ 398, 149, 40 },
		{ 398, 150, 40 },
		{ 398, 151, 40 },
		{ 398, 152, 40 },
		{ 398, 153, 40 },
		{ 398, 154, 40 },
		{ 398, 155, 40 },
		{ 398, 156, 40 },
		{ 398, 157, 40 },
		{ 398, 158, 40 },
		{ 398, 159, 40 },
		{ 398, 160, 40 },
		{ 398, 161, 40 },
		{ 398, 162, 40 },
		{ 398, 163, 40 },
		{ 398, 164, 40 },
		{ 398, 165, 40 },
		{ 398, 166, 40 },
		{ 398, 167, 40 },
		{ 398, 168, 40 },
		{ 399, 133, 40 },
		{ 399, 134, 40 },
		{ 399, 135, 40 },
		{ 399, 136, 40 },
		{ 399, 137, 40 },
		{ 399, 138, 40 },
		{ 399, 139, 40 },
		{ 399, 140, 40 },
		{ 399, 141, 40 },
		{ 399, 142, 40 },
		{ 399, 143, 40 },
		{ 399, 144, 40 },
		{ 399, 145, 40 },
		{ 399, 146, 40 },
		{ 399, 147, 40 },
		{ 399, 148, 40 },
		{ 399, 149, 40 },
		{ 399, 150, 40 },
		{ 399, 151, 40 },
		{ 399, 152, 40 },
		{ 399, 153, 40 },
		{ 399, 154, 40 },
		{ 399, 155, 40 },
		{ 399, 156, 40 },
		{ 399, 157, 40 },
		{ 399, 158, 40 },
		{ 399, 159, 40 },
		{ 399, 160, 40 },
		{ 399, 161, 40 },
		{ 399, 162, 40 },
		{ 399, 163, 40 },
		{ 399, 164, 40 },
		{ 399, 165, 40 },
		{ 399, 166, 40 },
		{ 399, 167, 40 },
		{ 399, 168, 40 },
		{ 400, 132, 40 },
		{ 400, 133, 40 },
		{ 400, 134, 40 },
		{ 400, 135, 40 },
		{ 400, 136, 40 },
		{ 400, 137, 40 },
		{ 400, 138, 40 },
		{ 400, 139, 40 },
		{ 400, 140, 40 },
		{ 400, 141, 40 },
		{ 400, 142, 40 },
		{ 400, 143, 40 },
		{ 400, 144, 40 },
		{ 400, 145, 40 },
		{ 400, 146, 40 },
		{ 400, 147, 40 },
		{ 400, 148, 40 },
		{ 400, 149, 40 },
		{ 400, 150, 40 },
		{ 400, 151, 40 },
		{ 400, 152, 40 },
		{ 400, 153, 40 },
		{ 400, 154, 40 },
		{ 400, 155, 40 },
		{ 400, 156, 40 },
		{ 400, 157, 40 },
		{ 400, 158, 40 },
		{ 400, 159, 40 },
		{ 400, 160, 40 },
		{ 400, 161, 40 },
		{ 400, 162, 40 },
		{ 400, 163, 40 },
		{ 400, 164, 40 },
		{ 400, 165, 40 },
		{ 400, 166, 40 },
		{ 400, 167, 40 },
		{ 401, 132, 40 },
		{ 401, 133, 40 },
		{ 401, 134, 40 },
		{ 401, 135, 40 },
		{ 401, 136, 40 },
		{ 401, 137, 40 },
		{ 401, 138, 40 },
		{ 401, 139, 40 },
		{ 401, 140, 40 },
		{ 401, 141, 40 },
		{ 401, 142, 40 },
		{ 401, 143, 40 },
		{ 401, 144, 40 },
		{ 401, 145, 40 },
		{ 401, 146, 40 },
		{ 401, 147, 40 },
		{ 401, 148, 40 },
		{ 401, 149, 40 },
		{ 401, 150, 40 },
		{ 401, 151, 40 },
		{ 401, 152, 40 },
		{ 401, 153, 40 },
		{ 401, 154, 40 },
		{ 401, 155, 40 },
		{ 401, 156, 40 },
		{ 401, 157, 40 },
		{ 401, 158, 40 },
		{ 401, 159, 40 },
		{ 401, 160, 40 },
		{ 401, 161, 40 },
		{ 401, 162, 40 },
		{ 401, 163, 40 },
		{ 401, 164, 40 },
		{ 401, 165, 40 },
		{ 401, 166, 40 },
		{ 402, 131, 40 },
		{ 402, 132, 40 },
		{ 402, 133, 40 },
		{ 402, 134, 40 },
		{ 402, 135, 40 },
		{ 402, 136, 40 },
		{ 402, 137, 40 },
		{ 402, 138, 40 },
		{ 402, 139, 40 },
		{ 402, 140, 40 },
		{ 402, 141, 40 },
		{ 402, 142, 40 },
		{ 402, 143, 40 },
		{ 402, 144, 40 },
		{ 402, 145, 40 },
		{ 402, 146, 40 },
		{ 402, 147, 40 },
		{ 402, 148, 40 },
		{ 402, 149, 40 },
		{ 402, 150, 40 },
		{ 402, 151, 40 },
		{ 402, 152, 40 },
		{ 402, 153, 40 },
		{ 402, 154, 40 },
		{ 402, 155, 40 },
		{ 402, 156, 40 },
		{ 402, 157, 40 },
		{ 402, 158, 40 },
		{ 402, 159, 40 },
		{ 402, 160, 40 },
		{ 402, 161, 40 },
		{ 402, 162, 40 },
		{ 402, 163, 40 },
		{ 402, 164, 40 },
		{ 402, 165, 40 },
		{ 402, 166, 40 },
		{ 403, 130, 40 },
		{ 403, 131, 40 },
		{ 403, 132, 40 },
		{ 403, 133, 40 },
		{ 403, 134, 40 },
		{ 403, 135, 40 },
		{ 403, 136, 40 },
		{ 403, 137, 40 },
		{ 403, 138, 40 },
		{ 403, 139, 40 },
		{ 403, 140, 40 },
		{ 403, 141, 40 },
		{ 403, 142, 40 },
		{ 403, 143, 40 },
		{ 403, 144, 40 },
		{ 403, 145, 40 },
		{ 403, 146, 40 },
		{ 403, 147, 40 },
		{ 403, 148, 40 },
		{ 403, 149, 40 },
		{ 403, 150, 40 },
		{ 403, 151, 40 },
		{ 403, 152, 40 },
		{ 403, 153, 40 },
		{ 403, 154, 40 },
		{ 403, 155, 40 },
		{ 403, 156, 40 },
		{ 403, 157, 40 },
		{ 403, 158, 40 },
		{ 403, 159, 40 },
		{ 403, 160, 40 },
		{ 403, 161, 40 },
		{ 403, 162, 40 },
		{ 403, 163, 40 },
		{ 403, 164, 40 },
		{ 403, 165, 40 },
		{ 404, 129, 40 },
		{ 404, 130, 40 },
		{ 404, 131, 40 },
		{ 404, 132, 40 },
		{ 404, 133, 40 },
		{ 404, 134, 40 },
		{ 404, 135, 40 },
		{ 404, 136, 40 },
		{ 404, 137, 40 },
		{ 404, 138, 40 },
		{ 404, 139, 40 },
		{ 404, 140, 40 },
		{ 404, 141, 40 },
		{ 404, 142, 40 },
		{ 404, 143, 40 },
		{ 404, 144, 40 },
		{ 404, 145, 40 },
		{ 404, 146, 40 },
		{ 404, 147, 40 },
		{ 404, 148, 40 },
		{ 404, 149, 40 },
		{ 404, 150, 40 },
		{ 404, 151, 40 },
		{ 404, 152, 40 },
		{ 404, 153, 40 },
		{ 404, 154, 40 },
		{ 404, 155, 40 },
		{ 404, 156, 40 },
		{ 404, 157, 40 },
		{ 404, 158, 40 },
		{ 404, 159, 40 },
		{ 404, 160, 40 },
		{ 404, 161, 40 },
		{ 404, 162, 40 },
		{ 404, 163, 40 },
		{ 404, 164, 40 },
		{ 404, 165, 40 },
		{ 405, 128, 40 },
		{ 405, 129, 40 },
		{ 405, 130, 40 },
		{ 405, 131, 40 },
		{ 405, 132, 40 },
		{ 405, 133, 40 },
		{ 405, 134, 40 },
		{ 405, 135, 40 },
		{ 405, 136, 40 },
		{ 405, 137, 40 },
		{ 405, 138, 40 },
		{ 405, 139, 40 },
		{ 405, 140, 40 },
		{ 405, 141, 40 },
		{ 405, 142, 40 },
		{ 405, 143, 40 },
		{ 405, 144, 40 },
		{ 405, 145, 40 },
		{ 405, 146, 40 },
		{ 405, 147, 40 },
		{ 405, 148, 40 },
		{ 405, 149, 40 },
		{ 405, 150, 40 },
		{ 405, 151, 40 },
		{ 405, 152, 40 },
		{ 405, 153, 40 },
		{ 405, 154, 40 },
		{ 405, 155, 40 },
		{ 405, 156, 40 },
		{ 405, 157, 40 },
		{ 405, 158, 40 },
		{ 405, 159, 40 },
		{ 405, 160, 40 },
		{ 405, 161, 40 },
		{ 405, 162, 40 },
		{ 405, 163, 40 },
		{ 405, 164, 40 },
		{ 406, 127, 40 },
		{ 406, 128, 40 },
		{ 406, 129, 40 },
		{ 406, 130, 40 },
		{ 406, 131, 40 },
		{ 406, 132, 40 },
		{ 406, 133, 40 },
		{ 406, 134, 40 },
		{ 406, 135, 40 },
		{ 406, 136, 40 },
		{ 406, 137, 40 },
		{ 406, 138, 40 },
		{ 406, 139, 40 },
		{ 406, 140, 40 },
		{ 406, 141, 40 },
		{ 406, 142, 40 },
		{ 406, 143, 40 },
		{ 406, 144, 40 },
		{ 406, 145, 40 },
		{ 406, 146, 40 },
		{ 406, 147, 40 },
		{ 406, 148, 40 },
		{ 406, 149, 40 },
		{ 406, 150, 40 },
		{ 406, 151, 40 },
		{ 406, 152, 40 },
		{ 406, 153, 40 },
		{ 406, 154, 40 },
		{ 406, 155, 40 },
		{ 406, 156, 40 },
		{ 406, 157, 40 },
		{ 406, 158, 40 },
		{ 406, 159, 40 },
		{ 406, 160, 40 },
		{ 406, 161, 40 },
		{ 406, 162, 40 },
		{ 406, 163, 40 },
		{ 407, 126, 40 },
		{ 407, 127, 40 },
		{ 407, 128, 40 },
		{ 407, 129, 40 },
		{ 407, 130, 40 },
		{ 407, 131, 40 },
		{ 407, 132, 40 },
		{ 407, 133, 40 },
		{ 407, 134, 40 },
		{ 407, 135, 40 },
		{ 407, 136, 40 },
		{ 407, 137, 40 },
		{ 407, 138, 40 },
		{ 407, 139, 40 },
		{ 407, 140, 40 },
		{ 407, 141, 40 },
		{ 407, 142, 40 },
		{ 407, 143, 40 },
		{ 407, 144, 40 },
		{ 407, 145, 40 },
		{ 407, 146, 40 },
		{ 407, 147, 40 },
		{ 407, 148, 40 },
		{ 407, 149, 40 },
		{ 407, 150, 40 },
		{ 407, 151, 40 },
		{ 407, 152, 40 },
		{ 407, 153, 40 },
		{ 407, 154, 40 },
		{ 407, 155, 40 },
		{ 407, 156, 40 },
		{ 407, 157, 40 },
		{ 407, 158, 40 },
		{ 407, 159, 40 },
		{ 407, 160, 40 },
		{ 407, 161, 40 },
		{ 407, 162, 40 },
		{ 407, 163, 40 },
		{ 408, 125, 40 },
		{ 408, 126, 40 },
		{ 408, 127, 40 },
		{ 408, 128, 40 },
		{ 408, 129, 40 },
		{ 408, 130, 40 },
		{ 408, 131, 40 },
		{ 408, 132, 40 },
		{ 408, 133, 40 },
		{ 408, 134, 40 },
		{ 408, 135, 40 },
		{ 408, 136, 40 },
		{ 408, 137, 40 },
		{ 408, 138, 40 },
		{ 408, 139, 40 },
		{ 408, 140, 40 },
		{ 408, 141, 40 },
		{ 408, 142, 40 },
		{ 408, 143, 40 },
		{ 408, 144, 40 },
		{ 408, 145, 40 },
		{ 408, 146, 40 },
		{ 408, 147, 40 },
		{ 408, 148, 40 },
		{ 408, 149, 40 },
		{ 408, 150, 40 },
		{ 408, 151, 40 },
		{ 408, 152, 40 },
		{ 408, 153, 40 },
		{ 408, 154, 40 },
		{ 408, 155, 40 },
		{ 408, 156, 40 },
		{ 408, 157, 40 },
		{ 408, 158, 40 },
		{ 408, 159, 40 },
		{ 408, 160, 40 },
		{ 408, 161, 40 },
		{ 408, 162, 40 },
		{ 409, 125, 40 },
		{ 409, 126, 40 },
		{ 409, 127, 40 },
		{ 409, 128, 40 },
		{ 409, 129, 40 },
		{ 409, 130, 40 },
		{ 409, 131, 40 },
		{ 409, 132, 40 },
		{ 409, 133, 40 },
		{ 409, 134, 40 },
		{ 409, 135, 40 },
		{ 409, 136, 40 },
		{ 409, 137, 40 },
		{ 409, 138, 40 },
		{ 409, 139, 40 },
		{ 409, 140, 40 },
		{ 409, 141, 40 },
		{ 409, 142, 40 },
		{ 409, 143, 40 },
		{ 409, 144, 40 },
		{ 409, 145, 40 },
		{ 409, 146, 40 },
		{ 409, 147, 40 },
		{ 409, 148, 40 },
		{ 409, 149, 40 },
		{ 409, 150, 40 },
		{ 409, 151, 40 },
		{ 409, 152, 40 },
		{ 409, 153, 40 },
		{ 409, 154, 40 },
		{ 409, 155, 40 },
		{ 409, 156, 40 },
		{ 409, 157, 40 },
		{ 409, 158, 40 },
		{ 409, 159, 40 },
		{ 409, 160, 40 },
		{ 409, 161, 40 },
		{ 410, 124, 40 },
		{ 410, 125, 40 },
		{ 410, 126, 40 },
		{ 410, 127, 40 },
		{ 410, 128, 40 },
		{ 410, 129, 40 },
		{ 410, 130, 40 },
		{ 410, 131, 40 },
		{ 410, 132, 40 },
		{ 410, 133, 40 },
		{ 410, 134, 40 },
		{ 410, 135, 40 },
		{ 410, 136, 40 },
		{ 410, 137, 40 },
		{ 410, 138, 40 },
		{ 410, 139, 40 },
		{ 410, 140, 40 },
		{ 410, 141, 40 },
		{ 410, 142, 40 },
		{ 410, 143, 40 },
		{ 410, 144, 40 },
		{ 410, 145, 40 },
		{ 410, 146, 40 },
		{ 410, 147, 40 },
		{ 410, 148, 40 },
		{ 410, 149, 40 },
		{ 410, 150, 40 },
		{ 410, 151, 40 },
		{ 410, 152, 40 },
		{ 410, 153, 40 },
		{ 410, 154, 40 },
		{ 410, 155, 40 },
		{ 410, 156, 40 },
		{ 410, 157, 40 },
		{ 410, 158, 40 },
		{ 410, 159, 40 },
		{ 410, 160, 40 },
		{ 410, 161, 40 },
		{ 411, 123, 40 },
		{ 411, 124, 40 },
		{ 411, 125, 40 },
		{ 411, 126, 40 },
		{ 411, 127, 40 },
		{ 411, 128, 40 },
		{ 411, 129, 40 },
		{ 411, 130, 40 },
		{ 411, 131, 40 },
		{ 411, 132, 40 },
		{ 411, 133, 40 },
		{ 411, 134, 40 },
		{ 411, 135, 40 },
		{ 411, 136, 40 },
		{ 411, 137, 40 },
		{ 411, 138, 40 },
		{ 411, 139, 40 },
		{ 411, 140, 40 },
		{ 411, 141, 40 },
		{ 411, 142, 40 },
		{ 411, 143, 40 },
		{ 411, 144, 40 },
		{ 411, 145, 40 },
		{ 411, 146, 40 },
		{ 411, 147, 40 },
		{ 411, 148, 40 },
		{ 411, 149, 40 },
		{ 411, 150, 40 },
		{ 411, 151, 40 },
		{ 411, 152, 40 },
		{ 411, 153, 40 },
		{ 411, 154, 40 },
		{ 411, 155, 40 },
		{ 411, 156, 40 },
		{ 411, 157, 40 },
		{ 411, 158, 40 },
		{ 411, 159, 40 },
		{ 411, 160, 40 },
		{ 412, 122, 40 },
		{ 412, 123, 40 },
		{ 412, 124, 40 },
		{ 412, 125, 40 },
		{ 412, 126, 40 },
		{ 412, 127, 40 },
		{ 412, 128, 40 },
		{ 412, 129, 40 },
		{ 412, 130, 40 },
		{ 412, 131, 40 },
		{ 412, 132, 40 },
		{ 412, 133, 40 },
		{ 412, 134, 40 },
		{ 412, 135, 40 },
		{ 412, 136, 40 },
		{ 412, 137, 40 },
		{ 412, 138, 40 },
		{ 412, 139, 40 },
		{ 412, 140, 40 },
		{ 412, 141, 40 },
		{ 412, 142, 40 },
		{ 412, 143, 40 },
		{ 412, 144, 40 },
		{ 412, 145, 40 },
		{ 412, 146, 40 },
		{ 412, 147, 40 },
		{ 412, 148, 40 },
		{ 412, 149, 40 },
		{ 412, 150, 40 },
		{ 412, 151, 40 },
		{ 412, 152, 40 },
		{ 412, 153, 40 },
		{ 412, 154, 40 },
		{ 412, 155, 40 },
		{ 412, 156, 40 },
		{ 412, 157, 40 },
		{ 412, 158, 40 },
		{ 412, 159, 40 },
		{ 412, 160, 40 },
		{ 413, 121, 40 },
		{ 413, 122, 40 },
		{ 413, 123, 40 },
		{ 413, 124, 40 },
		{ 413, 125, 40 },
		{ 413, 126, 40 },
		{ 413, 127, 40 },
		{ 413, 128, 40 },
		{ 413, 129, 40 },
		{ 413, 130, 40 },
		{ 413, 131, 40 },
		{ 413, 132, 40 },
		{ 413, 133, 40 },
		{ 413, 134, 40 },
		{ 413, 135, 40 },
		{ 413, 136, 40 },
		{ 413, 137, 40 },
		{ 413, 138, 40 },
		{ 413, 139, 40 },
		{ 413, 140, 40 },
		{ 413, 141, 40 },
		{ 413, 142, 40 },
		{ 413, 143, 40 },
		{ 413, 144, 40 },
		{ 413, 145, 40 },
		{ 413, 146, 40 },
		{ 413, 147, 40 },
		{ 413, 148, 40 },
		{ 413, 149, 40 },
		{ 413, 150, 40 },
		{ 413, 151, 40 },
		{ 413, 152, 40 },
		{ 413, 153, 40 },
		{ 413, 154, 40 },
		{ 413, 155, 40 },
		{ 413, 156, 40 },
		{ 413, 157, 40 },
		{ 413, 158, 40 },
		{ 413, 159, 40 },
		{ 414, 120, 40 },
		{ 414, 121, 40 },
		{ 414, 122, 40 },
		{ 414, 123, 40 },
		{ 414, 124, 40 },
		{ 414, 125, 40 },
		{ 414, 126, 40 },
		{ 414, 127, 40 },
		{ 414, 128, 40 },
		{ 414, 129, 40 },
		{ 414, 130, 40 },
		{ 414, 131, 40 },
		{ 414, 132, 40 },
		{ 414, 133, 40 },
		{ 414, 134, 40 },
		{ 414, 135, 40 },
		{ 414, 136, 40 },
		{ 414, 137, 40 },
		{ 414, 138, 40 },
		{ 414, 139, 40 },
		{ 414, 140, 40 },
		{ 414, 141, 40 },
		{ 414, 142, 40 },
		{ 414, 143, 40 },
		{ 414, 144, 40 },
		{ 414, 145, 40 },
		{ 414, 146, 40 },
		{ 414, 147, 40 },
		{ 414, 148, 40 },
		{ 414, 149, 40 },
		{ 414, 150, 40 },
		{ 414, 151, 40 },
		{ 414, 152, 40 },
		{ 414, 153, 40 },
		{ 414, 154, 40 },
		{ 414, 155, 40 },
		{ 414, 156, 40 },
		{ 414, 157, 40 },
		{ 414, 158, 40 },
		{ 415, 119, 40 },
		{ 415, 120, 40 },
		{ 415, 121, 40 },
		{ 415, 122, 40 },
		{ 415, 123, 40 },
		{ 415, 124, 40 },
		{ 415, 125, 40 },
		{ 415, 126, 40 },
		{ 415, 127, 40 },
		{ 415, 128, 40 },
		{ 415, 129, 40 },
		{ 415, 130, 40 },
		{ 415, 131, 40 },
		{ 415, 132, 40 },
		{ 415, 133, 40 },
		{ 415, 134, 40 },
		{ 415, 135, 40 },
		{ 415, 136, 40 },
		{ 415, 137, 40 },
		{ 415, 138, 40 },
		{ 415, 139, 40 },
		{ 415, 140, 40 },
		{ 415, 141, 40 },
		{ 415, 142, 40 },
		{ 415, 143, 40 },
		{ 415, 144, 40 },
		{ 415, 145, 40 },
		{ 415, 146, 40 },
		{ 415, 147, 40 },
		{ 415, 148, 40 },
		{ 415, 149, 40 },
		{ 415, 150, 40 },
		{ 415, 151, 40 },
		{ 415, 152, 40 },
		{ 415, 153, 40 },
		{ 415, 154, 40 },
		{ 415, 155, 40 },
		{ 415, 156, 40 },
		{ 415, 157, 40 },
		{ 415, 158, 40 },
		{ 416, 118, 40 },
		{ 416, 119, 40 },
		{ 416, 120, 40 },
		{ 416, 121, 40 },
		{ 416, 122, 40 },
		{ 416, 123, 40 },
		{ 416, 124, 40 },
		{ 416, 125, 40 },
		{ 416, 126, 40 },
		{ 416, 127, 40 },
		{ 416, 128, 40 },
		{ 416, 129, 40 },
		{ 416, 130, 40 },
		{ 416, 131, 40 },
		{ 416, 132, 40 },
		{ 416, 133, 40 },
		{ 416, 134, 40 },
		{ 416, 135, 40 },
		{ 416, 136, 40 },
		{ 416, 137, 40 },
		{ 416, 138, 40 },
		{ 416, 139, 40 },
		{ 416, 140, 40 },
		{ 416, 141, 40 },
		{ 416, 142, 40 },
		{ 416, 143, 40 },
		{ 416, 144, 40 },
		{ 416, 145, 40 },
		{ 416, 146, 40 },
		{ 416, 147, 40 },
		{ 416, 148, 40 },
		{ 416, 149, 40 },
		{ 416, 150, 40 },
		{ 416, 151, 40 },
		{ 416, 152, 40 },
		{ 416, 153, 40 },
		{ 416, 154, 40 },
		{ 416, 155, 40 },
		{ 416, 156, 40 },
		{ 416, 157, 40 },
		{ 417, 118, 40 },
		{ 417, 119, 40 },
		{ 417, 120, 40 },
		{ 417, 121, 40 },
		{ 417, 122, 40 },
		{ 417, 123, 40 },
		{ 417, 124, 40 },
		{ 417, 125, 40 },
		{ 417, 126, 40 },
		{ 417, 127, 40 },
		{ 417, 128, 40 },
		{ 417, 129, 40 },
		{ 417, 130, 40 },
		{ 417, 131, 40 },
		{ 417, 132, 40 },
		{ 417, 133, 40 },
		{ 417, 134, 40 },
		{ 417, 135, 40 },
		{ 417, 136, 40 },
		{ 417, 137, 40 },
		{ 417, 138, 40 },
		{ 417, 139, 40 },
		{ 417, 140, 40 },
		{ 417, 141, 40 },
		{ 417, 142, 40 },
		{ 417, 143, 40 },
		{ 417, 144, 40 },
		{ 417, 145, 40 },
		{ 417, 146, 40 },
		{ 417, 147, 40 },
		{ 417, 148, 40 },
		{ 417, 149, 40 },
		{ 417, 150, 40 },
		{ 417, 151, 40 },
		{ 417, 152, 40 },
		{ 417, 153, 40 },
		{ 417, 154, 40 },
		{ 417, 155, 40 },
		{ 417, 156, 40 },
		{ 418, 117, 40 },
		{ 418, 118, 40 },
		{ 418, 119, 40 },
		{ 418, 120, 40 },
		{ 418, 121, 40 },
		{ 418, 122, 40 },
		{ 418, 123, 40 },
		{ 418, 124, 40 },
		{ 418, 125, 40 },
		{ 418, 126, 40 },
		{ 418, 127, 40 },
		{ 418, 128, 40 },
		{ 418, 129, 40 },
		{ 418, 130, 40 },
		{ 418, 131, 40 },
		{ 418, 132, 40 },
		{ 418, 133, 40 },
		{ 418, 134, 40 },
		{ 418, 135, 40 },
		{ 418, 136, 40 },
		{ 418, 137, 40 },
		{ 418, 138, 40 },
		{ 418, 139, 40 },
		{ 418, 140, 40 },
		{ 418, 141, 40 },
		{ 418, 142, 40 },
		{ 418, 143, 40 },
		{ 418, 144, 40 },
		{ 418, 145, 40 },
		{ 418, 146, 40 },
		{ 418, 147, 40 },
		{ 418, 148, 40 },
		{ 418, 149, 40 },
		{ 418, 150, 40 },
		{ 418, 151, 40 },
		{ 418, 152, 40 },
		{ 418, 153, 40 },
		{ 418, 154, 40 },
		{ 418, 155, 40 },
		{ 418, 156, 40 },
		{ 419, 116, 40 },
		{ 419, 117, 40 },
		{ 419, 118, 40 },
		{ 419, 119, 40 },
		{ 419, 120, 40 },
		{ 419, 121, 40 },
		{ 419, 122, 40 },
		{ 419, 123, 40 },
		{ 419, 124, 40 },
		{ 419, 125, 40 },
		{ 419, 126, 40 },
		{ 419, 127, 40 },
		{ 419, 128, 40 },
		{ 419, 129, 40 },
		{ 419, 130, 40 },
		{ 419, 131, 40 },
		{ 419, 132, 40 },
		{ 419, 133, 40 },
		{ 419, 134, 40 },
		{ 419, 135, 40 },
		{ 419, 136, 40 },
		{ 419, 137, 40 },
		{ 419, 138, 40 },
		{ 419, 139, 40 },
		{ 419, 140, 40 },
		{ 419, 141, 40 },
		{ 419, 142, 40 },
		{ 419, 143, 40 },
		{ 419, 144, 40 },
		{ 419, 145, 40 },
		{ 419, 146, 40 },
		{ 419, 147, 40 },
		{ 419, 148, 40 },
		{ 419, 149, 40 },
		{ 419, 150, 40 },
		{ 419, 151, 40 },
		{ 419, 152, 40 },
		{ 419, 153, 40 },
		{ 419, 154, 40 },
		{ 419, 155, 40 },
		{ 420, 115, 40 },
		{ 420, 116, 40 },
		{ 420, 117, 40 },
		{ 420, 118, 40 },
		{ 420, 119, 40 },
		{ 420, 120, 40 },
		{ 420, 121, 40 },
		{ 420, 122, 40 },
		{ 420, 123, 40 },
		{ 420, 124, 40 },
		{ 420, 125, 40 },
		{ 420, 126, 40 },
		{ 420, 127, 40 },
		{ 420, 128, 40 },
		{ 420, 129, 40 },
		{ 420, 130, 40 },
		{ 420, 131, 40 },
		{ 420, 132, 40 },
		{ 420, 133, 40 },
		{ 420, 134, 40 },
		{ 420, 135, 40 },
		{ 420, 136, 40 },
		{ 420, 137, 40 },
		{ 420, 138, 40 },
		{ 420, 139, 40 },
		{ 420, 140, 40 },
		{ 420, 141, 40 },
		{ 420, 142, 40 },
		{ 420, 143, 40 },
		{ 420, 144, 40 },
		{ 420, 145, 40 },
		{ 420, 146, 40 },
		{ 420, 147, 40 },
		{ 420, 148, 40 },
		{ 420, 149, 40 },
		{ 420, 150, 40 },
		{ 420, 151, 40 },
		{ 420, 152, 40 },
		{ 420, 153, 40 },
		{ 420, 154, 40 },
		{ 420, 155, 40 },
		{ 421, 114, 40 },
		{ 421, 115, 40 },
		{ 421, 116, 40 },
		{ 421, 117, 40 },
		{ 421, 118, 40 },
		{ 421, 119, 40 },
		{ 421, 120, 40 },
		{ 421, 121, 40 },
		{ 421, 122, 40 },
		{ 421, 123, 40 },
		{ 421, 124, 40 },
		{ 421, 125, 40 },
		{ 421, 126, 40 },
		{ 421, 127, 40 },
		{ 421, 128, 40 },
		{ 421, 129, 40 },
		{ 421, 130, 40 },
		{ 421, 131, 40 },
		{ 421, 132, 40 },
		{ 421, 133, 40 },
		{ 421, 134, 40 },
		{ 421, 135, 40 },
		{ 421, 136, 40 },
		{ 421, 137, 40 },
		{ 421, 138, 40 },
		{ 421, 139, 40 },
		{ 421, 140, 40 },
		{ 421, 141, 40 },
		{ 421, 142, 40 },
		{ 421, 143, 40 },
		{ 421, 144, 40 },
		{ 421, 145, 40 },
		{ 421, 146, 40 },
		{ 421, 147, 40 },
		{ 421, 148, 40 },
		{ 421, 149, 40 },
		{ 421, 150, 40 },
		{ 421, 151, 40 },
		{ 421, 152, 40 },
		{ 421, 153, 40 },
		{ 421, 154, 40 },
		{ 422, 113, 40 },
		{ 422, 114, 40 },
		{ 422, 115, 40 },
		{ 422, 116, 40 },
		{ 422, 117, 40 },
		{ 422, 118, 40 },
		{ 422, 119, 40 },
		{ 422, 120, 40 },
		{ 422, 121, 40 },
		{ 422, 122, 40 },
		{ 422, 123, 40 },
		{ 422, 124, 40 },
		{ 422, 125, 40 },
		{ 422, 126, 40 },
		{ 422, 127, 40 },
		{ 422, 128, 40 },
		{ 422, 129, 40 },
		{ 422, 130, 40 },
		{ 422, 131, 40 },
		{ 422, 132, 40 },
		{ 422, 133, 40 },
		{ 422, 134, 40 },
		{ 422, 135, 40 },
		{ 422, 136, 40 },
		{ 422, 137, 40 },
		{ 422, 138, 40 },
		{ 422, 139, 40 },
		{ 422, 140, 40 },
		{ 422, 141, 40 },
		{ 422, 142, 40 },
		{ 422, 143, 40 },
		{ 422, 144, 40 },
		{ 422, 145, 40 },
		{ 422, 146, 40 },
		{ 422, 147, 40 },
		{ 422, 148, 40 },
		{ 422, 149, 40 },
		{ 422, 150, 40 },
		{ 422, 151, 40 },
		{ 422, 152, 40 },
		{ 422, 153, 40 },
		{ 423, 112, 40 },
		{ 423, 113, 40 },
		{ 423, 114, 40 },
		{ 423, 115, 40 },
		{ 423, 116, 40 },
		{ 423, 117, 40 },
		{ 423, 118, 40 },
		{ 423, 119, 40 },
		{ 423, 120, 40 },
		{ 423, 121, 40 },
		{ 423, 122, 40 },
		{ 423, 123, 40 },
		{ 423, 124, 40 },
		{ 423, 125, 40 },
		{ 423, 126, 40 },
		{ 423, 127, 40 },
		{ 423, 128, 40 },
		{ 423, 129, 40 },
		{ 423, 130, 40 },
		{ 423, 131, 40 },
		{ 423, 132, 40 },
		{ 423, 133, 40 },
		{ 423, 134, 40 },
		{ 423, 135, 40 },
		{ 423, 136, 40 },
		{ 423, 137, 40 },
		{ 423, 138, 40 },
		{ 423, 139, 40 },
		{ 423, 140, 40 },
		{ 423, 141, 40 },
		{ 423, 142, 40 },
		{ 423, 143, 40 },
		{ 423, 144, 40 },
		{ 423, 145, 40 },
		{ 423, 146, 40 },
		{ 423, 147, 40 },
		{ 423, 148, 40 },
		{ 423, 149, 40 },
		{ 423, 150, 40 },
		{ 423, 151, 40 },
		{ 423, 152, 40 },
		{ 423, 153, 40 },
		{ 424, 111, 40 },
		{ 424, 112, 40 },
		{ 424, 113, 40 },
		{ 424, 114, 40 },
		{ 424, 115, 40 },
		{ 424, 116, 40 },
		{ 424, 117, 40 },
		{ 424, 118, 40 },
		{ 424, 119, 40 },
		{ 424, 120, 40 },
		{ 424, 121, 40 },
		{ 424, 122, 40 },
		{ 424, 123, 40 },
		{ 424, 124, 40 },
		{ 424, 125, 40 },
		{ 424, 126, 40 },
		{ 424, 127, 40 },
		{ 424, 128, 40 },
		{ 424, 129, 40 },
		{ 424, 130, 40 },
		{ 424, 131, 40 },
		{ 424, 132, 40 },
		{ 424, 133, 40 },
		{ 424, 134, 40 },
		{ 424, 135, 40 },
		{ 424, 136, 40 },
		{ 424, 137, 40 },
		{ 424, 138, 40 },
		{ 424, 139, 40 },
		{ 424, 140, 40 },
		{ 424, 141, 40 },
		{ 424, 142, 40 },
		{ 424, 143, 40 },
		{ 424, 144, 40 },
		{ 424, 145, 40 },
		{ 424, 146, 40 },
		{ 424, 147, 40 },
		{ 424, 148, 40 },
		{ 424, 149, 40 },
		{ 424, 150, 40 },
		{ 424, 151, 40 },
		{ 424, 152, 40 },
		{ 425, 111, 40 },
		{ 425, 112, 40 },
		{ 425, 113, 40 },
		{ 425, 114, 40 },
		{ 425, 115, 40 },
		{ 425, 116, 40 },
		{ 425, 117, 40 },
		{ 425, 118, 40 },
		{ 425, 119, 40 },
		{ 425, 120, 40 },
		{ 425, 121, 40 },
		{ 425, 122, 40 },
		{ 425, 123, 40 },
		{ 425, 124, 40 },
		{ 425, 125, 40 },
		{ 425, 126, 40 },
		{ 425, 127, 40 },
		{ 425, 128, 40 },
		{ 425, 129, 40 },
		{ 425, 130, 40 },
		{ 425, 131, 40 },
		{ 425, 132, 40 },
		{ 425, 133, 40 },
		{ 425, 134, 40 },
		{ 425, 135, 40 },
		{ 425, 136, 40 },
		{ 425, 137, 40 },
		{ 425, 138, 40 },
		{ 425, 139, 40 },
		{ 425, 140, 40 },
		{ 425, 141, 40 },
		{ 425, 142, 40 },
		{ 425, 143, 40 },
		{ 425, 144, 40 },
		{ 425, 145, 40 },
		{ 425, 146, 40 },
		{ 425, 147, 40 },
		{ 425, 148, 40 },
		{ 425, 149, 40 },
		{ 425, 150, 40 },
		{ 425, 151, 40 },
		{ 426, 110, 40 },
		{ 426, 111, 40 },
		{ 426, 112, 40 },
		{ 426, 113, 40 },
		{ 426, 114, 40 },
		{ 426, 115, 40 },
		{ 426, 116, 40 },
		{ 426, 117, 40 },
		{ 426, 118, 40 },
		{ 426, 119, 40 },
		{ 426, 120, 40 },
		{ 426, 121, 40 },
		{ 426, 122, 40 },
		{ 426, 123, 40 },
		{ 426, 124, 40 },
		{ 426, 125, 40 },
		{ 426, 126, 40 },
		{ 426, 127, 40 },
		{ 426, 128, 40 },
		{ 426, 129, 40 },
		{ 426, 130, 40 },
		{ 426, 131, 40 },
		{ 426, 132, 40 },
		{ 426, 133, 40 },
		{ 426, 134, 40 },
		{ 426, 135, 40 },
		{ 426, 136, 40 },
		{ 426, 137, 40 },
		{ 426, 138, 40 },
		{ 426, 139, 40 },
		{ 426, 140, 40 },
		{ 426, 141, 40 },
		{ 426, 142, 40 },
		{ 426, 143, 40 },
		{ 426, 144, 40 },
		{ 426, 145, 40 },
		{ 426, 146, 40 },
		{ 426, 147, 40 },
		{ 426, 148, 40 },
		{ 426, 149, 40 },
		{ 426, 150, 40 },
		{ 426, 151, 40 },
		{ 427, 109, 40 },
		{ 427, 110, 40 },
		{ 427, 111, 40 },
		{ 427, 112, 40 },
		{ 427, 113, 40 },
		{ 427, 114, 40 },
		{ 427, 115, 40 },
		{ 427, 116, 40 },
		{ 427, 117, 40 },
		{ 427, 118, 40 },
		{ 427, 119, 40 },
		{ 427, 120, 40 },
		{ 427, 121, 40 },
		{ 427, 122, 40 },
		{ 427, 123, 40 },
		{ 427, 124, 40 },
		{ 427, 125, 40 },
		{ 427, 126, 40 },
		{ 427, 127, 40 },
		{ 427, 128, 40 },
		{ 427, 129, 40 },
		{ 427, 130, 40 },
		{ 427, 131, 40 },
		{ 427, 132, 40 },
		{ 427, 133, 40 },
		{ 427, 134, 40 },
		{ 427, 135, 40 },
		{ 427, 136, 40 },
		{ 427, 137, 40 },
		{ 427, 138, 40 },
		{ 427, 139, 40 },
		{ 427, 140, 40 },
		{ 427, 141, 40 },
		{ 427, 142, 40 },
		{ 427, 143, 40 },
		{ 427, 144, 40 },
		{ 427, 145, 40 },
		{ 427, 146, 40 },
		{ 427, 147, 40 },
		{ 427, 148, 40 },
		{ 427, 149, 40 },
		{ 427, 150, 40 },
		{ 428, 108, 40 },
		{ 428, 109, 40 },
		{ 428, 110, 40 },
		{ 428, 111, 40 },
		{ 428, 112, 40 },
		{ 428, 113, 40 },
		{ 428, 114, 40 },
		{ 428, 115, 40 },
		{ 428, 116, 40 },
		{ 428, 117, 40 },
		{ 428, 118, 40 },
		{ 428, 119, 40 },
		{ 428, 120, 40 },
		{ 428, 121, 40 },
		{ 428, 122, 40 },
		{ 428, 123, 40 },
		{ 428, 124, 40 },
		{ 428, 125, 40 },
		{ 428, 126, 40 },
		{ 428, 127, 40 },
		{ 428, 128, 40 },
		{ 428, 129, 40 },
		{ 428, 130, 40 },
		{ 428, 131, 40 },
		{ 428, 132, 40 },
		{ 428, 133, 40 },
		{ 428, 134, 40 },
		{ 428, 135, 40 },
		{ 428, 136, 40 },
		{ 428, 137, 40 },
		{ 428, 138, 40 },
		{ 428, 139, 40 },
		{ 428, 140, 40 },
		{ 428, 141, 40 },
		{ 428, 142, 40 },
		{ 428, 143, 40 },
		{ 428, 144, 40 },
		{ 428, 145, 40 },
		{ 428, 146, 40 },
		{ 428, 147, 40 },
		{ 428, 148, 40 },
		{ 428, 149, 40 },
		{ 428, 150, 40 },
		{ 429, 107, 40 },
		{ 429, 108, 40 },
		{ 429, 109, 40 },
		{ 429, 110, 40 },
		{ 429, 111, 40 },
		{ 429, 112, 40 },
		{ 429, 113, 40 },
		{ 429, 114, 40 },
		{ 429, 115, 40 },
		{ 429, 116, 40 },
		{ 429, 117, 40 },
		{ 429, 118, 40 },
		{ 429, 119, 40 },
		{ 429, 120, 40 },
		{ 429, 121, 40 },
		{ 429, 122, 40 },
		{ 429, 123, 40 },
		{ 429, 124, 40 },
		{ 429, 125, 40 },
		{ 429, 126, 40 },
		{ 429, 127, 40 },
		{ 429, 128, 40 },
		{ 429, 129, 40 },
		{ 429, 130, 40 },
		{ 429, 131, 40 },
		{ 429, 132, 40 },
		{ 429, 133, 40 },
		{ 429, 134, 40 },
		{ 429, 135, 40 },
		{ 429, 136, 40 },
		{ 429, 137, 40 },
		{ 429, 138, 40 },
		{ 429, 139, 40 },
		{ 429, 140, 40 },
		{ 429, 141, 40 },
		{ 429, 142, 40 },
		{ 429, 143, 40 },
		{ 429, 144, 40 },
		{ 429, 145, 40 },
		{ 429, 146, 40 },
		{ 429, 147, 40 },
		{ 429, 148, 40 },
		{ 429, 149, 40 },
		{ 430, 106, 40 },
		{ 430, 107, 40 },
		{ 430, 108, 40 },
		{ 430, 109, 40 },
		{ 430, 110, 40 },
		{ 430, 111, 40 },
		{ 430, 112, 40 },
		{ 430, 113, 40 },
		{ 430, 114, 40 },
		{ 430, 115, 40 },
		{ 430, 116, 40 },
		{ 430, 117, 40 },
		{ 430, 118, 40 },
		{ 430, 119, 40 },
		{ 430, 120, 40 },
		{ 430, 121, 40 },
		{ 430, 122, 40 },
		{ 430, 123, 40 },
		{ 430, 124, 40 },
		{ 430, 125, 40 },
		{ 430, 126, 40 },
		{ 430, 127, 40 },
		{ 430, 128, 40 },
		{ 430, 129, 40 },
		{ 430, 130, 40 },
		{ 430, 131, 40 },
		{ 430, 132, 40 },
		{ 430, 133, 40 },
		{ 430, 134, 40 },
		{ 430, 135, 40 },
		{ 430, 136, 40 },
		{ 430, 137, 40 },
		{ 430, 138, 40 },
		{ 430, 139, 40 },
		{ 430, 140, 40 },
		{ 430, 141, 40 },
		{ 430, 142, 40 },
		{ 430, 143, 40 },
		{ 430, 144, 40 },
		{ 430, 145, 40 },
		{ 430, 146, 40 },
		{ 430, 147, 40 },
		{ 430, 148, 40 },
		{ 431, 105, 40 },
		{ 431, 106, 40 },
		{ 431, 107, 40 },
		{ 431, 108, 40 },
		{ 431, 109, 40 },
		{ 431, 110, 40 },
		{ 431, 111, 40 },
		{ 431, 112, 40 },
		{ 431, 113, 40 },
		{ 431, 114, 40 },
		{ 431, 115, 40 },
		{ 431, 116, 40 },
		{ 431, 117, 40 },
		{ 431, 118, 40 },
		{ 431, 119, 40 },
		{ 431, 120, 40 },
		{ 431, 121, 40 },
		{ 431, 122, 40 },
		{ 431, 123, 40 },
		{ 431, 124, 40 },
		{ 431, 125, 40 },
		{ 431, 126, 40 },
		{ 431, 127, 40 },
		{ 431, 128, 40 },
		{ 431, 129, 40 },
		{ 431, 130, 40 },
		{ 431, 131, 40 },
		{ 431, 132, 40 },
		{ 431, 133, 40 },
		{ 431, 134, 40 },
		{ 431, 135, 40 },
		{ 431, 136, 40 },
		{ 431, 137, 40 },
		{ 431, 138, 40 },
		{ 431, 139, 40 },
		{ 431, 140, 40 },
		{ 431, 141, 40 },
		{ 431, 142, 40 },
		{ 431, 143, 40 },
		{ 431, 144, 40 },
		{ 431, 145, 40 },
		{ 431, 146, 40 },
		{ 431, 147, 40 },
		{ 431, 148, 40 },
		{ 432, 104, 40 },
		{ 432, 105, 40 },
		{ 432, 106, 40 },
		{ 432, 107, 40 },
		{ 432, 108, 40 },
		{ 432, 109, 40 },
		{ 432, 110, 40 },
		{ 432, 111, 40 },
		{ 432, 112, 40 },
		{ 432, 113, 40 },
		{ 432, 114, 40 },
		{ 432, 115, 40 },
		{ 432, 116, 40 },
		{ 432, 117, 40 },
		{ 432, 118, 40 },
		{ 432, 119, 40 },
		{ 432, 120, 40 },
		{ 432, 121, 40 },
		{ 432, 122, 40 },
		{ 432, 123, 40 },
		{ 432, 124, 40 },
		{ 432, 125, 40 },
		{ 432, 126, 40 },
		{ 432, 127, 40 },
		{ 432, 128, 40 },
		{ 432, 129, 40 },
		{ 432, 130, 40 },
		{ 432, 131, 40 },
		{ 432, 132, 40 },
		{ 432, 133, 40 },
		{ 432, 134, 40 },
		{ 432, 135, 40 },
		{ 432, 136, 40 },
		{ 432, 137, 40 },
		{ 432, 138, 40 },
		{ 432, 139, 40 },
		{ 432, 140, 40 },
		{ 432, 141, 40 },
		{ 432, 142, 40 },
		{ 432, 143, 40 },
		{ 432, 144, 40 },
		{ 432, 145, 40 },
		{ 432, 146, 40 },
		{ 432, 147, 40 },
		{ 433, 104, 40 },
		{ 433, 105, 40 },
		{ 433, 106, 40 },
		{ 433, 107, 40 },
		{ 433, 108, 40 },
		{ 433, 109, 40 },
		{ 433, 110, 40 },
		{ 433, 111, 40 },
		{ 433, 112, 40 },
		{ 433, 113, 40 },
		{ 433, 114, 40 },
		{ 433, 115, 40 },
		{ 433, 116, 40 },
		{ 433, 117, 40 },
		{ 433, 118, 40 },
		{ 433, 119, 40 },
		{ 433, 120, 40 },
		{ 433, 121, 40 },
		{ 433, 122, 40 },
		{ 433, 123, 40 },
		{ 433, 124, 40 },
		{ 433, 125, 40 },
		{ 433, 126, 40 },
		{ 433, 127, 40 },
		{ 433, 128, 40 },
		{ 433, 129, 40 },
		{ 433, 130, 40 },
		{ 433, 131, 40 },
		{ 433, 132, 40 },
		{ 433, 133, 40 },
		{ 433, 134, 40 },
		{ 433, 135, 40 },
		{ 433, 136, 40 },
		{ 433, 137, 40 },
		{ 433, 138, 40 },
		{ 433, 139, 40 },
		{ 433, 140, 40 },
		{ 433, 141, 40 },
		{ 433, 142, 40 },
		{ 433, 143, 40 },
		{ 433, 144, 40 },
		{ 433, 145, 40 },
		{ 433, 146, 40 },
		{ 434, 103, 40 },
		{ 434, 104, 40 },
		{ 434, 105, 40 },
		{ 434, 106, 40 },
		{ 434, 107, 40 },
		{ 434, 108, 40 },
		{ 434, 109, 40 },
		{ 434, 110, 40 },
		{ 434, 111, 40 },
		{ 434, 112, 40 },
		{ 434, 113, 40 },
		{ 434, 114, 40 },
		{ 434, 115, 40 },
		{ 434, 116, 40 },
		{ 434, 117, 40 },
		{ 434, 118, 40 },
		{ 434, 119, 40 },
		{ 434, 120, 40 },
		{ 434, 121, 40 },
		{ 434, 122, 40 },
		{ 434, 123, 40 },
		{ 434, 124, 40 },
		{ 434, 125, 40 },
		{ 434, 126, 40 },
		{ 434, 127, 40 },
		{ 434, 128, 40 },
		{ 434, 129, 40 },
		{ 434, 130, 40 },
		{ 434, 131, 40 },
		{ 434, 132, 40 },
		{ 434, 133, 40 },
		{ 434, 134, 40 },
		{ 434, 135, 40 },
		{ 434, 136, 40 },
		{ 434, 137, 40 },
		{ 434, 138, 40 },
		{ 434, 139, 40 },
		{ 434, 140, 40 },
		{ 434, 141, 40 },
		{ 434, 142, 40 },
		{ 434, 143, 40 },
		{ 434, 144, 40 },
		{ 434, 145, 40 },
		{ 434, 146, 40 },
		{ 435, 102, 40 },
		{ 435, 103, 40 },
		{ 435, 104, 40 },
		{ 435, 105, 40 },
		{ 435, 106, 40 },
		{ 435, 107, 40 },
		{ 435, 108, 40 },
		{ 435, 109, 40 },
		{ 435, 110, 40 },
		{ 435, 111, 40 },
		{ 435, 112, 40 },
		{ 435, 113, 40 },
		{ 435, 114, 40 },
		{ 435, 115, 40 },
		{ 435, 116, 40 },
		{ 435, 117, 40 },
		{ 435, 118, 40 },
		{ 435, 119, 40 },
		{ 435, 120, 40 },
		{ 435, 121, 40 },
		{ 435, 122, 40 },
		{ 435, 123, 40 },
		{ 435, 124, 40 },
		{ 435, 125, 40 },
		{ 435, 126, 40 },
		{ 435, 127, 40 },
		{ 435, 128, 40 },
		{ 435, 129, 40 },
		{ 435, 130, 40 },
		{ 435, 131, 40 },
		{ 435, 132, 40 },
		{ 435, 133, 40 },
		{ 435, 134, 40 },
		{ 435, 135, 40 },
		{ 435, 136, 40 },
		{ 435, 137, 40 },
		{ 435, 138, 40 },
		{ 435, 139, 40 },
		{ 435, 140, 40 },
		{ 435, 141, 40 },
		{ 435, 142, 40 },
		{ 435, 143, 40 },
		{ 435, 144, 40 },
		{ 435, 145, 40 },
		{ 436, 101, 40 },
		{ 436, 102, 40 },
		{ 436, 103, 40 },
		{ 436, 104, 40 },
		{ 436, 105, 40 },
		{ 436, 106, 40 },
		{ 436, 107, 40 },
		{ 436, 108, 40 },
		{ 436, 109, 40 },
		{ 436, 110, 40 },
		{ 436, 111, 40 },
		{ 436, 112, 40 },
		{ 436, 113, 40 },
		{ 436, 114, 40 },
		{ 436, 115, 40 },
		{ 436, 116, 40 },
		{ 436, 117, 40 },
		{ 436, 118, 40 },
		{ 436, 119, 40 },
		{ 436, 120, 40 },
		{ 436, 121, 40 },
		{ 436, 122, 40 },
		{ 436, 123, 40 },
		{ 436, 124, 40 },
		{ 436, 125, 40 },
		{ 436, 126, 40 },
		{ 436, 127, 40 },
		{ 436, 128, 40 },
		{ 436, 129, 40 },
		{ 436, 130, 40 },
		{ 436, 131, 40 },
		{ 436, 132, 40 },
		{ 436, 133, 40 },
		{ 436, 134, 40 },
		{ 436, 135, 40 },
		{ 436, 136, 40 },
		{ 436, 137, 40 },
		{ 436, 138, 40 },
		{ 436, 139, 40 },
		{ 436, 140, 40 },
		{ 436, 141, 40 },
		{ 436, 142, 40 },
		{ 436, 143, 40 },
		{ 436, 144, 40 },
		{ 436, 145, 40 },
		{ 437, 100, 40 },
		{ 437, 101, 40 },
		{ 437, 102, 40 },
		{ 437, 103, 40 },
		{ 437, 104, 40 },
		{ 437, 105, 40 },
		{ 437, 106, 40 },
		{ 437, 107, 40 },
		{ 437, 108, 40 },
		{ 437, 109, 40 },
		{ 437, 110, 40 },
		{ 437, 111, 40 },
		{ 437, 112, 40 },
		{ 437, 113, 40 },
		{ 437, 114, 40 },
		{ 437, 115, 40 },
		{ 437, 116, 40 },
		{ 437, 117, 40 },
		{ 437, 118, 40 },
		{ 437, 119, 40 },
		{ 437, 120, 40 },
		{ 437, 121, 40 },
		{ 437, 122, 40 },
		{ 437, 123, 40 },
		{ 437, 124, 40 },
		{ 437, 125, 40 },
		{ 437, 126, 40 },
		{ 437, 127, 40 },
		{ 437, 128, 40 },
		{ 437, 129, 40 },
		{ 437, 130, 40 },
		{ 437, 131, 40 },
		{ 437, 132, 40 },
		{ 437, 133, 40 },
		{ 437, 134, 40 },
		{ 437, 135, 40 },
		{ 437, 136, 40 },
		{ 437, 137, 40 },
		{ 437, 138, 40 },
		{ 437, 139, 40 },
		{ 437, 140, 40 },
		{ 437, 141, 40 },
		{ 437, 142, 40 },
		{ 437, 143, 40 },
		{ 437, 144, 40 },
		{ 438, 99, 40 },
		{ 438, 100, 40 },
		{ 438, 101, 40 },
		{ 438, 102, 40 },
		{ 438, 103, 40 },
		{ 438, 104, 40 },
		{ 438, 105, 40 },
		{ 438, 106, 40 },
		{ 438, 107, 40 },
		{ 438, 108, 40 },
		{ 438, 109, 40 },
		{ 438, 110, 40 },
		{ 438, 111, 40 },
		{ 438, 112, 40 },
		{ 438, 113, 40 },
		{ 438, 114, 40 },
		{ 438, 115, 40 },
		{ 438, 116, 40 },
		{ 438, 117, 40 },
		{ 438, 118, 40 },
		{ 438, 119, 40 },
		{ 438, 120, 40 },
		{ 438, 121, 40 },
		{ 438, 122, 40 },
		{ 438, 123, 40 },
		{ 438, 124, 40 },
		{ 438, 125, 40 },
		{ 438, 126, 40 },
		{ 438, 127, 40 },
		{ 438, 128, 40 },
		{ 438, 129, 40 },
		{ 438, 130, 40 },
		{ 438, 131, 40 },
		{ 438, 132, 40 },
		{ 438, 133, 40 },
		{ 438, 134, 40 },
		{ 438, 135, 40 },
		{ 438, 136, 40 },
		{ 438, 137, 40 },
		{ 438, 138, 40 },
		{ 438, 139, 40 },
		{ 438, 140, 40 },
		{ 438, 141, 40 },
		{ 438, 142, 40 },
		{ 438, 143, 40 },
		{ 439, 98, 40 },
		{ 439, 99, 40 },
		{ 439, 100, 40 },
		{ 439, 101, 40 },
		{ 439, 102, 40 },
		{ 439, 103, 40 },
		{ 439, 104, 40 },
		{ 439, 105, 40 },
		{ 439, 106, 40 },
		{ 439, 107, 40 },
		{ 439, 108, 40 },
		{ 439, 109, 40 },
		{ 439, 110, 40 },
		{ 439, 111, 40 },
		{ 439, 112, 40 },
		{ 439, 113, 40 },
		{ 439, 114, 40 },
		{ 439, 115, 40 },
		{ 439, 116, 40 },
		{ 439, 117, 40 },
		{ 439, 118, 40 },
		{ 439, 119, 40 },
		{ 439, 120, 40 },
		{ 439, 121, 40 },
		{ 439, 122, 40 },
		{ 439, 123, 40 },
		{ 439, 124, 40 },
		{ 439, 125, 40 },
		{ 439, 126, 40 },
		{ 439, 127, 40 },
		{ 439, 128, 40 },
		{ 439, 129, 40 },
		{ 439, 130, 40 },
		{ 439, 131, 40 },
		{ 439, 132, 40 },
		{ 439, 133, 40 },
		{ 439, 134, 40 },
		{ 439, 135, 40 },
		{ 439, 136, 40 },
		{ 439, 137, 40 },
		{ 439, 138, 40 },
		{ 439, 139, 40 },
		{ 439, 140, 40 },
		{ 439, 141, 40 },
		{ 439, 142, 40 },
		{ 439, 143, 40 },
		{ 440, 97, 40 },
		{ 440, 98, 40 },
		{ 440, 99, 40 },
		{ 440, 100, 40 },
		{ 440, 101, 40 },
		{ 440, 102, 40 },
		{ 440, 103, 40 },
		{ 440, 104, 40 },
		{ 440, 105, 40 },
		{ 440, 106, 40 },
		{ 440, 107, 40 },
		{ 440, 108, 40 },
		{ 440, 109, 40 },
		{ 440, 110, 40 },
		{ 440, 111, 40 },
		{ 440, 112, 40 },
		{ 440, 113, 40 },
		{ 440, 114, 40 },
		{ 440, 115, 40 },
		{ 440, 116, 40 },
		{ 440, 117, 40 },
		{ 440, 118, 40 },
		{ 440, 119, 40 },
		{ 440, 120, 40 },
		{ 440, 121, 40 },
		{ 440, 122, 40 },
		{ 440, 123, 40 },
		{ 440, 124, 40 },
		{ 440, 125, 40 },
		{ 440, 126, 40 },
		{ 440, 127, 40 },
		{ 440, 128, 40 },
		{ 440, 129, 40 },
		{ 440, 130, 40 },
		{ 440, 131, 40 },
		{ 440, 132, 40 },
		{ 440, 133, 40 },
		{ 440, 134, 40 },
		{ 440, 135, 40 },
		{ 440, 136, 40 },
		{ 440, 137, 40 },
		{ 440, 138, 40 },
		{ 440, 139, 40 },
		{ 440, 140, 40 },
		{ 440, 141, 40 },
		{ 440, 142, 40 },
		{ 441, 97, 40 },
		{ 441, 98, 40 },
		{ 441, 99, 40 },
		{ 441, 100, 40 },
		{ 441, 101, 40 },
		{ 441, 102, 40 },
		{ 441, 103, 40 },
		{ 441, 104, 40 },
		{ 441, 105, 40 },
		{ 441, 106, 40 },
		{ 441, 107, 40 },
		{ 441, 108, 40 },
		{ 441, 109, 40 },
		{ 441, 110, 40 },
		{ 441, 111, 40 },
		{ 441, 112, 40 },
		{ 441, 113, 40 },
		{ 441, 114, 40 },
		{ 441, 115, 40 },
		{ 441, 116, 40 },
		{ 441, 117, 40 },
		{ 441, 118, 40 },
		{ 441, 119, 40 },
		{ 441, 120, 40 },
		{ 441, 121, 40 },
		{ 441, 122, 40 },
		{ 441, 123, 40 },
		{ 441, 124, 40 },
		{ 441, 125, 40 },
		{ 441, 126, 40 },
		{ 441, 127, 40 },
		{ 441, 128, 40 },
		{ 441, 129, 40 },
		{ 441, 130, 40 },
		{ 441, 131, 40 },
		{ 441, 132, 40 },
		{ 441, 133, 40 },
		{ 441, 134, 40 },
		{ 441, 135, 40 },
		{ 441, 136, 40 },
		{ 441, 137, 40 },
		{ 441, 138, 40 },
		{ 441, 139, 40 },
		{ 441, 140, 40 },
		{ 441, 141, 40 },
		{ 442, 96, 40 },
		{ 442, 97, 40 },
		{ 442, 98, 40 },
		{ 442, 99, 40 },
		{ 442, 100, 40 },
		{ 442, 101, 40 },
		{ 442, 102, 40 },
		{ 442, 103, 40 },
		{ 442, 104, 40 },
		{ 442, 105, 40 },
		{ 442, 106, 40 },
		{ 442, 107, 40 },
		{ 442, 108, 40 },
		{ 442, 109, 40 },
		{ 442, 110, 40 },
		{ 442, 111, 40 },
		{ 442, 112, 40 },
		{ 442, 113, 40 },
		{ 442, 114, 40 },
		{ 442, 115, 40 },
		{ 442, 116, 40 },
		{ 442, 117, 40 },
		{ 442, 118, 40 },
		{ 442, 119, 40 },
		{ 442, 120, 40 },
		{ 442, 121, 40 },
		{ 442, 122, 40 },
		{ 442, 123, 40 },
		{ 442, 124, 40 },
		{ 442, 125, 40 },
		{ 442, 126, 40 },
		{ 442, 127, 40 },
		{ 442, 128, 40 },
		{ 442, 129, 40 },
		{ 442, 130, 40 },
		{ 442, 131, 40 },
		{ 442, 132, 40 },
		{ 442, 133, 40 },
		{ 442, 134, 40 },
		{ 442, 135, 40 },
		{ 442, 136, 40 },
		{ 442, 137, 40 },
		{ 442, 138, 40 },
		{ 442, 139, 40 },
		{ 442, 140, 40 },
		{ 442, 141, 40 },
		{ 443, 95, 40 },
		{ 443, 96, 40 },
		{ 443, 97, 40 },
		{ 443, 98, 40 },
		{ 443, 99, 40 },
		{ 443, 100, 40 },
		{ 443, 101, 40 },
		{ 443, 102, 40 },
		{ 443, 103, 40 },
		{ 443, 104, 40 },
		{ 443, 105, 40 },
		{ 443, 106, 40 },
		{ 443, 107, 40 },
		{ 443, 108, 40 },
		{ 443, 109, 40 },
		{ 443, 110, 40 },
		{ 443, 111, 40 },
		{ 443, 112, 40 },
		{ 443, 113, 40 },
		{ 443, 114, 40 },
		{ 443, 115, 40 },
		{ 443, 116, 40 },
		{ 443, 117, 40 },
		{ 443, 118, 40 },
		{ 443, 119, 40 },
		{ 443, 120, 40 },
		{ 443, 121, 40 },
		{ 443, 122, 40 },
		{ 443, 123, 40 },
		{ 443, 124, 40 },
		{ 443, 125, 40 },
		{ 443, 126, 40 },
		{ 443, 127, 40 },
		{ 443, 128, 40 },
		{ 443, 129, 40 },
		{ 443, 130, 40 },
		{ 443, 131, 40 },
		{ 443, 132, 40 },
		{ 443, 133, 40 },
		{ 443, 134, 40 },
		{ 443, 135, 40 },
		{ 443, 136, 40 },
		{ 443, 137, 40 },
		{ 443, 138, 40 },
		{ 443, 139, 40 },
		{ 443, 140, 40 },
		{ 444, 94, 40 },
		{ 444, 95, 40 },
		{ 444, 96, 40 },
		{ 444, 97, 40 },
		{ 444, 98, 40 },
		{ 444, 99, 40 },
		{ 444, 100, 40 },
		{ 444, 101, 40 },
		{ 444, 102, 40 },
		{ 444, 103, 40 },
		{ 444, 104, 40 },
		{ 444, 105, 40 },
		{ 444, 106, 40 },
		{ 444, 107, 40 },
		{ 444, 108, 40 },
		{ 444, 109, 40 },
		{ 444, 110, 40 },
		{ 444, 111, 40 },
		{ 444, 112, 40 },
		{ 444, 113, 40 },
		{ 444, 114, 40 },
		{ 444, 115, 40 },
		{ 444, 116, 40 },
		{ 444, 117, 40 },
		{ 444, 118, 40 },
		{ 444, 119, 40 },
		{ 444, 120, 40 },
		{ 444, 121, 40 },
		{ 444, 122, 40 },
		{ 444, 123, 40 },
		{ 444, 124, 40 },
		{ 444, 125, 40 },
		{ 444, 126, 40 },
		{ 444, 127, 40 },
		{ 444, 128, 40 },
		{ 444, 129, 40 },
		{ 444, 130, 40 },
		{ 444, 131, 40 },
		{ 444, 132, 40 },
		{ 444, 133, 40 },
		{ 444, 134, 40 },
		{ 444, 135, 40 },
		{ 444, 136, 40 },
		{ 444, 137, 40 },
		{ 444, 138, 40 },
		{ 444, 139, 40 },
		{ 444, 140, 40 },
		{ 445, 93, 40 },
		{ 445, 94, 40 },
		{ 445, 95, 40 },
		{ 445, 96, 40 },
		{ 445, 97, 40 },
		{ 445, 98, 40 },
		{ 445, 99, 40 },
		{ 445, 100, 40 },
		{ 445, 101, 40 },
		{ 445, 102, 40 },
		{ 445, 103, 40 },
		{ 445, 104, 40 },
		{ 445, 105, 40 },
		{ 445, 106, 40 },
		{ 445, 107, 40 },
		{ 445, 108, 40 },
		{ 445, 109, 40 },
		{ 445, 110, 40 },
		{ 445, 111, 40 },
		{ 445, 112, 40 },
		{ 445, 113, 40 },
		{ 445, 114, 40 },
		{ 445, 115, 40 },
		{ 445, 116, 40 },
		{ 445, 117, 40 },
		{ 445, 118, 40 },
		{ 445, 119, 40 },
		{ 445, 120, 40 },
		{ 445, 121, 40 },
		{ 445, 122, 40 },
		{ 445, 123, 40 },
		{ 445, 124, 40 },
		{ 445, 125, 40 },
		{ 445, 126, 40 },
		{ 445, 127, 40 },
		{ 445, 128, 40 },
		{ 445, 129, 40 },
		{ 445, 130, 40 },
		{ 445, 131, 40 },
		{ 445, 132, 40 },
		{ 445, 133, 40 },
		{ 445, 134, 40 },
		{ 445, 135, 40 },
		{ 445, 136, 40 },
		{ 445, 137, 40 },
		{ 445, 138, 40 },
		{ 445, 139, 40 },
		{ 446, 92, 40 },
		{ 446, 93, 40 },
		{ 446, 94, 40 },
		{ 446, 95, 40 },
		{ 446, 96, 40 },
		{ 446, 97, 40 },
		{ 446, 98, 40 },
		{ 446, 99, 40 },
		{ 446, 100, 40 },
		{ 446, 101, 40 },
		{ 446, 102, 40 },
		{ 446, 103, 40 },
		{ 446, 104, 40 },
		{ 446, 105, 40 },
		{ 446, 106, 40 },
		{ 446, 107, 40 },
		{ 446, 108, 40 },
		{ 446, 109, 40 },
		{ 446, 110, 40 },
		{ 446, 111, 40 },
		{ 446, 112, 40 },
		{ 446, 113, 40 },
		{ 446, 114, 40 },
		{ 446, 115, 40 },
		{ 446, 116, 40 },
		{ 446, 117, 40 },
		{ 446, 118, 40 },
		{ 446, 119, 40 },
		{ 446, 120, 40 },
		{ 446, 121, 40 },
		{ 446, 122, 40 },
		{ 446, 123, 40 },
		{ 446, 124, 40 },
		{ 446, 125, 40 },
		{ 446, 126, 40 },
		{ 446, 127, 40 },
		{ 446, 128, 40 },
		{ 446, 129, 40 },
		{ 446, 130, 40 },
		{ 446, 131, 40 },
		{ 446, 132, 40 },
		{ 446, 133, 40 },
		{ 446, 134, 40 },
		{ 446, 135, 40 },
		{ 446, 136, 40 },
		{ 446, 137, 40 },
		{ 446, 138, 40 },
		{ 447, 91, 40 },
		{ 447, 92, 40 },
		{ 447, 93, 40 },
		{ 447, 94, 40 },
		{ 447, 95, 40 },
		{ 447, 96, 40 },
		{ 447, 97, 40 },
		{ 447, 98, 40 },
		{ 447, 99, 40 },
		{ 447, 100, 40 },
		{ 447, 101, 40 },
		{ 447, 102, 40 },
		{ 447, 103, 40 },
		{ 447, 104, 40 },
		{ 447, 105, 40 },
		{ 447, 106, 40 },
		{ 447, 107, 40 },
		{ 447, 108, 40 },
		{ 447, 109, 40 },
		{ 447, 110, 40 },
		{ 447, 111, 40 },
		{ 447, 112, 40 },
		{ 447, 113, 40 },
		{ 447, 114, 40 },
		{ 447, 115, 40 },
		{ 447, 116, 40 },
		{ 447, 117, 40 },
		{ 447, 118, 40 },
		{ 447, 119, 40 },
		{ 447, 120, 40 },
		{ 447, 121, 40 },
		{ 447, 122, 40 },
		{ 447, 123, 40 },
		{ 447, 124, 40 },
		{ 447, 125, 40 },
		{ 447, 126, 40 },
		{ 447, 127, 40 },
		{ 447, 128, 40 },
		{ 447, 129, 40 },
		{ 447, 130, 40 },
		{ 447, 131, 40 },
		{ 447, 132, 40 },
		{ 447, 133, 40 },
		{ 447, 134, 40 },
		{ 447, 135, 40 },
		{ 447, 136, 40 },
		{ 447, 137, 40 },
		{ 447, 138, 40 },
		{ 448, 90, 40 },
		{ 448, 91, 40 },
		{ 448, 92, 40 },
		{ 448, 93, 40 },
		{ 448, 94, 40 },
		{ 448, 95, 40 },
		{ 448, 96, 40 },
		{ 448, 97, 40 },
		{ 448, 98, 40 },
		{ 448, 99, 40 },
		{ 448, 100, 40 },
		{ 448, 101, 40 },
		{ 448, 102, 40 },
		{ 448, 103, 40 },
		{ 448, 104, 40 },
		{ 448, 105, 40 },
		{ 448, 106, 40 },
		{ 448, 107, 40 },
		{ 448, 108, 40 },
		{ 448, 109, 40 },
		{ 448, 110, 40 },
		{ 448, 111, 40 },
		{ 448, 112, 40 },
		{ 448, 113, 40 },
		{ 448, 114, 40 },
		{ 448, 115, 40 },
		{ 448, 116, 40 },
		{ 448, 117, 40 },
		{ 448, 118, 40 },
		{ 448, 119, 40 },
		{ 448, 120, 40 },
		{ 448, 121, 40 },
		{ 448, 122, 40 },
		{ 448, 123, 40 },
		{ 448, 124, 40 },
		{ 448, 125, 40 },
		{ 448, 126, 40 },
		{ 448, 127, 40 },
		{ 448, 128, 40 },
		{ 448, 129, 40 },
		{ 448, 130, 40 },
		{ 448, 131, 40 },
		{ 448, 132, 40 },
		{ 448, 133, 40 },
		{ 448, 134, 40 },
		{ 448, 135, 40 },
		{ 448, 136, 40 },
		{ 448, 137, 40 },
		{ 449, 90, 40 },
		{ 449, 91, 40 },
		{ 449, 92, 40 },
		{ 449, 93, 40 },
		{ 449, 94, 40 },
		{ 449, 95, 40 },
		{ 449, 96, 40 },
		{ 449, 97, 40 },
		{ 449, 98, 40 },
		{ 449, 99, 40 },
		{ 449, 100, 40 },
		{ 449, 101, 40 },
		{ 449, 102, 40 },
		{ 449, 103, 40 },
		{ 449, 104, 40 },
		{ 449, 105, 40 },
		{ 449, 106, 40 },
		{ 449, 107, 40 },
		{ 449, 108, 40 },
		{ 449, 109, 40 },
		{ 449, 110, 40 },
		{ 449, 111, 40 },
		{ 449, 112, 40 },
		{ 449, 113, 40 },
		{ 449, 114, 40 },
		{ 449, 115, 40 },
		{ 449, 116, 40 },
		{ 449, 117, 40 },
		{ 449, 118, 40 },
		{ 449, 119, 40 },
		{ 449, 120, 40 },
		{ 449, 121, 40 },
		{ 449, 122, 40 },
		{ 449, 123, 40 },
		{ 449, 124, 40 },
		{ 449, 125, 40 },
		{ 449, 126, 40 },
		{ 449, 127, 40 },
		{ 449, 128, 40 },
		{ 449, 129, 40 },
		{ 449, 130, 40 },
		{ 449, 131, 40 },
		{ 449, 132, 40 },
		{ 449, 133, 40 },
		{ 449, 134, 40 },
		{ 449, 135, 40 },
		{ 449, 136, 40 },
		{ 450, 89, 40 },
		{ 450, 90, 40 },
		{ 450, 91, 40 },
		{ 450, 92, 40 },
		{ 450, 93, 40 },
		{ 450, 94, 40 },
		{ 450, 95, 40 },
		{ 450, 96, 40 },
		{ 450, 97, 40 },
		{ 450, 98, 40 },
		{ 450, 99, 40 },
		{ 450, 100, 40 },
		{ 450, 101, 40 },
		{ 450, 102, 40 },
		{ 450, 103, 40 },
		{ 450, 104, 40 },
		{ 450, 105, 40 },
		{ 450, 106, 40 },
		{ 450, 107, 40 },
		{ 450, 108, 40 },
		{ 450, 109, 40 },
		{ 450, 110, 40 },
		{ 450, 111, 40 },
		{ 450, 112, 40 },
		{ 450, 113, 40 },
		{ 450, 114, 40 },
		{ 450, 115, 40 },
		{ 450, 116, 40 },
		{ 450, 117, 40 },
		{ 450, 118, 40 },
		{ 450, 119, 40 },
		{ 450, 120, 40 },
		{ 450, 121, 40 },
		{ 450, 122, 40 },
		{ 450, 123, 40 },
		{ 450, 124, 40 },
		{ 450, 125, 40 },
		{ 450, 126, 40 },
		{ 450, 127, 40 },
		{ 450, 128, 40 },
		{ 450, 129, 40 },
		{ 450, 130, 40 },
		{ 450, 131, 40 },
		{ 450, 132, 40 },
		{ 450, 133, 40 },
		{ 450, 134, 40 },
		{ 450, 135, 40 },
		{ 450, 136, 40 },
		{ 451, 88, 40 },
		{ 451, 89, 40 },
		{ 451, 90, 40 },
		{ 451, 91, 40 },
		{ 451, 92, 40 },
		{ 451, 93, 40 },
		{ 451, 94, 40 },
		{ 451, 95, 40 },
		{ 451, 96, 40 },
		{ 451, 97, 40 },
		{ 451, 98, 40 },
		{ 451, 99, 40 },
		{ 451, 100, 40 },
		{ 451, 101, 40 },
		{ 451, 102, 40 },
		{ 451, 103, 40 },
		{ 451, 104, 40 },
		{ 451, 105, 40 },
		{ 451, 106, 40 },
		{ 451, 107, 40 },
		{ 451, 108, 40 },
		{ 451, 109, 40 },
		{ 451, 110, 40 },
		{ 451, 111, 40 },
		{ 451, 112, 40 },
		{ 451, 113, 40 },
		{ 451, 114, 40 },
		{ 451, 115, 40 },
		{ 451, 116, 40 },
		{ 451, 117, 40 },
		{ 451, 118, 40 },
		{ 451, 119, 40 },
		{ 451, 120, 40 },
		{ 451, 121, 40 },
		{ 451, 122, 40 },
		{ 451, 123, 40 },
		{ 451, 124, 40 },
		{ 451, 125, 40 },
		{ 451, 126, 40 },
		{ 451, 127, 40 },
		{ 451, 128, 40 },
		{ 451, 129, 40 },
		{ 451, 130, 40 },
		{ 451, 131, 40 },
		{ 451, 132, 40 },
		{ 451, 133, 40 },
		{ 451, 134, 40 },
		{ 451, 135, 40 },
		{ 452, 87, 40 },
		{ 452, 88, 40 },
		{ 452, 89, 40 },
		{ 452, 90, 40 },
		{ 452, 91, 40 },
		{ 452, 92, 40 },
		{ 452, 93, 40 },
		{ 452, 94, 40 },
		{ 452, 95, 40 },
		{ 452, 96, 40 },
		{ 452, 97, 40 },
		{ 452, 98, 40 },
		{ 452, 99, 40 },
		{ 452, 100, 40 },
		{ 452, 101, 40 },
		{ 452, 102, 40 },
		{ 452, 103, 40 },
		{ 452, 104, 40 },
		{ 452, 105, 40 },
		{ 452, 106, 40 },
		{ 452, 107, 40 },
		{ 452, 108, 40 },
		{ 452, 109, 40 },
		{ 452, 110, 40 },
		{ 452, 111, 40 },
		{ 452, 112, 40 },
		{ 452, 113, 40 },
		{ 452, 114, 40 },
		{ 452, 115, 40 },
		{ 452, 116, 40 },
		{ 452, 117, 40 },
		{ 452, 118, 40 },
		{ 452, 119, 40 },
		{ 452, 120, 40 },
		{ 452, 121, 40 },
		{ 452, 122, 40 },
		{ 452, 123, 40 },
		{ 452, 124, 40 },
		{ 452, 125, 40 },
		{ 452, 126, 40 },
		{ 452, 127, 40 },
		{ 452, 128, 40 },
		{ 452, 129, 40 },
		{ 452, 130, 40 },
		{ 452, 131, 40 },
		{ 452, 132, 40 },
		{ 452, 133, 40 },
		{ 452, 134, 40 },
		{ 452, 135, 40 },
		{ 453, 86, 40 },
		{ 453, 87, 40 },
		{ 453, 88, 40 },
		{ 453, 89, 40 },
		{ 453, 90, 40 },
		{ 453, 91, 40 },
		{ 453, 92, 40 },
		{ 453, 93, 40 },
		{ 453, 94, 40 },
		{ 453, 95, 40 },
		{ 453, 96, 40 },
		{ 453, 97, 40 },
		{ 453, 98, 40 },
		{ 453, 99, 40 },
		{ 453, 100, 40 },
		{ 453, 101, 40 },
		{ 453, 102, 40 },
		{ 453, 103, 40 },
		{ 453, 104, 40 },
		{ 453, 105, 40 },
		{ 453, 106, 40 },
		{ 453, 107, 40 },
		{ 453, 108, 40 },
		{ 453, 109, 40 },
		{ 453, 110, 40 },
		{ 453, 111, 40 },
		{ 453, 112, 40 },
		{ 453, 113, 40 },
		{ 453, 114, 40 },
		{ 453, 115, 40 },
		{ 453, 116, 40 },
		{ 453, 117, 40 },
		{ 453, 118, 40 },
		{ 453, 119, 40 },
		{ 453, 120, 40 },
		{ 453, 121, 40 },
		{ 453, 122, 40 },
		{ 453, 123, 40 },
		{ 453, 124, 40 },
		{ 453, 125, 40 },
		{ 453, 126, 40 },
		{ 453, 127, 40 },
		{ 453, 128, 40 },
		{ 453, 129, 40 },
		{ 453, 130, 40 },
		{ 453, 131, 40 },
		{ 453, 132, 40 },
		{ 453, 133, 40 },
		{ 453, 134, 40 },
		{ 454, 85, 40 },
		{ 454, 86, 40 },
		{ 454, 87, 40 },
		{ 454, 88, 40 },
		{ 454, 89, 40 },
		{ 454, 90, 40 },
		{ 454, 91, 40 },
		{ 454, 92, 40 },
		{ 454, 93, 40 },
		{ 454, 94, 40 },
		{ 454, 95, 40 },
		{ 454, 96, 40 },
		{ 454, 97, 40 },
		{ 454, 98, 40 },
		{ 454, 99, 40 },
		{ 454, 100, 40 },
		{ 454, 101, 40 },
		{ 454, 102, 40 },
		{ 454, 103, 40 },
		{ 454, 104, 40 },
		{ 454, 105, 40 },
		{ 454, 106, 40 },
		{ 454, 107, 40 },
		{ 454, 108, 40 },
		{ 454, 109, 40 },
		{ 454, 110, 40 },
		{ 454, 111, 40 },
		{ 454, 112, 40 },
		{ 454, 113, 40 },
		{ 454, 114, 40 },
		{ 454, 115, 40 },
		{ 454, 116, 40 },
		{ 454, 117, 40 },
		{ 454, 118, 40 },
		{ 454, 119, 40 },
		{ 454, 120, 40 },
		{ 454, 121, 40 },
		{ 454, 122, 40 },
		{ 454, 123, 40 },
		{ 454, 124, 40 },
		{ 454, 125, 40 },
		{ 454, 126, 40 },
		{ 454, 127, 40 },
		{ 454, 128, 40 },
		{ 454, 129, 40 },
		{ 454, 130, 40 },
		{ 454, 131, 40 },
		{ 454, 132, 40 },
		{ 454, 133, 40 },
		{ 455, 84, 40 },
		{ 455, 85, 40 },
		{ 455, 86, 40 },
		{ 455, 87, 40 },
		{ 455, 88, 40 },
		{ 455, 89, 40 },
		{ 455, 90, 40 },
		{ 455, 91, 40 },
		{ 455, 92, 40 },
		{ 455, 93, 40 },
		{ 455, 94, 40 },
		{ 455, 95, 40 },
		{ 455, 96, 40 },
		{ 455, 97, 40 },
		{ 455, 98, 40 },
		{ 455, 99, 40 },
		{ 455, 100, 40 },
		{ 455, 101, 40 },
		{ 455, 102, 40 },
		{ 455, 103, 40 },
		{ 455, 104, 40 },
		{ 455, 105, 40 },
		{ 455, 106, 40 },
		{ 455, 107, 40 },
		{ 455, 108, 40 },
		{ 455, 109, 40 },
		{ 455, 110, 40 },
		{ 455, 111, 40 },
		{ 455, 112, 40 },
		{ 455, 113, 40 },
		{ 455, 114, 40 },
		{ 455, 115, 40 },
		{ 455, 116, 40 },
		{ 455, 117, 40 },
		{ 455, 118, 40 },
		{ 455, 119, 40 },
		{ 455, 120, 40 },
		{ 455, 121, 40 },
		{ 455, 122, 40 },
		{ 455, 123, 40 },
		{ 455, 124, 40 },
		{ 455, 125, 40 },
		{ 455, 126, 40 },
		{ 455, 127, 40 },
		{ 455, 128, 40 },
		{ 455, 129, 40 },
		{ 455, 130, 40 },
		{ 455, 131, 40 },
		{ 455, 132, 40 },
		{ 455, 133, 40 },
		{ 456, 83, 40 },
		{ 456, 84, 40 },
		{ 456, 85, 40 },
		{ 456, 86, 40 },
		{ 456, 87, 40 },
		{ 456, 88, 40 },
		{ 456, 89, 40 },
		{ 456, 90, 40 },
		{ 456, 91, 40 },
		{ 456, 92, 40 },
		{ 456, 93, 40 },
		{ 456, 94, 40 },
		{ 456, 95, 40 },
		{ 456, 96, 40 },
		{ 456, 97, 40 },
		{ 456, 98, 40 },
		{ 456, 99, 40 },
		{ 456, 100, 40 },
		{ 456, 101, 40 },
		{ 456, 102, 40 },
		{ 456, 103, 40 },
		{ 456, 104, 40 },
		{ 456, 105, 40 },
		{ 456, 106, 40 },
		{ 456, 107, 40 },
		{ 456, 108, 40 },
		{ 456, 109, 40 },
		{ 456, 110, 40 },
		{ 456, 111, 40 },
		{ 456, 112, 40 },
		{ 456, 113, 40 },
		{ 456, 114, 40 },
		{ 456, 115, 40 },
		{ 456, 116, 40 },
		{ 456, 117, 40 },
		{ 456, 118, 40 },
		{ 456, 119, 40 },
		{ 456, 120, 40 },
		{ 456, 121, 40 },
		{ 456, 122, 40 },
		{ 456, 123, 40 },
		{ 456, 124, 40 },
		{ 456, 125, 40 },
		{ 456, 126, 40 },
		{ 456, 127, 40 },
		{ 456, 128, 40 },
		{ 456, 129, 40 },
		{ 456, 130, 40 },
		{ 456, 131, 40 },
		{ 456, 132, 40 },
		{ 457, 83, 40 },
		{ 457, 84, 40 },
		{ 457, 85, 40 },
		{ 457, 86, 40 },
		{ 457, 87, 40 },
		{ 457, 88, 40 },
		{ 457, 89, 40 },
		{ 457, 90, 40 },
		{ 457, 91, 40 },
		{ 457, 92, 40 },
		{ 457, 93, 40 },
		{ 457, 94, 40 },
		{ 457, 95, 40 },
		{ 457, 96, 40 },
		{ 457, 97, 40 },
		{ 457, 98, 40 },
		{ 457, 99, 40 },
		{ 457, 100, 40 },
		{ 457, 101, 40 },
		{ 457, 102, 40 },
		{ 457, 103, 40 },
		{ 457, 104, 40 },
		{ 457, 105, 40 },
		{ 457, 106, 40 },
		{ 457, 107, 40 },
		{ 457, 108, 40 },
		{ 457, 109, 40 },
		{ 457, 110, 40 },
		{ 457, 111, 40 },
		{ 457, 112, 40 },
		{ 457, 113, 40 },
		{ 457, 114, 40 },
		{ 457, 115, 40 },
		{ 457, 116, 40 },
		{ 457, 117, 40 },
		{ 457, 118, 40 },
		{ 457, 119, 40 },
		{ 457, 120, 40 },
		{ 457, 121, 40 },
		{ 457, 122, 40 },
		{ 457, 123, 40 },
		{ 457, 124, 40 },
		{ 457, 125, 40 },
		{ 457, 126, 40 },
		{ 457, 127, 40 },
		{ 457, 128, 40 },
		{ 457, 129, 40 },
		{ 457, 130, 40 },
		{ 457, 131, 40 },
		{ 458, 82, 40 },
		{ 458, 83, 40 },
		{ 458, 84, 40 },
		{ 458, 85, 40 },
		{ 458, 86, 40 },
		{ 458, 87, 40 },
		{ 458, 88, 40 },
		{ 458, 89, 40 },
		{ 458, 90, 40 },
		{ 458, 91, 40 },
		{ 458, 92, 40 },
		{ 458, 93, 40 },
		{ 458, 94, 40 },
		{ 458, 95, 40 },
		{ 458, 96, 40 },
		{ 458, 97, 40 },
		{ 458, 98, 40 },
		{ 458, 99, 40 },
		{ 458, 100, 40 },
		{ 458, 101, 40 },
		{ 458, 102, 40 },
		{ 458, 103, 40 },
		{ 458, 104, 40 },
		{ 458, 105, 40 },
		{ 458, 106, 40 },
		{ 458, 107, 40 },
		{ 458, 108, 40 },
		{ 458, 109, 40 },
		{ 458, 110, 40 },
		{ 458, 111, 40 },
		{ 458, 112, 40 },
		{ 458, 113, 40 },
		{ 458, 114, 40 },
		{ 458, 115, 40 },
		{ 458, 116, 40 },
		{ 458, 117, 40 },
		{ 458, 118, 40 },
		{ 458, 119, 40 },
		{ 458, 120, 40 },
		{ 458, 121, 40 },
		{ 458, 122, 40 },
		{ 458, 123, 40 },
		{ 458, 124, 40 },
		{ 458, 125, 40 },
		{ 458, 126, 40 },
		{ 458, 127, 40 },
		{ 458, 128, 40 },
		{ 458, 129, 40 },
		{ 458, 130, 40 },
		{ 458, 131, 40 },
		{ 459, 81, 40 },
		{ 459, 82, 40 },
		{ 459, 83, 40 },
		{ 459, 84, 40 },
		{ 459, 85, 40 },
		{ 459, 86, 40 },
		{ 459, 87, 40 },
		{ 459, 88, 40 },
		{ 459, 89, 40 },
		{ 459, 90, 40 },
		{ 459, 91, 40 },
		{ 459, 92, 40 },
		{ 459, 93, 40 },
		{ 459, 94, 40 },
		{ 459, 95, 40 },
		{ 459, 96, 40 },
		{ 459, 97, 40 },
		{ 459, 98, 40 },
		{ 459, 99, 40 },
		{ 459, 100, 40 },
		{ 459, 101, 40 },
		{ 459, 102, 40 },
		{ 459, 103, 40 },
		{ 459, 104, 40 },
		{ 459, 105, 40 },
		{ 459, 106, 40 },
		{ 459, 107, 40 },
		{ 459, 108, 40 },
		{ 459, 109, 40 },
		{ 459, 110, 40 },
		{ 459, 111, 40 },
		{ 459, 112, 40 },
		{ 459, 113, 40 },
		{ 459, 114, 40 },
		{ 459, 115, 40 },
		{ 459, 116, 40 },
		{ 459, 117, 40 },
		{ 459, 118, 40 },
		{ 459, 119, 40 },
		{ 459, 120, 40 },
		{ 459, 121, 40 },
		{ 459, 122, 40 },
		{ 459, 123, 40 },
		{ 459, 124, 40 },
		{ 459, 125, 40 },
		{ 459, 126, 40 },
		{ 459, 127, 40 },
		{ 459, 128, 40 },
		{ 459, 129, 40 },
		{ 459, 130, 40 },
		{ 460, 80, 40 },
		{ 460, 81, 40 },
		{ 460, 82, 40 },
		{ 460, 83, 40 },
		{ 460, 84, 40 },
		{ 460, 85, 40 },
		{ 460, 86, 40 },
		{ 460, 87, 40 },
		{ 460, 88, 40 },
		{ 460, 89, 40 },
		{ 460, 90, 40 },
		{ 460, 91, 40 },
		{ 460, 92, 40 },
		{ 460, 93, 40 },
		{ 460, 94, 40 },
		{ 460, 95, 40 },
		{ 460, 96, 40 },
		{ 460, 97, 40 },
		{ 460, 98, 40 },
		{ 460, 99, 40 },
		{ 460, 100, 40 },
		{ 460, 101, 40 },
		{ 460, 102, 40 },
		{ 460, 103, 40 },
		{ 460, 104, 40 },
		{ 460, 105, 40 },
		{ 460, 106, 40 },
		{ 460, 107, 40 },
		{ 460, 108, 40 },
		{ 460, 109, 40 },
		{ 460, 110, 40 },
		{ 460, 111, 40 },
		{ 460, 112, 40 },
		{ 460, 113, 40 },
		{ 460, 114, 40 },
		{ 460, 115, 40 },
		{ 460, 116, 40 },
		{ 460, 117, 40 },
		{ 460, 118, 40 },
		{ 460, 119, 40 },
		{ 460, 120, 40 },
		{ 460, 121, 40 },
		{ 460, 122, 40 },
		{ 460, 123, 40 },
		{ 460, 124, 40 },
		{ 460, 125, 40 },
		{ 460, 126, 40 },
		{ 460, 127, 40 },
		{ 460, 128, 40 },
		{ 460, 129, 40 },
		{ 460, 130, 40 },
		{ 461, 79, 40 },
		{ 461, 80, 40 },
		{ 461, 81, 40 },
		{ 461, 82, 40 },
		{ 461, 83, 40 },
		{ 461, 84, 40 },
		{ 461, 85, 40 },
		{ 461, 86, 40 },
		{ 461, 87, 40 },
		{ 461, 88, 40 },
		{ 461, 89, 40 },
		{ 461, 90, 40 },
		{ 461, 91, 40 },
		{ 461, 92, 40 },
		{ 461, 93, 40 },
		{ 461, 94, 40 },
		{ 461, 95, 40 },
		{ 461, 96, 40 },
		{ 461, 97, 40 },
		{ 461, 98, 40 },
		{ 461, 99, 40 },
		{ 461, 100, 40 },
		{ 461, 101, 40 },
		{ 461, 102, 40 },
		{ 461, 103, 40 },
		{ 461, 104, 40 },
		{ 461, 105, 40 },
		{ 461, 106, 40 },
		{ 461, 107, 40 },
		{ 461, 108, 40 },
		{ 461, 109, 40 },
		{ 461, 110, 40 },
		{ 461, 111, 40 },
		{ 461, 112, 40 },
		{ 461, 113, 40 },
		{ 461, 114, 40 },
		{ 461, 115, 40 },
		{ 461, 116, 40 },
		{ 461, 117, 40 },
		{ 461, 118, 40 },
		{ 461, 119, 40 },
		{ 461, 120, 40 },
		{ 461, 121, 40 },
		{ 461, 122, 40 },
		{ 461, 123, 40 },
		{ 461, 124, 40 },
		{ 461, 125, 40 },
		{ 461, 126, 40 },
		{ 461, 127, 40 },
		{ 461, 128, 40 },
		{ 461, 129, 40 },
		{ 462, 78, 40 },
		{ 462, 79, 40 },
		{ 462, 80, 40 },
		{ 462, 81, 40 },
		{ 462, 82, 40 },
		{ 462, 83, 40 },
		{ 462, 84, 40 },
		{ 462, 85, 40 },
		{ 462, 86, 40 },
		{ 462, 87, 40 },
		{ 462, 88, 40 },
		{ 462, 89, 40 },
		{ 462, 90, 40 },
		{ 462, 91, 40 },
		{ 462, 92, 40 },
		{ 462, 93, 40 },
		{ 462, 94, 40 },
		{ 462, 95, 40 },
		{ 462, 96, 40 },
		{ 462, 97, 40 },
		{ 462, 98, 40 },
		{ 462, 99, 40 },
		{ 462, 100, 40 },
		{ 462, 101, 40 },
		{ 462, 102, 40 },
		{ 462, 103, 40 },
		{ 462, 104, 40 },
		{ 462, 105, 40 },
		{ 462, 106, 40 },
		{ 462, 107, 40 },
		{ 462, 108, 40 },
		{ 462, 109, 40 },
		{ 462, 110, 40 },
		{ 462, 111, 40 },
		{ 462, 112, 40 },
		{ 462, 113, 40 },
		{ 462, 114, 40 },
		{ 462, 115, 40 },
		{ 462, 116, 40 },
		{ 462, 117, 40 },
		{ 462, 118, 40 },
		{ 462, 119, 40 },
		{ 462, 120, 40 },
		{ 462, 121, 40 },
		{ 462, 122, 40 },
		{ 462, 123, 40 },
		{ 462, 124, 40 },
		{ 462, 125, 40 },
		{ 462, 126, 40 },
		{ 462, 127, 40 },
		{ 462, 128, 40 },
		{ 463, 77, 40 },
		{ 463, 78, 40 },
		{ 463, 79, 40 },
		{ 463, 80, 40 },
		{ 463, 81, 40 },
		{ 463, 82, 40 },
		{ 463, 83, 40 },
		{ 463, 84, 40 },
		{ 463, 85, 40 },
		{ 463, 86, 40 },
		{ 463, 87, 40 },
		{ 463, 88, 40 },
		{ 463, 89, 40 },
		{ 463, 90, 40 },
		{ 463, 91, 40 },
		{ 463, 92, 40 },
		{ 463, 93, 40 },
		{ 463, 94, 40 },
		{ 463, 95, 40 },
		{ 463, 96, 40 },
		{ 463, 97, 40 },
		{ 463, 98, 40 },
		{ 463, 99, 40 },
		{ 463, 100, 40 },
		{ 463, 101, 40 },
		{ 463, 102, 40 },
		{ 463, 103, 40 },
		{ 463, 104, 40 },
		{ 463, 105, 40 },
		{ 463, 106, 40 },
		{ 463, 107, 40 },
		{ 463, 108, 40 },
		{ 463, 109, 40 },
		{ 463, 110, 40 },
		{ 463, 111, 40 },
		{ 463, 112, 40 },
		{ 463, 113, 40 },
		{ 463, 114, 40 },
		{ 463, 115, 40 },
		{ 463, 116, 40 },
		{ 463, 117, 40 },
		{ 463, 118, 40 },
		{ 463, 119, 40 },
		{ 463, 120, 40 },
		{ 463, 121, 40 },
		{ 463, 122, 40 },
		{ 463, 123, 40 },
		{ 463, 124, 40 },
		{ 463, 125, 40 },
		{ 463, 126, 40 },
		{ 463, 127, 40 },
		{ 463, 128, 40 },
		{ 464, 76, 40 },
		{ 464, 77, 40 },
		{ 464, 78, 40 },
		{ 464, 79, 40 },
		{ 464, 80, 40 },
		{ 464, 81, 40 },
		{ 464, 82, 40 },
		{ 464, 83, 40 },
		{ 464, 84, 40 },
		{ 464, 85, 40 },
		{ 464, 86, 40 },
		{ 464, 87, 40 },
		{ 464, 88, 40 },
		{ 464, 89, 40 },
		{ 464, 90, 40 },
		{ 464, 91, 40 },
		{ 464, 92, 40 },
		{ 464, 93, 40 },
		{ 464, 94, 40 },
		{ 464, 95, 40 },
		{ 464, 96, 40 },
		{ 464, 97, 40 },
		{ 464, 98, 40 },
		{ 464, 99, 40 },
		{ 464, 100, 40 },
		{ 464, 101, 40 },
		{ 464, 102, 40 },
		{ 464, 103, 40 },
		{ 464, 104, 40 },
		{ 464, 105, 40 },
		{ 464, 106, 40 },
		{ 464, 107, 40 },
		{ 464, 108, 40 },
		{ 464, 109, 40 },
		{ 464, 110, 40 },
		{ 464, 111, 40 },
		{ 464, 112, 40 },
		{ 464, 113, 40 },
		{ 464, 114, 40 },
		{ 464, 115, 40 },
		{ 464, 116, 40 },
		{ 464, 117, 40 },
		{ 464, 118, 40 },
		{ 464, 119, 40 },
		{ 464, 120, 40 },
		{ 464, 121, 40 },
		{ 464, 122, 40 },
		{ 464, 123, 40 },
		{ 464, 124, 40 },
		{ 464, 125, 40 },
		{ 464, 126, 40 },
		{ 464, 127, 40 },
		{ 465, 76, 40 },
		{ 465, 77, 40 },
		{ 465, 78, 40 },
		{ 465, 79, 40 },
		{ 465, 80, 40 },
		{ 465, 81, 40 },
		{ 465, 82, 40 },
		{ 465, 83, 40 },
		{ 465, 84, 40 },
		{ 465, 85, 40 },
		{ 465, 86, 40 },
		{ 465, 87, 40 },
		{ 465, 88, 40 },
		{ 465, 89, 40 },
		{ 465, 90, 40 },
		{ 465, 91, 40 },
		{ 465, 92, 40 },
		{ 465, 93, 40 },
		{ 465, 94, 40 },
		{ 465, 95, 40 },
		{ 465, 96, 40 },
		{ 465, 97, 40 },
		{ 465, 98, 40 },
		{ 465, 99, 40 },
		{ 465, 100, 40 },
		{ 465, 101, 40 },
		{ 465, 102, 40 },
		{ 465, 103, 40 },
		{ 465, 104, 40 },
		{ 465, 105, 40 },
		{ 465, 106, 40 },
		{ 465, 107, 40 },
		{ 465, 108, 40 },
		{ 465, 109, 40 },
		{ 465, 110, 40 },
		{ 465, 111, 40 },
		{ 465, 112, 40 },
		{ 465, 113, 40 },
		{ 465, 114, 40 },
		{ 465, 115, 40 },
		{ 465, 116, 40 },
		{ 465, 117, 40 },
		{ 465, 118, 40 },
		{ 465, 119, 40 },
		{ 465, 120, 40 },
		{ 465, 121, 40 },
		{ 465, 122, 40 },
		{ 465, 123, 40 },
		{ 465, 124, 40 },
		{ 465, 125, 40 },
		{ 465, 126, 40 },
		{ 466, 75, 40 },
		{ 466, 76, 40 },
		{ 466, 77, 40 },
		{ 466, 78, 40 },
		{ 466, 79, 40 },
		{ 466, 80, 40 },
		{ 466, 81, 40 },
		{ 466, 82, 40 },
		{ 466, 83, 40 },
		{ 466, 84, 40 },
		{ 466, 85, 40 },
		{ 466, 86, 40 },
		{ 466, 87, 40 },
		{ 466, 88, 40 },
		{ 466, 89, 40 },
		{ 466, 90, 40 },
		{ 466, 91, 40 },
		{ 466, 92, 40 },
		{ 466, 93, 40 },
		{ 466, 94, 40 },
		{ 466, 95, 40 },
		{ 466, 96, 40 },
		{ 466, 97, 40 },
		{ 466, 98, 40 },
		{ 466, 99, 40 },
		{ 466, 100, 40 },
		{ 466, 101, 40 },
		{ 466, 102, 40 },
		{ 466, 103, 40 },
		{ 466, 104, 40 },
		{ 466, 105, 40 },
		{ 466, 106, 40 },
		{ 466, 107, 40 },
		{ 466, 108, 40 },
		{ 466, 109, 40 },
		{ 466, 110, 40 },
		{ 466, 111, 40 },
		{ 466, 112, 40 },
		{ 466, 113, 40 },
		{ 466, 114, 40 },
		{ 466, 115, 40 },
		{ 466, 116, 40 },
		{ 466, 117, 40 },
		{ 466, 118, 40 },
		{ 466, 119, 40 },
		{ 466, 120, 40 },
		{ 466, 121, 40 },
		{ 466, 122, 40 },
		{ 466, 123, 40 },
		{ 466, 124, 40 },
		{ 466, 125, 40 },
		{ 466, 126, 40 },
		{ 467, 74, 40 },
		{ 467, 75, 40 },
		{ 467, 76, 40 },
		{ 467, 77, 40 },
		{ 467, 78, 40 },
		{ 467, 79, 40 },
		{ 467, 80, 40 },
		{ 467, 81, 40 },
		{ 467, 82, 40 },
		{ 467, 83, 40 },
		{ 467, 84, 40 },
		{ 467, 85, 40 },
		{ 467, 86, 40 },
		{ 467, 87, 40 },
		{ 467, 88, 40 },
		{ 467, 89, 40 },
		{ 467, 90, 40 },
		{ 467, 91, 40 },
		{ 467, 92, 40 },
		{ 467, 93, 40 },
		{ 467, 94, 40 },
		{ 467, 95, 40 },
		{ 467, 96, 40 },
		{ 467, 97, 40 },
		{ 467, 98, 40 },
		{ 467, 99, 40 },
		{ 467, 100, 40 },
		{ 467, 101, 40 },
		{ 467, 102, 40 },
		{ 467, 103, 40 },
		{ 467, 104, 40 },
		{ 467, 105, 40 },
		{ 467, 106, 40 },
		{ 467, 107, 40 },
		{ 467, 108, 40 },
		{ 467, 109, 40 },
		{ 467, 110, 40 },
		{ 467, 111, 40 },
		{ 467, 112, 40 },
		{ 467, 113, 40 },
		{ 467, 114, 40 },
		{ 467, 115, 40 },
		{ 467, 116, 40 },
		{ 467, 117, 40 },
		{ 467, 118, 40 },
		{ 467, 119, 40 },
		{ 467, 120, 40 },
		{ 467, 121, 40 },
		{ 467, 122, 40 },
		{ 467, 123, 40 },
		{ 467, 124, 40 },
		{ 467, 125, 40 },
		{ 468, 73, 40 },
		{ 468, 74, 40 },
		{ 468, 75, 40 },
		{ 468, 76, 40 },
		{ 468, 77, 40 },
		{ 468, 78, 40 },
		{ 468, 79, 40 },
		{ 468, 80, 40 },
		{ 468, 81, 40 },
		{ 468, 82, 40 },
		{ 468, 83, 40 },
		{ 468, 84, 40 },
		{ 468, 85, 40 },
		{ 468, 86, 40 },
		{ 468, 87, 40 },
		{ 468, 88, 40 },
		{ 468, 89, 40 },
		{ 468, 90, 40 },
		{ 468, 91, 40 },
		{ 468, 92, 40 },
		{ 468, 93, 40 },
		{ 468, 94, 40 },
		{ 468, 95, 40 },
		{ 468, 96, 40 },
		{ 468, 97, 40 },
		{ 468, 98, 40 },
		{ 468, 99, 40 },
		{ 468, 100, 40 },
		{ 468, 101, 40 },
		{ 468, 102, 40 },
		{ 468, 103, 40 },
		{ 468, 104, 40 },
		{ 468, 105, 40 },
		{ 468, 106, 40 },
		{ 468, 107, 40 },
		{ 468, 108, 40 },
		{ 468, 109, 40 },
		{ 468, 110, 40 },
		{ 468, 111, 40 },
		{ 468, 112, 40 },
		{ 468, 113, 40 },
		{ 468, 114, 40 },
		{ 468, 115, 40 },
		{ 468, 116, 40 },
		{ 468, 117, 40 },
		{ 468, 118, 40 },
		{ 468, 119, 40 },
		{ 468, 120, 40 },
		{ 468, 121, 40 },
		{ 468, 122, 40 },
		{ 468, 123, 40 },
		{ 468, 124, 40 },
		{ 468, 125, 40 },
		{ 469, 72, 40 },
		{ 469, 73, 40 },
		{ 469, 74, 40 },
		{ 469, 75, 40 },
		{ 469, 76, 40 },
		{ 469, 77, 40 },
		{ 469, 78, 40 },
		{ 469, 79, 40 },
		{ 469, 80, 40 },
		{ 469, 81, 40 },
		{ 469, 82, 40 },
		{ 469, 83, 40 },
		{ 469, 84, 40 },
		{ 469, 85, 40 },
		{ 469, 86, 40 },
		{ 469, 87, 40 },
		{ 469, 88, 40 },
		{ 469, 89, 40 },
		{ 469, 90, 40 },
		{ 469, 91, 40 },
		{ 469, 92, 40 },
		{ 469, 93, 40 },
		{ 469, 94, 40 },
		{ 469, 95, 40 },
		{ 469, 96, 40 },
		{ 469, 97, 40 },
		{ 469, 98, 40 },
		{ 469, 99, 40 },
		{ 469, 100, 40 },
		{ 469, 101, 40 },
		{ 469, 102, 40 },
		{ 469, 103, 40 },
		{ 469, 104, 40 },
		{ 469, 105, 40 },
		{ 469, 106, 40 },
		{ 469, 107, 40 },
		{ 469, 108, 40 },
		{ 469, 109, 40 },
		{ 469, 110, 40 },
		{ 469, 111, 40 },
		{ 469, 112, 40 },
		{ 469, 113, 40 },
		{ 469, 114, 40 },
		{ 469, 115, 40 },
		{ 469, 116, 40 },
		{ 469, 117, 40 },
		{ 469, 118, 40 },
		{ 469, 119, 40 },
		{ 469, 120, 40 },
		{ 469, 121, 40 },
		{ 469, 122, 40 },
		{ 469, 123, 40 },
		{ 469, 124, 40 },
		{ 470, 71, 40 },
		{ 470, 72, 40 },
		{ 470, 73, 40 },
		{ 470, 74, 40 },
		{ 470, 75, 40 },
		{ 470, 76, 40 },
		{ 470, 77, 40 },
		{ 470, 78, 40 },
		{ 470, 79, 40 },
		{ 470, 80, 40 },
		{ 470, 81, 40 },
		{ 470, 82, 40 },
		{ 470, 83, 40 },
		{ 470, 84, 40 },
		{ 470, 85, 40 },
		{ 470, 86, 40 },
		{ 470, 87, 40 },
		{ 470, 88, 40 },
		{ 470, 89, 40 },
		{ 470, 90, 40 },
		{ 470, 91, 40 },
		{ 470, 92, 40 },
		{ 470, 93, 40 },
		{ 470, 94, 40 },
		{ 470, 95, 40 },
		{ 470, 96, 40 },
		{ 470, 97, 40 },
		{ 470, 98, 40 },
		{ 470, 99, 40 },
		{ 470, 100, 40 },
		{ 470, 101, 40 },
		{ 470, 102, 40 },
		{ 470, 103, 40 },
		{ 470, 104, 40 },
		{ 470, 105, 40 },
		{ 470, 106, 40 },
		{ 470, 107, 40 },
		{ 470, 108, 40 },
		{ 470, 109, 40 },
		{ 470, 110, 40 },
		{ 470, 111, 40 },
		{ 470, 112, 40 },
		{ 470, 113, 40 },
		{ 470, 114, 40 },
		{ 470, 115, 40 },
		{ 470, 116, 40 },
		{ 470, 117, 40 },
		{ 470, 118, 40 },
		{ 470, 119, 40 },
		{ 470, 120, 40 },
		{ 470, 121, 40 },
		{ 470, 122, 40 },
		{ 470, 123, 40 },
		{ 471, 70, 40 },
		{ 471, 71, 40 },
		{ 471, 72, 40 },
		{ 471, 73, 40 },
		{ 471, 74, 40 },
		{ 471, 75, 40 },
		{ 471, 76, 40 },
		{ 471, 77, 40 },
		{ 471, 78, 40 },
		{ 471, 79, 40 },
		{ 471, 80, 40 },
		{ 471, 81, 40 },
		{ 471, 82, 40 },
		{ 471, 83, 40 },
		{ 471, 84, 40 },
		{ 471, 85, 40 },
		{ 471, 86, 40 },
		{ 471, 87, 40 },
		{ 471, 88, 40 },
		{ 471, 89, 40 },
		{ 471, 90, 40 },
		{ 471, 91, 40 },
		{ 471, 92, 40 },
		{ 471, 93, 40 },
		{ 471, 94, 40 },
		{ 471, 95, 40 },
		{ 471, 96, 40 },
		{ 471, 97, 40 },
		{ 471, 98, 40 },
		{ 471, 99, 40 },
		{ 471, 100, 40 },
		{ 471, 101, 40 },
		{ 471, 102, 40 },
		{ 471, 103, 40 },
		{ 471, 104, 40 },
		{ 471, 105, 40 },
		{ 471, 106, 40 },
		{ 471, 107, 40 },
		{ 471, 108, 40 },
		{ 471, 109, 40 },
		{ 471, 110, 40 },
		{ 471, 111, 40 },
		{ 471, 112, 40 },
		{ 471, 113, 40 },
		{ 471, 114, 40 },
		{ 471, 115, 40 },
		{ 471, 116, 40 },
		{ 471, 117, 40 },
		{ 471, 118, 40 },
		{ 471, 119, 40 },
		{ 471, 120, 40 },
		{ 471, 121, 40 },
		{ 471, 122, 40 },
		{ 471, 123, 40 },
		{ 472, 69, 40 },
		{ 472, 70, 40 },
		{ 472, 71, 40 },
		{ 472, 72, 40 },
		{ 472, 73, 40 },
		{ 472, 74, 40 },
		{ 472, 75, 40 },
		{ 472, 76, 40 },
		{ 472, 77, 40 },
		{ 472, 78, 40 },
		{ 472, 79, 40 },
		{ 472, 80, 40 },
		{ 472, 81, 40 },
		{ 472, 82, 40 },
		{ 472, 83, 40 },
		{ 472, 84, 40 },
		{ 472, 85, 40 },
		{ 472, 86, 40 },
		{ 472, 87, 40 },
		{ 472, 88, 40 },
		{ 472, 89, 40 },
		{ 472, 90, 40 },
		{ 472, 91, 40 },
		{ 472, 92, 40 },
		{ 472, 93, 40 },
		{ 472, 94, 40 },
		{ 472, 95, 40 },
		{ 472, 96, 40 },
		{ 472, 97, 40 },
		{ 472, 98, 40 },
		{ 472, 99, 40 },
		{ 472, 100, 40 },
		{ 472, 101, 40 },
		{ 472, 102, 40 },
		{ 472, 103, 40 },
		{ 472, 104, 40 },
		{ 472, 105, 40 },
		{ 472, 106, 40 },
		{ 472, 107, 40 },
		{ 472, 108, 40 },
		{ 472, 109, 40 },
		{ 472, 110, 40 },
		{ 472, 111, 40 },
		{ 472, 112, 40 },
		{ 472, 113, 40 },
		{ 472, 114, 40 },
		{ 472, 115, 40 },
		{ 472, 116, 40 },
		{ 472, 117, 40 },
		{ 472, 118, 40 },
		{ 472, 119, 40 },
		{ 472, 120, 40 },
		{ 472, 121, 40 },
		{ 472, 122, 40 },
		{ 473, 69, 40 },
		{ 473, 70, 40 },
		{ 473, 71, 40 },
		{ 473, 72, 40 },
		{ 473, 73, 40 },
		{ 473, 74, 40 },
		{ 473, 75, 40 },
		{ 473, 76, 40 },
		{ 473, 77, 40 },
		{ 473, 78, 40 },
		{ 473, 79, 40 },
		{ 473, 80, 40 },
		{ 473, 81, 40 },
		{ 473, 82, 40 },
		{ 473, 83, 40 },
		{ 473, 84, 40 },
		{ 473, 85, 40 },
		{ 473, 86, 40 },
		{ 473, 87, 40 },
		{ 473, 88, 40 },
		{ 473, 89, 40 },
		{ 473, 90, 40 },
		{ 473, 91, 40 },
		{ 473, 92, 40 },
		{ 473, 93, 40 },
		{ 473, 94, 40 },
		{ 473, 95, 40 },
		{ 473, 96, 40 },
		{ 473, 97, 40 },
		{ 473, 98, 40 },
		{ 473, 99, 40 },
		{ 473, 100, 40 },
		{ 473, 101, 40 },
		{ 473, 102, 40 },
		{ 473, 103, 40 },
		{ 473, 104, 40 },
		{ 473, 105, 40 },
		{ 473, 106, 40 },
		{ 473, 107, 40 },
		{ 473, 108, 40 },
		{ 473, 109, 40 },
		{ 473, 110, 40 },
		{ 473, 111, 40 },
		{ 473, 112, 40 },
		{ 473, 113, 40 },
		{ 473, 114, 40 },
		{ 473, 115, 40 },
		{ 473, 116, 40 },
		{ 473, 117, 40 },
		{ 473, 118, 40 },
		{ 473, 119, 40 },
		{ 473, 120, 40 },
		{ 473, 121, 40 },
		{ 474, 68, 40 },
		{ 474, 69, 40 },
		{ 474, 70, 40 },
		{ 474, 71, 40 },
		{ 474, 72, 40 },
		{ 474, 73, 40 },
		{ 474, 74, 40 },
		{ 474, 75, 40 },
		{ 474, 76, 40 },
		{ 474, 77, 40 },
		{ 474, 78, 40 },
		{ 474, 79, 40 },
		{ 474, 80, 40 },
		{ 474, 81, 40 },
		{ 474, 82, 40 },
		{ 474, 83, 40 },
		{ 474, 84, 40 },
		{ 474, 85, 40 },
		{ 474, 86, 40 },
		{ 474, 87, 40 },
		{ 474, 88, 40 },
		{ 474, 89, 40 },
		{ 474, 90, 40 },
		{ 474, 91, 40 },
		{ 474, 92, 40 },
		{ 474, 93, 40 },
		{ 474, 94, 40 },
		{ 474, 95, 40 },
		{ 474, 96, 40 },
		{ 474, 97, 40 },
		{ 474, 98, 40 },
		{ 474, 99, 40 },
		{ 474, 100, 40 },
		{ 474, 101, 40 },
		{ 474, 102, 40 },
		{ 474, 103, 40 },
		{ 474, 104, 40 },
		{ 474, 105, 40 },
		{ 474, 106, 40 },
		{ 474, 107, 40 },
		{ 474, 108, 40 },
		{ 474, 109, 40 },
		{ 474, 110, 40 },
		{ 474, 111, 40 },
		{ 474, 112, 40 },
		{ 474, 113, 40 },
		{ 474, 114, 40 },
		{ 474, 115, 40 },
		{ 474, 116, 40 },
		{ 474, 117, 40 },
		{ 474, 118, 40 },
		{ 474, 119, 40 },
		{ 474, 120, 40 },
		{ 474, 121, 40 },
		{ 475, 67, 40 },
		{ 475, 68, 40 },
		{ 475, 69, 40 },
		{ 475, 70, 40 },
		{ 475, 71, 40 },
		{ 475, 72, 40 },
		{ 475, 73, 40 },
		{ 475, 74, 40 },
		{ 475, 75, 40 },
		{ 475, 76, 40 },
		{ 475, 77, 40 },
		{ 475, 78, 40 },
		{ 475, 79, 40 },
		{ 475, 80, 40 },
		{ 475, 81, 40 },
		{ 475, 82, 40 },
		{ 475, 83, 40 },
		{ 475, 84, 40 },
		{ 475, 85, 40 },
		{ 475, 86, 40 },
		{ 475, 87, 40 },
		{ 475, 88, 40 },
		{ 475, 89, 40 },
		{ 475, 90, 40 },
		{ 475, 91, 40 },
		{ 475, 92, 40 },
		{ 475, 93, 40 },
		{ 475, 94, 40 },
		{ 475, 95, 40 },
		{ 475, 96, 40 },
		{ 475, 97, 40 },
		{ 475, 98, 40 },
		{ 475, 99, 40 },
		{ 475, 100, 40 },
		{ 475, 101, 40 },
		{ 475, 102, 40 },
		{ 475, 103, 40 },
		{ 475, 104, 40 },
		{ 475, 105, 40 },
		{ 475, 106, 40 },
		{ 475, 107, 40 },
		{ 475, 108, 40 },
		{ 475, 109, 40 },
		{ 475, 110, 40 },
		{ 475, 111, 40 },
		{ 475, 112, 40 },
		{ 475, 113, 40 },
		{ 475, 114, 40 },
		{ 475, 115, 40 },
		{ 475, 116, 40 },
		{ 475, 117, 40 },
		{ 475, 118, 40 },
		{ 475, 119, 40 },
		{ 475, 120, 40 },
		{ 476, 66, 40 },
		{ 476, 67, 40 },
		{ 476, 68, 40 },
		{ 476, 69, 40 },
		{ 476, 70, 40 },
		{ 476, 71, 40 },
		{ 476, 72, 40 },
		{ 476, 73, 40 },
		{ 476, 74, 40 },
		{ 476, 75, 40 },
		{ 476, 76, 40 },
		{ 476, 77, 40 },
		{ 476, 78, 40 },
		{ 476, 79, 40 },
		{ 476, 80, 40 },
		{ 476, 81, 40 },
		{ 476, 82, 40 },
		{ 476, 83, 40 },
		{ 476, 84, 40 },
		{ 476, 85, 40 },
		{ 476, 86, 40 },
		{ 476, 87, 40 },
		{ 476, 88, 40 },
		{ 476, 89, 40 },
		{ 476, 90, 40 },
		{ 476, 91, 40 },
		{ 476, 92, 40 },
		{ 476, 93, 40 },
		{ 476, 94, 40 },
		{ 476, 95, 40 },
		{ 476, 96, 40 },
		{ 476, 97, 40 },
		{ 476, 98, 40 },
		{ 476, 99, 40 },
		{ 476, 100, 40 },
		{ 476, 101, 40 },
		{ 476, 102, 40 },
		{ 476, 103, 40 },
		{ 476, 104, 40 },
		{ 476, 105, 40 },
		{ 476, 106, 40 },
		{ 476, 107, 40 },
		{ 476, 108, 40 },
		{ 476, 109, 40 },
		{ 476, 110, 40 },
		{ 476, 111, 40 },
		{ 476, 112, 40 },
		{ 476, 113, 40 },
		{ 476, 114, 40 },
		{ 476, 115, 40 },
		{ 476, 116, 40 },
		{ 476, 117, 40 },
		{ 476, 118, 40 },
		{ 476, 119, 40 },
		{ 476, 120, 40 },
		{ 477, 65, 40 },
		{ 477, 66, 40 },
		{ 477, 67, 40 },
		{ 477, 68, 40 },
		{ 477, 69, 40 },
		{ 477, 70, 40 },
		{ 477, 71, 40 },
		{ 477, 72, 40 },
		{ 477, 73, 40 },
		{ 477, 74, 40 },
		{ 477, 75, 40 },
		{ 477, 76, 40 },
		{ 477, 77, 40 },
		{ 477, 78, 40 },
		{ 477, 79, 40 },
		{ 477, 80, 40 },
		{ 477, 81, 40 },
		{ 477, 82, 40 },
		{ 477, 83, 40 },
		{ 477, 84, 40 },
		{ 477, 85, 40 },
		{ 477, 86, 40 },
		{ 477, 87, 40 },
		{ 477, 88, 40 },
		{ 477, 89, 40 },
		{ 477, 90, 40 },
		{ 477, 91, 40 },
		{ 477, 92, 40 },
		{ 477, 93, 40 },
		{ 477, 94, 40 },
		{ 477, 95, 40 },
		{ 477, 96, 40 },
		{ 477, 97, 40 },
		{ 477, 98, 40 },
		{ 477, 99, 40 },
		{ 477, 100, 40 },
		{ 477, 101, 40 },
		{ 477, 102, 40 },
		{ 477, 103, 40 },
		{ 477, 104, 40 },
		{ 477, 105, 40 },
		{ 477, 106, 40 },
		{ 477, 107, 40 },
		{ 477, 108, 40 },
		{ 477, 109, 40 },
		{ 477, 110, 40 },
		{ 477, 111, 40 },
		{ 477, 112, 40 },
		{ 477, 113, 40 },
		{ 477, 114, 40 },
		{ 477, 115, 40 },
		{ 477, 116, 40 },
		{ 477, 117, 40 },
		{ 477, 118, 40 },
		{ 477, 119, 40 },
		{ 478, 64, 40 },
		{ 478, 65, 40 },
		{ 478, 66, 40 },
		{ 478, 67, 40 },
		{ 478, 68, 40 },
		{ 478, 69, 40 },
		{ 478, 70, 40 },
		{ 478, 71, 40 },
		{ 478, 72, 40 },
		{ 478, 73, 40 },
		{ 478, 74, 40 },
		{ 478, 75, 40 },
		{ 478, 76, 40 },
		{ 478, 77, 40 },
		{ 478, 78, 40 },
		{ 478, 79, 40 },
		{ 478, 80, 40 },
		{ 478, 81, 40 },
		{ 478, 82, 40 },
		{ 478, 83, 40 },
		{ 478, 84, 40 },
		{ 478, 85, 40 },
		{ 478, 86, 40 },
		{ 478, 87, 40 },
		{ 478, 88, 40 },
		{ 478, 89, 40 },
		{ 478, 90, 40 },
		{ 478, 91, 40 },
		{ 478, 92, 40 },
		{ 478, 93, 40 },
		{ 478, 94, 40 },
		{ 478, 95, 40 },
		{ 478, 96, 40 },
		{ 478, 97, 40 },
		{ 478, 98, 40 },
		{ 478, 99, 40 },
		{ 478, 100, 40 },
		{ 478, 101, 40 },
		{ 478, 102, 40 },
		{ 478, 103, 40 },
		{ 478, 104, 40 },
		{ 478, 105, 40 },
		{ 478, 106, 40 },
		{ 478, 107, 40 },
		{ 478, 108, 40 },
		{ 478, 109, 40 },
		{ 478, 110, 40 },
		{ 478, 111, 40 },
		{ 478, 112, 40 },
		{ 478, 113, 40 },
		{ 478, 114, 40 },
		{ 478, 115, 40 },
		{ 478, 116, 40 },
		{ 478, 117, 40 },
		{ 478, 118, 40 },
		{ 479, 63, 40 },
		{ 479, 64, 40 },
		{ 479, 65, 40 },
		{ 479, 66, 40 },
		{ 479, 67, 40 },
		{ 479, 68, 40 },
		{ 479, 69, 40 },
		{ 479, 70, 40 },
		{ 479, 71, 40 },
		{ 479, 72, 40 },
		{ 479, 73, 40 },
		{ 479, 74, 40 },
		{ 479, 75, 40 },
		{ 479, 76, 40 },
		{ 479, 77, 40 },
		{ 479, 78, 40 },
		{ 479, 79, 40 },
		{ 479, 80, 40 },
		{ 479, 81, 40 },
		{ 479, 82, 40 },
		{ 479, 83, 40 },
		{ 479, 84, 40 },
		{ 479, 85, 40 },
		{ 479, 86, 40 },
		{ 479, 87, 40 },
		{ 479, 88, 40 },
		{ 479, 89, 40 },
		{ 479, 90, 40 },
		{ 479, 91, 40 },
		{ 479, 92, 40 },
		{ 479, 93, 40 },
		{ 479, 94, 40 },
		{ 479, 95, 40 },
		{ 479, 96, 40 },
		{ 479, 97, 40 },
		{ 479, 98, 40 },
		{ 479, 99, 40 },
		{ 479, 100, 40 },
		{ 479, 101, 40 },
		{ 479, 102, 40 },
		{ 479, 103, 40 },
		{ 479, 104, 40 },
		{ 479, 105, 40 },
		{ 479, 106, 40 },
		{ 479, 107, 40 },
		{ 479, 108, 40 },
		{ 479, 109, 40 },
		{ 479, 110, 40 },
		{ 479, 111, 40 },
		{ 479, 112, 40 },
		{ 479, 113, 40 },
		{ 479, 114, 40 },
		{ 479, 115, 40 },
		{ 479, 116, 40 },
		{ 479, 117, 40 },
		{ 479, 118, 40 },
		{ 480, 62, 40 },
		{ 480, 63, 40 },
		{ 480, 64, 40 },
		{ 480, 65, 40 },
		{ 480, 66, 40 },
		{ 480, 67, 40 },
		{ 480, 68, 40 },
		{ 480, 69, 40 },
		{ 480, 70, 40 },
		{ 480, 71, 40 },
		{ 480, 72, 40 },
		{ 480, 73, 40 },
		{ 480, 74, 40 },
		{ 480, 75, 40 },
		{ 480, 76, 40 },
		{ 480, 77, 40 },
		{ 480, 78, 40 },
		{ 480, 79, 40 },
		{ 480, 80, 40 },
		{ 480, 81, 40 },
		{ 480, 82, 40 },
		{ 480, 83, 40 },
		{ 480, 84, 40 },
		{ 480, 85, 40 },
		{ 480, 86, 40 },
		{ 480, 87, 40 },
		{ 480, 88, 40 },
		{ 480, 89, 40 },
		{ 480, 90, 40 },
		{ 480, 91, 40 },
		{ 480, 92, 40 },
		{ 480, 93, 40 },
		{ 480, 94, 40 },
		{ 480, 95, 40 },
		{ 480, 96, 40 },
		{ 480, 97, 40 },
		{ 480, 98, 40 },
		{ 480, 99, 40 },
		{ 480, 100, 40 },
		{ 480, 101, 40 },
		{ 480, 102, 40 },
		{ 480, 103, 40 },
		{ 480, 104, 40 },
		{ 480, 105, 40 },
		{ 480, 106, 40 },
		{ 480, 107, 40 },
		{ 480, 108, 40 },
		{ 480, 109, 40 },
		{ 480, 110, 40 },
		{ 480, 111, 40 },
		{ 480, 112, 40 },
		{ 480, 113, 40 },
		{ 480, 114, 40 },
		{ 480, 115, 40 },
		{ 480, 116, 40 },
		{ 480, 117, 40 },
		{ 481, 62, 40 },
		{ 481, 63, 40 },
		{ 481, 64, 40 },
		{ 481, 65, 40 },
		{ 481, 66, 40 },
		{ 481, 67, 40 },
		{ 481, 68, 40 },
		{ 481, 69, 40 },
		{ 481, 70, 40 },
		{ 481, 71, 40 },
		{ 481, 72, 40 },
		{ 481, 73, 40 },
		{ 481, 74, 40 },
		{ 481, 75, 40 },
		{ 481, 76, 40 },
		{ 481, 77, 40 },
		{ 481, 78, 40 },
		{ 481, 79, 40 },
		{ 481, 80, 40 },
		{ 481, 81, 40 },
		{ 481, 82, 40 },
		{ 481, 83, 40 },
		{ 481, 84, 40 },
		{ 481, 85, 40 },
		{ 481, 86, 40 },
		{ 481, 87, 40 },
		{ 481, 88, 40 },
		{ 481, 89, 40 },
		{ 481, 90, 40 },
		{ 481, 91, 40 },
		{ 481, 92, 40 },
		{ 481, 93, 40 },
		{ 481, 94, 40 },
		{ 481, 95, 40 },
		{ 481, 96, 40 },
		{ 481, 97, 40 },
		{ 481, 98, 40 },
		{ 481, 99, 40 },
		{ 481, 100, 40 },
		{ 481, 101, 40 },
		{ 481, 102, 40 },
		{ 481, 103, 40 },
		{ 481, 104, 40 },
		{ 481, 105, 40 },
		{ 481, 106, 40 },
		{ 481, 107, 40 },
		{ 481, 108, 40 },
		{ 481, 109, 40 },
		{ 481, 110, 40 },
		{ 481, 111, 40 },
		{ 481, 112, 40 },
		{ 481, 113, 40 },
		{ 481, 114, 40 },
		{ 481, 115, 40 },
		{ 481, 116, 40 },
		{ 482, 61, 40 },
		{ 482, 62, 40 },
		{ 482, 63, 40 },
		{ 482, 64, 40 },
		{ 482, 65, 40 },
		{ 482, 66, 40 },
		{ 482, 67, 40 },
		{ 482, 68, 40 },
		{ 482, 69, 40 },
		{ 482, 70, 40 },
		{ 482, 71, 40 },
		{ 482, 72, 40 },
		{ 482, 73, 40 },
		{ 482, 74, 40 },
		{ 482, 75, 40 },
		{ 482, 76, 40 },
		{ 482, 77, 40 },
		{ 482, 78, 40 },
		{ 482, 79, 40 },
		{ 482, 80, 40 },
		{ 482, 81, 40 },
		{ 482, 82, 40 },
		{ 482, 83, 40 },
		{ 482, 84, 40 },
		{ 482, 85, 40 },
		{ 482, 86, 40 },
		{ 482, 87, 40 },
		{ 482, 88, 40 },
		{ 482, 89, 40 },
		{ 482, 90, 40 },
		{ 482, 91, 40 },
		{ 482, 92, 40 },
		{ 482, 93, 40 },
		{ 482, 94, 40 },
		{ 482, 95, 40 },
		{ 482, 96, 40 },
		{ 482, 97, 40 },
		{ 482, 98, 40 },
		{ 482, 99, 40 },
		{ 482, 100, 40 },
		{ 482, 101, 40 },
		{ 482, 102, 40 },
		{ 482, 103, 40 },
		{ 482, 104, 40 },
		{ 482, 105, 40 },
		{ 482, 106, 40 },
		{ 482, 107, 40 },
		{ 482, 108, 40 },
		{ 482, 109, 40 },
		{ 482, 110, 40 },
		{ 482, 111, 40 },
		{ 482, 112, 40 },
		{ 482, 113, 40 },
		{ 482, 114, 40 },
		{ 482, 115, 40 },
		{ 482, 116, 40 },
		{ 483, 60, 40 },
		{ 483, 61, 40 },
		{ 483, 62, 40 },
		{ 483, 63, 40 },
		{ 483, 64, 40 },
		{ 483, 65, 40 },
		{ 483, 66, 40 },
		{ 483, 67, 40 },
		{ 483, 68, 40 },
		{ 483, 69, 40 },
		{ 483, 70, 40 },
		{ 483, 71, 40 },
		{ 483, 72, 40 },
		{ 483, 73, 40 },
		{ 483, 74, 40 },
		{ 483, 75, 40 },
		{ 483, 76, 40 },
		{ 483, 77, 40 },
		{ 483, 78, 40 },
		{ 483, 79, 40 },
		{ 483, 80, 40 },
		{ 483, 81, 40 },
		{ 483, 82, 40 },
		{ 483, 83, 40 },
		{ 483, 84, 40 },
		{ 483, 85, 40 },
		{ 483, 86, 40 },
		{ 483, 87, 40 },
		{ 483, 88, 40 },
		{ 483, 89, 40 },
		{ 483, 90, 40 },
		{ 483, 91, 40 },
		{ 483, 92, 40 },
		{ 483, 93, 40 },
		{ 483, 94, 40 },
		{ 483, 95, 40 },
		{ 483, 96, 40 },
		{ 483, 97, 40 },
		{ 483, 98, 40 },
		{ 483, 99, 40 },
		{ 483, 100, 40 },
		{ 483, 101, 40 },
		{ 483, 102, 40 },
		{ 483, 103, 40 },
		{ 483, 104, 40 },
		{ 483, 105, 40 },
		{ 483, 106, 40 },
		{ 483, 107, 40 },
		{ 483, 108, 40 },
		{ 483, 109, 40 },
		{ 483, 110, 40 },
		{ 483, 111, 40 },
		{ 483, 112, 40 },
		{ 483, 113, 40 },
		{ 483, 114, 40 },
		{ 483, 115, 40 },
		{ 484, 59, 40 },
		{ 484, 60, 40 },
		{ 484, 61, 40 },
		{ 484, 62, 40 },
		{ 484, 63, 40 },
		{ 484, 64, 40 },
		{ 484, 65, 40 },
		{ 484, 66, 40 },
		{ 484, 67, 40 },
		{ 484, 68, 40 },
		{ 484, 69, 40 },
		{ 484, 70, 40 },
		{ 484, 71, 40 },
		{ 484, 72, 40 },
		{ 484, 73, 40 },
		{ 484, 74, 40 },
		{ 484, 75, 40 },
		{ 484, 76, 40 },
		{ 484, 77, 40 },
		{ 484, 78, 40 },
		{ 484, 79, 40 },
		{ 484, 80, 40 },
		{ 484, 81, 40 },
		{ 484, 82, 40 },
		{ 484, 83, 40 },
		{ 484, 84, 40 },
		{ 484, 85, 40 },
		{ 484, 86, 40 },
		{ 484, 87, 40 },
		{ 484, 88, 40 },
		{ 484, 89, 40 },
		{ 484, 90, 40 },
		{ 484, 91, 40 },
		{ 484, 92, 40 },
		{ 484, 93, 40 },
		{ 484, 94, 40 },
		{ 484, 95, 40 },
		{ 484, 96, 40 },
		{ 484, 97, 40 },
		{ 484, 98, 40 },
		{ 484, 99, 40 },
		{ 484, 100, 40 },
		{ 484, 101, 40 },
		{ 484, 102, 40 },
		{ 484, 103, 40 },
		{ 484, 104, 40 },
		{ 484, 105, 40 },
		{ 484, 106, 40 },
		{ 484, 107, 40 },
		{ 484, 108, 40 },
		{ 484, 109, 40 },
		{ 484, 110, 40 },
		{ 484, 111, 40 },
		{ 484, 112, 40 },
		{ 484, 113, 40 },
		{ 484, 114, 40 },
		{ 484, 115, 40 },
		{ 485, 58, 40 },
		{ 485, 59, 40 },
		{ 485, 60, 40 },
		{ 485, 61, 40 },
		{ 485, 62, 40 },
		{ 485, 63, 40 },
		{ 485, 64, 40 },
		{ 485, 65, 40 },
		{ 485, 66, 40 },
		{ 485, 67, 40 },
		{ 485, 68, 40 },
		{ 485, 69, 40 },
		{ 485, 70, 40 },
		{ 485, 71, 40 },
		{ 485, 72, 40 },
		{ 485, 73, 40 },
		{ 485, 74, 40 },
		{ 485, 75, 40 },
		{ 485, 76, 40 },
		{ 485, 77, 40 },
		{ 485, 78, 40 },
		{ 485, 79, 40 },
		{ 485, 80, 40 },
		{ 485, 81, 40 },
		{ 485, 82, 40 },
		{ 485, 83, 40 },
		{ 485, 84, 40 },
		{ 485, 85, 40 },
		{ 485, 86, 40 },
		{ 485, 87, 40 },
		{ 485, 88, 40 },
		{ 485, 89, 40 },
		{ 485, 90, 40 },
		{ 485, 91, 40 },
		{ 485, 92, 40 },
		{ 485, 93, 40 },
		{ 485, 94, 40 },
		{ 485, 95, 40 },
		{ 485, 96, 40 },
		{ 485, 97, 40 },
		{ 485, 98, 40 },
		{ 485, 99, 40 },
		{ 485, 100, 40 },
		{ 485, 101, 40 },
		{ 485, 102, 40 },
		{ 485, 103, 40 },
		{ 485, 104, 40 },
		{ 485, 105, 40 },
		{ 485, 106, 40 },
		{ 485, 107, 40 },
		{ 485, 108, 40 },
		{ 485, 109, 40 },
		{ 485, 110, 40 },
		{ 485, 111, 40 },
		{ 485, 112, 40 },
		{ 485, 113, 40 },
		{ 485, 114, 40 },
		{ 486, 57, 40 },
		{ 486, 58, 40 },
		{ 486, 59, 40 },
		{ 486, 60, 40 },
		{ 486, 61, 40 },
		{ 486, 62, 40 },
		{ 486, 63, 40 },
		{ 486, 64, 40 },
		{ 486, 65, 40 },
		{ 486, 66, 40 },
		{ 486, 67, 40 },
		{ 486, 68, 40 },
		{ 486, 69, 40 },
		{ 486, 70, 40 },
		{ 486, 71, 40 },
		{ 486, 72, 40 },
		{ 486, 73, 40 },
		{ 486, 74, 40 },
		{ 486, 75, 40 },
		{ 486, 76, 40 },
		{ 486, 77, 40 },
		{ 486, 78, 40 },
		{ 486, 79, 40 },
		{ 486, 80, 40 },
		{ 486, 81, 40 },
		{ 486, 82, 40 },
		{ 486, 83, 40 },
		{ 486, 84, 40 },
		{ 486, 85, 40 },
		{ 486, 86, 40 },
		{ 486, 87, 40 },
		{ 486, 88, 40 },
		{ 486, 89, 40 },
		{ 486, 90, 40 },
		{ 486, 91, 40 },
		{ 486, 92, 40 },
		{ 486, 93, 40 },
		{ 486, 94, 40 },
		{ 486, 95, 40 },
		{ 486, 96, 40 },
		{ 486, 97, 40 },
		{ 486, 98, 40 },
		{ 486, 99, 40 },
		{ 486, 100, 40 },
		{ 486, 101, 40 },
		{ 486, 102, 40 },
		{ 486, 103, 40 },
		{ 486, 104, 40 },
		{ 486, 105, 40 },
		{ 486, 106, 40 },
		{ 486, 107, 40 },
		{ 486, 108, 40 },
		{ 486, 109, 40 },
		{ 486, 110, 40 },
		{ 486, 111, 40 },
		{ 486, 112, 40 },
		{ 486, 113, 40 },
		{ 487, 56, 40 },
		{ 487, 57, 40 },
		{ 487, 58, 40 },
		{ 487, 59, 40 },
		{ 487, 60, 40 },
		{ 487, 61, 40 },
		{ 487, 62, 40 },
		{ 487, 63, 40 },
		{ 487, 64, 40 },
		{ 487, 65, 40 },
		{ 487, 66, 40 },
		{ 487, 67, 40 },
		{ 487, 68, 40 },
		{ 487, 69, 40 },
		{ 487, 70, 40 },
		{ 487, 71, 40 },
		{ 487, 72, 40 },
		{ 487, 73, 40 },
		{ 487, 74, 40 },
		{ 487, 75, 40 },
		{ 487, 76, 40 },
		{ 487, 77, 40 },
		{ 487, 78, 40 },
		{ 487, 79, 40 },
		{ 487, 80, 40 },
		{ 487, 81, 40 },
		{ 487, 82, 40 },
		{ 487, 83, 40 },
		{ 487, 84, 40 },
		{ 487, 85, 40 },
		{ 487, 86, 40 },
		{ 487, 87, 40 },
		{ 487, 88, 40 },
		{ 487, 89, 40 },
		{ 487, 90, 40 },
		{ 487, 91, 40 },
		{ 487, 92, 40 },
		{ 487, 93, 40 },
		{ 487, 94, 40 },
		{ 487, 95, 40 },
		{ 487, 96, 40 },
		{ 487, 97, 40 },
		{ 487, 98, 40 },
		{ 487, 99, 40 },
		{ 487, 100, 40 },
		{ 487, 101, 40 },
		{ 487, 102, 40 },
		{ 487, 103, 40 },
		{ 487, 104, 40 },
		{ 487, 105, 40 },
		{ 487, 106, 40 },
		{ 487, 107, 40 },
		{ 487, 108, 40 },
		{ 487, 109, 40 },
		{ 487, 110, 40 },
		{ 487, 111, 40 },
		{ 487, 112, 40 },
		{ 487, 113, 40 },
		{ 488, 55, 40 },
		{ 488, 56, 40 },
		{ 488, 57, 40 },
		{ 488, 58, 40 },
		{ 488, 59, 40 },
		{ 488, 60, 40 },
		{ 488, 61, 40 },
		{ 488, 62, 40 },
		{ 488, 63, 40 },
		{ 488, 64, 40 },
		{ 488, 65, 40 },
		{ 488, 66, 40 },
		{ 488, 67, 40 },
		{ 488, 68, 40 },
		{ 488, 69, 40 },
		{ 488, 70, 40 },
		{ 488, 71, 40 },
		{ 488, 72, 40 },
		{ 488, 73, 40 },
		{ 488, 74, 40 },
		{ 488, 75, 40 },
		{ 488, 76, 40 },
		{ 488, 77, 40 },
		{ 488, 78, 40 },
		{ 488, 79, 40 },
		{ 488, 80, 40 },
		{ 488, 81, 40 },
		{ 488, 82, 40 },
		{ 488, 83, 40 },
		{ 488, 84, 40 },
		{ 488, 85, 40 },
		{ 488, 86, 40 },
		{ 488, 87, 40 },
		{ 488, 88, 40 },
		{ 488, 89, 40 },
		{ 488, 90, 40 },
		{ 488, 91, 40 },
		{ 488, 92, 40 },
		{ 488, 93, 40 },
		{ 488, 94, 40 },
		{ 488, 95, 40 },
		{ 488, 96, 40 },
		{ 488, 97, 40 },
		{ 488, 98, 40 },
		{ 488, 99, 40 },
		{ 488, 100, 40 },
		{ 488, 101, 40 },
		{ 488, 102, 40 },
		{ 488, 103, 40 },
		{ 488, 104, 40 },
		{ 488, 105, 40 },
		{ 488, 106, 40 },
		{ 488, 107, 40 },
		{ 488, 108, 40 },
		{ 488, 109, 40 },
		{ 488, 110, 40 },
		{ 488, 111, 40 },
		{ 488, 112, 40 },
		{ 489, 55, 40 },
		{ 489, 56, 40 },
		{ 489, 57, 40 },
		{ 489, 58, 40 },
		{ 489, 59, 40 },
		{ 489, 60, 40 },
		{ 489, 61, 40 },
		{ 489, 62, 40 },
		{ 489, 63, 40 },
		{ 489, 64, 40 },
		{ 489, 65, 40 },
		{ 489, 66, 40 },
		{ 489, 67, 40 },
		{ 489, 68, 40 },
		{ 489, 69, 40 },
		{ 489, 70, 40 },
		{ 489, 71, 40 },
		{ 489, 72, 40 },
		{ 489, 73, 40 },
		{ 489, 74, 40 },
		{ 489, 75, 40 },
		{ 489, 76, 40 },
		{ 489, 77, 40 },
		{ 489, 78, 40 },
		{ 489, 79, 40 },
		{ 489, 80, 40 },
		{ 489, 81, 40 },
		{ 489, 82, 40 },
		{ 489, 83, 40 },
		{ 489, 84, 40 },
		{ 489, 85, 40 },
		{ 489, 86, 40 },
		{ 489, 87, 40 },
		{ 489, 88, 40 },
		{ 489, 89, 40 },
		{ 489, 90, 40 },
		{ 489, 91, 40 },
		{ 489, 92, 40 },
		{ 489, 93, 40 },
		{ 489, 94, 40 },
		{ 489, 95, 40 },
		{ 489, 96, 40 },
		{ 489, 97, 40 },
		{ 489, 98, 40 },
		{ 489, 99, 40 },
		{ 489, 100, 40 },
		{ 489, 101, 40 },
		{ 489, 102, 40 },
		{ 489, 103, 40 },
		{ 489, 104, 40 },
		{ 489, 105, 40 },
		{ 489, 106, 40 },
		{ 489, 107, 40 },
		{ 489, 108, 40 },
		{ 489, 109, 40 },
		{ 489, 110, 40 },
		{ 489, 111, 40 },
		{ 490, 54, 40 },
		{ 490, 55, 40 },
		{ 490, 56, 40 },
		{ 490, 57, 40 },
		{ 490, 58, 40 },
		{ 490, 59, 40 },
		{ 490, 60, 40 },
		{ 490, 61, 40 },
		{ 490, 62, 40 },
		{ 490, 63, 40 },
		{ 490, 64, 40 },
		{ 490, 65, 40 },
		{ 490, 66, 40 },
		{ 490, 67, 40 },
		{ 490, 68, 40 },
		{ 490, 69, 40 },
		{ 490, 70, 40 },
		{ 490, 71, 40 },
		{ 490, 72, 40 },
		{ 490, 73, 40 },
		{ 490, 74, 40 },
		{ 490, 75, 40 },
		{ 490, 76, 40 },
		{ 490, 77, 40 },
		{ 490, 78, 40 },
		{ 490, 79, 40 },
		{ 490, 80, 40 },
		{ 490, 81, 40 },
		{ 490, 82, 40 },
		{ 490, 83, 40 },
		{ 490, 84, 40 },
		{ 490, 85, 40 },
		{ 490, 86, 40 },
		{ 490, 87, 40 },
		{ 490, 88, 40 },
		{ 490, 89, 40 },
		{ 490, 90, 40 },
		{ 490, 91, 40 },
		{ 490, 92, 40 },
		{ 490, 93, 40 },
		{ 490, 94, 40 },
		{ 490, 95, 40 },
		{ 490, 96, 40 },
		{ 490, 97, 40 },
		{ 490, 98, 40 },
		{ 490, 99, 40 },
		{ 490, 100, 40 },
		{ 490, 101, 40 },
		{ 490, 102, 40 },
		{ 490, 103, 40 },
		{ 490, 104, 40 },
		{ 490, 105, 40 },
		{ 490, 106, 40 },
		{ 490, 107, 40 },
		{ 490, 108, 40 },
		{ 490, 109, 40 },
		{ 490, 110, 40 },
		{ 490, 111, 40 },
		{ 491, 53, 40 },
		{ 491, 54, 40 },
		{ 491, 55, 40 },
		{ 491, 56, 40 },
		{ 491, 57, 40 },
		{ 491, 58, 40 },
		{ 491, 59, 40 },
		{ 491, 60, 40 },
		{ 491, 61, 40 },
		{ 491, 62, 40 },
		{ 491, 63, 40 },
		{ 491, 64, 40 },
		{ 491, 65, 40 },
		{ 491, 66, 40 },
		{ 491, 67, 40 },
		{ 491, 68, 40 },
		{ 491, 69, 40 },
		{ 491, 70, 40 },
		{ 491, 71, 40 },
		{ 491, 72, 40 },
		{ 491, 73, 40 },
		{ 491, 74, 40 },
		{ 491, 75, 40 },
		{ 491, 76, 40 },
		{ 491, 77, 40 },
		{ 491, 78, 40 },
		{ 491, 79, 40 },
		{ 491, 80, 40 },
		{ 491, 81, 40 },
		{ 491, 82, 40 },
		{ 491, 83, 40 },
		{ 491, 84, 40 },
		{ 491, 85, 40 },
		{ 491, 86, 40 },
		{ 491, 87, 40 },
		{ 491, 88, 40 },
		{ 491, 89, 40 },
		{ 491, 90, 40 },
		{ 491, 91, 40 },
		{ 491, 92, 40 },
		{ 491, 93, 40 },
		{ 491, 94, 40 },
		{ 491, 95, 40 },
		{ 491, 96, 40 },
		{ 491, 97, 40 },
		{ 491, 98, 40 },
		{ 491, 99, 40 },
		{ 491, 100, 40 },
		{ 491, 101, 40 },
		{ 491, 102, 40 },
		{ 491, 103, 40 },
		{ 491, 104, 40 },
		{ 491, 105, 40 },
		{ 491, 106, 40 },
		{ 491, 107, 40 },
		{ 491, 108, 40 },
		{ 491, 109, 40 },
		{ 491, 110, 40 },
		{ 492, 52, 40 },
		{ 492, 53, 40 },
		{ 492, 54, 40 },
		{ 492, 55, 40 },
		{ 492, 56, 40 },
		{ 492, 57, 40 },
		{ 492, 58, 40 },
		{ 492, 59, 40 },
		{ 492, 60, 40 },
		{ 492, 61, 40 },
		{ 492, 62, 40 },
		{ 492, 63, 40 },
		{ 492, 64, 40 },
		{ 492, 65, 40 },
		{ 492, 66, 40 },
		{ 492, 67, 40 },
		{ 492, 68, 40 },
		{ 492, 69, 40 },
		{ 492, 70, 40 },
		{ 492, 71, 40 },
		{ 492, 72, 40 },
		{ 492, 73, 40 },
		{ 492, 74, 40 },
		{ 492, 75, 40 },
		{ 492, 76, 40 },
		{ 492, 77, 40 },
		{ 492, 78, 40 },
		{ 492, 79, 40 },
		{ 492, 80, 40 },
		{ 492, 81, 40 },
		{ 492, 82, 40 },
		{ 492, 83, 40 },
		{ 492, 84, 40 },
		{ 492, 85, 40 },
		{ 492, 86, 40 },
		{ 492, 87, 40 },
		{ 492, 88, 40 },
		{ 492, 89, 40 },
		{ 492, 90, 40 },
		{ 492, 91, 40 },
		{ 492, 92, 40 },
		{ 492, 93, 40 },
		{ 492, 94, 40 },
		{ 492, 95, 40 },
		{ 492, 96, 40 },
		{ 492, 97, 40 },
		{ 492, 98, 40 },
		{ 492, 99, 40 },
		{ 492, 100, 40 },
		{ 492, 101, 40 },
		{ 492, 102, 40 },
		{ 492, 103, 40 },
		{ 492, 104, 40 },
		{ 492, 105, 40 },
		{ 492, 106, 40 },
		{ 492, 107, 40 },
		{ 492, 108, 40 },
		{ 492, 109, 40 },
		{ 492, 110, 40 },
		{ 493, 51, 40 },
		{ 493, 52, 40 },
		{ 493, 53, 40 },
		{ 493, 54, 40 },
		{ 493, 55, 40 },
		{ 493, 56, 40 },
		{ 493, 57, 40 },
		{ 493, 58, 40 },
		{ 493, 59, 40 },
		{ 493, 60, 40 },
		{ 493, 61, 40 },
		{ 493, 62, 40 },
		{ 493, 63, 40 },
		{ 493, 64, 40 },
		{ 493, 65, 40 },
		{ 493, 66, 40 },
		{ 493, 67, 40 },
		{ 493, 68, 40 },
		{ 493, 69, 40 },
		{ 493, 70, 40 },
		{ 493, 71, 40 },
		{ 493, 72, 40 },
		{ 493, 73, 40 },
		{ 493, 74, 40 },
		{ 493, 75, 40 },
		{ 493, 76, 40 },
		{ 493, 77, 40 },
		{ 493, 78, 40 },
		{ 493, 79, 40 },
		{ 493, 80, 40 },
		{ 493, 81, 40 },
		{ 493, 82, 40 },
		{ 493, 83, 40 },
		{ 493, 84, 40 },
		{ 493, 85, 40 },
		{ 493, 86, 40 },
		{ 493, 87, 40 },
		{ 493, 88, 40 },
		{ 493, 89, 40 },
		{ 493, 90, 40 },
		{ 493, 91, 40 },
		{ 493, 92, 40 },
		{ 493, 93, 40 },
		{ 493, 94, 40 },
		{ 493, 95, 40 },
		{ 493, 96, 40 },
		{ 493, 97, 40 },
		{ 493, 98, 40 },
		{ 493, 99, 40 },
		{ 493, 100, 40 },
		{ 493, 101, 40 },
		{ 493, 102, 40 },
		{ 493, 103, 40 },
		{ 493, 104, 40 },
		{ 493, 105, 40 },
		{ 493, 106, 40 },
		{ 493, 107, 40 },
		{ 493, 108, 40 },
		{ 493, 109, 40 },
		{ 494, 50, 40 },
		{ 494, 51, 40 },
		{ 494, 52, 40 },
		{ 494, 53, 40 },
		{ 494, 54, 40 },
		{ 494, 55, 40 },
		{ 494, 56, 40 },
		{ 494, 57, 40 },
		{ 494, 58, 40 },
		{ 494, 59, 40 },
		{ 494, 60, 40 },
		{ 494, 61, 40 },
		{ 494, 62, 40 },
		{ 494, 63, 40 },
		{ 494, 64, 40 },
		{ 494, 65, 40 },
		{ 494, 66, 40 },
		{ 494, 67, 40 },
		{ 494, 68, 40 },
		{ 494, 69, 40 },
		{ 494, 70, 40 },
		{ 494, 71, 40 },
		{ 494, 72, 40 },
		{ 494, 73, 40 },
		{ 494, 74, 40 },
		{ 494, 75, 40 },
		{ 494, 76, 40 },
		{ 494, 77, 40 },
		{ 494, 78, 40 },
		{ 494, 79, 40 },
		{ 494, 80, 40 },
		{ 494, 81, 40 },
		{ 494, 82, 40 },
		{ 494, 83, 40 },
		{ 494, 84, 40 },
		{ 494, 85, 40 },
		{ 494, 86, 40 },
		{ 494, 87, 40 },
		{ 494, 88, 40 },
		{ 494, 89, 40 },
		{ 494, 90, 40 },
		{ 494, 91, 40 },
		{ 494, 92, 40 },
		{ 494, 93, 40 },
		{ 494, 94, 40 },
		{ 494, 95, 40 },
		{ 494, 96, 40 },
		{ 494, 97, 40 },
		{ 494, 98, 40 },
		{ 494, 99, 40 },
		{ 494, 100, 40 },
		{ 494, 101, 40 },
		{ 494, 102, 40 },
		{ 494, 103, 40 },
		{ 494, 104, 40 },
		{ 494, 105, 40 },
		{ 494, 106, 40 },
		{ 494, 107, 40 },
		{ 494, 108, 40 },
		{ 495, 49, 40 },
		{ 495, 50, 40 },
		{ 495, 51, 40 },
		{ 495, 52, 40 },
		{ 495, 53, 40 },
		{ 495, 54, 40 },
		{ 495, 55, 40 },
		{ 495, 56, 40 },
		{ 495, 57, 40 },
		{ 495, 58, 40 },
		{ 495, 59, 40 },
		{ 495, 60, 40 },
		{ 495, 61, 40 },
		{ 495, 62, 40 },
		{ 495, 63, 40 },
		{ 495, 64, 40 },
		{ 495, 65, 40 },
		{ 495, 66, 40 },
		{ 495, 67, 40 },
		{ 495, 68, 40 },
		{ 495, 69, 40 },
		{ 495, 70, 40 },
		{ 495, 71, 40 },
		{ 495, 72, 40 },
		{ 495, 73, 40 },
		{ 495, 74, 40 },
		{ 495, 75, 40 },
		{ 495, 76, 40 },
		{ 495, 77, 40 },
		{ 495, 78, 40 },
		{ 495, 79, 40 },
		{ 495, 80, 40 },
		{ 495, 81, 40 },
		{ 495, 82, 40 },
		{ 495, 83, 40 },
		{ 495, 84, 40 },
		{ 495, 85, 40 },
		{ 495, 86, 40 },
		{ 495, 87, 40 },
		{ 495, 88, 40 },
		{ 495, 89, 40 },
		{ 495, 90, 40 },
		{ 495, 91, 40 },
		{ 495, 92, 40 },
		{ 495, 93, 40 },
		{ 495, 94, 40 },
		{ 495, 95, 40 },
		{ 495, 96, 40 },
		{ 495, 97, 40 },
		{ 495, 98, 40 },
		{ 495, 99, 40 },
		{ 495, 100, 40 },
		{ 495, 101, 40 },
		{ 495, 102, 40 },
		{ 495, 103, 40 },
		{ 495, 104, 40 },
		{ 495, 105, 40 },
		{ 495, 106, 40 },
		{ 495, 107, 40 },
		{ 495, 108, 40 },
		{ 496, 48, 40 },
		{ 496, 49, 40 },
		{ 496, 50, 40 },
		{ 496, 51, 40 },
		{ 496, 52, 40 },
		{ 496, 53, 40 },
		{ 496, 54, 40 },
		{ 496, 55, 40 },
		{ 496, 56, 40 },
		{ 496, 57, 40 },
		{ 496, 58, 40 },
		{ 496, 59, 40 },
		{ 496, 60, 40 },
		{ 496, 61, 40 },
		{ 496, 62, 40 },
		{ 496, 63, 40 },
		{ 496, 64, 40 },
		{ 496, 65, 40 },
		{ 496, 66, 40 },
		{ 496, 67, 40 },
		{ 496, 68, 40 },
		{ 496, 69, 40 },
		{ 496, 70, 40 },
		{ 496, 71, 40 },
		{ 496, 72, 40 },
		{ 496, 73, 40 },
		{ 496, 74, 40 },
		{ 496, 75, 40 },
		{ 496, 76, 40 },
		{ 496, 77, 40 },
		{ 496, 78, 40 },
		{ 496, 79, 40 },
		{ 496, 80, 40 },
		{ 496, 81, 40 },
		{ 496, 82, 40 },
		{ 496, 83, 40 },
		{ 496, 84, 40 },
		{ 496, 85, 40 },
		{ 496, 86, 40 },
		{ 496, 87, 40 },
		{ 496, 88, 40 },
		{ 496, 89, 40 },
		{ 496, 90, 40 },
		{ 496, 91, 40 },
		{ 496, 92, 40 },
		{ 496, 93, 40 },
		{ 496, 94, 40 },
		{ 496, 95, 40 },
		{ 496, 96, 40 },
		{ 496, 97, 40 },
		{ 496, 98, 40 },
		{ 496, 99, 40 },
		{ 496, 100, 40 },
		{ 496, 101, 40 },
		{ 496, 102, 40 },
		{ 496, 103, 40 },
		{ 496, 104, 40 },
		{ 496, 105, 40 },
		{ 496, 106, 40 },
		{ 496, 107, 40 },
		{ 497, 48, 40 },
		{ 497, 49, 40 },
		{ 497, 50, 40 },
		{ 497, 51, 40 },
		{ 497, 52, 40 },
		{ 497, 53, 40 },
		{ 497, 54, 40 },
		{ 497, 55, 40 },
		{ 497, 56, 40 },
		{ 497, 57, 40 },
		{ 497, 58, 40 },
		{ 497, 59, 40 },
		{ 497, 60, 40 },
		{ 497, 61, 40 },
		{ 497, 62, 40 },
		{ 497, 63, 40 },
		{ 497, 64, 40 },
		{ 497, 65, 40 },
		{ 497, 66, 40 },
		{ 497, 67, 40 },
		{ 497, 68, 40 },
		{ 497, 69, 40 },
		{ 497, 70, 40 },
		{ 497, 71, 40 },
		{ 497, 72, 40 },
		{ 497, 73, 40 },
		{ 497, 74, 40 },
		{ 497, 75, 40 },
		{ 497, 76, 40 },
		{ 497, 77, 40 },
		{ 497, 78, 40 },
		{ 497, 79, 40 },
		{ 497, 80, 40 },
		{ 497, 81, 40 },
		{ 497, 82, 40 },
		{ 497, 83, 40 },
		{ 497, 84, 40 },
		{ 497, 85, 40 },
		{ 497, 86, 40 },
		{ 497, 87, 40 },
		{ 497, 88, 40 },
		{ 497, 89, 40 },
		{ 497, 90, 40 },
		{ 497, 91, 40 },
		{ 497, 92, 40 },
		{ 497, 93, 40 },
		{ 497, 94, 40 },
		{ 497, 95, 40 },
		{ 497, 96, 40 },
		{ 497, 97, 40 },
		{ 497, 98, 40 },
		{ 497, 99, 40 },
		{ 497, 100, 40 },
		{ 497, 101, 40 },
		{ 497, 102, 40 },
		{ 497, 103, 40 },
		{ 497, 104, 40 },
		{ 497, 105, 40 },
		{ 497, 106, 40 },
		{ 498, 47, 40 },
		{ 498, 48, 40 },
		{ 498, 49, 40 },
		{ 498, 50, 40 },
		{ 498, 51, 40 },
		{ 498, 52, 40 },
		{ 498, 53, 40 },
		{ 498, 54, 40 },
		{ 498, 55, 40 },
		{ 498, 56, 40 },
		{ 498, 57, 40 },
		{ 498, 58, 40 },
		{ 498, 59, 40 },
		{ 498, 60, 40 },
		{ 498, 61, 40 },
		{ 498, 62, 40 },
		{ 498, 63, 40 },
		{ 498, 64, 40 },
		{ 498, 65, 40 },
		{ 498, 66, 40 },
		{ 498, 67, 40 },
		{ 498, 68, 40 },
		{ 498, 69, 40 },
		{ 498, 70, 40 },
		{ 498, 71, 40 },
		{ 498, 72, 40 },
		{ 498, 73, 40 },
		{ 498, 74, 40 },
		{ 498, 75, 40 },
		{ 498, 76, 40 },
		{ 498, 77, 40 },
		{ 498, 78, 40 },
		{ 498, 79, 40 },
		{ 498, 80, 40 },
		{ 498, 81, 40 },
		{ 498, 82, 40 },
		{ 498, 83, 40 },
		{ 498, 84, 40 },
		{ 498, 85, 40 },
		{ 498, 86, 40 },
		{ 498, 87, 40 },
		{ 498, 88, 40 },
		{ 498, 89, 40 },
		{ 498, 90, 40 },
		{ 498, 91, 40 },
		{ 498, 92, 40 },
		{ 498, 93, 40 },
		{ 498, 94, 40 },
		{ 498, 95, 40 },
		{ 498, 96, 40 },
		{ 498, 97, 40 },
		{ 498, 98, 40 },
		{ 498, 99, 40 },
		{ 498, 100, 40 },
		{ 498, 101, 40 },
		{ 498, 102, 40 },
		{ 498, 103, 40 },
		{ 498, 104, 40 },
		{ 498, 105, 40 },
		{ 498, 106, 40 },
		{ 499, 46, 40 },
		{ 499, 47, 40 },
		{ 499, 48, 40 },
		{ 499, 49, 40 },
		{ 499, 50, 40 },
		{ 499, 51, 40 },
		{ 499, 52, 40 },
		{ 499, 53, 40 },
		{ 499, 54, 40 },
		{ 499, 55, 40 },
		{ 499, 56, 40 },
		{ 499, 57, 40 },
		{ 499, 58, 40 },
		{ 499, 59, 40 },
		{ 499, 60, 40 },
		{ 499, 61, 40 },
		{ 499, 62, 40 },
		{ 499, 63, 40 },
		{ 499, 64, 40 },
		{ 499, 65, 40 },
		{ 499, 66, 40 },
		{ 499, 67, 40 },
		{ 499, 68, 40 },
		{ 499, 69, 40 },
		{ 499, 70, 40 },
		{ 499, 71, 40 },
		{ 499, 72, 40 },
		{ 499, 73, 40 },
		{ 499, 74, 40 },
		{ 499, 75, 40 },
		{ 499, 76, 40 },
		{ 499, 77, 40 },
		{ 499, 78, 40 },
		{ 499, 79, 40 },
		{ 499, 80, 40 },
		{ 499, 81, 40 },
		{ 499, 82, 40 },
		{ 499, 83, 40 },
		{ 499, 84, 40 },
		{ 499, 85, 40 },
		{ 499, 86, 40 },
		{ 499, 87, 40 },
		{ 499, 88, 40 },
		{ 499, 89, 40 },
		{ 499, 90, 40 },
		{ 499, 91, 40 },
		{ 499, 92, 40 },
		{ 499, 93, 40 },
		{ 499, 94, 40 },
		{ 499, 95, 40 },
		{ 499, 96, 40 },
		{ 499, 97, 40 },
		{ 499, 98, 40 },
		{ 499, 99, 40 },
		{ 499, 100, 40 },
		{ 499, 101, 40 },
		{ 499, 102, 40 },
		{ 499, 103, 40 },
		{ 499, 104, 40 },
		{ 499, 105, 40 },
		{ 500, 45, 40 },
		{ 500, 46, 40 },
		{ 500, 47, 40 },
		{ 500, 48, 40 },
		{ 500, 49, 40 },
		{ 500, 50, 40 },
		{ 500, 51, 40 },
		{ 500, 52, 40 },
		{ 500, 53, 40 },
		{ 500, 54, 40 },
		{ 500, 55, 40 },
		{ 500, 56, 40 },
		{ 500, 57, 40 },
		{ 500, 58, 40 },
		{ 500, 59, 40 },
		{ 500, 60, 40 },
		{ 500, 61, 40 },
		{ 500, 62, 40 },
		{ 500, 63, 40 },
		{ 500, 64, 40 },
		{ 500, 65, 40 },
		{ 500, 66, 40 },
		{ 500, 67, 40 },
		{ 500, 68, 40 },
		{ 500, 69, 40 },
		{ 500, 70, 40 },
		{ 500, 71, 40 },
		{ 500, 72, 40 },
		{ 500, 73, 40 },
		{ 500, 74, 40 },
		{ 500, 75, 40 },
		{ 500, 76, 40 },
		{ 500, 77, 40 },
		{ 500, 78, 40 },
		{ 500, 79, 40 },
		{ 500, 80, 40 },
		{ 500, 81, 40 },
		{ 500, 82, 40 },
		{ 500, 83, 40 },
		{ 500, 84, 40 },
		{ 500, 85, 40 },
		{ 500, 86, 40 },
		{ 500, 87, 40 },
		{ 500, 88, 40 },
		{ 500, 89, 40 },
		{ 500, 90, 40 },
		{ 500, 91, 40 },
		{ 500, 92, 40 },
		{ 500, 93, 40 },
		{ 500, 94, 40 },
		{ 500, 95, 40 },
		{ 500, 96, 40 },
		{ 500, 97, 40 },
		{ 500, 98, 40 },
		{ 500, 99, 40 },
		{ 500, 100, 40 },
		{ 500, 101, 40 },
		{ 500, 102, 40 },
		{ 500, 103, 40 },
		{ 500, 104, 40 },
		{ 500, 105, 40 },
		{ 501, 44, 40 },
		{ 501, 45, 40 },
		{ 501, 46, 40 },
		{ 501, 47, 40 },
		{ 501, 48, 40 },
		{ 501, 49, 40 },
		{ 501, 50, 40 },
		{ 501, 51, 40 },
		{ 501, 52, 40 },
		{ 501, 53, 40 },
		{ 501, 54, 40 },
		{ 501, 55, 40 },
		{ 501, 56, 40 },
		{ 501, 57, 40 },
		{ 501, 58, 40 },
		{ 501, 59, 40 },
		{ 501, 60, 40 },
		{ 501, 61, 40 },
		{ 501, 62, 40 },
		{ 501, 63, 40 },
		{ 501, 64, 40 },
		{ 501, 65, 40 },
		{ 501, 66, 40 },
		{ 501, 67, 40 },
		{ 501, 68, 40 },
		{ 501, 69, 40 },
		{ 501, 70, 40 },
		{ 501, 71, 40 },
		{ 501, 72, 40 },
		{ 501, 73, 40 },
		{ 501, 74, 40 },
		{ 501, 75, 40 },
		{ 501, 76, 40 },
		{ 501, 77, 40 },
		{ 501, 78, 40 },
		{ 501, 79, 40 },
		{ 501, 80, 40 },
		{ 501, 81, 40 },
		{ 501, 82, 40 },
		{ 501, 83, 40 },
		{ 501, 84, 40 },
		{ 501, 85, 40 },
		{ 501, 86, 40 },
		{ 501, 87, 40 },
		{ 501, 88, 40 },
		{ 501, 89, 40 },
		{ 501, 90, 40 },
		{ 501, 91, 40 },
		{ 501, 92, 40 },
		{ 501, 93, 40 },
		{ 501, 94, 40 },
		{ 501, 95, 40 },
		{ 501, 96, 40 },
		{ 501, 97, 40 },
		{ 501, 98, 40 },
		{ 501, 99, 40 },
		{ 501, 100, 40 },
		{ 501, 101, 40 },
		{ 501, 102, 40 },
		{ 501, 103, 40 },
		{ 501, 104, 40 },
		{ 502, 43, 40 },
		{ 502, 44, 40 },
		{ 502, 45, 40 },
		{ 502, 46, 40 },
		{ 502, 47, 40 },
		{ 502, 48, 40 },
		{ 502, 49, 40 },
		{ 502, 50, 40 },
		{ 502, 51, 40 },
		{ 502, 52, 40 },
		{ 502, 53, 40 },
		{ 502, 54, 40 },
		{ 502, 55, 40 },
		{ 502, 56, 40 },
		{ 502, 57, 40 },
		{ 502, 58, 40 },
		{ 502, 59, 40 },
		{ 502, 60, 40 },
		{ 502, 61, 40 },
		{ 502, 62, 40 },
		{ 502, 63, 40 },
		{ 502, 64, 40 },
		{ 502, 65, 40 },
		{ 502, 66, 40 },
		{ 502, 67, 40 },
		{ 502, 68, 40 },
		{ 502, 69, 40 },
		{ 502, 70, 40 },
		{ 502, 71, 40 },
		{ 502, 72, 40 },
		{ 502, 73, 40 },
		{ 502, 74, 40 },
		{ 502, 75, 40 },
		{ 502, 76, 40 },
		{ 502, 77, 40 },
		{ 502, 78, 40 },
		{ 502, 79, 40 },
		{ 502, 80, 40 },
		{ 502, 81, 40 },
		{ 502, 82, 40 },
		{ 502, 83, 40 },
		{ 502, 84, 40 },
		{ 502, 85, 40 },
		{ 502, 86, 40 },
		{ 502, 87, 40 },
		{ 502, 88, 40 },
		{ 502, 89, 40 },
		{ 502, 90, 40 },
		{ 502, 91, 40 },
		{ 502, 92, 40 },
		{ 502, 93, 40 },
		{ 502, 94, 40 },
		{ 502, 95, 40 },
		{ 502, 96, 40 },
		{ 502, 97, 40 },
		{ 502, 98, 40 },
		{ 502, 99, 40 },
		{ 502, 100, 40 },
		{ 502, 101, 40 },
		{ 502, 102, 40 },
		{ 502, 103, 40 },
		{ 503, 42, 40 },
		{ 503, 43, 40 },
		{ 503, 44, 40 },
		{ 503, 45, 40 },
		{ 503, 46, 40 },
		{ 503, 47, 40 },
		{ 503, 48, 40 },
		{ 503, 49, 40 },
		{ 503, 50, 40 },
		{ 503, 51, 40 },
		{ 503, 52, 40 },
		{ 503, 53, 40 },
		{ 503, 54, 40 },
		{ 503, 55, 40 },
		{ 503, 56, 40 },
		{ 503, 57, 40 },
		{ 503, 58, 40 },
		{ 503, 59, 40 },
		{ 503, 60, 40 },
		{ 503, 61, 40 },
		{ 503, 62, 40 },
		{ 503, 63, 40 },
		{ 503, 64, 40 },
		{ 503, 65, 40 },
		{ 503, 66, 40 },
		{ 503, 67, 40 },
		{ 503, 68, 40 },
		{ 503, 69, 40 },
		{ 503, 70, 40 },
		{ 503, 71, 40 },
		{ 503, 72, 40 },
		{ 503, 73, 40 },
		{ 503, 74, 40 },
		{ 503, 75, 40 },
		{ 503, 76, 40 },
		{ 503, 77, 40 },
		{ 503, 78, 40 },
		{ 503, 79, 40 },
		{ 503, 80, 40 },
		{ 503, 81, 40 },
		{ 503, 82, 40 },
		{ 503, 83, 40 },
		{ 503, 84, 40 },
		{ 503, 85, 40 },
		{ 503, 86, 40 },
		{ 503, 87, 40 },
		{ 503, 88, 40 },
		{ 503, 89, 40 },
		{ 503, 90, 40 },
		{ 503, 91, 40 },
		{ 503, 92, 40 },
		{ 503, 93, 40 },
		{ 503, 94, 40 },
		{ 503, 95, 40 },
		{ 503, 96, 40 },
		{ 503, 97, 40 },
		{ 503, 98, 40 },
		{ 503, 99, 40 },
		{ 503, 100, 40 },
		{ 503, 101, 40 },
		{ 503, 102, 40 },
		{ 503, 103, 40 },
		{ 504, 41, 40 },
		{ 504, 42, 40 },
		{ 504, 43, 40 },
		{ 504, 44, 40 },
		{ 504, 45, 40 },
		{ 504, 46, 40 },
		{ 504, 47, 40 },
		{ 504, 48, 40 },
		{ 504, 49, 40 },
		{ 504, 50, 40 },
		{ 504, 51, 40 },
		{ 504, 52, 40 },
		{ 504, 53, 40 },
		{ 504, 54, 40 },
		{ 504, 55, 40 },
		{ 504, 56, 40 },
		{ 504, 57, 40 },
		{ 504, 58, 40 },
		{ 504, 59, 40 },
		{ 504, 60, 40 },
		{ 504, 61, 40 },
		{ 504, 62, 40 },
		{ 504, 63, 40 },
		{ 504, 64, 40 },
		{ 504, 65, 40 },
		{ 504, 66, 40 },
		{ 504, 67, 40 },
		{ 504, 68, 40 },
		{ 504, 69, 40 },
		{ 504, 70, 40 },
		{ 504, 71, 40 },
		{ 504, 72, 40 },
		{ 504, 73, 40 },
		{ 504, 74, 40 },
		{ 504, 75, 40 },
		{ 504, 76, 40 },
		{ 504, 77, 40 },
		{ 504, 78, 40 },
		{ 504, 79, 40 },
		{ 504, 80, 40 },
		{ 504, 81, 40 },
		{ 504, 82, 40 },
		{ 504, 83, 40 },
		{ 504, 84, 40 },
		{ 504, 85, 40 },
		{ 504, 86, 40 },
		{ 504, 87, 40 },
		{ 504, 88, 40 },
		{ 504, 89, 40 },
		{ 504, 90, 40 },
		{ 504, 91, 40 },
		{ 504, 92, 40 },
		{ 504, 93, 40 },
		{ 504, 94, 40 },
		{ 504, 95, 40 },
		{ 504, 96, 40 },
		{ 504, 97, 40 },
		{ 504, 98, 40 },
		{ 504, 99, 40 },
		{ 504, 100, 40 },
		{ 504, 101, 40 },
		{ 504, 102, 40 },
		{ 505, 41, 40 },
		{ 505, 42, 40 },
		{ 505, 43, 40 },
		{ 505, 44, 40 },
		{ 505, 45, 40 },
		{ 505, 46, 40 },
		{ 505, 47, 40 },
		{ 505, 48, 40 },
		{ 505, 49, 40 },
		{ 505, 50, 40 },
		{ 505, 51, 40 },
		{ 505, 52, 40 },
		{ 505, 53, 40 },
		{ 505, 54, 40 },
		{ 505, 55, 40 },
		{ 505, 56, 40 },
		{ 505, 57, 40 },
		{ 505, 58, 40 },
		{ 505, 59, 40 },
		{ 505, 60, 40 },
		{ 505, 61, 40 },
		{ 505, 62, 40 },
		{ 505, 63, 40 },
		{ 505, 64, 40 },
		{ 505, 65, 40 },
		{ 505, 66, 40 },
		{ 505, 67, 40 },
		{ 505, 68, 40 },
		{ 505, 69, 40 },
		{ 505, 70, 40 },
		{ 505, 71, 40 },
		{ 505, 72, 40 },
		{ 505, 73, 40 },
		{ 505, 74, 40 },
		{ 505, 75, 40 },
		{ 505, 76, 40 },
		{ 505, 77, 40 },
		{ 505, 78, 40 },
		{ 505, 79, 40 },
		{ 505, 80, 40 },
		{ 505, 81, 40 },
		{ 505, 82, 40 },
		{ 505, 83, 40 },
		{ 505, 84, 40 },
		{ 505, 85, 40 },
		{ 505, 86, 40 },
		{ 505, 87, 40 },
		{ 505, 88, 40 },
		{ 505, 89, 40 },
		{ 505, 90, 40 },
		{ 505, 91, 40 },
		{ 505, 92, 40 },
		{ 505, 93, 40 },
		{ 505, 94, 40 },
		{ 505, 95, 40 },
		{ 505, 96, 40 },
		{ 505, 97, 40 },
		{ 505, 98, 40 },
		{ 505, 99, 40 },
		{ 505, 100, 40 },
		{ 505, 101, 40 },
		{ 506, 40, 40 },
		{ 506, 41, 40 },
		{ 506, 42, 40 },
		{ 506, 43, 40 },
		{ 506, 44, 40 },
		{ 506, 45, 40 },
		{ 506, 46, 40 },
		{ 506, 47, 40 },
		{ 506, 48, 40 },
		{ 506, 49, 40 },
		{ 506, 50, 40 },
		{ 506, 51, 40 },
		{ 506, 52, 40 },
		{ 506, 53, 40 },
		{ 506, 54, 40 },
		{ 506, 55, 40 },
		{ 506, 56, 40 },
		{ 506, 57, 40 },
		{ 506, 58, 40 },
		{ 506, 59, 40 },
		{ 506, 60, 40 },
		{ 506, 61, 40 },
		{ 506, 62, 40 },
		{ 506, 63, 40 },
		{ 506, 64, 40 },
		{ 506, 65, 40 },
		{ 506, 66, 40 },
		{ 506, 67, 40 },
		{ 506, 68, 40 },
		{ 506, 69, 40 },
		{ 506, 70, 40 },
		{ 506, 71, 40 },
		{ 506, 72, 40 },
		{ 506, 73, 40 },
		{ 506, 74, 40 },
		{ 506, 75, 40 },
		{ 506, 76, 40 },
		{ 506, 77, 40 },
		{ 506, 78, 40 },
		{ 506, 79, 40 },
		{ 506, 80, 40 },
		{ 506, 81, 40 },
		{ 506, 82, 40 },
		{ 506, 83, 40 },
		{ 506, 84, 40 },
		{ 506, 85, 40 },
		{ 506, 86, 40 },
		{ 506, 87, 40 },
		{ 506, 88, 40 },
		{ 506, 89, 40 },
		{ 506, 90, 40 },
		{ 506, 91, 40 },
		{ 506, 92, 40 },
		{ 506, 93, 40 },
		{ 506, 94, 40 },
		{ 506, 95, 40 },
		{ 506, 96, 40 },
		{ 506, 97, 40 },
		{ 506, 98, 40 },
		{ 506, 99, 40 },
		{ 506, 100, 40 },
		{ 506, 101, 40 },
		{ 507, 39, 40 },
		{ 507, 40, 40 },
		{ 507, 41, 40 },
		{ 507, 42, 40 },
		{ 507, 43, 40 },
		{ 507, 44, 40 },
		{ 507, 45, 40 },
		{ 507, 46, 40 },
		{ 507, 47, 40 },
		{ 507, 48, 40 },
		{ 507, 49, 40 },
		{ 507, 50, 40 },
		{ 507, 51, 40 },
		{ 507, 52, 40 },
		{ 507, 53, 40 },
		{ 507, 54, 40 },
		{ 507, 55, 40 },
		{ 507, 56, 40 },
		{ 507, 57, 40 },
		{ 507, 58, 40 },
		{ 507, 59, 40 },
		{ 507, 60, 40 },
		{ 507, 61, 40 },
		{ 507, 62, 40 },
		{ 507, 63, 40 },
		{ 507, 64, 40 },
		{ 507, 65, 40 },
		{ 507, 66, 40 },
		{ 507, 67, 40 },
		{ 507, 68, 40 },
		{ 507, 69, 40 },
		{ 507, 70, 40 },
		{ 507, 71, 40 },
		{ 507, 72, 40 },
		{ 507, 73, 40 },
		{ 507, 74, 40 },
		{ 507, 75, 40 },
		{ 507, 76, 40 },
		{ 507, 77, 40 },
		{ 507, 78, 40 },
		{ 507, 79, 40 },
		{ 507, 80, 40 },
		{ 507, 81, 40 },
		{ 507, 82, 40 },
		{ 507, 83, 40 },
		{ 507, 84, 40 },
		{ 507, 85, 40 },
		{ 507, 86, 40 },
		{ 507, 87, 40 },
		{ 507, 88, 40 },
		{ 507, 89, 40 },
		{ 507, 90, 40 },
		{ 507, 91, 40 },
		{ 507, 92, 40 },
		{ 507, 93, 40 },
		{ 507, 94, 40 },
		{ 507, 95, 40 },
		{ 507, 96, 40 },
		{ 507, 97, 40 },
		{ 507, 98, 40 },
		{ 507, 99, 40 },
		{ 507, 100, 40 },
		{ 508, 38, 40 },
		{ 508, 39, 40 },
		{ 508, 40, 40 },
		{ 508, 41, 40 },
		{ 508, 42, 40 },
		{ 508, 43, 40 },
		{ 508, 44, 40 },
		{ 508, 45, 40 },
		{ 508, 46, 40 },
		{ 508, 47, 40 },
		{ 508, 48, 40 },
		{ 508, 49, 40 },
		{ 508, 50, 40 },
		{ 508, 51, 40 },
		{ 508, 52, 40 },
		{ 508, 53, 40 },
		{ 508, 54, 40 },
		{ 508, 55, 40 },
		{ 508, 56, 40 },
		{ 508, 57, 40 },
		{ 508, 58, 40 },
		{ 508, 59, 40 },
		{ 508, 60, 40 },
		{ 508, 61, 40 },
		{ 508, 62, 40 },
		{ 508, 63, 40 },
		{ 508, 64, 40 },
		{ 508, 65, 40 },
		{ 508, 66, 40 },
		{ 508, 67, 40 },
		{ 508, 68, 40 },
		{ 508, 69, 40 },
		{ 508, 70, 40 },
		{ 508, 71, 40 },
		{ 508, 72, 40 },
		{ 508, 73, 40 },
		{ 508, 74, 40 },
		{ 508, 75, 40 },
		{ 508, 76, 40 },
		{ 508, 77, 40 },
		{ 508, 78, 40 },
		{ 508, 79, 40 },
		{ 508, 80, 40 },
		{ 508, 81, 40 },
		{ 508, 82, 40 },
		{ 508, 83, 40 },
		{ 508, 84, 40 },
		{ 508, 85, 40 },
		{ 508, 86, 40 },
		{ 508, 87, 40 },
		{ 508, 88, 40 },
		{ 508, 89, 40 },
		{ 508, 90, 40 },
		{ 508, 91, 40 },
		{ 508, 92, 40 },
		{ 508, 93, 40 },
		{ 508, 94, 40 },
		{ 508, 95, 40 },
		{ 508, 96, 40 },
		{ 508, 97, 40 },
		{ 508, 98, 40 },
		{ 508, 99, 40 },
		{ 508, 100, 40 },
		{ 509, 37, 40 },
		{ 509, 38, 40 },
		{ 509, 39, 40 },
		{ 509, 40, 40 },
		{ 509, 41, 40 },
		{ 509, 42, 40 },
		{ 509, 43, 40 },
		{ 509, 44, 40 },
		{ 509, 45, 40 },
		{ 509, 46, 40 },
		{ 509, 47, 40 },
		{ 509, 48, 40 },
		{ 509, 49, 40 },
		{ 509, 50, 40 },
		{ 509, 51, 40 },
		{ 509, 52, 40 },
		{ 509, 53, 40 },
		{ 509, 54, 40 },
		{ 509, 55, 40 },
		{ 509, 56, 40 },
		{ 509, 57, 40 },
		{ 509, 58, 40 },
		{ 509, 59, 40 },
		{ 509, 60, 40 },
		{ 509, 61, 40 },
		{ 509, 62, 40 },
		{ 509, 63, 40 },
		{ 509, 64, 40 },
		{ 509, 65, 40 },
		{ 509, 66, 40 },
		{ 509, 67, 40 },
		{ 509, 68, 40 },
		{ 509, 69, 40 },
		{ 509, 70, 40 },
		{ 509, 71, 40 },
		{ 509, 72, 40 },
		{ 509, 73, 40 },
		{ 509, 74, 40 },
		{ 509, 75, 40 },
		{ 509, 76, 40 },
		{ 509, 77, 40 },
		{ 509, 78, 40 },
		{ 509, 79, 40 },
		{ 509, 80, 40 },
		{ 509, 81, 40 },
		{ 509, 82, 40 },
		{ 509, 83, 40 },
		{ 509, 84, 40 },
		{ 509, 85, 40 },
		{ 509, 86, 40 },
		{ 509, 87, 40 },
		{ 509, 88, 40 },
		{ 509, 89, 40 },
		{ 509, 90, 40 },
		{ 509, 91, 40 },
		{ 509, 92, 40 },
		{ 509, 93, 40 },
		{ 509, 94, 40 },
		{ 509, 95, 40 },
		{ 509, 96, 40 },
		{ 509, 97, 40 },
		{ 509, 98, 40 },
		{ 509, 99, 40 },
		{ 510, 36, 40 },
		{ 510, 37, 40 },
		{ 510, 38, 40 },
		{ 510, 39, 40 },
		{ 510, 40, 40 },
		{ 510, 41, 40 },
		{ 510, 42, 40 },
		{ 510, 43, 40 },
		{ 510, 44, 40 },
		{ 510, 45, 40 },
		{ 510, 46, 40 },
		{ 510, 47, 40 },
		{ 510, 48, 40 },
		{ 510, 49, 40 },
		{ 510, 50, 40 },
		{ 510, 51, 40 },
		{ 510, 52, 40 },
		{ 510, 53, 40 },
		{ 510, 54, 40 },
		{ 510, 55, 40 },
		{ 510, 56, 40 },
		{ 510, 57, 40 },
		{ 510, 58, 40 },
		{ 510, 59, 40 },
		{ 510, 60, 40 },
		{ 510, 61, 40 },
		{ 510, 62, 40 },
		{ 510, 63, 40 },
		{ 510, 64, 40 },
		{ 510, 65, 40 },
		{ 510, 66, 40 },
		{ 510, 67, 40 },
		{ 510, 68, 40 },
		{ 510, 69, 40 },
		{ 510, 70, 40 },
		{ 510, 71, 40 },
		{ 510, 72, 40 },
		{ 510, 73, 40 },
		{ 510, 74, 40 },
		{ 510, 75, 40 },
		{ 510, 76, 40 },
		{ 510, 77, 40 },
		{ 510, 78, 40 },
		{ 510, 79, 40 },
		{ 510, 80, 40 },
		{ 510, 81, 40 },
		{ 510, 82, 40 },
		{ 510, 83, 40 },
		{ 510, 84, 40 },
		{ 510, 85, 40 },
		{ 510, 86, 40 },
		{ 510, 87, 40 },
		{ 510, 88, 40 },
		{ 510, 89, 40 },
		{ 510, 90, 40 },
		{ 510, 91, 40 },
		{ 510, 92, 40 },
		{ 510, 93, 40 },
		{ 510, 94, 40 },
		{ 510, 95, 40 },
		{ 510, 96, 40 },
		{ 510, 97, 40 },
		{ 510, 98, 40 },
		{ 511, 35, 40 },
		{ 511, 36, 40 },
		{ 511, 37, 40 },
		{ 511, 38, 40 },
		{ 511, 39, 40 },
		{ 511, 40, 40 },
		{ 511, 41, 40 },
		{ 511, 42, 40 },
		{ 511, 43, 40 },
		{ 511, 44, 40 },
		{ 511, 45, 40 },
		{ 511, 46, 40 },
		{ 511, 47, 40 },
		{ 511, 48, 40 },
		{ 511, 49, 40 },
		{ 511, 50, 40 },
		{ 511, 51, 40 },
		{ 511, 52, 40 },
		{ 511, 53, 40 },
		{ 511, 54, 40 },
		{ 511, 55, 40 },
		{ 511, 56, 40 },
		{ 511, 57, 40 },
		{ 511, 58, 40 },
		{ 511, 59, 40 },
		{ 511, 60, 40 },
		{ 511, 61, 40 },
		{ 511, 62, 40 },
		{ 511, 63, 40 },
		{ 511, 64, 40 },
		{ 511, 65, 40 },
		{ 511, 66, 40 },
		{ 511, 67, 40 },
		{ 511, 68, 40 },
		{ 511, 69, 40 },
		{ 511, 70, 40 },
		{ 511, 71, 40 },
		{ 511, 72, 40 },
		{ 511, 73, 40 },
		{ 511, 74, 40 },
		{ 511, 75, 40 },
		{ 511, 76, 40 },
		{ 511, 77, 40 },
		{ 511, 78, 40 },
		{ 511, 79, 40 },
		{ 511, 80, 40 },
		{ 511, 81, 40 },
		{ 511, 82, 40 },
		{ 511, 83, 40 },
		{ 511, 84, 40 },
		{ 511, 85, 40 },
		{ 511, 86, 40 },
		{ 511, 87, 40 },
		{ 511, 88, 40 },
		{ 511, 89, 40 },
		{ 511, 90, 40 },
		{ 511, 91, 40 },
		{ 511, 92, 40 },
		{ 511, 93, 40 },
		{ 511, 94, 40 },
		{ 511, 95, 40 },
		{ 511, 96, 40 },
		{ 511, 97, 40 },
		{ 511, 98, 40 },
		{ 512, 34, 40 },
		{ 512, 35, 40 },
		{ 512, 36, 40 },
		{ 512, 37, 40 },
		{ 512, 38, 40 },
		{ 512, 39, 40 },
		{ 512, 40, 40 },
		{ 512, 41, 40 },
		{ 512, 42, 40 },
		{ 512, 43, 40 },
		{ 512, 44, 40 },
		{ 512, 45, 40 },
		{ 512, 46, 40 },
		{ 512, 47, 40 },
		{ 512, 48, 40 },
		{ 512, 49, 40 },
		{ 512, 50, 40 },
		{ 512, 51, 40 },
		{ 512, 52, 40 },
		{ 512, 53, 40 },
		{ 512, 54, 40 },
		{ 512, 55, 40 },
		{ 512, 56, 40 },
		{ 512, 57, 40 },
		{ 512, 58, 40 },
		{ 512, 59, 40 },
		{ 512, 60, 40 },
		{ 512, 61, 40 },
		{ 512, 62, 40 },
		{ 512, 63, 40 },
		{ 512, 64, 40 },
		{ 512, 65, 40 },
		{ 512, 66, 40 },
		{ 512, 67, 40 },
		{ 512, 68, 40 },
		{ 512, 69, 40 },
		{ 512, 70, 40 },
		{ 512, 71, 40 },
		{ 512, 72, 40 },
		{ 512, 73, 40 },
		{ 512, 74, 40 },
		{ 512, 75, 40 },
		{ 512, 76, 40 },
		{ 512, 77, 40 },
		{ 512, 78, 40 },
		{ 512, 79, 40 },
		{ 512, 80, 40 },
		{ 512, 81, 40 },
		{ 512, 82, 40 },
		{ 512, 83, 40 },
		{ 512, 84, 40 },
		{ 512, 85, 40 },
		{ 512, 86, 40 },
		{ 512, 87, 40 },
		{ 512, 88, 40 },
		{ 512, 89, 40 },
		{ 512, 90, 40 },
		{ 512, 91, 40 },
		{ 512, 92, 40 },
		{ 512, 93, 40 },
		{ 512, 94, 40 },
		{ 512, 95, 40 },
		{ 512, 96, 40 },
		{ 512, 97, 40 },
		{ 1, 386, 41 },
		{ 1, 387, 41 },
		{ 1, 388, 41 },
		{ 1, 389, 41 },
		{ 1, 390, 41 },
		{ 1, 391, 41 },
		{ 1, 392, 41 },
		{ 1, 393, 41 },
		{ 1, 394, 41 },
		{ 1, 395, 41 },
		{ 1, 396, 41 },
		{ 1, 397, 41 },
		{ 1, 398, 41 },
		{ 1, 399, 41 },
		{ 1, 400, 41 },
		{ 1, 401, 41 },
		{ 1, 402, 41 },
		{ 1, 403, 41 },
		{ 1, 404, 41 },
		{ 1, 405, 41 },
		{ 1, 406, 41 },
		{ 1, 407, 41 },
		{ 1, 408, 41 },
		{ 1, 409, 41 },
		{ 1, 410, 41 },
		{ 1, 411, 41 },
		{ 1, 412, 41 },
		{ 1, 413, 41 },
		{ 1, 414, 41 },
		{ 1, 415, 41 },
		{ 1, 416, 41 },
		{ 1, 417, 41 },
		{ 1, 418, 41 },
		{ 1, 419, 41 },
		{ 1, 420, 41 },
		{ 1, 421, 41 },
		{ 1, 422, 41 },
		{ 1, 423, 41 },
		{ 1, 424, 41 },
		{ 1, 425, 41 },
		{ 1, 426, 41 },
		{ 1, 427, 41 },
		{ 1, 428, 41 },
		{ 1, 429, 41 },
		{ 1, 430, 41 },
		{ 1, 431, 41 },
		{ 1, 432, 41 },
		{ 1, 433, 41 },
		{ 1, 434, 41 },
		{ 1, 435, 41 },
		{ 1, 436, 41 },
		{ 1, 437, 41 },
		{ 1, 438, 41 },
		{ 1, 439, 41 },
		{ 1, 440, 41 },
		{ 1, 441, 41 },
		{ 1, 442, 41 },
		{ 1, 443, 41 },
		{ 1, 444, 41 },
		{ 1, 445, 41 },
		{ 1, 446, 41 },
		{ 1, 447, 41 },
		{ 1, 448, 41 },
		{ 2, 385, 41 },
		{ 2, 386, 41 },
		{ 2, 387, 41 },
		{ 2, 388, 41 },
		{ 2, 389, 41 },
		{ 2, 390, 41 },
		{ 2, 391, 41 },
		{ 2, 392, 41 },
		{ 2, 393, 41 },
		{ 2, 394, 41 },
		{ 2, 395, 41 },
		{ 2, 396, 41 },
		{ 2, 397, 41 },
		{ 2, 398, 41 },
		{ 2, 399, 41 },
		{ 2, 400, 41 },
		{ 2, 401, 41 },
		{ 2, 402, 41 },
		{ 2, 403, 41 },
		{ 2, 404, 41 },
		{ 2, 405, 41 },
		{ 2, 406, 41 },
		{ 2, 407, 41 },
		{ 2, 408, 41 },
		{ 2, 409, 41 },
		{ 2, 410, 41 },
		{ 2, 411, 41 },
		{ 2, 412, 41 },
		{ 2, 413, 41 },
		{ 2, 414, 41 },
		{ 2, 415, 41 },
		{ 2, 416, 41 },
		{ 2, 417, 41 },
		{ 2, 418, 41 },
		{ 2, 419, 41 },
		{ 2, 420, 41 },
		{ 2, 421, 41 },
		{ 2, 422, 41 },
		{ 2, 423, 41 },
		{ 2, 424, 41 },
		{ 2, 425, 41 },
		{ 2, 426, 41 },
		{ 2, 427, 41 },
		{ 2, 428, 41 },
		{ 2, 429, 41 },
		{ 2, 430, 41 },
		{ 2, 431, 41 },
		{ 2, 432, 41 },
		{ 2, 433, 41 },
		{ 2, 434, 41 },
		{ 2, 435, 41 },
		{ 2, 436, 41 },
		{ 2, 437, 41 },
		{ 2, 438, 41 },
		{ 2, 439, 41 },
		{ 2, 440, 41 },
		{ 2, 441, 41 },
		{ 2, 442, 41 },
		{ 2, 443, 41 },
		{ 2, 444, 41 },
		{ 2, 445, 41 },
		{ 2, 446, 41 },
		{ 2, 447, 41 },
		{ 2, 448, 41 },
		{ 3, 385, 41 },
		{ 3, 386, 41 },
		{ 3, 387, 41 },
		{ 3, 388, 41 },
		{ 3, 389, 41 },
		{ 3, 390, 41 },
		{ 3, 391, 41 },
		{ 3, 392, 41 },
		{ 3, 393, 41 },
		{ 3, 394, 41 },
		{ 3, 395, 41 },
		{ 3, 396, 41 },
		{ 3, 397, 41 },
		{ 3, 398, 41 },
		{ 3, 399, 41 },
		{ 3, 400, 41 },
		{ 3, 401, 41 },
		{ 3, 402, 41 },
		{ 3, 403, 41 },
		{ 3, 404, 41 },
		{ 3, 405, 41 },
		{ 3, 406, 41 },
		{ 3, 407, 41 },
		{ 3, 408, 41 },
		{ 3, 409, 41 },
		{ 3, 410, 41 },
		{ 3, 411, 41 },
		{ 3, 412, 41 },
		{ 3, 413, 41 },
		{ 3, 414, 41 },
		{ 3, 415, 41 },
		{ 3, 416, 41 },
		{ 3, 417, 41 },
		{ 3, 418, 41 },
		{ 3, 419, 41 },
		{ 3, 420, 41 },
		{ 3, 421, 41 },
		{ 3, 422, 41 },
		{ 3, 423, 41 },
		{ 3, 424, 41 },
		{ 3, 425, 41 },
		{ 3, 426, 41 },
		{ 3, 427, 41 },
		{ 3, 428, 41 },
		{ 3, 429, 41 },
		{ 3, 430, 41 },
		{ 3, 431, 41 },
		{ 3, 432, 41 },
		{ 3, 433, 41 },
		{ 3, 434, 41 },
		{ 3, 435, 41 },
		{ 3, 436, 41 },
		{ 3, 437, 41 },
		{ 3, 438, 41 },
		{ 3, 439, 41 },
		{ 3, 440, 41 },
		{ 3, 441, 41 },
		{ 3, 442, 41 },
		{ 3, 443, 41 },
		{ 3, 444, 41 },
		{ 3, 445, 41 },
		{ 3, 446, 41 },
		{ 3, 447, 41 },
		{ 4, 384, 41 },
		{ 4, 385, 41 },
		{ 4, 386, 41 },
		{ 4, 387, 41 },
		{ 4, 388, 41 },
		{ 4, 389, 41 },
		{ 4, 390, 41 },
		{ 4, 391, 41 },
		{ 4, 392, 41 },
		{ 4, 393, 41 },
		{ 4, 394, 41 },
		{ 4, 395, 41 },
		{ 4, 396, 41 },
		{ 4, 397, 41 },
		{ 4, 398, 41 },
		{ 4, 399, 41 },
		{ 4, 400, 41 },
		{ 4, 401, 41 },
		{ 4, 402, 41 },
		{ 4, 403, 41 },
		{ 4, 404, 41 },
		{ 4, 405, 41 },
		{ 4, 406, 41 },
		{ 4, 407, 41 },
		{ 4, 408, 41 },
		{ 4, 409, 41 },
		{ 4, 410, 41 },
		{ 4, 411, 41 },
		{ 4, 412, 41 },
		{ 4, 413, 41 },
		{ 4, 414, 41 },
		{ 4, 415, 41 },
		{ 4, 416, 41 },
		{ 4, 417, 41 },
		{ 4, 418, 41 },
		{ 4, 419, 41 },
		{ 4, 420, 41 },
		{ 4, 421, 41 },
		{ 4, 422, 41 },
		{ 4, 423, 41 },
		{ 4, 424, 41 },
		{ 4, 425, 41 },
		{ 4, 426, 41 },
		{ 4, 427, 41 },
		{ 4, 428, 41 },
		{ 4, 429, 41 },
		{ 4, 430, 41 },
		{ 4, 431, 41 },
		{ 4, 432, 41 },
		{ 4, 433, 41 },
		{ 4, 434, 41 },
		{ 4, 435, 41 },
		{ 4, 436, 41 },
		{ 4, 437, 41 },
		{ 4, 438, 41 },
		{ 4, 439, 41 },
		{ 4, 440, 41 },
		{ 4, 441, 41 },
		{ 4, 442, 41 },
		{ 4, 443, 41 },
		{ 4, 444, 41 },
		{ 4, 445, 41 },
		{ 4, 446, 41 },
		{ 5, 384, 41 },
		{ 5, 385, 41 },
		{ 5, 386, 41 },
		{ 5, 387, 41 },
		{ 5, 388, 41 },
		{ 5, 389, 41 },
		{ 5, 390, 41 },
		{ 5, 391, 41 },
		{ 5, 392, 41 },
		{ 5, 393, 41 },
		{ 5, 394, 41 },
		{ 5, 395, 41 },
		{ 5, 396, 41 },
		{ 5, 397, 41 },
		{ 5, 398, 41 },
		{ 5, 399, 41 },
		{ 5, 400, 41 },
		{ 5, 401, 41 },
		{ 5, 402, 41 },
		{ 5, 403, 41 },
		{ 5, 404, 41 },
		{ 5, 405, 41 },
		{ 5, 406, 41 },
		{ 5, 407, 41 },
		{ 5, 408, 41 },
		{ 5, 409, 41 },
		{ 5, 410, 41 },
		{ 5, 411, 41 },
		{ 5, 412, 41 },
		{ 5, 413, 41 },
		{ 5, 414, 41 },
		{ 5, 415, 41 },
		{ 5, 416, 41 },
		{ 5, 417, 41 },
		{ 5, 418, 41 },
		{ 5, 419, 41 },
		{ 5, 420, 41 },
		{ 5, 421, 41 },
		{ 5, 422, 41 },
		{ 5, 423, 41 },
		{ 5, 424, 41 },
		{ 5, 425, 41 },
		{ 5, 426, 41 },
		{ 5, 427, 41 },
		{ 5, 428, 41 },
		{ 5, 429, 41 },
		{ 5, 430, 41 },
		{ 5, 431, 41 },
		{ 5, 432, 41 },
		{ 5, 433, 41 },
		{ 5, 434, 41 },
		{ 5, 435, 41 },
		{ 5, 436, 41 },
		{ 5, 437, 41 },
		{ 5, 438, 41 },
		{ 5, 439, 41 },
		{ 5, 440, 41 },
		{ 5, 441, 41 },
		{ 5, 442, 41 },
		{ 5, 443, 41 },
		{ 5, 444, 41 },
		{ 5, 445, 41 },
		{ 6, 383, 41 },
		{ 6, 384, 41 },
		{ 6, 385, 41 },
		{ 6, 386, 41 },
		{ 6, 387, 41 },
		{ 6, 388, 41 },
		{ 6, 389, 41 },
		{ 6, 390, 41 },
		{ 6, 391, 41 },
		{ 6, 392, 41 },
		{ 6, 393, 41 },
		{ 6, 394, 41 },
		{ 6, 395, 41 },
		{ 6, 396, 41 },
		{ 6, 397, 41 },
		{ 6, 398, 41 },
		{ 6, 399, 41 },
		{ 6, 400, 41 },
		{ 6, 401, 41 },
		{ 6, 402, 41 },
		{ 6, 403, 41 },
		{ 6, 404, 41 },
		{ 6, 405, 41 },
		{ 6, 406, 41 },
		{ 6, 407, 41 },
		{ 6, 408, 41 },
		{ 6, 409, 41 },
		{ 6, 410, 41 },
		{ 6, 411, 41 },
		{ 6, 412, 41 },
		{ 6, 413, 41 },
		{ 6, 414, 41 },
		{ 6, 415, 41 },
		{ 6, 416, 41 },
		{ 6, 417, 41 },
		{ 6, 418, 41 },
		{ 6, 419, 41 },
		{ 6, 420, 41 },
		{ 6, 421, 41 },
		{ 6, 422, 41 },
		{ 6, 423, 41 },
		{ 6, 424, 41 },
		{ 6, 425, 41 },
		{ 6, 426, 41 },
		{ 6, 427, 41 },
		{ 6, 428, 41 },
		{ 6, 429, 41 },
		{ 6, 430, 41 },
		{ 6, 431, 41 },
		{ 6, 432, 41 },
		{ 6, 433, 41 },
		{ 6, 434, 41 },
		{ 6, 435, 41 },
		{ 6, 436, 41 },
		{ 6, 437, 41 },
		{ 6, 438, 41 },
		{ 6, 439, 41 },
		{ 6, 440, 41 },
		{ 6, 441, 41 },
		{ 6, 442, 41 },
		{ 6, 443, 41 },
		{ 6, 444, 41 },
		{ 6, 445, 41 },
		{ 7, 383, 41 },
		{ 7, 384, 41 },
		{ 7, 385, 41 },
		{ 7, 386, 41 },
		{ 7, 387, 41 },
		{ 7, 388, 41 },
		{ 7, 389, 41 },
		{ 7, 390, 41 },
		{ 7, 391, 41 },
		{ 7, 392, 41 },
		{ 7, 393, 41 },
		{ 7, 394, 41 },
		{ 7, 395, 41 },
		{ 7, 396, 41 },
		{ 7, 397, 41 },
		{ 7, 398, 41 },
		{ 7, 399, 41 },
		{ 7, 400, 41 },
		{ 7, 401, 41 },
		{ 7, 402, 41 },
		{ 7, 403, 41 },
		{ 7, 404, 41 },
		{ 7, 405, 41 },
		{ 7, 406, 41 },
		{ 7, 407, 41 },
		{ 7, 408, 41 },
		{ 7, 409, 41 },
		{ 7, 410, 41 },
		{ 7, 411, 41 },
		{ 7, 412, 41 },
		{ 7, 413, 41 },
		{ 7, 414, 41 },
		{ 7, 415, 41 },
		{ 7, 416, 41 },
		{ 7, 417, 41 },
		{ 7, 418, 41 },
		{ 7, 419, 41 },
		{ 7, 420, 41 },
		{ 7, 421, 41 },
		{ 7, 422, 41 },
		{ 7, 423, 41 },
		{ 7, 424, 41 },
		{ 7, 425, 41 },
		{ 7, 426, 41 },
		{ 7, 427, 41 },
		{ 7, 428, 41 },
		{ 7, 429, 41 },
		{ 7, 430, 41 },
		{ 7, 431, 41 },
		{ 7, 432, 41 },
		{ 7, 433, 41 },
		{ 7, 434, 41 },
		{ 7, 435, 41 },
		{ 7, 436, 41 },
		{ 7, 437, 41 },
		{ 7, 438, 41 },
		{ 7, 439, 41 },
		{ 7, 440, 41 },
		{ 7, 441, 41 },
		{ 7, 442, 41 },
		{ 7, 443, 41 },
		{ 7, 444, 41 },
		{ 8, 382, 41 },
		{ 8, 383, 41 },
		{ 8, 384, 41 },
		{ 8, 385, 41 },
		{ 8, 386, 41 },
		{ 8, 387, 41 },
		{ 8, 388, 41 },
		{ 8, 389, 41 },
		{ 8, 390, 41 },
		{ 8, 391, 41 },
		{ 8, 392, 41 },
		{ 8, 393, 41 },
		{ 8, 394, 41 },
		{ 8, 395, 41 },
		{ 8, 396, 41 },
		{ 8, 397, 41 },
		{ 8, 398, 41 },
		{ 8, 399, 41 },
		{ 8, 400, 41 },
		{ 8, 401, 41 },
		{ 8, 402, 41 },
		{ 8, 403, 41 },
		{ 8, 404, 41 },
		{ 8, 405, 41 },
		{ 8, 406, 41 },
		{ 8, 407, 41 },
		{ 8, 408, 41 },
		{ 8, 409, 41 },
		{ 8, 410, 41 },
		{ 8, 411, 41 },
		{ 8, 412, 41 },
		{ 8, 413, 41 },
		{ 8, 414, 41 },
		{ 8, 415, 41 },
		{ 8, 416, 41 },
		{ 8, 417, 41 },
		{ 8, 418, 41 },
		{ 8, 419, 41 },
		{ 8, 420, 41 },
		{ 8, 421, 41 },
		{ 8, 422, 41 },
		{ 8, 423, 41 },
		{ 8, 424, 41 },
		{ 8, 425, 41 },
		{ 8, 426, 41 },
		{ 8, 427, 41 },
		{ 8, 428, 41 },
		{ 8, 429, 41 },
		{ 8, 430, 41 },
		{ 8, 431, 41 },
		{ 8, 432, 41 },
		{ 8, 433, 41 },
		{ 8, 434, 41 },
		{ 8, 435, 41 },
		{ 8, 436, 41 },
		{ 8, 437, 41 },
		{ 8, 438, 41 },
		{ 8, 439, 41 },
		{ 8, 440, 41 },
		{ 8, 441, 41 },
		{ 8, 442, 41 },
		{ 8, 443, 41 },
		{ 9, 382, 41 },
		{ 9, 383, 41 },
		{ 9, 384, 41 },
		{ 9, 385, 41 },
		{ 9, 386, 41 },
		{ 9, 387, 41 },
		{ 9, 388, 41 },
		{ 9, 389, 41 },
		{ 9, 390, 41 },
		{ 9, 391, 41 },
		{ 9, 392, 41 },
		{ 9, 393, 41 },
		{ 9, 394, 41 },
		{ 9, 395, 41 },
		{ 9, 396, 41 },
		{ 9, 397, 41 },
		{ 9, 398, 41 },
		{ 9, 399, 41 },
		{ 9, 400, 41 },
		{ 9, 401, 41 },
		{ 9, 402, 41 },
		{ 9, 403, 41 },
		{ 9, 404, 41 },
		{ 9, 405, 41 },
		{ 9, 406, 41 },
		{ 9, 407, 41 },
		{ 9, 408, 41 },
		{ 9, 409, 41 },
		{ 9, 410, 41 },
		{ 9, 411, 41 },
		{ 9, 412, 41 },
		{ 9, 413, 41 },
		{ 9, 414, 41 },
		{ 9, 415, 41 },
		{ 9, 416, 41 },
		{ 9, 417, 41 },
		{ 9, 418, 41 },
		{ 9, 419, 41 },
		{ 9, 420, 41 },
		{ 9, 421, 41 },
		{ 9, 422, 41 },
		{ 9, 423, 41 },
		{ 9, 424, 41 },
		{ 9, 425, 41 },
		{ 9, 426, 41 },
		{ 9, 427, 41 },
		{ 9, 428, 41 },
		{ 9, 429, 41 },
		{ 9, 430, 41 },
		{ 9, 431, 41 },
		{ 9, 432, 41 },
		{ 9, 433, 41 },
		{ 9, 434, 41 },
		{ 9, 435, 41 },
		{ 9, 436, 41 },
		{ 9, 437, 41 },
		{ 9, 438, 41 },
		{ 9, 439, 41 },
		{ 9, 440, 41 },
		{ 9, 441, 41 },
		{ 9, 442, 41 },
		{ 10, 381, 41 },
		{ 10, 382, 41 },
		{ 10, 383, 41 },
		{ 10, 384, 41 },
		{ 10, 385, 41 },
		{ 10, 386, 41 },
		{ 10, 387, 41 },
		{ 10, 388, 41 },
		{ 10, 389, 41 },
		{ 10, 390, 41 },
		{ 10, 391, 41 },
		{ 10, 392, 41 },
		{ 10, 393, 41 },
		{ 10, 394, 41 },
		{ 10, 395, 41 },
		{ 10, 396, 41 },
		{ 10, 397, 41 },
		{ 10, 398, 41 },
		{ 10, 399, 41 },
		{ 10, 400, 41 },
		{ 10, 401, 41 },
		{ 10, 402, 41 },
		{ 10, 403, 41 },
		{ 10, 404, 41 },
		{ 10, 405, 41 },
		{ 10, 406, 41 },
		{ 10, 407, 41 },
		{ 10, 408, 41 },
		{ 10, 409, 41 },
		{ 10, 410, 41 },
		{ 10, 411, 41 },
		{ 10, 412, 41 },
		{ 10, 413, 41 },
		{ 10, 414, 41 },
		{ 10, 415, 41 },
		{ 10, 416, 41 },
		{ 10, 417, 41 },
		{ 10, 418, 41 },
		{ 10, 419, 41 },
		{ 10, 420, 41 },
		{ 10, 421, 41 },
		{ 10, 422, 41 },
		{ 10, 423, 41 },
		{ 10, 424, 41 },
		{ 10, 425, 41 },
		{ 10, 426, 41 },
		{ 10, 427, 41 },
		{ 10, 428, 41 },
		{ 10, 429, 41 },
		{ 10, 430, 41 },
		{ 10, 431, 41 },
		{ 10, 432, 41 },
		{ 10, 433, 41 },
		{ 10, 434, 41 },
		{ 10, 435, 41 },
		{ 10, 436, 41 },
		{ 10, 437, 41 },
		{ 10, 438, 41 },
		{ 10, 439, 41 },
		{ 10, 440, 41 },
		{ 10, 441, 41 },
		{ 10, 442, 41 },
		{ 11, 381, 41 },
		{ 11, 382, 41 },
		{ 11, 383, 41 },
		{ 11, 384, 41 },
		{ 11, 385, 41 },
		{ 11, 386, 41 },
		{ 11, 387, 41 },
		{ 11, 388, 41 },
		{ 11, 389, 41 },
		{ 11, 390, 41 },
		{ 11, 391, 41 },
		{ 11, 392, 41 },
		{ 11, 393, 41 },
		{ 11, 394, 41 },
		{ 11, 395, 41 },
		{ 11, 396, 41 },
		{ 11, 397, 41 },
		{ 11, 398, 41 },
		{ 11, 399, 41 },
		{ 11, 400, 41 },
		{ 11, 401, 41 },
		{ 11, 402, 41 },
		{ 11, 403, 41 },
		{ 11, 404, 41 },
		{ 11, 405, 41 },
		{ 11, 406, 41 },
		{ 11, 407, 41 },
		{ 11, 408, 41 },
		{ 11, 409, 41 },
		{ 11, 410, 41 },
		{ 11, 411, 41 },
		{ 11, 412, 41 },
		{ 11, 413, 41 },
		{ 11, 414, 41 },
		{ 11, 415, 41 },
		{ 11, 416, 41 },
		{ 11, 417, 41 },
		{ 11, 418, 41 },
		{ 11, 419, 41 },
		{ 11, 420, 41 },
		{ 11, 421, 41 },
		{ 11, 422, 41 },
		{ 11, 423, 41 },
		{ 11, 424, 41 },
		{ 11, 425, 41 },
		{ 11, 426, 41 },
		{ 11, 427, 41 },
		{ 11, 428, 41 },
		{ 11, 429, 41 },
		{ 11, 430, 41 },
		{ 11, 431, 41 },
		{ 11, 432, 41 },
		{ 11, 433, 41 },
		{ 11, 434, 41 },
		{ 11, 435, 41 },
		{ 11, 436, 41 },
		{ 11, 437, 41 },
		{ 11, 438, 41 },
		{ 11, 439, 41 },
		{ 11, 440, 41 },
		{ 11, 441, 41 },
		{ 12, 380, 41 },
		{ 12, 381, 41 },
		{ 12, 382, 41 },
		{ 12, 383, 41 },
		{ 12, 384, 41 },
		{ 12, 385, 41 },
		{ 12, 386, 41 },
		{ 12, 387, 41 },
		{ 12, 388, 41 },
		{ 12, 389, 41 },
		{ 12, 390, 41 },
		{ 12, 391, 41 },
		{ 12, 392, 41 },
		{ 12, 393, 41 },
		{ 12, 394, 41 },
		{ 12, 395, 41 },
		{ 12, 396, 41 },
		{ 12, 397, 41 },
		{ 12, 398, 41 },
		{ 12, 399, 41 },
		{ 12, 400, 41 },
		{ 12, 401, 41 },
		{ 12, 402, 41 },
		{ 12, 403, 41 },
		{ 12, 404, 41 },
		{ 12, 405, 41 },
		{ 12, 406, 41 },
		{ 12, 407, 41 },
		{ 12, 408, 41 },
		{ 12, 409, 41 },
		{ 12, 410, 41 },
		{ 12, 411, 41 },
		{ 12, 412, 41 },
		{ 12, 413, 41 },
		{ 12, 414, 41 },
		{ 12, 415, 41 },
		{ 12, 416, 41 },
		{ 12, 417, 41 },
		{ 12, 418, 41 },
		{ 12, 419, 41 },
		{ 12, 420, 41 },
		{ 12, 421, 41 },
		{ 12, 422, 41 },
		{ 12, 423, 41 },
		{ 12, 424, 41 },
		{ 12, 425, 41 },
		{ 12, 426, 41 },
		{ 12, 427, 41 },
		{ 12, 428, 41 },
		{ 12, 429, 41 },
		{ 12, 430, 41 },
		{ 12, 431, 41 },
		{ 12, 432, 41 },
		{ 12, 433, 41 },
		{ 12, 434, 41 },
		{ 12, 435, 41 },
		{ 12, 436, 41 },
		{ 12, 437, 41 },
		{ 12, 438, 41 },
		{ 12, 439, 41 },
		{ 12, 440, 41 },
		{ 13, 380, 41 },
		{ 13, 381, 41 },
		{ 13, 382, 41 },
		{ 13, 383, 41 },
		{ 13, 384, 41 },
		{ 13, 385, 41 },
		{ 13, 386, 41 },
		{ 13, 387, 41 },
		{ 13, 388, 41 },
		{ 13, 389, 41 },
		{ 13, 390, 41 },
		{ 13, 391, 41 },
		{ 13, 392, 41 },
		{ 13, 393, 41 },
		{ 13, 394, 41 },
		{ 13, 395, 41 },
		{ 13, 396, 41 },
		{ 13, 397, 41 },
		{ 13, 398, 41 },
		{ 13, 399, 41 },
		{ 13, 400, 41 },
		{ 13, 401, 41 },
		{ 13, 402, 41 },
		{ 13, 403, 41 },
		{ 13, 404, 41 },
		{ 13, 405, 41 },
		{ 13, 406, 41 },
		{ 13, 407, 41 },
		{ 13, 408, 41 },
		{ 13, 409, 41 },
		{ 13, 410, 41 },
		{ 13, 411, 41 },
		{ 13, 412, 41 },
		{ 13, 413, 41 },
		{ 13, 414, 41 },
		{ 13, 415, 41 },
		{ 13, 416, 41 },
		{ 13, 417, 41 },
		{ 13, 418, 41 },
		{ 13, 419, 41 },
		{ 13, 420, 41 },
		{ 13, 421, 41 },
		{ 13, 422, 41 },
		{ 13, 423, 41 },
		{ 13, 424, 41 },
		{ 13, 425, 41 },
		{ 13, 426, 41 },
		{ 13, 427, 41 },
		{ 13, 428, 41 },
		{ 13, 429, 41 },
		{ 13, 430, 41 },
		{ 13, 431, 41 },
		{ 13, 432, 41 },
		{ 13, 433, 41 },
		{ 13, 434, 41 },
		{ 13, 435, 41 },
		{ 13, 436, 41 },
		{ 13, 437, 41 },
		{ 13, 438, 41 },
		{ 13, 439, 41 },
		{ 14, 379, 41 },
		{ 14, 380, 41 },
		{ 14, 381, 41 },
		{ 14, 382, 41 },
		{ 14, 383, 41 },
		{ 14, 384, 41 },
		{ 14, 385, 41 },
		{ 14, 386, 41 },
		{ 14, 387, 41 },
		{ 14, 388, 41 },
		{ 14, 389, 41 },
		{ 14, 390, 41 },
		{ 14, 391, 41 },
		{ 14, 392, 41 },
		{ 14, 393, 41 },
		{ 14, 394, 41 },
		{ 14, 395, 41 },
		{ 14, 396, 41 },
		{ 14, 397, 41 },
		{ 14, 398, 41 },
		{ 14, 399, 41 },
		{ 14, 400, 41 },
		{ 14, 401, 41 },
		{ 14, 402, 41 },
		{ 14, 403, 41 },
		{ 14, 404, 41 },
		{ 14, 405, 41 },
		{ 14, 406, 41 },
		{ 14, 407, 41 },
		{ 14, 408, 41 },
		{ 14, 409, 41 },
		{ 14, 410, 41 },
		{ 14, 411, 41 },
		{ 14, 412, 41 },
		{ 14, 413, 41 },
		{ 14, 414, 41 },
		{ 14, 415, 41 },
		{ 14, 416, 41 },
		{ 14, 417, 41 },
		{ 14, 418, 41 },
		{ 14, 419, 41 },
		{ 14, 420, 41 },
		{ 14, 421, 41 },
		{ 14, 422, 41 },
		{ 14, 423, 41 },
		{ 14, 424, 41 },
		{ 14, 425, 41 },
		{ 14, 426, 41 },
		{ 14, 427, 41 },
		{ 14, 428, 41 },
		{ 14, 429, 41 },
		{ 14, 430, 41 },
		{ 14, 431, 41 },
		{ 14, 432, 41 },
		{ 14, 433, 41 },
		{ 14, 434, 41 },
		{ 14, 435, 41 },
		{ 14, 436, 41 },
		{ 14, 437, 41 },
		{ 14, 438, 41 },
		{ 14, 439, 41 },
		{ 15, 379, 41 },
		{ 15, 380, 41 },
		{ 15, 381, 41 },
		{ 15, 382, 41 },
		{ 15, 383, 41 },
		{ 15, 384, 41 },
		{ 15, 385, 41 },
		{ 15, 386, 41 },
		{ 15, 387, 41 },
		{ 15, 388, 41 },
		{ 15, 389, 41 },
		{ 15, 390, 41 },
		{ 15, 391, 41 },
		{ 15, 392, 41 },
		{ 15, 393, 41 },
		{ 15, 394, 41 },
		{ 15, 395, 41 },
		{ 15, 396, 41 },
		{ 15, 397, 41 },
		{ 15, 398, 41 },
		{ 15, 399, 41 },
		{ 15, 400, 41 },
		{ 15, 401, 41 },
		{ 15, 402, 41 },
		{ 15, 403, 41 },
		{ 15, 404, 41 },
		{ 15, 405, 41 },
		{ 15, 406, 41 },
		{ 15, 407, 41 },
		{ 15, 408, 41 },
		{ 15, 409, 41 },
		{ 15, 410, 41 },
		{ 15, 411, 41 },
		{ 15, 412, 41 },
		{ 15, 413, 41 },
		{ 15, 414, 41 },
		{ 15, 415, 41 },
		{ 15, 416, 41 },
		{ 15, 417, 41 },
		{ 15, 418, 41 },
		{ 15, 419, 41 },
		{ 15, 420, 41 },
		{ 15, 421, 41 },
		{ 15, 422, 41 },
		{ 15, 423, 41 },
		{ 15, 424, 41 },
		{ 15, 425, 41 },
		{ 15, 426, 41 },
		{ 15, 427, 41 },
		{ 15, 428, 41 },
		{ 15, 429, 41 },
		{ 15, 430, 41 },
		{ 15, 431, 41 },
		{ 15, 432, 41 },
		{ 15, 433, 41 },
		{ 15, 434, 41 },
		{ 15, 435, 41 },
		{ 15, 436, 41 },
		{ 15, 437, 41 },
		{ 15, 438, 41 },
		{ 16, 378, 41 },
		{ 16, 379, 41 },
		{ 16, 380, 41 },
		{ 16, 381, 41 },
		{ 16, 382, 41 },
		{ 16, 383, 41 },
		{ 16, 384, 41 },
		{ 16, 385, 41 },
		{ 16, 386, 41 },
		{ 16, 387, 41 },
		{ 16, 388, 41 },
		{ 16, 389, 41 },
		{ 16, 390, 41 },
		{ 16, 391, 41 },
		{ 16, 392, 41 },
		{ 16, 393, 41 },
		{ 16, 394, 41 },
		{ 16, 395, 41 },
		{ 16, 396, 41 },
		{ 16, 397, 41 },
		{ 16, 398, 41 },
		{ 16, 399, 41 },
		{ 16, 400, 41 },
		{ 16, 401, 41 },
		{ 16, 402, 41 },
		{ 16, 403, 41 },
		{ 16, 404, 41 },
		{ 16, 405, 41 },
		{ 16, 406, 41 },
		{ 16, 407, 41 },
		{ 16, 408, 41 },
		{ 16, 409, 41 },
		{ 16, 410, 41 },
		{ 16, 411, 41 },
		{ 16, 412, 41 },
		{ 16, 413, 41 },
		{ 16, 414, 41 },
		{ 16, 415, 41 },
		{ 16, 416, 41 },
		{ 16, 417, 41 },
		{ 16, 418, 41 },
		{ 16, 419, 41 },
		{ 16, 420, 41 },
		{ 16, 421, 41 },
		{ 16, 422, 41 },
		{ 16, 423, 41 },
		{ 16, 424, 41 },
		{ 16, 425, 41 },
		{ 16, 426, 41 },
		{ 16, 427, 41 },
		{ 16, 428, 41 },
		{ 16, 429, 41 },
		{ 16, 430, 41 },
		{ 16, 431, 41 },
		{ 16, 432, 41 },
		{ 16, 433, 41 },
		{ 16, 434, 41 },
		{ 16, 435, 41 },
		{ 16, 436, 41 },
		{ 16, 437, 41 },
		{ 17, 378, 41 },
		{ 17, 379, 41 },
		{ 17, 380, 41 },
		{ 17, 381, 41 },
		{ 17, 382, 41 },
		{ 17, 383, 41 },
		{ 17, 384, 41 },
		{ 17, 385, 41 },
		{ 17, 386, 41 },
		{ 17, 387, 41 },
		{ 17, 388, 41 },
		{ 17, 389, 41 },
		{ 17, 390, 41 },
		{ 17, 391, 41 },
		{ 17, 392, 41 },
		{ 17, 393, 41 },
		{ 17, 394, 41 },
		{ 17, 395, 41 },
		{ 17, 396, 41 },
		{ 17, 397, 41 },
		{ 17, 398, 41 },
		{ 17, 399, 41 },
		{ 17, 400, 41 },
		{ 17, 401, 41 },
		{ 17, 402, 41 },
		{ 17, 403, 41 },
		{ 17, 404, 41 },
		{ 17, 405, 41 },
		{ 17, 406, 41 },
		{ 17, 407, 41 },
		{ 17, 408, 41 },
		{ 17, 409, 41 },
		{ 17, 410, 41 },
		{ 17, 411, 41 },
		{ 17, 412, 41 },
		{ 17, 413, 41 },
		{ 17, 414, 41 },
		{ 17, 415, 41 },
		{ 17, 416, 41 },
		{ 17, 417, 41 },
		{ 17, 418, 41 },
		{ 17, 419, 41 },
		{ 17, 420, 41 },
		{ 17, 421, 41 },
		{ 17, 422, 41 },
		{ 17, 423, 41 },
		{ 17, 424, 41 },
		{ 17, 425, 41 },
		{ 17, 426, 41 },
		{ 17, 427, 41 },
		{ 17, 428, 41 },
		{ 17, 429, 41 },
		{ 17, 430, 41 },
		{ 17, 431, 41 },
		{ 17, 432, 41 },
		{ 17, 433, 41 },
		{ 17, 434, 41 },
		{ 17, 435, 41 },
		{ 17, 436, 41 },
		{ 18, 377, 41 },
		{ 18, 378, 41 },
		{ 18, 379, 41 },
		{ 18, 380, 41 },
		{ 18, 381, 41 },
		{ 18, 382, 41 },
		{ 18, 383, 41 },
		{ 18, 384, 41 },
		{ 18, 385, 41 },
		{ 18, 386, 41 },
		{ 18, 387, 41 },
		{ 18, 388, 41 },
		{ 18, 389, 41 },
		{ 18, 390, 41 },
		{ 18, 391, 41 },
		{ 18, 392, 41 },
		{ 18, 393, 41 },
		{ 18, 394, 41 },
		{ 18, 395, 41 },
		{ 18, 396, 41 },
		{ 18, 397, 41 },
		{ 18, 398, 41 },
		{ 18, 399, 41 },
		{ 18, 400, 41 },
		{ 18, 401, 41 },
		{ 18, 402, 41 },
		{ 18, 403, 41 },
		{ 18, 404, 41 },
		{ 18, 405, 41 },
		{ 18, 406, 41 },
		{ 18, 407, 41 },
		{ 18, 408, 41 },
		{ 18, 409, 41 },
		{ 18, 410, 41 },
		{ 18, 411, 41 },
		{ 18, 412, 41 },
		{ 18, 413, 41 },
		{ 18, 414, 41 },
		{ 18, 415, 41 },
		{ 18, 416, 41 },
		{ 18, 417, 41 },
		{ 18, 418, 41 },
		{ 18, 419, 41 },
		{ 18, 420, 41 },
		{ 18, 421, 41 },
		{ 18, 422, 41 },
		{ 18, 423, 41 },
		{ 18, 424, 41 },
		{ 18, 425, 41 },
		{ 18, 426, 41 },
		{ 18, 427, 41 },
		{ 18, 428, 41 },
		{ 18, 429, 41 },
		{ 18, 430, 41 },
		{ 18, 431, 41 },
		{ 18, 432, 41 },
		{ 18, 433, 41 },
		{ 18, 434, 41 },
		{ 18, 435, 41 },
		{ 18, 436, 41 },
		{ 19, 377, 41 },
		{ 19, 378, 41 },
		{ 19, 379, 41 },
		{ 19, 380, 41 },
		{ 19, 381, 41 },
		{ 19, 382, 41 },
		{ 19, 383, 41 },
		{ 19, 384, 41 },
		{ 19, 385, 41 },
		{ 19, 386, 41 },
		{ 19, 387, 41 },
		{ 19, 388, 41 },
		{ 19, 389, 41 },
		{ 19, 390, 41 },
		{ 19, 391, 41 },
		{ 19, 392, 41 },
		{ 19, 393, 41 },
		{ 19, 394, 41 },
		{ 19, 395, 41 },
		{ 19, 396, 41 },
		{ 19, 397, 41 },
		{ 19, 398, 41 },
		{ 19, 399, 41 },
		{ 19, 400, 41 },
		{ 19, 401, 41 },
		{ 19, 402, 41 },
		{ 19, 403, 41 },
		{ 19, 404, 41 },
		{ 19, 405, 41 },
		{ 19, 406, 41 },
		{ 19, 407, 41 },
		{ 19, 408, 41 },
		{ 19, 409, 41 },
		{ 19, 410, 41 },
		{ 19, 411, 41 },
		{ 19, 412, 41 },
		{ 19, 413, 41 },
		{ 19, 414, 41 },
		{ 19, 415, 41 },
		{ 19, 416, 41 },
		{ 19, 417, 41 },
		{ 19, 418, 41 },
		{ 19, 419, 41 },
		{ 19, 420, 41 },
		{ 19, 421, 41 },
		{ 19, 422, 41 },
		{ 19, 423, 41 },
		{ 19, 424, 41 },
		{ 19, 425, 41 },
		{ 19, 426, 41 },
		{ 19, 427, 41 },
		{ 19, 428, 41 },
		{ 19, 429, 41 },
		{ 19, 430, 41 },
		{ 19, 431, 41 },
		{ 19, 432, 41 },
		{ 19, 433, 41 },
		{ 19, 434, 41 },
		{ 19, 435, 41 },
		{ 20, 376, 41 },
		{ 20, 377, 41 },
		{ 20, 378, 41 },
		{ 20, 379, 41 },
		{ 20, 380, 41 },
		{ 20, 381, 41 },
		{ 20, 382, 41 },
		{ 20, 383, 41 },
		{ 20, 384, 41 },
		{ 20, 385, 41 },
		{ 20, 386, 41 },
		{ 20, 387, 41 },
		{ 20, 388, 41 },
		{ 20, 389, 41 },
		{ 20, 390, 41 },
		{ 20, 391, 41 },
		{ 20, 392, 41 },
		{ 20, 393, 41 },
		{ 20, 394, 41 },
		{ 20, 395, 41 },
		{ 20, 396, 41 },
		{ 20, 397, 41 },
		{ 20, 398, 41 },
		{ 20, 399, 41 },
		{ 20, 400, 41 },
		{ 20, 401, 41 },
		{ 20, 402, 41 },
		{ 20, 403, 41 },
		{ 20, 404, 41 },
		{ 20, 405, 41 },
		{ 20, 406, 41 },
		{ 20, 407, 41 },
		{ 20, 408, 41 },
		{ 20, 409, 41 },
		{ 20, 410, 41 },
		{ 20, 411, 41 },
		{ 20, 412, 41 },
		{ 20, 413, 41 },
		{ 20, 414, 41 },
		{ 20, 415, 41 },
		{ 20, 416, 41 },
		{ 20, 417, 41 },
		{ 20, 418, 41 },
		{ 20, 419, 41 },
		{ 20, 420, 41 },
		{ 20, 421, 41 },
		{ 20, 422, 41 },
		{ 20, 423, 41 },
		{ 20, 424, 41 },
		{ 20, 425, 41 },
		{ 20, 426, 41 },
		{ 20, 427, 41 },
		{ 20, 428, 41 },
		{ 20, 429, 41 },
		{ 20, 430, 41 },
		{ 20, 431, 41 },
		{ 20, 432, 41 },
		{ 20, 433, 41 },
		{ 20, 434, 41 },
		{ 21, 376, 41 },
		{ 21, 377, 41 },
		{ 21, 378, 41 },
		{ 21, 379, 41 },
		{ 21, 380, 41 },
		{ 21, 381, 41 },
		{ 21, 382, 41 },
		{ 21, 383, 41 },
		{ 21, 384, 41 },
		{ 21, 385, 41 },
		{ 21, 386, 41 },
		{ 21, 387, 41 },
		{ 21, 388, 41 },
		{ 21, 389, 41 },
		{ 21, 390, 41 },
		{ 21, 391, 41 },
		{ 21, 392, 41 },
		{ 21, 393, 41 },
		{ 21, 394, 41 },
		{ 21, 395, 41 },
		{ 21, 396, 41 },
		{ 21, 397, 41 },
		{ 21, 398, 41 },
		{ 21, 399, 41 },
		{ 21, 400, 41 },
		{ 21, 401, 41 },
		{ 21, 402, 41 },
		{ 21, 403, 41 },
		{ 21, 404, 41 },
		{ 21, 405, 41 },
		{ 21, 406, 41 },
		{ 21, 407, 41 },
		{ 21, 408, 41 },
		{ 21, 409, 41 },
		{ 21, 410, 41 },
		{ 21, 411, 41 },
		{ 21, 412, 41 },
		{ 21, 413, 41 },
		{ 21, 414, 41 },
		{ 21, 415, 41 },
		{ 21, 416, 41 },
		{ 21, 417, 41 },
		{ 21, 418, 41 },
		{ 21, 419, 41 },
		{ 21, 420, 41 },
		{ 21, 421, 41 },
		{ 21, 422, 41 },
		{ 21, 423, 41 },
		{ 21, 424, 41 },
		{ 21, 425, 41 },
		{ 21, 426, 41 },
		{ 21, 427, 41 },
		{ 21, 428, 41 },
		{ 21, 429, 41 },
		{ 21, 430, 41 },
		{ 21, 431, 41 },
		{ 21, 432, 41 },
		{ 21, 433, 41 },
		{ 22, 375, 41 },
		{ 22, 376, 41 },
		{ 22, 377, 41 },
		{ 22, 378, 41 },
		{ 22, 379, 41 },
		{ 22, 380, 41 },
		{ 22, 381, 41 },
		{ 22, 382, 41 },
		{ 22, 383, 41 },
		{ 22, 384, 41 },
		{ 22, 385, 41 },
		{ 22, 386, 41 },
		{ 22, 387, 41 },
		{ 22, 388, 41 },
		{ 22, 389, 41 },
		{ 22, 390, 41 },
		{ 22, 391, 41 },
		{ 22, 392, 41 },
		{ 22, 393, 41 },
		{ 22, 394, 41 },
		{ 22, 395, 41 },
		{ 22, 396, 41 },
		{ 22, 397, 41 },
		{ 22, 398, 41 },
		{ 22, 399, 41 },
		{ 22, 400, 41 },
		{ 22, 401, 41 },
		{ 22, 402, 41 },
		{ 22, 403, 41 },
		{ 22, 404, 41 },
		{ 22, 405, 41 },
		{ 22, 406, 41 },
		{ 22, 407, 41 },
		{ 22, 408, 41 },
		{ 22, 409, 41 },
		{ 22, 410, 41 },
		{ 22, 411, 41 },
		{ 22, 412, 41 },
		{ 22, 413, 41 },
		{ 22, 414, 41 },
		{ 22, 415, 41 },
		{ 22, 416, 41 },
		{ 22, 417, 41 },
		{ 22, 418, 41 },
		{ 22, 419, 41 },
		{ 22, 420, 41 },
		{ 22, 421, 41 },
		{ 22, 422, 41 },
		{ 22, 423, 41 },
		{ 22, 424, 41 },
		{ 22, 425, 41 },
		{ 22, 426, 41 },
		{ 22, 427, 41 },
		{ 22, 428, 41 },
		{ 22, 429, 41 },
		{ 22, 430, 41 },
		{ 22, 431, 41 },
		{ 22, 432, 41 },
		{ 22, 433, 41 },
		{ 23, 375, 41 },
		{ 23, 376, 41 },
		{ 23, 377, 41 },
		{ 23, 378, 41 },
		{ 23, 379, 41 },
		{ 23, 380, 41 },
		{ 23, 381, 41 },
		{ 23, 382, 41 },
		{ 23, 383, 41 },
		{ 23, 384, 41 },
		{ 23, 385, 41 },
		{ 23, 386, 41 },
		{ 23, 387, 41 },
		{ 23, 388, 41 },
		{ 23, 389, 41 },
		{ 23, 390, 41 },
		{ 23, 391, 41 },
		{ 23, 392, 41 },
		{ 23, 393, 41 },
		{ 23, 394, 41 },
		{ 23, 395, 41 },
		{ 23, 396, 41 },
		{ 23, 397, 41 },
		{ 23, 398, 41 },
		{ 23, 399, 41 },
		{ 23, 400, 41 },
		{ 23, 401, 41 },
		{ 23, 402, 41 },
		{ 23, 403, 41 },
		{ 23, 404, 41 },
		{ 23, 405, 41 },
		{ 23, 406, 41 },
		{ 23, 407, 41 },
		{ 23, 408, 41 },
		{ 23, 409, 41 },
		{ 23, 410, 41 },
		{ 23, 411, 41 },
		{ 23, 412, 41 },
		{ 23, 413, 41 },
		{ 23, 414, 41 },
		{ 23, 415, 41 },
		{ 23, 416, 41 },
		{ 23, 417, 41 },
		{ 23, 418, 41 },
		{ 23, 419, 41 },
		{ 23, 420, 41 },
		{ 23, 421, 41 },
		{ 23, 422, 41 },
		{ 23, 423, 41 },
		{ 23, 424, 41 },
		{ 23, 425, 41 },
		{ 23, 426, 41 },
		{ 23, 427, 41 },
		{ 23, 428, 41 },
		{ 23, 429, 41 },
		{ 23, 430, 41 },
		{ 23, 431, 41 },
		{ 23, 432, 41 },
		{ 24, 374, 41 },
		{ 24, 375, 41 },
		{ 24, 376, 41 },
		{ 24, 377, 41 },
		{ 24, 378, 41 },
		{ 24, 379, 41 },
		{ 24, 380, 41 },
		{ 24, 381, 41 },
		{ 24, 382, 41 },
		{ 24, 383, 41 },
		{ 24, 384, 41 },
		{ 24, 385, 41 },
		{ 24, 386, 41 },
		{ 24, 387, 41 },
		{ 24, 388, 41 },
		{ 24, 389, 41 },
		{ 24, 390, 41 },
		{ 24, 391, 41 },
		{ 24, 392, 41 },
		{ 24, 393, 41 },
		{ 24, 394, 41 },
		{ 24, 395, 41 },
		{ 24, 396, 41 },
		{ 24, 397, 41 },
		{ 24, 398, 41 },
		{ 24, 399, 41 },
		{ 24, 400, 41 },
		{ 24, 401, 41 },
		{ 24, 402, 41 },
		{ 24, 403, 41 },
		{ 24, 404, 41 },
		{ 24, 405, 41 },
		{ 24, 406, 41 },
		{ 24, 407, 41 },
		{ 24, 408, 41 },
		{ 24, 409, 41 },
		{ 24, 410, 41 },
		{ 24, 411, 41 },
		{ 24, 412, 41 },
		{ 24, 413, 41 },
		{ 24, 414, 41 },
		{ 24, 415, 41 },
		{ 24, 416, 41 },
		{ 24, 417, 41 },
		{ 24, 418, 41 },
		{ 24, 419, 41 },
		{ 24, 420, 41 },
		{ 24, 421, 41 },
		{ 24, 422, 41 },
		{ 24, 423, 41 },
		{ 24, 424, 41 },
		{ 24, 425, 41 },
		{ 24, 426, 41 },
		{ 24, 427, 41 },
		{ 24, 428, 41 },
		{ 24, 429, 41 },
		{ 24, 430, 41 },
		{ 24, 431, 41 },
		{ 25, 374, 41 },
		{ 25, 375, 41 },
		{ 25, 376, 41 },
		{ 25, 377, 41 },
		{ 25, 378, 41 },
		{ 25, 379, 41 },
		{ 25, 380, 41 },
		{ 25, 381, 41 },
		{ 25, 382, 41 },
		{ 25, 383, 41 },
		{ 25, 384, 41 },
		{ 25, 385, 41 },
		{ 25, 386, 41 },
		{ 25, 387, 41 },
		{ 25, 388, 41 },
		{ 25, 389, 41 },
		{ 25, 390, 41 },
		{ 25, 391, 41 },
		{ 25, 392, 41 },
		{ 25, 393, 41 },
		{ 25, 394, 41 },
		{ 25, 395, 41 },
		{ 25, 396, 41 },
		{ 25, 397, 41 },
		{ 25, 398, 41 },
		{ 25, 399, 41 },
		{ 25, 400, 41 },
		{ 25, 401, 41 },
		{ 25, 402, 41 },
		{ 25, 403, 41 },
		{ 25, 404, 41 },
		{ 25, 405, 41 },
		{ 25, 406, 41 },
		{ 25, 407, 41 },
		{ 25, 408, 41 },
		{ 25, 409, 41 },
		{ 25, 410, 41 },
		{ 25, 411, 41 },
		{ 25, 412, 41 },
		{ 25, 413, 41 },
		{ 25, 414, 41 },
		{ 25, 415, 41 },
		{ 25, 416, 41 },
		{ 25, 417, 41 },
		{ 25, 418, 41 },
		{ 25, 419, 41 },
		{ 25, 420, 41 },
		{ 25, 421, 41 },
		{ 25, 422, 41 },
		{ 25, 423, 41 },
		{ 25, 424, 41 },
		{ 25, 425, 41 },
		{ 25, 426, 41 },
		{ 25, 427, 41 },
		{ 25, 428, 41 },
		{ 25, 429, 41 },
		{ 25, 430, 41 },
		{ 26, 373, 41 },
		{ 26, 374, 41 },
		{ 26, 375, 41 },
		{ 26, 376, 41 },
		{ 26, 377, 41 },
		{ 26, 378, 41 },
		{ 26, 379, 41 },
		{ 26, 380, 41 },
		{ 26, 381, 41 },
		{ 26, 382, 41 },
		{ 26, 383, 41 },
		{ 26, 384, 41 },
		{ 26, 385, 41 },
		{ 26, 386, 41 },
		{ 26, 387, 41 },
		{ 26, 388, 41 },
		{ 26, 389, 41 },
		{ 26, 390, 41 },
		{ 26, 391, 41 },
		{ 26, 392, 41 },
		{ 26, 393, 41 },
		{ 26, 394, 41 },
		{ 26, 395, 41 },
		{ 26, 396, 41 },
		{ 26, 397, 41 },
		{ 26, 398, 41 },
		{ 26, 399, 41 },
		{ 26, 400, 41 },
		{ 26, 401, 41 },
		{ 26, 402, 41 },
		{ 26, 403, 41 },
		{ 26, 404, 41 },
		{ 26, 405, 41 },
		{ 26, 406, 41 },
		{ 26, 407, 41 },
		{ 26, 408, 41 },
		{ 26, 409, 41 },
		{ 26, 410, 41 },
		{ 26, 411, 41 },
		{ 26, 412, 41 },
		{ 26, 413, 41 },
		{ 26, 414, 41 },
		{ 26, 415, 41 },
		{ 26, 416, 41 },
		{ 26, 417, 41 },
		{ 26, 418, 41 },
		{ 26, 419, 41 },
		{ 26, 420, 41 },
		{ 26, 421, 41 },
		{ 26, 422, 41 },
		{ 26, 423, 41 },
		{ 26, 424, 41 },
		{ 26, 425, 41 },
		{ 26, 426, 41 },
		{ 26, 427, 41 },
		{ 26, 428, 41 },
		{ 26, 429, 41 },
		{ 26, 430, 41 },
		{ 27, 373, 41 },
		{ 27, 374, 41 },
		{ 27, 375, 41 },
		{ 27, 376, 41 },
		{ 27, 377, 41 },
		{ 27, 378, 41 },
		{ 27, 379, 41 },
		{ 27, 380, 41 },
		{ 27, 381, 41 },
		{ 27, 382, 41 },
		{ 27, 383, 41 },
		{ 27, 384, 41 },
		{ 27, 385, 41 },
		{ 27, 386, 41 },
		{ 27, 387, 41 },
		{ 27, 388, 41 },
		{ 27, 389, 41 },
		{ 27, 390, 41 },
		{ 27, 391, 41 },
		{ 27, 392, 41 },
		{ 27, 393, 41 },
		{ 27, 394, 41 },
		{ 27, 395, 41 },
		{ 27, 396, 41 },
		{ 27, 397, 41 },
		{ 27, 398, 41 },
		{ 27, 399, 41 },
		{ 27, 400, 41 },
		{ 27, 401, 41 },
		{ 27, 402, 41 },
		{ 27, 403, 41 },
		{ 27, 404, 41 },
		{ 27, 405, 41 },
		{ 27, 406, 41 },
		{ 27, 407, 41 },
		{ 27, 408, 41 },
		{ 27, 409, 41 },
		{ 27, 410, 41 },
		{ 27, 411, 41 },
		{ 27, 412, 41 },
		{ 27, 413, 41 },
		{ 27, 414, 41 },
		{ 27, 415, 41 },
		{ 27, 416, 41 },
		{ 27, 417, 41 },
		{ 27, 418, 41 },
		{ 27, 419, 41 },
		{ 27, 420, 41 },
		{ 27, 421, 41 },
		{ 27, 422, 41 },
		{ 27, 423, 41 },
		{ 27, 424, 41 },
		{ 27, 425, 41 },
		{ 27, 426, 41 },
		{ 27, 427, 41 },
		{ 27, 428, 41 },
		{ 27, 429, 41 },
		{ 28, 372, 41 },
		{ 28, 373, 41 },
		{ 28, 374, 41 },
		{ 28, 375, 41 },
		{ 28, 376, 41 },
		{ 28, 377, 41 },
		{ 28, 378, 41 },
		{ 28, 379, 41 },
		{ 28, 380, 41 },
		{ 28, 381, 41 },
		{ 28, 382, 41 },
		{ 28, 383, 41 },
		{ 28, 384, 41 },
		{ 28, 385, 41 },
		{ 28, 386, 41 },
		{ 28, 387, 41 },
		{ 28, 388, 41 },
		{ 28, 389, 41 },
		{ 28, 390, 41 },
		{ 28, 391, 41 },
		{ 28, 392, 41 },
		{ 28, 393, 41 },
		{ 28, 394, 41 },
		{ 28, 395, 41 },
		{ 28, 396, 41 },
		{ 28, 397, 41 },
		{ 28, 398, 41 },
		{ 28, 399, 41 },
		{ 28, 400, 41 },
		{ 28, 401, 41 },
		{ 28, 402, 41 },
		{ 28, 403, 41 },
		{ 28, 404, 41 },
		{ 28, 405, 41 },
		{ 28, 406, 41 },
		{ 28, 407, 41 },
		{ 28, 408, 41 },
		{ 28, 409, 41 },
		{ 28, 410, 41 },
		{ 28, 411, 41 },
		{ 28, 412, 41 },
		{ 28, 413, 41 },
		{ 28, 414, 41 },
		{ 28, 415, 41 },
		{ 28, 416, 41 },
		{ 28, 417, 41 },
		{ 28, 418, 41 },
		{ 28, 419, 41 },
		{ 28, 420, 41 },
		{ 28, 421, 41 },
		{ 28, 422, 41 },
		{ 28, 423, 41 },
		{ 28, 424, 41 },
		{ 28, 425, 41 },
		{ 28, 426, 41 },
		{ 28, 427, 41 },
		{ 28, 428, 41 },
		{ 29, 372, 41 },
		{ 29, 373, 41 },
		{ 29, 374, 41 },
		{ 29, 375, 41 },
		{ 29, 376, 41 },
		{ 29, 377, 41 },
		{ 29, 378, 41 },
		{ 29, 379, 41 },
		{ 29, 380, 41 },
		{ 29, 381, 41 },
		{ 29, 382, 41 },
		{ 29, 383, 41 },
		{ 29, 384, 41 },
		{ 29, 385, 41 },
		{ 29, 386, 41 },
		{ 29, 387, 41 },
		{ 29, 388, 41 },
		{ 29, 389, 41 },
		{ 29, 390, 41 },
		{ 29, 391, 41 },
		{ 29, 392, 41 },
		{ 29, 393, 41 },
		{ 29, 394, 41 },
		{ 29, 395, 41 },
		{ 29, 396, 41 },
		{ 29, 397, 41 },
		{ 29, 398, 41 },
		{ 29, 399, 41 },
		{ 29, 400, 41 },
		{ 29, 401, 41 },
		{ 29, 402, 41 },
		{ 29, 403, 41 },
		{ 29, 404, 41 },
		{ 29, 405, 41 },
		{ 29, 406, 41 },
		{ 29, 407, 41 },
		{ 29, 408, 41 },
		{ 29, 409, 41 },
		{ 29, 410, 41 },
		{ 29, 411, 41 },
		{ 29, 412, 41 },
		{ 29, 413, 41 },
		{ 29, 414, 41 },
		{ 29, 415, 41 },
		{ 29, 416, 41 },
		{ 29, 417, 41 },
		{ 29, 418, 41 },
		{ 29, 419, 41 },
		{ 29, 420, 41 },
		{ 29, 421, 41 },
		{ 29, 422, 41 },
		{ 29, 423, 41 },
		{ 29, 424, 41 },
		{ 29, 425, 41 },
		{ 29, 426, 41 },
		{ 29, 427, 41 },
		{ 30, 371, 41 },
		{ 30, 372, 41 },
		{ 30, 373, 41 },
		{ 30, 374, 41 },
		{ 30, 375, 41 },
		{ 30, 376, 41 },
		{ 30, 377, 41 },
		{ 30, 378, 41 },
		{ 30, 379, 41 },
		{ 30, 380, 41 },
		{ 30, 381, 41 },
		{ 30, 382, 41 },
		{ 30, 383, 41 },
		{ 30, 384, 41 },
		{ 30, 385, 41 },
		{ 30, 386, 41 },
		{ 30, 387, 41 },
		{ 30, 388, 41 },
		{ 30, 389, 41 },
		{ 30, 390, 41 },
		{ 30, 391, 41 },
		{ 30, 392, 41 },
		{ 30, 393, 41 },
		{ 30, 394, 41 },
		{ 30, 395, 41 },
		{ 30, 396, 41 },
		{ 30, 397, 41 },
		{ 30, 398, 41 },
		{ 30, 399, 41 },
		{ 30, 400, 41 },
		{ 30, 401, 41 },
		{ 30, 402, 41 },
		{ 30, 403, 41 },
		{ 30, 404, 41 },
		{ 30, 405, 41 },
		{ 30, 406, 41 },
		{ 30, 407, 41 },
		{ 30, 408, 41 },
		{ 30, 409, 41 },
		{ 30, 410, 41 },
		{ 30, 411, 41 },
		{ 30, 412, 41 },
		{ 30, 413, 41 },
		{ 30, 414, 41 },
		{ 30, 415, 41 },
		{ 30, 416, 41 },
		{ 30, 417, 41 },
		{ 30, 418, 41 },
		{ 30, 419, 41 },
		{ 30, 420, 41 },
		{ 30, 421, 41 },
		{ 30, 422, 41 },
		{ 30, 423, 41 },
		{ 30, 424, 41 },
		{ 30, 425, 41 },
		{ 30, 426, 41 },
		{ 30, 427, 41 },
		{ 31, 371, 41 },
		{ 31, 372, 41 },
		{ 31, 373, 41 },
		{ 31, 374, 41 },
		{ 31, 375, 41 },
		{ 31, 376, 41 },
		{ 31, 377, 41 },
		{ 31, 378, 41 },
		{ 31, 379, 41 },
		{ 31, 380, 41 },
		{ 31, 381, 41 },
		{ 31, 382, 41 },
		{ 31, 383, 41 },
		{ 31, 384, 41 },
		{ 31, 385, 41 },
		{ 31, 386, 41 },
		{ 31, 387, 41 },
		{ 31, 388, 41 },
		{ 31, 389, 41 },
		{ 31, 390, 41 },
		{ 31, 391, 41 },
		{ 31, 392, 41 },
		{ 31, 393, 41 },
		{ 31, 394, 41 },
		{ 31, 395, 41 },
		{ 31, 396, 41 },
		{ 31, 397, 41 },
		{ 31, 398, 41 },
		{ 31, 399, 41 },
		{ 31, 400, 41 },
		{ 31, 401, 41 },
		{ 31, 402, 41 },
		{ 31, 403, 41 },
		{ 31, 404, 41 },
		{ 31, 405, 41 },
		{ 31, 406, 41 },
		{ 31, 407, 41 },
		{ 31, 408, 41 },
		{ 31, 409, 41 },
		{ 31, 410, 41 },
		{ 31, 411, 41 },
		{ 31, 412, 41 },
		{ 31, 413, 41 },
		{ 31, 414, 41 },
		{ 31, 415, 41 },
		{ 31, 416, 41 },
		{ 31, 417, 41 },
		{ 31, 418, 41 },
		{ 31, 419, 41 },
		{ 31, 420, 41 },
		{ 31, 421, 41 },
		{ 31, 422, 41 },
		{ 31, 423, 41 },
		{ 31, 424, 41 },
		{ 31, 425, 41 },
		{ 31, 426, 41 },
		{ 32, 370, 41 },
		{ 32, 371, 41 },
		{ 32, 372, 41 },
		{ 32, 373, 41 },
		{ 32, 374, 41 },
		{ 32, 375, 41 },
		{ 32, 376, 41 },
		{ 32, 377, 41 },
		{ 32, 378, 41 },
		{ 32, 379, 41 },
		{ 32, 380, 41 },
		{ 32, 381, 41 },
		{ 32, 382, 41 },
		{ 32, 383, 41 },
		{ 32, 384, 41 },
		{ 32, 385, 41 },
		{ 32, 386, 41 },
		{ 32, 387, 41 },
		{ 32, 388, 41 },
		{ 32, 389, 41 },
		{ 32, 390, 41 },
		{ 32, 391, 41 },
		{ 32, 392, 41 },
		{ 32, 393, 41 },
		{ 32, 394, 41 },
		{ 32, 395, 41 },
		{ 32, 396, 41 },
		{ 32, 397, 41 },
		{ 32, 398, 41 },
		{ 32, 399, 41 },
		{ 32, 400, 41 },
		{ 32, 401, 41 },
		{ 32, 402, 41 },
		{ 32, 403, 41 },
		{ 32, 404, 41 },
		{ 32, 405, 41 },
		{ 32, 406, 41 },
		{ 32, 407, 41 },
		{ 32, 408, 41 },
		{ 32, 409, 41 },
		{ 32, 410, 41 },
		{ 32, 411, 41 },
		{ 32, 412, 41 },
		{ 32, 413, 41 },
		{ 32, 414, 41 },
		{ 32, 415, 41 },
		{ 32, 416, 41 },
		{ 32, 417, 41 },
		{ 32, 418, 41 },
		{ 32, 419, 41 },
		{ 32, 420, 41 },
		{ 32, 421, 41 },
		{ 32, 422, 41 },
		{ 32, 423, 41 },
		{ 32, 424, 41 },
		{ 32, 425, 41 },
		{ 33, 370, 41 },
		{ 33, 371, 41 },
		{ 33, 372, 41 },
		{ 33, 373, 41 },
		{ 33, 374, 41 },
		{ 33, 375, 41 },
		{ 33, 376, 41 },
		{ 33, 377, 41 },
		{ 33, 378, 41 },
		{ 33, 379, 41 },
		{ 33, 380, 41 },
		{ 33, 381, 41 },
		{ 33, 382, 41 },
		{ 33, 383, 41 },
		{ 33, 384, 41 },
		{ 33, 385, 41 },
		{ 33, 386, 41 },
		{ 33, 387, 41 },
		{ 33, 388, 41 },
		{ 33, 389, 41 },
		{ 33, 390, 41 },
		{ 33, 391, 41 },
		{ 33, 392, 41 },
		{ 33, 393, 41 },
		{ 33, 394, 41 },
		{ 33, 395, 41 },
		{ 33, 396, 41 },
		{ 33, 397, 41 },
		{ 33, 398, 41 },
		{ 33, 399, 41 },
		{ 33, 400, 41 },
		{ 33, 401, 41 },
		{ 33, 402, 41 },
		{ 33, 403, 41 },
		{ 33, 404, 41 },
		{ 33, 405, 41 },
		{ 33, 406, 41 },
		{ 33, 407, 41 },
		{ 33, 408, 41 },
		{ 33, 409, 41 },
		{ 33, 410, 41 },
		{ 33, 411, 41 },
		{ 33, 412, 41 },
		{ 33, 413, 41 },
		{ 33, 414, 41 },
		{ 33, 415, 41 },
		{ 33, 416, 41 },
		{ 33, 417, 41 },
		{ 33, 418, 41 },
		{ 33, 419, 41 },
		{ 33, 420, 41 },
		{ 33, 421, 41 },
		{ 33, 422, 41 },
		{ 33, 423, 41 },
		{ 33, 424, 41 },
		{ 34, 369, 41 },
		{ 34, 370, 41 },
		{ 34, 371, 41 },
		{ 34, 372, 41 },
		{ 34, 373, 41 },
		{ 34, 374, 41 },
		{ 34, 375, 41 },
		{ 34, 376, 41 },
		{ 34, 377, 41 },
		{ 34, 378, 41 },
		{ 34, 379, 41 },
		{ 34, 380, 41 },
		{ 34, 381, 41 },
		{ 34, 382, 41 },
		{ 34, 383, 41 },
		{ 34, 384, 41 },
		{ 34, 385, 41 },
		{ 34, 386, 41 },
		{ 34, 387, 41 },
		{ 34, 388, 41 },
		{ 34, 389, 41 },
		{ 34, 390, 41 },
		{ 34, 391, 41 },
		{ 34, 392, 41 },
		{ 34, 393, 41 },
		{ 34, 394, 41 },
		{ 34, 395, 41 },
		{ 34, 396, 41 },
		{ 34, 397, 41 },
		{ 34, 398, 41 },
		{ 34, 399, 41 },
		{ 34, 400, 41 },
		{ 34, 401, 41 },
		{ 34, 402, 41 },
		{ 34, 403, 41 },
		{ 34, 404, 41 },
		{ 34, 405, 41 },
		{ 34, 406, 41 },
		{ 34, 407, 41 },
		{ 34, 408, 41 },
		{ 34, 409, 41 },
		{ 34, 410, 41 },
		{ 34, 411, 41 },
		{ 34, 412, 41 },
		{ 34, 413, 41 },
		{ 34, 414, 41 },
		{ 34, 415, 41 },
		{ 34, 416, 41 },
		{ 34, 417, 41 },
		{ 34, 418, 41 },
		{ 34, 419, 41 },
		{ 34, 420, 41 },
		{ 34, 421, 41 },
		{ 34, 422, 41 },
		{ 34, 423, 41 },
		{ 34, 424, 41 },
		{ 35, 369, 41 },
		{ 35, 370, 41 },
		{ 35, 371, 41 },
		{ 35, 372, 41 },
		{ 35, 373, 41 },
		{ 35, 374, 41 },
		{ 35, 375, 41 },
		{ 35, 376, 41 },
		{ 35, 377, 41 },
		{ 35, 378, 41 },
		{ 35, 379, 41 },
		{ 35, 380, 41 },
		{ 35, 381, 41 },
		{ 35, 382, 41 },
		{ 35, 383, 41 },
		{ 35, 384, 41 },
		{ 35, 385, 41 },
		{ 35, 386, 41 },
		{ 35, 387, 41 },
		{ 35, 388, 41 },
		{ 35, 389, 41 },
		{ 35, 390, 41 },
		{ 35, 391, 41 },
		{ 35, 392, 41 },
		{ 35, 393, 41 },
		{ 35, 394, 41 },
		{ 35, 395, 41 },
		{ 35, 396, 41 },
		{ 35, 397, 41 },
		{ 35, 398, 41 },
		{ 35, 399, 41 },
		{ 35, 400, 41 },
		{ 35, 401, 41 },
		{ 35, 402, 41 },
		{ 35, 403, 41 },
		{ 35, 404, 41 },
		{ 35, 405, 41 },
		{ 35, 406, 41 },
		{ 35, 407, 41 },
		{ 35, 408, 41 },
		{ 35, 409, 41 },
		{ 35, 410, 41 },
		{ 35, 411, 41 },
		{ 35, 412, 41 },
		{ 35, 413, 41 },
		{ 35, 414, 41 },
		{ 35, 415, 41 },
		{ 35, 416, 41 },
		{ 35, 417, 41 },
		{ 35, 418, 41 },
		{ 35, 419, 41 },
		{ 35, 420, 41 },
		{ 35, 421, 41 },
		{ 35, 422, 41 },
		{ 35, 423, 41 },
		{ 36, 368, 41 },
		{ 36, 369, 41 },
		{ 36, 370, 41 },
		{ 36, 371, 41 },
		{ 36, 372, 41 },
		{ 36, 373, 41 },
		{ 36, 374, 41 },
		{ 36, 375, 41 },
		{ 36, 376, 41 },
		{ 36, 377, 41 },
		{ 36, 378, 41 },
		{ 36, 379, 41 },
		{ 36, 380, 41 },
		{ 36, 381, 41 },
		{ 36, 382, 41 },
		{ 36, 383, 41 },
		{ 36, 384, 41 },
		{ 36, 385, 41 },
		{ 36, 386, 41 },
		{ 36, 387, 41 },
		{ 36, 388, 41 },
		{ 36, 389, 41 },
		{ 36, 390, 41 },
		{ 36, 391, 41 },
		{ 36, 392, 41 },
		{ 36, 393, 41 },
		{ 36, 394, 41 },
		{ 36, 395, 41 },
		{ 36, 396, 41 },
		{ 36, 397, 41 },
		{ 36, 398, 41 },
		{ 36, 399, 41 },
		{ 36, 400, 41 },
		{ 36, 401, 41 },
		{ 36, 402, 41 },
		{ 36, 403, 41 },
		{ 36, 404, 41 },
		{ 36, 405, 41 },
		{ 36, 406, 41 },
		{ 36, 407, 41 },
		{ 36, 408, 41 },
		{ 36, 409, 41 },
		{ 36, 410, 41 },
		{ 36, 411, 41 },
		{ 36, 412, 41 },
		{ 36, 413, 41 },
		{ 36, 414, 41 },
		{ 36, 415, 41 },
		{ 36, 416, 41 },
		{ 36, 417, 41 },
		{ 36, 418, 41 },
		{ 36, 419, 41 },
		{ 36, 420, 41 },
		{ 36, 421, 41 },
		{ 36, 422, 41 },
		{ 37, 368, 41 },
		{ 37, 369, 41 },
		{ 37, 370, 41 },
		{ 37, 371, 41 },
		{ 37, 372, 41 },
		{ 37, 373, 41 },
		{ 37, 374, 41 },
		{ 37, 375, 41 },
		{ 37, 376, 41 },
		{ 37, 377, 41 },
		{ 37, 378, 41 },
		{ 37, 379, 41 },
		{ 37, 380, 41 },
		{ 37, 381, 41 },
		{ 37, 382, 41 },
		{ 37, 383, 41 },
		{ 37, 384, 41 },
		{ 37, 385, 41 },
		{ 37, 386, 41 },
		{ 37, 387, 41 },
		{ 37, 388, 41 },
		{ 37, 389, 41 },
		{ 37, 390, 41 },
		{ 37, 391, 41 },
		{ 37, 392, 41 },
		{ 37, 393, 41 },
		{ 37, 394, 41 },
		{ 37, 395, 41 },
		{ 37, 396, 41 },
		{ 37, 397, 41 },
		{ 37, 398, 41 },
		{ 37, 399, 41 },
		{ 37, 400, 41 },
		{ 37, 401, 41 },
		{ 37, 402, 41 },
		{ 37, 403, 41 },
		{ 37, 404, 41 },
		{ 37, 405, 41 },
		{ 37, 406, 41 },
		{ 37, 407, 41 },
		{ 37, 408, 41 },
		{ 37, 409, 41 },
		{ 37, 410, 41 },
		{ 37, 411, 41 },
		{ 37, 412, 41 },
		{ 37, 413, 41 },
		{ 37, 414, 41 },
		{ 37, 415, 41 },
		{ 37, 416, 41 },
		{ 37, 417, 41 },
		{ 37, 418, 41 },
		{ 37, 419, 41 },
		{ 37, 420, 41 },
		{ 37, 421, 41 },
		{ 38, 367, 41 },
		{ 38, 368, 41 },
		{ 38, 369, 41 },
		{ 38, 370, 41 },
		{ 38, 371, 41 },
		{ 38, 372, 41 },
		{ 38, 373, 41 },
		{ 38, 374, 41 },
		{ 38, 375, 41 },
		{ 38, 376, 41 },
		{ 38, 377, 41 },
		{ 38, 378, 41 },
		{ 38, 379, 41 },
		{ 38, 380, 41 },
		{ 38, 381, 41 },
		{ 38, 382, 41 },
		{ 38, 383, 41 },
		{ 38, 384, 41 },
		{ 38, 385, 41 },
		{ 38, 386, 41 },
		{ 38, 387, 41 },
		{ 38, 388, 41 },
		{ 38, 389, 41 },
		{ 38, 390, 41 },
		{ 38, 391, 41 },
		{ 38, 392, 41 },
		{ 38, 393, 41 },
		{ 38, 394, 41 },
		{ 38, 395, 41 },
		{ 38, 396, 41 },
		{ 38, 397, 41 },
		{ 38, 398, 41 },
		{ 38, 399, 41 },
		{ 38, 400, 41 },
		{ 38, 401, 41 },
		{ 38, 402, 41 },
		{ 38, 403, 41 },
		{ 38, 404, 41 },
		{ 38, 405, 41 },
		{ 38, 406, 41 },
		{ 38, 407, 41 },
		{ 38, 408, 41 },
		{ 38, 409, 41 },
		{ 38, 410, 41 },
		{ 38, 411, 41 },
		{ 38, 412, 41 },
		{ 38, 413, 41 },
		{ 38, 414, 41 },
		{ 38, 415, 41 },
		{ 38, 416, 41 },
		{ 38, 417, 41 },
		{ 38, 418, 41 },
		{ 38, 419, 41 },
		{ 38, 420, 41 },
		{ 38, 421, 41 },
		{ 39, 367, 41 },
		{ 39, 368, 41 },
		{ 39, 369, 41 },
		{ 39, 370, 41 },
		{ 39, 371, 41 },
		{ 39, 372, 41 },
		{ 39, 373, 41 },
		{ 39, 374, 41 },
		{ 39, 375, 41 },
		{ 39, 376, 41 },
		{ 39, 377, 41 },
		{ 39, 378, 41 },
		{ 39, 379, 41 },
		{ 39, 380, 41 },
		{ 39, 381, 41 },
		{ 39, 382, 41 },
		{ 39, 383, 41 },
		{ 39, 384, 41 },
		{ 39, 385, 41 },
		{ 39, 386, 41 },
		{ 39, 387, 41 },
		{ 39, 388, 41 },
		{ 39, 389, 41 },
		{ 39, 390, 41 },
		{ 39, 391, 41 },
		{ 39, 392, 41 },
		{ 39, 393, 41 },
		{ 39, 394, 41 },
		{ 39, 395, 41 },
		{ 39, 396, 41 },
		{ 39, 397, 41 },
		{ 39, 398, 41 },
		{ 39, 399, 41 },
		{ 39, 400, 41 },
		{ 39, 401, 41 },
		{ 39, 402, 41 },
		{ 39, 403, 41 },
		{ 39, 404, 41 },
		{ 39, 405, 41 },
		{ 39, 406, 41 },
		{ 39, 407, 41 },
		{ 39, 408, 41 },
		{ 39, 409, 41 },
		{ 39, 410, 41 },
		{ 39, 411, 41 },
		{ 39, 412, 41 },
		{ 39, 413, 41 },
		{ 39, 414, 41 },
		{ 39, 415, 41 },
		{ 39, 416, 41 },
		{ 39, 417, 41 },
		{ 39, 418, 41 },
		{ 39, 419, 41 },
		{ 39, 420, 41 },
		{ 40, 366, 41 },
		{ 40, 367, 41 },
		{ 40, 368, 41 },
		{ 40, 369, 41 },
		{ 40, 370, 41 },
		{ 40, 371, 41 },
		{ 40, 372, 41 },
		{ 40, 373, 41 },
		{ 40, 374, 41 },
		{ 40, 375, 41 },
		{ 40, 376, 41 },
		{ 40, 377, 41 },
		{ 40, 378, 41 },
		{ 40, 379, 41 },
		{ 40, 380, 41 },
		{ 40, 381, 41 },
		{ 40, 382, 41 },
		{ 40, 383, 41 },
		{ 40, 384, 41 },
		{ 40, 385, 41 },
		{ 40, 386, 41 },
		{ 40, 387, 41 },
		{ 40, 388, 41 },
		{ 40, 389, 41 },
		{ 40, 390, 41 },
		{ 40, 391, 41 },
		{ 40, 392, 41 },
		{ 40, 393, 41 },
		{ 40, 394, 41 },
		{ 40, 395, 41 },
		{ 40, 396, 41 },
		{ 40, 397, 41 },
		{ 40, 398, 41 },
		{ 40, 399, 41 },
		{ 40, 400, 41 },
		{ 40, 401, 41 },
		{ 40, 402, 41 },
		{ 40, 403, 41 },
		{ 40, 404, 41 },
		{ 40, 405, 41 },
		{ 40, 406, 41 },
		{ 40, 407, 41 },
		{ 40, 408, 41 },
		{ 40, 409, 41 },
		{ 40, 410, 41 },
		{ 40, 411, 41 },
		{ 40, 412, 41 },
		{ 40, 413, 41 },
		{ 40, 414, 41 },
		{ 40, 415, 41 },
		{ 40, 416, 41 },
		{ 40, 417, 41 },
		{ 40, 418, 41 },
		{ 40, 419, 41 },
		{ 41, 366, 41 },
		{ 41, 367, 41 },
		{ 41, 368, 41 },
		{ 41, 369, 41 },
		{ 41, 370, 41 },
		{ 41, 371, 41 },
		{ 41, 372, 41 },
		{ 41, 373, 41 },
		{ 41, 374, 41 },
		{ 41, 375, 41 },
		{ 41, 376, 41 },
		{ 41, 377, 41 },
		{ 41, 378, 41 },
		{ 41, 379, 41 },
		{ 41, 380, 41 },
		{ 41, 381, 41 },
		{ 41, 382, 41 },
		{ 41, 383, 41 },
		{ 41, 384, 41 },
		{ 41, 385, 41 },
		{ 41, 386, 41 },
		{ 41, 387, 41 },
		{ 41, 388, 41 },
		{ 41, 389, 41 },
		{ 41, 390, 41 },
		{ 41, 391, 41 },
		{ 41, 392, 41 },
		{ 41, 393, 41 },
		{ 41, 394, 41 },
		{ 41, 395, 41 },
		{ 41, 396, 41 },
		{ 41, 397, 41 },
		{ 41, 398, 41 },
		{ 41, 399, 41 },
		{ 41, 400, 41 },
		{ 41, 401, 41 },
		{ 41, 402, 41 },
		{ 41, 403, 41 },
		{ 41, 404, 41 },
		{ 41, 405, 41 },
		{ 41, 406, 41 },
		{ 41, 407, 41 },
		{ 41, 408, 41 },
		{ 41, 409, 41 },
		{ 41, 410, 41 },
		{ 41, 411, 41 },
		{ 41, 412, 41 },
		{ 41, 413, 41 },
		{ 41, 414, 41 },
		{ 41, 415, 41 },
		{ 41, 416, 41 },
		{ 41, 417, 41 },
		{ 41, 418, 41 },
		{ 42, 365, 41 },
		{ 42, 366, 41 },
		{ 42, 367, 41 },
		{ 42, 368, 41 },
		{ 42, 369, 41 },
		{ 42, 370, 41 },
		{ 42, 371, 41 },
		{ 42, 372, 41 },
		{ 42, 373, 41 },
		{ 42, 374, 41 },
		{ 42, 375, 41 },
		{ 42, 376, 41 },
		{ 42, 377, 41 },
		{ 42, 378, 41 },
		{ 42, 379, 41 },
		{ 42, 380, 41 },
		{ 42, 381, 41 },
		{ 42, 382, 41 },
		{ 42, 383, 41 },
		{ 42, 384, 41 },
		{ 42, 385, 41 },
		{ 42, 386, 41 },
		{ 42, 387, 41 },
		{ 42, 388, 41 },
		{ 42, 389, 41 },
		{ 42, 390, 41 },
		{ 42, 391, 41 },
		{ 42, 392, 41 },
		{ 42, 393, 41 },
		{ 42, 394, 41 },
		{ 42, 395, 41 },
		{ 42, 396, 41 },
		{ 42, 397, 41 },
		{ 42, 398, 41 },
		{ 42, 399, 41 },
		{ 42, 400, 41 },
		{ 42, 401, 41 },
		{ 42, 402, 41 },
		{ 42, 403, 41 },
		{ 42, 404, 41 },
		{ 42, 405, 41 },
		{ 42, 406, 41 },
		{ 42, 407, 41 },
		{ 42, 408, 41 },
		{ 42, 409, 41 },
		{ 42, 410, 41 },
		{ 42, 411, 41 },
		{ 42, 412, 41 },
		{ 42, 413, 41 },
		{ 42, 414, 41 },
		{ 42, 415, 41 },
		{ 42, 416, 41 },
		{ 42, 417, 41 },
		{ 42, 418, 41 },
		{ 43, 365, 41 },
		{ 43, 366, 41 },
		{ 43, 367, 41 },
		{ 43, 368, 41 },
		{ 43, 369, 41 },
		{ 43, 370, 41 },
		{ 43, 371, 41 },
		{ 43, 372, 41 },
		{ 43, 373, 41 },
		{ 43, 374, 41 },
		{ 43, 375, 41 },
		{ 43, 376, 41 },
		{ 43, 377, 41 },
		{ 43, 378, 41 },
		{ 43, 379, 41 },
		{ 43, 380, 41 },
		{ 43, 381, 41 },
		{ 43, 382, 41 },
		{ 43, 383, 41 },
		{ 43, 384, 41 },
		{ 43, 385, 41 },
		{ 43, 386, 41 },
		{ 43, 387, 41 },
		{ 43, 388, 41 },
		{ 43, 389, 41 },
		{ 43, 390, 41 },
		{ 43, 391, 41 },
		{ 43, 392, 41 },
		{ 43, 393, 41 },
		{ 43, 394, 41 },
		{ 43, 395, 41 },
		{ 43, 396, 41 },
		{ 43, 397, 41 },
		{ 43, 398, 41 },
		{ 43, 399, 41 },
		{ 43, 400, 41 },
		{ 43, 401, 41 },
		{ 43, 402, 41 },
		{ 43, 403, 41 },
		{ 43, 404, 41 },
		{ 43, 405, 41 },
		{ 43, 406, 41 },
		{ 43, 407, 41 },
		{ 43, 408, 41 },
		{ 43, 409, 41 },
		{ 43, 410, 41 },
		{ 43, 411, 41 },
		{ 43, 412, 41 },
		{ 43, 413, 41 },
		{ 43, 414, 41 },
		{ 43, 415, 41 },
		{ 43, 416, 41 },
		{ 43, 417, 41 },
		{ 44, 364, 41 },
		{ 44, 365, 41 },
		{ 44, 366, 41 },
		{ 44, 367, 41 },
		{ 44, 368, 41 },
		{ 44, 369, 41 },
		{ 44, 370, 41 },
		{ 44, 371, 41 },
		{ 44, 372, 41 },
		{ 44, 373, 41 },
		{ 44, 374, 41 },
		{ 44, 375, 41 },
		{ 44, 376, 41 },
		{ 44, 377, 41 },
		{ 44, 378, 41 },
		{ 44, 379, 41 },
		{ 44, 380, 41 },
		{ 44, 381, 41 },
		{ 44, 382, 41 },
		{ 44, 383, 41 },
		{ 44, 384, 41 },
		{ 44, 385, 41 },
		{ 44, 386, 41 },
		{ 44, 387, 41 },
		{ 44, 388, 41 },
		{ 44, 389, 41 },
		{ 44, 390, 41 },
		{ 44, 391, 41 },
		{ 44, 392, 41 },
		{ 44, 393, 41 },
		{ 44, 394, 41 },
		{ 44, 395, 41 },
		{ 44, 396, 41 },
		{ 44, 397, 41 },
		{ 44, 398, 41 },
		{ 44, 399, 41 },
		{ 44, 400, 41 },
		{ 44, 401, 41 },
		{ 44, 402, 41 },
		{ 44, 403, 41 },
		{ 44, 404, 41 },
		{ 44, 405, 41 },
		{ 44, 406, 41 },
		{ 44, 407, 41 },
		{ 44, 408, 41 },
		{ 44, 409, 41 },
		{ 44, 410, 41 },
		{ 44, 411, 41 },
		{ 44, 412, 41 },
		{ 44, 413, 41 },
		{ 44, 414, 41 },
		{ 44, 415, 41 },
		{ 44, 416, 41 },
		{ 45, 364, 41 },
		{ 45, 365, 41 },
		{ 45, 366, 41 },
		{ 45, 367, 41 },
		{ 45, 368, 41 },
		{ 45, 369, 41 },
		{ 45, 370, 41 },
		{ 45, 371, 41 },
		{ 45, 372, 41 },
		{ 45, 373, 41 },
		{ 45, 374, 41 },
		{ 45, 375, 41 },
		{ 45, 376, 41 },
		{ 45, 377, 41 },
		{ 45, 378, 41 },
		{ 45, 379, 41 },
		{ 45, 380, 41 },
		{ 45, 381, 41 },
		{ 45, 382, 41 },
		{ 45, 383, 41 },
		{ 45, 384, 41 },
		{ 45, 385, 41 },
		{ 45, 386, 41 },
		{ 45, 387, 41 },
		{ 45, 388, 41 },
		{ 45, 389, 41 },
		{ 45, 390, 41 },
		{ 45, 391, 41 },
		{ 45, 392, 41 },
		{ 45, 393, 41 },
		{ 45, 394, 41 },
		{ 45, 395, 41 },
		{ 45, 396, 41 },
		{ 45, 397, 41 },
		{ 45, 398, 41 },
		{ 45, 399, 41 },
		{ 45, 400, 41 },
		{ 45, 401, 41 },
		{ 45, 402, 41 },
		{ 45, 403, 41 },
		{ 45, 404, 41 },
		{ 45, 405, 41 },
		{ 45, 406, 41 },
		{ 45, 407, 41 },
		{ 45, 408, 41 },
		{ 45, 409, 41 },
		{ 45, 410, 41 },
		{ 45, 411, 41 },
		{ 45, 412, 41 },
		{ 45, 413, 41 },
		{ 45, 414, 41 },
		{ 45, 415, 41 },
		{ 46, 363, 41 },
		{ 46, 364, 41 },
		{ 46, 365, 41 },
		{ 46, 366, 41 },
		{ 46, 367, 41 },
		{ 46, 368, 41 },
		{ 46, 369, 41 },
		{ 46, 370, 41 },
		{ 46, 371, 41 },
		{ 46, 372, 41 },
		{ 46, 373, 41 },
		{ 46, 374, 41 },
		{ 46, 375, 41 },
		{ 46, 376, 41 },
		{ 46, 377, 41 },
		{ 46, 378, 41 },
		{ 46, 379, 41 },
		{ 46, 380, 41 },
		{ 46, 381, 41 },
		{ 46, 382, 41 },
		{ 46, 383, 41 },
		{ 46, 384, 41 },
		{ 46, 385, 41 },
		{ 46, 386, 41 },
		{ 46, 387, 41 },
		{ 46, 388, 41 },
		{ 46, 389, 41 },
		{ 46, 390, 41 },
		{ 46, 391, 41 },
		{ 46, 392, 41 },
		{ 46, 393, 41 },
		{ 46, 394, 41 },
		{ 46, 395, 41 },
		{ 46, 396, 41 },
		{ 46, 397, 41 },
		{ 46, 398, 41 },
		{ 46, 399, 41 },
		{ 46, 400, 41 },
		{ 46, 401, 41 },
		{ 46, 402, 41 },
		{ 46, 403, 41 },
		{ 46, 404, 41 },
		{ 46, 405, 41 },
		{ 46, 406, 41 },
		{ 46, 407, 41 },
		{ 46, 408, 41 },
		{ 46, 409, 41 },
		{ 46, 410, 41 },
		{ 46, 411, 41 },
		{ 46, 412, 41 },
		{ 46, 413, 41 },
		{ 46, 414, 41 },
		{ 46, 415, 41 },
		{ 47, 363, 41 },
		{ 47, 364, 41 },
		{ 47, 365, 41 },
		{ 47, 366, 41 },
		{ 47, 367, 41 },
		{ 47, 368, 41 },
		{ 47, 369, 41 },
		{ 47, 370, 41 },
		{ 47, 371, 41 },
		{ 47, 372, 41 },
		{ 47, 373, 41 },
		{ 47, 374, 41 },
		{ 47, 375, 41 },
		{ 47, 376, 41 },
		{ 47, 377, 41 },
		{ 47, 378, 41 },
		{ 47, 379, 41 },
		{ 47, 380, 41 },
		{ 47, 381, 41 },
		{ 47, 382, 41 },
		{ 47, 383, 41 },
		{ 47, 384, 41 },
		{ 47, 385, 41 },
		{ 47, 386, 41 },
		{ 47, 387, 41 },
		{ 47, 388, 41 },
		{ 47, 389, 41 },
		{ 47, 390, 41 },
		{ 47, 391, 41 },
		{ 47, 392, 41 },
		{ 47, 393, 41 },
		{ 47, 394, 41 },
		{ 47, 395, 41 },
		{ 47, 396, 41 },
		{ 47, 397, 41 },
		{ 47, 398, 41 },
		{ 47, 399, 41 },
		{ 47, 400, 41 },
		{ 47, 401, 41 },
		{ 47, 402, 41 },
		{ 47, 403, 41 },
		{ 47, 404, 41 },
		{ 47, 405, 41 },
		{ 47, 406, 41 },
		{ 47, 407, 41 },
		{ 47, 408, 41 },
		{ 47, 409, 41 },
		{ 47, 410, 41 },
		{ 47, 411, 41 },
		{ 47, 412, 41 },
		{ 47, 413, 41 },
		{ 47, 414, 41 },
		{ 48, 362, 41 },
		{ 48, 363, 41 },
		{ 48, 364, 41 },
		{ 48, 365, 41 },
		{ 48, 366, 41 },
		{ 48, 367, 41 },
		{ 48, 368, 41 },
		{ 48, 369, 41 },
		{ 48, 370, 41 },
		{ 48, 371, 41 },
		{ 48, 372, 41 },
		{ 48, 373, 41 },
		{ 48, 374, 41 },
		{ 48, 375, 41 },
		{ 48, 376, 41 },
		{ 48, 377, 41 },
		{ 48, 378, 41 },
		{ 48, 379, 41 },
		{ 48, 380, 41 },
		{ 48, 381, 41 },
		{ 48, 382, 41 },
		{ 48, 383, 41 },
		{ 48, 384, 41 },
		{ 48, 385, 41 },
		{ 48, 386, 41 },
		{ 48, 387, 41 },
		{ 48, 388, 41 },
		{ 48, 389, 41 },
		{ 48, 390, 41 },
		{ 48, 391, 41 },
		{ 48, 392, 41 },
		{ 48, 393, 41 },
		{ 48, 394, 41 },
		{ 48, 395, 41 },
		{ 48, 396, 41 },
		{ 48, 397, 41 },
		{ 48, 398, 41 },
		{ 48, 399, 41 },
		{ 48, 400, 41 },
		{ 48, 401, 41 },
		{ 48, 402, 41 },
		{ 48, 403, 41 },
		{ 48, 404, 41 },
		{ 48, 405, 41 },
		{ 48, 406, 41 },
		{ 48, 407, 41 },
		{ 48, 408, 41 },
		{ 48, 409, 41 },
		{ 48, 410, 41 },
		{ 48, 411, 41 },
		{ 48, 412, 41 },
		{ 48, 413, 41 },
		{ 49, 362, 41 },
		{ 49, 363, 41 },
		{ 49, 364, 41 },
		{ 49, 365, 41 },
		{ 49, 366, 41 },
		{ 49, 367, 41 },
		{ 49, 368, 41 },
		{ 49, 369, 41 },
		{ 49, 370, 41 },
		{ 49, 371, 41 },
		{ 49, 372, 41 },
		{ 49, 373, 41 },
		{ 49, 374, 41 },
		{ 49, 375, 41 },
		{ 49, 376, 41 },
		{ 49, 377, 41 },
		{ 49, 378, 41 },
		{ 49, 379, 41 },
		{ 49, 380, 41 },
		{ 49, 381, 41 },
		{ 49, 382, 41 },
		{ 49, 383, 41 },
		{ 49, 384, 41 },
		{ 49, 385, 41 },
		{ 49, 386, 41 },
		{ 49, 387, 41 },
		{ 49, 388, 41 },
		{ 49, 389, 41 },
		{ 49, 390, 41 },
		{ 49, 391, 41 },
		{ 49, 392, 41 },
		{ 49, 393, 41 },
		{ 49, 394, 41 },
		{ 49, 395, 41 },
		{ 49, 396, 41 },
		{ 49, 397, 41 },
		{ 49, 398, 41 },
		{ 49, 399, 41 },
		{ 49, 400, 41 },
		{ 49, 401, 41 },
		{ 49, 402, 41 },
		{ 49, 403, 41 },
		{ 49, 404, 41 },
		{ 49, 405, 41 },
		{ 49, 406, 41 },
		{ 49, 407, 41 },
		{ 49, 408, 41 },
		{ 49, 409, 41 },
		{ 49, 410, 41 },
		{ 49, 411, 41 },
		{ 49, 412, 41 },
		{ 50, 361, 41 },
		{ 50, 362, 41 },
		{ 50, 363, 41 },
		{ 50, 364, 41 },
		{ 50, 365, 41 },
		{ 50, 366, 41 },
		{ 50, 367, 41 },
		{ 50, 368, 41 },
		{ 50, 369, 41 },
		{ 50, 370, 41 },
		{ 50, 371, 41 },
		{ 50, 372, 41 },
		{ 50, 373, 41 },
		{ 50, 374, 41 },
		{ 50, 375, 41 },
		{ 50, 376, 41 },
		{ 50, 377, 41 },
		{ 50, 378, 41 },
		{ 50, 379, 41 },
		{ 50, 380, 41 },
		{ 50, 381, 41 },
		{ 50, 382, 41 },
		{ 50, 383, 41 },
		{ 50, 384, 41 },
		{ 50, 385, 41 },
		{ 50, 386, 41 },
		{ 50, 387, 41 },
		{ 50, 388, 41 },
		{ 50, 389, 41 },
		{ 50, 390, 41 },
		{ 50, 391, 41 },
		{ 50, 392, 41 },
		{ 50, 393, 41 },
		{ 50, 394, 41 },
		{ 50, 395, 41 },
		{ 50, 396, 41 },
		{ 50, 397, 41 },
		{ 50, 398, 41 },
		{ 50, 399, 41 },
		{ 50, 400, 41 },
		{ 50, 401, 41 },
		{ 50, 402, 41 },
		{ 50, 403, 41 },
		{ 50, 404, 41 },
		{ 50, 405, 41 },
		{ 50, 406, 41 },
		{ 50, 407, 41 },
		{ 50, 408, 41 },
		{ 50, 409, 41 },
		{ 50, 410, 41 },
		{ 50, 411, 41 },
		{ 50, 412, 41 },
		{ 51, 361, 41 },
		{ 51, 362, 41 },
		{ 51, 363, 41 },
		{ 51, 364, 41 },
		{ 51, 365, 41 },
		{ 51, 366, 41 },
		{ 51, 367, 41 },
		{ 51, 368, 41 },
		{ 51, 369, 41 },
		{ 51, 370, 41 },
		{ 51, 371, 41 },
		{ 51, 372, 41 },
		{ 51, 373, 41 },
		{ 51, 374, 41 },
		{ 51, 375, 41 },
		{ 51, 376, 41 },
		{ 51, 377, 41 },
		{ 51, 378, 41 },
		{ 51, 379, 41 },
		{ 51, 380, 41 },
		{ 51, 381, 41 },
		{ 51, 382, 41 },
		{ 51, 383, 41 },
		{ 51, 384, 41 },
		{ 51, 385, 41 },
		{ 51, 386, 41 },
		{ 51, 387, 41 },
		{ 51, 388, 41 },
		{ 51, 389, 41 },
		{ 51, 390, 41 },
		{ 51, 391, 41 },
		{ 51, 392, 41 },
		{ 51, 393, 41 },
		{ 51, 394, 41 },
		{ 51, 395, 41 },
		{ 51, 396, 41 },
		{ 51, 397, 41 },
		{ 51, 398, 41 },
		{ 51, 399, 41 },
		{ 51, 400, 41 },
		{ 51, 401, 41 },
		{ 51, 402, 41 },
		{ 51, 403, 41 },
		{ 51, 404, 41 },
		{ 51, 405, 41 },
		{ 51, 406, 41 },
		{ 51, 407, 41 },
		{ 51, 408, 41 },
		{ 51, 409, 41 },
		{ 51, 410, 41 },
		{ 51, 411, 41 },
		{ 52, 360, 41 },
		{ 52, 361, 41 },
		{ 52, 362, 41 },
		{ 52, 363, 41 },
		{ 52, 364, 41 },
		{ 52, 365, 41 },
		{ 52, 366, 41 },
		{ 52, 367, 41 },
		{ 52, 368, 41 },
		{ 52, 369, 41 },
		{ 52, 370, 41 },
		{ 52, 371, 41 },
		{ 52, 372, 41 },
		{ 52, 373, 41 },
		{ 52, 374, 41 },
		{ 52, 375, 41 },
		{ 52, 376, 41 },
		{ 52, 377, 41 },
		{ 52, 378, 41 },
		{ 52, 379, 41 },
		{ 52, 380, 41 },
		{ 52, 381, 41 },
		{ 52, 382, 41 },
		{ 52, 383, 41 },
		{ 52, 384, 41 },
		{ 52, 385, 41 },
		{ 52, 386, 41 },
		{ 52, 387, 41 },
		{ 52, 388, 41 },
		{ 52, 389, 41 },
		{ 52, 390, 41 },
		{ 52, 391, 41 },
		{ 52, 392, 41 },
		{ 52, 393, 41 },
		{ 52, 394, 41 },
		{ 52, 395, 41 },
		{ 52, 396, 41 },
		{ 52, 397, 41 },
		{ 52, 398, 41 },
		{ 52, 399, 41 },
		{ 52, 400, 41 },
		{ 52, 401, 41 },
		{ 52, 402, 41 },
		{ 52, 403, 41 },
		{ 52, 404, 41 },
		{ 52, 405, 41 },
		{ 52, 406, 41 },
		{ 52, 407, 41 },
		{ 52, 408, 41 },
		{ 52, 409, 41 },
		{ 52, 410, 41 },
		{ 53, 360, 41 },
		{ 53, 361, 41 },
		{ 53, 362, 41 },
		{ 53, 363, 41 },
		{ 53, 364, 41 },
		{ 53, 365, 41 },
		{ 53, 366, 41 },
		{ 53, 367, 41 },
		{ 53, 368, 41 },
		{ 53, 369, 41 },
		{ 53, 370, 41 },
		{ 53, 371, 41 },
		{ 53, 372, 41 },
		{ 53, 373, 41 },
		{ 53, 374, 41 },
		{ 53, 375, 41 },
		{ 53, 376, 41 },
		{ 53, 377, 41 },
		{ 53, 378, 41 },
		{ 53, 379, 41 },
		{ 53, 380, 41 },
		{ 53, 381, 41 },
		{ 53, 382, 41 },
		{ 53, 383, 41 },
		{ 53, 384, 41 },
		{ 53, 385, 41 },
		{ 53, 386, 41 },
		{ 53, 387, 41 },
		{ 53, 388, 41 },
		{ 53, 389, 41 },
		{ 53, 390, 41 },
		{ 53, 391, 41 },
		{ 53, 392, 41 },
		{ 53, 393, 41 },
		{ 53, 394, 41 },
		{ 53, 395, 41 },
		{ 53, 396, 41 },
		{ 53, 397, 41 },
		{ 53, 398, 41 },
		{ 53, 399, 41 },
		{ 53, 400, 41 },
		{ 53, 401, 41 },
		{ 53, 402, 41 },
		{ 53, 403, 41 },
		{ 53, 404, 41 },
		{ 53, 405, 41 },
		{ 53, 406, 41 },
		{ 53, 407, 41 },
		{ 53, 408, 41 },
		{ 53, 409, 41 },
		{ 54, 359, 41 },
		{ 54, 360, 41 },
		{ 54, 361, 41 },
		{ 54, 362, 41 },
		{ 54, 363, 41 },
		{ 54, 364, 41 },
		{ 54, 365, 41 },
		{ 54, 366, 41 },
		{ 54, 367, 41 },
		{ 54, 368, 41 },
		{ 54, 369, 41 },
		{ 54, 370, 41 },
		{ 54, 371, 41 },
		{ 54, 372, 41 },
		{ 54, 373, 41 },
		{ 54, 374, 41 },
		{ 54, 375, 41 },
		{ 54, 376, 41 },
		{ 54, 377, 41 },
		{ 54, 378, 41 },
		{ 54, 379, 41 },
		{ 54, 380, 41 },
		{ 54, 381, 41 },
		{ 54, 382, 41 },
		{ 54, 383, 41 },
		{ 54, 384, 41 },
		{ 54, 385, 41 },
		{ 54, 386, 41 },
		{ 54, 387, 41 },
		{ 54, 388, 41 },
		{ 54, 389, 41 },
		{ 54, 390, 41 },
		{ 54, 391, 41 },
		{ 54, 392, 41 },
		{ 54, 393, 41 },
		{ 54, 394, 41 },
		{ 54, 395, 41 },
		{ 54, 396, 41 },
		{ 54, 397, 41 },
		{ 54, 398, 41 },
		{ 54, 399, 41 },
		{ 54, 400, 41 },
		{ 54, 401, 41 },
		{ 54, 402, 41 },
		{ 54, 403, 41 },
		{ 54, 404, 41 },
		{ 54, 405, 41 },
		{ 54, 406, 41 },
		{ 54, 407, 41 },
		{ 54, 408, 41 },
		{ 54, 409, 41 },
		{ 55, 359, 41 },
		{ 55, 360, 41 },
		{ 55, 361, 41 },
		{ 55, 362, 41 },
		{ 55, 363, 41 },
		{ 55, 364, 41 },
		{ 55, 365, 41 },
		{ 55, 366, 41 },
		{ 55, 367, 41 },
		{ 55, 368, 41 },
		{ 55, 369, 41 },
		{ 55, 370, 41 },
		{ 55, 371, 41 },
		{ 55, 372, 41 },
		{ 55, 373, 41 },
		{ 55, 374, 41 },
		{ 55, 375, 41 },
		{ 55, 376, 41 },
		{ 55, 377, 41 },
		{ 55, 378, 41 },
		{ 55, 379, 41 },
		{ 55, 380, 41 },
		{ 55, 381, 41 },
		{ 55, 382, 41 },
		{ 55, 383, 41 },
		{ 55, 384, 41 },
		{ 55, 385, 41 },
		{ 55, 386, 41 },
		{ 55, 387, 41 },
		{ 55, 388, 41 },
		{ 55, 389, 41 },
		{ 55, 390, 41 },
		{ 55, 391, 41 },
		{ 55, 392, 41 },
		{ 55, 393, 41 },
		{ 55, 394, 41 },
		{ 55, 395, 41 },
		{ 55, 396, 41 },
		{ 55, 397, 41 },
		{ 55, 398, 41 },
		{ 55, 399, 41 },
		{ 55, 400, 41 },
		{ 55, 401, 41 },
		{ 55, 402, 41 },
		{ 55, 403, 41 },
		{ 55, 404, 41 },
		{ 55, 405, 41 },
		{ 55, 406, 41 },
		{ 55, 407, 41 },
		{ 55, 408, 41 },
		{ 56, 358, 41 },
		{ 56, 359, 41 },
		{ 56, 360, 41 },
		{ 56, 361, 41 },
		{ 56, 362, 41 },
		{ 56, 363, 41 },
		{ 56, 364, 41 },
		{ 56, 365, 41 },
		{ 56, 366, 41 },
		{ 56, 367, 41 },
		{ 56, 368, 41 },
		{ 56, 369, 41 },
		{ 56, 370, 41 },
		{ 56, 371, 41 },
		{ 56, 372, 41 },
		{ 56, 373, 41 },
		{ 56, 374, 41 },
		{ 56, 375, 41 },
		{ 56, 376, 41 },
		{ 56, 377, 41 },
		{ 56, 378, 41 },
		{ 56, 379, 41 },
		{ 56, 380, 41 },
		{ 56, 381, 41 },
		{ 56, 382, 41 },
		{ 56, 383, 41 },
		{ 56, 384, 41 },
		{ 56, 385, 41 },
		{ 56, 386, 41 },
		{ 56, 387, 41 },
		{ 56, 388, 41 },
		{ 56, 389, 41 },
		{ 56, 390, 41 },
		{ 56, 391, 41 },
		{ 56, 392, 41 },
		{ 56, 393, 41 },
		{ 56, 394, 41 },
		{ 56, 395, 41 },
		{ 56, 396, 41 },
		{ 56, 397, 41 },
		{ 56, 398, 41 },
		{ 56, 399, 41 },
		{ 56, 400, 41 },
		{ 56, 401, 41 },
		{ 56, 402, 41 },
		{ 56, 403, 41 },
		{ 56, 404, 41 },
		{ 56, 405, 41 },
		{ 56, 406, 41 },
		{ 56, 407, 41 },
		{ 57, 358, 41 },
		{ 57, 359, 41 },
		{ 57, 360, 41 },
		{ 57, 361, 41 },
		{ 57, 362, 41 },
		{ 57, 363, 41 },
		{ 57, 364, 41 },
		{ 57, 365, 41 },
		{ 57, 366, 41 },
		{ 57, 367, 41 },
		{ 57, 368, 41 },
		{ 57, 369, 41 },
		{ 57, 370, 41 },
		{ 57, 371, 41 },
		{ 57, 372, 41 },
		{ 57, 373, 41 },
		{ 57, 374, 41 },
		{ 57, 375, 41 },
		{ 57, 376, 41 },
		{ 57, 377, 41 },
		{ 57, 378, 41 },
		{ 57, 379, 41 },
		{ 57, 380, 41 },
		{ 57, 381, 41 },
		{ 57, 382, 41 },
		{ 57, 383, 41 },
		{ 57, 384, 41 },
		{ 57, 385, 41 },
		{ 57, 386, 41 },
		{ 57, 387, 41 },
		{ 57, 388, 41 },
		{ 57, 389, 41 },
		{ 57, 390, 41 },
		{ 57, 391, 41 },
		{ 57, 392, 41 },
		{ 57, 393, 41 },
		{ 57, 394, 41 },
		{ 57, 395, 41 },
		{ 57, 396, 41 },
		{ 57, 397, 41 },
		{ 57, 398, 41 },
		{ 57, 399, 41 },
		{ 57, 400, 41 },
		{ 57, 401, 41 },
		{ 57, 402, 41 },
		{ 57, 403, 41 },
		{ 57, 404, 41 },
		{ 57, 405, 41 },
		{ 57, 406, 41 },
		{ 58, 357, 41 },
		{ 58, 358, 41 },
		{ 58, 359, 41 },
		{ 58, 360, 41 },
		{ 58, 361, 41 },
		{ 58, 362, 41 },
		{ 58, 363, 41 },
		{ 58, 364, 41 },
		{ 58, 365, 41 },
		{ 58, 366, 41 },
		{ 58, 367, 41 },
		{ 58, 368, 41 },
		{ 58, 369, 41 },
		{ 58, 370, 41 },
		{ 58, 371, 41 },
		{ 58, 372, 41 },
		{ 58, 373, 41 },
		{ 58, 374, 41 },
		{ 58, 375, 41 },
		{ 58, 376, 41 },
		{ 58, 377, 41 },
		{ 58, 378, 41 },
		{ 58, 379, 41 },
		{ 58, 380, 41 },
		{ 58, 381, 41 },
		{ 58, 382, 41 },
		{ 58, 383, 41 },
		{ 58, 384, 41 },
		{ 58, 385, 41 },
		{ 58, 386, 41 },
		{ 58, 387, 41 },
		{ 58, 388, 41 },
		{ 58, 389, 41 },
		{ 58, 390, 41 },
		{ 58, 391, 41 },
		{ 58, 392, 41 },
		{ 58, 393, 41 },
		{ 58, 394, 41 },
		{ 58, 395, 41 },
		{ 58, 396, 41 },
		{ 58, 397, 41 },
		{ 58, 398, 41 },
		{ 58, 399, 41 },
		{ 58, 400, 41 },
		{ 58, 401, 41 },
		{ 58, 402, 41 },
		{ 58, 403, 41 },
		{ 58, 404, 41 },
		{ 58, 405, 41 },
		{ 58, 406, 41 },
		{ 59, 357, 41 },
		{ 59, 358, 41 },
		{ 59, 359, 41 },
		{ 59, 360, 41 },
		{ 59, 361, 41 },
		{ 59, 362, 41 },
		{ 59, 363, 41 },
		{ 59, 364, 41 },
		{ 59, 365, 41 },
		{ 59, 366, 41 },
		{ 59, 367, 41 },
		{ 59, 368, 41 },
		{ 59, 369, 41 },
		{ 59, 370, 41 },
		{ 59, 371, 41 },
		{ 59, 372, 41 },
		{ 59, 373, 41 },
		{ 59, 374, 41 },
		{ 59, 375, 41 },
		{ 59, 376, 41 },
		{ 59, 377, 41 },
		{ 59, 378, 41 },
		{ 59, 379, 41 },
		{ 59, 380, 41 },
		{ 59, 381, 41 },
		{ 59, 382, 41 },
		{ 59, 383, 41 },
		{ 59, 384, 41 },
		{ 59, 385, 41 },
		{ 59, 386, 41 },
		{ 59, 387, 41 },
		{ 59, 388, 41 },
		{ 59, 389, 41 },
		{ 59, 390, 41 },
		{ 59, 391, 41 },
		{ 59, 392, 41 },
		{ 59, 393, 41 },
		{ 59, 394, 41 },
		{ 59, 395, 41 },
		{ 59, 396, 41 },
		{ 59, 397, 41 },
		{ 59, 398, 41 },
		{ 59, 399, 41 },
		{ 59, 400, 41 },
		{ 59, 401, 41 },
		{ 59, 402, 41 },
		{ 59, 403, 41 },
		{ 59, 404, 41 },
		{ 59, 405, 41 },
		{ 60, 356, 41 },
		{ 60, 357, 41 },
		{ 60, 358, 41 },
		{ 60, 359, 41 },
		{ 60, 360, 41 },
		{ 60, 361, 41 },
		{ 60, 362, 41 },
		{ 60, 363, 41 },
		{ 60, 364, 41 },
		{ 60, 365, 41 },
		{ 60, 366, 41 },
		{ 60, 367, 41 },
		{ 60, 368, 41 },
		{ 60, 369, 41 },
		{ 60, 370, 41 },
		{ 60, 371, 41 },
		{ 60, 372, 41 },
		{ 60, 373, 41 },
		{ 60, 374, 41 },
		{ 60, 375, 41 },
		{ 60, 376, 41 },
		{ 60, 377, 41 },
		{ 60, 378, 41 },
		{ 60, 379, 41 },
		{ 60, 380, 41 },
		{ 60, 381, 41 },
		{ 60, 382, 41 },
		{ 60, 383, 41 },
		{ 60, 384, 41 },
		{ 60, 385, 41 },
		{ 60, 386, 41 },
		{ 60, 387, 41 },
		{ 60, 388, 41 },
		{ 60, 389, 41 },
		{ 60, 390, 41 },
		{ 60, 391, 41 },
		{ 60, 392, 41 },
		{ 60, 393, 41 },
		{ 60, 394, 41 },
		{ 60, 395, 41 },
		{ 60, 396, 41 },
		{ 60, 397, 41 },
		{ 60, 398, 41 },
		{ 60, 399, 41 },
		{ 60, 400, 41 },
		{ 60, 401, 41 },
		{ 60, 402, 41 },
		{ 60, 403, 41 },
		{ 60, 404, 41 },
		{ 61, 356, 41 },
		{ 61, 357, 41 },
		{ 61, 358, 41 },
		{ 61, 359, 41 },
		{ 61, 360, 41 },
		{ 61, 361, 41 },
		{ 61, 362, 41 },
		{ 61, 363, 41 },
		{ 61, 364, 41 },
		{ 61, 365, 41 },
		{ 61, 366, 41 },
		{ 61, 367, 41 },
		{ 61, 368, 41 },
		{ 61, 369, 41 },
		{ 61, 370, 41 },
		{ 61, 371, 41 },
		{ 61, 372, 41 },
		{ 61, 373, 41 },
		{ 61, 374, 41 },
		{ 61, 375, 41 },
		{ 61, 376, 41 },
		{ 61, 377, 41 },
		{ 61, 378, 41 },
		{ 61, 379, 41 },
		{ 61, 380, 41 },
		{ 61, 381, 41 },
		{ 61, 382, 41 },
		{ 61, 383, 41 },
		{ 61, 384, 41 },
		{ 61, 385, 41 },
		{ 61, 386, 41 },
		{ 61, 387, 41 },
		{ 61, 388, 41 },
		{ 61, 389, 41 },
		{ 61, 390, 41 },
		{ 61, 391, 41 },
		{ 61, 392, 41 },
		{ 61, 393, 41 },
		{ 61, 394, 41 },
		{ 61, 395, 41 },
		{ 61, 396, 41 },
		{ 61, 397, 41 },
		{ 61, 398, 41 },
		{ 61, 399, 41 },
		{ 61, 400, 41 },
		{ 61, 401, 41 },
		{ 61, 402, 41 },
		{ 61, 403, 41 },
		{ 62, 355, 41 },
		{ 62, 356, 41 },
		{ 62, 357, 41 },
		{ 62, 358, 41 },
		{ 62, 359, 41 },
		{ 62, 360, 41 },
		{ 62, 361, 41 },
		{ 62, 362, 41 },
		{ 62, 363, 41 },
		{ 62, 364, 41 },
		{ 62, 365, 41 },
		{ 62, 366, 41 },
		{ 62, 367, 41 },
		{ 62, 368, 41 },
		{ 62, 369, 41 },
		{ 62, 370, 41 },
		{ 62, 371, 41 },
		{ 62, 372, 41 },
		{ 62, 373, 41 },
		{ 62, 374, 41 },
		{ 62, 375, 41 },
		{ 62, 376, 41 },
		{ 62, 377, 41 },
		{ 62, 378, 41 },
		{ 62, 379, 41 },
		{ 62, 380, 41 },
		{ 62, 381, 41 },
		{ 62, 382, 41 },
		{ 62, 383, 41 },
		{ 62, 384, 41 },
		{ 62, 385, 41 },
		{ 62, 386, 41 },
		{ 62, 387, 41 },
		{ 62, 388, 41 },
		{ 62, 389, 41 },
		{ 62, 390, 41 },
		{ 62, 391, 41 },
		{ 62, 392, 41 },
		{ 62, 393, 41 },
		{ 62, 394, 41 },
		{ 62, 395, 41 },
		{ 62, 396, 41 },
		{ 62, 397, 41 },
		{ 62, 398, 41 },
		{ 62, 399, 41 },
		{ 62, 400, 41 },
		{ 62, 401, 41 },
		{ 62, 402, 41 },
		{ 62, 403, 41 },
		{ 63, 355, 41 },
		{ 63, 356, 41 },
		{ 63, 357, 41 },
		{ 63, 358, 41 },
		{ 63, 359, 41 },
		{ 63, 360, 41 },
		{ 63, 361, 41 },
		{ 63, 362, 41 },
		{ 63, 363, 41 },
		{ 63, 364, 41 },
		{ 63, 365, 41 },
		{ 63, 366, 41 },
		{ 63, 367, 41 },
		{ 63, 368, 41 },
		{ 63, 369, 41 },
		{ 63, 370, 41 },
		{ 63, 371, 41 },
		{ 63, 372, 41 },
		{ 63, 373, 41 },
		{ 63, 374, 41 },
		{ 63, 375, 41 },
		{ 63, 376, 41 },
		{ 63, 377, 41 },
		{ 63, 378, 41 },
		{ 63, 379, 41 },
		{ 63, 380, 41 },
		{ 63, 381, 41 },
		{ 63, 382, 41 },
		{ 63, 383, 41 },
		{ 63, 384, 41 },
		{ 63, 385, 41 },
		{ 63, 386, 41 },
		{ 63, 387, 41 },
		{ 63, 388, 41 },
		{ 63, 389, 41 },
		{ 63, 390, 41 },
		{ 63, 391, 41 },
		{ 63, 392, 41 },
		{ 63, 393, 41 },
		{ 63, 394, 41 },
		{ 63, 395, 41 },
		{ 63, 396, 41 },
		{ 63, 397, 41 },
		{ 63, 398, 41 },
		{ 63, 399, 41 },
		{ 63, 400, 41 },
		{ 63, 401, 41 },
		{ 63, 402, 41 },
		{ 64, 354, 41 },
		{ 64, 355, 41 },
		{ 64, 356, 41 },
		{ 64, 357, 41 },
		{ 64, 358, 41 },
		{ 64, 359, 41 },
		{ 64, 360, 41 },
		{ 64, 361, 41 },
		{ 64, 362, 41 },
		{ 64, 363, 41 },
		{ 64, 364, 41 },
		{ 64, 365, 41 },
		{ 64, 366, 41 },
		{ 64, 367, 41 },
		{ 64, 368, 41 },
		{ 64, 369, 41 },
		{ 64, 370, 41 },
		{ 64, 371, 41 },
		{ 64, 372, 41 },
		{ 64, 373, 41 },
		{ 64, 374, 41 },
		{ 64, 375, 41 },
		{ 64, 376, 41 },
		{ 64, 377, 41 },
		{ 64, 378, 41 },
		{ 64, 379, 41 },
		{ 64, 380, 41 },
		{ 64, 381, 41 },
		{ 64, 382, 41 },
		{ 64, 383, 41 },
		{ 64, 384, 41 },
		{ 64, 385, 41 },
		{ 64, 386, 41 },
		{ 64, 387, 41 },
		{ 64, 388, 41 },
		{ 64, 389, 41 },
		{ 64, 390, 41 },
		{ 64, 391, 41 },
		{ 64, 392, 41 },
		{ 64, 393, 41 },
		{ 64, 394, 41 },
		{ 64, 395, 41 },
		{ 64, 396, 41 },
		{ 64, 397, 41 },
		{ 64, 398, 41 },
		{ 64, 399, 41 },
		{ 64, 400, 41 },
		{ 64, 401, 41 },
		{ 65, 354, 41 },
		{ 65, 355, 41 },
		{ 65, 356, 41 },
		{ 65, 357, 41 },
		{ 65, 358, 41 },
		{ 65, 359, 41 },
		{ 65, 360, 41 },
		{ 65, 361, 41 },
		{ 65, 362, 41 },
		{ 65, 363, 41 },
		{ 65, 364, 41 },
		{ 65, 365, 41 },
		{ 65, 366, 41 },
		{ 65, 367, 41 },
		{ 65, 368, 41 },
		{ 65, 369, 41 },
		{ 65, 370, 41 },
		{ 65, 371, 41 },
		{ 65, 372, 41 },
		{ 65, 373, 41 },
		{ 65, 374, 41 },
		{ 65, 375, 41 },
		{ 65, 376, 41 },
		{ 65, 377, 41 },
		{ 65, 378, 41 },
		{ 65, 379, 41 },
		{ 65, 380, 41 },
		{ 65, 381, 41 },
		{ 65, 382, 41 },
		{ 65, 383, 41 },
		{ 65, 384, 41 },
		{ 65, 385, 41 },
		{ 65, 386, 41 },
		{ 65, 387, 41 },
		{ 65, 388, 41 },
		{ 65, 389, 41 },
		{ 65, 390, 41 },
		{ 65, 391, 41 },
		{ 65, 392, 41 },
		{ 65, 393, 41 },
		{ 65, 394, 41 },
		{ 65, 395, 41 },
		{ 65, 396, 41 },
		{ 65, 397, 41 },
		{ 65, 398, 41 },
		{ 65, 399, 41 },
		{ 65, 400, 41 },
		{ 66, 353, 41 },
		{ 66, 354, 41 },
		{ 66, 355, 41 },
		{ 66, 356, 41 },
		{ 66, 357, 41 },
		{ 66, 358, 41 },
		{ 66, 359, 41 },
		{ 66, 360, 41 },
		{ 66, 361, 41 },
		{ 66, 362, 41 },
		{ 66, 363, 41 },
		{ 66, 364, 41 },
		{ 66, 365, 41 },
		{ 66, 366, 41 },
		{ 66, 367, 41 },
		{ 66, 368, 41 },
		{ 66, 369, 41 },
		{ 66, 370, 41 },
		{ 66, 371, 41 },
		{ 66, 372, 41 },
		{ 66, 373, 41 },
		{ 66, 374, 41 },
		{ 66, 375, 41 },
		{ 66, 376, 41 },
		{ 66, 377, 41 },
		{ 66, 378, 41 },
		{ 66, 379, 41 },
		{ 66, 380, 41 },
		{ 66, 381, 41 },
		{ 66, 382, 41 },
		{ 66, 383, 41 },
		{ 66, 384, 41 },
		{ 66, 385, 41 },
		{ 66, 386, 41 },
		{ 66, 387, 41 },
		{ 66, 388, 41 },
		{ 66, 389, 41 },
		{ 66, 390, 41 },
		{ 66, 391, 41 },
		{ 66, 392, 41 },
		{ 66, 393, 41 },
		{ 66, 394, 41 },
		{ 66, 395, 41 },
		{ 66, 396, 41 },
		{ 66, 397, 41 },
		{ 66, 398, 41 },
		{ 66, 399, 41 },
		{ 66, 400, 41 },
		{ 67, 353, 41 },
		{ 67, 354, 41 },
		{ 67, 355, 41 },
		{ 67, 356, 41 },
		{ 67, 357, 41 },
		{ 67, 358, 41 },
		{ 67, 359, 41 },
		{ 67, 360, 41 },
		{ 67, 361, 41 },
		{ 67, 362, 41 },
		{ 67, 363, 41 },
		{ 67, 364, 41 },
		{ 67, 365, 41 },
		{ 67, 366, 41 },
		{ 67, 367, 41 },
		{ 67, 368, 41 },
		{ 67, 369, 41 },
		{ 67, 370, 41 },
		{ 67, 371, 41 },
		{ 67, 372, 41 },
		{ 67, 373, 41 },
		{ 67, 374, 41 },
		{ 67, 375, 41 },
		{ 67, 376, 41 },
		{ 67, 377, 41 },
		{ 67, 378, 41 },
		{ 67, 379, 41 },
		{ 67, 380, 41 },
		{ 67, 381, 41 },
		{ 67, 382, 41 },
		{ 67, 383, 41 },
		{ 67, 384, 41 },
		{ 67, 385, 41 },
		{ 67, 386, 41 },
		{ 67, 387, 41 },
		{ 67, 388, 41 },
		{ 67, 389, 41 },
		{ 67, 390, 41 },
		{ 67, 391, 41 },
		{ 67, 392, 41 },
		{ 67, 393, 41 },
		{ 67, 394, 41 },
		{ 67, 395, 41 },
		{ 67, 396, 41 },
		{ 67, 397, 41 },
		{ 67, 398, 41 },
		{ 67, 399, 41 },
		{ 68, 352, 41 },
		{ 68, 353, 41 },
		{ 68, 354, 41 },
		{ 68, 355, 41 },
		{ 68, 356, 41 },
		{ 68, 357, 41 },
		{ 68, 358, 41 },
		{ 68, 359, 41 },
		{ 68, 360, 41 },
		{ 68, 361, 41 },
		{ 68, 362, 41 },
		{ 68, 363, 41 },
		{ 68, 364, 41 },
		{ 68, 365, 41 },
		{ 68, 366, 41 },
		{ 68, 367, 41 },
		{ 68, 368, 41 },
		{ 68, 369, 41 },
		{ 68, 370, 41 },
		{ 68, 371, 41 },
		{ 68, 372, 41 },
		{ 68, 373, 41 },
		{ 68, 374, 41 },
		{ 68, 375, 41 },
		{ 68, 376, 41 },
		{ 68, 377, 41 },
		{ 68, 378, 41 },
		{ 68, 379, 41 },
		{ 68, 380, 41 },
		{ 68, 381, 41 },
		{ 68, 382, 41 },
		{ 68, 383, 41 },
		{ 68, 384, 41 },
		{ 68, 385, 41 },
		{ 68, 386, 41 },
		{ 68, 387, 41 },
		{ 68, 388, 41 },
		{ 68, 389, 41 },
		{ 68, 390, 41 },
		{ 68, 391, 41 },
		{ 68, 392, 41 },
		{ 68, 393, 41 },
		{ 68, 394, 41 },
		{ 68, 395, 41 },
		{ 68, 396, 41 },
		{ 68, 397, 41 },
		{ 68, 398, 41 },
		{ 69, 352, 41 },
		{ 69, 353, 41 },
		{ 69, 354, 41 },
		{ 69, 355, 41 },
		{ 69, 356, 41 },
		{ 69, 357, 41 },
		{ 69, 358, 41 },
		{ 69, 359, 41 },
		{ 69, 360, 41 },
		{ 69, 361, 41 },
		{ 69, 362, 41 },
		{ 69, 363, 41 },
		{ 69, 364, 41 },
		{ 69, 365, 41 },
		{ 69, 366, 41 },
		{ 69, 367, 41 },
		{ 69, 368, 41 },
		{ 69, 369, 41 },
		{ 69, 370, 41 },
		{ 69, 371, 41 },
		{ 69, 372, 41 },
		{ 69, 373, 41 },
		{ 69, 374, 41 },
		{ 69, 375, 41 },
		{ 69, 376, 41 },
		{ 69, 377, 41 },
		{ 69, 378, 41 },
		{ 69, 379, 41 },
		{ 69, 380, 41 },
		{ 69, 381, 41 },
		{ 69, 382, 41 },
		{ 69, 383, 41 },
		{ 69, 384, 41 },
		{ 69, 385, 41 },
		{ 69, 386, 41 },
		{ 69, 387, 41 },
		{ 69, 388, 41 },
		{ 69, 389, 41 },
		{ 69, 390, 41 },
		{ 69, 391, 41 },
		{ 69, 392, 41 },
		{ 69, 393, 41 },
		{ 69, 394, 41 },
		{ 69, 395, 41 },
		{ 69, 396, 41 },
		{ 69, 397, 41 },
		{ 70, 351, 41 },
		{ 70, 352, 41 },
		{ 70, 353, 41 },
		{ 70, 354, 41 },
		{ 70, 355, 41 },
		{ 70, 356, 41 },
		{ 70, 357, 41 },
		{ 70, 358, 41 },
		{ 70, 359, 41 },
		{ 70, 360, 41 },
		{ 70, 361, 41 },
		{ 70, 362, 41 },
		{ 70, 363, 41 },
		{ 70, 364, 41 },
		{ 70, 365, 41 },
		{ 70, 366, 41 },
		{ 70, 367, 41 },
		{ 70, 368, 41 },
		{ 70, 369, 41 },
		{ 70, 370, 41 },
		{ 70, 371, 41 },
		{ 70, 372, 41 },
		{ 70, 373, 41 },
		{ 70, 374, 41 },
		{ 70, 375, 41 },
		{ 70, 376, 41 },
		{ 70, 377, 41 },
		{ 70, 378, 41 },
		{ 70, 379, 41 },
		{ 70, 380, 41 },
		{ 70, 381, 41 },
		{ 70, 382, 41 },
		{ 70, 383, 41 },
		{ 70, 384, 41 },
		{ 70, 385, 41 },
		{ 70, 386, 41 },
		{ 70, 387, 41 },
		{ 70, 388, 41 },
		{ 70, 389, 41 },
		{ 70, 390, 41 },
		{ 70, 391, 41 },
		{ 70, 392, 41 },
		{ 70, 393, 41 },
		{ 70, 394, 41 },
		{ 70, 395, 41 },
		{ 70, 396, 41 },
		{ 70, 397, 41 },
		{ 71, 351, 41 },
		{ 71, 352, 41 },
		{ 71, 353, 41 },
		{ 71, 354, 41 },
		{ 71, 355, 41 },
		{ 71, 356, 41 },
		{ 71, 357, 41 },
		{ 71, 358, 41 },
		{ 71, 359, 41 },
		{ 71, 360, 41 },
		{ 71, 361, 41 },
		{ 71, 362, 41 },
		{ 71, 363, 41 },
		{ 71, 364, 41 },
		{ 71, 365, 41 },
		{ 71, 366, 41 },
		{ 71, 367, 41 },
		{ 71, 368, 41 },
		{ 71, 369, 41 },
		{ 71, 370, 41 },
		{ 71, 371, 41 },
		{ 71, 372, 41 },
		{ 71, 373, 41 },
		{ 71, 374, 41 },
		{ 71, 375, 41 },
		{ 71, 376, 41 },
		{ 71, 377, 41 },
		{ 71, 378, 41 },
		{ 71, 379, 41 },
		{ 71, 380, 41 },
		{ 71, 381, 41 },
		{ 71, 382, 41 },
		{ 71, 383, 41 },
		{ 71, 384, 41 },
		{ 71, 385, 41 },
		{ 71, 386, 41 },
		{ 71, 387, 41 },
		{ 71, 388, 41 },
		{ 71, 389, 41 },
		{ 71, 390, 41 },
		{ 71, 391, 41 },
		{ 71, 392, 41 },
		{ 71, 393, 41 },
		{ 71, 394, 41 },
		{ 71, 395, 41 },
		{ 71, 396, 41 },
		{ 72, 350, 41 },
		{ 72, 351, 41 },
		{ 72, 352, 41 },
		{ 72, 353, 41 },
		{ 72, 354, 41 },
		{ 72, 355, 41 },
		{ 72, 356, 41 },
		{ 72, 357, 41 },
		{ 72, 358, 41 },
		{ 72, 359, 41 },
		{ 72, 360, 41 },
		{ 72, 361, 41 },
		{ 72, 362, 41 },
		{ 72, 363, 41 },
		{ 72, 364, 41 },
		{ 72, 365, 41 },
		{ 72, 366, 41 },
		{ 72, 367, 41 },
		{ 72, 368, 41 },
		{ 72, 369, 41 },
		{ 72, 370, 41 },
		{ 72, 371, 41 },
		{ 72, 372, 41 },
		{ 72, 373, 41 },
		{ 72, 374, 41 },
		{ 72, 375, 41 },
		{ 72, 376, 41 },
		{ 72, 377, 41 },
		{ 72, 378, 41 },
		{ 72, 379, 41 },
		{ 72, 380, 41 },
		{ 72, 381, 41 },
		{ 72, 382, 41 },
		{ 72, 383, 41 },
		{ 72, 384, 41 },
		{ 72, 385, 41 },
		{ 72, 386, 41 },
		{ 72, 387, 41 },
		{ 72, 388, 41 },
		{ 72, 389, 41 },
		{ 72, 390, 41 },
		{ 72, 391, 41 },
		{ 72, 392, 41 },
		{ 72, 393, 41 },
		{ 72, 394, 41 },
		{ 72, 395, 41 },
		{ 73, 350, 41 },
		{ 73, 351, 41 },
		{ 73, 352, 41 },
		{ 73, 353, 41 },
		{ 73, 354, 41 },
		{ 73, 355, 41 },
		{ 73, 356, 41 },
		{ 73, 357, 41 },
		{ 73, 358, 41 },
		{ 73, 359, 41 },
		{ 73, 360, 41 },
		{ 73, 361, 41 },
		{ 73, 362, 41 },
		{ 73, 363, 41 },
		{ 73, 364, 41 },
		{ 73, 365, 41 },
		{ 73, 366, 41 },
		{ 73, 367, 41 },
		{ 73, 368, 41 },
		{ 73, 369, 41 },
		{ 73, 370, 41 },
		{ 73, 371, 41 },
		{ 73, 372, 41 },
		{ 73, 373, 41 },
		{ 73, 374, 41 },
		{ 73, 375, 41 },
		{ 73, 376, 41 },
		{ 73, 377, 41 },
		{ 73, 378, 41 },
		{ 73, 379, 41 },
		{ 73, 380, 41 },
		{ 73, 381, 41 },
		{ 73, 382, 41 },
		{ 73, 383, 41 },
		{ 73, 384, 41 },
		{ 73, 385, 41 },
		{ 73, 386, 41 },
		{ 73, 387, 41 },
		{ 73, 388, 41 },
		{ 73, 389, 41 },
		{ 73, 390, 41 },
		{ 73, 391, 41 },
		{ 73, 392, 41 },
		{ 73, 393, 41 },
		{ 73, 394, 41 },
		{ 74, 349, 41 },
		{ 74, 350, 41 },
		{ 74, 351, 41 },
		{ 74, 352, 41 },
		{ 74, 353, 41 },
		{ 74, 354, 41 },
		{ 74, 355, 41 },
		{ 74, 356, 41 },
		{ 74, 357, 41 },
		{ 74, 358, 41 },
		{ 74, 359, 41 },
		{ 74, 360, 41 },
		{ 74, 361, 41 },
		{ 74, 362, 41 },
		{ 74, 363, 41 },
		{ 74, 364, 41 },
		{ 74, 365, 41 },
		{ 74, 366, 41 },
		{ 74, 367, 41 },
		{ 74, 368, 41 },
		{ 74, 369, 41 },
		{ 74, 370, 41 },
		{ 74, 371, 41 },
		{ 74, 372, 41 },
		{ 74, 373, 41 },
		{ 74, 374, 41 },
		{ 74, 375, 41 },
		{ 74, 376, 41 },
		{ 74, 377, 41 },
		{ 74, 378, 41 },
		{ 74, 379, 41 },
		{ 74, 380, 41 },
		{ 74, 381, 41 },
		{ 74, 382, 41 },
		{ 74, 383, 41 },
		{ 74, 384, 41 },
		{ 74, 385, 41 },
		{ 74, 386, 41 },
		{ 74, 387, 41 },
		{ 74, 388, 41 },
		{ 74, 389, 41 },
		{ 74, 390, 41 },
		{ 74, 391, 41 },
		{ 74, 392, 41 },
		{ 74, 393, 41 },
		{ 74, 394, 41 },
		{ 75, 349, 41 },
		{ 75, 350, 41 },
		{ 75, 351, 41 },
		{ 75, 352, 41 },
		{ 75, 353, 41 },
		{ 75, 354, 41 },
		{ 75, 355, 41 },
		{ 75, 356, 41 },
		{ 75, 357, 41 },
		{ 75, 358, 41 },
		{ 75, 359, 41 },
		{ 75, 360, 41 },
		{ 75, 361, 41 },
		{ 75, 362, 41 },
		{ 75, 363, 41 },
		{ 75, 364, 41 },
		{ 75, 365, 41 },
		{ 75, 366, 41 },
		{ 75, 367, 41 },
		{ 75, 368, 41 },
		{ 75, 369, 41 },
		{ 75, 370, 41 },
		{ 75, 371, 41 },
		{ 75, 372, 41 },
		{ 75, 373, 41 },
		{ 75, 374, 41 },
		{ 75, 375, 41 },
		{ 75, 376, 41 },
		{ 75, 377, 41 },
		{ 75, 378, 41 },
		{ 75, 379, 41 },
		{ 75, 380, 41 },
		{ 75, 381, 41 },
		{ 75, 382, 41 },
		{ 75, 383, 41 },
		{ 75, 384, 41 },
		{ 75, 385, 41 },
		{ 75, 386, 41 },
		{ 75, 387, 41 },
		{ 75, 388, 41 },
		{ 75, 389, 41 },
		{ 75, 390, 41 },
		{ 75, 391, 41 },
		{ 75, 392, 41 },
		{ 75, 393, 41 },
		{ 76, 348, 41 },
		{ 76, 349, 41 },
		{ 76, 350, 41 },
		{ 76, 351, 41 },
		{ 76, 352, 41 },
		{ 76, 353, 41 },
		{ 76, 354, 41 },
		{ 76, 355, 41 },
		{ 76, 356, 41 },
		{ 76, 357, 41 },
		{ 76, 358, 41 },
		{ 76, 359, 41 },
		{ 76, 360, 41 },
		{ 76, 361, 41 },
		{ 76, 362, 41 },
		{ 76, 363, 41 },
		{ 76, 364, 41 },
		{ 76, 365, 41 },
		{ 76, 366, 41 },
		{ 76, 367, 41 },
		{ 76, 368, 41 },
		{ 76, 369, 41 },
		{ 76, 370, 41 },
		{ 76, 371, 41 },
		{ 76, 372, 41 },
		{ 76, 373, 41 },
		{ 76, 374, 41 },
		{ 76, 375, 41 },
		{ 76, 376, 41 },
		{ 76, 377, 41 },
		{ 76, 378, 41 },
		{ 76, 379, 41 },
		{ 76, 380, 41 },
		{ 76, 381, 41 },
		{ 76, 382, 41 },
		{ 76, 383, 41 },
		{ 76, 384, 41 },
		{ 76, 385, 41 },
		{ 76, 386, 41 },
		{ 76, 387, 41 },
		{ 76, 388, 41 },
		{ 76, 389, 41 },
		{ 76, 390, 41 },
		{ 76, 391, 41 },
		{ 76, 392, 41 },
		{ 77, 348, 41 },
		{ 77, 349, 41 },
		{ 77, 350, 41 },
		{ 77, 351, 41 },
		{ 77, 352, 41 },
		{ 77, 353, 41 },
		{ 77, 354, 41 },
		{ 77, 355, 41 },
		{ 77, 356, 41 },
		{ 77, 357, 41 },
		{ 77, 358, 41 },
		{ 77, 359, 41 },
		{ 77, 360, 41 },
		{ 77, 361, 41 },
		{ 77, 362, 41 },
		{ 77, 363, 41 },
		{ 77, 364, 41 },
		{ 77, 365, 41 },
		{ 77, 366, 41 },
		{ 77, 367, 41 },
		{ 77, 368, 41 },
		{ 77, 369, 41 },
		{ 77, 370, 41 },
		{ 77, 371, 41 },
		{ 77, 372, 41 },
		{ 77, 373, 41 },
		{ 77, 374, 41 },
		{ 77, 375, 41 },
		{ 77, 376, 41 },
		{ 77, 377, 41 },
		{ 77, 378, 41 },
		{ 77, 379, 41 },
		{ 77, 380, 41 },
		{ 77, 381, 41 },
		{ 77, 382, 41 },
		{ 77, 383, 41 },
		{ 77, 384, 41 },
		{ 77, 385, 41 },
		{ 77, 386, 41 },
		{ 77, 387, 41 },
		{ 77, 388, 41 },
		{ 77, 389, 41 },
		{ 77, 390, 41 },
		{ 77, 391, 41 },
		{ 78, 347, 41 },
		{ 78, 348, 41 },
		{ 78, 349, 41 },
		{ 78, 350, 41 },
		{ 78, 351, 41 },
		{ 78, 352, 41 },
		{ 78, 353, 41 },
		{ 78, 354, 41 },
		{ 78, 355, 41 },
		{ 78, 356, 41 },
		{ 78, 357, 41 },
		{ 78, 358, 41 },
		{ 78, 359, 41 },
		{ 78, 360, 41 },
		{ 78, 361, 41 },
		{ 78, 362, 41 },
		{ 78, 363, 41 },
		{ 78, 364, 41 },
		{ 78, 365, 41 },
		{ 78, 366, 41 },
		{ 78, 367, 41 },
		{ 78, 368, 41 },
		{ 78, 369, 41 },
		{ 78, 370, 41 },
		{ 78, 371, 41 },
		{ 78, 372, 41 },
		{ 78, 373, 41 },
		{ 78, 374, 41 },
		{ 78, 375, 41 },
		{ 78, 376, 41 },
		{ 78, 377, 41 },
		{ 78, 378, 41 },
		{ 78, 379, 41 },
		{ 78, 380, 41 },
		{ 78, 381, 41 },
		{ 78, 382, 41 },
		{ 78, 383, 41 },
		{ 78, 384, 41 },
		{ 78, 385, 41 },
		{ 78, 386, 41 },
		{ 78, 387, 41 },
		{ 78, 388, 41 },
		{ 78, 389, 41 },
		{ 78, 390, 41 },
		{ 78, 391, 41 },
		{ 79, 347, 41 },
		{ 79, 348, 41 },
		{ 79, 349, 41 },
		{ 79, 350, 41 },
		{ 79, 351, 41 },
		{ 79, 352, 41 },
		{ 79, 353, 41 },
		{ 79, 354, 41 },
		{ 79, 355, 41 },
		{ 79, 356, 41 },
		{ 79, 357, 41 },
		{ 79, 358, 41 },
		{ 79, 359, 41 },
		{ 79, 360, 41 },
		{ 79, 361, 41 },
		{ 79, 362, 41 },
		{ 79, 363, 41 },
		{ 79, 364, 41 },
		{ 79, 365, 41 },
		{ 79, 366, 41 },
		{ 79, 367, 41 },
		{ 79, 368, 41 },
		{ 79, 369, 41 },
		{ 79, 370, 41 },
		{ 79, 371, 41 },
		{ 79, 372, 41 },
		{ 79, 373, 41 },
		{ 79, 374, 41 },
		{ 79, 375, 41 },
		{ 79, 376, 41 },
		{ 79, 377, 41 },
		{ 79, 378, 41 },
		{ 79, 379, 41 },
		{ 79, 380, 41 },
		{ 79, 381, 41 },
		{ 79, 382, 41 },
		{ 79, 383, 41 },
		{ 79, 384, 41 },
		{ 79, 385, 41 },
		{ 79, 386, 41 },
		{ 79, 387, 41 },
		{ 79, 388, 41 },
		{ 79, 389, 41 },
		{ 79, 390, 41 },
		{ 80, 346, 41 },
		{ 80, 347, 41 },
		{ 80, 348, 41 },
		{ 80, 349, 41 },
		{ 80, 350, 41 },
		{ 80, 351, 41 },
		{ 80, 352, 41 },
		{ 80, 353, 41 },
		{ 80, 354, 41 },
		{ 80, 355, 41 },
		{ 80, 356, 41 },
		{ 80, 357, 41 },
		{ 80, 358, 41 },
		{ 80, 359, 41 },
		{ 80, 360, 41 },
		{ 80, 361, 41 },
		{ 80, 362, 41 },
		{ 80, 363, 41 },
		{ 80, 364, 41 },
		{ 80, 365, 41 },
		{ 80, 366, 41 },
		{ 80, 367, 41 },
		{ 80, 368, 41 },
		{ 80, 369, 41 },
		{ 80, 370, 41 },
		{ 80, 371, 41 },
		{ 80, 372, 41 },
		{ 80, 373, 41 },
		{ 80, 374, 41 },
		{ 80, 375, 41 },
		{ 80, 376, 41 },
		{ 80, 377, 41 },
		{ 80, 378, 41 },
		{ 80, 379, 41 },
		{ 80, 380, 41 },
		{ 80, 381, 41 },
		{ 80, 382, 41 },
		{ 80, 383, 41 },
		{ 80, 384, 41 },
		{ 80, 385, 41 },
		{ 80, 386, 41 },
		{ 80, 387, 41 },
		{ 80, 388, 41 },
		{ 80, 389, 41 },
		{ 81, 346, 41 },
		{ 81, 347, 41 },
		{ 81, 348, 41 },
		{ 81, 349, 41 },
		{ 81, 350, 41 },
		{ 81, 351, 41 },
		{ 81, 352, 41 },
		{ 81, 353, 41 },
		{ 81, 354, 41 },
		{ 81, 355, 41 },
		{ 81, 356, 41 },
		{ 81, 357, 41 },
		{ 81, 358, 41 },
		{ 81, 359, 41 },
		{ 81, 360, 41 },
		{ 81, 361, 41 },
		{ 81, 362, 41 },
		{ 81, 363, 41 },
		{ 81, 364, 41 },
		{ 81, 365, 41 },
		{ 81, 366, 41 },
		{ 81, 367, 41 },
		{ 81, 368, 41 },
		{ 81, 369, 41 },
		{ 81, 370, 41 },
		{ 81, 371, 41 },
		{ 81, 372, 41 },
		{ 81, 373, 41 },
		{ 81, 374, 41 },
		{ 81, 375, 41 },
		{ 81, 376, 41 },
		{ 81, 377, 41 },
		{ 81, 378, 41 },
		{ 81, 379, 41 },
		{ 81, 380, 41 },
		{ 81, 381, 41 },
		{ 81, 382, 41 },
		{ 81, 383, 41 },
		{ 81, 384, 41 },
		{ 81, 385, 41 },
		{ 81, 386, 41 },
		{ 81, 387, 41 },
		{ 81, 388, 41 },
		{ 82, 345, 41 },
		{ 82, 346, 41 },
		{ 82, 347, 41 },
		{ 82, 348, 41 },
		{ 82, 349, 41 },
		{ 82, 350, 41 },
		{ 82, 351, 41 },
		{ 82, 352, 41 },
		{ 82, 353, 41 },
		{ 82, 354, 41 },
		{ 82, 355, 41 },
		{ 82, 356, 41 },
		{ 82, 357, 41 },
		{ 82, 358, 41 },
		{ 82, 359, 41 },
		{ 82, 360, 41 },
		{ 82, 361, 41 },
		{ 82, 362, 41 },
		{ 82, 363, 41 },
		{ 82, 364, 41 },
		{ 82, 365, 41 },
		{ 82, 366, 41 },
		{ 82, 367, 41 },
		{ 82, 368, 41 },
		{ 82, 369, 41 },
		{ 82, 370, 41 },
		{ 82, 371, 41 },
		{ 82, 372, 41 },
		{ 82, 373, 41 },
		{ 82, 374, 41 },
		{ 82, 375, 41 },
		{ 82, 376, 41 },
		{ 82, 377, 41 },
		{ 82, 378, 41 },
		{ 82, 379, 41 },
		{ 82, 380, 41 },
		{ 82, 381, 41 },
		{ 82, 382, 41 },
		{ 82, 383, 41 },
		{ 82, 384, 41 },
		{ 82, 385, 41 },
		{ 82, 386, 41 },
		{ 82, 387, 41 },
		{ 82, 388, 41 },
		{ 83, 345, 41 },
		{ 83, 346, 41 },
		{ 83, 347, 41 },
		{ 83, 348, 41 },
		{ 83, 349, 41 },
		{ 83, 350, 41 },
		{ 83, 351, 41 },
		{ 83, 352, 41 },
		{ 83, 353, 41 },
		{ 83, 354, 41 },
		{ 83, 355, 41 },
		{ 83, 356, 41 },
		{ 83, 357, 41 },
		{ 83, 358, 41 },
		{ 83, 359, 41 },
		{ 83, 360, 41 },
		{ 83, 361, 41 },
		{ 83, 362, 41 },
		{ 83, 363, 41 },
		{ 83, 364, 41 },
		{ 83, 365, 41 },
		{ 83, 366, 41 },
		{ 83, 367, 41 },
		{ 83, 368, 41 },
		{ 83, 369, 41 },
		{ 83, 370, 41 },
		{ 83, 371, 41 },
		{ 83, 372, 41 },
		{ 83, 373, 41 },
		{ 83, 374, 41 },
		{ 83, 375, 41 },
		{ 83, 376, 41 },
		{ 83, 377, 41 },
		{ 83, 378, 41 },
		{ 83, 379, 41 },
		{ 83, 380, 41 },
		{ 83, 381, 41 },
		{ 83, 382, 41 },
		{ 83, 383, 41 },
		{ 83, 384, 41 },
		{ 83, 385, 41 },
		{ 83, 386, 41 },
		{ 83, 387, 41 },
		{ 84, 344, 41 },
		{ 84, 345, 41 },
		{ 84, 346, 41 },
		{ 84, 347, 41 },
		{ 84, 348, 41 },
		{ 84, 349, 41 },
		{ 84, 350, 41 },
		{ 84, 351, 41 },
		{ 84, 352, 41 },
		{ 84, 353, 41 },
		{ 84, 354, 41 },
		{ 84, 355, 41 },
		{ 84, 356, 41 },
		{ 84, 357, 41 },
		{ 84, 358, 41 },
		{ 84, 359, 41 },
		{ 84, 360, 41 },
		{ 84, 361, 41 },
		{ 84, 362, 41 },
		{ 84, 363, 41 },
		{ 84, 364, 41 },
		{ 84, 365, 41 },
		{ 84, 366, 41 },
		{ 84, 367, 41 },
		{ 84, 368, 41 },
		{ 84, 369, 41 },
		{ 84, 370, 41 },
		{ 84, 371, 41 },
		{ 84, 372, 41 },
		{ 84, 373, 41 },
		{ 84, 374, 41 },
		{ 84, 375, 41 },
		{ 84, 376, 41 },
		{ 84, 377, 41 },
		{ 84, 378, 41 },
		{ 84, 379, 41 },
		{ 84, 380, 41 },
		{ 84, 381, 41 },
		{ 84, 382, 41 },
		{ 84, 383, 41 },
		{ 84, 384, 41 },
		{ 84, 385, 41 },
		{ 84, 386, 41 },
		{ 85, 344, 41 },
		{ 85, 345, 41 },
		{ 85, 346, 41 },
		{ 85, 347, 41 },
		{ 85, 348, 41 },
		{ 85, 349, 41 },
		{ 85, 350, 41 },
		{ 85, 351, 41 },
		{ 85, 352, 41 },
		{ 85, 353, 41 },
		{ 85, 354, 41 },
		{ 85, 355, 41 },
		{ 85, 356, 41 },
		{ 85, 357, 41 },
		{ 85, 358, 41 },
		{ 85, 359, 41 },
		{ 85, 360, 41 },
		{ 85, 361, 41 },
		{ 85, 362, 41 },
		{ 85, 363, 41 },
		{ 85, 364, 41 },
		{ 85, 365, 41 },
		{ 85, 366, 41 },
		{ 85, 367, 41 },
		{ 85, 368, 41 },
		{ 85, 369, 41 },
		{ 85, 370, 41 },
		{ 85, 371, 41 },
		{ 85, 372, 41 },
		{ 85, 373, 41 },
		{ 85, 374, 41 },
		{ 85, 375, 41 },
		{ 85, 376, 41 },
		{ 85, 377, 41 },
		{ 85, 378, 41 },
		{ 85, 379, 41 },
		{ 85, 380, 41 },
		{ 85, 381, 41 },
		{ 85, 382, 41 },
		{ 85, 383, 41 },
		{ 85, 384, 41 },
		{ 85, 385, 41 },
		{ 86, 343, 41 },
		{ 86, 344, 41 },
		{ 86, 345, 41 },
		{ 86, 346, 41 },
		{ 86, 347, 41 },
		{ 86, 348, 41 },
		{ 86, 349, 41 },
		{ 86, 350, 41 },
		{ 86, 351, 41 },
		{ 86, 352, 41 },
		{ 86, 353, 41 },
		{ 86, 354, 41 },
		{ 86, 355, 41 },
		{ 86, 356, 41 },
		{ 86, 357, 41 },
		{ 86, 358, 41 },
		{ 86, 359, 41 },
		{ 86, 360, 41 },
		{ 86, 361, 41 },
		{ 86, 362, 41 },
		{ 86, 363, 41 },
		{ 86, 364, 41 },
		{ 86, 365, 41 },
		{ 86, 366, 41 },
		{ 86, 367, 41 },
		{ 86, 368, 41 },
		{ 86, 369, 41 },
		{ 86, 370, 41 },
		{ 86, 371, 41 },
		{ 86, 372, 41 },
		{ 86, 373, 41 },
		{ 86, 374, 41 },
		{ 86, 375, 41 },
		{ 86, 376, 41 },
		{ 86, 377, 41 },
		{ 86, 378, 41 },
		{ 86, 379, 41 },
		{ 86, 380, 41 },
		{ 86, 381, 41 },
		{ 86, 382, 41 },
		{ 86, 383, 41 },
		{ 86, 384, 41 },
		{ 86, 385, 41 },
		{ 87, 343, 41 },
		{ 87, 344, 41 },
		{ 87, 345, 41 },
		{ 87, 346, 41 },
		{ 87, 347, 41 },
		{ 87, 348, 41 },
		{ 87, 349, 41 },
		{ 87, 350, 41 },
		{ 87, 351, 41 },
		{ 87, 352, 41 },
		{ 87, 353, 41 },
		{ 87, 354, 41 },
		{ 87, 355, 41 },
		{ 87, 356, 41 },
		{ 87, 357, 41 },
		{ 87, 358, 41 },
		{ 87, 359, 41 },
		{ 87, 360, 41 },
		{ 87, 361, 41 },
		{ 87, 362, 41 },
		{ 87, 363, 41 },
		{ 87, 364, 41 },
		{ 87, 365, 41 },
		{ 87, 366, 41 },
		{ 87, 367, 41 },
		{ 87, 368, 41 },
		{ 87, 369, 41 },
		{ 87, 370, 41 },
		{ 87, 371, 41 },
		{ 87, 372, 41 },
		{ 87, 373, 41 },
		{ 87, 374, 41 },
		{ 87, 375, 41 },
		{ 87, 376, 41 },
		{ 87, 377, 41 },
		{ 87, 378, 41 },
		{ 87, 379, 41 },
		{ 87, 380, 41 },
		{ 87, 381, 41 },
		{ 87, 382, 41 },
		{ 87, 383, 41 },
		{ 87, 384, 41 },
		{ 88, 342, 41 },
		{ 88, 343, 41 },
		{ 88, 344, 41 },
		{ 88, 345, 41 },
		{ 88, 346, 41 },
		{ 88, 347, 41 },
		{ 88, 348, 41 },
		{ 88, 349, 41 },
		{ 88, 350, 41 },
		{ 88, 351, 41 },
		{ 88, 352, 41 },
		{ 88, 353, 41 },
		{ 88, 354, 41 },
		{ 88, 355, 41 },
		{ 88, 356, 41 },
		{ 88, 357, 41 },
		{ 88, 358, 41 },
		{ 88, 359, 41 },
		{ 88, 360, 41 },
		{ 88, 361, 41 },
		{ 88, 362, 41 },
		{ 88, 363, 41 },
		{ 88, 364, 41 },
		{ 88, 365, 41 },
		{ 88, 366, 41 },
		{ 88, 367, 41 },
		{ 88, 368, 41 },
		{ 88, 369, 41 },
		{ 88, 370, 41 },
		{ 88, 371, 41 },
		{ 88, 372, 41 },
		{ 88, 373, 41 },
		{ 88, 374, 41 },
		{ 88, 375, 41 },
		{ 88, 376, 41 },
		{ 88, 377, 41 },
		{ 88, 378, 41 },
		{ 88, 379, 41 },
		{ 88, 380, 41 },
		{ 88, 381, 41 },
		{ 88, 382, 41 },
		{ 88, 383, 41 },
		{ 89, 342, 41 },
		{ 89, 343, 41 },
		{ 89, 344, 41 },
		{ 89, 345, 41 },
		{ 89, 346, 41 },
		{ 89, 347, 41 },
		{ 89, 348, 41 },
		{ 89, 349, 41 },
		{ 89, 350, 41 },
		{ 89, 351, 41 },
		{ 89, 352, 41 },
		{ 89, 353, 41 },
		{ 89, 354, 41 },
		{ 89, 355, 41 },
		{ 89, 356, 41 },
		{ 89, 357, 41 },
		{ 89, 358, 41 },
		{ 89, 359, 41 },
		{ 89, 360, 41 },
		{ 89, 361, 41 },
		{ 89, 362, 41 },
		{ 89, 363, 41 },
		{ 89, 364, 41 },
		{ 89, 365, 41 },
		{ 89, 366, 41 },
		{ 89, 367, 41 },
		{ 89, 368, 41 },
		{ 89, 369, 41 },
		{ 89, 370, 41 },
		{ 89, 371, 41 },
		{ 89, 372, 41 },
		{ 89, 373, 41 },
		{ 89, 374, 41 },
		{ 89, 375, 41 },
		{ 89, 376, 41 },
		{ 89, 377, 41 },
		{ 89, 378, 41 },
		{ 89, 379, 41 },
		{ 89, 380, 41 },
		{ 89, 381, 41 },
		{ 89, 382, 41 },
		{ 90, 341, 41 },
		{ 90, 342, 41 },
		{ 90, 343, 41 },
		{ 90, 344, 41 },
		{ 90, 345, 41 },
		{ 90, 346, 41 },
		{ 90, 347, 41 },
		{ 90, 348, 41 },
		{ 90, 349, 41 },
		{ 90, 350, 41 },
		{ 90, 351, 41 },
		{ 90, 352, 41 },
		{ 90, 353, 41 },
		{ 90, 354, 41 },
		{ 90, 355, 41 },
		{ 90, 356, 41 },
		{ 90, 357, 41 },
		{ 90, 358, 41 },
		{ 90, 359, 41 },
		{ 90, 360, 41 },
		{ 90, 361, 41 },
		{ 90, 362, 41 },
		{ 90, 363, 41 },
		{ 90, 364, 41 },
		{ 90, 365, 41 },
		{ 90, 366, 41 },
		{ 90, 367, 41 },
		{ 90, 368, 41 },
		{ 90, 369, 41 },
		{ 90, 370, 41 },
		{ 90, 371, 41 },
		{ 90, 372, 41 },
		{ 90, 373, 41 },
		{ 90, 374, 41 },
		{ 90, 375, 41 },
		{ 90, 376, 41 },
		{ 90, 377, 41 },
		{ 90, 378, 41 },
		{ 90, 379, 41 },
		{ 90, 380, 41 },
		{ 90, 381, 41 },
		{ 90, 382, 41 },
		{ 91, 341, 41 },
		{ 91, 342, 41 },
		{ 91, 343, 41 },
		{ 91, 344, 41 },
		{ 91, 345, 41 },
		{ 91, 346, 41 },
		{ 91, 347, 41 },
		{ 91, 348, 41 },
		{ 91, 349, 41 },
		{ 91, 350, 41 },
		{ 91, 351, 41 },
		{ 91, 352, 41 },
		{ 91, 353, 41 },
		{ 91, 354, 41 },
		{ 91, 355, 41 },
		{ 91, 356, 41 },
		{ 91, 357, 41 },
		{ 91, 358, 41 },
		{ 91, 359, 41 },
		{ 91, 360, 41 },
		{ 91, 361, 41 },
		{ 91, 362, 41 },
		{ 91, 363, 41 },
		{ 91, 364, 41 },
		{ 91, 365, 41 },
		{ 91, 366, 41 },
		{ 91, 367, 41 },
		{ 91, 368, 41 },
		{ 91, 369, 41 },
		{ 91, 370, 41 },
		{ 91, 371, 41 },
		{ 91, 372, 41 },
		{ 91, 373, 41 },
		{ 91, 374, 41 },
		{ 91, 375, 41 },
		{ 91, 376, 41 },
		{ 91, 377, 41 },
		{ 91, 378, 41 },
		{ 91, 379, 41 },
		{ 91, 380, 41 },
		{ 91, 381, 41 },
		{ 92, 340, 41 },
		{ 92, 341, 41 },
		{ 92, 342, 41 },
		{ 92, 343, 41 },
		{ 92, 344, 41 },
		{ 92, 345, 41 },
		{ 92, 346, 41 },
		{ 92, 347, 41 },
		{ 92, 348, 41 },
		{ 92, 349, 41 },
		{ 92, 350, 41 },
		{ 92, 351, 41 },
		{ 92, 352, 41 },
		{ 92, 353, 41 },
		{ 92, 354, 41 },
		{ 92, 355, 41 },
		{ 92, 356, 41 },
		{ 92, 357, 41 },
		{ 92, 358, 41 },
		{ 92, 359, 41 },
		{ 92, 360, 41 },
		{ 92, 361, 41 },
		{ 92, 362, 41 },
		{ 92, 363, 41 },
		{ 92, 364, 41 },
		{ 92, 365, 41 },
		{ 92, 366, 41 },
		{ 92, 367, 41 },
		{ 92, 368, 41 },
		{ 92, 369, 41 },
		{ 92, 370, 41 },
		{ 92, 371, 41 },
		{ 92, 372, 41 },
		{ 92, 373, 41 },
		{ 92, 374, 41 },
		{ 92, 375, 41 },
		{ 92, 376, 41 },
		{ 92, 377, 41 },
		{ 92, 378, 41 },
		{ 92, 379, 41 },
		{ 92, 380, 41 },
		{ 93, 340, 41 },
		{ 93, 341, 41 },
		{ 93, 342, 41 },
		{ 93, 343, 41 },
		{ 93, 344, 41 },
		{ 93, 345, 41 },
		{ 93, 346, 41 },
		{ 93, 347, 41 },
		{ 93, 348, 41 },
		{ 93, 349, 41 },
		{ 93, 350, 41 },
		{ 93, 351, 41 },
		{ 93, 352, 41 },
		{ 93, 353, 41 },
		{ 93, 354, 41 },
		{ 93, 355, 41 },
		{ 93, 356, 41 },
		{ 93, 357, 41 },
		{ 93, 358, 41 },
		{ 93, 359, 41 },
		{ 93, 360, 41 },
		{ 93, 361, 41 },
		{ 93, 362, 41 },
		{ 93, 363, 41 },
		{ 93, 364, 41 },
		{ 93, 365, 41 },
		{ 93, 366, 41 },
		{ 93, 367, 41 },
		{ 93, 368, 41 },
		{ 93, 369, 41 },
		{ 93, 370, 41 },
		{ 93, 371, 41 },
		{ 93, 372, 41 },
		{ 93, 373, 41 },
		{ 93, 374, 41 },
		{ 93, 375, 41 },
		{ 93, 376, 41 },
		{ 93, 377, 41 },
		{ 93, 378, 41 },
		{ 93, 379, 41 },
		{ 94, 339, 41 },
		{ 94, 340, 41 },
		{ 94, 341, 41 },
		{ 94, 342, 41 },
		{ 94, 343, 41 },
		{ 94, 344, 41 },
		{ 94, 345, 41 },
		{ 94, 346, 41 },
		{ 94, 347, 41 },
		{ 94, 348, 41 },
		{ 94, 349, 41 },
		{ 94, 350, 41 },
		{ 94, 351, 41 },
		{ 94, 352, 41 },
		{ 94, 353, 41 },
		{ 94, 354, 41 },
		{ 94, 355, 41 },
		{ 94, 356, 41 },
		{ 94, 357, 41 },
		{ 94, 358, 41 },
		{ 94, 359, 41 },
		{ 94, 360, 41 },
		{ 94, 361, 41 },
		{ 94, 362, 41 },
		{ 94, 363, 41 },
		{ 94, 364, 41 },
		{ 94, 365, 41 },
		{ 94, 366, 41 },
		{ 94, 367, 41 },
		{ 94, 368, 41 },
		{ 94, 369, 41 },
		{ 94, 370, 41 },
		{ 94, 371, 41 },
		{ 94, 372, 41 },
		{ 94, 373, 41 },
		{ 94, 374, 41 },
		{ 94, 375, 41 },
		{ 94, 376, 41 },
		{ 94, 377, 41 },
		{ 94, 378, 41 },
		{ 94, 379, 41 },
		{ 95, 339, 41 },
		{ 95, 340, 41 },
		{ 95, 341, 41 },
		{ 95, 342, 41 },
		{ 95, 343, 41 },
		{ 95, 344, 41 },
		{ 95, 345, 41 },
		{ 95, 346, 41 },
		{ 95, 347, 41 },
		{ 95, 348, 41 },
		{ 95, 349, 41 },
		{ 95, 350, 41 },
		{ 95, 351, 41 },
		{ 95, 352, 41 },
		{ 95, 353, 41 },
		{ 95, 354, 41 },
		{ 95, 355, 41 },
		{ 95, 356, 41 },
		{ 95, 357, 41 },
		{ 95, 358, 41 },
		{ 95, 359, 41 },
		{ 95, 360, 41 },
		{ 95, 361, 41 },
		{ 95, 362, 41 },
		{ 95, 363, 41 },
		{ 95, 364, 41 },
		{ 95, 365, 41 },
		{ 95, 366, 41 },
		{ 95, 367, 41 },
		{ 95, 368, 41 },
		{ 95, 369, 41 },
		{ 95, 370, 41 },
		{ 95, 371, 41 },
		{ 95, 372, 41 },
		{ 95, 373, 41 },
		{ 95, 374, 41 },
		{ 95, 375, 41 },
		{ 95, 376, 41 },
		{ 95, 377, 41 },
		{ 95, 378, 41 },
		{ 96, 338, 41 },
		{ 96, 339, 41 },
		{ 96, 340, 41 },
		{ 96, 341, 41 },
		{ 96, 342, 41 },
		{ 96, 343, 41 },
		{ 96, 344, 41 },
		{ 96, 345, 41 },
		{ 96, 346, 41 },
		{ 96, 347, 41 },
		{ 96, 348, 41 },
		{ 96, 349, 41 },
		{ 96, 350, 41 },
		{ 96, 351, 41 },
		{ 96, 352, 41 },
		{ 96, 353, 41 },
		{ 96, 354, 41 },
		{ 96, 355, 41 },
		{ 96, 356, 41 },
		{ 96, 357, 41 },
		{ 96, 358, 41 },
		{ 96, 359, 41 },
		{ 96, 360, 41 },
		{ 96, 361, 41 },
		{ 96, 362, 41 },
		{ 96, 363, 41 },
		{ 96, 364, 41 },
		{ 96, 365, 41 },
		{ 96, 366, 41 },
		{ 96, 367, 41 },
		{ 96, 368, 41 },
		{ 96, 369, 41 },
		{ 96, 370, 41 },
		{ 96, 371, 41 },
		{ 96, 372, 41 },
		{ 96, 373, 41 },
		{ 96, 374, 41 },
		{ 96, 375, 41 },
		{ 96, 376, 41 },
		{ 96, 377, 41 },
		{ 97, 338, 41 },
		{ 97, 339, 41 },
		{ 97, 340, 41 },
		{ 97, 341, 41 },
		{ 97, 342, 41 },
		{ 97, 343, 41 },
		{ 97, 344, 41 },
		{ 97, 345, 41 },
		{ 97, 346, 41 },
		{ 97, 347, 41 },
		{ 97, 348, 41 },
		{ 97, 349, 41 },
		{ 97, 350, 41 },
		{ 97, 351, 41 },
		{ 97, 352, 41 },
		{ 97, 353, 41 },
		{ 97, 354, 41 },
		{ 97, 355, 41 },
		{ 97, 356, 41 },
		{ 97, 357, 41 },
		{ 97, 358, 41 },
		{ 97, 359, 41 },
		{ 97, 360, 41 },
		{ 97, 361, 41 },
		{ 97, 362, 41 },
		{ 97, 363, 41 },
		{ 97, 364, 41 },
		{ 97, 365, 41 },
		{ 97, 366, 41 },
		{ 97, 367, 41 },
		{ 97, 368, 41 },
		{ 97, 369, 41 },
		{ 97, 370, 41 },
		{ 97, 371, 41 },
		{ 97, 372, 41 },
		{ 97, 373, 41 },
		{ 97, 374, 41 },
		{ 97, 375, 41 },
		{ 97, 376, 41 },
		{ 98, 337, 41 },
		{ 98, 338, 41 },
		{ 98, 339, 41 },
		{ 98, 340, 41 },
		{ 98, 341, 41 },
		{ 98, 342, 41 },
		{ 98, 343, 41 },
		{ 98, 344, 41 },
		{ 98, 345, 41 },
		{ 98, 346, 41 },
		{ 98, 347, 41 },
		{ 98, 348, 41 },
		{ 98, 349, 41 },
		{ 98, 350, 41 },
		{ 98, 351, 41 },
		{ 98, 352, 41 },
		{ 98, 353, 41 },
		{ 98, 354, 41 },
		{ 98, 355, 41 },
		{ 98, 356, 41 },
		{ 98, 357, 41 },
		{ 98, 358, 41 },
		{ 98, 359, 41 },
		{ 98, 360, 41 },
		{ 98, 361, 41 },
		{ 98, 362, 41 },
		{ 98, 363, 41 },
		{ 98, 364, 41 },
		{ 98, 365, 41 },
		{ 98, 366, 41 },
		{ 98, 367, 41 },
		{ 98, 368, 41 },
		{ 98, 369, 41 },
		{ 98, 370, 41 },
		{ 98, 371, 41 },
		{ 98, 372, 41 },
		{ 98, 373, 41 },
		{ 98, 374, 41 },
		{ 98, 375, 41 },
		{ 98, 376, 41 },
		{ 99, 337, 41 },
		{ 99, 338, 41 },
		{ 99, 339, 41 },
		{ 99, 340, 41 },
		{ 99, 341, 41 },
		{ 99, 342, 41 },
		{ 99, 343, 41 },
		{ 99, 344, 41 },
		{ 99, 345, 41 },
		{ 99, 346, 41 },
		{ 99, 347, 41 },
		{ 99, 348, 41 },
		{ 99, 349, 41 },
		{ 99, 350, 41 },
		{ 99, 351, 41 },
		{ 99, 352, 41 },
		{ 99, 353, 41 },
		{ 99, 354, 41 },
		{ 99, 355, 41 },
		{ 99, 356, 41 },
		{ 99, 357, 41 },
		{ 99, 358, 41 },
		{ 99, 359, 41 },
		{ 99, 360, 41 },
		{ 99, 361, 41 },
		{ 99, 362, 41 },
		{ 99, 363, 41 },
		{ 99, 364, 41 },
		{ 99, 365, 41 },
		{ 99, 366, 41 },
		{ 99, 367, 41 },
		{ 99, 368, 41 },
		{ 99, 369, 41 },
		{ 99, 370, 41 },
		{ 99, 371, 41 },
		{ 99, 372, 41 },
		{ 99, 373, 41 },
		{ 99, 374, 41 },
		{ 99, 375, 41 },
		{ 100, 336, 41 },
		{ 100, 337, 41 },
		{ 100, 338, 41 },
		{ 100, 339, 41 },
		{ 100, 340, 41 },
		{ 100, 341, 41 },
		{ 100, 342, 41 },
		{ 100, 343, 41 },
		{ 100, 344, 41 },
		{ 100, 345, 41 },
		{ 100, 346, 41 },
		{ 100, 347, 41 },
		{ 100, 348, 41 },
		{ 100, 349, 41 },
		{ 100, 350, 41 },
		{ 100, 351, 41 },
		{ 100, 352, 41 },
		{ 100, 353, 41 },
		{ 100, 354, 41 },
		{ 100, 355, 41 },
		{ 100, 356, 41 },
		{ 100, 357, 41 },
		{ 100, 358, 41 },
		{ 100, 359, 41 },
		{ 100, 360, 41 },
		{ 100, 361, 41 },
		{ 100, 362, 41 },
		{ 100, 363, 41 },
		{ 100, 364, 41 },
		{ 100, 365, 41 },
		{ 100, 366, 41 },
		{ 100, 367, 41 },
		{ 100, 368, 41 },
		{ 100, 369, 41 },
		{ 100, 370, 41 },
		{ 100, 371, 41 },
		{ 100, 372, 41 },
		{ 100, 373, 41 },
		{ 100, 374, 41 },
		{ 101, 336, 41 },
		{ 101, 337, 41 },
		{ 101, 338, 41 },
		{ 101, 339, 41 },
		{ 101, 340, 41 },
		{ 101, 341, 41 },
		{ 101, 342, 41 },
		{ 101, 343, 41 },
		{ 101, 344, 41 },
		{ 101, 345, 41 },
		{ 101, 346, 41 },
		{ 101, 347, 41 },
		{ 101, 348, 41 },
		{ 101, 349, 41 },
		{ 101, 350, 41 },
		{ 101, 351, 41 },
		{ 101, 352, 41 },
		{ 101, 353, 41 },
		{ 101, 354, 41 },
		{ 101, 355, 41 },
		{ 101, 356, 41 },
		{ 101, 357, 41 },
		{ 101, 358, 41 },
		{ 101, 359, 41 },
		{ 101, 360, 41 },
		{ 101, 361, 41 },
		{ 101, 362, 41 },
		{ 101, 363, 41 },
		{ 101, 364, 41 },
		{ 101, 365, 41 },
		{ 101, 366, 41 },
		{ 101, 367, 41 },
		{ 101, 368, 41 },
		{ 101, 369, 41 },
		{ 101, 370, 41 },
		{ 101, 371, 41 },
		{ 101, 372, 41 },
		{ 101, 373, 41 },
		{ 102, 335, 41 },
		{ 102, 336, 41 },
		{ 102, 337, 41 },
		{ 102, 338, 41 },
		{ 102, 339, 41 },
		{ 102, 340, 41 },
		{ 102, 341, 41 },
		{ 102, 342, 41 },
		{ 102, 343, 41 },
		{ 102, 344, 41 },
		{ 102, 345, 41 },
		{ 102, 346, 41 },
		{ 102, 347, 41 },
		{ 102, 348, 41 },
		{ 102, 349, 41 },
		{ 102, 350, 41 },
		{ 102, 351, 41 },
		{ 102, 352, 41 },
		{ 102, 353, 41 },
		{ 102, 354, 41 },
		{ 102, 355, 41 },
		{ 102, 356, 41 },
		{ 102, 357, 41 },
		{ 102, 358, 41 },
		{ 102, 359, 41 },
		{ 102, 360, 41 },
		{ 102, 361, 41 },
		{ 102, 362, 41 },
		{ 102, 363, 41 },
		{ 102, 364, 41 },
		{ 102, 365, 41 },
		{ 102, 366, 41 },
		{ 102, 367, 41 },
		{ 102, 368, 41 },
		{ 102, 369, 41 },
		{ 102, 370, 41 },
		{ 102, 371, 41 },
		{ 102, 372, 41 },
		{ 102, 373, 41 },
		{ 103, 335, 41 },
		{ 103, 336, 41 },
		{ 103, 337, 41 },
		{ 103, 338, 41 },
		{ 103, 339, 41 },
		{ 103, 340, 41 },
		{ 103, 341, 41 },
		{ 103, 342, 41 },
		{ 103, 343, 41 },
		{ 103, 344, 41 },
		{ 103, 345, 41 },
		{ 103, 346, 41 },
		{ 103, 347, 41 },
		{ 103, 348, 41 },
		{ 103, 349, 41 },
		{ 103, 350, 41 },
		{ 103, 351, 41 },
		{ 103, 352, 41 },
		{ 103, 353, 41 },
		{ 103, 354, 41 },
		{ 103, 355, 41 },
		{ 103, 356, 41 },
		{ 103, 357, 41 },
		{ 103, 358, 41 },
		{ 103, 359, 41 },
		{ 103, 360, 41 },
		{ 103, 361, 41 },
		{ 103, 362, 41 },
		{ 103, 363, 41 },
		{ 103, 364, 41 },
		{ 103, 365, 41 },
		{ 103, 366, 41 },
		{ 103, 367, 41 },
		{ 103, 368, 41 },
		{ 103, 369, 41 },
		{ 103, 370, 41 },
		{ 103, 371, 41 },
		{ 103, 372, 41 },
		{ 104, 334, 41 },
		{ 104, 335, 41 },
		{ 104, 336, 41 },
		{ 104, 337, 41 },
		{ 104, 338, 41 },
		{ 104, 339, 41 },
		{ 104, 340, 41 },
		{ 104, 341, 41 },
		{ 104, 342, 41 },
		{ 104, 343, 41 },
		{ 104, 344, 41 },
		{ 104, 345, 41 },
		{ 104, 346, 41 },
		{ 104, 347, 41 },
		{ 104, 348, 41 },
		{ 104, 349, 41 },
		{ 104, 350, 41 },
		{ 104, 351, 41 },
		{ 104, 352, 41 },
		{ 104, 353, 41 },
		{ 104, 354, 41 },
		{ 104, 355, 41 },
		{ 104, 356, 41 },
		{ 104, 357, 41 },
		{ 104, 358, 41 },
		{ 104, 359, 41 },
		{ 104, 360, 41 },
		{ 104, 361, 41 },
		{ 104, 362, 41 },
		{ 104, 363, 41 },
		{ 104, 364, 41 },
		{ 104, 365, 41 },
		{ 104, 366, 41 },
		{ 104, 367, 41 },
		{ 104, 368, 41 },
		{ 104, 369, 41 },
		{ 104, 370, 41 },
		{ 104, 371, 41 },
		{ 105, 334, 41 },
		{ 105, 335, 41 },
		{ 105, 336, 41 },
		{ 105, 337, 41 },
		{ 105, 338, 41 },
		{ 105, 339, 41 },
		{ 105, 340, 41 },
		{ 105, 341, 41 },
		{ 105, 342, 41 },
		{ 105, 343, 41 },
		{ 105, 344, 41 },
		{ 105, 345, 41 },
		{ 105, 346, 41 },
		{ 105, 347, 41 },
		{ 105, 348, 41 },
		{ 105, 349, 41 },
		{ 105, 350, 41 },
		{ 105, 351, 41 },
		{ 105, 352, 41 },
		{ 105, 353, 41 },
		{ 105, 354, 41 },
		{ 105, 355, 41 },
		{ 105, 356, 41 },
		{ 105, 357, 41 },
		{ 105, 358, 41 },
		{ 105, 359, 41 },
		{ 105, 360, 41 },
		{ 105, 361, 41 },
		{ 105, 362, 41 },
		{ 105, 363, 41 },
		{ 105, 364, 41 },
		{ 105, 365, 41 },
		{ 105, 366, 41 },
		{ 105, 367, 41 },
		{ 105, 368, 41 },
		{ 105, 369, 41 },
		{ 105, 370, 41 },
		{ 106, 333, 41 },
		{ 106, 334, 41 },
		{ 106, 335, 41 },
		{ 106, 336, 41 },
		{ 106, 337, 41 },
		{ 106, 338, 41 },
		{ 106, 339, 41 },
		{ 106, 340, 41 },
		{ 106, 341, 41 },
		{ 106, 342, 41 },
		{ 106, 343, 41 },
		{ 106, 344, 41 },
		{ 106, 345, 41 },
		{ 106, 346, 41 },
		{ 106, 347, 41 },
		{ 106, 348, 41 },
		{ 106, 349, 41 },
		{ 106, 350, 41 },
		{ 106, 351, 41 },
		{ 106, 352, 41 },
		{ 106, 353, 41 },
		{ 106, 354, 41 },
		{ 106, 355, 41 },
		{ 106, 356, 41 },
		{ 106, 357, 41 },
		{ 106, 358, 41 },
		{ 106, 359, 41 },
		{ 106, 360, 41 },
		{ 106, 361, 41 },
		{ 106, 362, 41 },
		{ 106, 363, 41 },
		{ 106, 364, 41 },
		{ 106, 365, 41 },
		{ 106, 366, 41 },
		{ 106, 367, 41 },
		{ 106, 368, 41 },
		{ 106, 369, 41 },
		{ 106, 370, 41 },
		{ 107, 333, 41 },
		{ 107, 334, 41 },
		{ 107, 335, 41 },
		{ 107, 336, 41 },
		{ 107, 337, 41 },
		{ 107, 338, 41 },
		{ 107, 339, 41 },
		{ 107, 340, 41 },
		{ 107, 341, 41 },
		{ 107, 342, 41 },
		{ 107, 343, 41 },
		{ 107, 344, 41 },
		{ 107, 345, 41 },
		{ 107, 346, 41 },
		{ 107, 347, 41 },
		{ 107, 348, 41 },
		{ 107, 349, 41 },
		{ 107, 350, 41 },
		{ 107, 351, 41 },
		{ 107, 352, 41 },
		{ 107, 353, 41 },
		{ 107, 354, 41 },
		{ 107, 355, 41 },
		{ 107, 356, 41 },
		{ 107, 357, 41 },
		{ 107, 358, 41 },
		{ 107, 359, 41 },
		{ 107, 360, 41 },
		{ 107, 361, 41 },
		{ 107, 362, 41 },
		{ 107, 363, 41 },
		{ 107, 364, 41 },
		{ 107, 365, 41 },
		{ 107, 366, 41 },
		{ 107, 367, 41 },
		{ 107, 368, 41 },
		{ 107, 369, 41 },
		{ 108, 332, 41 },
		{ 108, 333, 41 },
		{ 108, 334, 41 },
		{ 108, 335, 41 },
		{ 108, 336, 41 },
		{ 108, 337, 41 },
		{ 108, 338, 41 },
		{ 108, 339, 41 },
		{ 108, 340, 41 },
		{ 108, 341, 41 },
		{ 108, 342, 41 },
		{ 108, 343, 41 },
		{ 108, 344, 41 },
		{ 108, 345, 41 },
		{ 108, 346, 41 },
		{ 108, 347, 41 },
		{ 108, 348, 41 },
		{ 108, 349, 41 },
		{ 108, 350, 41 },
		{ 108, 351, 41 },
		{ 108, 352, 41 },
		{ 108, 353, 41 },
		{ 108, 354, 41 },
		{ 108, 355, 41 },
		{ 108, 356, 41 },
		{ 108, 357, 41 },
		{ 108, 358, 41 },
		{ 108, 359, 41 },
		{ 108, 360, 41 },
		{ 108, 361, 41 },
		{ 108, 362, 41 },
		{ 108, 363, 41 },
		{ 108, 364, 41 },
		{ 108, 365, 41 },
		{ 108, 366, 41 },
		{ 108, 367, 41 },
		{ 108, 368, 41 },
		{ 109, 332, 41 },
		{ 109, 333, 41 },
		{ 109, 334, 41 },
		{ 109, 335, 41 },
		{ 109, 336, 41 },
		{ 109, 337, 41 },
		{ 109, 338, 41 },
		{ 109, 339, 41 },
		{ 109, 340, 41 },
		{ 109, 341, 41 },
		{ 109, 342, 41 },
		{ 109, 343, 41 },
		{ 109, 344, 41 },
		{ 109, 345, 41 },
		{ 109, 346, 41 },
		{ 109, 347, 41 },
		{ 109, 348, 41 },
		{ 109, 349, 41 },
		{ 109, 350, 41 },
		{ 109, 351, 41 },
		{ 109, 352, 41 },
		{ 109, 353, 41 },
		{ 109, 354, 41 },
		{ 109, 355, 41 },
		{ 109, 356, 41 },
		{ 109, 357, 41 },
		{ 109, 358, 41 },
		{ 109, 359, 41 },
		{ 109, 360, 41 },
		{ 109, 361, 41 },
		{ 109, 362, 41 },
		{ 109, 363, 41 },
		{ 109, 364, 41 },
		{ 109, 365, 41 },
		{ 109, 366, 41 },
		{ 109, 367, 41 },
		{ 110, 331, 41 },
		{ 110, 332, 41 },
		{ 110, 333, 41 },
		{ 110, 334, 41 },
		{ 110, 335, 41 },
		{ 110, 336, 41 },
		{ 110, 337, 41 },
		{ 110, 338, 41 },
		{ 110, 339, 41 },
		{ 110, 340, 41 },
		{ 110, 341, 41 },
		{ 110, 342, 41 },
		{ 110, 343, 41 },
		{ 110, 344, 41 },
		{ 110, 345, 41 },
		{ 110, 346, 41 },
		{ 110, 347, 41 },
		{ 110, 348, 41 },
		{ 110, 349, 41 },
		{ 110, 350, 41 },
		{ 110, 351, 41 },
		{ 110, 352, 41 },
		{ 110, 353, 41 },
		{ 110, 354, 41 },
		{ 110, 355, 41 },
		{ 110, 356, 41 },
		{ 110, 357, 41 },
		{ 110, 358, 41 },
		{ 110, 359, 41 },
		{ 110, 360, 41 },
		{ 110, 361, 41 },
		{ 110, 362, 41 },
		{ 110, 363, 41 },
		{ 110, 364, 41 },
		{ 110, 365, 41 },
		{ 110, 366, 41 },
		{ 110, 367, 41 },
		{ 111, 331, 41 },
		{ 111, 332, 41 },
		{ 111, 333, 41 },
		{ 111, 334, 41 },
		{ 111, 335, 41 },
		{ 111, 336, 41 },
		{ 111, 337, 41 },
		{ 111, 338, 41 },
		{ 111, 339, 41 },
		{ 111, 340, 41 },
		{ 111, 341, 41 },
		{ 111, 342, 41 },
		{ 111, 343, 41 },
		{ 111, 344, 41 },
		{ 111, 345, 41 },
		{ 111, 346, 41 },
		{ 111, 347, 41 },
		{ 111, 348, 41 },
		{ 111, 349, 41 },
		{ 111, 350, 41 },
		{ 111, 351, 41 },
		{ 111, 352, 41 },
		{ 111, 353, 41 },
		{ 111, 354, 41 },
		{ 111, 355, 41 },
		{ 111, 356, 41 },
		{ 111, 357, 41 },
		{ 111, 358, 41 },
		{ 111, 359, 41 },
		{ 111, 360, 41 },
		{ 111, 361, 41 },
		{ 111, 362, 41 },
		{ 111, 363, 41 },
		{ 111, 364, 41 },
		{ 111, 365, 41 },
		{ 111, 366, 41 },
		{ 112, 330, 41 },
		{ 112, 331, 41 },
		{ 112, 332, 41 },
		{ 112, 333, 41 },
		{ 112, 334, 41 },
		{ 112, 335, 41 },
		{ 112, 336, 41 },
		{ 112, 337, 41 },
		{ 112, 338, 41 },
		{ 112, 339, 41 },
		{ 112, 340, 41 },
		{ 112, 341, 41 },
		{ 112, 342, 41 },
		{ 112, 343, 41 },
		{ 112, 344, 41 },
		{ 112, 345, 41 },
		{ 112, 346, 41 },
		{ 112, 347, 41 },
		{ 112, 348, 41 },
		{ 112, 349, 41 },
		{ 112, 350, 41 },
		{ 112, 351, 41 },
		{ 112, 352, 41 },
		{ 112, 353, 41 },
		{ 112, 354, 41 },
		{ 112, 355, 41 },
		{ 112, 356, 41 },
		{ 112, 357, 41 },
		{ 112, 358, 41 },
		{ 112, 359, 41 },
		{ 112, 360, 41 },
		{ 112, 361, 41 },
		{ 112, 362, 41 },
		{ 112, 363, 41 },
		{ 112, 364, 41 },
		{ 112, 365, 41 },
		{ 113, 330, 41 },
		{ 113, 331, 41 },
		{ 113, 332, 41 },
		{ 113, 333, 41 },
		{ 113, 334, 41 },
		{ 113, 335, 41 },
		{ 113, 336, 41 },
		{ 113, 337, 41 },
		{ 113, 338, 41 },
		{ 113, 339, 41 },
		{ 113, 340, 41 },
		{ 113, 341, 41 },
		{ 113, 342, 41 },
		{ 113, 343, 41 },
		{ 113, 344, 41 },
		{ 113, 345, 41 },
		{ 113, 346, 41 },
		{ 113, 347, 41 },
		{ 113, 348, 41 },
		{ 113, 349, 41 },
		{ 113, 350, 41 },
		{ 113, 351, 41 },
		{ 113, 352, 41 },
		{ 113, 353, 41 },
		{ 113, 354, 41 },
		{ 113, 355, 41 },
		{ 113, 356, 41 },
		{ 113, 357, 41 },
		{ 113, 358, 41 },
		{ 113, 359, 41 },
		{ 113, 360, 41 },
		{ 113, 361, 41 },
		{ 113, 362, 41 },
		{ 113, 363, 41 },
		{ 113, 364, 41 },
		{ 114, 329, 41 },
		{ 114, 330, 41 },
		{ 114, 331, 41 },
		{ 114, 332, 41 },
		{ 114, 333, 41 },
		{ 114, 334, 41 },
		{ 114, 335, 41 },
		{ 114, 336, 41 },
		{ 114, 337, 41 },
		{ 114, 338, 41 },
		{ 114, 339, 41 },
		{ 114, 340, 41 },
		{ 114, 341, 41 },
		{ 114, 342, 41 },
		{ 114, 343, 41 },
		{ 114, 344, 41 },
		{ 114, 345, 41 },
		{ 114, 346, 41 },
		{ 114, 347, 41 },
		{ 114, 348, 41 },
		{ 114, 349, 41 },
		{ 114, 350, 41 },
		{ 114, 351, 41 },
		{ 114, 352, 41 },
		{ 114, 353, 41 },
		{ 114, 354, 41 },
		{ 114, 355, 41 },
		{ 114, 356, 41 },
		{ 114, 357, 41 },
		{ 114, 358, 41 },
		{ 114, 359, 41 },
		{ 114, 360, 41 },
		{ 114, 361, 41 },
		{ 114, 362, 41 },
		{ 114, 363, 41 },
		{ 114, 364, 41 },
		{ 115, 329, 41 },
		{ 115, 330, 41 },
		{ 115, 331, 41 },
		{ 115, 332, 41 },
		{ 115, 333, 41 },
		{ 115, 334, 41 },
		{ 115, 335, 41 },
		{ 115, 336, 41 },
		{ 115, 337, 41 },
		{ 115, 338, 41 },
		{ 115, 339, 41 },
		{ 115, 340, 41 },
		{ 115, 341, 41 },
		{ 115, 342, 41 },
		{ 115, 343, 41 },
		{ 115, 344, 41 },
		{ 115, 345, 41 },
		{ 115, 346, 41 },
		{ 115, 347, 41 },
		{ 115, 348, 41 },
		{ 115, 349, 41 },
		{ 115, 350, 41 },
		{ 115, 351, 41 },
		{ 115, 352, 41 },
		{ 115, 353, 41 },
		{ 115, 354, 41 },
		{ 115, 355, 41 },
		{ 115, 356, 41 },
		{ 115, 357, 41 },
		{ 115, 358, 41 },
		{ 115, 359, 41 },
		{ 115, 360, 41 },
		{ 115, 361, 41 },
		{ 115, 362, 41 },
		{ 115, 363, 41 },
		{ 116, 328, 41 },
		{ 116, 329, 41 },
		{ 116, 330, 41 },
		{ 116, 331, 41 },
		{ 116, 332, 41 },
		{ 116, 333, 41 },
		{ 116, 334, 41 },
		{ 116, 335, 41 },
		{ 116, 336, 41 },
		{ 116, 337, 41 },
		{ 116, 338, 41 },
		{ 116, 339, 41 },
		{ 116, 340, 41 },
		{ 116, 341, 41 },
		{ 116, 342, 41 },
		{ 116, 343, 41 },
		{ 116, 344, 41 },
		{ 116, 345, 41 },
		{ 116, 346, 41 },
		{ 116, 347, 41 },
		{ 116, 348, 41 },
		{ 116, 349, 41 },
		{ 116, 350, 41 },
		{ 116, 351, 41 },
		{ 116, 352, 41 },
		{ 116, 353, 41 },
		{ 116, 354, 41 },
		{ 116, 355, 41 },
		{ 116, 356, 41 },
		{ 116, 357, 41 },
		{ 116, 358, 41 },
		{ 116, 359, 41 },
		{ 116, 360, 41 },
		{ 116, 361, 41 },
		{ 116, 362, 41 },
		{ 117, 328, 41 },
		{ 117, 329, 41 },
		{ 117, 330, 41 },
		{ 117, 331, 41 },
		{ 117, 332, 41 },
		{ 117, 333, 41 },
		{ 117, 334, 41 },
		{ 117, 335, 41 },
		{ 117, 336, 41 },
		{ 117, 337, 41 },
		{ 117, 338, 41 },
		{ 117, 339, 41 },
		{ 117, 340, 41 },
		{ 117, 341, 41 },
		{ 117, 342, 41 },
		{ 117, 343, 41 },
		{ 117, 344, 41 },
		{ 117, 345, 41 },
		{ 117, 346, 41 },
		{ 117, 347, 41 },
		{ 117, 348, 41 },
		{ 117, 349, 41 },
		{ 117, 350, 41 },
		{ 117, 351, 41 },
		{ 117, 352, 41 },
		{ 117, 353, 41 },
		{ 117, 354, 41 },
		{ 117, 355, 41 },
		{ 117, 356, 41 },
		{ 117, 357, 41 },
		{ 117, 358, 41 },
		{ 117, 359, 41 },
		{ 117, 360, 41 },
		{ 117, 361, 41 },
		{ 118, 327, 41 },
		{ 118, 328, 41 },
		{ 118, 329, 41 },
		{ 118, 330, 41 },
		{ 118, 331, 41 },
		{ 118, 332, 41 },
		{ 118, 333, 41 },
		{ 118, 334, 41 },
		{ 118, 335, 41 },
		{ 118, 336, 41 },
		{ 118, 337, 41 },
		{ 118, 338, 41 },
		{ 118, 339, 41 },
		{ 118, 340, 41 },
		{ 118, 341, 41 },
		{ 118, 342, 41 },
		{ 118, 343, 41 },
		{ 118, 344, 41 },
		{ 118, 345, 41 },
		{ 118, 346, 41 },
		{ 118, 347, 41 },
		{ 118, 348, 41 },
		{ 118, 349, 41 },
		{ 118, 350, 41 },
		{ 118, 351, 41 },
		{ 118, 352, 41 },
		{ 118, 353, 41 },
		{ 118, 354, 41 },
		{ 118, 355, 41 },
		{ 118, 356, 41 },
		{ 118, 357, 41 },
		{ 118, 358, 41 },
		{ 118, 359, 41 },
		{ 118, 360, 41 },
		{ 118, 361, 41 },
		{ 119, 327, 41 },
		{ 119, 328, 41 },
		{ 119, 329, 41 },
		{ 119, 330, 41 },
		{ 119, 331, 41 },
		{ 119, 332, 41 },
		{ 119, 333, 41 },
		{ 119, 334, 41 },
		{ 119, 335, 41 },
		{ 119, 336, 41 },
		{ 119, 337, 41 },
		{ 119, 338, 41 },
		{ 119, 339, 41 },
		{ 119, 340, 41 },
		{ 119, 341, 41 },
		{ 119, 342, 41 },
		{ 119, 343, 41 },
		{ 119, 344, 41 },
		{ 119, 345, 41 },
		{ 119, 346, 41 },
		{ 119, 347, 41 },
		{ 119, 348, 41 },
		{ 119, 349, 41 },
		{ 119, 350, 41 },
		{ 119, 351, 41 },
		{ 119, 352, 41 },
		{ 119, 353, 41 },
		{ 119, 354, 41 },
		{ 119, 355, 41 },
		{ 119, 356, 41 },
		{ 119, 357, 41 },
		{ 119, 358, 41 },
		{ 119, 359, 41 },
		{ 119, 360, 41 },
		{ 120, 326, 41 },
		{ 120, 327, 41 },
		{ 120, 328, 41 },
		{ 120, 329, 41 },
		{ 120, 330, 41 },
		{ 120, 331, 41 },
		{ 120, 332, 41 },
		{ 120, 333, 41 },
		{ 120, 334, 41 },
		{ 120, 335, 41 },
		{ 120, 336, 41 },
		{ 120, 337, 41 },
		{ 120, 338, 41 },
		{ 120, 339, 41 },
		{ 120, 340, 41 },
		{ 120, 341, 41 },
		{ 120, 342, 41 },
		{ 120, 343, 41 },
		{ 120, 344, 41 },
		{ 120, 345, 41 },
		{ 120, 346, 41 },
		{ 120, 347, 41 },
		{ 120, 348, 41 },
		{ 120, 349, 41 },
		{ 120, 350, 41 },
		{ 120, 351, 41 },
		{ 120, 352, 41 },
		{ 120, 353, 41 },
		{ 120, 354, 41 },
		{ 120, 355, 41 },
		{ 120, 356, 41 },
		{ 120, 357, 41 },
		{ 120, 358, 41 },
		{ 120, 359, 41 },
		{ 121, 326, 41 },
		{ 121, 327, 41 },
		{ 121, 328, 41 },
		{ 121, 329, 41 },
		{ 121, 330, 41 },
		{ 121, 331, 41 },
		{ 121, 332, 41 },
		{ 121, 333, 41 },
		{ 121, 334, 41 },
		{ 121, 335, 41 },
		{ 121, 336, 41 },
		{ 121, 337, 41 },
		{ 121, 338, 41 },
		{ 121, 339, 41 },
		{ 121, 340, 41 },
		{ 121, 341, 41 },
		{ 121, 342, 41 },
		{ 121, 343, 41 },
		{ 121, 344, 41 },
		{ 121, 345, 41 },
		{ 121, 346, 41 },
		{ 121, 347, 41 },
		{ 121, 348, 41 },
		{ 121, 349, 41 },
		{ 121, 350, 41 },
		{ 121, 351, 41 },
		{ 121, 352, 41 },
		{ 121, 353, 41 },
		{ 121, 354, 41 },
		{ 121, 355, 41 },
		{ 121, 356, 41 },
		{ 121, 357, 41 },
		{ 121, 358, 41 },
		{ 122, 325, 41 },
		{ 122, 326, 41 },
		{ 122, 327, 41 },
		{ 122, 328, 41 },
		{ 122, 329, 41 },
		{ 122, 330, 41 },
		{ 122, 331, 41 },
		{ 122, 332, 41 },
		{ 122, 333, 41 },
		{ 122, 334, 41 },
		{ 122, 335, 41 },
		{ 122, 336, 41 },
		{ 122, 337, 41 },
		{ 122, 338, 41 },
		{ 122, 339, 41 },
		{ 122, 340, 41 },
		{ 122, 341, 41 },
		{ 122, 342, 41 },
		{ 122, 343, 41 },
		{ 122, 344, 41 },
		{ 122, 345, 41 },
		{ 122, 346, 41 },
		{ 122, 347, 41 },
		{ 122, 348, 41 },
		{ 122, 349, 41 },
		{ 122, 350, 41 },
		{ 122, 351, 41 },
		{ 122, 352, 41 },
		{ 122, 353, 41 },
		{ 122, 354, 41 },
		{ 122, 355, 41 },
		{ 122, 356, 41 },
		{ 122, 357, 41 },
		{ 122, 358, 41 },
		{ 123, 325, 41 },
		{ 123, 326, 41 },
		{ 123, 327, 41 },
		{ 123, 328, 41 },
		{ 123, 329, 41 },
		{ 123, 330, 41 },
		{ 123, 331, 41 },
		{ 123, 332, 41 },
		{ 123, 333, 41 },
		{ 123, 334, 41 },
		{ 123, 335, 41 },
		{ 123, 336, 41 },
		{ 123, 337, 41 },
		{ 123, 338, 41 },
		{ 123, 339, 41 },
		{ 123, 340, 41 },
		{ 123, 341, 41 },
		{ 123, 342, 41 },
		{ 123, 343, 41 },
		{ 123, 344, 41 },
		{ 123, 345, 41 },
		{ 123, 346, 41 },
		{ 123, 347, 41 },
		{ 123, 348, 41 },
		{ 123, 349, 41 },
		{ 123, 350, 41 },
		{ 123, 351, 41 },
		{ 123, 352, 41 },
		{ 123, 353, 41 },
		{ 123, 354, 41 },
		{ 123, 355, 41 },
		{ 123, 356, 41 },
		{ 123, 357, 41 },
		{ 124, 324, 41 },
		{ 124, 325, 41 },
		{ 124, 326, 41 },
		{ 124, 327, 41 },
		{ 124, 328, 41 },
		{ 124, 329, 41 },
		{ 124, 330, 41 },
		{ 124, 331, 41 },
		{ 124, 332, 41 },
		{ 124, 333, 41 },
		{ 124, 334, 41 },
		{ 124, 335, 41 },
		{ 124, 336, 41 },
		{ 124, 337, 41 },
		{ 124, 338, 41 },
		{ 124, 339, 41 },
		{ 124, 340, 41 },
		{ 124, 341, 41 },
		{ 124, 342, 41 },
		{ 124, 343, 41 },
		{ 124, 344, 41 },
		{ 124, 345, 41 },
		{ 124, 346, 41 },
		{ 124, 347, 41 },
		{ 124, 348, 41 },
		{ 124, 349, 41 },
		{ 124, 350, 41 },
		{ 124, 351, 41 },
		{ 124, 352, 41 },
		{ 124, 353, 41 },
		{ 124, 354, 41 },
		{ 124, 355, 41 },
		{ 124, 356, 41 },
		{ 125, 324, 41 },
		{ 125, 325, 41 },
		{ 125, 326, 41 },
		{ 125, 327, 41 },
		{ 125, 328, 41 },
		{ 125, 329, 41 },
		{ 125, 330, 41 },
		{ 125, 331, 41 },
		{ 125, 332, 41 },
		{ 125, 333, 41 },
		{ 125, 334, 41 },
		{ 125, 335, 41 },
		{ 125, 336, 41 },
		{ 125, 337, 41 },
		{ 125, 338, 41 },
		{ 125, 339, 41 },
		{ 125, 340, 41 },
		{ 125, 341, 41 },
		{ 125, 342, 41 },
		{ 125, 343, 41 },
		{ 125, 344, 41 },
		{ 125, 345, 41 },
		{ 125, 346, 41 },
		{ 125, 347, 41 },
		{ 125, 348, 41 },
		{ 125, 349, 41 },
		{ 125, 350, 41 },
		{ 125, 351, 41 },
		{ 125, 352, 41 },
		{ 125, 353, 41 },
		{ 125, 354, 41 },
		{ 125, 355, 41 },
		{ 126, 323, 41 },
		{ 126, 324, 41 },
		{ 126, 325, 41 },
		{ 126, 326, 41 },
		{ 126, 327, 41 },
		{ 126, 328, 41 },
		{ 126, 329, 41 },
		{ 126, 330, 41 },
		{ 126, 331, 41 },
		{ 126, 332, 41 },
		{ 126, 333, 41 },
		{ 126, 334, 41 },
		{ 126, 335, 41 },
		{ 126, 336, 41 },
		{ 126, 337, 41 },
		{ 126, 338, 41 },
		{ 126, 339, 41 },
		{ 126, 340, 41 },
		{ 126, 341, 41 },
		{ 126, 342, 41 },
		{ 126, 343, 41 },
		{ 126, 344, 41 },
		{ 126, 345, 41 },
		{ 126, 346, 41 },
		{ 126, 347, 41 },
		{ 126, 348, 41 },
		{ 126, 349, 41 },
		{ 126, 350, 41 },
		{ 126, 351, 41 },
		{ 126, 352, 41 },
		{ 126, 353, 41 },
		{ 126, 354, 41 },
		{ 126, 355, 41 },
		{ 127, 323, 41 },
		{ 127, 324, 41 },
		{ 127, 325, 41 },
		{ 127, 326, 41 },
		{ 127, 327, 41 },
		{ 127, 328, 41 },
		{ 127, 329, 41 },
		{ 127, 330, 41 },
		{ 127, 331, 41 },
		{ 127, 332, 41 },
		{ 127, 333, 41 },
		{ 127, 334, 41 },
		{ 127, 335, 41 },
		{ 127, 336, 41 },
		{ 127, 337, 41 },
		{ 127, 338, 41 },
		{ 127, 339, 41 },
		{ 127, 340, 41 },
		{ 127, 341, 41 },
		{ 127, 342, 41 },
		{ 127, 343, 41 },
		{ 127, 344, 41 },
		{ 127, 345, 41 },
		{ 127, 346, 41 },
		{ 127, 347, 41 },
		{ 127, 348, 41 },
		{ 127, 349, 41 },
		{ 127, 350, 41 },
		{ 127, 351, 41 },
		{ 127, 352, 41 },
		{ 127, 353, 41 },
		{ 127, 354, 41 },
		{ 128, 322, 41 },
		{ 128, 323, 41 },
		{ 128, 324, 41 },
		{ 128, 325, 41 },
		{ 128, 326, 41 },
		{ 128, 327, 41 },
		{ 128, 328, 41 },
		{ 128, 329, 41 },
		{ 128, 330, 41 },
		{ 128, 331, 41 },
		{ 128, 332, 41 },
		{ 128, 333, 41 },
		{ 128, 334, 41 },
		{ 128, 335, 41 },
		{ 128, 336, 41 },
		{ 128, 337, 41 },
		{ 128, 338, 41 },
		{ 128, 339, 41 },
		{ 128, 340, 41 },
		{ 128, 341, 41 },
		{ 128, 342, 41 },
		{ 128, 343, 41 },
		{ 128, 344, 41 },
		{ 128, 345, 41 },
		{ 128, 346, 41 },
		{ 128, 347, 41 },
		{ 128, 348, 41 },
		{ 128, 349, 41 },
		{ 128, 350, 41 },
		{ 128, 351, 41 },
		{ 128, 352, 41 },
		{ 128, 353, 41 },
		{ 129, 322, 41 },
		{ 129, 323, 41 },
		{ 129, 324, 41 },
		{ 129, 325, 41 },
		{ 129, 326, 41 },
		{ 129, 327, 41 },
		{ 129, 328, 41 },
		{ 129, 329, 41 },
		{ 129, 330, 41 },
		{ 129, 331, 41 },
		{ 129, 332, 41 },
		{ 129, 333, 41 },
		{ 129, 334, 41 },
		{ 129, 335, 41 },
		{ 129, 336, 41 },
		{ 129, 337, 41 },
		{ 129, 338, 41 },
		{ 129, 339, 41 },
		{ 129, 340, 41 },
		{ 129, 341, 41 },
		{ 129, 342, 41 },
		{ 129, 343, 41 },
		{ 129, 344, 41 },
		{ 129, 345, 41 },
		{ 129, 346, 41 },
		{ 129, 347, 41 },
		{ 129, 348, 41 },
		{ 129, 349, 41 },
		{ 129, 350, 41 },
		{ 129, 351, 41 },
		{ 129, 352, 41 },
		{ 130, 321, 41 },
		{ 130, 322, 41 },
		{ 130, 323, 41 },
		{ 130, 324, 41 },
		{ 130, 325, 41 },
		{ 130, 326, 41 },
		{ 130, 327, 41 },
		{ 130, 328, 41 },
		{ 130, 329, 41 },
		{ 130, 330, 41 },
		{ 130, 331, 41 },
		{ 130, 332, 41 },
		{ 130, 333, 41 },
		{ 130, 334, 41 },
		{ 130, 335, 41 },
		{ 130, 336, 41 },
		{ 130, 337, 41 },
		{ 130, 338, 41 },
		{ 130, 339, 41 },
		{ 130, 340, 41 },
		{ 130, 341, 41 },
		{ 130, 342, 41 },
		{ 130, 343, 41 },
		{ 130, 344, 41 },
		{ 130, 345, 41 },
		{ 130, 346, 41 },
		{ 130, 347, 41 },
		{ 130, 348, 41 },
		{ 130, 349, 41 },
		{ 130, 350, 41 },
		{ 130, 351, 41 },
		{ 130, 352, 41 },
		{ 131, 321, 41 },
		{ 131, 322, 41 },
		{ 131, 323, 41 },
		{ 131, 324, 41 },
		{ 131, 325, 41 },
		{ 131, 326, 41 },
		{ 131, 327, 41 },
		{ 131, 328, 41 },
		{ 131, 329, 41 },
		{ 131, 330, 41 },
		{ 131, 331, 41 },
		{ 131, 332, 41 },
		{ 131, 333, 41 },
		{ 131, 334, 41 },
		{ 131, 335, 41 },
		{ 131, 336, 41 },
		{ 131, 337, 41 },
		{ 131, 338, 41 },
		{ 131, 339, 41 },
		{ 131, 340, 41 },
		{ 131, 341, 41 },
		{ 131, 342, 41 },
		{ 131, 343, 41 },
		{ 131, 344, 41 },
		{ 131, 345, 41 },
		{ 131, 346, 41 },
		{ 131, 347, 41 },
		{ 131, 348, 41 },
		{ 131, 349, 41 },
		{ 131, 350, 41 },
		{ 131, 351, 41 },
		{ 132, 320, 41 },
		{ 132, 321, 41 },
		{ 132, 322, 41 },
		{ 132, 323, 41 },
		{ 132, 324, 41 },
		{ 132, 325, 41 },
		{ 132, 326, 41 },
		{ 132, 327, 41 },
		{ 132, 328, 41 },
		{ 132, 329, 41 },
		{ 132, 330, 41 },
		{ 132, 331, 41 },
		{ 132, 332, 41 },
		{ 132, 333, 41 },
		{ 132, 334, 41 },
		{ 132, 335, 41 },
		{ 132, 336, 41 },
		{ 132, 337, 41 },
		{ 132, 338, 41 },
		{ 132, 339, 41 },
		{ 132, 340, 41 },
		{ 132, 341, 41 },
		{ 132, 342, 41 },
		{ 132, 343, 41 },
		{ 132, 344, 41 },
		{ 132, 345, 41 },
		{ 132, 346, 41 },
		{ 132, 347, 41 },
		{ 132, 348, 41 },
		{ 132, 349, 41 },
		{ 132, 350, 41 },
		{ 133, 320, 41 },
		{ 133, 321, 41 },
		{ 133, 322, 41 },
		{ 133, 323, 41 },
		{ 133, 324, 41 },
		{ 133, 325, 41 },
		{ 133, 326, 41 },
		{ 133, 327, 41 },
		{ 133, 328, 41 },
		{ 133, 329, 41 },
		{ 133, 330, 41 },
		{ 133, 331, 41 },
		{ 133, 332, 41 },
		{ 133, 333, 41 },
		{ 133, 334, 41 },
		{ 133, 335, 41 },
		{ 133, 336, 41 },
		{ 133, 337, 41 },
		{ 133, 338, 41 },
		{ 133, 339, 41 },
		{ 133, 340, 41 },
		{ 133, 341, 41 },
		{ 133, 342, 41 },
		{ 133, 343, 41 },
		{ 133, 344, 41 },
		{ 133, 345, 41 },
		{ 133, 346, 41 },
		{ 133, 347, 41 },
		{ 133, 348, 41 },
		{ 133, 349, 41 },
		{ 134, 319, 41 },
		{ 134, 320, 41 },
		{ 134, 321, 41 },
		{ 134, 322, 41 },
		{ 134, 323, 41 },
		{ 134, 324, 41 },
		{ 134, 325, 41 },
		{ 134, 326, 41 },
		{ 134, 327, 41 },
		{ 134, 328, 41 },
		{ 134, 329, 41 },
		{ 134, 330, 41 },
		{ 134, 331, 41 },
		{ 134, 332, 41 },
		{ 134, 333, 41 },
		{ 134, 334, 41 },
		{ 134, 335, 41 },
		{ 134, 336, 41 },
		{ 134, 337, 41 },
		{ 134, 338, 41 },
		{ 134, 339, 41 },
		{ 134, 340, 41 },
		{ 134, 341, 41 },
		{ 134, 342, 41 },
		{ 134, 343, 41 },
		{ 134, 344, 41 },
		{ 134, 345, 41 },
		{ 134, 346, 41 },
		{ 134, 347, 41 },
		{ 134, 348, 41 },
		{ 134, 349, 41 },
		{ 135, 319, 41 },
		{ 135, 320, 41 },
		{ 135, 321, 41 },
		{ 135, 322, 41 },
		{ 135, 323, 41 },
		{ 135, 324, 41 },
		{ 135, 325, 41 },
		{ 135, 326, 41 },
		{ 135, 327, 41 },
		{ 135, 328, 41 },
		{ 135, 329, 41 },
		{ 135, 330, 41 },
		{ 135, 331, 41 },
		{ 135, 332, 41 },
		{ 135, 333, 41 },
		{ 135, 334, 41 },
		{ 135, 335, 41 },
		{ 135, 336, 41 },
		{ 135, 337, 41 },
		{ 135, 338, 41 },
		{ 135, 339, 41 },
		{ 135, 340, 41 },
		{ 135, 341, 41 },
		{ 135, 342, 41 },
		{ 135, 343, 41 },
		{ 135, 344, 41 },
		{ 135, 345, 41 },
		{ 135, 346, 41 },
		{ 135, 347, 41 },
		{ 135, 348, 41 },
		{ 136, 318, 41 },
		{ 136, 319, 41 },
		{ 136, 320, 41 },
		{ 136, 321, 41 },
		{ 136, 322, 41 },
		{ 136, 323, 41 },
		{ 136, 324, 41 },
		{ 136, 325, 41 },
		{ 136, 326, 41 },
		{ 136, 327, 41 },
		{ 136, 328, 41 },
		{ 136, 329, 41 },
		{ 136, 330, 41 },
		{ 136, 331, 41 },
		{ 136, 332, 41 },
		{ 136, 333, 41 },
		{ 136, 334, 41 },
		{ 136, 335, 41 },
		{ 136, 336, 41 },
		{ 136, 337, 41 },
		{ 136, 338, 41 },
		{ 136, 339, 41 },
		{ 136, 340, 41 },
		{ 136, 341, 41 },
		{ 136, 342, 41 },
		{ 136, 343, 41 },
		{ 136, 344, 41 },
		{ 136, 345, 41 },
		{ 136, 346, 41 },
		{ 136, 347, 41 },
		{ 137, 318, 41 },
		{ 137, 319, 41 },
		{ 137, 320, 41 },
		{ 137, 321, 41 },
		{ 137, 322, 41 },
		{ 137, 323, 41 },
		{ 137, 324, 41 },
		{ 137, 325, 41 },
		{ 137, 326, 41 },
		{ 137, 327, 41 },
		{ 137, 328, 41 },
		{ 137, 329, 41 },
		{ 137, 330, 41 },
		{ 137, 331, 41 },
		{ 137, 332, 41 },
		{ 137, 333, 41 },
		{ 137, 334, 41 },
		{ 137, 335, 41 },
		{ 137, 336, 41 },
		{ 137, 337, 41 },
		{ 137, 338, 41 },
		{ 137, 339, 41 },
		{ 137, 340, 41 },
		{ 137, 341, 41 },
		{ 137, 342, 41 },
		{ 137, 343, 41 },
		{ 137, 344, 41 },
		{ 137, 345, 41 },
		{ 137, 346, 41 },
		{ 138, 317, 41 },
		{ 138, 318, 41 },
		{ 138, 319, 41 },
		{ 138, 320, 41 },
		{ 138, 321, 41 },
		{ 138, 322, 41 },
		{ 138, 323, 41 },
		{ 138, 324, 41 },
		{ 138, 325, 41 },
		{ 138, 326, 41 },
		{ 138, 327, 41 },
		{ 138, 328, 41 },
		{ 138, 329, 41 },
		{ 138, 330, 41 },
		{ 138, 331, 41 },
		{ 138, 332, 41 },
		{ 138, 333, 41 },
		{ 138, 334, 41 },
		{ 138, 335, 41 },
		{ 138, 336, 41 },
		{ 138, 337, 41 },
		{ 138, 338, 41 },
		{ 138, 339, 41 },
		{ 138, 340, 41 },
		{ 138, 341, 41 },
		{ 138, 342, 41 },
		{ 138, 343, 41 },
		{ 138, 344, 41 },
		{ 138, 345, 41 },
		{ 138, 346, 41 },
		{ 139, 317, 41 },
		{ 139, 318, 41 },
		{ 139, 319, 41 },
		{ 139, 320, 41 },
		{ 139, 321, 41 },
		{ 139, 322, 41 },
		{ 139, 323, 41 },
		{ 139, 324, 41 },
		{ 139, 325, 41 },
		{ 139, 326, 41 },
		{ 139, 327, 41 },
		{ 139, 328, 41 },
		{ 139, 329, 41 },
		{ 139, 330, 41 },
		{ 139, 331, 41 },
		{ 139, 332, 41 },
		{ 139, 333, 41 },
		{ 139, 334, 41 },
		{ 139, 335, 41 },
		{ 139, 336, 41 },
		{ 139, 337, 41 },
		{ 139, 338, 41 },
		{ 139, 339, 41 },
		{ 139, 340, 41 },
		{ 139, 341, 41 },
		{ 139, 342, 41 },
		{ 139, 343, 41 },
		{ 139, 344, 41 },
		{ 139, 345, 41 },
		{ 140, 316, 41 },
		{ 140, 317, 41 },
		{ 140, 318, 41 },
		{ 140, 319, 41 },
		{ 140, 320, 41 },
		{ 140, 321, 41 },
		{ 140, 322, 41 },
		{ 140, 323, 41 },
		{ 140, 324, 41 },
		{ 140, 325, 41 },
		{ 140, 326, 41 },
		{ 140, 327, 41 },
		{ 140, 328, 41 },
		{ 140, 329, 41 },
		{ 140, 330, 41 },
		{ 140, 331, 41 },
		{ 140, 332, 41 },
		{ 140, 333, 41 },
		{ 140, 334, 41 },
		{ 140, 335, 41 },
		{ 140, 336, 41 },
		{ 140, 337, 41 },
		{ 140, 338, 41 },
		{ 140, 339, 41 },
		{ 140, 340, 41 },
		{ 140, 341, 41 },
		{ 140, 342, 41 },
		{ 140, 343, 41 },
		{ 140, 344, 41 },
		{ 141, 316, 41 },
		{ 141, 317, 41 },
		{ 141, 318, 41 },
		{ 141, 319, 41 },
		{ 141, 320, 41 },
		{ 141, 321, 41 },
		{ 141, 322, 41 },
		{ 141, 323, 41 },
		{ 141, 324, 41 },
		{ 141, 325, 41 },
		{ 141, 326, 41 },
		{ 141, 327, 41 },
		{ 141, 328, 41 },
		{ 141, 329, 41 },
		{ 141, 330, 41 },
		{ 141, 331, 41 },
		{ 141, 332, 41 },
		{ 141, 333, 41 },
		{ 141, 334, 41 },
		{ 141, 335, 41 },
		{ 141, 336, 41 },
		{ 141, 337, 41 },
		{ 141, 338, 41 },
		{ 141, 339, 41 },
		{ 141, 340, 41 },
		{ 141, 341, 41 },
		{ 141, 342, 41 },
		{ 141, 343, 41 },
		{ 142, 315, 41 },
		{ 142, 316, 41 },
		{ 142, 317, 41 },
		{ 142, 318, 41 },
		{ 142, 319, 41 },
		{ 142, 320, 41 },
		{ 142, 321, 41 },
		{ 142, 322, 41 },
		{ 142, 323, 41 },
		{ 142, 324, 41 },
		{ 142, 325, 41 },
		{ 142, 326, 41 },
		{ 142, 327, 41 },
		{ 142, 328, 41 },
		{ 142, 329, 41 },
		{ 142, 330, 41 },
		{ 142, 331, 41 },
		{ 142, 332, 41 },
		{ 142, 333, 41 },
		{ 142, 334, 41 },
		{ 142, 335, 41 },
		{ 142, 336, 41 },
		{ 142, 337, 41 },
		{ 142, 338, 41 },
		{ 142, 339, 41 },
		{ 142, 340, 41 },
		{ 142, 341, 41 },
		{ 142, 342, 41 },
		{ 142, 343, 41 },
		{ 143, 315, 41 },
		{ 143, 316, 41 },
		{ 143, 317, 41 },
		{ 143, 318, 41 },
		{ 143, 319, 41 },
		{ 143, 320, 41 },
		{ 143, 321, 41 },
		{ 143, 322, 41 },
		{ 143, 323, 41 },
		{ 143, 324, 41 },
		{ 143, 325, 41 },
		{ 143, 326, 41 },
		{ 143, 327, 41 },
		{ 143, 328, 41 },
		{ 143, 329, 41 },
		{ 143, 330, 41 },
		{ 143, 331, 41 },
		{ 143, 332, 41 },
		{ 143, 333, 41 },
		{ 143, 334, 41 },
		{ 143, 335, 41 },
		{ 143, 336, 41 },
		{ 143, 337, 41 },
		{ 143, 338, 41 },
		{ 143, 339, 41 },
		{ 143, 340, 41 },
		{ 143, 341, 41 },
		{ 143, 342, 41 },
		{ 144, 314, 41 },
		{ 144, 315, 41 },
		{ 144, 316, 41 },
		{ 144, 317, 41 },
		{ 144, 318, 41 },
		{ 144, 319, 41 },
		{ 144, 320, 41 },
		{ 144, 321, 41 },
		{ 144, 322, 41 },
		{ 144, 323, 41 },
		{ 144, 324, 41 },
		{ 144, 325, 41 },
		{ 144, 326, 41 },
		{ 144, 327, 41 },
		{ 144, 328, 41 },
		{ 144, 329, 41 },
		{ 144, 330, 41 },
		{ 144, 331, 41 },
		{ 144, 332, 41 },
		{ 144, 333, 41 },
		{ 144, 334, 41 },
		{ 144, 335, 41 },
		{ 144, 336, 41 },
		{ 144, 337, 41 },
		{ 144, 338, 41 },
		{ 144, 339, 41 },
		{ 144, 340, 41 },
		{ 144, 341, 41 },
		{ 145, 314, 41 },
		{ 145, 315, 41 },
		{ 145, 316, 41 },
		{ 145, 317, 41 },
		{ 145, 318, 41 },
		{ 145, 319, 41 },
		{ 145, 320, 41 },
		{ 145, 321, 41 },
		{ 145, 322, 41 },
		{ 145, 323, 41 },
		{ 145, 324, 41 },
		{ 145, 325, 41 },
		{ 145, 326, 41 },
		{ 145, 327, 41 },
		{ 145, 328, 41 },
		{ 145, 329, 41 },
		{ 145, 330, 41 },
		{ 145, 331, 41 },
		{ 145, 332, 41 },
		{ 145, 333, 41 },
		{ 145, 334, 41 },
		{ 145, 335, 41 },
		{ 145, 336, 41 },
		{ 145, 337, 41 },
		{ 145, 338, 41 },
		{ 145, 339, 41 },
		{ 145, 340, 41 },
		{ 146, 313, 41 },
		{ 146, 314, 41 },
		{ 146, 315, 41 },
		{ 146, 316, 41 },
		{ 146, 317, 41 },
		{ 146, 318, 41 },
		{ 146, 319, 41 },
		{ 146, 320, 41 },
		{ 146, 321, 41 },
		{ 146, 322, 41 },
		{ 146, 323, 41 },
		{ 146, 324, 41 },
		{ 146, 325, 41 },
		{ 146, 326, 41 },
		{ 146, 327, 41 },
		{ 146, 328, 41 },
		{ 146, 329, 41 },
		{ 146, 330, 41 },
		{ 146, 331, 41 },
		{ 146, 332, 41 },
		{ 146, 333, 41 },
		{ 146, 334, 41 },
		{ 146, 335, 41 },
		{ 146, 336, 41 },
		{ 146, 337, 41 },
		{ 146, 338, 41 },
		{ 146, 339, 41 },
		{ 146, 340, 41 },
		{ 147, 313, 41 },
		{ 147, 314, 41 },
		{ 147, 315, 41 },
		{ 147, 316, 41 },
		{ 147, 317, 41 },
		{ 147, 318, 41 },
		{ 147, 319, 41 },
		{ 147, 320, 41 },
		{ 147, 321, 41 },
		{ 147, 322, 41 },
		{ 147, 323, 41 },
		{ 147, 324, 41 },
		{ 147, 325, 41 },
		{ 147, 326, 41 },
		{ 147, 327, 41 },
		{ 147, 328, 41 },
		{ 147, 329, 41 },
		{ 147, 330, 41 },
		{ 147, 331, 41 },
		{ 147, 332, 41 },
		{ 147, 333, 41 },
		{ 147, 334, 41 },
		{ 147, 335, 41 },
		{ 147, 336, 41 },
		{ 147, 337, 41 },
		{ 147, 338, 41 },
		{ 147, 339, 41 },
		{ 148, 312, 41 },
		{ 148, 313, 41 },
		{ 148, 314, 41 },
		{ 148, 315, 41 },
		{ 148, 316, 41 },
		{ 148, 317, 41 },
		{ 148, 318, 41 },
		{ 148, 319, 41 },
		{ 148, 320, 41 },
		{ 148, 321, 41 },
		{ 148, 322, 41 },
		{ 148, 323, 41 },
		{ 148, 324, 41 },
		{ 148, 325, 41 },
		{ 148, 326, 41 },
		{ 148, 327, 41 },
		{ 148, 328, 41 },
		{ 148, 329, 41 },
		{ 148, 330, 41 },
		{ 148, 331, 41 },
		{ 148, 332, 41 },
		{ 148, 333, 41 },
		{ 148, 334, 41 },
		{ 148, 335, 41 },
		{ 148, 336, 41 },
		{ 148, 337, 41 },
		{ 148, 338, 41 },
		{ 149, 312, 41 },
		{ 149, 313, 41 },
		{ 149, 314, 41 },
		{ 149, 315, 41 },
		{ 149, 316, 41 },
		{ 149, 317, 41 },
		{ 149, 318, 41 },
		{ 149, 319, 41 },
		{ 149, 320, 41 },
		{ 149, 321, 41 },
		{ 149, 322, 41 },
		{ 149, 323, 41 },
		{ 149, 324, 41 },
		{ 149, 325, 41 },
		{ 149, 326, 41 },
		{ 149, 327, 41 },
		{ 149, 328, 41 },
		{ 149, 329, 41 },
		{ 149, 330, 41 },
		{ 149, 331, 41 },
		{ 149, 332, 41 },
		{ 149, 333, 41 },
		{ 149, 334, 41 },
		{ 149, 335, 41 },
		{ 149, 336, 41 },
		{ 149, 337, 41 },
		{ 150, 311, 41 },
		{ 150, 312, 41 },
		{ 150, 313, 41 },
		{ 150, 314, 41 },
		{ 150, 315, 41 },
		{ 150, 316, 41 },
		{ 150, 317, 41 },
		{ 150, 318, 41 },
		{ 150, 319, 41 },
		{ 150, 320, 41 },
		{ 150, 321, 41 },
		{ 150, 322, 41 },
		{ 150, 323, 41 },
		{ 150, 324, 41 },
		{ 150, 325, 41 },
		{ 150, 326, 41 },
		{ 150, 327, 41 },
		{ 150, 328, 41 },
		{ 150, 329, 41 },
		{ 150, 330, 41 },
		{ 150, 331, 41 },
		{ 150, 332, 41 },
		{ 150, 333, 41 },
		{ 150, 334, 41 },
		{ 150, 335, 41 },
		{ 150, 336, 41 },
		{ 150, 337, 41 },
		{ 151, 311, 41 },
		{ 151, 312, 41 },
		{ 151, 313, 41 },
		{ 151, 314, 41 },
		{ 151, 315, 41 },
		{ 151, 316, 41 },
		{ 151, 317, 41 },
		{ 151, 318, 41 },
		{ 151, 319, 41 },
		{ 151, 320, 41 },
		{ 151, 321, 41 },
		{ 151, 322, 41 },
		{ 151, 323, 41 },
		{ 151, 324, 41 },
		{ 151, 325, 41 },
		{ 151, 326, 41 },
		{ 151, 327, 41 },
		{ 151, 328, 41 },
		{ 151, 329, 41 },
		{ 151, 330, 41 },
		{ 151, 331, 41 },
		{ 151, 332, 41 },
		{ 151, 333, 41 },
		{ 151, 334, 41 },
		{ 151, 335, 41 },
		{ 151, 336, 41 },
		{ 152, 310, 41 },
		{ 152, 311, 41 },
		{ 152, 312, 41 },
		{ 152, 313, 41 },
		{ 152, 314, 41 },
		{ 152, 315, 41 },
		{ 152, 316, 41 },
		{ 152, 317, 41 },
		{ 152, 318, 41 },
		{ 152, 319, 41 },
		{ 152, 320, 41 },
		{ 152, 321, 41 },
		{ 152, 322, 41 },
		{ 152, 323, 41 },
		{ 152, 324, 41 },
		{ 152, 325, 41 },
		{ 152, 326, 41 },
		{ 152, 327, 41 },
		{ 152, 328, 41 },
		{ 152, 329, 41 },
		{ 152, 330, 41 },
		{ 152, 331, 41 },
		{ 152, 332, 41 },
		{ 152, 333, 41 },
		{ 152, 334, 41 },
		{ 152, 335, 41 },
		{ 153, 310, 41 },
		{ 153, 311, 41 },
		{ 153, 312, 41 },
		{ 153, 313, 41 },
		{ 153, 314, 41 },
		{ 153, 315, 41 },
		{ 153, 316, 41 },
		{ 153, 317, 41 },
		{ 153, 318, 41 },
		{ 153, 319, 41 },
		{ 153, 320, 41 },
		{ 153, 321, 41 },
		{ 153, 322, 41 },
		{ 153, 323, 41 },
		{ 153, 324, 41 },
		{ 153, 325, 41 },
		{ 153, 326, 41 },
		{ 153, 327, 41 },
		{ 153, 328, 41 },
		{ 153, 329, 41 },
		{ 153, 330, 41 },
		{ 153, 331, 41 },
		{ 153, 332, 41 },
		{ 153, 333, 41 },
		{ 153, 334, 41 },
		{ 154, 309, 41 },
		{ 154, 310, 41 },
		{ 154, 311, 41 },
		{ 154, 312, 41 },
		{ 154, 313, 41 },
		{ 154, 314, 41 },
		{ 154, 315, 41 },
		{ 154, 316, 41 },
		{ 154, 317, 41 },
		{ 154, 318, 41 },
		{ 154, 319, 41 },
		{ 154, 320, 41 },
		{ 154, 321, 41 },
		{ 154, 322, 41 },
		{ 154, 323, 41 },
		{ 154, 324, 41 },
		{ 154, 325, 41 },
		{ 154, 326, 41 },
		{ 154, 327, 41 },
		{ 154, 328, 41 },
		{ 154, 329, 41 },
		{ 154, 330, 41 },
		{ 154, 331, 41 },
		{ 154, 332, 41 },
		{ 154, 333, 41 },
		{ 154, 334, 41 },
		{ 155, 309, 41 },
		{ 155, 310, 41 },
		{ 155, 311, 41 },
		{ 155, 312, 41 },
		{ 155, 313, 41 },
		{ 155, 314, 41 },
		{ 155, 315, 41 },
		{ 155, 316, 41 },
		{ 155, 317, 41 },
		{ 155, 318, 41 },
		{ 155, 319, 41 },
		{ 155, 320, 41 },
		{ 155, 321, 41 },
		{ 155, 322, 41 },
		{ 155, 323, 41 },
		{ 155, 324, 41 },
		{ 155, 325, 41 },
		{ 155, 326, 41 },
		{ 155, 327, 41 },
		{ 155, 328, 41 },
		{ 155, 329, 41 },
		{ 155, 330, 41 },
		{ 155, 331, 41 },
		{ 155, 332, 41 },
		{ 155, 333, 41 },
		{ 156, 308, 41 },
		{ 156, 309, 41 },
		{ 156, 310, 41 },
		{ 156, 311, 41 },
		{ 156, 312, 41 },
		{ 156, 313, 41 },
		{ 156, 314, 41 },
		{ 156, 315, 41 },
		{ 156, 316, 41 },
		{ 156, 317, 41 },
		{ 156, 318, 41 },
		{ 156, 319, 41 },
		{ 156, 320, 41 },
		{ 156, 321, 41 },
		{ 156, 322, 41 },
		{ 156, 323, 41 },
		{ 156, 324, 41 },
		{ 156, 325, 41 },
		{ 156, 326, 41 },
		{ 156, 327, 41 },
		{ 156, 328, 41 },
		{ 156, 329, 41 },
		{ 156, 330, 41 },
		{ 156, 331, 41 },
		{ 156, 332, 41 },
		{ 157, 308, 41 },
		{ 157, 309, 41 },
		{ 157, 310, 41 },
		{ 157, 311, 41 },
		{ 157, 312, 41 },
		{ 157, 313, 41 },
		{ 157, 314, 41 },
		{ 157, 315, 41 },
		{ 157, 316, 41 },
		{ 157, 317, 41 },
		{ 157, 318, 41 },
		{ 157, 319, 41 },
		{ 157, 320, 41 },
		{ 157, 321, 41 },
		{ 157, 322, 41 },
		{ 157, 323, 41 },
		{ 157, 324, 41 },
		{ 157, 325, 41 },
		{ 157, 326, 41 },
		{ 157, 327, 41 },
		{ 157, 328, 41 },
		{ 157, 329, 41 },
		{ 157, 330, 41 },
		{ 157, 331, 41 },
		{ 158, 307, 41 },
		{ 158, 308, 41 },
		{ 158, 309, 41 },
		{ 158, 310, 41 },
		{ 158, 311, 41 },
		{ 158, 312, 41 },
		{ 158, 313, 41 },
		{ 158, 314, 41 },
		{ 158, 315, 41 },
		{ 158, 316, 41 },
		{ 158, 317, 41 },
		{ 158, 318, 41 },
		{ 158, 319, 41 },
		{ 158, 320, 41 },
		{ 158, 321, 41 },
		{ 158, 322, 41 },
		{ 158, 323, 41 },
		{ 158, 324, 41 },
		{ 158, 325, 41 },
		{ 158, 326, 41 },
		{ 158, 327, 41 },
		{ 158, 328, 41 },
		{ 158, 329, 41 },
		{ 158, 330, 41 },
		{ 158, 331, 41 },
		{ 159, 307, 41 },
		{ 159, 308, 41 },
		{ 159, 309, 41 },
		{ 159, 310, 41 },
		{ 159, 311, 41 },
		{ 159, 312, 41 },
		{ 159, 313, 41 },
		{ 159, 314, 41 },
		{ 159, 315, 41 },
		{ 159, 316, 41 },
		{ 159, 317, 41 },
		{ 159, 318, 41 },
		{ 159, 319, 41 },
		{ 159, 320, 41 },
		{ 159, 321, 41 },
		{ 159, 322, 41 },
		{ 159, 323, 41 },
		{ 159, 324, 41 },
		{ 159, 325, 41 },
		{ 159, 326, 41 },
		{ 159, 327, 41 },
		{ 159, 328, 41 },
		{ 159, 329, 41 },
		{ 159, 330, 41 },
		{ 160, 306, 41 },
		{ 160, 307, 41 },
		{ 160, 308, 41 },
		{ 160, 309, 41 },
		{ 160, 310, 41 },
		{ 160, 311, 41 },
		{ 160, 312, 41 },
		{ 160, 313, 41 },
		{ 160, 314, 41 },
		{ 160, 315, 41 },
		{ 160, 316, 41 },
		{ 160, 317, 41 },
		{ 160, 318, 41 },
		{ 160, 319, 41 },
		{ 160, 320, 41 },
		{ 160, 321, 41 },
		{ 160, 322, 41 },
		{ 160, 323, 41 },
		{ 160, 324, 41 },
		{ 160, 325, 41 },
		{ 160, 326, 41 },
		{ 160, 327, 41 },
		{ 160, 328, 41 },
		{ 160, 329, 41 },
		{ 161, 306, 41 },
		{ 161, 307, 41 },
		{ 161, 308, 41 },
		{ 161, 309, 41 },
		{ 161, 310, 41 },
		{ 161, 311, 41 },
		{ 161, 312, 41 },
		{ 161, 313, 41 },
		{ 161, 314, 41 },
		{ 161, 315, 41 },
		{ 161, 316, 41 },
		{ 161, 317, 41 },
		{ 161, 318, 41 },
		{ 161, 319, 41 },
		{ 161, 320, 41 },
		{ 161, 321, 41 },
		{ 161, 322, 41 },
		{ 161, 323, 41 },
		{ 161, 324, 41 },
		{ 161, 325, 41 },
		{ 161, 326, 41 },
		{ 161, 327, 41 },
		{ 161, 328, 41 },
		{ 162, 305, 41 },
		{ 162, 306, 41 },
		{ 162, 307, 41 },
		{ 162, 308, 41 },
		{ 162, 309, 41 },
		{ 162, 310, 41 },
		{ 162, 311, 41 },
		{ 162, 312, 41 },
		{ 162, 313, 41 },
		{ 162, 314, 41 },
		{ 162, 315, 41 },
		{ 162, 316, 41 },
		{ 162, 317, 41 },
		{ 162, 318, 41 },
		{ 162, 319, 41 },
		{ 162, 320, 41 },
		{ 162, 321, 41 },
		{ 162, 322, 41 },
		{ 162, 323, 41 },
		{ 162, 324, 41 },
		{ 162, 325, 41 },
		{ 162, 326, 41 },
		{ 162, 327, 41 },
		{ 162, 328, 41 },
		{ 163, 305, 41 },
		{ 163, 306, 41 },
		{ 163, 307, 41 },
		{ 163, 308, 41 },
		{ 163, 309, 41 },
		{ 163, 310, 41 },
		{ 163, 311, 41 },
		{ 163, 312, 41 },
		{ 163, 313, 41 },
		{ 163, 314, 41 },
		{ 163, 315, 41 },
		{ 163, 316, 41 },
		{ 163, 317, 41 },
		{ 163, 318, 41 },
		{ 163, 319, 41 },
		{ 163, 320, 41 },
		{ 163, 321, 41 },
		{ 163, 322, 41 },
		{ 163, 323, 41 },
		{ 163, 324, 41 },
		{ 163, 325, 41 },
		{ 163, 326, 41 },
		{ 163, 327, 41 },
		{ 164, 304, 41 },
		{ 164, 305, 41 },
		{ 164, 306, 41 },
		{ 164, 307, 41 },
		{ 164, 308, 41 },
		{ 164, 309, 41 },
		{ 164, 310, 41 },
		{ 164, 311, 41 },
		{ 164, 312, 41 },
		{ 164, 313, 41 },
		{ 164, 314, 41 },
		{ 164, 315, 41 },
		{ 164, 316, 41 },
		{ 164, 317, 41 },
		{ 164, 318, 41 },
		{ 164, 319, 41 },
		{ 164, 320, 41 },
		{ 164, 321, 41 },
		{ 164, 322, 41 },
		{ 164, 323, 41 },
		{ 164, 324, 41 },
		{ 164, 325, 41 },
		{ 164, 326, 41 },
		{ 165, 304, 41 },
		{ 165, 305, 41 },
		{ 165, 306, 41 },
		{ 165, 307, 41 },
		{ 165, 308, 41 },
		{ 165, 309, 41 },
		{ 165, 310, 41 },
		{ 165, 311, 41 },
		{ 165, 312, 41 },
		{ 165, 313, 41 },
		{ 165, 314, 41 },
		{ 165, 315, 41 },
		{ 165, 316, 41 },
		{ 165, 317, 41 },
		{ 165, 318, 41 },
		{ 165, 319, 41 },
		{ 165, 320, 41 },
		{ 165, 321, 41 },
		{ 165, 322, 41 },
		{ 165, 323, 41 },
		{ 165, 324, 41 },
		{ 165, 325, 41 },
		{ 166, 303, 41 },
		{ 166, 304, 41 },
		{ 166, 305, 41 },
		{ 166, 306, 41 },
		{ 166, 307, 41 },
		{ 166, 308, 41 },
		{ 166, 309, 41 },
		{ 166, 310, 41 },
		{ 166, 311, 41 },
		{ 166, 312, 41 },
		{ 166, 313, 41 },
		{ 166, 314, 41 },
		{ 166, 315, 41 },
		{ 166, 316, 41 },
		{ 166, 317, 41 },
		{ 166, 318, 41 },
		{ 166, 319, 41 },
		{ 166, 320, 41 },
		{ 166, 321, 41 },
		{ 166, 322, 41 },
		{ 166, 323, 41 },
		{ 166, 324, 41 },
		{ 166, 325, 41 },
		{ 167, 303, 41 },
		{ 167, 304, 41 },
		{ 167, 305, 41 },
		{ 167, 306, 41 },
		{ 167, 307, 41 },
		{ 167, 308, 41 },
		{ 167, 309, 41 },
		{ 167, 310, 41 },
		{ 167, 311, 41 },
		{ 167, 312, 41 },
		{ 167, 313, 41 },
		{ 167, 314, 41 },
		{ 167, 315, 41 },
		{ 167, 316, 41 },
		{ 167, 317, 41 },
		{ 167, 318, 41 },
		{ 167, 319, 41 },
		{ 167, 320, 41 },
		{ 167, 321, 41 },
		{ 167, 322, 41 },
		{ 167, 323, 41 },
		{ 167, 324, 41 },
		{ 168, 302, 41 },
		{ 168, 303, 41 },
		{ 168, 304, 41 },
		{ 168, 305, 41 },
		{ 168, 306, 41 },
		{ 168, 307, 41 },
		{ 168, 308, 41 },
		{ 168, 309, 41 },
		{ 168, 310, 41 },
		{ 168, 311, 41 },
		{ 168, 312, 41 },
		{ 168, 313, 41 },
		{ 168, 314, 41 },
		{ 168, 315, 41 },
		{ 168, 316, 41 },
		{ 168, 317, 41 },
		{ 168, 318, 41 },
		{ 168, 319, 41 },
		{ 168, 320, 41 },
		{ 168, 321, 41 },
		{ 168, 322, 41 },
		{ 168, 323, 41 },
		{ 169, 302, 41 },
		{ 169, 303, 41 },
		{ 169, 304, 41 },
		{ 169, 305, 41 },
		{ 169, 306, 41 },
		{ 169, 307, 41 },
		{ 169, 308, 41 },
		{ 169, 309, 41 },
		{ 169, 310, 41 },
		{ 169, 311, 41 },
		{ 169, 312, 41 },
		{ 169, 313, 41 },
		{ 169, 314, 41 },
		{ 169, 315, 41 },
		{ 169, 316, 41 },
		{ 169, 317, 41 },
		{ 169, 318, 41 },
		{ 169, 319, 41 },
		{ 169, 320, 41 },
		{ 169, 321, 41 },
		{ 169, 322, 41 },
		{ 170, 301, 41 },
		{ 170, 302, 41 },
		{ 170, 303, 41 },
		{ 170, 304, 41 },
		{ 170, 305, 41 },
		{ 170, 306, 41 },
		{ 170, 307, 41 },
		{ 170, 308, 41 },
		{ 170, 309, 41 },
		{ 170, 310, 41 },
		{ 170, 311, 41 },
		{ 170, 312, 41 },
		{ 170, 313, 41 },
		{ 170, 314, 41 },
		{ 170, 315, 41 },
		{ 170, 316, 41 },
		{ 170, 317, 41 },
		{ 170, 318, 41 },
		{ 170, 319, 41 },
		{ 170, 320, 41 },
		{ 170, 321, 41 },
		{ 170, 322, 41 },
		{ 171, 301, 41 },
		{ 171, 302, 41 },
		{ 171, 303, 41 },
		{ 171, 304, 41 },
		{ 171, 305, 41 },
		{ 171, 306, 41 },
		{ 171, 307, 41 },
		{ 171, 308, 41 },
		{ 171, 309, 41 },
		{ 171, 310, 41 },
		{ 171, 311, 41 },
		{ 171, 312, 41 },
		{ 171, 313, 41 },
		{ 171, 314, 41 },
		{ 171, 315, 41 },
		{ 171, 316, 41 },
		{ 171, 317, 41 },
		{ 171, 318, 41 },
		{ 171, 319, 41 },
		{ 171, 320, 41 },
		{ 171, 321, 41 },
		{ 172, 300, 41 },
		{ 172, 301, 41 },
		{ 172, 302, 41 },
		{ 172, 303, 41 },
		{ 172, 304, 41 },
		{ 172, 305, 41 },
		{ 172, 306, 41 },
		{ 172, 307, 41 },
		{ 172, 308, 41 },
		{ 172, 309, 41 },
		{ 172, 310, 41 },
		{ 172, 311, 41 },
		{ 172, 312, 41 },
		{ 172, 313, 41 },
		{ 172, 314, 41 },
		{ 172, 315, 41 },
		{ 172, 316, 41 },
		{ 172, 317, 41 },
		{ 172, 318, 41 },
		{ 172, 319, 41 },
		{ 172, 320, 41 },
		{ 173, 300, 41 },
		{ 173, 301, 41 },
		{ 173, 302, 41 },
		{ 173, 303, 41 },
		{ 173, 304, 41 },
		{ 173, 305, 41 },
		{ 173, 306, 41 },
		{ 173, 307, 41 },
		{ 173, 308, 41 },
		{ 173, 309, 41 },
		{ 173, 310, 41 },
		{ 173, 311, 41 },
		{ 173, 312, 41 },
		{ 173, 313, 41 },
		{ 173, 314, 41 },
		{ 173, 315, 41 },
		{ 173, 316, 41 },
		{ 173, 317, 41 },
		{ 173, 318, 41 },
		{ 173, 319, 41 },
		{ 174, 299, 41 },
		{ 174, 300, 41 },
		{ 174, 301, 41 },
		{ 174, 302, 41 },
		{ 174, 303, 41 },
		{ 174, 304, 41 },
		{ 174, 305, 41 },
		{ 174, 306, 41 },
		{ 174, 307, 41 },
		{ 174, 308, 41 },
		{ 174, 309, 41 },
		{ 174, 310, 41 },
		{ 174, 311, 41 },
		{ 174, 312, 41 },
		{ 174, 313, 41 },
		{ 174, 314, 41 },
		{ 174, 315, 41 },
		{ 174, 316, 41 },
		{ 174, 317, 41 },
		{ 174, 318, 41 },
		{ 174, 319, 41 },
		{ 175, 299, 41 },
		{ 175, 300, 41 },
		{ 175, 301, 41 },
		{ 175, 302, 41 },
		{ 175, 303, 41 },
		{ 175, 304, 41 },
		{ 175, 305, 41 },
		{ 175, 306, 41 },
		{ 175, 307, 41 },
		{ 175, 308, 41 },
		{ 175, 309, 41 },
		{ 175, 310, 41 },
		{ 175, 311, 41 },
		{ 175, 312, 41 },
		{ 175, 313, 41 },
		{ 175, 314, 41 },
		{ 175, 315, 41 },
		{ 175, 316, 41 },
		{ 175, 317, 41 },
		{ 175, 318, 41 },
		{ 176, 298, 41 },
		{ 176, 299, 41 },
		{ 176, 300, 41 },
		{ 176, 301, 41 },
		{ 176, 302, 41 },
		{ 176, 303, 41 },
		{ 176, 304, 41 },
		{ 176, 305, 41 },
		{ 176, 306, 41 },
		{ 176, 307, 41 },
		{ 176, 308, 41 },
		{ 176, 309, 41 },
		{ 176, 310, 41 },
		{ 176, 311, 41 },
		{ 176, 312, 41 },
		{ 176, 313, 41 },
		{ 176, 314, 41 },
		{ 176, 315, 41 },
		{ 176, 316, 41 },
		{ 176, 317, 41 },
		{ 177, 298, 41 },
		{ 177, 299, 41 },
		{ 177, 300, 41 },
		{ 177, 301, 41 },
		{ 177, 302, 41 },
		{ 177, 303, 41 },
		{ 177, 304, 41 },
		{ 177, 305, 41 },
		{ 177, 306, 41 },
		{ 177, 307, 41 },
		{ 177, 308, 41 },
		{ 177, 309, 41 },
		{ 177, 310, 41 },
		{ 177, 311, 41 },
		{ 177, 312, 41 },
		{ 177, 313, 41 },
		{ 177, 314, 41 },
		{ 177, 315, 41 },
		{ 177, 316, 41 },
		{ 178, 297, 41 },
		{ 178, 298, 41 },
		{ 178, 299, 41 },
		{ 178, 300, 41 },
		{ 178, 301, 41 },
		{ 178, 302, 41 },
		{ 178, 303, 41 },
		{ 178, 304, 41 },
		{ 178, 305, 41 },
		{ 178, 306, 41 },
		{ 178, 307, 41 },
		{ 178, 308, 41 },
		{ 178, 309, 41 },
		{ 178, 310, 41 },
		{ 178, 311, 41 },
		{ 178, 312, 41 },
		{ 178, 313, 41 },
		{ 178, 314, 41 },
		{ 178, 315, 41 },
		{ 178, 316, 41 },
		{ 179, 297, 41 },
		{ 179, 298, 41 },
		{ 179, 299, 41 },
		{ 179, 300, 41 },
		{ 179, 301, 41 },
		{ 179, 302, 41 },
		{ 179, 303, 41 },
		{ 179, 304, 41 },
		{ 179, 305, 41 },
		{ 179, 306, 41 },
		{ 179, 307, 41 },
		{ 179, 308, 41 },
		{ 179, 309, 41 },
		{ 179, 310, 41 },
		{ 179, 311, 41 },
		{ 179, 312, 41 },
		{ 179, 313, 41 },
		{ 179, 314, 41 },
		{ 179, 315, 41 },
		{ 180, 296, 41 },
		{ 180, 297, 41 },
		{ 180, 298, 41 },
		{ 180, 299, 41 },
		{ 180, 300, 41 },
		{ 180, 301, 41 },
		{ 180, 302, 41 },
		{ 180, 303, 41 },
		{ 180, 304, 41 },
		{ 180, 305, 41 },
		{ 180, 306, 41 },
		{ 180, 307, 41 },
		{ 180, 308, 41 },
		{ 180, 309, 41 },
		{ 180, 310, 41 },
		{ 180, 311, 41 },
		{ 180, 312, 41 },
		{ 180, 313, 41 },
		{ 180, 314, 41 },
		{ 181, 296, 41 },
		{ 181, 297, 41 },
		{ 181, 298, 41 },
		{ 181, 299, 41 },
		{ 181, 300, 41 },
		{ 181, 301, 41 },
		{ 181, 302, 41 },
		{ 181, 303, 41 },
		{ 181, 304, 41 },
		{ 181, 305, 41 },
		{ 181, 306, 41 },
		{ 181, 307, 41 },
		{ 181, 308, 41 },
		{ 181, 309, 41 },
		{ 181, 310, 41 },
		{ 181, 311, 41 },
		{ 181, 312, 41 },
		{ 181, 313, 41 },
		{ 182, 295, 41 },
		{ 182, 296, 41 },
		{ 182, 297, 41 },
		{ 182, 298, 41 },
		{ 182, 299, 41 },
		{ 182, 300, 41 },
		{ 182, 301, 41 },
		{ 182, 302, 41 },
		{ 182, 303, 41 },
		{ 182, 304, 41 },
		{ 182, 305, 41 },
		{ 182, 306, 41 },
		{ 182, 307, 41 },
		{ 182, 308, 41 },
		{ 182, 309, 41 },
		{ 182, 310, 41 },
		{ 182, 311, 41 },
		{ 182, 312, 41 },
		{ 182, 313, 41 },
		{ 183, 295, 41 },
		{ 183, 296, 41 },
		{ 183, 297, 41 },
		{ 183, 298, 41 },
		{ 183, 299, 41 },
		{ 183, 300, 41 },
		{ 183, 301, 41 },
		{ 183, 302, 41 },
		{ 183, 303, 41 },
		{ 183, 304, 41 },
		{ 183, 305, 41 },
		{ 183, 306, 41 },
		{ 183, 307, 41 },
		{ 183, 308, 41 },
		{ 183, 309, 41 },
		{ 183, 310, 41 },
		{ 183, 311, 41 },
		{ 183, 312, 41 },
		{ 184, 294, 41 },
		{ 184, 295, 41 },
		{ 184, 296, 41 },
		{ 184, 297, 41 },
		{ 184, 298, 41 },
		{ 184, 299, 41 },
		{ 184, 300, 41 },
		{ 184, 301, 41 },
		{ 184, 302, 41 },
		{ 184, 303, 41 },
		{ 184, 304, 41 },
		{ 184, 305, 41 },
		{ 184, 306, 41 },
		{ 184, 307, 41 },
		{ 184, 308, 41 },
		{ 184, 309, 41 },
		{ 184, 310, 41 },
		{ 184, 311, 41 },
		{ 185, 294, 41 },
		{ 185, 295, 41 },
		{ 185, 296, 41 },
		{ 185, 297, 41 },
		{ 185, 298, 41 },
		{ 185, 299, 41 },
		{ 185, 300, 41 },
		{ 185, 301, 41 },
		{ 185, 302, 41 },
		{ 185, 303, 41 },
		{ 185, 304, 41 },
		{ 185, 305, 41 },
		{ 185, 306, 41 },
		{ 185, 307, 41 },
		{ 185, 308, 41 },
		{ 185, 309, 41 },
		{ 185, 310, 41 },
		{ 186, 293, 41 },
		{ 186, 294, 41 },
		{ 186, 295, 41 },
		{ 186, 296, 41 },
		{ 186, 297, 41 },
		{ 186, 298, 41 },
		{ 186, 299, 41 },
		{ 186, 300, 41 },
		{ 186, 301, 41 },
		{ 186, 302, 41 },
		{ 186, 303, 41 },
		{ 186, 304, 41 },
		{ 186, 305, 41 },
		{ 186, 306, 41 },
		{ 186, 307, 41 },
		{ 186, 308, 41 },
		{ 186, 309, 41 },
		{ 186, 310, 41 },
		{ 187, 293, 41 },
		{ 187, 294, 41 },
		{ 187, 295, 41 },
		{ 187, 296, 41 },
		{ 187, 297, 41 },
		{ 187, 298, 41 },
		{ 187, 299, 41 },
		{ 187, 300, 41 },
		{ 187, 301, 41 },
		{ 187, 302, 41 },
		{ 187, 303, 41 },
		{ 187, 304, 41 },
		{ 187, 305, 41 },
		{ 187, 306, 41 },
		{ 187, 307, 41 },
		{ 187, 308, 41 },
		{ 187, 309, 41 },
		{ 188, 292, 41 },
		{ 188, 293, 41 },
		{ 188, 294, 41 },
		{ 188, 295, 41 },
		{ 188, 296, 41 },
		{ 188, 297, 41 },
		{ 188, 298, 41 },
		{ 188, 299, 41 },
		{ 188, 300, 41 },
		{ 188, 301, 41 },
		{ 188, 302, 41 },
		{ 188, 303, 41 },
		{ 188, 304, 41 },
		{ 188, 305, 41 },
		{ 188, 306, 41 },
		{ 188, 307, 41 },
		{ 188, 308, 41 },
		{ 189, 292, 41 },
		{ 189, 293, 41 },
		{ 189, 294, 41 },
		{ 189, 295, 41 },
		{ 189, 296, 41 },
		{ 189, 297, 41 },
		{ 189, 298, 41 },
		{ 189, 299, 41 },
		{ 189, 300, 41 },
		{ 189, 301, 41 },
		{ 189, 302, 41 },
		{ 189, 303, 41 },
		{ 189, 304, 41 },
		{ 189, 305, 41 },
		{ 189, 306, 41 },
		{ 189, 307, 41 },
		{ 190, 292, 41 },
		{ 190, 293, 41 },
		{ 190, 294, 41 },
		{ 190, 295, 41 },
		{ 190, 296, 41 },
		{ 190, 297, 41 },
		{ 190, 298, 41 },
		{ 190, 299, 41 },
		{ 190, 300, 41 },
		{ 190, 301, 41 },
		{ 190, 302, 41 },
		{ 190, 303, 41 },
		{ 190, 304, 41 },
		{ 190, 305, 41 },
		{ 190, 306, 41 },
		{ 191, 292, 41 },
		{ 191, 293, 41 },
		{ 191, 294, 41 },
		{ 191, 295, 41 },
		{ 191, 296, 41 },
		{ 191, 297, 41 },
		{ 191, 298, 41 },
		{ 191, 299, 41 },
		{ 191, 300, 41 },
		{ 191, 301, 41 },
		{ 191, 302, 41 },
		{ 191, 303, 41 },
		{ 191, 304, 41 },
		{ 323, 210, 41 },
		{ 323, 211, 41 },
		{ 323, 212, 41 },
		{ 323, 213, 41 },
		{ 323, 214, 41 },
		{ 323, 215, 41 },
		{ 323, 216, 41 },
		{ 323, 217, 41 },
		{ 323, 218, 41 },
		{ 323, 219, 41 },
		{ 323, 220, 41 },
		{ 323, 221, 41 },
		{ 323, 222, 41 },
		{ 324, 208, 41 },
		{ 324, 209, 41 },
		{ 324, 210, 41 },
		{ 324, 211, 41 },
		{ 324, 212, 41 },
		{ 324, 213, 41 },
		{ 324, 214, 41 },
		{ 324, 215, 41 },
		{ 324, 216, 41 },
		{ 324, 217, 41 },
		{ 324, 218, 41 },
		{ 324, 219, 41 },
		{ 324, 220, 41 },
		{ 324, 221, 41 },
		{ 324, 222, 41 },
		{ 325, 207, 41 },
		{ 325, 208, 41 },
		{ 325, 209, 41 },
		{ 325, 210, 41 },
		{ 325, 211, 41 },
		{ 325, 212, 41 },
		{ 325, 213, 41 },
		{ 325, 214, 41 },
		{ 325, 215, 41 },
		{ 325, 216, 41 },
		{ 325, 217, 41 },
		{ 325, 218, 41 },
		{ 325, 219, 41 },
		{ 325, 220, 41 },
		{ 325, 221, 41 },
		{ 325, 222, 41 },
		{ 326, 206, 41 },
		{ 326, 207, 41 },
		{ 326, 208, 41 },
		{ 326, 209, 41 },
		{ 326, 210, 41 },
		{ 326, 211, 41 },
		{ 326, 212, 41 },
		{ 326, 213, 41 },
		{ 326, 214, 41 },
		{ 326, 215, 41 },
		{ 326, 216, 41 },
		{ 326, 217, 41 },
		{ 326, 218, 41 },
		{ 326, 219, 41 },
		{ 326, 220, 41 },
		{ 326, 221, 41 },
		{ 326, 222, 41 },
		{ 327, 205, 41 },
		{ 327, 206, 41 },
		{ 327, 207, 41 },
		{ 327, 208, 41 },
		{ 327, 209, 41 },
		{ 327, 210, 41 },
		{ 327, 211, 41 },
		{ 327, 212, 41 },
		{ 327, 213, 41 },
		{ 327, 214, 41 },
		{ 327, 215, 41 },
		{ 327, 216, 41 },
		{ 327, 217, 41 },
		{ 327, 218, 41 },
		{ 327, 219, 41 },
		{ 327, 220, 41 },
		{ 327, 221, 41 },
		{ 328, 204, 41 },
		{ 328, 205, 41 },
		{ 328, 206, 41 },
		{ 328, 207, 41 },
		{ 328, 208, 41 },
		{ 328, 209, 41 },
		{ 328, 210, 41 },
		{ 328, 211, 41 },
		{ 328, 212, 41 },
		{ 328, 213, 41 },
		{ 328, 214, 41 },
		{ 328, 215, 41 },
		{ 328, 216, 41 },
		{ 328, 217, 41 },
		{ 328, 218, 41 },
		{ 328, 219, 41 },
		{ 328, 220, 41 },
		{ 328, 221, 41 },
		{ 329, 204, 41 },
		{ 329, 205, 41 },
		{ 329, 206, 41 },
		{ 329, 207, 41 },
		{ 329, 208, 41 },
		{ 329, 209, 41 },
		{ 329, 210, 41 },
		{ 329, 211, 41 },
		{ 329, 212, 41 },
		{ 329, 213, 41 },
		{ 329, 214, 41 },
		{ 329, 215, 41 },
		{ 329, 216, 41 },
		{ 329, 217, 41 },
		{ 329, 218, 41 },
		{ 329, 219, 41 },
		{ 329, 220, 41 },
		{ 330, 203, 41 },
		{ 330, 204, 41 },
		{ 330, 205, 41 },
		{ 330, 206, 41 },
		{ 330, 207, 41 },
		{ 330, 208, 41 },
		{ 330, 209, 41 },
		{ 330, 210, 41 },
		{ 330, 211, 41 },
		{ 330, 212, 41 },
		{ 330, 213, 41 },
		{ 330, 214, 41 },
		{ 330, 215, 41 },
		{ 330, 216, 41 },
		{ 330, 217, 41 },
		{ 330, 218, 41 },
		{ 330, 219, 41 },
		{ 330, 220, 41 },
		{ 331, 202, 41 },
		{ 331, 203, 41 },
		{ 331, 204, 41 },
		{ 331, 205, 41 },
		{ 331, 206, 41 },
		{ 331, 207, 41 },
		{ 331, 208, 41 },
		{ 331, 209, 41 },
		{ 331, 210, 41 },
		{ 331, 211, 41 },
		{ 331, 212, 41 },
		{ 331, 213, 41 },
		{ 331, 214, 41 },
		{ 331, 215, 41 },
		{ 331, 216, 41 },
		{ 331, 217, 41 },
		{ 331, 218, 41 },
		{ 331, 219, 41 },
		{ 332, 201, 41 },
		{ 332, 202, 41 },
		{ 332, 203, 41 },
		{ 332, 204, 41 },
		{ 332, 205, 41 },
		{ 332, 206, 41 },
		{ 332, 207, 41 },
		{ 332, 208, 41 },
		{ 332, 209, 41 },
		{ 332, 210, 41 },
		{ 332, 211, 41 },
		{ 332, 212, 41 },
		{ 332, 213, 41 },
		{ 332, 214, 41 },
		{ 332, 215, 41 },
		{ 332, 216, 41 },
		{ 332, 217, 41 },
		{ 332, 218, 41 },
		{ 332, 219, 41 },
		{ 333, 201, 41 },
		{ 333, 202, 41 },
		{ 333, 203, 41 },
		{ 333, 204, 41 },
		{ 333, 205, 41 },
		{ 333, 206, 41 },
		{ 333, 207, 41 },
		{ 333, 208, 41 },
		{ 333, 209, 41 },
		{ 333, 210, 41 },
		{ 333, 211, 41 },
		{ 333, 212, 41 },
		{ 333, 213, 41 },
		{ 333, 214, 41 },
		{ 333, 215, 41 },
		{ 333, 216, 41 },
		{ 333, 217, 41 },
		{ 333, 218, 41 },
		{ 334, 200, 41 },
		{ 334, 201, 41 },
		{ 334, 202, 41 },
		{ 334, 203, 41 },
		{ 334, 204, 41 },
		{ 334, 205, 41 },
		{ 334, 206, 41 },
		{ 334, 207, 41 },
		{ 334, 208, 41 },
		{ 334, 209, 41 },
		{ 334, 210, 41 },
		{ 334, 211, 41 },
		{ 334, 212, 41 },
		{ 334, 213, 41 },
		{ 334, 214, 41 },
		{ 334, 215, 41 },
		{ 334, 216, 41 },
		{ 334, 217, 41 },
		{ 334, 218, 41 },
		{ 335, 199, 41 },
		{ 335, 200, 41 },
		{ 335, 201, 41 },
		{ 335, 202, 41 },
		{ 335, 203, 41 },
		{ 335, 204, 41 },
		{ 335, 205, 41 },
		{ 335, 206, 41 },
		{ 335, 207, 41 },
		{ 335, 208, 41 },
		{ 335, 209, 41 },
		{ 335, 210, 41 },
		{ 335, 211, 41 },
		{ 335, 212, 41 },
		{ 335, 213, 41 },
		{ 335, 214, 41 },
		{ 335, 215, 41 },
		{ 335, 216, 41 },
		{ 335, 217, 41 },
		{ 336, 198, 41 },
		{ 336, 199, 41 },
		{ 336, 200, 41 },
		{ 336, 201, 41 },
		{ 336, 202, 41 },
		{ 336, 203, 41 },
		{ 336, 204, 41 },
		{ 336, 205, 41 },
		{ 336, 206, 41 },
		{ 336, 207, 41 },
		{ 336, 208, 41 },
		{ 336, 209, 41 },
		{ 336, 210, 41 },
		{ 336, 211, 41 },
		{ 336, 212, 41 },
		{ 336, 213, 41 },
		{ 336, 214, 41 },
		{ 336, 215, 41 },
		{ 336, 216, 41 },
		{ 336, 217, 41 },
		{ 337, 198, 41 },
		{ 337, 199, 41 },
		{ 337, 200, 41 },
		{ 337, 201, 41 },
		{ 337, 202, 41 },
		{ 337, 203, 41 },
		{ 337, 204, 41 },
		{ 337, 205, 41 },
		{ 337, 206, 41 },
		{ 337, 207, 41 },
		{ 337, 208, 41 },
		{ 337, 209, 41 },
		{ 337, 210, 41 },
		{ 337, 211, 41 },
		{ 337, 212, 41 },
		{ 337, 213, 41 },
		{ 337, 214, 41 },
		{ 337, 215, 41 },
		{ 337, 216, 41 },
		{ 338, 197, 41 },
		{ 338, 198, 41 },
		{ 338, 199, 41 },
		{ 338, 200, 41 },
		{ 338, 201, 41 },
		{ 338, 202, 41 },
		{ 338, 203, 41 },
		{ 338, 204, 41 },
		{ 338, 205, 41 },
		{ 338, 206, 41 },
		{ 338, 207, 41 },
		{ 338, 208, 41 },
		{ 338, 209, 41 },
		{ 338, 210, 41 },
		{ 338, 211, 41 },
		{ 338, 212, 41 },
		{ 338, 213, 41 },
		{ 338, 214, 41 },
		{ 338, 215, 41 },
		{ 338, 216, 41 },
		{ 339, 196, 41 },
		{ 339, 197, 41 },
		{ 339, 198, 41 },
		{ 339, 199, 41 },
		{ 339, 200, 41 },
		{ 339, 201, 41 },
		{ 339, 202, 41 },
		{ 339, 203, 41 },
		{ 339, 204, 41 },
		{ 339, 205, 41 },
		{ 339, 206, 41 },
		{ 339, 207, 41 },
		{ 339, 208, 41 },
		{ 339, 209, 41 },
		{ 339, 210, 41 },
		{ 339, 211, 41 },
		{ 339, 212, 41 },
		{ 339, 213, 41 },
		{ 339, 214, 41 },
		{ 339, 215, 41 },
		{ 340, 195, 41 },
		{ 340, 196, 41 },
		{ 340, 197, 41 },
		{ 340, 198, 41 },
		{ 340, 199, 41 },
		{ 340, 200, 41 },
		{ 340, 201, 41 },
		{ 340, 202, 41 },
		{ 340, 203, 41 },
		{ 340, 204, 41 },
		{ 340, 205, 41 },
		{ 340, 206, 41 },
		{ 340, 207, 41 },
		{ 340, 208, 41 },
		{ 340, 209, 41 },
		{ 340, 210, 41 },
		{ 340, 211, 41 },
		{ 340, 212, 41 },
		{ 340, 213, 41 },
		{ 340, 214, 41 },
		{ 340, 215, 41 },
		{ 341, 195, 41 },
		{ 341, 196, 41 },
		{ 341, 197, 41 },
		{ 341, 198, 41 },
		{ 341, 199, 41 },
		{ 341, 200, 41 },
		{ 341, 201, 41 },
		{ 341, 202, 41 },
		{ 341, 203, 41 },
		{ 341, 204, 41 },
		{ 341, 205, 41 },
		{ 341, 206, 41 },
		{ 341, 207, 41 },
		{ 341, 208, 41 },
		{ 341, 209, 41 },
		{ 341, 210, 41 },
		{ 341, 211, 41 },
		{ 341, 212, 41 },
		{ 341, 213, 41 },
		{ 341, 214, 41 },
		{ 342, 194, 41 },
		{ 342, 195, 41 },
		{ 342, 196, 41 },
		{ 342, 197, 41 },
		{ 342, 198, 41 },
		{ 342, 199, 41 },
		{ 342, 200, 41 },
		{ 342, 201, 41 },
		{ 342, 202, 41 },
		{ 342, 203, 41 },
		{ 342, 204, 41 },
		{ 342, 205, 41 },
		{ 342, 206, 41 },
		{ 342, 207, 41 },
		{ 342, 208, 41 },
		{ 342, 209, 41 },
		{ 342, 210, 41 },
		{ 342, 211, 41 },
		{ 342, 212, 41 },
		{ 342, 213, 41 },
		{ 342, 214, 41 },
		{ 343, 193, 41 },
		{ 343, 194, 41 },
		{ 343, 195, 41 },
		{ 343, 196, 41 },
		{ 343, 197, 41 },
		{ 343, 198, 41 },
		{ 343, 199, 41 },
		{ 343, 200, 41 },
		{ 343, 201, 41 },
		{ 343, 202, 41 },
		{ 343, 203, 41 },
		{ 343, 204, 41 },
		{ 343, 205, 41 },
		{ 343, 206, 41 },
		{ 343, 207, 41 },
		{ 343, 208, 41 },
		{ 343, 209, 41 },
		{ 343, 210, 41 },
		{ 343, 211, 41 },
		{ 343, 212, 41 },
		{ 343, 213, 41 },
		{ 344, 192, 41 },
		{ 344, 193, 41 },
		{ 344, 194, 41 },
		{ 344, 195, 41 },
		{ 344, 196, 41 },
		{ 344, 197, 41 },
		{ 344, 198, 41 },
		{ 344, 199, 41 },
		{ 344, 200, 41 },
		{ 344, 201, 41 },
		{ 344, 202, 41 },
		{ 344, 203, 41 },
		{ 344, 204, 41 },
		{ 344, 205, 41 },
		{ 344, 206, 41 },
		{ 344, 207, 41 },
		{ 344, 208, 41 },
		{ 344, 209, 41 },
		{ 344, 210, 41 },
		{ 344, 211, 41 },
		{ 344, 212, 41 },
		{ 344, 213, 41 },
		{ 345, 192, 41 },
		{ 345, 193, 41 },
		{ 345, 194, 41 },
		{ 345, 195, 41 },
		{ 345, 196, 41 },
		{ 345, 197, 41 },
		{ 345, 198, 41 },
		{ 345, 199, 41 },
		{ 345, 200, 41 },
		{ 345, 201, 41 },
		{ 345, 202, 41 },
		{ 345, 203, 41 },
		{ 345, 204, 41 },
		{ 345, 205, 41 },
		{ 345, 206, 41 },
		{ 345, 207, 41 },
		{ 345, 208, 41 },
		{ 345, 209, 41 },
		{ 345, 210, 41 },
		{ 345, 211, 41 },
		{ 345, 212, 41 },
		{ 346, 191, 41 },
		{ 346, 192, 41 },
		{ 346, 193, 41 },
		{ 346, 194, 41 },
		{ 346, 195, 41 },
		{ 346, 196, 41 },
		{ 346, 197, 41 },
		{ 346, 198, 41 },
		{ 346, 199, 41 },
		{ 346, 200, 41 },
		{ 346, 201, 41 },
		{ 346, 202, 41 },
		{ 346, 203, 41 },
		{ 346, 204, 41 },
		{ 346, 205, 41 },
		{ 346, 206, 41 },
		{ 346, 207, 41 },
		{ 346, 208, 41 },
		{ 346, 209, 41 },
		{ 346, 210, 41 },
		{ 346, 211, 41 },
		{ 346, 212, 41 },
		{ 347, 190, 41 },
		{ 347, 191, 41 },
		{ 347, 192, 41 },
		{ 347, 193, 41 },
		{ 347, 194, 41 },
		{ 347, 195, 41 },
		{ 347, 196, 41 },
		{ 347, 197, 41 },
		{ 347, 198, 41 },
		{ 347, 199, 41 },
		{ 347, 200, 41 },
		{ 347, 201, 41 },
		{ 347, 202, 41 },
		{ 347, 203, 41 },
		{ 347, 204, 41 },
		{ 347, 205, 41 },
		{ 347, 206, 41 },
		{ 347, 207, 41 },
		{ 347, 208, 41 },
		{ 347, 209, 41 },
		{ 347, 210, 41 },
		{ 347, 211, 41 },
		{ 348, 189, 41 },
		{ 348, 190, 41 },
		{ 348, 191, 41 },
		{ 348, 192, 41 },
		{ 348, 193, 41 },
		{ 348, 194, 41 },
		{ 348, 195, 41 },
		{ 348, 196, 41 },
		{ 348, 197, 41 },
		{ 348, 198, 41 },
		{ 348, 199, 41 },
		{ 348, 200, 41 },
		{ 348, 201, 41 },
		{ 348, 202, 41 },
		{ 348, 203, 41 },
		{ 348, 204, 41 },
		{ 348, 205, 41 },
		{ 348, 206, 41 },
		{ 348, 207, 41 },
		{ 348, 208, 41 },
		{ 348, 209, 41 },
		{ 348, 210, 41 },
		{ 348, 211, 41 },
		{ 349, 189, 41 },
		{ 349, 190, 41 },
		{ 349, 191, 41 },
		{ 349, 192, 41 },
		{ 349, 193, 41 },
		{ 349, 194, 41 },
		{ 349, 195, 41 },
		{ 349, 196, 41 },
		{ 349, 197, 41 },
		{ 349, 198, 41 },
		{ 349, 199, 41 },
		{ 349, 200, 41 },
		{ 349, 201, 41 },
		{ 349, 202, 41 },
		{ 349, 203, 41 },
		{ 349, 204, 41 },
		{ 349, 205, 41 },
		{ 349, 206, 41 },
		{ 349, 207, 41 },
		{ 349, 208, 41 },
		{ 349, 209, 41 },
		{ 349, 210, 41 },
		{ 350, 188, 41 },
		{ 350, 189, 41 },
		{ 350, 190, 41 },
		{ 350, 191, 41 },
		{ 350, 192, 41 },
		{ 350, 193, 41 },
		{ 350, 194, 41 },
		{ 350, 195, 41 },
		{ 350, 196, 41 },
		{ 350, 197, 41 },
		{ 350, 198, 41 },
		{ 350, 199, 41 },
		{ 350, 200, 41 },
		{ 350, 201, 41 },
		{ 350, 202, 41 },
		{ 350, 203, 41 },
		{ 350, 204, 41 },
		{ 350, 205, 41 },
		{ 350, 206, 41 },
		{ 350, 207, 41 },
		{ 350, 208, 41 },
		{ 350, 209, 41 },
		{ 350, 210, 41 },
		{ 351, 187, 41 },
		{ 351, 188, 41 },
		{ 351, 189, 41 },
		{ 351, 190, 41 },
		{ 351, 191, 41 },
		{ 351, 192, 41 },
		{ 351, 193, 41 },
		{ 351, 194, 41 },
		{ 351, 195, 41 },
		{ 351, 196, 41 },
		{ 351, 197, 41 },
		{ 351, 198, 41 },
		{ 351, 199, 41 },
		{ 351, 200, 41 },
		{ 351, 201, 41 },
		{ 351, 202, 41 },
		{ 351, 203, 41 },
		{ 351, 204, 41 },
		{ 351, 205, 41 },
		{ 351, 206, 41 },
		{ 351, 207, 41 },
		{ 351, 208, 41 },
		{ 351, 209, 41 },
		{ 352, 186, 41 },
		{ 352, 187, 41 },
		{ 352, 188, 41 },
		{ 352, 189, 41 },
		{ 352, 190, 41 },
		{ 352, 191, 41 },
		{ 352, 192, 41 },
		{ 352, 193, 41 },
		{ 352, 194, 41 },
		{ 352, 195, 41 },
		{ 352, 196, 41 },
		{ 352, 197, 41 },
		{ 352, 198, 41 },
		{ 352, 199, 41 },
		{ 352, 200, 41 },
		{ 352, 201, 41 },
		{ 352, 202, 41 },
		{ 352, 203, 41 },
		{ 352, 204, 41 },
		{ 352, 205, 41 },
		{ 352, 206, 41 },
		{ 352, 207, 41 },
		{ 352, 208, 41 },
		{ 352, 209, 41 },
		{ 353, 186, 41 },
		{ 353, 187, 41 },
		{ 353, 188, 41 },
		{ 353, 189, 41 },
		{ 353, 190, 41 },
		{ 353, 191, 41 },
		{ 353, 192, 41 },
		{ 353, 193, 41 },
		{ 353, 194, 41 },
		{ 353, 195, 41 },
		{ 353, 196, 41 },
		{ 353, 197, 41 },
		{ 353, 198, 41 },
		{ 353, 199, 41 },
		{ 353, 200, 41 },
		{ 353, 201, 41 },
		{ 353, 202, 41 },
		{ 353, 203, 41 },
		{ 353, 204, 41 },
		{ 353, 205, 41 },
		{ 353, 206, 41 },
		{ 353, 207, 41 },
		{ 353, 208, 41 },
		{ 354, 185, 41 },
		{ 354, 186, 41 },
		{ 354, 187, 41 },
		{ 354, 188, 41 },
		{ 354, 189, 41 },
		{ 354, 190, 41 },
		{ 354, 191, 41 },
		{ 354, 192, 41 },
		{ 354, 193, 41 },
		{ 354, 194, 41 },
		{ 354, 195, 41 },
		{ 354, 196, 41 },
		{ 354, 197, 41 },
		{ 354, 198, 41 },
		{ 354, 199, 41 },
		{ 354, 200, 41 },
		{ 354, 201, 41 },
		{ 354, 202, 41 },
		{ 354, 203, 41 },
		{ 354, 204, 41 },
		{ 354, 205, 41 },
		{ 354, 206, 41 },
		{ 354, 207, 41 },
		{ 354, 208, 41 },
		{ 355, 184, 41 },
		{ 355, 185, 41 },
		{ 355, 186, 41 },
		{ 355, 187, 41 },
		{ 355, 188, 41 },
		{ 355, 189, 41 },
		{ 355, 190, 41 },
		{ 355, 191, 41 },
		{ 355, 192, 41 },
		{ 355, 193, 41 },
		{ 355, 194, 41 },
		{ 355, 195, 41 },
		{ 355, 196, 41 },
		{ 355, 197, 41 },
		{ 355, 198, 41 },
		{ 355, 199, 41 },
		{ 355, 200, 41 },
		{ 355, 201, 41 },
		{ 355, 202, 41 },
		{ 355, 203, 41 },
		{ 355, 204, 41 },
		{ 355, 205, 41 },
		{ 355, 206, 41 },
		{ 355, 207, 41 },
		{ 356, 183, 41 },
		{ 356, 184, 41 },
		{ 356, 185, 41 },
		{ 356, 186, 41 },
		{ 356, 187, 41 },
		{ 356, 188, 41 },
		{ 356, 189, 41 },
		{ 356, 190, 41 },
		{ 356, 191, 41 },
		{ 356, 192, 41 },
		{ 356, 193, 41 },
		{ 356, 194, 41 },
		{ 356, 195, 41 },
		{ 356, 196, 41 },
		{ 356, 197, 41 },
		{ 356, 198, 41 },
		{ 356, 199, 41 },
		{ 356, 200, 41 },
		{ 356, 201, 41 },
		{ 356, 202, 41 },
		{ 356, 203, 41 },
		{ 356, 204, 41 },
		{ 356, 205, 41 },
		{ 356, 206, 41 },
		{ 356, 207, 41 },
		{ 357, 183, 41 },
		{ 357, 184, 41 },
		{ 357, 185, 41 },
		{ 357, 186, 41 },
		{ 357, 187, 41 },
		{ 357, 188, 41 },
		{ 357, 189, 41 },
		{ 357, 190, 41 },
		{ 357, 191, 41 },
		{ 357, 192, 41 },
		{ 357, 193, 41 },
		{ 357, 194, 41 },
		{ 357, 195, 41 },
		{ 357, 196, 41 },
		{ 357, 197, 41 },
		{ 357, 198, 41 },
		{ 357, 199, 41 },
		{ 357, 200, 41 },
		{ 357, 201, 41 },
		{ 357, 202, 41 },
		{ 357, 203, 41 },
		{ 357, 204, 41 },
		{ 357, 205, 41 },
		{ 357, 206, 41 },
		{ 358, 182, 41 },
		{ 358, 183, 41 },
		{ 358, 184, 41 },
		{ 358, 185, 41 },
		{ 358, 186, 41 },
		{ 358, 187, 41 },
		{ 358, 188, 41 },
		{ 358, 189, 41 },
		{ 358, 190, 41 },
		{ 358, 191, 41 },
		{ 358, 192, 41 },
		{ 358, 193, 41 },
		{ 358, 194, 41 },
		{ 358, 195, 41 },
		{ 358, 196, 41 },
		{ 358, 197, 41 },
		{ 358, 198, 41 },
		{ 358, 199, 41 },
		{ 358, 200, 41 },
		{ 358, 201, 41 },
		{ 358, 202, 41 },
		{ 358, 203, 41 },
		{ 358, 204, 41 },
		{ 358, 205, 41 },
		{ 358, 206, 41 },
		{ 359, 181, 41 },
		{ 359, 182, 41 },
		{ 359, 183, 41 },
		{ 359, 184, 41 },
		{ 359, 185, 41 },
		{ 359, 186, 41 },
		{ 359, 187, 41 },
		{ 359, 188, 41 },
		{ 359, 189, 41 },
		{ 359, 190, 41 },
		{ 359, 191, 41 },
		{ 359, 192, 41 },
		{ 359, 193, 41 },
		{ 359, 194, 41 },
		{ 359, 195, 41 },
		{ 359, 196, 41 },
		{ 359, 197, 41 },
		{ 359, 198, 41 },
		{ 359, 199, 41 },
		{ 359, 200, 41 },
		{ 359, 201, 41 },
		{ 359, 202, 41 },
		{ 359, 203, 41 },
		{ 359, 204, 41 },
		{ 359, 205, 41 },
		{ 360, 180, 41 },
		{ 360, 181, 41 },
		{ 360, 182, 41 },
		{ 360, 183, 41 },
		{ 360, 184, 41 },
		{ 360, 185, 41 },
		{ 360, 186, 41 },
		{ 360, 187, 41 },
		{ 360, 188, 41 },
		{ 360, 189, 41 },
		{ 360, 190, 41 },
		{ 360, 191, 41 },
		{ 360, 192, 41 },
		{ 360, 193, 41 },
		{ 360, 194, 41 },
		{ 360, 195, 41 },
		{ 360, 196, 41 },
		{ 360, 197, 41 },
		{ 360, 198, 41 },
		{ 360, 199, 41 },
		{ 360, 200, 41 },
		{ 360, 201, 41 },
		{ 360, 202, 41 },
		{ 360, 203, 41 },
		{ 360, 204, 41 },
		{ 360, 205, 41 },
		{ 361, 180, 41 },
		{ 361, 181, 41 },
		{ 361, 182, 41 },
		{ 361, 183, 41 },
		{ 361, 184, 41 },
		{ 361, 185, 41 },
		{ 361, 186, 41 },
		{ 361, 187, 41 },
		{ 361, 188, 41 },
		{ 361, 189, 41 },
		{ 361, 190, 41 },
		{ 361, 191, 41 },
		{ 361, 192, 41 },
		{ 361, 193, 41 },
		{ 361, 194, 41 },
		{ 361, 195, 41 },
		{ 361, 196, 41 },
		{ 361, 197, 41 },
		{ 361, 198, 41 },
		{ 361, 199, 41 },
		{ 361, 200, 41 },
		{ 361, 201, 41 },
		{ 361, 202, 41 },
		{ 361, 203, 41 },
		{ 361, 204, 41 },
		{ 362, 179, 41 },
		{ 362, 180, 41 },
		{ 362, 181, 41 },
		{ 362, 182, 41 },
		{ 362, 183, 41 },
		{ 362, 184, 41 },
		{ 362, 185, 41 },
		{ 362, 186, 41 },
		{ 362, 187, 41 },
		{ 362, 188, 41 },
		{ 362, 189, 41 },
		{ 362, 190, 41 },
		{ 362, 191, 41 },
		{ 362, 192, 41 },
		{ 362, 193, 41 },
		{ 362, 194, 41 },
		{ 362, 195, 41 },
		{ 362, 196, 41 },
		{ 362, 197, 41 },
		{ 362, 198, 41 },
		{ 362, 199, 41 },
		{ 362, 200, 41 },
		{ 362, 201, 41 },
		{ 362, 202, 41 },
		{ 362, 203, 41 },
		{ 362, 204, 41 },
		{ 363, 178, 41 },
		{ 363, 179, 41 },
		{ 363, 180, 41 },
		{ 363, 181, 41 },
		{ 363, 182, 41 },
		{ 363, 183, 41 },
		{ 363, 184, 41 },
		{ 363, 185, 41 },
		{ 363, 186, 41 },
		{ 363, 187, 41 },
		{ 363, 188, 41 },
		{ 363, 189, 41 },
		{ 363, 190, 41 },
		{ 363, 191, 41 },
		{ 363, 192, 41 },
		{ 363, 193, 41 },
		{ 363, 194, 41 },
		{ 363, 195, 41 },
		{ 363, 196, 41 },
		{ 363, 197, 41 },
		{ 363, 198, 41 },
		{ 363, 199, 41 },
		{ 363, 200, 41 },
		{ 363, 201, 41 },
		{ 363, 202, 41 },
		{ 363, 203, 41 },
		{ 364, 177, 41 },
		{ 364, 178, 41 },
		{ 364, 179, 41 },
		{ 364, 180, 41 },
		{ 364, 181, 41 },
		{ 364, 182, 41 },
		{ 364, 183, 41 },
		{ 364, 184, 41 },
		{ 364, 185, 41 },
		{ 364, 186, 41 },
		{ 364, 187, 41 },
		{ 364, 188, 41 },
		{ 364, 189, 41 },
		{ 364, 190, 41 },
		{ 364, 191, 41 },
		{ 364, 192, 41 },
		{ 364, 193, 41 },
		{ 364, 194, 41 },
		{ 364, 195, 41 },
		{ 364, 196, 41 },
		{ 364, 197, 41 },
		{ 364, 198, 41 },
		{ 364, 199, 41 },
		{ 364, 200, 41 },
		{ 364, 201, 41 },
		{ 364, 202, 41 },
		{ 364, 203, 41 },
		{ 365, 177, 41 },
		{ 365, 178, 41 },
		{ 365, 179, 41 },
		{ 365, 180, 41 },
		{ 365, 181, 41 },
		{ 365, 182, 41 },
		{ 365, 183, 41 },
		{ 365, 184, 41 },
		{ 365, 185, 41 },
		{ 365, 186, 41 },
		{ 365, 187, 41 },
		{ 365, 188, 41 },
		{ 365, 189, 41 },
		{ 365, 190, 41 },
		{ 365, 191, 41 },
		{ 365, 192, 41 },
		{ 365, 193, 41 },
		{ 365, 194, 41 },
		{ 365, 195, 41 },
		{ 365, 196, 41 },
		{ 365, 197, 41 },
		{ 365, 198, 41 },
		{ 365, 199, 41 },
		{ 365, 200, 41 },
		{ 365, 201, 41 },
		{ 365, 202, 41 },
		{ 366, 176, 41 },
		{ 366, 177, 41 },
		{ 366, 178, 41 },
		{ 366, 179, 41 },
		{ 366, 180, 41 },
		{ 366, 181, 41 },
		{ 366, 182, 41 },
		{ 366, 183, 41 },
		{ 366, 184, 41 },
		{ 366, 185, 41 },
		{ 366, 186, 41 },
		{ 366, 187, 41 },
		{ 366, 188, 41 },
		{ 366, 189, 41 },
		{ 366, 190, 41 },
		{ 366, 191, 41 },
		{ 366, 192, 41 },
		{ 366, 193, 41 },
		{ 366, 194, 41 },
		{ 366, 195, 41 },
		{ 366, 196, 41 },
		{ 366, 197, 41 },
		{ 366, 198, 41 },
		{ 366, 199, 41 },
		{ 366, 200, 41 },
		{ 366, 201, 41 },
		{ 366, 202, 41 },
		{ 367, 175, 41 },
		{ 367, 176, 41 },
		{ 367, 177, 41 },
		{ 367, 178, 41 },
		{ 367, 179, 41 },
		{ 367, 180, 41 },
		{ 367, 181, 41 },
		{ 367, 182, 41 },
		{ 367, 183, 41 },
		{ 367, 184, 41 },
		{ 367, 185, 41 },
		{ 367, 186, 41 },
		{ 367, 187, 41 },
		{ 367, 188, 41 },
		{ 367, 189, 41 },
		{ 367, 190, 41 },
		{ 367, 191, 41 },
		{ 367, 192, 41 },
		{ 367, 193, 41 },
		{ 367, 194, 41 },
		{ 367, 195, 41 },
		{ 367, 196, 41 },
		{ 367, 197, 41 },
		{ 367, 198, 41 },
		{ 367, 199, 41 },
		{ 367, 200, 41 },
		{ 367, 201, 41 },
		{ 368, 174, 41 },
		{ 368, 175, 41 },
		{ 368, 176, 41 },
		{ 368, 177, 41 },
		{ 368, 178, 41 },
		{ 368, 179, 41 },
		{ 368, 180, 41 },
		{ 368, 181, 41 },
		{ 368, 182, 41 },
		{ 368, 183, 41 },
		{ 368, 184, 41 },
		{ 368, 185, 41 },
		{ 368, 186, 41 },
		{ 368, 187, 41 },
		{ 368, 188, 41 },
		{ 368, 189, 41 },
		{ 368, 190, 41 },
		{ 368, 191, 41 },
		{ 368, 192, 41 },
		{ 368, 193, 41 },
		{ 368, 194, 41 },
		{ 368, 195, 41 },
		{ 368, 196, 41 },
		{ 368, 197, 41 },
		{ 368, 198, 41 },
		{ 368, 199, 41 },
		{ 368, 200, 41 },
		{ 368, 201, 41 },
		{ 369, 174, 41 },
		{ 369, 175, 41 },
		{ 369, 176, 41 },
		{ 369, 177, 41 },
		{ 369, 178, 41 },
		{ 369, 179, 41 },
		{ 369, 180, 41 },
		{ 369, 181, 41 },
		{ 369, 182, 41 },
		{ 369, 183, 41 },
		{ 369, 184, 41 },
		{ 369, 185, 41 },
		{ 369, 186, 41 },
		{ 369, 187, 41 },
		{ 369, 188, 41 },
		{ 369, 189, 41 },
		{ 369, 190, 41 },
		{ 369, 191, 41 },
		{ 369, 192, 41 },
		{ 369, 193, 41 },
		{ 369, 194, 41 },
		{ 369, 195, 41 },
		{ 369, 196, 41 },
		{ 369, 197, 41 },
		{ 369, 198, 41 },
		{ 369, 199, 41 },
		{ 369, 200, 41 },
		{ 370, 173, 41 },
		{ 370, 174, 41 },
		{ 370, 175, 41 },
		{ 370, 176, 41 },
		{ 370, 177, 41 },
		{ 370, 178, 41 },
		{ 370, 179, 41 },
		{ 370, 180, 41 },
		{ 370, 181, 41 },
		{ 370, 182, 41 },
		{ 370, 183, 41 },
		{ 370, 184, 41 },
		{ 370, 185, 41 },
		{ 370, 186, 41 },
		{ 370, 187, 41 },
		{ 370, 188, 41 },
		{ 370, 189, 41 },
		{ 370, 190, 41 },
		{ 370, 191, 41 },
		{ 370, 192, 41 },
		{ 370, 193, 41 },
		{ 370, 194, 41 },
		{ 370, 195, 41 },
		{ 370, 196, 41 },
		{ 370, 197, 41 },
		{ 370, 198, 41 },
		{ 370, 199, 41 },
		{ 370, 200, 41 },
		{ 371, 172, 41 },
		{ 371, 173, 41 },
		{ 371, 174, 41 },
		{ 371, 175, 41 },
		{ 371, 176, 41 },
		{ 371, 177, 41 },
		{ 371, 178, 41 },
		{ 371, 179, 41 },
		{ 371, 180, 41 },
		{ 371, 181, 41 },
		{ 371, 182, 41 },
		{ 371, 183, 41 },
		{ 371, 184, 41 },
		{ 371, 185, 41 },
		{ 371, 186, 41 },
		{ 371, 187, 41 },
		{ 371, 188, 41 },
		{ 371, 189, 41 },
		{ 371, 190, 41 },
		{ 371, 191, 41 },
		{ 371, 192, 41 },
		{ 371, 193, 41 },
		{ 371, 194, 41 },
		{ 371, 195, 41 },
		{ 371, 196, 41 },
		{ 371, 197, 41 },
		{ 371, 198, 41 },
		{ 371, 199, 41 },
		{ 372, 171, 41 },
		{ 372, 172, 41 },
		{ 372, 173, 41 },
		{ 372, 174, 41 },
		{ 372, 175, 41 },
		{ 372, 176, 41 },
		{ 372, 177, 41 },
		{ 372, 178, 41 },
		{ 372, 179, 41 },
		{ 372, 180, 41 },
		{ 372, 181, 41 },
		{ 372, 182, 41 },
		{ 372, 183, 41 },
		{ 372, 184, 41 },
		{ 372, 185, 41 },
		{ 372, 186, 41 },
		{ 372, 187, 41 },
		{ 372, 188, 41 },
		{ 372, 189, 41 },
		{ 372, 190, 41 },
		{ 372, 191, 41 },
		{ 372, 192, 41 },
		{ 372, 193, 41 },
		{ 372, 194, 41 },
		{ 372, 195, 41 },
		{ 372, 196, 41 },
		{ 372, 197, 41 },
		{ 372, 198, 41 },
		{ 372, 199, 41 },
		{ 373, 171, 41 },
		{ 373, 172, 41 },
		{ 373, 173, 41 },
		{ 373, 174, 41 },
		{ 373, 175, 41 },
		{ 373, 176, 41 },
		{ 373, 177, 41 },
		{ 373, 178, 41 },
		{ 373, 179, 41 },
		{ 373, 180, 41 },
		{ 373, 181, 41 },
		{ 373, 182, 41 },
		{ 373, 183, 41 },
		{ 373, 184, 41 },
		{ 373, 185, 41 },
		{ 373, 186, 41 },
		{ 373, 187, 41 },
		{ 373, 188, 41 },
		{ 373, 189, 41 },
		{ 373, 190, 41 },
		{ 373, 191, 41 },
		{ 373, 192, 41 },
		{ 373, 193, 41 },
		{ 373, 194, 41 },
		{ 373, 195, 41 },
		{ 373, 196, 41 },
		{ 373, 197, 41 },
		{ 373, 198, 41 },
		{ 374, 170, 41 },
		{ 374, 171, 41 },
		{ 374, 172, 41 },
		{ 374, 173, 41 },
		{ 374, 174, 41 },
		{ 374, 175, 41 },
		{ 374, 176, 41 },
		{ 374, 177, 41 },
		{ 374, 178, 41 },
		{ 374, 179, 41 },
		{ 374, 180, 41 },
		{ 374, 181, 41 },
		{ 374, 182, 41 },
		{ 374, 183, 41 },
		{ 374, 184, 41 },
		{ 374, 185, 41 },
		{ 374, 186, 41 },
		{ 374, 187, 41 },
		{ 374, 188, 41 },
		{ 374, 189, 41 },
		{ 374, 190, 41 },
		{ 374, 191, 41 },
		{ 374, 192, 41 },
		{ 374, 193, 41 },
		{ 374, 194, 41 },
		{ 374, 195, 41 },
		{ 374, 196, 41 },
		{ 374, 197, 41 },
		{ 374, 198, 41 },
		{ 375, 169, 41 },
		{ 375, 170, 41 },
		{ 375, 171, 41 },
		{ 375, 172, 41 },
		{ 375, 173, 41 },
		{ 375, 174, 41 },
		{ 375, 175, 41 },
		{ 375, 176, 41 },
		{ 375, 177, 41 },
		{ 375, 178, 41 },
		{ 375, 179, 41 },
		{ 375, 180, 41 },
		{ 375, 181, 41 },
		{ 375, 182, 41 },
		{ 375, 183, 41 },
		{ 375, 184, 41 },
		{ 375, 185, 41 },
		{ 375, 186, 41 },
		{ 375, 187, 41 },
		{ 375, 188, 41 },
		{ 375, 189, 41 },
		{ 375, 190, 41 },
		{ 375, 191, 41 },
		{ 375, 192, 41 },
		{ 375, 193, 41 },
		{ 375, 194, 41 },
		{ 375, 195, 41 },
		{ 375, 196, 41 },
		{ 375, 197, 41 },
		{ 376, 168, 41 },
		{ 376, 169, 41 },
		{ 376, 170, 41 },
		{ 376, 171, 41 },
		{ 376, 172, 41 },
		{ 376, 173, 41 },
		{ 376, 174, 41 },
		{ 376, 175, 41 },
		{ 376, 176, 41 },
		{ 376, 177, 41 },
		{ 376, 178, 41 },
		{ 376, 179, 41 },
		{ 376, 180, 41 },
		{ 376, 181, 41 },
		{ 376, 182, 41 },
		{ 376, 183, 41 },
		{ 376, 184, 41 },
		{ 376, 185, 41 },
		{ 376, 186, 41 },
		{ 376, 187, 41 },
		{ 376, 188, 41 },
		{ 376, 189, 41 },
		{ 376, 190, 41 },
		{ 376, 191, 41 },
		{ 376, 192, 41 },
		{ 376, 193, 41 },
		{ 376, 194, 41 },
		{ 376, 195, 41 },
		{ 376, 196, 41 },
		{ 376, 197, 41 },
		{ 377, 168, 41 },
		{ 377, 169, 41 },
		{ 377, 170, 41 },
		{ 377, 171, 41 },
		{ 377, 172, 41 },
		{ 377, 173, 41 },
		{ 377, 174, 41 },
		{ 377, 175, 41 },
		{ 377, 176, 41 },
		{ 377, 177, 41 },
		{ 377, 178, 41 },
		{ 377, 179, 41 },
		{ 377, 180, 41 },
		{ 377, 181, 41 },
		{ 377, 182, 41 },
		{ 377, 183, 41 },
		{ 377, 184, 41 },
		{ 377, 185, 41 },
		{ 377, 186, 41 },
		{ 377, 187, 41 },
		{ 377, 188, 41 },
		{ 377, 189, 41 },
		{ 377, 190, 41 },
		{ 377, 191, 41 },
		{ 377, 192, 41 },
		{ 377, 193, 41 },
		{ 377, 194, 41 },
		{ 377, 195, 41 },
		{ 377, 196, 41 },
		{ 378, 167, 41 },
		{ 378, 168, 41 },
		{ 378, 169, 41 },
		{ 378, 170, 41 },
		{ 378, 171, 41 },
		{ 378, 172, 41 },
		{ 378, 173, 41 },
		{ 378, 174, 41 },
		{ 378, 175, 41 },
		{ 378, 176, 41 },
		{ 378, 177, 41 },
		{ 378, 178, 41 },
		{ 378, 179, 41 },
		{ 378, 180, 41 },
		{ 378, 181, 41 },
		{ 378, 182, 41 },
		{ 378, 183, 41 },
		{ 378, 184, 41 },
		{ 378, 185, 41 },
		{ 378, 186, 41 },
		{ 378, 187, 41 },
		{ 378, 188, 41 },
		{ 378, 189, 41 },
		{ 378, 190, 41 },
		{ 378, 191, 41 },
		{ 378, 192, 41 },
		{ 378, 193, 41 },
		{ 378, 194, 41 },
		{ 378, 195, 41 },
		{ 378, 196, 41 },
		{ 379, 166, 41 },
		{ 379, 167, 41 },
		{ 379, 168, 41 },
		{ 379, 169, 41 },
		{ 379, 170, 41 },
		{ 379, 171, 41 },
		{ 379, 172, 41 },
		{ 379, 173, 41 },
		{ 379, 174, 41 },
		{ 379, 175, 41 },
		{ 379, 176, 41 },
		{ 379, 177, 41 },
		{ 379, 178, 41 },
		{ 379, 179, 41 },
		{ 379, 180, 41 },
		{ 379, 181, 41 },
		{ 379, 182, 41 },
		{ 379, 183, 41 },
		{ 379, 184, 41 },
		{ 379, 185, 41 },
		{ 379, 186, 41 },
		{ 379, 187, 41 },
		{ 379, 188, 41 },
		{ 379, 189, 41 },
		{ 379, 190, 41 },
		{ 379, 191, 41 },
		{ 379, 192, 41 },
		{ 379, 193, 41 },
		{ 379, 194, 41 },
		{ 379, 195, 41 },
		{ 380, 165, 41 },
		{ 380, 166, 41 },
		{ 380, 167, 41 },
		{ 380, 168, 41 },
		{ 380, 169, 41 },
		{ 380, 170, 41 },
		{ 380, 171, 41 },
		{ 380, 172, 41 },
		{ 380, 173, 41 },
		{ 380, 174, 41 },
		{ 380, 175, 41 },
		{ 380, 176, 41 },
		{ 380, 177, 41 },
		{ 380, 178, 41 },
		{ 380, 179, 41 },
		{ 380, 180, 41 },
		{ 380, 181, 41 },
		{ 380, 182, 41 },
		{ 380, 183, 41 },
		{ 380, 184, 41 },
		{ 380, 185, 41 },
		{ 380, 186, 41 },
		{ 380, 187, 41 },
		{ 380, 188, 41 },
		{ 380, 189, 41 },
		{ 380, 190, 41 },
		{ 380, 191, 41 },
		{ 380, 192, 41 },
		{ 380, 193, 41 },
		{ 380, 194, 41 },
		{ 380, 195, 41 },
		{ 381, 165, 41 },
		{ 381, 166, 41 },
		{ 381, 167, 41 },
		{ 381, 168, 41 },
		{ 381, 169, 41 },
		{ 381, 170, 41 },
		{ 381, 171, 41 },
		{ 381, 172, 41 },
		{ 381, 173, 41 },
		{ 381, 174, 41 },
		{ 381, 175, 41 },
		{ 381, 176, 41 },
		{ 381, 177, 41 },
		{ 381, 178, 41 },
		{ 381, 179, 41 },
		{ 381, 180, 41 },
		{ 381, 181, 41 },
		{ 381, 182, 41 },
		{ 381, 183, 41 },
		{ 381, 184, 41 },
		{ 381, 185, 41 },
		{ 381, 186, 41 },
		{ 381, 187, 41 },
		{ 381, 188, 41 },
		{ 381, 189, 41 },
		{ 381, 190, 41 },
		{ 381, 191, 41 },
		{ 381, 192, 41 },
		{ 381, 193, 41 },
		{ 381, 194, 41 },
		{ 382, 164, 41 },
		{ 382, 165, 41 },
		{ 382, 166, 41 },
		{ 382, 167, 41 },
		{ 382, 168, 41 },
		{ 382, 169, 41 },
		{ 382, 170, 41 },
		{ 382, 171, 41 },
		{ 382, 172, 41 },
		{ 382, 173, 41 },
		{ 382, 174, 41 },
		{ 382, 175, 41 },
		{ 382, 176, 41 },
		{ 382, 177, 41 },
		{ 382, 178, 41 },
		{ 382, 179, 41 },
		{ 382, 180, 41 },
		{ 382, 181, 41 },
		{ 382, 182, 41 },
		{ 382, 183, 41 },
		{ 382, 184, 41 },
		{ 382, 185, 41 },
		{ 382, 186, 41 },
		{ 382, 187, 41 },
		{ 382, 188, 41 },
		{ 382, 189, 41 },
		{ 382, 190, 41 },
		{ 382, 191, 41 },
		{ 382, 192, 41 },
		{ 382, 193, 41 },
		{ 382, 194, 41 },
		{ 383, 163, 41 },
		{ 383, 164, 41 },
		{ 383, 165, 41 },
		{ 383, 166, 41 },
		{ 383, 167, 41 },
		{ 383, 168, 41 },
		{ 383, 169, 41 },
		{ 383, 170, 41 },
		{ 383, 171, 41 },
		{ 383, 172, 41 },
		{ 383, 173, 41 },
		{ 383, 174, 41 },
		{ 383, 175, 41 },
		{ 383, 176, 41 },
		{ 383, 177, 41 },
		{ 383, 178, 41 },
		{ 383, 179, 41 },
		{ 383, 180, 41 },
		{ 383, 181, 41 },
		{ 383, 182, 41 },
		{ 383, 183, 41 },
		{ 383, 184, 41 },
		{ 383, 185, 41 },
		{ 383, 186, 41 },
		{ 383, 187, 41 },
		{ 383, 188, 41 },
		{ 383, 189, 41 },
		{ 383, 190, 41 },
		{ 383, 191, 41 },
		{ 383, 192, 41 },
		{ 383, 193, 41 },
		{ 384, 162, 41 },
		{ 384, 163, 41 },
		{ 384, 164, 41 },
		{ 384, 165, 41 },
		{ 384, 166, 41 },
		{ 384, 167, 41 },
		{ 384, 168, 41 },
		{ 384, 169, 41 },
		{ 384, 170, 41 },
		{ 384, 171, 41 },
		{ 384, 172, 41 },
		{ 384, 173, 41 },
		{ 384, 174, 41 },
		{ 384, 175, 41 },
		{ 384, 176, 41 },
		{ 384, 177, 41 },
		{ 384, 178, 41 },
		{ 384, 179, 41 },
		{ 384, 180, 41 },
		{ 384, 181, 41 },
		{ 384, 182, 41 },
		{ 384, 183, 41 },
		{ 384, 184, 41 },
		{ 384, 185, 41 },
		{ 384, 186, 41 },
		{ 384, 187, 41 },
		{ 384, 188, 41 },
		{ 384, 189, 41 },
		{ 384, 190, 41 },
		{ 384, 191, 41 },
		{ 384, 192, 41 },
		{ 384, 193, 41 },
		{ 385, 162, 41 },
		{ 385, 163, 41 },
		{ 385, 164, 41 },
		{ 385, 165, 41 },
		{ 385, 166, 41 },
		{ 385, 167, 41 },
		{ 385, 168, 41 },
		{ 385, 169, 41 },
		{ 385, 170, 41 },
		{ 385, 171, 41 },
		{ 385, 172, 41 },
		{ 385, 173, 41 },
		{ 385, 174, 41 },
		{ 385, 175, 41 },
		{ 385, 176, 41 },
		{ 385, 177, 41 },
		{ 385, 178, 41 },
		{ 385, 179, 41 },
		{ 385, 180, 41 },
		{ 385, 181, 41 },
		{ 385, 182, 41 },
		{ 385, 183, 41 },
		{ 385, 184, 41 },
		{ 385, 185, 41 },
		{ 385, 186, 41 },
		{ 385, 187, 41 },
		{ 385, 188, 41 },
		{ 385, 189, 41 },
		{ 385, 190, 41 },
		{ 385, 191, 41 },
		{ 385, 192, 41 },
		{ 386, 161, 41 },
		{ 386, 162, 41 },
		{ 386, 163, 41 },
		{ 386, 164, 41 },
		{ 386, 165, 41 },
		{ 386, 166, 41 },
		{ 386, 167, 41 },
		{ 386, 168, 41 },
		{ 386, 169, 41 },
		{ 386, 170, 41 },
		{ 386, 171, 41 },
		{ 386, 172, 41 },
		{ 386, 173, 41 },
		{ 386, 174, 41 },
		{ 386, 175, 41 },
		{ 386, 176, 41 },
		{ 386, 177, 41 },
		{ 386, 178, 41 },
		{ 386, 179, 41 },
		{ 386, 180, 41 },
		{ 386, 181, 41 },
		{ 386, 182, 41 },
		{ 386, 183, 41 },
		{ 386, 184, 41 },
		{ 386, 185, 41 },
		{ 386, 186, 41 },
		{ 386, 187, 41 },
		{ 386, 188, 41 },
		{ 386, 189, 41 },
		{ 386, 190, 41 },
		{ 386, 191, 41 },
		{ 386, 192, 41 },
		{ 387, 160, 41 },
		{ 387, 161, 41 },
		{ 387, 162, 41 },
		{ 387, 163, 41 },
		{ 387, 164, 41 },
		{ 387, 165, 41 },
		{ 387, 166, 41 },
		{ 387, 167, 41 },
		{ 387, 168, 41 },
		{ 387, 169, 41 },
		{ 387, 170, 41 },
		{ 387, 171, 41 },
		{ 387, 172, 41 },
		{ 387, 173, 41 },
		{ 387, 174, 41 },
		{ 387, 175, 41 },
		{ 387, 176, 41 },
		{ 387, 177, 41 },
		{ 387, 178, 41 },
		{ 387, 179, 41 },
		{ 387, 180, 41 },
		{ 387, 181, 41 },
		{ 387, 182, 41 },
		{ 387, 183, 41 },
		{ 387, 184, 41 },
		{ 387, 185, 41 },
		{ 387, 186, 41 },
		{ 387, 187, 41 },
		{ 387, 188, 41 },
		{ 387, 189, 41 },
		{ 387, 190, 41 },
		{ 387, 191, 41 },
		{ 388, 159, 41 },
		{ 388, 160, 41 },
		{ 388, 161, 41 },
		{ 388, 162, 41 },
		{ 388, 163, 41 },
		{ 388, 164, 41 },
		{ 388, 165, 41 },
		{ 388, 166, 41 },
		{ 388, 167, 41 },
		{ 388, 168, 41 },
		{ 388, 169, 41 },
		{ 388, 170, 41 },
		{ 388, 171, 41 },
		{ 388, 172, 41 },
		{ 388, 173, 41 },
		{ 388, 174, 41 },
		{ 388, 175, 41 },
		{ 388, 176, 41 },
		{ 388, 177, 41 },
		{ 388, 178, 41 },
		{ 388, 179, 41 },
		{ 388, 180, 41 },
		{ 388, 181, 41 },
		{ 388, 182, 41 },
		{ 388, 183, 41 },
		{ 388, 184, 41 },
		{ 388, 185, 41 },
		{ 388, 186, 41 },
		{ 388, 187, 41 },
		{ 388, 188, 41 },
		{ 388, 189, 41 },
		{ 388, 190, 41 },
		{ 388, 191, 41 },
		{ 389, 159, 41 },
		{ 389, 160, 41 },
		{ 389, 161, 41 },
		{ 389, 162, 41 },
		{ 389, 163, 41 },
		{ 389, 164, 41 },
		{ 389, 165, 41 },
		{ 389, 166, 41 },
		{ 389, 167, 41 },
		{ 389, 168, 41 },
		{ 389, 169, 41 },
		{ 389, 170, 41 },
		{ 389, 171, 41 },
		{ 389, 172, 41 },
		{ 389, 173, 41 },
		{ 389, 174, 41 },
		{ 389, 175, 41 },
		{ 389, 176, 41 },
		{ 389, 177, 41 },
		{ 389, 178, 41 },
		{ 389, 179, 41 },
		{ 389, 180, 41 },
		{ 389, 181, 41 },
		{ 389, 182, 41 },
		{ 389, 183, 41 },
		{ 389, 184, 41 },
		{ 389, 185, 41 },
		{ 389, 186, 41 },
		{ 389, 187, 41 },
		{ 389, 188, 41 },
		{ 389, 189, 41 },
		{ 389, 190, 41 },
		{ 390, 158, 41 },
		{ 390, 159, 41 },
		{ 390, 160, 41 },
		{ 390, 161, 41 },
		{ 390, 162, 41 },
		{ 390, 163, 41 },
		{ 390, 164, 41 },
		{ 390, 165, 41 },
		{ 390, 166, 41 },
		{ 390, 167, 41 },
		{ 390, 168, 41 },
		{ 390, 169, 41 },
		{ 390, 170, 41 },
		{ 390, 171, 41 },
		{ 390, 172, 41 },
		{ 390, 173, 41 },
		{ 390, 174, 41 },
		{ 390, 175, 41 },
		{ 390, 176, 41 },
		{ 390, 177, 41 },
		{ 390, 178, 41 },
		{ 390, 179, 41 },
		{ 390, 180, 41 },
		{ 390, 181, 41 },
		{ 390, 182, 41 },
		{ 390, 183, 41 },
		{ 390, 184, 41 },
		{ 390, 185, 41 },
		{ 390, 186, 41 },
		{ 390, 187, 41 },
		{ 390, 188, 41 },
		{ 390, 189, 41 },
		{ 390, 190, 41 },
		{ 391, 157, 41 },
		{ 391, 158, 41 },
		{ 391, 159, 41 },
		{ 391, 160, 41 },
		{ 391, 161, 41 },
		{ 391, 162, 41 },
		{ 391, 163, 41 },
		{ 391, 164, 41 },
		{ 391, 165, 41 },
		{ 391, 166, 41 },
		{ 391, 167, 41 },
		{ 391, 168, 41 },
		{ 391, 169, 41 },
		{ 391, 170, 41 },
		{ 391, 171, 41 },
		{ 391, 172, 41 },
		{ 391, 173, 41 },
		{ 391, 174, 41 },
		{ 391, 175, 41 },
		{ 391, 176, 41 },
		{ 391, 177, 41 },
		{ 391, 178, 41 },
		{ 391, 179, 41 },
		{ 391, 180, 41 },
		{ 391, 181, 41 },
		{ 391, 182, 41 },
		{ 391, 183, 41 },
		{ 391, 184, 41 },
		{ 391, 185, 41 },
		{ 391, 186, 41 },
		{ 391, 187, 41 },
		{ 391, 188, 41 },
		{ 391, 189, 41 },
		{ 392, 156, 41 },
		{ 392, 157, 41 },
		{ 392, 158, 41 },
		{ 392, 159, 41 },
		{ 392, 160, 41 },
		{ 392, 161, 41 },
		{ 392, 162, 41 },
		{ 392, 163, 41 },
		{ 392, 164, 41 },
		{ 392, 165, 41 },
		{ 392, 166, 41 },
		{ 392, 167, 41 },
		{ 392, 168, 41 },
		{ 392, 169, 41 },
		{ 392, 170, 41 },
		{ 392, 171, 41 },
		{ 392, 172, 41 },
		{ 392, 173, 41 },
		{ 392, 174, 41 },
		{ 392, 175, 41 },
		{ 392, 176, 41 },
		{ 392, 177, 41 },
		{ 392, 178, 41 },
		{ 392, 179, 41 },
		{ 392, 180, 41 },
		{ 392, 181, 41 },
		{ 392, 182, 41 },
		{ 392, 183, 41 },
		{ 392, 184, 41 },
		{ 392, 185, 41 },
		{ 392, 186, 41 },
		{ 392, 187, 41 },
		{ 392, 188, 41 },
		{ 392, 189, 41 },
		{ 393, 156, 41 },
		{ 393, 157, 41 },
		{ 393, 158, 41 },
		{ 393, 159, 41 },
		{ 393, 160, 41 },
		{ 393, 161, 41 },
		{ 393, 162, 41 },
		{ 393, 163, 41 },
		{ 393, 164, 41 },
		{ 393, 165, 41 },
		{ 393, 166, 41 },
		{ 393, 167, 41 },
		{ 393, 168, 41 },
		{ 393, 169, 41 },
		{ 393, 170, 41 },
		{ 393, 171, 41 },
		{ 393, 172, 41 },
		{ 393, 173, 41 },
		{ 393, 174, 41 },
		{ 393, 175, 41 },
		{ 393, 176, 41 },
		{ 393, 177, 41 },
		{ 393, 178, 41 },
		{ 393, 179, 41 },
		{ 393, 180, 41 },
		{ 393, 181, 41 },
		{ 393, 182, 41 },
		{ 393, 183, 41 },
		{ 393, 184, 41 },
		{ 393, 185, 41 },
		{ 393, 186, 41 },
		{ 393, 187, 41 },
		{ 393, 188, 41 },
		{ 394, 155, 41 },
		{ 394, 156, 41 },
		{ 394, 157, 41 },
		{ 394, 158, 41 },
		{ 394, 159, 41 },
		{ 394, 160, 41 },
		{ 394, 161, 41 },
		{ 394, 162, 41 },
		{ 394, 163, 41 },
		{ 394, 164, 41 },
		{ 394, 165, 41 },
		{ 394, 166, 41 },
		{ 394, 167, 41 },
		{ 394, 168, 41 },
		{ 394, 169, 41 },
		{ 394, 170, 41 },
		{ 394, 171, 41 },
		{ 394, 172, 41 },
		{ 394, 173, 41 },
		{ 394, 174, 41 },
		{ 394, 175, 41 },
		{ 394, 176, 41 },
		{ 394, 177, 41 },
		{ 394, 178, 41 },
		{ 394, 179, 41 },
		{ 394, 180, 41 },
		{ 394, 181, 41 },
		{ 394, 182, 41 },
		{ 394, 183, 41 },
		{ 394, 184, 41 },
		{ 394, 185, 41 },
		{ 394, 186, 41 },
		{ 394, 187, 41 },
		{ 394, 188, 41 },
		{ 395, 154, 41 },
		{ 395, 155, 41 },
		{ 395, 156, 41 },
		{ 395, 157, 41 },
		{ 395, 158, 41 },
		{ 395, 159, 41 },
		{ 395, 160, 41 },
		{ 395, 161, 41 },
		{ 395, 162, 41 },
		{ 395, 163, 41 },
		{ 395, 164, 41 },
		{ 395, 165, 41 },
		{ 395, 166, 41 },
		{ 395, 167, 41 },
		{ 395, 168, 41 },
		{ 395, 169, 41 },
		{ 395, 170, 41 },
		{ 395, 171, 41 },
		{ 395, 172, 41 },
		{ 395, 173, 41 },
		{ 395, 174, 41 },
		{ 395, 175, 41 },
		{ 395, 176, 41 },
		{ 395, 177, 41 },
		{ 395, 178, 41 },
		{ 395, 179, 41 },
		{ 395, 180, 41 },
		{ 395, 181, 41 },
		{ 395, 182, 41 },
		{ 395, 183, 41 },
		{ 395, 184, 41 },
		{ 395, 185, 41 },
		{ 395, 186, 41 },
		{ 395, 187, 41 },
		{ 396, 153, 41 },
		{ 396, 154, 41 },
		{ 396, 155, 41 },
		{ 396, 156, 41 },
		{ 396, 157, 41 },
		{ 396, 158, 41 },
		{ 396, 159, 41 },
		{ 396, 160, 41 },
		{ 396, 161, 41 },
		{ 396, 162, 41 },
		{ 396, 163, 41 },
		{ 396, 164, 41 },
		{ 396, 165, 41 },
		{ 396, 166, 41 },
		{ 396, 167, 41 },
		{ 396, 168, 41 },
		{ 396, 169, 41 },
		{ 396, 170, 41 },
		{ 396, 171, 41 },
		{ 396, 172, 41 },
		{ 396, 173, 41 },
		{ 396, 174, 41 },
		{ 396, 175, 41 },
		{ 396, 176, 41 },
		{ 396, 177, 41 },
		{ 396, 178, 41 },
		{ 396, 179, 41 },
		{ 396, 180, 41 },
		{ 396, 181, 41 },
		{ 396, 182, 41 },
		{ 396, 183, 41 },
		{ 396, 184, 41 },
		{ 396, 185, 41 },
		{ 396, 186, 41 },
		{ 396, 187, 41 },
		{ 397, 153, 41 },
		{ 397, 154, 41 },
		{ 397, 155, 41 },
		{ 397, 156, 41 },
		{ 397, 157, 41 },
		{ 397, 158, 41 },
		{ 397, 159, 41 },
		{ 397, 160, 41 },
		{ 397, 161, 41 },
		{ 397, 162, 41 },
		{ 397, 163, 41 },
		{ 397, 164, 41 },
		{ 397, 165, 41 },
		{ 397, 166, 41 },
		{ 397, 167, 41 },
		{ 397, 168, 41 },
		{ 397, 169, 41 },
		{ 397, 170, 41 },
		{ 397, 171, 41 },
		{ 397, 172, 41 },
		{ 397, 173, 41 },
		{ 397, 174, 41 },
		{ 397, 175, 41 },
		{ 397, 176, 41 },
		{ 397, 177, 41 },
		{ 397, 178, 41 },
		{ 397, 179, 41 },
		{ 397, 180, 41 },
		{ 397, 181, 41 },
		{ 397, 182, 41 },
		{ 397, 183, 41 },
		{ 397, 184, 41 },
		{ 397, 185, 41 },
		{ 397, 186, 41 },
		{ 398, 152, 41 },
		{ 398, 153, 41 },
		{ 398, 154, 41 },
		{ 398, 155, 41 },
		{ 398, 156, 41 },
		{ 398, 157, 41 },
		{ 398, 158, 41 },
		{ 398, 159, 41 },
		{ 398, 160, 41 },
		{ 398, 161, 41 },
		{ 398, 162, 41 },
		{ 398, 163, 41 },
		{ 398, 164, 41 },
		{ 398, 165, 41 },
		{ 398, 166, 41 },
		{ 398, 167, 41 },
		{ 398, 168, 41 },
		{ 398, 169, 41 },
		{ 398, 170, 41 },
		{ 398, 171, 41 },
		{ 398, 172, 41 },
		{ 398, 173, 41 },
		{ 398, 174, 41 },
		{ 398, 175, 41 },
		{ 398, 176, 41 },
		{ 398, 177, 41 },
		{ 398, 178, 41 },
		{ 398, 179, 41 },
		{ 398, 180, 41 },
		{ 398, 181, 41 },
		{ 398, 182, 41 },
		{ 398, 183, 41 },
		{ 398, 184, 41 },
		{ 398, 185, 41 },
		{ 398, 186, 41 },
		{ 399, 151, 41 },
		{ 399, 152, 41 },
		{ 399, 153, 41 },
		{ 399, 154, 41 },
		{ 399, 155, 41 },
		{ 399, 156, 41 },
		{ 399, 157, 41 },
		{ 399, 158, 41 },
		{ 399, 159, 41 },
		{ 399, 160, 41 },
		{ 399, 161, 41 },
		{ 399, 162, 41 },
		{ 399, 163, 41 },
		{ 399, 164, 41 },
		{ 399, 165, 41 },
		{ 399, 166, 41 },
		{ 399, 167, 41 },
		{ 399, 168, 41 },
		{ 399, 169, 41 },
		{ 399, 170, 41 },
		{ 399, 171, 41 },
		{ 399, 172, 41 },
		{ 399, 173, 41 },
		{ 399, 174, 41 },
		{ 399, 175, 41 },
		{ 399, 176, 41 },
		{ 399, 177, 41 },
		{ 399, 178, 41 },
		{ 399, 179, 41 },
		{ 399, 180, 41 },
		{ 399, 181, 41 },
		{ 399, 182, 41 },
		{ 399, 183, 41 },
		{ 399, 184, 41 },
		{ 399, 185, 41 },
		{ 400, 150, 41 },
		{ 400, 151, 41 },
		{ 400, 152, 41 },
		{ 400, 153, 41 },
		{ 400, 154, 41 },
		{ 400, 155, 41 },
		{ 400, 156, 41 },
		{ 400, 157, 41 },
		{ 400, 158, 41 },
		{ 400, 159, 41 },
		{ 400, 160, 41 },
		{ 400, 161, 41 },
		{ 400, 162, 41 },
		{ 400, 163, 41 },
		{ 400, 164, 41 },
		{ 400, 165, 41 },
		{ 400, 166, 41 },
		{ 400, 167, 41 },
		{ 400, 168, 41 },
		{ 400, 169, 41 },
		{ 400, 170, 41 },
		{ 400, 171, 41 },
		{ 400, 172, 41 },
		{ 400, 173, 41 },
		{ 400, 174, 41 },
		{ 400, 175, 41 },
		{ 400, 176, 41 },
		{ 400, 177, 41 },
		{ 400, 178, 41 },
		{ 400, 179, 41 },
		{ 400, 180, 41 },
		{ 400, 181, 41 },
		{ 400, 182, 41 },
		{ 400, 183, 41 },
		{ 400, 184, 41 },
		{ 400, 185, 41 },
		{ 401, 150, 41 },
		{ 401, 151, 41 },
		{ 401, 152, 41 },
		{ 401, 153, 41 },
		{ 401, 154, 41 },
		{ 401, 155, 41 },
		{ 401, 156, 41 },
		{ 401, 157, 41 },
		{ 401, 158, 41 },
		{ 401, 159, 41 },
		{ 401, 160, 41 },
		{ 401, 161, 41 },
		{ 401, 162, 41 },
		{ 401, 163, 41 },
		{ 401, 164, 41 },
		{ 401, 165, 41 },
		{ 401, 166, 41 },
		{ 401, 167, 41 },
		{ 401, 168, 41 },
		{ 401, 169, 41 },
		{ 401, 170, 41 },
		{ 401, 171, 41 },
		{ 401, 172, 41 },
		{ 401, 173, 41 },
		{ 401, 174, 41 },
		{ 401, 175, 41 },
		{ 401, 176, 41 },
		{ 401, 177, 41 },
		{ 401, 178, 41 },
		{ 401, 179, 41 },
		{ 401, 180, 41 },
		{ 401, 181, 41 },
		{ 401, 182, 41 },
		{ 401, 183, 41 },
		{ 401, 184, 41 },
		{ 402, 149, 41 },
		{ 402, 150, 41 },
		{ 402, 151, 41 },
		{ 402, 152, 41 },
		{ 402, 153, 41 },
		{ 402, 154, 41 },
		{ 402, 155, 41 },
		{ 402, 156, 41 },
		{ 402, 157, 41 },
		{ 402, 158, 41 },
		{ 402, 159, 41 },
		{ 402, 160, 41 },
		{ 402, 161, 41 },
		{ 402, 162, 41 },
		{ 402, 163, 41 },
		{ 402, 164, 41 },
		{ 402, 165, 41 },
		{ 402, 166, 41 },
		{ 402, 167, 41 },
		{ 402, 168, 41 },
		{ 402, 169, 41 },
		{ 402, 170, 41 },
		{ 402, 171, 41 },
		{ 402, 172, 41 },
		{ 402, 173, 41 },
		{ 402, 174, 41 },
		{ 402, 175, 41 },
		{ 402, 176, 41 },
		{ 402, 177, 41 },
		{ 402, 178, 41 },
		{ 402, 179, 41 },
		{ 402, 180, 41 },
		{ 402, 181, 41 },
		{ 402, 182, 41 },
		{ 402, 183, 41 },
		{ 402, 184, 41 },
		{ 403, 148, 41 },
		{ 403, 149, 41 },
		{ 403, 150, 41 },
		{ 403, 151, 41 },
		{ 403, 152, 41 },
		{ 403, 153, 41 },
		{ 403, 154, 41 },
		{ 403, 155, 41 },
		{ 403, 156, 41 },
		{ 403, 157, 41 },
		{ 403, 158, 41 },
		{ 403, 159, 41 },
		{ 403, 160, 41 },
		{ 403, 161, 41 },
		{ 403, 162, 41 },
		{ 403, 163, 41 },
		{ 403, 164, 41 },
		{ 403, 165, 41 },
		{ 403, 166, 41 },
		{ 403, 167, 41 },
		{ 403, 168, 41 },
		{ 403, 169, 41 },
		{ 403, 170, 41 },
		{ 403, 171, 41 },
		{ 403, 172, 41 },
		{ 403, 173, 41 },
		{ 403, 174, 41 },
		{ 403, 175, 41 },
		{ 403, 176, 41 },
		{ 403, 177, 41 },
		{ 403, 178, 41 },
		{ 403, 179, 41 },
		{ 403, 180, 41 },
		{ 403, 181, 41 },
		{ 403, 182, 41 },
		{ 403, 183, 41 },
		{ 404, 147, 41 },
		{ 404, 148, 41 },
		{ 404, 149, 41 },
		{ 404, 150, 41 },
		{ 404, 151, 41 },
		{ 404, 152, 41 },
		{ 404, 153, 41 },
		{ 404, 154, 41 },
		{ 404, 155, 41 },
		{ 404, 156, 41 },
		{ 404, 157, 41 },
		{ 404, 158, 41 },
		{ 404, 159, 41 },
		{ 404, 160, 41 },
		{ 404, 161, 41 },
		{ 404, 162, 41 },
		{ 404, 163, 41 },
		{ 404, 164, 41 },
		{ 404, 165, 41 },
		{ 404, 166, 41 },
		{ 404, 167, 41 },
		{ 404, 168, 41 },
		{ 404, 169, 41 },
		{ 404, 170, 41 },
		{ 404, 171, 41 },
		{ 404, 172, 41 },
		{ 404, 173, 41 },
		{ 404, 174, 41 },
		{ 404, 175, 41 },
		{ 404, 176, 41 },
		{ 404, 177, 41 },
		{ 404, 178, 41 },
		{ 404, 179, 41 },
		{ 404, 180, 41 },
		{ 404, 181, 41 },
		{ 404, 182, 41 },
		{ 404, 183, 41 },
		{ 405, 147, 41 },
		{ 405, 148, 41 },
		{ 405, 149, 41 },
		{ 405, 150, 41 },
		{ 405, 151, 41 },
		{ 405, 152, 41 },
		{ 405, 153, 41 },
		{ 405, 154, 41 },
		{ 405, 155, 41 },
		{ 405, 156, 41 },
		{ 405, 157, 41 },
		{ 405, 158, 41 },
		{ 405, 159, 41 },
		{ 405, 160, 41 },
		{ 405, 161, 41 },
		{ 405, 162, 41 },
		{ 405, 163, 41 },
		{ 405, 164, 41 },
		{ 405, 165, 41 },
		{ 405, 166, 41 },
		{ 405, 167, 41 },
		{ 405, 168, 41 },
		{ 405, 169, 41 },
		{ 405, 170, 41 },
		{ 405, 171, 41 },
		{ 405, 172, 41 },
		{ 405, 173, 41 },
		{ 405, 174, 41 },
		{ 405, 175, 41 },
		{ 405, 176, 41 },
		{ 405, 177, 41 },
		{ 405, 178, 41 },
		{ 405, 179, 41 },
		{ 405, 180, 41 },
		{ 405, 181, 41 },
		{ 405, 182, 41 },
		{ 406, 146, 41 },
		{ 406, 147, 41 },
		{ 406, 148, 41 },
		{ 406, 149, 41 },
		{ 406, 150, 41 },
		{ 406, 151, 41 },
		{ 406, 152, 41 },
		{ 406, 153, 41 },
		{ 406, 154, 41 },
		{ 406, 155, 41 },
		{ 406, 156, 41 },
		{ 406, 157, 41 },
		{ 406, 158, 41 },
		{ 406, 159, 41 },
		{ 406, 160, 41 },
		{ 406, 161, 41 },
		{ 406, 162, 41 },
		{ 406, 163, 41 },
		{ 406, 164, 41 },
		{ 406, 165, 41 },
		{ 406, 166, 41 },
		{ 406, 167, 41 },
		{ 406, 168, 41 },
		{ 406, 169, 41 },
		{ 406, 170, 41 },
		{ 406, 171, 41 },
		{ 406, 172, 41 },
		{ 406, 173, 41 },
		{ 406, 174, 41 },
		{ 406, 175, 41 },
		{ 406, 176, 41 },
		{ 406, 177, 41 },
		{ 406, 178, 41 },
		{ 406, 179, 41 },
		{ 406, 180, 41 },
		{ 406, 181, 41 },
		{ 406, 182, 41 },
		{ 407, 145, 41 },
		{ 407, 146, 41 },
		{ 407, 147, 41 },
		{ 407, 148, 41 },
		{ 407, 149, 41 },
		{ 407, 150, 41 },
		{ 407, 151, 41 },
		{ 407, 152, 41 },
		{ 407, 153, 41 },
		{ 407, 154, 41 },
		{ 407, 155, 41 },
		{ 407, 156, 41 },
		{ 407, 157, 41 },
		{ 407, 158, 41 },
		{ 407, 159, 41 },
		{ 407, 160, 41 },
		{ 407, 161, 41 },
		{ 407, 162, 41 },
		{ 407, 163, 41 },
		{ 407, 164, 41 },
		{ 407, 165, 41 },
		{ 407, 166, 41 },
		{ 407, 167, 41 },
		{ 407, 168, 41 },
		{ 407, 169, 41 },
		{ 407, 170, 41 },
		{ 407, 171, 41 },
		{ 407, 172, 41 },
		{ 407, 173, 41 },
		{ 407, 174, 41 },
		{ 407, 175, 41 },
		{ 407, 176, 41 },
		{ 407, 177, 41 },
		{ 407, 178, 41 },
		{ 407, 179, 41 },
		{ 407, 180, 41 },
		{ 407, 181, 41 },
		{ 408, 144, 41 },
		{ 408, 145, 41 },
		{ 408, 146, 41 },
		{ 408, 147, 41 },
		{ 408, 148, 41 },
		{ 408, 149, 41 },
		{ 408, 150, 41 },
		{ 408, 151, 41 },
		{ 408, 152, 41 },
		{ 408, 153, 41 },
		{ 408, 154, 41 },
		{ 408, 155, 41 },
		{ 408, 156, 41 },
		{ 408, 157, 41 },
		{ 408, 158, 41 },
		{ 408, 159, 41 },
		{ 408, 160, 41 },
		{ 408, 161, 41 },
		{ 408, 162, 41 },
		{ 408, 163, 41 },
		{ 408, 164, 41 },
		{ 408, 165, 41 },
		{ 408, 166, 41 },
		{ 408, 167, 41 },
		{ 408, 168, 41 },
		{ 408, 169, 41 },
		{ 408, 170, 41 },
		{ 408, 171, 41 },
		{ 408, 172, 41 },
		{ 408, 173, 41 },
		{ 408, 174, 41 },
		{ 408, 175, 41 },
		{ 408, 176, 41 },
		{ 408, 177, 41 },
		{ 408, 178, 41 },
		{ 408, 179, 41 },
		{ 408, 180, 41 },
		{ 408, 181, 41 },
		{ 409, 144, 41 },
		{ 409, 145, 41 },
		{ 409, 146, 41 },
		{ 409, 147, 41 },
		{ 409, 148, 41 },
		{ 409, 149, 41 },
		{ 409, 150, 41 },
		{ 409, 151, 41 },
		{ 409, 152, 41 },
		{ 409, 153, 41 },
		{ 409, 154, 41 },
		{ 409, 155, 41 },
		{ 409, 156, 41 },
		{ 409, 157, 41 },
		{ 409, 158, 41 },
		{ 409, 159, 41 },
		{ 409, 160, 41 },
		{ 409, 161, 41 },
		{ 409, 162, 41 },
		{ 409, 163, 41 },
		{ 409, 164, 41 },
		{ 409, 165, 41 },
		{ 409, 166, 41 },
		{ 409, 167, 41 },
		{ 409, 168, 41 },
		{ 409, 169, 41 },
		{ 409, 170, 41 },
		{ 409, 171, 41 },
		{ 409, 172, 41 },
		{ 409, 173, 41 },
		{ 409, 174, 41 },
		{ 409, 175, 41 },
		{ 409, 176, 41 },
		{ 409, 177, 41 },
		{ 409, 178, 41 },
		{ 409, 179, 41 },
		{ 409, 180, 41 },
		{ 410, 143, 41 },
		{ 410, 144, 41 },
		{ 410, 145, 41 },
		{ 410, 146, 41 },
		{ 410, 147, 41 },
		{ 410, 148, 41 },
		{ 410, 149, 41 },
		{ 410, 150, 41 },
		{ 410, 151, 41 },
		{ 410, 152, 41 },
		{ 410, 153, 41 },
		{ 410, 154, 41 },
		{ 410, 155, 41 },
		{ 410, 156, 41 },
		{ 410, 157, 41 },
		{ 410, 158, 41 },
		{ 410, 159, 41 },
		{ 410, 160, 41 },
		{ 410, 161, 41 },
		{ 410, 162, 41 },
		{ 410, 163, 41 },
		{ 410, 164, 41 },
		{ 410, 165, 41 },
		{ 410, 166, 41 },
		{ 410, 167, 41 },
		{ 410, 168, 41 },
		{ 410, 169, 41 },
		{ 410, 170, 41 },
		{ 410, 171, 41 },
		{ 410, 172, 41 },
		{ 410, 173, 41 },
		{ 410, 174, 41 },
		{ 410, 175, 41 },
		{ 410, 176, 41 },
		{ 410, 177, 41 },
		{ 410, 178, 41 },
		{ 410, 179, 41 },
		{ 410, 180, 41 },
		{ 411, 142, 41 },
		{ 411, 143, 41 },
		{ 411, 144, 41 },
		{ 411, 145, 41 },
		{ 411, 146, 41 },
		{ 411, 147, 41 },
		{ 411, 148, 41 },
		{ 411, 149, 41 },
		{ 411, 150, 41 },
		{ 411, 151, 41 },
		{ 411, 152, 41 },
		{ 411, 153, 41 },
		{ 411, 154, 41 },
		{ 411, 155, 41 },
		{ 411, 156, 41 },
		{ 411, 157, 41 },
		{ 411, 158, 41 },
		{ 411, 159, 41 },
		{ 411, 160, 41 },
		{ 411, 161, 41 },
		{ 411, 162, 41 },
		{ 411, 163, 41 },
		{ 411, 164, 41 },
		{ 411, 165, 41 },
		{ 411, 166, 41 },
		{ 411, 167, 41 },
		{ 411, 168, 41 },
		{ 411, 169, 41 },
		{ 411, 170, 41 },
		{ 411, 171, 41 },
		{ 411, 172, 41 },
		{ 411, 173, 41 },
		{ 411, 174, 41 },
		{ 411, 175, 41 },
		{ 411, 176, 41 },
		{ 411, 177, 41 },
		{ 411, 178, 41 },
		{ 411, 179, 41 },
		{ 412, 141, 41 },
		{ 412, 142, 41 },
		{ 412, 143, 41 },
		{ 412, 144, 41 },
		{ 412, 145, 41 },
		{ 412, 146, 41 },
		{ 412, 147, 41 },
		{ 412, 148, 41 },
		{ 412, 149, 41 },
		{ 412, 150, 41 },
		{ 412, 151, 41 },
		{ 412, 152, 41 },
		{ 412, 153, 41 },
		{ 412, 154, 41 },
		{ 412, 155, 41 },
		{ 412, 156, 41 },
		{ 412, 157, 41 },
		{ 412, 158, 41 },
		{ 412, 159, 41 },
		{ 412, 160, 41 },
		{ 412, 161, 41 },
		{ 412, 162, 41 },
		{ 412, 163, 41 },
		{ 412, 164, 41 },
		{ 412, 165, 41 },
		{ 412, 166, 41 },
		{ 412, 167, 41 },
		{ 412, 168, 41 },
		{ 412, 169, 41 },
		{ 412, 170, 41 },
		{ 412, 171, 41 },
		{ 412, 172, 41 },
		{ 412, 173, 41 },
		{ 412, 174, 41 },
		{ 412, 175, 41 },
		{ 412, 176, 41 },
		{ 412, 177, 41 },
		{ 412, 178, 41 },
		{ 412, 179, 41 },
		{ 413, 141, 41 },
		{ 413, 142, 41 },
		{ 413, 143, 41 },
		{ 413, 144, 41 },
		{ 413, 145, 41 },
		{ 413, 146, 41 },
		{ 413, 147, 41 },
		{ 413, 148, 41 },
		{ 413, 149, 41 },
		{ 413, 150, 41 },
		{ 413, 151, 41 },
		{ 413, 152, 41 },
		{ 413, 153, 41 },
		{ 413, 154, 41 },
		{ 413, 155, 41 },
		{ 413, 156, 41 },
		{ 413, 157, 41 },
		{ 413, 158, 41 },
		{ 413, 159, 41 },
		{ 413, 160, 41 },
		{ 413, 161, 41 },
		{ 413, 162, 41 },
		{ 413, 163, 41 },
		{ 413, 164, 41 },
		{ 413, 165, 41 },
		{ 413, 166, 41 },
		{ 413, 167, 41 },
		{ 413, 168, 41 },
		{ 413, 169, 41 },
		{ 413, 170, 41 },
		{ 413, 171, 41 },
		{ 413, 172, 41 },
		{ 413, 173, 41 },
		{ 413, 174, 41 },
		{ 413, 175, 41 },
		{ 413, 176, 41 },
		{ 413, 177, 41 },
		{ 413, 178, 41 },
		{ 414, 140, 41 },
		{ 414, 141, 41 },
		{ 414, 142, 41 },
		{ 414, 143, 41 },
		{ 414, 144, 41 },
		{ 414, 145, 41 },
		{ 414, 146, 41 },
		{ 414, 147, 41 },
		{ 414, 148, 41 },
		{ 414, 149, 41 },
		{ 414, 150, 41 },
		{ 414, 151, 41 },
		{ 414, 152, 41 },
		{ 414, 153, 41 },
		{ 414, 154, 41 },
		{ 414, 155, 41 },
		{ 414, 156, 41 },
		{ 414, 157, 41 },
		{ 414, 158, 41 },
		{ 414, 159, 41 },
		{ 414, 160, 41 },
		{ 414, 161, 41 },
		{ 414, 162, 41 },
		{ 414, 163, 41 },
		{ 414, 164, 41 },
		{ 414, 165, 41 },
		{ 414, 166, 41 },
		{ 414, 167, 41 },
		{ 414, 168, 41 },
		{ 414, 169, 41 },
		{ 414, 170, 41 },
		{ 414, 171, 41 },
		{ 414, 172, 41 },
		{ 414, 173, 41 },
		{ 414, 174, 41 },
		{ 414, 175, 41 },
		{ 414, 176, 41 },
		{ 414, 177, 41 },
		{ 414, 178, 41 },
		{ 415, 139, 41 },
		{ 415, 140, 41 },
		{ 415, 141, 41 },
		{ 415, 142, 41 },
		{ 415, 143, 41 },
		{ 415, 144, 41 },
		{ 415, 145, 41 },
		{ 415, 146, 41 },
		{ 415, 147, 41 },
		{ 415, 148, 41 },
		{ 415, 149, 41 },
		{ 415, 150, 41 },
		{ 415, 151, 41 },
		{ 415, 152, 41 },
		{ 415, 153, 41 },
		{ 415, 154, 41 },
		{ 415, 155, 41 },
		{ 415, 156, 41 },
		{ 415, 157, 41 },
		{ 415, 158, 41 },
		{ 415, 159, 41 },
		{ 415, 160, 41 },
		{ 415, 161, 41 },
		{ 415, 162, 41 },
		{ 415, 163, 41 },
		{ 415, 164, 41 },
		{ 415, 165, 41 },
		{ 415, 166, 41 },
		{ 415, 167, 41 },
		{ 415, 168, 41 },
		{ 415, 169, 41 },
		{ 415, 170, 41 },
		{ 415, 171, 41 },
		{ 415, 172, 41 },
		{ 415, 173, 41 },
		{ 415, 174, 41 },
		{ 415, 175, 41 },
		{ 415, 176, 41 },
		{ 415, 177, 41 },
		{ 416, 138, 41 },
		{ 416, 139, 41 },
		{ 416, 140, 41 },
		{ 416, 141, 41 },
		{ 416, 142, 41 },
		{ 416, 143, 41 },
		{ 416, 144, 41 },
		{ 416, 145, 41 },
		{ 416, 146, 41 },
		{ 416, 147, 41 },
		{ 416, 148, 41 },
		{ 416, 149, 41 },
		{ 416, 150, 41 },
		{ 416, 151, 41 },
		{ 416, 152, 41 },
		{ 416, 153, 41 },
		{ 416, 154, 41 },
		{ 416, 155, 41 },
		{ 416, 156, 41 },
		{ 416, 157, 41 },
		{ 416, 158, 41 },
		{ 416, 159, 41 },
		{ 416, 160, 41 },
		{ 416, 161, 41 },
		{ 416, 162, 41 },
		{ 416, 163, 41 },
		{ 416, 164, 41 },
		{ 416, 165, 41 },
		{ 416, 166, 41 },
		{ 416, 167, 41 },
		{ 416, 168, 41 },
		{ 416, 169, 41 },
		{ 416, 170, 41 },
		{ 416, 171, 41 },
		{ 416, 172, 41 },
		{ 416, 173, 41 },
		{ 416, 174, 41 },
		{ 416, 175, 41 },
		{ 416, 176, 41 },
		{ 416, 177, 41 },
		{ 417, 138, 41 },
		{ 417, 139, 41 },
		{ 417, 140, 41 },
		{ 417, 141, 41 },
		{ 417, 142, 41 },
		{ 417, 143, 41 },
		{ 417, 144, 41 },
		{ 417, 145, 41 },
		{ 417, 146, 41 },
		{ 417, 147, 41 },
		{ 417, 148, 41 },
		{ 417, 149, 41 },
		{ 417, 150, 41 },
		{ 417, 151, 41 },
		{ 417, 152, 41 },
		{ 417, 153, 41 },
		{ 417, 154, 41 },
		{ 417, 155, 41 },
		{ 417, 156, 41 },
		{ 417, 157, 41 },
		{ 417, 158, 41 },
		{ 417, 159, 41 },
		{ 417, 160, 41 },
		{ 417, 161, 41 },
		{ 417, 162, 41 },
		{ 417, 163, 41 },
		{ 417, 164, 41 },
		{ 417, 165, 41 },
		{ 417, 166, 41 },
		{ 417, 167, 41 },
		{ 417, 168, 41 },
		{ 417, 169, 41 },
		{ 417, 170, 41 },
		{ 417, 171, 41 },
		{ 417, 172, 41 },
		{ 417, 173, 41 },
		{ 417, 174, 41 },
		{ 417, 175, 41 },
		{ 417, 176, 41 },
		{ 418, 137, 41 },
		{ 418, 138, 41 },
		{ 418, 139, 41 },
		{ 418, 140, 41 },
		{ 418, 141, 41 },
		{ 418, 142, 41 },
		{ 418, 143, 41 },
		{ 418, 144, 41 },
		{ 418, 145, 41 },
		{ 418, 146, 41 },
		{ 418, 147, 41 },
		{ 418, 148, 41 },
		{ 418, 149, 41 },
		{ 418, 150, 41 },
		{ 418, 151, 41 },
		{ 418, 152, 41 },
		{ 418, 153, 41 },
		{ 418, 154, 41 },
		{ 418, 155, 41 },
		{ 418, 156, 41 },
		{ 418, 157, 41 },
		{ 418, 158, 41 },
		{ 418, 159, 41 },
		{ 418, 160, 41 },
		{ 418, 161, 41 },
		{ 418, 162, 41 },
		{ 418, 163, 41 },
		{ 418, 164, 41 },
		{ 418, 165, 41 },
		{ 418, 166, 41 },
		{ 418, 167, 41 },
		{ 418, 168, 41 },
		{ 418, 169, 41 },
		{ 418, 170, 41 },
		{ 418, 171, 41 },
		{ 418, 172, 41 },
		{ 418, 173, 41 },
		{ 418, 174, 41 },
		{ 418, 175, 41 },
		{ 418, 176, 41 },
		{ 419, 136, 41 },
		{ 419, 137, 41 },
		{ 419, 138, 41 },
		{ 419, 139, 41 },
		{ 419, 140, 41 },
		{ 419, 141, 41 },
		{ 419, 142, 41 },
		{ 419, 143, 41 },
		{ 419, 144, 41 },
		{ 419, 145, 41 },
		{ 419, 146, 41 },
		{ 419, 147, 41 },
		{ 419, 148, 41 },
		{ 419, 149, 41 },
		{ 419, 150, 41 },
		{ 419, 151, 41 },
		{ 419, 152, 41 },
		{ 419, 153, 41 },
		{ 419, 154, 41 },
		{ 419, 155, 41 },
		{ 419, 156, 41 },
		{ 419, 157, 41 },
		{ 419, 158, 41 },
		{ 419, 159, 41 },
		{ 419, 160, 41 },
		{ 419, 161, 41 },
		{ 419, 162, 41 },
		{ 419, 163, 41 },
		{ 419, 164, 41 },
		{ 419, 165, 41 },
		{ 419, 166, 41 },
		{ 419, 167, 41 },
		{ 419, 168, 41 },
		{ 419, 169, 41 },
		{ 419, 170, 41 },
		{ 419, 171, 41 },
		{ 419, 172, 41 },
		{ 419, 173, 41 },
		{ 419, 174, 41 },
		{ 419, 175, 41 },
		{ 420, 135, 41 },
		{ 420, 136, 41 },
		{ 420, 137, 41 },
		{ 420, 138, 41 },
		{ 420, 139, 41 },
		{ 420, 140, 41 },
		{ 420, 141, 41 },
		{ 420, 142, 41 },
		{ 420, 143, 41 },
		{ 420, 144, 41 },
		{ 420, 145, 41 },
		{ 420, 146, 41 },
		{ 420, 147, 41 },
		{ 420, 148, 41 },
		{ 420, 149, 41 },
		{ 420, 150, 41 },
		{ 420, 151, 41 },
		{ 420, 152, 41 },
		{ 420, 153, 41 },
		{ 420, 154, 41 },
		{ 420, 155, 41 },
		{ 420, 156, 41 },
		{ 420, 157, 41 },
		{ 420, 158, 41 },
		{ 420, 159, 41 },
		{ 420, 160, 41 },
		{ 420, 161, 41 },
		{ 420, 162, 41 },
		{ 420, 163, 41 },
		{ 420, 164, 41 },
		{ 420, 165, 41 },
		{ 420, 166, 41 },
		{ 420, 167, 41 },
		{ 420, 168, 41 },
		{ 420, 169, 41 },
		{ 420, 170, 41 },
		{ 420, 171, 41 },
		{ 420, 172, 41 },
		{ 420, 173, 41 },
		{ 420, 174, 41 },
		{ 420, 175, 41 },
		{ 421, 135, 41 },
		{ 421, 136, 41 },
		{ 421, 137, 41 },
		{ 421, 138, 41 },
		{ 421, 139, 41 },
		{ 421, 140, 41 },
		{ 421, 141, 41 },
		{ 421, 142, 41 },
		{ 421, 143, 41 },
		{ 421, 144, 41 },
		{ 421, 145, 41 },
		{ 421, 146, 41 },
		{ 421, 147, 41 },
		{ 421, 148, 41 },
		{ 421, 149, 41 },
		{ 421, 150, 41 },
		{ 421, 151, 41 },
		{ 421, 152, 41 },
		{ 421, 153, 41 },
		{ 421, 154, 41 },
		{ 421, 155, 41 },
		{ 421, 156, 41 },
		{ 421, 157, 41 },
		{ 421, 158, 41 },
		{ 421, 159, 41 },
		{ 421, 160, 41 },
		{ 421, 161, 41 },
		{ 421, 162, 41 },
		{ 421, 163, 41 },
		{ 421, 164, 41 },
		{ 421, 165, 41 },
		{ 421, 166, 41 },
		{ 421, 167, 41 },
		{ 421, 168, 41 },
		{ 421, 169, 41 },
		{ 421, 170, 41 },
		{ 421, 171, 41 },
		{ 421, 172, 41 },
		{ 421, 173, 41 },
		{ 421, 174, 41 },
		{ 422, 134, 41 },
		{ 422, 135, 41 },
		{ 422, 136, 41 },
		{ 422, 137, 41 },
		{ 422, 138, 41 },
		{ 422, 139, 41 },
		{ 422, 140, 41 },
		{ 422, 141, 41 },
		{ 422, 142, 41 },
		{ 422, 143, 41 },
		{ 422, 144, 41 },
		{ 422, 145, 41 },
		{ 422, 146, 41 },
		{ 422, 147, 41 },
		{ 422, 148, 41 },
		{ 422, 149, 41 },
		{ 422, 150, 41 },
		{ 422, 151, 41 },
		{ 422, 152, 41 },
		{ 422, 153, 41 },
		{ 422, 154, 41 },
		{ 422, 155, 41 },
		{ 422, 156, 41 },
		{ 422, 157, 41 },
		{ 422, 158, 41 },
		{ 422, 159, 41 },
		{ 422, 160, 41 },
		{ 422, 161, 41 },
		{ 422, 162, 41 },
		{ 422, 163, 41 },
		{ 422, 164, 41 },
		{ 422, 165, 41 },
		{ 422, 166, 41 },
		{ 422, 167, 41 },
		{ 422, 168, 41 },
		{ 422, 169, 41 },
		{ 422, 170, 41 },
		{ 422, 171, 41 },
		{ 422, 172, 41 },
		{ 422, 173, 41 },
		{ 422, 174, 41 },
		{ 423, 133, 41 },
		{ 423, 134, 41 },
		{ 423, 135, 41 },
		{ 423, 136, 41 },
		{ 423, 137, 41 },
		{ 423, 138, 41 },
		{ 423, 139, 41 },
		{ 423, 140, 41 },
		{ 423, 141, 41 },
		{ 423, 142, 41 },
		{ 423, 143, 41 },
		{ 423, 144, 41 },
		{ 423, 145, 41 },
		{ 423, 146, 41 },
		{ 423, 147, 41 },
		{ 423, 148, 41 },
		{ 423, 149, 41 },
		{ 423, 150, 41 },
		{ 423, 151, 41 },
		{ 423, 152, 41 },
		{ 423, 153, 41 },
		{ 423, 154, 41 },
		{ 423, 155, 41 },
		{ 423, 156, 41 },
		{ 423, 157, 41 },
		{ 423, 158, 41 },
		{ 423, 159, 41 },
		{ 423, 160, 41 },
		{ 423, 161, 41 },
		{ 423, 162, 41 },
		{ 423, 163, 41 },
		{ 423, 164, 41 },
		{ 423, 165, 41 },
		{ 423, 166, 41 },
		{ 423, 167, 41 },
		{ 423, 168, 41 },
		{ 423, 169, 41 },
		{ 423, 170, 41 },
		{ 423, 171, 41 },
		{ 423, 172, 41 },
		{ 423, 173, 41 },
		{ 424, 132, 41 },
		{ 424, 133, 41 },
		{ 424, 134, 41 },
		{ 424, 135, 41 },
		{ 424, 136, 41 },
		{ 424, 137, 41 },
		{ 424, 138, 41 },
		{ 424, 139, 41 },
		{ 424, 140, 41 },
		{ 424, 141, 41 },
		{ 424, 142, 41 },
		{ 424, 143, 41 },
		{ 424, 144, 41 },
		{ 424, 145, 41 },
		{ 424, 146, 41 },
		{ 424, 147, 41 },
		{ 424, 148, 41 },
		{ 424, 149, 41 },
		{ 424, 150, 41 },
		{ 424, 151, 41 },
		{ 424, 152, 41 },
		{ 424, 153, 41 },
		{ 424, 154, 41 },
		{ 424, 155, 41 },
		{ 424, 156, 41 },
		{ 424, 157, 41 },
		{ 424, 158, 41 },
		{ 424, 159, 41 },
		{ 424, 160, 41 },
		{ 424, 161, 41 },
		{ 424, 162, 41 },
		{ 424, 163, 41 },
		{ 424, 164, 41 },
		{ 424, 165, 41 },
		{ 424, 166, 41 },
		{ 424, 167, 41 },
		{ 424, 168, 41 },
		{ 424, 169, 41 },
		{ 424, 170, 41 },
		{ 424, 171, 41 },
		{ 424, 172, 41 },
		{ 424, 173, 41 },
		{ 425, 132, 41 },
		{ 425, 133, 41 },
		{ 425, 134, 41 },
		{ 425, 135, 41 },
		{ 425, 136, 41 },
		{ 425, 137, 41 },
		{ 425, 138, 41 },
		{ 425, 139, 41 },
		{ 425, 140, 41 },
		{ 425, 141, 41 },
		{ 425, 142, 41 },
		{ 425, 143, 41 },
		{ 425, 144, 41 },
		{ 425, 145, 41 },
		{ 425, 146, 41 },
		{ 425, 147, 41 },
		{ 425, 148, 41 },
		{ 425, 149, 41 },
		{ 425, 150, 41 },
		{ 425, 151, 41 },
		{ 425, 152, 41 },
		{ 425, 153, 41 },
		{ 425, 154, 41 },
		{ 425, 155, 41 },
		{ 425, 156, 41 },
		{ 425, 157, 41 },
		{ 425, 158, 41 },
		{ 425, 159, 41 },
		{ 425, 160, 41 },
		{ 425, 161, 41 },
		{ 425, 162, 41 },
		{ 425, 163, 41 },
		{ 425, 164, 41 },
		{ 425, 165, 41 },
		{ 425, 166, 41 },
		{ 425, 167, 41 },
		{ 425, 168, 41 },
		{ 425, 169, 41 },
		{ 425, 170, 41 },
		{ 425, 171, 41 },
		{ 425, 172, 41 },
		{ 426, 131, 41 },
		{ 426, 132, 41 },
		{ 426, 133, 41 },
		{ 426, 134, 41 },
		{ 426, 135, 41 },
		{ 426, 136, 41 },
		{ 426, 137, 41 },
		{ 426, 138, 41 },
		{ 426, 139, 41 },
		{ 426, 140, 41 },
		{ 426, 141, 41 },
		{ 426, 142, 41 },
		{ 426, 143, 41 },
		{ 426, 144, 41 },
		{ 426, 145, 41 },
		{ 426, 146, 41 },
		{ 426, 147, 41 },
		{ 426, 148, 41 },
		{ 426, 149, 41 },
		{ 426, 150, 41 },
		{ 426, 151, 41 },
		{ 426, 152, 41 },
		{ 426, 153, 41 },
		{ 426, 154, 41 },
		{ 426, 155, 41 },
		{ 426, 156, 41 },
		{ 426, 157, 41 },
		{ 426, 158, 41 },
		{ 426, 159, 41 },
		{ 426, 160, 41 },
		{ 426, 161, 41 },
		{ 426, 162, 41 },
		{ 426, 163, 41 },
		{ 426, 164, 41 },
		{ 426, 165, 41 },
		{ 426, 166, 41 },
		{ 426, 167, 41 },
		{ 426, 168, 41 },
		{ 426, 169, 41 },
		{ 426, 170, 41 },
		{ 426, 171, 41 },
		{ 426, 172, 41 },
		{ 427, 130, 41 },
		{ 427, 131, 41 },
		{ 427, 132, 41 },
		{ 427, 133, 41 },
		{ 427, 134, 41 },
		{ 427, 135, 41 },
		{ 427, 136, 41 },
		{ 427, 137, 41 },
		{ 427, 138, 41 },
		{ 427, 139, 41 },
		{ 427, 140, 41 },
		{ 427, 141, 41 },
		{ 427, 142, 41 },
		{ 427, 143, 41 },
		{ 427, 144, 41 },
		{ 427, 145, 41 },
		{ 427, 146, 41 },
		{ 427, 147, 41 },
		{ 427, 148, 41 },
		{ 427, 149, 41 },
		{ 427, 150, 41 },
		{ 427, 151, 41 },
		{ 427, 152, 41 },
		{ 427, 153, 41 },
		{ 427, 154, 41 },
		{ 427, 155, 41 },
		{ 427, 156, 41 },
		{ 427, 157, 41 },
		{ 427, 158, 41 },
		{ 427, 159, 41 },
		{ 427, 160, 41 },
		{ 427, 161, 41 },
		{ 427, 162, 41 },
		{ 427, 163, 41 },
		{ 427, 164, 41 },
		{ 427, 165, 41 },
		{ 427, 166, 41 },
		{ 427, 167, 41 },
		{ 427, 168, 41 },
		{ 427, 169, 41 },
		{ 427, 170, 41 },
		{ 427, 171, 41 },
		{ 428, 129, 41 },
		{ 428, 130, 41 },
		{ 428, 131, 41 },
		{ 428, 132, 41 },
		{ 428, 133, 41 },
		{ 428, 134, 41 },
		{ 428, 135, 41 },
		{ 428, 136, 41 },
		{ 428, 137, 41 },
		{ 428, 138, 41 },
		{ 428, 139, 41 },
		{ 428, 140, 41 },
		{ 428, 141, 41 },
		{ 428, 142, 41 },
		{ 428, 143, 41 },
		{ 428, 144, 41 },
		{ 428, 145, 41 },
		{ 428, 146, 41 },
		{ 428, 147, 41 },
		{ 428, 148, 41 },
		{ 428, 149, 41 },
		{ 428, 150, 41 },
		{ 428, 151, 41 },
		{ 428, 152, 41 },
		{ 428, 153, 41 },
		{ 428, 154, 41 },
		{ 428, 155, 41 },
		{ 428, 156, 41 },
		{ 428, 157, 41 },
		{ 428, 158, 41 },
		{ 428, 159, 41 },
		{ 428, 160, 41 },
		{ 428, 161, 41 },
		{ 428, 162, 41 },
		{ 428, 163, 41 },
		{ 428, 164, 41 },
		{ 428, 165, 41 },
		{ 428, 166, 41 },
		{ 428, 167, 41 },
		{ 428, 168, 41 },
		{ 428, 169, 41 },
		{ 428, 170, 41 },
		{ 428, 171, 41 },
		{ 429, 129, 41 },
		{ 429, 130, 41 },
		{ 429, 131, 41 },
		{ 429, 132, 41 },
		{ 429, 133, 41 },
		{ 429, 134, 41 },
		{ 429, 135, 41 },
		{ 429, 136, 41 },
		{ 429, 137, 41 },
		{ 429, 138, 41 },
		{ 429, 139, 41 },
		{ 429, 140, 41 },
		{ 429, 141, 41 },
		{ 429, 142, 41 },
		{ 429, 143, 41 },
		{ 429, 144, 41 },
		{ 429, 145, 41 },
		{ 429, 146, 41 },
		{ 429, 147, 41 },
		{ 429, 148, 41 },
		{ 429, 149, 41 },
		{ 429, 150, 41 },
		{ 429, 151, 41 },
		{ 429, 152, 41 },
		{ 429, 153, 41 },
		{ 429, 154, 41 },
		{ 429, 155, 41 },
		{ 429, 156, 41 },
		{ 429, 157, 41 },
		{ 429, 158, 41 },
		{ 429, 159, 41 },
		{ 429, 160, 41 },
		{ 429, 161, 41 },
		{ 429, 162, 41 },
		{ 429, 163, 41 },
		{ 429, 164, 41 },
		{ 429, 165, 41 },
		{ 429, 166, 41 },
		{ 429, 167, 41 },
		{ 429, 168, 41 },
		{ 429, 169, 41 },
		{ 429, 170, 41 },
		{ 430, 128, 41 },
		{ 430, 129, 41 },
		{ 430, 130, 41 },
		{ 430, 131, 41 },
		{ 430, 132, 41 },
		{ 430, 133, 41 },
		{ 430, 134, 41 },
		{ 430, 135, 41 },
		{ 430, 136, 41 },
		{ 430, 137, 41 },
		{ 430, 138, 41 },
		{ 430, 139, 41 },
		{ 430, 140, 41 },
		{ 430, 141, 41 },
		{ 430, 142, 41 },
		{ 430, 143, 41 },
		{ 430, 144, 41 },
		{ 430, 145, 41 },
		{ 430, 146, 41 },
		{ 430, 147, 41 },
		{ 430, 148, 41 },
		{ 430, 149, 41 },
		{ 430, 150, 41 },
		{ 430, 151, 41 },
		{ 430, 152, 41 },
		{ 430, 153, 41 },
		{ 430, 154, 41 },
		{ 430, 155, 41 },
		{ 430, 156, 41 },
		{ 430, 157, 41 },
		{ 430, 158, 41 },
		{ 430, 159, 41 },
		{ 430, 160, 41 },
		{ 430, 161, 41 },
		{ 430, 162, 41 },
		{ 430, 163, 41 },
		{ 430, 164, 41 },
		{ 430, 165, 41 },
		{ 430, 166, 41 },
		{ 430, 167, 41 },
		{ 430, 168, 41 },
		{ 430, 169, 41 },
		{ 430, 170, 41 },
		{ 431, 127, 41 },
		{ 431, 128, 41 },
		{ 431, 129, 41 },
		{ 431, 130, 41 },
		{ 431, 131, 41 },
		{ 431, 132, 41 },
		{ 431, 133, 41 },
		{ 431, 134, 41 },
		{ 431, 135, 41 },
		{ 431, 136, 41 },
		{ 431, 137, 41 },
		{ 431, 138, 41 },
		{ 431, 139, 41 },
		{ 431, 140, 41 },
		{ 431, 141, 41 },
		{ 431, 142, 41 },
		{ 431, 143, 41 },
		{ 431, 144, 41 },
		{ 431, 145, 41 },
		{ 431, 146, 41 },
		{ 431, 147, 41 },
		{ 431, 148, 41 },
		{ 431, 149, 41 },
		{ 431, 150, 41 },
		{ 431, 151, 41 },
		{ 431, 152, 41 },
		{ 431, 153, 41 },
		{ 431, 154, 41 },
		{ 431, 155, 41 },
		{ 431, 156, 41 },
		{ 431, 157, 41 },
		{ 431, 158, 41 },
		{ 431, 159, 41 },
		{ 431, 160, 41 },
		{ 431, 161, 41 },
		{ 431, 162, 41 },
		{ 431, 163, 41 },
		{ 431, 164, 41 },
		{ 431, 165, 41 },
		{ 431, 166, 41 },
		{ 431, 167, 41 },
		{ 431, 168, 41 },
		{ 431, 169, 41 },
		{ 432, 126, 41 },
		{ 432, 127, 41 },
		{ 432, 128, 41 },
		{ 432, 129, 41 },
		{ 432, 130, 41 },
		{ 432, 131, 41 },
		{ 432, 132, 41 },
		{ 432, 133, 41 },
		{ 432, 134, 41 },
		{ 432, 135, 41 },
		{ 432, 136, 41 },
		{ 432, 137, 41 },
		{ 432, 138, 41 },
		{ 432, 139, 41 },
		{ 432, 140, 41 },
		{ 432, 141, 41 },
		{ 432, 142, 41 },
		{ 432, 143, 41 },
		{ 432, 144, 41 },
		{ 432, 145, 41 },
		{ 432, 146, 41 },
		{ 432, 147, 41 },
		{ 432, 148, 41 },
		{ 432, 149, 41 },
		{ 432, 150, 41 },
		{ 432, 151, 41 },
		{ 432, 152, 41 },
		{ 432, 153, 41 },
		{ 432, 154, 41 },
		{ 432, 155, 41 },
		{ 432, 156, 41 },
		{ 432, 157, 41 },
		{ 432, 158, 41 },
		{ 432, 159, 41 },
		{ 432, 160, 41 },
		{ 432, 161, 41 },
		{ 432, 162, 41 },
		{ 432, 163, 41 },
		{ 432, 164, 41 },
		{ 432, 165, 41 },
		{ 432, 166, 41 },
		{ 432, 167, 41 },
		{ 432, 168, 41 },
		{ 432, 169, 41 },
		{ 433, 126, 41 },
		{ 433, 127, 41 },
		{ 433, 128, 41 },
		{ 433, 129, 41 },
		{ 433, 130, 41 },
		{ 433, 131, 41 },
		{ 433, 132, 41 },
		{ 433, 133, 41 },
		{ 433, 134, 41 },
		{ 433, 135, 41 },
		{ 433, 136, 41 },
		{ 433, 137, 41 },
		{ 433, 138, 41 },
		{ 433, 139, 41 },
		{ 433, 140, 41 },
		{ 433, 141, 41 },
		{ 433, 142, 41 },
		{ 433, 143, 41 },
		{ 433, 144, 41 },
		{ 433, 145, 41 },
		{ 433, 146, 41 },
		{ 433, 147, 41 },
		{ 433, 148, 41 },
		{ 433, 149, 41 },
		{ 433, 150, 41 },
		{ 433, 151, 41 },
		{ 433, 152, 41 },
		{ 433, 153, 41 },
		{ 433, 154, 41 },
		{ 433, 155, 41 },
		{ 433, 156, 41 },
		{ 433, 157, 41 },
		{ 433, 158, 41 },
		{ 433, 159, 41 },
		{ 433, 160, 41 },
		{ 433, 161, 41 },
		{ 433, 162, 41 },
		{ 433, 163, 41 },
		{ 433, 164, 41 },
		{ 433, 165, 41 },
		{ 433, 166, 41 },
		{ 433, 167, 41 },
		{ 433, 168, 41 },
		{ 434, 125, 41 },
		{ 434, 126, 41 },
		{ 434, 127, 41 },
		{ 434, 128, 41 },
		{ 434, 129, 41 },
		{ 434, 130, 41 },
		{ 434, 131, 41 },
		{ 434, 132, 41 },
		{ 434, 133, 41 },
		{ 434, 134, 41 },
		{ 434, 135, 41 },
		{ 434, 136, 41 },
		{ 434, 137, 41 },
		{ 434, 138, 41 },
		{ 434, 139, 41 },
		{ 434, 140, 41 },
		{ 434, 141, 41 },
		{ 434, 142, 41 },
		{ 434, 143, 41 },
		{ 434, 144, 41 },
		{ 434, 145, 41 },
		{ 434, 146, 41 },
		{ 434, 147, 41 },
		{ 434, 148, 41 },
		{ 434, 149, 41 },
		{ 434, 150, 41 },
		{ 434, 151, 41 },
		{ 434, 152, 41 },
		{ 434, 153, 41 },
		{ 434, 154, 41 },
		{ 434, 155, 41 },
		{ 434, 156, 41 },
		{ 434, 157, 41 },
		{ 434, 158, 41 },
		{ 434, 159, 41 },
		{ 434, 160, 41 },
		{ 434, 161, 41 },
		{ 434, 162, 41 },
		{ 434, 163, 41 },
		{ 434, 164, 41 },
		{ 434, 165, 41 },
		{ 434, 166, 41 },
		{ 434, 167, 41 },
		{ 434, 168, 41 },
		{ 435, 124, 41 },
		{ 435, 125, 41 },
		{ 435, 126, 41 },
		{ 435, 127, 41 },
		{ 435, 128, 41 },
		{ 435, 129, 41 },
		{ 435, 130, 41 },
		{ 435, 131, 41 },
		{ 435, 132, 41 },
		{ 435, 133, 41 },
		{ 435, 134, 41 },
		{ 435, 135, 41 },
		{ 435, 136, 41 },
		{ 435, 137, 41 },
		{ 435, 138, 41 },
		{ 435, 139, 41 },
		{ 435, 140, 41 },
		{ 435, 141, 41 },
		{ 435, 142, 41 },
		{ 435, 143, 41 },
		{ 435, 144, 41 },
		{ 435, 145, 41 },
		{ 435, 146, 41 },
		{ 435, 147, 41 },
		{ 435, 148, 41 },
		{ 435, 149, 41 },
		{ 435, 150, 41 },
		{ 435, 151, 41 },
		{ 435, 152, 41 },
		{ 435, 153, 41 },
		{ 435, 154, 41 },
		{ 435, 155, 41 },
		{ 435, 156, 41 },
		{ 435, 157, 41 },
		{ 435, 158, 41 },
		{ 435, 159, 41 },
		{ 435, 160, 41 },
		{ 435, 161, 41 },
		{ 435, 162, 41 },
		{ 435, 163, 41 },
		{ 435, 164, 41 },
		{ 435, 165, 41 },
		{ 435, 166, 41 },
		{ 435, 167, 41 },
		{ 436, 123, 41 },
		{ 436, 124, 41 },
		{ 436, 125, 41 },
		{ 436, 126, 41 },
		{ 436, 127, 41 },
		{ 436, 128, 41 },
		{ 436, 129, 41 },
		{ 436, 130, 41 },
		{ 436, 131, 41 },
		{ 436, 132, 41 },
		{ 436, 133, 41 },
		{ 436, 134, 41 },
		{ 436, 135, 41 },
		{ 436, 136, 41 },
		{ 436, 137, 41 },
		{ 436, 138, 41 },
		{ 436, 139, 41 },
		{ 436, 140, 41 },
		{ 436, 141, 41 },
		{ 436, 142, 41 },
		{ 436, 143, 41 },
		{ 436, 144, 41 },
		{ 436, 145, 41 },
		{ 436, 146, 41 },
		{ 436, 147, 41 },
		{ 436, 148, 41 },
		{ 436, 149, 41 },
		{ 436, 150, 41 },
		{ 436, 151, 41 },
		{ 436, 152, 41 },
		{ 436, 153, 41 },
		{ 436, 154, 41 },
		{ 436, 155, 41 },
		{ 436, 156, 41 },
		{ 436, 157, 41 },
		{ 436, 158, 41 },
		{ 436, 159, 41 },
		{ 436, 160, 41 },
		{ 436, 161, 41 },
		{ 436, 162, 41 },
		{ 436, 163, 41 },
		{ 436, 164, 41 },
		{ 436, 165, 41 },
		{ 436, 166, 41 },
		{ 436, 167, 41 },
		{ 437, 123, 41 },
		{ 437, 124, 41 },
		{ 437, 125, 41 },
		{ 437, 126, 41 },
		{ 437, 127, 41 },
		{ 437, 128, 41 },
		{ 437, 129, 41 },
		{ 437, 130, 41 },
		{ 437, 131, 41 },
		{ 437, 132, 41 },
		{ 437, 133, 41 },
		{ 437, 134, 41 },
		{ 437, 135, 41 },
		{ 437, 136, 41 },
		{ 437, 137, 41 },
		{ 437, 138, 41 },
		{ 437, 139, 41 },
		{ 437, 140, 41 },
		{ 437, 141, 41 },
		{ 437, 142, 41 },
		{ 437, 143, 41 },
		{ 437, 144, 41 },
		{ 437, 145, 41 },
		{ 437, 146, 41 },
		{ 437, 147, 41 },
		{ 437, 148, 41 },
		{ 437, 149, 41 },
		{ 437, 150, 41 },
		{ 437, 151, 41 },
		{ 437, 152, 41 },
		{ 437, 153, 41 },
		{ 437, 154, 41 },
		{ 437, 155, 41 },
		{ 437, 156, 41 },
		{ 437, 157, 41 },
		{ 437, 158, 41 },
		{ 437, 159, 41 },
		{ 437, 160, 41 },
		{ 437, 161, 41 },
		{ 437, 162, 41 },
		{ 437, 163, 41 },
		{ 437, 164, 41 },
		{ 437, 165, 41 },
		{ 437, 166, 41 },
		{ 438, 122, 41 },
		{ 438, 123, 41 },
		{ 438, 124, 41 },
		{ 438, 125, 41 },
		{ 438, 126, 41 },
		{ 438, 127, 41 },
		{ 438, 128, 41 },
		{ 438, 129, 41 },
		{ 438, 130, 41 },
		{ 438, 131, 41 },
		{ 438, 132, 41 },
		{ 438, 133, 41 },
		{ 438, 134, 41 },
		{ 438, 135, 41 },
		{ 438, 136, 41 },
		{ 438, 137, 41 },
		{ 438, 138, 41 },
		{ 438, 139, 41 },
		{ 438, 140, 41 },
		{ 438, 141, 41 },
		{ 438, 142, 41 },
		{ 438, 143, 41 },
		{ 438, 144, 41 },
		{ 438, 145, 41 },
		{ 438, 146, 41 },
		{ 438, 147, 41 },
		{ 438, 148, 41 },
		{ 438, 149, 41 },
		{ 438, 150, 41 },
		{ 438, 151, 41 },
		{ 438, 152, 41 },
		{ 438, 153, 41 },
		{ 438, 154, 41 },
		{ 438, 155, 41 },
		{ 438, 156, 41 },
		{ 438, 157, 41 },
		{ 438, 158, 41 },
		{ 438, 159, 41 },
		{ 438, 160, 41 },
		{ 438, 161, 41 },
		{ 438, 162, 41 },
		{ 438, 163, 41 },
		{ 438, 164, 41 },
		{ 438, 165, 41 },
		{ 438, 166, 41 },
		{ 439, 121, 41 },
		{ 439, 122, 41 },
		{ 439, 123, 41 },
		{ 439, 124, 41 },
		{ 439, 125, 41 },
		{ 439, 126, 41 },
		{ 439, 127, 41 },
		{ 439, 128, 41 },
		{ 439, 129, 41 },
		{ 439, 130, 41 },
		{ 439, 131, 41 },
		{ 439, 132, 41 },
		{ 439, 133, 41 },
		{ 439, 134, 41 },
		{ 439, 135, 41 },
		{ 439, 136, 41 },
		{ 439, 137, 41 },
		{ 439, 138, 41 },
		{ 439, 139, 41 },
		{ 439, 140, 41 },
		{ 439, 141, 41 },
		{ 439, 142, 41 },
		{ 439, 143, 41 },
		{ 439, 144, 41 },
		{ 439, 145, 41 },
		{ 439, 146, 41 },
		{ 439, 147, 41 },
		{ 439, 148, 41 },
		{ 439, 149, 41 },
		{ 439, 150, 41 },
		{ 439, 151, 41 },
		{ 439, 152, 41 },
		{ 439, 153, 41 },
		{ 439, 154, 41 },
		{ 439, 155, 41 },
		{ 439, 156, 41 },
		{ 439, 157, 41 },
		{ 439, 158, 41 },
		{ 439, 159, 41 },
		{ 439, 160, 41 },
		{ 439, 161, 41 },
		{ 439, 162, 41 },
		{ 439, 163, 41 },
		{ 439, 164, 41 },
		{ 439, 165, 41 },
		{ 440, 120, 41 },
		{ 440, 121, 41 },
		{ 440, 122, 41 },
		{ 440, 123, 41 },
		{ 440, 124, 41 },
		{ 440, 125, 41 },
		{ 440, 126, 41 },
		{ 440, 127, 41 },
		{ 440, 128, 41 },
		{ 440, 129, 41 },
		{ 440, 130, 41 },
		{ 440, 131, 41 },
		{ 440, 132, 41 },
		{ 440, 133, 41 },
		{ 440, 134, 41 },
		{ 440, 135, 41 },
		{ 440, 136, 41 },
		{ 440, 137, 41 },
		{ 440, 138, 41 },
		{ 440, 139, 41 },
		{ 440, 140, 41 },
		{ 440, 141, 41 },
		{ 440, 142, 41 },
		{ 440, 143, 41 },
		{ 440, 144, 41 },
		{ 440, 145, 41 },
		{ 440, 146, 41 },
		{ 440, 147, 41 },
		{ 440, 148, 41 },
		{ 440, 149, 41 },
		{ 440, 150, 41 },
		{ 440, 151, 41 },
		{ 440, 152, 41 },
		{ 440, 153, 41 },
		{ 440, 154, 41 },
		{ 440, 155, 41 },
		{ 440, 156, 41 },
		{ 440, 157, 41 },
		{ 440, 158, 41 },
		{ 440, 159, 41 },
		{ 440, 160, 41 },
		{ 440, 161, 41 },
		{ 440, 162, 41 },
		{ 440, 163, 41 },
		{ 440, 164, 41 },
		{ 440, 165, 41 },
		{ 441, 120, 41 },
		{ 441, 121, 41 },
		{ 441, 122, 41 },
		{ 441, 123, 41 },
		{ 441, 124, 41 },
		{ 441, 125, 41 },
		{ 441, 126, 41 },
		{ 441, 127, 41 },
		{ 441, 128, 41 },
		{ 441, 129, 41 },
		{ 441, 130, 41 },
		{ 441, 131, 41 },
		{ 441, 132, 41 },
		{ 441, 133, 41 },
		{ 441, 134, 41 },
		{ 441, 135, 41 },
		{ 441, 136, 41 },
		{ 441, 137, 41 },
		{ 441, 138, 41 },
		{ 441, 139, 41 },
		{ 441, 140, 41 },
		{ 441, 141, 41 },
		{ 441, 142, 41 },
		{ 441, 143, 41 },
		{ 441, 144, 41 },
		{ 441, 145, 41 },
		{ 441, 146, 41 },
		{ 441, 147, 41 },
		{ 441, 148, 41 },
		{ 441, 149, 41 },
		{ 441, 150, 41 },
		{ 441, 151, 41 },
		{ 441, 152, 41 },
		{ 441, 153, 41 },
		{ 441, 154, 41 },
		{ 441, 155, 41 },
		{ 441, 156, 41 },
		{ 441, 157, 41 },
		{ 441, 158, 41 },
		{ 441, 159, 41 },
		{ 441, 160, 41 },
		{ 441, 161, 41 },
		{ 441, 162, 41 },
		{ 441, 163, 41 },
		{ 441, 164, 41 },
		{ 442, 119, 41 },
		{ 442, 120, 41 },
		{ 442, 121, 41 },
		{ 442, 122, 41 },
		{ 442, 123, 41 },
		{ 442, 124, 41 },
		{ 442, 125, 41 },
		{ 442, 126, 41 },
		{ 442, 127, 41 },
		{ 442, 128, 41 },
		{ 442, 129, 41 },
		{ 442, 130, 41 },
		{ 442, 131, 41 },
		{ 442, 132, 41 },
		{ 442, 133, 41 },
		{ 442, 134, 41 },
		{ 442, 135, 41 },
		{ 442, 136, 41 },
		{ 442, 137, 41 },
		{ 442, 138, 41 },
		{ 442, 139, 41 },
		{ 442, 140, 41 },
		{ 442, 141, 41 },
		{ 442, 142, 41 },
		{ 442, 143, 41 },
		{ 442, 144, 41 },
		{ 442, 145, 41 },
		{ 442, 146, 41 },
		{ 442, 147, 41 },
		{ 442, 148, 41 },
		{ 442, 149, 41 },
		{ 442, 150, 41 },
		{ 442, 151, 41 },
		{ 442, 152, 41 },
		{ 442, 153, 41 },
		{ 442, 154, 41 },
		{ 442, 155, 41 },
		{ 442, 156, 41 },
		{ 442, 157, 41 },
		{ 442, 158, 41 },
		{ 442, 159, 41 },
		{ 442, 160, 41 },
		{ 442, 161, 41 },
		{ 442, 162, 41 },
		{ 442, 163, 41 },
		{ 442, 164, 41 },
		{ 443, 118, 41 },
		{ 443, 119, 41 },
		{ 443, 120, 41 },
		{ 443, 121, 41 },
		{ 443, 122, 41 },
		{ 443, 123, 41 },
		{ 443, 124, 41 },
		{ 443, 125, 41 },
		{ 443, 126, 41 },
		{ 443, 127, 41 },
		{ 443, 128, 41 },
		{ 443, 129, 41 },
		{ 443, 130, 41 },
		{ 443, 131, 41 },
		{ 443, 132, 41 },
		{ 443, 133, 41 },
		{ 443, 134, 41 },
		{ 443, 135, 41 },
		{ 443, 136, 41 },
		{ 443, 137, 41 },
		{ 443, 138, 41 },
		{ 443, 139, 41 },
		{ 443, 140, 41 },
		{ 443, 141, 41 },
		{ 443, 142, 41 },
		{ 443, 143, 41 },
		{ 443, 144, 41 },
		{ 443, 145, 41 },
		{ 443, 146, 41 },
		{ 443, 147, 41 },
		{ 443, 148, 41 },
		{ 443, 149, 41 },
		{ 443, 150, 41 },
		{ 443, 151, 41 },
		{ 443, 152, 41 },
		{ 443, 153, 41 },
		{ 443, 154, 41 },
		{ 443, 155, 41 },
		{ 443, 156, 41 },
		{ 443, 157, 41 },
		{ 443, 158, 41 },
		{ 443, 159, 41 },
		{ 443, 160, 41 },
		{ 443, 161, 41 },
		{ 443, 162, 41 },
		{ 443, 163, 41 },
		{ 444, 117, 41 },
		{ 444, 118, 41 },
		{ 444, 119, 41 },
		{ 444, 120, 41 },
		{ 444, 121, 41 },
		{ 444, 122, 41 },
		{ 444, 123, 41 },
		{ 444, 124, 41 },
		{ 444, 125, 41 },
		{ 444, 126, 41 },
		{ 444, 127, 41 },
		{ 444, 128, 41 },
		{ 444, 129, 41 },
		{ 444, 130, 41 },
		{ 444, 131, 41 },
		{ 444, 132, 41 },
		{ 444, 133, 41 },
		{ 444, 134, 41 },
		{ 444, 135, 41 },
		{ 444, 136, 41 },
		{ 444, 137, 41 },
		{ 444, 138, 41 },
		{ 444, 139, 41 },
		{ 444, 140, 41 },
		{ 444, 141, 41 },
		{ 444, 142, 41 },
		{ 444, 143, 41 },
		{ 444, 144, 41 },
		{ 444, 145, 41 },
		{ 444, 146, 41 },
		{ 444, 147, 41 },
		{ 444, 148, 41 },
		{ 444, 149, 41 },
		{ 444, 150, 41 },
		{ 444, 151, 41 },
		{ 444, 152, 41 },
		{ 444, 153, 41 },
		{ 444, 154, 41 },
		{ 444, 155, 41 },
		{ 444, 156, 41 },
		{ 444, 157, 41 },
		{ 444, 158, 41 },
		{ 444, 159, 41 },
		{ 444, 160, 41 },
		{ 444, 161, 41 },
		{ 444, 162, 41 },
		{ 444, 163, 41 },
		{ 445, 117, 41 },
		{ 445, 118, 41 },
		{ 445, 119, 41 },
		{ 445, 120, 41 },
		{ 445, 121, 41 },
		{ 445, 122, 41 },
		{ 445, 123, 41 },
		{ 445, 124, 41 },
		{ 445, 125, 41 },
		{ 445, 126, 41 },
		{ 445, 127, 41 },
		{ 445, 128, 41 },
		{ 445, 129, 41 },
		{ 445, 130, 41 },
		{ 445, 131, 41 },
		{ 445, 132, 41 },
		{ 445, 133, 41 },
		{ 445, 134, 41 },
		{ 445, 135, 41 },
		{ 445, 136, 41 },
		{ 445, 137, 41 },
		{ 445, 138, 41 },
		{ 445, 139, 41 },
		{ 445, 140, 41 },
		{ 445, 141, 41 },
		{ 445, 142, 41 },
		{ 445, 143, 41 },
		{ 445, 144, 41 },
		{ 445, 145, 41 },
		{ 445, 146, 41 },
		{ 445, 147, 41 },
		{ 445, 148, 41 },
		{ 445, 149, 41 },
		{ 445, 150, 41 },
		{ 445, 151, 41 },
		{ 445, 152, 41 },
		{ 445, 153, 41 },
		{ 445, 154, 41 },
		{ 445, 155, 41 },
		{ 445, 156, 41 },
		{ 445, 157, 41 },
		{ 445, 158, 41 },
		{ 445, 159, 41 },
		{ 445, 160, 41 },
		{ 445, 161, 41 },
		{ 445, 162, 41 },
		{ 446, 116, 41 },
		{ 446, 117, 41 },
		{ 446, 118, 41 },
		{ 446, 119, 41 },
		{ 446, 120, 41 },
		{ 446, 121, 41 },
		{ 446, 122, 41 },
		{ 446, 123, 41 },
		{ 446, 124, 41 },
		{ 446, 125, 41 },
		{ 446, 126, 41 },
		{ 446, 127, 41 },
		{ 446, 128, 41 },
		{ 446, 129, 41 },
		{ 446, 130, 41 },
		{ 446, 131, 41 },
		{ 446, 132, 41 },
		{ 446, 133, 41 },
		{ 446, 134, 41 },
		{ 446, 135, 41 },
		{ 446, 136, 41 },
		{ 446, 137, 41 },
		{ 446, 138, 41 },
		{ 446, 139, 41 },
		{ 446, 140, 41 },
		{ 446, 141, 41 },
		{ 446, 142, 41 },
		{ 446, 143, 41 },
		{ 446, 144, 41 },
		{ 446, 145, 41 },
		{ 446, 146, 41 },
		{ 446, 147, 41 },
		{ 446, 148, 41 },
		{ 446, 149, 41 },
		{ 446, 150, 41 },
		{ 446, 151, 41 },
		{ 446, 152, 41 },
		{ 446, 153, 41 },
		{ 446, 154, 41 },
		{ 446, 155, 41 },
		{ 446, 156, 41 },
		{ 446, 157, 41 },
		{ 446, 158, 41 },
		{ 446, 159, 41 },
		{ 446, 160, 41 },
		{ 446, 161, 41 },
		{ 446, 162, 41 },
		{ 447, 115, 41 },
		{ 447, 116, 41 },
		{ 447, 117, 41 },
		{ 447, 118, 41 },
		{ 447, 119, 41 },
		{ 447, 120, 41 },
		{ 447, 121, 41 },
		{ 447, 122, 41 },
		{ 447, 123, 41 },
		{ 447, 124, 41 },
		{ 447, 125, 41 },
		{ 447, 126, 41 },
		{ 447, 127, 41 },
		{ 447, 128, 41 },
		{ 447, 129, 41 },
		{ 447, 130, 41 },
		{ 447, 131, 41 },
		{ 447, 132, 41 },
		{ 447, 133, 41 },
		{ 447, 134, 41 },
		{ 447, 135, 41 },
		{ 447, 136, 41 },
		{ 447, 137, 41 },
		{ 447, 138, 41 },
		{ 447, 139, 41 },
		{ 447, 140, 41 },
		{ 447, 141, 41 },
		{ 447, 142, 41 },
		{ 447, 143, 41 },
		{ 447, 144, 41 },
		{ 447, 145, 41 },
		{ 447, 146, 41 },
		{ 447, 147, 41 },
		{ 447, 148, 41 },
		{ 447, 149, 41 },
		{ 447, 150, 41 },
		{ 447, 151, 41 },
		{ 447, 152, 41 },
		{ 447, 153, 41 },
		{ 447, 154, 41 },
		{ 447, 155, 41 },
		{ 447, 156, 41 },
		{ 447, 157, 41 },
		{ 447, 158, 41 },
		{ 447, 159, 41 },
		{ 447, 160, 41 },
		{ 447, 161, 41 },
		{ 448, 114, 41 },
		{ 448, 115, 41 },
		{ 448, 116, 41 },
		{ 448, 117, 41 },
		{ 448, 118, 41 },
		{ 448, 119, 41 },
		{ 448, 120, 41 },
		{ 448, 121, 41 },
		{ 448, 122, 41 },
		{ 448, 123, 41 },
		{ 448, 124, 41 },
		{ 448, 125, 41 },
		{ 448, 126, 41 },
		{ 448, 127, 41 },
		{ 448, 128, 41 },
		{ 448, 129, 41 },
		{ 448, 130, 41 },
		{ 448, 131, 41 },
		{ 448, 132, 41 },
		{ 448, 133, 41 },
		{ 448, 134, 41 },
		{ 448, 135, 41 },
		{ 448, 136, 41 },
		{ 448, 137, 41 },
		{ 448, 138, 41 },
		{ 448, 139, 41 },
		{ 448, 140, 41 },
		{ 448, 141, 41 },
		{ 448, 142, 41 },
		{ 448, 143, 41 },
		{ 448, 144, 41 },
		{ 448, 145, 41 },
		{ 448, 146, 41 },
		{ 448, 147, 41 },
		{ 448, 148, 41 },
		{ 448, 149, 41 },
		{ 448, 150, 41 },
		{ 448, 151, 41 },
		{ 448, 152, 41 },
		{ 448, 153, 41 },
		{ 448, 154, 41 },
		{ 448, 155, 41 },
		{ 448, 156, 41 },
		{ 448, 157, 41 },
		{ 448, 158, 41 },
		{ 448, 159, 41 },
		{ 448, 160, 41 },
		{ 448, 161, 41 },
		{ 449, 114, 41 },
		{ 449, 115, 41 },
		{ 449, 116, 41 },
		{ 449, 117, 41 },
		{ 449, 118, 41 },
		{ 449, 119, 41 },
		{ 449, 120, 41 },
		{ 449, 121, 41 },
		{ 449, 122, 41 },
		{ 449, 123, 41 },
		{ 449, 124, 41 },
		{ 449, 125, 41 },
		{ 449, 126, 41 },
		{ 449, 127, 41 },
		{ 449, 128, 41 },
		{ 449, 129, 41 },
		{ 449, 130, 41 },
		{ 449, 131, 41 },
		{ 449, 132, 41 },
		{ 449, 133, 41 },
		{ 449, 134, 41 },
		{ 449, 135, 41 },
		{ 449, 136, 41 },
		{ 449, 137, 41 },
		{ 449, 138, 41 },
		{ 449, 139, 41 },
		{ 449, 140, 41 },
		{ 449, 141, 41 },
		{ 449, 142, 41 },
		{ 449, 143, 41 },
		{ 449, 144, 41 },
		{ 449, 145, 41 },
		{ 449, 146, 41 },
		{ 449, 147, 41 },
		{ 449, 148, 41 },
		{ 449, 149, 41 },
		{ 449, 150, 41 },
		{ 449, 151, 41 },
		{ 449, 152, 41 },
		{ 449, 153, 41 },
		{ 449, 154, 41 },
		{ 449, 155, 41 },
		{ 449, 156, 41 },
		{ 449, 157, 41 },
		{ 449, 158, 41 },
		{ 449, 159, 41 },
		{ 449, 160, 41 },
		{ 450, 113, 41 },
		{ 450, 114, 41 },
		{ 450, 115, 41 },
		{ 450, 116, 41 },
		{ 450, 117, 41 },
		{ 450, 118, 41 },
		{ 450, 119, 41 },
		{ 450, 120, 41 },
		{ 450, 121, 41 },
		{ 450, 122, 41 },
		{ 450, 123, 41 },
		{ 450, 124, 41 },
		{ 450, 125, 41 },
		{ 450, 126, 41 },
		{ 450, 127, 41 },
		{ 450, 128, 41 },
		{ 450, 129, 41 },
		{ 450, 130, 41 },
		{ 450, 131, 41 },
		{ 450, 132, 41 },
		{ 450, 133, 41 },
		{ 450, 134, 41 },
		{ 450, 135, 41 },
		{ 450, 136, 41 },
		{ 450, 137, 41 },
		{ 450, 138, 41 },
		{ 450, 139, 41 },
		{ 450, 140, 41 },
		{ 450, 141, 41 },
		{ 450, 142, 41 },
		{ 450, 143, 41 },
		{ 450, 144, 41 },
		{ 450, 145, 41 },
		{ 450, 146, 41 },
		{ 450, 147, 41 },
		{ 450, 148, 41 },
		{ 450, 149, 41 },
		{ 450, 150, 41 },
		{ 450, 151, 41 },
		{ 450, 152, 41 },
		{ 450, 153, 41 },
		{ 450, 154, 41 },
		{ 450, 155, 41 },
		{ 450, 156, 41 },
		{ 450, 157, 41 },
		{ 450, 158, 41 },
		{ 450, 159, 41 },
		{ 450, 160, 41 },
		{ 451, 112, 41 },
		{ 451, 113, 41 },
		{ 451, 114, 41 },
		{ 451, 115, 41 },
		{ 451, 116, 41 },
		{ 451, 117, 41 },
		{ 451, 118, 41 },
		{ 451, 119, 41 },
		{ 451, 120, 41 },
		{ 451, 121, 41 },
		{ 451, 122, 41 },
		{ 451, 123, 41 },
		{ 451, 124, 41 },
		{ 451, 125, 41 },
		{ 451, 126, 41 },
		{ 451, 127, 41 },
		{ 451, 128, 41 },
		{ 451, 129, 41 },
		{ 451, 130, 41 },
		{ 451, 131, 41 },
		{ 451, 132, 41 },
		{ 451, 133, 41 },
		{ 451, 134, 41 },
		{ 451, 135, 41 },
		{ 451, 136, 41 },
		{ 451, 137, 41 },
		{ 451, 138, 41 },
		{ 451, 139, 41 },
		{ 451, 140, 41 },
		{ 451, 141, 41 },
		{ 451, 142, 41 },
		{ 451, 143, 41 },
		{ 451, 144, 41 },
		{ 451, 145, 41 },
		{ 451, 146, 41 },
		{ 451, 147, 41 },
		{ 451, 148, 41 },
		{ 451, 149, 41 },
		{ 451, 150, 41 },
		{ 451, 151, 41 },
		{ 451, 152, 41 },
		{ 451, 153, 41 },
		{ 451, 154, 41 },
		{ 451, 155, 41 },
		{ 451, 156, 41 },
		{ 451, 157, 41 },
		{ 451, 158, 41 },
		{ 451, 159, 41 },
		{ 452, 111, 41 },
		{ 452, 112, 41 },
		{ 452, 113, 41 },
		{ 452, 114, 41 },
		{ 452, 115, 41 },
		{ 452, 116, 41 },
		{ 452, 117, 41 },
		{ 452, 118, 41 },
		{ 452, 119, 41 },
		{ 452, 120, 41 },
		{ 452, 121, 41 },
		{ 452, 122, 41 },
		{ 452, 123, 41 },
		{ 452, 124, 41 },
		{ 452, 125, 41 },
		{ 452, 126, 41 },
		{ 452, 127, 41 },
		{ 452, 128, 41 },
		{ 452, 129, 41 },
		{ 452, 130, 41 },
		{ 452, 131, 41 },
		{ 452, 132, 41 },
		{ 452, 133, 41 },
		{ 452, 134, 41 },
		{ 452, 135, 41 },
		{ 452, 136, 41 },
		{ 452, 137, 41 },
		{ 452, 138, 41 },
		{ 452, 139, 41 },
		{ 452, 140, 41 },
		{ 452, 141, 41 },
		{ 452, 142, 41 },
		{ 452, 143, 41 },
		{ 452, 144, 41 },
		{ 452, 145, 41 },
		{ 452, 146, 41 },
		{ 452, 147, 41 },
		{ 452, 148, 41 },
		{ 452, 149, 41 },
		{ 452, 150, 41 },
		{ 452, 151, 41 },
		{ 452, 152, 41 },
		{ 452, 153, 41 },
		{ 452, 154, 41 },
		{ 452, 155, 41 },
		{ 452, 156, 41 },
		{ 452, 157, 41 },
		{ 452, 158, 41 },
		{ 452, 159, 41 },
		{ 453, 111, 41 },
		{ 453, 112, 41 },
		{ 453, 113, 41 },
		{ 453, 114, 41 },
		{ 453, 115, 41 },
		{ 453, 116, 41 },
		{ 453, 117, 41 },
		{ 453, 118, 41 },
		{ 453, 119, 41 },
		{ 453, 120, 41 },
		{ 453, 121, 41 },
		{ 453, 122, 41 },
		{ 453, 123, 41 },
		{ 453, 124, 41 },
		{ 453, 125, 41 },
		{ 453, 126, 41 },
		{ 453, 127, 41 },
		{ 453, 128, 41 },
		{ 453, 129, 41 },
		{ 453, 130, 41 },
		{ 453, 131, 41 },
		{ 453, 132, 41 },
		{ 453, 133, 41 },
		{ 453, 134, 41 },
		{ 453, 135, 41 },
		{ 453, 136, 41 },
		{ 453, 137, 41 },
		{ 453, 138, 41 },
		{ 453, 139, 41 },
		{ 453, 140, 41 },
		{ 453, 141, 41 },
		{ 453, 142, 41 },
		{ 453, 143, 41 },
		{ 453, 144, 41 },
		{ 453, 145, 41 },
		{ 453, 146, 41 },
		{ 453, 147, 41 },
		{ 453, 148, 41 },
		{ 453, 149, 41 },
		{ 453, 150, 41 },
		{ 453, 151, 41 },
		{ 453, 152, 41 },
		{ 453, 153, 41 },
		{ 453, 154, 41 },
		{ 453, 155, 41 },
		{ 453, 156, 41 },
		{ 453, 157, 41 },
		{ 453, 158, 41 },
		{ 454, 110, 41 },
		{ 454, 111, 41 },
		{ 454, 112, 41 },
		{ 454, 113, 41 },
		{ 454, 114, 41 },
		{ 454, 115, 41 },
		{ 454, 116, 41 },
		{ 454, 117, 41 },
		{ 454, 118, 41 },
		{ 454, 119, 41 },
		{ 454, 120, 41 },
		{ 454, 121, 41 },
		{ 454, 122, 41 },
		{ 454, 123, 41 },
		{ 454, 124, 41 },
		{ 454, 125, 41 },
		{ 454, 126, 41 },
		{ 454, 127, 41 },
		{ 454, 128, 41 },
		{ 454, 129, 41 },
		{ 454, 130, 41 },
		{ 454, 131, 41 },
		{ 454, 132, 41 },
		{ 454, 133, 41 },
		{ 454, 134, 41 },
		{ 454, 135, 41 },
		{ 454, 136, 41 },
		{ 454, 137, 41 },
		{ 454, 138, 41 },
		{ 454, 139, 41 },
		{ 454, 140, 41 },
		{ 454, 141, 41 },
		{ 454, 142, 41 },
		{ 454, 143, 41 },
		{ 454, 144, 41 },
		{ 454, 145, 41 },
		{ 454, 146, 41 },
		{ 454, 147, 41 },
		{ 454, 148, 41 },
		{ 454, 149, 41 },
		{ 454, 150, 41 },
		{ 454, 151, 41 },
		{ 454, 152, 41 },
		{ 454, 153, 41 },
		{ 454, 154, 41 },
		{ 454, 155, 41 },
		{ 454, 156, 41 },
		{ 454, 157, 41 },
		{ 454, 158, 41 },
		{ 455, 109, 41 },
		{ 455, 110, 41 },
		{ 455, 111, 41 },
		{ 455, 112, 41 },
		{ 455, 113, 41 },
		{ 455, 114, 41 },
		{ 455, 115, 41 },
		{ 455, 116, 41 },
		{ 455, 117, 41 },
		{ 455, 118, 41 },
		{ 455, 119, 41 },
		{ 455, 120, 41 },
		{ 455, 121, 41 },
		{ 455, 122, 41 },
		{ 455, 123, 41 },
		{ 455, 124, 41 },
		{ 455, 125, 41 },
		{ 455, 126, 41 },
		{ 455, 127, 41 },
		{ 455, 128, 41 },
		{ 455, 129, 41 },
		{ 455, 130, 41 },
		{ 455, 131, 41 },
		{ 455, 132, 41 },
		{ 455, 133, 41 },
		{ 455, 134, 41 },
		{ 455, 135, 41 },
		{ 455, 136, 41 },
		{ 455, 137, 41 },
		{ 455, 138, 41 },
		{ 455, 139, 41 },
		{ 455, 140, 41 },
		{ 455, 141, 41 },
		{ 455, 142, 41 },
		{ 455, 143, 41 },
		{ 455, 144, 41 },
		{ 455, 145, 41 },
		{ 455, 146, 41 },
		{ 455, 147, 41 },
		{ 455, 148, 41 },
		{ 455, 149, 41 },
		{ 455, 150, 41 },
		{ 455, 151, 41 },
		{ 455, 152, 41 },
		{ 455, 153, 41 },
		{ 455, 154, 41 },
		{ 455, 155, 41 },
		{ 455, 156, 41 },
		{ 455, 157, 41 },
		{ 456, 108, 41 },
		{ 456, 109, 41 },
		{ 456, 110, 41 },
		{ 456, 111, 41 },
		{ 456, 112, 41 },
		{ 456, 113, 41 },
		{ 456, 114, 41 },
		{ 456, 115, 41 },
		{ 456, 116, 41 },
		{ 456, 117, 41 },
		{ 456, 118, 41 },
		{ 456, 119, 41 },
		{ 456, 120, 41 },
		{ 456, 121, 41 },
		{ 456, 122, 41 },
		{ 456, 123, 41 },
		{ 456, 124, 41 },
		{ 456, 125, 41 },
		{ 456, 126, 41 },
		{ 456, 127, 41 },
		{ 456, 128, 41 },
		{ 456, 129, 41 },
		{ 456, 130, 41 },
		{ 456, 131, 41 },
		{ 456, 132, 41 },
		{ 456, 133, 41 },
		{ 456, 134, 41 },
		{ 456, 135, 41 },
		{ 456, 136, 41 },
		{ 456, 137, 41 },
		{ 456, 138, 41 },
		{ 456, 139, 41 },
		{ 456, 140, 41 },
		{ 456, 141, 41 },
		{ 456, 142, 41 },
		{ 456, 143, 41 },
		{ 456, 144, 41 },
		{ 456, 145, 41 },
		{ 456, 146, 41 },
		{ 456, 147, 41 },
		{ 456, 148, 41 },
		{ 456, 149, 41 },
		{ 456, 150, 41 },
		{ 456, 151, 41 },
		{ 456, 152, 41 },
		{ 456, 153, 41 },
		{ 456, 154, 41 },
		{ 456, 155, 41 },
		{ 456, 156, 41 },
		{ 456, 157, 41 },
		{ 457, 108, 41 },
		{ 457, 109, 41 },
		{ 457, 110, 41 },
		{ 457, 111, 41 },
		{ 457, 112, 41 },
		{ 457, 113, 41 },
		{ 457, 114, 41 },
		{ 457, 115, 41 },
		{ 457, 116, 41 },
		{ 457, 117, 41 },
		{ 457, 118, 41 },
		{ 457, 119, 41 },
		{ 457, 120, 41 },
		{ 457, 121, 41 },
		{ 457, 122, 41 },
		{ 457, 123, 41 },
		{ 457, 124, 41 },
		{ 457, 125, 41 },
		{ 457, 126, 41 },
		{ 457, 127, 41 },
		{ 457, 128, 41 },
		{ 457, 129, 41 },
		{ 457, 130, 41 },
		{ 457, 131, 41 },
		{ 457, 132, 41 },
		{ 457, 133, 41 },
		{ 457, 134, 41 },
		{ 457, 135, 41 },
		{ 457, 136, 41 },
		{ 457, 137, 41 },
		{ 457, 138, 41 },
		{ 457, 139, 41 },
		{ 457, 140, 41 },
		{ 457, 141, 41 },
		{ 457, 142, 41 },
		{ 457, 143, 41 },
		{ 457, 144, 41 },
		{ 457, 145, 41 },
		{ 457, 146, 41 },
		{ 457, 147, 41 },
		{ 457, 148, 41 },
		{ 457, 149, 41 },
		{ 457, 150, 41 },
		{ 457, 151, 41 },
		{ 457, 152, 41 },
		{ 457, 153, 41 },
		{ 457, 154, 41 },
		{ 457, 155, 41 },
		{ 457, 156, 41 },
		{ 458, 107, 41 },
		{ 458, 108, 41 },
		{ 458, 109, 41 },
		{ 458, 110, 41 },
		{ 458, 111, 41 },
		{ 458, 112, 41 },
		{ 458, 113, 41 },
		{ 458, 114, 41 },
		{ 458, 115, 41 },
		{ 458, 116, 41 },
		{ 458, 117, 41 },
		{ 458, 118, 41 },
		{ 458, 119, 41 },
		{ 458, 120, 41 },
		{ 458, 121, 41 },
		{ 458, 122, 41 },
		{ 458, 123, 41 },
		{ 458, 124, 41 },
		{ 458, 125, 41 },
		{ 458, 126, 41 },
		{ 458, 127, 41 },
		{ 458, 128, 41 },
		{ 458, 129, 41 },
		{ 458, 130, 41 },
		{ 458, 131, 41 },
		{ 458, 132, 41 },
		{ 458, 133, 41 },
		{ 458, 134, 41 },
		{ 458, 135, 41 },
		{ 458, 136, 41 },
		{ 458, 137, 41 },
		{ 458, 138, 41 },
		{ 458, 139, 41 },
		{ 458, 140, 41 },
		{ 458, 141, 41 },
		{ 458, 142, 41 },
		{ 458, 143, 41 },
		{ 458, 144, 41 },
		{ 458, 145, 41 },
		{ 458, 146, 41 },
		{ 458, 147, 41 },
		{ 458, 148, 41 },
		{ 458, 149, 41 },
		{ 458, 150, 41 },
		{ 458, 151, 41 },
		{ 458, 152, 41 },
		{ 458, 153, 41 },
		{ 458, 154, 41 },
		{ 458, 155, 41 },
		{ 458, 156, 41 },
		{ 459, 106, 41 },
		{ 459, 107, 41 },
		{ 459, 108, 41 },
		{ 459, 109, 41 },
		{ 459, 110, 41 },
		{ 459, 111, 41 },
		{ 459, 112, 41 },
		{ 459, 113, 41 },
		{ 459, 114, 41 },
		{ 459, 115, 41 },
		{ 459, 116, 41 },
		{ 459, 117, 41 },
		{ 459, 118, 41 },
		{ 459, 119, 41 },
		{ 459, 120, 41 },
		{ 459, 121, 41 },
		{ 459, 122, 41 },
		{ 459, 123, 41 },
		{ 459, 124, 41 },
		{ 459, 125, 41 },
		{ 459, 126, 41 },
		{ 459, 127, 41 },
		{ 459, 128, 41 },
		{ 459, 129, 41 },
		{ 459, 130, 41 },
		{ 459, 131, 41 },
		{ 459, 132, 41 },
		{ 459, 133, 41 },
		{ 459, 134, 41 },
		{ 459, 135, 41 },
		{ 459, 136, 41 },
		{ 459, 137, 41 },
		{ 459, 138, 41 },
		{ 459, 139, 41 },
		{ 459, 140, 41 },
		{ 459, 141, 41 },
		{ 459, 142, 41 },
		{ 459, 143, 41 },
		{ 459, 144, 41 },
		{ 459, 145, 41 },
		{ 459, 146, 41 },
		{ 459, 147, 41 },
		{ 459, 148, 41 },
		{ 459, 149, 41 },
		{ 459, 150, 41 },
		{ 459, 151, 41 },
		{ 459, 152, 41 },
		{ 459, 153, 41 },
		{ 459, 154, 41 },
		{ 459, 155, 41 },
		{ 460, 105, 41 },
		{ 460, 106, 41 },
		{ 460, 107, 41 },
		{ 460, 108, 41 },
		{ 460, 109, 41 },
		{ 460, 110, 41 },
		{ 460, 111, 41 },
		{ 460, 112, 41 },
		{ 460, 113, 41 },
		{ 460, 114, 41 },
		{ 460, 115, 41 },
		{ 460, 116, 41 },
		{ 460, 117, 41 },
		{ 460, 118, 41 },
		{ 460, 119, 41 },
		{ 460, 120, 41 },
		{ 460, 121, 41 },
		{ 460, 122, 41 },
		{ 460, 123, 41 },
		{ 460, 124, 41 },
		{ 460, 125, 41 },
		{ 460, 126, 41 },
		{ 460, 127, 41 },
		{ 460, 128, 41 },
		{ 460, 129, 41 },
		{ 460, 130, 41 },
		{ 460, 131, 41 },
		{ 460, 132, 41 },
		{ 460, 133, 41 },
		{ 460, 134, 41 },
		{ 460, 135, 41 },
		{ 460, 136, 41 },
		{ 460, 137, 41 },
		{ 460, 138, 41 },
		{ 460, 139, 41 },
		{ 460, 140, 41 },
		{ 460, 141, 41 },
		{ 460, 142, 41 },
		{ 460, 143, 41 },
		{ 460, 144, 41 },
		{ 460, 145, 41 },
		{ 460, 146, 41 },
		{ 460, 147, 41 },
		{ 460, 148, 41 },
		{ 460, 149, 41 },
		{ 460, 150, 41 },
		{ 460, 151, 41 },
		{ 460, 152, 41 },
		{ 460, 153, 41 },
		{ 460, 154, 41 },
		{ 460, 155, 41 },
		{ 461, 105, 41 },
		{ 461, 106, 41 },
		{ 461, 107, 41 },
		{ 461, 108, 41 },
		{ 461, 109, 41 },
		{ 461, 110, 41 },
		{ 461, 111, 41 },
		{ 461, 112, 41 },
		{ 461, 113, 41 },
		{ 461, 114, 41 },
		{ 461, 115, 41 },
		{ 461, 116, 41 },
		{ 461, 117, 41 },
		{ 461, 118, 41 },
		{ 461, 119, 41 },
		{ 461, 120, 41 },
		{ 461, 121, 41 },
		{ 461, 122, 41 },
		{ 461, 123, 41 },
		{ 461, 124, 41 },
		{ 461, 125, 41 },
		{ 461, 126, 41 },
		{ 461, 127, 41 },
		{ 461, 128, 41 },
		{ 461, 129, 41 },
		{ 461, 130, 41 },
		{ 461, 131, 41 },
		{ 461, 132, 41 },
		{ 461, 133, 41 },
		{ 461, 134, 41 },
		{ 461, 135, 41 },
		{ 461, 136, 41 },
		{ 461, 137, 41 },
		{ 461, 138, 41 },
		{ 461, 139, 41 },
		{ 461, 140, 41 },
		{ 461, 141, 41 },
		{ 461, 142, 41 },
		{ 461, 143, 41 },
		{ 461, 144, 41 },
		{ 461, 145, 41 },
		{ 461, 146, 41 },
		{ 461, 147, 41 },
		{ 461, 148, 41 },
		{ 461, 149, 41 },
		{ 461, 150, 41 },
		{ 461, 151, 41 },
		{ 461, 152, 41 },
		{ 461, 153, 41 },
		{ 461, 154, 41 },
		{ 462, 104, 41 },
		{ 462, 105, 41 },
		{ 462, 106, 41 },
		{ 462, 107, 41 },
		{ 462, 108, 41 },
		{ 462, 109, 41 },
		{ 462, 110, 41 },
		{ 462, 111, 41 },
		{ 462, 112, 41 },
		{ 462, 113, 41 },
		{ 462, 114, 41 },
		{ 462, 115, 41 },
		{ 462, 116, 41 },
		{ 462, 117, 41 },
		{ 462, 118, 41 },
		{ 462, 119, 41 },
		{ 462, 120, 41 },
		{ 462, 121, 41 },
		{ 462, 122, 41 },
		{ 462, 123, 41 },
		{ 462, 124, 41 },
		{ 462, 125, 41 },
		{ 462, 126, 41 },
		{ 462, 127, 41 },
		{ 462, 128, 41 },
		{ 462, 129, 41 },
		{ 462, 130, 41 },
		{ 462, 131, 41 },
		{ 462, 132, 41 },
		{ 462, 133, 41 },
		{ 462, 134, 41 },
		{ 462, 135, 41 },
		{ 462, 136, 41 },
		{ 462, 137, 41 },
		{ 462, 138, 41 },
		{ 462, 139, 41 },
		{ 462, 140, 41 },
		{ 462, 141, 41 },
		{ 462, 142, 41 },
		{ 462, 143, 41 },
		{ 462, 144, 41 },
		{ 462, 145, 41 },
		{ 462, 146, 41 },
		{ 462, 147, 41 },
		{ 462, 148, 41 },
		{ 462, 149, 41 },
		{ 462, 150, 41 },
		{ 462, 151, 41 },
		{ 462, 152, 41 },
		{ 462, 153, 41 },
		{ 462, 154, 41 },
		{ 463, 103, 41 },
		{ 463, 104, 41 },
		{ 463, 105, 41 },
		{ 463, 106, 41 },
		{ 463, 107, 41 },
		{ 463, 108, 41 },
		{ 463, 109, 41 },
		{ 463, 110, 41 },
		{ 463, 111, 41 },
		{ 463, 112, 41 },
		{ 463, 113, 41 },
		{ 463, 114, 41 },
		{ 463, 115, 41 },
		{ 463, 116, 41 },
		{ 463, 117, 41 },
		{ 463, 118, 41 },
		{ 463, 119, 41 },
		{ 463, 120, 41 },
		{ 463, 121, 41 },
		{ 463, 122, 41 },
		{ 463, 123, 41 },
		{ 463, 124, 41 },
		{ 463, 125, 41 },
		{ 463, 126, 41 },
		{ 463, 127, 41 },
		{ 463, 128, 41 },
		{ 463, 129, 41 },
		{ 463, 130, 41 },
		{ 463, 131, 41 },
		{ 463, 132, 41 },
		{ 463, 133, 41 },
		{ 463, 134, 41 },
		{ 463, 135, 41 },
		{ 463, 136, 41 },
		{ 463, 137, 41 },
		{ 463, 138, 41 },
		{ 463, 139, 41 },
		{ 463, 140, 41 },
		{ 463, 141, 41 },
		{ 463, 142, 41 },
		{ 463, 143, 41 },
		{ 463, 144, 41 },
		{ 463, 145, 41 },
		{ 463, 146, 41 },
		{ 463, 147, 41 },
		{ 463, 148, 41 },
		{ 463, 149, 41 },
		{ 463, 150, 41 },
		{ 463, 151, 41 },
		{ 463, 152, 41 },
		{ 463, 153, 41 },
		{ 464, 102, 41 },
		{ 464, 103, 41 },
		{ 464, 104, 41 },
		{ 464, 105, 41 },
		{ 464, 106, 41 },
		{ 464, 107, 41 },
		{ 464, 108, 41 },
		{ 464, 109, 41 },
		{ 464, 110, 41 },
		{ 464, 111, 41 },
		{ 464, 112, 41 },
		{ 464, 113, 41 },
		{ 464, 114, 41 },
		{ 464, 115, 41 },
		{ 464, 116, 41 },
		{ 464, 117, 41 },
		{ 464, 118, 41 },
		{ 464, 119, 41 },
		{ 464, 120, 41 },
		{ 464, 121, 41 },
		{ 464, 122, 41 },
		{ 464, 123, 41 },
		{ 464, 124, 41 },
		{ 464, 125, 41 },
		{ 464, 126, 41 },
		{ 464, 127, 41 },
		{ 464, 128, 41 },
		{ 464, 129, 41 },
		{ 464, 130, 41 },
		{ 464, 131, 41 },
		{ 464, 132, 41 },
		{ 464, 133, 41 },
		{ 464, 134, 41 },
		{ 464, 135, 41 },
		{ 464, 136, 41 },
		{ 464, 137, 41 },
		{ 464, 138, 41 },
		{ 464, 139, 41 },
		{ 464, 140, 41 },
		{ 464, 141, 41 },
		{ 464, 142, 41 },
		{ 464, 143, 41 },
		{ 464, 144, 41 },
		{ 464, 145, 41 },
		{ 464, 146, 41 },
		{ 464, 147, 41 },
		{ 464, 148, 41 },
		{ 464, 149, 41 },
		{ 464, 150, 41 },
		{ 464, 151, 41 },
		{ 464, 152, 41 },
		{ 464, 153, 41 },
		{ 465, 102, 41 },
		{ 465, 103, 41 },
		{ 465, 104, 41 },
		{ 465, 105, 41 },
		{ 465, 106, 41 },
		{ 465, 107, 41 },
		{ 465, 108, 41 },
		{ 465, 109, 41 },
		{ 465, 110, 41 },
		{ 465, 111, 41 },
		{ 465, 112, 41 },
		{ 465, 113, 41 },
		{ 465, 114, 41 },
		{ 465, 115, 41 },
		{ 465, 116, 41 },
		{ 465, 117, 41 },
		{ 465, 118, 41 },
		{ 465, 119, 41 },
		{ 465, 120, 41 },
		{ 465, 121, 41 },
		{ 465, 122, 41 },
		{ 465, 123, 41 },
		{ 465, 124, 41 },
		{ 465, 125, 41 },
		{ 465, 126, 41 },
		{ 465, 127, 41 },
		{ 465, 128, 41 },
		{ 465, 129, 41 },
		{ 465, 130, 41 },
		{ 465, 131, 41 },
		{ 465, 132, 41 },
		{ 465, 133, 41 },
		{ 465, 134, 41 },
		{ 465, 135, 41 },
		{ 465, 136, 41 },
		{ 465, 137, 41 },
		{ 465, 138, 41 },
		{ 465, 139, 41 },
		{ 465, 140, 41 },
		{ 465, 141, 41 },
		{ 465, 142, 41 },
		{ 465, 143, 41 },
		{ 465, 144, 41 },
		{ 465, 145, 41 },
		{ 465, 146, 41 },
		{ 465, 147, 41 },
		{ 465, 148, 41 },
		{ 465, 149, 41 },
		{ 465, 150, 41 },
		{ 465, 151, 41 },
		{ 465, 152, 41 },
		{ 466, 101, 41 },
		{ 466, 102, 41 },
		{ 466, 103, 41 },
		{ 466, 104, 41 },
		{ 466, 105, 41 },
		{ 466, 106, 41 },
		{ 466, 107, 41 },
		{ 466, 108, 41 },
		{ 466, 109, 41 },
		{ 466, 110, 41 },
		{ 466, 111, 41 },
		{ 466, 112, 41 },
		{ 466, 113, 41 },
		{ 466, 114, 41 },
		{ 466, 115, 41 },
		{ 466, 116, 41 },
		{ 466, 117, 41 },
		{ 466, 118, 41 },
		{ 466, 119, 41 },
		{ 466, 120, 41 },
		{ 466, 121, 41 },
		{ 466, 122, 41 },
		{ 466, 123, 41 },
		{ 466, 124, 41 },
		{ 466, 125, 41 },
		{ 466, 126, 41 },
		{ 466, 127, 41 },
		{ 466, 128, 41 },
		{ 466, 129, 41 },
		{ 466, 130, 41 },
		{ 466, 131, 41 },
		{ 466, 132, 41 },
		{ 466, 133, 41 },
		{ 466, 134, 41 },
		{ 466, 135, 41 },
		{ 466, 136, 41 },
		{ 466, 137, 41 },
		{ 466, 138, 41 },
		{ 466, 139, 41 },
		{ 466, 140, 41 },
		{ 466, 141, 41 },
		{ 466, 142, 41 },
		{ 466, 143, 41 },
		{ 466, 144, 41 },
		{ 466, 145, 41 },
		{ 466, 146, 41 },
		{ 466, 147, 41 },
		{ 466, 148, 41 },
		{ 466, 149, 41 },
		{ 466, 150, 41 },
		{ 466, 151, 41 },
		{ 466, 152, 41 },
		{ 467, 100, 41 },
		{ 467, 101, 41 },
		{ 467, 102, 41 },
		{ 467, 103, 41 },
		{ 467, 104, 41 },
		{ 467, 105, 41 },
		{ 467, 106, 41 },
		{ 467, 107, 41 },
		{ 467, 108, 41 },
		{ 467, 109, 41 },
		{ 467, 110, 41 },
		{ 467, 111, 41 },
		{ 467, 112, 41 },
		{ 467, 113, 41 },
		{ 467, 114, 41 },
		{ 467, 115, 41 },
		{ 467, 116, 41 },
		{ 467, 117, 41 },
		{ 467, 118, 41 },
		{ 467, 119, 41 },
		{ 467, 120, 41 },
		{ 467, 121, 41 },
		{ 467, 122, 41 },
		{ 467, 123, 41 },
		{ 467, 124, 41 },
		{ 467, 125, 41 },
		{ 467, 126, 41 },
		{ 467, 127, 41 },
		{ 467, 128, 41 },
		{ 467, 129, 41 },
		{ 467, 130, 41 },
		{ 467, 131, 41 },
		{ 467, 132, 41 },
		{ 467, 133, 41 },
		{ 467, 134, 41 },
		{ 467, 135, 41 },
		{ 467, 136, 41 },
		{ 467, 137, 41 },
		{ 467, 138, 41 },
		{ 467, 139, 41 },
		{ 467, 140, 41 },
		{ 467, 141, 41 },
		{ 467, 142, 41 },
		{ 467, 143, 41 },
		{ 467, 144, 41 },
		{ 467, 145, 41 },
		{ 467, 146, 41 },
		{ 467, 147, 41 },
		{ 467, 148, 41 },
		{ 467, 149, 41 },
		{ 467, 150, 41 },
		{ 467, 151, 41 },
		{ 468, 99, 41 },
		{ 468, 100, 41 },
		{ 468, 101, 41 },
		{ 468, 102, 41 },
		{ 468, 103, 41 },
		{ 468, 104, 41 },
		{ 468, 105, 41 },
		{ 468, 106, 41 },
		{ 468, 107, 41 },
		{ 468, 108, 41 },
		{ 468, 109, 41 },
		{ 468, 110, 41 },
		{ 468, 111, 41 },
		{ 468, 112, 41 },
		{ 468, 113, 41 },
		{ 468, 114, 41 },
		{ 468, 115, 41 },
		{ 468, 116, 41 },
		{ 468, 117, 41 },
		{ 468, 118, 41 },
		{ 468, 119, 41 },
		{ 468, 120, 41 },
		{ 468, 121, 41 },
		{ 468, 122, 41 },
		{ 468, 123, 41 },
		{ 468, 124, 41 },
		{ 468, 125, 41 },
		{ 468, 126, 41 },
		{ 468, 127, 41 },
		{ 468, 128, 41 },
		{ 468, 129, 41 },
		{ 468, 130, 41 },
		{ 468, 131, 41 },
		{ 468, 132, 41 },
		{ 468, 133, 41 },
		{ 468, 134, 41 },
		{ 468, 135, 41 },
		{ 468, 136, 41 },
		{ 468, 137, 41 },
		{ 468, 138, 41 },
		{ 468, 139, 41 },
		{ 468, 140, 41 },
		{ 468, 141, 41 },
		{ 468, 142, 41 },
		{ 468, 143, 41 },
		{ 468, 144, 41 },
		{ 468, 145, 41 },
		{ 468, 146, 41 },
		{ 468, 147, 41 },
		{ 468, 148, 41 },
		{ 468, 149, 41 },
		{ 468, 150, 41 },
		{ 468, 151, 41 },
		{ 469, 99, 41 },
		{ 469, 100, 41 },
		{ 469, 101, 41 },
		{ 469, 102, 41 },
		{ 469, 103, 41 },
		{ 469, 104, 41 },
		{ 469, 105, 41 },
		{ 469, 106, 41 },
		{ 469, 107, 41 },
		{ 469, 108, 41 },
		{ 469, 109, 41 },
		{ 469, 110, 41 },
		{ 469, 111, 41 },
		{ 469, 112, 41 },
		{ 469, 113, 41 },
		{ 469, 114, 41 },
		{ 469, 115, 41 },
		{ 469, 116, 41 },
		{ 469, 117, 41 },
		{ 469, 118, 41 },
		{ 469, 119, 41 },
		{ 469, 120, 41 },
		{ 469, 121, 41 },
		{ 469, 122, 41 },
		{ 469, 123, 41 },
		{ 469, 124, 41 },
		{ 469, 125, 41 },
		{ 469, 126, 41 },
		{ 469, 127, 41 },
		{ 469, 128, 41 },
		{ 469, 129, 41 },
		{ 469, 130, 41 },
		{ 469, 131, 41 },
		{ 469, 132, 41 },
		{ 469, 133, 41 },
		{ 469, 134, 41 },
		{ 469, 135, 41 },
		{ 469, 136, 41 },
		{ 469, 137, 41 },
		{ 469, 138, 41 },
		{ 469, 139, 41 },
		{ 469, 140, 41 },
		{ 469, 141, 41 },
		{ 469, 142, 41 },
		{ 469, 143, 41 },
		{ 469, 144, 41 },
		{ 469, 145, 41 },
		{ 469, 146, 41 },
		{ 469, 147, 41 },
		{ 469, 148, 41 },
		{ 469, 149, 41 },
		{ 469, 150, 41 },
		{ 470, 98, 41 },
		{ 470, 99, 41 },
		{ 470, 100, 41 },
		{ 470, 101, 41 },
		{ 470, 102, 41 },
		{ 470, 103, 41 },
		{ 470, 104, 41 },
		{ 470, 105, 41 },
		{ 470, 106, 41 },
		{ 470, 107, 41 },
		{ 470, 108, 41 },
		{ 470, 109, 41 },
		{ 470, 110, 41 },
		{ 470, 111, 41 },
		{ 470, 112, 41 },
		{ 470, 113, 41 },
		{ 470, 114, 41 },
		{ 470, 115, 41 },
		{ 470, 116, 41 },
		{ 470, 117, 41 },
		{ 470, 118, 41 },
		{ 470, 119, 41 },
		{ 470, 120, 41 },
		{ 470, 121, 41 },
		{ 470, 122, 41 },
		{ 470, 123, 41 },
		{ 470, 124, 41 },
		{ 470, 125, 41 },
		{ 470, 126, 41 },
		{ 470, 127, 41 },
		{ 470, 128, 41 },
		{ 470, 129, 41 },
		{ 470, 130, 41 },
		{ 470, 131, 41 },
		{ 470, 132, 41 },
		{ 470, 133, 41 },
		{ 470, 134, 41 },
		{ 470, 135, 41 },
		{ 470, 136, 41 },
		{ 470, 137, 41 },
		{ 470, 138, 41 },
		{ 470, 139, 41 },
		{ 470, 140, 41 },
		{ 470, 141, 41 },
		{ 470, 142, 41 },
		{ 470, 143, 41 },
		{ 470, 144, 41 },
		{ 470, 145, 41 },
		{ 470, 146, 41 },
		{ 470, 147, 41 },
		{ 470, 148, 41 },
		{ 470, 149, 41 },
		{ 470, 150, 41 },
		{ 471, 97, 41 },
		{ 471, 98, 41 },
		{ 471, 99, 41 },
		{ 471, 100, 41 },
		{ 471, 101, 41 },
		{ 471, 102, 41 },
		{ 471, 103, 41 },
		{ 471, 104, 41 },
		{ 471, 105, 41 },
		{ 471, 106, 41 },
		{ 471, 107, 41 },
		{ 471, 108, 41 },
		{ 471, 109, 41 },
		{ 471, 110, 41 },
		{ 471, 111, 41 },
		{ 471, 112, 41 },
		{ 471, 113, 41 },
		{ 471, 114, 41 },
		{ 471, 115, 41 },
		{ 471, 116, 41 },
		{ 471, 117, 41 },
		{ 471, 118, 41 },
		{ 471, 119, 41 },
		{ 471, 120, 41 },
		{ 471, 121, 41 },
		{ 471, 122, 41 },
		{ 471, 123, 41 },
		{ 471, 124, 41 },
		{ 471, 125, 41 },
		{ 471, 126, 41 },
		{ 471, 127, 41 },
		{ 471, 128, 41 },
		{ 471, 129, 41 },
		{ 471, 130, 41 },
		{ 471, 131, 41 },
		{ 471, 132, 41 },
		{ 471, 133, 41 },
		{ 471, 134, 41 },
		{ 471, 135, 41 },
		{ 471, 136, 41 },
		{ 471, 137, 41 },
		{ 471, 138, 41 },
		{ 471, 139, 41 },
		{ 471, 140, 41 },
		{ 471, 141, 41 },
		{ 471, 142, 41 },
		{ 471, 143, 41 },
		{ 471, 144, 41 },
		{ 471, 145, 41 },
		{ 471, 146, 41 },
		{ 471, 147, 41 },
		{ 471, 148, 41 },
		{ 471, 149, 41 },
		{ 472, 96, 41 },
		{ 472, 97, 41 },
		{ 472, 98, 41 },
		{ 472, 99, 41 },
		{ 472, 100, 41 },
		{ 472, 101, 41 },
		{ 472, 102, 41 },
		{ 472, 103, 41 },
		{ 472, 104, 41 },
		{ 472, 105, 41 },
		{ 472, 106, 41 },
		{ 472, 107, 41 },
		{ 472, 108, 41 },
		{ 472, 109, 41 },
		{ 472, 110, 41 },
		{ 472, 111, 41 },
		{ 472, 112, 41 },
		{ 472, 113, 41 },
		{ 472, 114, 41 },
		{ 472, 115, 41 },
		{ 472, 116, 41 },
		{ 472, 117, 41 },
		{ 472, 118, 41 },
		{ 472, 119, 41 },
		{ 472, 120, 41 },
		{ 472, 121, 41 },
		{ 472, 122, 41 },
		{ 472, 123, 41 },
		{ 472, 124, 41 },
		{ 472, 125, 41 },
		{ 472, 126, 41 },
		{ 472, 127, 41 },
		{ 472, 128, 41 },
		{ 472, 129, 41 },
		{ 472, 130, 41 },
		{ 472, 131, 41 },
		{ 472, 132, 41 },
		{ 472, 133, 41 },
		{ 472, 134, 41 },
		{ 472, 135, 41 },
		{ 472, 136, 41 },
		{ 472, 137, 41 },
		{ 472, 138, 41 },
		{ 472, 139, 41 },
		{ 472, 140, 41 },
		{ 472, 141, 41 },
		{ 472, 142, 41 },
		{ 472, 143, 41 },
		{ 472, 144, 41 },
		{ 472, 145, 41 },
		{ 472, 146, 41 },
		{ 472, 147, 41 },
		{ 472, 148, 41 },
		{ 472, 149, 41 },
		{ 473, 96, 41 },
		{ 473, 97, 41 },
		{ 473, 98, 41 },
		{ 473, 99, 41 },
		{ 473, 100, 41 },
		{ 473, 101, 41 },
		{ 473, 102, 41 },
		{ 473, 103, 41 },
		{ 473, 104, 41 },
		{ 473, 105, 41 },
		{ 473, 106, 41 },
		{ 473, 107, 41 },
		{ 473, 108, 41 },
		{ 473, 109, 41 },
		{ 473, 110, 41 },
		{ 473, 111, 41 },
		{ 473, 112, 41 },
		{ 473, 113, 41 },
		{ 473, 114, 41 },
		{ 473, 115, 41 },
		{ 473, 116, 41 },
		{ 473, 117, 41 },
		{ 473, 118, 41 },
		{ 473, 119, 41 },
		{ 473, 120, 41 },
		{ 473, 121, 41 },
		{ 473, 122, 41 },
		{ 473, 123, 41 },
		{ 473, 124, 41 },
		{ 473, 125, 41 },
		{ 473, 126, 41 },
		{ 473, 127, 41 },
		{ 473, 128, 41 },
		{ 473, 129, 41 },
		{ 473, 130, 41 },
		{ 473, 131, 41 },
		{ 473, 132, 41 },
		{ 473, 133, 41 },
		{ 473, 134, 41 },
		{ 473, 135, 41 },
		{ 473, 136, 41 },
		{ 473, 137, 41 },
		{ 473, 138, 41 },
		{ 473, 139, 41 },
		{ 473, 140, 41 },
		{ 473, 141, 41 },
		{ 473, 142, 41 },
		{ 473, 143, 41 },
		{ 473, 144, 41 },
		{ 473, 145, 41 },
		{ 473, 146, 41 },
		{ 473, 147, 41 },
		{ 473, 148, 41 },
		{ 474, 95, 41 },
		{ 474, 96, 41 },
		{ 474, 97, 41 },
		{ 474, 98, 41 },
		{ 474, 99, 41 },
		{ 474, 100, 41 },
		{ 474, 101, 41 },
		{ 474, 102, 41 },
		{ 474, 103, 41 },
		{ 474, 104, 41 },
		{ 474, 105, 41 },
		{ 474, 106, 41 },
		{ 474, 107, 41 },
		{ 474, 108, 41 },
		{ 474, 109, 41 },
		{ 474, 110, 41 },
		{ 474, 111, 41 },
		{ 474, 112, 41 },
		{ 474, 113, 41 },
		{ 474, 114, 41 },
		{ 474, 115, 41 },
		{ 474, 116, 41 },
		{ 474, 117, 41 },
		{ 474, 118, 41 },
		{ 474, 119, 41 },
		{ 474, 120, 41 },
		{ 474, 121, 41 },
		{ 474, 122, 41 },
		{ 474, 123, 41 },
		{ 474, 124, 41 },
		{ 474, 125, 41 },
		{ 474, 126, 41 },
		{ 474, 127, 41 },
		{ 474, 128, 41 },
		{ 474, 129, 41 },
		{ 474, 130, 41 },
		{ 474, 131, 41 },
		{ 474, 132, 41 },
		{ 474, 133, 41 },
		{ 474, 134, 41 },
		{ 474, 135, 41 },
		{ 474, 136, 41 },
		{ 474, 137, 41 },
		{ 474, 138, 41 },
		{ 474, 139, 41 },
		{ 474, 140, 41 },
		{ 474, 141, 41 },
		{ 474, 142, 41 },
		{ 474, 143, 41 },
		{ 474, 144, 41 },
		{ 474, 145, 41 },
		{ 474, 146, 41 },
		{ 474, 147, 41 },
		{ 474, 148, 41 },
		{ 475, 94, 41 },
		{ 475, 95, 41 },
		{ 475, 96, 41 },
		{ 475, 97, 41 },
		{ 475, 98, 41 },
		{ 475, 99, 41 },
		{ 475, 100, 41 },
		{ 475, 101, 41 },
		{ 475, 102, 41 },
		{ 475, 103, 41 },
		{ 475, 104, 41 },
		{ 475, 105, 41 },
		{ 475, 106, 41 },
		{ 475, 107, 41 },
		{ 475, 108, 41 },
		{ 475, 109, 41 },
		{ 475, 110, 41 },
		{ 475, 111, 41 },
		{ 475, 112, 41 },
		{ 475, 113, 41 },
		{ 475, 114, 41 },
		{ 475, 115, 41 },
		{ 475, 116, 41 },
		{ 475, 117, 41 },
		{ 475, 118, 41 },
		{ 475, 119, 41 },
		{ 475, 120, 41 },
		{ 475, 121, 41 },
		{ 475, 122, 41 },
		{ 475, 123, 41 },
		{ 475, 124, 41 },
		{ 475, 125, 41 },
		{ 475, 126, 41 },
		{ 475, 127, 41 },
		{ 475, 128, 41 },
		{ 475, 129, 41 },
		{ 475, 130, 41 },
		{ 475, 131, 41 },
		{ 475, 132, 41 },
		{ 475, 133, 41 },
		{ 475, 134, 41 },
		{ 475, 135, 41 },
		{ 475, 136, 41 },
		{ 475, 137, 41 },
		{ 475, 138, 41 },
		{ 475, 139, 41 },
		{ 475, 140, 41 },
		{ 475, 141, 41 },
		{ 475, 142, 41 },
		{ 475, 143, 41 },
		{ 475, 144, 41 },
		{ 475, 145, 41 },
		{ 475, 146, 41 },
		{ 475, 147, 41 },
		{ 476, 93, 41 },
		{ 476, 94, 41 },
		{ 476, 95, 41 },
		{ 476, 96, 41 },
		{ 476, 97, 41 },
		{ 476, 98, 41 },
		{ 476, 99, 41 },
		{ 476, 100, 41 },
		{ 476, 101, 41 },
		{ 476, 102, 41 },
		{ 476, 103, 41 },
		{ 476, 104, 41 },
		{ 476, 105, 41 },
		{ 476, 106, 41 },
		{ 476, 107, 41 },
		{ 476, 108, 41 },
		{ 476, 109, 41 },
		{ 476, 110, 41 },
		{ 476, 111, 41 },
		{ 476, 112, 41 },
		{ 476, 113, 41 },
		{ 476, 114, 41 },
		{ 476, 115, 41 },
		{ 476, 116, 41 },
		{ 476, 117, 41 },
		{ 476, 118, 41 },
		{ 476, 119, 41 },
		{ 476, 120, 41 },
		{ 476, 121, 41 },
		{ 476, 122, 41 },
		{ 476, 123, 41 },
		{ 476, 124, 41 },
		{ 476, 125, 41 },
		{ 476, 126, 41 },
		{ 476, 127, 41 },
		{ 476, 128, 41 },
		{ 476, 129, 41 },
		{ 476, 130, 41 },
		{ 476, 131, 41 },
		{ 476, 132, 41 },
		{ 476, 133, 41 },
		{ 476, 134, 41 },
		{ 476, 135, 41 },
		{ 476, 136, 41 },
		{ 476, 137, 41 },
		{ 476, 138, 41 },
		{ 476, 139, 41 },
		{ 476, 140, 41 },
		{ 476, 141, 41 },
		{ 476, 142, 41 },
		{ 476, 143, 41 },
		{ 476, 144, 41 },
		{ 476, 145, 41 },
		{ 476, 146, 41 },
		{ 476, 147, 41 },
		{ 477, 93, 41 },
		{ 477, 94, 41 },
		{ 477, 95, 41 },
		{ 477, 96, 41 },
		{ 477, 97, 41 },
		{ 477, 98, 41 },
		{ 477, 99, 41 },
		{ 477, 100, 41 },
		{ 477, 101, 41 },
		{ 477, 102, 41 },
		{ 477, 103, 41 },
		{ 477, 104, 41 },
		{ 477, 105, 41 },
		{ 477, 106, 41 },
		{ 477, 107, 41 },
		{ 477, 108, 41 },
		{ 477, 109, 41 },
		{ 477, 110, 41 },
		{ 477, 111, 41 },
		{ 477, 112, 41 },
		{ 477, 113, 41 },
		{ 477, 114, 41 },
		{ 477, 115, 41 },
		{ 477, 116, 41 },
		{ 477, 117, 41 },
		{ 477, 118, 41 },
		{ 477, 119, 41 },
		{ 477, 120, 41 },
		{ 477, 121, 41 },
		{ 477, 122, 41 },
		{ 477, 123, 41 },
		{ 477, 124, 41 },
		{ 477, 125, 41 },
		{ 477, 126, 41 },
		{ 477, 127, 41 },
		{ 477, 128, 41 },
		{ 477, 129, 41 },
		{ 477, 130, 41 },
		{ 477, 131, 41 },
		{ 477, 132, 41 },
		{ 477, 133, 41 },
		{ 477, 134, 41 },
		{ 477, 135, 41 },
		{ 477, 136, 41 },
		{ 477, 137, 41 },
		{ 477, 138, 41 },
		{ 477, 139, 41 },
		{ 477, 140, 41 },
		{ 477, 141, 41 },
		{ 477, 142, 41 },
		{ 477, 143, 41 },
		{ 477, 144, 41 },
		{ 477, 145, 41 },
		{ 477, 146, 41 },
		{ 478, 92, 41 },
		{ 478, 93, 41 },
		{ 478, 94, 41 },
		{ 478, 95, 41 },
		{ 478, 96, 41 },
		{ 478, 97, 41 },
		{ 478, 98, 41 },
		{ 478, 99, 41 },
		{ 478, 100, 41 },
		{ 478, 101, 41 },
		{ 478, 102, 41 },
		{ 478, 103, 41 },
		{ 478, 104, 41 },
		{ 478, 105, 41 },
		{ 478, 106, 41 },
		{ 478, 107, 41 },
		{ 478, 108, 41 },
		{ 478, 109, 41 },
		{ 478, 110, 41 },
		{ 478, 111, 41 },
		{ 478, 112, 41 },
		{ 478, 113, 41 },
		{ 478, 114, 41 },
		{ 478, 115, 41 },
		{ 478, 116, 41 },
		{ 478, 117, 41 },
		{ 478, 118, 41 },
		{ 478, 119, 41 },
		{ 478, 120, 41 },
		{ 478, 121, 41 },
		{ 478, 122, 41 },
		{ 478, 123, 41 },
		{ 478, 124, 41 },
		{ 478, 125, 41 },
		{ 478, 126, 41 },
		{ 478, 127, 41 },
		{ 478, 128, 41 },
		{ 478, 129, 41 },
		{ 478, 130, 41 },
		{ 478, 131, 41 },
		{ 478, 132, 41 },
		{ 478, 133, 41 },
		{ 478, 134, 41 },
		{ 478, 135, 41 },
		{ 478, 136, 41 },
		{ 478, 137, 41 },
		{ 478, 138, 41 },
		{ 478, 139, 41 },
		{ 478, 140, 41 },
		{ 478, 141, 41 },
		{ 478, 142, 41 },
		{ 478, 143, 41 },
		{ 478, 144, 41 },
		{ 478, 145, 41 },
		{ 478, 146, 41 },
		{ 479, 91, 41 },
		{ 479, 92, 41 },
		{ 479, 93, 41 },
		{ 479, 94, 41 },
		{ 479, 95, 41 },
		{ 479, 96, 41 },
		{ 479, 97, 41 },
		{ 479, 98, 41 },
		{ 479, 99, 41 },
		{ 479, 100, 41 },
		{ 479, 101, 41 },
		{ 479, 102, 41 },
		{ 479, 103, 41 },
		{ 479, 104, 41 },
		{ 479, 105, 41 },
		{ 479, 106, 41 },
		{ 479, 107, 41 },
		{ 479, 108, 41 },
		{ 479, 109, 41 },
		{ 479, 110, 41 },
		{ 479, 111, 41 },
		{ 479, 112, 41 },
		{ 479, 113, 41 },
		{ 479, 114, 41 },
		{ 479, 115, 41 },
		{ 479, 116, 41 },
		{ 479, 117, 41 },
		{ 479, 118, 41 },
		{ 479, 119, 41 },
		{ 479, 120, 41 },
		{ 479, 121, 41 },
		{ 479, 122, 41 },
		{ 479, 123, 41 },
		{ 479, 124, 41 },
		{ 479, 125, 41 },
		{ 479, 126, 41 },
		{ 479, 127, 41 },
		{ 479, 128, 41 },
		{ 479, 129, 41 },
		{ 479, 130, 41 },
		{ 479, 131, 41 },
		{ 479, 132, 41 },
		{ 479, 133, 41 },
		{ 479, 134, 41 },
		{ 479, 135, 41 },
		{ 479, 136, 41 },
		{ 479, 137, 41 },
		{ 479, 138, 41 },
		{ 479, 139, 41 },
		{ 479, 140, 41 },
		{ 479, 141, 41 },
		{ 479, 142, 41 },
		{ 479, 143, 41 },
		{ 479, 144, 41 },
		{ 479, 145, 41 },
		{ 480, 90, 41 },
		{ 480, 91, 41 },
		{ 480, 92, 41 },
		{ 480, 93, 41 },
		{ 480, 94, 41 },
		{ 480, 95, 41 },
		{ 480, 96, 41 },
		{ 480, 97, 41 },
		{ 480, 98, 41 },
		{ 480, 99, 41 },
		{ 480, 100, 41 },
		{ 480, 101, 41 },
		{ 480, 102, 41 },
		{ 480, 103, 41 },
		{ 480, 104, 41 },
		{ 480, 105, 41 },
		{ 480, 106, 41 },
		{ 480, 107, 41 },
		{ 480, 108, 41 },
		{ 480, 109, 41 },
		{ 480, 110, 41 },
		{ 480, 111, 41 },
		{ 480, 112, 41 },
		{ 480, 113, 41 },
		{ 480, 114, 41 },
		{ 480, 115, 41 },
		{ 480, 116, 41 },
		{ 480, 117, 41 },
		{ 480, 118, 41 },
		{ 480, 119, 41 },
		{ 480, 120, 41 },
		{ 480, 121, 41 },
		{ 480, 122, 41 },
		{ 480, 123, 41 },
		{ 480, 124, 41 },
		{ 480, 125, 41 },
		{ 480, 126, 41 },
		{ 480, 127, 41 },
		{ 480, 128, 41 },
		{ 480, 129, 41 },
		{ 480, 130, 41 },
		{ 480, 131, 41 },
		{ 480, 132, 41 },
		{ 480, 133, 41 },
		{ 480, 134, 41 },
		{ 480, 135, 41 },
		{ 480, 136, 41 },
		{ 480, 137, 41 },
		{ 480, 138, 41 },
		{ 480, 139, 41 },
		{ 480, 140, 41 },
		{ 480, 141, 41 },
		{ 480, 142, 41 },
		{ 480, 143, 41 },
		{ 480, 144, 41 },
		{ 480, 145, 41 },
		{ 481, 90, 41 },
		{ 481, 91, 41 },
		{ 481, 92, 41 },
		{ 481, 93, 41 },
		{ 481, 94, 41 },
		{ 481, 95, 41 },
		{ 481, 96, 41 },
		{ 481, 97, 41 },
		{ 481, 98, 41 },
		{ 481, 99, 41 },
		{ 481, 100, 41 },
		{ 481, 101, 41 },
		{ 481, 102, 41 },
		{ 481, 103, 41 },
		{ 481, 104, 41 },
		{ 481, 105, 41 },
		{ 481, 106, 41 },
		{ 481, 107, 41 },
		{ 481, 108, 41 },
		{ 481, 109, 41 },
		{ 481, 110, 41 },
		{ 481, 111, 41 },
		{ 481, 112, 41 },
		{ 481, 113, 41 },
		{ 481, 114, 41 },
		{ 481, 115, 41 },
		{ 481, 116, 41 },
		{ 481, 117, 41 },
		{ 481, 118, 41 },
		{ 481, 119, 41 },
		{ 481, 120, 41 },
		{ 481, 121, 41 },
		{ 481, 122, 41 },
		{ 481, 123, 41 },
		{ 481, 124, 41 },
		{ 481, 125, 41 },
		{ 481, 126, 41 },
		{ 481, 127, 41 },
		{ 481, 128, 41 },
		{ 481, 129, 41 },
		{ 481, 130, 41 },
		{ 481, 131, 41 },
		{ 481, 132, 41 },
		{ 481, 133, 41 },
		{ 481, 134, 41 },
		{ 481, 135, 41 },
		{ 481, 136, 41 },
		{ 481, 137, 41 },
		{ 481, 138, 41 },
		{ 481, 139, 41 },
		{ 481, 140, 41 },
		{ 481, 141, 41 },
		{ 481, 142, 41 },
		{ 481, 143, 41 },
		{ 481, 144, 41 },
		{ 482, 89, 41 },
		{ 482, 90, 41 },
		{ 482, 91, 41 },
		{ 482, 92, 41 },
		{ 482, 93, 41 },
		{ 482, 94, 41 },
		{ 482, 95, 41 },
		{ 482, 96, 41 },
		{ 482, 97, 41 },
		{ 482, 98, 41 },
		{ 482, 99, 41 },
		{ 482, 100, 41 },
		{ 482, 101, 41 },
		{ 482, 102, 41 },
		{ 482, 103, 41 },
		{ 482, 104, 41 },
		{ 482, 105, 41 },
		{ 482, 106, 41 },
		{ 482, 107, 41 },
		{ 482, 108, 41 },
		{ 482, 109, 41 },
		{ 482, 110, 41 },
		{ 482, 111, 41 },
		{ 482, 112, 41 },
		{ 482, 113, 41 },
		{ 482, 114, 41 },
		{ 482, 115, 41 },
		{ 482, 116, 41 },
		{ 482, 117, 41 },
		{ 482, 118, 41 },
		{ 482, 119, 41 },
		{ 482, 120, 41 },
		{ 482, 121, 41 },
		{ 482, 122, 41 },
		{ 482, 123, 41 },
		{ 482, 124, 41 },
		{ 482, 125, 41 },
		{ 482, 126, 41 },
		{ 482, 127, 41 },
		{ 482, 128, 41 },
		{ 482, 129, 41 },
		{ 482, 130, 41 },
		{ 482, 131, 41 },
		{ 482, 132, 41 },
		{ 482, 133, 41 },
		{ 482, 134, 41 },
		{ 482, 135, 41 },
		{ 482, 136, 41 },
		{ 482, 137, 41 },
		{ 482, 138, 41 },
		{ 482, 139, 41 },
		{ 482, 140, 41 },
		{ 482, 141, 41 },
		{ 482, 142, 41 },
		{ 482, 143, 41 },
		{ 482, 144, 41 },
		{ 483, 88, 41 },
		{ 483, 89, 41 },
		{ 483, 90, 41 },
		{ 483, 91, 41 },
		{ 483, 92, 41 },
		{ 483, 93, 41 },
		{ 483, 94, 41 },
		{ 483, 95, 41 },
		{ 483, 96, 41 },
		{ 483, 97, 41 },
		{ 483, 98, 41 },
		{ 483, 99, 41 },
		{ 483, 100, 41 },
		{ 483, 101, 41 },
		{ 483, 102, 41 },
		{ 483, 103, 41 },
		{ 483, 104, 41 },
		{ 483, 105, 41 },
		{ 483, 106, 41 },
		{ 483, 107, 41 },
		{ 483, 108, 41 },
		{ 483, 109, 41 },
		{ 483, 110, 41 },
		{ 483, 111, 41 },
		{ 483, 112, 41 },
		{ 483, 113, 41 },
		{ 483, 114, 41 },
		{ 483, 115, 41 },
		{ 483, 116, 41 },
		{ 483, 117, 41 },
		{ 483, 118, 41 },
		{ 483, 119, 41 },
		{ 483, 120, 41 },
		{ 483, 121, 41 },
		{ 483, 122, 41 },
		{ 483, 123, 41 },
		{ 483, 124, 41 },
		{ 483, 125, 41 },
		{ 483, 126, 41 },
		{ 483, 127, 41 },
		{ 483, 128, 41 },
		{ 483, 129, 41 },
		{ 483, 130, 41 },
		{ 483, 131, 41 },
		{ 483, 132, 41 },
		{ 483, 133, 41 },
		{ 483, 134, 41 },
		{ 483, 135, 41 },
		{ 483, 136, 41 },
		{ 483, 137, 41 },
		{ 483, 138, 41 },
		{ 483, 139, 41 },
		{ 483, 140, 41 },
		{ 483, 141, 41 },
		{ 483, 142, 41 },
		{ 483, 143, 41 },
		{ 484, 87, 41 },
		{ 484, 88, 41 },
		{ 484, 89, 41 },
		{ 484, 90, 41 },
		{ 484, 91, 41 },
		{ 484, 92, 41 },
		{ 484, 93, 41 },
		{ 484, 94, 41 },
		{ 484, 95, 41 },
		{ 484, 96, 41 },
		{ 484, 97, 41 },
		{ 484, 98, 41 },
		{ 484, 99, 41 },
		{ 484, 100, 41 },
		{ 484, 101, 41 },
		{ 484, 102, 41 },
		{ 484, 103, 41 },
		{ 484, 104, 41 },
		{ 484, 105, 41 },
		{ 484, 106, 41 },
		{ 484, 107, 41 },
		{ 484, 108, 41 },
		{ 484, 109, 41 },
		{ 484, 110, 41 },
		{ 484, 111, 41 },
		{ 484, 112, 41 },
		{ 484, 113, 41 },
		{ 484, 114, 41 },
		{ 484, 115, 41 },
		{ 484, 116, 41 },
		{ 484, 117, 41 },
		{ 484, 118, 41 },
		{ 484, 119, 41 },
		{ 484, 120, 41 },
		{ 484, 121, 41 },
		{ 484, 122, 41 },
		{ 484, 123, 41 },
		{ 484, 124, 41 },
		{ 484, 125, 41 },
		{ 484, 126, 41 },
		{ 484, 127, 41 },
		{ 484, 128, 41 },
		{ 484, 129, 41 },
		{ 484, 130, 41 },
		{ 484, 131, 41 },
		{ 484, 132, 41 },
		{ 484, 133, 41 },
		{ 484, 134, 41 },
		{ 484, 135, 41 },
		{ 484, 136, 41 },
		{ 484, 137, 41 },
		{ 484, 138, 41 },
		{ 484, 139, 41 },
		{ 484, 140, 41 },
		{ 484, 141, 41 },
		{ 484, 142, 41 },
		{ 484, 143, 41 },
		{ 485, 87, 41 },
		{ 485, 88, 41 },
		{ 485, 89, 41 },
		{ 485, 90, 41 },
		{ 485, 91, 41 },
		{ 485, 92, 41 },
		{ 485, 93, 41 },
		{ 485, 94, 41 },
		{ 485, 95, 41 },
		{ 485, 96, 41 },
		{ 485, 97, 41 },
		{ 485, 98, 41 },
		{ 485, 99, 41 },
		{ 485, 100, 41 },
		{ 485, 101, 41 },
		{ 485, 102, 41 },
		{ 485, 103, 41 },
		{ 485, 104, 41 },
		{ 485, 105, 41 },
		{ 485, 106, 41 },
		{ 485, 107, 41 },
		{ 485, 108, 41 },
		{ 485, 109, 41 },
		{ 485, 110, 41 },
		{ 485, 111, 41 },
		{ 485, 112, 41 },
		{ 485, 113, 41 },
		{ 485, 114, 41 },
		{ 485, 115, 41 },
		{ 485, 116, 41 },
		{ 485, 117, 41 },
		{ 485, 118, 41 },
		{ 485, 119, 41 },
		{ 485, 120, 41 },
		{ 485, 121, 41 },
		{ 485, 122, 41 },
		{ 485, 123, 41 },
		{ 485, 124, 41 },
		{ 485, 125, 41 },
		{ 485, 126, 41 },
		{ 485, 127, 41 },
		{ 485, 128, 41 },
		{ 485, 129, 41 },
		{ 485, 130, 41 },
		{ 485, 131, 41 },
		{ 485, 132, 41 },
		{ 485, 133, 41 },
		{ 485, 134, 41 },
		{ 485, 135, 41 },
		{ 485, 136, 41 },
		{ 485, 137, 41 },
		{ 485, 138, 41 },
		{ 485, 139, 41 },
		{ 485, 140, 41 },
		{ 485, 141, 41 },
		{ 485, 142, 41 },
		{ 486, 86, 41 },
		{ 486, 87, 41 },
		{ 486, 88, 41 },
		{ 486, 89, 41 },
		{ 486, 90, 41 },
		{ 486, 91, 41 },
		{ 486, 92, 41 },
		{ 486, 93, 41 },
		{ 486, 94, 41 },
		{ 486, 95, 41 },
		{ 486, 96, 41 },
		{ 486, 97, 41 },
		{ 486, 98, 41 },
		{ 486, 99, 41 },
		{ 486, 100, 41 },
		{ 486, 101, 41 },
		{ 486, 102, 41 },
		{ 486, 103, 41 },
		{ 486, 104, 41 },
		{ 486, 105, 41 },
		{ 486, 106, 41 },
		{ 486, 107, 41 },
		{ 486, 108, 41 },
		{ 486, 109, 41 },
		{ 486, 110, 41 },
		{ 486, 111, 41 },
		{ 486, 112, 41 },
		{ 486, 113, 41 },
		{ 486, 114, 41 },
		{ 486, 115, 41 },
		{ 486, 116, 41 },
		{ 486, 117, 41 },
		{ 486, 118, 41 },
		{ 486, 119, 41 },
		{ 486, 120, 41 },
		{ 486, 121, 41 },
		{ 486, 122, 41 },
		{ 486, 123, 41 },
		{ 486, 124, 41 },
		{ 486, 125, 41 },
		{ 486, 126, 41 },
		{ 486, 127, 41 },
		{ 486, 128, 41 },
		{ 486, 129, 41 },
		{ 486, 130, 41 },
		{ 486, 131, 41 },
		{ 486, 132, 41 },
		{ 486, 133, 41 },
		{ 486, 134, 41 },
		{ 486, 135, 41 },
		{ 486, 136, 41 },
		{ 486, 137, 41 },
		{ 486, 138, 41 },
		{ 486, 139, 41 },
		{ 486, 140, 41 },
		{ 486, 141, 41 },
		{ 486, 142, 41 },
		{ 487, 85, 41 },
		{ 487, 86, 41 },
		{ 487, 87, 41 },
		{ 487, 88, 41 },
		{ 487, 89, 41 },
		{ 487, 90, 41 },
		{ 487, 91, 41 },
		{ 487, 92, 41 },
		{ 487, 93, 41 },
		{ 487, 94, 41 },
		{ 487, 95, 41 },
		{ 487, 96, 41 },
		{ 487, 97, 41 },
		{ 487, 98, 41 },
		{ 487, 99, 41 },
		{ 487, 100, 41 },
		{ 487, 101, 41 },
		{ 487, 102, 41 },
		{ 487, 103, 41 },
		{ 487, 104, 41 },
		{ 487, 105, 41 },
		{ 487, 106, 41 },
		{ 487, 107, 41 },
		{ 487, 108, 41 },
		{ 487, 109, 41 },
		{ 487, 110, 41 },
		{ 487, 111, 41 },
		{ 487, 112, 41 },
		{ 487, 113, 41 },
		{ 487, 114, 41 },
		{ 487, 115, 41 },
		{ 487, 116, 41 },
		{ 487, 117, 41 },
		{ 487, 118, 41 },
		{ 487, 119, 41 },
		{ 487, 120, 41 },
		{ 487, 121, 41 },
		{ 487, 122, 41 },
		{ 487, 123, 41 },
		{ 487, 124, 41 },
		{ 487, 125, 41 },
		{ 487, 126, 41 },
		{ 487, 127, 41 },
		{ 487, 128, 41 },
		{ 487, 129, 41 },
		{ 487, 130, 41 },
		{ 487, 131, 41 },
		{ 487, 132, 41 },
		{ 487, 133, 41 },
		{ 487, 134, 41 },
		{ 487, 135, 41 },
		{ 487, 136, 41 },
		{ 487, 137, 41 },
		{ 487, 138, 41 },
		{ 487, 139, 41 },
		{ 487, 140, 41 },
		{ 487, 141, 41 },
		{ 488, 84, 41 },
		{ 488, 85, 41 },
		{ 488, 86, 41 },
		{ 488, 87, 41 },
		{ 488, 88, 41 },
		{ 488, 89, 41 },
		{ 488, 90, 41 },
		{ 488, 91, 41 },
		{ 488, 92, 41 },
		{ 488, 93, 41 },
		{ 488, 94, 41 },
		{ 488, 95, 41 },
		{ 488, 96, 41 },
		{ 488, 97, 41 },
		{ 488, 98, 41 },
		{ 488, 99, 41 },
		{ 488, 100, 41 },
		{ 488, 101, 41 },
		{ 488, 102, 41 },
		{ 488, 103, 41 },
		{ 488, 104, 41 },
		{ 488, 105, 41 },
		{ 488, 106, 41 },
		{ 488, 107, 41 },
		{ 488, 108, 41 },
		{ 488, 109, 41 },
		{ 488, 110, 41 },
		{ 488, 111, 41 },
		{ 488, 112, 41 },
		{ 488, 113, 41 },
		{ 488, 114, 41 },
		{ 488, 115, 41 },
		{ 488, 116, 41 },
		{ 488, 117, 41 },
		{ 488, 118, 41 },
		{ 488, 119, 41 },
		{ 488, 120, 41 },
		{ 488, 121, 41 },
		{ 488, 122, 41 },
		{ 488, 123, 41 },
		{ 488, 124, 41 },
		{ 488, 125, 41 },
		{ 488, 126, 41 },
		{ 488, 127, 41 },
		{ 488, 128, 41 },
		{ 488, 129, 41 },
		{ 488, 130, 41 },
		{ 488, 131, 41 },
		{ 488, 132, 41 },
		{ 488, 133, 41 },
		{ 488, 134, 41 },
		{ 488, 135, 41 },
		{ 488, 136, 41 },
		{ 488, 137, 41 },
		{ 488, 138, 41 },
		{ 488, 139, 41 },
		{ 488, 140, 41 },
		{ 488, 141, 41 },
		{ 489, 84, 41 },
		{ 489, 85, 41 },
		{ 489, 86, 41 },
		{ 489, 87, 41 },
		{ 489, 88, 41 },
		{ 489, 89, 41 },
		{ 489, 90, 41 },
		{ 489, 91, 41 },
		{ 489, 92, 41 },
		{ 489, 93, 41 },
		{ 489, 94, 41 },
		{ 489, 95, 41 },
		{ 489, 96, 41 },
		{ 489, 97, 41 },
		{ 489, 98, 41 },
		{ 489, 99, 41 },
		{ 489, 100, 41 },
		{ 489, 101, 41 },
		{ 489, 102, 41 },
		{ 489, 103, 41 },
		{ 489, 104, 41 },
		{ 489, 105, 41 },
		{ 489, 106, 41 },
		{ 489, 107, 41 },
		{ 489, 108, 41 },
		{ 489, 109, 41 },
		{ 489, 110, 41 },
		{ 489, 111, 41 },
		{ 489, 112, 41 },
		{ 489, 113, 41 },
		{ 489, 114, 41 },
		{ 489, 115, 41 },
		{ 489, 116, 41 },
		{ 489, 117, 41 },
		{ 489, 118, 41 },
		{ 489, 119, 41 },
		{ 489, 120, 41 },
		{ 489, 121, 41 },
		{ 489, 122, 41 },
		{ 489, 123, 41 },
		{ 489, 124, 41 },
		{ 489, 125, 41 },
		{ 489, 126, 41 },
		{ 489, 127, 41 },
		{ 489, 128, 41 },
		{ 489, 129, 41 },
		{ 489, 130, 41 },
		{ 489, 131, 41 },
		{ 489, 132, 41 },
		{ 489, 133, 41 },
		{ 489, 134, 41 },
		{ 489, 135, 41 },
		{ 489, 136, 41 },
		{ 489, 137, 41 },
		{ 489, 138, 41 },
		{ 489, 139, 41 },
		{ 489, 140, 41 },
		{ 490, 83, 41 },
		{ 490, 84, 41 },
		{ 490, 85, 41 },
		{ 490, 86, 41 },
		{ 490, 87, 41 },
		{ 490, 88, 41 },
		{ 490, 89, 41 },
		{ 490, 90, 41 },
		{ 490, 91, 41 },
		{ 490, 92, 41 },
		{ 490, 93, 41 },
		{ 490, 94, 41 },
		{ 490, 95, 41 },
		{ 490, 96, 41 },
		{ 490, 97, 41 },
		{ 490, 98, 41 },
		{ 490, 99, 41 },
		{ 490, 100, 41 },
		{ 490, 101, 41 },
		{ 490, 102, 41 },
		{ 490, 103, 41 },
		{ 490, 104, 41 },
		{ 490, 105, 41 },
		{ 490, 106, 41 },
		{ 490, 107, 41 },
		{ 490, 108, 41 },
		{ 490, 109, 41 },
		{ 490, 110, 41 },
		{ 490, 111, 41 },
		{ 490, 112, 41 },
		{ 490, 113, 41 },
		{ 490, 114, 41 },
		{ 490, 115, 41 },
		{ 490, 116, 41 },
		{ 490, 117, 41 },
		{ 490, 118, 41 },
		{ 490, 119, 41 },
		{ 490, 120, 41 },
		{ 490, 121, 41 },
		{ 490, 122, 41 },
		{ 490, 123, 41 },
		{ 490, 124, 41 },
		{ 490, 125, 41 },
		{ 490, 126, 41 },
		{ 490, 127, 41 },
		{ 490, 128, 41 },
		{ 490, 129, 41 },
		{ 490, 130, 41 },
		{ 490, 131, 41 },
		{ 490, 132, 41 },
		{ 490, 133, 41 },
		{ 490, 134, 41 },
		{ 490, 135, 41 },
		{ 490, 136, 41 },
		{ 490, 137, 41 },
		{ 490, 138, 41 },
		{ 490, 139, 41 },
		{ 490, 140, 41 },
		{ 491, 82, 41 },
		{ 491, 83, 41 },
		{ 491, 84, 41 },
		{ 491, 85, 41 },
		{ 491, 86, 41 },
		{ 491, 87, 41 },
		{ 491, 88, 41 },
		{ 491, 89, 41 },
		{ 491, 90, 41 },
		{ 491, 91, 41 },
		{ 491, 92, 41 },
		{ 491, 93, 41 },
		{ 491, 94, 41 },
		{ 491, 95, 41 },
		{ 491, 96, 41 },
		{ 491, 97, 41 },
		{ 491, 98, 41 },
		{ 491, 99, 41 },
		{ 491, 100, 41 },
		{ 491, 101, 41 },
		{ 491, 102, 41 },
		{ 491, 103, 41 },
		{ 491, 104, 41 },
		{ 491, 105, 41 },
		{ 491, 106, 41 },
		{ 491, 107, 41 },
		{ 491, 108, 41 },
		{ 491, 109, 41 },
		{ 491, 110, 41 },
		{ 491, 111, 41 },
		{ 491, 112, 41 },
		{ 491, 113, 41 },
		{ 491, 114, 41 },
		{ 491, 115, 41 },
		{ 491, 116, 41 },
		{ 491, 117, 41 },
		{ 491, 118, 41 },
		{ 491, 119, 41 },
		{ 491, 120, 41 },
		{ 491, 121, 41 },
		{ 491, 122, 41 },
		{ 491, 123, 41 },
		{ 491, 124, 41 },
		{ 491, 125, 41 },
		{ 491, 126, 41 },
		{ 491, 127, 41 },
		{ 491, 128, 41 },
		{ 491, 129, 41 },
		{ 491, 130, 41 },
		{ 491, 131, 41 },
		{ 491, 132, 41 },
		{ 491, 133, 41 },
		{ 491, 134, 41 },
		{ 491, 135, 41 },
		{ 491, 136, 41 },
		{ 491, 137, 41 },
		{ 491, 138, 41 },
		{ 491, 139, 41 },
		{ 492, 81, 41 },
		{ 492, 82, 41 },
		{ 492, 83, 41 },
		{ 492, 84, 41 },
		{ 492, 85, 41 },
		{ 492, 86, 41 },
		{ 492, 87, 41 },
		{ 492, 88, 41 },
		{ 492, 89, 41 },
		{ 492, 90, 41 },
		{ 492, 91, 41 },
		{ 492, 92, 41 },
		{ 492, 93, 41 },
		{ 492, 94, 41 },
		{ 492, 95, 41 },
		{ 492, 96, 41 },
		{ 492, 97, 41 },
		{ 492, 98, 41 },
		{ 492, 99, 41 },
		{ 492, 100, 41 },
		{ 492, 101, 41 },
		{ 492, 102, 41 },
		{ 492, 103, 41 },
		{ 492, 104, 41 },
		{ 492, 105, 41 },
		{ 492, 106, 41 },
		{ 492, 107, 41 },
		{ 492, 108, 41 },
		{ 492, 109, 41 },
		{ 492, 110, 41 },
		{ 492, 111, 41 },
		{ 492, 112, 41 },
		{ 492, 113, 41 },
		{ 492, 114, 41 },
		{ 492, 115, 41 },
		{ 492, 116, 41 },
		{ 492, 117, 41 },
		{ 492, 118, 41 },
		{ 492, 119, 41 },
		{ 492, 120, 41 },
		{ 492, 121, 41 },
		{ 492, 122, 41 },
		{ 492, 123, 41 },
		{ 492, 124, 41 },
		{ 492, 125, 41 },
		{ 492, 126, 41 },
		{ 492, 127, 41 },
		{ 492, 128, 41 },
		{ 492, 129, 41 },
		{ 492, 130, 41 },
		{ 492, 131, 41 },
		{ 492, 132, 41 },
		{ 492, 133, 41 },
		{ 492, 134, 41 },
		{ 492, 135, 41 },
		{ 492, 136, 41 },
		{ 492, 137, 41 },
		{ 492, 138, 41 },
		{ 492, 139, 41 },
		{ 493, 81, 41 },
		{ 493, 82, 41 },
		{ 493, 83, 41 },
		{ 493, 84, 41 },
		{ 493, 85, 41 },
		{ 493, 86, 41 },
		{ 493, 87, 41 },
		{ 493, 88, 41 },
		{ 493, 89, 41 },
		{ 493, 90, 41 },
		{ 493, 91, 41 },
		{ 493, 92, 41 },
		{ 493, 93, 41 },
		{ 493, 94, 41 },
		{ 493, 95, 41 },
		{ 493, 96, 41 },
		{ 493, 97, 41 },
		{ 493, 98, 41 },
		{ 493, 99, 41 },
		{ 493, 100, 41 },
		{ 493, 101, 41 },
		{ 493, 102, 41 },
		{ 493, 103, 41 },
		{ 493, 104, 41 },
		{ 493, 105, 41 },
		{ 493, 106, 41 },
		{ 493, 107, 41 },
		{ 493, 108, 41 },
		{ 493, 109, 41 },
		{ 493, 110, 41 },
		{ 493, 111, 41 },
		{ 493, 112, 41 },
		{ 493, 113, 41 },
		{ 493, 114, 41 },
		{ 493, 115, 41 },
		{ 493, 116, 41 },
		{ 493, 117, 41 },
		{ 493, 118, 41 },
		{ 493, 119, 41 },
		{ 493, 120, 41 },
		{ 493, 121, 41 },
		{ 493, 122, 41 },
		{ 493, 123, 41 },
		{ 493, 124, 41 },
		{ 493, 125, 41 },
		{ 493, 126, 41 },
		{ 493, 127, 41 },
		{ 493, 128, 41 },
		{ 493, 129, 41 },
		{ 493, 130, 41 },
		{ 493, 131, 41 },
		{ 493, 132, 41 },
		{ 493, 133, 41 },
		{ 493, 134, 41 },
		{ 493, 135, 41 },
		{ 493, 136, 41 },
		{ 493, 137, 41 },
		{ 493, 138, 41 },
		{ 494, 80, 41 },
		{ 494, 81, 41 },
		{ 494, 82, 41 },
		{ 494, 83, 41 },
		{ 494, 84, 41 },
		{ 494, 85, 41 },
		{ 494, 86, 41 },
		{ 494, 87, 41 },
		{ 494, 88, 41 },
		{ 494, 89, 41 },
		{ 494, 90, 41 },
		{ 494, 91, 41 },
		{ 494, 92, 41 },
		{ 494, 93, 41 },
		{ 494, 94, 41 },
		{ 494, 95, 41 },
		{ 494, 96, 41 },
		{ 494, 97, 41 },
		{ 494, 98, 41 },
		{ 494, 99, 41 },
		{ 494, 100, 41 },
		{ 494, 101, 41 },
		{ 494, 102, 41 },
		{ 494, 103, 41 },
		{ 494, 104, 41 },
		{ 494, 105, 41 },
		{ 494, 106, 41 },
		{ 494, 107, 41 },
		{ 494, 108, 41 },
		{ 494, 109, 41 },
		{ 494, 110, 41 },
		{ 494, 111, 41 },
		{ 494, 112, 41 },
		{ 494, 113, 41 },
		{ 494, 114, 41 },
		{ 494, 115, 41 },
		{ 494, 116, 41 },
		{ 494, 117, 41 },
		{ 494, 118, 41 },
		{ 494, 119, 41 },
		{ 494, 120, 41 },
		{ 494, 121, 41 },
		{ 494, 122, 41 },
		{ 494, 123, 41 },
		{ 494, 124, 41 },
		{ 494, 125, 41 },
		{ 494, 126, 41 },
		{ 494, 127, 41 },
		{ 494, 128, 41 },
		{ 494, 129, 41 },
		{ 494, 130, 41 },
		{ 494, 131, 41 },
		{ 494, 132, 41 },
		{ 494, 133, 41 },
		{ 494, 134, 41 },
		{ 494, 135, 41 },
		{ 494, 136, 41 },
		{ 494, 137, 41 },
		{ 494, 138, 41 },
		{ 495, 79, 41 },
		{ 495, 80, 41 },
		{ 495, 81, 41 },
		{ 495, 82, 41 },
		{ 495, 83, 41 },
		{ 495, 84, 41 },
		{ 495, 85, 41 },
		{ 495, 86, 41 },
		{ 495, 87, 41 },
		{ 495, 88, 41 },
		{ 495, 89, 41 },
		{ 495, 90, 41 },
		{ 495, 91, 41 },
		{ 495, 92, 41 },
		{ 495, 93, 41 },
		{ 495, 94, 41 },
		{ 495, 95, 41 },
		{ 495, 96, 41 },
		{ 495, 97, 41 },
		{ 495, 98, 41 },
		{ 495, 99, 41 },
		{ 495, 100, 41 },
		{ 495, 101, 41 },
		{ 495, 102, 41 },
		{ 495, 103, 41 },
		{ 495, 104, 41 },
		{ 495, 105, 41 },
		{ 495, 106, 41 },
		{ 495, 107, 41 },
		{ 495, 108, 41 },
		{ 495, 109, 41 },
		{ 495, 110, 41 },
		{ 495, 111, 41 },
		{ 495, 112, 41 },
		{ 495, 113, 41 },
		{ 495, 114, 41 },
		{ 495, 115, 41 },
		{ 495, 116, 41 },
		{ 495, 117, 41 },
		{ 495, 118, 41 },
		{ 495, 119, 41 },
		{ 495, 120, 41 },
		{ 495, 121, 41 },
		{ 495, 122, 41 },
		{ 495, 123, 41 },
		{ 495, 124, 41 },
		{ 495, 125, 41 },
		{ 495, 126, 41 },
		{ 495, 127, 41 },
		{ 495, 128, 41 },
		{ 495, 129, 41 },
		{ 495, 130, 41 },
		{ 495, 131, 41 },
		{ 495, 132, 41 },
		{ 495, 133, 41 },
		{ 495, 134, 41 },
		{ 495, 135, 41 },
		{ 495, 136, 41 },
		{ 495, 137, 41 },
		{ 496, 78, 41 },
		{ 496, 79, 41 },
		{ 496, 80, 41 },
		{ 496, 81, 41 },
		{ 496, 82, 41 },
		{ 496, 83, 41 },
		{ 496, 84, 41 },
		{ 496, 85, 41 },
		{ 496, 86, 41 },
		{ 496, 87, 41 },
		{ 496, 88, 41 },
		{ 496, 89, 41 },
		{ 496, 90, 41 },
		{ 496, 91, 41 },
		{ 496, 92, 41 },
		{ 496, 93, 41 },
		{ 496, 94, 41 },
		{ 496, 95, 41 },
		{ 496, 96, 41 },
		{ 496, 97, 41 },
		{ 496, 98, 41 },
		{ 496, 99, 41 },
		{ 496, 100, 41 },
		{ 496, 101, 41 },
		{ 496, 102, 41 },
		{ 496, 103, 41 },
		{ 496, 104, 41 },
		{ 496, 105, 41 },
		{ 496, 106, 41 },
		{ 496, 107, 41 },
		{ 496, 108, 41 },
		{ 496, 109, 41 },
		{ 496, 110, 41 },
		{ 496, 111, 41 },
		{ 496, 112, 41 },
		{ 496, 113, 41 },
		{ 496, 114, 41 },
		{ 496, 115, 41 },
		{ 496, 116, 41 },
		{ 496, 117, 41 },
		{ 496, 118, 41 },
		{ 496, 119, 41 },
		{ 496, 120, 41 },
		{ 496, 121, 41 },
		{ 496, 122, 41 },
		{ 496, 123, 41 },
		{ 496, 124, 41 },
		{ 496, 125, 41 },
		{ 496, 126, 41 },
		{ 496, 127, 41 },
		{ 496, 128, 41 },
		{ 496, 129, 41 },
		{ 496, 130, 41 },
		{ 496, 131, 41 },
		{ 496, 132, 41 },
		{ 496, 133, 41 },
		{ 496, 134, 41 },
		{ 496, 135, 41 },
		{ 496, 136, 41 },
		{ 496, 137, 41 },
		{ 497, 78, 41 },
		{ 497, 79, 41 },
		{ 497, 80, 41 },
		{ 497, 81, 41 },
		{ 497, 82, 41 },
		{ 497, 83, 41 },
		{ 497, 84, 41 },
		{ 497, 85, 41 },
		{ 497, 86, 41 },
		{ 497, 87, 41 },
		{ 497, 88, 41 },
		{ 497, 89, 41 },
		{ 497, 90, 41 },
		{ 497, 91, 41 },
		{ 497, 92, 41 },
		{ 497, 93, 41 },
		{ 497, 94, 41 },
		{ 497, 95, 41 },
		{ 497, 96, 41 },
		{ 497, 97, 41 },
		{ 497, 98, 41 },
		{ 497, 99, 41 },
		{ 497, 100, 41 },
		{ 497, 101, 41 },
		{ 497, 102, 41 },
		{ 497, 103, 41 },
		{ 497, 104, 41 },
		{ 497, 105, 41 },
		{ 497, 106, 41 },
		{ 497, 107, 41 },
		{ 497, 108, 41 },
		{ 497, 109, 41 },
		{ 497, 110, 41 },
		{ 497, 111, 41 },
		{ 497, 112, 41 },
		{ 497, 113, 41 },
		{ 497, 114, 41 },
		{ 497, 115, 41 },
		{ 497, 116, 41 },
		{ 497, 117, 41 },
		{ 497, 118, 41 },
		{ 497, 119, 41 },
		{ 497, 120, 41 },
		{ 497, 121, 41 },
		{ 497, 122, 41 },
		{ 497, 123, 41 },
		{ 497, 124, 41 },
		{ 497, 125, 41 },
		{ 497, 126, 41 },
		{ 497, 127, 41 },
		{ 497, 128, 41 },
		{ 497, 129, 41 },
		{ 497, 130, 41 },
		{ 497, 131, 41 },
		{ 497, 132, 41 },
		{ 497, 133, 41 },
		{ 497, 134, 41 },
		{ 497, 135, 41 },
		{ 497, 136, 41 },
		{ 498, 77, 41 },
		{ 498, 78, 41 },
		{ 498, 79, 41 },
		{ 498, 80, 41 },
		{ 498, 81, 41 },
		{ 498, 82, 41 },
		{ 498, 83, 41 },
		{ 498, 84, 41 },
		{ 498, 85, 41 },
		{ 498, 86, 41 },
		{ 498, 87, 41 },
		{ 498, 88, 41 },
		{ 498, 89, 41 },
		{ 498, 90, 41 },
		{ 498, 91, 41 },
		{ 498, 92, 41 },
		{ 498, 93, 41 },
		{ 498, 94, 41 },
		{ 498, 95, 41 },
		{ 498, 96, 41 },
		{ 498, 97, 41 },
		{ 498, 98, 41 },
		{ 498, 99, 41 },
		{ 498, 100, 41 },
		{ 498, 101, 41 },
		{ 498, 102, 41 },
		{ 498, 103, 41 },
		{ 498, 104, 41 },
		{ 498, 105, 41 },
		{ 498, 106, 41 },
		{ 498, 107, 41 },
		{ 498, 108, 41 },
		{ 498, 109, 41 },
		{ 498, 110, 41 },
		{ 498, 111, 41 },
		{ 498, 112, 41 },
		{ 498, 113, 41 },
		{ 498, 114, 41 },
		{ 498, 115, 41 },
		{ 498, 116, 41 },
		{ 498, 117, 41 },
		{ 498, 118, 41 },
		{ 498, 119, 41 },
		{ 498, 120, 41 },
		{ 498, 121, 41 },
		{ 498, 122, 41 },
		{ 498, 123, 41 },
		{ 498, 124, 41 },
		{ 498, 125, 41 },
		{ 498, 126, 41 },
		{ 498, 127, 41 },
		{ 498, 128, 41 },
		{ 498, 129, 41 },
		{ 498, 130, 41 },
		{ 498, 131, 41 },
		{ 498, 132, 41 },
		{ 498, 133, 41 },
		{ 498, 134, 41 },
		{ 498, 135, 41 },
		{ 498, 136, 41 },
		{ 499, 76, 41 },
		{ 499, 77, 41 },
		{ 499, 78, 41 },
		{ 499, 79, 41 },
		{ 499, 80, 41 },
		{ 499, 81, 41 },
		{ 499, 82, 41 },
		{ 499, 83, 41 },
		{ 499, 84, 41 },
		{ 499, 85, 41 },
		{ 499, 86, 41 },
		{ 499, 87, 41 },
		{ 499, 88, 41 },
		{ 499, 89, 41 },
		{ 499, 90, 41 },
		{ 499, 91, 41 },
		{ 499, 92, 41 },
		{ 499, 93, 41 },
		{ 499, 94, 41 },
		{ 499, 95, 41 },
		{ 499, 96, 41 },
		{ 499, 97, 41 },
		{ 499, 98, 41 },
		{ 499, 99, 41 },
		{ 499, 100, 41 },
		{ 499, 101, 41 },
		{ 499, 102, 41 },
		{ 499, 103, 41 },
		{ 499, 104, 41 },
		{ 499, 105, 41 },
		{ 499, 106, 41 },
		{ 499, 107, 41 },
		{ 499, 108, 41 },
		{ 499, 109, 41 },
		{ 499, 110, 41 },
		{ 499, 111, 41 },
		{ 499, 112, 41 },
		{ 499, 113, 41 },
		{ 499, 114, 41 },
		{ 499, 115, 41 },
		{ 499, 116, 41 },
		{ 499, 117, 41 },
		{ 499, 118, 41 },
		{ 499, 119, 41 },
		{ 499, 120, 41 },
		{ 499, 121, 41 },
		{ 499, 122, 41 },
		{ 499, 123, 41 },
		{ 499, 124, 41 },
		{ 499, 125, 41 },
		{ 499, 126, 41 },
		{ 499, 127, 41 },
		{ 499, 128, 41 },
		{ 499, 129, 41 },
		{ 499, 130, 41 },
		{ 499, 131, 41 },
		{ 499, 132, 41 },
		{ 499, 133, 41 },
		{ 499, 134, 41 },
		{ 499, 135, 41 },
		{ 500, 75, 41 },
		{ 500, 76, 41 },
		{ 500, 77, 41 },
		{ 500, 78, 41 },
		{ 500, 79, 41 },
		{ 500, 80, 41 },
		{ 500, 81, 41 },
		{ 500, 82, 41 },
		{ 500, 83, 41 },
		{ 500, 84, 41 },
		{ 500, 85, 41 },
		{ 500, 86, 41 },
		{ 500, 87, 41 },
		{ 500, 88, 41 },
		{ 500, 89, 41 },
		{ 500, 90, 41 },
		{ 500, 91, 41 },
		{ 500, 92, 41 },
		{ 500, 93, 41 },
		{ 500, 94, 41 },
		{ 500, 95, 41 },
		{ 500, 96, 41 },
		{ 500, 97, 41 },
		{ 500, 98, 41 },
		{ 500, 99, 41 },
		{ 500, 100, 41 },
		{ 500, 101, 41 },
		{ 500, 102, 41 },
		{ 500, 103, 41 },
		{ 500, 104, 41 },
		{ 500, 105, 41 },
		{ 500, 106, 41 },
		{ 500, 107, 41 },
		{ 500, 108, 41 },
		{ 500, 109, 41 },
		{ 500, 110, 41 },
		{ 500, 111, 41 },
		{ 500, 112, 41 },
		{ 500, 113, 41 },
		{ 500, 114, 41 },
		{ 500, 115, 41 },
		{ 500, 116, 41 },
		{ 500, 117, 41 },
		{ 500, 118, 41 },
		{ 500, 119, 41 },
		{ 500, 120, 41 },
		{ 500, 121, 41 },
		{ 500, 122, 41 },
		{ 500, 123, 41 },
		{ 500, 124, 41 },
		{ 500, 125, 41 },
		{ 500, 126, 41 },
		{ 500, 127, 41 },
		{ 500, 128, 41 },
		{ 500, 129, 41 },
		{ 500, 130, 41 },
		{ 500, 131, 41 },
		{ 500, 132, 41 },
		{ 500, 133, 41 },
		{ 500, 134, 41 },
		{ 500, 135, 41 },
		{ 501, 75, 41 },
		{ 501, 76, 41 },
		{ 501, 77, 41 },
		{ 501, 78, 41 },
		{ 501, 79, 41 },
		{ 501, 80, 41 },
		{ 501, 81, 41 },
		{ 501, 82, 41 },
		{ 501, 83, 41 },
		{ 501, 84, 41 },
		{ 501, 85, 41 },
		{ 501, 86, 41 },
		{ 501, 87, 41 },
		{ 501, 88, 41 },
		{ 501, 89, 41 },
		{ 501, 90, 41 },
		{ 501, 91, 41 },
		{ 501, 92, 41 },
		{ 501, 93, 41 },
		{ 501, 94, 41 },
		{ 501, 95, 41 },
		{ 501, 96, 41 },
		{ 501, 97, 41 },
		{ 501, 98, 41 },
		{ 501, 99, 41 },
		{ 501, 100, 41 },
		{ 501, 101, 41 },
		{ 501, 102, 41 },
		{ 501, 103, 41 },
		{ 501, 104, 41 },
		{ 501, 105, 41 },
		{ 501, 106, 41 },
		{ 501, 107, 41 },
		{ 501, 108, 41 },
		{ 501, 109, 41 },
		{ 501, 110, 41 },
		{ 501, 111, 41 },
		{ 501, 112, 41 },
		{ 501, 113, 41 },
		{ 501, 114, 41 },
		{ 501, 115, 41 },
		{ 501, 116, 41 },
		{ 501, 117, 41 },
		{ 501, 118, 41 },
		{ 501, 119, 41 },
		{ 501, 120, 41 },
		{ 501, 121, 41 },
		{ 501, 122, 41 },
		{ 501, 123, 41 },
		{ 501, 124, 41 },
		{ 501, 125, 41 },
		{ 501, 126, 41 },
		{ 501, 127, 41 },
		{ 501, 128, 41 },
		{ 501, 129, 41 },
		{ 501, 130, 41 },
		{ 501, 131, 41 },
		{ 501, 132, 41 },
		{ 501, 133, 41 },
		{ 501, 134, 41 },
		{ 502, 74, 41 },
		{ 502, 75, 41 },
		{ 502, 76, 41 },
		{ 502, 77, 41 },
		{ 502, 78, 41 },
		{ 502, 79, 41 },
		{ 502, 80, 41 },
		{ 502, 81, 41 },
		{ 502, 82, 41 },
		{ 502, 83, 41 },
		{ 502, 84, 41 },
		{ 502, 85, 41 },
		{ 502, 86, 41 },
		{ 502, 87, 41 },
		{ 502, 88, 41 },
		{ 502, 89, 41 },
		{ 502, 90, 41 },
		{ 502, 91, 41 },
		{ 502, 92, 41 },
		{ 502, 93, 41 },
		{ 502, 94, 41 },
		{ 502, 95, 41 },
		{ 502, 96, 41 },
		{ 502, 97, 41 },
		{ 502, 98, 41 },
		{ 502, 99, 41 },
		{ 502, 100, 41 },
		{ 502, 101, 41 },
		{ 502, 102, 41 },
		{ 502, 103, 41 },
		{ 502, 104, 41 },
		{ 502, 105, 41 },
		{ 502, 106, 41 },
		{ 502, 107, 41 },
		{ 502, 108, 41 },
		{ 502, 109, 41 },
		{ 502, 110, 41 },
		{ 502, 111, 41 },
		{ 502, 112, 41 },
		{ 502, 113, 41 },
		{ 502, 114, 41 },
		{ 502, 115, 41 },
		{ 502, 116, 41 },
		{ 502, 117, 41 },
		{ 502, 118, 41 },
		{ 502, 119, 41 },
		{ 502, 120, 41 },
		{ 502, 121, 41 },
		{ 502, 122, 41 },
		{ 502, 123, 41 },
		{ 502, 124, 41 },
		{ 502, 125, 41 },
		{ 502, 126, 41 },
		{ 502, 127, 41 },
		{ 502, 128, 41 },
		{ 502, 129, 41 },
		{ 502, 130, 41 },
		{ 502, 131, 41 },
		{ 502, 132, 41 },
		{ 502, 133, 41 },
		{ 502, 134, 41 },
		{ 503, 73, 41 },
		{ 503, 74, 41 },
		{ 503, 75, 41 },
		{ 503, 76, 41 },
		{ 503, 77, 41 },
		{ 503, 78, 41 },
		{ 503, 79, 41 },
		{ 503, 80, 41 },
		{ 503, 81, 41 },
		{ 503, 82, 41 },
		{ 503, 83, 41 },
		{ 503, 84, 41 },
		{ 503, 85, 41 },
		{ 503, 86, 41 },
		{ 503, 87, 41 },
		{ 503, 88, 41 },
		{ 503, 89, 41 },
		{ 503, 90, 41 },
		{ 503, 91, 41 },
		{ 503, 92, 41 },
		{ 503, 93, 41 },
		{ 503, 94, 41 },
		{ 503, 95, 41 },
		{ 503, 96, 41 },
		{ 503, 97, 41 },
		{ 503, 98, 41 },
		{ 503, 99, 41 },
		{ 503, 100, 41 },
		{ 503, 101, 41 },
		{ 503, 102, 41 },
		{ 503, 103, 41 },
		{ 503, 104, 41 },
		{ 503, 105, 41 },
		{ 503, 106, 41 },
		{ 503, 107, 41 },
		{ 503, 108, 41 },
		{ 503, 109, 41 },
		{ 503, 110, 41 },
		{ 503, 111, 41 },
		{ 503, 112, 41 },
		{ 503, 113, 41 },
		{ 503, 114, 41 },
		{ 503, 115, 41 },
		{ 503, 116, 41 },
		{ 503, 117, 41 },
		{ 503, 118, 41 },
		{ 503, 119, 41 },
		{ 503, 120, 41 },
		{ 503, 121, 41 },
		{ 503, 122, 41 },
		{ 503, 123, 41 },
		{ 503, 124, 41 },
		{ 503, 125, 41 },
		{ 503, 126, 41 },
		{ 503, 127, 41 },
		{ 503, 128, 41 },
		{ 503, 129, 41 },
		{ 503, 130, 41 },
		{ 503, 131, 41 },
		{ 503, 132, 41 },
		{ 503, 133, 41 },
		{ 504, 72, 41 },
		{ 504, 73, 41 },
		{ 504, 74, 41 },
		{ 504, 75, 41 },
		{ 504, 76, 41 },
		{ 504, 77, 41 },
		{ 504, 78, 41 },
		{ 504, 79, 41 },
		{ 504, 80, 41 },
		{ 504, 81, 41 },
		{ 504, 82, 41 },
		{ 504, 83, 41 },
		{ 504, 84, 41 },
		{ 504, 85, 41 },
		{ 504, 86, 41 },
		{ 504, 87, 41 },
		{ 504, 88, 41 },
		{ 504, 89, 41 },
		{ 504, 90, 41 },
		{ 504, 91, 41 },
		{ 504, 92, 41 },
		{ 504, 93, 41 },
		{ 504, 94, 41 },
		{ 504, 95, 41 },
		{ 504, 96, 41 },
		{ 504, 97, 41 },
		{ 504, 98, 41 },
		{ 504, 99, 41 },
		{ 504, 100, 41 },
		{ 504, 101, 41 },
		{ 504, 102, 41 },
		{ 504, 103, 41 },
		{ 504, 104, 41 },
		{ 504, 105, 41 },
		{ 504, 106, 41 },
		{ 504, 107, 41 },
		{ 504, 108, 41 },
		{ 504, 109, 41 },
		{ 504, 110, 41 },
		{ 504, 111, 41 },
		{ 504, 112, 41 },
		{ 504, 113, 41 },
		{ 504, 114, 41 },
		{ 504, 115, 41 },
		{ 504, 116, 41 },
		{ 504, 117, 41 },
		{ 504, 118, 41 },
		{ 504, 119, 41 },
		{ 504, 120, 41 },
		{ 504, 121, 41 },
		{ 504, 122, 41 },
		{ 504, 123, 41 },
		{ 504, 124, 41 },
		{ 504, 125, 41 },
		{ 504, 126, 41 },
		{ 504, 127, 41 },
		{ 504, 128, 41 },
		{ 504, 129, 41 },
		{ 504, 130, 41 },
		{ 504, 131, 41 },
		{ 504, 132, 41 },
		{ 504, 133, 41 },
		{ 505, 72, 41 },
		{ 505, 73, 41 },
		{ 505, 74, 41 },
		{ 505, 75, 41 },
		{ 505, 76, 41 },
		{ 505, 77, 41 },
		{ 505, 78, 41 },
		{ 505, 79, 41 },
		{ 505, 80, 41 },
		{ 505, 81, 41 },
		{ 505, 82, 41 },
		{ 505, 83, 41 },
		{ 505, 84, 41 },
		{ 505, 85, 41 },
		{ 505, 86, 41 },
		{ 505, 87, 41 },
		{ 505, 88, 41 },
		{ 505, 89, 41 },
		{ 505, 90, 41 },
		{ 505, 91, 41 },
		{ 505, 92, 41 },
		{ 505, 93, 41 },
		{ 505, 94, 41 },
		{ 505, 95, 41 },
		{ 505, 96, 41 },
		{ 505, 97, 41 },
		{ 505, 98, 41 },
		{ 505, 99, 41 },
		{ 505, 100, 41 },
		{ 505, 101, 41 },
		{ 505, 102, 41 },
		{ 505, 103, 41 },
		{ 505, 104, 41 },
		{ 505, 105, 41 },
		{ 505, 106, 41 },
		{ 505, 107, 41 },
		{ 505, 108, 41 },
		{ 505, 109, 41 },
		{ 505, 110, 41 },
		{ 505, 111, 41 },
		{ 505, 112, 41 },
		{ 505, 113, 41 },
		{ 505, 114, 41 },
		{ 505, 115, 41 },
		{ 505, 116, 41 },
		{ 505, 117, 41 },
		{ 505, 118, 41 },
		{ 505, 119, 41 },
		{ 505, 120, 41 },
		{ 505, 121, 41 },
		{ 505, 122, 41 },
		{ 505, 123, 41 },
		{ 505, 124, 41 },
		{ 505, 125, 41 },
		{ 505, 126, 41 },
		{ 505, 127, 41 },
		{ 505, 128, 41 },
		{ 505, 129, 41 },
		{ 505, 130, 41 },
		{ 505, 131, 41 },
		{ 505, 132, 41 },
		{ 506, 71, 41 },
		{ 506, 72, 41 },
		{ 506, 73, 41 },
		{ 506, 74, 41 },
		{ 506, 75, 41 },
		{ 506, 76, 41 },
		{ 506, 77, 41 },
		{ 506, 78, 41 },
		{ 506, 79, 41 },
		{ 506, 80, 41 },
		{ 506, 81, 41 },
		{ 506, 82, 41 },
		{ 506, 83, 41 },
		{ 506, 84, 41 },
		{ 506, 85, 41 },
		{ 506, 86, 41 },
		{ 506, 87, 41 },
		{ 506, 88, 41 },
		{ 506, 89, 41 },
		{ 506, 90, 41 },
		{ 506, 91, 41 },
		{ 506, 92, 41 },
		{ 506, 93, 41 },
		{ 506, 94, 41 },
		{ 506, 95, 41 },
		{ 506, 96, 41 },
		{ 506, 97, 41 },
		{ 506, 98, 41 },
		{ 506, 99, 41 },
		{ 506, 100, 41 },
		{ 506, 101, 41 },
		{ 506, 102, 41 },
		{ 506, 103, 41 },
		{ 506, 104, 41 },
		{ 506, 105, 41 },
		{ 506, 106, 41 },
		{ 506, 107, 41 },
		{ 506, 108, 41 },
		{ 506, 109, 41 },
		{ 506, 110, 41 },
		{ 506, 111, 41 },
		{ 506, 112, 41 },
		{ 506, 113, 41 },
		{ 506, 114, 41 },
		{ 506, 115, 41 },
		{ 506, 116, 41 },
		{ 506, 117, 41 },
		{ 506, 118, 41 },
		{ 506, 119, 41 },
		{ 506, 120, 41 },
		{ 506, 121, 41 },
		{ 506, 122, 41 },
		{ 506, 123, 41 },
		{ 506, 124, 41 },
		{ 506, 125, 41 },
		{ 506, 126, 41 },
		{ 506, 127, 41 },
		{ 506, 128, 41 },
		{ 506, 129, 41 },
		{ 506, 130, 41 },
		{ 506, 131, 41 },
		{ 506, 132, 41 },
		{ 507, 70, 41 },
		{ 507, 71, 41 },
		{ 507, 72, 41 },
		{ 507, 73, 41 },
		{ 507, 74, 41 },
		{ 507, 75, 41 },
		{ 507, 76, 41 },
		{ 507, 77, 41 },
		{ 507, 78, 41 },
		{ 507, 79, 41 },
		{ 507, 80, 41 },
		{ 507, 81, 41 },
		{ 507, 82, 41 },
		{ 507, 83, 41 },
		{ 507, 84, 41 },
		{ 507, 85, 41 },
		{ 507, 86, 41 },
		{ 507, 87, 41 },
		{ 507, 88, 41 },
		{ 507, 89, 41 },
		{ 507, 90, 41 },
		{ 507, 91, 41 },
		{ 507, 92, 41 },
		{ 507, 93, 41 },
		{ 507, 94, 41 },
		{ 507, 95, 41 },
		{ 507, 96, 41 },
		{ 507, 97, 41 },
		{ 507, 98, 41 },
		{ 507, 99, 41 },
		{ 507, 100, 41 },
		{ 507, 101, 41 },
		{ 507, 102, 41 },
		{ 507, 103, 41 },
		{ 507, 104, 41 },
		{ 507, 105, 41 },
		{ 507, 106, 41 },
		{ 507, 107, 41 },
		{ 507, 108, 41 },
		{ 507, 109, 41 },
		{ 507, 110, 41 },
		{ 507, 111, 41 },
		{ 507, 112, 41 },
		{ 507, 113, 41 },
		{ 507, 114, 41 },
		{ 507, 115, 41 },
		{ 507, 116, 41 },
		{ 507, 117, 41 },
		{ 507, 118, 41 },
		{ 507, 119, 41 },
		{ 507, 120, 41 },
		{ 507, 121, 41 },
		{ 507, 122, 41 },
		{ 507, 123, 41 },
		{ 507, 124, 41 },
		{ 507, 125, 41 },
		{ 507, 126, 41 },
		{ 507, 127, 41 },
		{ 507, 128, 41 },
		{ 507, 129, 41 },
		{ 507, 130, 41 },
		{ 507, 131, 41 },
		{ 508, 69, 41 },
		{ 508, 70, 41 },
		{ 508, 71, 41 },
		{ 508, 72, 41 },
		{ 508, 73, 41 },
		{ 508, 74, 41 },
		{ 508, 75, 41 },
		{ 508, 76, 41 },
		{ 508, 77, 41 },
		{ 508, 78, 41 },
		{ 508, 79, 41 },
		{ 508, 80, 41 },
		{ 508, 81, 41 },
		{ 508, 82, 41 },
		{ 508, 83, 41 },
		{ 508, 84, 41 },
		{ 508, 85, 41 },
		{ 508, 86, 41 },
		{ 508, 87, 41 },
		{ 508, 88, 41 },
		{ 508, 89, 41 },
		{ 508, 90, 41 },
		{ 508, 91, 41 },
		{ 508, 92, 41 },
		{ 508, 93, 41 },
		{ 508, 94, 41 },
		{ 508, 95, 41 },
		{ 508, 96, 41 },
		{ 508, 97, 41 },
		{ 508, 98, 41 },
		{ 508, 99, 41 },
		{ 508, 100, 41 },
		{ 508, 101, 41 },
		{ 508, 102, 41 },
		{ 508, 103, 41 },
		{ 508, 104, 41 },
		{ 508, 105, 41 },
		{ 508, 106, 41 },
		{ 508, 107, 41 },
		{ 508, 108, 41 },
		{ 508, 109, 41 },
		{ 508, 110, 41 },
		{ 508, 111, 41 },
		{ 508, 112, 41 },
		{ 508, 113, 41 },
		{ 508, 114, 41 },
		{ 508, 115, 41 },
		{ 508, 116, 41 },
		{ 508, 117, 41 },
		{ 508, 118, 41 },
		{ 508, 119, 41 },
		{ 508, 120, 41 },
		{ 508, 121, 41 },
		{ 508, 122, 41 },
		{ 508, 123, 41 },
		{ 508, 124, 41 },
		{ 508, 125, 41 },
		{ 508, 126, 41 },
		{ 508, 127, 41 },
		{ 508, 128, 41 },
		{ 508, 129, 41 },
		{ 508, 130, 41 },
		{ 508, 131, 41 },
		{ 509, 69, 41 },
		{ 509, 70, 41 },
		{ 509, 71, 41 },
		{ 509, 72, 41 },
		{ 509, 73, 41 },
		{ 509, 74, 41 },
		{ 509, 75, 41 },
		{ 509, 76, 41 },
		{ 509, 77, 41 },
		{ 509, 78, 41 },
		{ 509, 79, 41 },
		{ 509, 80, 41 },
		{ 509, 81, 41 },
		{ 509, 82, 41 },
		{ 509, 83, 41 },
		{ 509, 84, 41 },
		{ 509, 85, 41 },
		{ 509, 86, 41 },
		{ 509, 87, 41 },
		{ 509, 88, 41 },
		{ 509, 89, 41 },
		{ 509, 90, 41 },
		{ 509, 91, 41 },
		{ 509, 92, 41 },
		{ 509, 93, 41 },
		{ 509, 94, 41 },
		{ 509, 95, 41 },
		{ 509, 96, 41 },
		{ 509, 97, 41 },
		{ 509, 98, 41 },
		{ 509, 99, 41 },
		{ 509, 100, 41 },
		{ 509, 101, 41 },
		{ 509, 102, 41 },
		{ 509, 103, 41 },
		{ 509, 104, 41 },
		{ 509, 105, 41 },
		{ 509, 106, 41 },
		{ 509, 107, 41 },
		{ 509, 108, 41 },
		{ 509, 109, 41 },
		{ 509, 110, 41 },
		{ 509, 111, 41 },
		{ 509, 112, 41 },
		{ 509, 113, 41 },
		{ 509, 114, 41 },
		{ 509, 115, 41 },
		{ 509, 116, 41 },
		{ 509, 117, 41 },
		{ 509, 118, 41 },
		{ 509, 119, 41 },
		{ 509, 120, 41 },
		{ 509, 121, 41 },
		{ 509, 122, 41 },
		{ 509, 123, 41 },
		{ 509, 124, 41 },
		{ 509, 125, 41 },
		{ 509, 126, 41 },
		{ 509, 127, 41 },
		{ 509, 128, 41 },
		{ 509, 129, 41 },
		{ 509, 130, 41 },
		{ 510, 68, 41 },
		{ 510, 69, 41 },
		{ 510, 70, 41 },
		{ 510, 71, 41 },
		{ 510, 72, 41 },
		{ 510, 73, 41 },
		{ 510, 74, 41 },
		{ 510, 75, 41 },
		{ 510, 76, 41 },
		{ 510, 77, 41 },
		{ 510, 78, 41 },
		{ 510, 79, 41 },
		{ 510, 80, 41 },
		{ 510, 81, 41 },
		{ 510, 82, 41 },
		{ 510, 83, 41 },
		{ 510, 84, 41 },
		{ 510, 85, 41 },
		{ 510, 86, 41 },
		{ 510, 87, 41 },
		{ 510, 88, 41 },
		{ 510, 89, 41 },
		{ 510, 90, 41 },
		{ 510, 91, 41 },
		{ 510, 92, 41 },
		{ 510, 93, 41 },
		{ 510, 94, 41 },
		{ 510, 95, 41 },
		{ 510, 96, 41 },
		{ 510, 97, 41 },
		{ 510, 98, 41 },
		{ 510, 99, 41 },
		{ 510, 100, 41 },
		{ 510, 101, 41 },
		{ 510, 102, 41 },
		{ 510, 103, 41 },
		{ 510, 104, 41 },
		{ 510, 105, 41 },
		{ 510, 106, 41 },
		{ 510, 107, 41 },
		{ 510, 108, 41 },
		{ 510, 109, 41 },
		{ 510, 110, 41 },
		{ 510, 111, 41 },
		{ 510, 112, 41 },
		{ 510, 113, 41 },
		{ 510, 114, 41 },
		{ 510, 115, 41 },
		{ 510, 116, 41 },
		{ 510, 117, 41 },
		{ 510, 118, 41 },
		{ 510, 119, 41 },
		{ 510, 120, 41 },
		{ 510, 121, 41 },
		{ 510, 122, 41 },
		{ 510, 123, 41 },
		{ 510, 124, 41 },
		{ 510, 125, 41 },
		{ 510, 126, 41 },
		{ 510, 127, 41 },
		{ 510, 128, 41 },
		{ 510, 129, 41 },
		{ 510, 130, 41 },
		{ 511, 67, 41 },
		{ 511, 68, 41 },
		{ 511, 69, 41 },
		{ 511, 70, 41 },
		{ 511, 71, 41 },
		{ 511, 72, 41 },
		{ 511, 73, 41 },
		{ 511, 74, 41 },
		{ 511, 75, 41 },
		{ 511, 76, 41 },
		{ 511, 77, 41 },
		{ 511, 78, 41 },
		{ 511, 79, 41 },
		{ 511, 80, 41 },
		{ 511, 81, 41 },
		{ 511, 82, 41 },
		{ 511, 83, 41 },
		{ 511, 84, 41 },
		{ 511, 85, 41 },
		{ 511, 86, 41 },
		{ 511, 87, 41 },
		{ 511, 88, 41 },
		{ 511, 89, 41 },
		{ 511, 90, 41 },
		{ 511, 91, 41 },
		{ 511, 92, 41 },
		{ 511, 93, 41 },
		{ 511, 94, 41 },
		{ 511, 95, 41 },
		{ 511, 96, 41 },
		{ 511, 97, 41 },
		{ 511, 98, 41 },
		{ 511, 99, 41 },
		{ 511, 100, 41 },
		{ 511, 101, 41 },
		{ 511, 102, 41 },
		{ 511, 103, 41 },
		{ 511, 104, 41 },
		{ 511, 105, 41 },
		{ 511, 106, 41 },
		{ 511, 107, 41 },
		{ 511, 108, 41 },
		{ 511, 109, 41 },
		{ 511, 110, 41 },
		{ 511, 111, 41 },
		{ 511, 112, 41 },
		{ 511, 113, 41 },
		{ 511, 114, 41 },
		{ 511, 115, 41 },
		{ 511, 116, 41 },
		{ 511, 117, 41 },
		{ 511, 118, 41 },
		{ 511, 119, 41 },
		{ 511, 120, 41 },
		{ 511, 121, 41 },
		{ 511, 122, 41 },
		{ 511, 123, 41 },
		{ 511, 124, 41 },
		{ 511, 125, 41 },
		{ 511, 126, 41 },
		{ 511, 127, 41 },
		{ 511, 128, 41 },
		{ 511, 129, 41 },
		{ 512, 66, 41 },
		{ 512, 67, 41 },
		{ 512, 68, 41 },
		{ 512, 69, 41 },
		{ 512, 70, 41 },
		{ 512, 71, 41 },
		{ 512, 72, 41 },
		{ 512, 73, 41 },
		{ 512, 74, 41 },
		{ 512, 75, 41 },
		{ 512, 76, 41 },
		{ 512, 77, 41 },
		{ 512, 78, 41 },
		{ 512, 79, 41 },
		{ 512, 80, 41 },
		{ 512, 81, 41 },
		{ 512, 82, 41 },
		{ 512, 83, 41 },
		{ 512, 84, 41 },
		{ 512, 85, 41 },
		{ 512, 86, 41 },
		{ 512, 87, 41 },
		{ 512, 88, 41 },
		{ 512, 89, 41 },
		{ 512, 90, 41 },
		{ 512, 91, 41 },
		{ 512, 92, 41 },
		{ 512, 93, 41 },
		{ 512, 94, 41 },
		{ 512, 95, 41 },
		{ 512, 96, 41 },
		{ 512, 97, 41 },
		{ 512, 98, 41 },
		{ 512, 99, 41 },
		{ 512, 100, 41 },
		{ 512, 101, 41 },
		{ 512, 102, 41 },
		{ 512, 103, 41 },
		{ 512, 104, 41 },
		{ 512, 105, 41 },
		{ 512, 106, 41 },
		{ 512, 107, 41 },
		{ 512, 108, 41 },
		{ 512, 109, 41 },
		{ 512, 110, 41 },
		{ 512, 111, 41 },
		{ 512, 112, 41 },
		{ 512, 113, 41 },
		{ 512, 114, 41 },
		{ 512, 115, 41 },
		{ 512, 116, 41 },
		{ 512, 117, 41 },
		{ 512, 118, 41 },
		{ 512, 119, 41 },
		{ 512, 120, 41 },
		{ 512, 121, 41 },
		{ 512, 122, 41 },
		{ 512, 123, 41 },
		{ 512, 124, 41 },
		{ 512, 125, 41 },
		{ 512, 126, 41 },
		{ 512, 127, 41 },
		{ 512, 128, 41 },
		{ 512, 129, 41 },
		{ 1, 354, 42 },
		{ 1, 355, 42 },
		{ 1, 356, 42 },
		{ 1, 357, 42 },
		{ 1, 358, 42 },
		{ 1, 359, 42 },
		{ 1, 360, 42 },
		{ 1, 361, 42 },
		{ 1, 362, 42 },
		{ 1, 363, 42 },
		{ 1, 364, 42 },
		{ 1, 365, 42 },
		{ 1, 366, 42 },
		{ 1, 367, 42 },
		{ 1, 368, 42 },
		{ 1, 369, 42 },
		{ 1, 370, 42 },
		{ 1, 371, 42 },
		{ 1, 372, 42 },
		{ 1, 373, 42 },
		{ 1, 374, 42 },
		{ 1, 375, 42 },
		{ 1, 376, 42 },
		{ 1, 377, 42 },
		{ 1, 378, 42 },
		{ 1, 379, 42 },
		{ 1, 380, 42 },
		{ 1, 381, 42 },
		{ 1, 382, 42 },
		{ 1, 383, 42 },
		{ 1, 384, 42 },
		{ 1, 385, 42 },
		{ 1, 386, 42 },
		{ 1, 387, 42 },
		{ 1, 388, 42 },
		{ 1, 389, 42 },
		{ 1, 390, 42 },
		{ 1, 391, 42 },
		{ 1, 392, 42 },
		{ 1, 393, 42 },
		{ 1, 394, 42 },
		{ 1, 395, 42 },
		{ 1, 396, 42 },
		{ 1, 397, 42 },
		{ 1, 398, 42 },
		{ 1, 399, 42 },
		{ 1, 400, 42 },
		{ 1, 401, 42 },
		{ 1, 402, 42 },
		{ 1, 403, 42 },
		{ 1, 404, 42 },
		{ 1, 405, 42 },
		{ 1, 406, 42 },
		{ 1, 407, 42 },
		{ 1, 408, 42 },
		{ 1, 409, 42 },
		{ 1, 410, 42 },
		{ 1, 411, 42 },
		{ 1, 412, 42 },
		{ 1, 413, 42 },
		{ 1, 414, 42 },
		{ 1, 415, 42 },
		{ 1, 416, 42 },
		{ 2, 353, 42 },
		{ 2, 354, 42 },
		{ 2, 355, 42 },
		{ 2, 356, 42 },
		{ 2, 357, 42 },
		{ 2, 358, 42 },
		{ 2, 359, 42 },
		{ 2, 360, 42 },
		{ 2, 361, 42 },
		{ 2, 362, 42 },
		{ 2, 363, 42 },
		{ 2, 364, 42 },
		{ 2, 365, 42 },
		{ 2, 366, 42 },
		{ 2, 367, 42 },
		{ 2, 368, 42 },
		{ 2, 369, 42 },
		{ 2, 370, 42 },
		{ 2, 371, 42 },
		{ 2, 372, 42 },
		{ 2, 373, 42 },
		{ 2, 374, 42 },
		{ 2, 375, 42 },
		{ 2, 376, 42 },
		{ 2, 377, 42 },
		{ 2, 378, 42 },
		{ 2, 379, 42 },
		{ 2, 380, 42 },
		{ 2, 381, 42 },
		{ 2, 382, 42 },
		{ 2, 383, 42 },
		{ 2, 384, 42 },
		{ 2, 385, 42 },
		{ 2, 386, 42 },
		{ 2, 387, 42 },
		{ 2, 388, 42 },
		{ 2, 389, 42 },
		{ 2, 390, 42 },
		{ 2, 391, 42 },
		{ 2, 392, 42 },
		{ 2, 393, 42 },
		{ 2, 394, 42 },
		{ 2, 395, 42 },
		{ 2, 396, 42 },
		{ 2, 397, 42 },
		{ 2, 398, 42 },
		{ 2, 399, 42 },
		{ 2, 400, 42 },
		{ 2, 401, 42 },
		{ 2, 402, 42 },
		{ 2, 403, 42 },
		{ 2, 404, 42 },
		{ 2, 405, 42 },
		{ 2, 406, 42 },
		{ 2, 407, 42 },
		{ 2, 408, 42 },
		{ 2, 409, 42 },
		{ 2, 410, 42 },
		{ 2, 411, 42 },
		{ 2, 412, 42 },
		{ 2, 413, 42 },
		{ 2, 414, 42 },
		{ 2, 415, 42 },
		{ 2, 416, 42 },
		{ 3, 353, 42 },
		{ 3, 354, 42 },
		{ 3, 355, 42 },
		{ 3, 356, 42 },
		{ 3, 357, 42 },
		{ 3, 358, 42 },
		{ 3, 359, 42 },
		{ 3, 360, 42 },
		{ 3, 361, 42 },
		{ 3, 362, 42 },
		{ 3, 363, 42 },
		{ 3, 364, 42 },
		{ 3, 365, 42 },
		{ 3, 366, 42 },
		{ 3, 367, 42 },
		{ 3, 368, 42 },
		{ 3, 369, 42 },
		{ 3, 370, 42 },
		{ 3, 371, 42 },
		{ 3, 372, 42 },
		{ 3, 373, 42 },
		{ 3, 374, 42 },
		{ 3, 375, 42 },
		{ 3, 376, 42 },
		{ 3, 377, 42 },
		{ 3, 378, 42 },
		{ 3, 379, 42 },
		{ 3, 380, 42 },
		{ 3, 381, 42 },
		{ 3, 382, 42 },
		{ 3, 383, 42 },
		{ 3, 384, 42 },
		{ 3, 385, 42 },
		{ 3, 386, 42 },
		{ 3, 387, 42 },
		{ 3, 388, 42 },
		{ 3, 389, 42 },
		{ 3, 390, 42 },
		{ 3, 391, 42 },
		{ 3, 392, 42 },
		{ 3, 393, 42 },
		{ 3, 394, 42 },
		{ 3, 395, 42 },
		{ 3, 396, 42 },
		{ 3, 397, 42 },
		{ 3, 398, 42 },
		{ 3, 399, 42 },
		{ 3, 400, 42 },
		{ 3, 401, 42 },
		{ 3, 402, 42 },
		{ 3, 403, 42 },
		{ 3, 404, 42 },
		{ 3, 405, 42 },
		{ 3, 406, 42 },
		{ 3, 407, 42 },
		{ 3, 408, 42 },
		{ 3, 409, 42 },
		{ 3, 410, 42 },
		{ 3, 411, 42 },
		{ 3, 412, 42 },
		{ 3, 413, 42 },
		{ 3, 414, 42 },
		{ 3, 415, 42 },
		{ 4, 352, 42 },
		{ 4, 353, 42 },
		{ 4, 354, 42 },
		{ 4, 355, 42 },
		{ 4, 356, 42 },
		{ 4, 357, 42 },
		{ 4, 358, 42 },
		{ 4, 359, 42 },
		{ 4, 360, 42 },
		{ 4, 361, 42 },
		{ 4, 362, 42 },
		{ 4, 363, 42 },
		{ 4, 364, 42 },
		{ 4, 365, 42 },
		{ 4, 366, 42 },
		{ 4, 367, 42 },
		{ 4, 368, 42 },
		{ 4, 369, 42 },
		{ 4, 370, 42 },
		{ 4, 371, 42 },
		{ 4, 372, 42 },
		{ 4, 373, 42 },
		{ 4, 374, 42 },
		{ 4, 375, 42 },
		{ 4, 376, 42 },
		{ 4, 377, 42 },
		{ 4, 378, 42 },
		{ 4, 379, 42 },
		{ 4, 380, 42 },
		{ 4, 381, 42 },
		{ 4, 382, 42 },
		{ 4, 383, 42 },
		{ 4, 384, 42 },
		{ 4, 385, 42 },
		{ 4, 386, 42 },
		{ 4, 387, 42 },
		{ 4, 388, 42 },
		{ 4, 389, 42 },
		{ 4, 390, 42 },
		{ 4, 391, 42 },
		{ 4, 392, 42 },
		{ 4, 393, 42 },
		{ 4, 394, 42 },
		{ 4, 395, 42 },
		{ 4, 396, 42 },
		{ 4, 397, 42 },
		{ 4, 398, 42 },
		{ 4, 399, 42 },
		{ 4, 400, 42 },
		{ 4, 401, 42 },
		{ 4, 402, 42 },
		{ 4, 403, 42 },
		{ 4, 404, 42 },
		{ 4, 405, 42 },
		{ 4, 406, 42 },
		{ 4, 407, 42 },
		{ 4, 408, 42 },
		{ 4, 409, 42 },
		{ 4, 410, 42 },
		{ 4, 411, 42 },
		{ 4, 412, 42 },
		{ 4, 413, 42 },
		{ 4, 414, 42 },
		{ 4, 415, 42 },
		{ 5, 352, 42 },
		{ 5, 353, 42 },
		{ 5, 354, 42 },
		{ 5, 355, 42 },
		{ 5, 356, 42 },
		{ 5, 357, 42 },
		{ 5, 358, 42 },
		{ 5, 359, 42 },
		{ 5, 360, 42 },
		{ 5, 361, 42 },
		{ 5, 362, 42 },
		{ 5, 363, 42 },
		{ 5, 364, 42 },
		{ 5, 365, 42 },
		{ 5, 366, 42 },
		{ 5, 367, 42 },
		{ 5, 368, 42 },
		{ 5, 369, 42 },
		{ 5, 370, 42 },
		{ 5, 371, 42 },
		{ 5, 372, 42 },
		{ 5, 373, 42 },
		{ 5, 374, 42 },
		{ 5, 375, 42 },
		{ 5, 376, 42 },
		{ 5, 377, 42 },
		{ 5, 378, 42 },
		{ 5, 379, 42 },
		{ 5, 380, 42 },
		{ 5, 381, 42 },
		{ 5, 382, 42 },
		{ 5, 383, 42 },
		{ 5, 384, 42 },
		{ 5, 385, 42 },
		{ 5, 386, 42 },
		{ 5, 387, 42 },
		{ 5, 388, 42 },
		{ 5, 389, 42 },
		{ 5, 390, 42 },
		{ 5, 391, 42 },
		{ 5, 392, 42 },
		{ 5, 393, 42 },
		{ 5, 394, 42 },
		{ 5, 395, 42 },
		{ 5, 396, 42 },
		{ 5, 397, 42 },
		{ 5, 398, 42 },
		{ 5, 399, 42 },
		{ 5, 400, 42 },
		{ 5, 401, 42 },
		{ 5, 402, 42 },
		{ 5, 403, 42 },
		{ 5, 404, 42 },
		{ 5, 405, 42 },
		{ 5, 406, 42 },
		{ 5, 407, 42 },
		{ 5, 408, 42 },
		{ 5, 409, 42 },
		{ 5, 410, 42 },
		{ 5, 411, 42 },
		{ 5, 412, 42 },
		{ 5, 413, 42 },
		{ 5, 414, 42 },
		{ 6, 352, 42 },
		{ 6, 353, 42 },
		{ 6, 354, 42 },
		{ 6, 355, 42 },
		{ 6, 356, 42 },
		{ 6, 357, 42 },
		{ 6, 358, 42 },
		{ 6, 359, 42 },
		{ 6, 360, 42 },
		{ 6, 361, 42 },
		{ 6, 362, 42 },
		{ 6, 363, 42 },
		{ 6, 364, 42 },
		{ 6, 365, 42 },
		{ 6, 366, 42 },
		{ 6, 367, 42 },
		{ 6, 368, 42 },
		{ 6, 369, 42 },
		{ 6, 370, 42 },
		{ 6, 371, 42 },
		{ 6, 372, 42 },
		{ 6, 373, 42 },
		{ 6, 374, 42 },
		{ 6, 375, 42 },
		{ 6, 376, 42 },
		{ 6, 377, 42 },
		{ 6, 378, 42 },
		{ 6, 379, 42 },
		{ 6, 380, 42 },
		{ 6, 381, 42 },
		{ 6, 382, 42 },
		{ 6, 383, 42 },
		{ 6, 384, 42 },
		{ 6, 385, 42 },
		{ 6, 386, 42 },
		{ 6, 387, 42 },
		{ 6, 388, 42 },
		{ 6, 389, 42 },
		{ 6, 390, 42 },
		{ 6, 391, 42 },
		{ 6, 392, 42 },
		{ 6, 393, 42 },
		{ 6, 394, 42 },
		{ 6, 395, 42 },
		{ 6, 396, 42 },
		{ 6, 397, 42 },
		{ 6, 398, 42 },
		{ 6, 399, 42 },
		{ 6, 400, 42 },
		{ 6, 401, 42 },
		{ 6, 402, 42 },
		{ 6, 403, 42 },
		{ 6, 404, 42 },
		{ 6, 405, 42 },
		{ 6, 406, 42 },
		{ 6, 407, 42 },
		{ 6, 408, 42 },
		{ 6, 409, 42 },
		{ 6, 410, 42 },
		{ 6, 411, 42 },
		{ 6, 412, 42 },
		{ 6, 413, 42 },
		{ 7, 351, 42 },
		{ 7, 352, 42 },
		{ 7, 353, 42 },
		{ 7, 354, 42 },
		{ 7, 355, 42 },
		{ 7, 356, 42 },
		{ 7, 357, 42 },
		{ 7, 358, 42 },
		{ 7, 359, 42 },
		{ 7, 360, 42 },
		{ 7, 361, 42 },
		{ 7, 362, 42 },
		{ 7, 363, 42 },
		{ 7, 364, 42 },
		{ 7, 365, 42 },
		{ 7, 366, 42 },
		{ 7, 367, 42 },
		{ 7, 368, 42 },
		{ 7, 369, 42 },
		{ 7, 370, 42 },
		{ 7, 371, 42 },
		{ 7, 372, 42 },
		{ 7, 373, 42 },
		{ 7, 374, 42 },
		{ 7, 375, 42 },
		{ 7, 376, 42 },
		{ 7, 377, 42 },
		{ 7, 378, 42 },
		{ 7, 379, 42 },
		{ 7, 380, 42 },
		{ 7, 381, 42 },
		{ 7, 382, 42 },
		{ 7, 383, 42 },
		{ 7, 384, 42 },
		{ 7, 385, 42 },
		{ 7, 386, 42 },
		{ 7, 387, 42 },
		{ 7, 388, 42 },
		{ 7, 389, 42 },
		{ 7, 390, 42 },
		{ 7, 391, 42 },
		{ 7, 392, 42 },
		{ 7, 393, 42 },
		{ 7, 394, 42 },
		{ 7, 395, 42 },
		{ 7, 396, 42 },
		{ 7, 397, 42 },
		{ 7, 398, 42 },
		{ 7, 399, 42 },
		{ 7, 400, 42 },
		{ 7, 401, 42 },
		{ 7, 402, 42 },
		{ 7, 403, 42 },
		{ 7, 404, 42 },
		{ 7, 405, 42 },
		{ 7, 406, 42 },
		{ 7, 407, 42 },
		{ 7, 408, 42 },
		{ 7, 409, 42 },
		{ 7, 410, 42 },
		{ 7, 411, 42 },
		{ 7, 412, 42 },
		{ 7, 413, 42 },
		{ 8, 351, 42 },
		{ 8, 352, 42 },
		{ 8, 353, 42 },
		{ 8, 354, 42 },
		{ 8, 355, 42 },
		{ 8, 356, 42 },
		{ 8, 357, 42 },
		{ 8, 358, 42 },
		{ 8, 359, 42 },
		{ 8, 360, 42 },
		{ 8, 361, 42 },
		{ 8, 362, 42 },
		{ 8, 363, 42 },
		{ 8, 364, 42 },
		{ 8, 365, 42 },
		{ 8, 366, 42 },
		{ 8, 367, 42 },
		{ 8, 368, 42 },
		{ 8, 369, 42 },
		{ 8, 370, 42 },
		{ 8, 371, 42 },
		{ 8, 372, 42 },
		{ 8, 373, 42 },
		{ 8, 374, 42 },
		{ 8, 375, 42 },
		{ 8, 376, 42 },
		{ 8, 377, 42 },
		{ 8, 378, 42 },
		{ 8, 379, 42 },
		{ 8, 380, 42 },
		{ 8, 381, 42 },
		{ 8, 382, 42 },
		{ 8, 383, 42 },
		{ 8, 384, 42 },
		{ 8, 385, 42 },
		{ 8, 386, 42 },
		{ 8, 387, 42 },
		{ 8, 388, 42 },
		{ 8, 389, 42 },
		{ 8, 390, 42 },
		{ 8, 391, 42 },
		{ 8, 392, 42 },
		{ 8, 393, 42 },
		{ 8, 394, 42 },
		{ 8, 395, 42 },
		{ 8, 396, 42 },
		{ 8, 397, 42 },
		{ 8, 398, 42 },
		{ 8, 399, 42 },
		{ 8, 400, 42 },
		{ 8, 401, 42 },
		{ 8, 402, 42 },
		{ 8, 403, 42 },
		{ 8, 404, 42 },
		{ 8, 405, 42 },
		{ 8, 406, 42 },
		{ 8, 407, 42 },
		{ 8, 408, 42 },
		{ 8, 409, 42 },
		{ 8, 410, 42 },
		{ 8, 411, 42 },
		{ 8, 412, 42 },
		{ 9, 351, 42 },
		{ 9, 352, 42 },
		{ 9, 353, 42 },
		{ 9, 354, 42 },
		{ 9, 355, 42 },
		{ 9, 356, 42 },
		{ 9, 357, 42 },
		{ 9, 358, 42 },
		{ 9, 359, 42 },
		{ 9, 360, 42 },
		{ 9, 361, 42 },
		{ 9, 362, 42 },
		{ 9, 363, 42 },
		{ 9, 364, 42 },
		{ 9, 365, 42 },
		{ 9, 366, 42 },
		{ 9, 367, 42 },
		{ 9, 368, 42 },
		{ 9, 369, 42 },
		{ 9, 370, 42 },
		{ 9, 371, 42 },
		{ 9, 372, 42 },
		{ 9, 373, 42 },
		{ 9, 374, 42 },
		{ 9, 375, 42 },
		{ 9, 376, 42 },
		{ 9, 377, 42 },
		{ 9, 378, 42 },
		{ 9, 379, 42 },
		{ 9, 380, 42 },
		{ 9, 381, 42 },
		{ 9, 382, 42 },
		{ 9, 383, 42 },
		{ 9, 384, 42 },
		{ 9, 385, 42 },
		{ 9, 386, 42 },
		{ 9, 387, 42 },
		{ 9, 388, 42 },
		{ 9, 389, 42 },
		{ 9, 390, 42 },
		{ 9, 391, 42 },
		{ 9, 392, 42 },
		{ 9, 393, 42 },
		{ 9, 394, 42 },
		{ 9, 395, 42 },
		{ 9, 396, 42 },
		{ 9, 397, 42 },
		{ 9, 398, 42 },
		{ 9, 399, 42 },
		{ 9, 400, 42 },
		{ 9, 401, 42 },
		{ 9, 402, 42 },
		{ 9, 403, 42 },
		{ 9, 404, 42 },
		{ 9, 405, 42 },
		{ 9, 406, 42 },
		{ 9, 407, 42 },
		{ 9, 408, 42 },
		{ 9, 409, 42 },
		{ 9, 410, 42 },
		{ 9, 411, 42 },
		{ 10, 350, 42 },
		{ 10, 351, 42 },
		{ 10, 352, 42 },
		{ 10, 353, 42 },
		{ 10, 354, 42 },
		{ 10, 355, 42 },
		{ 10, 356, 42 },
		{ 10, 357, 42 },
		{ 10, 358, 42 },
		{ 10, 359, 42 },
		{ 10, 360, 42 },
		{ 10, 361, 42 },
		{ 10, 362, 42 },
		{ 10, 363, 42 },
		{ 10, 364, 42 },
		{ 10, 365, 42 },
		{ 10, 366, 42 },
		{ 10, 367, 42 },
		{ 10, 368, 42 },
		{ 10, 369, 42 },
		{ 10, 370, 42 },
		{ 10, 371, 42 },
		{ 10, 372, 42 },
		{ 10, 373, 42 },
		{ 10, 374, 42 },
		{ 10, 375, 42 },
		{ 10, 376, 42 },
		{ 10, 377, 42 },
		{ 10, 378, 42 },
		{ 10, 379, 42 },
		{ 10, 380, 42 },
		{ 10, 381, 42 },
		{ 10, 382, 42 },
		{ 10, 383, 42 },
		{ 10, 384, 42 },
		{ 10, 385, 42 },
		{ 10, 386, 42 },
		{ 10, 387, 42 },
		{ 10, 388, 42 },
		{ 10, 389, 42 },
		{ 10, 390, 42 },
		{ 10, 391, 42 },
		{ 10, 392, 42 },
		{ 10, 393, 42 },
		{ 10, 394, 42 },
		{ 10, 395, 42 },
		{ 10, 396, 42 },
		{ 10, 397, 42 },
		{ 10, 398, 42 },
		{ 10, 399, 42 },
		{ 10, 400, 42 },
		{ 10, 401, 42 },
		{ 10, 402, 42 },
		{ 10, 403, 42 },
		{ 10, 404, 42 },
		{ 10, 405, 42 },
		{ 10, 406, 42 },
		{ 10, 407, 42 },
		{ 10, 408, 42 },
		{ 10, 409, 42 },
		{ 10, 410, 42 },
		{ 10, 411, 42 },
		{ 11, 350, 42 },
		{ 11, 351, 42 },
		{ 11, 352, 42 },
		{ 11, 353, 42 },
		{ 11, 354, 42 },
		{ 11, 355, 42 },
		{ 11, 356, 42 },
		{ 11, 357, 42 },
		{ 11, 358, 42 },
		{ 11, 359, 42 },
		{ 11, 360, 42 },
		{ 11, 361, 42 },
		{ 11, 362, 42 },
		{ 11, 363, 42 },
		{ 11, 364, 42 },
		{ 11, 365, 42 },
		{ 11, 366, 42 },
		{ 11, 367, 42 },
		{ 11, 368, 42 },
		{ 11, 369, 42 },
		{ 11, 370, 42 },
		{ 11, 371, 42 },
		{ 11, 372, 42 },
		{ 11, 373, 42 },
		{ 11, 374, 42 },
		{ 11, 375, 42 },
		{ 11, 376, 42 },
		{ 11, 377, 42 },
		{ 11, 378, 42 },
		{ 11, 379, 42 },
		{ 11, 380, 42 },
		{ 11, 381, 42 },
		{ 11, 382, 42 },
		{ 11, 383, 42 },
		{ 11, 384, 42 },
		{ 11, 385, 42 },
		{ 11, 386, 42 },
		{ 11, 387, 42 },
		{ 11, 388, 42 },
		{ 11, 389, 42 },
		{ 11, 390, 42 },
		{ 11, 391, 42 },
		{ 11, 392, 42 },
		{ 11, 393, 42 },
		{ 11, 394, 42 },
		{ 11, 395, 42 },
		{ 11, 396, 42 },
		{ 11, 397, 42 },
		{ 11, 398, 42 },
		{ 11, 399, 42 },
		{ 11, 400, 42 },
		{ 11, 401, 42 },
		{ 11, 402, 42 },
		{ 11, 403, 42 },
		{ 11, 404, 42 },
		{ 11, 405, 42 },
		{ 11, 406, 42 },
		{ 11, 407, 42 },
		{ 11, 408, 42 },
		{ 11, 409, 42 },
		{ 11, 410, 42 },
		{ 12, 349, 42 },
		{ 12, 350, 42 },
		{ 12, 351, 42 },
		{ 12, 352, 42 },
		{ 12, 353, 42 },
		{ 12, 354, 42 },
		{ 12, 355, 42 },
		{ 12, 356, 42 },
		{ 12, 357, 42 },
		{ 12, 358, 42 },
		{ 12, 359, 42 },
		{ 12, 360, 42 },
		{ 12, 361, 42 },
		{ 12, 362, 42 },
		{ 12, 363, 42 },
		{ 12, 364, 42 },
		{ 12, 365, 42 },
		{ 12, 366, 42 },
		{ 12, 367, 42 },
		{ 12, 368, 42 },
		{ 12, 369, 42 },
		{ 12, 370, 42 },
		{ 12, 371, 42 },
		{ 12, 372, 42 },
		{ 12, 373, 42 },
		{ 12, 374, 42 },
		{ 12, 375, 42 },
		{ 12, 376, 42 },
		{ 12, 377, 42 },
		{ 12, 378, 42 },
		{ 12, 379, 42 },
		{ 12, 380, 42 },
		{ 12, 381, 42 },
		{ 12, 382, 42 },
		{ 12, 383, 42 },
		{ 12, 384, 42 },
		{ 12, 385, 42 },
		{ 12, 386, 42 },
		{ 12, 387, 42 },
		{ 12, 388, 42 },
		{ 12, 389, 42 },
		{ 12, 390, 42 },
		{ 12, 391, 42 },
		{ 12, 392, 42 },
		{ 12, 393, 42 },
		{ 12, 394, 42 },
		{ 12, 395, 42 },
		{ 12, 396, 42 },
		{ 12, 397, 42 },
		{ 12, 398, 42 },
		{ 12, 399, 42 },
		{ 12, 400, 42 },
		{ 12, 401, 42 },
		{ 12, 402, 42 },
		{ 12, 403, 42 },
		{ 12, 404, 42 },
		{ 12, 405, 42 },
		{ 12, 406, 42 },
		{ 12, 407, 42 },
		{ 12, 408, 42 },
		{ 12, 409, 42 },
		{ 12, 410, 42 },
		{ 13, 349, 42 },
		{ 13, 350, 42 },
		{ 13, 351, 42 },
		{ 13, 352, 42 },
		{ 13, 353, 42 },
		{ 13, 354, 42 },
		{ 13, 355, 42 },
		{ 13, 356, 42 },
		{ 13, 357, 42 },
		{ 13, 358, 42 },
		{ 13, 359, 42 },
		{ 13, 360, 42 },
		{ 13, 361, 42 },
		{ 13, 362, 42 },
		{ 13, 363, 42 },
		{ 13, 364, 42 },
		{ 13, 365, 42 },
		{ 13, 366, 42 },
		{ 13, 367, 42 },
		{ 13, 368, 42 },
		{ 13, 369, 42 },
		{ 13, 370, 42 },
		{ 13, 371, 42 },
		{ 13, 372, 42 },
		{ 13, 373, 42 },
		{ 13, 374, 42 },
		{ 13, 375, 42 },
		{ 13, 376, 42 },
		{ 13, 377, 42 },
		{ 13, 378, 42 },
		{ 13, 379, 42 },
		{ 13, 380, 42 },
		{ 13, 381, 42 },
		{ 13, 382, 42 },
		{ 13, 383, 42 },
		{ 13, 384, 42 },
		{ 13, 385, 42 },
		{ 13, 386, 42 },
		{ 13, 387, 42 },
		{ 13, 388, 42 },
		{ 13, 389, 42 },
		{ 13, 390, 42 },
		{ 13, 391, 42 },
		{ 13, 392, 42 },
		{ 13, 393, 42 },
		{ 13, 394, 42 },
		{ 13, 395, 42 },
		{ 13, 396, 42 },
		{ 13, 397, 42 },
		{ 13, 398, 42 },
		{ 13, 399, 42 },
		{ 13, 400, 42 },
		{ 13, 401, 42 },
		{ 13, 402, 42 },
		{ 13, 403, 42 },
		{ 13, 404, 42 },
		{ 13, 405, 42 },
		{ 13, 406, 42 },
		{ 13, 407, 42 },
		{ 13, 408, 42 },
		{ 13, 409, 42 },
		{ 14, 349, 42 },
		{ 14, 350, 42 },
		{ 14, 351, 42 },
		{ 14, 352, 42 },
		{ 14, 353, 42 },
		{ 14, 354, 42 },
		{ 14, 355, 42 },
		{ 14, 356, 42 },
		{ 14, 357, 42 },
		{ 14, 358, 42 },
		{ 14, 359, 42 },
		{ 14, 360, 42 },
		{ 14, 361, 42 },
		{ 14, 362, 42 },
		{ 14, 363, 42 },
		{ 14, 364, 42 },
		{ 14, 365, 42 },
		{ 14, 366, 42 },
		{ 14, 367, 42 },
		{ 14, 368, 42 },
		{ 14, 369, 42 },
		{ 14, 370, 42 },
		{ 14, 371, 42 },
		{ 14, 372, 42 },
		{ 14, 373, 42 },
		{ 14, 374, 42 },
		{ 14, 375, 42 },
		{ 14, 376, 42 },
		{ 14, 377, 42 },
		{ 14, 378, 42 },
		{ 14, 379, 42 },
		{ 14, 380, 42 },
		{ 14, 381, 42 },
		{ 14, 382, 42 },
		{ 14, 383, 42 },
		{ 14, 384, 42 },
		{ 14, 385, 42 },
		{ 14, 386, 42 },
		{ 14, 387, 42 },
		{ 14, 388, 42 },
		{ 14, 389, 42 },
		{ 14, 390, 42 },
		{ 14, 391, 42 },
		{ 14, 392, 42 },
		{ 14, 393, 42 },
		{ 14, 394, 42 },
		{ 14, 395, 42 },
		{ 14, 396, 42 },
		{ 14, 397, 42 },
		{ 14, 398, 42 },
		{ 14, 399, 42 },
		{ 14, 400, 42 },
		{ 14, 401, 42 },
		{ 14, 402, 42 },
		{ 14, 403, 42 },
		{ 14, 404, 42 },
		{ 14, 405, 42 },
		{ 14, 406, 42 },
		{ 14, 407, 42 },
		{ 14, 408, 42 },
		{ 15, 348, 42 },
		{ 15, 349, 42 },
		{ 15, 350, 42 },
		{ 15, 351, 42 },
		{ 15, 352, 42 },
		{ 15, 353, 42 },
		{ 15, 354, 42 },
		{ 15, 355, 42 },
		{ 15, 356, 42 },
		{ 15, 357, 42 },
		{ 15, 358, 42 },
		{ 15, 359, 42 },
		{ 15, 360, 42 },
		{ 15, 361, 42 },
		{ 15, 362, 42 },
		{ 15, 363, 42 },
		{ 15, 364, 42 },
		{ 15, 365, 42 },
		{ 15, 366, 42 },
		{ 15, 367, 42 },
		{ 15, 368, 42 },
		{ 15, 369, 42 },
		{ 15, 370, 42 },
		{ 15, 371, 42 },
		{ 15, 372, 42 },
		{ 15, 373, 42 },
		{ 15, 374, 42 },
		{ 15, 375, 42 },
		{ 15, 376, 42 },
		{ 15, 377, 42 },
		{ 15, 378, 42 },
		{ 15, 379, 42 },
		{ 15, 380, 42 },
		{ 15, 381, 42 },
		{ 15, 382, 42 },
		{ 15, 383, 42 },
		{ 15, 384, 42 },
		{ 15, 385, 42 },
		{ 15, 386, 42 },
		{ 15, 387, 42 },
		{ 15, 388, 42 },
		{ 15, 389, 42 },
		{ 15, 390, 42 },
		{ 15, 391, 42 },
		{ 15, 392, 42 },
		{ 15, 393, 42 },
		{ 15, 394, 42 },
		{ 15, 395, 42 },
		{ 15, 396, 42 },
		{ 15, 397, 42 },
		{ 15, 398, 42 },
		{ 15, 399, 42 },
		{ 15, 400, 42 },
		{ 15, 401, 42 },
		{ 15, 402, 42 },
		{ 15, 403, 42 },
		{ 15, 404, 42 },
		{ 15, 405, 42 },
		{ 15, 406, 42 },
		{ 15, 407, 42 },
		{ 15, 408, 42 },
		{ 16, 348, 42 },
		{ 16, 349, 42 },
		{ 16, 350, 42 },
		{ 16, 351, 42 },
		{ 16, 352, 42 },
		{ 16, 353, 42 },
		{ 16, 354, 42 },
		{ 16, 355, 42 },
		{ 16, 356, 42 },
		{ 16, 357, 42 },
		{ 16, 358, 42 },
		{ 16, 359, 42 },
		{ 16, 360, 42 },
		{ 16, 361, 42 },
		{ 16, 362, 42 },
		{ 16, 363, 42 },
		{ 16, 364, 42 },
		{ 16, 365, 42 },
		{ 16, 366, 42 },
		{ 16, 367, 42 },
		{ 16, 368, 42 },
		{ 16, 369, 42 },
		{ 16, 370, 42 },
		{ 16, 371, 42 },
		{ 16, 372, 42 },
		{ 16, 373, 42 },
		{ 16, 374, 42 },
		{ 16, 375, 42 },
		{ 16, 376, 42 },
		{ 16, 377, 42 },
		{ 16, 378, 42 },
		{ 16, 379, 42 },
		{ 16, 380, 42 },
		{ 16, 381, 42 },
		{ 16, 382, 42 },
		{ 16, 383, 42 },
		{ 16, 384, 42 },
		{ 16, 385, 42 },
		{ 16, 386, 42 },
		{ 16, 387, 42 },
		{ 16, 388, 42 },
		{ 16, 389, 42 },
		{ 16, 390, 42 },
		{ 16, 391, 42 },
		{ 16, 392, 42 },
		{ 16, 393, 42 },
		{ 16, 394, 42 },
		{ 16, 395, 42 },
		{ 16, 396, 42 },
		{ 16, 397, 42 },
		{ 16, 398, 42 },
		{ 16, 399, 42 },
		{ 16, 400, 42 },
		{ 16, 401, 42 },
		{ 16, 402, 42 },
		{ 16, 403, 42 },
		{ 16, 404, 42 },
		{ 16, 405, 42 },
		{ 16, 406, 42 },
		{ 16, 407, 42 },
		{ 17, 348, 42 },
		{ 17, 349, 42 },
		{ 17, 350, 42 },
		{ 17, 351, 42 },
		{ 17, 352, 42 },
		{ 17, 353, 42 },
		{ 17, 354, 42 },
		{ 17, 355, 42 },
		{ 17, 356, 42 },
		{ 17, 357, 42 },
		{ 17, 358, 42 },
		{ 17, 359, 42 },
		{ 17, 360, 42 },
		{ 17, 361, 42 },
		{ 17, 362, 42 },
		{ 17, 363, 42 },
		{ 17, 364, 42 },
		{ 17, 365, 42 },
		{ 17, 366, 42 },
		{ 17, 367, 42 },
		{ 17, 368, 42 },
		{ 17, 369, 42 },
		{ 17, 370, 42 },
		{ 17, 371, 42 },
		{ 17, 372, 42 },
		{ 17, 373, 42 },
		{ 17, 374, 42 },
		{ 17, 375, 42 },
		{ 17, 376, 42 },
		{ 17, 377, 42 },
		{ 17, 378, 42 },
		{ 17, 379, 42 },
		{ 17, 380, 42 },
		{ 17, 381, 42 },
		{ 17, 382, 42 },
		{ 17, 383, 42 },
		{ 17, 384, 42 },
		{ 17, 385, 42 },
		{ 17, 386, 42 },
		{ 17, 387, 42 },
		{ 17, 388, 42 },
		{ 17, 389, 42 },
		{ 17, 390, 42 },
		{ 17, 391, 42 },
		{ 17, 392, 42 },
		{ 17, 393, 42 },
		{ 17, 394, 42 },
		{ 17, 395, 42 },
		{ 17, 396, 42 },
		{ 17, 397, 42 },
		{ 17, 398, 42 },
		{ 17, 399, 42 },
		{ 17, 400, 42 },
		{ 17, 401, 42 },
		{ 17, 402, 42 },
		{ 17, 403, 42 },
		{ 17, 404, 42 },
		{ 17, 405, 42 },
		{ 17, 406, 42 },
		{ 18, 347, 42 },
		{ 18, 348, 42 },
		{ 18, 349, 42 },
		{ 18, 350, 42 },
		{ 18, 351, 42 },
		{ 18, 352, 42 },
		{ 18, 353, 42 },
		{ 18, 354, 42 },
		{ 18, 355, 42 },
		{ 18, 356, 42 },
		{ 18, 357, 42 },
		{ 18, 358, 42 },
		{ 18, 359, 42 },
		{ 18, 360, 42 },
		{ 18, 361, 42 },
		{ 18, 362, 42 },
		{ 18, 363, 42 },
		{ 18, 364, 42 },
		{ 18, 365, 42 },
		{ 18, 366, 42 },
		{ 18, 367, 42 },
		{ 18, 368, 42 },
		{ 18, 369, 42 },
		{ 18, 370, 42 },
		{ 18, 371, 42 },
		{ 18, 372, 42 },
		{ 18, 373, 42 },
		{ 18, 374, 42 },
		{ 18, 375, 42 },
		{ 18, 376, 42 },
		{ 18, 377, 42 },
		{ 18, 378, 42 },
		{ 18, 379, 42 },
		{ 18, 380, 42 },
		{ 18, 381, 42 },
		{ 18, 382, 42 },
		{ 18, 383, 42 },
		{ 18, 384, 42 },
		{ 18, 385, 42 },
		{ 18, 386, 42 },
		{ 18, 387, 42 },
		{ 18, 388, 42 },
		{ 18, 389, 42 },
		{ 18, 390, 42 },
		{ 18, 391, 42 },
		{ 18, 392, 42 },
		{ 18, 393, 42 },
		{ 18, 394, 42 },
		{ 18, 395, 42 },
		{ 18, 396, 42 },
		{ 18, 397, 42 },
		{ 18, 398, 42 },
		{ 18, 399, 42 },
		{ 18, 400, 42 },
		{ 18, 401, 42 },
		{ 18, 402, 42 },
		{ 18, 403, 42 },
		{ 18, 404, 42 },
		{ 18, 405, 42 },
		{ 18, 406, 42 },
		{ 19, 347, 42 },
		{ 19, 348, 42 },
		{ 19, 349, 42 },
		{ 19, 350, 42 },
		{ 19, 351, 42 },
		{ 19, 352, 42 },
		{ 19, 353, 42 },
		{ 19, 354, 42 },
		{ 19, 355, 42 },
		{ 19, 356, 42 },
		{ 19, 357, 42 },
		{ 19, 358, 42 },
		{ 19, 359, 42 },
		{ 19, 360, 42 },
		{ 19, 361, 42 },
		{ 19, 362, 42 },
		{ 19, 363, 42 },
		{ 19, 364, 42 },
		{ 19, 365, 42 },
		{ 19, 366, 42 },
		{ 19, 367, 42 },
		{ 19, 368, 42 },
		{ 19, 369, 42 },
		{ 19, 370, 42 },
		{ 19, 371, 42 },
		{ 19, 372, 42 },
		{ 19, 373, 42 },
		{ 19, 374, 42 },
		{ 19, 375, 42 },
		{ 19, 376, 42 },
		{ 19, 377, 42 },
		{ 19, 378, 42 },
		{ 19, 379, 42 },
		{ 19, 380, 42 },
		{ 19, 381, 42 },
		{ 19, 382, 42 },
		{ 19, 383, 42 },
		{ 19, 384, 42 },
		{ 19, 385, 42 },
		{ 19, 386, 42 },
		{ 19, 387, 42 },
		{ 19, 388, 42 },
		{ 19, 389, 42 },
		{ 19, 390, 42 },
		{ 19, 391, 42 },
		{ 19, 392, 42 },
		{ 19, 393, 42 },
		{ 19, 394, 42 },
		{ 19, 395, 42 },
		{ 19, 396, 42 },
		{ 19, 397, 42 },
		{ 19, 398, 42 },
		{ 19, 399, 42 },
		{ 19, 400, 42 },
		{ 19, 401, 42 },
		{ 19, 402, 42 },
		{ 19, 403, 42 },
		{ 19, 404, 42 },
		{ 19, 405, 42 },
		{ 20, 346, 42 },
		{ 20, 347, 42 },
		{ 20, 348, 42 },
		{ 20, 349, 42 },
		{ 20, 350, 42 },
		{ 20, 351, 42 },
		{ 20, 352, 42 },
		{ 20, 353, 42 },
		{ 20, 354, 42 },
		{ 20, 355, 42 },
		{ 20, 356, 42 },
		{ 20, 357, 42 },
		{ 20, 358, 42 },
		{ 20, 359, 42 },
		{ 20, 360, 42 },
		{ 20, 361, 42 },
		{ 20, 362, 42 },
		{ 20, 363, 42 },
		{ 20, 364, 42 },
		{ 20, 365, 42 },
		{ 20, 366, 42 },
		{ 20, 367, 42 },
		{ 20, 368, 42 },
		{ 20, 369, 42 },
		{ 20, 370, 42 },
		{ 20, 371, 42 },
		{ 20, 372, 42 },
		{ 20, 373, 42 },
		{ 20, 374, 42 },
		{ 20, 375, 42 },
		{ 20, 376, 42 },
		{ 20, 377, 42 },
		{ 20, 378, 42 },
		{ 20, 379, 42 },
		{ 20, 380, 42 },
		{ 20, 381, 42 },
		{ 20, 382, 42 },
		{ 20, 383, 42 },
		{ 20, 384, 42 },
		{ 20, 385, 42 },
		{ 20, 386, 42 },
		{ 20, 387, 42 },
		{ 20, 388, 42 },
		{ 20, 389, 42 },
		{ 20, 390, 42 },
		{ 20, 391, 42 },
		{ 20, 392, 42 },
		{ 20, 393, 42 },
		{ 20, 394, 42 },
		{ 20, 395, 42 },
		{ 20, 396, 42 },
		{ 20, 397, 42 },
		{ 20, 398, 42 },
		{ 20, 399, 42 },
		{ 20, 400, 42 },
		{ 20, 401, 42 },
		{ 20, 402, 42 },
		{ 20, 403, 42 },
		{ 20, 404, 42 },
		{ 20, 405, 42 },
		{ 21, 346, 42 },
		{ 21, 347, 42 },
		{ 21, 348, 42 },
		{ 21, 349, 42 },
		{ 21, 350, 42 },
		{ 21, 351, 42 },
		{ 21, 352, 42 },
		{ 21, 353, 42 },
		{ 21, 354, 42 },
		{ 21, 355, 42 },
		{ 21, 356, 42 },
		{ 21, 357, 42 },
		{ 21, 358, 42 },
		{ 21, 359, 42 },
		{ 21, 360, 42 },
		{ 21, 361, 42 },
		{ 21, 362, 42 },
		{ 21, 363, 42 },
		{ 21, 364, 42 },
		{ 21, 365, 42 },
		{ 21, 366, 42 },
		{ 21, 367, 42 },
		{ 21, 368, 42 },
		{ 21, 369, 42 },
		{ 21, 370, 42 },
		{ 21, 371, 42 },
		{ 21, 372, 42 },
		{ 21, 373, 42 },
		{ 21, 374, 42 },
		{ 21, 375, 42 },
		{ 21, 376, 42 },
		{ 21, 377, 42 },
		{ 21, 378, 42 },
		{ 21, 379, 42 },
		{ 21, 380, 42 },
		{ 21, 381, 42 },
		{ 21, 382, 42 },
		{ 21, 383, 42 },
		{ 21, 384, 42 },
		{ 21, 385, 42 },
		{ 21, 386, 42 },
		{ 21, 387, 42 },
		{ 21, 388, 42 },
		{ 21, 389, 42 },
		{ 21, 390, 42 },
		{ 21, 391, 42 },
		{ 21, 392, 42 },
		{ 21, 393, 42 },
		{ 21, 394, 42 },
		{ 21, 395, 42 },
		{ 21, 396, 42 },
		{ 21, 397, 42 },
		{ 21, 398, 42 },
		{ 21, 399, 42 },
		{ 21, 400, 42 },
		{ 21, 401, 42 },
		{ 21, 402, 42 },
		{ 21, 403, 42 },
		{ 21, 404, 42 },
		{ 22, 346, 42 },
		{ 22, 347, 42 },
		{ 22, 348, 42 },
		{ 22, 349, 42 },
		{ 22, 350, 42 },
		{ 22, 351, 42 },
		{ 22, 352, 42 },
		{ 22, 353, 42 },
		{ 22, 354, 42 },
		{ 22, 355, 42 },
		{ 22, 356, 42 },
		{ 22, 357, 42 },
		{ 22, 358, 42 },
		{ 22, 359, 42 },
		{ 22, 360, 42 },
		{ 22, 361, 42 },
		{ 22, 362, 42 },
		{ 22, 363, 42 },
		{ 22, 364, 42 },
		{ 22, 365, 42 },
		{ 22, 366, 42 },
		{ 22, 367, 42 },
		{ 22, 368, 42 },
		{ 22, 369, 42 },
		{ 22, 370, 42 },
		{ 22, 371, 42 },
		{ 22, 372, 42 },
		{ 22, 373, 42 },
		{ 22, 374, 42 },
		{ 22, 375, 42 },
		{ 22, 376, 42 },
		{ 22, 377, 42 },
		{ 22, 378, 42 },
		{ 22, 379, 42 },
		{ 22, 380, 42 },
		{ 22, 381, 42 },
		{ 22, 382, 42 },
		{ 22, 383, 42 },
		{ 22, 384, 42 },
		{ 22, 385, 42 },
		{ 22, 386, 42 },
		{ 22, 387, 42 },
		{ 22, 388, 42 },
		{ 22, 389, 42 },
		{ 22, 390, 42 },
		{ 22, 391, 42 },
		{ 22, 392, 42 },
		{ 22, 393, 42 },
		{ 22, 394, 42 },
		{ 22, 395, 42 },
		{ 22, 396, 42 },
		{ 22, 397, 42 },
		{ 22, 398, 42 },
		{ 22, 399, 42 },
		{ 22, 400, 42 },
		{ 22, 401, 42 },
		{ 22, 402, 42 },
		{ 22, 403, 42 },
		{ 23, 345, 42 },
		{ 23, 346, 42 },
		{ 23, 347, 42 },
		{ 23, 348, 42 },
		{ 23, 349, 42 },
		{ 23, 350, 42 },
		{ 23, 351, 42 },
		{ 23, 352, 42 },
		{ 23, 353, 42 },
		{ 23, 354, 42 },
		{ 23, 355, 42 },
		{ 23, 356, 42 },
		{ 23, 357, 42 },
		{ 23, 358, 42 },
		{ 23, 359, 42 },
		{ 23, 360, 42 },
		{ 23, 361, 42 },
		{ 23, 362, 42 },
		{ 23, 363, 42 },
		{ 23, 364, 42 },
		{ 23, 365, 42 },
		{ 23, 366, 42 },
		{ 23, 367, 42 },
		{ 23, 368, 42 },
		{ 23, 369, 42 },
		{ 23, 370, 42 },
		{ 23, 371, 42 },
		{ 23, 372, 42 },
		{ 23, 373, 42 },
		{ 23, 374, 42 },
		{ 23, 375, 42 },
		{ 23, 376, 42 },
		{ 23, 377, 42 },
		{ 23, 378, 42 },
		{ 23, 379, 42 },
		{ 23, 380, 42 },
		{ 23, 381, 42 },
		{ 23, 382, 42 },
		{ 23, 383, 42 },
		{ 23, 384, 42 },
		{ 23, 385, 42 },
		{ 23, 386, 42 },
		{ 23, 387, 42 },
		{ 23, 388, 42 },
		{ 23, 389, 42 },
		{ 23, 390, 42 },
		{ 23, 391, 42 },
		{ 23, 392, 42 },
		{ 23, 393, 42 },
		{ 23, 394, 42 },
		{ 23, 395, 42 },
		{ 23, 396, 42 },
		{ 23, 397, 42 },
		{ 23, 398, 42 },
		{ 23, 399, 42 },
		{ 23, 400, 42 },
		{ 23, 401, 42 },
		{ 23, 402, 42 },
		{ 23, 403, 42 },
		{ 24, 345, 42 },
		{ 24, 346, 42 },
		{ 24, 347, 42 },
		{ 24, 348, 42 },
		{ 24, 349, 42 },
		{ 24, 350, 42 },
		{ 24, 351, 42 },
		{ 24, 352, 42 },
		{ 24, 353, 42 },
		{ 24, 354, 42 },
		{ 24, 355, 42 },
		{ 24, 356, 42 },
		{ 24, 357, 42 },
		{ 24, 358, 42 },
		{ 24, 359, 42 },
		{ 24, 360, 42 },
		{ 24, 361, 42 },
		{ 24, 362, 42 },
		{ 24, 363, 42 },
		{ 24, 364, 42 },
		{ 24, 365, 42 },
		{ 24, 366, 42 },
		{ 24, 367, 42 },
		{ 24, 368, 42 },
		{ 24, 369, 42 },
		{ 24, 370, 42 },
		{ 24, 371, 42 },
		{ 24, 372, 42 },
		{ 24, 373, 42 },
		{ 24, 374, 42 },
		{ 24, 375, 42 },
		{ 24, 376, 42 },
		{ 24, 377, 42 },
		{ 24, 378, 42 },
		{ 24, 379, 42 },
		{ 24, 380, 42 },
		{ 24, 381, 42 },
		{ 24, 382, 42 },
		{ 24, 383, 42 },
		{ 24, 384, 42 },
		{ 24, 385, 42 },
		{ 24, 386, 42 },
		{ 24, 387, 42 },
		{ 24, 388, 42 },
		{ 24, 389, 42 },
		{ 24, 390, 42 },
		{ 24, 391, 42 },
		{ 24, 392, 42 },
		{ 24, 393, 42 },
		{ 24, 394, 42 },
		{ 24, 395, 42 },
		{ 24, 396, 42 },
		{ 24, 397, 42 },
		{ 24, 398, 42 },
		{ 24, 399, 42 },
		{ 24, 400, 42 },
		{ 24, 401, 42 },
		{ 24, 402, 42 },
		{ 25, 345, 42 },
		{ 25, 346, 42 },
		{ 25, 347, 42 },
		{ 25, 348, 42 },
		{ 25, 349, 42 },
		{ 25, 350, 42 },
		{ 25, 351, 42 },
		{ 25, 352, 42 },
		{ 25, 353, 42 },
		{ 25, 354, 42 },
		{ 25, 355, 42 },
		{ 25, 356, 42 },
		{ 25, 357, 42 },
		{ 25, 358, 42 },
		{ 25, 359, 42 },
		{ 25, 360, 42 },
		{ 25, 361, 42 },
		{ 25, 362, 42 },
		{ 25, 363, 42 },
		{ 25, 364, 42 },
		{ 25, 365, 42 },
		{ 25, 366, 42 },
		{ 25, 367, 42 },
		{ 25, 368, 42 },
		{ 25, 369, 42 },
		{ 25, 370, 42 },
		{ 25, 371, 42 },
		{ 25, 372, 42 },
		{ 25, 373, 42 },
		{ 25, 374, 42 },
		{ 25, 375, 42 },
		{ 25, 376, 42 },
		{ 25, 377, 42 },
		{ 25, 378, 42 },
		{ 25, 379, 42 },
		{ 25, 380, 42 },
		{ 25, 381, 42 },
		{ 25, 382, 42 },
		{ 25, 383, 42 },
		{ 25, 384, 42 },
		{ 25, 385, 42 },
		{ 25, 386, 42 },
		{ 25, 387, 42 },
		{ 25, 388, 42 },
		{ 25, 389, 42 },
		{ 25, 390, 42 },
		{ 25, 391, 42 },
		{ 25, 392, 42 },
		{ 25, 393, 42 },
		{ 25, 394, 42 },
		{ 25, 395, 42 },
		{ 25, 396, 42 },
		{ 25, 397, 42 },
		{ 25, 398, 42 },
		{ 25, 399, 42 },
		{ 25, 400, 42 },
		{ 25, 401, 42 },
		{ 26, 344, 42 },
		{ 26, 345, 42 },
		{ 26, 346, 42 },
		{ 26, 347, 42 },
		{ 26, 348, 42 },
		{ 26, 349, 42 },
		{ 26, 350, 42 },
		{ 26, 351, 42 },
		{ 26, 352, 42 },
		{ 26, 353, 42 },
		{ 26, 354, 42 },
		{ 26, 355, 42 },
		{ 26, 356, 42 },
		{ 26, 357, 42 },
		{ 26, 358, 42 },
		{ 26, 359, 42 },
		{ 26, 360, 42 },
		{ 26, 361, 42 },
		{ 26, 362, 42 },
		{ 26, 363, 42 },
		{ 26, 364, 42 },
		{ 26, 365, 42 },
		{ 26, 366, 42 },
		{ 26, 367, 42 },
		{ 26, 368, 42 },
		{ 26, 369, 42 },
		{ 26, 370, 42 },
		{ 26, 371, 42 },
		{ 26, 372, 42 },
		{ 26, 373, 42 },
		{ 26, 374, 42 },
		{ 26, 375, 42 },
		{ 26, 376, 42 },
		{ 26, 377, 42 },
		{ 26, 378, 42 },
		{ 26, 379, 42 },
		{ 26, 380, 42 },
		{ 26, 381, 42 },
		{ 26, 382, 42 },
		{ 26, 383, 42 },
		{ 26, 384, 42 },
		{ 26, 385, 42 },
		{ 26, 386, 42 },
		{ 26, 387, 42 },
		{ 26, 388, 42 },
		{ 26, 389, 42 },
		{ 26, 390, 42 },
		{ 26, 391, 42 },
		{ 26, 392, 42 },
		{ 26, 393, 42 },
		{ 26, 394, 42 },
		{ 26, 395, 42 },
		{ 26, 396, 42 },
		{ 26, 397, 42 },
		{ 26, 398, 42 },
		{ 26, 399, 42 },
		{ 26, 400, 42 },
		{ 26, 401, 42 },
		{ 27, 344, 42 },
		{ 27, 345, 42 },
		{ 27, 346, 42 },
		{ 27, 347, 42 },
		{ 27, 348, 42 },
		{ 27, 349, 42 },
		{ 27, 350, 42 },
		{ 27, 351, 42 },
		{ 27, 352, 42 },
		{ 27, 353, 42 },
		{ 27, 354, 42 },
		{ 27, 355, 42 },
		{ 27, 356, 42 },
		{ 27, 357, 42 },
		{ 27, 358, 42 },
		{ 27, 359, 42 },
		{ 27, 360, 42 },
		{ 27, 361, 42 },
		{ 27, 362, 42 },
		{ 27, 363, 42 },
		{ 27, 364, 42 },
		{ 27, 365, 42 },
		{ 27, 366, 42 },
		{ 27, 367, 42 },
		{ 27, 368, 42 },
		{ 27, 369, 42 },
		{ 27, 370, 42 },
		{ 27, 371, 42 },
		{ 27, 372, 42 },
		{ 27, 373, 42 },
		{ 27, 374, 42 },
		{ 27, 375, 42 },
		{ 27, 376, 42 },
		{ 27, 377, 42 },
		{ 27, 378, 42 },
		{ 27, 379, 42 },
		{ 27, 380, 42 },
		{ 27, 381, 42 },
		{ 27, 382, 42 },
		{ 27, 383, 42 },
		{ 27, 384, 42 },
		{ 27, 385, 42 },
		{ 27, 386, 42 },
		{ 27, 387, 42 },
		{ 27, 388, 42 },
		{ 27, 389, 42 },
		{ 27, 390, 42 },
		{ 27, 391, 42 },
		{ 27, 392, 42 },
		{ 27, 393, 42 },
		{ 27, 394, 42 },
		{ 27, 395, 42 },
		{ 27, 396, 42 },
		{ 27, 397, 42 },
		{ 27, 398, 42 },
		{ 27, 399, 42 },
		{ 27, 400, 42 },
		{ 28, 343, 42 },
		{ 28, 344, 42 },
		{ 28, 345, 42 },
		{ 28, 346, 42 },
		{ 28, 347, 42 },
		{ 28, 348, 42 },
		{ 28, 349, 42 },
		{ 28, 350, 42 },
		{ 28, 351, 42 },
		{ 28, 352, 42 },
		{ 28, 353, 42 },
		{ 28, 354, 42 },
		{ 28, 355, 42 },
		{ 28, 356, 42 },
		{ 28, 357, 42 },
		{ 28, 358, 42 },
		{ 28, 359, 42 },
		{ 28, 360, 42 },
		{ 28, 361, 42 },
		{ 28, 362, 42 },
		{ 28, 363, 42 },
		{ 28, 364, 42 },
		{ 28, 365, 42 },
		{ 28, 366, 42 },
		{ 28, 367, 42 },
		{ 28, 368, 42 },
		{ 28, 369, 42 },
		{ 28, 370, 42 },
		{ 28, 371, 42 },
		{ 28, 372, 42 },
		{ 28, 373, 42 },
		{ 28, 374, 42 },
		{ 28, 375, 42 },
		{ 28, 376, 42 },
		{ 28, 377, 42 },
		{ 28, 378, 42 },
		{ 28, 379, 42 },
		{ 28, 380, 42 },
		{ 28, 381, 42 },
		{ 28, 382, 42 },
		{ 28, 383, 42 },
		{ 28, 384, 42 },
		{ 28, 385, 42 },
		{ 28, 386, 42 },
		{ 28, 387, 42 },
		{ 28, 388, 42 },
		{ 28, 389, 42 },
		{ 28, 390, 42 },
		{ 28, 391, 42 },
		{ 28, 392, 42 },
		{ 28, 393, 42 },
		{ 28, 394, 42 },
		{ 28, 395, 42 },
		{ 28, 396, 42 },
		{ 28, 397, 42 },
		{ 28, 398, 42 },
		{ 28, 399, 42 },
		{ 28, 400, 42 },
		{ 29, 343, 42 },
		{ 29, 344, 42 },
		{ 29, 345, 42 },
		{ 29, 346, 42 },
		{ 29, 347, 42 },
		{ 29, 348, 42 },
		{ 29, 349, 42 },
		{ 29, 350, 42 },
		{ 29, 351, 42 },
		{ 29, 352, 42 },
		{ 29, 353, 42 },
		{ 29, 354, 42 },
		{ 29, 355, 42 },
		{ 29, 356, 42 },
		{ 29, 357, 42 },
		{ 29, 358, 42 },
		{ 29, 359, 42 },
		{ 29, 360, 42 },
		{ 29, 361, 42 },
		{ 29, 362, 42 },
		{ 29, 363, 42 },
		{ 29, 364, 42 },
		{ 29, 365, 42 },
		{ 29, 366, 42 },
		{ 29, 367, 42 },
		{ 29, 368, 42 },
		{ 29, 369, 42 },
		{ 29, 370, 42 },
		{ 29, 371, 42 },
		{ 29, 372, 42 },
		{ 29, 373, 42 },
		{ 29, 374, 42 },
		{ 29, 375, 42 },
		{ 29, 376, 42 },
		{ 29, 377, 42 },
		{ 29, 378, 42 },
		{ 29, 379, 42 },
		{ 29, 380, 42 },
		{ 29, 381, 42 },
		{ 29, 382, 42 },
		{ 29, 383, 42 },
		{ 29, 384, 42 },
		{ 29, 385, 42 },
		{ 29, 386, 42 },
		{ 29, 387, 42 },
		{ 29, 388, 42 },
		{ 29, 389, 42 },
		{ 29, 390, 42 },
		{ 29, 391, 42 },
		{ 29, 392, 42 },
		{ 29, 393, 42 },
		{ 29, 394, 42 },
		{ 29, 395, 42 },
		{ 29, 396, 42 },
		{ 29, 397, 42 },
		{ 29, 398, 42 },
		{ 29, 399, 42 },
		{ 30, 343, 42 },
		{ 30, 344, 42 },
		{ 30, 345, 42 },
		{ 30, 346, 42 },
		{ 30, 347, 42 },
		{ 30, 348, 42 },
		{ 30, 349, 42 },
		{ 30, 350, 42 },
		{ 30, 351, 42 },
		{ 30, 352, 42 },
		{ 30, 353, 42 },
		{ 30, 354, 42 },
		{ 30, 355, 42 },
		{ 30, 356, 42 },
		{ 30, 357, 42 },
		{ 30, 358, 42 },
		{ 30, 359, 42 },
		{ 30, 360, 42 },
		{ 30, 361, 42 },
		{ 30, 362, 42 },
		{ 30, 363, 42 },
		{ 30, 364, 42 },
		{ 30, 365, 42 },
		{ 30, 366, 42 },
		{ 30, 367, 42 },
		{ 30, 368, 42 },
		{ 30, 369, 42 },
		{ 30, 370, 42 },
		{ 30, 371, 42 },
		{ 30, 372, 42 },
		{ 30, 373, 42 },
		{ 30, 374, 42 },
		{ 30, 375, 42 },
		{ 30, 376, 42 },
		{ 30, 377, 42 },
		{ 30, 378, 42 },
		{ 30, 379, 42 },
		{ 30, 380, 42 },
		{ 30, 381, 42 },
		{ 30, 382, 42 },
		{ 30, 383, 42 },
		{ 30, 384, 42 },
		{ 30, 385, 42 },
		{ 30, 386, 42 },
		{ 30, 387, 42 },
		{ 30, 388, 42 },
		{ 30, 389, 42 },
		{ 30, 390, 42 },
		{ 30, 391, 42 },
		{ 30, 392, 42 },
		{ 30, 393, 42 },
		{ 30, 394, 42 },
		{ 30, 395, 42 },
		{ 30, 396, 42 },
		{ 30, 397, 42 },
		{ 30, 398, 42 },
		{ 31, 342, 42 },
		{ 31, 343, 42 },
		{ 31, 344, 42 },
		{ 31, 345, 42 },
		{ 31, 346, 42 },
		{ 31, 347, 42 },
		{ 31, 348, 42 },
		{ 31, 349, 42 },
		{ 31, 350, 42 },
		{ 31, 351, 42 },
		{ 31, 352, 42 },
		{ 31, 353, 42 },
		{ 31, 354, 42 },
		{ 31, 355, 42 },
		{ 31, 356, 42 },
		{ 31, 357, 42 },
		{ 31, 358, 42 },
		{ 31, 359, 42 },
		{ 31, 360, 42 },
		{ 31, 361, 42 },
		{ 31, 362, 42 },
		{ 31, 363, 42 },
		{ 31, 364, 42 },
		{ 31, 365, 42 },
		{ 31, 366, 42 },
		{ 31, 367, 42 },
		{ 31, 368, 42 },
		{ 31, 369, 42 },
		{ 31, 370, 42 },
		{ 31, 371, 42 },
		{ 31, 372, 42 },
		{ 31, 373, 42 },
		{ 31, 374, 42 },
		{ 31, 375, 42 },
		{ 31, 376, 42 },
		{ 31, 377, 42 },
		{ 31, 378, 42 },
		{ 31, 379, 42 },
		{ 31, 380, 42 },
		{ 31, 381, 42 },
		{ 31, 382, 42 },
		{ 31, 383, 42 },
		{ 31, 384, 42 },
		{ 31, 385, 42 },
		{ 31, 386, 42 },
		{ 31, 387, 42 },
		{ 31, 388, 42 },
		{ 31, 389, 42 },
		{ 31, 390, 42 },
		{ 31, 391, 42 },
		{ 31, 392, 42 },
		{ 31, 393, 42 },
		{ 31, 394, 42 },
		{ 31, 395, 42 },
		{ 31, 396, 42 },
		{ 31, 397, 42 },
		{ 31, 398, 42 },
		{ 32, 342, 42 },
		{ 32, 343, 42 },
		{ 32, 344, 42 },
		{ 32, 345, 42 },
		{ 32, 346, 42 },
		{ 32, 347, 42 },
		{ 32, 348, 42 },
		{ 32, 349, 42 },
		{ 32, 350, 42 },
		{ 32, 351, 42 },
		{ 32, 352, 42 },
		{ 32, 353, 42 },
		{ 32, 354, 42 },
		{ 32, 355, 42 },
		{ 32, 356, 42 },
		{ 32, 357, 42 },
		{ 32, 358, 42 },
		{ 32, 359, 42 },
		{ 32, 360, 42 },
		{ 32, 361, 42 },
		{ 32, 362, 42 },
		{ 32, 363, 42 },
		{ 32, 364, 42 },
		{ 32, 365, 42 },
		{ 32, 366, 42 },
		{ 32, 367, 42 },
		{ 32, 368, 42 },
		{ 32, 369, 42 },
		{ 32, 370, 42 },
		{ 32, 371, 42 },
		{ 32, 372, 42 },
		{ 32, 373, 42 },
		{ 32, 374, 42 },
		{ 32, 375, 42 },
		{ 32, 376, 42 },
		{ 32, 377, 42 },
		{ 32, 378, 42 },
		{ 32, 379, 42 },
		{ 32, 380, 42 },
		{ 32, 381, 42 },
		{ 32, 382, 42 },
		{ 32, 383, 42 },
		{ 32, 384, 42 },
		{ 32, 385, 42 },
		{ 32, 386, 42 },
		{ 32, 387, 42 },
		{ 32, 388, 42 },
		{ 32, 389, 42 },
		{ 32, 390, 42 },
		{ 32, 391, 42 },
		{ 32, 392, 42 },
		{ 32, 393, 42 },
		{ 32, 394, 42 },
		{ 32, 395, 42 },
		{ 32, 396, 42 },
		{ 32, 397, 42 },
		{ 33, 342, 42 },
		{ 33, 343, 42 },
		{ 33, 344, 42 },
		{ 33, 345, 42 },
		{ 33, 346, 42 },
		{ 33, 347, 42 },
		{ 33, 348, 42 },
		{ 33, 349, 42 },
		{ 33, 350, 42 },
		{ 33, 351, 42 },
		{ 33, 352, 42 },
		{ 33, 353, 42 },
		{ 33, 354, 42 },
		{ 33, 355, 42 },
		{ 33, 356, 42 },
		{ 33, 357, 42 },
		{ 33, 358, 42 },
		{ 33, 359, 42 },
		{ 33, 360, 42 },
		{ 33, 361, 42 },
		{ 33, 362, 42 },
		{ 33, 363, 42 },
		{ 33, 364, 42 },
		{ 33, 365, 42 },
		{ 33, 366, 42 },
		{ 33, 367, 42 },
		{ 33, 368, 42 },
		{ 33, 369, 42 },
		{ 33, 370, 42 },
		{ 33, 371, 42 },
		{ 33, 372, 42 },
		{ 33, 373, 42 },
		{ 33, 374, 42 },
		{ 33, 375, 42 },
		{ 33, 376, 42 },
		{ 33, 377, 42 },
		{ 33, 378, 42 },
		{ 33, 379, 42 },
		{ 33, 380, 42 },
		{ 33, 381, 42 },
		{ 33, 382, 42 },
		{ 33, 383, 42 },
		{ 33, 384, 42 },
		{ 33, 385, 42 },
		{ 33, 386, 42 },
		{ 33, 387, 42 },
		{ 33, 388, 42 },
		{ 33, 389, 42 },
		{ 33, 390, 42 },
		{ 33, 391, 42 },
		{ 33, 392, 42 },
		{ 33, 393, 42 },
		{ 33, 394, 42 },
		{ 33, 395, 42 },
		{ 33, 396, 42 },
		{ 34, 341, 42 },
		{ 34, 342, 42 },
		{ 34, 343, 42 },
		{ 34, 344, 42 },
		{ 34, 345, 42 },
		{ 34, 346, 42 },
		{ 34, 347, 42 },
		{ 34, 348, 42 },
		{ 34, 349, 42 },
		{ 34, 350, 42 },
		{ 34, 351, 42 },
		{ 34, 352, 42 },
		{ 34, 353, 42 },
		{ 34, 354, 42 },
		{ 34, 355, 42 },
		{ 34, 356, 42 },
		{ 34, 357, 42 },
		{ 34, 358, 42 },
		{ 34, 359, 42 },
		{ 34, 360, 42 },
		{ 34, 361, 42 },
		{ 34, 362, 42 },
		{ 34, 363, 42 },
		{ 34, 364, 42 },
		{ 34, 365, 42 },
		{ 34, 366, 42 },
		{ 34, 367, 42 },
		{ 34, 368, 42 },
		{ 34, 369, 42 },
		{ 34, 370, 42 },
		{ 34, 371, 42 },
		{ 34, 372, 42 },
		{ 34, 373, 42 },
		{ 34, 374, 42 },
		{ 34, 375, 42 },
		{ 34, 376, 42 },
		{ 34, 377, 42 },
		{ 34, 378, 42 },
		{ 34, 379, 42 },
		{ 34, 380, 42 },
		{ 34, 381, 42 },
		{ 34, 382, 42 },
		{ 34, 383, 42 },
		{ 34, 384, 42 },
		{ 34, 385, 42 },
		{ 34, 386, 42 },
		{ 34, 387, 42 },
		{ 34, 388, 42 },
		{ 34, 389, 42 },
		{ 34, 390, 42 },
		{ 34, 391, 42 },
		{ 34, 392, 42 },
		{ 34, 393, 42 },
		{ 34, 394, 42 },
		{ 34, 395, 42 },
		{ 34, 396, 42 },
		{ 35, 341, 42 },
		{ 35, 342, 42 },
		{ 35, 343, 42 },
		{ 35, 344, 42 },
		{ 35, 345, 42 },
		{ 35, 346, 42 },
		{ 35, 347, 42 },
		{ 35, 348, 42 },
		{ 35, 349, 42 },
		{ 35, 350, 42 },
		{ 35, 351, 42 },
		{ 35, 352, 42 },
		{ 35, 353, 42 },
		{ 35, 354, 42 },
		{ 35, 355, 42 },
		{ 35, 356, 42 },
		{ 35, 357, 42 },
		{ 35, 358, 42 },
		{ 35, 359, 42 },
		{ 35, 360, 42 },
		{ 35, 361, 42 },
		{ 35, 362, 42 },
		{ 35, 363, 42 },
		{ 35, 364, 42 },
		{ 35, 365, 42 },
		{ 35, 366, 42 },
		{ 35, 367, 42 },
		{ 35, 368, 42 },
		{ 35, 369, 42 },
		{ 35, 370, 42 },
		{ 35, 371, 42 },
		{ 35, 372, 42 },
		{ 35, 373, 42 },
		{ 35, 374, 42 },
		{ 35, 375, 42 },
		{ 35, 376, 42 },
		{ 35, 377, 42 },
		{ 35, 378, 42 },
		{ 35, 379, 42 },
		{ 35, 380, 42 },
		{ 35, 381, 42 },
		{ 35, 382, 42 },
		{ 35, 383, 42 },
		{ 35, 384, 42 },
		{ 35, 385, 42 },
		{ 35, 386, 42 },
		{ 35, 387, 42 },
		{ 35, 388, 42 },
		{ 35, 389, 42 },
		{ 35, 390, 42 },
		{ 35, 391, 42 },
		{ 35, 392, 42 },
		{ 35, 393, 42 },
		{ 35, 394, 42 },
		{ 35, 395, 42 },
		{ 36, 340, 42 },
		{ 36, 341, 42 },
		{ 36, 342, 42 },
		{ 36, 343, 42 },
		{ 36, 344, 42 },
		{ 36, 345, 42 },
		{ 36, 346, 42 },
		{ 36, 347, 42 },
		{ 36, 348, 42 },
		{ 36, 349, 42 },
		{ 36, 350, 42 },
		{ 36, 351, 42 },
		{ 36, 352, 42 },
		{ 36, 353, 42 },
		{ 36, 354, 42 },
		{ 36, 355, 42 },
		{ 36, 356, 42 },
		{ 36, 357, 42 },
		{ 36, 358, 42 },
		{ 36, 359, 42 },
		{ 36, 360, 42 },
		{ 36, 361, 42 },
		{ 36, 362, 42 },
		{ 36, 363, 42 },
		{ 36, 364, 42 },
		{ 36, 365, 42 },
		{ 36, 366, 42 },
		{ 36, 367, 42 },
		{ 36, 368, 42 },
		{ 36, 369, 42 },
		{ 36, 370, 42 },
		{ 36, 371, 42 },
		{ 36, 372, 42 },
		{ 36, 373, 42 },
		{ 36, 374, 42 },
		{ 36, 375, 42 },
		{ 36, 376, 42 },
		{ 36, 377, 42 },
		{ 36, 378, 42 },
		{ 36, 379, 42 },
		{ 36, 380, 42 },
		{ 36, 381, 42 },
		{ 36, 382, 42 },
		{ 36, 383, 42 },
		{ 36, 384, 42 },
		{ 36, 385, 42 },
		{ 36, 386, 42 },
		{ 36, 387, 42 },
		{ 36, 388, 42 },
		{ 36, 389, 42 },
		{ 36, 390, 42 },
		{ 36, 391, 42 },
		{ 36, 392, 42 },
		{ 36, 393, 42 },
		{ 36, 394, 42 },
		{ 36, 395, 42 },
		{ 37, 340, 42 },
		{ 37, 341, 42 },
		{ 37, 342, 42 },
		{ 37, 343, 42 },
		{ 37, 344, 42 },
		{ 37, 345, 42 },
		{ 37, 346, 42 },
		{ 37, 347, 42 },
		{ 37, 348, 42 },
		{ 37, 349, 42 },
		{ 37, 350, 42 },
		{ 37, 351, 42 },
		{ 37, 352, 42 },
		{ 37, 353, 42 },
		{ 37, 354, 42 },
		{ 37, 355, 42 },
		{ 37, 356, 42 },
		{ 37, 357, 42 },
		{ 37, 358, 42 },
		{ 37, 359, 42 },
		{ 37, 360, 42 },
		{ 37, 361, 42 },
		{ 37, 362, 42 },
		{ 37, 363, 42 },
		{ 37, 364, 42 },
		{ 37, 365, 42 },
		{ 37, 366, 42 },
		{ 37, 367, 42 },
		{ 37, 368, 42 },
		{ 37, 369, 42 },
		{ 37, 370, 42 },
		{ 37, 371, 42 },
		{ 37, 372, 42 },
		{ 37, 373, 42 },
		{ 37, 374, 42 },
		{ 37, 375, 42 },
		{ 37, 376, 42 },
		{ 37, 377, 42 },
		{ 37, 378, 42 },
		{ 37, 379, 42 },
		{ 37, 380, 42 },
		{ 37, 381, 42 },
		{ 37, 382, 42 },
		{ 37, 383, 42 },
		{ 37, 384, 42 },
		{ 37, 385, 42 },
		{ 37, 386, 42 },
		{ 37, 387, 42 },
		{ 37, 388, 42 },
		{ 37, 389, 42 },
		{ 37, 390, 42 },
		{ 37, 391, 42 },
		{ 37, 392, 42 },
		{ 37, 393, 42 },
		{ 37, 394, 42 },
		{ 38, 340, 42 },
		{ 38, 341, 42 },
		{ 38, 342, 42 },
		{ 38, 343, 42 },
		{ 38, 344, 42 },
		{ 38, 345, 42 },
		{ 38, 346, 42 },
		{ 38, 347, 42 },
		{ 38, 348, 42 },
		{ 38, 349, 42 },
		{ 38, 350, 42 },
		{ 38, 351, 42 },
		{ 38, 352, 42 },
		{ 38, 353, 42 },
		{ 38, 354, 42 },
		{ 38, 355, 42 },
		{ 38, 356, 42 },
		{ 38, 357, 42 },
		{ 38, 358, 42 },
		{ 38, 359, 42 },
		{ 38, 360, 42 },
		{ 38, 361, 42 },
		{ 38, 362, 42 },
		{ 38, 363, 42 },
		{ 38, 364, 42 },
		{ 38, 365, 42 },
		{ 38, 366, 42 },
		{ 38, 367, 42 },
		{ 38, 368, 42 },
		{ 38, 369, 42 },
		{ 38, 370, 42 },
		{ 38, 371, 42 },
		{ 38, 372, 42 },
		{ 38, 373, 42 },
		{ 38, 374, 42 },
		{ 38, 375, 42 },
		{ 38, 376, 42 },
		{ 38, 377, 42 },
		{ 38, 378, 42 },
		{ 38, 379, 42 },
		{ 38, 380, 42 },
		{ 38, 381, 42 },
		{ 38, 382, 42 },
		{ 38, 383, 42 },
		{ 38, 384, 42 },
		{ 38, 385, 42 },
		{ 38, 386, 42 },
		{ 38, 387, 42 },
		{ 38, 388, 42 },
		{ 38, 389, 42 },
		{ 38, 390, 42 },
		{ 38, 391, 42 },
		{ 38, 392, 42 },
		{ 38, 393, 42 },
		{ 39, 339, 42 },
		{ 39, 340, 42 },
		{ 39, 341, 42 },
		{ 39, 342, 42 },
		{ 39, 343, 42 },
		{ 39, 344, 42 },
		{ 39, 345, 42 },
		{ 39, 346, 42 },
		{ 39, 347, 42 },
		{ 39, 348, 42 },
		{ 39, 349, 42 },
		{ 39, 350, 42 },
		{ 39, 351, 42 },
		{ 39, 352, 42 },
		{ 39, 353, 42 },
		{ 39, 354, 42 },
		{ 39, 355, 42 },
		{ 39, 356, 42 },
		{ 39, 357, 42 },
		{ 39, 358, 42 },
		{ 39, 359, 42 },
		{ 39, 360, 42 },
		{ 39, 361, 42 },
		{ 39, 362, 42 },
		{ 39, 363, 42 },
		{ 39, 364, 42 },
		{ 39, 365, 42 },
		{ 39, 366, 42 },
		{ 39, 367, 42 },
		{ 39, 368, 42 },
		{ 39, 369, 42 },
		{ 39, 370, 42 },
		{ 39, 371, 42 },
		{ 39, 372, 42 },
		{ 39, 373, 42 },
		{ 39, 374, 42 },
		{ 39, 375, 42 },
		{ 39, 376, 42 },
		{ 39, 377, 42 },
		{ 39, 378, 42 },
		{ 39, 379, 42 },
		{ 39, 380, 42 },
		{ 39, 381, 42 },
		{ 39, 382, 42 },
		{ 39, 383, 42 },
		{ 39, 384, 42 },
		{ 39, 385, 42 },
		{ 39, 386, 42 },
		{ 39, 387, 42 },
		{ 39, 388, 42 },
		{ 39, 389, 42 },
		{ 39, 390, 42 },
		{ 39, 391, 42 },
		{ 39, 392, 42 },
		{ 39, 393, 42 },
		{ 40, 339, 42 },
		{ 40, 340, 42 },
		{ 40, 341, 42 },
		{ 40, 342, 42 },
		{ 40, 343, 42 },
		{ 40, 344, 42 },
		{ 40, 345, 42 },
		{ 40, 346, 42 },
		{ 40, 347, 42 },
		{ 40, 348, 42 },
		{ 40, 349, 42 },
		{ 40, 350, 42 },
		{ 40, 351, 42 },
		{ 40, 352, 42 },
		{ 40, 353, 42 },
		{ 40, 354, 42 },
		{ 40, 355, 42 },
		{ 40, 356, 42 },
		{ 40, 357, 42 },
		{ 40, 358, 42 },
		{ 40, 359, 42 },
		{ 40, 360, 42 },
		{ 40, 361, 42 },
		{ 40, 362, 42 },
		{ 40, 363, 42 },
		{ 40, 364, 42 },
		{ 40, 365, 42 },
		{ 40, 366, 42 },
		{ 40, 367, 42 },
		{ 40, 368, 42 },
		{ 40, 369, 42 },
		{ 40, 370, 42 },
		{ 40, 371, 42 },
		{ 40, 372, 42 },
		{ 40, 373, 42 },
		{ 40, 374, 42 },
		{ 40, 375, 42 },
		{ 40, 376, 42 },
		{ 40, 377, 42 },
		{ 40, 378, 42 },
		{ 40, 379, 42 },
		{ 40, 380, 42 },
		{ 40, 381, 42 },
		{ 40, 382, 42 },
		{ 40, 383, 42 },
		{ 40, 384, 42 },
		{ 40, 385, 42 },
		{ 40, 386, 42 },
		{ 40, 387, 42 },
		{ 40, 388, 42 },
		{ 40, 389, 42 },
		{ 40, 390, 42 },
		{ 40, 391, 42 },
		{ 40, 392, 42 },
		{ 41, 339, 42 },
		{ 41, 340, 42 },
		{ 41, 341, 42 },
		{ 41, 342, 42 },
		{ 41, 343, 42 },
		{ 41, 344, 42 },
		{ 41, 345, 42 },
		{ 41, 346, 42 },
		{ 41, 347, 42 },
		{ 41, 348, 42 },
		{ 41, 349, 42 },
		{ 41, 350, 42 },
		{ 41, 351, 42 },
		{ 41, 352, 42 },
		{ 41, 353, 42 },
		{ 41, 354, 42 },
		{ 41, 355, 42 },
		{ 41, 356, 42 },
		{ 41, 357, 42 },
		{ 41, 358, 42 },
		{ 41, 359, 42 },
		{ 41, 360, 42 },
		{ 41, 361, 42 },
		{ 41, 362, 42 },
		{ 41, 363, 42 },
		{ 41, 364, 42 },
		{ 41, 365, 42 },
		{ 41, 366, 42 },
		{ 41, 367, 42 },
		{ 41, 368, 42 },
		{ 41, 369, 42 },
		{ 41, 370, 42 },
		{ 41, 371, 42 },
		{ 41, 372, 42 },
		{ 41, 373, 42 },
		{ 41, 374, 42 },
		{ 41, 375, 42 },
		{ 41, 376, 42 },
		{ 41, 377, 42 },
		{ 41, 378, 42 },
		{ 41, 379, 42 },
		{ 41, 380, 42 },
		{ 41, 381, 42 },
		{ 41, 382, 42 },
		{ 41, 383, 42 },
		{ 41, 384, 42 },
		{ 41, 385, 42 },
		{ 41, 386, 42 },
		{ 41, 387, 42 },
		{ 41, 388, 42 },
		{ 41, 389, 42 },
		{ 41, 390, 42 },
		{ 41, 391, 42 },
		{ 42, 338, 42 },
		{ 42, 339, 42 },
		{ 42, 340, 42 },
		{ 42, 341, 42 },
		{ 42, 342, 42 },
		{ 42, 343, 42 },
		{ 42, 344, 42 },
		{ 42, 345, 42 },
		{ 42, 346, 42 },
		{ 42, 347, 42 },
		{ 42, 348, 42 },
		{ 42, 349, 42 },
		{ 42, 350, 42 },
		{ 42, 351, 42 },
		{ 42, 352, 42 },
		{ 42, 353, 42 },
		{ 42, 354, 42 },
		{ 42, 355, 42 },
		{ 42, 356, 42 },
		{ 42, 357, 42 },
		{ 42, 358, 42 },
		{ 42, 359, 42 },
		{ 42, 360, 42 },
		{ 42, 361, 42 },
		{ 42, 362, 42 },
		{ 42, 363, 42 },
		{ 42, 364, 42 },
		{ 42, 365, 42 },
		{ 42, 366, 42 },
		{ 42, 367, 42 },
		{ 42, 368, 42 },
		{ 42, 369, 42 },
		{ 42, 370, 42 },
		{ 42, 371, 42 },
		{ 42, 372, 42 },
		{ 42, 373, 42 },
		{ 42, 374, 42 },
		{ 42, 375, 42 },
		{ 42, 376, 42 },
		{ 42, 377, 42 },
		{ 42, 378, 42 },
		{ 42, 379, 42 },
		{ 42, 380, 42 },
		{ 42, 381, 42 },
		{ 42, 382, 42 },
		{ 42, 383, 42 },
		{ 42, 384, 42 },
		{ 42, 385, 42 },
		{ 42, 386, 42 },
		{ 42, 387, 42 },
		{ 42, 388, 42 },
		{ 42, 389, 42 },
		{ 42, 390, 42 },
		{ 42, 391, 42 },
		{ 43, 338, 42 },
		{ 43, 339, 42 },
		{ 43, 340, 42 },
		{ 43, 341, 42 },
		{ 43, 342, 42 },
		{ 43, 343, 42 },
		{ 43, 344, 42 },
		{ 43, 345, 42 },
		{ 43, 346, 42 },
		{ 43, 347, 42 },
		{ 43, 348, 42 },
		{ 43, 349, 42 },
		{ 43, 350, 42 },
		{ 43, 351, 42 },
		{ 43, 352, 42 },
		{ 43, 353, 42 },
		{ 43, 354, 42 },
		{ 43, 355, 42 },
		{ 43, 356, 42 },
		{ 43, 357, 42 },
		{ 43, 358, 42 },
		{ 43, 359, 42 },
		{ 43, 360, 42 },
		{ 43, 361, 42 },
		{ 43, 362, 42 },
		{ 43, 363, 42 },
		{ 43, 364, 42 },
		{ 43, 365, 42 },
		{ 43, 366, 42 },
		{ 43, 367, 42 },
		{ 43, 368, 42 },
		{ 43, 369, 42 },
		{ 43, 370, 42 },
		{ 43, 371, 42 },
		{ 43, 372, 42 },
		{ 43, 373, 42 },
		{ 43, 374, 42 },
		{ 43, 375, 42 },
		{ 43, 376, 42 },
		{ 43, 377, 42 },
		{ 43, 378, 42 },
		{ 43, 379, 42 },
		{ 43, 380, 42 },
		{ 43, 381, 42 },
		{ 43, 382, 42 },
		{ 43, 383, 42 },
		{ 43, 384, 42 },
		{ 43, 385, 42 },
		{ 43, 386, 42 },
		{ 43, 387, 42 },
		{ 43, 388, 42 },
		{ 43, 389, 42 },
		{ 43, 390, 42 },
		{ 44, 337, 42 },
		{ 44, 338, 42 },
		{ 44, 339, 42 },
		{ 44, 340, 42 },
		{ 44, 341, 42 },
		{ 44, 342, 42 },
		{ 44, 343, 42 },
		{ 44, 344, 42 },
		{ 44, 345, 42 },
		{ 44, 346, 42 },
		{ 44, 347, 42 },
		{ 44, 348, 42 },
		{ 44, 349, 42 },
		{ 44, 350, 42 },
		{ 44, 351, 42 },
		{ 44, 352, 42 },
		{ 44, 353, 42 },
		{ 44, 354, 42 },
		{ 44, 355, 42 },
		{ 44, 356, 42 },
		{ 44, 357, 42 },
		{ 44, 358, 42 },
		{ 44, 359, 42 },
		{ 44, 360, 42 },
		{ 44, 361, 42 },
		{ 44, 362, 42 },
		{ 44, 363, 42 },
		{ 44, 364, 42 },
		{ 44, 365, 42 },
		{ 44, 366, 42 },
		{ 44, 367, 42 },
		{ 44, 368, 42 },
		{ 44, 369, 42 },
		{ 44, 370, 42 },
		{ 44, 371, 42 },
		{ 44, 372, 42 },
		{ 44, 373, 42 },
		{ 44, 374, 42 },
		{ 44, 375, 42 },
		{ 44, 376, 42 },
		{ 44, 377, 42 },
		{ 44, 378, 42 },
		{ 44, 379, 42 },
		{ 44, 380, 42 },
		{ 44, 381, 42 },
		{ 44, 382, 42 },
		{ 44, 383, 42 },
		{ 44, 384, 42 },
		{ 44, 385, 42 },
		{ 44, 386, 42 },
		{ 44, 387, 42 },
		{ 44, 388, 42 },
		{ 44, 389, 42 },
		{ 44, 390, 42 },
		{ 45, 337, 42 },
		{ 45, 338, 42 },
		{ 45, 339, 42 },
		{ 45, 340, 42 },
		{ 45, 341, 42 },
		{ 45, 342, 42 },
		{ 45, 343, 42 },
		{ 45, 344, 42 },
		{ 45, 345, 42 },
		{ 45, 346, 42 },
		{ 45, 347, 42 },
		{ 45, 348, 42 },
		{ 45, 349, 42 },
		{ 45, 350, 42 },
		{ 45, 351, 42 },
		{ 45, 352, 42 },
		{ 45, 353, 42 },
		{ 45, 354, 42 },
		{ 45, 355, 42 },
		{ 45, 356, 42 },
		{ 45, 357, 42 },
		{ 45, 358, 42 },
		{ 45, 359, 42 },
		{ 45, 360, 42 },
		{ 45, 361, 42 },
		{ 45, 362, 42 },
		{ 45, 363, 42 },
		{ 45, 364, 42 },
		{ 45, 365, 42 },
		{ 45, 366, 42 },
		{ 45, 367, 42 },
		{ 45, 368, 42 },
		{ 45, 369, 42 },
		{ 45, 370, 42 },
		{ 45, 371, 42 },
		{ 45, 372, 42 },
		{ 45, 373, 42 },
		{ 45, 374, 42 },
		{ 45, 375, 42 },
		{ 45, 376, 42 },
		{ 45, 377, 42 },
		{ 45, 378, 42 },
		{ 45, 379, 42 },
		{ 45, 380, 42 },
		{ 45, 381, 42 },
		{ 45, 382, 42 },
		{ 45, 383, 42 },
		{ 45, 384, 42 },
		{ 45, 385, 42 },
		{ 45, 386, 42 },
		{ 45, 387, 42 },
		{ 45, 388, 42 },
		{ 45, 389, 42 },
		{ 46, 337, 42 },
		{ 46, 338, 42 },
		{ 46, 339, 42 },
		{ 46, 340, 42 },
		{ 46, 341, 42 },
		{ 46, 342, 42 },
		{ 46, 343, 42 },
		{ 46, 344, 42 },
		{ 46, 345, 42 },
		{ 46, 346, 42 },
		{ 46, 347, 42 },
		{ 46, 348, 42 },
		{ 46, 349, 42 },
		{ 46, 350, 42 },
		{ 46, 351, 42 },
		{ 46, 352, 42 },
		{ 46, 353, 42 },
		{ 46, 354, 42 },
		{ 46, 355, 42 },
		{ 46, 356, 42 },
		{ 46, 357, 42 },
		{ 46, 358, 42 },
		{ 46, 359, 42 },
		{ 46, 360, 42 },
		{ 46, 361, 42 },
		{ 46, 362, 42 },
		{ 46, 363, 42 },
		{ 46, 364, 42 },
		{ 46, 365, 42 },
		{ 46, 366, 42 },
		{ 46, 367, 42 },
		{ 46, 368, 42 },
		{ 46, 369, 42 },
		{ 46, 370, 42 },
		{ 46, 371, 42 },
		{ 46, 372, 42 },
		{ 46, 373, 42 },
		{ 46, 374, 42 },
		{ 46, 375, 42 },
		{ 46, 376, 42 },
		{ 46, 377, 42 },
		{ 46, 378, 42 },
		{ 46, 379, 42 },
		{ 46, 380, 42 },
		{ 46, 381, 42 },
		{ 46, 382, 42 },
		{ 46, 383, 42 },
		{ 46, 384, 42 },
		{ 46, 385, 42 },
		{ 46, 386, 42 },
		{ 46, 387, 42 },
		{ 46, 388, 42 },
		{ 47, 336, 42 },
		{ 47, 337, 42 },
		{ 47, 338, 42 },
		{ 47, 339, 42 },
		{ 47, 340, 42 },
		{ 47, 341, 42 },
		{ 47, 342, 42 },
		{ 47, 343, 42 },
		{ 47, 344, 42 },
		{ 47, 345, 42 },
		{ 47, 346, 42 },
		{ 47, 347, 42 },
		{ 47, 348, 42 },
		{ 47, 349, 42 },
		{ 47, 350, 42 },
		{ 47, 351, 42 },
		{ 47, 352, 42 },
		{ 47, 353, 42 },
		{ 47, 354, 42 },
		{ 47, 355, 42 },
		{ 47, 356, 42 },
		{ 47, 357, 42 },
		{ 47, 358, 42 },
		{ 47, 359, 42 },
		{ 47, 360, 42 },
		{ 47, 361, 42 },
		{ 47, 362, 42 },
		{ 47, 363, 42 },
		{ 47, 364, 42 },
		{ 47, 365, 42 },
		{ 47, 366, 42 },
		{ 47, 367, 42 },
		{ 47, 368, 42 },
		{ 47, 369, 42 },
		{ 47, 370, 42 },
		{ 47, 371, 42 },
		{ 47, 372, 42 },
		{ 47, 373, 42 },
		{ 47, 374, 42 },
		{ 47, 375, 42 },
		{ 47, 376, 42 },
		{ 47, 377, 42 },
		{ 47, 378, 42 },
		{ 47, 379, 42 },
		{ 47, 380, 42 },
		{ 47, 381, 42 },
		{ 47, 382, 42 },
		{ 47, 383, 42 },
		{ 47, 384, 42 },
		{ 47, 385, 42 },
		{ 47, 386, 42 },
		{ 47, 387, 42 },
		{ 47, 388, 42 },
		{ 48, 336, 42 },
		{ 48, 337, 42 },
		{ 48, 338, 42 },
		{ 48, 339, 42 },
		{ 48, 340, 42 },
		{ 48, 341, 42 },
		{ 48, 342, 42 },
		{ 48, 343, 42 },
		{ 48, 344, 42 },
		{ 48, 345, 42 },
		{ 48, 346, 42 },
		{ 48, 347, 42 },
		{ 48, 348, 42 },
		{ 48, 349, 42 },
		{ 48, 350, 42 },
		{ 48, 351, 42 },
		{ 48, 352, 42 },
		{ 48, 353, 42 },
		{ 48, 354, 42 },
		{ 48, 355, 42 },
		{ 48, 356, 42 },
		{ 48, 357, 42 },
		{ 48, 358, 42 },
		{ 48, 359, 42 },
		{ 48, 360, 42 },
		{ 48, 361, 42 },
		{ 48, 362, 42 },
		{ 48, 363, 42 },
		{ 48, 364, 42 },
		{ 48, 365, 42 },
		{ 48, 366, 42 },
		{ 48, 367, 42 },
		{ 48, 368, 42 },
		{ 48, 369, 42 },
		{ 48, 370, 42 },
		{ 48, 371, 42 },
		{ 48, 372, 42 },
		{ 48, 373, 42 },
		{ 48, 374, 42 },
		{ 48, 375, 42 },
		{ 48, 376, 42 },
		{ 48, 377, 42 },
		{ 48, 378, 42 },
		{ 48, 379, 42 },
		{ 48, 380, 42 },
		{ 48, 381, 42 },
		{ 48, 382, 42 },
		{ 48, 383, 42 },
		{ 48, 384, 42 },
		{ 48, 385, 42 },
		{ 48, 386, 42 },
		{ 48, 387, 42 },
		{ 49, 336, 42 },
		{ 49, 337, 42 },
		{ 49, 338, 42 },
		{ 49, 339, 42 },
		{ 49, 340, 42 },
		{ 49, 341, 42 },
		{ 49, 342, 42 },
		{ 49, 343, 42 },
		{ 49, 344, 42 },
		{ 49, 345, 42 },
		{ 49, 346, 42 },
		{ 49, 347, 42 },
		{ 49, 348, 42 },
		{ 49, 349, 42 },
		{ 49, 350, 42 },
		{ 49, 351, 42 },
		{ 49, 352, 42 },
		{ 49, 353, 42 },
		{ 49, 354, 42 },
		{ 49, 355, 42 },
		{ 49, 356, 42 },
		{ 49, 357, 42 },
		{ 49, 358, 42 },
		{ 49, 359, 42 },
		{ 49, 360, 42 },
		{ 49, 361, 42 },
		{ 49, 362, 42 },
		{ 49, 363, 42 },
		{ 49, 364, 42 },
		{ 49, 365, 42 },
		{ 49, 366, 42 },
		{ 49, 367, 42 },
		{ 49, 368, 42 },
		{ 49, 369, 42 },
		{ 49, 370, 42 },
		{ 49, 371, 42 },
		{ 49, 372, 42 },
		{ 49, 373, 42 },
		{ 49, 374, 42 },
		{ 49, 375, 42 },
		{ 49, 376, 42 },
		{ 49, 377, 42 },
		{ 49, 378, 42 },
		{ 49, 379, 42 },
		{ 49, 380, 42 },
		{ 49, 381, 42 },
		{ 49, 382, 42 },
		{ 49, 383, 42 },
		{ 49, 384, 42 },
		{ 49, 385, 42 },
		{ 49, 386, 42 },
		{ 50, 335, 42 },
		{ 50, 336, 42 },
		{ 50, 337, 42 },
		{ 50, 338, 42 },
		{ 50, 339, 42 },
		{ 50, 340, 42 },
		{ 50, 341, 42 },
		{ 50, 342, 42 },
		{ 50, 343, 42 },
		{ 50, 344, 42 },
		{ 50, 345, 42 },
		{ 50, 346, 42 },
		{ 50, 347, 42 },
		{ 50, 348, 42 },
		{ 50, 349, 42 },
		{ 50, 350, 42 },
		{ 50, 351, 42 },
		{ 50, 352, 42 },
		{ 50, 353, 42 },
		{ 50, 354, 42 },
		{ 50, 355, 42 },
		{ 50, 356, 42 },
		{ 50, 357, 42 },
		{ 50, 358, 42 },
		{ 50, 359, 42 },
		{ 50, 360, 42 },
		{ 50, 361, 42 },
		{ 50, 362, 42 },
		{ 50, 363, 42 },
		{ 50, 364, 42 },
		{ 50, 365, 42 },
		{ 50, 366, 42 },
		{ 50, 367, 42 },
		{ 50, 368, 42 },
		{ 50, 369, 42 },
		{ 50, 370, 42 },
		{ 50, 371, 42 },
		{ 50, 372, 42 },
		{ 50, 373, 42 },
		{ 50, 374, 42 },
		{ 50, 375, 42 },
		{ 50, 376, 42 },
		{ 50, 377, 42 },
		{ 50, 378, 42 },
		{ 50, 379, 42 },
		{ 50, 380, 42 },
		{ 50, 381, 42 },
		{ 50, 382, 42 },
		{ 50, 383, 42 },
		{ 50, 384, 42 },
		{ 50, 385, 42 },
		{ 50, 386, 42 },
		{ 51, 335, 42 },
		{ 51, 336, 42 },
		{ 51, 337, 42 },
		{ 51, 338, 42 },
		{ 51, 339, 42 },
		{ 51, 340, 42 },
		{ 51, 341, 42 },
		{ 51, 342, 42 },
		{ 51, 343, 42 },
		{ 51, 344, 42 },
		{ 51, 345, 42 },
		{ 51, 346, 42 },
		{ 51, 347, 42 },
		{ 51, 348, 42 },
		{ 51, 349, 42 },
		{ 51, 350, 42 },
		{ 51, 351, 42 },
		{ 51, 352, 42 },
		{ 51, 353, 42 },
		{ 51, 354, 42 },
		{ 51, 355, 42 },
		{ 51, 356, 42 },
		{ 51, 357, 42 },
		{ 51, 358, 42 },
		{ 51, 359, 42 },
		{ 51, 360, 42 },
		{ 51, 361, 42 },
		{ 51, 362, 42 },
		{ 51, 363, 42 },
		{ 51, 364, 42 },
		{ 51, 365, 42 },
		{ 51, 366, 42 },
		{ 51, 367, 42 },
		{ 51, 368, 42 },
		{ 51, 369, 42 },
		{ 51, 370, 42 },
		{ 51, 371, 42 },
		{ 51, 372, 42 },
		{ 51, 373, 42 },
		{ 51, 374, 42 },
		{ 51, 375, 42 },
		{ 51, 376, 42 },
		{ 51, 377, 42 },
		{ 51, 378, 42 },
		{ 51, 379, 42 },
		{ 51, 380, 42 },
		{ 51, 381, 42 },
		{ 51, 382, 42 },
		{ 51, 383, 42 },
		{ 51, 384, 42 },
		{ 51, 385, 42 },
		{ 52, 334, 42 },
		{ 52, 335, 42 },
		{ 52, 336, 42 },
		{ 52, 337, 42 },
		{ 52, 338, 42 },
		{ 52, 339, 42 },
		{ 52, 340, 42 },
		{ 52, 341, 42 },
		{ 52, 342, 42 },
		{ 52, 343, 42 },
		{ 52, 344, 42 },
		{ 52, 345, 42 },
		{ 52, 346, 42 },
		{ 52, 347, 42 },
		{ 52, 348, 42 },
		{ 52, 349, 42 },
		{ 52, 350, 42 },
		{ 52, 351, 42 },
		{ 52, 352, 42 },
		{ 52, 353, 42 },
		{ 52, 354, 42 },
		{ 52, 355, 42 },
		{ 52, 356, 42 },
		{ 52, 357, 42 },
		{ 52, 358, 42 },
		{ 52, 359, 42 },
		{ 52, 360, 42 },
		{ 52, 361, 42 },
		{ 52, 362, 42 },
		{ 52, 363, 42 },
		{ 52, 364, 42 },
		{ 52, 365, 42 },
		{ 52, 366, 42 },
		{ 52, 367, 42 },
		{ 52, 368, 42 },
		{ 52, 369, 42 },
		{ 52, 370, 42 },
		{ 52, 371, 42 },
		{ 52, 372, 42 },
		{ 52, 373, 42 },
		{ 52, 374, 42 },
		{ 52, 375, 42 },
		{ 52, 376, 42 },
		{ 52, 377, 42 },
		{ 52, 378, 42 },
		{ 52, 379, 42 },
		{ 52, 380, 42 },
		{ 52, 381, 42 },
		{ 52, 382, 42 },
		{ 52, 383, 42 },
		{ 52, 384, 42 },
		{ 52, 385, 42 },
		{ 53, 334, 42 },
		{ 53, 335, 42 },
		{ 53, 336, 42 },
		{ 53, 337, 42 },
		{ 53, 338, 42 },
		{ 53, 339, 42 },
		{ 53, 340, 42 },
		{ 53, 341, 42 },
		{ 53, 342, 42 },
		{ 53, 343, 42 },
		{ 53, 344, 42 },
		{ 53, 345, 42 },
		{ 53, 346, 42 },
		{ 53, 347, 42 },
		{ 53, 348, 42 },
		{ 53, 349, 42 },
		{ 53, 350, 42 },
		{ 53, 351, 42 },
		{ 53, 352, 42 },
		{ 53, 353, 42 },
		{ 53, 354, 42 },
		{ 53, 355, 42 },
		{ 53, 356, 42 },
		{ 53, 357, 42 },
		{ 53, 358, 42 },
		{ 53, 359, 42 },
		{ 53, 360, 42 },
		{ 53, 361, 42 },
		{ 53, 362, 42 },
		{ 53, 363, 42 },
		{ 53, 364, 42 },
		{ 53, 365, 42 },
		{ 53, 366, 42 },
		{ 53, 367, 42 },
		{ 53, 368, 42 },
		{ 53, 369, 42 },
		{ 53, 370, 42 },
		{ 53, 371, 42 },
		{ 53, 372, 42 },
		{ 53, 373, 42 },
		{ 53, 374, 42 },
		{ 53, 375, 42 },
		{ 53, 376, 42 },
		{ 53, 377, 42 },
		{ 53, 378, 42 },
		{ 53, 379, 42 },
		{ 53, 380, 42 },
		{ 53, 381, 42 },
		{ 53, 382, 42 },
		{ 53, 383, 42 },
		{ 53, 384, 42 },
		{ 54, 334, 42 },
		{ 54, 335, 42 },
		{ 54, 336, 42 },
		{ 54, 337, 42 },
		{ 54, 338, 42 },
		{ 54, 339, 42 },
		{ 54, 340, 42 },
		{ 54, 341, 42 },
		{ 54, 342, 42 },
		{ 54, 343, 42 },
		{ 54, 344, 42 },
		{ 54, 345, 42 },
		{ 54, 346, 42 },
		{ 54, 347, 42 },
		{ 54, 348, 42 },
		{ 54, 349, 42 },
		{ 54, 350, 42 },
		{ 54, 351, 42 },
		{ 54, 352, 42 },
		{ 54, 353, 42 },
		{ 54, 354, 42 },
		{ 54, 355, 42 },
		{ 54, 356, 42 },
		{ 54, 357, 42 },
		{ 54, 358, 42 },
		{ 54, 359, 42 },
		{ 54, 360, 42 },
		{ 54, 361, 42 },
		{ 54, 362, 42 },
		{ 54, 363, 42 },
		{ 54, 364, 42 },
		{ 54, 365, 42 },
		{ 54, 366, 42 },
		{ 54, 367, 42 },
		{ 54, 368, 42 },
		{ 54, 369, 42 },
		{ 54, 370, 42 },
		{ 54, 371, 42 },
		{ 54, 372, 42 },
		{ 54, 373, 42 },
		{ 54, 374, 42 },
		{ 54, 375, 42 },
		{ 54, 376, 42 },
		{ 54, 377, 42 },
		{ 54, 378, 42 },
		{ 54, 379, 42 },
		{ 54, 380, 42 },
		{ 54, 381, 42 },
		{ 54, 382, 42 },
		{ 54, 383, 42 },
		{ 55, 333, 42 },
		{ 55, 334, 42 },
		{ 55, 335, 42 },
		{ 55, 336, 42 },
		{ 55, 337, 42 },
		{ 55, 338, 42 },
		{ 55, 339, 42 },
		{ 55, 340, 42 },
		{ 55, 341, 42 },
		{ 55, 342, 42 },
		{ 55, 343, 42 },
		{ 55, 344, 42 },
		{ 55, 345, 42 },
		{ 55, 346, 42 },
		{ 55, 347, 42 },
		{ 55, 348, 42 },
		{ 55, 349, 42 },
		{ 55, 350, 42 },
		{ 55, 351, 42 },
		{ 55, 352, 42 },
		{ 55, 353, 42 },
		{ 55, 354, 42 },
		{ 55, 355, 42 },
		{ 55, 356, 42 },
		{ 55, 357, 42 },
		{ 55, 358, 42 },
		{ 55, 359, 42 },
		{ 55, 360, 42 },
		{ 55, 361, 42 },
		{ 55, 362, 42 },
		{ 55, 363, 42 },
		{ 55, 364, 42 },
		{ 55, 365, 42 },
		{ 55, 366, 42 },
		{ 55, 367, 42 },
		{ 55, 368, 42 },
		{ 55, 369, 42 },
		{ 55, 370, 42 },
		{ 55, 371, 42 },
		{ 55, 372, 42 },
		{ 55, 373, 42 },
		{ 55, 374, 42 },
		{ 55, 375, 42 },
		{ 55, 376, 42 },
		{ 55, 377, 42 },
		{ 55, 378, 42 },
		{ 55, 379, 42 },
		{ 55, 380, 42 },
		{ 55, 381, 42 },
		{ 55, 382, 42 },
		{ 55, 383, 42 },
		{ 56, 333, 42 },
		{ 56, 334, 42 },
		{ 56, 335, 42 },
		{ 56, 336, 42 },
		{ 56, 337, 42 },
		{ 56, 338, 42 },
		{ 56, 339, 42 },
		{ 56, 340, 42 },
		{ 56, 341, 42 },
		{ 56, 342, 42 },
		{ 56, 343, 42 },
		{ 56, 344, 42 },
		{ 56, 345, 42 },
		{ 56, 346, 42 },
		{ 56, 347, 42 },
		{ 56, 348, 42 },
		{ 56, 349, 42 },
		{ 56, 350, 42 },
		{ 56, 351, 42 },
		{ 56, 352, 42 },
		{ 56, 353, 42 },
		{ 56, 354, 42 },
		{ 56, 355, 42 },
		{ 56, 356, 42 },
		{ 56, 357, 42 },
		{ 56, 358, 42 },
		{ 56, 359, 42 },
		{ 56, 360, 42 },
		{ 56, 361, 42 },
		{ 56, 362, 42 },
		{ 56, 363, 42 },
		{ 56, 364, 42 },
		{ 56, 365, 42 },
		{ 56, 366, 42 },
		{ 56, 367, 42 },
		{ 56, 368, 42 },
		{ 56, 369, 42 },
		{ 56, 370, 42 },
		{ 56, 371, 42 },
		{ 56, 372, 42 },
		{ 56, 373, 42 },
		{ 56, 374, 42 },
		{ 56, 375, 42 },
		{ 56, 376, 42 },
		{ 56, 377, 42 },
		{ 56, 378, 42 },
		{ 56, 379, 42 },
		{ 56, 380, 42 },
		{ 56, 381, 42 },
		{ 56, 382, 42 },
		{ 57, 333, 42 },
		{ 57, 334, 42 },
		{ 57, 335, 42 },
		{ 57, 336, 42 },
		{ 57, 337, 42 },
		{ 57, 338, 42 },
		{ 57, 339, 42 },
		{ 57, 340, 42 },
		{ 57, 341, 42 },
		{ 57, 342, 42 },
		{ 57, 343, 42 },
		{ 57, 344, 42 },
		{ 57, 345, 42 },
		{ 57, 346, 42 },
		{ 57, 347, 42 },
		{ 57, 348, 42 },
		{ 57, 349, 42 },
		{ 57, 350, 42 },
		{ 57, 351, 42 },
		{ 57, 352, 42 },
		{ 57, 353, 42 },
		{ 57, 354, 42 },
		{ 57, 355, 42 },
		{ 57, 356, 42 },
		{ 57, 357, 42 },
		{ 57, 358, 42 },
		{ 57, 359, 42 },
		{ 57, 360, 42 },
		{ 57, 361, 42 },
		{ 57, 362, 42 },
		{ 57, 363, 42 },
		{ 57, 364, 42 },
		{ 57, 365, 42 },
		{ 57, 366, 42 },
		{ 57, 367, 42 },
		{ 57, 368, 42 },
		{ 57, 369, 42 },
		{ 57, 370, 42 },
		{ 57, 371, 42 },
		{ 57, 372, 42 },
		{ 57, 373, 42 },
		{ 57, 374, 42 },
		{ 57, 375, 42 },
		{ 57, 376, 42 },
		{ 57, 377, 42 },
		{ 57, 378, 42 },
		{ 57, 379, 42 },
		{ 57, 380, 42 },
		{ 57, 381, 42 },
		{ 58, 332, 42 },
		{ 58, 333, 42 },
		{ 58, 334, 42 },
		{ 58, 335, 42 },
		{ 58, 336, 42 },
		{ 58, 337, 42 },
		{ 58, 338, 42 },
		{ 58, 339, 42 },
		{ 58, 340, 42 },
		{ 58, 341, 42 },
		{ 58, 342, 42 },
		{ 58, 343, 42 },
		{ 58, 344, 42 },
		{ 58, 345, 42 },
		{ 58, 346, 42 },
		{ 58, 347, 42 },
		{ 58, 348, 42 },
		{ 58, 349, 42 },
		{ 58, 350, 42 },
		{ 58, 351, 42 },
		{ 58, 352, 42 },
		{ 58, 353, 42 },
		{ 58, 354, 42 },
		{ 58, 355, 42 },
		{ 58, 356, 42 },
		{ 58, 357, 42 },
		{ 58, 358, 42 },
		{ 58, 359, 42 },
		{ 58, 360, 42 },
		{ 58, 361, 42 },
		{ 58, 362, 42 },
		{ 58, 363, 42 },
		{ 58, 364, 42 },
		{ 58, 365, 42 },
		{ 58, 366, 42 },
		{ 58, 367, 42 },
		{ 58, 368, 42 },
		{ 58, 369, 42 },
		{ 58, 370, 42 },
		{ 58, 371, 42 },
		{ 58, 372, 42 },
		{ 58, 373, 42 },
		{ 58, 374, 42 },
		{ 58, 375, 42 },
		{ 58, 376, 42 },
		{ 58, 377, 42 },
		{ 58, 378, 42 },
		{ 58, 379, 42 },
		{ 58, 380, 42 },
		{ 58, 381, 42 },
		{ 59, 332, 42 },
		{ 59, 333, 42 },
		{ 59, 334, 42 },
		{ 59, 335, 42 },
		{ 59, 336, 42 },
		{ 59, 337, 42 },
		{ 59, 338, 42 },
		{ 59, 339, 42 },
		{ 59, 340, 42 },
		{ 59, 341, 42 },
		{ 59, 342, 42 },
		{ 59, 343, 42 },
		{ 59, 344, 42 },
		{ 59, 345, 42 },
		{ 59, 346, 42 },
		{ 59, 347, 42 },
		{ 59, 348, 42 },
		{ 59, 349, 42 },
		{ 59, 350, 42 },
		{ 59, 351, 42 },
		{ 59, 352, 42 },
		{ 59, 353, 42 },
		{ 59, 354, 42 },
		{ 59, 355, 42 },
		{ 59, 356, 42 },
		{ 59, 357, 42 },
		{ 59, 358, 42 },
		{ 59, 359, 42 },
		{ 59, 360, 42 },
		{ 59, 361, 42 },
		{ 59, 362, 42 },
		{ 59, 363, 42 },
		{ 59, 364, 42 },
		{ 59, 365, 42 },
		{ 59, 366, 42 },
		{ 59, 367, 42 },
		{ 59, 368, 42 },
		{ 59, 369, 42 },
		{ 59, 370, 42 },
		{ 59, 371, 42 },
		{ 59, 372, 42 },
		{ 59, 373, 42 },
		{ 59, 374, 42 },
		{ 59, 375, 42 },
		{ 59, 376, 42 },
		{ 59, 377, 42 },
		{ 59, 378, 42 },
		{ 59, 379, 42 },
		{ 59, 380, 42 },
		{ 60, 331, 42 },
		{ 60, 332, 42 },
		{ 60, 333, 42 },
		{ 60, 334, 42 },
		{ 60, 335, 42 },
		{ 60, 336, 42 },
		{ 60, 337, 42 },
		{ 60, 338, 42 },
		{ 60, 339, 42 },
		{ 60, 340, 42 },
		{ 60, 341, 42 },
		{ 60, 342, 42 },
		{ 60, 343, 42 },
		{ 60, 344, 42 },
		{ 60, 345, 42 },
		{ 60, 346, 42 },
		{ 60, 347, 42 },
		{ 60, 348, 42 },
		{ 60, 349, 42 },
		{ 60, 350, 42 },
		{ 60, 351, 42 },
		{ 60, 352, 42 },
		{ 60, 353, 42 },
		{ 60, 354, 42 },
		{ 60, 355, 42 },
		{ 60, 356, 42 },
		{ 60, 357, 42 },
		{ 60, 358, 42 },
		{ 60, 359, 42 },
		{ 60, 360, 42 },
		{ 60, 361, 42 },
		{ 60, 362, 42 },
		{ 60, 363, 42 },
		{ 60, 364, 42 },
		{ 60, 365, 42 },
		{ 60, 366, 42 },
		{ 60, 367, 42 },
		{ 60, 368, 42 },
		{ 60, 369, 42 },
		{ 60, 370, 42 },
		{ 60, 371, 42 },
		{ 60, 372, 42 },
		{ 60, 373, 42 },
		{ 60, 374, 42 },
		{ 60, 375, 42 },
		{ 60, 376, 42 },
		{ 60, 377, 42 },
		{ 60, 378, 42 },
		{ 60, 379, 42 },
		{ 60, 380, 42 },
		{ 61, 331, 42 },
		{ 61, 332, 42 },
		{ 61, 333, 42 },
		{ 61, 334, 42 },
		{ 61, 335, 42 },
		{ 61, 336, 42 },
		{ 61, 337, 42 },
		{ 61, 338, 42 },
		{ 61, 339, 42 },
		{ 61, 340, 42 },
		{ 61, 341, 42 },
		{ 61, 342, 42 },
		{ 61, 343, 42 },
		{ 61, 344, 42 },
		{ 61, 345, 42 },
		{ 61, 346, 42 },
		{ 61, 347, 42 },
		{ 61, 348, 42 },
		{ 61, 349, 42 },
		{ 61, 350, 42 },
		{ 61, 351, 42 },
		{ 61, 352, 42 },
		{ 61, 353, 42 },
		{ 61, 354, 42 },
		{ 61, 355, 42 },
		{ 61, 356, 42 },
		{ 61, 357, 42 },
		{ 61, 358, 42 },
		{ 61, 359, 42 },
		{ 61, 360, 42 },
		{ 61, 361, 42 },
		{ 61, 362, 42 },
		{ 61, 363, 42 },
		{ 61, 364, 42 },
		{ 61, 365, 42 },
		{ 61, 366, 42 },
		{ 61, 367, 42 },
		{ 61, 368, 42 },
		{ 61, 369, 42 },
		{ 61, 370, 42 },
		{ 61, 371, 42 },
		{ 61, 372, 42 },
		{ 61, 373, 42 },
		{ 61, 374, 42 },
		{ 61, 375, 42 },
		{ 61, 376, 42 },
		{ 61, 377, 42 },
		{ 61, 378, 42 },
		{ 61, 379, 42 },
		{ 62, 331, 42 },
		{ 62, 332, 42 },
		{ 62, 333, 42 },
		{ 62, 334, 42 },
		{ 62, 335, 42 },
		{ 62, 336, 42 },
		{ 62, 337, 42 },
		{ 62, 338, 42 },
		{ 62, 339, 42 },
		{ 62, 340, 42 },
		{ 62, 341, 42 },
		{ 62, 342, 42 },
		{ 62, 343, 42 },
		{ 62, 344, 42 },
		{ 62, 345, 42 },
		{ 62, 346, 42 },
		{ 62, 347, 42 },
		{ 62, 348, 42 },
		{ 62, 349, 42 },
		{ 62, 350, 42 },
		{ 62, 351, 42 },
		{ 62, 352, 42 },
		{ 62, 353, 42 },
		{ 62, 354, 42 },
		{ 62, 355, 42 },
		{ 62, 356, 42 },
		{ 62, 357, 42 },
		{ 62, 358, 42 },
		{ 62, 359, 42 },
		{ 62, 360, 42 },
		{ 62, 361, 42 },
		{ 62, 362, 42 },
		{ 62, 363, 42 },
		{ 62, 364, 42 },
		{ 62, 365, 42 },
		{ 62, 366, 42 },
		{ 62, 367, 42 },
		{ 62, 368, 42 },
		{ 62, 369, 42 },
		{ 62, 370, 42 },
		{ 62, 371, 42 },
		{ 62, 372, 42 },
		{ 62, 373, 42 },
		{ 62, 374, 42 },
		{ 62, 375, 42 },
		{ 62, 376, 42 },
		{ 62, 377, 42 },
		{ 62, 378, 42 },
		{ 63, 330, 42 },
		{ 63, 331, 42 },
		{ 63, 332, 42 },
		{ 63, 333, 42 },
		{ 63, 334, 42 },
		{ 63, 335, 42 },
		{ 63, 336, 42 },
		{ 63, 337, 42 },
		{ 63, 338, 42 },
		{ 63, 339, 42 },
		{ 63, 340, 42 },
		{ 63, 341, 42 },
		{ 63, 342, 42 },
		{ 63, 343, 42 },
		{ 63, 344, 42 },
		{ 63, 345, 42 },
		{ 63, 346, 42 },
		{ 63, 347, 42 },
		{ 63, 348, 42 },
		{ 63, 349, 42 },
		{ 63, 350, 42 },
		{ 63, 351, 42 },
		{ 63, 352, 42 },
		{ 63, 353, 42 },
		{ 63, 354, 42 },
		{ 63, 355, 42 },
		{ 63, 356, 42 },
		{ 63, 357, 42 },
		{ 63, 358, 42 },
		{ 63, 359, 42 },
		{ 63, 360, 42 },
		{ 63, 361, 42 },
		{ 63, 362, 42 },
		{ 63, 363, 42 },
		{ 63, 364, 42 },
		{ 63, 365, 42 },
		{ 63, 366, 42 },
		{ 63, 367, 42 },
		{ 63, 368, 42 },
		{ 63, 369, 42 },
		{ 63, 370, 42 },
		{ 63, 371, 42 },
		{ 63, 372, 42 },
		{ 63, 373, 42 },
		{ 63, 374, 42 },
		{ 63, 375, 42 },
		{ 63, 376, 42 },
		{ 63, 377, 42 },
		{ 63, 378, 42 },
		{ 64, 330, 42 },
		{ 64, 331, 42 },
		{ 64, 332, 42 },
		{ 64, 333, 42 },
		{ 64, 334, 42 },
		{ 64, 335, 42 },
		{ 64, 336, 42 },
		{ 64, 337, 42 },
		{ 64, 338, 42 },
		{ 64, 339, 42 },
		{ 64, 340, 42 },
		{ 64, 341, 42 },
		{ 64, 342, 42 },
		{ 64, 343, 42 },
		{ 64, 344, 42 },
		{ 64, 345, 42 },
		{ 64, 346, 42 },
		{ 64, 347, 42 },
		{ 64, 348, 42 },
		{ 64, 349, 42 },
		{ 64, 350, 42 },
		{ 64, 351, 42 },
		{ 64, 352, 42 },
		{ 64, 353, 42 },
		{ 64, 354, 42 },
		{ 64, 355, 42 },
		{ 64, 356, 42 },
		{ 64, 357, 42 },
		{ 64, 358, 42 },
		{ 64, 359, 42 },
		{ 64, 360, 42 },
		{ 64, 361, 42 },
		{ 64, 362, 42 },
		{ 64, 363, 42 },
		{ 64, 364, 42 },
		{ 64, 365, 42 },
		{ 64, 366, 42 },
		{ 64, 367, 42 },
		{ 64, 368, 42 },
		{ 64, 369, 42 },
		{ 64, 370, 42 },
		{ 64, 371, 42 },
		{ 64, 372, 42 },
		{ 64, 373, 42 },
		{ 64, 374, 42 },
		{ 64, 375, 42 },
		{ 64, 376, 42 },
		{ 64, 377, 42 },
		{ 65, 330, 42 },
		{ 65, 331, 42 },
		{ 65, 332, 42 },
		{ 65, 333, 42 },
		{ 65, 334, 42 },
		{ 65, 335, 42 },
		{ 65, 336, 42 },
		{ 65, 337, 42 },
		{ 65, 338, 42 },
		{ 65, 339, 42 },
		{ 65, 340, 42 },
		{ 65, 341, 42 },
		{ 65, 342, 42 },
		{ 65, 343, 42 },
		{ 65, 344, 42 },
		{ 65, 345, 42 },
		{ 65, 346, 42 },
		{ 65, 347, 42 },
		{ 65, 348, 42 },
		{ 65, 349, 42 },
		{ 65, 350, 42 },
		{ 65, 351, 42 },
		{ 65, 352, 42 },
		{ 65, 353, 42 },
		{ 65, 354, 42 },
		{ 65, 355, 42 },
		{ 65, 356, 42 },
		{ 65, 357, 42 },
		{ 65, 358, 42 },
		{ 65, 359, 42 },
		{ 65, 360, 42 },
		{ 65, 361, 42 },
		{ 65, 362, 42 },
		{ 65, 363, 42 },
		{ 65, 364, 42 },
		{ 65, 365, 42 },
		{ 65, 366, 42 },
		{ 65, 367, 42 },
		{ 65, 368, 42 },
		{ 65, 369, 42 },
		{ 65, 370, 42 },
		{ 65, 371, 42 },
		{ 65, 372, 42 },
		{ 65, 373, 42 },
		{ 65, 374, 42 },
		{ 65, 375, 42 },
		{ 65, 376, 42 },
		{ 66, 329, 42 },
		{ 66, 330, 42 },
		{ 66, 331, 42 },
		{ 66, 332, 42 },
		{ 66, 333, 42 },
		{ 66, 334, 42 },
		{ 66, 335, 42 },
		{ 66, 336, 42 },
		{ 66, 337, 42 },
		{ 66, 338, 42 },
		{ 66, 339, 42 },
		{ 66, 340, 42 },
		{ 66, 341, 42 },
		{ 66, 342, 42 },
		{ 66, 343, 42 },
		{ 66, 344, 42 },
		{ 66, 345, 42 },
		{ 66, 346, 42 },
		{ 66, 347, 42 },
		{ 66, 348, 42 },
		{ 66, 349, 42 },
		{ 66, 350, 42 },
		{ 66, 351, 42 },
		{ 66, 352, 42 },
		{ 66, 353, 42 },
		{ 66, 354, 42 },
		{ 66, 355, 42 },
		{ 66, 356, 42 },
		{ 66, 357, 42 },
		{ 66, 358, 42 },
		{ 66, 359, 42 },
		{ 66, 360, 42 },
		{ 66, 361, 42 },
		{ 66, 362, 42 },
		{ 66, 363, 42 },
		{ 66, 364, 42 },
		{ 66, 365, 42 },
		{ 66, 366, 42 },
		{ 66, 367, 42 },
		{ 66, 368, 42 },
		{ 66, 369, 42 },
		{ 66, 370, 42 },
		{ 66, 371, 42 },
		{ 66, 372, 42 },
		{ 66, 373, 42 },
		{ 66, 374, 42 },
		{ 66, 375, 42 },
		{ 66, 376, 42 },
		{ 67, 329, 42 },
		{ 67, 330, 42 },
		{ 67, 331, 42 },
		{ 67, 332, 42 },
		{ 67, 333, 42 },
		{ 67, 334, 42 },
		{ 67, 335, 42 },
		{ 67, 336, 42 },
		{ 67, 337, 42 },
		{ 67, 338, 42 },
		{ 67, 339, 42 },
		{ 67, 340, 42 },
		{ 67, 341, 42 },
		{ 67, 342, 42 },
		{ 67, 343, 42 },
		{ 67, 344, 42 },
		{ 67, 345, 42 },
		{ 67, 346, 42 },
		{ 67, 347, 42 },
		{ 67, 348, 42 },
		{ 67, 349, 42 },
		{ 67, 350, 42 },
		{ 67, 351, 42 },
		{ 67, 352, 42 },
		{ 67, 353, 42 },
		{ 67, 354, 42 },
		{ 67, 355, 42 },
		{ 67, 356, 42 },
		{ 67, 357, 42 },
		{ 67, 358, 42 },
		{ 67, 359, 42 },
		{ 67, 360, 42 },
		{ 67, 361, 42 },
		{ 67, 362, 42 },
		{ 67, 363, 42 },
		{ 67, 364, 42 },
		{ 67, 365, 42 },
		{ 67, 366, 42 },
		{ 67, 367, 42 },
		{ 67, 368, 42 },
		{ 67, 369, 42 },
		{ 67, 370, 42 },
		{ 67, 371, 42 },
		{ 67, 372, 42 },
		{ 67, 373, 42 },
		{ 67, 374, 42 },
		{ 67, 375, 42 },
		{ 68, 328, 42 },
		{ 68, 329, 42 },
		{ 68, 330, 42 },
		{ 68, 331, 42 },
		{ 68, 332, 42 },
		{ 68, 333, 42 },
		{ 68, 334, 42 },
		{ 68, 335, 42 },
		{ 68, 336, 42 },
		{ 68, 337, 42 },
		{ 68, 338, 42 },
		{ 68, 339, 42 },
		{ 68, 340, 42 },
		{ 68, 341, 42 },
		{ 68, 342, 42 },
		{ 68, 343, 42 },
		{ 68, 344, 42 },
		{ 68, 345, 42 },
		{ 68, 346, 42 },
		{ 68, 347, 42 },
		{ 68, 348, 42 },
		{ 68, 349, 42 },
		{ 68, 350, 42 },
		{ 68, 351, 42 },
		{ 68, 352, 42 },
		{ 68, 353, 42 },
		{ 68, 354, 42 },
		{ 68, 355, 42 },
		{ 68, 356, 42 },
		{ 68, 357, 42 },
		{ 68, 358, 42 },
		{ 68, 359, 42 },
		{ 68, 360, 42 },
		{ 68, 361, 42 },
		{ 68, 362, 42 },
		{ 68, 363, 42 },
		{ 68, 364, 42 },
		{ 68, 365, 42 },
		{ 68, 366, 42 },
		{ 68, 367, 42 },
		{ 68, 368, 42 },
		{ 68, 369, 42 },
		{ 68, 370, 42 },
		{ 68, 371, 42 },
		{ 68, 372, 42 },
		{ 68, 373, 42 },
		{ 68, 374, 42 },
		{ 68, 375, 42 },
		{ 69, 328, 42 },
		{ 69, 329, 42 },
		{ 69, 330, 42 },
		{ 69, 331, 42 },
		{ 69, 332, 42 },
		{ 69, 333, 42 },
		{ 69, 334, 42 },
		{ 69, 335, 42 },
		{ 69, 336, 42 },
		{ 69, 337, 42 },
		{ 69, 338, 42 },
		{ 69, 339, 42 },
		{ 69, 340, 42 },
		{ 69, 341, 42 },
		{ 69, 342, 42 },
		{ 69, 343, 42 },
		{ 69, 344, 42 },
		{ 69, 345, 42 },
		{ 69, 346, 42 },
		{ 69, 347, 42 },
		{ 69, 348, 42 },
		{ 69, 349, 42 },
		{ 69, 350, 42 },
		{ 69, 351, 42 },
		{ 69, 352, 42 },
		{ 69, 353, 42 },
		{ 69, 354, 42 },
		{ 69, 355, 42 },
		{ 69, 356, 42 },
		{ 69, 357, 42 },
		{ 69, 358, 42 },
		{ 69, 359, 42 },
		{ 69, 360, 42 },
		{ 69, 361, 42 },
		{ 69, 362, 42 },
		{ 69, 363, 42 },
		{ 69, 364, 42 },
		{ 69, 365, 42 },
		{ 69, 366, 42 },
		{ 69, 367, 42 },
		{ 69, 368, 42 },
		{ 69, 369, 42 },
		{ 69, 370, 42 },
		{ 69, 371, 42 },
		{ 69, 372, 42 },
		{ 69, 373, 42 },
		{ 69, 374, 42 },
		{ 70, 328, 42 },
		{ 70, 329, 42 },
		{ 70, 330, 42 },
		{ 70, 331, 42 },
		{ 70, 332, 42 },
		{ 70, 333, 42 },
		{ 70, 334, 42 },
		{ 70, 335, 42 },
		{ 70, 336, 42 },
		{ 70, 337, 42 },
		{ 70, 338, 42 },
		{ 70, 339, 42 },
		{ 70, 340, 42 },
		{ 70, 341, 42 },
		{ 70, 342, 42 },
		{ 70, 343, 42 },
		{ 70, 344, 42 },
		{ 70, 345, 42 },
		{ 70, 346, 42 },
		{ 70, 347, 42 },
		{ 70, 348, 42 },
		{ 70, 349, 42 },
		{ 70, 350, 42 },
		{ 70, 351, 42 },
		{ 70, 352, 42 },
		{ 70, 353, 42 },
		{ 70, 354, 42 },
		{ 70, 355, 42 },
		{ 70, 356, 42 },
		{ 70, 357, 42 },
		{ 70, 358, 42 },
		{ 70, 359, 42 },
		{ 70, 360, 42 },
		{ 70, 361, 42 },
		{ 70, 362, 42 },
		{ 70, 363, 42 },
		{ 70, 364, 42 },
		{ 70, 365, 42 },
		{ 70, 366, 42 },
		{ 70, 367, 42 },
		{ 70, 368, 42 },
		{ 70, 369, 42 },
		{ 70, 370, 42 },
		{ 70, 371, 42 },
		{ 70, 372, 42 },
		{ 70, 373, 42 },
		{ 71, 327, 42 },
		{ 71, 328, 42 },
		{ 71, 329, 42 },
		{ 71, 330, 42 },
		{ 71, 331, 42 },
		{ 71, 332, 42 },
		{ 71, 333, 42 },
		{ 71, 334, 42 },
		{ 71, 335, 42 },
		{ 71, 336, 42 },
		{ 71, 337, 42 },
		{ 71, 338, 42 },
		{ 71, 339, 42 },
		{ 71, 340, 42 },
		{ 71, 341, 42 },
		{ 71, 342, 42 },
		{ 71, 343, 42 },
		{ 71, 344, 42 },
		{ 71, 345, 42 },
		{ 71, 346, 42 },
		{ 71, 347, 42 },
		{ 71, 348, 42 },
		{ 71, 349, 42 },
		{ 71, 350, 42 },
		{ 71, 351, 42 },
		{ 71, 352, 42 },
		{ 71, 353, 42 },
		{ 71, 354, 42 },
		{ 71, 355, 42 },
		{ 71, 356, 42 },
		{ 71, 357, 42 },
		{ 71, 358, 42 },
		{ 71, 359, 42 },
		{ 71, 360, 42 },
		{ 71, 361, 42 },
		{ 71, 362, 42 },
		{ 71, 363, 42 },
		{ 71, 364, 42 },
		{ 71, 365, 42 },
		{ 71, 366, 42 },
		{ 71, 367, 42 },
		{ 71, 368, 42 },
		{ 71, 369, 42 },
		{ 71, 370, 42 },
		{ 71, 371, 42 },
		{ 71, 372, 42 },
		{ 71, 373, 42 },
		{ 72, 327, 42 },
		{ 72, 328, 42 },
		{ 72, 329, 42 },
		{ 72, 330, 42 },
		{ 72, 331, 42 },
		{ 72, 332, 42 },
		{ 72, 333, 42 },
		{ 72, 334, 42 },
		{ 72, 335, 42 },
		{ 72, 336, 42 },
		{ 72, 337, 42 },
		{ 72, 338, 42 },
		{ 72, 339, 42 },
		{ 72, 340, 42 },
		{ 72, 341, 42 },
		{ 72, 342, 42 },
		{ 72, 343, 42 },
		{ 72, 344, 42 },
		{ 72, 345, 42 },
		{ 72, 346, 42 },
		{ 72, 347, 42 },
		{ 72, 348, 42 },
		{ 72, 349, 42 },
		{ 72, 350, 42 },
		{ 72, 351, 42 },
		{ 72, 352, 42 },
		{ 72, 353, 42 },
		{ 72, 354, 42 },
		{ 72, 355, 42 },
		{ 72, 356, 42 },
		{ 72, 357, 42 },
		{ 72, 358, 42 },
		{ 72, 359, 42 },
		{ 72, 360, 42 },
		{ 72, 361, 42 },
		{ 72, 362, 42 },
		{ 72, 363, 42 },
		{ 72, 364, 42 },
		{ 72, 365, 42 },
		{ 72, 366, 42 },
		{ 72, 367, 42 },
		{ 72, 368, 42 },
		{ 72, 369, 42 },
		{ 72, 370, 42 },
		{ 72, 371, 42 },
		{ 72, 372, 42 },
		{ 73, 327, 42 },
		{ 73, 328, 42 },
		{ 73, 329, 42 },
		{ 73, 330, 42 },
		{ 73, 331, 42 },
		{ 73, 332, 42 },
		{ 73, 333, 42 },
		{ 73, 334, 42 },
		{ 73, 335, 42 },
		{ 73, 336, 42 },
		{ 73, 337, 42 },
		{ 73, 338, 42 },
		{ 73, 339, 42 },
		{ 73, 340, 42 },
		{ 73, 341, 42 },
		{ 73, 342, 42 },
		{ 73, 343, 42 },
		{ 73, 344, 42 },
		{ 73, 345, 42 },
		{ 73, 346, 42 },
		{ 73, 347, 42 },
		{ 73, 348, 42 },
		{ 73, 349, 42 },
		{ 73, 350, 42 },
		{ 73, 351, 42 },
		{ 73, 352, 42 },
		{ 73, 353, 42 },
		{ 73, 354, 42 },
		{ 73, 355, 42 },
		{ 73, 356, 42 },
		{ 73, 357, 42 },
		{ 73, 358, 42 },
		{ 73, 359, 42 },
		{ 73, 360, 42 },
		{ 73, 361, 42 },
		{ 73, 362, 42 },
		{ 73, 363, 42 },
		{ 73, 364, 42 },
		{ 73, 365, 42 },
		{ 73, 366, 42 },
		{ 73, 367, 42 },
		{ 73, 368, 42 },
		{ 73, 369, 42 },
		{ 73, 370, 42 },
		{ 73, 371, 42 },
		{ 74, 326, 42 },
		{ 74, 327, 42 },
		{ 74, 328, 42 },
		{ 74, 329, 42 },
		{ 74, 330, 42 },
		{ 74, 331, 42 },
		{ 74, 332, 42 },
		{ 74, 333, 42 },
		{ 74, 334, 42 },
		{ 74, 335, 42 },
		{ 74, 336, 42 },
		{ 74, 337, 42 },
		{ 74, 338, 42 },
		{ 74, 339, 42 },
		{ 74, 340, 42 },
		{ 74, 341, 42 },
		{ 74, 342, 42 },
		{ 74, 343, 42 },
		{ 74, 344, 42 },
		{ 74, 345, 42 },
		{ 74, 346, 42 },
		{ 74, 347, 42 },
		{ 74, 348, 42 },
		{ 74, 349, 42 },
		{ 74, 350, 42 },
		{ 74, 351, 42 },
		{ 74, 352, 42 },
		{ 74, 353, 42 },
		{ 74, 354, 42 },
		{ 74, 355, 42 },
		{ 74, 356, 42 },
		{ 74, 357, 42 },
		{ 74, 358, 42 },
		{ 74, 359, 42 },
		{ 74, 360, 42 },
		{ 74, 361, 42 },
		{ 74, 362, 42 },
		{ 74, 363, 42 },
		{ 74, 364, 42 },
		{ 74, 365, 42 },
		{ 74, 366, 42 },
		{ 74, 367, 42 },
		{ 74, 368, 42 },
		{ 74, 369, 42 },
		{ 74, 370, 42 },
		{ 74, 371, 42 },
		{ 75, 326, 42 },
		{ 75, 327, 42 },
		{ 75, 328, 42 },
		{ 75, 329, 42 },
		{ 75, 330, 42 },
		{ 75, 331, 42 },
		{ 75, 332, 42 },
		{ 75, 333, 42 },
		{ 75, 334, 42 },
		{ 75, 335, 42 },
		{ 75, 336, 42 },
		{ 75, 337, 42 },
		{ 75, 338, 42 },
		{ 75, 339, 42 },
		{ 75, 340, 42 },
		{ 75, 341, 42 },
		{ 75, 342, 42 },
		{ 75, 343, 42 },
		{ 75, 344, 42 },
		{ 75, 345, 42 },
		{ 75, 346, 42 },
		{ 75, 347, 42 },
		{ 75, 348, 42 },
		{ 75, 349, 42 },
		{ 75, 350, 42 },
		{ 75, 351, 42 },
		{ 75, 352, 42 },
		{ 75, 353, 42 },
		{ 75, 354, 42 },
		{ 75, 355, 42 },
		{ 75, 356, 42 },
		{ 75, 357, 42 },
		{ 75, 358, 42 },
		{ 75, 359, 42 },
		{ 75, 360, 42 },
		{ 75, 361, 42 },
		{ 75, 362, 42 },
		{ 75, 363, 42 },
		{ 75, 364, 42 },
		{ 75, 365, 42 },
		{ 75, 366, 42 },
		{ 75, 367, 42 },
		{ 75, 368, 42 },
		{ 75, 369, 42 },
		{ 75, 370, 42 },
		{ 76, 325, 42 },
		{ 76, 326, 42 },
		{ 76, 327, 42 },
		{ 76, 328, 42 },
		{ 76, 329, 42 },
		{ 76, 330, 42 },
		{ 76, 331, 42 },
		{ 76, 332, 42 },
		{ 76, 333, 42 },
		{ 76, 334, 42 },
		{ 76, 335, 42 },
		{ 76, 336, 42 },
		{ 76, 337, 42 },
		{ 76, 338, 42 },
		{ 76, 339, 42 },
		{ 76, 340, 42 },
		{ 76, 341, 42 },
		{ 76, 342, 42 },
		{ 76, 343, 42 },
		{ 76, 344, 42 },
		{ 76, 345, 42 },
		{ 76, 346, 42 },
		{ 76, 347, 42 },
		{ 76, 348, 42 },
		{ 76, 349, 42 },
		{ 76, 350, 42 },
		{ 76, 351, 42 },
		{ 76, 352, 42 },
		{ 76, 353, 42 },
		{ 76, 354, 42 },
		{ 76, 355, 42 },
		{ 76, 356, 42 },
		{ 76, 357, 42 },
		{ 76, 358, 42 },
		{ 76, 359, 42 },
		{ 76, 360, 42 },
		{ 76, 361, 42 },
		{ 76, 362, 42 },
		{ 76, 363, 42 },
		{ 76, 364, 42 },
		{ 76, 365, 42 },
		{ 76, 366, 42 },
		{ 76, 367, 42 },
		{ 76, 368, 42 },
		{ 76, 369, 42 },
		{ 76, 370, 42 },
		{ 77, 325, 42 },
		{ 77, 326, 42 },
		{ 77, 327, 42 },
		{ 77, 328, 42 },
		{ 77, 329, 42 },
		{ 77, 330, 42 },
		{ 77, 331, 42 },
		{ 77, 332, 42 },
		{ 77, 333, 42 },
		{ 77, 334, 42 },
		{ 77, 335, 42 },
		{ 77, 336, 42 },
		{ 77, 337, 42 },
		{ 77, 338, 42 },
		{ 77, 339, 42 },
		{ 77, 340, 42 },
		{ 77, 341, 42 },
		{ 77, 342, 42 },
		{ 77, 343, 42 },
		{ 77, 344, 42 },
		{ 77, 345, 42 },
		{ 77, 346, 42 },
		{ 77, 347, 42 },
		{ 77, 348, 42 },
		{ 77, 349, 42 },
		{ 77, 350, 42 },
		{ 77, 351, 42 },
		{ 77, 352, 42 },
		{ 77, 353, 42 },
		{ 77, 354, 42 },
		{ 77, 355, 42 },
		{ 77, 356, 42 },
		{ 77, 357, 42 },
		{ 77, 358, 42 },
		{ 77, 359, 42 },
		{ 77, 360, 42 },
		{ 77, 361, 42 },
		{ 77, 362, 42 },
		{ 77, 363, 42 },
		{ 77, 364, 42 },
		{ 77, 365, 42 },
		{ 77, 366, 42 },
		{ 77, 367, 42 },
		{ 77, 368, 42 },
		{ 77, 369, 42 },
		{ 78, 325, 42 },
		{ 78, 326, 42 },
		{ 78, 327, 42 },
		{ 78, 328, 42 },
		{ 78, 329, 42 },
		{ 78, 330, 42 },
		{ 78, 331, 42 },
		{ 78, 332, 42 },
		{ 78, 333, 42 },
		{ 78, 334, 42 },
		{ 78, 335, 42 },
		{ 78, 336, 42 },
		{ 78, 337, 42 },
		{ 78, 338, 42 },
		{ 78, 339, 42 },
		{ 78, 340, 42 },
		{ 78, 341, 42 },
		{ 78, 342, 42 },
		{ 78, 343, 42 },
		{ 78, 344, 42 },
		{ 78, 345, 42 },
		{ 78, 346, 42 },
		{ 78, 347, 42 },
		{ 78, 348, 42 },
		{ 78, 349, 42 },
		{ 78, 350, 42 },
		{ 78, 351, 42 },
		{ 78, 352, 42 },
		{ 78, 353, 42 },
		{ 78, 354, 42 },
		{ 78, 355, 42 },
		{ 78, 356, 42 },
		{ 78, 357, 42 },
		{ 78, 358, 42 },
		{ 78, 359, 42 },
		{ 78, 360, 42 },
		{ 78, 361, 42 },
		{ 78, 362, 42 },
		{ 78, 363, 42 },
		{ 78, 364, 42 },
		{ 78, 365, 42 },
		{ 78, 366, 42 },
		{ 78, 367, 42 },
		{ 78, 368, 42 },
		{ 79, 324, 42 },
		{ 79, 325, 42 },
		{ 79, 326, 42 },
		{ 79, 327, 42 },
		{ 79, 328, 42 },
		{ 79, 329, 42 },
		{ 79, 330, 42 },
		{ 79, 331, 42 },
		{ 79, 332, 42 },
		{ 79, 333, 42 },
		{ 79, 334, 42 },
		{ 79, 335, 42 },
		{ 79, 336, 42 },
		{ 79, 337, 42 },
		{ 79, 338, 42 },
		{ 79, 339, 42 },
		{ 79, 340, 42 },
		{ 79, 341, 42 },
		{ 79, 342, 42 },
		{ 79, 343, 42 },
		{ 79, 344, 42 },
		{ 79, 345, 42 },
		{ 79, 346, 42 },
		{ 79, 347, 42 },
		{ 79, 348, 42 },
		{ 79, 349, 42 },
		{ 79, 350, 42 },
		{ 79, 351, 42 },
		{ 79, 352, 42 },
		{ 79, 353, 42 },
		{ 79, 354, 42 },
		{ 79, 355, 42 },
		{ 79, 356, 42 },
		{ 79, 357, 42 },
		{ 79, 358, 42 },
		{ 79, 359, 42 },
		{ 79, 360, 42 },
		{ 79, 361, 42 },
		{ 79, 362, 42 },
		{ 79, 363, 42 },
		{ 79, 364, 42 },
		{ 79, 365, 42 },
		{ 79, 366, 42 },
		{ 79, 367, 42 },
		{ 79, 368, 42 },
		{ 80, 324, 42 },
		{ 80, 325, 42 },
		{ 80, 326, 42 },
		{ 80, 327, 42 },
		{ 80, 328, 42 },
		{ 80, 329, 42 },
		{ 80, 330, 42 },
		{ 80, 331, 42 },
		{ 80, 332, 42 },
		{ 80, 333, 42 },
		{ 80, 334, 42 },
		{ 80, 335, 42 },
		{ 80, 336, 42 },
		{ 80, 337, 42 },
		{ 80, 338, 42 },
		{ 80, 339, 42 },
		{ 80, 340, 42 },
		{ 80, 341, 42 },
		{ 80, 342, 42 },
		{ 80, 343, 42 },
		{ 80, 344, 42 },
		{ 80, 345, 42 },
		{ 80, 346, 42 },
		{ 80, 347, 42 },
		{ 80, 348, 42 },
		{ 80, 349, 42 },
		{ 80, 350, 42 },
		{ 80, 351, 42 },
		{ 80, 352, 42 },
		{ 80, 353, 42 },
		{ 80, 354, 42 },
		{ 80, 355, 42 },
		{ 80, 356, 42 },
		{ 80, 357, 42 },
		{ 80, 358, 42 },
		{ 80, 359, 42 },
		{ 80, 360, 42 },
		{ 80, 361, 42 },
		{ 80, 362, 42 },
		{ 80, 363, 42 },
		{ 80, 364, 42 },
		{ 80, 365, 42 },
		{ 80, 366, 42 },
		{ 80, 367, 42 },
		{ 81, 324, 42 },
		{ 81, 325, 42 },
		{ 81, 326, 42 },
		{ 81, 327, 42 },
		{ 81, 328, 42 },
		{ 81, 329, 42 },
		{ 81, 330, 42 },
		{ 81, 331, 42 },
		{ 81, 332, 42 },
		{ 81, 333, 42 },
		{ 81, 334, 42 },
		{ 81, 335, 42 },
		{ 81, 336, 42 },
		{ 81, 337, 42 },
		{ 81, 338, 42 },
		{ 81, 339, 42 },
		{ 81, 340, 42 },
		{ 81, 341, 42 },
		{ 81, 342, 42 },
		{ 81, 343, 42 },
		{ 81, 344, 42 },
		{ 81, 345, 42 },
		{ 81, 346, 42 },
		{ 81, 347, 42 },
		{ 81, 348, 42 },
		{ 81, 349, 42 },
		{ 81, 350, 42 },
		{ 81, 351, 42 },
		{ 81, 352, 42 },
		{ 81, 353, 42 },
		{ 81, 354, 42 },
		{ 81, 355, 42 },
		{ 81, 356, 42 },
		{ 81, 357, 42 },
		{ 81, 358, 42 },
		{ 81, 359, 42 },
		{ 81, 360, 42 },
		{ 81, 361, 42 },
		{ 81, 362, 42 },
		{ 81, 363, 42 },
		{ 81, 364, 42 },
		{ 81, 365, 42 },
		{ 81, 366, 42 },
		{ 82, 323, 42 },
		{ 82, 324, 42 },
		{ 82, 325, 42 },
		{ 82, 326, 42 },
		{ 82, 327, 42 },
		{ 82, 328, 42 },
		{ 82, 329, 42 },
		{ 82, 330, 42 },
		{ 82, 331, 42 },
		{ 82, 332, 42 },
		{ 82, 333, 42 },
		{ 82, 334, 42 },
		{ 82, 335, 42 },
		{ 82, 336, 42 },
		{ 82, 337, 42 },
		{ 82, 338, 42 },
		{ 82, 339, 42 },
		{ 82, 340, 42 },
		{ 82, 341, 42 },
		{ 82, 342, 42 },
		{ 82, 343, 42 },
		{ 82, 344, 42 },
		{ 82, 345, 42 },
		{ 82, 346, 42 },
		{ 82, 347, 42 },
		{ 82, 348, 42 },
		{ 82, 349, 42 },
		{ 82, 350, 42 },
		{ 82, 351, 42 },
		{ 82, 352, 42 },
		{ 82, 353, 42 },
		{ 82, 354, 42 },
		{ 82, 355, 42 },
		{ 82, 356, 42 },
		{ 82, 357, 42 },
		{ 82, 358, 42 },
		{ 82, 359, 42 },
		{ 82, 360, 42 },
		{ 82, 361, 42 },
		{ 82, 362, 42 },
		{ 82, 363, 42 },
		{ 82, 364, 42 },
		{ 82, 365, 42 },
		{ 82, 366, 42 },
		{ 83, 323, 42 },
		{ 83, 324, 42 },
		{ 83, 325, 42 },
		{ 83, 326, 42 },
		{ 83, 327, 42 },
		{ 83, 328, 42 },
		{ 83, 329, 42 },
		{ 83, 330, 42 },
		{ 83, 331, 42 },
		{ 83, 332, 42 },
		{ 83, 333, 42 },
		{ 83, 334, 42 },
		{ 83, 335, 42 },
		{ 83, 336, 42 },
		{ 83, 337, 42 },
		{ 83, 338, 42 },
		{ 83, 339, 42 },
		{ 83, 340, 42 },
		{ 83, 341, 42 },
		{ 83, 342, 42 },
		{ 83, 343, 42 },
		{ 83, 344, 42 },
		{ 83, 345, 42 },
		{ 83, 346, 42 },
		{ 83, 347, 42 },
		{ 83, 348, 42 },
		{ 83, 349, 42 },
		{ 83, 350, 42 },
		{ 83, 351, 42 },
		{ 83, 352, 42 },
		{ 83, 353, 42 },
		{ 83, 354, 42 },
		{ 83, 355, 42 },
		{ 83, 356, 42 },
		{ 83, 357, 42 },
		{ 83, 358, 42 },
		{ 83, 359, 42 },
		{ 83, 360, 42 },
		{ 83, 361, 42 },
		{ 83, 362, 42 },
		{ 83, 363, 42 },
		{ 83, 364, 42 },
		{ 83, 365, 42 },
		{ 84, 322, 42 },
		{ 84, 323, 42 },
		{ 84, 324, 42 },
		{ 84, 325, 42 },
		{ 84, 326, 42 },
		{ 84, 327, 42 },
		{ 84, 328, 42 },
		{ 84, 329, 42 },
		{ 84, 330, 42 },
		{ 84, 331, 42 },
		{ 84, 332, 42 },
		{ 84, 333, 42 },
		{ 84, 334, 42 },
		{ 84, 335, 42 },
		{ 84, 336, 42 },
		{ 84, 337, 42 },
		{ 84, 338, 42 },
		{ 84, 339, 42 },
		{ 84, 340, 42 },
		{ 84, 341, 42 },
		{ 84, 342, 42 },
		{ 84, 343, 42 },
		{ 84, 344, 42 },
		{ 84, 345, 42 },
		{ 84, 346, 42 },
		{ 84, 347, 42 },
		{ 84, 348, 42 },
		{ 84, 349, 42 },
		{ 84, 350, 42 },
		{ 84, 351, 42 },
		{ 84, 352, 42 },
		{ 84, 353, 42 },
		{ 84, 354, 42 },
		{ 84, 355, 42 },
		{ 84, 356, 42 },
		{ 84, 357, 42 },
		{ 84, 358, 42 },
		{ 84, 359, 42 },
		{ 84, 360, 42 },
		{ 84, 361, 42 },
		{ 84, 362, 42 },
		{ 84, 363, 42 },
		{ 84, 364, 42 },
		{ 84, 365, 42 },
		{ 85, 322, 42 },
		{ 85, 323, 42 },
		{ 85, 324, 42 },
		{ 85, 325, 42 },
		{ 85, 326, 42 },
		{ 85, 327, 42 },
		{ 85, 328, 42 },
		{ 85, 329, 42 },
		{ 85, 330, 42 },
		{ 85, 331, 42 },
		{ 85, 332, 42 },
		{ 85, 333, 42 },
		{ 85, 334, 42 },
		{ 85, 335, 42 },
		{ 85, 336, 42 },
		{ 85, 337, 42 },
		{ 85, 338, 42 },
		{ 85, 339, 42 },
		{ 85, 340, 42 },
		{ 85, 341, 42 },
		{ 85, 342, 42 },
		{ 85, 343, 42 },
		{ 85, 344, 42 },
		{ 85, 345, 42 },
		{ 85, 346, 42 },
		{ 85, 347, 42 },
		{ 85, 348, 42 },
		{ 85, 349, 42 },
		{ 85, 350, 42 },
		{ 85, 351, 42 },
		{ 85, 352, 42 },
		{ 85, 353, 42 },
		{ 85, 354, 42 },
		{ 85, 355, 42 },
		{ 85, 356, 42 },
		{ 85, 357, 42 },
		{ 85, 358, 42 },
		{ 85, 359, 42 },
		{ 85, 360, 42 },
		{ 85, 361, 42 },
		{ 85, 362, 42 },
		{ 85, 363, 42 },
		{ 85, 364, 42 },
		{ 86, 322, 42 },
		{ 86, 323, 42 },
		{ 86, 324, 42 },
		{ 86, 325, 42 },
		{ 86, 326, 42 },
		{ 86, 327, 42 },
		{ 86, 328, 42 },
		{ 86, 329, 42 },
		{ 86, 330, 42 },
		{ 86, 331, 42 },
		{ 86, 332, 42 },
		{ 86, 333, 42 },
		{ 86, 334, 42 },
		{ 86, 335, 42 },
		{ 86, 336, 42 },
		{ 86, 337, 42 },
		{ 86, 338, 42 },
		{ 86, 339, 42 },
		{ 86, 340, 42 },
		{ 86, 341, 42 },
		{ 86, 342, 42 },
		{ 86, 343, 42 },
		{ 86, 344, 42 },
		{ 86, 345, 42 },
		{ 86, 346, 42 },
		{ 86, 347, 42 },
		{ 86, 348, 42 },
		{ 86, 349, 42 },
		{ 86, 350, 42 },
		{ 86, 351, 42 },
		{ 86, 352, 42 },
		{ 86, 353, 42 },
		{ 86, 354, 42 },
		{ 86, 355, 42 },
		{ 86, 356, 42 },
		{ 86, 357, 42 },
		{ 86, 358, 42 },
		{ 86, 359, 42 },
		{ 86, 360, 42 },
		{ 86, 361, 42 },
		{ 86, 362, 42 },
		{ 86, 363, 42 },
		{ 87, 321, 42 },
		{ 87, 322, 42 },
		{ 87, 323, 42 },
		{ 87, 324, 42 },
		{ 87, 325, 42 },
		{ 87, 326, 42 },
		{ 87, 327, 42 },
		{ 87, 328, 42 },
		{ 87, 329, 42 },
		{ 87, 330, 42 },
		{ 87, 331, 42 },
		{ 87, 332, 42 },
		{ 87, 333, 42 },
		{ 87, 334, 42 },
		{ 87, 335, 42 },
		{ 87, 336, 42 },
		{ 87, 337, 42 },
		{ 87, 338, 42 },
		{ 87, 339, 42 },
		{ 87, 340, 42 },
		{ 87, 341, 42 },
		{ 87, 342, 42 },
		{ 87, 343, 42 },
		{ 87, 344, 42 },
		{ 87, 345, 42 },
		{ 87, 346, 42 },
		{ 87, 347, 42 },
		{ 87, 348, 42 },
		{ 87, 349, 42 },
		{ 87, 350, 42 },
		{ 87, 351, 42 },
		{ 87, 352, 42 },
		{ 87, 353, 42 },
		{ 87, 354, 42 },
		{ 87, 355, 42 },
		{ 87, 356, 42 },
		{ 87, 357, 42 },
		{ 87, 358, 42 },
		{ 87, 359, 42 },
		{ 87, 360, 42 },
		{ 87, 361, 42 },
		{ 87, 362, 42 },
		{ 87, 363, 42 },
		{ 88, 321, 42 },
		{ 88, 322, 42 },
		{ 88, 323, 42 },
		{ 88, 324, 42 },
		{ 88, 325, 42 },
		{ 88, 326, 42 },
		{ 88, 327, 42 },
		{ 88, 328, 42 },
		{ 88, 329, 42 },
		{ 88, 330, 42 },
		{ 88, 331, 42 },
		{ 88, 332, 42 },
		{ 88, 333, 42 },
		{ 88, 334, 42 },
		{ 88, 335, 42 },
		{ 88, 336, 42 },
		{ 88, 337, 42 },
		{ 88, 338, 42 },
		{ 88, 339, 42 },
		{ 88, 340, 42 },
		{ 88, 341, 42 },
		{ 88, 342, 42 },
		{ 88, 343, 42 },
		{ 88, 344, 42 },
		{ 88, 345, 42 },
		{ 88, 346, 42 },
		{ 88, 347, 42 },
		{ 88, 348, 42 },
		{ 88, 349, 42 },
		{ 88, 350, 42 },
		{ 88, 351, 42 },
		{ 88, 352, 42 },
		{ 88, 353, 42 },
		{ 88, 354, 42 },
		{ 88, 355, 42 },
		{ 88, 356, 42 },
		{ 88, 357, 42 },
		{ 88, 358, 42 },
		{ 88, 359, 42 },
		{ 88, 360, 42 },
		{ 88, 361, 42 },
		{ 88, 362, 42 },
		{ 89, 321, 42 },
		{ 89, 322, 42 },
		{ 89, 323, 42 },
		{ 89, 324, 42 },
		{ 89, 325, 42 },
		{ 89, 326, 42 },
		{ 89, 327, 42 },
		{ 89, 328, 42 },
		{ 89, 329, 42 },
		{ 89, 330, 42 },
		{ 89, 331, 42 },
		{ 89, 332, 42 },
		{ 89, 333, 42 },
		{ 89, 334, 42 },
		{ 89, 335, 42 },
		{ 89, 336, 42 },
		{ 89, 337, 42 },
		{ 89, 338, 42 },
		{ 89, 339, 42 },
		{ 89, 340, 42 },
		{ 89, 341, 42 },
		{ 89, 342, 42 },
		{ 89, 343, 42 },
		{ 89, 344, 42 },
		{ 89, 345, 42 },
		{ 89, 346, 42 },
		{ 89, 347, 42 },
		{ 89, 348, 42 },
		{ 89, 349, 42 },
		{ 89, 350, 42 },
		{ 89, 351, 42 },
		{ 89, 352, 42 },
		{ 89, 353, 42 },
		{ 89, 354, 42 },
		{ 89, 355, 42 },
		{ 89, 356, 42 },
		{ 89, 357, 42 },
		{ 89, 358, 42 },
		{ 89, 359, 42 },
		{ 89, 360, 42 },
		{ 89, 361, 42 },
		{ 90, 320, 42 },
		{ 90, 321, 42 },
		{ 90, 322, 42 },
		{ 90, 323, 42 },
		{ 90, 324, 42 },
		{ 90, 325, 42 },
		{ 90, 326, 42 },
		{ 90, 327, 42 },
		{ 90, 328, 42 },
		{ 90, 329, 42 },
		{ 90, 330, 42 },
		{ 90, 331, 42 },
		{ 90, 332, 42 },
		{ 90, 333, 42 },
		{ 90, 334, 42 },
		{ 90, 335, 42 },
		{ 90, 336, 42 },
		{ 90, 337, 42 },
		{ 90, 338, 42 },
		{ 90, 339, 42 },
		{ 90, 340, 42 },
		{ 90, 341, 42 },
		{ 90, 342, 42 },
		{ 90, 343, 42 },
		{ 90, 344, 42 },
		{ 90, 345, 42 },
		{ 90, 346, 42 },
		{ 90, 347, 42 },
		{ 90, 348, 42 },
		{ 90, 349, 42 },
		{ 90, 350, 42 },
		{ 90, 351, 42 },
		{ 90, 352, 42 },
		{ 90, 353, 42 },
		{ 90, 354, 42 },
		{ 90, 355, 42 },
		{ 90, 356, 42 },
		{ 90, 357, 42 },
		{ 90, 358, 42 },
		{ 90, 359, 42 },
		{ 90, 360, 42 },
		{ 90, 361, 42 },
		{ 91, 320, 42 },
		{ 91, 321, 42 },
		{ 91, 322, 42 },
		{ 91, 323, 42 },
		{ 91, 324, 42 },
		{ 91, 325, 42 },
		{ 91, 326, 42 },
		{ 91, 327, 42 },
		{ 91, 328, 42 },
		{ 91, 329, 42 },
		{ 91, 330, 42 },
		{ 91, 331, 42 },
		{ 91, 332, 42 },
		{ 91, 333, 42 },
		{ 91, 334, 42 },
		{ 91, 335, 42 },
		{ 91, 336, 42 },
		{ 91, 337, 42 },
		{ 91, 338, 42 },
		{ 91, 339, 42 },
		{ 91, 340, 42 },
		{ 91, 341, 42 },
		{ 91, 342, 42 },
		{ 91, 343, 42 },
		{ 91, 344, 42 },
		{ 91, 345, 42 },
		{ 91, 346, 42 },
		{ 91, 347, 42 },
		{ 91, 348, 42 },
		{ 91, 349, 42 },
		{ 91, 350, 42 },
		{ 91, 351, 42 },
		{ 91, 352, 42 },
		{ 91, 353, 42 },
		{ 91, 354, 42 },
		{ 91, 355, 42 },
		{ 91, 356, 42 },
		{ 91, 357, 42 },
		{ 91, 358, 42 },
		{ 91, 359, 42 },
		{ 91, 360, 42 },
		{ 92, 319, 42 },
		{ 92, 320, 42 },
		{ 92, 321, 42 },
		{ 92, 322, 42 },
		{ 92, 323, 42 },
		{ 92, 324, 42 },
		{ 92, 325, 42 },
		{ 92, 326, 42 },
		{ 92, 327, 42 },
		{ 92, 328, 42 },
		{ 92, 329, 42 },
		{ 92, 330, 42 },
		{ 92, 331, 42 },
		{ 92, 332, 42 },
		{ 92, 333, 42 },
		{ 92, 334, 42 },
		{ 92, 335, 42 },
		{ 92, 336, 42 },
		{ 92, 337, 42 },
		{ 92, 338, 42 },
		{ 92, 339, 42 },
		{ 92, 340, 42 },
		{ 92, 341, 42 },
		{ 92, 342, 42 },
		{ 92, 343, 42 },
		{ 92, 344, 42 },
		{ 92, 345, 42 },
		{ 92, 346, 42 },
		{ 92, 347, 42 },
		{ 92, 348, 42 },
		{ 92, 349, 42 },
		{ 92, 350, 42 },
		{ 92, 351, 42 },
		{ 92, 352, 42 },
		{ 92, 353, 42 },
		{ 92, 354, 42 },
		{ 92, 355, 42 },
		{ 92, 356, 42 },
		{ 92, 357, 42 },
		{ 92, 358, 42 },
		{ 92, 359, 42 },
		{ 92, 360, 42 },
		{ 93, 319, 42 },
		{ 93, 320, 42 },
		{ 93, 321, 42 },
		{ 93, 322, 42 },
		{ 93, 323, 42 },
		{ 93, 324, 42 },
		{ 93, 325, 42 },
		{ 93, 326, 42 },
		{ 93, 327, 42 },
		{ 93, 328, 42 },
		{ 93, 329, 42 },
		{ 93, 330, 42 },
		{ 93, 331, 42 },
		{ 93, 332, 42 },
		{ 93, 333, 42 },
		{ 93, 334, 42 },
		{ 93, 335, 42 },
		{ 93, 336, 42 },
		{ 93, 337, 42 },
		{ 93, 338, 42 },
		{ 93, 339, 42 },
		{ 93, 340, 42 },
		{ 93, 341, 42 },
		{ 93, 342, 42 },
		{ 93, 343, 42 },
		{ 93, 344, 42 },
		{ 93, 345, 42 },
		{ 93, 346, 42 },
		{ 93, 347, 42 },
		{ 93, 348, 42 },
		{ 93, 349, 42 },
		{ 93, 350, 42 },
		{ 93, 351, 42 },
		{ 93, 352, 42 },
		{ 93, 353, 42 },
		{ 93, 354, 42 },
		{ 93, 355, 42 },
		{ 93, 356, 42 },
		{ 93, 357, 42 },
		{ 93, 358, 42 },
		{ 93, 359, 42 },
		{ 94, 319, 42 },
		{ 94, 320, 42 },
		{ 94, 321, 42 },
		{ 94, 322, 42 },
		{ 94, 323, 42 },
		{ 94, 324, 42 },
		{ 94, 325, 42 },
		{ 94, 326, 42 },
		{ 94, 327, 42 },
		{ 94, 328, 42 },
		{ 94, 329, 42 },
		{ 94, 330, 42 },
		{ 94, 331, 42 },
		{ 94, 332, 42 },
		{ 94, 333, 42 },
		{ 94, 334, 42 },
		{ 94, 335, 42 },
		{ 94, 336, 42 },
		{ 94, 337, 42 },
		{ 94, 338, 42 },
		{ 94, 339, 42 },
		{ 94, 340, 42 },
		{ 94, 341, 42 },
		{ 94, 342, 42 },
		{ 94, 343, 42 },
		{ 94, 344, 42 },
		{ 94, 345, 42 },
		{ 94, 346, 42 },
		{ 94, 347, 42 },
		{ 94, 348, 42 },
		{ 94, 349, 42 },
		{ 94, 350, 42 },
		{ 94, 351, 42 },
		{ 94, 352, 42 },
		{ 94, 353, 42 },
		{ 94, 354, 42 },
		{ 94, 355, 42 },
		{ 94, 356, 42 },
		{ 94, 357, 42 },
		{ 94, 358, 42 },
		{ 95, 318, 42 },
		{ 95, 319, 42 },
		{ 95, 320, 42 },
		{ 95, 321, 42 },
		{ 95, 322, 42 },
		{ 95, 323, 42 },
		{ 95, 324, 42 },
		{ 95, 325, 42 },
		{ 95, 326, 42 },
		{ 95, 327, 42 },
		{ 95, 328, 42 },
		{ 95, 329, 42 },
		{ 95, 330, 42 },
		{ 95, 331, 42 },
		{ 95, 332, 42 },
		{ 95, 333, 42 },
		{ 95, 334, 42 },
		{ 95, 335, 42 },
		{ 95, 336, 42 },
		{ 95, 337, 42 },
		{ 95, 338, 42 },
		{ 95, 339, 42 },
		{ 95, 340, 42 },
		{ 95, 341, 42 },
		{ 95, 342, 42 },
		{ 95, 343, 42 },
		{ 95, 344, 42 },
		{ 95, 345, 42 },
		{ 95, 346, 42 },
		{ 95, 347, 42 },
		{ 95, 348, 42 },
		{ 95, 349, 42 },
		{ 95, 350, 42 },
		{ 95, 351, 42 },
		{ 95, 352, 42 },
		{ 95, 353, 42 },
		{ 95, 354, 42 },
		{ 95, 355, 42 },
		{ 95, 356, 42 },
		{ 95, 357, 42 },
		{ 95, 358, 42 },
		{ 96, 318, 42 },
		{ 96, 319, 42 },
		{ 96, 320, 42 },
		{ 96, 321, 42 },
		{ 96, 322, 42 },
		{ 96, 323, 42 },
		{ 96, 324, 42 },
		{ 96, 325, 42 },
		{ 96, 326, 42 },
		{ 96, 327, 42 },
		{ 96, 328, 42 },
		{ 96, 329, 42 },
		{ 96, 330, 42 },
		{ 96, 331, 42 },
		{ 96, 332, 42 },
		{ 96, 333, 42 },
		{ 96, 334, 42 },
		{ 96, 335, 42 },
		{ 96, 336, 42 },
		{ 96, 337, 42 },
		{ 96, 338, 42 },
		{ 96, 339, 42 },
		{ 96, 340, 42 },
		{ 96, 341, 42 },
		{ 96, 342, 42 },
		{ 96, 343, 42 },
		{ 96, 344, 42 },
		{ 96, 345, 42 },
		{ 96, 346, 42 },
		{ 96, 347, 42 },
		{ 96, 348, 42 },
		{ 96, 349, 42 },
		{ 96, 350, 42 },
		{ 96, 351, 42 },
		{ 96, 352, 42 },
		{ 96, 353, 42 },
		{ 96, 354, 42 },
		{ 96, 355, 42 },
		{ 96, 356, 42 },
		{ 96, 357, 42 },
		{ 97, 318, 42 },
		{ 97, 319, 42 },
		{ 97, 320, 42 },
		{ 97, 321, 42 },
		{ 97, 322, 42 },
		{ 97, 323, 42 },
		{ 97, 324, 42 },
		{ 97, 325, 42 },
		{ 97, 326, 42 },
		{ 97, 327, 42 },
		{ 97, 328, 42 },
		{ 97, 329, 42 },
		{ 97, 330, 42 },
		{ 97, 331, 42 },
		{ 97, 332, 42 },
		{ 97, 333, 42 },
		{ 97, 334, 42 },
		{ 97, 335, 42 },
		{ 97, 336, 42 },
		{ 97, 337, 42 },
		{ 97, 338, 42 },
		{ 97, 339, 42 },
		{ 97, 340, 42 },
		{ 97, 341, 42 },
		{ 97, 342, 42 },
		{ 97, 343, 42 },
		{ 97, 344, 42 },
		{ 97, 345, 42 },
		{ 97, 346, 42 },
		{ 97, 347, 42 },
		{ 97, 348, 42 },
		{ 97, 349, 42 },
		{ 97, 350, 42 },
		{ 97, 351, 42 },
		{ 97, 352, 42 },
		{ 97, 353, 42 },
		{ 97, 354, 42 },
		{ 97, 355, 42 },
		{ 97, 356, 42 },
		{ 98, 317, 42 },
		{ 98, 318, 42 },
		{ 98, 319, 42 },
		{ 98, 320, 42 },
		{ 98, 321, 42 },
		{ 98, 322, 42 },
		{ 98, 323, 42 },
		{ 98, 324, 42 },
		{ 98, 325, 42 },
		{ 98, 326, 42 },
		{ 98, 327, 42 },
		{ 98, 328, 42 },
		{ 98, 329, 42 },
		{ 98, 330, 42 },
		{ 98, 331, 42 },
		{ 98, 332, 42 },
		{ 98, 333, 42 },
		{ 98, 334, 42 },
		{ 98, 335, 42 },
		{ 98, 336, 42 },
		{ 98, 337, 42 },
		{ 98, 338, 42 },
		{ 98, 339, 42 },
		{ 98, 340, 42 },
		{ 98, 341, 42 },
		{ 98, 342, 42 },
		{ 98, 343, 42 },
		{ 98, 344, 42 },
		{ 98, 345, 42 },
		{ 98, 346, 42 },
		{ 98, 347, 42 },
		{ 98, 348, 42 },
		{ 98, 349, 42 },
		{ 98, 350, 42 },
		{ 98, 351, 42 },
		{ 98, 352, 42 },
		{ 98, 353, 42 },
		{ 98, 354, 42 },
		{ 98, 355, 42 },
		{ 98, 356, 42 },
		{ 99, 317, 42 },
		{ 99, 318, 42 },
		{ 99, 319, 42 },
		{ 99, 320, 42 },
		{ 99, 321, 42 },
		{ 99, 322, 42 },
		{ 99, 323, 42 },
		{ 99, 324, 42 },
		{ 99, 325, 42 },
		{ 99, 326, 42 },
		{ 99, 327, 42 },
		{ 99, 328, 42 },
		{ 99, 329, 42 },
		{ 99, 330, 42 },
		{ 99, 331, 42 },
		{ 99, 332, 42 },
		{ 99, 333, 42 },
		{ 99, 334, 42 },
		{ 99, 335, 42 },
		{ 99, 336, 42 },
		{ 99, 337, 42 },
		{ 99, 338, 42 },
		{ 99, 339, 42 },
		{ 99, 340, 42 },
		{ 99, 341, 42 },
		{ 99, 342, 42 },
		{ 99, 343, 42 },
		{ 99, 344, 42 },
		{ 99, 345, 42 },
		{ 99, 346, 42 },
		{ 99, 347, 42 },
		{ 99, 348, 42 },
		{ 99, 349, 42 },
		{ 99, 350, 42 },
		{ 99, 351, 42 },
		{ 99, 352, 42 },
		{ 99, 353, 42 },
		{ 99, 354, 42 },
		{ 99, 355, 42 },
		{ 100, 316, 42 },
		{ 100, 317, 42 },
		{ 100, 318, 42 },
		{ 100, 319, 42 },
		{ 100, 320, 42 },
		{ 100, 321, 42 },
		{ 100, 322, 42 },
		{ 100, 323, 42 },
		{ 100, 324, 42 },
		{ 100, 325, 42 },
		{ 100, 326, 42 },
		{ 100, 327, 42 },
		{ 100, 328, 42 },
		{ 100, 329, 42 },
		{ 100, 330, 42 },
		{ 100, 331, 42 },
		{ 100, 332, 42 },
		{ 100, 333, 42 },
		{ 100, 334, 42 },
		{ 100, 335, 42 },
		{ 100, 336, 42 },
		{ 100, 337, 42 },
		{ 100, 338, 42 },
		{ 100, 339, 42 },
		{ 100, 340, 42 },
		{ 100, 341, 42 },
		{ 100, 342, 42 },
		{ 100, 343, 42 },
		{ 100, 344, 42 },
		{ 100, 345, 42 },
		{ 100, 346, 42 },
		{ 100, 347, 42 },
		{ 100, 348, 42 },
		{ 100, 349, 42 },
		{ 100, 350, 42 },
		{ 100, 351, 42 },
		{ 100, 352, 42 },
		{ 100, 353, 42 },
		{ 100, 354, 42 },
		{ 100, 355, 42 },
		{ 101, 316, 42 },
		{ 101, 317, 42 },
		{ 101, 318, 42 },
		{ 101, 319, 42 },
		{ 101, 320, 42 },
		{ 101, 321, 42 },
		{ 101, 322, 42 },
		{ 101, 323, 42 },
		{ 101, 324, 42 },
		{ 101, 325, 42 },
		{ 101, 326, 42 },
		{ 101, 327, 42 },
		{ 101, 328, 42 },
		{ 101, 329, 42 },
		{ 101, 330, 42 },
		{ 101, 331, 42 },
		{ 101, 332, 42 },
		{ 101, 333, 42 },
		{ 101, 334, 42 },
		{ 101, 335, 42 },
		{ 101, 336, 42 },
		{ 101, 337, 42 },
		{ 101, 338, 42 },
		{ 101, 339, 42 },
		{ 101, 340, 42 },
		{ 101, 341, 42 },
		{ 101, 342, 42 },
		{ 101, 343, 42 },
		{ 101, 344, 42 },
		{ 101, 345, 42 },
		{ 101, 346, 42 },
		{ 101, 347, 42 },
		{ 101, 348, 42 },
		{ 101, 349, 42 },
		{ 101, 350, 42 },
		{ 101, 351, 42 },
		{ 101, 352, 42 },
		{ 101, 353, 42 },
		{ 101, 354, 42 },
		{ 102, 316, 42 },
		{ 102, 317, 42 },
		{ 102, 318, 42 },
		{ 102, 319, 42 },
		{ 102, 320, 42 },
		{ 102, 321, 42 },
		{ 102, 322, 42 },
		{ 102, 323, 42 },
		{ 102, 324, 42 },
		{ 102, 325, 42 },
		{ 102, 326, 42 },
		{ 102, 327, 42 },
		{ 102, 328, 42 },
		{ 102, 329, 42 },
		{ 102, 330, 42 },
		{ 102, 331, 42 },
		{ 102, 332, 42 },
		{ 102, 333, 42 },
		{ 102, 334, 42 },
		{ 102, 335, 42 },
		{ 102, 336, 42 },
		{ 102, 337, 42 },
		{ 102, 338, 42 },
		{ 102, 339, 42 },
		{ 102, 340, 42 },
		{ 102, 341, 42 },
		{ 102, 342, 42 },
		{ 102, 343, 42 },
		{ 102, 344, 42 },
		{ 102, 345, 42 },
		{ 102, 346, 42 },
		{ 102, 347, 42 },
		{ 102, 348, 42 },
		{ 102, 349, 42 },
		{ 102, 350, 42 },
		{ 102, 351, 42 },
		{ 102, 352, 42 },
		{ 102, 353, 42 },
		{ 103, 315, 42 },
		{ 103, 316, 42 },
		{ 103, 317, 42 },
		{ 103, 318, 42 },
		{ 103, 319, 42 },
		{ 103, 320, 42 },
		{ 103, 321, 42 },
		{ 103, 322, 42 },
		{ 103, 323, 42 },
		{ 103, 324, 42 },
		{ 103, 325, 42 },
		{ 103, 326, 42 },
		{ 103, 327, 42 },
		{ 103, 328, 42 },
		{ 103, 329, 42 },
		{ 103, 330, 42 },
		{ 103, 331, 42 },
		{ 103, 332, 42 },
		{ 103, 333, 42 },
		{ 103, 334, 42 },
		{ 103, 335, 42 },
		{ 103, 336, 42 },
		{ 103, 337, 42 },
		{ 103, 338, 42 },
		{ 103, 339, 42 },
		{ 103, 340, 42 },
		{ 103, 341, 42 },
		{ 103, 342, 42 },
		{ 103, 343, 42 },
		{ 103, 344, 42 },
		{ 103, 345, 42 },
		{ 103, 346, 42 },
		{ 103, 347, 42 },
		{ 103, 348, 42 },
		{ 103, 349, 42 },
		{ 103, 350, 42 },
		{ 103, 351, 42 },
		{ 103, 352, 42 },
		{ 103, 353, 42 },
		{ 104, 315, 42 },
		{ 104, 316, 42 },
		{ 104, 317, 42 },
		{ 104, 318, 42 },
		{ 104, 319, 42 },
		{ 104, 320, 42 },
		{ 104, 321, 42 },
		{ 104, 322, 42 },
		{ 104, 323, 42 },
		{ 104, 324, 42 },
		{ 104, 325, 42 },
		{ 104, 326, 42 },
		{ 104, 327, 42 },
		{ 104, 328, 42 },
		{ 104, 329, 42 },
		{ 104, 330, 42 },
		{ 104, 331, 42 },
		{ 104, 332, 42 },
		{ 104, 333, 42 },
		{ 104, 334, 42 },
		{ 104, 335, 42 },
		{ 104, 336, 42 },
		{ 104, 337, 42 },
		{ 104, 338, 42 },
		{ 104, 339, 42 },
		{ 104, 340, 42 },
		{ 104, 341, 42 },
		{ 104, 342, 42 },
		{ 104, 343, 42 },
		{ 104, 344, 42 },
		{ 104, 345, 42 },
		{ 104, 346, 42 },
		{ 104, 347, 42 },
		{ 104, 348, 42 },
		{ 104, 349, 42 },
		{ 104, 350, 42 },
		{ 104, 351, 42 },
		{ 104, 352, 42 },
		{ 105, 315, 42 },
		{ 105, 316, 42 },
		{ 105, 317, 42 },
		{ 105, 318, 42 },
		{ 105, 319, 42 },
		{ 105, 320, 42 },
		{ 105, 321, 42 },
		{ 105, 322, 42 },
		{ 105, 323, 42 },
		{ 105, 324, 42 },
		{ 105, 325, 42 },
		{ 105, 326, 42 },
		{ 105, 327, 42 },
		{ 105, 328, 42 },
		{ 105, 329, 42 },
		{ 105, 330, 42 },
		{ 105, 331, 42 },
		{ 105, 332, 42 },
		{ 105, 333, 42 },
		{ 105, 334, 42 },
		{ 105, 335, 42 },
		{ 105, 336, 42 },
		{ 105, 337, 42 },
		{ 105, 338, 42 },
		{ 105, 339, 42 },
		{ 105, 340, 42 },
		{ 105, 341, 42 },
		{ 105, 342, 42 },
		{ 105, 343, 42 },
		{ 105, 344, 42 },
		{ 105, 345, 42 },
		{ 105, 346, 42 },
		{ 105, 347, 42 },
		{ 105, 348, 42 },
		{ 105, 349, 42 },
		{ 105, 350, 42 },
		{ 105, 351, 42 },
		{ 106, 314, 42 },
		{ 106, 315, 42 },
		{ 106, 316, 42 },
		{ 106, 317, 42 },
		{ 106, 318, 42 },
		{ 106, 319, 42 },
		{ 106, 320, 42 },
		{ 106, 321, 42 },
		{ 106, 322, 42 },
		{ 106, 323, 42 },
		{ 106, 324, 42 },
		{ 106, 325, 42 },
		{ 106, 326, 42 },
		{ 106, 327, 42 },
		{ 106, 328, 42 },
		{ 106, 329, 42 },
		{ 106, 330, 42 },
		{ 106, 331, 42 },
		{ 106, 332, 42 },
		{ 106, 333, 42 },
		{ 106, 334, 42 },
		{ 106, 335, 42 },
		{ 106, 336, 42 },
		{ 106, 337, 42 },
		{ 106, 338, 42 },
		{ 106, 339, 42 },
		{ 106, 340, 42 },
		{ 106, 341, 42 },
		{ 106, 342, 42 },
		{ 106, 343, 42 },
		{ 106, 344, 42 },
		{ 106, 345, 42 },
		{ 106, 346, 42 },
		{ 106, 347, 42 },
		{ 106, 348, 42 },
		{ 106, 349, 42 },
		{ 106, 350, 42 },
		{ 106, 351, 42 },
		{ 107, 314, 42 },
		{ 107, 315, 42 },
		{ 107, 316, 42 },
		{ 107, 317, 42 },
		{ 107, 318, 42 },
		{ 107, 319, 42 },
		{ 107, 320, 42 },
		{ 107, 321, 42 },
		{ 107, 322, 42 },
		{ 107, 323, 42 },
		{ 107, 324, 42 },
		{ 107, 325, 42 },
		{ 107, 326, 42 },
		{ 107, 327, 42 },
		{ 107, 328, 42 },
		{ 107, 329, 42 },
		{ 107, 330, 42 },
		{ 107, 331, 42 },
		{ 107, 332, 42 },
		{ 107, 333, 42 },
		{ 107, 334, 42 },
		{ 107, 335, 42 },
		{ 107, 336, 42 },
		{ 107, 337, 42 },
		{ 107, 338, 42 },
		{ 107, 339, 42 },
		{ 107, 340, 42 },
		{ 107, 341, 42 },
		{ 107, 342, 42 },
		{ 107, 343, 42 },
		{ 107, 344, 42 },
		{ 107, 345, 42 },
		{ 107, 346, 42 },
		{ 107, 347, 42 },
		{ 107, 348, 42 },
		{ 107, 349, 42 },
		{ 107, 350, 42 },
		{ 108, 313, 42 },
		{ 108, 314, 42 },
		{ 108, 315, 42 },
		{ 108, 316, 42 },
		{ 108, 317, 42 },
		{ 108, 318, 42 },
		{ 108, 319, 42 },
		{ 108, 320, 42 },
		{ 108, 321, 42 },
		{ 108, 322, 42 },
		{ 108, 323, 42 },
		{ 108, 324, 42 },
		{ 108, 325, 42 },
		{ 108, 326, 42 },
		{ 108, 327, 42 },
		{ 108, 328, 42 },
		{ 108, 329, 42 },
		{ 108, 330, 42 },
		{ 108, 331, 42 },
		{ 108, 332, 42 },
		{ 108, 333, 42 },
		{ 108, 334, 42 },
		{ 108, 335, 42 },
		{ 108, 336, 42 },
		{ 108, 337, 42 },
		{ 108, 338, 42 },
		{ 108, 339, 42 },
		{ 108, 340, 42 },
		{ 108, 341, 42 },
		{ 108, 342, 42 },
		{ 108, 343, 42 },
		{ 108, 344, 42 },
		{ 108, 345, 42 },
		{ 108, 346, 42 },
		{ 108, 347, 42 },
		{ 108, 348, 42 },
		{ 108, 349, 42 },
		{ 108, 350, 42 },
		{ 109, 313, 42 },
		{ 109, 314, 42 },
		{ 109, 315, 42 },
		{ 109, 316, 42 },
		{ 109, 317, 42 },
		{ 109, 318, 42 },
		{ 109, 319, 42 },
		{ 109, 320, 42 },
		{ 109, 321, 42 },
		{ 109, 322, 42 },
		{ 109, 323, 42 },
		{ 109, 324, 42 },
		{ 109, 325, 42 },
		{ 109, 326, 42 },
		{ 109, 327, 42 },
		{ 109, 328, 42 },
		{ 109, 329, 42 },
		{ 109, 330, 42 },
		{ 109, 331, 42 },
		{ 109, 332, 42 },
		{ 109, 333, 42 },
		{ 109, 334, 42 },
		{ 109, 335, 42 },
		{ 109, 336, 42 },
		{ 109, 337, 42 },
		{ 109, 338, 42 },
		{ 109, 339, 42 },
		{ 109, 340, 42 },
		{ 109, 341, 42 },
		{ 109, 342, 42 },
		{ 109, 343, 42 },
		{ 109, 344, 42 },
		{ 109, 345, 42 },
		{ 109, 346, 42 },
		{ 109, 347, 42 },
		{ 109, 348, 42 },
		{ 109, 349, 42 },
		{ 110, 313, 42 },
		{ 110, 314, 42 },
		{ 110, 315, 42 },
		{ 110, 316, 42 },
		{ 110, 317, 42 },
		{ 110, 318, 42 },
		{ 110, 319, 42 },
		{ 110, 320, 42 },
		{ 110, 321, 42 },
		{ 110, 322, 42 },
		{ 110, 323, 42 },
		{ 110, 324, 42 },
		{ 110, 325, 42 },
		{ 110, 326, 42 },
		{ 110, 327, 42 },
		{ 110, 328, 42 },
		{ 110, 329, 42 },
		{ 110, 330, 42 },
		{ 110, 331, 42 },
		{ 110, 332, 42 },
		{ 110, 333, 42 },
		{ 110, 334, 42 },
		{ 110, 335, 42 },
		{ 110, 336, 42 },
		{ 110, 337, 42 },
		{ 110, 338, 42 },
		{ 110, 339, 42 },
		{ 110, 340, 42 },
		{ 110, 341, 42 },
		{ 110, 342, 42 },
		{ 110, 343, 42 },
		{ 110, 344, 42 },
		{ 110, 345, 42 },
		{ 110, 346, 42 },
		{ 110, 347, 42 },
		{ 110, 348, 42 },
		{ 111, 312, 42 },
		{ 111, 313, 42 },
		{ 111, 314, 42 },
		{ 111, 315, 42 },
		{ 111, 316, 42 },
		{ 111, 317, 42 },
		{ 111, 318, 42 },
		{ 111, 319, 42 },
		{ 111, 320, 42 },
		{ 111, 321, 42 },
		{ 111, 322, 42 },
		{ 111, 323, 42 },
		{ 111, 324, 42 },
		{ 111, 325, 42 },
		{ 111, 326, 42 },
		{ 111, 327, 42 },
		{ 111, 328, 42 },
		{ 111, 329, 42 },
		{ 111, 330, 42 },
		{ 111, 331, 42 },
		{ 111, 332, 42 },
		{ 111, 333, 42 },
		{ 111, 334, 42 },
		{ 111, 335, 42 },
		{ 111, 336, 42 },
		{ 111, 337, 42 },
		{ 111, 338, 42 },
		{ 111, 339, 42 },
		{ 111, 340, 42 },
		{ 111, 341, 42 },
		{ 111, 342, 42 },
		{ 111, 343, 42 },
		{ 111, 344, 42 },
		{ 111, 345, 42 },
		{ 111, 346, 42 },
		{ 111, 347, 42 },
		{ 111, 348, 42 },
		{ 112, 312, 42 },
		{ 112, 313, 42 },
		{ 112, 314, 42 },
		{ 112, 315, 42 },
		{ 112, 316, 42 },
		{ 112, 317, 42 },
		{ 112, 318, 42 },
		{ 112, 319, 42 },
		{ 112, 320, 42 },
		{ 112, 321, 42 },
		{ 112, 322, 42 },
		{ 112, 323, 42 },
		{ 112, 324, 42 },
		{ 112, 325, 42 },
		{ 112, 326, 42 },
		{ 112, 327, 42 },
		{ 112, 328, 42 },
		{ 112, 329, 42 },
		{ 112, 330, 42 },
		{ 112, 331, 42 },
		{ 112, 332, 42 },
		{ 112, 333, 42 },
		{ 112, 334, 42 },
		{ 112, 335, 42 },
		{ 112, 336, 42 },
		{ 112, 337, 42 },
		{ 112, 338, 42 },
		{ 112, 339, 42 },
		{ 112, 340, 42 },
		{ 112, 341, 42 },
		{ 112, 342, 42 },
		{ 112, 343, 42 },
		{ 112, 344, 42 },
		{ 112, 345, 42 },
		{ 112, 346, 42 },
		{ 112, 347, 42 },
		{ 113, 312, 42 },
		{ 113, 313, 42 },
		{ 113, 314, 42 },
		{ 113, 315, 42 },
		{ 113, 316, 42 },
		{ 113, 317, 42 },
		{ 113, 318, 42 },
		{ 113, 319, 42 },
		{ 113, 320, 42 },
		{ 113, 321, 42 },
		{ 113, 322, 42 },
		{ 113, 323, 42 },
		{ 113, 324, 42 },
		{ 113, 325, 42 },
		{ 113, 326, 42 },
		{ 113, 327, 42 },
		{ 113, 328, 42 },
		{ 113, 329, 42 },
		{ 113, 330, 42 },
		{ 113, 331, 42 },
		{ 113, 332, 42 },
		{ 113, 333, 42 },
		{ 113, 334, 42 },
		{ 113, 335, 42 },
		{ 113, 336, 42 },
		{ 113, 337, 42 },
		{ 113, 338, 42 },
		{ 113, 339, 42 },
		{ 113, 340, 42 },
		{ 113, 341, 42 },
		{ 113, 342, 42 },
		{ 113, 343, 42 },
		{ 113, 344, 42 },
		{ 113, 345, 42 },
		{ 113, 346, 42 },
		{ 114, 311, 42 },
		{ 114, 312, 42 },
		{ 114, 313, 42 },
		{ 114, 314, 42 },
		{ 114, 315, 42 },
		{ 114, 316, 42 },
		{ 114, 317, 42 },
		{ 114, 318, 42 },
		{ 114, 319, 42 },
		{ 114, 320, 42 },
		{ 114, 321, 42 },
		{ 114, 322, 42 },
		{ 114, 323, 42 },
		{ 114, 324, 42 },
		{ 114, 325, 42 },
		{ 114, 326, 42 },
		{ 114, 327, 42 },
		{ 114, 328, 42 },
		{ 114, 329, 42 },
		{ 114, 330, 42 },
		{ 114, 331, 42 },
		{ 114, 332, 42 },
		{ 114, 333, 42 },
		{ 114, 334, 42 },
		{ 114, 335, 42 },
		{ 114, 336, 42 },
		{ 114, 337, 42 },
		{ 114, 338, 42 },
		{ 114, 339, 42 },
		{ 114, 340, 42 },
		{ 114, 341, 42 },
		{ 114, 342, 42 },
		{ 114, 343, 42 },
		{ 114, 344, 42 },
		{ 114, 345, 42 },
		{ 114, 346, 42 },
		{ 115, 311, 42 },
		{ 115, 312, 42 },
		{ 115, 313, 42 },
		{ 115, 314, 42 },
		{ 115, 315, 42 },
		{ 115, 316, 42 },
		{ 115, 317, 42 },
		{ 115, 318, 42 },
		{ 115, 319, 42 },
		{ 115, 320, 42 },
		{ 115, 321, 42 },
		{ 115, 322, 42 },
		{ 115, 323, 42 },
		{ 115, 324, 42 },
		{ 115, 325, 42 },
		{ 115, 326, 42 },
		{ 115, 327, 42 },
		{ 115, 328, 42 },
		{ 115, 329, 42 },
		{ 115, 330, 42 },
		{ 115, 331, 42 },
		{ 115, 332, 42 },
		{ 115, 333, 42 },
		{ 115, 334, 42 },
		{ 115, 335, 42 },
		{ 115, 336, 42 },
		{ 115, 337, 42 },
		{ 115, 338, 42 },
		{ 115, 339, 42 },
		{ 115, 340, 42 },
		{ 115, 341, 42 },
		{ 115, 342, 42 },
		{ 115, 343, 42 },
		{ 115, 344, 42 },
		{ 115, 345, 42 },
		{ 116, 310, 42 },
		{ 116, 311, 42 },
		{ 116, 312, 42 },
		{ 116, 313, 42 },
		{ 116, 314, 42 },
		{ 116, 315, 42 },
		{ 116, 316, 42 },
		{ 116, 317, 42 },
		{ 116, 318, 42 },
		{ 116, 319, 42 },
		{ 116, 320, 42 },
		{ 116, 321, 42 },
		{ 116, 322, 42 },
		{ 116, 323, 42 },
		{ 116, 324, 42 },
		{ 116, 325, 42 },
		{ 116, 326, 42 },
		{ 116, 327, 42 },
		{ 116, 328, 42 },
		{ 116, 329, 42 },
		{ 116, 330, 42 },
		{ 116, 331, 42 },
		{ 116, 332, 42 },
		{ 116, 333, 42 },
		{ 116, 334, 42 },
		{ 116, 335, 42 },
		{ 116, 336, 42 },
		{ 116, 337, 42 },
		{ 116, 338, 42 },
		{ 116, 339, 42 },
		{ 116, 340, 42 },
		{ 116, 341, 42 },
		{ 116, 342, 42 },
		{ 116, 343, 42 },
		{ 116, 344, 42 },
		{ 116, 345, 42 },
		{ 117, 310, 42 },
		{ 117, 311, 42 },
		{ 117, 312, 42 },
		{ 117, 313, 42 },
		{ 117, 314, 42 },
		{ 117, 315, 42 },
		{ 117, 316, 42 },
		{ 117, 317, 42 },
		{ 117, 318, 42 },
		{ 117, 319, 42 },
		{ 117, 320, 42 },
		{ 117, 321, 42 },
		{ 117, 322, 42 },
		{ 117, 323, 42 },
		{ 117, 324, 42 },
		{ 117, 325, 42 },
		{ 117, 326, 42 },
		{ 117, 327, 42 },
		{ 117, 328, 42 },
		{ 117, 329, 42 },
		{ 117, 330, 42 },
		{ 117, 331, 42 },
		{ 117, 332, 42 },
		{ 117, 333, 42 },
		{ 117, 334, 42 },
		{ 117, 335, 42 },
		{ 117, 336, 42 },
		{ 117, 337, 42 },
		{ 117, 338, 42 },
		{ 117, 339, 42 },
		{ 117, 340, 42 },
		{ 117, 341, 42 },
		{ 117, 342, 42 },
		{ 117, 343, 42 },
		{ 117, 344, 42 },
		{ 118, 310, 42 },
		{ 118, 311, 42 },
		{ 118, 312, 42 },
		{ 118, 313, 42 },
		{ 118, 314, 42 },
		{ 118, 315, 42 },
		{ 118, 316, 42 },
		{ 118, 317, 42 },
		{ 118, 318, 42 },
		{ 118, 319, 42 },
		{ 118, 320, 42 },
		{ 118, 321, 42 },
		{ 118, 322, 42 },
		{ 118, 323, 42 },
		{ 118, 324, 42 },
		{ 118, 325, 42 },
		{ 118, 326, 42 },
		{ 118, 327, 42 },
		{ 118, 328, 42 },
		{ 118, 329, 42 },
		{ 118, 330, 42 },
		{ 118, 331, 42 },
		{ 118, 332, 42 },
		{ 118, 333, 42 },
		{ 118, 334, 42 },
		{ 118, 335, 42 },
		{ 118, 336, 42 },
		{ 118, 337, 42 },
		{ 118, 338, 42 },
		{ 118, 339, 42 },
		{ 118, 340, 42 },
		{ 118, 341, 42 },
		{ 118, 342, 42 },
		{ 118, 343, 42 },
		{ 119, 309, 42 },
		{ 119, 310, 42 },
		{ 119, 311, 42 },
		{ 119, 312, 42 },
		{ 119, 313, 42 },
		{ 119, 314, 42 },
		{ 119, 315, 42 },
		{ 119, 316, 42 },
		{ 119, 317, 42 },
		{ 119, 318, 42 },
		{ 119, 319, 42 },
		{ 119, 320, 42 },
		{ 119, 321, 42 },
		{ 119, 322, 42 },
		{ 119, 323, 42 },
		{ 119, 324, 42 },
		{ 119, 325, 42 },
		{ 119, 326, 42 },
		{ 119, 327, 42 },
		{ 119, 328, 42 },
		{ 119, 329, 42 },
		{ 119, 330, 42 },
		{ 119, 331, 42 },
		{ 119, 332, 42 },
		{ 119, 333, 42 },
		{ 119, 334, 42 },
		{ 119, 335, 42 },
		{ 119, 336, 42 },
		{ 119, 337, 42 },
		{ 119, 338, 42 },
		{ 119, 339, 42 },
		{ 119, 340, 42 },
		{ 119, 341, 42 },
		{ 119, 342, 42 },
		{ 119, 343, 42 },
		{ 120, 309, 42 },
		{ 120, 310, 42 },
		{ 120, 311, 42 },
		{ 120, 312, 42 },
		{ 120, 313, 42 },
		{ 120, 314, 42 },
		{ 120, 315, 42 },
		{ 120, 316, 42 },
		{ 120, 317, 42 },
		{ 120, 318, 42 },
		{ 120, 319, 42 },
		{ 120, 320, 42 },
		{ 120, 321, 42 },
		{ 120, 322, 42 },
		{ 120, 323, 42 },
		{ 120, 324, 42 },
		{ 120, 325, 42 },
		{ 120, 326, 42 },
		{ 120, 327, 42 },
		{ 120, 328, 42 },
		{ 120, 329, 42 },
		{ 120, 330, 42 },
		{ 120, 331, 42 },
		{ 120, 332, 42 },
		{ 120, 333, 42 },
		{ 120, 334, 42 },
		{ 120, 335, 42 },
		{ 120, 336, 42 },
		{ 120, 337, 42 },
		{ 120, 338, 42 },
		{ 120, 339, 42 },
		{ 120, 340, 42 },
		{ 120, 341, 42 },
		{ 120, 342, 42 },
		{ 121, 309, 42 },
		{ 121, 310, 42 },
		{ 121, 311, 42 },
		{ 121, 312, 42 },
		{ 121, 313, 42 },
		{ 121, 314, 42 },
		{ 121, 315, 42 },
		{ 121, 316, 42 },
		{ 121, 317, 42 },
		{ 121, 318, 42 },
		{ 121, 319, 42 },
		{ 121, 320, 42 },
		{ 121, 321, 42 },
		{ 121, 322, 42 },
		{ 121, 323, 42 },
		{ 121, 324, 42 },
		{ 121, 325, 42 },
		{ 121, 326, 42 },
		{ 121, 327, 42 },
		{ 121, 328, 42 },
		{ 121, 329, 42 },
		{ 121, 330, 42 },
		{ 121, 331, 42 },
		{ 121, 332, 42 },
		{ 121, 333, 42 },
		{ 121, 334, 42 },
		{ 121, 335, 42 },
		{ 121, 336, 42 },
		{ 121, 337, 42 },
		{ 121, 338, 42 },
		{ 121, 339, 42 },
		{ 121, 340, 42 },
		{ 121, 341, 42 },
		{ 122, 308, 42 },
		{ 122, 309, 42 },
		{ 122, 310, 42 },
		{ 122, 311, 42 },
		{ 122, 312, 42 },
		{ 122, 313, 42 },
		{ 122, 314, 42 },
		{ 122, 315, 42 },
		{ 122, 316, 42 },
		{ 122, 317, 42 },
		{ 122, 318, 42 },
		{ 122, 319, 42 },
		{ 122, 320, 42 },
		{ 122, 321, 42 },
		{ 122, 322, 42 },
		{ 122, 323, 42 },
		{ 122, 324, 42 },
		{ 122, 325, 42 },
		{ 122, 326, 42 },
		{ 122, 327, 42 },
		{ 122, 328, 42 },
		{ 122, 329, 42 },
		{ 122, 330, 42 },
		{ 122, 331, 42 },
		{ 122, 332, 42 },
		{ 122, 333, 42 },
		{ 122, 334, 42 },
		{ 122, 335, 42 },
		{ 122, 336, 42 },
		{ 122, 337, 42 },
		{ 122, 338, 42 },
		{ 122, 339, 42 },
		{ 122, 340, 42 },
		{ 122, 341, 42 },
		{ 123, 308, 42 },
		{ 123, 309, 42 },
		{ 123, 310, 42 },
		{ 123, 311, 42 },
		{ 123, 312, 42 },
		{ 123, 313, 42 },
		{ 123, 314, 42 },
		{ 123, 315, 42 },
		{ 123, 316, 42 },
		{ 123, 317, 42 },
		{ 123, 318, 42 },
		{ 123, 319, 42 },
		{ 123, 320, 42 },
		{ 123, 321, 42 },
		{ 123, 322, 42 },
		{ 123, 323, 42 },
		{ 123, 324, 42 },
		{ 123, 325, 42 },
		{ 123, 326, 42 },
		{ 123, 327, 42 },
		{ 123, 328, 42 },
		{ 123, 329, 42 },
		{ 123, 330, 42 },
		{ 123, 331, 42 },
		{ 123, 332, 42 },
		{ 123, 333, 42 },
		{ 123, 334, 42 },
		{ 123, 335, 42 },
		{ 123, 336, 42 },
		{ 123, 337, 42 },
		{ 123, 338, 42 },
		{ 123, 339, 42 },
		{ 123, 340, 42 },
		{ 124, 307, 42 },
		{ 124, 308, 42 },
		{ 124, 309, 42 },
		{ 124, 310, 42 },
		{ 124, 311, 42 },
		{ 124, 312, 42 },
		{ 124, 313, 42 },
		{ 124, 314, 42 },
		{ 124, 315, 42 },
		{ 124, 316, 42 },
		{ 124, 317, 42 },
		{ 124, 318, 42 },
		{ 124, 319, 42 },
		{ 124, 320, 42 },
		{ 124, 321, 42 },
		{ 124, 322, 42 },
		{ 124, 323, 42 },
		{ 124, 324, 42 },
		{ 124, 325, 42 },
		{ 124, 326, 42 },
		{ 124, 327, 42 },
		{ 124, 328, 42 },
		{ 124, 329, 42 },
		{ 124, 330, 42 },
		{ 124, 331, 42 },
		{ 124, 332, 42 },
		{ 124, 333, 42 },
		{ 124, 334, 42 },
		{ 124, 335, 42 },
		{ 124, 336, 42 },
		{ 124, 337, 42 },
		{ 124, 338, 42 },
		{ 124, 339, 42 },
		{ 124, 340, 42 },
		{ 125, 307, 42 },
		{ 125, 308, 42 },
		{ 125, 309, 42 },
		{ 125, 310, 42 },
		{ 125, 311, 42 },
		{ 125, 312, 42 },
		{ 125, 313, 42 },
		{ 125, 314, 42 },
		{ 125, 315, 42 },
		{ 125, 316, 42 },
		{ 125, 317, 42 },
		{ 125, 318, 42 },
		{ 125, 319, 42 },
		{ 125, 320, 42 },
		{ 125, 321, 42 },
		{ 125, 322, 42 },
		{ 125, 323, 42 },
		{ 125, 324, 42 },
		{ 125, 325, 42 },
		{ 125, 326, 42 },
		{ 125, 327, 42 },
		{ 125, 328, 42 },
		{ 125, 329, 42 },
		{ 125, 330, 42 },
		{ 125, 331, 42 },
		{ 125, 332, 42 },
		{ 125, 333, 42 },
		{ 125, 334, 42 },
		{ 125, 335, 42 },
		{ 125, 336, 42 },
		{ 125, 337, 42 },
		{ 125, 338, 42 },
		{ 125, 339, 42 },
		{ 126, 307, 42 },
		{ 126, 308, 42 },
		{ 126, 309, 42 },
		{ 126, 310, 42 },
		{ 126, 311, 42 },
		{ 126, 312, 42 },
		{ 126, 313, 42 },
		{ 126, 314, 42 },
		{ 126, 315, 42 },
		{ 126, 316, 42 },
		{ 126, 317, 42 },
		{ 126, 318, 42 },
		{ 126, 319, 42 },
		{ 126, 320, 42 },
		{ 126, 321, 42 },
		{ 126, 322, 42 },
		{ 126, 323, 42 },
		{ 126, 324, 42 },
		{ 126, 325, 42 },
		{ 126, 326, 42 },
		{ 126, 327, 42 },
		{ 126, 328, 42 },
		{ 126, 329, 42 },
		{ 126, 330, 42 },
		{ 126, 331, 42 },
		{ 126, 332, 42 },
		{ 126, 333, 42 },
		{ 126, 334, 42 },
		{ 126, 335, 42 },
		{ 126, 336, 42 },
		{ 126, 337, 42 },
		{ 126, 338, 42 },
		{ 127, 306, 42 },
		{ 127, 307, 42 },
		{ 127, 308, 42 },
		{ 127, 309, 42 },
		{ 127, 310, 42 },
		{ 127, 311, 42 },
		{ 127, 312, 42 },
		{ 127, 313, 42 },
		{ 127, 314, 42 },
		{ 127, 315, 42 },
		{ 127, 316, 42 },
		{ 127, 317, 42 },
		{ 127, 318, 42 },
		{ 127, 319, 42 },
		{ 127, 320, 42 },
		{ 127, 321, 42 },
		{ 127, 322, 42 },
		{ 127, 323, 42 },
		{ 127, 324, 42 },
		{ 127, 325, 42 },
		{ 127, 326, 42 },
		{ 127, 327, 42 },
		{ 127, 328, 42 },
		{ 127, 329, 42 },
		{ 127, 330, 42 },
		{ 127, 331, 42 },
		{ 127, 332, 42 },
		{ 127, 333, 42 },
		{ 127, 334, 42 },
		{ 127, 335, 42 },
		{ 127, 336, 42 },
		{ 127, 337, 42 },
		{ 127, 338, 42 },
		{ 128, 306, 42 },
		{ 128, 307, 42 },
		{ 128, 308, 42 },
		{ 128, 309, 42 },
		{ 128, 310, 42 },
		{ 128, 311, 42 },
		{ 128, 312, 42 },
		{ 128, 313, 42 },
		{ 128, 314, 42 },
		{ 128, 315, 42 },
		{ 128, 316, 42 },
		{ 128, 317, 42 },
		{ 128, 318, 42 },
		{ 128, 319, 42 },
		{ 128, 320, 42 },
		{ 128, 321, 42 },
		{ 128, 322, 42 },
		{ 128, 323, 42 },
		{ 128, 324, 42 },
		{ 128, 325, 42 },
		{ 128, 326, 42 },
		{ 128, 327, 42 },
		{ 128, 328, 42 },
		{ 128, 329, 42 },
		{ 128, 330, 42 },
		{ 128, 331, 42 },
		{ 128, 332, 42 },
		{ 128, 333, 42 },
		{ 128, 334, 42 },
		{ 128, 335, 42 },
		{ 128, 336, 42 },
		{ 128, 337, 42 },
		{ 129, 306, 42 },
		{ 129, 307, 42 },
		{ 129, 308, 42 },
		{ 129, 309, 42 },
		{ 129, 310, 42 },
		{ 129, 311, 42 },
		{ 129, 312, 42 },
		{ 129, 313, 42 },
		{ 129, 314, 42 },
		{ 129, 315, 42 },
		{ 129, 316, 42 },
		{ 129, 317, 42 },
		{ 129, 318, 42 },
		{ 129, 319, 42 },
		{ 129, 320, 42 },
		{ 129, 321, 42 },
		{ 129, 322, 42 },
		{ 129, 323, 42 },
		{ 129, 324, 42 },
		{ 129, 325, 42 },
		{ 129, 326, 42 },
		{ 129, 327, 42 },
		{ 129, 328, 42 },
		{ 129, 329, 42 },
		{ 129, 330, 42 },
		{ 129, 331, 42 },
		{ 129, 332, 42 },
		{ 129, 333, 42 },
		{ 129, 334, 42 },
		{ 129, 335, 42 },
		{ 129, 336, 42 },
		{ 130, 305, 42 },
		{ 130, 306, 42 },
		{ 130, 307, 42 },
		{ 130, 308, 42 },
		{ 130, 309, 42 },
		{ 130, 310, 42 },
		{ 130, 311, 42 },
		{ 130, 312, 42 },
		{ 130, 313, 42 },
		{ 130, 314, 42 },
		{ 130, 315, 42 },
		{ 130, 316, 42 },
		{ 130, 317, 42 },
		{ 130, 318, 42 },
		{ 130, 319, 42 },
		{ 130, 320, 42 },
		{ 130, 321, 42 },
		{ 130, 322, 42 },
		{ 130, 323, 42 },
		{ 130, 324, 42 },
		{ 130, 325, 42 },
		{ 130, 326, 42 },
		{ 130, 327, 42 },
		{ 130, 328, 42 },
		{ 130, 329, 42 },
		{ 130, 330, 42 },
		{ 130, 331, 42 },
		{ 130, 332, 42 },
		{ 130, 333, 42 },
		{ 130, 334, 42 },
		{ 130, 335, 42 },
		{ 130, 336, 42 },
		{ 131, 305, 42 },
		{ 131, 306, 42 },
		{ 131, 307, 42 },
		{ 131, 308, 42 },
		{ 131, 309, 42 },
		{ 131, 310, 42 },
		{ 131, 311, 42 },
		{ 131, 312, 42 },
		{ 131, 313, 42 },
		{ 131, 314, 42 },
		{ 131, 315, 42 },
		{ 131, 316, 42 },
		{ 131, 317, 42 },
		{ 131, 318, 42 },
		{ 131, 319, 42 },
		{ 131, 320, 42 },
		{ 131, 321, 42 },
		{ 131, 322, 42 },
		{ 131, 323, 42 },
		{ 131, 324, 42 },
		{ 131, 325, 42 },
		{ 131, 326, 42 },
		{ 131, 327, 42 },
		{ 131, 328, 42 },
		{ 131, 329, 42 },
		{ 131, 330, 42 },
		{ 131, 331, 42 },
		{ 131, 332, 42 },
		{ 131, 333, 42 },
		{ 131, 334, 42 },
		{ 131, 335, 42 },
		{ 132, 304, 42 },
		{ 132, 305, 42 },
		{ 132, 306, 42 },
		{ 132, 307, 42 },
		{ 132, 308, 42 },
		{ 132, 309, 42 },
		{ 132, 310, 42 },
		{ 132, 311, 42 },
		{ 132, 312, 42 },
		{ 132, 313, 42 },
		{ 132, 314, 42 },
		{ 132, 315, 42 },
		{ 132, 316, 42 },
		{ 132, 317, 42 },
		{ 132, 318, 42 },
		{ 132, 319, 42 },
		{ 132, 320, 42 },
		{ 132, 321, 42 },
		{ 132, 322, 42 },
		{ 132, 323, 42 },
		{ 132, 324, 42 },
		{ 132, 325, 42 },
		{ 132, 326, 42 },
		{ 132, 327, 42 },
		{ 132, 328, 42 },
		{ 132, 329, 42 },
		{ 132, 330, 42 },
		{ 132, 331, 42 },
		{ 132, 332, 42 },
		{ 132, 333, 42 },
		{ 132, 334, 42 },
		{ 132, 335, 42 },
		{ 133, 304, 42 },
		{ 133, 305, 42 },
		{ 133, 306, 42 },
		{ 133, 307, 42 },
		{ 133, 308, 42 },
		{ 133, 309, 42 },
		{ 133, 310, 42 },
		{ 133, 311, 42 },
		{ 133, 312, 42 },
		{ 133, 313, 42 },
		{ 133, 314, 42 },
		{ 133, 315, 42 },
		{ 133, 316, 42 },
		{ 133, 317, 42 },
		{ 133, 318, 42 },
		{ 133, 319, 42 },
		{ 133, 320, 42 },
		{ 133, 321, 42 },
		{ 133, 322, 42 },
		{ 133, 323, 42 },
		{ 133, 324, 42 },
		{ 133, 325, 42 },
		{ 133, 326, 42 },
		{ 133, 327, 42 },
		{ 133, 328, 42 },
		{ 133, 329, 42 },
		{ 133, 330, 42 },
		{ 133, 331, 42 },
		{ 133, 332, 42 },
		{ 133, 333, 42 },
		{ 133, 334, 42 },
		{ 134, 304, 42 },
		{ 134, 305, 42 },
		{ 134, 306, 42 },
		{ 134, 307, 42 },
		{ 134, 308, 42 },
		{ 134, 309, 42 },
		{ 134, 310, 42 },
		{ 134, 311, 42 },
		{ 134, 312, 42 },
		{ 134, 313, 42 },
		{ 134, 314, 42 },
		{ 134, 315, 42 },
		{ 134, 316, 42 },
		{ 134, 317, 42 },
		{ 134, 318, 42 },
		{ 134, 319, 42 },
		{ 134, 320, 42 },
		{ 134, 321, 42 },
		{ 134, 322, 42 },
		{ 134, 323, 42 },
		{ 134, 324, 42 },
		{ 134, 325, 42 },
		{ 134, 326, 42 },
		{ 134, 327, 42 },
		{ 134, 328, 42 },
		{ 134, 329, 42 },
		{ 134, 330, 42 },
		{ 134, 331, 42 },
		{ 134, 332, 42 },
		{ 134, 333, 42 },
		{ 135, 303, 42 },
		{ 135, 304, 42 },
		{ 135, 305, 42 },
		{ 135, 306, 42 },
		{ 135, 307, 42 },
		{ 135, 308, 42 },
		{ 135, 309, 42 },
		{ 135, 310, 42 },
		{ 135, 311, 42 },
		{ 135, 312, 42 },
		{ 135, 313, 42 },
		{ 135, 314, 42 },
		{ 135, 315, 42 },
		{ 135, 316, 42 },
		{ 135, 317, 42 },
		{ 135, 318, 42 },
		{ 135, 319, 42 },
		{ 135, 320, 42 },
		{ 135, 321, 42 },
		{ 135, 322, 42 },
		{ 135, 323, 42 },
		{ 135, 324, 42 },
		{ 135, 325, 42 },
		{ 135, 326, 42 },
		{ 135, 327, 42 },
		{ 135, 328, 42 },
		{ 135, 329, 42 },
		{ 135, 330, 42 },
		{ 135, 331, 42 },
		{ 135, 332, 42 },
		{ 135, 333, 42 },
		{ 136, 303, 42 },
		{ 136, 304, 42 },
		{ 136, 305, 42 },
		{ 136, 306, 42 },
		{ 136, 307, 42 },
		{ 136, 308, 42 },
		{ 136, 309, 42 },
		{ 136, 310, 42 },
		{ 136, 311, 42 },
		{ 136, 312, 42 },
		{ 136, 313, 42 },
		{ 136, 314, 42 },
		{ 136, 315, 42 },
		{ 136, 316, 42 },
		{ 136, 317, 42 },
		{ 136, 318, 42 },
		{ 136, 319, 42 },
		{ 136, 320, 42 },
		{ 136, 321, 42 },
		{ 136, 322, 42 },
		{ 136, 323, 42 },
		{ 136, 324, 42 },
		{ 136, 325, 42 },
		{ 136, 326, 42 },
		{ 136, 327, 42 },
		{ 136, 328, 42 },
		{ 136, 329, 42 },
		{ 136, 330, 42 },
		{ 136, 331, 42 },
		{ 136, 332, 42 },
		{ 137, 303, 42 },
		{ 137, 304, 42 },
		{ 137, 305, 42 },
		{ 137, 306, 42 },
		{ 137, 307, 42 },
		{ 137, 308, 42 },
		{ 137, 309, 42 },
		{ 137, 310, 42 },
		{ 137, 311, 42 },
		{ 137, 312, 42 },
		{ 137, 313, 42 },
		{ 137, 314, 42 },
		{ 137, 315, 42 },
		{ 137, 316, 42 },
		{ 137, 317, 42 },
		{ 137, 318, 42 },
		{ 137, 319, 42 },
		{ 137, 320, 42 },
		{ 137, 321, 42 },
		{ 137, 322, 42 },
		{ 137, 323, 42 },
		{ 137, 324, 42 },
		{ 137, 325, 42 },
		{ 137, 326, 42 },
		{ 137, 327, 42 },
		{ 137, 328, 42 },
		{ 137, 329, 42 },
		{ 137, 330, 42 },
		{ 137, 331, 42 },
		{ 138, 302, 42 },
		{ 138, 303, 42 },
		{ 138, 304, 42 },
		{ 138, 305, 42 },
		{ 138, 306, 42 },
		{ 138, 307, 42 },
		{ 138, 308, 42 },
		{ 138, 309, 42 },
		{ 138, 310, 42 },
		{ 138, 311, 42 },
		{ 138, 312, 42 },
		{ 138, 313, 42 },
		{ 138, 314, 42 },
		{ 138, 315, 42 },
		{ 138, 316, 42 },
		{ 138, 317, 42 },
		{ 138, 318, 42 },
		{ 138, 319, 42 },
		{ 138, 320, 42 },
		{ 138, 321, 42 },
		{ 138, 322, 42 },
		{ 138, 323, 42 },
		{ 138, 324, 42 },
		{ 138, 325, 42 },
		{ 138, 326, 42 },
		{ 138, 327, 42 },
		{ 138, 328, 42 },
		{ 138, 329, 42 },
		{ 138, 330, 42 },
		{ 138, 331, 42 },
		{ 139, 302, 42 },
		{ 139, 303, 42 },
		{ 139, 304, 42 },
		{ 139, 305, 42 },
		{ 139, 306, 42 },
		{ 139, 307, 42 },
		{ 139, 308, 42 },
		{ 139, 309, 42 },
		{ 139, 310, 42 },
		{ 139, 311, 42 },
		{ 139, 312, 42 },
		{ 139, 313, 42 },
		{ 139, 314, 42 },
		{ 139, 315, 42 },
		{ 139, 316, 42 },
		{ 139, 317, 42 },
		{ 139, 318, 42 },
		{ 139, 319, 42 },
		{ 139, 320, 42 },
		{ 139, 321, 42 },
		{ 139, 322, 42 },
		{ 139, 323, 42 },
		{ 139, 324, 42 },
		{ 139, 325, 42 },
		{ 139, 326, 42 },
		{ 139, 327, 42 },
		{ 139, 328, 42 },
		{ 139, 329, 42 },
		{ 139, 330, 42 },
		{ 140, 301, 42 },
		{ 140, 302, 42 },
		{ 140, 303, 42 },
		{ 140, 304, 42 },
		{ 140, 305, 42 },
		{ 140, 306, 42 },
		{ 140, 307, 42 },
		{ 140, 308, 42 },
		{ 140, 309, 42 },
		{ 140, 310, 42 },
		{ 140, 311, 42 },
		{ 140, 312, 42 },
		{ 140, 313, 42 },
		{ 140, 314, 42 },
		{ 140, 315, 42 },
		{ 140, 316, 42 },
		{ 140, 317, 42 },
		{ 140, 318, 42 },
		{ 140, 319, 42 },
		{ 140, 320, 42 },
		{ 140, 321, 42 },
		{ 140, 322, 42 },
		{ 140, 323, 42 },
		{ 140, 324, 42 },
		{ 140, 325, 42 },
		{ 140, 326, 42 },
		{ 140, 327, 42 },
		{ 140, 328, 42 },
		{ 140, 329, 42 },
		{ 140, 330, 42 },
		{ 141, 301, 42 },
		{ 141, 302, 42 },
		{ 141, 303, 42 },
		{ 141, 304, 42 },
		{ 141, 305, 42 },
		{ 141, 306, 42 },
		{ 141, 307, 42 },
		{ 141, 308, 42 },
		{ 141, 309, 42 },
		{ 141, 310, 42 },
		{ 141, 311, 42 },
		{ 141, 312, 42 },
		{ 141, 313, 42 },
		{ 141, 314, 42 },
		{ 141, 315, 42 },
		{ 141, 316, 42 },
		{ 141, 317, 42 },
		{ 141, 318, 42 },
		{ 141, 319, 42 },
		{ 141, 320, 42 },
		{ 141, 321, 42 },
		{ 141, 322, 42 },
		{ 141, 323, 42 },
		{ 141, 324, 42 },
		{ 141, 325, 42 },
		{ 141, 326, 42 },
		{ 141, 327, 42 },
		{ 141, 328, 42 },
		{ 141, 329, 42 },
		{ 142, 301, 42 },
		{ 142, 302, 42 },
		{ 142, 303, 42 },
		{ 142, 304, 42 },
		{ 142, 305, 42 },
		{ 142, 306, 42 },
		{ 142, 307, 42 },
		{ 142, 308, 42 },
		{ 142, 309, 42 },
		{ 142, 310, 42 },
		{ 142, 311, 42 },
		{ 142, 312, 42 },
		{ 142, 313, 42 },
		{ 142, 314, 42 },
		{ 142, 315, 42 },
		{ 142, 316, 42 },
		{ 142, 317, 42 },
		{ 142, 318, 42 },
		{ 142, 319, 42 },
		{ 142, 320, 42 },
		{ 142, 321, 42 },
		{ 142, 322, 42 },
		{ 142, 323, 42 },
		{ 142, 324, 42 },
		{ 142, 325, 42 },
		{ 142, 326, 42 },
		{ 142, 327, 42 },
		{ 142, 328, 42 },
		{ 143, 300, 42 },
		{ 143, 301, 42 },
		{ 143, 302, 42 },
		{ 143, 303, 42 },
		{ 143, 304, 42 },
		{ 143, 305, 42 },
		{ 143, 306, 42 },
		{ 143, 307, 42 },
		{ 143, 308, 42 },
		{ 143, 309, 42 },
		{ 143, 310, 42 },
		{ 143, 311, 42 },
		{ 143, 312, 42 },
		{ 143, 313, 42 },
		{ 143, 314, 42 },
		{ 143, 315, 42 },
		{ 143, 316, 42 },
		{ 143, 317, 42 },
		{ 143, 318, 42 },
		{ 143, 319, 42 },
		{ 143, 320, 42 },
		{ 143, 321, 42 },
		{ 143, 322, 42 },
		{ 143, 323, 42 },
		{ 143, 324, 42 },
		{ 143, 325, 42 },
		{ 143, 326, 42 },
		{ 143, 327, 42 },
		{ 143, 328, 42 },
		{ 144, 300, 42 },
		{ 144, 301, 42 },
		{ 144, 302, 42 },
		{ 144, 303, 42 },
		{ 144, 304, 42 },
		{ 144, 305, 42 },
		{ 144, 306, 42 },
		{ 144, 307, 42 },
		{ 144, 308, 42 },
		{ 144, 309, 42 },
		{ 144, 310, 42 },
		{ 144, 311, 42 },
		{ 144, 312, 42 },
		{ 144, 313, 42 },
		{ 144, 314, 42 },
		{ 144, 315, 42 },
		{ 144, 316, 42 },
		{ 144, 317, 42 },
		{ 144, 318, 42 },
		{ 144, 319, 42 },
		{ 144, 320, 42 },
		{ 144, 321, 42 },
		{ 144, 322, 42 },
		{ 144, 323, 42 },
		{ 144, 324, 42 },
		{ 144, 325, 42 },
		{ 144, 326, 42 },
		{ 144, 327, 42 },
		{ 145, 300, 42 },
		{ 145, 301, 42 },
		{ 145, 302, 42 },
		{ 145, 303, 42 },
		{ 145, 304, 42 },
		{ 145, 305, 42 },
		{ 145, 306, 42 },
		{ 145, 307, 42 },
		{ 145, 308, 42 },
		{ 145, 309, 42 },
		{ 145, 310, 42 },
		{ 145, 311, 42 },
		{ 145, 312, 42 },
		{ 145, 313, 42 },
		{ 145, 314, 42 },
		{ 145, 315, 42 },
		{ 145, 316, 42 },
		{ 145, 317, 42 },
		{ 145, 318, 42 },
		{ 145, 319, 42 },
		{ 145, 320, 42 },
		{ 145, 321, 42 },
		{ 145, 322, 42 },
		{ 145, 323, 42 },
		{ 145, 324, 42 },
		{ 145, 325, 42 },
		{ 145, 326, 42 },
		{ 146, 299, 42 },
		{ 146, 300, 42 },
		{ 146, 301, 42 },
		{ 146, 302, 42 },
		{ 146, 303, 42 },
		{ 146, 304, 42 },
		{ 146, 305, 42 },
		{ 146, 306, 42 },
		{ 146, 307, 42 },
		{ 146, 308, 42 },
		{ 146, 309, 42 },
		{ 146, 310, 42 },
		{ 146, 311, 42 },
		{ 146, 312, 42 },
		{ 146, 313, 42 },
		{ 146, 314, 42 },
		{ 146, 315, 42 },
		{ 146, 316, 42 },
		{ 146, 317, 42 },
		{ 146, 318, 42 },
		{ 146, 319, 42 },
		{ 146, 320, 42 },
		{ 146, 321, 42 },
		{ 146, 322, 42 },
		{ 146, 323, 42 },
		{ 146, 324, 42 },
		{ 146, 325, 42 },
		{ 146, 326, 42 },
		{ 147, 299, 42 },
		{ 147, 300, 42 },
		{ 147, 301, 42 },
		{ 147, 302, 42 },
		{ 147, 303, 42 },
		{ 147, 304, 42 },
		{ 147, 305, 42 },
		{ 147, 306, 42 },
		{ 147, 307, 42 },
		{ 147, 308, 42 },
		{ 147, 309, 42 },
		{ 147, 310, 42 },
		{ 147, 311, 42 },
		{ 147, 312, 42 },
		{ 147, 313, 42 },
		{ 147, 314, 42 },
		{ 147, 315, 42 },
		{ 147, 316, 42 },
		{ 147, 317, 42 },
		{ 147, 318, 42 },
		{ 147, 319, 42 },
		{ 147, 320, 42 },
		{ 147, 321, 42 },
		{ 147, 322, 42 },
		{ 147, 323, 42 },
		{ 147, 324, 42 },
		{ 147, 325, 42 },
		{ 148, 298, 42 },
		{ 148, 299, 42 },
		{ 148, 300, 42 },
		{ 148, 301, 42 },
		{ 148, 302, 42 },
		{ 148, 303, 42 },
		{ 148, 304, 42 },
		{ 148, 305, 42 },
		{ 148, 306, 42 },
		{ 148, 307, 42 },
		{ 148, 308, 42 },
		{ 148, 309, 42 },
		{ 148, 310, 42 },
		{ 148, 311, 42 },
		{ 148, 312, 42 },
		{ 148, 313, 42 },
		{ 148, 314, 42 },
		{ 148, 315, 42 },
		{ 148, 316, 42 },
		{ 148, 317, 42 },
		{ 148, 318, 42 },
		{ 148, 319, 42 },
		{ 148, 320, 42 },
		{ 148, 321, 42 },
		{ 148, 322, 42 },
		{ 148, 323, 42 },
		{ 148, 324, 42 },
		{ 148, 325, 42 },
		{ 149, 298, 42 },
		{ 149, 299, 42 },
		{ 149, 300, 42 },
		{ 149, 301, 42 },
		{ 149, 302, 42 },
		{ 149, 303, 42 },
		{ 149, 304, 42 },
		{ 149, 305, 42 },
		{ 149, 306, 42 },
		{ 149, 307, 42 },
		{ 149, 308, 42 },
		{ 149, 309, 42 },
		{ 149, 310, 42 },
		{ 149, 311, 42 },
		{ 149, 312, 42 },
		{ 149, 313, 42 },
		{ 149, 314, 42 },
		{ 149, 315, 42 },
		{ 149, 316, 42 },
		{ 149, 317, 42 },
		{ 149, 318, 42 },
		{ 149, 319, 42 },
		{ 149, 320, 42 },
		{ 149, 321, 42 },
		{ 149, 322, 42 },
		{ 149, 323, 42 },
		{ 149, 324, 42 },
		{ 150, 298, 42 },
		{ 150, 299, 42 },
		{ 150, 300, 42 },
		{ 150, 301, 42 },
		{ 150, 302, 42 },
		{ 150, 303, 42 },
		{ 150, 304, 42 },
		{ 150, 305, 42 },
		{ 150, 306, 42 },
		{ 150, 307, 42 },
		{ 150, 308, 42 },
		{ 150, 309, 42 },
		{ 150, 310, 42 },
		{ 150, 311, 42 },
		{ 150, 312, 42 },
		{ 150, 313, 42 },
		{ 150, 314, 42 },
		{ 150, 315, 42 },
		{ 150, 316, 42 },
		{ 150, 317, 42 },
		{ 150, 318, 42 },
		{ 150, 319, 42 },
		{ 150, 320, 42 },
		{ 150, 321, 42 },
		{ 150, 322, 42 },
		{ 150, 323, 42 },
		{ 151, 297, 42 },
		{ 151, 298, 42 },
		{ 151, 299, 42 },
		{ 151, 300, 42 },
		{ 151, 301, 42 },
		{ 151, 302, 42 },
		{ 151, 303, 42 },
		{ 151, 304, 42 },
		{ 151, 305, 42 },
		{ 151, 306, 42 },
		{ 151, 307, 42 },
		{ 151, 308, 42 },
		{ 151, 309, 42 },
		{ 151, 310, 42 },
		{ 151, 311, 42 },
		{ 151, 312, 42 },
		{ 151, 313, 42 },
		{ 151, 314, 42 },
		{ 151, 315, 42 },
		{ 151, 316, 42 },
		{ 151, 317, 42 },
		{ 151, 318, 42 },
		{ 151, 319, 42 },
		{ 151, 320, 42 },
		{ 151, 321, 42 },
		{ 151, 322, 42 },
		{ 151, 323, 42 },
		{ 152, 297, 42 },
		{ 152, 298, 42 },
		{ 152, 299, 42 },
		{ 152, 300, 42 },
		{ 152, 301, 42 },
		{ 152, 302, 42 },
		{ 152, 303, 42 },
		{ 152, 304, 42 },
		{ 152, 305, 42 },
		{ 152, 306, 42 },
		{ 152, 307, 42 },
		{ 152, 308, 42 },
		{ 152, 309, 42 },
		{ 152, 310, 42 },
		{ 152, 311, 42 },
		{ 152, 312, 42 },
		{ 152, 313, 42 },
		{ 152, 314, 42 },
		{ 152, 315, 42 },
		{ 152, 316, 42 },
		{ 152, 317, 42 },
		{ 152, 318, 42 },
		{ 152, 319, 42 },
		{ 152, 320, 42 },
		{ 152, 321, 42 },
		{ 152, 322, 42 },
		{ 153, 297, 42 },
		{ 153, 298, 42 },
		{ 153, 299, 42 },
		{ 153, 300, 42 },
		{ 153, 301, 42 },
		{ 153, 302, 42 },
		{ 153, 303, 42 },
		{ 153, 304, 42 },
		{ 153, 305, 42 },
		{ 153, 306, 42 },
		{ 153, 307, 42 },
		{ 153, 308, 42 },
		{ 153, 309, 42 },
		{ 153, 310, 42 },
		{ 153, 311, 42 },
		{ 153, 312, 42 },
		{ 153, 313, 42 },
		{ 153, 314, 42 },
		{ 153, 315, 42 },
		{ 153, 316, 42 },
		{ 153, 317, 42 },
		{ 153, 318, 42 },
		{ 153, 319, 42 },
		{ 153, 320, 42 },
		{ 153, 321, 42 },
		{ 154, 296, 42 },
		{ 154, 297, 42 },
		{ 154, 298, 42 },
		{ 154, 299, 42 },
		{ 154, 300, 42 },
		{ 154, 301, 42 },
		{ 154, 302, 42 },
		{ 154, 303, 42 },
		{ 154, 304, 42 },
		{ 154, 305, 42 },
		{ 154, 306, 42 },
		{ 154, 307, 42 },
		{ 154, 308, 42 },
		{ 154, 309, 42 },
		{ 154, 310, 42 },
		{ 154, 311, 42 },
		{ 154, 312, 42 },
		{ 154, 313, 42 },
		{ 154, 314, 42 },
		{ 154, 315, 42 },
		{ 154, 316, 42 },
		{ 154, 317, 42 },
		{ 154, 318, 42 },
		{ 154, 319, 42 },
		{ 154, 320, 42 },
		{ 154, 321, 42 },
		{ 155, 296, 42 },
		{ 155, 297, 42 },
		{ 155, 298, 42 },
		{ 155, 299, 42 },
		{ 155, 300, 42 },
		{ 155, 301, 42 },
		{ 155, 302, 42 },
		{ 155, 303, 42 },
		{ 155, 304, 42 },
		{ 155, 305, 42 },
		{ 155, 306, 42 },
		{ 155, 307, 42 },
		{ 155, 308, 42 },
		{ 155, 309, 42 },
		{ 155, 310, 42 },
		{ 155, 311, 42 },
		{ 155, 312, 42 },
		{ 155, 313, 42 },
		{ 155, 314, 42 },
		{ 155, 315, 42 },
		{ 155, 316, 42 },
		{ 155, 317, 42 },
		{ 155, 318, 42 },
		{ 155, 319, 42 },
		{ 155, 320, 42 },
		{ 156, 295, 42 },
		{ 156, 296, 42 },
		{ 156, 297, 42 },
		{ 156, 298, 42 },
		{ 156, 299, 42 },
		{ 156, 300, 42 },
		{ 156, 301, 42 },
		{ 156, 302, 42 },
		{ 156, 303, 42 },
		{ 156, 304, 42 },
		{ 156, 305, 42 },
		{ 156, 306, 42 },
		{ 156, 307, 42 },
		{ 156, 308, 42 },
		{ 156, 309, 42 },
		{ 156, 310, 42 },
		{ 156, 311, 42 },
		{ 156, 312, 42 },
		{ 156, 313, 42 },
		{ 156, 314, 42 },
		{ 156, 315, 42 },
		{ 156, 316, 42 },
		{ 156, 317, 42 },
		{ 156, 318, 42 },
		{ 156, 319, 42 },
		{ 156, 320, 42 },
		{ 157, 295, 42 },
		{ 157, 296, 42 },
		{ 157, 297, 42 },
		{ 157, 298, 42 },
		{ 157, 299, 42 },
		{ 157, 300, 42 },
		{ 157, 301, 42 },
		{ 157, 302, 42 },
		{ 157, 303, 42 },
		{ 157, 304, 42 },
		{ 157, 305, 42 },
		{ 157, 306, 42 },
		{ 157, 307, 42 },
		{ 157, 308, 42 },
		{ 157, 309, 42 },
		{ 157, 310, 42 },
		{ 157, 311, 42 },
		{ 157, 312, 42 },
		{ 157, 313, 42 },
		{ 157, 314, 42 },
		{ 157, 315, 42 },
		{ 157, 316, 42 },
		{ 157, 317, 42 },
		{ 157, 318, 42 },
		{ 157, 319, 42 },
		{ 158, 295, 42 },
		{ 158, 296, 42 },
		{ 158, 297, 42 },
		{ 158, 298, 42 },
		{ 158, 299, 42 },
		{ 158, 300, 42 },
		{ 158, 301, 42 },
		{ 158, 302, 42 },
		{ 158, 303, 42 },
		{ 158, 304, 42 },
		{ 158, 305, 42 },
		{ 158, 306, 42 },
		{ 158, 307, 42 },
		{ 158, 308, 42 },
		{ 158, 309, 42 },
		{ 158, 310, 42 },
		{ 158, 311, 42 },
		{ 158, 312, 42 },
		{ 158, 313, 42 },
		{ 158, 314, 42 },
		{ 158, 315, 42 },
		{ 158, 316, 42 },
		{ 158, 317, 42 },
		{ 158, 318, 42 },
		{ 159, 294, 42 },
		{ 159, 295, 42 },
		{ 159, 296, 42 },
		{ 159, 297, 42 },
		{ 159, 298, 42 },
		{ 159, 299, 42 },
		{ 159, 300, 42 },
		{ 159, 301, 42 },
		{ 159, 302, 42 },
		{ 159, 303, 42 },
		{ 159, 304, 42 },
		{ 159, 305, 42 },
		{ 159, 306, 42 },
		{ 159, 307, 42 },
		{ 159, 308, 42 },
		{ 159, 309, 42 },
		{ 159, 310, 42 },
		{ 159, 311, 42 },
		{ 159, 312, 42 },
		{ 159, 313, 42 },
		{ 159, 314, 42 },
		{ 159, 315, 42 },
		{ 159, 316, 42 },
		{ 159, 317, 42 },
		{ 159, 318, 42 },
		{ 160, 294, 42 },
		{ 160, 295, 42 },
		{ 160, 296, 42 },
		{ 160, 297, 42 },
		{ 160, 298, 42 },
		{ 160, 299, 42 },
		{ 160, 300, 42 },
		{ 160, 301, 42 },
		{ 160, 302, 42 },
		{ 160, 303, 42 },
		{ 160, 304, 42 },
		{ 160, 305, 42 },
		{ 160, 306, 42 },
		{ 160, 307, 42 },
		{ 160, 308, 42 },
		{ 160, 309, 42 },
		{ 160, 310, 42 },
		{ 160, 311, 42 },
		{ 160, 312, 42 },
		{ 160, 313, 42 },
		{ 160, 314, 42 },
		{ 160, 315, 42 },
		{ 160, 316, 42 },
		{ 160, 317, 42 },
		{ 161, 294, 42 },
		{ 161, 295, 42 },
		{ 161, 296, 42 },
		{ 161, 297, 42 },
		{ 161, 298, 42 },
		{ 161, 299, 42 },
		{ 161, 300, 42 },
		{ 161, 301, 42 },
		{ 161, 302, 42 },
		{ 161, 303, 42 },
		{ 161, 304, 42 },
		{ 161, 305, 42 },
		{ 161, 306, 42 },
		{ 161, 307, 42 },
		{ 161, 308, 42 },
		{ 161, 309, 42 },
		{ 161, 310, 42 },
		{ 161, 311, 42 },
		{ 161, 312, 42 },
		{ 161, 313, 42 },
		{ 161, 314, 42 },
		{ 161, 315, 42 },
		{ 161, 316, 42 },
		{ 162, 293, 42 },
		{ 162, 294, 42 },
		{ 162, 295, 42 },
		{ 162, 296, 42 },
		{ 162, 297, 42 },
		{ 162, 298, 42 },
		{ 162, 299, 42 },
		{ 162, 300, 42 },
		{ 162, 301, 42 },
		{ 162, 302, 42 },
		{ 162, 303, 42 },
		{ 162, 304, 42 },
		{ 162, 305, 42 },
		{ 162, 306, 42 },
		{ 162, 307, 42 },
		{ 162, 308, 42 },
		{ 162, 309, 42 },
		{ 162, 310, 42 },
		{ 162, 311, 42 },
		{ 162, 312, 42 },
		{ 162, 313, 42 },
		{ 162, 314, 42 },
		{ 162, 315, 42 },
		{ 162, 316, 42 },
		{ 163, 293, 42 },
		{ 163, 294, 42 },
		{ 163, 295, 42 },
		{ 163, 296, 42 },
		{ 163, 297, 42 },
		{ 163, 298, 42 },
		{ 163, 299, 42 },
		{ 163, 300, 42 },
		{ 163, 301, 42 },
		{ 163, 302, 42 },
		{ 163, 303, 42 },
		{ 163, 304, 42 },
		{ 163, 305, 42 },
		{ 163, 306, 42 },
		{ 163, 307, 42 },
		{ 163, 308, 42 },
		{ 163, 309, 42 },
		{ 163, 310, 42 },
		{ 163, 311, 42 },
		{ 163, 312, 42 },
		{ 163, 313, 42 },
		{ 163, 314, 42 },
		{ 163, 315, 42 },
		{ 164, 292, 42 },
		{ 164, 293, 42 },
		{ 164, 294, 42 },
		{ 164, 295, 42 },
		{ 164, 296, 42 },
		{ 164, 297, 42 },
		{ 164, 298, 42 },
		{ 164, 299, 42 },
		{ 164, 300, 42 },
		{ 164, 301, 42 },
		{ 164, 302, 42 },
		{ 164, 303, 42 },
		{ 164, 304, 42 },
		{ 164, 305, 42 },
		{ 164, 306, 42 },
		{ 164, 307, 42 },
		{ 164, 308, 42 },
		{ 164, 309, 42 },
		{ 164, 310, 42 },
		{ 164, 311, 42 },
		{ 164, 312, 42 },
		{ 164, 313, 42 },
		{ 164, 314, 42 },
		{ 164, 315, 42 },
		{ 165, 292, 42 },
		{ 165, 293, 42 },
		{ 165, 294, 42 },
		{ 165, 295, 42 },
		{ 165, 296, 42 },
		{ 165, 297, 42 },
		{ 165, 298, 42 },
		{ 165, 299, 42 },
		{ 165, 300, 42 },
		{ 165, 301, 42 },
		{ 165, 302, 42 },
		{ 165, 303, 42 },
		{ 165, 304, 42 },
		{ 165, 305, 42 },
		{ 165, 306, 42 },
		{ 165, 307, 42 },
		{ 165, 308, 42 },
		{ 165, 309, 42 },
		{ 165, 310, 42 },
		{ 165, 311, 42 },
		{ 165, 312, 42 },
		{ 165, 313, 42 },
		{ 165, 314, 42 },
		{ 166, 292, 42 },
		{ 166, 293, 42 },
		{ 166, 294, 42 },
		{ 166, 295, 42 },
		{ 166, 296, 42 },
		{ 166, 297, 42 },
		{ 166, 298, 42 },
		{ 166, 299, 42 },
		{ 166, 300, 42 },
		{ 166, 301, 42 },
		{ 166, 302, 42 },
		{ 166, 303, 42 },
		{ 166, 304, 42 },
		{ 166, 305, 42 },
		{ 166, 306, 42 },
		{ 166, 307, 42 },
		{ 166, 308, 42 },
		{ 166, 309, 42 },
		{ 166, 310, 42 },
		{ 166, 311, 42 },
		{ 166, 312, 42 },
		{ 166, 313, 42 },
		{ 167, 291, 42 },
		{ 167, 292, 42 },
		{ 167, 293, 42 },
		{ 167, 294, 42 },
		{ 167, 295, 42 },
		{ 167, 296, 42 },
		{ 167, 297, 42 },
		{ 167, 298, 42 },
		{ 167, 299, 42 },
		{ 167, 300, 42 },
		{ 167, 301, 42 },
		{ 167, 302, 42 },
		{ 167, 303, 42 },
		{ 167, 304, 42 },
		{ 167, 305, 42 },
		{ 167, 306, 42 },
		{ 167, 307, 42 },
		{ 167, 308, 42 },
		{ 167, 309, 42 },
		{ 167, 310, 42 },
		{ 167, 311, 42 },
		{ 167, 312, 42 },
		{ 167, 313, 42 },
		{ 168, 291, 42 },
		{ 168, 292, 42 },
		{ 168, 293, 42 },
		{ 168, 294, 42 },
		{ 168, 295, 42 },
		{ 168, 296, 42 },
		{ 168, 297, 42 },
		{ 168, 298, 42 },
		{ 168, 299, 42 },
		{ 168, 300, 42 },
		{ 168, 301, 42 },
		{ 168, 302, 42 },
		{ 168, 303, 42 },
		{ 168, 304, 42 },
		{ 168, 305, 42 },
		{ 168, 306, 42 },
		{ 168, 307, 42 },
		{ 168, 308, 42 },
		{ 168, 309, 42 },
		{ 168, 310, 42 },
		{ 168, 311, 42 },
		{ 168, 312, 42 },
		{ 169, 291, 42 },
		{ 169, 292, 42 },
		{ 169, 293, 42 },
		{ 169, 294, 42 },
		{ 169, 295, 42 },
		{ 169, 296, 42 },
		{ 169, 297, 42 },
		{ 169, 298, 42 },
		{ 169, 299, 42 },
		{ 169, 300, 42 },
		{ 169, 301, 42 },
		{ 169, 302, 42 },
		{ 169, 303, 42 },
		{ 169, 304, 42 },
		{ 169, 305, 42 },
		{ 169, 306, 42 },
		{ 169, 307, 42 },
		{ 169, 308, 42 },
		{ 169, 309, 42 },
		{ 169, 310, 42 },
		{ 169, 311, 42 },
		{ 170, 290, 42 },
		{ 170, 291, 42 },
		{ 170, 292, 42 },
		{ 170, 293, 42 },
		{ 170, 294, 42 },
		{ 170, 295, 42 },
		{ 170, 296, 42 },
		{ 170, 297, 42 },
		{ 170, 298, 42 },
		{ 170, 299, 42 },
		{ 170, 300, 42 },
		{ 170, 301, 42 },
		{ 170, 302, 42 },
		{ 170, 303, 42 },
		{ 170, 304, 42 },
		{ 170, 305, 42 },
		{ 170, 306, 42 },
		{ 170, 307, 42 },
		{ 170, 308, 42 },
		{ 170, 309, 42 },
		{ 170, 310, 42 },
		{ 170, 311, 42 },
		{ 171, 290, 42 },
		{ 171, 291, 42 },
		{ 171, 292, 42 },
		{ 171, 293, 42 },
		{ 171, 294, 42 },
		{ 171, 295, 42 },
		{ 171, 296, 42 },
		{ 171, 297, 42 },
		{ 171, 298, 42 },
		{ 171, 299, 42 },
		{ 171, 300, 42 },
		{ 171, 301, 42 },
		{ 171, 302, 42 },
		{ 171, 303, 42 },
		{ 171, 304, 42 },
		{ 171, 305, 42 },
		{ 171, 306, 42 },
		{ 171, 307, 42 },
		{ 171, 308, 42 },
		{ 171, 309, 42 },
		{ 171, 310, 42 },
		{ 172, 289, 42 },
		{ 172, 290, 42 },
		{ 172, 291, 42 },
		{ 172, 292, 42 },
		{ 172, 293, 42 },
		{ 172, 294, 42 },
		{ 172, 295, 42 },
		{ 172, 296, 42 },
		{ 172, 297, 42 },
		{ 172, 298, 42 },
		{ 172, 299, 42 },
		{ 172, 300, 42 },
		{ 172, 301, 42 },
		{ 172, 302, 42 },
		{ 172, 303, 42 },
		{ 172, 304, 42 },
		{ 172, 305, 42 },
		{ 172, 306, 42 },
		{ 172, 307, 42 },
		{ 172, 308, 42 },
		{ 172, 309, 42 },
		{ 172, 310, 42 },
		{ 173, 289, 42 },
		{ 173, 290, 42 },
		{ 173, 291, 42 },
		{ 173, 292, 42 },
		{ 173, 293, 42 },
		{ 173, 294, 42 },
		{ 173, 295, 42 },
		{ 173, 296, 42 },
		{ 173, 297, 42 },
		{ 173, 298, 42 },
		{ 173, 299, 42 },
		{ 173, 300, 42 },
		{ 173, 301, 42 },
		{ 173, 302, 42 },
		{ 173, 303, 42 },
		{ 173, 304, 42 },
		{ 173, 305, 42 },
		{ 173, 306, 42 },
		{ 173, 307, 42 },
		{ 173, 308, 42 },
		{ 173, 309, 42 },
		{ 174, 289, 42 },
		{ 174, 290, 42 },
		{ 174, 291, 42 },
		{ 174, 292, 42 },
		{ 174, 293, 42 },
		{ 174, 294, 42 },
		{ 174, 295, 42 },
		{ 174, 296, 42 },
		{ 174, 297, 42 },
		{ 174, 298, 42 },
		{ 174, 299, 42 },
		{ 174, 300, 42 },
		{ 174, 301, 42 },
		{ 174, 302, 42 },
		{ 174, 303, 42 },
		{ 174, 304, 42 },
		{ 174, 305, 42 },
		{ 174, 306, 42 },
		{ 174, 307, 42 },
		{ 174, 308, 42 },
		{ 175, 288, 42 },
		{ 175, 289, 42 },
		{ 175, 290, 42 },
		{ 175, 291, 42 },
		{ 175, 292, 42 },
		{ 175, 293, 42 },
		{ 175, 294, 42 },
		{ 175, 295, 42 },
		{ 175, 296, 42 },
		{ 175, 297, 42 },
		{ 175, 298, 42 },
		{ 175, 299, 42 },
		{ 175, 300, 42 },
		{ 175, 301, 42 },
		{ 175, 302, 42 },
		{ 175, 303, 42 },
		{ 175, 304, 42 },
		{ 175, 305, 42 },
		{ 175, 306, 42 },
		{ 175, 307, 42 },
		{ 175, 308, 42 },
		{ 176, 288, 42 },
		{ 176, 289, 42 },
		{ 176, 290, 42 },
		{ 176, 291, 42 },
		{ 176, 292, 42 },
		{ 176, 293, 42 },
		{ 176, 294, 42 },
		{ 176, 295, 42 },
		{ 176, 296, 42 },
		{ 176, 297, 42 },
		{ 176, 298, 42 },
		{ 176, 299, 42 },
		{ 176, 300, 42 },
		{ 176, 301, 42 },
		{ 176, 302, 42 },
		{ 176, 303, 42 },
		{ 176, 304, 42 },
		{ 176, 305, 42 },
		{ 176, 306, 42 },
		{ 176, 307, 42 },
		{ 177, 288, 42 },
		{ 177, 289, 42 },
		{ 177, 290, 42 },
		{ 177, 291, 42 },
		{ 177, 292, 42 },
		{ 177, 293, 42 },
		{ 177, 294, 42 },
		{ 177, 295, 42 },
		{ 177, 296, 42 },
		{ 177, 297, 42 },
		{ 177, 298, 42 },
		{ 177, 299, 42 },
		{ 177, 300, 42 },
		{ 177, 301, 42 },
		{ 177, 302, 42 },
		{ 177, 303, 42 },
		{ 177, 304, 42 },
		{ 177, 305, 42 },
		{ 177, 306, 42 },
		{ 178, 287, 42 },
		{ 178, 288, 42 },
		{ 178, 289, 42 },
		{ 178, 290, 42 },
		{ 178, 291, 42 },
		{ 178, 292, 42 },
		{ 178, 293, 42 },
		{ 178, 294, 42 },
		{ 178, 295, 42 },
		{ 178, 296, 42 },
		{ 178, 297, 42 },
		{ 178, 298, 42 },
		{ 178, 299, 42 },
		{ 178, 300, 42 },
		{ 178, 301, 42 },
		{ 178, 302, 42 },
		{ 178, 303, 42 },
		{ 178, 304, 42 },
		{ 178, 305, 42 },
		{ 178, 306, 42 },
		{ 179, 287, 42 },
		{ 179, 288, 42 },
		{ 179, 289, 42 },
		{ 179, 290, 42 },
		{ 179, 291, 42 },
		{ 179, 292, 42 },
		{ 179, 293, 42 },
		{ 179, 294, 42 },
		{ 179, 295, 42 },
		{ 179, 296, 42 },
		{ 179, 297, 42 },
		{ 179, 298, 42 },
		{ 179, 299, 42 },
		{ 179, 300, 42 },
		{ 179, 301, 42 },
		{ 179, 302, 42 },
		{ 179, 303, 42 },
		{ 179, 304, 42 },
		{ 179, 305, 42 },
		{ 180, 286, 42 },
		{ 180, 287, 42 },
		{ 180, 288, 42 },
		{ 180, 289, 42 },
		{ 180, 290, 42 },
		{ 180, 291, 42 },
		{ 180, 292, 42 },
		{ 180, 293, 42 },
		{ 180, 294, 42 },
		{ 180, 295, 42 },
		{ 180, 296, 42 },
		{ 180, 297, 42 },
		{ 180, 298, 42 },
		{ 180, 299, 42 },
		{ 180, 300, 42 },
		{ 180, 301, 42 },
		{ 180, 302, 42 },
		{ 180, 303, 42 },
		{ 180, 304, 42 },
		{ 180, 305, 42 },
		{ 181, 286, 42 },
		{ 181, 287, 42 },
		{ 181, 288, 42 },
		{ 181, 289, 42 },
		{ 181, 290, 42 },
		{ 181, 291, 42 },
		{ 181, 292, 42 },
		{ 181, 293, 42 },
		{ 181, 294, 42 },
		{ 181, 295, 42 },
		{ 181, 296, 42 },
		{ 181, 297, 42 },
		{ 181, 298, 42 },
		{ 181, 299, 42 },
		{ 181, 300, 42 },
		{ 181, 301, 42 },
		{ 181, 302, 42 },
		{ 181, 303, 42 },
		{ 181, 304, 42 },
		{ 182, 286, 42 },
		{ 182, 287, 42 },
		{ 182, 288, 42 },
		{ 182, 289, 42 },
		{ 182, 290, 42 },
		{ 182, 291, 42 },
		{ 182, 292, 42 },
		{ 182, 293, 42 },
		{ 182, 294, 42 },
		{ 182, 295, 42 },
		{ 182, 296, 42 },
		{ 182, 297, 42 },
		{ 182, 298, 42 },
		{ 182, 299, 42 },
		{ 182, 300, 42 },
		{ 182, 301, 42 },
		{ 182, 302, 42 },
		{ 182, 303, 42 },
		{ 183, 285, 42 },
		{ 183, 286, 42 },
		{ 183, 287, 42 },
		{ 183, 288, 42 },
		{ 183, 289, 42 },
		{ 183, 290, 42 },
		{ 183, 291, 42 },
		{ 183, 292, 42 },
		{ 183, 293, 42 },
		{ 183, 294, 42 },
		{ 183, 295, 42 },
		{ 183, 296, 42 },
		{ 183, 297, 42 },
		{ 183, 298, 42 },
		{ 183, 299, 42 },
		{ 183, 300, 42 },
		{ 183, 301, 42 },
		{ 183, 302, 42 },
		{ 183, 303, 42 },
		{ 184, 285, 42 },
		{ 184, 286, 42 },
		{ 184, 287, 42 },
		{ 184, 288, 42 },
		{ 184, 289, 42 },
		{ 184, 290, 42 },
		{ 184, 291, 42 },
		{ 184, 292, 42 },
		{ 184, 293, 42 },
		{ 184, 294, 42 },
		{ 184, 295, 42 },
		{ 184, 296, 42 },
		{ 184, 297, 42 },
		{ 184, 298, 42 },
		{ 184, 299, 42 },
		{ 184, 300, 42 },
		{ 184, 301, 42 },
		{ 184, 302, 42 },
		{ 185, 285, 42 },
		{ 185, 286, 42 },
		{ 185, 287, 42 },
		{ 185, 288, 42 },
		{ 185, 289, 42 },
		{ 185, 290, 42 },
		{ 185, 291, 42 },
		{ 185, 292, 42 },
		{ 185, 293, 42 },
		{ 185, 294, 42 },
		{ 185, 295, 42 },
		{ 185, 296, 42 },
		{ 185, 297, 42 },
		{ 185, 298, 42 },
		{ 185, 299, 42 },
		{ 185, 300, 42 },
		{ 185, 301, 42 },
		{ 186, 284, 42 },
		{ 186, 285, 42 },
		{ 186, 286, 42 },
		{ 186, 287, 42 },
		{ 186, 288, 42 },
		{ 186, 289, 42 },
		{ 186, 290, 42 },
		{ 186, 291, 42 },
		{ 186, 292, 42 },
		{ 186, 293, 42 },
		{ 186, 294, 42 },
		{ 186, 295, 42 },
		{ 186, 296, 42 },
		{ 186, 297, 42 },
		{ 186, 298, 42 },
		{ 186, 299, 42 },
		{ 186, 300, 42 },
		{ 186, 301, 42 },
		{ 187, 284, 42 },
		{ 187, 285, 42 },
		{ 187, 286, 42 },
		{ 187, 287, 42 },
		{ 187, 288, 42 },
		{ 187, 289, 42 },
		{ 187, 290, 42 },
		{ 187, 291, 42 },
		{ 187, 292, 42 },
		{ 187, 293, 42 },
		{ 187, 294, 42 },
		{ 187, 295, 42 },
		{ 187, 296, 42 },
		{ 187, 297, 42 },
		{ 187, 298, 42 },
		{ 187, 299, 42 },
		{ 187, 300, 42 },
		{ 188, 284, 42 },
		{ 188, 285, 42 },
		{ 188, 286, 42 },
		{ 188, 287, 42 },
		{ 188, 288, 42 },
		{ 188, 289, 42 },
		{ 188, 290, 42 },
		{ 188, 291, 42 },
		{ 188, 292, 42 },
		{ 188, 293, 42 },
		{ 188, 294, 42 },
		{ 188, 295, 42 },
		{ 188, 296, 42 },
		{ 188, 297, 42 },
		{ 188, 298, 42 },
		{ 188, 299, 42 },
		{ 189, 283, 42 },
		{ 189, 284, 42 },
		{ 189, 285, 42 },
		{ 189, 286, 42 },
		{ 189, 287, 42 },
		{ 189, 288, 42 },
		{ 189, 289, 42 },
		{ 189, 290, 42 },
		{ 189, 291, 42 },
		{ 189, 292, 42 },
		{ 189, 293, 42 },
		{ 189, 294, 42 },
		{ 189, 295, 42 },
		{ 189, 296, 42 },
		{ 189, 297, 42 },
		{ 189, 298, 42 },
		{ 189, 299, 42 },
		{ 190, 283, 42 },
		{ 190, 284, 42 },
		{ 190, 285, 42 },
		{ 190, 286, 42 },
		{ 190, 287, 42 },
		{ 190, 288, 42 },
		{ 190, 289, 42 },
		{ 190, 290, 42 },
		{ 190, 291, 42 },
		{ 190, 292, 42 },
		{ 190, 293, 42 },
		{ 190, 294, 42 },
		{ 190, 295, 42 },
		{ 190, 296, 42 },
		{ 190, 297, 42 },
		{ 190, 298, 42 },
		{ 191, 284, 42 },
		{ 191, 285, 42 },
		{ 191, 286, 42 },
		{ 191, 287, 42 },
		{ 191, 288, 42 },
		{ 191, 289, 42 },
		{ 191, 290, 42 },
		{ 191, 291, 42 },
		{ 191, 292, 42 },
		{ 191, 293, 42 },
		{ 191, 294, 42 },
		{ 191, 295, 42 },
		{ 191, 296, 42 },
		{ 323, 218, 42 },
		{ 323, 219, 42 },
		{ 323, 220, 42 },
		{ 323, 221, 42 },
		{ 323, 222, 42 },
		{ 323, 223, 42 },
		{ 323, 224, 42 },
		{ 323, 225, 42 },
		{ 323, 226, 42 },
		{ 323, 227, 42 },
		{ 323, 228, 42 },
		{ 323, 229, 42 },
		{ 323, 230, 42 },
		{ 324, 216, 42 },
		{ 324, 217, 42 },
		{ 324, 218, 42 },
		{ 324, 219, 42 },
		{ 324, 220, 42 },
		{ 324, 221, 42 },
		{ 324, 222, 42 },
		{ 324, 223, 42 },
		{ 324, 224, 42 },
		{ 324, 225, 42 },
		{ 324, 226, 42 },
		{ 324, 227, 42 },
		{ 324, 228, 42 },
		{ 324, 229, 42 },
		{ 324, 230, 42 },
		{ 324, 231, 42 },
		{ 325, 215, 42 },
		{ 325, 216, 42 },
		{ 325, 217, 42 },
		{ 325, 218, 42 },
		{ 325, 219, 42 },
		{ 325, 220, 42 },
		{ 325, 221, 42 },
		{ 325, 222, 42 },
		{ 325, 223, 42 },
		{ 325, 224, 42 },
		{ 325, 225, 42 },
		{ 325, 226, 42 },
		{ 325, 227, 42 },
		{ 325, 228, 42 },
		{ 325, 229, 42 },
		{ 325, 230, 42 },
		{ 325, 231, 42 },
		{ 326, 215, 42 },
		{ 326, 216, 42 },
		{ 326, 217, 42 },
		{ 326, 218, 42 },
		{ 326, 219, 42 },
		{ 326, 220, 42 },
		{ 326, 221, 42 },
		{ 326, 222, 42 },
		{ 326, 223, 42 },
		{ 326, 224, 42 },
		{ 326, 225, 42 },
		{ 326, 226, 42 },
		{ 326, 227, 42 },
		{ 326, 228, 42 },
		{ 326, 229, 42 },
		{ 326, 230, 42 },
		{ 327, 214, 42 },
		{ 327, 215, 42 },
		{ 327, 216, 42 },
		{ 327, 217, 42 },
		{ 327, 218, 42 },
		{ 327, 219, 42 },
		{ 327, 220, 42 },
		{ 327, 221, 42 },
		{ 327, 222, 42 },
		{ 327, 223, 42 },
		{ 327, 224, 42 },
		{ 327, 225, 42 },
		{ 327, 226, 42 },
		{ 327, 227, 42 },
		{ 327, 228, 42 },
		{ 327, 229, 42 },
		{ 327, 230, 42 },
		{ 328, 213, 42 },
		{ 328, 214, 42 },
		{ 328, 215, 42 },
		{ 328, 216, 42 },
		{ 328, 217, 42 },
		{ 328, 218, 42 },
		{ 328, 219, 42 },
		{ 328, 220, 42 },
		{ 328, 221, 42 },
		{ 328, 222, 42 },
		{ 328, 223, 42 },
		{ 328, 224, 42 },
		{ 328, 225, 42 },
		{ 328, 226, 42 },
		{ 328, 227, 42 },
		{ 328, 228, 42 },
		{ 328, 229, 42 },
		{ 328, 230, 42 },
		{ 329, 213, 42 },
		{ 329, 214, 42 },
		{ 329, 215, 42 },
		{ 329, 216, 42 },
		{ 329, 217, 42 },
		{ 329, 218, 42 },
		{ 329, 219, 42 },
		{ 329, 220, 42 },
		{ 329, 221, 42 },
		{ 329, 222, 42 },
		{ 329, 223, 42 },
		{ 329, 224, 42 },
		{ 329, 225, 42 },
		{ 329, 226, 42 },
		{ 329, 227, 42 },
		{ 329, 228, 42 },
		{ 329, 229, 42 },
		{ 330, 212, 42 },
		{ 330, 213, 42 },
		{ 330, 214, 42 },
		{ 330, 215, 42 },
		{ 330, 216, 42 },
		{ 330, 217, 42 },
		{ 330, 218, 42 },
		{ 330, 219, 42 },
		{ 330, 220, 42 },
		{ 330, 221, 42 },
		{ 330, 222, 42 },
		{ 330, 223, 42 },
		{ 330, 224, 42 },
		{ 330, 225, 42 },
		{ 330, 226, 42 },
		{ 330, 227, 42 },
		{ 330, 228, 42 },
		{ 330, 229, 42 },
		{ 331, 211, 42 },
		{ 331, 212, 42 },
		{ 331, 213, 42 },
		{ 331, 214, 42 },
		{ 331, 215, 42 },
		{ 331, 216, 42 },
		{ 331, 217, 42 },
		{ 331, 218, 42 },
		{ 331, 219, 42 },
		{ 331, 220, 42 },
		{ 331, 221, 42 },
		{ 331, 222, 42 },
		{ 331, 223, 42 },
		{ 331, 224, 42 },
		{ 331, 225, 42 },
		{ 331, 226, 42 },
		{ 331, 227, 42 },
		{ 331, 228, 42 },
		{ 331, 229, 42 },
		{ 332, 211, 42 },
		{ 332, 212, 42 },
		{ 332, 213, 42 },
		{ 332, 214, 42 },
		{ 332, 215, 42 },
		{ 332, 216, 42 },
		{ 332, 217, 42 },
		{ 332, 218, 42 },
		{ 332, 219, 42 },
		{ 332, 220, 42 },
		{ 332, 221, 42 },
		{ 332, 222, 42 },
		{ 332, 223, 42 },
		{ 332, 224, 42 },
		{ 332, 225, 42 },
		{ 332, 226, 42 },
		{ 332, 227, 42 },
		{ 332, 228, 42 },
		{ 333, 210, 42 },
		{ 333, 211, 42 },
		{ 333, 212, 42 },
		{ 333, 213, 42 },
		{ 333, 214, 42 },
		{ 333, 215, 42 },
		{ 333, 216, 42 },
		{ 333, 217, 42 },
		{ 333, 218, 42 },
		{ 333, 219, 42 },
		{ 333, 220, 42 },
		{ 333, 221, 42 },
		{ 333, 222, 42 },
		{ 333, 223, 42 },
		{ 333, 224, 42 },
		{ 333, 225, 42 },
		{ 333, 226, 42 },
		{ 333, 227, 42 },
		{ 333, 228, 42 },
		{ 334, 209, 42 },
		{ 334, 210, 42 },
		{ 334, 211, 42 },
		{ 334, 212, 42 },
		{ 334, 213, 42 },
		{ 334, 214, 42 },
		{ 334, 215, 42 },
		{ 334, 216, 42 },
		{ 334, 217, 42 },
		{ 334, 218, 42 },
		{ 334, 219, 42 },
		{ 334, 220, 42 },
		{ 334, 221, 42 },
		{ 334, 222, 42 },
		{ 334, 223, 42 },
		{ 334, 224, 42 },
		{ 334, 225, 42 },
		{ 334, 226, 42 },
		{ 334, 227, 42 },
		{ 334, 228, 42 },
		{ 335, 209, 42 },
		{ 335, 210, 42 },
		{ 335, 211, 42 },
		{ 335, 212, 42 },
		{ 335, 213, 42 },
		{ 335, 214, 42 },
		{ 335, 215, 42 },
		{ 335, 216, 42 },
		{ 335, 217, 42 },
		{ 335, 218, 42 },
		{ 335, 219, 42 },
		{ 335, 220, 42 },
		{ 335, 221, 42 },
		{ 335, 222, 42 },
		{ 335, 223, 42 },
		{ 335, 224, 42 },
		{ 335, 225, 42 },
		{ 335, 226, 42 },
		{ 335, 227, 42 },
		{ 336, 208, 42 },
		{ 336, 209, 42 },
		{ 336, 210, 42 },
		{ 336, 211, 42 },
		{ 336, 212, 42 },
		{ 336, 213, 42 },
		{ 336, 214, 42 },
		{ 336, 215, 42 },
		{ 336, 216, 42 },
		{ 336, 217, 42 },
		{ 336, 218, 42 },
		{ 336, 219, 42 },
		{ 336, 220, 42 },
		{ 336, 221, 42 },
		{ 336, 222, 42 },
		{ 336, 223, 42 },
		{ 336, 224, 42 },
		{ 336, 225, 42 },
		{ 336, 226, 42 },
		{ 336, 227, 42 },
		{ 337, 208, 42 },
		{ 337, 209, 42 },
		{ 337, 210, 42 },
		{ 337, 211, 42 },
		{ 337, 212, 42 },
		{ 337, 213, 42 },
		{ 337, 214, 42 },
		{ 337, 215, 42 },
		{ 337, 216, 42 },
		{ 337, 217, 42 },
		{ 337, 218, 42 },
		{ 337, 219, 42 },
		{ 337, 220, 42 },
		{ 337, 221, 42 },
		{ 337, 222, 42 },
		{ 337, 223, 42 },
		{ 337, 224, 42 },
		{ 337, 225, 42 },
		{ 337, 226, 42 },
		{ 338, 207, 42 },
		{ 338, 208, 42 },
		{ 338, 209, 42 },
		{ 338, 210, 42 },
		{ 338, 211, 42 },
		{ 338, 212, 42 },
		{ 338, 213, 42 },
		{ 338, 214, 42 },
		{ 338, 215, 42 },
		{ 338, 216, 42 },
		{ 338, 217, 42 },
		{ 338, 218, 42 },
		{ 338, 219, 42 },
		{ 338, 220, 42 },
		{ 338, 221, 42 },
		{ 338, 222, 42 },
		{ 338, 223, 42 },
		{ 338, 224, 42 },
		{ 338, 225, 42 },
		{ 338, 226, 42 },
		{ 339, 206, 42 },
		{ 339, 207, 42 },
		{ 339, 208, 42 },
		{ 339, 209, 42 },
		{ 339, 210, 42 },
		{ 339, 211, 42 },
		{ 339, 212, 42 },
		{ 339, 213, 42 },
		{ 339, 214, 42 },
		{ 339, 215, 42 },
		{ 339, 216, 42 },
		{ 339, 217, 42 },
		{ 339, 218, 42 },
		{ 339, 219, 42 },
		{ 339, 220, 42 },
		{ 339, 221, 42 },
		{ 339, 222, 42 },
		{ 339, 223, 42 },
		{ 339, 224, 42 },
		{ 339, 225, 42 },
		{ 339, 226, 42 },
		{ 340, 206, 42 },
		{ 340, 207, 42 },
		{ 340, 208, 42 },
		{ 340, 209, 42 },
		{ 340, 210, 42 },
		{ 340, 211, 42 },
		{ 340, 212, 42 },
		{ 340, 213, 42 },
		{ 340, 214, 42 },
		{ 340, 215, 42 },
		{ 340, 216, 42 },
		{ 340, 217, 42 },
		{ 340, 218, 42 },
		{ 340, 219, 42 },
		{ 340, 220, 42 },
		{ 340, 221, 42 },
		{ 340, 222, 42 },
		{ 340, 223, 42 },
		{ 340, 224, 42 },
		{ 340, 225, 42 },
		{ 341, 205, 42 },
		{ 341, 206, 42 },
		{ 341, 207, 42 },
		{ 341, 208, 42 },
		{ 341, 209, 42 },
		{ 341, 210, 42 },
		{ 341, 211, 42 },
		{ 341, 212, 42 },
		{ 341, 213, 42 },
		{ 341, 214, 42 },
		{ 341, 215, 42 },
		{ 341, 216, 42 },
		{ 341, 217, 42 },
		{ 341, 218, 42 },
		{ 341, 219, 42 },
		{ 341, 220, 42 },
		{ 341, 221, 42 },
		{ 341, 222, 42 },
		{ 341, 223, 42 },
		{ 341, 224, 42 },
		{ 341, 225, 42 },
		{ 342, 204, 42 },
		{ 342, 205, 42 },
		{ 342, 206, 42 },
		{ 342, 207, 42 },
		{ 342, 208, 42 },
		{ 342, 209, 42 },
		{ 342, 210, 42 },
		{ 342, 211, 42 },
		{ 342, 212, 42 },
		{ 342, 213, 42 },
		{ 342, 214, 42 },
		{ 342, 215, 42 },
		{ 342, 216, 42 },
		{ 342, 217, 42 },
		{ 342, 218, 42 },
		{ 342, 219, 42 },
		{ 342, 220, 42 },
		{ 342, 221, 42 },
		{ 342, 222, 42 },
		{ 342, 223, 42 },
		{ 342, 224, 42 },
		{ 342, 225, 42 },
		{ 343, 204, 42 },
		{ 343, 205, 42 },
		{ 343, 206, 42 },
		{ 343, 207, 42 },
		{ 343, 208, 42 },
		{ 343, 209, 42 },
		{ 343, 210, 42 },
		{ 343, 211, 42 },
		{ 343, 212, 42 },
		{ 343, 213, 42 },
		{ 343, 214, 42 },
		{ 343, 215, 42 },
		{ 343, 216, 42 },
		{ 343, 217, 42 },
		{ 343, 218, 42 },
		{ 343, 219, 42 },
		{ 343, 220, 42 },
		{ 343, 221, 42 },
		{ 343, 222, 42 },
		{ 343, 223, 42 },
		{ 343, 224, 42 },
		{ 344, 203, 42 },
		{ 344, 204, 42 },
		{ 344, 205, 42 },
		{ 344, 206, 42 },
		{ 344, 207, 42 },
		{ 344, 208, 42 },
		{ 344, 209, 42 },
		{ 344, 210, 42 },
		{ 344, 211, 42 },
		{ 344, 212, 42 },
		{ 344, 213, 42 },
		{ 344, 214, 42 },
		{ 344, 215, 42 },
		{ 344, 216, 42 },
		{ 344, 217, 42 },
		{ 344, 218, 42 },
		{ 344, 219, 42 },
		{ 344, 220, 42 },
		{ 344, 221, 42 },
		{ 344, 222, 42 },
		{ 344, 223, 42 },
		{ 344, 224, 42 },
		{ 345, 203, 42 },
		{ 345, 204, 42 },
		{ 345, 205, 42 },
		{ 345, 206, 42 },
		{ 345, 207, 42 },
		{ 345, 208, 42 },
		{ 345, 209, 42 },
		{ 345, 210, 42 },
		{ 345, 211, 42 },
		{ 345, 212, 42 },
		{ 345, 213, 42 },
		{ 345, 214, 42 },
		{ 345, 215, 42 },
		{ 345, 216, 42 },
		{ 345, 217, 42 },
		{ 345, 218, 42 },
		{ 345, 219, 42 },
		{ 345, 220, 42 },
		{ 345, 221, 42 },
		{ 345, 222, 42 },
		{ 345, 223, 42 },
		{ 346, 202, 42 },
		{ 346, 203, 42 },
		{ 346, 204, 42 },
		{ 346, 205, 42 },
		{ 346, 206, 42 },
		{ 346, 207, 42 },
		{ 346, 208, 42 },
		{ 346, 209, 42 },
		{ 346, 210, 42 },
		{ 346, 211, 42 },
		{ 346, 212, 42 },
		{ 346, 213, 42 },
		{ 346, 214, 42 },
		{ 346, 215, 42 },
		{ 346, 216, 42 },
		{ 346, 217, 42 },
		{ 346, 218, 42 },
		{ 346, 219, 42 },
		{ 346, 220, 42 },
		{ 346, 221, 42 },
		{ 346, 222, 42 },
		{ 346, 223, 42 },
		{ 347, 201, 42 },
		{ 347, 202, 42 },
		{ 347, 203, 42 },
		{ 347, 204, 42 },
		{ 347, 205, 42 },
		{ 347, 206, 42 },
		{ 347, 207, 42 },
		{ 347, 208, 42 },
		{ 347, 209, 42 },
		{ 347, 210, 42 },
		{ 347, 211, 42 },
		{ 347, 212, 42 },
		{ 347, 213, 42 },
		{ 347, 214, 42 },
		{ 347, 215, 42 },
		{ 347, 216, 42 },
		{ 347, 217, 42 },
		{ 347, 218, 42 },
		{ 347, 219, 42 },
		{ 347, 220, 42 },
		{ 347, 221, 42 },
		{ 347, 222, 42 },
		{ 347, 223, 42 },
		{ 348, 201, 42 },
		{ 348, 202, 42 },
		{ 348, 203, 42 },
		{ 348, 204, 42 },
		{ 348, 205, 42 },
		{ 348, 206, 42 },
		{ 348, 207, 42 },
		{ 348, 208, 42 },
		{ 348, 209, 42 },
		{ 348, 210, 42 },
		{ 348, 211, 42 },
		{ 348, 212, 42 },
		{ 348, 213, 42 },
		{ 348, 214, 42 },
		{ 348, 215, 42 },
		{ 348, 216, 42 },
		{ 348, 217, 42 },
		{ 348, 218, 42 },
		{ 348, 219, 42 },
		{ 348, 220, 42 },
		{ 348, 221, 42 },
		{ 348, 222, 42 },
		{ 349, 200, 42 },
		{ 349, 201, 42 },
		{ 349, 202, 42 },
		{ 349, 203, 42 },
		{ 349, 204, 42 },
		{ 349, 205, 42 },
		{ 349, 206, 42 },
		{ 349, 207, 42 },
		{ 349, 208, 42 },
		{ 349, 209, 42 },
		{ 349, 210, 42 },
		{ 349, 211, 42 },
		{ 349, 212, 42 },
		{ 349, 213, 42 },
		{ 349, 214, 42 },
		{ 349, 215, 42 },
		{ 349, 216, 42 },
		{ 349, 217, 42 },
		{ 349, 218, 42 },
		{ 349, 219, 42 },
		{ 349, 220, 42 },
		{ 349, 221, 42 },
		{ 349, 222, 42 },
		{ 350, 199, 42 },
		{ 350, 200, 42 },
		{ 350, 201, 42 },
		{ 350, 202, 42 },
		{ 350, 203, 42 },
		{ 350, 204, 42 },
		{ 350, 205, 42 },
		{ 350, 206, 42 },
		{ 350, 207, 42 },
		{ 350, 208, 42 },
		{ 350, 209, 42 },
		{ 350, 210, 42 },
		{ 350, 211, 42 },
		{ 350, 212, 42 },
		{ 350, 213, 42 },
		{ 350, 214, 42 },
		{ 350, 215, 42 },
		{ 350, 216, 42 },
		{ 350, 217, 42 },
		{ 350, 218, 42 },
		{ 350, 219, 42 },
		{ 350, 220, 42 },
		{ 350, 221, 42 },
		{ 350, 222, 42 },
		{ 351, 199, 42 },
		{ 351, 200, 42 },
		{ 351, 201, 42 },
		{ 351, 202, 42 },
		{ 351, 203, 42 },
		{ 351, 204, 42 },
		{ 351, 205, 42 },
		{ 351, 206, 42 },
		{ 351, 207, 42 },
		{ 351, 208, 42 },
		{ 351, 209, 42 },
		{ 351, 210, 42 },
		{ 351, 211, 42 },
		{ 351, 212, 42 },
		{ 351, 213, 42 },
		{ 351, 214, 42 },
		{ 351, 215, 42 },
		{ 351, 216, 42 },
		{ 351, 217, 42 },
		{ 351, 218, 42 },
		{ 351, 219, 42 },
		{ 351, 220, 42 },
		{ 351, 221, 42 },
		{ 352, 198, 42 },
		{ 352, 199, 42 },
		{ 352, 200, 42 },
		{ 352, 201, 42 },
		{ 352, 202, 42 },
		{ 352, 203, 42 },
		{ 352, 204, 42 },
		{ 352, 205, 42 },
		{ 352, 206, 42 },
		{ 352, 207, 42 },
		{ 352, 208, 42 },
		{ 352, 209, 42 },
		{ 352, 210, 42 },
		{ 352, 211, 42 },
		{ 352, 212, 42 },
		{ 352, 213, 42 },
		{ 352, 214, 42 },
		{ 352, 215, 42 },
		{ 352, 216, 42 },
		{ 352, 217, 42 },
		{ 352, 218, 42 },
		{ 352, 219, 42 },
		{ 352, 220, 42 },
		{ 352, 221, 42 },
		{ 353, 198, 42 },
		{ 353, 199, 42 },
		{ 353, 200, 42 },
		{ 353, 201, 42 },
		{ 353, 202, 42 },
		{ 353, 203, 42 },
		{ 353, 204, 42 },
		{ 353, 205, 42 },
		{ 353, 206, 42 },
		{ 353, 207, 42 },
		{ 353, 208, 42 },
		{ 353, 209, 42 },
		{ 353, 210, 42 },
		{ 353, 211, 42 },
		{ 353, 212, 42 },
		{ 353, 213, 42 },
		{ 353, 214, 42 },
		{ 353, 215, 42 },
		{ 353, 216, 42 },
		{ 353, 217, 42 },
		{ 353, 218, 42 },
		{ 353, 219, 42 },
		{ 353, 220, 42 },
		{ 354, 197, 42 },
		{ 354, 198, 42 },
		{ 354, 199, 42 },
		{ 354, 200, 42 },
		{ 354, 201, 42 },
		{ 354, 202, 42 },
		{ 354, 203, 42 },
		{ 354, 204, 42 },
		{ 354, 205, 42 },
		{ 354, 206, 42 },
		{ 354, 207, 42 },
		{ 354, 208, 42 },
		{ 354, 209, 42 },
		{ 354, 210, 42 },
		{ 354, 211, 42 },
		{ 354, 212, 42 },
		{ 354, 213, 42 },
		{ 354, 214, 42 },
		{ 354, 215, 42 },
		{ 354, 216, 42 },
		{ 354, 217, 42 },
		{ 354, 218, 42 },
		{ 354, 219, 42 },
		{ 354, 220, 42 },
		{ 355, 196, 42 },
		{ 355, 197, 42 },
		{ 355, 198, 42 },
		{ 355, 199, 42 },
		{ 355, 200, 42 },
		{ 355, 201, 42 },
		{ 355, 202, 42 },
		{ 355, 203, 42 },
		{ 355, 204, 42 },
		{ 355, 205, 42 },
		{ 355, 206, 42 },
		{ 355, 207, 42 },
		{ 355, 208, 42 },
		{ 355, 209, 42 },
		{ 355, 210, 42 },
		{ 355, 211, 42 },
		{ 355, 212, 42 },
		{ 355, 213, 42 },
		{ 355, 214, 42 },
		{ 355, 215, 42 },
		{ 355, 216, 42 },
		{ 355, 217, 42 },
		{ 355, 218, 42 },
		{ 355, 219, 42 },
		{ 355, 220, 42 },
		{ 356, 196, 42 },
		{ 356, 197, 42 },
		{ 356, 198, 42 },
		{ 356, 199, 42 },
		{ 356, 200, 42 },
		{ 356, 201, 42 },
		{ 356, 202, 42 },
		{ 356, 203, 42 },
		{ 356, 204, 42 },
		{ 356, 205, 42 },
		{ 356, 206, 42 },
		{ 356, 207, 42 },
		{ 356, 208, 42 },
		{ 356, 209, 42 },
		{ 356, 210, 42 },
		{ 356, 211, 42 },
		{ 356, 212, 42 },
		{ 356, 213, 42 },
		{ 356, 214, 42 },
		{ 356, 215, 42 },
		{ 356, 216, 42 },
		{ 356, 217, 42 },
		{ 356, 218, 42 },
		{ 356, 219, 42 },
		{ 357, 195, 42 },
		{ 357, 196, 42 },
		{ 357, 197, 42 },
		{ 357, 198, 42 },
		{ 357, 199, 42 },
		{ 357, 200, 42 },
		{ 357, 201, 42 },
		{ 357, 202, 42 },
		{ 357, 203, 42 },
		{ 357, 204, 42 },
		{ 357, 205, 42 },
		{ 357, 206, 42 },
		{ 357, 207, 42 },
		{ 357, 208, 42 },
		{ 357, 209, 42 },
		{ 357, 210, 42 },
		{ 357, 211, 42 },
		{ 357, 212, 42 },
		{ 357, 213, 42 },
		{ 357, 214, 42 },
		{ 357, 215, 42 },
		{ 357, 216, 42 },
		{ 357, 217, 42 },
		{ 357, 218, 42 },
		{ 357, 219, 42 },
		{ 358, 194, 42 },
		{ 358, 195, 42 },
		{ 358, 196, 42 },
		{ 358, 197, 42 },
		{ 358, 198, 42 },
		{ 358, 199, 42 },
		{ 358, 200, 42 },
		{ 358, 201, 42 },
		{ 358, 202, 42 },
		{ 358, 203, 42 },
		{ 358, 204, 42 },
		{ 358, 205, 42 },
		{ 358, 206, 42 },
		{ 358, 207, 42 },
		{ 358, 208, 42 },
		{ 358, 209, 42 },
		{ 358, 210, 42 },
		{ 358, 211, 42 },
		{ 358, 212, 42 },
		{ 358, 213, 42 },
		{ 358, 214, 42 },
		{ 358, 215, 42 },
		{ 358, 216, 42 },
		{ 358, 217, 42 },
		{ 358, 218, 42 },
		{ 358, 219, 42 },
		{ 359, 194, 42 },
		{ 359, 195, 42 },
		{ 359, 196, 42 },
		{ 359, 197, 42 },
		{ 359, 198, 42 },
		{ 359, 199, 42 },
		{ 359, 200, 42 },
		{ 359, 201, 42 },
		{ 359, 202, 42 },
		{ 359, 203, 42 },
		{ 359, 204, 42 },
		{ 359, 205, 42 },
		{ 359, 206, 42 },
		{ 359, 207, 42 },
		{ 359, 208, 42 },
		{ 359, 209, 42 },
		{ 359, 210, 42 },
		{ 359, 211, 42 },
		{ 359, 212, 42 },
		{ 359, 213, 42 },
		{ 359, 214, 42 },
		{ 359, 215, 42 },
		{ 359, 216, 42 },
		{ 359, 217, 42 },
		{ 359, 218, 42 },
		{ 360, 193, 42 },
		{ 360, 194, 42 },
		{ 360, 195, 42 },
		{ 360, 196, 42 },
		{ 360, 197, 42 },
		{ 360, 198, 42 },
		{ 360, 199, 42 },
		{ 360, 200, 42 },
		{ 360, 201, 42 },
		{ 360, 202, 42 },
		{ 360, 203, 42 },
		{ 360, 204, 42 },
		{ 360, 205, 42 },
		{ 360, 206, 42 },
		{ 360, 207, 42 },
		{ 360, 208, 42 },
		{ 360, 209, 42 },
		{ 360, 210, 42 },
		{ 360, 211, 42 },
		{ 360, 212, 42 },
		{ 360, 213, 42 },
		{ 360, 214, 42 },
		{ 360, 215, 42 },
		{ 360, 216, 42 },
		{ 360, 217, 42 },
		{ 360, 218, 42 },
		{ 361, 193, 42 },
		{ 361, 194, 42 },
		{ 361, 195, 42 },
		{ 361, 196, 42 },
		{ 361, 197, 42 },
		{ 361, 198, 42 },
		{ 361, 199, 42 },
		{ 361, 200, 42 },
		{ 361, 201, 42 },
		{ 361, 202, 42 },
		{ 361, 203, 42 },
		{ 361, 204, 42 },
		{ 361, 205, 42 },
		{ 361, 206, 42 },
		{ 361, 207, 42 },
		{ 361, 208, 42 },
		{ 361, 209, 42 },
		{ 361, 210, 42 },
		{ 361, 211, 42 },
		{ 361, 212, 42 },
		{ 361, 213, 42 },
		{ 361, 214, 42 },
		{ 361, 215, 42 },
		{ 361, 216, 42 },
		{ 361, 217, 42 },
		{ 362, 192, 42 },
		{ 362, 193, 42 },
		{ 362, 194, 42 },
		{ 362, 195, 42 },
		{ 362, 196, 42 },
		{ 362, 197, 42 },
		{ 362, 198, 42 },
		{ 362, 199, 42 },
		{ 362, 200, 42 },
		{ 362, 201, 42 },
		{ 362, 202, 42 },
		{ 362, 203, 42 },
		{ 362, 204, 42 },
		{ 362, 205, 42 },
		{ 362, 206, 42 },
		{ 362, 207, 42 },
		{ 362, 208, 42 },
		{ 362, 209, 42 },
		{ 362, 210, 42 },
		{ 362, 211, 42 },
		{ 362, 212, 42 },
		{ 362, 213, 42 },
		{ 362, 214, 42 },
		{ 362, 215, 42 },
		{ 362, 216, 42 },
		{ 362, 217, 42 },
		{ 363, 191, 42 },
		{ 363, 192, 42 },
		{ 363, 193, 42 },
		{ 363, 194, 42 },
		{ 363, 195, 42 },
		{ 363, 196, 42 },
		{ 363, 197, 42 },
		{ 363, 198, 42 },
		{ 363, 199, 42 },
		{ 363, 200, 42 },
		{ 363, 201, 42 },
		{ 363, 202, 42 },
		{ 363, 203, 42 },
		{ 363, 204, 42 },
		{ 363, 205, 42 },
		{ 363, 206, 42 },
		{ 363, 207, 42 },
		{ 363, 208, 42 },
		{ 363, 209, 42 },
		{ 363, 210, 42 },
		{ 363, 211, 42 },
		{ 363, 212, 42 },
		{ 363, 213, 42 },
		{ 363, 214, 42 },
		{ 363, 215, 42 },
		{ 363, 216, 42 },
		{ 363, 217, 42 },
		{ 364, 191, 42 },
		{ 364, 192, 42 },
		{ 364, 193, 42 },
		{ 364, 194, 42 },
		{ 364, 195, 42 },
		{ 364, 196, 42 },
		{ 364, 197, 42 },
		{ 364, 198, 42 },
		{ 364, 199, 42 },
		{ 364, 200, 42 },
		{ 364, 201, 42 },
		{ 364, 202, 42 },
		{ 364, 203, 42 },
		{ 364, 204, 42 },
		{ 364, 205, 42 },
		{ 364, 206, 42 },
		{ 364, 207, 42 },
		{ 364, 208, 42 },
		{ 364, 209, 42 },
		{ 364, 210, 42 },
		{ 364, 211, 42 },
		{ 364, 212, 42 },
		{ 364, 213, 42 },
		{ 364, 214, 42 },
		{ 364, 215, 42 },
		{ 364, 216, 42 },
		{ 365, 190, 42 },
		{ 365, 191, 42 },
		{ 365, 192, 42 },
		{ 365, 193, 42 },
		{ 365, 194, 42 },
		{ 365, 195, 42 },
		{ 365, 196, 42 },
		{ 365, 197, 42 },
		{ 365, 198, 42 },
		{ 365, 199, 42 },
		{ 365, 200, 42 },
		{ 365, 201, 42 },
		{ 365, 202, 42 },
		{ 365, 203, 42 },
		{ 365, 204, 42 },
		{ 365, 205, 42 },
		{ 365, 206, 42 },
		{ 365, 207, 42 },
		{ 365, 208, 42 },
		{ 365, 209, 42 },
		{ 365, 210, 42 },
		{ 365, 211, 42 },
		{ 365, 212, 42 },
		{ 365, 213, 42 },
		{ 365, 214, 42 },
		{ 365, 215, 42 },
		{ 365, 216, 42 },
		{ 366, 189, 42 },
		{ 366, 190, 42 },
		{ 366, 191, 42 },
		{ 366, 192, 42 },
		{ 366, 193, 42 },
		{ 366, 194, 42 },
		{ 366, 195, 42 },
		{ 366, 196, 42 },
		{ 366, 197, 42 },
		{ 366, 198, 42 },
		{ 366, 199, 42 },
		{ 366, 200, 42 },
		{ 366, 201, 42 },
		{ 366, 202, 42 },
		{ 366, 203, 42 },
		{ 366, 204, 42 },
		{ 366, 205, 42 },
		{ 366, 206, 42 },
		{ 366, 207, 42 },
		{ 366, 208, 42 },
		{ 366, 209, 42 },
		{ 366, 210, 42 },
		{ 366, 211, 42 },
		{ 366, 212, 42 },
		{ 366, 213, 42 },
		{ 366, 214, 42 },
		{ 366, 215, 42 },
		{ 366, 216, 42 },
		{ 367, 189, 42 },
		{ 367, 190, 42 },
		{ 367, 191, 42 },
		{ 367, 192, 42 },
		{ 367, 193, 42 },
		{ 367, 194, 42 },
		{ 367, 195, 42 },
		{ 367, 196, 42 },
		{ 367, 197, 42 },
		{ 367, 198, 42 },
		{ 367, 199, 42 },
		{ 367, 200, 42 },
		{ 367, 201, 42 },
		{ 367, 202, 42 },
		{ 367, 203, 42 },
		{ 367, 204, 42 },
		{ 367, 205, 42 },
		{ 367, 206, 42 },
		{ 367, 207, 42 },
		{ 367, 208, 42 },
		{ 367, 209, 42 },
		{ 367, 210, 42 },
		{ 367, 211, 42 },
		{ 367, 212, 42 },
		{ 367, 213, 42 },
		{ 367, 214, 42 },
		{ 367, 215, 42 },
		{ 368, 188, 42 },
		{ 368, 189, 42 },
		{ 368, 190, 42 },
		{ 368, 191, 42 },
		{ 368, 192, 42 },
		{ 368, 193, 42 },
		{ 368, 194, 42 },
		{ 368, 195, 42 },
		{ 368, 196, 42 },
		{ 368, 197, 42 },
		{ 368, 198, 42 },
		{ 368, 199, 42 },
		{ 368, 200, 42 },
		{ 368, 201, 42 },
		{ 368, 202, 42 },
		{ 368, 203, 42 },
		{ 368, 204, 42 },
		{ 368, 205, 42 },
		{ 368, 206, 42 },
		{ 368, 207, 42 },
		{ 368, 208, 42 },
		{ 368, 209, 42 },
		{ 368, 210, 42 },
		{ 368, 211, 42 },
		{ 368, 212, 42 },
		{ 368, 213, 42 },
		{ 368, 214, 42 },
		{ 368, 215, 42 },
		{ 369, 188, 42 },
		{ 369, 189, 42 },
		{ 369, 190, 42 },
		{ 369, 191, 42 },
		{ 369, 192, 42 },
		{ 369, 193, 42 },
		{ 369, 194, 42 },
		{ 369, 195, 42 },
		{ 369, 196, 42 },
		{ 369, 197, 42 },
		{ 369, 198, 42 },
		{ 369, 199, 42 },
		{ 369, 200, 42 },
		{ 369, 201, 42 },
		{ 369, 202, 42 },
		{ 369, 203, 42 },
		{ 369, 204, 42 },
		{ 369, 205, 42 },
		{ 369, 206, 42 },
		{ 369, 207, 42 },
		{ 369, 208, 42 },
		{ 369, 209, 42 },
		{ 369, 210, 42 },
		{ 369, 211, 42 },
		{ 369, 212, 42 },
		{ 369, 213, 42 },
		{ 369, 214, 42 },
		{ 370, 187, 42 },
		{ 370, 188, 42 },
		{ 370, 189, 42 },
		{ 370, 190, 42 },
		{ 370, 191, 42 },
		{ 370, 192, 42 },
		{ 370, 193, 42 },
		{ 370, 194, 42 },
		{ 370, 195, 42 },
		{ 370, 196, 42 },
		{ 370, 197, 42 },
		{ 370, 198, 42 },
		{ 370, 199, 42 },
		{ 370, 200, 42 },
		{ 370, 201, 42 },
		{ 370, 202, 42 },
		{ 370, 203, 42 },
		{ 370, 204, 42 },
		{ 370, 205, 42 },
		{ 370, 206, 42 },
		{ 370, 207, 42 },
		{ 370, 208, 42 },
		{ 370, 209, 42 },
		{ 370, 210, 42 },
		{ 370, 211, 42 },
		{ 370, 212, 42 },
		{ 370, 213, 42 },
		{ 370, 214, 42 },
		{ 371, 186, 42 },
		{ 371, 187, 42 },
		{ 371, 188, 42 },
		{ 371, 189, 42 },
		{ 371, 190, 42 },
		{ 371, 191, 42 },
		{ 371, 192, 42 },
		{ 371, 193, 42 },
		{ 371, 194, 42 },
		{ 371, 195, 42 },
		{ 371, 196, 42 },
		{ 371, 197, 42 },
		{ 371, 198, 42 },
		{ 371, 199, 42 },
		{ 371, 200, 42 },
		{ 371, 201, 42 },
		{ 371, 202, 42 },
		{ 371, 203, 42 },
		{ 371, 204, 42 },
		{ 371, 205, 42 },
		{ 371, 206, 42 },
		{ 371, 207, 42 },
		{ 371, 208, 42 },
		{ 371, 209, 42 },
		{ 371, 210, 42 },
		{ 371, 211, 42 },
		{ 371, 212, 42 },
		{ 371, 213, 42 },
		{ 371, 214, 42 },
		{ 372, 186, 42 },
		{ 372, 187, 42 },
		{ 372, 188, 42 },
		{ 372, 189, 42 },
		{ 372, 190, 42 },
		{ 372, 191, 42 },
		{ 372, 192, 42 },
		{ 372, 193, 42 },
		{ 372, 194, 42 },
		{ 372, 195, 42 },
		{ 372, 196, 42 },
		{ 372, 197, 42 },
		{ 372, 198, 42 },
		{ 372, 199, 42 },
		{ 372, 200, 42 },
		{ 372, 201, 42 },
		{ 372, 202, 42 },
		{ 372, 203, 42 },
		{ 372, 204, 42 },
		{ 372, 205, 42 },
		{ 372, 206, 42 },
		{ 372, 207, 42 },
		{ 372, 208, 42 },
		{ 372, 209, 42 },
		{ 372, 210, 42 },
		{ 372, 211, 42 },
		{ 372, 212, 42 },
		{ 372, 213, 42 },
		{ 373, 185, 42 },
		{ 373, 186, 42 },
		{ 373, 187, 42 },
		{ 373, 188, 42 },
		{ 373, 189, 42 },
		{ 373, 190, 42 },
		{ 373, 191, 42 },
		{ 373, 192, 42 },
		{ 373, 193, 42 },
		{ 373, 194, 42 },
		{ 373, 195, 42 },
		{ 373, 196, 42 },
		{ 373, 197, 42 },
		{ 373, 198, 42 },
		{ 373, 199, 42 },
		{ 373, 200, 42 },
		{ 373, 201, 42 },
		{ 373, 202, 42 },
		{ 373, 203, 42 },
		{ 373, 204, 42 },
		{ 373, 205, 42 },
		{ 373, 206, 42 },
		{ 373, 207, 42 },
		{ 373, 208, 42 },
		{ 373, 209, 42 },
		{ 373, 210, 42 },
		{ 373, 211, 42 },
		{ 373, 212, 42 },
		{ 373, 213, 42 },
		{ 374, 184, 42 },
		{ 374, 185, 42 },
		{ 374, 186, 42 },
		{ 374, 187, 42 },
		{ 374, 188, 42 },
		{ 374, 189, 42 },
		{ 374, 190, 42 },
		{ 374, 191, 42 },
		{ 374, 192, 42 },
		{ 374, 193, 42 },
		{ 374, 194, 42 },
		{ 374, 195, 42 },
		{ 374, 196, 42 },
		{ 374, 197, 42 },
		{ 374, 198, 42 },
		{ 374, 199, 42 },
		{ 374, 200, 42 },
		{ 374, 201, 42 },
		{ 374, 202, 42 },
		{ 374, 203, 42 },
		{ 374, 204, 42 },
		{ 374, 205, 42 },
		{ 374, 206, 42 },
		{ 374, 207, 42 },
		{ 374, 208, 42 },
		{ 374, 209, 42 },
		{ 374, 210, 42 },
		{ 374, 211, 42 },
		{ 374, 212, 42 },
		{ 374, 213, 42 },
		{ 375, 184, 42 },
		{ 375, 185, 42 },
		{ 375, 186, 42 },
		{ 375, 187, 42 },
		{ 375, 188, 42 },
		{ 375, 189, 42 },
		{ 375, 190, 42 },
		{ 375, 191, 42 },
		{ 375, 192, 42 },
		{ 375, 193, 42 },
		{ 375, 194, 42 },
		{ 375, 195, 42 },
		{ 375, 196, 42 },
		{ 375, 197, 42 },
		{ 375, 198, 42 },
		{ 375, 199, 42 },
		{ 375, 200, 42 },
		{ 375, 201, 42 },
		{ 375, 202, 42 },
		{ 375, 203, 42 },
		{ 375, 204, 42 },
		{ 375, 205, 42 },
		{ 375, 206, 42 },
		{ 375, 207, 42 },
		{ 375, 208, 42 },
		{ 375, 209, 42 },
		{ 375, 210, 42 },
		{ 375, 211, 42 },
		{ 375, 212, 42 },
		{ 376, 183, 42 },
		{ 376, 184, 42 },
		{ 376, 185, 42 },
		{ 376, 186, 42 },
		{ 376, 187, 42 },
		{ 376, 188, 42 },
		{ 376, 189, 42 },
		{ 376, 190, 42 },
		{ 376, 191, 42 },
		{ 376, 192, 42 },
		{ 376, 193, 42 },
		{ 376, 194, 42 },
		{ 376, 195, 42 },
		{ 376, 196, 42 },
		{ 376, 197, 42 },
		{ 376, 198, 42 },
		{ 376, 199, 42 },
		{ 376, 200, 42 },
		{ 376, 201, 42 },
		{ 376, 202, 42 },
		{ 376, 203, 42 },
		{ 376, 204, 42 },
		{ 376, 205, 42 },
		{ 376, 206, 42 },
		{ 376, 207, 42 },
		{ 376, 208, 42 },
		{ 376, 209, 42 },
		{ 376, 210, 42 },
		{ 376, 211, 42 },
		{ 376, 212, 42 },
		{ 377, 183, 42 },
		{ 377, 184, 42 },
		{ 377, 185, 42 },
		{ 377, 186, 42 },
		{ 377, 187, 42 },
		{ 377, 188, 42 },
		{ 377, 189, 42 },
		{ 377, 190, 42 },
		{ 377, 191, 42 },
		{ 377, 192, 42 },
		{ 377, 193, 42 },
		{ 377, 194, 42 },
		{ 377, 195, 42 },
		{ 377, 196, 42 },
		{ 377, 197, 42 },
		{ 377, 198, 42 },
		{ 377, 199, 42 },
		{ 377, 200, 42 },
		{ 377, 201, 42 },
		{ 377, 202, 42 },
		{ 377, 203, 42 },
		{ 377, 204, 42 },
		{ 377, 205, 42 },
		{ 377, 206, 42 },
		{ 377, 207, 42 },
		{ 377, 208, 42 },
		{ 377, 209, 42 },
		{ 377, 210, 42 },
		{ 377, 211, 42 },
		{ 378, 182, 42 },
		{ 378, 183, 42 },
		{ 378, 184, 42 },
		{ 378, 185, 42 },
		{ 378, 186, 42 },
		{ 378, 187, 42 },
		{ 378, 188, 42 },
		{ 378, 189, 42 },
		{ 378, 190, 42 },
		{ 378, 191, 42 },
		{ 378, 192, 42 },
		{ 378, 193, 42 },
		{ 378, 194, 42 },
		{ 378, 195, 42 },
		{ 378, 196, 42 },
		{ 378, 197, 42 },
		{ 378, 198, 42 },
		{ 378, 199, 42 },
		{ 378, 200, 42 },
		{ 378, 201, 42 },
		{ 378, 202, 42 },
		{ 378, 203, 42 },
		{ 378, 204, 42 },
		{ 378, 205, 42 },
		{ 378, 206, 42 },
		{ 378, 207, 42 },
		{ 378, 208, 42 },
		{ 378, 209, 42 },
		{ 378, 210, 42 },
		{ 378, 211, 42 },
		{ 379, 181, 42 },
		{ 379, 182, 42 },
		{ 379, 183, 42 },
		{ 379, 184, 42 },
		{ 379, 185, 42 },
		{ 379, 186, 42 },
		{ 379, 187, 42 },
		{ 379, 188, 42 },
		{ 379, 189, 42 },
		{ 379, 190, 42 },
		{ 379, 191, 42 },
		{ 379, 192, 42 },
		{ 379, 193, 42 },
		{ 379, 194, 42 },
		{ 379, 195, 42 },
		{ 379, 196, 42 },
		{ 379, 197, 42 },
		{ 379, 198, 42 },
		{ 379, 199, 42 },
		{ 379, 200, 42 },
		{ 379, 201, 42 },
		{ 379, 202, 42 },
		{ 379, 203, 42 },
		{ 379, 204, 42 },
		{ 379, 205, 42 },
		{ 379, 206, 42 },
		{ 379, 207, 42 },
		{ 379, 208, 42 },
		{ 379, 209, 42 },
		{ 379, 210, 42 },
		{ 379, 211, 42 },
		{ 380, 181, 42 },
		{ 380, 182, 42 },
		{ 380, 183, 42 },
		{ 380, 184, 42 },
		{ 380, 185, 42 },
		{ 380, 186, 42 },
		{ 380, 187, 42 },
		{ 380, 188, 42 },
		{ 380, 189, 42 },
		{ 380, 190, 42 },
		{ 380, 191, 42 },
		{ 380, 192, 42 },
		{ 380, 193, 42 },
		{ 380, 194, 42 },
		{ 380, 195, 42 },
		{ 380, 196, 42 },
		{ 380, 197, 42 },
		{ 380, 198, 42 },
		{ 380, 199, 42 },
		{ 380, 200, 42 },
		{ 380, 201, 42 },
		{ 380, 202, 42 },
		{ 380, 203, 42 },
		{ 380, 204, 42 },
		{ 380, 205, 42 },
		{ 380, 206, 42 },
		{ 380, 207, 42 },
		{ 380, 208, 42 },
		{ 380, 209, 42 },
		{ 380, 210, 42 },
		{ 381, 180, 42 },
		{ 381, 181, 42 },
		{ 381, 182, 42 },
		{ 381, 183, 42 },
		{ 381, 184, 42 },
		{ 381, 185, 42 },
		{ 381, 186, 42 },
		{ 381, 187, 42 },
		{ 381, 188, 42 },
		{ 381, 189, 42 },
		{ 381, 190, 42 },
		{ 381, 191, 42 },
		{ 381, 192, 42 },
		{ 381, 193, 42 },
		{ 381, 194, 42 },
		{ 381, 195, 42 },
		{ 381, 196, 42 },
		{ 381, 197, 42 },
		{ 381, 198, 42 },
		{ 381, 199, 42 },
		{ 381, 200, 42 },
		{ 381, 201, 42 },
		{ 381, 202, 42 },
		{ 381, 203, 42 },
		{ 381, 204, 42 },
		{ 381, 205, 42 },
		{ 381, 206, 42 },
		{ 381, 207, 42 },
		{ 381, 208, 42 },
		{ 381, 209, 42 },
		{ 381, 210, 42 },
		{ 382, 179, 42 },
		{ 382, 180, 42 },
		{ 382, 181, 42 },
		{ 382, 182, 42 },
		{ 382, 183, 42 },
		{ 382, 184, 42 },
		{ 382, 185, 42 },
		{ 382, 186, 42 },
		{ 382, 187, 42 },
		{ 382, 188, 42 },
		{ 382, 189, 42 },
		{ 382, 190, 42 },
		{ 382, 191, 42 },
		{ 382, 192, 42 },
		{ 382, 193, 42 },
		{ 382, 194, 42 },
		{ 382, 195, 42 },
		{ 382, 196, 42 },
		{ 382, 197, 42 },
		{ 382, 198, 42 },
		{ 382, 199, 42 },
		{ 382, 200, 42 },
		{ 382, 201, 42 },
		{ 382, 202, 42 },
		{ 382, 203, 42 },
		{ 382, 204, 42 },
		{ 382, 205, 42 },
		{ 382, 206, 42 },
		{ 382, 207, 42 },
		{ 382, 208, 42 },
		{ 382, 209, 42 },
		{ 382, 210, 42 },
		{ 383, 179, 42 },
		{ 383, 180, 42 },
		{ 383, 181, 42 },
		{ 383, 182, 42 },
		{ 383, 183, 42 },
		{ 383, 184, 42 },
		{ 383, 185, 42 },
		{ 383, 186, 42 },
		{ 383, 187, 42 },
		{ 383, 188, 42 },
		{ 383, 189, 42 },
		{ 383, 190, 42 },
		{ 383, 191, 42 },
		{ 383, 192, 42 },
		{ 383, 193, 42 },
		{ 383, 194, 42 },
		{ 383, 195, 42 },
		{ 383, 196, 42 },
		{ 383, 197, 42 },
		{ 383, 198, 42 },
		{ 383, 199, 42 },
		{ 383, 200, 42 },
		{ 383, 201, 42 },
		{ 383, 202, 42 },
		{ 383, 203, 42 },
		{ 383, 204, 42 },
		{ 383, 205, 42 },
		{ 383, 206, 42 },
		{ 383, 207, 42 },
		{ 383, 208, 42 },
		{ 383, 209, 42 },
		{ 384, 178, 42 },
		{ 384, 179, 42 },
		{ 384, 180, 42 },
		{ 384, 181, 42 },
		{ 384, 182, 42 },
		{ 384, 183, 42 },
		{ 384, 184, 42 },
		{ 384, 185, 42 },
		{ 384, 186, 42 },
		{ 384, 187, 42 },
		{ 384, 188, 42 },
		{ 384, 189, 42 },
		{ 384, 190, 42 },
		{ 384, 191, 42 },
		{ 384, 192, 42 },
		{ 384, 193, 42 },
		{ 384, 194, 42 },
		{ 384, 195, 42 },
		{ 384, 196, 42 },
		{ 384, 197, 42 },
		{ 384, 198, 42 },
		{ 384, 199, 42 },
		{ 384, 200, 42 },
		{ 384, 201, 42 },
		{ 384, 202, 42 },
		{ 384, 203, 42 },
		{ 384, 204, 42 },
		{ 384, 205, 42 },
		{ 384, 206, 42 },
		{ 384, 207, 42 },
		{ 384, 208, 42 },
		{ 384, 209, 42 },
		{ 385, 178, 42 },
		{ 385, 179, 42 },
		{ 385, 180, 42 },
		{ 385, 181, 42 },
		{ 385, 182, 42 },
		{ 385, 183, 42 },
		{ 385, 184, 42 },
		{ 385, 185, 42 },
		{ 385, 186, 42 },
		{ 385, 187, 42 },
		{ 385, 188, 42 },
		{ 385, 189, 42 },
		{ 385, 190, 42 },
		{ 385, 191, 42 },
		{ 385, 192, 42 },
		{ 385, 193, 42 },
		{ 385, 194, 42 },
		{ 385, 195, 42 },
		{ 385, 196, 42 },
		{ 385, 197, 42 },
		{ 385, 198, 42 },
		{ 385, 199, 42 },
		{ 385, 200, 42 },
		{ 385, 201, 42 },
		{ 385, 202, 42 },
		{ 385, 203, 42 },
		{ 385, 204, 42 },
		{ 385, 205, 42 },
		{ 385, 206, 42 },
		{ 385, 207, 42 },
		{ 385, 208, 42 },
		{ 386, 177, 42 },
		{ 386, 178, 42 },
		{ 386, 179, 42 },
		{ 386, 180, 42 },
		{ 386, 181, 42 },
		{ 386, 182, 42 },
		{ 386, 183, 42 },
		{ 386, 184, 42 },
		{ 386, 185, 42 },
		{ 386, 186, 42 },
		{ 386, 187, 42 },
		{ 386, 188, 42 },
		{ 386, 189, 42 },
		{ 386, 190, 42 },
		{ 386, 191, 42 },
		{ 386, 192, 42 },
		{ 386, 193, 42 },
		{ 386, 194, 42 },
		{ 386, 195, 42 },
		{ 386, 196, 42 },
		{ 386, 197, 42 },
		{ 386, 198, 42 },
		{ 386, 199, 42 },
		{ 386, 200, 42 },
		{ 386, 201, 42 },
		{ 386, 202, 42 },
		{ 386, 203, 42 },
		{ 386, 204, 42 },
		{ 386, 205, 42 },
		{ 386, 206, 42 },
		{ 386, 207, 42 },
		{ 386, 208, 42 },
		{ 387, 176, 42 },
		{ 387, 177, 42 },
		{ 387, 178, 42 },
		{ 387, 179, 42 },
		{ 387, 180, 42 },
		{ 387, 181, 42 },
		{ 387, 182, 42 },
		{ 387, 183, 42 },
		{ 387, 184, 42 },
		{ 387, 185, 42 },
		{ 387, 186, 42 },
		{ 387, 187, 42 },
		{ 387, 188, 42 },
		{ 387, 189, 42 },
		{ 387, 190, 42 },
		{ 387, 191, 42 },
		{ 387, 192, 42 },
		{ 387, 193, 42 },
		{ 387, 194, 42 },
		{ 387, 195, 42 },
		{ 387, 196, 42 },
		{ 387, 197, 42 },
		{ 387, 198, 42 },
		{ 387, 199, 42 },
		{ 387, 200, 42 },
		{ 387, 201, 42 },
		{ 387, 202, 42 },
		{ 387, 203, 42 },
		{ 387, 204, 42 },
		{ 387, 205, 42 },
		{ 387, 206, 42 },
		{ 387, 207, 42 },
		{ 387, 208, 42 },
		{ 388, 176, 42 },
		{ 388, 177, 42 },
		{ 388, 178, 42 },
		{ 388, 179, 42 },
		{ 388, 180, 42 },
		{ 388, 181, 42 },
		{ 388, 182, 42 },
		{ 388, 183, 42 },
		{ 388, 184, 42 },
		{ 388, 185, 42 },
		{ 388, 186, 42 },
		{ 388, 187, 42 },
		{ 388, 188, 42 },
		{ 388, 189, 42 },
		{ 388, 190, 42 },
		{ 388, 191, 42 },
		{ 388, 192, 42 },
		{ 388, 193, 42 },
		{ 388, 194, 42 },
		{ 388, 195, 42 },
		{ 388, 196, 42 },
		{ 388, 197, 42 },
		{ 388, 198, 42 },
		{ 388, 199, 42 },
		{ 388, 200, 42 },
		{ 388, 201, 42 },
		{ 388, 202, 42 },
		{ 388, 203, 42 },
		{ 388, 204, 42 },
		{ 388, 205, 42 },
		{ 388, 206, 42 },
		{ 388, 207, 42 },
		{ 389, 175, 42 },
		{ 389, 176, 42 },
		{ 389, 177, 42 },
		{ 389, 178, 42 },
		{ 389, 179, 42 },
		{ 389, 180, 42 },
		{ 389, 181, 42 },
		{ 389, 182, 42 },
		{ 389, 183, 42 },
		{ 389, 184, 42 },
		{ 389, 185, 42 },
		{ 389, 186, 42 },
		{ 389, 187, 42 },
		{ 389, 188, 42 },
		{ 389, 189, 42 },
		{ 389, 190, 42 },
		{ 389, 191, 42 },
		{ 389, 192, 42 },
		{ 389, 193, 42 },
		{ 389, 194, 42 },
		{ 389, 195, 42 },
		{ 389, 196, 42 },
		{ 389, 197, 42 },
		{ 389, 198, 42 },
		{ 389, 199, 42 },
		{ 389, 200, 42 },
		{ 389, 201, 42 },
		{ 389, 202, 42 },
		{ 389, 203, 42 },
		{ 389, 204, 42 },
		{ 389, 205, 42 },
		{ 389, 206, 42 },
		{ 389, 207, 42 },
		{ 390, 174, 42 },
		{ 390, 175, 42 },
		{ 390, 176, 42 },
		{ 390, 177, 42 },
		{ 390, 178, 42 },
		{ 390, 179, 42 },
		{ 390, 180, 42 },
		{ 390, 181, 42 },
		{ 390, 182, 42 },
		{ 390, 183, 42 },
		{ 390, 184, 42 },
		{ 390, 185, 42 },
		{ 390, 186, 42 },
		{ 390, 187, 42 },
		{ 390, 188, 42 },
		{ 390, 189, 42 },
		{ 390, 190, 42 },
		{ 390, 191, 42 },
		{ 390, 192, 42 },
		{ 390, 193, 42 },
		{ 390, 194, 42 },
		{ 390, 195, 42 },
		{ 390, 196, 42 },
		{ 390, 197, 42 },
		{ 390, 198, 42 },
		{ 390, 199, 42 },
		{ 390, 200, 42 },
		{ 390, 201, 42 },
		{ 390, 202, 42 },
		{ 390, 203, 42 },
		{ 390, 204, 42 },
		{ 390, 205, 42 },
		{ 390, 206, 42 },
		{ 390, 207, 42 },
		{ 391, 174, 42 },
		{ 391, 175, 42 },
		{ 391, 176, 42 },
		{ 391, 177, 42 },
		{ 391, 178, 42 },
		{ 391, 179, 42 },
		{ 391, 180, 42 },
		{ 391, 181, 42 },
		{ 391, 182, 42 },
		{ 391, 183, 42 },
		{ 391, 184, 42 },
		{ 391, 185, 42 },
		{ 391, 186, 42 },
		{ 391, 187, 42 },
		{ 391, 188, 42 },
		{ 391, 189, 42 },
		{ 391, 190, 42 },
		{ 391, 191, 42 },
		{ 391, 192, 42 },
		{ 391, 193, 42 },
		{ 391, 194, 42 },
		{ 391, 195, 42 },
		{ 391, 196, 42 },
		{ 391, 197, 42 },
		{ 391, 198, 42 },
		{ 391, 199, 42 },
		{ 391, 200, 42 },
		{ 391, 201, 42 },
		{ 391, 202, 42 },
		{ 391, 203, 42 },
		{ 391, 204, 42 },
		{ 391, 205, 42 },
		{ 391, 206, 42 },
		{ 392, 173, 42 },
		{ 392, 174, 42 },
		{ 392, 175, 42 },
		{ 392, 176, 42 },
		{ 392, 177, 42 },
		{ 392, 178, 42 },
		{ 392, 179, 42 },
		{ 392, 180, 42 },
		{ 392, 181, 42 },
		{ 392, 182, 42 },
		{ 392, 183, 42 },
		{ 392, 184, 42 },
		{ 392, 185, 42 },
		{ 392, 186, 42 },
		{ 392, 187, 42 },
		{ 392, 188, 42 },
		{ 392, 189, 42 },
		{ 392, 190, 42 },
		{ 392, 191, 42 },
		{ 392, 192, 42 },
		{ 392, 193, 42 },
		{ 392, 194, 42 },
		{ 392, 195, 42 },
		{ 392, 196, 42 },
		{ 392, 197, 42 },
		{ 392, 198, 42 },
		{ 392, 199, 42 },
		{ 392, 200, 42 },
		{ 392, 201, 42 },
		{ 392, 202, 42 },
		{ 392, 203, 42 },
		{ 392, 204, 42 },
		{ 392, 205, 42 },
		{ 392, 206, 42 },
		{ 393, 173, 42 },
		{ 393, 174, 42 },
		{ 393, 175, 42 },
		{ 393, 176, 42 },
		{ 393, 177, 42 },
		{ 393, 178, 42 },
		{ 393, 179, 42 },
		{ 393, 180, 42 },
		{ 393, 181, 42 },
		{ 393, 182, 42 },
		{ 393, 183, 42 },
		{ 393, 184, 42 },
		{ 393, 185, 42 },
		{ 393, 186, 42 },
		{ 393, 187, 42 },
		{ 393, 188, 42 },
		{ 393, 189, 42 },
		{ 393, 190, 42 },
		{ 393, 191, 42 },
		{ 393, 192, 42 },
		{ 393, 193, 42 },
		{ 393, 194, 42 },
		{ 393, 195, 42 },
		{ 393, 196, 42 },
		{ 393, 197, 42 },
		{ 393, 198, 42 },
		{ 393, 199, 42 },
		{ 393, 200, 42 },
		{ 393, 201, 42 },
		{ 393, 202, 42 },
		{ 393, 203, 42 },
		{ 393, 204, 42 },
		{ 393, 205, 42 },
		{ 394, 172, 42 },
		{ 394, 173, 42 },
		{ 394, 174, 42 },
		{ 394, 175, 42 },
		{ 394, 176, 42 },
		{ 394, 177, 42 },
		{ 394, 178, 42 },
		{ 394, 179, 42 },
		{ 394, 180, 42 },
		{ 394, 181, 42 },
		{ 394, 182, 42 },
		{ 394, 183, 42 },
		{ 394, 184, 42 },
		{ 394, 185, 42 },
		{ 394, 186, 42 },
		{ 394, 187, 42 },
		{ 394, 188, 42 },
		{ 394, 189, 42 },
		{ 394, 190, 42 },
		{ 394, 191, 42 },
		{ 394, 192, 42 },
		{ 394, 193, 42 },
		{ 394, 194, 42 },
		{ 394, 195, 42 },
		{ 394, 196, 42 },
		{ 394, 197, 42 },
		{ 394, 198, 42 },
		{ 394, 199, 42 },
		{ 394, 200, 42 },
		{ 394, 201, 42 },
		{ 394, 202, 42 },
		{ 394, 203, 42 },
		{ 394, 204, 42 },
		{ 394, 205, 42 },
		{ 395, 171, 42 },
		{ 395, 172, 42 },
		{ 395, 173, 42 },
		{ 395, 174, 42 },
		{ 395, 175, 42 },
		{ 395, 176, 42 },
		{ 395, 177, 42 },
		{ 395, 178, 42 },
		{ 395, 179, 42 },
		{ 395, 180, 42 },
		{ 395, 181, 42 },
		{ 395, 182, 42 },
		{ 395, 183, 42 },
		{ 395, 184, 42 },
		{ 395, 185, 42 },
		{ 395, 186, 42 },
		{ 395, 187, 42 },
		{ 395, 188, 42 },
		{ 395, 189, 42 },
		{ 395, 190, 42 },
		{ 395, 191, 42 },
		{ 395, 192, 42 },
		{ 395, 193, 42 },
		{ 395, 194, 42 },
		{ 395, 195, 42 },
		{ 395, 196, 42 },
		{ 395, 197, 42 },
		{ 395, 198, 42 },
		{ 395, 199, 42 },
		{ 395, 200, 42 },
		{ 395, 201, 42 },
		{ 395, 202, 42 },
		{ 395, 203, 42 },
		{ 395, 204, 42 },
		{ 395, 205, 42 },
		{ 396, 171, 42 },
		{ 396, 172, 42 },
		{ 396, 173, 42 },
		{ 396, 174, 42 },
		{ 396, 175, 42 },
		{ 396, 176, 42 },
		{ 396, 177, 42 },
		{ 396, 178, 42 },
		{ 396, 179, 42 },
		{ 396, 180, 42 },
		{ 396, 181, 42 },
		{ 396, 182, 42 },
		{ 396, 183, 42 },
		{ 396, 184, 42 },
		{ 396, 185, 42 },
		{ 396, 186, 42 },
		{ 396, 187, 42 },
		{ 396, 188, 42 },
		{ 396, 189, 42 },
		{ 396, 190, 42 },
		{ 396, 191, 42 },
		{ 396, 192, 42 },
		{ 396, 193, 42 },
		{ 396, 194, 42 },
		{ 396, 195, 42 },
		{ 396, 196, 42 },
		{ 396, 197, 42 },
		{ 396, 198, 42 },
		{ 396, 199, 42 },
		{ 396, 200, 42 },
		{ 396, 201, 42 },
		{ 396, 202, 42 },
		{ 396, 203, 42 },
		{ 396, 204, 42 },
		{ 397, 170, 42 },
		{ 397, 171, 42 },
		{ 397, 172, 42 },
		{ 397, 173, 42 },
		{ 397, 174, 42 },
		{ 397, 175, 42 },
		{ 397, 176, 42 },
		{ 397, 177, 42 },
		{ 397, 178, 42 },
		{ 397, 179, 42 },
		{ 397, 180, 42 },
		{ 397, 181, 42 },
		{ 397, 182, 42 },
		{ 397, 183, 42 },
		{ 397, 184, 42 },
		{ 397, 185, 42 },
		{ 397, 186, 42 },
		{ 397, 187, 42 },
		{ 397, 188, 42 },
		{ 397, 189, 42 },
		{ 397, 190, 42 },
		{ 397, 191, 42 },
		{ 397, 192, 42 },
		{ 397, 193, 42 },
		{ 397, 194, 42 },
		{ 397, 195, 42 },
		{ 397, 196, 42 },
		{ 397, 197, 42 },
		{ 397, 198, 42 },
		{ 397, 199, 42 },
		{ 397, 200, 42 },
		{ 397, 201, 42 },
		{ 397, 202, 42 },
		{ 397, 203, 42 },
		{ 397, 204, 42 },
		{ 398, 169, 42 },
		{ 398, 170, 42 },
		{ 398, 171, 42 },
		{ 398, 172, 42 },
		{ 398, 173, 42 },
		{ 398, 174, 42 },
		{ 398, 175, 42 },
		{ 398, 176, 42 },
		{ 398, 177, 42 },
		{ 398, 178, 42 },
		{ 398, 179, 42 },
		{ 398, 180, 42 },
		{ 398, 181, 42 },
		{ 398, 182, 42 },
		{ 398, 183, 42 },
		{ 398, 184, 42 },
		{ 398, 185, 42 },
		{ 398, 186, 42 },
		{ 398, 187, 42 },
		{ 398, 188, 42 },
		{ 398, 189, 42 },
		{ 398, 190, 42 },
		{ 398, 191, 42 },
		{ 398, 192, 42 },
		{ 398, 193, 42 },
		{ 398, 194, 42 },
		{ 398, 195, 42 },
		{ 398, 196, 42 },
		{ 398, 197, 42 },
		{ 398, 198, 42 },
		{ 398, 199, 42 },
		{ 398, 200, 42 },
		{ 398, 201, 42 },
		{ 398, 202, 42 },
		{ 398, 203, 42 },
		{ 398, 204, 42 },
		{ 399, 169, 42 },
		{ 399, 170, 42 },
		{ 399, 171, 42 },
		{ 399, 172, 42 },
		{ 399, 173, 42 },
		{ 399, 174, 42 },
		{ 399, 175, 42 },
		{ 399, 176, 42 },
		{ 399, 177, 42 },
		{ 399, 178, 42 },
		{ 399, 179, 42 },
		{ 399, 180, 42 },
		{ 399, 181, 42 },
		{ 399, 182, 42 },
		{ 399, 183, 42 },
		{ 399, 184, 42 },
		{ 399, 185, 42 },
		{ 399, 186, 42 },
		{ 399, 187, 42 },
		{ 399, 188, 42 },
		{ 399, 189, 42 },
		{ 399, 190, 42 },
		{ 399, 191, 42 },
		{ 399, 192, 42 },
		{ 399, 193, 42 },
		{ 399, 194, 42 },
		{ 399, 195, 42 },
		{ 399, 196, 42 },
		{ 399, 197, 42 },
		{ 399, 198, 42 },
		{ 399, 199, 42 },
		{ 399, 200, 42 },
		{ 399, 201, 42 },
		{ 399, 202, 42 },
		{ 399, 203, 42 },
		{ 400, 168, 42 },
		{ 400, 169, 42 },
		{ 400, 170, 42 },
		{ 400, 171, 42 },
		{ 400, 172, 42 },
		{ 400, 173, 42 },
		{ 400, 174, 42 },
		{ 400, 175, 42 },
		{ 400, 176, 42 },
		{ 400, 177, 42 },
		{ 400, 178, 42 },
		{ 400, 179, 42 },
		{ 400, 180, 42 },
		{ 400, 181, 42 },
		{ 400, 182, 42 },
		{ 400, 183, 42 },
		{ 400, 184, 42 },
		{ 400, 185, 42 },
		{ 400, 186, 42 },
		{ 400, 187, 42 },
		{ 400, 188, 42 },
		{ 400, 189, 42 },
		{ 400, 190, 42 },
		{ 400, 191, 42 },
		{ 400, 192, 42 },
		{ 400, 193, 42 },
		{ 400, 194, 42 },
		{ 400, 195, 42 },
		{ 400, 196, 42 },
		{ 400, 197, 42 },
		{ 400, 198, 42 },
		{ 400, 199, 42 },
		{ 400, 200, 42 },
		{ 400, 201, 42 },
		{ 400, 202, 42 },
		{ 400, 203, 42 },
		{ 401, 168, 42 },
		{ 401, 169, 42 },
		{ 401, 170, 42 },
		{ 401, 171, 42 },
		{ 401, 172, 42 },
		{ 401, 173, 42 },
		{ 401, 174, 42 },
		{ 401, 175, 42 },
		{ 401, 176, 42 },
		{ 401, 177, 42 },
		{ 401, 178, 42 },
		{ 401, 179, 42 },
		{ 401, 180, 42 },
		{ 401, 181, 42 },
		{ 401, 182, 42 },
		{ 401, 183, 42 },
		{ 401, 184, 42 },
		{ 401, 185, 42 },
		{ 401, 186, 42 },
		{ 401, 187, 42 },
		{ 401, 188, 42 },
		{ 401, 189, 42 },
		{ 401, 190, 42 },
		{ 401, 191, 42 },
		{ 401, 192, 42 },
		{ 401, 193, 42 },
		{ 401, 194, 42 },
		{ 401, 195, 42 },
		{ 401, 196, 42 },
		{ 401, 197, 42 },
		{ 401, 198, 42 },
		{ 401, 199, 42 },
		{ 401, 200, 42 },
		{ 401, 201, 42 },
		{ 401, 202, 42 },
		{ 402, 167, 42 },
		{ 402, 168, 42 },
		{ 402, 169, 42 },
		{ 402, 170, 42 },
		{ 402, 171, 42 },
		{ 402, 172, 42 },
		{ 402, 173, 42 },
		{ 402, 174, 42 },
		{ 402, 175, 42 },
		{ 402, 176, 42 },
		{ 402, 177, 42 },
		{ 402, 178, 42 },
		{ 402, 179, 42 },
		{ 402, 180, 42 },
		{ 402, 181, 42 },
		{ 402, 182, 42 },
		{ 402, 183, 42 },
		{ 402, 184, 42 },
		{ 402, 185, 42 },
		{ 402, 186, 42 },
		{ 402, 187, 42 },
		{ 402, 188, 42 },
		{ 402, 189, 42 },
		{ 402, 190, 42 },
		{ 402, 191, 42 },
		{ 402, 192, 42 },
		{ 402, 193, 42 },
		{ 402, 194, 42 },
		{ 402, 195, 42 },
		{ 402, 196, 42 },
		{ 402, 197, 42 },
		{ 402, 198, 42 },
		{ 402, 199, 42 },
		{ 402, 200, 42 },
		{ 402, 201, 42 },
		{ 402, 202, 42 },
		{ 403, 166, 42 },
		{ 403, 167, 42 },
		{ 403, 168, 42 },
		{ 403, 169, 42 },
		{ 403, 170, 42 },
		{ 403, 171, 42 },
		{ 403, 172, 42 },
		{ 403, 173, 42 },
		{ 403, 174, 42 },
		{ 403, 175, 42 },
		{ 403, 176, 42 },
		{ 403, 177, 42 },
		{ 403, 178, 42 },
		{ 403, 179, 42 },
		{ 403, 180, 42 },
		{ 403, 181, 42 },
		{ 403, 182, 42 },
		{ 403, 183, 42 },
		{ 403, 184, 42 },
		{ 403, 185, 42 },
		{ 403, 186, 42 },
		{ 403, 187, 42 },
		{ 403, 188, 42 },
		{ 403, 189, 42 },
		{ 403, 190, 42 },
		{ 403, 191, 42 },
		{ 403, 192, 42 },
		{ 403, 193, 42 },
		{ 403, 194, 42 },
		{ 403, 195, 42 },
		{ 403, 196, 42 },
		{ 403, 197, 42 },
		{ 403, 198, 42 },
		{ 403, 199, 42 },
		{ 403, 200, 42 },
		{ 403, 201, 42 },
		{ 403, 202, 42 },
		{ 404, 166, 42 },
		{ 404, 167, 42 },
		{ 404, 168, 42 },
		{ 404, 169, 42 },
		{ 404, 170, 42 },
		{ 404, 171, 42 },
		{ 404, 172, 42 },
		{ 404, 173, 42 },
		{ 404, 174, 42 },
		{ 404, 175, 42 },
		{ 404, 176, 42 },
		{ 404, 177, 42 },
		{ 404, 178, 42 },
		{ 404, 179, 42 },
		{ 404, 180, 42 },
		{ 404, 181, 42 },
		{ 404, 182, 42 },
		{ 404, 183, 42 },
		{ 404, 184, 42 },
		{ 404, 185, 42 },
		{ 404, 186, 42 },
		{ 404, 187, 42 },
		{ 404, 188, 42 },
		{ 404, 189, 42 },
		{ 404, 190, 42 },
		{ 404, 191, 42 },
		{ 404, 192, 42 },
		{ 404, 193, 42 },
		{ 404, 194, 42 },
		{ 404, 195, 42 },
		{ 404, 196, 42 },
		{ 404, 197, 42 },
		{ 404, 198, 42 },
		{ 404, 199, 42 },
		{ 404, 200, 42 },
		{ 404, 201, 42 },
		{ 405, 165, 42 },
		{ 405, 166, 42 },
		{ 405, 167, 42 },
		{ 405, 168, 42 },
		{ 405, 169, 42 },
		{ 405, 170, 42 },
		{ 405, 171, 42 },
		{ 405, 172, 42 },
		{ 405, 173, 42 },
		{ 405, 174, 42 },
		{ 405, 175, 42 },
		{ 405, 176, 42 },
		{ 405, 177, 42 },
		{ 405, 178, 42 },
		{ 405, 179, 42 },
		{ 405, 180, 42 },
		{ 405, 181, 42 },
		{ 405, 182, 42 },
		{ 405, 183, 42 },
		{ 405, 184, 42 },
		{ 405, 185, 42 },
		{ 405, 186, 42 },
		{ 405, 187, 42 },
		{ 405, 188, 42 },
		{ 405, 189, 42 },
		{ 405, 190, 42 },
		{ 405, 191, 42 },
		{ 405, 192, 42 },
		{ 405, 193, 42 },
		{ 405, 194, 42 },
		{ 405, 195, 42 },
		{ 405, 196, 42 },
		{ 405, 197, 42 },
		{ 405, 198, 42 },
		{ 405, 199, 42 },
		{ 405, 200, 42 },
		{ 405, 201, 42 },
		{ 406, 164, 42 },
		{ 406, 165, 42 },
		{ 406, 166, 42 },
		{ 406, 167, 42 },
		{ 406, 168, 42 },
		{ 406, 169, 42 },
		{ 406, 170, 42 },
		{ 406, 171, 42 },
		{ 406, 172, 42 },
		{ 406, 173, 42 },
		{ 406, 174, 42 },
		{ 406, 175, 42 },
		{ 406, 176, 42 },
		{ 406, 177, 42 },
		{ 406, 178, 42 },
		{ 406, 179, 42 },
		{ 406, 180, 42 },
		{ 406, 181, 42 },
		{ 406, 182, 42 },
		{ 406, 183, 42 },
		{ 406, 184, 42 },
		{ 406, 185, 42 },
		{ 406, 186, 42 },
		{ 406, 187, 42 },
		{ 406, 188, 42 },
		{ 406, 189, 42 },
		{ 406, 190, 42 },
		{ 406, 191, 42 },
		{ 406, 192, 42 },
		{ 406, 193, 42 },
		{ 406, 194, 42 },
		{ 406, 195, 42 },
		{ 406, 196, 42 },
		{ 406, 197, 42 },
		{ 406, 198, 42 },
		{ 406, 199, 42 },
		{ 406, 200, 42 },
		{ 406, 201, 42 },
		{ 407, 164, 42 },
		{ 407, 165, 42 },
		{ 407, 166, 42 },
		{ 407, 167, 42 },
		{ 407, 168, 42 },
		{ 407, 169, 42 },
		{ 407, 170, 42 },
		{ 407, 171, 42 },
		{ 407, 172, 42 },
		{ 407, 173, 42 },
		{ 407, 174, 42 },
		{ 407, 175, 42 },
		{ 407, 176, 42 },
		{ 407, 177, 42 },
		{ 407, 178, 42 },
		{ 407, 179, 42 },
		{ 407, 180, 42 },
		{ 407, 181, 42 },
		{ 407, 182, 42 },
		{ 407, 183, 42 },
		{ 407, 184, 42 },
		{ 407, 185, 42 },
		{ 407, 186, 42 },
		{ 407, 187, 42 },
		{ 407, 188, 42 },
		{ 407, 189, 42 },
		{ 407, 190, 42 },
		{ 407, 191, 42 },
		{ 407, 192, 42 },
		{ 407, 193, 42 },
		{ 407, 194, 42 },
		{ 407, 195, 42 },
		{ 407, 196, 42 },
		{ 407, 197, 42 },
		{ 407, 198, 42 },
		{ 407, 199, 42 },
		{ 407, 200, 42 },
		{ 408, 163, 42 },
		{ 408, 164, 42 },
		{ 408, 165, 42 },
		{ 408, 166, 42 },
		{ 408, 167, 42 },
		{ 408, 168, 42 },
		{ 408, 169, 42 },
		{ 408, 170, 42 },
		{ 408, 171, 42 },
		{ 408, 172, 42 },
		{ 408, 173, 42 },
		{ 408, 174, 42 },
		{ 408, 175, 42 },
		{ 408, 176, 42 },
		{ 408, 177, 42 },
		{ 408, 178, 42 },
		{ 408, 179, 42 },
		{ 408, 180, 42 },
		{ 408, 181, 42 },
		{ 408, 182, 42 },
		{ 408, 183, 42 },
		{ 408, 184, 42 },
		{ 408, 185, 42 },
		{ 408, 186, 42 },
		{ 408, 187, 42 },
		{ 408, 188, 42 },
		{ 408, 189, 42 },
		{ 408, 190, 42 },
		{ 408, 191, 42 },
		{ 408, 192, 42 },
		{ 408, 193, 42 },
		{ 408, 194, 42 },
		{ 408, 195, 42 },
		{ 408, 196, 42 },
		{ 408, 197, 42 },
		{ 408, 198, 42 },
		{ 408, 199, 42 },
		{ 408, 200, 42 },
		{ 409, 163, 42 },
		{ 409, 164, 42 },
		{ 409, 165, 42 },
		{ 409, 166, 42 },
		{ 409, 167, 42 },
		{ 409, 168, 42 },
		{ 409, 169, 42 },
		{ 409, 170, 42 },
		{ 409, 171, 42 },
		{ 409, 172, 42 },
		{ 409, 173, 42 },
		{ 409, 174, 42 },
		{ 409, 175, 42 },
		{ 409, 176, 42 },
		{ 409, 177, 42 },
		{ 409, 178, 42 },
		{ 409, 179, 42 },
		{ 409, 180, 42 },
		{ 409, 181, 42 },
		{ 409, 182, 42 },
		{ 409, 183, 42 },
		{ 409, 184, 42 },
		{ 409, 185, 42 },
		{ 409, 186, 42 },
		{ 409, 187, 42 },
		{ 409, 188, 42 },
		{ 409, 189, 42 },
		{ 409, 190, 42 },
		{ 409, 191, 42 },
		{ 409, 192, 42 },
		{ 409, 193, 42 },
		{ 409, 194, 42 },
		{ 409, 195, 42 },
		{ 409, 196, 42 },
		{ 409, 197, 42 },
		{ 409, 198, 42 },
		{ 409, 199, 42 },
		{ 410, 162, 42 },
		{ 410, 163, 42 },
		{ 410, 164, 42 },
		{ 410, 165, 42 },
		{ 410, 166, 42 },
		{ 410, 167, 42 },
		{ 410, 168, 42 },
		{ 410, 169, 42 },
		{ 410, 170, 42 },
		{ 410, 171, 42 },
		{ 410, 172, 42 },
		{ 410, 173, 42 },
		{ 410, 174, 42 },
		{ 410, 175, 42 },
		{ 410, 176, 42 },
		{ 410, 177, 42 },
		{ 410, 178, 42 },
		{ 410, 179, 42 },
		{ 410, 180, 42 },
		{ 410, 181, 42 },
		{ 410, 182, 42 },
		{ 410, 183, 42 },
		{ 410, 184, 42 },
		{ 410, 185, 42 },
		{ 410, 186, 42 },
		{ 410, 187, 42 },
		{ 410, 188, 42 },
		{ 410, 189, 42 },
		{ 410, 190, 42 },
		{ 410, 191, 42 },
		{ 410, 192, 42 },
		{ 410, 193, 42 },
		{ 410, 194, 42 },
		{ 410, 195, 42 },
		{ 410, 196, 42 },
		{ 410, 197, 42 },
		{ 410, 198, 42 },
		{ 410, 199, 42 },
		{ 411, 161, 42 },
		{ 411, 162, 42 },
		{ 411, 163, 42 },
		{ 411, 164, 42 },
		{ 411, 165, 42 },
		{ 411, 166, 42 },
		{ 411, 167, 42 },
		{ 411, 168, 42 },
		{ 411, 169, 42 },
		{ 411, 170, 42 },
		{ 411, 171, 42 },
		{ 411, 172, 42 },
		{ 411, 173, 42 },
		{ 411, 174, 42 },
		{ 411, 175, 42 },
		{ 411, 176, 42 },
		{ 411, 177, 42 },
		{ 411, 178, 42 },
		{ 411, 179, 42 },
		{ 411, 180, 42 },
		{ 411, 181, 42 },
		{ 411, 182, 42 },
		{ 411, 183, 42 },
		{ 411, 184, 42 },
		{ 411, 185, 42 },
		{ 411, 186, 42 },
		{ 411, 187, 42 },
		{ 411, 188, 42 },
		{ 411, 189, 42 },
		{ 411, 190, 42 },
		{ 411, 191, 42 },
		{ 411, 192, 42 },
		{ 411, 193, 42 },
		{ 411, 194, 42 },
		{ 411, 195, 42 },
		{ 411, 196, 42 },
		{ 411, 197, 42 },
		{ 411, 198, 42 },
		{ 411, 199, 42 },
		{ 412, 161, 42 },
		{ 412, 162, 42 },
		{ 412, 163, 42 },
		{ 412, 164, 42 },
		{ 412, 165, 42 },
		{ 412, 166, 42 },
		{ 412, 167, 42 },
		{ 412, 168, 42 },
		{ 412, 169, 42 },
		{ 412, 170, 42 },
		{ 412, 171, 42 },
		{ 412, 172, 42 },
		{ 412, 173, 42 },
		{ 412, 174, 42 },
		{ 412, 175, 42 },
		{ 412, 176, 42 },
		{ 412, 177, 42 },
		{ 412, 178, 42 },
		{ 412, 179, 42 },
		{ 412, 180, 42 },
		{ 412, 181, 42 },
		{ 412, 182, 42 },
		{ 412, 183, 42 },
		{ 412, 184, 42 },
		{ 412, 185, 42 },
		{ 412, 186, 42 },
		{ 412, 187, 42 },
		{ 412, 188, 42 },
		{ 412, 189, 42 },
		{ 412, 190, 42 },
		{ 412, 191, 42 },
		{ 412, 192, 42 },
		{ 412, 193, 42 },
		{ 412, 194, 42 },
		{ 412, 195, 42 },
		{ 412, 196, 42 },
		{ 412, 197, 42 },
		{ 412, 198, 42 },
		{ 413, 160, 42 },
		{ 413, 161, 42 },
		{ 413, 162, 42 },
		{ 413, 163, 42 },
		{ 413, 164, 42 },
		{ 413, 165, 42 },
		{ 413, 166, 42 },
		{ 413, 167, 42 },
		{ 413, 168, 42 },
		{ 413, 169, 42 },
		{ 413, 170, 42 },
		{ 413, 171, 42 },
		{ 413, 172, 42 },
		{ 413, 173, 42 },
		{ 413, 174, 42 },
		{ 413, 175, 42 },
		{ 413, 176, 42 },
		{ 413, 177, 42 },
		{ 413, 178, 42 },
		{ 413, 179, 42 },
		{ 413, 180, 42 },
		{ 413, 181, 42 },
		{ 413, 182, 42 },
		{ 413, 183, 42 },
		{ 413, 184, 42 },
		{ 413, 185, 42 },
		{ 413, 186, 42 },
		{ 413, 187, 42 },
		{ 413, 188, 42 },
		{ 413, 189, 42 },
		{ 413, 190, 42 },
		{ 413, 191, 42 },
		{ 413, 192, 42 },
		{ 413, 193, 42 },
		{ 413, 194, 42 },
		{ 413, 195, 42 },
		{ 413, 196, 42 },
		{ 413, 197, 42 },
		{ 413, 198, 42 },
		{ 414, 159, 42 },
		{ 414, 160, 42 },
		{ 414, 161, 42 },
		{ 414, 162, 42 },
		{ 414, 163, 42 },
		{ 414, 164, 42 },
		{ 414, 165, 42 },
		{ 414, 166, 42 },
		{ 414, 167, 42 },
		{ 414, 168, 42 },
		{ 414, 169, 42 },
		{ 414, 170, 42 },
		{ 414, 171, 42 },
		{ 414, 172, 42 },
		{ 414, 173, 42 },
		{ 414, 174, 42 },
		{ 414, 175, 42 },
		{ 414, 176, 42 },
		{ 414, 177, 42 },
		{ 414, 178, 42 },
		{ 414, 179, 42 },
		{ 414, 180, 42 },
		{ 414, 181, 42 },
		{ 414, 182, 42 },
		{ 414, 183, 42 },
		{ 414, 184, 42 },
		{ 414, 185, 42 },
		{ 414, 186, 42 },
		{ 414, 187, 42 },
		{ 414, 188, 42 },
		{ 414, 189, 42 },
		{ 414, 190, 42 },
		{ 414, 191, 42 },
		{ 414, 192, 42 },
		{ 414, 193, 42 },
		{ 414, 194, 42 },
		{ 414, 195, 42 },
		{ 414, 196, 42 },
		{ 414, 197, 42 },
		{ 414, 198, 42 },
		{ 415, 159, 42 },
		{ 415, 160, 42 },
		{ 415, 161, 42 },
		{ 415, 162, 42 },
		{ 415, 163, 42 },
		{ 415, 164, 42 },
		{ 415, 165, 42 },
		{ 415, 166, 42 },
		{ 415, 167, 42 },
		{ 415, 168, 42 },
		{ 415, 169, 42 },
		{ 415, 170, 42 },
		{ 415, 171, 42 },
		{ 415, 172, 42 },
		{ 415, 173, 42 },
		{ 415, 174, 42 },
		{ 415, 175, 42 },
		{ 415, 176, 42 },
		{ 415, 177, 42 },
		{ 415, 178, 42 },
		{ 415, 179, 42 },
		{ 415, 180, 42 },
		{ 415, 181, 42 },
		{ 415, 182, 42 },
		{ 415, 183, 42 },
		{ 415, 184, 42 },
		{ 415, 185, 42 },
		{ 415, 186, 42 },
		{ 415, 187, 42 },
		{ 415, 188, 42 },
		{ 415, 189, 42 },
		{ 415, 190, 42 },
		{ 415, 191, 42 },
		{ 415, 192, 42 },
		{ 415, 193, 42 },
		{ 415, 194, 42 },
		{ 415, 195, 42 },
		{ 415, 196, 42 },
		{ 415, 197, 42 },
		{ 416, 158, 42 },
		{ 416, 159, 42 },
		{ 416, 160, 42 },
		{ 416, 161, 42 },
		{ 416, 162, 42 },
		{ 416, 163, 42 },
		{ 416, 164, 42 },
		{ 416, 165, 42 },
		{ 416, 166, 42 },
		{ 416, 167, 42 },
		{ 416, 168, 42 },
		{ 416, 169, 42 },
		{ 416, 170, 42 },
		{ 416, 171, 42 },
		{ 416, 172, 42 },
		{ 416, 173, 42 },
		{ 416, 174, 42 },
		{ 416, 175, 42 },
		{ 416, 176, 42 },
		{ 416, 177, 42 },
		{ 416, 178, 42 },
		{ 416, 179, 42 },
		{ 416, 180, 42 },
		{ 416, 181, 42 },
		{ 416, 182, 42 },
		{ 416, 183, 42 },
		{ 416, 184, 42 },
		{ 416, 185, 42 },
		{ 416, 186, 42 },
		{ 416, 187, 42 },
		{ 416, 188, 42 },
		{ 416, 189, 42 },
		{ 416, 190, 42 },
		{ 416, 191, 42 },
		{ 416, 192, 42 },
		{ 416, 193, 42 },
		{ 416, 194, 42 },
		{ 416, 195, 42 },
		{ 416, 196, 42 },
		{ 416, 197, 42 },
		{ 417, 158, 42 },
		{ 417, 159, 42 },
		{ 417, 160, 42 },
		{ 417, 161, 42 },
		{ 417, 162, 42 },
		{ 417, 163, 42 },
		{ 417, 164, 42 },
		{ 417, 165, 42 },
		{ 417, 166, 42 },
		{ 417, 167, 42 },
		{ 417, 168, 42 },
		{ 417, 169, 42 },
		{ 417, 170, 42 },
		{ 417, 171, 42 },
		{ 417, 172, 42 },
		{ 417, 173, 42 },
		{ 417, 174, 42 },
		{ 417, 175, 42 },
		{ 417, 176, 42 },
		{ 417, 177, 42 },
		{ 417, 178, 42 },
		{ 417, 179, 42 },
		{ 417, 180, 42 },
		{ 417, 181, 42 },
		{ 417, 182, 42 },
		{ 417, 183, 42 },
		{ 417, 184, 42 },
		{ 417, 185, 42 },
		{ 417, 186, 42 },
		{ 417, 187, 42 },
		{ 417, 188, 42 },
		{ 417, 189, 42 },
		{ 417, 190, 42 },
		{ 417, 191, 42 },
		{ 417, 192, 42 },
		{ 417, 193, 42 },
		{ 417, 194, 42 },
		{ 417, 195, 42 },
		{ 417, 196, 42 },
		{ 418, 157, 42 },
		{ 418, 158, 42 },
		{ 418, 159, 42 },
		{ 418, 160, 42 },
		{ 418, 161, 42 },
		{ 418, 162, 42 },
		{ 418, 163, 42 },
		{ 418, 164, 42 },
		{ 418, 165, 42 },
		{ 418, 166, 42 },
		{ 418, 167, 42 },
		{ 418, 168, 42 },
		{ 418, 169, 42 },
		{ 418, 170, 42 },
		{ 418, 171, 42 },
		{ 418, 172, 42 },
		{ 418, 173, 42 },
		{ 418, 174, 42 },
		{ 418, 175, 42 },
		{ 418, 176, 42 },
		{ 418, 177, 42 },
		{ 418, 178, 42 },
		{ 418, 179, 42 },
		{ 418, 180, 42 },
		{ 418, 181, 42 },
		{ 418, 182, 42 },
		{ 418, 183, 42 },
		{ 418, 184, 42 },
		{ 418, 185, 42 },
		{ 418, 186, 42 },
		{ 418, 187, 42 },
		{ 418, 188, 42 },
		{ 418, 189, 42 },
		{ 418, 190, 42 },
		{ 418, 191, 42 },
		{ 418, 192, 42 },
		{ 418, 193, 42 },
		{ 418, 194, 42 },
		{ 418, 195, 42 },
		{ 418, 196, 42 },
		{ 419, 156, 42 },
		{ 419, 157, 42 },
		{ 419, 158, 42 },
		{ 419, 159, 42 },
		{ 419, 160, 42 },
		{ 419, 161, 42 },
		{ 419, 162, 42 },
		{ 419, 163, 42 },
		{ 419, 164, 42 },
		{ 419, 165, 42 },
		{ 419, 166, 42 },
		{ 419, 167, 42 },
		{ 419, 168, 42 },
		{ 419, 169, 42 },
		{ 419, 170, 42 },
		{ 419, 171, 42 },
		{ 419, 172, 42 },
		{ 419, 173, 42 },
		{ 419, 174, 42 },
		{ 419, 175, 42 },
		{ 419, 176, 42 },
		{ 419, 177, 42 },
		{ 419, 178, 42 },
		{ 419, 179, 42 },
		{ 419, 180, 42 },
		{ 419, 181, 42 },
		{ 419, 182, 42 },
		{ 419, 183, 42 },
		{ 419, 184, 42 },
		{ 419, 185, 42 },
		{ 419, 186, 42 },
		{ 419, 187, 42 },
		{ 419, 188, 42 },
		{ 419, 189, 42 },
		{ 419, 190, 42 },
		{ 419, 191, 42 },
		{ 419, 192, 42 },
		{ 419, 193, 42 },
		{ 419, 194, 42 },
		{ 419, 195, 42 },
		{ 419, 196, 42 },
		{ 420, 156, 42 },
		{ 420, 157, 42 },
		{ 420, 158, 42 },
		{ 420, 159, 42 },
		{ 420, 160, 42 },
		{ 420, 161, 42 },
		{ 420, 162, 42 },
		{ 420, 163, 42 },
		{ 420, 164, 42 },
		{ 420, 165, 42 },
		{ 420, 166, 42 },
		{ 420, 167, 42 },
		{ 420, 168, 42 },
		{ 420, 169, 42 },
		{ 420, 170, 42 },
		{ 420, 171, 42 },
		{ 420, 172, 42 },
		{ 420, 173, 42 },
		{ 420, 174, 42 },
		{ 420, 175, 42 },
		{ 420, 176, 42 },
		{ 420, 177, 42 },
		{ 420, 178, 42 },
		{ 420, 179, 42 },
		{ 420, 180, 42 },
		{ 420, 181, 42 },
		{ 420, 182, 42 },
		{ 420, 183, 42 },
		{ 420, 184, 42 },
		{ 420, 185, 42 },
		{ 420, 186, 42 },
		{ 420, 187, 42 },
		{ 420, 188, 42 },
		{ 420, 189, 42 },
		{ 420, 190, 42 },
		{ 420, 191, 42 },
		{ 420, 192, 42 },
		{ 420, 193, 42 },
		{ 420, 194, 42 },
		{ 420, 195, 42 },
		{ 421, 155, 42 },
		{ 421, 156, 42 },
		{ 421, 157, 42 },
		{ 421, 158, 42 },
		{ 421, 159, 42 },
		{ 421, 160, 42 },
		{ 421, 161, 42 },
		{ 421, 162, 42 },
		{ 421, 163, 42 },
		{ 421, 164, 42 },
		{ 421, 165, 42 },
		{ 421, 166, 42 },
		{ 421, 167, 42 },
		{ 421, 168, 42 },
		{ 421, 169, 42 },
		{ 421, 170, 42 },
		{ 421, 171, 42 },
		{ 421, 172, 42 },
		{ 421, 173, 42 },
		{ 421, 174, 42 },
		{ 421, 175, 42 },
		{ 421, 176, 42 },
		{ 421, 177, 42 },
		{ 421, 178, 42 },
		{ 421, 179, 42 },
		{ 421, 180, 42 },
		{ 421, 181, 42 },
		{ 421, 182, 42 },
		{ 421, 183, 42 },
		{ 421, 184, 42 },
		{ 421, 185, 42 },
		{ 421, 186, 42 },
		{ 421, 187, 42 },
		{ 421, 188, 42 },
		{ 421, 189, 42 },
		{ 421, 190, 42 },
		{ 421, 191, 42 },
		{ 421, 192, 42 },
		{ 421, 193, 42 },
		{ 421, 194, 42 },
		{ 421, 195, 42 },
		{ 422, 154, 42 },
		{ 422, 155, 42 },
		{ 422, 156, 42 },
		{ 422, 157, 42 },
		{ 422, 158, 42 },
		{ 422, 159, 42 },
		{ 422, 160, 42 },
		{ 422, 161, 42 },
		{ 422, 162, 42 },
		{ 422, 163, 42 },
		{ 422, 164, 42 },
		{ 422, 165, 42 },
		{ 422, 166, 42 },
		{ 422, 167, 42 },
		{ 422, 168, 42 },
		{ 422, 169, 42 },
		{ 422, 170, 42 },
		{ 422, 171, 42 },
		{ 422, 172, 42 },
		{ 422, 173, 42 },
		{ 422, 174, 42 },
		{ 422, 175, 42 },
		{ 422, 176, 42 },
		{ 422, 177, 42 },
		{ 422, 178, 42 },
		{ 422, 179, 42 },
		{ 422, 180, 42 },
		{ 422, 181, 42 },
		{ 422, 182, 42 },
		{ 422, 183, 42 },
		{ 422, 184, 42 },
		{ 422, 185, 42 },
		{ 422, 186, 42 },
		{ 422, 187, 42 },
		{ 422, 188, 42 },
		{ 422, 189, 42 },
		{ 422, 190, 42 },
		{ 422, 191, 42 },
		{ 422, 192, 42 },
		{ 422, 193, 42 },
		{ 422, 194, 42 },
		{ 422, 195, 42 },
		{ 423, 154, 42 },
		{ 423, 155, 42 },
		{ 423, 156, 42 },
		{ 423, 157, 42 },
		{ 423, 158, 42 },
		{ 423, 159, 42 },
		{ 423, 160, 42 },
		{ 423, 161, 42 },
		{ 423, 162, 42 },
		{ 423, 163, 42 },
		{ 423, 164, 42 },
		{ 423, 165, 42 },
		{ 423, 166, 42 },
		{ 423, 167, 42 },
		{ 423, 168, 42 },
		{ 423, 169, 42 },
		{ 423, 170, 42 },
		{ 423, 171, 42 },
		{ 423, 172, 42 },
		{ 423, 173, 42 },
		{ 423, 174, 42 },
		{ 423, 175, 42 },
		{ 423, 176, 42 },
		{ 423, 177, 42 },
		{ 423, 178, 42 },
		{ 423, 179, 42 },
		{ 423, 180, 42 },
		{ 423, 181, 42 },
		{ 423, 182, 42 },
		{ 423, 183, 42 },
		{ 423, 184, 42 },
		{ 423, 185, 42 },
		{ 423, 186, 42 },
		{ 423, 187, 42 },
		{ 423, 188, 42 },
		{ 423, 189, 42 },
		{ 423, 190, 42 },
		{ 423, 191, 42 },
		{ 423, 192, 42 },
		{ 423, 193, 42 },
		{ 423, 194, 42 },
		{ 424, 153, 42 },
		{ 424, 154, 42 },
		{ 424, 155, 42 },
		{ 424, 156, 42 },
		{ 424, 157, 42 },
		{ 424, 158, 42 },
		{ 424, 159, 42 },
		{ 424, 160, 42 },
		{ 424, 161, 42 },
		{ 424, 162, 42 },
		{ 424, 163, 42 },
		{ 424, 164, 42 },
		{ 424, 165, 42 },
		{ 424, 166, 42 },
		{ 424, 167, 42 },
		{ 424, 168, 42 },
		{ 424, 169, 42 },
		{ 424, 170, 42 },
		{ 424, 171, 42 },
		{ 424, 172, 42 },
		{ 424, 173, 42 },
		{ 424, 174, 42 },
		{ 424, 175, 42 },
		{ 424, 176, 42 },
		{ 424, 177, 42 },
		{ 424, 178, 42 },
		{ 424, 179, 42 },
		{ 424, 180, 42 },
		{ 424, 181, 42 },
		{ 424, 182, 42 },
		{ 424, 183, 42 },
		{ 424, 184, 42 },
		{ 424, 185, 42 },
		{ 424, 186, 42 },
		{ 424, 187, 42 },
		{ 424, 188, 42 },
		{ 424, 189, 42 },
		{ 424, 190, 42 },
		{ 424, 191, 42 },
		{ 424, 192, 42 },
		{ 424, 193, 42 },
		{ 424, 194, 42 },
		{ 425, 153, 42 },
		{ 425, 154, 42 },
		{ 425, 155, 42 },
		{ 425, 156, 42 },
		{ 425, 157, 42 },
		{ 425, 158, 42 },
		{ 425, 159, 42 },
		{ 425, 160, 42 },
		{ 425, 161, 42 },
		{ 425, 162, 42 },
		{ 425, 163, 42 },
		{ 425, 164, 42 },
		{ 425, 165, 42 },
		{ 425, 166, 42 },
		{ 425, 167, 42 },
		{ 425, 168, 42 },
		{ 425, 169, 42 },
		{ 425, 170, 42 },
		{ 425, 171, 42 },
		{ 425, 172, 42 },
		{ 425, 173, 42 },
		{ 425, 174, 42 },
		{ 425, 175, 42 },
		{ 425, 176, 42 },
		{ 425, 177, 42 },
		{ 425, 178, 42 },
		{ 425, 179, 42 },
		{ 425, 180, 42 },
		{ 425, 181, 42 },
		{ 425, 182, 42 },
		{ 425, 183, 42 },
		{ 425, 184, 42 },
		{ 425, 185, 42 },
		{ 425, 186, 42 },
		{ 425, 187, 42 },
		{ 425, 188, 42 },
		{ 425, 189, 42 },
		{ 425, 190, 42 },
		{ 425, 191, 42 },
		{ 425, 192, 42 },
		{ 425, 193, 42 },
		{ 426, 152, 42 },
		{ 426, 153, 42 },
		{ 426, 154, 42 },
		{ 426, 155, 42 },
		{ 426, 156, 42 },
		{ 426, 157, 42 },
		{ 426, 158, 42 },
		{ 426, 159, 42 },
		{ 426, 160, 42 },
		{ 426, 161, 42 },
		{ 426, 162, 42 },
		{ 426, 163, 42 },
		{ 426, 164, 42 },
		{ 426, 165, 42 },
		{ 426, 166, 42 },
		{ 426, 167, 42 },
		{ 426, 168, 42 },
		{ 426, 169, 42 },
		{ 426, 170, 42 },
		{ 426, 171, 42 },
		{ 426, 172, 42 },
		{ 426, 173, 42 },
		{ 426, 174, 42 },
		{ 426, 175, 42 },
		{ 426, 176, 42 },
		{ 426, 177, 42 },
		{ 426, 178, 42 },
		{ 426, 179, 42 },
		{ 426, 180, 42 },
		{ 426, 181, 42 },
		{ 426, 182, 42 },
		{ 426, 183, 42 },
		{ 426, 184, 42 },
		{ 426, 185, 42 },
		{ 426, 186, 42 },
		{ 426, 187, 42 },
		{ 426, 188, 42 },
		{ 426, 189, 42 },
		{ 426, 190, 42 },
		{ 426, 191, 42 },
		{ 426, 192, 42 },
		{ 426, 193, 42 },
		{ 427, 151, 42 },
		{ 427, 152, 42 },
		{ 427, 153, 42 },
		{ 427, 154, 42 },
		{ 427, 155, 42 },
		{ 427, 156, 42 },
		{ 427, 157, 42 },
		{ 427, 158, 42 },
		{ 427, 159, 42 },
		{ 427, 160, 42 },
		{ 427, 161, 42 },
		{ 427, 162, 42 },
		{ 427, 163, 42 },
		{ 427, 164, 42 },
		{ 427, 165, 42 },
		{ 427, 166, 42 },
		{ 427, 167, 42 },
		{ 427, 168, 42 },
		{ 427, 169, 42 },
		{ 427, 170, 42 },
		{ 427, 171, 42 },
		{ 427, 172, 42 },
		{ 427, 173, 42 },
		{ 427, 174, 42 },
		{ 427, 175, 42 },
		{ 427, 176, 42 },
		{ 427, 177, 42 },
		{ 427, 178, 42 },
		{ 427, 179, 42 },
		{ 427, 180, 42 },
		{ 427, 181, 42 },
		{ 427, 182, 42 },
		{ 427, 183, 42 },
		{ 427, 184, 42 },
		{ 427, 185, 42 },
		{ 427, 186, 42 },
		{ 427, 187, 42 },
		{ 427, 188, 42 },
		{ 427, 189, 42 },
		{ 427, 190, 42 },
		{ 427, 191, 42 },
		{ 427, 192, 42 },
		{ 427, 193, 42 },
		{ 428, 151, 42 },
		{ 428, 152, 42 },
		{ 428, 153, 42 },
		{ 428, 154, 42 },
		{ 428, 155, 42 },
		{ 428, 156, 42 },
		{ 428, 157, 42 },
		{ 428, 158, 42 },
		{ 428, 159, 42 },
		{ 428, 160, 42 },
		{ 428, 161, 42 },
		{ 428, 162, 42 },
		{ 428, 163, 42 },
		{ 428, 164, 42 },
		{ 428, 165, 42 },
		{ 428, 166, 42 },
		{ 428, 167, 42 },
		{ 428, 168, 42 },
		{ 428, 169, 42 },
		{ 428, 170, 42 },
		{ 428, 171, 42 },
		{ 428, 172, 42 },
		{ 428, 173, 42 },
		{ 428, 174, 42 },
		{ 428, 175, 42 },
		{ 428, 176, 42 },
		{ 428, 177, 42 },
		{ 428, 178, 42 },
		{ 428, 179, 42 },
		{ 428, 180, 42 },
		{ 428, 181, 42 },
		{ 428, 182, 42 },
		{ 428, 183, 42 },
		{ 428, 184, 42 },
		{ 428, 185, 42 },
		{ 428, 186, 42 },
		{ 428, 187, 42 },
		{ 428, 188, 42 },
		{ 428, 189, 42 },
		{ 428, 190, 42 },
		{ 428, 191, 42 },
		{ 428, 192, 42 },
		{ 429, 150, 42 },
		{ 429, 151, 42 },
		{ 429, 152, 42 },
		{ 429, 153, 42 },
		{ 429, 154, 42 },
		{ 429, 155, 42 },
		{ 429, 156, 42 },
		{ 429, 157, 42 },
		{ 429, 158, 42 },
		{ 429, 159, 42 },
		{ 429, 160, 42 },
		{ 429, 161, 42 },
		{ 429, 162, 42 },
		{ 429, 163, 42 },
		{ 429, 164, 42 },
		{ 429, 165, 42 },
		{ 429, 166, 42 },
		{ 429, 167, 42 },
		{ 429, 168, 42 },
		{ 429, 169, 42 },
		{ 429, 170, 42 },
		{ 429, 171, 42 },
		{ 429, 172, 42 },
		{ 429, 173, 42 },
		{ 429, 174, 42 },
		{ 429, 175, 42 },
		{ 429, 176, 42 },
		{ 429, 177, 42 },
		{ 429, 178, 42 },
		{ 429, 179, 42 },
		{ 429, 180, 42 },
		{ 429, 181, 42 },
		{ 429, 182, 42 },
		{ 429, 183, 42 },
		{ 429, 184, 42 },
		{ 429, 185, 42 },
		{ 429, 186, 42 },
		{ 429, 187, 42 },
		{ 429, 188, 42 },
		{ 429, 189, 42 },
		{ 429, 190, 42 },
		{ 429, 191, 42 },
		{ 429, 192, 42 },
		{ 430, 149, 42 },
		{ 430, 150, 42 },
		{ 430, 151, 42 },
		{ 430, 152, 42 },
		{ 430, 153, 42 },
		{ 430, 154, 42 },
		{ 430, 155, 42 },
		{ 430, 156, 42 },
		{ 430, 157, 42 },
		{ 430, 158, 42 },
		{ 430, 159, 42 },
		{ 430, 160, 42 },
		{ 430, 161, 42 },
		{ 430, 162, 42 },
		{ 430, 163, 42 },
		{ 430, 164, 42 },
		{ 430, 165, 42 },
		{ 430, 166, 42 },
		{ 430, 167, 42 },
		{ 430, 168, 42 },
		{ 430, 169, 42 },
		{ 430, 170, 42 },
		{ 430, 171, 42 },
		{ 430, 172, 42 },
		{ 430, 173, 42 },
		{ 430, 174, 42 },
		{ 430, 175, 42 },
		{ 430, 176, 42 },
		{ 430, 177, 42 },
		{ 430, 178, 42 },
		{ 430, 179, 42 },
		{ 430, 180, 42 },
		{ 430, 181, 42 },
		{ 430, 182, 42 },
		{ 430, 183, 42 },
		{ 430, 184, 42 },
		{ 430, 185, 42 },
		{ 430, 186, 42 },
		{ 430, 187, 42 },
		{ 430, 188, 42 },
		{ 430, 189, 42 },
		{ 430, 190, 42 },
		{ 430, 191, 42 },
		{ 430, 192, 42 },
		{ 431, 149, 42 },
		{ 431, 150, 42 },
		{ 431, 151, 42 },
		{ 431, 152, 42 },
		{ 431, 153, 42 },
		{ 431, 154, 42 },
		{ 431, 155, 42 },
		{ 431, 156, 42 },
		{ 431, 157, 42 },
		{ 431, 158, 42 },
		{ 431, 159, 42 },
		{ 431, 160, 42 },
		{ 431, 161, 42 },
		{ 431, 162, 42 },
		{ 431, 163, 42 },
		{ 431, 164, 42 },
		{ 431, 165, 42 },
		{ 431, 166, 42 },
		{ 431, 167, 42 },
		{ 431, 168, 42 },
		{ 431, 169, 42 },
		{ 431, 170, 42 },
		{ 431, 171, 42 },
		{ 431, 172, 42 },
		{ 431, 173, 42 },
		{ 431, 174, 42 },
		{ 431, 175, 42 },
		{ 431, 176, 42 },
		{ 431, 177, 42 },
		{ 431, 178, 42 },
		{ 431, 179, 42 },
		{ 431, 180, 42 },
		{ 431, 181, 42 },
		{ 431, 182, 42 },
		{ 431, 183, 42 },
		{ 431, 184, 42 },
		{ 431, 185, 42 },
		{ 431, 186, 42 },
		{ 431, 187, 42 },
		{ 431, 188, 42 },
		{ 431, 189, 42 },
		{ 431, 190, 42 },
		{ 431, 191, 42 },
		{ 432, 148, 42 },
		{ 432, 149, 42 },
		{ 432, 150, 42 },
		{ 432, 151, 42 },
		{ 432, 152, 42 },
		{ 432, 153, 42 },
		{ 432, 154, 42 },
		{ 432, 155, 42 },
		{ 432, 156, 42 },
		{ 432, 157, 42 },
		{ 432, 158, 42 },
		{ 432, 159, 42 },
		{ 432, 160, 42 },
		{ 432, 161, 42 },
		{ 432, 162, 42 },
		{ 432, 163, 42 },
		{ 432, 164, 42 },
		{ 432, 165, 42 },
		{ 432, 166, 42 },
		{ 432, 167, 42 },
		{ 432, 168, 42 },
		{ 432, 169, 42 },
		{ 432, 170, 42 },
		{ 432, 171, 42 },
		{ 432, 172, 42 },
		{ 432, 173, 42 },
		{ 432, 174, 42 },
		{ 432, 175, 42 },
		{ 432, 176, 42 },
		{ 432, 177, 42 },
		{ 432, 178, 42 },
		{ 432, 179, 42 },
		{ 432, 180, 42 },
		{ 432, 181, 42 },
		{ 432, 182, 42 },
		{ 432, 183, 42 },
		{ 432, 184, 42 },
		{ 432, 185, 42 },
		{ 432, 186, 42 },
		{ 432, 187, 42 },
		{ 432, 188, 42 },
		{ 432, 189, 42 },
		{ 432, 190, 42 },
		{ 432, 191, 42 },
		{ 433, 148, 42 },
		{ 433, 149, 42 },
		{ 433, 150, 42 },
		{ 433, 151, 42 },
		{ 433, 152, 42 },
		{ 433, 153, 42 },
		{ 433, 154, 42 },
		{ 433, 155, 42 },
		{ 433, 156, 42 },
		{ 433, 157, 42 },
		{ 433, 158, 42 },
		{ 433, 159, 42 },
		{ 433, 160, 42 },
		{ 433, 161, 42 },
		{ 433, 162, 42 },
		{ 433, 163, 42 },
		{ 433, 164, 42 },
		{ 433, 165, 42 },
		{ 433, 166, 42 },
		{ 433, 167, 42 },
		{ 433, 168, 42 },
		{ 433, 169, 42 },
		{ 433, 170, 42 },
		{ 433, 171, 42 },
		{ 433, 172, 42 },
		{ 433, 173, 42 },
		{ 433, 174, 42 },
		{ 433, 175, 42 },
		{ 433, 176, 42 },
		{ 433, 177, 42 },
		{ 433, 178, 42 },
		{ 433, 179, 42 },
		{ 433, 180, 42 },
		{ 433, 181, 42 },
		{ 433, 182, 42 },
		{ 433, 183, 42 },
		{ 433, 184, 42 },
		{ 433, 185, 42 },
		{ 433, 186, 42 },
		{ 433, 187, 42 },
		{ 433, 188, 42 },
		{ 433, 189, 42 },
		{ 433, 190, 42 },
		{ 434, 147, 42 },
		{ 434, 148, 42 },
		{ 434, 149, 42 },
		{ 434, 150, 42 },
		{ 434, 151, 42 },
		{ 434, 152, 42 },
		{ 434, 153, 42 },
		{ 434, 154, 42 },
		{ 434, 155, 42 },
		{ 434, 156, 42 },
		{ 434, 157, 42 },
		{ 434, 158, 42 },
		{ 434, 159, 42 },
		{ 434, 160, 42 },
		{ 434, 161, 42 },
		{ 434, 162, 42 },
		{ 434, 163, 42 },
		{ 434, 164, 42 },
		{ 434, 165, 42 },
		{ 434, 166, 42 },
		{ 434, 167, 42 },
		{ 434, 168, 42 },
		{ 434, 169, 42 },
		{ 434, 170, 42 },
		{ 434, 171, 42 },
		{ 434, 172, 42 },
		{ 434, 173, 42 },
		{ 434, 174, 42 },
		{ 434, 175, 42 },
		{ 434, 176, 42 },
		{ 434, 177, 42 },
		{ 434, 178, 42 },
		{ 434, 179, 42 },
		{ 434, 180, 42 },
		{ 434, 181, 42 },
		{ 434, 182, 42 },
		{ 434, 183, 42 },
		{ 434, 184, 42 },
		{ 434, 185, 42 },
		{ 434, 186, 42 },
		{ 434, 187, 42 },
		{ 434, 188, 42 },
		{ 434, 189, 42 },
		{ 434, 190, 42 },
		{ 435, 146, 42 },
		{ 435, 147, 42 },
		{ 435, 148, 42 },
		{ 435, 149, 42 },
		{ 435, 150, 42 },
		{ 435, 151, 42 },
		{ 435, 152, 42 },
		{ 435, 153, 42 },
		{ 435, 154, 42 },
		{ 435, 155, 42 },
		{ 435, 156, 42 },
		{ 435, 157, 42 },
		{ 435, 158, 42 },
		{ 435, 159, 42 },
		{ 435, 160, 42 },
		{ 435, 161, 42 },
		{ 435, 162, 42 },
		{ 435, 163, 42 },
		{ 435, 164, 42 },
		{ 435, 165, 42 },
		{ 435, 166, 42 },
		{ 435, 167, 42 },
		{ 435, 168, 42 },
		{ 435, 169, 42 },
		{ 435, 170, 42 },
		{ 435, 171, 42 },
		{ 435, 172, 42 },
		{ 435, 173, 42 },
		{ 435, 174, 42 },
		{ 435, 175, 42 },
		{ 435, 176, 42 },
		{ 435, 177, 42 },
		{ 435, 178, 42 },
		{ 435, 179, 42 },
		{ 435, 180, 42 },
		{ 435, 181, 42 },
		{ 435, 182, 42 },
		{ 435, 183, 42 },
		{ 435, 184, 42 },
		{ 435, 185, 42 },
		{ 435, 186, 42 },
		{ 435, 187, 42 },
		{ 435, 188, 42 },
		{ 435, 189, 42 },
		{ 435, 190, 42 },
		{ 436, 146, 42 },
		{ 436, 147, 42 },
		{ 436, 148, 42 },
		{ 436, 149, 42 },
		{ 436, 150, 42 },
		{ 436, 151, 42 },
		{ 436, 152, 42 },
		{ 436, 153, 42 },
		{ 436, 154, 42 },
		{ 436, 155, 42 },
		{ 436, 156, 42 },
		{ 436, 157, 42 },
		{ 436, 158, 42 },
		{ 436, 159, 42 },
		{ 436, 160, 42 },
		{ 436, 161, 42 },
		{ 436, 162, 42 },
		{ 436, 163, 42 },
		{ 436, 164, 42 },
		{ 436, 165, 42 },
		{ 436, 166, 42 },
		{ 436, 167, 42 },
		{ 436, 168, 42 },
		{ 436, 169, 42 },
		{ 436, 170, 42 },
		{ 436, 171, 42 },
		{ 436, 172, 42 },
		{ 436, 173, 42 },
		{ 436, 174, 42 },
		{ 436, 175, 42 },
		{ 436, 176, 42 },
		{ 436, 177, 42 },
		{ 436, 178, 42 },
		{ 436, 179, 42 },
		{ 436, 180, 42 },
		{ 436, 181, 42 },
		{ 436, 182, 42 },
		{ 436, 183, 42 },
		{ 436, 184, 42 },
		{ 436, 185, 42 },
		{ 436, 186, 42 },
		{ 436, 187, 42 },
		{ 436, 188, 42 },
		{ 436, 189, 42 },
		{ 437, 145, 42 },
		{ 437, 146, 42 },
		{ 437, 147, 42 },
		{ 437, 148, 42 },
		{ 437, 149, 42 },
		{ 437, 150, 42 },
		{ 437, 151, 42 },
		{ 437, 152, 42 },
		{ 437, 153, 42 },
		{ 437, 154, 42 },
		{ 437, 155, 42 },
		{ 437, 156, 42 },
		{ 437, 157, 42 },
		{ 437, 158, 42 },
		{ 437, 159, 42 },
		{ 437, 160, 42 },
		{ 437, 161, 42 },
		{ 437, 162, 42 },
		{ 437, 163, 42 },
		{ 437, 164, 42 },
		{ 437, 165, 42 },
		{ 437, 166, 42 },
		{ 437, 167, 42 },
		{ 437, 168, 42 },
		{ 437, 169, 42 },
		{ 437, 170, 42 },
		{ 437, 171, 42 },
		{ 437, 172, 42 },
		{ 437, 173, 42 },
		{ 437, 174, 42 },
		{ 437, 175, 42 },
		{ 437, 176, 42 },
		{ 437, 177, 42 },
		{ 437, 178, 42 },
		{ 437, 179, 42 },
		{ 437, 180, 42 },
		{ 437, 181, 42 },
		{ 437, 182, 42 },
		{ 437, 183, 42 },
		{ 437, 184, 42 },
		{ 437, 185, 42 },
		{ 437, 186, 42 },
		{ 437, 187, 42 },
		{ 437, 188, 42 },
		{ 437, 189, 42 },
		{ 438, 144, 42 },
		{ 438, 145, 42 },
		{ 438, 146, 42 },
		{ 438, 147, 42 },
		{ 438, 148, 42 },
		{ 438, 149, 42 },
		{ 438, 150, 42 },
		{ 438, 151, 42 },
		{ 438, 152, 42 },
		{ 438, 153, 42 },
		{ 438, 154, 42 },
		{ 438, 155, 42 },
		{ 438, 156, 42 },
		{ 438, 157, 42 },
		{ 438, 158, 42 },
		{ 438, 159, 42 },
		{ 438, 160, 42 },
		{ 438, 161, 42 },
		{ 438, 162, 42 },
		{ 438, 163, 42 },
		{ 438, 164, 42 },
		{ 438, 165, 42 },
		{ 438, 166, 42 },
		{ 438, 167, 42 },
		{ 438, 168, 42 },
		{ 438, 169, 42 },
		{ 438, 170, 42 },
		{ 438, 171, 42 },
		{ 438, 172, 42 },
		{ 438, 173, 42 },
		{ 438, 174, 42 },
		{ 438, 175, 42 },
		{ 438, 176, 42 },
		{ 438, 177, 42 },
		{ 438, 178, 42 },
		{ 438, 179, 42 },
		{ 438, 180, 42 },
		{ 438, 181, 42 },
		{ 438, 182, 42 },
		{ 438, 183, 42 },
		{ 438, 184, 42 },
		{ 438, 185, 42 },
		{ 438, 186, 42 },
		{ 438, 187, 42 },
		{ 438, 188, 42 },
		{ 438, 189, 42 },
		{ 439, 144, 42 },
		{ 439, 145, 42 },
		{ 439, 146, 42 },
		{ 439, 147, 42 },
		{ 439, 148, 42 },
		{ 439, 149, 42 },
		{ 439, 150, 42 },
		{ 439, 151, 42 },
		{ 439, 152, 42 },
		{ 439, 153, 42 },
		{ 439, 154, 42 },
		{ 439, 155, 42 },
		{ 439, 156, 42 },
		{ 439, 157, 42 },
		{ 439, 158, 42 },
		{ 439, 159, 42 },
		{ 439, 160, 42 },
		{ 439, 161, 42 },
		{ 439, 162, 42 },
		{ 439, 163, 42 },
		{ 439, 164, 42 },
		{ 439, 165, 42 },
		{ 439, 166, 42 },
		{ 439, 167, 42 },
		{ 439, 168, 42 },
		{ 439, 169, 42 },
		{ 439, 170, 42 },
		{ 439, 171, 42 },
		{ 439, 172, 42 },
		{ 439, 173, 42 },
		{ 439, 174, 42 },
		{ 439, 175, 42 },
		{ 439, 176, 42 },
		{ 439, 177, 42 },
		{ 439, 178, 42 },
		{ 439, 179, 42 },
		{ 439, 180, 42 },
		{ 439, 181, 42 },
		{ 439, 182, 42 },
		{ 439, 183, 42 },
		{ 439, 184, 42 },
		{ 439, 185, 42 },
		{ 439, 186, 42 },
		{ 439, 187, 42 },
		{ 439, 188, 42 },
		{ 440, 143, 42 },
		{ 440, 144, 42 },
		{ 440, 145, 42 },
		{ 440, 146, 42 },
		{ 440, 147, 42 },
		{ 440, 148, 42 },
		{ 440, 149, 42 },
		{ 440, 150, 42 },
		{ 440, 151, 42 },
		{ 440, 152, 42 },
		{ 440, 153, 42 },
		{ 440, 154, 42 },
		{ 440, 155, 42 },
		{ 440, 156, 42 },
		{ 440, 157, 42 },
		{ 440, 158, 42 },
		{ 440, 159, 42 },
		{ 440, 160, 42 },
		{ 440, 161, 42 },
		{ 440, 162, 42 },
		{ 440, 163, 42 },
		{ 440, 164, 42 },
		{ 440, 165, 42 },
		{ 440, 166, 42 },
		{ 440, 167, 42 },
		{ 440, 168, 42 },
		{ 440, 169, 42 },
		{ 440, 170, 42 },
		{ 440, 171, 42 },
		{ 440, 172, 42 },
		{ 440, 173, 42 },
		{ 440, 174, 42 },
		{ 440, 175, 42 },
		{ 440, 176, 42 },
		{ 440, 177, 42 },
		{ 440, 178, 42 },
		{ 440, 179, 42 },
		{ 440, 180, 42 },
		{ 440, 181, 42 },
		{ 440, 182, 42 },
		{ 440, 183, 42 },
		{ 440, 184, 42 },
		{ 440, 185, 42 },
		{ 440, 186, 42 },
		{ 440, 187, 42 },
		{ 440, 188, 42 },
		{ 441, 143, 42 },
		{ 441, 144, 42 },
		{ 441, 145, 42 },
		{ 441, 146, 42 },
		{ 441, 147, 42 },
		{ 441, 148, 42 },
		{ 441, 149, 42 },
		{ 441, 150, 42 },
		{ 441, 151, 42 },
		{ 441, 152, 42 },
		{ 441, 153, 42 },
		{ 441, 154, 42 },
		{ 441, 155, 42 },
		{ 441, 156, 42 },
		{ 441, 157, 42 },
		{ 441, 158, 42 },
		{ 441, 159, 42 },
		{ 441, 160, 42 },
		{ 441, 161, 42 },
		{ 441, 162, 42 },
		{ 441, 163, 42 },
		{ 441, 164, 42 },
		{ 441, 165, 42 },
		{ 441, 166, 42 },
		{ 441, 167, 42 },
		{ 441, 168, 42 },
		{ 441, 169, 42 },
		{ 441, 170, 42 },
		{ 441, 171, 42 },
		{ 441, 172, 42 },
		{ 441, 173, 42 },
		{ 441, 174, 42 },
		{ 441, 175, 42 },
		{ 441, 176, 42 },
		{ 441, 177, 42 },
		{ 441, 178, 42 },
		{ 441, 179, 42 },
		{ 441, 180, 42 },
		{ 441, 181, 42 },
		{ 441, 182, 42 },
		{ 441, 183, 42 },
		{ 441, 184, 42 },
		{ 441, 185, 42 },
		{ 441, 186, 42 },
		{ 441, 187, 42 },
		{ 442, 142, 42 },
		{ 442, 143, 42 },
		{ 442, 144, 42 },
		{ 442, 145, 42 },
		{ 442, 146, 42 },
		{ 442, 147, 42 },
		{ 442, 148, 42 },
		{ 442, 149, 42 },
		{ 442, 150, 42 },
		{ 442, 151, 42 },
		{ 442, 152, 42 },
		{ 442, 153, 42 },
		{ 442, 154, 42 },
		{ 442, 155, 42 },
		{ 442, 156, 42 },
		{ 442, 157, 42 },
		{ 442, 158, 42 },
		{ 442, 159, 42 },
		{ 442, 160, 42 },
		{ 442, 161, 42 },
		{ 442, 162, 42 },
		{ 442, 163, 42 },
		{ 442, 164, 42 },
		{ 442, 165, 42 },
		{ 442, 166, 42 },
		{ 442, 167, 42 },
		{ 442, 168, 42 },
		{ 442, 169, 42 },
		{ 442, 170, 42 },
		{ 442, 171, 42 },
		{ 442, 172, 42 },
		{ 442, 173, 42 },
		{ 442, 174, 42 },
		{ 442, 175, 42 },
		{ 442, 176, 42 },
		{ 442, 177, 42 },
		{ 442, 178, 42 },
		{ 442, 179, 42 },
		{ 442, 180, 42 },
		{ 442, 181, 42 },
		{ 442, 182, 42 },
		{ 442, 183, 42 },
		{ 442, 184, 42 },
		{ 442, 185, 42 },
		{ 442, 186, 42 },
		{ 442, 187, 42 },
		{ 443, 141, 42 },
		{ 443, 142, 42 },
		{ 443, 143, 42 },
		{ 443, 144, 42 },
		{ 443, 145, 42 },
		{ 443, 146, 42 },
		{ 443, 147, 42 },
		{ 443, 148, 42 },
		{ 443, 149, 42 },
		{ 443, 150, 42 },
		{ 443, 151, 42 },
		{ 443, 152, 42 },
		{ 443, 153, 42 },
		{ 443, 154, 42 },
		{ 443, 155, 42 },
		{ 443, 156, 42 },
		{ 443, 157, 42 },
		{ 443, 158, 42 },
		{ 443, 159, 42 },
		{ 443, 160, 42 },
		{ 443, 161, 42 },
		{ 443, 162, 42 },
		{ 443, 163, 42 },
		{ 443, 164, 42 },
		{ 443, 165, 42 },
		{ 443, 166, 42 },
		{ 443, 167, 42 },
		{ 443, 168, 42 },
		{ 443, 169, 42 },
		{ 443, 170, 42 },
		{ 443, 171, 42 },
		{ 443, 172, 42 },
		{ 443, 173, 42 },
		{ 443, 174, 42 },
		{ 443, 175, 42 },
		{ 443, 176, 42 },
		{ 443, 177, 42 },
		{ 443, 178, 42 },
		{ 443, 179, 42 },
		{ 443, 180, 42 },
		{ 443, 181, 42 },
		{ 443, 182, 42 },
		{ 443, 183, 42 },
		{ 443, 184, 42 },
		{ 443, 185, 42 },
		{ 443, 186, 42 },
		{ 443, 187, 42 },
		{ 444, 141, 42 },
		{ 444, 142, 42 },
		{ 444, 143, 42 },
		{ 444, 144, 42 },
		{ 444, 145, 42 },
		{ 444, 146, 42 },
		{ 444, 147, 42 },
		{ 444, 148, 42 },
		{ 444, 149, 42 },
		{ 444, 150, 42 },
		{ 444, 151, 42 },
		{ 444, 152, 42 },
		{ 444, 153, 42 },
		{ 444, 154, 42 },
		{ 444, 155, 42 },
		{ 444, 156, 42 },
		{ 444, 157, 42 },
		{ 444, 158, 42 },
		{ 444, 159, 42 },
		{ 444, 160, 42 },
		{ 444, 161, 42 },
		{ 444, 162, 42 },
		{ 444, 163, 42 },
		{ 444, 164, 42 },
		{ 444, 165, 42 },
		{ 444, 166, 42 },
		{ 444, 167, 42 },
		{ 444, 168, 42 },
		{ 444, 169, 42 },
		{ 444, 170, 42 },
		{ 444, 171, 42 },
		{ 444, 172, 42 },
		{ 444, 173, 42 },
		{ 444, 174, 42 },
		{ 444, 175, 42 },
		{ 444, 176, 42 },
		{ 444, 177, 42 },
		{ 444, 178, 42 },
		{ 444, 179, 42 },
		{ 444, 180, 42 },
		{ 444, 181, 42 },
		{ 444, 182, 42 },
		{ 444, 183, 42 },
		{ 444, 184, 42 },
		{ 444, 185, 42 },
		{ 444, 186, 42 },
		{ 445, 140, 42 },
		{ 445, 141, 42 },
		{ 445, 142, 42 },
		{ 445, 143, 42 },
		{ 445, 144, 42 },
		{ 445, 145, 42 },
		{ 445, 146, 42 },
		{ 445, 147, 42 },
		{ 445, 148, 42 },
		{ 445, 149, 42 },
		{ 445, 150, 42 },
		{ 445, 151, 42 },
		{ 445, 152, 42 },
		{ 445, 153, 42 },
		{ 445, 154, 42 },
		{ 445, 155, 42 },
		{ 445, 156, 42 },
		{ 445, 157, 42 },
		{ 445, 158, 42 },
		{ 445, 159, 42 },
		{ 445, 160, 42 },
		{ 445, 161, 42 },
		{ 445, 162, 42 },
		{ 445, 163, 42 },
		{ 445, 164, 42 },
		{ 445, 165, 42 },
		{ 445, 166, 42 },
		{ 445, 167, 42 },
		{ 445, 168, 42 },
		{ 445, 169, 42 },
		{ 445, 170, 42 },
		{ 445, 171, 42 },
		{ 445, 172, 42 },
		{ 445, 173, 42 },
		{ 445, 174, 42 },
		{ 445, 175, 42 },
		{ 445, 176, 42 },
		{ 445, 177, 42 },
		{ 445, 178, 42 },
		{ 445, 179, 42 },
		{ 445, 180, 42 },
		{ 445, 181, 42 },
		{ 445, 182, 42 },
		{ 445, 183, 42 },
		{ 445, 184, 42 },
		{ 445, 185, 42 },
		{ 445, 186, 42 },
		{ 446, 139, 42 },
		{ 446, 140, 42 },
		{ 446, 141, 42 },
		{ 446, 142, 42 },
		{ 446, 143, 42 },
		{ 446, 144, 42 },
		{ 446, 145, 42 },
		{ 446, 146, 42 },
		{ 446, 147, 42 },
		{ 446, 148, 42 },
		{ 446, 149, 42 },
		{ 446, 150, 42 },
		{ 446, 151, 42 },
		{ 446, 152, 42 },
		{ 446, 153, 42 },
		{ 446, 154, 42 },
		{ 446, 155, 42 },
		{ 446, 156, 42 },
		{ 446, 157, 42 },
		{ 446, 158, 42 },
		{ 446, 159, 42 },
		{ 446, 160, 42 },
		{ 446, 161, 42 },
		{ 446, 162, 42 },
		{ 446, 163, 42 },
		{ 446, 164, 42 },
		{ 446, 165, 42 },
		{ 446, 166, 42 },
		{ 446, 167, 42 },
		{ 446, 168, 42 },
		{ 446, 169, 42 },
		{ 446, 170, 42 },
		{ 446, 171, 42 },
		{ 446, 172, 42 },
		{ 446, 173, 42 },
		{ 446, 174, 42 },
		{ 446, 175, 42 },
		{ 446, 176, 42 },
		{ 446, 177, 42 },
		{ 446, 178, 42 },
		{ 446, 179, 42 },
		{ 446, 180, 42 },
		{ 446, 181, 42 },
		{ 446, 182, 42 },
		{ 446, 183, 42 },
		{ 446, 184, 42 },
		{ 446, 185, 42 },
		{ 446, 186, 42 },
		{ 447, 139, 42 },
		{ 447, 140, 42 },
		{ 447, 141, 42 },
		{ 447, 142, 42 },
		{ 447, 143, 42 },
		{ 447, 144, 42 },
		{ 447, 145, 42 },
		{ 447, 146, 42 },
		{ 447, 147, 42 },
		{ 447, 148, 42 },
		{ 447, 149, 42 },
		{ 447, 150, 42 },
		{ 447, 151, 42 },
		{ 447, 152, 42 },
		{ 447, 153, 42 },
		{ 447, 154, 42 },
		{ 447, 155, 42 },
		{ 447, 156, 42 },
		{ 447, 157, 42 },
		{ 447, 158, 42 },
		{ 447, 159, 42 },
		{ 447, 160, 42 },
		{ 447, 161, 42 },
		{ 447, 162, 42 },
		{ 447, 163, 42 },
		{ 447, 164, 42 },
		{ 447, 165, 42 },
		{ 447, 166, 42 },
		{ 447, 167, 42 },
		{ 447, 168, 42 },
		{ 447, 169, 42 },
		{ 447, 170, 42 },
		{ 447, 171, 42 },
		{ 447, 172, 42 },
		{ 447, 173, 42 },
		{ 447, 174, 42 },
		{ 447, 175, 42 },
		{ 447, 176, 42 },
		{ 447, 177, 42 },
		{ 447, 178, 42 },
		{ 447, 179, 42 },
		{ 447, 180, 42 },
		{ 447, 181, 42 },
		{ 447, 182, 42 },
		{ 447, 183, 42 },
		{ 447, 184, 42 },
		{ 447, 185, 42 },
		{ 448, 138, 42 },
		{ 448, 139, 42 },
		{ 448, 140, 42 },
		{ 448, 141, 42 },
		{ 448, 142, 42 },
		{ 448, 143, 42 },
		{ 448, 144, 42 },
		{ 448, 145, 42 },
		{ 448, 146, 42 },
		{ 448, 147, 42 },
		{ 448, 148, 42 },
		{ 448, 149, 42 },
		{ 448, 150, 42 },
		{ 448, 151, 42 },
		{ 448, 152, 42 },
		{ 448, 153, 42 },
		{ 448, 154, 42 },
		{ 448, 155, 42 },
		{ 448, 156, 42 },
		{ 448, 157, 42 },
		{ 448, 158, 42 },
		{ 448, 159, 42 },
		{ 448, 160, 42 },
		{ 448, 161, 42 },
		{ 448, 162, 42 },
		{ 448, 163, 42 },
		{ 448, 164, 42 },
		{ 448, 165, 42 },
		{ 448, 166, 42 },
		{ 448, 167, 42 },
		{ 448, 168, 42 },
		{ 448, 169, 42 },
		{ 448, 170, 42 },
		{ 448, 171, 42 },
		{ 448, 172, 42 },
		{ 448, 173, 42 },
		{ 448, 174, 42 },
		{ 448, 175, 42 },
		{ 448, 176, 42 },
		{ 448, 177, 42 },
		{ 448, 178, 42 },
		{ 448, 179, 42 },
		{ 448, 180, 42 },
		{ 448, 181, 42 },
		{ 448, 182, 42 },
		{ 448, 183, 42 },
		{ 448, 184, 42 },
		{ 448, 185, 42 },
		{ 449, 138, 42 },
		{ 449, 139, 42 },
		{ 449, 140, 42 },
		{ 449, 141, 42 },
		{ 449, 142, 42 },
		{ 449, 143, 42 },
		{ 449, 144, 42 },
		{ 449, 145, 42 },
		{ 449, 146, 42 },
		{ 449, 147, 42 },
		{ 449, 148, 42 },
		{ 449, 149, 42 },
		{ 449, 150, 42 },
		{ 449, 151, 42 },
		{ 449, 152, 42 },
		{ 449, 153, 42 },
		{ 449, 154, 42 },
		{ 449, 155, 42 },
		{ 449, 156, 42 },
		{ 449, 157, 42 },
		{ 449, 158, 42 },
		{ 449, 159, 42 },
		{ 449, 160, 42 },
		{ 449, 161, 42 },
		{ 449, 162, 42 },
		{ 449, 163, 42 },
		{ 449, 164, 42 },
		{ 449, 165, 42 },
		{ 449, 166, 42 },
		{ 449, 167, 42 },
		{ 449, 168, 42 },
		{ 449, 169, 42 },
		{ 449, 170, 42 },
		{ 449, 171, 42 },
		{ 449, 172, 42 },
		{ 449, 173, 42 },
		{ 449, 174, 42 },
		{ 449, 175, 42 },
		{ 449, 176, 42 },
		{ 449, 177, 42 },
		{ 449, 178, 42 },
		{ 449, 179, 42 },
		{ 449, 180, 42 },
		{ 449, 181, 42 },
		{ 449, 182, 42 },
		{ 449, 183, 42 },
		{ 449, 184, 42 },
		{ 450, 137, 42 },
		{ 450, 138, 42 },
		{ 450, 139, 42 },
		{ 450, 140, 42 },
		{ 450, 141, 42 },
		{ 450, 142, 42 },
		{ 450, 143, 42 },
		{ 450, 144, 42 },
		{ 450, 145, 42 },
		{ 450, 146, 42 },
		{ 450, 147, 42 },
		{ 450, 148, 42 },
		{ 450, 149, 42 },
		{ 450, 150, 42 },
		{ 450, 151, 42 },
		{ 450, 152, 42 },
		{ 450, 153, 42 },
		{ 450, 154, 42 },
		{ 450, 155, 42 },
		{ 450, 156, 42 },
		{ 450, 157, 42 },
		{ 450, 158, 42 },
		{ 450, 159, 42 },
		{ 450, 160, 42 },
		{ 450, 161, 42 },
		{ 450, 162, 42 },
		{ 450, 163, 42 },
		{ 450, 164, 42 },
		{ 450, 165, 42 },
		{ 450, 166, 42 },
		{ 450, 167, 42 },
		{ 450, 168, 42 },
		{ 450, 169, 42 },
		{ 450, 170, 42 },
		{ 450, 171, 42 },
		{ 450, 172, 42 },
		{ 450, 173, 42 },
		{ 450, 174, 42 },
		{ 450, 175, 42 },
		{ 450, 176, 42 },
		{ 450, 177, 42 },
		{ 450, 178, 42 },
		{ 450, 179, 42 },
		{ 450, 180, 42 },
		{ 450, 181, 42 },
		{ 450, 182, 42 },
		{ 450, 183, 42 },
		{ 450, 184, 42 },
		{ 451, 136, 42 },
		{ 451, 137, 42 },
		{ 451, 138, 42 },
		{ 451, 139, 42 },
		{ 451, 140, 42 },
		{ 451, 141, 42 },
		{ 451, 142, 42 },
		{ 451, 143, 42 },
		{ 451, 144, 42 },
		{ 451, 145, 42 },
		{ 451, 146, 42 },
		{ 451, 147, 42 },
		{ 451, 148, 42 },
		{ 451, 149, 42 },
		{ 451, 150, 42 },
		{ 451, 151, 42 },
		{ 451, 152, 42 },
		{ 451, 153, 42 },
		{ 451, 154, 42 },
		{ 451, 155, 42 },
		{ 451, 156, 42 },
		{ 451, 157, 42 },
		{ 451, 158, 42 },
		{ 451, 159, 42 },
		{ 451, 160, 42 },
		{ 451, 161, 42 },
		{ 451, 162, 42 },
		{ 451, 163, 42 },
		{ 451, 164, 42 },
		{ 451, 165, 42 },
		{ 451, 166, 42 },
		{ 451, 167, 42 },
		{ 451, 168, 42 },
		{ 451, 169, 42 },
		{ 451, 170, 42 },
		{ 451, 171, 42 },
		{ 451, 172, 42 },
		{ 451, 173, 42 },
		{ 451, 174, 42 },
		{ 451, 175, 42 },
		{ 451, 176, 42 },
		{ 451, 177, 42 },
		{ 451, 178, 42 },
		{ 451, 179, 42 },
		{ 451, 180, 42 },
		{ 451, 181, 42 },
		{ 451, 182, 42 },
		{ 451, 183, 42 },
		{ 451, 184, 42 },
		{ 452, 136, 42 },
		{ 452, 137, 42 },
		{ 452, 138, 42 },
		{ 452, 139, 42 },
		{ 452, 140, 42 },
		{ 452, 141, 42 },
		{ 452, 142, 42 },
		{ 452, 143, 42 },
		{ 452, 144, 42 },
		{ 452, 145, 42 },
		{ 452, 146, 42 },
		{ 452, 147, 42 },
		{ 452, 148, 42 },
		{ 452, 149, 42 },
		{ 452, 150, 42 },
		{ 452, 151, 42 },
		{ 452, 152, 42 },
		{ 452, 153, 42 },
		{ 452, 154, 42 },
		{ 452, 155, 42 },
		{ 452, 156, 42 },
		{ 452, 157, 42 },
		{ 452, 158, 42 },
		{ 452, 159, 42 },
		{ 452, 160, 42 },
		{ 452, 161, 42 },
		{ 452, 162, 42 },
		{ 452, 163, 42 },
		{ 452, 164, 42 },
		{ 452, 165, 42 },
		{ 452, 166, 42 },
		{ 452, 167, 42 },
		{ 452, 168, 42 },
		{ 452, 169, 42 },
		{ 452, 170, 42 },
		{ 452, 171, 42 },
		{ 452, 172, 42 },
		{ 452, 173, 42 },
		{ 452, 174, 42 },
		{ 452, 175, 42 },
		{ 452, 176, 42 },
		{ 452, 177, 42 },
		{ 452, 178, 42 },
		{ 452, 179, 42 },
		{ 452, 180, 42 },
		{ 452, 181, 42 },
		{ 452, 182, 42 },
		{ 452, 183, 42 },
		{ 453, 135, 42 },
		{ 453, 136, 42 },
		{ 453, 137, 42 },
		{ 453, 138, 42 },
		{ 453, 139, 42 },
		{ 453, 140, 42 },
		{ 453, 141, 42 },
		{ 453, 142, 42 },
		{ 453, 143, 42 },
		{ 453, 144, 42 },
		{ 453, 145, 42 },
		{ 453, 146, 42 },
		{ 453, 147, 42 },
		{ 453, 148, 42 },
		{ 453, 149, 42 },
		{ 453, 150, 42 },
		{ 453, 151, 42 },
		{ 453, 152, 42 },
		{ 453, 153, 42 },
		{ 453, 154, 42 },
		{ 453, 155, 42 },
		{ 453, 156, 42 },
		{ 453, 157, 42 },
		{ 453, 158, 42 },
		{ 453, 159, 42 },
		{ 453, 160, 42 },
		{ 453, 161, 42 },
		{ 453, 162, 42 },
		{ 453, 163, 42 },
		{ 453, 164, 42 },
		{ 453, 165, 42 },
		{ 453, 166, 42 },
		{ 453, 167, 42 },
		{ 453, 168, 42 },
		{ 453, 169, 42 },
		{ 453, 170, 42 },
		{ 453, 171, 42 },
		{ 453, 172, 42 },
		{ 453, 173, 42 },
		{ 453, 174, 42 },
		{ 453, 175, 42 },
		{ 453, 176, 42 },
		{ 453, 177, 42 },
		{ 453, 178, 42 },
		{ 453, 179, 42 },
		{ 453, 180, 42 },
		{ 453, 181, 42 },
		{ 453, 182, 42 },
		{ 453, 183, 42 },
		{ 454, 134, 42 },
		{ 454, 135, 42 },
		{ 454, 136, 42 },
		{ 454, 137, 42 },
		{ 454, 138, 42 },
		{ 454, 139, 42 },
		{ 454, 140, 42 },
		{ 454, 141, 42 },
		{ 454, 142, 42 },
		{ 454, 143, 42 },
		{ 454, 144, 42 },
		{ 454, 145, 42 },
		{ 454, 146, 42 },
		{ 454, 147, 42 },
		{ 454, 148, 42 },
		{ 454, 149, 42 },
		{ 454, 150, 42 },
		{ 454, 151, 42 },
		{ 454, 152, 42 },
		{ 454, 153, 42 },
		{ 454, 154, 42 },
		{ 454, 155, 42 },
		{ 454, 156, 42 },
		{ 454, 157, 42 },
		{ 454, 158, 42 },
		{ 454, 159, 42 },
		{ 454, 160, 42 },
		{ 454, 161, 42 },
		{ 454, 162, 42 },
		{ 454, 163, 42 },
		{ 454, 164, 42 },
		{ 454, 165, 42 },
		{ 454, 166, 42 },
		{ 454, 167, 42 },
		{ 454, 168, 42 },
		{ 454, 169, 42 },
		{ 454, 170, 42 },
		{ 454, 171, 42 },
		{ 454, 172, 42 },
		{ 454, 173, 42 },
		{ 454, 174, 42 },
		{ 454, 175, 42 },
		{ 454, 176, 42 },
		{ 454, 177, 42 },
		{ 454, 178, 42 },
		{ 454, 179, 42 },
		{ 454, 180, 42 },
		{ 454, 181, 42 },
		{ 454, 182, 42 },
		{ 454, 183, 42 },
		{ 455, 134, 42 },
		{ 455, 135, 42 },
		{ 455, 136, 42 },
		{ 455, 137, 42 },
		{ 455, 138, 42 },
		{ 455, 139, 42 },
		{ 455, 140, 42 },
		{ 455, 141, 42 },
		{ 455, 142, 42 },
		{ 455, 143, 42 },
		{ 455, 144, 42 },
		{ 455, 145, 42 },
		{ 455, 146, 42 },
		{ 455, 147, 42 },
		{ 455, 148, 42 },
		{ 455, 149, 42 },
		{ 455, 150, 42 },
		{ 455, 151, 42 },
		{ 455, 152, 42 },
		{ 455, 153, 42 },
		{ 455, 154, 42 },
		{ 455, 155, 42 },
		{ 455, 156, 42 },
		{ 455, 157, 42 },
		{ 455, 158, 42 },
		{ 455, 159, 42 },
		{ 455, 160, 42 },
		{ 455, 161, 42 },
		{ 455, 162, 42 },
		{ 455, 163, 42 },
		{ 455, 164, 42 },
		{ 455, 165, 42 },
		{ 455, 166, 42 },
		{ 455, 167, 42 },
		{ 455, 168, 42 },
		{ 455, 169, 42 },
		{ 455, 170, 42 },
		{ 455, 171, 42 },
		{ 455, 172, 42 },
		{ 455, 173, 42 },
		{ 455, 174, 42 },
		{ 455, 175, 42 },
		{ 455, 176, 42 },
		{ 455, 177, 42 },
		{ 455, 178, 42 },
		{ 455, 179, 42 },
		{ 455, 180, 42 },
		{ 455, 181, 42 },
		{ 455, 182, 42 },
		{ 456, 133, 42 },
		{ 456, 134, 42 },
		{ 456, 135, 42 },
		{ 456, 136, 42 },
		{ 456, 137, 42 },
		{ 456, 138, 42 },
		{ 456, 139, 42 },
		{ 456, 140, 42 },
		{ 456, 141, 42 },
		{ 456, 142, 42 },
		{ 456, 143, 42 },
		{ 456, 144, 42 },
		{ 456, 145, 42 },
		{ 456, 146, 42 },
		{ 456, 147, 42 },
		{ 456, 148, 42 },
		{ 456, 149, 42 },
		{ 456, 150, 42 },
		{ 456, 151, 42 },
		{ 456, 152, 42 },
		{ 456, 153, 42 },
		{ 456, 154, 42 },
		{ 456, 155, 42 },
		{ 456, 156, 42 },
		{ 456, 157, 42 },
		{ 456, 158, 42 },
		{ 456, 159, 42 },
		{ 456, 160, 42 },
		{ 456, 161, 42 },
		{ 456, 162, 42 },
		{ 456, 163, 42 },
		{ 456, 164, 42 },
		{ 456, 165, 42 },
		{ 456, 166, 42 },
		{ 456, 167, 42 },
		{ 456, 168, 42 },
		{ 456, 169, 42 },
		{ 456, 170, 42 },
		{ 456, 171, 42 },
		{ 456, 172, 42 },
		{ 456, 173, 42 },
		{ 456, 174, 42 },
		{ 456, 175, 42 },
		{ 456, 176, 42 },
		{ 456, 177, 42 },
		{ 456, 178, 42 },
		{ 456, 179, 42 },
		{ 456, 180, 42 },
		{ 456, 181, 42 },
		{ 456, 182, 42 },
		{ 457, 133, 42 },
		{ 457, 134, 42 },
		{ 457, 135, 42 },
		{ 457, 136, 42 },
		{ 457, 137, 42 },
		{ 457, 138, 42 },
		{ 457, 139, 42 },
		{ 457, 140, 42 },
		{ 457, 141, 42 },
		{ 457, 142, 42 },
		{ 457, 143, 42 },
		{ 457, 144, 42 },
		{ 457, 145, 42 },
		{ 457, 146, 42 },
		{ 457, 147, 42 },
		{ 457, 148, 42 },
		{ 457, 149, 42 },
		{ 457, 150, 42 },
		{ 457, 151, 42 },
		{ 457, 152, 42 },
		{ 457, 153, 42 },
		{ 457, 154, 42 },
		{ 457, 155, 42 },
		{ 457, 156, 42 },
		{ 457, 157, 42 },
		{ 457, 158, 42 },
		{ 457, 159, 42 },
		{ 457, 160, 42 },
		{ 457, 161, 42 },
		{ 457, 162, 42 },
		{ 457, 163, 42 },
		{ 457, 164, 42 },
		{ 457, 165, 42 },
		{ 457, 166, 42 },
		{ 457, 167, 42 },
		{ 457, 168, 42 },
		{ 457, 169, 42 },
		{ 457, 170, 42 },
		{ 457, 171, 42 },
		{ 457, 172, 42 },
		{ 457, 173, 42 },
		{ 457, 174, 42 },
		{ 457, 175, 42 },
		{ 457, 176, 42 },
		{ 457, 177, 42 },
		{ 457, 178, 42 },
		{ 457, 179, 42 },
		{ 457, 180, 42 },
		{ 457, 181, 42 },
		{ 458, 132, 42 },
		{ 458, 133, 42 },
		{ 458, 134, 42 },
		{ 458, 135, 42 },
		{ 458, 136, 42 },
		{ 458, 137, 42 },
		{ 458, 138, 42 },
		{ 458, 139, 42 },
		{ 458, 140, 42 },
		{ 458, 141, 42 },
		{ 458, 142, 42 },
		{ 458, 143, 42 },
		{ 458, 144, 42 },
		{ 458, 145, 42 },
		{ 458, 146, 42 },
		{ 458, 147, 42 },
		{ 458, 148, 42 },
		{ 458, 149, 42 },
		{ 458, 150, 42 },
		{ 458, 151, 42 },
		{ 458, 152, 42 },
		{ 458, 153, 42 },
		{ 458, 154, 42 },
		{ 458, 155, 42 },
		{ 458, 156, 42 },
		{ 458, 157, 42 },
		{ 458, 158, 42 },
		{ 458, 159, 42 },
		{ 458, 160, 42 },
		{ 458, 161, 42 },
		{ 458, 162, 42 },
		{ 458, 163, 42 },
		{ 458, 164, 42 },
		{ 458, 165, 42 },
		{ 458, 166, 42 },
		{ 458, 167, 42 },
		{ 458, 168, 42 },
		{ 458, 169, 42 },
		{ 458, 170, 42 },
		{ 458, 171, 42 },
		{ 458, 172, 42 },
		{ 458, 173, 42 },
		{ 458, 174, 42 },
		{ 458, 175, 42 },
		{ 458, 176, 42 },
		{ 458, 177, 42 },
		{ 458, 178, 42 },
		{ 458, 179, 42 },
		{ 458, 180, 42 },
		{ 458, 181, 42 },
		{ 459, 131, 42 },
		{ 459, 132, 42 },
		{ 459, 133, 42 },
		{ 459, 134, 42 },
		{ 459, 135, 42 },
		{ 459, 136, 42 },
		{ 459, 137, 42 },
		{ 459, 138, 42 },
		{ 459, 139, 42 },
		{ 459, 140, 42 },
		{ 459, 141, 42 },
		{ 459, 142, 42 },
		{ 459, 143, 42 },
		{ 459, 144, 42 },
		{ 459, 145, 42 },
		{ 459, 146, 42 },
		{ 459, 147, 42 },
		{ 459, 148, 42 },
		{ 459, 149, 42 },
		{ 459, 150, 42 },
		{ 459, 151, 42 },
		{ 459, 152, 42 },
		{ 459, 153, 42 },
		{ 459, 154, 42 },
		{ 459, 155, 42 },
		{ 459, 156, 42 },
		{ 459, 157, 42 },
		{ 459, 158, 42 },
		{ 459, 159, 42 },
		{ 459, 160, 42 },
		{ 459, 161, 42 },
		{ 459, 162, 42 },
		{ 459, 163, 42 },
		{ 459, 164, 42 },
		{ 459, 165, 42 },
		{ 459, 166, 42 },
		{ 459, 167, 42 },
		{ 459, 168, 42 },
		{ 459, 169, 42 },
		{ 459, 170, 42 },
		{ 459, 171, 42 },
		{ 459, 172, 42 },
		{ 459, 173, 42 },
		{ 459, 174, 42 },
		{ 459, 175, 42 },
		{ 459, 176, 42 },
		{ 459, 177, 42 },
		{ 459, 178, 42 },
		{ 459, 179, 42 },
		{ 459, 180, 42 },
		{ 459, 181, 42 },
		{ 460, 131, 42 },
		{ 460, 132, 42 },
		{ 460, 133, 42 },
		{ 460, 134, 42 },
		{ 460, 135, 42 },
		{ 460, 136, 42 },
		{ 460, 137, 42 },
		{ 460, 138, 42 },
		{ 460, 139, 42 },
		{ 460, 140, 42 },
		{ 460, 141, 42 },
		{ 460, 142, 42 },
		{ 460, 143, 42 },
		{ 460, 144, 42 },
		{ 460, 145, 42 },
		{ 460, 146, 42 },
		{ 460, 147, 42 },
		{ 460, 148, 42 },
		{ 460, 149, 42 },
		{ 460, 150, 42 },
		{ 460, 151, 42 },
		{ 460, 152, 42 },
		{ 460, 153, 42 },
		{ 460, 154, 42 },
		{ 460, 155, 42 },
		{ 460, 156, 42 },
		{ 460, 157, 42 },
		{ 460, 158, 42 },
		{ 460, 159, 42 },
		{ 460, 160, 42 },
		{ 460, 161, 42 },
		{ 460, 162, 42 },
		{ 460, 163, 42 },
		{ 460, 164, 42 },
		{ 460, 165, 42 },
		{ 460, 166, 42 },
		{ 460, 167, 42 },
		{ 460, 168, 42 },
		{ 460, 169, 42 },
		{ 460, 170, 42 },
		{ 460, 171, 42 },
		{ 460, 172, 42 },
		{ 460, 173, 42 },
		{ 460, 174, 42 },
		{ 460, 175, 42 },
		{ 460, 176, 42 },
		{ 460, 177, 42 },
		{ 460, 178, 42 },
		{ 460, 179, 42 },
		{ 460, 180, 42 },
		{ 461, 130, 42 },
		{ 461, 131, 42 },
		{ 461, 132, 42 },
		{ 461, 133, 42 },
		{ 461, 134, 42 },
		{ 461, 135, 42 },
		{ 461, 136, 42 },
		{ 461, 137, 42 },
		{ 461, 138, 42 },
		{ 461, 139, 42 },
		{ 461, 140, 42 },
		{ 461, 141, 42 },
		{ 461, 142, 42 },
		{ 461, 143, 42 },
		{ 461, 144, 42 },
		{ 461, 145, 42 },
		{ 461, 146, 42 },
		{ 461, 147, 42 },
		{ 461, 148, 42 },
		{ 461, 149, 42 },
		{ 461, 150, 42 },
		{ 461, 151, 42 },
		{ 461, 152, 42 },
		{ 461, 153, 42 },
		{ 461, 154, 42 },
		{ 461, 155, 42 },
		{ 461, 156, 42 },
		{ 461, 157, 42 },
		{ 461, 158, 42 },
		{ 461, 159, 42 },
		{ 461, 160, 42 },
		{ 461, 161, 42 },
		{ 461, 162, 42 },
		{ 461, 163, 42 },
		{ 461, 164, 42 },
		{ 461, 165, 42 },
		{ 461, 166, 42 },
		{ 461, 167, 42 },
		{ 461, 168, 42 },
		{ 461, 169, 42 },
		{ 461, 170, 42 },
		{ 461, 171, 42 },
		{ 461, 172, 42 },
		{ 461, 173, 42 },
		{ 461, 174, 42 },
		{ 461, 175, 42 },
		{ 461, 176, 42 },
		{ 461, 177, 42 },
		{ 461, 178, 42 },
		{ 461, 179, 42 },
		{ 461, 180, 42 },
		{ 462, 129, 42 },
		{ 462, 130, 42 },
		{ 462, 131, 42 },
		{ 462, 132, 42 },
		{ 462, 133, 42 },
		{ 462, 134, 42 },
		{ 462, 135, 42 },
		{ 462, 136, 42 },
		{ 462, 137, 42 },
		{ 462, 138, 42 },
		{ 462, 139, 42 },
		{ 462, 140, 42 },
		{ 462, 141, 42 },
		{ 462, 142, 42 },
		{ 462, 143, 42 },
		{ 462, 144, 42 },
		{ 462, 145, 42 },
		{ 462, 146, 42 },
		{ 462, 147, 42 },
		{ 462, 148, 42 },
		{ 462, 149, 42 },
		{ 462, 150, 42 },
		{ 462, 151, 42 },
		{ 462, 152, 42 },
		{ 462, 153, 42 },
		{ 462, 154, 42 },
		{ 462, 155, 42 },
		{ 462, 156, 42 },
		{ 462, 157, 42 },
		{ 462, 158, 42 },
		{ 462, 159, 42 },
		{ 462, 160, 42 },
		{ 462, 161, 42 },
		{ 462, 162, 42 },
		{ 462, 163, 42 },
		{ 462, 164, 42 },
		{ 462, 165, 42 },
		{ 462, 166, 42 },
		{ 462, 167, 42 },
		{ 462, 168, 42 },
		{ 462, 169, 42 },
		{ 462, 170, 42 },
		{ 462, 171, 42 },
		{ 462, 172, 42 },
		{ 462, 173, 42 },
		{ 462, 174, 42 },
		{ 462, 175, 42 },
		{ 462, 176, 42 },
		{ 462, 177, 42 },
		{ 462, 178, 42 },
		{ 462, 179, 42 },
		{ 462, 180, 42 },
		{ 463, 129, 42 },
		{ 463, 130, 42 },
		{ 463, 131, 42 },
		{ 463, 132, 42 },
		{ 463, 133, 42 },
		{ 463, 134, 42 },
		{ 463, 135, 42 },
		{ 463, 136, 42 },
		{ 463, 137, 42 },
		{ 463, 138, 42 },
		{ 463, 139, 42 },
		{ 463, 140, 42 },
		{ 463, 141, 42 },
		{ 463, 142, 42 },
		{ 463, 143, 42 },
		{ 463, 144, 42 },
		{ 463, 145, 42 },
		{ 463, 146, 42 },
		{ 463, 147, 42 },
		{ 463, 148, 42 },
		{ 463, 149, 42 },
		{ 463, 150, 42 },
		{ 463, 151, 42 },
		{ 463, 152, 42 },
		{ 463, 153, 42 },
		{ 463, 154, 42 },
		{ 463, 155, 42 },
		{ 463, 156, 42 },
		{ 463, 157, 42 },
		{ 463, 158, 42 },
		{ 463, 159, 42 },
		{ 463, 160, 42 },
		{ 463, 161, 42 },
		{ 463, 162, 42 },
		{ 463, 163, 42 },
		{ 463, 164, 42 },
		{ 463, 165, 42 },
		{ 463, 166, 42 },
		{ 463, 167, 42 },
		{ 463, 168, 42 },
		{ 463, 169, 42 },
		{ 463, 170, 42 },
		{ 463, 171, 42 },
		{ 463, 172, 42 },
		{ 463, 173, 42 },
		{ 463, 174, 42 },
		{ 463, 175, 42 },
		{ 463, 176, 42 },
		{ 463, 177, 42 },
		{ 463, 178, 42 },
		{ 463, 179, 42 },
		{ 464, 128, 42 },
		{ 464, 129, 42 },
		{ 464, 130, 42 },
		{ 464, 131, 42 },
		{ 464, 132, 42 },
		{ 464, 133, 42 },
		{ 464, 134, 42 },
		{ 464, 135, 42 },
		{ 464, 136, 42 },
		{ 464, 137, 42 },
		{ 464, 138, 42 },
		{ 464, 139, 42 },
		{ 464, 140, 42 },
		{ 464, 141, 42 },
		{ 464, 142, 42 },
		{ 464, 143, 42 },
		{ 464, 144, 42 },
		{ 464, 145, 42 },
		{ 464, 146, 42 },
		{ 464, 147, 42 },
		{ 464, 148, 42 },
		{ 464, 149, 42 },
		{ 464, 150, 42 },
		{ 464, 151, 42 },
		{ 464, 152, 42 },
		{ 464, 153, 42 },
		{ 464, 154, 42 },
		{ 464, 155, 42 },
		{ 464, 156, 42 },
		{ 464, 157, 42 },
		{ 464, 158, 42 },
		{ 464, 159, 42 },
		{ 464, 160, 42 },
		{ 464, 161, 42 },
		{ 464, 162, 42 },
		{ 464, 163, 42 },
		{ 464, 164, 42 },
		{ 464, 165, 42 },
		{ 464, 166, 42 },
		{ 464, 167, 42 },
		{ 464, 168, 42 },
		{ 464, 169, 42 },
		{ 464, 170, 42 },
		{ 464, 171, 42 },
		{ 464, 172, 42 },
		{ 464, 173, 42 },
		{ 464, 174, 42 },
		{ 464, 175, 42 },
		{ 464, 176, 42 },
		{ 464, 177, 42 },
		{ 464, 178, 42 },
		{ 464, 179, 42 },
		{ 465, 128, 42 },
		{ 465, 129, 42 },
		{ 465, 130, 42 },
		{ 465, 131, 42 },
		{ 465, 132, 42 },
		{ 465, 133, 42 },
		{ 465, 134, 42 },
		{ 465, 135, 42 },
		{ 465, 136, 42 },
		{ 465, 137, 42 },
		{ 465, 138, 42 },
		{ 465, 139, 42 },
		{ 465, 140, 42 },
		{ 465, 141, 42 },
		{ 465, 142, 42 },
		{ 465, 143, 42 },
		{ 465, 144, 42 },
		{ 465, 145, 42 },
		{ 465, 146, 42 },
		{ 465, 147, 42 },
		{ 465, 148, 42 },
		{ 465, 149, 42 },
		{ 465, 150, 42 },
		{ 465, 151, 42 },
		{ 465, 152, 42 },
		{ 465, 153, 42 },
		{ 465, 154, 42 },
		{ 465, 155, 42 },
		{ 465, 156, 42 },
		{ 465, 157, 42 },
		{ 465, 158, 42 },
		{ 465, 159, 42 },
		{ 465, 160, 42 },
		{ 465, 161, 42 },
		{ 465, 162, 42 },
		{ 465, 163, 42 },
		{ 465, 164, 42 },
		{ 465, 165, 42 },
		{ 465, 166, 42 },
		{ 465, 167, 42 },
		{ 465, 168, 42 },
		{ 465, 169, 42 },
		{ 465, 170, 42 },
		{ 465, 171, 42 },
		{ 465, 172, 42 },
		{ 465, 173, 42 },
		{ 465, 174, 42 },
		{ 465, 175, 42 },
		{ 465, 176, 42 },
		{ 465, 177, 42 },
		{ 465, 178, 42 },
		{ 466, 127, 42 },
		{ 466, 128, 42 },
		{ 466, 129, 42 },
		{ 466, 130, 42 },
		{ 466, 131, 42 },
		{ 466, 132, 42 },
		{ 466, 133, 42 },
		{ 466, 134, 42 },
		{ 466, 135, 42 },
		{ 466, 136, 42 },
		{ 466, 137, 42 },
		{ 466, 138, 42 },
		{ 466, 139, 42 },
		{ 466, 140, 42 },
		{ 466, 141, 42 },
		{ 466, 142, 42 },
		{ 466, 143, 42 },
		{ 466, 144, 42 },
		{ 466, 145, 42 },
		{ 466, 146, 42 },
		{ 466, 147, 42 },
		{ 466, 148, 42 },
		{ 466, 149, 42 },
		{ 466, 150, 42 },
		{ 466, 151, 42 },
		{ 466, 152, 42 },
		{ 466, 153, 42 },
		{ 466, 154, 42 },
		{ 466, 155, 42 },
		{ 466, 156, 42 },
		{ 466, 157, 42 },
		{ 466, 158, 42 },
		{ 466, 159, 42 },
		{ 466, 160, 42 },
		{ 466, 161, 42 },
		{ 466, 162, 42 },
		{ 466, 163, 42 },
		{ 466, 164, 42 },
		{ 466, 165, 42 },
		{ 466, 166, 42 },
		{ 466, 167, 42 },
		{ 466, 168, 42 },
		{ 466, 169, 42 },
		{ 466, 170, 42 },
		{ 466, 171, 42 },
		{ 466, 172, 42 },
		{ 466, 173, 42 },
		{ 466, 174, 42 },
		{ 466, 175, 42 },
		{ 466, 176, 42 },
		{ 466, 177, 42 },
		{ 466, 178, 42 },
		{ 467, 126, 42 },
		{ 467, 127, 42 },
		{ 467, 128, 42 },
		{ 467, 129, 42 },
		{ 467, 130, 42 },
		{ 467, 131, 42 },
		{ 467, 132, 42 },
		{ 467, 133, 42 },
		{ 467, 134, 42 },
		{ 467, 135, 42 },
		{ 467, 136, 42 },
		{ 467, 137, 42 },
		{ 467, 138, 42 },
		{ 467, 139, 42 },
		{ 467, 140, 42 },
		{ 467, 141, 42 },
		{ 467, 142, 42 },
		{ 467, 143, 42 },
		{ 467, 144, 42 },
		{ 467, 145, 42 },
		{ 467, 146, 42 },
		{ 467, 147, 42 },
		{ 467, 148, 42 },
		{ 467, 149, 42 },
		{ 467, 150, 42 },
		{ 467, 151, 42 },
		{ 467, 152, 42 },
		{ 467, 153, 42 },
		{ 467, 154, 42 },
		{ 467, 155, 42 },
		{ 467, 156, 42 },
		{ 467, 157, 42 },
		{ 467, 158, 42 },
		{ 467, 159, 42 },
		{ 467, 160, 42 },
		{ 467, 161, 42 },
		{ 467, 162, 42 },
		{ 467, 163, 42 },
		{ 467, 164, 42 },
		{ 467, 165, 42 },
		{ 467, 166, 42 },
		{ 467, 167, 42 },
		{ 467, 168, 42 },
		{ 467, 169, 42 },
		{ 467, 170, 42 },
		{ 467, 171, 42 },
		{ 467, 172, 42 },
		{ 467, 173, 42 },
		{ 467, 174, 42 },
		{ 467, 175, 42 },
		{ 467, 176, 42 },
		{ 467, 177, 42 },
		{ 467, 178, 42 },
		{ 468, 126, 42 },
		{ 468, 127, 42 },
		{ 468, 128, 42 },
		{ 468, 129, 42 },
		{ 468, 130, 42 },
		{ 468, 131, 42 },
		{ 468, 132, 42 },
		{ 468, 133, 42 },
		{ 468, 134, 42 },
		{ 468, 135, 42 },
		{ 468, 136, 42 },
		{ 468, 137, 42 },
		{ 468, 138, 42 },
		{ 468, 139, 42 },
		{ 468, 140, 42 },
		{ 468, 141, 42 },
		{ 468, 142, 42 },
		{ 468, 143, 42 },
		{ 468, 144, 42 },
		{ 468, 145, 42 },
		{ 468, 146, 42 },
		{ 468, 147, 42 },
		{ 468, 148, 42 },
		{ 468, 149, 42 },
		{ 468, 150, 42 },
		{ 468, 151, 42 },
		{ 468, 152, 42 },
		{ 468, 153, 42 },
		{ 468, 154, 42 },
		{ 468, 155, 42 },
		{ 468, 156, 42 },
		{ 468, 157, 42 },
		{ 468, 158, 42 },
		{ 468, 159, 42 },
		{ 468, 160, 42 },
		{ 468, 161, 42 },
		{ 468, 162, 42 },
		{ 468, 163, 42 },
		{ 468, 164, 42 },
		{ 468, 165, 42 },
		{ 468, 166, 42 },
		{ 468, 167, 42 },
		{ 468, 168, 42 },
		{ 468, 169, 42 },
		{ 468, 170, 42 },
		{ 468, 171, 42 },
		{ 468, 172, 42 },
		{ 468, 173, 42 },
		{ 468, 174, 42 },
		{ 468, 175, 42 },
		{ 468, 176, 42 },
		{ 468, 177, 42 },
		{ 469, 125, 42 },
		{ 469, 126, 42 },
		{ 469, 127, 42 },
		{ 469, 128, 42 },
		{ 469, 129, 42 },
		{ 469, 130, 42 },
		{ 469, 131, 42 },
		{ 469, 132, 42 },
		{ 469, 133, 42 },
		{ 469, 134, 42 },
		{ 469, 135, 42 },
		{ 469, 136, 42 },
		{ 469, 137, 42 },
		{ 469, 138, 42 },
		{ 469, 139, 42 },
		{ 469, 140, 42 },
		{ 469, 141, 42 },
		{ 469, 142, 42 },
		{ 469, 143, 42 },
		{ 469, 144, 42 },
		{ 469, 145, 42 },
		{ 469, 146, 42 },
		{ 469, 147, 42 },
		{ 469, 148, 42 },
		{ 469, 149, 42 },
		{ 469, 150, 42 },
		{ 469, 151, 42 },
		{ 469, 152, 42 },
		{ 469, 153, 42 },
		{ 469, 154, 42 },
		{ 469, 155, 42 },
		{ 469, 156, 42 },
		{ 469, 157, 42 },
		{ 469, 158, 42 },
		{ 469, 159, 42 },
		{ 469, 160, 42 },
		{ 469, 161, 42 },
		{ 469, 162, 42 },
		{ 469, 163, 42 },
		{ 469, 164, 42 },
		{ 469, 165, 42 },
		{ 469, 166, 42 },
		{ 469, 167, 42 },
		{ 469, 168, 42 },
		{ 469, 169, 42 },
		{ 469, 170, 42 },
		{ 469, 171, 42 },
		{ 469, 172, 42 },
		{ 469, 173, 42 },
		{ 469, 174, 42 },
		{ 469, 175, 42 },
		{ 469, 176, 42 },
		{ 469, 177, 42 },
		{ 470, 124, 42 },
		{ 470, 125, 42 },
		{ 470, 126, 42 },
		{ 470, 127, 42 },
		{ 470, 128, 42 },
		{ 470, 129, 42 },
		{ 470, 130, 42 },
		{ 470, 131, 42 },
		{ 470, 132, 42 },
		{ 470, 133, 42 },
		{ 470, 134, 42 },
		{ 470, 135, 42 },
		{ 470, 136, 42 },
		{ 470, 137, 42 },
		{ 470, 138, 42 },
		{ 470, 139, 42 },
		{ 470, 140, 42 },
		{ 470, 141, 42 },
		{ 470, 142, 42 },
		{ 470, 143, 42 },
		{ 470, 144, 42 },
		{ 470, 145, 42 },
		{ 470, 146, 42 },
		{ 470, 147, 42 },
		{ 470, 148, 42 },
		{ 470, 149, 42 },
		{ 470, 150, 42 },
		{ 470, 151, 42 },
		{ 470, 152, 42 },
		{ 470, 153, 42 },
		{ 470, 154, 42 },
		{ 470, 155, 42 },
		{ 470, 156, 42 },
		{ 470, 157, 42 },
		{ 470, 158, 42 },
		{ 470, 159, 42 },
		{ 470, 160, 42 },
		{ 470, 161, 42 },
		{ 470, 162, 42 },
		{ 470, 163, 42 },
		{ 470, 164, 42 },
		{ 470, 165, 42 },
		{ 470, 166, 42 },
		{ 470, 167, 42 },
		{ 470, 168, 42 },
		{ 470, 169, 42 },
		{ 470, 170, 42 },
		{ 470, 171, 42 },
		{ 470, 172, 42 },
		{ 470, 173, 42 },
		{ 470, 174, 42 },
		{ 470, 175, 42 },
		{ 470, 176, 42 },
		{ 470, 177, 42 },
		{ 471, 124, 42 },
		{ 471, 125, 42 },
		{ 471, 126, 42 },
		{ 471, 127, 42 },
		{ 471, 128, 42 },
		{ 471, 129, 42 },
		{ 471, 130, 42 },
		{ 471, 131, 42 },
		{ 471, 132, 42 },
		{ 471, 133, 42 },
		{ 471, 134, 42 },
		{ 471, 135, 42 },
		{ 471, 136, 42 },
		{ 471, 137, 42 },
		{ 471, 138, 42 },
		{ 471, 139, 42 },
		{ 471, 140, 42 },
		{ 471, 141, 42 },
		{ 471, 142, 42 },
		{ 471, 143, 42 },
		{ 471, 144, 42 },
		{ 471, 145, 42 },
		{ 471, 146, 42 },
		{ 471, 147, 42 },
		{ 471, 148, 42 },
		{ 471, 149, 42 },
		{ 471, 150, 42 },
		{ 471, 151, 42 },
		{ 471, 152, 42 },
		{ 471, 153, 42 },
		{ 471, 154, 42 },
		{ 471, 155, 42 },
		{ 471, 156, 42 },
		{ 471, 157, 42 },
		{ 471, 158, 42 },
		{ 471, 159, 42 },
		{ 471, 160, 42 },
		{ 471, 161, 42 },
		{ 471, 162, 42 },
		{ 471, 163, 42 },
		{ 471, 164, 42 },
		{ 471, 165, 42 },
		{ 471, 166, 42 },
		{ 471, 167, 42 },
		{ 471, 168, 42 },
		{ 471, 169, 42 },
		{ 471, 170, 42 },
		{ 471, 171, 42 },
		{ 471, 172, 42 },
		{ 471, 173, 42 },
		{ 471, 174, 42 },
		{ 471, 175, 42 },
		{ 471, 176, 42 },
		{ 472, 123, 42 },
		{ 472, 124, 42 },
		{ 472, 125, 42 },
		{ 472, 126, 42 },
		{ 472, 127, 42 },
		{ 472, 128, 42 },
		{ 472, 129, 42 },
		{ 472, 130, 42 },
		{ 472, 131, 42 },
		{ 472, 132, 42 },
		{ 472, 133, 42 },
		{ 472, 134, 42 },
		{ 472, 135, 42 },
		{ 472, 136, 42 },
		{ 472, 137, 42 },
		{ 472, 138, 42 },
		{ 472, 139, 42 },
		{ 472, 140, 42 },
		{ 472, 141, 42 },
		{ 472, 142, 42 },
		{ 472, 143, 42 },
		{ 472, 144, 42 },
		{ 472, 145, 42 },
		{ 472, 146, 42 },
		{ 472, 147, 42 },
		{ 472, 148, 42 },
		{ 472, 149, 42 },
		{ 472, 150, 42 },
		{ 472, 151, 42 },
		{ 472, 152, 42 },
		{ 472, 153, 42 },
		{ 472, 154, 42 },
		{ 472, 155, 42 },
		{ 472, 156, 42 },
		{ 472, 157, 42 },
		{ 472, 158, 42 },
		{ 472, 159, 42 },
		{ 472, 160, 42 },
		{ 472, 161, 42 },
		{ 472, 162, 42 },
		{ 472, 163, 42 },
		{ 472, 164, 42 },
		{ 472, 165, 42 },
		{ 472, 166, 42 },
		{ 472, 167, 42 },
		{ 472, 168, 42 },
		{ 472, 169, 42 },
		{ 472, 170, 42 },
		{ 472, 171, 42 },
		{ 472, 172, 42 },
		{ 472, 173, 42 },
		{ 472, 174, 42 },
		{ 472, 175, 42 },
		{ 472, 176, 42 },
		{ 473, 123, 42 },
		{ 473, 124, 42 },
		{ 473, 125, 42 },
		{ 473, 126, 42 },
		{ 473, 127, 42 },
		{ 473, 128, 42 },
		{ 473, 129, 42 },
		{ 473, 130, 42 },
		{ 473, 131, 42 },
		{ 473, 132, 42 },
		{ 473, 133, 42 },
		{ 473, 134, 42 },
		{ 473, 135, 42 },
		{ 473, 136, 42 },
		{ 473, 137, 42 },
		{ 473, 138, 42 },
		{ 473, 139, 42 },
		{ 473, 140, 42 },
		{ 473, 141, 42 },
		{ 473, 142, 42 },
		{ 473, 143, 42 },
		{ 473, 144, 42 },
		{ 473, 145, 42 },
		{ 473, 146, 42 },
		{ 473, 147, 42 },
		{ 473, 148, 42 },
		{ 473, 149, 42 },
		{ 473, 150, 42 },
		{ 473, 151, 42 },
		{ 473, 152, 42 },
		{ 473, 153, 42 },
		{ 473, 154, 42 },
		{ 473, 155, 42 },
		{ 473, 156, 42 },
		{ 473, 157, 42 },
		{ 473, 158, 42 },
		{ 473, 159, 42 },
		{ 473, 160, 42 },
		{ 473, 161, 42 },
		{ 473, 162, 42 },
		{ 473, 163, 42 },
		{ 473, 164, 42 },
		{ 473, 165, 42 },
		{ 473, 166, 42 },
		{ 473, 167, 42 },
		{ 473, 168, 42 },
		{ 473, 169, 42 },
		{ 473, 170, 42 },
		{ 473, 171, 42 },
		{ 473, 172, 42 },
		{ 473, 173, 42 },
		{ 473, 174, 42 },
		{ 473, 175, 42 },
		{ 474, 122, 42 },
		{ 474, 123, 42 },
		{ 474, 124, 42 },
		{ 474, 125, 42 },
		{ 474, 126, 42 },
		{ 474, 127, 42 },
		{ 474, 128, 42 },
		{ 474, 129, 42 },
		{ 474, 130, 42 },
		{ 474, 131, 42 },
		{ 474, 132, 42 },
		{ 474, 133, 42 },
		{ 474, 134, 42 },
		{ 474, 135, 42 },
		{ 474, 136, 42 },
		{ 474, 137, 42 },
		{ 474, 138, 42 },
		{ 474, 139, 42 },
		{ 474, 140, 42 },
		{ 474, 141, 42 },
		{ 474, 142, 42 },
		{ 474, 143, 42 },
		{ 474, 144, 42 },
		{ 474, 145, 42 },
		{ 474, 146, 42 },
		{ 474, 147, 42 },
		{ 474, 148, 42 },
		{ 474, 149, 42 },
		{ 474, 150, 42 },
		{ 474, 151, 42 },
		{ 474, 152, 42 },
		{ 474, 153, 42 },
		{ 474, 154, 42 },
		{ 474, 155, 42 },
		{ 474, 156, 42 },
		{ 474, 157, 42 },
		{ 474, 158, 42 },
		{ 474, 159, 42 },
		{ 474, 160, 42 },
		{ 474, 161, 42 },
		{ 474, 162, 42 },
		{ 474, 163, 42 },
		{ 474, 164, 42 },
		{ 474, 165, 42 },
		{ 474, 166, 42 },
		{ 474, 167, 42 },
		{ 474, 168, 42 },
		{ 474, 169, 42 },
		{ 474, 170, 42 },
		{ 474, 171, 42 },
		{ 474, 172, 42 },
		{ 474, 173, 42 },
		{ 474, 174, 42 },
		{ 474, 175, 42 },
		{ 475, 121, 42 },
		{ 475, 122, 42 },
		{ 475, 123, 42 },
		{ 475, 124, 42 },
		{ 475, 125, 42 },
		{ 475, 126, 42 },
		{ 475, 127, 42 },
		{ 475, 128, 42 },
		{ 475, 129, 42 },
		{ 475, 130, 42 },
		{ 475, 131, 42 },
		{ 475, 132, 42 },
		{ 475, 133, 42 },
		{ 475, 134, 42 },
		{ 475, 135, 42 },
		{ 475, 136, 42 },
		{ 475, 137, 42 },
		{ 475, 138, 42 },
		{ 475, 139, 42 },
		{ 475, 140, 42 },
		{ 475, 141, 42 },
		{ 475, 142, 42 },
		{ 475, 143, 42 },
		{ 475, 144, 42 },
		{ 475, 145, 42 },
		{ 475, 146, 42 },
		{ 475, 147, 42 },
		{ 475, 148, 42 },
		{ 475, 149, 42 },
		{ 475, 150, 42 },
		{ 475, 151, 42 },
		{ 475, 152, 42 },
		{ 475, 153, 42 },
		{ 475, 154, 42 },
		{ 475, 155, 42 },
		{ 475, 156, 42 },
		{ 475, 157, 42 },
		{ 475, 158, 42 },
		{ 475, 159, 42 },
		{ 475, 160, 42 },
		{ 475, 161, 42 },
		{ 475, 162, 42 },
		{ 475, 163, 42 },
		{ 475, 164, 42 },
		{ 475, 165, 42 },
		{ 475, 166, 42 },
		{ 475, 167, 42 },
		{ 475, 168, 42 },
		{ 475, 169, 42 },
		{ 475, 170, 42 },
		{ 475, 171, 42 },
		{ 475, 172, 42 },
		{ 475, 173, 42 },
		{ 475, 174, 42 },
		{ 475, 175, 42 },
		{ 476, 121, 42 },
		{ 476, 122, 42 },
		{ 476, 123, 42 },
		{ 476, 124, 42 },
		{ 476, 125, 42 },
		{ 476, 126, 42 },
		{ 476, 127, 42 },
		{ 476, 128, 42 },
		{ 476, 129, 42 },
		{ 476, 130, 42 },
		{ 476, 131, 42 },
		{ 476, 132, 42 },
		{ 476, 133, 42 },
		{ 476, 134, 42 },
		{ 476, 135, 42 },
		{ 476, 136, 42 },
		{ 476, 137, 42 },
		{ 476, 138, 42 },
		{ 476, 139, 42 },
		{ 476, 140, 42 },
		{ 476, 141, 42 },
		{ 476, 142, 42 },
		{ 476, 143, 42 },
		{ 476, 144, 42 },
		{ 476, 145, 42 },
		{ 476, 146, 42 },
		{ 476, 147, 42 },
		{ 476, 148, 42 },
		{ 476, 149, 42 },
		{ 476, 150, 42 },
		{ 476, 151, 42 },
		{ 476, 152, 42 },
		{ 476, 153, 42 },
		{ 476, 154, 42 },
		{ 476, 155, 42 },
		{ 476, 156, 42 },
		{ 476, 157, 42 },
		{ 476, 158, 42 },
		{ 476, 159, 42 },
		{ 476, 160, 42 },
		{ 476, 161, 42 },
		{ 476, 162, 42 },
		{ 476, 163, 42 },
		{ 476, 164, 42 },
		{ 476, 165, 42 },
		{ 476, 166, 42 },
		{ 476, 167, 42 },
		{ 476, 168, 42 },
		{ 476, 169, 42 },
		{ 476, 170, 42 },
		{ 476, 171, 42 },
		{ 476, 172, 42 },
		{ 476, 173, 42 },
		{ 476, 174, 42 },
		{ 477, 120, 42 },
		{ 477, 121, 42 },
		{ 477, 122, 42 },
		{ 477, 123, 42 },
		{ 477, 124, 42 },
		{ 477, 125, 42 },
		{ 477, 126, 42 },
		{ 477, 127, 42 },
		{ 477, 128, 42 },
		{ 477, 129, 42 },
		{ 477, 130, 42 },
		{ 477, 131, 42 },
		{ 477, 132, 42 },
		{ 477, 133, 42 },
		{ 477, 134, 42 },
		{ 477, 135, 42 },
		{ 477, 136, 42 },
		{ 477, 137, 42 },
		{ 477, 138, 42 },
		{ 477, 139, 42 },
		{ 477, 140, 42 },
		{ 477, 141, 42 },
		{ 477, 142, 42 },
		{ 477, 143, 42 },
		{ 477, 144, 42 },
		{ 477, 145, 42 },
		{ 477, 146, 42 },
		{ 477, 147, 42 },
		{ 477, 148, 42 },
		{ 477, 149, 42 },
		{ 477, 150, 42 },
		{ 477, 151, 42 },
		{ 477, 152, 42 },
		{ 477, 153, 42 },
		{ 477, 154, 42 },
		{ 477, 155, 42 },
		{ 477, 156, 42 },
		{ 477, 157, 42 },
		{ 477, 158, 42 },
		{ 477, 159, 42 },
		{ 477, 160, 42 },
		{ 477, 161, 42 },
		{ 477, 162, 42 },
		{ 477, 163, 42 },
		{ 477, 164, 42 },
		{ 477, 165, 42 },
		{ 477, 166, 42 },
		{ 477, 167, 42 },
		{ 477, 168, 42 },
		{ 477, 169, 42 },
		{ 477, 170, 42 },
		{ 477, 171, 42 },
		{ 477, 172, 42 },
		{ 477, 173, 42 },
		{ 477, 174, 42 },
		{ 478, 119, 42 },
		{ 478, 120, 42 },
		{ 478, 121, 42 },
		{ 478, 122, 42 },
		{ 478, 123, 42 },
		{ 478, 124, 42 },
		{ 478, 125, 42 },
		{ 478, 126, 42 },
		{ 478, 127, 42 },
		{ 478, 128, 42 },
		{ 478, 129, 42 },
		{ 478, 130, 42 },
		{ 478, 131, 42 },
		{ 478, 132, 42 },
		{ 478, 133, 42 },
		{ 478, 134, 42 },
		{ 478, 135, 42 },
		{ 478, 136, 42 },
		{ 478, 137, 42 },
		{ 478, 138, 42 },
		{ 478, 139, 42 },
		{ 478, 140, 42 },
		{ 478, 141, 42 },
		{ 478, 142, 42 },
		{ 478, 143, 42 },
		{ 478, 144, 42 },
		{ 478, 145, 42 },
		{ 478, 146, 42 },
		{ 478, 147, 42 },
		{ 478, 148, 42 },
		{ 478, 149, 42 },
		{ 478, 150, 42 },
		{ 478, 151, 42 },
		{ 478, 152, 42 },
		{ 478, 153, 42 },
		{ 478, 154, 42 },
		{ 478, 155, 42 },
		{ 478, 156, 42 },
		{ 478, 157, 42 },
		{ 478, 158, 42 },
		{ 478, 159, 42 },
		{ 478, 160, 42 },
		{ 478, 161, 42 },
		{ 478, 162, 42 },
		{ 478, 163, 42 },
		{ 478, 164, 42 },
		{ 478, 165, 42 },
		{ 478, 166, 42 },
		{ 478, 167, 42 },
		{ 478, 168, 42 },
		{ 478, 169, 42 },
		{ 478, 170, 42 },
		{ 478, 171, 42 },
		{ 478, 172, 42 },
		{ 478, 173, 42 },
		{ 478, 174, 42 },
		{ 479, 119, 42 },
		{ 479, 120, 42 },
		{ 479, 121, 42 },
		{ 479, 122, 42 },
		{ 479, 123, 42 },
		{ 479, 124, 42 },
		{ 479, 125, 42 },
		{ 479, 126, 42 },
		{ 479, 127, 42 },
		{ 479, 128, 42 },
		{ 479, 129, 42 },
		{ 479, 130, 42 },
		{ 479, 131, 42 },
		{ 479, 132, 42 },
		{ 479, 133, 42 },
		{ 479, 134, 42 },
		{ 479, 135, 42 },
		{ 479, 136, 42 },
		{ 479, 137, 42 },
		{ 479, 138, 42 },
		{ 479, 139, 42 },
		{ 479, 140, 42 },
		{ 479, 141, 42 },
		{ 479, 142, 42 },
		{ 479, 143, 42 },
		{ 479, 144, 42 },
		{ 479, 145, 42 },
		{ 479, 146, 42 },
		{ 479, 147, 42 },
		{ 479, 148, 42 },
		{ 479, 149, 42 },
		{ 479, 150, 42 },
		{ 479, 151, 42 },
		{ 479, 152, 42 },
		{ 479, 153, 42 },
		{ 479, 154, 42 },
		{ 479, 155, 42 },
		{ 479, 156, 42 },
		{ 479, 157, 42 },
		{ 479, 158, 42 },
		{ 479, 159, 42 },
		{ 479, 160, 42 },
		{ 479, 161, 42 },
		{ 479, 162, 42 },
		{ 479, 163, 42 },
		{ 479, 164, 42 },
		{ 479, 165, 42 },
		{ 479, 166, 42 },
		{ 479, 167, 42 },
		{ 479, 168, 42 },
		{ 479, 169, 42 },
		{ 479, 170, 42 },
		{ 479, 171, 42 },
		{ 479, 172, 42 },
		{ 479, 173, 42 },
		{ 480, 118, 42 },
		{ 480, 119, 42 },
		{ 480, 120, 42 },
		{ 480, 121, 42 },
		{ 480, 122, 42 },
		{ 480, 123, 42 },
		{ 480, 124, 42 },
		{ 480, 125, 42 },
		{ 480, 126, 42 },
		{ 480, 127, 42 },
		{ 480, 128, 42 },
		{ 480, 129, 42 },
		{ 480, 130, 42 },
		{ 480, 131, 42 },
		{ 480, 132, 42 },
		{ 480, 133, 42 },
		{ 480, 134, 42 },
		{ 480, 135, 42 },
		{ 480, 136, 42 },
		{ 480, 137, 42 },
		{ 480, 138, 42 },
		{ 480, 139, 42 },
		{ 480, 140, 42 },
		{ 480, 141, 42 },
		{ 480, 142, 42 },
		{ 480, 143, 42 },
		{ 480, 144, 42 },
		{ 480, 145, 42 },
		{ 480, 146, 42 },
		{ 480, 147, 42 },
		{ 480, 148, 42 },
		{ 480, 149, 42 },
		{ 480, 150, 42 },
		{ 480, 151, 42 },
		{ 480, 152, 42 },
		{ 480, 153, 42 },
		{ 480, 154, 42 },
		{ 480, 155, 42 },
		{ 480, 156, 42 },
		{ 480, 157, 42 },
		{ 480, 158, 42 },
		{ 480, 159, 42 },
		{ 480, 160, 42 },
		{ 480, 161, 42 },
		{ 480, 162, 42 },
		{ 480, 163, 42 },
		{ 480, 164, 42 },
		{ 480, 165, 42 },
		{ 480, 166, 42 },
		{ 480, 167, 42 },
		{ 480, 168, 42 },
		{ 480, 169, 42 },
		{ 480, 170, 42 },
		{ 480, 171, 42 },
		{ 480, 172, 42 },
		{ 480, 173, 42 },
		{ 481, 118, 42 },
		{ 481, 119, 42 },
		{ 481, 120, 42 },
		{ 481, 121, 42 },
		{ 481, 122, 42 },
		{ 481, 123, 42 },
		{ 481, 124, 42 },
		{ 481, 125, 42 },
		{ 481, 126, 42 },
		{ 481, 127, 42 },
		{ 481, 128, 42 },
		{ 481, 129, 42 },
		{ 481, 130, 42 },
		{ 481, 131, 42 },
		{ 481, 132, 42 },
		{ 481, 133, 42 },
		{ 481, 134, 42 },
		{ 481, 135, 42 },
		{ 481, 136, 42 },
		{ 481, 137, 42 },
		{ 481, 138, 42 },
		{ 481, 139, 42 },
		{ 481, 140, 42 },
		{ 481, 141, 42 },
		{ 481, 142, 42 },
		{ 481, 143, 42 },
		{ 481, 144, 42 },
		{ 481, 145, 42 },
		{ 481, 146, 42 },
		{ 481, 147, 42 },
		{ 481, 148, 42 },
		{ 481, 149, 42 },
		{ 481, 150, 42 },
		{ 481, 151, 42 },
		{ 481, 152, 42 },
		{ 481, 153, 42 },
		{ 481, 154, 42 },
		{ 481, 155, 42 },
		{ 481, 156, 42 },
		{ 481, 157, 42 },
		{ 481, 158, 42 },
		{ 481, 159, 42 },
		{ 481, 160, 42 },
		{ 481, 161, 42 },
		{ 481, 162, 42 },
		{ 481, 163, 42 },
		{ 481, 164, 42 },
		{ 481, 165, 42 },
		{ 481, 166, 42 },
		{ 481, 167, 42 },
		{ 481, 168, 42 },
		{ 481, 169, 42 },
		{ 481, 170, 42 },
		{ 481, 171, 42 },
		{ 481, 172, 42 },
		{ 482, 117, 42 },
		{ 482, 118, 42 },
		{ 482, 119, 42 },
		{ 482, 120, 42 },
		{ 482, 121, 42 },
		{ 482, 122, 42 },
		{ 482, 123, 42 },
		{ 482, 124, 42 },
		{ 482, 125, 42 },
		{ 482, 126, 42 },
		{ 482, 127, 42 },
		{ 482, 128, 42 },
		{ 482, 129, 42 },
		{ 482, 130, 42 },
		{ 482, 131, 42 },
		{ 482, 132, 42 },
		{ 482, 133, 42 },
		{ 482, 134, 42 },
		{ 482, 135, 42 },
		{ 482, 136, 42 },
		{ 482, 137, 42 },
		{ 482, 138, 42 },
		{ 482, 139, 42 },
		{ 482, 140, 42 },
		{ 482, 141, 42 },
		{ 482, 142, 42 },
		{ 482, 143, 42 },
		{ 482, 144, 42 },
		{ 482, 145, 42 },
		{ 482, 146, 42 },
		{ 482, 147, 42 },
		{ 482, 148, 42 },
		{ 482, 149, 42 },
		{ 482, 150, 42 },
		{ 482, 151, 42 },
		{ 482, 152, 42 },
		{ 482, 153, 42 },
		{ 482, 154, 42 },
		{ 482, 155, 42 },
		{ 482, 156, 42 },
		{ 482, 157, 42 },
		{ 482, 158, 42 },
		{ 482, 159, 42 },
		{ 482, 160, 42 },
		{ 482, 161, 42 },
		{ 482, 162, 42 },
		{ 482, 163, 42 },
		{ 482, 164, 42 },
		{ 482, 165, 42 },
		{ 482, 166, 42 },
		{ 482, 167, 42 },
		{ 482, 168, 42 },
		{ 482, 169, 42 },
		{ 482, 170, 42 },
		{ 482, 171, 42 },
		{ 482, 172, 42 },
		{ 483, 116, 42 },
		{ 483, 117, 42 },
		{ 483, 118, 42 },
		{ 483, 119, 42 },
		{ 483, 120, 42 },
		{ 483, 121, 42 },
		{ 483, 122, 42 },
		{ 483, 123, 42 },
		{ 483, 124, 42 },
		{ 483, 125, 42 },
		{ 483, 126, 42 },
		{ 483, 127, 42 },
		{ 483, 128, 42 },
		{ 483, 129, 42 },
		{ 483, 130, 42 },
		{ 483, 131, 42 },
		{ 483, 132, 42 },
		{ 483, 133, 42 },
		{ 483, 134, 42 },
		{ 483, 135, 42 },
		{ 483, 136, 42 },
		{ 483, 137, 42 },
		{ 483, 138, 42 },
		{ 483, 139, 42 },
		{ 483, 140, 42 },
		{ 483, 141, 42 },
		{ 483, 142, 42 },
		{ 483, 143, 42 },
		{ 483, 144, 42 },
		{ 483, 145, 42 },
		{ 483, 146, 42 },
		{ 483, 147, 42 },
		{ 483, 148, 42 },
		{ 483, 149, 42 },
		{ 483, 150, 42 },
		{ 483, 151, 42 },
		{ 483, 152, 42 },
		{ 483, 153, 42 },
		{ 483, 154, 42 },
		{ 483, 155, 42 },
		{ 483, 156, 42 },
		{ 483, 157, 42 },
		{ 483, 158, 42 },
		{ 483, 159, 42 },
		{ 483, 160, 42 },
		{ 483, 161, 42 },
		{ 483, 162, 42 },
		{ 483, 163, 42 },
		{ 483, 164, 42 },
		{ 483, 165, 42 },
		{ 483, 166, 42 },
		{ 483, 167, 42 },
		{ 483, 168, 42 },
		{ 483, 169, 42 },
		{ 483, 170, 42 },
		{ 483, 171, 42 },
		{ 483, 172, 42 },
		{ 484, 116, 42 },
		{ 484, 117, 42 },
		{ 484, 118, 42 },
		{ 484, 119, 42 },
		{ 484, 120, 42 },
		{ 484, 121, 42 },
		{ 484, 122, 42 },
		{ 484, 123, 42 },
		{ 484, 124, 42 },
		{ 484, 125, 42 },
		{ 484, 126, 42 },
		{ 484, 127, 42 },
		{ 484, 128, 42 },
		{ 484, 129, 42 },
		{ 484, 130, 42 },
		{ 484, 131, 42 },
		{ 484, 132, 42 },
		{ 484, 133, 42 },
		{ 484, 134, 42 },
		{ 484, 135, 42 },
		{ 484, 136, 42 },
		{ 484, 137, 42 },
		{ 484, 138, 42 },
		{ 484, 139, 42 },
		{ 484, 140, 42 },
		{ 484, 141, 42 },
		{ 484, 142, 42 },
		{ 484, 143, 42 },
		{ 484, 144, 42 },
		{ 484, 145, 42 },
		{ 484, 146, 42 },
		{ 484, 147, 42 },
		{ 484, 148, 42 },
		{ 484, 149, 42 },
		{ 484, 150, 42 },
		{ 484, 151, 42 },
		{ 484, 152, 42 },
		{ 484, 153, 42 },
		{ 484, 154, 42 },
		{ 484, 155, 42 },
		{ 484, 156, 42 },
		{ 484, 157, 42 },
		{ 484, 158, 42 },
		{ 484, 159, 42 },
		{ 484, 160, 42 },
		{ 484, 161, 42 },
		{ 484, 162, 42 },
		{ 484, 163, 42 },
		{ 484, 164, 42 },
		{ 484, 165, 42 },
		{ 484, 166, 42 },
		{ 484, 167, 42 },
		{ 484, 168, 42 },
		{ 484, 169, 42 },
		{ 484, 170, 42 },
		{ 484, 171, 42 },
		{ 485, 115, 42 },
		{ 485, 116, 42 },
		{ 485, 117, 42 },
		{ 485, 118, 42 },
		{ 485, 119, 42 },
		{ 485, 120, 42 },
		{ 485, 121, 42 },
		{ 485, 122, 42 },
		{ 485, 123, 42 },
		{ 485, 124, 42 },
		{ 485, 125, 42 },
		{ 485, 126, 42 },
		{ 485, 127, 42 },
		{ 485, 128, 42 },
		{ 485, 129, 42 },
		{ 485, 130, 42 },
		{ 485, 131, 42 },
		{ 485, 132, 42 },
		{ 485, 133, 42 },
		{ 485, 134, 42 },
		{ 485, 135, 42 },
		{ 485, 136, 42 },
		{ 485, 137, 42 },
		{ 485, 138, 42 },
		{ 485, 139, 42 },
		{ 485, 140, 42 },
		{ 485, 141, 42 },
		{ 485, 142, 42 },
		{ 485, 143, 42 },
		{ 485, 144, 42 },
		{ 485, 145, 42 },
		{ 485, 146, 42 },
		{ 485, 147, 42 },
		{ 485, 148, 42 },
		{ 485, 149, 42 },
		{ 485, 150, 42 },
		{ 485, 151, 42 },
		{ 485, 152, 42 },
		{ 485, 153, 42 },
		{ 485, 154, 42 },
		{ 485, 155, 42 },
		{ 485, 156, 42 },
		{ 485, 157, 42 },
		{ 485, 158, 42 },
		{ 485, 159, 42 },
		{ 485, 160, 42 },
		{ 485, 161, 42 },
		{ 485, 162, 42 },
		{ 485, 163, 42 },
		{ 485, 164, 42 },
		{ 485, 165, 42 },
		{ 485, 166, 42 },
		{ 485, 167, 42 },
		{ 485, 168, 42 },
		{ 485, 169, 42 },
		{ 485, 170, 42 },
		{ 485, 171, 42 },
		{ 486, 114, 42 },
		{ 486, 115, 42 },
		{ 486, 116, 42 },
		{ 486, 117, 42 },
		{ 486, 118, 42 },
		{ 486, 119, 42 },
		{ 486, 120, 42 },
		{ 486, 121, 42 },
		{ 486, 122, 42 },
		{ 486, 123, 42 },
		{ 486, 124, 42 },
		{ 486, 125, 42 },
		{ 486, 126, 42 },
		{ 486, 127, 42 },
		{ 486, 128, 42 },
		{ 486, 129, 42 },
		{ 486, 130, 42 },
		{ 486, 131, 42 },
		{ 486, 132, 42 },
		{ 486, 133, 42 },
		{ 486, 134, 42 },
		{ 486, 135, 42 },
		{ 486, 136, 42 },
		{ 486, 137, 42 },
		{ 486, 138, 42 },
		{ 486, 139, 42 },
		{ 486, 140, 42 },
		{ 486, 141, 42 },
		{ 486, 142, 42 },
		{ 486, 143, 42 },
		{ 486, 144, 42 },
		{ 486, 145, 42 },
		{ 486, 146, 42 },
		{ 486, 147, 42 },
		{ 486, 148, 42 },
		{ 486, 149, 42 },
		{ 486, 150, 42 },
		{ 486, 151, 42 },
		{ 486, 152, 42 },
		{ 486, 153, 42 },
		{ 486, 154, 42 },
		{ 486, 155, 42 },
		{ 486, 156, 42 },
		{ 486, 157, 42 },
		{ 486, 158, 42 },
		{ 486, 159, 42 },
		{ 486, 160, 42 },
		{ 486, 161, 42 },
		{ 486, 162, 42 },
		{ 486, 163, 42 },
		{ 486, 164, 42 },
		{ 486, 165, 42 },
		{ 486, 166, 42 },
		{ 486, 167, 42 },
		{ 486, 168, 42 },
		{ 486, 169, 42 },
		{ 486, 170, 42 },
		{ 486, 171, 42 },
		{ 487, 114, 42 },
		{ 487, 115, 42 },
		{ 487, 116, 42 },
		{ 487, 117, 42 },
		{ 487, 118, 42 },
		{ 487, 119, 42 },
		{ 487, 120, 42 },
		{ 487, 121, 42 },
		{ 487, 122, 42 },
		{ 487, 123, 42 },
		{ 487, 124, 42 },
		{ 487, 125, 42 },
		{ 487, 126, 42 },
		{ 487, 127, 42 },
		{ 487, 128, 42 },
		{ 487, 129, 42 },
		{ 487, 130, 42 },
		{ 487, 131, 42 },
		{ 487, 132, 42 },
		{ 487, 133, 42 },
		{ 487, 134, 42 },
		{ 487, 135, 42 },
		{ 487, 136, 42 },
		{ 487, 137, 42 },
		{ 487, 138, 42 },
		{ 487, 139, 42 },
		{ 487, 140, 42 },
		{ 487, 141, 42 },
		{ 487, 142, 42 },
		{ 487, 143, 42 },
		{ 487, 144, 42 },
		{ 487, 145, 42 },
		{ 487, 146, 42 },
		{ 487, 147, 42 },
		{ 487, 148, 42 },
		{ 487, 149, 42 },
		{ 487, 150, 42 },
		{ 487, 151, 42 },
		{ 487, 152, 42 },
		{ 487, 153, 42 },
		{ 487, 154, 42 },
		{ 487, 155, 42 },
		{ 487, 156, 42 },
		{ 487, 157, 42 },
		{ 487, 158, 42 },
		{ 487, 159, 42 },
		{ 487, 160, 42 },
		{ 487, 161, 42 },
		{ 487, 162, 42 },
		{ 487, 163, 42 },
		{ 487, 164, 42 },
		{ 487, 165, 42 },
		{ 487, 166, 42 },
		{ 487, 167, 42 },
		{ 487, 168, 42 },
		{ 487, 169, 42 },
		{ 487, 170, 42 },
		{ 488, 113, 42 },
		{ 488, 114, 42 },
		{ 488, 115, 42 },
		{ 488, 116, 42 },
		{ 488, 117, 42 },
		{ 488, 118, 42 },
		{ 488, 119, 42 },
		{ 488, 120, 42 },
		{ 488, 121, 42 },
		{ 488, 122, 42 },
		{ 488, 123, 42 },
		{ 488, 124, 42 },
		{ 488, 125, 42 },
		{ 488, 126, 42 },
		{ 488, 127, 42 },
		{ 488, 128, 42 },
		{ 488, 129, 42 },
		{ 488, 130, 42 },
		{ 488, 131, 42 },
		{ 488, 132, 42 },
		{ 488, 133, 42 },
		{ 488, 134, 42 },
		{ 488, 135, 42 },
		{ 488, 136, 42 },
		{ 488, 137, 42 },
		{ 488, 138, 42 },
		{ 488, 139, 42 },
		{ 488, 140, 42 },
		{ 488, 141, 42 },
		{ 488, 142, 42 },
		{ 488, 143, 42 },
		{ 488, 144, 42 },
		{ 488, 145, 42 },
		{ 488, 146, 42 },
		{ 488, 147, 42 },
		{ 488, 148, 42 },
		{ 488, 149, 42 },
		{ 488, 150, 42 },
		{ 488, 151, 42 },
		{ 488, 152, 42 },
		{ 488, 153, 42 },
		{ 488, 154, 42 },
		{ 488, 155, 42 },
		{ 488, 156, 42 },
		{ 488, 157, 42 },
		{ 488, 158, 42 },
		{ 488, 159, 42 },
		{ 488, 160, 42 },
		{ 488, 161, 42 },
		{ 488, 162, 42 },
		{ 488, 163, 42 },
		{ 488, 164, 42 },
		{ 488, 165, 42 },
		{ 488, 166, 42 },
		{ 488, 167, 42 },
		{ 488, 168, 42 },
		{ 488, 169, 42 },
		{ 488, 170, 42 },
		{ 489, 113, 42 },
		{ 489, 114, 42 },
		{ 489, 115, 42 },
		{ 489, 116, 42 },
		{ 489, 117, 42 },
		{ 489, 118, 42 },
		{ 489, 119, 42 },
		{ 489, 120, 42 },
		{ 489, 121, 42 },
		{ 489, 122, 42 },
		{ 489, 123, 42 },
		{ 489, 124, 42 },
		{ 489, 125, 42 },
		{ 489, 126, 42 },
		{ 489, 127, 42 },
		{ 489, 128, 42 },
		{ 489, 129, 42 },
		{ 489, 130, 42 },
		{ 489, 131, 42 },
		{ 489, 132, 42 },
		{ 489, 133, 42 },
		{ 489, 134, 42 },
		{ 489, 135, 42 },
		{ 489, 136, 42 },
		{ 489, 137, 42 },
		{ 489, 138, 42 },
		{ 489, 139, 42 },
		{ 489, 140, 42 },
		{ 489, 141, 42 },
		{ 489, 142, 42 },
		{ 489, 143, 42 },
		{ 489, 144, 42 },
		{ 489, 145, 42 },
		{ 489, 146, 42 },
		{ 489, 147, 42 },
		{ 489, 148, 42 },
		{ 489, 149, 42 },
		{ 489, 150, 42 },
		{ 489, 151, 42 },
		{ 489, 152, 42 },
		{ 489, 153, 42 },
		{ 489, 154, 42 },
		{ 489, 155, 42 },
		{ 489, 156, 42 },
		{ 489, 157, 42 },
		{ 489, 158, 42 },
		{ 489, 159, 42 },
		{ 489, 160, 42 },
		{ 489, 161, 42 },
		{ 489, 162, 42 },
		{ 489, 163, 42 },
		{ 489, 164, 42 },
		{ 489, 165, 42 },
		{ 489, 166, 42 },
		{ 489, 167, 42 },
		{ 489, 168, 42 },
		{ 489, 169, 42 },
		{ 490, 112, 42 },
		{ 490, 113, 42 },
		{ 490, 114, 42 },
		{ 490, 115, 42 },
		{ 490, 116, 42 },
		{ 490, 117, 42 },
		{ 490, 118, 42 },
		{ 490, 119, 42 },
		{ 490, 120, 42 },
		{ 490, 121, 42 },
		{ 490, 122, 42 },
		{ 490, 123, 42 },
		{ 490, 124, 42 },
		{ 490, 125, 42 },
		{ 490, 126, 42 },
		{ 490, 127, 42 },
		{ 490, 128, 42 },
		{ 490, 129, 42 },
		{ 490, 130, 42 },
		{ 490, 131, 42 },
		{ 490, 132, 42 },
		{ 490, 133, 42 },
		{ 490, 134, 42 },
		{ 490, 135, 42 },
		{ 490, 136, 42 },
		{ 490, 137, 42 },
		{ 490, 138, 42 },
		{ 490, 139, 42 },
		{ 490, 140, 42 },
		{ 490, 141, 42 },
		{ 490, 142, 42 },
		{ 490, 143, 42 },
		{ 490, 144, 42 },
		{ 490, 145, 42 },
		{ 490, 146, 42 },
		{ 490, 147, 42 },
		{ 490, 148, 42 },
		{ 490, 149, 42 },
		{ 490, 150, 42 },
		{ 490, 151, 42 },
		{ 490, 152, 42 },
		{ 490, 153, 42 },
		{ 490, 154, 42 },
		{ 490, 155, 42 },
		{ 490, 156, 42 },
		{ 490, 157, 42 },
		{ 490, 158, 42 },
		{ 490, 159, 42 },
		{ 490, 160, 42 },
		{ 490, 161, 42 },
		{ 490, 162, 42 },
		{ 490, 163, 42 },
		{ 490, 164, 42 },
		{ 490, 165, 42 },
		{ 490, 166, 42 },
		{ 490, 167, 42 },
		{ 490, 168, 42 },
		{ 490, 169, 42 },
		{ 491, 111, 42 },
		{ 491, 112, 42 },
		{ 491, 113, 42 },
		{ 491, 114, 42 },
		{ 491, 115, 42 },
		{ 491, 116, 42 },
		{ 491, 117, 42 },
		{ 491, 118, 42 },
		{ 491, 119, 42 },
		{ 491, 120, 42 },
		{ 491, 121, 42 },
		{ 491, 122, 42 },
		{ 491, 123, 42 },
		{ 491, 124, 42 },
		{ 491, 125, 42 },
		{ 491, 126, 42 },
		{ 491, 127, 42 },
		{ 491, 128, 42 },
		{ 491, 129, 42 },
		{ 491, 130, 42 },
		{ 491, 131, 42 },
		{ 491, 132, 42 },
		{ 491, 133, 42 },
		{ 491, 134, 42 },
		{ 491, 135, 42 },
		{ 491, 136, 42 },
		{ 491, 137, 42 },
		{ 491, 138, 42 },
		{ 491, 139, 42 },
		{ 491, 140, 42 },
		{ 491, 141, 42 },
		{ 491, 142, 42 },
		{ 491, 143, 42 },
		{ 491, 144, 42 },
		{ 491, 145, 42 },
		{ 491, 146, 42 },
		{ 491, 147, 42 },
		{ 491, 148, 42 },
		{ 491, 149, 42 },
		{ 491, 150, 42 },
		{ 491, 151, 42 },
		{ 491, 152, 42 },
		{ 491, 153, 42 },
		{ 491, 154, 42 },
		{ 491, 155, 42 },
		{ 491, 156, 42 },
		{ 491, 157, 42 },
		{ 491, 158, 42 },
		{ 491, 159, 42 },
		{ 491, 160, 42 },
		{ 491, 161, 42 },
		{ 491, 162, 42 },
		{ 491, 163, 42 },
		{ 491, 164, 42 },
		{ 491, 165, 42 },
		{ 491, 166, 42 },
		{ 491, 167, 42 },
		{ 491, 168, 42 },
		{ 491, 169, 42 },
		{ 492, 111, 42 },
		{ 492, 112, 42 },
		{ 492, 113, 42 },
		{ 492, 114, 42 },
		{ 492, 115, 42 },
		{ 492, 116, 42 },
		{ 492, 117, 42 },
		{ 492, 118, 42 },
		{ 492, 119, 42 },
		{ 492, 120, 42 },
		{ 492, 121, 42 },
		{ 492, 122, 42 },
		{ 492, 123, 42 },
		{ 492, 124, 42 },
		{ 492, 125, 42 },
		{ 492, 126, 42 },
		{ 492, 127, 42 },
		{ 492, 128, 42 },
		{ 492, 129, 42 },
		{ 492, 130, 42 },
		{ 492, 131, 42 },
		{ 492, 132, 42 },
		{ 492, 133, 42 },
		{ 492, 134, 42 },
		{ 492, 135, 42 },
		{ 492, 136, 42 },
		{ 492, 137, 42 },
		{ 492, 138, 42 },
		{ 492, 139, 42 },
		{ 492, 140, 42 },
		{ 492, 141, 42 },
		{ 492, 142, 42 },
		{ 492, 143, 42 },
		{ 492, 144, 42 },
		{ 492, 145, 42 },
		{ 492, 146, 42 },
		{ 492, 147, 42 },
		{ 492, 148, 42 },
		{ 492, 149, 42 },
		{ 492, 150, 42 },
		{ 492, 151, 42 },
		{ 492, 152, 42 },
		{ 492, 153, 42 },
		{ 492, 154, 42 },
		{ 492, 155, 42 },
		{ 492, 156, 42 },
		{ 492, 157, 42 },
		{ 492, 158, 42 },
		{ 492, 159, 42 },
		{ 492, 160, 42 },
		{ 492, 161, 42 },
		{ 492, 162, 42 },
		{ 492, 163, 42 },
		{ 492, 164, 42 },
		{ 492, 165, 42 },
		{ 492, 166, 42 },
		{ 492, 167, 42 },
		{ 492, 168, 42 },
		{ 493, 110, 42 },
		{ 493, 111, 42 },
		{ 493, 112, 42 },
		{ 493, 113, 42 },
		{ 493, 114, 42 },
		{ 493, 115, 42 },
		{ 493, 116, 42 },
		{ 493, 117, 42 },
		{ 493, 118, 42 },
		{ 493, 119, 42 },
		{ 493, 120, 42 },
		{ 493, 121, 42 },
		{ 493, 122, 42 },
		{ 493, 123, 42 },
		{ 493, 124, 42 },
		{ 493, 125, 42 },
		{ 493, 126, 42 },
		{ 493, 127, 42 },
		{ 493, 128, 42 },
		{ 493, 129, 42 },
		{ 493, 130, 42 },
		{ 493, 131, 42 },
		{ 493, 132, 42 },
		{ 493, 133, 42 },
		{ 493, 134, 42 },
		{ 493, 135, 42 },
		{ 493, 136, 42 },
		{ 493, 137, 42 },
		{ 493, 138, 42 },
		{ 493, 139, 42 },
		{ 493, 140, 42 },
		{ 493, 141, 42 },
		{ 493, 142, 42 },
		{ 493, 143, 42 },
		{ 493, 144, 42 },
		{ 493, 145, 42 },
		{ 493, 146, 42 },
		{ 493, 147, 42 },
		{ 493, 148, 42 },
		{ 493, 149, 42 },
		{ 493, 150, 42 },
		{ 493, 151, 42 },
		{ 493, 152, 42 },
		{ 493, 153, 42 },
		{ 493, 154, 42 },
		{ 493, 155, 42 },
		{ 493, 156, 42 },
		{ 493, 157, 42 },
		{ 493, 158, 42 },
		{ 493, 159, 42 },
		{ 493, 160, 42 },
		{ 493, 161, 42 },
		{ 493, 162, 42 },
		{ 493, 163, 42 },
		{ 493, 164, 42 },
		{ 493, 165, 42 },
		{ 493, 166, 42 },
		{ 493, 167, 42 },
		{ 493, 168, 42 },
		{ 494, 109, 42 },
		{ 494, 110, 42 },
		{ 494, 111, 42 },
		{ 494, 112, 42 },
		{ 494, 113, 42 },
		{ 494, 114, 42 },
		{ 494, 115, 42 },
		{ 494, 116, 42 },
		{ 494, 117, 42 },
		{ 494, 118, 42 },
		{ 494, 119, 42 },
		{ 494, 120, 42 },
		{ 494, 121, 42 },
		{ 494, 122, 42 },
		{ 494, 123, 42 },
		{ 494, 124, 42 },
		{ 494, 125, 42 },
		{ 494, 126, 42 },
		{ 494, 127, 42 },
		{ 494, 128, 42 },
		{ 494, 129, 42 },
		{ 494, 130, 42 },
		{ 494, 131, 42 },
		{ 494, 132, 42 },
		{ 494, 133, 42 },
		{ 494, 134, 42 },
		{ 494, 135, 42 },
		{ 494, 136, 42 },
		{ 494, 137, 42 },
		{ 494, 138, 42 },
		{ 494, 139, 42 },
		{ 494, 140, 42 },
		{ 494, 141, 42 },
		{ 494, 142, 42 },
		{ 494, 143, 42 },
		{ 494, 144, 42 },
		{ 494, 145, 42 },
		{ 494, 146, 42 },
		{ 494, 147, 42 },
		{ 494, 148, 42 },
		{ 494, 149, 42 },
		{ 494, 150, 42 },
		{ 494, 151, 42 },
		{ 494, 152, 42 },
		{ 494, 153, 42 },
		{ 494, 154, 42 },
		{ 494, 155, 42 },
		{ 494, 156, 42 },
		{ 494, 157, 42 },
		{ 494, 158, 42 },
		{ 494, 159, 42 },
		{ 494, 160, 42 },
		{ 494, 161, 42 },
		{ 494, 162, 42 },
		{ 494, 163, 42 },
		{ 494, 164, 42 },
		{ 494, 165, 42 },
		{ 494, 166, 42 },
		{ 494, 167, 42 },
		{ 494, 168, 42 },
		{ 495, 109, 42 },
		{ 495, 110, 42 },
		{ 495, 111, 42 },
		{ 495, 112, 42 },
		{ 495, 113, 42 },
		{ 495, 114, 42 },
		{ 495, 115, 42 },
		{ 495, 116, 42 },
		{ 495, 117, 42 },
		{ 495, 118, 42 },
		{ 495, 119, 42 },
		{ 495, 120, 42 },
		{ 495, 121, 42 },
		{ 495, 122, 42 },
		{ 495, 123, 42 },
		{ 495, 124, 42 },
		{ 495, 125, 42 },
		{ 495, 126, 42 },
		{ 495, 127, 42 },
		{ 495, 128, 42 },
		{ 495, 129, 42 },
		{ 495, 130, 42 },
		{ 495, 131, 42 },
		{ 495, 132, 42 },
		{ 495, 133, 42 },
		{ 495, 134, 42 },
		{ 495, 135, 42 },
		{ 495, 136, 42 },
		{ 495, 137, 42 },
		{ 495, 138, 42 },
		{ 495, 139, 42 },
		{ 495, 140, 42 },
		{ 495, 141, 42 },
		{ 495, 142, 42 },
		{ 495, 143, 42 },
		{ 495, 144, 42 },
		{ 495, 145, 42 },
		{ 495, 146, 42 },
		{ 495, 147, 42 },
		{ 495, 148, 42 },
		{ 495, 149, 42 },
		{ 495, 150, 42 },
		{ 495, 151, 42 },
		{ 495, 152, 42 },
		{ 495, 153, 42 },
		{ 495, 154, 42 },
		{ 495, 155, 42 },
		{ 495, 156, 42 },
		{ 495, 157, 42 },
		{ 495, 158, 42 },
		{ 495, 159, 42 },
		{ 495, 160, 42 },
		{ 495, 161, 42 },
		{ 495, 162, 42 },
		{ 495, 163, 42 },
		{ 495, 164, 42 },
		{ 495, 165, 42 },
		{ 495, 166, 42 },
		{ 495, 167, 42 },
		{ 496, 108, 42 },
		{ 496, 109, 42 },
		{ 496, 110, 42 },
		{ 496, 111, 42 },
		{ 496, 112, 42 },
		{ 496, 113, 42 },
		{ 496, 114, 42 },
		{ 496, 115, 42 },
		{ 496, 116, 42 },
		{ 496, 117, 42 },
		{ 496, 118, 42 },
		{ 496, 119, 42 },
		{ 496, 120, 42 },
		{ 496, 121, 42 },
		{ 496, 122, 42 },
		{ 496, 123, 42 },
		{ 496, 124, 42 },
		{ 496, 125, 42 },
		{ 496, 126, 42 },
		{ 496, 127, 42 },
		{ 496, 128, 42 },
		{ 496, 129, 42 },
		{ 496, 130, 42 },
		{ 496, 131, 42 },
		{ 496, 132, 42 },
		{ 496, 133, 42 },
		{ 496, 134, 42 },
		{ 496, 135, 42 },
		{ 496, 136, 42 },
		{ 496, 137, 42 },
		{ 496, 138, 42 },
		{ 496, 139, 42 },
		{ 496, 140, 42 },
		{ 496, 141, 42 },
		{ 496, 142, 42 },
		{ 496, 143, 42 },
		{ 496, 144, 42 },
		{ 496, 145, 42 },
		{ 496, 146, 42 },
		{ 496, 147, 42 },
		{ 496, 148, 42 },
		{ 496, 149, 42 },
		{ 496, 150, 42 },
		{ 496, 151, 42 },
		{ 496, 152, 42 },
		{ 496, 153, 42 },
		{ 496, 154, 42 },
		{ 496, 155, 42 },
		{ 496, 156, 42 },
		{ 496, 157, 42 },
		{ 496, 158, 42 },
		{ 496, 159, 42 },
		{ 496, 160, 42 },
		{ 496, 161, 42 },
		{ 496, 162, 42 },
		{ 496, 163, 42 },
		{ 496, 164, 42 },
		{ 496, 165, 42 },
		{ 496, 166, 42 },
		{ 496, 167, 42 },
		{ 497, 108, 42 },
		{ 497, 109, 42 },
		{ 497, 110, 42 },
		{ 497, 111, 42 },
		{ 497, 112, 42 },
		{ 497, 113, 42 },
		{ 497, 114, 42 },
		{ 497, 115, 42 },
		{ 497, 116, 42 },
		{ 497, 117, 42 },
		{ 497, 118, 42 },
		{ 497, 119, 42 },
		{ 497, 120, 42 },
		{ 497, 121, 42 },
		{ 497, 122, 42 },
		{ 497, 123, 42 },
		{ 497, 124, 42 },
		{ 497, 125, 42 },
		{ 497, 126, 42 },
		{ 497, 127, 42 },
		{ 497, 128, 42 },
		{ 497, 129, 42 },
		{ 497, 130, 42 },
		{ 497, 131, 42 },
		{ 497, 132, 42 },
		{ 497, 133, 42 },
		{ 497, 134, 42 },
		{ 497, 135, 42 },
		{ 497, 136, 42 },
		{ 497, 137, 42 },
		{ 497, 138, 42 },
		{ 497, 139, 42 },
		{ 497, 140, 42 },
		{ 497, 141, 42 },
		{ 497, 142, 42 },
		{ 497, 143, 42 },
		{ 497, 144, 42 },
		{ 497, 145, 42 },
		{ 497, 146, 42 },
		{ 497, 147, 42 },
		{ 497, 148, 42 },
		{ 497, 149, 42 },
		{ 497, 150, 42 },
		{ 497, 151, 42 },
		{ 497, 152, 42 },
		{ 497, 153, 42 },
		{ 497, 154, 42 },
		{ 497, 155, 42 },
		{ 497, 156, 42 },
		{ 497, 157, 42 },
		{ 497, 158, 42 },
		{ 497, 159, 42 },
		{ 497, 160, 42 },
		{ 497, 161, 42 },
		{ 497, 162, 42 },
		{ 497, 163, 42 },
		{ 497, 164, 42 },
		{ 497, 165, 42 },
		{ 497, 166, 42 },
		{ 498, 107, 42 },
		{ 498, 108, 42 },
		{ 498, 109, 42 },
		{ 498, 110, 42 },
		{ 498, 111, 42 },
		{ 498, 112, 42 },
		{ 498, 113, 42 },
		{ 498, 114, 42 },
		{ 498, 115, 42 },
		{ 498, 116, 42 },
		{ 498, 117, 42 },
		{ 498, 118, 42 },
		{ 498, 119, 42 },
		{ 498, 120, 42 },
		{ 498, 121, 42 },
		{ 498, 122, 42 },
		{ 498, 123, 42 },
		{ 498, 124, 42 },
		{ 498, 125, 42 },
		{ 498, 126, 42 },
		{ 498, 127, 42 },
		{ 498, 128, 42 },
		{ 498, 129, 42 },
		{ 498, 130, 42 },
		{ 498, 131, 42 },
		{ 498, 132, 42 },
		{ 498, 133, 42 },
		{ 498, 134, 42 },
		{ 498, 135, 42 },
		{ 498, 136, 42 },
		{ 498, 137, 42 },
		{ 498, 138, 42 },
		{ 498, 139, 42 },
		{ 498, 140, 42 },
		{ 498, 141, 42 },
		{ 498, 142, 42 },
		{ 498, 143, 42 },
		{ 498, 144, 42 },
		{ 498, 145, 42 },
		{ 498, 146, 42 },
		{ 498, 147, 42 },
		{ 498, 148, 42 },
		{ 498, 149, 42 },
		{ 498, 150, 42 },
		{ 498, 151, 42 },
		{ 498, 152, 42 },
		{ 498, 153, 42 },
		{ 498, 154, 42 },
		{ 498, 155, 42 },
		{ 498, 156, 42 },
		{ 498, 157, 42 },
		{ 498, 158, 42 },
		{ 498, 159, 42 },
		{ 498, 160, 42 },
		{ 498, 161, 42 },
		{ 498, 162, 42 },
		{ 498, 163, 42 },
		{ 498, 164, 42 },
		{ 498, 165, 42 },
		{ 498, 166, 42 },
		{ 499, 106, 42 },
		{ 499, 107, 42 },
		{ 499, 108, 42 },
		{ 499, 109, 42 },
		{ 499, 110, 42 },
		{ 499, 111, 42 },
		{ 499, 112, 42 },
		{ 499, 113, 42 },
		{ 499, 114, 42 },
		{ 499, 115, 42 },
		{ 499, 116, 42 },
		{ 499, 117, 42 },
		{ 499, 118, 42 },
		{ 499, 119, 42 },
		{ 499, 120, 42 },
		{ 499, 121, 42 },
		{ 499, 122, 42 },
		{ 499, 123, 42 },
		{ 499, 124, 42 },
		{ 499, 125, 42 },
		{ 499, 126, 42 },
		{ 499, 127, 42 },
		{ 499, 128, 42 },
		{ 499, 129, 42 },
		{ 499, 130, 42 },
		{ 499, 131, 42 },
		{ 499, 132, 42 },
		{ 499, 133, 42 },
		{ 499, 134, 42 },
		{ 499, 135, 42 },
		{ 499, 136, 42 },
		{ 499, 137, 42 },
		{ 499, 138, 42 },
		{ 499, 139, 42 },
		{ 499, 140, 42 },
		{ 499, 141, 42 },
		{ 499, 142, 42 },
		{ 499, 143, 42 },
		{ 499, 144, 42 },
		{ 499, 145, 42 },
		{ 499, 146, 42 },
		{ 499, 147, 42 },
		{ 499, 148, 42 },
		{ 499, 149, 42 },
		{ 499, 150, 42 },
		{ 499, 151, 42 },
		{ 499, 152, 42 },
		{ 499, 153, 42 },
		{ 499, 154, 42 },
		{ 499, 155, 42 },
		{ 499, 156, 42 },
		{ 499, 157, 42 },
		{ 499, 158, 42 },
		{ 499, 159, 42 },
		{ 499, 160, 42 },
		{ 499, 161, 42 },
		{ 499, 162, 42 },
		{ 499, 163, 42 },
		{ 499, 164, 42 },
		{ 499, 165, 42 },
		{ 499, 166, 42 },
		{ 500, 106, 42 },
		{ 500, 107, 42 },
		{ 500, 108, 42 },
		{ 500, 109, 42 },
		{ 500, 110, 42 },
		{ 500, 111, 42 },
		{ 500, 112, 42 },
		{ 500, 113, 42 },
		{ 500, 114, 42 },
		{ 500, 115, 42 },
		{ 500, 116, 42 },
		{ 500, 117, 42 },
		{ 500, 118, 42 },
		{ 500, 119, 42 },
		{ 500, 120, 42 },
		{ 500, 121, 42 },
		{ 500, 122, 42 },
		{ 500, 123, 42 },
		{ 500, 124, 42 },
		{ 500, 125, 42 },
		{ 500, 126, 42 },
		{ 500, 127, 42 },
		{ 500, 128, 42 },
		{ 500, 129, 42 },
		{ 500, 130, 42 },
		{ 500, 131, 42 },
		{ 500, 132, 42 },
		{ 500, 133, 42 },
		{ 500, 134, 42 },
		{ 500, 135, 42 },
		{ 500, 136, 42 },
		{ 500, 137, 42 },
		{ 500, 138, 42 },
		{ 500, 139, 42 },
		{ 500, 140, 42 },
		{ 500, 141, 42 },
		{ 500, 142, 42 },
		{ 500, 143, 42 },
		{ 500, 144, 42 },
		{ 500, 145, 42 },
		{ 500, 146, 42 },
		{ 500, 147, 42 },
		{ 500, 148, 42 },
		{ 500, 149, 42 },
		{ 500, 150, 42 },
		{ 500, 151, 42 },
		{ 500, 152, 42 },
		{ 500, 153, 42 },
		{ 500, 154, 42 },
		{ 500, 155, 42 },
		{ 500, 156, 42 },
		{ 500, 157, 42 },
		{ 500, 158, 42 },
		{ 500, 159, 42 },
		{ 500, 160, 42 },
		{ 500, 161, 42 },
		{ 500, 162, 42 },
		{ 500, 163, 42 },
		{ 500, 164, 42 },
		{ 500, 165, 42 },
		{ 501, 105, 42 },
		{ 501, 106, 42 },
		{ 501, 107, 42 },
		{ 501, 108, 42 },
		{ 501, 109, 42 },
		{ 501, 110, 42 },
		{ 501, 111, 42 },
		{ 501, 112, 42 },
		{ 501, 113, 42 },
		{ 501, 114, 42 },
		{ 501, 115, 42 },
		{ 501, 116, 42 },
		{ 501, 117, 42 },
		{ 501, 118, 42 },
		{ 501, 119, 42 },
		{ 501, 120, 42 },
		{ 501, 121, 42 },
		{ 501, 122, 42 },
		{ 501, 123, 42 },
		{ 501, 124, 42 },
		{ 501, 125, 42 },
		{ 501, 126, 42 },
		{ 501, 127, 42 },
		{ 501, 128, 42 },
		{ 501, 129, 42 },
		{ 501, 130, 42 },
		{ 501, 131, 42 },
		{ 501, 132, 42 },
		{ 501, 133, 42 },
		{ 501, 134, 42 },
		{ 501, 135, 42 },
		{ 501, 136, 42 },
		{ 501, 137, 42 },
		{ 501, 138, 42 },
		{ 501, 139, 42 },
		{ 501, 140, 42 },
		{ 501, 141, 42 },
		{ 501, 142, 42 },
		{ 501, 143, 42 },
		{ 501, 144, 42 },
		{ 501, 145, 42 },
		{ 501, 146, 42 },
		{ 501, 147, 42 },
		{ 501, 148, 42 },
		{ 501, 149, 42 },
		{ 501, 150, 42 },
		{ 501, 151, 42 },
		{ 501, 152, 42 },
		{ 501, 153, 42 },
		{ 501, 154, 42 },
		{ 501, 155, 42 },
		{ 501, 156, 42 },
		{ 501, 157, 42 },
		{ 501, 158, 42 },
		{ 501, 159, 42 },
		{ 501, 160, 42 },
		{ 501, 161, 42 },
		{ 501, 162, 42 },
		{ 501, 163, 42 },
		{ 501, 164, 42 },
		{ 501, 165, 42 },
		{ 502, 104, 42 },
		{ 502, 105, 42 },
		{ 502, 106, 42 },
		{ 502, 107, 42 },
		{ 502, 108, 42 },
		{ 502, 109, 42 },
		{ 502, 110, 42 },
		{ 502, 111, 42 },
		{ 502, 112, 42 },
		{ 502, 113, 42 },
		{ 502, 114, 42 },
		{ 502, 115, 42 },
		{ 502, 116, 42 },
		{ 502, 117, 42 },
		{ 502, 118, 42 },
		{ 502, 119, 42 },
		{ 502, 120, 42 },
		{ 502, 121, 42 },
		{ 502, 122, 42 },
		{ 502, 123, 42 },
		{ 502, 124, 42 },
		{ 502, 125, 42 },
		{ 502, 126, 42 },
		{ 502, 127, 42 },
		{ 502, 128, 42 },
		{ 502, 129, 42 },
		{ 502, 130, 42 },
		{ 502, 131, 42 },
		{ 502, 132, 42 },
		{ 502, 133, 42 },
		{ 502, 134, 42 },
		{ 502, 135, 42 },
		{ 502, 136, 42 },
		{ 502, 137, 42 },
		{ 502, 138, 42 },
		{ 502, 139, 42 },
		{ 502, 140, 42 },
		{ 502, 141, 42 },
		{ 502, 142, 42 },
		{ 502, 143, 42 },
		{ 502, 144, 42 },
		{ 502, 145, 42 },
		{ 502, 146, 42 },
		{ 502, 147, 42 },
		{ 502, 148, 42 },
		{ 502, 149, 42 },
		{ 502, 150, 42 },
		{ 502, 151, 42 },
		{ 502, 152, 42 },
		{ 502, 153, 42 },
		{ 502, 154, 42 },
		{ 502, 155, 42 },
		{ 502, 156, 42 },
		{ 502, 157, 42 },
		{ 502, 158, 42 },
		{ 502, 159, 42 },
		{ 502, 160, 42 },
		{ 502, 161, 42 },
		{ 502, 162, 42 },
		{ 502, 163, 42 },
		{ 502, 164, 42 },
		{ 502, 165, 42 },
		{ 503, 104, 42 },
		{ 503, 105, 42 },
		{ 503, 106, 42 },
		{ 503, 107, 42 },
		{ 503, 108, 42 },
		{ 503, 109, 42 },
		{ 503, 110, 42 },
		{ 503, 111, 42 },
		{ 503, 112, 42 },
		{ 503, 113, 42 },
		{ 503, 114, 42 },
		{ 503, 115, 42 },
		{ 503, 116, 42 },
		{ 503, 117, 42 },
		{ 503, 118, 42 },
		{ 503, 119, 42 },
		{ 503, 120, 42 },
		{ 503, 121, 42 },
		{ 503, 122, 42 },
		{ 503, 123, 42 },
		{ 503, 124, 42 },
		{ 503, 125, 42 },
		{ 503, 126, 42 },
		{ 503, 127, 42 },
		{ 503, 128, 42 },
		{ 503, 129, 42 },
		{ 503, 130, 42 },
		{ 503, 131, 42 },
		{ 503, 132, 42 },
		{ 503, 133, 42 },
		{ 503, 134, 42 },
		{ 503, 135, 42 },
		{ 503, 136, 42 },
		{ 503, 137, 42 },
		{ 503, 138, 42 },
		{ 503, 139, 42 },
		{ 503, 140, 42 },
		{ 503, 141, 42 },
		{ 503, 142, 42 },
		{ 503, 143, 42 },
		{ 503, 144, 42 },
		{ 503, 145, 42 },
		{ 503, 146, 42 },
		{ 503, 147, 42 },
		{ 503, 148, 42 },
		{ 503, 149, 42 },
		{ 503, 150, 42 },
		{ 503, 151, 42 },
		{ 503, 152, 42 },
		{ 503, 153, 42 },
		{ 503, 154, 42 },
		{ 503, 155, 42 },
		{ 503, 156, 42 },
		{ 503, 157, 42 },
		{ 503, 158, 42 },
		{ 503, 159, 42 },
		{ 503, 160, 42 },
		{ 503, 161, 42 },
		{ 503, 162, 42 },
		{ 503, 163, 42 },
		{ 503, 164, 42 },
		{ 504, 103, 42 },
		{ 504, 104, 42 },
		{ 504, 105, 42 },
		{ 504, 106, 42 },
		{ 504, 107, 42 },
		{ 504, 108, 42 },
		{ 504, 109, 42 },
		{ 504, 110, 42 },
		{ 504, 111, 42 },
		{ 504, 112, 42 },
		{ 504, 113, 42 },
		{ 504, 114, 42 },
		{ 504, 115, 42 },
		{ 504, 116, 42 },
		{ 504, 117, 42 },
		{ 504, 118, 42 },
		{ 504, 119, 42 },
		{ 504, 120, 42 },
		{ 504, 121, 42 },
		{ 504, 122, 42 },
		{ 504, 123, 42 },
		{ 504, 124, 42 },
		{ 504, 125, 42 },
		{ 504, 126, 42 },
		{ 504, 127, 42 },
		{ 504, 128, 42 },
		{ 504, 129, 42 },
		{ 504, 130, 42 },
		{ 504, 131, 42 },
		{ 504, 132, 42 },
		{ 504, 133, 42 },
		{ 504, 134, 42 },
		{ 504, 135, 42 },
		{ 504, 136, 42 },
		{ 504, 137, 42 },
		{ 504, 138, 42 },
		{ 504, 139, 42 },
		{ 504, 140, 42 },
		{ 504, 141, 42 },
		{ 504, 142, 42 },
		{ 504, 143, 42 },
		{ 504, 144, 42 },
		{ 504, 145, 42 },
		{ 504, 146, 42 },
		{ 504, 147, 42 },
		{ 504, 148, 42 },
		{ 504, 149, 42 },
		{ 504, 150, 42 },
		{ 504, 151, 42 },
		{ 504, 152, 42 },
		{ 504, 153, 42 },
		{ 504, 154, 42 },
		{ 504, 155, 42 },
		{ 504, 156, 42 },
		{ 504, 157, 42 },
		{ 504, 158, 42 },
		{ 504, 159, 42 },
		{ 504, 160, 42 },
		{ 504, 161, 42 },
		{ 504, 162, 42 },
		{ 504, 163, 42 },
		{ 504, 164, 42 },
		{ 505, 103, 42 },
		{ 505, 104, 42 },
		{ 505, 105, 42 },
		{ 505, 106, 42 },
		{ 505, 107, 42 },
		{ 505, 108, 42 },
		{ 505, 109, 42 },
		{ 505, 110, 42 },
		{ 505, 111, 42 },
		{ 505, 112, 42 },
		{ 505, 113, 42 },
		{ 505, 114, 42 },
		{ 505, 115, 42 },
		{ 505, 116, 42 },
		{ 505, 117, 42 },
		{ 505, 118, 42 },
		{ 505, 119, 42 },
		{ 505, 120, 42 },
		{ 505, 121, 42 },
		{ 505, 122, 42 },
		{ 505, 123, 42 },
		{ 505, 124, 42 },
		{ 505, 125, 42 },
		{ 505, 126, 42 },
		{ 505, 127, 42 },
		{ 505, 128, 42 },
		{ 505, 129, 42 },
		{ 505, 130, 42 },
		{ 505, 131, 42 },
		{ 505, 132, 42 },
		{ 505, 133, 42 },
		{ 505, 134, 42 },
		{ 505, 135, 42 },
		{ 505, 136, 42 },
		{ 505, 137, 42 },
		{ 505, 138, 42 },
		{ 505, 139, 42 },
		{ 505, 140, 42 },
		{ 505, 141, 42 },
		{ 505, 142, 42 },
		{ 505, 143, 42 },
		{ 505, 144, 42 },
		{ 505, 145, 42 },
		{ 505, 146, 42 },
		{ 505, 147, 42 },
		{ 505, 148, 42 },
		{ 505, 149, 42 },
		{ 505, 150, 42 },
		{ 505, 151, 42 },
		{ 505, 152, 42 },
		{ 505, 153, 42 },
		{ 505, 154, 42 },
		{ 505, 155, 42 },
		{ 505, 156, 42 },
		{ 505, 157, 42 },
		{ 505, 158, 42 },
		{ 505, 159, 42 },
		{ 505, 160, 42 },
		{ 505, 161, 42 },
		{ 505, 162, 42 },
		{ 505, 163, 42 },
		{ 506, 102, 42 },
		{ 506, 103, 42 },
		{ 506, 104, 42 },
		{ 506, 105, 42 },
		{ 506, 106, 42 },
		{ 506, 107, 42 },
		{ 506, 108, 42 },
		{ 506, 109, 42 },
		{ 506, 110, 42 },
		{ 506, 111, 42 },
		{ 506, 112, 42 },
		{ 506, 113, 42 },
		{ 506, 114, 42 },
		{ 506, 115, 42 },
		{ 506, 116, 42 },
		{ 506, 117, 42 },
		{ 506, 118, 42 },
		{ 506, 119, 42 },
		{ 506, 120, 42 },
		{ 506, 121, 42 },
		{ 506, 122, 42 },
		{ 506, 123, 42 },
		{ 506, 124, 42 },
		{ 506, 125, 42 },
		{ 506, 126, 42 },
		{ 506, 127, 42 },
		{ 506, 128, 42 },
		{ 506, 129, 42 },
		{ 506, 130, 42 },
		{ 506, 131, 42 },
		{ 506, 132, 42 },
		{ 506, 133, 42 },
		{ 506, 134, 42 },
		{ 506, 135, 42 },
		{ 506, 136, 42 },
		{ 506, 137, 42 },
		{ 506, 138, 42 },
		{ 506, 139, 42 },
		{ 506, 140, 42 },
		{ 506, 141, 42 },
		{ 506, 142, 42 },
		{ 506, 143, 42 },
		{ 506, 144, 42 },
		{ 506, 145, 42 },
		{ 506, 146, 42 },
		{ 506, 147, 42 },
		{ 506, 148, 42 },
		{ 506, 149, 42 },
		{ 506, 150, 42 },
		{ 506, 151, 42 },
		{ 506, 152, 42 },
		{ 506, 153, 42 },
		{ 506, 154, 42 },
		{ 506, 155, 42 },
		{ 506, 156, 42 },
		{ 506, 157, 42 },
		{ 506, 158, 42 },
		{ 506, 159, 42 },
		{ 506, 160, 42 },
		{ 506, 161, 42 },
		{ 506, 162, 42 },
		{ 506, 163, 42 },
		{ 507, 101, 42 },
		{ 507, 102, 42 },
		{ 507, 103, 42 },
		{ 507, 104, 42 },
		{ 507, 105, 42 },
		{ 507, 106, 42 },
		{ 507, 107, 42 },
		{ 507, 108, 42 },
		{ 507, 109, 42 },
		{ 507, 110, 42 },
		{ 507, 111, 42 },
		{ 507, 112, 42 },
		{ 507, 113, 42 },
		{ 507, 114, 42 },
		{ 507, 115, 42 },
		{ 507, 116, 42 },
		{ 507, 117, 42 },
		{ 507, 118, 42 },
		{ 507, 119, 42 },
		{ 507, 120, 42 },
		{ 507, 121, 42 },
		{ 507, 122, 42 },
		{ 507, 123, 42 },
		{ 507, 124, 42 },
		{ 507, 125, 42 },
		{ 507, 126, 42 },
		{ 507, 127, 42 },
		{ 507, 128, 42 },
		{ 507, 129, 42 },
		{ 507, 130, 42 },
		{ 507, 131, 42 },
		{ 507, 132, 42 },
		{ 507, 133, 42 },
		{ 507, 134, 42 },
		{ 507, 135, 42 },
		{ 507, 136, 42 },
		{ 507, 137, 42 },
		{ 507, 138, 42 },
		{ 507, 139, 42 },
		{ 507, 140, 42 },
		{ 507, 141, 42 },
		{ 507, 142, 42 },
		{ 507, 143, 42 },
		{ 507, 144, 42 },
		{ 507, 145, 42 },
		{ 507, 146, 42 },
		{ 507, 147, 42 },
		{ 507, 148, 42 },
		{ 507, 149, 42 },
		{ 507, 150, 42 },
		{ 507, 151, 42 },
		{ 507, 152, 42 },
		{ 507, 153, 42 },
		{ 507, 154, 42 },
		{ 507, 155, 42 },
		{ 507, 156, 42 },
		{ 507, 157, 42 },
		{ 507, 158, 42 },
		{ 507, 159, 42 },
		{ 507, 160, 42 },
		{ 507, 161, 42 },
		{ 507, 162, 42 },
		{ 507, 163, 42 },
		{ 508, 101, 42 },
		{ 508, 102, 42 },
		{ 508, 103, 42 },
		{ 508, 104, 42 },
		{ 508, 105, 42 },
		{ 508, 106, 42 },
		{ 508, 107, 42 },
		{ 508, 108, 42 },
		{ 508, 109, 42 },
		{ 508, 110, 42 },
		{ 508, 111, 42 },
		{ 508, 112, 42 },
		{ 508, 113, 42 },
		{ 508, 114, 42 },
		{ 508, 115, 42 },
		{ 508, 116, 42 },
		{ 508, 117, 42 },
		{ 508, 118, 42 },
		{ 508, 119, 42 },
		{ 508, 120, 42 },
		{ 508, 121, 42 },
		{ 508, 122, 42 },
		{ 508, 123, 42 },
		{ 508, 124, 42 },
		{ 508, 125, 42 },
		{ 508, 126, 42 },
		{ 508, 127, 42 },
		{ 508, 128, 42 },
		{ 508, 129, 42 },
		{ 508, 130, 42 },
		{ 508, 131, 42 },
		{ 508, 132, 42 },
		{ 508, 133, 42 },
		{ 508, 134, 42 },
		{ 508, 135, 42 },
		{ 508, 136, 42 },
		{ 508, 137, 42 },
		{ 508, 138, 42 },
		{ 508, 139, 42 },
		{ 508, 140, 42 },
		{ 508, 141, 42 },
		{ 508, 142, 42 },
		{ 508, 143, 42 },
		{ 508, 144, 42 },
		{ 508, 145, 42 },
		{ 508, 146, 42 },
		{ 508, 147, 42 },
		{ 508, 148, 42 },
		{ 508, 149, 42 },
		{ 508, 150, 42 },
		{ 508, 151, 42 },
		{ 508, 152, 42 },
		{ 508, 153, 42 },
		{ 508, 154, 42 },
		{ 508, 155, 42 },
		{ 508, 156, 42 },
		{ 508, 157, 42 },
		{ 508, 158, 42 },
		{ 508, 159, 42 },
		{ 508, 160, 42 },
		{ 508, 161, 42 },
		{ 508, 162, 42 },
		{ 509, 100, 42 },
		{ 509, 101, 42 },
		{ 509, 102, 42 },
		{ 509, 103, 42 },
		{ 509, 104, 42 },
		{ 509, 105, 42 },
		{ 509, 106, 42 },
		{ 509, 107, 42 },
		{ 509, 108, 42 },
		{ 509, 109, 42 },
		{ 509, 110, 42 },
		{ 509, 111, 42 },
		{ 509, 112, 42 },
		{ 509, 113, 42 },
		{ 509, 114, 42 },
		{ 509, 115, 42 },
		{ 509, 116, 42 },
		{ 509, 117, 42 },
		{ 509, 118, 42 },
		{ 509, 119, 42 },
		{ 509, 120, 42 },
		{ 509, 121, 42 },
		{ 509, 122, 42 },
		{ 509, 123, 42 },
		{ 509, 124, 42 },
		{ 509, 125, 42 },
		{ 509, 126, 42 },
		{ 509, 127, 42 },
		{ 509, 128, 42 },
		{ 509, 129, 42 },
		{ 509, 130, 42 },
		{ 509, 131, 42 },
		{ 509, 132, 42 },
		{ 509, 133, 42 },
		{ 509, 134, 42 },
		{ 509, 135, 42 },
		{ 509, 136, 42 },
		{ 509, 137, 42 },
		{ 509, 138, 42 },
		{ 509, 139, 42 },
		{ 509, 140, 42 },
		{ 509, 141, 42 },
		{ 509, 142, 42 },
		{ 509, 143, 42 },
		{ 509, 144, 42 },
		{ 509, 145, 42 },
		{ 509, 146, 42 },
		{ 509, 147, 42 },
		{ 509, 148, 42 },
		{ 509, 149, 42 },
		{ 509, 150, 42 },
		{ 509, 151, 42 },
		{ 509, 152, 42 },
		{ 509, 153, 42 },
		{ 509, 154, 42 },
		{ 509, 155, 42 },
		{ 509, 156, 42 },
		{ 509, 157, 42 },
		{ 509, 158, 42 },
		{ 509, 159, 42 },
		{ 509, 160, 42 },
		{ 509, 161, 42 },
		{ 509, 162, 42 },
		{ 510, 99, 42 },
		{ 510, 100, 42 },
		{ 510, 101, 42 },
		{ 510, 102, 42 },
		{ 510, 103, 42 },
		{ 510, 104, 42 },
		{ 510, 105, 42 },
		{ 510, 106, 42 },
		{ 510, 107, 42 },
		{ 510, 108, 42 },
		{ 510, 109, 42 },
		{ 510, 110, 42 },
		{ 510, 111, 42 },
		{ 510, 112, 42 },
		{ 510, 113, 42 },
		{ 510, 114, 42 },
		{ 510, 115, 42 },
		{ 510, 116, 42 },
		{ 510, 117, 42 },
		{ 510, 118, 42 },
		{ 510, 119, 42 },
		{ 510, 120, 42 },
		{ 510, 121, 42 },
		{ 510, 122, 42 },
		{ 510, 123, 42 },
		{ 510, 124, 42 },
		{ 510, 125, 42 },
		{ 510, 126, 42 },
		{ 510, 127, 42 },
		{ 510, 128, 42 },
		{ 510, 129, 42 },
		{ 510, 130, 42 },
		{ 510, 131, 42 },
		{ 510, 132, 42 },
		{ 510, 133, 42 },
		{ 510, 134, 42 },
		{ 510, 135, 42 },
		{ 510, 136, 42 },
		{ 510, 137, 42 },
		{ 510, 138, 42 },
		{ 510, 139, 42 },
		{ 510, 140, 42 },
		{ 510, 141, 42 },
		{ 510, 142, 42 },
		{ 510, 143, 42 },
		{ 510, 144, 42 },
		{ 510, 145, 42 },
		{ 510, 146, 42 },
		{ 510, 147, 42 },
		{ 510, 148, 42 },
		{ 510, 149, 42 },
		{ 510, 150, 42 },
		{ 510, 151, 42 },
		{ 510, 152, 42 },
		{ 510, 153, 42 },
		{ 510, 154, 42 },
		{ 510, 155, 42 },
		{ 510, 156, 42 },
		{ 510, 157, 42 },
		{ 510, 158, 42 },
		{ 510, 159, 42 },
		{ 510, 160, 42 },
		{ 510, 161, 42 },
		{ 510, 162, 42 },
		{ 511, 99, 42 },
		{ 511, 100, 42 },
		{ 511, 101, 42 },
		{ 511, 102, 42 },
		{ 511, 103, 42 },
		{ 511, 104, 42 },
		{ 511, 105, 42 },
		{ 511, 106, 42 },
		{ 511, 107, 42 },
		{ 511, 108, 42 },
		{ 511, 109, 42 },
		{ 511, 110, 42 },
		{ 511, 111, 42 },
		{ 511, 112, 42 },
		{ 511, 113, 42 },
		{ 511, 114, 42 },
		{ 511, 115, 42 },
		{ 511, 116, 42 },
		{ 511, 117, 42 },
		{ 511, 118, 42 },
		{ 511, 119, 42 },
		{ 511, 120, 42 },
		{ 511, 121, 42 },
		{ 511, 122, 42 },
		{ 511, 123, 42 },
		{ 511, 124, 42 },
		{ 511, 125, 42 },
		{ 511, 126, 42 },
		{ 511, 127, 42 },
		{ 511, 128, 42 },
		{ 511, 129, 42 },
		{ 511, 130, 42 },
		{ 511, 131, 42 },
		{ 511, 132, 42 },
		{ 511, 133, 42 },
		{ 511, 134, 42 },
		{ 511, 135, 42 },
		{ 511, 136, 42 },
		{ 511, 137, 42 },
		{ 511, 138, 42 },
		{ 511, 139, 42 },
		{ 511, 140, 42 },
		{ 511, 141, 42 },
		{ 511, 142, 42 },
		{ 511, 143, 42 },
		{ 511, 144, 42 },
		{ 511, 145, 42 },
		{ 511, 146, 42 },
		{ 511, 147, 42 },
		{ 511, 148, 42 },
		{ 511, 149, 42 },
		{ 511, 150, 42 },
		{ 511, 151, 42 },
		{ 511, 152, 42 },
		{ 511, 153, 42 },
		{ 511, 154, 42 },
		{ 511, 155, 42 },
		{ 511, 156, 42 },
		{ 511, 157, 42 },
		{ 511, 158, 42 },
		{ 511, 159, 42 },
		{ 511, 160, 42 },
		{ 511, 161, 42 },
		{ 512, 98, 42 },
		{ 512, 99, 42 },
		{ 512, 100, 42 },
		{ 512, 101, 42 },
		{ 512, 102, 42 },
		{ 512, 103, 42 },
		{ 512, 104, 42 },
		{ 512, 105, 42 },
		{ 512, 106, 42 },
		{ 512, 107, 42 },
		{ 512, 108, 42 },
		{ 512, 109, 42 },
		{ 512, 110, 42 },
		{ 512, 111, 42 },
		{ 512, 112, 42 },
		{ 512, 113, 42 },
		{ 512, 114, 42 },
		{ 512, 115, 42 },
		{ 512, 116, 42 },
		{ 512, 117, 42 },
		{ 512, 118, 42 },
		{ 512, 119, 42 },
		{ 512, 120, 42 },
		{ 512, 121, 42 },
		{ 512, 122, 42 },
		{ 512, 123, 42 },
		{ 512, 124, 42 },
		{ 512, 125, 42 },
		{ 512, 126, 42 },
		{ 512, 127, 42 },
		{ 512, 128, 42 },
		{ 512, 129, 42 },
		{ 512, 130, 42 },
		{ 512, 131, 42 },
		{ 512, 132, 42 },
		{ 512, 133, 42 },
		{ 512, 134, 42 },
		{ 512, 135, 42 },
		{ 512, 136, 42 },
		{ 512, 137, 42 },
		{ 512, 138, 42 },
		{ 512, 139, 42 },
		{ 512, 140, 42 },
		{ 512, 141, 42 },
		{ 512, 142, 42 },
		{ 512, 143, 42 },
		{ 512, 144, 42 },
		{ 512, 145, 42 },
		{ 512, 146, 42 },
		{ 512, 147, 42 },
		{ 512, 148, 42 },
		{ 512, 149, 42 },
		{ 512, 150, 42 },
		{ 512, 151, 42 },
		{ 512, 152, 42 },
		{ 512, 153, 42 },
		{ 512, 154, 42 },
		{ 512, 155, 42 },
		{ 512, 156, 42 },
		{ 512, 157, 42 },
		{ 512, 158, 42 },
		{ 512, 159, 42 },
		{ 512, 160, 42 },
		{ 512, 161, 42 },
		{ 1, 322, 43 },
		{ 1, 323, 43 },
		{ 1, 324, 43 },
		{ 1, 325, 43 },
		{ 1, 326, 43 },
		{ 1, 327, 43 },
		{ 1, 328, 43 },
		{ 1, 329, 43 },
		{ 1, 330, 43 },
		{ 1, 331, 43 },
		{ 1, 332, 43 },
		{ 1, 333, 43 },
		{ 1, 334, 43 },
		{ 1, 335, 43 },
		{ 1, 336, 43 },
		{ 1, 337, 43 },
		{ 1, 338, 43 },
		{ 1, 339, 43 },
		{ 1, 340, 43 },
		{ 1, 341, 43 },
		{ 1, 342, 43 },
		{ 1, 343, 43 },
		{ 1, 344, 43 },
		{ 1, 345, 43 },
		{ 1, 346, 43 },
		{ 1, 347, 43 },
		{ 1, 348, 43 },
		{ 1, 349, 43 },
		{ 1, 350, 43 },
		{ 1, 351, 43 },
		{ 1, 352, 43 },
		{ 1, 353, 43 },
		{ 1, 354, 43 },
		{ 1, 355, 43 },
		{ 1, 356, 43 },
		{ 1, 357, 43 },
		{ 1, 358, 43 },
		{ 1, 359, 43 },
		{ 1, 360, 43 },
		{ 1, 361, 43 },
		{ 1, 362, 43 },
		{ 1, 363, 43 },
		{ 1, 364, 43 },
		{ 1, 365, 43 },
		{ 1, 366, 43 },
		{ 1, 367, 43 },
		{ 1, 368, 43 },
		{ 1, 369, 43 },
		{ 1, 370, 43 },
		{ 1, 371, 43 },
		{ 1, 372, 43 },
		{ 1, 373, 43 },
		{ 1, 374, 43 },
		{ 1, 375, 43 },
		{ 1, 376, 43 },
		{ 1, 377, 43 },
		{ 1, 378, 43 },
		{ 1, 379, 43 },
		{ 1, 380, 43 },
		{ 1, 381, 43 },
		{ 1, 382, 43 },
		{ 1, 383, 43 },
		{ 1, 384, 43 },
		{ 2, 321, 43 },
		{ 2, 322, 43 },
		{ 2, 323, 43 },
		{ 2, 324, 43 },
		{ 2, 325, 43 },
		{ 2, 326, 43 },
		{ 2, 327, 43 },
		{ 2, 328, 43 },
		{ 2, 329, 43 },
		{ 2, 330, 43 },
		{ 2, 331, 43 },
		{ 2, 332, 43 },
		{ 2, 333, 43 },
		{ 2, 334, 43 },
		{ 2, 335, 43 },
		{ 2, 336, 43 },
		{ 2, 337, 43 },
		{ 2, 338, 43 },
		{ 2, 339, 43 },
		{ 2, 340, 43 },
		{ 2, 341, 43 },
		{ 2, 342, 43 },
		{ 2, 343, 43 },
		{ 2, 344, 43 },
		{ 2, 345, 43 },
		{ 2, 346, 43 },
		{ 2, 347, 43 },
		{ 2, 348, 43 },
		{ 2, 349, 43 },
		{ 2, 350, 43 },
		{ 2, 351, 43 },
		{ 2, 352, 43 },
		{ 2, 353, 43 },
		{ 2, 354, 43 },
		{ 2, 355, 43 },
		{ 2, 356, 43 },
		{ 2, 357, 43 },
		{ 2, 358, 43 },
		{ 2, 359, 43 },
		{ 2, 360, 43 },
		{ 2, 361, 43 },
		{ 2, 362, 43 },
		{ 2, 363, 43 },
		{ 2, 364, 43 },
		{ 2, 365, 43 },
		{ 2, 366, 43 },
		{ 2, 367, 43 },
		{ 2, 368, 43 },
		{ 2, 369, 43 },
		{ 2, 370, 43 },
		{ 2, 371, 43 },
		{ 2, 372, 43 },
		{ 2, 373, 43 },
		{ 2, 374, 43 },
		{ 2, 375, 43 },
		{ 2, 376, 43 },
		{ 2, 377, 43 },
		{ 2, 378, 43 },
		{ 2, 379, 43 },
		{ 2, 380, 43 },
		{ 2, 381, 43 },
		{ 2, 382, 43 },
		{ 2, 383, 43 },
		{ 2, 384, 43 },
		{ 3, 321, 43 },
		{ 3, 322, 43 },
		{ 3, 323, 43 },
		{ 3, 324, 43 },
		{ 3, 325, 43 },
		{ 3, 326, 43 },
		{ 3, 327, 43 },
		{ 3, 328, 43 },
		{ 3, 329, 43 },
		{ 3, 330, 43 },
		{ 3, 331, 43 },
		{ 3, 332, 43 },
		{ 3, 333, 43 },
		{ 3, 334, 43 },
		{ 3, 335, 43 },
		{ 3, 336, 43 },
		{ 3, 337, 43 },
		{ 3, 338, 43 },
		{ 3, 339, 43 },
		{ 3, 340, 43 },
		{ 3, 341, 43 },
		{ 3, 342, 43 },
		{ 3, 343, 43 },
		{ 3, 344, 43 },
		{ 3, 345, 43 },
		{ 3, 346, 43 },
		{ 3, 347, 43 },
		{ 3, 348, 43 },
		{ 3, 349, 43 },
		{ 3, 350, 43 },
		{ 3, 351, 43 },
		{ 3, 352, 43 },
		{ 3, 353, 43 },
		{ 3, 354, 43 },
		{ 3, 355, 43 },
		{ 3, 356, 43 },
		{ 3, 357, 43 },
		{ 3, 358, 43 },
		{ 3, 359, 43 },
		{ 3, 360, 43 },
		{ 3, 361, 43 },
		{ 3, 362, 43 },
		{ 3, 363, 43 },
		{ 3, 364, 43 },
		{ 3, 365, 43 },
		{ 3, 366, 43 },
		{ 3, 367, 43 },
		{ 3, 368, 43 },
		{ 3, 369, 43 },
		{ 3, 370, 43 },
		{ 3, 371, 43 },
		{ 3, 372, 43 },
		{ 3, 373, 43 },
		{ 3, 374, 43 },
		{ 3, 375, 43 },
		{ 3, 376, 43 },
		{ 3, 377, 43 },
		{ 3, 378, 43 },
		{ 3, 379, 43 },
		{ 3, 380, 43 },
		{ 3, 381, 43 },
		{ 3, 382, 43 },
		{ 3, 383, 43 },
		{ 4, 321, 43 },
		{ 4, 322, 43 },
		{ 4, 323, 43 },
		{ 4, 324, 43 },
		{ 4, 325, 43 },
		{ 4, 326, 43 },
		{ 4, 327, 43 },
		{ 4, 328, 43 },
		{ 4, 329, 43 },
		{ 4, 330, 43 },
		{ 4, 331, 43 },
		{ 4, 332, 43 },
		{ 4, 333, 43 },
		{ 4, 334, 43 },
		{ 4, 335, 43 },
		{ 4, 336, 43 },
		{ 4, 337, 43 },
		{ 4, 338, 43 },
		{ 4, 339, 43 },
		{ 4, 340, 43 },
		{ 4, 341, 43 },
		{ 4, 342, 43 },
		{ 4, 343, 43 },
		{ 4, 344, 43 },
		{ 4, 345, 43 },
		{ 4, 346, 43 },
		{ 4, 347, 43 },
		{ 4, 348, 43 },
		{ 4, 349, 43 },
		{ 4, 350, 43 },
		{ 4, 351, 43 },
		{ 4, 352, 43 },
		{ 4, 353, 43 },
		{ 4, 354, 43 },
		{ 4, 355, 43 },
		{ 4, 356, 43 },
		{ 4, 357, 43 },
		{ 4, 358, 43 },
		{ 4, 359, 43 },
		{ 4, 360, 43 },
		{ 4, 361, 43 },
		{ 4, 362, 43 },
		{ 4, 363, 43 },
		{ 4, 364, 43 },
		{ 4, 365, 43 },
		{ 4, 366, 43 },
		{ 4, 367, 43 },
		{ 4, 368, 43 },
		{ 4, 369, 43 },
		{ 4, 370, 43 },
		{ 4, 371, 43 },
		{ 4, 372, 43 },
		{ 4, 373, 43 },
		{ 4, 374, 43 },
		{ 4, 375, 43 },
		{ 4, 376, 43 },
		{ 4, 377, 43 },
		{ 4, 378, 43 },
		{ 4, 379, 43 },
		{ 4, 380, 43 },
		{ 4, 381, 43 },
		{ 4, 382, 43 },
		{ 4, 383, 43 },
		{ 5, 321, 43 },
		{ 5, 322, 43 },
		{ 5, 323, 43 },
		{ 5, 324, 43 },
		{ 5, 325, 43 },
		{ 5, 326, 43 },
		{ 5, 327, 43 },
		{ 5, 328, 43 },
		{ 5, 329, 43 },
		{ 5, 330, 43 },
		{ 5, 331, 43 },
		{ 5, 332, 43 },
		{ 5, 333, 43 },
		{ 5, 334, 43 },
		{ 5, 335, 43 },
		{ 5, 336, 43 },
		{ 5, 337, 43 },
		{ 5, 338, 43 },
		{ 5, 339, 43 },
		{ 5, 340, 43 },
		{ 5, 341, 43 },
		{ 5, 342, 43 },
		{ 5, 343, 43 },
		{ 5, 344, 43 },
		{ 5, 345, 43 },
		{ 5, 346, 43 },
		{ 5, 347, 43 },
		{ 5, 348, 43 },
		{ 5, 349, 43 },
		{ 5, 350, 43 },
		{ 5, 351, 43 },
		{ 5, 352, 43 },
		{ 5, 353, 43 },
		{ 5, 354, 43 },
		{ 5, 355, 43 },
		{ 5, 356, 43 },
		{ 5, 357, 43 },
		{ 5, 358, 43 },
		{ 5, 359, 43 },
		{ 5, 360, 43 },
		{ 5, 361, 43 },
		{ 5, 362, 43 },
		{ 5, 363, 43 },
		{ 5, 364, 43 },
		{ 5, 365, 43 },
		{ 5, 366, 43 },
		{ 5, 367, 43 },
		{ 5, 368, 43 },
		{ 5, 369, 43 },
		{ 5, 370, 43 },
		{ 5, 371, 43 },
		{ 5, 372, 43 },
		{ 5, 373, 43 },
		{ 5, 374, 43 },
		{ 5, 375, 43 },
		{ 5, 376, 43 },
		{ 5, 377, 43 },
		{ 5, 378, 43 },
		{ 5, 379, 43 },
		{ 5, 380, 43 },
		{ 5, 381, 43 },
		{ 5, 382, 43 },
		{ 6, 320, 43 },
		{ 6, 321, 43 },
		{ 6, 322, 43 },
		{ 6, 323, 43 },
		{ 6, 324, 43 },
		{ 6, 325, 43 },
		{ 6, 326, 43 },
		{ 6, 327, 43 },
		{ 6, 328, 43 },
		{ 6, 329, 43 },
		{ 6, 330, 43 },
		{ 6, 331, 43 },
		{ 6, 332, 43 },
		{ 6, 333, 43 },
		{ 6, 334, 43 },
		{ 6, 335, 43 },
		{ 6, 336, 43 },
		{ 6, 337, 43 },
		{ 6, 338, 43 },
		{ 6, 339, 43 },
		{ 6, 340, 43 },
		{ 6, 341, 43 },
		{ 6, 342, 43 },
		{ 6, 343, 43 },
		{ 6, 344, 43 },
		{ 6, 345, 43 },
		{ 6, 346, 43 },
		{ 6, 347, 43 },
		{ 6, 348, 43 },
		{ 6, 349, 43 },
		{ 6, 350, 43 },
		{ 6, 351, 43 },
		{ 6, 352, 43 },
		{ 6, 353, 43 },
		{ 6, 354, 43 },
		{ 6, 355, 43 },
		{ 6, 356, 43 },
		{ 6, 357, 43 },
		{ 6, 358, 43 },
		{ 6, 359, 43 },
		{ 6, 360, 43 },
		{ 6, 361, 43 },
		{ 6, 362, 43 },
		{ 6, 363, 43 },
		{ 6, 364, 43 },
		{ 6, 365, 43 },
		{ 6, 366, 43 },
		{ 6, 367, 43 },
		{ 6, 368, 43 },
		{ 6, 369, 43 },
		{ 6, 370, 43 },
		{ 6, 371, 43 },
		{ 6, 372, 43 },
		{ 6, 373, 43 },
		{ 6, 374, 43 },
		{ 6, 375, 43 },
		{ 6, 376, 43 },
		{ 6, 377, 43 },
		{ 6, 378, 43 },
		{ 6, 379, 43 },
		{ 6, 380, 43 },
		{ 6, 381, 43 },
		{ 6, 382, 43 },
		{ 7, 320, 43 },
		{ 7, 321, 43 },
		{ 7, 322, 43 },
		{ 7, 323, 43 },
		{ 7, 324, 43 },
		{ 7, 325, 43 },
		{ 7, 326, 43 },
		{ 7, 327, 43 },
		{ 7, 328, 43 },
		{ 7, 329, 43 },
		{ 7, 330, 43 },
		{ 7, 331, 43 },
		{ 7, 332, 43 },
		{ 7, 333, 43 },
		{ 7, 334, 43 },
		{ 7, 335, 43 },
		{ 7, 336, 43 },
		{ 7, 337, 43 },
		{ 7, 338, 43 },
		{ 7, 339, 43 },
		{ 7, 340, 43 },
		{ 7, 341, 43 },
		{ 7, 342, 43 },
		{ 7, 343, 43 },
		{ 7, 344, 43 },
		{ 7, 345, 43 },
		{ 7, 346, 43 },
		{ 7, 347, 43 },
		{ 7, 348, 43 },
		{ 7, 349, 43 },
		{ 7, 350, 43 },
		{ 7, 351, 43 },
		{ 7, 352, 43 },
		{ 7, 353, 43 },
		{ 7, 354, 43 },
		{ 7, 355, 43 },
		{ 7, 356, 43 },
		{ 7, 357, 43 },
		{ 7, 358, 43 },
		{ 7, 359, 43 },
		{ 7, 360, 43 },
		{ 7, 361, 43 },
		{ 7, 362, 43 },
		{ 7, 363, 43 },
		{ 7, 364, 43 },
		{ 7, 365, 43 },
		{ 7, 366, 43 },
		{ 7, 367, 43 },
		{ 7, 368, 43 },
		{ 7, 369, 43 },
		{ 7, 370, 43 },
		{ 7, 371, 43 },
		{ 7, 372, 43 },
		{ 7, 373, 43 },
		{ 7, 374, 43 },
		{ 7, 375, 43 },
		{ 7, 376, 43 },
		{ 7, 377, 43 },
		{ 7, 378, 43 },
		{ 7, 379, 43 },
		{ 7, 380, 43 },
		{ 7, 381, 43 },
		{ 8, 320, 43 },
		{ 8, 321, 43 },
		{ 8, 322, 43 },
		{ 8, 323, 43 },
		{ 8, 324, 43 },
		{ 8, 325, 43 },
		{ 8, 326, 43 },
		{ 8, 327, 43 },
		{ 8, 328, 43 },
		{ 8, 329, 43 },
		{ 8, 330, 43 },
		{ 8, 331, 43 },
		{ 8, 332, 43 },
		{ 8, 333, 43 },
		{ 8, 334, 43 },
		{ 8, 335, 43 },
		{ 8, 336, 43 },
		{ 8, 337, 43 },
		{ 8, 338, 43 },
		{ 8, 339, 43 },
		{ 8, 340, 43 },
		{ 8, 341, 43 },
		{ 8, 342, 43 },
		{ 8, 343, 43 },
		{ 8, 344, 43 },
		{ 8, 345, 43 },
		{ 8, 346, 43 },
		{ 8, 347, 43 },
		{ 8, 348, 43 },
		{ 8, 349, 43 },
		{ 8, 350, 43 },
		{ 8, 351, 43 },
		{ 8, 352, 43 },
		{ 8, 353, 43 },
		{ 8, 354, 43 },
		{ 8, 355, 43 },
		{ 8, 356, 43 },
		{ 8, 357, 43 },
		{ 8, 358, 43 },
		{ 8, 359, 43 },
		{ 8, 360, 43 },
		{ 8, 361, 43 },
		{ 8, 362, 43 },
		{ 8, 363, 43 },
		{ 8, 364, 43 },
		{ 8, 365, 43 },
		{ 8, 366, 43 },
		{ 8, 367, 43 },
		{ 8, 368, 43 },
		{ 8, 369, 43 },
		{ 8, 370, 43 },
		{ 8, 371, 43 },
		{ 8, 372, 43 },
		{ 8, 373, 43 },
		{ 8, 374, 43 },
		{ 8, 375, 43 },
		{ 8, 376, 43 },
		{ 8, 377, 43 },
		{ 8, 378, 43 },
		{ 8, 379, 43 },
		{ 8, 380, 43 },
		{ 8, 381, 43 },
		{ 9, 320, 43 },
		{ 9, 321, 43 },
		{ 9, 322, 43 },
		{ 9, 323, 43 },
		{ 9, 324, 43 },
		{ 9, 325, 43 },
		{ 9, 326, 43 },
		{ 9, 327, 43 },
		{ 9, 328, 43 },
		{ 9, 329, 43 },
		{ 9, 330, 43 },
		{ 9, 331, 43 },
		{ 9, 332, 43 },
		{ 9, 333, 43 },
		{ 9, 334, 43 },
		{ 9, 335, 43 },
		{ 9, 336, 43 },
		{ 9, 337, 43 },
		{ 9, 338, 43 },
		{ 9, 339, 43 },
		{ 9, 340, 43 },
		{ 9, 341, 43 },
		{ 9, 342, 43 },
		{ 9, 343, 43 },
		{ 9, 344, 43 },
		{ 9, 345, 43 },
		{ 9, 346, 43 },
		{ 9, 347, 43 },
		{ 9, 348, 43 },
		{ 9, 349, 43 },
		{ 9, 350, 43 },
		{ 9, 351, 43 },
		{ 9, 352, 43 },
		{ 9, 353, 43 },
		{ 9, 354, 43 },
		{ 9, 355, 43 },
		{ 9, 356, 43 },
		{ 9, 357, 43 },
		{ 9, 358, 43 },
		{ 9, 359, 43 },
		{ 9, 360, 43 },
		{ 9, 361, 43 },
		{ 9, 362, 43 },
		{ 9, 363, 43 },
		{ 9, 364, 43 },
		{ 9, 365, 43 },
		{ 9, 366, 43 },
		{ 9, 367, 43 },
		{ 9, 368, 43 },
		{ 9, 369, 43 },
		{ 9, 370, 43 },
		{ 9, 371, 43 },
		{ 9, 372, 43 },
		{ 9, 373, 43 },
		{ 9, 374, 43 },
		{ 9, 375, 43 },
		{ 9, 376, 43 },
		{ 9, 377, 43 },
		{ 9, 378, 43 },
		{ 9, 379, 43 },
		{ 9, 380, 43 },
		{ 10, 319, 43 },
		{ 10, 320, 43 },
		{ 10, 321, 43 },
		{ 10, 322, 43 },
		{ 10, 323, 43 },
		{ 10, 324, 43 },
		{ 10, 325, 43 },
		{ 10, 326, 43 },
		{ 10, 327, 43 },
		{ 10, 328, 43 },
		{ 10, 329, 43 },
		{ 10, 330, 43 },
		{ 10, 331, 43 },
		{ 10, 332, 43 },
		{ 10, 333, 43 },
		{ 10, 334, 43 },
		{ 10, 335, 43 },
		{ 10, 336, 43 },
		{ 10, 337, 43 },
		{ 10, 338, 43 },
		{ 10, 339, 43 },
		{ 10, 340, 43 },
		{ 10, 341, 43 },
		{ 10, 342, 43 },
		{ 10, 343, 43 },
		{ 10, 344, 43 },
		{ 10, 345, 43 },
		{ 10, 346, 43 },
		{ 10, 347, 43 },
		{ 10, 348, 43 },
		{ 10, 349, 43 },
		{ 10, 350, 43 },
		{ 10, 351, 43 },
		{ 10, 352, 43 },
		{ 10, 353, 43 },
		{ 10, 354, 43 },
		{ 10, 355, 43 },
		{ 10, 356, 43 },
		{ 10, 357, 43 },
		{ 10, 358, 43 },
		{ 10, 359, 43 },
		{ 10, 360, 43 },
		{ 10, 361, 43 },
		{ 10, 362, 43 },
		{ 10, 363, 43 },
		{ 10, 364, 43 },
		{ 10, 365, 43 },
		{ 10, 366, 43 },
		{ 10, 367, 43 },
		{ 10, 368, 43 },
		{ 10, 369, 43 },
		{ 10, 370, 43 },
		{ 10, 371, 43 },
		{ 10, 372, 43 },
		{ 10, 373, 43 },
		{ 10, 374, 43 },
		{ 10, 375, 43 },
		{ 10, 376, 43 },
		{ 10, 377, 43 },
		{ 10, 378, 43 },
		{ 10, 379, 43 },
		{ 10, 380, 43 },
		{ 11, 319, 43 },
		{ 11, 320, 43 },
		{ 11, 321, 43 },
		{ 11, 322, 43 },
		{ 11, 323, 43 },
		{ 11, 324, 43 },
		{ 11, 325, 43 },
		{ 11, 326, 43 },
		{ 11, 327, 43 },
		{ 11, 328, 43 },
		{ 11, 329, 43 },
		{ 11, 330, 43 },
		{ 11, 331, 43 },
		{ 11, 332, 43 },
		{ 11, 333, 43 },
		{ 11, 334, 43 },
		{ 11, 335, 43 },
		{ 11, 336, 43 },
		{ 11, 337, 43 },
		{ 11, 338, 43 },
		{ 11, 339, 43 },
		{ 11, 340, 43 },
		{ 11, 341, 43 },
		{ 11, 342, 43 },
		{ 11, 343, 43 },
		{ 11, 344, 43 },
		{ 11, 345, 43 },
		{ 11, 346, 43 },
		{ 11, 347, 43 },
		{ 11, 348, 43 },
		{ 11, 349, 43 },
		{ 11, 350, 43 },
		{ 11, 351, 43 },
		{ 11, 352, 43 },
		{ 11, 353, 43 },
		{ 11, 354, 43 },
		{ 11, 355, 43 },
		{ 11, 356, 43 },
		{ 11, 357, 43 },
		{ 11, 358, 43 },
		{ 11, 359, 43 },
		{ 11, 360, 43 },
		{ 11, 361, 43 },
		{ 11, 362, 43 },
		{ 11, 363, 43 },
		{ 11, 364, 43 },
		{ 11, 365, 43 },
		{ 11, 366, 43 },
		{ 11, 367, 43 },
		{ 11, 368, 43 },
		{ 11, 369, 43 },
		{ 11, 370, 43 },
		{ 11, 371, 43 },
		{ 11, 372, 43 },
		{ 11, 373, 43 },
		{ 11, 374, 43 },
		{ 11, 375, 43 },
		{ 11, 376, 43 },
		{ 11, 377, 43 },
		{ 11, 378, 43 },
		{ 11, 379, 43 },
		{ 12, 319, 43 },
		{ 12, 320, 43 },
		{ 12, 321, 43 },
		{ 12, 322, 43 },
		{ 12, 323, 43 },
		{ 12, 324, 43 },
		{ 12, 325, 43 },
		{ 12, 326, 43 },
		{ 12, 327, 43 },
		{ 12, 328, 43 },
		{ 12, 329, 43 },
		{ 12, 330, 43 },
		{ 12, 331, 43 },
		{ 12, 332, 43 },
		{ 12, 333, 43 },
		{ 12, 334, 43 },
		{ 12, 335, 43 },
		{ 12, 336, 43 },
		{ 12, 337, 43 },
		{ 12, 338, 43 },
		{ 12, 339, 43 },
		{ 12, 340, 43 },
		{ 12, 341, 43 },
		{ 12, 342, 43 },
		{ 12, 343, 43 },
		{ 12, 344, 43 },
		{ 12, 345, 43 },
		{ 12, 346, 43 },
		{ 12, 347, 43 },
		{ 12, 348, 43 },
		{ 12, 349, 43 },
		{ 12, 350, 43 },
		{ 12, 351, 43 },
		{ 12, 352, 43 },
		{ 12, 353, 43 },
		{ 12, 354, 43 },
		{ 12, 355, 43 },
		{ 12, 356, 43 },
		{ 12, 357, 43 },
		{ 12, 358, 43 },
		{ 12, 359, 43 },
		{ 12, 360, 43 },
		{ 12, 361, 43 },
		{ 12, 362, 43 },
		{ 12, 363, 43 },
		{ 12, 364, 43 },
		{ 12, 365, 43 },
		{ 12, 366, 43 },
		{ 12, 367, 43 },
		{ 12, 368, 43 },
		{ 12, 369, 43 },
		{ 12, 370, 43 },
		{ 12, 371, 43 },
		{ 12, 372, 43 },
		{ 12, 373, 43 },
		{ 12, 374, 43 },
		{ 12, 375, 43 },
		{ 12, 376, 43 },
		{ 12, 377, 43 },
		{ 12, 378, 43 },
		{ 12, 379, 43 },
		{ 13, 319, 43 },
		{ 13, 320, 43 },
		{ 13, 321, 43 },
		{ 13, 322, 43 },
		{ 13, 323, 43 },
		{ 13, 324, 43 },
		{ 13, 325, 43 },
		{ 13, 326, 43 },
		{ 13, 327, 43 },
		{ 13, 328, 43 },
		{ 13, 329, 43 },
		{ 13, 330, 43 },
		{ 13, 331, 43 },
		{ 13, 332, 43 },
		{ 13, 333, 43 },
		{ 13, 334, 43 },
		{ 13, 335, 43 },
		{ 13, 336, 43 },
		{ 13, 337, 43 },
		{ 13, 338, 43 },
		{ 13, 339, 43 },
		{ 13, 340, 43 },
		{ 13, 341, 43 },
		{ 13, 342, 43 },
		{ 13, 343, 43 },
		{ 13, 344, 43 },
		{ 13, 345, 43 },
		{ 13, 346, 43 },
		{ 13, 347, 43 },
		{ 13, 348, 43 },
		{ 13, 349, 43 },
		{ 13, 350, 43 },
		{ 13, 351, 43 },
		{ 13, 352, 43 },
		{ 13, 353, 43 },
		{ 13, 354, 43 },
		{ 13, 355, 43 },
		{ 13, 356, 43 },
		{ 13, 357, 43 },
		{ 13, 358, 43 },
		{ 13, 359, 43 },
		{ 13, 360, 43 },
		{ 13, 361, 43 },
		{ 13, 362, 43 },
		{ 13, 363, 43 },
		{ 13, 364, 43 },
		{ 13, 365, 43 },
		{ 13, 366, 43 },
		{ 13, 367, 43 },
		{ 13, 368, 43 },
		{ 13, 369, 43 },
		{ 13, 370, 43 },
		{ 13, 371, 43 },
		{ 13, 372, 43 },
		{ 13, 373, 43 },
		{ 13, 374, 43 },
		{ 13, 375, 43 },
		{ 13, 376, 43 },
		{ 13, 377, 43 },
		{ 13, 378, 43 },
		{ 14, 318, 43 },
		{ 14, 319, 43 },
		{ 14, 320, 43 },
		{ 14, 321, 43 },
		{ 14, 322, 43 },
		{ 14, 323, 43 },
		{ 14, 324, 43 },
		{ 14, 325, 43 },
		{ 14, 326, 43 },
		{ 14, 327, 43 },
		{ 14, 328, 43 },
		{ 14, 329, 43 },
		{ 14, 330, 43 },
		{ 14, 331, 43 },
		{ 14, 332, 43 },
		{ 14, 333, 43 },
		{ 14, 334, 43 },
		{ 14, 335, 43 },
		{ 14, 336, 43 },
		{ 14, 337, 43 },
		{ 14, 338, 43 },
		{ 14, 339, 43 },
		{ 14, 340, 43 },
		{ 14, 341, 43 },
		{ 14, 342, 43 },
		{ 14, 343, 43 },
		{ 14, 344, 43 },
		{ 14, 345, 43 },
		{ 14, 346, 43 },
		{ 14, 347, 43 },
		{ 14, 348, 43 },
		{ 14, 349, 43 },
		{ 14, 350, 43 },
		{ 14, 351, 43 },
		{ 14, 352, 43 },
		{ 14, 353, 43 },
		{ 14, 354, 43 },
		{ 14, 355, 43 },
		{ 14, 356, 43 },
		{ 14, 357, 43 },
		{ 14, 358, 43 },
		{ 14, 359, 43 },
		{ 14, 360, 43 },
		{ 14, 361, 43 },
		{ 14, 362, 43 },
		{ 14, 363, 43 },
		{ 14, 364, 43 },
		{ 14, 365, 43 },
		{ 14, 366, 43 },
		{ 14, 367, 43 },
		{ 14, 368, 43 },
		{ 14, 369, 43 },
		{ 14, 370, 43 },
		{ 14, 371, 43 },
		{ 14, 372, 43 },
		{ 14, 373, 43 },
		{ 14, 374, 43 },
		{ 14, 375, 43 },
		{ 14, 376, 43 },
		{ 14, 377, 43 },
		{ 14, 378, 43 },
		{ 15, 318, 43 },
		{ 15, 319, 43 },
		{ 15, 320, 43 },
		{ 15, 321, 43 },
		{ 15, 322, 43 },
		{ 15, 323, 43 },
		{ 15, 324, 43 },
		{ 15, 325, 43 },
		{ 15, 326, 43 },
		{ 15, 327, 43 },
		{ 15, 328, 43 },
		{ 15, 329, 43 },
		{ 15, 330, 43 },
		{ 15, 331, 43 },
		{ 15, 332, 43 },
		{ 15, 333, 43 },
		{ 15, 334, 43 },
		{ 15, 335, 43 },
		{ 15, 336, 43 },
		{ 15, 337, 43 },
		{ 15, 338, 43 },
		{ 15, 339, 43 },
		{ 15, 340, 43 },
		{ 15, 341, 43 },
		{ 15, 342, 43 },
		{ 15, 343, 43 },
		{ 15, 344, 43 },
		{ 15, 345, 43 },
		{ 15, 346, 43 },
		{ 15, 347, 43 },
		{ 15, 348, 43 },
		{ 15, 349, 43 },
		{ 15, 350, 43 },
		{ 15, 351, 43 },
		{ 15, 352, 43 },
		{ 15, 353, 43 },
		{ 15, 354, 43 },
		{ 15, 355, 43 },
		{ 15, 356, 43 },
		{ 15, 357, 43 },
		{ 15, 358, 43 },
		{ 15, 359, 43 },
		{ 15, 360, 43 },
		{ 15, 361, 43 },
		{ 15, 362, 43 },
		{ 15, 363, 43 },
		{ 15, 364, 43 },
		{ 15, 365, 43 },
		{ 15, 366, 43 },
		{ 15, 367, 43 },
		{ 15, 368, 43 },
		{ 15, 369, 43 },
		{ 15, 370, 43 },
		{ 15, 371, 43 },
		{ 15, 372, 43 },
		{ 15, 373, 43 },
		{ 15, 374, 43 },
		{ 15, 375, 43 },
		{ 15, 376, 43 },
		{ 15, 377, 43 },
		{ 16, 318, 43 },
		{ 16, 319, 43 },
		{ 16, 320, 43 },
		{ 16, 321, 43 },
		{ 16, 322, 43 },
		{ 16, 323, 43 },
		{ 16, 324, 43 },
		{ 16, 325, 43 },
		{ 16, 326, 43 },
		{ 16, 327, 43 },
		{ 16, 328, 43 },
		{ 16, 329, 43 },
		{ 16, 330, 43 },
		{ 16, 331, 43 },
		{ 16, 332, 43 },
		{ 16, 333, 43 },
		{ 16, 334, 43 },
		{ 16, 335, 43 },
		{ 16, 336, 43 },
		{ 16, 337, 43 },
		{ 16, 338, 43 },
		{ 16, 339, 43 },
		{ 16, 340, 43 },
		{ 16, 341, 43 },
		{ 16, 342, 43 },
		{ 16, 343, 43 },
		{ 16, 344, 43 },
		{ 16, 345, 43 },
		{ 16, 346, 43 },
		{ 16, 347, 43 },
		{ 16, 348, 43 },
		{ 16, 349, 43 },
		{ 16, 350, 43 },
		{ 16, 351, 43 },
		{ 16, 352, 43 },
		{ 16, 353, 43 },
		{ 16, 354, 43 },
		{ 16, 355, 43 },
		{ 16, 356, 43 },
		{ 16, 357, 43 },
		{ 16, 358, 43 },
		{ 16, 359, 43 },
		{ 16, 360, 43 },
		{ 16, 361, 43 },
		{ 16, 362, 43 },
		{ 16, 363, 43 },
		{ 16, 364, 43 },
		{ 16, 365, 43 },
		{ 16, 366, 43 },
		{ 16, 367, 43 },
		{ 16, 368, 43 },
		{ 16, 369, 43 },
		{ 16, 370, 43 },
		{ 16, 371, 43 },
		{ 16, 372, 43 },
		{ 16, 373, 43 },
		{ 16, 374, 43 },
		{ 16, 375, 43 },
		{ 16, 376, 43 },
		{ 16, 377, 43 },
		{ 17, 318, 43 },
		{ 17, 319, 43 },
		{ 17, 320, 43 },
		{ 17, 321, 43 },
		{ 17, 322, 43 },
		{ 17, 323, 43 },
		{ 17, 324, 43 },
		{ 17, 325, 43 },
		{ 17, 326, 43 },
		{ 17, 327, 43 },
		{ 17, 328, 43 },
		{ 17, 329, 43 },
		{ 17, 330, 43 },
		{ 17, 331, 43 },
		{ 17, 332, 43 },
		{ 17, 333, 43 },
		{ 17, 334, 43 },
		{ 17, 335, 43 },
		{ 17, 336, 43 },
		{ 17, 337, 43 },
		{ 17, 338, 43 },
		{ 17, 339, 43 },
		{ 17, 340, 43 },
		{ 17, 341, 43 },
		{ 17, 342, 43 },
		{ 17, 343, 43 },
		{ 17, 344, 43 },
		{ 17, 345, 43 },
		{ 17, 346, 43 },
		{ 17, 347, 43 },
		{ 17, 348, 43 },
		{ 17, 349, 43 },
		{ 17, 350, 43 },
		{ 17, 351, 43 },
		{ 17, 352, 43 },
		{ 17, 353, 43 },
		{ 17, 354, 43 },
		{ 17, 355, 43 },
		{ 17, 356, 43 },
		{ 17, 357, 43 },
		{ 17, 358, 43 },
		{ 17, 359, 43 },
		{ 17, 360, 43 },
		{ 17, 361, 43 },
		{ 17, 362, 43 },
		{ 17, 363, 43 },
		{ 17, 364, 43 },
		{ 17, 365, 43 },
		{ 17, 366, 43 },
		{ 17, 367, 43 },
		{ 17, 368, 43 },
		{ 17, 369, 43 },
		{ 17, 370, 43 },
		{ 17, 371, 43 },
		{ 17, 372, 43 },
		{ 17, 373, 43 },
		{ 17, 374, 43 },
		{ 17, 375, 43 },
		{ 17, 376, 43 },
		{ 18, 317, 43 },
		{ 18, 318, 43 },
		{ 18, 319, 43 },
		{ 18, 320, 43 },
		{ 18, 321, 43 },
		{ 18, 322, 43 },
		{ 18, 323, 43 },
		{ 18, 324, 43 },
		{ 18, 325, 43 },
		{ 18, 326, 43 },
		{ 18, 327, 43 },
		{ 18, 328, 43 },
		{ 18, 329, 43 },
		{ 18, 330, 43 },
		{ 18, 331, 43 },
		{ 18, 332, 43 },
		{ 18, 333, 43 },
		{ 18, 334, 43 },
		{ 18, 335, 43 },
		{ 18, 336, 43 },
		{ 18, 337, 43 },
		{ 18, 338, 43 },
		{ 18, 339, 43 },
		{ 18, 340, 43 },
		{ 18, 341, 43 },
		{ 18, 342, 43 },
		{ 18, 343, 43 },
		{ 18, 344, 43 },
		{ 18, 345, 43 },
		{ 18, 346, 43 },
		{ 18, 347, 43 },
		{ 18, 348, 43 },
		{ 18, 349, 43 },
		{ 18, 350, 43 },
		{ 18, 351, 43 },
		{ 18, 352, 43 },
		{ 18, 353, 43 },
		{ 18, 354, 43 },
		{ 18, 355, 43 },
		{ 18, 356, 43 },
		{ 18, 357, 43 },
		{ 18, 358, 43 },
		{ 18, 359, 43 },
		{ 18, 360, 43 },
		{ 18, 361, 43 },
		{ 18, 362, 43 },
		{ 18, 363, 43 },
		{ 18, 364, 43 },
		{ 18, 365, 43 },
		{ 18, 366, 43 },
		{ 18, 367, 43 },
		{ 18, 368, 43 },
		{ 18, 369, 43 },
		{ 18, 370, 43 },
		{ 18, 371, 43 },
		{ 18, 372, 43 },
		{ 18, 373, 43 },
		{ 18, 374, 43 },
		{ 18, 375, 43 },
		{ 18, 376, 43 },
		{ 19, 317, 43 },
		{ 19, 318, 43 },
		{ 19, 319, 43 },
		{ 19, 320, 43 },
		{ 19, 321, 43 },
		{ 19, 322, 43 },
		{ 19, 323, 43 },
		{ 19, 324, 43 },
		{ 19, 325, 43 },
		{ 19, 326, 43 },
		{ 19, 327, 43 },
		{ 19, 328, 43 },
		{ 19, 329, 43 },
		{ 19, 330, 43 },
		{ 19, 331, 43 },
		{ 19, 332, 43 },
		{ 19, 333, 43 },
		{ 19, 334, 43 },
		{ 19, 335, 43 },
		{ 19, 336, 43 },
		{ 19, 337, 43 },
		{ 19, 338, 43 },
		{ 19, 339, 43 },
		{ 19, 340, 43 },
		{ 19, 341, 43 },
		{ 19, 342, 43 },
		{ 19, 343, 43 },
		{ 19, 344, 43 },
		{ 19, 345, 43 },
		{ 19, 346, 43 },
		{ 19, 347, 43 },
		{ 19, 348, 43 },
		{ 19, 349, 43 },
		{ 19, 350, 43 },
		{ 19, 351, 43 },
		{ 19, 352, 43 },
		{ 19, 353, 43 },
		{ 19, 354, 43 },
		{ 19, 355, 43 },
		{ 19, 356, 43 },
		{ 19, 357, 43 },
		{ 19, 358, 43 },
		{ 19, 359, 43 },
		{ 19, 360, 43 },
		{ 19, 361, 43 },
		{ 19, 362, 43 },
		{ 19, 363, 43 },
		{ 19, 364, 43 },
		{ 19, 365, 43 },
		{ 19, 366, 43 },
		{ 19, 367, 43 },
		{ 19, 368, 43 },
		{ 19, 369, 43 },
		{ 19, 370, 43 },
		{ 19, 371, 43 },
		{ 19, 372, 43 },
		{ 19, 373, 43 },
		{ 19, 374, 43 },
		{ 19, 375, 43 },
		{ 20, 317, 43 },
		{ 20, 318, 43 },
		{ 20, 319, 43 },
		{ 20, 320, 43 },
		{ 20, 321, 43 },
		{ 20, 322, 43 },
		{ 20, 323, 43 },
		{ 20, 324, 43 },
		{ 20, 325, 43 },
		{ 20, 326, 43 },
		{ 20, 327, 43 },
		{ 20, 328, 43 },
		{ 20, 329, 43 },
		{ 20, 330, 43 },
		{ 20, 331, 43 },
		{ 20, 332, 43 },
		{ 20, 333, 43 },
		{ 20, 334, 43 },
		{ 20, 335, 43 },
		{ 20, 336, 43 },
		{ 20, 337, 43 },
		{ 20, 338, 43 },
		{ 20, 339, 43 },
		{ 20, 340, 43 },
		{ 20, 341, 43 },
		{ 20, 342, 43 },
		{ 20, 343, 43 },
		{ 20, 344, 43 },
		{ 20, 345, 43 },
		{ 20, 346, 43 },
		{ 20, 347, 43 },
		{ 20, 348, 43 },
		{ 20, 349, 43 },
		{ 20, 350, 43 },
		{ 20, 351, 43 },
		{ 20, 352, 43 },
		{ 20, 353, 43 },
		{ 20, 354, 43 },
		{ 20, 355, 43 },
		{ 20, 356, 43 },
		{ 20, 357, 43 },
		{ 20, 358, 43 },
		{ 20, 359, 43 },
		{ 20, 360, 43 },
		{ 20, 361, 43 },
		{ 20, 362, 43 },
		{ 20, 363, 43 },
		{ 20, 364, 43 },
		{ 20, 365, 43 },
		{ 20, 366, 43 },
		{ 20, 367, 43 },
		{ 20, 368, 43 },
		{ 20, 369, 43 },
		{ 20, 370, 43 },
		{ 20, 371, 43 },
		{ 20, 372, 43 },
		{ 20, 373, 43 },
		{ 20, 374, 43 },
		{ 20, 375, 43 },
		{ 21, 317, 43 },
		{ 21, 318, 43 },
		{ 21, 319, 43 },
		{ 21, 320, 43 },
		{ 21, 321, 43 },
		{ 21, 322, 43 },
		{ 21, 323, 43 },
		{ 21, 324, 43 },
		{ 21, 325, 43 },
		{ 21, 326, 43 },
		{ 21, 327, 43 },
		{ 21, 328, 43 },
		{ 21, 329, 43 },
		{ 21, 330, 43 },
		{ 21, 331, 43 },
		{ 21, 332, 43 },
		{ 21, 333, 43 },
		{ 21, 334, 43 },
		{ 21, 335, 43 },
		{ 21, 336, 43 },
		{ 21, 337, 43 },
		{ 21, 338, 43 },
		{ 21, 339, 43 },
		{ 21, 340, 43 },
		{ 21, 341, 43 },
		{ 21, 342, 43 },
		{ 21, 343, 43 },
		{ 21, 344, 43 },
		{ 21, 345, 43 },
		{ 21, 346, 43 },
		{ 21, 347, 43 },
		{ 21, 348, 43 },
		{ 21, 349, 43 },
		{ 21, 350, 43 },
		{ 21, 351, 43 },
		{ 21, 352, 43 },
		{ 21, 353, 43 },
		{ 21, 354, 43 },
		{ 21, 355, 43 },
		{ 21, 356, 43 },
		{ 21, 357, 43 },
		{ 21, 358, 43 },
		{ 21, 359, 43 },
		{ 21, 360, 43 },
		{ 21, 361, 43 },
		{ 21, 362, 43 },
		{ 21, 363, 43 },
		{ 21, 364, 43 },
		{ 21, 365, 43 },
		{ 21, 366, 43 },
		{ 21, 367, 43 },
		{ 21, 368, 43 },
		{ 21, 369, 43 },
		{ 21, 370, 43 },
		{ 21, 371, 43 },
		{ 21, 372, 43 },
		{ 21, 373, 43 },
		{ 21, 374, 43 },
		{ 22, 316, 43 },
		{ 22, 317, 43 },
		{ 22, 318, 43 },
		{ 22, 319, 43 },
		{ 22, 320, 43 },
		{ 22, 321, 43 },
		{ 22, 322, 43 },
		{ 22, 323, 43 },
		{ 22, 324, 43 },
		{ 22, 325, 43 },
		{ 22, 326, 43 },
		{ 22, 327, 43 },
		{ 22, 328, 43 },
		{ 22, 329, 43 },
		{ 22, 330, 43 },
		{ 22, 331, 43 },
		{ 22, 332, 43 },
		{ 22, 333, 43 },
		{ 22, 334, 43 },
		{ 22, 335, 43 },
		{ 22, 336, 43 },
		{ 22, 337, 43 },
		{ 22, 338, 43 },
		{ 22, 339, 43 },
		{ 22, 340, 43 },
		{ 22, 341, 43 },
		{ 22, 342, 43 },
		{ 22, 343, 43 },
		{ 22, 344, 43 },
		{ 22, 345, 43 },
		{ 22, 346, 43 },
		{ 22, 347, 43 },
		{ 22, 348, 43 },
		{ 22, 349, 43 },
		{ 22, 350, 43 },
		{ 22, 351, 43 },
		{ 22, 352, 43 },
		{ 22, 353, 43 },
		{ 22, 354, 43 },
		{ 22, 355, 43 },
		{ 22, 356, 43 },
		{ 22, 357, 43 },
		{ 22, 358, 43 },
		{ 22, 359, 43 },
		{ 22, 360, 43 },
		{ 22, 361, 43 },
		{ 22, 362, 43 },
		{ 22, 363, 43 },
		{ 22, 364, 43 },
		{ 22, 365, 43 },
		{ 22, 366, 43 },
		{ 22, 367, 43 },
		{ 22, 368, 43 },
		{ 22, 369, 43 },
		{ 22, 370, 43 },
		{ 22, 371, 43 },
		{ 22, 372, 43 },
		{ 22, 373, 43 },
		{ 22, 374, 43 },
		{ 23, 316, 43 },
		{ 23, 317, 43 },
		{ 23, 318, 43 },
		{ 23, 319, 43 },
		{ 23, 320, 43 },
		{ 23, 321, 43 },
		{ 23, 322, 43 },
		{ 23, 323, 43 },
		{ 23, 324, 43 },
		{ 23, 325, 43 },
		{ 23, 326, 43 },
		{ 23, 327, 43 },
		{ 23, 328, 43 },
		{ 23, 329, 43 },
		{ 23, 330, 43 },
		{ 23, 331, 43 },
		{ 23, 332, 43 },
		{ 23, 333, 43 },
		{ 23, 334, 43 },
		{ 23, 335, 43 },
		{ 23, 336, 43 },
		{ 23, 337, 43 },
		{ 23, 338, 43 },
		{ 23, 339, 43 },
		{ 23, 340, 43 },
		{ 23, 341, 43 },
		{ 23, 342, 43 },
		{ 23, 343, 43 },
		{ 23, 344, 43 },
		{ 23, 345, 43 },
		{ 23, 346, 43 },
		{ 23, 347, 43 },
		{ 23, 348, 43 },
		{ 23, 349, 43 },
		{ 23, 350, 43 },
		{ 23, 351, 43 },
		{ 23, 352, 43 },
		{ 23, 353, 43 },
		{ 23, 354, 43 },
		{ 23, 355, 43 },
		{ 23, 356, 43 },
		{ 23, 357, 43 },
		{ 23, 358, 43 },
		{ 23, 359, 43 },
		{ 23, 360, 43 },
		{ 23, 361, 43 },
		{ 23, 362, 43 },
		{ 23, 363, 43 },
		{ 23, 364, 43 },
		{ 23, 365, 43 },
		{ 23, 366, 43 },
		{ 23, 367, 43 },
		{ 23, 368, 43 },
		{ 23, 369, 43 },
		{ 23, 370, 43 },
		{ 23, 371, 43 },
		{ 23, 372, 43 },
		{ 23, 373, 43 },
		{ 24, 316, 43 },
		{ 24, 317, 43 },
		{ 24, 318, 43 },
		{ 24, 319, 43 },
		{ 24, 320, 43 },
		{ 24, 321, 43 },
		{ 24, 322, 43 },
		{ 24, 323, 43 },
		{ 24, 324, 43 },
		{ 24, 325, 43 },
		{ 24, 326, 43 },
		{ 24, 327, 43 },
		{ 24, 328, 43 },
		{ 24, 329, 43 },
		{ 24, 330, 43 },
		{ 24, 331, 43 },
		{ 24, 332, 43 },
		{ 24, 333, 43 },
		{ 24, 334, 43 },
		{ 24, 335, 43 },
		{ 24, 336, 43 },
		{ 24, 337, 43 },
		{ 24, 338, 43 },
		{ 24, 339, 43 },
		{ 24, 340, 43 },
		{ 24, 341, 43 },
		{ 24, 342, 43 },
		{ 24, 343, 43 },
		{ 24, 344, 43 },
		{ 24, 345, 43 },
		{ 24, 346, 43 },
		{ 24, 347, 43 },
		{ 24, 348, 43 },
		{ 24, 349, 43 },
		{ 24, 350, 43 },
		{ 24, 351, 43 },
		{ 24, 352, 43 },
		{ 24, 353, 43 },
		{ 24, 354, 43 },
		{ 24, 355, 43 },
		{ 24, 356, 43 },
		{ 24, 357, 43 },
		{ 24, 358, 43 },
		{ 24, 359, 43 },
		{ 24, 360, 43 },
		{ 24, 361, 43 },
		{ 24, 362, 43 },
		{ 24, 363, 43 },
		{ 24, 364, 43 },
		{ 24, 365, 43 },
		{ 24, 366, 43 },
		{ 24, 367, 43 },
		{ 24, 368, 43 },
		{ 24, 369, 43 },
		{ 24, 370, 43 },
		{ 24, 371, 43 },
		{ 24, 372, 43 },
		{ 24, 373, 43 },
		{ 25, 316, 43 },
		{ 25, 317, 43 },
		{ 25, 318, 43 },
		{ 25, 319, 43 },
		{ 25, 320, 43 },
		{ 25, 321, 43 },
		{ 25, 322, 43 },
		{ 25, 323, 43 },
		{ 25, 324, 43 },
		{ 25, 325, 43 },
		{ 25, 326, 43 },
		{ 25, 327, 43 },
		{ 25, 328, 43 },
		{ 25, 329, 43 },
		{ 25, 330, 43 },
		{ 25, 331, 43 },
		{ 25, 332, 43 },
		{ 25, 333, 43 },
		{ 25, 334, 43 },
		{ 25, 335, 43 },
		{ 25, 336, 43 },
		{ 25, 337, 43 },
		{ 25, 338, 43 },
		{ 25, 339, 43 },
		{ 25, 340, 43 },
		{ 25, 341, 43 },
		{ 25, 342, 43 },
		{ 25, 343, 43 },
		{ 25, 344, 43 },
		{ 25, 345, 43 },
		{ 25, 346, 43 },
		{ 25, 347, 43 },
		{ 25, 348, 43 },
		{ 25, 349, 43 },
		{ 25, 350, 43 },
		{ 25, 351, 43 },
		{ 25, 352, 43 },
		{ 25, 353, 43 },
		{ 25, 354, 43 },
		{ 25, 355, 43 },
		{ 25, 356, 43 },
		{ 25, 357, 43 },
		{ 25, 358, 43 },
		{ 25, 359, 43 },
		{ 25, 360, 43 },
		{ 25, 361, 43 },
		{ 25, 362, 43 },
		{ 25, 363, 43 },
		{ 25, 364, 43 },
		{ 25, 365, 43 },
		{ 25, 366, 43 },
		{ 25, 367, 43 },
		{ 25, 368, 43 },
		{ 25, 369, 43 },
		{ 25, 370, 43 },
		{ 25, 371, 43 },
		{ 25, 372, 43 },
		{ 26, 315, 43 },
		{ 26, 316, 43 },
		{ 26, 317, 43 },
		{ 26, 318, 43 },
		{ 26, 319, 43 },
		{ 26, 320, 43 },
		{ 26, 321, 43 },
		{ 26, 322, 43 },
		{ 26, 323, 43 },
		{ 26, 324, 43 },
		{ 26, 325, 43 },
		{ 26, 326, 43 },
		{ 26, 327, 43 },
		{ 26, 328, 43 },
		{ 26, 329, 43 },
		{ 26, 330, 43 },
		{ 26, 331, 43 },
		{ 26, 332, 43 },
		{ 26, 333, 43 },
		{ 26, 334, 43 },
		{ 26, 335, 43 },
		{ 26, 336, 43 },
		{ 26, 337, 43 },
		{ 26, 338, 43 },
		{ 26, 339, 43 },
		{ 26, 340, 43 },
		{ 26, 341, 43 },
		{ 26, 342, 43 },
		{ 26, 343, 43 },
		{ 26, 344, 43 },
		{ 26, 345, 43 },
		{ 26, 346, 43 },
		{ 26, 347, 43 },
		{ 26, 348, 43 },
		{ 26, 349, 43 },
		{ 26, 350, 43 },
		{ 26, 351, 43 },
		{ 26, 352, 43 },
		{ 26, 353, 43 },
		{ 26, 354, 43 },
		{ 26, 355, 43 },
		{ 26, 356, 43 },
		{ 26, 357, 43 },
		{ 26, 358, 43 },
		{ 26, 359, 43 },
		{ 26, 360, 43 },
		{ 26, 361, 43 },
		{ 26, 362, 43 },
		{ 26, 363, 43 },
		{ 26, 364, 43 },
		{ 26, 365, 43 },
		{ 26, 366, 43 },
		{ 26, 367, 43 },
		{ 26, 368, 43 },
		{ 26, 369, 43 },
		{ 26, 370, 43 },
		{ 26, 371, 43 },
		{ 26, 372, 43 },
		{ 27, 315, 43 },
		{ 27, 316, 43 },
		{ 27, 317, 43 },
		{ 27, 318, 43 },
		{ 27, 319, 43 },
		{ 27, 320, 43 },
		{ 27, 321, 43 },
		{ 27, 322, 43 },
		{ 27, 323, 43 },
		{ 27, 324, 43 },
		{ 27, 325, 43 },
		{ 27, 326, 43 },
		{ 27, 327, 43 },
		{ 27, 328, 43 },
		{ 27, 329, 43 },
		{ 27, 330, 43 },
		{ 27, 331, 43 },
		{ 27, 332, 43 },
		{ 27, 333, 43 },
		{ 27, 334, 43 },
		{ 27, 335, 43 },
		{ 27, 336, 43 },
		{ 27, 337, 43 },
		{ 27, 338, 43 },
		{ 27, 339, 43 },
		{ 27, 340, 43 },
		{ 27, 341, 43 },
		{ 27, 342, 43 },
		{ 27, 343, 43 },
		{ 27, 344, 43 },
		{ 27, 345, 43 },
		{ 27, 346, 43 },
		{ 27, 347, 43 },
		{ 27, 348, 43 },
		{ 27, 349, 43 },
		{ 27, 350, 43 },
		{ 27, 351, 43 },
		{ 27, 352, 43 },
		{ 27, 353, 43 },
		{ 27, 354, 43 },
		{ 27, 355, 43 },
		{ 27, 356, 43 },
		{ 27, 357, 43 },
		{ 27, 358, 43 },
		{ 27, 359, 43 },
		{ 27, 360, 43 },
		{ 27, 361, 43 },
		{ 27, 362, 43 },
		{ 27, 363, 43 },
		{ 27, 364, 43 },
		{ 27, 365, 43 },
		{ 27, 366, 43 },
		{ 27, 367, 43 },
		{ 27, 368, 43 },
		{ 27, 369, 43 },
		{ 27, 370, 43 },
		{ 27, 371, 43 },
		{ 28, 315, 43 },
		{ 28, 316, 43 },
		{ 28, 317, 43 },
		{ 28, 318, 43 },
		{ 28, 319, 43 },
		{ 28, 320, 43 },
		{ 28, 321, 43 },
		{ 28, 322, 43 },
		{ 28, 323, 43 },
		{ 28, 324, 43 },
		{ 28, 325, 43 },
		{ 28, 326, 43 },
		{ 28, 327, 43 },
		{ 28, 328, 43 },
		{ 28, 329, 43 },
		{ 28, 330, 43 },
		{ 28, 331, 43 },
		{ 28, 332, 43 },
		{ 28, 333, 43 },
		{ 28, 334, 43 },
		{ 28, 335, 43 },
		{ 28, 336, 43 },
		{ 28, 337, 43 },
		{ 28, 338, 43 },
		{ 28, 339, 43 },
		{ 28, 340, 43 },
		{ 28, 341, 43 },
		{ 28, 342, 43 },
		{ 28, 343, 43 },
		{ 28, 344, 43 },
		{ 28, 345, 43 },
		{ 28, 346, 43 },
		{ 28, 347, 43 },
		{ 28, 348, 43 },
		{ 28, 349, 43 },
		{ 28, 350, 43 },
		{ 28, 351, 43 },
		{ 28, 352, 43 },
		{ 28, 353, 43 },
		{ 28, 354, 43 },
		{ 28, 355, 43 },
		{ 28, 356, 43 },
		{ 28, 357, 43 },
		{ 28, 358, 43 },
		{ 28, 359, 43 },
		{ 28, 360, 43 },
		{ 28, 361, 43 },
		{ 28, 362, 43 },
		{ 28, 363, 43 },
		{ 28, 364, 43 },
		{ 28, 365, 43 },
		{ 28, 366, 43 },
		{ 28, 367, 43 },
		{ 28, 368, 43 },
		{ 28, 369, 43 },
		{ 28, 370, 43 },
		{ 28, 371, 43 },
		{ 29, 315, 43 },
		{ 29, 316, 43 },
		{ 29, 317, 43 },
		{ 29, 318, 43 },
		{ 29, 319, 43 },
		{ 29, 320, 43 },
		{ 29, 321, 43 },
		{ 29, 322, 43 },
		{ 29, 323, 43 },
		{ 29, 324, 43 },
		{ 29, 325, 43 },
		{ 29, 326, 43 },
		{ 29, 327, 43 },
		{ 29, 328, 43 },
		{ 29, 329, 43 },
		{ 29, 330, 43 },
		{ 29, 331, 43 },
		{ 29, 332, 43 },
		{ 29, 333, 43 },
		{ 29, 334, 43 },
		{ 29, 335, 43 },
		{ 29, 336, 43 },
		{ 29, 337, 43 },
		{ 29, 338, 43 },
		{ 29, 339, 43 },
		{ 29, 340, 43 },
		{ 29, 341, 43 },
		{ 29, 342, 43 },
		{ 29, 343, 43 },
		{ 29, 344, 43 },
		{ 29, 345, 43 },
		{ 29, 346, 43 },
		{ 29, 347, 43 },
		{ 29, 348, 43 },
		{ 29, 349, 43 },
		{ 29, 350, 43 },
		{ 29, 351, 43 },
		{ 29, 352, 43 },
		{ 29, 353, 43 },
		{ 29, 354, 43 },
		{ 29, 355, 43 },
		{ 29, 356, 43 },
		{ 29, 357, 43 },
		{ 29, 358, 43 },
		{ 29, 359, 43 },
		{ 29, 360, 43 },
		{ 29, 361, 43 },
		{ 29, 362, 43 },
		{ 29, 363, 43 },
		{ 29, 364, 43 },
		{ 29, 365, 43 },
		{ 29, 366, 43 },
		{ 29, 367, 43 },
		{ 29, 368, 43 },
		{ 29, 369, 43 },
		{ 29, 370, 43 },
		{ 30, 314, 43 },
		{ 30, 315, 43 },
		{ 30, 316, 43 },
		{ 30, 317, 43 },
		{ 30, 318, 43 },
		{ 30, 319, 43 },
		{ 30, 320, 43 },
		{ 30, 321, 43 },
		{ 30, 322, 43 },
		{ 30, 323, 43 },
		{ 30, 324, 43 },
		{ 30, 325, 43 },
		{ 30, 326, 43 },
		{ 30, 327, 43 },
		{ 30, 328, 43 },
		{ 30, 329, 43 },
		{ 30, 330, 43 },
		{ 30, 331, 43 },
		{ 30, 332, 43 },
		{ 30, 333, 43 },
		{ 30, 334, 43 },
		{ 30, 335, 43 },
		{ 30, 336, 43 },
		{ 30, 337, 43 },
		{ 30, 338, 43 },
		{ 30, 339, 43 },
		{ 30, 340, 43 },
		{ 30, 341, 43 },
		{ 30, 342, 43 },
		{ 30, 343, 43 },
		{ 30, 344, 43 },
		{ 30, 345, 43 },
		{ 30, 346, 43 },
		{ 30, 347, 43 },
		{ 30, 348, 43 },
		{ 30, 349, 43 },
		{ 30, 350, 43 },
		{ 30, 351, 43 },
		{ 30, 352, 43 },
		{ 30, 353, 43 },
		{ 30, 354, 43 },
		{ 30, 355, 43 },
		{ 30, 356, 43 },
		{ 30, 357, 43 },
		{ 30, 358, 43 },
		{ 30, 359, 43 },
		{ 30, 360, 43 },
		{ 30, 361, 43 },
		{ 30, 362, 43 },
		{ 30, 363, 43 },
		{ 30, 364, 43 },
		{ 30, 365, 43 },
		{ 30, 366, 43 },
		{ 30, 367, 43 },
		{ 30, 368, 43 },
		{ 30, 369, 43 },
		{ 30, 370, 43 },
		{ 31, 314, 43 },
		{ 31, 315, 43 },
		{ 31, 316, 43 },
		{ 31, 317, 43 },
		{ 31, 318, 43 },
		{ 31, 319, 43 },
		{ 31, 320, 43 },
		{ 31, 321, 43 },
		{ 31, 322, 43 },
		{ 31, 323, 43 },
		{ 31, 324, 43 },
		{ 31, 325, 43 },
		{ 31, 326, 43 },
		{ 31, 327, 43 },
		{ 31, 328, 43 },
		{ 31, 329, 43 },
		{ 31, 330, 43 },
		{ 31, 331, 43 },
		{ 31, 332, 43 },
		{ 31, 333, 43 },
		{ 31, 334, 43 },
		{ 31, 335, 43 },
		{ 31, 336, 43 },
		{ 31, 337, 43 },
		{ 31, 338, 43 },
		{ 31, 339, 43 },
		{ 31, 340, 43 },
		{ 31, 341, 43 },
		{ 31, 342, 43 },
		{ 31, 343, 43 },
		{ 31, 344, 43 },
		{ 31, 345, 43 },
		{ 31, 346, 43 },
		{ 31, 347, 43 },
		{ 31, 348, 43 },
		{ 31, 349, 43 },
		{ 31, 350, 43 },
		{ 31, 351, 43 },
		{ 31, 352, 43 },
		{ 31, 353, 43 },
		{ 31, 354, 43 },
		{ 31, 355, 43 },
		{ 31, 356, 43 },
		{ 31, 357, 43 },
		{ 31, 358, 43 },
		{ 31, 359, 43 },
		{ 31, 360, 43 },
		{ 31, 361, 43 },
		{ 31, 362, 43 },
		{ 31, 363, 43 },
		{ 31, 364, 43 },
		{ 31, 365, 43 },
		{ 31, 366, 43 },
		{ 31, 367, 43 },
		{ 31, 368, 43 },
		{ 31, 369, 43 },
		{ 32, 314, 43 },
		{ 32, 315, 43 },
		{ 32, 316, 43 },
		{ 32, 317, 43 },
		{ 32, 318, 43 },
		{ 32, 319, 43 },
		{ 32, 320, 43 },
		{ 32, 321, 43 },
		{ 32, 322, 43 },
		{ 32, 323, 43 },
		{ 32, 324, 43 },
		{ 32, 325, 43 },
		{ 32, 326, 43 },
		{ 32, 327, 43 },
		{ 32, 328, 43 },
		{ 32, 329, 43 },
		{ 32, 330, 43 },
		{ 32, 331, 43 },
		{ 32, 332, 43 },
		{ 32, 333, 43 },
		{ 32, 334, 43 },
		{ 32, 335, 43 },
		{ 32, 336, 43 },
		{ 32, 337, 43 },
		{ 32, 338, 43 },
		{ 32, 339, 43 },
		{ 32, 340, 43 },
		{ 32, 341, 43 },
		{ 32, 342, 43 },
		{ 32, 343, 43 },
		{ 32, 344, 43 },
		{ 32, 345, 43 },
		{ 32, 346, 43 },
		{ 32, 347, 43 },
		{ 32, 348, 43 },
		{ 32, 349, 43 },
		{ 32, 350, 43 },
		{ 32, 351, 43 },
		{ 32, 352, 43 },
		{ 32, 353, 43 },
		{ 32, 354, 43 },
		{ 32, 355, 43 },
		{ 32, 356, 43 },
		{ 32, 357, 43 },
		{ 32, 358, 43 },
		{ 32, 359, 43 },
		{ 32, 360, 43 },
		{ 32, 361, 43 },
		{ 32, 362, 43 },
		{ 32, 363, 43 },
		{ 32, 364, 43 },
		{ 32, 365, 43 },
		{ 32, 366, 43 },
		{ 32, 367, 43 },
		{ 32, 368, 43 },
		{ 32, 369, 43 },
		{ 33, 314, 43 },
		{ 33, 315, 43 },
		{ 33, 316, 43 },
		{ 33, 317, 43 },
		{ 33, 318, 43 },
		{ 33, 319, 43 },
		{ 33, 320, 43 },
		{ 33, 321, 43 },
		{ 33, 322, 43 },
		{ 33, 323, 43 },
		{ 33, 324, 43 },
		{ 33, 325, 43 },
		{ 33, 326, 43 },
		{ 33, 327, 43 },
		{ 33, 328, 43 },
		{ 33, 329, 43 },
		{ 33, 330, 43 },
		{ 33, 331, 43 },
		{ 33, 332, 43 },
		{ 33, 333, 43 },
		{ 33, 334, 43 },
		{ 33, 335, 43 },
		{ 33, 336, 43 },
		{ 33, 337, 43 },
		{ 33, 338, 43 },
		{ 33, 339, 43 },
		{ 33, 340, 43 },
		{ 33, 341, 43 },
		{ 33, 342, 43 },
		{ 33, 343, 43 },
		{ 33, 344, 43 },
		{ 33, 345, 43 },
		{ 33, 346, 43 },
		{ 33, 347, 43 },
		{ 33, 348, 43 },
		{ 33, 349, 43 },
		{ 33, 350, 43 },
		{ 33, 351, 43 },
		{ 33, 352, 43 },
		{ 33, 353, 43 },
		{ 33, 354, 43 },
		{ 33, 355, 43 },
		{ 33, 356, 43 },
		{ 33, 357, 43 },
		{ 33, 358, 43 },
		{ 33, 359, 43 },
		{ 33, 360, 43 },
		{ 33, 361, 43 },
		{ 33, 362, 43 },
		{ 33, 363, 43 },
		{ 33, 364, 43 },
		{ 33, 365, 43 },
		{ 33, 366, 43 },
		{ 33, 367, 43 },
		{ 33, 368, 43 },
		{ 34, 313, 43 },
		{ 34, 314, 43 },
		{ 34, 315, 43 },
		{ 34, 316, 43 },
		{ 34, 317, 43 },
		{ 34, 318, 43 },
		{ 34, 319, 43 },
		{ 34, 320, 43 },
		{ 34, 321, 43 },
		{ 34, 322, 43 },
		{ 34, 323, 43 },
		{ 34, 324, 43 },
		{ 34, 325, 43 },
		{ 34, 326, 43 },
		{ 34, 327, 43 },
		{ 34, 328, 43 },
		{ 34, 329, 43 },
		{ 34, 330, 43 },
		{ 34, 331, 43 },
		{ 34, 332, 43 },
		{ 34, 333, 43 },
		{ 34, 334, 43 },
		{ 34, 335, 43 },
		{ 34, 336, 43 },
		{ 34, 337, 43 },
		{ 34, 338, 43 },
		{ 34, 339, 43 },
		{ 34, 340, 43 },
		{ 34, 341, 43 },
		{ 34, 342, 43 },
		{ 34, 343, 43 },
		{ 34, 344, 43 },
		{ 34, 345, 43 },
		{ 34, 346, 43 },
		{ 34, 347, 43 },
		{ 34, 348, 43 },
		{ 34, 349, 43 },
		{ 34, 350, 43 },
		{ 34, 351, 43 },
		{ 34, 352, 43 },
		{ 34, 353, 43 },
		{ 34, 354, 43 },
		{ 34, 355, 43 },
		{ 34, 356, 43 },
		{ 34, 357, 43 },
		{ 34, 358, 43 },
		{ 34, 359, 43 },
		{ 34, 360, 43 },
		{ 34, 361, 43 },
		{ 34, 362, 43 },
		{ 34, 363, 43 },
		{ 34, 364, 43 },
		{ 34, 365, 43 },
		{ 34, 366, 43 },
		{ 34, 367, 43 },
		{ 34, 368, 43 },
		{ 35, 313, 43 },
		{ 35, 314, 43 },
		{ 35, 315, 43 },
		{ 35, 316, 43 },
		{ 35, 317, 43 },
		{ 35, 318, 43 },
		{ 35, 319, 43 },
		{ 35, 320, 43 },
		{ 35, 321, 43 },
		{ 35, 322, 43 },
		{ 35, 323, 43 },
		{ 35, 324, 43 },
		{ 35, 325, 43 },
		{ 35, 326, 43 },
		{ 35, 327, 43 },
		{ 35, 328, 43 },
		{ 35, 329, 43 },
		{ 35, 330, 43 },
		{ 35, 331, 43 },
		{ 35, 332, 43 },
		{ 35, 333, 43 },
		{ 35, 334, 43 },
		{ 35, 335, 43 },
		{ 35, 336, 43 },
		{ 35, 337, 43 },
		{ 35, 338, 43 },
		{ 35, 339, 43 },
		{ 35, 340, 43 },
		{ 35, 341, 43 },
		{ 35, 342, 43 },
		{ 35, 343, 43 },
		{ 35, 344, 43 },
		{ 35, 345, 43 },
		{ 35, 346, 43 },
		{ 35, 347, 43 },
		{ 35, 348, 43 },
		{ 35, 349, 43 },
		{ 35, 350, 43 },
		{ 35, 351, 43 },
		{ 35, 352, 43 },
		{ 35, 353, 43 },
		{ 35, 354, 43 },
		{ 35, 355, 43 },
		{ 35, 356, 43 },
		{ 35, 357, 43 },
		{ 35, 358, 43 },
		{ 35, 359, 43 },
		{ 35, 360, 43 },
		{ 35, 361, 43 },
		{ 35, 362, 43 },
		{ 35, 363, 43 },
		{ 35, 364, 43 },
		{ 35, 365, 43 },
		{ 35, 366, 43 },
		{ 35, 367, 43 },
		{ 36, 313, 43 },
		{ 36, 314, 43 },
		{ 36, 315, 43 },
		{ 36, 316, 43 },
		{ 36, 317, 43 },
		{ 36, 318, 43 },
		{ 36, 319, 43 },
		{ 36, 320, 43 },
		{ 36, 321, 43 },
		{ 36, 322, 43 },
		{ 36, 323, 43 },
		{ 36, 324, 43 },
		{ 36, 325, 43 },
		{ 36, 326, 43 },
		{ 36, 327, 43 },
		{ 36, 328, 43 },
		{ 36, 329, 43 },
		{ 36, 330, 43 },
		{ 36, 331, 43 },
		{ 36, 332, 43 },
		{ 36, 333, 43 },
		{ 36, 334, 43 },
		{ 36, 335, 43 },
		{ 36, 336, 43 },
		{ 36, 337, 43 },
		{ 36, 338, 43 },
		{ 36, 339, 43 },
		{ 36, 340, 43 },
		{ 36, 341, 43 },
		{ 36, 342, 43 },
		{ 36, 343, 43 },
		{ 36, 344, 43 },
		{ 36, 345, 43 },
		{ 36, 346, 43 },
		{ 36, 347, 43 },
		{ 36, 348, 43 },
		{ 36, 349, 43 },
		{ 36, 350, 43 },
		{ 36, 351, 43 },
		{ 36, 352, 43 },
		{ 36, 353, 43 },
		{ 36, 354, 43 },
		{ 36, 355, 43 },
		{ 36, 356, 43 },
		{ 36, 357, 43 },
		{ 36, 358, 43 },
		{ 36, 359, 43 },
		{ 36, 360, 43 },
		{ 36, 361, 43 },
		{ 36, 362, 43 },
		{ 36, 363, 43 },
		{ 36, 364, 43 },
		{ 36, 365, 43 },
		{ 36, 366, 43 },
		{ 36, 367, 43 },
		{ 37, 313, 43 },
		{ 37, 314, 43 },
		{ 37, 315, 43 },
		{ 37, 316, 43 },
		{ 37, 317, 43 },
		{ 37, 318, 43 },
		{ 37, 319, 43 },
		{ 37, 320, 43 },
		{ 37, 321, 43 },
		{ 37, 322, 43 },
		{ 37, 323, 43 },
		{ 37, 324, 43 },
		{ 37, 325, 43 },
		{ 37, 326, 43 },
		{ 37, 327, 43 },
		{ 37, 328, 43 },
		{ 37, 329, 43 },
		{ 37, 330, 43 },
		{ 37, 331, 43 },
		{ 37, 332, 43 },
		{ 37, 333, 43 },
		{ 37, 334, 43 },
		{ 37, 335, 43 },
		{ 37, 336, 43 },
		{ 37, 337, 43 },
		{ 37, 338, 43 },
		{ 37, 339, 43 },
		{ 37, 340, 43 },
		{ 37, 341, 43 },
		{ 37, 342, 43 },
		{ 37, 343, 43 },
		{ 37, 344, 43 },
		{ 37, 345, 43 },
		{ 37, 346, 43 },
		{ 37, 347, 43 },
		{ 37, 348, 43 },
		{ 37, 349, 43 },
		{ 37, 350, 43 },
		{ 37, 351, 43 },
		{ 37, 352, 43 },
		{ 37, 353, 43 },
		{ 37, 354, 43 },
		{ 37, 355, 43 },
		{ 37, 356, 43 },
		{ 37, 357, 43 },
		{ 37, 358, 43 },
		{ 37, 359, 43 },
		{ 37, 360, 43 },
		{ 37, 361, 43 },
		{ 37, 362, 43 },
		{ 37, 363, 43 },
		{ 37, 364, 43 },
		{ 37, 365, 43 },
		{ 37, 366, 43 },
		{ 38, 312, 43 },
		{ 38, 313, 43 },
		{ 38, 314, 43 },
		{ 38, 315, 43 },
		{ 38, 316, 43 },
		{ 38, 317, 43 },
		{ 38, 318, 43 },
		{ 38, 319, 43 },
		{ 38, 320, 43 },
		{ 38, 321, 43 },
		{ 38, 322, 43 },
		{ 38, 323, 43 },
		{ 38, 324, 43 },
		{ 38, 325, 43 },
		{ 38, 326, 43 },
		{ 38, 327, 43 },
		{ 38, 328, 43 },
		{ 38, 329, 43 },
		{ 38, 330, 43 },
		{ 38, 331, 43 },
		{ 38, 332, 43 },
		{ 38, 333, 43 },
		{ 38, 334, 43 },
		{ 38, 335, 43 },
		{ 38, 336, 43 },
		{ 38, 337, 43 },
		{ 38, 338, 43 },
		{ 38, 339, 43 },
		{ 38, 340, 43 },
		{ 38, 341, 43 },
		{ 38, 342, 43 },
		{ 38, 343, 43 },
		{ 38, 344, 43 },
		{ 38, 345, 43 },
		{ 38, 346, 43 },
		{ 38, 347, 43 },
		{ 38, 348, 43 },
		{ 38, 349, 43 },
		{ 38, 350, 43 },
		{ 38, 351, 43 },
		{ 38, 352, 43 },
		{ 38, 353, 43 },
		{ 38, 354, 43 },
		{ 38, 355, 43 },
		{ 38, 356, 43 },
		{ 38, 357, 43 },
		{ 38, 358, 43 },
		{ 38, 359, 43 },
		{ 38, 360, 43 },
		{ 38, 361, 43 },
		{ 38, 362, 43 },
		{ 38, 363, 43 },
		{ 38, 364, 43 },
		{ 38, 365, 43 },
		{ 38, 366, 43 },
		{ 39, 312, 43 },
		{ 39, 313, 43 },
		{ 39, 314, 43 },
		{ 39, 315, 43 },
		{ 39, 316, 43 },
		{ 39, 317, 43 },
		{ 39, 318, 43 },
		{ 39, 319, 43 },
		{ 39, 320, 43 },
		{ 39, 321, 43 },
		{ 39, 322, 43 },
		{ 39, 323, 43 },
		{ 39, 324, 43 },
		{ 39, 325, 43 },
		{ 39, 326, 43 },
		{ 39, 327, 43 },
		{ 39, 328, 43 },
		{ 39, 329, 43 },
		{ 39, 330, 43 },
		{ 39, 331, 43 },
		{ 39, 332, 43 },
		{ 39, 333, 43 },
		{ 39, 334, 43 },
		{ 39, 335, 43 },
		{ 39, 336, 43 },
		{ 39, 337, 43 },
		{ 39, 338, 43 },
		{ 39, 339, 43 },
		{ 39, 340, 43 },
		{ 39, 341, 43 },
		{ 39, 342, 43 },
		{ 39, 343, 43 },
		{ 39, 344, 43 },
		{ 39, 345, 43 },
		{ 39, 346, 43 },
		{ 39, 347, 43 },
		{ 39, 348, 43 },
		{ 39, 349, 43 },
		{ 39, 350, 43 },
		{ 39, 351, 43 },
		{ 39, 352, 43 },
		{ 39, 353, 43 },
		{ 39, 354, 43 },
		{ 39, 355, 43 },
		{ 39, 356, 43 },
		{ 39, 357, 43 },
		{ 39, 358, 43 },
		{ 39, 359, 43 },
		{ 39, 360, 43 },
		{ 39, 361, 43 },
		{ 39, 362, 43 },
		{ 39, 363, 43 },
		{ 39, 364, 43 },
		{ 39, 365, 43 },
		{ 40, 312, 43 },
		{ 40, 313, 43 },
		{ 40, 314, 43 },
		{ 40, 315, 43 },
		{ 40, 316, 43 },
		{ 40, 317, 43 },
		{ 40, 318, 43 },
		{ 40, 319, 43 },
		{ 40, 320, 43 },
		{ 40, 321, 43 },
		{ 40, 322, 43 },
		{ 40, 323, 43 },
		{ 40, 324, 43 },
		{ 40, 325, 43 },
		{ 40, 326, 43 },
		{ 40, 327, 43 },
		{ 40, 328, 43 },
		{ 40, 329, 43 },
		{ 40, 330, 43 },
		{ 40, 331, 43 },
		{ 40, 332, 43 },
		{ 40, 333, 43 },
		{ 40, 334, 43 },
		{ 40, 335, 43 },
		{ 40, 336, 43 },
		{ 40, 337, 43 },
		{ 40, 338, 43 },
		{ 40, 339, 43 },
		{ 40, 340, 43 },
		{ 40, 341, 43 },
		{ 40, 342, 43 },
		{ 40, 343, 43 },
		{ 40, 344, 43 },
		{ 40, 345, 43 },
		{ 40, 346, 43 },
		{ 40, 347, 43 },
		{ 40, 348, 43 },
		{ 40, 349, 43 },
		{ 40, 350, 43 },
		{ 40, 351, 43 },
		{ 40, 352, 43 },
		{ 40, 353, 43 },
		{ 40, 354, 43 },
		{ 40, 355, 43 },
		{ 40, 356, 43 },
		{ 40, 357, 43 },
		{ 40, 358, 43 },
		{ 40, 359, 43 },
		{ 40, 360, 43 },
		{ 40, 361, 43 },
		{ 40, 362, 43 },
		{ 40, 363, 43 },
		{ 40, 364, 43 },
		{ 40, 365, 43 },
		{ 41, 312, 43 },
		{ 41, 313, 43 },
		{ 41, 314, 43 },
		{ 41, 315, 43 },
		{ 41, 316, 43 },
		{ 41, 317, 43 },
		{ 41, 318, 43 },
		{ 41, 319, 43 },
		{ 41, 320, 43 },
		{ 41, 321, 43 },
		{ 41, 322, 43 },
		{ 41, 323, 43 },
		{ 41, 324, 43 },
		{ 41, 325, 43 },
		{ 41, 326, 43 },
		{ 41, 327, 43 },
		{ 41, 328, 43 },
		{ 41, 329, 43 },
		{ 41, 330, 43 },
		{ 41, 331, 43 },
		{ 41, 332, 43 },
		{ 41, 333, 43 },
		{ 41, 334, 43 },
		{ 41, 335, 43 },
		{ 41, 336, 43 },
		{ 41, 337, 43 },
		{ 41, 338, 43 },
		{ 41, 339, 43 },
		{ 41, 340, 43 },
		{ 41, 341, 43 },
		{ 41, 342, 43 },
		{ 41, 343, 43 },
		{ 41, 344, 43 },
		{ 41, 345, 43 },
		{ 41, 346, 43 },
		{ 41, 347, 43 },
		{ 41, 348, 43 },
		{ 41, 349, 43 },
		{ 41, 350, 43 },
		{ 41, 351, 43 },
		{ 41, 352, 43 },
		{ 41, 353, 43 },
		{ 41, 354, 43 },
		{ 41, 355, 43 },
		{ 41, 356, 43 },
		{ 41, 357, 43 },
		{ 41, 358, 43 },
		{ 41, 359, 43 },
		{ 41, 360, 43 },
		{ 41, 361, 43 },
		{ 41, 362, 43 },
		{ 41, 363, 43 },
		{ 41, 364, 43 },
		{ 42, 311, 43 },
		{ 42, 312, 43 },
		{ 42, 313, 43 },
		{ 42, 314, 43 },
		{ 42, 315, 43 },
		{ 42, 316, 43 },
		{ 42, 317, 43 },
		{ 42, 318, 43 },
		{ 42, 319, 43 },
		{ 42, 320, 43 },
		{ 42, 321, 43 },
		{ 42, 322, 43 },
		{ 42, 323, 43 },
		{ 42, 324, 43 },
		{ 42, 325, 43 },
		{ 42, 326, 43 },
		{ 42, 327, 43 },
		{ 42, 328, 43 },
		{ 42, 329, 43 },
		{ 42, 330, 43 },
		{ 42, 331, 43 },
		{ 42, 332, 43 },
		{ 42, 333, 43 },
		{ 42, 334, 43 },
		{ 42, 335, 43 },
		{ 42, 336, 43 },
		{ 42, 337, 43 },
		{ 42, 338, 43 },
		{ 42, 339, 43 },
		{ 42, 340, 43 },
		{ 42, 341, 43 },
		{ 42, 342, 43 },
		{ 42, 343, 43 },
		{ 42, 344, 43 },
		{ 42, 345, 43 },
		{ 42, 346, 43 },
		{ 42, 347, 43 },
		{ 42, 348, 43 },
		{ 42, 349, 43 },
		{ 42, 350, 43 },
		{ 42, 351, 43 },
		{ 42, 352, 43 },
		{ 42, 353, 43 },
		{ 42, 354, 43 },
		{ 42, 355, 43 },
		{ 42, 356, 43 },
		{ 42, 357, 43 },
		{ 42, 358, 43 },
		{ 42, 359, 43 },
		{ 42, 360, 43 },
		{ 42, 361, 43 },
		{ 42, 362, 43 },
		{ 42, 363, 43 },
		{ 42, 364, 43 },
		{ 43, 311, 43 },
		{ 43, 312, 43 },
		{ 43, 313, 43 },
		{ 43, 314, 43 },
		{ 43, 315, 43 },
		{ 43, 316, 43 },
		{ 43, 317, 43 },
		{ 43, 318, 43 },
		{ 43, 319, 43 },
		{ 43, 320, 43 },
		{ 43, 321, 43 },
		{ 43, 322, 43 },
		{ 43, 323, 43 },
		{ 43, 324, 43 },
		{ 43, 325, 43 },
		{ 43, 326, 43 },
		{ 43, 327, 43 },
		{ 43, 328, 43 },
		{ 43, 329, 43 },
		{ 43, 330, 43 },
		{ 43, 331, 43 },
		{ 43, 332, 43 },
		{ 43, 333, 43 },
		{ 43, 334, 43 },
		{ 43, 335, 43 },
		{ 43, 336, 43 },
		{ 43, 337, 43 },
		{ 43, 338, 43 },
		{ 43, 339, 43 },
		{ 43, 340, 43 },
		{ 43, 341, 43 },
		{ 43, 342, 43 },
		{ 43, 343, 43 },
		{ 43, 344, 43 },
		{ 43, 345, 43 },
		{ 43, 346, 43 },
		{ 43, 347, 43 },
		{ 43, 348, 43 },
		{ 43, 349, 43 },
		{ 43, 350, 43 },
		{ 43, 351, 43 },
		{ 43, 352, 43 },
		{ 43, 353, 43 },
		{ 43, 354, 43 },
		{ 43, 355, 43 },
		{ 43, 356, 43 },
		{ 43, 357, 43 },
		{ 43, 358, 43 },
		{ 43, 359, 43 },
		{ 43, 360, 43 },
		{ 43, 361, 43 },
		{ 43, 362, 43 },
		{ 43, 363, 43 },
		{ 44, 311, 43 },
		{ 44, 312, 43 },
		{ 44, 313, 43 },
		{ 44, 314, 43 },
		{ 44, 315, 43 },
		{ 44, 316, 43 },
		{ 44, 317, 43 },
		{ 44, 318, 43 },
		{ 44, 319, 43 },
		{ 44, 320, 43 },
		{ 44, 321, 43 },
		{ 44, 322, 43 },
		{ 44, 323, 43 },
		{ 44, 324, 43 },
		{ 44, 325, 43 },
		{ 44, 326, 43 },
		{ 44, 327, 43 },
		{ 44, 328, 43 },
		{ 44, 329, 43 },
		{ 44, 330, 43 },
		{ 44, 331, 43 },
		{ 44, 332, 43 },
		{ 44, 333, 43 },
		{ 44, 334, 43 },
		{ 44, 335, 43 },
		{ 44, 336, 43 },
		{ 44, 337, 43 },
		{ 44, 338, 43 },
		{ 44, 339, 43 },
		{ 44, 340, 43 },
		{ 44, 341, 43 },
		{ 44, 342, 43 },
		{ 44, 343, 43 },
		{ 44, 344, 43 },
		{ 44, 345, 43 },
		{ 44, 346, 43 },
		{ 44, 347, 43 },
		{ 44, 348, 43 },
		{ 44, 349, 43 },
		{ 44, 350, 43 },
		{ 44, 351, 43 },
		{ 44, 352, 43 },
		{ 44, 353, 43 },
		{ 44, 354, 43 },
		{ 44, 355, 43 },
		{ 44, 356, 43 },
		{ 44, 357, 43 },
		{ 44, 358, 43 },
		{ 44, 359, 43 },
		{ 44, 360, 43 },
		{ 44, 361, 43 },
		{ 44, 362, 43 },
		{ 44, 363, 43 },
		{ 45, 311, 43 },
		{ 45, 312, 43 },
		{ 45, 313, 43 },
		{ 45, 314, 43 },
		{ 45, 315, 43 },
		{ 45, 316, 43 },
		{ 45, 317, 43 },
		{ 45, 318, 43 },
		{ 45, 319, 43 },
		{ 45, 320, 43 },
		{ 45, 321, 43 },
		{ 45, 322, 43 },
		{ 45, 323, 43 },
		{ 45, 324, 43 },
		{ 45, 325, 43 },
		{ 45, 326, 43 },
		{ 45, 327, 43 },
		{ 45, 328, 43 },
		{ 45, 329, 43 },
		{ 45, 330, 43 },
		{ 45, 331, 43 },
		{ 45, 332, 43 },
		{ 45, 333, 43 },
		{ 45, 334, 43 },
		{ 45, 335, 43 },
		{ 45, 336, 43 },
		{ 45, 337, 43 },
		{ 45, 338, 43 },
		{ 45, 339, 43 },
		{ 45, 340, 43 },
		{ 45, 341, 43 },
		{ 45, 342, 43 },
		{ 45, 343, 43 },
		{ 45, 344, 43 },
		{ 45, 345, 43 },
		{ 45, 346, 43 },
		{ 45, 347, 43 },
		{ 45, 348, 43 },
		{ 45, 349, 43 },
		{ 45, 350, 43 },
		{ 45, 351, 43 },
		{ 45, 352, 43 },
		{ 45, 353, 43 },
		{ 45, 354, 43 },
		{ 45, 355, 43 },
		{ 45, 356, 43 },
		{ 45, 357, 43 },
		{ 45, 358, 43 },
		{ 45, 359, 43 },
		{ 45, 360, 43 },
		{ 45, 361, 43 },
		{ 45, 362, 43 },
		{ 46, 310, 43 },
		{ 46, 311, 43 },
		{ 46, 312, 43 },
		{ 46, 313, 43 },
		{ 46, 314, 43 },
		{ 46, 315, 43 },
		{ 46, 316, 43 },
		{ 46, 317, 43 },
		{ 46, 318, 43 },
		{ 46, 319, 43 },
		{ 46, 320, 43 },
		{ 46, 321, 43 },
		{ 46, 322, 43 },
		{ 46, 323, 43 },
		{ 46, 324, 43 },
		{ 46, 325, 43 },
		{ 46, 326, 43 },
		{ 46, 327, 43 },
		{ 46, 328, 43 },
		{ 46, 329, 43 },
		{ 46, 330, 43 },
		{ 46, 331, 43 },
		{ 46, 332, 43 },
		{ 46, 333, 43 },
		{ 46, 334, 43 },
		{ 46, 335, 43 },
		{ 46, 336, 43 },
		{ 46, 337, 43 },
		{ 46, 338, 43 },
		{ 46, 339, 43 },
		{ 46, 340, 43 },
		{ 46, 341, 43 },
		{ 46, 342, 43 },
		{ 46, 343, 43 },
		{ 46, 344, 43 },
		{ 46, 345, 43 },
		{ 46, 346, 43 },
		{ 46, 347, 43 },
		{ 46, 348, 43 },
		{ 46, 349, 43 },
		{ 46, 350, 43 },
		{ 46, 351, 43 },
		{ 46, 352, 43 },
		{ 46, 353, 43 },
		{ 46, 354, 43 },
		{ 46, 355, 43 },
		{ 46, 356, 43 },
		{ 46, 357, 43 },
		{ 46, 358, 43 },
		{ 46, 359, 43 },
		{ 46, 360, 43 },
		{ 46, 361, 43 },
		{ 46, 362, 43 },
		{ 47, 310, 43 },
		{ 47, 311, 43 },
		{ 47, 312, 43 },
		{ 47, 313, 43 },
		{ 47, 314, 43 },
		{ 47, 315, 43 },
		{ 47, 316, 43 },
		{ 47, 317, 43 },
		{ 47, 318, 43 },
		{ 47, 319, 43 },
		{ 47, 320, 43 },
		{ 47, 321, 43 },
		{ 47, 322, 43 },
		{ 47, 323, 43 },
		{ 47, 324, 43 },
		{ 47, 325, 43 },
		{ 47, 326, 43 },
		{ 47, 327, 43 },
		{ 47, 328, 43 },
		{ 47, 329, 43 },
		{ 47, 330, 43 },
		{ 47, 331, 43 },
		{ 47, 332, 43 },
		{ 47, 333, 43 },
		{ 47, 334, 43 },
		{ 47, 335, 43 },
		{ 47, 336, 43 },
		{ 47, 337, 43 },
		{ 47, 338, 43 },
		{ 47, 339, 43 },
		{ 47, 340, 43 },
		{ 47, 341, 43 },
		{ 47, 342, 43 },
		{ 47, 343, 43 },
		{ 47, 344, 43 },
		{ 47, 345, 43 },
		{ 47, 346, 43 },
		{ 47, 347, 43 },
		{ 47, 348, 43 },
		{ 47, 349, 43 },
		{ 47, 350, 43 },
		{ 47, 351, 43 },
		{ 47, 352, 43 },
		{ 47, 353, 43 },
		{ 47, 354, 43 },
		{ 47, 355, 43 },
		{ 47, 356, 43 },
		{ 47, 357, 43 },
		{ 47, 358, 43 },
		{ 47, 359, 43 },
		{ 47, 360, 43 },
		{ 47, 361, 43 },
		{ 48, 310, 43 },
		{ 48, 311, 43 },
		{ 48, 312, 43 },
		{ 48, 313, 43 },
		{ 48, 314, 43 },
		{ 48, 315, 43 },
		{ 48, 316, 43 },
		{ 48, 317, 43 },
		{ 48, 318, 43 },
		{ 48, 319, 43 },
		{ 48, 320, 43 },
		{ 48, 321, 43 },
		{ 48, 322, 43 },
		{ 48, 323, 43 },
		{ 48, 324, 43 },
		{ 48, 325, 43 },
		{ 48, 326, 43 },
		{ 48, 327, 43 },
		{ 48, 328, 43 },
		{ 48, 329, 43 },
		{ 48, 330, 43 },
		{ 48, 331, 43 },
		{ 48, 332, 43 },
		{ 48, 333, 43 },
		{ 48, 334, 43 },
		{ 48, 335, 43 },
		{ 48, 336, 43 },
		{ 48, 337, 43 },
		{ 48, 338, 43 },
		{ 48, 339, 43 },
		{ 48, 340, 43 },
		{ 48, 341, 43 },
		{ 48, 342, 43 },
		{ 48, 343, 43 },
		{ 48, 344, 43 },
		{ 48, 345, 43 },
		{ 48, 346, 43 },
		{ 48, 347, 43 },
		{ 48, 348, 43 },
		{ 48, 349, 43 },
		{ 48, 350, 43 },
		{ 48, 351, 43 },
		{ 48, 352, 43 },
		{ 48, 353, 43 },
		{ 48, 354, 43 },
		{ 48, 355, 43 },
		{ 48, 356, 43 },
		{ 48, 357, 43 },
		{ 48, 358, 43 },
		{ 48, 359, 43 },
		{ 48, 360, 43 },
		{ 48, 361, 43 },
		{ 49, 310, 43 },
		{ 49, 311, 43 },
		{ 49, 312, 43 },
		{ 49, 313, 43 },
		{ 49, 314, 43 },
		{ 49, 315, 43 },
		{ 49, 316, 43 },
		{ 49, 317, 43 },
		{ 49, 318, 43 },
		{ 49, 319, 43 },
		{ 49, 320, 43 },
		{ 49, 321, 43 },
		{ 49, 322, 43 },
		{ 49, 323, 43 },
		{ 49, 324, 43 },
		{ 49, 325, 43 },
		{ 49, 326, 43 },
		{ 49, 327, 43 },
		{ 49, 328, 43 },
		{ 49, 329, 43 },
		{ 49, 330, 43 },
		{ 49, 331, 43 },
		{ 49, 332, 43 },
		{ 49, 333, 43 },
		{ 49, 334, 43 },
		{ 49, 335, 43 },
		{ 49, 336, 43 },
		{ 49, 337, 43 },
		{ 49, 338, 43 },
		{ 49, 339, 43 },
		{ 49, 340, 43 },
		{ 49, 341, 43 },
		{ 49, 342, 43 },
		{ 49, 343, 43 },
		{ 49, 344, 43 },
		{ 49, 345, 43 },
		{ 49, 346, 43 },
		{ 49, 347, 43 },
		{ 49, 348, 43 },
		{ 49, 349, 43 },
		{ 49, 350, 43 },
		{ 49, 351, 43 },
		{ 49, 352, 43 },
		{ 49, 353, 43 },
		{ 49, 354, 43 },
		{ 49, 355, 43 },
		{ 49, 356, 43 },
		{ 49, 357, 43 },
		{ 49, 358, 43 },
		{ 49, 359, 43 },
		{ 49, 360, 43 },
		{ 50, 309, 43 },
		{ 50, 310, 43 },
		{ 50, 311, 43 },
		{ 50, 312, 43 },
		{ 50, 313, 43 },
		{ 50, 314, 43 },
		{ 50, 315, 43 },
		{ 50, 316, 43 },
		{ 50, 317, 43 },
		{ 50, 318, 43 },
		{ 50, 319, 43 },
		{ 50, 320, 43 },
		{ 50, 321, 43 },
		{ 50, 322, 43 },
		{ 50, 323, 43 },
		{ 50, 324, 43 },
		{ 50, 325, 43 },
		{ 50, 326, 43 },
		{ 50, 327, 43 },
		{ 50, 328, 43 },
		{ 50, 329, 43 },
		{ 50, 330, 43 },
		{ 50, 331, 43 },
		{ 50, 332, 43 },
		{ 50, 333, 43 },
		{ 50, 334, 43 },
		{ 50, 335, 43 },
		{ 50, 336, 43 },
		{ 50, 337, 43 },
		{ 50, 338, 43 },
		{ 50, 339, 43 },
		{ 50, 340, 43 },
		{ 50, 341, 43 },
		{ 50, 342, 43 },
		{ 50, 343, 43 },
		{ 50, 344, 43 },
		{ 50, 345, 43 },
		{ 50, 346, 43 },
		{ 50, 347, 43 },
		{ 50, 348, 43 },
		{ 50, 349, 43 },
		{ 50, 350, 43 },
		{ 50, 351, 43 },
		{ 50, 352, 43 },
		{ 50, 353, 43 },
		{ 50, 354, 43 },
		{ 50, 355, 43 },
		{ 50, 356, 43 },
		{ 50, 357, 43 },
		{ 50, 358, 43 },
		{ 50, 359, 43 },
		{ 50, 360, 43 },
		{ 51, 309, 43 },
		{ 51, 310, 43 },
		{ 51, 311, 43 },
		{ 51, 312, 43 },
		{ 51, 313, 43 },
		{ 51, 314, 43 },
		{ 51, 315, 43 },
		{ 51, 316, 43 },
		{ 51, 317, 43 },
		{ 51, 318, 43 },
		{ 51, 319, 43 },
		{ 51, 320, 43 },
		{ 51, 321, 43 },
		{ 51, 322, 43 },
		{ 51, 323, 43 },
		{ 51, 324, 43 },
		{ 51, 325, 43 },
		{ 51, 326, 43 },
		{ 51, 327, 43 },
		{ 51, 328, 43 },
		{ 51, 329, 43 },
		{ 51, 330, 43 },
		{ 51, 331, 43 },
		{ 51, 332, 43 },
		{ 51, 333, 43 },
		{ 51, 334, 43 },
		{ 51, 335, 43 },
		{ 51, 336, 43 },
		{ 51, 337, 43 },
		{ 51, 338, 43 },
		{ 51, 339, 43 },
		{ 51, 340, 43 },
		{ 51, 341, 43 },
		{ 51, 342, 43 },
		{ 51, 343, 43 },
		{ 51, 344, 43 },
		{ 51, 345, 43 },
		{ 51, 346, 43 },
		{ 51, 347, 43 },
		{ 51, 348, 43 },
		{ 51, 349, 43 },
		{ 51, 350, 43 },
		{ 51, 351, 43 },
		{ 51, 352, 43 },
		{ 51, 353, 43 },
		{ 51, 354, 43 },
		{ 51, 355, 43 },
		{ 51, 356, 43 },
		{ 51, 357, 43 },
		{ 51, 358, 43 },
		{ 51, 359, 43 },
		{ 52, 309, 43 },
		{ 52, 310, 43 },
		{ 52, 311, 43 },
		{ 52, 312, 43 },
		{ 52, 313, 43 },
		{ 52, 314, 43 },
		{ 52, 315, 43 },
		{ 52, 316, 43 },
		{ 52, 317, 43 },
		{ 52, 318, 43 },
		{ 52, 319, 43 },
		{ 52, 320, 43 },
		{ 52, 321, 43 },
		{ 52, 322, 43 },
		{ 52, 323, 43 },
		{ 52, 324, 43 },
		{ 52, 325, 43 },
		{ 52, 326, 43 },
		{ 52, 327, 43 },
		{ 52, 328, 43 },
		{ 52, 329, 43 },
		{ 52, 330, 43 },
		{ 52, 331, 43 },
		{ 52, 332, 43 },
		{ 52, 333, 43 },
		{ 52, 334, 43 },
		{ 52, 335, 43 },
		{ 52, 336, 43 },
		{ 52, 337, 43 },
		{ 52, 338, 43 },
		{ 52, 339, 43 },
		{ 52, 340, 43 },
		{ 52, 341, 43 },
		{ 52, 342, 43 },
		{ 52, 343, 43 },
		{ 52, 344, 43 },
		{ 52, 345, 43 },
		{ 52, 346, 43 },
		{ 52, 347, 43 },
		{ 52, 348, 43 },
		{ 52, 349, 43 },
		{ 52, 350, 43 },
		{ 52, 351, 43 },
		{ 52, 352, 43 },
		{ 52, 353, 43 },
		{ 52, 354, 43 },
		{ 52, 355, 43 },
		{ 52, 356, 43 },
		{ 52, 357, 43 },
		{ 52, 358, 43 },
		{ 52, 359, 43 },
		{ 53, 309, 43 },
		{ 53, 310, 43 },
		{ 53, 311, 43 },
		{ 53, 312, 43 },
		{ 53, 313, 43 },
		{ 53, 314, 43 },
		{ 53, 315, 43 },
		{ 53, 316, 43 },
		{ 53, 317, 43 },
		{ 53, 318, 43 },
		{ 53, 319, 43 },
		{ 53, 320, 43 },
		{ 53, 321, 43 },
		{ 53, 322, 43 },
		{ 53, 323, 43 },
		{ 53, 324, 43 },
		{ 53, 325, 43 },
		{ 53, 326, 43 },
		{ 53, 327, 43 },
		{ 53, 328, 43 },
		{ 53, 329, 43 },
		{ 53, 330, 43 },
		{ 53, 331, 43 },
		{ 53, 332, 43 },
		{ 53, 333, 43 },
		{ 53, 334, 43 },
		{ 53, 335, 43 },
		{ 53, 336, 43 },
		{ 53, 337, 43 },
		{ 53, 338, 43 },
		{ 53, 339, 43 },
		{ 53, 340, 43 },
		{ 53, 341, 43 },
		{ 53, 342, 43 },
		{ 53, 343, 43 },
		{ 53, 344, 43 },
		{ 53, 345, 43 },
		{ 53, 346, 43 },
		{ 53, 347, 43 },
		{ 53, 348, 43 },
		{ 53, 349, 43 },
		{ 53, 350, 43 },
		{ 53, 351, 43 },
		{ 53, 352, 43 },
		{ 53, 353, 43 },
		{ 53, 354, 43 },
		{ 53, 355, 43 },
		{ 53, 356, 43 },
		{ 53, 357, 43 },
		{ 53, 358, 43 },
		{ 54, 308, 43 },
		{ 54, 309, 43 },
		{ 54, 310, 43 },
		{ 54, 311, 43 },
		{ 54, 312, 43 },
		{ 54, 313, 43 },
		{ 54, 314, 43 },
		{ 54, 315, 43 },
		{ 54, 316, 43 },
		{ 54, 317, 43 },
		{ 54, 318, 43 },
		{ 54, 319, 43 },
		{ 54, 320, 43 },
		{ 54, 321, 43 },
		{ 54, 322, 43 },
		{ 54, 323, 43 },
		{ 54, 324, 43 },
		{ 54, 325, 43 },
		{ 54, 326, 43 },
		{ 54, 327, 43 },
		{ 54, 328, 43 },
		{ 54, 329, 43 },
		{ 54, 330, 43 },
		{ 54, 331, 43 },
		{ 54, 332, 43 },
		{ 54, 333, 43 },
		{ 54, 334, 43 },
		{ 54, 335, 43 },
		{ 54, 336, 43 },
		{ 54, 337, 43 },
		{ 54, 338, 43 },
		{ 54, 339, 43 },
		{ 54, 340, 43 },
		{ 54, 341, 43 },
		{ 54, 342, 43 },
		{ 54, 343, 43 },
		{ 54, 344, 43 },
		{ 54, 345, 43 },
		{ 54, 346, 43 },
		{ 54, 347, 43 },
		{ 54, 348, 43 },
		{ 54, 349, 43 },
		{ 54, 350, 43 },
		{ 54, 351, 43 },
		{ 54, 352, 43 },
		{ 54, 353, 43 },
		{ 54, 354, 43 },
		{ 54, 355, 43 },
		{ 54, 356, 43 },
		{ 54, 357, 43 },
		{ 54, 358, 43 },
		{ 55, 308, 43 },
		{ 55, 309, 43 },
		{ 55, 310, 43 },
		{ 55, 311, 43 },
		{ 55, 312, 43 },
		{ 55, 313, 43 },
		{ 55, 314, 43 },
		{ 55, 315, 43 },
		{ 55, 316, 43 },
		{ 55, 317, 43 },
		{ 55, 318, 43 },
		{ 55, 319, 43 },
		{ 55, 320, 43 },
		{ 55, 321, 43 },
		{ 55, 322, 43 },
		{ 55, 323, 43 },
		{ 55, 324, 43 },
		{ 55, 325, 43 },
		{ 55, 326, 43 },
		{ 55, 327, 43 },
		{ 55, 328, 43 },
		{ 55, 329, 43 },
		{ 55, 330, 43 },
		{ 55, 331, 43 },
		{ 55, 332, 43 },
		{ 55, 333, 43 },
		{ 55, 334, 43 },
		{ 55, 335, 43 },
		{ 55, 336, 43 },
		{ 55, 337, 43 },
		{ 55, 338, 43 },
		{ 55, 339, 43 },
		{ 55, 340, 43 },
		{ 55, 341, 43 },
		{ 55, 342, 43 },
		{ 55, 343, 43 },
		{ 55, 344, 43 },
		{ 55, 345, 43 },
		{ 55, 346, 43 },
		{ 55, 347, 43 },
		{ 55, 348, 43 },
		{ 55, 349, 43 },
		{ 55, 350, 43 },
		{ 55, 351, 43 },
		{ 55, 352, 43 },
		{ 55, 353, 43 },
		{ 55, 354, 43 },
		{ 55, 355, 43 },
		{ 55, 356, 43 },
		{ 55, 357, 43 },
		{ 56, 308, 43 },
		{ 56, 309, 43 },
		{ 56, 310, 43 },
		{ 56, 311, 43 },
		{ 56, 312, 43 },
		{ 56, 313, 43 },
		{ 56, 314, 43 },
		{ 56, 315, 43 },
		{ 56, 316, 43 },
		{ 56, 317, 43 },
		{ 56, 318, 43 },
		{ 56, 319, 43 },
		{ 56, 320, 43 },
		{ 56, 321, 43 },
		{ 56, 322, 43 },
		{ 56, 323, 43 },
		{ 56, 324, 43 },
		{ 56, 325, 43 },
		{ 56, 326, 43 },
		{ 56, 327, 43 },
		{ 56, 328, 43 },
		{ 56, 329, 43 },
		{ 56, 330, 43 },
		{ 56, 331, 43 },
		{ 56, 332, 43 },
		{ 56, 333, 43 },
		{ 56, 334, 43 },
		{ 56, 335, 43 },
		{ 56, 336, 43 },
		{ 56, 337, 43 },
		{ 56, 338, 43 },
		{ 56, 339, 43 },
		{ 56, 340, 43 },
		{ 56, 341, 43 },
		{ 56, 342, 43 },
		{ 56, 343, 43 },
		{ 56, 344, 43 },
		{ 56, 345, 43 },
		{ 56, 346, 43 },
		{ 56, 347, 43 },
		{ 56, 348, 43 },
		{ 56, 349, 43 },
		{ 56, 350, 43 },
		{ 56, 351, 43 },
		{ 56, 352, 43 },
		{ 56, 353, 43 },
		{ 56, 354, 43 },
		{ 56, 355, 43 },
		{ 56, 356, 43 },
		{ 56, 357, 43 },
		{ 57, 308, 43 },
		{ 57, 309, 43 },
		{ 57, 310, 43 },
		{ 57, 311, 43 },
		{ 57, 312, 43 },
		{ 57, 313, 43 },
		{ 57, 314, 43 },
		{ 57, 315, 43 },
		{ 57, 316, 43 },
		{ 57, 317, 43 },
		{ 57, 318, 43 },
		{ 57, 319, 43 },
		{ 57, 320, 43 },
		{ 57, 321, 43 },
		{ 57, 322, 43 },
		{ 57, 323, 43 },
		{ 57, 324, 43 },
		{ 57, 325, 43 },
		{ 57, 326, 43 },
		{ 57, 327, 43 },
		{ 57, 328, 43 },
		{ 57, 329, 43 },
		{ 57, 330, 43 },
		{ 57, 331, 43 },
		{ 57, 332, 43 },
		{ 57, 333, 43 },
		{ 57, 334, 43 },
		{ 57, 335, 43 },
		{ 57, 336, 43 },
		{ 57, 337, 43 },
		{ 57, 338, 43 },
		{ 57, 339, 43 },
		{ 57, 340, 43 },
		{ 57, 341, 43 },
		{ 57, 342, 43 },
		{ 57, 343, 43 },
		{ 57, 344, 43 },
		{ 57, 345, 43 },
		{ 57, 346, 43 },
		{ 57, 347, 43 },
		{ 57, 348, 43 },
		{ 57, 349, 43 },
		{ 57, 350, 43 },
		{ 57, 351, 43 },
		{ 57, 352, 43 },
		{ 57, 353, 43 },
		{ 57, 354, 43 },
		{ 57, 355, 43 },
		{ 57, 356, 43 },
		{ 58, 307, 43 },
		{ 58, 308, 43 },
		{ 58, 309, 43 },
		{ 58, 310, 43 },
		{ 58, 311, 43 },
		{ 58, 312, 43 },
		{ 58, 313, 43 },
		{ 58, 314, 43 },
		{ 58, 315, 43 },
		{ 58, 316, 43 },
		{ 58, 317, 43 },
		{ 58, 318, 43 },
		{ 58, 319, 43 },
		{ 58, 320, 43 },
		{ 58, 321, 43 },
		{ 58, 322, 43 },
		{ 58, 323, 43 },
		{ 58, 324, 43 },
		{ 58, 325, 43 },
		{ 58, 326, 43 },
		{ 58, 327, 43 },
		{ 58, 328, 43 },
		{ 58, 329, 43 },
		{ 58, 330, 43 },
		{ 58, 331, 43 },
		{ 58, 332, 43 },
		{ 58, 333, 43 },
		{ 58, 334, 43 },
		{ 58, 335, 43 },
		{ 58, 336, 43 },
		{ 58, 337, 43 },
		{ 58, 338, 43 },
		{ 58, 339, 43 },
		{ 58, 340, 43 },
		{ 58, 341, 43 },
		{ 58, 342, 43 },
		{ 58, 343, 43 },
		{ 58, 344, 43 },
		{ 58, 345, 43 },
		{ 58, 346, 43 },
		{ 58, 347, 43 },
		{ 58, 348, 43 },
		{ 58, 349, 43 },
		{ 58, 350, 43 },
		{ 58, 351, 43 },
		{ 58, 352, 43 },
		{ 58, 353, 43 },
		{ 58, 354, 43 },
		{ 58, 355, 43 },
		{ 58, 356, 43 },
		{ 59, 307, 43 },
		{ 59, 308, 43 },
		{ 59, 309, 43 },
		{ 59, 310, 43 },
		{ 59, 311, 43 },
		{ 59, 312, 43 },
		{ 59, 313, 43 },
		{ 59, 314, 43 },
		{ 59, 315, 43 },
		{ 59, 316, 43 },
		{ 59, 317, 43 },
		{ 59, 318, 43 },
		{ 59, 319, 43 },
		{ 59, 320, 43 },
		{ 59, 321, 43 },
		{ 59, 322, 43 },
		{ 59, 323, 43 },
		{ 59, 324, 43 },
		{ 59, 325, 43 },
		{ 59, 326, 43 },
		{ 59, 327, 43 },
		{ 59, 328, 43 },
		{ 59, 329, 43 },
		{ 59, 330, 43 },
		{ 59, 331, 43 },
		{ 59, 332, 43 },
		{ 59, 333, 43 },
		{ 59, 334, 43 },
		{ 59, 335, 43 },
		{ 59, 336, 43 },
		{ 59, 337, 43 },
		{ 59, 338, 43 },
		{ 59, 339, 43 },
		{ 59, 340, 43 },
		{ 59, 341, 43 },
		{ 59, 342, 43 },
		{ 59, 343, 43 },
		{ 59, 344, 43 },
		{ 59, 345, 43 },
		{ 59, 346, 43 },
		{ 59, 347, 43 },
		{ 59, 348, 43 },
		{ 59, 349, 43 },
		{ 59, 350, 43 },
		{ 59, 351, 43 },
		{ 59, 352, 43 },
		{ 59, 353, 43 },
		{ 59, 354, 43 },
		{ 59, 355, 43 },
		{ 60, 307, 43 },
		{ 60, 308, 43 },
		{ 60, 309, 43 },
		{ 60, 310, 43 },
		{ 60, 311, 43 },
		{ 60, 312, 43 },
		{ 60, 313, 43 },
		{ 60, 314, 43 },
		{ 60, 315, 43 },
		{ 60, 316, 43 },
		{ 60, 317, 43 },
		{ 60, 318, 43 },
		{ 60, 319, 43 },
		{ 60, 320, 43 },
		{ 60, 321, 43 },
		{ 60, 322, 43 },
		{ 60, 323, 43 },
		{ 60, 324, 43 },
		{ 60, 325, 43 },
		{ 60, 326, 43 },
		{ 60, 327, 43 },
		{ 60, 328, 43 },
		{ 60, 329, 43 },
		{ 60, 330, 43 },
		{ 60, 331, 43 },
		{ 60, 332, 43 },
		{ 60, 333, 43 },
		{ 60, 334, 43 },
		{ 60, 335, 43 },
		{ 60, 336, 43 },
		{ 60, 337, 43 },
		{ 60, 338, 43 },
		{ 60, 339, 43 },
		{ 60, 340, 43 },
		{ 60, 341, 43 },
		{ 60, 342, 43 },
		{ 60, 343, 43 },
		{ 60, 344, 43 },
		{ 60, 345, 43 },
		{ 60, 346, 43 },
		{ 60, 347, 43 },
		{ 60, 348, 43 },
		{ 60, 349, 43 },
		{ 60, 350, 43 },
		{ 60, 351, 43 },
		{ 60, 352, 43 },
		{ 60, 353, 43 },
		{ 60, 354, 43 },
		{ 60, 355, 43 },
		{ 61, 307, 43 },
		{ 61, 308, 43 },
		{ 61, 309, 43 },
		{ 61, 310, 43 },
		{ 61, 311, 43 },
		{ 61, 312, 43 },
		{ 61, 313, 43 },
		{ 61, 314, 43 },
		{ 61, 315, 43 },
		{ 61, 316, 43 },
		{ 61, 317, 43 },
		{ 61, 318, 43 },
		{ 61, 319, 43 },
		{ 61, 320, 43 },
		{ 61, 321, 43 },
		{ 61, 322, 43 },
		{ 61, 323, 43 },
		{ 61, 324, 43 },
		{ 61, 325, 43 },
		{ 61, 326, 43 },
		{ 61, 327, 43 },
		{ 61, 328, 43 },
		{ 61, 329, 43 },
		{ 61, 330, 43 },
		{ 61, 331, 43 },
		{ 61, 332, 43 },
		{ 61, 333, 43 },
		{ 61, 334, 43 },
		{ 61, 335, 43 },
		{ 61, 336, 43 },
		{ 61, 337, 43 },
		{ 61, 338, 43 },
		{ 61, 339, 43 },
		{ 61, 340, 43 },
		{ 61, 341, 43 },
		{ 61, 342, 43 },
		{ 61, 343, 43 },
		{ 61, 344, 43 },
		{ 61, 345, 43 },
		{ 61, 346, 43 },
		{ 61, 347, 43 },
		{ 61, 348, 43 },
		{ 61, 349, 43 },
		{ 61, 350, 43 },
		{ 61, 351, 43 },
		{ 61, 352, 43 },
		{ 61, 353, 43 },
		{ 61, 354, 43 },
		{ 62, 306, 43 },
		{ 62, 307, 43 },
		{ 62, 308, 43 },
		{ 62, 309, 43 },
		{ 62, 310, 43 },
		{ 62, 311, 43 },
		{ 62, 312, 43 },
		{ 62, 313, 43 },
		{ 62, 314, 43 },
		{ 62, 315, 43 },
		{ 62, 316, 43 },
		{ 62, 317, 43 },
		{ 62, 318, 43 },
		{ 62, 319, 43 },
		{ 62, 320, 43 },
		{ 62, 321, 43 },
		{ 62, 322, 43 },
		{ 62, 323, 43 },
		{ 62, 324, 43 },
		{ 62, 325, 43 },
		{ 62, 326, 43 },
		{ 62, 327, 43 },
		{ 62, 328, 43 },
		{ 62, 329, 43 },
		{ 62, 330, 43 },
		{ 62, 331, 43 },
		{ 62, 332, 43 },
		{ 62, 333, 43 },
		{ 62, 334, 43 },
		{ 62, 335, 43 },
		{ 62, 336, 43 },
		{ 62, 337, 43 },
		{ 62, 338, 43 },
		{ 62, 339, 43 },
		{ 62, 340, 43 },
		{ 62, 341, 43 },
		{ 62, 342, 43 },
		{ 62, 343, 43 },
		{ 62, 344, 43 },
		{ 62, 345, 43 },
		{ 62, 346, 43 },
		{ 62, 347, 43 },
		{ 62, 348, 43 },
		{ 62, 349, 43 },
		{ 62, 350, 43 },
		{ 62, 351, 43 },
		{ 62, 352, 43 },
		{ 62, 353, 43 },
		{ 62, 354, 43 },
		{ 63, 306, 43 },
		{ 63, 307, 43 },
		{ 63, 308, 43 },
		{ 63, 309, 43 },
		{ 63, 310, 43 },
		{ 63, 311, 43 },
		{ 63, 312, 43 },
		{ 63, 313, 43 },
		{ 63, 314, 43 },
		{ 63, 315, 43 },
		{ 63, 316, 43 },
		{ 63, 317, 43 },
		{ 63, 318, 43 },
		{ 63, 319, 43 },
		{ 63, 320, 43 },
		{ 63, 321, 43 },
		{ 63, 322, 43 },
		{ 63, 323, 43 },
		{ 63, 324, 43 },
		{ 63, 325, 43 },
		{ 63, 326, 43 },
		{ 63, 327, 43 },
		{ 63, 328, 43 },
		{ 63, 329, 43 },
		{ 63, 330, 43 },
		{ 63, 331, 43 },
		{ 63, 332, 43 },
		{ 63, 333, 43 },
		{ 63, 334, 43 },
		{ 63, 335, 43 },
		{ 63, 336, 43 },
		{ 63, 337, 43 },
		{ 63, 338, 43 },
		{ 63, 339, 43 },
		{ 63, 340, 43 },
		{ 63, 341, 43 },
		{ 63, 342, 43 },
		{ 63, 343, 43 },
		{ 63, 344, 43 },
		{ 63, 345, 43 },
		{ 63, 346, 43 },
		{ 63, 347, 43 },
		{ 63, 348, 43 },
		{ 63, 349, 43 },
		{ 63, 350, 43 },
		{ 63, 351, 43 },
		{ 63, 352, 43 },
		{ 63, 353, 43 },
		{ 64, 306, 43 },
		{ 64, 307, 43 },
		{ 64, 308, 43 },
		{ 64, 309, 43 },
		{ 64, 310, 43 },
		{ 64, 311, 43 },
		{ 64, 312, 43 },
		{ 64, 313, 43 },
		{ 64, 314, 43 },
		{ 64, 315, 43 },
		{ 64, 316, 43 },
		{ 64, 317, 43 },
		{ 64, 318, 43 },
		{ 64, 319, 43 },
		{ 64, 320, 43 },
		{ 64, 321, 43 },
		{ 64, 322, 43 },
		{ 64, 323, 43 },
		{ 64, 324, 43 },
		{ 64, 325, 43 },
		{ 64, 326, 43 },
		{ 64, 327, 43 },
		{ 64, 328, 43 },
		{ 64, 329, 43 },
		{ 64, 330, 43 },
		{ 64, 331, 43 },
		{ 64, 332, 43 },
		{ 64, 333, 43 },
		{ 64, 334, 43 },
		{ 64, 335, 43 },
		{ 64, 336, 43 },
		{ 64, 337, 43 },
		{ 64, 338, 43 },
		{ 64, 339, 43 },
		{ 64, 340, 43 },
		{ 64, 341, 43 },
		{ 64, 342, 43 },
		{ 64, 343, 43 },
		{ 64, 344, 43 },
		{ 64, 345, 43 },
		{ 64, 346, 43 },
		{ 64, 347, 43 },
		{ 64, 348, 43 },
		{ 64, 349, 43 },
		{ 64, 350, 43 },
		{ 64, 351, 43 },
		{ 64, 352, 43 },
		{ 64, 353, 43 },
		{ 65, 306, 43 },
		{ 65, 307, 43 },
		{ 65, 308, 43 },
		{ 65, 309, 43 },
		{ 65, 310, 43 },
		{ 65, 311, 43 },
		{ 65, 312, 43 },
		{ 65, 313, 43 },
		{ 65, 314, 43 },
		{ 65, 315, 43 },
		{ 65, 316, 43 },
		{ 65, 317, 43 },
		{ 65, 318, 43 },
		{ 65, 319, 43 },
		{ 65, 320, 43 },
		{ 65, 321, 43 },
		{ 65, 322, 43 },
		{ 65, 323, 43 },
		{ 65, 324, 43 },
		{ 65, 325, 43 },
		{ 65, 326, 43 },
		{ 65, 327, 43 },
		{ 65, 328, 43 },
		{ 65, 329, 43 },
		{ 65, 330, 43 },
		{ 65, 331, 43 },
		{ 65, 332, 43 },
		{ 65, 333, 43 },
		{ 65, 334, 43 },
		{ 65, 335, 43 },
		{ 65, 336, 43 },
		{ 65, 337, 43 },
		{ 65, 338, 43 },
		{ 65, 339, 43 },
		{ 65, 340, 43 },
		{ 65, 341, 43 },
		{ 65, 342, 43 },
		{ 65, 343, 43 },
		{ 65, 344, 43 },
		{ 65, 345, 43 },
		{ 65, 346, 43 },
		{ 65, 347, 43 },
		{ 65, 348, 43 },
		{ 65, 349, 43 },
		{ 65, 350, 43 },
		{ 65, 351, 43 },
		{ 65, 352, 43 },
		{ 66, 305, 43 },
		{ 66, 306, 43 },
		{ 66, 307, 43 },
		{ 66, 308, 43 },
		{ 66, 309, 43 },
		{ 66, 310, 43 },
		{ 66, 311, 43 },
		{ 66, 312, 43 },
		{ 66, 313, 43 },
		{ 66, 314, 43 },
		{ 66, 315, 43 },
		{ 66, 316, 43 },
		{ 66, 317, 43 },
		{ 66, 318, 43 },
		{ 66, 319, 43 },
		{ 66, 320, 43 },
		{ 66, 321, 43 },
		{ 66, 322, 43 },
		{ 66, 323, 43 },
		{ 66, 324, 43 },
		{ 66, 325, 43 },
		{ 66, 326, 43 },
		{ 66, 327, 43 },
		{ 66, 328, 43 },
		{ 66, 329, 43 },
		{ 66, 330, 43 },
		{ 66, 331, 43 },
		{ 66, 332, 43 },
		{ 66, 333, 43 },
		{ 66, 334, 43 },
		{ 66, 335, 43 },
		{ 66, 336, 43 },
		{ 66, 337, 43 },
		{ 66, 338, 43 },
		{ 66, 339, 43 },
		{ 66, 340, 43 },
		{ 66, 341, 43 },
		{ 66, 342, 43 },
		{ 66, 343, 43 },
		{ 66, 344, 43 },
		{ 66, 345, 43 },
		{ 66, 346, 43 },
		{ 66, 347, 43 },
		{ 66, 348, 43 },
		{ 66, 349, 43 },
		{ 66, 350, 43 },
		{ 66, 351, 43 },
		{ 66, 352, 43 },
		{ 67, 305, 43 },
		{ 67, 306, 43 },
		{ 67, 307, 43 },
		{ 67, 308, 43 },
		{ 67, 309, 43 },
		{ 67, 310, 43 },
		{ 67, 311, 43 },
		{ 67, 312, 43 },
		{ 67, 313, 43 },
		{ 67, 314, 43 },
		{ 67, 315, 43 },
		{ 67, 316, 43 },
		{ 67, 317, 43 },
		{ 67, 318, 43 },
		{ 67, 319, 43 },
		{ 67, 320, 43 },
		{ 67, 321, 43 },
		{ 67, 322, 43 },
		{ 67, 323, 43 },
		{ 67, 324, 43 },
		{ 67, 325, 43 },
		{ 67, 326, 43 },
		{ 67, 327, 43 },
		{ 67, 328, 43 },
		{ 67, 329, 43 },
		{ 67, 330, 43 },
		{ 67, 331, 43 },
		{ 67, 332, 43 },
		{ 67, 333, 43 },
		{ 67, 334, 43 },
		{ 67, 335, 43 },
		{ 67, 336, 43 },
		{ 67, 337, 43 },
		{ 67, 338, 43 },
		{ 67, 339, 43 },
		{ 67, 340, 43 },
		{ 67, 341, 43 },
		{ 67, 342, 43 },
		{ 67, 343, 43 },
		{ 67, 344, 43 },
		{ 67, 345, 43 },
		{ 67, 346, 43 },
		{ 67, 347, 43 },
		{ 67, 348, 43 },
		{ 67, 349, 43 },
		{ 67, 350, 43 },
		{ 67, 351, 43 },
		{ 68, 305, 43 },
		{ 68, 306, 43 },
		{ 68, 307, 43 },
		{ 68, 308, 43 },
		{ 68, 309, 43 },
		{ 68, 310, 43 },
		{ 68, 311, 43 },
		{ 68, 312, 43 },
		{ 68, 313, 43 },
		{ 68, 314, 43 },
		{ 68, 315, 43 },
		{ 68, 316, 43 },
		{ 68, 317, 43 },
		{ 68, 318, 43 },
		{ 68, 319, 43 },
		{ 68, 320, 43 },
		{ 68, 321, 43 },
		{ 68, 322, 43 },
		{ 68, 323, 43 },
		{ 68, 324, 43 },
		{ 68, 325, 43 },
		{ 68, 326, 43 },
		{ 68, 327, 43 },
		{ 68, 328, 43 },
		{ 68, 329, 43 },
		{ 68, 330, 43 },
		{ 68, 331, 43 },
		{ 68, 332, 43 },
		{ 68, 333, 43 },
		{ 68, 334, 43 },
		{ 68, 335, 43 },
		{ 68, 336, 43 },
		{ 68, 337, 43 },
		{ 68, 338, 43 },
		{ 68, 339, 43 },
		{ 68, 340, 43 },
		{ 68, 341, 43 },
		{ 68, 342, 43 },
		{ 68, 343, 43 },
		{ 68, 344, 43 },
		{ 68, 345, 43 },
		{ 68, 346, 43 },
		{ 68, 347, 43 },
		{ 68, 348, 43 },
		{ 68, 349, 43 },
		{ 68, 350, 43 },
		{ 68, 351, 43 },
		{ 69, 305, 43 },
		{ 69, 306, 43 },
		{ 69, 307, 43 },
		{ 69, 308, 43 },
		{ 69, 309, 43 },
		{ 69, 310, 43 },
		{ 69, 311, 43 },
		{ 69, 312, 43 },
		{ 69, 313, 43 },
		{ 69, 314, 43 },
		{ 69, 315, 43 },
		{ 69, 316, 43 },
		{ 69, 317, 43 },
		{ 69, 318, 43 },
		{ 69, 319, 43 },
		{ 69, 320, 43 },
		{ 69, 321, 43 },
		{ 69, 322, 43 },
		{ 69, 323, 43 },
		{ 69, 324, 43 },
		{ 69, 325, 43 },
		{ 69, 326, 43 },
		{ 69, 327, 43 },
		{ 69, 328, 43 },
		{ 69, 329, 43 },
		{ 69, 330, 43 },
		{ 69, 331, 43 },
		{ 69, 332, 43 },
		{ 69, 333, 43 },
		{ 69, 334, 43 },
		{ 69, 335, 43 },
		{ 69, 336, 43 },
		{ 69, 337, 43 },
		{ 69, 338, 43 },
		{ 69, 339, 43 },
		{ 69, 340, 43 },
		{ 69, 341, 43 },
		{ 69, 342, 43 },
		{ 69, 343, 43 },
		{ 69, 344, 43 },
		{ 69, 345, 43 },
		{ 69, 346, 43 },
		{ 69, 347, 43 },
		{ 69, 348, 43 },
		{ 69, 349, 43 },
		{ 69, 350, 43 },
		{ 70, 304, 43 },
		{ 70, 305, 43 },
		{ 70, 306, 43 },
		{ 70, 307, 43 },
		{ 70, 308, 43 },
		{ 70, 309, 43 },
		{ 70, 310, 43 },
		{ 70, 311, 43 },
		{ 70, 312, 43 },
		{ 70, 313, 43 },
		{ 70, 314, 43 },
		{ 70, 315, 43 },
		{ 70, 316, 43 },
		{ 70, 317, 43 },
		{ 70, 318, 43 },
		{ 70, 319, 43 },
		{ 70, 320, 43 },
		{ 70, 321, 43 },
		{ 70, 322, 43 },
		{ 70, 323, 43 },
		{ 70, 324, 43 },
		{ 70, 325, 43 },
		{ 70, 326, 43 },
		{ 70, 327, 43 },
		{ 70, 328, 43 },
		{ 70, 329, 43 },
		{ 70, 330, 43 },
		{ 70, 331, 43 },
		{ 70, 332, 43 },
		{ 70, 333, 43 },
		{ 70, 334, 43 },
		{ 70, 335, 43 },
		{ 70, 336, 43 },
		{ 70, 337, 43 },
		{ 70, 338, 43 },
		{ 70, 339, 43 },
		{ 70, 340, 43 },
		{ 70, 341, 43 },
		{ 70, 342, 43 },
		{ 70, 343, 43 },
		{ 70, 344, 43 },
		{ 70, 345, 43 },
		{ 70, 346, 43 },
		{ 70, 347, 43 },
		{ 70, 348, 43 },
		{ 70, 349, 43 },
		{ 70, 350, 43 },
		{ 71, 304, 43 },
		{ 71, 305, 43 },
		{ 71, 306, 43 },
		{ 71, 307, 43 },
		{ 71, 308, 43 },
		{ 71, 309, 43 },
		{ 71, 310, 43 },
		{ 71, 311, 43 },
		{ 71, 312, 43 },
		{ 71, 313, 43 },
		{ 71, 314, 43 },
		{ 71, 315, 43 },
		{ 71, 316, 43 },
		{ 71, 317, 43 },
		{ 71, 318, 43 },
		{ 71, 319, 43 },
		{ 71, 320, 43 },
		{ 71, 321, 43 },
		{ 71, 322, 43 },
		{ 71, 323, 43 },
		{ 71, 324, 43 },
		{ 71, 325, 43 },
		{ 71, 326, 43 },
		{ 71, 327, 43 },
		{ 71, 328, 43 },
		{ 71, 329, 43 },
		{ 71, 330, 43 },
		{ 71, 331, 43 },
		{ 71, 332, 43 },
		{ 71, 333, 43 },
		{ 71, 334, 43 },
		{ 71, 335, 43 },
		{ 71, 336, 43 },
		{ 71, 337, 43 },
		{ 71, 338, 43 },
		{ 71, 339, 43 },
		{ 71, 340, 43 },
		{ 71, 341, 43 },
		{ 71, 342, 43 },
		{ 71, 343, 43 },
		{ 71, 344, 43 },
		{ 71, 345, 43 },
		{ 71, 346, 43 },
		{ 71, 347, 43 },
		{ 71, 348, 43 },
		{ 71, 349, 43 },
		{ 72, 304, 43 },
		{ 72, 305, 43 },
		{ 72, 306, 43 },
		{ 72, 307, 43 },
		{ 72, 308, 43 },
		{ 72, 309, 43 },
		{ 72, 310, 43 },
		{ 72, 311, 43 },
		{ 72, 312, 43 },
		{ 72, 313, 43 },
		{ 72, 314, 43 },
		{ 72, 315, 43 },
		{ 72, 316, 43 },
		{ 72, 317, 43 },
		{ 72, 318, 43 },
		{ 72, 319, 43 },
		{ 72, 320, 43 },
		{ 72, 321, 43 },
		{ 72, 322, 43 },
		{ 72, 323, 43 },
		{ 72, 324, 43 },
		{ 72, 325, 43 },
		{ 72, 326, 43 },
		{ 72, 327, 43 },
		{ 72, 328, 43 },
		{ 72, 329, 43 },
		{ 72, 330, 43 },
		{ 72, 331, 43 },
		{ 72, 332, 43 },
		{ 72, 333, 43 },
		{ 72, 334, 43 },
		{ 72, 335, 43 },
		{ 72, 336, 43 },
		{ 72, 337, 43 },
		{ 72, 338, 43 },
		{ 72, 339, 43 },
		{ 72, 340, 43 },
		{ 72, 341, 43 },
		{ 72, 342, 43 },
		{ 72, 343, 43 },
		{ 72, 344, 43 },
		{ 72, 345, 43 },
		{ 72, 346, 43 },
		{ 72, 347, 43 },
		{ 72, 348, 43 },
		{ 72, 349, 43 },
		{ 73, 304, 43 },
		{ 73, 305, 43 },
		{ 73, 306, 43 },
		{ 73, 307, 43 },
		{ 73, 308, 43 },
		{ 73, 309, 43 },
		{ 73, 310, 43 },
		{ 73, 311, 43 },
		{ 73, 312, 43 },
		{ 73, 313, 43 },
		{ 73, 314, 43 },
		{ 73, 315, 43 },
		{ 73, 316, 43 },
		{ 73, 317, 43 },
		{ 73, 318, 43 },
		{ 73, 319, 43 },
		{ 73, 320, 43 },
		{ 73, 321, 43 },
		{ 73, 322, 43 },
		{ 73, 323, 43 },
		{ 73, 324, 43 },
		{ 73, 325, 43 },
		{ 73, 326, 43 },
		{ 73, 327, 43 },
		{ 73, 328, 43 },
		{ 73, 329, 43 },
		{ 73, 330, 43 },
		{ 73, 331, 43 },
		{ 73, 332, 43 },
		{ 73, 333, 43 },
		{ 73, 334, 43 },
		{ 73, 335, 43 },
		{ 73, 336, 43 },
		{ 73, 337, 43 },
		{ 73, 338, 43 },
		{ 73, 339, 43 },
		{ 73, 340, 43 },
		{ 73, 341, 43 },
		{ 73, 342, 43 },
		{ 73, 343, 43 },
		{ 73, 344, 43 },
		{ 73, 345, 43 },
		{ 73, 346, 43 },
		{ 73, 347, 43 },
		{ 73, 348, 43 },
		{ 74, 303, 43 },
		{ 74, 304, 43 },
		{ 74, 305, 43 },
		{ 74, 306, 43 },
		{ 74, 307, 43 },
		{ 74, 308, 43 },
		{ 74, 309, 43 },
		{ 74, 310, 43 },
		{ 74, 311, 43 },
		{ 74, 312, 43 },
		{ 74, 313, 43 },
		{ 74, 314, 43 },
		{ 74, 315, 43 },
		{ 74, 316, 43 },
		{ 74, 317, 43 },
		{ 74, 318, 43 },
		{ 74, 319, 43 },
		{ 74, 320, 43 },
		{ 74, 321, 43 },
		{ 74, 322, 43 },
		{ 74, 323, 43 },
		{ 74, 324, 43 },
		{ 74, 325, 43 },
		{ 74, 326, 43 },
		{ 74, 327, 43 },
		{ 74, 328, 43 },
		{ 74, 329, 43 },
		{ 74, 330, 43 },
		{ 74, 331, 43 },
		{ 74, 332, 43 },
		{ 74, 333, 43 },
		{ 74, 334, 43 },
		{ 74, 335, 43 },
		{ 74, 336, 43 },
		{ 74, 337, 43 },
		{ 74, 338, 43 },
		{ 74, 339, 43 },
		{ 74, 340, 43 },
		{ 74, 341, 43 },
		{ 74, 342, 43 },
		{ 74, 343, 43 },
		{ 74, 344, 43 },
		{ 74, 345, 43 },
		{ 74, 346, 43 },
		{ 74, 347, 43 },
		{ 74, 348, 43 },
		{ 75, 303, 43 },
		{ 75, 304, 43 },
		{ 75, 305, 43 },
		{ 75, 306, 43 },
		{ 75, 307, 43 },
		{ 75, 308, 43 },
		{ 75, 309, 43 },
		{ 75, 310, 43 },
		{ 75, 311, 43 },
		{ 75, 312, 43 },
		{ 75, 313, 43 },
		{ 75, 314, 43 },
		{ 75, 315, 43 },
		{ 75, 316, 43 },
		{ 75, 317, 43 },
		{ 75, 318, 43 },
		{ 75, 319, 43 },
		{ 75, 320, 43 },
		{ 75, 321, 43 },
		{ 75, 322, 43 },
		{ 75, 323, 43 },
		{ 75, 324, 43 },
		{ 75, 325, 43 },
		{ 75, 326, 43 },
		{ 75, 327, 43 },
		{ 75, 328, 43 },
		{ 75, 329, 43 },
		{ 75, 330, 43 },
		{ 75, 331, 43 },
		{ 75, 332, 43 },
		{ 75, 333, 43 },
		{ 75, 334, 43 },
		{ 75, 335, 43 },
		{ 75, 336, 43 },
		{ 75, 337, 43 },
		{ 75, 338, 43 },
		{ 75, 339, 43 },
		{ 75, 340, 43 },
		{ 75, 341, 43 },
		{ 75, 342, 43 },
		{ 75, 343, 43 },
		{ 75, 344, 43 },
		{ 75, 345, 43 },
		{ 75, 346, 43 },
		{ 75, 347, 43 },
		{ 76, 303, 43 },
		{ 76, 304, 43 },
		{ 76, 305, 43 },
		{ 76, 306, 43 },
		{ 76, 307, 43 },
		{ 76, 308, 43 },
		{ 76, 309, 43 },
		{ 76, 310, 43 },
		{ 76, 311, 43 },
		{ 76, 312, 43 },
		{ 76, 313, 43 },
		{ 76, 314, 43 },
		{ 76, 315, 43 },
		{ 76, 316, 43 },
		{ 76, 317, 43 },
		{ 76, 318, 43 },
		{ 76, 319, 43 },
		{ 76, 320, 43 },
		{ 76, 321, 43 },
		{ 76, 322, 43 },
		{ 76, 323, 43 },
		{ 76, 324, 43 },
		{ 76, 325, 43 },
		{ 76, 326, 43 },
		{ 76, 327, 43 },
		{ 76, 328, 43 },
		{ 76, 329, 43 },
		{ 76, 330, 43 },
		{ 76, 331, 43 },
		{ 76, 332, 43 },
		{ 76, 333, 43 },
		{ 76, 334, 43 },
		{ 76, 335, 43 },
		{ 76, 336, 43 },
		{ 76, 337, 43 },
		{ 76, 338, 43 },
		{ 76, 339, 43 },
		{ 76, 340, 43 },
		{ 76, 341, 43 },
		{ 76, 342, 43 },
		{ 76, 343, 43 },
		{ 76, 344, 43 },
		{ 76, 345, 43 },
		{ 76, 346, 43 },
		{ 76, 347, 43 },
		{ 77, 303, 43 },
		{ 77, 304, 43 },
		{ 77, 305, 43 },
		{ 77, 306, 43 },
		{ 77, 307, 43 },
		{ 77, 308, 43 },
		{ 77, 309, 43 },
		{ 77, 310, 43 },
		{ 77, 311, 43 },
		{ 77, 312, 43 },
		{ 77, 313, 43 },
		{ 77, 314, 43 },
		{ 77, 315, 43 },
		{ 77, 316, 43 },
		{ 77, 317, 43 },
		{ 77, 318, 43 },
		{ 77, 319, 43 },
		{ 77, 320, 43 },
		{ 77, 321, 43 },
		{ 77, 322, 43 },
		{ 77, 323, 43 },
		{ 77, 324, 43 },
		{ 77, 325, 43 },
		{ 77, 326, 43 },
		{ 77, 327, 43 },
		{ 77, 328, 43 },
		{ 77, 329, 43 },
		{ 77, 330, 43 },
		{ 77, 331, 43 },
		{ 77, 332, 43 },
		{ 77, 333, 43 },
		{ 77, 334, 43 },
		{ 77, 335, 43 },
		{ 77, 336, 43 },
		{ 77, 337, 43 },
		{ 77, 338, 43 },
		{ 77, 339, 43 },
		{ 77, 340, 43 },
		{ 77, 341, 43 },
		{ 77, 342, 43 },
		{ 77, 343, 43 },
		{ 77, 344, 43 },
		{ 77, 345, 43 },
		{ 77, 346, 43 },
		{ 78, 302, 43 },
		{ 78, 303, 43 },
		{ 78, 304, 43 },
		{ 78, 305, 43 },
		{ 78, 306, 43 },
		{ 78, 307, 43 },
		{ 78, 308, 43 },
		{ 78, 309, 43 },
		{ 78, 310, 43 },
		{ 78, 311, 43 },
		{ 78, 312, 43 },
		{ 78, 313, 43 },
		{ 78, 314, 43 },
		{ 78, 315, 43 },
		{ 78, 316, 43 },
		{ 78, 317, 43 },
		{ 78, 318, 43 },
		{ 78, 319, 43 },
		{ 78, 320, 43 },
		{ 78, 321, 43 },
		{ 78, 322, 43 },
		{ 78, 323, 43 },
		{ 78, 324, 43 },
		{ 78, 325, 43 },
		{ 78, 326, 43 },
		{ 78, 327, 43 },
		{ 78, 328, 43 },
		{ 78, 329, 43 },
		{ 78, 330, 43 },
		{ 78, 331, 43 },
		{ 78, 332, 43 },
		{ 78, 333, 43 },
		{ 78, 334, 43 },
		{ 78, 335, 43 },
		{ 78, 336, 43 },
		{ 78, 337, 43 },
		{ 78, 338, 43 },
		{ 78, 339, 43 },
		{ 78, 340, 43 },
		{ 78, 341, 43 },
		{ 78, 342, 43 },
		{ 78, 343, 43 },
		{ 78, 344, 43 },
		{ 78, 345, 43 },
		{ 78, 346, 43 },
		{ 79, 302, 43 },
		{ 79, 303, 43 },
		{ 79, 304, 43 },
		{ 79, 305, 43 },
		{ 79, 306, 43 },
		{ 79, 307, 43 },
		{ 79, 308, 43 },
		{ 79, 309, 43 },
		{ 79, 310, 43 },
		{ 79, 311, 43 },
		{ 79, 312, 43 },
		{ 79, 313, 43 },
		{ 79, 314, 43 },
		{ 79, 315, 43 },
		{ 79, 316, 43 },
		{ 79, 317, 43 },
		{ 79, 318, 43 },
		{ 79, 319, 43 },
		{ 79, 320, 43 },
		{ 79, 321, 43 },
		{ 79, 322, 43 },
		{ 79, 323, 43 },
		{ 79, 324, 43 },
		{ 79, 325, 43 },
		{ 79, 326, 43 },
		{ 79, 327, 43 },
		{ 79, 328, 43 },
		{ 79, 329, 43 },
		{ 79, 330, 43 },
		{ 79, 331, 43 },
		{ 79, 332, 43 },
		{ 79, 333, 43 },
		{ 79, 334, 43 },
		{ 79, 335, 43 },
		{ 79, 336, 43 },
		{ 79, 337, 43 },
		{ 79, 338, 43 },
		{ 79, 339, 43 },
		{ 79, 340, 43 },
		{ 79, 341, 43 },
		{ 79, 342, 43 },
		{ 79, 343, 43 },
		{ 79, 344, 43 },
		{ 79, 345, 43 },
		{ 80, 302, 43 },
		{ 80, 303, 43 },
		{ 80, 304, 43 },
		{ 80, 305, 43 },
		{ 80, 306, 43 },
		{ 80, 307, 43 },
		{ 80, 308, 43 },
		{ 80, 309, 43 },
		{ 80, 310, 43 },
		{ 80, 311, 43 },
		{ 80, 312, 43 },
		{ 80, 313, 43 },
		{ 80, 314, 43 },
		{ 80, 315, 43 },
		{ 80, 316, 43 },
		{ 80, 317, 43 },
		{ 80, 318, 43 },
		{ 80, 319, 43 },
		{ 80, 320, 43 },
		{ 80, 321, 43 },
		{ 80, 322, 43 },
		{ 80, 323, 43 },
		{ 80, 324, 43 },
		{ 80, 325, 43 },
		{ 80, 326, 43 },
		{ 80, 327, 43 },
		{ 80, 328, 43 },
		{ 80, 329, 43 },
		{ 80, 330, 43 },
		{ 80, 331, 43 },
		{ 80, 332, 43 },
		{ 80, 333, 43 },
		{ 80, 334, 43 },
		{ 80, 335, 43 },
		{ 80, 336, 43 },
		{ 80, 337, 43 },
		{ 80, 338, 43 },
		{ 80, 339, 43 },
		{ 80, 340, 43 },
		{ 80, 341, 43 },
		{ 80, 342, 43 },
		{ 80, 343, 43 },
		{ 80, 344, 43 },
		{ 80, 345, 43 },
		{ 81, 302, 43 },
		{ 81, 303, 43 },
		{ 81, 304, 43 },
		{ 81, 305, 43 },
		{ 81, 306, 43 },
		{ 81, 307, 43 },
		{ 81, 308, 43 },
		{ 81, 309, 43 },
		{ 81, 310, 43 },
		{ 81, 311, 43 },
		{ 81, 312, 43 },
		{ 81, 313, 43 },
		{ 81, 314, 43 },
		{ 81, 315, 43 },
		{ 81, 316, 43 },
		{ 81, 317, 43 },
		{ 81, 318, 43 },
		{ 81, 319, 43 },
		{ 81, 320, 43 },
		{ 81, 321, 43 },
		{ 81, 322, 43 },
		{ 81, 323, 43 },
		{ 81, 324, 43 },
		{ 81, 325, 43 },
		{ 81, 326, 43 },
		{ 81, 327, 43 },
		{ 81, 328, 43 },
		{ 81, 329, 43 },
		{ 81, 330, 43 },
		{ 81, 331, 43 },
		{ 81, 332, 43 },
		{ 81, 333, 43 },
		{ 81, 334, 43 },
		{ 81, 335, 43 },
		{ 81, 336, 43 },
		{ 81, 337, 43 },
		{ 81, 338, 43 },
		{ 81, 339, 43 },
		{ 81, 340, 43 },
		{ 81, 341, 43 },
		{ 81, 342, 43 },
		{ 81, 343, 43 },
		{ 81, 344, 43 },
		{ 82, 301, 43 },
		{ 82, 302, 43 },
		{ 82, 303, 43 },
		{ 82, 304, 43 },
		{ 82, 305, 43 },
		{ 82, 306, 43 },
		{ 82, 307, 43 },
		{ 82, 308, 43 },
		{ 82, 309, 43 },
		{ 82, 310, 43 },
		{ 82, 311, 43 },
		{ 82, 312, 43 },
		{ 82, 313, 43 },
		{ 82, 314, 43 },
		{ 82, 315, 43 },
		{ 82, 316, 43 },
		{ 82, 317, 43 },
		{ 82, 318, 43 },
		{ 82, 319, 43 },
		{ 82, 320, 43 },
		{ 82, 321, 43 },
		{ 82, 322, 43 },
		{ 82, 323, 43 },
		{ 82, 324, 43 },
		{ 82, 325, 43 },
		{ 82, 326, 43 },
		{ 82, 327, 43 },
		{ 82, 328, 43 },
		{ 82, 329, 43 },
		{ 82, 330, 43 },
		{ 82, 331, 43 },
		{ 82, 332, 43 },
		{ 82, 333, 43 },
		{ 82, 334, 43 },
		{ 82, 335, 43 },
		{ 82, 336, 43 },
		{ 82, 337, 43 },
		{ 82, 338, 43 },
		{ 82, 339, 43 },
		{ 82, 340, 43 },
		{ 82, 341, 43 },
		{ 82, 342, 43 },
		{ 82, 343, 43 },
		{ 82, 344, 43 },
		{ 83, 301, 43 },
		{ 83, 302, 43 },
		{ 83, 303, 43 },
		{ 83, 304, 43 },
		{ 83, 305, 43 },
		{ 83, 306, 43 },
		{ 83, 307, 43 },
		{ 83, 308, 43 },
		{ 83, 309, 43 },
		{ 83, 310, 43 },
		{ 83, 311, 43 },
		{ 83, 312, 43 },
		{ 83, 313, 43 },
		{ 83, 314, 43 },
		{ 83, 315, 43 },
		{ 83, 316, 43 },
		{ 83, 317, 43 },
		{ 83, 318, 43 },
		{ 83, 319, 43 },
		{ 83, 320, 43 },
		{ 83, 321, 43 },
		{ 83, 322, 43 },
		{ 83, 323, 43 },
		{ 83, 324, 43 },
		{ 83, 325, 43 },
		{ 83, 326, 43 },
		{ 83, 327, 43 },
		{ 83, 328, 43 },
		{ 83, 329, 43 },
		{ 83, 330, 43 },
		{ 83, 331, 43 },
		{ 83, 332, 43 },
		{ 83, 333, 43 },
		{ 83, 334, 43 },
		{ 83, 335, 43 },
		{ 83, 336, 43 },
		{ 83, 337, 43 },
		{ 83, 338, 43 },
		{ 83, 339, 43 },
		{ 83, 340, 43 },
		{ 83, 341, 43 },
		{ 83, 342, 43 },
		{ 83, 343, 43 },
		{ 84, 301, 43 },
		{ 84, 302, 43 },
		{ 84, 303, 43 },
		{ 84, 304, 43 },
		{ 84, 305, 43 },
		{ 84, 306, 43 },
		{ 84, 307, 43 },
		{ 84, 308, 43 },
		{ 84, 309, 43 },
		{ 84, 310, 43 },
		{ 84, 311, 43 },
		{ 84, 312, 43 },
		{ 84, 313, 43 },
		{ 84, 314, 43 },
		{ 84, 315, 43 },
		{ 84, 316, 43 },
		{ 84, 317, 43 },
		{ 84, 318, 43 },
		{ 84, 319, 43 },
		{ 84, 320, 43 },
		{ 84, 321, 43 },
		{ 84, 322, 43 },
		{ 84, 323, 43 },
		{ 84, 324, 43 },
		{ 84, 325, 43 },
		{ 84, 326, 43 },
		{ 84, 327, 43 },
		{ 84, 328, 43 },
		{ 84, 329, 43 },
		{ 84, 330, 43 },
		{ 84, 331, 43 },
		{ 84, 332, 43 },
		{ 84, 333, 43 },
		{ 84, 334, 43 },
		{ 84, 335, 43 },
		{ 84, 336, 43 },
		{ 84, 337, 43 },
		{ 84, 338, 43 },
		{ 84, 339, 43 },
		{ 84, 340, 43 },
		{ 84, 341, 43 },
		{ 84, 342, 43 },
		{ 84, 343, 43 },
		{ 85, 301, 43 },
		{ 85, 302, 43 },
		{ 85, 303, 43 },
		{ 85, 304, 43 },
		{ 85, 305, 43 },
		{ 85, 306, 43 },
		{ 85, 307, 43 },
		{ 85, 308, 43 },
		{ 85, 309, 43 },
		{ 85, 310, 43 },
		{ 85, 311, 43 },
		{ 85, 312, 43 },
		{ 85, 313, 43 },
		{ 85, 314, 43 },
		{ 85, 315, 43 },
		{ 85, 316, 43 },
		{ 85, 317, 43 },
		{ 85, 318, 43 },
		{ 85, 319, 43 },
		{ 85, 320, 43 },
		{ 85, 321, 43 },
		{ 85, 322, 43 },
		{ 85, 323, 43 },
		{ 85, 324, 43 },
		{ 85, 325, 43 },
		{ 85, 326, 43 },
		{ 85, 327, 43 },
		{ 85, 328, 43 },
		{ 85, 329, 43 },
		{ 85, 330, 43 },
		{ 85, 331, 43 },
		{ 85, 332, 43 },
		{ 85, 333, 43 },
		{ 85, 334, 43 },
		{ 85, 335, 43 },
		{ 85, 336, 43 },
		{ 85, 337, 43 },
		{ 85, 338, 43 },
		{ 85, 339, 43 },
		{ 85, 340, 43 },
		{ 85, 341, 43 },
		{ 85, 342, 43 },
		{ 86, 300, 43 },
		{ 86, 301, 43 },
		{ 86, 302, 43 },
		{ 86, 303, 43 },
		{ 86, 304, 43 },
		{ 86, 305, 43 },
		{ 86, 306, 43 },
		{ 86, 307, 43 },
		{ 86, 308, 43 },
		{ 86, 309, 43 },
		{ 86, 310, 43 },
		{ 86, 311, 43 },
		{ 86, 312, 43 },
		{ 86, 313, 43 },
		{ 86, 314, 43 },
		{ 86, 315, 43 },
		{ 86, 316, 43 },
		{ 86, 317, 43 },
		{ 86, 318, 43 },
		{ 86, 319, 43 },
		{ 86, 320, 43 },
		{ 86, 321, 43 },
		{ 86, 322, 43 },
		{ 86, 323, 43 },
		{ 86, 324, 43 },
		{ 86, 325, 43 },
		{ 86, 326, 43 },
		{ 86, 327, 43 },
		{ 86, 328, 43 },
		{ 86, 329, 43 },
		{ 86, 330, 43 },
		{ 86, 331, 43 },
		{ 86, 332, 43 },
		{ 86, 333, 43 },
		{ 86, 334, 43 },
		{ 86, 335, 43 },
		{ 86, 336, 43 },
		{ 86, 337, 43 },
		{ 86, 338, 43 },
		{ 86, 339, 43 },
		{ 86, 340, 43 },
		{ 86, 341, 43 },
		{ 86, 342, 43 },
		{ 87, 300, 43 },
		{ 87, 301, 43 },
		{ 87, 302, 43 },
		{ 87, 303, 43 },
		{ 87, 304, 43 },
		{ 87, 305, 43 },
		{ 87, 306, 43 },
		{ 87, 307, 43 },
		{ 87, 308, 43 },
		{ 87, 309, 43 },
		{ 87, 310, 43 },
		{ 87, 311, 43 },
		{ 87, 312, 43 },
		{ 87, 313, 43 },
		{ 87, 314, 43 },
		{ 87, 315, 43 },
		{ 87, 316, 43 },
		{ 87, 317, 43 },
		{ 87, 318, 43 },
		{ 87, 319, 43 },
		{ 87, 320, 43 },
		{ 87, 321, 43 },
		{ 87, 322, 43 },
		{ 87, 323, 43 },
		{ 87, 324, 43 },
		{ 87, 325, 43 },
		{ 87, 326, 43 },
		{ 87, 327, 43 },
		{ 87, 328, 43 },
		{ 87, 329, 43 },
		{ 87, 330, 43 },
		{ 87, 331, 43 },
		{ 87, 332, 43 },
		{ 87, 333, 43 },
		{ 87, 334, 43 },
		{ 87, 335, 43 },
		{ 87, 336, 43 },
		{ 87, 337, 43 },
		{ 87, 338, 43 },
		{ 87, 339, 43 },
		{ 87, 340, 43 },
		{ 87, 341, 43 },
		{ 88, 300, 43 },
		{ 88, 301, 43 },
		{ 88, 302, 43 },
		{ 88, 303, 43 },
		{ 88, 304, 43 },
		{ 88, 305, 43 },
		{ 88, 306, 43 },
		{ 88, 307, 43 },
		{ 88, 308, 43 },
		{ 88, 309, 43 },
		{ 88, 310, 43 },
		{ 88, 311, 43 },
		{ 88, 312, 43 },
		{ 88, 313, 43 },
		{ 88, 314, 43 },
		{ 88, 315, 43 },
		{ 88, 316, 43 },
		{ 88, 317, 43 },
		{ 88, 318, 43 },
		{ 88, 319, 43 },
		{ 88, 320, 43 },
		{ 88, 321, 43 },
		{ 88, 322, 43 },
		{ 88, 323, 43 },
		{ 88, 324, 43 },
		{ 88, 325, 43 },
		{ 88, 326, 43 },
		{ 88, 327, 43 },
		{ 88, 328, 43 },
		{ 88, 329, 43 },
		{ 88, 330, 43 },
		{ 88, 331, 43 },
		{ 88, 332, 43 },
		{ 88, 333, 43 },
		{ 88, 334, 43 },
		{ 88, 335, 43 },
		{ 88, 336, 43 },
		{ 88, 337, 43 },
		{ 88, 338, 43 },
		{ 88, 339, 43 },
		{ 88, 340, 43 },
		{ 88, 341, 43 },
		{ 89, 300, 43 },
		{ 89, 301, 43 },
		{ 89, 302, 43 },
		{ 89, 303, 43 },
		{ 89, 304, 43 },
		{ 89, 305, 43 },
		{ 89, 306, 43 },
		{ 89, 307, 43 },
		{ 89, 308, 43 },
		{ 89, 309, 43 },
		{ 89, 310, 43 },
		{ 89, 311, 43 },
		{ 89, 312, 43 },
		{ 89, 313, 43 },
		{ 89, 314, 43 },
		{ 89, 315, 43 },
		{ 89, 316, 43 },
		{ 89, 317, 43 },
		{ 89, 318, 43 },
		{ 89, 319, 43 },
		{ 89, 320, 43 },
		{ 89, 321, 43 },
		{ 89, 322, 43 },
		{ 89, 323, 43 },
		{ 89, 324, 43 },
		{ 89, 325, 43 },
		{ 89, 326, 43 },
		{ 89, 327, 43 },
		{ 89, 328, 43 },
		{ 89, 329, 43 },
		{ 89, 330, 43 },
		{ 89, 331, 43 },
		{ 89, 332, 43 },
		{ 89, 333, 43 },
		{ 89, 334, 43 },
		{ 89, 335, 43 },
		{ 89, 336, 43 },
		{ 89, 337, 43 },
		{ 89, 338, 43 },
		{ 89, 339, 43 },
		{ 89, 340, 43 },
		{ 90, 299, 43 },
		{ 90, 300, 43 },
		{ 90, 301, 43 },
		{ 90, 302, 43 },
		{ 90, 303, 43 },
		{ 90, 304, 43 },
		{ 90, 305, 43 },
		{ 90, 306, 43 },
		{ 90, 307, 43 },
		{ 90, 308, 43 },
		{ 90, 309, 43 },
		{ 90, 310, 43 },
		{ 90, 311, 43 },
		{ 90, 312, 43 },
		{ 90, 313, 43 },
		{ 90, 314, 43 },
		{ 90, 315, 43 },
		{ 90, 316, 43 },
		{ 90, 317, 43 },
		{ 90, 318, 43 },
		{ 90, 319, 43 },
		{ 90, 320, 43 },
		{ 90, 321, 43 },
		{ 90, 322, 43 },
		{ 90, 323, 43 },
		{ 90, 324, 43 },
		{ 90, 325, 43 },
		{ 90, 326, 43 },
		{ 90, 327, 43 },
		{ 90, 328, 43 },
		{ 90, 329, 43 },
		{ 90, 330, 43 },
		{ 90, 331, 43 },
		{ 90, 332, 43 },
		{ 90, 333, 43 },
		{ 90, 334, 43 },
		{ 90, 335, 43 },
		{ 90, 336, 43 },
		{ 90, 337, 43 },
		{ 90, 338, 43 },
		{ 90, 339, 43 },
		{ 90, 340, 43 },
		{ 91, 299, 43 },
		{ 91, 300, 43 },
		{ 91, 301, 43 },
		{ 91, 302, 43 },
		{ 91, 303, 43 },
		{ 91, 304, 43 },
		{ 91, 305, 43 },
		{ 91, 306, 43 },
		{ 91, 307, 43 },
		{ 91, 308, 43 },
		{ 91, 309, 43 },
		{ 91, 310, 43 },
		{ 91, 311, 43 },
		{ 91, 312, 43 },
		{ 91, 313, 43 },
		{ 91, 314, 43 },
		{ 91, 315, 43 },
		{ 91, 316, 43 },
		{ 91, 317, 43 },
		{ 91, 318, 43 },
		{ 91, 319, 43 },
		{ 91, 320, 43 },
		{ 91, 321, 43 },
		{ 91, 322, 43 },
		{ 91, 323, 43 },
		{ 91, 324, 43 },
		{ 91, 325, 43 },
		{ 91, 326, 43 },
		{ 91, 327, 43 },
		{ 91, 328, 43 },
		{ 91, 329, 43 },
		{ 91, 330, 43 },
		{ 91, 331, 43 },
		{ 91, 332, 43 },
		{ 91, 333, 43 },
		{ 91, 334, 43 },
		{ 91, 335, 43 },
		{ 91, 336, 43 },
		{ 91, 337, 43 },
		{ 91, 338, 43 },
		{ 91, 339, 43 },
		{ 92, 299, 43 },
		{ 92, 300, 43 },
		{ 92, 301, 43 },
		{ 92, 302, 43 },
		{ 92, 303, 43 },
		{ 92, 304, 43 },
		{ 92, 305, 43 },
		{ 92, 306, 43 },
		{ 92, 307, 43 },
		{ 92, 308, 43 },
		{ 92, 309, 43 },
		{ 92, 310, 43 },
		{ 92, 311, 43 },
		{ 92, 312, 43 },
		{ 92, 313, 43 },
		{ 92, 314, 43 },
		{ 92, 315, 43 },
		{ 92, 316, 43 },
		{ 92, 317, 43 },
		{ 92, 318, 43 },
		{ 92, 319, 43 },
		{ 92, 320, 43 },
		{ 92, 321, 43 },
		{ 92, 322, 43 },
		{ 92, 323, 43 },
		{ 92, 324, 43 },
		{ 92, 325, 43 },
		{ 92, 326, 43 },
		{ 92, 327, 43 },
		{ 92, 328, 43 },
		{ 92, 329, 43 },
		{ 92, 330, 43 },
		{ 92, 331, 43 },
		{ 92, 332, 43 },
		{ 92, 333, 43 },
		{ 92, 334, 43 },
		{ 92, 335, 43 },
		{ 92, 336, 43 },
		{ 92, 337, 43 },
		{ 92, 338, 43 },
		{ 92, 339, 43 },
		{ 93, 299, 43 },
		{ 93, 300, 43 },
		{ 93, 301, 43 },
		{ 93, 302, 43 },
		{ 93, 303, 43 },
		{ 93, 304, 43 },
		{ 93, 305, 43 },
		{ 93, 306, 43 },
		{ 93, 307, 43 },
		{ 93, 308, 43 },
		{ 93, 309, 43 },
		{ 93, 310, 43 },
		{ 93, 311, 43 },
		{ 93, 312, 43 },
		{ 93, 313, 43 },
		{ 93, 314, 43 },
		{ 93, 315, 43 },
		{ 93, 316, 43 },
		{ 93, 317, 43 },
		{ 93, 318, 43 },
		{ 93, 319, 43 },
		{ 93, 320, 43 },
		{ 93, 321, 43 },
		{ 93, 322, 43 },
		{ 93, 323, 43 },
		{ 93, 324, 43 },
		{ 93, 325, 43 },
		{ 93, 326, 43 },
		{ 93, 327, 43 },
		{ 93, 328, 43 },
		{ 93, 329, 43 },
		{ 93, 330, 43 },
		{ 93, 331, 43 },
		{ 93, 332, 43 },
		{ 93, 333, 43 },
		{ 93, 334, 43 },
		{ 93, 335, 43 },
		{ 93, 336, 43 },
		{ 93, 337, 43 },
		{ 93, 338, 43 },
		{ 94, 298, 43 },
		{ 94, 299, 43 },
		{ 94, 300, 43 },
		{ 94, 301, 43 },
		{ 94, 302, 43 },
		{ 94, 303, 43 },
		{ 94, 304, 43 },
		{ 94, 305, 43 },
		{ 94, 306, 43 },
		{ 94, 307, 43 },
		{ 94, 308, 43 },
		{ 94, 309, 43 },
		{ 94, 310, 43 },
		{ 94, 311, 43 },
		{ 94, 312, 43 },
		{ 94, 313, 43 },
		{ 94, 314, 43 },
		{ 94, 315, 43 },
		{ 94, 316, 43 },
		{ 94, 317, 43 },
		{ 94, 318, 43 },
		{ 94, 319, 43 },
		{ 94, 320, 43 },
		{ 94, 321, 43 },
		{ 94, 322, 43 },
		{ 94, 323, 43 },
		{ 94, 324, 43 },
		{ 94, 325, 43 },
		{ 94, 326, 43 },
		{ 94, 327, 43 },
		{ 94, 328, 43 },
		{ 94, 329, 43 },
		{ 94, 330, 43 },
		{ 94, 331, 43 },
		{ 94, 332, 43 },
		{ 94, 333, 43 },
		{ 94, 334, 43 },
		{ 94, 335, 43 },
		{ 94, 336, 43 },
		{ 94, 337, 43 },
		{ 94, 338, 43 },
		{ 95, 298, 43 },
		{ 95, 299, 43 },
		{ 95, 300, 43 },
		{ 95, 301, 43 },
		{ 95, 302, 43 },
		{ 95, 303, 43 },
		{ 95, 304, 43 },
		{ 95, 305, 43 },
		{ 95, 306, 43 },
		{ 95, 307, 43 },
		{ 95, 308, 43 },
		{ 95, 309, 43 },
		{ 95, 310, 43 },
		{ 95, 311, 43 },
		{ 95, 312, 43 },
		{ 95, 313, 43 },
		{ 95, 314, 43 },
		{ 95, 315, 43 },
		{ 95, 316, 43 },
		{ 95, 317, 43 },
		{ 95, 318, 43 },
		{ 95, 319, 43 },
		{ 95, 320, 43 },
		{ 95, 321, 43 },
		{ 95, 322, 43 },
		{ 95, 323, 43 },
		{ 95, 324, 43 },
		{ 95, 325, 43 },
		{ 95, 326, 43 },
		{ 95, 327, 43 },
		{ 95, 328, 43 },
		{ 95, 329, 43 },
		{ 95, 330, 43 },
		{ 95, 331, 43 },
		{ 95, 332, 43 },
		{ 95, 333, 43 },
		{ 95, 334, 43 },
		{ 95, 335, 43 },
		{ 95, 336, 43 },
		{ 95, 337, 43 },
		{ 96, 298, 43 },
		{ 96, 299, 43 },
		{ 96, 300, 43 },
		{ 96, 301, 43 },
		{ 96, 302, 43 },
		{ 96, 303, 43 },
		{ 96, 304, 43 },
		{ 96, 305, 43 },
		{ 96, 306, 43 },
		{ 96, 307, 43 },
		{ 96, 308, 43 },
		{ 96, 309, 43 },
		{ 96, 310, 43 },
		{ 96, 311, 43 },
		{ 96, 312, 43 },
		{ 96, 313, 43 },
		{ 96, 314, 43 },
		{ 96, 315, 43 },
		{ 96, 316, 43 },
		{ 96, 317, 43 },
		{ 96, 318, 43 },
		{ 96, 319, 43 },
		{ 96, 320, 43 },
		{ 96, 321, 43 },
		{ 96, 322, 43 },
		{ 96, 323, 43 },
		{ 96, 324, 43 },
		{ 96, 325, 43 },
		{ 96, 326, 43 },
		{ 96, 327, 43 },
		{ 96, 328, 43 },
		{ 96, 329, 43 },
		{ 96, 330, 43 },
		{ 96, 331, 43 },
		{ 96, 332, 43 },
		{ 96, 333, 43 },
		{ 96, 334, 43 },
		{ 96, 335, 43 },
		{ 96, 336, 43 },
		{ 96, 337, 43 },
		{ 97, 298, 43 },
		{ 97, 299, 43 },
		{ 97, 300, 43 },
		{ 97, 301, 43 },
		{ 97, 302, 43 },
		{ 97, 303, 43 },
		{ 97, 304, 43 },
		{ 97, 305, 43 },
		{ 97, 306, 43 },
		{ 97, 307, 43 },
		{ 97, 308, 43 },
		{ 97, 309, 43 },
		{ 97, 310, 43 },
		{ 97, 311, 43 },
		{ 97, 312, 43 },
		{ 97, 313, 43 },
		{ 97, 314, 43 },
		{ 97, 315, 43 },
		{ 97, 316, 43 },
		{ 97, 317, 43 },
		{ 97, 318, 43 },
		{ 97, 319, 43 },
		{ 97, 320, 43 },
		{ 97, 321, 43 },
		{ 97, 322, 43 },
		{ 97, 323, 43 },
		{ 97, 324, 43 },
		{ 97, 325, 43 },
		{ 97, 326, 43 },
		{ 97, 327, 43 },
		{ 97, 328, 43 },
		{ 97, 329, 43 },
		{ 97, 330, 43 },
		{ 97, 331, 43 },
		{ 97, 332, 43 },
		{ 97, 333, 43 },
		{ 97, 334, 43 },
		{ 97, 335, 43 },
		{ 97, 336, 43 },
		{ 98, 297, 43 },
		{ 98, 298, 43 },
		{ 98, 299, 43 },
		{ 98, 300, 43 },
		{ 98, 301, 43 },
		{ 98, 302, 43 },
		{ 98, 303, 43 },
		{ 98, 304, 43 },
		{ 98, 305, 43 },
		{ 98, 306, 43 },
		{ 98, 307, 43 },
		{ 98, 308, 43 },
		{ 98, 309, 43 },
		{ 98, 310, 43 },
		{ 98, 311, 43 },
		{ 98, 312, 43 },
		{ 98, 313, 43 },
		{ 98, 314, 43 },
		{ 98, 315, 43 },
		{ 98, 316, 43 },
		{ 98, 317, 43 },
		{ 98, 318, 43 },
		{ 98, 319, 43 },
		{ 98, 320, 43 },
		{ 98, 321, 43 },
		{ 98, 322, 43 },
		{ 98, 323, 43 },
		{ 98, 324, 43 },
		{ 98, 325, 43 },
		{ 98, 326, 43 },
		{ 98, 327, 43 },
		{ 98, 328, 43 },
		{ 98, 329, 43 },
		{ 98, 330, 43 },
		{ 98, 331, 43 },
		{ 98, 332, 43 },
		{ 98, 333, 43 },
		{ 98, 334, 43 },
		{ 98, 335, 43 },
		{ 98, 336, 43 },
		{ 99, 297, 43 },
		{ 99, 298, 43 },
		{ 99, 299, 43 },
		{ 99, 300, 43 },
		{ 99, 301, 43 },
		{ 99, 302, 43 },
		{ 99, 303, 43 },
		{ 99, 304, 43 },
		{ 99, 305, 43 },
		{ 99, 306, 43 },
		{ 99, 307, 43 },
		{ 99, 308, 43 },
		{ 99, 309, 43 },
		{ 99, 310, 43 },
		{ 99, 311, 43 },
		{ 99, 312, 43 },
		{ 99, 313, 43 },
		{ 99, 314, 43 },
		{ 99, 315, 43 },
		{ 99, 316, 43 },
		{ 99, 317, 43 },
		{ 99, 318, 43 },
		{ 99, 319, 43 },
		{ 99, 320, 43 },
		{ 99, 321, 43 },
		{ 99, 322, 43 },
		{ 99, 323, 43 },
		{ 99, 324, 43 },
		{ 99, 325, 43 },
		{ 99, 326, 43 },
		{ 99, 327, 43 },
		{ 99, 328, 43 },
		{ 99, 329, 43 },
		{ 99, 330, 43 },
		{ 99, 331, 43 },
		{ 99, 332, 43 },
		{ 99, 333, 43 },
		{ 99, 334, 43 },
		{ 99, 335, 43 },
		{ 100, 297, 43 },
		{ 100, 298, 43 },
		{ 100, 299, 43 },
		{ 100, 300, 43 },
		{ 100, 301, 43 },
		{ 100, 302, 43 },
		{ 100, 303, 43 },
		{ 100, 304, 43 },
		{ 100, 305, 43 },
		{ 100, 306, 43 },
		{ 100, 307, 43 },
		{ 100, 308, 43 },
		{ 100, 309, 43 },
		{ 100, 310, 43 },
		{ 100, 311, 43 },
		{ 100, 312, 43 },
		{ 100, 313, 43 },
		{ 100, 314, 43 },
		{ 100, 315, 43 },
		{ 100, 316, 43 },
		{ 100, 317, 43 },
		{ 100, 318, 43 },
		{ 100, 319, 43 },
		{ 100, 320, 43 },
		{ 100, 321, 43 },
		{ 100, 322, 43 },
		{ 100, 323, 43 },
		{ 100, 324, 43 },
		{ 100, 325, 43 },
		{ 100, 326, 43 },
		{ 100, 327, 43 },
		{ 100, 328, 43 },
		{ 100, 329, 43 },
		{ 100, 330, 43 },
		{ 100, 331, 43 },
		{ 100, 332, 43 },
		{ 100, 333, 43 },
		{ 100, 334, 43 },
		{ 100, 335, 43 },
		{ 101, 297, 43 },
		{ 101, 298, 43 },
		{ 101, 299, 43 },
		{ 101, 300, 43 },
		{ 101, 301, 43 },
		{ 101, 302, 43 },
		{ 101, 303, 43 },
		{ 101, 304, 43 },
		{ 101, 305, 43 },
		{ 101, 306, 43 },
		{ 101, 307, 43 },
		{ 101, 308, 43 },
		{ 101, 309, 43 },
		{ 101, 310, 43 },
		{ 101, 311, 43 },
		{ 101, 312, 43 },
		{ 101, 313, 43 },
		{ 101, 314, 43 },
		{ 101, 315, 43 },
		{ 101, 316, 43 },
		{ 101, 317, 43 },
		{ 101, 318, 43 },
		{ 101, 319, 43 },
		{ 101, 320, 43 },
		{ 101, 321, 43 },
		{ 101, 322, 43 },
		{ 101, 323, 43 },
		{ 101, 324, 43 },
		{ 101, 325, 43 },
		{ 101, 326, 43 },
		{ 101, 327, 43 },
		{ 101, 328, 43 },
		{ 101, 329, 43 },
		{ 101, 330, 43 },
		{ 101, 331, 43 },
		{ 101, 332, 43 },
		{ 101, 333, 43 },
		{ 101, 334, 43 },
		{ 102, 296, 43 },
		{ 102, 297, 43 },
		{ 102, 298, 43 },
		{ 102, 299, 43 },
		{ 102, 300, 43 },
		{ 102, 301, 43 },
		{ 102, 302, 43 },
		{ 102, 303, 43 },
		{ 102, 304, 43 },
		{ 102, 305, 43 },
		{ 102, 306, 43 },
		{ 102, 307, 43 },
		{ 102, 308, 43 },
		{ 102, 309, 43 },
		{ 102, 310, 43 },
		{ 102, 311, 43 },
		{ 102, 312, 43 },
		{ 102, 313, 43 },
		{ 102, 314, 43 },
		{ 102, 315, 43 },
		{ 102, 316, 43 },
		{ 102, 317, 43 },
		{ 102, 318, 43 },
		{ 102, 319, 43 },
		{ 102, 320, 43 },
		{ 102, 321, 43 },
		{ 102, 322, 43 },
		{ 102, 323, 43 },
		{ 102, 324, 43 },
		{ 102, 325, 43 },
		{ 102, 326, 43 },
		{ 102, 327, 43 },
		{ 102, 328, 43 },
		{ 102, 329, 43 },
		{ 102, 330, 43 },
		{ 102, 331, 43 },
		{ 102, 332, 43 },
		{ 102, 333, 43 },
		{ 102, 334, 43 },
		{ 103, 296, 43 },
		{ 103, 297, 43 },
		{ 103, 298, 43 },
		{ 103, 299, 43 },
		{ 103, 300, 43 },
		{ 103, 301, 43 },
		{ 103, 302, 43 },
		{ 103, 303, 43 },
		{ 103, 304, 43 },
		{ 103, 305, 43 },
		{ 103, 306, 43 },
		{ 103, 307, 43 },
		{ 103, 308, 43 },
		{ 103, 309, 43 },
		{ 103, 310, 43 },
		{ 103, 311, 43 },
		{ 103, 312, 43 },
		{ 103, 313, 43 },
		{ 103, 314, 43 },
		{ 103, 315, 43 },
		{ 103, 316, 43 },
		{ 103, 317, 43 },
		{ 103, 318, 43 },
		{ 103, 319, 43 },
		{ 103, 320, 43 },
		{ 103, 321, 43 },
		{ 103, 322, 43 },
		{ 103, 323, 43 },
		{ 103, 324, 43 },
		{ 103, 325, 43 },
		{ 103, 326, 43 },
		{ 103, 327, 43 },
		{ 103, 328, 43 },
		{ 103, 329, 43 },
		{ 103, 330, 43 },
		{ 103, 331, 43 },
		{ 103, 332, 43 },
		{ 103, 333, 43 },
		{ 104, 296, 43 },
		{ 104, 297, 43 },
		{ 104, 298, 43 },
		{ 104, 299, 43 },
		{ 104, 300, 43 },
		{ 104, 301, 43 },
		{ 104, 302, 43 },
		{ 104, 303, 43 },
		{ 104, 304, 43 },
		{ 104, 305, 43 },
		{ 104, 306, 43 },
		{ 104, 307, 43 },
		{ 104, 308, 43 },
		{ 104, 309, 43 },
		{ 104, 310, 43 },
		{ 104, 311, 43 },
		{ 104, 312, 43 },
		{ 104, 313, 43 },
		{ 104, 314, 43 },
		{ 104, 315, 43 },
		{ 104, 316, 43 },
		{ 104, 317, 43 },
		{ 104, 318, 43 },
		{ 104, 319, 43 },
		{ 104, 320, 43 },
		{ 104, 321, 43 },
		{ 104, 322, 43 },
		{ 104, 323, 43 },
		{ 104, 324, 43 },
		{ 104, 325, 43 },
		{ 104, 326, 43 },
		{ 104, 327, 43 },
		{ 104, 328, 43 },
		{ 104, 329, 43 },
		{ 104, 330, 43 },
		{ 104, 331, 43 },
		{ 104, 332, 43 },
		{ 104, 333, 43 },
		{ 105, 296, 43 },
		{ 105, 297, 43 },
		{ 105, 298, 43 },
		{ 105, 299, 43 },
		{ 105, 300, 43 },
		{ 105, 301, 43 },
		{ 105, 302, 43 },
		{ 105, 303, 43 },
		{ 105, 304, 43 },
		{ 105, 305, 43 },
		{ 105, 306, 43 },
		{ 105, 307, 43 },
		{ 105, 308, 43 },
		{ 105, 309, 43 },
		{ 105, 310, 43 },
		{ 105, 311, 43 },
		{ 105, 312, 43 },
		{ 105, 313, 43 },
		{ 105, 314, 43 },
		{ 105, 315, 43 },
		{ 105, 316, 43 },
		{ 105, 317, 43 },
		{ 105, 318, 43 },
		{ 105, 319, 43 },
		{ 105, 320, 43 },
		{ 105, 321, 43 },
		{ 105, 322, 43 },
		{ 105, 323, 43 },
		{ 105, 324, 43 },
		{ 105, 325, 43 },
		{ 105, 326, 43 },
		{ 105, 327, 43 },
		{ 105, 328, 43 },
		{ 105, 329, 43 },
		{ 105, 330, 43 },
		{ 105, 331, 43 },
		{ 105, 332, 43 },
		{ 106, 295, 43 },
		{ 106, 296, 43 },
		{ 106, 297, 43 },
		{ 106, 298, 43 },
		{ 106, 299, 43 },
		{ 106, 300, 43 },
		{ 106, 301, 43 },
		{ 106, 302, 43 },
		{ 106, 303, 43 },
		{ 106, 304, 43 },
		{ 106, 305, 43 },
		{ 106, 306, 43 },
		{ 106, 307, 43 },
		{ 106, 308, 43 },
		{ 106, 309, 43 },
		{ 106, 310, 43 },
		{ 106, 311, 43 },
		{ 106, 312, 43 },
		{ 106, 313, 43 },
		{ 106, 314, 43 },
		{ 106, 315, 43 },
		{ 106, 316, 43 },
		{ 106, 317, 43 },
		{ 106, 318, 43 },
		{ 106, 319, 43 },
		{ 106, 320, 43 },
		{ 106, 321, 43 },
		{ 106, 322, 43 },
		{ 106, 323, 43 },
		{ 106, 324, 43 },
		{ 106, 325, 43 },
		{ 106, 326, 43 },
		{ 106, 327, 43 },
		{ 106, 328, 43 },
		{ 106, 329, 43 },
		{ 106, 330, 43 },
		{ 106, 331, 43 },
		{ 106, 332, 43 },
		{ 107, 295, 43 },
		{ 107, 296, 43 },
		{ 107, 297, 43 },
		{ 107, 298, 43 },
		{ 107, 299, 43 },
		{ 107, 300, 43 },
		{ 107, 301, 43 },
		{ 107, 302, 43 },
		{ 107, 303, 43 },
		{ 107, 304, 43 },
		{ 107, 305, 43 },
		{ 107, 306, 43 },
		{ 107, 307, 43 },
		{ 107, 308, 43 },
		{ 107, 309, 43 },
		{ 107, 310, 43 },
		{ 107, 311, 43 },
		{ 107, 312, 43 },
		{ 107, 313, 43 },
		{ 107, 314, 43 },
		{ 107, 315, 43 },
		{ 107, 316, 43 },
		{ 107, 317, 43 },
		{ 107, 318, 43 },
		{ 107, 319, 43 },
		{ 107, 320, 43 },
		{ 107, 321, 43 },
		{ 107, 322, 43 },
		{ 107, 323, 43 },
		{ 107, 324, 43 },
		{ 107, 325, 43 },
		{ 107, 326, 43 },
		{ 107, 327, 43 },
		{ 107, 328, 43 },
		{ 107, 329, 43 },
		{ 107, 330, 43 },
		{ 107, 331, 43 },
		{ 108, 295, 43 },
		{ 108, 296, 43 },
		{ 108, 297, 43 },
		{ 108, 298, 43 },
		{ 108, 299, 43 },
		{ 108, 300, 43 },
		{ 108, 301, 43 },
		{ 108, 302, 43 },
		{ 108, 303, 43 },
		{ 108, 304, 43 },
		{ 108, 305, 43 },
		{ 108, 306, 43 },
		{ 108, 307, 43 },
		{ 108, 308, 43 },
		{ 108, 309, 43 },
		{ 108, 310, 43 },
		{ 108, 311, 43 },
		{ 108, 312, 43 },
		{ 108, 313, 43 },
		{ 108, 314, 43 },
		{ 108, 315, 43 },
		{ 108, 316, 43 },
		{ 108, 317, 43 },
		{ 108, 318, 43 },
		{ 108, 319, 43 },
		{ 108, 320, 43 },
		{ 108, 321, 43 },
		{ 108, 322, 43 },
		{ 108, 323, 43 },
		{ 108, 324, 43 },
		{ 108, 325, 43 },
		{ 108, 326, 43 },
		{ 108, 327, 43 },
		{ 108, 328, 43 },
		{ 108, 329, 43 },
		{ 108, 330, 43 },
		{ 108, 331, 43 },
		{ 109, 295, 43 },
		{ 109, 296, 43 },
		{ 109, 297, 43 },
		{ 109, 298, 43 },
		{ 109, 299, 43 },
		{ 109, 300, 43 },
		{ 109, 301, 43 },
		{ 109, 302, 43 },
		{ 109, 303, 43 },
		{ 109, 304, 43 },
		{ 109, 305, 43 },
		{ 109, 306, 43 },
		{ 109, 307, 43 },
		{ 109, 308, 43 },
		{ 109, 309, 43 },
		{ 109, 310, 43 },
		{ 109, 311, 43 },
		{ 109, 312, 43 },
		{ 109, 313, 43 },
		{ 109, 314, 43 },
		{ 109, 315, 43 },
		{ 109, 316, 43 },
		{ 109, 317, 43 },
		{ 109, 318, 43 },
		{ 109, 319, 43 },
		{ 109, 320, 43 },
		{ 109, 321, 43 },
		{ 109, 322, 43 },
		{ 109, 323, 43 },
		{ 109, 324, 43 },
		{ 109, 325, 43 },
		{ 109, 326, 43 },
		{ 109, 327, 43 },
		{ 109, 328, 43 },
		{ 109, 329, 43 },
		{ 109, 330, 43 },
		{ 110, 294, 43 },
		{ 110, 295, 43 },
		{ 110, 296, 43 },
		{ 110, 297, 43 },
		{ 110, 298, 43 },
		{ 110, 299, 43 },
		{ 110, 300, 43 },
		{ 110, 301, 43 },
		{ 110, 302, 43 },
		{ 110, 303, 43 },
		{ 110, 304, 43 },
		{ 110, 305, 43 },
		{ 110, 306, 43 },
		{ 110, 307, 43 },
		{ 110, 308, 43 },
		{ 110, 309, 43 },
		{ 110, 310, 43 },
		{ 110, 311, 43 },
		{ 110, 312, 43 },
		{ 110, 313, 43 },
		{ 110, 314, 43 },
		{ 110, 315, 43 },
		{ 110, 316, 43 },
		{ 110, 317, 43 },
		{ 110, 318, 43 },
		{ 110, 319, 43 },
		{ 110, 320, 43 },
		{ 110, 321, 43 },
		{ 110, 322, 43 },
		{ 110, 323, 43 },
		{ 110, 324, 43 },
		{ 110, 325, 43 },
		{ 110, 326, 43 },
		{ 110, 327, 43 },
		{ 110, 328, 43 },
		{ 110, 329, 43 },
		{ 110, 330, 43 },
		{ 111, 294, 43 },
		{ 111, 295, 43 },
		{ 111, 296, 43 },
		{ 111, 297, 43 },
		{ 111, 298, 43 },
		{ 111, 299, 43 },
		{ 111, 300, 43 },
		{ 111, 301, 43 },
		{ 111, 302, 43 },
		{ 111, 303, 43 },
		{ 111, 304, 43 },
		{ 111, 305, 43 },
		{ 111, 306, 43 },
		{ 111, 307, 43 },
		{ 111, 308, 43 },
		{ 111, 309, 43 },
		{ 111, 310, 43 },
		{ 111, 311, 43 },
		{ 111, 312, 43 },
		{ 111, 313, 43 },
		{ 111, 314, 43 },
		{ 111, 315, 43 },
		{ 111, 316, 43 },
		{ 111, 317, 43 },
		{ 111, 318, 43 },
		{ 111, 319, 43 },
		{ 111, 320, 43 },
		{ 111, 321, 43 },
		{ 111, 322, 43 },
		{ 111, 323, 43 },
		{ 111, 324, 43 },
		{ 111, 325, 43 },
		{ 111, 326, 43 },
		{ 111, 327, 43 },
		{ 111, 328, 43 },
		{ 111, 329, 43 },
		{ 112, 294, 43 },
		{ 112, 295, 43 },
		{ 112, 296, 43 },
		{ 112, 297, 43 },
		{ 112, 298, 43 },
		{ 112, 299, 43 },
		{ 112, 300, 43 },
		{ 112, 301, 43 },
		{ 112, 302, 43 },
		{ 112, 303, 43 },
		{ 112, 304, 43 },
		{ 112, 305, 43 },
		{ 112, 306, 43 },
		{ 112, 307, 43 },
		{ 112, 308, 43 },
		{ 112, 309, 43 },
		{ 112, 310, 43 },
		{ 112, 311, 43 },
		{ 112, 312, 43 },
		{ 112, 313, 43 },
		{ 112, 314, 43 },
		{ 112, 315, 43 },
		{ 112, 316, 43 },
		{ 112, 317, 43 },
		{ 112, 318, 43 },
		{ 112, 319, 43 },
		{ 112, 320, 43 },
		{ 112, 321, 43 },
		{ 112, 322, 43 },
		{ 112, 323, 43 },
		{ 112, 324, 43 },
		{ 112, 325, 43 },
		{ 112, 326, 43 },
		{ 112, 327, 43 },
		{ 112, 328, 43 },
		{ 112, 329, 43 },
		{ 113, 294, 43 },
		{ 113, 295, 43 },
		{ 113, 296, 43 },
		{ 113, 297, 43 },
		{ 113, 298, 43 },
		{ 113, 299, 43 },
		{ 113, 300, 43 },
		{ 113, 301, 43 },
		{ 113, 302, 43 },
		{ 113, 303, 43 },
		{ 113, 304, 43 },
		{ 113, 305, 43 },
		{ 113, 306, 43 },
		{ 113, 307, 43 },
		{ 113, 308, 43 },
		{ 113, 309, 43 },
		{ 113, 310, 43 },
		{ 113, 311, 43 },
		{ 113, 312, 43 },
		{ 113, 313, 43 },
		{ 113, 314, 43 },
		{ 113, 315, 43 },
		{ 113, 316, 43 },
		{ 113, 317, 43 },
		{ 113, 318, 43 },
		{ 113, 319, 43 },
		{ 113, 320, 43 },
		{ 113, 321, 43 },
		{ 113, 322, 43 },
		{ 113, 323, 43 },
		{ 113, 324, 43 },
		{ 113, 325, 43 },
		{ 113, 326, 43 },
		{ 113, 327, 43 },
		{ 113, 328, 43 },
		{ 114, 293, 43 },
		{ 114, 294, 43 },
		{ 114, 295, 43 },
		{ 114, 296, 43 },
		{ 114, 297, 43 },
		{ 114, 298, 43 },
		{ 114, 299, 43 },
		{ 114, 300, 43 },
		{ 114, 301, 43 },
		{ 114, 302, 43 },
		{ 114, 303, 43 },
		{ 114, 304, 43 },
		{ 114, 305, 43 },
		{ 114, 306, 43 },
		{ 114, 307, 43 },
		{ 114, 308, 43 },
		{ 114, 309, 43 },
		{ 114, 310, 43 },
		{ 114, 311, 43 },
		{ 114, 312, 43 },
		{ 114, 313, 43 },
		{ 114, 314, 43 },
		{ 114, 315, 43 },
		{ 114, 316, 43 },
		{ 114, 317, 43 },
		{ 114, 318, 43 },
		{ 114, 319, 43 },
		{ 114, 320, 43 },
		{ 114, 321, 43 },
		{ 114, 322, 43 },
		{ 114, 323, 43 },
		{ 114, 324, 43 },
		{ 114, 325, 43 },
		{ 114, 326, 43 },
		{ 114, 327, 43 },
		{ 114, 328, 43 },
		{ 115, 293, 43 },
		{ 115, 294, 43 },
		{ 115, 295, 43 },
		{ 115, 296, 43 },
		{ 115, 297, 43 },
		{ 115, 298, 43 },
		{ 115, 299, 43 },
		{ 115, 300, 43 },
		{ 115, 301, 43 },
		{ 115, 302, 43 },
		{ 115, 303, 43 },
		{ 115, 304, 43 },
		{ 115, 305, 43 },
		{ 115, 306, 43 },
		{ 115, 307, 43 },
		{ 115, 308, 43 },
		{ 115, 309, 43 },
		{ 115, 310, 43 },
		{ 115, 311, 43 },
		{ 115, 312, 43 },
		{ 115, 313, 43 },
		{ 115, 314, 43 },
		{ 115, 315, 43 },
		{ 115, 316, 43 },
		{ 115, 317, 43 },
		{ 115, 318, 43 },
		{ 115, 319, 43 },
		{ 115, 320, 43 },
		{ 115, 321, 43 },
		{ 115, 322, 43 },
		{ 115, 323, 43 },
		{ 115, 324, 43 },
		{ 115, 325, 43 },
		{ 115, 326, 43 },
		{ 115, 327, 43 },
		{ 116, 293, 43 },
		{ 116, 294, 43 },
		{ 116, 295, 43 },
		{ 116, 296, 43 },
		{ 116, 297, 43 },
		{ 116, 298, 43 },
		{ 116, 299, 43 },
		{ 116, 300, 43 },
		{ 116, 301, 43 },
		{ 116, 302, 43 },
		{ 116, 303, 43 },
		{ 116, 304, 43 },
		{ 116, 305, 43 },
		{ 116, 306, 43 },
		{ 116, 307, 43 },
		{ 116, 308, 43 },
		{ 116, 309, 43 },
		{ 116, 310, 43 },
		{ 116, 311, 43 },
		{ 116, 312, 43 },
		{ 116, 313, 43 },
		{ 116, 314, 43 },
		{ 116, 315, 43 },
		{ 116, 316, 43 },
		{ 116, 317, 43 },
		{ 116, 318, 43 },
		{ 116, 319, 43 },
		{ 116, 320, 43 },
		{ 116, 321, 43 },
		{ 116, 322, 43 },
		{ 116, 323, 43 },
		{ 116, 324, 43 },
		{ 116, 325, 43 },
		{ 116, 326, 43 },
		{ 116, 327, 43 },
		{ 117, 293, 43 },
		{ 117, 294, 43 },
		{ 117, 295, 43 },
		{ 117, 296, 43 },
		{ 117, 297, 43 },
		{ 117, 298, 43 },
		{ 117, 299, 43 },
		{ 117, 300, 43 },
		{ 117, 301, 43 },
		{ 117, 302, 43 },
		{ 117, 303, 43 },
		{ 117, 304, 43 },
		{ 117, 305, 43 },
		{ 117, 306, 43 },
		{ 117, 307, 43 },
		{ 117, 308, 43 },
		{ 117, 309, 43 },
		{ 117, 310, 43 },
		{ 117, 311, 43 },
		{ 117, 312, 43 },
		{ 117, 313, 43 },
		{ 117, 314, 43 },
		{ 117, 315, 43 },
		{ 117, 316, 43 },
		{ 117, 317, 43 },
		{ 117, 318, 43 },
		{ 117, 319, 43 },
		{ 117, 320, 43 },
		{ 117, 321, 43 },
		{ 117, 322, 43 },
		{ 117, 323, 43 },
		{ 117, 324, 43 },
		{ 117, 325, 43 },
		{ 117, 326, 43 },
		{ 118, 292, 43 },
		{ 118, 293, 43 },
		{ 118, 294, 43 },
		{ 118, 295, 43 },
		{ 118, 296, 43 },
		{ 118, 297, 43 },
		{ 118, 298, 43 },
		{ 118, 299, 43 },
		{ 118, 300, 43 },
		{ 118, 301, 43 },
		{ 118, 302, 43 },
		{ 118, 303, 43 },
		{ 118, 304, 43 },
		{ 118, 305, 43 },
		{ 118, 306, 43 },
		{ 118, 307, 43 },
		{ 118, 308, 43 },
		{ 118, 309, 43 },
		{ 118, 310, 43 },
		{ 118, 311, 43 },
		{ 118, 312, 43 },
		{ 118, 313, 43 },
		{ 118, 314, 43 },
		{ 118, 315, 43 },
		{ 118, 316, 43 },
		{ 118, 317, 43 },
		{ 118, 318, 43 },
		{ 118, 319, 43 },
		{ 118, 320, 43 },
		{ 118, 321, 43 },
		{ 118, 322, 43 },
		{ 118, 323, 43 },
		{ 118, 324, 43 },
		{ 118, 325, 43 },
		{ 118, 326, 43 },
		{ 119, 292, 43 },
		{ 119, 293, 43 },
		{ 119, 294, 43 },
		{ 119, 295, 43 },
		{ 119, 296, 43 },
		{ 119, 297, 43 },
		{ 119, 298, 43 },
		{ 119, 299, 43 },
		{ 119, 300, 43 },
		{ 119, 301, 43 },
		{ 119, 302, 43 },
		{ 119, 303, 43 },
		{ 119, 304, 43 },
		{ 119, 305, 43 },
		{ 119, 306, 43 },
		{ 119, 307, 43 },
		{ 119, 308, 43 },
		{ 119, 309, 43 },
		{ 119, 310, 43 },
		{ 119, 311, 43 },
		{ 119, 312, 43 },
		{ 119, 313, 43 },
		{ 119, 314, 43 },
		{ 119, 315, 43 },
		{ 119, 316, 43 },
		{ 119, 317, 43 },
		{ 119, 318, 43 },
		{ 119, 319, 43 },
		{ 119, 320, 43 },
		{ 119, 321, 43 },
		{ 119, 322, 43 },
		{ 119, 323, 43 },
		{ 119, 324, 43 },
		{ 119, 325, 43 },
		{ 120, 292, 43 },
		{ 120, 293, 43 },
		{ 120, 294, 43 },
		{ 120, 295, 43 },
		{ 120, 296, 43 },
		{ 120, 297, 43 },
		{ 120, 298, 43 },
		{ 120, 299, 43 },
		{ 120, 300, 43 },
		{ 120, 301, 43 },
		{ 120, 302, 43 },
		{ 120, 303, 43 },
		{ 120, 304, 43 },
		{ 120, 305, 43 },
		{ 120, 306, 43 },
		{ 120, 307, 43 },
		{ 120, 308, 43 },
		{ 120, 309, 43 },
		{ 120, 310, 43 },
		{ 120, 311, 43 },
		{ 120, 312, 43 },
		{ 120, 313, 43 },
		{ 120, 314, 43 },
		{ 120, 315, 43 },
		{ 120, 316, 43 },
		{ 120, 317, 43 },
		{ 120, 318, 43 },
		{ 120, 319, 43 },
		{ 120, 320, 43 },
		{ 120, 321, 43 },
		{ 120, 322, 43 },
		{ 120, 323, 43 },
		{ 120, 324, 43 },
		{ 120, 325, 43 },
		{ 121, 292, 43 },
		{ 121, 293, 43 },
		{ 121, 294, 43 },
		{ 121, 295, 43 },
		{ 121, 296, 43 },
		{ 121, 297, 43 },
		{ 121, 298, 43 },
		{ 121, 299, 43 },
		{ 121, 300, 43 },
		{ 121, 301, 43 },
		{ 121, 302, 43 },
		{ 121, 303, 43 },
		{ 121, 304, 43 },
		{ 121, 305, 43 },
		{ 121, 306, 43 },
		{ 121, 307, 43 },
		{ 121, 308, 43 },
		{ 121, 309, 43 },
		{ 121, 310, 43 },
		{ 121, 311, 43 },
		{ 121, 312, 43 },
		{ 121, 313, 43 },
		{ 121, 314, 43 },
		{ 121, 315, 43 },
		{ 121, 316, 43 },
		{ 121, 317, 43 },
		{ 121, 318, 43 },
		{ 121, 319, 43 },
		{ 121, 320, 43 },
		{ 121, 321, 43 },
		{ 121, 322, 43 },
		{ 121, 323, 43 },
		{ 121, 324, 43 },
		{ 122, 291, 43 },
		{ 122, 292, 43 },
		{ 122, 293, 43 },
		{ 122, 294, 43 },
		{ 122, 295, 43 },
		{ 122, 296, 43 },
		{ 122, 297, 43 },
		{ 122, 298, 43 },
		{ 122, 299, 43 },
		{ 122, 300, 43 },
		{ 122, 301, 43 },
		{ 122, 302, 43 },
		{ 122, 303, 43 },
		{ 122, 304, 43 },
		{ 122, 305, 43 },
		{ 122, 306, 43 },
		{ 122, 307, 43 },
		{ 122, 308, 43 },
		{ 122, 309, 43 },
		{ 122, 310, 43 },
		{ 122, 311, 43 },
		{ 122, 312, 43 },
		{ 122, 313, 43 },
		{ 122, 314, 43 },
		{ 122, 315, 43 },
		{ 122, 316, 43 },
		{ 122, 317, 43 },
		{ 122, 318, 43 },
		{ 122, 319, 43 },
		{ 122, 320, 43 },
		{ 122, 321, 43 },
		{ 122, 322, 43 },
		{ 122, 323, 43 },
		{ 122, 324, 43 },
		{ 123, 291, 43 },
		{ 123, 292, 43 },
		{ 123, 293, 43 },
		{ 123, 294, 43 },
		{ 123, 295, 43 },
		{ 123, 296, 43 },
		{ 123, 297, 43 },
		{ 123, 298, 43 },
		{ 123, 299, 43 },
		{ 123, 300, 43 },
		{ 123, 301, 43 },
		{ 123, 302, 43 },
		{ 123, 303, 43 },
		{ 123, 304, 43 },
		{ 123, 305, 43 },
		{ 123, 306, 43 },
		{ 123, 307, 43 },
		{ 123, 308, 43 },
		{ 123, 309, 43 },
		{ 123, 310, 43 },
		{ 123, 311, 43 },
		{ 123, 312, 43 },
		{ 123, 313, 43 },
		{ 123, 314, 43 },
		{ 123, 315, 43 },
		{ 123, 316, 43 },
		{ 123, 317, 43 },
		{ 123, 318, 43 },
		{ 123, 319, 43 },
		{ 123, 320, 43 },
		{ 123, 321, 43 },
		{ 123, 322, 43 },
		{ 123, 323, 43 },
		{ 124, 291, 43 },
		{ 124, 292, 43 },
		{ 124, 293, 43 },
		{ 124, 294, 43 },
		{ 124, 295, 43 },
		{ 124, 296, 43 },
		{ 124, 297, 43 },
		{ 124, 298, 43 },
		{ 124, 299, 43 },
		{ 124, 300, 43 },
		{ 124, 301, 43 },
		{ 124, 302, 43 },
		{ 124, 303, 43 },
		{ 124, 304, 43 },
		{ 124, 305, 43 },
		{ 124, 306, 43 },
		{ 124, 307, 43 },
		{ 124, 308, 43 },
		{ 124, 309, 43 },
		{ 124, 310, 43 },
		{ 124, 311, 43 },
		{ 124, 312, 43 },
		{ 124, 313, 43 },
		{ 124, 314, 43 },
		{ 124, 315, 43 },
		{ 124, 316, 43 },
		{ 124, 317, 43 },
		{ 124, 318, 43 },
		{ 124, 319, 43 },
		{ 124, 320, 43 },
		{ 124, 321, 43 },
		{ 124, 322, 43 },
		{ 124, 323, 43 },
		{ 125, 291, 43 },
		{ 125, 292, 43 },
		{ 125, 293, 43 },
		{ 125, 294, 43 },
		{ 125, 295, 43 },
		{ 125, 296, 43 },
		{ 125, 297, 43 },
		{ 125, 298, 43 },
		{ 125, 299, 43 },
		{ 125, 300, 43 },
		{ 125, 301, 43 },
		{ 125, 302, 43 },
		{ 125, 303, 43 },
		{ 125, 304, 43 },
		{ 125, 305, 43 },
		{ 125, 306, 43 },
		{ 125, 307, 43 },
		{ 125, 308, 43 },
		{ 125, 309, 43 },
		{ 125, 310, 43 },
		{ 125, 311, 43 },
		{ 125, 312, 43 },
		{ 125, 313, 43 },
		{ 125, 314, 43 },
		{ 125, 315, 43 },
		{ 125, 316, 43 },
		{ 125, 317, 43 },
		{ 125, 318, 43 },
		{ 125, 319, 43 },
		{ 125, 320, 43 },
		{ 125, 321, 43 },
		{ 125, 322, 43 },
		{ 126, 290, 43 },
		{ 126, 291, 43 },
		{ 126, 292, 43 },
		{ 126, 293, 43 },
		{ 126, 294, 43 },
		{ 126, 295, 43 },
		{ 126, 296, 43 },
		{ 126, 297, 43 },
		{ 126, 298, 43 },
		{ 126, 299, 43 },
		{ 126, 300, 43 },
		{ 126, 301, 43 },
		{ 126, 302, 43 },
		{ 126, 303, 43 },
		{ 126, 304, 43 },
		{ 126, 305, 43 },
		{ 126, 306, 43 },
		{ 126, 307, 43 },
		{ 126, 308, 43 },
		{ 126, 309, 43 },
		{ 126, 310, 43 },
		{ 126, 311, 43 },
		{ 126, 312, 43 },
		{ 126, 313, 43 },
		{ 126, 314, 43 },
		{ 126, 315, 43 },
		{ 126, 316, 43 },
		{ 126, 317, 43 },
		{ 126, 318, 43 },
		{ 126, 319, 43 },
		{ 126, 320, 43 },
		{ 126, 321, 43 },
		{ 126, 322, 43 },
		{ 127, 290, 43 },
		{ 127, 291, 43 },
		{ 127, 292, 43 },
		{ 127, 293, 43 },
		{ 127, 294, 43 },
		{ 127, 295, 43 },
		{ 127, 296, 43 },
		{ 127, 297, 43 },
		{ 127, 298, 43 },
		{ 127, 299, 43 },
		{ 127, 300, 43 },
		{ 127, 301, 43 },
		{ 127, 302, 43 },
		{ 127, 303, 43 },
		{ 127, 304, 43 },
		{ 127, 305, 43 },
		{ 127, 306, 43 },
		{ 127, 307, 43 },
		{ 127, 308, 43 },
		{ 127, 309, 43 },
		{ 127, 310, 43 },
		{ 127, 311, 43 },
		{ 127, 312, 43 },
		{ 127, 313, 43 },
		{ 127, 314, 43 },
		{ 127, 315, 43 },
		{ 127, 316, 43 },
		{ 127, 317, 43 },
		{ 127, 318, 43 },
		{ 127, 319, 43 },
		{ 127, 320, 43 },
		{ 127, 321, 43 },
		{ 128, 290, 43 },
		{ 128, 291, 43 },
		{ 128, 292, 43 },
		{ 128, 293, 43 },
		{ 128, 294, 43 },
		{ 128, 295, 43 },
		{ 128, 296, 43 },
		{ 128, 297, 43 },
		{ 128, 298, 43 },
		{ 128, 299, 43 },
		{ 128, 300, 43 },
		{ 128, 301, 43 },
		{ 128, 302, 43 },
		{ 128, 303, 43 },
		{ 128, 304, 43 },
		{ 128, 305, 43 },
		{ 128, 306, 43 },
		{ 128, 307, 43 },
		{ 128, 308, 43 },
		{ 128, 309, 43 },
		{ 128, 310, 43 },
		{ 128, 311, 43 },
		{ 128, 312, 43 },
		{ 128, 313, 43 },
		{ 128, 314, 43 },
		{ 128, 315, 43 },
		{ 128, 316, 43 },
		{ 128, 317, 43 },
		{ 128, 318, 43 },
		{ 128, 319, 43 },
		{ 128, 320, 43 },
		{ 128, 321, 43 },
		{ 129, 290, 43 },
		{ 129, 291, 43 },
		{ 129, 292, 43 },
		{ 129, 293, 43 },
		{ 129, 294, 43 },
		{ 129, 295, 43 },
		{ 129, 296, 43 },
		{ 129, 297, 43 },
		{ 129, 298, 43 },
		{ 129, 299, 43 },
		{ 129, 300, 43 },
		{ 129, 301, 43 },
		{ 129, 302, 43 },
		{ 129, 303, 43 },
		{ 129, 304, 43 },
		{ 129, 305, 43 },
		{ 129, 306, 43 },
		{ 129, 307, 43 },
		{ 129, 308, 43 },
		{ 129, 309, 43 },
		{ 129, 310, 43 },
		{ 129, 311, 43 },
		{ 129, 312, 43 },
		{ 129, 313, 43 },
		{ 129, 314, 43 },
		{ 129, 315, 43 },
		{ 129, 316, 43 },
		{ 129, 317, 43 },
		{ 129, 318, 43 },
		{ 129, 319, 43 },
		{ 129, 320, 43 },
		{ 130, 289, 43 },
		{ 130, 290, 43 },
		{ 130, 291, 43 },
		{ 130, 292, 43 },
		{ 130, 293, 43 },
		{ 130, 294, 43 },
		{ 130, 295, 43 },
		{ 130, 296, 43 },
		{ 130, 297, 43 },
		{ 130, 298, 43 },
		{ 130, 299, 43 },
		{ 130, 300, 43 },
		{ 130, 301, 43 },
		{ 130, 302, 43 },
		{ 130, 303, 43 },
		{ 130, 304, 43 },
		{ 130, 305, 43 },
		{ 130, 306, 43 },
		{ 130, 307, 43 },
		{ 130, 308, 43 },
		{ 130, 309, 43 },
		{ 130, 310, 43 },
		{ 130, 311, 43 },
		{ 130, 312, 43 },
		{ 130, 313, 43 },
		{ 130, 314, 43 },
		{ 130, 315, 43 },
		{ 130, 316, 43 },
		{ 130, 317, 43 },
		{ 130, 318, 43 },
		{ 130, 319, 43 },
		{ 130, 320, 43 },
		{ 131, 289, 43 },
		{ 131, 290, 43 },
		{ 131, 291, 43 },
		{ 131, 292, 43 },
		{ 131, 293, 43 },
		{ 131, 294, 43 },
		{ 131, 295, 43 },
		{ 131, 296, 43 },
		{ 131, 297, 43 },
		{ 131, 298, 43 },
		{ 131, 299, 43 },
		{ 131, 300, 43 },
		{ 131, 301, 43 },
		{ 131, 302, 43 },
		{ 131, 303, 43 },
		{ 131, 304, 43 },
		{ 131, 305, 43 },
		{ 131, 306, 43 },
		{ 131, 307, 43 },
		{ 131, 308, 43 },
		{ 131, 309, 43 },
		{ 131, 310, 43 },
		{ 131, 311, 43 },
		{ 131, 312, 43 },
		{ 131, 313, 43 },
		{ 131, 314, 43 },
		{ 131, 315, 43 },
		{ 131, 316, 43 },
		{ 131, 317, 43 },
		{ 131, 318, 43 },
		{ 131, 319, 43 },
		{ 132, 289, 43 },
		{ 132, 290, 43 },
		{ 132, 291, 43 },
		{ 132, 292, 43 },
		{ 132, 293, 43 },
		{ 132, 294, 43 },
		{ 132, 295, 43 },
		{ 132, 296, 43 },
		{ 132, 297, 43 },
		{ 132, 298, 43 },
		{ 132, 299, 43 },
		{ 132, 300, 43 },
		{ 132, 301, 43 },
		{ 132, 302, 43 },
		{ 132, 303, 43 },
		{ 132, 304, 43 },
		{ 132, 305, 43 },
		{ 132, 306, 43 },
		{ 132, 307, 43 },
		{ 132, 308, 43 },
		{ 132, 309, 43 },
		{ 132, 310, 43 },
		{ 132, 311, 43 },
		{ 132, 312, 43 },
		{ 132, 313, 43 },
		{ 132, 314, 43 },
		{ 132, 315, 43 },
		{ 132, 316, 43 },
		{ 132, 317, 43 },
		{ 132, 318, 43 },
		{ 132, 319, 43 },
		{ 133, 289, 43 },
		{ 133, 290, 43 },
		{ 133, 291, 43 },
		{ 133, 292, 43 },
		{ 133, 293, 43 },
		{ 133, 294, 43 },
		{ 133, 295, 43 },
		{ 133, 296, 43 },
		{ 133, 297, 43 },
		{ 133, 298, 43 },
		{ 133, 299, 43 },
		{ 133, 300, 43 },
		{ 133, 301, 43 },
		{ 133, 302, 43 },
		{ 133, 303, 43 },
		{ 133, 304, 43 },
		{ 133, 305, 43 },
		{ 133, 306, 43 },
		{ 133, 307, 43 },
		{ 133, 308, 43 },
		{ 133, 309, 43 },
		{ 133, 310, 43 },
		{ 133, 311, 43 },
		{ 133, 312, 43 },
		{ 133, 313, 43 },
		{ 133, 314, 43 },
		{ 133, 315, 43 },
		{ 133, 316, 43 },
		{ 133, 317, 43 },
		{ 133, 318, 43 },
		{ 134, 288, 43 },
		{ 134, 289, 43 },
		{ 134, 290, 43 },
		{ 134, 291, 43 },
		{ 134, 292, 43 },
		{ 134, 293, 43 },
		{ 134, 294, 43 },
		{ 134, 295, 43 },
		{ 134, 296, 43 },
		{ 134, 297, 43 },
		{ 134, 298, 43 },
		{ 134, 299, 43 },
		{ 134, 300, 43 },
		{ 134, 301, 43 },
		{ 134, 302, 43 },
		{ 134, 303, 43 },
		{ 134, 304, 43 },
		{ 134, 305, 43 },
		{ 134, 306, 43 },
		{ 134, 307, 43 },
		{ 134, 308, 43 },
		{ 134, 309, 43 },
		{ 134, 310, 43 },
		{ 134, 311, 43 },
		{ 134, 312, 43 },
		{ 134, 313, 43 },
		{ 134, 314, 43 },
		{ 134, 315, 43 },
		{ 134, 316, 43 },
		{ 134, 317, 43 },
		{ 134, 318, 43 },
		{ 135, 288, 43 },
		{ 135, 289, 43 },
		{ 135, 290, 43 },
		{ 135, 291, 43 },
		{ 135, 292, 43 },
		{ 135, 293, 43 },
		{ 135, 294, 43 },
		{ 135, 295, 43 },
		{ 135, 296, 43 },
		{ 135, 297, 43 },
		{ 135, 298, 43 },
		{ 135, 299, 43 },
		{ 135, 300, 43 },
		{ 135, 301, 43 },
		{ 135, 302, 43 },
		{ 135, 303, 43 },
		{ 135, 304, 43 },
		{ 135, 305, 43 },
		{ 135, 306, 43 },
		{ 135, 307, 43 },
		{ 135, 308, 43 },
		{ 135, 309, 43 },
		{ 135, 310, 43 },
		{ 135, 311, 43 },
		{ 135, 312, 43 },
		{ 135, 313, 43 },
		{ 135, 314, 43 },
		{ 135, 315, 43 },
		{ 135, 316, 43 },
		{ 135, 317, 43 },
		{ 136, 288, 43 },
		{ 136, 289, 43 },
		{ 136, 290, 43 },
		{ 136, 291, 43 },
		{ 136, 292, 43 },
		{ 136, 293, 43 },
		{ 136, 294, 43 },
		{ 136, 295, 43 },
		{ 136, 296, 43 },
		{ 136, 297, 43 },
		{ 136, 298, 43 },
		{ 136, 299, 43 },
		{ 136, 300, 43 },
		{ 136, 301, 43 },
		{ 136, 302, 43 },
		{ 136, 303, 43 },
		{ 136, 304, 43 },
		{ 136, 305, 43 },
		{ 136, 306, 43 },
		{ 136, 307, 43 },
		{ 136, 308, 43 },
		{ 136, 309, 43 },
		{ 136, 310, 43 },
		{ 136, 311, 43 },
		{ 136, 312, 43 },
		{ 136, 313, 43 },
		{ 136, 314, 43 },
		{ 136, 315, 43 },
		{ 136, 316, 43 },
		{ 136, 317, 43 },
		{ 137, 288, 43 },
		{ 137, 289, 43 },
		{ 137, 290, 43 },
		{ 137, 291, 43 },
		{ 137, 292, 43 },
		{ 137, 293, 43 },
		{ 137, 294, 43 },
		{ 137, 295, 43 },
		{ 137, 296, 43 },
		{ 137, 297, 43 },
		{ 137, 298, 43 },
		{ 137, 299, 43 },
		{ 137, 300, 43 },
		{ 137, 301, 43 },
		{ 137, 302, 43 },
		{ 137, 303, 43 },
		{ 137, 304, 43 },
		{ 137, 305, 43 },
		{ 137, 306, 43 },
		{ 137, 307, 43 },
		{ 137, 308, 43 },
		{ 137, 309, 43 },
		{ 137, 310, 43 },
		{ 137, 311, 43 },
		{ 137, 312, 43 },
		{ 137, 313, 43 },
		{ 137, 314, 43 },
		{ 137, 315, 43 },
		{ 137, 316, 43 },
		{ 138, 287, 43 },
		{ 138, 288, 43 },
		{ 138, 289, 43 },
		{ 138, 290, 43 },
		{ 138, 291, 43 },
		{ 138, 292, 43 },
		{ 138, 293, 43 },
		{ 138, 294, 43 },
		{ 138, 295, 43 },
		{ 138, 296, 43 },
		{ 138, 297, 43 },
		{ 138, 298, 43 },
		{ 138, 299, 43 },
		{ 138, 300, 43 },
		{ 138, 301, 43 },
		{ 138, 302, 43 },
		{ 138, 303, 43 },
		{ 138, 304, 43 },
		{ 138, 305, 43 },
		{ 138, 306, 43 },
		{ 138, 307, 43 },
		{ 138, 308, 43 },
		{ 138, 309, 43 },
		{ 138, 310, 43 },
		{ 138, 311, 43 },
		{ 138, 312, 43 },
		{ 138, 313, 43 },
		{ 138, 314, 43 },
		{ 138, 315, 43 },
		{ 138, 316, 43 },
		{ 139, 287, 43 },
		{ 139, 288, 43 },
		{ 139, 289, 43 },
		{ 139, 290, 43 },
		{ 139, 291, 43 },
		{ 139, 292, 43 },
		{ 139, 293, 43 },
		{ 139, 294, 43 },
		{ 139, 295, 43 },
		{ 139, 296, 43 },
		{ 139, 297, 43 },
		{ 139, 298, 43 },
		{ 139, 299, 43 },
		{ 139, 300, 43 },
		{ 139, 301, 43 },
		{ 139, 302, 43 },
		{ 139, 303, 43 },
		{ 139, 304, 43 },
		{ 139, 305, 43 },
		{ 139, 306, 43 },
		{ 139, 307, 43 },
		{ 139, 308, 43 },
		{ 139, 309, 43 },
		{ 139, 310, 43 },
		{ 139, 311, 43 },
		{ 139, 312, 43 },
		{ 139, 313, 43 },
		{ 139, 314, 43 },
		{ 139, 315, 43 },
		{ 140, 287, 43 },
		{ 140, 288, 43 },
		{ 140, 289, 43 },
		{ 140, 290, 43 },
		{ 140, 291, 43 },
		{ 140, 292, 43 },
		{ 140, 293, 43 },
		{ 140, 294, 43 },
		{ 140, 295, 43 },
		{ 140, 296, 43 },
		{ 140, 297, 43 },
		{ 140, 298, 43 },
		{ 140, 299, 43 },
		{ 140, 300, 43 },
		{ 140, 301, 43 },
		{ 140, 302, 43 },
		{ 140, 303, 43 },
		{ 140, 304, 43 },
		{ 140, 305, 43 },
		{ 140, 306, 43 },
		{ 140, 307, 43 },
		{ 140, 308, 43 },
		{ 140, 309, 43 },
		{ 140, 310, 43 },
		{ 140, 311, 43 },
		{ 140, 312, 43 },
		{ 140, 313, 43 },
		{ 140, 314, 43 },
		{ 140, 315, 43 },
		{ 141, 287, 43 },
		{ 141, 288, 43 },
		{ 141, 289, 43 },
		{ 141, 290, 43 },
		{ 141, 291, 43 },
		{ 141, 292, 43 },
		{ 141, 293, 43 },
		{ 141, 294, 43 },
		{ 141, 295, 43 },
		{ 141, 296, 43 },
		{ 141, 297, 43 },
		{ 141, 298, 43 },
		{ 141, 299, 43 },
		{ 141, 300, 43 },
		{ 141, 301, 43 },
		{ 141, 302, 43 },
		{ 141, 303, 43 },
		{ 141, 304, 43 },
		{ 141, 305, 43 },
		{ 141, 306, 43 },
		{ 141, 307, 43 },
		{ 141, 308, 43 },
		{ 141, 309, 43 },
		{ 141, 310, 43 },
		{ 141, 311, 43 },
		{ 141, 312, 43 },
		{ 141, 313, 43 },
		{ 141, 314, 43 },
		{ 142, 286, 43 },
		{ 142, 287, 43 },
		{ 142, 288, 43 },
		{ 142, 289, 43 },
		{ 142, 290, 43 },
		{ 142, 291, 43 },
		{ 142, 292, 43 },
		{ 142, 293, 43 },
		{ 142, 294, 43 },
		{ 142, 295, 43 },
		{ 142, 296, 43 },
		{ 142, 297, 43 },
		{ 142, 298, 43 },
		{ 142, 299, 43 },
		{ 142, 300, 43 },
		{ 142, 301, 43 },
		{ 142, 302, 43 },
		{ 142, 303, 43 },
		{ 142, 304, 43 },
		{ 142, 305, 43 },
		{ 142, 306, 43 },
		{ 142, 307, 43 },
		{ 142, 308, 43 },
		{ 142, 309, 43 },
		{ 142, 310, 43 },
		{ 142, 311, 43 },
		{ 142, 312, 43 },
		{ 142, 313, 43 },
		{ 142, 314, 43 },
		{ 143, 286, 43 },
		{ 143, 287, 43 },
		{ 143, 288, 43 },
		{ 143, 289, 43 },
		{ 143, 290, 43 },
		{ 143, 291, 43 },
		{ 143, 292, 43 },
		{ 143, 293, 43 },
		{ 143, 294, 43 },
		{ 143, 295, 43 },
		{ 143, 296, 43 },
		{ 143, 297, 43 },
		{ 143, 298, 43 },
		{ 143, 299, 43 },
		{ 143, 300, 43 },
		{ 143, 301, 43 },
		{ 143, 302, 43 },
		{ 143, 303, 43 },
		{ 143, 304, 43 },
		{ 143, 305, 43 },
		{ 143, 306, 43 },
		{ 143, 307, 43 },
		{ 143, 308, 43 },
		{ 143, 309, 43 },
		{ 143, 310, 43 },
		{ 143, 311, 43 },
		{ 143, 312, 43 },
		{ 143, 313, 43 },
		{ 144, 286, 43 },
		{ 144, 287, 43 },
		{ 144, 288, 43 },
		{ 144, 289, 43 },
		{ 144, 290, 43 },
		{ 144, 291, 43 },
		{ 144, 292, 43 },
		{ 144, 293, 43 },
		{ 144, 294, 43 },
		{ 144, 295, 43 },
		{ 144, 296, 43 },
		{ 144, 297, 43 },
		{ 144, 298, 43 },
		{ 144, 299, 43 },
		{ 144, 300, 43 },
		{ 144, 301, 43 },
		{ 144, 302, 43 },
		{ 144, 303, 43 },
		{ 144, 304, 43 },
		{ 144, 305, 43 },
		{ 144, 306, 43 },
		{ 144, 307, 43 },
		{ 144, 308, 43 },
		{ 144, 309, 43 },
		{ 144, 310, 43 },
		{ 144, 311, 43 },
		{ 144, 312, 43 },
		{ 144, 313, 43 },
		{ 145, 286, 43 },
		{ 145, 287, 43 },
		{ 145, 288, 43 },
		{ 145, 289, 43 },
		{ 145, 290, 43 },
		{ 145, 291, 43 },
		{ 145, 292, 43 },
		{ 145, 293, 43 },
		{ 145, 294, 43 },
		{ 145, 295, 43 },
		{ 145, 296, 43 },
		{ 145, 297, 43 },
		{ 145, 298, 43 },
		{ 145, 299, 43 },
		{ 145, 300, 43 },
		{ 145, 301, 43 },
		{ 145, 302, 43 },
		{ 145, 303, 43 },
		{ 145, 304, 43 },
		{ 145, 305, 43 },
		{ 145, 306, 43 },
		{ 145, 307, 43 },
		{ 145, 308, 43 },
		{ 145, 309, 43 },
		{ 145, 310, 43 },
		{ 145, 311, 43 },
		{ 145, 312, 43 },
		{ 146, 285, 43 },
		{ 146, 286, 43 },
		{ 146, 287, 43 },
		{ 146, 288, 43 },
		{ 146, 289, 43 },
		{ 146, 290, 43 },
		{ 146, 291, 43 },
		{ 146, 292, 43 },
		{ 146, 293, 43 },
		{ 146, 294, 43 },
		{ 146, 295, 43 },
		{ 146, 296, 43 },
		{ 146, 297, 43 },
		{ 146, 298, 43 },
		{ 146, 299, 43 },
		{ 146, 300, 43 },
		{ 146, 301, 43 },
		{ 146, 302, 43 },
		{ 146, 303, 43 },
		{ 146, 304, 43 },
		{ 146, 305, 43 },
		{ 146, 306, 43 },
		{ 146, 307, 43 },
		{ 146, 308, 43 },
		{ 146, 309, 43 },
		{ 146, 310, 43 },
		{ 146, 311, 43 },
		{ 146, 312, 43 },
		{ 147, 285, 43 },
		{ 147, 286, 43 },
		{ 147, 287, 43 },
		{ 147, 288, 43 },
		{ 147, 289, 43 },
		{ 147, 290, 43 },
		{ 147, 291, 43 },
		{ 147, 292, 43 },
		{ 147, 293, 43 },
		{ 147, 294, 43 },
		{ 147, 295, 43 },
		{ 147, 296, 43 },
		{ 147, 297, 43 },
		{ 147, 298, 43 },
		{ 147, 299, 43 },
		{ 147, 300, 43 },
		{ 147, 301, 43 },
		{ 147, 302, 43 },
		{ 147, 303, 43 },
		{ 147, 304, 43 },
		{ 147, 305, 43 },
		{ 147, 306, 43 },
		{ 147, 307, 43 },
		{ 147, 308, 43 },
		{ 147, 309, 43 },
		{ 147, 310, 43 },
		{ 147, 311, 43 },
		{ 148, 285, 43 },
		{ 148, 286, 43 },
		{ 148, 287, 43 },
		{ 148, 288, 43 },
		{ 148, 289, 43 },
		{ 148, 290, 43 },
		{ 148, 291, 43 },
		{ 148, 292, 43 },
		{ 148, 293, 43 },
		{ 148, 294, 43 },
		{ 148, 295, 43 },
		{ 148, 296, 43 },
		{ 148, 297, 43 },
		{ 148, 298, 43 },
		{ 148, 299, 43 },
		{ 148, 300, 43 },
		{ 148, 301, 43 },
		{ 148, 302, 43 },
		{ 148, 303, 43 },
		{ 148, 304, 43 },
		{ 148, 305, 43 },
		{ 148, 306, 43 },
		{ 148, 307, 43 },
		{ 148, 308, 43 },
		{ 148, 309, 43 },
		{ 148, 310, 43 },
		{ 148, 311, 43 },
		{ 149, 285, 43 },
		{ 149, 286, 43 },
		{ 149, 287, 43 },
		{ 149, 288, 43 },
		{ 149, 289, 43 },
		{ 149, 290, 43 },
		{ 149, 291, 43 },
		{ 149, 292, 43 },
		{ 149, 293, 43 },
		{ 149, 294, 43 },
		{ 149, 295, 43 },
		{ 149, 296, 43 },
		{ 149, 297, 43 },
		{ 149, 298, 43 },
		{ 149, 299, 43 },
		{ 149, 300, 43 },
		{ 149, 301, 43 },
		{ 149, 302, 43 },
		{ 149, 303, 43 },
		{ 149, 304, 43 },
		{ 149, 305, 43 },
		{ 149, 306, 43 },
		{ 149, 307, 43 },
		{ 149, 308, 43 },
		{ 149, 309, 43 },
		{ 149, 310, 43 },
		{ 150, 284, 43 },
		{ 150, 285, 43 },
		{ 150, 286, 43 },
		{ 150, 287, 43 },
		{ 150, 288, 43 },
		{ 150, 289, 43 },
		{ 150, 290, 43 },
		{ 150, 291, 43 },
		{ 150, 292, 43 },
		{ 150, 293, 43 },
		{ 150, 294, 43 },
		{ 150, 295, 43 },
		{ 150, 296, 43 },
		{ 150, 297, 43 },
		{ 150, 298, 43 },
		{ 150, 299, 43 },
		{ 150, 300, 43 },
		{ 150, 301, 43 },
		{ 150, 302, 43 },
		{ 150, 303, 43 },
		{ 150, 304, 43 },
		{ 150, 305, 43 },
		{ 150, 306, 43 },
		{ 150, 307, 43 },
		{ 150, 308, 43 },
		{ 150, 309, 43 },
		{ 150, 310, 43 },
		{ 151, 284, 43 },
		{ 151, 285, 43 },
		{ 151, 286, 43 },
		{ 151, 287, 43 },
		{ 151, 288, 43 },
		{ 151, 289, 43 },
		{ 151, 290, 43 },
		{ 151, 291, 43 },
		{ 151, 292, 43 },
		{ 151, 293, 43 },
		{ 151, 294, 43 },
		{ 151, 295, 43 },
		{ 151, 296, 43 },
		{ 151, 297, 43 },
		{ 151, 298, 43 },
		{ 151, 299, 43 },
		{ 151, 300, 43 },
		{ 151, 301, 43 },
		{ 151, 302, 43 },
		{ 151, 303, 43 },
		{ 151, 304, 43 },
		{ 151, 305, 43 },
		{ 151, 306, 43 },
		{ 151, 307, 43 },
		{ 151, 308, 43 },
		{ 151, 309, 43 },
		{ 152, 284, 43 },
		{ 152, 285, 43 },
		{ 152, 286, 43 },
		{ 152, 287, 43 },
		{ 152, 288, 43 },
		{ 152, 289, 43 },
		{ 152, 290, 43 },
		{ 152, 291, 43 },
		{ 152, 292, 43 },
		{ 152, 293, 43 },
		{ 152, 294, 43 },
		{ 152, 295, 43 },
		{ 152, 296, 43 },
		{ 152, 297, 43 },
		{ 152, 298, 43 },
		{ 152, 299, 43 },
		{ 152, 300, 43 },
		{ 152, 301, 43 },
		{ 152, 302, 43 },
		{ 152, 303, 43 },
		{ 152, 304, 43 },
		{ 152, 305, 43 },
		{ 152, 306, 43 },
		{ 152, 307, 43 },
		{ 152, 308, 43 },
		{ 152, 309, 43 },
		{ 153, 284, 43 },
		{ 153, 285, 43 },
		{ 153, 286, 43 },
		{ 153, 287, 43 },
		{ 153, 288, 43 },
		{ 153, 289, 43 },
		{ 153, 290, 43 },
		{ 153, 291, 43 },
		{ 153, 292, 43 },
		{ 153, 293, 43 },
		{ 153, 294, 43 },
		{ 153, 295, 43 },
		{ 153, 296, 43 },
		{ 153, 297, 43 },
		{ 153, 298, 43 },
		{ 153, 299, 43 },
		{ 153, 300, 43 },
		{ 153, 301, 43 },
		{ 153, 302, 43 },
		{ 153, 303, 43 },
		{ 153, 304, 43 },
		{ 153, 305, 43 },
		{ 153, 306, 43 },
		{ 153, 307, 43 },
		{ 153, 308, 43 },
		{ 154, 283, 43 },
		{ 154, 284, 43 },
		{ 154, 285, 43 },
		{ 154, 286, 43 },
		{ 154, 287, 43 },
		{ 154, 288, 43 },
		{ 154, 289, 43 },
		{ 154, 290, 43 },
		{ 154, 291, 43 },
		{ 154, 292, 43 },
		{ 154, 293, 43 },
		{ 154, 294, 43 },
		{ 154, 295, 43 },
		{ 154, 296, 43 },
		{ 154, 297, 43 },
		{ 154, 298, 43 },
		{ 154, 299, 43 },
		{ 154, 300, 43 },
		{ 154, 301, 43 },
		{ 154, 302, 43 },
		{ 154, 303, 43 },
		{ 154, 304, 43 },
		{ 154, 305, 43 },
		{ 154, 306, 43 },
		{ 154, 307, 43 },
		{ 154, 308, 43 },
		{ 155, 283, 43 },
		{ 155, 284, 43 },
		{ 155, 285, 43 },
		{ 155, 286, 43 },
		{ 155, 287, 43 },
		{ 155, 288, 43 },
		{ 155, 289, 43 },
		{ 155, 290, 43 },
		{ 155, 291, 43 },
		{ 155, 292, 43 },
		{ 155, 293, 43 },
		{ 155, 294, 43 },
		{ 155, 295, 43 },
		{ 155, 296, 43 },
		{ 155, 297, 43 },
		{ 155, 298, 43 },
		{ 155, 299, 43 },
		{ 155, 300, 43 },
		{ 155, 301, 43 },
		{ 155, 302, 43 },
		{ 155, 303, 43 },
		{ 155, 304, 43 },
		{ 155, 305, 43 },
		{ 155, 306, 43 },
		{ 155, 307, 43 },
		{ 156, 283, 43 },
		{ 156, 284, 43 },
		{ 156, 285, 43 },
		{ 156, 286, 43 },
		{ 156, 287, 43 },
		{ 156, 288, 43 },
		{ 156, 289, 43 },
		{ 156, 290, 43 },
		{ 156, 291, 43 },
		{ 156, 292, 43 },
		{ 156, 293, 43 },
		{ 156, 294, 43 },
		{ 156, 295, 43 },
		{ 156, 296, 43 },
		{ 156, 297, 43 },
		{ 156, 298, 43 },
		{ 156, 299, 43 },
		{ 156, 300, 43 },
		{ 156, 301, 43 },
		{ 156, 302, 43 },
		{ 156, 303, 43 },
		{ 156, 304, 43 },
		{ 156, 305, 43 },
		{ 156, 306, 43 },
		{ 156, 307, 43 },
		{ 157, 283, 43 },
		{ 157, 284, 43 },
		{ 157, 285, 43 },
		{ 157, 286, 43 },
		{ 157, 287, 43 },
		{ 157, 288, 43 },
		{ 157, 289, 43 },
		{ 157, 290, 43 },
		{ 157, 291, 43 },
		{ 157, 292, 43 },
		{ 157, 293, 43 },
		{ 157, 294, 43 },
		{ 157, 295, 43 },
		{ 157, 296, 43 },
		{ 157, 297, 43 },
		{ 157, 298, 43 },
		{ 157, 299, 43 },
		{ 157, 300, 43 },
		{ 157, 301, 43 },
		{ 157, 302, 43 },
		{ 157, 303, 43 },
		{ 157, 304, 43 },
		{ 157, 305, 43 },
		{ 157, 306, 43 },
		{ 158, 282, 43 },
		{ 158, 283, 43 },
		{ 158, 284, 43 },
		{ 158, 285, 43 },
		{ 158, 286, 43 },
		{ 158, 287, 43 },
		{ 158, 288, 43 },
		{ 158, 289, 43 },
		{ 158, 290, 43 },
		{ 158, 291, 43 },
		{ 158, 292, 43 },
		{ 158, 293, 43 },
		{ 158, 294, 43 },
		{ 158, 295, 43 },
		{ 158, 296, 43 },
		{ 158, 297, 43 },
		{ 158, 298, 43 },
		{ 158, 299, 43 },
		{ 158, 300, 43 },
		{ 158, 301, 43 },
		{ 158, 302, 43 },
		{ 158, 303, 43 },
		{ 158, 304, 43 },
		{ 158, 305, 43 },
		{ 158, 306, 43 },
		{ 159, 282, 43 },
		{ 159, 283, 43 },
		{ 159, 284, 43 },
		{ 159, 285, 43 },
		{ 159, 286, 43 },
		{ 159, 287, 43 },
		{ 159, 288, 43 },
		{ 159, 289, 43 },
		{ 159, 290, 43 },
		{ 159, 291, 43 },
		{ 159, 292, 43 },
		{ 159, 293, 43 },
		{ 159, 294, 43 },
		{ 159, 295, 43 },
		{ 159, 296, 43 },
		{ 159, 297, 43 },
		{ 159, 298, 43 },
		{ 159, 299, 43 },
		{ 159, 300, 43 },
		{ 159, 301, 43 },
		{ 159, 302, 43 },
		{ 159, 303, 43 },
		{ 159, 304, 43 },
		{ 159, 305, 43 },
		{ 160, 282, 43 },
		{ 160, 283, 43 },
		{ 160, 284, 43 },
		{ 160, 285, 43 },
		{ 160, 286, 43 },
		{ 160, 287, 43 },
		{ 160, 288, 43 },
		{ 160, 289, 43 },
		{ 160, 290, 43 },
		{ 160, 291, 43 },
		{ 160, 292, 43 },
		{ 160, 293, 43 },
		{ 160, 294, 43 },
		{ 160, 295, 43 },
		{ 160, 296, 43 },
		{ 160, 297, 43 },
		{ 160, 298, 43 },
		{ 160, 299, 43 },
		{ 160, 300, 43 },
		{ 160, 301, 43 },
		{ 160, 302, 43 },
		{ 160, 303, 43 },
		{ 160, 304, 43 },
		{ 160, 305, 43 },
		{ 161, 282, 43 },
		{ 161, 283, 43 },
		{ 161, 284, 43 },
		{ 161, 285, 43 },
		{ 161, 286, 43 },
		{ 161, 287, 43 },
		{ 161, 288, 43 },
		{ 161, 289, 43 },
		{ 161, 290, 43 },
		{ 161, 291, 43 },
		{ 161, 292, 43 },
		{ 161, 293, 43 },
		{ 161, 294, 43 },
		{ 161, 295, 43 },
		{ 161, 296, 43 },
		{ 161, 297, 43 },
		{ 161, 298, 43 },
		{ 161, 299, 43 },
		{ 161, 300, 43 },
		{ 161, 301, 43 },
		{ 161, 302, 43 },
		{ 161, 303, 43 },
		{ 161, 304, 43 },
		{ 162, 281, 43 },
		{ 162, 282, 43 },
		{ 162, 283, 43 },
		{ 162, 284, 43 },
		{ 162, 285, 43 },
		{ 162, 286, 43 },
		{ 162, 287, 43 },
		{ 162, 288, 43 },
		{ 162, 289, 43 },
		{ 162, 290, 43 },
		{ 162, 291, 43 },
		{ 162, 292, 43 },
		{ 162, 293, 43 },
		{ 162, 294, 43 },
		{ 162, 295, 43 },
		{ 162, 296, 43 },
		{ 162, 297, 43 },
		{ 162, 298, 43 },
		{ 162, 299, 43 },
		{ 162, 300, 43 },
		{ 162, 301, 43 },
		{ 162, 302, 43 },
		{ 162, 303, 43 },
		{ 162, 304, 43 },
		{ 163, 281, 43 },
		{ 163, 282, 43 },
		{ 163, 283, 43 },
		{ 163, 284, 43 },
		{ 163, 285, 43 },
		{ 163, 286, 43 },
		{ 163, 287, 43 },
		{ 163, 288, 43 },
		{ 163, 289, 43 },
		{ 163, 290, 43 },
		{ 163, 291, 43 },
		{ 163, 292, 43 },
		{ 163, 293, 43 },
		{ 163, 294, 43 },
		{ 163, 295, 43 },
		{ 163, 296, 43 },
		{ 163, 297, 43 },
		{ 163, 298, 43 },
		{ 163, 299, 43 },
		{ 163, 300, 43 },
		{ 163, 301, 43 },
		{ 163, 302, 43 },
		{ 163, 303, 43 },
		{ 164, 281, 43 },
		{ 164, 282, 43 },
		{ 164, 283, 43 },
		{ 164, 284, 43 },
		{ 164, 285, 43 },
		{ 164, 286, 43 },
		{ 164, 287, 43 },
		{ 164, 288, 43 },
		{ 164, 289, 43 },
		{ 164, 290, 43 },
		{ 164, 291, 43 },
		{ 164, 292, 43 },
		{ 164, 293, 43 },
		{ 164, 294, 43 },
		{ 164, 295, 43 },
		{ 164, 296, 43 },
		{ 164, 297, 43 },
		{ 164, 298, 43 },
		{ 164, 299, 43 },
		{ 164, 300, 43 },
		{ 164, 301, 43 },
		{ 164, 302, 43 },
		{ 164, 303, 43 },
		{ 165, 281, 43 },
		{ 165, 282, 43 },
		{ 165, 283, 43 },
		{ 165, 284, 43 },
		{ 165, 285, 43 },
		{ 165, 286, 43 },
		{ 165, 287, 43 },
		{ 165, 288, 43 },
		{ 165, 289, 43 },
		{ 165, 290, 43 },
		{ 165, 291, 43 },
		{ 165, 292, 43 },
		{ 165, 293, 43 },
		{ 165, 294, 43 },
		{ 165, 295, 43 },
		{ 165, 296, 43 },
		{ 165, 297, 43 },
		{ 165, 298, 43 },
		{ 165, 299, 43 },
		{ 165, 300, 43 },
		{ 165, 301, 43 },
		{ 165, 302, 43 },
		{ 166, 280, 43 },
		{ 166, 281, 43 },
		{ 166, 282, 43 },
		{ 166, 283, 43 },
		{ 166, 284, 43 },
		{ 166, 285, 43 },
		{ 166, 286, 43 },
		{ 166, 287, 43 },
		{ 166, 288, 43 },
		{ 166, 289, 43 },
		{ 166, 290, 43 },
		{ 166, 291, 43 },
		{ 166, 292, 43 },
		{ 166, 293, 43 },
		{ 166, 294, 43 },
		{ 166, 295, 43 },
		{ 166, 296, 43 },
		{ 166, 297, 43 },
		{ 166, 298, 43 },
		{ 166, 299, 43 },
		{ 166, 300, 43 },
		{ 166, 301, 43 },
		{ 166, 302, 43 },
		{ 167, 280, 43 },
		{ 167, 281, 43 },
		{ 167, 282, 43 },
		{ 167, 283, 43 },
		{ 167, 284, 43 },
		{ 167, 285, 43 },
		{ 167, 286, 43 },
		{ 167, 287, 43 },
		{ 167, 288, 43 },
		{ 167, 289, 43 },
		{ 167, 290, 43 },
		{ 167, 291, 43 },
		{ 167, 292, 43 },
		{ 167, 293, 43 },
		{ 167, 294, 43 },
		{ 167, 295, 43 },
		{ 167, 296, 43 },
		{ 167, 297, 43 },
		{ 167, 298, 43 },
		{ 167, 299, 43 },
		{ 167, 300, 43 },
		{ 167, 301, 43 },
		{ 168, 280, 43 },
		{ 168, 281, 43 },
		{ 168, 282, 43 },
		{ 168, 283, 43 },
		{ 168, 284, 43 },
		{ 168, 285, 43 },
		{ 168, 286, 43 },
		{ 168, 287, 43 },
		{ 168, 288, 43 },
		{ 168, 289, 43 },
		{ 168, 290, 43 },
		{ 168, 291, 43 },
		{ 168, 292, 43 },
		{ 168, 293, 43 },
		{ 168, 294, 43 },
		{ 168, 295, 43 },
		{ 168, 296, 43 },
		{ 168, 297, 43 },
		{ 168, 298, 43 },
		{ 168, 299, 43 },
		{ 168, 300, 43 },
		{ 168, 301, 43 },
		{ 169, 280, 43 },
		{ 169, 281, 43 },
		{ 169, 282, 43 },
		{ 169, 283, 43 },
		{ 169, 284, 43 },
		{ 169, 285, 43 },
		{ 169, 286, 43 },
		{ 169, 287, 43 },
		{ 169, 288, 43 },
		{ 169, 289, 43 },
		{ 169, 290, 43 },
		{ 169, 291, 43 },
		{ 169, 292, 43 },
		{ 169, 293, 43 },
		{ 169, 294, 43 },
		{ 169, 295, 43 },
		{ 169, 296, 43 },
		{ 169, 297, 43 },
		{ 169, 298, 43 },
		{ 169, 299, 43 },
		{ 169, 300, 43 },
		{ 170, 279, 43 },
		{ 170, 280, 43 },
		{ 170, 281, 43 },
		{ 170, 282, 43 },
		{ 170, 283, 43 },
		{ 170, 284, 43 },
		{ 170, 285, 43 },
		{ 170, 286, 43 },
		{ 170, 287, 43 },
		{ 170, 288, 43 },
		{ 170, 289, 43 },
		{ 170, 290, 43 },
		{ 170, 291, 43 },
		{ 170, 292, 43 },
		{ 170, 293, 43 },
		{ 170, 294, 43 },
		{ 170, 295, 43 },
		{ 170, 296, 43 },
		{ 170, 297, 43 },
		{ 170, 298, 43 },
		{ 170, 299, 43 },
		{ 170, 300, 43 },
		{ 171, 279, 43 },
		{ 171, 280, 43 },
		{ 171, 281, 43 },
		{ 171, 282, 43 },
		{ 171, 283, 43 },
		{ 171, 284, 43 },
		{ 171, 285, 43 },
		{ 171, 286, 43 },
		{ 171, 287, 43 },
		{ 171, 288, 43 },
		{ 171, 289, 43 },
		{ 171, 290, 43 },
		{ 171, 291, 43 },
		{ 171, 292, 43 },
		{ 171, 293, 43 },
		{ 171, 294, 43 },
		{ 171, 295, 43 },
		{ 171, 296, 43 },
		{ 171, 297, 43 },
		{ 171, 298, 43 },
		{ 171, 299, 43 },
		{ 172, 279, 43 },
		{ 172, 280, 43 },
		{ 172, 281, 43 },
		{ 172, 282, 43 },
		{ 172, 283, 43 },
		{ 172, 284, 43 },
		{ 172, 285, 43 },
		{ 172, 286, 43 },
		{ 172, 287, 43 },
		{ 172, 288, 43 },
		{ 172, 289, 43 },
		{ 172, 290, 43 },
		{ 172, 291, 43 },
		{ 172, 292, 43 },
		{ 172, 293, 43 },
		{ 172, 294, 43 },
		{ 172, 295, 43 },
		{ 172, 296, 43 },
		{ 172, 297, 43 },
		{ 172, 298, 43 },
		{ 172, 299, 43 },
		{ 173, 279, 43 },
		{ 173, 280, 43 },
		{ 173, 281, 43 },
		{ 173, 282, 43 },
		{ 173, 283, 43 },
		{ 173, 284, 43 },
		{ 173, 285, 43 },
		{ 173, 286, 43 },
		{ 173, 287, 43 },
		{ 173, 288, 43 },
		{ 173, 289, 43 },
		{ 173, 290, 43 },
		{ 173, 291, 43 },
		{ 173, 292, 43 },
		{ 173, 293, 43 },
		{ 173, 294, 43 },
		{ 173, 295, 43 },
		{ 173, 296, 43 },
		{ 173, 297, 43 },
		{ 173, 298, 43 },
		{ 174, 278, 43 },
		{ 174, 279, 43 },
		{ 174, 280, 43 },
		{ 174, 281, 43 },
		{ 174, 282, 43 },
		{ 174, 283, 43 },
		{ 174, 284, 43 },
		{ 174, 285, 43 },
		{ 174, 286, 43 },
		{ 174, 287, 43 },
		{ 174, 288, 43 },
		{ 174, 289, 43 },
		{ 174, 290, 43 },
		{ 174, 291, 43 },
		{ 174, 292, 43 },
		{ 174, 293, 43 },
		{ 174, 294, 43 },
		{ 174, 295, 43 },
		{ 174, 296, 43 },
		{ 174, 297, 43 },
		{ 174, 298, 43 },
		{ 175, 278, 43 },
		{ 175, 279, 43 },
		{ 175, 280, 43 },
		{ 175, 281, 43 },
		{ 175, 282, 43 },
		{ 175, 283, 43 },
		{ 175, 284, 43 },
		{ 175, 285, 43 },
		{ 175, 286, 43 },
		{ 175, 287, 43 },
		{ 175, 288, 43 },
		{ 175, 289, 43 },
		{ 175, 290, 43 },
		{ 175, 291, 43 },
		{ 175, 292, 43 },
		{ 175, 293, 43 },
		{ 175, 294, 43 },
		{ 175, 295, 43 },
		{ 175, 296, 43 },
		{ 175, 297, 43 },
		{ 176, 278, 43 },
		{ 176, 279, 43 },
		{ 176, 280, 43 },
		{ 176, 281, 43 },
		{ 176, 282, 43 },
		{ 176, 283, 43 },
		{ 176, 284, 43 },
		{ 176, 285, 43 },
		{ 176, 286, 43 },
		{ 176, 287, 43 },
		{ 176, 288, 43 },
		{ 176, 289, 43 },
		{ 176, 290, 43 },
		{ 176, 291, 43 },
		{ 176, 292, 43 },
		{ 176, 293, 43 },
		{ 176, 294, 43 },
		{ 176, 295, 43 },
		{ 176, 296, 43 },
		{ 176, 297, 43 },
		{ 177, 278, 43 },
		{ 177, 279, 43 },
		{ 177, 280, 43 },
		{ 177, 281, 43 },
		{ 177, 282, 43 },
		{ 177, 283, 43 },
		{ 177, 284, 43 },
		{ 177, 285, 43 },
		{ 177, 286, 43 },
		{ 177, 287, 43 },
		{ 177, 288, 43 },
		{ 177, 289, 43 },
		{ 177, 290, 43 },
		{ 177, 291, 43 },
		{ 177, 292, 43 },
		{ 177, 293, 43 },
		{ 177, 294, 43 },
		{ 177, 295, 43 },
		{ 177, 296, 43 },
		{ 178, 277, 43 },
		{ 178, 278, 43 },
		{ 178, 279, 43 },
		{ 178, 280, 43 },
		{ 178, 281, 43 },
		{ 178, 282, 43 },
		{ 178, 283, 43 },
		{ 178, 284, 43 },
		{ 178, 285, 43 },
		{ 178, 286, 43 },
		{ 178, 287, 43 },
		{ 178, 288, 43 },
		{ 178, 289, 43 },
		{ 178, 290, 43 },
		{ 178, 291, 43 },
		{ 178, 292, 43 },
		{ 178, 293, 43 },
		{ 178, 294, 43 },
		{ 178, 295, 43 },
		{ 178, 296, 43 },
		{ 179, 277, 43 },
		{ 179, 278, 43 },
		{ 179, 279, 43 },
		{ 179, 280, 43 },
		{ 179, 281, 43 },
		{ 179, 282, 43 },
		{ 179, 283, 43 },
		{ 179, 284, 43 },
		{ 179, 285, 43 },
		{ 179, 286, 43 },
		{ 179, 287, 43 },
		{ 179, 288, 43 },
		{ 179, 289, 43 },
		{ 179, 290, 43 },
		{ 179, 291, 43 },
		{ 179, 292, 43 },
		{ 179, 293, 43 },
		{ 179, 294, 43 },
		{ 179, 295, 43 },
		{ 180, 277, 43 },
		{ 180, 278, 43 },
		{ 180, 279, 43 },
		{ 180, 280, 43 },
		{ 180, 281, 43 },
		{ 180, 282, 43 },
		{ 180, 283, 43 },
		{ 180, 284, 43 },
		{ 180, 285, 43 },
		{ 180, 286, 43 },
		{ 180, 287, 43 },
		{ 180, 288, 43 },
		{ 180, 289, 43 },
		{ 180, 290, 43 },
		{ 180, 291, 43 },
		{ 180, 292, 43 },
		{ 180, 293, 43 },
		{ 180, 294, 43 },
		{ 180, 295, 43 },
		{ 181, 277, 43 },
		{ 181, 278, 43 },
		{ 181, 279, 43 },
		{ 181, 280, 43 },
		{ 181, 281, 43 },
		{ 181, 282, 43 },
		{ 181, 283, 43 },
		{ 181, 284, 43 },
		{ 181, 285, 43 },
		{ 181, 286, 43 },
		{ 181, 287, 43 },
		{ 181, 288, 43 },
		{ 181, 289, 43 },
		{ 181, 290, 43 },
		{ 181, 291, 43 },
		{ 181, 292, 43 },
		{ 181, 293, 43 },
		{ 181, 294, 43 },
		{ 182, 276, 43 },
		{ 182, 277, 43 },
		{ 182, 278, 43 },
		{ 182, 279, 43 },
		{ 182, 280, 43 },
		{ 182, 281, 43 },
		{ 182, 282, 43 },
		{ 182, 283, 43 },
		{ 182, 284, 43 },
		{ 182, 285, 43 },
		{ 182, 286, 43 },
		{ 182, 287, 43 },
		{ 182, 288, 43 },
		{ 182, 289, 43 },
		{ 182, 290, 43 },
		{ 182, 291, 43 },
		{ 182, 292, 43 },
		{ 182, 293, 43 },
		{ 182, 294, 43 },
		{ 183, 276, 43 },
		{ 183, 277, 43 },
		{ 183, 278, 43 },
		{ 183, 279, 43 },
		{ 183, 280, 43 },
		{ 183, 281, 43 },
		{ 183, 282, 43 },
		{ 183, 283, 43 },
		{ 183, 284, 43 },
		{ 183, 285, 43 },
		{ 183, 286, 43 },
		{ 183, 287, 43 },
		{ 183, 288, 43 },
		{ 183, 289, 43 },
		{ 183, 290, 43 },
		{ 183, 291, 43 },
		{ 183, 292, 43 },
		{ 183, 293, 43 },
		{ 184, 276, 43 },
		{ 184, 277, 43 },
		{ 184, 278, 43 },
		{ 184, 279, 43 },
		{ 184, 280, 43 },
		{ 184, 281, 43 },
		{ 184, 282, 43 },
		{ 184, 283, 43 },
		{ 184, 284, 43 },
		{ 184, 285, 43 },
		{ 184, 286, 43 },
		{ 184, 287, 43 },
		{ 184, 288, 43 },
		{ 184, 289, 43 },
		{ 184, 290, 43 },
		{ 184, 291, 43 },
		{ 184, 292, 43 },
		{ 184, 293, 43 },
		{ 185, 276, 43 },
		{ 185, 277, 43 },
		{ 185, 278, 43 },
		{ 185, 279, 43 },
		{ 185, 280, 43 },
		{ 185, 281, 43 },
		{ 185, 282, 43 },
		{ 185, 283, 43 },
		{ 185, 284, 43 },
		{ 185, 285, 43 },
		{ 185, 286, 43 },
		{ 185, 287, 43 },
		{ 185, 288, 43 },
		{ 185, 289, 43 },
		{ 185, 290, 43 },
		{ 185, 291, 43 },
		{ 185, 292, 43 },
		{ 186, 275, 43 },
		{ 186, 276, 43 },
		{ 186, 277, 43 },
		{ 186, 278, 43 },
		{ 186, 279, 43 },
		{ 186, 280, 43 },
		{ 186, 281, 43 },
		{ 186, 282, 43 },
		{ 186, 283, 43 },
		{ 186, 284, 43 },
		{ 186, 285, 43 },
		{ 186, 286, 43 },
		{ 186, 287, 43 },
		{ 186, 288, 43 },
		{ 186, 289, 43 },
		{ 186, 290, 43 },
		{ 186, 291, 43 },
		{ 186, 292, 43 },
		{ 187, 275, 43 },
		{ 187, 276, 43 },
		{ 187, 277, 43 },
		{ 187, 278, 43 },
		{ 187, 279, 43 },
		{ 187, 280, 43 },
		{ 187, 281, 43 },
		{ 187, 282, 43 },
		{ 187, 283, 43 },
		{ 187, 284, 43 },
		{ 187, 285, 43 },
		{ 187, 286, 43 },
		{ 187, 287, 43 },
		{ 187, 288, 43 },
		{ 187, 289, 43 },
		{ 187, 290, 43 },
		{ 187, 291, 43 },
		{ 188, 275, 43 },
		{ 188, 276, 43 },
		{ 188, 277, 43 },
		{ 188, 278, 43 },
		{ 188, 279, 43 },
		{ 188, 280, 43 },
		{ 188, 281, 43 },
		{ 188, 282, 43 },
		{ 188, 283, 43 },
		{ 188, 284, 43 },
		{ 188, 285, 43 },
		{ 188, 286, 43 },
		{ 188, 287, 43 },
		{ 188, 288, 43 },
		{ 188, 289, 43 },
		{ 188, 290, 43 },
		{ 188, 291, 43 },
		{ 189, 275, 43 },
		{ 189, 276, 43 },
		{ 189, 277, 43 },
		{ 189, 278, 43 },
		{ 189, 279, 43 },
		{ 189, 280, 43 },
		{ 189, 281, 43 },
		{ 189, 282, 43 },
		{ 189, 283, 43 },
		{ 189, 284, 43 },
		{ 189, 285, 43 },
		{ 189, 286, 43 },
		{ 189, 287, 43 },
		{ 189, 288, 43 },
		{ 189, 289, 43 },
		{ 189, 290, 43 },
		{ 190, 275, 43 },
		{ 190, 276, 43 },
		{ 190, 277, 43 },
		{ 190, 278, 43 },
		{ 190, 279, 43 },
		{ 190, 280, 43 },
		{ 190, 281, 43 },
		{ 190, 282, 43 },
		{ 190, 283, 43 },
		{ 190, 284, 43 },
		{ 190, 285, 43 },
		{ 190, 286, 43 },
		{ 190, 287, 43 },
		{ 190, 288, 43 },
		{ 190, 289, 43 },
		{ 191, 276, 43 },
		{ 191, 277, 43 },
		{ 191, 278, 43 },
		{ 191, 279, 43 },
		{ 191, 280, 43 },
		{ 191, 281, 43 },
		{ 191, 282, 43 },
		{ 191, 283, 43 },
		{ 191, 284, 43 },
		{ 191, 285, 43 },
		{ 191, 286, 43 },
		{ 191, 287, 43 },
		{ 191, 288, 43 },
		{ 323, 226, 43 },
		{ 323, 227, 43 },
		{ 323, 228, 43 },
		{ 323, 229, 43 },
		{ 323, 230, 43 },
		{ 323, 231, 43 },
		{ 323, 232, 43 },
		{ 323, 233, 43 },
		{ 323, 234, 43 },
		{ 323, 235, 43 },
		{ 323, 236, 43 },
		{ 323, 237, 43 },
		{ 323, 238, 43 },
		{ 324, 225, 43 },
		{ 324, 226, 43 },
		{ 324, 227, 43 },
		{ 324, 228, 43 },
		{ 324, 229, 43 },
		{ 324, 230, 43 },
		{ 324, 231, 43 },
		{ 324, 232, 43 },
		{ 324, 233, 43 },
		{ 324, 234, 43 },
		{ 324, 235, 43 },
		{ 324, 236, 43 },
		{ 324, 237, 43 },
		{ 324, 238, 43 },
		{ 324, 239, 43 },
		{ 325, 224, 43 },
		{ 325, 225, 43 },
		{ 325, 226, 43 },
		{ 325, 227, 43 },
		{ 325, 228, 43 },
		{ 325, 229, 43 },
		{ 325, 230, 43 },
		{ 325, 231, 43 },
		{ 325, 232, 43 },
		{ 325, 233, 43 },
		{ 325, 234, 43 },
		{ 325, 235, 43 },
		{ 325, 236, 43 },
		{ 325, 237, 43 },
		{ 325, 238, 43 },
		{ 325, 239, 43 },
		{ 326, 223, 43 },
		{ 326, 224, 43 },
		{ 326, 225, 43 },
		{ 326, 226, 43 },
		{ 326, 227, 43 },
		{ 326, 228, 43 },
		{ 326, 229, 43 },
		{ 326, 230, 43 },
		{ 326, 231, 43 },
		{ 326, 232, 43 },
		{ 326, 233, 43 },
		{ 326, 234, 43 },
		{ 326, 235, 43 },
		{ 326, 236, 43 },
		{ 326, 237, 43 },
		{ 326, 238, 43 },
		{ 326, 239, 43 },
		{ 327, 223, 43 },
		{ 327, 224, 43 },
		{ 327, 225, 43 },
		{ 327, 226, 43 },
		{ 327, 227, 43 },
		{ 327, 228, 43 },
		{ 327, 229, 43 },
		{ 327, 230, 43 },
		{ 327, 231, 43 },
		{ 327, 232, 43 },
		{ 327, 233, 43 },
		{ 327, 234, 43 },
		{ 327, 235, 43 },
		{ 327, 236, 43 },
		{ 327, 237, 43 },
		{ 327, 238, 43 },
		{ 327, 239, 43 },
		{ 328, 222, 43 },
		{ 328, 223, 43 },
		{ 328, 224, 43 },
		{ 328, 225, 43 },
		{ 328, 226, 43 },
		{ 328, 227, 43 },
		{ 328, 228, 43 },
		{ 328, 229, 43 },
		{ 328, 230, 43 },
		{ 328, 231, 43 },
		{ 328, 232, 43 },
		{ 328, 233, 43 },
		{ 328, 234, 43 },
		{ 328, 235, 43 },
		{ 328, 236, 43 },
		{ 328, 237, 43 },
		{ 328, 238, 43 },
		{ 328, 239, 43 },
		{ 329, 222, 43 },
		{ 329, 223, 43 },
		{ 329, 224, 43 },
		{ 329, 225, 43 },
		{ 329, 226, 43 },
		{ 329, 227, 43 },
		{ 329, 228, 43 },
		{ 329, 229, 43 },
		{ 329, 230, 43 },
		{ 329, 231, 43 },
		{ 329, 232, 43 },
		{ 329, 233, 43 },
		{ 329, 234, 43 },
		{ 329, 235, 43 },
		{ 329, 236, 43 },
		{ 329, 237, 43 },
		{ 329, 238, 43 },
		{ 330, 221, 43 },
		{ 330, 222, 43 },
		{ 330, 223, 43 },
		{ 330, 224, 43 },
		{ 330, 225, 43 },
		{ 330, 226, 43 },
		{ 330, 227, 43 },
		{ 330, 228, 43 },
		{ 330, 229, 43 },
		{ 330, 230, 43 },
		{ 330, 231, 43 },
		{ 330, 232, 43 },
		{ 330, 233, 43 },
		{ 330, 234, 43 },
		{ 330, 235, 43 },
		{ 330, 236, 43 },
		{ 330, 237, 43 },
		{ 330, 238, 43 },
		{ 331, 221, 43 },
		{ 331, 222, 43 },
		{ 331, 223, 43 },
		{ 331, 224, 43 },
		{ 331, 225, 43 },
		{ 331, 226, 43 },
		{ 331, 227, 43 },
		{ 331, 228, 43 },
		{ 331, 229, 43 },
		{ 331, 230, 43 },
		{ 331, 231, 43 },
		{ 331, 232, 43 },
		{ 331, 233, 43 },
		{ 331, 234, 43 },
		{ 331, 235, 43 },
		{ 331, 236, 43 },
		{ 331, 237, 43 },
		{ 331, 238, 43 },
		{ 332, 220, 43 },
		{ 332, 221, 43 },
		{ 332, 222, 43 },
		{ 332, 223, 43 },
		{ 332, 224, 43 },
		{ 332, 225, 43 },
		{ 332, 226, 43 },
		{ 332, 227, 43 },
		{ 332, 228, 43 },
		{ 332, 229, 43 },
		{ 332, 230, 43 },
		{ 332, 231, 43 },
		{ 332, 232, 43 },
		{ 332, 233, 43 },
		{ 332, 234, 43 },
		{ 332, 235, 43 },
		{ 332, 236, 43 },
		{ 332, 237, 43 },
		{ 332, 238, 43 },
		{ 333, 220, 43 },
		{ 333, 221, 43 },
		{ 333, 222, 43 },
		{ 333, 223, 43 },
		{ 333, 224, 43 },
		{ 333, 225, 43 },
		{ 333, 226, 43 },
		{ 333, 227, 43 },
		{ 333, 228, 43 },
		{ 333, 229, 43 },
		{ 333, 230, 43 },
		{ 333, 231, 43 },
		{ 333, 232, 43 },
		{ 333, 233, 43 },
		{ 333, 234, 43 },
		{ 333, 235, 43 },
		{ 333, 236, 43 },
		{ 333, 237, 43 },
		{ 334, 219, 43 },
		{ 334, 220, 43 },
		{ 334, 221, 43 },
		{ 334, 222, 43 },
		{ 334, 223, 43 },
		{ 334, 224, 43 },
		{ 334, 225, 43 },
		{ 334, 226, 43 },
		{ 334, 227, 43 },
		{ 334, 228, 43 },
		{ 334, 229, 43 },
		{ 334, 230, 43 },
		{ 334, 231, 43 },
		{ 334, 232, 43 },
		{ 334, 233, 43 },
		{ 334, 234, 43 },
		{ 334, 235, 43 },
		{ 334, 236, 43 },
		{ 334, 237, 43 },
		{ 335, 219, 43 },
		{ 335, 220, 43 },
		{ 335, 221, 43 },
		{ 335, 222, 43 },
		{ 335, 223, 43 },
		{ 335, 224, 43 },
		{ 335, 225, 43 },
		{ 335, 226, 43 },
		{ 335, 227, 43 },
		{ 335, 228, 43 },
		{ 335, 229, 43 },
		{ 335, 230, 43 },
		{ 335, 231, 43 },
		{ 335, 232, 43 },
		{ 335, 233, 43 },
		{ 335, 234, 43 },
		{ 335, 235, 43 },
		{ 335, 236, 43 },
		{ 335, 237, 43 },
		{ 336, 218, 43 },
		{ 336, 219, 43 },
		{ 336, 220, 43 },
		{ 336, 221, 43 },
		{ 336, 222, 43 },
		{ 336, 223, 43 },
		{ 336, 224, 43 },
		{ 336, 225, 43 },
		{ 336, 226, 43 },
		{ 336, 227, 43 },
		{ 336, 228, 43 },
		{ 336, 229, 43 },
		{ 336, 230, 43 },
		{ 336, 231, 43 },
		{ 336, 232, 43 },
		{ 336, 233, 43 },
		{ 336, 234, 43 },
		{ 336, 235, 43 },
		{ 336, 236, 43 },
		{ 336, 237, 43 },
		{ 337, 218, 43 },
		{ 337, 219, 43 },
		{ 337, 220, 43 },
		{ 337, 221, 43 },
		{ 337, 222, 43 },
		{ 337, 223, 43 },
		{ 337, 224, 43 },
		{ 337, 225, 43 },
		{ 337, 226, 43 },
		{ 337, 227, 43 },
		{ 337, 228, 43 },
		{ 337, 229, 43 },
		{ 337, 230, 43 },
		{ 337, 231, 43 },
		{ 337, 232, 43 },
		{ 337, 233, 43 },
		{ 337, 234, 43 },
		{ 337, 235, 43 },
		{ 337, 236, 43 },
		{ 338, 217, 43 },
		{ 338, 218, 43 },
		{ 338, 219, 43 },
		{ 338, 220, 43 },
		{ 338, 221, 43 },
		{ 338, 222, 43 },
		{ 338, 223, 43 },
		{ 338, 224, 43 },
		{ 338, 225, 43 },
		{ 338, 226, 43 },
		{ 338, 227, 43 },
		{ 338, 228, 43 },
		{ 338, 229, 43 },
		{ 338, 230, 43 },
		{ 338, 231, 43 },
		{ 338, 232, 43 },
		{ 338, 233, 43 },
		{ 338, 234, 43 },
		{ 338, 235, 43 },
		{ 338, 236, 43 },
		{ 339, 217, 43 },
		{ 339, 218, 43 },
		{ 339, 219, 43 },
		{ 339, 220, 43 },
		{ 339, 221, 43 },
		{ 339, 222, 43 },
		{ 339, 223, 43 },
		{ 339, 224, 43 },
		{ 339, 225, 43 },
		{ 339, 226, 43 },
		{ 339, 227, 43 },
		{ 339, 228, 43 },
		{ 339, 229, 43 },
		{ 339, 230, 43 },
		{ 339, 231, 43 },
		{ 339, 232, 43 },
		{ 339, 233, 43 },
		{ 339, 234, 43 },
		{ 339, 235, 43 },
		{ 339, 236, 43 },
		{ 340, 216, 43 },
		{ 340, 217, 43 },
		{ 340, 218, 43 },
		{ 340, 219, 43 },
		{ 340, 220, 43 },
		{ 340, 221, 43 },
		{ 340, 222, 43 },
		{ 340, 223, 43 },
		{ 340, 224, 43 },
		{ 340, 225, 43 },
		{ 340, 226, 43 },
		{ 340, 227, 43 },
		{ 340, 228, 43 },
		{ 340, 229, 43 },
		{ 340, 230, 43 },
		{ 340, 231, 43 },
		{ 340, 232, 43 },
		{ 340, 233, 43 },
		{ 340, 234, 43 },
		{ 340, 235, 43 },
		{ 340, 236, 43 },
		{ 341, 216, 43 },
		{ 341, 217, 43 },
		{ 341, 218, 43 },
		{ 341, 219, 43 },
		{ 341, 220, 43 },
		{ 341, 221, 43 },
		{ 341, 222, 43 },
		{ 341, 223, 43 },
		{ 341, 224, 43 },
		{ 341, 225, 43 },
		{ 341, 226, 43 },
		{ 341, 227, 43 },
		{ 341, 228, 43 },
		{ 341, 229, 43 },
		{ 341, 230, 43 },
		{ 341, 231, 43 },
		{ 341, 232, 43 },
		{ 341, 233, 43 },
		{ 341, 234, 43 },
		{ 341, 235, 43 },
		{ 342, 215, 43 },
		{ 342, 216, 43 },
		{ 342, 217, 43 },
		{ 342, 218, 43 },
		{ 342, 219, 43 },
		{ 342, 220, 43 },
		{ 342, 221, 43 },
		{ 342, 222, 43 },
		{ 342, 223, 43 },
		{ 342, 224, 43 },
		{ 342, 225, 43 },
		{ 342, 226, 43 },
		{ 342, 227, 43 },
		{ 342, 228, 43 },
		{ 342, 229, 43 },
		{ 342, 230, 43 },
		{ 342, 231, 43 },
		{ 342, 232, 43 },
		{ 342, 233, 43 },
		{ 342, 234, 43 },
		{ 342, 235, 43 },
		{ 343, 215, 43 },
		{ 343, 216, 43 },
		{ 343, 217, 43 },
		{ 343, 218, 43 },
		{ 343, 219, 43 },
		{ 343, 220, 43 },
		{ 343, 221, 43 },
		{ 343, 222, 43 },
		{ 343, 223, 43 },
		{ 343, 224, 43 },
		{ 343, 225, 43 },
		{ 343, 226, 43 },
		{ 343, 227, 43 },
		{ 343, 228, 43 },
		{ 343, 229, 43 },
		{ 343, 230, 43 },
		{ 343, 231, 43 },
		{ 343, 232, 43 },
		{ 343, 233, 43 },
		{ 343, 234, 43 },
		{ 343, 235, 43 },
		{ 344, 214, 43 },
		{ 344, 215, 43 },
		{ 344, 216, 43 },
		{ 344, 217, 43 },
		{ 344, 218, 43 },
		{ 344, 219, 43 },
		{ 344, 220, 43 },
		{ 344, 221, 43 },
		{ 344, 222, 43 },
		{ 344, 223, 43 },
		{ 344, 224, 43 },
		{ 344, 225, 43 },
		{ 344, 226, 43 },
		{ 344, 227, 43 },
		{ 344, 228, 43 },
		{ 344, 229, 43 },
		{ 344, 230, 43 },
		{ 344, 231, 43 },
		{ 344, 232, 43 },
		{ 344, 233, 43 },
		{ 344, 234, 43 },
		{ 344, 235, 43 },
		{ 345, 214, 43 },
		{ 345, 215, 43 },
		{ 345, 216, 43 },
		{ 345, 217, 43 },
		{ 345, 218, 43 },
		{ 345, 219, 43 },
		{ 345, 220, 43 },
		{ 345, 221, 43 },
		{ 345, 222, 43 },
		{ 345, 223, 43 },
		{ 345, 224, 43 },
		{ 345, 225, 43 },
		{ 345, 226, 43 },
		{ 345, 227, 43 },
		{ 345, 228, 43 },
		{ 345, 229, 43 },
		{ 345, 230, 43 },
		{ 345, 231, 43 },
		{ 345, 232, 43 },
		{ 345, 233, 43 },
		{ 345, 234, 43 },
		{ 346, 213, 43 },
		{ 346, 214, 43 },
		{ 346, 215, 43 },
		{ 346, 216, 43 },
		{ 346, 217, 43 },
		{ 346, 218, 43 },
		{ 346, 219, 43 },
		{ 346, 220, 43 },
		{ 346, 221, 43 },
		{ 346, 222, 43 },
		{ 346, 223, 43 },
		{ 346, 224, 43 },
		{ 346, 225, 43 },
		{ 346, 226, 43 },
		{ 346, 227, 43 },
		{ 346, 228, 43 },
		{ 346, 229, 43 },
		{ 346, 230, 43 },
		{ 346, 231, 43 },
		{ 346, 232, 43 },
		{ 346, 233, 43 },
		{ 346, 234, 43 },
		{ 347, 213, 43 },
		{ 347, 214, 43 },
		{ 347, 215, 43 },
		{ 347, 216, 43 },
		{ 347, 217, 43 },
		{ 347, 218, 43 },
		{ 347, 219, 43 },
		{ 347, 220, 43 },
		{ 347, 221, 43 },
		{ 347, 222, 43 },
		{ 347, 223, 43 },
		{ 347, 224, 43 },
		{ 347, 225, 43 },
		{ 347, 226, 43 },
		{ 347, 227, 43 },
		{ 347, 228, 43 },
		{ 347, 229, 43 },
		{ 347, 230, 43 },
		{ 347, 231, 43 },
		{ 347, 232, 43 },
		{ 347, 233, 43 },
		{ 347, 234, 43 },
		{ 348, 212, 43 },
		{ 348, 213, 43 },
		{ 348, 214, 43 },
		{ 348, 215, 43 },
		{ 348, 216, 43 },
		{ 348, 217, 43 },
		{ 348, 218, 43 },
		{ 348, 219, 43 },
		{ 348, 220, 43 },
		{ 348, 221, 43 },
		{ 348, 222, 43 },
		{ 348, 223, 43 },
		{ 348, 224, 43 },
		{ 348, 225, 43 },
		{ 348, 226, 43 },
		{ 348, 227, 43 },
		{ 348, 228, 43 },
		{ 348, 229, 43 },
		{ 348, 230, 43 },
		{ 348, 231, 43 },
		{ 348, 232, 43 },
		{ 348, 233, 43 },
		{ 348, 234, 43 },
		{ 349, 212, 43 },
		{ 349, 213, 43 },
		{ 349, 214, 43 },
		{ 349, 215, 43 },
		{ 349, 216, 43 },
		{ 349, 217, 43 },
		{ 349, 218, 43 },
		{ 349, 219, 43 },
		{ 349, 220, 43 },
		{ 349, 221, 43 },
		{ 349, 222, 43 },
		{ 349, 223, 43 },
		{ 349, 224, 43 },
		{ 349, 225, 43 },
		{ 349, 226, 43 },
		{ 349, 227, 43 },
		{ 349, 228, 43 },
		{ 349, 229, 43 },
		{ 349, 230, 43 },
		{ 349, 231, 43 },
		{ 349, 232, 43 },
		{ 349, 233, 43 },
		{ 350, 211, 43 },
		{ 350, 212, 43 },
		{ 350, 213, 43 },
		{ 350, 214, 43 },
		{ 350, 215, 43 },
		{ 350, 216, 43 },
		{ 350, 217, 43 },
		{ 350, 218, 43 },
		{ 350, 219, 43 },
		{ 350, 220, 43 },
		{ 350, 221, 43 },
		{ 350, 222, 43 },
		{ 350, 223, 43 },
		{ 350, 224, 43 },
		{ 350, 225, 43 },
		{ 350, 226, 43 },
		{ 350, 227, 43 },
		{ 350, 228, 43 },
		{ 350, 229, 43 },
		{ 350, 230, 43 },
		{ 350, 231, 43 },
		{ 350, 232, 43 },
		{ 350, 233, 43 },
		{ 351, 211, 43 },
		{ 351, 212, 43 },
		{ 351, 213, 43 },
		{ 351, 214, 43 },
		{ 351, 215, 43 },
		{ 351, 216, 43 },
		{ 351, 217, 43 },
		{ 351, 218, 43 },
		{ 351, 219, 43 },
		{ 351, 220, 43 },
		{ 351, 221, 43 },
		{ 351, 222, 43 },
		{ 351, 223, 43 },
		{ 351, 224, 43 },
		{ 351, 225, 43 },
		{ 351, 226, 43 },
		{ 351, 227, 43 },
		{ 351, 228, 43 },
		{ 351, 229, 43 },
		{ 351, 230, 43 },
		{ 351, 231, 43 },
		{ 351, 232, 43 },
		{ 351, 233, 43 },
		{ 352, 210, 43 },
		{ 352, 211, 43 },
		{ 352, 212, 43 },
		{ 352, 213, 43 },
		{ 352, 214, 43 },
		{ 352, 215, 43 },
		{ 352, 216, 43 },
		{ 352, 217, 43 },
		{ 352, 218, 43 },
		{ 352, 219, 43 },
		{ 352, 220, 43 },
		{ 352, 221, 43 },
		{ 352, 222, 43 },
		{ 352, 223, 43 },
		{ 352, 224, 43 },
		{ 352, 225, 43 },
		{ 352, 226, 43 },
		{ 352, 227, 43 },
		{ 352, 228, 43 },
		{ 352, 229, 43 },
		{ 352, 230, 43 },
		{ 352, 231, 43 },
		{ 352, 232, 43 },
		{ 352, 233, 43 },
		{ 353, 210, 43 },
		{ 353, 211, 43 },
		{ 353, 212, 43 },
		{ 353, 213, 43 },
		{ 353, 214, 43 },
		{ 353, 215, 43 },
		{ 353, 216, 43 },
		{ 353, 217, 43 },
		{ 353, 218, 43 },
		{ 353, 219, 43 },
		{ 353, 220, 43 },
		{ 353, 221, 43 },
		{ 353, 222, 43 },
		{ 353, 223, 43 },
		{ 353, 224, 43 },
		{ 353, 225, 43 },
		{ 353, 226, 43 },
		{ 353, 227, 43 },
		{ 353, 228, 43 },
		{ 353, 229, 43 },
		{ 353, 230, 43 },
		{ 353, 231, 43 },
		{ 353, 232, 43 },
		{ 354, 209, 43 },
		{ 354, 210, 43 },
		{ 354, 211, 43 },
		{ 354, 212, 43 },
		{ 354, 213, 43 },
		{ 354, 214, 43 },
		{ 354, 215, 43 },
		{ 354, 216, 43 },
		{ 354, 217, 43 },
		{ 354, 218, 43 },
		{ 354, 219, 43 },
		{ 354, 220, 43 },
		{ 354, 221, 43 },
		{ 354, 222, 43 },
		{ 354, 223, 43 },
		{ 354, 224, 43 },
		{ 354, 225, 43 },
		{ 354, 226, 43 },
		{ 354, 227, 43 },
		{ 354, 228, 43 },
		{ 354, 229, 43 },
		{ 354, 230, 43 },
		{ 354, 231, 43 },
		{ 354, 232, 43 },
		{ 355, 209, 43 },
		{ 355, 210, 43 },
		{ 355, 211, 43 },
		{ 355, 212, 43 },
		{ 355, 213, 43 },
		{ 355, 214, 43 },
		{ 355, 215, 43 },
		{ 355, 216, 43 },
		{ 355, 217, 43 },
		{ 355, 218, 43 },
		{ 355, 219, 43 },
		{ 355, 220, 43 },
		{ 355, 221, 43 },
		{ 355, 222, 43 },
		{ 355, 223, 43 },
		{ 355, 224, 43 },
		{ 355, 225, 43 },
		{ 355, 226, 43 },
		{ 355, 227, 43 },
		{ 355, 228, 43 },
		{ 355, 229, 43 },
		{ 355, 230, 43 },
		{ 355, 231, 43 },
		{ 355, 232, 43 },
		{ 356, 208, 43 },
		{ 356, 209, 43 },
		{ 356, 210, 43 },
		{ 356, 211, 43 },
		{ 356, 212, 43 },
		{ 356, 213, 43 },
		{ 356, 214, 43 },
		{ 356, 215, 43 },
		{ 356, 216, 43 },
		{ 356, 217, 43 },
		{ 356, 218, 43 },
		{ 356, 219, 43 },
		{ 356, 220, 43 },
		{ 356, 221, 43 },
		{ 356, 222, 43 },
		{ 356, 223, 43 },
		{ 356, 224, 43 },
		{ 356, 225, 43 },
		{ 356, 226, 43 },
		{ 356, 227, 43 },
		{ 356, 228, 43 },
		{ 356, 229, 43 },
		{ 356, 230, 43 },
		{ 356, 231, 43 },
		{ 356, 232, 43 },
		{ 357, 208, 43 },
		{ 357, 209, 43 },
		{ 357, 210, 43 },
		{ 357, 211, 43 },
		{ 357, 212, 43 },
		{ 357, 213, 43 },
		{ 357, 214, 43 },
		{ 357, 215, 43 },
		{ 357, 216, 43 },
		{ 357, 217, 43 },
		{ 357, 218, 43 },
		{ 357, 219, 43 },
		{ 357, 220, 43 },
		{ 357, 221, 43 },
		{ 357, 222, 43 },
		{ 357, 223, 43 },
		{ 357, 224, 43 },
		{ 357, 225, 43 },
		{ 357, 226, 43 },
		{ 357, 227, 43 },
		{ 357, 228, 43 },
		{ 357, 229, 43 },
		{ 357, 230, 43 },
		{ 357, 231, 43 },
		{ 358, 207, 43 },
		{ 358, 208, 43 },
		{ 358, 209, 43 },
		{ 358, 210, 43 },
		{ 358, 211, 43 },
		{ 358, 212, 43 },
		{ 358, 213, 43 },
		{ 358, 214, 43 },
		{ 358, 215, 43 },
		{ 358, 216, 43 },
		{ 358, 217, 43 },
		{ 358, 218, 43 },
		{ 358, 219, 43 },
		{ 358, 220, 43 },
		{ 358, 221, 43 },
		{ 358, 222, 43 },
		{ 358, 223, 43 },
		{ 358, 224, 43 },
		{ 358, 225, 43 },
		{ 358, 226, 43 },
		{ 358, 227, 43 },
		{ 358, 228, 43 },
		{ 358, 229, 43 },
		{ 358, 230, 43 },
		{ 358, 231, 43 },
		{ 359, 207, 43 },
		{ 359, 208, 43 },
		{ 359, 209, 43 },
		{ 359, 210, 43 },
		{ 359, 211, 43 },
		{ 359, 212, 43 },
		{ 359, 213, 43 },
		{ 359, 214, 43 },
		{ 359, 215, 43 },
		{ 359, 216, 43 },
		{ 359, 217, 43 },
		{ 359, 218, 43 },
		{ 359, 219, 43 },
		{ 359, 220, 43 },
		{ 359, 221, 43 },
		{ 359, 222, 43 },
		{ 359, 223, 43 },
		{ 359, 224, 43 },
		{ 359, 225, 43 },
		{ 359, 226, 43 },
		{ 359, 227, 43 },
		{ 359, 228, 43 },
		{ 359, 229, 43 },
		{ 359, 230, 43 },
		{ 359, 231, 43 },
		{ 360, 206, 43 },
		{ 360, 207, 43 },
		{ 360, 208, 43 },
		{ 360, 209, 43 },
		{ 360, 210, 43 },
		{ 360, 211, 43 },
		{ 360, 212, 43 },
		{ 360, 213, 43 },
		{ 360, 214, 43 },
		{ 360, 215, 43 },
		{ 360, 216, 43 },
		{ 360, 217, 43 },
		{ 360, 218, 43 },
		{ 360, 219, 43 },
		{ 360, 220, 43 },
		{ 360, 221, 43 },
		{ 360, 222, 43 },
		{ 360, 223, 43 },
		{ 360, 224, 43 },
		{ 360, 225, 43 },
		{ 360, 226, 43 },
		{ 360, 227, 43 },
		{ 360, 228, 43 },
		{ 360, 229, 43 },
		{ 360, 230, 43 },
		{ 360, 231, 43 },
		{ 361, 206, 43 },
		{ 361, 207, 43 },
		{ 361, 208, 43 },
		{ 361, 209, 43 },
		{ 361, 210, 43 },
		{ 361, 211, 43 },
		{ 361, 212, 43 },
		{ 361, 213, 43 },
		{ 361, 214, 43 },
		{ 361, 215, 43 },
		{ 361, 216, 43 },
		{ 361, 217, 43 },
		{ 361, 218, 43 },
		{ 361, 219, 43 },
		{ 361, 220, 43 },
		{ 361, 221, 43 },
		{ 361, 222, 43 },
		{ 361, 223, 43 },
		{ 361, 224, 43 },
		{ 361, 225, 43 },
		{ 361, 226, 43 },
		{ 361, 227, 43 },
		{ 361, 228, 43 },
		{ 361, 229, 43 },
		{ 361, 230, 43 },
		{ 362, 205, 43 },
		{ 362, 206, 43 },
		{ 362, 207, 43 },
		{ 362, 208, 43 },
		{ 362, 209, 43 },
		{ 362, 210, 43 },
		{ 362, 211, 43 },
		{ 362, 212, 43 },
		{ 362, 213, 43 },
		{ 362, 214, 43 },
		{ 362, 215, 43 },
		{ 362, 216, 43 },
		{ 362, 217, 43 },
		{ 362, 218, 43 },
		{ 362, 219, 43 },
		{ 362, 220, 43 },
		{ 362, 221, 43 },
		{ 362, 222, 43 },
		{ 362, 223, 43 },
		{ 362, 224, 43 },
		{ 362, 225, 43 },
		{ 362, 226, 43 },
		{ 362, 227, 43 },
		{ 362, 228, 43 },
		{ 362, 229, 43 },
		{ 362, 230, 43 },
		{ 363, 205, 43 },
		{ 363, 206, 43 },
		{ 363, 207, 43 },
		{ 363, 208, 43 },
		{ 363, 209, 43 },
		{ 363, 210, 43 },
		{ 363, 211, 43 },
		{ 363, 212, 43 },
		{ 363, 213, 43 },
		{ 363, 214, 43 },
		{ 363, 215, 43 },
		{ 363, 216, 43 },
		{ 363, 217, 43 },
		{ 363, 218, 43 },
		{ 363, 219, 43 },
		{ 363, 220, 43 },
		{ 363, 221, 43 },
		{ 363, 222, 43 },
		{ 363, 223, 43 },
		{ 363, 224, 43 },
		{ 363, 225, 43 },
		{ 363, 226, 43 },
		{ 363, 227, 43 },
		{ 363, 228, 43 },
		{ 363, 229, 43 },
		{ 363, 230, 43 },
		{ 364, 204, 43 },
		{ 364, 205, 43 },
		{ 364, 206, 43 },
		{ 364, 207, 43 },
		{ 364, 208, 43 },
		{ 364, 209, 43 },
		{ 364, 210, 43 },
		{ 364, 211, 43 },
		{ 364, 212, 43 },
		{ 364, 213, 43 },
		{ 364, 214, 43 },
		{ 364, 215, 43 },
		{ 364, 216, 43 },
		{ 364, 217, 43 },
		{ 364, 218, 43 },
		{ 364, 219, 43 },
		{ 364, 220, 43 },
		{ 364, 221, 43 },
		{ 364, 222, 43 },
		{ 364, 223, 43 },
		{ 364, 224, 43 },
		{ 364, 225, 43 },
		{ 364, 226, 43 },
		{ 364, 227, 43 },
		{ 364, 228, 43 },
		{ 364, 229, 43 },
		{ 364, 230, 43 },
		{ 365, 204, 43 },
		{ 365, 205, 43 },
		{ 365, 206, 43 },
		{ 365, 207, 43 },
		{ 365, 208, 43 },
		{ 365, 209, 43 },
		{ 365, 210, 43 },
		{ 365, 211, 43 },
		{ 365, 212, 43 },
		{ 365, 213, 43 },
		{ 365, 214, 43 },
		{ 365, 215, 43 },
		{ 365, 216, 43 },
		{ 365, 217, 43 },
		{ 365, 218, 43 },
		{ 365, 219, 43 },
		{ 365, 220, 43 },
		{ 365, 221, 43 },
		{ 365, 222, 43 },
		{ 365, 223, 43 },
		{ 365, 224, 43 },
		{ 365, 225, 43 },
		{ 365, 226, 43 },
		{ 365, 227, 43 },
		{ 365, 228, 43 },
		{ 365, 229, 43 },
		{ 366, 203, 43 },
		{ 366, 204, 43 },
		{ 366, 205, 43 },
		{ 366, 206, 43 },
		{ 366, 207, 43 },
		{ 366, 208, 43 },
		{ 366, 209, 43 },
		{ 366, 210, 43 },
		{ 366, 211, 43 },
		{ 366, 212, 43 },
		{ 366, 213, 43 },
		{ 366, 214, 43 },
		{ 366, 215, 43 },
		{ 366, 216, 43 },
		{ 366, 217, 43 },
		{ 366, 218, 43 },
		{ 366, 219, 43 },
		{ 366, 220, 43 },
		{ 366, 221, 43 },
		{ 366, 222, 43 },
		{ 366, 223, 43 },
		{ 366, 224, 43 },
		{ 366, 225, 43 },
		{ 366, 226, 43 },
		{ 366, 227, 43 },
		{ 366, 228, 43 },
		{ 366, 229, 43 },
		{ 367, 203, 43 },
		{ 367, 204, 43 },
		{ 367, 205, 43 },
		{ 367, 206, 43 },
		{ 367, 207, 43 },
		{ 367, 208, 43 },
		{ 367, 209, 43 },
		{ 367, 210, 43 },
		{ 367, 211, 43 },
		{ 367, 212, 43 },
		{ 367, 213, 43 },
		{ 367, 214, 43 },
		{ 367, 215, 43 },
		{ 367, 216, 43 },
		{ 367, 217, 43 },
		{ 367, 218, 43 },
		{ 367, 219, 43 },
		{ 367, 220, 43 },
		{ 367, 221, 43 },
		{ 367, 222, 43 },
		{ 367, 223, 43 },
		{ 367, 224, 43 },
		{ 367, 225, 43 },
		{ 367, 226, 43 },
		{ 367, 227, 43 },
		{ 367, 228, 43 },
		{ 367, 229, 43 },
		{ 368, 202, 43 },
		{ 368, 203, 43 },
		{ 368, 204, 43 },
		{ 368, 205, 43 },
		{ 368, 206, 43 },
		{ 368, 207, 43 },
		{ 368, 208, 43 },
		{ 368, 209, 43 },
		{ 368, 210, 43 },
		{ 368, 211, 43 },
		{ 368, 212, 43 },
		{ 368, 213, 43 },
		{ 368, 214, 43 },
		{ 368, 215, 43 },
		{ 368, 216, 43 },
		{ 368, 217, 43 },
		{ 368, 218, 43 },
		{ 368, 219, 43 },
		{ 368, 220, 43 },
		{ 368, 221, 43 },
		{ 368, 222, 43 },
		{ 368, 223, 43 },
		{ 368, 224, 43 },
		{ 368, 225, 43 },
		{ 368, 226, 43 },
		{ 368, 227, 43 },
		{ 368, 228, 43 },
		{ 368, 229, 43 },
		{ 369, 202, 43 },
		{ 369, 203, 43 },
		{ 369, 204, 43 },
		{ 369, 205, 43 },
		{ 369, 206, 43 },
		{ 369, 207, 43 },
		{ 369, 208, 43 },
		{ 369, 209, 43 },
		{ 369, 210, 43 },
		{ 369, 211, 43 },
		{ 369, 212, 43 },
		{ 369, 213, 43 },
		{ 369, 214, 43 },
		{ 369, 215, 43 },
		{ 369, 216, 43 },
		{ 369, 217, 43 },
		{ 369, 218, 43 },
		{ 369, 219, 43 },
		{ 369, 220, 43 },
		{ 369, 221, 43 },
		{ 369, 222, 43 },
		{ 369, 223, 43 },
		{ 369, 224, 43 },
		{ 369, 225, 43 },
		{ 369, 226, 43 },
		{ 369, 227, 43 },
		{ 369, 228, 43 },
		{ 370, 201, 43 },
		{ 370, 202, 43 },
		{ 370, 203, 43 },
		{ 370, 204, 43 },
		{ 370, 205, 43 },
		{ 370, 206, 43 },
		{ 370, 207, 43 },
		{ 370, 208, 43 },
		{ 370, 209, 43 },
		{ 370, 210, 43 },
		{ 370, 211, 43 },
		{ 370, 212, 43 },
		{ 370, 213, 43 },
		{ 370, 214, 43 },
		{ 370, 215, 43 },
		{ 370, 216, 43 },
		{ 370, 217, 43 },
		{ 370, 218, 43 },
		{ 370, 219, 43 },
		{ 370, 220, 43 },
		{ 370, 221, 43 },
		{ 370, 222, 43 },
		{ 370, 223, 43 },
		{ 370, 224, 43 },
		{ 370, 225, 43 },
		{ 370, 226, 43 },
		{ 370, 227, 43 },
		{ 370, 228, 43 },
		{ 371, 201, 43 },
		{ 371, 202, 43 },
		{ 371, 203, 43 },
		{ 371, 204, 43 },
		{ 371, 205, 43 },
		{ 371, 206, 43 },
		{ 371, 207, 43 },
		{ 371, 208, 43 },
		{ 371, 209, 43 },
		{ 371, 210, 43 },
		{ 371, 211, 43 },
		{ 371, 212, 43 },
		{ 371, 213, 43 },
		{ 371, 214, 43 },
		{ 371, 215, 43 },
		{ 371, 216, 43 },
		{ 371, 217, 43 },
		{ 371, 218, 43 },
		{ 371, 219, 43 },
		{ 371, 220, 43 },
		{ 371, 221, 43 },
		{ 371, 222, 43 },
		{ 371, 223, 43 },
		{ 371, 224, 43 },
		{ 371, 225, 43 },
		{ 371, 226, 43 },
		{ 371, 227, 43 },
		{ 371, 228, 43 },
		{ 372, 200, 43 },
		{ 372, 201, 43 },
		{ 372, 202, 43 },
		{ 372, 203, 43 },
		{ 372, 204, 43 },
		{ 372, 205, 43 },
		{ 372, 206, 43 },
		{ 372, 207, 43 },
		{ 372, 208, 43 },
		{ 372, 209, 43 },
		{ 372, 210, 43 },
		{ 372, 211, 43 },
		{ 372, 212, 43 },
		{ 372, 213, 43 },
		{ 372, 214, 43 },
		{ 372, 215, 43 },
		{ 372, 216, 43 },
		{ 372, 217, 43 },
		{ 372, 218, 43 },
		{ 372, 219, 43 },
		{ 372, 220, 43 },
		{ 372, 221, 43 },
		{ 372, 222, 43 },
		{ 372, 223, 43 },
		{ 372, 224, 43 },
		{ 372, 225, 43 },
		{ 372, 226, 43 },
		{ 372, 227, 43 },
		{ 372, 228, 43 },
		{ 373, 200, 43 },
		{ 373, 201, 43 },
		{ 373, 202, 43 },
		{ 373, 203, 43 },
		{ 373, 204, 43 },
		{ 373, 205, 43 },
		{ 373, 206, 43 },
		{ 373, 207, 43 },
		{ 373, 208, 43 },
		{ 373, 209, 43 },
		{ 373, 210, 43 },
		{ 373, 211, 43 },
		{ 373, 212, 43 },
		{ 373, 213, 43 },
		{ 373, 214, 43 },
		{ 373, 215, 43 },
		{ 373, 216, 43 },
		{ 373, 217, 43 },
		{ 373, 218, 43 },
		{ 373, 219, 43 },
		{ 373, 220, 43 },
		{ 373, 221, 43 },
		{ 373, 222, 43 },
		{ 373, 223, 43 },
		{ 373, 224, 43 },
		{ 373, 225, 43 },
		{ 373, 226, 43 },
		{ 373, 227, 43 },
		{ 374, 199, 43 },
		{ 374, 200, 43 },
		{ 374, 201, 43 },
		{ 374, 202, 43 },
		{ 374, 203, 43 },
		{ 374, 204, 43 },
		{ 374, 205, 43 },
		{ 374, 206, 43 },
		{ 374, 207, 43 },
		{ 374, 208, 43 },
		{ 374, 209, 43 },
		{ 374, 210, 43 },
		{ 374, 211, 43 },
		{ 374, 212, 43 },
		{ 374, 213, 43 },
		{ 374, 214, 43 },
		{ 374, 215, 43 },
		{ 374, 216, 43 },
		{ 374, 217, 43 },
		{ 374, 218, 43 },
		{ 374, 219, 43 },
		{ 374, 220, 43 },
		{ 374, 221, 43 },
		{ 374, 222, 43 },
		{ 374, 223, 43 },
		{ 374, 224, 43 },
		{ 374, 225, 43 },
		{ 374, 226, 43 },
		{ 374, 227, 43 },
		{ 375, 199, 43 },
		{ 375, 200, 43 },
		{ 375, 201, 43 },
		{ 375, 202, 43 },
		{ 375, 203, 43 },
		{ 375, 204, 43 },
		{ 375, 205, 43 },
		{ 375, 206, 43 },
		{ 375, 207, 43 },
		{ 375, 208, 43 },
		{ 375, 209, 43 },
		{ 375, 210, 43 },
		{ 375, 211, 43 },
		{ 375, 212, 43 },
		{ 375, 213, 43 },
		{ 375, 214, 43 },
		{ 375, 215, 43 },
		{ 375, 216, 43 },
		{ 375, 217, 43 },
		{ 375, 218, 43 },
		{ 375, 219, 43 },
		{ 375, 220, 43 },
		{ 375, 221, 43 },
		{ 375, 222, 43 },
		{ 375, 223, 43 },
		{ 375, 224, 43 },
		{ 375, 225, 43 },
		{ 375, 226, 43 },
		{ 375, 227, 43 },
		{ 376, 198, 43 },
		{ 376, 199, 43 },
		{ 376, 200, 43 },
		{ 376, 201, 43 },
		{ 376, 202, 43 },
		{ 376, 203, 43 },
		{ 376, 204, 43 },
		{ 376, 205, 43 },
		{ 376, 206, 43 },
		{ 376, 207, 43 },
		{ 376, 208, 43 },
		{ 376, 209, 43 },
		{ 376, 210, 43 },
		{ 376, 211, 43 },
		{ 376, 212, 43 },
		{ 376, 213, 43 },
		{ 376, 214, 43 },
		{ 376, 215, 43 },
		{ 376, 216, 43 },
		{ 376, 217, 43 },
		{ 376, 218, 43 },
		{ 376, 219, 43 },
		{ 376, 220, 43 },
		{ 376, 221, 43 },
		{ 376, 222, 43 },
		{ 376, 223, 43 },
		{ 376, 224, 43 },
		{ 376, 225, 43 },
		{ 376, 226, 43 },
		{ 376, 227, 43 },
		{ 377, 198, 43 },
		{ 377, 199, 43 },
		{ 377, 200, 43 },
		{ 377, 201, 43 },
		{ 377, 202, 43 },
		{ 377, 203, 43 },
		{ 377, 204, 43 },
		{ 377, 205, 43 },
		{ 377, 206, 43 },
		{ 377, 207, 43 },
		{ 377, 208, 43 },
		{ 377, 209, 43 },
		{ 377, 210, 43 },
		{ 377, 211, 43 },
		{ 377, 212, 43 },
		{ 377, 213, 43 },
		{ 377, 214, 43 },
		{ 377, 215, 43 },
		{ 377, 216, 43 },
		{ 377, 217, 43 },
		{ 377, 218, 43 },
		{ 377, 219, 43 },
		{ 377, 220, 43 },
		{ 377, 221, 43 },
		{ 377, 222, 43 },
		{ 377, 223, 43 },
		{ 377, 224, 43 },
		{ 377, 225, 43 },
		{ 377, 226, 43 },
		{ 378, 197, 43 },
		{ 378, 198, 43 },
		{ 378, 199, 43 },
		{ 378, 200, 43 },
		{ 378, 201, 43 },
		{ 378, 202, 43 },
		{ 378, 203, 43 },
		{ 378, 204, 43 },
		{ 378, 205, 43 },
		{ 378, 206, 43 },
		{ 378, 207, 43 },
		{ 378, 208, 43 },
		{ 378, 209, 43 },
		{ 378, 210, 43 },
		{ 378, 211, 43 },
		{ 378, 212, 43 },
		{ 378, 213, 43 },
		{ 378, 214, 43 },
		{ 378, 215, 43 },
		{ 378, 216, 43 },
		{ 378, 217, 43 },
		{ 378, 218, 43 },
		{ 378, 219, 43 },
		{ 378, 220, 43 },
		{ 378, 221, 43 },
		{ 378, 222, 43 },
		{ 378, 223, 43 },
		{ 378, 224, 43 },
		{ 378, 225, 43 },
		{ 378, 226, 43 },
		{ 379, 197, 43 },
		{ 379, 198, 43 },
		{ 379, 199, 43 },
		{ 379, 200, 43 },
		{ 379, 201, 43 },
		{ 379, 202, 43 },
		{ 379, 203, 43 },
		{ 379, 204, 43 },
		{ 379, 205, 43 },
		{ 379, 206, 43 },
		{ 379, 207, 43 },
		{ 379, 208, 43 },
		{ 379, 209, 43 },
		{ 379, 210, 43 },
		{ 379, 211, 43 },
		{ 379, 212, 43 },
		{ 379, 213, 43 },
		{ 379, 214, 43 },
		{ 379, 215, 43 },
		{ 379, 216, 43 },
		{ 379, 217, 43 },
		{ 379, 218, 43 },
		{ 379, 219, 43 },
		{ 379, 220, 43 },
		{ 379, 221, 43 },
		{ 379, 222, 43 },
		{ 379, 223, 43 },
		{ 379, 224, 43 },
		{ 379, 225, 43 },
		{ 379, 226, 43 },
		{ 380, 196, 43 },
		{ 380, 197, 43 },
		{ 380, 198, 43 },
		{ 380, 199, 43 },
		{ 380, 200, 43 },
		{ 380, 201, 43 },
		{ 380, 202, 43 },
		{ 380, 203, 43 },
		{ 380, 204, 43 },
		{ 380, 205, 43 },
		{ 380, 206, 43 },
		{ 380, 207, 43 },
		{ 380, 208, 43 },
		{ 380, 209, 43 },
		{ 380, 210, 43 },
		{ 380, 211, 43 },
		{ 380, 212, 43 },
		{ 380, 213, 43 },
		{ 380, 214, 43 },
		{ 380, 215, 43 },
		{ 380, 216, 43 },
		{ 380, 217, 43 },
		{ 380, 218, 43 },
		{ 380, 219, 43 },
		{ 380, 220, 43 },
		{ 380, 221, 43 },
		{ 380, 222, 43 },
		{ 380, 223, 43 },
		{ 380, 224, 43 },
		{ 380, 225, 43 },
		{ 380, 226, 43 },
		{ 381, 196, 43 },
		{ 381, 197, 43 },
		{ 381, 198, 43 },
		{ 381, 199, 43 },
		{ 381, 200, 43 },
		{ 381, 201, 43 },
		{ 381, 202, 43 },
		{ 381, 203, 43 },
		{ 381, 204, 43 },
		{ 381, 205, 43 },
		{ 381, 206, 43 },
		{ 381, 207, 43 },
		{ 381, 208, 43 },
		{ 381, 209, 43 },
		{ 381, 210, 43 },
		{ 381, 211, 43 },
		{ 381, 212, 43 },
		{ 381, 213, 43 },
		{ 381, 214, 43 },
		{ 381, 215, 43 },
		{ 381, 216, 43 },
		{ 381, 217, 43 },
		{ 381, 218, 43 },
		{ 381, 219, 43 },
		{ 381, 220, 43 },
		{ 381, 221, 43 },
		{ 381, 222, 43 },
		{ 381, 223, 43 },
		{ 381, 224, 43 },
		{ 381, 225, 43 },
		{ 382, 195, 43 },
		{ 382, 196, 43 },
		{ 382, 197, 43 },
		{ 382, 198, 43 },
		{ 382, 199, 43 },
		{ 382, 200, 43 },
		{ 382, 201, 43 },
		{ 382, 202, 43 },
		{ 382, 203, 43 },
		{ 382, 204, 43 },
		{ 382, 205, 43 },
		{ 382, 206, 43 },
		{ 382, 207, 43 },
		{ 382, 208, 43 },
		{ 382, 209, 43 },
		{ 382, 210, 43 },
		{ 382, 211, 43 },
		{ 382, 212, 43 },
		{ 382, 213, 43 },
		{ 382, 214, 43 },
		{ 382, 215, 43 },
		{ 382, 216, 43 },
		{ 382, 217, 43 },
		{ 382, 218, 43 },
		{ 382, 219, 43 },
		{ 382, 220, 43 },
		{ 382, 221, 43 },
		{ 382, 222, 43 },
		{ 382, 223, 43 },
		{ 382, 224, 43 },
		{ 382, 225, 43 },
		{ 383, 195, 43 },
		{ 383, 196, 43 },
		{ 383, 197, 43 },
		{ 383, 198, 43 },
		{ 383, 199, 43 },
		{ 383, 200, 43 },
		{ 383, 201, 43 },
		{ 383, 202, 43 },
		{ 383, 203, 43 },
		{ 383, 204, 43 },
		{ 383, 205, 43 },
		{ 383, 206, 43 },
		{ 383, 207, 43 },
		{ 383, 208, 43 },
		{ 383, 209, 43 },
		{ 383, 210, 43 },
		{ 383, 211, 43 },
		{ 383, 212, 43 },
		{ 383, 213, 43 },
		{ 383, 214, 43 },
		{ 383, 215, 43 },
		{ 383, 216, 43 },
		{ 383, 217, 43 },
		{ 383, 218, 43 },
		{ 383, 219, 43 },
		{ 383, 220, 43 },
		{ 383, 221, 43 },
		{ 383, 222, 43 },
		{ 383, 223, 43 },
		{ 383, 224, 43 },
		{ 383, 225, 43 },
		{ 384, 194, 43 },
		{ 384, 195, 43 },
		{ 384, 196, 43 },
		{ 384, 197, 43 },
		{ 384, 198, 43 },
		{ 384, 199, 43 },
		{ 384, 200, 43 },
		{ 384, 201, 43 },
		{ 384, 202, 43 },
		{ 384, 203, 43 },
		{ 384, 204, 43 },
		{ 384, 205, 43 },
		{ 384, 206, 43 },
		{ 384, 207, 43 },
		{ 384, 208, 43 },
		{ 384, 209, 43 },
		{ 384, 210, 43 },
		{ 384, 211, 43 },
		{ 384, 212, 43 },
		{ 384, 213, 43 },
		{ 384, 214, 43 },
		{ 384, 215, 43 },
		{ 384, 216, 43 },
		{ 384, 217, 43 },
		{ 384, 218, 43 },
		{ 384, 219, 43 },
		{ 384, 220, 43 },
		{ 384, 221, 43 },
		{ 384, 222, 43 },
		{ 384, 223, 43 },
		{ 384, 224, 43 },
		{ 384, 225, 43 },
		{ 385, 194, 43 },
		{ 385, 195, 43 },
		{ 385, 196, 43 },
		{ 385, 197, 43 },
		{ 385, 198, 43 },
		{ 385, 199, 43 },
		{ 385, 200, 43 },
		{ 385, 201, 43 },
		{ 385, 202, 43 },
		{ 385, 203, 43 },
		{ 385, 204, 43 },
		{ 385, 205, 43 },
		{ 385, 206, 43 },
		{ 385, 207, 43 },
		{ 385, 208, 43 },
		{ 385, 209, 43 },
		{ 385, 210, 43 },
		{ 385, 211, 43 },
		{ 385, 212, 43 },
		{ 385, 213, 43 },
		{ 385, 214, 43 },
		{ 385, 215, 43 },
		{ 385, 216, 43 },
		{ 385, 217, 43 },
		{ 385, 218, 43 },
		{ 385, 219, 43 },
		{ 385, 220, 43 },
		{ 385, 221, 43 },
		{ 385, 222, 43 },
		{ 385, 223, 43 },
		{ 385, 224, 43 },
		{ 386, 193, 43 },
		{ 386, 194, 43 },
		{ 386, 195, 43 },
		{ 386, 196, 43 },
		{ 386, 197, 43 },
		{ 386, 198, 43 },
		{ 386, 199, 43 },
		{ 386, 200, 43 },
		{ 386, 201, 43 },
		{ 386, 202, 43 },
		{ 386, 203, 43 },
		{ 386, 204, 43 },
		{ 386, 205, 43 },
		{ 386, 206, 43 },
		{ 386, 207, 43 },
		{ 386, 208, 43 },
		{ 386, 209, 43 },
		{ 386, 210, 43 },
		{ 386, 211, 43 },
		{ 386, 212, 43 },
		{ 386, 213, 43 },
		{ 386, 214, 43 },
		{ 386, 215, 43 },
		{ 386, 216, 43 },
		{ 386, 217, 43 },
		{ 386, 218, 43 },
		{ 386, 219, 43 },
		{ 386, 220, 43 },
		{ 386, 221, 43 },
		{ 386, 222, 43 },
		{ 386, 223, 43 },
		{ 386, 224, 43 },
		{ 387, 193, 43 },
		{ 387, 194, 43 },
		{ 387, 195, 43 },
		{ 387, 196, 43 },
		{ 387, 197, 43 },
		{ 387, 198, 43 },
		{ 387, 199, 43 },
		{ 387, 200, 43 },
		{ 387, 201, 43 },
		{ 387, 202, 43 },
		{ 387, 203, 43 },
		{ 387, 204, 43 },
		{ 387, 205, 43 },
		{ 387, 206, 43 },
		{ 387, 207, 43 },
		{ 387, 208, 43 },
		{ 387, 209, 43 },
		{ 387, 210, 43 },
		{ 387, 211, 43 },
		{ 387, 212, 43 },
		{ 387, 213, 43 },
		{ 387, 214, 43 },
		{ 387, 215, 43 },
		{ 387, 216, 43 },
		{ 387, 217, 43 },
		{ 387, 218, 43 },
		{ 387, 219, 43 },
		{ 387, 220, 43 },
		{ 387, 221, 43 },
		{ 387, 222, 43 },
		{ 387, 223, 43 },
		{ 387, 224, 43 },
		{ 388, 192, 43 },
		{ 388, 193, 43 },
		{ 388, 194, 43 },
		{ 388, 195, 43 },
		{ 388, 196, 43 },
		{ 388, 197, 43 },
		{ 388, 198, 43 },
		{ 388, 199, 43 },
		{ 388, 200, 43 },
		{ 388, 201, 43 },
		{ 388, 202, 43 },
		{ 388, 203, 43 },
		{ 388, 204, 43 },
		{ 388, 205, 43 },
		{ 388, 206, 43 },
		{ 388, 207, 43 },
		{ 388, 208, 43 },
		{ 388, 209, 43 },
		{ 388, 210, 43 },
		{ 388, 211, 43 },
		{ 388, 212, 43 },
		{ 388, 213, 43 },
		{ 388, 214, 43 },
		{ 388, 215, 43 },
		{ 388, 216, 43 },
		{ 388, 217, 43 },
		{ 388, 218, 43 },
		{ 388, 219, 43 },
		{ 388, 220, 43 },
		{ 388, 221, 43 },
		{ 388, 222, 43 },
		{ 388, 223, 43 },
		{ 388, 224, 43 },
		{ 389, 192, 43 },
		{ 389, 193, 43 },
		{ 389, 194, 43 },
		{ 389, 195, 43 },
		{ 389, 196, 43 },
		{ 389, 197, 43 },
		{ 389, 198, 43 },
		{ 389, 199, 43 },
		{ 389, 200, 43 },
		{ 389, 201, 43 },
		{ 389, 202, 43 },
		{ 389, 203, 43 },
		{ 389, 204, 43 },
		{ 389, 205, 43 },
		{ 389, 206, 43 },
		{ 389, 207, 43 },
		{ 389, 208, 43 },
		{ 389, 209, 43 },
		{ 389, 210, 43 },
		{ 389, 211, 43 },
		{ 389, 212, 43 },
		{ 389, 213, 43 },
		{ 389, 214, 43 },
		{ 389, 215, 43 },
		{ 389, 216, 43 },
		{ 389, 217, 43 },
		{ 389, 218, 43 },
		{ 389, 219, 43 },
		{ 389, 220, 43 },
		{ 389, 221, 43 },
		{ 389, 222, 43 },
		{ 389, 223, 43 },
		{ 390, 191, 43 },
		{ 390, 192, 43 },
		{ 390, 193, 43 },
		{ 390, 194, 43 },
		{ 390, 195, 43 },
		{ 390, 196, 43 },
		{ 390, 197, 43 },
		{ 390, 198, 43 },
		{ 390, 199, 43 },
		{ 390, 200, 43 },
		{ 390, 201, 43 },
		{ 390, 202, 43 },
		{ 390, 203, 43 },
		{ 390, 204, 43 },
		{ 390, 205, 43 },
		{ 390, 206, 43 },
		{ 390, 207, 43 },
		{ 390, 208, 43 },
		{ 390, 209, 43 },
		{ 390, 210, 43 },
		{ 390, 211, 43 },
		{ 390, 212, 43 },
		{ 390, 213, 43 },
		{ 390, 214, 43 },
		{ 390, 215, 43 },
		{ 390, 216, 43 },
		{ 390, 217, 43 },
		{ 390, 218, 43 },
		{ 390, 219, 43 },
		{ 390, 220, 43 },
		{ 390, 221, 43 },
		{ 390, 222, 43 },
		{ 390, 223, 43 },
		{ 391, 191, 43 },
		{ 391, 192, 43 },
		{ 391, 193, 43 },
		{ 391, 194, 43 },
		{ 391, 195, 43 },
		{ 391, 196, 43 },
		{ 391, 197, 43 },
		{ 391, 198, 43 },
		{ 391, 199, 43 },
		{ 391, 200, 43 },
		{ 391, 201, 43 },
		{ 391, 202, 43 },
		{ 391, 203, 43 },
		{ 391, 204, 43 },
		{ 391, 205, 43 },
		{ 391, 206, 43 },
		{ 391, 207, 43 },
		{ 391, 208, 43 },
		{ 391, 209, 43 },
		{ 391, 210, 43 },
		{ 391, 211, 43 },
		{ 391, 212, 43 },
		{ 391, 213, 43 },
		{ 391, 214, 43 },
		{ 391, 215, 43 },
		{ 391, 216, 43 },
		{ 391, 217, 43 },
		{ 391, 218, 43 },
		{ 391, 219, 43 },
		{ 391, 220, 43 },
		{ 391, 221, 43 },
		{ 391, 222, 43 },
		{ 391, 223, 43 },
		{ 392, 190, 43 },
		{ 392, 191, 43 },
		{ 392, 192, 43 },
		{ 392, 193, 43 },
		{ 392, 194, 43 },
		{ 392, 195, 43 },
		{ 392, 196, 43 },
		{ 392, 197, 43 },
		{ 392, 198, 43 },
		{ 392, 199, 43 },
		{ 392, 200, 43 },
		{ 392, 201, 43 },
		{ 392, 202, 43 },
		{ 392, 203, 43 },
		{ 392, 204, 43 },
		{ 392, 205, 43 },
		{ 392, 206, 43 },
		{ 392, 207, 43 },
		{ 392, 208, 43 },
		{ 392, 209, 43 },
		{ 392, 210, 43 },
		{ 392, 211, 43 },
		{ 392, 212, 43 },
		{ 392, 213, 43 },
		{ 392, 214, 43 },
		{ 392, 215, 43 },
		{ 392, 216, 43 },
		{ 392, 217, 43 },
		{ 392, 218, 43 },
		{ 392, 219, 43 },
		{ 392, 220, 43 },
		{ 392, 221, 43 },
		{ 392, 222, 43 },
		{ 392, 223, 43 },
		{ 393, 190, 43 },
		{ 393, 191, 43 },
		{ 393, 192, 43 },
		{ 393, 193, 43 },
		{ 393, 194, 43 },
		{ 393, 195, 43 },
		{ 393, 196, 43 },
		{ 393, 197, 43 },
		{ 393, 198, 43 },
		{ 393, 199, 43 },
		{ 393, 200, 43 },
		{ 393, 201, 43 },
		{ 393, 202, 43 },
		{ 393, 203, 43 },
		{ 393, 204, 43 },
		{ 393, 205, 43 },
		{ 393, 206, 43 },
		{ 393, 207, 43 },
		{ 393, 208, 43 },
		{ 393, 209, 43 },
		{ 393, 210, 43 },
		{ 393, 211, 43 },
		{ 393, 212, 43 },
		{ 393, 213, 43 },
		{ 393, 214, 43 },
		{ 393, 215, 43 },
		{ 393, 216, 43 },
		{ 393, 217, 43 },
		{ 393, 218, 43 },
		{ 393, 219, 43 },
		{ 393, 220, 43 },
		{ 393, 221, 43 },
		{ 393, 222, 43 },
		{ 394, 189, 43 },
		{ 394, 190, 43 },
		{ 394, 191, 43 },
		{ 394, 192, 43 },
		{ 394, 193, 43 },
		{ 394, 194, 43 },
		{ 394, 195, 43 },
		{ 394, 196, 43 },
		{ 394, 197, 43 },
		{ 394, 198, 43 },
		{ 394, 199, 43 },
		{ 394, 200, 43 },
		{ 394, 201, 43 },
		{ 394, 202, 43 },
		{ 394, 203, 43 },
		{ 394, 204, 43 },
		{ 394, 205, 43 },
		{ 394, 206, 43 },
		{ 394, 207, 43 },
		{ 394, 208, 43 },
		{ 394, 209, 43 },
		{ 394, 210, 43 },
		{ 394, 211, 43 },
		{ 394, 212, 43 },
		{ 394, 213, 43 },
		{ 394, 214, 43 },
		{ 394, 215, 43 },
		{ 394, 216, 43 },
		{ 394, 217, 43 },
		{ 394, 218, 43 },
		{ 394, 219, 43 },
		{ 394, 220, 43 },
		{ 394, 221, 43 },
		{ 394, 222, 43 },
		{ 395, 189, 43 },
		{ 395, 190, 43 },
		{ 395, 191, 43 },
		{ 395, 192, 43 },
		{ 395, 193, 43 },
		{ 395, 194, 43 },
		{ 395, 195, 43 },
		{ 395, 196, 43 },
		{ 395, 197, 43 },
		{ 395, 198, 43 },
		{ 395, 199, 43 },
		{ 395, 200, 43 },
		{ 395, 201, 43 },
		{ 395, 202, 43 },
		{ 395, 203, 43 },
		{ 395, 204, 43 },
		{ 395, 205, 43 },
		{ 395, 206, 43 },
		{ 395, 207, 43 },
		{ 395, 208, 43 },
		{ 395, 209, 43 },
		{ 395, 210, 43 },
		{ 395, 211, 43 },
		{ 395, 212, 43 },
		{ 395, 213, 43 },
		{ 395, 214, 43 },
		{ 395, 215, 43 },
		{ 395, 216, 43 },
		{ 395, 217, 43 },
		{ 395, 218, 43 },
		{ 395, 219, 43 },
		{ 395, 220, 43 },
		{ 395, 221, 43 },
		{ 395, 222, 43 },
		{ 396, 188, 43 },
		{ 396, 189, 43 },
		{ 396, 190, 43 },
		{ 396, 191, 43 },
		{ 396, 192, 43 },
		{ 396, 193, 43 },
		{ 396, 194, 43 },
		{ 396, 195, 43 },
		{ 396, 196, 43 },
		{ 396, 197, 43 },
		{ 396, 198, 43 },
		{ 396, 199, 43 },
		{ 396, 200, 43 },
		{ 396, 201, 43 },
		{ 396, 202, 43 },
		{ 396, 203, 43 },
		{ 396, 204, 43 },
		{ 396, 205, 43 },
		{ 396, 206, 43 },
		{ 396, 207, 43 },
		{ 396, 208, 43 },
		{ 396, 209, 43 },
		{ 396, 210, 43 },
		{ 396, 211, 43 },
		{ 396, 212, 43 },
		{ 396, 213, 43 },
		{ 396, 214, 43 },
		{ 396, 215, 43 },
		{ 396, 216, 43 },
		{ 396, 217, 43 },
		{ 396, 218, 43 },
		{ 396, 219, 43 },
		{ 396, 220, 43 },
		{ 396, 221, 43 },
		{ 396, 222, 43 },
		{ 397, 188, 43 },
		{ 397, 189, 43 },
		{ 397, 190, 43 },
		{ 397, 191, 43 },
		{ 397, 192, 43 },
		{ 397, 193, 43 },
		{ 397, 194, 43 },
		{ 397, 195, 43 },
		{ 397, 196, 43 },
		{ 397, 197, 43 },
		{ 397, 198, 43 },
		{ 397, 199, 43 },
		{ 397, 200, 43 },
		{ 397, 201, 43 },
		{ 397, 202, 43 },
		{ 397, 203, 43 },
		{ 397, 204, 43 },
		{ 397, 205, 43 },
		{ 397, 206, 43 },
		{ 397, 207, 43 },
		{ 397, 208, 43 },
		{ 397, 209, 43 },
		{ 397, 210, 43 },
		{ 397, 211, 43 },
		{ 397, 212, 43 },
		{ 397, 213, 43 },
		{ 397, 214, 43 },
		{ 397, 215, 43 },
		{ 397, 216, 43 },
		{ 397, 217, 43 },
		{ 397, 218, 43 },
		{ 397, 219, 43 },
		{ 397, 220, 43 },
		{ 397, 221, 43 },
		{ 398, 187, 43 },
		{ 398, 188, 43 },
		{ 398, 189, 43 },
		{ 398, 190, 43 },
		{ 398, 191, 43 },
		{ 398, 192, 43 },
		{ 398, 193, 43 },
		{ 398, 194, 43 },
		{ 398, 195, 43 },
		{ 398, 196, 43 },
		{ 398, 197, 43 },
		{ 398, 198, 43 },
		{ 398, 199, 43 },
		{ 398, 200, 43 },
		{ 398, 201, 43 },
		{ 398, 202, 43 },
		{ 398, 203, 43 },
		{ 398, 204, 43 },
		{ 398, 205, 43 },
		{ 398, 206, 43 },
		{ 398, 207, 43 },
		{ 398, 208, 43 },
		{ 398, 209, 43 },
		{ 398, 210, 43 },
		{ 398, 211, 43 },
		{ 398, 212, 43 },
		{ 398, 213, 43 },
		{ 398, 214, 43 },
		{ 398, 215, 43 },
		{ 398, 216, 43 },
		{ 398, 217, 43 },
		{ 398, 218, 43 },
		{ 398, 219, 43 },
		{ 398, 220, 43 },
		{ 398, 221, 43 },
		{ 399, 187, 43 },
		{ 399, 188, 43 },
		{ 399, 189, 43 },
		{ 399, 190, 43 },
		{ 399, 191, 43 },
		{ 399, 192, 43 },
		{ 399, 193, 43 },
		{ 399, 194, 43 },
		{ 399, 195, 43 },
		{ 399, 196, 43 },
		{ 399, 197, 43 },
		{ 399, 198, 43 },
		{ 399, 199, 43 },
		{ 399, 200, 43 },
		{ 399, 201, 43 },
		{ 399, 202, 43 },
		{ 399, 203, 43 },
		{ 399, 204, 43 },
		{ 399, 205, 43 },
		{ 399, 206, 43 },
		{ 399, 207, 43 },
		{ 399, 208, 43 },
		{ 399, 209, 43 },
		{ 399, 210, 43 },
		{ 399, 211, 43 },
		{ 399, 212, 43 },
		{ 399, 213, 43 },
		{ 399, 214, 43 },
		{ 399, 215, 43 },
		{ 399, 216, 43 },
		{ 399, 217, 43 },
		{ 399, 218, 43 },
		{ 399, 219, 43 },
		{ 399, 220, 43 },
		{ 399, 221, 43 },
		{ 400, 186, 43 },
		{ 400, 187, 43 },
		{ 400, 188, 43 },
		{ 400, 189, 43 },
		{ 400, 190, 43 },
		{ 400, 191, 43 },
		{ 400, 192, 43 },
		{ 400, 193, 43 },
		{ 400, 194, 43 },
		{ 400, 195, 43 },
		{ 400, 196, 43 },
		{ 400, 197, 43 },
		{ 400, 198, 43 },
		{ 400, 199, 43 },
		{ 400, 200, 43 },
		{ 400, 201, 43 },
		{ 400, 202, 43 },
		{ 400, 203, 43 },
		{ 400, 204, 43 },
		{ 400, 205, 43 },
		{ 400, 206, 43 },
		{ 400, 207, 43 },
		{ 400, 208, 43 },
		{ 400, 209, 43 },
		{ 400, 210, 43 },
		{ 400, 211, 43 },
		{ 400, 212, 43 },
		{ 400, 213, 43 },
		{ 400, 214, 43 },
		{ 400, 215, 43 },
		{ 400, 216, 43 },
		{ 400, 217, 43 },
		{ 400, 218, 43 },
		{ 400, 219, 43 },
		{ 400, 220, 43 },
		{ 400, 221, 43 },
		{ 401, 186, 43 },
		{ 401, 187, 43 },
		{ 401, 188, 43 },
		{ 401, 189, 43 },
		{ 401, 190, 43 },
		{ 401, 191, 43 },
		{ 401, 192, 43 },
		{ 401, 193, 43 },
		{ 401, 194, 43 },
		{ 401, 195, 43 },
		{ 401, 196, 43 },
		{ 401, 197, 43 },
		{ 401, 198, 43 },
		{ 401, 199, 43 },
		{ 401, 200, 43 },
		{ 401, 201, 43 },
		{ 401, 202, 43 },
		{ 401, 203, 43 },
		{ 401, 204, 43 },
		{ 401, 205, 43 },
		{ 401, 206, 43 },
		{ 401, 207, 43 },
		{ 401, 208, 43 },
		{ 401, 209, 43 },
		{ 401, 210, 43 },
		{ 401, 211, 43 },
		{ 401, 212, 43 },
		{ 401, 213, 43 },
		{ 401, 214, 43 },
		{ 401, 215, 43 },
		{ 401, 216, 43 },
		{ 401, 217, 43 },
		{ 401, 218, 43 },
		{ 401, 219, 43 },
		{ 401, 220, 43 },
		{ 402, 185, 43 },
		{ 402, 186, 43 },
		{ 402, 187, 43 },
		{ 402, 188, 43 },
		{ 402, 189, 43 },
		{ 402, 190, 43 },
		{ 402, 191, 43 },
		{ 402, 192, 43 },
		{ 402, 193, 43 },
		{ 402, 194, 43 },
		{ 402, 195, 43 },
		{ 402, 196, 43 },
		{ 402, 197, 43 },
		{ 402, 198, 43 },
		{ 402, 199, 43 },
		{ 402, 200, 43 },
		{ 402, 201, 43 },
		{ 402, 202, 43 },
		{ 402, 203, 43 },
		{ 402, 204, 43 },
		{ 402, 205, 43 },
		{ 402, 206, 43 },
		{ 402, 207, 43 },
		{ 402, 208, 43 },
		{ 402, 209, 43 },
		{ 402, 210, 43 },
		{ 402, 211, 43 },
		{ 402, 212, 43 },
		{ 402, 213, 43 },
		{ 402, 214, 43 },
		{ 402, 215, 43 },
		{ 402, 216, 43 },
		{ 402, 217, 43 },
		{ 402, 218, 43 },
		{ 402, 219, 43 },
		{ 402, 220, 43 },
		{ 403, 185, 43 },
		{ 403, 186, 43 },
		{ 403, 187, 43 },
		{ 403, 188, 43 },
		{ 403, 189, 43 },
		{ 403, 190, 43 },
		{ 403, 191, 43 },
		{ 403, 192, 43 },
		{ 403, 193, 43 },
		{ 403, 194, 43 },
		{ 403, 195, 43 },
		{ 403, 196, 43 },
		{ 403, 197, 43 },
		{ 403, 198, 43 },
		{ 403, 199, 43 },
		{ 403, 200, 43 },
		{ 403, 201, 43 },
		{ 403, 202, 43 },
		{ 403, 203, 43 },
		{ 403, 204, 43 },
		{ 403, 205, 43 },
		{ 403, 206, 43 },
		{ 403, 207, 43 },
		{ 403, 208, 43 },
		{ 403, 209, 43 },
		{ 403, 210, 43 },
		{ 403, 211, 43 },
		{ 403, 212, 43 },
		{ 403, 213, 43 },
		{ 403, 214, 43 },
		{ 403, 215, 43 },
		{ 403, 216, 43 },
		{ 403, 217, 43 },
		{ 403, 218, 43 },
		{ 403, 219, 43 },
		{ 403, 220, 43 },
		{ 404, 184, 43 },
		{ 404, 185, 43 },
		{ 404, 186, 43 },
		{ 404, 187, 43 },
		{ 404, 188, 43 },
		{ 404, 189, 43 },
		{ 404, 190, 43 },
		{ 404, 191, 43 },
		{ 404, 192, 43 },
		{ 404, 193, 43 },
		{ 404, 194, 43 },
		{ 404, 195, 43 },
		{ 404, 196, 43 },
		{ 404, 197, 43 },
		{ 404, 198, 43 },
		{ 404, 199, 43 },
		{ 404, 200, 43 },
		{ 404, 201, 43 },
		{ 404, 202, 43 },
		{ 404, 203, 43 },
		{ 404, 204, 43 },
		{ 404, 205, 43 },
		{ 404, 206, 43 },
		{ 404, 207, 43 },
		{ 404, 208, 43 },
		{ 404, 209, 43 },
		{ 404, 210, 43 },
		{ 404, 211, 43 },
		{ 404, 212, 43 },
		{ 404, 213, 43 },
		{ 404, 214, 43 },
		{ 404, 215, 43 },
		{ 404, 216, 43 },
		{ 404, 217, 43 },
		{ 404, 218, 43 },
		{ 404, 219, 43 },
		{ 404, 220, 43 },
		{ 405, 184, 43 },
		{ 405, 185, 43 },
		{ 405, 186, 43 },
		{ 405, 187, 43 },
		{ 405, 188, 43 },
		{ 405, 189, 43 },
		{ 405, 190, 43 },
		{ 405, 191, 43 },
		{ 405, 192, 43 },
		{ 405, 193, 43 },
		{ 405, 194, 43 },
		{ 405, 195, 43 },
		{ 405, 196, 43 },
		{ 405, 197, 43 },
		{ 405, 198, 43 },
		{ 405, 199, 43 },
		{ 405, 200, 43 },
		{ 405, 201, 43 },
		{ 405, 202, 43 },
		{ 405, 203, 43 },
		{ 405, 204, 43 },
		{ 405, 205, 43 },
		{ 405, 206, 43 },
		{ 405, 207, 43 },
		{ 405, 208, 43 },
		{ 405, 209, 43 },
		{ 405, 210, 43 },
		{ 405, 211, 43 },
		{ 405, 212, 43 },
		{ 405, 213, 43 },
		{ 405, 214, 43 },
		{ 405, 215, 43 },
		{ 405, 216, 43 },
		{ 405, 217, 43 },
		{ 405, 218, 43 },
		{ 405, 219, 43 },
		{ 406, 183, 43 },
		{ 406, 184, 43 },
		{ 406, 185, 43 },
		{ 406, 186, 43 },
		{ 406, 187, 43 },
		{ 406, 188, 43 },
		{ 406, 189, 43 },
		{ 406, 190, 43 },
		{ 406, 191, 43 },
		{ 406, 192, 43 },
		{ 406, 193, 43 },
		{ 406, 194, 43 },
		{ 406, 195, 43 },
		{ 406, 196, 43 },
		{ 406, 197, 43 },
		{ 406, 198, 43 },
		{ 406, 199, 43 },
		{ 406, 200, 43 },
		{ 406, 201, 43 },
		{ 406, 202, 43 },
		{ 406, 203, 43 },
		{ 406, 204, 43 },
		{ 406, 205, 43 },
		{ 406, 206, 43 },
		{ 406, 207, 43 },
		{ 406, 208, 43 },
		{ 406, 209, 43 },
		{ 406, 210, 43 },
		{ 406, 211, 43 },
		{ 406, 212, 43 },
		{ 406, 213, 43 },
		{ 406, 214, 43 },
		{ 406, 215, 43 },
		{ 406, 216, 43 },
		{ 406, 217, 43 },
		{ 406, 218, 43 },
		{ 406, 219, 43 },
		{ 407, 183, 43 },
		{ 407, 184, 43 },
		{ 407, 185, 43 },
		{ 407, 186, 43 },
		{ 407, 187, 43 },
		{ 407, 188, 43 },
		{ 407, 189, 43 },
		{ 407, 190, 43 },
		{ 407, 191, 43 },
		{ 407, 192, 43 },
		{ 407, 193, 43 },
		{ 407, 194, 43 },
		{ 407, 195, 43 },
		{ 407, 196, 43 },
		{ 407, 197, 43 },
		{ 407, 198, 43 },
		{ 407, 199, 43 },
		{ 407, 200, 43 },
		{ 407, 201, 43 },
		{ 407, 202, 43 },
		{ 407, 203, 43 },
		{ 407, 204, 43 },
		{ 407, 205, 43 },
		{ 407, 206, 43 },
		{ 407, 207, 43 },
		{ 407, 208, 43 },
		{ 407, 209, 43 },
		{ 407, 210, 43 },
		{ 407, 211, 43 },
		{ 407, 212, 43 },
		{ 407, 213, 43 },
		{ 407, 214, 43 },
		{ 407, 215, 43 },
		{ 407, 216, 43 },
		{ 407, 217, 43 },
		{ 407, 218, 43 },
		{ 407, 219, 43 },
		{ 408, 182, 43 },
		{ 408, 183, 43 },
		{ 408, 184, 43 },
		{ 408, 185, 43 },
		{ 408, 186, 43 },
		{ 408, 187, 43 },
		{ 408, 188, 43 },
		{ 408, 189, 43 },
		{ 408, 190, 43 },
		{ 408, 191, 43 },
		{ 408, 192, 43 },
		{ 408, 193, 43 },
		{ 408, 194, 43 },
		{ 408, 195, 43 },
		{ 408, 196, 43 },
		{ 408, 197, 43 },
		{ 408, 198, 43 },
		{ 408, 199, 43 },
		{ 408, 200, 43 },
		{ 408, 201, 43 },
		{ 408, 202, 43 },
		{ 408, 203, 43 },
		{ 408, 204, 43 },
		{ 408, 205, 43 },
		{ 408, 206, 43 },
		{ 408, 207, 43 },
		{ 408, 208, 43 },
		{ 408, 209, 43 },
		{ 408, 210, 43 },
		{ 408, 211, 43 },
		{ 408, 212, 43 },
		{ 408, 213, 43 },
		{ 408, 214, 43 },
		{ 408, 215, 43 },
		{ 408, 216, 43 },
		{ 408, 217, 43 },
		{ 408, 218, 43 },
		{ 408, 219, 43 },
		{ 409, 182, 43 },
		{ 409, 183, 43 },
		{ 409, 184, 43 },
		{ 409, 185, 43 },
		{ 409, 186, 43 },
		{ 409, 187, 43 },
		{ 409, 188, 43 },
		{ 409, 189, 43 },
		{ 409, 190, 43 },
		{ 409, 191, 43 },
		{ 409, 192, 43 },
		{ 409, 193, 43 },
		{ 409, 194, 43 },
		{ 409, 195, 43 },
		{ 409, 196, 43 },
		{ 409, 197, 43 },
		{ 409, 198, 43 },
		{ 409, 199, 43 },
		{ 409, 200, 43 },
		{ 409, 201, 43 },
		{ 409, 202, 43 },
		{ 409, 203, 43 },
		{ 409, 204, 43 },
		{ 409, 205, 43 },
		{ 409, 206, 43 },
		{ 409, 207, 43 },
		{ 409, 208, 43 },
		{ 409, 209, 43 },
		{ 409, 210, 43 },
		{ 409, 211, 43 },
		{ 409, 212, 43 },
		{ 409, 213, 43 },
		{ 409, 214, 43 },
		{ 409, 215, 43 },
		{ 409, 216, 43 },
		{ 409, 217, 43 },
		{ 409, 218, 43 },
		{ 410, 181, 43 },
		{ 410, 182, 43 },
		{ 410, 183, 43 },
		{ 410, 184, 43 },
		{ 410, 185, 43 },
		{ 410, 186, 43 },
		{ 410, 187, 43 },
		{ 410, 188, 43 },
		{ 410, 189, 43 },
		{ 410, 190, 43 },
		{ 410, 191, 43 },
		{ 410, 192, 43 },
		{ 410, 193, 43 },
		{ 410, 194, 43 },
		{ 410, 195, 43 },
		{ 410, 196, 43 },
		{ 410, 197, 43 },
		{ 410, 198, 43 },
		{ 410, 199, 43 },
		{ 410, 200, 43 },
		{ 410, 201, 43 },
		{ 410, 202, 43 },
		{ 410, 203, 43 },
		{ 410, 204, 43 },
		{ 410, 205, 43 },
		{ 410, 206, 43 },
		{ 410, 207, 43 },
		{ 410, 208, 43 },
		{ 410, 209, 43 },
		{ 410, 210, 43 },
		{ 410, 211, 43 },
		{ 410, 212, 43 },
		{ 410, 213, 43 },
		{ 410, 214, 43 },
		{ 410, 215, 43 },
		{ 410, 216, 43 },
		{ 410, 217, 43 },
		{ 410, 218, 43 },
		{ 411, 181, 43 },
		{ 411, 182, 43 },
		{ 411, 183, 43 },
		{ 411, 184, 43 },
		{ 411, 185, 43 },
		{ 411, 186, 43 },
		{ 411, 187, 43 },
		{ 411, 188, 43 },
		{ 411, 189, 43 },
		{ 411, 190, 43 },
		{ 411, 191, 43 },
		{ 411, 192, 43 },
		{ 411, 193, 43 },
		{ 411, 194, 43 },
		{ 411, 195, 43 },
		{ 411, 196, 43 },
		{ 411, 197, 43 },
		{ 411, 198, 43 },
		{ 411, 199, 43 },
		{ 411, 200, 43 },
		{ 411, 201, 43 },
		{ 411, 202, 43 },
		{ 411, 203, 43 },
		{ 411, 204, 43 },
		{ 411, 205, 43 },
		{ 411, 206, 43 },
		{ 411, 207, 43 },
		{ 411, 208, 43 },
		{ 411, 209, 43 },
		{ 411, 210, 43 },
		{ 411, 211, 43 },
		{ 411, 212, 43 },
		{ 411, 213, 43 },
		{ 411, 214, 43 },
		{ 411, 215, 43 },
		{ 411, 216, 43 },
		{ 411, 217, 43 },
		{ 411, 218, 43 },
		{ 412, 180, 43 },
		{ 412, 181, 43 },
		{ 412, 182, 43 },
		{ 412, 183, 43 },
		{ 412, 184, 43 },
		{ 412, 185, 43 },
		{ 412, 186, 43 },
		{ 412, 187, 43 },
		{ 412, 188, 43 },
		{ 412, 189, 43 },
		{ 412, 190, 43 },
		{ 412, 191, 43 },
		{ 412, 192, 43 },
		{ 412, 193, 43 },
		{ 412, 194, 43 },
		{ 412, 195, 43 },
		{ 412, 196, 43 },
		{ 412, 197, 43 },
		{ 412, 198, 43 },
		{ 412, 199, 43 },
		{ 412, 200, 43 },
		{ 412, 201, 43 },
		{ 412, 202, 43 },
		{ 412, 203, 43 },
		{ 412, 204, 43 },
		{ 412, 205, 43 },
		{ 412, 206, 43 },
		{ 412, 207, 43 },
		{ 412, 208, 43 },
		{ 412, 209, 43 },
		{ 412, 210, 43 },
		{ 412, 211, 43 },
		{ 412, 212, 43 },
		{ 412, 213, 43 },
		{ 412, 214, 43 },
		{ 412, 215, 43 },
		{ 412, 216, 43 },
		{ 412, 217, 43 },
		{ 412, 218, 43 },
		{ 413, 180, 43 },
		{ 413, 181, 43 },
		{ 413, 182, 43 },
		{ 413, 183, 43 },
		{ 413, 184, 43 },
		{ 413, 185, 43 },
		{ 413, 186, 43 },
		{ 413, 187, 43 },
		{ 413, 188, 43 },
		{ 413, 189, 43 },
		{ 413, 190, 43 },
		{ 413, 191, 43 },
		{ 413, 192, 43 },
		{ 413, 193, 43 },
		{ 413, 194, 43 },
		{ 413, 195, 43 },
		{ 413, 196, 43 },
		{ 413, 197, 43 },
		{ 413, 198, 43 },
		{ 413, 199, 43 },
		{ 413, 200, 43 },
		{ 413, 201, 43 },
		{ 413, 202, 43 },
		{ 413, 203, 43 },
		{ 413, 204, 43 },
		{ 413, 205, 43 },
		{ 413, 206, 43 },
		{ 413, 207, 43 },
		{ 413, 208, 43 },
		{ 413, 209, 43 },
		{ 413, 210, 43 },
		{ 413, 211, 43 },
		{ 413, 212, 43 },
		{ 413, 213, 43 },
		{ 413, 214, 43 },
		{ 413, 215, 43 },
		{ 413, 216, 43 },
		{ 413, 217, 43 },
		{ 414, 179, 43 },
		{ 414, 180, 43 },
		{ 414, 181, 43 },
		{ 414, 182, 43 },
		{ 414, 183, 43 },
		{ 414, 184, 43 },
		{ 414, 185, 43 },
		{ 414, 186, 43 },
		{ 414, 187, 43 },
		{ 414, 188, 43 },
		{ 414, 189, 43 },
		{ 414, 190, 43 },
		{ 414, 191, 43 },
		{ 414, 192, 43 },
		{ 414, 193, 43 },
		{ 414, 194, 43 },
		{ 414, 195, 43 },
		{ 414, 196, 43 },
		{ 414, 197, 43 },
		{ 414, 198, 43 },
		{ 414, 199, 43 },
		{ 414, 200, 43 },
		{ 414, 201, 43 },
		{ 414, 202, 43 },
		{ 414, 203, 43 },
		{ 414, 204, 43 },
		{ 414, 205, 43 },
		{ 414, 206, 43 },
		{ 414, 207, 43 },
		{ 414, 208, 43 },
		{ 414, 209, 43 },
		{ 414, 210, 43 },
		{ 414, 211, 43 },
		{ 414, 212, 43 },
		{ 414, 213, 43 },
		{ 414, 214, 43 },
		{ 414, 215, 43 },
		{ 414, 216, 43 },
		{ 414, 217, 43 },
		{ 415, 179, 43 },
		{ 415, 180, 43 },
		{ 415, 181, 43 },
		{ 415, 182, 43 },
		{ 415, 183, 43 },
		{ 415, 184, 43 },
		{ 415, 185, 43 },
		{ 415, 186, 43 },
		{ 415, 187, 43 },
		{ 415, 188, 43 },
		{ 415, 189, 43 },
		{ 415, 190, 43 },
		{ 415, 191, 43 },
		{ 415, 192, 43 },
		{ 415, 193, 43 },
		{ 415, 194, 43 },
		{ 415, 195, 43 },
		{ 415, 196, 43 },
		{ 415, 197, 43 },
		{ 415, 198, 43 },
		{ 415, 199, 43 },
		{ 415, 200, 43 },
		{ 415, 201, 43 },
		{ 415, 202, 43 },
		{ 415, 203, 43 },
		{ 415, 204, 43 },
		{ 415, 205, 43 },
		{ 415, 206, 43 },
		{ 415, 207, 43 },
		{ 415, 208, 43 },
		{ 415, 209, 43 },
		{ 415, 210, 43 },
		{ 415, 211, 43 },
		{ 415, 212, 43 },
		{ 415, 213, 43 },
		{ 415, 214, 43 },
		{ 415, 215, 43 },
		{ 415, 216, 43 },
		{ 415, 217, 43 },
		{ 416, 178, 43 },
		{ 416, 179, 43 },
		{ 416, 180, 43 },
		{ 416, 181, 43 },
		{ 416, 182, 43 },
		{ 416, 183, 43 },
		{ 416, 184, 43 },
		{ 416, 185, 43 },
		{ 416, 186, 43 },
		{ 416, 187, 43 },
		{ 416, 188, 43 },
		{ 416, 189, 43 },
		{ 416, 190, 43 },
		{ 416, 191, 43 },
		{ 416, 192, 43 },
		{ 416, 193, 43 },
		{ 416, 194, 43 },
		{ 416, 195, 43 },
		{ 416, 196, 43 },
		{ 416, 197, 43 },
		{ 416, 198, 43 },
		{ 416, 199, 43 },
		{ 416, 200, 43 },
		{ 416, 201, 43 },
		{ 416, 202, 43 },
		{ 416, 203, 43 },
		{ 416, 204, 43 },
		{ 416, 205, 43 },
		{ 416, 206, 43 },
		{ 416, 207, 43 },
		{ 416, 208, 43 },
		{ 416, 209, 43 },
		{ 416, 210, 43 },
		{ 416, 211, 43 },
		{ 416, 212, 43 },
		{ 416, 213, 43 },
		{ 416, 214, 43 },
		{ 416, 215, 43 },
		{ 416, 216, 43 },
		{ 416, 217, 43 },
		{ 417, 178, 43 },
		{ 417, 179, 43 },
		{ 417, 180, 43 },
		{ 417, 181, 43 },
		{ 417, 182, 43 },
		{ 417, 183, 43 },
		{ 417, 184, 43 },
		{ 417, 185, 43 },
		{ 417, 186, 43 },
		{ 417, 187, 43 },
		{ 417, 188, 43 },
		{ 417, 189, 43 },
		{ 417, 190, 43 },
		{ 417, 191, 43 },
		{ 417, 192, 43 },
		{ 417, 193, 43 },
		{ 417, 194, 43 },
		{ 417, 195, 43 },
		{ 417, 196, 43 },
		{ 417, 197, 43 },
		{ 417, 198, 43 },
		{ 417, 199, 43 },
		{ 417, 200, 43 },
		{ 417, 201, 43 },
		{ 417, 202, 43 },
		{ 417, 203, 43 },
		{ 417, 204, 43 },
		{ 417, 205, 43 },
		{ 417, 206, 43 },
		{ 417, 207, 43 },
		{ 417, 208, 43 },
		{ 417, 209, 43 },
		{ 417, 210, 43 },
		{ 417, 211, 43 },
		{ 417, 212, 43 },
		{ 417, 213, 43 },
		{ 417, 214, 43 },
		{ 417, 215, 43 },
		{ 417, 216, 43 },
		{ 418, 177, 43 },
		{ 418, 178, 43 },
		{ 418, 179, 43 },
		{ 418, 180, 43 },
		{ 418, 181, 43 },
		{ 418, 182, 43 },
		{ 418, 183, 43 },
		{ 418, 184, 43 },
		{ 418, 185, 43 },
		{ 418, 186, 43 },
		{ 418, 187, 43 },
		{ 418, 188, 43 },
		{ 418, 189, 43 },
		{ 418, 190, 43 },
		{ 418, 191, 43 },
		{ 418, 192, 43 },
		{ 418, 193, 43 },
		{ 418, 194, 43 },
		{ 418, 195, 43 },
		{ 418, 196, 43 },
		{ 418, 197, 43 },
		{ 418, 198, 43 },
		{ 418, 199, 43 },
		{ 418, 200, 43 },
		{ 418, 201, 43 },
		{ 418, 202, 43 },
		{ 418, 203, 43 },
		{ 418, 204, 43 },
		{ 418, 205, 43 },
		{ 418, 206, 43 },
		{ 418, 207, 43 },
		{ 418, 208, 43 },
		{ 418, 209, 43 },
		{ 418, 210, 43 },
		{ 418, 211, 43 },
		{ 418, 212, 43 },
		{ 418, 213, 43 },
		{ 418, 214, 43 },
		{ 418, 215, 43 },
		{ 418, 216, 43 },
		{ 419, 177, 43 },
		{ 419, 178, 43 },
		{ 419, 179, 43 },
		{ 419, 180, 43 },
		{ 419, 181, 43 },
		{ 419, 182, 43 },
		{ 419, 183, 43 },
		{ 419, 184, 43 },
		{ 419, 185, 43 },
		{ 419, 186, 43 },
		{ 419, 187, 43 },
		{ 419, 188, 43 },
		{ 419, 189, 43 },
		{ 419, 190, 43 },
		{ 419, 191, 43 },
		{ 419, 192, 43 },
		{ 419, 193, 43 },
		{ 419, 194, 43 },
		{ 419, 195, 43 },
		{ 419, 196, 43 },
		{ 419, 197, 43 },
		{ 419, 198, 43 },
		{ 419, 199, 43 },
		{ 419, 200, 43 },
		{ 419, 201, 43 },
		{ 419, 202, 43 },
		{ 419, 203, 43 },
		{ 419, 204, 43 },
		{ 419, 205, 43 },
		{ 419, 206, 43 },
		{ 419, 207, 43 },
		{ 419, 208, 43 },
		{ 419, 209, 43 },
		{ 419, 210, 43 },
		{ 419, 211, 43 },
		{ 419, 212, 43 },
		{ 419, 213, 43 },
		{ 419, 214, 43 },
		{ 419, 215, 43 },
		{ 419, 216, 43 },
		{ 420, 176, 43 },
		{ 420, 177, 43 },
		{ 420, 178, 43 },
		{ 420, 179, 43 },
		{ 420, 180, 43 },
		{ 420, 181, 43 },
		{ 420, 182, 43 },
		{ 420, 183, 43 },
		{ 420, 184, 43 },
		{ 420, 185, 43 },
		{ 420, 186, 43 },
		{ 420, 187, 43 },
		{ 420, 188, 43 },
		{ 420, 189, 43 },
		{ 420, 190, 43 },
		{ 420, 191, 43 },
		{ 420, 192, 43 },
		{ 420, 193, 43 },
		{ 420, 194, 43 },
		{ 420, 195, 43 },
		{ 420, 196, 43 },
		{ 420, 197, 43 },
		{ 420, 198, 43 },
		{ 420, 199, 43 },
		{ 420, 200, 43 },
		{ 420, 201, 43 },
		{ 420, 202, 43 },
		{ 420, 203, 43 },
		{ 420, 204, 43 },
		{ 420, 205, 43 },
		{ 420, 206, 43 },
		{ 420, 207, 43 },
		{ 420, 208, 43 },
		{ 420, 209, 43 },
		{ 420, 210, 43 },
		{ 420, 211, 43 },
		{ 420, 212, 43 },
		{ 420, 213, 43 },
		{ 420, 214, 43 },
		{ 420, 215, 43 },
		{ 420, 216, 43 },
		{ 421, 176, 43 },
		{ 421, 177, 43 },
		{ 421, 178, 43 },
		{ 421, 179, 43 },
		{ 421, 180, 43 },
		{ 421, 181, 43 },
		{ 421, 182, 43 },
		{ 421, 183, 43 },
		{ 421, 184, 43 },
		{ 421, 185, 43 },
		{ 421, 186, 43 },
		{ 421, 187, 43 },
		{ 421, 188, 43 },
		{ 421, 189, 43 },
		{ 421, 190, 43 },
		{ 421, 191, 43 },
		{ 421, 192, 43 },
		{ 421, 193, 43 },
		{ 421, 194, 43 },
		{ 421, 195, 43 },
		{ 421, 196, 43 },
		{ 421, 197, 43 },
		{ 421, 198, 43 },
		{ 421, 199, 43 },
		{ 421, 200, 43 },
		{ 421, 201, 43 },
		{ 421, 202, 43 },
		{ 421, 203, 43 },
		{ 421, 204, 43 },
		{ 421, 205, 43 },
		{ 421, 206, 43 },
		{ 421, 207, 43 },
		{ 421, 208, 43 },
		{ 421, 209, 43 },
		{ 421, 210, 43 },
		{ 421, 211, 43 },
		{ 421, 212, 43 },
		{ 421, 213, 43 },
		{ 421, 214, 43 },
		{ 421, 215, 43 },
		{ 422, 175, 43 },
		{ 422, 176, 43 },
		{ 422, 177, 43 },
		{ 422, 178, 43 },
		{ 422, 179, 43 },
		{ 422, 180, 43 },
		{ 422, 181, 43 },
		{ 422, 182, 43 },
		{ 422, 183, 43 },
		{ 422, 184, 43 },
		{ 422, 185, 43 },
		{ 422, 186, 43 },
		{ 422, 187, 43 },
		{ 422, 188, 43 },
		{ 422, 189, 43 },
		{ 422, 190, 43 },
		{ 422, 191, 43 },
		{ 422, 192, 43 },
		{ 422, 193, 43 },
		{ 422, 194, 43 },
		{ 422, 195, 43 },
		{ 422, 196, 43 },
		{ 422, 197, 43 },
		{ 422, 198, 43 },
		{ 422, 199, 43 },
		{ 422, 200, 43 },
		{ 422, 201, 43 },
		{ 422, 202, 43 },
		{ 422, 203, 43 },
		{ 422, 204, 43 },
		{ 422, 205, 43 },
		{ 422, 206, 43 },
		{ 422, 207, 43 },
		{ 422, 208, 43 },
		{ 422, 209, 43 },
		{ 422, 210, 43 },
		{ 422, 211, 43 },
		{ 422, 212, 43 },
		{ 422, 213, 43 },
		{ 422, 214, 43 },
		{ 422, 215, 43 },
		{ 423, 175, 43 },
		{ 423, 176, 43 },
		{ 423, 177, 43 },
		{ 423, 178, 43 },
		{ 423, 179, 43 },
		{ 423, 180, 43 },
		{ 423, 181, 43 },
		{ 423, 182, 43 },
		{ 423, 183, 43 },
		{ 423, 184, 43 },
		{ 423, 185, 43 },
		{ 423, 186, 43 },
		{ 423, 187, 43 },
		{ 423, 188, 43 },
		{ 423, 189, 43 },
		{ 423, 190, 43 },
		{ 423, 191, 43 },
		{ 423, 192, 43 },
		{ 423, 193, 43 },
		{ 423, 194, 43 },
		{ 423, 195, 43 },
		{ 423, 196, 43 },
		{ 423, 197, 43 },
		{ 423, 198, 43 },
		{ 423, 199, 43 },
		{ 423, 200, 43 },
		{ 423, 201, 43 },
		{ 423, 202, 43 },
		{ 423, 203, 43 },
		{ 423, 204, 43 },
		{ 423, 205, 43 },
		{ 423, 206, 43 },
		{ 423, 207, 43 },
		{ 423, 208, 43 },
		{ 423, 209, 43 },
		{ 423, 210, 43 },
		{ 423, 211, 43 },
		{ 423, 212, 43 },
		{ 423, 213, 43 },
		{ 423, 214, 43 },
		{ 423, 215, 43 },
		{ 424, 174, 43 },
		{ 424, 175, 43 },
		{ 424, 176, 43 },
		{ 424, 177, 43 },
		{ 424, 178, 43 },
		{ 424, 179, 43 },
		{ 424, 180, 43 },
		{ 424, 181, 43 },
		{ 424, 182, 43 },
		{ 424, 183, 43 },
		{ 424, 184, 43 },
		{ 424, 185, 43 },
		{ 424, 186, 43 },
		{ 424, 187, 43 },
		{ 424, 188, 43 },
		{ 424, 189, 43 },
		{ 424, 190, 43 },
		{ 424, 191, 43 },
		{ 424, 192, 43 },
		{ 424, 193, 43 },
		{ 424, 194, 43 },
		{ 424, 195, 43 },
		{ 424, 196, 43 },
		{ 424, 197, 43 },
		{ 424, 198, 43 },
		{ 424, 199, 43 },
		{ 424, 200, 43 },
		{ 424, 201, 43 },
		{ 424, 202, 43 },
		{ 424, 203, 43 },
		{ 424, 204, 43 },
		{ 424, 205, 43 },
		{ 424, 206, 43 },
		{ 424, 207, 43 },
		{ 424, 208, 43 },
		{ 424, 209, 43 },
		{ 424, 210, 43 },
		{ 424, 211, 43 },
		{ 424, 212, 43 },
		{ 424, 213, 43 },
		{ 424, 214, 43 },
		{ 424, 215, 43 },
		{ 425, 174, 43 },
		{ 425, 175, 43 },
		{ 425, 176, 43 },
		{ 425, 177, 43 },
		{ 425, 178, 43 },
		{ 425, 179, 43 },
		{ 425, 180, 43 },
		{ 425, 181, 43 },
		{ 425, 182, 43 },
		{ 425, 183, 43 },
		{ 425, 184, 43 },
		{ 425, 185, 43 },
		{ 425, 186, 43 },
		{ 425, 187, 43 },
		{ 425, 188, 43 },
		{ 425, 189, 43 },
		{ 425, 190, 43 },
		{ 425, 191, 43 },
		{ 425, 192, 43 },
		{ 425, 193, 43 },
		{ 425, 194, 43 },
		{ 425, 195, 43 },
		{ 425, 196, 43 },
		{ 425, 197, 43 },
		{ 425, 198, 43 },
		{ 425, 199, 43 },
		{ 425, 200, 43 },
		{ 425, 201, 43 },
		{ 425, 202, 43 },
		{ 425, 203, 43 },
		{ 425, 204, 43 },
		{ 425, 205, 43 },
		{ 425, 206, 43 },
		{ 425, 207, 43 },
		{ 425, 208, 43 },
		{ 425, 209, 43 },
		{ 425, 210, 43 },
		{ 425, 211, 43 },
		{ 425, 212, 43 },
		{ 425, 213, 43 },
		{ 425, 214, 43 },
		{ 426, 173, 43 },
		{ 426, 174, 43 },
		{ 426, 175, 43 },
		{ 426, 176, 43 },
		{ 426, 177, 43 },
		{ 426, 178, 43 },
		{ 426, 179, 43 },
		{ 426, 180, 43 },
		{ 426, 181, 43 },
		{ 426, 182, 43 },
		{ 426, 183, 43 },
		{ 426, 184, 43 },
		{ 426, 185, 43 },
		{ 426, 186, 43 },
		{ 426, 187, 43 },
		{ 426, 188, 43 },
		{ 426, 189, 43 },
		{ 426, 190, 43 },
		{ 426, 191, 43 },
		{ 426, 192, 43 },
		{ 426, 193, 43 },
		{ 426, 194, 43 },
		{ 426, 195, 43 },
		{ 426, 196, 43 },
		{ 426, 197, 43 },
		{ 426, 198, 43 },
		{ 426, 199, 43 },
		{ 426, 200, 43 },
		{ 426, 201, 43 },
		{ 426, 202, 43 },
		{ 426, 203, 43 },
		{ 426, 204, 43 },
		{ 426, 205, 43 },
		{ 426, 206, 43 },
		{ 426, 207, 43 },
		{ 426, 208, 43 },
		{ 426, 209, 43 },
		{ 426, 210, 43 },
		{ 426, 211, 43 },
		{ 426, 212, 43 },
		{ 426, 213, 43 },
		{ 426, 214, 43 },
		{ 427, 173, 43 },
		{ 427, 174, 43 },
		{ 427, 175, 43 },
		{ 427, 176, 43 },
		{ 427, 177, 43 },
		{ 427, 178, 43 },
		{ 427, 179, 43 },
		{ 427, 180, 43 },
		{ 427, 181, 43 },
		{ 427, 182, 43 },
		{ 427, 183, 43 },
		{ 427, 184, 43 },
		{ 427, 185, 43 },
		{ 427, 186, 43 },
		{ 427, 187, 43 },
		{ 427, 188, 43 },
		{ 427, 189, 43 },
		{ 427, 190, 43 },
		{ 427, 191, 43 },
		{ 427, 192, 43 },
		{ 427, 193, 43 },
		{ 427, 194, 43 },
		{ 427, 195, 43 },
		{ 427, 196, 43 },
		{ 427, 197, 43 },
		{ 427, 198, 43 },
		{ 427, 199, 43 },
		{ 427, 200, 43 },
		{ 427, 201, 43 },
		{ 427, 202, 43 },
		{ 427, 203, 43 },
		{ 427, 204, 43 },
		{ 427, 205, 43 },
		{ 427, 206, 43 },
		{ 427, 207, 43 },
		{ 427, 208, 43 },
		{ 427, 209, 43 },
		{ 427, 210, 43 },
		{ 427, 211, 43 },
		{ 427, 212, 43 },
		{ 427, 213, 43 },
		{ 427, 214, 43 },
		{ 428, 172, 43 },
		{ 428, 173, 43 },
		{ 428, 174, 43 },
		{ 428, 175, 43 },
		{ 428, 176, 43 },
		{ 428, 177, 43 },
		{ 428, 178, 43 },
		{ 428, 179, 43 },
		{ 428, 180, 43 },
		{ 428, 181, 43 },
		{ 428, 182, 43 },
		{ 428, 183, 43 },
		{ 428, 184, 43 },
		{ 428, 185, 43 },
		{ 428, 186, 43 },
		{ 428, 187, 43 },
		{ 428, 188, 43 },
		{ 428, 189, 43 },
		{ 428, 190, 43 },
		{ 428, 191, 43 },
		{ 428, 192, 43 },
		{ 428, 193, 43 },
		{ 428, 194, 43 },
		{ 428, 195, 43 },
		{ 428, 196, 43 },
		{ 428, 197, 43 },
		{ 428, 198, 43 },
		{ 428, 199, 43 },
		{ 428, 200, 43 },
		{ 428, 201, 43 },
		{ 428, 202, 43 },
		{ 428, 203, 43 },
		{ 428, 204, 43 },
		{ 428, 205, 43 },
		{ 428, 206, 43 },
		{ 428, 207, 43 },
		{ 428, 208, 43 },
		{ 428, 209, 43 },
		{ 428, 210, 43 },
		{ 428, 211, 43 },
		{ 428, 212, 43 },
		{ 428, 213, 43 },
		{ 428, 214, 43 },
		{ 429, 172, 43 },
		{ 429, 173, 43 },
		{ 429, 174, 43 },
		{ 429, 175, 43 },
		{ 429, 176, 43 },
		{ 429, 177, 43 },
		{ 429, 178, 43 },
		{ 429, 179, 43 },
		{ 429, 180, 43 },
		{ 429, 181, 43 },
		{ 429, 182, 43 },
		{ 429, 183, 43 },
		{ 429, 184, 43 },
		{ 429, 185, 43 },
		{ 429, 186, 43 },
		{ 429, 187, 43 },
		{ 429, 188, 43 },
		{ 429, 189, 43 },
		{ 429, 190, 43 },
		{ 429, 191, 43 },
		{ 429, 192, 43 },
		{ 429, 193, 43 },
		{ 429, 194, 43 },
		{ 429, 195, 43 },
		{ 429, 196, 43 },
		{ 429, 197, 43 },
		{ 429, 198, 43 },
		{ 429, 199, 43 },
		{ 429, 200, 43 },
		{ 429, 201, 43 },
		{ 429, 202, 43 },
		{ 429, 203, 43 },
		{ 429, 204, 43 },
		{ 429, 205, 43 },
		{ 429, 206, 43 },
		{ 429, 207, 43 },
		{ 429, 208, 43 },
		{ 429, 209, 43 },
		{ 429, 210, 43 },
		{ 429, 211, 43 },
		{ 429, 212, 43 },
		{ 429, 213, 43 },
		{ 430, 171, 43 },
		{ 430, 172, 43 },
		{ 430, 173, 43 },
		{ 430, 174, 43 },
		{ 430, 175, 43 },
		{ 430, 176, 43 },
		{ 430, 177, 43 },
		{ 430, 178, 43 },
		{ 430, 179, 43 },
		{ 430, 180, 43 },
		{ 430, 181, 43 },
		{ 430, 182, 43 },
		{ 430, 183, 43 },
		{ 430, 184, 43 },
		{ 430, 185, 43 },
		{ 430, 186, 43 },
		{ 430, 187, 43 },
		{ 430, 188, 43 },
		{ 430, 189, 43 },
		{ 430, 190, 43 },
		{ 430, 191, 43 },
		{ 430, 192, 43 },
		{ 430, 193, 43 },
		{ 430, 194, 43 },
		{ 430, 195, 43 },
		{ 430, 196, 43 },
		{ 430, 197, 43 },
		{ 430, 198, 43 },
		{ 430, 199, 43 },
		{ 430, 200, 43 },
		{ 430, 201, 43 },
		{ 430, 202, 43 },
		{ 430, 203, 43 },
		{ 430, 204, 43 },
		{ 430, 205, 43 },
		{ 430, 206, 43 },
		{ 430, 207, 43 },
		{ 430, 208, 43 },
		{ 430, 209, 43 },
		{ 430, 210, 43 },
		{ 430, 211, 43 },
		{ 430, 212, 43 },
		{ 430, 213, 43 },
		{ 431, 171, 43 },
		{ 431, 172, 43 },
		{ 431, 173, 43 },
		{ 431, 174, 43 },
		{ 431, 175, 43 },
		{ 431, 176, 43 },
		{ 431, 177, 43 },
		{ 431, 178, 43 },
		{ 431, 179, 43 },
		{ 431, 180, 43 },
		{ 431, 181, 43 },
		{ 431, 182, 43 },
		{ 431, 183, 43 },
		{ 431, 184, 43 },
		{ 431, 185, 43 },
		{ 431, 186, 43 },
		{ 431, 187, 43 },
		{ 431, 188, 43 },
		{ 431, 189, 43 },
		{ 431, 190, 43 },
		{ 431, 191, 43 },
		{ 431, 192, 43 },
		{ 431, 193, 43 },
		{ 431, 194, 43 },
		{ 431, 195, 43 },
		{ 431, 196, 43 },
		{ 431, 197, 43 },
		{ 431, 198, 43 },
		{ 431, 199, 43 },
		{ 431, 200, 43 },
		{ 431, 201, 43 },
		{ 431, 202, 43 },
		{ 431, 203, 43 },
		{ 431, 204, 43 },
		{ 431, 205, 43 },
		{ 431, 206, 43 },
		{ 431, 207, 43 },
		{ 431, 208, 43 },
		{ 431, 209, 43 },
		{ 431, 210, 43 },
		{ 431, 211, 43 },
		{ 431, 212, 43 },
		{ 431, 213, 43 },
		{ 432, 170, 43 },
		{ 432, 171, 43 },
		{ 432, 172, 43 },
		{ 432, 173, 43 },
		{ 432, 174, 43 },
		{ 432, 175, 43 },
		{ 432, 176, 43 },
		{ 432, 177, 43 },
		{ 432, 178, 43 },
		{ 432, 179, 43 },
		{ 432, 180, 43 },
		{ 432, 181, 43 },
		{ 432, 182, 43 },
		{ 432, 183, 43 },
		{ 432, 184, 43 },
		{ 432, 185, 43 },
		{ 432, 186, 43 },
		{ 432, 187, 43 },
		{ 432, 188, 43 },
		{ 432, 189, 43 },
		{ 432, 190, 43 },
		{ 432, 191, 43 },
		{ 432, 192, 43 },
		{ 432, 193, 43 },
		{ 432, 194, 43 },
		{ 432, 195, 43 },
		{ 432, 196, 43 },
		{ 432, 197, 43 },
		{ 432, 198, 43 },
		{ 432, 199, 43 },
		{ 432, 200, 43 },
		{ 432, 201, 43 },
		{ 432, 202, 43 },
		{ 432, 203, 43 },
		{ 432, 204, 43 },
		{ 432, 205, 43 },
		{ 432, 206, 43 },
		{ 432, 207, 43 },
		{ 432, 208, 43 },
		{ 432, 209, 43 },
		{ 432, 210, 43 },
		{ 432, 211, 43 },
		{ 432, 212, 43 },
		{ 432, 213, 43 },
		{ 433, 170, 43 },
		{ 433, 171, 43 },
		{ 433, 172, 43 },
		{ 433, 173, 43 },
		{ 433, 174, 43 },
		{ 433, 175, 43 },
		{ 433, 176, 43 },
		{ 433, 177, 43 },
		{ 433, 178, 43 },
		{ 433, 179, 43 },
		{ 433, 180, 43 },
		{ 433, 181, 43 },
		{ 433, 182, 43 },
		{ 433, 183, 43 },
		{ 433, 184, 43 },
		{ 433, 185, 43 },
		{ 433, 186, 43 },
		{ 433, 187, 43 },
		{ 433, 188, 43 },
		{ 433, 189, 43 },
		{ 433, 190, 43 },
		{ 433, 191, 43 },
		{ 433, 192, 43 },
		{ 433, 193, 43 },
		{ 433, 194, 43 },
		{ 433, 195, 43 },
		{ 433, 196, 43 },
		{ 433, 197, 43 },
		{ 433, 198, 43 },
		{ 433, 199, 43 },
		{ 433, 200, 43 },
		{ 433, 201, 43 },
		{ 433, 202, 43 },
		{ 433, 203, 43 },
		{ 433, 204, 43 },
		{ 433, 205, 43 },
		{ 433, 206, 43 },
		{ 433, 207, 43 },
		{ 433, 208, 43 },
		{ 433, 209, 43 },
		{ 433, 210, 43 },
		{ 433, 211, 43 },
		{ 433, 212, 43 },
		{ 434, 169, 43 },
		{ 434, 170, 43 },
		{ 434, 171, 43 },
		{ 434, 172, 43 },
		{ 434, 173, 43 },
		{ 434, 174, 43 },
		{ 434, 175, 43 },
		{ 434, 176, 43 },
		{ 434, 177, 43 },
		{ 434, 178, 43 },
		{ 434, 179, 43 },
		{ 434, 180, 43 },
		{ 434, 181, 43 },
		{ 434, 182, 43 },
		{ 434, 183, 43 },
		{ 434, 184, 43 },
		{ 434, 185, 43 },
		{ 434, 186, 43 },
		{ 434, 187, 43 },
		{ 434, 188, 43 },
		{ 434, 189, 43 },
		{ 434, 190, 43 },
		{ 434, 191, 43 },
		{ 434, 192, 43 },
		{ 434, 193, 43 },
		{ 434, 194, 43 },
		{ 434, 195, 43 },
		{ 434, 196, 43 },
		{ 434, 197, 43 },
		{ 434, 198, 43 },
		{ 434, 199, 43 },
		{ 434, 200, 43 },
		{ 434, 201, 43 },
		{ 434, 202, 43 },
		{ 434, 203, 43 },
		{ 434, 204, 43 },
		{ 434, 205, 43 },
		{ 434, 206, 43 },
		{ 434, 207, 43 },
		{ 434, 208, 43 },
		{ 434, 209, 43 },
		{ 434, 210, 43 },
		{ 434, 211, 43 },
		{ 434, 212, 43 },
		{ 435, 169, 43 },
		{ 435, 170, 43 },
		{ 435, 171, 43 },
		{ 435, 172, 43 },
		{ 435, 173, 43 },
		{ 435, 174, 43 },
		{ 435, 175, 43 },
		{ 435, 176, 43 },
		{ 435, 177, 43 },
		{ 435, 178, 43 },
		{ 435, 179, 43 },
		{ 435, 180, 43 },
		{ 435, 181, 43 },
		{ 435, 182, 43 },
		{ 435, 183, 43 },
		{ 435, 184, 43 },
		{ 435, 185, 43 },
		{ 435, 186, 43 },
		{ 435, 187, 43 },
		{ 435, 188, 43 },
		{ 435, 189, 43 },
		{ 435, 190, 43 },
		{ 435, 191, 43 },
		{ 435, 192, 43 },
		{ 435, 193, 43 },
		{ 435, 194, 43 },
		{ 435, 195, 43 },
		{ 435, 196, 43 },
		{ 435, 197, 43 },
		{ 435, 198, 43 },
		{ 435, 199, 43 },
		{ 435, 200, 43 },
		{ 435, 201, 43 },
		{ 435, 202, 43 },
		{ 435, 203, 43 },
		{ 435, 204, 43 },
		{ 435, 205, 43 },
		{ 435, 206, 43 },
		{ 435, 207, 43 },
		{ 435, 208, 43 },
		{ 435, 209, 43 },
		{ 435, 210, 43 },
		{ 435, 211, 43 },
		{ 435, 212, 43 },
		{ 436, 168, 43 },
		{ 436, 169, 43 },
		{ 436, 170, 43 },
		{ 436, 171, 43 },
		{ 436, 172, 43 },
		{ 436, 173, 43 },
		{ 436, 174, 43 },
		{ 436, 175, 43 },
		{ 436, 176, 43 },
		{ 436, 177, 43 },
		{ 436, 178, 43 },
		{ 436, 179, 43 },
		{ 436, 180, 43 },
		{ 436, 181, 43 },
		{ 436, 182, 43 },
		{ 436, 183, 43 },
		{ 436, 184, 43 },
		{ 436, 185, 43 },
		{ 436, 186, 43 },
		{ 436, 187, 43 },
		{ 436, 188, 43 },
		{ 436, 189, 43 },
		{ 436, 190, 43 },
		{ 436, 191, 43 },
		{ 436, 192, 43 },
		{ 436, 193, 43 },
		{ 436, 194, 43 },
		{ 436, 195, 43 },
		{ 436, 196, 43 },
		{ 436, 197, 43 },
		{ 436, 198, 43 },
		{ 436, 199, 43 },
		{ 436, 200, 43 },
		{ 436, 201, 43 },
		{ 436, 202, 43 },
		{ 436, 203, 43 },
		{ 436, 204, 43 },
		{ 436, 205, 43 },
		{ 436, 206, 43 },
		{ 436, 207, 43 },
		{ 436, 208, 43 },
		{ 436, 209, 43 },
		{ 436, 210, 43 },
		{ 436, 211, 43 },
		{ 436, 212, 43 },
		{ 437, 168, 43 },
		{ 437, 169, 43 },
		{ 437, 170, 43 },
		{ 437, 171, 43 },
		{ 437, 172, 43 },
		{ 437, 173, 43 },
		{ 437, 174, 43 },
		{ 437, 175, 43 },
		{ 437, 176, 43 },
		{ 437, 177, 43 },
		{ 437, 178, 43 },
		{ 437, 179, 43 },
		{ 437, 180, 43 },
		{ 437, 181, 43 },
		{ 437, 182, 43 },
		{ 437, 183, 43 },
		{ 437, 184, 43 },
		{ 437, 185, 43 },
		{ 437, 186, 43 },
		{ 437, 187, 43 },
		{ 437, 188, 43 },
		{ 437, 189, 43 },
		{ 437, 190, 43 },
		{ 437, 191, 43 },
		{ 437, 192, 43 },
		{ 437, 193, 43 },
		{ 437, 194, 43 },
		{ 437, 195, 43 },
		{ 437, 196, 43 },
		{ 437, 197, 43 },
		{ 437, 198, 43 },
		{ 437, 199, 43 },
		{ 437, 200, 43 },
		{ 437, 201, 43 },
		{ 437, 202, 43 },
		{ 437, 203, 43 },
		{ 437, 204, 43 },
		{ 437, 205, 43 },
		{ 437, 206, 43 },
		{ 437, 207, 43 },
		{ 437, 208, 43 },
		{ 437, 209, 43 },
		{ 437, 210, 43 },
		{ 437, 211, 43 },
		{ 438, 167, 43 },
		{ 438, 168, 43 },
		{ 438, 169, 43 },
		{ 438, 170, 43 },
		{ 438, 171, 43 },
		{ 438, 172, 43 },
		{ 438, 173, 43 },
		{ 438, 174, 43 },
		{ 438, 175, 43 },
		{ 438, 176, 43 },
		{ 438, 177, 43 },
		{ 438, 178, 43 },
		{ 438, 179, 43 },
		{ 438, 180, 43 },
		{ 438, 181, 43 },
		{ 438, 182, 43 },
		{ 438, 183, 43 },
		{ 438, 184, 43 },
		{ 438, 185, 43 },
		{ 438, 186, 43 },
		{ 438, 187, 43 },
		{ 438, 188, 43 },
		{ 438, 189, 43 },
		{ 438, 190, 43 },
		{ 438, 191, 43 },
		{ 438, 192, 43 },
		{ 438, 193, 43 },
		{ 438, 194, 43 },
		{ 438, 195, 43 },
		{ 438, 196, 43 },
		{ 438, 197, 43 },
		{ 438, 198, 43 },
		{ 438, 199, 43 },
		{ 438, 200, 43 },
		{ 438, 201, 43 },
		{ 438, 202, 43 },
		{ 438, 203, 43 },
		{ 438, 204, 43 },
		{ 438, 205, 43 },
		{ 438, 206, 43 },
		{ 438, 207, 43 },
		{ 438, 208, 43 },
		{ 438, 209, 43 },
		{ 438, 210, 43 },
		{ 438, 211, 43 },
		{ 439, 167, 43 },
		{ 439, 168, 43 },
		{ 439, 169, 43 },
		{ 439, 170, 43 },
		{ 439, 171, 43 },
		{ 439, 172, 43 },
		{ 439, 173, 43 },
		{ 439, 174, 43 },
		{ 439, 175, 43 },
		{ 439, 176, 43 },
		{ 439, 177, 43 },
		{ 439, 178, 43 },
		{ 439, 179, 43 },
		{ 439, 180, 43 },
		{ 439, 181, 43 },
		{ 439, 182, 43 },
		{ 439, 183, 43 },
		{ 439, 184, 43 },
		{ 439, 185, 43 },
		{ 439, 186, 43 },
		{ 439, 187, 43 },
		{ 439, 188, 43 },
		{ 439, 189, 43 },
		{ 439, 190, 43 },
		{ 439, 191, 43 },
		{ 439, 192, 43 },
		{ 439, 193, 43 },
		{ 439, 194, 43 },
		{ 439, 195, 43 },
		{ 439, 196, 43 },
		{ 439, 197, 43 },
		{ 439, 198, 43 },
		{ 439, 199, 43 },
		{ 439, 200, 43 },
		{ 439, 201, 43 },
		{ 439, 202, 43 },
		{ 439, 203, 43 },
		{ 439, 204, 43 },
		{ 439, 205, 43 },
		{ 439, 206, 43 },
		{ 439, 207, 43 },
		{ 439, 208, 43 },
		{ 439, 209, 43 },
		{ 439, 210, 43 },
		{ 439, 211, 43 },
		{ 440, 166, 43 },
		{ 440, 167, 43 },
		{ 440, 168, 43 },
		{ 440, 169, 43 },
		{ 440, 170, 43 },
		{ 440, 171, 43 },
		{ 440, 172, 43 },
		{ 440, 173, 43 },
		{ 440, 174, 43 },
		{ 440, 175, 43 },
		{ 440, 176, 43 },
		{ 440, 177, 43 },
		{ 440, 178, 43 },
		{ 440, 179, 43 },
		{ 440, 180, 43 },
		{ 440, 181, 43 },
		{ 440, 182, 43 },
		{ 440, 183, 43 },
		{ 440, 184, 43 },
		{ 440, 185, 43 },
		{ 440, 186, 43 },
		{ 440, 187, 43 },
		{ 440, 188, 43 },
		{ 440, 189, 43 },
		{ 440, 190, 43 },
		{ 440, 191, 43 },
		{ 440, 192, 43 },
		{ 440, 193, 43 },
		{ 440, 194, 43 },
		{ 440, 195, 43 },
		{ 440, 196, 43 },
		{ 440, 197, 43 },
		{ 440, 198, 43 },
		{ 440, 199, 43 },
		{ 440, 200, 43 },
		{ 440, 201, 43 },
		{ 440, 202, 43 },
		{ 440, 203, 43 },
		{ 440, 204, 43 },
		{ 440, 205, 43 },
		{ 440, 206, 43 },
		{ 440, 207, 43 },
		{ 440, 208, 43 },
		{ 440, 209, 43 },
		{ 440, 210, 43 },
		{ 440, 211, 43 },
		{ 441, 166, 43 },
		{ 441, 167, 43 },
		{ 441, 168, 43 },
		{ 441, 169, 43 },
		{ 441, 170, 43 },
		{ 441, 171, 43 },
		{ 441, 172, 43 },
		{ 441, 173, 43 },
		{ 441, 174, 43 },
		{ 441, 175, 43 },
		{ 441, 176, 43 },
		{ 441, 177, 43 },
		{ 441, 178, 43 },
		{ 441, 179, 43 },
		{ 441, 180, 43 },
		{ 441, 181, 43 },
		{ 441, 182, 43 },
		{ 441, 183, 43 },
		{ 441, 184, 43 },
		{ 441, 185, 43 },
		{ 441, 186, 43 },
		{ 441, 187, 43 },
		{ 441, 188, 43 },
		{ 441, 189, 43 },
		{ 441, 190, 43 },
		{ 441, 191, 43 },
		{ 441, 192, 43 },
		{ 441, 193, 43 },
		{ 441, 194, 43 },
		{ 441, 195, 43 },
		{ 441, 196, 43 },
		{ 441, 197, 43 },
		{ 441, 198, 43 },
		{ 441, 199, 43 },
		{ 441, 200, 43 },
		{ 441, 201, 43 },
		{ 441, 202, 43 },
		{ 441, 203, 43 },
		{ 441, 204, 43 },
		{ 441, 205, 43 },
		{ 441, 206, 43 },
		{ 441, 207, 43 },
		{ 441, 208, 43 },
		{ 441, 209, 43 },
		{ 441, 210, 43 },
		{ 442, 165, 43 },
		{ 442, 166, 43 },
		{ 442, 167, 43 },
		{ 442, 168, 43 },
		{ 442, 169, 43 },
		{ 442, 170, 43 },
		{ 442, 171, 43 },
		{ 442, 172, 43 },
		{ 442, 173, 43 },
		{ 442, 174, 43 },
		{ 442, 175, 43 },
		{ 442, 176, 43 },
		{ 442, 177, 43 },
		{ 442, 178, 43 },
		{ 442, 179, 43 },
		{ 442, 180, 43 },
		{ 442, 181, 43 },
		{ 442, 182, 43 },
		{ 442, 183, 43 },
		{ 442, 184, 43 },
		{ 442, 185, 43 },
		{ 442, 186, 43 },
		{ 442, 187, 43 },
		{ 442, 188, 43 },
		{ 442, 189, 43 },
		{ 442, 190, 43 },
		{ 442, 191, 43 },
		{ 442, 192, 43 },
		{ 442, 193, 43 },
		{ 442, 194, 43 },
		{ 442, 195, 43 },
		{ 442, 196, 43 },
		{ 442, 197, 43 },
		{ 442, 198, 43 },
		{ 442, 199, 43 },
		{ 442, 200, 43 },
		{ 442, 201, 43 },
		{ 442, 202, 43 },
		{ 442, 203, 43 },
		{ 442, 204, 43 },
		{ 442, 205, 43 },
		{ 442, 206, 43 },
		{ 442, 207, 43 },
		{ 442, 208, 43 },
		{ 442, 209, 43 },
		{ 442, 210, 43 },
		{ 443, 165, 43 },
		{ 443, 166, 43 },
		{ 443, 167, 43 },
		{ 443, 168, 43 },
		{ 443, 169, 43 },
		{ 443, 170, 43 },
		{ 443, 171, 43 },
		{ 443, 172, 43 },
		{ 443, 173, 43 },
		{ 443, 174, 43 },
		{ 443, 175, 43 },
		{ 443, 176, 43 },
		{ 443, 177, 43 },
		{ 443, 178, 43 },
		{ 443, 179, 43 },
		{ 443, 180, 43 },
		{ 443, 181, 43 },
		{ 443, 182, 43 },
		{ 443, 183, 43 },
		{ 443, 184, 43 },
		{ 443, 185, 43 },
		{ 443, 186, 43 },
		{ 443, 187, 43 },
		{ 443, 188, 43 },
		{ 443, 189, 43 },
		{ 443, 190, 43 },
		{ 443, 191, 43 },
		{ 443, 192, 43 },
		{ 443, 193, 43 },
		{ 443, 194, 43 },
		{ 443, 195, 43 },
		{ 443, 196, 43 },
		{ 443, 197, 43 },
		{ 443, 198, 43 },
		{ 443, 199, 43 },
		{ 443, 200, 43 },
		{ 443, 201, 43 },
		{ 443, 202, 43 },
		{ 443, 203, 43 },
		{ 443, 204, 43 },
		{ 443, 205, 43 },
		{ 443, 206, 43 },
		{ 443, 207, 43 },
		{ 443, 208, 43 },
		{ 443, 209, 43 },
		{ 443, 210, 43 },
		{ 444, 164, 43 },
		{ 444, 165, 43 },
		{ 444, 166, 43 },
		{ 444, 167, 43 },
		{ 444, 168, 43 },
		{ 444, 169, 43 },
		{ 444, 170, 43 },
		{ 444, 171, 43 },
		{ 444, 172, 43 },
		{ 444, 173, 43 },
		{ 444, 174, 43 },
		{ 444, 175, 43 },
		{ 444, 176, 43 },
		{ 444, 177, 43 },
		{ 444, 178, 43 },
		{ 444, 179, 43 },
		{ 444, 180, 43 },
		{ 444, 181, 43 },
		{ 444, 182, 43 },
		{ 444, 183, 43 },
		{ 444, 184, 43 },
		{ 444, 185, 43 },
		{ 444, 186, 43 },
		{ 444, 187, 43 },
		{ 444, 188, 43 },
		{ 444, 189, 43 },
		{ 444, 190, 43 },
		{ 444, 191, 43 },
		{ 444, 192, 43 },
		{ 444, 193, 43 },
		{ 444, 194, 43 },
		{ 444, 195, 43 },
		{ 444, 196, 43 },
		{ 444, 197, 43 },
		{ 444, 198, 43 },
		{ 444, 199, 43 },
		{ 444, 200, 43 },
		{ 444, 201, 43 },
		{ 444, 202, 43 },
		{ 444, 203, 43 },
		{ 444, 204, 43 },
		{ 444, 205, 43 },
		{ 444, 206, 43 },
		{ 444, 207, 43 },
		{ 444, 208, 43 },
		{ 444, 209, 43 },
		{ 444, 210, 43 },
		{ 445, 164, 43 },
		{ 445, 165, 43 },
		{ 445, 166, 43 },
		{ 445, 167, 43 },
		{ 445, 168, 43 },
		{ 445, 169, 43 },
		{ 445, 170, 43 },
		{ 445, 171, 43 },
		{ 445, 172, 43 },
		{ 445, 173, 43 },
		{ 445, 174, 43 },
		{ 445, 175, 43 },
		{ 445, 176, 43 },
		{ 445, 177, 43 },
		{ 445, 178, 43 },
		{ 445, 179, 43 },
		{ 445, 180, 43 },
		{ 445, 181, 43 },
		{ 445, 182, 43 },
		{ 445, 183, 43 },
		{ 445, 184, 43 },
		{ 445, 185, 43 },
		{ 445, 186, 43 },
		{ 445, 187, 43 },
		{ 445, 188, 43 },
		{ 445, 189, 43 },
		{ 445, 190, 43 },
		{ 445, 191, 43 },
		{ 445, 192, 43 },
		{ 445, 193, 43 },
		{ 445, 194, 43 },
		{ 445, 195, 43 },
		{ 445, 196, 43 },
		{ 445, 197, 43 },
		{ 445, 198, 43 },
		{ 445, 199, 43 },
		{ 445, 200, 43 },
		{ 445, 201, 43 },
		{ 445, 202, 43 },
		{ 445, 203, 43 },
		{ 445, 204, 43 },
		{ 445, 205, 43 },
		{ 445, 206, 43 },
		{ 445, 207, 43 },
		{ 445, 208, 43 },
		{ 445, 209, 43 },
		{ 446, 163, 43 },
		{ 446, 164, 43 },
		{ 446, 165, 43 },
		{ 446, 166, 43 },
		{ 446, 167, 43 },
		{ 446, 168, 43 },
		{ 446, 169, 43 },
		{ 446, 170, 43 },
		{ 446, 171, 43 },
		{ 446, 172, 43 },
		{ 446, 173, 43 },
		{ 446, 174, 43 },
		{ 446, 175, 43 },
		{ 446, 176, 43 },
		{ 446, 177, 43 },
		{ 446, 178, 43 },
		{ 446, 179, 43 },
		{ 446, 180, 43 },
		{ 446, 181, 43 },
		{ 446, 182, 43 },
		{ 446, 183, 43 },
		{ 446, 184, 43 },
		{ 446, 185, 43 },
		{ 446, 186, 43 },
		{ 446, 187, 43 },
		{ 446, 188, 43 },
		{ 446, 189, 43 },
		{ 446, 190, 43 },
		{ 446, 191, 43 },
		{ 446, 192, 43 },
		{ 446, 193, 43 },
		{ 446, 194, 43 },
		{ 446, 195, 43 },
		{ 446, 196, 43 },
		{ 446, 197, 43 },
		{ 446, 198, 43 },
		{ 446, 199, 43 },
		{ 446, 200, 43 },
		{ 446, 201, 43 },
		{ 446, 202, 43 },
		{ 446, 203, 43 },
		{ 446, 204, 43 },
		{ 446, 205, 43 },
		{ 446, 206, 43 },
		{ 446, 207, 43 },
		{ 446, 208, 43 },
		{ 446, 209, 43 },
		{ 447, 163, 43 },
		{ 447, 164, 43 },
		{ 447, 165, 43 },
		{ 447, 166, 43 },
		{ 447, 167, 43 },
		{ 447, 168, 43 },
		{ 447, 169, 43 },
		{ 447, 170, 43 },
		{ 447, 171, 43 },
		{ 447, 172, 43 },
		{ 447, 173, 43 },
		{ 447, 174, 43 },
		{ 447, 175, 43 },
		{ 447, 176, 43 },
		{ 447, 177, 43 },
		{ 447, 178, 43 },
		{ 447, 179, 43 },
		{ 447, 180, 43 },
		{ 447, 181, 43 },
		{ 447, 182, 43 },
		{ 447, 183, 43 },
		{ 447, 184, 43 },
		{ 447, 185, 43 },
		{ 447, 186, 43 },
		{ 447, 187, 43 },
		{ 447, 188, 43 },
		{ 447, 189, 43 },
		{ 447, 190, 43 },
		{ 447, 191, 43 },
		{ 447, 192, 43 },
		{ 447, 193, 43 },
		{ 447, 194, 43 },
		{ 447, 195, 43 },
		{ 447, 196, 43 },
		{ 447, 197, 43 },
		{ 447, 198, 43 },
		{ 447, 199, 43 },
		{ 447, 200, 43 },
		{ 447, 201, 43 },
		{ 447, 202, 43 },
		{ 447, 203, 43 },
		{ 447, 204, 43 },
		{ 447, 205, 43 },
		{ 447, 206, 43 },
		{ 447, 207, 43 },
		{ 447, 208, 43 },
		{ 447, 209, 43 },
		{ 448, 162, 43 },
		{ 448, 163, 43 },
		{ 448, 164, 43 },
		{ 448, 165, 43 },
		{ 448, 166, 43 },
		{ 448, 167, 43 },
		{ 448, 168, 43 },
		{ 448, 169, 43 },
		{ 448, 170, 43 },
		{ 448, 171, 43 },
		{ 448, 172, 43 },
		{ 448, 173, 43 },
		{ 448, 174, 43 },
		{ 448, 175, 43 },
		{ 448, 176, 43 },
		{ 448, 177, 43 },
		{ 448, 178, 43 },
		{ 448, 179, 43 },
		{ 448, 180, 43 },
		{ 448, 181, 43 },
		{ 448, 182, 43 },
		{ 448, 183, 43 },
		{ 448, 184, 43 },
		{ 448, 185, 43 },
		{ 448, 186, 43 },
		{ 448, 187, 43 },
		{ 448, 188, 43 },
		{ 448, 189, 43 },
		{ 448, 190, 43 },
		{ 448, 191, 43 },
		{ 448, 192, 43 },
		{ 448, 193, 43 },
		{ 448, 194, 43 },
		{ 448, 195, 43 },
		{ 448, 196, 43 },
		{ 448, 197, 43 },
		{ 448, 198, 43 },
		{ 448, 199, 43 },
		{ 448, 200, 43 },
		{ 448, 201, 43 },
		{ 448, 202, 43 },
		{ 448, 203, 43 },
		{ 448, 204, 43 },
		{ 448, 205, 43 },
		{ 448, 206, 43 },
		{ 448, 207, 43 },
		{ 448, 208, 43 },
		{ 448, 209, 43 },
		{ 449, 162, 43 },
		{ 449, 163, 43 },
		{ 449, 164, 43 },
		{ 449, 165, 43 },
		{ 449, 166, 43 },
		{ 449, 167, 43 },
		{ 449, 168, 43 },
		{ 449, 169, 43 },
		{ 449, 170, 43 },
		{ 449, 171, 43 },
		{ 449, 172, 43 },
		{ 449, 173, 43 },
		{ 449, 174, 43 },
		{ 449, 175, 43 },
		{ 449, 176, 43 },
		{ 449, 177, 43 },
		{ 449, 178, 43 },
		{ 449, 179, 43 },
		{ 449, 180, 43 },
		{ 449, 181, 43 },
		{ 449, 182, 43 },
		{ 449, 183, 43 },
		{ 449, 184, 43 },
		{ 449, 185, 43 },
		{ 449, 186, 43 },
		{ 449, 187, 43 },
		{ 449, 188, 43 },
		{ 449, 189, 43 },
		{ 449, 190, 43 },
		{ 449, 191, 43 },
		{ 449, 192, 43 },
		{ 449, 193, 43 },
		{ 449, 194, 43 },
		{ 449, 195, 43 },
		{ 449, 196, 43 },
		{ 449, 197, 43 },
		{ 449, 198, 43 },
		{ 449, 199, 43 },
		{ 449, 200, 43 },
		{ 449, 201, 43 },
		{ 449, 202, 43 },
		{ 449, 203, 43 },
		{ 449, 204, 43 },
		{ 449, 205, 43 },
		{ 449, 206, 43 },
		{ 449, 207, 43 },
		{ 449, 208, 43 },
		{ 450, 161, 43 },
		{ 450, 162, 43 },
		{ 450, 163, 43 },
		{ 450, 164, 43 },
		{ 450, 165, 43 },
		{ 450, 166, 43 },
		{ 450, 167, 43 },
		{ 450, 168, 43 },
		{ 450, 169, 43 },
		{ 450, 170, 43 },
		{ 450, 171, 43 },
		{ 450, 172, 43 },
		{ 450, 173, 43 },
		{ 450, 174, 43 },
		{ 450, 175, 43 },
		{ 450, 176, 43 },
		{ 450, 177, 43 },
		{ 450, 178, 43 },
		{ 450, 179, 43 },
		{ 450, 180, 43 },
		{ 450, 181, 43 },
		{ 450, 182, 43 },
		{ 450, 183, 43 },
		{ 450, 184, 43 },
		{ 450, 185, 43 },
		{ 450, 186, 43 },
		{ 450, 187, 43 },
		{ 450, 188, 43 },
		{ 450, 189, 43 },
		{ 450, 190, 43 },
		{ 450, 191, 43 },
		{ 450, 192, 43 },
		{ 450, 193, 43 },
		{ 450, 194, 43 },
		{ 450, 195, 43 },
		{ 450, 196, 43 },
		{ 450, 197, 43 },
		{ 450, 198, 43 },
		{ 450, 199, 43 },
		{ 450, 200, 43 },
		{ 450, 201, 43 },
		{ 450, 202, 43 },
		{ 450, 203, 43 },
		{ 450, 204, 43 },
		{ 450, 205, 43 },
		{ 450, 206, 43 },
		{ 450, 207, 43 },
		{ 450, 208, 43 },
		{ 451, 161, 43 },
		{ 451, 162, 43 },
		{ 451, 163, 43 },
		{ 451, 164, 43 },
		{ 451, 165, 43 },
		{ 451, 166, 43 },
		{ 451, 167, 43 },
		{ 451, 168, 43 },
		{ 451, 169, 43 },
		{ 451, 170, 43 },
		{ 451, 171, 43 },
		{ 451, 172, 43 },
		{ 451, 173, 43 },
		{ 451, 174, 43 },
		{ 451, 175, 43 },
		{ 451, 176, 43 },
		{ 451, 177, 43 },
		{ 451, 178, 43 },
		{ 451, 179, 43 },
		{ 451, 180, 43 },
		{ 451, 181, 43 },
		{ 451, 182, 43 },
		{ 451, 183, 43 },
		{ 451, 184, 43 },
		{ 451, 185, 43 },
		{ 451, 186, 43 },
		{ 451, 187, 43 },
		{ 451, 188, 43 },
		{ 451, 189, 43 },
		{ 451, 190, 43 },
		{ 451, 191, 43 },
		{ 451, 192, 43 },
		{ 451, 193, 43 },
		{ 451, 194, 43 },
		{ 451, 195, 43 },
		{ 451, 196, 43 },
		{ 451, 197, 43 },
		{ 451, 198, 43 },
		{ 451, 199, 43 },
		{ 451, 200, 43 },
		{ 451, 201, 43 },
		{ 451, 202, 43 },
		{ 451, 203, 43 },
		{ 451, 204, 43 },
		{ 451, 205, 43 },
		{ 451, 206, 43 },
		{ 451, 207, 43 },
		{ 451, 208, 43 },
		{ 452, 160, 43 },
		{ 452, 161, 43 },
		{ 452, 162, 43 },
		{ 452, 163, 43 },
		{ 452, 164, 43 },
		{ 452, 165, 43 },
		{ 452, 166, 43 },
		{ 452, 167, 43 },
		{ 452, 168, 43 },
		{ 452, 169, 43 },
		{ 452, 170, 43 },
		{ 452, 171, 43 },
		{ 452, 172, 43 },
		{ 452, 173, 43 },
		{ 452, 174, 43 },
		{ 452, 175, 43 },
		{ 452, 176, 43 },
		{ 452, 177, 43 },
		{ 452, 178, 43 },
		{ 452, 179, 43 },
		{ 452, 180, 43 },
		{ 452, 181, 43 },
		{ 452, 182, 43 },
		{ 452, 183, 43 },
		{ 452, 184, 43 },
		{ 452, 185, 43 },
		{ 452, 186, 43 },
		{ 452, 187, 43 },
		{ 452, 188, 43 },
		{ 452, 189, 43 },
		{ 452, 190, 43 },
		{ 452, 191, 43 },
		{ 452, 192, 43 },
		{ 452, 193, 43 },
		{ 452, 194, 43 },
		{ 452, 195, 43 },
		{ 452, 196, 43 },
		{ 452, 197, 43 },
		{ 452, 198, 43 },
		{ 452, 199, 43 },
		{ 452, 200, 43 },
		{ 452, 201, 43 },
		{ 452, 202, 43 },
		{ 452, 203, 43 },
		{ 452, 204, 43 },
		{ 452, 205, 43 },
		{ 452, 206, 43 },
		{ 452, 207, 43 },
		{ 452, 208, 43 },
		{ 453, 160, 43 },
		{ 453, 161, 43 },
		{ 453, 162, 43 },
		{ 453, 163, 43 },
		{ 453, 164, 43 },
		{ 453, 165, 43 },
		{ 453, 166, 43 },
		{ 453, 167, 43 },
		{ 453, 168, 43 },
		{ 453, 169, 43 },
		{ 453, 170, 43 },
		{ 453, 171, 43 },
		{ 453, 172, 43 },
		{ 453, 173, 43 },
		{ 453, 174, 43 },
		{ 453, 175, 43 },
		{ 453, 176, 43 },
		{ 453, 177, 43 },
		{ 453, 178, 43 },
		{ 453, 179, 43 },
		{ 453, 180, 43 },
		{ 453, 181, 43 },
		{ 453, 182, 43 },
		{ 453, 183, 43 },
		{ 453, 184, 43 },
		{ 453, 185, 43 },
		{ 453, 186, 43 },
		{ 453, 187, 43 },
		{ 453, 188, 43 },
		{ 453, 189, 43 },
		{ 453, 190, 43 },
		{ 453, 191, 43 },
		{ 453, 192, 43 },
		{ 453, 193, 43 },
		{ 453, 194, 43 },
		{ 453, 195, 43 },
		{ 453, 196, 43 },
		{ 453, 197, 43 },
		{ 453, 198, 43 },
		{ 453, 199, 43 },
		{ 453, 200, 43 },
		{ 453, 201, 43 },
		{ 453, 202, 43 },
		{ 453, 203, 43 },
		{ 453, 204, 43 },
		{ 453, 205, 43 },
		{ 453, 206, 43 },
		{ 453, 207, 43 },
		{ 454, 159, 43 },
		{ 454, 160, 43 },
		{ 454, 161, 43 },
		{ 454, 162, 43 },
		{ 454, 163, 43 },
		{ 454, 164, 43 },
		{ 454, 165, 43 },
		{ 454, 166, 43 },
		{ 454, 167, 43 },
		{ 454, 168, 43 },
		{ 454, 169, 43 },
		{ 454, 170, 43 },
		{ 454, 171, 43 },
		{ 454, 172, 43 },
		{ 454, 173, 43 },
		{ 454, 174, 43 },
		{ 454, 175, 43 },
		{ 454, 176, 43 },
		{ 454, 177, 43 },
		{ 454, 178, 43 },
		{ 454, 179, 43 },
		{ 454, 180, 43 },
		{ 454, 181, 43 },
		{ 454, 182, 43 },
		{ 454, 183, 43 },
		{ 454, 184, 43 },
		{ 454, 185, 43 },
		{ 454, 186, 43 },
		{ 454, 187, 43 },
		{ 454, 188, 43 },
		{ 454, 189, 43 },
		{ 454, 190, 43 },
		{ 454, 191, 43 },
		{ 454, 192, 43 },
		{ 454, 193, 43 },
		{ 454, 194, 43 },
		{ 454, 195, 43 },
		{ 454, 196, 43 },
		{ 454, 197, 43 },
		{ 454, 198, 43 },
		{ 454, 199, 43 },
		{ 454, 200, 43 },
		{ 454, 201, 43 },
		{ 454, 202, 43 },
		{ 454, 203, 43 },
		{ 454, 204, 43 },
		{ 454, 205, 43 },
		{ 454, 206, 43 },
		{ 454, 207, 43 },
		{ 455, 159, 43 },
		{ 455, 160, 43 },
		{ 455, 161, 43 },
		{ 455, 162, 43 },
		{ 455, 163, 43 },
		{ 455, 164, 43 },
		{ 455, 165, 43 },
		{ 455, 166, 43 },
		{ 455, 167, 43 },
		{ 455, 168, 43 },
		{ 455, 169, 43 },
		{ 455, 170, 43 },
		{ 455, 171, 43 },
		{ 455, 172, 43 },
		{ 455, 173, 43 },
		{ 455, 174, 43 },
		{ 455, 175, 43 },
		{ 455, 176, 43 },
		{ 455, 177, 43 },
		{ 455, 178, 43 },
		{ 455, 179, 43 },
		{ 455, 180, 43 },
		{ 455, 181, 43 },
		{ 455, 182, 43 },
		{ 455, 183, 43 },
		{ 455, 184, 43 },
		{ 455, 185, 43 },
		{ 455, 186, 43 },
		{ 455, 187, 43 },
		{ 455, 188, 43 },
		{ 455, 189, 43 },
		{ 455, 190, 43 },
		{ 455, 191, 43 },
		{ 455, 192, 43 },
		{ 455, 193, 43 },
		{ 455, 194, 43 },
		{ 455, 195, 43 },
		{ 455, 196, 43 },
		{ 455, 197, 43 },
		{ 455, 198, 43 },
		{ 455, 199, 43 },
		{ 455, 200, 43 },
		{ 455, 201, 43 },
		{ 455, 202, 43 },
		{ 455, 203, 43 },
		{ 455, 204, 43 },
		{ 455, 205, 43 },
		{ 455, 206, 43 },
		{ 455, 207, 43 },
		{ 456, 158, 43 },
		{ 456, 159, 43 },
		{ 456, 160, 43 },
		{ 456, 161, 43 },
		{ 456, 162, 43 },
		{ 456, 163, 43 },
		{ 456, 164, 43 },
		{ 456, 165, 43 },
		{ 456, 166, 43 },
		{ 456, 167, 43 },
		{ 456, 168, 43 },
		{ 456, 169, 43 },
		{ 456, 170, 43 },
		{ 456, 171, 43 },
		{ 456, 172, 43 },
		{ 456, 173, 43 },
		{ 456, 174, 43 },
		{ 456, 175, 43 },
		{ 456, 176, 43 },
		{ 456, 177, 43 },
		{ 456, 178, 43 },
		{ 456, 179, 43 },
		{ 456, 180, 43 },
		{ 456, 181, 43 },
		{ 456, 182, 43 },
		{ 456, 183, 43 },
		{ 456, 184, 43 },
		{ 456, 185, 43 },
		{ 456, 186, 43 },
		{ 456, 187, 43 },
		{ 456, 188, 43 },
		{ 456, 189, 43 },
		{ 456, 190, 43 },
		{ 456, 191, 43 },
		{ 456, 192, 43 },
		{ 456, 193, 43 },
		{ 456, 194, 43 },
		{ 456, 195, 43 },
		{ 456, 196, 43 },
		{ 456, 197, 43 },
		{ 456, 198, 43 },
		{ 456, 199, 43 },
		{ 456, 200, 43 },
		{ 456, 201, 43 },
		{ 456, 202, 43 },
		{ 456, 203, 43 },
		{ 456, 204, 43 },
		{ 456, 205, 43 },
		{ 456, 206, 43 },
		{ 456, 207, 43 },
		{ 457, 158, 43 },
		{ 457, 159, 43 },
		{ 457, 160, 43 },
		{ 457, 161, 43 },
		{ 457, 162, 43 },
		{ 457, 163, 43 },
		{ 457, 164, 43 },
		{ 457, 165, 43 },
		{ 457, 166, 43 },
		{ 457, 167, 43 },
		{ 457, 168, 43 },
		{ 457, 169, 43 },
		{ 457, 170, 43 },
		{ 457, 171, 43 },
		{ 457, 172, 43 },
		{ 457, 173, 43 },
		{ 457, 174, 43 },
		{ 457, 175, 43 },
		{ 457, 176, 43 },
		{ 457, 177, 43 },
		{ 457, 178, 43 },
		{ 457, 179, 43 },
		{ 457, 180, 43 },
		{ 457, 181, 43 },
		{ 457, 182, 43 },
		{ 457, 183, 43 },
		{ 457, 184, 43 },
		{ 457, 185, 43 },
		{ 457, 186, 43 },
		{ 457, 187, 43 },
		{ 457, 188, 43 },
		{ 457, 189, 43 },
		{ 457, 190, 43 },
		{ 457, 191, 43 },
		{ 457, 192, 43 },
		{ 457, 193, 43 },
		{ 457, 194, 43 },
		{ 457, 195, 43 },
		{ 457, 196, 43 },
		{ 457, 197, 43 },
		{ 457, 198, 43 },
		{ 457, 199, 43 },
		{ 457, 200, 43 },
		{ 457, 201, 43 },
		{ 457, 202, 43 },
		{ 457, 203, 43 },
		{ 457, 204, 43 },
		{ 457, 205, 43 },
		{ 457, 206, 43 },
		{ 458, 157, 43 },
		{ 458, 158, 43 },
		{ 458, 159, 43 },
		{ 458, 160, 43 },
		{ 458, 161, 43 },
		{ 458, 162, 43 },
		{ 458, 163, 43 },
		{ 458, 164, 43 },
		{ 458, 165, 43 },
		{ 458, 166, 43 },
		{ 458, 167, 43 },
		{ 458, 168, 43 },
		{ 458, 169, 43 },
		{ 458, 170, 43 },
		{ 458, 171, 43 },
		{ 458, 172, 43 },
		{ 458, 173, 43 },
		{ 458, 174, 43 },
		{ 458, 175, 43 },
		{ 458, 176, 43 },
		{ 458, 177, 43 },
		{ 458, 178, 43 },
		{ 458, 179, 43 },
		{ 458, 180, 43 },
		{ 458, 181, 43 },
		{ 458, 182, 43 },
		{ 458, 183, 43 },
		{ 458, 184, 43 },
		{ 458, 185, 43 },
		{ 458, 186, 43 },
		{ 458, 187, 43 },
		{ 458, 188, 43 },
		{ 458, 189, 43 },
		{ 458, 190, 43 },
		{ 458, 191, 43 },
		{ 458, 192, 43 },
		{ 458, 193, 43 },
		{ 458, 194, 43 },
		{ 458, 195, 43 },
		{ 458, 196, 43 },
		{ 458, 197, 43 },
		{ 458, 198, 43 },
		{ 458, 199, 43 },
		{ 458, 200, 43 },
		{ 458, 201, 43 },
		{ 458, 202, 43 },
		{ 458, 203, 43 },
		{ 458, 204, 43 },
		{ 458, 205, 43 },
		{ 458, 206, 43 },
		{ 459, 157, 43 },
		{ 459, 158, 43 },
		{ 459, 159, 43 },
		{ 459, 160, 43 },
		{ 459, 161, 43 },
		{ 459, 162, 43 },
		{ 459, 163, 43 },
		{ 459, 164, 43 },
		{ 459, 165, 43 },
		{ 459, 166, 43 },
		{ 459, 167, 43 },
		{ 459, 168, 43 },
		{ 459, 169, 43 },
		{ 459, 170, 43 },
		{ 459, 171, 43 },
		{ 459, 172, 43 },
		{ 459, 173, 43 },
		{ 459, 174, 43 },
		{ 459, 175, 43 },
		{ 459, 176, 43 },
		{ 459, 177, 43 },
		{ 459, 178, 43 },
		{ 459, 179, 43 },
		{ 459, 180, 43 },
		{ 459, 181, 43 },
		{ 459, 182, 43 },
		{ 459, 183, 43 },
		{ 459, 184, 43 },
		{ 459, 185, 43 },
		{ 459, 186, 43 },
		{ 459, 187, 43 },
		{ 459, 188, 43 },
		{ 459, 189, 43 },
		{ 459, 190, 43 },
		{ 459, 191, 43 },
		{ 459, 192, 43 },
		{ 459, 193, 43 },
		{ 459, 194, 43 },
		{ 459, 195, 43 },
		{ 459, 196, 43 },
		{ 459, 197, 43 },
		{ 459, 198, 43 },
		{ 459, 199, 43 },
		{ 459, 200, 43 },
		{ 459, 201, 43 },
		{ 459, 202, 43 },
		{ 459, 203, 43 },
		{ 459, 204, 43 },
		{ 459, 205, 43 },
		{ 459, 206, 43 },
		{ 460, 156, 43 },
		{ 460, 157, 43 },
		{ 460, 158, 43 },
		{ 460, 159, 43 },
		{ 460, 160, 43 },
		{ 460, 161, 43 },
		{ 460, 162, 43 },
		{ 460, 163, 43 },
		{ 460, 164, 43 },
		{ 460, 165, 43 },
		{ 460, 166, 43 },
		{ 460, 167, 43 },
		{ 460, 168, 43 },
		{ 460, 169, 43 },
		{ 460, 170, 43 },
		{ 460, 171, 43 },
		{ 460, 172, 43 },
		{ 460, 173, 43 },
		{ 460, 174, 43 },
		{ 460, 175, 43 },
		{ 460, 176, 43 },
		{ 460, 177, 43 },
		{ 460, 178, 43 },
		{ 460, 179, 43 },
		{ 460, 180, 43 },
		{ 460, 181, 43 },
		{ 460, 182, 43 },
		{ 460, 183, 43 },
		{ 460, 184, 43 },
		{ 460, 185, 43 },
		{ 460, 186, 43 },
		{ 460, 187, 43 },
		{ 460, 188, 43 },
		{ 460, 189, 43 },
		{ 460, 190, 43 },
		{ 460, 191, 43 },
		{ 460, 192, 43 },
		{ 460, 193, 43 },
		{ 460, 194, 43 },
		{ 460, 195, 43 },
		{ 460, 196, 43 },
		{ 460, 197, 43 },
		{ 460, 198, 43 },
		{ 460, 199, 43 },
		{ 460, 200, 43 },
		{ 460, 201, 43 },
		{ 460, 202, 43 },
		{ 460, 203, 43 },
		{ 460, 204, 43 },
		{ 460, 205, 43 },
		{ 460, 206, 43 },
		{ 461, 156, 43 },
		{ 461, 157, 43 },
		{ 461, 158, 43 },
		{ 461, 159, 43 },
		{ 461, 160, 43 },
		{ 461, 161, 43 },
		{ 461, 162, 43 },
		{ 461, 163, 43 },
		{ 461, 164, 43 },
		{ 461, 165, 43 },
		{ 461, 166, 43 },
		{ 461, 167, 43 },
		{ 461, 168, 43 },
		{ 461, 169, 43 },
		{ 461, 170, 43 },
		{ 461, 171, 43 },
		{ 461, 172, 43 },
		{ 461, 173, 43 },
		{ 461, 174, 43 },
		{ 461, 175, 43 },
		{ 461, 176, 43 },
		{ 461, 177, 43 },
		{ 461, 178, 43 },
		{ 461, 179, 43 },
		{ 461, 180, 43 },
		{ 461, 181, 43 },
		{ 461, 182, 43 },
		{ 461, 183, 43 },
		{ 461, 184, 43 },
		{ 461, 185, 43 },
		{ 461, 186, 43 },
		{ 461, 187, 43 },
		{ 461, 188, 43 },
		{ 461, 189, 43 },
		{ 461, 190, 43 },
		{ 461, 191, 43 },
		{ 461, 192, 43 },
		{ 461, 193, 43 },
		{ 461, 194, 43 },
		{ 461, 195, 43 },
		{ 461, 196, 43 },
		{ 461, 197, 43 },
		{ 461, 198, 43 },
		{ 461, 199, 43 },
		{ 461, 200, 43 },
		{ 461, 201, 43 },
		{ 461, 202, 43 },
		{ 461, 203, 43 },
		{ 461, 204, 43 },
		{ 461, 205, 43 },
		{ 462, 155, 43 },
		{ 462, 156, 43 },
		{ 462, 157, 43 },
		{ 462, 158, 43 },
		{ 462, 159, 43 },
		{ 462, 160, 43 },
		{ 462, 161, 43 },
		{ 462, 162, 43 },
		{ 462, 163, 43 },
		{ 462, 164, 43 },
		{ 462, 165, 43 },
		{ 462, 166, 43 },
		{ 462, 167, 43 },
		{ 462, 168, 43 },
		{ 462, 169, 43 },
		{ 462, 170, 43 },
		{ 462, 171, 43 },
		{ 462, 172, 43 },
		{ 462, 173, 43 },
		{ 462, 174, 43 },
		{ 462, 175, 43 },
		{ 462, 176, 43 },
		{ 462, 177, 43 },
		{ 462, 178, 43 },
		{ 462, 179, 43 },
		{ 462, 180, 43 },
		{ 462, 181, 43 },
		{ 462, 182, 43 },
		{ 462, 183, 43 },
		{ 462, 184, 43 },
		{ 462, 185, 43 },
		{ 462, 186, 43 },
		{ 462, 187, 43 },
		{ 462, 188, 43 },
		{ 462, 189, 43 },
		{ 462, 190, 43 },
		{ 462, 191, 43 },
		{ 462, 192, 43 },
		{ 462, 193, 43 },
		{ 462, 194, 43 },
		{ 462, 195, 43 },
		{ 462, 196, 43 },
		{ 462, 197, 43 },
		{ 462, 198, 43 },
		{ 462, 199, 43 },
		{ 462, 200, 43 },
		{ 462, 201, 43 },
		{ 462, 202, 43 },
		{ 462, 203, 43 },
		{ 462, 204, 43 },
		{ 462, 205, 43 },
		{ 463, 155, 43 },
		{ 463, 156, 43 },
		{ 463, 157, 43 },
		{ 463, 158, 43 },
		{ 463, 159, 43 },
		{ 463, 160, 43 },
		{ 463, 161, 43 },
		{ 463, 162, 43 },
		{ 463, 163, 43 },
		{ 463, 164, 43 },
		{ 463, 165, 43 },
		{ 463, 166, 43 },
		{ 463, 167, 43 },
		{ 463, 168, 43 },
		{ 463, 169, 43 },
		{ 463, 170, 43 },
		{ 463, 171, 43 },
		{ 463, 172, 43 },
		{ 463, 173, 43 },
		{ 463, 174, 43 },
		{ 463, 175, 43 },
		{ 463, 176, 43 },
		{ 463, 177, 43 },
		{ 463, 178, 43 },
		{ 463, 179, 43 },
		{ 463, 180, 43 },
		{ 463, 181, 43 },
		{ 463, 182, 43 },
		{ 463, 183, 43 },
		{ 463, 184, 43 },
		{ 463, 185, 43 },
		{ 463, 186, 43 },
		{ 463, 187, 43 },
		{ 463, 188, 43 },
		{ 463, 189, 43 },
		{ 463, 190, 43 },
		{ 463, 191, 43 },
		{ 463, 192, 43 },
		{ 463, 193, 43 },
		{ 463, 194, 43 },
		{ 463, 195, 43 },
		{ 463, 196, 43 },
		{ 463, 197, 43 },
		{ 463, 198, 43 },
		{ 463, 199, 43 },
		{ 463, 200, 43 },
		{ 463, 201, 43 },
		{ 463, 202, 43 },
		{ 463, 203, 43 },
		{ 463, 204, 43 },
		{ 463, 205, 43 },
		{ 464, 154, 43 },
		{ 464, 155, 43 },
		{ 464, 156, 43 },
		{ 464, 157, 43 },
		{ 464, 158, 43 },
		{ 464, 159, 43 },
		{ 464, 160, 43 },
		{ 464, 161, 43 },
		{ 464, 162, 43 },
		{ 464, 163, 43 },
		{ 464, 164, 43 },
		{ 464, 165, 43 },
		{ 464, 166, 43 },
		{ 464, 167, 43 },
		{ 464, 168, 43 },
		{ 464, 169, 43 },
		{ 464, 170, 43 },
		{ 464, 171, 43 },
		{ 464, 172, 43 },
		{ 464, 173, 43 },
		{ 464, 174, 43 },
		{ 464, 175, 43 },
		{ 464, 176, 43 },
		{ 464, 177, 43 },
		{ 464, 178, 43 },
		{ 464, 179, 43 },
		{ 464, 180, 43 },
		{ 464, 181, 43 },
		{ 464, 182, 43 },
		{ 464, 183, 43 },
		{ 464, 184, 43 },
		{ 464, 185, 43 },
		{ 464, 186, 43 },
		{ 464, 187, 43 },
		{ 464, 188, 43 },
		{ 464, 189, 43 },
		{ 464, 190, 43 },
		{ 464, 191, 43 },
		{ 464, 192, 43 },
		{ 464, 193, 43 },
		{ 464, 194, 43 },
		{ 464, 195, 43 },
		{ 464, 196, 43 },
		{ 464, 197, 43 },
		{ 464, 198, 43 },
		{ 464, 199, 43 },
		{ 464, 200, 43 },
		{ 464, 201, 43 },
		{ 464, 202, 43 },
		{ 464, 203, 43 },
		{ 464, 204, 43 },
		{ 464, 205, 43 },
		{ 465, 154, 43 },
		{ 465, 155, 43 },
		{ 465, 156, 43 },
		{ 465, 157, 43 },
		{ 465, 158, 43 },
		{ 465, 159, 43 },
		{ 465, 160, 43 },
		{ 465, 161, 43 },
		{ 465, 162, 43 },
		{ 465, 163, 43 },
		{ 465, 164, 43 },
		{ 465, 165, 43 },
		{ 465, 166, 43 },
		{ 465, 167, 43 },
		{ 465, 168, 43 },
		{ 465, 169, 43 },
		{ 465, 170, 43 },
		{ 465, 171, 43 },
		{ 465, 172, 43 },
		{ 465, 173, 43 },
		{ 465, 174, 43 },
		{ 465, 175, 43 },
		{ 465, 176, 43 },
		{ 465, 177, 43 },
		{ 465, 178, 43 },
		{ 465, 179, 43 },
		{ 465, 180, 43 },
		{ 465, 181, 43 },
		{ 465, 182, 43 },
		{ 465, 183, 43 },
		{ 465, 184, 43 },
		{ 465, 185, 43 },
		{ 465, 186, 43 },
		{ 465, 187, 43 },
		{ 465, 188, 43 },
		{ 465, 189, 43 },
		{ 465, 190, 43 },
		{ 465, 191, 43 },
		{ 465, 192, 43 },
		{ 465, 193, 43 },
		{ 465, 194, 43 },
		{ 465, 195, 43 },
		{ 465, 196, 43 },
		{ 465, 197, 43 },
		{ 465, 198, 43 },
		{ 465, 199, 43 },
		{ 465, 200, 43 },
		{ 465, 201, 43 },
		{ 465, 202, 43 },
		{ 465, 203, 43 },
		{ 465, 204, 43 },
		{ 466, 153, 43 },
		{ 466, 154, 43 },
		{ 466, 155, 43 },
		{ 466, 156, 43 },
		{ 466, 157, 43 },
		{ 466, 158, 43 },
		{ 466, 159, 43 },
		{ 466, 160, 43 },
		{ 466, 161, 43 },
		{ 466, 162, 43 },
		{ 466, 163, 43 },
		{ 466, 164, 43 },
		{ 466, 165, 43 },
		{ 466, 166, 43 },
		{ 466, 167, 43 },
		{ 466, 168, 43 },
		{ 466, 169, 43 },
		{ 466, 170, 43 },
		{ 466, 171, 43 },
		{ 466, 172, 43 },
		{ 466, 173, 43 },
		{ 466, 174, 43 },
		{ 466, 175, 43 },
		{ 466, 176, 43 },
		{ 466, 177, 43 },
		{ 466, 178, 43 },
		{ 466, 179, 43 },
		{ 466, 180, 43 },
		{ 466, 181, 43 },
		{ 466, 182, 43 },
		{ 466, 183, 43 },
		{ 466, 184, 43 },
		{ 466, 185, 43 },
		{ 466, 186, 43 },
		{ 466, 187, 43 },
		{ 466, 188, 43 },
		{ 466, 189, 43 },
		{ 466, 190, 43 },
		{ 466, 191, 43 },
		{ 466, 192, 43 },
		{ 466, 193, 43 },
		{ 466, 194, 43 },
		{ 466, 195, 43 },
		{ 466, 196, 43 },
		{ 466, 197, 43 },
		{ 466, 198, 43 },
		{ 466, 199, 43 },
		{ 466, 200, 43 },
		{ 466, 201, 43 },
		{ 466, 202, 43 },
		{ 466, 203, 43 },
		{ 466, 204, 43 },
		{ 467, 153, 43 },
		{ 467, 154, 43 },
		{ 467, 155, 43 },
		{ 467, 156, 43 },
		{ 467, 157, 43 },
		{ 467, 158, 43 },
		{ 467, 159, 43 },
		{ 467, 160, 43 },
		{ 467, 161, 43 },
		{ 467, 162, 43 },
		{ 467, 163, 43 },
		{ 467, 164, 43 },
		{ 467, 165, 43 },
		{ 467, 166, 43 },
		{ 467, 167, 43 },
		{ 467, 168, 43 },
		{ 467, 169, 43 },
		{ 467, 170, 43 },
		{ 467, 171, 43 },
		{ 467, 172, 43 },
		{ 467, 173, 43 },
		{ 467, 174, 43 },
		{ 467, 175, 43 },
		{ 467, 176, 43 },
		{ 467, 177, 43 },
		{ 467, 178, 43 },
		{ 467, 179, 43 },
		{ 467, 180, 43 },
		{ 467, 181, 43 },
		{ 467, 182, 43 },
		{ 467, 183, 43 },
		{ 467, 184, 43 },
		{ 467, 185, 43 },
		{ 467, 186, 43 },
		{ 467, 187, 43 },
		{ 467, 188, 43 },
		{ 467, 189, 43 },
		{ 467, 190, 43 },
		{ 467, 191, 43 },
		{ 467, 192, 43 },
		{ 467, 193, 43 },
		{ 467, 194, 43 },
		{ 467, 195, 43 },
		{ 467, 196, 43 },
		{ 467, 197, 43 },
		{ 467, 198, 43 },
		{ 467, 199, 43 },
		{ 467, 200, 43 },
		{ 467, 201, 43 },
		{ 467, 202, 43 },
		{ 467, 203, 43 },
		{ 467, 204, 43 },
		{ 468, 152, 43 },
		{ 468, 153, 43 },
		{ 468, 154, 43 },
		{ 468, 155, 43 },
		{ 468, 156, 43 },
		{ 468, 157, 43 },
		{ 468, 158, 43 },
		{ 468, 159, 43 },
		{ 468, 160, 43 },
		{ 468, 161, 43 },
		{ 468, 162, 43 },
		{ 468, 163, 43 },
		{ 468, 164, 43 },
		{ 468, 165, 43 },
		{ 468, 166, 43 },
		{ 468, 167, 43 },
		{ 468, 168, 43 },
		{ 468, 169, 43 },
		{ 468, 170, 43 },
		{ 468, 171, 43 },
		{ 468, 172, 43 },
		{ 468, 173, 43 },
		{ 468, 174, 43 },
		{ 468, 175, 43 },
		{ 468, 176, 43 },
		{ 468, 177, 43 },
		{ 468, 178, 43 },
		{ 468, 179, 43 },
		{ 468, 180, 43 },
		{ 468, 181, 43 },
		{ 468, 182, 43 },
		{ 468, 183, 43 },
		{ 468, 184, 43 },
		{ 468, 185, 43 },
		{ 468, 186, 43 },
		{ 468, 187, 43 },
		{ 468, 188, 43 },
		{ 468, 189, 43 },
		{ 468, 190, 43 },
		{ 468, 191, 43 },
		{ 468, 192, 43 },
		{ 468, 193, 43 },
		{ 468, 194, 43 },
		{ 468, 195, 43 },
		{ 468, 196, 43 },
		{ 468, 197, 43 },
		{ 468, 198, 43 },
		{ 468, 199, 43 },
		{ 468, 200, 43 },
		{ 468, 201, 43 },
		{ 468, 202, 43 },
		{ 468, 203, 43 },
		{ 468, 204, 43 },
		{ 469, 152, 43 },
		{ 469, 153, 43 },
		{ 469, 154, 43 },
		{ 469, 155, 43 },
		{ 469, 156, 43 },
		{ 469, 157, 43 },
		{ 469, 158, 43 },
		{ 469, 159, 43 },
		{ 469, 160, 43 },
		{ 469, 161, 43 },
		{ 469, 162, 43 },
		{ 469, 163, 43 },
		{ 469, 164, 43 },
		{ 469, 165, 43 },
		{ 469, 166, 43 },
		{ 469, 167, 43 },
		{ 469, 168, 43 },
		{ 469, 169, 43 },
		{ 469, 170, 43 },
		{ 469, 171, 43 },
		{ 469, 172, 43 },
		{ 469, 173, 43 },
		{ 469, 174, 43 },
		{ 469, 175, 43 },
		{ 469, 176, 43 },
		{ 469, 177, 43 },
		{ 469, 178, 43 },
		{ 469, 179, 43 },
		{ 469, 180, 43 },
		{ 469, 181, 43 },
		{ 469, 182, 43 },
		{ 469, 183, 43 },
		{ 469, 184, 43 },
		{ 469, 185, 43 },
		{ 469, 186, 43 },
		{ 469, 187, 43 },
		{ 469, 188, 43 },
		{ 469, 189, 43 },
		{ 469, 190, 43 },
		{ 469, 191, 43 },
		{ 469, 192, 43 },
		{ 469, 193, 43 },
		{ 469, 194, 43 },
		{ 469, 195, 43 },
		{ 469, 196, 43 },
		{ 469, 197, 43 },
		{ 469, 198, 43 },
		{ 469, 199, 43 },
		{ 469, 200, 43 },
		{ 469, 201, 43 },
		{ 469, 202, 43 },
		{ 469, 203, 43 },
		{ 470, 151, 43 },
		{ 470, 152, 43 },
		{ 470, 153, 43 },
		{ 470, 154, 43 },
		{ 470, 155, 43 },
		{ 470, 156, 43 },
		{ 470, 157, 43 },
		{ 470, 158, 43 },
		{ 470, 159, 43 },
		{ 470, 160, 43 },
		{ 470, 161, 43 },
		{ 470, 162, 43 },
		{ 470, 163, 43 },
		{ 470, 164, 43 },
		{ 470, 165, 43 },
		{ 470, 166, 43 },
		{ 470, 167, 43 },
		{ 470, 168, 43 },
		{ 470, 169, 43 },
		{ 470, 170, 43 },
		{ 470, 171, 43 },
		{ 470, 172, 43 },
		{ 470, 173, 43 },
		{ 470, 174, 43 },
		{ 470, 175, 43 },
		{ 470, 176, 43 },
		{ 470, 177, 43 },
		{ 470, 178, 43 },
		{ 470, 179, 43 },
		{ 470, 180, 43 },
		{ 470, 181, 43 },
		{ 470, 182, 43 },
		{ 470, 183, 43 },
		{ 470, 184, 43 },
		{ 470, 185, 43 },
		{ 470, 186, 43 },
		{ 470, 187, 43 },
		{ 470, 188, 43 },
		{ 470, 189, 43 },
		{ 470, 190, 43 },
		{ 470, 191, 43 },
		{ 470, 192, 43 },
		{ 470, 193, 43 },
		{ 470, 194, 43 },
		{ 470, 195, 43 },
		{ 470, 196, 43 },
		{ 470, 197, 43 },
		{ 470, 198, 43 },
		{ 470, 199, 43 },
		{ 470, 200, 43 },
		{ 470, 201, 43 },
		{ 470, 202, 43 },
		{ 470, 203, 43 },
		{ 471, 151, 43 },
		{ 471, 152, 43 },
		{ 471, 153, 43 },
		{ 471, 154, 43 },
		{ 471, 155, 43 },
		{ 471, 156, 43 },
		{ 471, 157, 43 },
		{ 471, 158, 43 },
		{ 471, 159, 43 },
		{ 471, 160, 43 },
		{ 471, 161, 43 },
		{ 471, 162, 43 },
		{ 471, 163, 43 },
		{ 471, 164, 43 },
		{ 471, 165, 43 },
		{ 471, 166, 43 },
		{ 471, 167, 43 },
		{ 471, 168, 43 },
		{ 471, 169, 43 },
		{ 471, 170, 43 },
		{ 471, 171, 43 },
		{ 471, 172, 43 },
		{ 471, 173, 43 },
		{ 471, 174, 43 },
		{ 471, 175, 43 },
		{ 471, 176, 43 },
		{ 471, 177, 43 },
		{ 471, 178, 43 },
		{ 471, 179, 43 },
		{ 471, 180, 43 },
		{ 471, 181, 43 },
		{ 471, 182, 43 },
		{ 471, 183, 43 },
		{ 471, 184, 43 },
		{ 471, 185, 43 },
		{ 471, 186, 43 },
		{ 471, 187, 43 },
		{ 471, 188, 43 },
		{ 471, 189, 43 },
		{ 471, 190, 43 },
		{ 471, 191, 43 },
		{ 471, 192, 43 },
		{ 471, 193, 43 },
		{ 471, 194, 43 },
		{ 471, 195, 43 },
		{ 471, 196, 43 },
		{ 471, 197, 43 },
		{ 471, 198, 43 },
		{ 471, 199, 43 },
		{ 471, 200, 43 },
		{ 471, 201, 43 },
		{ 471, 202, 43 },
		{ 471, 203, 43 },
		{ 472, 150, 43 },
		{ 472, 151, 43 },
		{ 472, 152, 43 },
		{ 472, 153, 43 },
		{ 472, 154, 43 },
		{ 472, 155, 43 },
		{ 472, 156, 43 },
		{ 472, 157, 43 },
		{ 472, 158, 43 },
		{ 472, 159, 43 },
		{ 472, 160, 43 },
		{ 472, 161, 43 },
		{ 472, 162, 43 },
		{ 472, 163, 43 },
		{ 472, 164, 43 },
		{ 472, 165, 43 },
		{ 472, 166, 43 },
		{ 472, 167, 43 },
		{ 472, 168, 43 },
		{ 472, 169, 43 },
		{ 472, 170, 43 },
		{ 472, 171, 43 },
		{ 472, 172, 43 },
		{ 472, 173, 43 },
		{ 472, 174, 43 },
		{ 472, 175, 43 },
		{ 472, 176, 43 },
		{ 472, 177, 43 },
		{ 472, 178, 43 },
		{ 472, 179, 43 },
		{ 472, 180, 43 },
		{ 472, 181, 43 },
		{ 472, 182, 43 },
		{ 472, 183, 43 },
		{ 472, 184, 43 },
		{ 472, 185, 43 },
		{ 472, 186, 43 },
		{ 472, 187, 43 },
		{ 472, 188, 43 },
		{ 472, 189, 43 },
		{ 472, 190, 43 },
		{ 472, 191, 43 },
		{ 472, 192, 43 },
		{ 472, 193, 43 },
		{ 472, 194, 43 },
		{ 472, 195, 43 },
		{ 472, 196, 43 },
		{ 472, 197, 43 },
		{ 472, 198, 43 },
		{ 472, 199, 43 },
		{ 472, 200, 43 },
		{ 472, 201, 43 },
		{ 472, 202, 43 },
		{ 472, 203, 43 },
		{ 473, 150, 43 },
		{ 473, 151, 43 },
		{ 473, 152, 43 },
		{ 473, 153, 43 },
		{ 473, 154, 43 },
		{ 473, 155, 43 },
		{ 473, 156, 43 },
		{ 473, 157, 43 },
		{ 473, 158, 43 },
		{ 473, 159, 43 },
		{ 473, 160, 43 },
		{ 473, 161, 43 },
		{ 473, 162, 43 },
		{ 473, 163, 43 },
		{ 473, 164, 43 },
		{ 473, 165, 43 },
		{ 473, 166, 43 },
		{ 473, 167, 43 },
		{ 473, 168, 43 },
		{ 473, 169, 43 },
		{ 473, 170, 43 },
		{ 473, 171, 43 },
		{ 473, 172, 43 },
		{ 473, 173, 43 },
		{ 473, 174, 43 },
		{ 473, 175, 43 },
		{ 473, 176, 43 },
		{ 473, 177, 43 },
		{ 473, 178, 43 },
		{ 473, 179, 43 },
		{ 473, 180, 43 },
		{ 473, 181, 43 },
		{ 473, 182, 43 },
		{ 473, 183, 43 },
		{ 473, 184, 43 },
		{ 473, 185, 43 },
		{ 473, 186, 43 },
		{ 473, 187, 43 },
		{ 473, 188, 43 },
		{ 473, 189, 43 },
		{ 473, 190, 43 },
		{ 473, 191, 43 },
		{ 473, 192, 43 },
		{ 473, 193, 43 },
		{ 473, 194, 43 },
		{ 473, 195, 43 },
		{ 473, 196, 43 },
		{ 473, 197, 43 },
		{ 473, 198, 43 },
		{ 473, 199, 43 },
		{ 473, 200, 43 },
		{ 473, 201, 43 },
		{ 473, 202, 43 },
		{ 474, 149, 43 },
		{ 474, 150, 43 },
		{ 474, 151, 43 },
		{ 474, 152, 43 },
		{ 474, 153, 43 },
		{ 474, 154, 43 },
		{ 474, 155, 43 },
		{ 474, 156, 43 },
		{ 474, 157, 43 },
		{ 474, 158, 43 },
		{ 474, 159, 43 },
		{ 474, 160, 43 },
		{ 474, 161, 43 },
		{ 474, 162, 43 },
		{ 474, 163, 43 },
		{ 474, 164, 43 },
		{ 474, 165, 43 },
		{ 474, 166, 43 },
		{ 474, 167, 43 },
		{ 474, 168, 43 },
		{ 474, 169, 43 },
		{ 474, 170, 43 },
		{ 474, 171, 43 },
		{ 474, 172, 43 },
		{ 474, 173, 43 },
		{ 474, 174, 43 },
		{ 474, 175, 43 },
		{ 474, 176, 43 },
		{ 474, 177, 43 },
		{ 474, 178, 43 },
		{ 474, 179, 43 },
		{ 474, 180, 43 },
		{ 474, 181, 43 },
		{ 474, 182, 43 },
		{ 474, 183, 43 },
		{ 474, 184, 43 },
		{ 474, 185, 43 },
		{ 474, 186, 43 },
		{ 474, 187, 43 },
		{ 474, 188, 43 },
		{ 474, 189, 43 },
		{ 474, 190, 43 },
		{ 474, 191, 43 },
		{ 474, 192, 43 },
		{ 474, 193, 43 },
		{ 474, 194, 43 },
		{ 474, 195, 43 },
		{ 474, 196, 43 },
		{ 474, 197, 43 },
		{ 474, 198, 43 },
		{ 474, 199, 43 },
		{ 474, 200, 43 },
		{ 474, 201, 43 },
		{ 474, 202, 43 },
		{ 475, 149, 43 },
		{ 475, 150, 43 },
		{ 475, 151, 43 },
		{ 475, 152, 43 },
		{ 475, 153, 43 },
		{ 475, 154, 43 },
		{ 475, 155, 43 },
		{ 475, 156, 43 },
		{ 475, 157, 43 },
		{ 475, 158, 43 },
		{ 475, 159, 43 },
		{ 475, 160, 43 },
		{ 475, 161, 43 },
		{ 475, 162, 43 },
		{ 475, 163, 43 },
		{ 475, 164, 43 },
		{ 475, 165, 43 },
		{ 475, 166, 43 },
		{ 475, 167, 43 },
		{ 475, 168, 43 },
		{ 475, 169, 43 },
		{ 475, 170, 43 },
		{ 475, 171, 43 },
		{ 475, 172, 43 },
		{ 475, 173, 43 },
		{ 475, 174, 43 },
		{ 475, 175, 43 },
		{ 475, 176, 43 },
		{ 475, 177, 43 },
		{ 475, 178, 43 },
		{ 475, 179, 43 },
		{ 475, 180, 43 },
		{ 475, 181, 43 },
		{ 475, 182, 43 },
		{ 475, 183, 43 },
		{ 475, 184, 43 },
		{ 475, 185, 43 },
		{ 475, 186, 43 },
		{ 475, 187, 43 },
		{ 475, 188, 43 },
		{ 475, 189, 43 },
		{ 475, 190, 43 },
		{ 475, 191, 43 },
		{ 475, 192, 43 },
		{ 475, 193, 43 },
		{ 475, 194, 43 },
		{ 475, 195, 43 },
		{ 475, 196, 43 },
		{ 475, 197, 43 },
		{ 475, 198, 43 },
		{ 475, 199, 43 },
		{ 475, 200, 43 },
		{ 475, 201, 43 },
		{ 475, 202, 43 },
		{ 476, 148, 43 },
		{ 476, 149, 43 },
		{ 476, 150, 43 },
		{ 476, 151, 43 },
		{ 476, 152, 43 },
		{ 476, 153, 43 },
		{ 476, 154, 43 },
		{ 476, 155, 43 },
		{ 476, 156, 43 },
		{ 476, 157, 43 },
		{ 476, 158, 43 },
		{ 476, 159, 43 },
		{ 476, 160, 43 },
		{ 476, 161, 43 },
		{ 476, 162, 43 },
		{ 476, 163, 43 },
		{ 476, 164, 43 },
		{ 476, 165, 43 },
		{ 476, 166, 43 },
		{ 476, 167, 43 },
		{ 476, 168, 43 },
		{ 476, 169, 43 },
		{ 476, 170, 43 },
		{ 476, 171, 43 },
		{ 476, 172, 43 },
		{ 476, 173, 43 },
		{ 476, 174, 43 },
		{ 476, 175, 43 },
		{ 476, 176, 43 },
		{ 476, 177, 43 },
		{ 476, 178, 43 },
		{ 476, 179, 43 },
		{ 476, 180, 43 },
		{ 476, 181, 43 },
		{ 476, 182, 43 },
		{ 476, 183, 43 },
		{ 476, 184, 43 },
		{ 476, 185, 43 },
		{ 476, 186, 43 },
		{ 476, 187, 43 },
		{ 476, 188, 43 },
		{ 476, 189, 43 },
		{ 476, 190, 43 },
		{ 476, 191, 43 },
		{ 476, 192, 43 },
		{ 476, 193, 43 },
		{ 476, 194, 43 },
		{ 476, 195, 43 },
		{ 476, 196, 43 },
		{ 476, 197, 43 },
		{ 476, 198, 43 },
		{ 476, 199, 43 },
		{ 476, 200, 43 },
		{ 476, 201, 43 },
		{ 476, 202, 43 },
		{ 477, 148, 43 },
		{ 477, 149, 43 },
		{ 477, 150, 43 },
		{ 477, 151, 43 },
		{ 477, 152, 43 },
		{ 477, 153, 43 },
		{ 477, 154, 43 },
		{ 477, 155, 43 },
		{ 477, 156, 43 },
		{ 477, 157, 43 },
		{ 477, 158, 43 },
		{ 477, 159, 43 },
		{ 477, 160, 43 },
		{ 477, 161, 43 },
		{ 477, 162, 43 },
		{ 477, 163, 43 },
		{ 477, 164, 43 },
		{ 477, 165, 43 },
		{ 477, 166, 43 },
		{ 477, 167, 43 },
		{ 477, 168, 43 },
		{ 477, 169, 43 },
		{ 477, 170, 43 },
		{ 477, 171, 43 },
		{ 477, 172, 43 },
		{ 477, 173, 43 },
		{ 477, 174, 43 },
		{ 477, 175, 43 },
		{ 477, 176, 43 },
		{ 477, 177, 43 },
		{ 477, 178, 43 },
		{ 477, 179, 43 },
		{ 477, 180, 43 },
		{ 477, 181, 43 },
		{ 477, 182, 43 },
		{ 477, 183, 43 },
		{ 477, 184, 43 },
		{ 477, 185, 43 },
		{ 477, 186, 43 },
		{ 477, 187, 43 },
		{ 477, 188, 43 },
		{ 477, 189, 43 },
		{ 477, 190, 43 },
		{ 477, 191, 43 },
		{ 477, 192, 43 },
		{ 477, 193, 43 },
		{ 477, 194, 43 },
		{ 477, 195, 43 },
		{ 477, 196, 43 },
		{ 477, 197, 43 },
		{ 477, 198, 43 },
		{ 477, 199, 43 },
		{ 477, 200, 43 },
		{ 477, 201, 43 },
		{ 478, 147, 43 },
		{ 478, 148, 43 },
		{ 478, 149, 43 },
		{ 478, 150, 43 },
		{ 478, 151, 43 },
		{ 478, 152, 43 },
		{ 478, 153, 43 },
		{ 478, 154, 43 },
		{ 478, 155, 43 },
		{ 478, 156, 43 },
		{ 478, 157, 43 },
		{ 478, 158, 43 },
		{ 478, 159, 43 },
		{ 478, 160, 43 },
		{ 478, 161, 43 },
		{ 478, 162, 43 },
		{ 478, 163, 43 },
		{ 478, 164, 43 },
		{ 478, 165, 43 },
		{ 478, 166, 43 },
		{ 478, 167, 43 },
		{ 478, 168, 43 },
		{ 478, 169, 43 },
		{ 478, 170, 43 },
		{ 478, 171, 43 },
		{ 478, 172, 43 },
		{ 478, 173, 43 },
		{ 478, 174, 43 },
		{ 478, 175, 43 },
		{ 478, 176, 43 },
		{ 478, 177, 43 },
		{ 478, 178, 43 },
		{ 478, 179, 43 },
		{ 478, 180, 43 },
		{ 478, 181, 43 },
		{ 478, 182, 43 },
		{ 478, 183, 43 },
		{ 478, 184, 43 },
		{ 478, 185, 43 },
		{ 478, 186, 43 },
		{ 478, 187, 43 },
		{ 478, 188, 43 },
		{ 478, 189, 43 },
		{ 478, 190, 43 },
		{ 478, 191, 43 },
		{ 478, 192, 43 },
		{ 478, 193, 43 },
		{ 478, 194, 43 },
		{ 478, 195, 43 },
		{ 478, 196, 43 },
		{ 478, 197, 43 },
		{ 478, 198, 43 },
		{ 478, 199, 43 },
		{ 478, 200, 43 },
		{ 478, 201, 43 },
		{ 479, 147, 43 },
		{ 479, 148, 43 },
		{ 479, 149, 43 },
		{ 479, 150, 43 },
		{ 479, 151, 43 },
		{ 479, 152, 43 },
		{ 479, 153, 43 },
		{ 479, 154, 43 },
		{ 479, 155, 43 },
		{ 479, 156, 43 },
		{ 479, 157, 43 },
		{ 479, 158, 43 },
		{ 479, 159, 43 },
		{ 479, 160, 43 },
		{ 479, 161, 43 },
		{ 479, 162, 43 },
		{ 479, 163, 43 },
		{ 479, 164, 43 },
		{ 479, 165, 43 },
		{ 479, 166, 43 },
		{ 479, 167, 43 },
		{ 479, 168, 43 },
		{ 479, 169, 43 },
		{ 479, 170, 43 },
		{ 479, 171, 43 },
		{ 479, 172, 43 },
		{ 479, 173, 43 },
		{ 479, 174, 43 },
		{ 479, 175, 43 },
		{ 479, 176, 43 },
		{ 479, 177, 43 },
		{ 479, 178, 43 },
		{ 479, 179, 43 },
		{ 479, 180, 43 },
		{ 479, 181, 43 },
		{ 479, 182, 43 },
		{ 479, 183, 43 },
		{ 479, 184, 43 },
		{ 479, 185, 43 },
		{ 479, 186, 43 },
		{ 479, 187, 43 },
		{ 479, 188, 43 },
		{ 479, 189, 43 },
		{ 479, 190, 43 },
		{ 479, 191, 43 },
		{ 479, 192, 43 },
		{ 479, 193, 43 },
		{ 479, 194, 43 },
		{ 479, 195, 43 },
		{ 479, 196, 43 },
		{ 479, 197, 43 },
		{ 479, 198, 43 },
		{ 479, 199, 43 },
		{ 479, 200, 43 },
		{ 479, 201, 43 },
		{ 480, 146, 43 },
		{ 480, 147, 43 },
		{ 480, 148, 43 },
		{ 480, 149, 43 },
		{ 480, 150, 43 },
		{ 480, 151, 43 },
		{ 480, 152, 43 },
		{ 480, 153, 43 },
		{ 480, 154, 43 },
		{ 480, 155, 43 },
		{ 480, 156, 43 },
		{ 480, 157, 43 },
		{ 480, 158, 43 },
		{ 480, 159, 43 },
		{ 480, 160, 43 },
		{ 480, 161, 43 },
		{ 480, 162, 43 },
		{ 480, 163, 43 },
		{ 480, 164, 43 },
		{ 480, 165, 43 },
		{ 480, 166, 43 },
		{ 480, 167, 43 },
		{ 480, 168, 43 },
		{ 480, 169, 43 },
		{ 480, 170, 43 },
		{ 480, 171, 43 },
		{ 480, 172, 43 },
		{ 480, 173, 43 },
		{ 480, 174, 43 },
		{ 480, 175, 43 },
		{ 480, 176, 43 },
		{ 480, 177, 43 },
		{ 480, 178, 43 },
		{ 480, 179, 43 },
		{ 480, 180, 43 },
		{ 480, 181, 43 },
		{ 480, 182, 43 },
		{ 480, 183, 43 },
		{ 480, 184, 43 },
		{ 480, 185, 43 },
		{ 480, 186, 43 },
		{ 480, 187, 43 },
		{ 480, 188, 43 },
		{ 480, 189, 43 },
		{ 480, 190, 43 },
		{ 480, 191, 43 },
		{ 480, 192, 43 },
		{ 480, 193, 43 },
		{ 480, 194, 43 },
		{ 480, 195, 43 },
		{ 480, 196, 43 },
		{ 480, 197, 43 },
		{ 480, 198, 43 },
		{ 480, 199, 43 },
		{ 480, 200, 43 },
		{ 480, 201, 43 },
		{ 481, 146, 43 },
		{ 481, 147, 43 },
		{ 481, 148, 43 },
		{ 481, 149, 43 },
		{ 481, 150, 43 },
		{ 481, 151, 43 },
		{ 481, 152, 43 },
		{ 481, 153, 43 },
		{ 481, 154, 43 },
		{ 481, 155, 43 },
		{ 481, 156, 43 },
		{ 481, 157, 43 },
		{ 481, 158, 43 },
		{ 481, 159, 43 },
		{ 481, 160, 43 },
		{ 481, 161, 43 },
		{ 481, 162, 43 },
		{ 481, 163, 43 },
		{ 481, 164, 43 },
		{ 481, 165, 43 },
		{ 481, 166, 43 },
		{ 481, 167, 43 },
		{ 481, 168, 43 },
		{ 481, 169, 43 },
		{ 481, 170, 43 },
		{ 481, 171, 43 },
		{ 481, 172, 43 },
		{ 481, 173, 43 },
		{ 481, 174, 43 },
		{ 481, 175, 43 },
		{ 481, 176, 43 },
		{ 481, 177, 43 },
		{ 481, 178, 43 },
		{ 481, 179, 43 },
		{ 481, 180, 43 },
		{ 481, 181, 43 },
		{ 481, 182, 43 },
		{ 481, 183, 43 },
		{ 481, 184, 43 },
		{ 481, 185, 43 },
		{ 481, 186, 43 },
		{ 481, 187, 43 },
		{ 481, 188, 43 },
		{ 481, 189, 43 },
		{ 481, 190, 43 },
		{ 481, 191, 43 },
		{ 481, 192, 43 },
		{ 481, 193, 43 },
		{ 481, 194, 43 },
		{ 481, 195, 43 },
		{ 481, 196, 43 },
		{ 481, 197, 43 },
		{ 481, 198, 43 },
		{ 481, 199, 43 },
		{ 481, 200, 43 },
		{ 482, 145, 43 },
		{ 482, 146, 43 },
		{ 482, 147, 43 },
		{ 482, 148, 43 },
		{ 482, 149, 43 },
		{ 482, 150, 43 },
		{ 482, 151, 43 },
		{ 482, 152, 43 },
		{ 482, 153, 43 },
		{ 482, 154, 43 },
		{ 482, 155, 43 },
		{ 482, 156, 43 },
		{ 482, 157, 43 },
		{ 482, 158, 43 },
		{ 482, 159, 43 },
		{ 482, 160, 43 },
		{ 482, 161, 43 },
		{ 482, 162, 43 },
		{ 482, 163, 43 },
		{ 482, 164, 43 },
		{ 482, 165, 43 },
		{ 482, 166, 43 },
		{ 482, 167, 43 },
		{ 482, 168, 43 },
		{ 482, 169, 43 },
		{ 482, 170, 43 },
		{ 482, 171, 43 },
		{ 482, 172, 43 },
		{ 482, 173, 43 },
		{ 482, 174, 43 },
		{ 482, 175, 43 },
		{ 482, 176, 43 },
		{ 482, 177, 43 },
		{ 482, 178, 43 },
		{ 482, 179, 43 },
		{ 482, 180, 43 },
		{ 482, 181, 43 },
		{ 482, 182, 43 },
		{ 482, 183, 43 },
		{ 482, 184, 43 },
		{ 482, 185, 43 },
		{ 482, 186, 43 },
		{ 482, 187, 43 },
		{ 482, 188, 43 },
		{ 482, 189, 43 },
		{ 482, 190, 43 },
		{ 482, 191, 43 },
		{ 482, 192, 43 },
		{ 482, 193, 43 },
		{ 482, 194, 43 },
		{ 482, 195, 43 },
		{ 482, 196, 43 },
		{ 482, 197, 43 },
		{ 482, 198, 43 },
		{ 482, 199, 43 },
		{ 482, 200, 43 },
		{ 483, 145, 43 },
		{ 483, 146, 43 },
		{ 483, 147, 43 },
		{ 483, 148, 43 },
		{ 483, 149, 43 },
		{ 483, 150, 43 },
		{ 483, 151, 43 },
		{ 483, 152, 43 },
		{ 483, 153, 43 },
		{ 483, 154, 43 },
		{ 483, 155, 43 },
		{ 483, 156, 43 },
		{ 483, 157, 43 },
		{ 483, 158, 43 },
		{ 483, 159, 43 },
		{ 483, 160, 43 },
		{ 483, 161, 43 },
		{ 483, 162, 43 },
		{ 483, 163, 43 },
		{ 483, 164, 43 },
		{ 483, 165, 43 },
		{ 483, 166, 43 },
		{ 483, 167, 43 },
		{ 483, 168, 43 },
		{ 483, 169, 43 },
		{ 483, 170, 43 },
		{ 483, 171, 43 },
		{ 483, 172, 43 },
		{ 483, 173, 43 },
		{ 483, 174, 43 },
		{ 483, 175, 43 },
		{ 483, 176, 43 },
		{ 483, 177, 43 },
		{ 483, 178, 43 },
		{ 483, 179, 43 },
		{ 483, 180, 43 },
		{ 483, 181, 43 },
		{ 483, 182, 43 },
		{ 483, 183, 43 },
		{ 483, 184, 43 },
		{ 483, 185, 43 },
		{ 483, 186, 43 },
		{ 483, 187, 43 },
		{ 483, 188, 43 },
		{ 483, 189, 43 },
		{ 483, 190, 43 },
		{ 483, 191, 43 },
		{ 483, 192, 43 },
		{ 483, 193, 43 },
		{ 483, 194, 43 },
		{ 483, 195, 43 },
		{ 483, 196, 43 },
		{ 483, 197, 43 },
		{ 483, 198, 43 },
		{ 483, 199, 43 },
		{ 483, 200, 43 },
		{ 484, 144, 43 },
		{ 484, 145, 43 },
		{ 484, 146, 43 },
		{ 484, 147, 43 },
		{ 484, 148, 43 },
		{ 484, 149, 43 },
		{ 484, 150, 43 },
		{ 484, 151, 43 },
		{ 484, 152, 43 },
		{ 484, 153, 43 },
		{ 484, 154, 43 },
		{ 484, 155, 43 },
		{ 484, 156, 43 },
		{ 484, 157, 43 },
		{ 484, 158, 43 },
		{ 484, 159, 43 },
		{ 484, 160, 43 },
		{ 484, 161, 43 },
		{ 484, 162, 43 },
		{ 484, 163, 43 },
		{ 484, 164, 43 },
		{ 484, 165, 43 },
		{ 484, 166, 43 },
		{ 484, 167, 43 },
		{ 484, 168, 43 },
		{ 484, 169, 43 },
		{ 484, 170, 43 },
		{ 484, 171, 43 },
		{ 484, 172, 43 },
		{ 484, 173, 43 },
		{ 484, 174, 43 },
		{ 484, 175, 43 },
		{ 484, 176, 43 },
		{ 484, 177, 43 },
		{ 484, 178, 43 },
		{ 484, 179, 43 },
		{ 484, 180, 43 },
		{ 484, 181, 43 },
		{ 484, 182, 43 },
		{ 484, 183, 43 },
		{ 484, 184, 43 },
		{ 484, 185, 43 },
		{ 484, 186, 43 },
		{ 484, 187, 43 },
		{ 484, 188, 43 },
		{ 484, 189, 43 },
		{ 484, 190, 43 },
		{ 484, 191, 43 },
		{ 484, 192, 43 },
		{ 484, 193, 43 },
		{ 484, 194, 43 },
		{ 484, 195, 43 },
		{ 484, 196, 43 },
		{ 484, 197, 43 },
		{ 484, 198, 43 },
		{ 484, 199, 43 },
		{ 484, 200, 43 },
		{ 485, 144, 43 },
		{ 485, 145, 43 },
		{ 485, 146, 43 },
		{ 485, 147, 43 },
		{ 485, 148, 43 },
		{ 485, 149, 43 },
		{ 485, 150, 43 },
		{ 485, 151, 43 },
		{ 485, 152, 43 },
		{ 485, 153, 43 },
		{ 485, 154, 43 },
		{ 485, 155, 43 },
		{ 485, 156, 43 },
		{ 485, 157, 43 },
		{ 485, 158, 43 },
		{ 485, 159, 43 },
		{ 485, 160, 43 },
		{ 485, 161, 43 },
		{ 485, 162, 43 },
		{ 485, 163, 43 },
		{ 485, 164, 43 },
		{ 485, 165, 43 },
		{ 485, 166, 43 },
		{ 485, 167, 43 },
		{ 485, 168, 43 },
		{ 485, 169, 43 },
		{ 485, 170, 43 },
		{ 485, 171, 43 },
		{ 485, 172, 43 },
		{ 485, 173, 43 },
		{ 485, 174, 43 },
		{ 485, 175, 43 },
		{ 485, 176, 43 },
		{ 485, 177, 43 },
		{ 485, 178, 43 },
		{ 485, 179, 43 },
		{ 485, 180, 43 },
		{ 485, 181, 43 },
		{ 485, 182, 43 },
		{ 485, 183, 43 },
		{ 485, 184, 43 },
		{ 485, 185, 43 },
		{ 485, 186, 43 },
		{ 485, 187, 43 },
		{ 485, 188, 43 },
		{ 485, 189, 43 },
		{ 485, 190, 43 },
		{ 485, 191, 43 },
		{ 485, 192, 43 },
		{ 485, 193, 43 },
		{ 485, 194, 43 },
		{ 485, 195, 43 },
		{ 485, 196, 43 },
		{ 485, 197, 43 },
		{ 485, 198, 43 },
		{ 485, 199, 43 },
		{ 486, 143, 43 },
		{ 486, 144, 43 },
		{ 486, 145, 43 },
		{ 486, 146, 43 },
		{ 486, 147, 43 },
		{ 486, 148, 43 },
		{ 486, 149, 43 },
		{ 486, 150, 43 },
		{ 486, 151, 43 },
		{ 486, 152, 43 },
		{ 486, 153, 43 },
		{ 486, 154, 43 },
		{ 486, 155, 43 },
		{ 486, 156, 43 },
		{ 486, 157, 43 },
		{ 486, 158, 43 },
		{ 486, 159, 43 },
		{ 486, 160, 43 },
		{ 486, 161, 43 },
		{ 486, 162, 43 },
		{ 486, 163, 43 },
		{ 486, 164, 43 },
		{ 486, 165, 43 },
		{ 486, 166, 43 },
		{ 486, 167, 43 },
		{ 486, 168, 43 },
		{ 486, 169, 43 },
		{ 486, 170, 43 },
		{ 486, 171, 43 },
		{ 486, 172, 43 },
		{ 486, 173, 43 },
		{ 486, 174, 43 },
		{ 486, 175, 43 },
		{ 486, 176, 43 },
		{ 486, 177, 43 },
		{ 486, 178, 43 },
		{ 486, 179, 43 },
		{ 486, 180, 43 },
		{ 486, 181, 43 },
		{ 486, 182, 43 },
		{ 486, 183, 43 },
		{ 486, 184, 43 },
		{ 486, 185, 43 },
		{ 486, 186, 43 },
		{ 486, 187, 43 },
		{ 486, 188, 43 },
		{ 486, 189, 43 },
		{ 486, 190, 43 },
		{ 486, 191, 43 },
		{ 486, 192, 43 },
		{ 486, 193, 43 },
		{ 486, 194, 43 },
		{ 486, 195, 43 },
		{ 486, 196, 43 },
		{ 486, 197, 43 },
		{ 486, 198, 43 },
		{ 486, 199, 43 },
		{ 487, 143, 43 },
		{ 487, 144, 43 },
		{ 487, 145, 43 },
		{ 487, 146, 43 },
		{ 487, 147, 43 },
		{ 487, 148, 43 },
		{ 487, 149, 43 },
		{ 487, 150, 43 },
		{ 487, 151, 43 },
		{ 487, 152, 43 },
		{ 487, 153, 43 },
		{ 487, 154, 43 },
		{ 487, 155, 43 },
		{ 487, 156, 43 },
		{ 487, 157, 43 },
		{ 487, 158, 43 },
		{ 487, 159, 43 },
		{ 487, 160, 43 },
		{ 487, 161, 43 },
		{ 487, 162, 43 },
		{ 487, 163, 43 },
		{ 487, 164, 43 },
		{ 487, 165, 43 },
		{ 487, 166, 43 },
		{ 487, 167, 43 },
		{ 487, 168, 43 },
		{ 487, 169, 43 },
		{ 487, 170, 43 },
		{ 487, 171, 43 },
		{ 487, 172, 43 },
		{ 487, 173, 43 },
		{ 487, 174, 43 },
		{ 487, 175, 43 },
		{ 487, 176, 43 },
		{ 487, 177, 43 },
		{ 487, 178, 43 },
		{ 487, 179, 43 },
		{ 487, 180, 43 },
		{ 487, 181, 43 },
		{ 487, 182, 43 },
		{ 487, 183, 43 },
		{ 487, 184, 43 },
		{ 487, 185, 43 },
		{ 487, 186, 43 },
		{ 487, 187, 43 },
		{ 487, 188, 43 },
		{ 487, 189, 43 },
		{ 487, 190, 43 },
		{ 487, 191, 43 },
		{ 487, 192, 43 },
		{ 487, 193, 43 },
		{ 487, 194, 43 },
		{ 487, 195, 43 },
		{ 487, 196, 43 },
		{ 487, 197, 43 },
		{ 487, 198, 43 },
		{ 487, 199, 43 },
		{ 488, 142, 43 },
		{ 488, 143, 43 },
		{ 488, 144, 43 },
		{ 488, 145, 43 },
		{ 488, 146, 43 },
		{ 488, 147, 43 },
		{ 488, 148, 43 },
		{ 488, 149, 43 },
		{ 488, 150, 43 },
		{ 488, 151, 43 },
		{ 488, 152, 43 },
		{ 488, 153, 43 },
		{ 488, 154, 43 },
		{ 488, 155, 43 },
		{ 488, 156, 43 },
		{ 488, 157, 43 },
		{ 488, 158, 43 },
		{ 488, 159, 43 },
		{ 488, 160, 43 },
		{ 488, 161, 43 },
		{ 488, 162, 43 },
		{ 488, 163, 43 },
		{ 488, 164, 43 },
		{ 488, 165, 43 },
		{ 488, 166, 43 },
		{ 488, 167, 43 },
		{ 488, 168, 43 },
		{ 488, 169, 43 },
		{ 488, 170, 43 },
		{ 488, 171, 43 },
		{ 488, 172, 43 },
		{ 488, 173, 43 },
		{ 488, 174, 43 },
		{ 488, 175, 43 },
		{ 488, 176, 43 },
		{ 488, 177, 43 },
		{ 488, 178, 43 },
		{ 488, 179, 43 },
		{ 488, 180, 43 },
		{ 488, 181, 43 },
		{ 488, 182, 43 },
		{ 488, 183, 43 },
		{ 488, 184, 43 },
		{ 488, 185, 43 },
		{ 488, 186, 43 },
		{ 488, 187, 43 },
		{ 488, 188, 43 },
		{ 488, 189, 43 },
		{ 488, 190, 43 },
		{ 488, 191, 43 },
		{ 488, 192, 43 },
		{ 488, 193, 43 },
		{ 488, 194, 43 },
		{ 488, 195, 43 },
		{ 488, 196, 43 },
		{ 488, 197, 43 },
		{ 488, 198, 43 },
		{ 488, 199, 43 },
		{ 489, 142, 43 },
		{ 489, 143, 43 },
		{ 489, 144, 43 },
		{ 489, 145, 43 },
		{ 489, 146, 43 },
		{ 489, 147, 43 },
		{ 489, 148, 43 },
		{ 489, 149, 43 },
		{ 489, 150, 43 },
		{ 489, 151, 43 },
		{ 489, 152, 43 },
		{ 489, 153, 43 },
		{ 489, 154, 43 },
		{ 489, 155, 43 },
		{ 489, 156, 43 },
		{ 489, 157, 43 },
		{ 489, 158, 43 },
		{ 489, 159, 43 },
		{ 489, 160, 43 },
		{ 489, 161, 43 },
		{ 489, 162, 43 },
		{ 489, 163, 43 },
		{ 489, 164, 43 },
		{ 489, 165, 43 },
		{ 489, 166, 43 },
		{ 489, 167, 43 },
		{ 489, 168, 43 },
		{ 489, 169, 43 },
		{ 489, 170, 43 },
		{ 489, 171, 43 },
		{ 489, 172, 43 },
		{ 489, 173, 43 },
		{ 489, 174, 43 },
		{ 489, 175, 43 },
		{ 489, 176, 43 },
		{ 489, 177, 43 },
		{ 489, 178, 43 },
		{ 489, 179, 43 },
		{ 489, 180, 43 },
		{ 489, 181, 43 },
		{ 489, 182, 43 },
		{ 489, 183, 43 },
		{ 489, 184, 43 },
		{ 489, 185, 43 },
		{ 489, 186, 43 },
		{ 489, 187, 43 },
		{ 489, 188, 43 },
		{ 489, 189, 43 },
		{ 489, 190, 43 },
		{ 489, 191, 43 },
		{ 489, 192, 43 },
		{ 489, 193, 43 },
		{ 489, 194, 43 },
		{ 489, 195, 43 },
		{ 489, 196, 43 },
		{ 489, 197, 43 },
		{ 489, 198, 43 },
		{ 490, 141, 43 },
		{ 490, 142, 43 },
		{ 490, 143, 43 },
		{ 490, 144, 43 },
		{ 490, 145, 43 },
		{ 490, 146, 43 },
		{ 490, 147, 43 },
		{ 490, 148, 43 },
		{ 490, 149, 43 },
		{ 490, 150, 43 },
		{ 490, 151, 43 },
		{ 490, 152, 43 },
		{ 490, 153, 43 },
		{ 490, 154, 43 },
		{ 490, 155, 43 },
		{ 490, 156, 43 },
		{ 490, 157, 43 },
		{ 490, 158, 43 },
		{ 490, 159, 43 },
		{ 490, 160, 43 },
		{ 490, 161, 43 },
		{ 490, 162, 43 },
		{ 490, 163, 43 },
		{ 490, 164, 43 },
		{ 490, 165, 43 },
		{ 490, 166, 43 },
		{ 490, 167, 43 },
		{ 490, 168, 43 },
		{ 490, 169, 43 },
		{ 490, 170, 43 },
		{ 490, 171, 43 },
		{ 490, 172, 43 },
		{ 490, 173, 43 },
		{ 490, 174, 43 },
		{ 490, 175, 43 },
		{ 490, 176, 43 },
		{ 490, 177, 43 },
		{ 490, 178, 43 },
		{ 490, 179, 43 },
		{ 490, 180, 43 },
		{ 490, 181, 43 },
		{ 490, 182, 43 },
		{ 490, 183, 43 },
		{ 490, 184, 43 },
		{ 490, 185, 43 },
		{ 490, 186, 43 },
		{ 490, 187, 43 },
		{ 490, 188, 43 },
		{ 490, 189, 43 },
		{ 490, 190, 43 },
		{ 490, 191, 43 },
		{ 490, 192, 43 },
		{ 490, 193, 43 },
		{ 490, 194, 43 },
		{ 490, 195, 43 },
		{ 490, 196, 43 },
		{ 490, 197, 43 },
		{ 490, 198, 43 },
		{ 491, 141, 43 },
		{ 491, 142, 43 },
		{ 491, 143, 43 },
		{ 491, 144, 43 },
		{ 491, 145, 43 },
		{ 491, 146, 43 },
		{ 491, 147, 43 },
		{ 491, 148, 43 },
		{ 491, 149, 43 },
		{ 491, 150, 43 },
		{ 491, 151, 43 },
		{ 491, 152, 43 },
		{ 491, 153, 43 },
		{ 491, 154, 43 },
		{ 491, 155, 43 },
		{ 491, 156, 43 },
		{ 491, 157, 43 },
		{ 491, 158, 43 },
		{ 491, 159, 43 },
		{ 491, 160, 43 },
		{ 491, 161, 43 },
		{ 491, 162, 43 },
		{ 491, 163, 43 },
		{ 491, 164, 43 },
		{ 491, 165, 43 },
		{ 491, 166, 43 },
		{ 491, 167, 43 },
		{ 491, 168, 43 },
		{ 491, 169, 43 },
		{ 491, 170, 43 },
		{ 491, 171, 43 },
		{ 491, 172, 43 },
		{ 491, 173, 43 },
		{ 491, 174, 43 },
		{ 491, 175, 43 },
		{ 491, 176, 43 },
		{ 491, 177, 43 },
		{ 491, 178, 43 },
		{ 491, 179, 43 },
		{ 491, 180, 43 },
		{ 491, 181, 43 },
		{ 491, 182, 43 },
		{ 491, 183, 43 },
		{ 491, 184, 43 },
		{ 491, 185, 43 },
		{ 491, 186, 43 },
		{ 491, 187, 43 },
		{ 491, 188, 43 },
		{ 491, 189, 43 },
		{ 491, 190, 43 },
		{ 491, 191, 43 },
		{ 491, 192, 43 },
		{ 491, 193, 43 },
		{ 491, 194, 43 },
		{ 491, 195, 43 },
		{ 491, 196, 43 },
		{ 491, 197, 43 },
		{ 491, 198, 43 },
		{ 492, 140, 43 },
		{ 492, 141, 43 },
		{ 492, 142, 43 },
		{ 492, 143, 43 },
		{ 492, 144, 43 },
		{ 492, 145, 43 },
		{ 492, 146, 43 },
		{ 492, 147, 43 },
		{ 492, 148, 43 },
		{ 492, 149, 43 },
		{ 492, 150, 43 },
		{ 492, 151, 43 },
		{ 492, 152, 43 },
		{ 492, 153, 43 },
		{ 492, 154, 43 },
		{ 492, 155, 43 },
		{ 492, 156, 43 },
		{ 492, 157, 43 },
		{ 492, 158, 43 },
		{ 492, 159, 43 },
		{ 492, 160, 43 },
		{ 492, 161, 43 },
		{ 492, 162, 43 },
		{ 492, 163, 43 },
		{ 492, 164, 43 },
		{ 492, 165, 43 },
		{ 492, 166, 43 },
		{ 492, 167, 43 },
		{ 492, 168, 43 },
		{ 492, 169, 43 },
		{ 492, 170, 43 },
		{ 492, 171, 43 },
		{ 492, 172, 43 },
		{ 492, 173, 43 },
		{ 492, 174, 43 },
		{ 492, 175, 43 },
		{ 492, 176, 43 },
		{ 492, 177, 43 },
		{ 492, 178, 43 },
		{ 492, 179, 43 },
		{ 492, 180, 43 },
		{ 492, 181, 43 },
		{ 492, 182, 43 },
		{ 492, 183, 43 },
		{ 492, 184, 43 },
		{ 492, 185, 43 },
		{ 492, 186, 43 },
		{ 492, 187, 43 },
		{ 492, 188, 43 },
		{ 492, 189, 43 },
		{ 492, 190, 43 },
		{ 492, 191, 43 },
		{ 492, 192, 43 },
		{ 492, 193, 43 },
		{ 492, 194, 43 },
		{ 492, 195, 43 },
		{ 492, 196, 43 },
		{ 492, 197, 43 },
		{ 492, 198, 43 },
		{ 493, 140, 43 },
		{ 493, 141, 43 },
		{ 493, 142, 43 },
		{ 493, 143, 43 },
		{ 493, 144, 43 },
		{ 493, 145, 43 },
		{ 493, 146, 43 },
		{ 493, 147, 43 },
		{ 493, 148, 43 },
		{ 493, 149, 43 },
		{ 493, 150, 43 },
		{ 493, 151, 43 },
		{ 493, 152, 43 },
		{ 493, 153, 43 },
		{ 493, 154, 43 },
		{ 493, 155, 43 },
		{ 493, 156, 43 },
		{ 493, 157, 43 },
		{ 493, 158, 43 },
		{ 493, 159, 43 },
		{ 493, 160, 43 },
		{ 493, 161, 43 },
		{ 493, 162, 43 },
		{ 493, 163, 43 },
		{ 493, 164, 43 },
		{ 493, 165, 43 },
		{ 493, 166, 43 },
		{ 493, 167, 43 },
		{ 493, 168, 43 },
		{ 493, 169, 43 },
		{ 493, 170, 43 },
		{ 493, 171, 43 },
		{ 493, 172, 43 },
		{ 493, 173, 43 },
		{ 493, 174, 43 },
		{ 493, 175, 43 },
		{ 493, 176, 43 },
		{ 493, 177, 43 },
		{ 493, 178, 43 },
		{ 493, 179, 43 },
		{ 493, 180, 43 },
		{ 493, 181, 43 },
		{ 493, 182, 43 },
		{ 493, 183, 43 },
		{ 493, 184, 43 },
		{ 493, 185, 43 },
		{ 493, 186, 43 },
		{ 493, 187, 43 },
		{ 493, 188, 43 },
		{ 493, 189, 43 },
		{ 493, 190, 43 },
		{ 493, 191, 43 },
		{ 493, 192, 43 },
		{ 493, 193, 43 },
		{ 493, 194, 43 },
		{ 493, 195, 43 },
		{ 493, 196, 43 },
		{ 493, 197, 43 },
		{ 494, 139, 43 },
		{ 494, 140, 43 },
		{ 494, 141, 43 },
		{ 494, 142, 43 },
		{ 494, 143, 43 },
		{ 494, 144, 43 },
		{ 494, 145, 43 },
		{ 494, 146, 43 },
		{ 494, 147, 43 },
		{ 494, 148, 43 },
		{ 494, 149, 43 },
		{ 494, 150, 43 },
		{ 494, 151, 43 },
		{ 494, 152, 43 },
		{ 494, 153, 43 },
		{ 494, 154, 43 },
		{ 494, 155, 43 },
		{ 494, 156, 43 },
		{ 494, 157, 43 },
		{ 494, 158, 43 },
		{ 494, 159, 43 },
		{ 494, 160, 43 },
		{ 494, 161, 43 },
		{ 494, 162, 43 },
		{ 494, 163, 43 },
		{ 494, 164, 43 },
		{ 494, 165, 43 },
		{ 494, 166, 43 },
		{ 494, 167, 43 },
		{ 494, 168, 43 },
		{ 494, 169, 43 },
		{ 494, 170, 43 },
		{ 494, 171, 43 },
		{ 494, 172, 43 },
		{ 494, 173, 43 },
		{ 494, 174, 43 },
		{ 494, 175, 43 },
		{ 494, 176, 43 },
		{ 494, 177, 43 },
		{ 494, 178, 43 },
		{ 494, 179, 43 },
		{ 494, 180, 43 },
		{ 494, 181, 43 },
		{ 494, 182, 43 },
		{ 494, 183, 43 },
		{ 494, 184, 43 },
		{ 494, 185, 43 },
		{ 494, 186, 43 },
		{ 494, 187, 43 },
		{ 494, 188, 43 },
		{ 494, 189, 43 },
		{ 494, 190, 43 },
		{ 494, 191, 43 },
		{ 494, 192, 43 },
		{ 494, 193, 43 },
		{ 494, 194, 43 },
		{ 494, 195, 43 },
		{ 494, 196, 43 },
		{ 494, 197, 43 },
		{ 495, 139, 43 },
		{ 495, 140, 43 },
		{ 495, 141, 43 },
		{ 495, 142, 43 },
		{ 495, 143, 43 },
		{ 495, 144, 43 },
		{ 495, 145, 43 },
		{ 495, 146, 43 },
		{ 495, 147, 43 },
		{ 495, 148, 43 },
		{ 495, 149, 43 },
		{ 495, 150, 43 },
		{ 495, 151, 43 },
		{ 495, 152, 43 },
		{ 495, 153, 43 },
		{ 495, 154, 43 },
		{ 495, 155, 43 },
		{ 495, 156, 43 },
		{ 495, 157, 43 },
		{ 495, 158, 43 },
		{ 495, 159, 43 },
		{ 495, 160, 43 },
		{ 495, 161, 43 },
		{ 495, 162, 43 },
		{ 495, 163, 43 },
		{ 495, 164, 43 },
		{ 495, 165, 43 },
		{ 495, 166, 43 },
		{ 495, 167, 43 },
		{ 495, 168, 43 },
		{ 495, 169, 43 },
		{ 495, 170, 43 },
		{ 495, 171, 43 },
		{ 495, 172, 43 },
		{ 495, 173, 43 },
		{ 495, 174, 43 },
		{ 495, 175, 43 },
		{ 495, 176, 43 },
		{ 495, 177, 43 },
		{ 495, 178, 43 },
		{ 495, 179, 43 },
		{ 495, 180, 43 },
		{ 495, 181, 43 },
		{ 495, 182, 43 },
		{ 495, 183, 43 },
		{ 495, 184, 43 },
		{ 495, 185, 43 },
		{ 495, 186, 43 },
		{ 495, 187, 43 },
		{ 495, 188, 43 },
		{ 495, 189, 43 },
		{ 495, 190, 43 },
		{ 495, 191, 43 },
		{ 495, 192, 43 },
		{ 495, 193, 43 },
		{ 495, 194, 43 },
		{ 495, 195, 43 },
		{ 495, 196, 43 },
		{ 495, 197, 43 },
		{ 496, 138, 43 },
		{ 496, 139, 43 },
		{ 496, 140, 43 },
		{ 496, 141, 43 },
		{ 496, 142, 43 },
		{ 496, 143, 43 },
		{ 496, 144, 43 },
		{ 496, 145, 43 },
		{ 496, 146, 43 },
		{ 496, 147, 43 },
		{ 496, 148, 43 },
		{ 496, 149, 43 },
		{ 496, 150, 43 },
		{ 496, 151, 43 },
		{ 496, 152, 43 },
		{ 496, 153, 43 },
		{ 496, 154, 43 },
		{ 496, 155, 43 },
		{ 496, 156, 43 },
		{ 496, 157, 43 },
		{ 496, 158, 43 },
		{ 496, 159, 43 },
		{ 496, 160, 43 },
		{ 496, 161, 43 },
		{ 496, 162, 43 },
		{ 496, 163, 43 },
		{ 496, 164, 43 },
		{ 496, 165, 43 },
		{ 496, 166, 43 },
		{ 496, 167, 43 },
		{ 496, 168, 43 },
		{ 496, 169, 43 },
		{ 496, 170, 43 },
		{ 496, 171, 43 },
		{ 496, 172, 43 },
		{ 496, 173, 43 },
		{ 496, 174, 43 },
		{ 496, 175, 43 },
		{ 496, 176, 43 },
		{ 496, 177, 43 },
		{ 496, 178, 43 },
		{ 496, 179, 43 },
		{ 496, 180, 43 },
		{ 496, 181, 43 },
		{ 496, 182, 43 },
		{ 496, 183, 43 },
		{ 496, 184, 43 },
		{ 496, 185, 43 },
		{ 496, 186, 43 },
		{ 496, 187, 43 },
		{ 496, 188, 43 },
		{ 496, 189, 43 },
		{ 496, 190, 43 },
		{ 496, 191, 43 },
		{ 496, 192, 43 },
		{ 496, 193, 43 },
		{ 496, 194, 43 },
		{ 496, 195, 43 },
		{ 496, 196, 43 },
		{ 496, 197, 43 },
		{ 497, 138, 43 },
		{ 497, 139, 43 },
		{ 497, 140, 43 },
		{ 497, 141, 43 },
		{ 497, 142, 43 },
		{ 497, 143, 43 },
		{ 497, 144, 43 },
		{ 497, 145, 43 },
		{ 497, 146, 43 },
		{ 497, 147, 43 },
		{ 497, 148, 43 },
		{ 497, 149, 43 },
		{ 497, 150, 43 },
		{ 497, 151, 43 },
		{ 497, 152, 43 },
		{ 497, 153, 43 },
		{ 497, 154, 43 },
		{ 497, 155, 43 },
		{ 497, 156, 43 },
		{ 497, 157, 43 },
		{ 497, 158, 43 },
		{ 497, 159, 43 },
		{ 497, 160, 43 },
		{ 497, 161, 43 },
		{ 497, 162, 43 },
		{ 497, 163, 43 },
		{ 497, 164, 43 },
		{ 497, 165, 43 },
		{ 497, 166, 43 },
		{ 497, 167, 43 },
		{ 497, 168, 43 },
		{ 497, 169, 43 },
		{ 497, 170, 43 },
		{ 497, 171, 43 },
		{ 497, 172, 43 },
		{ 497, 173, 43 },
		{ 497, 174, 43 },
		{ 497, 175, 43 },
		{ 497, 176, 43 },
		{ 497, 177, 43 },
		{ 497, 178, 43 },
		{ 497, 179, 43 },
		{ 497, 180, 43 },
		{ 497, 181, 43 },
		{ 497, 182, 43 },
		{ 497, 183, 43 },
		{ 497, 184, 43 },
		{ 497, 185, 43 },
		{ 497, 186, 43 },
		{ 497, 187, 43 },
		{ 497, 188, 43 },
		{ 497, 189, 43 },
		{ 497, 190, 43 },
		{ 497, 191, 43 },
		{ 497, 192, 43 },
		{ 497, 193, 43 },
		{ 497, 194, 43 },
		{ 497, 195, 43 },
		{ 497, 196, 43 },
		{ 498, 137, 43 },
		{ 498, 138, 43 },
		{ 498, 139, 43 },
		{ 498, 140, 43 },
		{ 498, 141, 43 },
		{ 498, 142, 43 },
		{ 498, 143, 43 },
		{ 498, 144, 43 },
		{ 498, 145, 43 },
		{ 498, 146, 43 },
		{ 498, 147, 43 },
		{ 498, 148, 43 },
		{ 498, 149, 43 },
		{ 498, 150, 43 },
		{ 498, 151, 43 },
		{ 498, 152, 43 },
		{ 498, 153, 43 },
		{ 498, 154, 43 },
		{ 498, 155, 43 },
		{ 498, 156, 43 },
		{ 498, 157, 43 },
		{ 498, 158, 43 },
		{ 498, 159, 43 },
		{ 498, 160, 43 },
		{ 498, 161, 43 },
		{ 498, 162, 43 },
		{ 498, 163, 43 },
		{ 498, 164, 43 },
		{ 498, 165, 43 },
		{ 498, 166, 43 },
		{ 498, 167, 43 },
		{ 498, 168, 43 },
		{ 498, 169, 43 },
		{ 498, 170, 43 },
		{ 498, 171, 43 },
		{ 498, 172, 43 },
		{ 498, 173, 43 },
		{ 498, 174, 43 },
		{ 498, 175, 43 },
		{ 498, 176, 43 },
		{ 498, 177, 43 },
		{ 498, 178, 43 },
		{ 498, 179, 43 },
		{ 498, 180, 43 },
		{ 498, 181, 43 },
		{ 498, 182, 43 },
		{ 498, 183, 43 },
		{ 498, 184, 43 },
		{ 498, 185, 43 },
		{ 498, 186, 43 },
		{ 498, 187, 43 },
		{ 498, 188, 43 },
		{ 498, 189, 43 },
		{ 498, 190, 43 },
		{ 498, 191, 43 },
		{ 498, 192, 43 },
		{ 498, 193, 43 },
		{ 498, 194, 43 },
		{ 498, 195, 43 },
		{ 498, 196, 43 },
		{ 499, 137, 43 },
		{ 499, 138, 43 },
		{ 499, 139, 43 },
		{ 499, 140, 43 },
		{ 499, 141, 43 },
		{ 499, 142, 43 },
		{ 499, 143, 43 },
		{ 499, 144, 43 },
		{ 499, 145, 43 },
		{ 499, 146, 43 },
		{ 499, 147, 43 },
		{ 499, 148, 43 },
		{ 499, 149, 43 },
		{ 499, 150, 43 },
		{ 499, 151, 43 },
		{ 499, 152, 43 },
		{ 499, 153, 43 },
		{ 499, 154, 43 },
		{ 499, 155, 43 },
		{ 499, 156, 43 },
		{ 499, 157, 43 },
		{ 499, 158, 43 },
		{ 499, 159, 43 },
		{ 499, 160, 43 },
		{ 499, 161, 43 },
		{ 499, 162, 43 },
		{ 499, 163, 43 },
		{ 499, 164, 43 },
		{ 499, 165, 43 },
		{ 499, 166, 43 },
		{ 499, 167, 43 },
		{ 499, 168, 43 },
		{ 499, 169, 43 },
		{ 499, 170, 43 },
		{ 499, 171, 43 },
		{ 499, 172, 43 },
		{ 499, 173, 43 },
		{ 499, 174, 43 },
		{ 499, 175, 43 },
		{ 499, 176, 43 },
		{ 499, 177, 43 },
		{ 499, 178, 43 },
		{ 499, 179, 43 },
		{ 499, 180, 43 },
		{ 499, 181, 43 },
		{ 499, 182, 43 },
		{ 499, 183, 43 },
		{ 499, 184, 43 },
		{ 499, 185, 43 },
		{ 499, 186, 43 },
		{ 499, 187, 43 },
		{ 499, 188, 43 },
		{ 499, 189, 43 },
		{ 499, 190, 43 },
		{ 499, 191, 43 },
		{ 499, 192, 43 },
		{ 499, 193, 43 },
		{ 499, 194, 43 },
		{ 499, 195, 43 },
		{ 499, 196, 43 },
		{ 500, 136, 43 },
		{ 500, 137, 43 },
		{ 500, 138, 43 },
		{ 500, 139, 43 },
		{ 500, 140, 43 },
		{ 500, 141, 43 },
		{ 500, 142, 43 },
		{ 500, 143, 43 },
		{ 500, 144, 43 },
		{ 500, 145, 43 },
		{ 500, 146, 43 },
		{ 500, 147, 43 },
		{ 500, 148, 43 },
		{ 500, 149, 43 },
		{ 500, 150, 43 },
		{ 500, 151, 43 },
		{ 500, 152, 43 },
		{ 500, 153, 43 },
		{ 500, 154, 43 },
		{ 500, 155, 43 },
		{ 500, 156, 43 },
		{ 500, 157, 43 },
		{ 500, 158, 43 },
		{ 500, 159, 43 },
		{ 500, 160, 43 },
		{ 500, 161, 43 },
		{ 500, 162, 43 },
		{ 500, 163, 43 },
		{ 500, 164, 43 },
		{ 500, 165, 43 },
		{ 500, 166, 43 },
		{ 500, 167, 43 },
		{ 500, 168, 43 },
		{ 500, 169, 43 },
		{ 500, 170, 43 },
		{ 500, 171, 43 },
		{ 500, 172, 43 },
		{ 500, 173, 43 },
		{ 500, 174, 43 },
		{ 500, 175, 43 },
		{ 500, 176, 43 },
		{ 500, 177, 43 },
		{ 500, 178, 43 },
		{ 500, 179, 43 },
		{ 500, 180, 43 },
		{ 500, 181, 43 },
		{ 500, 182, 43 },
		{ 500, 183, 43 },
		{ 500, 184, 43 },
		{ 500, 185, 43 },
		{ 500, 186, 43 },
		{ 500, 187, 43 },
		{ 500, 188, 43 },
		{ 500, 189, 43 },
		{ 500, 190, 43 },
		{ 500, 191, 43 },
		{ 500, 192, 43 },
		{ 500, 193, 43 },
		{ 500, 194, 43 },
		{ 500, 195, 43 },
		{ 500, 196, 43 },
		{ 501, 136, 43 },
		{ 501, 137, 43 },
		{ 501, 138, 43 },
		{ 501, 139, 43 },
		{ 501, 140, 43 },
		{ 501, 141, 43 },
		{ 501, 142, 43 },
		{ 501, 143, 43 },
		{ 501, 144, 43 },
		{ 501, 145, 43 },
		{ 501, 146, 43 },
		{ 501, 147, 43 },
		{ 501, 148, 43 },
		{ 501, 149, 43 },
		{ 501, 150, 43 },
		{ 501, 151, 43 },
		{ 501, 152, 43 },
		{ 501, 153, 43 },
		{ 501, 154, 43 },
		{ 501, 155, 43 },
		{ 501, 156, 43 },
		{ 501, 157, 43 },
		{ 501, 158, 43 },
		{ 501, 159, 43 },
		{ 501, 160, 43 },
		{ 501, 161, 43 },
		{ 501, 162, 43 },
		{ 501, 163, 43 },
		{ 501, 164, 43 },
		{ 501, 165, 43 },
		{ 501, 166, 43 },
		{ 501, 167, 43 },
		{ 501, 168, 43 },
		{ 501, 169, 43 },
		{ 501, 170, 43 },
		{ 501, 171, 43 },
		{ 501, 172, 43 },
		{ 501, 173, 43 },
		{ 501, 174, 43 },
		{ 501, 175, 43 },
		{ 501, 176, 43 },
		{ 501, 177, 43 },
		{ 501, 178, 43 },
		{ 501, 179, 43 },
		{ 501, 180, 43 },
		{ 501, 181, 43 },
		{ 501, 182, 43 },
		{ 501, 183, 43 },
		{ 501, 184, 43 },
		{ 501, 185, 43 },
		{ 501, 186, 43 },
		{ 501, 187, 43 },
		{ 501, 188, 43 },
		{ 501, 189, 43 },
		{ 501, 190, 43 },
		{ 501, 191, 43 },
		{ 501, 192, 43 },
		{ 501, 193, 43 },
		{ 501, 194, 43 },
		{ 501, 195, 43 },
		{ 502, 135, 43 },
		{ 502, 136, 43 },
		{ 502, 137, 43 },
		{ 502, 138, 43 },
		{ 502, 139, 43 },
		{ 502, 140, 43 },
		{ 502, 141, 43 },
		{ 502, 142, 43 },
		{ 502, 143, 43 },
		{ 502, 144, 43 },
		{ 502, 145, 43 },
		{ 502, 146, 43 },
		{ 502, 147, 43 },
		{ 502, 148, 43 },
		{ 502, 149, 43 },
		{ 502, 150, 43 },
		{ 502, 151, 43 },
		{ 502, 152, 43 },
		{ 502, 153, 43 },
		{ 502, 154, 43 },
		{ 502, 155, 43 },
		{ 502, 156, 43 },
		{ 502, 157, 43 },
		{ 502, 158, 43 },
		{ 502, 159, 43 },
		{ 502, 160, 43 },
		{ 502, 161, 43 },
		{ 502, 162, 43 },
		{ 502, 163, 43 },
		{ 502, 164, 43 },
		{ 502, 165, 43 },
		{ 502, 166, 43 },
		{ 502, 167, 43 },
		{ 502, 168, 43 },
		{ 502, 169, 43 },
		{ 502, 170, 43 },
		{ 502, 171, 43 },
		{ 502, 172, 43 },
		{ 502, 173, 43 },
		{ 502, 174, 43 },
		{ 502, 175, 43 },
		{ 502, 176, 43 },
		{ 502, 177, 43 },
		{ 502, 178, 43 },
		{ 502, 179, 43 },
		{ 502, 180, 43 },
		{ 502, 181, 43 },
		{ 502, 182, 43 },
		{ 502, 183, 43 },
		{ 502, 184, 43 },
		{ 502, 185, 43 },
		{ 502, 186, 43 },
		{ 502, 187, 43 },
		{ 502, 188, 43 },
		{ 502, 189, 43 },
		{ 502, 190, 43 },
		{ 502, 191, 43 },
		{ 502, 192, 43 },
		{ 502, 193, 43 },
		{ 502, 194, 43 },
		{ 502, 195, 43 },
		{ 503, 135, 43 },
		{ 503, 136, 43 },
		{ 503, 137, 43 },
		{ 503, 138, 43 },
		{ 503, 139, 43 },
		{ 503, 140, 43 },
		{ 503, 141, 43 },
		{ 503, 142, 43 },
		{ 503, 143, 43 },
		{ 503, 144, 43 },
		{ 503, 145, 43 },
		{ 503, 146, 43 },
		{ 503, 147, 43 },
		{ 503, 148, 43 },
		{ 503, 149, 43 },
		{ 503, 150, 43 },
		{ 503, 151, 43 },
		{ 503, 152, 43 },
		{ 503, 153, 43 },
		{ 503, 154, 43 },
		{ 503, 155, 43 },
		{ 503, 156, 43 },
		{ 503, 157, 43 },
		{ 503, 158, 43 },
		{ 503, 159, 43 },
		{ 503, 160, 43 },
		{ 503, 161, 43 },
		{ 503, 162, 43 },
		{ 503, 163, 43 },
		{ 503, 164, 43 },
		{ 503, 165, 43 },
		{ 503, 166, 43 },
		{ 503, 167, 43 },
		{ 503, 168, 43 },
		{ 503, 169, 43 },
		{ 503, 170, 43 },
		{ 503, 171, 43 },
		{ 503, 172, 43 },
		{ 503, 173, 43 },
		{ 503, 174, 43 },
		{ 503, 175, 43 },
		{ 503, 176, 43 },
		{ 503, 177, 43 },
		{ 503, 178, 43 },
		{ 503, 179, 43 },
		{ 503, 180, 43 },
		{ 503, 181, 43 },
		{ 503, 182, 43 },
		{ 503, 183, 43 },
		{ 503, 184, 43 },
		{ 503, 185, 43 },
		{ 503, 186, 43 },
		{ 503, 187, 43 },
		{ 503, 188, 43 },
		{ 503, 189, 43 },
		{ 503, 190, 43 },
		{ 503, 191, 43 },
		{ 503, 192, 43 },
		{ 503, 193, 43 },
		{ 503, 194, 43 },
		{ 503, 195, 43 },
		{ 504, 134, 43 },
		{ 504, 135, 43 },
		{ 504, 136, 43 },
		{ 504, 137, 43 },
		{ 504, 138, 43 },
		{ 504, 139, 43 },
		{ 504, 140, 43 },
		{ 504, 141, 43 },
		{ 504, 142, 43 },
		{ 504, 143, 43 },
		{ 504, 144, 43 },
		{ 504, 145, 43 },
		{ 504, 146, 43 },
		{ 504, 147, 43 },
		{ 504, 148, 43 },
		{ 504, 149, 43 },
		{ 504, 150, 43 },
		{ 504, 151, 43 },
		{ 504, 152, 43 },
		{ 504, 153, 43 },
		{ 504, 154, 43 },
		{ 504, 155, 43 },
		{ 504, 156, 43 },
		{ 504, 157, 43 },
		{ 504, 158, 43 },
		{ 504, 159, 43 },
		{ 504, 160, 43 },
		{ 504, 161, 43 },
		{ 504, 162, 43 },
		{ 504, 163, 43 },
		{ 504, 164, 43 },
		{ 504, 165, 43 },
		{ 504, 166, 43 },
		{ 504, 167, 43 },
		{ 504, 168, 43 },
		{ 504, 169, 43 },
		{ 504, 170, 43 },
		{ 504, 171, 43 },
		{ 504, 172, 43 },
		{ 504, 173, 43 },
		{ 504, 174, 43 },
		{ 504, 175, 43 },
		{ 504, 176, 43 },
		{ 504, 177, 43 },
		{ 504, 178, 43 },
		{ 504, 179, 43 },
		{ 504, 180, 43 },
		{ 504, 181, 43 },
		{ 504, 182, 43 },
		{ 504, 183, 43 },
		{ 504, 184, 43 },
		{ 504, 185, 43 },
		{ 504, 186, 43 },
		{ 504, 187, 43 },
		{ 504, 188, 43 },
		{ 504, 189, 43 },
		{ 504, 190, 43 },
		{ 504, 191, 43 },
		{ 504, 192, 43 },
		{ 504, 193, 43 },
		{ 504, 194, 43 },
		{ 504, 195, 43 },
		{ 505, 134, 43 },
		{ 505, 135, 43 },
		{ 505, 136, 43 },
		{ 505, 137, 43 },
		{ 505, 138, 43 },
		{ 505, 139, 43 },
		{ 505, 140, 43 },
		{ 505, 141, 43 },
		{ 505, 142, 43 },
		{ 505, 143, 43 },
		{ 505, 144, 43 },
		{ 505, 145, 43 },
		{ 505, 146, 43 },
		{ 505, 147, 43 },
		{ 505, 148, 43 },
		{ 505, 149, 43 },
		{ 505, 150, 43 },
		{ 505, 151, 43 },
		{ 505, 152, 43 },
		{ 505, 153, 43 },
		{ 505, 154, 43 },
		{ 505, 155, 43 },
		{ 505, 156, 43 },
		{ 505, 157, 43 },
		{ 505, 158, 43 },
		{ 505, 159, 43 },
		{ 505, 160, 43 },
		{ 505, 161, 43 },
		{ 505, 162, 43 },
		{ 505, 163, 43 },
		{ 505, 164, 43 },
		{ 505, 165, 43 },
		{ 505, 166, 43 },
		{ 505, 167, 43 },
		{ 505, 168, 43 },
		{ 505, 169, 43 },
		{ 505, 170, 43 },
		{ 505, 171, 43 },
		{ 505, 172, 43 },
		{ 505, 173, 43 },
		{ 505, 174, 43 },
		{ 505, 175, 43 },
		{ 505, 176, 43 },
		{ 505, 177, 43 },
		{ 505, 178, 43 },
		{ 505, 179, 43 },
		{ 505, 180, 43 },
		{ 505, 181, 43 },
		{ 505, 182, 43 },
		{ 505, 183, 43 },
		{ 505, 184, 43 },
		{ 505, 185, 43 },
		{ 505, 186, 43 },
		{ 505, 187, 43 },
		{ 505, 188, 43 },
		{ 505, 189, 43 },
		{ 505, 190, 43 },
		{ 505, 191, 43 },
		{ 505, 192, 43 },
		{ 505, 193, 43 },
		{ 505, 194, 43 },
		{ 506, 133, 43 },
		{ 506, 134, 43 },
		{ 506, 135, 43 },
		{ 506, 136, 43 },
		{ 506, 137, 43 },
		{ 506, 138, 43 },
		{ 506, 139, 43 },
		{ 506, 140, 43 },
		{ 506, 141, 43 },
		{ 506, 142, 43 },
		{ 506, 143, 43 },
		{ 506, 144, 43 },
		{ 506, 145, 43 },
		{ 506, 146, 43 },
		{ 506, 147, 43 },
		{ 506, 148, 43 },
		{ 506, 149, 43 },
		{ 506, 150, 43 },
		{ 506, 151, 43 },
		{ 506, 152, 43 },
		{ 506, 153, 43 },
		{ 506, 154, 43 },
		{ 506, 155, 43 },
		{ 506, 156, 43 },
		{ 506, 157, 43 },
		{ 506, 158, 43 },
		{ 506, 159, 43 },
		{ 506, 160, 43 },
		{ 506, 161, 43 },
		{ 506, 162, 43 },
		{ 506, 163, 43 },
		{ 506, 164, 43 },
		{ 506, 165, 43 },
		{ 506, 166, 43 },
		{ 506, 167, 43 },
		{ 506, 168, 43 },
		{ 506, 169, 43 },
		{ 506, 170, 43 },
		{ 506, 171, 43 },
		{ 506, 172, 43 },
		{ 506, 173, 43 },
		{ 506, 174, 43 },
		{ 506, 175, 43 },
		{ 506, 176, 43 },
		{ 506, 177, 43 },
		{ 506, 178, 43 },
		{ 506, 179, 43 },
		{ 506, 180, 43 },
		{ 506, 181, 43 },
		{ 506, 182, 43 },
		{ 506, 183, 43 },
		{ 506, 184, 43 },
		{ 506, 185, 43 },
		{ 506, 186, 43 },
		{ 506, 187, 43 },
		{ 506, 188, 43 },
		{ 506, 189, 43 },
		{ 506, 190, 43 },
		{ 506, 191, 43 },
		{ 506, 192, 43 },
		{ 506, 193, 43 },
		{ 506, 194, 43 },
		{ 507, 133, 43 },
		{ 507, 134, 43 },
		{ 507, 135, 43 },
		{ 507, 136, 43 },
		{ 507, 137, 43 },
		{ 507, 138, 43 },
		{ 507, 139, 43 },
		{ 507, 140, 43 },
		{ 507, 141, 43 },
		{ 507, 142, 43 },
		{ 507, 143, 43 },
		{ 507, 144, 43 },
		{ 507, 145, 43 },
		{ 507, 146, 43 },
		{ 507, 147, 43 },
		{ 507, 148, 43 },
		{ 507, 149, 43 },
		{ 507, 150, 43 },
		{ 507, 151, 43 },
		{ 507, 152, 43 },
		{ 507, 153, 43 },
		{ 507, 154, 43 },
		{ 507, 155, 43 },
		{ 507, 156, 43 },
		{ 507, 157, 43 },
		{ 507, 158, 43 },
		{ 507, 159, 43 },
		{ 507, 160, 43 },
		{ 507, 161, 43 },
		{ 507, 162, 43 },
		{ 507, 163, 43 },
		{ 507, 164, 43 },
		{ 507, 165, 43 },
		{ 507, 166, 43 },
		{ 507, 167, 43 },
		{ 507, 168, 43 },
		{ 507, 169, 43 },
		{ 507, 170, 43 },
		{ 507, 171, 43 },
		{ 507, 172, 43 },
		{ 507, 173, 43 },
		{ 507, 174, 43 },
		{ 507, 175, 43 },
		{ 507, 176, 43 },
		{ 507, 177, 43 },
		{ 507, 178, 43 },
		{ 507, 179, 43 },
		{ 507, 180, 43 },
		{ 507, 181, 43 },
		{ 507, 182, 43 },
		{ 507, 183, 43 },
		{ 507, 184, 43 },
		{ 507, 185, 43 },
		{ 507, 186, 43 },
		{ 507, 187, 43 },
		{ 507, 188, 43 },
		{ 507, 189, 43 },
		{ 507, 190, 43 },
		{ 507, 191, 43 },
		{ 507, 192, 43 },
		{ 507, 193, 43 },
		{ 507, 194, 43 },
		{ 508, 132, 43 },
		{ 508, 133, 43 },
		{ 508, 134, 43 },
		{ 508, 135, 43 },
		{ 508, 136, 43 },
		{ 508, 137, 43 },
		{ 508, 138, 43 },
		{ 508, 139, 43 },
		{ 508, 140, 43 },
		{ 508, 141, 43 },
		{ 508, 142, 43 },
		{ 508, 143, 43 },
		{ 508, 144, 43 },
		{ 508, 145, 43 },
		{ 508, 146, 43 },
		{ 508, 147, 43 },
		{ 508, 148, 43 },
		{ 508, 149, 43 },
		{ 508, 150, 43 },
		{ 508, 151, 43 },
		{ 508, 152, 43 },
		{ 508, 153, 43 },
		{ 508, 154, 43 },
		{ 508, 155, 43 },
		{ 508, 156, 43 },
		{ 508, 157, 43 },
		{ 508, 158, 43 },
		{ 508, 159, 43 },
		{ 508, 160, 43 },
		{ 508, 161, 43 },
		{ 508, 162, 43 },
		{ 508, 163, 43 },
		{ 508, 164, 43 },
		{ 508, 165, 43 },
		{ 508, 166, 43 },
		{ 508, 167, 43 },
		{ 508, 168, 43 },
		{ 508, 169, 43 },
		{ 508, 170, 43 },
		{ 508, 171, 43 },
		{ 508, 172, 43 },
		{ 508, 173, 43 },
		{ 508, 174, 43 },
		{ 508, 175, 43 },
		{ 508, 176, 43 },
		{ 508, 177, 43 },
		{ 508, 178, 43 },
		{ 508, 179, 43 },
		{ 508, 180, 43 },
		{ 508, 181, 43 },
		{ 508, 182, 43 },
		{ 508, 183, 43 },
		{ 508, 184, 43 },
		{ 508, 185, 43 },
		{ 508, 186, 43 },
		{ 508, 187, 43 },
		{ 508, 188, 43 },
		{ 508, 189, 43 },
		{ 508, 190, 43 },
		{ 508, 191, 43 },
		{ 508, 192, 43 },
		{ 508, 193, 43 },
		{ 508, 194, 43 },
		{ 509, 132, 43 },
		{ 509, 133, 43 },
		{ 509, 134, 43 },
		{ 509, 135, 43 },
		{ 509, 136, 43 },
		{ 509, 137, 43 },
		{ 509, 138, 43 },
		{ 509, 139, 43 },
		{ 509, 140, 43 },
		{ 509, 141, 43 },
		{ 509, 142, 43 },
		{ 509, 143, 43 },
		{ 509, 144, 43 },
		{ 509, 145, 43 },
		{ 509, 146, 43 },
		{ 509, 147, 43 },
		{ 509, 148, 43 },
		{ 509, 149, 43 },
		{ 509, 150, 43 },
		{ 509, 151, 43 },
		{ 509, 152, 43 },
		{ 509, 153, 43 },
		{ 509, 154, 43 },
		{ 509, 155, 43 },
		{ 509, 156, 43 },
		{ 509, 157, 43 },
		{ 509, 158, 43 },
		{ 509, 159, 43 },
		{ 509, 160, 43 },
		{ 509, 161, 43 },
		{ 509, 162, 43 },
		{ 509, 163, 43 },
		{ 509, 164, 43 },
		{ 509, 165, 43 },
		{ 509, 166, 43 },
		{ 509, 167, 43 },
		{ 509, 168, 43 },
		{ 509, 169, 43 },
		{ 509, 170, 43 },
		{ 509, 171, 43 },
		{ 509, 172, 43 },
		{ 509, 173, 43 },
		{ 509, 174, 43 },
		{ 509, 175, 43 },
		{ 509, 176, 43 },
		{ 509, 177, 43 },
		{ 509, 178, 43 },
		{ 509, 179, 43 },
		{ 509, 180, 43 },
		{ 509, 181, 43 },
		{ 509, 182, 43 },
		{ 509, 183, 43 },
		{ 509, 184, 43 },
		{ 509, 185, 43 },
		{ 509, 186, 43 },
		{ 509, 187, 43 },
		{ 509, 188, 43 },
		{ 509, 189, 43 },
		{ 509, 190, 43 },
		{ 509, 191, 43 },
		{ 509, 192, 43 },
		{ 509, 193, 43 },
		{ 510, 131, 43 },
		{ 510, 132, 43 },
		{ 510, 133, 43 },
		{ 510, 134, 43 },
		{ 510, 135, 43 },
		{ 510, 136, 43 },
		{ 510, 137, 43 },
		{ 510, 138, 43 },
		{ 510, 139, 43 },
		{ 510, 140, 43 },
		{ 510, 141, 43 },
		{ 510, 142, 43 },
		{ 510, 143, 43 },
		{ 510, 144, 43 },
		{ 510, 145, 43 },
		{ 510, 146, 43 },
		{ 510, 147, 43 },
		{ 510, 148, 43 },
		{ 510, 149, 43 },
		{ 510, 150, 43 },
		{ 510, 151, 43 },
		{ 510, 152, 43 },
		{ 510, 153, 43 },
		{ 510, 154, 43 },
		{ 510, 155, 43 },
		{ 510, 156, 43 },
		{ 510, 157, 43 },
		{ 510, 158, 43 },
		{ 510, 159, 43 },
		{ 510, 160, 43 },
		{ 510, 161, 43 },
		{ 510, 162, 43 },
		{ 510, 163, 43 },
		{ 510, 164, 43 },
		{ 510, 165, 43 },
		{ 510, 166, 43 },
		{ 510, 167, 43 },
		{ 510, 168, 43 },
		{ 510, 169, 43 },
		{ 510, 170, 43 },
		{ 510, 171, 43 },
		{ 510, 172, 43 },
		{ 510, 173, 43 },
		{ 510, 174, 43 },
		{ 510, 175, 43 },
		{ 510, 176, 43 },
		{ 510, 177, 43 },
		{ 510, 178, 43 },
		{ 510, 179, 43 },
		{ 510, 180, 43 },
		{ 510, 181, 43 },
		{ 510, 182, 43 },
		{ 510, 183, 43 },
		{ 510, 184, 43 },
		{ 510, 185, 43 },
		{ 510, 186, 43 },
		{ 510, 187, 43 },
		{ 510, 188, 43 },
		{ 510, 189, 43 },
		{ 510, 190, 43 },
		{ 510, 191, 43 },
		{ 510, 192, 43 },
		{ 510, 193, 43 },
		{ 511, 131, 43 },
		{ 511, 132, 43 },
		{ 511, 133, 43 },
		{ 511, 134, 43 },
		{ 511, 135, 43 },
		{ 511, 136, 43 },
		{ 511, 137, 43 },
		{ 511, 138, 43 },
		{ 511, 139, 43 },
		{ 511, 140, 43 },
		{ 511, 141, 43 },
		{ 511, 142, 43 },
		{ 511, 143, 43 },
		{ 511, 144, 43 },
		{ 511, 145, 43 },
		{ 511, 146, 43 },
		{ 511, 147, 43 },
		{ 511, 148, 43 },
		{ 511, 149, 43 },
		{ 511, 150, 43 },
		{ 511, 151, 43 },
		{ 511, 152, 43 },
		{ 511, 153, 43 },
		{ 511, 154, 43 },
		{ 511, 155, 43 },
		{ 511, 156, 43 },
		{ 511, 157, 43 },
		{ 511, 158, 43 },
		{ 511, 159, 43 },
		{ 511, 160, 43 },
		{ 511, 161, 43 },
		{ 511, 162, 43 },
		{ 511, 163, 43 },
		{ 511, 164, 43 },
		{ 511, 165, 43 },
		{ 511, 166, 43 },
		{ 511, 167, 43 },
		{ 511, 168, 43 },
		{ 511, 169, 43 },
		{ 511, 170, 43 },
		{ 511, 171, 43 },
		{ 511, 172, 43 },
		{ 511, 173, 43 },
		{ 511, 174, 43 },
		{ 511, 175, 43 },
		{ 511, 176, 43 },
		{ 511, 177, 43 },
		{ 511, 178, 43 },
		{ 511, 179, 43 },
		{ 511, 180, 43 },
		{ 511, 181, 43 },
		{ 511, 182, 43 },
		{ 511, 183, 43 },
		{ 511, 184, 43 },
		{ 511, 185, 43 },
		{ 511, 186, 43 },
		{ 511, 187, 43 },
		{ 511, 188, 43 },
		{ 511, 189, 43 },
		{ 511, 190, 43 },
		{ 511, 191, 43 },
		{ 511, 192, 43 },
		{ 511, 193, 43 },
		{ 512, 130, 43 },
		{ 512, 131, 43 },
		{ 512, 132, 43 },
		{ 512, 133, 43 },
		{ 512, 134, 43 },
		{ 512, 135, 43 },
		{ 512, 136, 43 },
		{ 512, 137, 43 },
		{ 512, 138, 43 },
		{ 512, 139, 43 },
		{ 512, 140, 43 },
		{ 512, 141, 43 },
		{ 512, 142, 43 },
		{ 512, 143, 43 },
		{ 512, 144, 43 },
		{ 512, 145, 43 },
		{ 512, 146, 43 },
		{ 512, 147, 43 },
		{ 512, 148, 43 },
		{ 512, 149, 43 },
		{ 512, 150, 43 },
		{ 512, 151, 43 },
		{ 512, 152, 43 },
		{ 512, 153, 43 },
		{ 512, 154, 43 },
		{ 512, 155, 43 },
		{ 512, 156, 43 },
		{ 512, 157, 43 },
		{ 512, 158, 43 },
		{ 512, 159, 43 },
		{ 512, 160, 43 },
		{ 512, 161, 43 },
		{ 512, 162, 43 },
		{ 512, 163, 43 },
		{ 512, 164, 43 },
		{ 512, 165, 43 },
		{ 512, 166, 43 },
		{ 512, 167, 43 },
		{ 512, 168, 43 },
		{ 512, 169, 43 },
		{ 512, 170, 43 },
		{ 512, 171, 43 },
		{ 512, 172, 43 },
		{ 512, 173, 43 },
		{ 512, 174, 43 },
		{ 512, 175, 43 },
		{ 512, 176, 43 },
		{ 512, 177, 43 },
		{ 512, 178, 43 },
		{ 512, 179, 43 },
		{ 512, 180, 43 },
		{ 512, 181, 43 },
		{ 512, 182, 43 },
		{ 512, 183, 43 },
		{ 512, 184, 43 },
		{ 512, 185, 43 },
		{ 512, 186, 43 },
		{ 512, 187, 43 },
		{ 512, 188, 43 },
		{ 512, 189, 43 },
		{ 512, 190, 43 },
		{ 512, 191, 43 },
		{ 512, 192, 43 },
		{ 512, 193, 43 },
		{ 1, 290, 44 },
		{ 1, 291, 44 },
		{ 1, 292, 44 },
		{ 1, 293, 44 },
		{ 1, 294, 44 },
		{ 1, 295, 44 },
		{ 1, 296, 44 },
		{ 1, 297, 44 },
		{ 1, 298, 44 },
		{ 1, 299, 44 },
		{ 1, 300, 44 },
		{ 1, 301, 44 },
		{ 1, 302, 44 },
		{ 1, 303, 44 },
		{ 1, 304, 44 },
		{ 1, 305, 44 },
		{ 1, 306, 44 },
		{ 1, 307, 44 },
		{ 1, 308, 44 },
		{ 1, 309, 44 },
		{ 1, 310, 44 },
		{ 1, 311, 44 },
		{ 1, 312, 44 },
		{ 1, 313, 44 },
		{ 1, 314, 44 },
		{ 1, 315, 44 },
		{ 1, 316, 44 },
		{ 1, 317, 44 },
		{ 1, 318, 44 },
		{ 1, 319, 44 },
		{ 1, 320, 44 },
		{ 1, 321, 44 },
		{ 1, 322, 44 },
		{ 1, 323, 44 },
		{ 1, 324, 44 },
		{ 1, 325, 44 },
		{ 1, 326, 44 },
		{ 1, 327, 44 },
		{ 1, 328, 44 },
		{ 1, 329, 44 },
		{ 1, 330, 44 },
		{ 1, 331, 44 },
		{ 1, 332, 44 },
		{ 1, 333, 44 },
		{ 1, 334, 44 },
		{ 1, 335, 44 },
		{ 1, 336, 44 },
		{ 1, 337, 44 },
		{ 1, 338, 44 },
		{ 1, 339, 44 },
		{ 1, 340, 44 },
		{ 1, 341, 44 },
		{ 1, 342, 44 },
		{ 1, 343, 44 },
		{ 1, 344, 44 },
		{ 1, 345, 44 },
		{ 1, 346, 44 },
		{ 1, 347, 44 },
		{ 1, 348, 44 },
		{ 1, 349, 44 },
		{ 1, 350, 44 },
		{ 1, 351, 44 },
		{ 1, 352, 44 },
		{ 2, 289, 44 },
		{ 2, 290, 44 },
		{ 2, 291, 44 },
		{ 2, 292, 44 },
		{ 2, 293, 44 },
		{ 2, 294, 44 },
		{ 2, 295, 44 },
		{ 2, 296, 44 },
		{ 2, 297, 44 },
		{ 2, 298, 44 },
		{ 2, 299, 44 },
		{ 2, 300, 44 },
		{ 2, 301, 44 },
		{ 2, 302, 44 },
		{ 2, 303, 44 },
		{ 2, 304, 44 },
		{ 2, 305, 44 },
		{ 2, 306, 44 },
		{ 2, 307, 44 },
		{ 2, 308, 44 },
		{ 2, 309, 44 },
		{ 2, 310, 44 },
		{ 2, 311, 44 },
		{ 2, 312, 44 },
		{ 2, 313, 44 },
		{ 2, 314, 44 },
		{ 2, 315, 44 },
		{ 2, 316, 44 },
		{ 2, 317, 44 },
		{ 2, 318, 44 },
		{ 2, 319, 44 },
		{ 2, 320, 44 },
		{ 2, 321, 44 },
		{ 2, 322, 44 },
		{ 2, 323, 44 },
		{ 2, 324, 44 },
		{ 2, 325, 44 },
		{ 2, 326, 44 },
		{ 2, 327, 44 },
		{ 2, 328, 44 },
		{ 2, 329, 44 },
		{ 2, 330, 44 },
		{ 2, 331, 44 },
		{ 2, 332, 44 },
		{ 2, 333, 44 },
		{ 2, 334, 44 },
		{ 2, 335, 44 },
		{ 2, 336, 44 },
		{ 2, 337, 44 },
		{ 2, 338, 44 },
		{ 2, 339, 44 },
		{ 2, 340, 44 },
		{ 2, 341, 44 },
		{ 2, 342, 44 },
		{ 2, 343, 44 },
		{ 2, 344, 44 },
		{ 2, 345, 44 },
		{ 2, 346, 44 },
		{ 2, 347, 44 },
		{ 2, 348, 44 },
		{ 2, 349, 44 },
		{ 2, 350, 44 },
		{ 2, 351, 44 },
		{ 2, 352, 44 },
		{ 3, 289, 44 },
		{ 3, 290, 44 },
		{ 3, 291, 44 },
		{ 3, 292, 44 },
		{ 3, 293, 44 },
		{ 3, 294, 44 },
		{ 3, 295, 44 },
		{ 3, 296, 44 },
		{ 3, 297, 44 },
		{ 3, 298, 44 },
		{ 3, 299, 44 },
		{ 3, 300, 44 },
		{ 3, 301, 44 },
		{ 3, 302, 44 },
		{ 3, 303, 44 },
		{ 3, 304, 44 },
		{ 3, 305, 44 },
		{ 3, 306, 44 },
		{ 3, 307, 44 },
		{ 3, 308, 44 },
		{ 3, 309, 44 },
		{ 3, 310, 44 },
		{ 3, 311, 44 },
		{ 3, 312, 44 },
		{ 3, 313, 44 },
		{ 3, 314, 44 },
		{ 3, 315, 44 },
		{ 3, 316, 44 },
		{ 3, 317, 44 },
		{ 3, 318, 44 },
		{ 3, 319, 44 },
		{ 3, 320, 44 },
		{ 3, 321, 44 },
		{ 3, 322, 44 },
		{ 3, 323, 44 },
		{ 3, 324, 44 },
		{ 3, 325, 44 },
		{ 3, 326, 44 },
		{ 3, 327, 44 },
		{ 3, 328, 44 },
		{ 3, 329, 44 },
		{ 3, 330, 44 },
		{ 3, 331, 44 },
		{ 3, 332, 44 },
		{ 3, 333, 44 },
		{ 3, 334, 44 },
		{ 3, 335, 44 },
		{ 3, 336, 44 },
		{ 3, 337, 44 },
		{ 3, 338, 44 },
		{ 3, 339, 44 },
		{ 3, 340, 44 },
		{ 3, 341, 44 },
		{ 3, 342, 44 },
		{ 3, 343, 44 },
		{ 3, 344, 44 },
		{ 3, 345, 44 },
		{ 3, 346, 44 },
		{ 3, 347, 44 },
		{ 3, 348, 44 },
		{ 3, 349, 44 },
		{ 3, 350, 44 },
		{ 3, 351, 44 },
		{ 3, 352, 44 },
		{ 4, 289, 44 },
		{ 4, 290, 44 },
		{ 4, 291, 44 },
		{ 4, 292, 44 },
		{ 4, 293, 44 },
		{ 4, 294, 44 },
		{ 4, 295, 44 },
		{ 4, 296, 44 },
		{ 4, 297, 44 },
		{ 4, 298, 44 },
		{ 4, 299, 44 },
		{ 4, 300, 44 },
		{ 4, 301, 44 },
		{ 4, 302, 44 },
		{ 4, 303, 44 },
		{ 4, 304, 44 },
		{ 4, 305, 44 },
		{ 4, 306, 44 },
		{ 4, 307, 44 },
		{ 4, 308, 44 },
		{ 4, 309, 44 },
		{ 4, 310, 44 },
		{ 4, 311, 44 },
		{ 4, 312, 44 },
		{ 4, 313, 44 },
		{ 4, 314, 44 },
		{ 4, 315, 44 },
		{ 4, 316, 44 },
		{ 4, 317, 44 },
		{ 4, 318, 44 },
		{ 4, 319, 44 },
		{ 4, 320, 44 },
		{ 4, 321, 44 },
		{ 4, 322, 44 },
		{ 4, 323, 44 },
		{ 4, 324, 44 },
		{ 4, 325, 44 },
		{ 4, 326, 44 },
		{ 4, 327, 44 },
		{ 4, 328, 44 },
		{ 4, 329, 44 },
		{ 4, 330, 44 },
		{ 4, 331, 44 },
		{ 4, 332, 44 },
		{ 4, 333, 44 },
		{ 4, 334, 44 },
		{ 4, 335, 44 },
		{ 4, 336, 44 },
		{ 4, 337, 44 },
		{ 4, 338, 44 },
		{ 4, 339, 44 },
		{ 4, 340, 44 },
		{ 4, 341, 44 },
		{ 4, 342, 44 },
		{ 4, 343, 44 },
		{ 4, 344, 44 },
		{ 4, 345, 44 },
		{ 4, 346, 44 },
		{ 4, 347, 44 },
		{ 4, 348, 44 },
		{ 4, 349, 44 },
		{ 4, 350, 44 },
		{ 4, 351, 44 },
		{ 5, 289, 44 },
		{ 5, 290, 44 },
		{ 5, 291, 44 },
		{ 5, 292, 44 },
		{ 5, 293, 44 },
		{ 5, 294, 44 },
		{ 5, 295, 44 },
		{ 5, 296, 44 },
		{ 5, 297, 44 },
		{ 5, 298, 44 },
		{ 5, 299, 44 },
		{ 5, 300, 44 },
		{ 5, 301, 44 },
		{ 5, 302, 44 },
		{ 5, 303, 44 },
		{ 5, 304, 44 },
		{ 5, 305, 44 },
		{ 5, 306, 44 },
		{ 5, 307, 44 },
		{ 5, 308, 44 },
		{ 5, 309, 44 },
		{ 5, 310, 44 },
		{ 5, 311, 44 },
		{ 5, 312, 44 },
		{ 5, 313, 44 },
		{ 5, 314, 44 },
		{ 5, 315, 44 },
		{ 5, 316, 44 },
		{ 5, 317, 44 },
		{ 5, 318, 44 },
		{ 5, 319, 44 },
		{ 5, 320, 44 },
		{ 5, 321, 44 },
		{ 5, 322, 44 },
		{ 5, 323, 44 },
		{ 5, 324, 44 },
		{ 5, 325, 44 },
		{ 5, 326, 44 },
		{ 5, 327, 44 },
		{ 5, 328, 44 },
		{ 5, 329, 44 },
		{ 5, 330, 44 },
		{ 5, 331, 44 },
		{ 5, 332, 44 },
		{ 5, 333, 44 },
		{ 5, 334, 44 },
		{ 5, 335, 44 },
		{ 5, 336, 44 },
		{ 5, 337, 44 },
		{ 5, 338, 44 },
		{ 5, 339, 44 },
		{ 5, 340, 44 },
		{ 5, 341, 44 },
		{ 5, 342, 44 },
		{ 5, 343, 44 },
		{ 5, 344, 44 },
		{ 5, 345, 44 },
		{ 5, 346, 44 },
		{ 5, 347, 44 },
		{ 5, 348, 44 },
		{ 5, 349, 44 },
		{ 5, 350, 44 },
		{ 5, 351, 44 },
		{ 6, 289, 44 },
		{ 6, 290, 44 },
		{ 6, 291, 44 },
		{ 6, 292, 44 },
		{ 6, 293, 44 },
		{ 6, 294, 44 },
		{ 6, 295, 44 },
		{ 6, 296, 44 },
		{ 6, 297, 44 },
		{ 6, 298, 44 },
		{ 6, 299, 44 },
		{ 6, 300, 44 },
		{ 6, 301, 44 },
		{ 6, 302, 44 },
		{ 6, 303, 44 },
		{ 6, 304, 44 },
		{ 6, 305, 44 },
		{ 6, 306, 44 },
		{ 6, 307, 44 },
		{ 6, 308, 44 },
		{ 6, 309, 44 },
		{ 6, 310, 44 },
		{ 6, 311, 44 },
		{ 6, 312, 44 },
		{ 6, 313, 44 },
		{ 6, 314, 44 },
		{ 6, 315, 44 },
		{ 6, 316, 44 },
		{ 6, 317, 44 },
		{ 6, 318, 44 },
		{ 6, 319, 44 },
		{ 6, 320, 44 },
		{ 6, 321, 44 },
		{ 6, 322, 44 },
		{ 6, 323, 44 },
		{ 6, 324, 44 },
		{ 6, 325, 44 },
		{ 6, 326, 44 },
		{ 6, 327, 44 },
		{ 6, 328, 44 },
		{ 6, 329, 44 },
		{ 6, 330, 44 },
		{ 6, 331, 44 },
		{ 6, 332, 44 },
		{ 6, 333, 44 },
		{ 6, 334, 44 },
		{ 6, 335, 44 },
		{ 6, 336, 44 },
		{ 6, 337, 44 },
		{ 6, 338, 44 },
		{ 6, 339, 44 },
		{ 6, 340, 44 },
		{ 6, 341, 44 },
		{ 6, 342, 44 },
		{ 6, 343, 44 },
		{ 6, 344, 44 },
		{ 6, 345, 44 },
		{ 6, 346, 44 },
		{ 6, 347, 44 },
		{ 6, 348, 44 },
		{ 6, 349, 44 },
		{ 6, 350, 44 },
		{ 6, 351, 44 },
		{ 7, 289, 44 },
		{ 7, 290, 44 },
		{ 7, 291, 44 },
		{ 7, 292, 44 },
		{ 7, 293, 44 },
		{ 7, 294, 44 },
		{ 7, 295, 44 },
		{ 7, 296, 44 },
		{ 7, 297, 44 },
		{ 7, 298, 44 },
		{ 7, 299, 44 },
		{ 7, 300, 44 },
		{ 7, 301, 44 },
		{ 7, 302, 44 },
		{ 7, 303, 44 },
		{ 7, 304, 44 },
		{ 7, 305, 44 },
		{ 7, 306, 44 },
		{ 7, 307, 44 },
		{ 7, 308, 44 },
		{ 7, 309, 44 },
		{ 7, 310, 44 },
		{ 7, 311, 44 },
		{ 7, 312, 44 },
		{ 7, 313, 44 },
		{ 7, 314, 44 },
		{ 7, 315, 44 },
		{ 7, 316, 44 },
		{ 7, 317, 44 },
		{ 7, 318, 44 },
		{ 7, 319, 44 },
		{ 7, 320, 44 },
		{ 7, 321, 44 },
		{ 7, 322, 44 },
		{ 7, 323, 44 },
		{ 7, 324, 44 },
		{ 7, 325, 44 },
		{ 7, 326, 44 },
		{ 7, 327, 44 },
		{ 7, 328, 44 },
		{ 7, 329, 44 },
		{ 7, 330, 44 },
		{ 7, 331, 44 },
		{ 7, 332, 44 },
		{ 7, 333, 44 },
		{ 7, 334, 44 },
		{ 7, 335, 44 },
		{ 7, 336, 44 },
		{ 7, 337, 44 },
		{ 7, 338, 44 },
		{ 7, 339, 44 },
		{ 7, 340, 44 },
		{ 7, 341, 44 },
		{ 7, 342, 44 },
		{ 7, 343, 44 },
		{ 7, 344, 44 },
		{ 7, 345, 44 },
		{ 7, 346, 44 },
		{ 7, 347, 44 },
		{ 7, 348, 44 },
		{ 7, 349, 44 },
		{ 7, 350, 44 },
		{ 8, 289, 44 },
		{ 8, 290, 44 },
		{ 8, 291, 44 },
		{ 8, 292, 44 },
		{ 8, 293, 44 },
		{ 8, 294, 44 },
		{ 8, 295, 44 },
		{ 8, 296, 44 },
		{ 8, 297, 44 },
		{ 8, 298, 44 },
		{ 8, 299, 44 },
		{ 8, 300, 44 },
		{ 8, 301, 44 },
		{ 8, 302, 44 },
		{ 8, 303, 44 },
		{ 8, 304, 44 },
		{ 8, 305, 44 },
		{ 8, 306, 44 },
		{ 8, 307, 44 },
		{ 8, 308, 44 },
		{ 8, 309, 44 },
		{ 8, 310, 44 },
		{ 8, 311, 44 },
		{ 8, 312, 44 },
		{ 8, 313, 44 },
		{ 8, 314, 44 },
		{ 8, 315, 44 },
		{ 8, 316, 44 },
		{ 8, 317, 44 },
		{ 8, 318, 44 },
		{ 8, 319, 44 },
		{ 8, 320, 44 },
		{ 8, 321, 44 },
		{ 8, 322, 44 },
		{ 8, 323, 44 },
		{ 8, 324, 44 },
		{ 8, 325, 44 },
		{ 8, 326, 44 },
		{ 8, 327, 44 },
		{ 8, 328, 44 },
		{ 8, 329, 44 },
		{ 8, 330, 44 },
		{ 8, 331, 44 },
		{ 8, 332, 44 },
		{ 8, 333, 44 },
		{ 8, 334, 44 },
		{ 8, 335, 44 },
		{ 8, 336, 44 },
		{ 8, 337, 44 },
		{ 8, 338, 44 },
		{ 8, 339, 44 },
		{ 8, 340, 44 },
		{ 8, 341, 44 },
		{ 8, 342, 44 },
		{ 8, 343, 44 },
		{ 8, 344, 44 },
		{ 8, 345, 44 },
		{ 8, 346, 44 },
		{ 8, 347, 44 },
		{ 8, 348, 44 },
		{ 8, 349, 44 },
		{ 8, 350, 44 },
		{ 9, 289, 44 },
		{ 9, 290, 44 },
		{ 9, 291, 44 },
		{ 9, 292, 44 },
		{ 9, 293, 44 },
		{ 9, 294, 44 },
		{ 9, 295, 44 },
		{ 9, 296, 44 },
		{ 9, 297, 44 },
		{ 9, 298, 44 },
		{ 9, 299, 44 },
		{ 9, 300, 44 },
		{ 9, 301, 44 },
		{ 9, 302, 44 },
		{ 9, 303, 44 },
		{ 9, 304, 44 },
		{ 9, 305, 44 },
		{ 9, 306, 44 },
		{ 9, 307, 44 },
		{ 9, 308, 44 },
		{ 9, 309, 44 },
		{ 9, 310, 44 },
		{ 9, 311, 44 },
		{ 9, 312, 44 },
		{ 9, 313, 44 },
		{ 9, 314, 44 },
		{ 9, 315, 44 },
		{ 9, 316, 44 },
		{ 9, 317, 44 },
		{ 9, 318, 44 },
		{ 9, 319, 44 },
		{ 9, 320, 44 },
		{ 9, 321, 44 },
		{ 9, 322, 44 },
		{ 9, 323, 44 },
		{ 9, 324, 44 },
		{ 9, 325, 44 },
		{ 9, 326, 44 },
		{ 9, 327, 44 },
		{ 9, 328, 44 },
		{ 9, 329, 44 },
		{ 9, 330, 44 },
		{ 9, 331, 44 },
		{ 9, 332, 44 },
		{ 9, 333, 44 },
		{ 9, 334, 44 },
		{ 9, 335, 44 },
		{ 9, 336, 44 },
		{ 9, 337, 44 },
		{ 9, 338, 44 },
		{ 9, 339, 44 },
		{ 9, 340, 44 },
		{ 9, 341, 44 },
		{ 9, 342, 44 },
		{ 9, 343, 44 },
		{ 9, 344, 44 },
		{ 9, 345, 44 },
		{ 9, 346, 44 },
		{ 9, 347, 44 },
		{ 9, 348, 44 },
		{ 9, 349, 44 },
		{ 10, 288, 44 },
		{ 10, 289, 44 },
		{ 10, 290, 44 },
		{ 10, 291, 44 },
		{ 10, 292, 44 },
		{ 10, 293, 44 },
		{ 10, 294, 44 },
		{ 10, 295, 44 },
		{ 10, 296, 44 },
		{ 10, 297, 44 },
		{ 10, 298, 44 },
		{ 10, 299, 44 },
		{ 10, 300, 44 },
		{ 10, 301, 44 },
		{ 10, 302, 44 },
		{ 10, 303, 44 },
		{ 10, 304, 44 },
		{ 10, 305, 44 },
		{ 10, 306, 44 },
		{ 10, 307, 44 },
		{ 10, 308, 44 },
		{ 10, 309, 44 },
		{ 10, 310, 44 },
		{ 10, 311, 44 },
		{ 10, 312, 44 },
		{ 10, 313, 44 },
		{ 10, 314, 44 },
		{ 10, 315, 44 },
		{ 10, 316, 44 },
		{ 10, 317, 44 },
		{ 10, 318, 44 },
		{ 10, 319, 44 },
		{ 10, 320, 44 },
		{ 10, 321, 44 },
		{ 10, 322, 44 },
		{ 10, 323, 44 },
		{ 10, 324, 44 },
		{ 10, 325, 44 },
		{ 10, 326, 44 },
		{ 10, 327, 44 },
		{ 10, 328, 44 },
		{ 10, 329, 44 },
		{ 10, 330, 44 },
		{ 10, 331, 44 },
		{ 10, 332, 44 },
		{ 10, 333, 44 },
		{ 10, 334, 44 },
		{ 10, 335, 44 },
		{ 10, 336, 44 },
		{ 10, 337, 44 },
		{ 10, 338, 44 },
		{ 10, 339, 44 },
		{ 10, 340, 44 },
		{ 10, 341, 44 },
		{ 10, 342, 44 },
		{ 10, 343, 44 },
		{ 10, 344, 44 },
		{ 10, 345, 44 },
		{ 10, 346, 44 },
		{ 10, 347, 44 },
		{ 10, 348, 44 },
		{ 10, 349, 44 },
		{ 11, 288, 44 },
		{ 11, 289, 44 },
		{ 11, 290, 44 },
		{ 11, 291, 44 },
		{ 11, 292, 44 },
		{ 11, 293, 44 },
		{ 11, 294, 44 },
		{ 11, 295, 44 },
		{ 11, 296, 44 },
		{ 11, 297, 44 },
		{ 11, 298, 44 },
		{ 11, 299, 44 },
		{ 11, 300, 44 },
		{ 11, 301, 44 },
		{ 11, 302, 44 },
		{ 11, 303, 44 },
		{ 11, 304, 44 },
		{ 11, 305, 44 },
		{ 11, 306, 44 },
		{ 11, 307, 44 },
		{ 11, 308, 44 },
		{ 11, 309, 44 },
		{ 11, 310, 44 },
		{ 11, 311, 44 },
		{ 11, 312, 44 },
		{ 11, 313, 44 },
		{ 11, 314, 44 },
		{ 11, 315, 44 },
		{ 11, 316, 44 },
		{ 11, 317, 44 },
		{ 11, 318, 44 },
		{ 11, 319, 44 },
		{ 11, 320, 44 },
		{ 11, 321, 44 },
		{ 11, 322, 44 },
		{ 11, 323, 44 },
		{ 11, 324, 44 },
		{ 11, 325, 44 },
		{ 11, 326, 44 },
		{ 11, 327, 44 },
		{ 11, 328, 44 },
		{ 11, 329, 44 },
		{ 11, 330, 44 },
		{ 11, 331, 44 },
		{ 11, 332, 44 },
		{ 11, 333, 44 },
		{ 11, 334, 44 },
		{ 11, 335, 44 },
		{ 11, 336, 44 },
		{ 11, 337, 44 },
		{ 11, 338, 44 },
		{ 11, 339, 44 },
		{ 11, 340, 44 },
		{ 11, 341, 44 },
		{ 11, 342, 44 },
		{ 11, 343, 44 },
		{ 11, 344, 44 },
		{ 11, 345, 44 },
		{ 11, 346, 44 },
		{ 11, 347, 44 },
		{ 11, 348, 44 },
		{ 11, 349, 44 },
		{ 12, 288, 44 },
		{ 12, 289, 44 },
		{ 12, 290, 44 },
		{ 12, 291, 44 },
		{ 12, 292, 44 },
		{ 12, 293, 44 },
		{ 12, 294, 44 },
		{ 12, 295, 44 },
		{ 12, 296, 44 },
		{ 12, 297, 44 },
		{ 12, 298, 44 },
		{ 12, 299, 44 },
		{ 12, 300, 44 },
		{ 12, 301, 44 },
		{ 12, 302, 44 },
		{ 12, 303, 44 },
		{ 12, 304, 44 },
		{ 12, 305, 44 },
		{ 12, 306, 44 },
		{ 12, 307, 44 },
		{ 12, 308, 44 },
		{ 12, 309, 44 },
		{ 12, 310, 44 },
		{ 12, 311, 44 },
		{ 12, 312, 44 },
		{ 12, 313, 44 },
		{ 12, 314, 44 },
		{ 12, 315, 44 },
		{ 12, 316, 44 },
		{ 12, 317, 44 },
		{ 12, 318, 44 },
		{ 12, 319, 44 },
		{ 12, 320, 44 },
		{ 12, 321, 44 },
		{ 12, 322, 44 },
		{ 12, 323, 44 },
		{ 12, 324, 44 },
		{ 12, 325, 44 },
		{ 12, 326, 44 },
		{ 12, 327, 44 },
		{ 12, 328, 44 },
		{ 12, 329, 44 },
		{ 12, 330, 44 },
		{ 12, 331, 44 },
		{ 12, 332, 44 },
		{ 12, 333, 44 },
		{ 12, 334, 44 },
		{ 12, 335, 44 },
		{ 12, 336, 44 },
		{ 12, 337, 44 },
		{ 12, 338, 44 },
		{ 12, 339, 44 },
		{ 12, 340, 44 },
		{ 12, 341, 44 },
		{ 12, 342, 44 },
		{ 12, 343, 44 },
		{ 12, 344, 44 },
		{ 12, 345, 44 },
		{ 12, 346, 44 },
		{ 12, 347, 44 },
		{ 12, 348, 44 },
		{ 13, 288, 44 },
		{ 13, 289, 44 },
		{ 13, 290, 44 },
		{ 13, 291, 44 },
		{ 13, 292, 44 },
		{ 13, 293, 44 },
		{ 13, 294, 44 },
		{ 13, 295, 44 },
		{ 13, 296, 44 },
		{ 13, 297, 44 },
		{ 13, 298, 44 },
		{ 13, 299, 44 },
		{ 13, 300, 44 },
		{ 13, 301, 44 },
		{ 13, 302, 44 },
		{ 13, 303, 44 },
		{ 13, 304, 44 },
		{ 13, 305, 44 },
		{ 13, 306, 44 },
		{ 13, 307, 44 },
		{ 13, 308, 44 },
		{ 13, 309, 44 },
		{ 13, 310, 44 },
		{ 13, 311, 44 },
		{ 13, 312, 44 },
		{ 13, 313, 44 },
		{ 13, 314, 44 },
		{ 13, 315, 44 },
		{ 13, 316, 44 },
		{ 13, 317, 44 },
		{ 13, 318, 44 },
		{ 13, 319, 44 },
		{ 13, 320, 44 },
		{ 13, 321, 44 },
		{ 13, 322, 44 },
		{ 13, 323, 44 },
		{ 13, 324, 44 },
		{ 13, 325, 44 },
		{ 13, 326, 44 },
		{ 13, 327, 44 },
		{ 13, 328, 44 },
		{ 13, 329, 44 },
		{ 13, 330, 44 },
		{ 13, 331, 44 },
		{ 13, 332, 44 },
		{ 13, 333, 44 },
		{ 13, 334, 44 },
		{ 13, 335, 44 },
		{ 13, 336, 44 },
		{ 13, 337, 44 },
		{ 13, 338, 44 },
		{ 13, 339, 44 },
		{ 13, 340, 44 },
		{ 13, 341, 44 },
		{ 13, 342, 44 },
		{ 13, 343, 44 },
		{ 13, 344, 44 },
		{ 13, 345, 44 },
		{ 13, 346, 44 },
		{ 13, 347, 44 },
		{ 13, 348, 44 },
		{ 14, 288, 44 },
		{ 14, 289, 44 },
		{ 14, 290, 44 },
		{ 14, 291, 44 },
		{ 14, 292, 44 },
		{ 14, 293, 44 },
		{ 14, 294, 44 },
		{ 14, 295, 44 },
		{ 14, 296, 44 },
		{ 14, 297, 44 },
		{ 14, 298, 44 },
		{ 14, 299, 44 },
		{ 14, 300, 44 },
		{ 14, 301, 44 },
		{ 14, 302, 44 },
		{ 14, 303, 44 },
		{ 14, 304, 44 },
		{ 14, 305, 44 },
		{ 14, 306, 44 },
		{ 14, 307, 44 },
		{ 14, 308, 44 },
		{ 14, 309, 44 },
		{ 14, 310, 44 },
		{ 14, 311, 44 },
		{ 14, 312, 44 },
		{ 14, 313, 44 },
		{ 14, 314, 44 },
		{ 14, 315, 44 },
		{ 14, 316, 44 },
		{ 14, 317, 44 },
		{ 14, 318, 44 },
		{ 14, 319, 44 },
		{ 14, 320, 44 },
		{ 14, 321, 44 },
		{ 14, 322, 44 },
		{ 14, 323, 44 },
		{ 14, 324, 44 },
		{ 14, 325, 44 },
		{ 14, 326, 44 },
		{ 14, 327, 44 },
		{ 14, 328, 44 },
		{ 14, 329, 44 },
		{ 14, 330, 44 },
		{ 14, 331, 44 },
		{ 14, 332, 44 },
		{ 14, 333, 44 },
		{ 14, 334, 44 },
		{ 14, 335, 44 },
		{ 14, 336, 44 },
		{ 14, 337, 44 },
		{ 14, 338, 44 },
		{ 14, 339, 44 },
		{ 14, 340, 44 },
		{ 14, 341, 44 },
		{ 14, 342, 44 },
		{ 14, 343, 44 },
		{ 14, 344, 44 },
		{ 14, 345, 44 },
		{ 14, 346, 44 },
		{ 14, 347, 44 },
		{ 14, 348, 44 },
		{ 15, 288, 44 },
		{ 15, 289, 44 },
		{ 15, 290, 44 },
		{ 15, 291, 44 },
		{ 15, 292, 44 },
		{ 15, 293, 44 },
		{ 15, 294, 44 },
		{ 15, 295, 44 },
		{ 15, 296, 44 },
		{ 15, 297, 44 },
		{ 15, 298, 44 },
		{ 15, 299, 44 },
		{ 15, 300, 44 },
		{ 15, 301, 44 },
		{ 15, 302, 44 },
		{ 15, 303, 44 },
		{ 15, 304, 44 },
		{ 15, 305, 44 },
		{ 15, 306, 44 },
		{ 15, 307, 44 },
		{ 15, 308, 44 },
		{ 15, 309, 44 },
		{ 15, 310, 44 },
		{ 15, 311, 44 },
		{ 15, 312, 44 },
		{ 15, 313, 44 },
		{ 15, 314, 44 },
		{ 15, 315, 44 },
		{ 15, 316, 44 },
		{ 15, 317, 44 },
		{ 15, 318, 44 },
		{ 15, 319, 44 },
		{ 15, 320, 44 },
		{ 15, 321, 44 },
		{ 15, 322, 44 },
		{ 15, 323, 44 },
		{ 15, 324, 44 },
		{ 15, 325, 44 },
		{ 15, 326, 44 },
		{ 15, 327, 44 },
		{ 15, 328, 44 },
		{ 15, 329, 44 },
		{ 15, 330, 44 },
		{ 15, 331, 44 },
		{ 15, 332, 44 },
		{ 15, 333, 44 },
		{ 15, 334, 44 },
		{ 15, 335, 44 },
		{ 15, 336, 44 },
		{ 15, 337, 44 },
		{ 15, 338, 44 },
		{ 15, 339, 44 },
		{ 15, 340, 44 },
		{ 15, 341, 44 },
		{ 15, 342, 44 },
		{ 15, 343, 44 },
		{ 15, 344, 44 },
		{ 15, 345, 44 },
		{ 15, 346, 44 },
		{ 15, 347, 44 },
		{ 16, 288, 44 },
		{ 16, 289, 44 },
		{ 16, 290, 44 },
		{ 16, 291, 44 },
		{ 16, 292, 44 },
		{ 16, 293, 44 },
		{ 16, 294, 44 },
		{ 16, 295, 44 },
		{ 16, 296, 44 },
		{ 16, 297, 44 },
		{ 16, 298, 44 },
		{ 16, 299, 44 },
		{ 16, 300, 44 },
		{ 16, 301, 44 },
		{ 16, 302, 44 },
		{ 16, 303, 44 },
		{ 16, 304, 44 },
		{ 16, 305, 44 },
		{ 16, 306, 44 },
		{ 16, 307, 44 },
		{ 16, 308, 44 },
		{ 16, 309, 44 },
		{ 16, 310, 44 },
		{ 16, 311, 44 },
		{ 16, 312, 44 },
		{ 16, 313, 44 },
		{ 16, 314, 44 },
		{ 16, 315, 44 },
		{ 16, 316, 44 },
		{ 16, 317, 44 },
		{ 16, 318, 44 },
		{ 16, 319, 44 },
		{ 16, 320, 44 },
		{ 16, 321, 44 },
		{ 16, 322, 44 },
		{ 16, 323, 44 },
		{ 16, 324, 44 },
		{ 16, 325, 44 },
		{ 16, 326, 44 },
		{ 16, 327, 44 },
		{ 16, 328, 44 },
		{ 16, 329, 44 },
		{ 16, 330, 44 },
		{ 16, 331, 44 },
		{ 16, 332, 44 },
		{ 16, 333, 44 },
		{ 16, 334, 44 },
		{ 16, 335, 44 },
		{ 16, 336, 44 },
		{ 16, 337, 44 },
		{ 16, 338, 44 },
		{ 16, 339, 44 },
		{ 16, 340, 44 },
		{ 16, 341, 44 },
		{ 16, 342, 44 },
		{ 16, 343, 44 },
		{ 16, 344, 44 },
		{ 16, 345, 44 },
		{ 16, 346, 44 },
		{ 16, 347, 44 },
		{ 17, 288, 44 },
		{ 17, 289, 44 },
		{ 17, 290, 44 },
		{ 17, 291, 44 },
		{ 17, 292, 44 },
		{ 17, 293, 44 },
		{ 17, 294, 44 },
		{ 17, 295, 44 },
		{ 17, 296, 44 },
		{ 17, 297, 44 },
		{ 17, 298, 44 },
		{ 17, 299, 44 },
		{ 17, 300, 44 },
		{ 17, 301, 44 },
		{ 17, 302, 44 },
		{ 17, 303, 44 },
		{ 17, 304, 44 },
		{ 17, 305, 44 },
		{ 17, 306, 44 },
		{ 17, 307, 44 },
		{ 17, 308, 44 },
		{ 17, 309, 44 },
		{ 17, 310, 44 },
		{ 17, 311, 44 },
		{ 17, 312, 44 },
		{ 17, 313, 44 },
		{ 17, 314, 44 },
		{ 17, 315, 44 },
		{ 17, 316, 44 },
		{ 17, 317, 44 },
		{ 17, 318, 44 },
		{ 17, 319, 44 },
		{ 17, 320, 44 },
		{ 17, 321, 44 },
		{ 17, 322, 44 },
		{ 17, 323, 44 },
		{ 17, 324, 44 },
		{ 17, 325, 44 },
		{ 17, 326, 44 },
		{ 17, 327, 44 },
		{ 17, 328, 44 },
		{ 17, 329, 44 },
		{ 17, 330, 44 },
		{ 17, 331, 44 },
		{ 17, 332, 44 },
		{ 17, 333, 44 },
		{ 17, 334, 44 },
		{ 17, 335, 44 },
		{ 17, 336, 44 },
		{ 17, 337, 44 },
		{ 17, 338, 44 },
		{ 17, 339, 44 },
		{ 17, 340, 44 },
		{ 17, 341, 44 },
		{ 17, 342, 44 },
		{ 17, 343, 44 },
		{ 17, 344, 44 },
		{ 17, 345, 44 },
		{ 17, 346, 44 },
		{ 18, 287, 44 },
		{ 18, 288, 44 },
		{ 18, 289, 44 },
		{ 18, 290, 44 },
		{ 18, 291, 44 },
		{ 18, 292, 44 },
		{ 18, 293, 44 },
		{ 18, 294, 44 },
		{ 18, 295, 44 },
		{ 18, 296, 44 },
		{ 18, 297, 44 },
		{ 18, 298, 44 },
		{ 18, 299, 44 },
		{ 18, 300, 44 },
		{ 18, 301, 44 },
		{ 18, 302, 44 },
		{ 18, 303, 44 },
		{ 18, 304, 44 },
		{ 18, 305, 44 },
		{ 18, 306, 44 },
		{ 18, 307, 44 },
		{ 18, 308, 44 },
		{ 18, 309, 44 },
		{ 18, 310, 44 },
		{ 18, 311, 44 },
		{ 18, 312, 44 },
		{ 18, 313, 44 },
		{ 18, 314, 44 },
		{ 18, 315, 44 },
		{ 18, 316, 44 },
		{ 18, 317, 44 },
		{ 18, 318, 44 },
		{ 18, 319, 44 },
		{ 18, 320, 44 },
		{ 18, 321, 44 },
		{ 18, 322, 44 },
		{ 18, 323, 44 },
		{ 18, 324, 44 },
		{ 18, 325, 44 },
		{ 18, 326, 44 },
		{ 18, 327, 44 },
		{ 18, 328, 44 },
		{ 18, 329, 44 },
		{ 18, 330, 44 },
		{ 18, 331, 44 },
		{ 18, 332, 44 },
		{ 18, 333, 44 },
		{ 18, 334, 44 },
		{ 18, 335, 44 },
		{ 18, 336, 44 },
		{ 18, 337, 44 },
		{ 18, 338, 44 },
		{ 18, 339, 44 },
		{ 18, 340, 44 },
		{ 18, 341, 44 },
		{ 18, 342, 44 },
		{ 18, 343, 44 },
		{ 18, 344, 44 },
		{ 18, 345, 44 },
		{ 18, 346, 44 },
		{ 19, 287, 44 },
		{ 19, 288, 44 },
		{ 19, 289, 44 },
		{ 19, 290, 44 },
		{ 19, 291, 44 },
		{ 19, 292, 44 },
		{ 19, 293, 44 },
		{ 19, 294, 44 },
		{ 19, 295, 44 },
		{ 19, 296, 44 },
		{ 19, 297, 44 },
		{ 19, 298, 44 },
		{ 19, 299, 44 },
		{ 19, 300, 44 },
		{ 19, 301, 44 },
		{ 19, 302, 44 },
		{ 19, 303, 44 },
		{ 19, 304, 44 },
		{ 19, 305, 44 },
		{ 19, 306, 44 },
		{ 19, 307, 44 },
		{ 19, 308, 44 },
		{ 19, 309, 44 },
		{ 19, 310, 44 },
		{ 19, 311, 44 },
		{ 19, 312, 44 },
		{ 19, 313, 44 },
		{ 19, 314, 44 },
		{ 19, 315, 44 },
		{ 19, 316, 44 },
		{ 19, 317, 44 },
		{ 19, 318, 44 },
		{ 19, 319, 44 },
		{ 19, 320, 44 },
		{ 19, 321, 44 },
		{ 19, 322, 44 },
		{ 19, 323, 44 },
		{ 19, 324, 44 },
		{ 19, 325, 44 },
		{ 19, 326, 44 },
		{ 19, 327, 44 },
		{ 19, 328, 44 },
		{ 19, 329, 44 },
		{ 19, 330, 44 },
		{ 19, 331, 44 },
		{ 19, 332, 44 },
		{ 19, 333, 44 },
		{ 19, 334, 44 },
		{ 19, 335, 44 },
		{ 19, 336, 44 },
		{ 19, 337, 44 },
		{ 19, 338, 44 },
		{ 19, 339, 44 },
		{ 19, 340, 44 },
		{ 19, 341, 44 },
		{ 19, 342, 44 },
		{ 19, 343, 44 },
		{ 19, 344, 44 },
		{ 19, 345, 44 },
		{ 19, 346, 44 },
		{ 20, 287, 44 },
		{ 20, 288, 44 },
		{ 20, 289, 44 },
		{ 20, 290, 44 },
		{ 20, 291, 44 },
		{ 20, 292, 44 },
		{ 20, 293, 44 },
		{ 20, 294, 44 },
		{ 20, 295, 44 },
		{ 20, 296, 44 },
		{ 20, 297, 44 },
		{ 20, 298, 44 },
		{ 20, 299, 44 },
		{ 20, 300, 44 },
		{ 20, 301, 44 },
		{ 20, 302, 44 },
		{ 20, 303, 44 },
		{ 20, 304, 44 },
		{ 20, 305, 44 },
		{ 20, 306, 44 },
		{ 20, 307, 44 },
		{ 20, 308, 44 },
		{ 20, 309, 44 },
		{ 20, 310, 44 },
		{ 20, 311, 44 },
		{ 20, 312, 44 },
		{ 20, 313, 44 },
		{ 20, 314, 44 },
		{ 20, 315, 44 },
		{ 20, 316, 44 },
		{ 20, 317, 44 },
		{ 20, 318, 44 },
		{ 20, 319, 44 },
		{ 20, 320, 44 },
		{ 20, 321, 44 },
		{ 20, 322, 44 },
		{ 20, 323, 44 },
		{ 20, 324, 44 },
		{ 20, 325, 44 },
		{ 20, 326, 44 },
		{ 20, 327, 44 },
		{ 20, 328, 44 },
		{ 20, 329, 44 },
		{ 20, 330, 44 },
		{ 20, 331, 44 },
		{ 20, 332, 44 },
		{ 20, 333, 44 },
		{ 20, 334, 44 },
		{ 20, 335, 44 },
		{ 20, 336, 44 },
		{ 20, 337, 44 },
		{ 20, 338, 44 },
		{ 20, 339, 44 },
		{ 20, 340, 44 },
		{ 20, 341, 44 },
		{ 20, 342, 44 },
		{ 20, 343, 44 },
		{ 20, 344, 44 },
		{ 20, 345, 44 },
		{ 21, 287, 44 },
		{ 21, 288, 44 },
		{ 21, 289, 44 },
		{ 21, 290, 44 },
		{ 21, 291, 44 },
		{ 21, 292, 44 },
		{ 21, 293, 44 },
		{ 21, 294, 44 },
		{ 21, 295, 44 },
		{ 21, 296, 44 },
		{ 21, 297, 44 },
		{ 21, 298, 44 },
		{ 21, 299, 44 },
		{ 21, 300, 44 },
		{ 21, 301, 44 },
		{ 21, 302, 44 },
		{ 21, 303, 44 },
		{ 21, 304, 44 },
		{ 21, 305, 44 },
		{ 21, 306, 44 },
		{ 21, 307, 44 },
		{ 21, 308, 44 },
		{ 21, 309, 44 },
		{ 21, 310, 44 },
		{ 21, 311, 44 },
		{ 21, 312, 44 },
		{ 21, 313, 44 },
		{ 21, 314, 44 },
		{ 21, 315, 44 },
		{ 21, 316, 44 },
		{ 21, 317, 44 },
		{ 21, 318, 44 },
		{ 21, 319, 44 },
		{ 21, 320, 44 },
		{ 21, 321, 44 },
		{ 21, 322, 44 },
		{ 21, 323, 44 },
		{ 21, 324, 44 },
		{ 21, 325, 44 },
		{ 21, 326, 44 },
		{ 21, 327, 44 },
		{ 21, 328, 44 },
		{ 21, 329, 44 },
		{ 21, 330, 44 },
		{ 21, 331, 44 },
		{ 21, 332, 44 },
		{ 21, 333, 44 },
		{ 21, 334, 44 },
		{ 21, 335, 44 },
		{ 21, 336, 44 },
		{ 21, 337, 44 },
		{ 21, 338, 44 },
		{ 21, 339, 44 },
		{ 21, 340, 44 },
		{ 21, 341, 44 },
		{ 21, 342, 44 },
		{ 21, 343, 44 },
		{ 21, 344, 44 },
		{ 21, 345, 44 },
		{ 22, 287, 44 },
		{ 22, 288, 44 },
		{ 22, 289, 44 },
		{ 22, 290, 44 },
		{ 22, 291, 44 },
		{ 22, 292, 44 },
		{ 22, 293, 44 },
		{ 22, 294, 44 },
		{ 22, 295, 44 },
		{ 22, 296, 44 },
		{ 22, 297, 44 },
		{ 22, 298, 44 },
		{ 22, 299, 44 },
		{ 22, 300, 44 },
		{ 22, 301, 44 },
		{ 22, 302, 44 },
		{ 22, 303, 44 },
		{ 22, 304, 44 },
		{ 22, 305, 44 },
		{ 22, 306, 44 },
		{ 22, 307, 44 },
		{ 22, 308, 44 },
		{ 22, 309, 44 },
		{ 22, 310, 44 },
		{ 22, 311, 44 },
		{ 22, 312, 44 },
		{ 22, 313, 44 },
		{ 22, 314, 44 },
		{ 22, 315, 44 },
		{ 22, 316, 44 },
		{ 22, 317, 44 },
		{ 22, 318, 44 },
		{ 22, 319, 44 },
		{ 22, 320, 44 },
		{ 22, 321, 44 },
		{ 22, 322, 44 },
		{ 22, 323, 44 },
		{ 22, 324, 44 },
		{ 22, 325, 44 },
		{ 22, 326, 44 },
		{ 22, 327, 44 },
		{ 22, 328, 44 },
		{ 22, 329, 44 },
		{ 22, 330, 44 },
		{ 22, 331, 44 },
		{ 22, 332, 44 },
		{ 22, 333, 44 },
		{ 22, 334, 44 },
		{ 22, 335, 44 },
		{ 22, 336, 44 },
		{ 22, 337, 44 },
		{ 22, 338, 44 },
		{ 22, 339, 44 },
		{ 22, 340, 44 },
		{ 22, 341, 44 },
		{ 22, 342, 44 },
		{ 22, 343, 44 },
		{ 22, 344, 44 },
		{ 22, 345, 44 },
		{ 23, 287, 44 },
		{ 23, 288, 44 },
		{ 23, 289, 44 },
		{ 23, 290, 44 },
		{ 23, 291, 44 },
		{ 23, 292, 44 },
		{ 23, 293, 44 },
		{ 23, 294, 44 },
		{ 23, 295, 44 },
		{ 23, 296, 44 },
		{ 23, 297, 44 },
		{ 23, 298, 44 },
		{ 23, 299, 44 },
		{ 23, 300, 44 },
		{ 23, 301, 44 },
		{ 23, 302, 44 },
		{ 23, 303, 44 },
		{ 23, 304, 44 },
		{ 23, 305, 44 },
		{ 23, 306, 44 },
		{ 23, 307, 44 },
		{ 23, 308, 44 },
		{ 23, 309, 44 },
		{ 23, 310, 44 },
		{ 23, 311, 44 },
		{ 23, 312, 44 },
		{ 23, 313, 44 },
		{ 23, 314, 44 },
		{ 23, 315, 44 },
		{ 23, 316, 44 },
		{ 23, 317, 44 },
		{ 23, 318, 44 },
		{ 23, 319, 44 },
		{ 23, 320, 44 },
		{ 23, 321, 44 },
		{ 23, 322, 44 },
		{ 23, 323, 44 },
		{ 23, 324, 44 },
		{ 23, 325, 44 },
		{ 23, 326, 44 },
		{ 23, 327, 44 },
		{ 23, 328, 44 },
		{ 23, 329, 44 },
		{ 23, 330, 44 },
		{ 23, 331, 44 },
		{ 23, 332, 44 },
		{ 23, 333, 44 },
		{ 23, 334, 44 },
		{ 23, 335, 44 },
		{ 23, 336, 44 },
		{ 23, 337, 44 },
		{ 23, 338, 44 },
		{ 23, 339, 44 },
		{ 23, 340, 44 },
		{ 23, 341, 44 },
		{ 23, 342, 44 },
		{ 23, 343, 44 },
		{ 23, 344, 44 },
		{ 24, 287, 44 },
		{ 24, 288, 44 },
		{ 24, 289, 44 },
		{ 24, 290, 44 },
		{ 24, 291, 44 },
		{ 24, 292, 44 },
		{ 24, 293, 44 },
		{ 24, 294, 44 },
		{ 24, 295, 44 },
		{ 24, 296, 44 },
		{ 24, 297, 44 },
		{ 24, 298, 44 },
		{ 24, 299, 44 },
		{ 24, 300, 44 },
		{ 24, 301, 44 },
		{ 24, 302, 44 },
		{ 24, 303, 44 },
		{ 24, 304, 44 },
		{ 24, 305, 44 },
		{ 24, 306, 44 },
		{ 24, 307, 44 },
		{ 24, 308, 44 },
		{ 24, 309, 44 },
		{ 24, 310, 44 },
		{ 24, 311, 44 },
		{ 24, 312, 44 },
		{ 24, 313, 44 },
		{ 24, 314, 44 },
		{ 24, 315, 44 },
		{ 24, 316, 44 },
		{ 24, 317, 44 },
		{ 24, 318, 44 },
		{ 24, 319, 44 },
		{ 24, 320, 44 },
		{ 24, 321, 44 },
		{ 24, 322, 44 },
		{ 24, 323, 44 },
		{ 24, 324, 44 },
		{ 24, 325, 44 },
		{ 24, 326, 44 },
		{ 24, 327, 44 },
		{ 24, 328, 44 },
		{ 24, 329, 44 },
		{ 24, 330, 44 },
		{ 24, 331, 44 },
		{ 24, 332, 44 },
		{ 24, 333, 44 },
		{ 24, 334, 44 },
		{ 24, 335, 44 },
		{ 24, 336, 44 },
		{ 24, 337, 44 },
		{ 24, 338, 44 },
		{ 24, 339, 44 },
		{ 24, 340, 44 },
		{ 24, 341, 44 },
		{ 24, 342, 44 },
		{ 24, 343, 44 },
		{ 24, 344, 44 },
		{ 25, 287, 44 },
		{ 25, 288, 44 },
		{ 25, 289, 44 },
		{ 25, 290, 44 },
		{ 25, 291, 44 },
		{ 25, 292, 44 },
		{ 25, 293, 44 },
		{ 25, 294, 44 },
		{ 25, 295, 44 },
		{ 25, 296, 44 },
		{ 25, 297, 44 },
		{ 25, 298, 44 },
		{ 25, 299, 44 },
		{ 25, 300, 44 },
		{ 25, 301, 44 },
		{ 25, 302, 44 },
		{ 25, 303, 44 },
		{ 25, 304, 44 },
		{ 25, 305, 44 },
		{ 25, 306, 44 },
		{ 25, 307, 44 },
		{ 25, 308, 44 },
		{ 25, 309, 44 },
		{ 25, 310, 44 },
		{ 25, 311, 44 },
		{ 25, 312, 44 },
		{ 25, 313, 44 },
		{ 25, 314, 44 },
		{ 25, 315, 44 },
		{ 25, 316, 44 },
		{ 25, 317, 44 },
		{ 25, 318, 44 },
		{ 25, 319, 44 },
		{ 25, 320, 44 },
		{ 25, 321, 44 },
		{ 25, 322, 44 },
		{ 25, 323, 44 },
		{ 25, 324, 44 },
		{ 25, 325, 44 },
		{ 25, 326, 44 },
		{ 25, 327, 44 },
		{ 25, 328, 44 },
		{ 25, 329, 44 },
		{ 25, 330, 44 },
		{ 25, 331, 44 },
		{ 25, 332, 44 },
		{ 25, 333, 44 },
		{ 25, 334, 44 },
		{ 25, 335, 44 },
		{ 25, 336, 44 },
		{ 25, 337, 44 },
		{ 25, 338, 44 },
		{ 25, 339, 44 },
		{ 25, 340, 44 },
		{ 25, 341, 44 },
		{ 25, 342, 44 },
		{ 25, 343, 44 },
		{ 26, 286, 44 },
		{ 26, 287, 44 },
		{ 26, 288, 44 },
		{ 26, 289, 44 },
		{ 26, 290, 44 },
		{ 26, 291, 44 },
		{ 26, 292, 44 },
		{ 26, 293, 44 },
		{ 26, 294, 44 },
		{ 26, 295, 44 },
		{ 26, 296, 44 },
		{ 26, 297, 44 },
		{ 26, 298, 44 },
		{ 26, 299, 44 },
		{ 26, 300, 44 },
		{ 26, 301, 44 },
		{ 26, 302, 44 },
		{ 26, 303, 44 },
		{ 26, 304, 44 },
		{ 26, 305, 44 },
		{ 26, 306, 44 },
		{ 26, 307, 44 },
		{ 26, 308, 44 },
		{ 26, 309, 44 },
		{ 26, 310, 44 },
		{ 26, 311, 44 },
		{ 26, 312, 44 },
		{ 26, 313, 44 },
		{ 26, 314, 44 },
		{ 26, 315, 44 },
		{ 26, 316, 44 },
		{ 26, 317, 44 },
		{ 26, 318, 44 },
		{ 26, 319, 44 },
		{ 26, 320, 44 },
		{ 26, 321, 44 },
		{ 26, 322, 44 },
		{ 26, 323, 44 },
		{ 26, 324, 44 },
		{ 26, 325, 44 },
		{ 26, 326, 44 },
		{ 26, 327, 44 },
		{ 26, 328, 44 },
		{ 26, 329, 44 },
		{ 26, 330, 44 },
		{ 26, 331, 44 },
		{ 26, 332, 44 },
		{ 26, 333, 44 },
		{ 26, 334, 44 },
		{ 26, 335, 44 },
		{ 26, 336, 44 },
		{ 26, 337, 44 },
		{ 26, 338, 44 },
		{ 26, 339, 44 },
		{ 26, 340, 44 },
		{ 26, 341, 44 },
		{ 26, 342, 44 },
		{ 26, 343, 44 },
		{ 27, 286, 44 },
		{ 27, 287, 44 },
		{ 27, 288, 44 },
		{ 27, 289, 44 },
		{ 27, 290, 44 },
		{ 27, 291, 44 },
		{ 27, 292, 44 },
		{ 27, 293, 44 },
		{ 27, 294, 44 },
		{ 27, 295, 44 },
		{ 27, 296, 44 },
		{ 27, 297, 44 },
		{ 27, 298, 44 },
		{ 27, 299, 44 },
		{ 27, 300, 44 },
		{ 27, 301, 44 },
		{ 27, 302, 44 },
		{ 27, 303, 44 },
		{ 27, 304, 44 },
		{ 27, 305, 44 },
		{ 27, 306, 44 },
		{ 27, 307, 44 },
		{ 27, 308, 44 },
		{ 27, 309, 44 },
		{ 27, 310, 44 },
		{ 27, 311, 44 },
		{ 27, 312, 44 },
		{ 27, 313, 44 },
		{ 27, 314, 44 },
		{ 27, 315, 44 },
		{ 27, 316, 44 },
		{ 27, 317, 44 },
		{ 27, 318, 44 },
		{ 27, 319, 44 },
		{ 27, 320, 44 },
		{ 27, 321, 44 },
		{ 27, 322, 44 },
		{ 27, 323, 44 },
		{ 27, 324, 44 },
		{ 27, 325, 44 },
		{ 27, 326, 44 },
		{ 27, 327, 44 },
		{ 27, 328, 44 },
		{ 27, 329, 44 },
		{ 27, 330, 44 },
		{ 27, 331, 44 },
		{ 27, 332, 44 },
		{ 27, 333, 44 },
		{ 27, 334, 44 },
		{ 27, 335, 44 },
		{ 27, 336, 44 },
		{ 27, 337, 44 },
		{ 27, 338, 44 },
		{ 27, 339, 44 },
		{ 27, 340, 44 },
		{ 27, 341, 44 },
		{ 27, 342, 44 },
		{ 27, 343, 44 },
		{ 28, 286, 44 },
		{ 28, 287, 44 },
		{ 28, 288, 44 },
		{ 28, 289, 44 },
		{ 28, 290, 44 },
		{ 28, 291, 44 },
		{ 28, 292, 44 },
		{ 28, 293, 44 },
		{ 28, 294, 44 },
		{ 28, 295, 44 },
		{ 28, 296, 44 },
		{ 28, 297, 44 },
		{ 28, 298, 44 },
		{ 28, 299, 44 },
		{ 28, 300, 44 },
		{ 28, 301, 44 },
		{ 28, 302, 44 },
		{ 28, 303, 44 },
		{ 28, 304, 44 },
		{ 28, 305, 44 },
		{ 28, 306, 44 },
		{ 28, 307, 44 },
		{ 28, 308, 44 },
		{ 28, 309, 44 },
		{ 28, 310, 44 },
		{ 28, 311, 44 },
		{ 28, 312, 44 },
		{ 28, 313, 44 },
		{ 28, 314, 44 },
		{ 28, 315, 44 },
		{ 28, 316, 44 },
		{ 28, 317, 44 },
		{ 28, 318, 44 },
		{ 28, 319, 44 },
		{ 28, 320, 44 },
		{ 28, 321, 44 },
		{ 28, 322, 44 },
		{ 28, 323, 44 },
		{ 28, 324, 44 },
		{ 28, 325, 44 },
		{ 28, 326, 44 },
		{ 28, 327, 44 },
		{ 28, 328, 44 },
		{ 28, 329, 44 },
		{ 28, 330, 44 },
		{ 28, 331, 44 },
		{ 28, 332, 44 },
		{ 28, 333, 44 },
		{ 28, 334, 44 },
		{ 28, 335, 44 },
		{ 28, 336, 44 },
		{ 28, 337, 44 },
		{ 28, 338, 44 },
		{ 28, 339, 44 },
		{ 28, 340, 44 },
		{ 28, 341, 44 },
		{ 28, 342, 44 },
		{ 29, 286, 44 },
		{ 29, 287, 44 },
		{ 29, 288, 44 },
		{ 29, 289, 44 },
		{ 29, 290, 44 },
		{ 29, 291, 44 },
		{ 29, 292, 44 },
		{ 29, 293, 44 },
		{ 29, 294, 44 },
		{ 29, 295, 44 },
		{ 29, 296, 44 },
		{ 29, 297, 44 },
		{ 29, 298, 44 },
		{ 29, 299, 44 },
		{ 29, 300, 44 },
		{ 29, 301, 44 },
		{ 29, 302, 44 },
		{ 29, 303, 44 },
		{ 29, 304, 44 },
		{ 29, 305, 44 },
		{ 29, 306, 44 },
		{ 29, 307, 44 },
		{ 29, 308, 44 },
		{ 29, 309, 44 },
		{ 29, 310, 44 },
		{ 29, 311, 44 },
		{ 29, 312, 44 },
		{ 29, 313, 44 },
		{ 29, 314, 44 },
		{ 29, 315, 44 },
		{ 29, 316, 44 },
		{ 29, 317, 44 },
		{ 29, 318, 44 },
		{ 29, 319, 44 },
		{ 29, 320, 44 },
		{ 29, 321, 44 },
		{ 29, 322, 44 },
		{ 29, 323, 44 },
		{ 29, 324, 44 },
		{ 29, 325, 44 },
		{ 29, 326, 44 },
		{ 29, 327, 44 },
		{ 29, 328, 44 },
		{ 29, 329, 44 },
		{ 29, 330, 44 },
		{ 29, 331, 44 },
		{ 29, 332, 44 },
		{ 29, 333, 44 },
		{ 29, 334, 44 },
		{ 29, 335, 44 },
		{ 29, 336, 44 },
		{ 29, 337, 44 },
		{ 29, 338, 44 },
		{ 29, 339, 44 },
		{ 29, 340, 44 },
		{ 29, 341, 44 },
		{ 29, 342, 44 },
		{ 30, 286, 44 },
		{ 30, 287, 44 },
		{ 30, 288, 44 },
		{ 30, 289, 44 },
		{ 30, 290, 44 },
		{ 30, 291, 44 },
		{ 30, 292, 44 },
		{ 30, 293, 44 },
		{ 30, 294, 44 },
		{ 30, 295, 44 },
		{ 30, 296, 44 },
		{ 30, 297, 44 },
		{ 30, 298, 44 },
		{ 30, 299, 44 },
		{ 30, 300, 44 },
		{ 30, 301, 44 },
		{ 30, 302, 44 },
		{ 30, 303, 44 },
		{ 30, 304, 44 },
		{ 30, 305, 44 },
		{ 30, 306, 44 },
		{ 30, 307, 44 },
		{ 30, 308, 44 },
		{ 30, 309, 44 },
		{ 30, 310, 44 },
		{ 30, 311, 44 },
		{ 30, 312, 44 },
		{ 30, 313, 44 },
		{ 30, 314, 44 },
		{ 30, 315, 44 },
		{ 30, 316, 44 },
		{ 30, 317, 44 },
		{ 30, 318, 44 },
		{ 30, 319, 44 },
		{ 30, 320, 44 },
		{ 30, 321, 44 },
		{ 30, 322, 44 },
		{ 30, 323, 44 },
		{ 30, 324, 44 },
		{ 30, 325, 44 },
		{ 30, 326, 44 },
		{ 30, 327, 44 },
		{ 30, 328, 44 },
		{ 30, 329, 44 },
		{ 30, 330, 44 },
		{ 30, 331, 44 },
		{ 30, 332, 44 },
		{ 30, 333, 44 },
		{ 30, 334, 44 },
		{ 30, 335, 44 },
		{ 30, 336, 44 },
		{ 30, 337, 44 },
		{ 30, 338, 44 },
		{ 30, 339, 44 },
		{ 30, 340, 44 },
		{ 30, 341, 44 },
		{ 30, 342, 44 },
		{ 31, 286, 44 },
		{ 31, 287, 44 },
		{ 31, 288, 44 },
		{ 31, 289, 44 },
		{ 31, 290, 44 },
		{ 31, 291, 44 },
		{ 31, 292, 44 },
		{ 31, 293, 44 },
		{ 31, 294, 44 },
		{ 31, 295, 44 },
		{ 31, 296, 44 },
		{ 31, 297, 44 },
		{ 31, 298, 44 },
		{ 31, 299, 44 },
		{ 31, 300, 44 },
		{ 31, 301, 44 },
		{ 31, 302, 44 },
		{ 31, 303, 44 },
		{ 31, 304, 44 },
		{ 31, 305, 44 },
		{ 31, 306, 44 },
		{ 31, 307, 44 },
		{ 31, 308, 44 },
		{ 31, 309, 44 },
		{ 31, 310, 44 },
		{ 31, 311, 44 },
		{ 31, 312, 44 },
		{ 31, 313, 44 },
		{ 31, 314, 44 },
		{ 31, 315, 44 },
		{ 31, 316, 44 },
		{ 31, 317, 44 },
		{ 31, 318, 44 },
		{ 31, 319, 44 },
		{ 31, 320, 44 },
		{ 31, 321, 44 },
		{ 31, 322, 44 },
		{ 31, 323, 44 },
		{ 31, 324, 44 },
		{ 31, 325, 44 },
		{ 31, 326, 44 },
		{ 31, 327, 44 },
		{ 31, 328, 44 },
		{ 31, 329, 44 },
		{ 31, 330, 44 },
		{ 31, 331, 44 },
		{ 31, 332, 44 },
		{ 31, 333, 44 },
		{ 31, 334, 44 },
		{ 31, 335, 44 },
		{ 31, 336, 44 },
		{ 31, 337, 44 },
		{ 31, 338, 44 },
		{ 31, 339, 44 },
		{ 31, 340, 44 },
		{ 31, 341, 44 },
		{ 32, 286, 44 },
		{ 32, 287, 44 },
		{ 32, 288, 44 },
		{ 32, 289, 44 },
		{ 32, 290, 44 },
		{ 32, 291, 44 },
		{ 32, 292, 44 },
		{ 32, 293, 44 },
		{ 32, 294, 44 },
		{ 32, 295, 44 },
		{ 32, 296, 44 },
		{ 32, 297, 44 },
		{ 32, 298, 44 },
		{ 32, 299, 44 },
		{ 32, 300, 44 },
		{ 32, 301, 44 },
		{ 32, 302, 44 },
		{ 32, 303, 44 },
		{ 32, 304, 44 },
		{ 32, 305, 44 },
		{ 32, 306, 44 },
		{ 32, 307, 44 },
		{ 32, 308, 44 },
		{ 32, 309, 44 },
		{ 32, 310, 44 },
		{ 32, 311, 44 },
		{ 32, 312, 44 },
		{ 32, 313, 44 },
		{ 32, 314, 44 },
		{ 32, 315, 44 },
		{ 32, 316, 44 },
		{ 32, 317, 44 },
		{ 32, 318, 44 },
		{ 32, 319, 44 },
		{ 32, 320, 44 },
		{ 32, 321, 44 },
		{ 32, 322, 44 },
		{ 32, 323, 44 },
		{ 32, 324, 44 },
		{ 32, 325, 44 },
		{ 32, 326, 44 },
		{ 32, 327, 44 },
		{ 32, 328, 44 },
		{ 32, 329, 44 },
		{ 32, 330, 44 },
		{ 32, 331, 44 },
		{ 32, 332, 44 },
		{ 32, 333, 44 },
		{ 32, 334, 44 },
		{ 32, 335, 44 },
		{ 32, 336, 44 },
		{ 32, 337, 44 },
		{ 32, 338, 44 },
		{ 32, 339, 44 },
		{ 32, 340, 44 },
		{ 32, 341, 44 },
		{ 33, 286, 44 },
		{ 33, 287, 44 },
		{ 33, 288, 44 },
		{ 33, 289, 44 },
		{ 33, 290, 44 },
		{ 33, 291, 44 },
		{ 33, 292, 44 },
		{ 33, 293, 44 },
		{ 33, 294, 44 },
		{ 33, 295, 44 },
		{ 33, 296, 44 },
		{ 33, 297, 44 },
		{ 33, 298, 44 },
		{ 33, 299, 44 },
		{ 33, 300, 44 },
		{ 33, 301, 44 },
		{ 33, 302, 44 },
		{ 33, 303, 44 },
		{ 33, 304, 44 },
		{ 33, 305, 44 },
		{ 33, 306, 44 },
		{ 33, 307, 44 },
		{ 33, 308, 44 },
		{ 33, 309, 44 },
		{ 33, 310, 44 },
		{ 33, 311, 44 },
		{ 33, 312, 44 },
		{ 33, 313, 44 },
		{ 33, 314, 44 },
		{ 33, 315, 44 },
		{ 33, 316, 44 },
		{ 33, 317, 44 },
		{ 33, 318, 44 },
		{ 33, 319, 44 },
		{ 33, 320, 44 },
		{ 33, 321, 44 },
		{ 33, 322, 44 },
		{ 33, 323, 44 },
		{ 33, 324, 44 },
		{ 33, 325, 44 },
		{ 33, 326, 44 },
		{ 33, 327, 44 },
		{ 33, 328, 44 },
		{ 33, 329, 44 },
		{ 33, 330, 44 },
		{ 33, 331, 44 },
		{ 33, 332, 44 },
		{ 33, 333, 44 },
		{ 33, 334, 44 },
		{ 33, 335, 44 },
		{ 33, 336, 44 },
		{ 33, 337, 44 },
		{ 33, 338, 44 },
		{ 33, 339, 44 },
		{ 33, 340, 44 },
		{ 34, 285, 44 },
		{ 34, 286, 44 },
		{ 34, 287, 44 },
		{ 34, 288, 44 },
		{ 34, 289, 44 },
		{ 34, 290, 44 },
		{ 34, 291, 44 },
		{ 34, 292, 44 },
		{ 34, 293, 44 },
		{ 34, 294, 44 },
		{ 34, 295, 44 },
		{ 34, 296, 44 },
		{ 34, 297, 44 },
		{ 34, 298, 44 },
		{ 34, 299, 44 },
		{ 34, 300, 44 },
		{ 34, 301, 44 },
		{ 34, 302, 44 },
		{ 34, 303, 44 },
		{ 34, 304, 44 },
		{ 34, 305, 44 },
		{ 34, 306, 44 },
		{ 34, 307, 44 },
		{ 34, 308, 44 },
		{ 34, 309, 44 },
		{ 34, 310, 44 },
		{ 34, 311, 44 },
		{ 34, 312, 44 },
		{ 34, 313, 44 },
		{ 34, 314, 44 },
		{ 34, 315, 44 },
		{ 34, 316, 44 },
		{ 34, 317, 44 },
		{ 34, 318, 44 },
		{ 34, 319, 44 },
		{ 34, 320, 44 },
		{ 34, 321, 44 },
		{ 34, 322, 44 },
		{ 34, 323, 44 },
		{ 34, 324, 44 },
		{ 34, 325, 44 },
		{ 34, 326, 44 },
		{ 34, 327, 44 },
		{ 34, 328, 44 },
		{ 34, 329, 44 },
		{ 34, 330, 44 },
		{ 34, 331, 44 },
		{ 34, 332, 44 },
		{ 34, 333, 44 },
		{ 34, 334, 44 },
		{ 34, 335, 44 },
		{ 34, 336, 44 },
		{ 34, 337, 44 },
		{ 34, 338, 44 },
		{ 34, 339, 44 },
		{ 34, 340, 44 },
		{ 35, 285, 44 },
		{ 35, 286, 44 },
		{ 35, 287, 44 },
		{ 35, 288, 44 },
		{ 35, 289, 44 },
		{ 35, 290, 44 },
		{ 35, 291, 44 },
		{ 35, 292, 44 },
		{ 35, 293, 44 },
		{ 35, 294, 44 },
		{ 35, 295, 44 },
		{ 35, 296, 44 },
		{ 35, 297, 44 },
		{ 35, 298, 44 },
		{ 35, 299, 44 },
		{ 35, 300, 44 },
		{ 35, 301, 44 },
		{ 35, 302, 44 },
		{ 35, 303, 44 },
		{ 35, 304, 44 },
		{ 35, 305, 44 },
		{ 35, 306, 44 },
		{ 35, 307, 44 },
		{ 35, 308, 44 },
		{ 35, 309, 44 },
		{ 35, 310, 44 },
		{ 35, 311, 44 },
		{ 35, 312, 44 },
		{ 35, 313, 44 },
		{ 35, 314, 44 },
		{ 35, 315, 44 },
		{ 35, 316, 44 },
		{ 35, 317, 44 },
		{ 35, 318, 44 },
		{ 35, 319, 44 },
		{ 35, 320, 44 },
		{ 35, 321, 44 },
		{ 35, 322, 44 },
		{ 35, 323, 44 },
		{ 35, 324, 44 },
		{ 35, 325, 44 },
		{ 35, 326, 44 },
		{ 35, 327, 44 },
		{ 35, 328, 44 },
		{ 35, 329, 44 },
		{ 35, 330, 44 },
		{ 35, 331, 44 },
		{ 35, 332, 44 },
		{ 35, 333, 44 },
		{ 35, 334, 44 },
		{ 35, 335, 44 },
		{ 35, 336, 44 },
		{ 35, 337, 44 },
		{ 35, 338, 44 },
		{ 35, 339, 44 },
		{ 35, 340, 44 },
		{ 36, 285, 44 },
		{ 36, 286, 44 },
		{ 36, 287, 44 },
		{ 36, 288, 44 },
		{ 36, 289, 44 },
		{ 36, 290, 44 },
		{ 36, 291, 44 },
		{ 36, 292, 44 },
		{ 36, 293, 44 },
		{ 36, 294, 44 },
		{ 36, 295, 44 },
		{ 36, 296, 44 },
		{ 36, 297, 44 },
		{ 36, 298, 44 },
		{ 36, 299, 44 },
		{ 36, 300, 44 },
		{ 36, 301, 44 },
		{ 36, 302, 44 },
		{ 36, 303, 44 },
		{ 36, 304, 44 },
		{ 36, 305, 44 },
		{ 36, 306, 44 },
		{ 36, 307, 44 },
		{ 36, 308, 44 },
		{ 36, 309, 44 },
		{ 36, 310, 44 },
		{ 36, 311, 44 },
		{ 36, 312, 44 },
		{ 36, 313, 44 },
		{ 36, 314, 44 },
		{ 36, 315, 44 },
		{ 36, 316, 44 },
		{ 36, 317, 44 },
		{ 36, 318, 44 },
		{ 36, 319, 44 },
		{ 36, 320, 44 },
		{ 36, 321, 44 },
		{ 36, 322, 44 },
		{ 36, 323, 44 },
		{ 36, 324, 44 },
		{ 36, 325, 44 },
		{ 36, 326, 44 },
		{ 36, 327, 44 },
		{ 36, 328, 44 },
		{ 36, 329, 44 },
		{ 36, 330, 44 },
		{ 36, 331, 44 },
		{ 36, 332, 44 },
		{ 36, 333, 44 },
		{ 36, 334, 44 },
		{ 36, 335, 44 },
		{ 36, 336, 44 },
		{ 36, 337, 44 },
		{ 36, 338, 44 },
		{ 36, 339, 44 },
		{ 37, 285, 44 },
		{ 37, 286, 44 },
		{ 37, 287, 44 },
		{ 37, 288, 44 },
		{ 37, 289, 44 },
		{ 37, 290, 44 },
		{ 37, 291, 44 },
		{ 37, 292, 44 },
		{ 37, 293, 44 },
		{ 37, 294, 44 },
		{ 37, 295, 44 },
		{ 37, 296, 44 },
		{ 37, 297, 44 },
		{ 37, 298, 44 },
		{ 37, 299, 44 },
		{ 37, 300, 44 },
		{ 37, 301, 44 },
		{ 37, 302, 44 },
		{ 37, 303, 44 },
		{ 37, 304, 44 },
		{ 37, 305, 44 },
		{ 37, 306, 44 },
		{ 37, 307, 44 },
		{ 37, 308, 44 },
		{ 37, 309, 44 },
		{ 37, 310, 44 },
		{ 37, 311, 44 },
		{ 37, 312, 44 },
		{ 37, 313, 44 },
		{ 37, 314, 44 },
		{ 37, 315, 44 },
		{ 37, 316, 44 },
		{ 37, 317, 44 },
		{ 37, 318, 44 },
		{ 37, 319, 44 },
		{ 37, 320, 44 },
		{ 37, 321, 44 },
		{ 37, 322, 44 },
		{ 37, 323, 44 },
		{ 37, 324, 44 },
		{ 37, 325, 44 },
		{ 37, 326, 44 },
		{ 37, 327, 44 },
		{ 37, 328, 44 },
		{ 37, 329, 44 },
		{ 37, 330, 44 },
		{ 37, 331, 44 },
		{ 37, 332, 44 },
		{ 37, 333, 44 },
		{ 37, 334, 44 },
		{ 37, 335, 44 },
		{ 37, 336, 44 },
		{ 37, 337, 44 },
		{ 37, 338, 44 },
		{ 37, 339, 44 },
		{ 38, 285, 44 },
		{ 38, 286, 44 },
		{ 38, 287, 44 },
		{ 38, 288, 44 },
		{ 38, 289, 44 },
		{ 38, 290, 44 },
		{ 38, 291, 44 },
		{ 38, 292, 44 },
		{ 38, 293, 44 },
		{ 38, 294, 44 },
		{ 38, 295, 44 },
		{ 38, 296, 44 },
		{ 38, 297, 44 },
		{ 38, 298, 44 },
		{ 38, 299, 44 },
		{ 38, 300, 44 },
		{ 38, 301, 44 },
		{ 38, 302, 44 },
		{ 38, 303, 44 },
		{ 38, 304, 44 },
		{ 38, 305, 44 },
		{ 38, 306, 44 },
		{ 38, 307, 44 },
		{ 38, 308, 44 },
		{ 38, 309, 44 },
		{ 38, 310, 44 },
		{ 38, 311, 44 },
		{ 38, 312, 44 },
		{ 38, 313, 44 },
		{ 38, 314, 44 },
		{ 38, 315, 44 },
		{ 38, 316, 44 },
		{ 38, 317, 44 },
		{ 38, 318, 44 },
		{ 38, 319, 44 },
		{ 38, 320, 44 },
		{ 38, 321, 44 },
		{ 38, 322, 44 },
		{ 38, 323, 44 },
		{ 38, 324, 44 },
		{ 38, 325, 44 },
		{ 38, 326, 44 },
		{ 38, 327, 44 },
		{ 38, 328, 44 },
		{ 38, 329, 44 },
		{ 38, 330, 44 },
		{ 38, 331, 44 },
		{ 38, 332, 44 },
		{ 38, 333, 44 },
		{ 38, 334, 44 },
		{ 38, 335, 44 },
		{ 38, 336, 44 },
		{ 38, 337, 44 },
		{ 38, 338, 44 },
		{ 38, 339, 44 },
		{ 39, 285, 44 },
		{ 39, 286, 44 },
		{ 39, 287, 44 },
		{ 39, 288, 44 },
		{ 39, 289, 44 },
		{ 39, 290, 44 },
		{ 39, 291, 44 },
		{ 39, 292, 44 },
		{ 39, 293, 44 },
		{ 39, 294, 44 },
		{ 39, 295, 44 },
		{ 39, 296, 44 },
		{ 39, 297, 44 },
		{ 39, 298, 44 },
		{ 39, 299, 44 },
		{ 39, 300, 44 },
		{ 39, 301, 44 },
		{ 39, 302, 44 },
		{ 39, 303, 44 },
		{ 39, 304, 44 },
		{ 39, 305, 44 },
		{ 39, 306, 44 },
		{ 39, 307, 44 },
		{ 39, 308, 44 },
		{ 39, 309, 44 },
		{ 39, 310, 44 },
		{ 39, 311, 44 },
		{ 39, 312, 44 },
		{ 39, 313, 44 },
		{ 39, 314, 44 },
		{ 39, 315, 44 },
		{ 39, 316, 44 },
		{ 39, 317, 44 },
		{ 39, 318, 44 },
		{ 39, 319, 44 },
		{ 39, 320, 44 },
		{ 39, 321, 44 },
		{ 39, 322, 44 },
		{ 39, 323, 44 },
		{ 39, 324, 44 },
		{ 39, 325, 44 },
		{ 39, 326, 44 },
		{ 39, 327, 44 },
		{ 39, 328, 44 },
		{ 39, 329, 44 },
		{ 39, 330, 44 },
		{ 39, 331, 44 },
		{ 39, 332, 44 },
		{ 39, 333, 44 },
		{ 39, 334, 44 },
		{ 39, 335, 44 },
		{ 39, 336, 44 },
		{ 39, 337, 44 },
		{ 39, 338, 44 },
		{ 40, 285, 44 },
		{ 40, 286, 44 },
		{ 40, 287, 44 },
		{ 40, 288, 44 },
		{ 40, 289, 44 },
		{ 40, 290, 44 },
		{ 40, 291, 44 },
		{ 40, 292, 44 },
		{ 40, 293, 44 },
		{ 40, 294, 44 },
		{ 40, 295, 44 },
		{ 40, 296, 44 },
		{ 40, 297, 44 },
		{ 40, 298, 44 },
		{ 40, 299, 44 },
		{ 40, 300, 44 },
		{ 40, 301, 44 },
		{ 40, 302, 44 },
		{ 40, 303, 44 },
		{ 40, 304, 44 },
		{ 40, 305, 44 },
		{ 40, 306, 44 },
		{ 40, 307, 44 },
		{ 40, 308, 44 },
		{ 40, 309, 44 },
		{ 40, 310, 44 },
		{ 40, 311, 44 },
		{ 40, 312, 44 },
		{ 40, 313, 44 },
		{ 40, 314, 44 },
		{ 40, 315, 44 },
		{ 40, 316, 44 },
		{ 40, 317, 44 },
		{ 40, 318, 44 },
		{ 40, 319, 44 },
		{ 40, 320, 44 },
		{ 40, 321, 44 },
		{ 40, 322, 44 },
		{ 40, 323, 44 },
		{ 40, 324, 44 },
		{ 40, 325, 44 },
		{ 40, 326, 44 },
		{ 40, 327, 44 },
		{ 40, 328, 44 },
		{ 40, 329, 44 },
		{ 40, 330, 44 },
		{ 40, 331, 44 },
		{ 40, 332, 44 },
		{ 40, 333, 44 },
		{ 40, 334, 44 },
		{ 40, 335, 44 },
		{ 40, 336, 44 },
		{ 40, 337, 44 },
		{ 40, 338, 44 },
		{ 41, 285, 44 },
		{ 41, 286, 44 },
		{ 41, 287, 44 },
		{ 41, 288, 44 },
		{ 41, 289, 44 },
		{ 41, 290, 44 },
		{ 41, 291, 44 },
		{ 41, 292, 44 },
		{ 41, 293, 44 },
		{ 41, 294, 44 },
		{ 41, 295, 44 },
		{ 41, 296, 44 },
		{ 41, 297, 44 },
		{ 41, 298, 44 },
		{ 41, 299, 44 },
		{ 41, 300, 44 },
		{ 41, 301, 44 },
		{ 41, 302, 44 },
		{ 41, 303, 44 },
		{ 41, 304, 44 },
		{ 41, 305, 44 },
		{ 41, 306, 44 },
		{ 41, 307, 44 },
		{ 41, 308, 44 },
		{ 41, 309, 44 },
		{ 41, 310, 44 },
		{ 41, 311, 44 },
		{ 41, 312, 44 },
		{ 41, 313, 44 },
		{ 41, 314, 44 },
		{ 41, 315, 44 },
		{ 41, 316, 44 },
		{ 41, 317, 44 },
		{ 41, 318, 44 },
		{ 41, 319, 44 },
		{ 41, 320, 44 },
		{ 41, 321, 44 },
		{ 41, 322, 44 },
		{ 41, 323, 44 },
		{ 41, 324, 44 },
		{ 41, 325, 44 },
		{ 41, 326, 44 },
		{ 41, 327, 44 },
		{ 41, 328, 44 },
		{ 41, 329, 44 },
		{ 41, 330, 44 },
		{ 41, 331, 44 },
		{ 41, 332, 44 },
		{ 41, 333, 44 },
		{ 41, 334, 44 },
		{ 41, 335, 44 },
		{ 41, 336, 44 },
		{ 41, 337, 44 },
		{ 42, 284, 44 },
		{ 42, 285, 44 },
		{ 42, 286, 44 },
		{ 42, 287, 44 },
		{ 42, 288, 44 },
		{ 42, 289, 44 },
		{ 42, 290, 44 },
		{ 42, 291, 44 },
		{ 42, 292, 44 },
		{ 42, 293, 44 },
		{ 42, 294, 44 },
		{ 42, 295, 44 },
		{ 42, 296, 44 },
		{ 42, 297, 44 },
		{ 42, 298, 44 },
		{ 42, 299, 44 },
		{ 42, 300, 44 },
		{ 42, 301, 44 },
		{ 42, 302, 44 },
		{ 42, 303, 44 },
		{ 42, 304, 44 },
		{ 42, 305, 44 },
		{ 42, 306, 44 },
		{ 42, 307, 44 },
		{ 42, 308, 44 },
		{ 42, 309, 44 },
		{ 42, 310, 44 },
		{ 42, 311, 44 },
		{ 42, 312, 44 },
		{ 42, 313, 44 },
		{ 42, 314, 44 },
		{ 42, 315, 44 },
		{ 42, 316, 44 },
		{ 42, 317, 44 },
		{ 42, 318, 44 },
		{ 42, 319, 44 },
		{ 42, 320, 44 },
		{ 42, 321, 44 },
		{ 42, 322, 44 },
		{ 42, 323, 44 },
		{ 42, 324, 44 },
		{ 42, 325, 44 },
		{ 42, 326, 44 },
		{ 42, 327, 44 },
		{ 42, 328, 44 },
		{ 42, 329, 44 },
		{ 42, 330, 44 },
		{ 42, 331, 44 },
		{ 42, 332, 44 },
		{ 42, 333, 44 },
		{ 42, 334, 44 },
		{ 42, 335, 44 },
		{ 42, 336, 44 },
		{ 42, 337, 44 },
		{ 43, 284, 44 },
		{ 43, 285, 44 },
		{ 43, 286, 44 },
		{ 43, 287, 44 },
		{ 43, 288, 44 },
		{ 43, 289, 44 },
		{ 43, 290, 44 },
		{ 43, 291, 44 },
		{ 43, 292, 44 },
		{ 43, 293, 44 },
		{ 43, 294, 44 },
		{ 43, 295, 44 },
		{ 43, 296, 44 },
		{ 43, 297, 44 },
		{ 43, 298, 44 },
		{ 43, 299, 44 },
		{ 43, 300, 44 },
		{ 43, 301, 44 },
		{ 43, 302, 44 },
		{ 43, 303, 44 },
		{ 43, 304, 44 },
		{ 43, 305, 44 },
		{ 43, 306, 44 },
		{ 43, 307, 44 },
		{ 43, 308, 44 },
		{ 43, 309, 44 },
		{ 43, 310, 44 },
		{ 43, 311, 44 },
		{ 43, 312, 44 },
		{ 43, 313, 44 },
		{ 43, 314, 44 },
		{ 43, 315, 44 },
		{ 43, 316, 44 },
		{ 43, 317, 44 },
		{ 43, 318, 44 },
		{ 43, 319, 44 },
		{ 43, 320, 44 },
		{ 43, 321, 44 },
		{ 43, 322, 44 },
		{ 43, 323, 44 },
		{ 43, 324, 44 },
		{ 43, 325, 44 },
		{ 43, 326, 44 },
		{ 43, 327, 44 },
		{ 43, 328, 44 },
		{ 43, 329, 44 },
		{ 43, 330, 44 },
		{ 43, 331, 44 },
		{ 43, 332, 44 },
		{ 43, 333, 44 },
		{ 43, 334, 44 },
		{ 43, 335, 44 },
		{ 43, 336, 44 },
		{ 43, 337, 44 },
		{ 44, 284, 44 },
		{ 44, 285, 44 },
		{ 44, 286, 44 },
		{ 44, 287, 44 },
		{ 44, 288, 44 },
		{ 44, 289, 44 },
		{ 44, 290, 44 },
		{ 44, 291, 44 },
		{ 44, 292, 44 },
		{ 44, 293, 44 },
		{ 44, 294, 44 },
		{ 44, 295, 44 },
		{ 44, 296, 44 },
		{ 44, 297, 44 },
		{ 44, 298, 44 },
		{ 44, 299, 44 },
		{ 44, 300, 44 },
		{ 44, 301, 44 },
		{ 44, 302, 44 },
		{ 44, 303, 44 },
		{ 44, 304, 44 },
		{ 44, 305, 44 },
		{ 44, 306, 44 },
		{ 44, 307, 44 },
		{ 44, 308, 44 },
		{ 44, 309, 44 },
		{ 44, 310, 44 },
		{ 44, 311, 44 },
		{ 44, 312, 44 },
		{ 44, 313, 44 },
		{ 44, 314, 44 },
		{ 44, 315, 44 },
		{ 44, 316, 44 },
		{ 44, 317, 44 },
		{ 44, 318, 44 },
		{ 44, 319, 44 },
		{ 44, 320, 44 },
		{ 44, 321, 44 },
		{ 44, 322, 44 },
		{ 44, 323, 44 },
		{ 44, 324, 44 },
		{ 44, 325, 44 },
		{ 44, 326, 44 },
		{ 44, 327, 44 },
		{ 44, 328, 44 },
		{ 44, 329, 44 },
		{ 44, 330, 44 },
		{ 44, 331, 44 },
		{ 44, 332, 44 },
		{ 44, 333, 44 },
		{ 44, 334, 44 },
		{ 44, 335, 44 },
		{ 44, 336, 44 },
		{ 45, 284, 44 },
		{ 45, 285, 44 },
		{ 45, 286, 44 },
		{ 45, 287, 44 },
		{ 45, 288, 44 },
		{ 45, 289, 44 },
		{ 45, 290, 44 },
		{ 45, 291, 44 },
		{ 45, 292, 44 },
		{ 45, 293, 44 },
		{ 45, 294, 44 },
		{ 45, 295, 44 },
		{ 45, 296, 44 },
		{ 45, 297, 44 },
		{ 45, 298, 44 },
		{ 45, 299, 44 },
		{ 45, 300, 44 },
		{ 45, 301, 44 },
		{ 45, 302, 44 },
		{ 45, 303, 44 },
		{ 45, 304, 44 },
		{ 45, 305, 44 },
		{ 45, 306, 44 },
		{ 45, 307, 44 },
		{ 45, 308, 44 },
		{ 45, 309, 44 },
		{ 45, 310, 44 },
		{ 45, 311, 44 },
		{ 45, 312, 44 },
		{ 45, 313, 44 },
		{ 45, 314, 44 },
		{ 45, 315, 44 },
		{ 45, 316, 44 },
		{ 45, 317, 44 },
		{ 45, 318, 44 },
		{ 45, 319, 44 },
		{ 45, 320, 44 },
		{ 45, 321, 44 },
		{ 45, 322, 44 },
		{ 45, 323, 44 },
		{ 45, 324, 44 },
		{ 45, 325, 44 },
		{ 45, 326, 44 },
		{ 45, 327, 44 },
		{ 45, 328, 44 },
		{ 45, 329, 44 },
		{ 45, 330, 44 },
		{ 45, 331, 44 },
		{ 45, 332, 44 },
		{ 45, 333, 44 },
		{ 45, 334, 44 },
		{ 45, 335, 44 },
		{ 45, 336, 44 },
		{ 46, 284, 44 },
		{ 46, 285, 44 },
		{ 46, 286, 44 },
		{ 46, 287, 44 },
		{ 46, 288, 44 },
		{ 46, 289, 44 },
		{ 46, 290, 44 },
		{ 46, 291, 44 },
		{ 46, 292, 44 },
		{ 46, 293, 44 },
		{ 46, 294, 44 },
		{ 46, 295, 44 },
		{ 46, 296, 44 },
		{ 46, 297, 44 },
		{ 46, 298, 44 },
		{ 46, 299, 44 },
		{ 46, 300, 44 },
		{ 46, 301, 44 },
		{ 46, 302, 44 },
		{ 46, 303, 44 },
		{ 46, 304, 44 },
		{ 46, 305, 44 },
		{ 46, 306, 44 },
		{ 46, 307, 44 },
		{ 46, 308, 44 },
		{ 46, 309, 44 },
		{ 46, 310, 44 },
		{ 46, 311, 44 },
		{ 46, 312, 44 },
		{ 46, 313, 44 },
		{ 46, 314, 44 },
		{ 46, 315, 44 },
		{ 46, 316, 44 },
		{ 46, 317, 44 },
		{ 46, 318, 44 },
		{ 46, 319, 44 },
		{ 46, 320, 44 },
		{ 46, 321, 44 },
		{ 46, 322, 44 },
		{ 46, 323, 44 },
		{ 46, 324, 44 },
		{ 46, 325, 44 },
		{ 46, 326, 44 },
		{ 46, 327, 44 },
		{ 46, 328, 44 },
		{ 46, 329, 44 },
		{ 46, 330, 44 },
		{ 46, 331, 44 },
		{ 46, 332, 44 },
		{ 46, 333, 44 },
		{ 46, 334, 44 },
		{ 46, 335, 44 },
		{ 46, 336, 44 },
		{ 47, 284, 44 },
		{ 47, 285, 44 },
		{ 47, 286, 44 },
		{ 47, 287, 44 },
		{ 47, 288, 44 },
		{ 47, 289, 44 },
		{ 47, 290, 44 },
		{ 47, 291, 44 },
		{ 47, 292, 44 },
		{ 47, 293, 44 },
		{ 47, 294, 44 },
		{ 47, 295, 44 },
		{ 47, 296, 44 },
		{ 47, 297, 44 },
		{ 47, 298, 44 },
		{ 47, 299, 44 },
		{ 47, 300, 44 },
		{ 47, 301, 44 },
		{ 47, 302, 44 },
		{ 47, 303, 44 },
		{ 47, 304, 44 },
		{ 47, 305, 44 },
		{ 47, 306, 44 },
		{ 47, 307, 44 },
		{ 47, 308, 44 },
		{ 47, 309, 44 },
		{ 47, 310, 44 },
		{ 47, 311, 44 },
		{ 47, 312, 44 },
		{ 47, 313, 44 },
		{ 47, 314, 44 },
		{ 47, 315, 44 },
		{ 47, 316, 44 },
		{ 47, 317, 44 },
		{ 47, 318, 44 },
		{ 47, 319, 44 },
		{ 47, 320, 44 },
		{ 47, 321, 44 },
		{ 47, 322, 44 },
		{ 47, 323, 44 },
		{ 47, 324, 44 },
		{ 47, 325, 44 },
		{ 47, 326, 44 },
		{ 47, 327, 44 },
		{ 47, 328, 44 },
		{ 47, 329, 44 },
		{ 47, 330, 44 },
		{ 47, 331, 44 },
		{ 47, 332, 44 },
		{ 47, 333, 44 },
		{ 47, 334, 44 },
		{ 47, 335, 44 },
		{ 48, 284, 44 },
		{ 48, 285, 44 },
		{ 48, 286, 44 },
		{ 48, 287, 44 },
		{ 48, 288, 44 },
		{ 48, 289, 44 },
		{ 48, 290, 44 },
		{ 48, 291, 44 },
		{ 48, 292, 44 },
		{ 48, 293, 44 },
		{ 48, 294, 44 },
		{ 48, 295, 44 },
		{ 48, 296, 44 },
		{ 48, 297, 44 },
		{ 48, 298, 44 },
		{ 48, 299, 44 },
		{ 48, 300, 44 },
		{ 48, 301, 44 },
		{ 48, 302, 44 },
		{ 48, 303, 44 },
		{ 48, 304, 44 },
		{ 48, 305, 44 },
		{ 48, 306, 44 },
		{ 48, 307, 44 },
		{ 48, 308, 44 },
		{ 48, 309, 44 },
		{ 48, 310, 44 },
		{ 48, 311, 44 },
		{ 48, 312, 44 },
		{ 48, 313, 44 },
		{ 48, 314, 44 },
		{ 48, 315, 44 },
		{ 48, 316, 44 },
		{ 48, 317, 44 },
		{ 48, 318, 44 },
		{ 48, 319, 44 },
		{ 48, 320, 44 },
		{ 48, 321, 44 },
		{ 48, 322, 44 },
		{ 48, 323, 44 },
		{ 48, 324, 44 },
		{ 48, 325, 44 },
		{ 48, 326, 44 },
		{ 48, 327, 44 },
		{ 48, 328, 44 },
		{ 48, 329, 44 },
		{ 48, 330, 44 },
		{ 48, 331, 44 },
		{ 48, 332, 44 },
		{ 48, 333, 44 },
		{ 48, 334, 44 },
		{ 48, 335, 44 },
		{ 49, 284, 44 },
		{ 49, 285, 44 },
		{ 49, 286, 44 },
		{ 49, 287, 44 },
		{ 49, 288, 44 },
		{ 49, 289, 44 },
		{ 49, 290, 44 },
		{ 49, 291, 44 },
		{ 49, 292, 44 },
		{ 49, 293, 44 },
		{ 49, 294, 44 },
		{ 49, 295, 44 },
		{ 49, 296, 44 },
		{ 49, 297, 44 },
		{ 49, 298, 44 },
		{ 49, 299, 44 },
		{ 49, 300, 44 },
		{ 49, 301, 44 },
		{ 49, 302, 44 },
		{ 49, 303, 44 },
		{ 49, 304, 44 },
		{ 49, 305, 44 },
		{ 49, 306, 44 },
		{ 49, 307, 44 },
		{ 49, 308, 44 },
		{ 49, 309, 44 },
		{ 49, 310, 44 },
		{ 49, 311, 44 },
		{ 49, 312, 44 },
		{ 49, 313, 44 },
		{ 49, 314, 44 },
		{ 49, 315, 44 },
		{ 49, 316, 44 },
		{ 49, 317, 44 },
		{ 49, 318, 44 },
		{ 49, 319, 44 },
		{ 49, 320, 44 },
		{ 49, 321, 44 },
		{ 49, 322, 44 },
		{ 49, 323, 44 },
		{ 49, 324, 44 },
		{ 49, 325, 44 },
		{ 49, 326, 44 },
		{ 49, 327, 44 },
		{ 49, 328, 44 },
		{ 49, 329, 44 },
		{ 49, 330, 44 },
		{ 49, 331, 44 },
		{ 49, 332, 44 },
		{ 49, 333, 44 },
		{ 49, 334, 44 },
		{ 50, 283, 44 },
		{ 50, 284, 44 },
		{ 50, 285, 44 },
		{ 50, 286, 44 },
		{ 50, 287, 44 },
		{ 50, 288, 44 },
		{ 50, 289, 44 },
		{ 50, 290, 44 },
		{ 50, 291, 44 },
		{ 50, 292, 44 },
		{ 50, 293, 44 },
		{ 50, 294, 44 },
		{ 50, 295, 44 },
		{ 50, 296, 44 },
		{ 50, 297, 44 },
		{ 50, 298, 44 },
		{ 50, 299, 44 },
		{ 50, 300, 44 },
		{ 50, 301, 44 },
		{ 50, 302, 44 },
		{ 50, 303, 44 },
		{ 50, 304, 44 },
		{ 50, 305, 44 },
		{ 50, 306, 44 },
		{ 50, 307, 44 },
		{ 50, 308, 44 },
		{ 50, 309, 44 },
		{ 50, 310, 44 },
		{ 50, 311, 44 },
		{ 50, 312, 44 },
		{ 50, 313, 44 },
		{ 50, 314, 44 },
		{ 50, 315, 44 },
		{ 50, 316, 44 },
		{ 50, 317, 44 },
		{ 50, 318, 44 },
		{ 50, 319, 44 },
		{ 50, 320, 44 },
		{ 50, 321, 44 },
		{ 50, 322, 44 },
		{ 50, 323, 44 },
		{ 50, 324, 44 },
		{ 50, 325, 44 },
		{ 50, 326, 44 },
		{ 50, 327, 44 },
		{ 50, 328, 44 },
		{ 50, 329, 44 },
		{ 50, 330, 44 },
		{ 50, 331, 44 },
		{ 50, 332, 44 },
		{ 50, 333, 44 },
		{ 50, 334, 44 },
		{ 51, 283, 44 },
		{ 51, 284, 44 },
		{ 51, 285, 44 },
		{ 51, 286, 44 },
		{ 51, 287, 44 },
		{ 51, 288, 44 },
		{ 51, 289, 44 },
		{ 51, 290, 44 },
		{ 51, 291, 44 },
		{ 51, 292, 44 },
		{ 51, 293, 44 },
		{ 51, 294, 44 },
		{ 51, 295, 44 },
		{ 51, 296, 44 },
		{ 51, 297, 44 },
		{ 51, 298, 44 },
		{ 51, 299, 44 },
		{ 51, 300, 44 },
		{ 51, 301, 44 },
		{ 51, 302, 44 },
		{ 51, 303, 44 },
		{ 51, 304, 44 },
		{ 51, 305, 44 },
		{ 51, 306, 44 },
		{ 51, 307, 44 },
		{ 51, 308, 44 },
		{ 51, 309, 44 },
		{ 51, 310, 44 },
		{ 51, 311, 44 },
		{ 51, 312, 44 },
		{ 51, 313, 44 },
		{ 51, 314, 44 },
		{ 51, 315, 44 },
		{ 51, 316, 44 },
		{ 51, 317, 44 },
		{ 51, 318, 44 },
		{ 51, 319, 44 },
		{ 51, 320, 44 },
		{ 51, 321, 44 },
		{ 51, 322, 44 },
		{ 51, 323, 44 },
		{ 51, 324, 44 },
		{ 51, 325, 44 },
		{ 51, 326, 44 },
		{ 51, 327, 44 },
		{ 51, 328, 44 },
		{ 51, 329, 44 },
		{ 51, 330, 44 },
		{ 51, 331, 44 },
		{ 51, 332, 44 },
		{ 51, 333, 44 },
		{ 51, 334, 44 },
		{ 52, 283, 44 },
		{ 52, 284, 44 },
		{ 52, 285, 44 },
		{ 52, 286, 44 },
		{ 52, 287, 44 },
		{ 52, 288, 44 },
		{ 52, 289, 44 },
		{ 52, 290, 44 },
		{ 52, 291, 44 },
		{ 52, 292, 44 },
		{ 52, 293, 44 },
		{ 52, 294, 44 },
		{ 52, 295, 44 },
		{ 52, 296, 44 },
		{ 52, 297, 44 },
		{ 52, 298, 44 },
		{ 52, 299, 44 },
		{ 52, 300, 44 },
		{ 52, 301, 44 },
		{ 52, 302, 44 },
		{ 52, 303, 44 },
		{ 52, 304, 44 },
		{ 52, 305, 44 },
		{ 52, 306, 44 },
		{ 52, 307, 44 },
		{ 52, 308, 44 },
		{ 52, 309, 44 },
		{ 52, 310, 44 },
		{ 52, 311, 44 },
		{ 52, 312, 44 },
		{ 52, 313, 44 },
		{ 52, 314, 44 },
		{ 52, 315, 44 },
		{ 52, 316, 44 },
		{ 52, 317, 44 },
		{ 52, 318, 44 },
		{ 52, 319, 44 },
		{ 52, 320, 44 },
		{ 52, 321, 44 },
		{ 52, 322, 44 },
		{ 52, 323, 44 },
		{ 52, 324, 44 },
		{ 52, 325, 44 },
		{ 52, 326, 44 },
		{ 52, 327, 44 },
		{ 52, 328, 44 },
		{ 52, 329, 44 },
		{ 52, 330, 44 },
		{ 52, 331, 44 },
		{ 52, 332, 44 },
		{ 52, 333, 44 },
		{ 53, 283, 44 },
		{ 53, 284, 44 },
		{ 53, 285, 44 },
		{ 53, 286, 44 },
		{ 53, 287, 44 },
		{ 53, 288, 44 },
		{ 53, 289, 44 },
		{ 53, 290, 44 },
		{ 53, 291, 44 },
		{ 53, 292, 44 },
		{ 53, 293, 44 },
		{ 53, 294, 44 },
		{ 53, 295, 44 },
		{ 53, 296, 44 },
		{ 53, 297, 44 },
		{ 53, 298, 44 },
		{ 53, 299, 44 },
		{ 53, 300, 44 },
		{ 53, 301, 44 },
		{ 53, 302, 44 },
		{ 53, 303, 44 },
		{ 53, 304, 44 },
		{ 53, 305, 44 },
		{ 53, 306, 44 },
		{ 53, 307, 44 },
		{ 53, 308, 44 },
		{ 53, 309, 44 },
		{ 53, 310, 44 },
		{ 53, 311, 44 },
		{ 53, 312, 44 },
		{ 53, 313, 44 },
		{ 53, 314, 44 },
		{ 53, 315, 44 },
		{ 53, 316, 44 },
		{ 53, 317, 44 },
		{ 53, 318, 44 },
		{ 53, 319, 44 },
		{ 53, 320, 44 },
		{ 53, 321, 44 },
		{ 53, 322, 44 },
		{ 53, 323, 44 },
		{ 53, 324, 44 },
		{ 53, 325, 44 },
		{ 53, 326, 44 },
		{ 53, 327, 44 },
		{ 53, 328, 44 },
		{ 53, 329, 44 },
		{ 53, 330, 44 },
		{ 53, 331, 44 },
		{ 53, 332, 44 },
		{ 53, 333, 44 },
		{ 54, 283, 44 },
		{ 54, 284, 44 },
		{ 54, 285, 44 },
		{ 54, 286, 44 },
		{ 54, 287, 44 },
		{ 54, 288, 44 },
		{ 54, 289, 44 },
		{ 54, 290, 44 },
		{ 54, 291, 44 },
		{ 54, 292, 44 },
		{ 54, 293, 44 },
		{ 54, 294, 44 },
		{ 54, 295, 44 },
		{ 54, 296, 44 },
		{ 54, 297, 44 },
		{ 54, 298, 44 },
		{ 54, 299, 44 },
		{ 54, 300, 44 },
		{ 54, 301, 44 },
		{ 54, 302, 44 },
		{ 54, 303, 44 },
		{ 54, 304, 44 },
		{ 54, 305, 44 },
		{ 54, 306, 44 },
		{ 54, 307, 44 },
		{ 54, 308, 44 },
		{ 54, 309, 44 },
		{ 54, 310, 44 },
		{ 54, 311, 44 },
		{ 54, 312, 44 },
		{ 54, 313, 44 },
		{ 54, 314, 44 },
		{ 54, 315, 44 },
		{ 54, 316, 44 },
		{ 54, 317, 44 },
		{ 54, 318, 44 },
		{ 54, 319, 44 },
		{ 54, 320, 44 },
		{ 54, 321, 44 },
		{ 54, 322, 44 },
		{ 54, 323, 44 },
		{ 54, 324, 44 },
		{ 54, 325, 44 },
		{ 54, 326, 44 },
		{ 54, 327, 44 },
		{ 54, 328, 44 },
		{ 54, 329, 44 },
		{ 54, 330, 44 },
		{ 54, 331, 44 },
		{ 54, 332, 44 },
		{ 54, 333, 44 },
		{ 55, 283, 44 },
		{ 55, 284, 44 },
		{ 55, 285, 44 },
		{ 55, 286, 44 },
		{ 55, 287, 44 },
		{ 55, 288, 44 },
		{ 55, 289, 44 },
		{ 55, 290, 44 },
		{ 55, 291, 44 },
		{ 55, 292, 44 },
		{ 55, 293, 44 },
		{ 55, 294, 44 },
		{ 55, 295, 44 },
		{ 55, 296, 44 },
		{ 55, 297, 44 },
		{ 55, 298, 44 },
		{ 55, 299, 44 },
		{ 55, 300, 44 },
		{ 55, 301, 44 },
		{ 55, 302, 44 },
		{ 55, 303, 44 },
		{ 55, 304, 44 },
		{ 55, 305, 44 },
		{ 55, 306, 44 },
		{ 55, 307, 44 },
		{ 55, 308, 44 },
		{ 55, 309, 44 },
		{ 55, 310, 44 },
		{ 55, 311, 44 },
		{ 55, 312, 44 },
		{ 55, 313, 44 },
		{ 55, 314, 44 },
		{ 55, 315, 44 },
		{ 55, 316, 44 },
		{ 55, 317, 44 },
		{ 55, 318, 44 },
		{ 55, 319, 44 },
		{ 55, 320, 44 },
		{ 55, 321, 44 },
		{ 55, 322, 44 },
		{ 55, 323, 44 },
		{ 55, 324, 44 },
		{ 55, 325, 44 },
		{ 55, 326, 44 },
		{ 55, 327, 44 },
		{ 55, 328, 44 },
		{ 55, 329, 44 },
		{ 55, 330, 44 },
		{ 55, 331, 44 },
		{ 55, 332, 44 },
		{ 56, 283, 44 },
		{ 56, 284, 44 },
		{ 56, 285, 44 },
		{ 56, 286, 44 },
		{ 56, 287, 44 },
		{ 56, 288, 44 },
		{ 56, 289, 44 },
		{ 56, 290, 44 },
		{ 56, 291, 44 },
		{ 56, 292, 44 },
		{ 56, 293, 44 },
		{ 56, 294, 44 },
		{ 56, 295, 44 },
		{ 56, 296, 44 },
		{ 56, 297, 44 },
		{ 56, 298, 44 },
		{ 56, 299, 44 },
		{ 56, 300, 44 },
		{ 56, 301, 44 },
		{ 56, 302, 44 },
		{ 56, 303, 44 },
		{ 56, 304, 44 },
		{ 56, 305, 44 },
		{ 56, 306, 44 },
		{ 56, 307, 44 },
		{ 56, 308, 44 },
		{ 56, 309, 44 },
		{ 56, 310, 44 },
		{ 56, 311, 44 },
		{ 56, 312, 44 },
		{ 56, 313, 44 },
		{ 56, 314, 44 },
		{ 56, 315, 44 },
		{ 56, 316, 44 },
		{ 56, 317, 44 },
		{ 56, 318, 44 },
		{ 56, 319, 44 },
		{ 56, 320, 44 },
		{ 56, 321, 44 },
		{ 56, 322, 44 },
		{ 56, 323, 44 },
		{ 56, 324, 44 },
		{ 56, 325, 44 },
		{ 56, 326, 44 },
		{ 56, 327, 44 },
		{ 56, 328, 44 },
		{ 56, 329, 44 },
		{ 56, 330, 44 },
		{ 56, 331, 44 },
		{ 56, 332, 44 },
		{ 57, 283, 44 },
		{ 57, 284, 44 },
		{ 57, 285, 44 },
		{ 57, 286, 44 },
		{ 57, 287, 44 },
		{ 57, 288, 44 },
		{ 57, 289, 44 },
		{ 57, 290, 44 },
		{ 57, 291, 44 },
		{ 57, 292, 44 },
		{ 57, 293, 44 },
		{ 57, 294, 44 },
		{ 57, 295, 44 },
		{ 57, 296, 44 },
		{ 57, 297, 44 },
		{ 57, 298, 44 },
		{ 57, 299, 44 },
		{ 57, 300, 44 },
		{ 57, 301, 44 },
		{ 57, 302, 44 },
		{ 57, 303, 44 },
		{ 57, 304, 44 },
		{ 57, 305, 44 },
		{ 57, 306, 44 },
		{ 57, 307, 44 },
		{ 57, 308, 44 },
		{ 57, 309, 44 },
		{ 57, 310, 44 },
		{ 57, 311, 44 },
		{ 57, 312, 44 },
		{ 57, 313, 44 },
		{ 57, 314, 44 },
		{ 57, 315, 44 },
		{ 57, 316, 44 },
		{ 57, 317, 44 },
		{ 57, 318, 44 },
		{ 57, 319, 44 },
		{ 57, 320, 44 },
		{ 57, 321, 44 },
		{ 57, 322, 44 },
		{ 57, 323, 44 },
		{ 57, 324, 44 },
		{ 57, 325, 44 },
		{ 57, 326, 44 },
		{ 57, 327, 44 },
		{ 57, 328, 44 },
		{ 57, 329, 44 },
		{ 57, 330, 44 },
		{ 57, 331, 44 },
		{ 58, 282, 44 },
		{ 58, 283, 44 },
		{ 58, 284, 44 },
		{ 58, 285, 44 },
		{ 58, 286, 44 },
		{ 58, 287, 44 },
		{ 58, 288, 44 },
		{ 58, 289, 44 },
		{ 58, 290, 44 },
		{ 58, 291, 44 },
		{ 58, 292, 44 },
		{ 58, 293, 44 },
		{ 58, 294, 44 },
		{ 58, 295, 44 },
		{ 58, 296, 44 },
		{ 58, 297, 44 },
		{ 58, 298, 44 },
		{ 58, 299, 44 },
		{ 58, 300, 44 },
		{ 58, 301, 44 },
		{ 58, 302, 44 },
		{ 58, 303, 44 },
		{ 58, 304, 44 },
		{ 58, 305, 44 },
		{ 58, 306, 44 },
		{ 58, 307, 44 },
		{ 58, 308, 44 },
		{ 58, 309, 44 },
		{ 58, 310, 44 },
		{ 58, 311, 44 },
		{ 58, 312, 44 },
		{ 58, 313, 44 },
		{ 58, 314, 44 },
		{ 58, 315, 44 },
		{ 58, 316, 44 },
		{ 58, 317, 44 },
		{ 58, 318, 44 },
		{ 58, 319, 44 },
		{ 58, 320, 44 },
		{ 58, 321, 44 },
		{ 58, 322, 44 },
		{ 58, 323, 44 },
		{ 58, 324, 44 },
		{ 58, 325, 44 },
		{ 58, 326, 44 },
		{ 58, 327, 44 },
		{ 58, 328, 44 },
		{ 58, 329, 44 },
		{ 58, 330, 44 },
		{ 58, 331, 44 },
		{ 59, 282, 44 },
		{ 59, 283, 44 },
		{ 59, 284, 44 },
		{ 59, 285, 44 },
		{ 59, 286, 44 },
		{ 59, 287, 44 },
		{ 59, 288, 44 },
		{ 59, 289, 44 },
		{ 59, 290, 44 },
		{ 59, 291, 44 },
		{ 59, 292, 44 },
		{ 59, 293, 44 },
		{ 59, 294, 44 },
		{ 59, 295, 44 },
		{ 59, 296, 44 },
		{ 59, 297, 44 },
		{ 59, 298, 44 },
		{ 59, 299, 44 },
		{ 59, 300, 44 },
		{ 59, 301, 44 },
		{ 59, 302, 44 },
		{ 59, 303, 44 },
		{ 59, 304, 44 },
		{ 59, 305, 44 },
		{ 59, 306, 44 },
		{ 59, 307, 44 },
		{ 59, 308, 44 },
		{ 59, 309, 44 },
		{ 59, 310, 44 },
		{ 59, 311, 44 },
		{ 59, 312, 44 },
		{ 59, 313, 44 },
		{ 59, 314, 44 },
		{ 59, 315, 44 },
		{ 59, 316, 44 },
		{ 59, 317, 44 },
		{ 59, 318, 44 },
		{ 59, 319, 44 },
		{ 59, 320, 44 },
		{ 59, 321, 44 },
		{ 59, 322, 44 },
		{ 59, 323, 44 },
		{ 59, 324, 44 },
		{ 59, 325, 44 },
		{ 59, 326, 44 },
		{ 59, 327, 44 },
		{ 59, 328, 44 },
		{ 59, 329, 44 },
		{ 59, 330, 44 },
		{ 59, 331, 44 },
		{ 60, 282, 44 },
		{ 60, 283, 44 },
		{ 60, 284, 44 },
		{ 60, 285, 44 },
		{ 60, 286, 44 },
		{ 60, 287, 44 },
		{ 60, 288, 44 },
		{ 60, 289, 44 },
		{ 60, 290, 44 },
		{ 60, 291, 44 },
		{ 60, 292, 44 },
		{ 60, 293, 44 },
		{ 60, 294, 44 },
		{ 60, 295, 44 },
		{ 60, 296, 44 },
		{ 60, 297, 44 },
		{ 60, 298, 44 },
		{ 60, 299, 44 },
		{ 60, 300, 44 },
		{ 60, 301, 44 },
		{ 60, 302, 44 },
		{ 60, 303, 44 },
		{ 60, 304, 44 },
		{ 60, 305, 44 },
		{ 60, 306, 44 },
		{ 60, 307, 44 },
		{ 60, 308, 44 },
		{ 60, 309, 44 },
		{ 60, 310, 44 },
		{ 60, 311, 44 },
		{ 60, 312, 44 },
		{ 60, 313, 44 },
		{ 60, 314, 44 },
		{ 60, 315, 44 },
		{ 60, 316, 44 },
		{ 60, 317, 44 },
		{ 60, 318, 44 },
		{ 60, 319, 44 },
		{ 60, 320, 44 },
		{ 60, 321, 44 },
		{ 60, 322, 44 },
		{ 60, 323, 44 },
		{ 60, 324, 44 },
		{ 60, 325, 44 },
		{ 60, 326, 44 },
		{ 60, 327, 44 },
		{ 60, 328, 44 },
		{ 60, 329, 44 },
		{ 60, 330, 44 },
		{ 61, 282, 44 },
		{ 61, 283, 44 },
		{ 61, 284, 44 },
		{ 61, 285, 44 },
		{ 61, 286, 44 },
		{ 61, 287, 44 },
		{ 61, 288, 44 },
		{ 61, 289, 44 },
		{ 61, 290, 44 },
		{ 61, 291, 44 },
		{ 61, 292, 44 },
		{ 61, 293, 44 },
		{ 61, 294, 44 },
		{ 61, 295, 44 },
		{ 61, 296, 44 },
		{ 61, 297, 44 },
		{ 61, 298, 44 },
		{ 61, 299, 44 },
		{ 61, 300, 44 },
		{ 61, 301, 44 },
		{ 61, 302, 44 },
		{ 61, 303, 44 },
		{ 61, 304, 44 },
		{ 61, 305, 44 },
		{ 61, 306, 44 },
		{ 61, 307, 44 },
		{ 61, 308, 44 },
		{ 61, 309, 44 },
		{ 61, 310, 44 },
		{ 61, 311, 44 },
		{ 61, 312, 44 },
		{ 61, 313, 44 },
		{ 61, 314, 44 },
		{ 61, 315, 44 },
		{ 61, 316, 44 },
		{ 61, 317, 44 },
		{ 61, 318, 44 },
		{ 61, 319, 44 },
		{ 61, 320, 44 },
		{ 61, 321, 44 },
		{ 61, 322, 44 },
		{ 61, 323, 44 },
		{ 61, 324, 44 },
		{ 61, 325, 44 },
		{ 61, 326, 44 },
		{ 61, 327, 44 },
		{ 61, 328, 44 },
		{ 61, 329, 44 },
		{ 61, 330, 44 },
		{ 62, 282, 44 },
		{ 62, 283, 44 },
		{ 62, 284, 44 },
		{ 62, 285, 44 },
		{ 62, 286, 44 },
		{ 62, 287, 44 },
		{ 62, 288, 44 },
		{ 62, 289, 44 },
		{ 62, 290, 44 },
		{ 62, 291, 44 },
		{ 62, 292, 44 },
		{ 62, 293, 44 },
		{ 62, 294, 44 },
		{ 62, 295, 44 },
		{ 62, 296, 44 },
		{ 62, 297, 44 },
		{ 62, 298, 44 },
		{ 62, 299, 44 },
		{ 62, 300, 44 },
		{ 62, 301, 44 },
		{ 62, 302, 44 },
		{ 62, 303, 44 },
		{ 62, 304, 44 },
		{ 62, 305, 44 },
		{ 62, 306, 44 },
		{ 62, 307, 44 },
		{ 62, 308, 44 },
		{ 62, 309, 44 },
		{ 62, 310, 44 },
		{ 62, 311, 44 },
		{ 62, 312, 44 },
		{ 62, 313, 44 },
		{ 62, 314, 44 },
		{ 62, 315, 44 },
		{ 62, 316, 44 },
		{ 62, 317, 44 },
		{ 62, 318, 44 },
		{ 62, 319, 44 },
		{ 62, 320, 44 },
		{ 62, 321, 44 },
		{ 62, 322, 44 },
		{ 62, 323, 44 },
		{ 62, 324, 44 },
		{ 62, 325, 44 },
		{ 62, 326, 44 },
		{ 62, 327, 44 },
		{ 62, 328, 44 },
		{ 62, 329, 44 },
		{ 62, 330, 44 },
		{ 63, 282, 44 },
		{ 63, 283, 44 },
		{ 63, 284, 44 },
		{ 63, 285, 44 },
		{ 63, 286, 44 },
		{ 63, 287, 44 },
		{ 63, 288, 44 },
		{ 63, 289, 44 },
		{ 63, 290, 44 },
		{ 63, 291, 44 },
		{ 63, 292, 44 },
		{ 63, 293, 44 },
		{ 63, 294, 44 },
		{ 63, 295, 44 },
		{ 63, 296, 44 },
		{ 63, 297, 44 },
		{ 63, 298, 44 },
		{ 63, 299, 44 },
		{ 63, 300, 44 },
		{ 63, 301, 44 },
		{ 63, 302, 44 },
		{ 63, 303, 44 },
		{ 63, 304, 44 },
		{ 63, 305, 44 },
		{ 63, 306, 44 },
		{ 63, 307, 44 },
		{ 63, 308, 44 },
		{ 63, 309, 44 },
		{ 63, 310, 44 },
		{ 63, 311, 44 },
		{ 63, 312, 44 },
		{ 63, 313, 44 },
		{ 63, 314, 44 },
		{ 63, 315, 44 },
		{ 63, 316, 44 },
		{ 63, 317, 44 },
		{ 63, 318, 44 },
		{ 63, 319, 44 },
		{ 63, 320, 44 },
		{ 63, 321, 44 },
		{ 63, 322, 44 },
		{ 63, 323, 44 },
		{ 63, 324, 44 },
		{ 63, 325, 44 },
		{ 63, 326, 44 },
		{ 63, 327, 44 },
		{ 63, 328, 44 },
		{ 63, 329, 44 },
		{ 64, 282, 44 },
		{ 64, 283, 44 },
		{ 64, 284, 44 },
		{ 64, 285, 44 },
		{ 64, 286, 44 },
		{ 64, 287, 44 },
		{ 64, 288, 44 },
		{ 64, 289, 44 },
		{ 64, 290, 44 },
		{ 64, 291, 44 },
		{ 64, 292, 44 },
		{ 64, 293, 44 },
		{ 64, 294, 44 },
		{ 64, 295, 44 },
		{ 64, 296, 44 },
		{ 64, 297, 44 },
		{ 64, 298, 44 },
		{ 64, 299, 44 },
		{ 64, 300, 44 },
		{ 64, 301, 44 },
		{ 64, 302, 44 },
		{ 64, 303, 44 },
		{ 64, 304, 44 },
		{ 64, 305, 44 },
		{ 64, 306, 44 },
		{ 64, 307, 44 },
		{ 64, 308, 44 },
		{ 64, 309, 44 },
		{ 64, 310, 44 },
		{ 64, 311, 44 },
		{ 64, 312, 44 },
		{ 64, 313, 44 },
		{ 64, 314, 44 },
		{ 64, 315, 44 },
		{ 64, 316, 44 },
		{ 64, 317, 44 },
		{ 64, 318, 44 },
		{ 64, 319, 44 },
		{ 64, 320, 44 },
		{ 64, 321, 44 },
		{ 64, 322, 44 },
		{ 64, 323, 44 },
		{ 64, 324, 44 },
		{ 64, 325, 44 },
		{ 64, 326, 44 },
		{ 64, 327, 44 },
		{ 64, 328, 44 },
		{ 64, 329, 44 },
		{ 65, 282, 44 },
		{ 65, 283, 44 },
		{ 65, 284, 44 },
		{ 65, 285, 44 },
		{ 65, 286, 44 },
		{ 65, 287, 44 },
		{ 65, 288, 44 },
		{ 65, 289, 44 },
		{ 65, 290, 44 },
		{ 65, 291, 44 },
		{ 65, 292, 44 },
		{ 65, 293, 44 },
		{ 65, 294, 44 },
		{ 65, 295, 44 },
		{ 65, 296, 44 },
		{ 65, 297, 44 },
		{ 65, 298, 44 },
		{ 65, 299, 44 },
		{ 65, 300, 44 },
		{ 65, 301, 44 },
		{ 65, 302, 44 },
		{ 65, 303, 44 },
		{ 65, 304, 44 },
		{ 65, 305, 44 },
		{ 65, 306, 44 },
		{ 65, 307, 44 },
		{ 65, 308, 44 },
		{ 65, 309, 44 },
		{ 65, 310, 44 },
		{ 65, 311, 44 },
		{ 65, 312, 44 },
		{ 65, 313, 44 },
		{ 65, 314, 44 },
		{ 65, 315, 44 },
		{ 65, 316, 44 },
		{ 65, 317, 44 },
		{ 65, 318, 44 },
		{ 65, 319, 44 },
		{ 65, 320, 44 },
		{ 65, 321, 44 },
		{ 65, 322, 44 },
		{ 65, 323, 44 },
		{ 65, 324, 44 },
		{ 65, 325, 44 },
		{ 65, 326, 44 },
		{ 65, 327, 44 },
		{ 65, 328, 44 },
		{ 66, 281, 44 },
		{ 66, 282, 44 },
		{ 66, 283, 44 },
		{ 66, 284, 44 },
		{ 66, 285, 44 },
		{ 66, 286, 44 },
		{ 66, 287, 44 },
		{ 66, 288, 44 },
		{ 66, 289, 44 },
		{ 66, 290, 44 },
		{ 66, 291, 44 },
		{ 66, 292, 44 },
		{ 66, 293, 44 },
		{ 66, 294, 44 },
		{ 66, 295, 44 },
		{ 66, 296, 44 },
		{ 66, 297, 44 },
		{ 66, 298, 44 },
		{ 66, 299, 44 },
		{ 66, 300, 44 },
		{ 66, 301, 44 },
		{ 66, 302, 44 },
		{ 66, 303, 44 },
		{ 66, 304, 44 },
		{ 66, 305, 44 },
		{ 66, 306, 44 },
		{ 66, 307, 44 },
		{ 66, 308, 44 },
		{ 66, 309, 44 },
		{ 66, 310, 44 },
		{ 66, 311, 44 },
		{ 66, 312, 44 },
		{ 66, 313, 44 },
		{ 66, 314, 44 },
		{ 66, 315, 44 },
		{ 66, 316, 44 },
		{ 66, 317, 44 },
		{ 66, 318, 44 },
		{ 66, 319, 44 },
		{ 66, 320, 44 },
		{ 66, 321, 44 },
		{ 66, 322, 44 },
		{ 66, 323, 44 },
		{ 66, 324, 44 },
		{ 66, 325, 44 },
		{ 66, 326, 44 },
		{ 66, 327, 44 },
		{ 66, 328, 44 },
		{ 67, 281, 44 },
		{ 67, 282, 44 },
		{ 67, 283, 44 },
		{ 67, 284, 44 },
		{ 67, 285, 44 },
		{ 67, 286, 44 },
		{ 67, 287, 44 },
		{ 67, 288, 44 },
		{ 67, 289, 44 },
		{ 67, 290, 44 },
		{ 67, 291, 44 },
		{ 67, 292, 44 },
		{ 67, 293, 44 },
		{ 67, 294, 44 },
		{ 67, 295, 44 },
		{ 67, 296, 44 },
		{ 67, 297, 44 },
		{ 67, 298, 44 },
		{ 67, 299, 44 },
		{ 67, 300, 44 },
		{ 67, 301, 44 },
		{ 67, 302, 44 },
		{ 67, 303, 44 },
		{ 67, 304, 44 },
		{ 67, 305, 44 },
		{ 67, 306, 44 },
		{ 67, 307, 44 },
		{ 67, 308, 44 },
		{ 67, 309, 44 },
		{ 67, 310, 44 },
		{ 67, 311, 44 },
		{ 67, 312, 44 },
		{ 67, 313, 44 },
		{ 67, 314, 44 },
		{ 67, 315, 44 },
		{ 67, 316, 44 },
		{ 67, 317, 44 },
		{ 67, 318, 44 },
		{ 67, 319, 44 },
		{ 67, 320, 44 },
		{ 67, 321, 44 },
		{ 67, 322, 44 },
		{ 67, 323, 44 },
		{ 67, 324, 44 },
		{ 67, 325, 44 },
		{ 67, 326, 44 },
		{ 67, 327, 44 },
		{ 67, 328, 44 },
		{ 68, 281, 44 },
		{ 68, 282, 44 },
		{ 68, 283, 44 },
		{ 68, 284, 44 },
		{ 68, 285, 44 },
		{ 68, 286, 44 },
		{ 68, 287, 44 },
		{ 68, 288, 44 },
		{ 68, 289, 44 },
		{ 68, 290, 44 },
		{ 68, 291, 44 },
		{ 68, 292, 44 },
		{ 68, 293, 44 },
		{ 68, 294, 44 },
		{ 68, 295, 44 },
		{ 68, 296, 44 },
		{ 68, 297, 44 },
		{ 68, 298, 44 },
		{ 68, 299, 44 },
		{ 68, 300, 44 },
		{ 68, 301, 44 },
		{ 68, 302, 44 },
		{ 68, 303, 44 },
		{ 68, 304, 44 },
		{ 68, 305, 44 },
		{ 68, 306, 44 },
		{ 68, 307, 44 },
		{ 68, 308, 44 },
		{ 68, 309, 44 },
		{ 68, 310, 44 },
		{ 68, 311, 44 },
		{ 68, 312, 44 },
		{ 68, 313, 44 },
		{ 68, 314, 44 },
		{ 68, 315, 44 },
		{ 68, 316, 44 },
		{ 68, 317, 44 },
		{ 68, 318, 44 },
		{ 68, 319, 44 },
		{ 68, 320, 44 },
		{ 68, 321, 44 },
		{ 68, 322, 44 },
		{ 68, 323, 44 },
		{ 68, 324, 44 },
		{ 68, 325, 44 },
		{ 68, 326, 44 },
		{ 68, 327, 44 },
		{ 69, 281, 44 },
		{ 69, 282, 44 },
		{ 69, 283, 44 },
		{ 69, 284, 44 },
		{ 69, 285, 44 },
		{ 69, 286, 44 },
		{ 69, 287, 44 },
		{ 69, 288, 44 },
		{ 69, 289, 44 },
		{ 69, 290, 44 },
		{ 69, 291, 44 },
		{ 69, 292, 44 },
		{ 69, 293, 44 },
		{ 69, 294, 44 },
		{ 69, 295, 44 },
		{ 69, 296, 44 },
		{ 69, 297, 44 },
		{ 69, 298, 44 },
		{ 69, 299, 44 },
		{ 69, 300, 44 },
		{ 69, 301, 44 },
		{ 69, 302, 44 },
		{ 69, 303, 44 },
		{ 69, 304, 44 },
		{ 69, 305, 44 },
		{ 69, 306, 44 },
		{ 69, 307, 44 },
		{ 69, 308, 44 },
		{ 69, 309, 44 },
		{ 69, 310, 44 },
		{ 69, 311, 44 },
		{ 69, 312, 44 },
		{ 69, 313, 44 },
		{ 69, 314, 44 },
		{ 69, 315, 44 },
		{ 69, 316, 44 },
		{ 69, 317, 44 },
		{ 69, 318, 44 },
		{ 69, 319, 44 },
		{ 69, 320, 44 },
		{ 69, 321, 44 },
		{ 69, 322, 44 },
		{ 69, 323, 44 },
		{ 69, 324, 44 },
		{ 69, 325, 44 },
		{ 69, 326, 44 },
		{ 69, 327, 44 },
		{ 70, 281, 44 },
		{ 70, 282, 44 },
		{ 70, 283, 44 },
		{ 70, 284, 44 },
		{ 70, 285, 44 },
		{ 70, 286, 44 },
		{ 70, 287, 44 },
		{ 70, 288, 44 },
		{ 70, 289, 44 },
		{ 70, 290, 44 },
		{ 70, 291, 44 },
		{ 70, 292, 44 },
		{ 70, 293, 44 },
		{ 70, 294, 44 },
		{ 70, 295, 44 },
		{ 70, 296, 44 },
		{ 70, 297, 44 },
		{ 70, 298, 44 },
		{ 70, 299, 44 },
		{ 70, 300, 44 },
		{ 70, 301, 44 },
		{ 70, 302, 44 },
		{ 70, 303, 44 },
		{ 70, 304, 44 },
		{ 70, 305, 44 },
		{ 70, 306, 44 },
		{ 70, 307, 44 },
		{ 70, 308, 44 },
		{ 70, 309, 44 },
		{ 70, 310, 44 },
		{ 70, 311, 44 },
		{ 70, 312, 44 },
		{ 70, 313, 44 },
		{ 70, 314, 44 },
		{ 70, 315, 44 },
		{ 70, 316, 44 },
		{ 70, 317, 44 },
		{ 70, 318, 44 },
		{ 70, 319, 44 },
		{ 70, 320, 44 },
		{ 70, 321, 44 },
		{ 70, 322, 44 },
		{ 70, 323, 44 },
		{ 70, 324, 44 },
		{ 70, 325, 44 },
		{ 70, 326, 44 },
		{ 70, 327, 44 },
		{ 71, 281, 44 },
		{ 71, 282, 44 },
		{ 71, 283, 44 },
		{ 71, 284, 44 },
		{ 71, 285, 44 },
		{ 71, 286, 44 },
		{ 71, 287, 44 },
		{ 71, 288, 44 },
		{ 71, 289, 44 },
		{ 71, 290, 44 },
		{ 71, 291, 44 },
		{ 71, 292, 44 },
		{ 71, 293, 44 },
		{ 71, 294, 44 },
		{ 71, 295, 44 },
		{ 71, 296, 44 },
		{ 71, 297, 44 },
		{ 71, 298, 44 },
		{ 71, 299, 44 },
		{ 71, 300, 44 },
		{ 71, 301, 44 },
		{ 71, 302, 44 },
		{ 71, 303, 44 },
		{ 71, 304, 44 },
		{ 71, 305, 44 },
		{ 71, 306, 44 },
		{ 71, 307, 44 },
		{ 71, 308, 44 },
		{ 71, 309, 44 },
		{ 71, 310, 44 },
		{ 71, 311, 44 },
		{ 71, 312, 44 },
		{ 71, 313, 44 },
		{ 71, 314, 44 },
		{ 71, 315, 44 },
		{ 71, 316, 44 },
		{ 71, 317, 44 },
		{ 71, 318, 44 },
		{ 71, 319, 44 },
		{ 71, 320, 44 },
		{ 71, 321, 44 },
		{ 71, 322, 44 },
		{ 71, 323, 44 },
		{ 71, 324, 44 },
		{ 71, 325, 44 },
		{ 71, 326, 44 },
		{ 72, 281, 44 },
		{ 72, 282, 44 },
		{ 72, 283, 44 },
		{ 72, 284, 44 },
		{ 72, 285, 44 },
		{ 72, 286, 44 },
		{ 72, 287, 44 },
		{ 72, 288, 44 },
		{ 72, 289, 44 },
		{ 72, 290, 44 },
		{ 72, 291, 44 },
		{ 72, 292, 44 },
		{ 72, 293, 44 },
		{ 72, 294, 44 },
		{ 72, 295, 44 },
		{ 72, 296, 44 },
		{ 72, 297, 44 },
		{ 72, 298, 44 },
		{ 72, 299, 44 },
		{ 72, 300, 44 },
		{ 72, 301, 44 },
		{ 72, 302, 44 },
		{ 72, 303, 44 },
		{ 72, 304, 44 },
		{ 72, 305, 44 },
		{ 72, 306, 44 },
		{ 72, 307, 44 },
		{ 72, 308, 44 },
		{ 72, 309, 44 },
		{ 72, 310, 44 },
		{ 72, 311, 44 },
		{ 72, 312, 44 },
		{ 72, 313, 44 },
		{ 72, 314, 44 },
		{ 72, 315, 44 },
		{ 72, 316, 44 },
		{ 72, 317, 44 },
		{ 72, 318, 44 },
		{ 72, 319, 44 },
		{ 72, 320, 44 },
		{ 72, 321, 44 },
		{ 72, 322, 44 },
		{ 72, 323, 44 },
		{ 72, 324, 44 },
		{ 72, 325, 44 },
		{ 72, 326, 44 },
		{ 73, 281, 44 },
		{ 73, 282, 44 },
		{ 73, 283, 44 },
		{ 73, 284, 44 },
		{ 73, 285, 44 },
		{ 73, 286, 44 },
		{ 73, 287, 44 },
		{ 73, 288, 44 },
		{ 73, 289, 44 },
		{ 73, 290, 44 },
		{ 73, 291, 44 },
		{ 73, 292, 44 },
		{ 73, 293, 44 },
		{ 73, 294, 44 },
		{ 73, 295, 44 },
		{ 73, 296, 44 },
		{ 73, 297, 44 },
		{ 73, 298, 44 },
		{ 73, 299, 44 },
		{ 73, 300, 44 },
		{ 73, 301, 44 },
		{ 73, 302, 44 },
		{ 73, 303, 44 },
		{ 73, 304, 44 },
		{ 73, 305, 44 },
		{ 73, 306, 44 },
		{ 73, 307, 44 },
		{ 73, 308, 44 },
		{ 73, 309, 44 },
		{ 73, 310, 44 },
		{ 73, 311, 44 },
		{ 73, 312, 44 },
		{ 73, 313, 44 },
		{ 73, 314, 44 },
		{ 73, 315, 44 },
		{ 73, 316, 44 },
		{ 73, 317, 44 },
		{ 73, 318, 44 },
		{ 73, 319, 44 },
		{ 73, 320, 44 },
		{ 73, 321, 44 },
		{ 73, 322, 44 },
		{ 73, 323, 44 },
		{ 73, 324, 44 },
		{ 73, 325, 44 },
		{ 74, 280, 44 },
		{ 74, 281, 44 },
		{ 74, 282, 44 },
		{ 74, 283, 44 },
		{ 74, 284, 44 },
		{ 74, 285, 44 },
		{ 74, 286, 44 },
		{ 74, 287, 44 },
		{ 74, 288, 44 },
		{ 74, 289, 44 },
		{ 74, 290, 44 },
		{ 74, 291, 44 },
		{ 74, 292, 44 },
		{ 74, 293, 44 },
		{ 74, 294, 44 },
		{ 74, 295, 44 },
		{ 74, 296, 44 },
		{ 74, 297, 44 },
		{ 74, 298, 44 },
		{ 74, 299, 44 },
		{ 74, 300, 44 },
		{ 74, 301, 44 },
		{ 74, 302, 44 },
		{ 74, 303, 44 },
		{ 74, 304, 44 },
		{ 74, 305, 44 },
		{ 74, 306, 44 },
		{ 74, 307, 44 },
		{ 74, 308, 44 },
		{ 74, 309, 44 },
		{ 74, 310, 44 },
		{ 74, 311, 44 },
		{ 74, 312, 44 },
		{ 74, 313, 44 },
		{ 74, 314, 44 },
		{ 74, 315, 44 },
		{ 74, 316, 44 },
		{ 74, 317, 44 },
		{ 74, 318, 44 },
		{ 74, 319, 44 },
		{ 74, 320, 44 },
		{ 74, 321, 44 },
		{ 74, 322, 44 },
		{ 74, 323, 44 },
		{ 74, 324, 44 },
		{ 74, 325, 44 },
		{ 75, 280, 44 },
		{ 75, 281, 44 },
		{ 75, 282, 44 },
		{ 75, 283, 44 },
		{ 75, 284, 44 },
		{ 75, 285, 44 },
		{ 75, 286, 44 },
		{ 75, 287, 44 },
		{ 75, 288, 44 },
		{ 75, 289, 44 },
		{ 75, 290, 44 },
		{ 75, 291, 44 },
		{ 75, 292, 44 },
		{ 75, 293, 44 },
		{ 75, 294, 44 },
		{ 75, 295, 44 },
		{ 75, 296, 44 },
		{ 75, 297, 44 },
		{ 75, 298, 44 },
		{ 75, 299, 44 },
		{ 75, 300, 44 },
		{ 75, 301, 44 },
		{ 75, 302, 44 },
		{ 75, 303, 44 },
		{ 75, 304, 44 },
		{ 75, 305, 44 },
		{ 75, 306, 44 },
		{ 75, 307, 44 },
		{ 75, 308, 44 },
		{ 75, 309, 44 },
		{ 75, 310, 44 },
		{ 75, 311, 44 },
		{ 75, 312, 44 },
		{ 75, 313, 44 },
		{ 75, 314, 44 },
		{ 75, 315, 44 },
		{ 75, 316, 44 },
		{ 75, 317, 44 },
		{ 75, 318, 44 },
		{ 75, 319, 44 },
		{ 75, 320, 44 },
		{ 75, 321, 44 },
		{ 75, 322, 44 },
		{ 75, 323, 44 },
		{ 75, 324, 44 },
		{ 75, 325, 44 },
		{ 76, 280, 44 },
		{ 76, 281, 44 },
		{ 76, 282, 44 },
		{ 76, 283, 44 },
		{ 76, 284, 44 },
		{ 76, 285, 44 },
		{ 76, 286, 44 },
		{ 76, 287, 44 },
		{ 76, 288, 44 },
		{ 76, 289, 44 },
		{ 76, 290, 44 },
		{ 76, 291, 44 },
		{ 76, 292, 44 },
		{ 76, 293, 44 },
		{ 76, 294, 44 },
		{ 76, 295, 44 },
		{ 76, 296, 44 },
		{ 76, 297, 44 },
		{ 76, 298, 44 },
		{ 76, 299, 44 },
		{ 76, 300, 44 },
		{ 76, 301, 44 },
		{ 76, 302, 44 },
		{ 76, 303, 44 },
		{ 76, 304, 44 },
		{ 76, 305, 44 },
		{ 76, 306, 44 },
		{ 76, 307, 44 },
		{ 76, 308, 44 },
		{ 76, 309, 44 },
		{ 76, 310, 44 },
		{ 76, 311, 44 },
		{ 76, 312, 44 },
		{ 76, 313, 44 },
		{ 76, 314, 44 },
		{ 76, 315, 44 },
		{ 76, 316, 44 },
		{ 76, 317, 44 },
		{ 76, 318, 44 },
		{ 76, 319, 44 },
		{ 76, 320, 44 },
		{ 76, 321, 44 },
		{ 76, 322, 44 },
		{ 76, 323, 44 },
		{ 76, 324, 44 },
		{ 77, 280, 44 },
		{ 77, 281, 44 },
		{ 77, 282, 44 },
		{ 77, 283, 44 },
		{ 77, 284, 44 },
		{ 77, 285, 44 },
		{ 77, 286, 44 },
		{ 77, 287, 44 },
		{ 77, 288, 44 },
		{ 77, 289, 44 },
		{ 77, 290, 44 },
		{ 77, 291, 44 },
		{ 77, 292, 44 },
		{ 77, 293, 44 },
		{ 77, 294, 44 },
		{ 77, 295, 44 },
		{ 77, 296, 44 },
		{ 77, 297, 44 },
		{ 77, 298, 44 },
		{ 77, 299, 44 },
		{ 77, 300, 44 },
		{ 77, 301, 44 },
		{ 77, 302, 44 },
		{ 77, 303, 44 },
		{ 77, 304, 44 },
		{ 77, 305, 44 },
		{ 77, 306, 44 },
		{ 77, 307, 44 },
		{ 77, 308, 44 },
		{ 77, 309, 44 },
		{ 77, 310, 44 },
		{ 77, 311, 44 },
		{ 77, 312, 44 },
		{ 77, 313, 44 },
		{ 77, 314, 44 },
		{ 77, 315, 44 },
		{ 77, 316, 44 },
		{ 77, 317, 44 },
		{ 77, 318, 44 },
		{ 77, 319, 44 },
		{ 77, 320, 44 },
		{ 77, 321, 44 },
		{ 77, 322, 44 },
		{ 77, 323, 44 },
		{ 77, 324, 44 },
		{ 78, 280, 44 },
		{ 78, 281, 44 },
		{ 78, 282, 44 },
		{ 78, 283, 44 },
		{ 78, 284, 44 },
		{ 78, 285, 44 },
		{ 78, 286, 44 },
		{ 78, 287, 44 },
		{ 78, 288, 44 },
		{ 78, 289, 44 },
		{ 78, 290, 44 },
		{ 78, 291, 44 },
		{ 78, 292, 44 },
		{ 78, 293, 44 },
		{ 78, 294, 44 },
		{ 78, 295, 44 },
		{ 78, 296, 44 },
		{ 78, 297, 44 },
		{ 78, 298, 44 },
		{ 78, 299, 44 },
		{ 78, 300, 44 },
		{ 78, 301, 44 },
		{ 78, 302, 44 },
		{ 78, 303, 44 },
		{ 78, 304, 44 },
		{ 78, 305, 44 },
		{ 78, 306, 44 },
		{ 78, 307, 44 },
		{ 78, 308, 44 },
		{ 78, 309, 44 },
		{ 78, 310, 44 },
		{ 78, 311, 44 },
		{ 78, 312, 44 },
		{ 78, 313, 44 },
		{ 78, 314, 44 },
		{ 78, 315, 44 },
		{ 78, 316, 44 },
		{ 78, 317, 44 },
		{ 78, 318, 44 },
		{ 78, 319, 44 },
		{ 78, 320, 44 },
		{ 78, 321, 44 },
		{ 78, 322, 44 },
		{ 78, 323, 44 },
		{ 78, 324, 44 },
		{ 79, 280, 44 },
		{ 79, 281, 44 },
		{ 79, 282, 44 },
		{ 79, 283, 44 },
		{ 79, 284, 44 },
		{ 79, 285, 44 },
		{ 79, 286, 44 },
		{ 79, 287, 44 },
		{ 79, 288, 44 },
		{ 79, 289, 44 },
		{ 79, 290, 44 },
		{ 79, 291, 44 },
		{ 79, 292, 44 },
		{ 79, 293, 44 },
		{ 79, 294, 44 },
		{ 79, 295, 44 },
		{ 79, 296, 44 },
		{ 79, 297, 44 },
		{ 79, 298, 44 },
		{ 79, 299, 44 },
		{ 79, 300, 44 },
		{ 79, 301, 44 },
		{ 79, 302, 44 },
		{ 79, 303, 44 },
		{ 79, 304, 44 },
		{ 79, 305, 44 },
		{ 79, 306, 44 },
		{ 79, 307, 44 },
		{ 79, 308, 44 },
		{ 79, 309, 44 },
		{ 79, 310, 44 },
		{ 79, 311, 44 },
		{ 79, 312, 44 },
		{ 79, 313, 44 },
		{ 79, 314, 44 },
		{ 79, 315, 44 },
		{ 79, 316, 44 },
		{ 79, 317, 44 },
		{ 79, 318, 44 },
		{ 79, 319, 44 },
		{ 79, 320, 44 },
		{ 79, 321, 44 },
		{ 79, 322, 44 },
		{ 79, 323, 44 },
		{ 80, 280, 44 },
		{ 80, 281, 44 },
		{ 80, 282, 44 },
		{ 80, 283, 44 },
		{ 80, 284, 44 },
		{ 80, 285, 44 },
		{ 80, 286, 44 },
		{ 80, 287, 44 },
		{ 80, 288, 44 },
		{ 80, 289, 44 },
		{ 80, 290, 44 },
		{ 80, 291, 44 },
		{ 80, 292, 44 },
		{ 80, 293, 44 },
		{ 80, 294, 44 },
		{ 80, 295, 44 },
		{ 80, 296, 44 },
		{ 80, 297, 44 },
		{ 80, 298, 44 },
		{ 80, 299, 44 },
		{ 80, 300, 44 },
		{ 80, 301, 44 },
		{ 80, 302, 44 },
		{ 80, 303, 44 },
		{ 80, 304, 44 },
		{ 80, 305, 44 },
		{ 80, 306, 44 },
		{ 80, 307, 44 },
		{ 80, 308, 44 },
		{ 80, 309, 44 },
		{ 80, 310, 44 },
		{ 80, 311, 44 },
		{ 80, 312, 44 },
		{ 80, 313, 44 },
		{ 80, 314, 44 },
		{ 80, 315, 44 },
		{ 80, 316, 44 },
		{ 80, 317, 44 },
		{ 80, 318, 44 },
		{ 80, 319, 44 },
		{ 80, 320, 44 },
		{ 80, 321, 44 },
		{ 80, 322, 44 },
		{ 80, 323, 44 },
		{ 81, 280, 44 },
		{ 81, 281, 44 },
		{ 81, 282, 44 },
		{ 81, 283, 44 },
		{ 81, 284, 44 },
		{ 81, 285, 44 },
		{ 81, 286, 44 },
		{ 81, 287, 44 },
		{ 81, 288, 44 },
		{ 81, 289, 44 },
		{ 81, 290, 44 },
		{ 81, 291, 44 },
		{ 81, 292, 44 },
		{ 81, 293, 44 },
		{ 81, 294, 44 },
		{ 81, 295, 44 },
		{ 81, 296, 44 },
		{ 81, 297, 44 },
		{ 81, 298, 44 },
		{ 81, 299, 44 },
		{ 81, 300, 44 },
		{ 81, 301, 44 },
		{ 81, 302, 44 },
		{ 81, 303, 44 },
		{ 81, 304, 44 },
		{ 81, 305, 44 },
		{ 81, 306, 44 },
		{ 81, 307, 44 },
		{ 81, 308, 44 },
		{ 81, 309, 44 },
		{ 81, 310, 44 },
		{ 81, 311, 44 },
		{ 81, 312, 44 },
		{ 81, 313, 44 },
		{ 81, 314, 44 },
		{ 81, 315, 44 },
		{ 81, 316, 44 },
		{ 81, 317, 44 },
		{ 81, 318, 44 },
		{ 81, 319, 44 },
		{ 81, 320, 44 },
		{ 81, 321, 44 },
		{ 81, 322, 44 },
		{ 82, 279, 44 },
		{ 82, 280, 44 },
		{ 82, 281, 44 },
		{ 82, 282, 44 },
		{ 82, 283, 44 },
		{ 82, 284, 44 },
		{ 82, 285, 44 },
		{ 82, 286, 44 },
		{ 82, 287, 44 },
		{ 82, 288, 44 },
		{ 82, 289, 44 },
		{ 82, 290, 44 },
		{ 82, 291, 44 },
		{ 82, 292, 44 },
		{ 82, 293, 44 },
		{ 82, 294, 44 },
		{ 82, 295, 44 },
		{ 82, 296, 44 },
		{ 82, 297, 44 },
		{ 82, 298, 44 },
		{ 82, 299, 44 },
		{ 82, 300, 44 },
		{ 82, 301, 44 },
		{ 82, 302, 44 },
		{ 82, 303, 44 },
		{ 82, 304, 44 },
		{ 82, 305, 44 },
		{ 82, 306, 44 },
		{ 82, 307, 44 },
		{ 82, 308, 44 },
		{ 82, 309, 44 },
		{ 82, 310, 44 },
		{ 82, 311, 44 },
		{ 82, 312, 44 },
		{ 82, 313, 44 },
		{ 82, 314, 44 },
		{ 82, 315, 44 },
		{ 82, 316, 44 },
		{ 82, 317, 44 },
		{ 82, 318, 44 },
		{ 82, 319, 44 },
		{ 82, 320, 44 },
		{ 82, 321, 44 },
		{ 82, 322, 44 },
		{ 83, 279, 44 },
		{ 83, 280, 44 },
		{ 83, 281, 44 },
		{ 83, 282, 44 },
		{ 83, 283, 44 },
		{ 83, 284, 44 },
		{ 83, 285, 44 },
		{ 83, 286, 44 },
		{ 83, 287, 44 },
		{ 83, 288, 44 },
		{ 83, 289, 44 },
		{ 83, 290, 44 },
		{ 83, 291, 44 },
		{ 83, 292, 44 },
		{ 83, 293, 44 },
		{ 83, 294, 44 },
		{ 83, 295, 44 },
		{ 83, 296, 44 },
		{ 83, 297, 44 },
		{ 83, 298, 44 },
		{ 83, 299, 44 },
		{ 83, 300, 44 },
		{ 83, 301, 44 },
		{ 83, 302, 44 },
		{ 83, 303, 44 },
		{ 83, 304, 44 },
		{ 83, 305, 44 },
		{ 83, 306, 44 },
		{ 83, 307, 44 },
		{ 83, 308, 44 },
		{ 83, 309, 44 },
		{ 83, 310, 44 },
		{ 83, 311, 44 },
		{ 83, 312, 44 },
		{ 83, 313, 44 },
		{ 83, 314, 44 },
		{ 83, 315, 44 },
		{ 83, 316, 44 },
		{ 83, 317, 44 },
		{ 83, 318, 44 },
		{ 83, 319, 44 },
		{ 83, 320, 44 },
		{ 83, 321, 44 },
		{ 83, 322, 44 },
		{ 84, 279, 44 },
		{ 84, 280, 44 },
		{ 84, 281, 44 },
		{ 84, 282, 44 },
		{ 84, 283, 44 },
		{ 84, 284, 44 },
		{ 84, 285, 44 },
		{ 84, 286, 44 },
		{ 84, 287, 44 },
		{ 84, 288, 44 },
		{ 84, 289, 44 },
		{ 84, 290, 44 },
		{ 84, 291, 44 },
		{ 84, 292, 44 },
		{ 84, 293, 44 },
		{ 84, 294, 44 },
		{ 84, 295, 44 },
		{ 84, 296, 44 },
		{ 84, 297, 44 },
		{ 84, 298, 44 },
		{ 84, 299, 44 },
		{ 84, 300, 44 },
		{ 84, 301, 44 },
		{ 84, 302, 44 },
		{ 84, 303, 44 },
		{ 84, 304, 44 },
		{ 84, 305, 44 },
		{ 84, 306, 44 },
		{ 84, 307, 44 },
		{ 84, 308, 44 },
		{ 84, 309, 44 },
		{ 84, 310, 44 },
		{ 84, 311, 44 },
		{ 84, 312, 44 },
		{ 84, 313, 44 },
		{ 84, 314, 44 },
		{ 84, 315, 44 },
		{ 84, 316, 44 },
		{ 84, 317, 44 },
		{ 84, 318, 44 },
		{ 84, 319, 44 },
		{ 84, 320, 44 },
		{ 84, 321, 44 },
		{ 85, 279, 44 },
		{ 85, 280, 44 },
		{ 85, 281, 44 },
		{ 85, 282, 44 },
		{ 85, 283, 44 },
		{ 85, 284, 44 },
		{ 85, 285, 44 },
		{ 85, 286, 44 },
		{ 85, 287, 44 },
		{ 85, 288, 44 },
		{ 85, 289, 44 },
		{ 85, 290, 44 },
		{ 85, 291, 44 },
		{ 85, 292, 44 },
		{ 85, 293, 44 },
		{ 85, 294, 44 },
		{ 85, 295, 44 },
		{ 85, 296, 44 },
		{ 85, 297, 44 },
		{ 85, 298, 44 },
		{ 85, 299, 44 },
		{ 85, 300, 44 },
		{ 85, 301, 44 },
		{ 85, 302, 44 },
		{ 85, 303, 44 },
		{ 85, 304, 44 },
		{ 85, 305, 44 },
		{ 85, 306, 44 },
		{ 85, 307, 44 },
		{ 85, 308, 44 },
		{ 85, 309, 44 },
		{ 85, 310, 44 },
		{ 85, 311, 44 },
		{ 85, 312, 44 },
		{ 85, 313, 44 },
		{ 85, 314, 44 },
		{ 85, 315, 44 },
		{ 85, 316, 44 },
		{ 85, 317, 44 },
		{ 85, 318, 44 },
		{ 85, 319, 44 },
		{ 85, 320, 44 },
		{ 85, 321, 44 },
		{ 86, 279, 44 },
		{ 86, 280, 44 },
		{ 86, 281, 44 },
		{ 86, 282, 44 },
		{ 86, 283, 44 },
		{ 86, 284, 44 },
		{ 86, 285, 44 },
		{ 86, 286, 44 },
		{ 86, 287, 44 },
		{ 86, 288, 44 },
		{ 86, 289, 44 },
		{ 86, 290, 44 },
		{ 86, 291, 44 },
		{ 86, 292, 44 },
		{ 86, 293, 44 },
		{ 86, 294, 44 },
		{ 86, 295, 44 },
		{ 86, 296, 44 },
		{ 86, 297, 44 },
		{ 86, 298, 44 },
		{ 86, 299, 44 },
		{ 86, 300, 44 },
		{ 86, 301, 44 },
		{ 86, 302, 44 },
		{ 86, 303, 44 },
		{ 86, 304, 44 },
		{ 86, 305, 44 },
		{ 86, 306, 44 },
		{ 86, 307, 44 },
		{ 86, 308, 44 },
		{ 86, 309, 44 },
		{ 86, 310, 44 },
		{ 86, 311, 44 },
		{ 86, 312, 44 },
		{ 86, 313, 44 },
		{ 86, 314, 44 },
		{ 86, 315, 44 },
		{ 86, 316, 44 },
		{ 86, 317, 44 },
		{ 86, 318, 44 },
		{ 86, 319, 44 },
		{ 86, 320, 44 },
		{ 86, 321, 44 },
		{ 87, 279, 44 },
		{ 87, 280, 44 },
		{ 87, 281, 44 },
		{ 87, 282, 44 },
		{ 87, 283, 44 },
		{ 87, 284, 44 },
		{ 87, 285, 44 },
		{ 87, 286, 44 },
		{ 87, 287, 44 },
		{ 87, 288, 44 },
		{ 87, 289, 44 },
		{ 87, 290, 44 },
		{ 87, 291, 44 },
		{ 87, 292, 44 },
		{ 87, 293, 44 },
		{ 87, 294, 44 },
		{ 87, 295, 44 },
		{ 87, 296, 44 },
		{ 87, 297, 44 },
		{ 87, 298, 44 },
		{ 87, 299, 44 },
		{ 87, 300, 44 },
		{ 87, 301, 44 },
		{ 87, 302, 44 },
		{ 87, 303, 44 },
		{ 87, 304, 44 },
		{ 87, 305, 44 },
		{ 87, 306, 44 },
		{ 87, 307, 44 },
		{ 87, 308, 44 },
		{ 87, 309, 44 },
		{ 87, 310, 44 },
		{ 87, 311, 44 },
		{ 87, 312, 44 },
		{ 87, 313, 44 },
		{ 87, 314, 44 },
		{ 87, 315, 44 },
		{ 87, 316, 44 },
		{ 87, 317, 44 },
		{ 87, 318, 44 },
		{ 87, 319, 44 },
		{ 87, 320, 44 },
		{ 88, 279, 44 },
		{ 88, 280, 44 },
		{ 88, 281, 44 },
		{ 88, 282, 44 },
		{ 88, 283, 44 },
		{ 88, 284, 44 },
		{ 88, 285, 44 },
		{ 88, 286, 44 },
		{ 88, 287, 44 },
		{ 88, 288, 44 },
		{ 88, 289, 44 },
		{ 88, 290, 44 },
		{ 88, 291, 44 },
		{ 88, 292, 44 },
		{ 88, 293, 44 },
		{ 88, 294, 44 },
		{ 88, 295, 44 },
		{ 88, 296, 44 },
		{ 88, 297, 44 },
		{ 88, 298, 44 },
		{ 88, 299, 44 },
		{ 88, 300, 44 },
		{ 88, 301, 44 },
		{ 88, 302, 44 },
		{ 88, 303, 44 },
		{ 88, 304, 44 },
		{ 88, 305, 44 },
		{ 88, 306, 44 },
		{ 88, 307, 44 },
		{ 88, 308, 44 },
		{ 88, 309, 44 },
		{ 88, 310, 44 },
		{ 88, 311, 44 },
		{ 88, 312, 44 },
		{ 88, 313, 44 },
		{ 88, 314, 44 },
		{ 88, 315, 44 },
		{ 88, 316, 44 },
		{ 88, 317, 44 },
		{ 88, 318, 44 },
		{ 88, 319, 44 },
		{ 88, 320, 44 },
		{ 89, 279, 44 },
		{ 89, 280, 44 },
		{ 89, 281, 44 },
		{ 89, 282, 44 },
		{ 89, 283, 44 },
		{ 89, 284, 44 },
		{ 89, 285, 44 },
		{ 89, 286, 44 },
		{ 89, 287, 44 },
		{ 89, 288, 44 },
		{ 89, 289, 44 },
		{ 89, 290, 44 },
		{ 89, 291, 44 },
		{ 89, 292, 44 },
		{ 89, 293, 44 },
		{ 89, 294, 44 },
		{ 89, 295, 44 },
		{ 89, 296, 44 },
		{ 89, 297, 44 },
		{ 89, 298, 44 },
		{ 89, 299, 44 },
		{ 89, 300, 44 },
		{ 89, 301, 44 },
		{ 89, 302, 44 },
		{ 89, 303, 44 },
		{ 89, 304, 44 },
		{ 89, 305, 44 },
		{ 89, 306, 44 },
		{ 89, 307, 44 },
		{ 89, 308, 44 },
		{ 89, 309, 44 },
		{ 89, 310, 44 },
		{ 89, 311, 44 },
		{ 89, 312, 44 },
		{ 89, 313, 44 },
		{ 89, 314, 44 },
		{ 89, 315, 44 },
		{ 89, 316, 44 },
		{ 89, 317, 44 },
		{ 89, 318, 44 },
		{ 89, 319, 44 },
		{ 90, 278, 44 },
		{ 90, 279, 44 },
		{ 90, 280, 44 },
		{ 90, 281, 44 },
		{ 90, 282, 44 },
		{ 90, 283, 44 },
		{ 90, 284, 44 },
		{ 90, 285, 44 },
		{ 90, 286, 44 },
		{ 90, 287, 44 },
		{ 90, 288, 44 },
		{ 90, 289, 44 },
		{ 90, 290, 44 },
		{ 90, 291, 44 },
		{ 90, 292, 44 },
		{ 90, 293, 44 },
		{ 90, 294, 44 },
		{ 90, 295, 44 },
		{ 90, 296, 44 },
		{ 90, 297, 44 },
		{ 90, 298, 44 },
		{ 90, 299, 44 },
		{ 90, 300, 44 },
		{ 90, 301, 44 },
		{ 90, 302, 44 },
		{ 90, 303, 44 },
		{ 90, 304, 44 },
		{ 90, 305, 44 },
		{ 90, 306, 44 },
		{ 90, 307, 44 },
		{ 90, 308, 44 },
		{ 90, 309, 44 },
		{ 90, 310, 44 },
		{ 90, 311, 44 },
		{ 90, 312, 44 },
		{ 90, 313, 44 },
		{ 90, 314, 44 },
		{ 90, 315, 44 },
		{ 90, 316, 44 },
		{ 90, 317, 44 },
		{ 90, 318, 44 },
		{ 90, 319, 44 },
		{ 91, 278, 44 },
		{ 91, 279, 44 },
		{ 91, 280, 44 },
		{ 91, 281, 44 },
		{ 91, 282, 44 },
		{ 91, 283, 44 },
		{ 91, 284, 44 },
		{ 91, 285, 44 },
		{ 91, 286, 44 },
		{ 91, 287, 44 },
		{ 91, 288, 44 },
		{ 91, 289, 44 },
		{ 91, 290, 44 },
		{ 91, 291, 44 },
		{ 91, 292, 44 },
		{ 91, 293, 44 },
		{ 91, 294, 44 },
		{ 91, 295, 44 },
		{ 91, 296, 44 },
		{ 91, 297, 44 },
		{ 91, 298, 44 },
		{ 91, 299, 44 },
		{ 91, 300, 44 },
		{ 91, 301, 44 },
		{ 91, 302, 44 },
		{ 91, 303, 44 },
		{ 91, 304, 44 },
		{ 91, 305, 44 },
		{ 91, 306, 44 },
		{ 91, 307, 44 },
		{ 91, 308, 44 },
		{ 91, 309, 44 },
		{ 91, 310, 44 },
		{ 91, 311, 44 },
		{ 91, 312, 44 },
		{ 91, 313, 44 },
		{ 91, 314, 44 },
		{ 91, 315, 44 },
		{ 91, 316, 44 },
		{ 91, 317, 44 },
		{ 91, 318, 44 },
		{ 91, 319, 44 },
		{ 92, 278, 44 },
		{ 92, 279, 44 },
		{ 92, 280, 44 },
		{ 92, 281, 44 },
		{ 92, 282, 44 },
		{ 92, 283, 44 },
		{ 92, 284, 44 },
		{ 92, 285, 44 },
		{ 92, 286, 44 },
		{ 92, 287, 44 },
		{ 92, 288, 44 },
		{ 92, 289, 44 },
		{ 92, 290, 44 },
		{ 92, 291, 44 },
		{ 92, 292, 44 },
		{ 92, 293, 44 },
		{ 92, 294, 44 },
		{ 92, 295, 44 },
		{ 92, 296, 44 },
		{ 92, 297, 44 },
		{ 92, 298, 44 },
		{ 92, 299, 44 },
		{ 92, 300, 44 },
		{ 92, 301, 44 },
		{ 92, 302, 44 },
		{ 92, 303, 44 },
		{ 92, 304, 44 },
		{ 92, 305, 44 },
		{ 92, 306, 44 },
		{ 92, 307, 44 },
		{ 92, 308, 44 },
		{ 92, 309, 44 },
		{ 92, 310, 44 },
		{ 92, 311, 44 },
		{ 92, 312, 44 },
		{ 92, 313, 44 },
		{ 92, 314, 44 },
		{ 92, 315, 44 },
		{ 92, 316, 44 },
		{ 92, 317, 44 },
		{ 92, 318, 44 },
		{ 93, 278, 44 },
		{ 93, 279, 44 },
		{ 93, 280, 44 },
		{ 93, 281, 44 },
		{ 93, 282, 44 },
		{ 93, 283, 44 },
		{ 93, 284, 44 },
		{ 93, 285, 44 },
		{ 93, 286, 44 },
		{ 93, 287, 44 },
		{ 93, 288, 44 },
		{ 93, 289, 44 },
		{ 93, 290, 44 },
		{ 93, 291, 44 },
		{ 93, 292, 44 },
		{ 93, 293, 44 },
		{ 93, 294, 44 },
		{ 93, 295, 44 },
		{ 93, 296, 44 },
		{ 93, 297, 44 },
		{ 93, 298, 44 },
		{ 93, 299, 44 },
		{ 93, 300, 44 },
		{ 93, 301, 44 },
		{ 93, 302, 44 },
		{ 93, 303, 44 },
		{ 93, 304, 44 },
		{ 93, 305, 44 },
		{ 93, 306, 44 },
		{ 93, 307, 44 },
		{ 93, 308, 44 },
		{ 93, 309, 44 },
		{ 93, 310, 44 },
		{ 93, 311, 44 },
		{ 93, 312, 44 },
		{ 93, 313, 44 },
		{ 93, 314, 44 },
		{ 93, 315, 44 },
		{ 93, 316, 44 },
		{ 93, 317, 44 },
		{ 93, 318, 44 },
		{ 94, 278, 44 },
		{ 94, 279, 44 },
		{ 94, 280, 44 },
		{ 94, 281, 44 },
		{ 94, 282, 44 },
		{ 94, 283, 44 },
		{ 94, 284, 44 },
		{ 94, 285, 44 },
		{ 94, 286, 44 },
		{ 94, 287, 44 },
		{ 94, 288, 44 },
		{ 94, 289, 44 },
		{ 94, 290, 44 },
		{ 94, 291, 44 },
		{ 94, 292, 44 },
		{ 94, 293, 44 },
		{ 94, 294, 44 },
		{ 94, 295, 44 },
		{ 94, 296, 44 },
		{ 94, 297, 44 },
		{ 94, 298, 44 },
		{ 94, 299, 44 },
		{ 94, 300, 44 },
		{ 94, 301, 44 },
		{ 94, 302, 44 },
		{ 94, 303, 44 },
		{ 94, 304, 44 },
		{ 94, 305, 44 },
		{ 94, 306, 44 },
		{ 94, 307, 44 },
		{ 94, 308, 44 },
		{ 94, 309, 44 },
		{ 94, 310, 44 },
		{ 94, 311, 44 },
		{ 94, 312, 44 },
		{ 94, 313, 44 },
		{ 94, 314, 44 },
		{ 94, 315, 44 },
		{ 94, 316, 44 },
		{ 94, 317, 44 },
		{ 94, 318, 44 },
		{ 95, 278, 44 },
		{ 95, 279, 44 },
		{ 95, 280, 44 },
		{ 95, 281, 44 },
		{ 95, 282, 44 },
		{ 95, 283, 44 },
		{ 95, 284, 44 },
		{ 95, 285, 44 },
		{ 95, 286, 44 },
		{ 95, 287, 44 },
		{ 95, 288, 44 },
		{ 95, 289, 44 },
		{ 95, 290, 44 },
		{ 95, 291, 44 },
		{ 95, 292, 44 },
		{ 95, 293, 44 },
		{ 95, 294, 44 },
		{ 95, 295, 44 },
		{ 95, 296, 44 },
		{ 95, 297, 44 },
		{ 95, 298, 44 },
		{ 95, 299, 44 },
		{ 95, 300, 44 },
		{ 95, 301, 44 },
		{ 95, 302, 44 },
		{ 95, 303, 44 },
		{ 95, 304, 44 },
		{ 95, 305, 44 },
		{ 95, 306, 44 },
		{ 95, 307, 44 },
		{ 95, 308, 44 },
		{ 95, 309, 44 },
		{ 95, 310, 44 },
		{ 95, 311, 44 },
		{ 95, 312, 44 },
		{ 95, 313, 44 },
		{ 95, 314, 44 },
		{ 95, 315, 44 },
		{ 95, 316, 44 },
		{ 95, 317, 44 },
		{ 96, 278, 44 },
		{ 96, 279, 44 },
		{ 96, 280, 44 },
		{ 96, 281, 44 },
		{ 96, 282, 44 },
		{ 96, 283, 44 },
		{ 96, 284, 44 },
		{ 96, 285, 44 },
		{ 96, 286, 44 },
		{ 96, 287, 44 },
		{ 96, 288, 44 },
		{ 96, 289, 44 },
		{ 96, 290, 44 },
		{ 96, 291, 44 },
		{ 96, 292, 44 },
		{ 96, 293, 44 },
		{ 96, 294, 44 },
		{ 96, 295, 44 },
		{ 96, 296, 44 },
		{ 96, 297, 44 },
		{ 96, 298, 44 },
		{ 96, 299, 44 },
		{ 96, 300, 44 },
		{ 96, 301, 44 },
		{ 96, 302, 44 },
		{ 96, 303, 44 },
		{ 96, 304, 44 },
		{ 96, 305, 44 },
		{ 96, 306, 44 },
		{ 96, 307, 44 },
		{ 96, 308, 44 },
		{ 96, 309, 44 },
		{ 96, 310, 44 },
		{ 96, 311, 44 },
		{ 96, 312, 44 },
		{ 96, 313, 44 },
		{ 96, 314, 44 },
		{ 96, 315, 44 },
		{ 96, 316, 44 },
		{ 96, 317, 44 },
		{ 97, 278, 44 },
		{ 97, 279, 44 },
		{ 97, 280, 44 },
		{ 97, 281, 44 },
		{ 97, 282, 44 },
		{ 97, 283, 44 },
		{ 97, 284, 44 },
		{ 97, 285, 44 },
		{ 97, 286, 44 },
		{ 97, 287, 44 },
		{ 97, 288, 44 },
		{ 97, 289, 44 },
		{ 97, 290, 44 },
		{ 97, 291, 44 },
		{ 97, 292, 44 },
		{ 97, 293, 44 },
		{ 97, 294, 44 },
		{ 97, 295, 44 },
		{ 97, 296, 44 },
		{ 97, 297, 44 },
		{ 97, 298, 44 },
		{ 97, 299, 44 },
		{ 97, 300, 44 },
		{ 97, 301, 44 },
		{ 97, 302, 44 },
		{ 97, 303, 44 },
		{ 97, 304, 44 },
		{ 97, 305, 44 },
		{ 97, 306, 44 },
		{ 97, 307, 44 },
		{ 97, 308, 44 },
		{ 97, 309, 44 },
		{ 97, 310, 44 },
		{ 97, 311, 44 },
		{ 97, 312, 44 },
		{ 97, 313, 44 },
		{ 97, 314, 44 },
		{ 97, 315, 44 },
		{ 97, 316, 44 },
		{ 98, 277, 44 },
		{ 98, 278, 44 },
		{ 98, 279, 44 },
		{ 98, 280, 44 },
		{ 98, 281, 44 },
		{ 98, 282, 44 },
		{ 98, 283, 44 },
		{ 98, 284, 44 },
		{ 98, 285, 44 },
		{ 98, 286, 44 },
		{ 98, 287, 44 },
		{ 98, 288, 44 },
		{ 98, 289, 44 },
		{ 98, 290, 44 },
		{ 98, 291, 44 },
		{ 98, 292, 44 },
		{ 98, 293, 44 },
		{ 98, 294, 44 },
		{ 98, 295, 44 },
		{ 98, 296, 44 },
		{ 98, 297, 44 },
		{ 98, 298, 44 },
		{ 98, 299, 44 },
		{ 98, 300, 44 },
		{ 98, 301, 44 },
		{ 98, 302, 44 },
		{ 98, 303, 44 },
		{ 98, 304, 44 },
		{ 98, 305, 44 },
		{ 98, 306, 44 },
		{ 98, 307, 44 },
		{ 98, 308, 44 },
		{ 98, 309, 44 },
		{ 98, 310, 44 },
		{ 98, 311, 44 },
		{ 98, 312, 44 },
		{ 98, 313, 44 },
		{ 98, 314, 44 },
		{ 98, 315, 44 },
		{ 98, 316, 44 },
		{ 99, 277, 44 },
		{ 99, 278, 44 },
		{ 99, 279, 44 },
		{ 99, 280, 44 },
		{ 99, 281, 44 },
		{ 99, 282, 44 },
		{ 99, 283, 44 },
		{ 99, 284, 44 },
		{ 99, 285, 44 },
		{ 99, 286, 44 },
		{ 99, 287, 44 },
		{ 99, 288, 44 },
		{ 99, 289, 44 },
		{ 99, 290, 44 },
		{ 99, 291, 44 },
		{ 99, 292, 44 },
		{ 99, 293, 44 },
		{ 99, 294, 44 },
		{ 99, 295, 44 },
		{ 99, 296, 44 },
		{ 99, 297, 44 },
		{ 99, 298, 44 },
		{ 99, 299, 44 },
		{ 99, 300, 44 },
		{ 99, 301, 44 },
		{ 99, 302, 44 },
		{ 99, 303, 44 },
		{ 99, 304, 44 },
		{ 99, 305, 44 },
		{ 99, 306, 44 },
		{ 99, 307, 44 },
		{ 99, 308, 44 },
		{ 99, 309, 44 },
		{ 99, 310, 44 },
		{ 99, 311, 44 },
		{ 99, 312, 44 },
		{ 99, 313, 44 },
		{ 99, 314, 44 },
		{ 99, 315, 44 },
		{ 99, 316, 44 },
		{ 100, 277, 44 },
		{ 100, 278, 44 },
		{ 100, 279, 44 },
		{ 100, 280, 44 },
		{ 100, 281, 44 },
		{ 100, 282, 44 },
		{ 100, 283, 44 },
		{ 100, 284, 44 },
		{ 100, 285, 44 },
		{ 100, 286, 44 },
		{ 100, 287, 44 },
		{ 100, 288, 44 },
		{ 100, 289, 44 },
		{ 100, 290, 44 },
		{ 100, 291, 44 },
		{ 100, 292, 44 },
		{ 100, 293, 44 },
		{ 100, 294, 44 },
		{ 100, 295, 44 },
		{ 100, 296, 44 },
		{ 100, 297, 44 },
		{ 100, 298, 44 },
		{ 100, 299, 44 },
		{ 100, 300, 44 },
		{ 100, 301, 44 },
		{ 100, 302, 44 },
		{ 100, 303, 44 },
		{ 100, 304, 44 },
		{ 100, 305, 44 },
		{ 100, 306, 44 },
		{ 100, 307, 44 },
		{ 100, 308, 44 },
		{ 100, 309, 44 },
		{ 100, 310, 44 },
		{ 100, 311, 44 },
		{ 100, 312, 44 },
		{ 100, 313, 44 },
		{ 100, 314, 44 },
		{ 100, 315, 44 },
		{ 101, 277, 44 },
		{ 101, 278, 44 },
		{ 101, 279, 44 },
		{ 101, 280, 44 },
		{ 101, 281, 44 },
		{ 101, 282, 44 },
		{ 101, 283, 44 },
		{ 101, 284, 44 },
		{ 101, 285, 44 },
		{ 101, 286, 44 },
		{ 101, 287, 44 },
		{ 101, 288, 44 },
		{ 101, 289, 44 },
		{ 101, 290, 44 },
		{ 101, 291, 44 },
		{ 101, 292, 44 },
		{ 101, 293, 44 },
		{ 101, 294, 44 },
		{ 101, 295, 44 },
		{ 101, 296, 44 },
		{ 101, 297, 44 },
		{ 101, 298, 44 },
		{ 101, 299, 44 },
		{ 101, 300, 44 },
		{ 101, 301, 44 },
		{ 101, 302, 44 },
		{ 101, 303, 44 },
		{ 101, 304, 44 },
		{ 101, 305, 44 },
		{ 101, 306, 44 },
		{ 101, 307, 44 },
		{ 101, 308, 44 },
		{ 101, 309, 44 },
		{ 101, 310, 44 },
		{ 101, 311, 44 },
		{ 101, 312, 44 },
		{ 101, 313, 44 },
		{ 101, 314, 44 },
		{ 101, 315, 44 },
		{ 102, 277, 44 },
		{ 102, 278, 44 },
		{ 102, 279, 44 },
		{ 102, 280, 44 },
		{ 102, 281, 44 },
		{ 102, 282, 44 },
		{ 102, 283, 44 },
		{ 102, 284, 44 },
		{ 102, 285, 44 },
		{ 102, 286, 44 },
		{ 102, 287, 44 },
		{ 102, 288, 44 },
		{ 102, 289, 44 },
		{ 102, 290, 44 },
		{ 102, 291, 44 },
		{ 102, 292, 44 },
		{ 102, 293, 44 },
		{ 102, 294, 44 },
		{ 102, 295, 44 },
		{ 102, 296, 44 },
		{ 102, 297, 44 },
		{ 102, 298, 44 },
		{ 102, 299, 44 },
		{ 102, 300, 44 },
		{ 102, 301, 44 },
		{ 102, 302, 44 },
		{ 102, 303, 44 },
		{ 102, 304, 44 },
		{ 102, 305, 44 },
		{ 102, 306, 44 },
		{ 102, 307, 44 },
		{ 102, 308, 44 },
		{ 102, 309, 44 },
		{ 102, 310, 44 },
		{ 102, 311, 44 },
		{ 102, 312, 44 },
		{ 102, 313, 44 },
		{ 102, 314, 44 },
		{ 102, 315, 44 },
		{ 103, 277, 44 },
		{ 103, 278, 44 },
		{ 103, 279, 44 },
		{ 103, 280, 44 },
		{ 103, 281, 44 },
		{ 103, 282, 44 },
		{ 103, 283, 44 },
		{ 103, 284, 44 },
		{ 103, 285, 44 },
		{ 103, 286, 44 },
		{ 103, 287, 44 },
		{ 103, 288, 44 },
		{ 103, 289, 44 },
		{ 103, 290, 44 },
		{ 103, 291, 44 },
		{ 103, 292, 44 },
		{ 103, 293, 44 },
		{ 103, 294, 44 },
		{ 103, 295, 44 },
		{ 103, 296, 44 },
		{ 103, 297, 44 },
		{ 103, 298, 44 },
		{ 103, 299, 44 },
		{ 103, 300, 44 },
		{ 103, 301, 44 },
		{ 103, 302, 44 },
		{ 103, 303, 44 },
		{ 103, 304, 44 },
		{ 103, 305, 44 },
		{ 103, 306, 44 },
		{ 103, 307, 44 },
		{ 103, 308, 44 },
		{ 103, 309, 44 },
		{ 103, 310, 44 },
		{ 103, 311, 44 },
		{ 103, 312, 44 },
		{ 103, 313, 44 },
		{ 103, 314, 44 },
		{ 104, 277, 44 },
		{ 104, 278, 44 },
		{ 104, 279, 44 },
		{ 104, 280, 44 },
		{ 104, 281, 44 },
		{ 104, 282, 44 },
		{ 104, 283, 44 },
		{ 104, 284, 44 },
		{ 104, 285, 44 },
		{ 104, 286, 44 },
		{ 104, 287, 44 },
		{ 104, 288, 44 },
		{ 104, 289, 44 },
		{ 104, 290, 44 },
		{ 104, 291, 44 },
		{ 104, 292, 44 },
		{ 104, 293, 44 },
		{ 104, 294, 44 },
		{ 104, 295, 44 },
		{ 104, 296, 44 },
		{ 104, 297, 44 },
		{ 104, 298, 44 },
		{ 104, 299, 44 },
		{ 104, 300, 44 },
		{ 104, 301, 44 },
		{ 104, 302, 44 },
		{ 104, 303, 44 },
		{ 104, 304, 44 },
		{ 104, 305, 44 },
		{ 104, 306, 44 },
		{ 104, 307, 44 },
		{ 104, 308, 44 },
		{ 104, 309, 44 },
		{ 104, 310, 44 },
		{ 104, 311, 44 },
		{ 104, 312, 44 },
		{ 104, 313, 44 },
		{ 104, 314, 44 },
		{ 105, 277, 44 },
		{ 105, 278, 44 },
		{ 105, 279, 44 },
		{ 105, 280, 44 },
		{ 105, 281, 44 },
		{ 105, 282, 44 },
		{ 105, 283, 44 },
		{ 105, 284, 44 },
		{ 105, 285, 44 },
		{ 105, 286, 44 },
		{ 105, 287, 44 },
		{ 105, 288, 44 },
		{ 105, 289, 44 },
		{ 105, 290, 44 },
		{ 105, 291, 44 },
		{ 105, 292, 44 },
		{ 105, 293, 44 },
		{ 105, 294, 44 },
		{ 105, 295, 44 },
		{ 105, 296, 44 },
		{ 105, 297, 44 },
		{ 105, 298, 44 },
		{ 105, 299, 44 },
		{ 105, 300, 44 },
		{ 105, 301, 44 },
		{ 105, 302, 44 },
		{ 105, 303, 44 },
		{ 105, 304, 44 },
		{ 105, 305, 44 },
		{ 105, 306, 44 },
		{ 105, 307, 44 },
		{ 105, 308, 44 },
		{ 105, 309, 44 },
		{ 105, 310, 44 },
		{ 105, 311, 44 },
		{ 105, 312, 44 },
		{ 105, 313, 44 },
		{ 106, 276, 44 },
		{ 106, 277, 44 },
		{ 106, 278, 44 },
		{ 106, 279, 44 },
		{ 106, 280, 44 },
		{ 106, 281, 44 },
		{ 106, 282, 44 },
		{ 106, 283, 44 },
		{ 106, 284, 44 },
		{ 106, 285, 44 },
		{ 106, 286, 44 },
		{ 106, 287, 44 },
		{ 106, 288, 44 },
		{ 106, 289, 44 },
		{ 106, 290, 44 },
		{ 106, 291, 44 },
		{ 106, 292, 44 },
		{ 106, 293, 44 },
		{ 106, 294, 44 },
		{ 106, 295, 44 },
		{ 106, 296, 44 },
		{ 106, 297, 44 },
		{ 106, 298, 44 },
		{ 106, 299, 44 },
		{ 106, 300, 44 },
		{ 106, 301, 44 },
		{ 106, 302, 44 },
		{ 106, 303, 44 },
		{ 106, 304, 44 },
		{ 106, 305, 44 },
		{ 106, 306, 44 },
		{ 106, 307, 44 },
		{ 106, 308, 44 },
		{ 106, 309, 44 },
		{ 106, 310, 44 },
		{ 106, 311, 44 },
		{ 106, 312, 44 },
		{ 106, 313, 44 },
		{ 107, 276, 44 },
		{ 107, 277, 44 },
		{ 107, 278, 44 },
		{ 107, 279, 44 },
		{ 107, 280, 44 },
		{ 107, 281, 44 },
		{ 107, 282, 44 },
		{ 107, 283, 44 },
		{ 107, 284, 44 },
		{ 107, 285, 44 },
		{ 107, 286, 44 },
		{ 107, 287, 44 },
		{ 107, 288, 44 },
		{ 107, 289, 44 },
		{ 107, 290, 44 },
		{ 107, 291, 44 },
		{ 107, 292, 44 },
		{ 107, 293, 44 },
		{ 107, 294, 44 },
		{ 107, 295, 44 },
		{ 107, 296, 44 },
		{ 107, 297, 44 },
		{ 107, 298, 44 },
		{ 107, 299, 44 },
		{ 107, 300, 44 },
		{ 107, 301, 44 },
		{ 107, 302, 44 },
		{ 107, 303, 44 },
		{ 107, 304, 44 },
		{ 107, 305, 44 },
		{ 107, 306, 44 },
		{ 107, 307, 44 },
		{ 107, 308, 44 },
		{ 107, 309, 44 },
		{ 107, 310, 44 },
		{ 107, 311, 44 },
		{ 107, 312, 44 },
		{ 107, 313, 44 },
		{ 108, 276, 44 },
		{ 108, 277, 44 },
		{ 108, 278, 44 },
		{ 108, 279, 44 },
		{ 108, 280, 44 },
		{ 108, 281, 44 },
		{ 108, 282, 44 },
		{ 108, 283, 44 },
		{ 108, 284, 44 },
		{ 108, 285, 44 },
		{ 108, 286, 44 },
		{ 108, 287, 44 },
		{ 108, 288, 44 },
		{ 108, 289, 44 },
		{ 108, 290, 44 },
		{ 108, 291, 44 },
		{ 108, 292, 44 },
		{ 108, 293, 44 },
		{ 108, 294, 44 },
		{ 108, 295, 44 },
		{ 108, 296, 44 },
		{ 108, 297, 44 },
		{ 108, 298, 44 },
		{ 108, 299, 44 },
		{ 108, 300, 44 },
		{ 108, 301, 44 },
		{ 108, 302, 44 },
		{ 108, 303, 44 },
		{ 108, 304, 44 },
		{ 108, 305, 44 },
		{ 108, 306, 44 },
		{ 108, 307, 44 },
		{ 108, 308, 44 },
		{ 108, 309, 44 },
		{ 108, 310, 44 },
		{ 108, 311, 44 },
		{ 108, 312, 44 },
		{ 109, 276, 44 },
		{ 109, 277, 44 },
		{ 109, 278, 44 },
		{ 109, 279, 44 },
		{ 109, 280, 44 },
		{ 109, 281, 44 },
		{ 109, 282, 44 },
		{ 109, 283, 44 },
		{ 109, 284, 44 },
		{ 109, 285, 44 },
		{ 109, 286, 44 },
		{ 109, 287, 44 },
		{ 109, 288, 44 },
		{ 109, 289, 44 },
		{ 109, 290, 44 },
		{ 109, 291, 44 },
		{ 109, 292, 44 },
		{ 109, 293, 44 },
		{ 109, 294, 44 },
		{ 109, 295, 44 },
		{ 109, 296, 44 },
		{ 109, 297, 44 },
		{ 109, 298, 44 },
		{ 109, 299, 44 },
		{ 109, 300, 44 },
		{ 109, 301, 44 },
		{ 109, 302, 44 },
		{ 109, 303, 44 },
		{ 109, 304, 44 },
		{ 109, 305, 44 },
		{ 109, 306, 44 },
		{ 109, 307, 44 },
		{ 109, 308, 44 },
		{ 109, 309, 44 },
		{ 109, 310, 44 },
		{ 109, 311, 44 },
		{ 109, 312, 44 },
		{ 110, 276, 44 },
		{ 110, 277, 44 },
		{ 110, 278, 44 },
		{ 110, 279, 44 },
		{ 110, 280, 44 },
		{ 110, 281, 44 },
		{ 110, 282, 44 },
		{ 110, 283, 44 },
		{ 110, 284, 44 },
		{ 110, 285, 44 },
		{ 110, 286, 44 },
		{ 110, 287, 44 },
		{ 110, 288, 44 },
		{ 110, 289, 44 },
		{ 110, 290, 44 },
		{ 110, 291, 44 },
		{ 110, 292, 44 },
		{ 110, 293, 44 },
		{ 110, 294, 44 },
		{ 110, 295, 44 },
		{ 110, 296, 44 },
		{ 110, 297, 44 },
		{ 110, 298, 44 },
		{ 110, 299, 44 },
		{ 110, 300, 44 },
		{ 110, 301, 44 },
		{ 110, 302, 44 },
		{ 110, 303, 44 },
		{ 110, 304, 44 },
		{ 110, 305, 44 },
		{ 110, 306, 44 },
		{ 110, 307, 44 },
		{ 110, 308, 44 },
		{ 110, 309, 44 },
		{ 110, 310, 44 },
		{ 110, 311, 44 },
		{ 110, 312, 44 },
		{ 111, 276, 44 },
		{ 111, 277, 44 },
		{ 111, 278, 44 },
		{ 111, 279, 44 },
		{ 111, 280, 44 },
		{ 111, 281, 44 },
		{ 111, 282, 44 },
		{ 111, 283, 44 },
		{ 111, 284, 44 },
		{ 111, 285, 44 },
		{ 111, 286, 44 },
		{ 111, 287, 44 },
		{ 111, 288, 44 },
		{ 111, 289, 44 },
		{ 111, 290, 44 },
		{ 111, 291, 44 },
		{ 111, 292, 44 },
		{ 111, 293, 44 },
		{ 111, 294, 44 },
		{ 111, 295, 44 },
		{ 111, 296, 44 },
		{ 111, 297, 44 },
		{ 111, 298, 44 },
		{ 111, 299, 44 },
		{ 111, 300, 44 },
		{ 111, 301, 44 },
		{ 111, 302, 44 },
		{ 111, 303, 44 },
		{ 111, 304, 44 },
		{ 111, 305, 44 },
		{ 111, 306, 44 },
		{ 111, 307, 44 },
		{ 111, 308, 44 },
		{ 111, 309, 44 },
		{ 111, 310, 44 },
		{ 111, 311, 44 },
		{ 112, 276, 44 },
		{ 112, 277, 44 },
		{ 112, 278, 44 },
		{ 112, 279, 44 },
		{ 112, 280, 44 },
		{ 112, 281, 44 },
		{ 112, 282, 44 },
		{ 112, 283, 44 },
		{ 112, 284, 44 },
		{ 112, 285, 44 },
		{ 112, 286, 44 },
		{ 112, 287, 44 },
		{ 112, 288, 44 },
		{ 112, 289, 44 },
		{ 112, 290, 44 },
		{ 112, 291, 44 },
		{ 112, 292, 44 },
		{ 112, 293, 44 },
		{ 112, 294, 44 },
		{ 112, 295, 44 },
		{ 112, 296, 44 },
		{ 112, 297, 44 },
		{ 112, 298, 44 },
		{ 112, 299, 44 },
		{ 112, 300, 44 },
		{ 112, 301, 44 },
		{ 112, 302, 44 },
		{ 112, 303, 44 },
		{ 112, 304, 44 },
		{ 112, 305, 44 },
		{ 112, 306, 44 },
		{ 112, 307, 44 },
		{ 112, 308, 44 },
		{ 112, 309, 44 },
		{ 112, 310, 44 },
		{ 112, 311, 44 },
		{ 113, 276, 44 },
		{ 113, 277, 44 },
		{ 113, 278, 44 },
		{ 113, 279, 44 },
		{ 113, 280, 44 },
		{ 113, 281, 44 },
		{ 113, 282, 44 },
		{ 113, 283, 44 },
		{ 113, 284, 44 },
		{ 113, 285, 44 },
		{ 113, 286, 44 },
		{ 113, 287, 44 },
		{ 113, 288, 44 },
		{ 113, 289, 44 },
		{ 113, 290, 44 },
		{ 113, 291, 44 },
		{ 113, 292, 44 },
		{ 113, 293, 44 },
		{ 113, 294, 44 },
		{ 113, 295, 44 },
		{ 113, 296, 44 },
		{ 113, 297, 44 },
		{ 113, 298, 44 },
		{ 113, 299, 44 },
		{ 113, 300, 44 },
		{ 113, 301, 44 },
		{ 113, 302, 44 },
		{ 113, 303, 44 },
		{ 113, 304, 44 },
		{ 113, 305, 44 },
		{ 113, 306, 44 },
		{ 113, 307, 44 },
		{ 113, 308, 44 },
		{ 113, 309, 44 },
		{ 113, 310, 44 },
		{ 114, 275, 44 },
		{ 114, 276, 44 },
		{ 114, 277, 44 },
		{ 114, 278, 44 },
		{ 114, 279, 44 },
		{ 114, 280, 44 },
		{ 114, 281, 44 },
		{ 114, 282, 44 },
		{ 114, 283, 44 },
		{ 114, 284, 44 },
		{ 114, 285, 44 },
		{ 114, 286, 44 },
		{ 114, 287, 44 },
		{ 114, 288, 44 },
		{ 114, 289, 44 },
		{ 114, 290, 44 },
		{ 114, 291, 44 },
		{ 114, 292, 44 },
		{ 114, 293, 44 },
		{ 114, 294, 44 },
		{ 114, 295, 44 },
		{ 114, 296, 44 },
		{ 114, 297, 44 },
		{ 114, 298, 44 },
		{ 114, 299, 44 },
		{ 114, 300, 44 },
		{ 114, 301, 44 },
		{ 114, 302, 44 },
		{ 114, 303, 44 },
		{ 114, 304, 44 },
		{ 114, 305, 44 },
		{ 114, 306, 44 },
		{ 114, 307, 44 },
		{ 114, 308, 44 },
		{ 114, 309, 44 },
		{ 114, 310, 44 },
		{ 115, 275, 44 },
		{ 115, 276, 44 },
		{ 115, 277, 44 },
		{ 115, 278, 44 },
		{ 115, 279, 44 },
		{ 115, 280, 44 },
		{ 115, 281, 44 },
		{ 115, 282, 44 },
		{ 115, 283, 44 },
		{ 115, 284, 44 },
		{ 115, 285, 44 },
		{ 115, 286, 44 },
		{ 115, 287, 44 },
		{ 115, 288, 44 },
		{ 115, 289, 44 },
		{ 115, 290, 44 },
		{ 115, 291, 44 },
		{ 115, 292, 44 },
		{ 115, 293, 44 },
		{ 115, 294, 44 },
		{ 115, 295, 44 },
		{ 115, 296, 44 },
		{ 115, 297, 44 },
		{ 115, 298, 44 },
		{ 115, 299, 44 },
		{ 115, 300, 44 },
		{ 115, 301, 44 },
		{ 115, 302, 44 },
		{ 115, 303, 44 },
		{ 115, 304, 44 },
		{ 115, 305, 44 },
		{ 115, 306, 44 },
		{ 115, 307, 44 },
		{ 115, 308, 44 },
		{ 115, 309, 44 },
		{ 115, 310, 44 },
		{ 116, 275, 44 },
		{ 116, 276, 44 },
		{ 116, 277, 44 },
		{ 116, 278, 44 },
		{ 116, 279, 44 },
		{ 116, 280, 44 },
		{ 116, 281, 44 },
		{ 116, 282, 44 },
		{ 116, 283, 44 },
		{ 116, 284, 44 },
		{ 116, 285, 44 },
		{ 116, 286, 44 },
		{ 116, 287, 44 },
		{ 116, 288, 44 },
		{ 116, 289, 44 },
		{ 116, 290, 44 },
		{ 116, 291, 44 },
		{ 116, 292, 44 },
		{ 116, 293, 44 },
		{ 116, 294, 44 },
		{ 116, 295, 44 },
		{ 116, 296, 44 },
		{ 116, 297, 44 },
		{ 116, 298, 44 },
		{ 116, 299, 44 },
		{ 116, 300, 44 },
		{ 116, 301, 44 },
		{ 116, 302, 44 },
		{ 116, 303, 44 },
		{ 116, 304, 44 },
		{ 116, 305, 44 },
		{ 116, 306, 44 },
		{ 116, 307, 44 },
		{ 116, 308, 44 },
		{ 116, 309, 44 },
		{ 117, 275, 44 },
		{ 117, 276, 44 },
		{ 117, 277, 44 },
		{ 117, 278, 44 },
		{ 117, 279, 44 },
		{ 117, 280, 44 },
		{ 117, 281, 44 },
		{ 117, 282, 44 },
		{ 117, 283, 44 },
		{ 117, 284, 44 },
		{ 117, 285, 44 },
		{ 117, 286, 44 },
		{ 117, 287, 44 },
		{ 117, 288, 44 },
		{ 117, 289, 44 },
		{ 117, 290, 44 },
		{ 117, 291, 44 },
		{ 117, 292, 44 },
		{ 117, 293, 44 },
		{ 117, 294, 44 },
		{ 117, 295, 44 },
		{ 117, 296, 44 },
		{ 117, 297, 44 },
		{ 117, 298, 44 },
		{ 117, 299, 44 },
		{ 117, 300, 44 },
		{ 117, 301, 44 },
		{ 117, 302, 44 },
		{ 117, 303, 44 },
		{ 117, 304, 44 },
		{ 117, 305, 44 },
		{ 117, 306, 44 },
		{ 117, 307, 44 },
		{ 117, 308, 44 },
		{ 117, 309, 44 },
		{ 118, 275, 44 },
		{ 118, 276, 44 },
		{ 118, 277, 44 },
		{ 118, 278, 44 },
		{ 118, 279, 44 },
		{ 118, 280, 44 },
		{ 118, 281, 44 },
		{ 118, 282, 44 },
		{ 118, 283, 44 },
		{ 118, 284, 44 },
		{ 118, 285, 44 },
		{ 118, 286, 44 },
		{ 118, 287, 44 },
		{ 118, 288, 44 },
		{ 118, 289, 44 },
		{ 118, 290, 44 },
		{ 118, 291, 44 },
		{ 118, 292, 44 },
		{ 118, 293, 44 },
		{ 118, 294, 44 },
		{ 118, 295, 44 },
		{ 118, 296, 44 },
		{ 118, 297, 44 },
		{ 118, 298, 44 },
		{ 118, 299, 44 },
		{ 118, 300, 44 },
		{ 118, 301, 44 },
		{ 118, 302, 44 },
		{ 118, 303, 44 },
		{ 118, 304, 44 },
		{ 118, 305, 44 },
		{ 118, 306, 44 },
		{ 118, 307, 44 },
		{ 118, 308, 44 },
		{ 118, 309, 44 },
		{ 119, 275, 44 },
		{ 119, 276, 44 },
		{ 119, 277, 44 },
		{ 119, 278, 44 },
		{ 119, 279, 44 },
		{ 119, 280, 44 },
		{ 119, 281, 44 },
		{ 119, 282, 44 },
		{ 119, 283, 44 },
		{ 119, 284, 44 },
		{ 119, 285, 44 },
		{ 119, 286, 44 },
		{ 119, 287, 44 },
		{ 119, 288, 44 },
		{ 119, 289, 44 },
		{ 119, 290, 44 },
		{ 119, 291, 44 },
		{ 119, 292, 44 },
		{ 119, 293, 44 },
		{ 119, 294, 44 },
		{ 119, 295, 44 },
		{ 119, 296, 44 },
		{ 119, 297, 44 },
		{ 119, 298, 44 },
		{ 119, 299, 44 },
		{ 119, 300, 44 },
		{ 119, 301, 44 },
		{ 119, 302, 44 },
		{ 119, 303, 44 },
		{ 119, 304, 44 },
		{ 119, 305, 44 },
		{ 119, 306, 44 },
		{ 119, 307, 44 },
		{ 119, 308, 44 },
		{ 120, 275, 44 },
		{ 120, 276, 44 },
		{ 120, 277, 44 },
		{ 120, 278, 44 },
		{ 120, 279, 44 },
		{ 120, 280, 44 },
		{ 120, 281, 44 },
		{ 120, 282, 44 },
		{ 120, 283, 44 },
		{ 120, 284, 44 },
		{ 120, 285, 44 },
		{ 120, 286, 44 },
		{ 120, 287, 44 },
		{ 120, 288, 44 },
		{ 120, 289, 44 },
		{ 120, 290, 44 },
		{ 120, 291, 44 },
		{ 120, 292, 44 },
		{ 120, 293, 44 },
		{ 120, 294, 44 },
		{ 120, 295, 44 },
		{ 120, 296, 44 },
		{ 120, 297, 44 },
		{ 120, 298, 44 },
		{ 120, 299, 44 },
		{ 120, 300, 44 },
		{ 120, 301, 44 },
		{ 120, 302, 44 },
		{ 120, 303, 44 },
		{ 120, 304, 44 },
		{ 120, 305, 44 },
		{ 120, 306, 44 },
		{ 120, 307, 44 },
		{ 120, 308, 44 },
		{ 121, 275, 44 },
		{ 121, 276, 44 },
		{ 121, 277, 44 },
		{ 121, 278, 44 },
		{ 121, 279, 44 },
		{ 121, 280, 44 },
		{ 121, 281, 44 },
		{ 121, 282, 44 },
		{ 121, 283, 44 },
		{ 121, 284, 44 },
		{ 121, 285, 44 },
		{ 121, 286, 44 },
		{ 121, 287, 44 },
		{ 121, 288, 44 },
		{ 121, 289, 44 },
		{ 121, 290, 44 },
		{ 121, 291, 44 },
		{ 121, 292, 44 },
		{ 121, 293, 44 },
		{ 121, 294, 44 },
		{ 121, 295, 44 },
		{ 121, 296, 44 },
		{ 121, 297, 44 },
		{ 121, 298, 44 },
		{ 121, 299, 44 },
		{ 121, 300, 44 },
		{ 121, 301, 44 },
		{ 121, 302, 44 },
		{ 121, 303, 44 },
		{ 121, 304, 44 },
		{ 121, 305, 44 },
		{ 121, 306, 44 },
		{ 121, 307, 44 },
		{ 122, 274, 44 },
		{ 122, 275, 44 },
		{ 122, 276, 44 },
		{ 122, 277, 44 },
		{ 122, 278, 44 },
		{ 122, 279, 44 },
		{ 122, 280, 44 },
		{ 122, 281, 44 },
		{ 122, 282, 44 },
		{ 122, 283, 44 },
		{ 122, 284, 44 },
		{ 122, 285, 44 },
		{ 122, 286, 44 },
		{ 122, 287, 44 },
		{ 122, 288, 44 },
		{ 122, 289, 44 },
		{ 122, 290, 44 },
		{ 122, 291, 44 },
		{ 122, 292, 44 },
		{ 122, 293, 44 },
		{ 122, 294, 44 },
		{ 122, 295, 44 },
		{ 122, 296, 44 },
		{ 122, 297, 44 },
		{ 122, 298, 44 },
		{ 122, 299, 44 },
		{ 122, 300, 44 },
		{ 122, 301, 44 },
		{ 122, 302, 44 },
		{ 122, 303, 44 },
		{ 122, 304, 44 },
		{ 122, 305, 44 },
		{ 122, 306, 44 },
		{ 122, 307, 44 },
		{ 123, 274, 44 },
		{ 123, 275, 44 },
		{ 123, 276, 44 },
		{ 123, 277, 44 },
		{ 123, 278, 44 },
		{ 123, 279, 44 },
		{ 123, 280, 44 },
		{ 123, 281, 44 },
		{ 123, 282, 44 },
		{ 123, 283, 44 },
		{ 123, 284, 44 },
		{ 123, 285, 44 },
		{ 123, 286, 44 },
		{ 123, 287, 44 },
		{ 123, 288, 44 },
		{ 123, 289, 44 },
		{ 123, 290, 44 },
		{ 123, 291, 44 },
		{ 123, 292, 44 },
		{ 123, 293, 44 },
		{ 123, 294, 44 },
		{ 123, 295, 44 },
		{ 123, 296, 44 },
		{ 123, 297, 44 },
		{ 123, 298, 44 },
		{ 123, 299, 44 },
		{ 123, 300, 44 },
		{ 123, 301, 44 },
		{ 123, 302, 44 },
		{ 123, 303, 44 },
		{ 123, 304, 44 },
		{ 123, 305, 44 },
		{ 123, 306, 44 },
		{ 123, 307, 44 },
		{ 124, 274, 44 },
		{ 124, 275, 44 },
		{ 124, 276, 44 },
		{ 124, 277, 44 },
		{ 124, 278, 44 },
		{ 124, 279, 44 },
		{ 124, 280, 44 },
		{ 124, 281, 44 },
		{ 124, 282, 44 },
		{ 124, 283, 44 },
		{ 124, 284, 44 },
		{ 124, 285, 44 },
		{ 124, 286, 44 },
		{ 124, 287, 44 },
		{ 124, 288, 44 },
		{ 124, 289, 44 },
		{ 124, 290, 44 },
		{ 124, 291, 44 },
		{ 124, 292, 44 },
		{ 124, 293, 44 },
		{ 124, 294, 44 },
		{ 124, 295, 44 },
		{ 124, 296, 44 },
		{ 124, 297, 44 },
		{ 124, 298, 44 },
		{ 124, 299, 44 },
		{ 124, 300, 44 },
		{ 124, 301, 44 },
		{ 124, 302, 44 },
		{ 124, 303, 44 },
		{ 124, 304, 44 },
		{ 124, 305, 44 },
		{ 124, 306, 44 },
		{ 125, 274, 44 },
		{ 125, 275, 44 },
		{ 125, 276, 44 },
		{ 125, 277, 44 },
		{ 125, 278, 44 },
		{ 125, 279, 44 },
		{ 125, 280, 44 },
		{ 125, 281, 44 },
		{ 125, 282, 44 },
		{ 125, 283, 44 },
		{ 125, 284, 44 },
		{ 125, 285, 44 },
		{ 125, 286, 44 },
		{ 125, 287, 44 },
		{ 125, 288, 44 },
		{ 125, 289, 44 },
		{ 125, 290, 44 },
		{ 125, 291, 44 },
		{ 125, 292, 44 },
		{ 125, 293, 44 },
		{ 125, 294, 44 },
		{ 125, 295, 44 },
		{ 125, 296, 44 },
		{ 125, 297, 44 },
		{ 125, 298, 44 },
		{ 125, 299, 44 },
		{ 125, 300, 44 },
		{ 125, 301, 44 },
		{ 125, 302, 44 },
		{ 125, 303, 44 },
		{ 125, 304, 44 },
		{ 125, 305, 44 },
		{ 125, 306, 44 },
		{ 126, 274, 44 },
		{ 126, 275, 44 },
		{ 126, 276, 44 },
		{ 126, 277, 44 },
		{ 126, 278, 44 },
		{ 126, 279, 44 },
		{ 126, 280, 44 },
		{ 126, 281, 44 },
		{ 126, 282, 44 },
		{ 126, 283, 44 },
		{ 126, 284, 44 },
		{ 126, 285, 44 },
		{ 126, 286, 44 },
		{ 126, 287, 44 },
		{ 126, 288, 44 },
		{ 126, 289, 44 },
		{ 126, 290, 44 },
		{ 126, 291, 44 },
		{ 126, 292, 44 },
		{ 126, 293, 44 },
		{ 126, 294, 44 },
		{ 126, 295, 44 },
		{ 126, 296, 44 },
		{ 126, 297, 44 },
		{ 126, 298, 44 },
		{ 126, 299, 44 },
		{ 126, 300, 44 },
		{ 126, 301, 44 },
		{ 126, 302, 44 },
		{ 126, 303, 44 },
		{ 126, 304, 44 },
		{ 126, 305, 44 },
		{ 126, 306, 44 },
		{ 127, 274, 44 },
		{ 127, 275, 44 },
		{ 127, 276, 44 },
		{ 127, 277, 44 },
		{ 127, 278, 44 },
		{ 127, 279, 44 },
		{ 127, 280, 44 },
		{ 127, 281, 44 },
		{ 127, 282, 44 },
		{ 127, 283, 44 },
		{ 127, 284, 44 },
		{ 127, 285, 44 },
		{ 127, 286, 44 },
		{ 127, 287, 44 },
		{ 127, 288, 44 },
		{ 127, 289, 44 },
		{ 127, 290, 44 },
		{ 127, 291, 44 },
		{ 127, 292, 44 },
		{ 127, 293, 44 },
		{ 127, 294, 44 },
		{ 127, 295, 44 },
		{ 127, 296, 44 },
		{ 127, 297, 44 },
		{ 127, 298, 44 },
		{ 127, 299, 44 },
		{ 127, 300, 44 },
		{ 127, 301, 44 },
		{ 127, 302, 44 },
		{ 127, 303, 44 },
		{ 127, 304, 44 },
		{ 127, 305, 44 },
		{ 128, 274, 44 },
		{ 128, 275, 44 },
		{ 128, 276, 44 },
		{ 128, 277, 44 },
		{ 128, 278, 44 },
		{ 128, 279, 44 },
		{ 128, 280, 44 },
		{ 128, 281, 44 },
		{ 128, 282, 44 },
		{ 128, 283, 44 },
		{ 128, 284, 44 },
		{ 128, 285, 44 },
		{ 128, 286, 44 },
		{ 128, 287, 44 },
		{ 128, 288, 44 },
		{ 128, 289, 44 },
		{ 128, 290, 44 },
		{ 128, 291, 44 },
		{ 128, 292, 44 },
		{ 128, 293, 44 },
		{ 128, 294, 44 },
		{ 128, 295, 44 },
		{ 128, 296, 44 },
		{ 128, 297, 44 },
		{ 128, 298, 44 },
		{ 128, 299, 44 },
		{ 128, 300, 44 },
		{ 128, 301, 44 },
		{ 128, 302, 44 },
		{ 128, 303, 44 },
		{ 128, 304, 44 },
		{ 128, 305, 44 },
		{ 129, 274, 44 },
		{ 129, 275, 44 },
		{ 129, 276, 44 },
		{ 129, 277, 44 },
		{ 129, 278, 44 },
		{ 129, 279, 44 },
		{ 129, 280, 44 },
		{ 129, 281, 44 },
		{ 129, 282, 44 },
		{ 129, 283, 44 },
		{ 129, 284, 44 },
		{ 129, 285, 44 },
		{ 129, 286, 44 },
		{ 129, 287, 44 },
		{ 129, 288, 44 },
		{ 129, 289, 44 },
		{ 129, 290, 44 },
		{ 129, 291, 44 },
		{ 129, 292, 44 },
		{ 129, 293, 44 },
		{ 129, 294, 44 },
		{ 129, 295, 44 },
		{ 129, 296, 44 },
		{ 129, 297, 44 },
		{ 129, 298, 44 },
		{ 129, 299, 44 },
		{ 129, 300, 44 },
		{ 129, 301, 44 },
		{ 129, 302, 44 },
		{ 129, 303, 44 },
		{ 129, 304, 44 },
		{ 130, 273, 44 },
		{ 130, 274, 44 },
		{ 130, 275, 44 },
		{ 130, 276, 44 },
		{ 130, 277, 44 },
		{ 130, 278, 44 },
		{ 130, 279, 44 },
		{ 130, 280, 44 },
		{ 130, 281, 44 },
		{ 130, 282, 44 },
		{ 130, 283, 44 },
		{ 130, 284, 44 },
		{ 130, 285, 44 },
		{ 130, 286, 44 },
		{ 130, 287, 44 },
		{ 130, 288, 44 },
		{ 130, 289, 44 },
		{ 130, 290, 44 },
		{ 130, 291, 44 },
		{ 130, 292, 44 },
		{ 130, 293, 44 },
		{ 130, 294, 44 },
		{ 130, 295, 44 },
		{ 130, 296, 44 },
		{ 130, 297, 44 },
		{ 130, 298, 44 },
		{ 130, 299, 44 },
		{ 130, 300, 44 },
		{ 130, 301, 44 },
		{ 130, 302, 44 },
		{ 130, 303, 44 },
		{ 130, 304, 44 },
		{ 131, 273, 44 },
		{ 131, 274, 44 },
		{ 131, 275, 44 },
		{ 131, 276, 44 },
		{ 131, 277, 44 },
		{ 131, 278, 44 },
		{ 131, 279, 44 },
		{ 131, 280, 44 },
		{ 131, 281, 44 },
		{ 131, 282, 44 },
		{ 131, 283, 44 },
		{ 131, 284, 44 },
		{ 131, 285, 44 },
		{ 131, 286, 44 },
		{ 131, 287, 44 },
		{ 131, 288, 44 },
		{ 131, 289, 44 },
		{ 131, 290, 44 },
		{ 131, 291, 44 },
		{ 131, 292, 44 },
		{ 131, 293, 44 },
		{ 131, 294, 44 },
		{ 131, 295, 44 },
		{ 131, 296, 44 },
		{ 131, 297, 44 },
		{ 131, 298, 44 },
		{ 131, 299, 44 },
		{ 131, 300, 44 },
		{ 131, 301, 44 },
		{ 131, 302, 44 },
		{ 131, 303, 44 },
		{ 131, 304, 44 },
		{ 132, 273, 44 },
		{ 132, 274, 44 },
		{ 132, 275, 44 },
		{ 132, 276, 44 },
		{ 132, 277, 44 },
		{ 132, 278, 44 },
		{ 132, 279, 44 },
		{ 132, 280, 44 },
		{ 132, 281, 44 },
		{ 132, 282, 44 },
		{ 132, 283, 44 },
		{ 132, 284, 44 },
		{ 132, 285, 44 },
		{ 132, 286, 44 },
		{ 132, 287, 44 },
		{ 132, 288, 44 },
		{ 132, 289, 44 },
		{ 132, 290, 44 },
		{ 132, 291, 44 },
		{ 132, 292, 44 },
		{ 132, 293, 44 },
		{ 132, 294, 44 },
		{ 132, 295, 44 },
		{ 132, 296, 44 },
		{ 132, 297, 44 },
		{ 132, 298, 44 },
		{ 132, 299, 44 },
		{ 132, 300, 44 },
		{ 132, 301, 44 },
		{ 132, 302, 44 },
		{ 132, 303, 44 },
		{ 133, 273, 44 },
		{ 133, 274, 44 },
		{ 133, 275, 44 },
		{ 133, 276, 44 },
		{ 133, 277, 44 },
		{ 133, 278, 44 },
		{ 133, 279, 44 },
		{ 133, 280, 44 },
		{ 133, 281, 44 },
		{ 133, 282, 44 },
		{ 133, 283, 44 },
		{ 133, 284, 44 },
		{ 133, 285, 44 },
		{ 133, 286, 44 },
		{ 133, 287, 44 },
		{ 133, 288, 44 },
		{ 133, 289, 44 },
		{ 133, 290, 44 },
		{ 133, 291, 44 },
		{ 133, 292, 44 },
		{ 133, 293, 44 },
		{ 133, 294, 44 },
		{ 133, 295, 44 },
		{ 133, 296, 44 },
		{ 133, 297, 44 },
		{ 133, 298, 44 },
		{ 133, 299, 44 },
		{ 133, 300, 44 },
		{ 133, 301, 44 },
		{ 133, 302, 44 },
		{ 133, 303, 44 },
		{ 134, 273, 44 },
		{ 134, 274, 44 },
		{ 134, 275, 44 },
		{ 134, 276, 44 },
		{ 134, 277, 44 },
		{ 134, 278, 44 },
		{ 134, 279, 44 },
		{ 134, 280, 44 },
		{ 134, 281, 44 },
		{ 134, 282, 44 },
		{ 134, 283, 44 },
		{ 134, 284, 44 },
		{ 134, 285, 44 },
		{ 134, 286, 44 },
		{ 134, 287, 44 },
		{ 134, 288, 44 },
		{ 134, 289, 44 },
		{ 134, 290, 44 },
		{ 134, 291, 44 },
		{ 134, 292, 44 },
		{ 134, 293, 44 },
		{ 134, 294, 44 },
		{ 134, 295, 44 },
		{ 134, 296, 44 },
		{ 134, 297, 44 },
		{ 134, 298, 44 },
		{ 134, 299, 44 },
		{ 134, 300, 44 },
		{ 134, 301, 44 },
		{ 134, 302, 44 },
		{ 134, 303, 44 },
		{ 135, 273, 44 },
		{ 135, 274, 44 },
		{ 135, 275, 44 },
		{ 135, 276, 44 },
		{ 135, 277, 44 },
		{ 135, 278, 44 },
		{ 135, 279, 44 },
		{ 135, 280, 44 },
		{ 135, 281, 44 },
		{ 135, 282, 44 },
		{ 135, 283, 44 },
		{ 135, 284, 44 },
		{ 135, 285, 44 },
		{ 135, 286, 44 },
		{ 135, 287, 44 },
		{ 135, 288, 44 },
		{ 135, 289, 44 },
		{ 135, 290, 44 },
		{ 135, 291, 44 },
		{ 135, 292, 44 },
		{ 135, 293, 44 },
		{ 135, 294, 44 },
		{ 135, 295, 44 },
		{ 135, 296, 44 },
		{ 135, 297, 44 },
		{ 135, 298, 44 },
		{ 135, 299, 44 },
		{ 135, 300, 44 },
		{ 135, 301, 44 },
		{ 135, 302, 44 },
		{ 136, 273, 44 },
		{ 136, 274, 44 },
		{ 136, 275, 44 },
		{ 136, 276, 44 },
		{ 136, 277, 44 },
		{ 136, 278, 44 },
		{ 136, 279, 44 },
		{ 136, 280, 44 },
		{ 136, 281, 44 },
		{ 136, 282, 44 },
		{ 136, 283, 44 },
		{ 136, 284, 44 },
		{ 136, 285, 44 },
		{ 136, 286, 44 },
		{ 136, 287, 44 },
		{ 136, 288, 44 },
		{ 136, 289, 44 },
		{ 136, 290, 44 },
		{ 136, 291, 44 },
		{ 136, 292, 44 },
		{ 136, 293, 44 },
		{ 136, 294, 44 },
		{ 136, 295, 44 },
		{ 136, 296, 44 },
		{ 136, 297, 44 },
		{ 136, 298, 44 },
		{ 136, 299, 44 },
		{ 136, 300, 44 },
		{ 136, 301, 44 },
		{ 136, 302, 44 },
		{ 137, 273, 44 },
		{ 137, 274, 44 },
		{ 137, 275, 44 },
		{ 137, 276, 44 },
		{ 137, 277, 44 },
		{ 137, 278, 44 },
		{ 137, 279, 44 },
		{ 137, 280, 44 },
		{ 137, 281, 44 },
		{ 137, 282, 44 },
		{ 137, 283, 44 },
		{ 137, 284, 44 },
		{ 137, 285, 44 },
		{ 137, 286, 44 },
		{ 137, 287, 44 },
		{ 137, 288, 44 },
		{ 137, 289, 44 },
		{ 137, 290, 44 },
		{ 137, 291, 44 },
		{ 137, 292, 44 },
		{ 137, 293, 44 },
		{ 137, 294, 44 },
		{ 137, 295, 44 },
		{ 137, 296, 44 },
		{ 137, 297, 44 },
		{ 137, 298, 44 },
		{ 137, 299, 44 },
		{ 137, 300, 44 },
		{ 137, 301, 44 },
		{ 138, 272, 44 },
		{ 138, 273, 44 },
		{ 138, 274, 44 },
		{ 138, 275, 44 },
		{ 138, 276, 44 },
		{ 138, 277, 44 },
		{ 138, 278, 44 },
		{ 138, 279, 44 },
		{ 138, 280, 44 },
		{ 138, 281, 44 },
		{ 138, 282, 44 },
		{ 138, 283, 44 },
		{ 138, 284, 44 },
		{ 138, 285, 44 },
		{ 138, 286, 44 },
		{ 138, 287, 44 },
		{ 138, 288, 44 },
		{ 138, 289, 44 },
		{ 138, 290, 44 },
		{ 138, 291, 44 },
		{ 138, 292, 44 },
		{ 138, 293, 44 },
		{ 138, 294, 44 },
		{ 138, 295, 44 },
		{ 138, 296, 44 },
		{ 138, 297, 44 },
		{ 138, 298, 44 },
		{ 138, 299, 44 },
		{ 138, 300, 44 },
		{ 138, 301, 44 },
		{ 139, 272, 44 },
		{ 139, 273, 44 },
		{ 139, 274, 44 },
		{ 139, 275, 44 },
		{ 139, 276, 44 },
		{ 139, 277, 44 },
		{ 139, 278, 44 },
		{ 139, 279, 44 },
		{ 139, 280, 44 },
		{ 139, 281, 44 },
		{ 139, 282, 44 },
		{ 139, 283, 44 },
		{ 139, 284, 44 },
		{ 139, 285, 44 },
		{ 139, 286, 44 },
		{ 139, 287, 44 },
		{ 139, 288, 44 },
		{ 139, 289, 44 },
		{ 139, 290, 44 },
		{ 139, 291, 44 },
		{ 139, 292, 44 },
		{ 139, 293, 44 },
		{ 139, 294, 44 },
		{ 139, 295, 44 },
		{ 139, 296, 44 },
		{ 139, 297, 44 },
		{ 139, 298, 44 },
		{ 139, 299, 44 },
		{ 139, 300, 44 },
		{ 139, 301, 44 },
		{ 140, 272, 44 },
		{ 140, 273, 44 },
		{ 140, 274, 44 },
		{ 140, 275, 44 },
		{ 140, 276, 44 },
		{ 140, 277, 44 },
		{ 140, 278, 44 },
		{ 140, 279, 44 },
		{ 140, 280, 44 },
		{ 140, 281, 44 },
		{ 140, 282, 44 },
		{ 140, 283, 44 },
		{ 140, 284, 44 },
		{ 140, 285, 44 },
		{ 140, 286, 44 },
		{ 140, 287, 44 },
		{ 140, 288, 44 },
		{ 140, 289, 44 },
		{ 140, 290, 44 },
		{ 140, 291, 44 },
		{ 140, 292, 44 },
		{ 140, 293, 44 },
		{ 140, 294, 44 },
		{ 140, 295, 44 },
		{ 140, 296, 44 },
		{ 140, 297, 44 },
		{ 140, 298, 44 },
		{ 140, 299, 44 },
		{ 140, 300, 44 },
		{ 141, 272, 44 },
		{ 141, 273, 44 },
		{ 141, 274, 44 },
		{ 141, 275, 44 },
		{ 141, 276, 44 },
		{ 141, 277, 44 },
		{ 141, 278, 44 },
		{ 141, 279, 44 },
		{ 141, 280, 44 },
		{ 141, 281, 44 },
		{ 141, 282, 44 },
		{ 141, 283, 44 },
		{ 141, 284, 44 },
		{ 141, 285, 44 },
		{ 141, 286, 44 },
		{ 141, 287, 44 },
		{ 141, 288, 44 },
		{ 141, 289, 44 },
		{ 141, 290, 44 },
		{ 141, 291, 44 },
		{ 141, 292, 44 },
		{ 141, 293, 44 },
		{ 141, 294, 44 },
		{ 141, 295, 44 },
		{ 141, 296, 44 },
		{ 141, 297, 44 },
		{ 141, 298, 44 },
		{ 141, 299, 44 },
		{ 141, 300, 44 },
		{ 142, 272, 44 },
		{ 142, 273, 44 },
		{ 142, 274, 44 },
		{ 142, 275, 44 },
		{ 142, 276, 44 },
		{ 142, 277, 44 },
		{ 142, 278, 44 },
		{ 142, 279, 44 },
		{ 142, 280, 44 },
		{ 142, 281, 44 },
		{ 142, 282, 44 },
		{ 142, 283, 44 },
		{ 142, 284, 44 },
		{ 142, 285, 44 },
		{ 142, 286, 44 },
		{ 142, 287, 44 },
		{ 142, 288, 44 },
		{ 142, 289, 44 },
		{ 142, 290, 44 },
		{ 142, 291, 44 },
		{ 142, 292, 44 },
		{ 142, 293, 44 },
		{ 142, 294, 44 },
		{ 142, 295, 44 },
		{ 142, 296, 44 },
		{ 142, 297, 44 },
		{ 142, 298, 44 },
		{ 142, 299, 44 },
		{ 142, 300, 44 },
		{ 143, 272, 44 },
		{ 143, 273, 44 },
		{ 143, 274, 44 },
		{ 143, 275, 44 },
		{ 143, 276, 44 },
		{ 143, 277, 44 },
		{ 143, 278, 44 },
		{ 143, 279, 44 },
		{ 143, 280, 44 },
		{ 143, 281, 44 },
		{ 143, 282, 44 },
		{ 143, 283, 44 },
		{ 143, 284, 44 },
		{ 143, 285, 44 },
		{ 143, 286, 44 },
		{ 143, 287, 44 },
		{ 143, 288, 44 },
		{ 143, 289, 44 },
		{ 143, 290, 44 },
		{ 143, 291, 44 },
		{ 143, 292, 44 },
		{ 143, 293, 44 },
		{ 143, 294, 44 },
		{ 143, 295, 44 },
		{ 143, 296, 44 },
		{ 143, 297, 44 },
		{ 143, 298, 44 },
		{ 143, 299, 44 },
		{ 144, 272, 44 },
		{ 144, 273, 44 },
		{ 144, 274, 44 },
		{ 144, 275, 44 },
		{ 144, 276, 44 },
		{ 144, 277, 44 },
		{ 144, 278, 44 },
		{ 144, 279, 44 },
		{ 144, 280, 44 },
		{ 144, 281, 44 },
		{ 144, 282, 44 },
		{ 144, 283, 44 },
		{ 144, 284, 44 },
		{ 144, 285, 44 },
		{ 144, 286, 44 },
		{ 144, 287, 44 },
		{ 144, 288, 44 },
		{ 144, 289, 44 },
		{ 144, 290, 44 },
		{ 144, 291, 44 },
		{ 144, 292, 44 },
		{ 144, 293, 44 },
		{ 144, 294, 44 },
		{ 144, 295, 44 },
		{ 144, 296, 44 },
		{ 144, 297, 44 },
		{ 144, 298, 44 },
		{ 144, 299, 44 },
		{ 145, 272, 44 },
		{ 145, 273, 44 },
		{ 145, 274, 44 },
		{ 145, 275, 44 },
		{ 145, 276, 44 },
		{ 145, 277, 44 },
		{ 145, 278, 44 },
		{ 145, 279, 44 },
		{ 145, 280, 44 },
		{ 145, 281, 44 },
		{ 145, 282, 44 },
		{ 145, 283, 44 },
		{ 145, 284, 44 },
		{ 145, 285, 44 },
		{ 145, 286, 44 },
		{ 145, 287, 44 },
		{ 145, 288, 44 },
		{ 145, 289, 44 },
		{ 145, 290, 44 },
		{ 145, 291, 44 },
		{ 145, 292, 44 },
		{ 145, 293, 44 },
		{ 145, 294, 44 },
		{ 145, 295, 44 },
		{ 145, 296, 44 },
		{ 145, 297, 44 },
		{ 145, 298, 44 },
		{ 146, 271, 44 },
		{ 146, 272, 44 },
		{ 146, 273, 44 },
		{ 146, 274, 44 },
		{ 146, 275, 44 },
		{ 146, 276, 44 },
		{ 146, 277, 44 },
		{ 146, 278, 44 },
		{ 146, 279, 44 },
		{ 146, 280, 44 },
		{ 146, 281, 44 },
		{ 146, 282, 44 },
		{ 146, 283, 44 },
		{ 146, 284, 44 },
		{ 146, 285, 44 },
		{ 146, 286, 44 },
		{ 146, 287, 44 },
		{ 146, 288, 44 },
		{ 146, 289, 44 },
		{ 146, 290, 44 },
		{ 146, 291, 44 },
		{ 146, 292, 44 },
		{ 146, 293, 44 },
		{ 146, 294, 44 },
		{ 146, 295, 44 },
		{ 146, 296, 44 },
		{ 146, 297, 44 },
		{ 146, 298, 44 },
		{ 147, 271, 44 },
		{ 147, 272, 44 },
		{ 147, 273, 44 },
		{ 147, 274, 44 },
		{ 147, 275, 44 },
		{ 147, 276, 44 },
		{ 147, 277, 44 },
		{ 147, 278, 44 },
		{ 147, 279, 44 },
		{ 147, 280, 44 },
		{ 147, 281, 44 },
		{ 147, 282, 44 },
		{ 147, 283, 44 },
		{ 147, 284, 44 },
		{ 147, 285, 44 },
		{ 147, 286, 44 },
		{ 147, 287, 44 },
		{ 147, 288, 44 },
		{ 147, 289, 44 },
		{ 147, 290, 44 },
		{ 147, 291, 44 },
		{ 147, 292, 44 },
		{ 147, 293, 44 },
		{ 147, 294, 44 },
		{ 147, 295, 44 },
		{ 147, 296, 44 },
		{ 147, 297, 44 },
		{ 147, 298, 44 },
		{ 148, 271, 44 },
		{ 148, 272, 44 },
		{ 148, 273, 44 },
		{ 148, 274, 44 },
		{ 148, 275, 44 },
		{ 148, 276, 44 },
		{ 148, 277, 44 },
		{ 148, 278, 44 },
		{ 148, 279, 44 },
		{ 148, 280, 44 },
		{ 148, 281, 44 },
		{ 148, 282, 44 },
		{ 148, 283, 44 },
		{ 148, 284, 44 },
		{ 148, 285, 44 },
		{ 148, 286, 44 },
		{ 148, 287, 44 },
		{ 148, 288, 44 },
		{ 148, 289, 44 },
		{ 148, 290, 44 },
		{ 148, 291, 44 },
		{ 148, 292, 44 },
		{ 148, 293, 44 },
		{ 148, 294, 44 },
		{ 148, 295, 44 },
		{ 148, 296, 44 },
		{ 148, 297, 44 },
		{ 149, 271, 44 },
		{ 149, 272, 44 },
		{ 149, 273, 44 },
		{ 149, 274, 44 },
		{ 149, 275, 44 },
		{ 149, 276, 44 },
		{ 149, 277, 44 },
		{ 149, 278, 44 },
		{ 149, 279, 44 },
		{ 149, 280, 44 },
		{ 149, 281, 44 },
		{ 149, 282, 44 },
		{ 149, 283, 44 },
		{ 149, 284, 44 },
		{ 149, 285, 44 },
		{ 149, 286, 44 },
		{ 149, 287, 44 },
		{ 149, 288, 44 },
		{ 149, 289, 44 },
		{ 149, 290, 44 },
		{ 149, 291, 44 },
		{ 149, 292, 44 },
		{ 149, 293, 44 },
		{ 149, 294, 44 },
		{ 149, 295, 44 },
		{ 149, 296, 44 },
		{ 149, 297, 44 },
		{ 150, 271, 44 },
		{ 150, 272, 44 },
		{ 150, 273, 44 },
		{ 150, 274, 44 },
		{ 150, 275, 44 },
		{ 150, 276, 44 },
		{ 150, 277, 44 },
		{ 150, 278, 44 },
		{ 150, 279, 44 },
		{ 150, 280, 44 },
		{ 150, 281, 44 },
		{ 150, 282, 44 },
		{ 150, 283, 44 },
		{ 150, 284, 44 },
		{ 150, 285, 44 },
		{ 150, 286, 44 },
		{ 150, 287, 44 },
		{ 150, 288, 44 },
		{ 150, 289, 44 },
		{ 150, 290, 44 },
		{ 150, 291, 44 },
		{ 150, 292, 44 },
		{ 150, 293, 44 },
		{ 150, 294, 44 },
		{ 150, 295, 44 },
		{ 150, 296, 44 },
		{ 150, 297, 44 },
		{ 151, 271, 44 },
		{ 151, 272, 44 },
		{ 151, 273, 44 },
		{ 151, 274, 44 },
		{ 151, 275, 44 },
		{ 151, 276, 44 },
		{ 151, 277, 44 },
		{ 151, 278, 44 },
		{ 151, 279, 44 },
		{ 151, 280, 44 },
		{ 151, 281, 44 },
		{ 151, 282, 44 },
		{ 151, 283, 44 },
		{ 151, 284, 44 },
		{ 151, 285, 44 },
		{ 151, 286, 44 },
		{ 151, 287, 44 },
		{ 151, 288, 44 },
		{ 151, 289, 44 },
		{ 151, 290, 44 },
		{ 151, 291, 44 },
		{ 151, 292, 44 },
		{ 151, 293, 44 },
		{ 151, 294, 44 },
		{ 151, 295, 44 },
		{ 151, 296, 44 },
		{ 152, 271, 44 },
		{ 152, 272, 44 },
		{ 152, 273, 44 },
		{ 152, 274, 44 },
		{ 152, 275, 44 },
		{ 152, 276, 44 },
		{ 152, 277, 44 },
		{ 152, 278, 44 },
		{ 152, 279, 44 },
		{ 152, 280, 44 },
		{ 152, 281, 44 },
		{ 152, 282, 44 },
		{ 152, 283, 44 },
		{ 152, 284, 44 },
		{ 152, 285, 44 },
		{ 152, 286, 44 },
		{ 152, 287, 44 },
		{ 152, 288, 44 },
		{ 152, 289, 44 },
		{ 152, 290, 44 },
		{ 152, 291, 44 },
		{ 152, 292, 44 },
		{ 152, 293, 44 },
		{ 152, 294, 44 },
		{ 152, 295, 44 },
		{ 152, 296, 44 },
		{ 153, 271, 44 },
		{ 153, 272, 44 },
		{ 153, 273, 44 },
		{ 153, 274, 44 },
		{ 153, 275, 44 },
		{ 153, 276, 44 },
		{ 153, 277, 44 },
		{ 153, 278, 44 },
		{ 153, 279, 44 },
		{ 153, 280, 44 },
		{ 153, 281, 44 },
		{ 153, 282, 44 },
		{ 153, 283, 44 },
		{ 153, 284, 44 },
		{ 153, 285, 44 },
		{ 153, 286, 44 },
		{ 153, 287, 44 },
		{ 153, 288, 44 },
		{ 153, 289, 44 },
		{ 153, 290, 44 },
		{ 153, 291, 44 },
		{ 153, 292, 44 },
		{ 153, 293, 44 },
		{ 153, 294, 44 },
		{ 153, 295, 44 },
		{ 154, 270, 44 },
		{ 154, 271, 44 },
		{ 154, 272, 44 },
		{ 154, 273, 44 },
		{ 154, 274, 44 },
		{ 154, 275, 44 },
		{ 154, 276, 44 },
		{ 154, 277, 44 },
		{ 154, 278, 44 },
		{ 154, 279, 44 },
		{ 154, 280, 44 },
		{ 154, 281, 44 },
		{ 154, 282, 44 },
		{ 154, 283, 44 },
		{ 154, 284, 44 },
		{ 154, 285, 44 },
		{ 154, 286, 44 },
		{ 154, 287, 44 },
		{ 154, 288, 44 },
		{ 154, 289, 44 },
		{ 154, 290, 44 },
		{ 154, 291, 44 },
		{ 154, 292, 44 },
		{ 154, 293, 44 },
		{ 154, 294, 44 },
		{ 154, 295, 44 },
		{ 155, 270, 44 },
		{ 155, 271, 44 },
		{ 155, 272, 44 },
		{ 155, 273, 44 },
		{ 155, 274, 44 },
		{ 155, 275, 44 },
		{ 155, 276, 44 },
		{ 155, 277, 44 },
		{ 155, 278, 44 },
		{ 155, 279, 44 },
		{ 155, 280, 44 },
		{ 155, 281, 44 },
		{ 155, 282, 44 },
		{ 155, 283, 44 },
		{ 155, 284, 44 },
		{ 155, 285, 44 },
		{ 155, 286, 44 },
		{ 155, 287, 44 },
		{ 155, 288, 44 },
		{ 155, 289, 44 },
		{ 155, 290, 44 },
		{ 155, 291, 44 },
		{ 155, 292, 44 },
		{ 155, 293, 44 },
		{ 155, 294, 44 },
		{ 155, 295, 44 },
		{ 156, 270, 44 },
		{ 156, 271, 44 },
		{ 156, 272, 44 },
		{ 156, 273, 44 },
		{ 156, 274, 44 },
		{ 156, 275, 44 },
		{ 156, 276, 44 },
		{ 156, 277, 44 },
		{ 156, 278, 44 },
		{ 156, 279, 44 },
		{ 156, 280, 44 },
		{ 156, 281, 44 },
		{ 156, 282, 44 },
		{ 156, 283, 44 },
		{ 156, 284, 44 },
		{ 156, 285, 44 },
		{ 156, 286, 44 },
		{ 156, 287, 44 },
		{ 156, 288, 44 },
		{ 156, 289, 44 },
		{ 156, 290, 44 },
		{ 156, 291, 44 },
		{ 156, 292, 44 },
		{ 156, 293, 44 },
		{ 156, 294, 44 },
		{ 157, 270, 44 },
		{ 157, 271, 44 },
		{ 157, 272, 44 },
		{ 157, 273, 44 },
		{ 157, 274, 44 },
		{ 157, 275, 44 },
		{ 157, 276, 44 },
		{ 157, 277, 44 },
		{ 157, 278, 44 },
		{ 157, 279, 44 },
		{ 157, 280, 44 },
		{ 157, 281, 44 },
		{ 157, 282, 44 },
		{ 157, 283, 44 },
		{ 157, 284, 44 },
		{ 157, 285, 44 },
		{ 157, 286, 44 },
		{ 157, 287, 44 },
		{ 157, 288, 44 },
		{ 157, 289, 44 },
		{ 157, 290, 44 },
		{ 157, 291, 44 },
		{ 157, 292, 44 },
		{ 157, 293, 44 },
		{ 157, 294, 44 },
		{ 158, 270, 44 },
		{ 158, 271, 44 },
		{ 158, 272, 44 },
		{ 158, 273, 44 },
		{ 158, 274, 44 },
		{ 158, 275, 44 },
		{ 158, 276, 44 },
		{ 158, 277, 44 },
		{ 158, 278, 44 },
		{ 158, 279, 44 },
		{ 158, 280, 44 },
		{ 158, 281, 44 },
		{ 158, 282, 44 },
		{ 158, 283, 44 },
		{ 158, 284, 44 },
		{ 158, 285, 44 },
		{ 158, 286, 44 },
		{ 158, 287, 44 },
		{ 158, 288, 44 },
		{ 158, 289, 44 },
		{ 158, 290, 44 },
		{ 158, 291, 44 },
		{ 158, 292, 44 },
		{ 158, 293, 44 },
		{ 158, 294, 44 },
		{ 159, 270, 44 },
		{ 159, 271, 44 },
		{ 159, 272, 44 },
		{ 159, 273, 44 },
		{ 159, 274, 44 },
		{ 159, 275, 44 },
		{ 159, 276, 44 },
		{ 159, 277, 44 },
		{ 159, 278, 44 },
		{ 159, 279, 44 },
		{ 159, 280, 44 },
		{ 159, 281, 44 },
		{ 159, 282, 44 },
		{ 159, 283, 44 },
		{ 159, 284, 44 },
		{ 159, 285, 44 },
		{ 159, 286, 44 },
		{ 159, 287, 44 },
		{ 159, 288, 44 },
		{ 159, 289, 44 },
		{ 159, 290, 44 },
		{ 159, 291, 44 },
		{ 159, 292, 44 },
		{ 159, 293, 44 },
		{ 160, 270, 44 },
		{ 160, 271, 44 },
		{ 160, 272, 44 },
		{ 160, 273, 44 },
		{ 160, 274, 44 },
		{ 160, 275, 44 },
		{ 160, 276, 44 },
		{ 160, 277, 44 },
		{ 160, 278, 44 },
		{ 160, 279, 44 },
		{ 160, 280, 44 },
		{ 160, 281, 44 },
		{ 160, 282, 44 },
		{ 160, 283, 44 },
		{ 160, 284, 44 },
		{ 160, 285, 44 },
		{ 160, 286, 44 },
		{ 160, 287, 44 },
		{ 160, 288, 44 },
		{ 160, 289, 44 },
		{ 160, 290, 44 },
		{ 160, 291, 44 },
		{ 160, 292, 44 },
		{ 160, 293, 44 },
		{ 161, 270, 44 },
		{ 161, 271, 44 },
		{ 161, 272, 44 },
		{ 161, 273, 44 },
		{ 161, 274, 44 },
		{ 161, 275, 44 },
		{ 161, 276, 44 },
		{ 161, 277, 44 },
		{ 161, 278, 44 },
		{ 161, 279, 44 },
		{ 161, 280, 44 },
		{ 161, 281, 44 },
		{ 161, 282, 44 },
		{ 161, 283, 44 },
		{ 161, 284, 44 },
		{ 161, 285, 44 },
		{ 161, 286, 44 },
		{ 161, 287, 44 },
		{ 161, 288, 44 },
		{ 161, 289, 44 },
		{ 161, 290, 44 },
		{ 161, 291, 44 },
		{ 161, 292, 44 },
		{ 162, 269, 44 },
		{ 162, 270, 44 },
		{ 162, 271, 44 },
		{ 162, 272, 44 },
		{ 162, 273, 44 },
		{ 162, 274, 44 },
		{ 162, 275, 44 },
		{ 162, 276, 44 },
		{ 162, 277, 44 },
		{ 162, 278, 44 },
		{ 162, 279, 44 },
		{ 162, 280, 44 },
		{ 162, 281, 44 },
		{ 162, 282, 44 },
		{ 162, 283, 44 },
		{ 162, 284, 44 },
		{ 162, 285, 44 },
		{ 162, 286, 44 },
		{ 162, 287, 44 },
		{ 162, 288, 44 },
		{ 162, 289, 44 },
		{ 162, 290, 44 },
		{ 162, 291, 44 },
		{ 162, 292, 44 },
		{ 163, 269, 44 },
		{ 163, 270, 44 },
		{ 163, 271, 44 },
		{ 163, 272, 44 },
		{ 163, 273, 44 },
		{ 163, 274, 44 },
		{ 163, 275, 44 },
		{ 163, 276, 44 },
		{ 163, 277, 44 },
		{ 163, 278, 44 },
		{ 163, 279, 44 },
		{ 163, 280, 44 },
		{ 163, 281, 44 },
		{ 163, 282, 44 },
		{ 163, 283, 44 },
		{ 163, 284, 44 },
		{ 163, 285, 44 },
		{ 163, 286, 44 },
		{ 163, 287, 44 },
		{ 163, 288, 44 },
		{ 163, 289, 44 },
		{ 163, 290, 44 },
		{ 163, 291, 44 },
		{ 163, 292, 44 },
		{ 164, 269, 44 },
		{ 164, 270, 44 },
		{ 164, 271, 44 },
		{ 164, 272, 44 },
		{ 164, 273, 44 },
		{ 164, 274, 44 },
		{ 164, 275, 44 },
		{ 164, 276, 44 },
		{ 164, 277, 44 },
		{ 164, 278, 44 },
		{ 164, 279, 44 },
		{ 164, 280, 44 },
		{ 164, 281, 44 },
		{ 164, 282, 44 },
		{ 164, 283, 44 },
		{ 164, 284, 44 },
		{ 164, 285, 44 },
		{ 164, 286, 44 },
		{ 164, 287, 44 },
		{ 164, 288, 44 },
		{ 164, 289, 44 },
		{ 164, 290, 44 },
		{ 164, 291, 44 },
		{ 165, 269, 44 },
		{ 165, 270, 44 },
		{ 165, 271, 44 },
		{ 165, 272, 44 },
		{ 165, 273, 44 },
		{ 165, 274, 44 },
		{ 165, 275, 44 },
		{ 165, 276, 44 },
		{ 165, 277, 44 },
		{ 165, 278, 44 },
		{ 165, 279, 44 },
		{ 165, 280, 44 },
		{ 165, 281, 44 },
		{ 165, 282, 44 },
		{ 165, 283, 44 },
		{ 165, 284, 44 },
		{ 165, 285, 44 },
		{ 165, 286, 44 },
		{ 165, 287, 44 },
		{ 165, 288, 44 },
		{ 165, 289, 44 },
		{ 165, 290, 44 },
		{ 165, 291, 44 },
		{ 166, 269, 44 },
		{ 166, 270, 44 },
		{ 166, 271, 44 },
		{ 166, 272, 44 },
		{ 166, 273, 44 },
		{ 166, 274, 44 },
		{ 166, 275, 44 },
		{ 166, 276, 44 },
		{ 166, 277, 44 },
		{ 166, 278, 44 },
		{ 166, 279, 44 },
		{ 166, 280, 44 },
		{ 166, 281, 44 },
		{ 166, 282, 44 },
		{ 166, 283, 44 },
		{ 166, 284, 44 },
		{ 166, 285, 44 },
		{ 166, 286, 44 },
		{ 166, 287, 44 },
		{ 166, 288, 44 },
		{ 166, 289, 44 },
		{ 166, 290, 44 },
		{ 166, 291, 44 },
		{ 167, 269, 44 },
		{ 167, 270, 44 },
		{ 167, 271, 44 },
		{ 167, 272, 44 },
		{ 167, 273, 44 },
		{ 167, 274, 44 },
		{ 167, 275, 44 },
		{ 167, 276, 44 },
		{ 167, 277, 44 },
		{ 167, 278, 44 },
		{ 167, 279, 44 },
		{ 167, 280, 44 },
		{ 167, 281, 44 },
		{ 167, 282, 44 },
		{ 167, 283, 44 },
		{ 167, 284, 44 },
		{ 167, 285, 44 },
		{ 167, 286, 44 },
		{ 167, 287, 44 },
		{ 167, 288, 44 },
		{ 167, 289, 44 },
		{ 167, 290, 44 },
		{ 168, 269, 44 },
		{ 168, 270, 44 },
		{ 168, 271, 44 },
		{ 168, 272, 44 },
		{ 168, 273, 44 },
		{ 168, 274, 44 },
		{ 168, 275, 44 },
		{ 168, 276, 44 },
		{ 168, 277, 44 },
		{ 168, 278, 44 },
		{ 168, 279, 44 },
		{ 168, 280, 44 },
		{ 168, 281, 44 },
		{ 168, 282, 44 },
		{ 168, 283, 44 },
		{ 168, 284, 44 },
		{ 168, 285, 44 },
		{ 168, 286, 44 },
		{ 168, 287, 44 },
		{ 168, 288, 44 },
		{ 168, 289, 44 },
		{ 168, 290, 44 },
		{ 169, 269, 44 },
		{ 169, 270, 44 },
		{ 169, 271, 44 },
		{ 169, 272, 44 },
		{ 169, 273, 44 },
		{ 169, 274, 44 },
		{ 169, 275, 44 },
		{ 169, 276, 44 },
		{ 169, 277, 44 },
		{ 169, 278, 44 },
		{ 169, 279, 44 },
		{ 169, 280, 44 },
		{ 169, 281, 44 },
		{ 169, 282, 44 },
		{ 169, 283, 44 },
		{ 169, 284, 44 },
		{ 169, 285, 44 },
		{ 169, 286, 44 },
		{ 169, 287, 44 },
		{ 169, 288, 44 },
		{ 169, 289, 44 },
		{ 170, 268, 44 },
		{ 170, 269, 44 },
		{ 170, 270, 44 },
		{ 170, 271, 44 },
		{ 170, 272, 44 },
		{ 170, 273, 44 },
		{ 170, 274, 44 },
		{ 170, 275, 44 },
		{ 170, 276, 44 },
		{ 170, 277, 44 },
		{ 170, 278, 44 },
		{ 170, 279, 44 },
		{ 170, 280, 44 },
		{ 170, 281, 44 },
		{ 170, 282, 44 },
		{ 170, 283, 44 },
		{ 170, 284, 44 },
		{ 170, 285, 44 },
		{ 170, 286, 44 },
		{ 170, 287, 44 },
		{ 170, 288, 44 },
		{ 170, 289, 44 },
		{ 171, 268, 44 },
		{ 171, 269, 44 },
		{ 171, 270, 44 },
		{ 171, 271, 44 },
		{ 171, 272, 44 },
		{ 171, 273, 44 },
		{ 171, 274, 44 },
		{ 171, 275, 44 },
		{ 171, 276, 44 },
		{ 171, 277, 44 },
		{ 171, 278, 44 },
		{ 171, 279, 44 },
		{ 171, 280, 44 },
		{ 171, 281, 44 },
		{ 171, 282, 44 },
		{ 171, 283, 44 },
		{ 171, 284, 44 },
		{ 171, 285, 44 },
		{ 171, 286, 44 },
		{ 171, 287, 44 },
		{ 171, 288, 44 },
		{ 171, 289, 44 },
		{ 172, 268, 44 },
		{ 172, 269, 44 },
		{ 172, 270, 44 },
		{ 172, 271, 44 },
		{ 172, 272, 44 },
		{ 172, 273, 44 },
		{ 172, 274, 44 },
		{ 172, 275, 44 },
		{ 172, 276, 44 },
		{ 172, 277, 44 },
		{ 172, 278, 44 },
		{ 172, 279, 44 },
		{ 172, 280, 44 },
		{ 172, 281, 44 },
		{ 172, 282, 44 },
		{ 172, 283, 44 },
		{ 172, 284, 44 },
		{ 172, 285, 44 },
		{ 172, 286, 44 },
		{ 172, 287, 44 },
		{ 172, 288, 44 },
		{ 173, 268, 44 },
		{ 173, 269, 44 },
		{ 173, 270, 44 },
		{ 173, 271, 44 },
		{ 173, 272, 44 },
		{ 173, 273, 44 },
		{ 173, 274, 44 },
		{ 173, 275, 44 },
		{ 173, 276, 44 },
		{ 173, 277, 44 },
		{ 173, 278, 44 },
		{ 173, 279, 44 },
		{ 173, 280, 44 },
		{ 173, 281, 44 },
		{ 173, 282, 44 },
		{ 173, 283, 44 },
		{ 173, 284, 44 },
		{ 173, 285, 44 },
		{ 173, 286, 44 },
		{ 173, 287, 44 },
		{ 173, 288, 44 },
		{ 174, 268, 44 },
		{ 174, 269, 44 },
		{ 174, 270, 44 },
		{ 174, 271, 44 },
		{ 174, 272, 44 },
		{ 174, 273, 44 },
		{ 174, 274, 44 },
		{ 174, 275, 44 },
		{ 174, 276, 44 },
		{ 174, 277, 44 },
		{ 174, 278, 44 },
		{ 174, 279, 44 },
		{ 174, 280, 44 },
		{ 174, 281, 44 },
		{ 174, 282, 44 },
		{ 174, 283, 44 },
		{ 174, 284, 44 },
		{ 174, 285, 44 },
		{ 174, 286, 44 },
		{ 174, 287, 44 },
		{ 174, 288, 44 },
		{ 175, 268, 44 },
		{ 175, 269, 44 },
		{ 175, 270, 44 },
		{ 175, 271, 44 },
		{ 175, 272, 44 },
		{ 175, 273, 44 },
		{ 175, 274, 44 },
		{ 175, 275, 44 },
		{ 175, 276, 44 },
		{ 175, 277, 44 },
		{ 175, 278, 44 },
		{ 175, 279, 44 },
		{ 175, 280, 44 },
		{ 175, 281, 44 },
		{ 175, 282, 44 },
		{ 175, 283, 44 },
		{ 175, 284, 44 },
		{ 175, 285, 44 },
		{ 175, 286, 44 },
		{ 175, 287, 44 },
		{ 176, 268, 44 },
		{ 176, 269, 44 },
		{ 176, 270, 44 },
		{ 176, 271, 44 },
		{ 176, 272, 44 },
		{ 176, 273, 44 },
		{ 176, 274, 44 },
		{ 176, 275, 44 },
		{ 176, 276, 44 },
		{ 176, 277, 44 },
		{ 176, 278, 44 },
		{ 176, 279, 44 },
		{ 176, 280, 44 },
		{ 176, 281, 44 },
		{ 176, 282, 44 },
		{ 176, 283, 44 },
		{ 176, 284, 44 },
		{ 176, 285, 44 },
		{ 176, 286, 44 },
		{ 176, 287, 44 },
		{ 177, 268, 44 },
		{ 177, 269, 44 },
		{ 177, 270, 44 },
		{ 177, 271, 44 },
		{ 177, 272, 44 },
		{ 177, 273, 44 },
		{ 177, 274, 44 },
		{ 177, 275, 44 },
		{ 177, 276, 44 },
		{ 177, 277, 44 },
		{ 177, 278, 44 },
		{ 177, 279, 44 },
		{ 177, 280, 44 },
		{ 177, 281, 44 },
		{ 177, 282, 44 },
		{ 177, 283, 44 },
		{ 177, 284, 44 },
		{ 177, 285, 44 },
		{ 177, 286, 44 },
		{ 178, 267, 44 },
		{ 178, 268, 44 },
		{ 178, 269, 44 },
		{ 178, 270, 44 },
		{ 178, 271, 44 },
		{ 178, 272, 44 },
		{ 178, 273, 44 },
		{ 178, 274, 44 },
		{ 178, 275, 44 },
		{ 178, 276, 44 },
		{ 178, 277, 44 },
		{ 178, 278, 44 },
		{ 178, 279, 44 },
		{ 178, 280, 44 },
		{ 178, 281, 44 },
		{ 178, 282, 44 },
		{ 178, 283, 44 },
		{ 178, 284, 44 },
		{ 178, 285, 44 },
		{ 178, 286, 44 },
		{ 179, 267, 44 },
		{ 179, 268, 44 },
		{ 179, 269, 44 },
		{ 179, 270, 44 },
		{ 179, 271, 44 },
		{ 179, 272, 44 },
		{ 179, 273, 44 },
		{ 179, 274, 44 },
		{ 179, 275, 44 },
		{ 179, 276, 44 },
		{ 179, 277, 44 },
		{ 179, 278, 44 },
		{ 179, 279, 44 },
		{ 179, 280, 44 },
		{ 179, 281, 44 },
		{ 179, 282, 44 },
		{ 179, 283, 44 },
		{ 179, 284, 44 },
		{ 179, 285, 44 },
		{ 179, 286, 44 },
		{ 180, 267, 44 },
		{ 180, 268, 44 },
		{ 180, 269, 44 },
		{ 180, 270, 44 },
		{ 180, 271, 44 },
		{ 180, 272, 44 },
		{ 180, 273, 44 },
		{ 180, 274, 44 },
		{ 180, 275, 44 },
		{ 180, 276, 44 },
		{ 180, 277, 44 },
		{ 180, 278, 44 },
		{ 180, 279, 44 },
		{ 180, 280, 44 },
		{ 180, 281, 44 },
		{ 180, 282, 44 },
		{ 180, 283, 44 },
		{ 180, 284, 44 },
		{ 180, 285, 44 },
		{ 181, 267, 44 },
		{ 181, 268, 44 },
		{ 181, 269, 44 },
		{ 181, 270, 44 },
		{ 181, 271, 44 },
		{ 181, 272, 44 },
		{ 181, 273, 44 },
		{ 181, 274, 44 },
		{ 181, 275, 44 },
		{ 181, 276, 44 },
		{ 181, 277, 44 },
		{ 181, 278, 44 },
		{ 181, 279, 44 },
		{ 181, 280, 44 },
		{ 181, 281, 44 },
		{ 181, 282, 44 },
		{ 181, 283, 44 },
		{ 181, 284, 44 },
		{ 181, 285, 44 },
		{ 182, 267, 44 },
		{ 182, 268, 44 },
		{ 182, 269, 44 },
		{ 182, 270, 44 },
		{ 182, 271, 44 },
		{ 182, 272, 44 },
		{ 182, 273, 44 },
		{ 182, 274, 44 },
		{ 182, 275, 44 },
		{ 182, 276, 44 },
		{ 182, 277, 44 },
		{ 182, 278, 44 },
		{ 182, 279, 44 },
		{ 182, 280, 44 },
		{ 182, 281, 44 },
		{ 182, 282, 44 },
		{ 182, 283, 44 },
		{ 182, 284, 44 },
		{ 182, 285, 44 },
		{ 183, 267, 44 },
		{ 183, 268, 44 },
		{ 183, 269, 44 },
		{ 183, 270, 44 },
		{ 183, 271, 44 },
		{ 183, 272, 44 },
		{ 183, 273, 44 },
		{ 183, 274, 44 },
		{ 183, 275, 44 },
		{ 183, 276, 44 },
		{ 183, 277, 44 },
		{ 183, 278, 44 },
		{ 183, 279, 44 },
		{ 183, 280, 44 },
		{ 183, 281, 44 },
		{ 183, 282, 44 },
		{ 183, 283, 44 },
		{ 183, 284, 44 },
		{ 184, 267, 44 },
		{ 184, 268, 44 },
		{ 184, 269, 44 },
		{ 184, 270, 44 },
		{ 184, 271, 44 },
		{ 184, 272, 44 },
		{ 184, 273, 44 },
		{ 184, 274, 44 },
		{ 184, 275, 44 },
		{ 184, 276, 44 },
		{ 184, 277, 44 },
		{ 184, 278, 44 },
		{ 184, 279, 44 },
		{ 184, 280, 44 },
		{ 184, 281, 44 },
		{ 184, 282, 44 },
		{ 184, 283, 44 },
		{ 184, 284, 44 },
		{ 185, 267, 44 },
		{ 185, 268, 44 },
		{ 185, 269, 44 },
		{ 185, 270, 44 },
		{ 185, 271, 44 },
		{ 185, 272, 44 },
		{ 185, 273, 44 },
		{ 185, 274, 44 },
		{ 185, 275, 44 },
		{ 185, 276, 44 },
		{ 185, 277, 44 },
		{ 185, 278, 44 },
		{ 185, 279, 44 },
		{ 185, 280, 44 },
		{ 185, 281, 44 },
		{ 185, 282, 44 },
		{ 185, 283, 44 },
		{ 186, 267, 44 },
		{ 186, 268, 44 },
		{ 186, 269, 44 },
		{ 186, 270, 44 },
		{ 186, 271, 44 },
		{ 186, 272, 44 },
		{ 186, 273, 44 },
		{ 186, 274, 44 },
		{ 186, 275, 44 },
		{ 186, 276, 44 },
		{ 186, 277, 44 },
		{ 186, 278, 44 },
		{ 186, 279, 44 },
		{ 186, 280, 44 },
		{ 186, 281, 44 },
		{ 186, 282, 44 },
		{ 186, 283, 44 },
		{ 187, 266, 44 },
		{ 187, 267, 44 },
		{ 187, 268, 44 },
		{ 187, 269, 44 },
		{ 187, 270, 44 },
		{ 187, 271, 44 },
		{ 187, 272, 44 },
		{ 187, 273, 44 },
		{ 187, 274, 44 },
		{ 187, 275, 44 },
		{ 187, 276, 44 },
		{ 187, 277, 44 },
		{ 187, 278, 44 },
		{ 187, 279, 44 },
		{ 187, 280, 44 },
		{ 187, 281, 44 },
		{ 187, 282, 44 },
		{ 187, 283, 44 },
		{ 188, 266, 44 },
		{ 188, 267, 44 },
		{ 188, 268, 44 },
		{ 188, 269, 44 },
		{ 188, 270, 44 },
		{ 188, 271, 44 },
		{ 188, 272, 44 },
		{ 188, 273, 44 },
		{ 188, 274, 44 },
		{ 188, 275, 44 },
		{ 188, 276, 44 },
		{ 188, 277, 44 },
		{ 188, 278, 44 },
		{ 188, 279, 44 },
		{ 188, 280, 44 },
		{ 188, 281, 44 },
		{ 188, 282, 44 },
		{ 189, 266, 44 },
		{ 189, 267, 44 },
		{ 189, 268, 44 },
		{ 189, 269, 44 },
		{ 189, 270, 44 },
		{ 189, 271, 44 },
		{ 189, 272, 44 },
		{ 189, 273, 44 },
		{ 189, 274, 44 },
		{ 189, 275, 44 },
		{ 189, 276, 44 },
		{ 189, 277, 44 },
		{ 189, 278, 44 },
		{ 189, 279, 44 },
		{ 189, 280, 44 },
		{ 189, 281, 44 },
		{ 189, 282, 44 },
		{ 190, 267, 44 },
		{ 190, 268, 44 },
		{ 190, 269, 44 },
		{ 190, 270, 44 },
		{ 190, 271, 44 },
		{ 190, 272, 44 },
		{ 190, 273, 44 },
		{ 190, 274, 44 },
		{ 190, 275, 44 },
		{ 190, 276, 44 },
		{ 190, 277, 44 },
		{ 190, 278, 44 },
		{ 190, 279, 44 },
		{ 190, 280, 44 },
		{ 190, 281, 44 },
		{ 191, 268, 44 },
		{ 191, 269, 44 },
		{ 191, 270, 44 },
		{ 191, 271, 44 },
		{ 191, 272, 44 },
		{ 191, 273, 44 },
		{ 191, 274, 44 },
		{ 191, 275, 44 },
		{ 191, 276, 44 },
		{ 191, 277, 44 },
		{ 191, 278, 44 },
		{ 191, 279, 44 },
		{ 323, 235, 44 },
		{ 323, 236, 44 },
		{ 323, 237, 44 },
		{ 323, 238, 44 },
		{ 323, 239, 44 },
		{ 323, 240, 44 },
		{ 323, 241, 44 },
		{ 323, 242, 44 },
		{ 323, 243, 44 },
		{ 323, 244, 44 },
		{ 323, 245, 44 },
		{ 323, 246, 44 },
		{ 324, 233, 44 },
		{ 324, 234, 44 },
		{ 324, 235, 44 },
		{ 324, 236, 44 },
		{ 324, 237, 44 },
		{ 324, 238, 44 },
		{ 324, 239, 44 },
		{ 324, 240, 44 },
		{ 324, 241, 44 },
		{ 324, 242, 44 },
		{ 324, 243, 44 },
		{ 324, 244, 44 },
		{ 324, 245, 44 },
		{ 324, 246, 44 },
		{ 324, 247, 44 },
		{ 325, 232, 44 },
		{ 325, 233, 44 },
		{ 325, 234, 44 },
		{ 325, 235, 44 },
		{ 325, 236, 44 },
		{ 325, 237, 44 },
		{ 325, 238, 44 },
		{ 325, 239, 44 },
		{ 325, 240, 44 },
		{ 325, 241, 44 },
		{ 325, 242, 44 },
		{ 325, 243, 44 },
		{ 325, 244, 44 },
		{ 325, 245, 44 },
		{ 325, 246, 44 },
		{ 325, 247, 44 },
		{ 325, 248, 44 },
		{ 326, 232, 44 },
		{ 326, 233, 44 },
		{ 326, 234, 44 },
		{ 326, 235, 44 },
		{ 326, 236, 44 },
		{ 326, 237, 44 },
		{ 326, 238, 44 },
		{ 326, 239, 44 },
		{ 326, 240, 44 },
		{ 326, 241, 44 },
		{ 326, 242, 44 },
		{ 326, 243, 44 },
		{ 326, 244, 44 },
		{ 326, 245, 44 },
		{ 326, 246, 44 },
		{ 326, 247, 44 },
		{ 326, 248, 44 },
		{ 327, 231, 44 },
		{ 327, 232, 44 },
		{ 327, 233, 44 },
		{ 327, 234, 44 },
		{ 327, 235, 44 },
		{ 327, 236, 44 },
		{ 327, 237, 44 },
		{ 327, 238, 44 },
		{ 327, 239, 44 },
		{ 327, 240, 44 },
		{ 327, 241, 44 },
		{ 327, 242, 44 },
		{ 327, 243, 44 },
		{ 327, 244, 44 },
		{ 327, 245, 44 },
		{ 327, 246, 44 },
		{ 327, 247, 44 },
		{ 327, 248, 44 },
		{ 328, 231, 44 },
		{ 328, 232, 44 },
		{ 328, 233, 44 },
		{ 328, 234, 44 },
		{ 328, 235, 44 },
		{ 328, 236, 44 },
		{ 328, 237, 44 },
		{ 328, 238, 44 },
		{ 328, 239, 44 },
		{ 328, 240, 44 },
		{ 328, 241, 44 },
		{ 328, 242, 44 },
		{ 328, 243, 44 },
		{ 328, 244, 44 },
		{ 328, 245, 44 },
		{ 328, 246, 44 },
		{ 328, 247, 44 },
		{ 329, 231, 44 },
		{ 329, 232, 44 },
		{ 329, 233, 44 },
		{ 329, 234, 44 },
		{ 329, 235, 44 },
		{ 329, 236, 44 },
		{ 329, 237, 44 },
		{ 329, 238, 44 },
		{ 329, 239, 44 },
		{ 329, 240, 44 },
		{ 329, 241, 44 },
		{ 329, 242, 44 },
		{ 329, 243, 44 },
		{ 329, 244, 44 },
		{ 329, 245, 44 },
		{ 329, 246, 44 },
		{ 329, 247, 44 },
		{ 330, 230, 44 },
		{ 330, 231, 44 },
		{ 330, 232, 44 },
		{ 330, 233, 44 },
		{ 330, 234, 44 },
		{ 330, 235, 44 },
		{ 330, 236, 44 },
		{ 330, 237, 44 },
		{ 330, 238, 44 },
		{ 330, 239, 44 },
		{ 330, 240, 44 },
		{ 330, 241, 44 },
		{ 330, 242, 44 },
		{ 330, 243, 44 },
		{ 330, 244, 44 },
		{ 330, 245, 44 },
		{ 330, 246, 44 },
		{ 330, 247, 44 },
		{ 331, 230, 44 },
		{ 331, 231, 44 },
		{ 331, 232, 44 },
		{ 331, 233, 44 },
		{ 331, 234, 44 },
		{ 331, 235, 44 },
		{ 331, 236, 44 },
		{ 331, 237, 44 },
		{ 331, 238, 44 },
		{ 331, 239, 44 },
		{ 331, 240, 44 },
		{ 331, 241, 44 },
		{ 331, 242, 44 },
		{ 331, 243, 44 },
		{ 331, 244, 44 },
		{ 331, 245, 44 },
		{ 331, 246, 44 },
		{ 331, 247, 44 },
		{ 332, 229, 44 },
		{ 332, 230, 44 },
		{ 332, 231, 44 },
		{ 332, 232, 44 },
		{ 332, 233, 44 },
		{ 332, 234, 44 },
		{ 332, 235, 44 },
		{ 332, 236, 44 },
		{ 332, 237, 44 },
		{ 332, 238, 44 },
		{ 332, 239, 44 },
		{ 332, 240, 44 },
		{ 332, 241, 44 },
		{ 332, 242, 44 },
		{ 332, 243, 44 },
		{ 332, 244, 44 },
		{ 332, 245, 44 },
		{ 332, 246, 44 },
		{ 332, 247, 44 },
		{ 333, 229, 44 },
		{ 333, 230, 44 },
		{ 333, 231, 44 },
		{ 333, 232, 44 },
		{ 333, 233, 44 },
		{ 333, 234, 44 },
		{ 333, 235, 44 },
		{ 333, 236, 44 },
		{ 333, 237, 44 },
		{ 333, 238, 44 },
		{ 333, 239, 44 },
		{ 333, 240, 44 },
		{ 333, 241, 44 },
		{ 333, 242, 44 },
		{ 333, 243, 44 },
		{ 333, 244, 44 },
		{ 333, 245, 44 },
		{ 333, 246, 44 },
		{ 333, 247, 44 },
		{ 334, 229, 44 },
		{ 334, 230, 44 },
		{ 334, 231, 44 },
		{ 334, 232, 44 },
		{ 334, 233, 44 },
		{ 334, 234, 44 },
		{ 334, 235, 44 },
		{ 334, 236, 44 },
		{ 334, 237, 44 },
		{ 334, 238, 44 },
		{ 334, 239, 44 },
		{ 334, 240, 44 },
		{ 334, 241, 44 },
		{ 334, 242, 44 },
		{ 334, 243, 44 },
		{ 334, 244, 44 },
		{ 334, 245, 44 },
		{ 334, 246, 44 },
		{ 334, 247, 44 },
		{ 335, 228, 44 },
		{ 335, 229, 44 },
		{ 335, 230, 44 },
		{ 335, 231, 44 },
		{ 335, 232, 44 },
		{ 335, 233, 44 },
		{ 335, 234, 44 },
		{ 335, 235, 44 },
		{ 335, 236, 44 },
		{ 335, 237, 44 },
		{ 335, 238, 44 },
		{ 335, 239, 44 },
		{ 335, 240, 44 },
		{ 335, 241, 44 },
		{ 335, 242, 44 },
		{ 335, 243, 44 },
		{ 335, 244, 44 },
		{ 335, 245, 44 },
		{ 335, 246, 44 },
		{ 335, 247, 44 },
		{ 336, 228, 44 },
		{ 336, 229, 44 },
		{ 336, 230, 44 },
		{ 336, 231, 44 },
		{ 336, 232, 44 },
		{ 336, 233, 44 },
		{ 336, 234, 44 },
		{ 336, 235, 44 },
		{ 336, 236, 44 },
		{ 336, 237, 44 },
		{ 336, 238, 44 },
		{ 336, 239, 44 },
		{ 336, 240, 44 },
		{ 336, 241, 44 },
		{ 336, 242, 44 },
		{ 336, 243, 44 },
		{ 336, 244, 44 },
		{ 336, 245, 44 },
		{ 336, 246, 44 },
		{ 336, 247, 44 },
		{ 337, 228, 44 },
		{ 337, 229, 44 },
		{ 337, 230, 44 },
		{ 337, 231, 44 },
		{ 337, 232, 44 },
		{ 337, 233, 44 },
		{ 337, 234, 44 },
		{ 337, 235, 44 },
		{ 337, 236, 44 },
		{ 337, 237, 44 },
		{ 337, 238, 44 },
		{ 337, 239, 44 },
		{ 337, 240, 44 },
		{ 337, 241, 44 },
		{ 337, 242, 44 },
		{ 337, 243, 44 },
		{ 337, 244, 44 },
		{ 337, 245, 44 },
		{ 337, 246, 44 },
		{ 338, 227, 44 },
		{ 338, 228, 44 },
		{ 338, 229, 44 },
		{ 338, 230, 44 },
		{ 338, 231, 44 },
		{ 338, 232, 44 },
		{ 338, 233, 44 },
		{ 338, 234, 44 },
		{ 338, 235, 44 },
		{ 338, 236, 44 },
		{ 338, 237, 44 },
		{ 338, 238, 44 },
		{ 338, 239, 44 },
		{ 338, 240, 44 },
		{ 338, 241, 44 },
		{ 338, 242, 44 },
		{ 338, 243, 44 },
		{ 338, 244, 44 },
		{ 338, 245, 44 },
		{ 338, 246, 44 },
		{ 339, 227, 44 },
		{ 339, 228, 44 },
		{ 339, 229, 44 },
		{ 339, 230, 44 },
		{ 339, 231, 44 },
		{ 339, 232, 44 },
		{ 339, 233, 44 },
		{ 339, 234, 44 },
		{ 339, 235, 44 },
		{ 339, 236, 44 },
		{ 339, 237, 44 },
		{ 339, 238, 44 },
		{ 339, 239, 44 },
		{ 339, 240, 44 },
		{ 339, 241, 44 },
		{ 339, 242, 44 },
		{ 339, 243, 44 },
		{ 339, 244, 44 },
		{ 339, 245, 44 },
		{ 339, 246, 44 },
		{ 340, 226, 44 },
		{ 340, 227, 44 },
		{ 340, 228, 44 },
		{ 340, 229, 44 },
		{ 340, 230, 44 },
		{ 340, 231, 44 },
		{ 340, 232, 44 },
		{ 340, 233, 44 },
		{ 340, 234, 44 },
		{ 340, 235, 44 },
		{ 340, 236, 44 },
		{ 340, 237, 44 },
		{ 340, 238, 44 },
		{ 340, 239, 44 },
		{ 340, 240, 44 },
		{ 340, 241, 44 },
		{ 340, 242, 44 },
		{ 340, 243, 44 },
		{ 340, 244, 44 },
		{ 340, 245, 44 },
		{ 340, 246, 44 },
		{ 341, 226, 44 },
		{ 341, 227, 44 },
		{ 341, 228, 44 },
		{ 341, 229, 44 },
		{ 341, 230, 44 },
		{ 341, 231, 44 },
		{ 341, 232, 44 },
		{ 341, 233, 44 },
		{ 341, 234, 44 },
		{ 341, 235, 44 },
		{ 341, 236, 44 },
		{ 341, 237, 44 },
		{ 341, 238, 44 },
		{ 341, 239, 44 },
		{ 341, 240, 44 },
		{ 341, 241, 44 },
		{ 341, 242, 44 },
		{ 341, 243, 44 },
		{ 341, 244, 44 },
		{ 341, 245, 44 },
		{ 341, 246, 44 },
		{ 342, 226, 44 },
		{ 342, 227, 44 },
		{ 342, 228, 44 },
		{ 342, 229, 44 },
		{ 342, 230, 44 },
		{ 342, 231, 44 },
		{ 342, 232, 44 },
		{ 342, 233, 44 },
		{ 342, 234, 44 },
		{ 342, 235, 44 },
		{ 342, 236, 44 },
		{ 342, 237, 44 },
		{ 342, 238, 44 },
		{ 342, 239, 44 },
		{ 342, 240, 44 },
		{ 342, 241, 44 },
		{ 342, 242, 44 },
		{ 342, 243, 44 },
		{ 342, 244, 44 },
		{ 342, 245, 44 },
		{ 342, 246, 44 },
		{ 343, 225, 44 },
		{ 343, 226, 44 },
		{ 343, 227, 44 },
		{ 343, 228, 44 },
		{ 343, 229, 44 },
		{ 343, 230, 44 },
		{ 343, 231, 44 },
		{ 343, 232, 44 },
		{ 343, 233, 44 },
		{ 343, 234, 44 },
		{ 343, 235, 44 },
		{ 343, 236, 44 },
		{ 343, 237, 44 },
		{ 343, 238, 44 },
		{ 343, 239, 44 },
		{ 343, 240, 44 },
		{ 343, 241, 44 },
		{ 343, 242, 44 },
		{ 343, 243, 44 },
		{ 343, 244, 44 },
		{ 343, 245, 44 },
		{ 343, 246, 44 },
		{ 344, 225, 44 },
		{ 344, 226, 44 },
		{ 344, 227, 44 },
		{ 344, 228, 44 },
		{ 344, 229, 44 },
		{ 344, 230, 44 },
		{ 344, 231, 44 },
		{ 344, 232, 44 },
		{ 344, 233, 44 },
		{ 344, 234, 44 },
		{ 344, 235, 44 },
		{ 344, 236, 44 },
		{ 344, 237, 44 },
		{ 344, 238, 44 },
		{ 344, 239, 44 },
		{ 344, 240, 44 },
		{ 344, 241, 44 },
		{ 344, 242, 44 },
		{ 344, 243, 44 },
		{ 344, 244, 44 },
		{ 344, 245, 44 },
		{ 344, 246, 44 },
		{ 345, 225, 44 },
		{ 345, 226, 44 },
		{ 345, 227, 44 },
		{ 345, 228, 44 },
		{ 345, 229, 44 },
		{ 345, 230, 44 },
		{ 345, 231, 44 },
		{ 345, 232, 44 },
		{ 345, 233, 44 },
		{ 345, 234, 44 },
		{ 345, 235, 44 },
		{ 345, 236, 44 },
		{ 345, 237, 44 },
		{ 345, 238, 44 },
		{ 345, 239, 44 },
		{ 345, 240, 44 },
		{ 345, 241, 44 },
		{ 345, 242, 44 },
		{ 345, 243, 44 },
		{ 345, 244, 44 },
		{ 345, 245, 44 },
		{ 346, 224, 44 },
		{ 346, 225, 44 },
		{ 346, 226, 44 },
		{ 346, 227, 44 },
		{ 346, 228, 44 },
		{ 346, 229, 44 },
		{ 346, 230, 44 },
		{ 346, 231, 44 },
		{ 346, 232, 44 },
		{ 346, 233, 44 },
		{ 346, 234, 44 },
		{ 346, 235, 44 },
		{ 346, 236, 44 },
		{ 346, 237, 44 },
		{ 346, 238, 44 },
		{ 346, 239, 44 },
		{ 346, 240, 44 },
		{ 346, 241, 44 },
		{ 346, 242, 44 },
		{ 346, 243, 44 },
		{ 346, 244, 44 },
		{ 346, 245, 44 },
		{ 347, 224, 44 },
		{ 347, 225, 44 },
		{ 347, 226, 44 },
		{ 347, 227, 44 },
		{ 347, 228, 44 },
		{ 347, 229, 44 },
		{ 347, 230, 44 },
		{ 347, 231, 44 },
		{ 347, 232, 44 },
		{ 347, 233, 44 },
		{ 347, 234, 44 },
		{ 347, 235, 44 },
		{ 347, 236, 44 },
		{ 347, 237, 44 },
		{ 347, 238, 44 },
		{ 347, 239, 44 },
		{ 347, 240, 44 },
		{ 347, 241, 44 },
		{ 347, 242, 44 },
		{ 347, 243, 44 },
		{ 347, 244, 44 },
		{ 347, 245, 44 },
		{ 348, 223, 44 },
		{ 348, 224, 44 },
		{ 348, 225, 44 },
		{ 348, 226, 44 },
		{ 348, 227, 44 },
		{ 348, 228, 44 },
		{ 348, 229, 44 },
		{ 348, 230, 44 },
		{ 348, 231, 44 },
		{ 348, 232, 44 },
		{ 348, 233, 44 },
		{ 348, 234, 44 },
		{ 348, 235, 44 },
		{ 348, 236, 44 },
		{ 348, 237, 44 },
		{ 348, 238, 44 },
		{ 348, 239, 44 },
		{ 348, 240, 44 },
		{ 348, 241, 44 },
		{ 348, 242, 44 },
		{ 348, 243, 44 },
		{ 348, 244, 44 },
		{ 348, 245, 44 },
		{ 349, 223, 44 },
		{ 349, 224, 44 },
		{ 349, 225, 44 },
		{ 349, 226, 44 },
		{ 349, 227, 44 },
		{ 349, 228, 44 },
		{ 349, 229, 44 },
		{ 349, 230, 44 },
		{ 349, 231, 44 },
		{ 349, 232, 44 },
		{ 349, 233, 44 },
		{ 349, 234, 44 },
		{ 349, 235, 44 },
		{ 349, 236, 44 },
		{ 349, 237, 44 },
		{ 349, 238, 44 },
		{ 349, 239, 44 },
		{ 349, 240, 44 },
		{ 349, 241, 44 },
		{ 349, 242, 44 },
		{ 349, 243, 44 },
		{ 349, 244, 44 },
		{ 349, 245, 44 },
		{ 350, 223, 44 },
		{ 350, 224, 44 },
		{ 350, 225, 44 },
		{ 350, 226, 44 },
		{ 350, 227, 44 },
		{ 350, 228, 44 },
		{ 350, 229, 44 },
		{ 350, 230, 44 },
		{ 350, 231, 44 },
		{ 350, 232, 44 },
		{ 350, 233, 44 },
		{ 350, 234, 44 },
		{ 350, 235, 44 },
		{ 350, 236, 44 },
		{ 350, 237, 44 },
		{ 350, 238, 44 },
		{ 350, 239, 44 },
		{ 350, 240, 44 },
		{ 350, 241, 44 },
		{ 350, 242, 44 },
		{ 350, 243, 44 },
		{ 350, 244, 44 },
		{ 350, 245, 44 },
		{ 351, 222, 44 },
		{ 351, 223, 44 },
		{ 351, 224, 44 },
		{ 351, 225, 44 },
		{ 351, 226, 44 },
		{ 351, 227, 44 },
		{ 351, 228, 44 },
		{ 351, 229, 44 },
		{ 351, 230, 44 },
		{ 351, 231, 44 },
		{ 351, 232, 44 },
		{ 351, 233, 44 },
		{ 351, 234, 44 },
		{ 351, 235, 44 },
		{ 351, 236, 44 },
		{ 351, 237, 44 },
		{ 351, 238, 44 },
		{ 351, 239, 44 },
		{ 351, 240, 44 },
		{ 351, 241, 44 },
		{ 351, 242, 44 },
		{ 351, 243, 44 },
		{ 351, 244, 44 },
		{ 351, 245, 44 },
		{ 352, 222, 44 },
		{ 352, 223, 44 },
		{ 352, 224, 44 },
		{ 352, 225, 44 },
		{ 352, 226, 44 },
		{ 352, 227, 44 },
		{ 352, 228, 44 },
		{ 352, 229, 44 },
		{ 352, 230, 44 },
		{ 352, 231, 44 },
		{ 352, 232, 44 },
		{ 352, 233, 44 },
		{ 352, 234, 44 },
		{ 352, 235, 44 },
		{ 352, 236, 44 },
		{ 352, 237, 44 },
		{ 352, 238, 44 },
		{ 352, 239, 44 },
		{ 352, 240, 44 },
		{ 352, 241, 44 },
		{ 352, 242, 44 },
		{ 352, 243, 44 },
		{ 352, 244, 44 },
		{ 352, 245, 44 },
		{ 353, 222, 44 },
		{ 353, 223, 44 },
		{ 353, 224, 44 },
		{ 353, 225, 44 },
		{ 353, 226, 44 },
		{ 353, 227, 44 },
		{ 353, 228, 44 },
		{ 353, 229, 44 },
		{ 353, 230, 44 },
		{ 353, 231, 44 },
		{ 353, 232, 44 },
		{ 353, 233, 44 },
		{ 353, 234, 44 },
		{ 353, 235, 44 },
		{ 353, 236, 44 },
		{ 353, 237, 44 },
		{ 353, 238, 44 },
		{ 353, 239, 44 },
		{ 353, 240, 44 },
		{ 353, 241, 44 },
		{ 353, 242, 44 },
		{ 353, 243, 44 },
		{ 353, 244, 44 },
		{ 354, 221, 44 },
		{ 354, 222, 44 },
		{ 354, 223, 44 },
		{ 354, 224, 44 },
		{ 354, 225, 44 },
		{ 354, 226, 44 },
		{ 354, 227, 44 },
		{ 354, 228, 44 },
		{ 354, 229, 44 },
		{ 354, 230, 44 },
		{ 354, 231, 44 },
		{ 354, 232, 44 },
		{ 354, 233, 44 },
		{ 354, 234, 44 },
		{ 354, 235, 44 },
		{ 354, 236, 44 },
		{ 354, 237, 44 },
		{ 354, 238, 44 },
		{ 354, 239, 44 },
		{ 354, 240, 44 },
		{ 354, 241, 44 },
		{ 354, 242, 44 },
		{ 354, 243, 44 },
		{ 354, 244, 44 },
		{ 355, 221, 44 },
		{ 355, 222, 44 },
		{ 355, 223, 44 },
		{ 355, 224, 44 },
		{ 355, 225, 44 },
		{ 355, 226, 44 },
		{ 355, 227, 44 },
		{ 355, 228, 44 },
		{ 355, 229, 44 },
		{ 355, 230, 44 },
		{ 355, 231, 44 },
		{ 355, 232, 44 },
		{ 355, 233, 44 },
		{ 355, 234, 44 },
		{ 355, 235, 44 },
		{ 355, 236, 44 },
		{ 355, 237, 44 },
		{ 355, 238, 44 },
		{ 355, 239, 44 },
		{ 355, 240, 44 },
		{ 355, 241, 44 },
		{ 355, 242, 44 },
		{ 355, 243, 44 },
		{ 355, 244, 44 },
		{ 356, 220, 44 },
		{ 356, 221, 44 },
		{ 356, 222, 44 },
		{ 356, 223, 44 },
		{ 356, 224, 44 },
		{ 356, 225, 44 },
		{ 356, 226, 44 },
		{ 356, 227, 44 },
		{ 356, 228, 44 },
		{ 356, 229, 44 },
		{ 356, 230, 44 },
		{ 356, 231, 44 },
		{ 356, 232, 44 },
		{ 356, 233, 44 },
		{ 356, 234, 44 },
		{ 356, 235, 44 },
		{ 356, 236, 44 },
		{ 356, 237, 44 },
		{ 356, 238, 44 },
		{ 356, 239, 44 },
		{ 356, 240, 44 },
		{ 356, 241, 44 },
		{ 356, 242, 44 },
		{ 356, 243, 44 },
		{ 356, 244, 44 },
		{ 357, 220, 44 },
		{ 357, 221, 44 },
		{ 357, 222, 44 },
		{ 357, 223, 44 },
		{ 357, 224, 44 },
		{ 357, 225, 44 },
		{ 357, 226, 44 },
		{ 357, 227, 44 },
		{ 357, 228, 44 },
		{ 357, 229, 44 },
		{ 357, 230, 44 },
		{ 357, 231, 44 },
		{ 357, 232, 44 },
		{ 357, 233, 44 },
		{ 357, 234, 44 },
		{ 357, 235, 44 },
		{ 357, 236, 44 },
		{ 357, 237, 44 },
		{ 357, 238, 44 },
		{ 357, 239, 44 },
		{ 357, 240, 44 },
		{ 357, 241, 44 },
		{ 357, 242, 44 },
		{ 357, 243, 44 },
		{ 357, 244, 44 },
		{ 358, 220, 44 },
		{ 358, 221, 44 },
		{ 358, 222, 44 },
		{ 358, 223, 44 },
		{ 358, 224, 44 },
		{ 358, 225, 44 },
		{ 358, 226, 44 },
		{ 358, 227, 44 },
		{ 358, 228, 44 },
		{ 358, 229, 44 },
		{ 358, 230, 44 },
		{ 358, 231, 44 },
		{ 358, 232, 44 },
		{ 358, 233, 44 },
		{ 358, 234, 44 },
		{ 358, 235, 44 },
		{ 358, 236, 44 },
		{ 358, 237, 44 },
		{ 358, 238, 44 },
		{ 358, 239, 44 },
		{ 358, 240, 44 },
		{ 358, 241, 44 },
		{ 358, 242, 44 },
		{ 358, 243, 44 },
		{ 358, 244, 44 },
		{ 359, 219, 44 },
		{ 359, 220, 44 },
		{ 359, 221, 44 },
		{ 359, 222, 44 },
		{ 359, 223, 44 },
		{ 359, 224, 44 },
		{ 359, 225, 44 },
		{ 359, 226, 44 },
		{ 359, 227, 44 },
		{ 359, 228, 44 },
		{ 359, 229, 44 },
		{ 359, 230, 44 },
		{ 359, 231, 44 },
		{ 359, 232, 44 },
		{ 359, 233, 44 },
		{ 359, 234, 44 },
		{ 359, 235, 44 },
		{ 359, 236, 44 },
		{ 359, 237, 44 },
		{ 359, 238, 44 },
		{ 359, 239, 44 },
		{ 359, 240, 44 },
		{ 359, 241, 44 },
		{ 359, 242, 44 },
		{ 359, 243, 44 },
		{ 359, 244, 44 },
		{ 360, 219, 44 },
		{ 360, 220, 44 },
		{ 360, 221, 44 },
		{ 360, 222, 44 },
		{ 360, 223, 44 },
		{ 360, 224, 44 },
		{ 360, 225, 44 },
		{ 360, 226, 44 },
		{ 360, 227, 44 },
		{ 360, 228, 44 },
		{ 360, 229, 44 },
		{ 360, 230, 44 },
		{ 360, 231, 44 },
		{ 360, 232, 44 },
		{ 360, 233, 44 },
		{ 360, 234, 44 },
		{ 360, 235, 44 },
		{ 360, 236, 44 },
		{ 360, 237, 44 },
		{ 360, 238, 44 },
		{ 360, 239, 44 },
		{ 360, 240, 44 },
		{ 360, 241, 44 },
		{ 360, 242, 44 },
		{ 360, 243, 44 },
		{ 360, 244, 44 },
		{ 361, 219, 44 },
		{ 361, 220, 44 },
		{ 361, 221, 44 },
		{ 361, 222, 44 },
		{ 361, 223, 44 },
		{ 361, 224, 44 },
		{ 361, 225, 44 },
		{ 361, 226, 44 },
		{ 361, 227, 44 },
		{ 361, 228, 44 },
		{ 361, 229, 44 },
		{ 361, 230, 44 },
		{ 361, 231, 44 },
		{ 361, 232, 44 },
		{ 361, 233, 44 },
		{ 361, 234, 44 },
		{ 361, 235, 44 },
		{ 361, 236, 44 },
		{ 361, 237, 44 },
		{ 361, 238, 44 },
		{ 361, 239, 44 },
		{ 361, 240, 44 },
		{ 361, 241, 44 },
		{ 361, 242, 44 },
		{ 361, 243, 44 },
		{ 362, 218, 44 },
		{ 362, 219, 44 },
		{ 362, 220, 44 },
		{ 362, 221, 44 },
		{ 362, 222, 44 },
		{ 362, 223, 44 },
		{ 362, 224, 44 },
		{ 362, 225, 44 },
		{ 362, 226, 44 },
		{ 362, 227, 44 },
		{ 362, 228, 44 },
		{ 362, 229, 44 },
		{ 362, 230, 44 },
		{ 362, 231, 44 },
		{ 362, 232, 44 },
		{ 362, 233, 44 },
		{ 362, 234, 44 },
		{ 362, 235, 44 },
		{ 362, 236, 44 },
		{ 362, 237, 44 },
		{ 362, 238, 44 },
		{ 362, 239, 44 },
		{ 362, 240, 44 },
		{ 362, 241, 44 },
		{ 362, 242, 44 },
		{ 362, 243, 44 },
		{ 363, 218, 44 },
		{ 363, 219, 44 },
		{ 363, 220, 44 },
		{ 363, 221, 44 },
		{ 363, 222, 44 },
		{ 363, 223, 44 },
		{ 363, 224, 44 },
		{ 363, 225, 44 },
		{ 363, 226, 44 },
		{ 363, 227, 44 },
		{ 363, 228, 44 },
		{ 363, 229, 44 },
		{ 363, 230, 44 },
		{ 363, 231, 44 },
		{ 363, 232, 44 },
		{ 363, 233, 44 },
		{ 363, 234, 44 },
		{ 363, 235, 44 },
		{ 363, 236, 44 },
		{ 363, 237, 44 },
		{ 363, 238, 44 },
		{ 363, 239, 44 },
		{ 363, 240, 44 },
		{ 363, 241, 44 },
		{ 363, 242, 44 },
		{ 363, 243, 44 },
		{ 364, 217, 44 },
		{ 364, 218, 44 },
		{ 364, 219, 44 },
		{ 364, 220, 44 },
		{ 364, 221, 44 },
		{ 364, 222, 44 },
		{ 364, 223, 44 },
		{ 364, 224, 44 },
		{ 364, 225, 44 },
		{ 364, 226, 44 },
		{ 364, 227, 44 },
		{ 364, 228, 44 },
		{ 364, 229, 44 },
		{ 364, 230, 44 },
		{ 364, 231, 44 },
		{ 364, 232, 44 },
		{ 364, 233, 44 },
		{ 364, 234, 44 },
		{ 364, 235, 44 },
		{ 364, 236, 44 },
		{ 364, 237, 44 },
		{ 364, 238, 44 },
		{ 364, 239, 44 },
		{ 364, 240, 44 },
		{ 364, 241, 44 },
		{ 364, 242, 44 },
		{ 364, 243, 44 },
		{ 365, 217, 44 },
		{ 365, 218, 44 },
		{ 365, 219, 44 },
		{ 365, 220, 44 },
		{ 365, 221, 44 },
		{ 365, 222, 44 },
		{ 365, 223, 44 },
		{ 365, 224, 44 },
		{ 365, 225, 44 },
		{ 365, 226, 44 },
		{ 365, 227, 44 },
		{ 365, 228, 44 },
		{ 365, 229, 44 },
		{ 365, 230, 44 },
		{ 365, 231, 44 },
		{ 365, 232, 44 },
		{ 365, 233, 44 },
		{ 365, 234, 44 },
		{ 365, 235, 44 },
		{ 365, 236, 44 },
		{ 365, 237, 44 },
		{ 365, 238, 44 },
		{ 365, 239, 44 },
		{ 365, 240, 44 },
		{ 365, 241, 44 },
		{ 365, 242, 44 },
		{ 365, 243, 44 },
		{ 366, 217, 44 },
		{ 366, 218, 44 },
		{ 366, 219, 44 },
		{ 366, 220, 44 },
		{ 366, 221, 44 },
		{ 366, 222, 44 },
		{ 366, 223, 44 },
		{ 366, 224, 44 },
		{ 366, 225, 44 },
		{ 366, 226, 44 },
		{ 366, 227, 44 },
		{ 366, 228, 44 },
		{ 366, 229, 44 },
		{ 366, 230, 44 },
		{ 366, 231, 44 },
		{ 366, 232, 44 },
		{ 366, 233, 44 },
		{ 366, 234, 44 },
		{ 366, 235, 44 },
		{ 366, 236, 44 },
		{ 366, 237, 44 },
		{ 366, 238, 44 },
		{ 366, 239, 44 },
		{ 366, 240, 44 },
		{ 366, 241, 44 },
		{ 366, 242, 44 },
		{ 366, 243, 44 },
		{ 367, 216, 44 },
		{ 367, 217, 44 },
		{ 367, 218, 44 },
		{ 367, 219, 44 },
		{ 367, 220, 44 },
		{ 367, 221, 44 },
		{ 367, 222, 44 },
		{ 367, 223, 44 },
		{ 367, 224, 44 },
		{ 367, 225, 44 },
		{ 367, 226, 44 },
		{ 367, 227, 44 },
		{ 367, 228, 44 },
		{ 367, 229, 44 },
		{ 367, 230, 44 },
		{ 367, 231, 44 },
		{ 367, 232, 44 },
		{ 367, 233, 44 },
		{ 367, 234, 44 },
		{ 367, 235, 44 },
		{ 367, 236, 44 },
		{ 367, 237, 44 },
		{ 367, 238, 44 },
		{ 367, 239, 44 },
		{ 367, 240, 44 },
		{ 367, 241, 44 },
		{ 367, 242, 44 },
		{ 367, 243, 44 },
		{ 368, 216, 44 },
		{ 368, 217, 44 },
		{ 368, 218, 44 },
		{ 368, 219, 44 },
		{ 368, 220, 44 },
		{ 368, 221, 44 },
		{ 368, 222, 44 },
		{ 368, 223, 44 },
		{ 368, 224, 44 },
		{ 368, 225, 44 },
		{ 368, 226, 44 },
		{ 368, 227, 44 },
		{ 368, 228, 44 },
		{ 368, 229, 44 },
		{ 368, 230, 44 },
		{ 368, 231, 44 },
		{ 368, 232, 44 },
		{ 368, 233, 44 },
		{ 368, 234, 44 },
		{ 368, 235, 44 },
		{ 368, 236, 44 },
		{ 368, 237, 44 },
		{ 368, 238, 44 },
		{ 368, 239, 44 },
		{ 368, 240, 44 },
		{ 368, 241, 44 },
		{ 368, 242, 44 },
		{ 368, 243, 44 },
		{ 369, 216, 44 },
		{ 369, 217, 44 },
		{ 369, 218, 44 },
		{ 369, 219, 44 },
		{ 369, 220, 44 },
		{ 369, 221, 44 },
		{ 369, 222, 44 },
		{ 369, 223, 44 },
		{ 369, 224, 44 },
		{ 369, 225, 44 },
		{ 369, 226, 44 },
		{ 369, 227, 44 },
		{ 369, 228, 44 },
		{ 369, 229, 44 },
		{ 369, 230, 44 },
		{ 369, 231, 44 },
		{ 369, 232, 44 },
		{ 369, 233, 44 },
		{ 369, 234, 44 },
		{ 369, 235, 44 },
		{ 369, 236, 44 },
		{ 369, 237, 44 },
		{ 369, 238, 44 },
		{ 369, 239, 44 },
		{ 369, 240, 44 },
		{ 369, 241, 44 },
		{ 369, 242, 44 },
		{ 370, 215, 44 },
		{ 370, 216, 44 },
		{ 370, 217, 44 },
		{ 370, 218, 44 },
		{ 370, 219, 44 },
		{ 370, 220, 44 },
		{ 370, 221, 44 },
		{ 370, 222, 44 },
		{ 370, 223, 44 },
		{ 370, 224, 44 },
		{ 370, 225, 44 },
		{ 370, 226, 44 },
		{ 370, 227, 44 },
		{ 370, 228, 44 },
		{ 370, 229, 44 },
		{ 370, 230, 44 },
		{ 370, 231, 44 },
		{ 370, 232, 44 },
		{ 370, 233, 44 },
		{ 370, 234, 44 },
		{ 370, 235, 44 },
		{ 370, 236, 44 },
		{ 370, 237, 44 },
		{ 370, 238, 44 },
		{ 370, 239, 44 },
		{ 370, 240, 44 },
		{ 370, 241, 44 },
		{ 370, 242, 44 },
		{ 371, 215, 44 },
		{ 371, 216, 44 },
		{ 371, 217, 44 },
		{ 371, 218, 44 },
		{ 371, 219, 44 },
		{ 371, 220, 44 },
		{ 371, 221, 44 },
		{ 371, 222, 44 },
		{ 371, 223, 44 },
		{ 371, 224, 44 },
		{ 371, 225, 44 },
		{ 371, 226, 44 },
		{ 371, 227, 44 },
		{ 371, 228, 44 },
		{ 371, 229, 44 },
		{ 371, 230, 44 },
		{ 371, 231, 44 },
		{ 371, 232, 44 },
		{ 371, 233, 44 },
		{ 371, 234, 44 },
		{ 371, 235, 44 },
		{ 371, 236, 44 },
		{ 371, 237, 44 },
		{ 371, 238, 44 },
		{ 371, 239, 44 },
		{ 371, 240, 44 },
		{ 371, 241, 44 },
		{ 371, 242, 44 },
		{ 372, 214, 44 },
		{ 372, 215, 44 },
		{ 372, 216, 44 },
		{ 372, 217, 44 },
		{ 372, 218, 44 },
		{ 372, 219, 44 },
		{ 372, 220, 44 },
		{ 372, 221, 44 },
		{ 372, 222, 44 },
		{ 372, 223, 44 },
		{ 372, 224, 44 },
		{ 372, 225, 44 },
		{ 372, 226, 44 },
		{ 372, 227, 44 },
		{ 372, 228, 44 },
		{ 372, 229, 44 },
		{ 372, 230, 44 },
		{ 372, 231, 44 },
		{ 372, 232, 44 },
		{ 372, 233, 44 },
		{ 372, 234, 44 },
		{ 372, 235, 44 },
		{ 372, 236, 44 },
		{ 372, 237, 44 },
		{ 372, 238, 44 },
		{ 372, 239, 44 },
		{ 372, 240, 44 },
		{ 372, 241, 44 },
		{ 372, 242, 44 },
		{ 373, 214, 44 },
		{ 373, 215, 44 },
		{ 373, 216, 44 },
		{ 373, 217, 44 },
		{ 373, 218, 44 },
		{ 373, 219, 44 },
		{ 373, 220, 44 },
		{ 373, 221, 44 },
		{ 373, 222, 44 },
		{ 373, 223, 44 },
		{ 373, 224, 44 },
		{ 373, 225, 44 },
		{ 373, 226, 44 },
		{ 373, 227, 44 },
		{ 373, 228, 44 },
		{ 373, 229, 44 },
		{ 373, 230, 44 },
		{ 373, 231, 44 },
		{ 373, 232, 44 },
		{ 373, 233, 44 },
		{ 373, 234, 44 },
		{ 373, 235, 44 },
		{ 373, 236, 44 },
		{ 373, 237, 44 },
		{ 373, 238, 44 },
		{ 373, 239, 44 },
		{ 373, 240, 44 },
		{ 373, 241, 44 },
		{ 373, 242, 44 },
		{ 374, 214, 44 },
		{ 374, 215, 44 },
		{ 374, 216, 44 },
		{ 374, 217, 44 },
		{ 374, 218, 44 },
		{ 374, 219, 44 },
		{ 374, 220, 44 },
		{ 374, 221, 44 },
		{ 374, 222, 44 },
		{ 374, 223, 44 },
		{ 374, 224, 44 },
		{ 374, 225, 44 },
		{ 374, 226, 44 },
		{ 374, 227, 44 },
		{ 374, 228, 44 },
		{ 374, 229, 44 },
		{ 374, 230, 44 },
		{ 374, 231, 44 },
		{ 374, 232, 44 },
		{ 374, 233, 44 },
		{ 374, 234, 44 },
		{ 374, 235, 44 },
		{ 374, 236, 44 },
		{ 374, 237, 44 },
		{ 374, 238, 44 },
		{ 374, 239, 44 },
		{ 374, 240, 44 },
		{ 374, 241, 44 },
		{ 374, 242, 44 },
		{ 375, 213, 44 },
		{ 375, 214, 44 },
		{ 375, 215, 44 },
		{ 375, 216, 44 },
		{ 375, 217, 44 },
		{ 375, 218, 44 },
		{ 375, 219, 44 },
		{ 375, 220, 44 },
		{ 375, 221, 44 },
		{ 375, 222, 44 },
		{ 375, 223, 44 },
		{ 375, 224, 44 },
		{ 375, 225, 44 },
		{ 375, 226, 44 },
		{ 375, 227, 44 },
		{ 375, 228, 44 },
		{ 375, 229, 44 },
		{ 375, 230, 44 },
		{ 375, 231, 44 },
		{ 375, 232, 44 },
		{ 375, 233, 44 },
		{ 375, 234, 44 },
		{ 375, 235, 44 },
		{ 375, 236, 44 },
		{ 375, 237, 44 },
		{ 375, 238, 44 },
		{ 375, 239, 44 },
		{ 375, 240, 44 },
		{ 375, 241, 44 },
		{ 375, 242, 44 },
		{ 376, 213, 44 },
		{ 376, 214, 44 },
		{ 376, 215, 44 },
		{ 376, 216, 44 },
		{ 376, 217, 44 },
		{ 376, 218, 44 },
		{ 376, 219, 44 },
		{ 376, 220, 44 },
		{ 376, 221, 44 },
		{ 376, 222, 44 },
		{ 376, 223, 44 },
		{ 376, 224, 44 },
		{ 376, 225, 44 },
		{ 376, 226, 44 },
		{ 376, 227, 44 },
		{ 376, 228, 44 },
		{ 376, 229, 44 },
		{ 376, 230, 44 },
		{ 376, 231, 44 },
		{ 376, 232, 44 },
		{ 376, 233, 44 },
		{ 376, 234, 44 },
		{ 376, 235, 44 },
		{ 376, 236, 44 },
		{ 376, 237, 44 },
		{ 376, 238, 44 },
		{ 376, 239, 44 },
		{ 376, 240, 44 },
		{ 376, 241, 44 },
		{ 376, 242, 44 },
		{ 377, 213, 44 },
		{ 377, 214, 44 },
		{ 377, 215, 44 },
		{ 377, 216, 44 },
		{ 377, 217, 44 },
		{ 377, 218, 44 },
		{ 377, 219, 44 },
		{ 377, 220, 44 },
		{ 377, 221, 44 },
		{ 377, 222, 44 },
		{ 377, 223, 44 },
		{ 377, 224, 44 },
		{ 377, 225, 44 },
		{ 377, 226, 44 },
		{ 377, 227, 44 },
		{ 377, 228, 44 },
		{ 377, 229, 44 },
		{ 377, 230, 44 },
		{ 377, 231, 44 },
		{ 377, 232, 44 },
		{ 377, 233, 44 },
		{ 377, 234, 44 },
		{ 377, 235, 44 },
		{ 377, 236, 44 },
		{ 377, 237, 44 },
		{ 377, 238, 44 },
		{ 377, 239, 44 },
		{ 377, 240, 44 },
		{ 377, 241, 44 },
		{ 378, 212, 44 },
		{ 378, 213, 44 },
		{ 378, 214, 44 },
		{ 378, 215, 44 },
		{ 378, 216, 44 },
		{ 378, 217, 44 },
		{ 378, 218, 44 },
		{ 378, 219, 44 },
		{ 378, 220, 44 },
		{ 378, 221, 44 },
		{ 378, 222, 44 },
		{ 378, 223, 44 },
		{ 378, 224, 44 },
		{ 378, 225, 44 },
		{ 378, 226, 44 },
		{ 378, 227, 44 },
		{ 378, 228, 44 },
		{ 378, 229, 44 },
		{ 378, 230, 44 },
		{ 378, 231, 44 },
		{ 378, 232, 44 },
		{ 378, 233, 44 },
		{ 378, 234, 44 },
		{ 378, 235, 44 },
		{ 378, 236, 44 },
		{ 378, 237, 44 },
		{ 378, 238, 44 },
		{ 378, 239, 44 },
		{ 378, 240, 44 },
		{ 378, 241, 44 },
		{ 379, 212, 44 },
		{ 379, 213, 44 },
		{ 379, 214, 44 },
		{ 379, 215, 44 },
		{ 379, 216, 44 },
		{ 379, 217, 44 },
		{ 379, 218, 44 },
		{ 379, 219, 44 },
		{ 379, 220, 44 },
		{ 379, 221, 44 },
		{ 379, 222, 44 },
		{ 379, 223, 44 },
		{ 379, 224, 44 },
		{ 379, 225, 44 },
		{ 379, 226, 44 },
		{ 379, 227, 44 },
		{ 379, 228, 44 },
		{ 379, 229, 44 },
		{ 379, 230, 44 },
		{ 379, 231, 44 },
		{ 379, 232, 44 },
		{ 379, 233, 44 },
		{ 379, 234, 44 },
		{ 379, 235, 44 },
		{ 379, 236, 44 },
		{ 379, 237, 44 },
		{ 379, 238, 44 },
		{ 379, 239, 44 },
		{ 379, 240, 44 },
		{ 379, 241, 44 },
		{ 380, 211, 44 },
		{ 380, 212, 44 },
		{ 380, 213, 44 },
		{ 380, 214, 44 },
		{ 380, 215, 44 },
		{ 380, 216, 44 },
		{ 380, 217, 44 },
		{ 380, 218, 44 },
		{ 380, 219, 44 },
		{ 380, 220, 44 },
		{ 380, 221, 44 },
		{ 380, 222, 44 },
		{ 380, 223, 44 },
		{ 380, 224, 44 },
		{ 380, 225, 44 },
		{ 380, 226, 44 },
		{ 380, 227, 44 },
		{ 380, 228, 44 },
		{ 380, 229, 44 },
		{ 380, 230, 44 },
		{ 380, 231, 44 },
		{ 380, 232, 44 },
		{ 380, 233, 44 },
		{ 380, 234, 44 },
		{ 380, 235, 44 },
		{ 380, 236, 44 },
		{ 380, 237, 44 },
		{ 380, 238, 44 },
		{ 380, 239, 44 },
		{ 380, 240, 44 },
		{ 380, 241, 44 },
		{ 381, 211, 44 },
		{ 381, 212, 44 },
		{ 381, 213, 44 },
		{ 381, 214, 44 },
		{ 381, 215, 44 },
		{ 381, 216, 44 },
		{ 381, 217, 44 },
		{ 381, 218, 44 },
		{ 381, 219, 44 },
		{ 381, 220, 44 },
		{ 381, 221, 44 },
		{ 381, 222, 44 },
		{ 381, 223, 44 },
		{ 381, 224, 44 },
		{ 381, 225, 44 },
		{ 381, 226, 44 },
		{ 381, 227, 44 },
		{ 381, 228, 44 },
		{ 381, 229, 44 },
		{ 381, 230, 44 },
		{ 381, 231, 44 },
		{ 381, 232, 44 },
		{ 381, 233, 44 },
		{ 381, 234, 44 },
		{ 381, 235, 44 },
		{ 381, 236, 44 },
		{ 381, 237, 44 },
		{ 381, 238, 44 },
		{ 381, 239, 44 },
		{ 381, 240, 44 },
		{ 381, 241, 44 },
		{ 382, 211, 44 },
		{ 382, 212, 44 },
		{ 382, 213, 44 },
		{ 382, 214, 44 },
		{ 382, 215, 44 },
		{ 382, 216, 44 },
		{ 382, 217, 44 },
		{ 382, 218, 44 },
		{ 382, 219, 44 },
		{ 382, 220, 44 },
		{ 382, 221, 44 },
		{ 382, 222, 44 },
		{ 382, 223, 44 },
		{ 382, 224, 44 },
		{ 382, 225, 44 },
		{ 382, 226, 44 },
		{ 382, 227, 44 },
		{ 382, 228, 44 },
		{ 382, 229, 44 },
		{ 382, 230, 44 },
		{ 382, 231, 44 },
		{ 382, 232, 44 },
		{ 382, 233, 44 },
		{ 382, 234, 44 },
		{ 382, 235, 44 },
		{ 382, 236, 44 },
		{ 382, 237, 44 },
		{ 382, 238, 44 },
		{ 382, 239, 44 },
		{ 382, 240, 44 },
		{ 382, 241, 44 },
		{ 383, 210, 44 },
		{ 383, 211, 44 },
		{ 383, 212, 44 },
		{ 383, 213, 44 },
		{ 383, 214, 44 },
		{ 383, 215, 44 },
		{ 383, 216, 44 },
		{ 383, 217, 44 },
		{ 383, 218, 44 },
		{ 383, 219, 44 },
		{ 383, 220, 44 },
		{ 383, 221, 44 },
		{ 383, 222, 44 },
		{ 383, 223, 44 },
		{ 383, 224, 44 },
		{ 383, 225, 44 },
		{ 383, 226, 44 },
		{ 383, 227, 44 },
		{ 383, 228, 44 },
		{ 383, 229, 44 },
		{ 383, 230, 44 },
		{ 383, 231, 44 },
		{ 383, 232, 44 },
		{ 383, 233, 44 },
		{ 383, 234, 44 },
		{ 383, 235, 44 },
		{ 383, 236, 44 },
		{ 383, 237, 44 },
		{ 383, 238, 44 },
		{ 383, 239, 44 },
		{ 383, 240, 44 },
		{ 383, 241, 44 },
		{ 384, 210, 44 },
		{ 384, 211, 44 },
		{ 384, 212, 44 },
		{ 384, 213, 44 },
		{ 384, 214, 44 },
		{ 384, 215, 44 },
		{ 384, 216, 44 },
		{ 384, 217, 44 },
		{ 384, 218, 44 },
		{ 384, 219, 44 },
		{ 384, 220, 44 },
		{ 384, 221, 44 },
		{ 384, 222, 44 },
		{ 384, 223, 44 },
		{ 384, 224, 44 },
		{ 384, 225, 44 },
		{ 384, 226, 44 },
		{ 384, 227, 44 },
		{ 384, 228, 44 },
		{ 384, 229, 44 },
		{ 384, 230, 44 },
		{ 384, 231, 44 },
		{ 384, 232, 44 },
		{ 384, 233, 44 },
		{ 384, 234, 44 },
		{ 384, 235, 44 },
		{ 384, 236, 44 },
		{ 384, 237, 44 },
		{ 384, 238, 44 },
		{ 384, 239, 44 },
		{ 384, 240, 44 },
		{ 384, 241, 44 },
		{ 385, 210, 44 },
		{ 385, 211, 44 },
		{ 385, 212, 44 },
		{ 385, 213, 44 },
		{ 385, 214, 44 },
		{ 385, 215, 44 },
		{ 385, 216, 44 },
		{ 385, 217, 44 },
		{ 385, 218, 44 },
		{ 385, 219, 44 },
		{ 385, 220, 44 },
		{ 385, 221, 44 },
		{ 385, 222, 44 },
		{ 385, 223, 44 },
		{ 385, 224, 44 },
		{ 385, 225, 44 },
		{ 385, 226, 44 },
		{ 385, 227, 44 },
		{ 385, 228, 44 },
		{ 385, 229, 44 },
		{ 385, 230, 44 },
		{ 385, 231, 44 },
		{ 385, 232, 44 },
		{ 385, 233, 44 },
		{ 385, 234, 44 },
		{ 385, 235, 44 },
		{ 385, 236, 44 },
		{ 385, 237, 44 },
		{ 385, 238, 44 },
		{ 385, 239, 44 },
		{ 385, 240, 44 },
		{ 386, 209, 44 },
		{ 386, 210, 44 },
		{ 386, 211, 44 },
		{ 386, 212, 44 },
		{ 386, 213, 44 },
		{ 386, 214, 44 },
		{ 386, 215, 44 },
		{ 386, 216, 44 },
		{ 386, 217, 44 },
		{ 386, 218, 44 },
		{ 386, 219, 44 },
		{ 386, 220, 44 },
		{ 386, 221, 44 },
		{ 386, 222, 44 },
		{ 386, 223, 44 },
		{ 386, 224, 44 },
		{ 386, 225, 44 },
		{ 386, 226, 44 },
		{ 386, 227, 44 },
		{ 386, 228, 44 },
		{ 386, 229, 44 },
		{ 386, 230, 44 },
		{ 386, 231, 44 },
		{ 386, 232, 44 },
		{ 386, 233, 44 },
		{ 386, 234, 44 },
		{ 386, 235, 44 },
		{ 386, 236, 44 },
		{ 386, 237, 44 },
		{ 386, 238, 44 },
		{ 386, 239, 44 },
		{ 386, 240, 44 },
		{ 387, 209, 44 },
		{ 387, 210, 44 },
		{ 387, 211, 44 },
		{ 387, 212, 44 },
		{ 387, 213, 44 },
		{ 387, 214, 44 },
		{ 387, 215, 44 },
		{ 387, 216, 44 },
		{ 387, 217, 44 },
		{ 387, 218, 44 },
		{ 387, 219, 44 },
		{ 387, 220, 44 },
		{ 387, 221, 44 },
		{ 387, 222, 44 },
		{ 387, 223, 44 },
		{ 387, 224, 44 },
		{ 387, 225, 44 },
		{ 387, 226, 44 },
		{ 387, 227, 44 },
		{ 387, 228, 44 },
		{ 387, 229, 44 },
		{ 387, 230, 44 },
		{ 387, 231, 44 },
		{ 387, 232, 44 },
		{ 387, 233, 44 },
		{ 387, 234, 44 },
		{ 387, 235, 44 },
		{ 387, 236, 44 },
		{ 387, 237, 44 },
		{ 387, 238, 44 },
		{ 387, 239, 44 },
		{ 387, 240, 44 },
		{ 388, 208, 44 },
		{ 388, 209, 44 },
		{ 388, 210, 44 },
		{ 388, 211, 44 },
		{ 388, 212, 44 },
		{ 388, 213, 44 },
		{ 388, 214, 44 },
		{ 388, 215, 44 },
		{ 388, 216, 44 },
		{ 388, 217, 44 },
		{ 388, 218, 44 },
		{ 388, 219, 44 },
		{ 388, 220, 44 },
		{ 388, 221, 44 },
		{ 388, 222, 44 },
		{ 388, 223, 44 },
		{ 388, 224, 44 },
		{ 388, 225, 44 },
		{ 388, 226, 44 },
		{ 388, 227, 44 },
		{ 388, 228, 44 },
		{ 388, 229, 44 },
		{ 388, 230, 44 },
		{ 388, 231, 44 },
		{ 388, 232, 44 },
		{ 388, 233, 44 },
		{ 388, 234, 44 },
		{ 388, 235, 44 },
		{ 388, 236, 44 },
		{ 388, 237, 44 },
		{ 388, 238, 44 },
		{ 388, 239, 44 },
		{ 388, 240, 44 },
		{ 389, 208, 44 },
		{ 389, 209, 44 },
		{ 389, 210, 44 },
		{ 389, 211, 44 },
		{ 389, 212, 44 },
		{ 389, 213, 44 },
		{ 389, 214, 44 },
		{ 389, 215, 44 },
		{ 389, 216, 44 },
		{ 389, 217, 44 },
		{ 389, 218, 44 },
		{ 389, 219, 44 },
		{ 389, 220, 44 },
		{ 389, 221, 44 },
		{ 389, 222, 44 },
		{ 389, 223, 44 },
		{ 389, 224, 44 },
		{ 389, 225, 44 },
		{ 389, 226, 44 },
		{ 389, 227, 44 },
		{ 389, 228, 44 },
		{ 389, 229, 44 },
		{ 389, 230, 44 },
		{ 389, 231, 44 },
		{ 389, 232, 44 },
		{ 389, 233, 44 },
		{ 389, 234, 44 },
		{ 389, 235, 44 },
		{ 389, 236, 44 },
		{ 389, 237, 44 },
		{ 389, 238, 44 },
		{ 389, 239, 44 },
		{ 389, 240, 44 },
		{ 390, 208, 44 },
		{ 390, 209, 44 },
		{ 390, 210, 44 },
		{ 390, 211, 44 },
		{ 390, 212, 44 },
		{ 390, 213, 44 },
		{ 390, 214, 44 },
		{ 390, 215, 44 },
		{ 390, 216, 44 },
		{ 390, 217, 44 },
		{ 390, 218, 44 },
		{ 390, 219, 44 },
		{ 390, 220, 44 },
		{ 390, 221, 44 },
		{ 390, 222, 44 },
		{ 390, 223, 44 },
		{ 390, 224, 44 },
		{ 390, 225, 44 },
		{ 390, 226, 44 },
		{ 390, 227, 44 },
		{ 390, 228, 44 },
		{ 390, 229, 44 },
		{ 390, 230, 44 },
		{ 390, 231, 44 },
		{ 390, 232, 44 },
		{ 390, 233, 44 },
		{ 390, 234, 44 },
		{ 390, 235, 44 },
		{ 390, 236, 44 },
		{ 390, 237, 44 },
		{ 390, 238, 44 },
		{ 390, 239, 44 },
		{ 390, 240, 44 },
		{ 391, 207, 44 },
		{ 391, 208, 44 },
		{ 391, 209, 44 },
		{ 391, 210, 44 },
		{ 391, 211, 44 },
		{ 391, 212, 44 },
		{ 391, 213, 44 },
		{ 391, 214, 44 },
		{ 391, 215, 44 },
		{ 391, 216, 44 },
		{ 391, 217, 44 },
		{ 391, 218, 44 },
		{ 391, 219, 44 },
		{ 391, 220, 44 },
		{ 391, 221, 44 },
		{ 391, 222, 44 },
		{ 391, 223, 44 },
		{ 391, 224, 44 },
		{ 391, 225, 44 },
		{ 391, 226, 44 },
		{ 391, 227, 44 },
		{ 391, 228, 44 },
		{ 391, 229, 44 },
		{ 391, 230, 44 },
		{ 391, 231, 44 },
		{ 391, 232, 44 },
		{ 391, 233, 44 },
		{ 391, 234, 44 },
		{ 391, 235, 44 },
		{ 391, 236, 44 },
		{ 391, 237, 44 },
		{ 391, 238, 44 },
		{ 391, 239, 44 },
		{ 391, 240, 44 },
		{ 392, 207, 44 },
		{ 392, 208, 44 },
		{ 392, 209, 44 },
		{ 392, 210, 44 },
		{ 392, 211, 44 },
		{ 392, 212, 44 },
		{ 392, 213, 44 },
		{ 392, 214, 44 },
		{ 392, 215, 44 },
		{ 392, 216, 44 },
		{ 392, 217, 44 },
		{ 392, 218, 44 },
		{ 392, 219, 44 },
		{ 392, 220, 44 },
		{ 392, 221, 44 },
		{ 392, 222, 44 },
		{ 392, 223, 44 },
		{ 392, 224, 44 },
		{ 392, 225, 44 },
		{ 392, 226, 44 },
		{ 392, 227, 44 },
		{ 392, 228, 44 },
		{ 392, 229, 44 },
		{ 392, 230, 44 },
		{ 392, 231, 44 },
		{ 392, 232, 44 },
		{ 392, 233, 44 },
		{ 392, 234, 44 },
		{ 392, 235, 44 },
		{ 392, 236, 44 },
		{ 392, 237, 44 },
		{ 392, 238, 44 },
		{ 392, 239, 44 },
		{ 392, 240, 44 },
		{ 393, 207, 44 },
		{ 393, 208, 44 },
		{ 393, 209, 44 },
		{ 393, 210, 44 },
		{ 393, 211, 44 },
		{ 393, 212, 44 },
		{ 393, 213, 44 },
		{ 393, 214, 44 },
		{ 393, 215, 44 },
		{ 393, 216, 44 },
		{ 393, 217, 44 },
		{ 393, 218, 44 },
		{ 393, 219, 44 },
		{ 393, 220, 44 },
		{ 393, 221, 44 },
		{ 393, 222, 44 },
		{ 393, 223, 44 },
		{ 393, 224, 44 },
		{ 393, 225, 44 },
		{ 393, 226, 44 },
		{ 393, 227, 44 },
		{ 393, 228, 44 },
		{ 393, 229, 44 },
		{ 393, 230, 44 },
		{ 393, 231, 44 },
		{ 393, 232, 44 },
		{ 393, 233, 44 },
		{ 393, 234, 44 },
		{ 393, 235, 44 },
		{ 393, 236, 44 },
		{ 393, 237, 44 },
		{ 393, 238, 44 },
		{ 393, 239, 44 },
		{ 394, 206, 44 },
		{ 394, 207, 44 },
		{ 394, 208, 44 },
		{ 394, 209, 44 },
		{ 394, 210, 44 },
		{ 394, 211, 44 },
		{ 394, 212, 44 },
		{ 394, 213, 44 },
		{ 394, 214, 44 },
		{ 394, 215, 44 },
		{ 394, 216, 44 },
		{ 394, 217, 44 },
		{ 394, 218, 44 },
		{ 394, 219, 44 },
		{ 394, 220, 44 },
		{ 394, 221, 44 },
		{ 394, 222, 44 },
		{ 394, 223, 44 },
		{ 394, 224, 44 },
		{ 394, 225, 44 },
		{ 394, 226, 44 },
		{ 394, 227, 44 },
		{ 394, 228, 44 },
		{ 394, 229, 44 },
		{ 394, 230, 44 },
		{ 394, 231, 44 },
		{ 394, 232, 44 },
		{ 394, 233, 44 },
		{ 394, 234, 44 },
		{ 394, 235, 44 },
		{ 394, 236, 44 },
		{ 394, 237, 44 },
		{ 394, 238, 44 },
		{ 394, 239, 44 },
		{ 395, 206, 44 },
		{ 395, 207, 44 },
		{ 395, 208, 44 },
		{ 395, 209, 44 },
		{ 395, 210, 44 },
		{ 395, 211, 44 },
		{ 395, 212, 44 },
		{ 395, 213, 44 },
		{ 395, 214, 44 },
		{ 395, 215, 44 },
		{ 395, 216, 44 },
		{ 395, 217, 44 },
		{ 395, 218, 44 },
		{ 395, 219, 44 },
		{ 395, 220, 44 },
		{ 395, 221, 44 },
		{ 395, 222, 44 },
		{ 395, 223, 44 },
		{ 395, 224, 44 },
		{ 395, 225, 44 },
		{ 395, 226, 44 },
		{ 395, 227, 44 },
		{ 395, 228, 44 },
		{ 395, 229, 44 },
		{ 395, 230, 44 },
		{ 395, 231, 44 },
		{ 395, 232, 44 },
		{ 395, 233, 44 },
		{ 395, 234, 44 },
		{ 395, 235, 44 },
		{ 395, 236, 44 },
		{ 395, 237, 44 },
		{ 395, 238, 44 },
		{ 395, 239, 44 },
		{ 396, 205, 44 },
		{ 396, 206, 44 },
		{ 396, 207, 44 },
		{ 396, 208, 44 },
		{ 396, 209, 44 },
		{ 396, 210, 44 },
		{ 396, 211, 44 },
		{ 396, 212, 44 },
		{ 396, 213, 44 },
		{ 396, 214, 44 },
		{ 396, 215, 44 },
		{ 396, 216, 44 },
		{ 396, 217, 44 },
		{ 396, 218, 44 },
		{ 396, 219, 44 },
		{ 396, 220, 44 },
		{ 396, 221, 44 },
		{ 396, 222, 44 },
		{ 396, 223, 44 },
		{ 396, 224, 44 },
		{ 396, 225, 44 },
		{ 396, 226, 44 },
		{ 396, 227, 44 },
		{ 396, 228, 44 },
		{ 396, 229, 44 },
		{ 396, 230, 44 },
		{ 396, 231, 44 },
		{ 396, 232, 44 },
		{ 396, 233, 44 },
		{ 396, 234, 44 },
		{ 396, 235, 44 },
		{ 396, 236, 44 },
		{ 396, 237, 44 },
		{ 396, 238, 44 },
		{ 396, 239, 44 },
		{ 397, 205, 44 },
		{ 397, 206, 44 },
		{ 397, 207, 44 },
		{ 397, 208, 44 },
		{ 397, 209, 44 },
		{ 397, 210, 44 },
		{ 397, 211, 44 },
		{ 397, 212, 44 },
		{ 397, 213, 44 },
		{ 397, 214, 44 },
		{ 397, 215, 44 },
		{ 397, 216, 44 },
		{ 397, 217, 44 },
		{ 397, 218, 44 },
		{ 397, 219, 44 },
		{ 397, 220, 44 },
		{ 397, 221, 44 },
		{ 397, 222, 44 },
		{ 397, 223, 44 },
		{ 397, 224, 44 },
		{ 397, 225, 44 },
		{ 397, 226, 44 },
		{ 397, 227, 44 },
		{ 397, 228, 44 },
		{ 397, 229, 44 },
		{ 397, 230, 44 },
		{ 397, 231, 44 },
		{ 397, 232, 44 },
		{ 397, 233, 44 },
		{ 397, 234, 44 },
		{ 397, 235, 44 },
		{ 397, 236, 44 },
		{ 397, 237, 44 },
		{ 397, 238, 44 },
		{ 397, 239, 44 },
		{ 398, 205, 44 },
		{ 398, 206, 44 },
		{ 398, 207, 44 },
		{ 398, 208, 44 },
		{ 398, 209, 44 },
		{ 398, 210, 44 },
		{ 398, 211, 44 },
		{ 398, 212, 44 },
		{ 398, 213, 44 },
		{ 398, 214, 44 },
		{ 398, 215, 44 },
		{ 398, 216, 44 },
		{ 398, 217, 44 },
		{ 398, 218, 44 },
		{ 398, 219, 44 },
		{ 398, 220, 44 },
		{ 398, 221, 44 },
		{ 398, 222, 44 },
		{ 398, 223, 44 },
		{ 398, 224, 44 },
		{ 398, 225, 44 },
		{ 398, 226, 44 },
		{ 398, 227, 44 },
		{ 398, 228, 44 },
		{ 398, 229, 44 },
		{ 398, 230, 44 },
		{ 398, 231, 44 },
		{ 398, 232, 44 },
		{ 398, 233, 44 },
		{ 398, 234, 44 },
		{ 398, 235, 44 },
		{ 398, 236, 44 },
		{ 398, 237, 44 },
		{ 398, 238, 44 },
		{ 398, 239, 44 },
		{ 399, 204, 44 },
		{ 399, 205, 44 },
		{ 399, 206, 44 },
		{ 399, 207, 44 },
		{ 399, 208, 44 },
		{ 399, 209, 44 },
		{ 399, 210, 44 },
		{ 399, 211, 44 },
		{ 399, 212, 44 },
		{ 399, 213, 44 },
		{ 399, 214, 44 },
		{ 399, 215, 44 },
		{ 399, 216, 44 },
		{ 399, 217, 44 },
		{ 399, 218, 44 },
		{ 399, 219, 44 },
		{ 399, 220, 44 },
		{ 399, 221, 44 },
		{ 399, 222, 44 },
		{ 399, 223, 44 },
		{ 399, 224, 44 },
		{ 399, 225, 44 },
		{ 399, 226, 44 },
		{ 399, 227, 44 },
		{ 399, 228, 44 },
		{ 399, 229, 44 },
		{ 399, 230, 44 },
		{ 399, 231, 44 },
		{ 399, 232, 44 },
		{ 399, 233, 44 },
		{ 399, 234, 44 },
		{ 399, 235, 44 },
		{ 399, 236, 44 },
		{ 399, 237, 44 },
		{ 399, 238, 44 },
		{ 399, 239, 44 },
		{ 400, 204, 44 },
		{ 400, 205, 44 },
		{ 400, 206, 44 },
		{ 400, 207, 44 },
		{ 400, 208, 44 },
		{ 400, 209, 44 },
		{ 400, 210, 44 },
		{ 400, 211, 44 },
		{ 400, 212, 44 },
		{ 400, 213, 44 },
		{ 400, 214, 44 },
		{ 400, 215, 44 },
		{ 400, 216, 44 },
		{ 400, 217, 44 },
		{ 400, 218, 44 },
		{ 400, 219, 44 },
		{ 400, 220, 44 },
		{ 400, 221, 44 },
		{ 400, 222, 44 },
		{ 400, 223, 44 },
		{ 400, 224, 44 },
		{ 400, 225, 44 },
		{ 400, 226, 44 },
		{ 400, 227, 44 },
		{ 400, 228, 44 },
		{ 400, 229, 44 },
		{ 400, 230, 44 },
		{ 400, 231, 44 },
		{ 400, 232, 44 },
		{ 400, 233, 44 },
		{ 400, 234, 44 },
		{ 400, 235, 44 },
		{ 400, 236, 44 },
		{ 400, 237, 44 },
		{ 400, 238, 44 },
		{ 400, 239, 44 },
		{ 401, 204, 44 },
		{ 401, 205, 44 },
		{ 401, 206, 44 },
		{ 401, 207, 44 },
		{ 401, 208, 44 },
		{ 401, 209, 44 },
		{ 401, 210, 44 },
		{ 401, 211, 44 },
		{ 401, 212, 44 },
		{ 401, 213, 44 },
		{ 401, 214, 44 },
		{ 401, 215, 44 },
		{ 401, 216, 44 },
		{ 401, 217, 44 },
		{ 401, 218, 44 },
		{ 401, 219, 44 },
		{ 401, 220, 44 },
		{ 401, 221, 44 },
		{ 401, 222, 44 },
		{ 401, 223, 44 },
		{ 401, 224, 44 },
		{ 401, 225, 44 },
		{ 401, 226, 44 },
		{ 401, 227, 44 },
		{ 401, 228, 44 },
		{ 401, 229, 44 },
		{ 401, 230, 44 },
		{ 401, 231, 44 },
		{ 401, 232, 44 },
		{ 401, 233, 44 },
		{ 401, 234, 44 },
		{ 401, 235, 44 },
		{ 401, 236, 44 },
		{ 401, 237, 44 },
		{ 401, 238, 44 },
		{ 402, 203, 44 },
		{ 402, 204, 44 },
		{ 402, 205, 44 },
		{ 402, 206, 44 },
		{ 402, 207, 44 },
		{ 402, 208, 44 },
		{ 402, 209, 44 },
		{ 402, 210, 44 },
		{ 402, 211, 44 },
		{ 402, 212, 44 },
		{ 402, 213, 44 },
		{ 402, 214, 44 },
		{ 402, 215, 44 },
		{ 402, 216, 44 },
		{ 402, 217, 44 },
		{ 402, 218, 44 },
		{ 402, 219, 44 },
		{ 402, 220, 44 },
		{ 402, 221, 44 },
		{ 402, 222, 44 },
		{ 402, 223, 44 },
		{ 402, 224, 44 },
		{ 402, 225, 44 },
		{ 402, 226, 44 },
		{ 402, 227, 44 },
		{ 402, 228, 44 },
		{ 402, 229, 44 },
		{ 402, 230, 44 },
		{ 402, 231, 44 },
		{ 402, 232, 44 },
		{ 402, 233, 44 },
		{ 402, 234, 44 },
		{ 402, 235, 44 },
		{ 402, 236, 44 },
		{ 402, 237, 44 },
		{ 402, 238, 44 },
		{ 403, 203, 44 },
		{ 403, 204, 44 },
		{ 403, 205, 44 },
		{ 403, 206, 44 },
		{ 403, 207, 44 },
		{ 403, 208, 44 },
		{ 403, 209, 44 },
		{ 403, 210, 44 },
		{ 403, 211, 44 },
		{ 403, 212, 44 },
		{ 403, 213, 44 },
		{ 403, 214, 44 },
		{ 403, 215, 44 },
		{ 403, 216, 44 },
		{ 403, 217, 44 },
		{ 403, 218, 44 },
		{ 403, 219, 44 },
		{ 403, 220, 44 },
		{ 403, 221, 44 },
		{ 403, 222, 44 },
		{ 403, 223, 44 },
		{ 403, 224, 44 },
		{ 403, 225, 44 },
		{ 403, 226, 44 },
		{ 403, 227, 44 },
		{ 403, 228, 44 },
		{ 403, 229, 44 },
		{ 403, 230, 44 },
		{ 403, 231, 44 },
		{ 403, 232, 44 },
		{ 403, 233, 44 },
		{ 403, 234, 44 },
		{ 403, 235, 44 },
		{ 403, 236, 44 },
		{ 403, 237, 44 },
		{ 403, 238, 44 },
		{ 404, 202, 44 },
		{ 404, 203, 44 },
		{ 404, 204, 44 },
		{ 404, 205, 44 },
		{ 404, 206, 44 },
		{ 404, 207, 44 },
		{ 404, 208, 44 },
		{ 404, 209, 44 },
		{ 404, 210, 44 },
		{ 404, 211, 44 },
		{ 404, 212, 44 },
		{ 404, 213, 44 },
		{ 404, 214, 44 },
		{ 404, 215, 44 },
		{ 404, 216, 44 },
		{ 404, 217, 44 },
		{ 404, 218, 44 },
		{ 404, 219, 44 },
		{ 404, 220, 44 },
		{ 404, 221, 44 },
		{ 404, 222, 44 },
		{ 404, 223, 44 },
		{ 404, 224, 44 },
		{ 404, 225, 44 },
		{ 404, 226, 44 },
		{ 404, 227, 44 },
		{ 404, 228, 44 },
		{ 404, 229, 44 },
		{ 404, 230, 44 },
		{ 404, 231, 44 },
		{ 404, 232, 44 },
		{ 404, 233, 44 },
		{ 404, 234, 44 },
		{ 404, 235, 44 },
		{ 404, 236, 44 },
		{ 404, 237, 44 },
		{ 404, 238, 44 },
		{ 405, 202, 44 },
		{ 405, 203, 44 },
		{ 405, 204, 44 },
		{ 405, 205, 44 },
		{ 405, 206, 44 },
		{ 405, 207, 44 },
		{ 405, 208, 44 },
		{ 405, 209, 44 },
		{ 405, 210, 44 },
		{ 405, 211, 44 },
		{ 405, 212, 44 },
		{ 405, 213, 44 },
		{ 405, 214, 44 },
		{ 405, 215, 44 },
		{ 405, 216, 44 },
		{ 405, 217, 44 },
		{ 405, 218, 44 },
		{ 405, 219, 44 },
		{ 405, 220, 44 },
		{ 405, 221, 44 },
		{ 405, 222, 44 },
		{ 405, 223, 44 },
		{ 405, 224, 44 },
		{ 405, 225, 44 },
		{ 405, 226, 44 },
		{ 405, 227, 44 },
		{ 405, 228, 44 },
		{ 405, 229, 44 },
		{ 405, 230, 44 },
		{ 405, 231, 44 },
		{ 405, 232, 44 },
		{ 405, 233, 44 },
		{ 405, 234, 44 },
		{ 405, 235, 44 },
		{ 405, 236, 44 },
		{ 405, 237, 44 },
		{ 405, 238, 44 },
		{ 406, 202, 44 },
		{ 406, 203, 44 },
		{ 406, 204, 44 },
		{ 406, 205, 44 },
		{ 406, 206, 44 },
		{ 406, 207, 44 },
		{ 406, 208, 44 },
		{ 406, 209, 44 },
		{ 406, 210, 44 },
		{ 406, 211, 44 },
		{ 406, 212, 44 },
		{ 406, 213, 44 },
		{ 406, 214, 44 },
		{ 406, 215, 44 },
		{ 406, 216, 44 },
		{ 406, 217, 44 },
		{ 406, 218, 44 },
		{ 406, 219, 44 },
		{ 406, 220, 44 },
		{ 406, 221, 44 },
		{ 406, 222, 44 },
		{ 406, 223, 44 },
		{ 406, 224, 44 },
		{ 406, 225, 44 },
		{ 406, 226, 44 },
		{ 406, 227, 44 },
		{ 406, 228, 44 },
		{ 406, 229, 44 },
		{ 406, 230, 44 },
		{ 406, 231, 44 },
		{ 406, 232, 44 },
		{ 406, 233, 44 },
		{ 406, 234, 44 },
		{ 406, 235, 44 },
		{ 406, 236, 44 },
		{ 406, 237, 44 },
		{ 406, 238, 44 },
		{ 407, 201, 44 },
		{ 407, 202, 44 },
		{ 407, 203, 44 },
		{ 407, 204, 44 },
		{ 407, 205, 44 },
		{ 407, 206, 44 },
		{ 407, 207, 44 },
		{ 407, 208, 44 },
		{ 407, 209, 44 },
		{ 407, 210, 44 },
		{ 407, 211, 44 },
		{ 407, 212, 44 },
		{ 407, 213, 44 },
		{ 407, 214, 44 },
		{ 407, 215, 44 },
		{ 407, 216, 44 },
		{ 407, 217, 44 },
		{ 407, 218, 44 },
		{ 407, 219, 44 },
		{ 407, 220, 44 },
		{ 407, 221, 44 },
		{ 407, 222, 44 },
		{ 407, 223, 44 },
		{ 407, 224, 44 },
		{ 407, 225, 44 },
		{ 407, 226, 44 },
		{ 407, 227, 44 },
		{ 407, 228, 44 },
		{ 407, 229, 44 },
		{ 407, 230, 44 },
		{ 407, 231, 44 },
		{ 407, 232, 44 },
		{ 407, 233, 44 },
		{ 407, 234, 44 },
		{ 407, 235, 44 },
		{ 407, 236, 44 },
		{ 407, 237, 44 },
		{ 407, 238, 44 },
		{ 408, 201, 44 },
		{ 408, 202, 44 },
		{ 408, 203, 44 },
		{ 408, 204, 44 },
		{ 408, 205, 44 },
		{ 408, 206, 44 },
		{ 408, 207, 44 },
		{ 408, 208, 44 },
		{ 408, 209, 44 },
		{ 408, 210, 44 },
		{ 408, 211, 44 },
		{ 408, 212, 44 },
		{ 408, 213, 44 },
		{ 408, 214, 44 },
		{ 408, 215, 44 },
		{ 408, 216, 44 },
		{ 408, 217, 44 },
		{ 408, 218, 44 },
		{ 408, 219, 44 },
		{ 408, 220, 44 },
		{ 408, 221, 44 },
		{ 408, 222, 44 },
		{ 408, 223, 44 },
		{ 408, 224, 44 },
		{ 408, 225, 44 },
		{ 408, 226, 44 },
		{ 408, 227, 44 },
		{ 408, 228, 44 },
		{ 408, 229, 44 },
		{ 408, 230, 44 },
		{ 408, 231, 44 },
		{ 408, 232, 44 },
		{ 408, 233, 44 },
		{ 408, 234, 44 },
		{ 408, 235, 44 },
		{ 408, 236, 44 },
		{ 408, 237, 44 },
		{ 408, 238, 44 },
		{ 409, 201, 44 },
		{ 409, 202, 44 },
		{ 409, 203, 44 },
		{ 409, 204, 44 },
		{ 409, 205, 44 },
		{ 409, 206, 44 },
		{ 409, 207, 44 },
		{ 409, 208, 44 },
		{ 409, 209, 44 },
		{ 409, 210, 44 },
		{ 409, 211, 44 },
		{ 409, 212, 44 },
		{ 409, 213, 44 },
		{ 409, 214, 44 },
		{ 409, 215, 44 },
		{ 409, 216, 44 },
		{ 409, 217, 44 },
		{ 409, 218, 44 },
		{ 409, 219, 44 },
		{ 409, 220, 44 },
		{ 409, 221, 44 },
		{ 409, 222, 44 },
		{ 409, 223, 44 },
		{ 409, 224, 44 },
		{ 409, 225, 44 },
		{ 409, 226, 44 },
		{ 409, 227, 44 },
		{ 409, 228, 44 },
		{ 409, 229, 44 },
		{ 409, 230, 44 },
		{ 409, 231, 44 },
		{ 409, 232, 44 },
		{ 409, 233, 44 },
		{ 409, 234, 44 },
		{ 409, 235, 44 },
		{ 409, 236, 44 },
		{ 409, 237, 44 },
		{ 410, 200, 44 },
		{ 410, 201, 44 },
		{ 410, 202, 44 },
		{ 410, 203, 44 },
		{ 410, 204, 44 },
		{ 410, 205, 44 },
		{ 410, 206, 44 },
		{ 410, 207, 44 },
		{ 410, 208, 44 },
		{ 410, 209, 44 },
		{ 410, 210, 44 },
		{ 410, 211, 44 },
		{ 410, 212, 44 },
		{ 410, 213, 44 },
		{ 410, 214, 44 },
		{ 410, 215, 44 },
		{ 410, 216, 44 },
		{ 410, 217, 44 },
		{ 410, 218, 44 },
		{ 410, 219, 44 },
		{ 410, 220, 44 },
		{ 410, 221, 44 },
		{ 410, 222, 44 },
		{ 410, 223, 44 },
		{ 410, 224, 44 },
		{ 410, 225, 44 },
		{ 410, 226, 44 },
		{ 410, 227, 44 },
		{ 410, 228, 44 },
		{ 410, 229, 44 },
		{ 410, 230, 44 },
		{ 410, 231, 44 },
		{ 410, 232, 44 },
		{ 410, 233, 44 },
		{ 410, 234, 44 },
		{ 410, 235, 44 },
		{ 410, 236, 44 },
		{ 410, 237, 44 },
		{ 411, 200, 44 },
		{ 411, 201, 44 },
		{ 411, 202, 44 },
		{ 411, 203, 44 },
		{ 411, 204, 44 },
		{ 411, 205, 44 },
		{ 411, 206, 44 },
		{ 411, 207, 44 },
		{ 411, 208, 44 },
		{ 411, 209, 44 },
		{ 411, 210, 44 },
		{ 411, 211, 44 },
		{ 411, 212, 44 },
		{ 411, 213, 44 },
		{ 411, 214, 44 },
		{ 411, 215, 44 },
		{ 411, 216, 44 },
		{ 411, 217, 44 },
		{ 411, 218, 44 },
		{ 411, 219, 44 },
		{ 411, 220, 44 },
		{ 411, 221, 44 },
		{ 411, 222, 44 },
		{ 411, 223, 44 },
		{ 411, 224, 44 },
		{ 411, 225, 44 },
		{ 411, 226, 44 },
		{ 411, 227, 44 },
		{ 411, 228, 44 },
		{ 411, 229, 44 },
		{ 411, 230, 44 },
		{ 411, 231, 44 },
		{ 411, 232, 44 },
		{ 411, 233, 44 },
		{ 411, 234, 44 },
		{ 411, 235, 44 },
		{ 411, 236, 44 },
		{ 411, 237, 44 },
		{ 412, 199, 44 },
		{ 412, 200, 44 },
		{ 412, 201, 44 },
		{ 412, 202, 44 },
		{ 412, 203, 44 },
		{ 412, 204, 44 },
		{ 412, 205, 44 },
		{ 412, 206, 44 },
		{ 412, 207, 44 },
		{ 412, 208, 44 },
		{ 412, 209, 44 },
		{ 412, 210, 44 },
		{ 412, 211, 44 },
		{ 412, 212, 44 },
		{ 412, 213, 44 },
		{ 412, 214, 44 },
		{ 412, 215, 44 },
		{ 412, 216, 44 },
		{ 412, 217, 44 },
		{ 412, 218, 44 },
		{ 412, 219, 44 },
		{ 412, 220, 44 },
		{ 412, 221, 44 },
		{ 412, 222, 44 },
		{ 412, 223, 44 },
		{ 412, 224, 44 },
		{ 412, 225, 44 },
		{ 412, 226, 44 },
		{ 412, 227, 44 },
		{ 412, 228, 44 },
		{ 412, 229, 44 },
		{ 412, 230, 44 },
		{ 412, 231, 44 },
		{ 412, 232, 44 },
		{ 412, 233, 44 },
		{ 412, 234, 44 },
		{ 412, 235, 44 },
		{ 412, 236, 44 },
		{ 412, 237, 44 },
		{ 413, 199, 44 },
		{ 413, 200, 44 },
		{ 413, 201, 44 },
		{ 413, 202, 44 },
		{ 413, 203, 44 },
		{ 413, 204, 44 },
		{ 413, 205, 44 },
		{ 413, 206, 44 },
		{ 413, 207, 44 },
		{ 413, 208, 44 },
		{ 413, 209, 44 },
		{ 413, 210, 44 },
		{ 413, 211, 44 },
		{ 413, 212, 44 },
		{ 413, 213, 44 },
		{ 413, 214, 44 },
		{ 413, 215, 44 },
		{ 413, 216, 44 },
		{ 413, 217, 44 },
		{ 413, 218, 44 },
		{ 413, 219, 44 },
		{ 413, 220, 44 },
		{ 413, 221, 44 },
		{ 413, 222, 44 },
		{ 413, 223, 44 },
		{ 413, 224, 44 },
		{ 413, 225, 44 },
		{ 413, 226, 44 },
		{ 413, 227, 44 },
		{ 413, 228, 44 },
		{ 413, 229, 44 },
		{ 413, 230, 44 },
		{ 413, 231, 44 },
		{ 413, 232, 44 },
		{ 413, 233, 44 },
		{ 413, 234, 44 },
		{ 413, 235, 44 },
		{ 413, 236, 44 },
		{ 413, 237, 44 },
		{ 414, 199, 44 },
		{ 414, 200, 44 },
		{ 414, 201, 44 },
		{ 414, 202, 44 },
		{ 414, 203, 44 },
		{ 414, 204, 44 },
		{ 414, 205, 44 },
		{ 414, 206, 44 },
		{ 414, 207, 44 },
		{ 414, 208, 44 },
		{ 414, 209, 44 },
		{ 414, 210, 44 },
		{ 414, 211, 44 },
		{ 414, 212, 44 },
		{ 414, 213, 44 },
		{ 414, 214, 44 },
		{ 414, 215, 44 },
		{ 414, 216, 44 },
		{ 414, 217, 44 },
		{ 414, 218, 44 },
		{ 414, 219, 44 },
		{ 414, 220, 44 },
		{ 414, 221, 44 },
		{ 414, 222, 44 },
		{ 414, 223, 44 },
		{ 414, 224, 44 },
		{ 414, 225, 44 },
		{ 414, 226, 44 },
		{ 414, 227, 44 },
		{ 414, 228, 44 },
		{ 414, 229, 44 },
		{ 414, 230, 44 },
		{ 414, 231, 44 },
		{ 414, 232, 44 },
		{ 414, 233, 44 },
		{ 414, 234, 44 },
		{ 414, 235, 44 },
		{ 414, 236, 44 },
		{ 414, 237, 44 },
		{ 415, 198, 44 },
		{ 415, 199, 44 },
		{ 415, 200, 44 },
		{ 415, 201, 44 },
		{ 415, 202, 44 },
		{ 415, 203, 44 },
		{ 415, 204, 44 },
		{ 415, 205, 44 },
		{ 415, 206, 44 },
		{ 415, 207, 44 },
		{ 415, 208, 44 },
		{ 415, 209, 44 },
		{ 415, 210, 44 },
		{ 415, 211, 44 },
		{ 415, 212, 44 },
		{ 415, 213, 44 },
		{ 415, 214, 44 },
		{ 415, 215, 44 },
		{ 415, 216, 44 },
		{ 415, 217, 44 },
		{ 415, 218, 44 },
		{ 415, 219, 44 },
		{ 415, 220, 44 },
		{ 415, 221, 44 },
		{ 415, 222, 44 },
		{ 415, 223, 44 },
		{ 415, 224, 44 },
		{ 415, 225, 44 },
		{ 415, 226, 44 },
		{ 415, 227, 44 },
		{ 415, 228, 44 },
		{ 415, 229, 44 },
		{ 415, 230, 44 },
		{ 415, 231, 44 },
		{ 415, 232, 44 },
		{ 415, 233, 44 },
		{ 415, 234, 44 },
		{ 415, 235, 44 },
		{ 415, 236, 44 },
		{ 415, 237, 44 },
		{ 416, 198, 44 },
		{ 416, 199, 44 },
		{ 416, 200, 44 },
		{ 416, 201, 44 },
		{ 416, 202, 44 },
		{ 416, 203, 44 },
		{ 416, 204, 44 },
		{ 416, 205, 44 },
		{ 416, 206, 44 },
		{ 416, 207, 44 },
		{ 416, 208, 44 },
		{ 416, 209, 44 },
		{ 416, 210, 44 },
		{ 416, 211, 44 },
		{ 416, 212, 44 },
		{ 416, 213, 44 },
		{ 416, 214, 44 },
		{ 416, 215, 44 },
		{ 416, 216, 44 },
		{ 416, 217, 44 },
		{ 416, 218, 44 },
		{ 416, 219, 44 },
		{ 416, 220, 44 },
		{ 416, 221, 44 },
		{ 416, 222, 44 },
		{ 416, 223, 44 },
		{ 416, 224, 44 },
		{ 416, 225, 44 },
		{ 416, 226, 44 },
		{ 416, 227, 44 },
		{ 416, 228, 44 },
		{ 416, 229, 44 },
		{ 416, 230, 44 },
		{ 416, 231, 44 },
		{ 416, 232, 44 },
		{ 416, 233, 44 },
		{ 416, 234, 44 },
		{ 416, 235, 44 },
		{ 416, 236, 44 },
		{ 416, 237, 44 },
		{ 417, 198, 44 },
		{ 417, 199, 44 },
		{ 417, 200, 44 },
		{ 417, 201, 44 },
		{ 417, 202, 44 },
		{ 417, 203, 44 },
		{ 417, 204, 44 },
		{ 417, 205, 44 },
		{ 417, 206, 44 },
		{ 417, 207, 44 },
		{ 417, 208, 44 },
		{ 417, 209, 44 },
		{ 417, 210, 44 },
		{ 417, 211, 44 },
		{ 417, 212, 44 },
		{ 417, 213, 44 },
		{ 417, 214, 44 },
		{ 417, 215, 44 },
		{ 417, 216, 44 },
		{ 417, 217, 44 },
		{ 417, 218, 44 },
		{ 417, 219, 44 },
		{ 417, 220, 44 },
		{ 417, 221, 44 },
		{ 417, 222, 44 },
		{ 417, 223, 44 },
		{ 417, 224, 44 },
		{ 417, 225, 44 },
		{ 417, 226, 44 },
		{ 417, 227, 44 },
		{ 417, 228, 44 },
		{ 417, 229, 44 },
		{ 417, 230, 44 },
		{ 417, 231, 44 },
		{ 417, 232, 44 },
		{ 417, 233, 44 },
		{ 417, 234, 44 },
		{ 417, 235, 44 },
		{ 417, 236, 44 },
		{ 418, 197, 44 },
		{ 418, 198, 44 },
		{ 418, 199, 44 },
		{ 418, 200, 44 },
		{ 418, 201, 44 },
		{ 418, 202, 44 },
		{ 418, 203, 44 },
		{ 418, 204, 44 },
		{ 418, 205, 44 },
		{ 418, 206, 44 },
		{ 418, 207, 44 },
		{ 418, 208, 44 },
		{ 418, 209, 44 },
		{ 418, 210, 44 },
		{ 418, 211, 44 },
		{ 418, 212, 44 },
		{ 418, 213, 44 },
		{ 418, 214, 44 },
		{ 418, 215, 44 },
		{ 418, 216, 44 },
		{ 418, 217, 44 },
		{ 418, 218, 44 },
		{ 418, 219, 44 },
		{ 418, 220, 44 },
		{ 418, 221, 44 },
		{ 418, 222, 44 },
		{ 418, 223, 44 },
		{ 418, 224, 44 },
		{ 418, 225, 44 },
		{ 418, 226, 44 },
		{ 418, 227, 44 },
		{ 418, 228, 44 },
		{ 418, 229, 44 },
		{ 418, 230, 44 },
		{ 418, 231, 44 },
		{ 418, 232, 44 },
		{ 418, 233, 44 },
		{ 418, 234, 44 },
		{ 418, 235, 44 },
		{ 418, 236, 44 },
		{ 419, 197, 44 },
		{ 419, 198, 44 },
		{ 419, 199, 44 },
		{ 419, 200, 44 },
		{ 419, 201, 44 },
		{ 419, 202, 44 },
		{ 419, 203, 44 },
		{ 419, 204, 44 },
		{ 419, 205, 44 },
		{ 419, 206, 44 },
		{ 419, 207, 44 },
		{ 419, 208, 44 },
		{ 419, 209, 44 },
		{ 419, 210, 44 },
		{ 419, 211, 44 },
		{ 419, 212, 44 },
		{ 419, 213, 44 },
		{ 419, 214, 44 },
		{ 419, 215, 44 },
		{ 419, 216, 44 },
		{ 419, 217, 44 },
		{ 419, 218, 44 },
		{ 419, 219, 44 },
		{ 419, 220, 44 },
		{ 419, 221, 44 },
		{ 419, 222, 44 },
		{ 419, 223, 44 },
		{ 419, 224, 44 },
		{ 419, 225, 44 },
		{ 419, 226, 44 },
		{ 419, 227, 44 },
		{ 419, 228, 44 },
		{ 419, 229, 44 },
		{ 419, 230, 44 },
		{ 419, 231, 44 },
		{ 419, 232, 44 },
		{ 419, 233, 44 },
		{ 419, 234, 44 },
		{ 419, 235, 44 },
		{ 419, 236, 44 },
		{ 420, 196, 44 },
		{ 420, 197, 44 },
		{ 420, 198, 44 },
		{ 420, 199, 44 },
		{ 420, 200, 44 },
		{ 420, 201, 44 },
		{ 420, 202, 44 },
		{ 420, 203, 44 },
		{ 420, 204, 44 },
		{ 420, 205, 44 },
		{ 420, 206, 44 },
		{ 420, 207, 44 },
		{ 420, 208, 44 },
		{ 420, 209, 44 },
		{ 420, 210, 44 },
		{ 420, 211, 44 },
		{ 420, 212, 44 },
		{ 420, 213, 44 },
		{ 420, 214, 44 },
		{ 420, 215, 44 },
		{ 420, 216, 44 },
		{ 420, 217, 44 },
		{ 420, 218, 44 },
		{ 420, 219, 44 },
		{ 420, 220, 44 },
		{ 420, 221, 44 },
		{ 420, 222, 44 },
		{ 420, 223, 44 },
		{ 420, 224, 44 },
		{ 420, 225, 44 },
		{ 420, 226, 44 },
		{ 420, 227, 44 },
		{ 420, 228, 44 },
		{ 420, 229, 44 },
		{ 420, 230, 44 },
		{ 420, 231, 44 },
		{ 420, 232, 44 },
		{ 420, 233, 44 },
		{ 420, 234, 44 },
		{ 420, 235, 44 },
		{ 420, 236, 44 },
		{ 421, 196, 44 },
		{ 421, 197, 44 },
		{ 421, 198, 44 },
		{ 421, 199, 44 },
		{ 421, 200, 44 },
		{ 421, 201, 44 },
		{ 421, 202, 44 },
		{ 421, 203, 44 },
		{ 421, 204, 44 },
		{ 421, 205, 44 },
		{ 421, 206, 44 },
		{ 421, 207, 44 },
		{ 421, 208, 44 },
		{ 421, 209, 44 },
		{ 421, 210, 44 },
		{ 421, 211, 44 },
		{ 421, 212, 44 },
		{ 421, 213, 44 },
		{ 421, 214, 44 },
		{ 421, 215, 44 },
		{ 421, 216, 44 },
		{ 421, 217, 44 },
		{ 421, 218, 44 },
		{ 421, 219, 44 },
		{ 421, 220, 44 },
		{ 421, 221, 44 },
		{ 421, 222, 44 },
		{ 421, 223, 44 },
		{ 421, 224, 44 },
		{ 421, 225, 44 },
		{ 421, 226, 44 },
		{ 421, 227, 44 },
		{ 421, 228, 44 },
		{ 421, 229, 44 },
		{ 421, 230, 44 },
		{ 421, 231, 44 },
		{ 421, 232, 44 },
		{ 421, 233, 44 },
		{ 421, 234, 44 },
		{ 421, 235, 44 },
		{ 421, 236, 44 },
		{ 422, 196, 44 },
		{ 422, 197, 44 },
		{ 422, 198, 44 },
		{ 422, 199, 44 },
		{ 422, 200, 44 },
		{ 422, 201, 44 },
		{ 422, 202, 44 },
		{ 422, 203, 44 },
		{ 422, 204, 44 },
		{ 422, 205, 44 },
		{ 422, 206, 44 },
		{ 422, 207, 44 },
		{ 422, 208, 44 },
		{ 422, 209, 44 },
		{ 422, 210, 44 },
		{ 422, 211, 44 },
		{ 422, 212, 44 },
		{ 422, 213, 44 },
		{ 422, 214, 44 },
		{ 422, 215, 44 },
		{ 422, 216, 44 },
		{ 422, 217, 44 },
		{ 422, 218, 44 },
		{ 422, 219, 44 },
		{ 422, 220, 44 },
		{ 422, 221, 44 },
		{ 422, 222, 44 },
		{ 422, 223, 44 },
		{ 422, 224, 44 },
		{ 422, 225, 44 },
		{ 422, 226, 44 },
		{ 422, 227, 44 },
		{ 422, 228, 44 },
		{ 422, 229, 44 },
		{ 422, 230, 44 },
		{ 422, 231, 44 },
		{ 422, 232, 44 },
		{ 422, 233, 44 },
		{ 422, 234, 44 },
		{ 422, 235, 44 },
		{ 422, 236, 44 },
		{ 423, 195, 44 },
		{ 423, 196, 44 },
		{ 423, 197, 44 },
		{ 423, 198, 44 },
		{ 423, 199, 44 },
		{ 423, 200, 44 },
		{ 423, 201, 44 },
		{ 423, 202, 44 },
		{ 423, 203, 44 },
		{ 423, 204, 44 },
		{ 423, 205, 44 },
		{ 423, 206, 44 },
		{ 423, 207, 44 },
		{ 423, 208, 44 },
		{ 423, 209, 44 },
		{ 423, 210, 44 },
		{ 423, 211, 44 },
		{ 423, 212, 44 },
		{ 423, 213, 44 },
		{ 423, 214, 44 },
		{ 423, 215, 44 },
		{ 423, 216, 44 },
		{ 423, 217, 44 },
		{ 423, 218, 44 },
		{ 423, 219, 44 },
		{ 423, 220, 44 },
		{ 423, 221, 44 },
		{ 423, 222, 44 },
		{ 423, 223, 44 },
		{ 423, 224, 44 },
		{ 423, 225, 44 },
		{ 423, 226, 44 },
		{ 423, 227, 44 },
		{ 423, 228, 44 },
		{ 423, 229, 44 },
		{ 423, 230, 44 },
		{ 423, 231, 44 },
		{ 423, 232, 44 },
		{ 423, 233, 44 },
		{ 423, 234, 44 },
		{ 423, 235, 44 },
		{ 423, 236, 44 },
		{ 424, 195, 44 },
		{ 424, 196, 44 },
		{ 424, 197, 44 },
		{ 424, 198, 44 },
		{ 424, 199, 44 },
		{ 424, 200, 44 },
		{ 424, 201, 44 },
		{ 424, 202, 44 },
		{ 424, 203, 44 },
		{ 424, 204, 44 },
		{ 424, 205, 44 },
		{ 424, 206, 44 },
		{ 424, 207, 44 },
		{ 424, 208, 44 },
		{ 424, 209, 44 },
		{ 424, 210, 44 },
		{ 424, 211, 44 },
		{ 424, 212, 44 },
		{ 424, 213, 44 },
		{ 424, 214, 44 },
		{ 424, 215, 44 },
		{ 424, 216, 44 },
		{ 424, 217, 44 },
		{ 424, 218, 44 },
		{ 424, 219, 44 },
		{ 424, 220, 44 },
		{ 424, 221, 44 },
		{ 424, 222, 44 },
		{ 424, 223, 44 },
		{ 424, 224, 44 },
		{ 424, 225, 44 },
		{ 424, 226, 44 },
		{ 424, 227, 44 },
		{ 424, 228, 44 },
		{ 424, 229, 44 },
		{ 424, 230, 44 },
		{ 424, 231, 44 },
		{ 424, 232, 44 },
		{ 424, 233, 44 },
		{ 424, 234, 44 },
		{ 424, 235, 44 },
		{ 424, 236, 44 },
		{ 425, 195, 44 },
		{ 425, 196, 44 },
		{ 425, 197, 44 },
		{ 425, 198, 44 },
		{ 425, 199, 44 },
		{ 425, 200, 44 },
		{ 425, 201, 44 },
		{ 425, 202, 44 },
		{ 425, 203, 44 },
		{ 425, 204, 44 },
		{ 425, 205, 44 },
		{ 425, 206, 44 },
		{ 425, 207, 44 },
		{ 425, 208, 44 },
		{ 425, 209, 44 },
		{ 425, 210, 44 },
		{ 425, 211, 44 },
		{ 425, 212, 44 },
		{ 425, 213, 44 },
		{ 425, 214, 44 },
		{ 425, 215, 44 },
		{ 425, 216, 44 },
		{ 425, 217, 44 },
		{ 425, 218, 44 },
		{ 425, 219, 44 },
		{ 425, 220, 44 },
		{ 425, 221, 44 },
		{ 425, 222, 44 },
		{ 425, 223, 44 },
		{ 425, 224, 44 },
		{ 425, 225, 44 },
		{ 425, 226, 44 },
		{ 425, 227, 44 },
		{ 425, 228, 44 },
		{ 425, 229, 44 },
		{ 425, 230, 44 },
		{ 425, 231, 44 },
		{ 425, 232, 44 },
		{ 425, 233, 44 },
		{ 425, 234, 44 },
		{ 425, 235, 44 },
		{ 426, 194, 44 },
		{ 426, 195, 44 },
		{ 426, 196, 44 },
		{ 426, 197, 44 },
		{ 426, 198, 44 },
		{ 426, 199, 44 },
		{ 426, 200, 44 },
		{ 426, 201, 44 },
		{ 426, 202, 44 },
		{ 426, 203, 44 },
		{ 426, 204, 44 },
		{ 426, 205, 44 },
		{ 426, 206, 44 },
		{ 426, 207, 44 },
		{ 426, 208, 44 },
		{ 426, 209, 44 },
		{ 426, 210, 44 },
		{ 426, 211, 44 },
		{ 426, 212, 44 },
		{ 426, 213, 44 },
		{ 426, 214, 44 },
		{ 426, 215, 44 },
		{ 426, 216, 44 },
		{ 426, 217, 44 },
		{ 426, 218, 44 },
		{ 426, 219, 44 },
		{ 426, 220, 44 },
		{ 426, 221, 44 },
		{ 426, 222, 44 },
		{ 426, 223, 44 },
		{ 426, 224, 44 },
		{ 426, 225, 44 },
		{ 426, 226, 44 },
		{ 426, 227, 44 },
		{ 426, 228, 44 },
		{ 426, 229, 44 },
		{ 426, 230, 44 },
		{ 426, 231, 44 },
		{ 426, 232, 44 },
		{ 426, 233, 44 },
		{ 426, 234, 44 },
		{ 426, 235, 44 },
		{ 427, 194, 44 },
		{ 427, 195, 44 },
		{ 427, 196, 44 },
		{ 427, 197, 44 },
		{ 427, 198, 44 },
		{ 427, 199, 44 },
		{ 427, 200, 44 },
		{ 427, 201, 44 },
		{ 427, 202, 44 },
		{ 427, 203, 44 },
		{ 427, 204, 44 },
		{ 427, 205, 44 },
		{ 427, 206, 44 },
		{ 427, 207, 44 },
		{ 427, 208, 44 },
		{ 427, 209, 44 },
		{ 427, 210, 44 },
		{ 427, 211, 44 },
		{ 427, 212, 44 },
		{ 427, 213, 44 },
		{ 427, 214, 44 },
		{ 427, 215, 44 },
		{ 427, 216, 44 },
		{ 427, 217, 44 },
		{ 427, 218, 44 },
		{ 427, 219, 44 },
		{ 427, 220, 44 },
		{ 427, 221, 44 },
		{ 427, 222, 44 },
		{ 427, 223, 44 },
		{ 427, 224, 44 },
		{ 427, 225, 44 },
		{ 427, 226, 44 },
		{ 427, 227, 44 },
		{ 427, 228, 44 },
		{ 427, 229, 44 },
		{ 427, 230, 44 },
		{ 427, 231, 44 },
		{ 427, 232, 44 },
		{ 427, 233, 44 },
		{ 427, 234, 44 },
		{ 427, 235, 44 },
		{ 428, 193, 44 },
		{ 428, 194, 44 },
		{ 428, 195, 44 },
		{ 428, 196, 44 },
		{ 428, 197, 44 },
		{ 428, 198, 44 },
		{ 428, 199, 44 },
		{ 428, 200, 44 },
		{ 428, 201, 44 },
		{ 428, 202, 44 },
		{ 428, 203, 44 },
		{ 428, 204, 44 },
		{ 428, 205, 44 },
		{ 428, 206, 44 },
		{ 428, 207, 44 },
		{ 428, 208, 44 },
		{ 428, 209, 44 },
		{ 428, 210, 44 },
		{ 428, 211, 44 },
		{ 428, 212, 44 },
		{ 428, 213, 44 },
		{ 428, 214, 44 },
		{ 428, 215, 44 },
		{ 428, 216, 44 },
		{ 428, 217, 44 },
		{ 428, 218, 44 },
		{ 428, 219, 44 },
		{ 428, 220, 44 },
		{ 428, 221, 44 },
		{ 428, 222, 44 },
		{ 428, 223, 44 },
		{ 428, 224, 44 },
		{ 428, 225, 44 },
		{ 428, 226, 44 },
		{ 428, 227, 44 },
		{ 428, 228, 44 },
		{ 428, 229, 44 },
		{ 428, 230, 44 },
		{ 428, 231, 44 },
		{ 428, 232, 44 },
		{ 428, 233, 44 },
		{ 428, 234, 44 },
		{ 428, 235, 44 },
		{ 429, 193, 44 },
		{ 429, 194, 44 },
		{ 429, 195, 44 },
		{ 429, 196, 44 },
		{ 429, 197, 44 },
		{ 429, 198, 44 },
		{ 429, 199, 44 },
		{ 429, 200, 44 },
		{ 429, 201, 44 },
		{ 429, 202, 44 },
		{ 429, 203, 44 },
		{ 429, 204, 44 },
		{ 429, 205, 44 },
		{ 429, 206, 44 },
		{ 429, 207, 44 },
		{ 429, 208, 44 },
		{ 429, 209, 44 },
		{ 429, 210, 44 },
		{ 429, 211, 44 },
		{ 429, 212, 44 },
		{ 429, 213, 44 },
		{ 429, 214, 44 },
		{ 429, 215, 44 },
		{ 429, 216, 44 },
		{ 429, 217, 44 },
		{ 429, 218, 44 },
		{ 429, 219, 44 },
		{ 429, 220, 44 },
		{ 429, 221, 44 },
		{ 429, 222, 44 },
		{ 429, 223, 44 },
		{ 429, 224, 44 },
		{ 429, 225, 44 },
		{ 429, 226, 44 },
		{ 429, 227, 44 },
		{ 429, 228, 44 },
		{ 429, 229, 44 },
		{ 429, 230, 44 },
		{ 429, 231, 44 },
		{ 429, 232, 44 },
		{ 429, 233, 44 },
		{ 429, 234, 44 },
		{ 429, 235, 44 },
		{ 430, 193, 44 },
		{ 430, 194, 44 },
		{ 430, 195, 44 },
		{ 430, 196, 44 },
		{ 430, 197, 44 },
		{ 430, 198, 44 },
		{ 430, 199, 44 },
		{ 430, 200, 44 },
		{ 430, 201, 44 },
		{ 430, 202, 44 },
		{ 430, 203, 44 },
		{ 430, 204, 44 },
		{ 430, 205, 44 },
		{ 430, 206, 44 },
		{ 430, 207, 44 },
		{ 430, 208, 44 },
		{ 430, 209, 44 },
		{ 430, 210, 44 },
		{ 430, 211, 44 },
		{ 430, 212, 44 },
		{ 430, 213, 44 },
		{ 430, 214, 44 },
		{ 430, 215, 44 },
		{ 430, 216, 44 },
		{ 430, 217, 44 },
		{ 430, 218, 44 },
		{ 430, 219, 44 },
		{ 430, 220, 44 },
		{ 430, 221, 44 },
		{ 430, 222, 44 },
		{ 430, 223, 44 },
		{ 430, 224, 44 },
		{ 430, 225, 44 },
		{ 430, 226, 44 },
		{ 430, 227, 44 },
		{ 430, 228, 44 },
		{ 430, 229, 44 },
		{ 430, 230, 44 },
		{ 430, 231, 44 },
		{ 430, 232, 44 },
		{ 430, 233, 44 },
		{ 430, 234, 44 },
		{ 430, 235, 44 },
		{ 431, 192, 44 },
		{ 431, 193, 44 },
		{ 431, 194, 44 },
		{ 431, 195, 44 },
		{ 431, 196, 44 },
		{ 431, 197, 44 },
		{ 431, 198, 44 },
		{ 431, 199, 44 },
		{ 431, 200, 44 },
		{ 431, 201, 44 },
		{ 431, 202, 44 },
		{ 431, 203, 44 },
		{ 431, 204, 44 },
		{ 431, 205, 44 },
		{ 431, 206, 44 },
		{ 431, 207, 44 },
		{ 431, 208, 44 },
		{ 431, 209, 44 },
		{ 431, 210, 44 },
		{ 431, 211, 44 },
		{ 431, 212, 44 },
		{ 431, 213, 44 },
		{ 431, 214, 44 },
		{ 431, 215, 44 },
		{ 431, 216, 44 },
		{ 431, 217, 44 },
		{ 431, 218, 44 },
		{ 431, 219, 44 },
		{ 431, 220, 44 },
		{ 431, 221, 44 },
		{ 431, 222, 44 },
		{ 431, 223, 44 },
		{ 431, 224, 44 },
		{ 431, 225, 44 },
		{ 431, 226, 44 },
		{ 431, 227, 44 },
		{ 431, 228, 44 },
		{ 431, 229, 44 },
		{ 431, 230, 44 },
		{ 431, 231, 44 },
		{ 431, 232, 44 },
		{ 431, 233, 44 },
		{ 431, 234, 44 },
		{ 431, 235, 44 },
		{ 432, 192, 44 },
		{ 432, 193, 44 },
		{ 432, 194, 44 },
		{ 432, 195, 44 },
		{ 432, 196, 44 },
		{ 432, 197, 44 },
		{ 432, 198, 44 },
		{ 432, 199, 44 },
		{ 432, 200, 44 },
		{ 432, 201, 44 },
		{ 432, 202, 44 },
		{ 432, 203, 44 },
		{ 432, 204, 44 },
		{ 432, 205, 44 },
		{ 432, 206, 44 },
		{ 432, 207, 44 },
		{ 432, 208, 44 },
		{ 432, 209, 44 },
		{ 432, 210, 44 },
		{ 432, 211, 44 },
		{ 432, 212, 44 },
		{ 432, 213, 44 },
		{ 432, 214, 44 },
		{ 432, 215, 44 },
		{ 432, 216, 44 },
		{ 432, 217, 44 },
		{ 432, 218, 44 },
		{ 432, 219, 44 },
		{ 432, 220, 44 },
		{ 432, 221, 44 },
		{ 432, 222, 44 },
		{ 432, 223, 44 },
		{ 432, 224, 44 },
		{ 432, 225, 44 },
		{ 432, 226, 44 },
		{ 432, 227, 44 },
		{ 432, 228, 44 },
		{ 432, 229, 44 },
		{ 432, 230, 44 },
		{ 432, 231, 44 },
		{ 432, 232, 44 },
		{ 432, 233, 44 },
		{ 432, 234, 44 },
		{ 432, 235, 44 },
		{ 433, 192, 44 },
		{ 433, 193, 44 },
		{ 433, 194, 44 },
		{ 433, 195, 44 },
		{ 433, 196, 44 },
		{ 433, 197, 44 },
		{ 433, 198, 44 },
		{ 433, 199, 44 },
		{ 433, 200, 44 },
		{ 433, 201, 44 },
		{ 433, 202, 44 },
		{ 433, 203, 44 },
		{ 433, 204, 44 },
		{ 433, 205, 44 },
		{ 433, 206, 44 },
		{ 433, 207, 44 },
		{ 433, 208, 44 },
		{ 433, 209, 44 },
		{ 433, 210, 44 },
		{ 433, 211, 44 },
		{ 433, 212, 44 },
		{ 433, 213, 44 },
		{ 433, 214, 44 },
		{ 433, 215, 44 },
		{ 433, 216, 44 },
		{ 433, 217, 44 },
		{ 433, 218, 44 },
		{ 433, 219, 44 },
		{ 433, 220, 44 },
		{ 433, 221, 44 },
		{ 433, 222, 44 },
		{ 433, 223, 44 },
		{ 433, 224, 44 },
		{ 433, 225, 44 },
		{ 433, 226, 44 },
		{ 433, 227, 44 },
		{ 433, 228, 44 },
		{ 433, 229, 44 },
		{ 433, 230, 44 },
		{ 433, 231, 44 },
		{ 433, 232, 44 },
		{ 433, 233, 44 },
		{ 433, 234, 44 },
		{ 434, 191, 44 },
		{ 434, 192, 44 },
		{ 434, 193, 44 },
		{ 434, 194, 44 },
		{ 434, 195, 44 },
		{ 434, 196, 44 },
		{ 434, 197, 44 },
		{ 434, 198, 44 },
		{ 434, 199, 44 },
		{ 434, 200, 44 },
		{ 434, 201, 44 },
		{ 434, 202, 44 },
		{ 434, 203, 44 },
		{ 434, 204, 44 },
		{ 434, 205, 44 },
		{ 434, 206, 44 },
		{ 434, 207, 44 },
		{ 434, 208, 44 },
		{ 434, 209, 44 },
		{ 434, 210, 44 },
		{ 434, 211, 44 },
		{ 434, 212, 44 },
		{ 434, 213, 44 },
		{ 434, 214, 44 },
		{ 434, 215, 44 },
		{ 434, 216, 44 },
		{ 434, 217, 44 },
		{ 434, 218, 44 },
		{ 434, 219, 44 },
		{ 434, 220, 44 },
		{ 434, 221, 44 },
		{ 434, 222, 44 },
		{ 434, 223, 44 },
		{ 434, 224, 44 },
		{ 434, 225, 44 },
		{ 434, 226, 44 },
		{ 434, 227, 44 },
		{ 434, 228, 44 },
		{ 434, 229, 44 },
		{ 434, 230, 44 },
		{ 434, 231, 44 },
		{ 434, 232, 44 },
		{ 434, 233, 44 },
		{ 434, 234, 44 },
		{ 435, 191, 44 },
		{ 435, 192, 44 },
		{ 435, 193, 44 },
		{ 435, 194, 44 },
		{ 435, 195, 44 },
		{ 435, 196, 44 },
		{ 435, 197, 44 },
		{ 435, 198, 44 },
		{ 435, 199, 44 },
		{ 435, 200, 44 },
		{ 435, 201, 44 },
		{ 435, 202, 44 },
		{ 435, 203, 44 },
		{ 435, 204, 44 },
		{ 435, 205, 44 },
		{ 435, 206, 44 },
		{ 435, 207, 44 },
		{ 435, 208, 44 },
		{ 435, 209, 44 },
		{ 435, 210, 44 },
		{ 435, 211, 44 },
		{ 435, 212, 44 },
		{ 435, 213, 44 },
		{ 435, 214, 44 },
		{ 435, 215, 44 },
		{ 435, 216, 44 },
		{ 435, 217, 44 },
		{ 435, 218, 44 },
		{ 435, 219, 44 },
		{ 435, 220, 44 },
		{ 435, 221, 44 },
		{ 435, 222, 44 },
		{ 435, 223, 44 },
		{ 435, 224, 44 },
		{ 435, 225, 44 },
		{ 435, 226, 44 },
		{ 435, 227, 44 },
		{ 435, 228, 44 },
		{ 435, 229, 44 },
		{ 435, 230, 44 },
		{ 435, 231, 44 },
		{ 435, 232, 44 },
		{ 435, 233, 44 },
		{ 435, 234, 44 },
		{ 436, 190, 44 },
		{ 436, 191, 44 },
		{ 436, 192, 44 },
		{ 436, 193, 44 },
		{ 436, 194, 44 },
		{ 436, 195, 44 },
		{ 436, 196, 44 },
		{ 436, 197, 44 },
		{ 436, 198, 44 },
		{ 436, 199, 44 },
		{ 436, 200, 44 },
		{ 436, 201, 44 },
		{ 436, 202, 44 },
		{ 436, 203, 44 },
		{ 436, 204, 44 },
		{ 436, 205, 44 },
		{ 436, 206, 44 },
		{ 436, 207, 44 },
		{ 436, 208, 44 },
		{ 436, 209, 44 },
		{ 436, 210, 44 },
		{ 436, 211, 44 },
		{ 436, 212, 44 },
		{ 436, 213, 44 },
		{ 436, 214, 44 },
		{ 436, 215, 44 },
		{ 436, 216, 44 },
		{ 436, 217, 44 },
		{ 436, 218, 44 },
		{ 436, 219, 44 },
		{ 436, 220, 44 },
		{ 436, 221, 44 },
		{ 436, 222, 44 },
		{ 436, 223, 44 },
		{ 436, 224, 44 },
		{ 436, 225, 44 },
		{ 436, 226, 44 },
		{ 436, 227, 44 },
		{ 436, 228, 44 },
		{ 436, 229, 44 },
		{ 436, 230, 44 },
		{ 436, 231, 44 },
		{ 436, 232, 44 },
		{ 436, 233, 44 },
		{ 436, 234, 44 },
		{ 437, 190, 44 },
		{ 437, 191, 44 },
		{ 437, 192, 44 },
		{ 437, 193, 44 },
		{ 437, 194, 44 },
		{ 437, 195, 44 },
		{ 437, 196, 44 },
		{ 437, 197, 44 },
		{ 437, 198, 44 },
		{ 437, 199, 44 },
		{ 437, 200, 44 },
		{ 437, 201, 44 },
		{ 437, 202, 44 },
		{ 437, 203, 44 },
		{ 437, 204, 44 },
		{ 437, 205, 44 },
		{ 437, 206, 44 },
		{ 437, 207, 44 },
		{ 437, 208, 44 },
		{ 437, 209, 44 },
		{ 437, 210, 44 },
		{ 437, 211, 44 },
		{ 437, 212, 44 },
		{ 437, 213, 44 },
		{ 437, 214, 44 },
		{ 437, 215, 44 },
		{ 437, 216, 44 },
		{ 437, 217, 44 },
		{ 437, 218, 44 },
		{ 437, 219, 44 },
		{ 437, 220, 44 },
		{ 437, 221, 44 },
		{ 437, 222, 44 },
		{ 437, 223, 44 },
		{ 437, 224, 44 },
		{ 437, 225, 44 },
		{ 437, 226, 44 },
		{ 437, 227, 44 },
		{ 437, 228, 44 },
		{ 437, 229, 44 },
		{ 437, 230, 44 },
		{ 437, 231, 44 },
		{ 437, 232, 44 },
		{ 437, 233, 44 },
		{ 437, 234, 44 },
		{ 438, 190, 44 },
		{ 438, 191, 44 },
		{ 438, 192, 44 },
		{ 438, 193, 44 },
		{ 438, 194, 44 },
		{ 438, 195, 44 },
		{ 438, 196, 44 },
		{ 438, 197, 44 },
		{ 438, 198, 44 },
		{ 438, 199, 44 },
		{ 438, 200, 44 },
		{ 438, 201, 44 },
		{ 438, 202, 44 },
		{ 438, 203, 44 },
		{ 438, 204, 44 },
		{ 438, 205, 44 },
		{ 438, 206, 44 },
		{ 438, 207, 44 },
		{ 438, 208, 44 },
		{ 438, 209, 44 },
		{ 438, 210, 44 },
		{ 438, 211, 44 },
		{ 438, 212, 44 },
		{ 438, 213, 44 },
		{ 438, 214, 44 },
		{ 438, 215, 44 },
		{ 438, 216, 44 },
		{ 438, 217, 44 },
		{ 438, 218, 44 },
		{ 438, 219, 44 },
		{ 438, 220, 44 },
		{ 438, 221, 44 },
		{ 438, 222, 44 },
		{ 438, 223, 44 },
		{ 438, 224, 44 },
		{ 438, 225, 44 },
		{ 438, 226, 44 },
		{ 438, 227, 44 },
		{ 438, 228, 44 },
		{ 438, 229, 44 },
		{ 438, 230, 44 },
		{ 438, 231, 44 },
		{ 438, 232, 44 },
		{ 438, 233, 44 },
		{ 438, 234, 44 },
		{ 439, 189, 44 },
		{ 439, 190, 44 },
		{ 439, 191, 44 },
		{ 439, 192, 44 },
		{ 439, 193, 44 },
		{ 439, 194, 44 },
		{ 439, 195, 44 },
		{ 439, 196, 44 },
		{ 439, 197, 44 },
		{ 439, 198, 44 },
		{ 439, 199, 44 },
		{ 439, 200, 44 },
		{ 439, 201, 44 },
		{ 439, 202, 44 },
		{ 439, 203, 44 },
		{ 439, 204, 44 },
		{ 439, 205, 44 },
		{ 439, 206, 44 },
		{ 439, 207, 44 },
		{ 439, 208, 44 },
		{ 439, 209, 44 },
		{ 439, 210, 44 },
		{ 439, 211, 44 },
		{ 439, 212, 44 },
		{ 439, 213, 44 },
		{ 439, 214, 44 },
		{ 439, 215, 44 },
		{ 439, 216, 44 },
		{ 439, 217, 44 },
		{ 439, 218, 44 },
		{ 439, 219, 44 },
		{ 439, 220, 44 },
		{ 439, 221, 44 },
		{ 439, 222, 44 },
		{ 439, 223, 44 },
		{ 439, 224, 44 },
		{ 439, 225, 44 },
		{ 439, 226, 44 },
		{ 439, 227, 44 },
		{ 439, 228, 44 },
		{ 439, 229, 44 },
		{ 439, 230, 44 },
		{ 439, 231, 44 },
		{ 439, 232, 44 },
		{ 439, 233, 44 },
		{ 439, 234, 44 },
		{ 440, 189, 44 },
		{ 440, 190, 44 },
		{ 440, 191, 44 },
		{ 440, 192, 44 },
		{ 440, 193, 44 },
		{ 440, 194, 44 },
		{ 440, 195, 44 },
		{ 440, 196, 44 },
		{ 440, 197, 44 },
		{ 440, 198, 44 },
		{ 440, 199, 44 },
		{ 440, 200, 44 },
		{ 440, 201, 44 },
		{ 440, 202, 44 },
		{ 440, 203, 44 },
		{ 440, 204, 44 },
		{ 440, 205, 44 },
		{ 440, 206, 44 },
		{ 440, 207, 44 },
		{ 440, 208, 44 },
		{ 440, 209, 44 },
		{ 440, 210, 44 },
		{ 440, 211, 44 },
		{ 440, 212, 44 },
		{ 440, 213, 44 },
		{ 440, 214, 44 },
		{ 440, 215, 44 },
		{ 440, 216, 44 },
		{ 440, 217, 44 },
		{ 440, 218, 44 },
		{ 440, 219, 44 },
		{ 440, 220, 44 },
		{ 440, 221, 44 },
		{ 440, 222, 44 },
		{ 440, 223, 44 },
		{ 440, 224, 44 },
		{ 440, 225, 44 },
		{ 440, 226, 44 },
		{ 440, 227, 44 },
		{ 440, 228, 44 },
		{ 440, 229, 44 },
		{ 440, 230, 44 },
		{ 440, 231, 44 },
		{ 440, 232, 44 },
		{ 440, 233, 44 },
		{ 440, 234, 44 },
		{ 441, 189, 44 },
		{ 441, 190, 44 },
		{ 441, 191, 44 },
		{ 441, 192, 44 },
		{ 441, 193, 44 },
		{ 441, 194, 44 },
		{ 441, 195, 44 },
		{ 441, 196, 44 },
		{ 441, 197, 44 },
		{ 441, 198, 44 },
		{ 441, 199, 44 },
		{ 441, 200, 44 },
		{ 441, 201, 44 },
		{ 441, 202, 44 },
		{ 441, 203, 44 },
		{ 441, 204, 44 },
		{ 441, 205, 44 },
		{ 441, 206, 44 },
		{ 441, 207, 44 },
		{ 441, 208, 44 },
		{ 441, 209, 44 },
		{ 441, 210, 44 },
		{ 441, 211, 44 },
		{ 441, 212, 44 },
		{ 441, 213, 44 },
		{ 441, 214, 44 },
		{ 441, 215, 44 },
		{ 441, 216, 44 },
		{ 441, 217, 44 },
		{ 441, 218, 44 },
		{ 441, 219, 44 },
		{ 441, 220, 44 },
		{ 441, 221, 44 },
		{ 441, 222, 44 },
		{ 441, 223, 44 },
		{ 441, 224, 44 },
		{ 441, 225, 44 },
		{ 441, 226, 44 },
		{ 441, 227, 44 },
		{ 441, 228, 44 },
		{ 441, 229, 44 },
		{ 441, 230, 44 },
		{ 441, 231, 44 },
		{ 441, 232, 44 },
		{ 441, 233, 44 },
		{ 442, 188, 44 },
		{ 442, 189, 44 },
		{ 442, 190, 44 },
		{ 442, 191, 44 },
		{ 442, 192, 44 },
		{ 442, 193, 44 },
		{ 442, 194, 44 },
		{ 442, 195, 44 },
		{ 442, 196, 44 },
		{ 442, 197, 44 },
		{ 442, 198, 44 },
		{ 442, 199, 44 },
		{ 442, 200, 44 },
		{ 442, 201, 44 },
		{ 442, 202, 44 },
		{ 442, 203, 44 },
		{ 442, 204, 44 },
		{ 442, 205, 44 },
		{ 442, 206, 44 },
		{ 442, 207, 44 },
		{ 442, 208, 44 },
		{ 442, 209, 44 },
		{ 442, 210, 44 },
		{ 442, 211, 44 },
		{ 442, 212, 44 },
		{ 442, 213, 44 },
		{ 442, 214, 44 },
		{ 442, 215, 44 },
		{ 442, 216, 44 },
		{ 442, 217, 44 },
		{ 442, 218, 44 },
		{ 442, 219, 44 },
		{ 442, 220, 44 },
		{ 442, 221, 44 },
		{ 442, 222, 44 },
		{ 442, 223, 44 },
		{ 442, 224, 44 },
		{ 442, 225, 44 },
		{ 442, 226, 44 },
		{ 442, 227, 44 },
		{ 442, 228, 44 },
		{ 442, 229, 44 },
		{ 442, 230, 44 },
		{ 442, 231, 44 },
		{ 442, 232, 44 },
		{ 442, 233, 44 },
		{ 443, 188, 44 },
		{ 443, 189, 44 },
		{ 443, 190, 44 },
		{ 443, 191, 44 },
		{ 443, 192, 44 },
		{ 443, 193, 44 },
		{ 443, 194, 44 },
		{ 443, 195, 44 },
		{ 443, 196, 44 },
		{ 443, 197, 44 },
		{ 443, 198, 44 },
		{ 443, 199, 44 },
		{ 443, 200, 44 },
		{ 443, 201, 44 },
		{ 443, 202, 44 },
		{ 443, 203, 44 },
		{ 443, 204, 44 },
		{ 443, 205, 44 },
		{ 443, 206, 44 },
		{ 443, 207, 44 },
		{ 443, 208, 44 },
		{ 443, 209, 44 },
		{ 443, 210, 44 },
		{ 443, 211, 44 },
		{ 443, 212, 44 },
		{ 443, 213, 44 },
		{ 443, 214, 44 },
		{ 443, 215, 44 },
		{ 443, 216, 44 },
		{ 443, 217, 44 },
		{ 443, 218, 44 },
		{ 443, 219, 44 },
		{ 443, 220, 44 },
		{ 443, 221, 44 },
		{ 443, 222, 44 },
		{ 443, 223, 44 },
		{ 443, 224, 44 },
		{ 443, 225, 44 },
		{ 443, 226, 44 },
		{ 443, 227, 44 },
		{ 443, 228, 44 },
		{ 443, 229, 44 },
		{ 443, 230, 44 },
		{ 443, 231, 44 },
		{ 443, 232, 44 },
		{ 443, 233, 44 },
		{ 444, 187, 44 },
		{ 444, 188, 44 },
		{ 444, 189, 44 },
		{ 444, 190, 44 },
		{ 444, 191, 44 },
		{ 444, 192, 44 },
		{ 444, 193, 44 },
		{ 444, 194, 44 },
		{ 444, 195, 44 },
		{ 444, 196, 44 },
		{ 444, 197, 44 },
		{ 444, 198, 44 },
		{ 444, 199, 44 },
		{ 444, 200, 44 },
		{ 444, 201, 44 },
		{ 444, 202, 44 },
		{ 444, 203, 44 },
		{ 444, 204, 44 },
		{ 444, 205, 44 },
		{ 444, 206, 44 },
		{ 444, 207, 44 },
		{ 444, 208, 44 },
		{ 444, 209, 44 },
		{ 444, 210, 44 },
		{ 444, 211, 44 },
		{ 444, 212, 44 },
		{ 444, 213, 44 },
		{ 444, 214, 44 },
		{ 444, 215, 44 },
		{ 444, 216, 44 },
		{ 444, 217, 44 },
		{ 444, 218, 44 },
		{ 444, 219, 44 },
		{ 444, 220, 44 },
		{ 444, 221, 44 },
		{ 444, 222, 44 },
		{ 444, 223, 44 },
		{ 444, 224, 44 },
		{ 444, 225, 44 },
		{ 444, 226, 44 },
		{ 444, 227, 44 },
		{ 444, 228, 44 },
		{ 444, 229, 44 },
		{ 444, 230, 44 },
		{ 444, 231, 44 },
		{ 444, 232, 44 },
		{ 444, 233, 44 },
		{ 445, 187, 44 },
		{ 445, 188, 44 },
		{ 445, 189, 44 },
		{ 445, 190, 44 },
		{ 445, 191, 44 },
		{ 445, 192, 44 },
		{ 445, 193, 44 },
		{ 445, 194, 44 },
		{ 445, 195, 44 },
		{ 445, 196, 44 },
		{ 445, 197, 44 },
		{ 445, 198, 44 },
		{ 445, 199, 44 },
		{ 445, 200, 44 },
		{ 445, 201, 44 },
		{ 445, 202, 44 },
		{ 445, 203, 44 },
		{ 445, 204, 44 },
		{ 445, 205, 44 },
		{ 445, 206, 44 },
		{ 445, 207, 44 },
		{ 445, 208, 44 },
		{ 445, 209, 44 },
		{ 445, 210, 44 },
		{ 445, 211, 44 },
		{ 445, 212, 44 },
		{ 445, 213, 44 },
		{ 445, 214, 44 },
		{ 445, 215, 44 },
		{ 445, 216, 44 },
		{ 445, 217, 44 },
		{ 445, 218, 44 },
		{ 445, 219, 44 },
		{ 445, 220, 44 },
		{ 445, 221, 44 },
		{ 445, 222, 44 },
		{ 445, 223, 44 },
		{ 445, 224, 44 },
		{ 445, 225, 44 },
		{ 445, 226, 44 },
		{ 445, 227, 44 },
		{ 445, 228, 44 },
		{ 445, 229, 44 },
		{ 445, 230, 44 },
		{ 445, 231, 44 },
		{ 445, 232, 44 },
		{ 445, 233, 44 },
		{ 446, 187, 44 },
		{ 446, 188, 44 },
		{ 446, 189, 44 },
		{ 446, 190, 44 },
		{ 446, 191, 44 },
		{ 446, 192, 44 },
		{ 446, 193, 44 },
		{ 446, 194, 44 },
		{ 446, 195, 44 },
		{ 446, 196, 44 },
		{ 446, 197, 44 },
		{ 446, 198, 44 },
		{ 446, 199, 44 },
		{ 446, 200, 44 },
		{ 446, 201, 44 },
		{ 446, 202, 44 },
		{ 446, 203, 44 },
		{ 446, 204, 44 },
		{ 446, 205, 44 },
		{ 446, 206, 44 },
		{ 446, 207, 44 },
		{ 446, 208, 44 },
		{ 446, 209, 44 },
		{ 446, 210, 44 },
		{ 446, 211, 44 },
		{ 446, 212, 44 },
		{ 446, 213, 44 },
		{ 446, 214, 44 },
		{ 446, 215, 44 },
		{ 446, 216, 44 },
		{ 446, 217, 44 },
		{ 446, 218, 44 },
		{ 446, 219, 44 },
		{ 446, 220, 44 },
		{ 446, 221, 44 },
		{ 446, 222, 44 },
		{ 446, 223, 44 },
		{ 446, 224, 44 },
		{ 446, 225, 44 },
		{ 446, 226, 44 },
		{ 446, 227, 44 },
		{ 446, 228, 44 },
		{ 446, 229, 44 },
		{ 446, 230, 44 },
		{ 446, 231, 44 },
		{ 446, 232, 44 },
		{ 446, 233, 44 },
		{ 447, 186, 44 },
		{ 447, 187, 44 },
		{ 447, 188, 44 },
		{ 447, 189, 44 },
		{ 447, 190, 44 },
		{ 447, 191, 44 },
		{ 447, 192, 44 },
		{ 447, 193, 44 },
		{ 447, 194, 44 },
		{ 447, 195, 44 },
		{ 447, 196, 44 },
		{ 447, 197, 44 },
		{ 447, 198, 44 },
		{ 447, 199, 44 },
		{ 447, 200, 44 },
		{ 447, 201, 44 },
		{ 447, 202, 44 },
		{ 447, 203, 44 },
		{ 447, 204, 44 },
		{ 447, 205, 44 },
		{ 447, 206, 44 },
		{ 447, 207, 44 },
		{ 447, 208, 44 },
		{ 447, 209, 44 },
		{ 447, 210, 44 },
		{ 447, 211, 44 },
		{ 447, 212, 44 },
		{ 447, 213, 44 },
		{ 447, 214, 44 },
		{ 447, 215, 44 },
		{ 447, 216, 44 },
		{ 447, 217, 44 },
		{ 447, 218, 44 },
		{ 447, 219, 44 },
		{ 447, 220, 44 },
		{ 447, 221, 44 },
		{ 447, 222, 44 },
		{ 447, 223, 44 },
		{ 447, 224, 44 },
		{ 447, 225, 44 },
		{ 447, 226, 44 },
		{ 447, 227, 44 },
		{ 447, 228, 44 },
		{ 447, 229, 44 },
		{ 447, 230, 44 },
		{ 447, 231, 44 },
		{ 447, 232, 44 },
		{ 447, 233, 44 },
		{ 448, 186, 44 },
		{ 448, 187, 44 },
		{ 448, 188, 44 },
		{ 448, 189, 44 },
		{ 448, 190, 44 },
		{ 448, 191, 44 },
		{ 448, 192, 44 },
		{ 448, 193, 44 },
		{ 448, 194, 44 },
		{ 448, 195, 44 },
		{ 448, 196, 44 },
		{ 448, 197, 44 },
		{ 448, 198, 44 },
		{ 448, 199, 44 },
		{ 448, 200, 44 },
		{ 448, 201, 44 },
		{ 448, 202, 44 },
		{ 448, 203, 44 },
		{ 448, 204, 44 },
		{ 448, 205, 44 },
		{ 448, 206, 44 },
		{ 448, 207, 44 },
		{ 448, 208, 44 },
		{ 448, 209, 44 },
		{ 448, 210, 44 },
		{ 448, 211, 44 },
		{ 448, 212, 44 },
		{ 448, 213, 44 },
		{ 448, 214, 44 },
		{ 448, 215, 44 },
		{ 448, 216, 44 },
		{ 448, 217, 44 },
		{ 448, 218, 44 },
		{ 448, 219, 44 },
		{ 448, 220, 44 },
		{ 448, 221, 44 },
		{ 448, 222, 44 },
		{ 448, 223, 44 },
		{ 448, 224, 44 },
		{ 448, 225, 44 },
		{ 448, 226, 44 },
		{ 448, 227, 44 },
		{ 448, 228, 44 },
		{ 448, 229, 44 },
		{ 448, 230, 44 },
		{ 448, 231, 44 },
		{ 448, 232, 44 },
		{ 448, 233, 44 },
		{ 449, 186, 44 },
		{ 449, 187, 44 },
		{ 449, 188, 44 },
		{ 449, 189, 44 },
		{ 449, 190, 44 },
		{ 449, 191, 44 },
		{ 449, 192, 44 },
		{ 449, 193, 44 },
		{ 449, 194, 44 },
		{ 449, 195, 44 },
		{ 449, 196, 44 },
		{ 449, 197, 44 },
		{ 449, 198, 44 },
		{ 449, 199, 44 },
		{ 449, 200, 44 },
		{ 449, 201, 44 },
		{ 449, 202, 44 },
		{ 449, 203, 44 },
		{ 449, 204, 44 },
		{ 449, 205, 44 },
		{ 449, 206, 44 },
		{ 449, 207, 44 },
		{ 449, 208, 44 },
		{ 449, 209, 44 },
		{ 449, 210, 44 },
		{ 449, 211, 44 },
		{ 449, 212, 44 },
		{ 449, 213, 44 },
		{ 449, 214, 44 },
		{ 449, 215, 44 },
		{ 449, 216, 44 },
		{ 449, 217, 44 },
		{ 449, 218, 44 },
		{ 449, 219, 44 },
		{ 449, 220, 44 },
		{ 449, 221, 44 },
		{ 449, 222, 44 },
		{ 449, 223, 44 },
		{ 449, 224, 44 },
		{ 449, 225, 44 },
		{ 449, 226, 44 },
		{ 449, 227, 44 },
		{ 449, 228, 44 },
		{ 449, 229, 44 },
		{ 449, 230, 44 },
		{ 449, 231, 44 },
		{ 449, 232, 44 },
		{ 450, 185, 44 },
		{ 450, 186, 44 },
		{ 450, 187, 44 },
		{ 450, 188, 44 },
		{ 450, 189, 44 },
		{ 450, 190, 44 },
		{ 450, 191, 44 },
		{ 450, 192, 44 },
		{ 450, 193, 44 },
		{ 450, 194, 44 },
		{ 450, 195, 44 },
		{ 450, 196, 44 },
		{ 450, 197, 44 },
		{ 450, 198, 44 },
		{ 450, 199, 44 },
		{ 450, 200, 44 },
		{ 450, 201, 44 },
		{ 450, 202, 44 },
		{ 450, 203, 44 },
		{ 450, 204, 44 },
		{ 450, 205, 44 },
		{ 450, 206, 44 },
		{ 450, 207, 44 },
		{ 450, 208, 44 },
		{ 450, 209, 44 },
		{ 450, 210, 44 },
		{ 450, 211, 44 },
		{ 450, 212, 44 },
		{ 450, 213, 44 },
		{ 450, 214, 44 },
		{ 450, 215, 44 },
		{ 450, 216, 44 },
		{ 450, 217, 44 },
		{ 450, 218, 44 },
		{ 450, 219, 44 },
		{ 450, 220, 44 },
		{ 450, 221, 44 },
		{ 450, 222, 44 },
		{ 450, 223, 44 },
		{ 450, 224, 44 },
		{ 450, 225, 44 },
		{ 450, 226, 44 },
		{ 450, 227, 44 },
		{ 450, 228, 44 },
		{ 450, 229, 44 },
		{ 450, 230, 44 },
		{ 450, 231, 44 },
		{ 450, 232, 44 },
		{ 451, 185, 44 },
		{ 451, 186, 44 },
		{ 451, 187, 44 },
		{ 451, 188, 44 },
		{ 451, 189, 44 },
		{ 451, 190, 44 },
		{ 451, 191, 44 },
		{ 451, 192, 44 },
		{ 451, 193, 44 },
		{ 451, 194, 44 },
		{ 451, 195, 44 },
		{ 451, 196, 44 },
		{ 451, 197, 44 },
		{ 451, 198, 44 },
		{ 451, 199, 44 },
		{ 451, 200, 44 },
		{ 451, 201, 44 },
		{ 451, 202, 44 },
		{ 451, 203, 44 },
		{ 451, 204, 44 },
		{ 451, 205, 44 },
		{ 451, 206, 44 },
		{ 451, 207, 44 },
		{ 451, 208, 44 },
		{ 451, 209, 44 },
		{ 451, 210, 44 },
		{ 451, 211, 44 },
		{ 451, 212, 44 },
		{ 451, 213, 44 },
		{ 451, 214, 44 },
		{ 451, 215, 44 },
		{ 451, 216, 44 },
		{ 451, 217, 44 },
		{ 451, 218, 44 },
		{ 451, 219, 44 },
		{ 451, 220, 44 },
		{ 451, 221, 44 },
		{ 451, 222, 44 },
		{ 451, 223, 44 },
		{ 451, 224, 44 },
		{ 451, 225, 44 },
		{ 451, 226, 44 },
		{ 451, 227, 44 },
		{ 451, 228, 44 },
		{ 451, 229, 44 },
		{ 451, 230, 44 },
		{ 451, 231, 44 },
		{ 451, 232, 44 },
		{ 452, 184, 44 },
		{ 452, 185, 44 },
		{ 452, 186, 44 },
		{ 452, 187, 44 },
		{ 452, 188, 44 },
		{ 452, 189, 44 },
		{ 452, 190, 44 },
		{ 452, 191, 44 },
		{ 452, 192, 44 },
		{ 452, 193, 44 },
		{ 452, 194, 44 },
		{ 452, 195, 44 },
		{ 452, 196, 44 },
		{ 452, 197, 44 },
		{ 452, 198, 44 },
		{ 452, 199, 44 },
		{ 452, 200, 44 },
		{ 452, 201, 44 },
		{ 452, 202, 44 },
		{ 452, 203, 44 },
		{ 452, 204, 44 },
		{ 452, 205, 44 },
		{ 452, 206, 44 },
		{ 452, 207, 44 },
		{ 452, 208, 44 },
		{ 452, 209, 44 },
		{ 452, 210, 44 },
		{ 452, 211, 44 },
		{ 452, 212, 44 },
		{ 452, 213, 44 },
		{ 452, 214, 44 },
		{ 452, 215, 44 },
		{ 452, 216, 44 },
		{ 452, 217, 44 },
		{ 452, 218, 44 },
		{ 452, 219, 44 },
		{ 452, 220, 44 },
		{ 452, 221, 44 },
		{ 452, 222, 44 },
		{ 452, 223, 44 },
		{ 452, 224, 44 },
		{ 452, 225, 44 },
		{ 452, 226, 44 },
		{ 452, 227, 44 },
		{ 452, 228, 44 },
		{ 452, 229, 44 },
		{ 452, 230, 44 },
		{ 452, 231, 44 },
		{ 452, 232, 44 },
		{ 453, 184, 44 },
		{ 453, 185, 44 },
		{ 453, 186, 44 },
		{ 453, 187, 44 },
		{ 453, 188, 44 },
		{ 453, 189, 44 },
		{ 453, 190, 44 },
		{ 453, 191, 44 },
		{ 453, 192, 44 },
		{ 453, 193, 44 },
		{ 453, 194, 44 },
		{ 453, 195, 44 },
		{ 453, 196, 44 },
		{ 453, 197, 44 },
		{ 453, 198, 44 },
		{ 453, 199, 44 },
		{ 453, 200, 44 },
		{ 453, 201, 44 },
		{ 453, 202, 44 },
		{ 453, 203, 44 },
		{ 453, 204, 44 },
		{ 453, 205, 44 },
		{ 453, 206, 44 },
		{ 453, 207, 44 },
		{ 453, 208, 44 },
		{ 453, 209, 44 },
		{ 453, 210, 44 },
		{ 453, 211, 44 },
		{ 453, 212, 44 },
		{ 453, 213, 44 },
		{ 453, 214, 44 },
		{ 453, 215, 44 },
		{ 453, 216, 44 },
		{ 453, 217, 44 },
		{ 453, 218, 44 },
		{ 453, 219, 44 },
		{ 453, 220, 44 },
		{ 453, 221, 44 },
		{ 453, 222, 44 },
		{ 453, 223, 44 },
		{ 453, 224, 44 },
		{ 453, 225, 44 },
		{ 453, 226, 44 },
		{ 453, 227, 44 },
		{ 453, 228, 44 },
		{ 453, 229, 44 },
		{ 453, 230, 44 },
		{ 453, 231, 44 },
		{ 453, 232, 44 },
		{ 454, 184, 44 },
		{ 454, 185, 44 },
		{ 454, 186, 44 },
		{ 454, 187, 44 },
		{ 454, 188, 44 },
		{ 454, 189, 44 },
		{ 454, 190, 44 },
		{ 454, 191, 44 },
		{ 454, 192, 44 },
		{ 454, 193, 44 },
		{ 454, 194, 44 },
		{ 454, 195, 44 },
		{ 454, 196, 44 },
		{ 454, 197, 44 },
		{ 454, 198, 44 },
		{ 454, 199, 44 },
		{ 454, 200, 44 },
		{ 454, 201, 44 },
		{ 454, 202, 44 },
		{ 454, 203, 44 },
		{ 454, 204, 44 },
		{ 454, 205, 44 },
		{ 454, 206, 44 },
		{ 454, 207, 44 },
		{ 454, 208, 44 },
		{ 454, 209, 44 },
		{ 454, 210, 44 },
		{ 454, 211, 44 },
		{ 454, 212, 44 },
		{ 454, 213, 44 },
		{ 454, 214, 44 },
		{ 454, 215, 44 },
		{ 454, 216, 44 },
		{ 454, 217, 44 },
		{ 454, 218, 44 },
		{ 454, 219, 44 },
		{ 454, 220, 44 },
		{ 454, 221, 44 },
		{ 454, 222, 44 },
		{ 454, 223, 44 },
		{ 454, 224, 44 },
		{ 454, 225, 44 },
		{ 454, 226, 44 },
		{ 454, 227, 44 },
		{ 454, 228, 44 },
		{ 454, 229, 44 },
		{ 454, 230, 44 },
		{ 454, 231, 44 },
		{ 454, 232, 44 },
		{ 455, 183, 44 },
		{ 455, 184, 44 },
		{ 455, 185, 44 },
		{ 455, 186, 44 },
		{ 455, 187, 44 },
		{ 455, 188, 44 },
		{ 455, 189, 44 },
		{ 455, 190, 44 },
		{ 455, 191, 44 },
		{ 455, 192, 44 },
		{ 455, 193, 44 },
		{ 455, 194, 44 },
		{ 455, 195, 44 },
		{ 455, 196, 44 },
		{ 455, 197, 44 },
		{ 455, 198, 44 },
		{ 455, 199, 44 },
		{ 455, 200, 44 },
		{ 455, 201, 44 },
		{ 455, 202, 44 },
		{ 455, 203, 44 },
		{ 455, 204, 44 },
		{ 455, 205, 44 },
		{ 455, 206, 44 },
		{ 455, 207, 44 },
		{ 455, 208, 44 },
		{ 455, 209, 44 },
		{ 455, 210, 44 },
		{ 455, 211, 44 },
		{ 455, 212, 44 },
		{ 455, 213, 44 },
		{ 455, 214, 44 },
		{ 455, 215, 44 },
		{ 455, 216, 44 },
		{ 455, 217, 44 },
		{ 455, 218, 44 },
		{ 455, 219, 44 },
		{ 455, 220, 44 },
		{ 455, 221, 44 },
		{ 455, 222, 44 },
		{ 455, 223, 44 },
		{ 455, 224, 44 },
		{ 455, 225, 44 },
		{ 455, 226, 44 },
		{ 455, 227, 44 },
		{ 455, 228, 44 },
		{ 455, 229, 44 },
		{ 455, 230, 44 },
		{ 455, 231, 44 },
		{ 455, 232, 44 },
		{ 456, 183, 44 },
		{ 456, 184, 44 },
		{ 456, 185, 44 },
		{ 456, 186, 44 },
		{ 456, 187, 44 },
		{ 456, 188, 44 },
		{ 456, 189, 44 },
		{ 456, 190, 44 },
		{ 456, 191, 44 },
		{ 456, 192, 44 },
		{ 456, 193, 44 },
		{ 456, 194, 44 },
		{ 456, 195, 44 },
		{ 456, 196, 44 },
		{ 456, 197, 44 },
		{ 456, 198, 44 },
		{ 456, 199, 44 },
		{ 456, 200, 44 },
		{ 456, 201, 44 },
		{ 456, 202, 44 },
		{ 456, 203, 44 },
		{ 456, 204, 44 },
		{ 456, 205, 44 },
		{ 456, 206, 44 },
		{ 456, 207, 44 },
		{ 456, 208, 44 },
		{ 456, 209, 44 },
		{ 456, 210, 44 },
		{ 456, 211, 44 },
		{ 456, 212, 44 },
		{ 456, 213, 44 },
		{ 456, 214, 44 },
		{ 456, 215, 44 },
		{ 456, 216, 44 },
		{ 456, 217, 44 },
		{ 456, 218, 44 },
		{ 456, 219, 44 },
		{ 456, 220, 44 },
		{ 456, 221, 44 },
		{ 456, 222, 44 },
		{ 456, 223, 44 },
		{ 456, 224, 44 },
		{ 456, 225, 44 },
		{ 456, 226, 44 },
		{ 456, 227, 44 },
		{ 456, 228, 44 },
		{ 456, 229, 44 },
		{ 456, 230, 44 },
		{ 456, 231, 44 },
		{ 456, 232, 44 },
		{ 457, 183, 44 },
		{ 457, 184, 44 },
		{ 457, 185, 44 },
		{ 457, 186, 44 },
		{ 457, 187, 44 },
		{ 457, 188, 44 },
		{ 457, 189, 44 },
		{ 457, 190, 44 },
		{ 457, 191, 44 },
		{ 457, 192, 44 },
		{ 457, 193, 44 },
		{ 457, 194, 44 },
		{ 457, 195, 44 },
		{ 457, 196, 44 },
		{ 457, 197, 44 },
		{ 457, 198, 44 },
		{ 457, 199, 44 },
		{ 457, 200, 44 },
		{ 457, 201, 44 },
		{ 457, 202, 44 },
		{ 457, 203, 44 },
		{ 457, 204, 44 },
		{ 457, 205, 44 },
		{ 457, 206, 44 },
		{ 457, 207, 44 },
		{ 457, 208, 44 },
		{ 457, 209, 44 },
		{ 457, 210, 44 },
		{ 457, 211, 44 },
		{ 457, 212, 44 },
		{ 457, 213, 44 },
		{ 457, 214, 44 },
		{ 457, 215, 44 },
		{ 457, 216, 44 },
		{ 457, 217, 44 },
		{ 457, 218, 44 },
		{ 457, 219, 44 },
		{ 457, 220, 44 },
		{ 457, 221, 44 },
		{ 457, 222, 44 },
		{ 457, 223, 44 },
		{ 457, 224, 44 },
		{ 457, 225, 44 },
		{ 457, 226, 44 },
		{ 457, 227, 44 },
		{ 457, 228, 44 },
		{ 457, 229, 44 },
		{ 457, 230, 44 },
		{ 457, 231, 44 },
		{ 458, 182, 44 },
		{ 458, 183, 44 },
		{ 458, 184, 44 },
		{ 458, 185, 44 },
		{ 458, 186, 44 },
		{ 458, 187, 44 },
		{ 458, 188, 44 },
		{ 458, 189, 44 },
		{ 458, 190, 44 },
		{ 458, 191, 44 },
		{ 458, 192, 44 },
		{ 458, 193, 44 },
		{ 458, 194, 44 },
		{ 458, 195, 44 },
		{ 458, 196, 44 },
		{ 458, 197, 44 },
		{ 458, 198, 44 },
		{ 458, 199, 44 },
		{ 458, 200, 44 },
		{ 458, 201, 44 },
		{ 458, 202, 44 },
		{ 458, 203, 44 },
		{ 458, 204, 44 },
		{ 458, 205, 44 },
		{ 458, 206, 44 },
		{ 458, 207, 44 },
		{ 458, 208, 44 },
		{ 458, 209, 44 },
		{ 458, 210, 44 },
		{ 458, 211, 44 },
		{ 458, 212, 44 },
		{ 458, 213, 44 },
		{ 458, 214, 44 },
		{ 458, 215, 44 },
		{ 458, 216, 44 },
		{ 458, 217, 44 },
		{ 458, 218, 44 },
		{ 458, 219, 44 },
		{ 458, 220, 44 },
		{ 458, 221, 44 },
		{ 458, 222, 44 },
		{ 458, 223, 44 },
		{ 458, 224, 44 },
		{ 458, 225, 44 },
		{ 458, 226, 44 },
		{ 458, 227, 44 },
		{ 458, 228, 44 },
		{ 458, 229, 44 },
		{ 458, 230, 44 },
		{ 458, 231, 44 },
		{ 459, 182, 44 },
		{ 459, 183, 44 },
		{ 459, 184, 44 },
		{ 459, 185, 44 },
		{ 459, 186, 44 },
		{ 459, 187, 44 },
		{ 459, 188, 44 },
		{ 459, 189, 44 },
		{ 459, 190, 44 },
		{ 459, 191, 44 },
		{ 459, 192, 44 },
		{ 459, 193, 44 },
		{ 459, 194, 44 },
		{ 459, 195, 44 },
		{ 459, 196, 44 },
		{ 459, 197, 44 },
		{ 459, 198, 44 },
		{ 459, 199, 44 },
		{ 459, 200, 44 },
		{ 459, 201, 44 },
		{ 459, 202, 44 },
		{ 459, 203, 44 },
		{ 459, 204, 44 },
		{ 459, 205, 44 },
		{ 459, 206, 44 },
		{ 459, 207, 44 },
		{ 459, 208, 44 },
		{ 459, 209, 44 },
		{ 459, 210, 44 },
		{ 459, 211, 44 },
		{ 459, 212, 44 },
		{ 459, 213, 44 },
		{ 459, 214, 44 },
		{ 459, 215, 44 },
		{ 459, 216, 44 },
		{ 459, 217, 44 },
		{ 459, 218, 44 },
		{ 459, 219, 44 },
		{ 459, 220, 44 },
		{ 459, 221, 44 },
		{ 459, 222, 44 },
		{ 459, 223, 44 },
		{ 459, 224, 44 },
		{ 459, 225, 44 },
		{ 459, 226, 44 },
		{ 459, 227, 44 },
		{ 459, 228, 44 },
		{ 459, 229, 44 },
		{ 459, 230, 44 },
		{ 459, 231, 44 },
		{ 460, 181, 44 },
		{ 460, 182, 44 },
		{ 460, 183, 44 },
		{ 460, 184, 44 },
		{ 460, 185, 44 },
		{ 460, 186, 44 },
		{ 460, 187, 44 },
		{ 460, 188, 44 },
		{ 460, 189, 44 },
		{ 460, 190, 44 },
		{ 460, 191, 44 },
		{ 460, 192, 44 },
		{ 460, 193, 44 },
		{ 460, 194, 44 },
		{ 460, 195, 44 },
		{ 460, 196, 44 },
		{ 460, 197, 44 },
		{ 460, 198, 44 },
		{ 460, 199, 44 },
		{ 460, 200, 44 },
		{ 460, 201, 44 },
		{ 460, 202, 44 },
		{ 460, 203, 44 },
		{ 460, 204, 44 },
		{ 460, 205, 44 },
		{ 460, 206, 44 },
		{ 460, 207, 44 },
		{ 460, 208, 44 },
		{ 460, 209, 44 },
		{ 460, 210, 44 },
		{ 460, 211, 44 },
		{ 460, 212, 44 },
		{ 460, 213, 44 },
		{ 460, 214, 44 },
		{ 460, 215, 44 },
		{ 460, 216, 44 },
		{ 460, 217, 44 },
		{ 460, 218, 44 },
		{ 460, 219, 44 },
		{ 460, 220, 44 },
		{ 460, 221, 44 },
		{ 460, 222, 44 },
		{ 460, 223, 44 },
		{ 460, 224, 44 },
		{ 460, 225, 44 },
		{ 460, 226, 44 },
		{ 460, 227, 44 },
		{ 460, 228, 44 },
		{ 460, 229, 44 },
		{ 460, 230, 44 },
		{ 460, 231, 44 },
		{ 461, 181, 44 },
		{ 461, 182, 44 },
		{ 461, 183, 44 },
		{ 461, 184, 44 },
		{ 461, 185, 44 },
		{ 461, 186, 44 },
		{ 461, 187, 44 },
		{ 461, 188, 44 },
		{ 461, 189, 44 },
		{ 461, 190, 44 },
		{ 461, 191, 44 },
		{ 461, 192, 44 },
		{ 461, 193, 44 },
		{ 461, 194, 44 },
		{ 461, 195, 44 },
		{ 461, 196, 44 },
		{ 461, 197, 44 },
		{ 461, 198, 44 },
		{ 461, 199, 44 },
		{ 461, 200, 44 },
		{ 461, 201, 44 },
		{ 461, 202, 44 },
		{ 461, 203, 44 },
		{ 461, 204, 44 },
		{ 461, 205, 44 },
		{ 461, 206, 44 },
		{ 461, 207, 44 },
		{ 461, 208, 44 },
		{ 461, 209, 44 },
		{ 461, 210, 44 },
		{ 461, 211, 44 },
		{ 461, 212, 44 },
		{ 461, 213, 44 },
		{ 461, 214, 44 },
		{ 461, 215, 44 },
		{ 461, 216, 44 },
		{ 461, 217, 44 },
		{ 461, 218, 44 },
		{ 461, 219, 44 },
		{ 461, 220, 44 },
		{ 461, 221, 44 },
		{ 461, 222, 44 },
		{ 461, 223, 44 },
		{ 461, 224, 44 },
		{ 461, 225, 44 },
		{ 461, 226, 44 },
		{ 461, 227, 44 },
		{ 461, 228, 44 },
		{ 461, 229, 44 },
		{ 461, 230, 44 },
		{ 461, 231, 44 },
		{ 462, 181, 44 },
		{ 462, 182, 44 },
		{ 462, 183, 44 },
		{ 462, 184, 44 },
		{ 462, 185, 44 },
		{ 462, 186, 44 },
		{ 462, 187, 44 },
		{ 462, 188, 44 },
		{ 462, 189, 44 },
		{ 462, 190, 44 },
		{ 462, 191, 44 },
		{ 462, 192, 44 },
		{ 462, 193, 44 },
		{ 462, 194, 44 },
		{ 462, 195, 44 },
		{ 462, 196, 44 },
		{ 462, 197, 44 },
		{ 462, 198, 44 },
		{ 462, 199, 44 },
		{ 462, 200, 44 },
		{ 462, 201, 44 },
		{ 462, 202, 44 },
		{ 462, 203, 44 },
		{ 462, 204, 44 },
		{ 462, 205, 44 },
		{ 462, 206, 44 },
		{ 462, 207, 44 },
		{ 462, 208, 44 },
		{ 462, 209, 44 },
		{ 462, 210, 44 },
		{ 462, 211, 44 },
		{ 462, 212, 44 },
		{ 462, 213, 44 },
		{ 462, 214, 44 },
		{ 462, 215, 44 },
		{ 462, 216, 44 },
		{ 462, 217, 44 },
		{ 462, 218, 44 },
		{ 462, 219, 44 },
		{ 462, 220, 44 },
		{ 462, 221, 44 },
		{ 462, 222, 44 },
		{ 462, 223, 44 },
		{ 462, 224, 44 },
		{ 462, 225, 44 },
		{ 462, 226, 44 },
		{ 462, 227, 44 },
		{ 462, 228, 44 },
		{ 462, 229, 44 },
		{ 462, 230, 44 },
		{ 462, 231, 44 },
		{ 463, 180, 44 },
		{ 463, 181, 44 },
		{ 463, 182, 44 },
		{ 463, 183, 44 },
		{ 463, 184, 44 },
		{ 463, 185, 44 },
		{ 463, 186, 44 },
		{ 463, 187, 44 },
		{ 463, 188, 44 },
		{ 463, 189, 44 },
		{ 463, 190, 44 },
		{ 463, 191, 44 },
		{ 463, 192, 44 },
		{ 463, 193, 44 },
		{ 463, 194, 44 },
		{ 463, 195, 44 },
		{ 463, 196, 44 },
		{ 463, 197, 44 },
		{ 463, 198, 44 },
		{ 463, 199, 44 },
		{ 463, 200, 44 },
		{ 463, 201, 44 },
		{ 463, 202, 44 },
		{ 463, 203, 44 },
		{ 463, 204, 44 },
		{ 463, 205, 44 },
		{ 463, 206, 44 },
		{ 463, 207, 44 },
		{ 463, 208, 44 },
		{ 463, 209, 44 },
		{ 463, 210, 44 },
		{ 463, 211, 44 },
		{ 463, 212, 44 },
		{ 463, 213, 44 },
		{ 463, 214, 44 },
		{ 463, 215, 44 },
		{ 463, 216, 44 },
		{ 463, 217, 44 },
		{ 463, 218, 44 },
		{ 463, 219, 44 },
		{ 463, 220, 44 },
		{ 463, 221, 44 },
		{ 463, 222, 44 },
		{ 463, 223, 44 },
		{ 463, 224, 44 },
		{ 463, 225, 44 },
		{ 463, 226, 44 },
		{ 463, 227, 44 },
		{ 463, 228, 44 },
		{ 463, 229, 44 },
		{ 463, 230, 44 },
		{ 463, 231, 44 },
		{ 464, 180, 44 },
		{ 464, 181, 44 },
		{ 464, 182, 44 },
		{ 464, 183, 44 },
		{ 464, 184, 44 },
		{ 464, 185, 44 },
		{ 464, 186, 44 },
		{ 464, 187, 44 },
		{ 464, 188, 44 },
		{ 464, 189, 44 },
		{ 464, 190, 44 },
		{ 464, 191, 44 },
		{ 464, 192, 44 },
		{ 464, 193, 44 },
		{ 464, 194, 44 },
		{ 464, 195, 44 },
		{ 464, 196, 44 },
		{ 464, 197, 44 },
		{ 464, 198, 44 },
		{ 464, 199, 44 },
		{ 464, 200, 44 },
		{ 464, 201, 44 },
		{ 464, 202, 44 },
		{ 464, 203, 44 },
		{ 464, 204, 44 },
		{ 464, 205, 44 },
		{ 464, 206, 44 },
		{ 464, 207, 44 },
		{ 464, 208, 44 },
		{ 464, 209, 44 },
		{ 464, 210, 44 },
		{ 464, 211, 44 },
		{ 464, 212, 44 },
		{ 464, 213, 44 },
		{ 464, 214, 44 },
		{ 464, 215, 44 },
		{ 464, 216, 44 },
		{ 464, 217, 44 },
		{ 464, 218, 44 },
		{ 464, 219, 44 },
		{ 464, 220, 44 },
		{ 464, 221, 44 },
		{ 464, 222, 44 },
		{ 464, 223, 44 },
		{ 464, 224, 44 },
		{ 464, 225, 44 },
		{ 464, 226, 44 },
		{ 464, 227, 44 },
		{ 464, 228, 44 },
		{ 464, 229, 44 },
		{ 464, 230, 44 },
		{ 464, 231, 44 },
		{ 465, 180, 44 },
		{ 465, 181, 44 },
		{ 465, 182, 44 },
		{ 465, 183, 44 },
		{ 465, 184, 44 },
		{ 465, 185, 44 },
		{ 465, 186, 44 },
		{ 465, 187, 44 },
		{ 465, 188, 44 },
		{ 465, 189, 44 },
		{ 465, 190, 44 },
		{ 465, 191, 44 },
		{ 465, 192, 44 },
		{ 465, 193, 44 },
		{ 465, 194, 44 },
		{ 465, 195, 44 },
		{ 465, 196, 44 },
		{ 465, 197, 44 },
		{ 465, 198, 44 },
		{ 465, 199, 44 },
		{ 465, 200, 44 },
		{ 465, 201, 44 },
		{ 465, 202, 44 },
		{ 465, 203, 44 },
		{ 465, 204, 44 },
		{ 465, 205, 44 },
		{ 465, 206, 44 },
		{ 465, 207, 44 },
		{ 465, 208, 44 },
		{ 465, 209, 44 },
		{ 465, 210, 44 },
		{ 465, 211, 44 },
		{ 465, 212, 44 },
		{ 465, 213, 44 },
		{ 465, 214, 44 },
		{ 465, 215, 44 },
		{ 465, 216, 44 },
		{ 465, 217, 44 },
		{ 465, 218, 44 },
		{ 465, 219, 44 },
		{ 465, 220, 44 },
		{ 465, 221, 44 },
		{ 465, 222, 44 },
		{ 465, 223, 44 },
		{ 465, 224, 44 },
		{ 465, 225, 44 },
		{ 465, 226, 44 },
		{ 465, 227, 44 },
		{ 465, 228, 44 },
		{ 465, 229, 44 },
		{ 465, 230, 44 },
		{ 466, 179, 44 },
		{ 466, 180, 44 },
		{ 466, 181, 44 },
		{ 466, 182, 44 },
		{ 466, 183, 44 },
		{ 466, 184, 44 },
		{ 466, 185, 44 },
		{ 466, 186, 44 },
		{ 466, 187, 44 },
		{ 466, 188, 44 },
		{ 466, 189, 44 },
		{ 466, 190, 44 },
		{ 466, 191, 44 },
		{ 466, 192, 44 },
		{ 466, 193, 44 },
		{ 466, 194, 44 },
		{ 466, 195, 44 },
		{ 466, 196, 44 },
		{ 466, 197, 44 },
		{ 466, 198, 44 },
		{ 466, 199, 44 },
		{ 466, 200, 44 },
		{ 466, 201, 44 },
		{ 466, 202, 44 },
		{ 466, 203, 44 },
		{ 466, 204, 44 },
		{ 466, 205, 44 },
		{ 466, 206, 44 },
		{ 466, 207, 44 },
		{ 466, 208, 44 },
		{ 466, 209, 44 },
		{ 466, 210, 44 },
		{ 466, 211, 44 },
		{ 466, 212, 44 },
		{ 466, 213, 44 },
		{ 466, 214, 44 },
		{ 466, 215, 44 },
		{ 466, 216, 44 },
		{ 466, 217, 44 },
		{ 466, 218, 44 },
		{ 466, 219, 44 },
		{ 466, 220, 44 },
		{ 466, 221, 44 },
		{ 466, 222, 44 },
		{ 466, 223, 44 },
		{ 466, 224, 44 },
		{ 466, 225, 44 },
		{ 466, 226, 44 },
		{ 466, 227, 44 },
		{ 466, 228, 44 },
		{ 466, 229, 44 },
		{ 466, 230, 44 },
		{ 467, 179, 44 },
		{ 467, 180, 44 },
		{ 467, 181, 44 },
		{ 467, 182, 44 },
		{ 467, 183, 44 },
		{ 467, 184, 44 },
		{ 467, 185, 44 },
		{ 467, 186, 44 },
		{ 467, 187, 44 },
		{ 467, 188, 44 },
		{ 467, 189, 44 },
		{ 467, 190, 44 },
		{ 467, 191, 44 },
		{ 467, 192, 44 },
		{ 467, 193, 44 },
		{ 467, 194, 44 },
		{ 467, 195, 44 },
		{ 467, 196, 44 },
		{ 467, 197, 44 },
		{ 467, 198, 44 },
		{ 467, 199, 44 },
		{ 467, 200, 44 },
		{ 467, 201, 44 },
		{ 467, 202, 44 },
		{ 467, 203, 44 },
		{ 467, 204, 44 },
		{ 467, 205, 44 },
		{ 467, 206, 44 },
		{ 467, 207, 44 },
		{ 467, 208, 44 },
		{ 467, 209, 44 },
		{ 467, 210, 44 },
		{ 467, 211, 44 },
		{ 467, 212, 44 },
		{ 467, 213, 44 },
		{ 467, 214, 44 },
		{ 467, 215, 44 },
		{ 467, 216, 44 },
		{ 467, 217, 44 },
		{ 467, 218, 44 },
		{ 467, 219, 44 },
		{ 467, 220, 44 },
		{ 467, 221, 44 },
		{ 467, 222, 44 },
		{ 467, 223, 44 },
		{ 467, 224, 44 },
		{ 467, 225, 44 },
		{ 467, 226, 44 },
		{ 467, 227, 44 },
		{ 467, 228, 44 },
		{ 467, 229, 44 },
		{ 467, 230, 44 },
		{ 468, 178, 44 },
		{ 468, 179, 44 },
		{ 468, 180, 44 },
		{ 468, 181, 44 },
		{ 468, 182, 44 },
		{ 468, 183, 44 },
		{ 468, 184, 44 },
		{ 468, 185, 44 },
		{ 468, 186, 44 },
		{ 468, 187, 44 },
		{ 468, 188, 44 },
		{ 468, 189, 44 },
		{ 468, 190, 44 },
		{ 468, 191, 44 },
		{ 468, 192, 44 },
		{ 468, 193, 44 },
		{ 468, 194, 44 },
		{ 468, 195, 44 },
		{ 468, 196, 44 },
		{ 468, 197, 44 },
		{ 468, 198, 44 },
		{ 468, 199, 44 },
		{ 468, 200, 44 },
		{ 468, 201, 44 },
		{ 468, 202, 44 },
		{ 468, 203, 44 },
		{ 468, 204, 44 },
		{ 468, 205, 44 },
		{ 468, 206, 44 },
		{ 468, 207, 44 },
		{ 468, 208, 44 },
		{ 468, 209, 44 },
		{ 468, 210, 44 },
		{ 468, 211, 44 },
		{ 468, 212, 44 },
		{ 468, 213, 44 },
		{ 468, 214, 44 },
		{ 468, 215, 44 },
		{ 468, 216, 44 },
		{ 468, 217, 44 },
		{ 468, 218, 44 },
		{ 468, 219, 44 },
		{ 468, 220, 44 },
		{ 468, 221, 44 },
		{ 468, 222, 44 },
		{ 468, 223, 44 },
		{ 468, 224, 44 },
		{ 468, 225, 44 },
		{ 468, 226, 44 },
		{ 468, 227, 44 },
		{ 468, 228, 44 },
		{ 468, 229, 44 },
		{ 468, 230, 44 },
		{ 469, 178, 44 },
		{ 469, 179, 44 },
		{ 469, 180, 44 },
		{ 469, 181, 44 },
		{ 469, 182, 44 },
		{ 469, 183, 44 },
		{ 469, 184, 44 },
		{ 469, 185, 44 },
		{ 469, 186, 44 },
		{ 469, 187, 44 },
		{ 469, 188, 44 },
		{ 469, 189, 44 },
		{ 469, 190, 44 },
		{ 469, 191, 44 },
		{ 469, 192, 44 },
		{ 469, 193, 44 },
		{ 469, 194, 44 },
		{ 469, 195, 44 },
		{ 469, 196, 44 },
		{ 469, 197, 44 },
		{ 469, 198, 44 },
		{ 469, 199, 44 },
		{ 469, 200, 44 },
		{ 469, 201, 44 },
		{ 469, 202, 44 },
		{ 469, 203, 44 },
		{ 469, 204, 44 },
		{ 469, 205, 44 },
		{ 469, 206, 44 },
		{ 469, 207, 44 },
		{ 469, 208, 44 },
		{ 469, 209, 44 },
		{ 469, 210, 44 },
		{ 469, 211, 44 },
		{ 469, 212, 44 },
		{ 469, 213, 44 },
		{ 469, 214, 44 },
		{ 469, 215, 44 },
		{ 469, 216, 44 },
		{ 469, 217, 44 },
		{ 469, 218, 44 },
		{ 469, 219, 44 },
		{ 469, 220, 44 },
		{ 469, 221, 44 },
		{ 469, 222, 44 },
		{ 469, 223, 44 },
		{ 469, 224, 44 },
		{ 469, 225, 44 },
		{ 469, 226, 44 },
		{ 469, 227, 44 },
		{ 469, 228, 44 },
		{ 469, 229, 44 },
		{ 469, 230, 44 },
		{ 470, 178, 44 },
		{ 470, 179, 44 },
		{ 470, 180, 44 },
		{ 470, 181, 44 },
		{ 470, 182, 44 },
		{ 470, 183, 44 },
		{ 470, 184, 44 },
		{ 470, 185, 44 },
		{ 470, 186, 44 },
		{ 470, 187, 44 },
		{ 470, 188, 44 },
		{ 470, 189, 44 },
		{ 470, 190, 44 },
		{ 470, 191, 44 },
		{ 470, 192, 44 },
		{ 470, 193, 44 },
		{ 470, 194, 44 },
		{ 470, 195, 44 },
		{ 470, 196, 44 },
		{ 470, 197, 44 },
		{ 470, 198, 44 },
		{ 470, 199, 44 },
		{ 470, 200, 44 },
		{ 470, 201, 44 },
		{ 470, 202, 44 },
		{ 470, 203, 44 },
		{ 470, 204, 44 },
		{ 470, 205, 44 },
		{ 470, 206, 44 },
		{ 470, 207, 44 },
		{ 470, 208, 44 },
		{ 470, 209, 44 },
		{ 470, 210, 44 },
		{ 470, 211, 44 },
		{ 470, 212, 44 },
		{ 470, 213, 44 },
		{ 470, 214, 44 },
		{ 470, 215, 44 },
		{ 470, 216, 44 },
		{ 470, 217, 44 },
		{ 470, 218, 44 },
		{ 470, 219, 44 },
		{ 470, 220, 44 },
		{ 470, 221, 44 },
		{ 470, 222, 44 },
		{ 470, 223, 44 },
		{ 470, 224, 44 },
		{ 470, 225, 44 },
		{ 470, 226, 44 },
		{ 470, 227, 44 },
		{ 470, 228, 44 },
		{ 470, 229, 44 },
		{ 470, 230, 44 },
		{ 471, 177, 44 },
		{ 471, 178, 44 },
		{ 471, 179, 44 },
		{ 471, 180, 44 },
		{ 471, 181, 44 },
		{ 471, 182, 44 },
		{ 471, 183, 44 },
		{ 471, 184, 44 },
		{ 471, 185, 44 },
		{ 471, 186, 44 },
		{ 471, 187, 44 },
		{ 471, 188, 44 },
		{ 471, 189, 44 },
		{ 471, 190, 44 },
		{ 471, 191, 44 },
		{ 471, 192, 44 },
		{ 471, 193, 44 },
		{ 471, 194, 44 },
		{ 471, 195, 44 },
		{ 471, 196, 44 },
		{ 471, 197, 44 },
		{ 471, 198, 44 },
		{ 471, 199, 44 },
		{ 471, 200, 44 },
		{ 471, 201, 44 },
		{ 471, 202, 44 },
		{ 471, 203, 44 },
		{ 471, 204, 44 },
		{ 471, 205, 44 },
		{ 471, 206, 44 },
		{ 471, 207, 44 },
		{ 471, 208, 44 },
		{ 471, 209, 44 },
		{ 471, 210, 44 },
		{ 471, 211, 44 },
		{ 471, 212, 44 },
		{ 471, 213, 44 },
		{ 471, 214, 44 },
		{ 471, 215, 44 },
		{ 471, 216, 44 },
		{ 471, 217, 44 },
		{ 471, 218, 44 },
		{ 471, 219, 44 },
		{ 471, 220, 44 },
		{ 471, 221, 44 },
		{ 471, 222, 44 },
		{ 471, 223, 44 },
		{ 471, 224, 44 },
		{ 471, 225, 44 },
		{ 471, 226, 44 },
		{ 471, 227, 44 },
		{ 471, 228, 44 },
		{ 471, 229, 44 },
		{ 471, 230, 44 },
		{ 472, 177, 44 },
		{ 472, 178, 44 },
		{ 472, 179, 44 },
		{ 472, 180, 44 },
		{ 472, 181, 44 },
		{ 472, 182, 44 },
		{ 472, 183, 44 },
		{ 472, 184, 44 },
		{ 472, 185, 44 },
		{ 472, 186, 44 },
		{ 472, 187, 44 },
		{ 472, 188, 44 },
		{ 472, 189, 44 },
		{ 472, 190, 44 },
		{ 472, 191, 44 },
		{ 472, 192, 44 },
		{ 472, 193, 44 },
		{ 472, 194, 44 },
		{ 472, 195, 44 },
		{ 472, 196, 44 },
		{ 472, 197, 44 },
		{ 472, 198, 44 },
		{ 472, 199, 44 },
		{ 472, 200, 44 },
		{ 472, 201, 44 },
		{ 472, 202, 44 },
		{ 472, 203, 44 },
		{ 472, 204, 44 },
		{ 472, 205, 44 },
		{ 472, 206, 44 },
		{ 472, 207, 44 },
		{ 472, 208, 44 },
		{ 472, 209, 44 },
		{ 472, 210, 44 },
		{ 472, 211, 44 },
		{ 472, 212, 44 },
		{ 472, 213, 44 },
		{ 472, 214, 44 },
		{ 472, 215, 44 },
		{ 472, 216, 44 },
		{ 472, 217, 44 },
		{ 472, 218, 44 },
		{ 472, 219, 44 },
		{ 472, 220, 44 },
		{ 472, 221, 44 },
		{ 472, 222, 44 },
		{ 472, 223, 44 },
		{ 472, 224, 44 },
		{ 472, 225, 44 },
		{ 472, 226, 44 },
		{ 472, 227, 44 },
		{ 472, 228, 44 },
		{ 472, 229, 44 },
		{ 472, 230, 44 },
		{ 473, 177, 44 },
		{ 473, 178, 44 },
		{ 473, 179, 44 },
		{ 473, 180, 44 },
		{ 473, 181, 44 },
		{ 473, 182, 44 },
		{ 473, 183, 44 },
		{ 473, 184, 44 },
		{ 473, 185, 44 },
		{ 473, 186, 44 },
		{ 473, 187, 44 },
		{ 473, 188, 44 },
		{ 473, 189, 44 },
		{ 473, 190, 44 },
		{ 473, 191, 44 },
		{ 473, 192, 44 },
		{ 473, 193, 44 },
		{ 473, 194, 44 },
		{ 473, 195, 44 },
		{ 473, 196, 44 },
		{ 473, 197, 44 },
		{ 473, 198, 44 },
		{ 473, 199, 44 },
		{ 473, 200, 44 },
		{ 473, 201, 44 },
		{ 473, 202, 44 },
		{ 473, 203, 44 },
		{ 473, 204, 44 },
		{ 473, 205, 44 },
		{ 473, 206, 44 },
		{ 473, 207, 44 },
		{ 473, 208, 44 },
		{ 473, 209, 44 },
		{ 473, 210, 44 },
		{ 473, 211, 44 },
		{ 473, 212, 44 },
		{ 473, 213, 44 },
		{ 473, 214, 44 },
		{ 473, 215, 44 },
		{ 473, 216, 44 },
		{ 473, 217, 44 },
		{ 473, 218, 44 },
		{ 473, 219, 44 },
		{ 473, 220, 44 },
		{ 473, 221, 44 },
		{ 473, 222, 44 },
		{ 473, 223, 44 },
		{ 473, 224, 44 },
		{ 473, 225, 44 },
		{ 473, 226, 44 },
		{ 473, 227, 44 },
		{ 473, 228, 44 },
		{ 473, 229, 44 },
		{ 474, 176, 44 },
		{ 474, 177, 44 },
		{ 474, 178, 44 },
		{ 474, 179, 44 },
		{ 474, 180, 44 },
		{ 474, 181, 44 },
		{ 474, 182, 44 },
		{ 474, 183, 44 },
		{ 474, 184, 44 },
		{ 474, 185, 44 },
		{ 474, 186, 44 },
		{ 474, 187, 44 },
		{ 474, 188, 44 },
		{ 474, 189, 44 },
		{ 474, 190, 44 },
		{ 474, 191, 44 },
		{ 474, 192, 44 },
		{ 474, 193, 44 },
		{ 474, 194, 44 },
		{ 474, 195, 44 },
		{ 474, 196, 44 },
		{ 474, 197, 44 },
		{ 474, 198, 44 },
		{ 474, 199, 44 },
		{ 474, 200, 44 },
		{ 474, 201, 44 },
		{ 474, 202, 44 },
		{ 474, 203, 44 },
		{ 474, 204, 44 },
		{ 474, 205, 44 },
		{ 474, 206, 44 },
		{ 474, 207, 44 },
		{ 474, 208, 44 },
		{ 474, 209, 44 },
		{ 474, 210, 44 },
		{ 474, 211, 44 },
		{ 474, 212, 44 },
		{ 474, 213, 44 },
		{ 474, 214, 44 },
		{ 474, 215, 44 },
		{ 474, 216, 44 },
		{ 474, 217, 44 },
		{ 474, 218, 44 },
		{ 474, 219, 44 },
		{ 474, 220, 44 },
		{ 474, 221, 44 },
		{ 474, 222, 44 },
		{ 474, 223, 44 },
		{ 474, 224, 44 },
		{ 474, 225, 44 },
		{ 474, 226, 44 },
		{ 474, 227, 44 },
		{ 474, 228, 44 },
		{ 474, 229, 44 },
		{ 475, 176, 44 },
		{ 475, 177, 44 },
		{ 475, 178, 44 },
		{ 475, 179, 44 },
		{ 475, 180, 44 },
		{ 475, 181, 44 },
		{ 475, 182, 44 },
		{ 475, 183, 44 },
		{ 475, 184, 44 },
		{ 475, 185, 44 },
		{ 475, 186, 44 },
		{ 475, 187, 44 },
		{ 475, 188, 44 },
		{ 475, 189, 44 },
		{ 475, 190, 44 },
		{ 475, 191, 44 },
		{ 475, 192, 44 },
		{ 475, 193, 44 },
		{ 475, 194, 44 },
		{ 475, 195, 44 },
		{ 475, 196, 44 },
		{ 475, 197, 44 },
		{ 475, 198, 44 },
		{ 475, 199, 44 },
		{ 475, 200, 44 },
		{ 475, 201, 44 },
		{ 475, 202, 44 },
		{ 475, 203, 44 },
		{ 475, 204, 44 },
		{ 475, 205, 44 },
		{ 475, 206, 44 },
		{ 475, 207, 44 },
		{ 475, 208, 44 },
		{ 475, 209, 44 },
		{ 475, 210, 44 },
		{ 475, 211, 44 },
		{ 475, 212, 44 },
		{ 475, 213, 44 },
		{ 475, 214, 44 },
		{ 475, 215, 44 },
		{ 475, 216, 44 },
		{ 475, 217, 44 },
		{ 475, 218, 44 },
		{ 475, 219, 44 },
		{ 475, 220, 44 },
		{ 475, 221, 44 },
		{ 475, 222, 44 },
		{ 475, 223, 44 },
		{ 475, 224, 44 },
		{ 475, 225, 44 },
		{ 475, 226, 44 },
		{ 475, 227, 44 },
		{ 475, 228, 44 },
		{ 475, 229, 44 },
		{ 476, 175, 44 },
		{ 476, 176, 44 },
		{ 476, 177, 44 },
		{ 476, 178, 44 },
		{ 476, 179, 44 },
		{ 476, 180, 44 },
		{ 476, 181, 44 },
		{ 476, 182, 44 },
		{ 476, 183, 44 },
		{ 476, 184, 44 },
		{ 476, 185, 44 },
		{ 476, 186, 44 },
		{ 476, 187, 44 },
		{ 476, 188, 44 },
		{ 476, 189, 44 },
		{ 476, 190, 44 },
		{ 476, 191, 44 },
		{ 476, 192, 44 },
		{ 476, 193, 44 },
		{ 476, 194, 44 },
		{ 476, 195, 44 },
		{ 476, 196, 44 },
		{ 476, 197, 44 },
		{ 476, 198, 44 },
		{ 476, 199, 44 },
		{ 476, 200, 44 },
		{ 476, 201, 44 },
		{ 476, 202, 44 },
		{ 476, 203, 44 },
		{ 476, 204, 44 },
		{ 476, 205, 44 },
		{ 476, 206, 44 },
		{ 476, 207, 44 },
		{ 476, 208, 44 },
		{ 476, 209, 44 },
		{ 476, 210, 44 },
		{ 476, 211, 44 },
		{ 476, 212, 44 },
		{ 476, 213, 44 },
		{ 476, 214, 44 },
		{ 476, 215, 44 },
		{ 476, 216, 44 },
		{ 476, 217, 44 },
		{ 476, 218, 44 },
		{ 476, 219, 44 },
		{ 476, 220, 44 },
		{ 476, 221, 44 },
		{ 476, 222, 44 },
		{ 476, 223, 44 },
		{ 476, 224, 44 },
		{ 476, 225, 44 },
		{ 476, 226, 44 },
		{ 476, 227, 44 },
		{ 476, 228, 44 },
		{ 476, 229, 44 },
		{ 477, 175, 44 },
		{ 477, 176, 44 },
		{ 477, 177, 44 },
		{ 477, 178, 44 },
		{ 477, 179, 44 },
		{ 477, 180, 44 },
		{ 477, 181, 44 },
		{ 477, 182, 44 },
		{ 477, 183, 44 },
		{ 477, 184, 44 },
		{ 477, 185, 44 },
		{ 477, 186, 44 },
		{ 477, 187, 44 },
		{ 477, 188, 44 },
		{ 477, 189, 44 },
		{ 477, 190, 44 },
		{ 477, 191, 44 },
		{ 477, 192, 44 },
		{ 477, 193, 44 },
		{ 477, 194, 44 },
		{ 477, 195, 44 },
		{ 477, 196, 44 },
		{ 477, 197, 44 },
		{ 477, 198, 44 },
		{ 477, 199, 44 },
		{ 477, 200, 44 },
		{ 477, 201, 44 },
		{ 477, 202, 44 },
		{ 477, 203, 44 },
		{ 477, 204, 44 },
		{ 477, 205, 44 },
		{ 477, 206, 44 },
		{ 477, 207, 44 },
		{ 477, 208, 44 },
		{ 477, 209, 44 },
		{ 477, 210, 44 },
		{ 477, 211, 44 },
		{ 477, 212, 44 },
		{ 477, 213, 44 },
		{ 477, 214, 44 },
		{ 477, 215, 44 },
		{ 477, 216, 44 },
		{ 477, 217, 44 },
		{ 477, 218, 44 },
		{ 477, 219, 44 },
		{ 477, 220, 44 },
		{ 477, 221, 44 },
		{ 477, 222, 44 },
		{ 477, 223, 44 },
		{ 477, 224, 44 },
		{ 477, 225, 44 },
		{ 477, 226, 44 },
		{ 477, 227, 44 },
		{ 477, 228, 44 },
		{ 477, 229, 44 },
		{ 478, 175, 44 },
		{ 478, 176, 44 },
		{ 478, 177, 44 },
		{ 478, 178, 44 },
		{ 478, 179, 44 },
		{ 478, 180, 44 },
		{ 478, 181, 44 },
		{ 478, 182, 44 },
		{ 478, 183, 44 },
		{ 478, 184, 44 },
		{ 478, 185, 44 },
		{ 478, 186, 44 },
		{ 478, 187, 44 },
		{ 478, 188, 44 },
		{ 478, 189, 44 },
		{ 478, 190, 44 },
		{ 478, 191, 44 },
		{ 478, 192, 44 },
		{ 478, 193, 44 },
		{ 478, 194, 44 },
		{ 478, 195, 44 },
		{ 478, 196, 44 },
		{ 478, 197, 44 },
		{ 478, 198, 44 },
		{ 478, 199, 44 },
		{ 478, 200, 44 },
		{ 478, 201, 44 },
		{ 478, 202, 44 },
		{ 478, 203, 44 },
		{ 478, 204, 44 },
		{ 478, 205, 44 },
		{ 478, 206, 44 },
		{ 478, 207, 44 },
		{ 478, 208, 44 },
		{ 478, 209, 44 },
		{ 478, 210, 44 },
		{ 478, 211, 44 },
		{ 478, 212, 44 },
		{ 478, 213, 44 },
		{ 478, 214, 44 },
		{ 478, 215, 44 },
		{ 478, 216, 44 },
		{ 478, 217, 44 },
		{ 478, 218, 44 },
		{ 478, 219, 44 },
		{ 478, 220, 44 },
		{ 478, 221, 44 },
		{ 478, 222, 44 },
		{ 478, 223, 44 },
		{ 478, 224, 44 },
		{ 478, 225, 44 },
		{ 478, 226, 44 },
		{ 478, 227, 44 },
		{ 478, 228, 44 },
		{ 478, 229, 44 },
		{ 479, 174, 44 },
		{ 479, 175, 44 },
		{ 479, 176, 44 },
		{ 479, 177, 44 },
		{ 479, 178, 44 },
		{ 479, 179, 44 },
		{ 479, 180, 44 },
		{ 479, 181, 44 },
		{ 479, 182, 44 },
		{ 479, 183, 44 },
		{ 479, 184, 44 },
		{ 479, 185, 44 },
		{ 479, 186, 44 },
		{ 479, 187, 44 },
		{ 479, 188, 44 },
		{ 479, 189, 44 },
		{ 479, 190, 44 },
		{ 479, 191, 44 },
		{ 479, 192, 44 },
		{ 479, 193, 44 },
		{ 479, 194, 44 },
		{ 479, 195, 44 },
		{ 479, 196, 44 },
		{ 479, 197, 44 },
		{ 479, 198, 44 },
		{ 479, 199, 44 },
		{ 479, 200, 44 },
		{ 479, 201, 44 },
		{ 479, 202, 44 },
		{ 479, 203, 44 },
		{ 479, 204, 44 },
		{ 479, 205, 44 },
		{ 479, 206, 44 },
		{ 479, 207, 44 },
		{ 479, 208, 44 },
		{ 479, 209, 44 },
		{ 479, 210, 44 },
		{ 479, 211, 44 },
		{ 479, 212, 44 },
		{ 479, 213, 44 },
		{ 479, 214, 44 },
		{ 479, 215, 44 },
		{ 479, 216, 44 },
		{ 479, 217, 44 },
		{ 479, 218, 44 },
		{ 479, 219, 44 },
		{ 479, 220, 44 },
		{ 479, 221, 44 },
		{ 479, 222, 44 },
		{ 479, 223, 44 },
		{ 479, 224, 44 },
		{ 479, 225, 44 },
		{ 479, 226, 44 },
		{ 479, 227, 44 },
		{ 479, 228, 44 },
		{ 479, 229, 44 },
		{ 480, 174, 44 },
		{ 480, 175, 44 },
		{ 480, 176, 44 },
		{ 480, 177, 44 },
		{ 480, 178, 44 },
		{ 480, 179, 44 },
		{ 480, 180, 44 },
		{ 480, 181, 44 },
		{ 480, 182, 44 },
		{ 480, 183, 44 },
		{ 480, 184, 44 },
		{ 480, 185, 44 },
		{ 480, 186, 44 },
		{ 480, 187, 44 },
		{ 480, 188, 44 },
		{ 480, 189, 44 },
		{ 480, 190, 44 },
		{ 480, 191, 44 },
		{ 480, 192, 44 },
		{ 480, 193, 44 },
		{ 480, 194, 44 },
		{ 480, 195, 44 },
		{ 480, 196, 44 },
		{ 480, 197, 44 },
		{ 480, 198, 44 },
		{ 480, 199, 44 },
		{ 480, 200, 44 },
		{ 480, 201, 44 },
		{ 480, 202, 44 },
		{ 480, 203, 44 },
		{ 480, 204, 44 },
		{ 480, 205, 44 },
		{ 480, 206, 44 },
		{ 480, 207, 44 },
		{ 480, 208, 44 },
		{ 480, 209, 44 },
		{ 480, 210, 44 },
		{ 480, 211, 44 },
		{ 480, 212, 44 },
		{ 480, 213, 44 },
		{ 480, 214, 44 },
		{ 480, 215, 44 },
		{ 480, 216, 44 },
		{ 480, 217, 44 },
		{ 480, 218, 44 },
		{ 480, 219, 44 },
		{ 480, 220, 44 },
		{ 480, 221, 44 },
		{ 480, 222, 44 },
		{ 480, 223, 44 },
		{ 480, 224, 44 },
		{ 480, 225, 44 },
		{ 480, 226, 44 },
		{ 480, 227, 44 },
		{ 480, 228, 44 },
		{ 480, 229, 44 },
		{ 481, 174, 44 },
		{ 481, 175, 44 },
		{ 481, 176, 44 },
		{ 481, 177, 44 },
		{ 481, 178, 44 },
		{ 481, 179, 44 },
		{ 481, 180, 44 },
		{ 481, 181, 44 },
		{ 481, 182, 44 },
		{ 481, 183, 44 },
		{ 481, 184, 44 },
		{ 481, 185, 44 },
		{ 481, 186, 44 },
		{ 481, 187, 44 },
		{ 481, 188, 44 },
		{ 481, 189, 44 },
		{ 481, 190, 44 },
		{ 481, 191, 44 },
		{ 481, 192, 44 },
		{ 481, 193, 44 },
		{ 481, 194, 44 },
		{ 481, 195, 44 },
		{ 481, 196, 44 },
		{ 481, 197, 44 },
		{ 481, 198, 44 },
		{ 481, 199, 44 },
		{ 481, 200, 44 },
		{ 481, 201, 44 },
		{ 481, 202, 44 },
		{ 481, 203, 44 },
		{ 481, 204, 44 },
		{ 481, 205, 44 },
		{ 481, 206, 44 },
		{ 481, 207, 44 },
		{ 481, 208, 44 },
		{ 481, 209, 44 },
		{ 481, 210, 44 },
		{ 481, 211, 44 },
		{ 481, 212, 44 },
		{ 481, 213, 44 },
		{ 481, 214, 44 },
		{ 481, 215, 44 },
		{ 481, 216, 44 },
		{ 481, 217, 44 },
		{ 481, 218, 44 },
		{ 481, 219, 44 },
		{ 481, 220, 44 },
		{ 481, 221, 44 },
		{ 481, 222, 44 },
		{ 481, 223, 44 },
		{ 481, 224, 44 },
		{ 481, 225, 44 },
		{ 481, 226, 44 },
		{ 481, 227, 44 },
		{ 481, 228, 44 },
		{ 482, 173, 44 },
		{ 482, 174, 44 },
		{ 482, 175, 44 },
		{ 482, 176, 44 },
		{ 482, 177, 44 },
		{ 482, 178, 44 },
		{ 482, 179, 44 },
		{ 482, 180, 44 },
		{ 482, 181, 44 },
		{ 482, 182, 44 },
		{ 482, 183, 44 },
		{ 482, 184, 44 },
		{ 482, 185, 44 },
		{ 482, 186, 44 },
		{ 482, 187, 44 },
		{ 482, 188, 44 },
		{ 482, 189, 44 },
		{ 482, 190, 44 },
		{ 482, 191, 44 },
		{ 482, 192, 44 },
		{ 482, 193, 44 },
		{ 482, 194, 44 },
		{ 482, 195, 44 },
		{ 482, 196, 44 },
		{ 482, 197, 44 },
		{ 482, 198, 44 },
		{ 482, 199, 44 },
		{ 482, 200, 44 },
		{ 482, 201, 44 },
		{ 482, 202, 44 },
		{ 482, 203, 44 },
		{ 482, 204, 44 },
		{ 482, 205, 44 },
		{ 482, 206, 44 },
		{ 482, 207, 44 },
		{ 482, 208, 44 },
		{ 482, 209, 44 },
		{ 482, 210, 44 },
		{ 482, 211, 44 },
		{ 482, 212, 44 },
		{ 482, 213, 44 },
		{ 482, 214, 44 },
		{ 482, 215, 44 },
		{ 482, 216, 44 },
		{ 482, 217, 44 },
		{ 482, 218, 44 },
		{ 482, 219, 44 },
		{ 482, 220, 44 },
		{ 482, 221, 44 },
		{ 482, 222, 44 },
		{ 482, 223, 44 },
		{ 482, 224, 44 },
		{ 482, 225, 44 },
		{ 482, 226, 44 },
		{ 482, 227, 44 },
		{ 482, 228, 44 },
		{ 483, 173, 44 },
		{ 483, 174, 44 },
		{ 483, 175, 44 },
		{ 483, 176, 44 },
		{ 483, 177, 44 },
		{ 483, 178, 44 },
		{ 483, 179, 44 },
		{ 483, 180, 44 },
		{ 483, 181, 44 },
		{ 483, 182, 44 },
		{ 483, 183, 44 },
		{ 483, 184, 44 },
		{ 483, 185, 44 },
		{ 483, 186, 44 },
		{ 483, 187, 44 },
		{ 483, 188, 44 },
		{ 483, 189, 44 },
		{ 483, 190, 44 },
		{ 483, 191, 44 },
		{ 483, 192, 44 },
		{ 483, 193, 44 },
		{ 483, 194, 44 },
		{ 483, 195, 44 },
		{ 483, 196, 44 },
		{ 483, 197, 44 },
		{ 483, 198, 44 },
		{ 483, 199, 44 },
		{ 483, 200, 44 },
		{ 483, 201, 44 },
		{ 483, 202, 44 },
		{ 483, 203, 44 },
		{ 483, 204, 44 },
		{ 483, 205, 44 },
		{ 483, 206, 44 },
		{ 483, 207, 44 },
		{ 483, 208, 44 },
		{ 483, 209, 44 },
		{ 483, 210, 44 },
		{ 483, 211, 44 },
		{ 483, 212, 44 },
		{ 483, 213, 44 },
		{ 483, 214, 44 },
		{ 483, 215, 44 },
		{ 483, 216, 44 },
		{ 483, 217, 44 },
		{ 483, 218, 44 },
		{ 483, 219, 44 },
		{ 483, 220, 44 },
		{ 483, 221, 44 },
		{ 483, 222, 44 },
		{ 483, 223, 44 },
		{ 483, 224, 44 },
		{ 483, 225, 44 },
		{ 483, 226, 44 },
		{ 483, 227, 44 },
		{ 483, 228, 44 },
		{ 484, 172, 44 },
		{ 484, 173, 44 },
		{ 484, 174, 44 },
		{ 484, 175, 44 },
		{ 484, 176, 44 },
		{ 484, 177, 44 },
		{ 484, 178, 44 },
		{ 484, 179, 44 },
		{ 484, 180, 44 },
		{ 484, 181, 44 },
		{ 484, 182, 44 },
		{ 484, 183, 44 },
		{ 484, 184, 44 },
		{ 484, 185, 44 },
		{ 484, 186, 44 },
		{ 484, 187, 44 },
		{ 484, 188, 44 },
		{ 484, 189, 44 },
		{ 484, 190, 44 },
		{ 484, 191, 44 },
		{ 484, 192, 44 },
		{ 484, 193, 44 },
		{ 484, 194, 44 },
		{ 484, 195, 44 },
		{ 484, 196, 44 },
		{ 484, 197, 44 },
		{ 484, 198, 44 },
		{ 484, 199, 44 },
		{ 484, 200, 44 },
		{ 484, 201, 44 },
		{ 484, 202, 44 },
		{ 484, 203, 44 },
		{ 484, 204, 44 },
		{ 484, 205, 44 },
		{ 484, 206, 44 },
		{ 484, 207, 44 },
		{ 484, 208, 44 },
		{ 484, 209, 44 },
		{ 484, 210, 44 },
		{ 484, 211, 44 },
		{ 484, 212, 44 },
		{ 484, 213, 44 },
		{ 484, 214, 44 },
		{ 484, 215, 44 },
		{ 484, 216, 44 },
		{ 484, 217, 44 },
		{ 484, 218, 44 },
		{ 484, 219, 44 },
		{ 484, 220, 44 },
		{ 484, 221, 44 },
		{ 484, 222, 44 },
		{ 484, 223, 44 },
		{ 484, 224, 44 },
		{ 484, 225, 44 },
		{ 484, 226, 44 },
		{ 484, 227, 44 },
		{ 484, 228, 44 },
		{ 485, 172, 44 },
		{ 485, 173, 44 },
		{ 485, 174, 44 },
		{ 485, 175, 44 },
		{ 485, 176, 44 },
		{ 485, 177, 44 },
		{ 485, 178, 44 },
		{ 485, 179, 44 },
		{ 485, 180, 44 },
		{ 485, 181, 44 },
		{ 485, 182, 44 },
		{ 485, 183, 44 },
		{ 485, 184, 44 },
		{ 485, 185, 44 },
		{ 485, 186, 44 },
		{ 485, 187, 44 },
		{ 485, 188, 44 },
		{ 485, 189, 44 },
		{ 485, 190, 44 },
		{ 485, 191, 44 },
		{ 485, 192, 44 },
		{ 485, 193, 44 },
		{ 485, 194, 44 },
		{ 485, 195, 44 },
		{ 485, 196, 44 },
		{ 485, 197, 44 },
		{ 485, 198, 44 },
		{ 485, 199, 44 },
		{ 485, 200, 44 },
		{ 485, 201, 44 },
		{ 485, 202, 44 },
		{ 485, 203, 44 },
		{ 485, 204, 44 },
		{ 485, 205, 44 },
		{ 485, 206, 44 },
		{ 485, 207, 44 },
		{ 485, 208, 44 },
		{ 485, 209, 44 },
		{ 485, 210, 44 },
		{ 485, 211, 44 },
		{ 485, 212, 44 },
		{ 485, 213, 44 },
		{ 485, 214, 44 },
		{ 485, 215, 44 },
		{ 485, 216, 44 },
		{ 485, 217, 44 },
		{ 485, 218, 44 },
		{ 485, 219, 44 },
		{ 485, 220, 44 },
		{ 485, 221, 44 },
		{ 485, 222, 44 },
		{ 485, 223, 44 },
		{ 485, 224, 44 },
		{ 485, 225, 44 },
		{ 485, 226, 44 },
		{ 485, 227, 44 },
		{ 485, 228, 44 },
		{ 486, 172, 44 },
		{ 486, 173, 44 },
		{ 486, 174, 44 },
		{ 486, 175, 44 },
		{ 486, 176, 44 },
		{ 486, 177, 44 },
		{ 486, 178, 44 },
		{ 486, 179, 44 },
		{ 486, 180, 44 },
		{ 486, 181, 44 },
		{ 486, 182, 44 },
		{ 486, 183, 44 },
		{ 486, 184, 44 },
		{ 486, 185, 44 },
		{ 486, 186, 44 },
		{ 486, 187, 44 },
		{ 486, 188, 44 },
		{ 486, 189, 44 },
		{ 486, 190, 44 },
		{ 486, 191, 44 },
		{ 486, 192, 44 },
		{ 486, 193, 44 },
		{ 486, 194, 44 },
		{ 486, 195, 44 },
		{ 486, 196, 44 },
		{ 486, 197, 44 },
		{ 486, 198, 44 },
		{ 486, 199, 44 },
		{ 486, 200, 44 },
		{ 486, 201, 44 },
		{ 486, 202, 44 },
		{ 486, 203, 44 },
		{ 486, 204, 44 },
		{ 486, 205, 44 },
		{ 486, 206, 44 },
		{ 486, 207, 44 },
		{ 486, 208, 44 },
		{ 486, 209, 44 },
		{ 486, 210, 44 },
		{ 486, 211, 44 },
		{ 486, 212, 44 },
		{ 486, 213, 44 },
		{ 486, 214, 44 },
		{ 486, 215, 44 },
		{ 486, 216, 44 },
		{ 486, 217, 44 },
		{ 486, 218, 44 },
		{ 486, 219, 44 },
		{ 486, 220, 44 },
		{ 486, 221, 44 },
		{ 486, 222, 44 },
		{ 486, 223, 44 },
		{ 486, 224, 44 },
		{ 486, 225, 44 },
		{ 486, 226, 44 },
		{ 486, 227, 44 },
		{ 486, 228, 44 },
		{ 487, 171, 44 },
		{ 487, 172, 44 },
		{ 487, 173, 44 },
		{ 487, 174, 44 },
		{ 487, 175, 44 },
		{ 487, 176, 44 },
		{ 487, 177, 44 },
		{ 487, 178, 44 },
		{ 487, 179, 44 },
		{ 487, 180, 44 },
		{ 487, 181, 44 },
		{ 487, 182, 44 },
		{ 487, 183, 44 },
		{ 487, 184, 44 },
		{ 487, 185, 44 },
		{ 487, 186, 44 },
		{ 487, 187, 44 },
		{ 487, 188, 44 },
		{ 487, 189, 44 },
		{ 487, 190, 44 },
		{ 487, 191, 44 },
		{ 487, 192, 44 },
		{ 487, 193, 44 },
		{ 487, 194, 44 },
		{ 487, 195, 44 },
		{ 487, 196, 44 },
		{ 487, 197, 44 },
		{ 487, 198, 44 },
		{ 487, 199, 44 },
		{ 487, 200, 44 },
		{ 487, 201, 44 },
		{ 487, 202, 44 },
		{ 487, 203, 44 },
		{ 487, 204, 44 },
		{ 487, 205, 44 },
		{ 487, 206, 44 },
		{ 487, 207, 44 },
		{ 487, 208, 44 },
		{ 487, 209, 44 },
		{ 487, 210, 44 },
		{ 487, 211, 44 },
		{ 487, 212, 44 },
		{ 487, 213, 44 },
		{ 487, 214, 44 },
		{ 487, 215, 44 },
		{ 487, 216, 44 },
		{ 487, 217, 44 },
		{ 487, 218, 44 },
		{ 487, 219, 44 },
		{ 487, 220, 44 },
		{ 487, 221, 44 },
		{ 487, 222, 44 },
		{ 487, 223, 44 },
		{ 487, 224, 44 },
		{ 487, 225, 44 },
		{ 487, 226, 44 },
		{ 487, 227, 44 },
		{ 487, 228, 44 },
		{ 488, 171, 44 },
		{ 488, 172, 44 },
		{ 488, 173, 44 },
		{ 488, 174, 44 },
		{ 488, 175, 44 },
		{ 488, 176, 44 },
		{ 488, 177, 44 },
		{ 488, 178, 44 },
		{ 488, 179, 44 },
		{ 488, 180, 44 },
		{ 488, 181, 44 },
		{ 488, 182, 44 },
		{ 488, 183, 44 },
		{ 488, 184, 44 },
		{ 488, 185, 44 },
		{ 488, 186, 44 },
		{ 488, 187, 44 },
		{ 488, 188, 44 },
		{ 488, 189, 44 },
		{ 488, 190, 44 },
		{ 488, 191, 44 },
		{ 488, 192, 44 },
		{ 488, 193, 44 },
		{ 488, 194, 44 },
		{ 488, 195, 44 },
		{ 488, 196, 44 },
		{ 488, 197, 44 },
		{ 488, 198, 44 },
		{ 488, 199, 44 },
		{ 488, 200, 44 },
		{ 488, 201, 44 },
		{ 488, 202, 44 },
		{ 488, 203, 44 },
		{ 488, 204, 44 },
		{ 488, 205, 44 },
		{ 488, 206, 44 },
		{ 488, 207, 44 },
		{ 488, 208, 44 },
		{ 488, 209, 44 },
		{ 488, 210, 44 },
		{ 488, 211, 44 },
		{ 488, 212, 44 },
		{ 488, 213, 44 },
		{ 488, 214, 44 },
		{ 488, 215, 44 },
		{ 488, 216, 44 },
		{ 488, 217, 44 },
		{ 488, 218, 44 },
		{ 488, 219, 44 },
		{ 488, 220, 44 },
		{ 488, 221, 44 },
		{ 488, 222, 44 },
		{ 488, 223, 44 },
		{ 488, 224, 44 },
		{ 488, 225, 44 },
		{ 488, 226, 44 },
		{ 488, 227, 44 },
		{ 488, 228, 44 },
		{ 489, 171, 44 },
		{ 489, 172, 44 },
		{ 489, 173, 44 },
		{ 489, 174, 44 },
		{ 489, 175, 44 },
		{ 489, 176, 44 },
		{ 489, 177, 44 },
		{ 489, 178, 44 },
		{ 489, 179, 44 },
		{ 489, 180, 44 },
		{ 489, 181, 44 },
		{ 489, 182, 44 },
		{ 489, 183, 44 },
		{ 489, 184, 44 },
		{ 489, 185, 44 },
		{ 489, 186, 44 },
		{ 489, 187, 44 },
		{ 489, 188, 44 },
		{ 489, 189, 44 },
		{ 489, 190, 44 },
		{ 489, 191, 44 },
		{ 489, 192, 44 },
		{ 489, 193, 44 },
		{ 489, 194, 44 },
		{ 489, 195, 44 },
		{ 489, 196, 44 },
		{ 489, 197, 44 },
		{ 489, 198, 44 },
		{ 489, 199, 44 },
		{ 489, 200, 44 },
		{ 489, 201, 44 },
		{ 489, 202, 44 },
		{ 489, 203, 44 },
		{ 489, 204, 44 },
		{ 489, 205, 44 },
		{ 489, 206, 44 },
		{ 489, 207, 44 },
		{ 489, 208, 44 },
		{ 489, 209, 44 },
		{ 489, 210, 44 },
		{ 489, 211, 44 },
		{ 489, 212, 44 },
		{ 489, 213, 44 },
		{ 489, 214, 44 },
		{ 489, 215, 44 },
		{ 489, 216, 44 },
		{ 489, 217, 44 },
		{ 489, 218, 44 },
		{ 489, 219, 44 },
		{ 489, 220, 44 },
		{ 489, 221, 44 },
		{ 489, 222, 44 },
		{ 489, 223, 44 },
		{ 489, 224, 44 },
		{ 489, 225, 44 },
		{ 489, 226, 44 },
		{ 489, 227, 44 },
		{ 490, 170, 44 },
		{ 490, 171, 44 },
		{ 490, 172, 44 },
		{ 490, 173, 44 },
		{ 490, 174, 44 },
		{ 490, 175, 44 },
		{ 490, 176, 44 },
		{ 490, 177, 44 },
		{ 490, 178, 44 },
		{ 490, 179, 44 },
		{ 490, 180, 44 },
		{ 490, 181, 44 },
		{ 490, 182, 44 },
		{ 490, 183, 44 },
		{ 490, 184, 44 },
		{ 490, 185, 44 },
		{ 490, 186, 44 },
		{ 490, 187, 44 },
		{ 490, 188, 44 },
		{ 490, 189, 44 },
		{ 490, 190, 44 },
		{ 490, 191, 44 },
		{ 490, 192, 44 },
		{ 490, 193, 44 },
		{ 490, 194, 44 },
		{ 490, 195, 44 },
		{ 490, 196, 44 },
		{ 490, 197, 44 },
		{ 490, 198, 44 },
		{ 490, 199, 44 },
		{ 490, 200, 44 },
		{ 490, 201, 44 },
		{ 490, 202, 44 },
		{ 490, 203, 44 },
		{ 490, 204, 44 },
		{ 490, 205, 44 },
		{ 490, 206, 44 },
		{ 490, 207, 44 },
		{ 490, 208, 44 },
		{ 490, 209, 44 },
		{ 490, 210, 44 },
		{ 490, 211, 44 },
		{ 490, 212, 44 },
		{ 490, 213, 44 },
		{ 490, 214, 44 },
		{ 490, 215, 44 },
		{ 490, 216, 44 },
		{ 490, 217, 44 },
		{ 490, 218, 44 },
		{ 490, 219, 44 },
		{ 490, 220, 44 },
		{ 490, 221, 44 },
		{ 490, 222, 44 },
		{ 490, 223, 44 },
		{ 490, 224, 44 },
		{ 490, 225, 44 },
		{ 490, 226, 44 },
		{ 490, 227, 44 },
		{ 491, 170, 44 },
		{ 491, 171, 44 },
		{ 491, 172, 44 },
		{ 491, 173, 44 },
		{ 491, 174, 44 },
		{ 491, 175, 44 },
		{ 491, 176, 44 },
		{ 491, 177, 44 },
		{ 491, 178, 44 },
		{ 491, 179, 44 },
		{ 491, 180, 44 },
		{ 491, 181, 44 },
		{ 491, 182, 44 },
		{ 491, 183, 44 },
		{ 491, 184, 44 },
		{ 491, 185, 44 },
		{ 491, 186, 44 },
		{ 491, 187, 44 },
		{ 491, 188, 44 },
		{ 491, 189, 44 },
		{ 491, 190, 44 },
		{ 491, 191, 44 },
		{ 491, 192, 44 },
		{ 491, 193, 44 },
		{ 491, 194, 44 },
		{ 491, 195, 44 },
		{ 491, 196, 44 },
		{ 491, 197, 44 },
		{ 491, 198, 44 },
		{ 491, 199, 44 },
		{ 491, 200, 44 },
		{ 491, 201, 44 },
		{ 491, 202, 44 },
		{ 491, 203, 44 },
		{ 491, 204, 44 },
		{ 491, 205, 44 },
		{ 491, 206, 44 },
		{ 491, 207, 44 },
		{ 491, 208, 44 },
		{ 491, 209, 44 },
		{ 491, 210, 44 },
		{ 491, 211, 44 },
		{ 491, 212, 44 },
		{ 491, 213, 44 },
		{ 491, 214, 44 },
		{ 491, 215, 44 },
		{ 491, 216, 44 },
		{ 491, 217, 44 },
		{ 491, 218, 44 },
		{ 491, 219, 44 },
		{ 491, 220, 44 },
		{ 491, 221, 44 },
		{ 491, 222, 44 },
		{ 491, 223, 44 },
		{ 491, 224, 44 },
		{ 491, 225, 44 },
		{ 491, 226, 44 },
		{ 491, 227, 44 },
		{ 492, 169, 44 },
		{ 492, 170, 44 },
		{ 492, 171, 44 },
		{ 492, 172, 44 },
		{ 492, 173, 44 },
		{ 492, 174, 44 },
		{ 492, 175, 44 },
		{ 492, 176, 44 },
		{ 492, 177, 44 },
		{ 492, 178, 44 },
		{ 492, 179, 44 },
		{ 492, 180, 44 },
		{ 492, 181, 44 },
		{ 492, 182, 44 },
		{ 492, 183, 44 },
		{ 492, 184, 44 },
		{ 492, 185, 44 },
		{ 492, 186, 44 },
		{ 492, 187, 44 },
		{ 492, 188, 44 },
		{ 492, 189, 44 },
		{ 492, 190, 44 },
		{ 492, 191, 44 },
		{ 492, 192, 44 },
		{ 492, 193, 44 },
		{ 492, 194, 44 },
		{ 492, 195, 44 },
		{ 492, 196, 44 },
		{ 492, 197, 44 },
		{ 492, 198, 44 },
		{ 492, 199, 44 },
		{ 492, 200, 44 },
		{ 492, 201, 44 },
		{ 492, 202, 44 },
		{ 492, 203, 44 },
		{ 492, 204, 44 },
		{ 492, 205, 44 },
		{ 492, 206, 44 },
		{ 492, 207, 44 },
		{ 492, 208, 44 },
		{ 492, 209, 44 },
		{ 492, 210, 44 },
		{ 492, 211, 44 },
		{ 492, 212, 44 },
		{ 492, 213, 44 },
		{ 492, 214, 44 },
		{ 492, 215, 44 },
		{ 492, 216, 44 },
		{ 492, 217, 44 },
		{ 492, 218, 44 },
		{ 492, 219, 44 },
		{ 492, 220, 44 },
		{ 492, 221, 44 },
		{ 492, 222, 44 },
		{ 492, 223, 44 },
		{ 492, 224, 44 },
		{ 492, 225, 44 },
		{ 492, 226, 44 },
		{ 492, 227, 44 },
		{ 493, 169, 44 },
		{ 493, 170, 44 },
		{ 493, 171, 44 },
		{ 493, 172, 44 },
		{ 493, 173, 44 },
		{ 493, 174, 44 },
		{ 493, 175, 44 },
		{ 493, 176, 44 },
		{ 493, 177, 44 },
		{ 493, 178, 44 },
		{ 493, 179, 44 },
		{ 493, 180, 44 },
		{ 493, 181, 44 },
		{ 493, 182, 44 },
		{ 493, 183, 44 },
		{ 493, 184, 44 },
		{ 493, 185, 44 },
		{ 493, 186, 44 },
		{ 493, 187, 44 },
		{ 493, 188, 44 },
		{ 493, 189, 44 },
		{ 493, 190, 44 },
		{ 493, 191, 44 },
		{ 493, 192, 44 },
		{ 493, 193, 44 },
		{ 493, 194, 44 },
		{ 493, 195, 44 },
		{ 493, 196, 44 },
		{ 493, 197, 44 },
		{ 493, 198, 44 },
		{ 493, 199, 44 },
		{ 493, 200, 44 },
		{ 493, 201, 44 },
		{ 493, 202, 44 },
		{ 493, 203, 44 },
		{ 493, 204, 44 },
		{ 493, 205, 44 },
		{ 493, 206, 44 },
		{ 493, 207, 44 },
		{ 493, 208, 44 },
		{ 493, 209, 44 },
		{ 493, 210, 44 },
		{ 493, 211, 44 },
		{ 493, 212, 44 },
		{ 493, 213, 44 },
		{ 493, 214, 44 },
		{ 493, 215, 44 },
		{ 493, 216, 44 },
		{ 493, 217, 44 },
		{ 493, 218, 44 },
		{ 493, 219, 44 },
		{ 493, 220, 44 },
		{ 493, 221, 44 },
		{ 493, 222, 44 },
		{ 493, 223, 44 },
		{ 493, 224, 44 },
		{ 493, 225, 44 },
		{ 493, 226, 44 },
		{ 493, 227, 44 },
		{ 494, 169, 44 },
		{ 494, 170, 44 },
		{ 494, 171, 44 },
		{ 494, 172, 44 },
		{ 494, 173, 44 },
		{ 494, 174, 44 },
		{ 494, 175, 44 },
		{ 494, 176, 44 },
		{ 494, 177, 44 },
		{ 494, 178, 44 },
		{ 494, 179, 44 },
		{ 494, 180, 44 },
		{ 494, 181, 44 },
		{ 494, 182, 44 },
		{ 494, 183, 44 },
		{ 494, 184, 44 },
		{ 494, 185, 44 },
		{ 494, 186, 44 },
		{ 494, 187, 44 },
		{ 494, 188, 44 },
		{ 494, 189, 44 },
		{ 494, 190, 44 },
		{ 494, 191, 44 },
		{ 494, 192, 44 },
		{ 494, 193, 44 },
		{ 494, 194, 44 },
		{ 494, 195, 44 },
		{ 494, 196, 44 },
		{ 494, 197, 44 },
		{ 494, 198, 44 },
		{ 494, 199, 44 },
		{ 494, 200, 44 },
		{ 494, 201, 44 },
		{ 494, 202, 44 },
		{ 494, 203, 44 },
		{ 494, 204, 44 },
		{ 494, 205, 44 },
		{ 494, 206, 44 },
		{ 494, 207, 44 },
		{ 494, 208, 44 },
		{ 494, 209, 44 },
		{ 494, 210, 44 },
		{ 494, 211, 44 },
		{ 494, 212, 44 },
		{ 494, 213, 44 },
		{ 494, 214, 44 },
		{ 494, 215, 44 },
		{ 494, 216, 44 },
		{ 494, 217, 44 },
		{ 494, 218, 44 },
		{ 494, 219, 44 },
		{ 494, 220, 44 },
		{ 494, 221, 44 },
		{ 494, 222, 44 },
		{ 494, 223, 44 },
		{ 494, 224, 44 },
		{ 494, 225, 44 },
		{ 494, 226, 44 },
		{ 494, 227, 44 },
		{ 495, 168, 44 },
		{ 495, 169, 44 },
		{ 495, 170, 44 },
		{ 495, 171, 44 },
		{ 495, 172, 44 },
		{ 495, 173, 44 },
		{ 495, 174, 44 },
		{ 495, 175, 44 },
		{ 495, 176, 44 },
		{ 495, 177, 44 },
		{ 495, 178, 44 },
		{ 495, 179, 44 },
		{ 495, 180, 44 },
		{ 495, 181, 44 },
		{ 495, 182, 44 },
		{ 495, 183, 44 },
		{ 495, 184, 44 },
		{ 495, 185, 44 },
		{ 495, 186, 44 },
		{ 495, 187, 44 },
		{ 495, 188, 44 },
		{ 495, 189, 44 },
		{ 495, 190, 44 },
		{ 495, 191, 44 },
		{ 495, 192, 44 },
		{ 495, 193, 44 },
		{ 495, 194, 44 },
		{ 495, 195, 44 },
		{ 495, 196, 44 },
		{ 495, 197, 44 },
		{ 495, 198, 44 },
		{ 495, 199, 44 },
		{ 495, 200, 44 },
		{ 495, 201, 44 },
		{ 495, 202, 44 },
		{ 495, 203, 44 },
		{ 495, 204, 44 },
		{ 495, 205, 44 },
		{ 495, 206, 44 },
		{ 495, 207, 44 },
		{ 495, 208, 44 },
		{ 495, 209, 44 },
		{ 495, 210, 44 },
		{ 495, 211, 44 },
		{ 495, 212, 44 },
		{ 495, 213, 44 },
		{ 495, 214, 44 },
		{ 495, 215, 44 },
		{ 495, 216, 44 },
		{ 495, 217, 44 },
		{ 495, 218, 44 },
		{ 495, 219, 44 },
		{ 495, 220, 44 },
		{ 495, 221, 44 },
		{ 495, 222, 44 },
		{ 495, 223, 44 },
		{ 495, 224, 44 },
		{ 495, 225, 44 },
		{ 495, 226, 44 },
		{ 495, 227, 44 },
		{ 496, 168, 44 },
		{ 496, 169, 44 },
		{ 496, 170, 44 },
		{ 496, 171, 44 },
		{ 496, 172, 44 },
		{ 496, 173, 44 },
		{ 496, 174, 44 },
		{ 496, 175, 44 },
		{ 496, 176, 44 },
		{ 496, 177, 44 },
		{ 496, 178, 44 },
		{ 496, 179, 44 },
		{ 496, 180, 44 },
		{ 496, 181, 44 },
		{ 496, 182, 44 },
		{ 496, 183, 44 },
		{ 496, 184, 44 },
		{ 496, 185, 44 },
		{ 496, 186, 44 },
		{ 496, 187, 44 },
		{ 496, 188, 44 },
		{ 496, 189, 44 },
		{ 496, 190, 44 },
		{ 496, 191, 44 },
		{ 496, 192, 44 },
		{ 496, 193, 44 },
		{ 496, 194, 44 },
		{ 496, 195, 44 },
		{ 496, 196, 44 },
		{ 496, 197, 44 },
		{ 496, 198, 44 },
		{ 496, 199, 44 },
		{ 496, 200, 44 },
		{ 496, 201, 44 },
		{ 496, 202, 44 },
		{ 496, 203, 44 },
		{ 496, 204, 44 },
		{ 496, 205, 44 },
		{ 496, 206, 44 },
		{ 496, 207, 44 },
		{ 496, 208, 44 },
		{ 496, 209, 44 },
		{ 496, 210, 44 },
		{ 496, 211, 44 },
		{ 496, 212, 44 },
		{ 496, 213, 44 },
		{ 496, 214, 44 },
		{ 496, 215, 44 },
		{ 496, 216, 44 },
		{ 496, 217, 44 },
		{ 496, 218, 44 },
		{ 496, 219, 44 },
		{ 496, 220, 44 },
		{ 496, 221, 44 },
		{ 496, 222, 44 },
		{ 496, 223, 44 },
		{ 496, 224, 44 },
		{ 496, 225, 44 },
		{ 496, 226, 44 },
		{ 496, 227, 44 },
		{ 497, 168, 44 },
		{ 497, 169, 44 },
		{ 497, 170, 44 },
		{ 497, 171, 44 },
		{ 497, 172, 44 },
		{ 497, 173, 44 },
		{ 497, 174, 44 },
		{ 497, 175, 44 },
		{ 497, 176, 44 },
		{ 497, 177, 44 },
		{ 497, 178, 44 },
		{ 497, 179, 44 },
		{ 497, 180, 44 },
		{ 497, 181, 44 },
		{ 497, 182, 44 },
		{ 497, 183, 44 },
		{ 497, 184, 44 },
		{ 497, 185, 44 },
		{ 497, 186, 44 },
		{ 497, 187, 44 },
		{ 497, 188, 44 },
		{ 497, 189, 44 },
		{ 497, 190, 44 },
		{ 497, 191, 44 },
		{ 497, 192, 44 },
		{ 497, 193, 44 },
		{ 497, 194, 44 },
		{ 497, 195, 44 },
		{ 497, 196, 44 },
		{ 497, 197, 44 },
		{ 497, 198, 44 },
		{ 497, 199, 44 },
		{ 497, 200, 44 },
		{ 497, 201, 44 },
		{ 497, 202, 44 },
		{ 497, 203, 44 },
		{ 497, 204, 44 },
		{ 497, 205, 44 },
		{ 497, 206, 44 },
		{ 497, 207, 44 },
		{ 497, 208, 44 },
		{ 497, 209, 44 },
		{ 497, 210, 44 },
		{ 497, 211, 44 },
		{ 497, 212, 44 },
		{ 497, 213, 44 },
		{ 497, 214, 44 },
		{ 497, 215, 44 },
		{ 497, 216, 44 },
		{ 497, 217, 44 },
		{ 497, 218, 44 },
		{ 497, 219, 44 },
		{ 497, 220, 44 },
		{ 497, 221, 44 },
		{ 497, 222, 44 },
		{ 497, 223, 44 },
		{ 497, 224, 44 },
		{ 497, 225, 44 },
		{ 497, 226, 44 },
		{ 498, 167, 44 },
		{ 498, 168, 44 },
		{ 498, 169, 44 },
		{ 498, 170, 44 },
		{ 498, 171, 44 },
		{ 498, 172, 44 },
		{ 498, 173, 44 },
		{ 498, 174, 44 },
		{ 498, 175, 44 },
		{ 498, 176, 44 },
		{ 498, 177, 44 },
		{ 498, 178, 44 },
		{ 498, 179, 44 },
		{ 498, 180, 44 },
		{ 498, 181, 44 },
		{ 498, 182, 44 },
		{ 498, 183, 44 },
		{ 498, 184, 44 },
		{ 498, 185, 44 },
		{ 498, 186, 44 },
		{ 498, 187, 44 },
		{ 498, 188, 44 },
		{ 498, 189, 44 },
		{ 498, 190, 44 },
		{ 498, 191, 44 },
		{ 498, 192, 44 },
		{ 498, 193, 44 },
		{ 498, 194, 44 },
		{ 498, 195, 44 },
		{ 498, 196, 44 },
		{ 498, 197, 44 },
		{ 498, 198, 44 },
		{ 498, 199, 44 },
		{ 498, 200, 44 },
		{ 498, 201, 44 },
		{ 498, 202, 44 },
		{ 498, 203, 44 },
		{ 498, 204, 44 },
		{ 498, 205, 44 },
		{ 498, 206, 44 },
		{ 498, 207, 44 },
		{ 498, 208, 44 },
		{ 498, 209, 44 },
		{ 498, 210, 44 },
		{ 498, 211, 44 },
		{ 498, 212, 44 },
		{ 498, 213, 44 },
		{ 498, 214, 44 },
		{ 498, 215, 44 },
		{ 498, 216, 44 },
		{ 498, 217, 44 },
		{ 498, 218, 44 },
		{ 498, 219, 44 },
		{ 498, 220, 44 },
		{ 498, 221, 44 },
		{ 498, 222, 44 },
		{ 498, 223, 44 },
		{ 498, 224, 44 },
		{ 498, 225, 44 },
		{ 498, 226, 44 },
		{ 499, 167, 44 },
		{ 499, 168, 44 },
		{ 499, 169, 44 },
		{ 499, 170, 44 },
		{ 499, 171, 44 },
		{ 499, 172, 44 },
		{ 499, 173, 44 },
		{ 499, 174, 44 },
		{ 499, 175, 44 },
		{ 499, 176, 44 },
		{ 499, 177, 44 },
		{ 499, 178, 44 },
		{ 499, 179, 44 },
		{ 499, 180, 44 },
		{ 499, 181, 44 },
		{ 499, 182, 44 },
		{ 499, 183, 44 },
		{ 499, 184, 44 },
		{ 499, 185, 44 },
		{ 499, 186, 44 },
		{ 499, 187, 44 },
		{ 499, 188, 44 },
		{ 499, 189, 44 },
		{ 499, 190, 44 },
		{ 499, 191, 44 },
		{ 499, 192, 44 },
		{ 499, 193, 44 },
		{ 499, 194, 44 },
		{ 499, 195, 44 },
		{ 499, 196, 44 },
		{ 499, 197, 44 },
		{ 499, 198, 44 },
		{ 499, 199, 44 },
		{ 499, 200, 44 },
		{ 499, 201, 44 },
		{ 499, 202, 44 },
		{ 499, 203, 44 },
		{ 499, 204, 44 },
		{ 499, 205, 44 },
		{ 499, 206, 44 },
		{ 499, 207, 44 },
		{ 499, 208, 44 },
		{ 499, 209, 44 },
		{ 499, 210, 44 },
		{ 499, 211, 44 },
		{ 499, 212, 44 },
		{ 499, 213, 44 },
		{ 499, 214, 44 },
		{ 499, 215, 44 },
		{ 499, 216, 44 },
		{ 499, 217, 44 },
		{ 499, 218, 44 },
		{ 499, 219, 44 },
		{ 499, 220, 44 },
		{ 499, 221, 44 },
		{ 499, 222, 44 },
		{ 499, 223, 44 },
		{ 499, 224, 44 },
		{ 499, 225, 44 },
		{ 499, 226, 44 },
		{ 500, 166, 44 },
		{ 500, 167, 44 },
		{ 500, 168, 44 },
		{ 500, 169, 44 },
		{ 500, 170, 44 },
		{ 500, 171, 44 },
		{ 500, 172, 44 },
		{ 500, 173, 44 },
		{ 500, 174, 44 },
		{ 500, 175, 44 },
		{ 500, 176, 44 },
		{ 500, 177, 44 },
		{ 500, 178, 44 },
		{ 500, 179, 44 },
		{ 500, 180, 44 },
		{ 500, 181, 44 },
		{ 500, 182, 44 },
		{ 500, 183, 44 },
		{ 500, 184, 44 },
		{ 500, 185, 44 },
		{ 500, 186, 44 },
		{ 500, 187, 44 },
		{ 500, 188, 44 },
		{ 500, 189, 44 },
		{ 500, 190, 44 },
		{ 500, 191, 44 },
		{ 500, 192, 44 },
		{ 500, 193, 44 },
		{ 500, 194, 44 },
		{ 500, 195, 44 },
		{ 500, 196, 44 },
		{ 500, 197, 44 },
		{ 500, 198, 44 },
		{ 500, 199, 44 },
		{ 500, 200, 44 },
		{ 500, 201, 44 },
		{ 500, 202, 44 },
		{ 500, 203, 44 },
		{ 500, 204, 44 },
		{ 500, 205, 44 },
		{ 500, 206, 44 },
		{ 500, 207, 44 },
		{ 500, 208, 44 },
		{ 500, 209, 44 },
		{ 500, 210, 44 },
		{ 500, 211, 44 },
		{ 500, 212, 44 },
		{ 500, 213, 44 },
		{ 500, 214, 44 },
		{ 500, 215, 44 },
		{ 500, 216, 44 },
		{ 500, 217, 44 },
		{ 500, 218, 44 },
		{ 500, 219, 44 },
		{ 500, 220, 44 },
		{ 500, 221, 44 },
		{ 500, 222, 44 },
		{ 500, 223, 44 },
		{ 500, 224, 44 },
		{ 500, 225, 44 },
		{ 500, 226, 44 },
		{ 501, 166, 44 },
		{ 501, 167, 44 },
		{ 501, 168, 44 },
		{ 501, 169, 44 },
		{ 501, 170, 44 },
		{ 501, 171, 44 },
		{ 501, 172, 44 },
		{ 501, 173, 44 },
		{ 501, 174, 44 },
		{ 501, 175, 44 },
		{ 501, 176, 44 },
		{ 501, 177, 44 },
		{ 501, 178, 44 },
		{ 501, 179, 44 },
		{ 501, 180, 44 },
		{ 501, 181, 44 },
		{ 501, 182, 44 },
		{ 501, 183, 44 },
		{ 501, 184, 44 },
		{ 501, 185, 44 },
		{ 501, 186, 44 },
		{ 501, 187, 44 },
		{ 501, 188, 44 },
		{ 501, 189, 44 },
		{ 501, 190, 44 },
		{ 501, 191, 44 },
		{ 501, 192, 44 },
		{ 501, 193, 44 },
		{ 501, 194, 44 },
		{ 501, 195, 44 },
		{ 501, 196, 44 },
		{ 501, 197, 44 },
		{ 501, 198, 44 },
		{ 501, 199, 44 },
		{ 501, 200, 44 },
		{ 501, 201, 44 },
		{ 501, 202, 44 },
		{ 501, 203, 44 },
		{ 501, 204, 44 },
		{ 501, 205, 44 },
		{ 501, 206, 44 },
		{ 501, 207, 44 },
		{ 501, 208, 44 },
		{ 501, 209, 44 },
		{ 501, 210, 44 },
		{ 501, 211, 44 },
		{ 501, 212, 44 },
		{ 501, 213, 44 },
		{ 501, 214, 44 },
		{ 501, 215, 44 },
		{ 501, 216, 44 },
		{ 501, 217, 44 },
		{ 501, 218, 44 },
		{ 501, 219, 44 },
		{ 501, 220, 44 },
		{ 501, 221, 44 },
		{ 501, 222, 44 },
		{ 501, 223, 44 },
		{ 501, 224, 44 },
		{ 501, 225, 44 },
		{ 501, 226, 44 },
		{ 502, 166, 44 },
		{ 502, 167, 44 },
		{ 502, 168, 44 },
		{ 502, 169, 44 },
		{ 502, 170, 44 },
		{ 502, 171, 44 },
		{ 502, 172, 44 },
		{ 502, 173, 44 },
		{ 502, 174, 44 },
		{ 502, 175, 44 },
		{ 502, 176, 44 },
		{ 502, 177, 44 },
		{ 502, 178, 44 },
		{ 502, 179, 44 },
		{ 502, 180, 44 },
		{ 502, 181, 44 },
		{ 502, 182, 44 },
		{ 502, 183, 44 },
		{ 502, 184, 44 },
		{ 502, 185, 44 },
		{ 502, 186, 44 },
		{ 502, 187, 44 },
		{ 502, 188, 44 },
		{ 502, 189, 44 },
		{ 502, 190, 44 },
		{ 502, 191, 44 },
		{ 502, 192, 44 },
		{ 502, 193, 44 },
		{ 502, 194, 44 },
		{ 502, 195, 44 },
		{ 502, 196, 44 },
		{ 502, 197, 44 },
		{ 502, 198, 44 },
		{ 502, 199, 44 },
		{ 502, 200, 44 },
		{ 502, 201, 44 },
		{ 502, 202, 44 },
		{ 502, 203, 44 },
		{ 502, 204, 44 },
		{ 502, 205, 44 },
		{ 502, 206, 44 },
		{ 502, 207, 44 },
		{ 502, 208, 44 },
		{ 502, 209, 44 },
		{ 502, 210, 44 },
		{ 502, 211, 44 },
		{ 502, 212, 44 },
		{ 502, 213, 44 },
		{ 502, 214, 44 },
		{ 502, 215, 44 },
		{ 502, 216, 44 },
		{ 502, 217, 44 },
		{ 502, 218, 44 },
		{ 502, 219, 44 },
		{ 502, 220, 44 },
		{ 502, 221, 44 },
		{ 502, 222, 44 },
		{ 502, 223, 44 },
		{ 502, 224, 44 },
		{ 502, 225, 44 },
		{ 502, 226, 44 },
		{ 503, 165, 44 },
		{ 503, 166, 44 },
		{ 503, 167, 44 },
		{ 503, 168, 44 },
		{ 503, 169, 44 },
		{ 503, 170, 44 },
		{ 503, 171, 44 },
		{ 503, 172, 44 },
		{ 503, 173, 44 },
		{ 503, 174, 44 },
		{ 503, 175, 44 },
		{ 503, 176, 44 },
		{ 503, 177, 44 },
		{ 503, 178, 44 },
		{ 503, 179, 44 },
		{ 503, 180, 44 },
		{ 503, 181, 44 },
		{ 503, 182, 44 },
		{ 503, 183, 44 },
		{ 503, 184, 44 },
		{ 503, 185, 44 },
		{ 503, 186, 44 },
		{ 503, 187, 44 },
		{ 503, 188, 44 },
		{ 503, 189, 44 },
		{ 503, 190, 44 },
		{ 503, 191, 44 },
		{ 503, 192, 44 },
		{ 503, 193, 44 },
		{ 503, 194, 44 },
		{ 503, 195, 44 },
		{ 503, 196, 44 },
		{ 503, 197, 44 },
		{ 503, 198, 44 },
		{ 503, 199, 44 },
		{ 503, 200, 44 },
		{ 503, 201, 44 },
		{ 503, 202, 44 },
		{ 503, 203, 44 },
		{ 503, 204, 44 },
		{ 503, 205, 44 },
		{ 503, 206, 44 },
		{ 503, 207, 44 },
		{ 503, 208, 44 },
		{ 503, 209, 44 },
		{ 503, 210, 44 },
		{ 503, 211, 44 },
		{ 503, 212, 44 },
		{ 503, 213, 44 },
		{ 503, 214, 44 },
		{ 503, 215, 44 },
		{ 503, 216, 44 },
		{ 503, 217, 44 },
		{ 503, 218, 44 },
		{ 503, 219, 44 },
		{ 503, 220, 44 },
		{ 503, 221, 44 },
		{ 503, 222, 44 },
		{ 503, 223, 44 },
		{ 503, 224, 44 },
		{ 503, 225, 44 },
		{ 503, 226, 44 },
		{ 504, 165, 44 },
		{ 504, 166, 44 },
		{ 504, 167, 44 },
		{ 504, 168, 44 },
		{ 504, 169, 44 },
		{ 504, 170, 44 },
		{ 504, 171, 44 },
		{ 504, 172, 44 },
		{ 504, 173, 44 },
		{ 504, 174, 44 },
		{ 504, 175, 44 },
		{ 504, 176, 44 },
		{ 504, 177, 44 },
		{ 504, 178, 44 },
		{ 504, 179, 44 },
		{ 504, 180, 44 },
		{ 504, 181, 44 },
		{ 504, 182, 44 },
		{ 504, 183, 44 },
		{ 504, 184, 44 },
		{ 504, 185, 44 },
		{ 504, 186, 44 },
		{ 504, 187, 44 },
		{ 504, 188, 44 },
		{ 504, 189, 44 },
		{ 504, 190, 44 },
		{ 504, 191, 44 },
		{ 504, 192, 44 },
		{ 504, 193, 44 },
		{ 504, 194, 44 },
		{ 504, 195, 44 },
		{ 504, 196, 44 },
		{ 504, 197, 44 },
		{ 504, 198, 44 },
		{ 504, 199, 44 },
		{ 504, 200, 44 },
		{ 504, 201, 44 },
		{ 504, 202, 44 },
		{ 504, 203, 44 },
		{ 504, 204, 44 },
		{ 504, 205, 44 },
		{ 504, 206, 44 },
		{ 504, 207, 44 },
		{ 504, 208, 44 },
		{ 504, 209, 44 },
		{ 504, 210, 44 },
		{ 504, 211, 44 },
		{ 504, 212, 44 },
		{ 504, 213, 44 },
		{ 504, 214, 44 },
		{ 504, 215, 44 },
		{ 504, 216, 44 },
		{ 504, 217, 44 },
		{ 504, 218, 44 },
		{ 504, 219, 44 },
		{ 504, 220, 44 },
		{ 504, 221, 44 },
		{ 504, 222, 44 },
		{ 504, 223, 44 },
		{ 504, 224, 44 },
		{ 504, 225, 44 },
		{ 504, 226, 44 },
		{ 505, 165, 44 },
		{ 505, 166, 44 },
		{ 505, 167, 44 },
		{ 505, 168, 44 },
		{ 505, 169, 44 },
		{ 505, 170, 44 },
		{ 505, 171, 44 },
		{ 505, 172, 44 },
		{ 505, 173, 44 },
		{ 505, 174, 44 },
		{ 505, 175, 44 },
		{ 505, 176, 44 },
		{ 505, 177, 44 },
		{ 505, 178, 44 },
		{ 505, 179, 44 },
		{ 505, 180, 44 },
		{ 505, 181, 44 },
		{ 505, 182, 44 },
		{ 505, 183, 44 },
		{ 505, 184, 44 },
		{ 505, 185, 44 },
		{ 505, 186, 44 },
		{ 505, 187, 44 },
		{ 505, 188, 44 },
		{ 505, 189, 44 },
		{ 505, 190, 44 },
		{ 505, 191, 44 },
		{ 505, 192, 44 },
		{ 505, 193, 44 },
		{ 505, 194, 44 },
		{ 505, 195, 44 },
		{ 505, 196, 44 },
		{ 505, 197, 44 },
		{ 505, 198, 44 },
		{ 505, 199, 44 },
		{ 505, 200, 44 },
		{ 505, 201, 44 },
		{ 505, 202, 44 },
		{ 505, 203, 44 },
		{ 505, 204, 44 },
		{ 505, 205, 44 },
		{ 505, 206, 44 },
		{ 505, 207, 44 },
		{ 505, 208, 44 },
		{ 505, 209, 44 },
		{ 505, 210, 44 },
		{ 505, 211, 44 },
		{ 505, 212, 44 },
		{ 505, 213, 44 },
		{ 505, 214, 44 },
		{ 505, 215, 44 },
		{ 505, 216, 44 },
		{ 505, 217, 44 },
		{ 505, 218, 44 },
		{ 505, 219, 44 },
		{ 505, 220, 44 },
		{ 505, 221, 44 },
		{ 505, 222, 44 },
		{ 505, 223, 44 },
		{ 505, 224, 44 },
		{ 505, 225, 44 },
		{ 506, 164, 44 },
		{ 506, 165, 44 },
		{ 506, 166, 44 },
		{ 506, 167, 44 },
		{ 506, 168, 44 },
		{ 506, 169, 44 },
		{ 506, 170, 44 },
		{ 506, 171, 44 },
		{ 506, 172, 44 },
		{ 506, 173, 44 },
		{ 506, 174, 44 },
		{ 506, 175, 44 },
		{ 506, 176, 44 },
		{ 506, 177, 44 },
		{ 506, 178, 44 },
		{ 506, 179, 44 },
		{ 506, 180, 44 },
		{ 506, 181, 44 },
		{ 506, 182, 44 },
		{ 506, 183, 44 },
		{ 506, 184, 44 },
		{ 506, 185, 44 },
		{ 506, 186, 44 },
		{ 506, 187, 44 },
		{ 506, 188, 44 },
		{ 506, 189, 44 },
		{ 506, 190, 44 },
		{ 506, 191, 44 },
		{ 506, 192, 44 },
		{ 506, 193, 44 },
		{ 506, 194, 44 },
		{ 506, 195, 44 },
		{ 506, 196, 44 },
		{ 506, 197, 44 },
		{ 506, 198, 44 },
		{ 506, 199, 44 },
		{ 506, 200, 44 },
		{ 506, 201, 44 },
		{ 506, 202, 44 },
		{ 506, 203, 44 },
		{ 506, 204, 44 },
		{ 506, 205, 44 },
		{ 506, 206, 44 },
		{ 506, 207, 44 },
		{ 506, 208, 44 },
		{ 506, 209, 44 },
		{ 506, 210, 44 },
		{ 506, 211, 44 },
		{ 506, 212, 44 },
		{ 506, 213, 44 },
		{ 506, 214, 44 },
		{ 506, 215, 44 },
		{ 506, 216, 44 },
		{ 506, 217, 44 },
		{ 506, 218, 44 },
		{ 506, 219, 44 },
		{ 506, 220, 44 },
		{ 506, 221, 44 },
		{ 506, 222, 44 },
		{ 506, 223, 44 },
		{ 506, 224, 44 },
		{ 506, 225, 44 },
		{ 507, 164, 44 },
		{ 507, 165, 44 },
		{ 507, 166, 44 },
		{ 507, 167, 44 },
		{ 507, 168, 44 },
		{ 507, 169, 44 },
		{ 507, 170, 44 },
		{ 507, 171, 44 },
		{ 507, 172, 44 },
		{ 507, 173, 44 },
		{ 507, 174, 44 },
		{ 507, 175, 44 },
		{ 507, 176, 44 },
		{ 507, 177, 44 },
		{ 507, 178, 44 },
		{ 507, 179, 44 },
		{ 507, 180, 44 },
		{ 507, 181, 44 },
		{ 507, 182, 44 },
		{ 507, 183, 44 },
		{ 507, 184, 44 },
		{ 507, 185, 44 },
		{ 507, 186, 44 },
		{ 507, 187, 44 },
		{ 507, 188, 44 },
		{ 507, 189, 44 },
		{ 507, 190, 44 },
		{ 507, 191, 44 },
		{ 507, 192, 44 },
		{ 507, 193, 44 },
		{ 507, 194, 44 },
		{ 507, 195, 44 },
		{ 507, 196, 44 },
		{ 507, 197, 44 },
		{ 507, 198, 44 },
		{ 507, 199, 44 },
		{ 507, 200, 44 },
		{ 507, 201, 44 },
		{ 507, 202, 44 },
		{ 507, 203, 44 },
		{ 507, 204, 44 },
		{ 507, 205, 44 },
		{ 507, 206, 44 },
		{ 507, 207, 44 },
		{ 507, 208, 44 },
		{ 507, 209, 44 },
		{ 507, 210, 44 },
		{ 507, 211, 44 },
		{ 507, 212, 44 },
		{ 507, 213, 44 },
		{ 507, 214, 44 },
		{ 507, 215, 44 },
		{ 507, 216, 44 },
		{ 507, 217, 44 },
		{ 507, 218, 44 },
		{ 507, 219, 44 },
		{ 507, 220, 44 },
		{ 507, 221, 44 },
		{ 507, 222, 44 },
		{ 507, 223, 44 },
		{ 507, 224, 44 },
		{ 507, 225, 44 },
		{ 508, 163, 44 },
		{ 508, 164, 44 },
		{ 508, 165, 44 },
		{ 508, 166, 44 },
		{ 508, 167, 44 },
		{ 508, 168, 44 },
		{ 508, 169, 44 },
		{ 508, 170, 44 },
		{ 508, 171, 44 },
		{ 508, 172, 44 },
		{ 508, 173, 44 },
		{ 508, 174, 44 },
		{ 508, 175, 44 },
		{ 508, 176, 44 },
		{ 508, 177, 44 },
		{ 508, 178, 44 },
		{ 508, 179, 44 },
		{ 508, 180, 44 },
		{ 508, 181, 44 },
		{ 508, 182, 44 },
		{ 508, 183, 44 },
		{ 508, 184, 44 },
		{ 508, 185, 44 },
		{ 508, 186, 44 },
		{ 508, 187, 44 },
		{ 508, 188, 44 },
		{ 508, 189, 44 },
		{ 508, 190, 44 },
		{ 508, 191, 44 },
		{ 508, 192, 44 },
		{ 508, 193, 44 },
		{ 508, 194, 44 },
		{ 508, 195, 44 },
		{ 508, 196, 44 },
		{ 508, 197, 44 },
		{ 508, 198, 44 },
		{ 508, 199, 44 },
		{ 508, 200, 44 },
		{ 508, 201, 44 },
		{ 508, 202, 44 },
		{ 508, 203, 44 },
		{ 508, 204, 44 },
		{ 508, 205, 44 },
		{ 508, 206, 44 },
		{ 508, 207, 44 },
		{ 508, 208, 44 },
		{ 508, 209, 44 },
		{ 508, 210, 44 },
		{ 508, 211, 44 },
		{ 508, 212, 44 },
		{ 508, 213, 44 },
		{ 508, 214, 44 },
		{ 508, 215, 44 },
		{ 508, 216, 44 },
		{ 508, 217, 44 },
		{ 508, 218, 44 },
		{ 508, 219, 44 },
		{ 508, 220, 44 },
		{ 508, 221, 44 },
		{ 508, 222, 44 },
		{ 508, 223, 44 },
		{ 508, 224, 44 },
		{ 508, 225, 44 },
		{ 509, 163, 44 },
		{ 509, 164, 44 },
		{ 509, 165, 44 },
		{ 509, 166, 44 },
		{ 509, 167, 44 },
		{ 509, 168, 44 },
		{ 509, 169, 44 },
		{ 509, 170, 44 },
		{ 509, 171, 44 },
		{ 509, 172, 44 },
		{ 509, 173, 44 },
		{ 509, 174, 44 },
		{ 509, 175, 44 },
		{ 509, 176, 44 },
		{ 509, 177, 44 },
		{ 509, 178, 44 },
		{ 509, 179, 44 },
		{ 509, 180, 44 },
		{ 509, 181, 44 },
		{ 509, 182, 44 },
		{ 509, 183, 44 },
		{ 509, 184, 44 },
		{ 509, 185, 44 },
		{ 509, 186, 44 },
		{ 509, 187, 44 },
		{ 509, 188, 44 },
		{ 509, 189, 44 },
		{ 509, 190, 44 },
		{ 509, 191, 44 },
		{ 509, 192, 44 },
		{ 509, 193, 44 },
		{ 509, 194, 44 },
		{ 509, 195, 44 },
		{ 509, 196, 44 },
		{ 509, 197, 44 },
		{ 509, 198, 44 },
		{ 509, 199, 44 },
		{ 509, 200, 44 },
		{ 509, 201, 44 },
		{ 509, 202, 44 },
		{ 509, 203, 44 },
		{ 509, 204, 44 },
		{ 509, 205, 44 },
		{ 509, 206, 44 },
		{ 509, 207, 44 },
		{ 509, 208, 44 },
		{ 509, 209, 44 },
		{ 509, 210, 44 },
		{ 509, 211, 44 },
		{ 509, 212, 44 },
		{ 509, 213, 44 },
		{ 509, 214, 44 },
		{ 509, 215, 44 },
		{ 509, 216, 44 },
		{ 509, 217, 44 },
		{ 509, 218, 44 },
		{ 509, 219, 44 },
		{ 509, 220, 44 },
		{ 509, 221, 44 },
		{ 509, 222, 44 },
		{ 509, 223, 44 },
		{ 509, 224, 44 },
		{ 509, 225, 44 },
		{ 510, 163, 44 },
		{ 510, 164, 44 },
		{ 510, 165, 44 },
		{ 510, 166, 44 },
		{ 510, 167, 44 },
		{ 510, 168, 44 },
		{ 510, 169, 44 },
		{ 510, 170, 44 },
		{ 510, 171, 44 },
		{ 510, 172, 44 },
		{ 510, 173, 44 },
		{ 510, 174, 44 },
		{ 510, 175, 44 },
		{ 510, 176, 44 },
		{ 510, 177, 44 },
		{ 510, 178, 44 },
		{ 510, 179, 44 },
		{ 510, 180, 44 },
		{ 510, 181, 44 },
		{ 510, 182, 44 },
		{ 510, 183, 44 },
		{ 510, 184, 44 },
		{ 510, 185, 44 },
		{ 510, 186, 44 },
		{ 510, 187, 44 },
		{ 510, 188, 44 },
		{ 510, 189, 44 },
		{ 510, 190, 44 },
		{ 510, 191, 44 },
		{ 510, 192, 44 },
		{ 510, 193, 44 },
		{ 510, 194, 44 },
		{ 510, 195, 44 },
		{ 510, 196, 44 },
		{ 510, 197, 44 },
		{ 510, 198, 44 },
		{ 510, 199, 44 },
		{ 510, 200, 44 },
		{ 510, 201, 44 },
		{ 510, 202, 44 },
		{ 510, 203, 44 },
		{ 510, 204, 44 },
		{ 510, 205, 44 },
		{ 510, 206, 44 },
		{ 510, 207, 44 },
		{ 510, 208, 44 },
		{ 510, 209, 44 },
		{ 510, 210, 44 },
		{ 510, 211, 44 },
		{ 510, 212, 44 },
		{ 510, 213, 44 },
		{ 510, 214, 44 },
		{ 510, 215, 44 },
		{ 510, 216, 44 },
		{ 510, 217, 44 },
		{ 510, 218, 44 },
		{ 510, 219, 44 },
		{ 510, 220, 44 },
		{ 510, 221, 44 },
		{ 510, 222, 44 },
		{ 510, 223, 44 },
		{ 510, 224, 44 },
		{ 510, 225, 44 },
		{ 511, 162, 44 },
		{ 511, 163, 44 },
		{ 511, 164, 44 },
		{ 511, 165, 44 },
		{ 511, 166, 44 },
		{ 511, 167, 44 },
		{ 511, 168, 44 },
		{ 511, 169, 44 },
		{ 511, 170, 44 },
		{ 511, 171, 44 },
		{ 511, 172, 44 },
		{ 511, 173, 44 },
		{ 511, 174, 44 },
		{ 511, 175, 44 },
		{ 511, 176, 44 },
		{ 511, 177, 44 },
		{ 511, 178, 44 },
		{ 511, 179, 44 },
		{ 511, 180, 44 },
		{ 511, 181, 44 },
		{ 511, 182, 44 },
		{ 511, 183, 44 },
		{ 511, 184, 44 },
		{ 511, 185, 44 },
		{ 511, 186, 44 },
		{ 511, 187, 44 },
		{ 511, 188, 44 },
		{ 511, 189, 44 },
		{ 511, 190, 44 },
		{ 511, 191, 44 },
		{ 511, 192, 44 },
		{ 511, 193, 44 },
		{ 511, 194, 44 },
		{ 511, 195, 44 },
		{ 511, 196, 44 },
		{ 511, 197, 44 },
		{ 511, 198, 44 },
		{ 511, 199, 44 },
		{ 511, 200, 44 },
		{ 511, 201, 44 },
		{ 511, 202, 44 },
		{ 511, 203, 44 },
		{ 511, 204, 44 },
		{ 511, 205, 44 },
		{ 511, 206, 44 },
		{ 511, 207, 44 },
		{ 511, 208, 44 },
		{ 511, 209, 44 },
		{ 511, 210, 44 },
		{ 511, 211, 44 },
		{ 511, 212, 44 },
		{ 511, 213, 44 },
		{ 511, 214, 44 },
		{ 511, 215, 44 },
		{ 511, 216, 44 },
		{ 511, 217, 44 },
		{ 511, 218, 44 },
		{ 511, 219, 44 },
		{ 511, 220, 44 },
		{ 511, 221, 44 },
		{ 511, 222, 44 },
		{ 511, 223, 44 },
		{ 511, 224, 44 },
		{ 511, 225, 44 },
		{ 512, 162, 44 },
		{ 512, 163, 44 },
		{ 512, 164, 44 },
		{ 512, 165, 44 },
		{ 512, 166, 44 },
		{ 512, 167, 44 },
		{ 512, 168, 44 },
		{ 512, 169, 44 },
		{ 512, 170, 44 },
		{ 512, 171, 44 },
		{ 512, 172, 44 },
		{ 512, 173, 44 },
		{ 512, 174, 44 },
		{ 512, 175, 44 },
		{ 512, 176, 44 },
		{ 512, 177, 44 },
		{ 512, 178, 44 },
		{ 512, 179, 44 },
		{ 512, 180, 44 },
		{ 512, 181, 44 },
		{ 512, 182, 44 },
		{ 512, 183, 44 },
		{ 512, 184, 44 },
		{ 512, 185, 44 },
		{ 512, 186, 44 },
		{ 512, 187, 44 },
		{ 512, 188, 44 },
		{ 512, 189, 44 },
		{ 512, 190, 44 },
		{ 512, 191, 44 },
		{ 512, 192, 44 },
		{ 512, 193, 44 },
		{ 512, 194, 44 },
		{ 512, 195, 44 },
		{ 512, 196, 44 },
		{ 512, 197, 44 },
		{ 512, 198, 44 },
		{ 512, 199, 44 },
		{ 512, 200, 44 },
		{ 512, 201, 44 },
		{ 512, 202, 44 },
		{ 512, 203, 44 },
		{ 512, 204, 44 },
		{ 512, 205, 44 },
		{ 512, 206, 44 },
		{ 512, 207, 44 },
		{ 512, 208, 44 },
		{ 512, 209, 44 },
		{ 512, 210, 44 },
		{ 512, 211, 44 },
		{ 512, 212, 44 },
		{ 512, 213, 44 },
		{ 512, 214, 44 },
		{ 512, 215, 44 },
		{ 512, 216, 44 },
		{ 512, 217, 44 },
		{ 512, 218, 44 },
		{ 512, 219, 44 },
		{ 512, 220, 44 },
		{ 512, 221, 44 },
		{ 512, 222, 44 },
		{ 512, 223, 44 },
		{ 512, 224, 44 },
		{ 512, 225, 44 },
		{ 1, 258, 45 },
		{ 1, 259, 45 },
		{ 1, 260, 45 },
		{ 1, 261, 45 },
		{ 1, 262, 45 },
		{ 1, 263, 45 },
		{ 1, 264, 45 },
		{ 1, 265, 45 },
		{ 1, 266, 45 },
		{ 1, 267, 45 },
		{ 1, 268, 45 },
		{ 1, 269, 45 },
		{ 1, 270, 45 },
		{ 1, 271, 45 },
		{ 1, 272, 45 },
		{ 1, 273, 45 },
		{ 1, 274, 45 },
		{ 1, 275, 45 },
		{ 1, 276, 45 },
		{ 1, 277, 45 },
		{ 1, 278, 45 },
		{ 1, 279, 45 },
		{ 1, 280, 45 },
		{ 1, 281, 45 },
		{ 1, 282, 45 },
		{ 1, 283, 45 },
		{ 1, 284, 45 },
		{ 1, 285, 45 },
		{ 1, 286, 45 },
		{ 1, 287, 45 },
		{ 1, 288, 45 },
		{ 1, 289, 45 },
		{ 1, 290, 45 },
		{ 1, 291, 45 },
		{ 1, 292, 45 },
		{ 1, 293, 45 },
		{ 1, 294, 45 },
		{ 1, 295, 45 },
		{ 1, 296, 45 },
		{ 1, 297, 45 },
		{ 1, 298, 45 },
		{ 1, 299, 45 },
		{ 1, 300, 45 },
		{ 1, 301, 45 },
		{ 1, 302, 45 },
		{ 1, 303, 45 },
		{ 1, 304, 45 },
		{ 1, 305, 45 },
		{ 1, 306, 45 },
		{ 1, 307, 45 },
		{ 1, 308, 45 },
		{ 1, 309, 45 },
		{ 1, 310, 45 },
		{ 1, 311, 45 },
		{ 1, 312, 45 },
		{ 1, 313, 45 },
		{ 1, 314, 45 },
		{ 1, 315, 45 },
		{ 1, 316, 45 },
		{ 1, 317, 45 },
		{ 1, 318, 45 },
		{ 1, 319, 45 },
		{ 1, 320, 45 },
		{ 2, 258, 45 },
		{ 2, 259, 45 },
		{ 2, 260, 45 },
		{ 2, 261, 45 },
		{ 2, 262, 45 },
		{ 2, 263, 45 },
		{ 2, 264, 45 },
		{ 2, 265, 45 },
		{ 2, 266, 45 },
		{ 2, 267, 45 },
		{ 2, 268, 45 },
		{ 2, 269, 45 },
		{ 2, 270, 45 },
		{ 2, 271, 45 },
		{ 2, 272, 45 },
		{ 2, 273, 45 },
		{ 2, 274, 45 },
		{ 2, 275, 45 },
		{ 2, 276, 45 },
		{ 2, 277, 45 },
		{ 2, 278, 45 },
		{ 2, 279, 45 },
		{ 2, 280, 45 },
		{ 2, 281, 45 },
		{ 2, 282, 45 },
		{ 2, 283, 45 },
		{ 2, 284, 45 },
		{ 2, 285, 45 },
		{ 2, 286, 45 },
		{ 2, 287, 45 },
		{ 2, 288, 45 },
		{ 2, 289, 45 },
		{ 2, 290, 45 },
		{ 2, 291, 45 },
		{ 2, 292, 45 },
		{ 2, 293, 45 },
		{ 2, 294, 45 },
		{ 2, 295, 45 },
		{ 2, 296, 45 },
		{ 2, 297, 45 },
		{ 2, 298, 45 },
		{ 2, 299, 45 },
		{ 2, 300, 45 },
		{ 2, 301, 45 },
		{ 2, 302, 45 },
		{ 2, 303, 45 },
		{ 2, 304, 45 },
		{ 2, 305, 45 },
		{ 2, 306, 45 },
		{ 2, 307, 45 },
		{ 2, 308, 45 },
		{ 2, 309, 45 },
		{ 2, 310, 45 },
		{ 2, 311, 45 },
		{ 2, 312, 45 },
		{ 2, 313, 45 },
		{ 2, 314, 45 },
		{ 2, 315, 45 },
		{ 2, 316, 45 },
		{ 2, 317, 45 },
		{ 2, 318, 45 },
		{ 2, 319, 45 },
		{ 2, 320, 45 },
		{ 3, 258, 45 },
		{ 3, 259, 45 },
		{ 3, 260, 45 },
		{ 3, 261, 45 },
		{ 3, 262, 45 },
		{ 3, 263, 45 },
		{ 3, 264, 45 },
		{ 3, 265, 45 },
		{ 3, 266, 45 },
		{ 3, 267, 45 },
		{ 3, 268, 45 },
		{ 3, 269, 45 },
		{ 3, 270, 45 },
		{ 3, 271, 45 },
		{ 3, 272, 45 },
		{ 3, 273, 45 },
		{ 3, 274, 45 },
		{ 3, 275, 45 },
		{ 3, 276, 45 },
		{ 3, 277, 45 },
		{ 3, 278, 45 },
		{ 3, 279, 45 },
		{ 3, 280, 45 },
		{ 3, 281, 45 },
		{ 3, 282, 45 },
		{ 3, 283, 45 },
		{ 3, 284, 45 },
		{ 3, 285, 45 },
		{ 3, 286, 45 },
		{ 3, 287, 45 },
		{ 3, 288, 45 },
		{ 3, 289, 45 },
		{ 3, 290, 45 },
		{ 3, 291, 45 },
		{ 3, 292, 45 },
		{ 3, 293, 45 },
		{ 3, 294, 45 },
		{ 3, 295, 45 },
		{ 3, 296, 45 },
		{ 3, 297, 45 },
		{ 3, 298, 45 },
		{ 3, 299, 45 },
		{ 3, 300, 45 },
		{ 3, 301, 45 },
		{ 3, 302, 45 },
		{ 3, 303, 45 },
		{ 3, 304, 45 },
		{ 3, 305, 45 },
		{ 3, 306, 45 },
		{ 3, 307, 45 },
		{ 3, 308, 45 },
		{ 3, 309, 45 },
		{ 3, 310, 45 },
		{ 3, 311, 45 },
		{ 3, 312, 45 },
		{ 3, 313, 45 },
		{ 3, 314, 45 },
		{ 3, 315, 45 },
		{ 3, 316, 45 },
		{ 3, 317, 45 },
		{ 3, 318, 45 },
		{ 3, 319, 45 },
		{ 3, 320, 45 },
		{ 4, 258, 45 },
		{ 4, 259, 45 },
		{ 4, 260, 45 },
		{ 4, 261, 45 },
		{ 4, 262, 45 },
		{ 4, 263, 45 },
		{ 4, 264, 45 },
		{ 4, 265, 45 },
		{ 4, 266, 45 },
		{ 4, 267, 45 },
		{ 4, 268, 45 },
		{ 4, 269, 45 },
		{ 4, 270, 45 },
		{ 4, 271, 45 },
		{ 4, 272, 45 },
		{ 4, 273, 45 },
		{ 4, 274, 45 },
		{ 4, 275, 45 },
		{ 4, 276, 45 },
		{ 4, 277, 45 },
		{ 4, 278, 45 },
		{ 4, 279, 45 },
		{ 4, 280, 45 },
		{ 4, 281, 45 },
		{ 4, 282, 45 },
		{ 4, 283, 45 },
		{ 4, 284, 45 },
		{ 4, 285, 45 },
		{ 4, 286, 45 },
		{ 4, 287, 45 },
		{ 4, 288, 45 },
		{ 4, 289, 45 },
		{ 4, 290, 45 },
		{ 4, 291, 45 },
		{ 4, 292, 45 },
		{ 4, 293, 45 },
		{ 4, 294, 45 },
		{ 4, 295, 45 },
		{ 4, 296, 45 },
		{ 4, 297, 45 },
		{ 4, 298, 45 },
		{ 4, 299, 45 },
		{ 4, 300, 45 },
		{ 4, 301, 45 },
		{ 4, 302, 45 },
		{ 4, 303, 45 },
		{ 4, 304, 45 },
		{ 4, 305, 45 },
		{ 4, 306, 45 },
		{ 4, 307, 45 },
		{ 4, 308, 45 },
		{ 4, 309, 45 },
		{ 4, 310, 45 },
		{ 4, 311, 45 },
		{ 4, 312, 45 },
		{ 4, 313, 45 },
		{ 4, 314, 45 },
		{ 4, 315, 45 },
		{ 4, 316, 45 },
		{ 4, 317, 45 },
		{ 4, 318, 45 },
		{ 4, 319, 45 },
		{ 4, 320, 45 },
		{ 5, 258, 45 },
		{ 5, 259, 45 },
		{ 5, 260, 45 },
		{ 5, 261, 45 },
		{ 5, 262, 45 },
		{ 5, 263, 45 },
		{ 5, 264, 45 },
		{ 5, 265, 45 },
		{ 5, 266, 45 },
		{ 5, 267, 45 },
		{ 5, 268, 45 },
		{ 5, 269, 45 },
		{ 5, 270, 45 },
		{ 5, 271, 45 },
		{ 5, 272, 45 },
		{ 5, 273, 45 },
		{ 5, 274, 45 },
		{ 5, 275, 45 },
		{ 5, 276, 45 },
		{ 5, 277, 45 },
		{ 5, 278, 45 },
		{ 5, 279, 45 },
		{ 5, 280, 45 },
		{ 5, 281, 45 },
		{ 5, 282, 45 },
		{ 5, 283, 45 },
		{ 5, 284, 45 },
		{ 5, 285, 45 },
		{ 5, 286, 45 },
		{ 5, 287, 45 },
		{ 5, 288, 45 },
		{ 5, 289, 45 },
		{ 5, 290, 45 },
		{ 5, 291, 45 },
		{ 5, 292, 45 },
		{ 5, 293, 45 },
		{ 5, 294, 45 },
		{ 5, 295, 45 },
		{ 5, 296, 45 },
		{ 5, 297, 45 },
		{ 5, 298, 45 },
		{ 5, 299, 45 },
		{ 5, 300, 45 },
		{ 5, 301, 45 },
		{ 5, 302, 45 },
		{ 5, 303, 45 },
		{ 5, 304, 45 },
		{ 5, 305, 45 },
		{ 5, 306, 45 },
		{ 5, 307, 45 },
		{ 5, 308, 45 },
		{ 5, 309, 45 },
		{ 5, 310, 45 },
		{ 5, 311, 45 },
		{ 5, 312, 45 },
		{ 5, 313, 45 },
		{ 5, 314, 45 },
		{ 5, 315, 45 },
		{ 5, 316, 45 },
		{ 5, 317, 45 },
		{ 5, 318, 45 },
		{ 5, 319, 45 },
		{ 6, 258, 45 },
		{ 6, 259, 45 },
		{ 6, 260, 45 },
		{ 6, 261, 45 },
		{ 6, 262, 45 },
		{ 6, 263, 45 },
		{ 6, 264, 45 },
		{ 6, 265, 45 },
		{ 6, 266, 45 },
		{ 6, 267, 45 },
		{ 6, 268, 45 },
		{ 6, 269, 45 },
		{ 6, 270, 45 },
		{ 6, 271, 45 },
		{ 6, 272, 45 },
		{ 6, 273, 45 },
		{ 6, 274, 45 },
		{ 6, 275, 45 },
		{ 6, 276, 45 },
		{ 6, 277, 45 },
		{ 6, 278, 45 },
		{ 6, 279, 45 },
		{ 6, 280, 45 },
		{ 6, 281, 45 },
		{ 6, 282, 45 },
		{ 6, 283, 45 },
		{ 6, 284, 45 },
		{ 6, 285, 45 },
		{ 6, 286, 45 },
		{ 6, 287, 45 },
		{ 6, 288, 45 },
		{ 6, 289, 45 },
		{ 6, 290, 45 },
		{ 6, 291, 45 },
		{ 6, 292, 45 },
		{ 6, 293, 45 },
		{ 6, 294, 45 },
		{ 6, 295, 45 },
		{ 6, 296, 45 },
		{ 6, 297, 45 },
		{ 6, 298, 45 },
		{ 6, 299, 45 },
		{ 6, 300, 45 },
		{ 6, 301, 45 },
		{ 6, 302, 45 },
		{ 6, 303, 45 },
		{ 6, 304, 45 },
		{ 6, 305, 45 },
		{ 6, 306, 45 },
		{ 6, 307, 45 },
		{ 6, 308, 45 },
		{ 6, 309, 45 },
		{ 6, 310, 45 },
		{ 6, 311, 45 },
		{ 6, 312, 45 },
		{ 6, 313, 45 },
		{ 6, 314, 45 },
		{ 6, 315, 45 },
		{ 6, 316, 45 },
		{ 6, 317, 45 },
		{ 6, 318, 45 },
		{ 6, 319, 45 },
		{ 7, 258, 45 },
		{ 7, 259, 45 },
		{ 7, 260, 45 },
		{ 7, 261, 45 },
		{ 7, 262, 45 },
		{ 7, 263, 45 },
		{ 7, 264, 45 },
		{ 7, 265, 45 },
		{ 7, 266, 45 },
		{ 7, 267, 45 },
		{ 7, 268, 45 },
		{ 7, 269, 45 },
		{ 7, 270, 45 },
		{ 7, 271, 45 },
		{ 7, 272, 45 },
		{ 7, 273, 45 },
		{ 7, 274, 45 },
		{ 7, 275, 45 },
		{ 7, 276, 45 },
		{ 7, 277, 45 },
		{ 7, 278, 45 },
		{ 7, 279, 45 },
		{ 7, 280, 45 },
		{ 7, 281, 45 },
		{ 7, 282, 45 },
		{ 7, 283, 45 },
		{ 7, 284, 45 },
		{ 7, 285, 45 },
		{ 7, 286, 45 },
		{ 7, 287, 45 },
		{ 7, 288, 45 },
		{ 7, 289, 45 },
		{ 7, 290, 45 },
		{ 7, 291, 45 },
		{ 7, 292, 45 },
		{ 7, 293, 45 },
		{ 7, 294, 45 },
		{ 7, 295, 45 },
		{ 7, 296, 45 },
		{ 7, 297, 45 },
		{ 7, 298, 45 },
		{ 7, 299, 45 },
		{ 7, 300, 45 },
		{ 7, 301, 45 },
		{ 7, 302, 45 },
		{ 7, 303, 45 },
		{ 7, 304, 45 },
		{ 7, 305, 45 },
		{ 7, 306, 45 },
		{ 7, 307, 45 },
		{ 7, 308, 45 },
		{ 7, 309, 45 },
		{ 7, 310, 45 },
		{ 7, 311, 45 },
		{ 7, 312, 45 },
		{ 7, 313, 45 },
		{ 7, 314, 45 },
		{ 7, 315, 45 },
		{ 7, 316, 45 },
		{ 7, 317, 45 },
		{ 7, 318, 45 },
		{ 7, 319, 45 },
		{ 8, 258, 45 },
		{ 8, 259, 45 },
		{ 8, 260, 45 },
		{ 8, 261, 45 },
		{ 8, 262, 45 },
		{ 8, 263, 45 },
		{ 8, 264, 45 },
		{ 8, 265, 45 },
		{ 8, 266, 45 },
		{ 8, 267, 45 },
		{ 8, 268, 45 },
		{ 8, 269, 45 },
		{ 8, 270, 45 },
		{ 8, 271, 45 },
		{ 8, 272, 45 },
		{ 8, 273, 45 },
		{ 8, 274, 45 },
		{ 8, 275, 45 },
		{ 8, 276, 45 },
		{ 8, 277, 45 },
		{ 8, 278, 45 },
		{ 8, 279, 45 },
		{ 8, 280, 45 },
		{ 8, 281, 45 },
		{ 8, 282, 45 },
		{ 8, 283, 45 },
		{ 8, 284, 45 },
		{ 8, 285, 45 },
		{ 8, 286, 45 },
		{ 8, 287, 45 },
		{ 8, 288, 45 },
		{ 8, 289, 45 },
		{ 8, 290, 45 },
		{ 8, 291, 45 },
		{ 8, 292, 45 },
		{ 8, 293, 45 },
		{ 8, 294, 45 },
		{ 8, 295, 45 },
		{ 8, 296, 45 },
		{ 8, 297, 45 },
		{ 8, 298, 45 },
		{ 8, 299, 45 },
		{ 8, 300, 45 },
		{ 8, 301, 45 },
		{ 8, 302, 45 },
		{ 8, 303, 45 },
		{ 8, 304, 45 },
		{ 8, 305, 45 },
		{ 8, 306, 45 },
		{ 8, 307, 45 },
		{ 8, 308, 45 },
		{ 8, 309, 45 },
		{ 8, 310, 45 },
		{ 8, 311, 45 },
		{ 8, 312, 45 },
		{ 8, 313, 45 },
		{ 8, 314, 45 },
		{ 8, 315, 45 },
		{ 8, 316, 45 },
		{ 8, 317, 45 },
		{ 8, 318, 45 },
		{ 8, 319, 45 },
		{ 9, 258, 45 },
		{ 9, 259, 45 },
		{ 9, 260, 45 },
		{ 9, 261, 45 },
		{ 9, 262, 45 },
		{ 9, 263, 45 },
		{ 9, 264, 45 },
		{ 9, 265, 45 },
		{ 9, 266, 45 },
		{ 9, 267, 45 },
		{ 9, 268, 45 },
		{ 9, 269, 45 },
		{ 9, 270, 45 },
		{ 9, 271, 45 },
		{ 9, 272, 45 },
		{ 9, 273, 45 },
		{ 9, 274, 45 },
		{ 9, 275, 45 },
		{ 9, 276, 45 },
		{ 9, 277, 45 },
		{ 9, 278, 45 },
		{ 9, 279, 45 },
		{ 9, 280, 45 },
		{ 9, 281, 45 },
		{ 9, 282, 45 },
		{ 9, 283, 45 },
		{ 9, 284, 45 },
		{ 9, 285, 45 },
		{ 9, 286, 45 },
		{ 9, 287, 45 },
		{ 9, 288, 45 },
		{ 9, 289, 45 },
		{ 9, 290, 45 },
		{ 9, 291, 45 },
		{ 9, 292, 45 },
		{ 9, 293, 45 },
		{ 9, 294, 45 },
		{ 9, 295, 45 },
		{ 9, 296, 45 },
		{ 9, 297, 45 },
		{ 9, 298, 45 },
		{ 9, 299, 45 },
		{ 9, 300, 45 },
		{ 9, 301, 45 },
		{ 9, 302, 45 },
		{ 9, 303, 45 },
		{ 9, 304, 45 },
		{ 9, 305, 45 },
		{ 9, 306, 45 },
		{ 9, 307, 45 },
		{ 9, 308, 45 },
		{ 9, 309, 45 },
		{ 9, 310, 45 },
		{ 9, 311, 45 },
		{ 9, 312, 45 },
		{ 9, 313, 45 },
		{ 9, 314, 45 },
		{ 9, 315, 45 },
		{ 9, 316, 45 },
		{ 9, 317, 45 },
		{ 9, 318, 45 },
		{ 10, 258, 45 },
		{ 10, 259, 45 },
		{ 10, 260, 45 },
		{ 10, 261, 45 },
		{ 10, 262, 45 },
		{ 10, 263, 45 },
		{ 10, 264, 45 },
		{ 10, 265, 45 },
		{ 10, 266, 45 },
		{ 10, 267, 45 },
		{ 10, 268, 45 },
		{ 10, 269, 45 },
		{ 10, 270, 45 },
		{ 10, 271, 45 },
		{ 10, 272, 45 },
		{ 10, 273, 45 },
		{ 10, 274, 45 },
		{ 10, 275, 45 },
		{ 10, 276, 45 },
		{ 10, 277, 45 },
		{ 10, 278, 45 },
		{ 10, 279, 45 },
		{ 10, 280, 45 },
		{ 10, 281, 45 },
		{ 10, 282, 45 },
		{ 10, 283, 45 },
		{ 10, 284, 45 },
		{ 10, 285, 45 },
		{ 10, 286, 45 },
		{ 10, 287, 45 },
		{ 10, 288, 45 },
		{ 10, 289, 45 },
		{ 10, 290, 45 },
		{ 10, 291, 45 },
		{ 10, 292, 45 },
		{ 10, 293, 45 },
		{ 10, 294, 45 },
		{ 10, 295, 45 },
		{ 10, 296, 45 },
		{ 10, 297, 45 },
		{ 10, 298, 45 },
		{ 10, 299, 45 },
		{ 10, 300, 45 },
		{ 10, 301, 45 },
		{ 10, 302, 45 },
		{ 10, 303, 45 },
		{ 10, 304, 45 },
		{ 10, 305, 45 },
		{ 10, 306, 45 },
		{ 10, 307, 45 },
		{ 10, 308, 45 },
		{ 10, 309, 45 },
		{ 10, 310, 45 },
		{ 10, 311, 45 },
		{ 10, 312, 45 },
		{ 10, 313, 45 },
		{ 10, 314, 45 },
		{ 10, 315, 45 },
		{ 10, 316, 45 },
		{ 10, 317, 45 },
		{ 10, 318, 45 },
		{ 11, 258, 45 },
		{ 11, 259, 45 },
		{ 11, 260, 45 },
		{ 11, 261, 45 },
		{ 11, 262, 45 },
		{ 11, 263, 45 },
		{ 11, 264, 45 },
		{ 11, 265, 45 },
		{ 11, 266, 45 },
		{ 11, 267, 45 },
		{ 11, 268, 45 },
		{ 11, 269, 45 },
		{ 11, 270, 45 },
		{ 11, 271, 45 },
		{ 11, 272, 45 },
		{ 11, 273, 45 },
		{ 11, 274, 45 },
		{ 11, 275, 45 },
		{ 11, 276, 45 },
		{ 11, 277, 45 },
		{ 11, 278, 45 },
		{ 11, 279, 45 },
		{ 11, 280, 45 },
		{ 11, 281, 45 },
		{ 11, 282, 45 },
		{ 11, 283, 45 },
		{ 11, 284, 45 },
		{ 11, 285, 45 },
		{ 11, 286, 45 },
		{ 11, 287, 45 },
		{ 11, 288, 45 },
		{ 11, 289, 45 },
		{ 11, 290, 45 },
		{ 11, 291, 45 },
		{ 11, 292, 45 },
		{ 11, 293, 45 },
		{ 11, 294, 45 },
		{ 11, 295, 45 },
		{ 11, 296, 45 },
		{ 11, 297, 45 },
		{ 11, 298, 45 },
		{ 11, 299, 45 },
		{ 11, 300, 45 },
		{ 11, 301, 45 },
		{ 11, 302, 45 },
		{ 11, 303, 45 },
		{ 11, 304, 45 },
		{ 11, 305, 45 },
		{ 11, 306, 45 },
		{ 11, 307, 45 },
		{ 11, 308, 45 },
		{ 11, 309, 45 },
		{ 11, 310, 45 },
		{ 11, 311, 45 },
		{ 11, 312, 45 },
		{ 11, 313, 45 },
		{ 11, 314, 45 },
		{ 11, 315, 45 },
		{ 11, 316, 45 },
		{ 11, 317, 45 },
		{ 11, 318, 45 },
		{ 12, 258, 45 },
		{ 12, 259, 45 },
		{ 12, 260, 45 },
		{ 12, 261, 45 },
		{ 12, 262, 45 },
		{ 12, 263, 45 },
		{ 12, 264, 45 },
		{ 12, 265, 45 },
		{ 12, 266, 45 },
		{ 12, 267, 45 },
		{ 12, 268, 45 },
		{ 12, 269, 45 },
		{ 12, 270, 45 },
		{ 12, 271, 45 },
		{ 12, 272, 45 },
		{ 12, 273, 45 },
		{ 12, 274, 45 },
		{ 12, 275, 45 },
		{ 12, 276, 45 },
		{ 12, 277, 45 },
		{ 12, 278, 45 },
		{ 12, 279, 45 },
		{ 12, 280, 45 },
		{ 12, 281, 45 },
		{ 12, 282, 45 },
		{ 12, 283, 45 },
		{ 12, 284, 45 },
		{ 12, 285, 45 },
		{ 12, 286, 45 },
		{ 12, 287, 45 },
		{ 12, 288, 45 },
		{ 12, 289, 45 },
		{ 12, 290, 45 },
		{ 12, 291, 45 },
		{ 12, 292, 45 },
		{ 12, 293, 45 },
		{ 12, 294, 45 },
		{ 12, 295, 45 },
		{ 12, 296, 45 },
		{ 12, 297, 45 },
		{ 12, 298, 45 },
		{ 12, 299, 45 },
		{ 12, 300, 45 },
		{ 12, 301, 45 },
		{ 12, 302, 45 },
		{ 12, 303, 45 },
		{ 12, 304, 45 },
		{ 12, 305, 45 },
		{ 12, 306, 45 },
		{ 12, 307, 45 },
		{ 12, 308, 45 },
		{ 12, 309, 45 },
		{ 12, 310, 45 },
		{ 12, 311, 45 },
		{ 12, 312, 45 },
		{ 12, 313, 45 },
		{ 12, 314, 45 },
		{ 12, 315, 45 },
		{ 12, 316, 45 },
		{ 12, 317, 45 },
		{ 12, 318, 45 },
		{ 13, 258, 45 },
		{ 13, 259, 45 },
		{ 13, 260, 45 },
		{ 13, 261, 45 },
		{ 13, 262, 45 },
		{ 13, 263, 45 },
		{ 13, 264, 45 },
		{ 13, 265, 45 },
		{ 13, 266, 45 },
		{ 13, 267, 45 },
		{ 13, 268, 45 },
		{ 13, 269, 45 },
		{ 13, 270, 45 },
		{ 13, 271, 45 },
		{ 13, 272, 45 },
		{ 13, 273, 45 },
		{ 13, 274, 45 },
		{ 13, 275, 45 },
		{ 13, 276, 45 },
		{ 13, 277, 45 },
		{ 13, 278, 45 },
		{ 13, 279, 45 },
		{ 13, 280, 45 },
		{ 13, 281, 45 },
		{ 13, 282, 45 },
		{ 13, 283, 45 },
		{ 13, 284, 45 },
		{ 13, 285, 45 },
		{ 13, 286, 45 },
		{ 13, 287, 45 },
		{ 13, 288, 45 },
		{ 13, 289, 45 },
		{ 13, 290, 45 },
		{ 13, 291, 45 },
		{ 13, 292, 45 },
		{ 13, 293, 45 },
		{ 13, 294, 45 },
		{ 13, 295, 45 },
		{ 13, 296, 45 },
		{ 13, 297, 45 },
		{ 13, 298, 45 },
		{ 13, 299, 45 },
		{ 13, 300, 45 },
		{ 13, 301, 45 },
		{ 13, 302, 45 },
		{ 13, 303, 45 },
		{ 13, 304, 45 },
		{ 13, 305, 45 },
		{ 13, 306, 45 },
		{ 13, 307, 45 },
		{ 13, 308, 45 },
		{ 13, 309, 45 },
		{ 13, 310, 45 },
		{ 13, 311, 45 },
		{ 13, 312, 45 },
		{ 13, 313, 45 },
		{ 13, 314, 45 },
		{ 13, 315, 45 },
		{ 13, 316, 45 },
		{ 13, 317, 45 },
		{ 14, 258, 45 },
		{ 14, 259, 45 },
		{ 14, 260, 45 },
		{ 14, 261, 45 },
		{ 14, 262, 45 },
		{ 14, 263, 45 },
		{ 14, 264, 45 },
		{ 14, 265, 45 },
		{ 14, 266, 45 },
		{ 14, 267, 45 },
		{ 14, 268, 45 },
		{ 14, 269, 45 },
		{ 14, 270, 45 },
		{ 14, 271, 45 },
		{ 14, 272, 45 },
		{ 14, 273, 45 },
		{ 14, 274, 45 },
		{ 14, 275, 45 },
		{ 14, 276, 45 },
		{ 14, 277, 45 },
		{ 14, 278, 45 },
		{ 14, 279, 45 },
		{ 14, 280, 45 },
		{ 14, 281, 45 },
		{ 14, 282, 45 },
		{ 14, 283, 45 },
		{ 14, 284, 45 },
		{ 14, 285, 45 },
		{ 14, 286, 45 },
		{ 14, 287, 45 },
		{ 14, 288, 45 },
		{ 14, 289, 45 },
		{ 14, 290, 45 },
		{ 14, 291, 45 },
		{ 14, 292, 45 },
		{ 14, 293, 45 },
		{ 14, 294, 45 },
		{ 14, 295, 45 },
		{ 14, 296, 45 },
		{ 14, 297, 45 },
		{ 14, 298, 45 },
		{ 14, 299, 45 },
		{ 14, 300, 45 },
		{ 14, 301, 45 },
		{ 14, 302, 45 },
		{ 14, 303, 45 },
		{ 14, 304, 45 },
		{ 14, 305, 45 },
		{ 14, 306, 45 },
		{ 14, 307, 45 },
		{ 14, 308, 45 },
		{ 14, 309, 45 },
		{ 14, 310, 45 },
		{ 14, 311, 45 },
		{ 14, 312, 45 },
		{ 14, 313, 45 },
		{ 14, 314, 45 },
		{ 14, 315, 45 },
		{ 14, 316, 45 },
		{ 14, 317, 45 },
		{ 15, 258, 45 },
		{ 15, 259, 45 },
		{ 15, 260, 45 },
		{ 15, 261, 45 },
		{ 15, 262, 45 },
		{ 15, 263, 45 },
		{ 15, 264, 45 },
		{ 15, 265, 45 },
		{ 15, 266, 45 },
		{ 15, 267, 45 },
		{ 15, 268, 45 },
		{ 15, 269, 45 },
		{ 15, 270, 45 },
		{ 15, 271, 45 },
		{ 15, 272, 45 },
		{ 15, 273, 45 },
		{ 15, 274, 45 },
		{ 15, 275, 45 },
		{ 15, 276, 45 },
		{ 15, 277, 45 },
		{ 15, 278, 45 },
		{ 15, 279, 45 },
		{ 15, 280, 45 },
		{ 15, 281, 45 },
		{ 15, 282, 45 },
		{ 15, 283, 45 },
		{ 15, 284, 45 },
		{ 15, 285, 45 },
		{ 15, 286, 45 },
		{ 15, 287, 45 },
		{ 15, 288, 45 },
		{ 15, 289, 45 },
		{ 15, 290, 45 },
		{ 15, 291, 45 },
		{ 15, 292, 45 },
		{ 15, 293, 45 },
		{ 15, 294, 45 },
		{ 15, 295, 45 },
		{ 15, 296, 45 },
		{ 15, 297, 45 },
		{ 15, 298, 45 },
		{ 15, 299, 45 },
		{ 15, 300, 45 },
		{ 15, 301, 45 },
		{ 15, 302, 45 },
		{ 15, 303, 45 },
		{ 15, 304, 45 },
		{ 15, 305, 45 },
		{ 15, 306, 45 },
		{ 15, 307, 45 },
		{ 15, 308, 45 },
		{ 15, 309, 45 },
		{ 15, 310, 45 },
		{ 15, 311, 45 },
		{ 15, 312, 45 },
		{ 15, 313, 45 },
		{ 15, 314, 45 },
		{ 15, 315, 45 },
		{ 15, 316, 45 },
		{ 15, 317, 45 },
		{ 16, 258, 45 },
		{ 16, 259, 45 },
		{ 16, 260, 45 },
		{ 16, 261, 45 },
		{ 16, 262, 45 },
		{ 16, 263, 45 },
		{ 16, 264, 45 },
		{ 16, 265, 45 },
		{ 16, 266, 45 },
		{ 16, 267, 45 },
		{ 16, 268, 45 },
		{ 16, 269, 45 },
		{ 16, 270, 45 },
		{ 16, 271, 45 },
		{ 16, 272, 45 },
		{ 16, 273, 45 },
		{ 16, 274, 45 },
		{ 16, 275, 45 },
		{ 16, 276, 45 },
		{ 16, 277, 45 },
		{ 16, 278, 45 },
		{ 16, 279, 45 },
		{ 16, 280, 45 },
		{ 16, 281, 45 },
		{ 16, 282, 45 },
		{ 16, 283, 45 },
		{ 16, 284, 45 },
		{ 16, 285, 45 },
		{ 16, 286, 45 },
		{ 16, 287, 45 },
		{ 16, 288, 45 },
		{ 16, 289, 45 },
		{ 16, 290, 45 },
		{ 16, 291, 45 },
		{ 16, 292, 45 },
		{ 16, 293, 45 },
		{ 16, 294, 45 },
		{ 16, 295, 45 },
		{ 16, 296, 45 },
		{ 16, 297, 45 },
		{ 16, 298, 45 },
		{ 16, 299, 45 },
		{ 16, 300, 45 },
		{ 16, 301, 45 },
		{ 16, 302, 45 },
		{ 16, 303, 45 },
		{ 16, 304, 45 },
		{ 16, 305, 45 },
		{ 16, 306, 45 },
		{ 16, 307, 45 },
		{ 16, 308, 45 },
		{ 16, 309, 45 },
		{ 16, 310, 45 },
		{ 16, 311, 45 },
		{ 16, 312, 45 },
		{ 16, 313, 45 },
		{ 16, 314, 45 },
		{ 16, 315, 45 },
		{ 16, 316, 45 },
		{ 16, 317, 45 },
		{ 17, 258, 45 },
		{ 17, 259, 45 },
		{ 17, 260, 45 },
		{ 17, 261, 45 },
		{ 17, 262, 45 },
		{ 17, 263, 45 },
		{ 17, 264, 45 },
		{ 17, 265, 45 },
		{ 17, 266, 45 },
		{ 17, 267, 45 },
		{ 17, 268, 45 },
		{ 17, 269, 45 },
		{ 17, 270, 45 },
		{ 17, 271, 45 },
		{ 17, 272, 45 },
		{ 17, 273, 45 },
		{ 17, 274, 45 },
		{ 17, 275, 45 },
		{ 17, 276, 45 },
		{ 17, 277, 45 },
		{ 17, 278, 45 },
		{ 17, 279, 45 },
		{ 17, 280, 45 },
		{ 17, 281, 45 },
		{ 17, 282, 45 },
		{ 17, 283, 45 },
		{ 17, 284, 45 },
		{ 17, 285, 45 },
		{ 17, 286, 45 },
		{ 17, 287, 45 },
		{ 17, 288, 45 },
		{ 17, 289, 45 },
		{ 17, 290, 45 },
		{ 17, 291, 45 },
		{ 17, 292, 45 },
		{ 17, 293, 45 },
		{ 17, 294, 45 },
		{ 17, 295, 45 },
		{ 17, 296, 45 },
		{ 17, 297, 45 },
		{ 17, 298, 45 },
		{ 17, 299, 45 },
		{ 17, 300, 45 },
		{ 17, 301, 45 },
		{ 17, 302, 45 },
		{ 17, 303, 45 },
		{ 17, 304, 45 },
		{ 17, 305, 45 },
		{ 17, 306, 45 },
		{ 17, 307, 45 },
		{ 17, 308, 45 },
		{ 17, 309, 45 },
		{ 17, 310, 45 },
		{ 17, 311, 45 },
		{ 17, 312, 45 },
		{ 17, 313, 45 },
		{ 17, 314, 45 },
		{ 17, 315, 45 },
		{ 17, 316, 45 },
		{ 18, 258, 45 },
		{ 18, 259, 45 },
		{ 18, 260, 45 },
		{ 18, 261, 45 },
		{ 18, 262, 45 },
		{ 18, 263, 45 },
		{ 18, 264, 45 },
		{ 18, 265, 45 },
		{ 18, 266, 45 },
		{ 18, 267, 45 },
		{ 18, 268, 45 },
		{ 18, 269, 45 },
		{ 18, 270, 45 },
		{ 18, 271, 45 },
		{ 18, 272, 45 },
		{ 18, 273, 45 },
		{ 18, 274, 45 },
		{ 18, 275, 45 },
		{ 18, 276, 45 },
		{ 18, 277, 45 },
		{ 18, 278, 45 },
		{ 18, 279, 45 },
		{ 18, 280, 45 },
		{ 18, 281, 45 },
		{ 18, 282, 45 },
		{ 18, 283, 45 },
		{ 18, 284, 45 },
		{ 18, 285, 45 },
		{ 18, 286, 45 },
		{ 18, 287, 45 },
		{ 18, 288, 45 },
		{ 18, 289, 45 },
		{ 18, 290, 45 },
		{ 18, 291, 45 },
		{ 18, 292, 45 },
		{ 18, 293, 45 },
		{ 18, 294, 45 },
		{ 18, 295, 45 },
		{ 18, 296, 45 },
		{ 18, 297, 45 },
		{ 18, 298, 45 },
		{ 18, 299, 45 },
		{ 18, 300, 45 },
		{ 18, 301, 45 },
		{ 18, 302, 45 },
		{ 18, 303, 45 },
		{ 18, 304, 45 },
		{ 18, 305, 45 },
		{ 18, 306, 45 },
		{ 18, 307, 45 },
		{ 18, 308, 45 },
		{ 18, 309, 45 },
		{ 18, 310, 45 },
		{ 18, 311, 45 },
		{ 18, 312, 45 },
		{ 18, 313, 45 },
		{ 18, 314, 45 },
		{ 18, 315, 45 },
		{ 18, 316, 45 },
		{ 19, 258, 45 },
		{ 19, 259, 45 },
		{ 19, 260, 45 },
		{ 19, 261, 45 },
		{ 19, 262, 45 },
		{ 19, 263, 45 },
		{ 19, 264, 45 },
		{ 19, 265, 45 },
		{ 19, 266, 45 },
		{ 19, 267, 45 },
		{ 19, 268, 45 },
		{ 19, 269, 45 },
		{ 19, 270, 45 },
		{ 19, 271, 45 },
		{ 19, 272, 45 },
		{ 19, 273, 45 },
		{ 19, 274, 45 },
		{ 19, 275, 45 },
		{ 19, 276, 45 },
		{ 19, 277, 45 },
		{ 19, 278, 45 },
		{ 19, 279, 45 },
		{ 19, 280, 45 },
		{ 19, 281, 45 },
		{ 19, 282, 45 },
		{ 19, 283, 45 },
		{ 19, 284, 45 },
		{ 19, 285, 45 },
		{ 19, 286, 45 },
		{ 19, 287, 45 },
		{ 19, 288, 45 },
		{ 19, 289, 45 },
		{ 19, 290, 45 },
		{ 19, 291, 45 },
		{ 19, 292, 45 },
		{ 19, 293, 45 },
		{ 19, 294, 45 },
		{ 19, 295, 45 },
		{ 19, 296, 45 },
		{ 19, 297, 45 },
		{ 19, 298, 45 },
		{ 19, 299, 45 },
		{ 19, 300, 45 },
		{ 19, 301, 45 },
		{ 19, 302, 45 },
		{ 19, 303, 45 },
		{ 19, 304, 45 },
		{ 19, 305, 45 },
		{ 19, 306, 45 },
		{ 19, 307, 45 },
		{ 19, 308, 45 },
		{ 19, 309, 45 },
		{ 19, 310, 45 },
		{ 19, 311, 45 },
		{ 19, 312, 45 },
		{ 19, 313, 45 },
		{ 19, 314, 45 },
		{ 19, 315, 45 },
		{ 19, 316, 45 },
		{ 20, 258, 45 },
		{ 20, 259, 45 },
		{ 20, 260, 45 },
		{ 20, 261, 45 },
		{ 20, 262, 45 },
		{ 20, 263, 45 },
		{ 20, 264, 45 },
		{ 20, 265, 45 },
		{ 20, 266, 45 },
		{ 20, 267, 45 },
		{ 20, 268, 45 },
		{ 20, 269, 45 },
		{ 20, 270, 45 },
		{ 20, 271, 45 },
		{ 20, 272, 45 },
		{ 20, 273, 45 },
		{ 20, 274, 45 },
		{ 20, 275, 45 },
		{ 20, 276, 45 },
		{ 20, 277, 45 },
		{ 20, 278, 45 },
		{ 20, 279, 45 },
		{ 20, 280, 45 },
		{ 20, 281, 45 },
		{ 20, 282, 45 },
		{ 20, 283, 45 },
		{ 20, 284, 45 },
		{ 20, 285, 45 },
		{ 20, 286, 45 },
		{ 20, 287, 45 },
		{ 20, 288, 45 },
		{ 20, 289, 45 },
		{ 20, 290, 45 },
		{ 20, 291, 45 },
		{ 20, 292, 45 },
		{ 20, 293, 45 },
		{ 20, 294, 45 },
		{ 20, 295, 45 },
		{ 20, 296, 45 },
		{ 20, 297, 45 },
		{ 20, 298, 45 },
		{ 20, 299, 45 },
		{ 20, 300, 45 },
		{ 20, 301, 45 },
		{ 20, 302, 45 },
		{ 20, 303, 45 },
		{ 20, 304, 45 },
		{ 20, 305, 45 },
		{ 20, 306, 45 },
		{ 20, 307, 45 },
		{ 20, 308, 45 },
		{ 20, 309, 45 },
		{ 20, 310, 45 },
		{ 20, 311, 45 },
		{ 20, 312, 45 },
		{ 20, 313, 45 },
		{ 20, 314, 45 },
		{ 20, 315, 45 },
		{ 20, 316, 45 },
		{ 21, 258, 45 },
		{ 21, 259, 45 },
		{ 21, 260, 45 },
		{ 21, 261, 45 },
		{ 21, 262, 45 },
		{ 21, 263, 45 },
		{ 21, 264, 45 },
		{ 21, 265, 45 },
		{ 21, 266, 45 },
		{ 21, 267, 45 },
		{ 21, 268, 45 },
		{ 21, 269, 45 },
		{ 21, 270, 45 },
		{ 21, 271, 45 },
		{ 21, 272, 45 },
		{ 21, 273, 45 },
		{ 21, 274, 45 },
		{ 21, 275, 45 },
		{ 21, 276, 45 },
		{ 21, 277, 45 },
		{ 21, 278, 45 },
		{ 21, 279, 45 },
		{ 21, 280, 45 },
		{ 21, 281, 45 },
		{ 21, 282, 45 },
		{ 21, 283, 45 },
		{ 21, 284, 45 },
		{ 21, 285, 45 },
		{ 21, 286, 45 },
		{ 21, 287, 45 },
		{ 21, 288, 45 },
		{ 21, 289, 45 },
		{ 21, 290, 45 },
		{ 21, 291, 45 },
		{ 21, 292, 45 },
		{ 21, 293, 45 },
		{ 21, 294, 45 },
		{ 21, 295, 45 },
		{ 21, 296, 45 },
		{ 21, 297, 45 },
		{ 21, 298, 45 },
		{ 21, 299, 45 },
		{ 21, 300, 45 },
		{ 21, 301, 45 },
		{ 21, 302, 45 },
		{ 21, 303, 45 },
		{ 21, 304, 45 },
		{ 21, 305, 45 },
		{ 21, 306, 45 },
		{ 21, 307, 45 },
		{ 21, 308, 45 },
		{ 21, 309, 45 },
		{ 21, 310, 45 },
		{ 21, 311, 45 },
		{ 21, 312, 45 },
		{ 21, 313, 45 },
		{ 21, 314, 45 },
		{ 21, 315, 45 },
		{ 22, 258, 45 },
		{ 22, 259, 45 },
		{ 22, 260, 45 },
		{ 22, 261, 45 },
		{ 22, 262, 45 },
		{ 22, 263, 45 },
		{ 22, 264, 45 },
		{ 22, 265, 45 },
		{ 22, 266, 45 },
		{ 22, 267, 45 },
		{ 22, 268, 45 },
		{ 22, 269, 45 },
		{ 22, 270, 45 },
		{ 22, 271, 45 },
		{ 22, 272, 45 },
		{ 22, 273, 45 },
		{ 22, 274, 45 },
		{ 22, 275, 45 },
		{ 22, 276, 45 },
		{ 22, 277, 45 },
		{ 22, 278, 45 },
		{ 22, 279, 45 },
		{ 22, 280, 45 },
		{ 22, 281, 45 },
		{ 22, 282, 45 },
		{ 22, 283, 45 },
		{ 22, 284, 45 },
		{ 22, 285, 45 },
		{ 22, 286, 45 },
		{ 22, 287, 45 },
		{ 22, 288, 45 },
		{ 22, 289, 45 },
		{ 22, 290, 45 },
		{ 22, 291, 45 },
		{ 22, 292, 45 },
		{ 22, 293, 45 },
		{ 22, 294, 45 },
		{ 22, 295, 45 },
		{ 22, 296, 45 },
		{ 22, 297, 45 },
		{ 22, 298, 45 },
		{ 22, 299, 45 },
		{ 22, 300, 45 },
		{ 22, 301, 45 },
		{ 22, 302, 45 },
		{ 22, 303, 45 },
		{ 22, 304, 45 },
		{ 22, 305, 45 },
		{ 22, 306, 45 },
		{ 22, 307, 45 },
		{ 22, 308, 45 },
		{ 22, 309, 45 },
		{ 22, 310, 45 },
		{ 22, 311, 45 },
		{ 22, 312, 45 },
		{ 22, 313, 45 },
		{ 22, 314, 45 },
		{ 22, 315, 45 },
		{ 23, 258, 45 },
		{ 23, 259, 45 },
		{ 23, 260, 45 },
		{ 23, 261, 45 },
		{ 23, 262, 45 },
		{ 23, 263, 45 },
		{ 23, 264, 45 },
		{ 23, 265, 45 },
		{ 23, 266, 45 },
		{ 23, 267, 45 },
		{ 23, 268, 45 },
		{ 23, 269, 45 },
		{ 23, 270, 45 },
		{ 23, 271, 45 },
		{ 23, 272, 45 },
		{ 23, 273, 45 },
		{ 23, 274, 45 },
		{ 23, 275, 45 },
		{ 23, 276, 45 },
		{ 23, 277, 45 },
		{ 23, 278, 45 },
		{ 23, 279, 45 },
		{ 23, 280, 45 },
		{ 23, 281, 45 },
		{ 23, 282, 45 },
		{ 23, 283, 45 },
		{ 23, 284, 45 },
		{ 23, 285, 45 },
		{ 23, 286, 45 },
		{ 23, 287, 45 },
		{ 23, 288, 45 },
		{ 23, 289, 45 },
		{ 23, 290, 45 },
		{ 23, 291, 45 },
		{ 23, 292, 45 },
		{ 23, 293, 45 },
		{ 23, 294, 45 },
		{ 23, 295, 45 },
		{ 23, 296, 45 },
		{ 23, 297, 45 },
		{ 23, 298, 45 },
		{ 23, 299, 45 },
		{ 23, 300, 45 },
		{ 23, 301, 45 },
		{ 23, 302, 45 },
		{ 23, 303, 45 },
		{ 23, 304, 45 },
		{ 23, 305, 45 },
		{ 23, 306, 45 },
		{ 23, 307, 45 },
		{ 23, 308, 45 },
		{ 23, 309, 45 },
		{ 23, 310, 45 },
		{ 23, 311, 45 },
		{ 23, 312, 45 },
		{ 23, 313, 45 },
		{ 23, 314, 45 },
		{ 23, 315, 45 },
		{ 24, 258, 45 },
		{ 24, 259, 45 },
		{ 24, 260, 45 },
		{ 24, 261, 45 },
		{ 24, 262, 45 },
		{ 24, 263, 45 },
		{ 24, 264, 45 },
		{ 24, 265, 45 },
		{ 24, 266, 45 },
		{ 24, 267, 45 },
		{ 24, 268, 45 },
		{ 24, 269, 45 },
		{ 24, 270, 45 },
		{ 24, 271, 45 },
		{ 24, 272, 45 },
		{ 24, 273, 45 },
		{ 24, 274, 45 },
		{ 24, 275, 45 },
		{ 24, 276, 45 },
		{ 24, 277, 45 },
		{ 24, 278, 45 },
		{ 24, 279, 45 },
		{ 24, 280, 45 },
		{ 24, 281, 45 },
		{ 24, 282, 45 },
		{ 24, 283, 45 },
		{ 24, 284, 45 },
		{ 24, 285, 45 },
		{ 24, 286, 45 },
		{ 24, 287, 45 },
		{ 24, 288, 45 },
		{ 24, 289, 45 },
		{ 24, 290, 45 },
		{ 24, 291, 45 },
		{ 24, 292, 45 },
		{ 24, 293, 45 },
		{ 24, 294, 45 },
		{ 24, 295, 45 },
		{ 24, 296, 45 },
		{ 24, 297, 45 },
		{ 24, 298, 45 },
		{ 24, 299, 45 },
		{ 24, 300, 45 },
		{ 24, 301, 45 },
		{ 24, 302, 45 },
		{ 24, 303, 45 },
		{ 24, 304, 45 },
		{ 24, 305, 45 },
		{ 24, 306, 45 },
		{ 24, 307, 45 },
		{ 24, 308, 45 },
		{ 24, 309, 45 },
		{ 24, 310, 45 },
		{ 24, 311, 45 },
		{ 24, 312, 45 },
		{ 24, 313, 45 },
		{ 24, 314, 45 },
		{ 24, 315, 45 },
		{ 25, 258, 45 },
		{ 25, 259, 45 },
		{ 25, 260, 45 },
		{ 25, 261, 45 },
		{ 25, 262, 45 },
		{ 25, 263, 45 },
		{ 25, 264, 45 },
		{ 25, 265, 45 },
		{ 25, 266, 45 },
		{ 25, 267, 45 },
		{ 25, 268, 45 },
		{ 25, 269, 45 },
		{ 25, 270, 45 },
		{ 25, 271, 45 },
		{ 25, 272, 45 },
		{ 25, 273, 45 },
		{ 25, 274, 45 },
		{ 25, 275, 45 },
		{ 25, 276, 45 },
		{ 25, 277, 45 },
		{ 25, 278, 45 },
		{ 25, 279, 45 },
		{ 25, 280, 45 },
		{ 25, 281, 45 },
		{ 25, 282, 45 },
		{ 25, 283, 45 },
		{ 25, 284, 45 },
		{ 25, 285, 45 },
		{ 25, 286, 45 },
		{ 25, 287, 45 },
		{ 25, 288, 45 },
		{ 25, 289, 45 },
		{ 25, 290, 45 },
		{ 25, 291, 45 },
		{ 25, 292, 45 },
		{ 25, 293, 45 },
		{ 25, 294, 45 },
		{ 25, 295, 45 },
		{ 25, 296, 45 },
		{ 25, 297, 45 },
		{ 25, 298, 45 },
		{ 25, 299, 45 },
		{ 25, 300, 45 },
		{ 25, 301, 45 },
		{ 25, 302, 45 },
		{ 25, 303, 45 },
		{ 25, 304, 45 },
		{ 25, 305, 45 },
		{ 25, 306, 45 },
		{ 25, 307, 45 },
		{ 25, 308, 45 },
		{ 25, 309, 45 },
		{ 25, 310, 45 },
		{ 25, 311, 45 },
		{ 25, 312, 45 },
		{ 25, 313, 45 },
		{ 25, 314, 45 },
		{ 26, 258, 45 },
		{ 26, 259, 45 },
		{ 26, 260, 45 },
		{ 26, 261, 45 },
		{ 26, 262, 45 },
		{ 26, 263, 45 },
		{ 26, 264, 45 },
		{ 26, 265, 45 },
		{ 26, 266, 45 },
		{ 26, 267, 45 },
		{ 26, 268, 45 },
		{ 26, 269, 45 },
		{ 26, 270, 45 },
		{ 26, 271, 45 },
		{ 26, 272, 45 },
		{ 26, 273, 45 },
		{ 26, 274, 45 },
		{ 26, 275, 45 },
		{ 26, 276, 45 },
		{ 26, 277, 45 },
		{ 26, 278, 45 },
		{ 26, 279, 45 },
		{ 26, 280, 45 },
		{ 26, 281, 45 },
		{ 26, 282, 45 },
		{ 26, 283, 45 },
		{ 26, 284, 45 },
		{ 26, 285, 45 },
		{ 26, 286, 45 },
		{ 26, 287, 45 },
		{ 26, 288, 45 },
		{ 26, 289, 45 },
		{ 26, 290, 45 },
		{ 26, 291, 45 },
		{ 26, 292, 45 },
		{ 26, 293, 45 },
		{ 26, 294, 45 },
		{ 26, 295, 45 },
		{ 26, 296, 45 },
		{ 26, 297, 45 },
		{ 26, 298, 45 },
		{ 26, 299, 45 },
		{ 26, 300, 45 },
		{ 26, 301, 45 },
		{ 26, 302, 45 },
		{ 26, 303, 45 },
		{ 26, 304, 45 },
		{ 26, 305, 45 },
		{ 26, 306, 45 },
		{ 26, 307, 45 },
		{ 26, 308, 45 },
		{ 26, 309, 45 },
		{ 26, 310, 45 },
		{ 26, 311, 45 },
		{ 26, 312, 45 },
		{ 26, 313, 45 },
		{ 26, 314, 45 },
		{ 27, 258, 45 },
		{ 27, 259, 45 },
		{ 27, 260, 45 },
		{ 27, 261, 45 },
		{ 27, 262, 45 },
		{ 27, 263, 45 },
		{ 27, 264, 45 },
		{ 27, 265, 45 },
		{ 27, 266, 45 },
		{ 27, 267, 45 },
		{ 27, 268, 45 },
		{ 27, 269, 45 },
		{ 27, 270, 45 },
		{ 27, 271, 45 },
		{ 27, 272, 45 },
		{ 27, 273, 45 },
		{ 27, 274, 45 },
		{ 27, 275, 45 },
		{ 27, 276, 45 },
		{ 27, 277, 45 },
		{ 27, 278, 45 },
		{ 27, 279, 45 },
		{ 27, 280, 45 },
		{ 27, 281, 45 },
		{ 27, 282, 45 },
		{ 27, 283, 45 },
		{ 27, 284, 45 },
		{ 27, 285, 45 },
		{ 27, 286, 45 },
		{ 27, 287, 45 },
		{ 27, 288, 45 },
		{ 27, 289, 45 },
		{ 27, 290, 45 },
		{ 27, 291, 45 },
		{ 27, 292, 45 },
		{ 27, 293, 45 },
		{ 27, 294, 45 },
		{ 27, 295, 45 },
		{ 27, 296, 45 },
		{ 27, 297, 45 },
		{ 27, 298, 45 },
		{ 27, 299, 45 },
		{ 27, 300, 45 },
		{ 27, 301, 45 },
		{ 27, 302, 45 },
		{ 27, 303, 45 },
		{ 27, 304, 45 },
		{ 27, 305, 45 },
		{ 27, 306, 45 },
		{ 27, 307, 45 },
		{ 27, 308, 45 },
		{ 27, 309, 45 },
		{ 27, 310, 45 },
		{ 27, 311, 45 },
		{ 27, 312, 45 },
		{ 27, 313, 45 },
		{ 27, 314, 45 },
		{ 28, 258, 45 },
		{ 28, 259, 45 },
		{ 28, 260, 45 },
		{ 28, 261, 45 },
		{ 28, 262, 45 },
		{ 28, 263, 45 },
		{ 28, 264, 45 },
		{ 28, 265, 45 },
		{ 28, 266, 45 },
		{ 28, 267, 45 },
		{ 28, 268, 45 },
		{ 28, 269, 45 },
		{ 28, 270, 45 },
		{ 28, 271, 45 },
		{ 28, 272, 45 },
		{ 28, 273, 45 },
		{ 28, 274, 45 },
		{ 28, 275, 45 },
		{ 28, 276, 45 },
		{ 28, 277, 45 },
		{ 28, 278, 45 },
		{ 28, 279, 45 },
		{ 28, 280, 45 },
		{ 28, 281, 45 },
		{ 28, 282, 45 },
		{ 28, 283, 45 },
		{ 28, 284, 45 },
		{ 28, 285, 45 },
		{ 28, 286, 45 },
		{ 28, 287, 45 },
		{ 28, 288, 45 },
		{ 28, 289, 45 },
		{ 28, 290, 45 },
		{ 28, 291, 45 },
		{ 28, 292, 45 },
		{ 28, 293, 45 },
		{ 28, 294, 45 },
		{ 28, 295, 45 },
		{ 28, 296, 45 },
		{ 28, 297, 45 },
		{ 28, 298, 45 },
		{ 28, 299, 45 },
		{ 28, 300, 45 },
		{ 28, 301, 45 },
		{ 28, 302, 45 },
		{ 28, 303, 45 },
		{ 28, 304, 45 },
		{ 28, 305, 45 },
		{ 28, 306, 45 },
		{ 28, 307, 45 },
		{ 28, 308, 45 },
		{ 28, 309, 45 },
		{ 28, 310, 45 },
		{ 28, 311, 45 },
		{ 28, 312, 45 },
		{ 28, 313, 45 },
		{ 28, 314, 45 },
		{ 29, 258, 45 },
		{ 29, 259, 45 },
		{ 29, 260, 45 },
		{ 29, 261, 45 },
		{ 29, 262, 45 },
		{ 29, 263, 45 },
		{ 29, 264, 45 },
		{ 29, 265, 45 },
		{ 29, 266, 45 },
		{ 29, 267, 45 },
		{ 29, 268, 45 },
		{ 29, 269, 45 },
		{ 29, 270, 45 },
		{ 29, 271, 45 },
		{ 29, 272, 45 },
		{ 29, 273, 45 },
		{ 29, 274, 45 },
		{ 29, 275, 45 },
		{ 29, 276, 45 },
		{ 29, 277, 45 },
		{ 29, 278, 45 },
		{ 29, 279, 45 },
		{ 29, 280, 45 },
		{ 29, 281, 45 },
		{ 29, 282, 45 },
		{ 29, 283, 45 },
		{ 29, 284, 45 },
		{ 29, 285, 45 },
		{ 29, 286, 45 },
		{ 29, 287, 45 },
		{ 29, 288, 45 },
		{ 29, 289, 45 },
		{ 29, 290, 45 },
		{ 29, 291, 45 },
		{ 29, 292, 45 },
		{ 29, 293, 45 },
		{ 29, 294, 45 },
		{ 29, 295, 45 },
		{ 29, 296, 45 },
		{ 29, 297, 45 },
		{ 29, 298, 45 },
		{ 29, 299, 45 },
		{ 29, 300, 45 },
		{ 29, 301, 45 },
		{ 29, 302, 45 },
		{ 29, 303, 45 },
		{ 29, 304, 45 },
		{ 29, 305, 45 },
		{ 29, 306, 45 },
		{ 29, 307, 45 },
		{ 29, 308, 45 },
		{ 29, 309, 45 },
		{ 29, 310, 45 },
		{ 29, 311, 45 },
		{ 29, 312, 45 },
		{ 29, 313, 45 },
		{ 30, 258, 45 },
		{ 30, 259, 45 },
		{ 30, 260, 45 },
		{ 30, 261, 45 },
		{ 30, 262, 45 },
		{ 30, 263, 45 },
		{ 30, 264, 45 },
		{ 30, 265, 45 },
		{ 30, 266, 45 },
		{ 30, 267, 45 },
		{ 30, 268, 45 },
		{ 30, 269, 45 },
		{ 30, 270, 45 },
		{ 30, 271, 45 },
		{ 30, 272, 45 },
		{ 30, 273, 45 },
		{ 30, 274, 45 },
		{ 30, 275, 45 },
		{ 30, 276, 45 },
		{ 30, 277, 45 },
		{ 30, 278, 45 },
		{ 30, 279, 45 },
		{ 30, 280, 45 },
		{ 30, 281, 45 },
		{ 30, 282, 45 },
		{ 30, 283, 45 },
		{ 30, 284, 45 },
		{ 30, 285, 45 },
		{ 30, 286, 45 },
		{ 30, 287, 45 },
		{ 30, 288, 45 },
		{ 30, 289, 45 },
		{ 30, 290, 45 },
		{ 30, 291, 45 },
		{ 30, 292, 45 },
		{ 30, 293, 45 },
		{ 30, 294, 45 },
		{ 30, 295, 45 },
		{ 30, 296, 45 },
		{ 30, 297, 45 },
		{ 30, 298, 45 },
		{ 30, 299, 45 },
		{ 30, 300, 45 },
		{ 30, 301, 45 },
		{ 30, 302, 45 },
		{ 30, 303, 45 },
		{ 30, 304, 45 },
		{ 30, 305, 45 },
		{ 30, 306, 45 },
		{ 30, 307, 45 },
		{ 30, 308, 45 },
		{ 30, 309, 45 },
		{ 30, 310, 45 },
		{ 30, 311, 45 },
		{ 30, 312, 45 },
		{ 30, 313, 45 },
		{ 31, 258, 45 },
		{ 31, 259, 45 },
		{ 31, 260, 45 },
		{ 31, 261, 45 },
		{ 31, 262, 45 },
		{ 31, 263, 45 },
		{ 31, 264, 45 },
		{ 31, 265, 45 },
		{ 31, 266, 45 },
		{ 31, 267, 45 },
		{ 31, 268, 45 },
		{ 31, 269, 45 },
		{ 31, 270, 45 },
		{ 31, 271, 45 },
		{ 31, 272, 45 },
		{ 31, 273, 45 },
		{ 31, 274, 45 },
		{ 31, 275, 45 },
		{ 31, 276, 45 },
		{ 31, 277, 45 },
		{ 31, 278, 45 },
		{ 31, 279, 45 },
		{ 31, 280, 45 },
		{ 31, 281, 45 },
		{ 31, 282, 45 },
		{ 31, 283, 45 },
		{ 31, 284, 45 },
		{ 31, 285, 45 },
		{ 31, 286, 45 },
		{ 31, 287, 45 },
		{ 31, 288, 45 },
		{ 31, 289, 45 },
		{ 31, 290, 45 },
		{ 31, 291, 45 },
		{ 31, 292, 45 },
		{ 31, 293, 45 },
		{ 31, 294, 45 },
		{ 31, 295, 45 },
		{ 31, 296, 45 },
		{ 31, 297, 45 },
		{ 31, 298, 45 },
		{ 31, 299, 45 },
		{ 31, 300, 45 },
		{ 31, 301, 45 },
		{ 31, 302, 45 },
		{ 31, 303, 45 },
		{ 31, 304, 45 },
		{ 31, 305, 45 },
		{ 31, 306, 45 },
		{ 31, 307, 45 },
		{ 31, 308, 45 },
		{ 31, 309, 45 },
		{ 31, 310, 45 },
		{ 31, 311, 45 },
		{ 31, 312, 45 },
		{ 31, 313, 45 },
		{ 32, 258, 45 },
		{ 32, 259, 45 },
		{ 32, 260, 45 },
		{ 32, 261, 45 },
		{ 32, 262, 45 },
		{ 32, 263, 45 },
		{ 32, 264, 45 },
		{ 32, 265, 45 },
		{ 32, 266, 45 },
		{ 32, 267, 45 },
		{ 32, 268, 45 },
		{ 32, 269, 45 },
		{ 32, 270, 45 },
		{ 32, 271, 45 },
		{ 32, 272, 45 },
		{ 32, 273, 45 },
		{ 32, 274, 45 },
		{ 32, 275, 45 },
		{ 32, 276, 45 },
		{ 32, 277, 45 },
		{ 32, 278, 45 },
		{ 32, 279, 45 },
		{ 32, 280, 45 },
		{ 32, 281, 45 },
		{ 32, 282, 45 },
		{ 32, 283, 45 },
		{ 32, 284, 45 },
		{ 32, 285, 45 },
		{ 32, 286, 45 },
		{ 32, 287, 45 },
		{ 32, 288, 45 },
		{ 32, 289, 45 },
		{ 32, 290, 45 },
		{ 32, 291, 45 },
		{ 32, 292, 45 },
		{ 32, 293, 45 },
		{ 32, 294, 45 },
		{ 32, 295, 45 },
		{ 32, 296, 45 },
		{ 32, 297, 45 },
		{ 32, 298, 45 },
		{ 32, 299, 45 },
		{ 32, 300, 45 },
		{ 32, 301, 45 },
		{ 32, 302, 45 },
		{ 32, 303, 45 },
		{ 32, 304, 45 },
		{ 32, 305, 45 },
		{ 32, 306, 45 },
		{ 32, 307, 45 },
		{ 32, 308, 45 },
		{ 32, 309, 45 },
		{ 32, 310, 45 },
		{ 32, 311, 45 },
		{ 32, 312, 45 },
		{ 32, 313, 45 },
		{ 33, 258, 45 },
		{ 33, 259, 45 },
		{ 33, 260, 45 },
		{ 33, 261, 45 },
		{ 33, 262, 45 },
		{ 33, 263, 45 },
		{ 33, 264, 45 },
		{ 33, 265, 45 },
		{ 33, 266, 45 },
		{ 33, 267, 45 },
		{ 33, 268, 45 },
		{ 33, 269, 45 },
		{ 33, 270, 45 },
		{ 33, 271, 45 },
		{ 33, 272, 45 },
		{ 33, 273, 45 },
		{ 33, 274, 45 },
		{ 33, 275, 45 },
		{ 33, 276, 45 },
		{ 33, 277, 45 },
		{ 33, 278, 45 },
		{ 33, 279, 45 },
		{ 33, 280, 45 },
		{ 33, 281, 45 },
		{ 33, 282, 45 },
		{ 33, 283, 45 },
		{ 33, 284, 45 },
		{ 33, 285, 45 },
		{ 33, 286, 45 },
		{ 33, 287, 45 },
		{ 33, 288, 45 },
		{ 33, 289, 45 },
		{ 33, 290, 45 },
		{ 33, 291, 45 },
		{ 33, 292, 45 },
		{ 33, 293, 45 },
		{ 33, 294, 45 },
		{ 33, 295, 45 },
		{ 33, 296, 45 },
		{ 33, 297, 45 },
		{ 33, 298, 45 },
		{ 33, 299, 45 },
		{ 33, 300, 45 },
		{ 33, 301, 45 },
		{ 33, 302, 45 },
		{ 33, 303, 45 },
		{ 33, 304, 45 },
		{ 33, 305, 45 },
		{ 33, 306, 45 },
		{ 33, 307, 45 },
		{ 33, 308, 45 },
		{ 33, 309, 45 },
		{ 33, 310, 45 },
		{ 33, 311, 45 },
		{ 33, 312, 45 },
		{ 34, 258, 45 },
		{ 34, 259, 45 },
		{ 34, 260, 45 },
		{ 34, 261, 45 },
		{ 34, 262, 45 },
		{ 34, 263, 45 },
		{ 34, 264, 45 },
		{ 34, 265, 45 },
		{ 34, 266, 45 },
		{ 34, 267, 45 },
		{ 34, 268, 45 },
		{ 34, 269, 45 },
		{ 34, 270, 45 },
		{ 34, 271, 45 },
		{ 34, 272, 45 },
		{ 34, 273, 45 },
		{ 34, 274, 45 },
		{ 34, 275, 45 },
		{ 34, 276, 45 },
		{ 34, 277, 45 },
		{ 34, 278, 45 },
		{ 34, 279, 45 },
		{ 34, 280, 45 },
		{ 34, 281, 45 },
		{ 34, 282, 45 },
		{ 34, 283, 45 },
		{ 34, 284, 45 },
		{ 34, 285, 45 },
		{ 34, 286, 45 },
		{ 34, 287, 45 },
		{ 34, 288, 45 },
		{ 34, 289, 45 },
		{ 34, 290, 45 },
		{ 34, 291, 45 },
		{ 34, 292, 45 },
		{ 34, 293, 45 },
		{ 34, 294, 45 },
		{ 34, 295, 45 },
		{ 34, 296, 45 },
		{ 34, 297, 45 },
		{ 34, 298, 45 },
		{ 34, 299, 45 },
		{ 34, 300, 45 },
		{ 34, 301, 45 },
		{ 34, 302, 45 },
		{ 34, 303, 45 },
		{ 34, 304, 45 },
		{ 34, 305, 45 },
		{ 34, 306, 45 },
		{ 34, 307, 45 },
		{ 34, 308, 45 },
		{ 34, 309, 45 },
		{ 34, 310, 45 },
		{ 34, 311, 45 },
		{ 34, 312, 45 },
		{ 35, 258, 45 },
		{ 35, 259, 45 },
		{ 35, 260, 45 },
		{ 35, 261, 45 },
		{ 35, 262, 45 },
		{ 35, 263, 45 },
		{ 35, 264, 45 },
		{ 35, 265, 45 },
		{ 35, 266, 45 },
		{ 35, 267, 45 },
		{ 35, 268, 45 },
		{ 35, 269, 45 },
		{ 35, 270, 45 },
		{ 35, 271, 45 },
		{ 35, 272, 45 },
		{ 35, 273, 45 },
		{ 35, 274, 45 },
		{ 35, 275, 45 },
		{ 35, 276, 45 },
		{ 35, 277, 45 },
		{ 35, 278, 45 },
		{ 35, 279, 45 },
		{ 35, 280, 45 },
		{ 35, 281, 45 },
		{ 35, 282, 45 },
		{ 35, 283, 45 },
		{ 35, 284, 45 },
		{ 35, 285, 45 },
		{ 35, 286, 45 },
		{ 35, 287, 45 },
		{ 35, 288, 45 },
		{ 35, 289, 45 },
		{ 35, 290, 45 },
		{ 35, 291, 45 },
		{ 35, 292, 45 },
		{ 35, 293, 45 },
		{ 35, 294, 45 },
		{ 35, 295, 45 },
		{ 35, 296, 45 },
		{ 35, 297, 45 },
		{ 35, 298, 45 },
		{ 35, 299, 45 },
		{ 35, 300, 45 },
		{ 35, 301, 45 },
		{ 35, 302, 45 },
		{ 35, 303, 45 },
		{ 35, 304, 45 },
		{ 35, 305, 45 },
		{ 35, 306, 45 },
		{ 35, 307, 45 },
		{ 35, 308, 45 },
		{ 35, 309, 45 },
		{ 35, 310, 45 },
		{ 35, 311, 45 },
		{ 35, 312, 45 },
		{ 36, 258, 45 },
		{ 36, 259, 45 },
		{ 36, 260, 45 },
		{ 36, 261, 45 },
		{ 36, 262, 45 },
		{ 36, 263, 45 },
		{ 36, 264, 45 },
		{ 36, 265, 45 },
		{ 36, 266, 45 },
		{ 36, 267, 45 },
		{ 36, 268, 45 },
		{ 36, 269, 45 },
		{ 36, 270, 45 },
		{ 36, 271, 45 },
		{ 36, 272, 45 },
		{ 36, 273, 45 },
		{ 36, 274, 45 },
		{ 36, 275, 45 },
		{ 36, 276, 45 },
		{ 36, 277, 45 },
		{ 36, 278, 45 },
		{ 36, 279, 45 },
		{ 36, 280, 45 },
		{ 36, 281, 45 },
		{ 36, 282, 45 },
		{ 36, 283, 45 },
		{ 36, 284, 45 },
		{ 36, 285, 45 },
		{ 36, 286, 45 },
		{ 36, 287, 45 },
		{ 36, 288, 45 },
		{ 36, 289, 45 },
		{ 36, 290, 45 },
		{ 36, 291, 45 },
		{ 36, 292, 45 },
		{ 36, 293, 45 },
		{ 36, 294, 45 },
		{ 36, 295, 45 },
		{ 36, 296, 45 },
		{ 36, 297, 45 },
		{ 36, 298, 45 },
		{ 36, 299, 45 },
		{ 36, 300, 45 },
		{ 36, 301, 45 },
		{ 36, 302, 45 },
		{ 36, 303, 45 },
		{ 36, 304, 45 },
		{ 36, 305, 45 },
		{ 36, 306, 45 },
		{ 36, 307, 45 },
		{ 36, 308, 45 },
		{ 36, 309, 45 },
		{ 36, 310, 45 },
		{ 36, 311, 45 },
		{ 36, 312, 45 },
		{ 37, 258, 45 },
		{ 37, 259, 45 },
		{ 37, 260, 45 },
		{ 37, 261, 45 },
		{ 37, 262, 45 },
		{ 37, 263, 45 },
		{ 37, 264, 45 },
		{ 37, 265, 45 },
		{ 37, 266, 45 },
		{ 37, 267, 45 },
		{ 37, 268, 45 },
		{ 37, 269, 45 },
		{ 37, 270, 45 },
		{ 37, 271, 45 },
		{ 37, 272, 45 },
		{ 37, 273, 45 },
		{ 37, 274, 45 },
		{ 37, 275, 45 },
		{ 37, 276, 45 },
		{ 37, 277, 45 },
		{ 37, 278, 45 },
		{ 37, 279, 45 },
		{ 37, 280, 45 },
		{ 37, 281, 45 },
		{ 37, 282, 45 },
		{ 37, 283, 45 },
		{ 37, 284, 45 },
		{ 37, 285, 45 },
		{ 37, 286, 45 },
		{ 37, 287, 45 },
		{ 37, 288, 45 },
		{ 37, 289, 45 },
		{ 37, 290, 45 },
		{ 37, 291, 45 },
		{ 37, 292, 45 },
		{ 37, 293, 45 },
		{ 37, 294, 45 },
		{ 37, 295, 45 },
		{ 37, 296, 45 },
		{ 37, 297, 45 },
		{ 37, 298, 45 },
		{ 37, 299, 45 },
		{ 37, 300, 45 },
		{ 37, 301, 45 },
		{ 37, 302, 45 },
		{ 37, 303, 45 },
		{ 37, 304, 45 },
		{ 37, 305, 45 },
		{ 37, 306, 45 },
		{ 37, 307, 45 },
		{ 37, 308, 45 },
		{ 37, 309, 45 },
		{ 37, 310, 45 },
		{ 37, 311, 45 },
		{ 38, 258, 45 },
		{ 38, 259, 45 },
		{ 38, 260, 45 },
		{ 38, 261, 45 },
		{ 38, 262, 45 },
		{ 38, 263, 45 },
		{ 38, 264, 45 },
		{ 38, 265, 45 },
		{ 38, 266, 45 },
		{ 38, 267, 45 },
		{ 38, 268, 45 },
		{ 38, 269, 45 },
		{ 38, 270, 45 },
		{ 38, 271, 45 },
		{ 38, 272, 45 },
		{ 38, 273, 45 },
		{ 38, 274, 45 },
		{ 38, 275, 45 },
		{ 38, 276, 45 },
		{ 38, 277, 45 },
		{ 38, 278, 45 },
		{ 38, 279, 45 },
		{ 38, 280, 45 },
		{ 38, 281, 45 },
		{ 38, 282, 45 },
		{ 38, 283, 45 },
		{ 38, 284, 45 },
		{ 38, 285, 45 },
		{ 38, 286, 45 },
		{ 38, 287, 45 },
		{ 38, 288, 45 },
		{ 38, 289, 45 },
		{ 38, 290, 45 },
		{ 38, 291, 45 },
		{ 38, 292, 45 },
		{ 38, 293, 45 },
		{ 38, 294, 45 },
		{ 38, 295, 45 },
		{ 38, 296, 45 },
		{ 38, 297, 45 },
		{ 38, 298, 45 },
		{ 38, 299, 45 },
		{ 38, 300, 45 },
		{ 38, 301, 45 },
		{ 38, 302, 45 },
		{ 38, 303, 45 },
		{ 38, 304, 45 },
		{ 38, 305, 45 },
		{ 38, 306, 45 },
		{ 38, 307, 45 },
		{ 38, 308, 45 },
		{ 38, 309, 45 },
		{ 38, 310, 45 },
		{ 38, 311, 45 },
		{ 39, 258, 45 },
		{ 39, 259, 45 },
		{ 39, 260, 45 },
		{ 39, 261, 45 },
		{ 39, 262, 45 },
		{ 39, 263, 45 },
		{ 39, 264, 45 },
		{ 39, 265, 45 },
		{ 39, 266, 45 },
		{ 39, 267, 45 },
		{ 39, 268, 45 },
		{ 39, 269, 45 },
		{ 39, 270, 45 },
		{ 39, 271, 45 },
		{ 39, 272, 45 },
		{ 39, 273, 45 },
		{ 39, 274, 45 },
		{ 39, 275, 45 },
		{ 39, 276, 45 },
		{ 39, 277, 45 },
		{ 39, 278, 45 },
		{ 39, 279, 45 },
		{ 39, 280, 45 },
		{ 39, 281, 45 },
		{ 39, 282, 45 },
		{ 39, 283, 45 },
		{ 39, 284, 45 },
		{ 39, 285, 45 },
		{ 39, 286, 45 },
		{ 39, 287, 45 },
		{ 39, 288, 45 },
		{ 39, 289, 45 },
		{ 39, 290, 45 },
		{ 39, 291, 45 },
		{ 39, 292, 45 },
		{ 39, 293, 45 },
		{ 39, 294, 45 },
		{ 39, 295, 45 },
		{ 39, 296, 45 },
		{ 39, 297, 45 },
		{ 39, 298, 45 },
		{ 39, 299, 45 },
		{ 39, 300, 45 },
		{ 39, 301, 45 },
		{ 39, 302, 45 },
		{ 39, 303, 45 },
		{ 39, 304, 45 },
		{ 39, 305, 45 },
		{ 39, 306, 45 },
		{ 39, 307, 45 },
		{ 39, 308, 45 },
		{ 39, 309, 45 },
		{ 39, 310, 45 },
		{ 39, 311, 45 },
		{ 40, 258, 45 },
		{ 40, 259, 45 },
		{ 40, 260, 45 },
		{ 40, 261, 45 },
		{ 40, 262, 45 },
		{ 40, 263, 45 },
		{ 40, 264, 45 },
		{ 40, 265, 45 },
		{ 40, 266, 45 },
		{ 40, 267, 45 },
		{ 40, 268, 45 },
		{ 40, 269, 45 },
		{ 40, 270, 45 },
		{ 40, 271, 45 },
		{ 40, 272, 45 },
		{ 40, 273, 45 },
		{ 40, 274, 45 },
		{ 40, 275, 45 },
		{ 40, 276, 45 },
		{ 40, 277, 45 },
		{ 40, 278, 45 },
		{ 40, 279, 45 },
		{ 40, 280, 45 },
		{ 40, 281, 45 },
		{ 40, 282, 45 },
		{ 40, 283, 45 },
		{ 40, 284, 45 },
		{ 40, 285, 45 },
		{ 40, 286, 45 },
		{ 40, 287, 45 },
		{ 40, 288, 45 },
		{ 40, 289, 45 },
		{ 40, 290, 45 },
		{ 40, 291, 45 },
		{ 40, 292, 45 },
		{ 40, 293, 45 },
		{ 40, 294, 45 },
		{ 40, 295, 45 },
		{ 40, 296, 45 },
		{ 40, 297, 45 },
		{ 40, 298, 45 },
		{ 40, 299, 45 },
		{ 40, 300, 45 },
		{ 40, 301, 45 },
		{ 40, 302, 45 },
		{ 40, 303, 45 },
		{ 40, 304, 45 },
		{ 40, 305, 45 },
		{ 40, 306, 45 },
		{ 40, 307, 45 },
		{ 40, 308, 45 },
		{ 40, 309, 45 },
		{ 40, 310, 45 },
		{ 40, 311, 45 },
		{ 41, 258, 45 },
		{ 41, 259, 45 },
		{ 41, 260, 45 },
		{ 41, 261, 45 },
		{ 41, 262, 45 },
		{ 41, 263, 45 },
		{ 41, 264, 45 },
		{ 41, 265, 45 },
		{ 41, 266, 45 },
		{ 41, 267, 45 },
		{ 41, 268, 45 },
		{ 41, 269, 45 },
		{ 41, 270, 45 },
		{ 41, 271, 45 },
		{ 41, 272, 45 },
		{ 41, 273, 45 },
		{ 41, 274, 45 },
		{ 41, 275, 45 },
		{ 41, 276, 45 },
		{ 41, 277, 45 },
		{ 41, 278, 45 },
		{ 41, 279, 45 },
		{ 41, 280, 45 },
		{ 41, 281, 45 },
		{ 41, 282, 45 },
		{ 41, 283, 45 },
		{ 41, 284, 45 },
		{ 41, 285, 45 },
		{ 41, 286, 45 },
		{ 41, 287, 45 },
		{ 41, 288, 45 },
		{ 41, 289, 45 },
		{ 41, 290, 45 },
		{ 41, 291, 45 },
		{ 41, 292, 45 },
		{ 41, 293, 45 },
		{ 41, 294, 45 },
		{ 41, 295, 45 },
		{ 41, 296, 45 },
		{ 41, 297, 45 },
		{ 41, 298, 45 },
		{ 41, 299, 45 },
		{ 41, 300, 45 },
		{ 41, 301, 45 },
		{ 41, 302, 45 },
		{ 41, 303, 45 },
		{ 41, 304, 45 },
		{ 41, 305, 45 },
		{ 41, 306, 45 },
		{ 41, 307, 45 },
		{ 41, 308, 45 },
		{ 41, 309, 45 },
		{ 41, 310, 45 },
		{ 42, 258, 45 },
		{ 42, 259, 45 },
		{ 42, 260, 45 },
		{ 42, 261, 45 },
		{ 42, 262, 45 },
		{ 42, 263, 45 },
		{ 42, 264, 45 },
		{ 42, 265, 45 },
		{ 42, 266, 45 },
		{ 42, 267, 45 },
		{ 42, 268, 45 },
		{ 42, 269, 45 },
		{ 42, 270, 45 },
		{ 42, 271, 45 },
		{ 42, 272, 45 },
		{ 42, 273, 45 },
		{ 42, 274, 45 },
		{ 42, 275, 45 },
		{ 42, 276, 45 },
		{ 42, 277, 45 },
		{ 42, 278, 45 },
		{ 42, 279, 45 },
		{ 42, 280, 45 },
		{ 42, 281, 45 },
		{ 42, 282, 45 },
		{ 42, 283, 45 },
		{ 42, 284, 45 },
		{ 42, 285, 45 },
		{ 42, 286, 45 },
		{ 42, 287, 45 },
		{ 42, 288, 45 },
		{ 42, 289, 45 },
		{ 42, 290, 45 },
		{ 42, 291, 45 },
		{ 42, 292, 45 },
		{ 42, 293, 45 },
		{ 42, 294, 45 },
		{ 42, 295, 45 },
		{ 42, 296, 45 },
		{ 42, 297, 45 },
		{ 42, 298, 45 },
		{ 42, 299, 45 },
		{ 42, 300, 45 },
		{ 42, 301, 45 },
		{ 42, 302, 45 },
		{ 42, 303, 45 },
		{ 42, 304, 45 },
		{ 42, 305, 45 },
		{ 42, 306, 45 },
		{ 42, 307, 45 },
		{ 42, 308, 45 },
		{ 42, 309, 45 },
		{ 42, 310, 45 },
		{ 43, 258, 45 },
		{ 43, 259, 45 },
		{ 43, 260, 45 },
		{ 43, 261, 45 },
		{ 43, 262, 45 },
		{ 43, 263, 45 },
		{ 43, 264, 45 },
		{ 43, 265, 45 },
		{ 43, 266, 45 },
		{ 43, 267, 45 },
		{ 43, 268, 45 },
		{ 43, 269, 45 },
		{ 43, 270, 45 },
		{ 43, 271, 45 },
		{ 43, 272, 45 },
		{ 43, 273, 45 },
		{ 43, 274, 45 },
		{ 43, 275, 45 },
		{ 43, 276, 45 },
		{ 43, 277, 45 },
		{ 43, 278, 45 },
		{ 43, 279, 45 },
		{ 43, 280, 45 },
		{ 43, 281, 45 },
		{ 43, 282, 45 },
		{ 43, 283, 45 },
		{ 43, 284, 45 },
		{ 43, 285, 45 },
		{ 43, 286, 45 },
		{ 43, 287, 45 },
		{ 43, 288, 45 },
		{ 43, 289, 45 },
		{ 43, 290, 45 },
		{ 43, 291, 45 },
		{ 43, 292, 45 },
		{ 43, 293, 45 },
		{ 43, 294, 45 },
		{ 43, 295, 45 },
		{ 43, 296, 45 },
		{ 43, 297, 45 },
		{ 43, 298, 45 },
		{ 43, 299, 45 },
		{ 43, 300, 45 },
		{ 43, 301, 45 },
		{ 43, 302, 45 },
		{ 43, 303, 45 },
		{ 43, 304, 45 },
		{ 43, 305, 45 },
		{ 43, 306, 45 },
		{ 43, 307, 45 },
		{ 43, 308, 45 },
		{ 43, 309, 45 },
		{ 43, 310, 45 },
		{ 44, 258, 45 },
		{ 44, 259, 45 },
		{ 44, 260, 45 },
		{ 44, 261, 45 },
		{ 44, 262, 45 },
		{ 44, 263, 45 },
		{ 44, 264, 45 },
		{ 44, 265, 45 },
		{ 44, 266, 45 },
		{ 44, 267, 45 },
		{ 44, 268, 45 },
		{ 44, 269, 45 },
		{ 44, 270, 45 },
		{ 44, 271, 45 },
		{ 44, 272, 45 },
		{ 44, 273, 45 },
		{ 44, 274, 45 },
		{ 44, 275, 45 },
		{ 44, 276, 45 },
		{ 44, 277, 45 },
		{ 44, 278, 45 },
		{ 44, 279, 45 },
		{ 44, 280, 45 },
		{ 44, 281, 45 },
		{ 44, 282, 45 },
		{ 44, 283, 45 },
		{ 44, 284, 45 },
		{ 44, 285, 45 },
		{ 44, 286, 45 },
		{ 44, 287, 45 },
		{ 44, 288, 45 },
		{ 44, 289, 45 },
		{ 44, 290, 45 },
		{ 44, 291, 45 },
		{ 44, 292, 45 },
		{ 44, 293, 45 },
		{ 44, 294, 45 },
		{ 44, 295, 45 },
		{ 44, 296, 45 },
		{ 44, 297, 45 },
		{ 44, 298, 45 },
		{ 44, 299, 45 },
		{ 44, 300, 45 },
		{ 44, 301, 45 },
		{ 44, 302, 45 },
		{ 44, 303, 45 },
		{ 44, 304, 45 },
		{ 44, 305, 45 },
		{ 44, 306, 45 },
		{ 44, 307, 45 },
		{ 44, 308, 45 },
		{ 44, 309, 45 },
		{ 44, 310, 45 },
		{ 45, 258, 45 },
		{ 45, 259, 45 },
		{ 45, 260, 45 },
		{ 45, 261, 45 },
		{ 45, 262, 45 },
		{ 45, 263, 45 },
		{ 45, 264, 45 },
		{ 45, 265, 45 },
		{ 45, 266, 45 },
		{ 45, 267, 45 },
		{ 45, 268, 45 },
		{ 45, 269, 45 },
		{ 45, 270, 45 },
		{ 45, 271, 45 },
		{ 45, 272, 45 },
		{ 45, 273, 45 },
		{ 45, 274, 45 },
		{ 45, 275, 45 },
		{ 45, 276, 45 },
		{ 45, 277, 45 },
		{ 45, 278, 45 },
		{ 45, 279, 45 },
		{ 45, 280, 45 },
		{ 45, 281, 45 },
		{ 45, 282, 45 },
		{ 45, 283, 45 },
		{ 45, 284, 45 },
		{ 45, 285, 45 },
		{ 45, 286, 45 },
		{ 45, 287, 45 },
		{ 45, 288, 45 },
		{ 45, 289, 45 },
		{ 45, 290, 45 },
		{ 45, 291, 45 },
		{ 45, 292, 45 },
		{ 45, 293, 45 },
		{ 45, 294, 45 },
		{ 45, 295, 45 },
		{ 45, 296, 45 },
		{ 45, 297, 45 },
		{ 45, 298, 45 },
		{ 45, 299, 45 },
		{ 45, 300, 45 },
		{ 45, 301, 45 },
		{ 45, 302, 45 },
		{ 45, 303, 45 },
		{ 45, 304, 45 },
		{ 45, 305, 45 },
		{ 45, 306, 45 },
		{ 45, 307, 45 },
		{ 45, 308, 45 },
		{ 45, 309, 45 },
		{ 46, 258, 45 },
		{ 46, 259, 45 },
		{ 46, 260, 45 },
		{ 46, 261, 45 },
		{ 46, 262, 45 },
		{ 46, 263, 45 },
		{ 46, 264, 45 },
		{ 46, 265, 45 },
		{ 46, 266, 45 },
		{ 46, 267, 45 },
		{ 46, 268, 45 },
		{ 46, 269, 45 },
		{ 46, 270, 45 },
		{ 46, 271, 45 },
		{ 46, 272, 45 },
		{ 46, 273, 45 },
		{ 46, 274, 45 },
		{ 46, 275, 45 },
		{ 46, 276, 45 },
		{ 46, 277, 45 },
		{ 46, 278, 45 },
		{ 46, 279, 45 },
		{ 46, 280, 45 },
		{ 46, 281, 45 },
		{ 46, 282, 45 },
		{ 46, 283, 45 },
		{ 46, 284, 45 },
		{ 46, 285, 45 },
		{ 46, 286, 45 },
		{ 46, 287, 45 },
		{ 46, 288, 45 },
		{ 46, 289, 45 },
		{ 46, 290, 45 },
		{ 46, 291, 45 },
		{ 46, 292, 45 },
		{ 46, 293, 45 },
		{ 46, 294, 45 },
		{ 46, 295, 45 },
		{ 46, 296, 45 },
		{ 46, 297, 45 },
		{ 46, 298, 45 },
		{ 46, 299, 45 },
		{ 46, 300, 45 },
		{ 46, 301, 45 },
		{ 46, 302, 45 },
		{ 46, 303, 45 },
		{ 46, 304, 45 },
		{ 46, 305, 45 },
		{ 46, 306, 45 },
		{ 46, 307, 45 },
		{ 46, 308, 45 },
		{ 46, 309, 45 },
		{ 47, 258, 45 },
		{ 47, 259, 45 },
		{ 47, 260, 45 },
		{ 47, 261, 45 },
		{ 47, 262, 45 },
		{ 47, 263, 45 },
		{ 47, 264, 45 },
		{ 47, 265, 45 },
		{ 47, 266, 45 },
		{ 47, 267, 45 },
		{ 47, 268, 45 },
		{ 47, 269, 45 },
		{ 47, 270, 45 },
		{ 47, 271, 45 },
		{ 47, 272, 45 },
		{ 47, 273, 45 },
		{ 47, 274, 45 },
		{ 47, 275, 45 },
		{ 47, 276, 45 },
		{ 47, 277, 45 },
		{ 47, 278, 45 },
		{ 47, 279, 45 },
		{ 47, 280, 45 },
		{ 47, 281, 45 },
		{ 47, 282, 45 },
		{ 47, 283, 45 },
		{ 47, 284, 45 },
		{ 47, 285, 45 },
		{ 47, 286, 45 },
		{ 47, 287, 45 },
		{ 47, 288, 45 },
		{ 47, 289, 45 },
		{ 47, 290, 45 },
		{ 47, 291, 45 },
		{ 47, 292, 45 },
		{ 47, 293, 45 },
		{ 47, 294, 45 },
		{ 47, 295, 45 },
		{ 47, 296, 45 },
		{ 47, 297, 45 },
		{ 47, 298, 45 },
		{ 47, 299, 45 },
		{ 47, 300, 45 },
		{ 47, 301, 45 },
		{ 47, 302, 45 },
		{ 47, 303, 45 },
		{ 47, 304, 45 },
		{ 47, 305, 45 },
		{ 47, 306, 45 },
		{ 47, 307, 45 },
		{ 47, 308, 45 },
		{ 47, 309, 45 },
		{ 48, 258, 45 },
		{ 48, 259, 45 },
		{ 48, 260, 45 },
		{ 48, 261, 45 },
		{ 48, 262, 45 },
		{ 48, 263, 45 },
		{ 48, 264, 45 },
		{ 48, 265, 45 },
		{ 48, 266, 45 },
		{ 48, 267, 45 },
		{ 48, 268, 45 },
		{ 48, 269, 45 },
		{ 48, 270, 45 },
		{ 48, 271, 45 },
		{ 48, 272, 45 },
		{ 48, 273, 45 },
		{ 48, 274, 45 },
		{ 48, 275, 45 },
		{ 48, 276, 45 },
		{ 48, 277, 45 },
		{ 48, 278, 45 },
		{ 48, 279, 45 },
		{ 48, 280, 45 },
		{ 48, 281, 45 },
		{ 48, 282, 45 },
		{ 48, 283, 45 },
		{ 48, 284, 45 },
		{ 48, 285, 45 },
		{ 48, 286, 45 },
		{ 48, 287, 45 },
		{ 48, 288, 45 },
		{ 48, 289, 45 },
		{ 48, 290, 45 },
		{ 48, 291, 45 },
		{ 48, 292, 45 },
		{ 48, 293, 45 },
		{ 48, 294, 45 },
		{ 48, 295, 45 },
		{ 48, 296, 45 },
		{ 48, 297, 45 },
		{ 48, 298, 45 },
		{ 48, 299, 45 },
		{ 48, 300, 45 },
		{ 48, 301, 45 },
		{ 48, 302, 45 },
		{ 48, 303, 45 },
		{ 48, 304, 45 },
		{ 48, 305, 45 },
		{ 48, 306, 45 },
		{ 48, 307, 45 },
		{ 48, 308, 45 },
		{ 48, 309, 45 },
		{ 49, 258, 45 },
		{ 49, 259, 45 },
		{ 49, 260, 45 },
		{ 49, 261, 45 },
		{ 49, 262, 45 },
		{ 49, 263, 45 },
		{ 49, 264, 45 },
		{ 49, 265, 45 },
		{ 49, 266, 45 },
		{ 49, 267, 45 },
		{ 49, 268, 45 },
		{ 49, 269, 45 },
		{ 49, 270, 45 },
		{ 49, 271, 45 },
		{ 49, 272, 45 },
		{ 49, 273, 45 },
		{ 49, 274, 45 },
		{ 49, 275, 45 },
		{ 49, 276, 45 },
		{ 49, 277, 45 },
		{ 49, 278, 45 },
		{ 49, 279, 45 },
		{ 49, 280, 45 },
		{ 49, 281, 45 },
		{ 49, 282, 45 },
		{ 49, 283, 45 },
		{ 49, 284, 45 },
		{ 49, 285, 45 },
		{ 49, 286, 45 },
		{ 49, 287, 45 },
		{ 49, 288, 45 },
		{ 49, 289, 45 },
		{ 49, 290, 45 },
		{ 49, 291, 45 },
		{ 49, 292, 45 },
		{ 49, 293, 45 },
		{ 49, 294, 45 },
		{ 49, 295, 45 },
		{ 49, 296, 45 },
		{ 49, 297, 45 },
		{ 49, 298, 45 },
		{ 49, 299, 45 },
		{ 49, 300, 45 },
		{ 49, 301, 45 },
		{ 49, 302, 45 },
		{ 49, 303, 45 },
		{ 49, 304, 45 },
		{ 49, 305, 45 },
		{ 49, 306, 45 },
		{ 49, 307, 45 },
		{ 49, 308, 45 },
		{ 50, 258, 45 },
		{ 50, 259, 45 },
		{ 50, 260, 45 },
		{ 50, 261, 45 },
		{ 50, 262, 45 },
		{ 50, 263, 45 },
		{ 50, 264, 45 },
		{ 50, 265, 45 },
		{ 50, 266, 45 },
		{ 50, 267, 45 },
		{ 50, 268, 45 },
		{ 50, 269, 45 },
		{ 50, 270, 45 },
		{ 50, 271, 45 },
		{ 50, 272, 45 },
		{ 50, 273, 45 },
		{ 50, 274, 45 },
		{ 50, 275, 45 },
		{ 50, 276, 45 },
		{ 50, 277, 45 },
		{ 50, 278, 45 },
		{ 50, 279, 45 },
		{ 50, 280, 45 },
		{ 50, 281, 45 },
		{ 50, 282, 45 },
		{ 50, 283, 45 },
		{ 50, 284, 45 },
		{ 50, 285, 45 },
		{ 50, 286, 45 },
		{ 50, 287, 45 },
		{ 50, 288, 45 },
		{ 50, 289, 45 },
		{ 50, 290, 45 },
		{ 50, 291, 45 },
		{ 50, 292, 45 },
		{ 50, 293, 45 },
		{ 50, 294, 45 },
		{ 50, 295, 45 },
		{ 50, 296, 45 },
		{ 50, 297, 45 },
		{ 50, 298, 45 },
		{ 50, 299, 45 },
		{ 50, 300, 45 },
		{ 50, 301, 45 },
		{ 50, 302, 45 },
		{ 50, 303, 45 },
		{ 50, 304, 45 },
		{ 50, 305, 45 },
		{ 50, 306, 45 },
		{ 50, 307, 45 },
		{ 50, 308, 45 },
		{ 51, 258, 45 },
		{ 51, 259, 45 },
		{ 51, 260, 45 },
		{ 51, 261, 45 },
		{ 51, 262, 45 },
		{ 51, 263, 45 },
		{ 51, 264, 45 },
		{ 51, 265, 45 },
		{ 51, 266, 45 },
		{ 51, 267, 45 },
		{ 51, 268, 45 },
		{ 51, 269, 45 },
		{ 51, 270, 45 },
		{ 51, 271, 45 },
		{ 51, 272, 45 },
		{ 51, 273, 45 },
		{ 51, 274, 45 },
		{ 51, 275, 45 },
		{ 51, 276, 45 },
		{ 51, 277, 45 },
		{ 51, 278, 45 },
		{ 51, 279, 45 },
		{ 51, 280, 45 },
		{ 51, 281, 45 },
		{ 51, 282, 45 },
		{ 51, 283, 45 },
		{ 51, 284, 45 },
		{ 51, 285, 45 },
		{ 51, 286, 45 },
		{ 51, 287, 45 },
		{ 51, 288, 45 },
		{ 51, 289, 45 },
		{ 51, 290, 45 },
		{ 51, 291, 45 },
		{ 51, 292, 45 },
		{ 51, 293, 45 },
		{ 51, 294, 45 },
		{ 51, 295, 45 },
		{ 51, 296, 45 },
		{ 51, 297, 45 },
		{ 51, 298, 45 },
		{ 51, 299, 45 },
		{ 51, 300, 45 },
		{ 51, 301, 45 },
		{ 51, 302, 45 },
		{ 51, 303, 45 },
		{ 51, 304, 45 },
		{ 51, 305, 45 },
		{ 51, 306, 45 },
		{ 51, 307, 45 },
		{ 51, 308, 45 },
		{ 52, 258, 45 },
		{ 52, 259, 45 },
		{ 52, 260, 45 },
		{ 52, 261, 45 },
		{ 52, 262, 45 },
		{ 52, 263, 45 },
		{ 52, 264, 45 },
		{ 52, 265, 45 },
		{ 52, 266, 45 },
		{ 52, 267, 45 },
		{ 52, 268, 45 },
		{ 52, 269, 45 },
		{ 52, 270, 45 },
		{ 52, 271, 45 },
		{ 52, 272, 45 },
		{ 52, 273, 45 },
		{ 52, 274, 45 },
		{ 52, 275, 45 },
		{ 52, 276, 45 },
		{ 52, 277, 45 },
		{ 52, 278, 45 },
		{ 52, 279, 45 },
		{ 52, 280, 45 },
		{ 52, 281, 45 },
		{ 52, 282, 45 },
		{ 52, 283, 45 },
		{ 52, 284, 45 },
		{ 52, 285, 45 },
		{ 52, 286, 45 },
		{ 52, 287, 45 },
		{ 52, 288, 45 },
		{ 52, 289, 45 },
		{ 52, 290, 45 },
		{ 52, 291, 45 },
		{ 52, 292, 45 },
		{ 52, 293, 45 },
		{ 52, 294, 45 },
		{ 52, 295, 45 },
		{ 52, 296, 45 },
		{ 52, 297, 45 },
		{ 52, 298, 45 },
		{ 52, 299, 45 },
		{ 52, 300, 45 },
		{ 52, 301, 45 },
		{ 52, 302, 45 },
		{ 52, 303, 45 },
		{ 52, 304, 45 },
		{ 52, 305, 45 },
		{ 52, 306, 45 },
		{ 52, 307, 45 },
		{ 52, 308, 45 },
		{ 53, 258, 45 },
		{ 53, 259, 45 },
		{ 53, 260, 45 },
		{ 53, 261, 45 },
		{ 53, 262, 45 },
		{ 53, 263, 45 },
		{ 53, 264, 45 },
		{ 53, 265, 45 },
		{ 53, 266, 45 },
		{ 53, 267, 45 },
		{ 53, 268, 45 },
		{ 53, 269, 45 },
		{ 53, 270, 45 },
		{ 53, 271, 45 },
		{ 53, 272, 45 },
		{ 53, 273, 45 },
		{ 53, 274, 45 },
		{ 53, 275, 45 },
		{ 53, 276, 45 },
		{ 53, 277, 45 },
		{ 53, 278, 45 },
		{ 53, 279, 45 },
		{ 53, 280, 45 },
		{ 53, 281, 45 },
		{ 53, 282, 45 },
		{ 53, 283, 45 },
		{ 53, 284, 45 },
		{ 53, 285, 45 },
		{ 53, 286, 45 },
		{ 53, 287, 45 },
		{ 53, 288, 45 },
		{ 53, 289, 45 },
		{ 53, 290, 45 },
		{ 53, 291, 45 },
		{ 53, 292, 45 },
		{ 53, 293, 45 },
		{ 53, 294, 45 },
		{ 53, 295, 45 },
		{ 53, 296, 45 },
		{ 53, 297, 45 },
		{ 53, 298, 45 },
		{ 53, 299, 45 },
		{ 53, 300, 45 },
		{ 53, 301, 45 },
		{ 53, 302, 45 },
		{ 53, 303, 45 },
		{ 53, 304, 45 },
		{ 53, 305, 45 },
		{ 53, 306, 45 },
		{ 53, 307, 45 },
		{ 54, 258, 45 },
		{ 54, 259, 45 },
		{ 54, 260, 45 },
		{ 54, 261, 45 },
		{ 54, 262, 45 },
		{ 54, 263, 45 },
		{ 54, 264, 45 },
		{ 54, 265, 45 },
		{ 54, 266, 45 },
		{ 54, 267, 45 },
		{ 54, 268, 45 },
		{ 54, 269, 45 },
		{ 54, 270, 45 },
		{ 54, 271, 45 },
		{ 54, 272, 45 },
		{ 54, 273, 45 },
		{ 54, 274, 45 },
		{ 54, 275, 45 },
		{ 54, 276, 45 },
		{ 54, 277, 45 },
		{ 54, 278, 45 },
		{ 54, 279, 45 },
		{ 54, 280, 45 },
		{ 54, 281, 45 },
		{ 54, 282, 45 },
		{ 54, 283, 45 },
		{ 54, 284, 45 },
		{ 54, 285, 45 },
		{ 54, 286, 45 },
		{ 54, 287, 45 },
		{ 54, 288, 45 },
		{ 54, 289, 45 },
		{ 54, 290, 45 },
		{ 54, 291, 45 },
		{ 54, 292, 45 },
		{ 54, 293, 45 },
		{ 54, 294, 45 },
		{ 54, 295, 45 },
		{ 54, 296, 45 },
		{ 54, 297, 45 },
		{ 54, 298, 45 },
		{ 54, 299, 45 },
		{ 54, 300, 45 },
		{ 54, 301, 45 },
		{ 54, 302, 45 },
		{ 54, 303, 45 },
		{ 54, 304, 45 },
		{ 54, 305, 45 },
		{ 54, 306, 45 },
		{ 54, 307, 45 },
		{ 55, 258, 45 },
		{ 55, 259, 45 },
		{ 55, 260, 45 },
		{ 55, 261, 45 },
		{ 55, 262, 45 },
		{ 55, 263, 45 },
		{ 55, 264, 45 },
		{ 55, 265, 45 },
		{ 55, 266, 45 },
		{ 55, 267, 45 },
		{ 55, 268, 45 },
		{ 55, 269, 45 },
		{ 55, 270, 45 },
		{ 55, 271, 45 },
		{ 55, 272, 45 },
		{ 55, 273, 45 },
		{ 55, 274, 45 },
		{ 55, 275, 45 },
		{ 55, 276, 45 },
		{ 55, 277, 45 },
		{ 55, 278, 45 },
		{ 55, 279, 45 },
		{ 55, 280, 45 },
		{ 55, 281, 45 },
		{ 55, 282, 45 },
		{ 55, 283, 45 },
		{ 55, 284, 45 },
		{ 55, 285, 45 },
		{ 55, 286, 45 },
		{ 55, 287, 45 },
		{ 55, 288, 45 },
		{ 55, 289, 45 },
		{ 55, 290, 45 },
		{ 55, 291, 45 },
		{ 55, 292, 45 },
		{ 55, 293, 45 },
		{ 55, 294, 45 },
		{ 55, 295, 45 },
		{ 55, 296, 45 },
		{ 55, 297, 45 },
		{ 55, 298, 45 },
		{ 55, 299, 45 },
		{ 55, 300, 45 },
		{ 55, 301, 45 },
		{ 55, 302, 45 },
		{ 55, 303, 45 },
		{ 55, 304, 45 },
		{ 55, 305, 45 },
		{ 55, 306, 45 },
		{ 55, 307, 45 },
		{ 56, 258, 45 },
		{ 56, 259, 45 },
		{ 56, 260, 45 },
		{ 56, 261, 45 },
		{ 56, 262, 45 },
		{ 56, 263, 45 },
		{ 56, 264, 45 },
		{ 56, 265, 45 },
		{ 56, 266, 45 },
		{ 56, 267, 45 },
		{ 56, 268, 45 },
		{ 56, 269, 45 },
		{ 56, 270, 45 },
		{ 56, 271, 45 },
		{ 56, 272, 45 },
		{ 56, 273, 45 },
		{ 56, 274, 45 },
		{ 56, 275, 45 },
		{ 56, 276, 45 },
		{ 56, 277, 45 },
		{ 56, 278, 45 },
		{ 56, 279, 45 },
		{ 56, 280, 45 },
		{ 56, 281, 45 },
		{ 56, 282, 45 },
		{ 56, 283, 45 },
		{ 56, 284, 45 },
		{ 56, 285, 45 },
		{ 56, 286, 45 },
		{ 56, 287, 45 },
		{ 56, 288, 45 },
		{ 56, 289, 45 },
		{ 56, 290, 45 },
		{ 56, 291, 45 },
		{ 56, 292, 45 },
		{ 56, 293, 45 },
		{ 56, 294, 45 },
		{ 56, 295, 45 },
		{ 56, 296, 45 },
		{ 56, 297, 45 },
		{ 56, 298, 45 },
		{ 56, 299, 45 },
		{ 56, 300, 45 },
		{ 56, 301, 45 },
		{ 56, 302, 45 },
		{ 56, 303, 45 },
		{ 56, 304, 45 },
		{ 56, 305, 45 },
		{ 56, 306, 45 },
		{ 56, 307, 45 },
		{ 57, 258, 45 },
		{ 57, 259, 45 },
		{ 57, 260, 45 },
		{ 57, 261, 45 },
		{ 57, 262, 45 },
		{ 57, 263, 45 },
		{ 57, 264, 45 },
		{ 57, 265, 45 },
		{ 57, 266, 45 },
		{ 57, 267, 45 },
		{ 57, 268, 45 },
		{ 57, 269, 45 },
		{ 57, 270, 45 },
		{ 57, 271, 45 },
		{ 57, 272, 45 },
		{ 57, 273, 45 },
		{ 57, 274, 45 },
		{ 57, 275, 45 },
		{ 57, 276, 45 },
		{ 57, 277, 45 },
		{ 57, 278, 45 },
		{ 57, 279, 45 },
		{ 57, 280, 45 },
		{ 57, 281, 45 },
		{ 57, 282, 45 },
		{ 57, 283, 45 },
		{ 57, 284, 45 },
		{ 57, 285, 45 },
		{ 57, 286, 45 },
		{ 57, 287, 45 },
		{ 57, 288, 45 },
		{ 57, 289, 45 },
		{ 57, 290, 45 },
		{ 57, 291, 45 },
		{ 57, 292, 45 },
		{ 57, 293, 45 },
		{ 57, 294, 45 },
		{ 57, 295, 45 },
		{ 57, 296, 45 },
		{ 57, 297, 45 },
		{ 57, 298, 45 },
		{ 57, 299, 45 },
		{ 57, 300, 45 },
		{ 57, 301, 45 },
		{ 57, 302, 45 },
		{ 57, 303, 45 },
		{ 57, 304, 45 },
		{ 57, 305, 45 },
		{ 57, 306, 45 },
		{ 58, 258, 45 },
		{ 58, 259, 45 },
		{ 58, 260, 45 },
		{ 58, 261, 45 },
		{ 58, 262, 45 },
		{ 58, 263, 45 },
		{ 58, 264, 45 },
		{ 58, 265, 45 },
		{ 58, 266, 45 },
		{ 58, 267, 45 },
		{ 58, 268, 45 },
		{ 58, 269, 45 },
		{ 58, 270, 45 },
		{ 58, 271, 45 },
		{ 58, 272, 45 },
		{ 58, 273, 45 },
		{ 58, 274, 45 },
		{ 58, 275, 45 },
		{ 58, 276, 45 },
		{ 58, 277, 45 },
		{ 58, 278, 45 },
		{ 58, 279, 45 },
		{ 58, 280, 45 },
		{ 58, 281, 45 },
		{ 58, 282, 45 },
		{ 58, 283, 45 },
		{ 58, 284, 45 },
		{ 58, 285, 45 },
		{ 58, 286, 45 },
		{ 58, 287, 45 },
		{ 58, 288, 45 },
		{ 58, 289, 45 },
		{ 58, 290, 45 },
		{ 58, 291, 45 },
		{ 58, 292, 45 },
		{ 58, 293, 45 },
		{ 58, 294, 45 },
		{ 58, 295, 45 },
		{ 58, 296, 45 },
		{ 58, 297, 45 },
		{ 58, 298, 45 },
		{ 58, 299, 45 },
		{ 58, 300, 45 },
		{ 58, 301, 45 },
		{ 58, 302, 45 },
		{ 58, 303, 45 },
		{ 58, 304, 45 },
		{ 58, 305, 45 },
		{ 58, 306, 45 },
		{ 59, 258, 45 },
		{ 59, 259, 45 },
		{ 59, 260, 45 },
		{ 59, 261, 45 },
		{ 59, 262, 45 },
		{ 59, 263, 45 },
		{ 59, 264, 45 },
		{ 59, 265, 45 },
		{ 59, 266, 45 },
		{ 59, 267, 45 },
		{ 59, 268, 45 },
		{ 59, 269, 45 },
		{ 59, 270, 45 },
		{ 59, 271, 45 },
		{ 59, 272, 45 },
		{ 59, 273, 45 },
		{ 59, 274, 45 },
		{ 59, 275, 45 },
		{ 59, 276, 45 },
		{ 59, 277, 45 },
		{ 59, 278, 45 },
		{ 59, 279, 45 },
		{ 59, 280, 45 },
		{ 59, 281, 45 },
		{ 59, 282, 45 },
		{ 59, 283, 45 },
		{ 59, 284, 45 },
		{ 59, 285, 45 },
		{ 59, 286, 45 },
		{ 59, 287, 45 },
		{ 59, 288, 45 },
		{ 59, 289, 45 },
		{ 59, 290, 45 },
		{ 59, 291, 45 },
		{ 59, 292, 45 },
		{ 59, 293, 45 },
		{ 59, 294, 45 },
		{ 59, 295, 45 },
		{ 59, 296, 45 },
		{ 59, 297, 45 },
		{ 59, 298, 45 },
		{ 59, 299, 45 },
		{ 59, 300, 45 },
		{ 59, 301, 45 },
		{ 59, 302, 45 },
		{ 59, 303, 45 },
		{ 59, 304, 45 },
		{ 59, 305, 45 },
		{ 59, 306, 45 },
		{ 60, 258, 45 },
		{ 60, 259, 45 },
		{ 60, 260, 45 },
		{ 60, 261, 45 },
		{ 60, 262, 45 },
		{ 60, 263, 45 },
		{ 60, 264, 45 },
		{ 60, 265, 45 },
		{ 60, 266, 45 },
		{ 60, 267, 45 },
		{ 60, 268, 45 },
		{ 60, 269, 45 },
		{ 60, 270, 45 },
		{ 60, 271, 45 },
		{ 60, 272, 45 },
		{ 60, 273, 45 },
		{ 60, 274, 45 },
		{ 60, 275, 45 },
		{ 60, 276, 45 },
		{ 60, 277, 45 },
		{ 60, 278, 45 },
		{ 60, 279, 45 },
		{ 60, 280, 45 },
		{ 60, 281, 45 },
		{ 60, 282, 45 },
		{ 60, 283, 45 },
		{ 60, 284, 45 },
		{ 60, 285, 45 },
		{ 60, 286, 45 },
		{ 60, 287, 45 },
		{ 60, 288, 45 },
		{ 60, 289, 45 },
		{ 60, 290, 45 },
		{ 60, 291, 45 },
		{ 60, 292, 45 },
		{ 60, 293, 45 },
		{ 60, 294, 45 },
		{ 60, 295, 45 },
		{ 60, 296, 45 },
		{ 60, 297, 45 },
		{ 60, 298, 45 },
		{ 60, 299, 45 },
		{ 60, 300, 45 },
		{ 60, 301, 45 },
		{ 60, 302, 45 },
		{ 60, 303, 45 },
		{ 60, 304, 45 },
		{ 60, 305, 45 },
		{ 60, 306, 45 },
		{ 61, 258, 45 },
		{ 61, 259, 45 },
		{ 61, 260, 45 },
		{ 61, 261, 45 },
		{ 61, 262, 45 },
		{ 61, 263, 45 },
		{ 61, 264, 45 },
		{ 61, 265, 45 },
		{ 61, 266, 45 },
		{ 61, 267, 45 },
		{ 61, 268, 45 },
		{ 61, 269, 45 },
		{ 61, 270, 45 },
		{ 61, 271, 45 },
		{ 61, 272, 45 },
		{ 61, 273, 45 },
		{ 61, 274, 45 },
		{ 61, 275, 45 },
		{ 61, 276, 45 },
		{ 61, 277, 45 },
		{ 61, 278, 45 },
		{ 61, 279, 45 },
		{ 61, 280, 45 },
		{ 61, 281, 45 },
		{ 61, 282, 45 },
		{ 61, 283, 45 },
		{ 61, 284, 45 },
		{ 61, 285, 45 },
		{ 61, 286, 45 },
		{ 61, 287, 45 },
		{ 61, 288, 45 },
		{ 61, 289, 45 },
		{ 61, 290, 45 },
		{ 61, 291, 45 },
		{ 61, 292, 45 },
		{ 61, 293, 45 },
		{ 61, 294, 45 },
		{ 61, 295, 45 },
		{ 61, 296, 45 },
		{ 61, 297, 45 },
		{ 61, 298, 45 },
		{ 61, 299, 45 },
		{ 61, 300, 45 },
		{ 61, 301, 45 },
		{ 61, 302, 45 },
		{ 61, 303, 45 },
		{ 61, 304, 45 },
		{ 61, 305, 45 },
		{ 62, 258, 45 },
		{ 62, 259, 45 },
		{ 62, 260, 45 },
		{ 62, 261, 45 },
		{ 62, 262, 45 },
		{ 62, 263, 45 },
		{ 62, 264, 45 },
		{ 62, 265, 45 },
		{ 62, 266, 45 },
		{ 62, 267, 45 },
		{ 62, 268, 45 },
		{ 62, 269, 45 },
		{ 62, 270, 45 },
		{ 62, 271, 45 },
		{ 62, 272, 45 },
		{ 62, 273, 45 },
		{ 62, 274, 45 },
		{ 62, 275, 45 },
		{ 62, 276, 45 },
		{ 62, 277, 45 },
		{ 62, 278, 45 },
		{ 62, 279, 45 },
		{ 62, 280, 45 },
		{ 62, 281, 45 },
		{ 62, 282, 45 },
		{ 62, 283, 45 },
		{ 62, 284, 45 },
		{ 62, 285, 45 },
		{ 62, 286, 45 },
		{ 62, 287, 45 },
		{ 62, 288, 45 },
		{ 62, 289, 45 },
		{ 62, 290, 45 },
		{ 62, 291, 45 },
		{ 62, 292, 45 },
		{ 62, 293, 45 },
		{ 62, 294, 45 },
		{ 62, 295, 45 },
		{ 62, 296, 45 },
		{ 62, 297, 45 },
		{ 62, 298, 45 },
		{ 62, 299, 45 },
		{ 62, 300, 45 },
		{ 62, 301, 45 },
		{ 62, 302, 45 },
		{ 62, 303, 45 },
		{ 62, 304, 45 },
		{ 62, 305, 45 },
		{ 63, 258, 45 },
		{ 63, 259, 45 },
		{ 63, 260, 45 },
		{ 63, 261, 45 },
		{ 63, 262, 45 },
		{ 63, 263, 45 },
		{ 63, 264, 45 },
		{ 63, 265, 45 },
		{ 63, 266, 45 },
		{ 63, 267, 45 },
		{ 63, 268, 45 },
		{ 63, 269, 45 },
		{ 63, 270, 45 },
		{ 63, 271, 45 },
		{ 63, 272, 45 },
		{ 63, 273, 45 },
		{ 63, 274, 45 },
		{ 63, 275, 45 },
		{ 63, 276, 45 },
		{ 63, 277, 45 },
		{ 63, 278, 45 },
		{ 63, 279, 45 },
		{ 63, 280, 45 },
		{ 63, 281, 45 },
		{ 63, 282, 45 },
		{ 63, 283, 45 },
		{ 63, 284, 45 },
		{ 63, 285, 45 },
		{ 63, 286, 45 },
		{ 63, 287, 45 },
		{ 63, 288, 45 },
		{ 63, 289, 45 },
		{ 63, 290, 45 },
		{ 63, 291, 45 },
		{ 63, 292, 45 },
		{ 63, 293, 45 },
		{ 63, 294, 45 },
		{ 63, 295, 45 },
		{ 63, 296, 45 },
		{ 63, 297, 45 },
		{ 63, 298, 45 },
		{ 63, 299, 45 },
		{ 63, 300, 45 },
		{ 63, 301, 45 },
		{ 63, 302, 45 },
		{ 63, 303, 45 },
		{ 63, 304, 45 },
		{ 63, 305, 45 },
		{ 64, 258, 45 },
		{ 64, 259, 45 },
		{ 64, 260, 45 },
		{ 64, 261, 45 },
		{ 64, 262, 45 },
		{ 64, 263, 45 },
		{ 64, 264, 45 },
		{ 64, 265, 45 },
		{ 64, 266, 45 },
		{ 64, 267, 45 },
		{ 64, 268, 45 },
		{ 64, 269, 45 },
		{ 64, 270, 45 },
		{ 64, 271, 45 },
		{ 64, 272, 45 },
		{ 64, 273, 45 },
		{ 64, 274, 45 },
		{ 64, 275, 45 },
		{ 64, 276, 45 },
		{ 64, 277, 45 },
		{ 64, 278, 45 },
		{ 64, 279, 45 },
		{ 64, 280, 45 },
		{ 64, 281, 45 },
		{ 64, 282, 45 },
		{ 64, 283, 45 },
		{ 64, 284, 45 },
		{ 64, 285, 45 },
		{ 64, 286, 45 },
		{ 64, 287, 45 },
		{ 64, 288, 45 },
		{ 64, 289, 45 },
		{ 64, 290, 45 },
		{ 64, 291, 45 },
		{ 64, 292, 45 },
		{ 64, 293, 45 },
		{ 64, 294, 45 },
		{ 64, 295, 45 },
		{ 64, 296, 45 },
		{ 64, 297, 45 },
		{ 64, 298, 45 },
		{ 64, 299, 45 },
		{ 64, 300, 45 },
		{ 64, 301, 45 },
		{ 64, 302, 45 },
		{ 64, 303, 45 },
		{ 64, 304, 45 },
		{ 64, 305, 45 },
		{ 65, 258, 45 },
		{ 65, 259, 45 },
		{ 65, 260, 45 },
		{ 65, 261, 45 },
		{ 65, 262, 45 },
		{ 65, 263, 45 },
		{ 65, 264, 45 },
		{ 65, 265, 45 },
		{ 65, 266, 45 },
		{ 65, 267, 45 },
		{ 65, 268, 45 },
		{ 65, 269, 45 },
		{ 65, 270, 45 },
		{ 65, 271, 45 },
		{ 65, 272, 45 },
		{ 65, 273, 45 },
		{ 65, 274, 45 },
		{ 65, 275, 45 },
		{ 65, 276, 45 },
		{ 65, 277, 45 },
		{ 65, 278, 45 },
		{ 65, 279, 45 },
		{ 65, 280, 45 },
		{ 65, 281, 45 },
		{ 65, 282, 45 },
		{ 65, 283, 45 },
		{ 65, 284, 45 },
		{ 65, 285, 45 },
		{ 65, 286, 45 },
		{ 65, 287, 45 },
		{ 65, 288, 45 },
		{ 65, 289, 45 },
		{ 65, 290, 45 },
		{ 65, 291, 45 },
		{ 65, 292, 45 },
		{ 65, 293, 45 },
		{ 65, 294, 45 },
		{ 65, 295, 45 },
		{ 65, 296, 45 },
		{ 65, 297, 45 },
		{ 65, 298, 45 },
		{ 65, 299, 45 },
		{ 65, 300, 45 },
		{ 65, 301, 45 },
		{ 65, 302, 45 },
		{ 65, 303, 45 },
		{ 65, 304, 45 },
		{ 66, 258, 45 },
		{ 66, 259, 45 },
		{ 66, 260, 45 },
		{ 66, 261, 45 },
		{ 66, 262, 45 },
		{ 66, 263, 45 },
		{ 66, 264, 45 },
		{ 66, 265, 45 },
		{ 66, 266, 45 },
		{ 66, 267, 45 },
		{ 66, 268, 45 },
		{ 66, 269, 45 },
		{ 66, 270, 45 },
		{ 66, 271, 45 },
		{ 66, 272, 45 },
		{ 66, 273, 45 },
		{ 66, 274, 45 },
		{ 66, 275, 45 },
		{ 66, 276, 45 },
		{ 66, 277, 45 },
		{ 66, 278, 45 },
		{ 66, 279, 45 },
		{ 66, 280, 45 },
		{ 66, 281, 45 },
		{ 66, 282, 45 },
		{ 66, 283, 45 },
		{ 66, 284, 45 },
		{ 66, 285, 45 },
		{ 66, 286, 45 },
		{ 66, 287, 45 },
		{ 66, 288, 45 },
		{ 66, 289, 45 },
		{ 66, 290, 45 },
		{ 66, 291, 45 },
		{ 66, 292, 45 },
		{ 66, 293, 45 },
		{ 66, 294, 45 },
		{ 66, 295, 45 },
		{ 66, 296, 45 },
		{ 66, 297, 45 },
		{ 66, 298, 45 },
		{ 66, 299, 45 },
		{ 66, 300, 45 },
		{ 66, 301, 45 },
		{ 66, 302, 45 },
		{ 66, 303, 45 },
		{ 66, 304, 45 },
		{ 67, 258, 45 },
		{ 67, 259, 45 },
		{ 67, 260, 45 },
		{ 67, 261, 45 },
		{ 67, 262, 45 },
		{ 67, 263, 45 },
		{ 67, 264, 45 },
		{ 67, 265, 45 },
		{ 67, 266, 45 },
		{ 67, 267, 45 },
		{ 67, 268, 45 },
		{ 67, 269, 45 },
		{ 67, 270, 45 },
		{ 67, 271, 45 },
		{ 67, 272, 45 },
		{ 67, 273, 45 },
		{ 67, 274, 45 },
		{ 67, 275, 45 },
		{ 67, 276, 45 },
		{ 67, 277, 45 },
		{ 67, 278, 45 },
		{ 67, 279, 45 },
		{ 67, 280, 45 },
		{ 67, 281, 45 },
		{ 67, 282, 45 },
		{ 67, 283, 45 },
		{ 67, 284, 45 },
		{ 67, 285, 45 },
		{ 67, 286, 45 },
		{ 67, 287, 45 },
		{ 67, 288, 45 },
		{ 67, 289, 45 },
		{ 67, 290, 45 },
		{ 67, 291, 45 },
		{ 67, 292, 45 },
		{ 67, 293, 45 },
		{ 67, 294, 45 },
		{ 67, 295, 45 },
		{ 67, 296, 45 },
		{ 67, 297, 45 },
		{ 67, 298, 45 },
		{ 67, 299, 45 },
		{ 67, 300, 45 },
		{ 67, 301, 45 },
		{ 67, 302, 45 },
		{ 67, 303, 45 },
		{ 67, 304, 45 },
		{ 68, 258, 45 },
		{ 68, 259, 45 },
		{ 68, 260, 45 },
		{ 68, 261, 45 },
		{ 68, 262, 45 },
		{ 68, 263, 45 },
		{ 68, 264, 45 },
		{ 68, 265, 45 },
		{ 68, 266, 45 },
		{ 68, 267, 45 },
		{ 68, 268, 45 },
		{ 68, 269, 45 },
		{ 68, 270, 45 },
		{ 68, 271, 45 },
		{ 68, 272, 45 },
		{ 68, 273, 45 },
		{ 68, 274, 45 },
		{ 68, 275, 45 },
		{ 68, 276, 45 },
		{ 68, 277, 45 },
		{ 68, 278, 45 },
		{ 68, 279, 45 },
		{ 68, 280, 45 },
		{ 68, 281, 45 },
		{ 68, 282, 45 },
		{ 68, 283, 45 },
		{ 68, 284, 45 },
		{ 68, 285, 45 },
		{ 68, 286, 45 },
		{ 68, 287, 45 },
		{ 68, 288, 45 },
		{ 68, 289, 45 },
		{ 68, 290, 45 },
		{ 68, 291, 45 },
		{ 68, 292, 45 },
		{ 68, 293, 45 },
		{ 68, 294, 45 },
		{ 68, 295, 45 },
		{ 68, 296, 45 },
		{ 68, 297, 45 },
		{ 68, 298, 45 },
		{ 68, 299, 45 },
		{ 68, 300, 45 },
		{ 68, 301, 45 },
		{ 68, 302, 45 },
		{ 68, 303, 45 },
		{ 68, 304, 45 },
		{ 69, 258, 45 },
		{ 69, 259, 45 },
		{ 69, 260, 45 },
		{ 69, 261, 45 },
		{ 69, 262, 45 },
		{ 69, 263, 45 },
		{ 69, 264, 45 },
		{ 69, 265, 45 },
		{ 69, 266, 45 },
		{ 69, 267, 45 },
		{ 69, 268, 45 },
		{ 69, 269, 45 },
		{ 69, 270, 45 },
		{ 69, 271, 45 },
		{ 69, 272, 45 },
		{ 69, 273, 45 },
		{ 69, 274, 45 },
		{ 69, 275, 45 },
		{ 69, 276, 45 },
		{ 69, 277, 45 },
		{ 69, 278, 45 },
		{ 69, 279, 45 },
		{ 69, 280, 45 },
		{ 69, 281, 45 },
		{ 69, 282, 45 },
		{ 69, 283, 45 },
		{ 69, 284, 45 },
		{ 69, 285, 45 },
		{ 69, 286, 45 },
		{ 69, 287, 45 },
		{ 69, 288, 45 },
		{ 69, 289, 45 },
		{ 69, 290, 45 },
		{ 69, 291, 45 },
		{ 69, 292, 45 },
		{ 69, 293, 45 },
		{ 69, 294, 45 },
		{ 69, 295, 45 },
		{ 69, 296, 45 },
		{ 69, 297, 45 },
		{ 69, 298, 45 },
		{ 69, 299, 45 },
		{ 69, 300, 45 },
		{ 69, 301, 45 },
		{ 69, 302, 45 },
		{ 69, 303, 45 },
		{ 70, 258, 45 },
		{ 70, 259, 45 },
		{ 70, 260, 45 },
		{ 70, 261, 45 },
		{ 70, 262, 45 },
		{ 70, 263, 45 },
		{ 70, 264, 45 },
		{ 70, 265, 45 },
		{ 70, 266, 45 },
		{ 70, 267, 45 },
		{ 70, 268, 45 },
		{ 70, 269, 45 },
		{ 70, 270, 45 },
		{ 70, 271, 45 },
		{ 70, 272, 45 },
		{ 70, 273, 45 },
		{ 70, 274, 45 },
		{ 70, 275, 45 },
		{ 70, 276, 45 },
		{ 70, 277, 45 },
		{ 70, 278, 45 },
		{ 70, 279, 45 },
		{ 70, 280, 45 },
		{ 70, 281, 45 },
		{ 70, 282, 45 },
		{ 70, 283, 45 },
		{ 70, 284, 45 },
		{ 70, 285, 45 },
		{ 70, 286, 45 },
		{ 70, 287, 45 },
		{ 70, 288, 45 },
		{ 70, 289, 45 },
		{ 70, 290, 45 },
		{ 70, 291, 45 },
		{ 70, 292, 45 },
		{ 70, 293, 45 },
		{ 70, 294, 45 },
		{ 70, 295, 45 },
		{ 70, 296, 45 },
		{ 70, 297, 45 },
		{ 70, 298, 45 },
		{ 70, 299, 45 },
		{ 70, 300, 45 },
		{ 70, 301, 45 },
		{ 70, 302, 45 },
		{ 70, 303, 45 },
		{ 71, 258, 45 },
		{ 71, 259, 45 },
		{ 71, 260, 45 },
		{ 71, 261, 45 },
		{ 71, 262, 45 },
		{ 71, 263, 45 },
		{ 71, 264, 45 },
		{ 71, 265, 45 },
		{ 71, 266, 45 },
		{ 71, 267, 45 },
		{ 71, 268, 45 },
		{ 71, 269, 45 },
		{ 71, 270, 45 },
		{ 71, 271, 45 },
		{ 71, 272, 45 },
		{ 71, 273, 45 },
		{ 71, 274, 45 },
		{ 71, 275, 45 },
		{ 71, 276, 45 },
		{ 71, 277, 45 },
		{ 71, 278, 45 },
		{ 71, 279, 45 },
		{ 71, 280, 45 },
		{ 71, 281, 45 },
		{ 71, 282, 45 },
		{ 71, 283, 45 },
		{ 71, 284, 45 },
		{ 71, 285, 45 },
		{ 71, 286, 45 },
		{ 71, 287, 45 },
		{ 71, 288, 45 },
		{ 71, 289, 45 },
		{ 71, 290, 45 },
		{ 71, 291, 45 },
		{ 71, 292, 45 },
		{ 71, 293, 45 },
		{ 71, 294, 45 },
		{ 71, 295, 45 },
		{ 71, 296, 45 },
		{ 71, 297, 45 },
		{ 71, 298, 45 },
		{ 71, 299, 45 },
		{ 71, 300, 45 },
		{ 71, 301, 45 },
		{ 71, 302, 45 },
		{ 71, 303, 45 },
		{ 72, 258, 45 },
		{ 72, 259, 45 },
		{ 72, 260, 45 },
		{ 72, 261, 45 },
		{ 72, 262, 45 },
		{ 72, 263, 45 },
		{ 72, 264, 45 },
		{ 72, 265, 45 },
		{ 72, 266, 45 },
		{ 72, 267, 45 },
		{ 72, 268, 45 },
		{ 72, 269, 45 },
		{ 72, 270, 45 },
		{ 72, 271, 45 },
		{ 72, 272, 45 },
		{ 72, 273, 45 },
		{ 72, 274, 45 },
		{ 72, 275, 45 },
		{ 72, 276, 45 },
		{ 72, 277, 45 },
		{ 72, 278, 45 },
		{ 72, 279, 45 },
		{ 72, 280, 45 },
		{ 72, 281, 45 },
		{ 72, 282, 45 },
		{ 72, 283, 45 },
		{ 72, 284, 45 },
		{ 72, 285, 45 },
		{ 72, 286, 45 },
		{ 72, 287, 45 },
		{ 72, 288, 45 },
		{ 72, 289, 45 },
		{ 72, 290, 45 },
		{ 72, 291, 45 },
		{ 72, 292, 45 },
		{ 72, 293, 45 },
		{ 72, 294, 45 },
		{ 72, 295, 45 },
		{ 72, 296, 45 },
		{ 72, 297, 45 },
		{ 72, 298, 45 },
		{ 72, 299, 45 },
		{ 72, 300, 45 },
		{ 72, 301, 45 },
		{ 72, 302, 45 },
		{ 72, 303, 45 },
		{ 73, 258, 45 },
		{ 73, 259, 45 },
		{ 73, 260, 45 },
		{ 73, 261, 45 },
		{ 73, 262, 45 },
		{ 73, 263, 45 },
		{ 73, 264, 45 },
		{ 73, 265, 45 },
		{ 73, 266, 45 },
		{ 73, 267, 45 },
		{ 73, 268, 45 },
		{ 73, 269, 45 },
		{ 73, 270, 45 },
		{ 73, 271, 45 },
		{ 73, 272, 45 },
		{ 73, 273, 45 },
		{ 73, 274, 45 },
		{ 73, 275, 45 },
		{ 73, 276, 45 },
		{ 73, 277, 45 },
		{ 73, 278, 45 },
		{ 73, 279, 45 },
		{ 73, 280, 45 },
		{ 73, 281, 45 },
		{ 73, 282, 45 },
		{ 73, 283, 45 },
		{ 73, 284, 45 },
		{ 73, 285, 45 },
		{ 73, 286, 45 },
		{ 73, 287, 45 },
		{ 73, 288, 45 },
		{ 73, 289, 45 },
		{ 73, 290, 45 },
		{ 73, 291, 45 },
		{ 73, 292, 45 },
		{ 73, 293, 45 },
		{ 73, 294, 45 },
		{ 73, 295, 45 },
		{ 73, 296, 45 },
		{ 73, 297, 45 },
		{ 73, 298, 45 },
		{ 73, 299, 45 },
		{ 73, 300, 45 },
		{ 73, 301, 45 },
		{ 73, 302, 45 },
		{ 74, 258, 45 },
		{ 74, 259, 45 },
		{ 74, 260, 45 },
		{ 74, 261, 45 },
		{ 74, 262, 45 },
		{ 74, 263, 45 },
		{ 74, 264, 45 },
		{ 74, 265, 45 },
		{ 74, 266, 45 },
		{ 74, 267, 45 },
		{ 74, 268, 45 },
		{ 74, 269, 45 },
		{ 74, 270, 45 },
		{ 74, 271, 45 },
		{ 74, 272, 45 },
		{ 74, 273, 45 },
		{ 74, 274, 45 },
		{ 74, 275, 45 },
		{ 74, 276, 45 },
		{ 74, 277, 45 },
		{ 74, 278, 45 },
		{ 74, 279, 45 },
		{ 74, 280, 45 },
		{ 74, 281, 45 },
		{ 74, 282, 45 },
		{ 74, 283, 45 },
		{ 74, 284, 45 },
		{ 74, 285, 45 },
		{ 74, 286, 45 },
		{ 74, 287, 45 },
		{ 74, 288, 45 },
		{ 74, 289, 45 },
		{ 74, 290, 45 },
		{ 74, 291, 45 },
		{ 74, 292, 45 },
		{ 74, 293, 45 },
		{ 74, 294, 45 },
		{ 74, 295, 45 },
		{ 74, 296, 45 },
		{ 74, 297, 45 },
		{ 74, 298, 45 },
		{ 74, 299, 45 },
		{ 74, 300, 45 },
		{ 74, 301, 45 },
		{ 74, 302, 45 },
		{ 75, 258, 45 },
		{ 75, 259, 45 },
		{ 75, 260, 45 },
		{ 75, 261, 45 },
		{ 75, 262, 45 },
		{ 75, 263, 45 },
		{ 75, 264, 45 },
		{ 75, 265, 45 },
		{ 75, 266, 45 },
		{ 75, 267, 45 },
		{ 75, 268, 45 },
		{ 75, 269, 45 },
		{ 75, 270, 45 },
		{ 75, 271, 45 },
		{ 75, 272, 45 },
		{ 75, 273, 45 },
		{ 75, 274, 45 },
		{ 75, 275, 45 },
		{ 75, 276, 45 },
		{ 75, 277, 45 },
		{ 75, 278, 45 },
		{ 75, 279, 45 },
		{ 75, 280, 45 },
		{ 75, 281, 45 },
		{ 75, 282, 45 },
		{ 75, 283, 45 },
		{ 75, 284, 45 },
		{ 75, 285, 45 },
		{ 75, 286, 45 },
		{ 75, 287, 45 },
		{ 75, 288, 45 },
		{ 75, 289, 45 },
		{ 75, 290, 45 },
		{ 75, 291, 45 },
		{ 75, 292, 45 },
		{ 75, 293, 45 },
		{ 75, 294, 45 },
		{ 75, 295, 45 },
		{ 75, 296, 45 },
		{ 75, 297, 45 },
		{ 75, 298, 45 },
		{ 75, 299, 45 },
		{ 75, 300, 45 },
		{ 75, 301, 45 },
		{ 75, 302, 45 },
		{ 76, 258, 45 },
		{ 76, 259, 45 },
		{ 76, 260, 45 },
		{ 76, 261, 45 },
		{ 76, 262, 45 },
		{ 76, 263, 45 },
		{ 76, 264, 45 },
		{ 76, 265, 45 },
		{ 76, 266, 45 },
		{ 76, 267, 45 },
		{ 76, 268, 45 },
		{ 76, 269, 45 },
		{ 76, 270, 45 },
		{ 76, 271, 45 },
		{ 76, 272, 45 },
		{ 76, 273, 45 },
		{ 76, 274, 45 },
		{ 76, 275, 45 },
		{ 76, 276, 45 },
		{ 76, 277, 45 },
		{ 76, 278, 45 },
		{ 76, 279, 45 },
		{ 76, 280, 45 },
		{ 76, 281, 45 },
		{ 76, 282, 45 },
		{ 76, 283, 45 },
		{ 76, 284, 45 },
		{ 76, 285, 45 },
		{ 76, 286, 45 },
		{ 76, 287, 45 },
		{ 76, 288, 45 },
		{ 76, 289, 45 },
		{ 76, 290, 45 },
		{ 76, 291, 45 },
		{ 76, 292, 45 },
		{ 76, 293, 45 },
		{ 76, 294, 45 },
		{ 76, 295, 45 },
		{ 76, 296, 45 },
		{ 76, 297, 45 },
		{ 76, 298, 45 },
		{ 76, 299, 45 },
		{ 76, 300, 45 },
		{ 76, 301, 45 },
		{ 76, 302, 45 },
		{ 77, 258, 45 },
		{ 77, 259, 45 },
		{ 77, 260, 45 },
		{ 77, 261, 45 },
		{ 77, 262, 45 },
		{ 77, 263, 45 },
		{ 77, 264, 45 },
		{ 77, 265, 45 },
		{ 77, 266, 45 },
		{ 77, 267, 45 },
		{ 77, 268, 45 },
		{ 77, 269, 45 },
		{ 77, 270, 45 },
		{ 77, 271, 45 },
		{ 77, 272, 45 },
		{ 77, 273, 45 },
		{ 77, 274, 45 },
		{ 77, 275, 45 },
		{ 77, 276, 45 },
		{ 77, 277, 45 },
		{ 77, 278, 45 },
		{ 77, 279, 45 },
		{ 77, 280, 45 },
		{ 77, 281, 45 },
		{ 77, 282, 45 },
		{ 77, 283, 45 },
		{ 77, 284, 45 },
		{ 77, 285, 45 },
		{ 77, 286, 45 },
		{ 77, 287, 45 },
		{ 77, 288, 45 },
		{ 77, 289, 45 },
		{ 77, 290, 45 },
		{ 77, 291, 45 },
		{ 77, 292, 45 },
		{ 77, 293, 45 },
		{ 77, 294, 45 },
		{ 77, 295, 45 },
		{ 77, 296, 45 },
		{ 77, 297, 45 },
		{ 77, 298, 45 },
		{ 77, 299, 45 },
		{ 77, 300, 45 },
		{ 77, 301, 45 },
		{ 78, 258, 45 },
		{ 78, 259, 45 },
		{ 78, 260, 45 },
		{ 78, 261, 45 },
		{ 78, 262, 45 },
		{ 78, 263, 45 },
		{ 78, 264, 45 },
		{ 78, 265, 45 },
		{ 78, 266, 45 },
		{ 78, 267, 45 },
		{ 78, 268, 45 },
		{ 78, 269, 45 },
		{ 78, 270, 45 },
		{ 78, 271, 45 },
		{ 78, 272, 45 },
		{ 78, 273, 45 },
		{ 78, 274, 45 },
		{ 78, 275, 45 },
		{ 78, 276, 45 },
		{ 78, 277, 45 },
		{ 78, 278, 45 },
		{ 78, 279, 45 },
		{ 78, 280, 45 },
		{ 78, 281, 45 },
		{ 78, 282, 45 },
		{ 78, 283, 45 },
		{ 78, 284, 45 },
		{ 78, 285, 45 },
		{ 78, 286, 45 },
		{ 78, 287, 45 },
		{ 78, 288, 45 },
		{ 78, 289, 45 },
		{ 78, 290, 45 },
		{ 78, 291, 45 },
		{ 78, 292, 45 },
		{ 78, 293, 45 },
		{ 78, 294, 45 },
		{ 78, 295, 45 },
		{ 78, 296, 45 },
		{ 78, 297, 45 },
		{ 78, 298, 45 },
		{ 78, 299, 45 },
		{ 78, 300, 45 },
		{ 78, 301, 45 },
		{ 79, 258, 45 },
		{ 79, 259, 45 },
		{ 79, 260, 45 },
		{ 79, 261, 45 },
		{ 79, 262, 45 },
		{ 79, 263, 45 },
		{ 79, 264, 45 },
		{ 79, 265, 45 },
		{ 79, 266, 45 },
		{ 79, 267, 45 },
		{ 79, 268, 45 },
		{ 79, 269, 45 },
		{ 79, 270, 45 },
		{ 79, 271, 45 },
		{ 79, 272, 45 },
		{ 79, 273, 45 },
		{ 79, 274, 45 },
		{ 79, 275, 45 },
		{ 79, 276, 45 },
		{ 79, 277, 45 },
		{ 79, 278, 45 },
		{ 79, 279, 45 },
		{ 79, 280, 45 },
		{ 79, 281, 45 },
		{ 79, 282, 45 },
		{ 79, 283, 45 },
		{ 79, 284, 45 },
		{ 79, 285, 45 },
		{ 79, 286, 45 },
		{ 79, 287, 45 },
		{ 79, 288, 45 },
		{ 79, 289, 45 },
		{ 79, 290, 45 },
		{ 79, 291, 45 },
		{ 79, 292, 45 },
		{ 79, 293, 45 },
		{ 79, 294, 45 },
		{ 79, 295, 45 },
		{ 79, 296, 45 },
		{ 79, 297, 45 },
		{ 79, 298, 45 },
		{ 79, 299, 45 },
		{ 79, 300, 45 },
		{ 79, 301, 45 },
		{ 80, 258, 45 },
		{ 80, 259, 45 },
		{ 80, 260, 45 },
		{ 80, 261, 45 },
		{ 80, 262, 45 },
		{ 80, 263, 45 },
		{ 80, 264, 45 },
		{ 80, 265, 45 },
		{ 80, 266, 45 },
		{ 80, 267, 45 },
		{ 80, 268, 45 },
		{ 80, 269, 45 },
		{ 80, 270, 45 },
		{ 80, 271, 45 },
		{ 80, 272, 45 },
		{ 80, 273, 45 },
		{ 80, 274, 45 },
		{ 80, 275, 45 },
		{ 80, 276, 45 },
		{ 80, 277, 45 },
		{ 80, 278, 45 },
		{ 80, 279, 45 },
		{ 80, 280, 45 },
		{ 80, 281, 45 },
		{ 80, 282, 45 },
		{ 80, 283, 45 },
		{ 80, 284, 45 },
		{ 80, 285, 45 },
		{ 80, 286, 45 },
		{ 80, 287, 45 },
		{ 80, 288, 45 },
		{ 80, 289, 45 },
		{ 80, 290, 45 },
		{ 80, 291, 45 },
		{ 80, 292, 45 },
		{ 80, 293, 45 },
		{ 80, 294, 45 },
		{ 80, 295, 45 },
		{ 80, 296, 45 },
		{ 80, 297, 45 },
		{ 80, 298, 45 },
		{ 80, 299, 45 },
		{ 80, 300, 45 },
		{ 80, 301, 45 },
		{ 81, 258, 45 },
		{ 81, 259, 45 },
		{ 81, 260, 45 },
		{ 81, 261, 45 },
		{ 81, 262, 45 },
		{ 81, 263, 45 },
		{ 81, 264, 45 },
		{ 81, 265, 45 },
		{ 81, 266, 45 },
		{ 81, 267, 45 },
		{ 81, 268, 45 },
		{ 81, 269, 45 },
		{ 81, 270, 45 },
		{ 81, 271, 45 },
		{ 81, 272, 45 },
		{ 81, 273, 45 },
		{ 81, 274, 45 },
		{ 81, 275, 45 },
		{ 81, 276, 45 },
		{ 81, 277, 45 },
		{ 81, 278, 45 },
		{ 81, 279, 45 },
		{ 81, 280, 45 },
		{ 81, 281, 45 },
		{ 81, 282, 45 },
		{ 81, 283, 45 },
		{ 81, 284, 45 },
		{ 81, 285, 45 },
		{ 81, 286, 45 },
		{ 81, 287, 45 },
		{ 81, 288, 45 },
		{ 81, 289, 45 },
		{ 81, 290, 45 },
		{ 81, 291, 45 },
		{ 81, 292, 45 },
		{ 81, 293, 45 },
		{ 81, 294, 45 },
		{ 81, 295, 45 },
		{ 81, 296, 45 },
		{ 81, 297, 45 },
		{ 81, 298, 45 },
		{ 81, 299, 45 },
		{ 81, 300, 45 },
		{ 82, 258, 45 },
		{ 82, 259, 45 },
		{ 82, 260, 45 },
		{ 82, 261, 45 },
		{ 82, 262, 45 },
		{ 82, 263, 45 },
		{ 82, 264, 45 },
		{ 82, 265, 45 },
		{ 82, 266, 45 },
		{ 82, 267, 45 },
		{ 82, 268, 45 },
		{ 82, 269, 45 },
		{ 82, 270, 45 },
		{ 82, 271, 45 },
		{ 82, 272, 45 },
		{ 82, 273, 45 },
		{ 82, 274, 45 },
		{ 82, 275, 45 },
		{ 82, 276, 45 },
		{ 82, 277, 45 },
		{ 82, 278, 45 },
		{ 82, 279, 45 },
		{ 82, 280, 45 },
		{ 82, 281, 45 },
		{ 82, 282, 45 },
		{ 82, 283, 45 },
		{ 82, 284, 45 },
		{ 82, 285, 45 },
		{ 82, 286, 45 },
		{ 82, 287, 45 },
		{ 82, 288, 45 },
		{ 82, 289, 45 },
		{ 82, 290, 45 },
		{ 82, 291, 45 },
		{ 82, 292, 45 },
		{ 82, 293, 45 },
		{ 82, 294, 45 },
		{ 82, 295, 45 },
		{ 82, 296, 45 },
		{ 82, 297, 45 },
		{ 82, 298, 45 },
		{ 82, 299, 45 },
		{ 82, 300, 45 },
		{ 83, 258, 45 },
		{ 83, 259, 45 },
		{ 83, 260, 45 },
		{ 83, 261, 45 },
		{ 83, 262, 45 },
		{ 83, 263, 45 },
		{ 83, 264, 45 },
		{ 83, 265, 45 },
		{ 83, 266, 45 },
		{ 83, 267, 45 },
		{ 83, 268, 45 },
		{ 83, 269, 45 },
		{ 83, 270, 45 },
		{ 83, 271, 45 },
		{ 83, 272, 45 },
		{ 83, 273, 45 },
		{ 83, 274, 45 },
		{ 83, 275, 45 },
		{ 83, 276, 45 },
		{ 83, 277, 45 },
		{ 83, 278, 45 },
		{ 83, 279, 45 },
		{ 83, 280, 45 },
		{ 83, 281, 45 },
		{ 83, 282, 45 },
		{ 83, 283, 45 },
		{ 83, 284, 45 },
		{ 83, 285, 45 },
		{ 83, 286, 45 },
		{ 83, 287, 45 },
		{ 83, 288, 45 },
		{ 83, 289, 45 },
		{ 83, 290, 45 },
		{ 83, 291, 45 },
		{ 83, 292, 45 },
		{ 83, 293, 45 },
		{ 83, 294, 45 },
		{ 83, 295, 45 },
		{ 83, 296, 45 },
		{ 83, 297, 45 },
		{ 83, 298, 45 },
		{ 83, 299, 45 },
		{ 83, 300, 45 },
		{ 84, 258, 45 },
		{ 84, 259, 45 },
		{ 84, 260, 45 },
		{ 84, 261, 45 },
		{ 84, 262, 45 },
		{ 84, 263, 45 },
		{ 84, 264, 45 },
		{ 84, 265, 45 },
		{ 84, 266, 45 },
		{ 84, 267, 45 },
		{ 84, 268, 45 },
		{ 84, 269, 45 },
		{ 84, 270, 45 },
		{ 84, 271, 45 },
		{ 84, 272, 45 },
		{ 84, 273, 45 },
		{ 84, 274, 45 },
		{ 84, 275, 45 },
		{ 84, 276, 45 },
		{ 84, 277, 45 },
		{ 84, 278, 45 },
		{ 84, 279, 45 },
		{ 84, 280, 45 },
		{ 84, 281, 45 },
		{ 84, 282, 45 },
		{ 84, 283, 45 },
		{ 84, 284, 45 },
		{ 84, 285, 45 },
		{ 84, 286, 45 },
		{ 84, 287, 45 },
		{ 84, 288, 45 },
		{ 84, 289, 45 },
		{ 84, 290, 45 },
		{ 84, 291, 45 },
		{ 84, 292, 45 },
		{ 84, 293, 45 },
		{ 84, 294, 45 },
		{ 84, 295, 45 },
		{ 84, 296, 45 },
		{ 84, 297, 45 },
		{ 84, 298, 45 },
		{ 84, 299, 45 },
		{ 84, 300, 45 },
		{ 85, 258, 45 },
		{ 85, 259, 45 },
		{ 85, 260, 45 },
		{ 85, 261, 45 },
		{ 85, 262, 45 },
		{ 85, 263, 45 },
		{ 85, 264, 45 },
		{ 85, 265, 45 },
		{ 85, 266, 45 },
		{ 85, 267, 45 },
		{ 85, 268, 45 },
		{ 85, 269, 45 },
		{ 85, 270, 45 },
		{ 85, 271, 45 },
		{ 85, 272, 45 },
		{ 85, 273, 45 },
		{ 85, 274, 45 },
		{ 85, 275, 45 },
		{ 85, 276, 45 },
		{ 85, 277, 45 },
		{ 85, 278, 45 },
		{ 85, 279, 45 },
		{ 85, 280, 45 },
		{ 85, 281, 45 },
		{ 85, 282, 45 },
		{ 85, 283, 45 },
		{ 85, 284, 45 },
		{ 85, 285, 45 },
		{ 85, 286, 45 },
		{ 85, 287, 45 },
		{ 85, 288, 45 },
		{ 85, 289, 45 },
		{ 85, 290, 45 },
		{ 85, 291, 45 },
		{ 85, 292, 45 },
		{ 85, 293, 45 },
		{ 85, 294, 45 },
		{ 85, 295, 45 },
		{ 85, 296, 45 },
		{ 85, 297, 45 },
		{ 85, 298, 45 },
		{ 85, 299, 45 },
		{ 86, 258, 45 },
		{ 86, 259, 45 },
		{ 86, 260, 45 },
		{ 86, 261, 45 },
		{ 86, 262, 45 },
		{ 86, 263, 45 },
		{ 86, 264, 45 },
		{ 86, 265, 45 },
		{ 86, 266, 45 },
		{ 86, 267, 45 },
		{ 86, 268, 45 },
		{ 86, 269, 45 },
		{ 86, 270, 45 },
		{ 86, 271, 45 },
		{ 86, 272, 45 },
		{ 86, 273, 45 },
		{ 86, 274, 45 },
		{ 86, 275, 45 },
		{ 86, 276, 45 },
		{ 86, 277, 45 },
		{ 86, 278, 45 },
		{ 86, 279, 45 },
		{ 86, 280, 45 },
		{ 86, 281, 45 },
		{ 86, 282, 45 },
		{ 86, 283, 45 },
		{ 86, 284, 45 },
		{ 86, 285, 45 },
		{ 86, 286, 45 },
		{ 86, 287, 45 },
		{ 86, 288, 45 },
		{ 86, 289, 45 },
		{ 86, 290, 45 },
		{ 86, 291, 45 },
		{ 86, 292, 45 },
		{ 86, 293, 45 },
		{ 86, 294, 45 },
		{ 86, 295, 45 },
		{ 86, 296, 45 },
		{ 86, 297, 45 },
		{ 86, 298, 45 },
		{ 86, 299, 45 },
		{ 87, 258, 45 },
		{ 87, 259, 45 },
		{ 87, 260, 45 },
		{ 87, 261, 45 },
		{ 87, 262, 45 },
		{ 87, 263, 45 },
		{ 87, 264, 45 },
		{ 87, 265, 45 },
		{ 87, 266, 45 },
		{ 87, 267, 45 },
		{ 87, 268, 45 },
		{ 87, 269, 45 },
		{ 87, 270, 45 },
		{ 87, 271, 45 },
		{ 87, 272, 45 },
		{ 87, 273, 45 },
		{ 87, 274, 45 },
		{ 87, 275, 45 },
		{ 87, 276, 45 },
		{ 87, 277, 45 },
		{ 87, 278, 45 },
		{ 87, 279, 45 },
		{ 87, 280, 45 },
		{ 87, 281, 45 },
		{ 87, 282, 45 },
		{ 87, 283, 45 },
		{ 87, 284, 45 },
		{ 87, 285, 45 },
		{ 87, 286, 45 },
		{ 87, 287, 45 },
		{ 87, 288, 45 },
		{ 87, 289, 45 },
		{ 87, 290, 45 },
		{ 87, 291, 45 },
		{ 87, 292, 45 },
		{ 87, 293, 45 },
		{ 87, 294, 45 },
		{ 87, 295, 45 },
		{ 87, 296, 45 },
		{ 87, 297, 45 },
		{ 87, 298, 45 },
		{ 87, 299, 45 },
		{ 88, 258, 45 },
		{ 88, 259, 45 },
		{ 88, 260, 45 },
		{ 88, 261, 45 },
		{ 88, 262, 45 },
		{ 88, 263, 45 },
		{ 88, 264, 45 },
		{ 88, 265, 45 },
		{ 88, 266, 45 },
		{ 88, 267, 45 },
		{ 88, 268, 45 },
		{ 88, 269, 45 },
		{ 88, 270, 45 },
		{ 88, 271, 45 },
		{ 88, 272, 45 },
		{ 88, 273, 45 },
		{ 88, 274, 45 },
		{ 88, 275, 45 },
		{ 88, 276, 45 },
		{ 88, 277, 45 },
		{ 88, 278, 45 },
		{ 88, 279, 45 },
		{ 88, 280, 45 },
		{ 88, 281, 45 },
		{ 88, 282, 45 },
		{ 88, 283, 45 },
		{ 88, 284, 45 },
		{ 88, 285, 45 },
		{ 88, 286, 45 },
		{ 88, 287, 45 },
		{ 88, 288, 45 },
		{ 88, 289, 45 },
		{ 88, 290, 45 },
		{ 88, 291, 45 },
		{ 88, 292, 45 },
		{ 88, 293, 45 },
		{ 88, 294, 45 },
		{ 88, 295, 45 },
		{ 88, 296, 45 },
		{ 88, 297, 45 },
		{ 88, 298, 45 },
		{ 88, 299, 45 },
		{ 89, 258, 45 },
		{ 89, 259, 45 },
		{ 89, 260, 45 },
		{ 89, 261, 45 },
		{ 89, 262, 45 },
		{ 89, 263, 45 },
		{ 89, 264, 45 },
		{ 89, 265, 45 },
		{ 89, 266, 45 },
		{ 89, 267, 45 },
		{ 89, 268, 45 },
		{ 89, 269, 45 },
		{ 89, 270, 45 },
		{ 89, 271, 45 },
		{ 89, 272, 45 },
		{ 89, 273, 45 },
		{ 89, 274, 45 },
		{ 89, 275, 45 },
		{ 89, 276, 45 },
		{ 89, 277, 45 },
		{ 89, 278, 45 },
		{ 89, 279, 45 },
		{ 89, 280, 45 },
		{ 89, 281, 45 },
		{ 89, 282, 45 },
		{ 89, 283, 45 },
		{ 89, 284, 45 },
		{ 89, 285, 45 },
		{ 89, 286, 45 },
		{ 89, 287, 45 },
		{ 89, 288, 45 },
		{ 89, 289, 45 },
		{ 89, 290, 45 },
		{ 89, 291, 45 },
		{ 89, 292, 45 },
		{ 89, 293, 45 },
		{ 89, 294, 45 },
		{ 89, 295, 45 },
		{ 89, 296, 45 },
		{ 89, 297, 45 },
		{ 89, 298, 45 },
		{ 90, 258, 45 },
		{ 90, 259, 45 },
		{ 90, 260, 45 },
		{ 90, 261, 45 },
		{ 90, 262, 45 },
		{ 90, 263, 45 },
		{ 90, 264, 45 },
		{ 90, 265, 45 },
		{ 90, 266, 45 },
		{ 90, 267, 45 },
		{ 90, 268, 45 },
		{ 90, 269, 45 },
		{ 90, 270, 45 },
		{ 90, 271, 45 },
		{ 90, 272, 45 },
		{ 90, 273, 45 },
		{ 90, 274, 45 },
		{ 90, 275, 45 },
		{ 90, 276, 45 },
		{ 90, 277, 45 },
		{ 90, 278, 45 },
		{ 90, 279, 45 },
		{ 90, 280, 45 },
		{ 90, 281, 45 },
		{ 90, 282, 45 },
		{ 90, 283, 45 },
		{ 90, 284, 45 },
		{ 90, 285, 45 },
		{ 90, 286, 45 },
		{ 90, 287, 45 },
		{ 90, 288, 45 },
		{ 90, 289, 45 },
		{ 90, 290, 45 },
		{ 90, 291, 45 },
		{ 90, 292, 45 },
		{ 90, 293, 45 },
		{ 90, 294, 45 },
		{ 90, 295, 45 },
		{ 90, 296, 45 },
		{ 90, 297, 45 },
		{ 90, 298, 45 },
		{ 91, 258, 45 },
		{ 91, 259, 45 },
		{ 91, 260, 45 },
		{ 91, 261, 45 },
		{ 91, 262, 45 },
		{ 91, 263, 45 },
		{ 91, 264, 45 },
		{ 91, 265, 45 },
		{ 91, 266, 45 },
		{ 91, 267, 45 },
		{ 91, 268, 45 },
		{ 91, 269, 45 },
		{ 91, 270, 45 },
		{ 91, 271, 45 },
		{ 91, 272, 45 },
		{ 91, 273, 45 },
		{ 91, 274, 45 },
		{ 91, 275, 45 },
		{ 91, 276, 45 },
		{ 91, 277, 45 },
		{ 91, 278, 45 },
		{ 91, 279, 45 },
		{ 91, 280, 45 },
		{ 91, 281, 45 },
		{ 91, 282, 45 },
		{ 91, 283, 45 },
		{ 91, 284, 45 },
		{ 91, 285, 45 },
		{ 91, 286, 45 },
		{ 91, 287, 45 },
		{ 91, 288, 45 },
		{ 91, 289, 45 },
		{ 91, 290, 45 },
		{ 91, 291, 45 },
		{ 91, 292, 45 },
		{ 91, 293, 45 },
		{ 91, 294, 45 },
		{ 91, 295, 45 },
		{ 91, 296, 45 },
		{ 91, 297, 45 },
		{ 91, 298, 45 },
		{ 92, 258, 45 },
		{ 92, 259, 45 },
		{ 92, 260, 45 },
		{ 92, 261, 45 },
		{ 92, 262, 45 },
		{ 92, 263, 45 },
		{ 92, 264, 45 },
		{ 92, 265, 45 },
		{ 92, 266, 45 },
		{ 92, 267, 45 },
		{ 92, 268, 45 },
		{ 92, 269, 45 },
		{ 92, 270, 45 },
		{ 92, 271, 45 },
		{ 92, 272, 45 },
		{ 92, 273, 45 },
		{ 92, 274, 45 },
		{ 92, 275, 45 },
		{ 92, 276, 45 },
		{ 92, 277, 45 },
		{ 92, 278, 45 },
		{ 92, 279, 45 },
		{ 92, 280, 45 },
		{ 92, 281, 45 },
		{ 92, 282, 45 },
		{ 92, 283, 45 },
		{ 92, 284, 45 },
		{ 92, 285, 45 },
		{ 92, 286, 45 },
		{ 92, 287, 45 },
		{ 92, 288, 45 },
		{ 92, 289, 45 },
		{ 92, 290, 45 },
		{ 92, 291, 45 },
		{ 92, 292, 45 },
		{ 92, 293, 45 },
		{ 92, 294, 45 },
		{ 92, 295, 45 },
		{ 92, 296, 45 },
		{ 92, 297, 45 },
		{ 92, 298, 45 },
		{ 93, 258, 45 },
		{ 93, 259, 45 },
		{ 93, 260, 45 },
		{ 93, 261, 45 },
		{ 93, 262, 45 },
		{ 93, 263, 45 },
		{ 93, 264, 45 },
		{ 93, 265, 45 },
		{ 93, 266, 45 },
		{ 93, 267, 45 },
		{ 93, 268, 45 },
		{ 93, 269, 45 },
		{ 93, 270, 45 },
		{ 93, 271, 45 },
		{ 93, 272, 45 },
		{ 93, 273, 45 },
		{ 93, 274, 45 },
		{ 93, 275, 45 },
		{ 93, 276, 45 },
		{ 93, 277, 45 },
		{ 93, 278, 45 },
		{ 93, 279, 45 },
		{ 93, 280, 45 },
		{ 93, 281, 45 },
		{ 93, 282, 45 },
		{ 93, 283, 45 },
		{ 93, 284, 45 },
		{ 93, 285, 45 },
		{ 93, 286, 45 },
		{ 93, 287, 45 },
		{ 93, 288, 45 },
		{ 93, 289, 45 },
		{ 93, 290, 45 },
		{ 93, 291, 45 },
		{ 93, 292, 45 },
		{ 93, 293, 45 },
		{ 93, 294, 45 },
		{ 93, 295, 45 },
		{ 93, 296, 45 },
		{ 93, 297, 45 },
		{ 94, 258, 45 },
		{ 94, 259, 45 },
		{ 94, 260, 45 },
		{ 94, 261, 45 },
		{ 94, 262, 45 },
		{ 94, 263, 45 },
		{ 94, 264, 45 },
		{ 94, 265, 45 },
		{ 94, 266, 45 },
		{ 94, 267, 45 },
		{ 94, 268, 45 },
		{ 94, 269, 45 },
		{ 94, 270, 45 },
		{ 94, 271, 45 },
		{ 94, 272, 45 },
		{ 94, 273, 45 },
		{ 94, 274, 45 },
		{ 94, 275, 45 },
		{ 94, 276, 45 },
		{ 94, 277, 45 },
		{ 94, 278, 45 },
		{ 94, 279, 45 },
		{ 94, 280, 45 },
		{ 94, 281, 45 },
		{ 94, 282, 45 },
		{ 94, 283, 45 },
		{ 94, 284, 45 },
		{ 94, 285, 45 },
		{ 94, 286, 45 },
		{ 94, 287, 45 },
		{ 94, 288, 45 },
		{ 94, 289, 45 },
		{ 94, 290, 45 },
		{ 94, 291, 45 },
		{ 94, 292, 45 },
		{ 94, 293, 45 },
		{ 94, 294, 45 },
		{ 94, 295, 45 },
		{ 94, 296, 45 },
		{ 94, 297, 45 },
		{ 95, 258, 45 },
		{ 95, 259, 45 },
		{ 95, 260, 45 },
		{ 95, 261, 45 },
		{ 95, 262, 45 },
		{ 95, 263, 45 },
		{ 95, 264, 45 },
		{ 95, 265, 45 },
		{ 95, 266, 45 },
		{ 95, 267, 45 },
		{ 95, 268, 45 },
		{ 95, 269, 45 },
		{ 95, 270, 45 },
		{ 95, 271, 45 },
		{ 95, 272, 45 },
		{ 95, 273, 45 },
		{ 95, 274, 45 },
		{ 95, 275, 45 },
		{ 95, 276, 45 },
		{ 95, 277, 45 },
		{ 95, 278, 45 },
		{ 95, 279, 45 },
		{ 95, 280, 45 },
		{ 95, 281, 45 },
		{ 95, 282, 45 },
		{ 95, 283, 45 },
		{ 95, 284, 45 },
		{ 95, 285, 45 },
		{ 95, 286, 45 },
		{ 95, 287, 45 },
		{ 95, 288, 45 },
		{ 95, 289, 45 },
		{ 95, 290, 45 },
		{ 95, 291, 45 },
		{ 95, 292, 45 },
		{ 95, 293, 45 },
		{ 95, 294, 45 },
		{ 95, 295, 45 },
		{ 95, 296, 45 },
		{ 95, 297, 45 },
		{ 96, 258, 45 },
		{ 96, 259, 45 },
		{ 96, 260, 45 },
		{ 96, 261, 45 },
		{ 96, 262, 45 },
		{ 96, 263, 45 },
		{ 96, 264, 45 },
		{ 96, 265, 45 },
		{ 96, 266, 45 },
		{ 96, 267, 45 },
		{ 96, 268, 45 },
		{ 96, 269, 45 },
		{ 96, 270, 45 },
		{ 96, 271, 45 },
		{ 96, 272, 45 },
		{ 96, 273, 45 },
		{ 96, 274, 45 },
		{ 96, 275, 45 },
		{ 96, 276, 45 },
		{ 96, 277, 45 },
		{ 96, 278, 45 },
		{ 96, 279, 45 },
		{ 96, 280, 45 },
		{ 96, 281, 45 },
		{ 96, 282, 45 },
		{ 96, 283, 45 },
		{ 96, 284, 45 },
		{ 96, 285, 45 },
		{ 96, 286, 45 },
		{ 96, 287, 45 },
		{ 96, 288, 45 },
		{ 96, 289, 45 },
		{ 96, 290, 45 },
		{ 96, 291, 45 },
		{ 96, 292, 45 },
		{ 96, 293, 45 },
		{ 96, 294, 45 },
		{ 96, 295, 45 },
		{ 96, 296, 45 },
		{ 96, 297, 45 },
		{ 97, 258, 45 },
		{ 97, 259, 45 },
		{ 97, 260, 45 },
		{ 97, 261, 45 },
		{ 97, 262, 45 },
		{ 97, 263, 45 },
		{ 97, 264, 45 },
		{ 97, 265, 45 },
		{ 97, 266, 45 },
		{ 97, 267, 45 },
		{ 97, 268, 45 },
		{ 97, 269, 45 },
		{ 97, 270, 45 },
		{ 97, 271, 45 },
		{ 97, 272, 45 },
		{ 97, 273, 45 },
		{ 97, 274, 45 },
		{ 97, 275, 45 },
		{ 97, 276, 45 },
		{ 97, 277, 45 },
		{ 97, 278, 45 },
		{ 97, 279, 45 },
		{ 97, 280, 45 },
		{ 97, 281, 45 },
		{ 97, 282, 45 },
		{ 97, 283, 45 },
		{ 97, 284, 45 },
		{ 97, 285, 45 },
		{ 97, 286, 45 },
		{ 97, 287, 45 },
		{ 97, 288, 45 },
		{ 97, 289, 45 },
		{ 97, 290, 45 },
		{ 97, 291, 45 },
		{ 97, 292, 45 },
		{ 97, 293, 45 },
		{ 97, 294, 45 },
		{ 97, 295, 45 },
		{ 97, 296, 45 },
		{ 98, 258, 45 },
		{ 98, 259, 45 },
		{ 98, 260, 45 },
		{ 98, 261, 45 },
		{ 98, 262, 45 },
		{ 98, 263, 45 },
		{ 98, 264, 45 },
		{ 98, 265, 45 },
		{ 98, 266, 45 },
		{ 98, 267, 45 },
		{ 98, 268, 45 },
		{ 98, 269, 45 },
		{ 98, 270, 45 },
		{ 98, 271, 45 },
		{ 98, 272, 45 },
		{ 98, 273, 45 },
		{ 98, 274, 45 },
		{ 98, 275, 45 },
		{ 98, 276, 45 },
		{ 98, 277, 45 },
		{ 98, 278, 45 },
		{ 98, 279, 45 },
		{ 98, 280, 45 },
		{ 98, 281, 45 },
		{ 98, 282, 45 },
		{ 98, 283, 45 },
		{ 98, 284, 45 },
		{ 98, 285, 45 },
		{ 98, 286, 45 },
		{ 98, 287, 45 },
		{ 98, 288, 45 },
		{ 98, 289, 45 },
		{ 98, 290, 45 },
		{ 98, 291, 45 },
		{ 98, 292, 45 },
		{ 98, 293, 45 },
		{ 98, 294, 45 },
		{ 98, 295, 45 },
		{ 98, 296, 45 },
		{ 99, 258, 45 },
		{ 99, 259, 45 },
		{ 99, 260, 45 },
		{ 99, 261, 45 },
		{ 99, 262, 45 },
		{ 99, 263, 45 },
		{ 99, 264, 45 },
		{ 99, 265, 45 },
		{ 99, 266, 45 },
		{ 99, 267, 45 },
		{ 99, 268, 45 },
		{ 99, 269, 45 },
		{ 99, 270, 45 },
		{ 99, 271, 45 },
		{ 99, 272, 45 },
		{ 99, 273, 45 },
		{ 99, 274, 45 },
		{ 99, 275, 45 },
		{ 99, 276, 45 },
		{ 99, 277, 45 },
		{ 99, 278, 45 },
		{ 99, 279, 45 },
		{ 99, 280, 45 },
		{ 99, 281, 45 },
		{ 99, 282, 45 },
		{ 99, 283, 45 },
		{ 99, 284, 45 },
		{ 99, 285, 45 },
		{ 99, 286, 45 },
		{ 99, 287, 45 },
		{ 99, 288, 45 },
		{ 99, 289, 45 },
		{ 99, 290, 45 },
		{ 99, 291, 45 },
		{ 99, 292, 45 },
		{ 99, 293, 45 },
		{ 99, 294, 45 },
		{ 99, 295, 45 },
		{ 99, 296, 45 },
		{ 100, 258, 45 },
		{ 100, 259, 45 },
		{ 100, 260, 45 },
		{ 100, 261, 45 },
		{ 100, 262, 45 },
		{ 100, 263, 45 },
		{ 100, 264, 45 },
		{ 100, 265, 45 },
		{ 100, 266, 45 },
		{ 100, 267, 45 },
		{ 100, 268, 45 },
		{ 100, 269, 45 },
		{ 100, 270, 45 },
		{ 100, 271, 45 },
		{ 100, 272, 45 },
		{ 100, 273, 45 },
		{ 100, 274, 45 },
		{ 100, 275, 45 },
		{ 100, 276, 45 },
		{ 100, 277, 45 },
		{ 100, 278, 45 },
		{ 100, 279, 45 },
		{ 100, 280, 45 },
		{ 100, 281, 45 },
		{ 100, 282, 45 },
		{ 100, 283, 45 },
		{ 100, 284, 45 },
		{ 100, 285, 45 },
		{ 100, 286, 45 },
		{ 100, 287, 45 },
		{ 100, 288, 45 },
		{ 100, 289, 45 },
		{ 100, 290, 45 },
		{ 100, 291, 45 },
		{ 100, 292, 45 },
		{ 100, 293, 45 },
		{ 100, 294, 45 },
		{ 100, 295, 45 },
		{ 100, 296, 45 },
		{ 101, 258, 45 },
		{ 101, 259, 45 },
		{ 101, 260, 45 },
		{ 101, 261, 45 },
		{ 101, 262, 45 },
		{ 101, 263, 45 },
		{ 101, 264, 45 },
		{ 101, 265, 45 },
		{ 101, 266, 45 },
		{ 101, 267, 45 },
		{ 101, 268, 45 },
		{ 101, 269, 45 },
		{ 101, 270, 45 },
		{ 101, 271, 45 },
		{ 101, 272, 45 },
		{ 101, 273, 45 },
		{ 101, 274, 45 },
		{ 101, 275, 45 },
		{ 101, 276, 45 },
		{ 101, 277, 45 },
		{ 101, 278, 45 },
		{ 101, 279, 45 },
		{ 101, 280, 45 },
		{ 101, 281, 45 },
		{ 101, 282, 45 },
		{ 101, 283, 45 },
		{ 101, 284, 45 },
		{ 101, 285, 45 },
		{ 101, 286, 45 },
		{ 101, 287, 45 },
		{ 101, 288, 45 },
		{ 101, 289, 45 },
		{ 101, 290, 45 },
		{ 101, 291, 45 },
		{ 101, 292, 45 },
		{ 101, 293, 45 },
		{ 101, 294, 45 },
		{ 101, 295, 45 },
		{ 102, 258, 45 },
		{ 102, 259, 45 },
		{ 102, 260, 45 },
		{ 102, 261, 45 },
		{ 102, 262, 45 },
		{ 102, 263, 45 },
		{ 102, 264, 45 },
		{ 102, 265, 45 },
		{ 102, 266, 45 },
		{ 102, 267, 45 },
		{ 102, 268, 45 },
		{ 102, 269, 45 },
		{ 102, 270, 45 },
		{ 102, 271, 45 },
		{ 102, 272, 45 },
		{ 102, 273, 45 },
		{ 102, 274, 45 },
		{ 102, 275, 45 },
		{ 102, 276, 45 },
		{ 102, 277, 45 },
		{ 102, 278, 45 },
		{ 102, 279, 45 },
		{ 102, 280, 45 },
		{ 102, 281, 45 },
		{ 102, 282, 45 },
		{ 102, 283, 45 },
		{ 102, 284, 45 },
		{ 102, 285, 45 },
		{ 102, 286, 45 },
		{ 102, 287, 45 },
		{ 102, 288, 45 },
		{ 102, 289, 45 },
		{ 102, 290, 45 },
		{ 102, 291, 45 },
		{ 102, 292, 45 },
		{ 102, 293, 45 },
		{ 102, 294, 45 },
		{ 102, 295, 45 },
		{ 103, 258, 45 },
		{ 103, 259, 45 },
		{ 103, 260, 45 },
		{ 103, 261, 45 },
		{ 103, 262, 45 },
		{ 103, 263, 45 },
		{ 103, 264, 45 },
		{ 103, 265, 45 },
		{ 103, 266, 45 },
		{ 103, 267, 45 },
		{ 103, 268, 45 },
		{ 103, 269, 45 },
		{ 103, 270, 45 },
		{ 103, 271, 45 },
		{ 103, 272, 45 },
		{ 103, 273, 45 },
		{ 103, 274, 45 },
		{ 103, 275, 45 },
		{ 103, 276, 45 },
		{ 103, 277, 45 },
		{ 103, 278, 45 },
		{ 103, 279, 45 },
		{ 103, 280, 45 },
		{ 103, 281, 45 },
		{ 103, 282, 45 },
		{ 103, 283, 45 },
		{ 103, 284, 45 },
		{ 103, 285, 45 },
		{ 103, 286, 45 },
		{ 103, 287, 45 },
		{ 103, 288, 45 },
		{ 103, 289, 45 },
		{ 103, 290, 45 },
		{ 103, 291, 45 },
		{ 103, 292, 45 },
		{ 103, 293, 45 },
		{ 103, 294, 45 },
		{ 103, 295, 45 },
		{ 104, 258, 45 },
		{ 104, 259, 45 },
		{ 104, 260, 45 },
		{ 104, 261, 45 },
		{ 104, 262, 45 },
		{ 104, 263, 45 },
		{ 104, 264, 45 },
		{ 104, 265, 45 },
		{ 104, 266, 45 },
		{ 104, 267, 45 },
		{ 104, 268, 45 },
		{ 104, 269, 45 },
		{ 104, 270, 45 },
		{ 104, 271, 45 },
		{ 104, 272, 45 },
		{ 104, 273, 45 },
		{ 104, 274, 45 },
		{ 104, 275, 45 },
		{ 104, 276, 45 },
		{ 104, 277, 45 },
		{ 104, 278, 45 },
		{ 104, 279, 45 },
		{ 104, 280, 45 },
		{ 104, 281, 45 },
		{ 104, 282, 45 },
		{ 104, 283, 45 },
		{ 104, 284, 45 },
		{ 104, 285, 45 },
		{ 104, 286, 45 },
		{ 104, 287, 45 },
		{ 104, 288, 45 },
		{ 104, 289, 45 },
		{ 104, 290, 45 },
		{ 104, 291, 45 },
		{ 104, 292, 45 },
		{ 104, 293, 45 },
		{ 104, 294, 45 },
		{ 104, 295, 45 },
		{ 105, 258, 45 },
		{ 105, 259, 45 },
		{ 105, 260, 45 },
		{ 105, 261, 45 },
		{ 105, 262, 45 },
		{ 105, 263, 45 },
		{ 105, 264, 45 },
		{ 105, 265, 45 },
		{ 105, 266, 45 },
		{ 105, 267, 45 },
		{ 105, 268, 45 },
		{ 105, 269, 45 },
		{ 105, 270, 45 },
		{ 105, 271, 45 },
		{ 105, 272, 45 },
		{ 105, 273, 45 },
		{ 105, 274, 45 },
		{ 105, 275, 45 },
		{ 105, 276, 45 },
		{ 105, 277, 45 },
		{ 105, 278, 45 },
		{ 105, 279, 45 },
		{ 105, 280, 45 },
		{ 105, 281, 45 },
		{ 105, 282, 45 },
		{ 105, 283, 45 },
		{ 105, 284, 45 },
		{ 105, 285, 45 },
		{ 105, 286, 45 },
		{ 105, 287, 45 },
		{ 105, 288, 45 },
		{ 105, 289, 45 },
		{ 105, 290, 45 },
		{ 105, 291, 45 },
		{ 105, 292, 45 },
		{ 105, 293, 45 },
		{ 105, 294, 45 },
		{ 106, 258, 45 },
		{ 106, 259, 45 },
		{ 106, 260, 45 },
		{ 106, 261, 45 },
		{ 106, 262, 45 },
		{ 106, 263, 45 },
		{ 106, 264, 45 },
		{ 106, 265, 45 },
		{ 106, 266, 45 },
		{ 106, 267, 45 },
		{ 106, 268, 45 },
		{ 106, 269, 45 },
		{ 106, 270, 45 },
		{ 106, 271, 45 },
		{ 106, 272, 45 },
		{ 106, 273, 45 },
		{ 106, 274, 45 },
		{ 106, 275, 45 },
		{ 106, 276, 45 },
		{ 106, 277, 45 },
		{ 106, 278, 45 },
		{ 106, 279, 45 },
		{ 106, 280, 45 },
		{ 106, 281, 45 },
		{ 106, 282, 45 },
		{ 106, 283, 45 },
		{ 106, 284, 45 },
		{ 106, 285, 45 },
		{ 106, 286, 45 },
		{ 106, 287, 45 },
		{ 106, 288, 45 },
		{ 106, 289, 45 },
		{ 106, 290, 45 },
		{ 106, 291, 45 },
		{ 106, 292, 45 },
		{ 106, 293, 45 },
		{ 106, 294, 45 },
		{ 107, 258, 45 },
		{ 107, 259, 45 },
		{ 107, 260, 45 },
		{ 107, 261, 45 },
		{ 107, 262, 45 },
		{ 107, 263, 45 },
		{ 107, 264, 45 },
		{ 107, 265, 45 },
		{ 107, 266, 45 },
		{ 107, 267, 45 },
		{ 107, 268, 45 },
		{ 107, 269, 45 },
		{ 107, 270, 45 },
		{ 107, 271, 45 },
		{ 107, 272, 45 },
		{ 107, 273, 45 },
		{ 107, 274, 45 },
		{ 107, 275, 45 },
		{ 107, 276, 45 },
		{ 107, 277, 45 },
		{ 107, 278, 45 },
		{ 107, 279, 45 },
		{ 107, 280, 45 },
		{ 107, 281, 45 },
		{ 107, 282, 45 },
		{ 107, 283, 45 },
		{ 107, 284, 45 },
		{ 107, 285, 45 },
		{ 107, 286, 45 },
		{ 107, 287, 45 },
		{ 107, 288, 45 },
		{ 107, 289, 45 },
		{ 107, 290, 45 },
		{ 107, 291, 45 },
		{ 107, 292, 45 },
		{ 107, 293, 45 },
		{ 107, 294, 45 },
		{ 108, 258, 45 },
		{ 108, 259, 45 },
		{ 108, 260, 45 },
		{ 108, 261, 45 },
		{ 108, 262, 45 },
		{ 108, 263, 45 },
		{ 108, 264, 45 },
		{ 108, 265, 45 },
		{ 108, 266, 45 },
		{ 108, 267, 45 },
		{ 108, 268, 45 },
		{ 108, 269, 45 },
		{ 108, 270, 45 },
		{ 108, 271, 45 },
		{ 108, 272, 45 },
		{ 108, 273, 45 },
		{ 108, 274, 45 },
		{ 108, 275, 45 },
		{ 108, 276, 45 },
		{ 108, 277, 45 },
		{ 108, 278, 45 },
		{ 108, 279, 45 },
		{ 108, 280, 45 },
		{ 108, 281, 45 },
		{ 108, 282, 45 },
		{ 108, 283, 45 },
		{ 108, 284, 45 },
		{ 108, 285, 45 },
		{ 108, 286, 45 },
		{ 108, 287, 45 },
		{ 108, 288, 45 },
		{ 108, 289, 45 },
		{ 108, 290, 45 },
		{ 108, 291, 45 },
		{ 108, 292, 45 },
		{ 108, 293, 45 },
		{ 108, 294, 45 },
		{ 109, 258, 45 },
		{ 109, 259, 45 },
		{ 109, 260, 45 },
		{ 109, 261, 45 },
		{ 109, 262, 45 },
		{ 109, 263, 45 },
		{ 109, 264, 45 },
		{ 109, 265, 45 },
		{ 109, 266, 45 },
		{ 109, 267, 45 },
		{ 109, 268, 45 },
		{ 109, 269, 45 },
		{ 109, 270, 45 },
		{ 109, 271, 45 },
		{ 109, 272, 45 },
		{ 109, 273, 45 },
		{ 109, 274, 45 },
		{ 109, 275, 45 },
		{ 109, 276, 45 },
		{ 109, 277, 45 },
		{ 109, 278, 45 },
		{ 109, 279, 45 },
		{ 109, 280, 45 },
		{ 109, 281, 45 },
		{ 109, 282, 45 },
		{ 109, 283, 45 },
		{ 109, 284, 45 },
		{ 109, 285, 45 },
		{ 109, 286, 45 },
		{ 109, 287, 45 },
		{ 109, 288, 45 },
		{ 109, 289, 45 },
		{ 109, 290, 45 },
		{ 109, 291, 45 },
		{ 109, 292, 45 },
		{ 109, 293, 45 },
		{ 110, 258, 45 },
		{ 110, 259, 45 },
		{ 110, 260, 45 },
		{ 110, 261, 45 },
		{ 110, 262, 45 },
		{ 110, 263, 45 },
		{ 110, 264, 45 },
		{ 110, 265, 45 },
		{ 110, 266, 45 },
		{ 110, 267, 45 },
		{ 110, 268, 45 },
		{ 110, 269, 45 },
		{ 110, 270, 45 },
		{ 110, 271, 45 },
		{ 110, 272, 45 },
		{ 110, 273, 45 },
		{ 110, 274, 45 },
		{ 110, 275, 45 },
		{ 110, 276, 45 },
		{ 110, 277, 45 },
		{ 110, 278, 45 },
		{ 110, 279, 45 },
		{ 110, 280, 45 },
		{ 110, 281, 45 },
		{ 110, 282, 45 },
		{ 110, 283, 45 },
		{ 110, 284, 45 },
		{ 110, 285, 45 },
		{ 110, 286, 45 },
		{ 110, 287, 45 },
		{ 110, 288, 45 },
		{ 110, 289, 45 },
		{ 110, 290, 45 },
		{ 110, 291, 45 },
		{ 110, 292, 45 },
		{ 110, 293, 45 },
		{ 111, 258, 45 },
		{ 111, 259, 45 },
		{ 111, 260, 45 },
		{ 111, 261, 45 },
		{ 111, 262, 45 },
		{ 111, 263, 45 },
		{ 111, 264, 45 },
		{ 111, 265, 45 },
		{ 111, 266, 45 },
		{ 111, 267, 45 },
		{ 111, 268, 45 },
		{ 111, 269, 45 },
		{ 111, 270, 45 },
		{ 111, 271, 45 },
		{ 111, 272, 45 },
		{ 111, 273, 45 },
		{ 111, 274, 45 },
		{ 111, 275, 45 },
		{ 111, 276, 45 },
		{ 111, 277, 45 },
		{ 111, 278, 45 },
		{ 111, 279, 45 },
		{ 111, 280, 45 },
		{ 111, 281, 45 },
		{ 111, 282, 45 },
		{ 111, 283, 45 },
		{ 111, 284, 45 },
		{ 111, 285, 45 },
		{ 111, 286, 45 },
		{ 111, 287, 45 },
		{ 111, 288, 45 },
		{ 111, 289, 45 },
		{ 111, 290, 45 },
		{ 111, 291, 45 },
		{ 111, 292, 45 },
		{ 111, 293, 45 },
		{ 112, 258, 45 },
		{ 112, 259, 45 },
		{ 112, 260, 45 },
		{ 112, 261, 45 },
		{ 112, 262, 45 },
		{ 112, 263, 45 },
		{ 112, 264, 45 },
		{ 112, 265, 45 },
		{ 112, 266, 45 },
		{ 112, 267, 45 },
		{ 112, 268, 45 },
		{ 112, 269, 45 },
		{ 112, 270, 45 },
		{ 112, 271, 45 },
		{ 112, 272, 45 },
		{ 112, 273, 45 },
		{ 112, 274, 45 },
		{ 112, 275, 45 },
		{ 112, 276, 45 },
		{ 112, 277, 45 },
		{ 112, 278, 45 },
		{ 112, 279, 45 },
		{ 112, 280, 45 },
		{ 112, 281, 45 },
		{ 112, 282, 45 },
		{ 112, 283, 45 },
		{ 112, 284, 45 },
		{ 112, 285, 45 },
		{ 112, 286, 45 },
		{ 112, 287, 45 },
		{ 112, 288, 45 },
		{ 112, 289, 45 },
		{ 112, 290, 45 },
		{ 112, 291, 45 },
		{ 112, 292, 45 },
		{ 112, 293, 45 },
		{ 113, 258, 45 },
		{ 113, 259, 45 },
		{ 113, 260, 45 },
		{ 113, 261, 45 },
		{ 113, 262, 45 },
		{ 113, 263, 45 },
		{ 113, 264, 45 },
		{ 113, 265, 45 },
		{ 113, 266, 45 },
		{ 113, 267, 45 },
		{ 113, 268, 45 },
		{ 113, 269, 45 },
		{ 113, 270, 45 },
		{ 113, 271, 45 },
		{ 113, 272, 45 },
		{ 113, 273, 45 },
		{ 113, 274, 45 },
		{ 113, 275, 45 },
		{ 113, 276, 45 },
		{ 113, 277, 45 },
		{ 113, 278, 45 },
		{ 113, 279, 45 },
		{ 113, 280, 45 },
		{ 113, 281, 45 },
		{ 113, 282, 45 },
		{ 113, 283, 45 },
		{ 113, 284, 45 },
		{ 113, 285, 45 },
		{ 113, 286, 45 },
		{ 113, 287, 45 },
		{ 113, 288, 45 },
		{ 113, 289, 45 },
		{ 113, 290, 45 },
		{ 113, 291, 45 },
		{ 113, 292, 45 },
		{ 114, 258, 45 },
		{ 114, 259, 45 },
		{ 114, 260, 45 },
		{ 114, 261, 45 },
		{ 114, 262, 45 },
		{ 114, 263, 45 },
		{ 114, 264, 45 },
		{ 114, 265, 45 },
		{ 114, 266, 45 },
		{ 114, 267, 45 },
		{ 114, 268, 45 },
		{ 114, 269, 45 },
		{ 114, 270, 45 },
		{ 114, 271, 45 },
		{ 114, 272, 45 },
		{ 114, 273, 45 },
		{ 114, 274, 45 },
		{ 114, 275, 45 },
		{ 114, 276, 45 },
		{ 114, 277, 45 },
		{ 114, 278, 45 },
		{ 114, 279, 45 },
		{ 114, 280, 45 },
		{ 114, 281, 45 },
		{ 114, 282, 45 },
		{ 114, 283, 45 },
		{ 114, 284, 45 },
		{ 114, 285, 45 },
		{ 114, 286, 45 },
		{ 114, 287, 45 },
		{ 114, 288, 45 },
		{ 114, 289, 45 },
		{ 114, 290, 45 },
		{ 114, 291, 45 },
		{ 114, 292, 45 },
		{ 115, 258, 45 },
		{ 115, 259, 45 },
		{ 115, 260, 45 },
		{ 115, 261, 45 },
		{ 115, 262, 45 },
		{ 115, 263, 45 },
		{ 115, 264, 45 },
		{ 115, 265, 45 },
		{ 115, 266, 45 },
		{ 115, 267, 45 },
		{ 115, 268, 45 },
		{ 115, 269, 45 },
		{ 115, 270, 45 },
		{ 115, 271, 45 },
		{ 115, 272, 45 },
		{ 115, 273, 45 },
		{ 115, 274, 45 },
		{ 115, 275, 45 },
		{ 115, 276, 45 },
		{ 115, 277, 45 },
		{ 115, 278, 45 },
		{ 115, 279, 45 },
		{ 115, 280, 45 },
		{ 115, 281, 45 },
		{ 115, 282, 45 },
		{ 115, 283, 45 },
		{ 115, 284, 45 },
		{ 115, 285, 45 },
		{ 115, 286, 45 },
		{ 115, 287, 45 },
		{ 115, 288, 45 },
		{ 115, 289, 45 },
		{ 115, 290, 45 },
		{ 115, 291, 45 },
		{ 115, 292, 45 },
		{ 116, 258, 45 },
		{ 116, 259, 45 },
		{ 116, 260, 45 },
		{ 116, 261, 45 },
		{ 116, 262, 45 },
		{ 116, 263, 45 },
		{ 116, 264, 45 },
		{ 116, 265, 45 },
		{ 116, 266, 45 },
		{ 116, 267, 45 },
		{ 116, 268, 45 },
		{ 116, 269, 45 },
		{ 116, 270, 45 },
		{ 116, 271, 45 },
		{ 116, 272, 45 },
		{ 116, 273, 45 },
		{ 116, 274, 45 },
		{ 116, 275, 45 },
		{ 116, 276, 45 },
		{ 116, 277, 45 },
		{ 116, 278, 45 },
		{ 116, 279, 45 },
		{ 116, 280, 45 },
		{ 116, 281, 45 },
		{ 116, 282, 45 },
		{ 116, 283, 45 },
		{ 116, 284, 45 },
		{ 116, 285, 45 },
		{ 116, 286, 45 },
		{ 116, 287, 45 },
		{ 116, 288, 45 },
		{ 116, 289, 45 },
		{ 116, 290, 45 },
		{ 116, 291, 45 },
		{ 116, 292, 45 },
		{ 117, 258, 45 },
		{ 117, 259, 45 },
		{ 117, 260, 45 },
		{ 117, 261, 45 },
		{ 117, 262, 45 },
		{ 117, 263, 45 },
		{ 117, 264, 45 },
		{ 117, 265, 45 },
		{ 117, 266, 45 },
		{ 117, 267, 45 },
		{ 117, 268, 45 },
		{ 117, 269, 45 },
		{ 117, 270, 45 },
		{ 117, 271, 45 },
		{ 117, 272, 45 },
		{ 117, 273, 45 },
		{ 117, 274, 45 },
		{ 117, 275, 45 },
		{ 117, 276, 45 },
		{ 117, 277, 45 },
		{ 117, 278, 45 },
		{ 117, 279, 45 },
		{ 117, 280, 45 },
		{ 117, 281, 45 },
		{ 117, 282, 45 },
		{ 117, 283, 45 },
		{ 117, 284, 45 },
		{ 117, 285, 45 },
		{ 117, 286, 45 },
		{ 117, 287, 45 },
		{ 117, 288, 45 },
		{ 117, 289, 45 },
		{ 117, 290, 45 },
		{ 117, 291, 45 },
		{ 118, 258, 45 },
		{ 118, 259, 45 },
		{ 118, 260, 45 },
		{ 118, 261, 45 },
		{ 118, 262, 45 },
		{ 118, 263, 45 },
		{ 118, 264, 45 },
		{ 118, 265, 45 },
		{ 118, 266, 45 },
		{ 118, 267, 45 },
		{ 118, 268, 45 },
		{ 118, 269, 45 },
		{ 118, 270, 45 },
		{ 118, 271, 45 },
		{ 118, 272, 45 },
		{ 118, 273, 45 },
		{ 118, 274, 45 },
		{ 118, 275, 45 },
		{ 118, 276, 45 },
		{ 118, 277, 45 },
		{ 118, 278, 45 },
		{ 118, 279, 45 },
		{ 118, 280, 45 },
		{ 118, 281, 45 },
		{ 118, 282, 45 },
		{ 118, 283, 45 },
		{ 118, 284, 45 },
		{ 118, 285, 45 },
		{ 118, 286, 45 },
		{ 118, 287, 45 },
		{ 118, 288, 45 },
		{ 118, 289, 45 },
		{ 118, 290, 45 },
		{ 118, 291, 45 },
		{ 119, 258, 45 },
		{ 119, 259, 45 },
		{ 119, 260, 45 },
		{ 119, 261, 45 },
		{ 119, 262, 45 },
		{ 119, 263, 45 },
		{ 119, 264, 45 },
		{ 119, 265, 45 },
		{ 119, 266, 45 },
		{ 119, 267, 45 },
		{ 119, 268, 45 },
		{ 119, 269, 45 },
		{ 119, 270, 45 },
		{ 119, 271, 45 },
		{ 119, 272, 45 },
		{ 119, 273, 45 },
		{ 119, 274, 45 },
		{ 119, 275, 45 },
		{ 119, 276, 45 },
		{ 119, 277, 45 },
		{ 119, 278, 45 },
		{ 119, 279, 45 },
		{ 119, 280, 45 },
		{ 119, 281, 45 },
		{ 119, 282, 45 },
		{ 119, 283, 45 },
		{ 119, 284, 45 },
		{ 119, 285, 45 },
		{ 119, 286, 45 },
		{ 119, 287, 45 },
		{ 119, 288, 45 },
		{ 119, 289, 45 },
		{ 119, 290, 45 },
		{ 119, 291, 45 },
		{ 120, 258, 45 },
		{ 120, 259, 45 },
		{ 120, 260, 45 },
		{ 120, 261, 45 },
		{ 120, 262, 45 },
		{ 120, 263, 45 },
		{ 120, 264, 45 },
		{ 120, 265, 45 },
		{ 120, 266, 45 },
		{ 120, 267, 45 },
		{ 120, 268, 45 },
		{ 120, 269, 45 },
		{ 120, 270, 45 },
		{ 120, 271, 45 },
		{ 120, 272, 45 },
		{ 120, 273, 45 },
		{ 120, 274, 45 },
		{ 120, 275, 45 },
		{ 120, 276, 45 },
		{ 120, 277, 45 },
		{ 120, 278, 45 },
		{ 120, 279, 45 },
		{ 120, 280, 45 },
		{ 120, 281, 45 },
		{ 120, 282, 45 },
		{ 120, 283, 45 },
		{ 120, 284, 45 },
		{ 120, 285, 45 },
		{ 120, 286, 45 },
		{ 120, 287, 45 },
		{ 120, 288, 45 },
		{ 120, 289, 45 },
		{ 120, 290, 45 },
		{ 120, 291, 45 },
		{ 121, 258, 45 },
		{ 121, 259, 45 },
		{ 121, 260, 45 },
		{ 121, 261, 45 },
		{ 121, 262, 45 },
		{ 121, 263, 45 },
		{ 121, 264, 45 },
		{ 121, 265, 45 },
		{ 121, 266, 45 },
		{ 121, 267, 45 },
		{ 121, 268, 45 },
		{ 121, 269, 45 },
		{ 121, 270, 45 },
		{ 121, 271, 45 },
		{ 121, 272, 45 },
		{ 121, 273, 45 },
		{ 121, 274, 45 },
		{ 121, 275, 45 },
		{ 121, 276, 45 },
		{ 121, 277, 45 },
		{ 121, 278, 45 },
		{ 121, 279, 45 },
		{ 121, 280, 45 },
		{ 121, 281, 45 },
		{ 121, 282, 45 },
		{ 121, 283, 45 },
		{ 121, 284, 45 },
		{ 121, 285, 45 },
		{ 121, 286, 45 },
		{ 121, 287, 45 },
		{ 121, 288, 45 },
		{ 121, 289, 45 },
		{ 121, 290, 45 },
		{ 122, 258, 45 },
		{ 122, 259, 45 },
		{ 122, 260, 45 },
		{ 122, 261, 45 },
		{ 122, 262, 45 },
		{ 122, 263, 45 },
		{ 122, 264, 45 },
		{ 122, 265, 45 },
		{ 122, 266, 45 },
		{ 122, 267, 45 },
		{ 122, 268, 45 },
		{ 122, 269, 45 },
		{ 122, 270, 45 },
		{ 122, 271, 45 },
		{ 122, 272, 45 },
		{ 122, 273, 45 },
		{ 122, 274, 45 },
		{ 122, 275, 45 },
		{ 122, 276, 45 },
		{ 122, 277, 45 },
		{ 122, 278, 45 },
		{ 122, 279, 45 },
		{ 122, 280, 45 },
		{ 122, 281, 45 },
		{ 122, 282, 45 },
		{ 122, 283, 45 },
		{ 122, 284, 45 },
		{ 122, 285, 45 },
		{ 122, 286, 45 },
		{ 122, 287, 45 },
		{ 122, 288, 45 },
		{ 122, 289, 45 },
		{ 122, 290, 45 },
		{ 123, 258, 45 },
		{ 123, 259, 45 },
		{ 123, 260, 45 },
		{ 123, 261, 45 },
		{ 123, 262, 45 },
		{ 123, 263, 45 },
		{ 123, 264, 45 },
		{ 123, 265, 45 },
		{ 123, 266, 45 },
		{ 123, 267, 45 },
		{ 123, 268, 45 },
		{ 123, 269, 45 },
		{ 123, 270, 45 },
		{ 123, 271, 45 },
		{ 123, 272, 45 },
		{ 123, 273, 45 },
		{ 123, 274, 45 },
		{ 123, 275, 45 },
		{ 123, 276, 45 },
		{ 123, 277, 45 },
		{ 123, 278, 45 },
		{ 123, 279, 45 },
		{ 123, 280, 45 },
		{ 123, 281, 45 },
		{ 123, 282, 45 },
		{ 123, 283, 45 },
		{ 123, 284, 45 },
		{ 123, 285, 45 },
		{ 123, 286, 45 },
		{ 123, 287, 45 },
		{ 123, 288, 45 },
		{ 123, 289, 45 },
		{ 123, 290, 45 },
		{ 124, 258, 45 },
		{ 124, 259, 45 },
		{ 124, 260, 45 },
		{ 124, 261, 45 },
		{ 124, 262, 45 },
		{ 124, 263, 45 },
		{ 124, 264, 45 },
		{ 124, 265, 45 },
		{ 124, 266, 45 },
		{ 124, 267, 45 },
		{ 124, 268, 45 },
		{ 124, 269, 45 },
		{ 124, 270, 45 },
		{ 124, 271, 45 },
		{ 124, 272, 45 },
		{ 124, 273, 45 },
		{ 124, 274, 45 },
		{ 124, 275, 45 },
		{ 124, 276, 45 },
		{ 124, 277, 45 },
		{ 124, 278, 45 },
		{ 124, 279, 45 },
		{ 124, 280, 45 },
		{ 124, 281, 45 },
		{ 124, 282, 45 },
		{ 124, 283, 45 },
		{ 124, 284, 45 },
		{ 124, 285, 45 },
		{ 124, 286, 45 },
		{ 124, 287, 45 },
		{ 124, 288, 45 },
		{ 124, 289, 45 },
		{ 124, 290, 45 },
		{ 125, 258, 45 },
		{ 125, 259, 45 },
		{ 125, 260, 45 },
		{ 125, 261, 45 },
		{ 125, 262, 45 },
		{ 125, 263, 45 },
		{ 125, 264, 45 },
		{ 125, 265, 45 },
		{ 125, 266, 45 },
		{ 125, 267, 45 },
		{ 125, 268, 45 },
		{ 125, 269, 45 },
		{ 125, 270, 45 },
		{ 125, 271, 45 },
		{ 125, 272, 45 },
		{ 125, 273, 45 },
		{ 125, 274, 45 },
		{ 125, 275, 45 },
		{ 125, 276, 45 },
		{ 125, 277, 45 },
		{ 125, 278, 45 },
		{ 125, 279, 45 },
		{ 125, 280, 45 },
		{ 125, 281, 45 },
		{ 125, 282, 45 },
		{ 125, 283, 45 },
		{ 125, 284, 45 },
		{ 125, 285, 45 },
		{ 125, 286, 45 },
		{ 125, 287, 45 },
		{ 125, 288, 45 },
		{ 125, 289, 45 },
		{ 126, 258, 45 },
		{ 126, 259, 45 },
		{ 126, 260, 45 },
		{ 126, 261, 45 },
		{ 126, 262, 45 },
		{ 126, 263, 45 },
		{ 126, 264, 45 },
		{ 126, 265, 45 },
		{ 126, 266, 45 },
		{ 126, 267, 45 },
		{ 126, 268, 45 },
		{ 126, 269, 45 },
		{ 126, 270, 45 },
		{ 126, 271, 45 },
		{ 126, 272, 45 },
		{ 126, 273, 45 },
		{ 126, 274, 45 },
		{ 126, 275, 45 },
		{ 126, 276, 45 },
		{ 126, 277, 45 },
		{ 126, 278, 45 },
		{ 126, 279, 45 },
		{ 126, 280, 45 },
		{ 126, 281, 45 },
		{ 126, 282, 45 },
		{ 126, 283, 45 },
		{ 126, 284, 45 },
		{ 126, 285, 45 },
		{ 126, 286, 45 },
		{ 126, 287, 45 },
		{ 126, 288, 45 },
		{ 126, 289, 45 },
		{ 127, 258, 45 },
		{ 127, 259, 45 },
		{ 127, 260, 45 },
		{ 127, 261, 45 },
		{ 127, 262, 45 },
		{ 127, 263, 45 },
		{ 127, 264, 45 },
		{ 127, 265, 45 },
		{ 127, 266, 45 },
		{ 127, 267, 45 },
		{ 127, 268, 45 },
		{ 127, 269, 45 },
		{ 127, 270, 45 },
		{ 127, 271, 45 },
		{ 127, 272, 45 },
		{ 127, 273, 45 },
		{ 127, 274, 45 },
		{ 127, 275, 45 },
		{ 127, 276, 45 },
		{ 127, 277, 45 },
		{ 127, 278, 45 },
		{ 127, 279, 45 },
		{ 127, 280, 45 },
		{ 127, 281, 45 },
		{ 127, 282, 45 },
		{ 127, 283, 45 },
		{ 127, 284, 45 },
		{ 127, 285, 45 },
		{ 127, 286, 45 },
		{ 127, 287, 45 },
		{ 127, 288, 45 },
		{ 127, 289, 45 },
		{ 128, 258, 45 },
		{ 128, 259, 45 },
		{ 128, 260, 45 },
		{ 128, 261, 45 },
		{ 128, 262, 45 },
		{ 128, 263, 45 },
		{ 128, 264, 45 },
		{ 128, 265, 45 },
		{ 128, 266, 45 },
		{ 128, 267, 45 },
		{ 128, 268, 45 },
		{ 128, 269, 45 },
		{ 128, 270, 45 },
		{ 128, 271, 45 },
		{ 128, 272, 45 },
		{ 128, 273, 45 },
		{ 128, 274, 45 },
		{ 128, 275, 45 },
		{ 128, 276, 45 },
		{ 128, 277, 45 },
		{ 128, 278, 45 },
		{ 128, 279, 45 },
		{ 128, 280, 45 },
		{ 128, 281, 45 },
		{ 128, 282, 45 },
		{ 128, 283, 45 },
		{ 128, 284, 45 },
		{ 128, 285, 45 },
		{ 128, 286, 45 },
		{ 128, 287, 45 },
		{ 128, 288, 45 },
		{ 128, 289, 45 },
		{ 129, 258, 45 },
		{ 129, 259, 45 },
		{ 129, 260, 45 },
		{ 129, 261, 45 },
		{ 129, 262, 45 },
		{ 129, 263, 45 },
		{ 129, 264, 45 },
		{ 129, 265, 45 },
		{ 129, 266, 45 },
		{ 129, 267, 45 },
		{ 129, 268, 45 },
		{ 129, 269, 45 },
		{ 129, 270, 45 },
		{ 129, 271, 45 },
		{ 129, 272, 45 },
		{ 129, 273, 45 },
		{ 129, 274, 45 },
		{ 129, 275, 45 },
		{ 129, 276, 45 },
		{ 129, 277, 45 },
		{ 129, 278, 45 },
		{ 129, 279, 45 },
		{ 129, 280, 45 },
		{ 129, 281, 45 },
		{ 129, 282, 45 },
		{ 129, 283, 45 },
		{ 129, 284, 45 },
		{ 129, 285, 45 },
		{ 129, 286, 45 },
		{ 129, 287, 45 },
		{ 129, 288, 45 },
		{ 130, 258, 45 },
		{ 130, 259, 45 },
		{ 130, 260, 45 },
		{ 130, 261, 45 },
		{ 130, 262, 45 },
		{ 130, 263, 45 },
		{ 130, 264, 45 },
		{ 130, 265, 45 },
		{ 130, 266, 45 },
		{ 130, 267, 45 },
		{ 130, 268, 45 },
		{ 130, 269, 45 },
		{ 130, 270, 45 },
		{ 130, 271, 45 },
		{ 130, 272, 45 },
		{ 130, 273, 45 },
		{ 130, 274, 45 },
		{ 130, 275, 45 },
		{ 130, 276, 45 },
		{ 130, 277, 45 },
		{ 130, 278, 45 },
		{ 130, 279, 45 },
		{ 130, 280, 45 },
		{ 130, 281, 45 },
		{ 130, 282, 45 },
		{ 130, 283, 45 },
		{ 130, 284, 45 },
		{ 130, 285, 45 },
		{ 130, 286, 45 },
		{ 130, 287, 45 },
		{ 130, 288, 45 },
		{ 131, 258, 45 },
		{ 131, 259, 45 },
		{ 131, 260, 45 },
		{ 131, 261, 45 },
		{ 131, 262, 45 },
		{ 131, 263, 45 },
		{ 131, 264, 45 },
		{ 131, 265, 45 },
		{ 131, 266, 45 },
		{ 131, 267, 45 },
		{ 131, 268, 45 },
		{ 131, 269, 45 },
		{ 131, 270, 45 },
		{ 131, 271, 45 },
		{ 131, 272, 45 },
		{ 131, 273, 45 },
		{ 131, 274, 45 },
		{ 131, 275, 45 },
		{ 131, 276, 45 },
		{ 131, 277, 45 },
		{ 131, 278, 45 },
		{ 131, 279, 45 },
		{ 131, 280, 45 },
		{ 131, 281, 45 },
		{ 131, 282, 45 },
		{ 131, 283, 45 },
		{ 131, 284, 45 },
		{ 131, 285, 45 },
		{ 131, 286, 45 },
		{ 131, 287, 45 },
		{ 131, 288, 45 },
		{ 132, 258, 45 },
		{ 132, 259, 45 },
		{ 132, 260, 45 },
		{ 132, 261, 45 },
		{ 132, 262, 45 },
		{ 132, 263, 45 },
		{ 132, 264, 45 },
		{ 132, 265, 45 },
		{ 132, 266, 45 },
		{ 132, 267, 45 },
		{ 132, 268, 45 },
		{ 132, 269, 45 },
		{ 132, 270, 45 },
		{ 132, 271, 45 },
		{ 132, 272, 45 },
		{ 132, 273, 45 },
		{ 132, 274, 45 },
		{ 132, 275, 45 },
		{ 132, 276, 45 },
		{ 132, 277, 45 },
		{ 132, 278, 45 },
		{ 132, 279, 45 },
		{ 132, 280, 45 },
		{ 132, 281, 45 },
		{ 132, 282, 45 },
		{ 132, 283, 45 },
		{ 132, 284, 45 },
		{ 132, 285, 45 },
		{ 132, 286, 45 },
		{ 132, 287, 45 },
		{ 132, 288, 45 },
		{ 133, 258, 45 },
		{ 133, 259, 45 },
		{ 133, 260, 45 },
		{ 133, 261, 45 },
		{ 133, 262, 45 },
		{ 133, 263, 45 },
		{ 133, 264, 45 },
		{ 133, 265, 45 },
		{ 133, 266, 45 },
		{ 133, 267, 45 },
		{ 133, 268, 45 },
		{ 133, 269, 45 },
		{ 133, 270, 45 },
		{ 133, 271, 45 },
		{ 133, 272, 45 },
		{ 133, 273, 45 },
		{ 133, 274, 45 },
		{ 133, 275, 45 },
		{ 133, 276, 45 },
		{ 133, 277, 45 },
		{ 133, 278, 45 },
		{ 133, 279, 45 },
		{ 133, 280, 45 },
		{ 133, 281, 45 },
		{ 133, 282, 45 },
		{ 133, 283, 45 },
		{ 133, 284, 45 },
		{ 133, 285, 45 },
		{ 133, 286, 45 },
		{ 133, 287, 45 },
		{ 134, 258, 45 },
		{ 134, 259, 45 },
		{ 134, 260, 45 },
		{ 134, 261, 45 },
		{ 134, 262, 45 },
		{ 134, 263, 45 },
		{ 134, 264, 45 },
		{ 134, 265, 45 },
		{ 134, 266, 45 },
		{ 134, 267, 45 },
		{ 134, 268, 45 },
		{ 134, 269, 45 },
		{ 134, 270, 45 },
		{ 134, 271, 45 },
		{ 134, 272, 45 },
		{ 134, 273, 45 },
		{ 134, 274, 45 },
		{ 134, 275, 45 },
		{ 134, 276, 45 },
		{ 134, 277, 45 },
		{ 134, 278, 45 },
		{ 134, 279, 45 },
		{ 134, 280, 45 },
		{ 134, 281, 45 },
		{ 134, 282, 45 },
		{ 134, 283, 45 },
		{ 134, 284, 45 },
		{ 134, 285, 45 },
		{ 134, 286, 45 },
		{ 134, 287, 45 },
		{ 135, 258, 45 },
		{ 135, 259, 45 },
		{ 135, 260, 45 },
		{ 135, 261, 45 },
		{ 135, 262, 45 },
		{ 135, 263, 45 },
		{ 135, 264, 45 },
		{ 135, 265, 45 },
		{ 135, 266, 45 },
		{ 135, 267, 45 },
		{ 135, 268, 45 },
		{ 135, 269, 45 },
		{ 135, 270, 45 },
		{ 135, 271, 45 },
		{ 135, 272, 45 },
		{ 135, 273, 45 },
		{ 135, 274, 45 },
		{ 135, 275, 45 },
		{ 135, 276, 45 },
		{ 135, 277, 45 },
		{ 135, 278, 45 },
		{ 135, 279, 45 },
		{ 135, 280, 45 },
		{ 135, 281, 45 },
		{ 135, 282, 45 },
		{ 135, 283, 45 },
		{ 135, 284, 45 },
		{ 135, 285, 45 },
		{ 135, 286, 45 },
		{ 135, 287, 45 },
		{ 136, 258, 45 },
		{ 136, 259, 45 },
		{ 136, 260, 45 },
		{ 136, 261, 45 },
		{ 136, 262, 45 },
		{ 136, 263, 45 },
		{ 136, 264, 45 },
		{ 136, 265, 45 },
		{ 136, 266, 45 },
		{ 136, 267, 45 },
		{ 136, 268, 45 },
		{ 136, 269, 45 },
		{ 136, 270, 45 },
		{ 136, 271, 45 },
		{ 136, 272, 45 },
		{ 136, 273, 45 },
		{ 136, 274, 45 },
		{ 136, 275, 45 },
		{ 136, 276, 45 },
		{ 136, 277, 45 },
		{ 136, 278, 45 },
		{ 136, 279, 45 },
		{ 136, 280, 45 },
		{ 136, 281, 45 },
		{ 136, 282, 45 },
		{ 136, 283, 45 },
		{ 136, 284, 45 },
		{ 136, 285, 45 },
		{ 136, 286, 45 },
		{ 136, 287, 45 },
		{ 137, 258, 45 },
		{ 137, 259, 45 },
		{ 137, 260, 45 },
		{ 137, 261, 45 },
		{ 137, 262, 45 },
		{ 137, 263, 45 },
		{ 137, 264, 45 },
		{ 137, 265, 45 },
		{ 137, 266, 45 },
		{ 137, 267, 45 },
		{ 137, 268, 45 },
		{ 137, 269, 45 },
		{ 137, 270, 45 },
		{ 137, 271, 45 },
		{ 137, 272, 45 },
		{ 137, 273, 45 },
		{ 137, 274, 45 },
		{ 137, 275, 45 },
		{ 137, 276, 45 },
		{ 137, 277, 45 },
		{ 137, 278, 45 },
		{ 137, 279, 45 },
		{ 137, 280, 45 },
		{ 137, 281, 45 },
		{ 137, 282, 45 },
		{ 137, 283, 45 },
		{ 137, 284, 45 },
		{ 137, 285, 45 },
		{ 137, 286, 45 },
		{ 138, 258, 45 },
		{ 138, 259, 45 },
		{ 138, 260, 45 },
		{ 138, 261, 45 },
		{ 138, 262, 45 },
		{ 138, 263, 45 },
		{ 138, 264, 45 },
		{ 138, 265, 45 },
		{ 138, 266, 45 },
		{ 138, 267, 45 },
		{ 138, 268, 45 },
		{ 138, 269, 45 },
		{ 138, 270, 45 },
		{ 138, 271, 45 },
		{ 138, 272, 45 },
		{ 138, 273, 45 },
		{ 138, 274, 45 },
		{ 138, 275, 45 },
		{ 138, 276, 45 },
		{ 138, 277, 45 },
		{ 138, 278, 45 },
		{ 138, 279, 45 },
		{ 138, 280, 45 },
		{ 138, 281, 45 },
		{ 138, 282, 45 },
		{ 138, 283, 45 },
		{ 138, 284, 45 },
		{ 138, 285, 45 },
		{ 138, 286, 45 },
		{ 139, 258, 45 },
		{ 139, 259, 45 },
		{ 139, 260, 45 },
		{ 139, 261, 45 },
		{ 139, 262, 45 },
		{ 139, 263, 45 },
		{ 139, 264, 45 },
		{ 139, 265, 45 },
		{ 139, 266, 45 },
		{ 139, 267, 45 },
		{ 139, 268, 45 },
		{ 139, 269, 45 },
		{ 139, 270, 45 },
		{ 139, 271, 45 },
		{ 139, 272, 45 },
		{ 139, 273, 45 },
		{ 139, 274, 45 },
		{ 139, 275, 45 },
		{ 139, 276, 45 },
		{ 139, 277, 45 },
		{ 139, 278, 45 },
		{ 139, 279, 45 },
		{ 139, 280, 45 },
		{ 139, 281, 45 },
		{ 139, 282, 45 },
		{ 139, 283, 45 },
		{ 139, 284, 45 },
		{ 139, 285, 45 },
		{ 139, 286, 45 },
		{ 140, 258, 45 },
		{ 140, 259, 45 },
		{ 140, 260, 45 },
		{ 140, 261, 45 },
		{ 140, 262, 45 },
		{ 140, 263, 45 },
		{ 140, 264, 45 },
		{ 140, 265, 45 },
		{ 140, 266, 45 },
		{ 140, 267, 45 },
		{ 140, 268, 45 },
		{ 140, 269, 45 },
		{ 140, 270, 45 },
		{ 140, 271, 45 },
		{ 140, 272, 45 },
		{ 140, 273, 45 },
		{ 140, 274, 45 },
		{ 140, 275, 45 },
		{ 140, 276, 45 },
		{ 140, 277, 45 },
		{ 140, 278, 45 },
		{ 140, 279, 45 },
		{ 140, 280, 45 },
		{ 140, 281, 45 },
		{ 140, 282, 45 },
		{ 140, 283, 45 },
		{ 140, 284, 45 },
		{ 140, 285, 45 },
		{ 140, 286, 45 },
		{ 141, 258, 45 },
		{ 141, 259, 45 },
		{ 141, 260, 45 },
		{ 141, 261, 45 },
		{ 141, 262, 45 },
		{ 141, 263, 45 },
		{ 141, 264, 45 },
		{ 141, 265, 45 },
		{ 141, 266, 45 },
		{ 141, 267, 45 },
		{ 141, 268, 45 },
		{ 141, 269, 45 },
		{ 141, 270, 45 },
		{ 141, 271, 45 },
		{ 141, 272, 45 },
		{ 141, 273, 45 },
		{ 141, 274, 45 },
		{ 141, 275, 45 },
		{ 141, 276, 45 },
		{ 141, 277, 45 },
		{ 141, 278, 45 },
		{ 141, 279, 45 },
		{ 141, 280, 45 },
		{ 141, 281, 45 },
		{ 141, 282, 45 },
		{ 141, 283, 45 },
		{ 141, 284, 45 },
		{ 141, 285, 45 },
		{ 142, 258, 45 },
		{ 142, 259, 45 },
		{ 142, 260, 45 },
		{ 142, 261, 45 },
		{ 142, 262, 45 },
		{ 142, 263, 45 },
		{ 142, 264, 45 },
		{ 142, 265, 45 },
		{ 142, 266, 45 },
		{ 142, 267, 45 },
		{ 142, 268, 45 },
		{ 142, 269, 45 },
		{ 142, 270, 45 },
		{ 142, 271, 45 },
		{ 142, 272, 45 },
		{ 142, 273, 45 },
		{ 142, 274, 45 },
		{ 142, 275, 45 },
		{ 142, 276, 45 },
		{ 142, 277, 45 },
		{ 142, 278, 45 },
		{ 142, 279, 45 },
		{ 142, 280, 45 },
		{ 142, 281, 45 },
		{ 142, 282, 45 },
		{ 142, 283, 45 },
		{ 142, 284, 45 },
		{ 142, 285, 45 },
		{ 143, 258, 45 },
		{ 143, 259, 45 },
		{ 143, 260, 45 },
		{ 143, 261, 45 },
		{ 143, 262, 45 },
		{ 143, 263, 45 },
		{ 143, 264, 45 },
		{ 143, 265, 45 },
		{ 143, 266, 45 },
		{ 143, 267, 45 },
		{ 143, 268, 45 },
		{ 143, 269, 45 },
		{ 143, 270, 45 },
		{ 143, 271, 45 },
		{ 143, 272, 45 },
		{ 143, 273, 45 },
		{ 143, 274, 45 },
		{ 143, 275, 45 },
		{ 143, 276, 45 },
		{ 143, 277, 45 },
		{ 143, 278, 45 },
		{ 143, 279, 45 },
		{ 143, 280, 45 },
		{ 143, 281, 45 },
		{ 143, 282, 45 },
		{ 143, 283, 45 },
		{ 143, 284, 45 },
		{ 143, 285, 45 },
		{ 144, 258, 45 },
		{ 144, 259, 45 },
		{ 144, 260, 45 },
		{ 144, 261, 45 },
		{ 144, 262, 45 },
		{ 144, 263, 45 },
		{ 144, 264, 45 },
		{ 144, 265, 45 },
		{ 144, 266, 45 },
		{ 144, 267, 45 },
		{ 144, 268, 45 },
		{ 144, 269, 45 },
		{ 144, 270, 45 },
		{ 144, 271, 45 },
		{ 144, 272, 45 },
		{ 144, 273, 45 },
		{ 144, 274, 45 },
		{ 144, 275, 45 },
		{ 144, 276, 45 },
		{ 144, 277, 45 },
		{ 144, 278, 45 },
		{ 144, 279, 45 },
		{ 144, 280, 45 },
		{ 144, 281, 45 },
		{ 144, 282, 45 },
		{ 144, 283, 45 },
		{ 144, 284, 45 },
		{ 144, 285, 45 },
		{ 145, 258, 45 },
		{ 145, 259, 45 },
		{ 145, 260, 45 },
		{ 145, 261, 45 },
		{ 145, 262, 45 },
		{ 145, 263, 45 },
		{ 145, 264, 45 },
		{ 145, 265, 45 },
		{ 145, 266, 45 },
		{ 145, 267, 45 },
		{ 145, 268, 45 },
		{ 145, 269, 45 },
		{ 145, 270, 45 },
		{ 145, 271, 45 },
		{ 145, 272, 45 },
		{ 145, 273, 45 },
		{ 145, 274, 45 },
		{ 145, 275, 45 },
		{ 145, 276, 45 },
		{ 145, 277, 45 },
		{ 145, 278, 45 },
		{ 145, 279, 45 },
		{ 145, 280, 45 },
		{ 145, 281, 45 },
		{ 145, 282, 45 },
		{ 145, 283, 45 },
		{ 145, 284, 45 },
		{ 146, 258, 45 },
		{ 146, 259, 45 },
		{ 146, 260, 45 },
		{ 146, 261, 45 },
		{ 146, 262, 45 },
		{ 146, 263, 45 },
		{ 146, 264, 45 },
		{ 146, 265, 45 },
		{ 146, 266, 45 },
		{ 146, 267, 45 },
		{ 146, 268, 45 },
		{ 146, 269, 45 },
		{ 146, 270, 45 },
		{ 146, 271, 45 },
		{ 146, 272, 45 },
		{ 146, 273, 45 },
		{ 146, 274, 45 },
		{ 146, 275, 45 },
		{ 146, 276, 45 },
		{ 146, 277, 45 },
		{ 146, 278, 45 },
		{ 146, 279, 45 },
		{ 146, 280, 45 },
		{ 146, 281, 45 },
		{ 146, 282, 45 },
		{ 146, 283, 45 },
		{ 146, 284, 45 },
		{ 147, 258, 45 },
		{ 147, 259, 45 },
		{ 147, 260, 45 },
		{ 147, 261, 45 },
		{ 147, 262, 45 },
		{ 147, 263, 45 },
		{ 147, 264, 45 },
		{ 147, 265, 45 },
		{ 147, 266, 45 },
		{ 147, 267, 45 },
		{ 147, 268, 45 },
		{ 147, 269, 45 },
		{ 147, 270, 45 },
		{ 147, 271, 45 },
		{ 147, 272, 45 },
		{ 147, 273, 45 },
		{ 147, 274, 45 },
		{ 147, 275, 45 },
		{ 147, 276, 45 },
		{ 147, 277, 45 },
		{ 147, 278, 45 },
		{ 147, 279, 45 },
		{ 147, 280, 45 },
		{ 147, 281, 45 },
		{ 147, 282, 45 },
		{ 147, 283, 45 },
		{ 147, 284, 45 },
		{ 148, 258, 45 },
		{ 148, 259, 45 },
		{ 148, 260, 45 },
		{ 148, 261, 45 },
		{ 148, 262, 45 },
		{ 148, 263, 45 },
		{ 148, 264, 45 },
		{ 148, 265, 45 },
		{ 148, 266, 45 },
		{ 148, 267, 45 },
		{ 148, 268, 45 },
		{ 148, 269, 45 },
		{ 148, 270, 45 },
		{ 148, 271, 45 },
		{ 148, 272, 45 },
		{ 148, 273, 45 },
		{ 148, 274, 45 },
		{ 148, 275, 45 },
		{ 148, 276, 45 },
		{ 148, 277, 45 },
		{ 148, 278, 45 },
		{ 148, 279, 45 },
		{ 148, 280, 45 },
		{ 148, 281, 45 },
		{ 148, 282, 45 },
		{ 148, 283, 45 },
		{ 148, 284, 45 },
		{ 149, 258, 45 },
		{ 149, 259, 45 },
		{ 149, 260, 45 },
		{ 149, 261, 45 },
		{ 149, 262, 45 },
		{ 149, 263, 45 },
		{ 149, 264, 45 },
		{ 149, 265, 45 },
		{ 149, 266, 45 },
		{ 149, 267, 45 },
		{ 149, 268, 45 },
		{ 149, 269, 45 },
		{ 149, 270, 45 },
		{ 149, 271, 45 },
		{ 149, 272, 45 },
		{ 149, 273, 45 },
		{ 149, 274, 45 },
		{ 149, 275, 45 },
		{ 149, 276, 45 },
		{ 149, 277, 45 },
		{ 149, 278, 45 },
		{ 149, 279, 45 },
		{ 149, 280, 45 },
		{ 149, 281, 45 },
		{ 149, 282, 45 },
		{ 149, 283, 45 },
		{ 150, 258, 45 },
		{ 150, 259, 45 },
		{ 150, 260, 45 },
		{ 150, 261, 45 },
		{ 150, 262, 45 },
		{ 150, 263, 45 },
		{ 150, 264, 45 },
		{ 150, 265, 45 },
		{ 150, 266, 45 },
		{ 150, 267, 45 },
		{ 150, 268, 45 },
		{ 150, 269, 45 },
		{ 150, 270, 45 },
		{ 150, 271, 45 },
		{ 150, 272, 45 },
		{ 150, 273, 45 },
		{ 150, 274, 45 },
		{ 150, 275, 45 },
		{ 150, 276, 45 },
		{ 150, 277, 45 },
		{ 150, 278, 45 },
		{ 150, 279, 45 },
		{ 150, 280, 45 },
		{ 150, 281, 45 },
		{ 150, 282, 45 },
		{ 150, 283, 45 },
		{ 151, 258, 45 },
		{ 151, 259, 45 },
		{ 151, 260, 45 },
		{ 151, 261, 45 },
		{ 151, 262, 45 },
		{ 151, 263, 45 },
		{ 151, 264, 45 },
		{ 151, 265, 45 },
		{ 151, 266, 45 },
		{ 151, 267, 45 },
		{ 151, 268, 45 },
		{ 151, 269, 45 },
		{ 151, 270, 45 },
		{ 151, 271, 45 },
		{ 151, 272, 45 },
		{ 151, 273, 45 },
		{ 151, 274, 45 },
		{ 151, 275, 45 },
		{ 151, 276, 45 },
		{ 151, 277, 45 },
		{ 151, 278, 45 },
		{ 151, 279, 45 },
		{ 151, 280, 45 },
		{ 151, 281, 45 },
		{ 151, 282, 45 },
		{ 151, 283, 45 },
		{ 152, 258, 45 },
		{ 152, 259, 45 },
		{ 152, 260, 45 },
		{ 152, 261, 45 },
		{ 152, 262, 45 },
		{ 152, 263, 45 },
		{ 152, 264, 45 },
		{ 152, 265, 45 },
		{ 152, 266, 45 },
		{ 152, 267, 45 },
		{ 152, 268, 45 },
		{ 152, 269, 45 },
		{ 152, 270, 45 },
		{ 152, 271, 45 },
		{ 152, 272, 45 },
		{ 152, 273, 45 },
		{ 152, 274, 45 },
		{ 152, 275, 45 },
		{ 152, 276, 45 },
		{ 152, 277, 45 },
		{ 152, 278, 45 },
		{ 152, 279, 45 },
		{ 152, 280, 45 },
		{ 152, 281, 45 },
		{ 152, 282, 45 },
		{ 152, 283, 45 },
		{ 153, 258, 45 },
		{ 153, 259, 45 },
		{ 153, 260, 45 },
		{ 153, 261, 45 },
		{ 153, 262, 45 },
		{ 153, 263, 45 },
		{ 153, 264, 45 },
		{ 153, 265, 45 },
		{ 153, 266, 45 },
		{ 153, 267, 45 },
		{ 153, 268, 45 },
		{ 153, 269, 45 },
		{ 153, 270, 45 },
		{ 153, 271, 45 },
		{ 153, 272, 45 },
		{ 153, 273, 45 },
		{ 153, 274, 45 },
		{ 153, 275, 45 },
		{ 153, 276, 45 },
		{ 153, 277, 45 },
		{ 153, 278, 45 },
		{ 153, 279, 45 },
		{ 153, 280, 45 },
		{ 153, 281, 45 },
		{ 153, 282, 45 },
		{ 154, 258, 45 },
		{ 154, 259, 45 },
		{ 154, 260, 45 },
		{ 154, 261, 45 },
		{ 154, 262, 45 },
		{ 154, 263, 45 },
		{ 154, 264, 45 },
		{ 154, 265, 45 },
		{ 154, 266, 45 },
		{ 154, 267, 45 },
		{ 154, 268, 45 },
		{ 154, 269, 45 },
		{ 154, 270, 45 },
		{ 154, 271, 45 },
		{ 154, 272, 45 },
		{ 154, 273, 45 },
		{ 154, 274, 45 },
		{ 154, 275, 45 },
		{ 154, 276, 45 },
		{ 154, 277, 45 },
		{ 154, 278, 45 },
		{ 154, 279, 45 },
		{ 154, 280, 45 },
		{ 154, 281, 45 },
		{ 154, 282, 45 },
		{ 155, 258, 45 },
		{ 155, 259, 45 },
		{ 155, 260, 45 },
		{ 155, 261, 45 },
		{ 155, 262, 45 },
		{ 155, 263, 45 },
		{ 155, 264, 45 },
		{ 155, 265, 45 },
		{ 155, 266, 45 },
		{ 155, 267, 45 },
		{ 155, 268, 45 },
		{ 155, 269, 45 },
		{ 155, 270, 45 },
		{ 155, 271, 45 },
		{ 155, 272, 45 },
		{ 155, 273, 45 },
		{ 155, 274, 45 },
		{ 155, 275, 45 },
		{ 155, 276, 45 },
		{ 155, 277, 45 },
		{ 155, 278, 45 },
		{ 155, 279, 45 },
		{ 155, 280, 45 },
		{ 155, 281, 45 },
		{ 155, 282, 45 },
		{ 156, 258, 45 },
		{ 156, 259, 45 },
		{ 156, 260, 45 },
		{ 156, 261, 45 },
		{ 156, 262, 45 },
		{ 156, 263, 45 },
		{ 156, 264, 45 },
		{ 156, 265, 45 },
		{ 156, 266, 45 },
		{ 156, 267, 45 },
		{ 156, 268, 45 },
		{ 156, 269, 45 },
		{ 156, 270, 45 },
		{ 156, 271, 45 },
		{ 156, 272, 45 },
		{ 156, 273, 45 },
		{ 156, 274, 45 },
		{ 156, 275, 45 },
		{ 156, 276, 45 },
		{ 156, 277, 45 },
		{ 156, 278, 45 },
		{ 156, 279, 45 },
		{ 156, 280, 45 },
		{ 156, 281, 45 },
		{ 156, 282, 45 },
		{ 157, 258, 45 },
		{ 157, 259, 45 },
		{ 157, 260, 45 },
		{ 157, 261, 45 },
		{ 157, 262, 45 },
		{ 157, 263, 45 },
		{ 157, 264, 45 },
		{ 157, 265, 45 },
		{ 157, 266, 45 },
		{ 157, 267, 45 },
		{ 157, 268, 45 },
		{ 157, 269, 45 },
		{ 157, 270, 45 },
		{ 157, 271, 45 },
		{ 157, 272, 45 },
		{ 157, 273, 45 },
		{ 157, 274, 45 },
		{ 157, 275, 45 },
		{ 157, 276, 45 },
		{ 157, 277, 45 },
		{ 157, 278, 45 },
		{ 157, 279, 45 },
		{ 157, 280, 45 },
		{ 157, 281, 45 },
		{ 158, 258, 45 },
		{ 158, 259, 45 },
		{ 158, 260, 45 },
		{ 158, 261, 45 },
		{ 158, 262, 45 },
		{ 158, 263, 45 },
		{ 158, 264, 45 },
		{ 158, 265, 45 },
		{ 158, 266, 45 },
		{ 158, 267, 45 },
		{ 158, 268, 45 },
		{ 158, 269, 45 },
		{ 158, 270, 45 },
		{ 158, 271, 45 },
		{ 158, 272, 45 },
		{ 158, 273, 45 },
		{ 158, 274, 45 },
		{ 158, 275, 45 },
		{ 158, 276, 45 },
		{ 158, 277, 45 },
		{ 158, 278, 45 },
		{ 158, 279, 45 },
		{ 158, 280, 45 },
		{ 158, 281, 45 },
		{ 159, 258, 45 },
		{ 159, 259, 45 },
		{ 159, 260, 45 },
		{ 159, 261, 45 },
		{ 159, 262, 45 },
		{ 159, 263, 45 },
		{ 159, 264, 45 },
		{ 159, 265, 45 },
		{ 159, 266, 45 },
		{ 159, 267, 45 },
		{ 159, 268, 45 },
		{ 159, 269, 45 },
		{ 159, 270, 45 },
		{ 159, 271, 45 },
		{ 159, 272, 45 },
		{ 159, 273, 45 },
		{ 159, 274, 45 },
		{ 159, 275, 45 },
		{ 159, 276, 45 },
		{ 159, 277, 45 },
		{ 159, 278, 45 },
		{ 159, 279, 45 },
		{ 159, 280, 45 },
		{ 159, 281, 45 },
		{ 160, 258, 45 },
		{ 160, 259, 45 },
		{ 160, 260, 45 },
		{ 160, 261, 45 },
		{ 160, 262, 45 },
		{ 160, 263, 45 },
		{ 160, 264, 45 },
		{ 160, 265, 45 },
		{ 160, 266, 45 },
		{ 160, 267, 45 },
		{ 160, 268, 45 },
		{ 160, 269, 45 },
		{ 160, 270, 45 },
		{ 160, 271, 45 },
		{ 160, 272, 45 },
		{ 160, 273, 45 },
		{ 160, 274, 45 },
		{ 160, 275, 45 },
		{ 160, 276, 45 },
		{ 160, 277, 45 },
		{ 160, 278, 45 },
		{ 160, 279, 45 },
		{ 160, 280, 45 },
		{ 160, 281, 45 },
		{ 161, 258, 45 },
		{ 161, 259, 45 },
		{ 161, 260, 45 },
		{ 161, 261, 45 },
		{ 161, 262, 45 },
		{ 161, 263, 45 },
		{ 161, 264, 45 },
		{ 161, 265, 45 },
		{ 161, 266, 45 },
		{ 161, 267, 45 },
		{ 161, 268, 45 },
		{ 161, 269, 45 },
		{ 161, 270, 45 },
		{ 161, 271, 45 },
		{ 161, 272, 45 },
		{ 161, 273, 45 },
		{ 161, 274, 45 },
		{ 161, 275, 45 },
		{ 161, 276, 45 },
		{ 161, 277, 45 },
		{ 161, 278, 45 },
		{ 161, 279, 45 },
		{ 161, 280, 45 },
		{ 162, 258, 45 },
		{ 162, 259, 45 },
		{ 162, 260, 45 },
		{ 162, 261, 45 },
		{ 162, 262, 45 },
		{ 162, 263, 45 },
		{ 162, 264, 45 },
		{ 162, 265, 45 },
		{ 162, 266, 45 },
		{ 162, 267, 45 },
		{ 162, 268, 45 },
		{ 162, 269, 45 },
		{ 162, 270, 45 },
		{ 162, 271, 45 },
		{ 162, 272, 45 },
		{ 162, 273, 45 },
		{ 162, 274, 45 },
		{ 162, 275, 45 },
		{ 162, 276, 45 },
		{ 162, 277, 45 },
		{ 162, 278, 45 },
		{ 162, 279, 45 },
		{ 162, 280, 45 },
		{ 163, 258, 45 },
		{ 163, 259, 45 },
		{ 163, 260, 45 },
		{ 163, 261, 45 },
		{ 163, 262, 45 },
		{ 163, 263, 45 },
		{ 163, 264, 45 },
		{ 163, 265, 45 },
		{ 163, 266, 45 },
		{ 163, 267, 45 },
		{ 163, 268, 45 },
		{ 163, 269, 45 },
		{ 163, 270, 45 },
		{ 163, 271, 45 },
		{ 163, 272, 45 },
		{ 163, 273, 45 },
		{ 163, 274, 45 },
		{ 163, 275, 45 },
		{ 163, 276, 45 },
		{ 163, 277, 45 },
		{ 163, 278, 45 },
		{ 163, 279, 45 },
		{ 163, 280, 45 },
		{ 164, 258, 45 },
		{ 164, 259, 45 },
		{ 164, 260, 45 },
		{ 164, 261, 45 },
		{ 164, 262, 45 },
		{ 164, 263, 45 },
		{ 164, 264, 45 },
		{ 164, 265, 45 },
		{ 164, 266, 45 },
		{ 164, 267, 45 },
		{ 164, 268, 45 },
		{ 164, 269, 45 },
		{ 164, 270, 45 },
		{ 164, 271, 45 },
		{ 164, 272, 45 },
		{ 164, 273, 45 },
		{ 164, 274, 45 },
		{ 164, 275, 45 },
		{ 164, 276, 45 },
		{ 164, 277, 45 },
		{ 164, 278, 45 },
		{ 164, 279, 45 },
		{ 164, 280, 45 },
		{ 165, 258, 45 },
		{ 165, 259, 45 },
		{ 165, 260, 45 },
		{ 165, 261, 45 },
		{ 165, 262, 45 },
		{ 165, 263, 45 },
		{ 165, 264, 45 },
		{ 165, 265, 45 },
		{ 165, 266, 45 },
		{ 165, 267, 45 },
		{ 165, 268, 45 },
		{ 165, 269, 45 },
		{ 165, 270, 45 },
		{ 165, 271, 45 },
		{ 165, 272, 45 },
		{ 165, 273, 45 },
		{ 165, 274, 45 },
		{ 165, 275, 45 },
		{ 165, 276, 45 },
		{ 165, 277, 45 },
		{ 165, 278, 45 },
		{ 165, 279, 45 },
		{ 166, 258, 45 },
		{ 166, 259, 45 },
		{ 166, 260, 45 },
		{ 166, 261, 45 },
		{ 166, 262, 45 },
		{ 166, 263, 45 },
		{ 166, 264, 45 },
		{ 166, 265, 45 },
		{ 166, 266, 45 },
		{ 166, 267, 45 },
		{ 166, 268, 45 },
		{ 166, 269, 45 },
		{ 166, 270, 45 },
		{ 166, 271, 45 },
		{ 166, 272, 45 },
		{ 166, 273, 45 },
		{ 166, 274, 45 },
		{ 166, 275, 45 },
		{ 166, 276, 45 },
		{ 166, 277, 45 },
		{ 166, 278, 45 },
		{ 166, 279, 45 },
		{ 167, 258, 45 },
		{ 167, 259, 45 },
		{ 167, 260, 45 },
		{ 167, 261, 45 },
		{ 167, 262, 45 },
		{ 167, 263, 45 },
		{ 167, 264, 45 },
		{ 167, 265, 45 },
		{ 167, 266, 45 },
		{ 167, 267, 45 },
		{ 167, 268, 45 },
		{ 167, 269, 45 },
		{ 167, 270, 45 },
		{ 167, 271, 45 },
		{ 167, 272, 45 },
		{ 167, 273, 45 },
		{ 167, 274, 45 },
		{ 167, 275, 45 },
		{ 167, 276, 45 },
		{ 167, 277, 45 },
		{ 167, 278, 45 },
		{ 167, 279, 45 },
		{ 168, 258, 45 },
		{ 168, 259, 45 },
		{ 168, 260, 45 },
		{ 168, 261, 45 },
		{ 168, 262, 45 },
		{ 168, 263, 45 },
		{ 168, 264, 45 },
		{ 168, 265, 45 },
		{ 168, 266, 45 },
		{ 168, 267, 45 },
		{ 168, 268, 45 },
		{ 168, 269, 45 },
		{ 168, 270, 45 },
		{ 168, 271, 45 },
		{ 168, 272, 45 },
		{ 168, 273, 45 },
		{ 168, 274, 45 },
		{ 168, 275, 45 },
		{ 168, 276, 45 },
		{ 168, 277, 45 },
		{ 168, 278, 45 },
		{ 168, 279, 45 },
		{ 169, 258, 45 },
		{ 169, 259, 45 },
		{ 169, 260, 45 },
		{ 169, 261, 45 },
		{ 169, 262, 45 },
		{ 169, 263, 45 },
		{ 169, 264, 45 },
		{ 169, 265, 45 },
		{ 169, 266, 45 },
		{ 169, 267, 45 },
		{ 169, 268, 45 },
		{ 169, 269, 45 },
		{ 169, 270, 45 },
		{ 169, 271, 45 },
		{ 169, 272, 45 },
		{ 169, 273, 45 },
		{ 169, 274, 45 },
		{ 169, 275, 45 },
		{ 169, 276, 45 },
		{ 169, 277, 45 },
		{ 169, 278, 45 },
		{ 170, 258, 45 },
		{ 170, 259, 45 },
		{ 170, 260, 45 },
		{ 170, 261, 45 },
		{ 170, 262, 45 },
		{ 170, 263, 45 },
		{ 170, 264, 45 },
		{ 170, 265, 45 },
		{ 170, 266, 45 },
		{ 170, 267, 45 },
		{ 170, 268, 45 },
		{ 170, 269, 45 },
		{ 170, 270, 45 },
		{ 170, 271, 45 },
		{ 170, 272, 45 },
		{ 170, 273, 45 },
		{ 170, 274, 45 },
		{ 170, 275, 45 },
		{ 170, 276, 45 },
		{ 170, 277, 45 },
		{ 170, 278, 45 },
		{ 171, 258, 45 },
		{ 171, 259, 45 },
		{ 171, 260, 45 },
		{ 171, 261, 45 },
		{ 171, 262, 45 },
		{ 171, 263, 45 },
		{ 171, 264, 45 },
		{ 171, 265, 45 },
		{ 171, 266, 45 },
		{ 171, 267, 45 },
		{ 171, 268, 45 },
		{ 171, 269, 45 },
		{ 171, 270, 45 },
		{ 171, 271, 45 },
		{ 171, 272, 45 },
		{ 171, 273, 45 },
		{ 171, 274, 45 },
		{ 171, 275, 45 },
		{ 171, 276, 45 },
		{ 171, 277, 45 },
		{ 171, 278, 45 },
		{ 172, 258, 45 },
		{ 172, 259, 45 },
		{ 172, 260, 45 },
		{ 172, 261, 45 },
		{ 172, 262, 45 },
		{ 172, 263, 45 },
		{ 172, 264, 45 },
		{ 172, 265, 45 },
		{ 172, 266, 45 },
		{ 172, 267, 45 },
		{ 172, 268, 45 },
		{ 172, 269, 45 },
		{ 172, 270, 45 },
		{ 172, 271, 45 },
		{ 172, 272, 45 },
		{ 172, 273, 45 },
		{ 172, 274, 45 },
		{ 172, 275, 45 },
		{ 172, 276, 45 },
		{ 172, 277, 45 },
		{ 172, 278, 45 },
		{ 173, 258, 45 },
		{ 173, 259, 45 },
		{ 173, 260, 45 },
		{ 173, 261, 45 },
		{ 173, 262, 45 },
		{ 173, 263, 45 },
		{ 173, 264, 45 },
		{ 173, 265, 45 },
		{ 173, 266, 45 },
		{ 173, 267, 45 },
		{ 173, 268, 45 },
		{ 173, 269, 45 },
		{ 173, 270, 45 },
		{ 173, 271, 45 },
		{ 173, 272, 45 },
		{ 173, 273, 45 },
		{ 173, 274, 45 },
		{ 173, 275, 45 },
		{ 173, 276, 45 },
		{ 173, 277, 45 },
		{ 174, 258, 45 },
		{ 174, 259, 45 },
		{ 174, 260, 45 },
		{ 174, 261, 45 },
		{ 174, 262, 45 },
		{ 174, 263, 45 },
		{ 174, 264, 45 },
		{ 174, 265, 45 },
		{ 174, 266, 45 },
		{ 174, 267, 45 },
		{ 174, 268, 45 },
		{ 174, 269, 45 },
		{ 174, 270, 45 },
		{ 174, 271, 45 },
		{ 174, 272, 45 },
		{ 174, 273, 45 },
		{ 174, 274, 45 },
		{ 174, 275, 45 },
		{ 174, 276, 45 },
		{ 174, 277, 45 },
		{ 175, 258, 45 },
		{ 175, 259, 45 },
		{ 175, 260, 45 },
		{ 175, 261, 45 },
		{ 175, 262, 45 },
		{ 175, 263, 45 },
		{ 175, 264, 45 },
		{ 175, 265, 45 },
		{ 175, 266, 45 },
		{ 175, 267, 45 },
		{ 175, 268, 45 },
		{ 175, 269, 45 },
		{ 175, 270, 45 },
		{ 175, 271, 45 },
		{ 175, 272, 45 },
		{ 175, 273, 45 },
		{ 175, 274, 45 },
		{ 175, 275, 45 },
		{ 175, 276, 45 },
		{ 175, 277, 45 },
		{ 176, 258, 45 },
		{ 176, 259, 45 },
		{ 176, 260, 45 },
		{ 176, 261, 45 },
		{ 176, 262, 45 },
		{ 176, 263, 45 },
		{ 176, 264, 45 },
		{ 176, 265, 45 },
		{ 176, 266, 45 },
		{ 176, 267, 45 },
		{ 176, 268, 45 },
		{ 176, 269, 45 },
		{ 176, 270, 45 },
		{ 176, 271, 45 },
		{ 176, 272, 45 },
		{ 176, 273, 45 },
		{ 176, 274, 45 },
		{ 176, 275, 45 },
		{ 176, 276, 45 },
		{ 176, 277, 45 },
		{ 177, 258, 45 },
		{ 177, 259, 45 },
		{ 177, 260, 45 },
		{ 177, 261, 45 },
		{ 177, 262, 45 },
		{ 177, 263, 45 },
		{ 177, 264, 45 },
		{ 177, 265, 45 },
		{ 177, 266, 45 },
		{ 177, 267, 45 },
		{ 177, 268, 45 },
		{ 177, 269, 45 },
		{ 177, 270, 45 },
		{ 177, 271, 45 },
		{ 177, 272, 45 },
		{ 177, 273, 45 },
		{ 177, 274, 45 },
		{ 177, 275, 45 },
		{ 177, 276, 45 },
		{ 178, 258, 45 },
		{ 178, 259, 45 },
		{ 178, 260, 45 },
		{ 178, 261, 45 },
		{ 178, 262, 45 },
		{ 178, 263, 45 },
		{ 178, 264, 45 },
		{ 178, 265, 45 },
		{ 178, 266, 45 },
		{ 178, 267, 45 },
		{ 178, 268, 45 },
		{ 178, 269, 45 },
		{ 178, 270, 45 },
		{ 178, 271, 45 },
		{ 178, 272, 45 },
		{ 178, 273, 45 },
		{ 178, 274, 45 },
		{ 178, 275, 45 },
		{ 178, 276, 45 },
		{ 179, 258, 45 },
		{ 179, 259, 45 },
		{ 179, 260, 45 },
		{ 179, 261, 45 },
		{ 179, 262, 45 },
		{ 179, 263, 45 },
		{ 179, 264, 45 },
		{ 179, 265, 45 },
		{ 179, 266, 45 },
		{ 179, 267, 45 },
		{ 179, 268, 45 },
		{ 179, 269, 45 },
		{ 179, 270, 45 },
		{ 179, 271, 45 },
		{ 179, 272, 45 },
		{ 179, 273, 45 },
		{ 179, 274, 45 },
		{ 179, 275, 45 },
		{ 179, 276, 45 },
		{ 180, 258, 45 },
		{ 180, 259, 45 },
		{ 180, 260, 45 },
		{ 180, 261, 45 },
		{ 180, 262, 45 },
		{ 180, 263, 45 },
		{ 180, 264, 45 },
		{ 180, 265, 45 },
		{ 180, 266, 45 },
		{ 180, 267, 45 },
		{ 180, 268, 45 },
		{ 180, 269, 45 },
		{ 180, 270, 45 },
		{ 180, 271, 45 },
		{ 180, 272, 45 },
		{ 180, 273, 45 },
		{ 180, 274, 45 },
		{ 180, 275, 45 },
		{ 180, 276, 45 },
		{ 181, 258, 45 },
		{ 181, 259, 45 },
		{ 181, 260, 45 },
		{ 181, 261, 45 },
		{ 181, 262, 45 },
		{ 181, 263, 45 },
		{ 181, 264, 45 },
		{ 181, 265, 45 },
		{ 181, 266, 45 },
		{ 181, 267, 45 },
		{ 181, 268, 45 },
		{ 181, 269, 45 },
		{ 181, 270, 45 },
		{ 181, 271, 45 },
		{ 181, 272, 45 },
		{ 181, 273, 45 },
		{ 181, 274, 45 },
		{ 181, 275, 45 },
		{ 182, 258, 45 },
		{ 182, 259, 45 },
		{ 182, 260, 45 },
		{ 182, 261, 45 },
		{ 182, 262, 45 },
		{ 182, 263, 45 },
		{ 182, 264, 45 },
		{ 182, 265, 45 },
		{ 182, 266, 45 },
		{ 182, 267, 45 },
		{ 182, 268, 45 },
		{ 182, 269, 45 },
		{ 182, 270, 45 },
		{ 182, 271, 45 },
		{ 182, 272, 45 },
		{ 182, 273, 45 },
		{ 182, 274, 45 },
		{ 182, 275, 45 },
		{ 183, 258, 45 },
		{ 183, 259, 45 },
		{ 183, 260, 45 },
		{ 183, 261, 45 },
		{ 183, 262, 45 },
		{ 183, 263, 45 },
		{ 183, 264, 45 },
		{ 183, 265, 45 },
		{ 183, 266, 45 },
		{ 183, 267, 45 },
		{ 183, 268, 45 },
		{ 183, 269, 45 },
		{ 183, 270, 45 },
		{ 183, 271, 45 },
		{ 183, 272, 45 },
		{ 183, 273, 45 },
		{ 183, 274, 45 },
		{ 183, 275, 45 },
		{ 184, 258, 45 },
		{ 184, 259, 45 },
		{ 184, 260, 45 },
		{ 184, 261, 45 },
		{ 184, 262, 45 },
		{ 184, 263, 45 },
		{ 184, 264, 45 },
		{ 184, 265, 45 },
		{ 184, 266, 45 },
		{ 184, 267, 45 },
		{ 184, 268, 45 },
		{ 184, 269, 45 },
		{ 184, 270, 45 },
		{ 184, 271, 45 },
		{ 184, 272, 45 },
		{ 184, 273, 45 },
		{ 184, 274, 45 },
		{ 184, 275, 45 },
		{ 185, 258, 45 },
		{ 185, 259, 45 },
		{ 185, 260, 45 },
		{ 185, 261, 45 },
		{ 185, 262, 45 },
		{ 185, 263, 45 },
		{ 185, 264, 45 },
		{ 185, 265, 45 },
		{ 185, 266, 45 },
		{ 185, 267, 45 },
		{ 185, 268, 45 },
		{ 185, 269, 45 },
		{ 185, 270, 45 },
		{ 185, 271, 45 },
		{ 185, 272, 45 },
		{ 185, 273, 45 },
		{ 185, 274, 45 },
		{ 186, 258, 45 },
		{ 186, 259, 45 },
		{ 186, 260, 45 },
		{ 186, 261, 45 },
		{ 186, 262, 45 },
		{ 186, 263, 45 },
		{ 186, 264, 45 },
		{ 186, 265, 45 },
		{ 186, 266, 45 },
		{ 186, 267, 45 },
		{ 186, 268, 45 },
		{ 186, 269, 45 },
		{ 186, 270, 45 },
		{ 186, 271, 45 },
		{ 186, 272, 45 },
		{ 186, 273, 45 },
		{ 186, 274, 45 },
		{ 187, 258, 45 },
		{ 187, 259, 45 },
		{ 187, 260, 45 },
		{ 187, 261, 45 },
		{ 187, 262, 45 },
		{ 187, 263, 45 },
		{ 187, 264, 45 },
		{ 187, 265, 45 },
		{ 187, 266, 45 },
		{ 187, 267, 45 },
		{ 187, 268, 45 },
		{ 187, 269, 45 },
		{ 187, 270, 45 },
		{ 187, 271, 45 },
		{ 187, 272, 45 },
		{ 187, 273, 45 },
		{ 187, 274, 45 },
		{ 188, 258, 45 },
		{ 188, 259, 45 },
		{ 188, 260, 45 },
		{ 188, 261, 45 },
		{ 188, 262, 45 },
		{ 188, 263, 45 },
		{ 188, 264, 45 },
		{ 188, 265, 45 },
		{ 188, 266, 45 },
		{ 188, 267, 45 },
		{ 188, 268, 45 },
		{ 188, 269, 45 },
		{ 188, 270, 45 },
		{ 188, 271, 45 },
		{ 188, 272, 45 },
		{ 188, 273, 45 },
		{ 189, 258, 45 },
		{ 189, 259, 45 },
		{ 189, 260, 45 },
		{ 189, 261, 45 },
		{ 189, 262, 45 },
		{ 189, 263, 45 },
		{ 189, 264, 45 },
		{ 189, 265, 45 },
		{ 189, 266, 45 },
		{ 189, 267, 45 },
		{ 189, 268, 45 },
		{ 189, 269, 45 },
		{ 189, 270, 45 },
		{ 189, 271, 45 },
		{ 189, 272, 45 },
		{ 189, 273, 45 },
		{ 190, 258, 45 },
		{ 190, 259, 45 },
		{ 190, 260, 45 },
		{ 190, 261, 45 },
		{ 190, 262, 45 },
		{ 190, 263, 45 },
		{ 190, 264, 45 },
		{ 190, 265, 45 },
		{ 190, 266, 45 },
		{ 190, 267, 45 },
		{ 190, 268, 45 },
		{ 190, 269, 45 },
		{ 190, 270, 45 },
		{ 190, 271, 45 },
		{ 190, 272, 45 },
		{ 190, 273, 45 },
		{ 191, 259, 45 },
		{ 191, 260, 45 },
		{ 191, 261, 45 },
		{ 191, 262, 45 },
		{ 191, 263, 45 },
		{ 191, 264, 45 },
		{ 191, 265, 45 },
		{ 191, 266, 45 },
		{ 191, 267, 45 },
		{ 191, 268, 45 },
		{ 191, 269, 45 },
		{ 191, 270, 45 },
		{ 191, 271, 45 },
		{ 323, 243, 45 },
		{ 323, 244, 45 },
		{ 323, 245, 45 },
		{ 323, 246, 45 },
		{ 323, 247, 45 },
		{ 323, 248, 45 },
		{ 323, 249, 45 },
		{ 323, 250, 45 },
		{ 323, 251, 45 },
		{ 323, 252, 45 },
		{ 323, 253, 45 },
		{ 323, 254, 45 },
		{ 323, 255, 45 },
		{ 324, 241, 45 },
		{ 324, 242, 45 },
		{ 324, 243, 45 },
		{ 324, 244, 45 },
		{ 324, 245, 45 },
		{ 324, 246, 45 },
		{ 324, 247, 45 },
		{ 324, 248, 45 },
		{ 324, 249, 45 },
		{ 324, 250, 45 },
		{ 324, 251, 45 },
		{ 324, 252, 45 },
		{ 324, 253, 45 },
		{ 324, 254, 45 },
		{ 324, 255, 45 },
		{ 324, 256, 45 },
		{ 325, 241, 45 },
		{ 325, 242, 45 },
		{ 325, 243, 45 },
		{ 325, 244, 45 },
		{ 325, 245, 45 },
		{ 325, 246, 45 },
		{ 325, 247, 45 },
		{ 325, 248, 45 },
		{ 325, 249, 45 },
		{ 325, 250, 45 },
		{ 325, 251, 45 },
		{ 325, 252, 45 },
		{ 325, 253, 45 },
		{ 325, 254, 45 },
		{ 325, 255, 45 },
		{ 325, 256, 45 },
		{ 326, 241, 45 },
		{ 326, 242, 45 },
		{ 326, 243, 45 },
		{ 326, 244, 45 },
		{ 326, 245, 45 },
		{ 326, 246, 45 },
		{ 326, 247, 45 },
		{ 326, 248, 45 },
		{ 326, 249, 45 },
		{ 326, 250, 45 },
		{ 326, 251, 45 },
		{ 326, 252, 45 },
		{ 326, 253, 45 },
		{ 326, 254, 45 },
		{ 326, 255, 45 },
		{ 326, 256, 45 },
		{ 327, 240, 45 },
		{ 327, 241, 45 },
		{ 327, 242, 45 },
		{ 327, 243, 45 },
		{ 327, 244, 45 },
		{ 327, 245, 45 },
		{ 327, 246, 45 },
		{ 327, 247, 45 },
		{ 327, 248, 45 },
		{ 327, 249, 45 },
		{ 327, 250, 45 },
		{ 327, 251, 45 },
		{ 327, 252, 45 },
		{ 327, 253, 45 },
		{ 327, 254, 45 },
		{ 327, 255, 45 },
		{ 327, 256, 45 },
		{ 328, 240, 45 },
		{ 328, 241, 45 },
		{ 328, 242, 45 },
		{ 328, 243, 45 },
		{ 328, 244, 45 },
		{ 328, 245, 45 },
		{ 328, 246, 45 },
		{ 328, 247, 45 },
		{ 328, 248, 45 },
		{ 328, 249, 45 },
		{ 328, 250, 45 },
		{ 328, 251, 45 },
		{ 328, 252, 45 },
		{ 328, 253, 45 },
		{ 328, 254, 45 },
		{ 328, 255, 45 },
		{ 328, 256, 45 },
		{ 329, 240, 45 },
		{ 329, 241, 45 },
		{ 329, 242, 45 },
		{ 329, 243, 45 },
		{ 329, 244, 45 },
		{ 329, 245, 45 },
		{ 329, 246, 45 },
		{ 329, 247, 45 },
		{ 329, 248, 45 },
		{ 329, 249, 45 },
		{ 329, 250, 45 },
		{ 329, 251, 45 },
		{ 329, 252, 45 },
		{ 329, 253, 45 },
		{ 329, 254, 45 },
		{ 329, 255, 45 },
		{ 329, 256, 45 },
		{ 330, 239, 45 },
		{ 330, 240, 45 },
		{ 330, 241, 45 },
		{ 330, 242, 45 },
		{ 330, 243, 45 },
		{ 330, 244, 45 },
		{ 330, 245, 45 },
		{ 330, 246, 45 },
		{ 330, 247, 45 },
		{ 330, 248, 45 },
		{ 330, 249, 45 },
		{ 330, 250, 45 },
		{ 330, 251, 45 },
		{ 330, 252, 45 },
		{ 330, 253, 45 },
		{ 330, 254, 45 },
		{ 330, 255, 45 },
		{ 330, 256, 45 },
		{ 331, 239, 45 },
		{ 331, 240, 45 },
		{ 331, 241, 45 },
		{ 331, 242, 45 },
		{ 331, 243, 45 },
		{ 331, 244, 45 },
		{ 331, 245, 45 },
		{ 331, 246, 45 },
		{ 331, 247, 45 },
		{ 331, 248, 45 },
		{ 331, 249, 45 },
		{ 331, 250, 45 },
		{ 331, 251, 45 },
		{ 331, 252, 45 },
		{ 331, 253, 45 },
		{ 331, 254, 45 },
		{ 331, 255, 45 },
		{ 331, 256, 45 },
		{ 332, 239, 45 },
		{ 332, 240, 45 },
		{ 332, 241, 45 },
		{ 332, 242, 45 },
		{ 332, 243, 45 },
		{ 332, 244, 45 },
		{ 332, 245, 45 },
		{ 332, 246, 45 },
		{ 332, 247, 45 },
		{ 332, 248, 45 },
		{ 332, 249, 45 },
		{ 332, 250, 45 },
		{ 332, 251, 45 },
		{ 332, 252, 45 },
		{ 332, 253, 45 },
		{ 332, 254, 45 },
		{ 332, 255, 45 },
		{ 332, 256, 45 },
		{ 333, 239, 45 },
		{ 333, 240, 45 },
		{ 333, 241, 45 },
		{ 333, 242, 45 },
		{ 333, 243, 45 },
		{ 333, 244, 45 },
		{ 333, 245, 45 },
		{ 333, 246, 45 },
		{ 333, 247, 45 },
		{ 333, 248, 45 },
		{ 333, 249, 45 },
		{ 333, 250, 45 },
		{ 333, 251, 45 },
		{ 333, 252, 45 },
		{ 333, 253, 45 },
		{ 333, 254, 45 },
		{ 333, 255, 45 },
		{ 333, 256, 45 },
		{ 334, 238, 45 },
		{ 334, 239, 45 },
		{ 334, 240, 45 },
		{ 334, 241, 45 },
		{ 334, 242, 45 },
		{ 334, 243, 45 },
		{ 334, 244, 45 },
		{ 334, 245, 45 },
		{ 334, 246, 45 },
		{ 334, 247, 45 },
		{ 334, 248, 45 },
		{ 334, 249, 45 },
		{ 334, 250, 45 },
		{ 334, 251, 45 },
		{ 334, 252, 45 },
		{ 334, 253, 45 },
		{ 334, 254, 45 },
		{ 334, 255, 45 },
		{ 334, 256, 45 },
		{ 335, 238, 45 },
		{ 335, 239, 45 },
		{ 335, 240, 45 },
		{ 335, 241, 45 },
		{ 335, 242, 45 },
		{ 335, 243, 45 },
		{ 335, 244, 45 },
		{ 335, 245, 45 },
		{ 335, 246, 45 },
		{ 335, 247, 45 },
		{ 335, 248, 45 },
		{ 335, 249, 45 },
		{ 335, 250, 45 },
		{ 335, 251, 45 },
		{ 335, 252, 45 },
		{ 335, 253, 45 },
		{ 335, 254, 45 },
		{ 335, 255, 45 },
		{ 335, 256, 45 },
		{ 336, 238, 45 },
		{ 336, 239, 45 },
		{ 336, 240, 45 },
		{ 336, 241, 45 },
		{ 336, 242, 45 },
		{ 336, 243, 45 },
		{ 336, 244, 45 },
		{ 336, 245, 45 },
		{ 336, 246, 45 },
		{ 336, 247, 45 },
		{ 336, 248, 45 },
		{ 336, 249, 45 },
		{ 336, 250, 45 },
		{ 336, 251, 45 },
		{ 336, 252, 45 },
		{ 336, 253, 45 },
		{ 336, 254, 45 },
		{ 336, 255, 45 },
		{ 336, 256, 45 },
		{ 337, 238, 45 },
		{ 337, 239, 45 },
		{ 337, 240, 45 },
		{ 337, 241, 45 },
		{ 337, 242, 45 },
		{ 337, 243, 45 },
		{ 337, 244, 45 },
		{ 337, 245, 45 },
		{ 337, 246, 45 },
		{ 337, 247, 45 },
		{ 337, 248, 45 },
		{ 337, 249, 45 },
		{ 337, 250, 45 },
		{ 337, 251, 45 },
		{ 337, 252, 45 },
		{ 337, 253, 45 },
		{ 337, 254, 45 },
		{ 337, 255, 45 },
		{ 337, 256, 45 },
		{ 338, 237, 45 },
		{ 338, 238, 45 },
		{ 338, 239, 45 },
		{ 338, 240, 45 },
		{ 338, 241, 45 },
		{ 338, 242, 45 },
		{ 338, 243, 45 },
		{ 338, 244, 45 },
		{ 338, 245, 45 },
		{ 338, 246, 45 },
		{ 338, 247, 45 },
		{ 338, 248, 45 },
		{ 338, 249, 45 },
		{ 338, 250, 45 },
		{ 338, 251, 45 },
		{ 338, 252, 45 },
		{ 338, 253, 45 },
		{ 338, 254, 45 },
		{ 338, 255, 45 },
		{ 338, 256, 45 },
		{ 339, 237, 45 },
		{ 339, 238, 45 },
		{ 339, 239, 45 },
		{ 339, 240, 45 },
		{ 339, 241, 45 },
		{ 339, 242, 45 },
		{ 339, 243, 45 },
		{ 339, 244, 45 },
		{ 339, 245, 45 },
		{ 339, 246, 45 },
		{ 339, 247, 45 },
		{ 339, 248, 45 },
		{ 339, 249, 45 },
		{ 339, 250, 45 },
		{ 339, 251, 45 },
		{ 339, 252, 45 },
		{ 339, 253, 45 },
		{ 339, 254, 45 },
		{ 339, 255, 45 },
		{ 339, 256, 45 },
		{ 340, 237, 45 },
		{ 340, 238, 45 },
		{ 340, 239, 45 },
		{ 340, 240, 45 },
		{ 340, 241, 45 },
		{ 340, 242, 45 },
		{ 340, 243, 45 },
		{ 340, 244, 45 },
		{ 340, 245, 45 },
		{ 340, 246, 45 },
		{ 340, 247, 45 },
		{ 340, 248, 45 },
		{ 340, 249, 45 },
		{ 340, 250, 45 },
		{ 340, 251, 45 },
		{ 340, 252, 45 },
		{ 340, 253, 45 },
		{ 340, 254, 45 },
		{ 340, 255, 45 },
		{ 340, 256, 45 },
		{ 341, 237, 45 },
		{ 341, 238, 45 },
		{ 341, 239, 45 },
		{ 341, 240, 45 },
		{ 341, 241, 45 },
		{ 341, 242, 45 },
		{ 341, 243, 45 },
		{ 341, 244, 45 },
		{ 341, 245, 45 },
		{ 341, 246, 45 },
		{ 341, 247, 45 },
		{ 341, 248, 45 },
		{ 341, 249, 45 },
		{ 341, 250, 45 },
		{ 341, 251, 45 },
		{ 341, 252, 45 },
		{ 341, 253, 45 },
		{ 341, 254, 45 },
		{ 341, 255, 45 },
		{ 341, 256, 45 },
		{ 342, 236, 45 },
		{ 342, 237, 45 },
		{ 342, 238, 45 },
		{ 342, 239, 45 },
		{ 342, 240, 45 },
		{ 342, 241, 45 },
		{ 342, 242, 45 },
		{ 342, 243, 45 },
		{ 342, 244, 45 },
		{ 342, 245, 45 },
		{ 342, 246, 45 },
		{ 342, 247, 45 },
		{ 342, 248, 45 },
		{ 342, 249, 45 },
		{ 342, 250, 45 },
		{ 342, 251, 45 },
		{ 342, 252, 45 },
		{ 342, 253, 45 },
		{ 342, 254, 45 },
		{ 342, 255, 45 },
		{ 342, 256, 45 },
		{ 343, 236, 45 },
		{ 343, 237, 45 },
		{ 343, 238, 45 },
		{ 343, 239, 45 },
		{ 343, 240, 45 },
		{ 343, 241, 45 },
		{ 343, 242, 45 },
		{ 343, 243, 45 },
		{ 343, 244, 45 },
		{ 343, 245, 45 },
		{ 343, 246, 45 },
		{ 343, 247, 45 },
		{ 343, 248, 45 },
		{ 343, 249, 45 },
		{ 343, 250, 45 },
		{ 343, 251, 45 },
		{ 343, 252, 45 },
		{ 343, 253, 45 },
		{ 343, 254, 45 },
		{ 343, 255, 45 },
		{ 343, 256, 45 },
		{ 344, 236, 45 },
		{ 344, 237, 45 },
		{ 344, 238, 45 },
		{ 344, 239, 45 },
		{ 344, 240, 45 },
		{ 344, 241, 45 },
		{ 344, 242, 45 },
		{ 344, 243, 45 },
		{ 344, 244, 45 },
		{ 344, 245, 45 },
		{ 344, 246, 45 },
		{ 344, 247, 45 },
		{ 344, 248, 45 },
		{ 344, 249, 45 },
		{ 344, 250, 45 },
		{ 344, 251, 45 },
		{ 344, 252, 45 },
		{ 344, 253, 45 },
		{ 344, 254, 45 },
		{ 344, 255, 45 },
		{ 344, 256, 45 },
		{ 345, 236, 45 },
		{ 345, 237, 45 },
		{ 345, 238, 45 },
		{ 345, 239, 45 },
		{ 345, 240, 45 },
		{ 345, 241, 45 },
		{ 345, 242, 45 },
		{ 345, 243, 45 },
		{ 345, 244, 45 },
		{ 345, 245, 45 },
		{ 345, 246, 45 },
		{ 345, 247, 45 },
		{ 345, 248, 45 },
		{ 345, 249, 45 },
		{ 345, 250, 45 },
		{ 345, 251, 45 },
		{ 345, 252, 45 },
		{ 345, 253, 45 },
		{ 345, 254, 45 },
		{ 345, 255, 45 },
		{ 345, 256, 45 },
		{ 346, 235, 45 },
		{ 346, 236, 45 },
		{ 346, 237, 45 },
		{ 346, 238, 45 },
		{ 346, 239, 45 },
		{ 346, 240, 45 },
		{ 346, 241, 45 },
		{ 346, 242, 45 },
		{ 346, 243, 45 },
		{ 346, 244, 45 },
		{ 346, 245, 45 },
		{ 346, 246, 45 },
		{ 346, 247, 45 },
		{ 346, 248, 45 },
		{ 346, 249, 45 },
		{ 346, 250, 45 },
		{ 346, 251, 45 },
		{ 346, 252, 45 },
		{ 346, 253, 45 },
		{ 346, 254, 45 },
		{ 346, 255, 45 },
		{ 346, 256, 45 },
		{ 347, 235, 45 },
		{ 347, 236, 45 },
		{ 347, 237, 45 },
		{ 347, 238, 45 },
		{ 347, 239, 45 },
		{ 347, 240, 45 },
		{ 347, 241, 45 },
		{ 347, 242, 45 },
		{ 347, 243, 45 },
		{ 347, 244, 45 },
		{ 347, 245, 45 },
		{ 347, 246, 45 },
		{ 347, 247, 45 },
		{ 347, 248, 45 },
		{ 347, 249, 45 },
		{ 347, 250, 45 },
		{ 347, 251, 45 },
		{ 347, 252, 45 },
		{ 347, 253, 45 },
		{ 347, 254, 45 },
		{ 347, 255, 45 },
		{ 347, 256, 45 },
		{ 348, 235, 45 },
		{ 348, 236, 45 },
		{ 348, 237, 45 },
		{ 348, 238, 45 },
		{ 348, 239, 45 },
		{ 348, 240, 45 },
		{ 348, 241, 45 },
		{ 348, 242, 45 },
		{ 348, 243, 45 },
		{ 348, 244, 45 },
		{ 348, 245, 45 },
		{ 348, 246, 45 },
		{ 348, 247, 45 },
		{ 348, 248, 45 },
		{ 348, 249, 45 },
		{ 348, 250, 45 },
		{ 348, 251, 45 },
		{ 348, 252, 45 },
		{ 348, 253, 45 },
		{ 348, 254, 45 },
		{ 348, 255, 45 },
		{ 348, 256, 45 },
		{ 349, 235, 45 },
		{ 349, 236, 45 },
		{ 349, 237, 45 },
		{ 349, 238, 45 },
		{ 349, 239, 45 },
		{ 349, 240, 45 },
		{ 349, 241, 45 },
		{ 349, 242, 45 },
		{ 349, 243, 45 },
		{ 349, 244, 45 },
		{ 349, 245, 45 },
		{ 349, 246, 45 },
		{ 349, 247, 45 },
		{ 349, 248, 45 },
		{ 349, 249, 45 },
		{ 349, 250, 45 },
		{ 349, 251, 45 },
		{ 349, 252, 45 },
		{ 349, 253, 45 },
		{ 349, 254, 45 },
		{ 349, 255, 45 },
		{ 349, 256, 45 },
		{ 350, 234, 45 },
		{ 350, 235, 45 },
		{ 350, 236, 45 },
		{ 350, 237, 45 },
		{ 350, 238, 45 },
		{ 350, 239, 45 },
		{ 350, 240, 45 },
		{ 350, 241, 45 },
		{ 350, 242, 45 },
		{ 350, 243, 45 },
		{ 350, 244, 45 },
		{ 350, 245, 45 },
		{ 350, 246, 45 },
		{ 350, 247, 45 },
		{ 350, 248, 45 },
		{ 350, 249, 45 },
		{ 350, 250, 45 },
		{ 350, 251, 45 },
		{ 350, 252, 45 },
		{ 350, 253, 45 },
		{ 350, 254, 45 },
		{ 350, 255, 45 },
		{ 350, 256, 45 },
		{ 351, 234, 45 },
		{ 351, 235, 45 },
		{ 351, 236, 45 },
		{ 351, 237, 45 },
		{ 351, 238, 45 },
		{ 351, 239, 45 },
		{ 351, 240, 45 },
		{ 351, 241, 45 },
		{ 351, 242, 45 },
		{ 351, 243, 45 },
		{ 351, 244, 45 },
		{ 351, 245, 45 },
		{ 351, 246, 45 },
		{ 351, 247, 45 },
		{ 351, 248, 45 },
		{ 351, 249, 45 },
		{ 351, 250, 45 },
		{ 351, 251, 45 },
		{ 351, 252, 45 },
		{ 351, 253, 45 },
		{ 351, 254, 45 },
		{ 351, 255, 45 },
		{ 351, 256, 45 },
		{ 352, 234, 45 },
		{ 352, 235, 45 },
		{ 352, 236, 45 },
		{ 352, 237, 45 },
		{ 352, 238, 45 },
		{ 352, 239, 45 },
		{ 352, 240, 45 },
		{ 352, 241, 45 },
		{ 352, 242, 45 },
		{ 352, 243, 45 },
		{ 352, 244, 45 },
		{ 352, 245, 45 },
		{ 352, 246, 45 },
		{ 352, 247, 45 },
		{ 352, 248, 45 },
		{ 352, 249, 45 },
		{ 352, 250, 45 },
		{ 352, 251, 45 },
		{ 352, 252, 45 },
		{ 352, 253, 45 },
		{ 352, 254, 45 },
		{ 352, 255, 45 },
		{ 352, 256, 45 },
		{ 353, 234, 45 },
		{ 353, 235, 45 },
		{ 353, 236, 45 },
		{ 353, 237, 45 },
		{ 353, 238, 45 },
		{ 353, 239, 45 },
		{ 353, 240, 45 },
		{ 353, 241, 45 },
		{ 353, 242, 45 },
		{ 353, 243, 45 },
		{ 353, 244, 45 },
		{ 353, 245, 45 },
		{ 353, 246, 45 },
		{ 353, 247, 45 },
		{ 353, 248, 45 },
		{ 353, 249, 45 },
		{ 353, 250, 45 },
		{ 353, 251, 45 },
		{ 353, 252, 45 },
		{ 353, 253, 45 },
		{ 353, 254, 45 },
		{ 353, 255, 45 },
		{ 353, 256, 45 },
		{ 354, 233, 45 },
		{ 354, 234, 45 },
		{ 354, 235, 45 },
		{ 354, 236, 45 },
		{ 354, 237, 45 },
		{ 354, 238, 45 },
		{ 354, 239, 45 },
		{ 354, 240, 45 },
		{ 354, 241, 45 },
		{ 354, 242, 45 },
		{ 354, 243, 45 },
		{ 354, 244, 45 },
		{ 354, 245, 45 },
		{ 354, 246, 45 },
		{ 354, 247, 45 },
		{ 354, 248, 45 },
		{ 354, 249, 45 },
		{ 354, 250, 45 },
		{ 354, 251, 45 },
		{ 354, 252, 45 },
		{ 354, 253, 45 },
		{ 354, 254, 45 },
		{ 354, 255, 45 },
		{ 354, 256, 45 },
		{ 355, 233, 45 },
		{ 355, 234, 45 },
		{ 355, 235, 45 },
		{ 355, 236, 45 },
		{ 355, 237, 45 },
		{ 355, 238, 45 },
		{ 355, 239, 45 },
		{ 355, 240, 45 },
		{ 355, 241, 45 },
		{ 355, 242, 45 },
		{ 355, 243, 45 },
		{ 355, 244, 45 },
		{ 355, 245, 45 },
		{ 355, 246, 45 },
		{ 355, 247, 45 },
		{ 355, 248, 45 },
		{ 355, 249, 45 },
		{ 355, 250, 45 },
		{ 355, 251, 45 },
		{ 355, 252, 45 },
		{ 355, 253, 45 },
		{ 355, 254, 45 },
		{ 355, 255, 45 },
		{ 355, 256, 45 },
		{ 356, 233, 45 },
		{ 356, 234, 45 },
		{ 356, 235, 45 },
		{ 356, 236, 45 },
		{ 356, 237, 45 },
		{ 356, 238, 45 },
		{ 356, 239, 45 },
		{ 356, 240, 45 },
		{ 356, 241, 45 },
		{ 356, 242, 45 },
		{ 356, 243, 45 },
		{ 356, 244, 45 },
		{ 356, 245, 45 },
		{ 356, 246, 45 },
		{ 356, 247, 45 },
		{ 356, 248, 45 },
		{ 356, 249, 45 },
		{ 356, 250, 45 },
		{ 356, 251, 45 },
		{ 356, 252, 45 },
		{ 356, 253, 45 },
		{ 356, 254, 45 },
		{ 356, 255, 45 },
		{ 356, 256, 45 },
		{ 357, 233, 45 },
		{ 357, 234, 45 },
		{ 357, 235, 45 },
		{ 357, 236, 45 },
		{ 357, 237, 45 },
		{ 357, 238, 45 },
		{ 357, 239, 45 },
		{ 357, 240, 45 },
		{ 357, 241, 45 },
		{ 357, 242, 45 },
		{ 357, 243, 45 },
		{ 357, 244, 45 },
		{ 357, 245, 45 },
		{ 357, 246, 45 },
		{ 357, 247, 45 },
		{ 357, 248, 45 },
		{ 357, 249, 45 },
		{ 357, 250, 45 },
		{ 357, 251, 45 },
		{ 357, 252, 45 },
		{ 357, 253, 45 },
		{ 357, 254, 45 },
		{ 357, 255, 45 },
		{ 357, 256, 45 },
		{ 358, 232, 45 },
		{ 358, 233, 45 },
		{ 358, 234, 45 },
		{ 358, 235, 45 },
		{ 358, 236, 45 },
		{ 358, 237, 45 },
		{ 358, 238, 45 },
		{ 358, 239, 45 },
		{ 358, 240, 45 },
		{ 358, 241, 45 },
		{ 358, 242, 45 },
		{ 358, 243, 45 },
		{ 358, 244, 45 },
		{ 358, 245, 45 },
		{ 358, 246, 45 },
		{ 358, 247, 45 },
		{ 358, 248, 45 },
		{ 358, 249, 45 },
		{ 358, 250, 45 },
		{ 358, 251, 45 },
		{ 358, 252, 45 },
		{ 358, 253, 45 },
		{ 358, 254, 45 },
		{ 358, 255, 45 },
		{ 358, 256, 45 },
		{ 359, 232, 45 },
		{ 359, 233, 45 },
		{ 359, 234, 45 },
		{ 359, 235, 45 },
		{ 359, 236, 45 },
		{ 359, 237, 45 },
		{ 359, 238, 45 },
		{ 359, 239, 45 },
		{ 359, 240, 45 },
		{ 359, 241, 45 },
		{ 359, 242, 45 },
		{ 359, 243, 45 },
		{ 359, 244, 45 },
		{ 359, 245, 45 },
		{ 359, 246, 45 },
		{ 359, 247, 45 },
		{ 359, 248, 45 },
		{ 359, 249, 45 },
		{ 359, 250, 45 },
		{ 359, 251, 45 },
		{ 359, 252, 45 },
		{ 359, 253, 45 },
		{ 359, 254, 45 },
		{ 359, 255, 45 },
		{ 359, 256, 45 },
		{ 360, 232, 45 },
		{ 360, 233, 45 },
		{ 360, 234, 45 },
		{ 360, 235, 45 },
		{ 360, 236, 45 },
		{ 360, 237, 45 },
		{ 360, 238, 45 },
		{ 360, 239, 45 },
		{ 360, 240, 45 },
		{ 360, 241, 45 },
		{ 360, 242, 45 },
		{ 360, 243, 45 },
		{ 360, 244, 45 },
		{ 360, 245, 45 },
		{ 360, 246, 45 },
		{ 360, 247, 45 },
		{ 360, 248, 45 },
		{ 360, 249, 45 },
		{ 360, 250, 45 },
		{ 360, 251, 45 },
		{ 360, 252, 45 },
		{ 360, 253, 45 },
		{ 360, 254, 45 },
		{ 360, 255, 45 },
		{ 360, 256, 45 },
		{ 361, 232, 45 },
		{ 361, 233, 45 },
		{ 361, 234, 45 },
		{ 361, 235, 45 },
		{ 361, 236, 45 },
		{ 361, 237, 45 },
		{ 361, 238, 45 },
		{ 361, 239, 45 },
		{ 361, 240, 45 },
		{ 361, 241, 45 },
		{ 361, 242, 45 },
		{ 361, 243, 45 },
		{ 361, 244, 45 },
		{ 361, 245, 45 },
		{ 361, 246, 45 },
		{ 361, 247, 45 },
		{ 361, 248, 45 },
		{ 361, 249, 45 },
		{ 361, 250, 45 },
		{ 361, 251, 45 },
		{ 361, 252, 45 },
		{ 361, 253, 45 },
		{ 361, 254, 45 },
		{ 361, 255, 45 },
		{ 361, 256, 45 },
		{ 362, 231, 45 },
		{ 362, 232, 45 },
		{ 362, 233, 45 },
		{ 362, 234, 45 },
		{ 362, 235, 45 },
		{ 362, 236, 45 },
		{ 362, 237, 45 },
		{ 362, 238, 45 },
		{ 362, 239, 45 },
		{ 362, 240, 45 },
		{ 362, 241, 45 },
		{ 362, 242, 45 },
		{ 362, 243, 45 },
		{ 362, 244, 45 },
		{ 362, 245, 45 },
		{ 362, 246, 45 },
		{ 362, 247, 45 },
		{ 362, 248, 45 },
		{ 362, 249, 45 },
		{ 362, 250, 45 },
		{ 362, 251, 45 },
		{ 362, 252, 45 },
		{ 362, 253, 45 },
		{ 362, 254, 45 },
		{ 362, 255, 45 },
		{ 362, 256, 45 },
		{ 363, 231, 45 },
		{ 363, 232, 45 },
		{ 363, 233, 45 },
		{ 363, 234, 45 },
		{ 363, 235, 45 },
		{ 363, 236, 45 },
		{ 363, 237, 45 },
		{ 363, 238, 45 },
		{ 363, 239, 45 },
		{ 363, 240, 45 },
		{ 363, 241, 45 },
		{ 363, 242, 45 },
		{ 363, 243, 45 },
		{ 363, 244, 45 },
		{ 363, 245, 45 },
		{ 363, 246, 45 },
		{ 363, 247, 45 },
		{ 363, 248, 45 },
		{ 363, 249, 45 },
		{ 363, 250, 45 },
		{ 363, 251, 45 },
		{ 363, 252, 45 },
		{ 363, 253, 45 },
		{ 363, 254, 45 },
		{ 363, 255, 45 },
		{ 363, 256, 45 },
		{ 364, 231, 45 },
		{ 364, 232, 45 },
		{ 364, 233, 45 },
		{ 364, 234, 45 },
		{ 364, 235, 45 },
		{ 364, 236, 45 },
		{ 364, 237, 45 },
		{ 364, 238, 45 },
		{ 364, 239, 45 },
		{ 364, 240, 45 },
		{ 364, 241, 45 },
		{ 364, 242, 45 },
		{ 364, 243, 45 },
		{ 364, 244, 45 },
		{ 364, 245, 45 },
		{ 364, 246, 45 },
		{ 364, 247, 45 },
		{ 364, 248, 45 },
		{ 364, 249, 45 },
		{ 364, 250, 45 },
		{ 364, 251, 45 },
		{ 364, 252, 45 },
		{ 364, 253, 45 },
		{ 364, 254, 45 },
		{ 364, 255, 45 },
		{ 364, 256, 45 },
		{ 365, 231, 45 },
		{ 365, 232, 45 },
		{ 365, 233, 45 },
		{ 365, 234, 45 },
		{ 365, 235, 45 },
		{ 365, 236, 45 },
		{ 365, 237, 45 },
		{ 365, 238, 45 },
		{ 365, 239, 45 },
		{ 365, 240, 45 },
		{ 365, 241, 45 },
		{ 365, 242, 45 },
		{ 365, 243, 45 },
		{ 365, 244, 45 },
		{ 365, 245, 45 },
		{ 365, 246, 45 },
		{ 365, 247, 45 },
		{ 365, 248, 45 },
		{ 365, 249, 45 },
		{ 365, 250, 45 },
		{ 365, 251, 45 },
		{ 365, 252, 45 },
		{ 365, 253, 45 },
		{ 365, 254, 45 },
		{ 365, 255, 45 },
		{ 365, 256, 45 },
		{ 366, 230, 45 },
		{ 366, 231, 45 },
		{ 366, 232, 45 },
		{ 366, 233, 45 },
		{ 366, 234, 45 },
		{ 366, 235, 45 },
		{ 366, 236, 45 },
		{ 366, 237, 45 },
		{ 366, 238, 45 },
		{ 366, 239, 45 },
		{ 366, 240, 45 },
		{ 366, 241, 45 },
		{ 366, 242, 45 },
		{ 366, 243, 45 },
		{ 366, 244, 45 },
		{ 366, 245, 45 },
		{ 366, 246, 45 },
		{ 366, 247, 45 },
		{ 366, 248, 45 },
		{ 366, 249, 45 },
		{ 366, 250, 45 },
		{ 366, 251, 45 },
		{ 366, 252, 45 },
		{ 366, 253, 45 },
		{ 366, 254, 45 },
		{ 366, 255, 45 },
		{ 366, 256, 45 },
		{ 367, 230, 45 },
		{ 367, 231, 45 },
		{ 367, 232, 45 },
		{ 367, 233, 45 },
		{ 367, 234, 45 },
		{ 367, 235, 45 },
		{ 367, 236, 45 },
		{ 367, 237, 45 },
		{ 367, 238, 45 },
		{ 367, 239, 45 },
		{ 367, 240, 45 },
		{ 367, 241, 45 },
		{ 367, 242, 45 },
		{ 367, 243, 45 },
		{ 367, 244, 45 },
		{ 367, 245, 45 },
		{ 367, 246, 45 },
		{ 367, 247, 45 },
		{ 367, 248, 45 },
		{ 367, 249, 45 },
		{ 367, 250, 45 },
		{ 367, 251, 45 },
		{ 367, 252, 45 },
		{ 367, 253, 45 },
		{ 367, 254, 45 },
		{ 367, 255, 45 },
		{ 367, 256, 45 },
		{ 368, 230, 45 },
		{ 368, 231, 45 },
		{ 368, 232, 45 },
		{ 368, 233, 45 },
		{ 368, 234, 45 },
		{ 368, 235, 45 },
		{ 368, 236, 45 },
		{ 368, 237, 45 },
		{ 368, 238, 45 },
		{ 368, 239, 45 },
		{ 368, 240, 45 },
		{ 368, 241, 45 },
		{ 368, 242, 45 },
		{ 368, 243, 45 },
		{ 368, 244, 45 },
		{ 368, 245, 45 },
		{ 368, 246, 45 },
		{ 368, 247, 45 },
		{ 368, 248, 45 },
		{ 368, 249, 45 },
		{ 368, 250, 45 },
		{ 368, 251, 45 },
		{ 368, 252, 45 },
		{ 368, 253, 45 },
		{ 368, 254, 45 },
		{ 368, 255, 45 },
		{ 368, 256, 45 },
		{ 369, 230, 45 },
		{ 369, 231, 45 },
		{ 369, 232, 45 },
		{ 369, 233, 45 },
		{ 369, 234, 45 },
		{ 369, 235, 45 },
		{ 369, 236, 45 },
		{ 369, 237, 45 },
		{ 369, 238, 45 },
		{ 369, 239, 45 },
		{ 369, 240, 45 },
		{ 369, 241, 45 },
		{ 369, 242, 45 },
		{ 369, 243, 45 },
		{ 369, 244, 45 },
		{ 369, 245, 45 },
		{ 369, 246, 45 },
		{ 369, 247, 45 },
		{ 369, 248, 45 },
		{ 369, 249, 45 },
		{ 369, 250, 45 },
		{ 369, 251, 45 },
		{ 369, 252, 45 },
		{ 369, 253, 45 },
		{ 369, 254, 45 },
		{ 369, 255, 45 },
		{ 369, 256, 45 },
		{ 370, 229, 45 },
		{ 370, 230, 45 },
		{ 370, 231, 45 },
		{ 370, 232, 45 },
		{ 370, 233, 45 },
		{ 370, 234, 45 },
		{ 370, 235, 45 },
		{ 370, 236, 45 },
		{ 370, 237, 45 },
		{ 370, 238, 45 },
		{ 370, 239, 45 },
		{ 370, 240, 45 },
		{ 370, 241, 45 },
		{ 370, 242, 45 },
		{ 370, 243, 45 },
		{ 370, 244, 45 },
		{ 370, 245, 45 },
		{ 370, 246, 45 },
		{ 370, 247, 45 },
		{ 370, 248, 45 },
		{ 370, 249, 45 },
		{ 370, 250, 45 },
		{ 370, 251, 45 },
		{ 370, 252, 45 },
		{ 370, 253, 45 },
		{ 370, 254, 45 },
		{ 370, 255, 45 },
		{ 370, 256, 45 },
		{ 371, 229, 45 },
		{ 371, 230, 45 },
		{ 371, 231, 45 },
		{ 371, 232, 45 },
		{ 371, 233, 45 },
		{ 371, 234, 45 },
		{ 371, 235, 45 },
		{ 371, 236, 45 },
		{ 371, 237, 45 },
		{ 371, 238, 45 },
		{ 371, 239, 45 },
		{ 371, 240, 45 },
		{ 371, 241, 45 },
		{ 371, 242, 45 },
		{ 371, 243, 45 },
		{ 371, 244, 45 },
		{ 371, 245, 45 },
		{ 371, 246, 45 },
		{ 371, 247, 45 },
		{ 371, 248, 45 },
		{ 371, 249, 45 },
		{ 371, 250, 45 },
		{ 371, 251, 45 },
		{ 371, 252, 45 },
		{ 371, 253, 45 },
		{ 371, 254, 45 },
		{ 371, 255, 45 },
		{ 371, 256, 45 },
		{ 372, 229, 45 },
		{ 372, 230, 45 },
		{ 372, 231, 45 },
		{ 372, 232, 45 },
		{ 372, 233, 45 },
		{ 372, 234, 45 },
		{ 372, 235, 45 },
		{ 372, 236, 45 },
		{ 372, 237, 45 },
		{ 372, 238, 45 },
		{ 372, 239, 45 },
		{ 372, 240, 45 },
		{ 372, 241, 45 },
		{ 372, 242, 45 },
		{ 372, 243, 45 },
		{ 372, 244, 45 },
		{ 372, 245, 45 },
		{ 372, 246, 45 },
		{ 372, 247, 45 },
		{ 372, 248, 45 },
		{ 372, 249, 45 },
		{ 372, 250, 45 },
		{ 372, 251, 45 },
		{ 372, 252, 45 },
		{ 372, 253, 45 },
		{ 372, 254, 45 },
		{ 372, 255, 45 },
		{ 372, 256, 45 },
		{ 373, 229, 45 },
		{ 373, 230, 45 },
		{ 373, 231, 45 },
		{ 373, 232, 45 },
		{ 373, 233, 45 },
		{ 373, 234, 45 },
		{ 373, 235, 45 },
		{ 373, 236, 45 },
		{ 373, 237, 45 },
		{ 373, 238, 45 },
		{ 373, 239, 45 },
		{ 373, 240, 45 },
		{ 373, 241, 45 },
		{ 373, 242, 45 },
		{ 373, 243, 45 },
		{ 373, 244, 45 },
		{ 373, 245, 45 },
		{ 373, 246, 45 },
		{ 373, 247, 45 },
		{ 373, 248, 45 },
		{ 373, 249, 45 },
		{ 373, 250, 45 },
		{ 373, 251, 45 },
		{ 373, 252, 45 },
		{ 373, 253, 45 },
		{ 373, 254, 45 },
		{ 373, 255, 45 },
		{ 373, 256, 45 },
		{ 374, 228, 45 },
		{ 374, 229, 45 },
		{ 374, 230, 45 },
		{ 374, 231, 45 },
		{ 374, 232, 45 },
		{ 374, 233, 45 },
		{ 374, 234, 45 },
		{ 374, 235, 45 },
		{ 374, 236, 45 },
		{ 374, 237, 45 },
		{ 374, 238, 45 },
		{ 374, 239, 45 },
		{ 374, 240, 45 },
		{ 374, 241, 45 },
		{ 374, 242, 45 },
		{ 374, 243, 45 },
		{ 374, 244, 45 },
		{ 374, 245, 45 },
		{ 374, 246, 45 },
		{ 374, 247, 45 },
		{ 374, 248, 45 },
		{ 374, 249, 45 },
		{ 374, 250, 45 },
		{ 374, 251, 45 },
		{ 374, 252, 45 },
		{ 374, 253, 45 },
		{ 374, 254, 45 },
		{ 374, 255, 45 },
		{ 374, 256, 45 },
		{ 375, 228, 45 },
		{ 375, 229, 45 },
		{ 375, 230, 45 },
		{ 375, 231, 45 },
		{ 375, 232, 45 },
		{ 375, 233, 45 },
		{ 375, 234, 45 },
		{ 375, 235, 45 },
		{ 375, 236, 45 },
		{ 375, 237, 45 },
		{ 375, 238, 45 },
		{ 375, 239, 45 },
		{ 375, 240, 45 },
		{ 375, 241, 45 },
		{ 375, 242, 45 },
		{ 375, 243, 45 },
		{ 375, 244, 45 },
		{ 375, 245, 45 },
		{ 375, 246, 45 },
		{ 375, 247, 45 },
		{ 375, 248, 45 },
		{ 375, 249, 45 },
		{ 375, 250, 45 },
		{ 375, 251, 45 },
		{ 375, 252, 45 },
		{ 375, 253, 45 },
		{ 375, 254, 45 },
		{ 375, 255, 45 },
		{ 375, 256, 45 },
		{ 376, 228, 45 },
		{ 376, 229, 45 },
		{ 376, 230, 45 },
		{ 376, 231, 45 },
		{ 376, 232, 45 },
		{ 376, 233, 45 },
		{ 376, 234, 45 },
		{ 376, 235, 45 },
		{ 376, 236, 45 },
		{ 376, 237, 45 },
		{ 376, 238, 45 },
		{ 376, 239, 45 },
		{ 376, 240, 45 },
		{ 376, 241, 45 },
		{ 376, 242, 45 },
		{ 376, 243, 45 },
		{ 376, 244, 45 },
		{ 376, 245, 45 },
		{ 376, 246, 45 },
		{ 376, 247, 45 },
		{ 376, 248, 45 },
		{ 376, 249, 45 },
		{ 376, 250, 45 },
		{ 376, 251, 45 },
		{ 376, 252, 45 },
		{ 376, 253, 45 },
		{ 376, 254, 45 },
		{ 376, 255, 45 },
		{ 376, 256, 45 },
		{ 377, 228, 45 },
		{ 377, 229, 45 },
		{ 377, 230, 45 },
		{ 377, 231, 45 },
		{ 377, 232, 45 },
		{ 377, 233, 45 },
		{ 377, 234, 45 },
		{ 377, 235, 45 },
		{ 377, 236, 45 },
		{ 377, 237, 45 },
		{ 377, 238, 45 },
		{ 377, 239, 45 },
		{ 377, 240, 45 },
		{ 377, 241, 45 },
		{ 377, 242, 45 },
		{ 377, 243, 45 },
		{ 377, 244, 45 },
		{ 377, 245, 45 },
		{ 377, 246, 45 },
		{ 377, 247, 45 },
		{ 377, 248, 45 },
		{ 377, 249, 45 },
		{ 377, 250, 45 },
		{ 377, 251, 45 },
		{ 377, 252, 45 },
		{ 377, 253, 45 },
		{ 377, 254, 45 },
		{ 377, 255, 45 },
		{ 377, 256, 45 },
		{ 378, 227, 45 },
		{ 378, 228, 45 },
		{ 378, 229, 45 },
		{ 378, 230, 45 },
		{ 378, 231, 45 },
		{ 378, 232, 45 },
		{ 378, 233, 45 },
		{ 378, 234, 45 },
		{ 378, 235, 45 },
		{ 378, 236, 45 },
		{ 378, 237, 45 },
		{ 378, 238, 45 },
		{ 378, 239, 45 },
		{ 378, 240, 45 },
		{ 378, 241, 45 },
		{ 378, 242, 45 },
		{ 378, 243, 45 },
		{ 378, 244, 45 },
		{ 378, 245, 45 },
		{ 378, 246, 45 },
		{ 378, 247, 45 },
		{ 378, 248, 45 },
		{ 378, 249, 45 },
		{ 378, 250, 45 },
		{ 378, 251, 45 },
		{ 378, 252, 45 },
		{ 378, 253, 45 },
		{ 378, 254, 45 },
		{ 378, 255, 45 },
		{ 378, 256, 45 },
		{ 379, 227, 45 },
		{ 379, 228, 45 },
		{ 379, 229, 45 },
		{ 379, 230, 45 },
		{ 379, 231, 45 },
		{ 379, 232, 45 },
		{ 379, 233, 45 },
		{ 379, 234, 45 },
		{ 379, 235, 45 },
		{ 379, 236, 45 },
		{ 379, 237, 45 },
		{ 379, 238, 45 },
		{ 379, 239, 45 },
		{ 379, 240, 45 },
		{ 379, 241, 45 },
		{ 379, 242, 45 },
		{ 379, 243, 45 },
		{ 379, 244, 45 },
		{ 379, 245, 45 },
		{ 379, 246, 45 },
		{ 379, 247, 45 },
		{ 379, 248, 45 },
		{ 379, 249, 45 },
		{ 379, 250, 45 },
		{ 379, 251, 45 },
		{ 379, 252, 45 },
		{ 379, 253, 45 },
		{ 379, 254, 45 },
		{ 379, 255, 45 },
		{ 379, 256, 45 },
		{ 380, 227, 45 },
		{ 380, 228, 45 },
		{ 380, 229, 45 },
		{ 380, 230, 45 },
		{ 380, 231, 45 },
		{ 380, 232, 45 },
		{ 380, 233, 45 },
		{ 380, 234, 45 },
		{ 380, 235, 45 },
		{ 380, 236, 45 },
		{ 380, 237, 45 },
		{ 380, 238, 45 },
		{ 380, 239, 45 },
		{ 380, 240, 45 },
		{ 380, 241, 45 },
		{ 380, 242, 45 },
		{ 380, 243, 45 },
		{ 380, 244, 45 },
		{ 380, 245, 45 },
		{ 380, 246, 45 },
		{ 380, 247, 45 },
		{ 380, 248, 45 },
		{ 380, 249, 45 },
		{ 380, 250, 45 },
		{ 380, 251, 45 },
		{ 380, 252, 45 },
		{ 380, 253, 45 },
		{ 380, 254, 45 },
		{ 380, 255, 45 },
		{ 380, 256, 45 },
		{ 381, 227, 45 },
		{ 381, 228, 45 },
		{ 381, 229, 45 },
		{ 381, 230, 45 },
		{ 381, 231, 45 },
		{ 381, 232, 45 },
		{ 381, 233, 45 },
		{ 381, 234, 45 },
		{ 381, 235, 45 },
		{ 381, 236, 45 },
		{ 381, 237, 45 },
		{ 381, 238, 45 },
		{ 381, 239, 45 },
		{ 381, 240, 45 },
		{ 381, 241, 45 },
		{ 381, 242, 45 },
		{ 381, 243, 45 },
		{ 381, 244, 45 },
		{ 381, 245, 45 },
		{ 381, 246, 45 },
		{ 381, 247, 45 },
		{ 381, 248, 45 },
		{ 381, 249, 45 },
		{ 381, 250, 45 },
		{ 381, 251, 45 },
		{ 381, 252, 45 },
		{ 381, 253, 45 },
		{ 381, 254, 45 },
		{ 381, 255, 45 },
		{ 381, 256, 45 },
		{ 382, 226, 45 },
		{ 382, 227, 45 },
		{ 382, 228, 45 },
		{ 382, 229, 45 },
		{ 382, 230, 45 },
		{ 382, 231, 45 },
		{ 382, 232, 45 },
		{ 382, 233, 45 },
		{ 382, 234, 45 },
		{ 382, 235, 45 },
		{ 382, 236, 45 },
		{ 382, 237, 45 },
		{ 382, 238, 45 },
		{ 382, 239, 45 },
		{ 382, 240, 45 },
		{ 382, 241, 45 },
		{ 382, 242, 45 },
		{ 382, 243, 45 },
		{ 382, 244, 45 },
		{ 382, 245, 45 },
		{ 382, 246, 45 },
		{ 382, 247, 45 },
		{ 382, 248, 45 },
		{ 382, 249, 45 },
		{ 382, 250, 45 },
		{ 382, 251, 45 },
		{ 382, 252, 45 },
		{ 382, 253, 45 },
		{ 382, 254, 45 },
		{ 382, 255, 45 },
		{ 382, 256, 45 },
		{ 383, 226, 45 },
		{ 383, 227, 45 },
		{ 383, 228, 45 },
		{ 383, 229, 45 },
		{ 383, 230, 45 },
		{ 383, 231, 45 },
		{ 383, 232, 45 },
		{ 383, 233, 45 },
		{ 383, 234, 45 },
		{ 383, 235, 45 },
		{ 383, 236, 45 },
		{ 383, 237, 45 },
		{ 383, 238, 45 },
		{ 383, 239, 45 },
		{ 383, 240, 45 },
		{ 383, 241, 45 },
		{ 383, 242, 45 },
		{ 383, 243, 45 },
		{ 383, 244, 45 },
		{ 383, 245, 45 },
		{ 383, 246, 45 },
		{ 383, 247, 45 },
		{ 383, 248, 45 },
		{ 383, 249, 45 },
		{ 383, 250, 45 },
		{ 383, 251, 45 },
		{ 383, 252, 45 },
		{ 383, 253, 45 },
		{ 383, 254, 45 },
		{ 383, 255, 45 },
		{ 383, 256, 45 },
		{ 384, 226, 45 },
		{ 384, 227, 45 },
		{ 384, 228, 45 },
		{ 384, 229, 45 },
		{ 384, 230, 45 },
		{ 384, 231, 45 },
		{ 384, 232, 45 },
		{ 384, 233, 45 },
		{ 384, 234, 45 },
		{ 384, 235, 45 },
		{ 384, 236, 45 },
		{ 384, 237, 45 },
		{ 384, 238, 45 },
		{ 384, 239, 45 },
		{ 384, 240, 45 },
		{ 384, 241, 45 },
		{ 384, 242, 45 },
		{ 384, 243, 45 },
		{ 384, 244, 45 },
		{ 384, 245, 45 },
		{ 384, 246, 45 },
		{ 384, 247, 45 },
		{ 384, 248, 45 },
		{ 384, 249, 45 },
		{ 384, 250, 45 },
		{ 384, 251, 45 },
		{ 384, 252, 45 },
		{ 384, 253, 45 },
		{ 384, 254, 45 },
		{ 384, 255, 45 },
		{ 384, 256, 45 },
		{ 385, 226, 45 },
		{ 385, 227, 45 },
		{ 385, 228, 45 },
		{ 385, 229, 45 },
		{ 385, 230, 45 },
		{ 385, 231, 45 },
		{ 385, 232, 45 },
		{ 385, 233, 45 },
		{ 385, 234, 45 },
		{ 385, 235, 45 },
		{ 385, 236, 45 },
		{ 385, 237, 45 },
		{ 385, 238, 45 },
		{ 385, 239, 45 },
		{ 385, 240, 45 },
		{ 385, 241, 45 },
		{ 385, 242, 45 },
		{ 385, 243, 45 },
		{ 385, 244, 45 },
		{ 385, 245, 45 },
		{ 385, 246, 45 },
		{ 385, 247, 45 },
		{ 385, 248, 45 },
		{ 385, 249, 45 },
		{ 385, 250, 45 },
		{ 385, 251, 45 },
		{ 385, 252, 45 },
		{ 385, 253, 45 },
		{ 385, 254, 45 },
		{ 385, 255, 45 },
		{ 385, 256, 45 },
		{ 386, 225, 45 },
		{ 386, 226, 45 },
		{ 386, 227, 45 },
		{ 386, 228, 45 },
		{ 386, 229, 45 },
		{ 386, 230, 45 },
		{ 386, 231, 45 },
		{ 386, 232, 45 },
		{ 386, 233, 45 },
		{ 386, 234, 45 },
		{ 386, 235, 45 },
		{ 386, 236, 45 },
		{ 386, 237, 45 },
		{ 386, 238, 45 },
		{ 386, 239, 45 },
		{ 386, 240, 45 },
		{ 386, 241, 45 },
		{ 386, 242, 45 },
		{ 386, 243, 45 },
		{ 386, 244, 45 },
		{ 386, 245, 45 },
		{ 386, 246, 45 },
		{ 386, 247, 45 },
		{ 386, 248, 45 },
		{ 386, 249, 45 },
		{ 386, 250, 45 },
		{ 386, 251, 45 },
		{ 386, 252, 45 },
		{ 386, 253, 45 },
		{ 386, 254, 45 },
		{ 386, 255, 45 },
		{ 386, 256, 45 },
		{ 387, 225, 45 },
		{ 387, 226, 45 },
		{ 387, 227, 45 },
		{ 387, 228, 45 },
		{ 387, 229, 45 },
		{ 387, 230, 45 },
		{ 387, 231, 45 },
		{ 387, 232, 45 },
		{ 387, 233, 45 },
		{ 387, 234, 45 },
		{ 387, 235, 45 },
		{ 387, 236, 45 },
		{ 387, 237, 45 },
		{ 387, 238, 45 },
		{ 387, 239, 45 },
		{ 387, 240, 45 },
		{ 387, 241, 45 },
		{ 387, 242, 45 },
		{ 387, 243, 45 },
		{ 387, 244, 45 },
		{ 387, 245, 45 },
		{ 387, 246, 45 },
		{ 387, 247, 45 },
		{ 387, 248, 45 },
		{ 387, 249, 45 },
		{ 387, 250, 45 },
		{ 387, 251, 45 },
		{ 387, 252, 45 },
		{ 387, 253, 45 },
		{ 387, 254, 45 },
		{ 387, 255, 45 },
		{ 387, 256, 45 },
		{ 388, 225, 45 },
		{ 388, 226, 45 },
		{ 388, 227, 45 },
		{ 388, 228, 45 },
		{ 388, 229, 45 },
		{ 388, 230, 45 },
		{ 388, 231, 45 },
		{ 388, 232, 45 },
		{ 388, 233, 45 },
		{ 388, 234, 45 },
		{ 388, 235, 45 },
		{ 388, 236, 45 },
		{ 388, 237, 45 },
		{ 388, 238, 45 },
		{ 388, 239, 45 },
		{ 388, 240, 45 },
		{ 388, 241, 45 },
		{ 388, 242, 45 },
		{ 388, 243, 45 },
		{ 388, 244, 45 },
		{ 388, 245, 45 },
		{ 388, 246, 45 },
		{ 388, 247, 45 },
		{ 388, 248, 45 },
		{ 388, 249, 45 },
		{ 388, 250, 45 },
		{ 388, 251, 45 },
		{ 388, 252, 45 },
		{ 388, 253, 45 },
		{ 388, 254, 45 },
		{ 388, 255, 45 },
		{ 388, 256, 45 },
		{ 389, 225, 45 },
		{ 389, 226, 45 },
		{ 389, 227, 45 },
		{ 389, 228, 45 },
		{ 389, 229, 45 },
		{ 389, 230, 45 },
		{ 389, 231, 45 },
		{ 389, 232, 45 },
		{ 389, 233, 45 },
		{ 389, 234, 45 },
		{ 389, 235, 45 },
		{ 389, 236, 45 },
		{ 389, 237, 45 },
		{ 389, 238, 45 },
		{ 389, 239, 45 },
		{ 389, 240, 45 },
		{ 389, 241, 45 },
		{ 389, 242, 45 },
		{ 389, 243, 45 },
		{ 389, 244, 45 },
		{ 389, 245, 45 },
		{ 389, 246, 45 },
		{ 389, 247, 45 },
		{ 389, 248, 45 },
		{ 389, 249, 45 },
		{ 389, 250, 45 },
		{ 389, 251, 45 },
		{ 389, 252, 45 },
		{ 389, 253, 45 },
		{ 389, 254, 45 },
		{ 389, 255, 45 },
		{ 389, 256, 45 },
		{ 390, 224, 45 },
		{ 390, 225, 45 },
		{ 390, 226, 45 },
		{ 390, 227, 45 },
		{ 390, 228, 45 },
		{ 390, 229, 45 },
		{ 390, 230, 45 },
		{ 390, 231, 45 },
		{ 390, 232, 45 },
		{ 390, 233, 45 },
		{ 390, 234, 45 },
		{ 390, 235, 45 },
		{ 390, 236, 45 },
		{ 390, 237, 45 },
		{ 390, 238, 45 },
		{ 390, 239, 45 },
		{ 390, 240, 45 },
		{ 390, 241, 45 },
		{ 390, 242, 45 },
		{ 390, 243, 45 },
		{ 390, 244, 45 },
		{ 390, 245, 45 },
		{ 390, 246, 45 },
		{ 390, 247, 45 },
		{ 390, 248, 45 },
		{ 390, 249, 45 },
		{ 390, 250, 45 },
		{ 390, 251, 45 },
		{ 390, 252, 45 },
		{ 390, 253, 45 },
		{ 390, 254, 45 },
		{ 390, 255, 45 },
		{ 390, 256, 45 },
		{ 391, 224, 45 },
		{ 391, 225, 45 },
		{ 391, 226, 45 },
		{ 391, 227, 45 },
		{ 391, 228, 45 },
		{ 391, 229, 45 },
		{ 391, 230, 45 },
		{ 391, 231, 45 },
		{ 391, 232, 45 },
		{ 391, 233, 45 },
		{ 391, 234, 45 },
		{ 391, 235, 45 },
		{ 391, 236, 45 },
		{ 391, 237, 45 },
		{ 391, 238, 45 },
		{ 391, 239, 45 },
		{ 391, 240, 45 },
		{ 391, 241, 45 },
		{ 391, 242, 45 },
		{ 391, 243, 45 },
		{ 391, 244, 45 },
		{ 391, 245, 45 },
		{ 391, 246, 45 },
		{ 391, 247, 45 },
		{ 391, 248, 45 },
		{ 391, 249, 45 },
		{ 391, 250, 45 },
		{ 391, 251, 45 },
		{ 391, 252, 45 },
		{ 391, 253, 45 },
		{ 391, 254, 45 },
		{ 391, 255, 45 },
		{ 391, 256, 45 },
		{ 392, 224, 45 },
		{ 392, 225, 45 },
		{ 392, 226, 45 },
		{ 392, 227, 45 },
		{ 392, 228, 45 },
		{ 392, 229, 45 },
		{ 392, 230, 45 },
		{ 392, 231, 45 },
		{ 392, 232, 45 },
		{ 392, 233, 45 },
		{ 392, 234, 45 },
		{ 392, 235, 45 },
		{ 392, 236, 45 },
		{ 392, 237, 45 },
		{ 392, 238, 45 },
		{ 392, 239, 45 },
		{ 392, 240, 45 },
		{ 392, 241, 45 },
		{ 392, 242, 45 },
		{ 392, 243, 45 },
		{ 392, 244, 45 },
		{ 392, 245, 45 },
		{ 392, 246, 45 },
		{ 392, 247, 45 },
		{ 392, 248, 45 },
		{ 392, 249, 45 },
		{ 392, 250, 45 },
		{ 392, 251, 45 },
		{ 392, 252, 45 },
		{ 392, 253, 45 },
		{ 392, 254, 45 },
		{ 392, 255, 45 },
		{ 392, 256, 45 },
		{ 393, 224, 45 },
		{ 393, 225, 45 },
		{ 393, 226, 45 },
		{ 393, 227, 45 },
		{ 393, 228, 45 },
		{ 393, 229, 45 },
		{ 393, 230, 45 },
		{ 393, 231, 45 },
		{ 393, 232, 45 },
		{ 393, 233, 45 },
		{ 393, 234, 45 },
		{ 393, 235, 45 },
		{ 393, 236, 45 },
		{ 393, 237, 45 },
		{ 393, 238, 45 },
		{ 393, 239, 45 },
		{ 393, 240, 45 },
		{ 393, 241, 45 },
		{ 393, 242, 45 },
		{ 393, 243, 45 },
		{ 393, 244, 45 },
		{ 393, 245, 45 },
		{ 393, 246, 45 },
		{ 393, 247, 45 },
		{ 393, 248, 45 },
		{ 393, 249, 45 },
		{ 393, 250, 45 },
		{ 393, 251, 45 },
		{ 393, 252, 45 },
		{ 393, 253, 45 },
		{ 393, 254, 45 },
		{ 393, 255, 45 },
		{ 393, 256, 45 },
		{ 394, 223, 45 },
		{ 394, 224, 45 },
		{ 394, 225, 45 },
		{ 394, 226, 45 },
		{ 394, 227, 45 },
		{ 394, 228, 45 },
		{ 394, 229, 45 },
		{ 394, 230, 45 },
		{ 394, 231, 45 },
		{ 394, 232, 45 },
		{ 394, 233, 45 },
		{ 394, 234, 45 },
		{ 394, 235, 45 },
		{ 394, 236, 45 },
		{ 394, 237, 45 },
		{ 394, 238, 45 },
		{ 394, 239, 45 },
		{ 394, 240, 45 },
		{ 394, 241, 45 },
		{ 394, 242, 45 },
		{ 394, 243, 45 },
		{ 394, 244, 45 },
		{ 394, 245, 45 },
		{ 394, 246, 45 },
		{ 394, 247, 45 },
		{ 394, 248, 45 },
		{ 394, 249, 45 },
		{ 394, 250, 45 },
		{ 394, 251, 45 },
		{ 394, 252, 45 },
		{ 394, 253, 45 },
		{ 394, 254, 45 },
		{ 394, 255, 45 },
		{ 394, 256, 45 },
		{ 395, 223, 45 },
		{ 395, 224, 45 },
		{ 395, 225, 45 },
		{ 395, 226, 45 },
		{ 395, 227, 45 },
		{ 395, 228, 45 },
		{ 395, 229, 45 },
		{ 395, 230, 45 },
		{ 395, 231, 45 },
		{ 395, 232, 45 },
		{ 395, 233, 45 },
		{ 395, 234, 45 },
		{ 395, 235, 45 },
		{ 395, 236, 45 },
		{ 395, 237, 45 },
		{ 395, 238, 45 },
		{ 395, 239, 45 },
		{ 395, 240, 45 },
		{ 395, 241, 45 },
		{ 395, 242, 45 },
		{ 395, 243, 45 },
		{ 395, 244, 45 },
		{ 395, 245, 45 },
		{ 395, 246, 45 },
		{ 395, 247, 45 },
		{ 395, 248, 45 },
		{ 395, 249, 45 },
		{ 395, 250, 45 },
		{ 395, 251, 45 },
		{ 395, 252, 45 },
		{ 395, 253, 45 },
		{ 395, 254, 45 },
		{ 395, 255, 45 },
		{ 395, 256, 45 },
		{ 396, 223, 45 },
		{ 396, 224, 45 },
		{ 396, 225, 45 },
		{ 396, 226, 45 },
		{ 396, 227, 45 },
		{ 396, 228, 45 },
		{ 396, 229, 45 },
		{ 396, 230, 45 },
		{ 396, 231, 45 },
		{ 396, 232, 45 },
		{ 396, 233, 45 },
		{ 396, 234, 45 },
		{ 396, 235, 45 },
		{ 396, 236, 45 },
		{ 396, 237, 45 },
		{ 396, 238, 45 },
		{ 396, 239, 45 },
		{ 396, 240, 45 },
		{ 396, 241, 45 },
		{ 396, 242, 45 },
		{ 396, 243, 45 },
		{ 396, 244, 45 },
		{ 396, 245, 45 },
		{ 396, 246, 45 },
		{ 396, 247, 45 },
		{ 396, 248, 45 },
		{ 396, 249, 45 },
		{ 396, 250, 45 },
		{ 396, 251, 45 },
		{ 396, 252, 45 },
		{ 396, 253, 45 },
		{ 396, 254, 45 },
		{ 396, 255, 45 },
		{ 396, 256, 45 },
		{ 397, 223, 45 },
		{ 397, 224, 45 },
		{ 397, 225, 45 },
		{ 397, 226, 45 },
		{ 397, 227, 45 },
		{ 397, 228, 45 },
		{ 397, 229, 45 },
		{ 397, 230, 45 },
		{ 397, 231, 45 },
		{ 397, 232, 45 },
		{ 397, 233, 45 },
		{ 397, 234, 45 },
		{ 397, 235, 45 },
		{ 397, 236, 45 },
		{ 397, 237, 45 },
		{ 397, 238, 45 },
		{ 397, 239, 45 },
		{ 397, 240, 45 },
		{ 397, 241, 45 },
		{ 397, 242, 45 },
		{ 397, 243, 45 },
		{ 397, 244, 45 },
		{ 397, 245, 45 },
		{ 397, 246, 45 },
		{ 397, 247, 45 },
		{ 397, 248, 45 },
		{ 397, 249, 45 },
		{ 397, 250, 45 },
		{ 397, 251, 45 },
		{ 397, 252, 45 },
		{ 397, 253, 45 },
		{ 397, 254, 45 },
		{ 397, 255, 45 },
		{ 397, 256, 45 },
		{ 398, 222, 45 },
		{ 398, 223, 45 },
		{ 398, 224, 45 },
		{ 398, 225, 45 },
		{ 398, 226, 45 },
		{ 398, 227, 45 },
		{ 398, 228, 45 },
		{ 398, 229, 45 },
		{ 398, 230, 45 },
		{ 398, 231, 45 },
		{ 398, 232, 45 },
		{ 398, 233, 45 },
		{ 398, 234, 45 },
		{ 398, 235, 45 },
		{ 398, 236, 45 },
		{ 398, 237, 45 },
		{ 398, 238, 45 },
		{ 398, 239, 45 },
		{ 398, 240, 45 },
		{ 398, 241, 45 },
		{ 398, 242, 45 },
		{ 398, 243, 45 },
		{ 398, 244, 45 },
		{ 398, 245, 45 },
		{ 398, 246, 45 },
		{ 398, 247, 45 },
		{ 398, 248, 45 },
		{ 398, 249, 45 },
		{ 398, 250, 45 },
		{ 398, 251, 45 },
		{ 398, 252, 45 },
		{ 398, 253, 45 },
		{ 398, 254, 45 },
		{ 398, 255, 45 },
		{ 398, 256, 45 },
		{ 399, 222, 45 },
		{ 399, 223, 45 },
		{ 399, 224, 45 },
		{ 399, 225, 45 },
		{ 399, 226, 45 },
		{ 399, 227, 45 },
		{ 399, 228, 45 },
		{ 399, 229, 45 },
		{ 399, 230, 45 },
		{ 399, 231, 45 },
		{ 399, 232, 45 },
		{ 399, 233, 45 },
		{ 399, 234, 45 },
		{ 399, 235, 45 },
		{ 399, 236, 45 },
		{ 399, 237, 45 },
		{ 399, 238, 45 },
		{ 399, 239, 45 },
		{ 399, 240, 45 },
		{ 399, 241, 45 },
		{ 399, 242, 45 },
		{ 399, 243, 45 },
		{ 399, 244, 45 },
		{ 399, 245, 45 },
		{ 399, 246, 45 },
		{ 399, 247, 45 },
		{ 399, 248, 45 },
		{ 399, 249, 45 },
		{ 399, 250, 45 },
		{ 399, 251, 45 },
		{ 399, 252, 45 },
		{ 399, 253, 45 },
		{ 399, 254, 45 },
		{ 399, 255, 45 },
		{ 399, 256, 45 },
		{ 400, 222, 45 },
		{ 400, 223, 45 },
		{ 400, 224, 45 },
		{ 400, 225, 45 },
		{ 400, 226, 45 },
		{ 400, 227, 45 },
		{ 400, 228, 45 },
		{ 400, 229, 45 },
		{ 400, 230, 45 },
		{ 400, 231, 45 },
		{ 400, 232, 45 },
		{ 400, 233, 45 },
		{ 400, 234, 45 },
		{ 400, 235, 45 },
		{ 400, 236, 45 },
		{ 400, 237, 45 },
		{ 400, 238, 45 },
		{ 400, 239, 45 },
		{ 400, 240, 45 },
		{ 400, 241, 45 },
		{ 400, 242, 45 },
		{ 400, 243, 45 },
		{ 400, 244, 45 },
		{ 400, 245, 45 },
		{ 400, 246, 45 },
		{ 400, 247, 45 },
		{ 400, 248, 45 },
		{ 400, 249, 45 },
		{ 400, 250, 45 },
		{ 400, 251, 45 },
		{ 400, 252, 45 },
		{ 400, 253, 45 },
		{ 400, 254, 45 },
		{ 400, 255, 45 },
		{ 400, 256, 45 },
		{ 401, 222, 45 },
		{ 401, 223, 45 },
		{ 401, 224, 45 },
		{ 401, 225, 45 },
		{ 401, 226, 45 },
		{ 401, 227, 45 },
		{ 401, 228, 45 },
		{ 401, 229, 45 },
		{ 401, 230, 45 },
		{ 401, 231, 45 },
		{ 401, 232, 45 },
		{ 401, 233, 45 },
		{ 401, 234, 45 },
		{ 401, 235, 45 },
		{ 401, 236, 45 },
		{ 401, 237, 45 },
		{ 401, 238, 45 },
		{ 401, 239, 45 },
		{ 401, 240, 45 },
		{ 401, 241, 45 },
		{ 401, 242, 45 },
		{ 401, 243, 45 },
		{ 401, 244, 45 },
		{ 401, 245, 45 },
		{ 401, 246, 45 },
		{ 401, 247, 45 },
		{ 401, 248, 45 },
		{ 401, 249, 45 },
		{ 401, 250, 45 },
		{ 401, 251, 45 },
		{ 401, 252, 45 },
		{ 401, 253, 45 },
		{ 401, 254, 45 },
		{ 401, 255, 45 },
		{ 401, 256, 45 },
		{ 402, 221, 45 },
		{ 402, 222, 45 },
		{ 402, 223, 45 },
		{ 402, 224, 45 },
		{ 402, 225, 45 },
		{ 402, 226, 45 },
		{ 402, 227, 45 },
		{ 402, 228, 45 },
		{ 402, 229, 45 },
		{ 402, 230, 45 },
		{ 402, 231, 45 },
		{ 402, 232, 45 },
		{ 402, 233, 45 },
		{ 402, 234, 45 },
		{ 402, 235, 45 },
		{ 402, 236, 45 },
		{ 402, 237, 45 },
		{ 402, 238, 45 },
		{ 402, 239, 45 },
		{ 402, 240, 45 },
		{ 402, 241, 45 },
		{ 402, 242, 45 },
		{ 402, 243, 45 },
		{ 402, 244, 45 },
		{ 402, 245, 45 },
		{ 402, 246, 45 },
		{ 402, 247, 45 },
		{ 402, 248, 45 },
		{ 402, 249, 45 },
		{ 402, 250, 45 },
		{ 402, 251, 45 },
		{ 402, 252, 45 },
		{ 402, 253, 45 },
		{ 402, 254, 45 },
		{ 402, 255, 45 },
		{ 402, 256, 45 },
		{ 403, 221, 45 },
		{ 403, 222, 45 },
		{ 403, 223, 45 },
		{ 403, 224, 45 },
		{ 403, 225, 45 },
		{ 403, 226, 45 },
		{ 403, 227, 45 },
		{ 403, 228, 45 },
		{ 403, 229, 45 },
		{ 403, 230, 45 },
		{ 403, 231, 45 },
		{ 403, 232, 45 },
		{ 403, 233, 45 },
		{ 403, 234, 45 },
		{ 403, 235, 45 },
		{ 403, 236, 45 },
		{ 403, 237, 45 },
		{ 403, 238, 45 },
		{ 403, 239, 45 },
		{ 403, 240, 45 },
		{ 403, 241, 45 },
		{ 403, 242, 45 },
		{ 403, 243, 45 },
		{ 403, 244, 45 },
		{ 403, 245, 45 },
		{ 403, 246, 45 },
		{ 403, 247, 45 },
		{ 403, 248, 45 },
		{ 403, 249, 45 },
		{ 403, 250, 45 },
		{ 403, 251, 45 },
		{ 403, 252, 45 },
		{ 403, 253, 45 },
		{ 403, 254, 45 },
		{ 403, 255, 45 },
		{ 403, 256, 45 },
		{ 404, 221, 45 },
		{ 404, 222, 45 },
		{ 404, 223, 45 },
		{ 404, 224, 45 },
		{ 404, 225, 45 },
		{ 404, 226, 45 },
		{ 404, 227, 45 },
		{ 404, 228, 45 },
		{ 404, 229, 45 },
		{ 404, 230, 45 },
		{ 404, 231, 45 },
		{ 404, 232, 45 },
		{ 404, 233, 45 },
		{ 404, 234, 45 },
		{ 404, 235, 45 },
		{ 404, 236, 45 },
		{ 404, 237, 45 },
		{ 404, 238, 45 },
		{ 404, 239, 45 },
		{ 404, 240, 45 },
		{ 404, 241, 45 },
		{ 404, 242, 45 },
		{ 404, 243, 45 },
		{ 404, 244, 45 },
		{ 404, 245, 45 },
		{ 404, 246, 45 },
		{ 404, 247, 45 },
		{ 404, 248, 45 },
		{ 404, 249, 45 },
		{ 404, 250, 45 },
		{ 404, 251, 45 },
		{ 404, 252, 45 },
		{ 404, 253, 45 },
		{ 404, 254, 45 },
		{ 404, 255, 45 },
		{ 404, 256, 45 },
		{ 405, 221, 45 },
		{ 405, 222, 45 },
		{ 405, 223, 45 },
		{ 405, 224, 45 },
		{ 405, 225, 45 },
		{ 405, 226, 45 },
		{ 405, 227, 45 },
		{ 405, 228, 45 },
		{ 405, 229, 45 },
		{ 405, 230, 45 },
		{ 405, 231, 45 },
		{ 405, 232, 45 },
		{ 405, 233, 45 },
		{ 405, 234, 45 },
		{ 405, 235, 45 },
		{ 405, 236, 45 },
		{ 405, 237, 45 },
		{ 405, 238, 45 },
		{ 405, 239, 45 },
		{ 405, 240, 45 },
		{ 405, 241, 45 },
		{ 405, 242, 45 },
		{ 405, 243, 45 },
		{ 405, 244, 45 },
		{ 405, 245, 45 },
		{ 405, 246, 45 },
		{ 405, 247, 45 },
		{ 405, 248, 45 },
		{ 405, 249, 45 },
		{ 405, 250, 45 },
		{ 405, 251, 45 },
		{ 405, 252, 45 },
		{ 405, 253, 45 },
		{ 405, 254, 45 },
		{ 405, 255, 45 },
		{ 405, 256, 45 },
		{ 406, 220, 45 },
		{ 406, 221, 45 },
		{ 406, 222, 45 },
		{ 406, 223, 45 },
		{ 406, 224, 45 },
		{ 406, 225, 45 },
		{ 406, 226, 45 },
		{ 406, 227, 45 },
		{ 406, 228, 45 },
		{ 406, 229, 45 },
		{ 406, 230, 45 },
		{ 406, 231, 45 },
		{ 406, 232, 45 },
		{ 406, 233, 45 },
		{ 406, 234, 45 },
		{ 406, 235, 45 },
		{ 406, 236, 45 },
		{ 406, 237, 45 },
		{ 406, 238, 45 },
		{ 406, 239, 45 },
		{ 406, 240, 45 },
		{ 406, 241, 45 },
		{ 406, 242, 45 },
		{ 406, 243, 45 },
		{ 406, 244, 45 },
		{ 406, 245, 45 },
		{ 406, 246, 45 },
		{ 406, 247, 45 },
		{ 406, 248, 45 },
		{ 406, 249, 45 },
		{ 406, 250, 45 },
		{ 406, 251, 45 },
		{ 406, 252, 45 },
		{ 406, 253, 45 },
		{ 406, 254, 45 },
		{ 406, 255, 45 },
		{ 406, 256, 45 },
		{ 407, 220, 45 },
		{ 407, 221, 45 },
		{ 407, 222, 45 },
		{ 407, 223, 45 },
		{ 407, 224, 45 },
		{ 407, 225, 45 },
		{ 407, 226, 45 },
		{ 407, 227, 45 },
		{ 407, 228, 45 },
		{ 407, 229, 45 },
		{ 407, 230, 45 },
		{ 407, 231, 45 },
		{ 407, 232, 45 },
		{ 407, 233, 45 },
		{ 407, 234, 45 },
		{ 407, 235, 45 },
		{ 407, 236, 45 },
		{ 407, 237, 45 },
		{ 407, 238, 45 },
		{ 407, 239, 45 },
		{ 407, 240, 45 },
		{ 407, 241, 45 },
		{ 407, 242, 45 },
		{ 407, 243, 45 },
		{ 407, 244, 45 },
		{ 407, 245, 45 },
		{ 407, 246, 45 },
		{ 407, 247, 45 },
		{ 407, 248, 45 },
		{ 407, 249, 45 },
		{ 407, 250, 45 },
		{ 407, 251, 45 },
		{ 407, 252, 45 },
		{ 407, 253, 45 },
		{ 407, 254, 45 },
		{ 407, 255, 45 },
		{ 407, 256, 45 },
		{ 408, 220, 45 },
		{ 408, 221, 45 },
		{ 408, 222, 45 },
		{ 408, 223, 45 },
		{ 408, 224, 45 },
		{ 408, 225, 45 },
		{ 408, 226, 45 },
		{ 408, 227, 45 },
		{ 408, 228, 45 },
		{ 408, 229, 45 },
		{ 408, 230, 45 },
		{ 408, 231, 45 },
		{ 408, 232, 45 },
		{ 408, 233, 45 },
		{ 408, 234, 45 },
		{ 408, 235, 45 },
		{ 408, 236, 45 },
		{ 408, 237, 45 },
		{ 408, 238, 45 },
		{ 408, 239, 45 },
		{ 408, 240, 45 },
		{ 408, 241, 45 },
		{ 408, 242, 45 },
		{ 408, 243, 45 },
		{ 408, 244, 45 },
		{ 408, 245, 45 },
		{ 408, 246, 45 },
		{ 408, 247, 45 },
		{ 408, 248, 45 },
		{ 408, 249, 45 },
		{ 408, 250, 45 },
		{ 408, 251, 45 },
		{ 408, 252, 45 },
		{ 408, 253, 45 },
		{ 408, 254, 45 },
		{ 408, 255, 45 },
		{ 408, 256, 45 },
		{ 409, 220, 45 },
		{ 409, 221, 45 },
		{ 409, 222, 45 },
		{ 409, 223, 45 },
		{ 409, 224, 45 },
		{ 409, 225, 45 },
		{ 409, 226, 45 },
		{ 409, 227, 45 },
		{ 409, 228, 45 },
		{ 409, 229, 45 },
		{ 409, 230, 45 },
		{ 409, 231, 45 },
		{ 409, 232, 45 },
		{ 409, 233, 45 },
		{ 409, 234, 45 },
		{ 409, 235, 45 },
		{ 409, 236, 45 },
		{ 409, 237, 45 },
		{ 409, 238, 45 },
		{ 409, 239, 45 },
		{ 409, 240, 45 },
		{ 409, 241, 45 },
		{ 409, 242, 45 },
		{ 409, 243, 45 },
		{ 409, 244, 45 },
		{ 409, 245, 45 },
		{ 409, 246, 45 },
		{ 409, 247, 45 },
		{ 409, 248, 45 },
		{ 409, 249, 45 },
		{ 409, 250, 45 },
		{ 409, 251, 45 },
		{ 409, 252, 45 },
		{ 409, 253, 45 },
		{ 409, 254, 45 },
		{ 409, 255, 45 },
		{ 409, 256, 45 },
		{ 410, 219, 45 },
		{ 410, 220, 45 },
		{ 410, 221, 45 },
		{ 410, 222, 45 },
		{ 410, 223, 45 },
		{ 410, 224, 45 },
		{ 410, 225, 45 },
		{ 410, 226, 45 },
		{ 410, 227, 45 },
		{ 410, 228, 45 },
		{ 410, 229, 45 },
		{ 410, 230, 45 },
		{ 410, 231, 45 },
		{ 410, 232, 45 },
		{ 410, 233, 45 },
		{ 410, 234, 45 },
		{ 410, 235, 45 },
		{ 410, 236, 45 },
		{ 410, 237, 45 },
		{ 410, 238, 45 },
		{ 410, 239, 45 },
		{ 410, 240, 45 },
		{ 410, 241, 45 },
		{ 410, 242, 45 },
		{ 410, 243, 45 },
		{ 410, 244, 45 },
		{ 410, 245, 45 },
		{ 410, 246, 45 },
		{ 410, 247, 45 },
		{ 410, 248, 45 },
		{ 410, 249, 45 },
		{ 410, 250, 45 },
		{ 410, 251, 45 },
		{ 410, 252, 45 },
		{ 410, 253, 45 },
		{ 410, 254, 45 },
		{ 410, 255, 45 },
		{ 410, 256, 45 },
		{ 411, 219, 45 },
		{ 411, 220, 45 },
		{ 411, 221, 45 },
		{ 411, 222, 45 },
		{ 411, 223, 45 },
		{ 411, 224, 45 },
		{ 411, 225, 45 },
		{ 411, 226, 45 },
		{ 411, 227, 45 },
		{ 411, 228, 45 },
		{ 411, 229, 45 },
		{ 411, 230, 45 },
		{ 411, 231, 45 },
		{ 411, 232, 45 },
		{ 411, 233, 45 },
		{ 411, 234, 45 },
		{ 411, 235, 45 },
		{ 411, 236, 45 },
		{ 411, 237, 45 },
		{ 411, 238, 45 },
		{ 411, 239, 45 },
		{ 411, 240, 45 },
		{ 411, 241, 45 },
		{ 411, 242, 45 },
		{ 411, 243, 45 },
		{ 411, 244, 45 },
		{ 411, 245, 45 },
		{ 411, 246, 45 },
		{ 411, 247, 45 },
		{ 411, 248, 45 },
		{ 411, 249, 45 },
		{ 411, 250, 45 },
		{ 411, 251, 45 },
		{ 411, 252, 45 },
		{ 411, 253, 45 },
		{ 411, 254, 45 },
		{ 411, 255, 45 },
		{ 411, 256, 45 },
		{ 412, 219, 45 },
		{ 412, 220, 45 },
		{ 412, 221, 45 },
		{ 412, 222, 45 },
		{ 412, 223, 45 },
		{ 412, 224, 45 },
		{ 412, 225, 45 },
		{ 412, 226, 45 },
		{ 412, 227, 45 },
		{ 412, 228, 45 },
		{ 412, 229, 45 },
		{ 412, 230, 45 },
		{ 412, 231, 45 },
		{ 412, 232, 45 },
		{ 412, 233, 45 },
		{ 412, 234, 45 },
		{ 412, 235, 45 },
		{ 412, 236, 45 },
		{ 412, 237, 45 },
		{ 412, 238, 45 },
		{ 412, 239, 45 },
		{ 412, 240, 45 },
		{ 412, 241, 45 },
		{ 412, 242, 45 },
		{ 412, 243, 45 },
		{ 412, 244, 45 },
		{ 412, 245, 45 },
		{ 412, 246, 45 },
		{ 412, 247, 45 },
		{ 412, 248, 45 },
		{ 412, 249, 45 },
		{ 412, 250, 45 },
		{ 412, 251, 45 },
		{ 412, 252, 45 },
		{ 412, 253, 45 },
		{ 412, 254, 45 },
		{ 412, 255, 45 },
		{ 412, 256, 45 },
		{ 413, 219, 45 },
		{ 413, 220, 45 },
		{ 413, 221, 45 },
		{ 413, 222, 45 },
		{ 413, 223, 45 },
		{ 413, 224, 45 },
		{ 413, 225, 45 },
		{ 413, 226, 45 },
		{ 413, 227, 45 },
		{ 413, 228, 45 },
		{ 413, 229, 45 },
		{ 413, 230, 45 },
		{ 413, 231, 45 },
		{ 413, 232, 45 },
		{ 413, 233, 45 },
		{ 413, 234, 45 },
		{ 413, 235, 45 },
		{ 413, 236, 45 },
		{ 413, 237, 45 },
		{ 413, 238, 45 },
		{ 413, 239, 45 },
		{ 413, 240, 45 },
		{ 413, 241, 45 },
		{ 413, 242, 45 },
		{ 413, 243, 45 },
		{ 413, 244, 45 },
		{ 413, 245, 45 },
		{ 413, 246, 45 },
		{ 413, 247, 45 },
		{ 413, 248, 45 },
		{ 413, 249, 45 },
		{ 413, 250, 45 },
		{ 413, 251, 45 },
		{ 413, 252, 45 },
		{ 413, 253, 45 },
		{ 413, 254, 45 },
		{ 413, 255, 45 },
		{ 413, 256, 45 },
		{ 414, 218, 45 },
		{ 414, 219, 45 },
		{ 414, 220, 45 },
		{ 414, 221, 45 },
		{ 414, 222, 45 },
		{ 414, 223, 45 },
		{ 414, 224, 45 },
		{ 414, 225, 45 },
		{ 414, 226, 45 },
		{ 414, 227, 45 },
		{ 414, 228, 45 },
		{ 414, 229, 45 },
		{ 414, 230, 45 },
		{ 414, 231, 45 },
		{ 414, 232, 45 },
		{ 414, 233, 45 },
		{ 414, 234, 45 },
		{ 414, 235, 45 },
		{ 414, 236, 45 },
		{ 414, 237, 45 },
		{ 414, 238, 45 },
		{ 414, 239, 45 },
		{ 414, 240, 45 },
		{ 414, 241, 45 },
		{ 414, 242, 45 },
		{ 414, 243, 45 },
		{ 414, 244, 45 },
		{ 414, 245, 45 },
		{ 414, 246, 45 },
		{ 414, 247, 45 },
		{ 414, 248, 45 },
		{ 414, 249, 45 },
		{ 414, 250, 45 },
		{ 414, 251, 45 },
		{ 414, 252, 45 },
		{ 414, 253, 45 },
		{ 414, 254, 45 },
		{ 414, 255, 45 },
		{ 414, 256, 45 },
		{ 415, 218, 45 },
		{ 415, 219, 45 },
		{ 415, 220, 45 },
		{ 415, 221, 45 },
		{ 415, 222, 45 },
		{ 415, 223, 45 },
		{ 415, 224, 45 },
		{ 415, 225, 45 },
		{ 415, 226, 45 },
		{ 415, 227, 45 },
		{ 415, 228, 45 },
		{ 415, 229, 45 },
		{ 415, 230, 45 },
		{ 415, 231, 45 },
		{ 415, 232, 45 },
		{ 415, 233, 45 },
		{ 415, 234, 45 },
		{ 415, 235, 45 },
		{ 415, 236, 45 },
		{ 415, 237, 45 },
		{ 415, 238, 45 },
		{ 415, 239, 45 },
		{ 415, 240, 45 },
		{ 415, 241, 45 },
		{ 415, 242, 45 },
		{ 415, 243, 45 },
		{ 415, 244, 45 },
		{ 415, 245, 45 },
		{ 415, 246, 45 },
		{ 415, 247, 45 },
		{ 415, 248, 45 },
		{ 415, 249, 45 },
		{ 415, 250, 45 },
		{ 415, 251, 45 },
		{ 415, 252, 45 },
		{ 415, 253, 45 },
		{ 415, 254, 45 },
		{ 415, 255, 45 },
		{ 415, 256, 45 },
		{ 416, 218, 45 },
		{ 416, 219, 45 },
		{ 416, 220, 45 },
		{ 416, 221, 45 },
		{ 416, 222, 45 },
		{ 416, 223, 45 },
		{ 416, 224, 45 },
		{ 416, 225, 45 },
		{ 416, 226, 45 },
		{ 416, 227, 45 },
		{ 416, 228, 45 },
		{ 416, 229, 45 },
		{ 416, 230, 45 },
		{ 416, 231, 45 },
		{ 416, 232, 45 },
		{ 416, 233, 45 },
		{ 416, 234, 45 },
		{ 416, 235, 45 },
		{ 416, 236, 45 },
		{ 416, 237, 45 },
		{ 416, 238, 45 },
		{ 416, 239, 45 },
		{ 416, 240, 45 },
		{ 416, 241, 45 },
		{ 416, 242, 45 },
		{ 416, 243, 45 },
		{ 416, 244, 45 },
		{ 416, 245, 45 },
		{ 416, 246, 45 },
		{ 416, 247, 45 },
		{ 416, 248, 45 },
		{ 416, 249, 45 },
		{ 416, 250, 45 },
		{ 416, 251, 45 },
		{ 416, 252, 45 },
		{ 416, 253, 45 },
		{ 416, 254, 45 },
		{ 416, 255, 45 },
		{ 416, 256, 45 },
		{ 417, 218, 45 },
		{ 417, 219, 45 },
		{ 417, 220, 45 },
		{ 417, 221, 45 },
		{ 417, 222, 45 },
		{ 417, 223, 45 },
		{ 417, 224, 45 },
		{ 417, 225, 45 },
		{ 417, 226, 45 },
		{ 417, 227, 45 },
		{ 417, 228, 45 },
		{ 417, 229, 45 },
		{ 417, 230, 45 },
		{ 417, 231, 45 },
		{ 417, 232, 45 },
		{ 417, 233, 45 },
		{ 417, 234, 45 },
		{ 417, 235, 45 },
		{ 417, 236, 45 },
		{ 417, 237, 45 },
		{ 417, 238, 45 },
		{ 417, 239, 45 },
		{ 417, 240, 45 },
		{ 417, 241, 45 },
		{ 417, 242, 45 },
		{ 417, 243, 45 },
		{ 417, 244, 45 },
		{ 417, 245, 45 },
		{ 417, 246, 45 },
		{ 417, 247, 45 },
		{ 417, 248, 45 },
		{ 417, 249, 45 },
		{ 417, 250, 45 },
		{ 417, 251, 45 },
		{ 417, 252, 45 },
		{ 417, 253, 45 },
		{ 417, 254, 45 },
		{ 417, 255, 45 },
		{ 417, 256, 45 },
		{ 418, 217, 45 },
		{ 418, 218, 45 },
		{ 418, 219, 45 },
		{ 418, 220, 45 },
		{ 418, 221, 45 },
		{ 418, 222, 45 },
		{ 418, 223, 45 },
		{ 418, 224, 45 },
		{ 418, 225, 45 },
		{ 418, 226, 45 },
		{ 418, 227, 45 },
		{ 418, 228, 45 },
		{ 418, 229, 45 },
		{ 418, 230, 45 },
		{ 418, 231, 45 },
		{ 418, 232, 45 },
		{ 418, 233, 45 },
		{ 418, 234, 45 },
		{ 418, 235, 45 },
		{ 418, 236, 45 },
		{ 418, 237, 45 },
		{ 418, 238, 45 },
		{ 418, 239, 45 },
		{ 418, 240, 45 },
		{ 418, 241, 45 },
		{ 418, 242, 45 },
		{ 418, 243, 45 },
		{ 418, 244, 45 },
		{ 418, 245, 45 },
		{ 418, 246, 45 },
		{ 418, 247, 45 },
		{ 418, 248, 45 },
		{ 418, 249, 45 },
		{ 418, 250, 45 },
		{ 418, 251, 45 },
		{ 418, 252, 45 },
		{ 418, 253, 45 },
		{ 418, 254, 45 },
		{ 418, 255, 45 },
		{ 418, 256, 45 },
		{ 419, 217, 45 },
		{ 419, 218, 45 },
		{ 419, 219, 45 },
		{ 419, 220, 45 },
		{ 419, 221, 45 },
		{ 419, 222, 45 },
		{ 419, 223, 45 },
		{ 419, 224, 45 },
		{ 419, 225, 45 },
		{ 419, 226, 45 },
		{ 419, 227, 45 },
		{ 419, 228, 45 },
		{ 419, 229, 45 },
		{ 419, 230, 45 },
		{ 419, 231, 45 },
		{ 419, 232, 45 },
		{ 419, 233, 45 },
		{ 419, 234, 45 },
		{ 419, 235, 45 },
		{ 419, 236, 45 },
		{ 419, 237, 45 },
		{ 419, 238, 45 },
		{ 419, 239, 45 },
		{ 419, 240, 45 },
		{ 419, 241, 45 },
		{ 419, 242, 45 },
		{ 419, 243, 45 },
		{ 419, 244, 45 },
		{ 419, 245, 45 },
		{ 419, 246, 45 },
		{ 419, 247, 45 },
		{ 419, 248, 45 },
		{ 419, 249, 45 },
		{ 419, 250, 45 },
		{ 419, 251, 45 },
		{ 419, 252, 45 },
		{ 419, 253, 45 },
		{ 419, 254, 45 },
		{ 419, 255, 45 },
		{ 419, 256, 45 },
		{ 420, 217, 45 },
		{ 420, 218, 45 },
		{ 420, 219, 45 },
		{ 420, 220, 45 },
		{ 420, 221, 45 },
		{ 420, 222, 45 },
		{ 420, 223, 45 },
		{ 420, 224, 45 },
		{ 420, 225, 45 },
		{ 420, 226, 45 },
		{ 420, 227, 45 },
		{ 420, 228, 45 },
		{ 420, 229, 45 },
		{ 420, 230, 45 },
		{ 420, 231, 45 },
		{ 420, 232, 45 },
		{ 420, 233, 45 },
		{ 420, 234, 45 },
		{ 420, 235, 45 },
		{ 420, 236, 45 },
		{ 420, 237, 45 },
		{ 420, 238, 45 },
		{ 420, 239, 45 },
		{ 420, 240, 45 },
		{ 420, 241, 45 },
		{ 420, 242, 45 },
		{ 420, 243, 45 },
		{ 420, 244, 45 },
		{ 420, 245, 45 },
		{ 420, 246, 45 },
		{ 420, 247, 45 },
		{ 420, 248, 45 },
		{ 420, 249, 45 },
		{ 420, 250, 45 },
		{ 420, 251, 45 },
		{ 420, 252, 45 },
		{ 420, 253, 45 },
		{ 420, 254, 45 },
		{ 420, 255, 45 },
		{ 420, 256, 45 },
		{ 421, 217, 45 },
		{ 421, 218, 45 },
		{ 421, 219, 45 },
		{ 421, 220, 45 },
		{ 421, 221, 45 },
		{ 421, 222, 45 },
		{ 421, 223, 45 },
		{ 421, 224, 45 },
		{ 421, 225, 45 },
		{ 421, 226, 45 },
		{ 421, 227, 45 },
		{ 421, 228, 45 },
		{ 421, 229, 45 },
		{ 421, 230, 45 },
		{ 421, 231, 45 },
		{ 421, 232, 45 },
		{ 421, 233, 45 },
		{ 421, 234, 45 },
		{ 421, 235, 45 },
		{ 421, 236, 45 },
		{ 421, 237, 45 },
		{ 421, 238, 45 },
		{ 421, 239, 45 },
		{ 421, 240, 45 },
		{ 421, 241, 45 },
		{ 421, 242, 45 },
		{ 421, 243, 45 },
		{ 421, 244, 45 },
		{ 421, 245, 45 },
		{ 421, 246, 45 },
		{ 421, 247, 45 },
		{ 421, 248, 45 },
		{ 421, 249, 45 },
		{ 421, 250, 45 },
		{ 421, 251, 45 },
		{ 421, 252, 45 },
		{ 421, 253, 45 },
		{ 421, 254, 45 },
		{ 421, 255, 45 },
		{ 421, 256, 45 },
		{ 422, 216, 45 },
		{ 422, 217, 45 },
		{ 422, 218, 45 },
		{ 422, 219, 45 },
		{ 422, 220, 45 },
		{ 422, 221, 45 },
		{ 422, 222, 45 },
		{ 422, 223, 45 },
		{ 422, 224, 45 },
		{ 422, 225, 45 },
		{ 422, 226, 45 },
		{ 422, 227, 45 },
		{ 422, 228, 45 },
		{ 422, 229, 45 },
		{ 422, 230, 45 },
		{ 422, 231, 45 },
		{ 422, 232, 45 },
		{ 422, 233, 45 },
		{ 422, 234, 45 },
		{ 422, 235, 45 },
		{ 422, 236, 45 },
		{ 422, 237, 45 },
		{ 422, 238, 45 },
		{ 422, 239, 45 },
		{ 422, 240, 45 },
		{ 422, 241, 45 },
		{ 422, 242, 45 },
		{ 422, 243, 45 },
		{ 422, 244, 45 },
		{ 422, 245, 45 },
		{ 422, 246, 45 },
		{ 422, 247, 45 },
		{ 422, 248, 45 },
		{ 422, 249, 45 },
		{ 422, 250, 45 },
		{ 422, 251, 45 },
		{ 422, 252, 45 },
		{ 422, 253, 45 },
		{ 422, 254, 45 },
		{ 422, 255, 45 },
		{ 422, 256, 45 },
		{ 423, 216, 45 },
		{ 423, 217, 45 },
		{ 423, 218, 45 },
		{ 423, 219, 45 },
		{ 423, 220, 45 },
		{ 423, 221, 45 },
		{ 423, 222, 45 },
		{ 423, 223, 45 },
		{ 423, 224, 45 },
		{ 423, 225, 45 },
		{ 423, 226, 45 },
		{ 423, 227, 45 },
		{ 423, 228, 45 },
		{ 423, 229, 45 },
		{ 423, 230, 45 },
		{ 423, 231, 45 },
		{ 423, 232, 45 },
		{ 423, 233, 45 },
		{ 423, 234, 45 },
		{ 423, 235, 45 },
		{ 423, 236, 45 },
		{ 423, 237, 45 },
		{ 423, 238, 45 },
		{ 423, 239, 45 },
		{ 423, 240, 45 },
		{ 423, 241, 45 },
		{ 423, 242, 45 },
		{ 423, 243, 45 },
		{ 423, 244, 45 },
		{ 423, 245, 45 },
		{ 423, 246, 45 },
		{ 423, 247, 45 },
		{ 423, 248, 45 },
		{ 423, 249, 45 },
		{ 423, 250, 45 },
		{ 423, 251, 45 },
		{ 423, 252, 45 },
		{ 423, 253, 45 },
		{ 423, 254, 45 },
		{ 423, 255, 45 },
		{ 423, 256, 45 },
		{ 424, 216, 45 },
		{ 424, 217, 45 },
		{ 424, 218, 45 },
		{ 424, 219, 45 },
		{ 424, 220, 45 },
		{ 424, 221, 45 },
		{ 424, 222, 45 },
		{ 424, 223, 45 },
		{ 424, 224, 45 },
		{ 424, 225, 45 },
		{ 424, 226, 45 },
		{ 424, 227, 45 },
		{ 424, 228, 45 },
		{ 424, 229, 45 },
		{ 424, 230, 45 },
		{ 424, 231, 45 },
		{ 424, 232, 45 },
		{ 424, 233, 45 },
		{ 424, 234, 45 },
		{ 424, 235, 45 },
		{ 424, 236, 45 },
		{ 424, 237, 45 },
		{ 424, 238, 45 },
		{ 424, 239, 45 },
		{ 424, 240, 45 },
		{ 424, 241, 45 },
		{ 424, 242, 45 },
		{ 424, 243, 45 },
		{ 424, 244, 45 },
		{ 424, 245, 45 },
		{ 424, 246, 45 },
		{ 424, 247, 45 },
		{ 424, 248, 45 },
		{ 424, 249, 45 },
		{ 424, 250, 45 },
		{ 424, 251, 45 },
		{ 424, 252, 45 },
		{ 424, 253, 45 },
		{ 424, 254, 45 },
		{ 424, 255, 45 },
		{ 424, 256, 45 },
		{ 425, 216, 45 },
		{ 425, 217, 45 },
		{ 425, 218, 45 },
		{ 425, 219, 45 },
		{ 425, 220, 45 },
		{ 425, 221, 45 },
		{ 425, 222, 45 },
		{ 425, 223, 45 },
		{ 425, 224, 45 },
		{ 425, 225, 45 },
		{ 425, 226, 45 },
		{ 425, 227, 45 },
		{ 425, 228, 45 },
		{ 425, 229, 45 },
		{ 425, 230, 45 },
		{ 425, 231, 45 },
		{ 425, 232, 45 },
		{ 425, 233, 45 },
		{ 425, 234, 45 },
		{ 425, 235, 45 },
		{ 425, 236, 45 },
		{ 425, 237, 45 },
		{ 425, 238, 45 },
		{ 425, 239, 45 },
		{ 425, 240, 45 },
		{ 425, 241, 45 },
		{ 425, 242, 45 },
		{ 425, 243, 45 },
		{ 425, 244, 45 },
		{ 425, 245, 45 },
		{ 425, 246, 45 },
		{ 425, 247, 45 },
		{ 425, 248, 45 },
		{ 425, 249, 45 },
		{ 425, 250, 45 },
		{ 425, 251, 45 },
		{ 425, 252, 45 },
		{ 425, 253, 45 },
		{ 425, 254, 45 },
		{ 425, 255, 45 },
		{ 425, 256, 45 },
		{ 426, 215, 45 },
		{ 426, 216, 45 },
		{ 426, 217, 45 },
		{ 426, 218, 45 },
		{ 426, 219, 45 },
		{ 426, 220, 45 },
		{ 426, 221, 45 },
		{ 426, 222, 45 },
		{ 426, 223, 45 },
		{ 426, 224, 45 },
		{ 426, 225, 45 },
		{ 426, 226, 45 },
		{ 426, 227, 45 },
		{ 426, 228, 45 },
		{ 426, 229, 45 },
		{ 426, 230, 45 },
		{ 426, 231, 45 },
		{ 426, 232, 45 },
		{ 426, 233, 45 },
		{ 426, 234, 45 },
		{ 426, 235, 45 },
		{ 426, 236, 45 },
		{ 426, 237, 45 },
		{ 426, 238, 45 },
		{ 426, 239, 45 },
		{ 426, 240, 45 },
		{ 426, 241, 45 },
		{ 426, 242, 45 },
		{ 426, 243, 45 },
		{ 426, 244, 45 },
		{ 426, 245, 45 },
		{ 426, 246, 45 },
		{ 426, 247, 45 },
		{ 426, 248, 45 },
		{ 426, 249, 45 },
		{ 426, 250, 45 },
		{ 426, 251, 45 },
		{ 426, 252, 45 },
		{ 426, 253, 45 },
		{ 426, 254, 45 },
		{ 426, 255, 45 },
		{ 426, 256, 45 },
		{ 427, 215, 45 },
		{ 427, 216, 45 },
		{ 427, 217, 45 },
		{ 427, 218, 45 },
		{ 427, 219, 45 },
		{ 427, 220, 45 },
		{ 427, 221, 45 },
		{ 427, 222, 45 },
		{ 427, 223, 45 },
		{ 427, 224, 45 },
		{ 427, 225, 45 },
		{ 427, 226, 45 },
		{ 427, 227, 45 },
		{ 427, 228, 45 },
		{ 427, 229, 45 },
		{ 427, 230, 45 },
		{ 427, 231, 45 },
		{ 427, 232, 45 },
		{ 427, 233, 45 },
		{ 427, 234, 45 },
		{ 427, 235, 45 },
		{ 427, 236, 45 },
		{ 427, 237, 45 },
		{ 427, 238, 45 },
		{ 427, 239, 45 },
		{ 427, 240, 45 },
		{ 427, 241, 45 },
		{ 427, 242, 45 },
		{ 427, 243, 45 },
		{ 427, 244, 45 },
		{ 427, 245, 45 },
		{ 427, 246, 45 },
		{ 427, 247, 45 },
		{ 427, 248, 45 },
		{ 427, 249, 45 },
		{ 427, 250, 45 },
		{ 427, 251, 45 },
		{ 427, 252, 45 },
		{ 427, 253, 45 },
		{ 427, 254, 45 },
		{ 427, 255, 45 },
		{ 427, 256, 45 },
		{ 428, 215, 45 },
		{ 428, 216, 45 },
		{ 428, 217, 45 },
		{ 428, 218, 45 },
		{ 428, 219, 45 },
		{ 428, 220, 45 },
		{ 428, 221, 45 },
		{ 428, 222, 45 },
		{ 428, 223, 45 },
		{ 428, 224, 45 },
		{ 428, 225, 45 },
		{ 428, 226, 45 },
		{ 428, 227, 45 },
		{ 428, 228, 45 },
		{ 428, 229, 45 },
		{ 428, 230, 45 },
		{ 428, 231, 45 },
		{ 428, 232, 45 },
		{ 428, 233, 45 },
		{ 428, 234, 45 },
		{ 428, 235, 45 },
		{ 428, 236, 45 },
		{ 428, 237, 45 },
		{ 428, 238, 45 },
		{ 428, 239, 45 },
		{ 428, 240, 45 },
		{ 428, 241, 45 },
		{ 428, 242, 45 },
		{ 428, 243, 45 },
		{ 428, 244, 45 },
		{ 428, 245, 45 },
		{ 428, 246, 45 },
		{ 428, 247, 45 },
		{ 428, 248, 45 },
		{ 428, 249, 45 },
		{ 428, 250, 45 },
		{ 428, 251, 45 },
		{ 428, 252, 45 },
		{ 428, 253, 45 },
		{ 428, 254, 45 },
		{ 428, 255, 45 },
		{ 428, 256, 45 },
		{ 429, 215, 45 },
		{ 429, 216, 45 },
		{ 429, 217, 45 },
		{ 429, 218, 45 },
		{ 429, 219, 45 },
		{ 429, 220, 45 },
		{ 429, 221, 45 },
		{ 429, 222, 45 },
		{ 429, 223, 45 },
		{ 429, 224, 45 },
		{ 429, 225, 45 },
		{ 429, 226, 45 },
		{ 429, 227, 45 },
		{ 429, 228, 45 },
		{ 429, 229, 45 },
		{ 429, 230, 45 },
		{ 429, 231, 45 },
		{ 429, 232, 45 },
		{ 429, 233, 45 },
		{ 429, 234, 45 },
		{ 429, 235, 45 },
		{ 429, 236, 45 },
		{ 429, 237, 45 },
		{ 429, 238, 45 },
		{ 429, 239, 45 },
		{ 429, 240, 45 },
		{ 429, 241, 45 },
		{ 429, 242, 45 },
		{ 429, 243, 45 },
		{ 429, 244, 45 },
		{ 429, 245, 45 },
		{ 429, 246, 45 },
		{ 429, 247, 45 },
		{ 429, 248, 45 },
		{ 429, 249, 45 },
		{ 429, 250, 45 },
		{ 429, 251, 45 },
		{ 429, 252, 45 },
		{ 429, 253, 45 },
		{ 429, 254, 45 },
		{ 429, 255, 45 },
		{ 429, 256, 45 },
		{ 430, 214, 45 },
		{ 430, 215, 45 },
		{ 430, 216, 45 },
		{ 430, 217, 45 },
		{ 430, 218, 45 },
		{ 430, 219, 45 },
		{ 430, 220, 45 },
		{ 430, 221, 45 },
		{ 430, 222, 45 },
		{ 430, 223, 45 },
		{ 430, 224, 45 },
		{ 430, 225, 45 },
		{ 430, 226, 45 },
		{ 430, 227, 45 },
		{ 430, 228, 45 },
		{ 430, 229, 45 },
		{ 430, 230, 45 },
		{ 430, 231, 45 },
		{ 430, 232, 45 },
		{ 430, 233, 45 },
		{ 430, 234, 45 },
		{ 430, 235, 45 },
		{ 430, 236, 45 },
		{ 430, 237, 45 },
		{ 430, 238, 45 },
		{ 430, 239, 45 },
		{ 430, 240, 45 },
		{ 430, 241, 45 },
		{ 430, 242, 45 },
		{ 430, 243, 45 },
		{ 430, 244, 45 },
		{ 430, 245, 45 },
		{ 430, 246, 45 },
		{ 430, 247, 45 },
		{ 430, 248, 45 },
		{ 430, 249, 45 },
		{ 430, 250, 45 },
		{ 430, 251, 45 },
		{ 430, 252, 45 },
		{ 430, 253, 45 },
		{ 430, 254, 45 },
		{ 430, 255, 45 },
		{ 430, 256, 45 },
		{ 431, 214, 45 },
		{ 431, 215, 45 },
		{ 431, 216, 45 },
		{ 431, 217, 45 },
		{ 431, 218, 45 },
		{ 431, 219, 45 },
		{ 431, 220, 45 },
		{ 431, 221, 45 },
		{ 431, 222, 45 },
		{ 431, 223, 45 },
		{ 431, 224, 45 },
		{ 431, 225, 45 },
		{ 431, 226, 45 },
		{ 431, 227, 45 },
		{ 431, 228, 45 },
		{ 431, 229, 45 },
		{ 431, 230, 45 },
		{ 431, 231, 45 },
		{ 431, 232, 45 },
		{ 431, 233, 45 },
		{ 431, 234, 45 },
		{ 431, 235, 45 },
		{ 431, 236, 45 },
		{ 431, 237, 45 },
		{ 431, 238, 45 },
		{ 431, 239, 45 },
		{ 431, 240, 45 },
		{ 431, 241, 45 },
		{ 431, 242, 45 },
		{ 431, 243, 45 },
		{ 431, 244, 45 },
		{ 431, 245, 45 },
		{ 431, 246, 45 },
		{ 431, 247, 45 },
		{ 431, 248, 45 },
		{ 431, 249, 45 },
		{ 431, 250, 45 },
		{ 431, 251, 45 },
		{ 431, 252, 45 },
		{ 431, 253, 45 },
		{ 431, 254, 45 },
		{ 431, 255, 45 },
		{ 431, 256, 45 },
		{ 432, 214, 45 },
		{ 432, 215, 45 },
		{ 432, 216, 45 },
		{ 432, 217, 45 },
		{ 432, 218, 45 },
		{ 432, 219, 45 },
		{ 432, 220, 45 },
		{ 432, 221, 45 },
		{ 432, 222, 45 },
		{ 432, 223, 45 },
		{ 432, 224, 45 },
		{ 432, 225, 45 },
		{ 432, 226, 45 },
		{ 432, 227, 45 },
		{ 432, 228, 45 },
		{ 432, 229, 45 },
		{ 432, 230, 45 },
		{ 432, 231, 45 },
		{ 432, 232, 45 },
		{ 432, 233, 45 },
		{ 432, 234, 45 },
		{ 432, 235, 45 },
		{ 432, 236, 45 },
		{ 432, 237, 45 },
		{ 432, 238, 45 },
		{ 432, 239, 45 },
		{ 432, 240, 45 },
		{ 432, 241, 45 },
		{ 432, 242, 45 },
		{ 432, 243, 45 },
		{ 432, 244, 45 },
		{ 432, 245, 45 },
		{ 432, 246, 45 },
		{ 432, 247, 45 },
		{ 432, 248, 45 },
		{ 432, 249, 45 },
		{ 432, 250, 45 },
		{ 432, 251, 45 },
		{ 432, 252, 45 },
		{ 432, 253, 45 },
		{ 432, 254, 45 },
		{ 432, 255, 45 },
		{ 432, 256, 45 },
		{ 433, 214, 45 },
		{ 433, 215, 45 },
		{ 433, 216, 45 },
		{ 433, 217, 45 },
		{ 433, 218, 45 },
		{ 433, 219, 45 },
		{ 433, 220, 45 },
		{ 433, 221, 45 },
		{ 433, 222, 45 },
		{ 433, 223, 45 },
		{ 433, 224, 45 },
		{ 433, 225, 45 },
		{ 433, 226, 45 },
		{ 433, 227, 45 },
		{ 433, 228, 45 },
		{ 433, 229, 45 },
		{ 433, 230, 45 },
		{ 433, 231, 45 },
		{ 433, 232, 45 },
		{ 433, 233, 45 },
		{ 433, 234, 45 },
		{ 433, 235, 45 },
		{ 433, 236, 45 },
		{ 433, 237, 45 },
		{ 433, 238, 45 },
		{ 433, 239, 45 },
		{ 433, 240, 45 },
		{ 433, 241, 45 },
		{ 433, 242, 45 },
		{ 433, 243, 45 },
		{ 433, 244, 45 },
		{ 433, 245, 45 },
		{ 433, 246, 45 },
		{ 433, 247, 45 },
		{ 433, 248, 45 },
		{ 433, 249, 45 },
		{ 433, 250, 45 },
		{ 433, 251, 45 },
		{ 433, 252, 45 },
		{ 433, 253, 45 },
		{ 433, 254, 45 },
		{ 433, 255, 45 },
		{ 433, 256, 45 },
		{ 434, 213, 45 },
		{ 434, 214, 45 },
		{ 434, 215, 45 },
		{ 434, 216, 45 },
		{ 434, 217, 45 },
		{ 434, 218, 45 },
		{ 434, 219, 45 },
		{ 434, 220, 45 },
		{ 434, 221, 45 },
		{ 434, 222, 45 },
		{ 434, 223, 45 },
		{ 434, 224, 45 },
		{ 434, 225, 45 },
		{ 434, 226, 45 },
		{ 434, 227, 45 },
		{ 434, 228, 45 },
		{ 434, 229, 45 },
		{ 434, 230, 45 },
		{ 434, 231, 45 },
		{ 434, 232, 45 },
		{ 434, 233, 45 },
		{ 434, 234, 45 },
		{ 434, 235, 45 },
		{ 434, 236, 45 },
		{ 434, 237, 45 },
		{ 434, 238, 45 },
		{ 434, 239, 45 },
		{ 434, 240, 45 },
		{ 434, 241, 45 },
		{ 434, 242, 45 },
		{ 434, 243, 45 },
		{ 434, 244, 45 },
		{ 434, 245, 45 },
		{ 434, 246, 45 },
		{ 434, 247, 45 },
		{ 434, 248, 45 },
		{ 434, 249, 45 },
		{ 434, 250, 45 },
		{ 434, 251, 45 },
		{ 434, 252, 45 },
		{ 434, 253, 45 },
		{ 434, 254, 45 },
		{ 434, 255, 45 },
		{ 434, 256, 45 },
		{ 435, 213, 45 },
		{ 435, 214, 45 },
		{ 435, 215, 45 },
		{ 435, 216, 45 },
		{ 435, 217, 45 },
		{ 435, 218, 45 },
		{ 435, 219, 45 },
		{ 435, 220, 45 },
		{ 435, 221, 45 },
		{ 435, 222, 45 },
		{ 435, 223, 45 },
		{ 435, 224, 45 },
		{ 435, 225, 45 },
		{ 435, 226, 45 },
		{ 435, 227, 45 },
		{ 435, 228, 45 },
		{ 435, 229, 45 },
		{ 435, 230, 45 },
		{ 435, 231, 45 },
		{ 435, 232, 45 },
		{ 435, 233, 45 },
		{ 435, 234, 45 },
		{ 435, 235, 45 },
		{ 435, 236, 45 },
		{ 435, 237, 45 },
		{ 435, 238, 45 },
		{ 435, 239, 45 },
		{ 435, 240, 45 },
		{ 435, 241, 45 },
		{ 435, 242, 45 },
		{ 435, 243, 45 },
		{ 435, 244, 45 },
		{ 435, 245, 45 },
		{ 435, 246, 45 },
		{ 435, 247, 45 },
		{ 435, 248, 45 },
		{ 435, 249, 45 },
		{ 435, 250, 45 },
		{ 435, 251, 45 },
		{ 435, 252, 45 },
		{ 435, 253, 45 },
		{ 435, 254, 45 },
		{ 435, 255, 45 },
		{ 435, 256, 45 },
		{ 436, 213, 45 },
		{ 436, 214, 45 },
		{ 436, 215, 45 },
		{ 436, 216, 45 },
		{ 436, 217, 45 },
		{ 436, 218, 45 },
		{ 436, 219, 45 },
		{ 436, 220, 45 },
		{ 436, 221, 45 },
		{ 436, 222, 45 },
		{ 436, 223, 45 },
		{ 436, 224, 45 },
		{ 436, 225, 45 },
		{ 436, 226, 45 },
		{ 436, 227, 45 },
		{ 436, 228, 45 },
		{ 436, 229, 45 },
		{ 436, 230, 45 },
		{ 436, 231, 45 },
		{ 436, 232, 45 },
		{ 436, 233, 45 },
		{ 436, 234, 45 },
		{ 436, 235, 45 },
		{ 436, 236, 45 },
		{ 436, 237, 45 },
		{ 436, 238, 45 },
		{ 436, 239, 45 },
		{ 436, 240, 45 },
		{ 436, 241, 45 },
		{ 436, 242, 45 },
		{ 436, 243, 45 },
		{ 436, 244, 45 },
		{ 436, 245, 45 },
		{ 436, 246, 45 },
		{ 436, 247, 45 },
		{ 436, 248, 45 },
		{ 436, 249, 45 },
		{ 436, 250, 45 },
		{ 436, 251, 45 },
		{ 436, 252, 45 },
		{ 436, 253, 45 },
		{ 436, 254, 45 },
		{ 436, 255, 45 },
		{ 436, 256, 45 },
		{ 437, 213, 45 },
		{ 437, 214, 45 },
		{ 437, 215, 45 },
		{ 437, 216, 45 },
		{ 437, 217, 45 },
		{ 437, 218, 45 },
		{ 437, 219, 45 },
		{ 437, 220, 45 },
		{ 437, 221, 45 },
		{ 437, 222, 45 },
		{ 437, 223, 45 },
		{ 437, 224, 45 },
		{ 437, 225, 45 },
		{ 437, 226, 45 },
		{ 437, 227, 45 },
		{ 437, 228, 45 },
		{ 437, 229, 45 },
		{ 437, 230, 45 },
		{ 437, 231, 45 },
		{ 437, 232, 45 },
		{ 437, 233, 45 },
		{ 437, 234, 45 },
		{ 437, 235, 45 },
		{ 437, 236, 45 },
		{ 437, 237, 45 },
		{ 437, 238, 45 },
		{ 437, 239, 45 },
		{ 437, 240, 45 },
		{ 437, 241, 45 },
		{ 437, 242, 45 },
		{ 437, 243, 45 },
		{ 437, 244, 45 },
		{ 437, 245, 45 },
		{ 437, 246, 45 },
		{ 437, 247, 45 },
		{ 437, 248, 45 },
		{ 437, 249, 45 },
		{ 437, 250, 45 },
		{ 437, 251, 45 },
		{ 437, 252, 45 },
		{ 437, 253, 45 },
		{ 437, 254, 45 },
		{ 437, 255, 45 },
		{ 437, 256, 45 },
		{ 438, 212, 45 },
		{ 438, 213, 45 },
		{ 438, 214, 45 },
		{ 438, 215, 45 },
		{ 438, 216, 45 },
		{ 438, 217, 45 },
		{ 438, 218, 45 },
		{ 438, 219, 45 },
		{ 438, 220, 45 },
		{ 438, 221, 45 },
		{ 438, 222, 45 },
		{ 438, 223, 45 },
		{ 438, 224, 45 },
		{ 438, 225, 45 },
		{ 438, 226, 45 },
		{ 438, 227, 45 },
		{ 438, 228, 45 },
		{ 438, 229, 45 },
		{ 438, 230, 45 },
		{ 438, 231, 45 },
		{ 438, 232, 45 },
		{ 438, 233, 45 },
		{ 438, 234, 45 },
		{ 438, 235, 45 },
		{ 438, 236, 45 },
		{ 438, 237, 45 },
		{ 438, 238, 45 },
		{ 438, 239, 45 },
		{ 438, 240, 45 },
		{ 438, 241, 45 },
		{ 438, 242, 45 },
		{ 438, 243, 45 },
		{ 438, 244, 45 },
		{ 438, 245, 45 },
		{ 438, 246, 45 },
		{ 438, 247, 45 },
		{ 438, 248, 45 },
		{ 438, 249, 45 },
		{ 438, 250, 45 },
		{ 438, 251, 45 },
		{ 438, 252, 45 },
		{ 438, 253, 45 },
		{ 438, 254, 45 },
		{ 438, 255, 45 },
		{ 438, 256, 45 },
		{ 439, 212, 45 },
		{ 439, 213, 45 },
		{ 439, 214, 45 },
		{ 439, 215, 45 },
		{ 439, 216, 45 },
		{ 439, 217, 45 },
		{ 439, 218, 45 },
		{ 439, 219, 45 },
		{ 439, 220, 45 },
		{ 439, 221, 45 },
		{ 439, 222, 45 },
		{ 439, 223, 45 },
		{ 439, 224, 45 },
		{ 439, 225, 45 },
		{ 439, 226, 45 },
		{ 439, 227, 45 },
		{ 439, 228, 45 },
		{ 439, 229, 45 },
		{ 439, 230, 45 },
		{ 439, 231, 45 },
		{ 439, 232, 45 },
		{ 439, 233, 45 },
		{ 439, 234, 45 },
		{ 439, 235, 45 },
		{ 439, 236, 45 },
		{ 439, 237, 45 },
		{ 439, 238, 45 },
		{ 439, 239, 45 },
		{ 439, 240, 45 },
		{ 439, 241, 45 },
		{ 439, 242, 45 },
		{ 439, 243, 45 },
		{ 439, 244, 45 },
		{ 439, 245, 45 },
		{ 439, 246, 45 },
		{ 439, 247, 45 },
		{ 439, 248, 45 },
		{ 439, 249, 45 },
		{ 439, 250, 45 },
		{ 439, 251, 45 },
		{ 439, 252, 45 },
		{ 439, 253, 45 },
		{ 439, 254, 45 },
		{ 439, 255, 45 },
		{ 439, 256, 45 },
		{ 440, 212, 45 },
		{ 440, 213, 45 },
		{ 440, 214, 45 },
		{ 440, 215, 45 },
		{ 440, 216, 45 },
		{ 440, 217, 45 },
		{ 440, 218, 45 },
		{ 440, 219, 45 },
		{ 440, 220, 45 },
		{ 440, 221, 45 },
		{ 440, 222, 45 },
		{ 440, 223, 45 },
		{ 440, 224, 45 },
		{ 440, 225, 45 },
		{ 440, 226, 45 },
		{ 440, 227, 45 },
		{ 440, 228, 45 },
		{ 440, 229, 45 },
		{ 440, 230, 45 },
		{ 440, 231, 45 },
		{ 440, 232, 45 },
		{ 440, 233, 45 },
		{ 440, 234, 45 },
		{ 440, 235, 45 },
		{ 440, 236, 45 },
		{ 440, 237, 45 },
		{ 440, 238, 45 },
		{ 440, 239, 45 },
		{ 440, 240, 45 },
		{ 440, 241, 45 },
		{ 440, 242, 45 },
		{ 440, 243, 45 },
		{ 440, 244, 45 },
		{ 440, 245, 45 },
		{ 440, 246, 45 },
		{ 440, 247, 45 },
		{ 440, 248, 45 },
		{ 440, 249, 45 },
		{ 440, 250, 45 },
		{ 440, 251, 45 },
		{ 440, 252, 45 },
		{ 440, 253, 45 },
		{ 440, 254, 45 },
		{ 440, 255, 45 },
		{ 440, 256, 45 },
		{ 441, 212, 45 },
		{ 441, 213, 45 },
		{ 441, 214, 45 },
		{ 441, 215, 45 },
		{ 441, 216, 45 },
		{ 441, 217, 45 },
		{ 441, 218, 45 },
		{ 441, 219, 45 },
		{ 441, 220, 45 },
		{ 441, 221, 45 },
		{ 441, 222, 45 },
		{ 441, 223, 45 },
		{ 441, 224, 45 },
		{ 441, 225, 45 },
		{ 441, 226, 45 },
		{ 441, 227, 45 },
		{ 441, 228, 45 },
		{ 441, 229, 45 },
		{ 441, 230, 45 },
		{ 441, 231, 45 },
		{ 441, 232, 45 },
		{ 441, 233, 45 },
		{ 441, 234, 45 },
		{ 441, 235, 45 },
		{ 441, 236, 45 },
		{ 441, 237, 45 },
		{ 441, 238, 45 },
		{ 441, 239, 45 },
		{ 441, 240, 45 },
		{ 441, 241, 45 },
		{ 441, 242, 45 },
		{ 441, 243, 45 },
		{ 441, 244, 45 },
		{ 441, 245, 45 },
		{ 441, 246, 45 },
		{ 441, 247, 45 },
		{ 441, 248, 45 },
		{ 441, 249, 45 },
		{ 441, 250, 45 },
		{ 441, 251, 45 },
		{ 441, 252, 45 },
		{ 441, 253, 45 },
		{ 441, 254, 45 },
		{ 441, 255, 45 },
		{ 441, 256, 45 },
		{ 442, 211, 45 },
		{ 442, 212, 45 },
		{ 442, 213, 45 },
		{ 442, 214, 45 },
		{ 442, 215, 45 },
		{ 442, 216, 45 },
		{ 442, 217, 45 },
		{ 442, 218, 45 },
		{ 442, 219, 45 },
		{ 442, 220, 45 },
		{ 442, 221, 45 },
		{ 442, 222, 45 },
		{ 442, 223, 45 },
		{ 442, 224, 45 },
		{ 442, 225, 45 },
		{ 442, 226, 45 },
		{ 442, 227, 45 },
		{ 442, 228, 45 },
		{ 442, 229, 45 },
		{ 442, 230, 45 },
		{ 442, 231, 45 },
		{ 442, 232, 45 },
		{ 442, 233, 45 },
		{ 442, 234, 45 },
		{ 442, 235, 45 },
		{ 442, 236, 45 },
		{ 442, 237, 45 },
		{ 442, 238, 45 },
		{ 442, 239, 45 },
		{ 442, 240, 45 },
		{ 442, 241, 45 },
		{ 442, 242, 45 },
		{ 442, 243, 45 },
		{ 442, 244, 45 },
		{ 442, 245, 45 },
		{ 442, 246, 45 },
		{ 442, 247, 45 },
		{ 442, 248, 45 },
		{ 442, 249, 45 },
		{ 442, 250, 45 },
		{ 442, 251, 45 },
		{ 442, 252, 45 },
		{ 442, 253, 45 },
		{ 442, 254, 45 },
		{ 442, 255, 45 },
		{ 442, 256, 45 },
		{ 443, 211, 45 },
		{ 443, 212, 45 },
		{ 443, 213, 45 },
		{ 443, 214, 45 },
		{ 443, 215, 45 },
		{ 443, 216, 45 },
		{ 443, 217, 45 },
		{ 443, 218, 45 },
		{ 443, 219, 45 },
		{ 443, 220, 45 },
		{ 443, 221, 45 },
		{ 443, 222, 45 },
		{ 443, 223, 45 },
		{ 443, 224, 45 },
		{ 443, 225, 45 },
		{ 443, 226, 45 },
		{ 443, 227, 45 },
		{ 443, 228, 45 },
		{ 443, 229, 45 },
		{ 443, 230, 45 },
		{ 443, 231, 45 },
		{ 443, 232, 45 },
		{ 443, 233, 45 },
		{ 443, 234, 45 },
		{ 443, 235, 45 },
		{ 443, 236, 45 },
		{ 443, 237, 45 },
		{ 443, 238, 45 },
		{ 443, 239, 45 },
		{ 443, 240, 45 },
		{ 443, 241, 45 },
		{ 443, 242, 45 },
		{ 443, 243, 45 },
		{ 443, 244, 45 },
		{ 443, 245, 45 },
		{ 443, 246, 45 },
		{ 443, 247, 45 },
		{ 443, 248, 45 },
		{ 443, 249, 45 },
		{ 443, 250, 45 },
		{ 443, 251, 45 },
		{ 443, 252, 45 },
		{ 443, 253, 45 },
		{ 443, 254, 45 },
		{ 443, 255, 45 },
		{ 443, 256, 45 },
		{ 444, 211, 45 },
		{ 444, 212, 45 },
		{ 444, 213, 45 },
		{ 444, 214, 45 },
		{ 444, 215, 45 },
		{ 444, 216, 45 },
		{ 444, 217, 45 },
		{ 444, 218, 45 },
		{ 444, 219, 45 },
		{ 444, 220, 45 },
		{ 444, 221, 45 },
		{ 444, 222, 45 },
		{ 444, 223, 45 },
		{ 444, 224, 45 },
		{ 444, 225, 45 },
		{ 444, 226, 45 },
		{ 444, 227, 45 },
		{ 444, 228, 45 },
		{ 444, 229, 45 },
		{ 444, 230, 45 },
		{ 444, 231, 45 },
		{ 444, 232, 45 },
		{ 444, 233, 45 },
		{ 444, 234, 45 },
		{ 444, 235, 45 },
		{ 444, 236, 45 },
		{ 444, 237, 45 },
		{ 444, 238, 45 },
		{ 444, 239, 45 },
		{ 444, 240, 45 },
		{ 444, 241, 45 },
		{ 444, 242, 45 },
		{ 444, 243, 45 },
		{ 444, 244, 45 },
		{ 444, 245, 45 },
		{ 444, 246, 45 },
		{ 444, 247, 45 },
		{ 444, 248, 45 },
		{ 444, 249, 45 },
		{ 444, 250, 45 },
		{ 444, 251, 45 },
		{ 444, 252, 45 },
		{ 444, 253, 45 },
		{ 444, 254, 45 },
		{ 444, 255, 45 },
		{ 444, 256, 45 },
		{ 445, 211, 45 },
		{ 445, 212, 45 },
		{ 445, 213, 45 },
		{ 445, 214, 45 },
		{ 445, 215, 45 },
		{ 445, 216, 45 },
		{ 445, 217, 45 },
		{ 445, 218, 45 },
		{ 445, 219, 45 },
		{ 445, 220, 45 },
		{ 445, 221, 45 },
		{ 445, 222, 45 },
		{ 445, 223, 45 },
		{ 445, 224, 45 },
		{ 445, 225, 45 },
		{ 445, 226, 45 },
		{ 445, 227, 45 },
		{ 445, 228, 45 },
		{ 445, 229, 45 },
		{ 445, 230, 45 },
		{ 445, 231, 45 },
		{ 445, 232, 45 },
		{ 445, 233, 45 },
		{ 445, 234, 45 },
		{ 445, 235, 45 },
		{ 445, 236, 45 },
		{ 445, 237, 45 },
		{ 445, 238, 45 },
		{ 445, 239, 45 },
		{ 445, 240, 45 },
		{ 445, 241, 45 },
		{ 445, 242, 45 },
		{ 445, 243, 45 },
		{ 445, 244, 45 },
		{ 445, 245, 45 },
		{ 445, 246, 45 },
		{ 445, 247, 45 },
		{ 445, 248, 45 },
		{ 445, 249, 45 },
		{ 445, 250, 45 },
		{ 445, 251, 45 },
		{ 445, 252, 45 },
		{ 445, 253, 45 },
		{ 445, 254, 45 },
		{ 445, 255, 45 },
		{ 445, 256, 45 },
		{ 446, 210, 45 },
		{ 446, 211, 45 },
		{ 446, 212, 45 },
		{ 446, 213, 45 },
		{ 446, 214, 45 },
		{ 446, 215, 45 },
		{ 446, 216, 45 },
		{ 446, 217, 45 },
		{ 446, 218, 45 },
		{ 446, 219, 45 },
		{ 446, 220, 45 },
		{ 446, 221, 45 },
		{ 446, 222, 45 },
		{ 446, 223, 45 },
		{ 446, 224, 45 },
		{ 446, 225, 45 },
		{ 446, 226, 45 },
		{ 446, 227, 45 },
		{ 446, 228, 45 },
		{ 446, 229, 45 },
		{ 446, 230, 45 },
		{ 446, 231, 45 },
		{ 446, 232, 45 },
		{ 446, 233, 45 },
		{ 446, 234, 45 },
		{ 446, 235, 45 },
		{ 446, 236, 45 },
		{ 446, 237, 45 },
		{ 446, 238, 45 },
		{ 446, 239, 45 },
		{ 446, 240, 45 },
		{ 446, 241, 45 },
		{ 446, 242, 45 },
		{ 446, 243, 45 },
		{ 446, 244, 45 },
		{ 446, 245, 45 },
		{ 446, 246, 45 },
		{ 446, 247, 45 },
		{ 446, 248, 45 },
		{ 446, 249, 45 },
		{ 446, 250, 45 },
		{ 446, 251, 45 },
		{ 446, 252, 45 },
		{ 446, 253, 45 },
		{ 446, 254, 45 },
		{ 446, 255, 45 },
		{ 446, 256, 45 },
		{ 447, 210, 45 },
		{ 447, 211, 45 },
		{ 447, 212, 45 },
		{ 447, 213, 45 },
		{ 447, 214, 45 },
		{ 447, 215, 45 },
		{ 447, 216, 45 },
		{ 447, 217, 45 },
		{ 447, 218, 45 },
		{ 447, 219, 45 },
		{ 447, 220, 45 },
		{ 447, 221, 45 },
		{ 447, 222, 45 },
		{ 447, 223, 45 },
		{ 447, 224, 45 },
		{ 447, 225, 45 },
		{ 447, 226, 45 },
		{ 447, 227, 45 },
		{ 447, 228, 45 },
		{ 447, 229, 45 },
		{ 447, 230, 45 },
		{ 447, 231, 45 },
		{ 447, 232, 45 },
		{ 447, 233, 45 },
		{ 447, 234, 45 },
		{ 447, 235, 45 },
		{ 447, 236, 45 },
		{ 447, 237, 45 },
		{ 447, 238, 45 },
		{ 447, 239, 45 },
		{ 447, 240, 45 },
		{ 447, 241, 45 },
		{ 447, 242, 45 },
		{ 447, 243, 45 },
		{ 447, 244, 45 },
		{ 447, 245, 45 },
		{ 447, 246, 45 },
		{ 447, 247, 45 },
		{ 447, 248, 45 },
		{ 447, 249, 45 },
		{ 447, 250, 45 },
		{ 447, 251, 45 },
		{ 447, 252, 45 },
		{ 447, 253, 45 },
		{ 447, 254, 45 },
		{ 447, 255, 45 },
		{ 447, 256, 45 },
		{ 448, 210, 45 },
		{ 448, 211, 45 },
		{ 448, 212, 45 },
		{ 448, 213, 45 },
		{ 448, 214, 45 },
		{ 448, 215, 45 },
		{ 448, 216, 45 },
		{ 448, 217, 45 },
		{ 448, 218, 45 },
		{ 448, 219, 45 },
		{ 448, 220, 45 },
		{ 448, 221, 45 },
		{ 448, 222, 45 },
		{ 448, 223, 45 },
		{ 448, 224, 45 },
		{ 448, 225, 45 },
		{ 448, 226, 45 },
		{ 448, 227, 45 },
		{ 448, 228, 45 },
		{ 448, 229, 45 },
		{ 448, 230, 45 },
		{ 448, 231, 45 },
		{ 448, 232, 45 },
		{ 448, 233, 45 },
		{ 448, 234, 45 },
		{ 448, 235, 45 },
		{ 448, 236, 45 },
		{ 448, 237, 45 },
		{ 448, 238, 45 },
		{ 448, 239, 45 },
		{ 448, 240, 45 },
		{ 448, 241, 45 },
		{ 448, 242, 45 },
		{ 448, 243, 45 },
		{ 448, 244, 45 },
		{ 448, 245, 45 },
		{ 448, 246, 45 },
		{ 448, 247, 45 },
		{ 448, 248, 45 },
		{ 448, 249, 45 },
		{ 448, 250, 45 },
		{ 448, 251, 45 },
		{ 448, 252, 45 },
		{ 448, 253, 45 },
		{ 448, 254, 45 },
		{ 448, 255, 45 },
		{ 448, 256, 45 },
		{ 449, 210, 45 },
		{ 449, 211, 45 },
		{ 449, 212, 45 },
		{ 449, 213, 45 },
		{ 449, 214, 45 },
		{ 449, 215, 45 },
		{ 449, 216, 45 },
		{ 449, 217, 45 },
		{ 449, 218, 45 },
		{ 449, 219, 45 },
		{ 449, 220, 45 },
		{ 449, 221, 45 },
		{ 449, 222, 45 },
		{ 449, 223, 45 },
		{ 449, 224, 45 },
		{ 449, 225, 45 },
		{ 449, 226, 45 },
		{ 449, 227, 45 },
		{ 449, 228, 45 },
		{ 449, 229, 45 },
		{ 449, 230, 45 },
		{ 449, 231, 45 },
		{ 449, 232, 45 },
		{ 449, 233, 45 },
		{ 449, 234, 45 },
		{ 449, 235, 45 },
		{ 449, 236, 45 },
		{ 449, 237, 45 },
		{ 449, 238, 45 },
		{ 449, 239, 45 },
		{ 449, 240, 45 },
		{ 449, 241, 45 },
		{ 449, 242, 45 },
		{ 449, 243, 45 },
		{ 449, 244, 45 },
		{ 449, 245, 45 },
		{ 449, 246, 45 },
		{ 449, 247, 45 },
		{ 449, 248, 45 },
		{ 449, 249, 45 },
		{ 449, 250, 45 },
		{ 449, 251, 45 },
		{ 449, 252, 45 },
		{ 449, 253, 45 },
		{ 449, 254, 45 },
		{ 449, 255, 45 },
		{ 449, 256, 45 },
		{ 450, 209, 45 },
		{ 450, 210, 45 },
		{ 450, 211, 45 },
		{ 450, 212, 45 },
		{ 450, 213, 45 },
		{ 450, 214, 45 },
		{ 450, 215, 45 },
		{ 450, 216, 45 },
		{ 450, 217, 45 },
		{ 450, 218, 45 },
		{ 450, 219, 45 },
		{ 450, 220, 45 },
		{ 450, 221, 45 },
		{ 450, 222, 45 },
		{ 450, 223, 45 },
		{ 450, 224, 45 },
		{ 450, 225, 45 },
		{ 450, 226, 45 },
		{ 450, 227, 45 },
		{ 450, 228, 45 },
		{ 450, 229, 45 },
		{ 450, 230, 45 },
		{ 450, 231, 45 },
		{ 450, 232, 45 },
		{ 450, 233, 45 },
		{ 450, 234, 45 },
		{ 450, 235, 45 },
		{ 450, 236, 45 },
		{ 450, 237, 45 },
		{ 450, 238, 45 },
		{ 450, 239, 45 },
		{ 450, 240, 45 },
		{ 450, 241, 45 },
		{ 450, 242, 45 },
		{ 450, 243, 45 },
		{ 450, 244, 45 },
		{ 450, 245, 45 },
		{ 450, 246, 45 },
		{ 450, 247, 45 },
		{ 450, 248, 45 },
		{ 450, 249, 45 },
		{ 450, 250, 45 },
		{ 450, 251, 45 },
		{ 450, 252, 45 },
		{ 450, 253, 45 },
		{ 450, 254, 45 },
		{ 450, 255, 45 },
		{ 450, 256, 45 },
		{ 451, 209, 45 },
		{ 451, 210, 45 },
		{ 451, 211, 45 },
		{ 451, 212, 45 },
		{ 451, 213, 45 },
		{ 451, 214, 45 },
		{ 451, 215, 45 },
		{ 451, 216, 45 },
		{ 451, 217, 45 },
		{ 451, 218, 45 },
		{ 451, 219, 45 },
		{ 451, 220, 45 },
		{ 451, 221, 45 },
		{ 451, 222, 45 },
		{ 451, 223, 45 },
		{ 451, 224, 45 },
		{ 451, 225, 45 },
		{ 451, 226, 45 },
		{ 451, 227, 45 },
		{ 451, 228, 45 },
		{ 451, 229, 45 },
		{ 451, 230, 45 },
		{ 451, 231, 45 },
		{ 451, 232, 45 },
		{ 451, 233, 45 },
		{ 451, 234, 45 },
		{ 451, 235, 45 },
		{ 451, 236, 45 },
		{ 451, 237, 45 },
		{ 451, 238, 45 },
		{ 451, 239, 45 },
		{ 451, 240, 45 },
		{ 451, 241, 45 },
		{ 451, 242, 45 },
		{ 451, 243, 45 },
		{ 451, 244, 45 },
		{ 451, 245, 45 },
		{ 451, 246, 45 },
		{ 451, 247, 45 },
		{ 451, 248, 45 },
		{ 451, 249, 45 },
		{ 451, 250, 45 },
		{ 451, 251, 45 },
		{ 451, 252, 45 },
		{ 451, 253, 45 },
		{ 451, 254, 45 },
		{ 451, 255, 45 },
		{ 451, 256, 45 },
		{ 452, 209, 45 },
		{ 452, 210, 45 },
		{ 452, 211, 45 },
		{ 452, 212, 45 },
		{ 452, 213, 45 },
		{ 452, 214, 45 },
		{ 452, 215, 45 },
		{ 452, 216, 45 },
		{ 452, 217, 45 },
		{ 452, 218, 45 },
		{ 452, 219, 45 },
		{ 452, 220, 45 },
		{ 452, 221, 45 },
		{ 452, 222, 45 },
		{ 452, 223, 45 },
		{ 452, 224, 45 },
		{ 452, 225, 45 },
		{ 452, 226, 45 },
		{ 452, 227, 45 },
		{ 452, 228, 45 },
		{ 452, 229, 45 },
		{ 452, 230, 45 },
		{ 452, 231, 45 },
		{ 452, 232, 45 },
		{ 452, 233, 45 },
		{ 452, 234, 45 },
		{ 452, 235, 45 },
		{ 452, 236, 45 },
		{ 452, 237, 45 },
		{ 452, 238, 45 },
		{ 452, 239, 45 },
		{ 452, 240, 45 },
		{ 452, 241, 45 },
		{ 452, 242, 45 },
		{ 452, 243, 45 },
		{ 452, 244, 45 },
		{ 452, 245, 45 },
		{ 452, 246, 45 },
		{ 452, 247, 45 },
		{ 452, 248, 45 },
		{ 452, 249, 45 },
		{ 452, 250, 45 },
		{ 452, 251, 45 },
		{ 452, 252, 45 },
		{ 452, 253, 45 },
		{ 452, 254, 45 },
		{ 452, 255, 45 },
		{ 452, 256, 45 },
		{ 453, 209, 45 },
		{ 453, 210, 45 },
		{ 453, 211, 45 },
		{ 453, 212, 45 },
		{ 453, 213, 45 },
		{ 453, 214, 45 },
		{ 453, 215, 45 },
		{ 453, 216, 45 },
		{ 453, 217, 45 },
		{ 453, 218, 45 },
		{ 453, 219, 45 },
		{ 453, 220, 45 },
		{ 453, 221, 45 },
		{ 453, 222, 45 },
		{ 453, 223, 45 },
		{ 453, 224, 45 },
		{ 453, 225, 45 },
		{ 453, 226, 45 },
		{ 453, 227, 45 },
		{ 453, 228, 45 },
		{ 453, 229, 45 },
		{ 453, 230, 45 },
		{ 453, 231, 45 },
		{ 453, 232, 45 },
		{ 453, 233, 45 },
		{ 453, 234, 45 },
		{ 453, 235, 45 },
		{ 453, 236, 45 },
		{ 453, 237, 45 },
		{ 453, 238, 45 },
		{ 453, 239, 45 },
		{ 453, 240, 45 },
		{ 453, 241, 45 },
		{ 453, 242, 45 },
		{ 453, 243, 45 },
		{ 453, 244, 45 },
		{ 453, 245, 45 },
		{ 453, 246, 45 },
		{ 453, 247, 45 },
		{ 453, 248, 45 },
		{ 453, 249, 45 },
		{ 453, 250, 45 },
		{ 453, 251, 45 },
		{ 453, 252, 45 },
		{ 453, 253, 45 },
		{ 453, 254, 45 },
		{ 453, 255, 45 },
		{ 453, 256, 45 },
		{ 454, 208, 45 },
		{ 454, 209, 45 },
		{ 454, 210, 45 },
		{ 454, 211, 45 },
		{ 454, 212, 45 },
		{ 454, 213, 45 },
		{ 454, 214, 45 },
		{ 454, 215, 45 },
		{ 454, 216, 45 },
		{ 454, 217, 45 },
		{ 454, 218, 45 },
		{ 454, 219, 45 },
		{ 454, 220, 45 },
		{ 454, 221, 45 },
		{ 454, 222, 45 },
		{ 454, 223, 45 },
		{ 454, 224, 45 },
		{ 454, 225, 45 },
		{ 454, 226, 45 },
		{ 454, 227, 45 },
		{ 454, 228, 45 },
		{ 454, 229, 45 },
		{ 454, 230, 45 },
		{ 454, 231, 45 },
		{ 454, 232, 45 },
		{ 454, 233, 45 },
		{ 454, 234, 45 },
		{ 454, 235, 45 },
		{ 454, 236, 45 },
		{ 454, 237, 45 },
		{ 454, 238, 45 },
		{ 454, 239, 45 },
		{ 454, 240, 45 },
		{ 454, 241, 45 },
		{ 454, 242, 45 },
		{ 454, 243, 45 },
		{ 454, 244, 45 },
		{ 454, 245, 45 },
		{ 454, 246, 45 },
		{ 454, 247, 45 },
		{ 454, 248, 45 },
		{ 454, 249, 45 },
		{ 454, 250, 45 },
		{ 454, 251, 45 },
		{ 454, 252, 45 },
		{ 454, 253, 45 },
		{ 454, 254, 45 },
		{ 454, 255, 45 },
		{ 454, 256, 45 },
		{ 455, 208, 45 },
		{ 455, 209, 45 },
		{ 455, 210, 45 },
		{ 455, 211, 45 },
		{ 455, 212, 45 },
		{ 455, 213, 45 },
		{ 455, 214, 45 },
		{ 455, 215, 45 },
		{ 455, 216, 45 },
		{ 455, 217, 45 },
		{ 455, 218, 45 },
		{ 455, 219, 45 },
		{ 455, 220, 45 },
		{ 455, 221, 45 },
		{ 455, 222, 45 },
		{ 455, 223, 45 },
		{ 455, 224, 45 },
		{ 455, 225, 45 },
		{ 455, 226, 45 },
		{ 455, 227, 45 },
		{ 455, 228, 45 },
		{ 455, 229, 45 },
		{ 455, 230, 45 },
		{ 455, 231, 45 },
		{ 455, 232, 45 },
		{ 455, 233, 45 },
		{ 455, 234, 45 },
		{ 455, 235, 45 },
		{ 455, 236, 45 },
		{ 455, 237, 45 },
		{ 455, 238, 45 },
		{ 455, 239, 45 },
		{ 455, 240, 45 },
		{ 455, 241, 45 },
		{ 455, 242, 45 },
		{ 455, 243, 45 },
		{ 455, 244, 45 },
		{ 455, 245, 45 },
		{ 455, 246, 45 },
		{ 455, 247, 45 },
		{ 455, 248, 45 },
		{ 455, 249, 45 },
		{ 455, 250, 45 },
		{ 455, 251, 45 },
		{ 455, 252, 45 },
		{ 455, 253, 45 },
		{ 455, 254, 45 },
		{ 455, 255, 45 },
		{ 455, 256, 45 },
		{ 456, 208, 45 },
		{ 456, 209, 45 },
		{ 456, 210, 45 },
		{ 456, 211, 45 },
		{ 456, 212, 45 },
		{ 456, 213, 45 },
		{ 456, 214, 45 },
		{ 456, 215, 45 },
		{ 456, 216, 45 },
		{ 456, 217, 45 },
		{ 456, 218, 45 },
		{ 456, 219, 45 },
		{ 456, 220, 45 },
		{ 456, 221, 45 },
		{ 456, 222, 45 },
		{ 456, 223, 45 },
		{ 456, 224, 45 },
		{ 456, 225, 45 },
		{ 456, 226, 45 },
		{ 456, 227, 45 },
		{ 456, 228, 45 },
		{ 456, 229, 45 },
		{ 456, 230, 45 },
		{ 456, 231, 45 },
		{ 456, 232, 45 },
		{ 456, 233, 45 },
		{ 456, 234, 45 },
		{ 456, 235, 45 },
		{ 456, 236, 45 },
		{ 456, 237, 45 },
		{ 456, 238, 45 },
		{ 456, 239, 45 },
		{ 456, 240, 45 },
		{ 456, 241, 45 },
		{ 456, 242, 45 },
		{ 456, 243, 45 },
		{ 456, 244, 45 },
		{ 456, 245, 45 },
		{ 456, 246, 45 },
		{ 456, 247, 45 },
		{ 456, 248, 45 },
		{ 456, 249, 45 },
		{ 456, 250, 45 },
		{ 456, 251, 45 },
		{ 456, 252, 45 },
		{ 456, 253, 45 },
		{ 456, 254, 45 },
		{ 456, 255, 45 },
		{ 456, 256, 45 },
		{ 457, 208, 45 },
		{ 457, 209, 45 },
		{ 457, 210, 45 },
		{ 457, 211, 45 },
		{ 457, 212, 45 },
		{ 457, 213, 45 },
		{ 457, 214, 45 },
		{ 457, 215, 45 },
		{ 457, 216, 45 },
		{ 457, 217, 45 },
		{ 457, 218, 45 },
		{ 457, 219, 45 },
		{ 457, 220, 45 },
		{ 457, 221, 45 },
		{ 457, 222, 45 },
		{ 457, 223, 45 },
		{ 457, 224, 45 },
		{ 457, 225, 45 },
		{ 457, 226, 45 },
		{ 457, 227, 45 },
		{ 457, 228, 45 },
		{ 457, 229, 45 },
		{ 457, 230, 45 },
		{ 457, 231, 45 },
		{ 457, 232, 45 },
		{ 457, 233, 45 },
		{ 457, 234, 45 },
		{ 457, 235, 45 },
		{ 457, 236, 45 },
		{ 457, 237, 45 },
		{ 457, 238, 45 },
		{ 457, 239, 45 },
		{ 457, 240, 45 },
		{ 457, 241, 45 },
		{ 457, 242, 45 },
		{ 457, 243, 45 },
		{ 457, 244, 45 },
		{ 457, 245, 45 },
		{ 457, 246, 45 },
		{ 457, 247, 45 },
		{ 457, 248, 45 },
		{ 457, 249, 45 },
		{ 457, 250, 45 },
		{ 457, 251, 45 },
		{ 457, 252, 45 },
		{ 457, 253, 45 },
		{ 457, 254, 45 },
		{ 457, 255, 45 },
		{ 457, 256, 45 },
		{ 458, 207, 45 },
		{ 458, 208, 45 },
		{ 458, 209, 45 },
		{ 458, 210, 45 },
		{ 458, 211, 45 },
		{ 458, 212, 45 },
		{ 458, 213, 45 },
		{ 458, 214, 45 },
		{ 458, 215, 45 },
		{ 458, 216, 45 },
		{ 458, 217, 45 },
		{ 458, 218, 45 },
		{ 458, 219, 45 },
		{ 458, 220, 45 },
		{ 458, 221, 45 },
		{ 458, 222, 45 },
		{ 458, 223, 45 },
		{ 458, 224, 45 },
		{ 458, 225, 45 },
		{ 458, 226, 45 },
		{ 458, 227, 45 },
		{ 458, 228, 45 },
		{ 458, 229, 45 },
		{ 458, 230, 45 },
		{ 458, 231, 45 },
		{ 458, 232, 45 },
		{ 458, 233, 45 },
		{ 458, 234, 45 },
		{ 458, 235, 45 },
		{ 458, 236, 45 },
		{ 458, 237, 45 },
		{ 458, 238, 45 },
		{ 458, 239, 45 },
		{ 458, 240, 45 },
		{ 458, 241, 45 },
		{ 458, 242, 45 },
		{ 458, 243, 45 },
		{ 458, 244, 45 },
		{ 458, 245, 45 },
		{ 458, 246, 45 },
		{ 458, 247, 45 },
		{ 458, 248, 45 },
		{ 458, 249, 45 },
		{ 458, 250, 45 },
		{ 458, 251, 45 },
		{ 458, 252, 45 },
		{ 458, 253, 45 },
		{ 458, 254, 45 },
		{ 458, 255, 45 },
		{ 458, 256, 45 },
		{ 459, 207, 45 },
		{ 459, 208, 45 },
		{ 459, 209, 45 },
		{ 459, 210, 45 },
		{ 459, 211, 45 },
		{ 459, 212, 45 },
		{ 459, 213, 45 },
		{ 459, 214, 45 },
		{ 459, 215, 45 },
		{ 459, 216, 45 },
		{ 459, 217, 45 },
		{ 459, 218, 45 },
		{ 459, 219, 45 },
		{ 459, 220, 45 },
		{ 459, 221, 45 },
		{ 459, 222, 45 },
		{ 459, 223, 45 },
		{ 459, 224, 45 },
		{ 459, 225, 45 },
		{ 459, 226, 45 },
		{ 459, 227, 45 },
		{ 459, 228, 45 },
		{ 459, 229, 45 },
		{ 459, 230, 45 },
		{ 459, 231, 45 },
		{ 459, 232, 45 },
		{ 459, 233, 45 },
		{ 459, 234, 45 },
		{ 459, 235, 45 },
		{ 459, 236, 45 },
		{ 459, 237, 45 },
		{ 459, 238, 45 },
		{ 459, 239, 45 },
		{ 459, 240, 45 },
		{ 459, 241, 45 },
		{ 459, 242, 45 },
		{ 459, 243, 45 },
		{ 459, 244, 45 },
		{ 459, 245, 45 },
		{ 459, 246, 45 },
		{ 459, 247, 45 },
		{ 459, 248, 45 },
		{ 459, 249, 45 },
		{ 459, 250, 45 },
		{ 459, 251, 45 },
		{ 459, 252, 45 },
		{ 459, 253, 45 },
		{ 459, 254, 45 },
		{ 459, 255, 45 },
		{ 459, 256, 45 },
		{ 460, 207, 45 },
		{ 460, 208, 45 },
		{ 460, 209, 45 },
		{ 460, 210, 45 },
		{ 460, 211, 45 },
		{ 460, 212, 45 },
		{ 460, 213, 45 },
		{ 460, 214, 45 },
		{ 460, 215, 45 },
		{ 460, 216, 45 },
		{ 460, 217, 45 },
		{ 460, 218, 45 },
		{ 460, 219, 45 },
		{ 460, 220, 45 },
		{ 460, 221, 45 },
		{ 460, 222, 45 },
		{ 460, 223, 45 },
		{ 460, 224, 45 },
		{ 460, 225, 45 },
		{ 460, 226, 45 },
		{ 460, 227, 45 },
		{ 460, 228, 45 },
		{ 460, 229, 45 },
		{ 460, 230, 45 },
		{ 460, 231, 45 },
		{ 460, 232, 45 },
		{ 460, 233, 45 },
		{ 460, 234, 45 },
		{ 460, 235, 45 },
		{ 460, 236, 45 },
		{ 460, 237, 45 },
		{ 460, 238, 45 },
		{ 460, 239, 45 },
		{ 460, 240, 45 },
		{ 460, 241, 45 },
		{ 460, 242, 45 },
		{ 460, 243, 45 },
		{ 460, 244, 45 },
		{ 460, 245, 45 },
		{ 460, 246, 45 },
		{ 460, 247, 45 },
		{ 460, 248, 45 },
		{ 460, 249, 45 },
		{ 460, 250, 45 },
		{ 460, 251, 45 },
		{ 460, 252, 45 },
		{ 460, 253, 45 },
		{ 460, 254, 45 },
		{ 460, 255, 45 },
		{ 460, 256, 45 },
		{ 461, 207, 45 },
		{ 461, 208, 45 },
		{ 461, 209, 45 },
		{ 461, 210, 45 },
		{ 461, 211, 45 },
		{ 461, 212, 45 },
		{ 461, 213, 45 },
		{ 461, 214, 45 },
		{ 461, 215, 45 },
		{ 461, 216, 45 },
		{ 461, 217, 45 },
		{ 461, 218, 45 },
		{ 461, 219, 45 },
		{ 461, 220, 45 },
		{ 461, 221, 45 },
		{ 461, 222, 45 },
		{ 461, 223, 45 },
		{ 461, 224, 45 },
		{ 461, 225, 45 },
		{ 461, 226, 45 },
		{ 461, 227, 45 },
		{ 461, 228, 45 },
		{ 461, 229, 45 },
		{ 461, 230, 45 },
		{ 461, 231, 45 },
		{ 461, 232, 45 },
		{ 461, 233, 45 },
		{ 461, 234, 45 },
		{ 461, 235, 45 },
		{ 461, 236, 45 },
		{ 461, 237, 45 },
		{ 461, 238, 45 },
		{ 461, 239, 45 },
		{ 461, 240, 45 },
		{ 461, 241, 45 },
		{ 461, 242, 45 },
		{ 461, 243, 45 },
		{ 461, 244, 45 },
		{ 461, 245, 45 },
		{ 461, 246, 45 },
		{ 461, 247, 45 },
		{ 461, 248, 45 },
		{ 461, 249, 45 },
		{ 461, 250, 45 },
		{ 461, 251, 45 },
		{ 461, 252, 45 },
		{ 461, 253, 45 },
		{ 461, 254, 45 },
		{ 461, 255, 45 },
		{ 461, 256, 45 },
		{ 462, 206, 45 },
		{ 462, 207, 45 },
		{ 462, 208, 45 },
		{ 462, 209, 45 },
		{ 462, 210, 45 },
		{ 462, 211, 45 },
		{ 462, 212, 45 },
		{ 462, 213, 45 },
		{ 462, 214, 45 },
		{ 462, 215, 45 },
		{ 462, 216, 45 },
		{ 462, 217, 45 },
		{ 462, 218, 45 },
		{ 462, 219, 45 },
		{ 462, 220, 45 },
		{ 462, 221, 45 },
		{ 462, 222, 45 },
		{ 462, 223, 45 },
		{ 462, 224, 45 },
		{ 462, 225, 45 },
		{ 462, 226, 45 },
		{ 462, 227, 45 },
		{ 462, 228, 45 },
		{ 462, 229, 45 },
		{ 462, 230, 45 },
		{ 462, 231, 45 },
		{ 462, 232, 45 },
		{ 462, 233, 45 },
		{ 462, 234, 45 },
		{ 462, 235, 45 },
		{ 462, 236, 45 },
		{ 462, 237, 45 },
		{ 462, 238, 45 },
		{ 462, 239, 45 },
		{ 462, 240, 45 },
		{ 462, 241, 45 },
		{ 462, 242, 45 },
		{ 462, 243, 45 },
		{ 462, 244, 45 },
		{ 462, 245, 45 },
		{ 462, 246, 45 },
		{ 462, 247, 45 },
		{ 462, 248, 45 },
		{ 462, 249, 45 },
		{ 462, 250, 45 },
		{ 462, 251, 45 },
		{ 462, 252, 45 },
		{ 462, 253, 45 },
		{ 462, 254, 45 },
		{ 462, 255, 45 },
		{ 462, 256, 45 },
		{ 463, 206, 45 },
		{ 463, 207, 45 },
		{ 463, 208, 45 },
		{ 463, 209, 45 },
		{ 463, 210, 45 },
		{ 463, 211, 45 },
		{ 463, 212, 45 },
		{ 463, 213, 45 },
		{ 463, 214, 45 },
		{ 463, 215, 45 },
		{ 463, 216, 45 },
		{ 463, 217, 45 },
		{ 463, 218, 45 },
		{ 463, 219, 45 },
		{ 463, 220, 45 },
		{ 463, 221, 45 },
		{ 463, 222, 45 },
		{ 463, 223, 45 },
		{ 463, 224, 45 },
		{ 463, 225, 45 },
		{ 463, 226, 45 },
		{ 463, 227, 45 },
		{ 463, 228, 45 },
		{ 463, 229, 45 },
		{ 463, 230, 45 },
		{ 463, 231, 45 },
		{ 463, 232, 45 },
		{ 463, 233, 45 },
		{ 463, 234, 45 },
		{ 463, 235, 45 },
		{ 463, 236, 45 },
		{ 463, 237, 45 },
		{ 463, 238, 45 },
		{ 463, 239, 45 },
		{ 463, 240, 45 },
		{ 463, 241, 45 },
		{ 463, 242, 45 },
		{ 463, 243, 45 },
		{ 463, 244, 45 },
		{ 463, 245, 45 },
		{ 463, 246, 45 },
		{ 463, 247, 45 },
		{ 463, 248, 45 },
		{ 463, 249, 45 },
		{ 463, 250, 45 },
		{ 463, 251, 45 },
		{ 463, 252, 45 },
		{ 463, 253, 45 },
		{ 463, 254, 45 },
		{ 463, 255, 45 },
		{ 463, 256, 45 },
		{ 464, 206, 45 },
		{ 464, 207, 45 },
		{ 464, 208, 45 },
		{ 464, 209, 45 },
		{ 464, 210, 45 },
		{ 464, 211, 45 },
		{ 464, 212, 45 },
		{ 464, 213, 45 },
		{ 464, 214, 45 },
		{ 464, 215, 45 },
		{ 464, 216, 45 },
		{ 464, 217, 45 },
		{ 464, 218, 45 },
		{ 464, 219, 45 },
		{ 464, 220, 45 },
		{ 464, 221, 45 },
		{ 464, 222, 45 },
		{ 464, 223, 45 },
		{ 464, 224, 45 },
		{ 464, 225, 45 },
		{ 464, 226, 45 },
		{ 464, 227, 45 },
		{ 464, 228, 45 },
		{ 464, 229, 45 },
		{ 464, 230, 45 },
		{ 464, 231, 45 },
		{ 464, 232, 45 },
		{ 464, 233, 45 },
		{ 464, 234, 45 },
		{ 464, 235, 45 },
		{ 464, 236, 45 },
		{ 464, 237, 45 },
		{ 464, 238, 45 },
		{ 464, 239, 45 },
		{ 464, 240, 45 },
		{ 464, 241, 45 },
		{ 464, 242, 45 },
		{ 464, 243, 45 },
		{ 464, 244, 45 },
		{ 464, 245, 45 },
		{ 464, 246, 45 },
		{ 464, 247, 45 },
		{ 464, 248, 45 },
		{ 464, 249, 45 },
		{ 464, 250, 45 },
		{ 464, 251, 45 },
		{ 464, 252, 45 },
		{ 464, 253, 45 },
		{ 464, 254, 45 },
		{ 464, 255, 45 },
		{ 464, 256, 45 },
		{ 465, 206, 45 },
		{ 465, 207, 45 },
		{ 465, 208, 45 },
		{ 465, 209, 45 },
		{ 465, 210, 45 },
		{ 465, 211, 45 },
		{ 465, 212, 45 },
		{ 465, 213, 45 },
		{ 465, 214, 45 },
		{ 465, 215, 45 },
		{ 465, 216, 45 },
		{ 465, 217, 45 },
		{ 465, 218, 45 },
		{ 465, 219, 45 },
		{ 465, 220, 45 },
		{ 465, 221, 45 },
		{ 465, 222, 45 },
		{ 465, 223, 45 },
		{ 465, 224, 45 },
		{ 465, 225, 45 },
		{ 465, 226, 45 },
		{ 465, 227, 45 },
		{ 465, 228, 45 },
		{ 465, 229, 45 },
		{ 465, 230, 45 },
		{ 465, 231, 45 },
		{ 465, 232, 45 },
		{ 465, 233, 45 },
		{ 465, 234, 45 },
		{ 465, 235, 45 },
		{ 465, 236, 45 },
		{ 465, 237, 45 },
		{ 465, 238, 45 },
		{ 465, 239, 45 },
		{ 465, 240, 45 },
		{ 465, 241, 45 },
		{ 465, 242, 45 },
		{ 465, 243, 45 },
		{ 465, 244, 45 },
		{ 465, 245, 45 },
		{ 465, 246, 45 },
		{ 465, 247, 45 },
		{ 465, 248, 45 },
		{ 465, 249, 45 },
		{ 465, 250, 45 },
		{ 465, 251, 45 },
		{ 465, 252, 45 },
		{ 465, 253, 45 },
		{ 465, 254, 45 },
		{ 465, 255, 45 },
		{ 465, 256, 45 },
		{ 466, 205, 45 },
		{ 466, 206, 45 },
		{ 466, 207, 45 },
		{ 466, 208, 45 },
		{ 466, 209, 45 },
		{ 466, 210, 45 },
		{ 466, 211, 45 },
		{ 466, 212, 45 },
		{ 466, 213, 45 },
		{ 466, 214, 45 },
		{ 466, 215, 45 },
		{ 466, 216, 45 },
		{ 466, 217, 45 },
		{ 466, 218, 45 },
		{ 466, 219, 45 },
		{ 466, 220, 45 },
		{ 466, 221, 45 },
		{ 466, 222, 45 },
		{ 466, 223, 45 },
		{ 466, 224, 45 },
		{ 466, 225, 45 },
		{ 466, 226, 45 },
		{ 466, 227, 45 },
		{ 466, 228, 45 },
		{ 466, 229, 45 },
		{ 466, 230, 45 },
		{ 466, 231, 45 },
		{ 466, 232, 45 },
		{ 466, 233, 45 },
		{ 466, 234, 45 },
		{ 466, 235, 45 },
		{ 466, 236, 45 },
		{ 466, 237, 45 },
		{ 466, 238, 45 },
		{ 466, 239, 45 },
		{ 466, 240, 45 },
		{ 466, 241, 45 },
		{ 466, 242, 45 },
		{ 466, 243, 45 },
		{ 466, 244, 45 },
		{ 466, 245, 45 },
		{ 466, 246, 45 },
		{ 466, 247, 45 },
		{ 466, 248, 45 },
		{ 466, 249, 45 },
		{ 466, 250, 45 },
		{ 466, 251, 45 },
		{ 466, 252, 45 },
		{ 466, 253, 45 },
		{ 466, 254, 45 },
		{ 466, 255, 45 },
		{ 466, 256, 45 },
		{ 467, 205, 45 },
		{ 467, 206, 45 },
		{ 467, 207, 45 },
		{ 467, 208, 45 },
		{ 467, 209, 45 },
		{ 467, 210, 45 },
		{ 467, 211, 45 },
		{ 467, 212, 45 },
		{ 467, 213, 45 },
		{ 467, 214, 45 },
		{ 467, 215, 45 },
		{ 467, 216, 45 },
		{ 467, 217, 45 },
		{ 467, 218, 45 },
		{ 467, 219, 45 },
		{ 467, 220, 45 },
		{ 467, 221, 45 },
		{ 467, 222, 45 },
		{ 467, 223, 45 },
		{ 467, 224, 45 },
		{ 467, 225, 45 },
		{ 467, 226, 45 },
		{ 467, 227, 45 },
		{ 467, 228, 45 },
		{ 467, 229, 45 },
		{ 467, 230, 45 },
		{ 467, 231, 45 },
		{ 467, 232, 45 },
		{ 467, 233, 45 },
		{ 467, 234, 45 },
		{ 467, 235, 45 },
		{ 467, 236, 45 },
		{ 467, 237, 45 },
		{ 467, 238, 45 },
		{ 467, 239, 45 },
		{ 467, 240, 45 },
		{ 467, 241, 45 },
		{ 467, 242, 45 },
		{ 467, 243, 45 },
		{ 467, 244, 45 },
		{ 467, 245, 45 },
		{ 467, 246, 45 },
		{ 467, 247, 45 },
		{ 467, 248, 45 },
		{ 467, 249, 45 },
		{ 467, 250, 45 },
		{ 467, 251, 45 },
		{ 467, 252, 45 },
		{ 467, 253, 45 },
		{ 467, 254, 45 },
		{ 467, 255, 45 },
		{ 467, 256, 45 },
		{ 468, 205, 45 },
		{ 468, 206, 45 },
		{ 468, 207, 45 },
		{ 468, 208, 45 },
		{ 468, 209, 45 },
		{ 468, 210, 45 },
		{ 468, 211, 45 },
		{ 468, 212, 45 },
		{ 468, 213, 45 },
		{ 468, 214, 45 },
		{ 468, 215, 45 },
		{ 468, 216, 45 },
		{ 468, 217, 45 },
		{ 468, 218, 45 },
		{ 468, 219, 45 },
		{ 468, 220, 45 },
		{ 468, 221, 45 },
		{ 468, 222, 45 },
		{ 468, 223, 45 },
		{ 468, 224, 45 },
		{ 468, 225, 45 },
		{ 468, 226, 45 },
		{ 468, 227, 45 },
		{ 468, 228, 45 },
		{ 468, 229, 45 },
		{ 468, 230, 45 },
		{ 468, 231, 45 },
		{ 468, 232, 45 },
		{ 468, 233, 45 },
		{ 468, 234, 45 },
		{ 468, 235, 45 },
		{ 468, 236, 45 },
		{ 468, 237, 45 },
		{ 468, 238, 45 },
		{ 468, 239, 45 },
		{ 468, 240, 45 },
		{ 468, 241, 45 },
		{ 468, 242, 45 },
		{ 468, 243, 45 },
		{ 468, 244, 45 },
		{ 468, 245, 45 },
		{ 468, 246, 45 },
		{ 468, 247, 45 },
		{ 468, 248, 45 },
		{ 468, 249, 45 },
		{ 468, 250, 45 },
		{ 468, 251, 45 },
		{ 468, 252, 45 },
		{ 468, 253, 45 },
		{ 468, 254, 45 },
		{ 468, 255, 45 },
		{ 468, 256, 45 },
		{ 469, 205, 45 },
		{ 469, 206, 45 },
		{ 469, 207, 45 },
		{ 469, 208, 45 },
		{ 469, 209, 45 },
		{ 469, 210, 45 },
		{ 469, 211, 45 },
		{ 469, 212, 45 },
		{ 469, 213, 45 },
		{ 469, 214, 45 },
		{ 469, 215, 45 },
		{ 469, 216, 45 },
		{ 469, 217, 45 },
		{ 469, 218, 45 },
		{ 469, 219, 45 },
		{ 469, 220, 45 },
		{ 469, 221, 45 },
		{ 469, 222, 45 },
		{ 469, 223, 45 },
		{ 469, 224, 45 },
		{ 469, 225, 45 },
		{ 469, 226, 45 },
		{ 469, 227, 45 },
		{ 469, 228, 45 },
		{ 469, 229, 45 },
		{ 469, 230, 45 },
		{ 469, 231, 45 },
		{ 469, 232, 45 },
		{ 469, 233, 45 },
		{ 469, 234, 45 },
		{ 469, 235, 45 },
		{ 469, 236, 45 },
		{ 469, 237, 45 },
		{ 469, 238, 45 },
		{ 469, 239, 45 },
		{ 469, 240, 45 },
		{ 469, 241, 45 },
		{ 469, 242, 45 },
		{ 469, 243, 45 },
		{ 469, 244, 45 },
		{ 469, 245, 45 },
		{ 469, 246, 45 },
		{ 469, 247, 45 },
		{ 469, 248, 45 },
		{ 469, 249, 45 },
		{ 469, 250, 45 },
		{ 469, 251, 45 },
		{ 469, 252, 45 },
		{ 469, 253, 45 },
		{ 469, 254, 45 },
		{ 469, 255, 45 },
		{ 469, 256, 45 },
		{ 470, 204, 45 },
		{ 470, 205, 45 },
		{ 470, 206, 45 },
		{ 470, 207, 45 },
		{ 470, 208, 45 },
		{ 470, 209, 45 },
		{ 470, 210, 45 },
		{ 470, 211, 45 },
		{ 470, 212, 45 },
		{ 470, 213, 45 },
		{ 470, 214, 45 },
		{ 470, 215, 45 },
		{ 470, 216, 45 },
		{ 470, 217, 45 },
		{ 470, 218, 45 },
		{ 470, 219, 45 },
		{ 470, 220, 45 },
		{ 470, 221, 45 },
		{ 470, 222, 45 },
		{ 470, 223, 45 },
		{ 470, 224, 45 },
		{ 470, 225, 45 },
		{ 470, 226, 45 },
		{ 470, 227, 45 },
		{ 470, 228, 45 },
		{ 470, 229, 45 },
		{ 470, 230, 45 },
		{ 470, 231, 45 },
		{ 470, 232, 45 },
		{ 470, 233, 45 },
		{ 470, 234, 45 },
		{ 470, 235, 45 },
		{ 470, 236, 45 },
		{ 470, 237, 45 },
		{ 470, 238, 45 },
		{ 470, 239, 45 },
		{ 470, 240, 45 },
		{ 470, 241, 45 },
		{ 470, 242, 45 },
		{ 470, 243, 45 },
		{ 470, 244, 45 },
		{ 470, 245, 45 },
		{ 470, 246, 45 },
		{ 470, 247, 45 },
		{ 470, 248, 45 },
		{ 470, 249, 45 },
		{ 470, 250, 45 },
		{ 470, 251, 45 },
		{ 470, 252, 45 },
		{ 470, 253, 45 },
		{ 470, 254, 45 },
		{ 470, 255, 45 },
		{ 470, 256, 45 },
		{ 471, 204, 45 },
		{ 471, 205, 45 },
		{ 471, 206, 45 },
		{ 471, 207, 45 },
		{ 471, 208, 45 },
		{ 471, 209, 45 },
		{ 471, 210, 45 },
		{ 471, 211, 45 },
		{ 471, 212, 45 },
		{ 471, 213, 45 },
		{ 471, 214, 45 },
		{ 471, 215, 45 },
		{ 471, 216, 45 },
		{ 471, 217, 45 },
		{ 471, 218, 45 },
		{ 471, 219, 45 },
		{ 471, 220, 45 },
		{ 471, 221, 45 },
		{ 471, 222, 45 },
		{ 471, 223, 45 },
		{ 471, 224, 45 },
		{ 471, 225, 45 },
		{ 471, 226, 45 },
		{ 471, 227, 45 },
		{ 471, 228, 45 },
		{ 471, 229, 45 },
		{ 471, 230, 45 },
		{ 471, 231, 45 },
		{ 471, 232, 45 },
		{ 471, 233, 45 },
		{ 471, 234, 45 },
		{ 471, 235, 45 },
		{ 471, 236, 45 },
		{ 471, 237, 45 },
		{ 471, 238, 45 },
		{ 471, 239, 45 },
		{ 471, 240, 45 },
		{ 471, 241, 45 },
		{ 471, 242, 45 },
		{ 471, 243, 45 },
		{ 471, 244, 45 },
		{ 471, 245, 45 },
		{ 471, 246, 45 },
		{ 471, 247, 45 },
		{ 471, 248, 45 },
		{ 471, 249, 45 },
		{ 471, 250, 45 },
		{ 471, 251, 45 },
		{ 471, 252, 45 },
		{ 471, 253, 45 },
		{ 471, 254, 45 },
		{ 471, 255, 45 },
		{ 471, 256, 45 },
		{ 472, 204, 45 },
		{ 472, 205, 45 },
		{ 472, 206, 45 },
		{ 472, 207, 45 },
		{ 472, 208, 45 },
		{ 472, 209, 45 },
		{ 472, 210, 45 },
		{ 472, 211, 45 },
		{ 472, 212, 45 },
		{ 472, 213, 45 },
		{ 472, 214, 45 },
		{ 472, 215, 45 },
		{ 472, 216, 45 },
		{ 472, 217, 45 },
		{ 472, 218, 45 },
		{ 472, 219, 45 },
		{ 472, 220, 45 },
		{ 472, 221, 45 },
		{ 472, 222, 45 },
		{ 472, 223, 45 },
		{ 472, 224, 45 },
		{ 472, 225, 45 },
		{ 472, 226, 45 },
		{ 472, 227, 45 },
		{ 472, 228, 45 },
		{ 472, 229, 45 },
		{ 472, 230, 45 },
		{ 472, 231, 45 },
		{ 472, 232, 45 },
		{ 472, 233, 45 },
		{ 472, 234, 45 },
		{ 472, 235, 45 },
		{ 472, 236, 45 },
		{ 472, 237, 45 },
		{ 472, 238, 45 },
		{ 472, 239, 45 },
		{ 472, 240, 45 },
		{ 472, 241, 45 },
		{ 472, 242, 45 },
		{ 472, 243, 45 },
		{ 472, 244, 45 },
		{ 472, 245, 45 },
		{ 472, 246, 45 },
		{ 472, 247, 45 },
		{ 472, 248, 45 },
		{ 472, 249, 45 },
		{ 472, 250, 45 },
		{ 472, 251, 45 },
		{ 472, 252, 45 },
		{ 472, 253, 45 },
		{ 472, 254, 45 },
		{ 472, 255, 45 },
		{ 472, 256, 45 },
		{ 473, 204, 45 },
		{ 473, 205, 45 },
		{ 473, 206, 45 },
		{ 473, 207, 45 },
		{ 473, 208, 45 },
		{ 473, 209, 45 },
		{ 473, 210, 45 },
		{ 473, 211, 45 },
		{ 473, 212, 45 },
		{ 473, 213, 45 },
		{ 473, 214, 45 },
		{ 473, 215, 45 },
		{ 473, 216, 45 },
		{ 473, 217, 45 },
		{ 473, 218, 45 },
		{ 473, 219, 45 },
		{ 473, 220, 45 },
		{ 473, 221, 45 },
		{ 473, 222, 45 },
		{ 473, 223, 45 },
		{ 473, 224, 45 },
		{ 473, 225, 45 },
		{ 473, 226, 45 },
		{ 473, 227, 45 },
		{ 473, 228, 45 },
		{ 473, 229, 45 },
		{ 473, 230, 45 },
		{ 473, 231, 45 },
		{ 473, 232, 45 },
		{ 473, 233, 45 },
		{ 473, 234, 45 },
		{ 473, 235, 45 },
		{ 473, 236, 45 },
		{ 473, 237, 45 },
		{ 473, 238, 45 },
		{ 473, 239, 45 },
		{ 473, 240, 45 },
		{ 473, 241, 45 },
		{ 473, 242, 45 },
		{ 473, 243, 45 },
		{ 473, 244, 45 },
		{ 473, 245, 45 },
		{ 473, 246, 45 },
		{ 473, 247, 45 },
		{ 473, 248, 45 },
		{ 473, 249, 45 },
		{ 473, 250, 45 },
		{ 473, 251, 45 },
		{ 473, 252, 45 },
		{ 473, 253, 45 },
		{ 473, 254, 45 },
		{ 473, 255, 45 },
		{ 473, 256, 45 },
		{ 474, 203, 45 },
		{ 474, 204, 45 },
		{ 474, 205, 45 },
		{ 474, 206, 45 },
		{ 474, 207, 45 },
		{ 474, 208, 45 },
		{ 474, 209, 45 },
		{ 474, 210, 45 },
		{ 474, 211, 45 },
		{ 474, 212, 45 },
		{ 474, 213, 45 },
		{ 474, 214, 45 },
		{ 474, 215, 45 },
		{ 474, 216, 45 },
		{ 474, 217, 45 },
		{ 474, 218, 45 },
		{ 474, 219, 45 },
		{ 474, 220, 45 },
		{ 474, 221, 45 },
		{ 474, 222, 45 },
		{ 474, 223, 45 },
		{ 474, 224, 45 },
		{ 474, 225, 45 },
		{ 474, 226, 45 },
		{ 474, 227, 45 },
		{ 474, 228, 45 },
		{ 474, 229, 45 },
		{ 474, 230, 45 },
		{ 474, 231, 45 },
		{ 474, 232, 45 },
		{ 474, 233, 45 },
		{ 474, 234, 45 },
		{ 474, 235, 45 },
		{ 474, 236, 45 },
		{ 474, 237, 45 },
		{ 474, 238, 45 },
		{ 474, 239, 45 },
		{ 474, 240, 45 },
		{ 474, 241, 45 },
		{ 474, 242, 45 },
		{ 474, 243, 45 },
		{ 474, 244, 45 },
		{ 474, 245, 45 },
		{ 474, 246, 45 },
		{ 474, 247, 45 },
		{ 474, 248, 45 },
		{ 474, 249, 45 },
		{ 474, 250, 45 },
		{ 474, 251, 45 },
		{ 474, 252, 45 },
		{ 474, 253, 45 },
		{ 474, 254, 45 },
		{ 474, 255, 45 },
		{ 474, 256, 45 },
		{ 475, 203, 45 },
		{ 475, 204, 45 },
		{ 475, 205, 45 },
		{ 475, 206, 45 },
		{ 475, 207, 45 },
		{ 475, 208, 45 },
		{ 475, 209, 45 },
		{ 475, 210, 45 },
		{ 475, 211, 45 },
		{ 475, 212, 45 },
		{ 475, 213, 45 },
		{ 475, 214, 45 },
		{ 475, 215, 45 },
		{ 475, 216, 45 },
		{ 475, 217, 45 },
		{ 475, 218, 45 },
		{ 475, 219, 45 },
		{ 475, 220, 45 },
		{ 475, 221, 45 },
		{ 475, 222, 45 },
		{ 475, 223, 45 },
		{ 475, 224, 45 },
		{ 475, 225, 45 },
		{ 475, 226, 45 },
		{ 475, 227, 45 },
		{ 475, 228, 45 },
		{ 475, 229, 45 },
		{ 475, 230, 45 },
		{ 475, 231, 45 },
		{ 475, 232, 45 },
		{ 475, 233, 45 },
		{ 475, 234, 45 },
		{ 475, 235, 45 },
		{ 475, 236, 45 },
		{ 475, 237, 45 },
		{ 475, 238, 45 },
		{ 475, 239, 45 },
		{ 475, 240, 45 },
		{ 475, 241, 45 },
		{ 475, 242, 45 },
		{ 475, 243, 45 },
		{ 475, 244, 45 },
		{ 475, 245, 45 },
		{ 475, 246, 45 },
		{ 475, 247, 45 },
		{ 475, 248, 45 },
		{ 475, 249, 45 },
		{ 475, 250, 45 },
		{ 475, 251, 45 },
		{ 475, 252, 45 },
		{ 475, 253, 45 },
		{ 475, 254, 45 },
		{ 475, 255, 45 },
		{ 475, 256, 45 },
		{ 476, 203, 45 },
		{ 476, 204, 45 },
		{ 476, 205, 45 },
		{ 476, 206, 45 },
		{ 476, 207, 45 },
		{ 476, 208, 45 },
		{ 476, 209, 45 },
		{ 476, 210, 45 },
		{ 476, 211, 45 },
		{ 476, 212, 45 },
		{ 476, 213, 45 },
		{ 476, 214, 45 },
		{ 476, 215, 45 },
		{ 476, 216, 45 },
		{ 476, 217, 45 },
		{ 476, 218, 45 },
		{ 476, 219, 45 },
		{ 476, 220, 45 },
		{ 476, 221, 45 },
		{ 476, 222, 45 },
		{ 476, 223, 45 },
		{ 476, 224, 45 },
		{ 476, 225, 45 },
		{ 476, 226, 45 },
		{ 476, 227, 45 },
		{ 476, 228, 45 },
		{ 476, 229, 45 },
		{ 476, 230, 45 },
		{ 476, 231, 45 },
		{ 476, 232, 45 },
		{ 476, 233, 45 },
		{ 476, 234, 45 },
		{ 476, 235, 45 },
		{ 476, 236, 45 },
		{ 476, 237, 45 },
		{ 476, 238, 45 },
		{ 476, 239, 45 },
		{ 476, 240, 45 },
		{ 476, 241, 45 },
		{ 476, 242, 45 },
		{ 476, 243, 45 },
		{ 476, 244, 45 },
		{ 476, 245, 45 },
		{ 476, 246, 45 },
		{ 476, 247, 45 },
		{ 476, 248, 45 },
		{ 476, 249, 45 },
		{ 476, 250, 45 },
		{ 476, 251, 45 },
		{ 476, 252, 45 },
		{ 476, 253, 45 },
		{ 476, 254, 45 },
		{ 476, 255, 45 },
		{ 476, 256, 45 },
		{ 477, 203, 45 },
		{ 477, 204, 45 },
		{ 477, 205, 45 },
		{ 477, 206, 45 },
		{ 477, 207, 45 },
		{ 477, 208, 45 },
		{ 477, 209, 45 },
		{ 477, 210, 45 },
		{ 477, 211, 45 },
		{ 477, 212, 45 },
		{ 477, 213, 45 },
		{ 477, 214, 45 },
		{ 477, 215, 45 },
		{ 477, 216, 45 },
		{ 477, 217, 45 },
		{ 477, 218, 45 },
		{ 477, 219, 45 },
		{ 477, 220, 45 },
		{ 477, 221, 45 },
		{ 477, 222, 45 },
		{ 477, 223, 45 },
		{ 477, 224, 45 },
		{ 477, 225, 45 },
		{ 477, 226, 45 },
		{ 477, 227, 45 },
		{ 477, 228, 45 },
		{ 477, 229, 45 },
		{ 477, 230, 45 },
		{ 477, 231, 45 },
		{ 477, 232, 45 },
		{ 477, 233, 45 },
		{ 477, 234, 45 },
		{ 477, 235, 45 },
		{ 477, 236, 45 },
		{ 477, 237, 45 },
		{ 477, 238, 45 },
		{ 477, 239, 45 },
		{ 477, 240, 45 },
		{ 477, 241, 45 },
		{ 477, 242, 45 },
		{ 477, 243, 45 },
		{ 477, 244, 45 },
		{ 477, 245, 45 },
		{ 477, 246, 45 },
		{ 477, 247, 45 },
		{ 477, 248, 45 },
		{ 477, 249, 45 },
		{ 477, 250, 45 },
		{ 477, 251, 45 },
		{ 477, 252, 45 },
		{ 477, 253, 45 },
		{ 477, 254, 45 },
		{ 477, 255, 45 },
		{ 477, 256, 45 },
		{ 478, 202, 45 },
		{ 478, 203, 45 },
		{ 478, 204, 45 },
		{ 478, 205, 45 },
		{ 478, 206, 45 },
		{ 478, 207, 45 },
		{ 478, 208, 45 },
		{ 478, 209, 45 },
		{ 478, 210, 45 },
		{ 478, 211, 45 },
		{ 478, 212, 45 },
		{ 478, 213, 45 },
		{ 478, 214, 45 },
		{ 478, 215, 45 },
		{ 478, 216, 45 },
		{ 478, 217, 45 },
		{ 478, 218, 45 },
		{ 478, 219, 45 },
		{ 478, 220, 45 },
		{ 478, 221, 45 },
		{ 478, 222, 45 },
		{ 478, 223, 45 },
		{ 478, 224, 45 },
		{ 478, 225, 45 },
		{ 478, 226, 45 },
		{ 478, 227, 45 },
		{ 478, 228, 45 },
		{ 478, 229, 45 },
		{ 478, 230, 45 },
		{ 478, 231, 45 },
		{ 478, 232, 45 },
		{ 478, 233, 45 },
		{ 478, 234, 45 },
		{ 478, 235, 45 },
		{ 478, 236, 45 },
		{ 478, 237, 45 },
		{ 478, 238, 45 },
		{ 478, 239, 45 },
		{ 478, 240, 45 },
		{ 478, 241, 45 },
		{ 478, 242, 45 },
		{ 478, 243, 45 },
		{ 478, 244, 45 },
		{ 478, 245, 45 },
		{ 478, 246, 45 },
		{ 478, 247, 45 },
		{ 478, 248, 45 },
		{ 478, 249, 45 },
		{ 478, 250, 45 },
		{ 478, 251, 45 },
		{ 478, 252, 45 },
		{ 478, 253, 45 },
		{ 478, 254, 45 },
		{ 478, 255, 45 },
		{ 478, 256, 45 },
		{ 479, 202, 45 },
		{ 479, 203, 45 },
		{ 479, 204, 45 },
		{ 479, 205, 45 },
		{ 479, 206, 45 },
		{ 479, 207, 45 },
		{ 479, 208, 45 },
		{ 479, 209, 45 },
		{ 479, 210, 45 },
		{ 479, 211, 45 },
		{ 479, 212, 45 },
		{ 479, 213, 45 },
		{ 479, 214, 45 },
		{ 479, 215, 45 },
		{ 479, 216, 45 },
		{ 479, 217, 45 },
		{ 479, 218, 45 },
		{ 479, 219, 45 },
		{ 479, 220, 45 },
		{ 479, 221, 45 },
		{ 479, 222, 45 },
		{ 479, 223, 45 },
		{ 479, 224, 45 },
		{ 479, 225, 45 },
		{ 479, 226, 45 },
		{ 479, 227, 45 },
		{ 479, 228, 45 },
		{ 479, 229, 45 },
		{ 479, 230, 45 },
		{ 479, 231, 45 },
		{ 479, 232, 45 },
		{ 479, 233, 45 },
		{ 479, 234, 45 },
		{ 479, 235, 45 },
		{ 479, 236, 45 },
		{ 479, 237, 45 },
		{ 479, 238, 45 },
		{ 479, 239, 45 },
		{ 479, 240, 45 },
		{ 479, 241, 45 },
		{ 479, 242, 45 },
		{ 479, 243, 45 },
		{ 479, 244, 45 },
		{ 479, 245, 45 },
		{ 479, 246, 45 },
		{ 479, 247, 45 },
		{ 479, 248, 45 },
		{ 479, 249, 45 },
		{ 479, 250, 45 },
		{ 479, 251, 45 },
		{ 479, 252, 45 },
		{ 479, 253, 45 },
		{ 479, 254, 45 },
		{ 479, 255, 45 },
		{ 479, 256, 45 },
		{ 480, 202, 45 },
		{ 480, 203, 45 },
		{ 480, 204, 45 },
		{ 480, 205, 45 },
		{ 480, 206, 45 },
		{ 480, 207, 45 },
		{ 480, 208, 45 },
		{ 480, 209, 45 },
		{ 480, 210, 45 },
		{ 480, 211, 45 },
		{ 480, 212, 45 },
		{ 480, 213, 45 },
		{ 480, 214, 45 },
		{ 480, 215, 45 },
		{ 480, 216, 45 },
		{ 480, 217, 45 },
		{ 480, 218, 45 },
		{ 480, 219, 45 },
		{ 480, 220, 45 },
		{ 480, 221, 45 },
		{ 480, 222, 45 },
		{ 480, 223, 45 },
		{ 480, 224, 45 },
		{ 480, 225, 45 },
		{ 480, 226, 45 },
		{ 480, 227, 45 },
		{ 480, 228, 45 },
		{ 480, 229, 45 },
		{ 480, 230, 45 },
		{ 480, 231, 45 },
		{ 480, 232, 45 },
		{ 480, 233, 45 },
		{ 480, 234, 45 },
		{ 480, 235, 45 },
		{ 480, 236, 45 },
		{ 480, 237, 45 },
		{ 480, 238, 45 },
		{ 480, 239, 45 },
		{ 480, 240, 45 },
		{ 480, 241, 45 },
		{ 480, 242, 45 },
		{ 480, 243, 45 },
		{ 480, 244, 45 },
		{ 480, 245, 45 },
		{ 480, 246, 45 },
		{ 480, 247, 45 },
		{ 480, 248, 45 },
		{ 480, 249, 45 },
		{ 480, 250, 45 },
		{ 480, 251, 45 },
		{ 480, 252, 45 },
		{ 480, 253, 45 },
		{ 480, 254, 45 },
		{ 480, 255, 45 },
		{ 480, 256, 45 },
		{ 481, 202, 45 },
		{ 481, 203, 45 },
		{ 481, 204, 45 },
		{ 481, 205, 45 },
		{ 481, 206, 45 },
		{ 481, 207, 45 },
		{ 481, 208, 45 },
		{ 481, 209, 45 },
		{ 481, 210, 45 },
		{ 481, 211, 45 },
		{ 481, 212, 45 },
		{ 481, 213, 45 },
		{ 481, 214, 45 },
		{ 481, 215, 45 },
		{ 481, 216, 45 },
		{ 481, 217, 45 },
		{ 481, 218, 45 },
		{ 481, 219, 45 },
		{ 481, 220, 45 },
		{ 481, 221, 45 },
		{ 481, 222, 45 },
		{ 481, 223, 45 },
		{ 481, 224, 45 },
		{ 481, 225, 45 },
		{ 481, 226, 45 },
		{ 481, 227, 45 },
		{ 481, 228, 45 },
		{ 481, 229, 45 },
		{ 481, 230, 45 },
		{ 481, 231, 45 },
		{ 481, 232, 45 },
		{ 481, 233, 45 },
		{ 481, 234, 45 },
		{ 481, 235, 45 },
		{ 481, 236, 45 },
		{ 481, 237, 45 },
		{ 481, 238, 45 },
		{ 481, 239, 45 },
		{ 481, 240, 45 },
		{ 481, 241, 45 },
		{ 481, 242, 45 },
		{ 481, 243, 45 },
		{ 481, 244, 45 },
		{ 481, 245, 45 },
		{ 481, 246, 45 },
		{ 481, 247, 45 },
		{ 481, 248, 45 },
		{ 481, 249, 45 },
		{ 481, 250, 45 },
		{ 481, 251, 45 },
		{ 481, 252, 45 },
		{ 481, 253, 45 },
		{ 481, 254, 45 },
		{ 481, 255, 45 },
		{ 481, 256, 45 },
		{ 482, 201, 45 },
		{ 482, 202, 45 },
		{ 482, 203, 45 },
		{ 482, 204, 45 },
		{ 482, 205, 45 },
		{ 482, 206, 45 },
		{ 482, 207, 45 },
		{ 482, 208, 45 },
		{ 482, 209, 45 },
		{ 482, 210, 45 },
		{ 482, 211, 45 },
		{ 482, 212, 45 },
		{ 482, 213, 45 },
		{ 482, 214, 45 },
		{ 482, 215, 45 },
		{ 482, 216, 45 },
		{ 482, 217, 45 },
		{ 482, 218, 45 },
		{ 482, 219, 45 },
		{ 482, 220, 45 },
		{ 482, 221, 45 },
		{ 482, 222, 45 },
		{ 482, 223, 45 },
		{ 482, 224, 45 },
		{ 482, 225, 45 },
		{ 482, 226, 45 },
		{ 482, 227, 45 },
		{ 482, 228, 45 },
		{ 482, 229, 45 },
		{ 482, 230, 45 },
		{ 482, 231, 45 },
		{ 482, 232, 45 },
		{ 482, 233, 45 },
		{ 482, 234, 45 },
		{ 482, 235, 45 },
		{ 482, 236, 45 },
		{ 482, 237, 45 },
		{ 482, 238, 45 },
		{ 482, 239, 45 },
		{ 482, 240, 45 },
		{ 482, 241, 45 },
		{ 482, 242, 45 },
		{ 482, 243, 45 },
		{ 482, 244, 45 },
		{ 482, 245, 45 },
		{ 482, 246, 45 },
		{ 482, 247, 45 },
		{ 482, 248, 45 },
		{ 482, 249, 45 },
		{ 482, 250, 45 },
		{ 482, 251, 45 },
		{ 482, 252, 45 },
		{ 482, 253, 45 },
		{ 482, 254, 45 },
		{ 482, 255, 45 },
		{ 482, 256, 45 },
		{ 483, 201, 45 },
		{ 483, 202, 45 },
		{ 483, 203, 45 },
		{ 483, 204, 45 },
		{ 483, 205, 45 },
		{ 483, 206, 45 },
		{ 483, 207, 45 },
		{ 483, 208, 45 },
		{ 483, 209, 45 },
		{ 483, 210, 45 },
		{ 483, 211, 45 },
		{ 483, 212, 45 },
		{ 483, 213, 45 },
		{ 483, 214, 45 },
		{ 483, 215, 45 },
		{ 483, 216, 45 },
		{ 483, 217, 45 },
		{ 483, 218, 45 },
		{ 483, 219, 45 },
		{ 483, 220, 45 },
		{ 483, 221, 45 },
		{ 483, 222, 45 },
		{ 483, 223, 45 },
		{ 483, 224, 45 },
		{ 483, 225, 45 },
		{ 483, 226, 45 },
		{ 483, 227, 45 },
		{ 483, 228, 45 },
		{ 483, 229, 45 },
		{ 483, 230, 45 },
		{ 483, 231, 45 },
		{ 483, 232, 45 },
		{ 483, 233, 45 },
		{ 483, 234, 45 },
		{ 483, 235, 45 },
		{ 483, 236, 45 },
		{ 483, 237, 45 },
		{ 483, 238, 45 },
		{ 483, 239, 45 },
		{ 483, 240, 45 },
		{ 483, 241, 45 },
		{ 483, 242, 45 },
		{ 483, 243, 45 },
		{ 483, 244, 45 },
		{ 483, 245, 45 },
		{ 483, 246, 45 },
		{ 483, 247, 45 },
		{ 483, 248, 45 },
		{ 483, 249, 45 },
		{ 483, 250, 45 },
		{ 483, 251, 45 },
		{ 483, 252, 45 },
		{ 483, 253, 45 },
		{ 483, 254, 45 },
		{ 483, 255, 45 },
		{ 483, 256, 45 },
		{ 484, 201, 45 },
		{ 484, 202, 45 },
		{ 484, 203, 45 },
		{ 484, 204, 45 },
		{ 484, 205, 45 },
		{ 484, 206, 45 },
		{ 484, 207, 45 },
		{ 484, 208, 45 },
		{ 484, 209, 45 },
		{ 484, 210, 45 },
		{ 484, 211, 45 },
		{ 484, 212, 45 },
		{ 484, 213, 45 },
		{ 484, 214, 45 },
		{ 484, 215, 45 },
		{ 484, 216, 45 },
		{ 484, 217, 45 },
		{ 484, 218, 45 },
		{ 484, 219, 45 },
		{ 484, 220, 45 },
		{ 484, 221, 45 },
		{ 484, 222, 45 },
		{ 484, 223, 45 },
		{ 484, 224, 45 },
		{ 484, 225, 45 },
		{ 484, 226, 45 },
		{ 484, 227, 45 },
		{ 484, 228, 45 },
		{ 484, 229, 45 },
		{ 484, 230, 45 },
		{ 484, 231, 45 },
		{ 484, 232, 45 },
		{ 484, 233, 45 },
		{ 484, 234, 45 },
		{ 484, 235, 45 },
		{ 484, 236, 45 },
		{ 484, 237, 45 },
		{ 484, 238, 45 },
		{ 484, 239, 45 },
		{ 484, 240, 45 },
		{ 484, 241, 45 },
		{ 484, 242, 45 },
		{ 484, 243, 45 },
		{ 484, 244, 45 },
		{ 484, 245, 45 },
		{ 484, 246, 45 },
		{ 484, 247, 45 },
		{ 484, 248, 45 },
		{ 484, 249, 45 },
		{ 484, 250, 45 },
		{ 484, 251, 45 },
		{ 484, 252, 45 },
		{ 484, 253, 45 },
		{ 484, 254, 45 },
		{ 484, 255, 45 },
		{ 484, 256, 45 },
		{ 485, 201, 45 },
		{ 485, 202, 45 },
		{ 485, 203, 45 },
		{ 485, 204, 45 },
		{ 485, 205, 45 },
		{ 485, 206, 45 },
		{ 485, 207, 45 },
		{ 485, 208, 45 },
		{ 485, 209, 45 },
		{ 485, 210, 45 },
		{ 485, 211, 45 },
		{ 485, 212, 45 },
		{ 485, 213, 45 },
		{ 485, 214, 45 },
		{ 485, 215, 45 },
		{ 485, 216, 45 },
		{ 485, 217, 45 },
		{ 485, 218, 45 },
		{ 485, 219, 45 },
		{ 485, 220, 45 },
		{ 485, 221, 45 },
		{ 485, 222, 45 },
		{ 485, 223, 45 },
		{ 485, 224, 45 },
		{ 485, 225, 45 },
		{ 485, 226, 45 },
		{ 485, 227, 45 },
		{ 485, 228, 45 },
		{ 485, 229, 45 },
		{ 485, 230, 45 },
		{ 485, 231, 45 },
		{ 485, 232, 45 },
		{ 485, 233, 45 },
		{ 485, 234, 45 },
		{ 485, 235, 45 },
		{ 485, 236, 45 },
		{ 485, 237, 45 },
		{ 485, 238, 45 },
		{ 485, 239, 45 },
		{ 485, 240, 45 },
		{ 485, 241, 45 },
		{ 485, 242, 45 },
		{ 485, 243, 45 },
		{ 485, 244, 45 },
		{ 485, 245, 45 },
		{ 485, 246, 45 },
		{ 485, 247, 45 },
		{ 485, 248, 45 },
		{ 485, 249, 45 },
		{ 485, 250, 45 },
		{ 485, 251, 45 },
		{ 485, 252, 45 },
		{ 485, 253, 45 },
		{ 485, 254, 45 },
		{ 485, 255, 45 },
		{ 485, 256, 45 },
		{ 486, 200, 45 },
		{ 486, 201, 45 },
		{ 486, 202, 45 },
		{ 486, 203, 45 },
		{ 486, 204, 45 },
		{ 486, 205, 45 },
		{ 486, 206, 45 },
		{ 486, 207, 45 },
		{ 486, 208, 45 },
		{ 486, 209, 45 },
		{ 486, 210, 45 },
		{ 486, 211, 45 },
		{ 486, 212, 45 },
		{ 486, 213, 45 },
		{ 486, 214, 45 },
		{ 486, 215, 45 },
		{ 486, 216, 45 },
		{ 486, 217, 45 },
		{ 486, 218, 45 },
		{ 486, 219, 45 },
		{ 486, 220, 45 },
		{ 486, 221, 45 },
		{ 486, 222, 45 },
		{ 486, 223, 45 },
		{ 486, 224, 45 },
		{ 486, 225, 45 },
		{ 486, 226, 45 },
		{ 486, 227, 45 },
		{ 486, 228, 45 },
		{ 486, 229, 45 },
		{ 486, 230, 45 },
		{ 486, 231, 45 },
		{ 486, 232, 45 },
		{ 486, 233, 45 },
		{ 486, 234, 45 },
		{ 486, 235, 45 },
		{ 486, 236, 45 },
		{ 486, 237, 45 },
		{ 486, 238, 45 },
		{ 486, 239, 45 },
		{ 486, 240, 45 },
		{ 486, 241, 45 },
		{ 486, 242, 45 },
		{ 486, 243, 45 },
		{ 486, 244, 45 },
		{ 486, 245, 45 },
		{ 486, 246, 45 },
		{ 486, 247, 45 },
		{ 486, 248, 45 },
		{ 486, 249, 45 },
		{ 486, 250, 45 },
		{ 486, 251, 45 },
		{ 486, 252, 45 },
		{ 486, 253, 45 },
		{ 486, 254, 45 },
		{ 486, 255, 45 },
		{ 486, 256, 45 },
		{ 487, 200, 45 },
		{ 487, 201, 45 },
		{ 487, 202, 45 },
		{ 487, 203, 45 },
		{ 487, 204, 45 },
		{ 487, 205, 45 },
		{ 487, 206, 45 },
		{ 487, 207, 45 },
		{ 487, 208, 45 },
		{ 487, 209, 45 },
		{ 487, 210, 45 },
		{ 487, 211, 45 },
		{ 487, 212, 45 },
		{ 487, 213, 45 },
		{ 487, 214, 45 },
		{ 487, 215, 45 },
		{ 487, 216, 45 },
		{ 487, 217, 45 },
		{ 487, 218, 45 },
		{ 487, 219, 45 },
		{ 487, 220, 45 },
		{ 487, 221, 45 },
		{ 487, 222, 45 },
		{ 487, 223, 45 },
		{ 487, 224, 45 },
		{ 487, 225, 45 },
		{ 487, 226, 45 },
		{ 487, 227, 45 },
		{ 487, 228, 45 },
		{ 487, 229, 45 },
		{ 487, 230, 45 },
		{ 487, 231, 45 },
		{ 487, 232, 45 },
		{ 487, 233, 45 },
		{ 487, 234, 45 },
		{ 487, 235, 45 },
		{ 487, 236, 45 },
		{ 487, 237, 45 },
		{ 487, 238, 45 },
		{ 487, 239, 45 },
		{ 487, 240, 45 },
		{ 487, 241, 45 },
		{ 487, 242, 45 },
		{ 487, 243, 45 },
		{ 487, 244, 45 },
		{ 487, 245, 45 },
		{ 487, 246, 45 },
		{ 487, 247, 45 },
		{ 487, 248, 45 },
		{ 487, 249, 45 },
		{ 487, 250, 45 },
		{ 487, 251, 45 },
		{ 487, 252, 45 },
		{ 487, 253, 45 },
		{ 487, 254, 45 },
		{ 487, 255, 45 },
		{ 487, 256, 45 },
		{ 488, 200, 45 },
		{ 488, 201, 45 },
		{ 488, 202, 45 },
		{ 488, 203, 45 },
		{ 488, 204, 45 },
		{ 488, 205, 45 },
		{ 488, 206, 45 },
		{ 488, 207, 45 },
		{ 488, 208, 45 },
		{ 488, 209, 45 },
		{ 488, 210, 45 },
		{ 488, 211, 45 },
		{ 488, 212, 45 },
		{ 488, 213, 45 },
		{ 488, 214, 45 },
		{ 488, 215, 45 },
		{ 488, 216, 45 },
		{ 488, 217, 45 },
		{ 488, 218, 45 },
		{ 488, 219, 45 },
		{ 488, 220, 45 },
		{ 488, 221, 45 },
		{ 488, 222, 45 },
		{ 488, 223, 45 },
		{ 488, 224, 45 },
		{ 488, 225, 45 },
		{ 488, 226, 45 },
		{ 488, 227, 45 },
		{ 488, 228, 45 },
		{ 488, 229, 45 },
		{ 488, 230, 45 },
		{ 488, 231, 45 },
		{ 488, 232, 45 },
		{ 488, 233, 45 },
		{ 488, 234, 45 },
		{ 488, 235, 45 },
		{ 488, 236, 45 },
		{ 488, 237, 45 },
		{ 488, 238, 45 },
		{ 488, 239, 45 },
		{ 488, 240, 45 },
		{ 488, 241, 45 },
		{ 488, 242, 45 },
		{ 488, 243, 45 },
		{ 488, 244, 45 },
		{ 488, 245, 45 },
		{ 488, 246, 45 },
		{ 488, 247, 45 },
		{ 488, 248, 45 },
		{ 488, 249, 45 },
		{ 488, 250, 45 },
		{ 488, 251, 45 },
		{ 488, 252, 45 },
		{ 488, 253, 45 },
		{ 488, 254, 45 },
		{ 488, 255, 45 },
		{ 488, 256, 45 },
		{ 489, 200, 45 },
		{ 489, 201, 45 },
		{ 489, 202, 45 },
		{ 489, 203, 45 },
		{ 489, 204, 45 },
		{ 489, 205, 45 },
		{ 489, 206, 45 },
		{ 489, 207, 45 },
		{ 489, 208, 45 },
		{ 489, 209, 45 },
		{ 489, 210, 45 },
		{ 489, 211, 45 },
		{ 489, 212, 45 },
		{ 489, 213, 45 },
		{ 489, 214, 45 },
		{ 489, 215, 45 },
		{ 489, 216, 45 },
		{ 489, 217, 45 },
		{ 489, 218, 45 },
		{ 489, 219, 45 },
		{ 489, 220, 45 },
		{ 489, 221, 45 },
		{ 489, 222, 45 },
		{ 489, 223, 45 },
		{ 489, 224, 45 },
		{ 489, 225, 45 },
		{ 489, 226, 45 },
		{ 489, 227, 45 },
		{ 489, 228, 45 },
		{ 489, 229, 45 },
		{ 489, 230, 45 },
		{ 489, 231, 45 },
		{ 489, 232, 45 },
		{ 489, 233, 45 },
		{ 489, 234, 45 },
		{ 489, 235, 45 },
		{ 489, 236, 45 },
		{ 489, 237, 45 },
		{ 489, 238, 45 },
		{ 489, 239, 45 },
		{ 489, 240, 45 },
		{ 489, 241, 45 },
		{ 489, 242, 45 },
		{ 489, 243, 45 },
		{ 489, 244, 45 },
		{ 489, 245, 45 },
		{ 489, 246, 45 },
		{ 489, 247, 45 },
		{ 489, 248, 45 },
		{ 489, 249, 45 },
		{ 489, 250, 45 },
		{ 489, 251, 45 },
		{ 489, 252, 45 },
		{ 489, 253, 45 },
		{ 489, 254, 45 },
		{ 489, 255, 45 },
		{ 489, 256, 45 },
		{ 490, 199, 45 },
		{ 490, 200, 45 },
		{ 490, 201, 45 },
		{ 490, 202, 45 },
		{ 490, 203, 45 },
		{ 490, 204, 45 },
		{ 490, 205, 45 },
		{ 490, 206, 45 },
		{ 490, 207, 45 },
		{ 490, 208, 45 },
		{ 490, 209, 45 },
		{ 490, 210, 45 },
		{ 490, 211, 45 },
		{ 490, 212, 45 },
		{ 490, 213, 45 },
		{ 490, 214, 45 },
		{ 490, 215, 45 },
		{ 490, 216, 45 },
		{ 490, 217, 45 },
		{ 490, 218, 45 },
		{ 490, 219, 45 },
		{ 490, 220, 45 },
		{ 490, 221, 45 },
		{ 490, 222, 45 },
		{ 490, 223, 45 },
		{ 490, 224, 45 },
		{ 490, 225, 45 },
		{ 490, 226, 45 },
		{ 490, 227, 45 },
		{ 490, 228, 45 },
		{ 490, 229, 45 },
		{ 490, 230, 45 },
		{ 490, 231, 45 },
		{ 490, 232, 45 },
		{ 490, 233, 45 },
		{ 490, 234, 45 },
		{ 490, 235, 45 },
		{ 490, 236, 45 },
		{ 490, 237, 45 },
		{ 490, 238, 45 },
		{ 490, 239, 45 },
		{ 490, 240, 45 },
		{ 490, 241, 45 },
		{ 490, 242, 45 },
		{ 490, 243, 45 },
		{ 490, 244, 45 },
		{ 490, 245, 45 },
		{ 490, 246, 45 },
		{ 490, 247, 45 },
		{ 490, 248, 45 },
		{ 490, 249, 45 },
		{ 490, 250, 45 },
		{ 490, 251, 45 },
		{ 490, 252, 45 },
		{ 490, 253, 45 },
		{ 490, 254, 45 },
		{ 490, 255, 45 },
		{ 490, 256, 45 },
		{ 491, 199, 45 },
		{ 491, 200, 45 },
		{ 491, 201, 45 },
		{ 491, 202, 45 },
		{ 491, 203, 45 },
		{ 491, 204, 45 },
		{ 491, 205, 45 },
		{ 491, 206, 45 },
		{ 491, 207, 45 },
		{ 491, 208, 45 },
		{ 491, 209, 45 },
		{ 491, 210, 45 },
		{ 491, 211, 45 },
		{ 491, 212, 45 },
		{ 491, 213, 45 },
		{ 491, 214, 45 },
		{ 491, 215, 45 },
		{ 491, 216, 45 },
		{ 491, 217, 45 },
		{ 491, 218, 45 },
		{ 491, 219, 45 },
		{ 491, 220, 45 },
		{ 491, 221, 45 },
		{ 491, 222, 45 },
		{ 491, 223, 45 },
		{ 491, 224, 45 },
		{ 491, 225, 45 },
		{ 491, 226, 45 },
		{ 491, 227, 45 },
		{ 491, 228, 45 },
		{ 491, 229, 45 },
		{ 491, 230, 45 },
		{ 491, 231, 45 },
		{ 491, 232, 45 },
		{ 491, 233, 45 },
		{ 491, 234, 45 },
		{ 491, 235, 45 },
		{ 491, 236, 45 },
		{ 491, 237, 45 },
		{ 491, 238, 45 },
		{ 491, 239, 45 },
		{ 491, 240, 45 },
		{ 491, 241, 45 },
		{ 491, 242, 45 },
		{ 491, 243, 45 },
		{ 491, 244, 45 },
		{ 491, 245, 45 },
		{ 491, 246, 45 },
		{ 491, 247, 45 },
		{ 491, 248, 45 },
		{ 491, 249, 45 },
		{ 491, 250, 45 },
		{ 491, 251, 45 },
		{ 491, 252, 45 },
		{ 491, 253, 45 },
		{ 491, 254, 45 },
		{ 491, 255, 45 },
		{ 491, 256, 45 },
		{ 492, 199, 45 },
		{ 492, 200, 45 },
		{ 492, 201, 45 },
		{ 492, 202, 45 },
		{ 492, 203, 45 },
		{ 492, 204, 45 },
		{ 492, 205, 45 },
		{ 492, 206, 45 },
		{ 492, 207, 45 },
		{ 492, 208, 45 },
		{ 492, 209, 45 },
		{ 492, 210, 45 },
		{ 492, 211, 45 },
		{ 492, 212, 45 },
		{ 492, 213, 45 },
		{ 492, 214, 45 },
		{ 492, 215, 45 },
		{ 492, 216, 45 },
		{ 492, 217, 45 },
		{ 492, 218, 45 },
		{ 492, 219, 45 },
		{ 492, 220, 45 },
		{ 492, 221, 45 },
		{ 492, 222, 45 },
		{ 492, 223, 45 },
		{ 492, 224, 45 },
		{ 492, 225, 45 },
		{ 492, 226, 45 },
		{ 492, 227, 45 },
		{ 492, 228, 45 },
		{ 492, 229, 45 },
		{ 492, 230, 45 },
		{ 492, 231, 45 },
		{ 492, 232, 45 },
		{ 492, 233, 45 },
		{ 492, 234, 45 },
		{ 492, 235, 45 },
		{ 492, 236, 45 },
		{ 492, 237, 45 },
		{ 492, 238, 45 },
		{ 492, 239, 45 },
		{ 492, 240, 45 },
		{ 492, 241, 45 },
		{ 492, 242, 45 },
		{ 492, 243, 45 },
		{ 492, 244, 45 },
		{ 492, 245, 45 },
		{ 492, 246, 45 },
		{ 492, 247, 45 },
		{ 492, 248, 45 },
		{ 492, 249, 45 },
		{ 492, 250, 45 },
		{ 492, 251, 45 },
		{ 492, 252, 45 },
		{ 492, 253, 45 },
		{ 492, 254, 45 },
		{ 492, 255, 45 },
		{ 492, 256, 45 },
		{ 493, 199, 45 },
		{ 493, 200, 45 },
		{ 493, 201, 45 },
		{ 493, 202, 45 },
		{ 493, 203, 45 },
		{ 493, 204, 45 },
		{ 493, 205, 45 },
		{ 493, 206, 45 },
		{ 493, 207, 45 },
		{ 493, 208, 45 },
		{ 493, 209, 45 },
		{ 493, 210, 45 },
		{ 493, 211, 45 },
		{ 493, 212, 45 },
		{ 493, 213, 45 },
		{ 493, 214, 45 },
		{ 493, 215, 45 },
		{ 493, 216, 45 },
		{ 493, 217, 45 },
		{ 493, 218, 45 },
		{ 493, 219, 45 },
		{ 493, 220, 45 },
		{ 493, 221, 45 },
		{ 493, 222, 45 },
		{ 493, 223, 45 },
		{ 493, 224, 45 },
		{ 493, 225, 45 },
		{ 493, 226, 45 },
		{ 493, 227, 45 },
		{ 493, 228, 45 },
		{ 493, 229, 45 },
		{ 493, 230, 45 },
		{ 493, 231, 45 },
		{ 493, 232, 45 },
		{ 493, 233, 45 },
		{ 493, 234, 45 },
		{ 493, 235, 45 },
		{ 493, 236, 45 },
		{ 493, 237, 45 },
		{ 493, 238, 45 },
		{ 493, 239, 45 },
		{ 493, 240, 45 },
		{ 493, 241, 45 },
		{ 493, 242, 45 },
		{ 493, 243, 45 },
		{ 493, 244, 45 },
		{ 493, 245, 45 },
		{ 493, 246, 45 },
		{ 493, 247, 45 },
		{ 493, 248, 45 },
		{ 493, 249, 45 },
		{ 493, 250, 45 },
		{ 493, 251, 45 },
		{ 493, 252, 45 },
		{ 493, 253, 45 },
		{ 493, 254, 45 },
		{ 493, 255, 45 },
		{ 493, 256, 45 },
		{ 494, 198, 45 },
		{ 494, 199, 45 },
		{ 494, 200, 45 },
		{ 494, 201, 45 },
		{ 494, 202, 45 },
		{ 494, 203, 45 },
		{ 494, 204, 45 },
		{ 494, 205, 45 },
		{ 494, 206, 45 },
		{ 494, 207, 45 },
		{ 494, 208, 45 },
		{ 494, 209, 45 },
		{ 494, 210, 45 },
		{ 494, 211, 45 },
		{ 494, 212, 45 },
		{ 494, 213, 45 },
		{ 494, 214, 45 },
		{ 494, 215, 45 },
		{ 494, 216, 45 },
		{ 494, 217, 45 },
		{ 494, 218, 45 },
		{ 494, 219, 45 },
		{ 494, 220, 45 },
		{ 494, 221, 45 },
		{ 494, 222, 45 },
		{ 494, 223, 45 },
		{ 494, 224, 45 },
		{ 494, 225, 45 },
		{ 494, 226, 45 },
		{ 494, 227, 45 },
		{ 494, 228, 45 },
		{ 494, 229, 45 },
		{ 494, 230, 45 },
		{ 494, 231, 45 },
		{ 494, 232, 45 },
		{ 494, 233, 45 },
		{ 494, 234, 45 },
		{ 494, 235, 45 },
		{ 494, 236, 45 },
		{ 494, 237, 45 },
		{ 494, 238, 45 },
		{ 494, 239, 45 },
		{ 494, 240, 45 },
		{ 494, 241, 45 },
		{ 494, 242, 45 },
		{ 494, 243, 45 },
		{ 494, 244, 45 },
		{ 494, 245, 45 },
		{ 494, 246, 45 },
		{ 494, 247, 45 },
		{ 494, 248, 45 },
		{ 494, 249, 45 },
		{ 494, 250, 45 },
		{ 494, 251, 45 },
		{ 494, 252, 45 },
		{ 494, 253, 45 },
		{ 494, 254, 45 },
		{ 494, 255, 45 },
		{ 494, 256, 45 },
		{ 495, 198, 45 },
		{ 495, 199, 45 },
		{ 495, 200, 45 },
		{ 495, 201, 45 },
		{ 495, 202, 45 },
		{ 495, 203, 45 },
		{ 495, 204, 45 },
		{ 495, 205, 45 },
		{ 495, 206, 45 },
		{ 495, 207, 45 },
		{ 495, 208, 45 },
		{ 495, 209, 45 },
		{ 495, 210, 45 },
		{ 495, 211, 45 },
		{ 495, 212, 45 },
		{ 495, 213, 45 },
		{ 495, 214, 45 },
		{ 495, 215, 45 },
		{ 495, 216, 45 },
		{ 495, 217, 45 },
		{ 495, 218, 45 },
		{ 495, 219, 45 },
		{ 495, 220, 45 },
		{ 495, 221, 45 },
		{ 495, 222, 45 },
		{ 495, 223, 45 },
		{ 495, 224, 45 },
		{ 495, 225, 45 },
		{ 495, 226, 45 },
		{ 495, 227, 45 },
		{ 495, 228, 45 },
		{ 495, 229, 45 },
		{ 495, 230, 45 },
		{ 495, 231, 45 },
		{ 495, 232, 45 },
		{ 495, 233, 45 },
		{ 495, 234, 45 },
		{ 495, 235, 45 },
		{ 495, 236, 45 },
		{ 495, 237, 45 },
		{ 495, 238, 45 },
		{ 495, 239, 45 },
		{ 495, 240, 45 },
		{ 495, 241, 45 },
		{ 495, 242, 45 },
		{ 495, 243, 45 },
		{ 495, 244, 45 },
		{ 495, 245, 45 },
		{ 495, 246, 45 },
		{ 495, 247, 45 },
		{ 495, 248, 45 },
		{ 495, 249, 45 },
		{ 495, 250, 45 },
		{ 495, 251, 45 },
		{ 495, 252, 45 },
		{ 495, 253, 45 },
		{ 495, 254, 45 },
		{ 495, 255, 45 },
		{ 495, 256, 45 },
		{ 496, 198, 45 },
		{ 496, 199, 45 },
		{ 496, 200, 45 },
		{ 496, 201, 45 },
		{ 496, 202, 45 },
		{ 496, 203, 45 },
		{ 496, 204, 45 },
		{ 496, 205, 45 },
		{ 496, 206, 45 },
		{ 496, 207, 45 },
		{ 496, 208, 45 },
		{ 496, 209, 45 },
		{ 496, 210, 45 },
		{ 496, 211, 45 },
		{ 496, 212, 45 },
		{ 496, 213, 45 },
		{ 496, 214, 45 },
		{ 496, 215, 45 },
		{ 496, 216, 45 },
		{ 496, 217, 45 },
		{ 496, 218, 45 },
		{ 496, 219, 45 },
		{ 496, 220, 45 },
		{ 496, 221, 45 },
		{ 496, 222, 45 },
		{ 496, 223, 45 },
		{ 496, 224, 45 },
		{ 496, 225, 45 },
		{ 496, 226, 45 },
		{ 496, 227, 45 },
		{ 496, 228, 45 },
		{ 496, 229, 45 },
		{ 496, 230, 45 },
		{ 496, 231, 45 },
		{ 496, 232, 45 },
		{ 496, 233, 45 },
		{ 496, 234, 45 },
		{ 496, 235, 45 },
		{ 496, 236, 45 },
		{ 496, 237, 45 },
		{ 496, 238, 45 },
		{ 496, 239, 45 },
		{ 496, 240, 45 },
		{ 496, 241, 45 },
		{ 496, 242, 45 },
		{ 496, 243, 45 },
		{ 496, 244, 45 },
		{ 496, 245, 45 },
		{ 496, 246, 45 },
		{ 496, 247, 45 },
		{ 496, 248, 45 },
		{ 496, 249, 45 },
		{ 496, 250, 45 },
		{ 496, 251, 45 },
		{ 496, 252, 45 },
		{ 496, 253, 45 },
		{ 496, 254, 45 },
		{ 496, 255, 45 },
		{ 496, 256, 45 },
		{ 497, 198, 45 },
		{ 497, 199, 45 },
		{ 497, 200, 45 },
		{ 497, 201, 45 },
		{ 497, 202, 45 },
		{ 497, 203, 45 },
		{ 497, 204, 45 },
		{ 497, 205, 45 },
		{ 497, 206, 45 },
		{ 497, 207, 45 },
		{ 497, 208, 45 },
		{ 497, 209, 45 },
		{ 497, 210, 45 },
		{ 497, 211, 45 },
		{ 497, 212, 45 },
		{ 497, 213, 45 },
		{ 497, 214, 45 },
		{ 497, 215, 45 },
		{ 497, 216, 45 },
		{ 497, 217, 45 },
		{ 497, 218, 45 },
		{ 497, 219, 45 },
		{ 497, 220, 45 },
		{ 497, 221, 45 },
		{ 497, 222, 45 },
		{ 497, 223, 45 },
		{ 497, 224, 45 },
		{ 497, 225, 45 },
		{ 497, 226, 45 },
		{ 497, 227, 45 },
		{ 497, 228, 45 },
		{ 497, 229, 45 },
		{ 497, 230, 45 },
		{ 497, 231, 45 },
		{ 497, 232, 45 },
		{ 497, 233, 45 },
		{ 497, 234, 45 },
		{ 497, 235, 45 },
		{ 497, 236, 45 },
		{ 497, 237, 45 },
		{ 497, 238, 45 },
		{ 497, 239, 45 },
		{ 497, 240, 45 },
		{ 497, 241, 45 },
		{ 497, 242, 45 },
		{ 497, 243, 45 },
		{ 497, 244, 45 },
		{ 497, 245, 45 },
		{ 497, 246, 45 },
		{ 497, 247, 45 },
		{ 497, 248, 45 },
		{ 497, 249, 45 },
		{ 497, 250, 45 },
		{ 497, 251, 45 },
		{ 497, 252, 45 },
		{ 497, 253, 45 },
		{ 497, 254, 45 },
		{ 497, 255, 45 },
		{ 497, 256, 45 },
		{ 498, 197, 45 },
		{ 498, 198, 45 },
		{ 498, 199, 45 },
		{ 498, 200, 45 },
		{ 498, 201, 45 },
		{ 498, 202, 45 },
		{ 498, 203, 45 },
		{ 498, 204, 45 },
		{ 498, 205, 45 },
		{ 498, 206, 45 },
		{ 498, 207, 45 },
		{ 498, 208, 45 },
		{ 498, 209, 45 },
		{ 498, 210, 45 },
		{ 498, 211, 45 },
		{ 498, 212, 45 },
		{ 498, 213, 45 },
		{ 498, 214, 45 },
		{ 498, 215, 45 },
		{ 498, 216, 45 },
		{ 498, 217, 45 },
		{ 498, 218, 45 },
		{ 498, 219, 45 },
		{ 498, 220, 45 },
		{ 498, 221, 45 },
		{ 498, 222, 45 },
		{ 498, 223, 45 },
		{ 498, 224, 45 },
		{ 498, 225, 45 },
		{ 498, 226, 45 },
		{ 498, 227, 45 },
		{ 498, 228, 45 },
		{ 498, 229, 45 },
		{ 498, 230, 45 },
		{ 498, 231, 45 },
		{ 498, 232, 45 },
		{ 498, 233, 45 },
		{ 498, 234, 45 },
		{ 498, 235, 45 },
		{ 498, 236, 45 },
		{ 498, 237, 45 },
		{ 498, 238, 45 },
		{ 498, 239, 45 },
		{ 498, 240, 45 },
		{ 498, 241, 45 },
		{ 498, 242, 45 },
		{ 498, 243, 45 },
		{ 498, 244, 45 },
		{ 498, 245, 45 },
		{ 498, 246, 45 },
		{ 498, 247, 45 },
		{ 498, 248, 45 },
		{ 498, 249, 45 },
		{ 498, 250, 45 },
		{ 498, 251, 45 },
		{ 498, 252, 45 },
		{ 498, 253, 45 },
		{ 498, 254, 45 },
		{ 498, 255, 45 },
		{ 498, 256, 45 },
		{ 499, 197, 45 },
		{ 499, 198, 45 },
		{ 499, 199, 45 },
		{ 499, 200, 45 },
		{ 499, 201, 45 },
		{ 499, 202, 45 },
		{ 499, 203, 45 },
		{ 499, 204, 45 },
		{ 499, 205, 45 },
		{ 499, 206, 45 },
		{ 499, 207, 45 },
		{ 499, 208, 45 },
		{ 499, 209, 45 },
		{ 499, 210, 45 },
		{ 499, 211, 45 },
		{ 499, 212, 45 },
		{ 499, 213, 45 },
		{ 499, 214, 45 },
		{ 499, 215, 45 },
		{ 499, 216, 45 },
		{ 499, 217, 45 },
		{ 499, 218, 45 },
		{ 499, 219, 45 },
		{ 499, 220, 45 },
		{ 499, 221, 45 },
		{ 499, 222, 45 },
		{ 499, 223, 45 },
		{ 499, 224, 45 },
		{ 499, 225, 45 },
		{ 499, 226, 45 },
		{ 499, 227, 45 },
		{ 499, 228, 45 },
		{ 499, 229, 45 },
		{ 499, 230, 45 },
		{ 499, 231, 45 },
		{ 499, 232, 45 },
		{ 499, 233, 45 },
		{ 499, 234, 45 },
		{ 499, 235, 45 },
		{ 499, 236, 45 },
		{ 499, 237, 45 },
		{ 499, 238, 45 },
		{ 499, 239, 45 },
		{ 499, 240, 45 },
		{ 499, 241, 45 },
		{ 499, 242, 45 },
		{ 499, 243, 45 },
		{ 499, 244, 45 },
		{ 499, 245, 45 },
		{ 499, 246, 45 },
		{ 499, 247, 45 },
		{ 499, 248, 45 },
		{ 499, 249, 45 },
		{ 499, 250, 45 },
		{ 499, 251, 45 },
		{ 499, 252, 45 },
		{ 499, 253, 45 },
		{ 499, 254, 45 },
		{ 499, 255, 45 },
		{ 499, 256, 45 },
		{ 500, 197, 45 },
		{ 500, 198, 45 },
		{ 500, 199, 45 },
		{ 500, 200, 45 },
		{ 500, 201, 45 },
		{ 500, 202, 45 },
		{ 500, 203, 45 },
		{ 500, 204, 45 },
		{ 500, 205, 45 },
		{ 500, 206, 45 },
		{ 500, 207, 45 },
		{ 500, 208, 45 },
		{ 500, 209, 45 },
		{ 500, 210, 45 },
		{ 500, 211, 45 },
		{ 500, 212, 45 },
		{ 500, 213, 45 },
		{ 500, 214, 45 },
		{ 500, 215, 45 },
		{ 500, 216, 45 },
		{ 500, 217, 45 },
		{ 500, 218, 45 },
		{ 500, 219, 45 },
		{ 500, 220, 45 },
		{ 500, 221, 45 },
		{ 500, 222, 45 },
		{ 500, 223, 45 },
		{ 500, 224, 45 },
		{ 500, 225, 45 },
		{ 500, 226, 45 },
		{ 500, 227, 45 },
		{ 500, 228, 45 },
		{ 500, 229, 45 },
		{ 500, 230, 45 },
		{ 500, 231, 45 },
		{ 500, 232, 45 },
		{ 500, 233, 45 },
		{ 500, 234, 45 },
		{ 500, 235, 45 },
		{ 500, 236, 45 },
		{ 500, 237, 45 },
		{ 500, 238, 45 },
		{ 500, 239, 45 },
		{ 500, 240, 45 },
		{ 500, 241, 45 },
		{ 500, 242, 45 },
		{ 500, 243, 45 },
		{ 500, 244, 45 },
		{ 500, 245, 45 },
		{ 500, 246, 45 },
		{ 500, 247, 45 },
		{ 500, 248, 45 },
		{ 500, 249, 45 },
		{ 500, 250, 45 },
		{ 500, 251, 45 },
		{ 500, 252, 45 },
		{ 500, 253, 45 },
		{ 500, 254, 45 },
		{ 500, 255, 45 },
		{ 500, 256, 45 },
		{ 501, 197, 45 },
		{ 501, 198, 45 },
		{ 501, 199, 45 },
		{ 501, 200, 45 },
		{ 501, 201, 45 },
		{ 501, 202, 45 },
		{ 501, 203, 45 },
		{ 501, 204, 45 },
		{ 501, 205, 45 },
		{ 501, 206, 45 },
		{ 501, 207, 45 },
		{ 501, 208, 45 },
		{ 501, 209, 45 },
		{ 501, 210, 45 },
		{ 501, 211, 45 },
		{ 501, 212, 45 },
		{ 501, 213, 45 },
		{ 501, 214, 45 },
		{ 501, 215, 45 },
		{ 501, 216, 45 },
		{ 501, 217, 45 },
		{ 501, 218, 45 },
		{ 501, 219, 45 },
		{ 501, 220, 45 },
		{ 501, 221, 45 },
		{ 501, 222, 45 },
		{ 501, 223, 45 },
		{ 501, 224, 45 },
		{ 501, 225, 45 },
		{ 501, 226, 45 },
		{ 501, 227, 45 },
		{ 501, 228, 45 },
		{ 501, 229, 45 },
		{ 501, 230, 45 },
		{ 501, 231, 45 },
		{ 501, 232, 45 },
		{ 501, 233, 45 },
		{ 501, 234, 45 },
		{ 501, 235, 45 },
		{ 501, 236, 45 },
		{ 501, 237, 45 },
		{ 501, 238, 45 },
		{ 501, 239, 45 },
		{ 501, 240, 45 },
		{ 501, 241, 45 },
		{ 501, 242, 45 },
		{ 501, 243, 45 },
		{ 501, 244, 45 },
		{ 501, 245, 45 },
		{ 501, 246, 45 },
		{ 501, 247, 45 },
		{ 501, 248, 45 },
		{ 501, 249, 45 },
		{ 501, 250, 45 },
		{ 501, 251, 45 },
		{ 501, 252, 45 },
		{ 501, 253, 45 },
		{ 501, 254, 45 },
		{ 501, 255, 45 },
		{ 501, 256, 45 },
		{ 502, 196, 45 },
		{ 502, 197, 45 },
		{ 502, 198, 45 },
		{ 502, 199, 45 },
		{ 502, 200, 45 },
		{ 502, 201, 45 },
		{ 502, 202, 45 },
		{ 502, 203, 45 },
		{ 502, 204, 45 },
		{ 502, 205, 45 },
		{ 502, 206, 45 },
		{ 502, 207, 45 },
		{ 502, 208, 45 },
		{ 502, 209, 45 },
		{ 502, 210, 45 },
		{ 502, 211, 45 },
		{ 502, 212, 45 },
		{ 502, 213, 45 },
		{ 502, 214, 45 },
		{ 502, 215, 45 },
		{ 502, 216, 45 },
		{ 502, 217, 45 },
		{ 502, 218, 45 },
		{ 502, 219, 45 },
		{ 502, 220, 45 },
		{ 502, 221, 45 },
		{ 502, 222, 45 },
		{ 502, 223, 45 },
		{ 502, 224, 45 },
		{ 502, 225, 45 },
		{ 502, 226, 45 },
		{ 502, 227, 45 },
		{ 502, 228, 45 },
		{ 502, 229, 45 },
		{ 502, 230, 45 },
		{ 502, 231, 45 },
		{ 502, 232, 45 },
		{ 502, 233, 45 },
		{ 502, 234, 45 },
		{ 502, 235, 45 },
		{ 502, 236, 45 },
		{ 502, 237, 45 },
		{ 502, 238, 45 },
		{ 502, 239, 45 },
		{ 502, 240, 45 },
		{ 502, 241, 45 },
		{ 502, 242, 45 },
		{ 502, 243, 45 },
		{ 502, 244, 45 },
		{ 502, 245, 45 },
		{ 502, 246, 45 },
		{ 502, 247, 45 },
		{ 502, 248, 45 },
		{ 502, 249, 45 },
		{ 502, 250, 45 },
		{ 502, 251, 45 },
		{ 502, 252, 45 },
		{ 502, 253, 45 },
		{ 502, 254, 45 },
		{ 502, 255, 45 },
		{ 502, 256, 45 },
		{ 503, 196, 45 },
		{ 503, 197, 45 },
		{ 503, 198, 45 },
		{ 503, 199, 45 },
		{ 503, 200, 45 },
		{ 503, 201, 45 },
		{ 503, 202, 45 },
		{ 503, 203, 45 },
		{ 503, 204, 45 },
		{ 503, 205, 45 },
		{ 503, 206, 45 },
		{ 503, 207, 45 },
		{ 503, 208, 45 },
		{ 503, 209, 45 },
		{ 503, 210, 45 },
		{ 503, 211, 45 },
		{ 503, 212, 45 },
		{ 503, 213, 45 },
		{ 503, 214, 45 },
		{ 503, 215, 45 },
		{ 503, 216, 45 },
		{ 503, 217, 45 },
		{ 503, 218, 45 },
		{ 503, 219, 45 },
		{ 503, 220, 45 },
		{ 503, 221, 45 },
		{ 503, 222, 45 },
		{ 503, 223, 45 },
		{ 503, 224, 45 },
		{ 503, 225, 45 },
		{ 503, 226, 45 },
		{ 503, 227, 45 },
		{ 503, 228, 45 },
		{ 503, 229, 45 },
		{ 503, 230, 45 },
		{ 503, 231, 45 },
		{ 503, 232, 45 },
		{ 503, 233, 45 },
		{ 503, 234, 45 },
		{ 503, 235, 45 },
		{ 503, 236, 45 },
		{ 503, 237, 45 },
		{ 503, 238, 45 },
		{ 503, 239, 45 },
		{ 503, 240, 45 },
		{ 503, 241, 45 },
		{ 503, 242, 45 },
		{ 503, 243, 45 },
		{ 503, 244, 45 },
		{ 503, 245, 45 },
		{ 503, 246, 45 },
		{ 503, 247, 45 },
		{ 503, 248, 45 },
		{ 503, 249, 45 },
		{ 503, 250, 45 },
		{ 503, 251, 45 },
		{ 503, 252, 45 },
		{ 503, 253, 45 },
		{ 503, 254, 45 },
		{ 503, 255, 45 },
		{ 503, 256, 45 },
		{ 504, 196, 45 },
		{ 504, 197, 45 },
		{ 504, 198, 45 },
		{ 504, 199, 45 },
		{ 504, 200, 45 },
		{ 504, 201, 45 },
		{ 504, 202, 45 },
		{ 504, 203, 45 },
		{ 504, 204, 45 },
		{ 504, 205, 45 },
		{ 504, 206, 45 },
		{ 504, 207, 45 },
		{ 504, 208, 45 },
		{ 504, 209, 45 },
		{ 504, 210, 45 },
		{ 504, 211, 45 },
		{ 504, 212, 45 },
		{ 504, 213, 45 },
		{ 504, 214, 45 },
		{ 504, 215, 45 },
		{ 504, 216, 45 },
		{ 504, 217, 45 },
		{ 504, 218, 45 },
		{ 504, 219, 45 },
		{ 504, 220, 45 },
		{ 504, 221, 45 },
		{ 504, 222, 45 },
		{ 504, 223, 45 },
		{ 504, 224, 45 },
		{ 504, 225, 45 },
		{ 504, 226, 45 },
		{ 504, 227, 45 },
		{ 504, 228, 45 },
		{ 504, 229, 45 },
		{ 504, 230, 45 },
		{ 504, 231, 45 },
		{ 504, 232, 45 },
		{ 504, 233, 45 },
		{ 504, 234, 45 },
		{ 504, 235, 45 },
		{ 504, 236, 45 },
		{ 504, 237, 45 },
		{ 504, 238, 45 },
		{ 504, 239, 45 },
		{ 504, 240, 45 },
		{ 504, 241, 45 },
		{ 504, 242, 45 },
		{ 504, 243, 45 },
		{ 504, 244, 45 },
		{ 504, 245, 45 },
		{ 504, 246, 45 },
		{ 504, 247, 45 },
		{ 504, 248, 45 },
		{ 504, 249, 45 },
		{ 504, 250, 45 },
		{ 504, 251, 45 },
		{ 504, 252, 45 },
		{ 504, 253, 45 },
		{ 504, 254, 45 },
		{ 504, 255, 45 },
		{ 504, 256, 45 },
		{ 505, 196, 45 },
		{ 505, 197, 45 },
		{ 505, 198, 45 },
		{ 505, 199, 45 },
		{ 505, 200, 45 },
		{ 505, 201, 45 },
		{ 505, 202, 45 },
		{ 505, 203, 45 },
		{ 505, 204, 45 },
		{ 505, 205, 45 },
		{ 505, 206, 45 },
		{ 505, 207, 45 },
		{ 505, 208, 45 },
		{ 505, 209, 45 },
		{ 505, 210, 45 },
		{ 505, 211, 45 },
		{ 505, 212, 45 },
		{ 505, 213, 45 },
		{ 505, 214, 45 },
		{ 505, 215, 45 },
		{ 505, 216, 45 },
		{ 505, 217, 45 },
		{ 505, 218, 45 },
		{ 505, 219, 45 },
		{ 505, 220, 45 },
		{ 505, 221, 45 },
		{ 505, 222, 45 },
		{ 505, 223, 45 },
		{ 505, 224, 45 },
		{ 505, 225, 45 },
		{ 505, 226, 45 },
		{ 505, 227, 45 },
		{ 505, 228, 45 },
		{ 505, 229, 45 },
		{ 505, 230, 45 },
		{ 505, 231, 45 },
		{ 505, 232, 45 },
		{ 505, 233, 45 },
		{ 505, 234, 45 },
		{ 505, 235, 45 },
		{ 505, 236, 45 },
		{ 505, 237, 45 },
		{ 505, 238, 45 },
		{ 505, 239, 45 },
		{ 505, 240, 45 },
		{ 505, 241, 45 },
		{ 505, 242, 45 },
		{ 505, 243, 45 },
		{ 505, 244, 45 },
		{ 505, 245, 45 },
		{ 505, 246, 45 },
		{ 505, 247, 45 },
		{ 505, 248, 45 },
		{ 505, 249, 45 },
		{ 505, 250, 45 },
		{ 505, 251, 45 },
		{ 505, 252, 45 },
		{ 505, 253, 45 },
		{ 505, 254, 45 },
		{ 505, 255, 45 },
		{ 505, 256, 45 },
		{ 506, 195, 45 },
		{ 506, 196, 45 },
		{ 506, 197, 45 },
		{ 506, 198, 45 },
		{ 506, 199, 45 },
		{ 506, 200, 45 },
		{ 506, 201, 45 },
		{ 506, 202, 45 },
		{ 506, 203, 45 },
		{ 506, 204, 45 },
		{ 506, 205, 45 },
		{ 506, 206, 45 },
		{ 506, 207, 45 },
		{ 506, 208, 45 },
		{ 506, 209, 45 },
		{ 506, 210, 45 },
		{ 506, 211, 45 },
		{ 506, 212, 45 },
		{ 506, 213, 45 },
		{ 506, 214, 45 },
		{ 506, 215, 45 },
		{ 506, 216, 45 },
		{ 506, 217, 45 },
		{ 506, 218, 45 },
		{ 506, 219, 45 },
		{ 506, 220, 45 },
		{ 506, 221, 45 },
		{ 506, 222, 45 },
		{ 506, 223, 45 },
		{ 506, 224, 45 },
		{ 506, 225, 45 },
		{ 506, 226, 45 },
		{ 506, 227, 45 },
		{ 506, 228, 45 },
		{ 506, 229, 45 },
		{ 506, 230, 45 },
		{ 506, 231, 45 },
		{ 506, 232, 45 },
		{ 506, 233, 45 },
		{ 506, 234, 45 },
		{ 506, 235, 45 },
		{ 506, 236, 45 },
		{ 506, 237, 45 },
		{ 506, 238, 45 },
		{ 506, 239, 45 },
		{ 506, 240, 45 },
		{ 506, 241, 45 },
		{ 506, 242, 45 },
		{ 506, 243, 45 },
		{ 506, 244, 45 },
		{ 506, 245, 45 },
		{ 506, 246, 45 },
		{ 506, 247, 45 },
		{ 506, 248, 45 },
		{ 506, 249, 45 },
		{ 506, 250, 45 },
		{ 506, 251, 45 },
		{ 506, 252, 45 },
		{ 506, 253, 45 },
		{ 506, 254, 45 },
		{ 506, 255, 45 },
		{ 506, 256, 45 },
		{ 507, 195, 45 },
		{ 507, 196, 45 },
		{ 507, 197, 45 },
		{ 507, 198, 45 },
		{ 507, 199, 45 },
		{ 507, 200, 45 },
		{ 507, 201, 45 },
		{ 507, 202, 45 },
		{ 507, 203, 45 },
		{ 507, 204, 45 },
		{ 507, 205, 45 },
		{ 507, 206, 45 },
		{ 507, 207, 45 },
		{ 507, 208, 45 },
		{ 507, 209, 45 },
		{ 507, 210, 45 },
		{ 507, 211, 45 },
		{ 507, 212, 45 },
		{ 507, 213, 45 },
		{ 507, 214, 45 },
		{ 507, 215, 45 },
		{ 507, 216, 45 },
		{ 507, 217, 45 },
		{ 507, 218, 45 },
		{ 507, 219, 45 },
		{ 507, 220, 45 },
		{ 507, 221, 45 },
		{ 507, 222, 45 },
		{ 507, 223, 45 },
		{ 507, 224, 45 },
		{ 507, 225, 45 },
		{ 507, 226, 45 },
		{ 507, 227, 45 },
		{ 507, 228, 45 },
		{ 507, 229, 45 },
		{ 507, 230, 45 },
		{ 507, 231, 45 },
		{ 507, 232, 45 },
		{ 507, 233, 45 },
		{ 507, 234, 45 },
		{ 507, 235, 45 },
		{ 507, 236, 45 },
		{ 507, 237, 45 },
		{ 507, 238, 45 },
		{ 507, 239, 45 },
		{ 507, 240, 45 },
		{ 507, 241, 45 },
		{ 507, 242, 45 },
		{ 507, 243, 45 },
		{ 507, 244, 45 },
		{ 507, 245, 45 },
		{ 507, 246, 45 },
		{ 507, 247, 45 },
		{ 507, 248, 45 },
		{ 507, 249, 45 },
		{ 507, 250, 45 },
		{ 507, 251, 45 },
		{ 507, 252, 45 },
		{ 507, 253, 45 },
		{ 507, 254, 45 },
		{ 507, 255, 45 },
		{ 507, 256, 45 },
		{ 508, 195, 45 },
		{ 508, 196, 45 },
		{ 508, 197, 45 },
		{ 508, 198, 45 },
		{ 508, 199, 45 },
		{ 508, 200, 45 },
		{ 508, 201, 45 },
		{ 508, 202, 45 },
		{ 508, 203, 45 },
		{ 508, 204, 45 },
		{ 508, 205, 45 },
		{ 508, 206, 45 },
		{ 508, 207, 45 },
		{ 508, 208, 45 },
		{ 508, 209, 45 },
		{ 508, 210, 45 },
		{ 508, 211, 45 },
		{ 508, 212, 45 },
		{ 508, 213, 45 },
		{ 508, 214, 45 },
		{ 508, 215, 45 },
		{ 508, 216, 45 },
		{ 508, 217, 45 },
		{ 508, 218, 45 },
		{ 508, 219, 45 },
		{ 508, 220, 45 },
		{ 508, 221, 45 },
		{ 508, 222, 45 },
		{ 508, 223, 45 },
		{ 508, 224, 45 },
		{ 508, 225, 45 },
		{ 508, 226, 45 },
		{ 508, 227, 45 },
		{ 508, 228, 45 },
		{ 508, 229, 45 },
		{ 508, 230, 45 },
		{ 508, 231, 45 },
		{ 508, 232, 45 },
		{ 508, 233, 45 },
		{ 508, 234, 45 },
		{ 508, 235, 45 },
		{ 508, 236, 45 },
		{ 508, 237, 45 },
		{ 508, 238, 45 },
		{ 508, 239, 45 },
		{ 508, 240, 45 },
		{ 508, 241, 45 },
		{ 508, 242, 45 },
		{ 508, 243, 45 },
		{ 508, 244, 45 },
		{ 508, 245, 45 },
		{ 508, 246, 45 },
		{ 508, 247, 45 },
		{ 508, 248, 45 },
		{ 508, 249, 45 },
		{ 508, 250, 45 },
		{ 508, 251, 45 },
		{ 508, 252, 45 },
		{ 508, 253, 45 },
		{ 508, 254, 45 },
		{ 508, 255, 45 },
		{ 508, 256, 45 },
		{ 509, 195, 45 },
		{ 509, 196, 45 },
		{ 509, 197, 45 },
		{ 509, 198, 45 },
		{ 509, 199, 45 },
		{ 509, 200, 45 },
		{ 509, 201, 45 },
		{ 509, 202, 45 },
		{ 509, 203, 45 },
		{ 509, 204, 45 },
		{ 509, 205, 45 },
		{ 509, 206, 45 },
		{ 509, 207, 45 },
		{ 509, 208, 45 },
		{ 509, 209, 45 },
		{ 509, 210, 45 },
		{ 509, 211, 45 },
		{ 509, 212, 45 },
		{ 509, 213, 45 },
		{ 509, 214, 45 },
		{ 509, 215, 45 },
		{ 509, 216, 45 },
		{ 509, 217, 45 },
		{ 509, 218, 45 },
		{ 509, 219, 45 },
		{ 509, 220, 45 },
		{ 509, 221, 45 },
		{ 509, 222, 45 },
		{ 509, 223, 45 },
		{ 509, 224, 45 },
		{ 509, 225, 45 },
		{ 509, 226, 45 },
		{ 509, 227, 45 },
		{ 509, 228, 45 },
		{ 509, 229, 45 },
		{ 509, 230, 45 },
		{ 509, 231, 45 },
		{ 509, 232, 45 },
		{ 509, 233, 45 },
		{ 509, 234, 45 },
		{ 509, 235, 45 },
		{ 509, 236, 45 },
		{ 509, 237, 45 },
		{ 509, 238, 45 },
		{ 509, 239, 45 },
		{ 509, 240, 45 },
		{ 509, 241, 45 },
		{ 509, 242, 45 },
		{ 509, 243, 45 },
		{ 509, 244, 45 },
		{ 509, 245, 45 },
		{ 509, 246, 45 },
		{ 509, 247, 45 },
		{ 509, 248, 45 },
		{ 509, 249, 45 },
		{ 509, 250, 45 },
		{ 509, 251, 45 },
		{ 509, 252, 45 },
		{ 509, 253, 45 },
		{ 509, 254, 45 },
		{ 509, 255, 45 },
		{ 509, 256, 45 },
		{ 510, 194, 45 },
		{ 510, 195, 45 },
		{ 510, 196, 45 },
		{ 510, 197, 45 },
		{ 510, 198, 45 },
		{ 510, 199, 45 },
		{ 510, 200, 45 },
		{ 510, 201, 45 },
		{ 510, 202, 45 },
		{ 510, 203, 45 },
		{ 510, 204, 45 },
		{ 510, 205, 45 },
		{ 510, 206, 45 },
		{ 510, 207, 45 },
		{ 510, 208, 45 },
		{ 510, 209, 45 },
		{ 510, 210, 45 },
		{ 510, 211, 45 },
		{ 510, 212, 45 },
		{ 510, 213, 45 },
		{ 510, 214, 45 },
		{ 510, 215, 45 },
		{ 510, 216, 45 },
		{ 510, 217, 45 },
		{ 510, 218, 45 },
		{ 510, 219, 45 },
		{ 510, 220, 45 },
		{ 510, 221, 45 },
		{ 510, 222, 45 },
		{ 510, 223, 45 },
		{ 510, 224, 45 },
		{ 510, 225, 45 },
		{ 510, 226, 45 },
		{ 510, 227, 45 },
		{ 510, 228, 45 },
		{ 510, 229, 45 },
		{ 510, 230, 45 },
		{ 510, 231, 45 },
		{ 510, 232, 45 },
		{ 510, 233, 45 },
		{ 510, 234, 45 },
		{ 510, 235, 45 },
		{ 510, 236, 45 },
		{ 510, 237, 45 },
		{ 510, 238, 45 },
		{ 510, 239, 45 },
		{ 510, 240, 45 },
		{ 510, 241, 45 },
		{ 510, 242, 45 },
		{ 510, 243, 45 },
		{ 510, 244, 45 },
		{ 510, 245, 45 },
		{ 510, 246, 45 },
		{ 510, 247, 45 },
		{ 510, 248, 45 },
		{ 510, 249, 45 },
		{ 510, 250, 45 },
		{ 510, 251, 45 },
		{ 510, 252, 45 },
		{ 510, 253, 45 },
		{ 510, 254, 45 },
		{ 510, 255, 45 },
		{ 510, 256, 45 },
		{ 511, 194, 45 },
		{ 511, 195, 45 },
		{ 511, 196, 45 },
		{ 511, 197, 45 },
		{ 511, 198, 45 },
		{ 511, 199, 45 },
		{ 511, 200, 45 },
		{ 511, 201, 45 },
		{ 511, 202, 45 },
		{ 511, 203, 45 },
		{ 511, 204, 45 },
		{ 511, 205, 45 },
		{ 511, 206, 45 },
		{ 511, 207, 45 },
		{ 511, 208, 45 },
		{ 511, 209, 45 },
		{ 511, 210, 45 },
		{ 511, 211, 45 },
		{ 511, 212, 45 },
		{ 511, 213, 45 },
		{ 511, 214, 45 },
		{ 511, 215, 45 },
		{ 511, 216, 45 },
		{ 511, 217, 45 },
		{ 511, 218, 45 },
		{ 511, 219, 45 },
		{ 511, 220, 45 },
		{ 511, 221, 45 },
		{ 511, 222, 45 },
		{ 511, 223, 45 },
		{ 511, 224, 45 },
		{ 511, 225, 45 },
		{ 511, 226, 45 },
		{ 511, 227, 45 },
		{ 511, 228, 45 },
		{ 511, 229, 45 },
		{ 511, 230, 45 },
		{ 511, 231, 45 },
		{ 511, 232, 45 },
		{ 511, 233, 45 },
		{ 511, 234, 45 },
		{ 511, 235, 45 },
		{ 511, 236, 45 },
		{ 511, 237, 45 },
		{ 511, 238, 45 },
		{ 511, 239, 45 },
		{ 511, 240, 45 },
		{ 511, 241, 45 },
		{ 511, 242, 45 },
		{ 511, 243, 45 },
		{ 511, 244, 45 },
		{ 511, 245, 45 },
		{ 511, 246, 45 },
		{ 511, 247, 45 },
		{ 511, 248, 45 },
		{ 511, 249, 45 },
		{ 511, 250, 45 },
		{ 511, 251, 45 },
		{ 511, 252, 45 },
		{ 511, 253, 45 },
		{ 511, 254, 45 },
		{ 511, 255, 45 },
		{ 511, 256, 45 },
		{ 512, 194, 45 },
		{ 512, 195, 45 },
		{ 512, 196, 45 },
		{ 512, 197, 45 },
		{ 512, 198, 45 },
		{ 512, 199, 45 },
		{ 512, 200, 45 },
		{ 512, 201, 45 },
		{ 512, 202, 45 },
		{ 512, 203, 45 },
		{ 512, 204, 45 },
		{ 512, 205, 45 },
		{ 512, 206, 45 },
		{ 512, 207, 45 },
		{ 512, 208, 45 },
		{ 512, 209, 45 },
		{ 512, 210, 45 },
		{ 512, 211, 45 },
		{ 512, 212, 45 },
		{ 512, 213, 45 },
		{ 512, 214, 45 },
		{ 512, 215, 45 },
		{ 512, 216, 45 },
		{ 512, 217, 45 },
		{ 512, 218, 45 },
		{ 512, 219, 45 },
		{ 512, 220, 45 },
		{ 512, 221, 45 },
		{ 512, 222, 45 },
		{ 512, 223, 45 },
		{ 512, 224, 45 },
		{ 512, 225, 45 },
		{ 512, 226, 45 },
		{ 512, 227, 45 },
		{ 512, 228, 45 },
		{ 512, 229, 45 },
		{ 512, 230, 45 },
		{ 512, 231, 45 },
		{ 512, 232, 45 },
		{ 512, 233, 45 },
		{ 512, 234, 45 },
		{ 512, 235, 45 },
		{ 512, 236, 45 },
		{ 512, 237, 45 },
		{ 512, 238, 45 },
		{ 512, 239, 45 },
		{ 512, 240, 45 },
		{ 512, 241, 45 },
		{ 512, 242, 45 },
		{ 512, 243, 45 },
		{ 512, 244, 45 },
		{ 512, 245, 45 },
		{ 512, 246, 45 },
		{ 512, 247, 45 },
		{ 512, 248, 45 },
		{ 512, 249, 45 },
		{ 512, 250, 45 },
		{ 512, 251, 45 },
		{ 512, 252, 45 },
		{ 512, 253, 45 },
		{ 512, 254, 45 },
		{ 512, 255, 45 },
		{ 512, 256, 45 },
		{ 1, 226, 46 },
		{ 1, 227, 46 },
		{ 1, 228, 46 },
		{ 1, 229, 46 },
		{ 1, 230, 46 },
		{ 1, 231, 46 },
		{ 1, 232, 46 },
		{ 1, 233, 46 },
		{ 1, 234, 46 },
		{ 1, 235, 46 },
		{ 1, 236, 46 },
		{ 1, 237, 46 },
		{ 1, 238, 46 },
		{ 1, 239, 46 },
		{ 1, 240, 46 },
		{ 1, 241, 46 },
		{ 1, 242, 46 },
		{ 1, 243, 46 },
		{ 1, 244, 46 },
		{ 1, 245, 46 },
		{ 1, 246, 46 },
		{ 1, 247, 46 },
		{ 1, 248, 46 },
		{ 1, 249, 46 },
		{ 1, 250, 46 },
		{ 1, 251, 46 },
		{ 1, 252, 46 },
		{ 1, 253, 46 },
		{ 1, 254, 46 },
		{ 1, 255, 46 },
		{ 1, 256, 46 },
		{ 1, 257, 46 },
		{ 1, 258, 46 },
		{ 1, 259, 46 },
		{ 1, 260, 46 },
		{ 1, 261, 46 },
		{ 1, 262, 46 },
		{ 1, 263, 46 },
		{ 1, 264, 46 },
		{ 1, 265, 46 },
		{ 1, 266, 46 },
		{ 1, 267, 46 },
		{ 1, 268, 46 },
		{ 1, 269, 46 },
		{ 1, 270, 46 },
		{ 1, 271, 46 },
		{ 1, 272, 46 },
		{ 1, 273, 46 },
		{ 1, 274, 46 },
		{ 1, 275, 46 },
		{ 1, 276, 46 },
		{ 1, 277, 46 },
		{ 1, 278, 46 },
		{ 1, 279, 46 },
		{ 1, 280, 46 },
		{ 1, 281, 46 },
		{ 1, 282, 46 },
		{ 1, 283, 46 },
		{ 1, 284, 46 },
		{ 1, 285, 46 },
		{ 1, 286, 46 },
		{ 1, 287, 46 },
		{ 1, 288, 46 },
		{ 2, 226, 46 },
		{ 2, 227, 46 },
		{ 2, 228, 46 },
		{ 2, 229, 46 },
		{ 2, 230, 46 },
		{ 2, 231, 46 },
		{ 2, 232, 46 },
		{ 2, 233, 46 },
		{ 2, 234, 46 },
		{ 2, 235, 46 },
		{ 2, 236, 46 },
		{ 2, 237, 46 },
		{ 2, 238, 46 },
		{ 2, 239, 46 },
		{ 2, 240, 46 },
		{ 2, 241, 46 },
		{ 2, 242, 46 },
		{ 2, 243, 46 },
		{ 2, 244, 46 },
		{ 2, 245, 46 },
		{ 2, 246, 46 },
		{ 2, 247, 46 },
		{ 2, 248, 46 },
		{ 2, 249, 46 },
		{ 2, 250, 46 },
		{ 2, 251, 46 },
		{ 2, 252, 46 },
		{ 2, 253, 46 },
		{ 2, 254, 46 },
		{ 2, 255, 46 },
		{ 2, 256, 46 },
		{ 2, 257, 46 },
		{ 2, 258, 46 },
		{ 2, 259, 46 },
		{ 2, 260, 46 },
		{ 2, 261, 46 },
		{ 2, 262, 46 },
		{ 2, 263, 46 },
		{ 2, 264, 46 },
		{ 2, 265, 46 },
		{ 2, 266, 46 },
		{ 2, 267, 46 },
		{ 2, 268, 46 },
		{ 2, 269, 46 },
		{ 2, 270, 46 },
		{ 2, 271, 46 },
		{ 2, 272, 46 },
		{ 2, 273, 46 },
		{ 2, 274, 46 },
		{ 2, 275, 46 },
		{ 2, 276, 46 },
		{ 2, 277, 46 },
		{ 2, 278, 46 },
		{ 2, 279, 46 },
		{ 2, 280, 46 },
		{ 2, 281, 46 },
		{ 2, 282, 46 },
		{ 2, 283, 46 },
		{ 2, 284, 46 },
		{ 2, 285, 46 },
		{ 2, 286, 46 },
		{ 2, 287, 46 },
		{ 2, 288, 46 },
		{ 3, 226, 46 },
		{ 3, 227, 46 },
		{ 3, 228, 46 },
		{ 3, 229, 46 },
		{ 3, 230, 46 },
		{ 3, 231, 46 },
		{ 3, 232, 46 },
		{ 3, 233, 46 },
		{ 3, 234, 46 },
		{ 3, 235, 46 },
		{ 3, 236, 46 },
		{ 3, 237, 46 },
		{ 3, 238, 46 },
		{ 3, 239, 46 },
		{ 3, 240, 46 },
		{ 3, 241, 46 },
		{ 3, 242, 46 },
		{ 3, 243, 46 },
		{ 3, 244, 46 },
		{ 3, 245, 46 },
		{ 3, 246, 46 },
		{ 3, 247, 46 },
		{ 3, 248, 46 },
		{ 3, 249, 46 },
		{ 3, 250, 46 },
		{ 3, 251, 46 },
		{ 3, 252, 46 },
		{ 3, 253, 46 },
		{ 3, 254, 46 },
		{ 3, 255, 46 },
		{ 3, 256, 46 },
		{ 3, 257, 46 },
		{ 3, 258, 46 },
		{ 3, 259, 46 },
		{ 3, 260, 46 },
		{ 3, 261, 46 },
		{ 3, 262, 46 },
		{ 3, 263, 46 },
		{ 3, 264, 46 },
		{ 3, 265, 46 },
		{ 3, 266, 46 },
		{ 3, 267, 46 },
		{ 3, 268, 46 },
		{ 3, 269, 46 },
		{ 3, 270, 46 },
		{ 3, 271, 46 },
		{ 3, 272, 46 },
		{ 3, 273, 46 },
		{ 3, 274, 46 },
		{ 3, 275, 46 },
		{ 3, 276, 46 },
		{ 3, 277, 46 },
		{ 3, 278, 46 },
		{ 3, 279, 46 },
		{ 3, 280, 46 },
		{ 3, 281, 46 },
		{ 3, 282, 46 },
		{ 3, 283, 46 },
		{ 3, 284, 46 },
		{ 3, 285, 46 },
		{ 3, 286, 46 },
		{ 3, 287, 46 },
		{ 3, 288, 46 },
		{ 4, 226, 46 },
		{ 4, 227, 46 },
		{ 4, 228, 46 },
		{ 4, 229, 46 },
		{ 4, 230, 46 },
		{ 4, 231, 46 },
		{ 4, 232, 46 },
		{ 4, 233, 46 },
		{ 4, 234, 46 },
		{ 4, 235, 46 },
		{ 4, 236, 46 },
		{ 4, 237, 46 },
		{ 4, 238, 46 },
		{ 4, 239, 46 },
		{ 4, 240, 46 },
		{ 4, 241, 46 },
		{ 4, 242, 46 },
		{ 4, 243, 46 },
		{ 4, 244, 46 },
		{ 4, 245, 46 },
		{ 4, 246, 46 },
		{ 4, 247, 46 },
		{ 4, 248, 46 },
		{ 4, 249, 46 },
		{ 4, 250, 46 },
		{ 4, 251, 46 },
		{ 4, 252, 46 },
		{ 4, 253, 46 },
		{ 4, 254, 46 },
		{ 4, 255, 46 },
		{ 4, 256, 46 },
		{ 4, 257, 46 },
		{ 4, 258, 46 },
		{ 4, 259, 46 },
		{ 4, 260, 46 },
		{ 4, 261, 46 },
		{ 4, 262, 46 },
		{ 4, 263, 46 },
		{ 4, 264, 46 },
		{ 4, 265, 46 },
		{ 4, 266, 46 },
		{ 4, 267, 46 },
		{ 4, 268, 46 },
		{ 4, 269, 46 },
		{ 4, 270, 46 },
		{ 4, 271, 46 },
		{ 4, 272, 46 },
		{ 4, 273, 46 },
		{ 4, 274, 46 },
		{ 4, 275, 46 },
		{ 4, 276, 46 },
		{ 4, 277, 46 },
		{ 4, 278, 46 },
		{ 4, 279, 46 },
		{ 4, 280, 46 },
		{ 4, 281, 46 },
		{ 4, 282, 46 },
		{ 4, 283, 46 },
		{ 4, 284, 46 },
		{ 4, 285, 46 },
		{ 4, 286, 46 },
		{ 4, 287, 46 },
		{ 4, 288, 46 },
		{ 5, 226, 46 },
		{ 5, 227, 46 },
		{ 5, 228, 46 },
		{ 5, 229, 46 },
		{ 5, 230, 46 },
		{ 5, 231, 46 },
		{ 5, 232, 46 },
		{ 5, 233, 46 },
		{ 5, 234, 46 },
		{ 5, 235, 46 },
		{ 5, 236, 46 },
		{ 5, 237, 46 },
		{ 5, 238, 46 },
		{ 5, 239, 46 },
		{ 5, 240, 46 },
		{ 5, 241, 46 },
		{ 5, 242, 46 },
		{ 5, 243, 46 },
		{ 5, 244, 46 },
		{ 5, 245, 46 },
		{ 5, 246, 46 },
		{ 5, 247, 46 },
		{ 5, 248, 46 },
		{ 5, 249, 46 },
		{ 5, 250, 46 },
		{ 5, 251, 46 },
		{ 5, 252, 46 },
		{ 5, 253, 46 },
		{ 5, 254, 46 },
		{ 5, 255, 46 },
		{ 5, 256, 46 },
		{ 5, 257, 46 },
		{ 5, 258, 46 },
		{ 5, 259, 46 },
		{ 5, 260, 46 },
		{ 5, 261, 46 },
		{ 5, 262, 46 },
		{ 5, 263, 46 },
		{ 5, 264, 46 },
		{ 5, 265, 46 },
		{ 5, 266, 46 },
		{ 5, 267, 46 },
		{ 5, 268, 46 },
		{ 5, 269, 46 },
		{ 5, 270, 46 },
		{ 5, 271, 46 },
		{ 5, 272, 46 },
		{ 5, 273, 46 },
		{ 5, 274, 46 },
		{ 5, 275, 46 },
		{ 5, 276, 46 },
		{ 5, 277, 46 },
		{ 5, 278, 46 },
		{ 5, 279, 46 },
		{ 5, 280, 46 },
		{ 5, 281, 46 },
		{ 5, 282, 46 },
		{ 5, 283, 46 },
		{ 5, 284, 46 },
		{ 5, 285, 46 },
		{ 5, 286, 46 },
		{ 5, 287, 46 },
		{ 5, 288, 46 },
		{ 6, 226, 46 },
		{ 6, 227, 46 },
		{ 6, 228, 46 },
		{ 6, 229, 46 },
		{ 6, 230, 46 },
		{ 6, 231, 46 },
		{ 6, 232, 46 },
		{ 6, 233, 46 },
		{ 6, 234, 46 },
		{ 6, 235, 46 },
		{ 6, 236, 46 },
		{ 6, 237, 46 },
		{ 6, 238, 46 },
		{ 6, 239, 46 },
		{ 6, 240, 46 },
		{ 6, 241, 46 },
		{ 6, 242, 46 },
		{ 6, 243, 46 },
		{ 6, 244, 46 },
		{ 6, 245, 46 },
		{ 6, 246, 46 },
		{ 6, 247, 46 },
		{ 6, 248, 46 },
		{ 6, 249, 46 },
		{ 6, 250, 46 },
		{ 6, 251, 46 },
		{ 6, 252, 46 },
		{ 6, 253, 46 },
		{ 6, 254, 46 },
		{ 6, 255, 46 },
		{ 6, 256, 46 },
		{ 6, 257, 46 },
		{ 6, 258, 46 },
		{ 6, 259, 46 },
		{ 6, 260, 46 },
		{ 6, 261, 46 },
		{ 6, 262, 46 },
		{ 6, 263, 46 },
		{ 6, 264, 46 },
		{ 6, 265, 46 },
		{ 6, 266, 46 },
		{ 6, 267, 46 },
		{ 6, 268, 46 },
		{ 6, 269, 46 },
		{ 6, 270, 46 },
		{ 6, 271, 46 },
		{ 6, 272, 46 },
		{ 6, 273, 46 },
		{ 6, 274, 46 },
		{ 6, 275, 46 },
		{ 6, 276, 46 },
		{ 6, 277, 46 },
		{ 6, 278, 46 },
		{ 6, 279, 46 },
		{ 6, 280, 46 },
		{ 6, 281, 46 },
		{ 6, 282, 46 },
		{ 6, 283, 46 },
		{ 6, 284, 46 },
		{ 6, 285, 46 },
		{ 6, 286, 46 },
		{ 6, 287, 46 },
		{ 6, 288, 46 },
		{ 7, 226, 46 },
		{ 7, 227, 46 },
		{ 7, 228, 46 },
		{ 7, 229, 46 },
		{ 7, 230, 46 },
		{ 7, 231, 46 },
		{ 7, 232, 46 },
		{ 7, 233, 46 },
		{ 7, 234, 46 },
		{ 7, 235, 46 },
		{ 7, 236, 46 },
		{ 7, 237, 46 },
		{ 7, 238, 46 },
		{ 7, 239, 46 },
		{ 7, 240, 46 },
		{ 7, 241, 46 },
		{ 7, 242, 46 },
		{ 7, 243, 46 },
		{ 7, 244, 46 },
		{ 7, 245, 46 },
		{ 7, 246, 46 },
		{ 7, 247, 46 },
		{ 7, 248, 46 },
		{ 7, 249, 46 },
		{ 7, 250, 46 },
		{ 7, 251, 46 },
		{ 7, 252, 46 },
		{ 7, 253, 46 },
		{ 7, 254, 46 },
		{ 7, 255, 46 },
		{ 7, 256, 46 },
		{ 7, 257, 46 },
		{ 7, 258, 46 },
		{ 7, 259, 46 },
		{ 7, 260, 46 },
		{ 7, 261, 46 },
		{ 7, 262, 46 },
		{ 7, 263, 46 },
		{ 7, 264, 46 },
		{ 7, 265, 46 },
		{ 7, 266, 46 },
		{ 7, 267, 46 },
		{ 7, 268, 46 },
		{ 7, 269, 46 },
		{ 7, 270, 46 },
		{ 7, 271, 46 },
		{ 7, 272, 46 },
		{ 7, 273, 46 },
		{ 7, 274, 46 },
		{ 7, 275, 46 },
		{ 7, 276, 46 },
		{ 7, 277, 46 },
		{ 7, 278, 46 },
		{ 7, 279, 46 },
		{ 7, 280, 46 },
		{ 7, 281, 46 },
		{ 7, 282, 46 },
		{ 7, 283, 46 },
		{ 7, 284, 46 },
		{ 7, 285, 46 },
		{ 7, 286, 46 },
		{ 7, 287, 46 },
		{ 7, 288, 46 },
		{ 8, 226, 46 },
		{ 8, 227, 46 },
		{ 8, 228, 46 },
		{ 8, 229, 46 },
		{ 8, 230, 46 },
		{ 8, 231, 46 },
		{ 8, 232, 46 },
		{ 8, 233, 46 },
		{ 8, 234, 46 },
		{ 8, 235, 46 },
		{ 8, 236, 46 },
		{ 8, 237, 46 },
		{ 8, 238, 46 },
		{ 8, 239, 46 },
		{ 8, 240, 46 },
		{ 8, 241, 46 },
		{ 8, 242, 46 },
		{ 8, 243, 46 },
		{ 8, 244, 46 },
		{ 8, 245, 46 },
		{ 8, 246, 46 },
		{ 8, 247, 46 },
		{ 8, 248, 46 },
		{ 8, 249, 46 },
		{ 8, 250, 46 },
		{ 8, 251, 46 },
		{ 8, 252, 46 },
		{ 8, 253, 46 },
		{ 8, 254, 46 },
		{ 8, 255, 46 },
		{ 8, 256, 46 },
		{ 8, 257, 46 },
		{ 8, 258, 46 },
		{ 8, 259, 46 },
		{ 8, 260, 46 },
		{ 8, 261, 46 },
		{ 8, 262, 46 },
		{ 8, 263, 46 },
		{ 8, 264, 46 },
		{ 8, 265, 46 },
		{ 8, 266, 46 },
		{ 8, 267, 46 },
		{ 8, 268, 46 },
		{ 8, 269, 46 },
		{ 8, 270, 46 },
		{ 8, 271, 46 },
		{ 8, 272, 46 },
		{ 8, 273, 46 },
		{ 8, 274, 46 },
		{ 8, 275, 46 },
		{ 8, 276, 46 },
		{ 8, 277, 46 },
		{ 8, 278, 46 },
		{ 8, 279, 46 },
		{ 8, 280, 46 },
		{ 8, 281, 46 },
		{ 8, 282, 46 },
		{ 8, 283, 46 },
		{ 8, 284, 46 },
		{ 8, 285, 46 },
		{ 8, 286, 46 },
		{ 8, 287, 46 },
		{ 8, 288, 46 },
		{ 9, 227, 46 },
		{ 9, 228, 46 },
		{ 9, 229, 46 },
		{ 9, 230, 46 },
		{ 9, 231, 46 },
		{ 9, 232, 46 },
		{ 9, 233, 46 },
		{ 9, 234, 46 },
		{ 9, 235, 46 },
		{ 9, 236, 46 },
		{ 9, 237, 46 },
		{ 9, 238, 46 },
		{ 9, 239, 46 },
		{ 9, 240, 46 },
		{ 9, 241, 46 },
		{ 9, 242, 46 },
		{ 9, 243, 46 },
		{ 9, 244, 46 },
		{ 9, 245, 46 },
		{ 9, 246, 46 },
		{ 9, 247, 46 },
		{ 9, 248, 46 },
		{ 9, 249, 46 },
		{ 9, 250, 46 },
		{ 9, 251, 46 },
		{ 9, 252, 46 },
		{ 9, 253, 46 },
		{ 9, 254, 46 },
		{ 9, 255, 46 },
		{ 9, 256, 46 },
		{ 9, 257, 46 },
		{ 9, 258, 46 },
		{ 9, 259, 46 },
		{ 9, 260, 46 },
		{ 9, 261, 46 },
		{ 9, 262, 46 },
		{ 9, 263, 46 },
		{ 9, 264, 46 },
		{ 9, 265, 46 },
		{ 9, 266, 46 },
		{ 9, 267, 46 },
		{ 9, 268, 46 },
		{ 9, 269, 46 },
		{ 9, 270, 46 },
		{ 9, 271, 46 },
		{ 9, 272, 46 },
		{ 9, 273, 46 },
		{ 9, 274, 46 },
		{ 9, 275, 46 },
		{ 9, 276, 46 },
		{ 9, 277, 46 },
		{ 9, 278, 46 },
		{ 9, 279, 46 },
		{ 9, 280, 46 },
		{ 9, 281, 46 },
		{ 9, 282, 46 },
		{ 9, 283, 46 },
		{ 9, 284, 46 },
		{ 9, 285, 46 },
		{ 9, 286, 46 },
		{ 9, 287, 46 },
		{ 10, 227, 46 },
		{ 10, 228, 46 },
		{ 10, 229, 46 },
		{ 10, 230, 46 },
		{ 10, 231, 46 },
		{ 10, 232, 46 },
		{ 10, 233, 46 },
		{ 10, 234, 46 },
		{ 10, 235, 46 },
		{ 10, 236, 46 },
		{ 10, 237, 46 },
		{ 10, 238, 46 },
		{ 10, 239, 46 },
		{ 10, 240, 46 },
		{ 10, 241, 46 },
		{ 10, 242, 46 },
		{ 10, 243, 46 },
		{ 10, 244, 46 },
		{ 10, 245, 46 },
		{ 10, 246, 46 },
		{ 10, 247, 46 },
		{ 10, 248, 46 },
		{ 10, 249, 46 },
		{ 10, 250, 46 },
		{ 10, 251, 46 },
		{ 10, 252, 46 },
		{ 10, 253, 46 },
		{ 10, 254, 46 },
		{ 10, 255, 46 },
		{ 10, 256, 46 },
		{ 10, 257, 46 },
		{ 10, 258, 46 },
		{ 10, 259, 46 },
		{ 10, 260, 46 },
		{ 10, 261, 46 },
		{ 10, 262, 46 },
		{ 10, 263, 46 },
		{ 10, 264, 46 },
		{ 10, 265, 46 },
		{ 10, 266, 46 },
		{ 10, 267, 46 },
		{ 10, 268, 46 },
		{ 10, 269, 46 },
		{ 10, 270, 46 },
		{ 10, 271, 46 },
		{ 10, 272, 46 },
		{ 10, 273, 46 },
		{ 10, 274, 46 },
		{ 10, 275, 46 },
		{ 10, 276, 46 },
		{ 10, 277, 46 },
		{ 10, 278, 46 },
		{ 10, 279, 46 },
		{ 10, 280, 46 },
		{ 10, 281, 46 },
		{ 10, 282, 46 },
		{ 10, 283, 46 },
		{ 10, 284, 46 },
		{ 10, 285, 46 },
		{ 10, 286, 46 },
		{ 10, 287, 46 },
		{ 11, 227, 46 },
		{ 11, 228, 46 },
		{ 11, 229, 46 },
		{ 11, 230, 46 },
		{ 11, 231, 46 },
		{ 11, 232, 46 },
		{ 11, 233, 46 },
		{ 11, 234, 46 },
		{ 11, 235, 46 },
		{ 11, 236, 46 },
		{ 11, 237, 46 },
		{ 11, 238, 46 },
		{ 11, 239, 46 },
		{ 11, 240, 46 },
		{ 11, 241, 46 },
		{ 11, 242, 46 },
		{ 11, 243, 46 },
		{ 11, 244, 46 },
		{ 11, 245, 46 },
		{ 11, 246, 46 },
		{ 11, 247, 46 },
		{ 11, 248, 46 },
		{ 11, 249, 46 },
		{ 11, 250, 46 },
		{ 11, 251, 46 },
		{ 11, 252, 46 },
		{ 11, 253, 46 },
		{ 11, 254, 46 },
		{ 11, 255, 46 },
		{ 11, 256, 46 },
		{ 11, 257, 46 },
		{ 11, 258, 46 },
		{ 11, 259, 46 },
		{ 11, 260, 46 },
		{ 11, 261, 46 },
		{ 11, 262, 46 },
		{ 11, 263, 46 },
		{ 11, 264, 46 },
		{ 11, 265, 46 },
		{ 11, 266, 46 },
		{ 11, 267, 46 },
		{ 11, 268, 46 },
		{ 11, 269, 46 },
		{ 11, 270, 46 },
		{ 11, 271, 46 },
		{ 11, 272, 46 },
		{ 11, 273, 46 },
		{ 11, 274, 46 },
		{ 11, 275, 46 },
		{ 11, 276, 46 },
		{ 11, 277, 46 },
		{ 11, 278, 46 },
		{ 11, 279, 46 },
		{ 11, 280, 46 },
		{ 11, 281, 46 },
		{ 11, 282, 46 },
		{ 11, 283, 46 },
		{ 11, 284, 46 },
		{ 11, 285, 46 },
		{ 11, 286, 46 },
		{ 11, 287, 46 },
		{ 12, 227, 46 },
		{ 12, 228, 46 },
		{ 12, 229, 46 },
		{ 12, 230, 46 },
		{ 12, 231, 46 },
		{ 12, 232, 46 },
		{ 12, 233, 46 },
		{ 12, 234, 46 },
		{ 12, 235, 46 },
		{ 12, 236, 46 },
		{ 12, 237, 46 },
		{ 12, 238, 46 },
		{ 12, 239, 46 },
		{ 12, 240, 46 },
		{ 12, 241, 46 },
		{ 12, 242, 46 },
		{ 12, 243, 46 },
		{ 12, 244, 46 },
		{ 12, 245, 46 },
		{ 12, 246, 46 },
		{ 12, 247, 46 },
		{ 12, 248, 46 },
		{ 12, 249, 46 },
		{ 12, 250, 46 },
		{ 12, 251, 46 },
		{ 12, 252, 46 },
		{ 12, 253, 46 },
		{ 12, 254, 46 },
		{ 12, 255, 46 },
		{ 12, 256, 46 },
		{ 12, 257, 46 },
		{ 12, 258, 46 },
		{ 12, 259, 46 },
		{ 12, 260, 46 },
		{ 12, 261, 46 },
		{ 12, 262, 46 },
		{ 12, 263, 46 },
		{ 12, 264, 46 },
		{ 12, 265, 46 },
		{ 12, 266, 46 },
		{ 12, 267, 46 },
		{ 12, 268, 46 },
		{ 12, 269, 46 },
		{ 12, 270, 46 },
		{ 12, 271, 46 },
		{ 12, 272, 46 },
		{ 12, 273, 46 },
		{ 12, 274, 46 },
		{ 12, 275, 46 },
		{ 12, 276, 46 },
		{ 12, 277, 46 },
		{ 12, 278, 46 },
		{ 12, 279, 46 },
		{ 12, 280, 46 },
		{ 12, 281, 46 },
		{ 12, 282, 46 },
		{ 12, 283, 46 },
		{ 12, 284, 46 },
		{ 12, 285, 46 },
		{ 12, 286, 46 },
		{ 12, 287, 46 },
		{ 13, 227, 46 },
		{ 13, 228, 46 },
		{ 13, 229, 46 },
		{ 13, 230, 46 },
		{ 13, 231, 46 },
		{ 13, 232, 46 },
		{ 13, 233, 46 },
		{ 13, 234, 46 },
		{ 13, 235, 46 },
		{ 13, 236, 46 },
		{ 13, 237, 46 },
		{ 13, 238, 46 },
		{ 13, 239, 46 },
		{ 13, 240, 46 },
		{ 13, 241, 46 },
		{ 13, 242, 46 },
		{ 13, 243, 46 },
		{ 13, 244, 46 },
		{ 13, 245, 46 },
		{ 13, 246, 46 },
		{ 13, 247, 46 },
		{ 13, 248, 46 },
		{ 13, 249, 46 },
		{ 13, 250, 46 },
		{ 13, 251, 46 },
		{ 13, 252, 46 },
		{ 13, 253, 46 },
		{ 13, 254, 46 },
		{ 13, 255, 46 },
		{ 13, 256, 46 },
		{ 13, 257, 46 },
		{ 13, 258, 46 },
		{ 13, 259, 46 },
		{ 13, 260, 46 },
		{ 13, 261, 46 },
		{ 13, 262, 46 },
		{ 13, 263, 46 },
		{ 13, 264, 46 },
		{ 13, 265, 46 },
		{ 13, 266, 46 },
		{ 13, 267, 46 },
		{ 13, 268, 46 },
		{ 13, 269, 46 },
		{ 13, 270, 46 },
		{ 13, 271, 46 },
		{ 13, 272, 46 },
		{ 13, 273, 46 },
		{ 13, 274, 46 },
		{ 13, 275, 46 },
		{ 13, 276, 46 },
		{ 13, 277, 46 },
		{ 13, 278, 46 },
		{ 13, 279, 46 },
		{ 13, 280, 46 },
		{ 13, 281, 46 },
		{ 13, 282, 46 },
		{ 13, 283, 46 },
		{ 13, 284, 46 },
		{ 13, 285, 46 },
		{ 13, 286, 46 },
		{ 13, 287, 46 },
		{ 14, 227, 46 },
		{ 14, 228, 46 },
		{ 14, 229, 46 },
		{ 14, 230, 46 },
		{ 14, 231, 46 },
		{ 14, 232, 46 },
		{ 14, 233, 46 },
		{ 14, 234, 46 },
		{ 14, 235, 46 },
		{ 14, 236, 46 },
		{ 14, 237, 46 },
		{ 14, 238, 46 },
		{ 14, 239, 46 },
		{ 14, 240, 46 },
		{ 14, 241, 46 },
		{ 14, 242, 46 },
		{ 14, 243, 46 },
		{ 14, 244, 46 },
		{ 14, 245, 46 },
		{ 14, 246, 46 },
		{ 14, 247, 46 },
		{ 14, 248, 46 },
		{ 14, 249, 46 },
		{ 14, 250, 46 },
		{ 14, 251, 46 },
		{ 14, 252, 46 },
		{ 14, 253, 46 },
		{ 14, 254, 46 },
		{ 14, 255, 46 },
		{ 14, 256, 46 },
		{ 14, 257, 46 },
		{ 14, 258, 46 },
		{ 14, 259, 46 },
		{ 14, 260, 46 },
		{ 14, 261, 46 },
		{ 14, 262, 46 },
		{ 14, 263, 46 },
		{ 14, 264, 46 },
		{ 14, 265, 46 },
		{ 14, 266, 46 },
		{ 14, 267, 46 },
		{ 14, 268, 46 },
		{ 14, 269, 46 },
		{ 14, 270, 46 },
		{ 14, 271, 46 },
		{ 14, 272, 46 },
		{ 14, 273, 46 },
		{ 14, 274, 46 },
		{ 14, 275, 46 },
		{ 14, 276, 46 },
		{ 14, 277, 46 },
		{ 14, 278, 46 },
		{ 14, 279, 46 },
		{ 14, 280, 46 },
		{ 14, 281, 46 },
		{ 14, 282, 46 },
		{ 14, 283, 46 },
		{ 14, 284, 46 },
		{ 14, 285, 46 },
		{ 14, 286, 46 },
		{ 14, 287, 46 },
		{ 15, 227, 46 },
		{ 15, 228, 46 },
		{ 15, 229, 46 },
		{ 15, 230, 46 },
		{ 15, 231, 46 },
		{ 15, 232, 46 },
		{ 15, 233, 46 },
		{ 15, 234, 46 },
		{ 15, 235, 46 },
		{ 15, 236, 46 },
		{ 15, 237, 46 },
		{ 15, 238, 46 },
		{ 15, 239, 46 },
		{ 15, 240, 46 },
		{ 15, 241, 46 },
		{ 15, 242, 46 },
		{ 15, 243, 46 },
		{ 15, 244, 46 },
		{ 15, 245, 46 },
		{ 15, 246, 46 },
		{ 15, 247, 46 },
		{ 15, 248, 46 },
		{ 15, 249, 46 },
		{ 15, 250, 46 },
		{ 15, 251, 46 },
		{ 15, 252, 46 },
		{ 15, 253, 46 },
		{ 15, 254, 46 },
		{ 15, 255, 46 },
		{ 15, 256, 46 },
		{ 15, 257, 46 },
		{ 15, 258, 46 },
		{ 15, 259, 46 },
		{ 15, 260, 46 },
		{ 15, 261, 46 },
		{ 15, 262, 46 },
		{ 15, 263, 46 },
		{ 15, 264, 46 },
		{ 15, 265, 46 },
		{ 15, 266, 46 },
		{ 15, 267, 46 },
		{ 15, 268, 46 },
		{ 15, 269, 46 },
		{ 15, 270, 46 },
		{ 15, 271, 46 },
		{ 15, 272, 46 },
		{ 15, 273, 46 },
		{ 15, 274, 46 },
		{ 15, 275, 46 },
		{ 15, 276, 46 },
		{ 15, 277, 46 },
		{ 15, 278, 46 },
		{ 15, 279, 46 },
		{ 15, 280, 46 },
		{ 15, 281, 46 },
		{ 15, 282, 46 },
		{ 15, 283, 46 },
		{ 15, 284, 46 },
		{ 15, 285, 46 },
		{ 15, 286, 46 },
		{ 15, 287, 46 },
		{ 16, 227, 46 },
		{ 16, 228, 46 },
		{ 16, 229, 46 },
		{ 16, 230, 46 },
		{ 16, 231, 46 },
		{ 16, 232, 46 },
		{ 16, 233, 46 },
		{ 16, 234, 46 },
		{ 16, 235, 46 },
		{ 16, 236, 46 },
		{ 16, 237, 46 },
		{ 16, 238, 46 },
		{ 16, 239, 46 },
		{ 16, 240, 46 },
		{ 16, 241, 46 },
		{ 16, 242, 46 },
		{ 16, 243, 46 },
		{ 16, 244, 46 },
		{ 16, 245, 46 },
		{ 16, 246, 46 },
		{ 16, 247, 46 },
		{ 16, 248, 46 },
		{ 16, 249, 46 },
		{ 16, 250, 46 },
		{ 16, 251, 46 },
		{ 16, 252, 46 },
		{ 16, 253, 46 },
		{ 16, 254, 46 },
		{ 16, 255, 46 },
		{ 16, 256, 46 },
		{ 16, 257, 46 },
		{ 16, 258, 46 },
		{ 16, 259, 46 },
		{ 16, 260, 46 },
		{ 16, 261, 46 },
		{ 16, 262, 46 },
		{ 16, 263, 46 },
		{ 16, 264, 46 },
		{ 16, 265, 46 },
		{ 16, 266, 46 },
		{ 16, 267, 46 },
		{ 16, 268, 46 },
		{ 16, 269, 46 },
		{ 16, 270, 46 },
		{ 16, 271, 46 },
		{ 16, 272, 46 },
		{ 16, 273, 46 },
		{ 16, 274, 46 },
		{ 16, 275, 46 },
		{ 16, 276, 46 },
		{ 16, 277, 46 },
		{ 16, 278, 46 },
		{ 16, 279, 46 },
		{ 16, 280, 46 },
		{ 16, 281, 46 },
		{ 16, 282, 46 },
		{ 16, 283, 46 },
		{ 16, 284, 46 },
		{ 16, 285, 46 },
		{ 16, 286, 46 },
		{ 16, 287, 46 },
		{ 17, 228, 46 },
		{ 17, 229, 46 },
		{ 17, 230, 46 },
		{ 17, 231, 46 },
		{ 17, 232, 46 },
		{ 17, 233, 46 },
		{ 17, 234, 46 },
		{ 17, 235, 46 },
		{ 17, 236, 46 },
		{ 17, 237, 46 },
		{ 17, 238, 46 },
		{ 17, 239, 46 },
		{ 17, 240, 46 },
		{ 17, 241, 46 },
		{ 17, 242, 46 },
		{ 17, 243, 46 },
		{ 17, 244, 46 },
		{ 17, 245, 46 },
		{ 17, 246, 46 },
		{ 17, 247, 46 },
		{ 17, 248, 46 },
		{ 17, 249, 46 },
		{ 17, 250, 46 },
		{ 17, 251, 46 },
		{ 17, 252, 46 },
		{ 17, 253, 46 },
		{ 17, 254, 46 },
		{ 17, 255, 46 },
		{ 17, 256, 46 },
		{ 17, 257, 46 },
		{ 17, 258, 46 },
		{ 17, 259, 46 },
		{ 17, 260, 46 },
		{ 17, 261, 46 },
		{ 17, 262, 46 },
		{ 17, 263, 46 },
		{ 17, 264, 46 },
		{ 17, 265, 46 },
		{ 17, 266, 46 },
		{ 17, 267, 46 },
		{ 17, 268, 46 },
		{ 17, 269, 46 },
		{ 17, 270, 46 },
		{ 17, 271, 46 },
		{ 17, 272, 46 },
		{ 17, 273, 46 },
		{ 17, 274, 46 },
		{ 17, 275, 46 },
		{ 17, 276, 46 },
		{ 17, 277, 46 },
		{ 17, 278, 46 },
		{ 17, 279, 46 },
		{ 17, 280, 46 },
		{ 17, 281, 46 },
		{ 17, 282, 46 },
		{ 17, 283, 46 },
		{ 17, 284, 46 },
		{ 17, 285, 46 },
		{ 17, 286, 46 },
		{ 18, 228, 46 },
		{ 18, 229, 46 },
		{ 18, 230, 46 },
		{ 18, 231, 46 },
		{ 18, 232, 46 },
		{ 18, 233, 46 },
		{ 18, 234, 46 },
		{ 18, 235, 46 },
		{ 18, 236, 46 },
		{ 18, 237, 46 },
		{ 18, 238, 46 },
		{ 18, 239, 46 },
		{ 18, 240, 46 },
		{ 18, 241, 46 },
		{ 18, 242, 46 },
		{ 18, 243, 46 },
		{ 18, 244, 46 },
		{ 18, 245, 46 },
		{ 18, 246, 46 },
		{ 18, 247, 46 },
		{ 18, 248, 46 },
		{ 18, 249, 46 },
		{ 18, 250, 46 },
		{ 18, 251, 46 },
		{ 18, 252, 46 },
		{ 18, 253, 46 },
		{ 18, 254, 46 },
		{ 18, 255, 46 },
		{ 18, 256, 46 },
		{ 18, 257, 46 },
		{ 18, 258, 46 },
		{ 18, 259, 46 },
		{ 18, 260, 46 },
		{ 18, 261, 46 },
		{ 18, 262, 46 },
		{ 18, 263, 46 },
		{ 18, 264, 46 },
		{ 18, 265, 46 },
		{ 18, 266, 46 },
		{ 18, 267, 46 },
		{ 18, 268, 46 },
		{ 18, 269, 46 },
		{ 18, 270, 46 },
		{ 18, 271, 46 },
		{ 18, 272, 46 },
		{ 18, 273, 46 },
		{ 18, 274, 46 },
		{ 18, 275, 46 },
		{ 18, 276, 46 },
		{ 18, 277, 46 },
		{ 18, 278, 46 },
		{ 18, 279, 46 },
		{ 18, 280, 46 },
		{ 18, 281, 46 },
		{ 18, 282, 46 },
		{ 18, 283, 46 },
		{ 18, 284, 46 },
		{ 18, 285, 46 },
		{ 18, 286, 46 },
		{ 19, 228, 46 },
		{ 19, 229, 46 },
		{ 19, 230, 46 },
		{ 19, 231, 46 },
		{ 19, 232, 46 },
		{ 19, 233, 46 },
		{ 19, 234, 46 },
		{ 19, 235, 46 },
		{ 19, 236, 46 },
		{ 19, 237, 46 },
		{ 19, 238, 46 },
		{ 19, 239, 46 },
		{ 19, 240, 46 },
		{ 19, 241, 46 },
		{ 19, 242, 46 },
		{ 19, 243, 46 },
		{ 19, 244, 46 },
		{ 19, 245, 46 },
		{ 19, 246, 46 },
		{ 19, 247, 46 },
		{ 19, 248, 46 },
		{ 19, 249, 46 },
		{ 19, 250, 46 },
		{ 19, 251, 46 },
		{ 19, 252, 46 },
		{ 19, 253, 46 },
		{ 19, 254, 46 },
		{ 19, 255, 46 },
		{ 19, 256, 46 },
		{ 19, 257, 46 },
		{ 19, 258, 46 },
		{ 19, 259, 46 },
		{ 19, 260, 46 },
		{ 19, 261, 46 },
		{ 19, 262, 46 },
		{ 19, 263, 46 },
		{ 19, 264, 46 },
		{ 19, 265, 46 },
		{ 19, 266, 46 },
		{ 19, 267, 46 },
		{ 19, 268, 46 },
		{ 19, 269, 46 },
		{ 19, 270, 46 },
		{ 19, 271, 46 },
		{ 19, 272, 46 },
		{ 19, 273, 46 },
		{ 19, 274, 46 },
		{ 19, 275, 46 },
		{ 19, 276, 46 },
		{ 19, 277, 46 },
		{ 19, 278, 46 },
		{ 19, 279, 46 },
		{ 19, 280, 46 },
		{ 19, 281, 46 },
		{ 19, 282, 46 },
		{ 19, 283, 46 },
		{ 19, 284, 46 },
		{ 19, 285, 46 },
		{ 19, 286, 46 },
		{ 20, 228, 46 },
		{ 20, 229, 46 },
		{ 20, 230, 46 },
		{ 20, 231, 46 },
		{ 20, 232, 46 },
		{ 20, 233, 46 },
		{ 20, 234, 46 },
		{ 20, 235, 46 },
		{ 20, 236, 46 },
		{ 20, 237, 46 },
		{ 20, 238, 46 },
		{ 20, 239, 46 },
		{ 20, 240, 46 },
		{ 20, 241, 46 },
		{ 20, 242, 46 },
		{ 20, 243, 46 },
		{ 20, 244, 46 },
		{ 20, 245, 46 },
		{ 20, 246, 46 },
		{ 20, 247, 46 },
		{ 20, 248, 46 },
		{ 20, 249, 46 },
		{ 20, 250, 46 },
		{ 20, 251, 46 },
		{ 20, 252, 46 },
		{ 20, 253, 46 },
		{ 20, 254, 46 },
		{ 20, 255, 46 },
		{ 20, 256, 46 },
		{ 20, 257, 46 },
		{ 20, 258, 46 },
		{ 20, 259, 46 },
		{ 20, 260, 46 },
		{ 20, 261, 46 },
		{ 20, 262, 46 },
		{ 20, 263, 46 },
		{ 20, 264, 46 },
		{ 20, 265, 46 },
		{ 20, 266, 46 },
		{ 20, 267, 46 },
		{ 20, 268, 46 },
		{ 20, 269, 46 },
		{ 20, 270, 46 },
		{ 20, 271, 46 },
		{ 20, 272, 46 },
		{ 20, 273, 46 },
		{ 20, 274, 46 },
		{ 20, 275, 46 },
		{ 20, 276, 46 },
		{ 20, 277, 46 },
		{ 20, 278, 46 },
		{ 20, 279, 46 },
		{ 20, 280, 46 },
		{ 20, 281, 46 },
		{ 20, 282, 46 },
		{ 20, 283, 46 },
		{ 20, 284, 46 },
		{ 20, 285, 46 },
		{ 20, 286, 46 },
		{ 21, 228, 46 },
		{ 21, 229, 46 },
		{ 21, 230, 46 },
		{ 21, 231, 46 },
		{ 21, 232, 46 },
		{ 21, 233, 46 },
		{ 21, 234, 46 },
		{ 21, 235, 46 },
		{ 21, 236, 46 },
		{ 21, 237, 46 },
		{ 21, 238, 46 },
		{ 21, 239, 46 },
		{ 21, 240, 46 },
		{ 21, 241, 46 },
		{ 21, 242, 46 },
		{ 21, 243, 46 },
		{ 21, 244, 46 },
		{ 21, 245, 46 },
		{ 21, 246, 46 },
		{ 21, 247, 46 },
		{ 21, 248, 46 },
		{ 21, 249, 46 },
		{ 21, 250, 46 },
		{ 21, 251, 46 },
		{ 21, 252, 46 },
		{ 21, 253, 46 },
		{ 21, 254, 46 },
		{ 21, 255, 46 },
		{ 21, 256, 46 },
		{ 21, 257, 46 },
		{ 21, 258, 46 },
		{ 21, 259, 46 },
		{ 21, 260, 46 },
		{ 21, 261, 46 },
		{ 21, 262, 46 },
		{ 21, 263, 46 },
		{ 21, 264, 46 },
		{ 21, 265, 46 },
		{ 21, 266, 46 },
		{ 21, 267, 46 },
		{ 21, 268, 46 },
		{ 21, 269, 46 },
		{ 21, 270, 46 },
		{ 21, 271, 46 },
		{ 21, 272, 46 },
		{ 21, 273, 46 },
		{ 21, 274, 46 },
		{ 21, 275, 46 },
		{ 21, 276, 46 },
		{ 21, 277, 46 },
		{ 21, 278, 46 },
		{ 21, 279, 46 },
		{ 21, 280, 46 },
		{ 21, 281, 46 },
		{ 21, 282, 46 },
		{ 21, 283, 46 },
		{ 21, 284, 46 },
		{ 21, 285, 46 },
		{ 21, 286, 46 },
		{ 22, 228, 46 },
		{ 22, 229, 46 },
		{ 22, 230, 46 },
		{ 22, 231, 46 },
		{ 22, 232, 46 },
		{ 22, 233, 46 },
		{ 22, 234, 46 },
		{ 22, 235, 46 },
		{ 22, 236, 46 },
		{ 22, 237, 46 },
		{ 22, 238, 46 },
		{ 22, 239, 46 },
		{ 22, 240, 46 },
		{ 22, 241, 46 },
		{ 22, 242, 46 },
		{ 22, 243, 46 },
		{ 22, 244, 46 },
		{ 22, 245, 46 },
		{ 22, 246, 46 },
		{ 22, 247, 46 },
		{ 22, 248, 46 },
		{ 22, 249, 46 },
		{ 22, 250, 46 },
		{ 22, 251, 46 },
		{ 22, 252, 46 },
		{ 22, 253, 46 },
		{ 22, 254, 46 },
		{ 22, 255, 46 },
		{ 22, 256, 46 },
		{ 22, 257, 46 },
		{ 22, 258, 46 },
		{ 22, 259, 46 },
		{ 22, 260, 46 },
		{ 22, 261, 46 },
		{ 22, 262, 46 },
		{ 22, 263, 46 },
		{ 22, 264, 46 },
		{ 22, 265, 46 },
		{ 22, 266, 46 },
		{ 22, 267, 46 },
		{ 22, 268, 46 },
		{ 22, 269, 46 },
		{ 22, 270, 46 },
		{ 22, 271, 46 },
		{ 22, 272, 46 },
		{ 22, 273, 46 },
		{ 22, 274, 46 },
		{ 22, 275, 46 },
		{ 22, 276, 46 },
		{ 22, 277, 46 },
		{ 22, 278, 46 },
		{ 22, 279, 46 },
		{ 22, 280, 46 },
		{ 22, 281, 46 },
		{ 22, 282, 46 },
		{ 22, 283, 46 },
		{ 22, 284, 46 },
		{ 22, 285, 46 },
		{ 22, 286, 46 },
		{ 23, 228, 46 },
		{ 23, 229, 46 },
		{ 23, 230, 46 },
		{ 23, 231, 46 },
		{ 23, 232, 46 },
		{ 23, 233, 46 },
		{ 23, 234, 46 },
		{ 23, 235, 46 },
		{ 23, 236, 46 },
		{ 23, 237, 46 },
		{ 23, 238, 46 },
		{ 23, 239, 46 },
		{ 23, 240, 46 },
		{ 23, 241, 46 },
		{ 23, 242, 46 },
		{ 23, 243, 46 },
		{ 23, 244, 46 },
		{ 23, 245, 46 },
		{ 23, 246, 46 },
		{ 23, 247, 46 },
		{ 23, 248, 46 },
		{ 23, 249, 46 },
		{ 23, 250, 46 },
		{ 23, 251, 46 },
		{ 23, 252, 46 },
		{ 23, 253, 46 },
		{ 23, 254, 46 },
		{ 23, 255, 46 },
		{ 23, 256, 46 },
		{ 23, 257, 46 },
		{ 23, 258, 46 },
		{ 23, 259, 46 },
		{ 23, 260, 46 },
		{ 23, 261, 46 },
		{ 23, 262, 46 },
		{ 23, 263, 46 },
		{ 23, 264, 46 },
		{ 23, 265, 46 },
		{ 23, 266, 46 },
		{ 23, 267, 46 },
		{ 23, 268, 46 },
		{ 23, 269, 46 },
		{ 23, 270, 46 },
		{ 23, 271, 46 },
		{ 23, 272, 46 },
		{ 23, 273, 46 },
		{ 23, 274, 46 },
		{ 23, 275, 46 },
		{ 23, 276, 46 },
		{ 23, 277, 46 },
		{ 23, 278, 46 },
		{ 23, 279, 46 },
		{ 23, 280, 46 },
		{ 23, 281, 46 },
		{ 23, 282, 46 },
		{ 23, 283, 46 },
		{ 23, 284, 46 },
		{ 23, 285, 46 },
		{ 23, 286, 46 },
		{ 24, 228, 46 },
		{ 24, 229, 46 },
		{ 24, 230, 46 },
		{ 24, 231, 46 },
		{ 24, 232, 46 },
		{ 24, 233, 46 },
		{ 24, 234, 46 },
		{ 24, 235, 46 },
		{ 24, 236, 46 },
		{ 24, 237, 46 },
		{ 24, 238, 46 },
		{ 24, 239, 46 },
		{ 24, 240, 46 },
		{ 24, 241, 46 },
		{ 24, 242, 46 },
		{ 24, 243, 46 },
		{ 24, 244, 46 },
		{ 24, 245, 46 },
		{ 24, 246, 46 },
		{ 24, 247, 46 },
		{ 24, 248, 46 },
		{ 24, 249, 46 },
		{ 24, 250, 46 },
		{ 24, 251, 46 },
		{ 24, 252, 46 },
		{ 24, 253, 46 },
		{ 24, 254, 46 },
		{ 24, 255, 46 },
		{ 24, 256, 46 },
		{ 24, 257, 46 },
		{ 24, 258, 46 },
		{ 24, 259, 46 },
		{ 24, 260, 46 },
		{ 24, 261, 46 },
		{ 24, 262, 46 },
		{ 24, 263, 46 },
		{ 24, 264, 46 },
		{ 24, 265, 46 },
		{ 24, 266, 46 },
		{ 24, 267, 46 },
		{ 24, 268, 46 },
		{ 24, 269, 46 },
		{ 24, 270, 46 },
		{ 24, 271, 46 },
		{ 24, 272, 46 },
		{ 24, 273, 46 },
		{ 24, 274, 46 },
		{ 24, 275, 46 },
		{ 24, 276, 46 },
		{ 24, 277, 46 },
		{ 24, 278, 46 },
		{ 24, 279, 46 },
		{ 24, 280, 46 },
		{ 24, 281, 46 },
		{ 24, 282, 46 },
		{ 24, 283, 46 },
		{ 24, 284, 46 },
		{ 24, 285, 46 },
		{ 24, 286, 46 },
		{ 25, 229, 46 },
		{ 25, 230, 46 },
		{ 25, 231, 46 },
		{ 25, 232, 46 },
		{ 25, 233, 46 },
		{ 25, 234, 46 },
		{ 25, 235, 46 },
		{ 25, 236, 46 },
		{ 25, 237, 46 },
		{ 25, 238, 46 },
		{ 25, 239, 46 },
		{ 25, 240, 46 },
		{ 25, 241, 46 },
		{ 25, 242, 46 },
		{ 25, 243, 46 },
		{ 25, 244, 46 },
		{ 25, 245, 46 },
		{ 25, 246, 46 },
		{ 25, 247, 46 },
		{ 25, 248, 46 },
		{ 25, 249, 46 },
		{ 25, 250, 46 },
		{ 25, 251, 46 },
		{ 25, 252, 46 },
		{ 25, 253, 46 },
		{ 25, 254, 46 },
		{ 25, 255, 46 },
		{ 25, 256, 46 },
		{ 25, 257, 46 },
		{ 25, 258, 46 },
		{ 25, 259, 46 },
		{ 25, 260, 46 },
		{ 25, 261, 46 },
		{ 25, 262, 46 },
		{ 25, 263, 46 },
		{ 25, 264, 46 },
		{ 25, 265, 46 },
		{ 25, 266, 46 },
		{ 25, 267, 46 },
		{ 25, 268, 46 },
		{ 25, 269, 46 },
		{ 25, 270, 46 },
		{ 25, 271, 46 },
		{ 25, 272, 46 },
		{ 25, 273, 46 },
		{ 25, 274, 46 },
		{ 25, 275, 46 },
		{ 25, 276, 46 },
		{ 25, 277, 46 },
		{ 25, 278, 46 },
		{ 25, 279, 46 },
		{ 25, 280, 46 },
		{ 25, 281, 46 },
		{ 25, 282, 46 },
		{ 25, 283, 46 },
		{ 25, 284, 46 },
		{ 25, 285, 46 },
		{ 26, 229, 46 },
		{ 26, 230, 46 },
		{ 26, 231, 46 },
		{ 26, 232, 46 },
		{ 26, 233, 46 },
		{ 26, 234, 46 },
		{ 26, 235, 46 },
		{ 26, 236, 46 },
		{ 26, 237, 46 },
		{ 26, 238, 46 },
		{ 26, 239, 46 },
		{ 26, 240, 46 },
		{ 26, 241, 46 },
		{ 26, 242, 46 },
		{ 26, 243, 46 },
		{ 26, 244, 46 },
		{ 26, 245, 46 },
		{ 26, 246, 46 },
		{ 26, 247, 46 },
		{ 26, 248, 46 },
		{ 26, 249, 46 },
		{ 26, 250, 46 },
		{ 26, 251, 46 },
		{ 26, 252, 46 },
		{ 26, 253, 46 },
		{ 26, 254, 46 },
		{ 26, 255, 46 },
		{ 26, 256, 46 },
		{ 26, 257, 46 },
		{ 26, 258, 46 },
		{ 26, 259, 46 },
		{ 26, 260, 46 },
		{ 26, 261, 46 },
		{ 26, 262, 46 },
		{ 26, 263, 46 },
		{ 26, 264, 46 },
		{ 26, 265, 46 },
		{ 26, 266, 46 },
		{ 26, 267, 46 },
		{ 26, 268, 46 },
		{ 26, 269, 46 },
		{ 26, 270, 46 },
		{ 26, 271, 46 },
		{ 26, 272, 46 },
		{ 26, 273, 46 },
		{ 26, 274, 46 },
		{ 26, 275, 46 },
		{ 26, 276, 46 },
		{ 26, 277, 46 },
		{ 26, 278, 46 },
		{ 26, 279, 46 },
		{ 26, 280, 46 },
		{ 26, 281, 46 },
		{ 26, 282, 46 },
		{ 26, 283, 46 },
		{ 26, 284, 46 },
		{ 26, 285, 46 },
		{ 27, 229, 46 },
		{ 27, 230, 46 },
		{ 27, 231, 46 },
		{ 27, 232, 46 },
		{ 27, 233, 46 },
		{ 27, 234, 46 },
		{ 27, 235, 46 },
		{ 27, 236, 46 },
		{ 27, 237, 46 },
		{ 27, 238, 46 },
		{ 27, 239, 46 },
		{ 27, 240, 46 },
		{ 27, 241, 46 },
		{ 27, 242, 46 },
		{ 27, 243, 46 },
		{ 27, 244, 46 },
		{ 27, 245, 46 },
		{ 27, 246, 46 },
		{ 27, 247, 46 },
		{ 27, 248, 46 },
		{ 27, 249, 46 },
		{ 27, 250, 46 },
		{ 27, 251, 46 },
		{ 27, 252, 46 },
		{ 27, 253, 46 },
		{ 27, 254, 46 },
		{ 27, 255, 46 },
		{ 27, 256, 46 },
		{ 27, 257, 46 },
		{ 27, 258, 46 },
		{ 27, 259, 46 },
		{ 27, 260, 46 },
		{ 27, 261, 46 },
		{ 27, 262, 46 },
		{ 27, 263, 46 },
		{ 27, 264, 46 },
		{ 27, 265, 46 },
		{ 27, 266, 46 },
		{ 27, 267, 46 },
		{ 27, 268, 46 },
		{ 27, 269, 46 },
		{ 27, 270, 46 },
		{ 27, 271, 46 },
		{ 27, 272, 46 },
		{ 27, 273, 46 },
		{ 27, 274, 46 },
		{ 27, 275, 46 },
		{ 27, 276, 46 },
		{ 27, 277, 46 },
		{ 27, 278, 46 },
		{ 27, 279, 46 },
		{ 27, 280, 46 },
		{ 27, 281, 46 },
		{ 27, 282, 46 },
		{ 27, 283, 46 },
		{ 27, 284, 46 },
		{ 27, 285, 46 },
		{ 28, 229, 46 },
		{ 28, 230, 46 },
		{ 28, 231, 46 },
		{ 28, 232, 46 },
		{ 28, 233, 46 },
		{ 28, 234, 46 },
		{ 28, 235, 46 },
		{ 28, 236, 46 },
		{ 28, 237, 46 },
		{ 28, 238, 46 },
		{ 28, 239, 46 },
		{ 28, 240, 46 },
		{ 28, 241, 46 },
		{ 28, 242, 46 },
		{ 28, 243, 46 },
		{ 28, 244, 46 },
		{ 28, 245, 46 },
		{ 28, 246, 46 },
		{ 28, 247, 46 },
		{ 28, 248, 46 },
		{ 28, 249, 46 },
		{ 28, 250, 46 },
		{ 28, 251, 46 },
		{ 28, 252, 46 },
		{ 28, 253, 46 },
		{ 28, 254, 46 },
		{ 28, 255, 46 },
		{ 28, 256, 46 },
		{ 28, 257, 46 },
		{ 28, 258, 46 },
		{ 28, 259, 46 },
		{ 28, 260, 46 },
		{ 28, 261, 46 },
		{ 28, 262, 46 },
		{ 28, 263, 46 },
		{ 28, 264, 46 },
		{ 28, 265, 46 },
		{ 28, 266, 46 },
		{ 28, 267, 46 },
		{ 28, 268, 46 },
		{ 28, 269, 46 },
		{ 28, 270, 46 },
		{ 28, 271, 46 },
		{ 28, 272, 46 },
		{ 28, 273, 46 },
		{ 28, 274, 46 },
		{ 28, 275, 46 },
		{ 28, 276, 46 },
		{ 28, 277, 46 },
		{ 28, 278, 46 },
		{ 28, 279, 46 },
		{ 28, 280, 46 },
		{ 28, 281, 46 },
		{ 28, 282, 46 },
		{ 28, 283, 46 },
		{ 28, 284, 46 },
		{ 28, 285, 46 },
		{ 29, 229, 46 },
		{ 29, 230, 46 },
		{ 29, 231, 46 },
		{ 29, 232, 46 },
		{ 29, 233, 46 },
		{ 29, 234, 46 },
		{ 29, 235, 46 },
		{ 29, 236, 46 },
		{ 29, 237, 46 },
		{ 29, 238, 46 },
		{ 29, 239, 46 },
		{ 29, 240, 46 },
		{ 29, 241, 46 },
		{ 29, 242, 46 },
		{ 29, 243, 46 },
		{ 29, 244, 46 },
		{ 29, 245, 46 },
		{ 29, 246, 46 },
		{ 29, 247, 46 },
		{ 29, 248, 46 },
		{ 29, 249, 46 },
		{ 29, 250, 46 },
		{ 29, 251, 46 },
		{ 29, 252, 46 },
		{ 29, 253, 46 },
		{ 29, 254, 46 },
		{ 29, 255, 46 },
		{ 29, 256, 46 },
		{ 29, 257, 46 },
		{ 29, 258, 46 },
		{ 29, 259, 46 },
		{ 29, 260, 46 },
		{ 29, 261, 46 },
		{ 29, 262, 46 },
		{ 29, 263, 46 },
		{ 29, 264, 46 },
		{ 29, 265, 46 },
		{ 29, 266, 46 },
		{ 29, 267, 46 },
		{ 29, 268, 46 },
		{ 29, 269, 46 },
		{ 29, 270, 46 },
		{ 29, 271, 46 },
		{ 29, 272, 46 },
		{ 29, 273, 46 },
		{ 29, 274, 46 },
		{ 29, 275, 46 },
		{ 29, 276, 46 },
		{ 29, 277, 46 },
		{ 29, 278, 46 },
		{ 29, 279, 46 },
		{ 29, 280, 46 },
		{ 29, 281, 46 },
		{ 29, 282, 46 },
		{ 29, 283, 46 },
		{ 29, 284, 46 },
		{ 29, 285, 46 },
		{ 30, 229, 46 },
		{ 30, 230, 46 },
		{ 30, 231, 46 },
		{ 30, 232, 46 },
		{ 30, 233, 46 },
		{ 30, 234, 46 },
		{ 30, 235, 46 },
		{ 30, 236, 46 },
		{ 30, 237, 46 },
		{ 30, 238, 46 },
		{ 30, 239, 46 },
		{ 30, 240, 46 },
		{ 30, 241, 46 },
		{ 30, 242, 46 },
		{ 30, 243, 46 },
		{ 30, 244, 46 },
		{ 30, 245, 46 },
		{ 30, 246, 46 },
		{ 30, 247, 46 },
		{ 30, 248, 46 },
		{ 30, 249, 46 },
		{ 30, 250, 46 },
		{ 30, 251, 46 },
		{ 30, 252, 46 },
		{ 30, 253, 46 },
		{ 30, 254, 46 },
		{ 30, 255, 46 },
		{ 30, 256, 46 },
		{ 30, 257, 46 },
		{ 30, 258, 46 },
		{ 30, 259, 46 },
		{ 30, 260, 46 },
		{ 30, 261, 46 },
		{ 30, 262, 46 },
		{ 30, 263, 46 },
		{ 30, 264, 46 },
		{ 30, 265, 46 },
		{ 30, 266, 46 },
		{ 30, 267, 46 },
		{ 30, 268, 46 },
		{ 30, 269, 46 },
		{ 30, 270, 46 },
		{ 30, 271, 46 },
		{ 30, 272, 46 },
		{ 30, 273, 46 },
		{ 30, 274, 46 },
		{ 30, 275, 46 },
		{ 30, 276, 46 },
		{ 30, 277, 46 },
		{ 30, 278, 46 },
		{ 30, 279, 46 },
		{ 30, 280, 46 },
		{ 30, 281, 46 },
		{ 30, 282, 46 },
		{ 30, 283, 46 },
		{ 30, 284, 46 },
		{ 30, 285, 46 },
		{ 31, 229, 46 },
		{ 31, 230, 46 },
		{ 31, 231, 46 },
		{ 31, 232, 46 },
		{ 31, 233, 46 },
		{ 31, 234, 46 },
		{ 31, 235, 46 },
		{ 31, 236, 46 },
		{ 31, 237, 46 },
		{ 31, 238, 46 },
		{ 31, 239, 46 },
		{ 31, 240, 46 },
		{ 31, 241, 46 },
		{ 31, 242, 46 },
		{ 31, 243, 46 },
		{ 31, 244, 46 },
		{ 31, 245, 46 },
		{ 31, 246, 46 },
		{ 31, 247, 46 },
		{ 31, 248, 46 },
		{ 31, 249, 46 },
		{ 31, 250, 46 },
		{ 31, 251, 46 },
		{ 31, 252, 46 },
		{ 31, 253, 46 },
		{ 31, 254, 46 },
		{ 31, 255, 46 },
		{ 31, 256, 46 },
		{ 31, 257, 46 },
		{ 31, 258, 46 },
		{ 31, 259, 46 },
		{ 31, 260, 46 },
		{ 31, 261, 46 },
		{ 31, 262, 46 },
		{ 31, 263, 46 },
		{ 31, 264, 46 },
		{ 31, 265, 46 },
		{ 31, 266, 46 },
		{ 31, 267, 46 },
		{ 31, 268, 46 },
		{ 31, 269, 46 },
		{ 31, 270, 46 },
		{ 31, 271, 46 },
		{ 31, 272, 46 },
		{ 31, 273, 46 },
		{ 31, 274, 46 },
		{ 31, 275, 46 },
		{ 31, 276, 46 },
		{ 31, 277, 46 },
		{ 31, 278, 46 },
		{ 31, 279, 46 },
		{ 31, 280, 46 },
		{ 31, 281, 46 },
		{ 31, 282, 46 },
		{ 31, 283, 46 },
		{ 31, 284, 46 },
		{ 31, 285, 46 },
		{ 32, 229, 46 },
		{ 32, 230, 46 },
		{ 32, 231, 46 },
		{ 32, 232, 46 },
		{ 32, 233, 46 },
		{ 32, 234, 46 },
		{ 32, 235, 46 },
		{ 32, 236, 46 },
		{ 32, 237, 46 },
		{ 32, 238, 46 },
		{ 32, 239, 46 },
		{ 32, 240, 46 },
		{ 32, 241, 46 },
		{ 32, 242, 46 },
		{ 32, 243, 46 },
		{ 32, 244, 46 },
		{ 32, 245, 46 },
		{ 32, 246, 46 },
		{ 32, 247, 46 },
		{ 32, 248, 46 },
		{ 32, 249, 46 },
		{ 32, 250, 46 },
		{ 32, 251, 46 },
		{ 32, 252, 46 },
		{ 32, 253, 46 },
		{ 32, 254, 46 },
		{ 32, 255, 46 },
		{ 32, 256, 46 },
		{ 32, 257, 46 },
		{ 32, 258, 46 },
		{ 32, 259, 46 },
		{ 32, 260, 46 },
		{ 32, 261, 46 },
		{ 32, 262, 46 },
		{ 32, 263, 46 },
		{ 32, 264, 46 },
		{ 32, 265, 46 },
		{ 32, 266, 46 },
		{ 32, 267, 46 },
		{ 32, 268, 46 },
		{ 32, 269, 46 },
		{ 32, 270, 46 },
		{ 32, 271, 46 },
		{ 32, 272, 46 },
		{ 32, 273, 46 },
		{ 32, 274, 46 },
		{ 32, 275, 46 },
		{ 32, 276, 46 },
		{ 32, 277, 46 },
		{ 32, 278, 46 },
		{ 32, 279, 46 },
		{ 32, 280, 46 },
		{ 32, 281, 46 },
		{ 32, 282, 46 },
		{ 32, 283, 46 },
		{ 32, 284, 46 },
		{ 32, 285, 46 },
		{ 33, 230, 46 },
		{ 33, 231, 46 },
		{ 33, 232, 46 },
		{ 33, 233, 46 },
		{ 33, 234, 46 },
		{ 33, 235, 46 },
		{ 33, 236, 46 },
		{ 33, 237, 46 },
		{ 33, 238, 46 },
		{ 33, 239, 46 },
		{ 33, 240, 46 },
		{ 33, 241, 46 },
		{ 33, 242, 46 },
		{ 33, 243, 46 },
		{ 33, 244, 46 },
		{ 33, 245, 46 },
		{ 33, 246, 46 },
		{ 33, 247, 46 },
		{ 33, 248, 46 },
		{ 33, 249, 46 },
		{ 33, 250, 46 },
		{ 33, 251, 46 },
		{ 33, 252, 46 },
		{ 33, 253, 46 },
		{ 33, 254, 46 },
		{ 33, 255, 46 },
		{ 33, 256, 46 },
		{ 33, 257, 46 },
		{ 33, 258, 46 },
		{ 33, 259, 46 },
		{ 33, 260, 46 },
		{ 33, 261, 46 },
		{ 33, 262, 46 },
		{ 33, 263, 46 },
		{ 33, 264, 46 },
		{ 33, 265, 46 },
		{ 33, 266, 46 },
		{ 33, 267, 46 },
		{ 33, 268, 46 },
		{ 33, 269, 46 },
		{ 33, 270, 46 },
		{ 33, 271, 46 },
		{ 33, 272, 46 },
		{ 33, 273, 46 },
		{ 33, 274, 46 },
		{ 33, 275, 46 },
		{ 33, 276, 46 },
		{ 33, 277, 46 },
		{ 33, 278, 46 },
		{ 33, 279, 46 },
		{ 33, 280, 46 },
		{ 33, 281, 46 },
		{ 33, 282, 46 },
		{ 33, 283, 46 },
		{ 33, 284, 46 },
		{ 34, 230, 46 },
		{ 34, 231, 46 },
		{ 34, 232, 46 },
		{ 34, 233, 46 },
		{ 34, 234, 46 },
		{ 34, 235, 46 },
		{ 34, 236, 46 },
		{ 34, 237, 46 },
		{ 34, 238, 46 },
		{ 34, 239, 46 },
		{ 34, 240, 46 },
		{ 34, 241, 46 },
		{ 34, 242, 46 },
		{ 34, 243, 46 },
		{ 34, 244, 46 },
		{ 34, 245, 46 },
		{ 34, 246, 46 },
		{ 34, 247, 46 },
		{ 34, 248, 46 },
		{ 34, 249, 46 },
		{ 34, 250, 46 },
		{ 34, 251, 46 },
		{ 34, 252, 46 },
		{ 34, 253, 46 },
		{ 34, 254, 46 },
		{ 34, 255, 46 },
		{ 34, 256, 46 },
		{ 34, 257, 46 },
		{ 34, 258, 46 },
		{ 34, 259, 46 },
		{ 34, 260, 46 },
		{ 34, 261, 46 },
		{ 34, 262, 46 },
		{ 34, 263, 46 },
		{ 34, 264, 46 },
		{ 34, 265, 46 },
		{ 34, 266, 46 },
		{ 34, 267, 46 },
		{ 34, 268, 46 },
		{ 34, 269, 46 },
		{ 34, 270, 46 },
		{ 34, 271, 46 },
		{ 34, 272, 46 },
		{ 34, 273, 46 },
		{ 34, 274, 46 },
		{ 34, 275, 46 },
		{ 34, 276, 46 },
		{ 34, 277, 46 },
		{ 34, 278, 46 },
		{ 34, 279, 46 },
		{ 34, 280, 46 },
		{ 34, 281, 46 },
		{ 34, 282, 46 },
		{ 34, 283, 46 },
		{ 34, 284, 46 },
		{ 35, 230, 46 },
		{ 35, 231, 46 },
		{ 35, 232, 46 },
		{ 35, 233, 46 },
		{ 35, 234, 46 },
		{ 35, 235, 46 },
		{ 35, 236, 46 },
		{ 35, 237, 46 },
		{ 35, 238, 46 },
		{ 35, 239, 46 },
		{ 35, 240, 46 },
		{ 35, 241, 46 },
		{ 35, 242, 46 },
		{ 35, 243, 46 },
		{ 35, 244, 46 },
		{ 35, 245, 46 },
		{ 35, 246, 46 },
		{ 35, 247, 46 },
		{ 35, 248, 46 },
		{ 35, 249, 46 },
		{ 35, 250, 46 },
		{ 35, 251, 46 },
		{ 35, 252, 46 },
		{ 35, 253, 46 },
		{ 35, 254, 46 },
		{ 35, 255, 46 },
		{ 35, 256, 46 },
		{ 35, 257, 46 },
		{ 35, 258, 46 },
		{ 35, 259, 46 },
		{ 35, 260, 46 },
		{ 35, 261, 46 },
		{ 35, 262, 46 },
		{ 35, 263, 46 },
		{ 35, 264, 46 },
		{ 35, 265, 46 },
		{ 35, 266, 46 },
		{ 35, 267, 46 },
		{ 35, 268, 46 },
		{ 35, 269, 46 },
		{ 35, 270, 46 },
		{ 35, 271, 46 },
		{ 35, 272, 46 },
		{ 35, 273, 46 },
		{ 35, 274, 46 },
		{ 35, 275, 46 },
		{ 35, 276, 46 },
		{ 35, 277, 46 },
		{ 35, 278, 46 },
		{ 35, 279, 46 },
		{ 35, 280, 46 },
		{ 35, 281, 46 },
		{ 35, 282, 46 },
		{ 35, 283, 46 },
		{ 35, 284, 46 },
		{ 36, 230, 46 },
		{ 36, 231, 46 },
		{ 36, 232, 46 },
		{ 36, 233, 46 },
		{ 36, 234, 46 },
		{ 36, 235, 46 },
		{ 36, 236, 46 },
		{ 36, 237, 46 },
		{ 36, 238, 46 },
		{ 36, 239, 46 },
		{ 36, 240, 46 },
		{ 36, 241, 46 },
		{ 36, 242, 46 },
		{ 36, 243, 46 },
		{ 36, 244, 46 },
		{ 36, 245, 46 },
		{ 36, 246, 46 },
		{ 36, 247, 46 },
		{ 36, 248, 46 },
		{ 36, 249, 46 },
		{ 36, 250, 46 },
		{ 36, 251, 46 },
		{ 36, 252, 46 },
		{ 36, 253, 46 },
		{ 36, 254, 46 },
		{ 36, 255, 46 },
		{ 36, 256, 46 },
		{ 36, 257, 46 },
		{ 36, 258, 46 },
		{ 36, 259, 46 },
		{ 36, 260, 46 },
		{ 36, 261, 46 },
		{ 36, 262, 46 },
		{ 36, 263, 46 },
		{ 36, 264, 46 },
		{ 36, 265, 46 },
		{ 36, 266, 46 },
		{ 36, 267, 46 },
		{ 36, 268, 46 },
		{ 36, 269, 46 },
		{ 36, 270, 46 },
		{ 36, 271, 46 },
		{ 36, 272, 46 },
		{ 36, 273, 46 },
		{ 36, 274, 46 },
		{ 36, 275, 46 },
		{ 36, 276, 46 },
		{ 36, 277, 46 },
		{ 36, 278, 46 },
		{ 36, 279, 46 },
		{ 36, 280, 46 },
		{ 36, 281, 46 },
		{ 36, 282, 46 },
		{ 36, 283, 46 },
		{ 36, 284, 46 },
		{ 37, 230, 46 },
		{ 37, 231, 46 },
		{ 37, 232, 46 },
		{ 37, 233, 46 },
		{ 37, 234, 46 },
		{ 37, 235, 46 },
		{ 37, 236, 46 },
		{ 37, 237, 46 },
		{ 37, 238, 46 },
		{ 37, 239, 46 },
		{ 37, 240, 46 },
		{ 37, 241, 46 },
		{ 37, 242, 46 },
		{ 37, 243, 46 },
		{ 37, 244, 46 },
		{ 37, 245, 46 },
		{ 37, 246, 46 },
		{ 37, 247, 46 },
		{ 37, 248, 46 },
		{ 37, 249, 46 },
		{ 37, 250, 46 },
		{ 37, 251, 46 },
		{ 37, 252, 46 },
		{ 37, 253, 46 },
		{ 37, 254, 46 },
		{ 37, 255, 46 },
		{ 37, 256, 46 },
		{ 37, 257, 46 },
		{ 37, 258, 46 },
		{ 37, 259, 46 },
		{ 37, 260, 46 },
		{ 37, 261, 46 },
		{ 37, 262, 46 },
		{ 37, 263, 46 },
		{ 37, 264, 46 },
		{ 37, 265, 46 },
		{ 37, 266, 46 },
		{ 37, 267, 46 },
		{ 37, 268, 46 },
		{ 37, 269, 46 },
		{ 37, 270, 46 },
		{ 37, 271, 46 },
		{ 37, 272, 46 },
		{ 37, 273, 46 },
		{ 37, 274, 46 },
		{ 37, 275, 46 },
		{ 37, 276, 46 },
		{ 37, 277, 46 },
		{ 37, 278, 46 },
		{ 37, 279, 46 },
		{ 37, 280, 46 },
		{ 37, 281, 46 },
		{ 37, 282, 46 },
		{ 37, 283, 46 },
		{ 37, 284, 46 },
		{ 38, 230, 46 },
		{ 38, 231, 46 },
		{ 38, 232, 46 },
		{ 38, 233, 46 },
		{ 38, 234, 46 },
		{ 38, 235, 46 },
		{ 38, 236, 46 },
		{ 38, 237, 46 },
		{ 38, 238, 46 },
		{ 38, 239, 46 },
		{ 38, 240, 46 },
		{ 38, 241, 46 },
		{ 38, 242, 46 },
		{ 38, 243, 46 },
		{ 38, 244, 46 },
		{ 38, 245, 46 },
		{ 38, 246, 46 },
		{ 38, 247, 46 },
		{ 38, 248, 46 },
		{ 38, 249, 46 },
		{ 38, 250, 46 },
		{ 38, 251, 46 },
		{ 38, 252, 46 },
		{ 38, 253, 46 },
		{ 38, 254, 46 },
		{ 38, 255, 46 },
		{ 38, 256, 46 },
		{ 38, 257, 46 },
		{ 38, 258, 46 },
		{ 38, 259, 46 },
		{ 38, 260, 46 },
		{ 38, 261, 46 },
		{ 38, 262, 46 },
		{ 38, 263, 46 },
		{ 38, 264, 46 },
		{ 38, 265, 46 },
		{ 38, 266, 46 },
		{ 38, 267, 46 },
		{ 38, 268, 46 },
		{ 38, 269, 46 },
		{ 38, 270, 46 },
		{ 38, 271, 46 },
		{ 38, 272, 46 },
		{ 38, 273, 46 },
		{ 38, 274, 46 },
		{ 38, 275, 46 },
		{ 38, 276, 46 },
		{ 38, 277, 46 },
		{ 38, 278, 46 },
		{ 38, 279, 46 },
		{ 38, 280, 46 },
		{ 38, 281, 46 },
		{ 38, 282, 46 },
		{ 38, 283, 46 },
		{ 38, 284, 46 },
		{ 39, 230, 46 },
		{ 39, 231, 46 },
		{ 39, 232, 46 },
		{ 39, 233, 46 },
		{ 39, 234, 46 },
		{ 39, 235, 46 },
		{ 39, 236, 46 },
		{ 39, 237, 46 },
		{ 39, 238, 46 },
		{ 39, 239, 46 },
		{ 39, 240, 46 },
		{ 39, 241, 46 },
		{ 39, 242, 46 },
		{ 39, 243, 46 },
		{ 39, 244, 46 },
		{ 39, 245, 46 },
		{ 39, 246, 46 },
		{ 39, 247, 46 },
		{ 39, 248, 46 },
		{ 39, 249, 46 },
		{ 39, 250, 46 },
		{ 39, 251, 46 },
		{ 39, 252, 46 },
		{ 39, 253, 46 },
		{ 39, 254, 46 },
		{ 39, 255, 46 },
		{ 39, 256, 46 },
		{ 39, 257, 46 },
		{ 39, 258, 46 },
		{ 39, 259, 46 },
		{ 39, 260, 46 },
		{ 39, 261, 46 },
		{ 39, 262, 46 },
		{ 39, 263, 46 },
		{ 39, 264, 46 },
		{ 39, 265, 46 },
		{ 39, 266, 46 },
		{ 39, 267, 46 },
		{ 39, 268, 46 },
		{ 39, 269, 46 },
		{ 39, 270, 46 },
		{ 39, 271, 46 },
		{ 39, 272, 46 },
		{ 39, 273, 46 },
		{ 39, 274, 46 },
		{ 39, 275, 46 },
		{ 39, 276, 46 },
		{ 39, 277, 46 },
		{ 39, 278, 46 },
		{ 39, 279, 46 },
		{ 39, 280, 46 },
		{ 39, 281, 46 },
		{ 39, 282, 46 },
		{ 39, 283, 46 },
		{ 39, 284, 46 },
		{ 40, 230, 46 },
		{ 40, 231, 46 },
		{ 40, 232, 46 },
		{ 40, 233, 46 },
		{ 40, 234, 46 },
		{ 40, 235, 46 },
		{ 40, 236, 46 },
		{ 40, 237, 46 },
		{ 40, 238, 46 },
		{ 40, 239, 46 },
		{ 40, 240, 46 },
		{ 40, 241, 46 },
		{ 40, 242, 46 },
		{ 40, 243, 46 },
		{ 40, 244, 46 },
		{ 40, 245, 46 },
		{ 40, 246, 46 },
		{ 40, 247, 46 },
		{ 40, 248, 46 },
		{ 40, 249, 46 },
		{ 40, 250, 46 },
		{ 40, 251, 46 },
		{ 40, 252, 46 },
		{ 40, 253, 46 },
		{ 40, 254, 46 },
		{ 40, 255, 46 },
		{ 40, 256, 46 },
		{ 40, 257, 46 },
		{ 40, 258, 46 },
		{ 40, 259, 46 },
		{ 40, 260, 46 },
		{ 40, 261, 46 },
		{ 40, 262, 46 },
		{ 40, 263, 46 },
		{ 40, 264, 46 },
		{ 40, 265, 46 },
		{ 40, 266, 46 },
		{ 40, 267, 46 },
		{ 40, 268, 46 },
		{ 40, 269, 46 },
		{ 40, 270, 46 },
		{ 40, 271, 46 },
		{ 40, 272, 46 },
		{ 40, 273, 46 },
		{ 40, 274, 46 },
		{ 40, 275, 46 },
		{ 40, 276, 46 },
		{ 40, 277, 46 },
		{ 40, 278, 46 },
		{ 40, 279, 46 },
		{ 40, 280, 46 },
		{ 40, 281, 46 },
		{ 40, 282, 46 },
		{ 40, 283, 46 },
		{ 40, 284, 46 },
		{ 41, 231, 46 },
		{ 41, 232, 46 },
		{ 41, 233, 46 },
		{ 41, 234, 46 },
		{ 41, 235, 46 },
		{ 41, 236, 46 },
		{ 41, 237, 46 },
		{ 41, 238, 46 },
		{ 41, 239, 46 },
		{ 41, 240, 46 },
		{ 41, 241, 46 },
		{ 41, 242, 46 },
		{ 41, 243, 46 },
		{ 41, 244, 46 },
		{ 41, 245, 46 },
		{ 41, 246, 46 },
		{ 41, 247, 46 },
		{ 41, 248, 46 },
		{ 41, 249, 46 },
		{ 41, 250, 46 },
		{ 41, 251, 46 },
		{ 41, 252, 46 },
		{ 41, 253, 46 },
		{ 41, 254, 46 },
		{ 41, 255, 46 },
		{ 41, 256, 46 },
		{ 41, 257, 46 },
		{ 41, 258, 46 },
		{ 41, 259, 46 },
		{ 41, 260, 46 },
		{ 41, 261, 46 },
		{ 41, 262, 46 },
		{ 41, 263, 46 },
		{ 41, 264, 46 },
		{ 41, 265, 46 },
		{ 41, 266, 46 },
		{ 41, 267, 46 },
		{ 41, 268, 46 },
		{ 41, 269, 46 },
		{ 41, 270, 46 },
		{ 41, 271, 46 },
		{ 41, 272, 46 },
		{ 41, 273, 46 },
		{ 41, 274, 46 },
		{ 41, 275, 46 },
		{ 41, 276, 46 },
		{ 41, 277, 46 },
		{ 41, 278, 46 },
		{ 41, 279, 46 },
		{ 41, 280, 46 },
		{ 41, 281, 46 },
		{ 41, 282, 46 },
		{ 41, 283, 46 },
		{ 42, 231, 46 },
		{ 42, 232, 46 },
		{ 42, 233, 46 },
		{ 42, 234, 46 },
		{ 42, 235, 46 },
		{ 42, 236, 46 },
		{ 42, 237, 46 },
		{ 42, 238, 46 },
		{ 42, 239, 46 },
		{ 42, 240, 46 },
		{ 42, 241, 46 },
		{ 42, 242, 46 },
		{ 42, 243, 46 },
		{ 42, 244, 46 },
		{ 42, 245, 46 },
		{ 42, 246, 46 },
		{ 42, 247, 46 },
		{ 42, 248, 46 },
		{ 42, 249, 46 },
		{ 42, 250, 46 },
		{ 42, 251, 46 },
		{ 42, 252, 46 },
		{ 42, 253, 46 },
		{ 42, 254, 46 },
		{ 42, 255, 46 },
		{ 42, 256, 46 },
		{ 42, 257, 46 },
		{ 42, 258, 46 },
		{ 42, 259, 46 },
		{ 42, 260, 46 },
		{ 42, 261, 46 },
		{ 42, 262, 46 },
		{ 42, 263, 46 },
		{ 42, 264, 46 },
		{ 42, 265, 46 },
		{ 42, 266, 46 },
		{ 42, 267, 46 },
		{ 42, 268, 46 },
		{ 42, 269, 46 },
		{ 42, 270, 46 },
		{ 42, 271, 46 },
		{ 42, 272, 46 },
		{ 42, 273, 46 },
		{ 42, 274, 46 },
		{ 42, 275, 46 },
		{ 42, 276, 46 },
		{ 42, 277, 46 },
		{ 42, 278, 46 },
		{ 42, 279, 46 },
		{ 42, 280, 46 },
		{ 42, 281, 46 },
		{ 42, 282, 46 },
		{ 42, 283, 46 },
		{ 43, 231, 46 },
		{ 43, 232, 46 },
		{ 43, 233, 46 },
		{ 43, 234, 46 },
		{ 43, 235, 46 },
		{ 43, 236, 46 },
		{ 43, 237, 46 },
		{ 43, 238, 46 },
		{ 43, 239, 46 },
		{ 43, 240, 46 },
		{ 43, 241, 46 },
		{ 43, 242, 46 },
		{ 43, 243, 46 },
		{ 43, 244, 46 },
		{ 43, 245, 46 },
		{ 43, 246, 46 },
		{ 43, 247, 46 },
		{ 43, 248, 46 },
		{ 43, 249, 46 },
		{ 43, 250, 46 },
		{ 43, 251, 46 },
		{ 43, 252, 46 },
		{ 43, 253, 46 },
		{ 43, 254, 46 },
		{ 43, 255, 46 },
		{ 43, 256, 46 },
		{ 43, 257, 46 },
		{ 43, 258, 46 },
		{ 43, 259, 46 },
		{ 43, 260, 46 },
		{ 43, 261, 46 },
		{ 43, 262, 46 },
		{ 43, 263, 46 },
		{ 43, 264, 46 },
		{ 43, 265, 46 },
		{ 43, 266, 46 },
		{ 43, 267, 46 },
		{ 43, 268, 46 },
		{ 43, 269, 46 },
		{ 43, 270, 46 },
		{ 43, 271, 46 },
		{ 43, 272, 46 },
		{ 43, 273, 46 },
		{ 43, 274, 46 },
		{ 43, 275, 46 },
		{ 43, 276, 46 },
		{ 43, 277, 46 },
		{ 43, 278, 46 },
		{ 43, 279, 46 },
		{ 43, 280, 46 },
		{ 43, 281, 46 },
		{ 43, 282, 46 },
		{ 43, 283, 46 },
		{ 44, 231, 46 },
		{ 44, 232, 46 },
		{ 44, 233, 46 },
		{ 44, 234, 46 },
		{ 44, 235, 46 },
		{ 44, 236, 46 },
		{ 44, 237, 46 },
		{ 44, 238, 46 },
		{ 44, 239, 46 },
		{ 44, 240, 46 },
		{ 44, 241, 46 },
		{ 44, 242, 46 },
		{ 44, 243, 46 },
		{ 44, 244, 46 },
		{ 44, 245, 46 },
		{ 44, 246, 46 },
		{ 44, 247, 46 },
		{ 44, 248, 46 },
		{ 44, 249, 46 },
		{ 44, 250, 46 },
		{ 44, 251, 46 },
		{ 44, 252, 46 },
		{ 44, 253, 46 },
		{ 44, 254, 46 },
		{ 44, 255, 46 },
		{ 44, 256, 46 },
		{ 44, 257, 46 },
		{ 44, 258, 46 },
		{ 44, 259, 46 },
		{ 44, 260, 46 },
		{ 44, 261, 46 },
		{ 44, 262, 46 },
		{ 44, 263, 46 },
		{ 44, 264, 46 },
		{ 44, 265, 46 },
		{ 44, 266, 46 },
		{ 44, 267, 46 },
		{ 44, 268, 46 },
		{ 44, 269, 46 },
		{ 44, 270, 46 },
		{ 44, 271, 46 },
		{ 44, 272, 46 },
		{ 44, 273, 46 },
		{ 44, 274, 46 },
		{ 44, 275, 46 },
		{ 44, 276, 46 },
		{ 44, 277, 46 },
		{ 44, 278, 46 },
		{ 44, 279, 46 },
		{ 44, 280, 46 },
		{ 44, 281, 46 },
		{ 44, 282, 46 },
		{ 44, 283, 46 },
		{ 45, 231, 46 },
		{ 45, 232, 46 },
		{ 45, 233, 46 },
		{ 45, 234, 46 },
		{ 45, 235, 46 },
		{ 45, 236, 46 },
		{ 45, 237, 46 },
		{ 45, 238, 46 },
		{ 45, 239, 46 },
		{ 45, 240, 46 },
		{ 45, 241, 46 },
		{ 45, 242, 46 },
		{ 45, 243, 46 },
		{ 45, 244, 46 },
		{ 45, 245, 46 },
		{ 45, 246, 46 },
		{ 45, 247, 46 },
		{ 45, 248, 46 },
		{ 45, 249, 46 },
		{ 45, 250, 46 },
		{ 45, 251, 46 },
		{ 45, 252, 46 },
		{ 45, 253, 46 },
		{ 45, 254, 46 },
		{ 45, 255, 46 },
		{ 45, 256, 46 },
		{ 45, 257, 46 },
		{ 45, 258, 46 },
		{ 45, 259, 46 },
		{ 45, 260, 46 },
		{ 45, 261, 46 },
		{ 45, 262, 46 },
		{ 45, 263, 46 },
		{ 45, 264, 46 },
		{ 45, 265, 46 },
		{ 45, 266, 46 },
		{ 45, 267, 46 },
		{ 45, 268, 46 },
		{ 45, 269, 46 },
		{ 45, 270, 46 },
		{ 45, 271, 46 },
		{ 45, 272, 46 },
		{ 45, 273, 46 },
		{ 45, 274, 46 },
		{ 45, 275, 46 },
		{ 45, 276, 46 },
		{ 45, 277, 46 },
		{ 45, 278, 46 },
		{ 45, 279, 46 },
		{ 45, 280, 46 },
		{ 45, 281, 46 },
		{ 45, 282, 46 },
		{ 45, 283, 46 },
		{ 46, 231, 46 },
		{ 46, 232, 46 },
		{ 46, 233, 46 },
		{ 46, 234, 46 },
		{ 46, 235, 46 },
		{ 46, 236, 46 },
		{ 46, 237, 46 },
		{ 46, 238, 46 },
		{ 46, 239, 46 },
		{ 46, 240, 46 },
		{ 46, 241, 46 },
		{ 46, 242, 46 },
		{ 46, 243, 46 },
		{ 46, 244, 46 },
		{ 46, 245, 46 },
		{ 46, 246, 46 },
		{ 46, 247, 46 },
		{ 46, 248, 46 },
		{ 46, 249, 46 },
		{ 46, 250, 46 },
		{ 46, 251, 46 },
		{ 46, 252, 46 },
		{ 46, 253, 46 },
		{ 46, 254, 46 },
		{ 46, 255, 46 },
		{ 46, 256, 46 },
		{ 46, 257, 46 },
		{ 46, 258, 46 },
		{ 46, 259, 46 },
		{ 46, 260, 46 },
		{ 46, 261, 46 },
		{ 46, 262, 46 },
		{ 46, 263, 46 },
		{ 46, 264, 46 },
		{ 46, 265, 46 },
		{ 46, 266, 46 },
		{ 46, 267, 46 },
		{ 46, 268, 46 },
		{ 46, 269, 46 },
		{ 46, 270, 46 },
		{ 46, 271, 46 },
		{ 46, 272, 46 },
		{ 46, 273, 46 },
		{ 46, 274, 46 },
		{ 46, 275, 46 },
		{ 46, 276, 46 },
		{ 46, 277, 46 },
		{ 46, 278, 46 },
		{ 46, 279, 46 },
		{ 46, 280, 46 },
		{ 46, 281, 46 },
		{ 46, 282, 46 },
		{ 46, 283, 46 },
		{ 47, 231, 46 },
		{ 47, 232, 46 },
		{ 47, 233, 46 },
		{ 47, 234, 46 },
		{ 47, 235, 46 },
		{ 47, 236, 46 },
		{ 47, 237, 46 },
		{ 47, 238, 46 },
		{ 47, 239, 46 },
		{ 47, 240, 46 },
		{ 47, 241, 46 },
		{ 47, 242, 46 },
		{ 47, 243, 46 },
		{ 47, 244, 46 },
		{ 47, 245, 46 },
		{ 47, 246, 46 },
		{ 47, 247, 46 },
		{ 47, 248, 46 },
		{ 47, 249, 46 },
		{ 47, 250, 46 },
		{ 47, 251, 46 },
		{ 47, 252, 46 },
		{ 47, 253, 46 },
		{ 47, 254, 46 },
		{ 47, 255, 46 },
		{ 47, 256, 46 },
		{ 47, 257, 46 },
		{ 47, 258, 46 },
		{ 47, 259, 46 },
		{ 47, 260, 46 },
		{ 47, 261, 46 },
		{ 47, 262, 46 },
		{ 47, 263, 46 },
		{ 47, 264, 46 },
		{ 47, 265, 46 },
		{ 47, 266, 46 },
		{ 47, 267, 46 },
		{ 47, 268, 46 },
		{ 47, 269, 46 },
		{ 47, 270, 46 },
		{ 47, 271, 46 },
		{ 47, 272, 46 },
		{ 47, 273, 46 },
		{ 47, 274, 46 },
		{ 47, 275, 46 },
		{ 47, 276, 46 },
		{ 47, 277, 46 },
		{ 47, 278, 46 },
		{ 47, 279, 46 },
		{ 47, 280, 46 },
		{ 47, 281, 46 },
		{ 47, 282, 46 },
		{ 47, 283, 46 },
		{ 48, 231, 46 },
		{ 48, 232, 46 },
		{ 48, 233, 46 },
		{ 48, 234, 46 },
		{ 48, 235, 46 },
		{ 48, 236, 46 },
		{ 48, 237, 46 },
		{ 48, 238, 46 },
		{ 48, 239, 46 },
		{ 48, 240, 46 },
		{ 48, 241, 46 },
		{ 48, 242, 46 },
		{ 48, 243, 46 },
		{ 48, 244, 46 },
		{ 48, 245, 46 },
		{ 48, 246, 46 },
		{ 48, 247, 46 },
		{ 48, 248, 46 },
		{ 48, 249, 46 },
		{ 48, 250, 46 },
		{ 48, 251, 46 },
		{ 48, 252, 46 },
		{ 48, 253, 46 },
		{ 48, 254, 46 },
		{ 48, 255, 46 },
		{ 48, 256, 46 },
		{ 48, 257, 46 },
		{ 48, 258, 46 },
		{ 48, 259, 46 },
		{ 48, 260, 46 },
		{ 48, 261, 46 },
		{ 48, 262, 46 },
		{ 48, 263, 46 },
		{ 48, 264, 46 },
		{ 48, 265, 46 },
		{ 48, 266, 46 },
		{ 48, 267, 46 },
		{ 48, 268, 46 },
		{ 48, 269, 46 },
		{ 48, 270, 46 },
		{ 48, 271, 46 },
		{ 48, 272, 46 },
		{ 48, 273, 46 },
		{ 48, 274, 46 },
		{ 48, 275, 46 },
		{ 48, 276, 46 },
		{ 48, 277, 46 },
		{ 48, 278, 46 },
		{ 48, 279, 46 },
		{ 48, 280, 46 },
		{ 48, 281, 46 },
		{ 48, 282, 46 },
		{ 48, 283, 46 },
		{ 49, 232, 46 },
		{ 49, 233, 46 },
		{ 49, 234, 46 },
		{ 49, 235, 46 },
		{ 49, 236, 46 },
		{ 49, 237, 46 },
		{ 49, 238, 46 },
		{ 49, 239, 46 },
		{ 49, 240, 46 },
		{ 49, 241, 46 },
		{ 49, 242, 46 },
		{ 49, 243, 46 },
		{ 49, 244, 46 },
		{ 49, 245, 46 },
		{ 49, 246, 46 },
		{ 49, 247, 46 },
		{ 49, 248, 46 },
		{ 49, 249, 46 },
		{ 49, 250, 46 },
		{ 49, 251, 46 },
		{ 49, 252, 46 },
		{ 49, 253, 46 },
		{ 49, 254, 46 },
		{ 49, 255, 46 },
		{ 49, 256, 46 },
		{ 49, 257, 46 },
		{ 49, 258, 46 },
		{ 49, 259, 46 },
		{ 49, 260, 46 },
		{ 49, 261, 46 },
		{ 49, 262, 46 },
		{ 49, 263, 46 },
		{ 49, 264, 46 },
		{ 49, 265, 46 },
		{ 49, 266, 46 },
		{ 49, 267, 46 },
		{ 49, 268, 46 },
		{ 49, 269, 46 },
		{ 49, 270, 46 },
		{ 49, 271, 46 },
		{ 49, 272, 46 },
		{ 49, 273, 46 },
		{ 49, 274, 46 },
		{ 49, 275, 46 },
		{ 49, 276, 46 },
		{ 49, 277, 46 },
		{ 49, 278, 46 },
		{ 49, 279, 46 },
		{ 49, 280, 46 },
		{ 49, 281, 46 },
		{ 49, 282, 46 },
		{ 50, 232, 46 },
		{ 50, 233, 46 },
		{ 50, 234, 46 },
		{ 50, 235, 46 },
		{ 50, 236, 46 },
		{ 50, 237, 46 },
		{ 50, 238, 46 },
		{ 50, 239, 46 },
		{ 50, 240, 46 },
		{ 50, 241, 46 },
		{ 50, 242, 46 },
		{ 50, 243, 46 },
		{ 50, 244, 46 },
		{ 50, 245, 46 },
		{ 50, 246, 46 },
		{ 50, 247, 46 },
		{ 50, 248, 46 },
		{ 50, 249, 46 },
		{ 50, 250, 46 },
		{ 50, 251, 46 },
		{ 50, 252, 46 },
		{ 50, 253, 46 },
		{ 50, 254, 46 },
		{ 50, 255, 46 },
		{ 50, 256, 46 },
		{ 50, 257, 46 },
		{ 50, 258, 46 },
		{ 50, 259, 46 },
		{ 50, 260, 46 },
		{ 50, 261, 46 },
		{ 50, 262, 46 },
		{ 50, 263, 46 },
		{ 50, 264, 46 },
		{ 50, 265, 46 },
		{ 50, 266, 46 },
		{ 50, 267, 46 },
		{ 50, 268, 46 },
		{ 50, 269, 46 },
		{ 50, 270, 46 },
		{ 50, 271, 46 },
		{ 50, 272, 46 },
		{ 50, 273, 46 },
		{ 50, 274, 46 },
		{ 50, 275, 46 },
		{ 50, 276, 46 },
		{ 50, 277, 46 },
		{ 50, 278, 46 },
		{ 50, 279, 46 },
		{ 50, 280, 46 },
		{ 50, 281, 46 },
		{ 50, 282, 46 },
		{ 51, 232, 46 },
		{ 51, 233, 46 },
		{ 51, 234, 46 },
		{ 51, 235, 46 },
		{ 51, 236, 46 },
		{ 51, 237, 46 },
		{ 51, 238, 46 },
		{ 51, 239, 46 },
		{ 51, 240, 46 },
		{ 51, 241, 46 },
		{ 51, 242, 46 },
		{ 51, 243, 46 },
		{ 51, 244, 46 },
		{ 51, 245, 46 },
		{ 51, 246, 46 },
		{ 51, 247, 46 },
		{ 51, 248, 46 },
		{ 51, 249, 46 },
		{ 51, 250, 46 },
		{ 51, 251, 46 },
		{ 51, 252, 46 },
		{ 51, 253, 46 },
		{ 51, 254, 46 },
		{ 51, 255, 46 },
		{ 51, 256, 46 },
		{ 51, 257, 46 },
		{ 51, 258, 46 },
		{ 51, 259, 46 },
		{ 51, 260, 46 },
		{ 51, 261, 46 },
		{ 51, 262, 46 },
		{ 51, 263, 46 },
		{ 51, 264, 46 },
		{ 51, 265, 46 },
		{ 51, 266, 46 },
		{ 51, 267, 46 },
		{ 51, 268, 46 },
		{ 51, 269, 46 },
		{ 51, 270, 46 },
		{ 51, 271, 46 },
		{ 51, 272, 46 },
		{ 51, 273, 46 },
		{ 51, 274, 46 },
		{ 51, 275, 46 },
		{ 51, 276, 46 },
		{ 51, 277, 46 },
		{ 51, 278, 46 },
		{ 51, 279, 46 },
		{ 51, 280, 46 },
		{ 51, 281, 46 },
		{ 51, 282, 46 },
		{ 52, 232, 46 },
		{ 52, 233, 46 },
		{ 52, 234, 46 },
		{ 52, 235, 46 },
		{ 52, 236, 46 },
		{ 52, 237, 46 },
		{ 52, 238, 46 },
		{ 52, 239, 46 },
		{ 52, 240, 46 },
		{ 52, 241, 46 },
		{ 52, 242, 46 },
		{ 52, 243, 46 },
		{ 52, 244, 46 },
		{ 52, 245, 46 },
		{ 52, 246, 46 },
		{ 52, 247, 46 },
		{ 52, 248, 46 },
		{ 52, 249, 46 },
		{ 52, 250, 46 },
		{ 52, 251, 46 },
		{ 52, 252, 46 },
		{ 52, 253, 46 },
		{ 52, 254, 46 },
		{ 52, 255, 46 },
		{ 52, 256, 46 },
		{ 52, 257, 46 },
		{ 52, 258, 46 },
		{ 52, 259, 46 },
		{ 52, 260, 46 },
		{ 52, 261, 46 },
		{ 52, 262, 46 },
		{ 52, 263, 46 },
		{ 52, 264, 46 },
		{ 52, 265, 46 },
		{ 52, 266, 46 },
		{ 52, 267, 46 },
		{ 52, 268, 46 },
		{ 52, 269, 46 },
		{ 52, 270, 46 },
		{ 52, 271, 46 },
		{ 52, 272, 46 },
		{ 52, 273, 46 },
		{ 52, 274, 46 },
		{ 52, 275, 46 },
		{ 52, 276, 46 },
		{ 52, 277, 46 },
		{ 52, 278, 46 },
		{ 52, 279, 46 },
		{ 52, 280, 46 },
		{ 52, 281, 46 },
		{ 52, 282, 46 },
		{ 53, 232, 46 },
		{ 53, 233, 46 },
		{ 53, 234, 46 },
		{ 53, 235, 46 },
		{ 53, 236, 46 },
		{ 53, 237, 46 },
		{ 53, 238, 46 },
		{ 53, 239, 46 },
		{ 53, 240, 46 },
		{ 53, 241, 46 },
		{ 53, 242, 46 },
		{ 53, 243, 46 },
		{ 53, 244, 46 },
		{ 53, 245, 46 },
		{ 53, 246, 46 },
		{ 53, 247, 46 },
		{ 53, 248, 46 },
		{ 53, 249, 46 },
		{ 53, 250, 46 },
		{ 53, 251, 46 },
		{ 53, 252, 46 },
		{ 53, 253, 46 },
		{ 53, 254, 46 },
		{ 53, 255, 46 },
		{ 53, 256, 46 },
		{ 53, 257, 46 },
		{ 53, 258, 46 },
		{ 53, 259, 46 },
		{ 53, 260, 46 },
		{ 53, 261, 46 },
		{ 53, 262, 46 },
		{ 53, 263, 46 },
		{ 53, 264, 46 },
		{ 53, 265, 46 },
		{ 53, 266, 46 },
		{ 53, 267, 46 },
		{ 53, 268, 46 },
		{ 53, 269, 46 },
		{ 53, 270, 46 },
		{ 53, 271, 46 },
		{ 53, 272, 46 },
		{ 53, 273, 46 },
		{ 53, 274, 46 },
		{ 53, 275, 46 },
		{ 53, 276, 46 },
		{ 53, 277, 46 },
		{ 53, 278, 46 },
		{ 53, 279, 46 },
		{ 53, 280, 46 },
		{ 53, 281, 46 },
		{ 53, 282, 46 },
		{ 54, 232, 46 },
		{ 54, 233, 46 },
		{ 54, 234, 46 },
		{ 54, 235, 46 },
		{ 54, 236, 46 },
		{ 54, 237, 46 },
		{ 54, 238, 46 },
		{ 54, 239, 46 },
		{ 54, 240, 46 },
		{ 54, 241, 46 },
		{ 54, 242, 46 },
		{ 54, 243, 46 },
		{ 54, 244, 46 },
		{ 54, 245, 46 },
		{ 54, 246, 46 },
		{ 54, 247, 46 },
		{ 54, 248, 46 },
		{ 54, 249, 46 },
		{ 54, 250, 46 },
		{ 54, 251, 46 },
		{ 54, 252, 46 },
		{ 54, 253, 46 },
		{ 54, 254, 46 },
		{ 54, 255, 46 },
		{ 54, 256, 46 },
		{ 54, 257, 46 },
		{ 54, 258, 46 },
		{ 54, 259, 46 },
		{ 54, 260, 46 },
		{ 54, 261, 46 },
		{ 54, 262, 46 },
		{ 54, 263, 46 },
		{ 54, 264, 46 },
		{ 54, 265, 46 },
		{ 54, 266, 46 },
		{ 54, 267, 46 },
		{ 54, 268, 46 },
		{ 54, 269, 46 },
		{ 54, 270, 46 },
		{ 54, 271, 46 },
		{ 54, 272, 46 },
		{ 54, 273, 46 },
		{ 54, 274, 46 },
		{ 54, 275, 46 },
		{ 54, 276, 46 },
		{ 54, 277, 46 },
		{ 54, 278, 46 },
		{ 54, 279, 46 },
		{ 54, 280, 46 },
		{ 54, 281, 46 },
		{ 54, 282, 46 },
		{ 55, 232, 46 },
		{ 55, 233, 46 },
		{ 55, 234, 46 },
		{ 55, 235, 46 },
		{ 55, 236, 46 },
		{ 55, 237, 46 },
		{ 55, 238, 46 },
		{ 55, 239, 46 },
		{ 55, 240, 46 },
		{ 55, 241, 46 },
		{ 55, 242, 46 },
		{ 55, 243, 46 },
		{ 55, 244, 46 },
		{ 55, 245, 46 },
		{ 55, 246, 46 },
		{ 55, 247, 46 },
		{ 55, 248, 46 },
		{ 55, 249, 46 },
		{ 55, 250, 46 },
		{ 55, 251, 46 },
		{ 55, 252, 46 },
		{ 55, 253, 46 },
		{ 55, 254, 46 },
		{ 55, 255, 46 },
		{ 55, 256, 46 },
		{ 55, 257, 46 },
		{ 55, 258, 46 },
		{ 55, 259, 46 },
		{ 55, 260, 46 },
		{ 55, 261, 46 },
		{ 55, 262, 46 },
		{ 55, 263, 46 },
		{ 55, 264, 46 },
		{ 55, 265, 46 },
		{ 55, 266, 46 },
		{ 55, 267, 46 },
		{ 55, 268, 46 },
		{ 55, 269, 46 },
		{ 55, 270, 46 },
		{ 55, 271, 46 },
		{ 55, 272, 46 },
		{ 55, 273, 46 },
		{ 55, 274, 46 },
		{ 55, 275, 46 },
		{ 55, 276, 46 },
		{ 55, 277, 46 },
		{ 55, 278, 46 },
		{ 55, 279, 46 },
		{ 55, 280, 46 },
		{ 55, 281, 46 },
		{ 55, 282, 46 },
		{ 56, 232, 46 },
		{ 56, 233, 46 },
		{ 56, 234, 46 },
		{ 56, 235, 46 },
		{ 56, 236, 46 },
		{ 56, 237, 46 },
		{ 56, 238, 46 },
		{ 56, 239, 46 },
		{ 56, 240, 46 },
		{ 56, 241, 46 },
		{ 56, 242, 46 },
		{ 56, 243, 46 },
		{ 56, 244, 46 },
		{ 56, 245, 46 },
		{ 56, 246, 46 },
		{ 56, 247, 46 },
		{ 56, 248, 46 },
		{ 56, 249, 46 },
		{ 56, 250, 46 },
		{ 56, 251, 46 },
		{ 56, 252, 46 },
		{ 56, 253, 46 },
		{ 56, 254, 46 },
		{ 56, 255, 46 },
		{ 56, 256, 46 },
		{ 56, 257, 46 },
		{ 56, 258, 46 },
		{ 56, 259, 46 },
		{ 56, 260, 46 },
		{ 56, 261, 46 },
		{ 56, 262, 46 },
		{ 56, 263, 46 },
		{ 56, 264, 46 },
		{ 56, 265, 46 },
		{ 56, 266, 46 },
		{ 56, 267, 46 },
		{ 56, 268, 46 },
		{ 56, 269, 46 },
		{ 56, 270, 46 },
		{ 56, 271, 46 },
		{ 56, 272, 46 },
		{ 56, 273, 46 },
		{ 56, 274, 46 },
		{ 56, 275, 46 },
		{ 56, 276, 46 },
		{ 56, 277, 46 },
		{ 56, 278, 46 },
		{ 56, 279, 46 },
		{ 56, 280, 46 },
		{ 56, 281, 46 },
		{ 56, 282, 46 },
		{ 57, 233, 46 },
		{ 57, 234, 46 },
		{ 57, 235, 46 },
		{ 57, 236, 46 },
		{ 57, 237, 46 },
		{ 57, 238, 46 },
		{ 57, 239, 46 },
		{ 57, 240, 46 },
		{ 57, 241, 46 },
		{ 57, 242, 46 },
		{ 57, 243, 46 },
		{ 57, 244, 46 },
		{ 57, 245, 46 },
		{ 57, 246, 46 },
		{ 57, 247, 46 },
		{ 57, 248, 46 },
		{ 57, 249, 46 },
		{ 57, 250, 46 },
		{ 57, 251, 46 },
		{ 57, 252, 46 },
		{ 57, 253, 46 },
		{ 57, 254, 46 },
		{ 57, 255, 46 },
		{ 57, 256, 46 },
		{ 57, 257, 46 },
		{ 57, 258, 46 },
		{ 57, 259, 46 },
		{ 57, 260, 46 },
		{ 57, 261, 46 },
		{ 57, 262, 46 },
		{ 57, 263, 46 },
		{ 57, 264, 46 },
		{ 57, 265, 46 },
		{ 57, 266, 46 },
		{ 57, 267, 46 },
		{ 57, 268, 46 },
		{ 57, 269, 46 },
		{ 57, 270, 46 },
		{ 57, 271, 46 },
		{ 57, 272, 46 },
		{ 57, 273, 46 },
		{ 57, 274, 46 },
		{ 57, 275, 46 },
		{ 57, 276, 46 },
		{ 57, 277, 46 },
		{ 57, 278, 46 },
		{ 57, 279, 46 },
		{ 57, 280, 46 },
		{ 57, 281, 46 },
		{ 58, 233, 46 },
		{ 58, 234, 46 },
		{ 58, 235, 46 },
		{ 58, 236, 46 },
		{ 58, 237, 46 },
		{ 58, 238, 46 },
		{ 58, 239, 46 },
		{ 58, 240, 46 },
		{ 58, 241, 46 },
		{ 58, 242, 46 },
		{ 58, 243, 46 },
		{ 58, 244, 46 },
		{ 58, 245, 46 },
		{ 58, 246, 46 },
		{ 58, 247, 46 },
		{ 58, 248, 46 },
		{ 58, 249, 46 },
		{ 58, 250, 46 },
		{ 58, 251, 46 },
		{ 58, 252, 46 },
		{ 58, 253, 46 },
		{ 58, 254, 46 },
		{ 58, 255, 46 },
		{ 58, 256, 46 },
		{ 58, 257, 46 },
		{ 58, 258, 46 },
		{ 58, 259, 46 },
		{ 58, 260, 46 },
		{ 58, 261, 46 },
		{ 58, 262, 46 },
		{ 58, 263, 46 },
		{ 58, 264, 46 },
		{ 58, 265, 46 },
		{ 58, 266, 46 },
		{ 58, 267, 46 },
		{ 58, 268, 46 },
		{ 58, 269, 46 },
		{ 58, 270, 46 },
		{ 58, 271, 46 },
		{ 58, 272, 46 },
		{ 58, 273, 46 },
		{ 58, 274, 46 },
		{ 58, 275, 46 },
		{ 58, 276, 46 },
		{ 58, 277, 46 },
		{ 58, 278, 46 },
		{ 58, 279, 46 },
		{ 58, 280, 46 },
		{ 58, 281, 46 },
		{ 59, 233, 46 },
		{ 59, 234, 46 },
		{ 59, 235, 46 },
		{ 59, 236, 46 },
		{ 59, 237, 46 },
		{ 59, 238, 46 },
		{ 59, 239, 46 },
		{ 59, 240, 46 },
		{ 59, 241, 46 },
		{ 59, 242, 46 },
		{ 59, 243, 46 },
		{ 59, 244, 46 },
		{ 59, 245, 46 },
		{ 59, 246, 46 },
		{ 59, 247, 46 },
		{ 59, 248, 46 },
		{ 59, 249, 46 },
		{ 59, 250, 46 },
		{ 59, 251, 46 },
		{ 59, 252, 46 },
		{ 59, 253, 46 },
		{ 59, 254, 46 },
		{ 59, 255, 46 },
		{ 59, 256, 46 },
		{ 59, 257, 46 },
		{ 59, 258, 46 },
		{ 59, 259, 46 },
		{ 59, 260, 46 },
		{ 59, 261, 46 },
		{ 59, 262, 46 },
		{ 59, 263, 46 },
		{ 59, 264, 46 },
		{ 59, 265, 46 },
		{ 59, 266, 46 },
		{ 59, 267, 46 },
		{ 59, 268, 46 },
		{ 59, 269, 46 },
		{ 59, 270, 46 },
		{ 59, 271, 46 },
		{ 59, 272, 46 },
		{ 59, 273, 46 },
		{ 59, 274, 46 },
		{ 59, 275, 46 },
		{ 59, 276, 46 },
		{ 59, 277, 46 },
		{ 59, 278, 46 },
		{ 59, 279, 46 },
		{ 59, 280, 46 },
		{ 59, 281, 46 },
		{ 60, 233, 46 },
		{ 60, 234, 46 },
		{ 60, 235, 46 },
		{ 60, 236, 46 },
		{ 60, 237, 46 },
		{ 60, 238, 46 },
		{ 60, 239, 46 },
		{ 60, 240, 46 },
		{ 60, 241, 46 },
		{ 60, 242, 46 },
		{ 60, 243, 46 },
		{ 60, 244, 46 },
		{ 60, 245, 46 },
		{ 60, 246, 46 },
		{ 60, 247, 46 },
		{ 60, 248, 46 },
		{ 60, 249, 46 },
		{ 60, 250, 46 },
		{ 60, 251, 46 },
		{ 60, 252, 46 },
		{ 60, 253, 46 },
		{ 60, 254, 46 },
		{ 60, 255, 46 },
		{ 60, 256, 46 },
		{ 60, 257, 46 },
		{ 60, 258, 46 },
		{ 60, 259, 46 },
		{ 60, 260, 46 },
		{ 60, 261, 46 },
		{ 60, 262, 46 },
		{ 60, 263, 46 },
		{ 60, 264, 46 },
		{ 60, 265, 46 },
		{ 60, 266, 46 },
		{ 60, 267, 46 },
		{ 60, 268, 46 },
		{ 60, 269, 46 },
		{ 60, 270, 46 },
		{ 60, 271, 46 },
		{ 60, 272, 46 },
		{ 60, 273, 46 },
		{ 60, 274, 46 },
		{ 60, 275, 46 },
		{ 60, 276, 46 },
		{ 60, 277, 46 },
		{ 60, 278, 46 },
		{ 60, 279, 46 },
		{ 60, 280, 46 },
		{ 60, 281, 46 },
		{ 61, 233, 46 },
		{ 61, 234, 46 },
		{ 61, 235, 46 },
		{ 61, 236, 46 },
		{ 61, 237, 46 },
		{ 61, 238, 46 },
		{ 61, 239, 46 },
		{ 61, 240, 46 },
		{ 61, 241, 46 },
		{ 61, 242, 46 },
		{ 61, 243, 46 },
		{ 61, 244, 46 },
		{ 61, 245, 46 },
		{ 61, 246, 46 },
		{ 61, 247, 46 },
		{ 61, 248, 46 },
		{ 61, 249, 46 },
		{ 61, 250, 46 },
		{ 61, 251, 46 },
		{ 61, 252, 46 },
		{ 61, 253, 46 },
		{ 61, 254, 46 },
		{ 61, 255, 46 },
		{ 61, 256, 46 },
		{ 61, 257, 46 },
		{ 61, 258, 46 },
		{ 61, 259, 46 },
		{ 61, 260, 46 },
		{ 61, 261, 46 },
		{ 61, 262, 46 },
		{ 61, 263, 46 },
		{ 61, 264, 46 },
		{ 61, 265, 46 },
		{ 61, 266, 46 },
		{ 61, 267, 46 },
		{ 61, 268, 46 },
		{ 61, 269, 46 },
		{ 61, 270, 46 },
		{ 61, 271, 46 },
		{ 61, 272, 46 },
		{ 61, 273, 46 },
		{ 61, 274, 46 },
		{ 61, 275, 46 },
		{ 61, 276, 46 },
		{ 61, 277, 46 },
		{ 61, 278, 46 },
		{ 61, 279, 46 },
		{ 61, 280, 46 },
		{ 61, 281, 46 },
		{ 62, 233, 46 },
		{ 62, 234, 46 },
		{ 62, 235, 46 },
		{ 62, 236, 46 },
		{ 62, 237, 46 },
		{ 62, 238, 46 },
		{ 62, 239, 46 },
		{ 62, 240, 46 },
		{ 62, 241, 46 },
		{ 62, 242, 46 },
		{ 62, 243, 46 },
		{ 62, 244, 46 },
		{ 62, 245, 46 },
		{ 62, 246, 46 },
		{ 62, 247, 46 },
		{ 62, 248, 46 },
		{ 62, 249, 46 },
		{ 62, 250, 46 },
		{ 62, 251, 46 },
		{ 62, 252, 46 },
		{ 62, 253, 46 },
		{ 62, 254, 46 },
		{ 62, 255, 46 },
		{ 62, 256, 46 },
		{ 62, 257, 46 },
		{ 62, 258, 46 },
		{ 62, 259, 46 },
		{ 62, 260, 46 },
		{ 62, 261, 46 },
		{ 62, 262, 46 },
		{ 62, 263, 46 },
		{ 62, 264, 46 },
		{ 62, 265, 46 },
		{ 62, 266, 46 },
		{ 62, 267, 46 },
		{ 62, 268, 46 },
		{ 62, 269, 46 },
		{ 62, 270, 46 },
		{ 62, 271, 46 },
		{ 62, 272, 46 },
		{ 62, 273, 46 },
		{ 62, 274, 46 },
		{ 62, 275, 46 },
		{ 62, 276, 46 },
		{ 62, 277, 46 },
		{ 62, 278, 46 },
		{ 62, 279, 46 },
		{ 62, 280, 46 },
		{ 62, 281, 46 },
		{ 63, 233, 46 },
		{ 63, 234, 46 },
		{ 63, 235, 46 },
		{ 63, 236, 46 },
		{ 63, 237, 46 },
		{ 63, 238, 46 },
		{ 63, 239, 46 },
		{ 63, 240, 46 },
		{ 63, 241, 46 },
		{ 63, 242, 46 },
		{ 63, 243, 46 },
		{ 63, 244, 46 },
		{ 63, 245, 46 },
		{ 63, 246, 46 },
		{ 63, 247, 46 },
		{ 63, 248, 46 },
		{ 63, 249, 46 },
		{ 63, 250, 46 },
		{ 63, 251, 46 },
		{ 63, 252, 46 },
		{ 63, 253, 46 },
		{ 63, 254, 46 },
		{ 63, 255, 46 },
		{ 63, 256, 46 },
		{ 63, 257, 46 },
		{ 63, 258, 46 },
		{ 63, 259, 46 },
		{ 63, 260, 46 },
		{ 63, 261, 46 },
		{ 63, 262, 46 },
		{ 63, 263, 46 },
		{ 63, 264, 46 },
		{ 63, 265, 46 },
		{ 63, 266, 46 },
		{ 63, 267, 46 },
		{ 63, 268, 46 },
		{ 63, 269, 46 },
		{ 63, 270, 46 },
		{ 63, 271, 46 },
		{ 63, 272, 46 },
		{ 63, 273, 46 },
		{ 63, 274, 46 },
		{ 63, 275, 46 },
		{ 63, 276, 46 },
		{ 63, 277, 46 },
		{ 63, 278, 46 },
		{ 63, 279, 46 },
		{ 63, 280, 46 },
		{ 63, 281, 46 },
		{ 64, 233, 46 },
		{ 64, 234, 46 },
		{ 64, 235, 46 },
		{ 64, 236, 46 },
		{ 64, 237, 46 },
		{ 64, 238, 46 },
		{ 64, 239, 46 },
		{ 64, 240, 46 },
		{ 64, 241, 46 },
		{ 64, 242, 46 },
		{ 64, 243, 46 },
		{ 64, 244, 46 },
		{ 64, 245, 46 },
		{ 64, 246, 46 },
		{ 64, 247, 46 },
		{ 64, 248, 46 },
		{ 64, 249, 46 },
		{ 64, 250, 46 },
		{ 64, 251, 46 },
		{ 64, 252, 46 },
		{ 64, 253, 46 },
		{ 64, 254, 46 },
		{ 64, 255, 46 },
		{ 64, 256, 46 },
		{ 64, 257, 46 },
		{ 64, 258, 46 },
		{ 64, 259, 46 },
		{ 64, 260, 46 },
		{ 64, 261, 46 },
		{ 64, 262, 46 },
		{ 64, 263, 46 },
		{ 64, 264, 46 },
		{ 64, 265, 46 },
		{ 64, 266, 46 },
		{ 64, 267, 46 },
		{ 64, 268, 46 },
		{ 64, 269, 46 },
		{ 64, 270, 46 },
		{ 64, 271, 46 },
		{ 64, 272, 46 },
		{ 64, 273, 46 },
		{ 64, 274, 46 },
		{ 64, 275, 46 },
		{ 64, 276, 46 },
		{ 64, 277, 46 },
		{ 64, 278, 46 },
		{ 64, 279, 46 },
		{ 64, 280, 46 },
		{ 64, 281, 46 },
		{ 65, 234, 46 },
		{ 65, 235, 46 },
		{ 65, 236, 46 },
		{ 65, 237, 46 },
		{ 65, 238, 46 },
		{ 65, 239, 46 },
		{ 65, 240, 46 },
		{ 65, 241, 46 },
		{ 65, 242, 46 },
		{ 65, 243, 46 },
		{ 65, 244, 46 },
		{ 65, 245, 46 },
		{ 65, 246, 46 },
		{ 65, 247, 46 },
		{ 65, 248, 46 },
		{ 65, 249, 46 },
		{ 65, 250, 46 },
		{ 65, 251, 46 },
		{ 65, 252, 46 },
		{ 65, 253, 46 },
		{ 65, 254, 46 },
		{ 65, 255, 46 },
		{ 65, 256, 46 },
		{ 65, 257, 46 },
		{ 65, 258, 46 },
		{ 65, 259, 46 },
		{ 65, 260, 46 },
		{ 65, 261, 46 },
		{ 65, 262, 46 },
		{ 65, 263, 46 },
		{ 65, 264, 46 },
		{ 65, 265, 46 },
		{ 65, 266, 46 },
		{ 65, 267, 46 },
		{ 65, 268, 46 },
		{ 65, 269, 46 },
		{ 65, 270, 46 },
		{ 65, 271, 46 },
		{ 65, 272, 46 },
		{ 65, 273, 46 },
		{ 65, 274, 46 },
		{ 65, 275, 46 },
		{ 65, 276, 46 },
		{ 65, 277, 46 },
		{ 65, 278, 46 },
		{ 65, 279, 46 },
		{ 65, 280, 46 },
		{ 66, 234, 46 },
		{ 66, 235, 46 },
		{ 66, 236, 46 },
		{ 66, 237, 46 },
		{ 66, 238, 46 },
		{ 66, 239, 46 },
		{ 66, 240, 46 },
		{ 66, 241, 46 },
		{ 66, 242, 46 },
		{ 66, 243, 46 },
		{ 66, 244, 46 },
		{ 66, 245, 46 },
		{ 66, 246, 46 },
		{ 66, 247, 46 },
		{ 66, 248, 46 },
		{ 66, 249, 46 },
		{ 66, 250, 46 },
		{ 66, 251, 46 },
		{ 66, 252, 46 },
		{ 66, 253, 46 },
		{ 66, 254, 46 },
		{ 66, 255, 46 },
		{ 66, 256, 46 },
		{ 66, 257, 46 },
		{ 66, 258, 46 },
		{ 66, 259, 46 },
		{ 66, 260, 46 },
		{ 66, 261, 46 },
		{ 66, 262, 46 },
		{ 66, 263, 46 },
		{ 66, 264, 46 },
		{ 66, 265, 46 },
		{ 66, 266, 46 },
		{ 66, 267, 46 },
		{ 66, 268, 46 },
		{ 66, 269, 46 },
		{ 66, 270, 46 },
		{ 66, 271, 46 },
		{ 66, 272, 46 },
		{ 66, 273, 46 },
		{ 66, 274, 46 },
		{ 66, 275, 46 },
		{ 66, 276, 46 },
		{ 66, 277, 46 },
		{ 66, 278, 46 },
		{ 66, 279, 46 },
		{ 66, 280, 46 },
		{ 67, 234, 46 },
		{ 67, 235, 46 },
		{ 67, 236, 46 },
		{ 67, 237, 46 },
		{ 67, 238, 46 },
		{ 67, 239, 46 },
		{ 67, 240, 46 },
		{ 67, 241, 46 },
		{ 67, 242, 46 },
		{ 67, 243, 46 },
		{ 67, 244, 46 },
		{ 67, 245, 46 },
		{ 67, 246, 46 },
		{ 67, 247, 46 },
		{ 67, 248, 46 },
		{ 67, 249, 46 },
		{ 67, 250, 46 },
		{ 67, 251, 46 },
		{ 67, 252, 46 },
		{ 67, 253, 46 },
		{ 67, 254, 46 },
		{ 67, 255, 46 },
		{ 67, 256, 46 },
		{ 67, 257, 46 },
		{ 67, 258, 46 },
		{ 67, 259, 46 },
		{ 67, 260, 46 },
		{ 67, 261, 46 },
		{ 67, 262, 46 },
		{ 67, 263, 46 },
		{ 67, 264, 46 },
		{ 67, 265, 46 },
		{ 67, 266, 46 },
		{ 67, 267, 46 },
		{ 67, 268, 46 },
		{ 67, 269, 46 },
		{ 67, 270, 46 },
		{ 67, 271, 46 },
		{ 67, 272, 46 },
		{ 67, 273, 46 },
		{ 67, 274, 46 },
		{ 67, 275, 46 },
		{ 67, 276, 46 },
		{ 67, 277, 46 },
		{ 67, 278, 46 },
		{ 67, 279, 46 },
		{ 67, 280, 46 },
		{ 68, 234, 46 },
		{ 68, 235, 46 },
		{ 68, 236, 46 },
		{ 68, 237, 46 },
		{ 68, 238, 46 },
		{ 68, 239, 46 },
		{ 68, 240, 46 },
		{ 68, 241, 46 },
		{ 68, 242, 46 },
		{ 68, 243, 46 },
		{ 68, 244, 46 },
		{ 68, 245, 46 },
		{ 68, 246, 46 },
		{ 68, 247, 46 },
		{ 68, 248, 46 },
		{ 68, 249, 46 },
		{ 68, 250, 46 },
		{ 68, 251, 46 },
		{ 68, 252, 46 },
		{ 68, 253, 46 },
		{ 68, 254, 46 },
		{ 68, 255, 46 },
		{ 68, 256, 46 },
		{ 68, 257, 46 },
		{ 68, 258, 46 },
		{ 68, 259, 46 },
		{ 68, 260, 46 },
		{ 68, 261, 46 },
		{ 68, 262, 46 },
		{ 68, 263, 46 },
		{ 68, 264, 46 },
		{ 68, 265, 46 },
		{ 68, 266, 46 },
		{ 68, 267, 46 },
		{ 68, 268, 46 },
		{ 68, 269, 46 },
		{ 68, 270, 46 },
		{ 68, 271, 46 },
		{ 68, 272, 46 },
		{ 68, 273, 46 },
		{ 68, 274, 46 },
		{ 68, 275, 46 },
		{ 68, 276, 46 },
		{ 68, 277, 46 },
		{ 68, 278, 46 },
		{ 68, 279, 46 },
		{ 68, 280, 46 },
		{ 69, 234, 46 },
		{ 69, 235, 46 },
		{ 69, 236, 46 },
		{ 69, 237, 46 },
		{ 69, 238, 46 },
		{ 69, 239, 46 },
		{ 69, 240, 46 },
		{ 69, 241, 46 },
		{ 69, 242, 46 },
		{ 69, 243, 46 },
		{ 69, 244, 46 },
		{ 69, 245, 46 },
		{ 69, 246, 46 },
		{ 69, 247, 46 },
		{ 69, 248, 46 },
		{ 69, 249, 46 },
		{ 69, 250, 46 },
		{ 69, 251, 46 },
		{ 69, 252, 46 },
		{ 69, 253, 46 },
		{ 69, 254, 46 },
		{ 69, 255, 46 },
		{ 69, 256, 46 },
		{ 69, 257, 46 },
		{ 69, 258, 46 },
		{ 69, 259, 46 },
		{ 69, 260, 46 },
		{ 69, 261, 46 },
		{ 69, 262, 46 },
		{ 69, 263, 46 },
		{ 69, 264, 46 },
		{ 69, 265, 46 },
		{ 69, 266, 46 },
		{ 69, 267, 46 },
		{ 69, 268, 46 },
		{ 69, 269, 46 },
		{ 69, 270, 46 },
		{ 69, 271, 46 },
		{ 69, 272, 46 },
		{ 69, 273, 46 },
		{ 69, 274, 46 },
		{ 69, 275, 46 },
		{ 69, 276, 46 },
		{ 69, 277, 46 },
		{ 69, 278, 46 },
		{ 69, 279, 46 },
		{ 69, 280, 46 },
		{ 70, 234, 46 },
		{ 70, 235, 46 },
		{ 70, 236, 46 },
		{ 70, 237, 46 },
		{ 70, 238, 46 },
		{ 70, 239, 46 },
		{ 70, 240, 46 },
		{ 70, 241, 46 },
		{ 70, 242, 46 },
		{ 70, 243, 46 },
		{ 70, 244, 46 },
		{ 70, 245, 46 },
		{ 70, 246, 46 },
		{ 70, 247, 46 },
		{ 70, 248, 46 },
		{ 70, 249, 46 },
		{ 70, 250, 46 },
		{ 70, 251, 46 },
		{ 70, 252, 46 },
		{ 70, 253, 46 },
		{ 70, 254, 46 },
		{ 70, 255, 46 },
		{ 70, 256, 46 },
		{ 70, 257, 46 },
		{ 70, 258, 46 },
		{ 70, 259, 46 },
		{ 70, 260, 46 },
		{ 70, 261, 46 },
		{ 70, 262, 46 },
		{ 70, 263, 46 },
		{ 70, 264, 46 },
		{ 70, 265, 46 },
		{ 70, 266, 46 },
		{ 70, 267, 46 },
		{ 70, 268, 46 },
		{ 70, 269, 46 },
		{ 70, 270, 46 },
		{ 70, 271, 46 },
		{ 70, 272, 46 },
		{ 70, 273, 46 },
		{ 70, 274, 46 },
		{ 70, 275, 46 },
		{ 70, 276, 46 },
		{ 70, 277, 46 },
		{ 70, 278, 46 },
		{ 70, 279, 46 },
		{ 70, 280, 46 },
		{ 71, 234, 46 },
		{ 71, 235, 46 },
		{ 71, 236, 46 },
		{ 71, 237, 46 },
		{ 71, 238, 46 },
		{ 71, 239, 46 },
		{ 71, 240, 46 },
		{ 71, 241, 46 },
		{ 71, 242, 46 },
		{ 71, 243, 46 },
		{ 71, 244, 46 },
		{ 71, 245, 46 },
		{ 71, 246, 46 },
		{ 71, 247, 46 },
		{ 71, 248, 46 },
		{ 71, 249, 46 },
		{ 71, 250, 46 },
		{ 71, 251, 46 },
		{ 71, 252, 46 },
		{ 71, 253, 46 },
		{ 71, 254, 46 },
		{ 71, 255, 46 },
		{ 71, 256, 46 },
		{ 71, 257, 46 },
		{ 71, 258, 46 },
		{ 71, 259, 46 },
		{ 71, 260, 46 },
		{ 71, 261, 46 },
		{ 71, 262, 46 },
		{ 71, 263, 46 },
		{ 71, 264, 46 },
		{ 71, 265, 46 },
		{ 71, 266, 46 },
		{ 71, 267, 46 },
		{ 71, 268, 46 },
		{ 71, 269, 46 },
		{ 71, 270, 46 },
		{ 71, 271, 46 },
		{ 71, 272, 46 },
		{ 71, 273, 46 },
		{ 71, 274, 46 },
		{ 71, 275, 46 },
		{ 71, 276, 46 },
		{ 71, 277, 46 },
		{ 71, 278, 46 },
		{ 71, 279, 46 },
		{ 71, 280, 46 },
		{ 72, 234, 46 },
		{ 72, 235, 46 },
		{ 72, 236, 46 },
		{ 72, 237, 46 },
		{ 72, 238, 46 },
		{ 72, 239, 46 },
		{ 72, 240, 46 },
		{ 72, 241, 46 },
		{ 72, 242, 46 },
		{ 72, 243, 46 },
		{ 72, 244, 46 },
		{ 72, 245, 46 },
		{ 72, 246, 46 },
		{ 72, 247, 46 },
		{ 72, 248, 46 },
		{ 72, 249, 46 },
		{ 72, 250, 46 },
		{ 72, 251, 46 },
		{ 72, 252, 46 },
		{ 72, 253, 46 },
		{ 72, 254, 46 },
		{ 72, 255, 46 },
		{ 72, 256, 46 },
		{ 72, 257, 46 },
		{ 72, 258, 46 },
		{ 72, 259, 46 },
		{ 72, 260, 46 },
		{ 72, 261, 46 },
		{ 72, 262, 46 },
		{ 72, 263, 46 },
		{ 72, 264, 46 },
		{ 72, 265, 46 },
		{ 72, 266, 46 },
		{ 72, 267, 46 },
		{ 72, 268, 46 },
		{ 72, 269, 46 },
		{ 72, 270, 46 },
		{ 72, 271, 46 },
		{ 72, 272, 46 },
		{ 72, 273, 46 },
		{ 72, 274, 46 },
		{ 72, 275, 46 },
		{ 72, 276, 46 },
		{ 72, 277, 46 },
		{ 72, 278, 46 },
		{ 72, 279, 46 },
		{ 72, 280, 46 },
		{ 73, 235, 46 },
		{ 73, 236, 46 },
		{ 73, 237, 46 },
		{ 73, 238, 46 },
		{ 73, 239, 46 },
		{ 73, 240, 46 },
		{ 73, 241, 46 },
		{ 73, 242, 46 },
		{ 73, 243, 46 },
		{ 73, 244, 46 },
		{ 73, 245, 46 },
		{ 73, 246, 46 },
		{ 73, 247, 46 },
		{ 73, 248, 46 },
		{ 73, 249, 46 },
		{ 73, 250, 46 },
		{ 73, 251, 46 },
		{ 73, 252, 46 },
		{ 73, 253, 46 },
		{ 73, 254, 46 },
		{ 73, 255, 46 },
		{ 73, 256, 46 },
		{ 73, 257, 46 },
		{ 73, 258, 46 },
		{ 73, 259, 46 },
		{ 73, 260, 46 },
		{ 73, 261, 46 },
		{ 73, 262, 46 },
		{ 73, 263, 46 },
		{ 73, 264, 46 },
		{ 73, 265, 46 },
		{ 73, 266, 46 },
		{ 73, 267, 46 },
		{ 73, 268, 46 },
		{ 73, 269, 46 },
		{ 73, 270, 46 },
		{ 73, 271, 46 },
		{ 73, 272, 46 },
		{ 73, 273, 46 },
		{ 73, 274, 46 },
		{ 73, 275, 46 },
		{ 73, 276, 46 },
		{ 73, 277, 46 },
		{ 73, 278, 46 },
		{ 73, 279, 46 },
		{ 74, 235, 46 },
		{ 74, 236, 46 },
		{ 74, 237, 46 },
		{ 74, 238, 46 },
		{ 74, 239, 46 },
		{ 74, 240, 46 },
		{ 74, 241, 46 },
		{ 74, 242, 46 },
		{ 74, 243, 46 },
		{ 74, 244, 46 },
		{ 74, 245, 46 },
		{ 74, 246, 46 },
		{ 74, 247, 46 },
		{ 74, 248, 46 },
		{ 74, 249, 46 },
		{ 74, 250, 46 },
		{ 74, 251, 46 },
		{ 74, 252, 46 },
		{ 74, 253, 46 },
		{ 74, 254, 46 },
		{ 74, 255, 46 },
		{ 74, 256, 46 },
		{ 74, 257, 46 },
		{ 74, 258, 46 },
		{ 74, 259, 46 },
		{ 74, 260, 46 },
		{ 74, 261, 46 },
		{ 74, 262, 46 },
		{ 74, 263, 46 },
		{ 74, 264, 46 },
		{ 74, 265, 46 },
		{ 74, 266, 46 },
		{ 74, 267, 46 },
		{ 74, 268, 46 },
		{ 74, 269, 46 },
		{ 74, 270, 46 },
		{ 74, 271, 46 },
		{ 74, 272, 46 },
		{ 74, 273, 46 },
		{ 74, 274, 46 },
		{ 74, 275, 46 },
		{ 74, 276, 46 },
		{ 74, 277, 46 },
		{ 74, 278, 46 },
		{ 74, 279, 46 },
		{ 75, 235, 46 },
		{ 75, 236, 46 },
		{ 75, 237, 46 },
		{ 75, 238, 46 },
		{ 75, 239, 46 },
		{ 75, 240, 46 },
		{ 75, 241, 46 },
		{ 75, 242, 46 },
		{ 75, 243, 46 },
		{ 75, 244, 46 },
		{ 75, 245, 46 },
		{ 75, 246, 46 },
		{ 75, 247, 46 },
		{ 75, 248, 46 },
		{ 75, 249, 46 },
		{ 75, 250, 46 },
		{ 75, 251, 46 },
		{ 75, 252, 46 },
		{ 75, 253, 46 },
		{ 75, 254, 46 },
		{ 75, 255, 46 },
		{ 75, 256, 46 },
		{ 75, 257, 46 },
		{ 75, 258, 46 },
		{ 75, 259, 46 },
		{ 75, 260, 46 },
		{ 75, 261, 46 },
		{ 75, 262, 46 },
		{ 75, 263, 46 },
		{ 75, 264, 46 },
		{ 75, 265, 46 },
		{ 75, 266, 46 },
		{ 75, 267, 46 },
		{ 75, 268, 46 },
		{ 75, 269, 46 },
		{ 75, 270, 46 },
		{ 75, 271, 46 },
		{ 75, 272, 46 },
		{ 75, 273, 46 },
		{ 75, 274, 46 },
		{ 75, 275, 46 },
		{ 75, 276, 46 },
		{ 75, 277, 46 },
		{ 75, 278, 46 },
		{ 75, 279, 46 },
		{ 76, 235, 46 },
		{ 76, 236, 46 },
		{ 76, 237, 46 },
		{ 76, 238, 46 },
		{ 76, 239, 46 },
		{ 76, 240, 46 },
		{ 76, 241, 46 },
		{ 76, 242, 46 },
		{ 76, 243, 46 },
		{ 76, 244, 46 },
		{ 76, 245, 46 },
		{ 76, 246, 46 },
		{ 76, 247, 46 },
		{ 76, 248, 46 },
		{ 76, 249, 46 },
		{ 76, 250, 46 },
		{ 76, 251, 46 },
		{ 76, 252, 46 },
		{ 76, 253, 46 },
		{ 76, 254, 46 },
		{ 76, 255, 46 },
		{ 76, 256, 46 },
		{ 76, 257, 46 },
		{ 76, 258, 46 },
		{ 76, 259, 46 },
		{ 76, 260, 46 },
		{ 76, 261, 46 },
		{ 76, 262, 46 },
		{ 76, 263, 46 },
		{ 76, 264, 46 },
		{ 76, 265, 46 },
		{ 76, 266, 46 },
		{ 76, 267, 46 },
		{ 76, 268, 46 },
		{ 76, 269, 46 },
		{ 76, 270, 46 },
		{ 76, 271, 46 },
		{ 76, 272, 46 },
		{ 76, 273, 46 },
		{ 76, 274, 46 },
		{ 76, 275, 46 },
		{ 76, 276, 46 },
		{ 76, 277, 46 },
		{ 76, 278, 46 },
		{ 76, 279, 46 },
		{ 77, 235, 46 },
		{ 77, 236, 46 },
		{ 77, 237, 46 },
		{ 77, 238, 46 },
		{ 77, 239, 46 },
		{ 77, 240, 46 },
		{ 77, 241, 46 },
		{ 77, 242, 46 },
		{ 77, 243, 46 },
		{ 77, 244, 46 },
		{ 77, 245, 46 },
		{ 77, 246, 46 },
		{ 77, 247, 46 },
		{ 77, 248, 46 },
		{ 77, 249, 46 },
		{ 77, 250, 46 },
		{ 77, 251, 46 },
		{ 77, 252, 46 },
		{ 77, 253, 46 },
		{ 77, 254, 46 },
		{ 77, 255, 46 },
		{ 77, 256, 46 },
		{ 77, 257, 46 },
		{ 77, 258, 46 },
		{ 77, 259, 46 },
		{ 77, 260, 46 },
		{ 77, 261, 46 },
		{ 77, 262, 46 },
		{ 77, 263, 46 },
		{ 77, 264, 46 },
		{ 77, 265, 46 },
		{ 77, 266, 46 },
		{ 77, 267, 46 },
		{ 77, 268, 46 },
		{ 77, 269, 46 },
		{ 77, 270, 46 },
		{ 77, 271, 46 },
		{ 77, 272, 46 },
		{ 77, 273, 46 },
		{ 77, 274, 46 },
		{ 77, 275, 46 },
		{ 77, 276, 46 },
		{ 77, 277, 46 },
		{ 77, 278, 46 },
		{ 77, 279, 46 },
		{ 78, 235, 46 },
		{ 78, 236, 46 },
		{ 78, 237, 46 },
		{ 78, 238, 46 },
		{ 78, 239, 46 },
		{ 78, 240, 46 },
		{ 78, 241, 46 },
		{ 78, 242, 46 },
		{ 78, 243, 46 },
		{ 78, 244, 46 },
		{ 78, 245, 46 },
		{ 78, 246, 46 },
		{ 78, 247, 46 },
		{ 78, 248, 46 },
		{ 78, 249, 46 },
		{ 78, 250, 46 },
		{ 78, 251, 46 },
		{ 78, 252, 46 },
		{ 78, 253, 46 },
		{ 78, 254, 46 },
		{ 78, 255, 46 },
		{ 78, 256, 46 },
		{ 78, 257, 46 },
		{ 78, 258, 46 },
		{ 78, 259, 46 },
		{ 78, 260, 46 },
		{ 78, 261, 46 },
		{ 78, 262, 46 },
		{ 78, 263, 46 },
		{ 78, 264, 46 },
		{ 78, 265, 46 },
		{ 78, 266, 46 },
		{ 78, 267, 46 },
		{ 78, 268, 46 },
		{ 78, 269, 46 },
		{ 78, 270, 46 },
		{ 78, 271, 46 },
		{ 78, 272, 46 },
		{ 78, 273, 46 },
		{ 78, 274, 46 },
		{ 78, 275, 46 },
		{ 78, 276, 46 },
		{ 78, 277, 46 },
		{ 78, 278, 46 },
		{ 78, 279, 46 },
		{ 79, 235, 46 },
		{ 79, 236, 46 },
		{ 79, 237, 46 },
		{ 79, 238, 46 },
		{ 79, 239, 46 },
		{ 79, 240, 46 },
		{ 79, 241, 46 },
		{ 79, 242, 46 },
		{ 79, 243, 46 },
		{ 79, 244, 46 },
		{ 79, 245, 46 },
		{ 79, 246, 46 },
		{ 79, 247, 46 },
		{ 79, 248, 46 },
		{ 79, 249, 46 },
		{ 79, 250, 46 },
		{ 79, 251, 46 },
		{ 79, 252, 46 },
		{ 79, 253, 46 },
		{ 79, 254, 46 },
		{ 79, 255, 46 },
		{ 79, 256, 46 },
		{ 79, 257, 46 },
		{ 79, 258, 46 },
		{ 79, 259, 46 },
		{ 79, 260, 46 },
		{ 79, 261, 46 },
		{ 79, 262, 46 },
		{ 79, 263, 46 },
		{ 79, 264, 46 },
		{ 79, 265, 46 },
		{ 79, 266, 46 },
		{ 79, 267, 46 },
		{ 79, 268, 46 },
		{ 79, 269, 46 },
		{ 79, 270, 46 },
		{ 79, 271, 46 },
		{ 79, 272, 46 },
		{ 79, 273, 46 },
		{ 79, 274, 46 },
		{ 79, 275, 46 },
		{ 79, 276, 46 },
		{ 79, 277, 46 },
		{ 79, 278, 46 },
		{ 79, 279, 46 },
		{ 80, 235, 46 },
		{ 80, 236, 46 },
		{ 80, 237, 46 },
		{ 80, 238, 46 },
		{ 80, 239, 46 },
		{ 80, 240, 46 },
		{ 80, 241, 46 },
		{ 80, 242, 46 },
		{ 80, 243, 46 },
		{ 80, 244, 46 },
		{ 80, 245, 46 },
		{ 80, 246, 46 },
		{ 80, 247, 46 },
		{ 80, 248, 46 },
		{ 80, 249, 46 },
		{ 80, 250, 46 },
		{ 80, 251, 46 },
		{ 80, 252, 46 },
		{ 80, 253, 46 },
		{ 80, 254, 46 },
		{ 80, 255, 46 },
		{ 80, 256, 46 },
		{ 80, 257, 46 },
		{ 80, 258, 46 },
		{ 80, 259, 46 },
		{ 80, 260, 46 },
		{ 80, 261, 46 },
		{ 80, 262, 46 },
		{ 80, 263, 46 },
		{ 80, 264, 46 },
		{ 80, 265, 46 },
		{ 80, 266, 46 },
		{ 80, 267, 46 },
		{ 80, 268, 46 },
		{ 80, 269, 46 },
		{ 80, 270, 46 },
		{ 80, 271, 46 },
		{ 80, 272, 46 },
		{ 80, 273, 46 },
		{ 80, 274, 46 },
		{ 80, 275, 46 },
		{ 80, 276, 46 },
		{ 80, 277, 46 },
		{ 80, 278, 46 },
		{ 80, 279, 46 },
		{ 81, 236, 46 },
		{ 81, 237, 46 },
		{ 81, 238, 46 },
		{ 81, 239, 46 },
		{ 81, 240, 46 },
		{ 81, 241, 46 },
		{ 81, 242, 46 },
		{ 81, 243, 46 },
		{ 81, 244, 46 },
		{ 81, 245, 46 },
		{ 81, 246, 46 },
		{ 81, 247, 46 },
		{ 81, 248, 46 },
		{ 81, 249, 46 },
		{ 81, 250, 46 },
		{ 81, 251, 46 },
		{ 81, 252, 46 },
		{ 81, 253, 46 },
		{ 81, 254, 46 },
		{ 81, 255, 46 },
		{ 81, 256, 46 },
		{ 81, 257, 46 },
		{ 81, 258, 46 },
		{ 81, 259, 46 },
		{ 81, 260, 46 },
		{ 81, 261, 46 },
		{ 81, 262, 46 },
		{ 81, 263, 46 },
		{ 81, 264, 46 },
		{ 81, 265, 46 },
		{ 81, 266, 46 },
		{ 81, 267, 46 },
		{ 81, 268, 46 },
		{ 81, 269, 46 },
		{ 81, 270, 46 },
		{ 81, 271, 46 },
		{ 81, 272, 46 },
		{ 81, 273, 46 },
		{ 81, 274, 46 },
		{ 81, 275, 46 },
		{ 81, 276, 46 },
		{ 81, 277, 46 },
		{ 81, 278, 46 },
		{ 82, 236, 46 },
		{ 82, 237, 46 },
		{ 82, 238, 46 },
		{ 82, 239, 46 },
		{ 82, 240, 46 },
		{ 82, 241, 46 },
		{ 82, 242, 46 },
		{ 82, 243, 46 },
		{ 82, 244, 46 },
		{ 82, 245, 46 },
		{ 82, 246, 46 },
		{ 82, 247, 46 },
		{ 82, 248, 46 },
		{ 82, 249, 46 },
		{ 82, 250, 46 },
		{ 82, 251, 46 },
		{ 82, 252, 46 },
		{ 82, 253, 46 },
		{ 82, 254, 46 },
		{ 82, 255, 46 },
		{ 82, 256, 46 },
		{ 82, 257, 46 },
		{ 82, 258, 46 },
		{ 82, 259, 46 },
		{ 82, 260, 46 },
		{ 82, 261, 46 },
		{ 82, 262, 46 },
		{ 82, 263, 46 },
		{ 82, 264, 46 },
		{ 82, 265, 46 },
		{ 82, 266, 46 },
		{ 82, 267, 46 },
		{ 82, 268, 46 },
		{ 82, 269, 46 },
		{ 82, 270, 46 },
		{ 82, 271, 46 },
		{ 82, 272, 46 },
		{ 82, 273, 46 },
		{ 82, 274, 46 },
		{ 82, 275, 46 },
		{ 82, 276, 46 },
		{ 82, 277, 46 },
		{ 82, 278, 46 },
		{ 83, 236, 46 },
		{ 83, 237, 46 },
		{ 83, 238, 46 },
		{ 83, 239, 46 },
		{ 83, 240, 46 },
		{ 83, 241, 46 },
		{ 83, 242, 46 },
		{ 83, 243, 46 },
		{ 83, 244, 46 },
		{ 83, 245, 46 },
		{ 83, 246, 46 },
		{ 83, 247, 46 },
		{ 83, 248, 46 },
		{ 83, 249, 46 },
		{ 83, 250, 46 },
		{ 83, 251, 46 },
		{ 83, 252, 46 },
		{ 83, 253, 46 },
		{ 83, 254, 46 },
		{ 83, 255, 46 },
		{ 83, 256, 46 },
		{ 83, 257, 46 },
		{ 83, 258, 46 },
		{ 83, 259, 46 },
		{ 83, 260, 46 },
		{ 83, 261, 46 },
		{ 83, 262, 46 },
		{ 83, 263, 46 },
		{ 83, 264, 46 },
		{ 83, 265, 46 },
		{ 83, 266, 46 },
		{ 83, 267, 46 },
		{ 83, 268, 46 },
		{ 83, 269, 46 },
		{ 83, 270, 46 },
		{ 83, 271, 46 },
		{ 83, 272, 46 },
		{ 83, 273, 46 },
		{ 83, 274, 46 },
		{ 83, 275, 46 },
		{ 83, 276, 46 },
		{ 83, 277, 46 },
		{ 83, 278, 46 },
		{ 84, 236, 46 },
		{ 84, 237, 46 },
		{ 84, 238, 46 },
		{ 84, 239, 46 },
		{ 84, 240, 46 },
		{ 84, 241, 46 },
		{ 84, 242, 46 },
		{ 84, 243, 46 },
		{ 84, 244, 46 },
		{ 84, 245, 46 },
		{ 84, 246, 46 },
		{ 84, 247, 46 },
		{ 84, 248, 46 },
		{ 84, 249, 46 },
		{ 84, 250, 46 },
		{ 84, 251, 46 },
		{ 84, 252, 46 },
		{ 84, 253, 46 },
		{ 84, 254, 46 },
		{ 84, 255, 46 },
		{ 84, 256, 46 },
		{ 84, 257, 46 },
		{ 84, 258, 46 },
		{ 84, 259, 46 },
		{ 84, 260, 46 },
		{ 84, 261, 46 },
		{ 84, 262, 46 },
		{ 84, 263, 46 },
		{ 84, 264, 46 },
		{ 84, 265, 46 },
		{ 84, 266, 46 },
		{ 84, 267, 46 },
		{ 84, 268, 46 },
		{ 84, 269, 46 },
		{ 84, 270, 46 },
		{ 84, 271, 46 },
		{ 84, 272, 46 },
		{ 84, 273, 46 },
		{ 84, 274, 46 },
		{ 84, 275, 46 },
		{ 84, 276, 46 },
		{ 84, 277, 46 },
		{ 84, 278, 46 },
		{ 85, 236, 46 },
		{ 85, 237, 46 },
		{ 85, 238, 46 },
		{ 85, 239, 46 },
		{ 85, 240, 46 },
		{ 85, 241, 46 },
		{ 85, 242, 46 },
		{ 85, 243, 46 },
		{ 85, 244, 46 },
		{ 85, 245, 46 },
		{ 85, 246, 46 },
		{ 85, 247, 46 },
		{ 85, 248, 46 },
		{ 85, 249, 46 },
		{ 85, 250, 46 },
		{ 85, 251, 46 },
		{ 85, 252, 46 },
		{ 85, 253, 46 },
		{ 85, 254, 46 },
		{ 85, 255, 46 },
		{ 85, 256, 46 },
		{ 85, 257, 46 },
		{ 85, 258, 46 },
		{ 85, 259, 46 },
		{ 85, 260, 46 },
		{ 85, 261, 46 },
		{ 85, 262, 46 },
		{ 85, 263, 46 },
		{ 85, 264, 46 },
		{ 85, 265, 46 },
		{ 85, 266, 46 },
		{ 85, 267, 46 },
		{ 85, 268, 46 },
		{ 85, 269, 46 },
		{ 85, 270, 46 },
		{ 85, 271, 46 },
		{ 85, 272, 46 },
		{ 85, 273, 46 },
		{ 85, 274, 46 },
		{ 85, 275, 46 },
		{ 85, 276, 46 },
		{ 85, 277, 46 },
		{ 85, 278, 46 },
		{ 86, 236, 46 },
		{ 86, 237, 46 },
		{ 86, 238, 46 },
		{ 86, 239, 46 },
		{ 86, 240, 46 },
		{ 86, 241, 46 },
		{ 86, 242, 46 },
		{ 86, 243, 46 },
		{ 86, 244, 46 },
		{ 86, 245, 46 },
		{ 86, 246, 46 },
		{ 86, 247, 46 },
		{ 86, 248, 46 },
		{ 86, 249, 46 },
		{ 86, 250, 46 },
		{ 86, 251, 46 },
		{ 86, 252, 46 },
		{ 86, 253, 46 },
		{ 86, 254, 46 },
		{ 86, 255, 46 },
		{ 86, 256, 46 },
		{ 86, 257, 46 },
		{ 86, 258, 46 },
		{ 86, 259, 46 },
		{ 86, 260, 46 },
		{ 86, 261, 46 },
		{ 86, 262, 46 },
		{ 86, 263, 46 },
		{ 86, 264, 46 },
		{ 86, 265, 46 },
		{ 86, 266, 46 },
		{ 86, 267, 46 },
		{ 86, 268, 46 },
		{ 86, 269, 46 },
		{ 86, 270, 46 },
		{ 86, 271, 46 },
		{ 86, 272, 46 },
		{ 86, 273, 46 },
		{ 86, 274, 46 },
		{ 86, 275, 46 },
		{ 86, 276, 46 },
		{ 86, 277, 46 },
		{ 86, 278, 46 },
		{ 87, 236, 46 },
		{ 87, 237, 46 },
		{ 87, 238, 46 },
		{ 87, 239, 46 },
		{ 87, 240, 46 },
		{ 87, 241, 46 },
		{ 87, 242, 46 },
		{ 87, 243, 46 },
		{ 87, 244, 46 },
		{ 87, 245, 46 },
		{ 87, 246, 46 },
		{ 87, 247, 46 },
		{ 87, 248, 46 },
		{ 87, 249, 46 },
		{ 87, 250, 46 },
		{ 87, 251, 46 },
		{ 87, 252, 46 },
		{ 87, 253, 46 },
		{ 87, 254, 46 },
		{ 87, 255, 46 },
		{ 87, 256, 46 },
		{ 87, 257, 46 },
		{ 87, 258, 46 },
		{ 87, 259, 46 },
		{ 87, 260, 46 },
		{ 87, 261, 46 },
		{ 87, 262, 46 },
		{ 87, 263, 46 },
		{ 87, 264, 46 },
		{ 87, 265, 46 },
		{ 87, 266, 46 },
		{ 87, 267, 46 },
		{ 87, 268, 46 },
		{ 87, 269, 46 },
		{ 87, 270, 46 },
		{ 87, 271, 46 },
		{ 87, 272, 46 },
		{ 87, 273, 46 },
		{ 87, 274, 46 },
		{ 87, 275, 46 },
		{ 87, 276, 46 },
		{ 87, 277, 46 },
		{ 87, 278, 46 },
		{ 88, 236, 46 },
		{ 88, 237, 46 },
		{ 88, 238, 46 },
		{ 88, 239, 46 },
		{ 88, 240, 46 },
		{ 88, 241, 46 },
		{ 88, 242, 46 },
		{ 88, 243, 46 },
		{ 88, 244, 46 },
		{ 88, 245, 46 },
		{ 88, 246, 46 },
		{ 88, 247, 46 },
		{ 88, 248, 46 },
		{ 88, 249, 46 },
		{ 88, 250, 46 },
		{ 88, 251, 46 },
		{ 88, 252, 46 },
		{ 88, 253, 46 },
		{ 88, 254, 46 },
		{ 88, 255, 46 },
		{ 88, 256, 46 },
		{ 88, 257, 46 },
		{ 88, 258, 46 },
		{ 88, 259, 46 },
		{ 88, 260, 46 },
		{ 88, 261, 46 },
		{ 88, 262, 46 },
		{ 88, 263, 46 },
		{ 88, 264, 46 },
		{ 88, 265, 46 },
		{ 88, 266, 46 },
		{ 88, 267, 46 },
		{ 88, 268, 46 },
		{ 88, 269, 46 },
		{ 88, 270, 46 },
		{ 88, 271, 46 },
		{ 88, 272, 46 },
		{ 88, 273, 46 },
		{ 88, 274, 46 },
		{ 88, 275, 46 },
		{ 88, 276, 46 },
		{ 88, 277, 46 },
		{ 88, 278, 46 },
		{ 89, 237, 46 },
		{ 89, 238, 46 },
		{ 89, 239, 46 },
		{ 89, 240, 46 },
		{ 89, 241, 46 },
		{ 89, 242, 46 },
		{ 89, 243, 46 },
		{ 89, 244, 46 },
		{ 89, 245, 46 },
		{ 89, 246, 46 },
		{ 89, 247, 46 },
		{ 89, 248, 46 },
		{ 89, 249, 46 },
		{ 89, 250, 46 },
		{ 89, 251, 46 },
		{ 89, 252, 46 },
		{ 89, 253, 46 },
		{ 89, 254, 46 },
		{ 89, 255, 46 },
		{ 89, 256, 46 },
		{ 89, 257, 46 },
		{ 89, 258, 46 },
		{ 89, 259, 46 },
		{ 89, 260, 46 },
		{ 89, 261, 46 },
		{ 89, 262, 46 },
		{ 89, 263, 46 },
		{ 89, 264, 46 },
		{ 89, 265, 46 },
		{ 89, 266, 46 },
		{ 89, 267, 46 },
		{ 89, 268, 46 },
		{ 89, 269, 46 },
		{ 89, 270, 46 },
		{ 89, 271, 46 },
		{ 89, 272, 46 },
		{ 89, 273, 46 },
		{ 89, 274, 46 },
		{ 89, 275, 46 },
		{ 89, 276, 46 },
		{ 89, 277, 46 },
		{ 90, 237, 46 },
		{ 90, 238, 46 },
		{ 90, 239, 46 },
		{ 90, 240, 46 },
		{ 90, 241, 46 },
		{ 90, 242, 46 },
		{ 90, 243, 46 },
		{ 90, 244, 46 },
		{ 90, 245, 46 },
		{ 90, 246, 46 },
		{ 90, 247, 46 },
		{ 90, 248, 46 },
		{ 90, 249, 46 },
		{ 90, 250, 46 },
		{ 90, 251, 46 },
		{ 90, 252, 46 },
		{ 90, 253, 46 },
		{ 90, 254, 46 },
		{ 90, 255, 46 },
		{ 90, 256, 46 },
		{ 90, 257, 46 },
		{ 90, 258, 46 },
		{ 90, 259, 46 },
		{ 90, 260, 46 },
		{ 90, 261, 46 },
		{ 90, 262, 46 },
		{ 90, 263, 46 },
		{ 90, 264, 46 },
		{ 90, 265, 46 },
		{ 90, 266, 46 },
		{ 90, 267, 46 },
		{ 90, 268, 46 },
		{ 90, 269, 46 },
		{ 90, 270, 46 },
		{ 90, 271, 46 },
		{ 90, 272, 46 },
		{ 90, 273, 46 },
		{ 90, 274, 46 },
		{ 90, 275, 46 },
		{ 90, 276, 46 },
		{ 90, 277, 46 },
		{ 91, 237, 46 },
		{ 91, 238, 46 },
		{ 91, 239, 46 },
		{ 91, 240, 46 },
		{ 91, 241, 46 },
		{ 91, 242, 46 },
		{ 91, 243, 46 },
		{ 91, 244, 46 },
		{ 91, 245, 46 },
		{ 91, 246, 46 },
		{ 91, 247, 46 },
		{ 91, 248, 46 },
		{ 91, 249, 46 },
		{ 91, 250, 46 },
		{ 91, 251, 46 },
		{ 91, 252, 46 },
		{ 91, 253, 46 },
		{ 91, 254, 46 },
		{ 91, 255, 46 },
		{ 91, 256, 46 },
		{ 91, 257, 46 },
		{ 91, 258, 46 },
		{ 91, 259, 46 },
		{ 91, 260, 46 },
		{ 91, 261, 46 },
		{ 91, 262, 46 },
		{ 91, 263, 46 },
		{ 91, 264, 46 },
		{ 91, 265, 46 },
		{ 91, 266, 46 },
		{ 91, 267, 46 },
		{ 91, 268, 46 },
		{ 91, 269, 46 },
		{ 91, 270, 46 },
		{ 91, 271, 46 },
		{ 91, 272, 46 },
		{ 91, 273, 46 },
		{ 91, 274, 46 },
		{ 91, 275, 46 },
		{ 91, 276, 46 },
		{ 91, 277, 46 },
		{ 92, 237, 46 },
		{ 92, 238, 46 },
		{ 92, 239, 46 },
		{ 92, 240, 46 },
		{ 92, 241, 46 },
		{ 92, 242, 46 },
		{ 92, 243, 46 },
		{ 92, 244, 46 },
		{ 92, 245, 46 },
		{ 92, 246, 46 },
		{ 92, 247, 46 },
		{ 92, 248, 46 },
		{ 92, 249, 46 },
		{ 92, 250, 46 },
		{ 92, 251, 46 },
		{ 92, 252, 46 },
		{ 92, 253, 46 },
		{ 92, 254, 46 },
		{ 92, 255, 46 },
		{ 92, 256, 46 },
		{ 92, 257, 46 },
		{ 92, 258, 46 },
		{ 92, 259, 46 },
		{ 92, 260, 46 },
		{ 92, 261, 46 },
		{ 92, 262, 46 },
		{ 92, 263, 46 },
		{ 92, 264, 46 },
		{ 92, 265, 46 },
		{ 92, 266, 46 },
		{ 92, 267, 46 },
		{ 92, 268, 46 },
		{ 92, 269, 46 },
		{ 92, 270, 46 },
		{ 92, 271, 46 },
		{ 92, 272, 46 },
		{ 92, 273, 46 },
		{ 92, 274, 46 },
		{ 92, 275, 46 },
		{ 92, 276, 46 },
		{ 92, 277, 46 },
		{ 93, 237, 46 },
		{ 93, 238, 46 },
		{ 93, 239, 46 },
		{ 93, 240, 46 },
		{ 93, 241, 46 },
		{ 93, 242, 46 },
		{ 93, 243, 46 },
		{ 93, 244, 46 },
		{ 93, 245, 46 },
		{ 93, 246, 46 },
		{ 93, 247, 46 },
		{ 93, 248, 46 },
		{ 93, 249, 46 },
		{ 93, 250, 46 },
		{ 93, 251, 46 },
		{ 93, 252, 46 },
		{ 93, 253, 46 },
		{ 93, 254, 46 },
		{ 93, 255, 46 },
		{ 93, 256, 46 },
		{ 93, 257, 46 },
		{ 93, 258, 46 },
		{ 93, 259, 46 },
		{ 93, 260, 46 },
		{ 93, 261, 46 },
		{ 93, 262, 46 },
		{ 93, 263, 46 },
		{ 93, 264, 46 },
		{ 93, 265, 46 },
		{ 93, 266, 46 },
		{ 93, 267, 46 },
		{ 93, 268, 46 },
		{ 93, 269, 46 },
		{ 93, 270, 46 },
		{ 93, 271, 46 },
		{ 93, 272, 46 },
		{ 93, 273, 46 },
		{ 93, 274, 46 },
		{ 93, 275, 46 },
		{ 93, 276, 46 },
		{ 93, 277, 46 },
		{ 94, 237, 46 },
		{ 94, 238, 46 },
		{ 94, 239, 46 },
		{ 94, 240, 46 },
		{ 94, 241, 46 },
		{ 94, 242, 46 },
		{ 94, 243, 46 },
		{ 94, 244, 46 },
		{ 94, 245, 46 },
		{ 94, 246, 46 },
		{ 94, 247, 46 },
		{ 94, 248, 46 },
		{ 94, 249, 46 },
		{ 94, 250, 46 },
		{ 94, 251, 46 },
		{ 94, 252, 46 },
		{ 94, 253, 46 },
		{ 94, 254, 46 },
		{ 94, 255, 46 },
		{ 94, 256, 46 },
		{ 94, 257, 46 },
		{ 94, 258, 46 },
		{ 94, 259, 46 },
		{ 94, 260, 46 },
		{ 94, 261, 46 },
		{ 94, 262, 46 },
		{ 94, 263, 46 },
		{ 94, 264, 46 },
		{ 94, 265, 46 },
		{ 94, 266, 46 },
		{ 94, 267, 46 },
		{ 94, 268, 46 },
		{ 94, 269, 46 },
		{ 94, 270, 46 },
		{ 94, 271, 46 },
		{ 94, 272, 46 },
		{ 94, 273, 46 },
		{ 94, 274, 46 },
		{ 94, 275, 46 },
		{ 94, 276, 46 },
		{ 94, 277, 46 },
		{ 95, 237, 46 },
		{ 95, 238, 46 },
		{ 95, 239, 46 },
		{ 95, 240, 46 },
		{ 95, 241, 46 },
		{ 95, 242, 46 },
		{ 95, 243, 46 },
		{ 95, 244, 46 },
		{ 95, 245, 46 },
		{ 95, 246, 46 },
		{ 95, 247, 46 },
		{ 95, 248, 46 },
		{ 95, 249, 46 },
		{ 95, 250, 46 },
		{ 95, 251, 46 },
		{ 95, 252, 46 },
		{ 95, 253, 46 },
		{ 95, 254, 46 },
		{ 95, 255, 46 },
		{ 95, 256, 46 },
		{ 95, 257, 46 },
		{ 95, 258, 46 },
		{ 95, 259, 46 },
		{ 95, 260, 46 },
		{ 95, 261, 46 },
		{ 95, 262, 46 },
		{ 95, 263, 46 },
		{ 95, 264, 46 },
		{ 95, 265, 46 },
		{ 95, 266, 46 },
		{ 95, 267, 46 },
		{ 95, 268, 46 },
		{ 95, 269, 46 },
		{ 95, 270, 46 },
		{ 95, 271, 46 },
		{ 95, 272, 46 },
		{ 95, 273, 46 },
		{ 95, 274, 46 },
		{ 95, 275, 46 },
		{ 95, 276, 46 },
		{ 95, 277, 46 },
		{ 96, 237, 46 },
		{ 96, 238, 46 },
		{ 96, 239, 46 },
		{ 96, 240, 46 },
		{ 96, 241, 46 },
		{ 96, 242, 46 },
		{ 96, 243, 46 },
		{ 96, 244, 46 },
		{ 96, 245, 46 },
		{ 96, 246, 46 },
		{ 96, 247, 46 },
		{ 96, 248, 46 },
		{ 96, 249, 46 },
		{ 96, 250, 46 },
		{ 96, 251, 46 },
		{ 96, 252, 46 },
		{ 96, 253, 46 },
		{ 96, 254, 46 },
		{ 96, 255, 46 },
		{ 96, 256, 46 },
		{ 96, 257, 46 },
		{ 96, 258, 46 },
		{ 96, 259, 46 },
		{ 96, 260, 46 },
		{ 96, 261, 46 },
		{ 96, 262, 46 },
		{ 96, 263, 46 },
		{ 96, 264, 46 },
		{ 96, 265, 46 },
		{ 96, 266, 46 },
		{ 96, 267, 46 },
		{ 96, 268, 46 },
		{ 96, 269, 46 },
		{ 96, 270, 46 },
		{ 96, 271, 46 },
		{ 96, 272, 46 },
		{ 96, 273, 46 },
		{ 96, 274, 46 },
		{ 96, 275, 46 },
		{ 96, 276, 46 },
		{ 96, 277, 46 },
		{ 97, 238, 46 },
		{ 97, 239, 46 },
		{ 97, 240, 46 },
		{ 97, 241, 46 },
		{ 97, 242, 46 },
		{ 97, 243, 46 },
		{ 97, 244, 46 },
		{ 97, 245, 46 },
		{ 97, 246, 46 },
		{ 97, 247, 46 },
		{ 97, 248, 46 },
		{ 97, 249, 46 },
		{ 97, 250, 46 },
		{ 97, 251, 46 },
		{ 97, 252, 46 },
		{ 97, 253, 46 },
		{ 97, 254, 46 },
		{ 97, 255, 46 },
		{ 97, 256, 46 },
		{ 97, 257, 46 },
		{ 97, 258, 46 },
		{ 97, 259, 46 },
		{ 97, 260, 46 },
		{ 97, 261, 46 },
		{ 97, 262, 46 },
		{ 97, 263, 46 },
		{ 97, 264, 46 },
		{ 97, 265, 46 },
		{ 97, 266, 46 },
		{ 97, 267, 46 },
		{ 97, 268, 46 },
		{ 97, 269, 46 },
		{ 97, 270, 46 },
		{ 97, 271, 46 },
		{ 97, 272, 46 },
		{ 97, 273, 46 },
		{ 97, 274, 46 },
		{ 97, 275, 46 },
		{ 97, 276, 46 },
		{ 98, 238, 46 },
		{ 98, 239, 46 },
		{ 98, 240, 46 },
		{ 98, 241, 46 },
		{ 98, 242, 46 },
		{ 98, 243, 46 },
		{ 98, 244, 46 },
		{ 98, 245, 46 },
		{ 98, 246, 46 },
		{ 98, 247, 46 },
		{ 98, 248, 46 },
		{ 98, 249, 46 },
		{ 98, 250, 46 },
		{ 98, 251, 46 },
		{ 98, 252, 46 },
		{ 98, 253, 46 },
		{ 98, 254, 46 },
		{ 98, 255, 46 },
		{ 98, 256, 46 },
		{ 98, 257, 46 },
		{ 98, 258, 46 },
		{ 98, 259, 46 },
		{ 98, 260, 46 },
		{ 98, 261, 46 },
		{ 98, 262, 46 },
		{ 98, 263, 46 },
		{ 98, 264, 46 },
		{ 98, 265, 46 },
		{ 98, 266, 46 },
		{ 98, 267, 46 },
		{ 98, 268, 46 },
		{ 98, 269, 46 },
		{ 98, 270, 46 },
		{ 98, 271, 46 },
		{ 98, 272, 46 },
		{ 98, 273, 46 },
		{ 98, 274, 46 },
		{ 98, 275, 46 },
		{ 98, 276, 46 },
		{ 99, 238, 46 },
		{ 99, 239, 46 },
		{ 99, 240, 46 },
		{ 99, 241, 46 },
		{ 99, 242, 46 },
		{ 99, 243, 46 },
		{ 99, 244, 46 },
		{ 99, 245, 46 },
		{ 99, 246, 46 },
		{ 99, 247, 46 },
		{ 99, 248, 46 },
		{ 99, 249, 46 },
		{ 99, 250, 46 },
		{ 99, 251, 46 },
		{ 99, 252, 46 },
		{ 99, 253, 46 },
		{ 99, 254, 46 },
		{ 99, 255, 46 },
		{ 99, 256, 46 },
		{ 99, 257, 46 },
		{ 99, 258, 46 },
		{ 99, 259, 46 },
		{ 99, 260, 46 },
		{ 99, 261, 46 },
		{ 99, 262, 46 },
		{ 99, 263, 46 },
		{ 99, 264, 46 },
		{ 99, 265, 46 },
		{ 99, 266, 46 },
		{ 99, 267, 46 },
		{ 99, 268, 46 },
		{ 99, 269, 46 },
		{ 99, 270, 46 },
		{ 99, 271, 46 },
		{ 99, 272, 46 },
		{ 99, 273, 46 },
		{ 99, 274, 46 },
		{ 99, 275, 46 },
		{ 99, 276, 46 },
		{ 100, 238, 46 },
		{ 100, 239, 46 },
		{ 100, 240, 46 },
		{ 100, 241, 46 },
		{ 100, 242, 46 },
		{ 100, 243, 46 },
		{ 100, 244, 46 },
		{ 100, 245, 46 },
		{ 100, 246, 46 },
		{ 100, 247, 46 },
		{ 100, 248, 46 },
		{ 100, 249, 46 },
		{ 100, 250, 46 },
		{ 100, 251, 46 },
		{ 100, 252, 46 },
		{ 100, 253, 46 },
		{ 100, 254, 46 },
		{ 100, 255, 46 },
		{ 100, 256, 46 },
		{ 100, 257, 46 },
		{ 100, 258, 46 },
		{ 100, 259, 46 },
		{ 100, 260, 46 },
		{ 100, 261, 46 },
		{ 100, 262, 46 },
		{ 100, 263, 46 },
		{ 100, 264, 46 },
		{ 100, 265, 46 },
		{ 100, 266, 46 },
		{ 100, 267, 46 },
		{ 100, 268, 46 },
		{ 100, 269, 46 },
		{ 100, 270, 46 },
		{ 100, 271, 46 },
		{ 100, 272, 46 },
		{ 100, 273, 46 },
		{ 100, 274, 46 },
		{ 100, 275, 46 },
		{ 100, 276, 46 },
		{ 101, 238, 46 },
		{ 101, 239, 46 },
		{ 101, 240, 46 },
		{ 101, 241, 46 },
		{ 101, 242, 46 },
		{ 101, 243, 46 },
		{ 101, 244, 46 },
		{ 101, 245, 46 },
		{ 101, 246, 46 },
		{ 101, 247, 46 },
		{ 101, 248, 46 },
		{ 101, 249, 46 },
		{ 101, 250, 46 },
		{ 101, 251, 46 },
		{ 101, 252, 46 },
		{ 101, 253, 46 },
		{ 101, 254, 46 },
		{ 101, 255, 46 },
		{ 101, 256, 46 },
		{ 101, 257, 46 },
		{ 101, 258, 46 },
		{ 101, 259, 46 },
		{ 101, 260, 46 },
		{ 101, 261, 46 },
		{ 101, 262, 46 },
		{ 101, 263, 46 },
		{ 101, 264, 46 },
		{ 101, 265, 46 },
		{ 101, 266, 46 },
		{ 101, 267, 46 },
		{ 101, 268, 46 },
		{ 101, 269, 46 },
		{ 101, 270, 46 },
		{ 101, 271, 46 },
		{ 101, 272, 46 },
		{ 101, 273, 46 },
		{ 101, 274, 46 },
		{ 101, 275, 46 },
		{ 101, 276, 46 },
		{ 102, 238, 46 },
		{ 102, 239, 46 },
		{ 102, 240, 46 },
		{ 102, 241, 46 },
		{ 102, 242, 46 },
		{ 102, 243, 46 },
		{ 102, 244, 46 },
		{ 102, 245, 46 },
		{ 102, 246, 46 },
		{ 102, 247, 46 },
		{ 102, 248, 46 },
		{ 102, 249, 46 },
		{ 102, 250, 46 },
		{ 102, 251, 46 },
		{ 102, 252, 46 },
		{ 102, 253, 46 },
		{ 102, 254, 46 },
		{ 102, 255, 46 },
		{ 102, 256, 46 },
		{ 102, 257, 46 },
		{ 102, 258, 46 },
		{ 102, 259, 46 },
		{ 102, 260, 46 },
		{ 102, 261, 46 },
		{ 102, 262, 46 },
		{ 102, 263, 46 },
		{ 102, 264, 46 },
		{ 102, 265, 46 },
		{ 102, 266, 46 },
		{ 102, 267, 46 },
		{ 102, 268, 46 },
		{ 102, 269, 46 },
		{ 102, 270, 46 },
		{ 102, 271, 46 },
		{ 102, 272, 46 },
		{ 102, 273, 46 },
		{ 102, 274, 46 },
		{ 102, 275, 46 },
		{ 102, 276, 46 },
		{ 103, 238, 46 },
		{ 103, 239, 46 },
		{ 103, 240, 46 },
		{ 103, 241, 46 },
		{ 103, 242, 46 },
		{ 103, 243, 46 },
		{ 103, 244, 46 },
		{ 103, 245, 46 },
		{ 103, 246, 46 },
		{ 103, 247, 46 },
		{ 103, 248, 46 },
		{ 103, 249, 46 },
		{ 103, 250, 46 },
		{ 103, 251, 46 },
		{ 103, 252, 46 },
		{ 103, 253, 46 },
		{ 103, 254, 46 },
		{ 103, 255, 46 },
		{ 103, 256, 46 },
		{ 103, 257, 46 },
		{ 103, 258, 46 },
		{ 103, 259, 46 },
		{ 103, 260, 46 },
		{ 103, 261, 46 },
		{ 103, 262, 46 },
		{ 103, 263, 46 },
		{ 103, 264, 46 },
		{ 103, 265, 46 },
		{ 103, 266, 46 },
		{ 103, 267, 46 },
		{ 103, 268, 46 },
		{ 103, 269, 46 },
		{ 103, 270, 46 },
		{ 103, 271, 46 },
		{ 103, 272, 46 },
		{ 103, 273, 46 },
		{ 103, 274, 46 },
		{ 103, 275, 46 },
		{ 103, 276, 46 },
		{ 104, 238, 46 },
		{ 104, 239, 46 },
		{ 104, 240, 46 },
		{ 104, 241, 46 },
		{ 104, 242, 46 },
		{ 104, 243, 46 },
		{ 104, 244, 46 },
		{ 104, 245, 46 },
		{ 104, 246, 46 },
		{ 104, 247, 46 },
		{ 104, 248, 46 },
		{ 104, 249, 46 },
		{ 104, 250, 46 },
		{ 104, 251, 46 },
		{ 104, 252, 46 },
		{ 104, 253, 46 },
		{ 104, 254, 46 },
		{ 104, 255, 46 },
		{ 104, 256, 46 },
		{ 104, 257, 46 },
		{ 104, 258, 46 },
		{ 104, 259, 46 },
		{ 104, 260, 46 },
		{ 104, 261, 46 },
		{ 104, 262, 46 },
		{ 104, 263, 46 },
		{ 104, 264, 46 },
		{ 104, 265, 46 },
		{ 104, 266, 46 },
		{ 104, 267, 46 },
		{ 104, 268, 46 },
		{ 104, 269, 46 },
		{ 104, 270, 46 },
		{ 104, 271, 46 },
		{ 104, 272, 46 },
		{ 104, 273, 46 },
		{ 104, 274, 46 },
		{ 104, 275, 46 },
		{ 104, 276, 46 },
		{ 105, 239, 46 },
		{ 105, 240, 46 },
		{ 105, 241, 46 },
		{ 105, 242, 46 },
		{ 105, 243, 46 },
		{ 105, 244, 46 },
		{ 105, 245, 46 },
		{ 105, 246, 46 },
		{ 105, 247, 46 },
		{ 105, 248, 46 },
		{ 105, 249, 46 },
		{ 105, 250, 46 },
		{ 105, 251, 46 },
		{ 105, 252, 46 },
		{ 105, 253, 46 },
		{ 105, 254, 46 },
		{ 105, 255, 46 },
		{ 105, 256, 46 },
		{ 105, 257, 46 },
		{ 105, 258, 46 },
		{ 105, 259, 46 },
		{ 105, 260, 46 },
		{ 105, 261, 46 },
		{ 105, 262, 46 },
		{ 105, 263, 46 },
		{ 105, 264, 46 },
		{ 105, 265, 46 },
		{ 105, 266, 46 },
		{ 105, 267, 46 },
		{ 105, 268, 46 },
		{ 105, 269, 46 },
		{ 105, 270, 46 },
		{ 105, 271, 46 },
		{ 105, 272, 46 },
		{ 105, 273, 46 },
		{ 105, 274, 46 },
		{ 105, 275, 46 },
		{ 106, 239, 46 },
		{ 106, 240, 46 },
		{ 106, 241, 46 },
		{ 106, 242, 46 },
		{ 106, 243, 46 },
		{ 106, 244, 46 },
		{ 106, 245, 46 },
		{ 106, 246, 46 },
		{ 106, 247, 46 },
		{ 106, 248, 46 },
		{ 106, 249, 46 },
		{ 106, 250, 46 },
		{ 106, 251, 46 },
		{ 106, 252, 46 },
		{ 106, 253, 46 },
		{ 106, 254, 46 },
		{ 106, 255, 46 },
		{ 106, 256, 46 },
		{ 106, 257, 46 },
		{ 106, 258, 46 },
		{ 106, 259, 46 },
		{ 106, 260, 46 },
		{ 106, 261, 46 },
		{ 106, 262, 46 },
		{ 106, 263, 46 },
		{ 106, 264, 46 },
		{ 106, 265, 46 },
		{ 106, 266, 46 },
		{ 106, 267, 46 },
		{ 106, 268, 46 },
		{ 106, 269, 46 },
		{ 106, 270, 46 },
		{ 106, 271, 46 },
		{ 106, 272, 46 },
		{ 106, 273, 46 },
		{ 106, 274, 46 },
		{ 106, 275, 46 },
		{ 107, 239, 46 },
		{ 107, 240, 46 },
		{ 107, 241, 46 },
		{ 107, 242, 46 },
		{ 107, 243, 46 },
		{ 107, 244, 46 },
		{ 107, 245, 46 },
		{ 107, 246, 46 },
		{ 107, 247, 46 },
		{ 107, 248, 46 },
		{ 107, 249, 46 },
		{ 107, 250, 46 },
		{ 107, 251, 46 },
		{ 107, 252, 46 },
		{ 107, 253, 46 },
		{ 107, 254, 46 },
		{ 107, 255, 46 },
		{ 107, 256, 46 },
		{ 107, 257, 46 },
		{ 107, 258, 46 },
		{ 107, 259, 46 },
		{ 107, 260, 46 },
		{ 107, 261, 46 },
		{ 107, 262, 46 },
		{ 107, 263, 46 },
		{ 107, 264, 46 },
		{ 107, 265, 46 },
		{ 107, 266, 46 },
		{ 107, 267, 46 },
		{ 107, 268, 46 },
		{ 107, 269, 46 },
		{ 107, 270, 46 },
		{ 107, 271, 46 },
		{ 107, 272, 46 },
		{ 107, 273, 46 },
		{ 107, 274, 46 },
		{ 107, 275, 46 },
		{ 108, 239, 46 },
		{ 108, 240, 46 },
		{ 108, 241, 46 },
		{ 108, 242, 46 },
		{ 108, 243, 46 },
		{ 108, 244, 46 },
		{ 108, 245, 46 },
		{ 108, 246, 46 },
		{ 108, 247, 46 },
		{ 108, 248, 46 },
		{ 108, 249, 46 },
		{ 108, 250, 46 },
		{ 108, 251, 46 },
		{ 108, 252, 46 },
		{ 108, 253, 46 },
		{ 108, 254, 46 },
		{ 108, 255, 46 },
		{ 108, 256, 46 },
		{ 108, 257, 46 },
		{ 108, 258, 46 },
		{ 108, 259, 46 },
		{ 108, 260, 46 },
		{ 108, 261, 46 },
		{ 108, 262, 46 },
		{ 108, 263, 46 },
		{ 108, 264, 46 },
		{ 108, 265, 46 },
		{ 108, 266, 46 },
		{ 108, 267, 46 },
		{ 108, 268, 46 },
		{ 108, 269, 46 },
		{ 108, 270, 46 },
		{ 108, 271, 46 },
		{ 108, 272, 46 },
		{ 108, 273, 46 },
		{ 108, 274, 46 },
		{ 108, 275, 46 },
		{ 109, 239, 46 },
		{ 109, 240, 46 },
		{ 109, 241, 46 },
		{ 109, 242, 46 },
		{ 109, 243, 46 },
		{ 109, 244, 46 },
		{ 109, 245, 46 },
		{ 109, 246, 46 },
		{ 109, 247, 46 },
		{ 109, 248, 46 },
		{ 109, 249, 46 },
		{ 109, 250, 46 },
		{ 109, 251, 46 },
		{ 109, 252, 46 },
		{ 109, 253, 46 },
		{ 109, 254, 46 },
		{ 109, 255, 46 },
		{ 109, 256, 46 },
		{ 109, 257, 46 },
		{ 109, 258, 46 },
		{ 109, 259, 46 },
		{ 109, 260, 46 },
		{ 109, 261, 46 },
		{ 109, 262, 46 },
		{ 109, 263, 46 },
		{ 109, 264, 46 },
		{ 109, 265, 46 },
		{ 109, 266, 46 },
		{ 109, 267, 46 },
		{ 109, 268, 46 },
		{ 109, 269, 46 },
		{ 109, 270, 46 },
		{ 109, 271, 46 },
		{ 109, 272, 46 },
		{ 109, 273, 46 },
		{ 109, 274, 46 },
		{ 109, 275, 46 },
		{ 110, 239, 46 },
		{ 110, 240, 46 },
		{ 110, 241, 46 },
		{ 110, 242, 46 },
		{ 110, 243, 46 },
		{ 110, 244, 46 },
		{ 110, 245, 46 },
		{ 110, 246, 46 },
		{ 110, 247, 46 },
		{ 110, 248, 46 },
		{ 110, 249, 46 },
		{ 110, 250, 46 },
		{ 110, 251, 46 },
		{ 110, 252, 46 },
		{ 110, 253, 46 },
		{ 110, 254, 46 },
		{ 110, 255, 46 },
		{ 110, 256, 46 },
		{ 110, 257, 46 },
		{ 110, 258, 46 },
		{ 110, 259, 46 },
		{ 110, 260, 46 },
		{ 110, 261, 46 },
		{ 110, 262, 46 },
		{ 110, 263, 46 },
		{ 110, 264, 46 },
		{ 110, 265, 46 },
		{ 110, 266, 46 },
		{ 110, 267, 46 },
		{ 110, 268, 46 },
		{ 110, 269, 46 },
		{ 110, 270, 46 },
		{ 110, 271, 46 },
		{ 110, 272, 46 },
		{ 110, 273, 46 },
		{ 110, 274, 46 },
		{ 110, 275, 46 },
		{ 111, 239, 46 },
		{ 111, 240, 46 },
		{ 111, 241, 46 },
		{ 111, 242, 46 },
		{ 111, 243, 46 },
		{ 111, 244, 46 },
		{ 111, 245, 46 },
		{ 111, 246, 46 },
		{ 111, 247, 46 },
		{ 111, 248, 46 },
		{ 111, 249, 46 },
		{ 111, 250, 46 },
		{ 111, 251, 46 },
		{ 111, 252, 46 },
		{ 111, 253, 46 },
		{ 111, 254, 46 },
		{ 111, 255, 46 },
		{ 111, 256, 46 },
		{ 111, 257, 46 },
		{ 111, 258, 46 },
		{ 111, 259, 46 },
		{ 111, 260, 46 },
		{ 111, 261, 46 },
		{ 111, 262, 46 },
		{ 111, 263, 46 },
		{ 111, 264, 46 },
		{ 111, 265, 46 },
		{ 111, 266, 46 },
		{ 111, 267, 46 },
		{ 111, 268, 46 },
		{ 111, 269, 46 },
		{ 111, 270, 46 },
		{ 111, 271, 46 },
		{ 111, 272, 46 },
		{ 111, 273, 46 },
		{ 111, 274, 46 },
		{ 111, 275, 46 },
		{ 112, 239, 46 },
		{ 112, 240, 46 },
		{ 112, 241, 46 },
		{ 112, 242, 46 },
		{ 112, 243, 46 },
		{ 112, 244, 46 },
		{ 112, 245, 46 },
		{ 112, 246, 46 },
		{ 112, 247, 46 },
		{ 112, 248, 46 },
		{ 112, 249, 46 },
		{ 112, 250, 46 },
		{ 112, 251, 46 },
		{ 112, 252, 46 },
		{ 112, 253, 46 },
		{ 112, 254, 46 },
		{ 112, 255, 46 },
		{ 112, 256, 46 },
		{ 112, 257, 46 },
		{ 112, 258, 46 },
		{ 112, 259, 46 },
		{ 112, 260, 46 },
		{ 112, 261, 46 },
		{ 112, 262, 46 },
		{ 112, 263, 46 },
		{ 112, 264, 46 },
		{ 112, 265, 46 },
		{ 112, 266, 46 },
		{ 112, 267, 46 },
		{ 112, 268, 46 },
		{ 112, 269, 46 },
		{ 112, 270, 46 },
		{ 112, 271, 46 },
		{ 112, 272, 46 },
		{ 112, 273, 46 },
		{ 112, 274, 46 },
		{ 112, 275, 46 },
		{ 113, 240, 46 },
		{ 113, 241, 46 },
		{ 113, 242, 46 },
		{ 113, 243, 46 },
		{ 113, 244, 46 },
		{ 113, 245, 46 },
		{ 113, 246, 46 },
		{ 113, 247, 46 },
		{ 113, 248, 46 },
		{ 113, 249, 46 },
		{ 113, 250, 46 },
		{ 113, 251, 46 },
		{ 113, 252, 46 },
		{ 113, 253, 46 },
		{ 113, 254, 46 },
		{ 113, 255, 46 },
		{ 113, 256, 46 },
		{ 113, 257, 46 },
		{ 113, 258, 46 },
		{ 113, 259, 46 },
		{ 113, 260, 46 },
		{ 113, 261, 46 },
		{ 113, 262, 46 },
		{ 113, 263, 46 },
		{ 113, 264, 46 },
		{ 113, 265, 46 },
		{ 113, 266, 46 },
		{ 113, 267, 46 },
		{ 113, 268, 46 },
		{ 113, 269, 46 },
		{ 113, 270, 46 },
		{ 113, 271, 46 },
		{ 113, 272, 46 },
		{ 113, 273, 46 },
		{ 113, 274, 46 },
		{ 114, 240, 46 },
		{ 114, 241, 46 },
		{ 114, 242, 46 },
		{ 114, 243, 46 },
		{ 114, 244, 46 },
		{ 114, 245, 46 },
		{ 114, 246, 46 },
		{ 114, 247, 46 },
		{ 114, 248, 46 },
		{ 114, 249, 46 },
		{ 114, 250, 46 },
		{ 114, 251, 46 },
		{ 114, 252, 46 },
		{ 114, 253, 46 },
		{ 114, 254, 46 },
		{ 114, 255, 46 },
		{ 114, 256, 46 },
		{ 114, 257, 46 },
		{ 114, 258, 46 },
		{ 114, 259, 46 },
		{ 114, 260, 46 },
		{ 114, 261, 46 },
		{ 114, 262, 46 },
		{ 114, 263, 46 },
		{ 114, 264, 46 },
		{ 114, 265, 46 },
		{ 114, 266, 46 },
		{ 114, 267, 46 },
		{ 114, 268, 46 },
		{ 114, 269, 46 },
		{ 114, 270, 46 },
		{ 114, 271, 46 },
		{ 114, 272, 46 },
		{ 114, 273, 46 },
		{ 114, 274, 46 },
		{ 115, 240, 46 },
		{ 115, 241, 46 },
		{ 115, 242, 46 },
		{ 115, 243, 46 },
		{ 115, 244, 46 },
		{ 115, 245, 46 },
		{ 115, 246, 46 },
		{ 115, 247, 46 },
		{ 115, 248, 46 },
		{ 115, 249, 46 },
		{ 115, 250, 46 },
		{ 115, 251, 46 },
		{ 115, 252, 46 },
		{ 115, 253, 46 },
		{ 115, 254, 46 },
		{ 115, 255, 46 },
		{ 115, 256, 46 },
		{ 115, 257, 46 },
		{ 115, 258, 46 },
		{ 115, 259, 46 },
		{ 115, 260, 46 },
		{ 115, 261, 46 },
		{ 115, 262, 46 },
		{ 115, 263, 46 },
		{ 115, 264, 46 },
		{ 115, 265, 46 },
		{ 115, 266, 46 },
		{ 115, 267, 46 },
		{ 115, 268, 46 },
		{ 115, 269, 46 },
		{ 115, 270, 46 },
		{ 115, 271, 46 },
		{ 115, 272, 46 },
		{ 115, 273, 46 },
		{ 115, 274, 46 },
		{ 116, 240, 46 },
		{ 116, 241, 46 },
		{ 116, 242, 46 },
		{ 116, 243, 46 },
		{ 116, 244, 46 },
		{ 116, 245, 46 },
		{ 116, 246, 46 },
		{ 116, 247, 46 },
		{ 116, 248, 46 },
		{ 116, 249, 46 },
		{ 116, 250, 46 },
		{ 116, 251, 46 },
		{ 116, 252, 46 },
		{ 116, 253, 46 },
		{ 116, 254, 46 },
		{ 116, 255, 46 },
		{ 116, 256, 46 },
		{ 116, 257, 46 },
		{ 116, 258, 46 },
		{ 116, 259, 46 },
		{ 116, 260, 46 },
		{ 116, 261, 46 },
		{ 116, 262, 46 },
		{ 116, 263, 46 },
		{ 116, 264, 46 },
		{ 116, 265, 46 },
		{ 116, 266, 46 },
		{ 116, 267, 46 },
		{ 116, 268, 46 },
		{ 116, 269, 46 },
		{ 116, 270, 46 },
		{ 116, 271, 46 },
		{ 116, 272, 46 },
		{ 116, 273, 46 },
		{ 116, 274, 46 },
		{ 117, 240, 46 },
		{ 117, 241, 46 },
		{ 117, 242, 46 },
		{ 117, 243, 46 },
		{ 117, 244, 46 },
		{ 117, 245, 46 },
		{ 117, 246, 46 },
		{ 117, 247, 46 },
		{ 117, 248, 46 },
		{ 117, 249, 46 },
		{ 117, 250, 46 },
		{ 117, 251, 46 },
		{ 117, 252, 46 },
		{ 117, 253, 46 },
		{ 117, 254, 46 },
		{ 117, 255, 46 },
		{ 117, 256, 46 },
		{ 117, 257, 46 },
		{ 117, 258, 46 },
		{ 117, 259, 46 },
		{ 117, 260, 46 },
		{ 117, 261, 46 },
		{ 117, 262, 46 },
		{ 117, 263, 46 },
		{ 117, 264, 46 },
		{ 117, 265, 46 },
		{ 117, 266, 46 },
		{ 117, 267, 46 },
		{ 117, 268, 46 },
		{ 117, 269, 46 },
		{ 117, 270, 46 },
		{ 117, 271, 46 },
		{ 117, 272, 46 },
		{ 117, 273, 46 },
		{ 117, 274, 46 },
		{ 118, 240, 46 },
		{ 118, 241, 46 },
		{ 118, 242, 46 },
		{ 118, 243, 46 },
		{ 118, 244, 46 },
		{ 118, 245, 46 },
		{ 118, 246, 46 },
		{ 118, 247, 46 },
		{ 118, 248, 46 },
		{ 118, 249, 46 },
		{ 118, 250, 46 },
		{ 118, 251, 46 },
		{ 118, 252, 46 },
		{ 118, 253, 46 },
		{ 118, 254, 46 },
		{ 118, 255, 46 },
		{ 118, 256, 46 },
		{ 118, 257, 46 },
		{ 118, 258, 46 },
		{ 118, 259, 46 },
		{ 118, 260, 46 },
		{ 118, 261, 46 },
		{ 118, 262, 46 },
		{ 118, 263, 46 },
		{ 118, 264, 46 },
		{ 118, 265, 46 },
		{ 118, 266, 46 },
		{ 118, 267, 46 },
		{ 118, 268, 46 },
		{ 118, 269, 46 },
		{ 118, 270, 46 },
		{ 118, 271, 46 },
		{ 118, 272, 46 },
		{ 118, 273, 46 },
		{ 118, 274, 46 },
		{ 119, 240, 46 },
		{ 119, 241, 46 },
		{ 119, 242, 46 },
		{ 119, 243, 46 },
		{ 119, 244, 46 },
		{ 119, 245, 46 },
		{ 119, 246, 46 },
		{ 119, 247, 46 },
		{ 119, 248, 46 },
		{ 119, 249, 46 },
		{ 119, 250, 46 },
		{ 119, 251, 46 },
		{ 119, 252, 46 },
		{ 119, 253, 46 },
		{ 119, 254, 46 },
		{ 119, 255, 46 },
		{ 119, 256, 46 },
		{ 119, 257, 46 },
		{ 119, 258, 46 },
		{ 119, 259, 46 },
		{ 119, 260, 46 },
		{ 119, 261, 46 },
		{ 119, 262, 46 },
		{ 119, 263, 46 },
		{ 119, 264, 46 },
		{ 119, 265, 46 },
		{ 119, 266, 46 },
		{ 119, 267, 46 },
		{ 119, 268, 46 },
		{ 119, 269, 46 },
		{ 119, 270, 46 },
		{ 119, 271, 46 },
		{ 119, 272, 46 },
		{ 119, 273, 46 },
		{ 119, 274, 46 },
		{ 120, 240, 46 },
		{ 120, 241, 46 },
		{ 120, 242, 46 },
		{ 120, 243, 46 },
		{ 120, 244, 46 },
		{ 120, 245, 46 },
		{ 120, 246, 46 },
		{ 120, 247, 46 },
		{ 120, 248, 46 },
		{ 120, 249, 46 },
		{ 120, 250, 46 },
		{ 120, 251, 46 },
		{ 120, 252, 46 },
		{ 120, 253, 46 },
		{ 120, 254, 46 },
		{ 120, 255, 46 },
		{ 120, 256, 46 },
		{ 120, 257, 46 },
		{ 120, 258, 46 },
		{ 120, 259, 46 },
		{ 120, 260, 46 },
		{ 120, 261, 46 },
		{ 120, 262, 46 },
		{ 120, 263, 46 },
		{ 120, 264, 46 },
		{ 120, 265, 46 },
		{ 120, 266, 46 },
		{ 120, 267, 46 },
		{ 120, 268, 46 },
		{ 120, 269, 46 },
		{ 120, 270, 46 },
		{ 120, 271, 46 },
		{ 120, 272, 46 },
		{ 120, 273, 46 },
		{ 120, 274, 46 },
		{ 121, 241, 46 },
		{ 121, 242, 46 },
		{ 121, 243, 46 },
		{ 121, 244, 46 },
		{ 121, 245, 46 },
		{ 121, 246, 46 },
		{ 121, 247, 46 },
		{ 121, 248, 46 },
		{ 121, 249, 46 },
		{ 121, 250, 46 },
		{ 121, 251, 46 },
		{ 121, 252, 46 },
		{ 121, 253, 46 },
		{ 121, 254, 46 },
		{ 121, 255, 46 },
		{ 121, 256, 46 },
		{ 121, 257, 46 },
		{ 121, 258, 46 },
		{ 121, 259, 46 },
		{ 121, 260, 46 },
		{ 121, 261, 46 },
		{ 121, 262, 46 },
		{ 121, 263, 46 },
		{ 121, 264, 46 },
		{ 121, 265, 46 },
		{ 121, 266, 46 },
		{ 121, 267, 46 },
		{ 121, 268, 46 },
		{ 121, 269, 46 },
		{ 121, 270, 46 },
		{ 121, 271, 46 },
		{ 121, 272, 46 },
		{ 121, 273, 46 },
		{ 122, 241, 46 },
		{ 122, 242, 46 },
		{ 122, 243, 46 },
		{ 122, 244, 46 },
		{ 122, 245, 46 },
		{ 122, 246, 46 },
		{ 122, 247, 46 },
		{ 122, 248, 46 },
		{ 122, 249, 46 },
		{ 122, 250, 46 },
		{ 122, 251, 46 },
		{ 122, 252, 46 },
		{ 122, 253, 46 },
		{ 122, 254, 46 },
		{ 122, 255, 46 },
		{ 122, 256, 46 },
		{ 122, 257, 46 },
		{ 122, 258, 46 },
		{ 122, 259, 46 },
		{ 122, 260, 46 },
		{ 122, 261, 46 },
		{ 122, 262, 46 },
		{ 122, 263, 46 },
		{ 122, 264, 46 },
		{ 122, 265, 46 },
		{ 122, 266, 46 },
		{ 122, 267, 46 },
		{ 122, 268, 46 },
		{ 122, 269, 46 },
		{ 122, 270, 46 },
		{ 122, 271, 46 },
		{ 122, 272, 46 },
		{ 122, 273, 46 },
		{ 123, 241, 46 },
		{ 123, 242, 46 },
		{ 123, 243, 46 },
		{ 123, 244, 46 },
		{ 123, 245, 46 },
		{ 123, 246, 46 },
		{ 123, 247, 46 },
		{ 123, 248, 46 },
		{ 123, 249, 46 },
		{ 123, 250, 46 },
		{ 123, 251, 46 },
		{ 123, 252, 46 },
		{ 123, 253, 46 },
		{ 123, 254, 46 },
		{ 123, 255, 46 },
		{ 123, 256, 46 },
		{ 123, 257, 46 },
		{ 123, 258, 46 },
		{ 123, 259, 46 },
		{ 123, 260, 46 },
		{ 123, 261, 46 },
		{ 123, 262, 46 },
		{ 123, 263, 46 },
		{ 123, 264, 46 },
		{ 123, 265, 46 },
		{ 123, 266, 46 },
		{ 123, 267, 46 },
		{ 123, 268, 46 },
		{ 123, 269, 46 },
		{ 123, 270, 46 },
		{ 123, 271, 46 },
		{ 123, 272, 46 },
		{ 123, 273, 46 },
		{ 124, 241, 46 },
		{ 124, 242, 46 },
		{ 124, 243, 46 },
		{ 124, 244, 46 },
		{ 124, 245, 46 },
		{ 124, 246, 46 },
		{ 124, 247, 46 },
		{ 124, 248, 46 },
		{ 124, 249, 46 },
		{ 124, 250, 46 },
		{ 124, 251, 46 },
		{ 124, 252, 46 },
		{ 124, 253, 46 },
		{ 124, 254, 46 },
		{ 124, 255, 46 },
		{ 124, 256, 46 },
		{ 124, 257, 46 },
		{ 124, 258, 46 },
		{ 124, 259, 46 },
		{ 124, 260, 46 },
		{ 124, 261, 46 },
		{ 124, 262, 46 },
		{ 124, 263, 46 },
		{ 124, 264, 46 },
		{ 124, 265, 46 },
		{ 124, 266, 46 },
		{ 124, 267, 46 },
		{ 124, 268, 46 },
		{ 124, 269, 46 },
		{ 124, 270, 46 },
		{ 124, 271, 46 },
		{ 124, 272, 46 },
		{ 124, 273, 46 },
		{ 125, 241, 46 },
		{ 125, 242, 46 },
		{ 125, 243, 46 },
		{ 125, 244, 46 },
		{ 125, 245, 46 },
		{ 125, 246, 46 },
		{ 125, 247, 46 },
		{ 125, 248, 46 },
		{ 125, 249, 46 },
		{ 125, 250, 46 },
		{ 125, 251, 46 },
		{ 125, 252, 46 },
		{ 125, 253, 46 },
		{ 125, 254, 46 },
		{ 125, 255, 46 },
		{ 125, 256, 46 },
		{ 125, 257, 46 },
		{ 125, 258, 46 },
		{ 125, 259, 46 },
		{ 125, 260, 46 },
		{ 125, 261, 46 },
		{ 125, 262, 46 },
		{ 125, 263, 46 },
		{ 125, 264, 46 },
		{ 125, 265, 46 },
		{ 125, 266, 46 },
		{ 125, 267, 46 },
		{ 125, 268, 46 },
		{ 125, 269, 46 },
		{ 125, 270, 46 },
		{ 125, 271, 46 },
		{ 125, 272, 46 },
		{ 125, 273, 46 },
		{ 126, 241, 46 },
		{ 126, 242, 46 },
		{ 126, 243, 46 },
		{ 126, 244, 46 },
		{ 126, 245, 46 },
		{ 126, 246, 46 },
		{ 126, 247, 46 },
		{ 126, 248, 46 },
		{ 126, 249, 46 },
		{ 126, 250, 46 },
		{ 126, 251, 46 },
		{ 126, 252, 46 },
		{ 126, 253, 46 },
		{ 126, 254, 46 },
		{ 126, 255, 46 },
		{ 126, 256, 46 },
		{ 126, 257, 46 },
		{ 126, 258, 46 },
		{ 126, 259, 46 },
		{ 126, 260, 46 },
		{ 126, 261, 46 },
		{ 126, 262, 46 },
		{ 126, 263, 46 },
		{ 126, 264, 46 },
		{ 126, 265, 46 },
		{ 126, 266, 46 },
		{ 126, 267, 46 },
		{ 126, 268, 46 },
		{ 126, 269, 46 },
		{ 126, 270, 46 },
		{ 126, 271, 46 },
		{ 126, 272, 46 },
		{ 126, 273, 46 },
		{ 127, 241, 46 },
		{ 127, 242, 46 },
		{ 127, 243, 46 },
		{ 127, 244, 46 },
		{ 127, 245, 46 },
		{ 127, 246, 46 },
		{ 127, 247, 46 },
		{ 127, 248, 46 },
		{ 127, 249, 46 },
		{ 127, 250, 46 },
		{ 127, 251, 46 },
		{ 127, 252, 46 },
		{ 127, 253, 46 },
		{ 127, 254, 46 },
		{ 127, 255, 46 },
		{ 127, 256, 46 },
		{ 127, 257, 46 },
		{ 127, 258, 46 },
		{ 127, 259, 46 },
		{ 127, 260, 46 },
		{ 127, 261, 46 },
		{ 127, 262, 46 },
		{ 127, 263, 46 },
		{ 127, 264, 46 },
		{ 127, 265, 46 },
		{ 127, 266, 46 },
		{ 127, 267, 46 },
		{ 127, 268, 46 },
		{ 127, 269, 46 },
		{ 127, 270, 46 },
		{ 127, 271, 46 },
		{ 127, 272, 46 },
		{ 127, 273, 46 },
		{ 128, 241, 46 },
		{ 128, 242, 46 },
		{ 128, 243, 46 },
		{ 128, 244, 46 },
		{ 128, 245, 46 },
		{ 128, 246, 46 },
		{ 128, 247, 46 },
		{ 128, 248, 46 },
		{ 128, 249, 46 },
		{ 128, 250, 46 },
		{ 128, 251, 46 },
		{ 128, 252, 46 },
		{ 128, 253, 46 },
		{ 128, 254, 46 },
		{ 128, 255, 46 },
		{ 128, 256, 46 },
		{ 128, 257, 46 },
		{ 128, 258, 46 },
		{ 128, 259, 46 },
		{ 128, 260, 46 },
		{ 128, 261, 46 },
		{ 128, 262, 46 },
		{ 128, 263, 46 },
		{ 128, 264, 46 },
		{ 128, 265, 46 },
		{ 128, 266, 46 },
		{ 128, 267, 46 },
		{ 128, 268, 46 },
		{ 128, 269, 46 },
		{ 128, 270, 46 },
		{ 128, 271, 46 },
		{ 128, 272, 46 },
		{ 128, 273, 46 },
		{ 129, 242, 46 },
		{ 129, 243, 46 },
		{ 129, 244, 46 },
		{ 129, 245, 46 },
		{ 129, 246, 46 },
		{ 129, 247, 46 },
		{ 129, 248, 46 },
		{ 129, 249, 46 },
		{ 129, 250, 46 },
		{ 129, 251, 46 },
		{ 129, 252, 46 },
		{ 129, 253, 46 },
		{ 129, 254, 46 },
		{ 129, 255, 46 },
		{ 129, 256, 46 },
		{ 129, 257, 46 },
		{ 129, 258, 46 },
		{ 129, 259, 46 },
		{ 129, 260, 46 },
		{ 129, 261, 46 },
		{ 129, 262, 46 },
		{ 129, 263, 46 },
		{ 129, 264, 46 },
		{ 129, 265, 46 },
		{ 129, 266, 46 },
		{ 129, 267, 46 },
		{ 129, 268, 46 },
		{ 129, 269, 46 },
		{ 129, 270, 46 },
		{ 129, 271, 46 },
		{ 129, 272, 46 },
		{ 130, 242, 46 },
		{ 130, 243, 46 },
		{ 130, 244, 46 },
		{ 130, 245, 46 },
		{ 130, 246, 46 },
		{ 130, 247, 46 },
		{ 130, 248, 46 },
		{ 130, 249, 46 },
		{ 130, 250, 46 },
		{ 130, 251, 46 },
		{ 130, 252, 46 },
		{ 130, 253, 46 },
		{ 130, 254, 46 },
		{ 130, 255, 46 },
		{ 130, 256, 46 },
		{ 130, 257, 46 },
		{ 130, 258, 46 },
		{ 130, 259, 46 },
		{ 130, 260, 46 },
		{ 130, 261, 46 },
		{ 130, 262, 46 },
		{ 130, 263, 46 },
		{ 130, 264, 46 },
		{ 130, 265, 46 },
		{ 130, 266, 46 },
		{ 130, 267, 46 },
		{ 130, 268, 46 },
		{ 130, 269, 46 },
		{ 130, 270, 46 },
		{ 130, 271, 46 },
		{ 130, 272, 46 },
		{ 131, 242, 46 },
		{ 131, 243, 46 },
		{ 131, 244, 46 },
		{ 131, 245, 46 },
		{ 131, 246, 46 },
		{ 131, 247, 46 },
		{ 131, 248, 46 },
		{ 131, 249, 46 },
		{ 131, 250, 46 },
		{ 131, 251, 46 },
		{ 131, 252, 46 },
		{ 131, 253, 46 },
		{ 131, 254, 46 },
		{ 131, 255, 46 },
		{ 131, 256, 46 },
		{ 131, 257, 46 },
		{ 131, 258, 46 },
		{ 131, 259, 46 },
		{ 131, 260, 46 },
		{ 131, 261, 46 },
		{ 131, 262, 46 },
		{ 131, 263, 46 },
		{ 131, 264, 46 },
		{ 131, 265, 46 },
		{ 131, 266, 46 },
		{ 131, 267, 46 },
		{ 131, 268, 46 },
		{ 131, 269, 46 },
		{ 131, 270, 46 },
		{ 131, 271, 46 },
		{ 131, 272, 46 },
		{ 132, 242, 46 },
		{ 132, 243, 46 },
		{ 132, 244, 46 },
		{ 132, 245, 46 },
		{ 132, 246, 46 },
		{ 132, 247, 46 },
		{ 132, 248, 46 },
		{ 132, 249, 46 },
		{ 132, 250, 46 },
		{ 132, 251, 46 },
		{ 132, 252, 46 },
		{ 132, 253, 46 },
		{ 132, 254, 46 },
		{ 132, 255, 46 },
		{ 132, 256, 46 },
		{ 132, 257, 46 },
		{ 132, 258, 46 },
		{ 132, 259, 46 },
		{ 132, 260, 46 },
		{ 132, 261, 46 },
		{ 132, 262, 46 },
		{ 132, 263, 46 },
		{ 132, 264, 46 },
		{ 132, 265, 46 },
		{ 132, 266, 46 },
		{ 132, 267, 46 },
		{ 132, 268, 46 },
		{ 132, 269, 46 },
		{ 132, 270, 46 },
		{ 132, 271, 46 },
		{ 132, 272, 46 },
		{ 133, 242, 46 },
		{ 133, 243, 46 },
		{ 133, 244, 46 },
		{ 133, 245, 46 },
		{ 133, 246, 46 },
		{ 133, 247, 46 },
		{ 133, 248, 46 },
		{ 133, 249, 46 },
		{ 133, 250, 46 },
		{ 133, 251, 46 },
		{ 133, 252, 46 },
		{ 133, 253, 46 },
		{ 133, 254, 46 },
		{ 133, 255, 46 },
		{ 133, 256, 46 },
		{ 133, 257, 46 },
		{ 133, 258, 46 },
		{ 133, 259, 46 },
		{ 133, 260, 46 },
		{ 133, 261, 46 },
		{ 133, 262, 46 },
		{ 133, 263, 46 },
		{ 133, 264, 46 },
		{ 133, 265, 46 },
		{ 133, 266, 46 },
		{ 133, 267, 46 },
		{ 133, 268, 46 },
		{ 133, 269, 46 },
		{ 133, 270, 46 },
		{ 133, 271, 46 },
		{ 133, 272, 46 },
		{ 134, 242, 46 },
		{ 134, 243, 46 },
		{ 134, 244, 46 },
		{ 134, 245, 46 },
		{ 134, 246, 46 },
		{ 134, 247, 46 },
		{ 134, 248, 46 },
		{ 134, 249, 46 },
		{ 134, 250, 46 },
		{ 134, 251, 46 },
		{ 134, 252, 46 },
		{ 134, 253, 46 },
		{ 134, 254, 46 },
		{ 134, 255, 46 },
		{ 134, 256, 46 },
		{ 134, 257, 46 },
		{ 134, 258, 46 },
		{ 134, 259, 46 },
		{ 134, 260, 46 },
		{ 134, 261, 46 },
		{ 134, 262, 46 },
		{ 134, 263, 46 },
		{ 134, 264, 46 },
		{ 134, 265, 46 },
		{ 134, 266, 46 },
		{ 134, 267, 46 },
		{ 134, 268, 46 },
		{ 134, 269, 46 },
		{ 134, 270, 46 },
		{ 134, 271, 46 },
		{ 134, 272, 46 },
		{ 135, 242, 46 },
		{ 135, 243, 46 },
		{ 135, 244, 46 },
		{ 135, 245, 46 },
		{ 135, 246, 46 },
		{ 135, 247, 46 },
		{ 135, 248, 46 },
		{ 135, 249, 46 },
		{ 135, 250, 46 },
		{ 135, 251, 46 },
		{ 135, 252, 46 },
		{ 135, 253, 46 },
		{ 135, 254, 46 },
		{ 135, 255, 46 },
		{ 135, 256, 46 },
		{ 135, 257, 46 },
		{ 135, 258, 46 },
		{ 135, 259, 46 },
		{ 135, 260, 46 },
		{ 135, 261, 46 },
		{ 135, 262, 46 },
		{ 135, 263, 46 },
		{ 135, 264, 46 },
		{ 135, 265, 46 },
		{ 135, 266, 46 },
		{ 135, 267, 46 },
		{ 135, 268, 46 },
		{ 135, 269, 46 },
		{ 135, 270, 46 },
		{ 135, 271, 46 },
		{ 135, 272, 46 },
		{ 136, 242, 46 },
		{ 136, 243, 46 },
		{ 136, 244, 46 },
		{ 136, 245, 46 },
		{ 136, 246, 46 },
		{ 136, 247, 46 },
		{ 136, 248, 46 },
		{ 136, 249, 46 },
		{ 136, 250, 46 },
		{ 136, 251, 46 },
		{ 136, 252, 46 },
		{ 136, 253, 46 },
		{ 136, 254, 46 },
		{ 136, 255, 46 },
		{ 136, 256, 46 },
		{ 136, 257, 46 },
		{ 136, 258, 46 },
		{ 136, 259, 46 },
		{ 136, 260, 46 },
		{ 136, 261, 46 },
		{ 136, 262, 46 },
		{ 136, 263, 46 },
		{ 136, 264, 46 },
		{ 136, 265, 46 },
		{ 136, 266, 46 },
		{ 136, 267, 46 },
		{ 136, 268, 46 },
		{ 136, 269, 46 },
		{ 136, 270, 46 },
		{ 136, 271, 46 },
		{ 136, 272, 46 },
		{ 137, 243, 46 },
		{ 137, 244, 46 },
		{ 137, 245, 46 },
		{ 137, 246, 46 },
		{ 137, 247, 46 },
		{ 137, 248, 46 },
		{ 137, 249, 46 },
		{ 137, 250, 46 },
		{ 137, 251, 46 },
		{ 137, 252, 46 },
		{ 137, 253, 46 },
		{ 137, 254, 46 },
		{ 137, 255, 46 },
		{ 137, 256, 46 },
		{ 137, 257, 46 },
		{ 137, 258, 46 },
		{ 137, 259, 46 },
		{ 137, 260, 46 },
		{ 137, 261, 46 },
		{ 137, 262, 46 },
		{ 137, 263, 46 },
		{ 137, 264, 46 },
		{ 137, 265, 46 },
		{ 137, 266, 46 },
		{ 137, 267, 46 },
		{ 137, 268, 46 },
		{ 137, 269, 46 },
		{ 137, 270, 46 },
		{ 137, 271, 46 },
		{ 138, 243, 46 },
		{ 138, 244, 46 },
		{ 138, 245, 46 },
		{ 138, 246, 46 },
		{ 138, 247, 46 },
		{ 138, 248, 46 },
		{ 138, 249, 46 },
		{ 138, 250, 46 },
		{ 138, 251, 46 },
		{ 138, 252, 46 },
		{ 138, 253, 46 },
		{ 138, 254, 46 },
		{ 138, 255, 46 },
		{ 138, 256, 46 },
		{ 138, 257, 46 },
		{ 138, 258, 46 },
		{ 138, 259, 46 },
		{ 138, 260, 46 },
		{ 138, 261, 46 },
		{ 138, 262, 46 },
		{ 138, 263, 46 },
		{ 138, 264, 46 },
		{ 138, 265, 46 },
		{ 138, 266, 46 },
		{ 138, 267, 46 },
		{ 138, 268, 46 },
		{ 138, 269, 46 },
		{ 138, 270, 46 },
		{ 138, 271, 46 },
		{ 139, 243, 46 },
		{ 139, 244, 46 },
		{ 139, 245, 46 },
		{ 139, 246, 46 },
		{ 139, 247, 46 },
		{ 139, 248, 46 },
		{ 139, 249, 46 },
		{ 139, 250, 46 },
		{ 139, 251, 46 },
		{ 139, 252, 46 },
		{ 139, 253, 46 },
		{ 139, 254, 46 },
		{ 139, 255, 46 },
		{ 139, 256, 46 },
		{ 139, 257, 46 },
		{ 139, 258, 46 },
		{ 139, 259, 46 },
		{ 139, 260, 46 },
		{ 139, 261, 46 },
		{ 139, 262, 46 },
		{ 139, 263, 46 },
		{ 139, 264, 46 },
		{ 139, 265, 46 },
		{ 139, 266, 46 },
		{ 139, 267, 46 },
		{ 139, 268, 46 },
		{ 139, 269, 46 },
		{ 139, 270, 46 },
		{ 139, 271, 46 },
		{ 140, 243, 46 },
		{ 140, 244, 46 },
		{ 140, 245, 46 },
		{ 140, 246, 46 },
		{ 140, 247, 46 },
		{ 140, 248, 46 },
		{ 140, 249, 46 },
		{ 140, 250, 46 },
		{ 140, 251, 46 },
		{ 140, 252, 46 },
		{ 140, 253, 46 },
		{ 140, 254, 46 },
		{ 140, 255, 46 },
		{ 140, 256, 46 },
		{ 140, 257, 46 },
		{ 140, 258, 46 },
		{ 140, 259, 46 },
		{ 140, 260, 46 },
		{ 140, 261, 46 },
		{ 140, 262, 46 },
		{ 140, 263, 46 },
		{ 140, 264, 46 },
		{ 140, 265, 46 },
		{ 140, 266, 46 },
		{ 140, 267, 46 },
		{ 140, 268, 46 },
		{ 140, 269, 46 },
		{ 140, 270, 46 },
		{ 140, 271, 46 },
		{ 141, 243, 46 },
		{ 141, 244, 46 },
		{ 141, 245, 46 },
		{ 141, 246, 46 },
		{ 141, 247, 46 },
		{ 141, 248, 46 },
		{ 141, 249, 46 },
		{ 141, 250, 46 },
		{ 141, 251, 46 },
		{ 141, 252, 46 },
		{ 141, 253, 46 },
		{ 141, 254, 46 },
		{ 141, 255, 46 },
		{ 141, 256, 46 },
		{ 141, 257, 46 },
		{ 141, 258, 46 },
		{ 141, 259, 46 },
		{ 141, 260, 46 },
		{ 141, 261, 46 },
		{ 141, 262, 46 },
		{ 141, 263, 46 },
		{ 141, 264, 46 },
		{ 141, 265, 46 },
		{ 141, 266, 46 },
		{ 141, 267, 46 },
		{ 141, 268, 46 },
		{ 141, 269, 46 },
		{ 141, 270, 46 },
		{ 141, 271, 46 },
		{ 142, 243, 46 },
		{ 142, 244, 46 },
		{ 142, 245, 46 },
		{ 142, 246, 46 },
		{ 142, 247, 46 },
		{ 142, 248, 46 },
		{ 142, 249, 46 },
		{ 142, 250, 46 },
		{ 142, 251, 46 },
		{ 142, 252, 46 },
		{ 142, 253, 46 },
		{ 142, 254, 46 },
		{ 142, 255, 46 },
		{ 142, 256, 46 },
		{ 142, 257, 46 },
		{ 142, 258, 46 },
		{ 142, 259, 46 },
		{ 142, 260, 46 },
		{ 142, 261, 46 },
		{ 142, 262, 46 },
		{ 142, 263, 46 },
		{ 142, 264, 46 },
		{ 142, 265, 46 },
		{ 142, 266, 46 },
		{ 142, 267, 46 },
		{ 142, 268, 46 },
		{ 142, 269, 46 },
		{ 142, 270, 46 },
		{ 142, 271, 46 },
		{ 143, 243, 46 },
		{ 143, 244, 46 },
		{ 143, 245, 46 },
		{ 143, 246, 46 },
		{ 143, 247, 46 },
		{ 143, 248, 46 },
		{ 143, 249, 46 },
		{ 143, 250, 46 },
		{ 143, 251, 46 },
		{ 143, 252, 46 },
		{ 143, 253, 46 },
		{ 143, 254, 46 },
		{ 143, 255, 46 },
		{ 143, 256, 46 },
		{ 143, 257, 46 },
		{ 143, 258, 46 },
		{ 143, 259, 46 },
		{ 143, 260, 46 },
		{ 143, 261, 46 },
		{ 143, 262, 46 },
		{ 143, 263, 46 },
		{ 143, 264, 46 },
		{ 143, 265, 46 },
		{ 143, 266, 46 },
		{ 143, 267, 46 },
		{ 143, 268, 46 },
		{ 143, 269, 46 },
		{ 143, 270, 46 },
		{ 143, 271, 46 },
		{ 144, 243, 46 },
		{ 144, 244, 46 },
		{ 144, 245, 46 },
		{ 144, 246, 46 },
		{ 144, 247, 46 },
		{ 144, 248, 46 },
		{ 144, 249, 46 },
		{ 144, 250, 46 },
		{ 144, 251, 46 },
		{ 144, 252, 46 },
		{ 144, 253, 46 },
		{ 144, 254, 46 },
		{ 144, 255, 46 },
		{ 144, 256, 46 },
		{ 144, 257, 46 },
		{ 144, 258, 46 },
		{ 144, 259, 46 },
		{ 144, 260, 46 },
		{ 144, 261, 46 },
		{ 144, 262, 46 },
		{ 144, 263, 46 },
		{ 144, 264, 46 },
		{ 144, 265, 46 },
		{ 144, 266, 46 },
		{ 144, 267, 46 },
		{ 144, 268, 46 },
		{ 144, 269, 46 },
		{ 144, 270, 46 },
		{ 144, 271, 46 },
		{ 145, 244, 46 },
		{ 145, 245, 46 },
		{ 145, 246, 46 },
		{ 145, 247, 46 },
		{ 145, 248, 46 },
		{ 145, 249, 46 },
		{ 145, 250, 46 },
		{ 145, 251, 46 },
		{ 145, 252, 46 },
		{ 145, 253, 46 },
		{ 145, 254, 46 },
		{ 145, 255, 46 },
		{ 145, 256, 46 },
		{ 145, 257, 46 },
		{ 145, 258, 46 },
		{ 145, 259, 46 },
		{ 145, 260, 46 },
		{ 145, 261, 46 },
		{ 145, 262, 46 },
		{ 145, 263, 46 },
		{ 145, 264, 46 },
		{ 145, 265, 46 },
		{ 145, 266, 46 },
		{ 145, 267, 46 },
		{ 145, 268, 46 },
		{ 145, 269, 46 },
		{ 145, 270, 46 },
		{ 146, 244, 46 },
		{ 146, 245, 46 },
		{ 146, 246, 46 },
		{ 146, 247, 46 },
		{ 146, 248, 46 },
		{ 146, 249, 46 },
		{ 146, 250, 46 },
		{ 146, 251, 46 },
		{ 146, 252, 46 },
		{ 146, 253, 46 },
		{ 146, 254, 46 },
		{ 146, 255, 46 },
		{ 146, 256, 46 },
		{ 146, 257, 46 },
		{ 146, 258, 46 },
		{ 146, 259, 46 },
		{ 146, 260, 46 },
		{ 146, 261, 46 },
		{ 146, 262, 46 },
		{ 146, 263, 46 },
		{ 146, 264, 46 },
		{ 146, 265, 46 },
		{ 146, 266, 46 },
		{ 146, 267, 46 },
		{ 146, 268, 46 },
		{ 146, 269, 46 },
		{ 146, 270, 46 },
		{ 147, 244, 46 },
		{ 147, 245, 46 },
		{ 147, 246, 46 },
		{ 147, 247, 46 },
		{ 147, 248, 46 },
		{ 147, 249, 46 },
		{ 147, 250, 46 },
		{ 147, 251, 46 },
		{ 147, 252, 46 },
		{ 147, 253, 46 },
		{ 147, 254, 46 },
		{ 147, 255, 46 },
		{ 147, 256, 46 },
		{ 147, 257, 46 },
		{ 147, 258, 46 },
		{ 147, 259, 46 },
		{ 147, 260, 46 },
		{ 147, 261, 46 },
		{ 147, 262, 46 },
		{ 147, 263, 46 },
		{ 147, 264, 46 },
		{ 147, 265, 46 },
		{ 147, 266, 46 },
		{ 147, 267, 46 },
		{ 147, 268, 46 },
		{ 147, 269, 46 },
		{ 147, 270, 46 },
		{ 148, 244, 46 },
		{ 148, 245, 46 },
		{ 148, 246, 46 },
		{ 148, 247, 46 },
		{ 148, 248, 46 },
		{ 148, 249, 46 },
		{ 148, 250, 46 },
		{ 148, 251, 46 },
		{ 148, 252, 46 },
		{ 148, 253, 46 },
		{ 148, 254, 46 },
		{ 148, 255, 46 },
		{ 148, 256, 46 },
		{ 148, 257, 46 },
		{ 148, 258, 46 },
		{ 148, 259, 46 },
		{ 148, 260, 46 },
		{ 148, 261, 46 },
		{ 148, 262, 46 },
		{ 148, 263, 46 },
		{ 148, 264, 46 },
		{ 148, 265, 46 },
		{ 148, 266, 46 },
		{ 148, 267, 46 },
		{ 148, 268, 46 },
		{ 148, 269, 46 },
		{ 148, 270, 46 },
		{ 149, 244, 46 },
		{ 149, 245, 46 },
		{ 149, 246, 46 },
		{ 149, 247, 46 },
		{ 149, 248, 46 },
		{ 149, 249, 46 },
		{ 149, 250, 46 },
		{ 149, 251, 46 },
		{ 149, 252, 46 },
		{ 149, 253, 46 },
		{ 149, 254, 46 },
		{ 149, 255, 46 },
		{ 149, 256, 46 },
		{ 149, 257, 46 },
		{ 149, 258, 46 },
		{ 149, 259, 46 },
		{ 149, 260, 46 },
		{ 149, 261, 46 },
		{ 149, 262, 46 },
		{ 149, 263, 46 },
		{ 149, 264, 46 },
		{ 149, 265, 46 },
		{ 149, 266, 46 },
		{ 149, 267, 46 },
		{ 149, 268, 46 },
		{ 149, 269, 46 },
		{ 149, 270, 46 },
		{ 150, 244, 46 },
		{ 150, 245, 46 },
		{ 150, 246, 46 },
		{ 150, 247, 46 },
		{ 150, 248, 46 },
		{ 150, 249, 46 },
		{ 150, 250, 46 },
		{ 150, 251, 46 },
		{ 150, 252, 46 },
		{ 150, 253, 46 },
		{ 150, 254, 46 },
		{ 150, 255, 46 },
		{ 150, 256, 46 },
		{ 150, 257, 46 },
		{ 150, 258, 46 },
		{ 150, 259, 46 },
		{ 150, 260, 46 },
		{ 150, 261, 46 },
		{ 150, 262, 46 },
		{ 150, 263, 46 },
		{ 150, 264, 46 },
		{ 150, 265, 46 },
		{ 150, 266, 46 },
		{ 150, 267, 46 },
		{ 150, 268, 46 },
		{ 150, 269, 46 },
		{ 150, 270, 46 },
		{ 151, 244, 46 },
		{ 151, 245, 46 },
		{ 151, 246, 46 },
		{ 151, 247, 46 },
		{ 151, 248, 46 },
		{ 151, 249, 46 },
		{ 151, 250, 46 },
		{ 151, 251, 46 },
		{ 151, 252, 46 },
		{ 151, 253, 46 },
		{ 151, 254, 46 },
		{ 151, 255, 46 },
		{ 151, 256, 46 },
		{ 151, 257, 46 },
		{ 151, 258, 46 },
		{ 151, 259, 46 },
		{ 151, 260, 46 },
		{ 151, 261, 46 },
		{ 151, 262, 46 },
		{ 151, 263, 46 },
		{ 151, 264, 46 },
		{ 151, 265, 46 },
		{ 151, 266, 46 },
		{ 151, 267, 46 },
		{ 151, 268, 46 },
		{ 151, 269, 46 },
		{ 151, 270, 46 },
		{ 152, 244, 46 },
		{ 152, 245, 46 },
		{ 152, 246, 46 },
		{ 152, 247, 46 },
		{ 152, 248, 46 },
		{ 152, 249, 46 },
		{ 152, 250, 46 },
		{ 152, 251, 46 },
		{ 152, 252, 46 },
		{ 152, 253, 46 },
		{ 152, 254, 46 },
		{ 152, 255, 46 },
		{ 152, 256, 46 },
		{ 152, 257, 46 },
		{ 152, 258, 46 },
		{ 152, 259, 46 },
		{ 152, 260, 46 },
		{ 152, 261, 46 },
		{ 152, 262, 46 },
		{ 152, 263, 46 },
		{ 152, 264, 46 },
		{ 152, 265, 46 },
		{ 152, 266, 46 },
		{ 152, 267, 46 },
		{ 152, 268, 46 },
		{ 152, 269, 46 },
		{ 152, 270, 46 },
		{ 153, 245, 46 },
		{ 153, 246, 46 },
		{ 153, 247, 46 },
		{ 153, 248, 46 },
		{ 153, 249, 46 },
		{ 153, 250, 46 },
		{ 153, 251, 46 },
		{ 153, 252, 46 },
		{ 153, 253, 46 },
		{ 153, 254, 46 },
		{ 153, 255, 46 },
		{ 153, 256, 46 },
		{ 153, 257, 46 },
		{ 153, 258, 46 },
		{ 153, 259, 46 },
		{ 153, 260, 46 },
		{ 153, 261, 46 },
		{ 153, 262, 46 },
		{ 153, 263, 46 },
		{ 153, 264, 46 },
		{ 153, 265, 46 },
		{ 153, 266, 46 },
		{ 153, 267, 46 },
		{ 153, 268, 46 },
		{ 153, 269, 46 },
		{ 154, 245, 46 },
		{ 154, 246, 46 },
		{ 154, 247, 46 },
		{ 154, 248, 46 },
		{ 154, 249, 46 },
		{ 154, 250, 46 },
		{ 154, 251, 46 },
		{ 154, 252, 46 },
		{ 154, 253, 46 },
		{ 154, 254, 46 },
		{ 154, 255, 46 },
		{ 154, 256, 46 },
		{ 154, 257, 46 },
		{ 154, 258, 46 },
		{ 154, 259, 46 },
		{ 154, 260, 46 },
		{ 154, 261, 46 },
		{ 154, 262, 46 },
		{ 154, 263, 46 },
		{ 154, 264, 46 },
		{ 154, 265, 46 },
		{ 154, 266, 46 },
		{ 154, 267, 46 },
		{ 154, 268, 46 },
		{ 154, 269, 46 },
		{ 155, 245, 46 },
		{ 155, 246, 46 },
		{ 155, 247, 46 },
		{ 155, 248, 46 },
		{ 155, 249, 46 },
		{ 155, 250, 46 },
		{ 155, 251, 46 },
		{ 155, 252, 46 },
		{ 155, 253, 46 },
		{ 155, 254, 46 },
		{ 155, 255, 46 },
		{ 155, 256, 46 },
		{ 155, 257, 46 },
		{ 155, 258, 46 },
		{ 155, 259, 46 },
		{ 155, 260, 46 },
		{ 155, 261, 46 },
		{ 155, 262, 46 },
		{ 155, 263, 46 },
		{ 155, 264, 46 },
		{ 155, 265, 46 },
		{ 155, 266, 46 },
		{ 155, 267, 46 },
		{ 155, 268, 46 },
		{ 155, 269, 46 },
		{ 156, 245, 46 },
		{ 156, 246, 46 },
		{ 156, 247, 46 },
		{ 156, 248, 46 },
		{ 156, 249, 46 },
		{ 156, 250, 46 },
		{ 156, 251, 46 },
		{ 156, 252, 46 },
		{ 156, 253, 46 },
		{ 156, 254, 46 },
		{ 156, 255, 46 },
		{ 156, 256, 46 },
		{ 156, 257, 46 },
		{ 156, 258, 46 },
		{ 156, 259, 46 },
		{ 156, 260, 46 },
		{ 156, 261, 46 },
		{ 156, 262, 46 },
		{ 156, 263, 46 },
		{ 156, 264, 46 },
		{ 156, 265, 46 },
		{ 156, 266, 46 },
		{ 156, 267, 46 },
		{ 156, 268, 46 },
		{ 156, 269, 46 },
		{ 157, 245, 46 },
		{ 157, 246, 46 },
		{ 157, 247, 46 },
		{ 157, 248, 46 },
		{ 157, 249, 46 },
		{ 157, 250, 46 },
		{ 157, 251, 46 },
		{ 157, 252, 46 },
		{ 157, 253, 46 },
		{ 157, 254, 46 },
		{ 157, 255, 46 },
		{ 157, 256, 46 },
		{ 157, 257, 46 },
		{ 157, 258, 46 },
		{ 157, 259, 46 },
		{ 157, 260, 46 },
		{ 157, 261, 46 },
		{ 157, 262, 46 },
		{ 157, 263, 46 },
		{ 157, 264, 46 },
		{ 157, 265, 46 },
		{ 157, 266, 46 },
		{ 157, 267, 46 },
		{ 157, 268, 46 },
		{ 157, 269, 46 },
		{ 158, 245, 46 },
		{ 158, 246, 46 },
		{ 158, 247, 46 },
		{ 158, 248, 46 },
		{ 158, 249, 46 },
		{ 158, 250, 46 },
		{ 158, 251, 46 },
		{ 158, 252, 46 },
		{ 158, 253, 46 },
		{ 158, 254, 46 },
		{ 158, 255, 46 },
		{ 158, 256, 46 },
		{ 158, 257, 46 },
		{ 158, 258, 46 },
		{ 158, 259, 46 },
		{ 158, 260, 46 },
		{ 158, 261, 46 },
		{ 158, 262, 46 },
		{ 158, 263, 46 },
		{ 158, 264, 46 },
		{ 158, 265, 46 },
		{ 158, 266, 46 },
		{ 158, 267, 46 },
		{ 158, 268, 46 },
		{ 158, 269, 46 },
		{ 159, 245, 46 },
		{ 159, 246, 46 },
		{ 159, 247, 46 },
		{ 159, 248, 46 },
		{ 159, 249, 46 },
		{ 159, 250, 46 },
		{ 159, 251, 46 },
		{ 159, 252, 46 },
		{ 159, 253, 46 },
		{ 159, 254, 46 },
		{ 159, 255, 46 },
		{ 159, 256, 46 },
		{ 159, 257, 46 },
		{ 159, 258, 46 },
		{ 159, 259, 46 },
		{ 159, 260, 46 },
		{ 159, 261, 46 },
		{ 159, 262, 46 },
		{ 159, 263, 46 },
		{ 159, 264, 46 },
		{ 159, 265, 46 },
		{ 159, 266, 46 },
		{ 159, 267, 46 },
		{ 159, 268, 46 },
		{ 159, 269, 46 },
		{ 160, 245, 46 },
		{ 160, 246, 46 },
		{ 160, 247, 46 },
		{ 160, 248, 46 },
		{ 160, 249, 46 },
		{ 160, 250, 46 },
		{ 160, 251, 46 },
		{ 160, 252, 46 },
		{ 160, 253, 46 },
		{ 160, 254, 46 },
		{ 160, 255, 46 },
		{ 160, 256, 46 },
		{ 160, 257, 46 },
		{ 160, 258, 46 },
		{ 160, 259, 46 },
		{ 160, 260, 46 },
		{ 160, 261, 46 },
		{ 160, 262, 46 },
		{ 160, 263, 46 },
		{ 160, 264, 46 },
		{ 160, 265, 46 },
		{ 160, 266, 46 },
		{ 160, 267, 46 },
		{ 160, 268, 46 },
		{ 160, 269, 46 },
		{ 161, 246, 46 },
		{ 161, 247, 46 },
		{ 161, 248, 46 },
		{ 161, 249, 46 },
		{ 161, 250, 46 },
		{ 161, 251, 46 },
		{ 161, 252, 46 },
		{ 161, 253, 46 },
		{ 161, 254, 46 },
		{ 161, 255, 46 },
		{ 161, 256, 46 },
		{ 161, 257, 46 },
		{ 161, 258, 46 },
		{ 161, 259, 46 },
		{ 161, 260, 46 },
		{ 161, 261, 46 },
		{ 161, 262, 46 },
		{ 161, 263, 46 },
		{ 161, 264, 46 },
		{ 161, 265, 46 },
		{ 161, 266, 46 },
		{ 161, 267, 46 },
		{ 161, 268, 46 },
		{ 162, 246, 46 },
		{ 162, 247, 46 },
		{ 162, 248, 46 },
		{ 162, 249, 46 },
		{ 162, 250, 46 },
		{ 162, 251, 46 },
		{ 162, 252, 46 },
		{ 162, 253, 46 },
		{ 162, 254, 46 },
		{ 162, 255, 46 },
		{ 162, 256, 46 },
		{ 162, 257, 46 },
		{ 162, 258, 46 },
		{ 162, 259, 46 },
		{ 162, 260, 46 },
		{ 162, 261, 46 },
		{ 162, 262, 46 },
		{ 162, 263, 46 },
		{ 162, 264, 46 },
		{ 162, 265, 46 },
		{ 162, 266, 46 },
		{ 162, 267, 46 },
		{ 162, 268, 46 },
		{ 163, 246, 46 },
		{ 163, 247, 46 },
		{ 163, 248, 46 },
		{ 163, 249, 46 },
		{ 163, 250, 46 },
		{ 163, 251, 46 },
		{ 163, 252, 46 },
		{ 163, 253, 46 },
		{ 163, 254, 46 },
		{ 163, 255, 46 },
		{ 163, 256, 46 },
		{ 163, 257, 46 },
		{ 163, 258, 46 },
		{ 163, 259, 46 },
		{ 163, 260, 46 },
		{ 163, 261, 46 },
		{ 163, 262, 46 },
		{ 163, 263, 46 },
		{ 163, 264, 46 },
		{ 163, 265, 46 },
		{ 163, 266, 46 },
		{ 163, 267, 46 },
		{ 163, 268, 46 },
		{ 164, 246, 46 },
		{ 164, 247, 46 },
		{ 164, 248, 46 },
		{ 164, 249, 46 },
		{ 164, 250, 46 },
		{ 164, 251, 46 },
		{ 164, 252, 46 },
		{ 164, 253, 46 },
		{ 164, 254, 46 },
		{ 164, 255, 46 },
		{ 164, 256, 46 },
		{ 164, 257, 46 },
		{ 164, 258, 46 },
		{ 164, 259, 46 },
		{ 164, 260, 46 },
		{ 164, 261, 46 },
		{ 164, 262, 46 },
		{ 164, 263, 46 },
		{ 164, 264, 46 },
		{ 164, 265, 46 },
		{ 164, 266, 46 },
		{ 164, 267, 46 },
		{ 164, 268, 46 },
		{ 165, 246, 46 },
		{ 165, 247, 46 },
		{ 165, 248, 46 },
		{ 165, 249, 46 },
		{ 165, 250, 46 },
		{ 165, 251, 46 },
		{ 165, 252, 46 },
		{ 165, 253, 46 },
		{ 165, 254, 46 },
		{ 165, 255, 46 },
		{ 165, 256, 46 },
		{ 165, 257, 46 },
		{ 165, 258, 46 },
		{ 165, 259, 46 },
		{ 165, 260, 46 },
		{ 165, 261, 46 },
		{ 165, 262, 46 },
		{ 165, 263, 46 },
		{ 165, 264, 46 },
		{ 165, 265, 46 },
		{ 165, 266, 46 },
		{ 165, 267, 46 },
		{ 165, 268, 46 },
		{ 166, 246, 46 },
		{ 166, 247, 46 },
		{ 166, 248, 46 },
		{ 166, 249, 46 },
		{ 166, 250, 46 },
		{ 166, 251, 46 },
		{ 166, 252, 46 },
		{ 166, 253, 46 },
		{ 166, 254, 46 },
		{ 166, 255, 46 },
		{ 166, 256, 46 },
		{ 166, 257, 46 },
		{ 166, 258, 46 },
		{ 166, 259, 46 },
		{ 166, 260, 46 },
		{ 166, 261, 46 },
		{ 166, 262, 46 },
		{ 166, 263, 46 },
		{ 166, 264, 46 },
		{ 166, 265, 46 },
		{ 166, 266, 46 },
		{ 166, 267, 46 },
		{ 166, 268, 46 },
		{ 167, 246, 46 },
		{ 167, 247, 46 },
		{ 167, 248, 46 },
		{ 167, 249, 46 },
		{ 167, 250, 46 },
		{ 167, 251, 46 },
		{ 167, 252, 46 },
		{ 167, 253, 46 },
		{ 167, 254, 46 },
		{ 167, 255, 46 },
		{ 167, 256, 46 },
		{ 167, 257, 46 },
		{ 167, 258, 46 },
		{ 167, 259, 46 },
		{ 167, 260, 46 },
		{ 167, 261, 46 },
		{ 167, 262, 46 },
		{ 167, 263, 46 },
		{ 167, 264, 46 },
		{ 167, 265, 46 },
		{ 167, 266, 46 },
		{ 167, 267, 46 },
		{ 167, 268, 46 },
		{ 168, 246, 46 },
		{ 168, 247, 46 },
		{ 168, 248, 46 },
		{ 168, 249, 46 },
		{ 168, 250, 46 },
		{ 168, 251, 46 },
		{ 168, 252, 46 },
		{ 168, 253, 46 },
		{ 168, 254, 46 },
		{ 168, 255, 46 },
		{ 168, 256, 46 },
		{ 168, 257, 46 },
		{ 168, 258, 46 },
		{ 168, 259, 46 },
		{ 168, 260, 46 },
		{ 168, 261, 46 },
		{ 168, 262, 46 },
		{ 168, 263, 46 },
		{ 168, 264, 46 },
		{ 168, 265, 46 },
		{ 168, 266, 46 },
		{ 168, 267, 46 },
		{ 168, 268, 46 },
		{ 169, 247, 46 },
		{ 169, 248, 46 },
		{ 169, 249, 46 },
		{ 169, 250, 46 },
		{ 169, 251, 46 },
		{ 169, 252, 46 },
		{ 169, 253, 46 },
		{ 169, 254, 46 },
		{ 169, 255, 46 },
		{ 169, 256, 46 },
		{ 169, 257, 46 },
		{ 169, 258, 46 },
		{ 169, 259, 46 },
		{ 169, 260, 46 },
		{ 169, 261, 46 },
		{ 169, 262, 46 },
		{ 169, 263, 46 },
		{ 169, 264, 46 },
		{ 169, 265, 46 },
		{ 169, 266, 46 },
		{ 169, 267, 46 },
		{ 170, 247, 46 },
		{ 170, 248, 46 },
		{ 170, 249, 46 },
		{ 170, 250, 46 },
		{ 170, 251, 46 },
		{ 170, 252, 46 },
		{ 170, 253, 46 },
		{ 170, 254, 46 },
		{ 170, 255, 46 },
		{ 170, 256, 46 },
		{ 170, 257, 46 },
		{ 170, 258, 46 },
		{ 170, 259, 46 },
		{ 170, 260, 46 },
		{ 170, 261, 46 },
		{ 170, 262, 46 },
		{ 170, 263, 46 },
		{ 170, 264, 46 },
		{ 170, 265, 46 },
		{ 170, 266, 46 },
		{ 170, 267, 46 },
		{ 171, 247, 46 },
		{ 171, 248, 46 },
		{ 171, 249, 46 },
		{ 171, 250, 46 },
		{ 171, 251, 46 },
		{ 171, 252, 46 },
		{ 171, 253, 46 },
		{ 171, 254, 46 },
		{ 171, 255, 46 },
		{ 171, 256, 46 },
		{ 171, 257, 46 },
		{ 171, 258, 46 },
		{ 171, 259, 46 },
		{ 171, 260, 46 },
		{ 171, 261, 46 },
		{ 171, 262, 46 },
		{ 171, 263, 46 },
		{ 171, 264, 46 },
		{ 171, 265, 46 },
		{ 171, 266, 46 },
		{ 171, 267, 46 },
		{ 172, 247, 46 },
		{ 172, 248, 46 },
		{ 172, 249, 46 },
		{ 172, 250, 46 },
		{ 172, 251, 46 },
		{ 172, 252, 46 },
		{ 172, 253, 46 },
		{ 172, 254, 46 },
		{ 172, 255, 46 },
		{ 172, 256, 46 },
		{ 172, 257, 46 },
		{ 172, 258, 46 },
		{ 172, 259, 46 },
		{ 172, 260, 46 },
		{ 172, 261, 46 },
		{ 172, 262, 46 },
		{ 172, 263, 46 },
		{ 172, 264, 46 },
		{ 172, 265, 46 },
		{ 172, 266, 46 },
		{ 172, 267, 46 },
		{ 173, 247, 46 },
		{ 173, 248, 46 },
		{ 173, 249, 46 },
		{ 173, 250, 46 },
		{ 173, 251, 46 },
		{ 173, 252, 46 },
		{ 173, 253, 46 },
		{ 173, 254, 46 },
		{ 173, 255, 46 },
		{ 173, 256, 46 },
		{ 173, 257, 46 },
		{ 173, 258, 46 },
		{ 173, 259, 46 },
		{ 173, 260, 46 },
		{ 173, 261, 46 },
		{ 173, 262, 46 },
		{ 173, 263, 46 },
		{ 173, 264, 46 },
		{ 173, 265, 46 },
		{ 173, 266, 46 },
		{ 173, 267, 46 },
		{ 174, 247, 46 },
		{ 174, 248, 46 },
		{ 174, 249, 46 },
		{ 174, 250, 46 },
		{ 174, 251, 46 },
		{ 174, 252, 46 },
		{ 174, 253, 46 },
		{ 174, 254, 46 },
		{ 174, 255, 46 },
		{ 174, 256, 46 },
		{ 174, 257, 46 },
		{ 174, 258, 46 },
		{ 174, 259, 46 },
		{ 174, 260, 46 },
		{ 174, 261, 46 },
		{ 174, 262, 46 },
		{ 174, 263, 46 },
		{ 174, 264, 46 },
		{ 174, 265, 46 },
		{ 174, 266, 46 },
		{ 174, 267, 46 },
		{ 175, 247, 46 },
		{ 175, 248, 46 },
		{ 175, 249, 46 },
		{ 175, 250, 46 },
		{ 175, 251, 46 },
		{ 175, 252, 46 },
		{ 175, 253, 46 },
		{ 175, 254, 46 },
		{ 175, 255, 46 },
		{ 175, 256, 46 },
		{ 175, 257, 46 },
		{ 175, 258, 46 },
		{ 175, 259, 46 },
		{ 175, 260, 46 },
		{ 175, 261, 46 },
		{ 175, 262, 46 },
		{ 175, 263, 46 },
		{ 175, 264, 46 },
		{ 175, 265, 46 },
		{ 175, 266, 46 },
		{ 175, 267, 46 },
		{ 176, 247, 46 },
		{ 176, 248, 46 },
		{ 176, 249, 46 },
		{ 176, 250, 46 },
		{ 176, 251, 46 },
		{ 176, 252, 46 },
		{ 176, 253, 46 },
		{ 176, 254, 46 },
		{ 176, 255, 46 },
		{ 176, 256, 46 },
		{ 176, 257, 46 },
		{ 176, 258, 46 },
		{ 176, 259, 46 },
		{ 176, 260, 46 },
		{ 176, 261, 46 },
		{ 176, 262, 46 },
		{ 176, 263, 46 },
		{ 176, 264, 46 },
		{ 176, 265, 46 },
		{ 176, 266, 46 },
		{ 176, 267, 46 },
		{ 177, 248, 46 },
		{ 177, 249, 46 },
		{ 177, 250, 46 },
		{ 177, 251, 46 },
		{ 177, 252, 46 },
		{ 177, 253, 46 },
		{ 177, 254, 46 },
		{ 177, 255, 46 },
		{ 177, 256, 46 },
		{ 177, 257, 46 },
		{ 177, 258, 46 },
		{ 177, 259, 46 },
		{ 177, 260, 46 },
		{ 177, 261, 46 },
		{ 177, 262, 46 },
		{ 177, 263, 46 },
		{ 177, 264, 46 },
		{ 177, 265, 46 },
		{ 177, 266, 46 },
		{ 178, 248, 46 },
		{ 178, 249, 46 },
		{ 178, 250, 46 },
		{ 178, 251, 46 },
		{ 178, 252, 46 },
		{ 178, 253, 46 },
		{ 178, 254, 46 },
		{ 178, 255, 46 },
		{ 178, 256, 46 },
		{ 178, 257, 46 },
		{ 178, 258, 46 },
		{ 178, 259, 46 },
		{ 178, 260, 46 },
		{ 178, 261, 46 },
		{ 178, 262, 46 },
		{ 178, 263, 46 },
		{ 178, 264, 46 },
		{ 178, 265, 46 },
		{ 178, 266, 46 },
		{ 179, 248, 46 },
		{ 179, 249, 46 },
		{ 179, 250, 46 },
		{ 179, 251, 46 },
		{ 179, 252, 46 },
		{ 179, 253, 46 },
		{ 179, 254, 46 },
		{ 179, 255, 46 },
		{ 179, 256, 46 },
		{ 179, 257, 46 },
		{ 179, 258, 46 },
		{ 179, 259, 46 },
		{ 179, 260, 46 },
		{ 179, 261, 46 },
		{ 179, 262, 46 },
		{ 179, 263, 46 },
		{ 179, 264, 46 },
		{ 179, 265, 46 },
		{ 179, 266, 46 },
		{ 180, 248, 46 },
		{ 180, 249, 46 },
		{ 180, 250, 46 },
		{ 180, 251, 46 },
		{ 180, 252, 46 },
		{ 180, 253, 46 },
		{ 180, 254, 46 },
		{ 180, 255, 46 },
		{ 180, 256, 46 },
		{ 180, 257, 46 },
		{ 180, 258, 46 },
		{ 180, 259, 46 },
		{ 180, 260, 46 },
		{ 180, 261, 46 },
		{ 180, 262, 46 },
		{ 180, 263, 46 },
		{ 180, 264, 46 },
		{ 180, 265, 46 },
		{ 180, 266, 46 },
		{ 181, 248, 46 },
		{ 181, 249, 46 },
		{ 181, 250, 46 },
		{ 181, 251, 46 },
		{ 181, 252, 46 },
		{ 181, 253, 46 },
		{ 181, 254, 46 },
		{ 181, 255, 46 },
		{ 181, 256, 46 },
		{ 181, 257, 46 },
		{ 181, 258, 46 },
		{ 181, 259, 46 },
		{ 181, 260, 46 },
		{ 181, 261, 46 },
		{ 181, 262, 46 },
		{ 181, 263, 46 },
		{ 181, 264, 46 },
		{ 181, 265, 46 },
		{ 181, 266, 46 },
		{ 182, 248, 46 },
		{ 182, 249, 46 },
		{ 182, 250, 46 },
		{ 182, 251, 46 },
		{ 182, 252, 46 },
		{ 182, 253, 46 },
		{ 182, 254, 46 },
		{ 182, 255, 46 },
		{ 182, 256, 46 },
		{ 182, 257, 46 },
		{ 182, 258, 46 },
		{ 182, 259, 46 },
		{ 182, 260, 46 },
		{ 182, 261, 46 },
		{ 182, 262, 46 },
		{ 182, 263, 46 },
		{ 182, 264, 46 },
		{ 182, 265, 46 },
		{ 182, 266, 46 },
		{ 183, 248, 46 },
		{ 183, 249, 46 },
		{ 183, 250, 46 },
		{ 183, 251, 46 },
		{ 183, 252, 46 },
		{ 183, 253, 46 },
		{ 183, 254, 46 },
		{ 183, 255, 46 },
		{ 183, 256, 46 },
		{ 183, 257, 46 },
		{ 183, 258, 46 },
		{ 183, 259, 46 },
		{ 183, 260, 46 },
		{ 183, 261, 46 },
		{ 183, 262, 46 },
		{ 183, 263, 46 },
		{ 183, 264, 46 },
		{ 183, 265, 46 },
		{ 183, 266, 46 },
		{ 184, 248, 46 },
		{ 184, 249, 46 },
		{ 184, 250, 46 },
		{ 184, 251, 46 },
		{ 184, 252, 46 },
		{ 184, 253, 46 },
		{ 184, 254, 46 },
		{ 184, 255, 46 },
		{ 184, 256, 46 },
		{ 184, 257, 46 },
		{ 184, 258, 46 },
		{ 184, 259, 46 },
		{ 184, 260, 46 },
		{ 184, 261, 46 },
		{ 184, 262, 46 },
		{ 184, 263, 46 },
		{ 184, 264, 46 },
		{ 184, 265, 46 },
		{ 184, 266, 46 },
		{ 185, 249, 46 },
		{ 185, 250, 46 },
		{ 185, 251, 46 },
		{ 185, 252, 46 },
		{ 185, 253, 46 },
		{ 185, 254, 46 },
		{ 185, 255, 46 },
		{ 185, 256, 46 },
		{ 185, 257, 46 },
		{ 185, 258, 46 },
		{ 185, 259, 46 },
		{ 185, 260, 46 },
		{ 185, 261, 46 },
		{ 185, 262, 46 },
		{ 185, 263, 46 },
		{ 185, 264, 46 },
		{ 185, 265, 46 },
		{ 186, 249, 46 },
		{ 186, 250, 46 },
		{ 186, 251, 46 },
		{ 186, 252, 46 },
		{ 186, 253, 46 },
		{ 186, 254, 46 },
		{ 186, 255, 46 },
		{ 186, 256, 46 },
		{ 186, 257, 46 },
		{ 186, 258, 46 },
		{ 186, 259, 46 },
		{ 186, 260, 46 },
		{ 186, 261, 46 },
		{ 186, 262, 46 },
		{ 186, 263, 46 },
		{ 186, 264, 46 },
		{ 186, 265, 46 },
		{ 187, 249, 46 },
		{ 187, 250, 46 },
		{ 187, 251, 46 },
		{ 187, 252, 46 },
		{ 187, 253, 46 },
		{ 187, 254, 46 },
		{ 187, 255, 46 },
		{ 187, 256, 46 },
		{ 187, 257, 46 },
		{ 187, 258, 46 },
		{ 187, 259, 46 },
		{ 187, 260, 46 },
		{ 187, 261, 46 },
		{ 187, 262, 46 },
		{ 187, 263, 46 },
		{ 187, 264, 46 },
		{ 187, 265, 46 },
		{ 188, 249, 46 },
		{ 188, 250, 46 },
		{ 188, 251, 46 },
		{ 188, 252, 46 },
		{ 188, 253, 46 },
		{ 188, 254, 46 },
		{ 188, 255, 46 },
		{ 188, 256, 46 },
		{ 188, 257, 46 },
		{ 188, 258, 46 },
		{ 188, 259, 46 },
		{ 188, 260, 46 },
		{ 188, 261, 46 },
		{ 188, 262, 46 },
		{ 188, 263, 46 },
		{ 188, 264, 46 },
		{ 188, 265, 46 },
		{ 189, 249, 46 },
		{ 189, 250, 46 },
		{ 189, 251, 46 },
		{ 189, 252, 46 },
		{ 189, 253, 46 },
		{ 189, 254, 46 },
		{ 189, 255, 46 },
		{ 189, 256, 46 },
		{ 189, 257, 46 },
		{ 189, 258, 46 },
		{ 189, 259, 46 },
		{ 189, 260, 46 },
		{ 189, 261, 46 },
		{ 189, 262, 46 },
		{ 189, 263, 46 },
		{ 189, 264, 46 },
		{ 189, 265, 46 },
		{ 190, 250, 46 },
		{ 190, 251, 46 },
		{ 190, 252, 46 },
		{ 190, 253, 46 },
		{ 190, 254, 46 },
		{ 190, 255, 46 },
		{ 190, 256, 46 },
		{ 190, 257, 46 },
		{ 190, 258, 46 },
		{ 190, 259, 46 },
		{ 190, 260, 46 },
		{ 190, 261, 46 },
		{ 190, 262, 46 },
		{ 190, 263, 46 },
		{ 190, 264, 46 },
		{ 191, 251, 46 },
		{ 191, 252, 46 },
		{ 191, 253, 46 },
		{ 191, 254, 46 },
		{ 191, 255, 46 },
		{ 191, 256, 46 },
		{ 191, 257, 46 },
		{ 191, 258, 46 },
		{ 191, 259, 46 },
		{ 191, 260, 46 },
		{ 191, 261, 46 },
		{ 191, 262, 46 },
		{ 191, 263, 46 },
		{ 323, 251, 46 },
		{ 323, 252, 46 },
		{ 323, 253, 46 },
		{ 323, 254, 46 },
		{ 323, 255, 46 },
		{ 323, 256, 46 },
		{ 323, 257, 46 },
		{ 323, 258, 46 },
		{ 323, 259, 46 },
		{ 323, 260, 46 },
		{ 323, 261, 46 },
		{ 323, 262, 46 },
		{ 323, 263, 46 },
		{ 324, 250, 46 },
		{ 324, 251, 46 },
		{ 324, 252, 46 },
		{ 324, 253, 46 },
		{ 324, 254, 46 },
		{ 324, 255, 46 },
		{ 324, 256, 46 },
		{ 324, 257, 46 },
		{ 324, 258, 46 },
		{ 324, 259, 46 },
		{ 324, 260, 46 },
		{ 324, 261, 46 },
		{ 324, 262, 46 },
		{ 324, 263, 46 },
		{ 324, 264, 46 },
		{ 325, 249, 46 },
		{ 325, 250, 46 },
		{ 325, 251, 46 },
		{ 325, 252, 46 },
		{ 325, 253, 46 },
		{ 325, 254, 46 },
		{ 325, 255, 46 },
		{ 325, 256, 46 },
		{ 325, 257, 46 },
		{ 325, 258, 46 },
		{ 325, 259, 46 },
		{ 325, 260, 46 },
		{ 325, 261, 46 },
		{ 325, 262, 46 },
		{ 325, 263, 46 },
		{ 325, 264, 46 },
		{ 325, 265, 46 },
		{ 326, 249, 46 },
		{ 326, 250, 46 },
		{ 326, 251, 46 },
		{ 326, 252, 46 },
		{ 326, 253, 46 },
		{ 326, 254, 46 },
		{ 326, 255, 46 },
		{ 326, 256, 46 },
		{ 326, 257, 46 },
		{ 326, 258, 46 },
		{ 326, 259, 46 },
		{ 326, 260, 46 },
		{ 326, 261, 46 },
		{ 326, 262, 46 },
		{ 326, 263, 46 },
		{ 326, 264, 46 },
		{ 326, 265, 46 },
		{ 327, 249, 46 },
		{ 327, 250, 46 },
		{ 327, 251, 46 },
		{ 327, 252, 46 },
		{ 327, 253, 46 },
		{ 327, 254, 46 },
		{ 327, 255, 46 },
		{ 327, 256, 46 },
		{ 327, 257, 46 },
		{ 327, 258, 46 },
		{ 327, 259, 46 },
		{ 327, 260, 46 },
		{ 327, 261, 46 },
		{ 327, 262, 46 },
		{ 327, 263, 46 },
		{ 327, 264, 46 },
		{ 327, 265, 46 },
		{ 328, 249, 46 },
		{ 328, 250, 46 },
		{ 328, 251, 46 },
		{ 328, 252, 46 },
		{ 328, 253, 46 },
		{ 328, 254, 46 },
		{ 328, 255, 46 },
		{ 328, 256, 46 },
		{ 328, 257, 46 },
		{ 328, 258, 46 },
		{ 328, 259, 46 },
		{ 328, 260, 46 },
		{ 328, 261, 46 },
		{ 328, 262, 46 },
		{ 328, 263, 46 },
		{ 328, 264, 46 },
		{ 328, 265, 46 },
		{ 329, 249, 46 },
		{ 329, 250, 46 },
		{ 329, 251, 46 },
		{ 329, 252, 46 },
		{ 329, 253, 46 },
		{ 329, 254, 46 },
		{ 329, 255, 46 },
		{ 329, 256, 46 },
		{ 329, 257, 46 },
		{ 329, 258, 46 },
		{ 329, 259, 46 },
		{ 329, 260, 46 },
		{ 329, 261, 46 },
		{ 329, 262, 46 },
		{ 329, 263, 46 },
		{ 329, 264, 46 },
		{ 329, 265, 46 },
		{ 330, 248, 46 },
		{ 330, 249, 46 },
		{ 330, 250, 46 },
		{ 330, 251, 46 },
		{ 330, 252, 46 },
		{ 330, 253, 46 },
		{ 330, 254, 46 },
		{ 330, 255, 46 },
		{ 330, 256, 46 },
		{ 330, 257, 46 },
		{ 330, 258, 46 },
		{ 330, 259, 46 },
		{ 330, 260, 46 },
		{ 330, 261, 46 },
		{ 330, 262, 46 },
		{ 330, 263, 46 },
		{ 330, 264, 46 },
		{ 330, 265, 46 },
		{ 330, 266, 46 },
		{ 331, 248, 46 },
		{ 331, 249, 46 },
		{ 331, 250, 46 },
		{ 331, 251, 46 },
		{ 331, 252, 46 },
		{ 331, 253, 46 },
		{ 331, 254, 46 },
		{ 331, 255, 46 },
		{ 331, 256, 46 },
		{ 331, 257, 46 },
		{ 331, 258, 46 },
		{ 331, 259, 46 },
		{ 331, 260, 46 },
		{ 331, 261, 46 },
		{ 331, 262, 46 },
		{ 331, 263, 46 },
		{ 331, 264, 46 },
		{ 331, 265, 46 },
		{ 331, 266, 46 },
		{ 332, 248, 46 },
		{ 332, 249, 46 },
		{ 332, 250, 46 },
		{ 332, 251, 46 },
		{ 332, 252, 46 },
		{ 332, 253, 46 },
		{ 332, 254, 46 },
		{ 332, 255, 46 },
		{ 332, 256, 46 },
		{ 332, 257, 46 },
		{ 332, 258, 46 },
		{ 332, 259, 46 },
		{ 332, 260, 46 },
		{ 332, 261, 46 },
		{ 332, 262, 46 },
		{ 332, 263, 46 },
		{ 332, 264, 46 },
		{ 332, 265, 46 },
		{ 332, 266, 46 },
		{ 333, 248, 46 },
		{ 333, 249, 46 },
		{ 333, 250, 46 },
		{ 333, 251, 46 },
		{ 333, 252, 46 },
		{ 333, 253, 46 },
		{ 333, 254, 46 },
		{ 333, 255, 46 },
		{ 333, 256, 46 },
		{ 333, 257, 46 },
		{ 333, 258, 46 },
		{ 333, 259, 46 },
		{ 333, 260, 46 },
		{ 333, 261, 46 },
		{ 333, 262, 46 },
		{ 333, 263, 46 },
		{ 333, 264, 46 },
		{ 333, 265, 46 },
		{ 333, 266, 46 },
		{ 334, 248, 46 },
		{ 334, 249, 46 },
		{ 334, 250, 46 },
		{ 334, 251, 46 },
		{ 334, 252, 46 },
		{ 334, 253, 46 },
		{ 334, 254, 46 },
		{ 334, 255, 46 },
		{ 334, 256, 46 },
		{ 334, 257, 46 },
		{ 334, 258, 46 },
		{ 334, 259, 46 },
		{ 334, 260, 46 },
		{ 334, 261, 46 },
		{ 334, 262, 46 },
		{ 334, 263, 46 },
		{ 334, 264, 46 },
		{ 334, 265, 46 },
		{ 334, 266, 46 },
		{ 335, 248, 46 },
		{ 335, 249, 46 },
		{ 335, 250, 46 },
		{ 335, 251, 46 },
		{ 335, 252, 46 },
		{ 335, 253, 46 },
		{ 335, 254, 46 },
		{ 335, 255, 46 },
		{ 335, 256, 46 },
		{ 335, 257, 46 },
		{ 335, 258, 46 },
		{ 335, 259, 46 },
		{ 335, 260, 46 },
		{ 335, 261, 46 },
		{ 335, 262, 46 },
		{ 335, 263, 46 },
		{ 335, 264, 46 },
		{ 335, 265, 46 },
		{ 335, 266, 46 },
		{ 336, 248, 46 },
		{ 336, 249, 46 },
		{ 336, 250, 46 },
		{ 336, 251, 46 },
		{ 336, 252, 46 },
		{ 336, 253, 46 },
		{ 336, 254, 46 },
		{ 336, 255, 46 },
		{ 336, 256, 46 },
		{ 336, 257, 46 },
		{ 336, 258, 46 },
		{ 336, 259, 46 },
		{ 336, 260, 46 },
		{ 336, 261, 46 },
		{ 336, 262, 46 },
		{ 336, 263, 46 },
		{ 336, 264, 46 },
		{ 336, 265, 46 },
		{ 336, 266, 46 },
		{ 337, 248, 46 },
		{ 337, 249, 46 },
		{ 337, 250, 46 },
		{ 337, 251, 46 },
		{ 337, 252, 46 },
		{ 337, 253, 46 },
		{ 337, 254, 46 },
		{ 337, 255, 46 },
		{ 337, 256, 46 },
		{ 337, 257, 46 },
		{ 337, 258, 46 },
		{ 337, 259, 46 },
		{ 337, 260, 46 },
		{ 337, 261, 46 },
		{ 337, 262, 46 },
		{ 337, 263, 46 },
		{ 337, 264, 46 },
		{ 337, 265, 46 },
		{ 337, 266, 46 },
		{ 338, 247, 46 },
		{ 338, 248, 46 },
		{ 338, 249, 46 },
		{ 338, 250, 46 },
		{ 338, 251, 46 },
		{ 338, 252, 46 },
		{ 338, 253, 46 },
		{ 338, 254, 46 },
		{ 338, 255, 46 },
		{ 338, 256, 46 },
		{ 338, 257, 46 },
		{ 338, 258, 46 },
		{ 338, 259, 46 },
		{ 338, 260, 46 },
		{ 338, 261, 46 },
		{ 338, 262, 46 },
		{ 338, 263, 46 },
		{ 338, 264, 46 },
		{ 338, 265, 46 },
		{ 338, 266, 46 },
		{ 338, 267, 46 },
		{ 339, 247, 46 },
		{ 339, 248, 46 },
		{ 339, 249, 46 },
		{ 339, 250, 46 },
		{ 339, 251, 46 },
		{ 339, 252, 46 },
		{ 339, 253, 46 },
		{ 339, 254, 46 },
		{ 339, 255, 46 },
		{ 339, 256, 46 },
		{ 339, 257, 46 },
		{ 339, 258, 46 },
		{ 339, 259, 46 },
		{ 339, 260, 46 },
		{ 339, 261, 46 },
		{ 339, 262, 46 },
		{ 339, 263, 46 },
		{ 339, 264, 46 },
		{ 339, 265, 46 },
		{ 339, 266, 46 },
		{ 339, 267, 46 },
		{ 340, 247, 46 },
		{ 340, 248, 46 },
		{ 340, 249, 46 },
		{ 340, 250, 46 },
		{ 340, 251, 46 },
		{ 340, 252, 46 },
		{ 340, 253, 46 },
		{ 340, 254, 46 },
		{ 340, 255, 46 },
		{ 340, 256, 46 },
		{ 340, 257, 46 },
		{ 340, 258, 46 },
		{ 340, 259, 46 },
		{ 340, 260, 46 },
		{ 340, 261, 46 },
		{ 340, 262, 46 },
		{ 340, 263, 46 },
		{ 340, 264, 46 },
		{ 340, 265, 46 },
		{ 340, 266, 46 },
		{ 340, 267, 46 },
		{ 341, 247, 46 },
		{ 341, 248, 46 },
		{ 341, 249, 46 },
		{ 341, 250, 46 },
		{ 341, 251, 46 },
		{ 341, 252, 46 },
		{ 341, 253, 46 },
		{ 341, 254, 46 },
		{ 341, 255, 46 },
		{ 341, 256, 46 },
		{ 341, 257, 46 },
		{ 341, 258, 46 },
		{ 341, 259, 46 },
		{ 341, 260, 46 },
		{ 341, 261, 46 },
		{ 341, 262, 46 },
		{ 341, 263, 46 },
		{ 341, 264, 46 },
		{ 341, 265, 46 },
		{ 341, 266, 46 },
		{ 341, 267, 46 },
		{ 342, 247, 46 },
		{ 342, 248, 46 },
		{ 342, 249, 46 },
		{ 342, 250, 46 },
		{ 342, 251, 46 },
		{ 342, 252, 46 },
		{ 342, 253, 46 },
		{ 342, 254, 46 },
		{ 342, 255, 46 },
		{ 342, 256, 46 },
		{ 342, 257, 46 },
		{ 342, 258, 46 },
		{ 342, 259, 46 },
		{ 342, 260, 46 },
		{ 342, 261, 46 },
		{ 342, 262, 46 },
		{ 342, 263, 46 },
		{ 342, 264, 46 },
		{ 342, 265, 46 },
		{ 342, 266, 46 },
		{ 342, 267, 46 },
		{ 343, 247, 46 },
		{ 343, 248, 46 },
		{ 343, 249, 46 },
		{ 343, 250, 46 },
		{ 343, 251, 46 },
		{ 343, 252, 46 },
		{ 343, 253, 46 },
		{ 343, 254, 46 },
		{ 343, 255, 46 },
		{ 343, 256, 46 },
		{ 343, 257, 46 },
		{ 343, 258, 46 },
		{ 343, 259, 46 },
		{ 343, 260, 46 },
		{ 343, 261, 46 },
		{ 343, 262, 46 },
		{ 343, 263, 46 },
		{ 343, 264, 46 },
		{ 343, 265, 46 },
		{ 343, 266, 46 },
		{ 343, 267, 46 },
		{ 344, 247, 46 },
		{ 344, 248, 46 },
		{ 344, 249, 46 },
		{ 344, 250, 46 },
		{ 344, 251, 46 },
		{ 344, 252, 46 },
		{ 344, 253, 46 },
		{ 344, 254, 46 },
		{ 344, 255, 46 },
		{ 344, 256, 46 },
		{ 344, 257, 46 },
		{ 344, 258, 46 },
		{ 344, 259, 46 },
		{ 344, 260, 46 },
		{ 344, 261, 46 },
		{ 344, 262, 46 },
		{ 344, 263, 46 },
		{ 344, 264, 46 },
		{ 344, 265, 46 },
		{ 344, 266, 46 },
		{ 344, 267, 46 },
		{ 345, 247, 46 },
		{ 345, 248, 46 },
		{ 345, 249, 46 },
		{ 345, 250, 46 },
		{ 345, 251, 46 },
		{ 345, 252, 46 },
		{ 345, 253, 46 },
		{ 345, 254, 46 },
		{ 345, 255, 46 },
		{ 345, 256, 46 },
		{ 345, 257, 46 },
		{ 345, 258, 46 },
		{ 345, 259, 46 },
		{ 345, 260, 46 },
		{ 345, 261, 46 },
		{ 345, 262, 46 },
		{ 345, 263, 46 },
		{ 345, 264, 46 },
		{ 345, 265, 46 },
		{ 345, 266, 46 },
		{ 345, 267, 46 },
		{ 346, 246, 46 },
		{ 346, 247, 46 },
		{ 346, 248, 46 },
		{ 346, 249, 46 },
		{ 346, 250, 46 },
		{ 346, 251, 46 },
		{ 346, 252, 46 },
		{ 346, 253, 46 },
		{ 346, 254, 46 },
		{ 346, 255, 46 },
		{ 346, 256, 46 },
		{ 346, 257, 46 },
		{ 346, 258, 46 },
		{ 346, 259, 46 },
		{ 346, 260, 46 },
		{ 346, 261, 46 },
		{ 346, 262, 46 },
		{ 346, 263, 46 },
		{ 346, 264, 46 },
		{ 346, 265, 46 },
		{ 346, 266, 46 },
		{ 346, 267, 46 },
		{ 346, 268, 46 },
		{ 347, 246, 46 },
		{ 347, 247, 46 },
		{ 347, 248, 46 },
		{ 347, 249, 46 },
		{ 347, 250, 46 },
		{ 347, 251, 46 },
		{ 347, 252, 46 },
		{ 347, 253, 46 },
		{ 347, 254, 46 },
		{ 347, 255, 46 },
		{ 347, 256, 46 },
		{ 347, 257, 46 },
		{ 347, 258, 46 },
		{ 347, 259, 46 },
		{ 347, 260, 46 },
		{ 347, 261, 46 },
		{ 347, 262, 46 },
		{ 347, 263, 46 },
		{ 347, 264, 46 },
		{ 347, 265, 46 },
		{ 347, 266, 46 },
		{ 347, 267, 46 },
		{ 347, 268, 46 },
		{ 348, 246, 46 },
		{ 348, 247, 46 },
		{ 348, 248, 46 },
		{ 348, 249, 46 },
		{ 348, 250, 46 },
		{ 348, 251, 46 },
		{ 348, 252, 46 },
		{ 348, 253, 46 },
		{ 348, 254, 46 },
		{ 348, 255, 46 },
		{ 348, 256, 46 },
		{ 348, 257, 46 },
		{ 348, 258, 46 },
		{ 348, 259, 46 },
		{ 348, 260, 46 },
		{ 348, 261, 46 },
		{ 348, 262, 46 },
		{ 348, 263, 46 },
		{ 348, 264, 46 },
		{ 348, 265, 46 },
		{ 348, 266, 46 },
		{ 348, 267, 46 },
		{ 348, 268, 46 },
		{ 349, 246, 46 },
		{ 349, 247, 46 },
		{ 349, 248, 46 },
		{ 349, 249, 46 },
		{ 349, 250, 46 },
		{ 349, 251, 46 },
		{ 349, 252, 46 },
		{ 349, 253, 46 },
		{ 349, 254, 46 },
		{ 349, 255, 46 },
		{ 349, 256, 46 },
		{ 349, 257, 46 },
		{ 349, 258, 46 },
		{ 349, 259, 46 },
		{ 349, 260, 46 },
		{ 349, 261, 46 },
		{ 349, 262, 46 },
		{ 349, 263, 46 },
		{ 349, 264, 46 },
		{ 349, 265, 46 },
		{ 349, 266, 46 },
		{ 349, 267, 46 },
		{ 349, 268, 46 },
		{ 350, 246, 46 },
		{ 350, 247, 46 },
		{ 350, 248, 46 },
		{ 350, 249, 46 },
		{ 350, 250, 46 },
		{ 350, 251, 46 },
		{ 350, 252, 46 },
		{ 350, 253, 46 },
		{ 350, 254, 46 },
		{ 350, 255, 46 },
		{ 350, 256, 46 },
		{ 350, 257, 46 },
		{ 350, 258, 46 },
		{ 350, 259, 46 },
		{ 350, 260, 46 },
		{ 350, 261, 46 },
		{ 350, 262, 46 },
		{ 350, 263, 46 },
		{ 350, 264, 46 },
		{ 350, 265, 46 },
		{ 350, 266, 46 },
		{ 350, 267, 46 },
		{ 350, 268, 46 },
		{ 351, 246, 46 },
		{ 351, 247, 46 },
		{ 351, 248, 46 },
		{ 351, 249, 46 },
		{ 351, 250, 46 },
		{ 351, 251, 46 },
		{ 351, 252, 46 },
		{ 351, 253, 46 },
		{ 351, 254, 46 },
		{ 351, 255, 46 },
		{ 351, 256, 46 },
		{ 351, 257, 46 },
		{ 351, 258, 46 },
		{ 351, 259, 46 },
		{ 351, 260, 46 },
		{ 351, 261, 46 },
		{ 351, 262, 46 },
		{ 351, 263, 46 },
		{ 351, 264, 46 },
		{ 351, 265, 46 },
		{ 351, 266, 46 },
		{ 351, 267, 46 },
		{ 351, 268, 46 },
		{ 352, 246, 46 },
		{ 352, 247, 46 },
		{ 352, 248, 46 },
		{ 352, 249, 46 },
		{ 352, 250, 46 },
		{ 352, 251, 46 },
		{ 352, 252, 46 },
		{ 352, 253, 46 },
		{ 352, 254, 46 },
		{ 352, 255, 46 },
		{ 352, 256, 46 },
		{ 352, 257, 46 },
		{ 352, 258, 46 },
		{ 352, 259, 46 },
		{ 352, 260, 46 },
		{ 352, 261, 46 },
		{ 352, 262, 46 },
		{ 352, 263, 46 },
		{ 352, 264, 46 },
		{ 352, 265, 46 },
		{ 352, 266, 46 },
		{ 352, 267, 46 },
		{ 352, 268, 46 },
		{ 353, 246, 46 },
		{ 353, 247, 46 },
		{ 353, 248, 46 },
		{ 353, 249, 46 },
		{ 353, 250, 46 },
		{ 353, 251, 46 },
		{ 353, 252, 46 },
		{ 353, 253, 46 },
		{ 353, 254, 46 },
		{ 353, 255, 46 },
		{ 353, 256, 46 },
		{ 353, 257, 46 },
		{ 353, 258, 46 },
		{ 353, 259, 46 },
		{ 353, 260, 46 },
		{ 353, 261, 46 },
		{ 353, 262, 46 },
		{ 353, 263, 46 },
		{ 353, 264, 46 },
		{ 353, 265, 46 },
		{ 353, 266, 46 },
		{ 353, 267, 46 },
		{ 353, 268, 46 },
		{ 354, 245, 46 },
		{ 354, 246, 46 },
		{ 354, 247, 46 },
		{ 354, 248, 46 },
		{ 354, 249, 46 },
		{ 354, 250, 46 },
		{ 354, 251, 46 },
		{ 354, 252, 46 },
		{ 354, 253, 46 },
		{ 354, 254, 46 },
		{ 354, 255, 46 },
		{ 354, 256, 46 },
		{ 354, 257, 46 },
		{ 354, 258, 46 },
		{ 354, 259, 46 },
		{ 354, 260, 46 },
		{ 354, 261, 46 },
		{ 354, 262, 46 },
		{ 354, 263, 46 },
		{ 354, 264, 46 },
		{ 354, 265, 46 },
		{ 354, 266, 46 },
		{ 354, 267, 46 },
		{ 354, 268, 46 },
		{ 354, 269, 46 },
		{ 355, 245, 46 },
		{ 355, 246, 46 },
		{ 355, 247, 46 },
		{ 355, 248, 46 },
		{ 355, 249, 46 },
		{ 355, 250, 46 },
		{ 355, 251, 46 },
		{ 355, 252, 46 },
		{ 355, 253, 46 },
		{ 355, 254, 46 },
		{ 355, 255, 46 },
		{ 355, 256, 46 },
		{ 355, 257, 46 },
		{ 355, 258, 46 },
		{ 355, 259, 46 },
		{ 355, 260, 46 },
		{ 355, 261, 46 },
		{ 355, 262, 46 },
		{ 355, 263, 46 },
		{ 355, 264, 46 },
		{ 355, 265, 46 },
		{ 355, 266, 46 },
		{ 355, 267, 46 },
		{ 355, 268, 46 },
		{ 355, 269, 46 },
		{ 356, 245, 46 },
		{ 356, 246, 46 },
		{ 356, 247, 46 },
		{ 356, 248, 46 },
		{ 356, 249, 46 },
		{ 356, 250, 46 },
		{ 356, 251, 46 },
		{ 356, 252, 46 },
		{ 356, 253, 46 },
		{ 356, 254, 46 },
		{ 356, 255, 46 },
		{ 356, 256, 46 },
		{ 356, 257, 46 },
		{ 356, 258, 46 },
		{ 356, 259, 46 },
		{ 356, 260, 46 },
		{ 356, 261, 46 },
		{ 356, 262, 46 },
		{ 356, 263, 46 },
		{ 356, 264, 46 },
		{ 356, 265, 46 },
		{ 356, 266, 46 },
		{ 356, 267, 46 },
		{ 356, 268, 46 },
		{ 356, 269, 46 },
		{ 357, 245, 46 },
		{ 357, 246, 46 },
		{ 357, 247, 46 },
		{ 357, 248, 46 },
		{ 357, 249, 46 },
		{ 357, 250, 46 },
		{ 357, 251, 46 },
		{ 357, 252, 46 },
		{ 357, 253, 46 },
		{ 357, 254, 46 },
		{ 357, 255, 46 },
		{ 357, 256, 46 },
		{ 357, 257, 46 },
		{ 357, 258, 46 },
		{ 357, 259, 46 },
		{ 357, 260, 46 },
		{ 357, 261, 46 },
		{ 357, 262, 46 },
		{ 357, 263, 46 },
		{ 357, 264, 46 },
		{ 357, 265, 46 },
		{ 357, 266, 46 },
		{ 357, 267, 46 },
		{ 357, 268, 46 },
		{ 357, 269, 46 },
		{ 358, 245, 46 },
		{ 358, 246, 46 },
		{ 358, 247, 46 },
		{ 358, 248, 46 },
		{ 358, 249, 46 },
		{ 358, 250, 46 },
		{ 358, 251, 46 },
		{ 358, 252, 46 },
		{ 358, 253, 46 },
		{ 358, 254, 46 },
		{ 358, 255, 46 },
		{ 358, 256, 46 },
		{ 358, 257, 46 },
		{ 358, 258, 46 },
		{ 358, 259, 46 },
		{ 358, 260, 46 },
		{ 358, 261, 46 },
		{ 358, 262, 46 },
		{ 358, 263, 46 },
		{ 358, 264, 46 },
		{ 358, 265, 46 },
		{ 358, 266, 46 },
		{ 358, 267, 46 },
		{ 358, 268, 46 },
		{ 358, 269, 46 },
		{ 359, 245, 46 },
		{ 359, 246, 46 },
		{ 359, 247, 46 },
		{ 359, 248, 46 },
		{ 359, 249, 46 },
		{ 359, 250, 46 },
		{ 359, 251, 46 },
		{ 359, 252, 46 },
		{ 359, 253, 46 },
		{ 359, 254, 46 },
		{ 359, 255, 46 },
		{ 359, 256, 46 },
		{ 359, 257, 46 },
		{ 359, 258, 46 },
		{ 359, 259, 46 },
		{ 359, 260, 46 },
		{ 359, 261, 46 },
		{ 359, 262, 46 },
		{ 359, 263, 46 },
		{ 359, 264, 46 },
		{ 359, 265, 46 },
		{ 359, 266, 46 },
		{ 359, 267, 46 },
		{ 359, 268, 46 },
		{ 359, 269, 46 },
		{ 360, 245, 46 },
		{ 360, 246, 46 },
		{ 360, 247, 46 },
		{ 360, 248, 46 },
		{ 360, 249, 46 },
		{ 360, 250, 46 },
		{ 360, 251, 46 },
		{ 360, 252, 46 },
		{ 360, 253, 46 },
		{ 360, 254, 46 },
		{ 360, 255, 46 },
		{ 360, 256, 46 },
		{ 360, 257, 46 },
		{ 360, 258, 46 },
		{ 360, 259, 46 },
		{ 360, 260, 46 },
		{ 360, 261, 46 },
		{ 360, 262, 46 },
		{ 360, 263, 46 },
		{ 360, 264, 46 },
		{ 360, 265, 46 },
		{ 360, 266, 46 },
		{ 360, 267, 46 },
		{ 360, 268, 46 },
		{ 360, 269, 46 },
		{ 361, 245, 46 },
		{ 361, 246, 46 },
		{ 361, 247, 46 },
		{ 361, 248, 46 },
		{ 361, 249, 46 },
		{ 361, 250, 46 },
		{ 361, 251, 46 },
		{ 361, 252, 46 },
		{ 361, 253, 46 },
		{ 361, 254, 46 },
		{ 361, 255, 46 },
		{ 361, 256, 46 },
		{ 361, 257, 46 },
		{ 361, 258, 46 },
		{ 361, 259, 46 },
		{ 361, 260, 46 },
		{ 361, 261, 46 },
		{ 361, 262, 46 },
		{ 361, 263, 46 },
		{ 361, 264, 46 },
		{ 361, 265, 46 },
		{ 361, 266, 46 },
		{ 361, 267, 46 },
		{ 361, 268, 46 },
		{ 361, 269, 46 },
		{ 362, 244, 46 },
		{ 362, 245, 46 },
		{ 362, 246, 46 },
		{ 362, 247, 46 },
		{ 362, 248, 46 },
		{ 362, 249, 46 },
		{ 362, 250, 46 },
		{ 362, 251, 46 },
		{ 362, 252, 46 },
		{ 362, 253, 46 },
		{ 362, 254, 46 },
		{ 362, 255, 46 },
		{ 362, 256, 46 },
		{ 362, 257, 46 },
		{ 362, 258, 46 },
		{ 362, 259, 46 },
		{ 362, 260, 46 },
		{ 362, 261, 46 },
		{ 362, 262, 46 },
		{ 362, 263, 46 },
		{ 362, 264, 46 },
		{ 362, 265, 46 },
		{ 362, 266, 46 },
		{ 362, 267, 46 },
		{ 362, 268, 46 },
		{ 362, 269, 46 },
		{ 362, 270, 46 },
		{ 363, 244, 46 },
		{ 363, 245, 46 },
		{ 363, 246, 46 },
		{ 363, 247, 46 },
		{ 363, 248, 46 },
		{ 363, 249, 46 },
		{ 363, 250, 46 },
		{ 363, 251, 46 },
		{ 363, 252, 46 },
		{ 363, 253, 46 },
		{ 363, 254, 46 },
		{ 363, 255, 46 },
		{ 363, 256, 46 },
		{ 363, 257, 46 },
		{ 363, 258, 46 },
		{ 363, 259, 46 },
		{ 363, 260, 46 },
		{ 363, 261, 46 },
		{ 363, 262, 46 },
		{ 363, 263, 46 },
		{ 363, 264, 46 },
		{ 363, 265, 46 },
		{ 363, 266, 46 },
		{ 363, 267, 46 },
		{ 363, 268, 46 },
		{ 363, 269, 46 },
		{ 363, 270, 46 },
		{ 364, 244, 46 },
		{ 364, 245, 46 },
		{ 364, 246, 46 },
		{ 364, 247, 46 },
		{ 364, 248, 46 },
		{ 364, 249, 46 },
		{ 364, 250, 46 },
		{ 364, 251, 46 },
		{ 364, 252, 46 },
		{ 364, 253, 46 },
		{ 364, 254, 46 },
		{ 364, 255, 46 },
		{ 364, 256, 46 },
		{ 364, 257, 46 },
		{ 364, 258, 46 },
		{ 364, 259, 46 },
		{ 364, 260, 46 },
		{ 364, 261, 46 },
		{ 364, 262, 46 },
		{ 364, 263, 46 },
		{ 364, 264, 46 },
		{ 364, 265, 46 },
		{ 364, 266, 46 },
		{ 364, 267, 46 },
		{ 364, 268, 46 },
		{ 364, 269, 46 },
		{ 364, 270, 46 },
		{ 365, 244, 46 },
		{ 365, 245, 46 },
		{ 365, 246, 46 },
		{ 365, 247, 46 },
		{ 365, 248, 46 },
		{ 365, 249, 46 },
		{ 365, 250, 46 },
		{ 365, 251, 46 },
		{ 365, 252, 46 },
		{ 365, 253, 46 },
		{ 365, 254, 46 },
		{ 365, 255, 46 },
		{ 365, 256, 46 },
		{ 365, 257, 46 },
		{ 365, 258, 46 },
		{ 365, 259, 46 },
		{ 365, 260, 46 },
		{ 365, 261, 46 },
		{ 365, 262, 46 },
		{ 365, 263, 46 },
		{ 365, 264, 46 },
		{ 365, 265, 46 },
		{ 365, 266, 46 },
		{ 365, 267, 46 },
		{ 365, 268, 46 },
		{ 365, 269, 46 },
		{ 365, 270, 46 },
		{ 366, 244, 46 },
		{ 366, 245, 46 },
		{ 366, 246, 46 },
		{ 366, 247, 46 },
		{ 366, 248, 46 },
		{ 366, 249, 46 },
		{ 366, 250, 46 },
		{ 366, 251, 46 },
		{ 366, 252, 46 },
		{ 366, 253, 46 },
		{ 366, 254, 46 },
		{ 366, 255, 46 },
		{ 366, 256, 46 },
		{ 366, 257, 46 },
		{ 366, 258, 46 },
		{ 366, 259, 46 },
		{ 366, 260, 46 },
		{ 366, 261, 46 },
		{ 366, 262, 46 },
		{ 366, 263, 46 },
		{ 366, 264, 46 },
		{ 366, 265, 46 },
		{ 366, 266, 46 },
		{ 366, 267, 46 },
		{ 366, 268, 46 },
		{ 366, 269, 46 },
		{ 366, 270, 46 },
		{ 367, 244, 46 },
		{ 367, 245, 46 },
		{ 367, 246, 46 },
		{ 367, 247, 46 },
		{ 367, 248, 46 },
		{ 367, 249, 46 },
		{ 367, 250, 46 },
		{ 367, 251, 46 },
		{ 367, 252, 46 },
		{ 367, 253, 46 },
		{ 367, 254, 46 },
		{ 367, 255, 46 },
		{ 367, 256, 46 },
		{ 367, 257, 46 },
		{ 367, 258, 46 },
		{ 367, 259, 46 },
		{ 367, 260, 46 },
		{ 367, 261, 46 },
		{ 367, 262, 46 },
		{ 367, 263, 46 },
		{ 367, 264, 46 },
		{ 367, 265, 46 },
		{ 367, 266, 46 },
		{ 367, 267, 46 },
		{ 367, 268, 46 },
		{ 367, 269, 46 },
		{ 367, 270, 46 },
		{ 368, 244, 46 },
		{ 368, 245, 46 },
		{ 368, 246, 46 },
		{ 368, 247, 46 },
		{ 368, 248, 46 },
		{ 368, 249, 46 },
		{ 368, 250, 46 },
		{ 368, 251, 46 },
		{ 368, 252, 46 },
		{ 368, 253, 46 },
		{ 368, 254, 46 },
		{ 368, 255, 46 },
		{ 368, 256, 46 },
		{ 368, 257, 46 },
		{ 368, 258, 46 },
		{ 368, 259, 46 },
		{ 368, 260, 46 },
		{ 368, 261, 46 },
		{ 368, 262, 46 },
		{ 368, 263, 46 },
		{ 368, 264, 46 },
		{ 368, 265, 46 },
		{ 368, 266, 46 },
		{ 368, 267, 46 },
		{ 368, 268, 46 },
		{ 368, 269, 46 },
		{ 368, 270, 46 },
		{ 369, 244, 46 },
		{ 369, 245, 46 },
		{ 369, 246, 46 },
		{ 369, 247, 46 },
		{ 369, 248, 46 },
		{ 369, 249, 46 },
		{ 369, 250, 46 },
		{ 369, 251, 46 },
		{ 369, 252, 46 },
		{ 369, 253, 46 },
		{ 369, 254, 46 },
		{ 369, 255, 46 },
		{ 369, 256, 46 },
		{ 369, 257, 46 },
		{ 369, 258, 46 },
		{ 369, 259, 46 },
		{ 369, 260, 46 },
		{ 369, 261, 46 },
		{ 369, 262, 46 },
		{ 369, 263, 46 },
		{ 369, 264, 46 },
		{ 369, 265, 46 },
		{ 369, 266, 46 },
		{ 369, 267, 46 },
		{ 369, 268, 46 },
		{ 369, 269, 46 },
		{ 369, 270, 46 },
		{ 370, 243, 46 },
		{ 370, 244, 46 },
		{ 370, 245, 46 },
		{ 370, 246, 46 },
		{ 370, 247, 46 },
		{ 370, 248, 46 },
		{ 370, 249, 46 },
		{ 370, 250, 46 },
		{ 370, 251, 46 },
		{ 370, 252, 46 },
		{ 370, 253, 46 },
		{ 370, 254, 46 },
		{ 370, 255, 46 },
		{ 370, 256, 46 },
		{ 370, 257, 46 },
		{ 370, 258, 46 },
		{ 370, 259, 46 },
		{ 370, 260, 46 },
		{ 370, 261, 46 },
		{ 370, 262, 46 },
		{ 370, 263, 46 },
		{ 370, 264, 46 },
		{ 370, 265, 46 },
		{ 370, 266, 46 },
		{ 370, 267, 46 },
		{ 370, 268, 46 },
		{ 370, 269, 46 },
		{ 370, 270, 46 },
		{ 370, 271, 46 },
		{ 371, 243, 46 },
		{ 371, 244, 46 },
		{ 371, 245, 46 },
		{ 371, 246, 46 },
		{ 371, 247, 46 },
		{ 371, 248, 46 },
		{ 371, 249, 46 },
		{ 371, 250, 46 },
		{ 371, 251, 46 },
		{ 371, 252, 46 },
		{ 371, 253, 46 },
		{ 371, 254, 46 },
		{ 371, 255, 46 },
		{ 371, 256, 46 },
		{ 371, 257, 46 },
		{ 371, 258, 46 },
		{ 371, 259, 46 },
		{ 371, 260, 46 },
		{ 371, 261, 46 },
		{ 371, 262, 46 },
		{ 371, 263, 46 },
		{ 371, 264, 46 },
		{ 371, 265, 46 },
		{ 371, 266, 46 },
		{ 371, 267, 46 },
		{ 371, 268, 46 },
		{ 371, 269, 46 },
		{ 371, 270, 46 },
		{ 371, 271, 46 },
		{ 372, 243, 46 },
		{ 372, 244, 46 },
		{ 372, 245, 46 },
		{ 372, 246, 46 },
		{ 372, 247, 46 },
		{ 372, 248, 46 },
		{ 372, 249, 46 },
		{ 372, 250, 46 },
		{ 372, 251, 46 },
		{ 372, 252, 46 },
		{ 372, 253, 46 },
		{ 372, 254, 46 },
		{ 372, 255, 46 },
		{ 372, 256, 46 },
		{ 372, 257, 46 },
		{ 372, 258, 46 },
		{ 372, 259, 46 },
		{ 372, 260, 46 },
		{ 372, 261, 46 },
		{ 372, 262, 46 },
		{ 372, 263, 46 },
		{ 372, 264, 46 },
		{ 372, 265, 46 },
		{ 372, 266, 46 },
		{ 372, 267, 46 },
		{ 372, 268, 46 },
		{ 372, 269, 46 },
		{ 372, 270, 46 },
		{ 372, 271, 46 },
		{ 373, 243, 46 },
		{ 373, 244, 46 },
		{ 373, 245, 46 },
		{ 373, 246, 46 },
		{ 373, 247, 46 },
		{ 373, 248, 46 },
		{ 373, 249, 46 },
		{ 373, 250, 46 },
		{ 373, 251, 46 },
		{ 373, 252, 46 },
		{ 373, 253, 46 },
		{ 373, 254, 46 },
		{ 373, 255, 46 },
		{ 373, 256, 46 },
		{ 373, 257, 46 },
		{ 373, 258, 46 },
		{ 373, 259, 46 },
		{ 373, 260, 46 },
		{ 373, 261, 46 },
		{ 373, 262, 46 },
		{ 373, 263, 46 },
		{ 373, 264, 46 },
		{ 373, 265, 46 },
		{ 373, 266, 46 },
		{ 373, 267, 46 },
		{ 373, 268, 46 },
		{ 373, 269, 46 },
		{ 373, 270, 46 },
		{ 373, 271, 46 },
		{ 374, 243, 46 },
		{ 374, 244, 46 },
		{ 374, 245, 46 },
		{ 374, 246, 46 },
		{ 374, 247, 46 },
		{ 374, 248, 46 },
		{ 374, 249, 46 },
		{ 374, 250, 46 },
		{ 374, 251, 46 },
		{ 374, 252, 46 },
		{ 374, 253, 46 },
		{ 374, 254, 46 },
		{ 374, 255, 46 },
		{ 374, 256, 46 },
		{ 374, 257, 46 },
		{ 374, 258, 46 },
		{ 374, 259, 46 },
		{ 374, 260, 46 },
		{ 374, 261, 46 },
		{ 374, 262, 46 },
		{ 374, 263, 46 },
		{ 374, 264, 46 },
		{ 374, 265, 46 },
		{ 374, 266, 46 },
		{ 374, 267, 46 },
		{ 374, 268, 46 },
		{ 374, 269, 46 },
		{ 374, 270, 46 },
		{ 374, 271, 46 },
		{ 375, 243, 46 },
		{ 375, 244, 46 },
		{ 375, 245, 46 },
		{ 375, 246, 46 },
		{ 375, 247, 46 },
		{ 375, 248, 46 },
		{ 375, 249, 46 },
		{ 375, 250, 46 },
		{ 375, 251, 46 },
		{ 375, 252, 46 },
		{ 375, 253, 46 },
		{ 375, 254, 46 },
		{ 375, 255, 46 },
		{ 375, 256, 46 },
		{ 375, 257, 46 },
		{ 375, 258, 46 },
		{ 375, 259, 46 },
		{ 375, 260, 46 },
		{ 375, 261, 46 },
		{ 375, 262, 46 },
		{ 375, 263, 46 },
		{ 375, 264, 46 },
		{ 375, 265, 46 },
		{ 375, 266, 46 },
		{ 375, 267, 46 },
		{ 375, 268, 46 },
		{ 375, 269, 46 },
		{ 375, 270, 46 },
		{ 375, 271, 46 },
		{ 376, 243, 46 },
		{ 376, 244, 46 },
		{ 376, 245, 46 },
		{ 376, 246, 46 },
		{ 376, 247, 46 },
		{ 376, 248, 46 },
		{ 376, 249, 46 },
		{ 376, 250, 46 },
		{ 376, 251, 46 },
		{ 376, 252, 46 },
		{ 376, 253, 46 },
		{ 376, 254, 46 },
		{ 376, 255, 46 },
		{ 376, 256, 46 },
		{ 376, 257, 46 },
		{ 376, 258, 46 },
		{ 376, 259, 46 },
		{ 376, 260, 46 },
		{ 376, 261, 46 },
		{ 376, 262, 46 },
		{ 376, 263, 46 },
		{ 376, 264, 46 },
		{ 376, 265, 46 },
		{ 376, 266, 46 },
		{ 376, 267, 46 },
		{ 376, 268, 46 },
		{ 376, 269, 46 },
		{ 376, 270, 46 },
		{ 376, 271, 46 },
		{ 377, 243, 46 },
		{ 377, 244, 46 },
		{ 377, 245, 46 },
		{ 377, 246, 46 },
		{ 377, 247, 46 },
		{ 377, 248, 46 },
		{ 377, 249, 46 },
		{ 377, 250, 46 },
		{ 377, 251, 46 },
		{ 377, 252, 46 },
		{ 377, 253, 46 },
		{ 377, 254, 46 },
		{ 377, 255, 46 },
		{ 377, 256, 46 },
		{ 377, 257, 46 },
		{ 377, 258, 46 },
		{ 377, 259, 46 },
		{ 377, 260, 46 },
		{ 377, 261, 46 },
		{ 377, 262, 46 },
		{ 377, 263, 46 },
		{ 377, 264, 46 },
		{ 377, 265, 46 },
		{ 377, 266, 46 },
		{ 377, 267, 46 },
		{ 377, 268, 46 },
		{ 377, 269, 46 },
		{ 377, 270, 46 },
		{ 377, 271, 46 },
		{ 378, 242, 46 },
		{ 378, 243, 46 },
		{ 378, 244, 46 },
		{ 378, 245, 46 },
		{ 378, 246, 46 },
		{ 378, 247, 46 },
		{ 378, 248, 46 },
		{ 378, 249, 46 },
		{ 378, 250, 46 },
		{ 378, 251, 46 },
		{ 378, 252, 46 },
		{ 378, 253, 46 },
		{ 378, 254, 46 },
		{ 378, 255, 46 },
		{ 378, 256, 46 },
		{ 378, 257, 46 },
		{ 378, 258, 46 },
		{ 378, 259, 46 },
		{ 378, 260, 46 },
		{ 378, 261, 46 },
		{ 378, 262, 46 },
		{ 378, 263, 46 },
		{ 378, 264, 46 },
		{ 378, 265, 46 },
		{ 378, 266, 46 },
		{ 378, 267, 46 },
		{ 378, 268, 46 },
		{ 378, 269, 46 },
		{ 378, 270, 46 },
		{ 378, 271, 46 },
		{ 378, 272, 46 },
		{ 379, 242, 46 },
		{ 379, 243, 46 },
		{ 379, 244, 46 },
		{ 379, 245, 46 },
		{ 379, 246, 46 },
		{ 379, 247, 46 },
		{ 379, 248, 46 },
		{ 379, 249, 46 },
		{ 379, 250, 46 },
		{ 379, 251, 46 },
		{ 379, 252, 46 },
		{ 379, 253, 46 },
		{ 379, 254, 46 },
		{ 379, 255, 46 },
		{ 379, 256, 46 },
		{ 379, 257, 46 },
		{ 379, 258, 46 },
		{ 379, 259, 46 },
		{ 379, 260, 46 },
		{ 379, 261, 46 },
		{ 379, 262, 46 },
		{ 379, 263, 46 },
		{ 379, 264, 46 },
		{ 379, 265, 46 },
		{ 379, 266, 46 },
		{ 379, 267, 46 },
		{ 379, 268, 46 },
		{ 379, 269, 46 },
		{ 379, 270, 46 },
		{ 379, 271, 46 },
		{ 379, 272, 46 },
		{ 380, 242, 46 },
		{ 380, 243, 46 },
		{ 380, 244, 46 },
		{ 380, 245, 46 },
		{ 380, 246, 46 },
		{ 380, 247, 46 },
		{ 380, 248, 46 },
		{ 380, 249, 46 },
		{ 380, 250, 46 },
		{ 380, 251, 46 },
		{ 380, 252, 46 },
		{ 380, 253, 46 },
		{ 380, 254, 46 },
		{ 380, 255, 46 },
		{ 380, 256, 46 },
		{ 380, 257, 46 },
		{ 380, 258, 46 },
		{ 380, 259, 46 },
		{ 380, 260, 46 },
		{ 380, 261, 46 },
		{ 380, 262, 46 },
		{ 380, 263, 46 },
		{ 380, 264, 46 },
		{ 380, 265, 46 },
		{ 380, 266, 46 },
		{ 380, 267, 46 },
		{ 380, 268, 46 },
		{ 380, 269, 46 },
		{ 380, 270, 46 },
		{ 380, 271, 46 },
		{ 380, 272, 46 },
		{ 381, 242, 46 },
		{ 381, 243, 46 },
		{ 381, 244, 46 },
		{ 381, 245, 46 },
		{ 381, 246, 46 },
		{ 381, 247, 46 },
		{ 381, 248, 46 },
		{ 381, 249, 46 },
		{ 381, 250, 46 },
		{ 381, 251, 46 },
		{ 381, 252, 46 },
		{ 381, 253, 46 },
		{ 381, 254, 46 },
		{ 381, 255, 46 },
		{ 381, 256, 46 },
		{ 381, 257, 46 },
		{ 381, 258, 46 },
		{ 381, 259, 46 },
		{ 381, 260, 46 },
		{ 381, 261, 46 },
		{ 381, 262, 46 },
		{ 381, 263, 46 },
		{ 381, 264, 46 },
		{ 381, 265, 46 },
		{ 381, 266, 46 },
		{ 381, 267, 46 },
		{ 381, 268, 46 },
		{ 381, 269, 46 },
		{ 381, 270, 46 },
		{ 381, 271, 46 },
		{ 381, 272, 46 },
		{ 382, 242, 46 },
		{ 382, 243, 46 },
		{ 382, 244, 46 },
		{ 382, 245, 46 },
		{ 382, 246, 46 },
		{ 382, 247, 46 },
		{ 382, 248, 46 },
		{ 382, 249, 46 },
		{ 382, 250, 46 },
		{ 382, 251, 46 },
		{ 382, 252, 46 },
		{ 382, 253, 46 },
		{ 382, 254, 46 },
		{ 382, 255, 46 },
		{ 382, 256, 46 },
		{ 382, 257, 46 },
		{ 382, 258, 46 },
		{ 382, 259, 46 },
		{ 382, 260, 46 },
		{ 382, 261, 46 },
		{ 382, 262, 46 },
		{ 382, 263, 46 },
		{ 382, 264, 46 },
		{ 382, 265, 46 },
		{ 382, 266, 46 },
		{ 382, 267, 46 },
		{ 382, 268, 46 },
		{ 382, 269, 46 },
		{ 382, 270, 46 },
		{ 382, 271, 46 },
		{ 382, 272, 46 },
		{ 383, 242, 46 },
		{ 383, 243, 46 },
		{ 383, 244, 46 },
		{ 383, 245, 46 },
		{ 383, 246, 46 },
		{ 383, 247, 46 },
		{ 383, 248, 46 },
		{ 383, 249, 46 },
		{ 383, 250, 46 },
		{ 383, 251, 46 },
		{ 383, 252, 46 },
		{ 383, 253, 46 },
		{ 383, 254, 46 },
		{ 383, 255, 46 },
		{ 383, 256, 46 },
		{ 383, 257, 46 },
		{ 383, 258, 46 },
		{ 383, 259, 46 },
		{ 383, 260, 46 },
		{ 383, 261, 46 },
		{ 383, 262, 46 },
		{ 383, 263, 46 },
		{ 383, 264, 46 },
		{ 383, 265, 46 },
		{ 383, 266, 46 },
		{ 383, 267, 46 },
		{ 383, 268, 46 },
		{ 383, 269, 46 },
		{ 383, 270, 46 },
		{ 383, 271, 46 },
		{ 383, 272, 46 },
		{ 384, 242, 46 },
		{ 384, 243, 46 },
		{ 384, 244, 46 },
		{ 384, 245, 46 },
		{ 384, 246, 46 },
		{ 384, 247, 46 },
		{ 384, 248, 46 },
		{ 384, 249, 46 },
		{ 384, 250, 46 },
		{ 384, 251, 46 },
		{ 384, 252, 46 },
		{ 384, 253, 46 },
		{ 384, 254, 46 },
		{ 384, 255, 46 },
		{ 384, 256, 46 },
		{ 384, 257, 46 },
		{ 384, 258, 46 },
		{ 384, 259, 46 },
		{ 384, 260, 46 },
		{ 384, 261, 46 },
		{ 384, 262, 46 },
		{ 384, 263, 46 },
		{ 384, 264, 46 },
		{ 384, 265, 46 },
		{ 384, 266, 46 },
		{ 384, 267, 46 },
		{ 384, 268, 46 },
		{ 384, 269, 46 },
		{ 384, 270, 46 },
		{ 384, 271, 46 },
		{ 384, 272, 46 },
		{ 385, 242, 46 },
		{ 385, 243, 46 },
		{ 385, 244, 46 },
		{ 385, 245, 46 },
		{ 385, 246, 46 },
		{ 385, 247, 46 },
		{ 385, 248, 46 },
		{ 385, 249, 46 },
		{ 385, 250, 46 },
		{ 385, 251, 46 },
		{ 385, 252, 46 },
		{ 385, 253, 46 },
		{ 385, 254, 46 },
		{ 385, 255, 46 },
		{ 385, 256, 46 },
		{ 385, 257, 46 },
		{ 385, 258, 46 },
		{ 385, 259, 46 },
		{ 385, 260, 46 },
		{ 385, 261, 46 },
		{ 385, 262, 46 },
		{ 385, 263, 46 },
		{ 385, 264, 46 },
		{ 385, 265, 46 },
		{ 385, 266, 46 },
		{ 385, 267, 46 },
		{ 385, 268, 46 },
		{ 385, 269, 46 },
		{ 385, 270, 46 },
		{ 385, 271, 46 },
		{ 385, 272, 46 },
		{ 386, 241, 46 },
		{ 386, 242, 46 },
		{ 386, 243, 46 },
		{ 386, 244, 46 },
		{ 386, 245, 46 },
		{ 386, 246, 46 },
		{ 386, 247, 46 },
		{ 386, 248, 46 },
		{ 386, 249, 46 },
		{ 386, 250, 46 },
		{ 386, 251, 46 },
		{ 386, 252, 46 },
		{ 386, 253, 46 },
		{ 386, 254, 46 },
		{ 386, 255, 46 },
		{ 386, 256, 46 },
		{ 386, 257, 46 },
		{ 386, 258, 46 },
		{ 386, 259, 46 },
		{ 386, 260, 46 },
		{ 386, 261, 46 },
		{ 386, 262, 46 },
		{ 386, 263, 46 },
		{ 386, 264, 46 },
		{ 386, 265, 46 },
		{ 386, 266, 46 },
		{ 386, 267, 46 },
		{ 386, 268, 46 },
		{ 386, 269, 46 },
		{ 386, 270, 46 },
		{ 386, 271, 46 },
		{ 386, 272, 46 },
		{ 386, 273, 46 },
		{ 387, 241, 46 },
		{ 387, 242, 46 },
		{ 387, 243, 46 },
		{ 387, 244, 46 },
		{ 387, 245, 46 },
		{ 387, 246, 46 },
		{ 387, 247, 46 },
		{ 387, 248, 46 },
		{ 387, 249, 46 },
		{ 387, 250, 46 },
		{ 387, 251, 46 },
		{ 387, 252, 46 },
		{ 387, 253, 46 },
		{ 387, 254, 46 },
		{ 387, 255, 46 },
		{ 387, 256, 46 },
		{ 387, 257, 46 },
		{ 387, 258, 46 },
		{ 387, 259, 46 },
		{ 387, 260, 46 },
		{ 387, 261, 46 },
		{ 387, 262, 46 },
		{ 387, 263, 46 },
		{ 387, 264, 46 },
		{ 387, 265, 46 },
		{ 387, 266, 46 },
		{ 387, 267, 46 },
		{ 387, 268, 46 },
		{ 387, 269, 46 },
		{ 387, 270, 46 },
		{ 387, 271, 46 },
		{ 387, 272, 46 },
		{ 387, 273, 46 },
		{ 388, 241, 46 },
		{ 388, 242, 46 },
		{ 388, 243, 46 },
		{ 388, 244, 46 },
		{ 388, 245, 46 },
		{ 388, 246, 46 },
		{ 388, 247, 46 },
		{ 388, 248, 46 },
		{ 388, 249, 46 },
		{ 388, 250, 46 },
		{ 388, 251, 46 },
		{ 388, 252, 46 },
		{ 388, 253, 46 },
		{ 388, 254, 46 },
		{ 388, 255, 46 },
		{ 388, 256, 46 },
		{ 388, 257, 46 },
		{ 388, 258, 46 },
		{ 388, 259, 46 },
		{ 388, 260, 46 },
		{ 388, 261, 46 },
		{ 388, 262, 46 },
		{ 388, 263, 46 },
		{ 388, 264, 46 },
		{ 388, 265, 46 },
		{ 388, 266, 46 },
		{ 388, 267, 46 },
		{ 388, 268, 46 },
		{ 388, 269, 46 },
		{ 388, 270, 46 },
		{ 388, 271, 46 },
		{ 388, 272, 46 },
		{ 388, 273, 46 },
		{ 389, 241, 46 },
		{ 389, 242, 46 },
		{ 389, 243, 46 },
		{ 389, 244, 46 },
		{ 389, 245, 46 },
		{ 389, 246, 46 },
		{ 389, 247, 46 },
		{ 389, 248, 46 },
		{ 389, 249, 46 },
		{ 389, 250, 46 },
		{ 389, 251, 46 },
		{ 389, 252, 46 },
		{ 389, 253, 46 },
		{ 389, 254, 46 },
		{ 389, 255, 46 },
		{ 389, 256, 46 },
		{ 389, 257, 46 },
		{ 389, 258, 46 },
		{ 389, 259, 46 },
		{ 389, 260, 46 },
		{ 389, 261, 46 },
		{ 389, 262, 46 },
		{ 389, 263, 46 },
		{ 389, 264, 46 },
		{ 389, 265, 46 },
		{ 389, 266, 46 },
		{ 389, 267, 46 },
		{ 389, 268, 46 },
		{ 389, 269, 46 },
		{ 389, 270, 46 },
		{ 389, 271, 46 },
		{ 389, 272, 46 },
		{ 389, 273, 46 },
		{ 390, 241, 46 },
		{ 390, 242, 46 },
		{ 390, 243, 46 },
		{ 390, 244, 46 },
		{ 390, 245, 46 },
		{ 390, 246, 46 },
		{ 390, 247, 46 },
		{ 390, 248, 46 },
		{ 390, 249, 46 },
		{ 390, 250, 46 },
		{ 390, 251, 46 },
		{ 390, 252, 46 },
		{ 390, 253, 46 },
		{ 390, 254, 46 },
		{ 390, 255, 46 },
		{ 390, 256, 46 },
		{ 390, 257, 46 },
		{ 390, 258, 46 },
		{ 390, 259, 46 },
		{ 390, 260, 46 },
		{ 390, 261, 46 },
		{ 390, 262, 46 },
		{ 390, 263, 46 },
		{ 390, 264, 46 },
		{ 390, 265, 46 },
		{ 390, 266, 46 },
		{ 390, 267, 46 },
		{ 390, 268, 46 },
		{ 390, 269, 46 },
		{ 390, 270, 46 },
		{ 390, 271, 46 },
		{ 390, 272, 46 },
		{ 390, 273, 46 },
		{ 391, 241, 46 },
		{ 391, 242, 46 },
		{ 391, 243, 46 },
		{ 391, 244, 46 },
		{ 391, 245, 46 },
		{ 391, 246, 46 },
		{ 391, 247, 46 },
		{ 391, 248, 46 },
		{ 391, 249, 46 },
		{ 391, 250, 46 },
		{ 391, 251, 46 },
		{ 391, 252, 46 },
		{ 391, 253, 46 },
		{ 391, 254, 46 },
		{ 391, 255, 46 },
		{ 391, 256, 46 },
		{ 391, 257, 46 },
		{ 391, 258, 46 },
		{ 391, 259, 46 },
		{ 391, 260, 46 },
		{ 391, 261, 46 },
		{ 391, 262, 46 },
		{ 391, 263, 46 },
		{ 391, 264, 46 },
		{ 391, 265, 46 },
		{ 391, 266, 46 },
		{ 391, 267, 46 },
		{ 391, 268, 46 },
		{ 391, 269, 46 },
		{ 391, 270, 46 },
		{ 391, 271, 46 },
		{ 391, 272, 46 },
		{ 391, 273, 46 },
		{ 392, 241, 46 },
		{ 392, 242, 46 },
		{ 392, 243, 46 },
		{ 392, 244, 46 },
		{ 392, 245, 46 },
		{ 392, 246, 46 },
		{ 392, 247, 46 },
		{ 392, 248, 46 },
		{ 392, 249, 46 },
		{ 392, 250, 46 },
		{ 392, 251, 46 },
		{ 392, 252, 46 },
		{ 392, 253, 46 },
		{ 392, 254, 46 },
		{ 392, 255, 46 },
		{ 392, 256, 46 },
		{ 392, 257, 46 },
		{ 392, 258, 46 },
		{ 392, 259, 46 },
		{ 392, 260, 46 },
		{ 392, 261, 46 },
		{ 392, 262, 46 },
		{ 392, 263, 46 },
		{ 392, 264, 46 },
		{ 392, 265, 46 },
		{ 392, 266, 46 },
		{ 392, 267, 46 },
		{ 392, 268, 46 },
		{ 392, 269, 46 },
		{ 392, 270, 46 },
		{ 392, 271, 46 },
		{ 392, 272, 46 },
		{ 392, 273, 46 },
		{ 393, 241, 46 },
		{ 393, 242, 46 },
		{ 393, 243, 46 },
		{ 393, 244, 46 },
		{ 393, 245, 46 },
		{ 393, 246, 46 },
		{ 393, 247, 46 },
		{ 393, 248, 46 },
		{ 393, 249, 46 },
		{ 393, 250, 46 },
		{ 393, 251, 46 },
		{ 393, 252, 46 },
		{ 393, 253, 46 },
		{ 393, 254, 46 },
		{ 393, 255, 46 },
		{ 393, 256, 46 },
		{ 393, 257, 46 },
		{ 393, 258, 46 },
		{ 393, 259, 46 },
		{ 393, 260, 46 },
		{ 393, 261, 46 },
		{ 393, 262, 46 },
		{ 393, 263, 46 },
		{ 393, 264, 46 },
		{ 393, 265, 46 },
		{ 393, 266, 46 },
		{ 393, 267, 46 },
		{ 393, 268, 46 },
		{ 393, 269, 46 },
		{ 393, 270, 46 },
		{ 393, 271, 46 },
		{ 393, 272, 46 },
		{ 393, 273, 46 },
		{ 394, 240, 46 },
		{ 394, 241, 46 },
		{ 394, 242, 46 },
		{ 394, 243, 46 },
		{ 394, 244, 46 },
		{ 394, 245, 46 },
		{ 394, 246, 46 },
		{ 394, 247, 46 },
		{ 394, 248, 46 },
		{ 394, 249, 46 },
		{ 394, 250, 46 },
		{ 394, 251, 46 },
		{ 394, 252, 46 },
		{ 394, 253, 46 },
		{ 394, 254, 46 },
		{ 394, 255, 46 },
		{ 394, 256, 46 },
		{ 394, 257, 46 },
		{ 394, 258, 46 },
		{ 394, 259, 46 },
		{ 394, 260, 46 },
		{ 394, 261, 46 },
		{ 394, 262, 46 },
		{ 394, 263, 46 },
		{ 394, 264, 46 },
		{ 394, 265, 46 },
		{ 394, 266, 46 },
		{ 394, 267, 46 },
		{ 394, 268, 46 },
		{ 394, 269, 46 },
		{ 394, 270, 46 },
		{ 394, 271, 46 },
		{ 394, 272, 46 },
		{ 394, 273, 46 },
		{ 394, 274, 46 },
		{ 395, 240, 46 },
		{ 395, 241, 46 },
		{ 395, 242, 46 },
		{ 395, 243, 46 },
		{ 395, 244, 46 },
		{ 395, 245, 46 },
		{ 395, 246, 46 },
		{ 395, 247, 46 },
		{ 395, 248, 46 },
		{ 395, 249, 46 },
		{ 395, 250, 46 },
		{ 395, 251, 46 },
		{ 395, 252, 46 },
		{ 395, 253, 46 },
		{ 395, 254, 46 },
		{ 395, 255, 46 },
		{ 395, 256, 46 },
		{ 395, 257, 46 },
		{ 395, 258, 46 },
		{ 395, 259, 46 },
		{ 395, 260, 46 },
		{ 395, 261, 46 },
		{ 395, 262, 46 },
		{ 395, 263, 46 },
		{ 395, 264, 46 },
		{ 395, 265, 46 },
		{ 395, 266, 46 },
		{ 395, 267, 46 },
		{ 395, 268, 46 },
		{ 395, 269, 46 },
		{ 395, 270, 46 },
		{ 395, 271, 46 },
		{ 395, 272, 46 },
		{ 395, 273, 46 },
		{ 395, 274, 46 },
		{ 396, 240, 46 },
		{ 396, 241, 46 },
		{ 396, 242, 46 },
		{ 396, 243, 46 },
		{ 396, 244, 46 },
		{ 396, 245, 46 },
		{ 396, 246, 46 },
		{ 396, 247, 46 },
		{ 396, 248, 46 },
		{ 396, 249, 46 },
		{ 396, 250, 46 },
		{ 396, 251, 46 },
		{ 396, 252, 46 },
		{ 396, 253, 46 },
		{ 396, 254, 46 },
		{ 396, 255, 46 },
		{ 396, 256, 46 },
		{ 396, 257, 46 },
		{ 396, 258, 46 },
		{ 396, 259, 46 },
		{ 396, 260, 46 },
		{ 396, 261, 46 },
		{ 396, 262, 46 },
		{ 396, 263, 46 },
		{ 396, 264, 46 },
		{ 396, 265, 46 },
		{ 396, 266, 46 },
		{ 396, 267, 46 },
		{ 396, 268, 46 },
		{ 396, 269, 46 },
		{ 396, 270, 46 },
		{ 396, 271, 46 },
		{ 396, 272, 46 },
		{ 396, 273, 46 },
		{ 396, 274, 46 },
		{ 397, 240, 46 },
		{ 397, 241, 46 },
		{ 397, 242, 46 },
		{ 397, 243, 46 },
		{ 397, 244, 46 },
		{ 397, 245, 46 },
		{ 397, 246, 46 },
		{ 397, 247, 46 },
		{ 397, 248, 46 },
		{ 397, 249, 46 },
		{ 397, 250, 46 },
		{ 397, 251, 46 },
		{ 397, 252, 46 },
		{ 397, 253, 46 },
		{ 397, 254, 46 },
		{ 397, 255, 46 },
		{ 397, 256, 46 },
		{ 397, 257, 46 },
		{ 397, 258, 46 },
		{ 397, 259, 46 },
		{ 397, 260, 46 },
		{ 397, 261, 46 },
		{ 397, 262, 46 },
		{ 397, 263, 46 },
		{ 397, 264, 46 },
		{ 397, 265, 46 },
		{ 397, 266, 46 },
		{ 397, 267, 46 },
		{ 397, 268, 46 },
		{ 397, 269, 46 },
		{ 397, 270, 46 },
		{ 397, 271, 46 },
		{ 397, 272, 46 },
		{ 397, 273, 46 },
		{ 397, 274, 46 },
		{ 398, 240, 46 },
		{ 398, 241, 46 },
		{ 398, 242, 46 },
		{ 398, 243, 46 },
		{ 398, 244, 46 },
		{ 398, 245, 46 },
		{ 398, 246, 46 },
		{ 398, 247, 46 },
		{ 398, 248, 46 },
		{ 398, 249, 46 },
		{ 398, 250, 46 },
		{ 398, 251, 46 },
		{ 398, 252, 46 },
		{ 398, 253, 46 },
		{ 398, 254, 46 },
		{ 398, 255, 46 },
		{ 398, 256, 46 },
		{ 398, 257, 46 },
		{ 398, 258, 46 },
		{ 398, 259, 46 },
		{ 398, 260, 46 },
		{ 398, 261, 46 },
		{ 398, 262, 46 },
		{ 398, 263, 46 },
		{ 398, 264, 46 },
		{ 398, 265, 46 },
		{ 398, 266, 46 },
		{ 398, 267, 46 },
		{ 398, 268, 46 },
		{ 398, 269, 46 },
		{ 398, 270, 46 },
		{ 398, 271, 46 },
		{ 398, 272, 46 },
		{ 398, 273, 46 },
		{ 398, 274, 46 },
		{ 399, 240, 46 },
		{ 399, 241, 46 },
		{ 399, 242, 46 },
		{ 399, 243, 46 },
		{ 399, 244, 46 },
		{ 399, 245, 46 },
		{ 399, 246, 46 },
		{ 399, 247, 46 },
		{ 399, 248, 46 },
		{ 399, 249, 46 },
		{ 399, 250, 46 },
		{ 399, 251, 46 },
		{ 399, 252, 46 },
		{ 399, 253, 46 },
		{ 399, 254, 46 },
		{ 399, 255, 46 },
		{ 399, 256, 46 },
		{ 399, 257, 46 },
		{ 399, 258, 46 },
		{ 399, 259, 46 },
		{ 399, 260, 46 },
		{ 399, 261, 46 },
		{ 399, 262, 46 },
		{ 399, 263, 46 },
		{ 399, 264, 46 },
		{ 399, 265, 46 },
		{ 399, 266, 46 },
		{ 399, 267, 46 },
		{ 399, 268, 46 },
		{ 399, 269, 46 },
		{ 399, 270, 46 },
		{ 399, 271, 46 },
		{ 399, 272, 46 },
		{ 399, 273, 46 },
		{ 399, 274, 46 },
		{ 400, 240, 46 },
		{ 400, 241, 46 },
		{ 400, 242, 46 },
		{ 400, 243, 46 },
		{ 400, 244, 46 },
		{ 400, 245, 46 },
		{ 400, 246, 46 },
		{ 400, 247, 46 },
		{ 400, 248, 46 },
		{ 400, 249, 46 },
		{ 400, 250, 46 },
		{ 400, 251, 46 },
		{ 400, 252, 46 },
		{ 400, 253, 46 },
		{ 400, 254, 46 },
		{ 400, 255, 46 },
		{ 400, 256, 46 },
		{ 400, 257, 46 },
		{ 400, 258, 46 },
		{ 400, 259, 46 },
		{ 400, 260, 46 },
		{ 400, 261, 46 },
		{ 400, 262, 46 },
		{ 400, 263, 46 },
		{ 400, 264, 46 },
		{ 400, 265, 46 },
		{ 400, 266, 46 },
		{ 400, 267, 46 },
		{ 400, 268, 46 },
		{ 400, 269, 46 },
		{ 400, 270, 46 },
		{ 400, 271, 46 },
		{ 400, 272, 46 },
		{ 400, 273, 46 },
		{ 400, 274, 46 },
		{ 401, 240, 46 },
		{ 401, 241, 46 },
		{ 401, 242, 46 },
		{ 401, 243, 46 },
		{ 401, 244, 46 },
		{ 401, 245, 46 },
		{ 401, 246, 46 },
		{ 401, 247, 46 },
		{ 401, 248, 46 },
		{ 401, 249, 46 },
		{ 401, 250, 46 },
		{ 401, 251, 46 },
		{ 401, 252, 46 },
		{ 401, 253, 46 },
		{ 401, 254, 46 },
		{ 401, 255, 46 },
		{ 401, 256, 46 },
		{ 401, 257, 46 },
		{ 401, 258, 46 },
		{ 401, 259, 46 },
		{ 401, 260, 46 },
		{ 401, 261, 46 },
		{ 401, 262, 46 },
		{ 401, 263, 46 },
		{ 401, 264, 46 },
		{ 401, 265, 46 },
		{ 401, 266, 46 },
		{ 401, 267, 46 },
		{ 401, 268, 46 },
		{ 401, 269, 46 },
		{ 401, 270, 46 },
		{ 401, 271, 46 },
		{ 401, 272, 46 },
		{ 401, 273, 46 },
		{ 401, 274, 46 },
		{ 402, 239, 46 },
		{ 402, 240, 46 },
		{ 402, 241, 46 },
		{ 402, 242, 46 },
		{ 402, 243, 46 },
		{ 402, 244, 46 },
		{ 402, 245, 46 },
		{ 402, 246, 46 },
		{ 402, 247, 46 },
		{ 402, 248, 46 },
		{ 402, 249, 46 },
		{ 402, 250, 46 },
		{ 402, 251, 46 },
		{ 402, 252, 46 },
		{ 402, 253, 46 },
		{ 402, 254, 46 },
		{ 402, 255, 46 },
		{ 402, 256, 46 },
		{ 402, 257, 46 },
		{ 402, 258, 46 },
		{ 402, 259, 46 },
		{ 402, 260, 46 },
		{ 402, 261, 46 },
		{ 402, 262, 46 },
		{ 402, 263, 46 },
		{ 402, 264, 46 },
		{ 402, 265, 46 },
		{ 402, 266, 46 },
		{ 402, 267, 46 },
		{ 402, 268, 46 },
		{ 402, 269, 46 },
		{ 402, 270, 46 },
		{ 402, 271, 46 },
		{ 402, 272, 46 },
		{ 402, 273, 46 },
		{ 402, 274, 46 },
		{ 402, 275, 46 },
		{ 403, 239, 46 },
		{ 403, 240, 46 },
		{ 403, 241, 46 },
		{ 403, 242, 46 },
		{ 403, 243, 46 },
		{ 403, 244, 46 },
		{ 403, 245, 46 },
		{ 403, 246, 46 },
		{ 403, 247, 46 },
		{ 403, 248, 46 },
		{ 403, 249, 46 },
		{ 403, 250, 46 },
		{ 403, 251, 46 },
		{ 403, 252, 46 },
		{ 403, 253, 46 },
		{ 403, 254, 46 },
		{ 403, 255, 46 },
		{ 403, 256, 46 },
		{ 403, 257, 46 },
		{ 403, 258, 46 },
		{ 403, 259, 46 },
		{ 403, 260, 46 },
		{ 403, 261, 46 },
		{ 403, 262, 46 },
		{ 403, 263, 46 },
		{ 403, 264, 46 },
		{ 403, 265, 46 },
		{ 403, 266, 46 },
		{ 403, 267, 46 },
		{ 403, 268, 46 },
		{ 403, 269, 46 },
		{ 403, 270, 46 },
		{ 403, 271, 46 },
		{ 403, 272, 46 },
		{ 403, 273, 46 },
		{ 403, 274, 46 },
		{ 403, 275, 46 },
		{ 404, 239, 46 },
		{ 404, 240, 46 },
		{ 404, 241, 46 },
		{ 404, 242, 46 },
		{ 404, 243, 46 },
		{ 404, 244, 46 },
		{ 404, 245, 46 },
		{ 404, 246, 46 },
		{ 404, 247, 46 },
		{ 404, 248, 46 },
		{ 404, 249, 46 },
		{ 404, 250, 46 },
		{ 404, 251, 46 },
		{ 404, 252, 46 },
		{ 404, 253, 46 },
		{ 404, 254, 46 },
		{ 404, 255, 46 },
		{ 404, 256, 46 },
		{ 404, 257, 46 },
		{ 404, 258, 46 },
		{ 404, 259, 46 },
		{ 404, 260, 46 },
		{ 404, 261, 46 },
		{ 404, 262, 46 },
		{ 404, 263, 46 },
		{ 404, 264, 46 },
		{ 404, 265, 46 },
		{ 404, 266, 46 },
		{ 404, 267, 46 },
		{ 404, 268, 46 },
		{ 404, 269, 46 },
		{ 404, 270, 46 },
		{ 404, 271, 46 },
		{ 404, 272, 46 },
		{ 404, 273, 46 },
		{ 404, 274, 46 },
		{ 404, 275, 46 },
		{ 405, 239, 46 },
		{ 405, 240, 46 },
		{ 405, 241, 46 },
		{ 405, 242, 46 },
		{ 405, 243, 46 },
		{ 405, 244, 46 },
		{ 405, 245, 46 },
		{ 405, 246, 46 },
		{ 405, 247, 46 },
		{ 405, 248, 46 },
		{ 405, 249, 46 },
		{ 405, 250, 46 },
		{ 405, 251, 46 },
		{ 405, 252, 46 },
		{ 405, 253, 46 },
		{ 405, 254, 46 },
		{ 405, 255, 46 },
		{ 405, 256, 46 },
		{ 405, 257, 46 },
		{ 405, 258, 46 },
		{ 405, 259, 46 },
		{ 405, 260, 46 },
		{ 405, 261, 46 },
		{ 405, 262, 46 },
		{ 405, 263, 46 },
		{ 405, 264, 46 },
		{ 405, 265, 46 },
		{ 405, 266, 46 },
		{ 405, 267, 46 },
		{ 405, 268, 46 },
		{ 405, 269, 46 },
		{ 405, 270, 46 },
		{ 405, 271, 46 },
		{ 405, 272, 46 },
		{ 405, 273, 46 },
		{ 405, 274, 46 },
		{ 405, 275, 46 },
		{ 406, 239, 46 },
		{ 406, 240, 46 },
		{ 406, 241, 46 },
		{ 406, 242, 46 },
		{ 406, 243, 46 },
		{ 406, 244, 46 },
		{ 406, 245, 46 },
		{ 406, 246, 46 },
		{ 406, 247, 46 },
		{ 406, 248, 46 },
		{ 406, 249, 46 },
		{ 406, 250, 46 },
		{ 406, 251, 46 },
		{ 406, 252, 46 },
		{ 406, 253, 46 },
		{ 406, 254, 46 },
		{ 406, 255, 46 },
		{ 406, 256, 46 },
		{ 406, 257, 46 },
		{ 406, 258, 46 },
		{ 406, 259, 46 },
		{ 406, 260, 46 },
		{ 406, 261, 46 },
		{ 406, 262, 46 },
		{ 406, 263, 46 },
		{ 406, 264, 46 },
		{ 406, 265, 46 },
		{ 406, 266, 46 },
		{ 406, 267, 46 },
		{ 406, 268, 46 },
		{ 406, 269, 46 },
		{ 406, 270, 46 },
		{ 406, 271, 46 },
		{ 406, 272, 46 },
		{ 406, 273, 46 },
		{ 406, 274, 46 },
		{ 406, 275, 46 },
		{ 407, 239, 46 },
		{ 407, 240, 46 },
		{ 407, 241, 46 },
		{ 407, 242, 46 },
		{ 407, 243, 46 },
		{ 407, 244, 46 },
		{ 407, 245, 46 },
		{ 407, 246, 46 },
		{ 407, 247, 46 },
		{ 407, 248, 46 },
		{ 407, 249, 46 },
		{ 407, 250, 46 },
		{ 407, 251, 46 },
		{ 407, 252, 46 },
		{ 407, 253, 46 },
		{ 407, 254, 46 },
		{ 407, 255, 46 },
		{ 407, 256, 46 },
		{ 407, 257, 46 },
		{ 407, 258, 46 },
		{ 407, 259, 46 },
		{ 407, 260, 46 },
		{ 407, 261, 46 },
		{ 407, 262, 46 },
		{ 407, 263, 46 },
		{ 407, 264, 46 },
		{ 407, 265, 46 },
		{ 407, 266, 46 },
		{ 407, 267, 46 },
		{ 407, 268, 46 },
		{ 407, 269, 46 },
		{ 407, 270, 46 },
		{ 407, 271, 46 },
		{ 407, 272, 46 },
		{ 407, 273, 46 },
		{ 407, 274, 46 },
		{ 407, 275, 46 },
		{ 408, 239, 46 },
		{ 408, 240, 46 },
		{ 408, 241, 46 },
		{ 408, 242, 46 },
		{ 408, 243, 46 },
		{ 408, 244, 46 },
		{ 408, 245, 46 },
		{ 408, 246, 46 },
		{ 408, 247, 46 },
		{ 408, 248, 46 },
		{ 408, 249, 46 },
		{ 408, 250, 46 },
		{ 408, 251, 46 },
		{ 408, 252, 46 },
		{ 408, 253, 46 },
		{ 408, 254, 46 },
		{ 408, 255, 46 },
		{ 408, 256, 46 },
		{ 408, 257, 46 },
		{ 408, 258, 46 },
		{ 408, 259, 46 },
		{ 408, 260, 46 },
		{ 408, 261, 46 },
		{ 408, 262, 46 },
		{ 408, 263, 46 },
		{ 408, 264, 46 },
		{ 408, 265, 46 },
		{ 408, 266, 46 },
		{ 408, 267, 46 },
		{ 408, 268, 46 },
		{ 408, 269, 46 },
		{ 408, 270, 46 },
		{ 408, 271, 46 },
		{ 408, 272, 46 },
		{ 408, 273, 46 },
		{ 408, 274, 46 },
		{ 408, 275, 46 },
		{ 409, 239, 46 },
		{ 409, 240, 46 },
		{ 409, 241, 46 },
		{ 409, 242, 46 },
		{ 409, 243, 46 },
		{ 409, 244, 46 },
		{ 409, 245, 46 },
		{ 409, 246, 46 },
		{ 409, 247, 46 },
		{ 409, 248, 46 },
		{ 409, 249, 46 },
		{ 409, 250, 46 },
		{ 409, 251, 46 },
		{ 409, 252, 46 },
		{ 409, 253, 46 },
		{ 409, 254, 46 },
		{ 409, 255, 46 },
		{ 409, 256, 46 },
		{ 409, 257, 46 },
		{ 409, 258, 46 },
		{ 409, 259, 46 },
		{ 409, 260, 46 },
		{ 409, 261, 46 },
		{ 409, 262, 46 },
		{ 409, 263, 46 },
		{ 409, 264, 46 },
		{ 409, 265, 46 },
		{ 409, 266, 46 },
		{ 409, 267, 46 },
		{ 409, 268, 46 },
		{ 409, 269, 46 },
		{ 409, 270, 46 },
		{ 409, 271, 46 },
		{ 409, 272, 46 },
		{ 409, 273, 46 },
		{ 409, 274, 46 },
		{ 409, 275, 46 },
		{ 410, 238, 46 },
		{ 410, 239, 46 },
		{ 410, 240, 46 },
		{ 410, 241, 46 },
		{ 410, 242, 46 },
		{ 410, 243, 46 },
		{ 410, 244, 46 },
		{ 410, 245, 46 },
		{ 410, 246, 46 },
		{ 410, 247, 46 },
		{ 410, 248, 46 },
		{ 410, 249, 46 },
		{ 410, 250, 46 },
		{ 410, 251, 46 },
		{ 410, 252, 46 },
		{ 410, 253, 46 },
		{ 410, 254, 46 },
		{ 410, 255, 46 },
		{ 410, 256, 46 },
		{ 410, 257, 46 },
		{ 410, 258, 46 },
		{ 410, 259, 46 },
		{ 410, 260, 46 },
		{ 410, 261, 46 },
		{ 410, 262, 46 },
		{ 410, 263, 46 },
		{ 410, 264, 46 },
		{ 410, 265, 46 },
		{ 410, 266, 46 },
		{ 410, 267, 46 },
		{ 410, 268, 46 },
		{ 410, 269, 46 },
		{ 410, 270, 46 },
		{ 410, 271, 46 },
		{ 410, 272, 46 },
		{ 410, 273, 46 },
		{ 410, 274, 46 },
		{ 410, 275, 46 },
		{ 410, 276, 46 },
		{ 411, 238, 46 },
		{ 411, 239, 46 },
		{ 411, 240, 46 },
		{ 411, 241, 46 },
		{ 411, 242, 46 },
		{ 411, 243, 46 },
		{ 411, 244, 46 },
		{ 411, 245, 46 },
		{ 411, 246, 46 },
		{ 411, 247, 46 },
		{ 411, 248, 46 },
		{ 411, 249, 46 },
		{ 411, 250, 46 },
		{ 411, 251, 46 },
		{ 411, 252, 46 },
		{ 411, 253, 46 },
		{ 411, 254, 46 },
		{ 411, 255, 46 },
		{ 411, 256, 46 },
		{ 411, 257, 46 },
		{ 411, 258, 46 },
		{ 411, 259, 46 },
		{ 411, 260, 46 },
		{ 411, 261, 46 },
		{ 411, 262, 46 },
		{ 411, 263, 46 },
		{ 411, 264, 46 },
		{ 411, 265, 46 },
		{ 411, 266, 46 },
		{ 411, 267, 46 },
		{ 411, 268, 46 },
		{ 411, 269, 46 },
		{ 411, 270, 46 },
		{ 411, 271, 46 },
		{ 411, 272, 46 },
		{ 411, 273, 46 },
		{ 411, 274, 46 },
		{ 411, 275, 46 },
		{ 411, 276, 46 },
		{ 412, 238, 46 },
		{ 412, 239, 46 },
		{ 412, 240, 46 },
		{ 412, 241, 46 },
		{ 412, 242, 46 },
		{ 412, 243, 46 },
		{ 412, 244, 46 },
		{ 412, 245, 46 },
		{ 412, 246, 46 },
		{ 412, 247, 46 },
		{ 412, 248, 46 },
		{ 412, 249, 46 },
		{ 412, 250, 46 },
		{ 412, 251, 46 },
		{ 412, 252, 46 },
		{ 412, 253, 46 },
		{ 412, 254, 46 },
		{ 412, 255, 46 },
		{ 412, 256, 46 },
		{ 412, 257, 46 },
		{ 412, 258, 46 },
		{ 412, 259, 46 },
		{ 412, 260, 46 },
		{ 412, 261, 46 },
		{ 412, 262, 46 },
		{ 412, 263, 46 },
		{ 412, 264, 46 },
		{ 412, 265, 46 },
		{ 412, 266, 46 },
		{ 412, 267, 46 },
		{ 412, 268, 46 },
		{ 412, 269, 46 },
		{ 412, 270, 46 },
		{ 412, 271, 46 },
		{ 412, 272, 46 },
		{ 412, 273, 46 },
		{ 412, 274, 46 },
		{ 412, 275, 46 },
		{ 412, 276, 46 },
		{ 413, 238, 46 },
		{ 413, 239, 46 },
		{ 413, 240, 46 },
		{ 413, 241, 46 },
		{ 413, 242, 46 },
		{ 413, 243, 46 },
		{ 413, 244, 46 },
		{ 413, 245, 46 },
		{ 413, 246, 46 },
		{ 413, 247, 46 },
		{ 413, 248, 46 },
		{ 413, 249, 46 },
		{ 413, 250, 46 },
		{ 413, 251, 46 },
		{ 413, 252, 46 },
		{ 413, 253, 46 },
		{ 413, 254, 46 },
		{ 413, 255, 46 },
		{ 413, 256, 46 },
		{ 413, 257, 46 },
		{ 413, 258, 46 },
		{ 413, 259, 46 },
		{ 413, 260, 46 },
		{ 413, 261, 46 },
		{ 413, 262, 46 },
		{ 413, 263, 46 },
		{ 413, 264, 46 },
		{ 413, 265, 46 },
		{ 413, 266, 46 },
		{ 413, 267, 46 },
		{ 413, 268, 46 },
		{ 413, 269, 46 },
		{ 413, 270, 46 },
		{ 413, 271, 46 },
		{ 413, 272, 46 },
		{ 413, 273, 46 },
		{ 413, 274, 46 },
		{ 413, 275, 46 },
		{ 413, 276, 46 },
		{ 414, 238, 46 },
		{ 414, 239, 46 },
		{ 414, 240, 46 },
		{ 414, 241, 46 },
		{ 414, 242, 46 },
		{ 414, 243, 46 },
		{ 414, 244, 46 },
		{ 414, 245, 46 },
		{ 414, 246, 46 },
		{ 414, 247, 46 },
		{ 414, 248, 46 },
		{ 414, 249, 46 },
		{ 414, 250, 46 },
		{ 414, 251, 46 },
		{ 414, 252, 46 },
		{ 414, 253, 46 },
		{ 414, 254, 46 },
		{ 414, 255, 46 },
		{ 414, 256, 46 },
		{ 414, 257, 46 },
		{ 414, 258, 46 },
		{ 414, 259, 46 },
		{ 414, 260, 46 },
		{ 414, 261, 46 },
		{ 414, 262, 46 },
		{ 414, 263, 46 },
		{ 414, 264, 46 },
		{ 414, 265, 46 },
		{ 414, 266, 46 },
		{ 414, 267, 46 },
		{ 414, 268, 46 },
		{ 414, 269, 46 },
		{ 414, 270, 46 },
		{ 414, 271, 46 },
		{ 414, 272, 46 },
		{ 414, 273, 46 },
		{ 414, 274, 46 },
		{ 414, 275, 46 },
		{ 414, 276, 46 },
		{ 415, 238, 46 },
		{ 415, 239, 46 },
		{ 415, 240, 46 },
		{ 415, 241, 46 },
		{ 415, 242, 46 },
		{ 415, 243, 46 },
		{ 415, 244, 46 },
		{ 415, 245, 46 },
		{ 415, 246, 46 },
		{ 415, 247, 46 },
		{ 415, 248, 46 },
		{ 415, 249, 46 },
		{ 415, 250, 46 },
		{ 415, 251, 46 },
		{ 415, 252, 46 },
		{ 415, 253, 46 },
		{ 415, 254, 46 },
		{ 415, 255, 46 },
		{ 415, 256, 46 },
		{ 415, 257, 46 },
		{ 415, 258, 46 },
		{ 415, 259, 46 },
		{ 415, 260, 46 },
		{ 415, 261, 46 },
		{ 415, 262, 46 },
		{ 415, 263, 46 },
		{ 415, 264, 46 },
		{ 415, 265, 46 },
		{ 415, 266, 46 },
		{ 415, 267, 46 },
		{ 415, 268, 46 },
		{ 415, 269, 46 },
		{ 415, 270, 46 },
		{ 415, 271, 46 },
		{ 415, 272, 46 },
		{ 415, 273, 46 },
		{ 415, 274, 46 },
		{ 415, 275, 46 },
		{ 415, 276, 46 },
		{ 416, 238, 46 },
		{ 416, 239, 46 },
		{ 416, 240, 46 },
		{ 416, 241, 46 },
		{ 416, 242, 46 },
		{ 416, 243, 46 },
		{ 416, 244, 46 },
		{ 416, 245, 46 },
		{ 416, 246, 46 },
		{ 416, 247, 46 },
		{ 416, 248, 46 },
		{ 416, 249, 46 },
		{ 416, 250, 46 },
		{ 416, 251, 46 },
		{ 416, 252, 46 },
		{ 416, 253, 46 },
		{ 416, 254, 46 },
		{ 416, 255, 46 },
		{ 416, 256, 46 },
		{ 416, 257, 46 },
		{ 416, 258, 46 },
		{ 416, 259, 46 },
		{ 416, 260, 46 },
		{ 416, 261, 46 },
		{ 416, 262, 46 },
		{ 416, 263, 46 },
		{ 416, 264, 46 },
		{ 416, 265, 46 },
		{ 416, 266, 46 },
		{ 416, 267, 46 },
		{ 416, 268, 46 },
		{ 416, 269, 46 },
		{ 416, 270, 46 },
		{ 416, 271, 46 },
		{ 416, 272, 46 },
		{ 416, 273, 46 },
		{ 416, 274, 46 },
		{ 416, 275, 46 },
		{ 416, 276, 46 },
		{ 417, 238, 46 },
		{ 417, 239, 46 },
		{ 417, 240, 46 },
		{ 417, 241, 46 },
		{ 417, 242, 46 },
		{ 417, 243, 46 },
		{ 417, 244, 46 },
		{ 417, 245, 46 },
		{ 417, 246, 46 },
		{ 417, 247, 46 },
		{ 417, 248, 46 },
		{ 417, 249, 46 },
		{ 417, 250, 46 },
		{ 417, 251, 46 },
		{ 417, 252, 46 },
		{ 417, 253, 46 },
		{ 417, 254, 46 },
		{ 417, 255, 46 },
		{ 417, 256, 46 },
		{ 417, 257, 46 },
		{ 417, 258, 46 },
		{ 417, 259, 46 },
		{ 417, 260, 46 },
		{ 417, 261, 46 },
		{ 417, 262, 46 },
		{ 417, 263, 46 },
		{ 417, 264, 46 },
		{ 417, 265, 46 },
		{ 417, 266, 46 },
		{ 417, 267, 46 },
		{ 417, 268, 46 },
		{ 417, 269, 46 },
		{ 417, 270, 46 },
		{ 417, 271, 46 },
		{ 417, 272, 46 },
		{ 417, 273, 46 },
		{ 417, 274, 46 },
		{ 417, 275, 46 },
		{ 417, 276, 46 },
		{ 418, 237, 46 },
		{ 418, 238, 46 },
		{ 418, 239, 46 },
		{ 418, 240, 46 },
		{ 418, 241, 46 },
		{ 418, 242, 46 },
		{ 418, 243, 46 },
		{ 418, 244, 46 },
		{ 418, 245, 46 },
		{ 418, 246, 46 },
		{ 418, 247, 46 },
		{ 418, 248, 46 },
		{ 418, 249, 46 },
		{ 418, 250, 46 },
		{ 418, 251, 46 },
		{ 418, 252, 46 },
		{ 418, 253, 46 },
		{ 418, 254, 46 },
		{ 418, 255, 46 },
		{ 418, 256, 46 },
		{ 418, 257, 46 },
		{ 418, 258, 46 },
		{ 418, 259, 46 },
		{ 418, 260, 46 },
		{ 418, 261, 46 },
		{ 418, 262, 46 },
		{ 418, 263, 46 },
		{ 418, 264, 46 },
		{ 418, 265, 46 },
		{ 418, 266, 46 },
		{ 418, 267, 46 },
		{ 418, 268, 46 },
		{ 418, 269, 46 },
		{ 418, 270, 46 },
		{ 418, 271, 46 },
		{ 418, 272, 46 },
		{ 418, 273, 46 },
		{ 418, 274, 46 },
		{ 418, 275, 46 },
		{ 418, 276, 46 },
		{ 418, 277, 46 },
		{ 419, 237, 46 },
		{ 419, 238, 46 },
		{ 419, 239, 46 },
		{ 419, 240, 46 },
		{ 419, 241, 46 },
		{ 419, 242, 46 },
		{ 419, 243, 46 },
		{ 419, 244, 46 },
		{ 419, 245, 46 },
		{ 419, 246, 46 },
		{ 419, 247, 46 },
		{ 419, 248, 46 },
		{ 419, 249, 46 },
		{ 419, 250, 46 },
		{ 419, 251, 46 },
		{ 419, 252, 46 },
		{ 419, 253, 46 },
		{ 419, 254, 46 },
		{ 419, 255, 46 },
		{ 419, 256, 46 },
		{ 419, 257, 46 },
		{ 419, 258, 46 },
		{ 419, 259, 46 },
		{ 419, 260, 46 },
		{ 419, 261, 46 },
		{ 419, 262, 46 },
		{ 419, 263, 46 },
		{ 419, 264, 46 },
		{ 419, 265, 46 },
		{ 419, 266, 46 },
		{ 419, 267, 46 },
		{ 419, 268, 46 },
		{ 419, 269, 46 },
		{ 419, 270, 46 },
		{ 419, 271, 46 },
		{ 419, 272, 46 },
		{ 419, 273, 46 },
		{ 419, 274, 46 },
		{ 419, 275, 46 },
		{ 419, 276, 46 },
		{ 419, 277, 46 },
		{ 420, 237, 46 },
		{ 420, 238, 46 },
		{ 420, 239, 46 },
		{ 420, 240, 46 },
		{ 420, 241, 46 },
		{ 420, 242, 46 },
		{ 420, 243, 46 },
		{ 420, 244, 46 },
		{ 420, 245, 46 },
		{ 420, 246, 46 },
		{ 420, 247, 46 },
		{ 420, 248, 46 },
		{ 420, 249, 46 },
		{ 420, 250, 46 },
		{ 420, 251, 46 },
		{ 420, 252, 46 },
		{ 420, 253, 46 },
		{ 420, 254, 46 },
		{ 420, 255, 46 },
		{ 420, 256, 46 },
		{ 420, 257, 46 },
		{ 420, 258, 46 },
		{ 420, 259, 46 },
		{ 420, 260, 46 },
		{ 420, 261, 46 },
		{ 420, 262, 46 },
		{ 420, 263, 46 },
		{ 420, 264, 46 },
		{ 420, 265, 46 },
		{ 420, 266, 46 },
		{ 420, 267, 46 },
		{ 420, 268, 46 },
		{ 420, 269, 46 },
		{ 420, 270, 46 },
		{ 420, 271, 46 },
		{ 420, 272, 46 },
		{ 420, 273, 46 },
		{ 420, 274, 46 },
		{ 420, 275, 46 },
		{ 420, 276, 46 },
		{ 420, 277, 46 },
		{ 421, 237, 46 },
		{ 421, 238, 46 },
		{ 421, 239, 46 },
		{ 421, 240, 46 },
		{ 421, 241, 46 },
		{ 421, 242, 46 },
		{ 421, 243, 46 },
		{ 421, 244, 46 },
		{ 421, 245, 46 },
		{ 421, 246, 46 },
		{ 421, 247, 46 },
		{ 421, 248, 46 },
		{ 421, 249, 46 },
		{ 421, 250, 46 },
		{ 421, 251, 46 },
		{ 421, 252, 46 },
		{ 421, 253, 46 },
		{ 421, 254, 46 },
		{ 421, 255, 46 },
		{ 421, 256, 46 },
		{ 421, 257, 46 },
		{ 421, 258, 46 },
		{ 421, 259, 46 },
		{ 421, 260, 46 },
		{ 421, 261, 46 },
		{ 421, 262, 46 },
		{ 421, 263, 46 },
		{ 421, 264, 46 },
		{ 421, 265, 46 },
		{ 421, 266, 46 },
		{ 421, 267, 46 },
		{ 421, 268, 46 },
		{ 421, 269, 46 },
		{ 421, 270, 46 },
		{ 421, 271, 46 },
		{ 421, 272, 46 },
		{ 421, 273, 46 },
		{ 421, 274, 46 },
		{ 421, 275, 46 },
		{ 421, 276, 46 },
		{ 421, 277, 46 },
		{ 422, 237, 46 },
		{ 422, 238, 46 },
		{ 422, 239, 46 },
		{ 422, 240, 46 },
		{ 422, 241, 46 },
		{ 422, 242, 46 },
		{ 422, 243, 46 },
		{ 422, 244, 46 },
		{ 422, 245, 46 },
		{ 422, 246, 46 },
		{ 422, 247, 46 },
		{ 422, 248, 46 },
		{ 422, 249, 46 },
		{ 422, 250, 46 },
		{ 422, 251, 46 },
		{ 422, 252, 46 },
		{ 422, 253, 46 },
		{ 422, 254, 46 },
		{ 422, 255, 46 },
		{ 422, 256, 46 },
		{ 422, 257, 46 },
		{ 422, 258, 46 },
		{ 422, 259, 46 },
		{ 422, 260, 46 },
		{ 422, 261, 46 },
		{ 422, 262, 46 },
		{ 422, 263, 46 },
		{ 422, 264, 46 },
		{ 422, 265, 46 },
		{ 422, 266, 46 },
		{ 422, 267, 46 },
		{ 422, 268, 46 },
		{ 422, 269, 46 },
		{ 422, 270, 46 },
		{ 422, 271, 46 },
		{ 422, 272, 46 },
		{ 422, 273, 46 },
		{ 422, 274, 46 },
		{ 422, 275, 46 },
		{ 422, 276, 46 },
		{ 422, 277, 46 },
		{ 423, 237, 46 },
		{ 423, 238, 46 },
		{ 423, 239, 46 },
		{ 423, 240, 46 },
		{ 423, 241, 46 },
		{ 423, 242, 46 },
		{ 423, 243, 46 },
		{ 423, 244, 46 },
		{ 423, 245, 46 },
		{ 423, 246, 46 },
		{ 423, 247, 46 },
		{ 423, 248, 46 },
		{ 423, 249, 46 },
		{ 423, 250, 46 },
		{ 423, 251, 46 },
		{ 423, 252, 46 },
		{ 423, 253, 46 },
		{ 423, 254, 46 },
		{ 423, 255, 46 },
		{ 423, 256, 46 },
		{ 423, 257, 46 },
		{ 423, 258, 46 },
		{ 423, 259, 46 },
		{ 423, 260, 46 },
		{ 423, 261, 46 },
		{ 423, 262, 46 },
		{ 423, 263, 46 },
		{ 423, 264, 46 },
		{ 423, 265, 46 },
		{ 423, 266, 46 },
		{ 423, 267, 46 },
		{ 423, 268, 46 },
		{ 423, 269, 46 },
		{ 423, 270, 46 },
		{ 423, 271, 46 },
		{ 423, 272, 46 },
		{ 423, 273, 46 },
		{ 423, 274, 46 },
		{ 423, 275, 46 },
		{ 423, 276, 46 },
		{ 423, 277, 46 },
		{ 424, 237, 46 },
		{ 424, 238, 46 },
		{ 424, 239, 46 },
		{ 424, 240, 46 },
		{ 424, 241, 46 },
		{ 424, 242, 46 },
		{ 424, 243, 46 },
		{ 424, 244, 46 },
		{ 424, 245, 46 },
		{ 424, 246, 46 },
		{ 424, 247, 46 },
		{ 424, 248, 46 },
		{ 424, 249, 46 },
		{ 424, 250, 46 },
		{ 424, 251, 46 },
		{ 424, 252, 46 },
		{ 424, 253, 46 },
		{ 424, 254, 46 },
		{ 424, 255, 46 },
		{ 424, 256, 46 },
		{ 424, 257, 46 },
		{ 424, 258, 46 },
		{ 424, 259, 46 },
		{ 424, 260, 46 },
		{ 424, 261, 46 },
		{ 424, 262, 46 },
		{ 424, 263, 46 },
		{ 424, 264, 46 },
		{ 424, 265, 46 },
		{ 424, 266, 46 },
		{ 424, 267, 46 },
		{ 424, 268, 46 },
		{ 424, 269, 46 },
		{ 424, 270, 46 },
		{ 424, 271, 46 },
		{ 424, 272, 46 },
		{ 424, 273, 46 },
		{ 424, 274, 46 },
		{ 424, 275, 46 },
		{ 424, 276, 46 },
		{ 424, 277, 46 },
		{ 425, 237, 46 },
		{ 425, 238, 46 },
		{ 425, 239, 46 },
		{ 425, 240, 46 },
		{ 425, 241, 46 },
		{ 425, 242, 46 },
		{ 425, 243, 46 },
		{ 425, 244, 46 },
		{ 425, 245, 46 },
		{ 425, 246, 46 },
		{ 425, 247, 46 },
		{ 425, 248, 46 },
		{ 425, 249, 46 },
		{ 425, 250, 46 },
		{ 425, 251, 46 },
		{ 425, 252, 46 },
		{ 425, 253, 46 },
		{ 425, 254, 46 },
		{ 425, 255, 46 },
		{ 425, 256, 46 },
		{ 425, 257, 46 },
		{ 425, 258, 46 },
		{ 425, 259, 46 },
		{ 425, 260, 46 },
		{ 425, 261, 46 },
		{ 425, 262, 46 },
		{ 425, 263, 46 },
		{ 425, 264, 46 },
		{ 425, 265, 46 },
		{ 425, 266, 46 },
		{ 425, 267, 46 },
		{ 425, 268, 46 },
		{ 425, 269, 46 },
		{ 425, 270, 46 },
		{ 425, 271, 46 },
		{ 425, 272, 46 },
		{ 425, 273, 46 },
		{ 425, 274, 46 },
		{ 425, 275, 46 },
		{ 425, 276, 46 },
		{ 425, 277, 46 },
		{ 426, 236, 46 },
		{ 426, 237, 46 },
		{ 426, 238, 46 },
		{ 426, 239, 46 },
		{ 426, 240, 46 },
		{ 426, 241, 46 },
		{ 426, 242, 46 },
		{ 426, 243, 46 },
		{ 426, 244, 46 },
		{ 426, 245, 46 },
		{ 426, 246, 46 },
		{ 426, 247, 46 },
		{ 426, 248, 46 },
		{ 426, 249, 46 },
		{ 426, 250, 46 },
		{ 426, 251, 46 },
		{ 426, 252, 46 },
		{ 426, 253, 46 },
		{ 426, 254, 46 },
		{ 426, 255, 46 },
		{ 426, 256, 46 },
		{ 426, 257, 46 },
		{ 426, 258, 46 },
		{ 426, 259, 46 },
		{ 426, 260, 46 },
		{ 426, 261, 46 },
		{ 426, 262, 46 },
		{ 426, 263, 46 },
		{ 426, 264, 46 },
		{ 426, 265, 46 },
		{ 426, 266, 46 },
		{ 426, 267, 46 },
		{ 426, 268, 46 },
		{ 426, 269, 46 },
		{ 426, 270, 46 },
		{ 426, 271, 46 },
		{ 426, 272, 46 },
		{ 426, 273, 46 },
		{ 426, 274, 46 },
		{ 426, 275, 46 },
		{ 426, 276, 46 },
		{ 426, 277, 46 },
		{ 426, 278, 46 },
		{ 427, 236, 46 },
		{ 427, 237, 46 },
		{ 427, 238, 46 },
		{ 427, 239, 46 },
		{ 427, 240, 46 },
		{ 427, 241, 46 },
		{ 427, 242, 46 },
		{ 427, 243, 46 },
		{ 427, 244, 46 },
		{ 427, 245, 46 },
		{ 427, 246, 46 },
		{ 427, 247, 46 },
		{ 427, 248, 46 },
		{ 427, 249, 46 },
		{ 427, 250, 46 },
		{ 427, 251, 46 },
		{ 427, 252, 46 },
		{ 427, 253, 46 },
		{ 427, 254, 46 },
		{ 427, 255, 46 },
		{ 427, 256, 46 },
		{ 427, 257, 46 },
		{ 427, 258, 46 },
		{ 427, 259, 46 },
		{ 427, 260, 46 },
		{ 427, 261, 46 },
		{ 427, 262, 46 },
		{ 427, 263, 46 },
		{ 427, 264, 46 },
		{ 427, 265, 46 },
		{ 427, 266, 46 },
		{ 427, 267, 46 },
		{ 427, 268, 46 },
		{ 427, 269, 46 },
		{ 427, 270, 46 },
		{ 427, 271, 46 },
		{ 427, 272, 46 },
		{ 427, 273, 46 },
		{ 427, 274, 46 },
		{ 427, 275, 46 },
		{ 427, 276, 46 },
		{ 427, 277, 46 },
		{ 427, 278, 46 },
		{ 428, 236, 46 },
		{ 428, 237, 46 },
		{ 428, 238, 46 },
		{ 428, 239, 46 },
		{ 428, 240, 46 },
		{ 428, 241, 46 },
		{ 428, 242, 46 },
		{ 428, 243, 46 },
		{ 428, 244, 46 },
		{ 428, 245, 46 },
		{ 428, 246, 46 },
		{ 428, 247, 46 },
		{ 428, 248, 46 },
		{ 428, 249, 46 },
		{ 428, 250, 46 },
		{ 428, 251, 46 },
		{ 428, 252, 46 },
		{ 428, 253, 46 },
		{ 428, 254, 46 },
		{ 428, 255, 46 },
		{ 428, 256, 46 },
		{ 428, 257, 46 },
		{ 428, 258, 46 },
		{ 428, 259, 46 },
		{ 428, 260, 46 },
		{ 428, 261, 46 },
		{ 428, 262, 46 },
		{ 428, 263, 46 },
		{ 428, 264, 46 },
		{ 428, 265, 46 },
		{ 428, 266, 46 },
		{ 428, 267, 46 },
		{ 428, 268, 46 },
		{ 428, 269, 46 },
		{ 428, 270, 46 },
		{ 428, 271, 46 },
		{ 428, 272, 46 },
		{ 428, 273, 46 },
		{ 428, 274, 46 },
		{ 428, 275, 46 },
		{ 428, 276, 46 },
		{ 428, 277, 46 },
		{ 428, 278, 46 },
		{ 429, 236, 46 },
		{ 429, 237, 46 },
		{ 429, 238, 46 },
		{ 429, 239, 46 },
		{ 429, 240, 46 },
		{ 429, 241, 46 },
		{ 429, 242, 46 },
		{ 429, 243, 46 },
		{ 429, 244, 46 },
		{ 429, 245, 46 },
		{ 429, 246, 46 },
		{ 429, 247, 46 },
		{ 429, 248, 46 },
		{ 429, 249, 46 },
		{ 429, 250, 46 },
		{ 429, 251, 46 },
		{ 429, 252, 46 },
		{ 429, 253, 46 },
		{ 429, 254, 46 },
		{ 429, 255, 46 },
		{ 429, 256, 46 },
		{ 429, 257, 46 },
		{ 429, 258, 46 },
		{ 429, 259, 46 },
		{ 429, 260, 46 },
		{ 429, 261, 46 },
		{ 429, 262, 46 },
		{ 429, 263, 46 },
		{ 429, 264, 46 },
		{ 429, 265, 46 },
		{ 429, 266, 46 },
		{ 429, 267, 46 },
		{ 429, 268, 46 },
		{ 429, 269, 46 },
		{ 429, 270, 46 },
		{ 429, 271, 46 },
		{ 429, 272, 46 },
		{ 429, 273, 46 },
		{ 429, 274, 46 },
		{ 429, 275, 46 },
		{ 429, 276, 46 },
		{ 429, 277, 46 },
		{ 429, 278, 46 },
		{ 430, 236, 46 },
		{ 430, 237, 46 },
		{ 430, 238, 46 },
		{ 430, 239, 46 },
		{ 430, 240, 46 },
		{ 430, 241, 46 },
		{ 430, 242, 46 },
		{ 430, 243, 46 },
		{ 430, 244, 46 },
		{ 430, 245, 46 },
		{ 430, 246, 46 },
		{ 430, 247, 46 },
		{ 430, 248, 46 },
		{ 430, 249, 46 },
		{ 430, 250, 46 },
		{ 430, 251, 46 },
		{ 430, 252, 46 },
		{ 430, 253, 46 },
		{ 430, 254, 46 },
		{ 430, 255, 46 },
		{ 430, 256, 46 },
		{ 430, 257, 46 },
		{ 430, 258, 46 },
		{ 430, 259, 46 },
		{ 430, 260, 46 },
		{ 430, 261, 46 },
		{ 430, 262, 46 },
		{ 430, 263, 46 },
		{ 430, 264, 46 },
		{ 430, 265, 46 },
		{ 430, 266, 46 },
		{ 430, 267, 46 },
		{ 430, 268, 46 },
		{ 430, 269, 46 },
		{ 430, 270, 46 },
		{ 430, 271, 46 },
		{ 430, 272, 46 },
		{ 430, 273, 46 },
		{ 430, 274, 46 },
		{ 430, 275, 46 },
		{ 430, 276, 46 },
		{ 430, 277, 46 },
		{ 430, 278, 46 },
		{ 431, 236, 46 },
		{ 431, 237, 46 },
		{ 431, 238, 46 },
		{ 431, 239, 46 },
		{ 431, 240, 46 },
		{ 431, 241, 46 },
		{ 431, 242, 46 },
		{ 431, 243, 46 },
		{ 431, 244, 46 },
		{ 431, 245, 46 },
		{ 431, 246, 46 },
		{ 431, 247, 46 },
		{ 431, 248, 46 },
		{ 431, 249, 46 },
		{ 431, 250, 46 },
		{ 431, 251, 46 },
		{ 431, 252, 46 },
		{ 431, 253, 46 },
		{ 431, 254, 46 },
		{ 431, 255, 46 },
		{ 431, 256, 46 },
		{ 431, 257, 46 },
		{ 431, 258, 46 },
		{ 431, 259, 46 },
		{ 431, 260, 46 },
		{ 431, 261, 46 },
		{ 431, 262, 46 },
		{ 431, 263, 46 },
		{ 431, 264, 46 },
		{ 431, 265, 46 },
		{ 431, 266, 46 },
		{ 431, 267, 46 },
		{ 431, 268, 46 },
		{ 431, 269, 46 },
		{ 431, 270, 46 },
		{ 431, 271, 46 },
		{ 431, 272, 46 },
		{ 431, 273, 46 },
		{ 431, 274, 46 },
		{ 431, 275, 46 },
		{ 431, 276, 46 },
		{ 431, 277, 46 },
		{ 431, 278, 46 },
		{ 432, 236, 46 },
		{ 432, 237, 46 },
		{ 432, 238, 46 },
		{ 432, 239, 46 },
		{ 432, 240, 46 },
		{ 432, 241, 46 },
		{ 432, 242, 46 },
		{ 432, 243, 46 },
		{ 432, 244, 46 },
		{ 432, 245, 46 },
		{ 432, 246, 46 },
		{ 432, 247, 46 },
		{ 432, 248, 46 },
		{ 432, 249, 46 },
		{ 432, 250, 46 },
		{ 432, 251, 46 },
		{ 432, 252, 46 },
		{ 432, 253, 46 },
		{ 432, 254, 46 },
		{ 432, 255, 46 },
		{ 432, 256, 46 },
		{ 432, 257, 46 },
		{ 432, 258, 46 },
		{ 432, 259, 46 },
		{ 432, 260, 46 },
		{ 432, 261, 46 },
		{ 432, 262, 46 },
		{ 432, 263, 46 },
		{ 432, 264, 46 },
		{ 432, 265, 46 },
		{ 432, 266, 46 },
		{ 432, 267, 46 },
		{ 432, 268, 46 },
		{ 432, 269, 46 },
		{ 432, 270, 46 },
		{ 432, 271, 46 },
		{ 432, 272, 46 },
		{ 432, 273, 46 },
		{ 432, 274, 46 },
		{ 432, 275, 46 },
		{ 432, 276, 46 },
		{ 432, 277, 46 },
		{ 432, 278, 46 },
		{ 433, 236, 46 },
		{ 433, 237, 46 },
		{ 433, 238, 46 },
		{ 433, 239, 46 },
		{ 433, 240, 46 },
		{ 433, 241, 46 },
		{ 433, 242, 46 },
		{ 433, 243, 46 },
		{ 433, 244, 46 },
		{ 433, 245, 46 },
		{ 433, 246, 46 },
		{ 433, 247, 46 },
		{ 433, 248, 46 },
		{ 433, 249, 46 },
		{ 433, 250, 46 },
		{ 433, 251, 46 },
		{ 433, 252, 46 },
		{ 433, 253, 46 },
		{ 433, 254, 46 },
		{ 433, 255, 46 },
		{ 433, 256, 46 },
		{ 433, 257, 46 },
		{ 433, 258, 46 },
		{ 433, 259, 46 },
		{ 433, 260, 46 },
		{ 433, 261, 46 },
		{ 433, 262, 46 },
		{ 433, 263, 46 },
		{ 433, 264, 46 },
		{ 433, 265, 46 },
		{ 433, 266, 46 },
		{ 433, 267, 46 },
		{ 433, 268, 46 },
		{ 433, 269, 46 },
		{ 433, 270, 46 },
		{ 433, 271, 46 },
		{ 433, 272, 46 },
		{ 433, 273, 46 },
		{ 433, 274, 46 },
		{ 433, 275, 46 },
		{ 433, 276, 46 },
		{ 433, 277, 46 },
		{ 433, 278, 46 },
		{ 434, 235, 46 },
		{ 434, 236, 46 },
		{ 434, 237, 46 },
		{ 434, 238, 46 },
		{ 434, 239, 46 },
		{ 434, 240, 46 },
		{ 434, 241, 46 },
		{ 434, 242, 46 },
		{ 434, 243, 46 },
		{ 434, 244, 46 },
		{ 434, 245, 46 },
		{ 434, 246, 46 },
		{ 434, 247, 46 },
		{ 434, 248, 46 },
		{ 434, 249, 46 },
		{ 434, 250, 46 },
		{ 434, 251, 46 },
		{ 434, 252, 46 },
		{ 434, 253, 46 },
		{ 434, 254, 46 },
		{ 434, 255, 46 },
		{ 434, 256, 46 },
		{ 434, 257, 46 },
		{ 434, 258, 46 },
		{ 434, 259, 46 },
		{ 434, 260, 46 },
		{ 434, 261, 46 },
		{ 434, 262, 46 },
		{ 434, 263, 46 },
		{ 434, 264, 46 },
		{ 434, 265, 46 },
		{ 434, 266, 46 },
		{ 434, 267, 46 },
		{ 434, 268, 46 },
		{ 434, 269, 46 },
		{ 434, 270, 46 },
		{ 434, 271, 46 },
		{ 434, 272, 46 },
		{ 434, 273, 46 },
		{ 434, 274, 46 },
		{ 434, 275, 46 },
		{ 434, 276, 46 },
		{ 434, 277, 46 },
		{ 434, 278, 46 },
		{ 434, 279, 46 },
		{ 435, 235, 46 },
		{ 435, 236, 46 },
		{ 435, 237, 46 },
		{ 435, 238, 46 },
		{ 435, 239, 46 },
		{ 435, 240, 46 },
		{ 435, 241, 46 },
		{ 435, 242, 46 },
		{ 435, 243, 46 },
		{ 435, 244, 46 },
		{ 435, 245, 46 },
		{ 435, 246, 46 },
		{ 435, 247, 46 },
		{ 435, 248, 46 },
		{ 435, 249, 46 },
		{ 435, 250, 46 },
		{ 435, 251, 46 },
		{ 435, 252, 46 },
		{ 435, 253, 46 },
		{ 435, 254, 46 },
		{ 435, 255, 46 },
		{ 435, 256, 46 },
		{ 435, 257, 46 },
		{ 435, 258, 46 },
		{ 435, 259, 46 },
		{ 435, 260, 46 },
		{ 435, 261, 46 },
		{ 435, 262, 46 },
		{ 435, 263, 46 },
		{ 435, 264, 46 },
		{ 435, 265, 46 },
		{ 435, 266, 46 },
		{ 435, 267, 46 },
		{ 435, 268, 46 },
		{ 435, 269, 46 },
		{ 435, 270, 46 },
		{ 435, 271, 46 },
		{ 435, 272, 46 },
		{ 435, 273, 46 },
		{ 435, 274, 46 },
		{ 435, 275, 46 },
		{ 435, 276, 46 },
		{ 435, 277, 46 },
		{ 435, 278, 46 },
		{ 435, 279, 46 },
		{ 436, 235, 46 },
		{ 436, 236, 46 },
		{ 436, 237, 46 },
		{ 436, 238, 46 },
		{ 436, 239, 46 },
		{ 436, 240, 46 },
		{ 436, 241, 46 },
		{ 436, 242, 46 },
		{ 436, 243, 46 },
		{ 436, 244, 46 },
		{ 436, 245, 46 },
		{ 436, 246, 46 },
		{ 436, 247, 46 },
		{ 436, 248, 46 },
		{ 436, 249, 46 },
		{ 436, 250, 46 },
		{ 436, 251, 46 },
		{ 436, 252, 46 },
		{ 436, 253, 46 },
		{ 436, 254, 46 },
		{ 436, 255, 46 },
		{ 436, 256, 46 },
		{ 436, 257, 46 },
		{ 436, 258, 46 },
		{ 436, 259, 46 },
		{ 436, 260, 46 },
		{ 436, 261, 46 },
		{ 436, 262, 46 },
		{ 436, 263, 46 },
		{ 436, 264, 46 },
		{ 436, 265, 46 },
		{ 436, 266, 46 },
		{ 436, 267, 46 },
		{ 436, 268, 46 },
		{ 436, 269, 46 },
		{ 436, 270, 46 },
		{ 436, 271, 46 },
		{ 436, 272, 46 },
		{ 436, 273, 46 },
		{ 436, 274, 46 },
		{ 436, 275, 46 },
		{ 436, 276, 46 },
		{ 436, 277, 46 },
		{ 436, 278, 46 },
		{ 436, 279, 46 },
		{ 437, 235, 46 },
		{ 437, 236, 46 },
		{ 437, 237, 46 },
		{ 437, 238, 46 },
		{ 437, 239, 46 },
		{ 437, 240, 46 },
		{ 437, 241, 46 },
		{ 437, 242, 46 },
		{ 437, 243, 46 },
		{ 437, 244, 46 },
		{ 437, 245, 46 },
		{ 437, 246, 46 },
		{ 437, 247, 46 },
		{ 437, 248, 46 },
		{ 437, 249, 46 },
		{ 437, 250, 46 },
		{ 437, 251, 46 },
		{ 437, 252, 46 },
		{ 437, 253, 46 },
		{ 437, 254, 46 },
		{ 437, 255, 46 },
		{ 437, 256, 46 },
		{ 437, 257, 46 },
		{ 437, 258, 46 },
		{ 437, 259, 46 },
		{ 437, 260, 46 },
		{ 437, 261, 46 },
		{ 437, 262, 46 },
		{ 437, 263, 46 },
		{ 437, 264, 46 },
		{ 437, 265, 46 },
		{ 437, 266, 46 },
		{ 437, 267, 46 },
		{ 437, 268, 46 },
		{ 437, 269, 46 },
		{ 437, 270, 46 },
		{ 437, 271, 46 },
		{ 437, 272, 46 },
		{ 437, 273, 46 },
		{ 437, 274, 46 },
		{ 437, 275, 46 },
		{ 437, 276, 46 },
		{ 437, 277, 46 },
		{ 437, 278, 46 },
		{ 437, 279, 46 },
		{ 438, 235, 46 },
		{ 438, 236, 46 },
		{ 438, 237, 46 },
		{ 438, 238, 46 },
		{ 438, 239, 46 },
		{ 438, 240, 46 },
		{ 438, 241, 46 },
		{ 438, 242, 46 },
		{ 438, 243, 46 },
		{ 438, 244, 46 },
		{ 438, 245, 46 },
		{ 438, 246, 46 },
		{ 438, 247, 46 },
		{ 438, 248, 46 },
		{ 438, 249, 46 },
		{ 438, 250, 46 },
		{ 438, 251, 46 },
		{ 438, 252, 46 },
		{ 438, 253, 46 },
		{ 438, 254, 46 },
		{ 438, 255, 46 },
		{ 438, 256, 46 },
		{ 438, 257, 46 },
		{ 438, 258, 46 },
		{ 438, 259, 46 },
		{ 438, 260, 46 },
		{ 438, 261, 46 },
		{ 438, 262, 46 },
		{ 438, 263, 46 },
		{ 438, 264, 46 },
		{ 438, 265, 46 },
		{ 438, 266, 46 },
		{ 438, 267, 46 },
		{ 438, 268, 46 },
		{ 438, 269, 46 },
		{ 438, 270, 46 },
		{ 438, 271, 46 },
		{ 438, 272, 46 },
		{ 438, 273, 46 },
		{ 438, 274, 46 },
		{ 438, 275, 46 },
		{ 438, 276, 46 },
		{ 438, 277, 46 },
		{ 438, 278, 46 },
		{ 438, 279, 46 },
		{ 439, 235, 46 },
		{ 439, 236, 46 },
		{ 439, 237, 46 },
		{ 439, 238, 46 },
		{ 439, 239, 46 },
		{ 439, 240, 46 },
		{ 439, 241, 46 },
		{ 439, 242, 46 },
		{ 439, 243, 46 },
		{ 439, 244, 46 },
		{ 439, 245, 46 },
		{ 439, 246, 46 },
		{ 439, 247, 46 },
		{ 439, 248, 46 },
		{ 439, 249, 46 },
		{ 439, 250, 46 },
		{ 439, 251, 46 },
		{ 439, 252, 46 },
		{ 439, 253, 46 },
		{ 439, 254, 46 },
		{ 439, 255, 46 },
		{ 439, 256, 46 },
		{ 439, 257, 46 },
		{ 439, 258, 46 },
		{ 439, 259, 46 },
		{ 439, 260, 46 },
		{ 439, 261, 46 },
		{ 439, 262, 46 },
		{ 439, 263, 46 },
		{ 439, 264, 46 },
		{ 439, 265, 46 },
		{ 439, 266, 46 },
		{ 439, 267, 46 },
		{ 439, 268, 46 },
		{ 439, 269, 46 },
		{ 439, 270, 46 },
		{ 439, 271, 46 },
		{ 439, 272, 46 },
		{ 439, 273, 46 },
		{ 439, 274, 46 },
		{ 439, 275, 46 },
		{ 439, 276, 46 },
		{ 439, 277, 46 },
		{ 439, 278, 46 },
		{ 439, 279, 46 },
		{ 440, 235, 46 },
		{ 440, 236, 46 },
		{ 440, 237, 46 },
		{ 440, 238, 46 },
		{ 440, 239, 46 },
		{ 440, 240, 46 },
		{ 440, 241, 46 },
		{ 440, 242, 46 },
		{ 440, 243, 46 },
		{ 440, 244, 46 },
		{ 440, 245, 46 },
		{ 440, 246, 46 },
		{ 440, 247, 46 },
		{ 440, 248, 46 },
		{ 440, 249, 46 },
		{ 440, 250, 46 },
		{ 440, 251, 46 },
		{ 440, 252, 46 },
		{ 440, 253, 46 },
		{ 440, 254, 46 },
		{ 440, 255, 46 },
		{ 440, 256, 46 },
		{ 440, 257, 46 },
		{ 440, 258, 46 },
		{ 440, 259, 46 },
		{ 440, 260, 46 },
		{ 440, 261, 46 },
		{ 440, 262, 46 },
		{ 440, 263, 46 },
		{ 440, 264, 46 },
		{ 440, 265, 46 },
		{ 440, 266, 46 },
		{ 440, 267, 46 },
		{ 440, 268, 46 },
		{ 440, 269, 46 },
		{ 440, 270, 46 },
		{ 440, 271, 46 },
		{ 440, 272, 46 },
		{ 440, 273, 46 },
		{ 440, 274, 46 },
		{ 440, 275, 46 },
		{ 440, 276, 46 },
		{ 440, 277, 46 },
		{ 440, 278, 46 },
		{ 440, 279, 46 },
		{ 441, 235, 46 },
		{ 441, 236, 46 },
		{ 441, 237, 46 },
		{ 441, 238, 46 },
		{ 441, 239, 46 },
		{ 441, 240, 46 },
		{ 441, 241, 46 },
		{ 441, 242, 46 },
		{ 441, 243, 46 },
		{ 441, 244, 46 },
		{ 441, 245, 46 },
		{ 441, 246, 46 },
		{ 441, 247, 46 },
		{ 441, 248, 46 },
		{ 441, 249, 46 },
		{ 441, 250, 46 },
		{ 441, 251, 46 },
		{ 441, 252, 46 },
		{ 441, 253, 46 },
		{ 441, 254, 46 },
		{ 441, 255, 46 },
		{ 441, 256, 46 },
		{ 441, 257, 46 },
		{ 441, 258, 46 },
		{ 441, 259, 46 },
		{ 441, 260, 46 },
		{ 441, 261, 46 },
		{ 441, 262, 46 },
		{ 441, 263, 46 },
		{ 441, 264, 46 },
		{ 441, 265, 46 },
		{ 441, 266, 46 },
		{ 441, 267, 46 },
		{ 441, 268, 46 },
		{ 441, 269, 46 },
		{ 441, 270, 46 },
		{ 441, 271, 46 },
		{ 441, 272, 46 },
		{ 441, 273, 46 },
		{ 441, 274, 46 },
		{ 441, 275, 46 },
		{ 441, 276, 46 },
		{ 441, 277, 46 },
		{ 441, 278, 46 },
		{ 441, 279, 46 },
		{ 442, 234, 46 },
		{ 442, 235, 46 },
		{ 442, 236, 46 },
		{ 442, 237, 46 },
		{ 442, 238, 46 },
		{ 442, 239, 46 },
		{ 442, 240, 46 },
		{ 442, 241, 46 },
		{ 442, 242, 46 },
		{ 442, 243, 46 },
		{ 442, 244, 46 },
		{ 442, 245, 46 },
		{ 442, 246, 46 },
		{ 442, 247, 46 },
		{ 442, 248, 46 },
		{ 442, 249, 46 },
		{ 442, 250, 46 },
		{ 442, 251, 46 },
		{ 442, 252, 46 },
		{ 442, 253, 46 },
		{ 442, 254, 46 },
		{ 442, 255, 46 },
		{ 442, 256, 46 },
		{ 442, 257, 46 },
		{ 442, 258, 46 },
		{ 442, 259, 46 },
		{ 442, 260, 46 },
		{ 442, 261, 46 },
		{ 442, 262, 46 },
		{ 442, 263, 46 },
		{ 442, 264, 46 },
		{ 442, 265, 46 },
		{ 442, 266, 46 },
		{ 442, 267, 46 },
		{ 442, 268, 46 },
		{ 442, 269, 46 },
		{ 442, 270, 46 },
		{ 442, 271, 46 },
		{ 442, 272, 46 },
		{ 442, 273, 46 },
		{ 442, 274, 46 },
		{ 442, 275, 46 },
		{ 442, 276, 46 },
		{ 442, 277, 46 },
		{ 442, 278, 46 },
		{ 442, 279, 46 },
		{ 442, 280, 46 },
		{ 443, 234, 46 },
		{ 443, 235, 46 },
		{ 443, 236, 46 },
		{ 443, 237, 46 },
		{ 443, 238, 46 },
		{ 443, 239, 46 },
		{ 443, 240, 46 },
		{ 443, 241, 46 },
		{ 443, 242, 46 },
		{ 443, 243, 46 },
		{ 443, 244, 46 },
		{ 443, 245, 46 },
		{ 443, 246, 46 },
		{ 443, 247, 46 },
		{ 443, 248, 46 },
		{ 443, 249, 46 },
		{ 443, 250, 46 },
		{ 443, 251, 46 },
		{ 443, 252, 46 },
		{ 443, 253, 46 },
		{ 443, 254, 46 },
		{ 443, 255, 46 },
		{ 443, 256, 46 },
		{ 443, 257, 46 },
		{ 443, 258, 46 },
		{ 443, 259, 46 },
		{ 443, 260, 46 },
		{ 443, 261, 46 },
		{ 443, 262, 46 },
		{ 443, 263, 46 },
		{ 443, 264, 46 },
		{ 443, 265, 46 },
		{ 443, 266, 46 },
		{ 443, 267, 46 },
		{ 443, 268, 46 },
		{ 443, 269, 46 },
		{ 443, 270, 46 },
		{ 443, 271, 46 },
		{ 443, 272, 46 },
		{ 443, 273, 46 },
		{ 443, 274, 46 },
		{ 443, 275, 46 },
		{ 443, 276, 46 },
		{ 443, 277, 46 },
		{ 443, 278, 46 },
		{ 443, 279, 46 },
		{ 443, 280, 46 },
		{ 444, 234, 46 },
		{ 444, 235, 46 },
		{ 444, 236, 46 },
		{ 444, 237, 46 },
		{ 444, 238, 46 },
		{ 444, 239, 46 },
		{ 444, 240, 46 },
		{ 444, 241, 46 },
		{ 444, 242, 46 },
		{ 444, 243, 46 },
		{ 444, 244, 46 },
		{ 444, 245, 46 },
		{ 444, 246, 46 },
		{ 444, 247, 46 },
		{ 444, 248, 46 },
		{ 444, 249, 46 },
		{ 444, 250, 46 },
		{ 444, 251, 46 },
		{ 444, 252, 46 },
		{ 444, 253, 46 },
		{ 444, 254, 46 },
		{ 444, 255, 46 },
		{ 444, 256, 46 },
		{ 444, 257, 46 },
		{ 444, 258, 46 },
		{ 444, 259, 46 },
		{ 444, 260, 46 },
		{ 444, 261, 46 },
		{ 444, 262, 46 },
		{ 444, 263, 46 },
		{ 444, 264, 46 },
		{ 444, 265, 46 },
		{ 444, 266, 46 },
		{ 444, 267, 46 },
		{ 444, 268, 46 },
		{ 444, 269, 46 },
		{ 444, 270, 46 },
		{ 444, 271, 46 },
		{ 444, 272, 46 },
		{ 444, 273, 46 },
		{ 444, 274, 46 },
		{ 444, 275, 46 },
		{ 444, 276, 46 },
		{ 444, 277, 46 },
		{ 444, 278, 46 },
		{ 444, 279, 46 },
		{ 444, 280, 46 },
		{ 445, 234, 46 },
		{ 445, 235, 46 },
		{ 445, 236, 46 },
		{ 445, 237, 46 },
		{ 445, 238, 46 },
		{ 445, 239, 46 },
		{ 445, 240, 46 },
		{ 445, 241, 46 },
		{ 445, 242, 46 },
		{ 445, 243, 46 },
		{ 445, 244, 46 },
		{ 445, 245, 46 },
		{ 445, 246, 46 },
		{ 445, 247, 46 },
		{ 445, 248, 46 },
		{ 445, 249, 46 },
		{ 445, 250, 46 },
		{ 445, 251, 46 },
		{ 445, 252, 46 },
		{ 445, 253, 46 },
		{ 445, 254, 46 },
		{ 445, 255, 46 },
		{ 445, 256, 46 },
		{ 445, 257, 46 },
		{ 445, 258, 46 },
		{ 445, 259, 46 },
		{ 445, 260, 46 },
		{ 445, 261, 46 },
		{ 445, 262, 46 },
		{ 445, 263, 46 },
		{ 445, 264, 46 },
		{ 445, 265, 46 },
		{ 445, 266, 46 },
		{ 445, 267, 46 },
		{ 445, 268, 46 },
		{ 445, 269, 46 },
		{ 445, 270, 46 },
		{ 445, 271, 46 },
		{ 445, 272, 46 },
		{ 445, 273, 46 },
		{ 445, 274, 46 },
		{ 445, 275, 46 },
		{ 445, 276, 46 },
		{ 445, 277, 46 },
		{ 445, 278, 46 },
		{ 445, 279, 46 },
		{ 445, 280, 46 },
		{ 446, 234, 46 },
		{ 446, 235, 46 },
		{ 446, 236, 46 },
		{ 446, 237, 46 },
		{ 446, 238, 46 },
		{ 446, 239, 46 },
		{ 446, 240, 46 },
		{ 446, 241, 46 },
		{ 446, 242, 46 },
		{ 446, 243, 46 },
		{ 446, 244, 46 },
		{ 446, 245, 46 },
		{ 446, 246, 46 },
		{ 446, 247, 46 },
		{ 446, 248, 46 },
		{ 446, 249, 46 },
		{ 446, 250, 46 },
		{ 446, 251, 46 },
		{ 446, 252, 46 },
		{ 446, 253, 46 },
		{ 446, 254, 46 },
		{ 446, 255, 46 },
		{ 446, 256, 46 },
		{ 446, 257, 46 },
		{ 446, 258, 46 },
		{ 446, 259, 46 },
		{ 446, 260, 46 },
		{ 446, 261, 46 },
		{ 446, 262, 46 },
		{ 446, 263, 46 },
		{ 446, 264, 46 },
		{ 446, 265, 46 },
		{ 446, 266, 46 },
		{ 446, 267, 46 },
		{ 446, 268, 46 },
		{ 446, 269, 46 },
		{ 446, 270, 46 },
		{ 446, 271, 46 },
		{ 446, 272, 46 },
		{ 446, 273, 46 },
		{ 446, 274, 46 },
		{ 446, 275, 46 },
		{ 446, 276, 46 },
		{ 446, 277, 46 },
		{ 446, 278, 46 },
		{ 446, 279, 46 },
		{ 446, 280, 46 },
		{ 447, 234, 46 },
		{ 447, 235, 46 },
		{ 447, 236, 46 },
		{ 447, 237, 46 },
		{ 447, 238, 46 },
		{ 447, 239, 46 },
		{ 447, 240, 46 },
		{ 447, 241, 46 },
		{ 447, 242, 46 },
		{ 447, 243, 46 },
		{ 447, 244, 46 },
		{ 447, 245, 46 },
		{ 447, 246, 46 },
		{ 447, 247, 46 },
		{ 447, 248, 46 },
		{ 447, 249, 46 },
		{ 447, 250, 46 },
		{ 447, 251, 46 },
		{ 447, 252, 46 },
		{ 447, 253, 46 },
		{ 447, 254, 46 },
		{ 447, 255, 46 },
		{ 447, 256, 46 },
		{ 447, 257, 46 },
		{ 447, 258, 46 },
		{ 447, 259, 46 },
		{ 447, 260, 46 },
		{ 447, 261, 46 },
		{ 447, 262, 46 },
		{ 447, 263, 46 },
		{ 447, 264, 46 },
		{ 447, 265, 46 },
		{ 447, 266, 46 },
		{ 447, 267, 46 },
		{ 447, 268, 46 },
		{ 447, 269, 46 },
		{ 447, 270, 46 },
		{ 447, 271, 46 },
		{ 447, 272, 46 },
		{ 447, 273, 46 },
		{ 447, 274, 46 },
		{ 447, 275, 46 },
		{ 447, 276, 46 },
		{ 447, 277, 46 },
		{ 447, 278, 46 },
		{ 447, 279, 46 },
		{ 447, 280, 46 },
		{ 448, 234, 46 },
		{ 448, 235, 46 },
		{ 448, 236, 46 },
		{ 448, 237, 46 },
		{ 448, 238, 46 },
		{ 448, 239, 46 },
		{ 448, 240, 46 },
		{ 448, 241, 46 },
		{ 448, 242, 46 },
		{ 448, 243, 46 },
		{ 448, 244, 46 },
		{ 448, 245, 46 },
		{ 448, 246, 46 },
		{ 448, 247, 46 },
		{ 448, 248, 46 },
		{ 448, 249, 46 },
		{ 448, 250, 46 },
		{ 448, 251, 46 },
		{ 448, 252, 46 },
		{ 448, 253, 46 },
		{ 448, 254, 46 },
		{ 448, 255, 46 },
		{ 448, 256, 46 },
		{ 448, 257, 46 },
		{ 448, 258, 46 },
		{ 448, 259, 46 },
		{ 448, 260, 46 },
		{ 448, 261, 46 },
		{ 448, 262, 46 },
		{ 448, 263, 46 },
		{ 448, 264, 46 },
		{ 448, 265, 46 },
		{ 448, 266, 46 },
		{ 448, 267, 46 },
		{ 448, 268, 46 },
		{ 448, 269, 46 },
		{ 448, 270, 46 },
		{ 448, 271, 46 },
		{ 448, 272, 46 },
		{ 448, 273, 46 },
		{ 448, 274, 46 },
		{ 448, 275, 46 },
		{ 448, 276, 46 },
		{ 448, 277, 46 },
		{ 448, 278, 46 },
		{ 448, 279, 46 },
		{ 448, 280, 46 },
		{ 449, 234, 46 },
		{ 449, 235, 46 },
		{ 449, 236, 46 },
		{ 449, 237, 46 },
		{ 449, 238, 46 },
		{ 449, 239, 46 },
		{ 449, 240, 46 },
		{ 449, 241, 46 },
		{ 449, 242, 46 },
		{ 449, 243, 46 },
		{ 449, 244, 46 },
		{ 449, 245, 46 },
		{ 449, 246, 46 },
		{ 449, 247, 46 },
		{ 449, 248, 46 },
		{ 449, 249, 46 },
		{ 449, 250, 46 },
		{ 449, 251, 46 },
		{ 449, 252, 46 },
		{ 449, 253, 46 },
		{ 449, 254, 46 },
		{ 449, 255, 46 },
		{ 449, 256, 46 },
		{ 449, 257, 46 },
		{ 449, 258, 46 },
		{ 449, 259, 46 },
		{ 449, 260, 46 },
		{ 449, 261, 46 },
		{ 449, 262, 46 },
		{ 449, 263, 46 },
		{ 449, 264, 46 },
		{ 449, 265, 46 },
		{ 449, 266, 46 },
		{ 449, 267, 46 },
		{ 449, 268, 46 },
		{ 449, 269, 46 },
		{ 449, 270, 46 },
		{ 449, 271, 46 },
		{ 449, 272, 46 },
		{ 449, 273, 46 },
		{ 449, 274, 46 },
		{ 449, 275, 46 },
		{ 449, 276, 46 },
		{ 449, 277, 46 },
		{ 449, 278, 46 },
		{ 449, 279, 46 },
		{ 449, 280, 46 },
		{ 450, 233, 46 },
		{ 450, 234, 46 },
		{ 450, 235, 46 },
		{ 450, 236, 46 },
		{ 450, 237, 46 },
		{ 450, 238, 46 },
		{ 450, 239, 46 },
		{ 450, 240, 46 },
		{ 450, 241, 46 },
		{ 450, 242, 46 },
		{ 450, 243, 46 },
		{ 450, 244, 46 },
		{ 450, 245, 46 },
		{ 450, 246, 46 },
		{ 450, 247, 46 },
		{ 450, 248, 46 },
		{ 450, 249, 46 },
		{ 450, 250, 46 },
		{ 450, 251, 46 },
		{ 450, 252, 46 },
		{ 450, 253, 46 },
		{ 450, 254, 46 },
		{ 450, 255, 46 },
		{ 450, 256, 46 },
		{ 450, 257, 46 },
		{ 450, 258, 46 },
		{ 450, 259, 46 },
		{ 450, 260, 46 },
		{ 450, 261, 46 },
		{ 450, 262, 46 },
		{ 450, 263, 46 },
		{ 450, 264, 46 },
		{ 450, 265, 46 },
		{ 450, 266, 46 },
		{ 450, 267, 46 },
		{ 450, 268, 46 },
		{ 450, 269, 46 },
		{ 450, 270, 46 },
		{ 450, 271, 46 },
		{ 450, 272, 46 },
		{ 450, 273, 46 },
		{ 450, 274, 46 },
		{ 450, 275, 46 },
		{ 450, 276, 46 },
		{ 450, 277, 46 },
		{ 450, 278, 46 },
		{ 450, 279, 46 },
		{ 450, 280, 46 },
		{ 450, 281, 46 },
		{ 451, 233, 46 },
		{ 451, 234, 46 },
		{ 451, 235, 46 },
		{ 451, 236, 46 },
		{ 451, 237, 46 },
		{ 451, 238, 46 },
		{ 451, 239, 46 },
		{ 451, 240, 46 },
		{ 451, 241, 46 },
		{ 451, 242, 46 },
		{ 451, 243, 46 },
		{ 451, 244, 46 },
		{ 451, 245, 46 },
		{ 451, 246, 46 },
		{ 451, 247, 46 },
		{ 451, 248, 46 },
		{ 451, 249, 46 },
		{ 451, 250, 46 },
		{ 451, 251, 46 },
		{ 451, 252, 46 },
		{ 451, 253, 46 },
		{ 451, 254, 46 },
		{ 451, 255, 46 },
		{ 451, 256, 46 },
		{ 451, 257, 46 },
		{ 451, 258, 46 },
		{ 451, 259, 46 },
		{ 451, 260, 46 },
		{ 451, 261, 46 },
		{ 451, 262, 46 },
		{ 451, 263, 46 },
		{ 451, 264, 46 },
		{ 451, 265, 46 },
		{ 451, 266, 46 },
		{ 451, 267, 46 },
		{ 451, 268, 46 },
		{ 451, 269, 46 },
		{ 451, 270, 46 },
		{ 451, 271, 46 },
		{ 451, 272, 46 },
		{ 451, 273, 46 },
		{ 451, 274, 46 },
		{ 451, 275, 46 },
		{ 451, 276, 46 },
		{ 451, 277, 46 },
		{ 451, 278, 46 },
		{ 451, 279, 46 },
		{ 451, 280, 46 },
		{ 451, 281, 46 },
		{ 452, 233, 46 },
		{ 452, 234, 46 },
		{ 452, 235, 46 },
		{ 452, 236, 46 },
		{ 452, 237, 46 },
		{ 452, 238, 46 },
		{ 452, 239, 46 },
		{ 452, 240, 46 },
		{ 452, 241, 46 },
		{ 452, 242, 46 },
		{ 452, 243, 46 },
		{ 452, 244, 46 },
		{ 452, 245, 46 },
		{ 452, 246, 46 },
		{ 452, 247, 46 },
		{ 452, 248, 46 },
		{ 452, 249, 46 },
		{ 452, 250, 46 },
		{ 452, 251, 46 },
		{ 452, 252, 46 },
		{ 452, 253, 46 },
		{ 452, 254, 46 },
		{ 452, 255, 46 },
		{ 452, 256, 46 },
		{ 452, 257, 46 },
		{ 452, 258, 46 },
		{ 452, 259, 46 },
		{ 452, 260, 46 },
		{ 452, 261, 46 },
		{ 452, 262, 46 },
		{ 452, 263, 46 },
		{ 452, 264, 46 },
		{ 452, 265, 46 },
		{ 452, 266, 46 },
		{ 452, 267, 46 },
		{ 452, 268, 46 },
		{ 452, 269, 46 },
		{ 452, 270, 46 },
		{ 452, 271, 46 },
		{ 452, 272, 46 },
		{ 452, 273, 46 },
		{ 452, 274, 46 },
		{ 452, 275, 46 },
		{ 452, 276, 46 },
		{ 452, 277, 46 },
		{ 452, 278, 46 },
		{ 452, 279, 46 },
		{ 452, 280, 46 },
		{ 452, 281, 46 },
		{ 453, 233, 46 },
		{ 453, 234, 46 },
		{ 453, 235, 46 },
		{ 453, 236, 46 },
		{ 453, 237, 46 },
		{ 453, 238, 46 },
		{ 453, 239, 46 },
		{ 453, 240, 46 },
		{ 453, 241, 46 },
		{ 453, 242, 46 },
		{ 453, 243, 46 },
		{ 453, 244, 46 },
		{ 453, 245, 46 },
		{ 453, 246, 46 },
		{ 453, 247, 46 },
		{ 453, 248, 46 },
		{ 453, 249, 46 },
		{ 453, 250, 46 },
		{ 453, 251, 46 },
		{ 453, 252, 46 },
		{ 453, 253, 46 },
		{ 453, 254, 46 },
		{ 453, 255, 46 },
		{ 453, 256, 46 },
		{ 453, 257, 46 },
		{ 453, 258, 46 },
		{ 453, 259, 46 },
		{ 453, 260, 46 },
		{ 453, 261, 46 },
		{ 453, 262, 46 },
		{ 453, 263, 46 },
		{ 453, 264, 46 },
		{ 453, 265, 46 },
		{ 453, 266, 46 },
		{ 453, 267, 46 },
		{ 453, 268, 46 },
		{ 453, 269, 46 },
		{ 453, 270, 46 },
		{ 453, 271, 46 },
		{ 453, 272, 46 },
		{ 453, 273, 46 },
		{ 453, 274, 46 },
		{ 453, 275, 46 },
		{ 453, 276, 46 },
		{ 453, 277, 46 },
		{ 453, 278, 46 },
		{ 453, 279, 46 },
		{ 453, 280, 46 },
		{ 453, 281, 46 },
		{ 454, 233, 46 },
		{ 454, 234, 46 },
		{ 454, 235, 46 },
		{ 454, 236, 46 },
		{ 454, 237, 46 },
		{ 454, 238, 46 },
		{ 454, 239, 46 },
		{ 454, 240, 46 },
		{ 454, 241, 46 },
		{ 454, 242, 46 },
		{ 454, 243, 46 },
		{ 454, 244, 46 },
		{ 454, 245, 46 },
		{ 454, 246, 46 },
		{ 454, 247, 46 },
		{ 454, 248, 46 },
		{ 454, 249, 46 },
		{ 454, 250, 46 },
		{ 454, 251, 46 },
		{ 454, 252, 46 },
		{ 454, 253, 46 },
		{ 454, 254, 46 },
		{ 454, 255, 46 },
		{ 454, 256, 46 },
		{ 454, 257, 46 },
		{ 454, 258, 46 },
		{ 454, 259, 46 },
		{ 454, 260, 46 },
		{ 454, 261, 46 },
		{ 454, 262, 46 },
		{ 454, 263, 46 },
		{ 454, 264, 46 },
		{ 454, 265, 46 },
		{ 454, 266, 46 },
		{ 454, 267, 46 },
		{ 454, 268, 46 },
		{ 454, 269, 46 },
		{ 454, 270, 46 },
		{ 454, 271, 46 },
		{ 454, 272, 46 },
		{ 454, 273, 46 },
		{ 454, 274, 46 },
		{ 454, 275, 46 },
		{ 454, 276, 46 },
		{ 454, 277, 46 },
		{ 454, 278, 46 },
		{ 454, 279, 46 },
		{ 454, 280, 46 },
		{ 454, 281, 46 },
		{ 455, 233, 46 },
		{ 455, 234, 46 },
		{ 455, 235, 46 },
		{ 455, 236, 46 },
		{ 455, 237, 46 },
		{ 455, 238, 46 },
		{ 455, 239, 46 },
		{ 455, 240, 46 },
		{ 455, 241, 46 },
		{ 455, 242, 46 },
		{ 455, 243, 46 },
		{ 455, 244, 46 },
		{ 455, 245, 46 },
		{ 455, 246, 46 },
		{ 455, 247, 46 },
		{ 455, 248, 46 },
		{ 455, 249, 46 },
		{ 455, 250, 46 },
		{ 455, 251, 46 },
		{ 455, 252, 46 },
		{ 455, 253, 46 },
		{ 455, 254, 46 },
		{ 455, 255, 46 },
		{ 455, 256, 46 },
		{ 455, 257, 46 },
		{ 455, 258, 46 },
		{ 455, 259, 46 },
		{ 455, 260, 46 },
		{ 455, 261, 46 },
		{ 455, 262, 46 },
		{ 455, 263, 46 },
		{ 455, 264, 46 },
		{ 455, 265, 46 },
		{ 455, 266, 46 },
		{ 455, 267, 46 },
		{ 455, 268, 46 },
		{ 455, 269, 46 },
		{ 455, 270, 46 },
		{ 455, 271, 46 },
		{ 455, 272, 46 },
		{ 455, 273, 46 },
		{ 455, 274, 46 },
		{ 455, 275, 46 },
		{ 455, 276, 46 },
		{ 455, 277, 46 },
		{ 455, 278, 46 },
		{ 455, 279, 46 },
		{ 455, 280, 46 },
		{ 455, 281, 46 },
		{ 456, 233, 46 },
		{ 456, 234, 46 },
		{ 456, 235, 46 },
		{ 456, 236, 46 },
		{ 456, 237, 46 },
		{ 456, 238, 46 },
		{ 456, 239, 46 },
		{ 456, 240, 46 },
		{ 456, 241, 46 },
		{ 456, 242, 46 },
		{ 456, 243, 46 },
		{ 456, 244, 46 },
		{ 456, 245, 46 },
		{ 456, 246, 46 },
		{ 456, 247, 46 },
		{ 456, 248, 46 },
		{ 456, 249, 46 },
		{ 456, 250, 46 },
		{ 456, 251, 46 },
		{ 456, 252, 46 },
		{ 456, 253, 46 },
		{ 456, 254, 46 },
		{ 456, 255, 46 },
		{ 456, 256, 46 },
		{ 456, 257, 46 },
		{ 456, 258, 46 },
		{ 456, 259, 46 },
		{ 456, 260, 46 },
		{ 456, 261, 46 },
		{ 456, 262, 46 },
		{ 456, 263, 46 },
		{ 456, 264, 46 },
		{ 456, 265, 46 },
		{ 456, 266, 46 },
		{ 456, 267, 46 },
		{ 456, 268, 46 },
		{ 456, 269, 46 },
		{ 456, 270, 46 },
		{ 456, 271, 46 },
		{ 456, 272, 46 },
		{ 456, 273, 46 },
		{ 456, 274, 46 },
		{ 456, 275, 46 },
		{ 456, 276, 46 },
		{ 456, 277, 46 },
		{ 456, 278, 46 },
		{ 456, 279, 46 },
		{ 456, 280, 46 },
		{ 456, 281, 46 },
		{ 457, 233, 46 },
		{ 457, 234, 46 },
		{ 457, 235, 46 },
		{ 457, 236, 46 },
		{ 457, 237, 46 },
		{ 457, 238, 46 },
		{ 457, 239, 46 },
		{ 457, 240, 46 },
		{ 457, 241, 46 },
		{ 457, 242, 46 },
		{ 457, 243, 46 },
		{ 457, 244, 46 },
		{ 457, 245, 46 },
		{ 457, 246, 46 },
		{ 457, 247, 46 },
		{ 457, 248, 46 },
		{ 457, 249, 46 },
		{ 457, 250, 46 },
		{ 457, 251, 46 },
		{ 457, 252, 46 },
		{ 457, 253, 46 },
		{ 457, 254, 46 },
		{ 457, 255, 46 },
		{ 457, 256, 46 },
		{ 457, 257, 46 },
		{ 457, 258, 46 },
		{ 457, 259, 46 },
		{ 457, 260, 46 },
		{ 457, 261, 46 },
		{ 457, 262, 46 },
		{ 457, 263, 46 },
		{ 457, 264, 46 },
		{ 457, 265, 46 },
		{ 457, 266, 46 },
		{ 457, 267, 46 },
		{ 457, 268, 46 },
		{ 457, 269, 46 },
		{ 457, 270, 46 },
		{ 457, 271, 46 },
		{ 457, 272, 46 },
		{ 457, 273, 46 },
		{ 457, 274, 46 },
		{ 457, 275, 46 },
		{ 457, 276, 46 },
		{ 457, 277, 46 },
		{ 457, 278, 46 },
		{ 457, 279, 46 },
		{ 457, 280, 46 },
		{ 457, 281, 46 },
		{ 458, 232, 46 },
		{ 458, 233, 46 },
		{ 458, 234, 46 },
		{ 458, 235, 46 },
		{ 458, 236, 46 },
		{ 458, 237, 46 },
		{ 458, 238, 46 },
		{ 458, 239, 46 },
		{ 458, 240, 46 },
		{ 458, 241, 46 },
		{ 458, 242, 46 },
		{ 458, 243, 46 },
		{ 458, 244, 46 },
		{ 458, 245, 46 },
		{ 458, 246, 46 },
		{ 458, 247, 46 },
		{ 458, 248, 46 },
		{ 458, 249, 46 },
		{ 458, 250, 46 },
		{ 458, 251, 46 },
		{ 458, 252, 46 },
		{ 458, 253, 46 },
		{ 458, 254, 46 },
		{ 458, 255, 46 },
		{ 458, 256, 46 },
		{ 458, 257, 46 },
		{ 458, 258, 46 },
		{ 458, 259, 46 },
		{ 458, 260, 46 },
		{ 458, 261, 46 },
		{ 458, 262, 46 },
		{ 458, 263, 46 },
		{ 458, 264, 46 },
		{ 458, 265, 46 },
		{ 458, 266, 46 },
		{ 458, 267, 46 },
		{ 458, 268, 46 },
		{ 458, 269, 46 },
		{ 458, 270, 46 },
		{ 458, 271, 46 },
		{ 458, 272, 46 },
		{ 458, 273, 46 },
		{ 458, 274, 46 },
		{ 458, 275, 46 },
		{ 458, 276, 46 },
		{ 458, 277, 46 },
		{ 458, 278, 46 },
		{ 458, 279, 46 },
		{ 458, 280, 46 },
		{ 458, 281, 46 },
		{ 458, 282, 46 },
		{ 459, 232, 46 },
		{ 459, 233, 46 },
		{ 459, 234, 46 },
		{ 459, 235, 46 },
		{ 459, 236, 46 },
		{ 459, 237, 46 },
		{ 459, 238, 46 },
		{ 459, 239, 46 },
		{ 459, 240, 46 },
		{ 459, 241, 46 },
		{ 459, 242, 46 },
		{ 459, 243, 46 },
		{ 459, 244, 46 },
		{ 459, 245, 46 },
		{ 459, 246, 46 },
		{ 459, 247, 46 },
		{ 459, 248, 46 },
		{ 459, 249, 46 },
		{ 459, 250, 46 },
		{ 459, 251, 46 },
		{ 459, 252, 46 },
		{ 459, 253, 46 },
		{ 459, 254, 46 },
		{ 459, 255, 46 },
		{ 459, 256, 46 },
		{ 459, 257, 46 },
		{ 459, 258, 46 },
		{ 459, 259, 46 },
		{ 459, 260, 46 },
		{ 459, 261, 46 },
		{ 459, 262, 46 },
		{ 459, 263, 46 },
		{ 459, 264, 46 },
		{ 459, 265, 46 },
		{ 459, 266, 46 },
		{ 459, 267, 46 },
		{ 459, 268, 46 },
		{ 459, 269, 46 },
		{ 459, 270, 46 },
		{ 459, 271, 46 },
		{ 459, 272, 46 },
		{ 459, 273, 46 },
		{ 459, 274, 46 },
		{ 459, 275, 46 },
		{ 459, 276, 46 },
		{ 459, 277, 46 },
		{ 459, 278, 46 },
		{ 459, 279, 46 },
		{ 459, 280, 46 },
		{ 459, 281, 46 },
		{ 459, 282, 46 },
		{ 460, 232, 46 },
		{ 460, 233, 46 },
		{ 460, 234, 46 },
		{ 460, 235, 46 },
		{ 460, 236, 46 },
		{ 460, 237, 46 },
		{ 460, 238, 46 },
		{ 460, 239, 46 },
		{ 460, 240, 46 },
		{ 460, 241, 46 },
		{ 460, 242, 46 },
		{ 460, 243, 46 },
		{ 460, 244, 46 },
		{ 460, 245, 46 },
		{ 460, 246, 46 },
		{ 460, 247, 46 },
		{ 460, 248, 46 },
		{ 460, 249, 46 },
		{ 460, 250, 46 },
		{ 460, 251, 46 },
		{ 460, 252, 46 },
		{ 460, 253, 46 },
		{ 460, 254, 46 },
		{ 460, 255, 46 },
		{ 460, 256, 46 },
		{ 460, 257, 46 },
		{ 460, 258, 46 },
		{ 460, 259, 46 },
		{ 460, 260, 46 },
		{ 460, 261, 46 },
		{ 460, 262, 46 },
		{ 460, 263, 46 },
		{ 460, 264, 46 },
		{ 460, 265, 46 },
		{ 460, 266, 46 },
		{ 460, 267, 46 },
		{ 460, 268, 46 },
		{ 460, 269, 46 },
		{ 460, 270, 46 },
		{ 460, 271, 46 },
		{ 460, 272, 46 },
		{ 460, 273, 46 },
		{ 460, 274, 46 },
		{ 460, 275, 46 },
		{ 460, 276, 46 },
		{ 460, 277, 46 },
		{ 460, 278, 46 },
		{ 460, 279, 46 },
		{ 460, 280, 46 },
		{ 460, 281, 46 },
		{ 460, 282, 46 },
		{ 461, 232, 46 },
		{ 461, 233, 46 },
		{ 461, 234, 46 },
		{ 461, 235, 46 },
		{ 461, 236, 46 },
		{ 461, 237, 46 },
		{ 461, 238, 46 },
		{ 461, 239, 46 },
		{ 461, 240, 46 },
		{ 461, 241, 46 },
		{ 461, 242, 46 },
		{ 461, 243, 46 },
		{ 461, 244, 46 },
		{ 461, 245, 46 },
		{ 461, 246, 46 },
		{ 461, 247, 46 },
		{ 461, 248, 46 },
		{ 461, 249, 46 },
		{ 461, 250, 46 },
		{ 461, 251, 46 },
		{ 461, 252, 46 },
		{ 461, 253, 46 },
		{ 461, 254, 46 },
		{ 461, 255, 46 },
		{ 461, 256, 46 },
		{ 461, 257, 46 },
		{ 461, 258, 46 },
		{ 461, 259, 46 },
		{ 461, 260, 46 },
		{ 461, 261, 46 },
		{ 461, 262, 46 },
		{ 461, 263, 46 },
		{ 461, 264, 46 },
		{ 461, 265, 46 },
		{ 461, 266, 46 },
		{ 461, 267, 46 },
		{ 461, 268, 46 },
		{ 461, 269, 46 },
		{ 461, 270, 46 },
		{ 461, 271, 46 },
		{ 461, 272, 46 },
		{ 461, 273, 46 },
		{ 461, 274, 46 },
		{ 461, 275, 46 },
		{ 461, 276, 46 },
		{ 461, 277, 46 },
		{ 461, 278, 46 },
		{ 461, 279, 46 },
		{ 461, 280, 46 },
		{ 461, 281, 46 },
		{ 461, 282, 46 },
		{ 462, 232, 46 },
		{ 462, 233, 46 },
		{ 462, 234, 46 },
		{ 462, 235, 46 },
		{ 462, 236, 46 },
		{ 462, 237, 46 },
		{ 462, 238, 46 },
		{ 462, 239, 46 },
		{ 462, 240, 46 },
		{ 462, 241, 46 },
		{ 462, 242, 46 },
		{ 462, 243, 46 },
		{ 462, 244, 46 },
		{ 462, 245, 46 },
		{ 462, 246, 46 },
		{ 462, 247, 46 },
		{ 462, 248, 46 },
		{ 462, 249, 46 },
		{ 462, 250, 46 },
		{ 462, 251, 46 },
		{ 462, 252, 46 },
		{ 462, 253, 46 },
		{ 462, 254, 46 },
		{ 462, 255, 46 },
		{ 462, 256, 46 },
		{ 462, 257, 46 },
		{ 462, 258, 46 },
		{ 462, 259, 46 },
		{ 462, 260, 46 },
		{ 462, 261, 46 },
		{ 462, 262, 46 },
		{ 462, 263, 46 },
		{ 462, 264, 46 },
		{ 462, 265, 46 },
		{ 462, 266, 46 },
		{ 462, 267, 46 },
		{ 462, 268, 46 },
		{ 462, 269, 46 },
		{ 462, 270, 46 },
		{ 462, 271, 46 },
		{ 462, 272, 46 },
		{ 462, 273, 46 },
		{ 462, 274, 46 },
		{ 462, 275, 46 },
		{ 462, 276, 46 },
		{ 462, 277, 46 },
		{ 462, 278, 46 },
		{ 462, 279, 46 },
		{ 462, 280, 46 },
		{ 462, 281, 46 },
		{ 462, 282, 46 },
		{ 463, 232, 46 },
		{ 463, 233, 46 },
		{ 463, 234, 46 },
		{ 463, 235, 46 },
		{ 463, 236, 46 },
		{ 463, 237, 46 },
		{ 463, 238, 46 },
		{ 463, 239, 46 },
		{ 463, 240, 46 },
		{ 463, 241, 46 },
		{ 463, 242, 46 },
		{ 463, 243, 46 },
		{ 463, 244, 46 },
		{ 463, 245, 46 },
		{ 463, 246, 46 },
		{ 463, 247, 46 },
		{ 463, 248, 46 },
		{ 463, 249, 46 },
		{ 463, 250, 46 },
		{ 463, 251, 46 },
		{ 463, 252, 46 },
		{ 463, 253, 46 },
		{ 463, 254, 46 },
		{ 463, 255, 46 },
		{ 463, 256, 46 },
		{ 463, 257, 46 },
		{ 463, 258, 46 },
		{ 463, 259, 46 },
		{ 463, 260, 46 },
		{ 463, 261, 46 },
		{ 463, 262, 46 },
		{ 463, 263, 46 },
		{ 463, 264, 46 },
		{ 463, 265, 46 },
		{ 463, 266, 46 },
		{ 463, 267, 46 },
		{ 463, 268, 46 },
		{ 463, 269, 46 },
		{ 463, 270, 46 },
		{ 463, 271, 46 },
		{ 463, 272, 46 },
		{ 463, 273, 46 },
		{ 463, 274, 46 },
		{ 463, 275, 46 },
		{ 463, 276, 46 },
		{ 463, 277, 46 },
		{ 463, 278, 46 },
		{ 463, 279, 46 },
		{ 463, 280, 46 },
		{ 463, 281, 46 },
		{ 463, 282, 46 },
		{ 464, 232, 46 },
		{ 464, 233, 46 },
		{ 464, 234, 46 },
		{ 464, 235, 46 },
		{ 464, 236, 46 },
		{ 464, 237, 46 },
		{ 464, 238, 46 },
		{ 464, 239, 46 },
		{ 464, 240, 46 },
		{ 464, 241, 46 },
		{ 464, 242, 46 },
		{ 464, 243, 46 },
		{ 464, 244, 46 },
		{ 464, 245, 46 },
		{ 464, 246, 46 },
		{ 464, 247, 46 },
		{ 464, 248, 46 },
		{ 464, 249, 46 },
		{ 464, 250, 46 },
		{ 464, 251, 46 },
		{ 464, 252, 46 },
		{ 464, 253, 46 },
		{ 464, 254, 46 },
		{ 464, 255, 46 },
		{ 464, 256, 46 },
		{ 464, 257, 46 },
		{ 464, 258, 46 },
		{ 464, 259, 46 },
		{ 464, 260, 46 },
		{ 464, 261, 46 },
		{ 464, 262, 46 },
		{ 464, 263, 46 },
		{ 464, 264, 46 },
		{ 464, 265, 46 },
		{ 464, 266, 46 },
		{ 464, 267, 46 },
		{ 464, 268, 46 },
		{ 464, 269, 46 },
		{ 464, 270, 46 },
		{ 464, 271, 46 },
		{ 464, 272, 46 },
		{ 464, 273, 46 },
		{ 464, 274, 46 },
		{ 464, 275, 46 },
		{ 464, 276, 46 },
		{ 464, 277, 46 },
		{ 464, 278, 46 },
		{ 464, 279, 46 },
		{ 464, 280, 46 },
		{ 464, 281, 46 },
		{ 464, 282, 46 },
		{ 465, 232, 46 },
		{ 465, 233, 46 },
		{ 465, 234, 46 },
		{ 465, 235, 46 },
		{ 465, 236, 46 },
		{ 465, 237, 46 },
		{ 465, 238, 46 },
		{ 465, 239, 46 },
		{ 465, 240, 46 },
		{ 465, 241, 46 },
		{ 465, 242, 46 },
		{ 465, 243, 46 },
		{ 465, 244, 46 },
		{ 465, 245, 46 },
		{ 465, 246, 46 },
		{ 465, 247, 46 },
		{ 465, 248, 46 },
		{ 465, 249, 46 },
		{ 465, 250, 46 },
		{ 465, 251, 46 },
		{ 465, 252, 46 },
		{ 465, 253, 46 },
		{ 465, 254, 46 },
		{ 465, 255, 46 },
		{ 465, 256, 46 },
		{ 465, 257, 46 },
		{ 465, 258, 46 },
		{ 465, 259, 46 },
		{ 465, 260, 46 },
		{ 465, 261, 46 },
		{ 465, 262, 46 },
		{ 465, 263, 46 },
		{ 465, 264, 46 },
		{ 465, 265, 46 },
		{ 465, 266, 46 },
		{ 465, 267, 46 },
		{ 465, 268, 46 },
		{ 465, 269, 46 },
		{ 465, 270, 46 },
		{ 465, 271, 46 },
		{ 465, 272, 46 },
		{ 465, 273, 46 },
		{ 465, 274, 46 },
		{ 465, 275, 46 },
		{ 465, 276, 46 },
		{ 465, 277, 46 },
		{ 465, 278, 46 },
		{ 465, 279, 46 },
		{ 465, 280, 46 },
		{ 465, 281, 46 },
		{ 465, 282, 46 },
		{ 466, 231, 46 },
		{ 466, 232, 46 },
		{ 466, 233, 46 },
		{ 466, 234, 46 },
		{ 466, 235, 46 },
		{ 466, 236, 46 },
		{ 466, 237, 46 },
		{ 466, 238, 46 },
		{ 466, 239, 46 },
		{ 466, 240, 46 },
		{ 466, 241, 46 },
		{ 466, 242, 46 },
		{ 466, 243, 46 },
		{ 466, 244, 46 },
		{ 466, 245, 46 },
		{ 466, 246, 46 },
		{ 466, 247, 46 },
		{ 466, 248, 46 },
		{ 466, 249, 46 },
		{ 466, 250, 46 },
		{ 466, 251, 46 },
		{ 466, 252, 46 },
		{ 466, 253, 46 },
		{ 466, 254, 46 },
		{ 466, 255, 46 },
		{ 466, 256, 46 },
		{ 466, 257, 46 },
		{ 466, 258, 46 },
		{ 466, 259, 46 },
		{ 466, 260, 46 },
		{ 466, 261, 46 },
		{ 466, 262, 46 },
		{ 466, 263, 46 },
		{ 466, 264, 46 },
		{ 466, 265, 46 },
		{ 466, 266, 46 },
		{ 466, 267, 46 },
		{ 466, 268, 46 },
		{ 466, 269, 46 },
		{ 466, 270, 46 },
		{ 466, 271, 46 },
		{ 466, 272, 46 },
		{ 466, 273, 46 },
		{ 466, 274, 46 },
		{ 466, 275, 46 },
		{ 466, 276, 46 },
		{ 466, 277, 46 },
		{ 466, 278, 46 },
		{ 466, 279, 46 },
		{ 466, 280, 46 },
		{ 466, 281, 46 },
		{ 466, 282, 46 },
		{ 466, 283, 46 },
		{ 467, 231, 46 },
		{ 467, 232, 46 },
		{ 467, 233, 46 },
		{ 467, 234, 46 },
		{ 467, 235, 46 },
		{ 467, 236, 46 },
		{ 467, 237, 46 },
		{ 467, 238, 46 },
		{ 467, 239, 46 },
		{ 467, 240, 46 },
		{ 467, 241, 46 },
		{ 467, 242, 46 },
		{ 467, 243, 46 },
		{ 467, 244, 46 },
		{ 467, 245, 46 },
		{ 467, 246, 46 },
		{ 467, 247, 46 },
		{ 467, 248, 46 },
		{ 467, 249, 46 },
		{ 467, 250, 46 },
		{ 467, 251, 46 },
		{ 467, 252, 46 },
		{ 467, 253, 46 },
		{ 467, 254, 46 },
		{ 467, 255, 46 },
		{ 467, 256, 46 },
		{ 467, 257, 46 },
		{ 467, 258, 46 },
		{ 467, 259, 46 },
		{ 467, 260, 46 },
		{ 467, 261, 46 },
		{ 467, 262, 46 },
		{ 467, 263, 46 },
		{ 467, 264, 46 },
		{ 467, 265, 46 },
		{ 467, 266, 46 },
		{ 467, 267, 46 },
		{ 467, 268, 46 },
		{ 467, 269, 46 },
		{ 467, 270, 46 },
		{ 467, 271, 46 },
		{ 467, 272, 46 },
		{ 467, 273, 46 },
		{ 467, 274, 46 },
		{ 467, 275, 46 },
		{ 467, 276, 46 },
		{ 467, 277, 46 },
		{ 467, 278, 46 },
		{ 467, 279, 46 },
		{ 467, 280, 46 },
		{ 467, 281, 46 },
		{ 467, 282, 46 },
		{ 467, 283, 46 },
		{ 468, 231, 46 },
		{ 468, 232, 46 },
		{ 468, 233, 46 },
		{ 468, 234, 46 },
		{ 468, 235, 46 },
		{ 468, 236, 46 },
		{ 468, 237, 46 },
		{ 468, 238, 46 },
		{ 468, 239, 46 },
		{ 468, 240, 46 },
		{ 468, 241, 46 },
		{ 468, 242, 46 },
		{ 468, 243, 46 },
		{ 468, 244, 46 },
		{ 468, 245, 46 },
		{ 468, 246, 46 },
		{ 468, 247, 46 },
		{ 468, 248, 46 },
		{ 468, 249, 46 },
		{ 468, 250, 46 },
		{ 468, 251, 46 },
		{ 468, 252, 46 },
		{ 468, 253, 46 },
		{ 468, 254, 46 },
		{ 468, 255, 46 },
		{ 468, 256, 46 },
		{ 468, 257, 46 },
		{ 468, 258, 46 },
		{ 468, 259, 46 },
		{ 468, 260, 46 },
		{ 468, 261, 46 },
		{ 468, 262, 46 },
		{ 468, 263, 46 },
		{ 468, 264, 46 },
		{ 468, 265, 46 },
		{ 468, 266, 46 },
		{ 468, 267, 46 },
		{ 468, 268, 46 },
		{ 468, 269, 46 },
		{ 468, 270, 46 },
		{ 468, 271, 46 },
		{ 468, 272, 46 },
		{ 468, 273, 46 },
		{ 468, 274, 46 },
		{ 468, 275, 46 },
		{ 468, 276, 46 },
		{ 468, 277, 46 },
		{ 468, 278, 46 },
		{ 468, 279, 46 },
		{ 468, 280, 46 },
		{ 468, 281, 46 },
		{ 468, 282, 46 },
		{ 468, 283, 46 },
		{ 469, 231, 46 },
		{ 469, 232, 46 },
		{ 469, 233, 46 },
		{ 469, 234, 46 },
		{ 469, 235, 46 },
		{ 469, 236, 46 },
		{ 469, 237, 46 },
		{ 469, 238, 46 },
		{ 469, 239, 46 },
		{ 469, 240, 46 },
		{ 469, 241, 46 },
		{ 469, 242, 46 },
		{ 469, 243, 46 },
		{ 469, 244, 46 },
		{ 469, 245, 46 },
		{ 469, 246, 46 },
		{ 469, 247, 46 },
		{ 469, 248, 46 },
		{ 469, 249, 46 },
		{ 469, 250, 46 },
		{ 469, 251, 46 },
		{ 469, 252, 46 },
		{ 469, 253, 46 },
		{ 469, 254, 46 },
		{ 469, 255, 46 },
		{ 469, 256, 46 },
		{ 469, 257, 46 },
		{ 469, 258, 46 },
		{ 469, 259, 46 },
		{ 469, 260, 46 },
		{ 469, 261, 46 },
		{ 469, 262, 46 },
		{ 469, 263, 46 },
		{ 469, 264, 46 },
		{ 469, 265, 46 },
		{ 469, 266, 46 },
		{ 469, 267, 46 },
		{ 469, 268, 46 },
		{ 469, 269, 46 },
		{ 469, 270, 46 },
		{ 469, 271, 46 },
		{ 469, 272, 46 },
		{ 469, 273, 46 },
		{ 469, 274, 46 },
		{ 469, 275, 46 },
		{ 469, 276, 46 },
		{ 469, 277, 46 },
		{ 469, 278, 46 },
		{ 469, 279, 46 },
		{ 469, 280, 46 },
		{ 469, 281, 46 },
		{ 469, 282, 46 },
		{ 469, 283, 46 },
		{ 470, 231, 46 },
		{ 470, 232, 46 },
		{ 470, 233, 46 },
		{ 470, 234, 46 },
		{ 470, 235, 46 },
		{ 470, 236, 46 },
		{ 470, 237, 46 },
		{ 470, 238, 46 },
		{ 470, 239, 46 },
		{ 470, 240, 46 },
		{ 470, 241, 46 },
		{ 470, 242, 46 },
		{ 470, 243, 46 },
		{ 470, 244, 46 },
		{ 470, 245, 46 },
		{ 470, 246, 46 },
		{ 470, 247, 46 },
		{ 470, 248, 46 },
		{ 470, 249, 46 },
		{ 470, 250, 46 },
		{ 470, 251, 46 },
		{ 470, 252, 46 },
		{ 470, 253, 46 },
		{ 470, 254, 46 },
		{ 470, 255, 46 },
		{ 470, 256, 46 },
		{ 470, 257, 46 },
		{ 470, 258, 46 },
		{ 470, 259, 46 },
		{ 470, 260, 46 },
		{ 470, 261, 46 },
		{ 470, 262, 46 },
		{ 470, 263, 46 },
		{ 470, 264, 46 },
		{ 470, 265, 46 },
		{ 470, 266, 46 },
		{ 470, 267, 46 },
		{ 470, 268, 46 },
		{ 470, 269, 46 },
		{ 470, 270, 46 },
		{ 470, 271, 46 },
		{ 470, 272, 46 },
		{ 470, 273, 46 },
		{ 470, 274, 46 },
		{ 470, 275, 46 },
		{ 470, 276, 46 },
		{ 470, 277, 46 },
		{ 470, 278, 46 },
		{ 470, 279, 46 },
		{ 470, 280, 46 },
		{ 470, 281, 46 },
		{ 470, 282, 46 },
		{ 470, 283, 46 },
		{ 471, 231, 46 },
		{ 471, 232, 46 },
		{ 471, 233, 46 },
		{ 471, 234, 46 },
		{ 471, 235, 46 },
		{ 471, 236, 46 },
		{ 471, 237, 46 },
		{ 471, 238, 46 },
		{ 471, 239, 46 },
		{ 471, 240, 46 },
		{ 471, 241, 46 },
		{ 471, 242, 46 },
		{ 471, 243, 46 },
		{ 471, 244, 46 },
		{ 471, 245, 46 },
		{ 471, 246, 46 },
		{ 471, 247, 46 },
		{ 471, 248, 46 },
		{ 471, 249, 46 },
		{ 471, 250, 46 },
		{ 471, 251, 46 },
		{ 471, 252, 46 },
		{ 471, 253, 46 },
		{ 471, 254, 46 },
		{ 471, 255, 46 },
		{ 471, 256, 46 },
		{ 471, 257, 46 },
		{ 471, 258, 46 },
		{ 471, 259, 46 },
		{ 471, 260, 46 },
		{ 471, 261, 46 },
		{ 471, 262, 46 },
		{ 471, 263, 46 },
		{ 471, 264, 46 },
		{ 471, 265, 46 },
		{ 471, 266, 46 },
		{ 471, 267, 46 },
		{ 471, 268, 46 },
		{ 471, 269, 46 },
		{ 471, 270, 46 },
		{ 471, 271, 46 },
		{ 471, 272, 46 },
		{ 471, 273, 46 },
		{ 471, 274, 46 },
		{ 471, 275, 46 },
		{ 471, 276, 46 },
		{ 471, 277, 46 },
		{ 471, 278, 46 },
		{ 471, 279, 46 },
		{ 471, 280, 46 },
		{ 471, 281, 46 },
		{ 471, 282, 46 },
		{ 471, 283, 46 },
		{ 472, 231, 46 },
		{ 472, 232, 46 },
		{ 472, 233, 46 },
		{ 472, 234, 46 },
		{ 472, 235, 46 },
		{ 472, 236, 46 },
		{ 472, 237, 46 },
		{ 472, 238, 46 },
		{ 472, 239, 46 },
		{ 472, 240, 46 },
		{ 472, 241, 46 },
		{ 472, 242, 46 },
		{ 472, 243, 46 },
		{ 472, 244, 46 },
		{ 472, 245, 46 },
		{ 472, 246, 46 },
		{ 472, 247, 46 },
		{ 472, 248, 46 },
		{ 472, 249, 46 },
		{ 472, 250, 46 },
		{ 472, 251, 46 },
		{ 472, 252, 46 },
		{ 472, 253, 46 },
		{ 472, 254, 46 },
		{ 472, 255, 46 },
		{ 472, 256, 46 },
		{ 472, 257, 46 },
		{ 472, 258, 46 },
		{ 472, 259, 46 },
		{ 472, 260, 46 },
		{ 472, 261, 46 },
		{ 472, 262, 46 },
		{ 472, 263, 46 },
		{ 472, 264, 46 },
		{ 472, 265, 46 },
		{ 472, 266, 46 },
		{ 472, 267, 46 },
		{ 472, 268, 46 },
		{ 472, 269, 46 },
		{ 472, 270, 46 },
		{ 472, 271, 46 },
		{ 472, 272, 46 },
		{ 472, 273, 46 },
		{ 472, 274, 46 },
		{ 472, 275, 46 },
		{ 472, 276, 46 },
		{ 472, 277, 46 },
		{ 472, 278, 46 },
		{ 472, 279, 46 },
		{ 472, 280, 46 },
		{ 472, 281, 46 },
		{ 472, 282, 46 },
		{ 472, 283, 46 },
		{ 473, 231, 46 },
		{ 473, 232, 46 },
		{ 473, 233, 46 },
		{ 473, 234, 46 },
		{ 473, 235, 46 },
		{ 473, 236, 46 },
		{ 473, 237, 46 },
		{ 473, 238, 46 },
		{ 473, 239, 46 },
		{ 473, 240, 46 },
		{ 473, 241, 46 },
		{ 473, 242, 46 },
		{ 473, 243, 46 },
		{ 473, 244, 46 },
		{ 473, 245, 46 },
		{ 473, 246, 46 },
		{ 473, 247, 46 },
		{ 473, 248, 46 },
		{ 473, 249, 46 },
		{ 473, 250, 46 },
		{ 473, 251, 46 },
		{ 473, 252, 46 },
		{ 473, 253, 46 },
		{ 473, 254, 46 },
		{ 473, 255, 46 },
		{ 473, 256, 46 },
		{ 473, 257, 46 },
		{ 473, 258, 46 },
		{ 473, 259, 46 },
		{ 473, 260, 46 },
		{ 473, 261, 46 },
		{ 473, 262, 46 },
		{ 473, 263, 46 },
		{ 473, 264, 46 },
		{ 473, 265, 46 },
		{ 473, 266, 46 },
		{ 473, 267, 46 },
		{ 473, 268, 46 },
		{ 473, 269, 46 },
		{ 473, 270, 46 },
		{ 473, 271, 46 },
		{ 473, 272, 46 },
		{ 473, 273, 46 },
		{ 473, 274, 46 },
		{ 473, 275, 46 },
		{ 473, 276, 46 },
		{ 473, 277, 46 },
		{ 473, 278, 46 },
		{ 473, 279, 46 },
		{ 473, 280, 46 },
		{ 473, 281, 46 },
		{ 473, 282, 46 },
		{ 473, 283, 46 },
		{ 474, 230, 46 },
		{ 474, 231, 46 },
		{ 474, 232, 46 },
		{ 474, 233, 46 },
		{ 474, 234, 46 },
		{ 474, 235, 46 },
		{ 474, 236, 46 },
		{ 474, 237, 46 },
		{ 474, 238, 46 },
		{ 474, 239, 46 },
		{ 474, 240, 46 },
		{ 474, 241, 46 },
		{ 474, 242, 46 },
		{ 474, 243, 46 },
		{ 474, 244, 46 },
		{ 474, 245, 46 },
		{ 474, 246, 46 },
		{ 474, 247, 46 },
		{ 474, 248, 46 },
		{ 474, 249, 46 },
		{ 474, 250, 46 },
		{ 474, 251, 46 },
		{ 474, 252, 46 },
		{ 474, 253, 46 },
		{ 474, 254, 46 },
		{ 474, 255, 46 },
		{ 474, 256, 46 },
		{ 474, 257, 46 },
		{ 474, 258, 46 },
		{ 474, 259, 46 },
		{ 474, 260, 46 },
		{ 474, 261, 46 },
		{ 474, 262, 46 },
		{ 474, 263, 46 },
		{ 474, 264, 46 },
		{ 474, 265, 46 },
		{ 474, 266, 46 },
		{ 474, 267, 46 },
		{ 474, 268, 46 },
		{ 474, 269, 46 },
		{ 474, 270, 46 },
		{ 474, 271, 46 },
		{ 474, 272, 46 },
		{ 474, 273, 46 },
		{ 474, 274, 46 },
		{ 474, 275, 46 },
		{ 474, 276, 46 },
		{ 474, 277, 46 },
		{ 474, 278, 46 },
		{ 474, 279, 46 },
		{ 474, 280, 46 },
		{ 474, 281, 46 },
		{ 474, 282, 46 },
		{ 474, 283, 46 },
		{ 474, 284, 46 },
		{ 475, 230, 46 },
		{ 475, 231, 46 },
		{ 475, 232, 46 },
		{ 475, 233, 46 },
		{ 475, 234, 46 },
		{ 475, 235, 46 },
		{ 475, 236, 46 },
		{ 475, 237, 46 },
		{ 475, 238, 46 },
		{ 475, 239, 46 },
		{ 475, 240, 46 },
		{ 475, 241, 46 },
		{ 475, 242, 46 },
		{ 475, 243, 46 },
		{ 475, 244, 46 },
		{ 475, 245, 46 },
		{ 475, 246, 46 },
		{ 475, 247, 46 },
		{ 475, 248, 46 },
		{ 475, 249, 46 },
		{ 475, 250, 46 },
		{ 475, 251, 46 },
		{ 475, 252, 46 },
		{ 475, 253, 46 },
		{ 475, 254, 46 },
		{ 475, 255, 46 },
		{ 475, 256, 46 },
		{ 475, 257, 46 },
		{ 475, 258, 46 },
		{ 475, 259, 46 },
		{ 475, 260, 46 },
		{ 475, 261, 46 },
		{ 475, 262, 46 },
		{ 475, 263, 46 },
		{ 475, 264, 46 },
		{ 475, 265, 46 },
		{ 475, 266, 46 },
		{ 475, 267, 46 },
		{ 475, 268, 46 },
		{ 475, 269, 46 },
		{ 475, 270, 46 },
		{ 475, 271, 46 },
		{ 475, 272, 46 },
		{ 475, 273, 46 },
		{ 475, 274, 46 },
		{ 475, 275, 46 },
		{ 475, 276, 46 },
		{ 475, 277, 46 },
		{ 475, 278, 46 },
		{ 475, 279, 46 },
		{ 475, 280, 46 },
		{ 475, 281, 46 },
		{ 475, 282, 46 },
		{ 475, 283, 46 },
		{ 475, 284, 46 },
		{ 476, 230, 46 },
		{ 476, 231, 46 },
		{ 476, 232, 46 },
		{ 476, 233, 46 },
		{ 476, 234, 46 },
		{ 476, 235, 46 },
		{ 476, 236, 46 },
		{ 476, 237, 46 },
		{ 476, 238, 46 },
		{ 476, 239, 46 },
		{ 476, 240, 46 },
		{ 476, 241, 46 },
		{ 476, 242, 46 },
		{ 476, 243, 46 },
		{ 476, 244, 46 },
		{ 476, 245, 46 },
		{ 476, 246, 46 },
		{ 476, 247, 46 },
		{ 476, 248, 46 },
		{ 476, 249, 46 },
		{ 476, 250, 46 },
		{ 476, 251, 46 },
		{ 476, 252, 46 },
		{ 476, 253, 46 },
		{ 476, 254, 46 },
		{ 476, 255, 46 },
		{ 476, 256, 46 },
		{ 476, 257, 46 },
		{ 476, 258, 46 },
		{ 476, 259, 46 },
		{ 476, 260, 46 },
		{ 476, 261, 46 },
		{ 476, 262, 46 },
		{ 476, 263, 46 },
		{ 476, 264, 46 },
		{ 476, 265, 46 },
		{ 476, 266, 46 },
		{ 476, 267, 46 },
		{ 476, 268, 46 },
		{ 476, 269, 46 },
		{ 476, 270, 46 },
		{ 476, 271, 46 },
		{ 476, 272, 46 },
		{ 476, 273, 46 },
		{ 476, 274, 46 },
		{ 476, 275, 46 },
		{ 476, 276, 46 },
		{ 476, 277, 46 },
		{ 476, 278, 46 },
		{ 476, 279, 46 },
		{ 476, 280, 46 },
		{ 476, 281, 46 },
		{ 476, 282, 46 },
		{ 476, 283, 46 },
		{ 476, 284, 46 },
		{ 477, 230, 46 },
		{ 477, 231, 46 },
		{ 477, 232, 46 },
		{ 477, 233, 46 },
		{ 477, 234, 46 },
		{ 477, 235, 46 },
		{ 477, 236, 46 },
		{ 477, 237, 46 },
		{ 477, 238, 46 },
		{ 477, 239, 46 },
		{ 477, 240, 46 },
		{ 477, 241, 46 },
		{ 477, 242, 46 },
		{ 477, 243, 46 },
		{ 477, 244, 46 },
		{ 477, 245, 46 },
		{ 477, 246, 46 },
		{ 477, 247, 46 },
		{ 477, 248, 46 },
		{ 477, 249, 46 },
		{ 477, 250, 46 },
		{ 477, 251, 46 },
		{ 477, 252, 46 },
		{ 477, 253, 46 },
		{ 477, 254, 46 },
		{ 477, 255, 46 },
		{ 477, 256, 46 },
		{ 477, 257, 46 },
		{ 477, 258, 46 },
		{ 477, 259, 46 },
		{ 477, 260, 46 },
		{ 477, 261, 46 },
		{ 477, 262, 46 },
		{ 477, 263, 46 },
		{ 477, 264, 46 },
		{ 477, 265, 46 },
		{ 477, 266, 46 },
		{ 477, 267, 46 },
		{ 477, 268, 46 },
		{ 477, 269, 46 },
		{ 477, 270, 46 },
		{ 477, 271, 46 },
		{ 477, 272, 46 },
		{ 477, 273, 46 },
		{ 477, 274, 46 },
		{ 477, 275, 46 },
		{ 477, 276, 46 },
		{ 477, 277, 46 },
		{ 477, 278, 46 },
		{ 477, 279, 46 },
		{ 477, 280, 46 },
		{ 477, 281, 46 },
		{ 477, 282, 46 },
		{ 477, 283, 46 },
		{ 477, 284, 46 },
		{ 478, 230, 46 },
		{ 478, 231, 46 },
		{ 478, 232, 46 },
		{ 478, 233, 46 },
		{ 478, 234, 46 },
		{ 478, 235, 46 },
		{ 478, 236, 46 },
		{ 478, 237, 46 },
		{ 478, 238, 46 },
		{ 478, 239, 46 },
		{ 478, 240, 46 },
		{ 478, 241, 46 },
		{ 478, 242, 46 },
		{ 478, 243, 46 },
		{ 478, 244, 46 },
		{ 478, 245, 46 },
		{ 478, 246, 46 },
		{ 478, 247, 46 },
		{ 478, 248, 46 },
		{ 478, 249, 46 },
		{ 478, 250, 46 },
		{ 478, 251, 46 },
		{ 478, 252, 46 },
		{ 478, 253, 46 },
		{ 478, 254, 46 },
		{ 478, 255, 46 },
		{ 478, 256, 46 },
		{ 478, 257, 46 },
		{ 478, 258, 46 },
		{ 478, 259, 46 },
		{ 478, 260, 46 },
		{ 478, 261, 46 },
		{ 478, 262, 46 },
		{ 478, 263, 46 },
		{ 478, 264, 46 },
		{ 478, 265, 46 },
		{ 478, 266, 46 },
		{ 478, 267, 46 },
		{ 478, 268, 46 },
		{ 478, 269, 46 },
		{ 478, 270, 46 },
		{ 478, 271, 46 },
		{ 478, 272, 46 },
		{ 478, 273, 46 },
		{ 478, 274, 46 },
		{ 478, 275, 46 },
		{ 478, 276, 46 },
		{ 478, 277, 46 },
		{ 478, 278, 46 },
		{ 478, 279, 46 },
		{ 478, 280, 46 },
		{ 478, 281, 46 },
		{ 478, 282, 46 },
		{ 478, 283, 46 },
		{ 478, 284, 46 },
		{ 479, 230, 46 },
		{ 479, 231, 46 },
		{ 479, 232, 46 },
		{ 479, 233, 46 },
		{ 479, 234, 46 },
		{ 479, 235, 46 },
		{ 479, 236, 46 },
		{ 479, 237, 46 },
		{ 479, 238, 46 },
		{ 479, 239, 46 },
		{ 479, 240, 46 },
		{ 479, 241, 46 },
		{ 479, 242, 46 },
		{ 479, 243, 46 },
		{ 479, 244, 46 },
		{ 479, 245, 46 },
		{ 479, 246, 46 },
		{ 479, 247, 46 },
		{ 479, 248, 46 },
		{ 479, 249, 46 },
		{ 479, 250, 46 },
		{ 479, 251, 46 },
		{ 479, 252, 46 },
		{ 479, 253, 46 },
		{ 479, 254, 46 },
		{ 479, 255, 46 },
		{ 479, 256, 46 },
		{ 479, 257, 46 },
		{ 479, 258, 46 },
		{ 479, 259, 46 },
		{ 479, 260, 46 },
		{ 479, 261, 46 },
		{ 479, 262, 46 },
		{ 479, 263, 46 },
		{ 479, 264, 46 },
		{ 479, 265, 46 },
		{ 479, 266, 46 },
		{ 479, 267, 46 },
		{ 479, 268, 46 },
		{ 479, 269, 46 },
		{ 479, 270, 46 },
		{ 479, 271, 46 },
		{ 479, 272, 46 },
		{ 479, 273, 46 },
		{ 479, 274, 46 },
		{ 479, 275, 46 },
		{ 479, 276, 46 },
		{ 479, 277, 46 },
		{ 479, 278, 46 },
		{ 479, 279, 46 },
		{ 479, 280, 46 },
		{ 479, 281, 46 },
		{ 479, 282, 46 },
		{ 479, 283, 46 },
		{ 479, 284, 46 },
		{ 480, 230, 46 },
		{ 480, 231, 46 },
		{ 480, 232, 46 },
		{ 480, 233, 46 },
		{ 480, 234, 46 },
		{ 480, 235, 46 },
		{ 480, 236, 46 },
		{ 480, 237, 46 },
		{ 480, 238, 46 },
		{ 480, 239, 46 },
		{ 480, 240, 46 },
		{ 480, 241, 46 },
		{ 480, 242, 46 },
		{ 480, 243, 46 },
		{ 480, 244, 46 },
		{ 480, 245, 46 },
		{ 480, 246, 46 },
		{ 480, 247, 46 },
		{ 480, 248, 46 },
		{ 480, 249, 46 },
		{ 480, 250, 46 },
		{ 480, 251, 46 },
		{ 480, 252, 46 },
		{ 480, 253, 46 },
		{ 480, 254, 46 },
		{ 480, 255, 46 },
		{ 480, 256, 46 },
		{ 480, 257, 46 },
		{ 480, 258, 46 },
		{ 480, 259, 46 },
		{ 480, 260, 46 },
		{ 480, 261, 46 },
		{ 480, 262, 46 },
		{ 480, 263, 46 },
		{ 480, 264, 46 },
		{ 480, 265, 46 },
		{ 480, 266, 46 },
		{ 480, 267, 46 },
		{ 480, 268, 46 },
		{ 480, 269, 46 },
		{ 480, 270, 46 },
		{ 480, 271, 46 },
		{ 480, 272, 46 },
		{ 480, 273, 46 },
		{ 480, 274, 46 },
		{ 480, 275, 46 },
		{ 480, 276, 46 },
		{ 480, 277, 46 },
		{ 480, 278, 46 },
		{ 480, 279, 46 },
		{ 480, 280, 46 },
		{ 480, 281, 46 },
		{ 480, 282, 46 },
		{ 480, 283, 46 },
		{ 480, 284, 46 },
		{ 481, 230, 46 },
		{ 481, 231, 46 },
		{ 481, 232, 46 },
		{ 481, 233, 46 },
		{ 481, 234, 46 },
		{ 481, 235, 46 },
		{ 481, 236, 46 },
		{ 481, 237, 46 },
		{ 481, 238, 46 },
		{ 481, 239, 46 },
		{ 481, 240, 46 },
		{ 481, 241, 46 },
		{ 481, 242, 46 },
		{ 481, 243, 46 },
		{ 481, 244, 46 },
		{ 481, 245, 46 },
		{ 481, 246, 46 },
		{ 481, 247, 46 },
		{ 481, 248, 46 },
		{ 481, 249, 46 },
		{ 481, 250, 46 },
		{ 481, 251, 46 },
		{ 481, 252, 46 },
		{ 481, 253, 46 },
		{ 481, 254, 46 },
		{ 481, 255, 46 },
		{ 481, 256, 46 },
		{ 481, 257, 46 },
		{ 481, 258, 46 },
		{ 481, 259, 46 },
		{ 481, 260, 46 },
		{ 481, 261, 46 },
		{ 481, 262, 46 },
		{ 481, 263, 46 },
		{ 481, 264, 46 },
		{ 481, 265, 46 },
		{ 481, 266, 46 },
		{ 481, 267, 46 },
		{ 481, 268, 46 },
		{ 481, 269, 46 },
		{ 481, 270, 46 },
		{ 481, 271, 46 },
		{ 481, 272, 46 },
		{ 481, 273, 46 },
		{ 481, 274, 46 },
		{ 481, 275, 46 },
		{ 481, 276, 46 },
		{ 481, 277, 46 },
		{ 481, 278, 46 },
		{ 481, 279, 46 },
		{ 481, 280, 46 },
		{ 481, 281, 46 },
		{ 481, 282, 46 },
		{ 481, 283, 46 },
		{ 481, 284, 46 },
		{ 482, 229, 46 },
		{ 482, 230, 46 },
		{ 482, 231, 46 },
		{ 482, 232, 46 },
		{ 482, 233, 46 },
		{ 482, 234, 46 },
		{ 482, 235, 46 },
		{ 482, 236, 46 },
		{ 482, 237, 46 },
		{ 482, 238, 46 },
		{ 482, 239, 46 },
		{ 482, 240, 46 },
		{ 482, 241, 46 },
		{ 482, 242, 46 },
		{ 482, 243, 46 },
		{ 482, 244, 46 },
		{ 482, 245, 46 },
		{ 482, 246, 46 },
		{ 482, 247, 46 },
		{ 482, 248, 46 },
		{ 482, 249, 46 },
		{ 482, 250, 46 },
		{ 482, 251, 46 },
		{ 482, 252, 46 },
		{ 482, 253, 46 },
		{ 482, 254, 46 },
		{ 482, 255, 46 },
		{ 482, 256, 46 },
		{ 482, 257, 46 },
		{ 482, 258, 46 },
		{ 482, 259, 46 },
		{ 482, 260, 46 },
		{ 482, 261, 46 },
		{ 482, 262, 46 },
		{ 482, 263, 46 },
		{ 482, 264, 46 },
		{ 482, 265, 46 },
		{ 482, 266, 46 },
		{ 482, 267, 46 },
		{ 482, 268, 46 },
		{ 482, 269, 46 },
		{ 482, 270, 46 },
		{ 482, 271, 46 },
		{ 482, 272, 46 },
		{ 482, 273, 46 },
		{ 482, 274, 46 },
		{ 482, 275, 46 },
		{ 482, 276, 46 },
		{ 482, 277, 46 },
		{ 482, 278, 46 },
		{ 482, 279, 46 },
		{ 482, 280, 46 },
		{ 482, 281, 46 },
		{ 482, 282, 46 },
		{ 482, 283, 46 },
		{ 482, 284, 46 },
		{ 482, 285, 46 },
		{ 483, 229, 46 },
		{ 483, 230, 46 },
		{ 483, 231, 46 },
		{ 483, 232, 46 },
		{ 483, 233, 46 },
		{ 483, 234, 46 },
		{ 483, 235, 46 },
		{ 483, 236, 46 },
		{ 483, 237, 46 },
		{ 483, 238, 46 },
		{ 483, 239, 46 },
		{ 483, 240, 46 },
		{ 483, 241, 46 },
		{ 483, 242, 46 },
		{ 483, 243, 46 },
		{ 483, 244, 46 },
		{ 483, 245, 46 },
		{ 483, 246, 46 },
		{ 483, 247, 46 },
		{ 483, 248, 46 },
		{ 483, 249, 46 },
		{ 483, 250, 46 },
		{ 483, 251, 46 },
		{ 483, 252, 46 },
		{ 483, 253, 46 },
		{ 483, 254, 46 },
		{ 483, 255, 46 },
		{ 483, 256, 46 },
		{ 483, 257, 46 },
		{ 483, 258, 46 },
		{ 483, 259, 46 },
		{ 483, 260, 46 },
		{ 483, 261, 46 },
		{ 483, 262, 46 },
		{ 483, 263, 46 },
		{ 483, 264, 46 },
		{ 483, 265, 46 },
		{ 483, 266, 46 },
		{ 483, 267, 46 },
		{ 483, 268, 46 },
		{ 483, 269, 46 },
		{ 483, 270, 46 },
		{ 483, 271, 46 },
		{ 483, 272, 46 },
		{ 483, 273, 46 },
		{ 483, 274, 46 },
		{ 483, 275, 46 },
		{ 483, 276, 46 },
		{ 483, 277, 46 },
		{ 483, 278, 46 },
		{ 483, 279, 46 },
		{ 483, 280, 46 },
		{ 483, 281, 46 },
		{ 483, 282, 46 },
		{ 483, 283, 46 },
		{ 483, 284, 46 },
		{ 483, 285, 46 },
		{ 484, 229, 46 },
		{ 484, 230, 46 },
		{ 484, 231, 46 },
		{ 484, 232, 46 },
		{ 484, 233, 46 },
		{ 484, 234, 46 },
		{ 484, 235, 46 },
		{ 484, 236, 46 },
		{ 484, 237, 46 },
		{ 484, 238, 46 },
		{ 484, 239, 46 },
		{ 484, 240, 46 },
		{ 484, 241, 46 },
		{ 484, 242, 46 },
		{ 484, 243, 46 },
		{ 484, 244, 46 },
		{ 484, 245, 46 },
		{ 484, 246, 46 },
		{ 484, 247, 46 },
		{ 484, 248, 46 },
		{ 484, 249, 46 },
		{ 484, 250, 46 },
		{ 484, 251, 46 },
		{ 484, 252, 46 },
		{ 484, 253, 46 },
		{ 484, 254, 46 },
		{ 484, 255, 46 },
		{ 484, 256, 46 },
		{ 484, 257, 46 },
		{ 484, 258, 46 },
		{ 484, 259, 46 },
		{ 484, 260, 46 },
		{ 484, 261, 46 },
		{ 484, 262, 46 },
		{ 484, 263, 46 },
		{ 484, 264, 46 },
		{ 484, 265, 46 },
		{ 484, 266, 46 },
		{ 484, 267, 46 },
		{ 484, 268, 46 },
		{ 484, 269, 46 },
		{ 484, 270, 46 },
		{ 484, 271, 46 },
		{ 484, 272, 46 },
		{ 484, 273, 46 },
		{ 484, 274, 46 },
		{ 484, 275, 46 },
		{ 484, 276, 46 },
		{ 484, 277, 46 },
		{ 484, 278, 46 },
		{ 484, 279, 46 },
		{ 484, 280, 46 },
		{ 484, 281, 46 },
		{ 484, 282, 46 },
		{ 484, 283, 46 },
		{ 484, 284, 46 },
		{ 484, 285, 46 },
		{ 485, 229, 46 },
		{ 485, 230, 46 },
		{ 485, 231, 46 },
		{ 485, 232, 46 },
		{ 485, 233, 46 },
		{ 485, 234, 46 },
		{ 485, 235, 46 },
		{ 485, 236, 46 },
		{ 485, 237, 46 },
		{ 485, 238, 46 },
		{ 485, 239, 46 },
		{ 485, 240, 46 },
		{ 485, 241, 46 },
		{ 485, 242, 46 },
		{ 485, 243, 46 },
		{ 485, 244, 46 },
		{ 485, 245, 46 },
		{ 485, 246, 46 },
		{ 485, 247, 46 },
		{ 485, 248, 46 },
		{ 485, 249, 46 },
		{ 485, 250, 46 },
		{ 485, 251, 46 },
		{ 485, 252, 46 },
		{ 485, 253, 46 },
		{ 485, 254, 46 },
		{ 485, 255, 46 },
		{ 485, 256, 46 },
		{ 485, 257, 46 },
		{ 485, 258, 46 },
		{ 485, 259, 46 },
		{ 485, 260, 46 },
		{ 485, 261, 46 },
		{ 485, 262, 46 },
		{ 485, 263, 46 },
		{ 485, 264, 46 },
		{ 485, 265, 46 },
		{ 485, 266, 46 },
		{ 485, 267, 46 },
		{ 485, 268, 46 },
		{ 485, 269, 46 },
		{ 485, 270, 46 },
		{ 485, 271, 46 },
		{ 485, 272, 46 },
		{ 485, 273, 46 },
		{ 485, 274, 46 },
		{ 485, 275, 46 },
		{ 485, 276, 46 },
		{ 485, 277, 46 },
		{ 485, 278, 46 },
		{ 485, 279, 46 },
		{ 485, 280, 46 },
		{ 485, 281, 46 },
		{ 485, 282, 46 },
		{ 485, 283, 46 },
		{ 485, 284, 46 },
		{ 485, 285, 46 },
		{ 486, 229, 46 },
		{ 486, 230, 46 },
		{ 486, 231, 46 },
		{ 486, 232, 46 },
		{ 486, 233, 46 },
		{ 486, 234, 46 },
		{ 486, 235, 46 },
		{ 486, 236, 46 },
		{ 486, 237, 46 },
		{ 486, 238, 46 },
		{ 486, 239, 46 },
		{ 486, 240, 46 },
		{ 486, 241, 46 },
		{ 486, 242, 46 },
		{ 486, 243, 46 },
		{ 486, 244, 46 },
		{ 486, 245, 46 },
		{ 486, 246, 46 },
		{ 486, 247, 46 },
		{ 486, 248, 46 },
		{ 486, 249, 46 },
		{ 486, 250, 46 },
		{ 486, 251, 46 },
		{ 486, 252, 46 },
		{ 486, 253, 46 },
		{ 486, 254, 46 },
		{ 486, 255, 46 },
		{ 486, 256, 46 },
		{ 486, 257, 46 },
		{ 486, 258, 46 },
		{ 486, 259, 46 },
		{ 486, 260, 46 },
		{ 486, 261, 46 },
		{ 486, 262, 46 },
		{ 486, 263, 46 },
		{ 486, 264, 46 },
		{ 486, 265, 46 },
		{ 486, 266, 46 },
		{ 486, 267, 46 },
		{ 486, 268, 46 },
		{ 486, 269, 46 },
		{ 486, 270, 46 },
		{ 486, 271, 46 },
		{ 486, 272, 46 },
		{ 486, 273, 46 },
		{ 486, 274, 46 },
		{ 486, 275, 46 },
		{ 486, 276, 46 },
		{ 486, 277, 46 },
		{ 486, 278, 46 },
		{ 486, 279, 46 },
		{ 486, 280, 46 },
		{ 486, 281, 46 },
		{ 486, 282, 46 },
		{ 486, 283, 46 },
		{ 486, 284, 46 },
		{ 486, 285, 46 },
		{ 487, 229, 46 },
		{ 487, 230, 46 },
		{ 487, 231, 46 },
		{ 487, 232, 46 },
		{ 487, 233, 46 },
		{ 487, 234, 46 },
		{ 487, 235, 46 },
		{ 487, 236, 46 },
		{ 487, 237, 46 },
		{ 487, 238, 46 },
		{ 487, 239, 46 },
		{ 487, 240, 46 },
		{ 487, 241, 46 },
		{ 487, 242, 46 },
		{ 487, 243, 46 },
		{ 487, 244, 46 },
		{ 487, 245, 46 },
		{ 487, 246, 46 },
		{ 487, 247, 46 },
		{ 487, 248, 46 },
		{ 487, 249, 46 },
		{ 487, 250, 46 },
		{ 487, 251, 46 },
		{ 487, 252, 46 },
		{ 487, 253, 46 },
		{ 487, 254, 46 },
		{ 487, 255, 46 },
		{ 487, 256, 46 },
		{ 487, 257, 46 },
		{ 487, 258, 46 },
		{ 487, 259, 46 },
		{ 487, 260, 46 },
		{ 487, 261, 46 },
		{ 487, 262, 46 },
		{ 487, 263, 46 },
		{ 487, 264, 46 },
		{ 487, 265, 46 },
		{ 487, 266, 46 },
		{ 487, 267, 46 },
		{ 487, 268, 46 },
		{ 487, 269, 46 },
		{ 487, 270, 46 },
		{ 487, 271, 46 },
		{ 487, 272, 46 },
		{ 487, 273, 46 },
		{ 487, 274, 46 },
		{ 487, 275, 46 },
		{ 487, 276, 46 },
		{ 487, 277, 46 },
		{ 487, 278, 46 },
		{ 487, 279, 46 },
		{ 487, 280, 46 },
		{ 487, 281, 46 },
		{ 487, 282, 46 },
		{ 487, 283, 46 },
		{ 487, 284, 46 },
		{ 487, 285, 46 },
		{ 488, 229, 46 },
		{ 488, 230, 46 },
		{ 488, 231, 46 },
		{ 488, 232, 46 },
		{ 488, 233, 46 },
		{ 488, 234, 46 },
		{ 488, 235, 46 },
		{ 488, 236, 46 },
		{ 488, 237, 46 },
		{ 488, 238, 46 },
		{ 488, 239, 46 },
		{ 488, 240, 46 },
		{ 488, 241, 46 },
		{ 488, 242, 46 },
		{ 488, 243, 46 },
		{ 488, 244, 46 },
		{ 488, 245, 46 },
		{ 488, 246, 46 },
		{ 488, 247, 46 },
		{ 488, 248, 46 },
		{ 488, 249, 46 },
		{ 488, 250, 46 },
		{ 488, 251, 46 },
		{ 488, 252, 46 },
		{ 488, 253, 46 },
		{ 488, 254, 46 },
		{ 488, 255, 46 },
		{ 488, 256, 46 },
		{ 488, 257, 46 },
		{ 488, 258, 46 },
		{ 488, 259, 46 },
		{ 488, 260, 46 },
		{ 488, 261, 46 },
		{ 488, 262, 46 },
		{ 488, 263, 46 },
		{ 488, 264, 46 },
		{ 488, 265, 46 },
		{ 488, 266, 46 },
		{ 488, 267, 46 },
		{ 488, 268, 46 },
		{ 488, 269, 46 },
		{ 488, 270, 46 },
		{ 488, 271, 46 },
		{ 488, 272, 46 },
		{ 488, 273, 46 },
		{ 488, 274, 46 },
		{ 488, 275, 46 },
		{ 488, 276, 46 },
		{ 488, 277, 46 },
		{ 488, 278, 46 },
		{ 488, 279, 46 },
		{ 488, 280, 46 },
		{ 488, 281, 46 },
		{ 488, 282, 46 },
		{ 488, 283, 46 },
		{ 488, 284, 46 },
		{ 488, 285, 46 },
		{ 489, 229, 46 },
		{ 489, 230, 46 },
		{ 489, 231, 46 },
		{ 489, 232, 46 },
		{ 489, 233, 46 },
		{ 489, 234, 46 },
		{ 489, 235, 46 },
		{ 489, 236, 46 },
		{ 489, 237, 46 },
		{ 489, 238, 46 },
		{ 489, 239, 46 },
		{ 489, 240, 46 },
		{ 489, 241, 46 },
		{ 489, 242, 46 },
		{ 489, 243, 46 },
		{ 489, 244, 46 },
		{ 489, 245, 46 },
		{ 489, 246, 46 },
		{ 489, 247, 46 },
		{ 489, 248, 46 },
		{ 489, 249, 46 },
		{ 489, 250, 46 },
		{ 489, 251, 46 },
		{ 489, 252, 46 },
		{ 489, 253, 46 },
		{ 489, 254, 46 },
		{ 489, 255, 46 },
		{ 489, 256, 46 },
		{ 489, 257, 46 },
		{ 489, 258, 46 },
		{ 489, 259, 46 },
		{ 489, 260, 46 },
		{ 489, 261, 46 },
		{ 489, 262, 46 },
		{ 489, 263, 46 },
		{ 489, 264, 46 },
		{ 489, 265, 46 },
		{ 489, 266, 46 },
		{ 489, 267, 46 },
		{ 489, 268, 46 },
		{ 489, 269, 46 },
		{ 489, 270, 46 },
		{ 489, 271, 46 },
		{ 489, 272, 46 },
		{ 489, 273, 46 },
		{ 489, 274, 46 },
		{ 489, 275, 46 },
		{ 489, 276, 46 },
		{ 489, 277, 46 },
		{ 489, 278, 46 },
		{ 489, 279, 46 },
		{ 489, 280, 46 },
		{ 489, 281, 46 },
		{ 489, 282, 46 },
		{ 489, 283, 46 },
		{ 489, 284, 46 },
		{ 489, 285, 46 },
		{ 490, 228, 46 },
		{ 490, 229, 46 },
		{ 490, 230, 46 },
		{ 490, 231, 46 },
		{ 490, 232, 46 },
		{ 490, 233, 46 },
		{ 490, 234, 46 },
		{ 490, 235, 46 },
		{ 490, 236, 46 },
		{ 490, 237, 46 },
		{ 490, 238, 46 },
		{ 490, 239, 46 },
		{ 490, 240, 46 },
		{ 490, 241, 46 },
		{ 490, 242, 46 },
		{ 490, 243, 46 },
		{ 490, 244, 46 },
		{ 490, 245, 46 },
		{ 490, 246, 46 },
		{ 490, 247, 46 },
		{ 490, 248, 46 },
		{ 490, 249, 46 },
		{ 490, 250, 46 },
		{ 490, 251, 46 },
		{ 490, 252, 46 },
		{ 490, 253, 46 },
		{ 490, 254, 46 },
		{ 490, 255, 46 },
		{ 490, 256, 46 },
		{ 490, 257, 46 },
		{ 490, 258, 46 },
		{ 490, 259, 46 },
		{ 490, 260, 46 },
		{ 490, 261, 46 },
		{ 490, 262, 46 },
		{ 490, 263, 46 },
		{ 490, 264, 46 },
		{ 490, 265, 46 },
		{ 490, 266, 46 },
		{ 490, 267, 46 },
		{ 490, 268, 46 },
		{ 490, 269, 46 },
		{ 490, 270, 46 },
		{ 490, 271, 46 },
		{ 490, 272, 46 },
		{ 490, 273, 46 },
		{ 490, 274, 46 },
		{ 490, 275, 46 },
		{ 490, 276, 46 },
		{ 490, 277, 46 },
		{ 490, 278, 46 },
		{ 490, 279, 46 },
		{ 490, 280, 46 },
		{ 490, 281, 46 },
		{ 490, 282, 46 },
		{ 490, 283, 46 },
		{ 490, 284, 46 },
		{ 490, 285, 46 },
		{ 490, 286, 46 },
		{ 491, 228, 46 },
		{ 491, 229, 46 },
		{ 491, 230, 46 },
		{ 491, 231, 46 },
		{ 491, 232, 46 },
		{ 491, 233, 46 },
		{ 491, 234, 46 },
		{ 491, 235, 46 },
		{ 491, 236, 46 },
		{ 491, 237, 46 },
		{ 491, 238, 46 },
		{ 491, 239, 46 },
		{ 491, 240, 46 },
		{ 491, 241, 46 },
		{ 491, 242, 46 },
		{ 491, 243, 46 },
		{ 491, 244, 46 },
		{ 491, 245, 46 },
		{ 491, 246, 46 },
		{ 491, 247, 46 },
		{ 491, 248, 46 },
		{ 491, 249, 46 },
		{ 491, 250, 46 },
		{ 491, 251, 46 },
		{ 491, 252, 46 },
		{ 491, 253, 46 },
		{ 491, 254, 46 },
		{ 491, 255, 46 },
		{ 491, 256, 46 },
		{ 491, 257, 46 },
		{ 491, 258, 46 },
		{ 491, 259, 46 },
		{ 491, 260, 46 },
		{ 491, 261, 46 },
		{ 491, 262, 46 },
		{ 491, 263, 46 },
		{ 491, 264, 46 },
		{ 491, 265, 46 },
		{ 491, 266, 46 },
		{ 491, 267, 46 },
		{ 491, 268, 46 },
		{ 491, 269, 46 },
		{ 491, 270, 46 },
		{ 491, 271, 46 },
		{ 491, 272, 46 },
		{ 491, 273, 46 },
		{ 491, 274, 46 },
		{ 491, 275, 46 },
		{ 491, 276, 46 },
		{ 491, 277, 46 },
		{ 491, 278, 46 },
		{ 491, 279, 46 },
		{ 491, 280, 46 },
		{ 491, 281, 46 },
		{ 491, 282, 46 },
		{ 491, 283, 46 },
		{ 491, 284, 46 },
		{ 491, 285, 46 },
		{ 491, 286, 46 },
		{ 492, 228, 46 },
		{ 492, 229, 46 },
		{ 492, 230, 46 },
		{ 492, 231, 46 },
		{ 492, 232, 46 },
		{ 492, 233, 46 },
		{ 492, 234, 46 },
		{ 492, 235, 46 },
		{ 492, 236, 46 },
		{ 492, 237, 46 },
		{ 492, 238, 46 },
		{ 492, 239, 46 },
		{ 492, 240, 46 },
		{ 492, 241, 46 },
		{ 492, 242, 46 },
		{ 492, 243, 46 },
		{ 492, 244, 46 },
		{ 492, 245, 46 },
		{ 492, 246, 46 },
		{ 492, 247, 46 },
		{ 492, 248, 46 },
		{ 492, 249, 46 },
		{ 492, 250, 46 },
		{ 492, 251, 46 },
		{ 492, 252, 46 },
		{ 492, 253, 46 },
		{ 492, 254, 46 },
		{ 492, 255, 46 },
		{ 492, 256, 46 },
		{ 492, 257, 46 },
		{ 492, 258, 46 },
		{ 492, 259, 46 },
		{ 492, 260, 46 },
		{ 492, 261, 46 },
		{ 492, 262, 46 },
		{ 492, 263, 46 },
		{ 492, 264, 46 },
		{ 492, 265, 46 },
		{ 492, 266, 46 },
		{ 492, 267, 46 },
		{ 492, 268, 46 },
		{ 492, 269, 46 },
		{ 492, 270, 46 },
		{ 492, 271, 46 },
		{ 492, 272, 46 },
		{ 492, 273, 46 },
		{ 492, 274, 46 },
		{ 492, 275, 46 },
		{ 492, 276, 46 },
		{ 492, 277, 46 },
		{ 492, 278, 46 },
		{ 492, 279, 46 },
		{ 492, 280, 46 },
		{ 492, 281, 46 },
		{ 492, 282, 46 },
		{ 492, 283, 46 },
		{ 492, 284, 46 },
		{ 492, 285, 46 },
		{ 492, 286, 46 },
		{ 493, 228, 46 },
		{ 493, 229, 46 },
		{ 493, 230, 46 },
		{ 493, 231, 46 },
		{ 493, 232, 46 },
		{ 493, 233, 46 },
		{ 493, 234, 46 },
		{ 493, 235, 46 },
		{ 493, 236, 46 },
		{ 493, 237, 46 },
		{ 493, 238, 46 },
		{ 493, 239, 46 },
		{ 493, 240, 46 },
		{ 493, 241, 46 },
		{ 493, 242, 46 },
		{ 493, 243, 46 },
		{ 493, 244, 46 },
		{ 493, 245, 46 },
		{ 493, 246, 46 },
		{ 493, 247, 46 },
		{ 493, 248, 46 },
		{ 493, 249, 46 },
		{ 493, 250, 46 },
		{ 493, 251, 46 },
		{ 493, 252, 46 },
		{ 493, 253, 46 },
		{ 493, 254, 46 },
		{ 493, 255, 46 },
		{ 493, 256, 46 },
		{ 493, 257, 46 },
		{ 493, 258, 46 },
		{ 493, 259, 46 },
		{ 493, 260, 46 },
		{ 493, 261, 46 },
		{ 493, 262, 46 },
		{ 493, 263, 46 },
		{ 493, 264, 46 },
		{ 493, 265, 46 },
		{ 493, 266, 46 },
		{ 493, 267, 46 },
		{ 493, 268, 46 },
		{ 493, 269, 46 },
		{ 493, 270, 46 },
		{ 493, 271, 46 },
		{ 493, 272, 46 },
		{ 493, 273, 46 },
		{ 493, 274, 46 },
		{ 493, 275, 46 },
		{ 493, 276, 46 },
		{ 493, 277, 46 },
		{ 493, 278, 46 },
		{ 493, 279, 46 },
		{ 493, 280, 46 },
		{ 493, 281, 46 },
		{ 493, 282, 46 },
		{ 493, 283, 46 },
		{ 493, 284, 46 },
		{ 493, 285, 46 },
		{ 493, 286, 46 },
		{ 494, 228, 46 },
		{ 494, 229, 46 },
		{ 494, 230, 46 },
		{ 494, 231, 46 },
		{ 494, 232, 46 },
		{ 494, 233, 46 },
		{ 494, 234, 46 },
		{ 494, 235, 46 },
		{ 494, 236, 46 },
		{ 494, 237, 46 },
		{ 494, 238, 46 },
		{ 494, 239, 46 },
		{ 494, 240, 46 },
		{ 494, 241, 46 },
		{ 494, 242, 46 },
		{ 494, 243, 46 },
		{ 494, 244, 46 },
		{ 494, 245, 46 },
		{ 494, 246, 46 },
		{ 494, 247, 46 },
		{ 494, 248, 46 },
		{ 494, 249, 46 },
		{ 494, 250, 46 },
		{ 494, 251, 46 },
		{ 494, 252, 46 },
		{ 494, 253, 46 },
		{ 494, 254, 46 },
		{ 494, 255, 46 },
		{ 494, 256, 46 },
		{ 494, 257, 46 },
		{ 494, 258, 46 },
		{ 494, 259, 46 },
		{ 494, 260, 46 },
		{ 494, 261, 46 },
		{ 494, 262, 46 },
		{ 494, 263, 46 },
		{ 494, 264, 46 },
		{ 494, 265, 46 },
		{ 494, 266, 46 },
		{ 494, 267, 46 },
		{ 494, 268, 46 },
		{ 494, 269, 46 },
		{ 494, 270, 46 },
		{ 494, 271, 46 },
		{ 494, 272, 46 },
		{ 494, 273, 46 },
		{ 494, 274, 46 },
		{ 494, 275, 46 },
		{ 494, 276, 46 },
		{ 494, 277, 46 },
		{ 494, 278, 46 },
		{ 494, 279, 46 },
		{ 494, 280, 46 },
		{ 494, 281, 46 },
		{ 494, 282, 46 },
		{ 494, 283, 46 },
		{ 494, 284, 46 },
		{ 494, 285, 46 },
		{ 494, 286, 46 },
		{ 495, 228, 46 },
		{ 495, 229, 46 },
		{ 495, 230, 46 },
		{ 495, 231, 46 },
		{ 495, 232, 46 },
		{ 495, 233, 46 },
		{ 495, 234, 46 },
		{ 495, 235, 46 },
		{ 495, 236, 46 },
		{ 495, 237, 46 },
		{ 495, 238, 46 },
		{ 495, 239, 46 },
		{ 495, 240, 46 },
		{ 495, 241, 46 },
		{ 495, 242, 46 },
		{ 495, 243, 46 },
		{ 495, 244, 46 },
		{ 495, 245, 46 },
		{ 495, 246, 46 },
		{ 495, 247, 46 },
		{ 495, 248, 46 },
		{ 495, 249, 46 },
		{ 495, 250, 46 },
		{ 495, 251, 46 },
		{ 495, 252, 46 },
		{ 495, 253, 46 },
		{ 495, 254, 46 },
		{ 495, 255, 46 },
		{ 495, 256, 46 },
		{ 495, 257, 46 },
		{ 495, 258, 46 },
		{ 495, 259, 46 },
		{ 495, 260, 46 },
		{ 495, 261, 46 },
		{ 495, 262, 46 },
		{ 495, 263, 46 },
		{ 495, 264, 46 },
		{ 495, 265, 46 },
		{ 495, 266, 46 },
		{ 495, 267, 46 },
		{ 495, 268, 46 },
		{ 495, 269, 46 },
		{ 495, 270, 46 },
		{ 495, 271, 46 },
		{ 495, 272, 46 },
		{ 495, 273, 46 },
		{ 495, 274, 46 },
		{ 495, 275, 46 },
		{ 495, 276, 46 },
		{ 495, 277, 46 },
		{ 495, 278, 46 },
		{ 495, 279, 46 },
		{ 495, 280, 46 },
		{ 495, 281, 46 },
		{ 495, 282, 46 },
		{ 495, 283, 46 },
		{ 495, 284, 46 },
		{ 495, 285, 46 },
		{ 495, 286, 46 },
		{ 496, 228, 46 },
		{ 496, 229, 46 },
		{ 496, 230, 46 },
		{ 496, 231, 46 },
		{ 496, 232, 46 },
		{ 496, 233, 46 },
		{ 496, 234, 46 },
		{ 496, 235, 46 },
		{ 496, 236, 46 },
		{ 496, 237, 46 },
		{ 496, 238, 46 },
		{ 496, 239, 46 },
		{ 496, 240, 46 },
		{ 496, 241, 46 },
		{ 496, 242, 46 },
		{ 496, 243, 46 },
		{ 496, 244, 46 },
		{ 496, 245, 46 },
		{ 496, 246, 46 },
		{ 496, 247, 46 },
		{ 496, 248, 46 },
		{ 496, 249, 46 },
		{ 496, 250, 46 },
		{ 496, 251, 46 },
		{ 496, 252, 46 },
		{ 496, 253, 46 },
		{ 496, 254, 46 },
		{ 496, 255, 46 },
		{ 496, 256, 46 },
		{ 496, 257, 46 },
		{ 496, 258, 46 },
		{ 496, 259, 46 },
		{ 496, 260, 46 },
		{ 496, 261, 46 },
		{ 496, 262, 46 },
		{ 496, 263, 46 },
		{ 496, 264, 46 },
		{ 496, 265, 46 },
		{ 496, 266, 46 },
		{ 496, 267, 46 },
		{ 496, 268, 46 },
		{ 496, 269, 46 },
		{ 496, 270, 46 },
		{ 496, 271, 46 },
		{ 496, 272, 46 },
		{ 496, 273, 46 },
		{ 496, 274, 46 },
		{ 496, 275, 46 },
		{ 496, 276, 46 },
		{ 496, 277, 46 },
		{ 496, 278, 46 },
		{ 496, 279, 46 },
		{ 496, 280, 46 },
		{ 496, 281, 46 },
		{ 496, 282, 46 },
		{ 496, 283, 46 },
		{ 496, 284, 46 },
		{ 496, 285, 46 },
		{ 496, 286, 46 },
		{ 497, 228, 46 },
		{ 497, 229, 46 },
		{ 497, 230, 46 },
		{ 497, 231, 46 },
		{ 497, 232, 46 },
		{ 497, 233, 46 },
		{ 497, 234, 46 },
		{ 497, 235, 46 },
		{ 497, 236, 46 },
		{ 497, 237, 46 },
		{ 497, 238, 46 },
		{ 497, 239, 46 },
		{ 497, 240, 46 },
		{ 497, 241, 46 },
		{ 497, 242, 46 },
		{ 497, 243, 46 },
		{ 497, 244, 46 },
		{ 497, 245, 46 },
		{ 497, 246, 46 },
		{ 497, 247, 46 },
		{ 497, 248, 46 },
		{ 497, 249, 46 },
		{ 497, 250, 46 },
		{ 497, 251, 46 },
		{ 497, 252, 46 },
		{ 497, 253, 46 },
		{ 497, 254, 46 },
		{ 497, 255, 46 },
		{ 497, 256, 46 },
		{ 497, 257, 46 },
		{ 497, 258, 46 },
		{ 497, 259, 46 },
		{ 497, 260, 46 },
		{ 497, 261, 46 },
		{ 497, 262, 46 },
		{ 497, 263, 46 },
		{ 497, 264, 46 },
		{ 497, 265, 46 },
		{ 497, 266, 46 },
		{ 497, 267, 46 },
		{ 497, 268, 46 },
		{ 497, 269, 46 },
		{ 497, 270, 46 },
		{ 497, 271, 46 },
		{ 497, 272, 46 },
		{ 497, 273, 46 },
		{ 497, 274, 46 },
		{ 497, 275, 46 },
		{ 497, 276, 46 },
		{ 497, 277, 46 },
		{ 497, 278, 46 },
		{ 497, 279, 46 },
		{ 497, 280, 46 },
		{ 497, 281, 46 },
		{ 497, 282, 46 },
		{ 497, 283, 46 },
		{ 497, 284, 46 },
		{ 497, 285, 46 },
		{ 497, 286, 46 },
		{ 498, 227, 46 },
		{ 498, 228, 46 },
		{ 498, 229, 46 },
		{ 498, 230, 46 },
		{ 498, 231, 46 },
		{ 498, 232, 46 },
		{ 498, 233, 46 },
		{ 498, 234, 46 },
		{ 498, 235, 46 },
		{ 498, 236, 46 },
		{ 498, 237, 46 },
		{ 498, 238, 46 },
		{ 498, 239, 46 },
		{ 498, 240, 46 },
		{ 498, 241, 46 },
		{ 498, 242, 46 },
		{ 498, 243, 46 },
		{ 498, 244, 46 },
		{ 498, 245, 46 },
		{ 498, 246, 46 },
		{ 498, 247, 46 },
		{ 498, 248, 46 },
		{ 498, 249, 46 },
		{ 498, 250, 46 },
		{ 498, 251, 46 },
		{ 498, 252, 46 },
		{ 498, 253, 46 },
		{ 498, 254, 46 },
		{ 498, 255, 46 },
		{ 498, 256, 46 },
		{ 498, 257, 46 },
		{ 498, 258, 46 },
		{ 498, 259, 46 },
		{ 498, 260, 46 },
		{ 498, 261, 46 },
		{ 498, 262, 46 },
		{ 498, 263, 46 },
		{ 498, 264, 46 },
		{ 498, 265, 46 },
		{ 498, 266, 46 },
		{ 498, 267, 46 },
		{ 498, 268, 46 },
		{ 498, 269, 46 },
		{ 498, 270, 46 },
		{ 498, 271, 46 },
		{ 498, 272, 46 },
		{ 498, 273, 46 },
		{ 498, 274, 46 },
		{ 498, 275, 46 },
		{ 498, 276, 46 },
		{ 498, 277, 46 },
		{ 498, 278, 46 },
		{ 498, 279, 46 },
		{ 498, 280, 46 },
		{ 498, 281, 46 },
		{ 498, 282, 46 },
		{ 498, 283, 46 },
		{ 498, 284, 46 },
		{ 498, 285, 46 },
		{ 498, 286, 46 },
		{ 498, 287, 46 },
		{ 499, 227, 46 },
		{ 499, 228, 46 },
		{ 499, 229, 46 },
		{ 499, 230, 46 },
		{ 499, 231, 46 },
		{ 499, 232, 46 },
		{ 499, 233, 46 },
		{ 499, 234, 46 },
		{ 499, 235, 46 },
		{ 499, 236, 46 },
		{ 499, 237, 46 },
		{ 499, 238, 46 },
		{ 499, 239, 46 },
		{ 499, 240, 46 },
		{ 499, 241, 46 },
		{ 499, 242, 46 },
		{ 499, 243, 46 },
		{ 499, 244, 46 },
		{ 499, 245, 46 },
		{ 499, 246, 46 },
		{ 499, 247, 46 },
		{ 499, 248, 46 },
		{ 499, 249, 46 },
		{ 499, 250, 46 },
		{ 499, 251, 46 },
		{ 499, 252, 46 },
		{ 499, 253, 46 },
		{ 499, 254, 46 },
		{ 499, 255, 46 },
		{ 499, 256, 46 },
		{ 499, 257, 46 },
		{ 499, 258, 46 },
		{ 499, 259, 46 },
		{ 499, 260, 46 },
		{ 499, 261, 46 },
		{ 499, 262, 46 },
		{ 499, 263, 46 },
		{ 499, 264, 46 },
		{ 499, 265, 46 },
		{ 499, 266, 46 },
		{ 499, 267, 46 },
		{ 499, 268, 46 },
		{ 499, 269, 46 },
		{ 499, 270, 46 },
		{ 499, 271, 46 },
		{ 499, 272, 46 },
		{ 499, 273, 46 },
		{ 499, 274, 46 },
		{ 499, 275, 46 },
		{ 499, 276, 46 },
		{ 499, 277, 46 },
		{ 499, 278, 46 },
		{ 499, 279, 46 },
		{ 499, 280, 46 },
		{ 499, 281, 46 },
		{ 499, 282, 46 },
		{ 499, 283, 46 },
		{ 499, 284, 46 },
		{ 499, 285, 46 },
		{ 499, 286, 46 },
		{ 499, 287, 46 },
		{ 500, 227, 46 },
		{ 500, 228, 46 },
		{ 500, 229, 46 },
		{ 500, 230, 46 },
		{ 500, 231, 46 },
		{ 500, 232, 46 },
		{ 500, 233, 46 },
		{ 500, 234, 46 },
		{ 500, 235, 46 },
		{ 500, 236, 46 },
		{ 500, 237, 46 },
		{ 500, 238, 46 },
		{ 500, 239, 46 },
		{ 500, 240, 46 },
		{ 500, 241, 46 },
		{ 500, 242, 46 },
		{ 500, 243, 46 },
		{ 500, 244, 46 },
		{ 500, 245, 46 },
		{ 500, 246, 46 },
		{ 500, 247, 46 },
		{ 500, 248, 46 },
		{ 500, 249, 46 },
		{ 500, 250, 46 },
		{ 500, 251, 46 },
		{ 500, 252, 46 },
		{ 500, 253, 46 },
		{ 500, 254, 46 },
		{ 500, 255, 46 },
		{ 500, 256, 46 },
		{ 500, 257, 46 },
		{ 500, 258, 46 },
		{ 500, 259, 46 },
		{ 500, 260, 46 },
		{ 500, 261, 46 },
		{ 500, 262, 46 },
		{ 500, 263, 46 },
		{ 500, 264, 46 },
		{ 500, 265, 46 },
		{ 500, 266, 46 },
		{ 500, 267, 46 },
		{ 500, 268, 46 },
		{ 500, 269, 46 },
		{ 500, 270, 46 },
		{ 500, 271, 46 },
		{ 500, 272, 46 },
		{ 500, 273, 46 },
		{ 500, 274, 46 },
		{ 500, 275, 46 },
		{ 500, 276, 46 },
		{ 500, 277, 46 },
		{ 500, 278, 46 },
		{ 500, 279, 46 },
		{ 500, 280, 46 },
		{ 500, 281, 46 },
		{ 500, 282, 46 },
		{ 500, 283, 46 },
		{ 500, 284, 46 },
		{ 500, 285, 46 },
		{ 500, 286, 46 },
		{ 500, 287, 46 },
		{ 501, 227, 46 },
		{ 501, 228, 46 },
		{ 501, 229, 46 },
		{ 501, 230, 46 },
		{ 501, 231, 46 },
		{ 501, 232, 46 },
		{ 501, 233, 46 },
		{ 501, 234, 46 },
		{ 501, 235, 46 },
		{ 501, 236, 46 },
		{ 501, 237, 46 },
		{ 501, 238, 46 },
		{ 501, 239, 46 },
		{ 501, 240, 46 },
		{ 501, 241, 46 },
		{ 501, 242, 46 },
		{ 501, 243, 46 },
		{ 501, 244, 46 },
		{ 501, 245, 46 },
		{ 501, 246, 46 },
		{ 501, 247, 46 },
		{ 501, 248, 46 },
		{ 501, 249, 46 },
		{ 501, 250, 46 },
		{ 501, 251, 46 },
		{ 501, 252, 46 },
		{ 501, 253, 46 },
		{ 501, 254, 46 },
		{ 501, 255, 46 },
		{ 501, 256, 46 },
		{ 501, 257, 46 },
		{ 501, 258, 46 },
		{ 501, 259, 46 },
		{ 501, 260, 46 },
		{ 501, 261, 46 },
		{ 501, 262, 46 },
		{ 501, 263, 46 },
		{ 501, 264, 46 },
		{ 501, 265, 46 },
		{ 501, 266, 46 },
		{ 501, 267, 46 },
		{ 501, 268, 46 },
		{ 501, 269, 46 },
		{ 501, 270, 46 },
		{ 501, 271, 46 },
		{ 501, 272, 46 },
		{ 501, 273, 46 },
		{ 501, 274, 46 },
		{ 501, 275, 46 },
		{ 501, 276, 46 },
		{ 501, 277, 46 },
		{ 501, 278, 46 },
		{ 501, 279, 46 },
		{ 501, 280, 46 },
		{ 501, 281, 46 },
		{ 501, 282, 46 },
		{ 501, 283, 46 },
		{ 501, 284, 46 },
		{ 501, 285, 46 },
		{ 501, 286, 46 },
		{ 501, 287, 46 },
		{ 502, 227, 46 },
		{ 502, 228, 46 },
		{ 502, 229, 46 },
		{ 502, 230, 46 },
		{ 502, 231, 46 },
		{ 502, 232, 46 },
		{ 502, 233, 46 },
		{ 502, 234, 46 },
		{ 502, 235, 46 },
		{ 502, 236, 46 },
		{ 502, 237, 46 },
		{ 502, 238, 46 },
		{ 502, 239, 46 },
		{ 502, 240, 46 },
		{ 502, 241, 46 },
		{ 502, 242, 46 },
		{ 502, 243, 46 },
		{ 502, 244, 46 },
		{ 502, 245, 46 },
		{ 502, 246, 46 },
		{ 502, 247, 46 },
		{ 502, 248, 46 },
		{ 502, 249, 46 },
		{ 502, 250, 46 },
		{ 502, 251, 46 },
		{ 502, 252, 46 },
		{ 502, 253, 46 },
		{ 502, 254, 46 },
		{ 502, 255, 46 },
		{ 502, 256, 46 },
		{ 502, 257, 46 },
		{ 502, 258, 46 },
		{ 502, 259, 46 },
		{ 502, 260, 46 },
		{ 502, 261, 46 },
		{ 502, 262, 46 },
		{ 502, 263, 46 },
		{ 502, 264, 46 },
		{ 502, 265, 46 },
		{ 502, 266, 46 },
		{ 502, 267, 46 },
		{ 502, 268, 46 },
		{ 502, 269, 46 },
		{ 502, 270, 46 },
		{ 502, 271, 46 },
		{ 502, 272, 46 },
		{ 502, 273, 46 },
		{ 502, 274, 46 },
		{ 502, 275, 46 },
		{ 502, 276, 46 },
		{ 502, 277, 46 },
		{ 502, 278, 46 },
		{ 502, 279, 46 },
		{ 502, 280, 46 },
		{ 502, 281, 46 },
		{ 502, 282, 46 },
		{ 502, 283, 46 },
		{ 502, 284, 46 },
		{ 502, 285, 46 },
		{ 502, 286, 46 },
		{ 502, 287, 46 },
		{ 503, 227, 46 },
		{ 503, 228, 46 },
		{ 503, 229, 46 },
		{ 503, 230, 46 },
		{ 503, 231, 46 },
		{ 503, 232, 46 },
		{ 503, 233, 46 },
		{ 503, 234, 46 },
		{ 503, 235, 46 },
		{ 503, 236, 46 },
		{ 503, 237, 46 },
		{ 503, 238, 46 },
		{ 503, 239, 46 },
		{ 503, 240, 46 },
		{ 503, 241, 46 },
		{ 503, 242, 46 },
		{ 503, 243, 46 },
		{ 503, 244, 46 },
		{ 503, 245, 46 },
		{ 503, 246, 46 },
		{ 503, 247, 46 },
		{ 503, 248, 46 },
		{ 503, 249, 46 },
		{ 503, 250, 46 },
		{ 503, 251, 46 },
		{ 503, 252, 46 },
		{ 503, 253, 46 },
		{ 503, 254, 46 },
		{ 503, 255, 46 },
		{ 503, 256, 46 },
		{ 503, 257, 46 },
		{ 503, 258, 46 },
		{ 503, 259, 46 },
		{ 503, 260, 46 },
		{ 503, 261, 46 },
		{ 503, 262, 46 },
		{ 503, 263, 46 },
		{ 503, 264, 46 },
		{ 503, 265, 46 },
		{ 503, 266, 46 },
		{ 503, 267, 46 },
		{ 503, 268, 46 },
		{ 503, 269, 46 },
		{ 503, 270, 46 },
		{ 503, 271, 46 },
		{ 503, 272, 46 },
		{ 503, 273, 46 },
		{ 503, 274, 46 },
		{ 503, 275, 46 },
		{ 503, 276, 46 },
		{ 503, 277, 46 },
		{ 503, 278, 46 },
		{ 503, 279, 46 },
		{ 503, 280, 46 },
		{ 503, 281, 46 },
		{ 503, 282, 46 },
		{ 503, 283, 46 },
		{ 503, 284, 46 },
		{ 503, 285, 46 },
		{ 503, 286, 46 },
		{ 503, 287, 46 },
		{ 504, 227, 46 },
		{ 504, 228, 46 },
		{ 504, 229, 46 },
		{ 504, 230, 46 },
		{ 504, 231, 46 },
		{ 504, 232, 46 },
		{ 504, 233, 46 },
		{ 504, 234, 46 },
		{ 504, 235, 46 },
		{ 504, 236, 46 },
		{ 504, 237, 46 },
		{ 504, 238, 46 },
		{ 504, 239, 46 },
		{ 504, 240, 46 },
		{ 504, 241, 46 },
		{ 504, 242, 46 },
		{ 504, 243, 46 },
		{ 504, 244, 46 },
		{ 504, 245, 46 },
		{ 504, 246, 46 },
		{ 504, 247, 46 },
		{ 504, 248, 46 },
		{ 504, 249, 46 },
		{ 504, 250, 46 },
		{ 504, 251, 46 },
		{ 504, 252, 46 },
		{ 504, 253, 46 },
		{ 504, 254, 46 },
		{ 504, 255, 46 },
		{ 504, 256, 46 },
		{ 504, 257, 46 },
		{ 504, 258, 46 },
		{ 504, 259, 46 },
		{ 504, 260, 46 },
		{ 504, 261, 46 },
		{ 504, 262, 46 },
		{ 504, 263, 46 },
		{ 504, 264, 46 },
		{ 504, 265, 46 },
		{ 504, 266, 46 },
		{ 504, 267, 46 },
		{ 504, 268, 46 },
		{ 504, 269, 46 },
		{ 504, 270, 46 },
		{ 504, 271, 46 },
		{ 504, 272, 46 },
		{ 504, 273, 46 },
		{ 504, 274, 46 },
		{ 504, 275, 46 },
		{ 504, 276, 46 },
		{ 504, 277, 46 },
		{ 504, 278, 46 },
		{ 504, 279, 46 },
		{ 504, 280, 46 },
		{ 504, 281, 46 },
		{ 504, 282, 46 },
		{ 504, 283, 46 },
		{ 504, 284, 46 },
		{ 504, 285, 46 },
		{ 504, 286, 46 },
		{ 504, 287, 46 },
		{ 505, 227, 46 },
		{ 505, 228, 46 },
		{ 505, 229, 46 },
		{ 505, 230, 46 },
		{ 505, 231, 46 },
		{ 505, 232, 46 },
		{ 505, 233, 46 },
		{ 505, 234, 46 },
		{ 505, 235, 46 },
		{ 505, 236, 46 },
		{ 505, 237, 46 },
		{ 505, 238, 46 },
		{ 505, 239, 46 },
		{ 505, 240, 46 },
		{ 505, 241, 46 },
		{ 505, 242, 46 },
		{ 505, 243, 46 },
		{ 505, 244, 46 },
		{ 505, 245, 46 },
		{ 505, 246, 46 },
		{ 505, 247, 46 },
		{ 505, 248, 46 },
		{ 505, 249, 46 },
		{ 505, 250, 46 },
		{ 505, 251, 46 },
		{ 505, 252, 46 },
		{ 505, 253, 46 },
		{ 505, 254, 46 },
		{ 505, 255, 46 },
		{ 505, 256, 46 },
		{ 505, 257, 46 },
		{ 505, 258, 46 },
		{ 505, 259, 46 },
		{ 505, 260, 46 },
		{ 505, 261, 46 },
		{ 505, 262, 46 },
		{ 505, 263, 46 },
		{ 505, 264, 46 },
		{ 505, 265, 46 },
		{ 505, 266, 46 },
		{ 505, 267, 46 },
		{ 505, 268, 46 },
		{ 505, 269, 46 },
		{ 505, 270, 46 },
		{ 505, 271, 46 },
		{ 505, 272, 46 },
		{ 505, 273, 46 },
		{ 505, 274, 46 },
		{ 505, 275, 46 },
		{ 505, 276, 46 },
		{ 505, 277, 46 },
		{ 505, 278, 46 },
		{ 505, 279, 46 },
		{ 505, 280, 46 },
		{ 505, 281, 46 },
		{ 505, 282, 46 },
		{ 505, 283, 46 },
		{ 505, 284, 46 },
		{ 505, 285, 46 },
		{ 505, 286, 46 },
		{ 505, 287, 46 },
		{ 506, 226, 46 },
		{ 506, 227, 46 },
		{ 506, 228, 46 },
		{ 506, 229, 46 },
		{ 506, 230, 46 },
		{ 506, 231, 46 },
		{ 506, 232, 46 },
		{ 506, 233, 46 },
		{ 506, 234, 46 },
		{ 506, 235, 46 },
		{ 506, 236, 46 },
		{ 506, 237, 46 },
		{ 506, 238, 46 },
		{ 506, 239, 46 },
		{ 506, 240, 46 },
		{ 506, 241, 46 },
		{ 506, 242, 46 },
		{ 506, 243, 46 },
		{ 506, 244, 46 },
		{ 506, 245, 46 },
		{ 506, 246, 46 },
		{ 506, 247, 46 },
		{ 506, 248, 46 },
		{ 506, 249, 46 },
		{ 506, 250, 46 },
		{ 506, 251, 46 },
		{ 506, 252, 46 },
		{ 506, 253, 46 },
		{ 506, 254, 46 },
		{ 506, 255, 46 },
		{ 506, 256, 46 },
		{ 506, 257, 46 },
		{ 506, 258, 46 },
		{ 506, 259, 46 },
		{ 506, 260, 46 },
		{ 506, 261, 46 },
		{ 506, 262, 46 },
		{ 506, 263, 46 },
		{ 506, 264, 46 },
		{ 506, 265, 46 },
		{ 506, 266, 46 },
		{ 506, 267, 46 },
		{ 506, 268, 46 },
		{ 506, 269, 46 },
		{ 506, 270, 46 },
		{ 506, 271, 46 },
		{ 506, 272, 46 },
		{ 506, 273, 46 },
		{ 506, 274, 46 },
		{ 506, 275, 46 },
		{ 506, 276, 46 },
		{ 506, 277, 46 },
		{ 506, 278, 46 },
		{ 506, 279, 46 },
		{ 506, 280, 46 },
		{ 506, 281, 46 },
		{ 506, 282, 46 },
		{ 506, 283, 46 },
		{ 506, 284, 46 },
		{ 506, 285, 46 },
		{ 506, 286, 46 },
		{ 506, 287, 46 },
		{ 506, 288, 46 },
		{ 507, 226, 46 },
		{ 507, 227, 46 },
		{ 507, 228, 46 },
		{ 507, 229, 46 },
		{ 507, 230, 46 },
		{ 507, 231, 46 },
		{ 507, 232, 46 },
		{ 507, 233, 46 },
		{ 507, 234, 46 },
		{ 507, 235, 46 },
		{ 507, 236, 46 },
		{ 507, 237, 46 },
		{ 507, 238, 46 },
		{ 507, 239, 46 },
		{ 507, 240, 46 },
		{ 507, 241, 46 },
		{ 507, 242, 46 },
		{ 507, 243, 46 },
		{ 507, 244, 46 },
		{ 507, 245, 46 },
		{ 507, 246, 46 },
		{ 507, 247, 46 },
		{ 507, 248, 46 },
		{ 507, 249, 46 },
		{ 507, 250, 46 },
		{ 507, 251, 46 },
		{ 507, 252, 46 },
		{ 507, 253, 46 },
		{ 507, 254, 46 },
		{ 507, 255, 46 },
		{ 507, 256, 46 },
		{ 507, 257, 46 },
		{ 507, 258, 46 },
		{ 507, 259, 46 },
		{ 507, 260, 46 },
		{ 507, 261, 46 },
		{ 507, 262, 46 },
		{ 507, 263, 46 },
		{ 507, 264, 46 },
		{ 507, 265, 46 },
		{ 507, 266, 46 },
		{ 507, 267, 46 },
		{ 507, 268, 46 },
		{ 507, 269, 46 },
		{ 507, 270, 46 },
		{ 507, 271, 46 },
		{ 507, 272, 46 },
		{ 507, 273, 46 },
		{ 507, 274, 46 },
		{ 507, 275, 46 },
		{ 507, 276, 46 },
		{ 507, 277, 46 },
		{ 507, 278, 46 },
		{ 507, 279, 46 },
		{ 507, 280, 46 },
		{ 507, 281, 46 },
		{ 507, 282, 46 },
		{ 507, 283, 46 },
		{ 507, 284, 46 },
		{ 507, 285, 46 },
		{ 507, 286, 46 },
		{ 507, 287, 46 },
		{ 507, 288, 46 },
		{ 508, 226, 46 },
		{ 508, 227, 46 },
		{ 508, 228, 46 },
		{ 508, 229, 46 },
		{ 508, 230, 46 },
		{ 508, 231, 46 },
		{ 508, 232, 46 },
		{ 508, 233, 46 },
		{ 508, 234, 46 },
		{ 508, 235, 46 },
		{ 508, 236, 46 },
		{ 508, 237, 46 },
		{ 508, 238, 46 },
		{ 508, 239, 46 },
		{ 508, 240, 46 },
		{ 508, 241, 46 },
		{ 508, 242, 46 },
		{ 508, 243, 46 },
		{ 508, 244, 46 },
		{ 508, 245, 46 },
		{ 508, 246, 46 },
		{ 508, 247, 46 },
		{ 508, 248, 46 },
		{ 508, 249, 46 },
		{ 508, 250, 46 },
		{ 508, 251, 46 },
		{ 508, 252, 46 },
		{ 508, 253, 46 },
		{ 508, 254, 46 },
		{ 508, 255, 46 },
		{ 508, 256, 46 },
		{ 508, 257, 46 },
		{ 508, 258, 46 },
		{ 508, 259, 46 },
		{ 508, 260, 46 },
		{ 508, 261, 46 },
		{ 508, 262, 46 },
		{ 508, 263, 46 },
		{ 508, 264, 46 },
		{ 508, 265, 46 },
		{ 508, 266, 46 },
		{ 508, 267, 46 },
		{ 508, 268, 46 },
		{ 508, 269, 46 },
		{ 508, 270, 46 },
		{ 508, 271, 46 },
		{ 508, 272, 46 },
		{ 508, 273, 46 },
		{ 508, 274, 46 },
		{ 508, 275, 46 },
		{ 508, 276, 46 },
		{ 508, 277, 46 },
		{ 508, 278, 46 },
		{ 508, 279, 46 },
		{ 508, 280, 46 },
		{ 508, 281, 46 },
		{ 508, 282, 46 },
		{ 508, 283, 46 },
		{ 508, 284, 46 },
		{ 508, 285, 46 },
		{ 508, 286, 46 },
		{ 508, 287, 46 },
		{ 508, 288, 46 },
		{ 509, 226, 46 },
		{ 509, 227, 46 },
		{ 509, 228, 46 },
		{ 509, 229, 46 },
		{ 509, 230, 46 },
		{ 509, 231, 46 },
		{ 509, 232, 46 },
		{ 509, 233, 46 },
		{ 509, 234, 46 },
		{ 509, 235, 46 },
		{ 509, 236, 46 },
		{ 509, 237, 46 },
		{ 509, 238, 46 },
		{ 509, 239, 46 },
		{ 509, 240, 46 },
		{ 509, 241, 46 },
		{ 509, 242, 46 },
		{ 509, 243, 46 },
		{ 509, 244, 46 },
		{ 509, 245, 46 },
		{ 509, 246, 46 },
		{ 509, 247, 46 },
		{ 509, 248, 46 },
		{ 509, 249, 46 },
		{ 509, 250, 46 },
		{ 509, 251, 46 },
		{ 509, 252, 46 },
		{ 509, 253, 46 },
		{ 509, 254, 46 },
		{ 509, 255, 46 },
		{ 509, 256, 46 },
		{ 509, 257, 46 },
		{ 509, 258, 46 },
		{ 509, 259, 46 },
		{ 509, 260, 46 },
		{ 509, 261, 46 },
		{ 509, 262, 46 },
		{ 509, 263, 46 },
		{ 509, 264, 46 },
		{ 509, 265, 46 },
		{ 509, 266, 46 },
		{ 509, 267, 46 },
		{ 509, 268, 46 },
		{ 509, 269, 46 },
		{ 509, 270, 46 },
		{ 509, 271, 46 },
		{ 509, 272, 46 },
		{ 509, 273, 46 },
		{ 509, 274, 46 },
		{ 509, 275, 46 },
		{ 509, 276, 46 },
		{ 509, 277, 46 },
		{ 509, 278, 46 },
		{ 509, 279, 46 },
		{ 509, 280, 46 },
		{ 509, 281, 46 },
		{ 509, 282, 46 },
		{ 509, 283, 46 },
		{ 509, 284, 46 },
		{ 509, 285, 46 },
		{ 509, 286, 46 },
		{ 509, 287, 46 },
		{ 509, 288, 46 },
		{ 510, 226, 46 },
		{ 510, 227, 46 },
		{ 510, 228, 46 },
		{ 510, 229, 46 },
		{ 510, 230, 46 },
		{ 510, 231, 46 },
		{ 510, 232, 46 },
		{ 510, 233, 46 },
		{ 510, 234, 46 },
		{ 510, 235, 46 },
		{ 510, 236, 46 },
		{ 510, 237, 46 },
		{ 510, 238, 46 },
		{ 510, 239, 46 },
		{ 510, 240, 46 },
		{ 510, 241, 46 },
		{ 510, 242, 46 },
		{ 510, 243, 46 },
		{ 510, 244, 46 },
		{ 510, 245, 46 },
		{ 510, 246, 46 },
		{ 510, 247, 46 },
		{ 510, 248, 46 },
		{ 510, 249, 46 },
		{ 510, 250, 46 },
		{ 510, 251, 46 },
		{ 510, 252, 46 },
		{ 510, 253, 46 },
		{ 510, 254, 46 },
		{ 510, 255, 46 },
		{ 510, 256, 46 },
		{ 510, 257, 46 },
		{ 510, 258, 46 },
		{ 510, 259, 46 },
		{ 510, 260, 46 },
		{ 510, 261, 46 },
		{ 510, 262, 46 },
		{ 510, 263, 46 },
		{ 510, 264, 46 },
		{ 510, 265, 46 },
		{ 510, 266, 46 },
		{ 510, 267, 46 },
		{ 510, 268, 46 },
		{ 510, 269, 46 },
		{ 510, 270, 46 },
		{ 510, 271, 46 },
		{ 510, 272, 46 },
		{ 510, 273, 46 },
		{ 510, 274, 46 },
		{ 510, 275, 46 },
		{ 510, 276, 46 },
		{ 510, 277, 46 },
		{ 510, 278, 46 },
		{ 510, 279, 46 },
		{ 510, 280, 46 },
		{ 510, 281, 46 },
		{ 510, 282, 46 },
		{ 510, 283, 46 },
		{ 510, 284, 46 },
		{ 510, 285, 46 },
		{ 510, 286, 46 },
		{ 510, 287, 46 },
		{ 510, 288, 46 },
		{ 511, 226, 46 },
		{ 511, 227, 46 },
		{ 511, 228, 46 },
		{ 511, 229, 46 },
		{ 511, 230, 46 },
		{ 511, 231, 46 },
		{ 511, 232, 46 },
		{ 511, 233, 46 },
		{ 511, 234, 46 },
		{ 511, 235, 46 },
		{ 511, 236, 46 },
		{ 511, 237, 46 },
		{ 511, 238, 46 },
		{ 511, 239, 46 },
		{ 511, 240, 46 },
		{ 511, 241, 46 },
		{ 511, 242, 46 },
		{ 511, 243, 46 },
		{ 511, 244, 46 },
		{ 511, 245, 46 },
		{ 511, 246, 46 },
		{ 511, 247, 46 },
		{ 511, 248, 46 },
		{ 511, 249, 46 },
		{ 511, 250, 46 },
		{ 511, 251, 46 },
		{ 511, 252, 46 },
		{ 511, 253, 46 },
		{ 511, 254, 46 },
		{ 511, 255, 46 },
		{ 511, 256, 46 },
		{ 511, 257, 46 },
		{ 511, 258, 46 },
		{ 511, 259, 46 },
		{ 511, 260, 46 },
		{ 511, 261, 46 },
		{ 511, 262, 46 },
		{ 511, 263, 46 },
		{ 511, 264, 46 },
		{ 511, 265, 46 },
		{ 511, 266, 46 },
		{ 511, 267, 46 },
		{ 511, 268, 46 },
		{ 511, 269, 46 },
		{ 511, 270, 46 },
		{ 511, 271, 46 },
		{ 511, 272, 46 },
		{ 511, 273, 46 },
		{ 511, 274, 46 },
		{ 511, 275, 46 },
		{ 511, 276, 46 },
		{ 511, 277, 46 },
		{ 511, 278, 46 },
		{ 511, 279, 46 },
		{ 511, 280, 46 },
		{ 511, 281, 46 },
		{ 511, 282, 46 },
		{ 511, 283, 46 },
		{ 511, 284, 46 },
		{ 511, 285, 46 },
		{ 511, 286, 46 },
		{ 511, 287, 46 },
		{ 511, 288, 46 },
		{ 512, 226, 46 },
		{ 512, 227, 46 },
		{ 512, 228, 46 },
		{ 512, 229, 46 },
		{ 512, 230, 46 },
		{ 512, 231, 46 },
		{ 512, 232, 46 },
		{ 512, 233, 46 },
		{ 512, 234, 46 },
		{ 512, 235, 46 },
		{ 512, 236, 46 },
		{ 512, 237, 46 },
		{ 512, 238, 46 },
		{ 512, 239, 46 },
		{ 512, 240, 46 },
		{ 512, 241, 46 },
		{ 512, 242, 46 },
		{ 512, 243, 46 },
		{ 512, 244, 46 },
		{ 512, 245, 46 },
		{ 512, 246, 46 },
		{ 512, 247, 46 },
		{ 512, 248, 46 },
		{ 512, 249, 46 },
		{ 512, 250, 46 },
		{ 512, 251, 46 },
		{ 512, 252, 46 },
		{ 512, 253, 46 },
		{ 512, 254, 46 },
		{ 512, 255, 46 },
		{ 512, 256, 46 },
		{ 512, 257, 46 },
		{ 512, 258, 46 },
		{ 512, 259, 46 },
		{ 512, 260, 46 },
		{ 512, 261, 46 },
		{ 512, 262, 46 },
		{ 512, 263, 46 },
		{ 512, 264, 46 },
		{ 512, 265, 46 },
		{ 512, 266, 46 },
		{ 512, 267, 46 },
		{ 512, 268, 46 },
		{ 512, 269, 46 },
		{ 512, 270, 46 },
		{ 512, 271, 46 },
		{ 512, 272, 46 },
		{ 512, 273, 46 },
		{ 512, 274, 46 },
		{ 512, 275, 46 },
		{ 512, 276, 46 },
		{ 512, 277, 46 },
		{ 512, 278, 46 },
		{ 512, 279, 46 },
		{ 512, 280, 46 },
		{ 512, 281, 46 },
		{ 512, 282, 46 },
		{ 512, 283, 46 },
		{ 512, 284, 46 },
		{ 512, 285, 46 },
		{ 512, 286, 46 },
		{ 512, 287, 46 },
		{ 512, 288, 46 },
		{ 1, 194, 47 },
		{ 1, 195, 47 },
		{ 1, 196, 47 },
		{ 1, 197, 47 },
		{ 1, 198, 47 },
		{ 1, 199, 47 },
		{ 1, 200, 47 },
		{ 1, 201, 47 },
		{ 1, 202, 47 },
		{ 1, 203, 47 },
		{ 1, 204, 47 },
		{ 1, 205, 47 },
		{ 1, 206, 47 },
		{ 1, 207, 47 },
		{ 1, 208, 47 },
		{ 1, 209, 47 },
		{ 1, 210, 47 },
		{ 1, 211, 47 },
		{ 1, 212, 47 },
		{ 1, 213, 47 },
		{ 1, 214, 47 },
		{ 1, 215, 47 },
		{ 1, 216, 47 },
		{ 1, 217, 47 },
		{ 1, 218, 47 },
		{ 1, 219, 47 },
		{ 1, 220, 47 },
		{ 1, 221, 47 },
		{ 1, 222, 47 },
		{ 1, 223, 47 },
		{ 1, 224, 47 },
		{ 1, 225, 47 },
		{ 1, 226, 47 },
		{ 1, 227, 47 },
		{ 1, 228, 47 },
		{ 1, 229, 47 },
		{ 1, 230, 47 },
		{ 1, 231, 47 },
		{ 1, 232, 47 },
		{ 1, 233, 47 },
		{ 1, 234, 47 },
		{ 1, 235, 47 },
		{ 1, 236, 47 },
		{ 1, 237, 47 },
		{ 1, 238, 47 },
		{ 1, 239, 47 },
		{ 1, 240, 47 },
		{ 1, 241, 47 },
		{ 1, 242, 47 },
		{ 1, 243, 47 },
		{ 1, 244, 47 },
		{ 1, 245, 47 },
		{ 1, 246, 47 },
		{ 1, 247, 47 },
		{ 1, 248, 47 },
		{ 1, 249, 47 },
		{ 1, 250, 47 },
		{ 1, 251, 47 },
		{ 1, 252, 47 },
		{ 1, 253, 47 },
		{ 1, 254, 47 },
		{ 1, 255, 47 },
		{ 1, 256, 47 },
		{ 2, 194, 47 },
		{ 2, 195, 47 },
		{ 2, 196, 47 },
		{ 2, 197, 47 },
		{ 2, 198, 47 },
		{ 2, 199, 47 },
		{ 2, 200, 47 },
		{ 2, 201, 47 },
		{ 2, 202, 47 },
		{ 2, 203, 47 },
		{ 2, 204, 47 },
		{ 2, 205, 47 },
		{ 2, 206, 47 },
		{ 2, 207, 47 },
		{ 2, 208, 47 },
		{ 2, 209, 47 },
		{ 2, 210, 47 },
		{ 2, 211, 47 },
		{ 2, 212, 47 },
		{ 2, 213, 47 },
		{ 2, 214, 47 },
		{ 2, 215, 47 },
		{ 2, 216, 47 },
		{ 2, 217, 47 },
		{ 2, 218, 47 },
		{ 2, 219, 47 },
		{ 2, 220, 47 },
		{ 2, 221, 47 },
		{ 2, 222, 47 },
		{ 2, 223, 47 },
		{ 2, 224, 47 },
		{ 2, 225, 47 },
		{ 2, 226, 47 },
		{ 2, 227, 47 },
		{ 2, 228, 47 },
		{ 2, 229, 47 },
		{ 2, 230, 47 },
		{ 2, 231, 47 },
		{ 2, 232, 47 },
		{ 2, 233, 47 },
		{ 2, 234, 47 },
		{ 2, 235, 47 },
		{ 2, 236, 47 },
		{ 2, 237, 47 },
		{ 2, 238, 47 },
		{ 2, 239, 47 },
		{ 2, 240, 47 },
		{ 2, 241, 47 },
		{ 2, 242, 47 },
		{ 2, 243, 47 },
		{ 2, 244, 47 },
		{ 2, 245, 47 },
		{ 2, 246, 47 },
		{ 2, 247, 47 },
		{ 2, 248, 47 },
		{ 2, 249, 47 },
		{ 2, 250, 47 },
		{ 2, 251, 47 },
		{ 2, 252, 47 },
		{ 2, 253, 47 },
		{ 2, 254, 47 },
		{ 2, 255, 47 },
		{ 2, 256, 47 },
		{ 3, 194, 47 },
		{ 3, 195, 47 },
		{ 3, 196, 47 },
		{ 3, 197, 47 },
		{ 3, 198, 47 },
		{ 3, 199, 47 },
		{ 3, 200, 47 },
		{ 3, 201, 47 },
		{ 3, 202, 47 },
		{ 3, 203, 47 },
		{ 3, 204, 47 },
		{ 3, 205, 47 },
		{ 3, 206, 47 },
		{ 3, 207, 47 },
		{ 3, 208, 47 },
		{ 3, 209, 47 },
		{ 3, 210, 47 },
		{ 3, 211, 47 },
		{ 3, 212, 47 },
		{ 3, 213, 47 },
		{ 3, 214, 47 },
		{ 3, 215, 47 },
		{ 3, 216, 47 },
		{ 3, 217, 47 },
		{ 3, 218, 47 },
		{ 3, 219, 47 },
		{ 3, 220, 47 },
		{ 3, 221, 47 },
		{ 3, 222, 47 },
		{ 3, 223, 47 },
		{ 3, 224, 47 },
		{ 3, 225, 47 },
		{ 3, 226, 47 },
		{ 3, 227, 47 },
		{ 3, 228, 47 },
		{ 3, 229, 47 },
		{ 3, 230, 47 },
		{ 3, 231, 47 },
		{ 3, 232, 47 },
		{ 3, 233, 47 },
		{ 3, 234, 47 },
		{ 3, 235, 47 },
		{ 3, 236, 47 },
		{ 3, 237, 47 },
		{ 3, 238, 47 },
		{ 3, 239, 47 },
		{ 3, 240, 47 },
		{ 3, 241, 47 },
		{ 3, 242, 47 },
		{ 3, 243, 47 },
		{ 3, 244, 47 },
		{ 3, 245, 47 },
		{ 3, 246, 47 },
		{ 3, 247, 47 },
		{ 3, 248, 47 },
		{ 3, 249, 47 },
		{ 3, 250, 47 },
		{ 3, 251, 47 },
		{ 3, 252, 47 },
		{ 3, 253, 47 },
		{ 3, 254, 47 },
		{ 3, 255, 47 },
		{ 3, 256, 47 },
		{ 4, 194, 47 },
		{ 4, 195, 47 },
		{ 4, 196, 47 },
		{ 4, 197, 47 },
		{ 4, 198, 47 },
		{ 4, 199, 47 },
		{ 4, 200, 47 },
		{ 4, 201, 47 },
		{ 4, 202, 47 },
		{ 4, 203, 47 },
		{ 4, 204, 47 },
		{ 4, 205, 47 },
		{ 4, 206, 47 },
		{ 4, 207, 47 },
		{ 4, 208, 47 },
		{ 4, 209, 47 },
		{ 4, 210, 47 },
		{ 4, 211, 47 },
		{ 4, 212, 47 },
		{ 4, 213, 47 },
		{ 4, 214, 47 },
		{ 4, 215, 47 },
		{ 4, 216, 47 },
		{ 4, 217, 47 },
		{ 4, 218, 47 },
		{ 4, 219, 47 },
		{ 4, 220, 47 },
		{ 4, 221, 47 },
		{ 4, 222, 47 },
		{ 4, 223, 47 },
		{ 4, 224, 47 },
		{ 4, 225, 47 },
		{ 4, 226, 47 },
		{ 4, 227, 47 },
		{ 4, 228, 47 },
		{ 4, 229, 47 },
		{ 4, 230, 47 },
		{ 4, 231, 47 },
		{ 4, 232, 47 },
		{ 4, 233, 47 },
		{ 4, 234, 47 },
		{ 4, 235, 47 },
		{ 4, 236, 47 },
		{ 4, 237, 47 },
		{ 4, 238, 47 },
		{ 4, 239, 47 },
		{ 4, 240, 47 },
		{ 4, 241, 47 },
		{ 4, 242, 47 },
		{ 4, 243, 47 },
		{ 4, 244, 47 },
		{ 4, 245, 47 },
		{ 4, 246, 47 },
		{ 4, 247, 47 },
		{ 4, 248, 47 },
		{ 4, 249, 47 },
		{ 4, 250, 47 },
		{ 4, 251, 47 },
		{ 4, 252, 47 },
		{ 4, 253, 47 },
		{ 4, 254, 47 },
		{ 4, 255, 47 },
		{ 4, 256, 47 },
		{ 5, 195, 47 },
		{ 5, 196, 47 },
		{ 5, 197, 47 },
		{ 5, 198, 47 },
		{ 5, 199, 47 },
		{ 5, 200, 47 },
		{ 5, 201, 47 },
		{ 5, 202, 47 },
		{ 5, 203, 47 },
		{ 5, 204, 47 },
		{ 5, 205, 47 },
		{ 5, 206, 47 },
		{ 5, 207, 47 },
		{ 5, 208, 47 },
		{ 5, 209, 47 },
		{ 5, 210, 47 },
		{ 5, 211, 47 },
		{ 5, 212, 47 },
		{ 5, 213, 47 },
		{ 5, 214, 47 },
		{ 5, 215, 47 },
		{ 5, 216, 47 },
		{ 5, 217, 47 },
		{ 5, 218, 47 },
		{ 5, 219, 47 },
		{ 5, 220, 47 },
		{ 5, 221, 47 },
		{ 5, 222, 47 },
		{ 5, 223, 47 },
		{ 5, 224, 47 },
		{ 5, 225, 47 },
		{ 5, 226, 47 },
		{ 5, 227, 47 },
		{ 5, 228, 47 },
		{ 5, 229, 47 },
		{ 5, 230, 47 },
		{ 5, 231, 47 },
		{ 5, 232, 47 },
		{ 5, 233, 47 },
		{ 5, 234, 47 },
		{ 5, 235, 47 },
		{ 5, 236, 47 },
		{ 5, 237, 47 },
		{ 5, 238, 47 },
		{ 5, 239, 47 },
		{ 5, 240, 47 },
		{ 5, 241, 47 },
		{ 5, 242, 47 },
		{ 5, 243, 47 },
		{ 5, 244, 47 },
		{ 5, 245, 47 },
		{ 5, 246, 47 },
		{ 5, 247, 47 },
		{ 5, 248, 47 },
		{ 5, 249, 47 },
		{ 5, 250, 47 },
		{ 5, 251, 47 },
		{ 5, 252, 47 },
		{ 5, 253, 47 },
		{ 5, 254, 47 },
		{ 5, 255, 47 },
		{ 5, 256, 47 },
		{ 6, 195, 47 },
		{ 6, 196, 47 },
		{ 6, 197, 47 },
		{ 6, 198, 47 },
		{ 6, 199, 47 },
		{ 6, 200, 47 },
		{ 6, 201, 47 },
		{ 6, 202, 47 },
		{ 6, 203, 47 },
		{ 6, 204, 47 },
		{ 6, 205, 47 },
		{ 6, 206, 47 },
		{ 6, 207, 47 },
		{ 6, 208, 47 },
		{ 6, 209, 47 },
		{ 6, 210, 47 },
		{ 6, 211, 47 },
		{ 6, 212, 47 },
		{ 6, 213, 47 },
		{ 6, 214, 47 },
		{ 6, 215, 47 },
		{ 6, 216, 47 },
		{ 6, 217, 47 },
		{ 6, 218, 47 },
		{ 6, 219, 47 },
		{ 6, 220, 47 },
		{ 6, 221, 47 },
		{ 6, 222, 47 },
		{ 6, 223, 47 },
		{ 6, 224, 47 },
		{ 6, 225, 47 },
		{ 6, 226, 47 },
		{ 6, 227, 47 },
		{ 6, 228, 47 },
		{ 6, 229, 47 },
		{ 6, 230, 47 },
		{ 6, 231, 47 },
		{ 6, 232, 47 },
		{ 6, 233, 47 },
		{ 6, 234, 47 },
		{ 6, 235, 47 },
		{ 6, 236, 47 },
		{ 6, 237, 47 },
		{ 6, 238, 47 },
		{ 6, 239, 47 },
		{ 6, 240, 47 },
		{ 6, 241, 47 },
		{ 6, 242, 47 },
		{ 6, 243, 47 },
		{ 6, 244, 47 },
		{ 6, 245, 47 },
		{ 6, 246, 47 },
		{ 6, 247, 47 },
		{ 6, 248, 47 },
		{ 6, 249, 47 },
		{ 6, 250, 47 },
		{ 6, 251, 47 },
		{ 6, 252, 47 },
		{ 6, 253, 47 },
		{ 6, 254, 47 },
		{ 6, 255, 47 },
		{ 6, 256, 47 },
		{ 7, 195, 47 },
		{ 7, 196, 47 },
		{ 7, 197, 47 },
		{ 7, 198, 47 },
		{ 7, 199, 47 },
		{ 7, 200, 47 },
		{ 7, 201, 47 },
		{ 7, 202, 47 },
		{ 7, 203, 47 },
		{ 7, 204, 47 },
		{ 7, 205, 47 },
		{ 7, 206, 47 },
		{ 7, 207, 47 },
		{ 7, 208, 47 },
		{ 7, 209, 47 },
		{ 7, 210, 47 },
		{ 7, 211, 47 },
		{ 7, 212, 47 },
		{ 7, 213, 47 },
		{ 7, 214, 47 },
		{ 7, 215, 47 },
		{ 7, 216, 47 },
		{ 7, 217, 47 },
		{ 7, 218, 47 },
		{ 7, 219, 47 },
		{ 7, 220, 47 },
		{ 7, 221, 47 },
		{ 7, 222, 47 },
		{ 7, 223, 47 },
		{ 7, 224, 47 },
		{ 7, 225, 47 },
		{ 7, 226, 47 },
		{ 7, 227, 47 },
		{ 7, 228, 47 },
		{ 7, 229, 47 },
		{ 7, 230, 47 },
		{ 7, 231, 47 },
		{ 7, 232, 47 },
		{ 7, 233, 47 },
		{ 7, 234, 47 },
		{ 7, 235, 47 },
		{ 7, 236, 47 },
		{ 7, 237, 47 },
		{ 7, 238, 47 },
		{ 7, 239, 47 },
		{ 7, 240, 47 },
		{ 7, 241, 47 },
		{ 7, 242, 47 },
		{ 7, 243, 47 },
		{ 7, 244, 47 },
		{ 7, 245, 47 },
		{ 7, 246, 47 },
		{ 7, 247, 47 },
		{ 7, 248, 47 },
		{ 7, 249, 47 },
		{ 7, 250, 47 },
		{ 7, 251, 47 },
		{ 7, 252, 47 },
		{ 7, 253, 47 },
		{ 7, 254, 47 },
		{ 7, 255, 47 },
		{ 7, 256, 47 },
		{ 8, 195, 47 },
		{ 8, 196, 47 },
		{ 8, 197, 47 },
		{ 8, 198, 47 },
		{ 8, 199, 47 },
		{ 8, 200, 47 },
		{ 8, 201, 47 },
		{ 8, 202, 47 },
		{ 8, 203, 47 },
		{ 8, 204, 47 },
		{ 8, 205, 47 },
		{ 8, 206, 47 },
		{ 8, 207, 47 },
		{ 8, 208, 47 },
		{ 8, 209, 47 },
		{ 8, 210, 47 },
		{ 8, 211, 47 },
		{ 8, 212, 47 },
		{ 8, 213, 47 },
		{ 8, 214, 47 },
		{ 8, 215, 47 },
		{ 8, 216, 47 },
		{ 8, 217, 47 },
		{ 8, 218, 47 },
		{ 8, 219, 47 },
		{ 8, 220, 47 },
		{ 8, 221, 47 },
		{ 8, 222, 47 },
		{ 8, 223, 47 },
		{ 8, 224, 47 },
		{ 8, 225, 47 },
		{ 8, 226, 47 },
		{ 8, 227, 47 },
		{ 8, 228, 47 },
		{ 8, 229, 47 },
		{ 8, 230, 47 },
		{ 8, 231, 47 },
		{ 8, 232, 47 },
		{ 8, 233, 47 },
		{ 8, 234, 47 },
		{ 8, 235, 47 },
		{ 8, 236, 47 },
		{ 8, 237, 47 },
		{ 8, 238, 47 },
		{ 8, 239, 47 },
		{ 8, 240, 47 },
		{ 8, 241, 47 },
		{ 8, 242, 47 },
		{ 8, 243, 47 },
		{ 8, 244, 47 },
		{ 8, 245, 47 },
		{ 8, 246, 47 },
		{ 8, 247, 47 },
		{ 8, 248, 47 },
		{ 8, 249, 47 },
		{ 8, 250, 47 },
		{ 8, 251, 47 },
		{ 8, 252, 47 },
		{ 8, 253, 47 },
		{ 8, 254, 47 },
		{ 8, 255, 47 },
		{ 8, 256, 47 },
		{ 9, 196, 47 },
		{ 9, 197, 47 },
		{ 9, 198, 47 },
		{ 9, 199, 47 },
		{ 9, 200, 47 },
		{ 9, 201, 47 },
		{ 9, 202, 47 },
		{ 9, 203, 47 },
		{ 9, 204, 47 },
		{ 9, 205, 47 },
		{ 9, 206, 47 },
		{ 9, 207, 47 },
		{ 9, 208, 47 },
		{ 9, 209, 47 },
		{ 9, 210, 47 },
		{ 9, 211, 47 },
		{ 9, 212, 47 },
		{ 9, 213, 47 },
		{ 9, 214, 47 },
		{ 9, 215, 47 },
		{ 9, 216, 47 },
		{ 9, 217, 47 },
		{ 9, 218, 47 },
		{ 9, 219, 47 },
		{ 9, 220, 47 },
		{ 9, 221, 47 },
		{ 9, 222, 47 },
		{ 9, 223, 47 },
		{ 9, 224, 47 },
		{ 9, 225, 47 },
		{ 9, 226, 47 },
		{ 9, 227, 47 },
		{ 9, 228, 47 },
		{ 9, 229, 47 },
		{ 9, 230, 47 },
		{ 9, 231, 47 },
		{ 9, 232, 47 },
		{ 9, 233, 47 },
		{ 9, 234, 47 },
		{ 9, 235, 47 },
		{ 9, 236, 47 },
		{ 9, 237, 47 },
		{ 9, 238, 47 },
		{ 9, 239, 47 },
		{ 9, 240, 47 },
		{ 9, 241, 47 },
		{ 9, 242, 47 },
		{ 9, 243, 47 },
		{ 9, 244, 47 },
		{ 9, 245, 47 },
		{ 9, 246, 47 },
		{ 9, 247, 47 },
		{ 9, 248, 47 },
		{ 9, 249, 47 },
		{ 9, 250, 47 },
		{ 9, 251, 47 },
		{ 9, 252, 47 },
		{ 9, 253, 47 },
		{ 9, 254, 47 },
		{ 9, 255, 47 },
		{ 9, 256, 47 },
		{ 10, 196, 47 },
		{ 10, 197, 47 },
		{ 10, 198, 47 },
		{ 10, 199, 47 },
		{ 10, 200, 47 },
		{ 10, 201, 47 },
		{ 10, 202, 47 },
		{ 10, 203, 47 },
		{ 10, 204, 47 },
		{ 10, 205, 47 },
		{ 10, 206, 47 },
		{ 10, 207, 47 },
		{ 10, 208, 47 },
		{ 10, 209, 47 },
		{ 10, 210, 47 },
		{ 10, 211, 47 },
		{ 10, 212, 47 },
		{ 10, 213, 47 },
		{ 10, 214, 47 },
		{ 10, 215, 47 },
		{ 10, 216, 47 },
		{ 10, 217, 47 },
		{ 10, 218, 47 },
		{ 10, 219, 47 },
		{ 10, 220, 47 },
		{ 10, 221, 47 },
		{ 10, 222, 47 },
		{ 10, 223, 47 },
		{ 10, 224, 47 },
		{ 10, 225, 47 },
		{ 10, 226, 47 },
		{ 10, 227, 47 },
		{ 10, 228, 47 },
		{ 10, 229, 47 },
		{ 10, 230, 47 },
		{ 10, 231, 47 },
		{ 10, 232, 47 },
		{ 10, 233, 47 },
		{ 10, 234, 47 },
		{ 10, 235, 47 },
		{ 10, 236, 47 },
		{ 10, 237, 47 },
		{ 10, 238, 47 },
		{ 10, 239, 47 },
		{ 10, 240, 47 },
		{ 10, 241, 47 },
		{ 10, 242, 47 },
		{ 10, 243, 47 },
		{ 10, 244, 47 },
		{ 10, 245, 47 },
		{ 10, 246, 47 },
		{ 10, 247, 47 },
		{ 10, 248, 47 },
		{ 10, 249, 47 },
		{ 10, 250, 47 },
		{ 10, 251, 47 },
		{ 10, 252, 47 },
		{ 10, 253, 47 },
		{ 10, 254, 47 },
		{ 10, 255, 47 },
		{ 10, 256, 47 },
		{ 11, 196, 47 },
		{ 11, 197, 47 },
		{ 11, 198, 47 },
		{ 11, 199, 47 },
		{ 11, 200, 47 },
		{ 11, 201, 47 },
		{ 11, 202, 47 },
		{ 11, 203, 47 },
		{ 11, 204, 47 },
		{ 11, 205, 47 },
		{ 11, 206, 47 },
		{ 11, 207, 47 },
		{ 11, 208, 47 },
		{ 11, 209, 47 },
		{ 11, 210, 47 },
		{ 11, 211, 47 },
		{ 11, 212, 47 },
		{ 11, 213, 47 },
		{ 11, 214, 47 },
		{ 11, 215, 47 },
		{ 11, 216, 47 },
		{ 11, 217, 47 },
		{ 11, 218, 47 },
		{ 11, 219, 47 },
		{ 11, 220, 47 },
		{ 11, 221, 47 },
		{ 11, 222, 47 },
		{ 11, 223, 47 },
		{ 11, 224, 47 },
		{ 11, 225, 47 },
		{ 11, 226, 47 },
		{ 11, 227, 47 },
		{ 11, 228, 47 },
		{ 11, 229, 47 },
		{ 11, 230, 47 },
		{ 11, 231, 47 },
		{ 11, 232, 47 },
		{ 11, 233, 47 },
		{ 11, 234, 47 },
		{ 11, 235, 47 },
		{ 11, 236, 47 },
		{ 11, 237, 47 },
		{ 11, 238, 47 },
		{ 11, 239, 47 },
		{ 11, 240, 47 },
		{ 11, 241, 47 },
		{ 11, 242, 47 },
		{ 11, 243, 47 },
		{ 11, 244, 47 },
		{ 11, 245, 47 },
		{ 11, 246, 47 },
		{ 11, 247, 47 },
		{ 11, 248, 47 },
		{ 11, 249, 47 },
		{ 11, 250, 47 },
		{ 11, 251, 47 },
		{ 11, 252, 47 },
		{ 11, 253, 47 },
		{ 11, 254, 47 },
		{ 11, 255, 47 },
		{ 11, 256, 47 },
		{ 12, 196, 47 },
		{ 12, 197, 47 },
		{ 12, 198, 47 },
		{ 12, 199, 47 },
		{ 12, 200, 47 },
		{ 12, 201, 47 },
		{ 12, 202, 47 },
		{ 12, 203, 47 },
		{ 12, 204, 47 },
		{ 12, 205, 47 },
		{ 12, 206, 47 },
		{ 12, 207, 47 },
		{ 12, 208, 47 },
		{ 12, 209, 47 },
		{ 12, 210, 47 },
		{ 12, 211, 47 },
		{ 12, 212, 47 },
		{ 12, 213, 47 },
		{ 12, 214, 47 },
		{ 12, 215, 47 },
		{ 12, 216, 47 },
		{ 12, 217, 47 },
		{ 12, 218, 47 },
		{ 12, 219, 47 },
		{ 12, 220, 47 },
		{ 12, 221, 47 },
		{ 12, 222, 47 },
		{ 12, 223, 47 },
		{ 12, 224, 47 },
		{ 12, 225, 47 },
		{ 12, 226, 47 },
		{ 12, 227, 47 },
		{ 12, 228, 47 },
		{ 12, 229, 47 },
		{ 12, 230, 47 },
		{ 12, 231, 47 },
		{ 12, 232, 47 },
		{ 12, 233, 47 },
		{ 12, 234, 47 },
		{ 12, 235, 47 },
		{ 12, 236, 47 },
		{ 12, 237, 47 },
		{ 12, 238, 47 },
		{ 12, 239, 47 },
		{ 12, 240, 47 },
		{ 12, 241, 47 },
		{ 12, 242, 47 },
		{ 12, 243, 47 },
		{ 12, 244, 47 },
		{ 12, 245, 47 },
		{ 12, 246, 47 },
		{ 12, 247, 47 },
		{ 12, 248, 47 },
		{ 12, 249, 47 },
		{ 12, 250, 47 },
		{ 12, 251, 47 },
		{ 12, 252, 47 },
		{ 12, 253, 47 },
		{ 12, 254, 47 },
		{ 12, 255, 47 },
		{ 12, 256, 47 },
		{ 13, 197, 47 },
		{ 13, 198, 47 },
		{ 13, 199, 47 },
		{ 13, 200, 47 },
		{ 13, 201, 47 },
		{ 13, 202, 47 },
		{ 13, 203, 47 },
		{ 13, 204, 47 },
		{ 13, 205, 47 },
		{ 13, 206, 47 },
		{ 13, 207, 47 },
		{ 13, 208, 47 },
		{ 13, 209, 47 },
		{ 13, 210, 47 },
		{ 13, 211, 47 },
		{ 13, 212, 47 },
		{ 13, 213, 47 },
		{ 13, 214, 47 },
		{ 13, 215, 47 },
		{ 13, 216, 47 },
		{ 13, 217, 47 },
		{ 13, 218, 47 },
		{ 13, 219, 47 },
		{ 13, 220, 47 },
		{ 13, 221, 47 },
		{ 13, 222, 47 },
		{ 13, 223, 47 },
		{ 13, 224, 47 },
		{ 13, 225, 47 },
		{ 13, 226, 47 },
		{ 13, 227, 47 },
		{ 13, 228, 47 },
		{ 13, 229, 47 },
		{ 13, 230, 47 },
		{ 13, 231, 47 },
		{ 13, 232, 47 },
		{ 13, 233, 47 },
		{ 13, 234, 47 },
		{ 13, 235, 47 },
		{ 13, 236, 47 },
		{ 13, 237, 47 },
		{ 13, 238, 47 },
		{ 13, 239, 47 },
		{ 13, 240, 47 },
		{ 13, 241, 47 },
		{ 13, 242, 47 },
		{ 13, 243, 47 },
		{ 13, 244, 47 },
		{ 13, 245, 47 },
		{ 13, 246, 47 },
		{ 13, 247, 47 },
		{ 13, 248, 47 },
		{ 13, 249, 47 },
		{ 13, 250, 47 },
		{ 13, 251, 47 },
		{ 13, 252, 47 },
		{ 13, 253, 47 },
		{ 13, 254, 47 },
		{ 13, 255, 47 },
		{ 13, 256, 47 },
		{ 14, 197, 47 },
		{ 14, 198, 47 },
		{ 14, 199, 47 },
		{ 14, 200, 47 },
		{ 14, 201, 47 },
		{ 14, 202, 47 },
		{ 14, 203, 47 },
		{ 14, 204, 47 },
		{ 14, 205, 47 },
		{ 14, 206, 47 },
		{ 14, 207, 47 },
		{ 14, 208, 47 },
		{ 14, 209, 47 },
		{ 14, 210, 47 },
		{ 14, 211, 47 },
		{ 14, 212, 47 },
		{ 14, 213, 47 },
		{ 14, 214, 47 },
		{ 14, 215, 47 },
		{ 14, 216, 47 },
		{ 14, 217, 47 },
		{ 14, 218, 47 },
		{ 14, 219, 47 },
		{ 14, 220, 47 },
		{ 14, 221, 47 },
		{ 14, 222, 47 },
		{ 14, 223, 47 },
		{ 14, 224, 47 },
		{ 14, 225, 47 },
		{ 14, 226, 47 },
		{ 14, 227, 47 },
		{ 14, 228, 47 },
		{ 14, 229, 47 },
		{ 14, 230, 47 },
		{ 14, 231, 47 },
		{ 14, 232, 47 },
		{ 14, 233, 47 },
		{ 14, 234, 47 },
		{ 14, 235, 47 },
		{ 14, 236, 47 },
		{ 14, 237, 47 },
		{ 14, 238, 47 },
		{ 14, 239, 47 },
		{ 14, 240, 47 },
		{ 14, 241, 47 },
		{ 14, 242, 47 },
		{ 14, 243, 47 },
		{ 14, 244, 47 },
		{ 14, 245, 47 },
		{ 14, 246, 47 },
		{ 14, 247, 47 },
		{ 14, 248, 47 },
		{ 14, 249, 47 },
		{ 14, 250, 47 },
		{ 14, 251, 47 },
		{ 14, 252, 47 },
		{ 14, 253, 47 },
		{ 14, 254, 47 },
		{ 14, 255, 47 },
		{ 14, 256, 47 },
		{ 15, 197, 47 },
		{ 15, 198, 47 },
		{ 15, 199, 47 },
		{ 15, 200, 47 },
		{ 15, 201, 47 },
		{ 15, 202, 47 },
		{ 15, 203, 47 },
		{ 15, 204, 47 },
		{ 15, 205, 47 },
		{ 15, 206, 47 },
		{ 15, 207, 47 },
		{ 15, 208, 47 },
		{ 15, 209, 47 },
		{ 15, 210, 47 },
		{ 15, 211, 47 },
		{ 15, 212, 47 },
		{ 15, 213, 47 },
		{ 15, 214, 47 },
		{ 15, 215, 47 },
		{ 15, 216, 47 },
		{ 15, 217, 47 },
		{ 15, 218, 47 },
		{ 15, 219, 47 },
		{ 15, 220, 47 },
		{ 15, 221, 47 },
		{ 15, 222, 47 },
		{ 15, 223, 47 },
		{ 15, 224, 47 },
		{ 15, 225, 47 },
		{ 15, 226, 47 },
		{ 15, 227, 47 },
		{ 15, 228, 47 },
		{ 15, 229, 47 },
		{ 15, 230, 47 },
		{ 15, 231, 47 },
		{ 15, 232, 47 },
		{ 15, 233, 47 },
		{ 15, 234, 47 },
		{ 15, 235, 47 },
		{ 15, 236, 47 },
		{ 15, 237, 47 },
		{ 15, 238, 47 },
		{ 15, 239, 47 },
		{ 15, 240, 47 },
		{ 15, 241, 47 },
		{ 15, 242, 47 },
		{ 15, 243, 47 },
		{ 15, 244, 47 },
		{ 15, 245, 47 },
		{ 15, 246, 47 },
		{ 15, 247, 47 },
		{ 15, 248, 47 },
		{ 15, 249, 47 },
		{ 15, 250, 47 },
		{ 15, 251, 47 },
		{ 15, 252, 47 },
		{ 15, 253, 47 },
		{ 15, 254, 47 },
		{ 15, 255, 47 },
		{ 15, 256, 47 },
		{ 16, 197, 47 },
		{ 16, 198, 47 },
		{ 16, 199, 47 },
		{ 16, 200, 47 },
		{ 16, 201, 47 },
		{ 16, 202, 47 },
		{ 16, 203, 47 },
		{ 16, 204, 47 },
		{ 16, 205, 47 },
		{ 16, 206, 47 },
		{ 16, 207, 47 },
		{ 16, 208, 47 },
		{ 16, 209, 47 },
		{ 16, 210, 47 },
		{ 16, 211, 47 },
		{ 16, 212, 47 },
		{ 16, 213, 47 },
		{ 16, 214, 47 },
		{ 16, 215, 47 },
		{ 16, 216, 47 },
		{ 16, 217, 47 },
		{ 16, 218, 47 },
		{ 16, 219, 47 },
		{ 16, 220, 47 },
		{ 16, 221, 47 },
		{ 16, 222, 47 },
		{ 16, 223, 47 },
		{ 16, 224, 47 },
		{ 16, 225, 47 },
		{ 16, 226, 47 },
		{ 16, 227, 47 },
		{ 16, 228, 47 },
		{ 16, 229, 47 },
		{ 16, 230, 47 },
		{ 16, 231, 47 },
		{ 16, 232, 47 },
		{ 16, 233, 47 },
		{ 16, 234, 47 },
		{ 16, 235, 47 },
		{ 16, 236, 47 },
		{ 16, 237, 47 },
		{ 16, 238, 47 },
		{ 16, 239, 47 },
		{ 16, 240, 47 },
		{ 16, 241, 47 },
		{ 16, 242, 47 },
		{ 16, 243, 47 },
		{ 16, 244, 47 },
		{ 16, 245, 47 },
		{ 16, 246, 47 },
		{ 16, 247, 47 },
		{ 16, 248, 47 },
		{ 16, 249, 47 },
		{ 16, 250, 47 },
		{ 16, 251, 47 },
		{ 16, 252, 47 },
		{ 16, 253, 47 },
		{ 16, 254, 47 },
		{ 16, 255, 47 },
		{ 16, 256, 47 },
		{ 17, 198, 47 },
		{ 17, 199, 47 },
		{ 17, 200, 47 },
		{ 17, 201, 47 },
		{ 17, 202, 47 },
		{ 17, 203, 47 },
		{ 17, 204, 47 },
		{ 17, 205, 47 },
		{ 17, 206, 47 },
		{ 17, 207, 47 },
		{ 17, 208, 47 },
		{ 17, 209, 47 },
		{ 17, 210, 47 },
		{ 17, 211, 47 },
		{ 17, 212, 47 },
		{ 17, 213, 47 },
		{ 17, 214, 47 },
		{ 17, 215, 47 },
		{ 17, 216, 47 },
		{ 17, 217, 47 },
		{ 17, 218, 47 },
		{ 17, 219, 47 },
		{ 17, 220, 47 },
		{ 17, 221, 47 },
		{ 17, 222, 47 },
		{ 17, 223, 47 },
		{ 17, 224, 47 },
		{ 17, 225, 47 },
		{ 17, 226, 47 },
		{ 17, 227, 47 },
		{ 17, 228, 47 },
		{ 17, 229, 47 },
		{ 17, 230, 47 },
		{ 17, 231, 47 },
		{ 17, 232, 47 },
		{ 17, 233, 47 },
		{ 17, 234, 47 },
		{ 17, 235, 47 },
		{ 17, 236, 47 },
		{ 17, 237, 47 },
		{ 17, 238, 47 },
		{ 17, 239, 47 },
		{ 17, 240, 47 },
		{ 17, 241, 47 },
		{ 17, 242, 47 },
		{ 17, 243, 47 },
		{ 17, 244, 47 },
		{ 17, 245, 47 },
		{ 17, 246, 47 },
		{ 17, 247, 47 },
		{ 17, 248, 47 },
		{ 17, 249, 47 },
		{ 17, 250, 47 },
		{ 17, 251, 47 },
		{ 17, 252, 47 },
		{ 17, 253, 47 },
		{ 17, 254, 47 },
		{ 17, 255, 47 },
		{ 17, 256, 47 },
		{ 18, 198, 47 },
		{ 18, 199, 47 },
		{ 18, 200, 47 },
		{ 18, 201, 47 },
		{ 18, 202, 47 },
		{ 18, 203, 47 },
		{ 18, 204, 47 },
		{ 18, 205, 47 },
		{ 18, 206, 47 },
		{ 18, 207, 47 },
		{ 18, 208, 47 },
		{ 18, 209, 47 },
		{ 18, 210, 47 },
		{ 18, 211, 47 },
		{ 18, 212, 47 },
		{ 18, 213, 47 },
		{ 18, 214, 47 },
		{ 18, 215, 47 },
		{ 18, 216, 47 },
		{ 18, 217, 47 },
		{ 18, 218, 47 },
		{ 18, 219, 47 },
		{ 18, 220, 47 },
		{ 18, 221, 47 },
		{ 18, 222, 47 },
		{ 18, 223, 47 },
		{ 18, 224, 47 },
		{ 18, 225, 47 },
		{ 18, 226, 47 },
		{ 18, 227, 47 },
		{ 18, 228, 47 },
		{ 18, 229, 47 },
		{ 18, 230, 47 },
		{ 18, 231, 47 },
		{ 18, 232, 47 },
		{ 18, 233, 47 },
		{ 18, 234, 47 },
		{ 18, 235, 47 },
		{ 18, 236, 47 },
		{ 18, 237, 47 },
		{ 18, 238, 47 },
		{ 18, 239, 47 },
		{ 18, 240, 47 },
		{ 18, 241, 47 },
		{ 18, 242, 47 },
		{ 18, 243, 47 },
		{ 18, 244, 47 },
		{ 18, 245, 47 },
		{ 18, 246, 47 },
		{ 18, 247, 47 },
		{ 18, 248, 47 },
		{ 18, 249, 47 },
		{ 18, 250, 47 },
		{ 18, 251, 47 },
		{ 18, 252, 47 },
		{ 18, 253, 47 },
		{ 18, 254, 47 },
		{ 18, 255, 47 },
		{ 18, 256, 47 },
		{ 19, 198, 47 },
		{ 19, 199, 47 },
		{ 19, 200, 47 },
		{ 19, 201, 47 },
		{ 19, 202, 47 },
		{ 19, 203, 47 },
		{ 19, 204, 47 },
		{ 19, 205, 47 },
		{ 19, 206, 47 },
		{ 19, 207, 47 },
		{ 19, 208, 47 },
		{ 19, 209, 47 },
		{ 19, 210, 47 },
		{ 19, 211, 47 },
		{ 19, 212, 47 },
		{ 19, 213, 47 },
		{ 19, 214, 47 },
		{ 19, 215, 47 },
		{ 19, 216, 47 },
		{ 19, 217, 47 },
		{ 19, 218, 47 },
		{ 19, 219, 47 },
		{ 19, 220, 47 },
		{ 19, 221, 47 },
		{ 19, 222, 47 },
		{ 19, 223, 47 },
		{ 19, 224, 47 },
		{ 19, 225, 47 },
		{ 19, 226, 47 },
		{ 19, 227, 47 },
		{ 19, 228, 47 },
		{ 19, 229, 47 },
		{ 19, 230, 47 },
		{ 19, 231, 47 },
		{ 19, 232, 47 },
		{ 19, 233, 47 },
		{ 19, 234, 47 },
		{ 19, 235, 47 },
		{ 19, 236, 47 },
		{ 19, 237, 47 },
		{ 19, 238, 47 },
		{ 19, 239, 47 },
		{ 19, 240, 47 },
		{ 19, 241, 47 },
		{ 19, 242, 47 },
		{ 19, 243, 47 },
		{ 19, 244, 47 },
		{ 19, 245, 47 },
		{ 19, 246, 47 },
		{ 19, 247, 47 },
		{ 19, 248, 47 },
		{ 19, 249, 47 },
		{ 19, 250, 47 },
		{ 19, 251, 47 },
		{ 19, 252, 47 },
		{ 19, 253, 47 },
		{ 19, 254, 47 },
		{ 19, 255, 47 },
		{ 19, 256, 47 },
		{ 20, 198, 47 },
		{ 20, 199, 47 },
		{ 20, 200, 47 },
		{ 20, 201, 47 },
		{ 20, 202, 47 },
		{ 20, 203, 47 },
		{ 20, 204, 47 },
		{ 20, 205, 47 },
		{ 20, 206, 47 },
		{ 20, 207, 47 },
		{ 20, 208, 47 },
		{ 20, 209, 47 },
		{ 20, 210, 47 },
		{ 20, 211, 47 },
		{ 20, 212, 47 },
		{ 20, 213, 47 },
		{ 20, 214, 47 },
		{ 20, 215, 47 },
		{ 20, 216, 47 },
		{ 20, 217, 47 },
		{ 20, 218, 47 },
		{ 20, 219, 47 },
		{ 20, 220, 47 },
		{ 20, 221, 47 },
		{ 20, 222, 47 },
		{ 20, 223, 47 },
		{ 20, 224, 47 },
		{ 20, 225, 47 },
		{ 20, 226, 47 },
		{ 20, 227, 47 },
		{ 20, 228, 47 },
		{ 20, 229, 47 },
		{ 20, 230, 47 },
		{ 20, 231, 47 },
		{ 20, 232, 47 },
		{ 20, 233, 47 },
		{ 20, 234, 47 },
		{ 20, 235, 47 },
		{ 20, 236, 47 },
		{ 20, 237, 47 },
		{ 20, 238, 47 },
		{ 20, 239, 47 },
		{ 20, 240, 47 },
		{ 20, 241, 47 },
		{ 20, 242, 47 },
		{ 20, 243, 47 },
		{ 20, 244, 47 },
		{ 20, 245, 47 },
		{ 20, 246, 47 },
		{ 20, 247, 47 },
		{ 20, 248, 47 },
		{ 20, 249, 47 },
		{ 20, 250, 47 },
		{ 20, 251, 47 },
		{ 20, 252, 47 },
		{ 20, 253, 47 },
		{ 20, 254, 47 },
		{ 20, 255, 47 },
		{ 20, 256, 47 },
		{ 21, 199, 47 },
		{ 21, 200, 47 },
		{ 21, 201, 47 },
		{ 21, 202, 47 },
		{ 21, 203, 47 },
		{ 21, 204, 47 },
		{ 21, 205, 47 },
		{ 21, 206, 47 },
		{ 21, 207, 47 },
		{ 21, 208, 47 },
		{ 21, 209, 47 },
		{ 21, 210, 47 },
		{ 21, 211, 47 },
		{ 21, 212, 47 },
		{ 21, 213, 47 },
		{ 21, 214, 47 },
		{ 21, 215, 47 },
		{ 21, 216, 47 },
		{ 21, 217, 47 },
		{ 21, 218, 47 },
		{ 21, 219, 47 },
		{ 21, 220, 47 },
		{ 21, 221, 47 },
		{ 21, 222, 47 },
		{ 21, 223, 47 },
		{ 21, 224, 47 },
		{ 21, 225, 47 },
		{ 21, 226, 47 },
		{ 21, 227, 47 },
		{ 21, 228, 47 },
		{ 21, 229, 47 },
		{ 21, 230, 47 },
		{ 21, 231, 47 },
		{ 21, 232, 47 },
		{ 21, 233, 47 },
		{ 21, 234, 47 },
		{ 21, 235, 47 },
		{ 21, 236, 47 },
		{ 21, 237, 47 },
		{ 21, 238, 47 },
		{ 21, 239, 47 },
		{ 21, 240, 47 },
		{ 21, 241, 47 },
		{ 21, 242, 47 },
		{ 21, 243, 47 },
		{ 21, 244, 47 },
		{ 21, 245, 47 },
		{ 21, 246, 47 },
		{ 21, 247, 47 },
		{ 21, 248, 47 },
		{ 21, 249, 47 },
		{ 21, 250, 47 },
		{ 21, 251, 47 },
		{ 21, 252, 47 },
		{ 21, 253, 47 },
		{ 21, 254, 47 },
		{ 21, 255, 47 },
		{ 21, 256, 47 },
		{ 22, 199, 47 },
		{ 22, 200, 47 },
		{ 22, 201, 47 },
		{ 22, 202, 47 },
		{ 22, 203, 47 },
		{ 22, 204, 47 },
		{ 22, 205, 47 },
		{ 22, 206, 47 },
		{ 22, 207, 47 },
		{ 22, 208, 47 },
		{ 22, 209, 47 },
		{ 22, 210, 47 },
		{ 22, 211, 47 },
		{ 22, 212, 47 },
		{ 22, 213, 47 },
		{ 22, 214, 47 },
		{ 22, 215, 47 },
		{ 22, 216, 47 },
		{ 22, 217, 47 },
		{ 22, 218, 47 },
		{ 22, 219, 47 },
		{ 22, 220, 47 },
		{ 22, 221, 47 },
		{ 22, 222, 47 },
		{ 22, 223, 47 },
		{ 22, 224, 47 },
		{ 22, 225, 47 },
		{ 22, 226, 47 },
		{ 22, 227, 47 },
		{ 22, 228, 47 },
		{ 22, 229, 47 },
		{ 22, 230, 47 },
		{ 22, 231, 47 },
		{ 22, 232, 47 },
		{ 22, 233, 47 },
		{ 22, 234, 47 },
		{ 22, 235, 47 },
		{ 22, 236, 47 },
		{ 22, 237, 47 },
		{ 22, 238, 47 },
		{ 22, 239, 47 },
		{ 22, 240, 47 },
		{ 22, 241, 47 },
		{ 22, 242, 47 },
		{ 22, 243, 47 },
		{ 22, 244, 47 },
		{ 22, 245, 47 },
		{ 22, 246, 47 },
		{ 22, 247, 47 },
		{ 22, 248, 47 },
		{ 22, 249, 47 },
		{ 22, 250, 47 },
		{ 22, 251, 47 },
		{ 22, 252, 47 },
		{ 22, 253, 47 },
		{ 22, 254, 47 },
		{ 22, 255, 47 },
		{ 22, 256, 47 },
		{ 23, 199, 47 },
		{ 23, 200, 47 },
		{ 23, 201, 47 },
		{ 23, 202, 47 },
		{ 23, 203, 47 },
		{ 23, 204, 47 },
		{ 23, 205, 47 },
		{ 23, 206, 47 },
		{ 23, 207, 47 },
		{ 23, 208, 47 },
		{ 23, 209, 47 },
		{ 23, 210, 47 },
		{ 23, 211, 47 },
		{ 23, 212, 47 },
		{ 23, 213, 47 },
		{ 23, 214, 47 },
		{ 23, 215, 47 },
		{ 23, 216, 47 },
		{ 23, 217, 47 },
		{ 23, 218, 47 },
		{ 23, 219, 47 },
		{ 23, 220, 47 },
		{ 23, 221, 47 },
		{ 23, 222, 47 },
		{ 23, 223, 47 },
		{ 23, 224, 47 },
		{ 23, 225, 47 },
		{ 23, 226, 47 },
		{ 23, 227, 47 },
		{ 23, 228, 47 },
		{ 23, 229, 47 },
		{ 23, 230, 47 },
		{ 23, 231, 47 },
		{ 23, 232, 47 },
		{ 23, 233, 47 },
		{ 23, 234, 47 },
		{ 23, 235, 47 },
		{ 23, 236, 47 },
		{ 23, 237, 47 },
		{ 23, 238, 47 },
		{ 23, 239, 47 },
		{ 23, 240, 47 },
		{ 23, 241, 47 },
		{ 23, 242, 47 },
		{ 23, 243, 47 },
		{ 23, 244, 47 },
		{ 23, 245, 47 },
		{ 23, 246, 47 },
		{ 23, 247, 47 },
		{ 23, 248, 47 },
		{ 23, 249, 47 },
		{ 23, 250, 47 },
		{ 23, 251, 47 },
		{ 23, 252, 47 },
		{ 23, 253, 47 },
		{ 23, 254, 47 },
		{ 23, 255, 47 },
		{ 23, 256, 47 },
		{ 24, 199, 47 },
		{ 24, 200, 47 },
		{ 24, 201, 47 },
		{ 24, 202, 47 },
		{ 24, 203, 47 },
		{ 24, 204, 47 },
		{ 24, 205, 47 },
		{ 24, 206, 47 },
		{ 24, 207, 47 },
		{ 24, 208, 47 },
		{ 24, 209, 47 },
		{ 24, 210, 47 },
		{ 24, 211, 47 },
		{ 24, 212, 47 },
		{ 24, 213, 47 },
		{ 24, 214, 47 },
		{ 24, 215, 47 },
		{ 24, 216, 47 },
		{ 24, 217, 47 },
		{ 24, 218, 47 },
		{ 24, 219, 47 },
		{ 24, 220, 47 },
		{ 24, 221, 47 },
		{ 24, 222, 47 },
		{ 24, 223, 47 },
		{ 24, 224, 47 },
		{ 24, 225, 47 },
		{ 24, 226, 47 },
		{ 24, 227, 47 },
		{ 24, 228, 47 },
		{ 24, 229, 47 },
		{ 24, 230, 47 },
		{ 24, 231, 47 },
		{ 24, 232, 47 },
		{ 24, 233, 47 },
		{ 24, 234, 47 },
		{ 24, 235, 47 },
		{ 24, 236, 47 },
		{ 24, 237, 47 },
		{ 24, 238, 47 },
		{ 24, 239, 47 },
		{ 24, 240, 47 },
		{ 24, 241, 47 },
		{ 24, 242, 47 },
		{ 24, 243, 47 },
		{ 24, 244, 47 },
		{ 24, 245, 47 },
		{ 24, 246, 47 },
		{ 24, 247, 47 },
		{ 24, 248, 47 },
		{ 24, 249, 47 },
		{ 24, 250, 47 },
		{ 24, 251, 47 },
		{ 24, 252, 47 },
		{ 24, 253, 47 },
		{ 24, 254, 47 },
		{ 24, 255, 47 },
		{ 24, 256, 47 },
		{ 25, 200, 47 },
		{ 25, 201, 47 },
		{ 25, 202, 47 },
		{ 25, 203, 47 },
		{ 25, 204, 47 },
		{ 25, 205, 47 },
		{ 25, 206, 47 },
		{ 25, 207, 47 },
		{ 25, 208, 47 },
		{ 25, 209, 47 },
		{ 25, 210, 47 },
		{ 25, 211, 47 },
		{ 25, 212, 47 },
		{ 25, 213, 47 },
		{ 25, 214, 47 },
		{ 25, 215, 47 },
		{ 25, 216, 47 },
		{ 25, 217, 47 },
		{ 25, 218, 47 },
		{ 25, 219, 47 },
		{ 25, 220, 47 },
		{ 25, 221, 47 },
		{ 25, 222, 47 },
		{ 25, 223, 47 },
		{ 25, 224, 47 },
		{ 25, 225, 47 },
		{ 25, 226, 47 },
		{ 25, 227, 47 },
		{ 25, 228, 47 },
		{ 25, 229, 47 },
		{ 25, 230, 47 },
		{ 25, 231, 47 },
		{ 25, 232, 47 },
		{ 25, 233, 47 },
		{ 25, 234, 47 },
		{ 25, 235, 47 },
		{ 25, 236, 47 },
		{ 25, 237, 47 },
		{ 25, 238, 47 },
		{ 25, 239, 47 },
		{ 25, 240, 47 },
		{ 25, 241, 47 },
		{ 25, 242, 47 },
		{ 25, 243, 47 },
		{ 25, 244, 47 },
		{ 25, 245, 47 },
		{ 25, 246, 47 },
		{ 25, 247, 47 },
		{ 25, 248, 47 },
		{ 25, 249, 47 },
		{ 25, 250, 47 },
		{ 25, 251, 47 },
		{ 25, 252, 47 },
		{ 25, 253, 47 },
		{ 25, 254, 47 },
		{ 25, 255, 47 },
		{ 25, 256, 47 },
		{ 26, 200, 47 },
		{ 26, 201, 47 },
		{ 26, 202, 47 },
		{ 26, 203, 47 },
		{ 26, 204, 47 },
		{ 26, 205, 47 },
		{ 26, 206, 47 },
		{ 26, 207, 47 },
		{ 26, 208, 47 },
		{ 26, 209, 47 },
		{ 26, 210, 47 },
		{ 26, 211, 47 },
		{ 26, 212, 47 },
		{ 26, 213, 47 },
		{ 26, 214, 47 },
		{ 26, 215, 47 },
		{ 26, 216, 47 },
		{ 26, 217, 47 },
		{ 26, 218, 47 },
		{ 26, 219, 47 },
		{ 26, 220, 47 },
		{ 26, 221, 47 },
		{ 26, 222, 47 },
		{ 26, 223, 47 },
		{ 26, 224, 47 },
		{ 26, 225, 47 },
		{ 26, 226, 47 },
		{ 26, 227, 47 },
		{ 26, 228, 47 },
		{ 26, 229, 47 },
		{ 26, 230, 47 },
		{ 26, 231, 47 },
		{ 26, 232, 47 },
		{ 26, 233, 47 },
		{ 26, 234, 47 },
		{ 26, 235, 47 },
		{ 26, 236, 47 },
		{ 26, 237, 47 },
		{ 26, 238, 47 },
		{ 26, 239, 47 },
		{ 26, 240, 47 },
		{ 26, 241, 47 },
		{ 26, 242, 47 },
		{ 26, 243, 47 },
		{ 26, 244, 47 },
		{ 26, 245, 47 },
		{ 26, 246, 47 },
		{ 26, 247, 47 },
		{ 26, 248, 47 },
		{ 26, 249, 47 },
		{ 26, 250, 47 },
		{ 26, 251, 47 },
		{ 26, 252, 47 },
		{ 26, 253, 47 },
		{ 26, 254, 47 },
		{ 26, 255, 47 },
		{ 26, 256, 47 },
		{ 27, 200, 47 },
		{ 27, 201, 47 },
		{ 27, 202, 47 },
		{ 27, 203, 47 },
		{ 27, 204, 47 },
		{ 27, 205, 47 },
		{ 27, 206, 47 },
		{ 27, 207, 47 },
		{ 27, 208, 47 },
		{ 27, 209, 47 },
		{ 27, 210, 47 },
		{ 27, 211, 47 },
		{ 27, 212, 47 },
		{ 27, 213, 47 },
		{ 27, 214, 47 },
		{ 27, 215, 47 },
		{ 27, 216, 47 },
		{ 27, 217, 47 },
		{ 27, 218, 47 },
		{ 27, 219, 47 },
		{ 27, 220, 47 },
		{ 27, 221, 47 },
		{ 27, 222, 47 },
		{ 27, 223, 47 },
		{ 27, 224, 47 },
		{ 27, 225, 47 },
		{ 27, 226, 47 },
		{ 27, 227, 47 },
		{ 27, 228, 47 },
		{ 27, 229, 47 },
		{ 27, 230, 47 },
		{ 27, 231, 47 },
		{ 27, 232, 47 },
		{ 27, 233, 47 },
		{ 27, 234, 47 },
		{ 27, 235, 47 },
		{ 27, 236, 47 },
		{ 27, 237, 47 },
		{ 27, 238, 47 },
		{ 27, 239, 47 },
		{ 27, 240, 47 },
		{ 27, 241, 47 },
		{ 27, 242, 47 },
		{ 27, 243, 47 },
		{ 27, 244, 47 },
		{ 27, 245, 47 },
		{ 27, 246, 47 },
		{ 27, 247, 47 },
		{ 27, 248, 47 },
		{ 27, 249, 47 },
		{ 27, 250, 47 },
		{ 27, 251, 47 },
		{ 27, 252, 47 },
		{ 27, 253, 47 },
		{ 27, 254, 47 },
		{ 27, 255, 47 },
		{ 27, 256, 47 },
		{ 28, 200, 47 },
		{ 28, 201, 47 },
		{ 28, 202, 47 },
		{ 28, 203, 47 },
		{ 28, 204, 47 },
		{ 28, 205, 47 },
		{ 28, 206, 47 },
		{ 28, 207, 47 },
		{ 28, 208, 47 },
		{ 28, 209, 47 },
		{ 28, 210, 47 },
		{ 28, 211, 47 },
		{ 28, 212, 47 },
		{ 28, 213, 47 },
		{ 28, 214, 47 },
		{ 28, 215, 47 },
		{ 28, 216, 47 },
		{ 28, 217, 47 },
		{ 28, 218, 47 },
		{ 28, 219, 47 },
		{ 28, 220, 47 },
		{ 28, 221, 47 },
		{ 28, 222, 47 },
		{ 28, 223, 47 },
		{ 28, 224, 47 },
		{ 28, 225, 47 },
		{ 28, 226, 47 },
		{ 28, 227, 47 },
		{ 28, 228, 47 },
		{ 28, 229, 47 },
		{ 28, 230, 47 },
		{ 28, 231, 47 },
		{ 28, 232, 47 },
		{ 28, 233, 47 },
		{ 28, 234, 47 },
		{ 28, 235, 47 },
		{ 28, 236, 47 },
		{ 28, 237, 47 },
		{ 28, 238, 47 },
		{ 28, 239, 47 },
		{ 28, 240, 47 },
		{ 28, 241, 47 },
		{ 28, 242, 47 },
		{ 28, 243, 47 },
		{ 28, 244, 47 },
		{ 28, 245, 47 },
		{ 28, 246, 47 },
		{ 28, 247, 47 },
		{ 28, 248, 47 },
		{ 28, 249, 47 },
		{ 28, 250, 47 },
		{ 28, 251, 47 },
		{ 28, 252, 47 },
		{ 28, 253, 47 },
		{ 28, 254, 47 },
		{ 28, 255, 47 },
		{ 28, 256, 47 },
		{ 29, 201, 47 },
		{ 29, 202, 47 },
		{ 29, 203, 47 },
		{ 29, 204, 47 },
		{ 29, 205, 47 },
		{ 29, 206, 47 },
		{ 29, 207, 47 },
		{ 29, 208, 47 },
		{ 29, 209, 47 },
		{ 29, 210, 47 },
		{ 29, 211, 47 },
		{ 29, 212, 47 },
		{ 29, 213, 47 },
		{ 29, 214, 47 },
		{ 29, 215, 47 },
		{ 29, 216, 47 },
		{ 29, 217, 47 },
		{ 29, 218, 47 },
		{ 29, 219, 47 },
		{ 29, 220, 47 },
		{ 29, 221, 47 },
		{ 29, 222, 47 },
		{ 29, 223, 47 },
		{ 29, 224, 47 },
		{ 29, 225, 47 },
		{ 29, 226, 47 },
		{ 29, 227, 47 },
		{ 29, 228, 47 },
		{ 29, 229, 47 },
		{ 29, 230, 47 },
		{ 29, 231, 47 },
		{ 29, 232, 47 },
		{ 29, 233, 47 },
		{ 29, 234, 47 },
		{ 29, 235, 47 },
		{ 29, 236, 47 },
		{ 29, 237, 47 },
		{ 29, 238, 47 },
		{ 29, 239, 47 },
		{ 29, 240, 47 },
		{ 29, 241, 47 },
		{ 29, 242, 47 },
		{ 29, 243, 47 },
		{ 29, 244, 47 },
		{ 29, 245, 47 },
		{ 29, 246, 47 },
		{ 29, 247, 47 },
		{ 29, 248, 47 },
		{ 29, 249, 47 },
		{ 29, 250, 47 },
		{ 29, 251, 47 },
		{ 29, 252, 47 },
		{ 29, 253, 47 },
		{ 29, 254, 47 },
		{ 29, 255, 47 },
		{ 29, 256, 47 },
		{ 30, 201, 47 },
		{ 30, 202, 47 },
		{ 30, 203, 47 },
		{ 30, 204, 47 },
		{ 30, 205, 47 },
		{ 30, 206, 47 },
		{ 30, 207, 47 },
		{ 30, 208, 47 },
		{ 30, 209, 47 },
		{ 30, 210, 47 },
		{ 30, 211, 47 },
		{ 30, 212, 47 },
		{ 30, 213, 47 },
		{ 30, 214, 47 },
		{ 30, 215, 47 },
		{ 30, 216, 47 },
		{ 30, 217, 47 },
		{ 30, 218, 47 },
		{ 30, 219, 47 },
		{ 30, 220, 47 },
		{ 30, 221, 47 },
		{ 30, 222, 47 },
		{ 30, 223, 47 },
		{ 30, 224, 47 },
		{ 30, 225, 47 },
		{ 30, 226, 47 },
		{ 30, 227, 47 },
		{ 30, 228, 47 },
		{ 30, 229, 47 },
		{ 30, 230, 47 },
		{ 30, 231, 47 },
		{ 30, 232, 47 },
		{ 30, 233, 47 },
		{ 30, 234, 47 },
		{ 30, 235, 47 },
		{ 30, 236, 47 },
		{ 30, 237, 47 },
		{ 30, 238, 47 },
		{ 30, 239, 47 },
		{ 30, 240, 47 },
		{ 30, 241, 47 },
		{ 30, 242, 47 },
		{ 30, 243, 47 },
		{ 30, 244, 47 },
		{ 30, 245, 47 },
		{ 30, 246, 47 },
		{ 30, 247, 47 },
		{ 30, 248, 47 },
		{ 30, 249, 47 },
		{ 30, 250, 47 },
		{ 30, 251, 47 },
		{ 30, 252, 47 },
		{ 30, 253, 47 },
		{ 30, 254, 47 },
		{ 30, 255, 47 },
		{ 30, 256, 47 },
		{ 31, 201, 47 },
		{ 31, 202, 47 },
		{ 31, 203, 47 },
		{ 31, 204, 47 },
		{ 31, 205, 47 },
		{ 31, 206, 47 },
		{ 31, 207, 47 },
		{ 31, 208, 47 },
		{ 31, 209, 47 },
		{ 31, 210, 47 },
		{ 31, 211, 47 },
		{ 31, 212, 47 },
		{ 31, 213, 47 },
		{ 31, 214, 47 },
		{ 31, 215, 47 },
		{ 31, 216, 47 },
		{ 31, 217, 47 },
		{ 31, 218, 47 },
		{ 31, 219, 47 },
		{ 31, 220, 47 },
		{ 31, 221, 47 },
		{ 31, 222, 47 },
		{ 31, 223, 47 },
		{ 31, 224, 47 },
		{ 31, 225, 47 },
		{ 31, 226, 47 },
		{ 31, 227, 47 },
		{ 31, 228, 47 },
		{ 31, 229, 47 },
		{ 31, 230, 47 },
		{ 31, 231, 47 },
		{ 31, 232, 47 },
		{ 31, 233, 47 },
		{ 31, 234, 47 },
		{ 31, 235, 47 },
		{ 31, 236, 47 },
		{ 31, 237, 47 },
		{ 31, 238, 47 },
		{ 31, 239, 47 },
		{ 31, 240, 47 },
		{ 31, 241, 47 },
		{ 31, 242, 47 },
		{ 31, 243, 47 },
		{ 31, 244, 47 },
		{ 31, 245, 47 },
		{ 31, 246, 47 },
		{ 31, 247, 47 },
		{ 31, 248, 47 },
		{ 31, 249, 47 },
		{ 31, 250, 47 },
		{ 31, 251, 47 },
		{ 31, 252, 47 },
		{ 31, 253, 47 },
		{ 31, 254, 47 },
		{ 31, 255, 47 },
		{ 31, 256, 47 },
		{ 32, 201, 47 },
		{ 32, 202, 47 },
		{ 32, 203, 47 },
		{ 32, 204, 47 },
		{ 32, 205, 47 },
		{ 32, 206, 47 },
		{ 32, 207, 47 },
		{ 32, 208, 47 },
		{ 32, 209, 47 },
		{ 32, 210, 47 },
		{ 32, 211, 47 },
		{ 32, 212, 47 },
		{ 32, 213, 47 },
		{ 32, 214, 47 },
		{ 32, 215, 47 },
		{ 32, 216, 47 },
		{ 32, 217, 47 },
		{ 32, 218, 47 },
		{ 32, 219, 47 },
		{ 32, 220, 47 },
		{ 32, 221, 47 },
		{ 32, 222, 47 },
		{ 32, 223, 47 },
		{ 32, 224, 47 },
		{ 32, 225, 47 },
		{ 32, 226, 47 },
		{ 32, 227, 47 },
		{ 32, 228, 47 },
		{ 32, 229, 47 },
		{ 32, 230, 47 },
		{ 32, 231, 47 },
		{ 32, 232, 47 },
		{ 32, 233, 47 },
		{ 32, 234, 47 },
		{ 32, 235, 47 },
		{ 32, 236, 47 },
		{ 32, 237, 47 },
		{ 32, 238, 47 },
		{ 32, 239, 47 },
		{ 32, 240, 47 },
		{ 32, 241, 47 },
		{ 32, 242, 47 },
		{ 32, 243, 47 },
		{ 32, 244, 47 },
		{ 32, 245, 47 },
		{ 32, 246, 47 },
		{ 32, 247, 47 },
		{ 32, 248, 47 },
		{ 32, 249, 47 },
		{ 32, 250, 47 },
		{ 32, 251, 47 },
		{ 32, 252, 47 },
		{ 32, 253, 47 },
		{ 32, 254, 47 },
		{ 32, 255, 47 },
		{ 32, 256, 47 },
		{ 33, 202, 47 },
		{ 33, 203, 47 },
		{ 33, 204, 47 },
		{ 33, 205, 47 },
		{ 33, 206, 47 },
		{ 33, 207, 47 },
		{ 33, 208, 47 },
		{ 33, 209, 47 },
		{ 33, 210, 47 },
		{ 33, 211, 47 },
		{ 33, 212, 47 },
		{ 33, 213, 47 },
		{ 33, 214, 47 },
		{ 33, 215, 47 },
		{ 33, 216, 47 },
		{ 33, 217, 47 },
		{ 33, 218, 47 },
		{ 33, 219, 47 },
		{ 33, 220, 47 },
		{ 33, 221, 47 },
		{ 33, 222, 47 },
		{ 33, 223, 47 },
		{ 33, 224, 47 },
		{ 33, 225, 47 },
		{ 33, 226, 47 },
		{ 33, 227, 47 },
		{ 33, 228, 47 },
		{ 33, 229, 47 },
		{ 33, 230, 47 },
		{ 33, 231, 47 },
		{ 33, 232, 47 },
		{ 33, 233, 47 },
		{ 33, 234, 47 },
		{ 33, 235, 47 },
		{ 33, 236, 47 },
		{ 33, 237, 47 },
		{ 33, 238, 47 },
		{ 33, 239, 47 },
		{ 33, 240, 47 },
		{ 33, 241, 47 },
		{ 33, 242, 47 },
		{ 33, 243, 47 },
		{ 33, 244, 47 },
		{ 33, 245, 47 },
		{ 33, 246, 47 },
		{ 33, 247, 47 },
		{ 33, 248, 47 },
		{ 33, 249, 47 },
		{ 33, 250, 47 },
		{ 33, 251, 47 },
		{ 33, 252, 47 },
		{ 33, 253, 47 },
		{ 33, 254, 47 },
		{ 33, 255, 47 },
		{ 33, 256, 47 },
		{ 34, 202, 47 },
		{ 34, 203, 47 },
		{ 34, 204, 47 },
		{ 34, 205, 47 },
		{ 34, 206, 47 },
		{ 34, 207, 47 },
		{ 34, 208, 47 },
		{ 34, 209, 47 },
		{ 34, 210, 47 },
		{ 34, 211, 47 },
		{ 34, 212, 47 },
		{ 34, 213, 47 },
		{ 34, 214, 47 },
		{ 34, 215, 47 },
		{ 34, 216, 47 },
		{ 34, 217, 47 },
		{ 34, 218, 47 },
		{ 34, 219, 47 },
		{ 34, 220, 47 },
		{ 34, 221, 47 },
		{ 34, 222, 47 },
		{ 34, 223, 47 },
		{ 34, 224, 47 },
		{ 34, 225, 47 },
		{ 34, 226, 47 },
		{ 34, 227, 47 },
		{ 34, 228, 47 },
		{ 34, 229, 47 },
		{ 34, 230, 47 },
		{ 34, 231, 47 },
		{ 34, 232, 47 },
		{ 34, 233, 47 },
		{ 34, 234, 47 },
		{ 34, 235, 47 },
		{ 34, 236, 47 },
		{ 34, 237, 47 },
		{ 34, 238, 47 },
		{ 34, 239, 47 },
		{ 34, 240, 47 },
		{ 34, 241, 47 },
		{ 34, 242, 47 },
		{ 34, 243, 47 },
		{ 34, 244, 47 },
		{ 34, 245, 47 },
		{ 34, 246, 47 },
		{ 34, 247, 47 },
		{ 34, 248, 47 },
		{ 34, 249, 47 },
		{ 34, 250, 47 },
		{ 34, 251, 47 },
		{ 34, 252, 47 },
		{ 34, 253, 47 },
		{ 34, 254, 47 },
		{ 34, 255, 47 },
		{ 34, 256, 47 },
		{ 35, 202, 47 },
		{ 35, 203, 47 },
		{ 35, 204, 47 },
		{ 35, 205, 47 },
		{ 35, 206, 47 },
		{ 35, 207, 47 },
		{ 35, 208, 47 },
		{ 35, 209, 47 },
		{ 35, 210, 47 },
		{ 35, 211, 47 },
		{ 35, 212, 47 },
		{ 35, 213, 47 },
		{ 35, 214, 47 },
		{ 35, 215, 47 },
		{ 35, 216, 47 },
		{ 35, 217, 47 },
		{ 35, 218, 47 },
		{ 35, 219, 47 },
		{ 35, 220, 47 },
		{ 35, 221, 47 },
		{ 35, 222, 47 },
		{ 35, 223, 47 },
		{ 35, 224, 47 },
		{ 35, 225, 47 },
		{ 35, 226, 47 },
		{ 35, 227, 47 },
		{ 35, 228, 47 },
		{ 35, 229, 47 },
		{ 35, 230, 47 },
		{ 35, 231, 47 },
		{ 35, 232, 47 },
		{ 35, 233, 47 },
		{ 35, 234, 47 },
		{ 35, 235, 47 },
		{ 35, 236, 47 },
		{ 35, 237, 47 },
		{ 35, 238, 47 },
		{ 35, 239, 47 },
		{ 35, 240, 47 },
		{ 35, 241, 47 },
		{ 35, 242, 47 },
		{ 35, 243, 47 },
		{ 35, 244, 47 },
		{ 35, 245, 47 },
		{ 35, 246, 47 },
		{ 35, 247, 47 },
		{ 35, 248, 47 },
		{ 35, 249, 47 },
		{ 35, 250, 47 },
		{ 35, 251, 47 },
		{ 35, 252, 47 },
		{ 35, 253, 47 },
		{ 35, 254, 47 },
		{ 35, 255, 47 },
		{ 35, 256, 47 },
		{ 36, 202, 47 },
		{ 36, 203, 47 },
		{ 36, 204, 47 },
		{ 36, 205, 47 },
		{ 36, 206, 47 },
		{ 36, 207, 47 },
		{ 36, 208, 47 },
		{ 36, 209, 47 },
		{ 36, 210, 47 },
		{ 36, 211, 47 },
		{ 36, 212, 47 },
		{ 36, 213, 47 },
		{ 36, 214, 47 },
		{ 36, 215, 47 },
		{ 36, 216, 47 },
		{ 36, 217, 47 },
		{ 36, 218, 47 },
		{ 36, 219, 47 },
		{ 36, 220, 47 },
		{ 36, 221, 47 },
		{ 36, 222, 47 },
		{ 36, 223, 47 },
		{ 36, 224, 47 },
		{ 36, 225, 47 },
		{ 36, 226, 47 },
		{ 36, 227, 47 },
		{ 36, 228, 47 },
		{ 36, 229, 47 },
		{ 36, 230, 47 },
		{ 36, 231, 47 },
		{ 36, 232, 47 },
		{ 36, 233, 47 },
		{ 36, 234, 47 },
		{ 36, 235, 47 },
		{ 36, 236, 47 },
		{ 36, 237, 47 },
		{ 36, 238, 47 },
		{ 36, 239, 47 },
		{ 36, 240, 47 },
		{ 36, 241, 47 },
		{ 36, 242, 47 },
		{ 36, 243, 47 },
		{ 36, 244, 47 },
		{ 36, 245, 47 },
		{ 36, 246, 47 },
		{ 36, 247, 47 },
		{ 36, 248, 47 },
		{ 36, 249, 47 },
		{ 36, 250, 47 },
		{ 36, 251, 47 },
		{ 36, 252, 47 },
		{ 36, 253, 47 },
		{ 36, 254, 47 },
		{ 36, 255, 47 },
		{ 36, 256, 47 },
		{ 37, 203, 47 },
		{ 37, 204, 47 },
		{ 37, 205, 47 },
		{ 37, 206, 47 },
		{ 37, 207, 47 },
		{ 37, 208, 47 },
		{ 37, 209, 47 },
		{ 37, 210, 47 },
		{ 37, 211, 47 },
		{ 37, 212, 47 },
		{ 37, 213, 47 },
		{ 37, 214, 47 },
		{ 37, 215, 47 },
		{ 37, 216, 47 },
		{ 37, 217, 47 },
		{ 37, 218, 47 },
		{ 37, 219, 47 },
		{ 37, 220, 47 },
		{ 37, 221, 47 },
		{ 37, 222, 47 },
		{ 37, 223, 47 },
		{ 37, 224, 47 },
		{ 37, 225, 47 },
		{ 37, 226, 47 },
		{ 37, 227, 47 },
		{ 37, 228, 47 },
		{ 37, 229, 47 },
		{ 37, 230, 47 },
		{ 37, 231, 47 },
		{ 37, 232, 47 },
		{ 37, 233, 47 },
		{ 37, 234, 47 },
		{ 37, 235, 47 },
		{ 37, 236, 47 },
		{ 37, 237, 47 },
		{ 37, 238, 47 },
		{ 37, 239, 47 },
		{ 37, 240, 47 },
		{ 37, 241, 47 },
		{ 37, 242, 47 },
		{ 37, 243, 47 },
		{ 37, 244, 47 },
		{ 37, 245, 47 },
		{ 37, 246, 47 },
		{ 37, 247, 47 },
		{ 37, 248, 47 },
		{ 37, 249, 47 },
		{ 37, 250, 47 },
		{ 37, 251, 47 },
		{ 37, 252, 47 },
		{ 37, 253, 47 },
		{ 37, 254, 47 },
		{ 37, 255, 47 },
		{ 37, 256, 47 },
		{ 38, 203, 47 },
		{ 38, 204, 47 },
		{ 38, 205, 47 },
		{ 38, 206, 47 },
		{ 38, 207, 47 },
		{ 38, 208, 47 },
		{ 38, 209, 47 },
		{ 38, 210, 47 },
		{ 38, 211, 47 },
		{ 38, 212, 47 },
		{ 38, 213, 47 },
		{ 38, 214, 47 },
		{ 38, 215, 47 },
		{ 38, 216, 47 },
		{ 38, 217, 47 },
		{ 38, 218, 47 },
		{ 38, 219, 47 },
		{ 38, 220, 47 },
		{ 38, 221, 47 },
		{ 38, 222, 47 },
		{ 38, 223, 47 },
		{ 38, 224, 47 },
		{ 38, 225, 47 },
		{ 38, 226, 47 },
		{ 38, 227, 47 },
		{ 38, 228, 47 },
		{ 38, 229, 47 },
		{ 38, 230, 47 },
		{ 38, 231, 47 },
		{ 38, 232, 47 },
		{ 38, 233, 47 },
		{ 38, 234, 47 },
		{ 38, 235, 47 },
		{ 38, 236, 47 },
		{ 38, 237, 47 },
		{ 38, 238, 47 },
		{ 38, 239, 47 },
		{ 38, 240, 47 },
		{ 38, 241, 47 },
		{ 38, 242, 47 },
		{ 38, 243, 47 },
		{ 38, 244, 47 },
		{ 38, 245, 47 },
		{ 38, 246, 47 },
		{ 38, 247, 47 },
		{ 38, 248, 47 },
		{ 38, 249, 47 },
		{ 38, 250, 47 },
		{ 38, 251, 47 },
		{ 38, 252, 47 },
		{ 38, 253, 47 },
		{ 38, 254, 47 },
		{ 38, 255, 47 },
		{ 38, 256, 47 },
		{ 39, 203, 47 },
		{ 39, 204, 47 },
		{ 39, 205, 47 },
		{ 39, 206, 47 },
		{ 39, 207, 47 },
		{ 39, 208, 47 },
		{ 39, 209, 47 },
		{ 39, 210, 47 },
		{ 39, 211, 47 },
		{ 39, 212, 47 },
		{ 39, 213, 47 },
		{ 39, 214, 47 },
		{ 39, 215, 47 },
		{ 39, 216, 47 },
		{ 39, 217, 47 },
		{ 39, 218, 47 },
		{ 39, 219, 47 },
		{ 39, 220, 47 },
		{ 39, 221, 47 },
		{ 39, 222, 47 },
		{ 39, 223, 47 },
		{ 39, 224, 47 },
		{ 39, 225, 47 },
		{ 39, 226, 47 },
		{ 39, 227, 47 },
		{ 39, 228, 47 },
		{ 39, 229, 47 },
		{ 39, 230, 47 },
		{ 39, 231, 47 },
		{ 39, 232, 47 },
		{ 39, 233, 47 },
		{ 39, 234, 47 },
		{ 39, 235, 47 },
		{ 39, 236, 47 },
		{ 39, 237, 47 },
		{ 39, 238, 47 },
		{ 39, 239, 47 },
		{ 39, 240, 47 },
		{ 39, 241, 47 },
		{ 39, 242, 47 },
		{ 39, 243, 47 },
		{ 39, 244, 47 },
		{ 39, 245, 47 },
		{ 39, 246, 47 },
		{ 39, 247, 47 },
		{ 39, 248, 47 },
		{ 39, 249, 47 },
		{ 39, 250, 47 },
		{ 39, 251, 47 },
		{ 39, 252, 47 },
		{ 39, 253, 47 },
		{ 39, 254, 47 },
		{ 39, 255, 47 },
		{ 39, 256, 47 },
		{ 40, 203, 47 },
		{ 40, 204, 47 },
		{ 40, 205, 47 },
		{ 40, 206, 47 },
		{ 40, 207, 47 },
		{ 40, 208, 47 },
		{ 40, 209, 47 },
		{ 40, 210, 47 },
		{ 40, 211, 47 },
		{ 40, 212, 47 },
		{ 40, 213, 47 },
		{ 40, 214, 47 },
		{ 40, 215, 47 },
		{ 40, 216, 47 },
		{ 40, 217, 47 },
		{ 40, 218, 47 },
		{ 40, 219, 47 },
		{ 40, 220, 47 },
		{ 40, 221, 47 },
		{ 40, 222, 47 },
		{ 40, 223, 47 },
		{ 40, 224, 47 },
		{ 40, 225, 47 },
		{ 40, 226, 47 },
		{ 40, 227, 47 },
		{ 40, 228, 47 },
		{ 40, 229, 47 },
		{ 40, 230, 47 },
		{ 40, 231, 47 },
		{ 40, 232, 47 },
		{ 40, 233, 47 },
		{ 40, 234, 47 },
		{ 40, 235, 47 },
		{ 40, 236, 47 },
		{ 40, 237, 47 },
		{ 40, 238, 47 },
		{ 40, 239, 47 },
		{ 40, 240, 47 },
		{ 40, 241, 47 },
		{ 40, 242, 47 },
		{ 40, 243, 47 },
		{ 40, 244, 47 },
		{ 40, 245, 47 },
		{ 40, 246, 47 },
		{ 40, 247, 47 },
		{ 40, 248, 47 },
		{ 40, 249, 47 },
		{ 40, 250, 47 },
		{ 40, 251, 47 },
		{ 40, 252, 47 },
		{ 40, 253, 47 },
		{ 40, 254, 47 },
		{ 40, 255, 47 },
		{ 40, 256, 47 },
		{ 41, 204, 47 },
		{ 41, 205, 47 },
		{ 41, 206, 47 },
		{ 41, 207, 47 },
		{ 41, 208, 47 },
		{ 41, 209, 47 },
		{ 41, 210, 47 },
		{ 41, 211, 47 },
		{ 41, 212, 47 },
		{ 41, 213, 47 },
		{ 41, 214, 47 },
		{ 41, 215, 47 },
		{ 41, 216, 47 },
		{ 41, 217, 47 },
		{ 41, 218, 47 },
		{ 41, 219, 47 },
		{ 41, 220, 47 },
		{ 41, 221, 47 },
		{ 41, 222, 47 },
		{ 41, 223, 47 },
		{ 41, 224, 47 },
		{ 41, 225, 47 },
		{ 41, 226, 47 },
		{ 41, 227, 47 },
		{ 41, 228, 47 },
		{ 41, 229, 47 },
		{ 41, 230, 47 },
		{ 41, 231, 47 },
		{ 41, 232, 47 },
		{ 41, 233, 47 },
		{ 41, 234, 47 },
		{ 41, 235, 47 },
		{ 41, 236, 47 },
		{ 41, 237, 47 },
		{ 41, 238, 47 },
		{ 41, 239, 47 },
		{ 41, 240, 47 },
		{ 41, 241, 47 },
		{ 41, 242, 47 },
		{ 41, 243, 47 },
		{ 41, 244, 47 },
		{ 41, 245, 47 },
		{ 41, 246, 47 },
		{ 41, 247, 47 },
		{ 41, 248, 47 },
		{ 41, 249, 47 },
		{ 41, 250, 47 },
		{ 41, 251, 47 },
		{ 41, 252, 47 },
		{ 41, 253, 47 },
		{ 41, 254, 47 },
		{ 41, 255, 47 },
		{ 41, 256, 47 },
		{ 42, 204, 47 },
		{ 42, 205, 47 },
		{ 42, 206, 47 },
		{ 42, 207, 47 },
		{ 42, 208, 47 },
		{ 42, 209, 47 },
		{ 42, 210, 47 },
		{ 42, 211, 47 },
		{ 42, 212, 47 },
		{ 42, 213, 47 },
		{ 42, 214, 47 },
		{ 42, 215, 47 },
		{ 42, 216, 47 },
		{ 42, 217, 47 },
		{ 42, 218, 47 },
		{ 42, 219, 47 },
		{ 42, 220, 47 },
		{ 42, 221, 47 },
		{ 42, 222, 47 },
		{ 42, 223, 47 },
		{ 42, 224, 47 },
		{ 42, 225, 47 },
		{ 42, 226, 47 },
		{ 42, 227, 47 },
		{ 42, 228, 47 },
		{ 42, 229, 47 },
		{ 42, 230, 47 },
		{ 42, 231, 47 },
		{ 42, 232, 47 },
		{ 42, 233, 47 },
		{ 42, 234, 47 },
		{ 42, 235, 47 },
		{ 42, 236, 47 },
		{ 42, 237, 47 },
		{ 42, 238, 47 },
		{ 42, 239, 47 },
		{ 42, 240, 47 },
		{ 42, 241, 47 },
		{ 42, 242, 47 },
		{ 42, 243, 47 },
		{ 42, 244, 47 },
		{ 42, 245, 47 },
		{ 42, 246, 47 },
		{ 42, 247, 47 },
		{ 42, 248, 47 },
		{ 42, 249, 47 },
		{ 42, 250, 47 },
		{ 42, 251, 47 },
		{ 42, 252, 47 },
		{ 42, 253, 47 },
		{ 42, 254, 47 },
		{ 42, 255, 47 },
		{ 42, 256, 47 },
		{ 43, 204, 47 },
		{ 43, 205, 47 },
		{ 43, 206, 47 },
		{ 43, 207, 47 },
		{ 43, 208, 47 },
		{ 43, 209, 47 },
		{ 43, 210, 47 },
		{ 43, 211, 47 },
		{ 43, 212, 47 },
		{ 43, 213, 47 },
		{ 43, 214, 47 },
		{ 43, 215, 47 },
		{ 43, 216, 47 },
		{ 43, 217, 47 },
		{ 43, 218, 47 },
		{ 43, 219, 47 },
		{ 43, 220, 47 },
		{ 43, 221, 47 },
		{ 43, 222, 47 },
		{ 43, 223, 47 },
		{ 43, 224, 47 },
		{ 43, 225, 47 },
		{ 43, 226, 47 },
		{ 43, 227, 47 },
		{ 43, 228, 47 },
		{ 43, 229, 47 },
		{ 43, 230, 47 },
		{ 43, 231, 47 },
		{ 43, 232, 47 },
		{ 43, 233, 47 },
		{ 43, 234, 47 },
		{ 43, 235, 47 },
		{ 43, 236, 47 },
		{ 43, 237, 47 },
		{ 43, 238, 47 },
		{ 43, 239, 47 },
		{ 43, 240, 47 },
		{ 43, 241, 47 },
		{ 43, 242, 47 },
		{ 43, 243, 47 },
		{ 43, 244, 47 },
		{ 43, 245, 47 },
		{ 43, 246, 47 },
		{ 43, 247, 47 },
		{ 43, 248, 47 },
		{ 43, 249, 47 },
		{ 43, 250, 47 },
		{ 43, 251, 47 },
		{ 43, 252, 47 },
		{ 43, 253, 47 },
		{ 43, 254, 47 },
		{ 43, 255, 47 },
		{ 43, 256, 47 },
		{ 44, 204, 47 },
		{ 44, 205, 47 },
		{ 44, 206, 47 },
		{ 44, 207, 47 },
		{ 44, 208, 47 },
		{ 44, 209, 47 },
		{ 44, 210, 47 },
		{ 44, 211, 47 },
		{ 44, 212, 47 },
		{ 44, 213, 47 },
		{ 44, 214, 47 },
		{ 44, 215, 47 },
		{ 44, 216, 47 },
		{ 44, 217, 47 },
		{ 44, 218, 47 },
		{ 44, 219, 47 },
		{ 44, 220, 47 },
		{ 44, 221, 47 },
		{ 44, 222, 47 },
		{ 44, 223, 47 },
		{ 44, 224, 47 },
		{ 44, 225, 47 },
		{ 44, 226, 47 },
		{ 44, 227, 47 },
		{ 44, 228, 47 },
		{ 44, 229, 47 },
		{ 44, 230, 47 },
		{ 44, 231, 47 },
		{ 44, 232, 47 },
		{ 44, 233, 47 },
		{ 44, 234, 47 },
		{ 44, 235, 47 },
		{ 44, 236, 47 },
		{ 44, 237, 47 },
		{ 44, 238, 47 },
		{ 44, 239, 47 },
		{ 44, 240, 47 },
		{ 44, 241, 47 },
		{ 44, 242, 47 },
		{ 44, 243, 47 },
		{ 44, 244, 47 },
		{ 44, 245, 47 },
		{ 44, 246, 47 },
		{ 44, 247, 47 },
		{ 44, 248, 47 },
		{ 44, 249, 47 },
		{ 44, 250, 47 },
		{ 44, 251, 47 },
		{ 44, 252, 47 },
		{ 44, 253, 47 },
		{ 44, 254, 47 },
		{ 44, 255, 47 },
		{ 44, 256, 47 },
		{ 45, 205, 47 },
		{ 45, 206, 47 },
		{ 45, 207, 47 },
		{ 45, 208, 47 },
		{ 45, 209, 47 },
		{ 45, 210, 47 },
		{ 45, 211, 47 },
		{ 45, 212, 47 },
		{ 45, 213, 47 },
		{ 45, 214, 47 },
		{ 45, 215, 47 },
		{ 45, 216, 47 },
		{ 45, 217, 47 },
		{ 45, 218, 47 },
		{ 45, 219, 47 },
		{ 45, 220, 47 },
		{ 45, 221, 47 },
		{ 45, 222, 47 },
		{ 45, 223, 47 },
		{ 45, 224, 47 },
		{ 45, 225, 47 },
		{ 45, 226, 47 },
		{ 45, 227, 47 },
		{ 45, 228, 47 },
		{ 45, 229, 47 },
		{ 45, 230, 47 },
		{ 45, 231, 47 },
		{ 45, 232, 47 },
		{ 45, 233, 47 },
		{ 45, 234, 47 },
		{ 45, 235, 47 },
		{ 45, 236, 47 },
		{ 45, 237, 47 },
		{ 45, 238, 47 },
		{ 45, 239, 47 },
		{ 45, 240, 47 },
		{ 45, 241, 47 },
		{ 45, 242, 47 },
		{ 45, 243, 47 },
		{ 45, 244, 47 },
		{ 45, 245, 47 },
		{ 45, 246, 47 },
		{ 45, 247, 47 },
		{ 45, 248, 47 },
		{ 45, 249, 47 },
		{ 45, 250, 47 },
		{ 45, 251, 47 },
		{ 45, 252, 47 },
		{ 45, 253, 47 },
		{ 45, 254, 47 },
		{ 45, 255, 47 },
		{ 45, 256, 47 },
		{ 46, 205, 47 },
		{ 46, 206, 47 },
		{ 46, 207, 47 },
		{ 46, 208, 47 },
		{ 46, 209, 47 },
		{ 46, 210, 47 },
		{ 46, 211, 47 },
		{ 46, 212, 47 },
		{ 46, 213, 47 },
		{ 46, 214, 47 },
		{ 46, 215, 47 },
		{ 46, 216, 47 },
		{ 46, 217, 47 },
		{ 46, 218, 47 },
		{ 46, 219, 47 },
		{ 46, 220, 47 },
		{ 46, 221, 47 },
		{ 46, 222, 47 },
		{ 46, 223, 47 },
		{ 46, 224, 47 },
		{ 46, 225, 47 },
		{ 46, 226, 47 },
		{ 46, 227, 47 },
		{ 46, 228, 47 },
		{ 46, 229, 47 },
		{ 46, 230, 47 },
		{ 46, 231, 47 },
		{ 46, 232, 47 },
		{ 46, 233, 47 },
		{ 46, 234, 47 },
		{ 46, 235, 47 },
		{ 46, 236, 47 },
		{ 46, 237, 47 },
		{ 46, 238, 47 },
		{ 46, 239, 47 },
		{ 46, 240, 47 },
		{ 46, 241, 47 },
		{ 46, 242, 47 },
		{ 46, 243, 47 },
		{ 46, 244, 47 },
		{ 46, 245, 47 },
		{ 46, 246, 47 },
		{ 46, 247, 47 },
		{ 46, 248, 47 },
		{ 46, 249, 47 },
		{ 46, 250, 47 },
		{ 46, 251, 47 },
		{ 46, 252, 47 },
		{ 46, 253, 47 },
		{ 46, 254, 47 },
		{ 46, 255, 47 },
		{ 46, 256, 47 },
		{ 47, 205, 47 },
		{ 47, 206, 47 },
		{ 47, 207, 47 },
		{ 47, 208, 47 },
		{ 47, 209, 47 },
		{ 47, 210, 47 },
		{ 47, 211, 47 },
		{ 47, 212, 47 },
		{ 47, 213, 47 },
		{ 47, 214, 47 },
		{ 47, 215, 47 },
		{ 47, 216, 47 },
		{ 47, 217, 47 },
		{ 47, 218, 47 },
		{ 47, 219, 47 },
		{ 47, 220, 47 },
		{ 47, 221, 47 },
		{ 47, 222, 47 },
		{ 47, 223, 47 },
		{ 47, 224, 47 },
		{ 47, 225, 47 },
		{ 47, 226, 47 },
		{ 47, 227, 47 },
		{ 47, 228, 47 },
		{ 47, 229, 47 },
		{ 47, 230, 47 },
		{ 47, 231, 47 },
		{ 47, 232, 47 },
		{ 47, 233, 47 },
		{ 47, 234, 47 },
		{ 47, 235, 47 },
		{ 47, 236, 47 },
		{ 47, 237, 47 },
		{ 47, 238, 47 },
		{ 47, 239, 47 },
		{ 47, 240, 47 },
		{ 47, 241, 47 },
		{ 47, 242, 47 },
		{ 47, 243, 47 },
		{ 47, 244, 47 },
		{ 47, 245, 47 },
		{ 47, 246, 47 },
		{ 47, 247, 47 },
		{ 47, 248, 47 },
		{ 47, 249, 47 },
		{ 47, 250, 47 },
		{ 47, 251, 47 },
		{ 47, 252, 47 },
		{ 47, 253, 47 },
		{ 47, 254, 47 },
		{ 47, 255, 47 },
		{ 47, 256, 47 },
		{ 48, 205, 47 },
		{ 48, 206, 47 },
		{ 48, 207, 47 },
		{ 48, 208, 47 },
		{ 48, 209, 47 },
		{ 48, 210, 47 },
		{ 48, 211, 47 },
		{ 48, 212, 47 },
		{ 48, 213, 47 },
		{ 48, 214, 47 },
		{ 48, 215, 47 },
		{ 48, 216, 47 },
		{ 48, 217, 47 },
		{ 48, 218, 47 },
		{ 48, 219, 47 },
		{ 48, 220, 47 },
		{ 48, 221, 47 },
		{ 48, 222, 47 },
		{ 48, 223, 47 },
		{ 48, 224, 47 },
		{ 48, 225, 47 },
		{ 48, 226, 47 },
		{ 48, 227, 47 },
		{ 48, 228, 47 },
		{ 48, 229, 47 },
		{ 48, 230, 47 },
		{ 48, 231, 47 },
		{ 48, 232, 47 },
		{ 48, 233, 47 },
		{ 48, 234, 47 },
		{ 48, 235, 47 },
		{ 48, 236, 47 },
		{ 48, 237, 47 },
		{ 48, 238, 47 },
		{ 48, 239, 47 },
		{ 48, 240, 47 },
		{ 48, 241, 47 },
		{ 48, 242, 47 },
		{ 48, 243, 47 },
		{ 48, 244, 47 },
		{ 48, 245, 47 },
		{ 48, 246, 47 },
		{ 48, 247, 47 },
		{ 48, 248, 47 },
		{ 48, 249, 47 },
		{ 48, 250, 47 },
		{ 48, 251, 47 },
		{ 48, 252, 47 },
		{ 48, 253, 47 },
		{ 48, 254, 47 },
		{ 48, 255, 47 },
		{ 48, 256, 47 },
		{ 49, 206, 47 },
		{ 49, 207, 47 },
		{ 49, 208, 47 },
		{ 49, 209, 47 },
		{ 49, 210, 47 },
		{ 49, 211, 47 },
		{ 49, 212, 47 },
		{ 49, 213, 47 },
		{ 49, 214, 47 },
		{ 49, 215, 47 },
		{ 49, 216, 47 },
		{ 49, 217, 47 },
		{ 49, 218, 47 },
		{ 49, 219, 47 },
		{ 49, 220, 47 },
		{ 49, 221, 47 },
		{ 49, 222, 47 },
		{ 49, 223, 47 },
		{ 49, 224, 47 },
		{ 49, 225, 47 },
		{ 49, 226, 47 },
		{ 49, 227, 47 },
		{ 49, 228, 47 },
		{ 49, 229, 47 },
		{ 49, 230, 47 },
		{ 49, 231, 47 },
		{ 49, 232, 47 },
		{ 49, 233, 47 },
		{ 49, 234, 47 },
		{ 49, 235, 47 },
		{ 49, 236, 47 },
		{ 49, 237, 47 },
		{ 49, 238, 47 },
		{ 49, 239, 47 },
		{ 49, 240, 47 },
		{ 49, 241, 47 },
		{ 49, 242, 47 },
		{ 49, 243, 47 },
		{ 49, 244, 47 },
		{ 49, 245, 47 },
		{ 49, 246, 47 },
		{ 49, 247, 47 },
		{ 49, 248, 47 },
		{ 49, 249, 47 },
		{ 49, 250, 47 },
		{ 49, 251, 47 },
		{ 49, 252, 47 },
		{ 49, 253, 47 },
		{ 49, 254, 47 },
		{ 49, 255, 47 },
		{ 49, 256, 47 },
		{ 50, 206, 47 },
		{ 50, 207, 47 },
		{ 50, 208, 47 },
		{ 50, 209, 47 },
		{ 50, 210, 47 },
		{ 50, 211, 47 },
		{ 50, 212, 47 },
		{ 50, 213, 47 },
		{ 50, 214, 47 },
		{ 50, 215, 47 },
		{ 50, 216, 47 },
		{ 50, 217, 47 },
		{ 50, 218, 47 },
		{ 50, 219, 47 },
		{ 50, 220, 47 },
		{ 50, 221, 47 },
		{ 50, 222, 47 },
		{ 50, 223, 47 },
		{ 50, 224, 47 },
		{ 50, 225, 47 },
		{ 50, 226, 47 },
		{ 50, 227, 47 },
		{ 50, 228, 47 },
		{ 50, 229, 47 },
		{ 50, 230, 47 },
		{ 50, 231, 47 },
		{ 50, 232, 47 },
		{ 50, 233, 47 },
		{ 50, 234, 47 },
		{ 50, 235, 47 },
		{ 50, 236, 47 },
		{ 50, 237, 47 },
		{ 50, 238, 47 },
		{ 50, 239, 47 },
		{ 50, 240, 47 },
		{ 50, 241, 47 },
		{ 50, 242, 47 },
		{ 50, 243, 47 },
		{ 50, 244, 47 },
		{ 50, 245, 47 },
		{ 50, 246, 47 },
		{ 50, 247, 47 },
		{ 50, 248, 47 },
		{ 50, 249, 47 },
		{ 50, 250, 47 },
		{ 50, 251, 47 },
		{ 50, 252, 47 },
		{ 50, 253, 47 },
		{ 50, 254, 47 },
		{ 50, 255, 47 },
		{ 50, 256, 47 },
		{ 51, 206, 47 },
		{ 51, 207, 47 },
		{ 51, 208, 47 },
		{ 51, 209, 47 },
		{ 51, 210, 47 },
		{ 51, 211, 47 },
		{ 51, 212, 47 },
		{ 51, 213, 47 },
		{ 51, 214, 47 },
		{ 51, 215, 47 },
		{ 51, 216, 47 },
		{ 51, 217, 47 },
		{ 51, 218, 47 },
		{ 51, 219, 47 },
		{ 51, 220, 47 },
		{ 51, 221, 47 },
		{ 51, 222, 47 },
		{ 51, 223, 47 },
		{ 51, 224, 47 },
		{ 51, 225, 47 },
		{ 51, 226, 47 },
		{ 51, 227, 47 },
		{ 51, 228, 47 },
		{ 51, 229, 47 },
		{ 51, 230, 47 },
		{ 51, 231, 47 },
		{ 51, 232, 47 },
		{ 51, 233, 47 },
		{ 51, 234, 47 },
		{ 51, 235, 47 },
		{ 51, 236, 47 },
		{ 51, 237, 47 },
		{ 51, 238, 47 },
		{ 51, 239, 47 },
		{ 51, 240, 47 },
		{ 51, 241, 47 },
		{ 51, 242, 47 },
		{ 51, 243, 47 },
		{ 51, 244, 47 },
		{ 51, 245, 47 },
		{ 51, 246, 47 },
		{ 51, 247, 47 },
		{ 51, 248, 47 },
		{ 51, 249, 47 },
		{ 51, 250, 47 },
		{ 51, 251, 47 },
		{ 51, 252, 47 },
		{ 51, 253, 47 },
		{ 51, 254, 47 },
		{ 51, 255, 47 },
		{ 51, 256, 47 },
		{ 52, 206, 47 },
		{ 52, 207, 47 },
		{ 52, 208, 47 },
		{ 52, 209, 47 },
		{ 52, 210, 47 },
		{ 52, 211, 47 },
		{ 52, 212, 47 },
		{ 52, 213, 47 },
		{ 52, 214, 47 },
		{ 52, 215, 47 },
		{ 52, 216, 47 },
		{ 52, 217, 47 },
		{ 52, 218, 47 },
		{ 52, 219, 47 },
		{ 52, 220, 47 },
		{ 52, 221, 47 },
		{ 52, 222, 47 },
		{ 52, 223, 47 },
		{ 52, 224, 47 },
		{ 52, 225, 47 },
		{ 52, 226, 47 },
		{ 52, 227, 47 },
		{ 52, 228, 47 },
		{ 52, 229, 47 },
		{ 52, 230, 47 },
		{ 52, 231, 47 },
		{ 52, 232, 47 },
		{ 52, 233, 47 },
		{ 52, 234, 47 },
		{ 52, 235, 47 },
		{ 52, 236, 47 },
		{ 52, 237, 47 },
		{ 52, 238, 47 },
		{ 52, 239, 47 },
		{ 52, 240, 47 },
		{ 52, 241, 47 },
		{ 52, 242, 47 },
		{ 52, 243, 47 },
		{ 52, 244, 47 },
		{ 52, 245, 47 },
		{ 52, 246, 47 },
		{ 52, 247, 47 },
		{ 52, 248, 47 },
		{ 52, 249, 47 },
		{ 52, 250, 47 },
		{ 52, 251, 47 },
		{ 52, 252, 47 },
		{ 52, 253, 47 },
		{ 52, 254, 47 },
		{ 52, 255, 47 },
		{ 52, 256, 47 },
		{ 53, 207, 47 },
		{ 53, 208, 47 },
		{ 53, 209, 47 },
		{ 53, 210, 47 },
		{ 53, 211, 47 },
		{ 53, 212, 47 },
		{ 53, 213, 47 },
		{ 53, 214, 47 },
		{ 53, 215, 47 },
		{ 53, 216, 47 },
		{ 53, 217, 47 },
		{ 53, 218, 47 },
		{ 53, 219, 47 },
		{ 53, 220, 47 },
		{ 53, 221, 47 },
		{ 53, 222, 47 },
		{ 53, 223, 47 },
		{ 53, 224, 47 },
		{ 53, 225, 47 },
		{ 53, 226, 47 },
		{ 53, 227, 47 },
		{ 53, 228, 47 },
		{ 53, 229, 47 },
		{ 53, 230, 47 },
		{ 53, 231, 47 },
		{ 53, 232, 47 },
		{ 53, 233, 47 },
		{ 53, 234, 47 },
		{ 53, 235, 47 },
		{ 53, 236, 47 },
		{ 53, 237, 47 },
		{ 53, 238, 47 },
		{ 53, 239, 47 },
		{ 53, 240, 47 },
		{ 53, 241, 47 },
		{ 53, 242, 47 },
		{ 53, 243, 47 },
		{ 53, 244, 47 },
		{ 53, 245, 47 },
		{ 53, 246, 47 },
		{ 53, 247, 47 },
		{ 53, 248, 47 },
		{ 53, 249, 47 },
		{ 53, 250, 47 },
		{ 53, 251, 47 },
		{ 53, 252, 47 },
		{ 53, 253, 47 },
		{ 53, 254, 47 },
		{ 53, 255, 47 },
		{ 53, 256, 47 },
		{ 54, 207, 47 },
		{ 54, 208, 47 },
		{ 54, 209, 47 },
		{ 54, 210, 47 },
		{ 54, 211, 47 },
		{ 54, 212, 47 },
		{ 54, 213, 47 },
		{ 54, 214, 47 },
		{ 54, 215, 47 },
		{ 54, 216, 47 },
		{ 54, 217, 47 },
		{ 54, 218, 47 },
		{ 54, 219, 47 },
		{ 54, 220, 47 },
		{ 54, 221, 47 },
		{ 54, 222, 47 },
		{ 54, 223, 47 },
		{ 54, 224, 47 },
		{ 54, 225, 47 },
		{ 54, 226, 47 },
		{ 54, 227, 47 },
		{ 54, 228, 47 },
		{ 54, 229, 47 },
		{ 54, 230, 47 },
		{ 54, 231, 47 },
		{ 54, 232, 47 },
		{ 54, 233, 47 },
		{ 54, 234, 47 },
		{ 54, 235, 47 },
		{ 54, 236, 47 },
		{ 54, 237, 47 },
		{ 54, 238, 47 },
		{ 54, 239, 47 },
		{ 54, 240, 47 },
		{ 54, 241, 47 },
		{ 54, 242, 47 },
		{ 54, 243, 47 },
		{ 54, 244, 47 },
		{ 54, 245, 47 },
		{ 54, 246, 47 },
		{ 54, 247, 47 },
		{ 54, 248, 47 },
		{ 54, 249, 47 },
		{ 54, 250, 47 },
		{ 54, 251, 47 },
		{ 54, 252, 47 },
		{ 54, 253, 47 },
		{ 54, 254, 47 },
		{ 54, 255, 47 },
		{ 54, 256, 47 },
		{ 55, 207, 47 },
		{ 55, 208, 47 },
		{ 55, 209, 47 },
		{ 55, 210, 47 },
		{ 55, 211, 47 },
		{ 55, 212, 47 },
		{ 55, 213, 47 },
		{ 55, 214, 47 },
		{ 55, 215, 47 },
		{ 55, 216, 47 },
		{ 55, 217, 47 },
		{ 55, 218, 47 },
		{ 55, 219, 47 },
		{ 55, 220, 47 },
		{ 55, 221, 47 },
		{ 55, 222, 47 },
		{ 55, 223, 47 },
		{ 55, 224, 47 },
		{ 55, 225, 47 },
		{ 55, 226, 47 },
		{ 55, 227, 47 },
		{ 55, 228, 47 },
		{ 55, 229, 47 },
		{ 55, 230, 47 },
		{ 55, 231, 47 },
		{ 55, 232, 47 },
		{ 55, 233, 47 },
		{ 55, 234, 47 },
		{ 55, 235, 47 },
		{ 55, 236, 47 },
		{ 55, 237, 47 },
		{ 55, 238, 47 },
		{ 55, 239, 47 },
		{ 55, 240, 47 },
		{ 55, 241, 47 },
		{ 55, 242, 47 },
		{ 55, 243, 47 },
		{ 55, 244, 47 },
		{ 55, 245, 47 },
		{ 55, 246, 47 },
		{ 55, 247, 47 },
		{ 55, 248, 47 },
		{ 55, 249, 47 },
		{ 55, 250, 47 },
		{ 55, 251, 47 },
		{ 55, 252, 47 },
		{ 55, 253, 47 },
		{ 55, 254, 47 },
		{ 55, 255, 47 },
		{ 55, 256, 47 },
		{ 56, 207, 47 },
		{ 56, 208, 47 },
		{ 56, 209, 47 },
		{ 56, 210, 47 },
		{ 56, 211, 47 },
		{ 56, 212, 47 },
		{ 56, 213, 47 },
		{ 56, 214, 47 },
		{ 56, 215, 47 },
		{ 56, 216, 47 },
		{ 56, 217, 47 },
		{ 56, 218, 47 },
		{ 56, 219, 47 },
		{ 56, 220, 47 },
		{ 56, 221, 47 },
		{ 56, 222, 47 },
		{ 56, 223, 47 },
		{ 56, 224, 47 },
		{ 56, 225, 47 },
		{ 56, 226, 47 },
		{ 56, 227, 47 },
		{ 56, 228, 47 },
		{ 56, 229, 47 },
		{ 56, 230, 47 },
		{ 56, 231, 47 },
		{ 56, 232, 47 },
		{ 56, 233, 47 },
		{ 56, 234, 47 },
		{ 56, 235, 47 },
		{ 56, 236, 47 },
		{ 56, 237, 47 },
		{ 56, 238, 47 },
		{ 56, 239, 47 },
		{ 56, 240, 47 },
		{ 56, 241, 47 },
		{ 56, 242, 47 },
		{ 56, 243, 47 },
		{ 56, 244, 47 },
		{ 56, 245, 47 },
		{ 56, 246, 47 },
		{ 56, 247, 47 },
		{ 56, 248, 47 },
		{ 56, 249, 47 },
		{ 56, 250, 47 },
		{ 56, 251, 47 },
		{ 56, 252, 47 },
		{ 56, 253, 47 },
		{ 56, 254, 47 },
		{ 56, 255, 47 },
		{ 56, 256, 47 },
		{ 57, 208, 47 },
		{ 57, 209, 47 },
		{ 57, 210, 47 },
		{ 57, 211, 47 },
		{ 57, 212, 47 },
		{ 57, 213, 47 },
		{ 57, 214, 47 },
		{ 57, 215, 47 },
		{ 57, 216, 47 },
		{ 57, 217, 47 },
		{ 57, 218, 47 },
		{ 57, 219, 47 },
		{ 57, 220, 47 },
		{ 57, 221, 47 },
		{ 57, 222, 47 },
		{ 57, 223, 47 },
		{ 57, 224, 47 },
		{ 57, 225, 47 },
		{ 57, 226, 47 },
		{ 57, 227, 47 },
		{ 57, 228, 47 },
		{ 57, 229, 47 },
		{ 57, 230, 47 },
		{ 57, 231, 47 },
		{ 57, 232, 47 },
		{ 57, 233, 47 },
		{ 57, 234, 47 },
		{ 57, 235, 47 },
		{ 57, 236, 47 },
		{ 57, 237, 47 },
		{ 57, 238, 47 },
		{ 57, 239, 47 },
		{ 57, 240, 47 },
		{ 57, 241, 47 },
		{ 57, 242, 47 },
		{ 57, 243, 47 },
		{ 57, 244, 47 },
		{ 57, 245, 47 },
		{ 57, 246, 47 },
		{ 57, 247, 47 },
		{ 57, 248, 47 },
		{ 57, 249, 47 },
		{ 57, 250, 47 },
		{ 57, 251, 47 },
		{ 57, 252, 47 },
		{ 57, 253, 47 },
		{ 57, 254, 47 },
		{ 57, 255, 47 },
		{ 57, 256, 47 },
		{ 58, 208, 47 },
		{ 58, 209, 47 },
		{ 58, 210, 47 },
		{ 58, 211, 47 },
		{ 58, 212, 47 },
		{ 58, 213, 47 },
		{ 58, 214, 47 },
		{ 58, 215, 47 },
		{ 58, 216, 47 },
		{ 58, 217, 47 },
		{ 58, 218, 47 },
		{ 58, 219, 47 },
		{ 58, 220, 47 },
		{ 58, 221, 47 },
		{ 58, 222, 47 },
		{ 58, 223, 47 },
		{ 58, 224, 47 },
		{ 58, 225, 47 },
		{ 58, 226, 47 },
		{ 58, 227, 47 },
		{ 58, 228, 47 },
		{ 58, 229, 47 },
		{ 58, 230, 47 },
		{ 58, 231, 47 },
		{ 58, 232, 47 },
		{ 58, 233, 47 },
		{ 58, 234, 47 },
		{ 58, 235, 47 },
		{ 58, 236, 47 },
		{ 58, 237, 47 },
		{ 58, 238, 47 },
		{ 58, 239, 47 },
		{ 58, 240, 47 },
		{ 58, 241, 47 },
		{ 58, 242, 47 },
		{ 58, 243, 47 },
		{ 58, 244, 47 },
		{ 58, 245, 47 },
		{ 58, 246, 47 },
		{ 58, 247, 47 },
		{ 58, 248, 47 },
		{ 58, 249, 47 },
		{ 58, 250, 47 },
		{ 58, 251, 47 },
		{ 58, 252, 47 },
		{ 58, 253, 47 },
		{ 58, 254, 47 },
		{ 58, 255, 47 },
		{ 58, 256, 47 },
		{ 59, 208, 47 },
		{ 59, 209, 47 },
		{ 59, 210, 47 },
		{ 59, 211, 47 },
		{ 59, 212, 47 },
		{ 59, 213, 47 },
		{ 59, 214, 47 },
		{ 59, 215, 47 },
		{ 59, 216, 47 },
		{ 59, 217, 47 },
		{ 59, 218, 47 },
		{ 59, 219, 47 },
		{ 59, 220, 47 },
		{ 59, 221, 47 },
		{ 59, 222, 47 },
		{ 59, 223, 47 },
		{ 59, 224, 47 },
		{ 59, 225, 47 },
		{ 59, 226, 47 },
		{ 59, 227, 47 },
		{ 59, 228, 47 },
		{ 59, 229, 47 },
		{ 59, 230, 47 },
		{ 59, 231, 47 },
		{ 59, 232, 47 },
		{ 59, 233, 47 },
		{ 59, 234, 47 },
		{ 59, 235, 47 },
		{ 59, 236, 47 },
		{ 59, 237, 47 },
		{ 59, 238, 47 },
		{ 59, 239, 47 },
		{ 59, 240, 47 },
		{ 59, 241, 47 },
		{ 59, 242, 47 },
		{ 59, 243, 47 },
		{ 59, 244, 47 },
		{ 59, 245, 47 },
		{ 59, 246, 47 },
		{ 59, 247, 47 },
		{ 59, 248, 47 },
		{ 59, 249, 47 },
		{ 59, 250, 47 },
		{ 59, 251, 47 },
		{ 59, 252, 47 },
		{ 59, 253, 47 },
		{ 59, 254, 47 },
		{ 59, 255, 47 },
		{ 59, 256, 47 },
		{ 60, 208, 47 },
		{ 60, 209, 47 },
		{ 60, 210, 47 },
		{ 60, 211, 47 },
		{ 60, 212, 47 },
		{ 60, 213, 47 },
		{ 60, 214, 47 },
		{ 60, 215, 47 },
		{ 60, 216, 47 },
		{ 60, 217, 47 },
		{ 60, 218, 47 },
		{ 60, 219, 47 },
		{ 60, 220, 47 },
		{ 60, 221, 47 },
		{ 60, 222, 47 },
		{ 60, 223, 47 },
		{ 60, 224, 47 },
		{ 60, 225, 47 },
		{ 60, 226, 47 },
		{ 60, 227, 47 },
		{ 60, 228, 47 },
		{ 60, 229, 47 },
		{ 60, 230, 47 },
		{ 60, 231, 47 },
		{ 60, 232, 47 },
		{ 60, 233, 47 },
		{ 60, 234, 47 },
		{ 60, 235, 47 },
		{ 60, 236, 47 },
		{ 60, 237, 47 },
		{ 60, 238, 47 },
		{ 60, 239, 47 },
		{ 60, 240, 47 },
		{ 60, 241, 47 },
		{ 60, 242, 47 },
		{ 60, 243, 47 },
		{ 60, 244, 47 },
		{ 60, 245, 47 },
		{ 60, 246, 47 },
		{ 60, 247, 47 },
		{ 60, 248, 47 },
		{ 60, 249, 47 },
		{ 60, 250, 47 },
		{ 60, 251, 47 },
		{ 60, 252, 47 },
		{ 60, 253, 47 },
		{ 60, 254, 47 },
		{ 60, 255, 47 },
		{ 60, 256, 47 },
		{ 61, 209, 47 },
		{ 61, 210, 47 },
		{ 61, 211, 47 },
		{ 61, 212, 47 },
		{ 61, 213, 47 },
		{ 61, 214, 47 },
		{ 61, 215, 47 },
		{ 61, 216, 47 },
		{ 61, 217, 47 },
		{ 61, 218, 47 },
		{ 61, 219, 47 },
		{ 61, 220, 47 },
		{ 61, 221, 47 },
		{ 61, 222, 47 },
		{ 61, 223, 47 },
		{ 61, 224, 47 },
		{ 61, 225, 47 },
		{ 61, 226, 47 },
		{ 61, 227, 47 },
		{ 61, 228, 47 },
		{ 61, 229, 47 },
		{ 61, 230, 47 },
		{ 61, 231, 47 },
		{ 61, 232, 47 },
		{ 61, 233, 47 },
		{ 61, 234, 47 },
		{ 61, 235, 47 },
		{ 61, 236, 47 },
		{ 61, 237, 47 },
		{ 61, 238, 47 },
		{ 61, 239, 47 },
		{ 61, 240, 47 },
		{ 61, 241, 47 },
		{ 61, 242, 47 },
		{ 61, 243, 47 },
		{ 61, 244, 47 },
		{ 61, 245, 47 },
		{ 61, 246, 47 },
		{ 61, 247, 47 },
		{ 61, 248, 47 },
		{ 61, 249, 47 },
		{ 61, 250, 47 },
		{ 61, 251, 47 },
		{ 61, 252, 47 },
		{ 61, 253, 47 },
		{ 61, 254, 47 },
		{ 61, 255, 47 },
		{ 61, 256, 47 },
		{ 62, 209, 47 },
		{ 62, 210, 47 },
		{ 62, 211, 47 },
		{ 62, 212, 47 },
		{ 62, 213, 47 },
		{ 62, 214, 47 },
		{ 62, 215, 47 },
		{ 62, 216, 47 },
		{ 62, 217, 47 },
		{ 62, 218, 47 },
		{ 62, 219, 47 },
		{ 62, 220, 47 },
		{ 62, 221, 47 },
		{ 62, 222, 47 },
		{ 62, 223, 47 },
		{ 62, 224, 47 },
		{ 62, 225, 47 },
		{ 62, 226, 47 },
		{ 62, 227, 47 },
		{ 62, 228, 47 },
		{ 62, 229, 47 },
		{ 62, 230, 47 },
		{ 62, 231, 47 },
		{ 62, 232, 47 },
		{ 62, 233, 47 },
		{ 62, 234, 47 },
		{ 62, 235, 47 },
		{ 62, 236, 47 },
		{ 62, 237, 47 },
		{ 62, 238, 47 },
		{ 62, 239, 47 },
		{ 62, 240, 47 },
		{ 62, 241, 47 },
		{ 62, 242, 47 },
		{ 62, 243, 47 },
		{ 62, 244, 47 },
		{ 62, 245, 47 },
		{ 62, 246, 47 },
		{ 62, 247, 47 },
		{ 62, 248, 47 },
		{ 62, 249, 47 },
		{ 62, 250, 47 },
		{ 62, 251, 47 },
		{ 62, 252, 47 },
		{ 62, 253, 47 },
		{ 62, 254, 47 },
		{ 62, 255, 47 },
		{ 62, 256, 47 },
		{ 63, 209, 47 },
		{ 63, 210, 47 },
		{ 63, 211, 47 },
		{ 63, 212, 47 },
		{ 63, 213, 47 },
		{ 63, 214, 47 },
		{ 63, 215, 47 },
		{ 63, 216, 47 },
		{ 63, 217, 47 },
		{ 63, 218, 47 },
		{ 63, 219, 47 },
		{ 63, 220, 47 },
		{ 63, 221, 47 },
		{ 63, 222, 47 },
		{ 63, 223, 47 },
		{ 63, 224, 47 },
		{ 63, 225, 47 },
		{ 63, 226, 47 },
		{ 63, 227, 47 },
		{ 63, 228, 47 },
		{ 63, 229, 47 },
		{ 63, 230, 47 },
		{ 63, 231, 47 },
		{ 63, 232, 47 },
		{ 63, 233, 47 },
		{ 63, 234, 47 },
		{ 63, 235, 47 },
		{ 63, 236, 47 },
		{ 63, 237, 47 },
		{ 63, 238, 47 },
		{ 63, 239, 47 },
		{ 63, 240, 47 },
		{ 63, 241, 47 },
		{ 63, 242, 47 },
		{ 63, 243, 47 },
		{ 63, 244, 47 },
		{ 63, 245, 47 },
		{ 63, 246, 47 },
		{ 63, 247, 47 },
		{ 63, 248, 47 },
		{ 63, 249, 47 },
		{ 63, 250, 47 },
		{ 63, 251, 47 },
		{ 63, 252, 47 },
		{ 63, 253, 47 },
		{ 63, 254, 47 },
		{ 63, 255, 47 },
		{ 63, 256, 47 },
		{ 64, 209, 47 },
		{ 64, 210, 47 },
		{ 64, 211, 47 },
		{ 64, 212, 47 },
		{ 64, 213, 47 },
		{ 64, 214, 47 },
		{ 64, 215, 47 },
		{ 64, 216, 47 },
		{ 64, 217, 47 },
		{ 64, 218, 47 },
		{ 64, 219, 47 },
		{ 64, 220, 47 },
		{ 64, 221, 47 },
		{ 64, 222, 47 },
		{ 64, 223, 47 },
		{ 64, 224, 47 },
		{ 64, 225, 47 },
		{ 64, 226, 47 },
		{ 64, 227, 47 },
		{ 64, 228, 47 },
		{ 64, 229, 47 },
		{ 64, 230, 47 },
		{ 64, 231, 47 },
		{ 64, 232, 47 },
		{ 64, 233, 47 },
		{ 64, 234, 47 },
		{ 64, 235, 47 },
		{ 64, 236, 47 },
		{ 64, 237, 47 },
		{ 64, 238, 47 },
		{ 64, 239, 47 },
		{ 64, 240, 47 },
		{ 64, 241, 47 },
		{ 64, 242, 47 },
		{ 64, 243, 47 },
		{ 64, 244, 47 },
		{ 64, 245, 47 },
		{ 64, 246, 47 },
		{ 64, 247, 47 },
		{ 64, 248, 47 },
		{ 64, 249, 47 },
		{ 64, 250, 47 },
		{ 64, 251, 47 },
		{ 64, 252, 47 },
		{ 64, 253, 47 },
		{ 64, 254, 47 },
		{ 64, 255, 47 },
		{ 64, 256, 47 },
		{ 65, 210, 47 },
		{ 65, 211, 47 },
		{ 65, 212, 47 },
		{ 65, 213, 47 },
		{ 65, 214, 47 },
		{ 65, 215, 47 },
		{ 65, 216, 47 },
		{ 65, 217, 47 },
		{ 65, 218, 47 },
		{ 65, 219, 47 },
		{ 65, 220, 47 },
		{ 65, 221, 47 },
		{ 65, 222, 47 },
		{ 65, 223, 47 },
		{ 65, 224, 47 },
		{ 65, 225, 47 },
		{ 65, 226, 47 },
		{ 65, 227, 47 },
		{ 65, 228, 47 },
		{ 65, 229, 47 },
		{ 65, 230, 47 },
		{ 65, 231, 47 },
		{ 65, 232, 47 },
		{ 65, 233, 47 },
		{ 65, 234, 47 },
		{ 65, 235, 47 },
		{ 65, 236, 47 },
		{ 65, 237, 47 },
		{ 65, 238, 47 },
		{ 65, 239, 47 },
		{ 65, 240, 47 },
		{ 65, 241, 47 },
		{ 65, 242, 47 },
		{ 65, 243, 47 },
		{ 65, 244, 47 },
		{ 65, 245, 47 },
		{ 65, 246, 47 },
		{ 65, 247, 47 },
		{ 65, 248, 47 },
		{ 65, 249, 47 },
		{ 65, 250, 47 },
		{ 65, 251, 47 },
		{ 65, 252, 47 },
		{ 65, 253, 47 },
		{ 65, 254, 47 },
		{ 65, 255, 47 },
		{ 65, 256, 47 },
		{ 66, 210, 47 },
		{ 66, 211, 47 },
		{ 66, 212, 47 },
		{ 66, 213, 47 },
		{ 66, 214, 47 },
		{ 66, 215, 47 },
		{ 66, 216, 47 },
		{ 66, 217, 47 },
		{ 66, 218, 47 },
		{ 66, 219, 47 },
		{ 66, 220, 47 },
		{ 66, 221, 47 },
		{ 66, 222, 47 },
		{ 66, 223, 47 },
		{ 66, 224, 47 },
		{ 66, 225, 47 },
		{ 66, 226, 47 },
		{ 66, 227, 47 },
		{ 66, 228, 47 },
		{ 66, 229, 47 },
		{ 66, 230, 47 },
		{ 66, 231, 47 },
		{ 66, 232, 47 },
		{ 66, 233, 47 },
		{ 66, 234, 47 },
		{ 66, 235, 47 },
		{ 66, 236, 47 },
		{ 66, 237, 47 },
		{ 66, 238, 47 },
		{ 66, 239, 47 },
		{ 66, 240, 47 },
		{ 66, 241, 47 },
		{ 66, 242, 47 },
		{ 66, 243, 47 },
		{ 66, 244, 47 },
		{ 66, 245, 47 },
		{ 66, 246, 47 },
		{ 66, 247, 47 },
		{ 66, 248, 47 },
		{ 66, 249, 47 },
		{ 66, 250, 47 },
		{ 66, 251, 47 },
		{ 66, 252, 47 },
		{ 66, 253, 47 },
		{ 66, 254, 47 },
		{ 66, 255, 47 },
		{ 66, 256, 47 },
		{ 67, 210, 47 },
		{ 67, 211, 47 },
		{ 67, 212, 47 },
		{ 67, 213, 47 },
		{ 67, 214, 47 },
		{ 67, 215, 47 },
		{ 67, 216, 47 },
		{ 67, 217, 47 },
		{ 67, 218, 47 },
		{ 67, 219, 47 },
		{ 67, 220, 47 },
		{ 67, 221, 47 },
		{ 67, 222, 47 },
		{ 67, 223, 47 },
		{ 67, 224, 47 },
		{ 67, 225, 47 },
		{ 67, 226, 47 },
		{ 67, 227, 47 },
		{ 67, 228, 47 },
		{ 67, 229, 47 },
		{ 67, 230, 47 },
		{ 67, 231, 47 },
		{ 67, 232, 47 },
		{ 67, 233, 47 },
		{ 67, 234, 47 },
		{ 67, 235, 47 },
		{ 67, 236, 47 },
		{ 67, 237, 47 },
		{ 67, 238, 47 },
		{ 67, 239, 47 },
		{ 67, 240, 47 },
		{ 67, 241, 47 },
		{ 67, 242, 47 },
		{ 67, 243, 47 },
		{ 67, 244, 47 },
		{ 67, 245, 47 },
		{ 67, 246, 47 },
		{ 67, 247, 47 },
		{ 67, 248, 47 },
		{ 67, 249, 47 },
		{ 67, 250, 47 },
		{ 67, 251, 47 },
		{ 67, 252, 47 },
		{ 67, 253, 47 },
		{ 67, 254, 47 },
		{ 67, 255, 47 },
		{ 67, 256, 47 },
		{ 68, 210, 47 },
		{ 68, 211, 47 },
		{ 68, 212, 47 },
		{ 68, 213, 47 },
		{ 68, 214, 47 },
		{ 68, 215, 47 },
		{ 68, 216, 47 },
		{ 68, 217, 47 },
		{ 68, 218, 47 },
		{ 68, 219, 47 },
		{ 68, 220, 47 },
		{ 68, 221, 47 },
		{ 68, 222, 47 },
		{ 68, 223, 47 },
		{ 68, 224, 47 },
		{ 68, 225, 47 },
		{ 68, 226, 47 },
		{ 68, 227, 47 },
		{ 68, 228, 47 },
		{ 68, 229, 47 },
		{ 68, 230, 47 },
		{ 68, 231, 47 },
		{ 68, 232, 47 },
		{ 68, 233, 47 },
		{ 68, 234, 47 },
		{ 68, 235, 47 },
		{ 68, 236, 47 },
		{ 68, 237, 47 },
		{ 68, 238, 47 },
		{ 68, 239, 47 },
		{ 68, 240, 47 },
		{ 68, 241, 47 },
		{ 68, 242, 47 },
		{ 68, 243, 47 },
		{ 68, 244, 47 },
		{ 68, 245, 47 },
		{ 68, 246, 47 },
		{ 68, 247, 47 },
		{ 68, 248, 47 },
		{ 68, 249, 47 },
		{ 68, 250, 47 },
		{ 68, 251, 47 },
		{ 68, 252, 47 },
		{ 68, 253, 47 },
		{ 68, 254, 47 },
		{ 68, 255, 47 },
		{ 68, 256, 47 },
		{ 69, 211, 47 },
		{ 69, 212, 47 },
		{ 69, 213, 47 },
		{ 69, 214, 47 },
		{ 69, 215, 47 },
		{ 69, 216, 47 },
		{ 69, 217, 47 },
		{ 69, 218, 47 },
		{ 69, 219, 47 },
		{ 69, 220, 47 },
		{ 69, 221, 47 },
		{ 69, 222, 47 },
		{ 69, 223, 47 },
		{ 69, 224, 47 },
		{ 69, 225, 47 },
		{ 69, 226, 47 },
		{ 69, 227, 47 },
		{ 69, 228, 47 },
		{ 69, 229, 47 },
		{ 69, 230, 47 },
		{ 69, 231, 47 },
		{ 69, 232, 47 },
		{ 69, 233, 47 },
		{ 69, 234, 47 },
		{ 69, 235, 47 },
		{ 69, 236, 47 },
		{ 69, 237, 47 },
		{ 69, 238, 47 },
		{ 69, 239, 47 },
		{ 69, 240, 47 },
		{ 69, 241, 47 },
		{ 69, 242, 47 },
		{ 69, 243, 47 },
		{ 69, 244, 47 },
		{ 69, 245, 47 },
		{ 69, 246, 47 },
		{ 69, 247, 47 },
		{ 69, 248, 47 },
		{ 69, 249, 47 },
		{ 69, 250, 47 },
		{ 69, 251, 47 },
		{ 69, 252, 47 },
		{ 69, 253, 47 },
		{ 69, 254, 47 },
		{ 69, 255, 47 },
		{ 69, 256, 47 },
		{ 70, 211, 47 },
		{ 70, 212, 47 },
		{ 70, 213, 47 },
		{ 70, 214, 47 },
		{ 70, 215, 47 },
		{ 70, 216, 47 },
		{ 70, 217, 47 },
		{ 70, 218, 47 },
		{ 70, 219, 47 },
		{ 70, 220, 47 },
		{ 70, 221, 47 },
		{ 70, 222, 47 },
		{ 70, 223, 47 },
		{ 70, 224, 47 },
		{ 70, 225, 47 },
		{ 70, 226, 47 },
		{ 70, 227, 47 },
		{ 70, 228, 47 },
		{ 70, 229, 47 },
		{ 70, 230, 47 },
		{ 70, 231, 47 },
		{ 70, 232, 47 },
		{ 70, 233, 47 },
		{ 70, 234, 47 },
		{ 70, 235, 47 },
		{ 70, 236, 47 },
		{ 70, 237, 47 },
		{ 70, 238, 47 },
		{ 70, 239, 47 },
		{ 70, 240, 47 },
		{ 70, 241, 47 },
		{ 70, 242, 47 },
		{ 70, 243, 47 },
		{ 70, 244, 47 },
		{ 70, 245, 47 },
		{ 70, 246, 47 },
		{ 70, 247, 47 },
		{ 70, 248, 47 },
		{ 70, 249, 47 },
		{ 70, 250, 47 },
		{ 70, 251, 47 },
		{ 70, 252, 47 },
		{ 70, 253, 47 },
		{ 70, 254, 47 },
		{ 70, 255, 47 },
		{ 70, 256, 47 },
		{ 71, 211, 47 },
		{ 71, 212, 47 },
		{ 71, 213, 47 },
		{ 71, 214, 47 },
		{ 71, 215, 47 },
		{ 71, 216, 47 },
		{ 71, 217, 47 },
		{ 71, 218, 47 },
		{ 71, 219, 47 },
		{ 71, 220, 47 },
		{ 71, 221, 47 },
		{ 71, 222, 47 },
		{ 71, 223, 47 },
		{ 71, 224, 47 },
		{ 71, 225, 47 },
		{ 71, 226, 47 },
		{ 71, 227, 47 },
		{ 71, 228, 47 },
		{ 71, 229, 47 },
		{ 71, 230, 47 },
		{ 71, 231, 47 },
		{ 71, 232, 47 },
		{ 71, 233, 47 },
		{ 71, 234, 47 },
		{ 71, 235, 47 },
		{ 71, 236, 47 },
		{ 71, 237, 47 },
		{ 71, 238, 47 },
		{ 71, 239, 47 },
		{ 71, 240, 47 },
		{ 71, 241, 47 },
		{ 71, 242, 47 },
		{ 71, 243, 47 },
		{ 71, 244, 47 },
		{ 71, 245, 47 },
		{ 71, 246, 47 },
		{ 71, 247, 47 },
		{ 71, 248, 47 },
		{ 71, 249, 47 },
		{ 71, 250, 47 },
		{ 71, 251, 47 },
		{ 71, 252, 47 },
		{ 71, 253, 47 },
		{ 71, 254, 47 },
		{ 71, 255, 47 },
		{ 71, 256, 47 },
		{ 72, 211, 47 },
		{ 72, 212, 47 },
		{ 72, 213, 47 },
		{ 72, 214, 47 },
		{ 72, 215, 47 },
		{ 72, 216, 47 },
		{ 72, 217, 47 },
		{ 72, 218, 47 },
		{ 72, 219, 47 },
		{ 72, 220, 47 },
		{ 72, 221, 47 },
		{ 72, 222, 47 },
		{ 72, 223, 47 },
		{ 72, 224, 47 },
		{ 72, 225, 47 },
		{ 72, 226, 47 },
		{ 72, 227, 47 },
		{ 72, 228, 47 },
		{ 72, 229, 47 },
		{ 72, 230, 47 },
		{ 72, 231, 47 },
		{ 72, 232, 47 },
		{ 72, 233, 47 },
		{ 72, 234, 47 },
		{ 72, 235, 47 },
		{ 72, 236, 47 },
		{ 72, 237, 47 },
		{ 72, 238, 47 },
		{ 72, 239, 47 },
		{ 72, 240, 47 },
		{ 72, 241, 47 },
		{ 72, 242, 47 },
		{ 72, 243, 47 },
		{ 72, 244, 47 },
		{ 72, 245, 47 },
		{ 72, 246, 47 },
		{ 72, 247, 47 },
		{ 72, 248, 47 },
		{ 72, 249, 47 },
		{ 72, 250, 47 },
		{ 72, 251, 47 },
		{ 72, 252, 47 },
		{ 72, 253, 47 },
		{ 72, 254, 47 },
		{ 72, 255, 47 },
		{ 72, 256, 47 },
		{ 73, 212, 47 },
		{ 73, 213, 47 },
		{ 73, 214, 47 },
		{ 73, 215, 47 },
		{ 73, 216, 47 },
		{ 73, 217, 47 },
		{ 73, 218, 47 },
		{ 73, 219, 47 },
		{ 73, 220, 47 },
		{ 73, 221, 47 },
		{ 73, 222, 47 },
		{ 73, 223, 47 },
		{ 73, 224, 47 },
		{ 73, 225, 47 },
		{ 73, 226, 47 },
		{ 73, 227, 47 },
		{ 73, 228, 47 },
		{ 73, 229, 47 },
		{ 73, 230, 47 },
		{ 73, 231, 47 },
		{ 73, 232, 47 },
		{ 73, 233, 47 },
		{ 73, 234, 47 },
		{ 73, 235, 47 },
		{ 73, 236, 47 },
		{ 73, 237, 47 },
		{ 73, 238, 47 },
		{ 73, 239, 47 },
		{ 73, 240, 47 },
		{ 73, 241, 47 },
		{ 73, 242, 47 },
		{ 73, 243, 47 },
		{ 73, 244, 47 },
		{ 73, 245, 47 },
		{ 73, 246, 47 },
		{ 73, 247, 47 },
		{ 73, 248, 47 },
		{ 73, 249, 47 },
		{ 73, 250, 47 },
		{ 73, 251, 47 },
		{ 73, 252, 47 },
		{ 73, 253, 47 },
		{ 73, 254, 47 },
		{ 73, 255, 47 },
		{ 73, 256, 47 },
		{ 74, 212, 47 },
		{ 74, 213, 47 },
		{ 74, 214, 47 },
		{ 74, 215, 47 },
		{ 74, 216, 47 },
		{ 74, 217, 47 },
		{ 74, 218, 47 },
		{ 74, 219, 47 },
		{ 74, 220, 47 },
		{ 74, 221, 47 },
		{ 74, 222, 47 },
		{ 74, 223, 47 },
		{ 74, 224, 47 },
		{ 74, 225, 47 },
		{ 74, 226, 47 },
		{ 74, 227, 47 },
		{ 74, 228, 47 },
		{ 74, 229, 47 },
		{ 74, 230, 47 },
		{ 74, 231, 47 },
		{ 74, 232, 47 },
		{ 74, 233, 47 },
		{ 74, 234, 47 },
		{ 74, 235, 47 },
		{ 74, 236, 47 },
		{ 74, 237, 47 },
		{ 74, 238, 47 },
		{ 74, 239, 47 },
		{ 74, 240, 47 },
		{ 74, 241, 47 },
		{ 74, 242, 47 },
		{ 74, 243, 47 },
		{ 74, 244, 47 },
		{ 74, 245, 47 },
		{ 74, 246, 47 },
		{ 74, 247, 47 },
		{ 74, 248, 47 },
		{ 74, 249, 47 },
		{ 74, 250, 47 },
		{ 74, 251, 47 },
		{ 74, 252, 47 },
		{ 74, 253, 47 },
		{ 74, 254, 47 },
		{ 74, 255, 47 },
		{ 74, 256, 47 },
		{ 75, 212, 47 },
		{ 75, 213, 47 },
		{ 75, 214, 47 },
		{ 75, 215, 47 },
		{ 75, 216, 47 },
		{ 75, 217, 47 },
		{ 75, 218, 47 },
		{ 75, 219, 47 },
		{ 75, 220, 47 },
		{ 75, 221, 47 },
		{ 75, 222, 47 },
		{ 75, 223, 47 },
		{ 75, 224, 47 },
		{ 75, 225, 47 },
		{ 75, 226, 47 },
		{ 75, 227, 47 },
		{ 75, 228, 47 },
		{ 75, 229, 47 },
		{ 75, 230, 47 },
		{ 75, 231, 47 },
		{ 75, 232, 47 },
		{ 75, 233, 47 },
		{ 75, 234, 47 },
		{ 75, 235, 47 },
		{ 75, 236, 47 },
		{ 75, 237, 47 },
		{ 75, 238, 47 },
		{ 75, 239, 47 },
		{ 75, 240, 47 },
		{ 75, 241, 47 },
		{ 75, 242, 47 },
		{ 75, 243, 47 },
		{ 75, 244, 47 },
		{ 75, 245, 47 },
		{ 75, 246, 47 },
		{ 75, 247, 47 },
		{ 75, 248, 47 },
		{ 75, 249, 47 },
		{ 75, 250, 47 },
		{ 75, 251, 47 },
		{ 75, 252, 47 },
		{ 75, 253, 47 },
		{ 75, 254, 47 },
		{ 75, 255, 47 },
		{ 75, 256, 47 },
		{ 76, 212, 47 },
		{ 76, 213, 47 },
		{ 76, 214, 47 },
		{ 76, 215, 47 },
		{ 76, 216, 47 },
		{ 76, 217, 47 },
		{ 76, 218, 47 },
		{ 76, 219, 47 },
		{ 76, 220, 47 },
		{ 76, 221, 47 },
		{ 76, 222, 47 },
		{ 76, 223, 47 },
		{ 76, 224, 47 },
		{ 76, 225, 47 },
		{ 76, 226, 47 },
		{ 76, 227, 47 },
		{ 76, 228, 47 },
		{ 76, 229, 47 },
		{ 76, 230, 47 },
		{ 76, 231, 47 },
		{ 76, 232, 47 },
		{ 76, 233, 47 },
		{ 76, 234, 47 },
		{ 76, 235, 47 },
		{ 76, 236, 47 },
		{ 76, 237, 47 },
		{ 76, 238, 47 },
		{ 76, 239, 47 },
		{ 76, 240, 47 },
		{ 76, 241, 47 },
		{ 76, 242, 47 },
		{ 76, 243, 47 },
		{ 76, 244, 47 },
		{ 76, 245, 47 },
		{ 76, 246, 47 },
		{ 76, 247, 47 },
		{ 76, 248, 47 },
		{ 76, 249, 47 },
		{ 76, 250, 47 },
		{ 76, 251, 47 },
		{ 76, 252, 47 },
		{ 76, 253, 47 },
		{ 76, 254, 47 },
		{ 76, 255, 47 },
		{ 76, 256, 47 },
		{ 77, 213, 47 },
		{ 77, 214, 47 },
		{ 77, 215, 47 },
		{ 77, 216, 47 },
		{ 77, 217, 47 },
		{ 77, 218, 47 },
		{ 77, 219, 47 },
		{ 77, 220, 47 },
		{ 77, 221, 47 },
		{ 77, 222, 47 },
		{ 77, 223, 47 },
		{ 77, 224, 47 },
		{ 77, 225, 47 },
		{ 77, 226, 47 },
		{ 77, 227, 47 },
		{ 77, 228, 47 },
		{ 77, 229, 47 },
		{ 77, 230, 47 },
		{ 77, 231, 47 },
		{ 77, 232, 47 },
		{ 77, 233, 47 },
		{ 77, 234, 47 },
		{ 77, 235, 47 },
		{ 77, 236, 47 },
		{ 77, 237, 47 },
		{ 77, 238, 47 },
		{ 77, 239, 47 },
		{ 77, 240, 47 },
		{ 77, 241, 47 },
		{ 77, 242, 47 },
		{ 77, 243, 47 },
		{ 77, 244, 47 },
		{ 77, 245, 47 },
		{ 77, 246, 47 },
		{ 77, 247, 47 },
		{ 77, 248, 47 },
		{ 77, 249, 47 },
		{ 77, 250, 47 },
		{ 77, 251, 47 },
		{ 77, 252, 47 },
		{ 77, 253, 47 },
		{ 77, 254, 47 },
		{ 77, 255, 47 },
		{ 77, 256, 47 },
		{ 78, 213, 47 },
		{ 78, 214, 47 },
		{ 78, 215, 47 },
		{ 78, 216, 47 },
		{ 78, 217, 47 },
		{ 78, 218, 47 },
		{ 78, 219, 47 },
		{ 78, 220, 47 },
		{ 78, 221, 47 },
		{ 78, 222, 47 },
		{ 78, 223, 47 },
		{ 78, 224, 47 },
		{ 78, 225, 47 },
		{ 78, 226, 47 },
		{ 78, 227, 47 },
		{ 78, 228, 47 },
		{ 78, 229, 47 },
		{ 78, 230, 47 },
		{ 78, 231, 47 },
		{ 78, 232, 47 },
		{ 78, 233, 47 },
		{ 78, 234, 47 },
		{ 78, 235, 47 },
		{ 78, 236, 47 },
		{ 78, 237, 47 },
		{ 78, 238, 47 },
		{ 78, 239, 47 },
		{ 78, 240, 47 },
		{ 78, 241, 47 },
		{ 78, 242, 47 },
		{ 78, 243, 47 },
		{ 78, 244, 47 },
		{ 78, 245, 47 },
		{ 78, 246, 47 },
		{ 78, 247, 47 },
		{ 78, 248, 47 },
		{ 78, 249, 47 },
		{ 78, 250, 47 },
		{ 78, 251, 47 },
		{ 78, 252, 47 },
		{ 78, 253, 47 },
		{ 78, 254, 47 },
		{ 78, 255, 47 },
		{ 78, 256, 47 },
		{ 79, 213, 47 },
		{ 79, 214, 47 },
		{ 79, 215, 47 },
		{ 79, 216, 47 },
		{ 79, 217, 47 },
		{ 79, 218, 47 },
		{ 79, 219, 47 },
		{ 79, 220, 47 },
		{ 79, 221, 47 },
		{ 79, 222, 47 },
		{ 79, 223, 47 },
		{ 79, 224, 47 },
		{ 79, 225, 47 },
		{ 79, 226, 47 },
		{ 79, 227, 47 },
		{ 79, 228, 47 },
		{ 79, 229, 47 },
		{ 79, 230, 47 },
		{ 79, 231, 47 },
		{ 79, 232, 47 },
		{ 79, 233, 47 },
		{ 79, 234, 47 },
		{ 79, 235, 47 },
		{ 79, 236, 47 },
		{ 79, 237, 47 },
		{ 79, 238, 47 },
		{ 79, 239, 47 },
		{ 79, 240, 47 },
		{ 79, 241, 47 },
		{ 79, 242, 47 },
		{ 79, 243, 47 },
		{ 79, 244, 47 },
		{ 79, 245, 47 },
		{ 79, 246, 47 },
		{ 79, 247, 47 },
		{ 79, 248, 47 },
		{ 79, 249, 47 },
		{ 79, 250, 47 },
		{ 79, 251, 47 },
		{ 79, 252, 47 },
		{ 79, 253, 47 },
		{ 79, 254, 47 },
		{ 79, 255, 47 },
		{ 79, 256, 47 },
		{ 80, 213, 47 },
		{ 80, 214, 47 },
		{ 80, 215, 47 },
		{ 80, 216, 47 },
		{ 80, 217, 47 },
		{ 80, 218, 47 },
		{ 80, 219, 47 },
		{ 80, 220, 47 },
		{ 80, 221, 47 },
		{ 80, 222, 47 },
		{ 80, 223, 47 },
		{ 80, 224, 47 },
		{ 80, 225, 47 },
		{ 80, 226, 47 },
		{ 80, 227, 47 },
		{ 80, 228, 47 },
		{ 80, 229, 47 },
		{ 80, 230, 47 },
		{ 80, 231, 47 },
		{ 80, 232, 47 },
		{ 80, 233, 47 },
		{ 80, 234, 47 },
		{ 80, 235, 47 },
		{ 80, 236, 47 },
		{ 80, 237, 47 },
		{ 80, 238, 47 },
		{ 80, 239, 47 },
		{ 80, 240, 47 },
		{ 80, 241, 47 },
		{ 80, 242, 47 },
		{ 80, 243, 47 },
		{ 80, 244, 47 },
		{ 80, 245, 47 },
		{ 80, 246, 47 },
		{ 80, 247, 47 },
		{ 80, 248, 47 },
		{ 80, 249, 47 },
		{ 80, 250, 47 },
		{ 80, 251, 47 },
		{ 80, 252, 47 },
		{ 80, 253, 47 },
		{ 80, 254, 47 },
		{ 80, 255, 47 },
		{ 80, 256, 47 },
		{ 81, 214, 47 },
		{ 81, 215, 47 },
		{ 81, 216, 47 },
		{ 81, 217, 47 },
		{ 81, 218, 47 },
		{ 81, 219, 47 },
		{ 81, 220, 47 },
		{ 81, 221, 47 },
		{ 81, 222, 47 },
		{ 81, 223, 47 },
		{ 81, 224, 47 },
		{ 81, 225, 47 },
		{ 81, 226, 47 },
		{ 81, 227, 47 },
		{ 81, 228, 47 },
		{ 81, 229, 47 },
		{ 81, 230, 47 },
		{ 81, 231, 47 },
		{ 81, 232, 47 },
		{ 81, 233, 47 },
		{ 81, 234, 47 },
		{ 81, 235, 47 },
		{ 81, 236, 47 },
		{ 81, 237, 47 },
		{ 81, 238, 47 },
		{ 81, 239, 47 },
		{ 81, 240, 47 },
		{ 81, 241, 47 },
		{ 81, 242, 47 },
		{ 81, 243, 47 },
		{ 81, 244, 47 },
		{ 81, 245, 47 },
		{ 81, 246, 47 },
		{ 81, 247, 47 },
		{ 81, 248, 47 },
		{ 81, 249, 47 },
		{ 81, 250, 47 },
		{ 81, 251, 47 },
		{ 81, 252, 47 },
		{ 81, 253, 47 },
		{ 81, 254, 47 },
		{ 81, 255, 47 },
		{ 81, 256, 47 },
		{ 82, 214, 47 },
		{ 82, 215, 47 },
		{ 82, 216, 47 },
		{ 82, 217, 47 },
		{ 82, 218, 47 },
		{ 82, 219, 47 },
		{ 82, 220, 47 },
		{ 82, 221, 47 },
		{ 82, 222, 47 },
		{ 82, 223, 47 },
		{ 82, 224, 47 },
		{ 82, 225, 47 },
		{ 82, 226, 47 },
		{ 82, 227, 47 },
		{ 82, 228, 47 },
		{ 82, 229, 47 },
		{ 82, 230, 47 },
		{ 82, 231, 47 },
		{ 82, 232, 47 },
		{ 82, 233, 47 },
		{ 82, 234, 47 },
		{ 82, 235, 47 },
		{ 82, 236, 47 },
		{ 82, 237, 47 },
		{ 82, 238, 47 },
		{ 82, 239, 47 },
		{ 82, 240, 47 },
		{ 82, 241, 47 },
		{ 82, 242, 47 },
		{ 82, 243, 47 },
		{ 82, 244, 47 },
		{ 82, 245, 47 },
		{ 82, 246, 47 },
		{ 82, 247, 47 },
		{ 82, 248, 47 },
		{ 82, 249, 47 },
		{ 82, 250, 47 },
		{ 82, 251, 47 },
		{ 82, 252, 47 },
		{ 82, 253, 47 },
		{ 82, 254, 47 },
		{ 82, 255, 47 },
		{ 82, 256, 47 },
		{ 83, 214, 47 },
		{ 83, 215, 47 },
		{ 83, 216, 47 },
		{ 83, 217, 47 },
		{ 83, 218, 47 },
		{ 83, 219, 47 },
		{ 83, 220, 47 },
		{ 83, 221, 47 },
		{ 83, 222, 47 },
		{ 83, 223, 47 },
		{ 83, 224, 47 },
		{ 83, 225, 47 },
		{ 83, 226, 47 },
		{ 83, 227, 47 },
		{ 83, 228, 47 },
		{ 83, 229, 47 },
		{ 83, 230, 47 },
		{ 83, 231, 47 },
		{ 83, 232, 47 },
		{ 83, 233, 47 },
		{ 83, 234, 47 },
		{ 83, 235, 47 },
		{ 83, 236, 47 },
		{ 83, 237, 47 },
		{ 83, 238, 47 },
		{ 83, 239, 47 },
		{ 83, 240, 47 },
		{ 83, 241, 47 },
		{ 83, 242, 47 },
		{ 83, 243, 47 },
		{ 83, 244, 47 },
		{ 83, 245, 47 },
		{ 83, 246, 47 },
		{ 83, 247, 47 },
		{ 83, 248, 47 },
		{ 83, 249, 47 },
		{ 83, 250, 47 },
		{ 83, 251, 47 },
		{ 83, 252, 47 },
		{ 83, 253, 47 },
		{ 83, 254, 47 },
		{ 83, 255, 47 },
		{ 83, 256, 47 },
		{ 84, 214, 47 },
		{ 84, 215, 47 },
		{ 84, 216, 47 },
		{ 84, 217, 47 },
		{ 84, 218, 47 },
		{ 84, 219, 47 },
		{ 84, 220, 47 },
		{ 84, 221, 47 },
		{ 84, 222, 47 },
		{ 84, 223, 47 },
		{ 84, 224, 47 },
		{ 84, 225, 47 },
		{ 84, 226, 47 },
		{ 84, 227, 47 },
		{ 84, 228, 47 },
		{ 84, 229, 47 },
		{ 84, 230, 47 },
		{ 84, 231, 47 },
		{ 84, 232, 47 },
		{ 84, 233, 47 },
		{ 84, 234, 47 },
		{ 84, 235, 47 },
		{ 84, 236, 47 },
		{ 84, 237, 47 },
		{ 84, 238, 47 },
		{ 84, 239, 47 },
		{ 84, 240, 47 },
		{ 84, 241, 47 },
		{ 84, 242, 47 },
		{ 84, 243, 47 },
		{ 84, 244, 47 },
		{ 84, 245, 47 },
		{ 84, 246, 47 },
		{ 84, 247, 47 },
		{ 84, 248, 47 },
		{ 84, 249, 47 },
		{ 84, 250, 47 },
		{ 84, 251, 47 },
		{ 84, 252, 47 },
		{ 84, 253, 47 },
		{ 84, 254, 47 },
		{ 84, 255, 47 },
		{ 84, 256, 47 },
		{ 85, 215, 47 },
		{ 85, 216, 47 },
		{ 85, 217, 47 },
		{ 85, 218, 47 },
		{ 85, 219, 47 },
		{ 85, 220, 47 },
		{ 85, 221, 47 },
		{ 85, 222, 47 },
		{ 85, 223, 47 },
		{ 85, 224, 47 },
		{ 85, 225, 47 },
		{ 85, 226, 47 },
		{ 85, 227, 47 },
		{ 85, 228, 47 },
		{ 85, 229, 47 },
		{ 85, 230, 47 },
		{ 85, 231, 47 },
		{ 85, 232, 47 },
		{ 85, 233, 47 },
		{ 85, 234, 47 },
		{ 85, 235, 47 },
		{ 85, 236, 47 },
		{ 85, 237, 47 },
		{ 85, 238, 47 },
		{ 85, 239, 47 },
		{ 85, 240, 47 },
		{ 85, 241, 47 },
		{ 85, 242, 47 },
		{ 85, 243, 47 },
		{ 85, 244, 47 },
		{ 85, 245, 47 },
		{ 85, 246, 47 },
		{ 85, 247, 47 },
		{ 85, 248, 47 },
		{ 85, 249, 47 },
		{ 85, 250, 47 },
		{ 85, 251, 47 },
		{ 85, 252, 47 },
		{ 85, 253, 47 },
		{ 85, 254, 47 },
		{ 85, 255, 47 },
		{ 85, 256, 47 },
		{ 86, 215, 47 },
		{ 86, 216, 47 },
		{ 86, 217, 47 },
		{ 86, 218, 47 },
		{ 86, 219, 47 },
		{ 86, 220, 47 },
		{ 86, 221, 47 },
		{ 86, 222, 47 },
		{ 86, 223, 47 },
		{ 86, 224, 47 },
		{ 86, 225, 47 },
		{ 86, 226, 47 },
		{ 86, 227, 47 },
		{ 86, 228, 47 },
		{ 86, 229, 47 },
		{ 86, 230, 47 },
		{ 86, 231, 47 },
		{ 86, 232, 47 },
		{ 86, 233, 47 },
		{ 86, 234, 47 },
		{ 86, 235, 47 },
		{ 86, 236, 47 },
		{ 86, 237, 47 },
		{ 86, 238, 47 },
		{ 86, 239, 47 },
		{ 86, 240, 47 },
		{ 86, 241, 47 },
		{ 86, 242, 47 },
		{ 86, 243, 47 },
		{ 86, 244, 47 },
		{ 86, 245, 47 },
		{ 86, 246, 47 },
		{ 86, 247, 47 },
		{ 86, 248, 47 },
		{ 86, 249, 47 },
		{ 86, 250, 47 },
		{ 86, 251, 47 },
		{ 86, 252, 47 },
		{ 86, 253, 47 },
		{ 86, 254, 47 },
		{ 86, 255, 47 },
		{ 86, 256, 47 },
		{ 87, 215, 47 },
		{ 87, 216, 47 },
		{ 87, 217, 47 },
		{ 87, 218, 47 },
		{ 87, 219, 47 },
		{ 87, 220, 47 },
		{ 87, 221, 47 },
		{ 87, 222, 47 },
		{ 87, 223, 47 },
		{ 87, 224, 47 },
		{ 87, 225, 47 },
		{ 87, 226, 47 },
		{ 87, 227, 47 },
		{ 87, 228, 47 },
		{ 87, 229, 47 },
		{ 87, 230, 47 },
		{ 87, 231, 47 },
		{ 87, 232, 47 },
		{ 87, 233, 47 },
		{ 87, 234, 47 },
		{ 87, 235, 47 },
		{ 87, 236, 47 },
		{ 87, 237, 47 },
		{ 87, 238, 47 },
		{ 87, 239, 47 },
		{ 87, 240, 47 },
		{ 87, 241, 47 },
		{ 87, 242, 47 },
		{ 87, 243, 47 },
		{ 87, 244, 47 },
		{ 87, 245, 47 },
		{ 87, 246, 47 },
		{ 87, 247, 47 },
		{ 87, 248, 47 },
		{ 87, 249, 47 },
		{ 87, 250, 47 },
		{ 87, 251, 47 },
		{ 87, 252, 47 },
		{ 87, 253, 47 },
		{ 87, 254, 47 },
		{ 87, 255, 47 },
		{ 87, 256, 47 },
		{ 88, 215, 47 },
		{ 88, 216, 47 },
		{ 88, 217, 47 },
		{ 88, 218, 47 },
		{ 88, 219, 47 },
		{ 88, 220, 47 },
		{ 88, 221, 47 },
		{ 88, 222, 47 },
		{ 88, 223, 47 },
		{ 88, 224, 47 },
		{ 88, 225, 47 },
		{ 88, 226, 47 },
		{ 88, 227, 47 },
		{ 88, 228, 47 },
		{ 88, 229, 47 },
		{ 88, 230, 47 },
		{ 88, 231, 47 },
		{ 88, 232, 47 },
		{ 88, 233, 47 },
		{ 88, 234, 47 },
		{ 88, 235, 47 },
		{ 88, 236, 47 },
		{ 88, 237, 47 },
		{ 88, 238, 47 },
		{ 88, 239, 47 },
		{ 88, 240, 47 },
		{ 88, 241, 47 },
		{ 88, 242, 47 },
		{ 88, 243, 47 },
		{ 88, 244, 47 },
		{ 88, 245, 47 },
		{ 88, 246, 47 },
		{ 88, 247, 47 },
		{ 88, 248, 47 },
		{ 88, 249, 47 },
		{ 88, 250, 47 },
		{ 88, 251, 47 },
		{ 88, 252, 47 },
		{ 88, 253, 47 },
		{ 88, 254, 47 },
		{ 88, 255, 47 },
		{ 88, 256, 47 },
		{ 89, 216, 47 },
		{ 89, 217, 47 },
		{ 89, 218, 47 },
		{ 89, 219, 47 },
		{ 89, 220, 47 },
		{ 89, 221, 47 },
		{ 89, 222, 47 },
		{ 89, 223, 47 },
		{ 89, 224, 47 },
		{ 89, 225, 47 },
		{ 89, 226, 47 },
		{ 89, 227, 47 },
		{ 89, 228, 47 },
		{ 89, 229, 47 },
		{ 89, 230, 47 },
		{ 89, 231, 47 },
		{ 89, 232, 47 },
		{ 89, 233, 47 },
		{ 89, 234, 47 },
		{ 89, 235, 47 },
		{ 89, 236, 47 },
		{ 89, 237, 47 },
		{ 89, 238, 47 },
		{ 89, 239, 47 },
		{ 89, 240, 47 },
		{ 89, 241, 47 },
		{ 89, 242, 47 },
		{ 89, 243, 47 },
		{ 89, 244, 47 },
		{ 89, 245, 47 },
		{ 89, 246, 47 },
		{ 89, 247, 47 },
		{ 89, 248, 47 },
		{ 89, 249, 47 },
		{ 89, 250, 47 },
		{ 89, 251, 47 },
		{ 89, 252, 47 },
		{ 89, 253, 47 },
		{ 89, 254, 47 },
		{ 89, 255, 47 },
		{ 89, 256, 47 },
		{ 90, 216, 47 },
		{ 90, 217, 47 },
		{ 90, 218, 47 },
		{ 90, 219, 47 },
		{ 90, 220, 47 },
		{ 90, 221, 47 },
		{ 90, 222, 47 },
		{ 90, 223, 47 },
		{ 90, 224, 47 },
		{ 90, 225, 47 },
		{ 90, 226, 47 },
		{ 90, 227, 47 },
		{ 90, 228, 47 },
		{ 90, 229, 47 },
		{ 90, 230, 47 },
		{ 90, 231, 47 },
		{ 90, 232, 47 },
		{ 90, 233, 47 },
		{ 90, 234, 47 },
		{ 90, 235, 47 },
		{ 90, 236, 47 },
		{ 90, 237, 47 },
		{ 90, 238, 47 },
		{ 90, 239, 47 },
		{ 90, 240, 47 },
		{ 90, 241, 47 },
		{ 90, 242, 47 },
		{ 90, 243, 47 },
		{ 90, 244, 47 },
		{ 90, 245, 47 },
		{ 90, 246, 47 },
		{ 90, 247, 47 },
		{ 90, 248, 47 },
		{ 90, 249, 47 },
		{ 90, 250, 47 },
		{ 90, 251, 47 },
		{ 90, 252, 47 },
		{ 90, 253, 47 },
		{ 90, 254, 47 },
		{ 90, 255, 47 },
		{ 90, 256, 47 },
		{ 91, 216, 47 },
		{ 91, 217, 47 },
		{ 91, 218, 47 },
		{ 91, 219, 47 },
		{ 91, 220, 47 },
		{ 91, 221, 47 },
		{ 91, 222, 47 },
		{ 91, 223, 47 },
		{ 91, 224, 47 },
		{ 91, 225, 47 },
		{ 91, 226, 47 },
		{ 91, 227, 47 },
		{ 91, 228, 47 },
		{ 91, 229, 47 },
		{ 91, 230, 47 },
		{ 91, 231, 47 },
		{ 91, 232, 47 },
		{ 91, 233, 47 },
		{ 91, 234, 47 },
		{ 91, 235, 47 },
		{ 91, 236, 47 },
		{ 91, 237, 47 },
		{ 91, 238, 47 },
		{ 91, 239, 47 },
		{ 91, 240, 47 },
		{ 91, 241, 47 },
		{ 91, 242, 47 },
		{ 91, 243, 47 },
		{ 91, 244, 47 },
		{ 91, 245, 47 },
		{ 91, 246, 47 },
		{ 91, 247, 47 },
		{ 91, 248, 47 },
		{ 91, 249, 47 },
		{ 91, 250, 47 },
		{ 91, 251, 47 },
		{ 91, 252, 47 },
		{ 91, 253, 47 },
		{ 91, 254, 47 },
		{ 91, 255, 47 },
		{ 91, 256, 47 },
		{ 92, 216, 47 },
		{ 92, 217, 47 },
		{ 92, 218, 47 },
		{ 92, 219, 47 },
		{ 92, 220, 47 },
		{ 92, 221, 47 },
		{ 92, 222, 47 },
		{ 92, 223, 47 },
		{ 92, 224, 47 },
		{ 92, 225, 47 },
		{ 92, 226, 47 },
		{ 92, 227, 47 },
		{ 92, 228, 47 },
		{ 92, 229, 47 },
		{ 92, 230, 47 },
		{ 92, 231, 47 },
		{ 92, 232, 47 },
		{ 92, 233, 47 },
		{ 92, 234, 47 },
		{ 92, 235, 47 },
		{ 92, 236, 47 },
		{ 92, 237, 47 },
		{ 92, 238, 47 },
		{ 92, 239, 47 },
		{ 92, 240, 47 },
		{ 92, 241, 47 },
		{ 92, 242, 47 },
		{ 92, 243, 47 },
		{ 92, 244, 47 },
		{ 92, 245, 47 },
		{ 92, 246, 47 },
		{ 92, 247, 47 },
		{ 92, 248, 47 },
		{ 92, 249, 47 },
		{ 92, 250, 47 },
		{ 92, 251, 47 },
		{ 92, 252, 47 },
		{ 92, 253, 47 },
		{ 92, 254, 47 },
		{ 92, 255, 47 },
		{ 92, 256, 47 },
		{ 93, 217, 47 },
		{ 93, 218, 47 },
		{ 93, 219, 47 },
		{ 93, 220, 47 },
		{ 93, 221, 47 },
		{ 93, 222, 47 },
		{ 93, 223, 47 },
		{ 93, 224, 47 },
		{ 93, 225, 47 },
		{ 93, 226, 47 },
		{ 93, 227, 47 },
		{ 93, 228, 47 },
		{ 93, 229, 47 },
		{ 93, 230, 47 },
		{ 93, 231, 47 },
		{ 93, 232, 47 },
		{ 93, 233, 47 },
		{ 93, 234, 47 },
		{ 93, 235, 47 },
		{ 93, 236, 47 },
		{ 93, 237, 47 },
		{ 93, 238, 47 },
		{ 93, 239, 47 },
		{ 93, 240, 47 },
		{ 93, 241, 47 },
		{ 93, 242, 47 },
		{ 93, 243, 47 },
		{ 93, 244, 47 },
		{ 93, 245, 47 },
		{ 93, 246, 47 },
		{ 93, 247, 47 },
		{ 93, 248, 47 },
		{ 93, 249, 47 },
		{ 93, 250, 47 },
		{ 93, 251, 47 },
		{ 93, 252, 47 },
		{ 93, 253, 47 },
		{ 93, 254, 47 },
		{ 93, 255, 47 },
		{ 93, 256, 47 },
		{ 94, 217, 47 },
		{ 94, 218, 47 },
		{ 94, 219, 47 },
		{ 94, 220, 47 },
		{ 94, 221, 47 },
		{ 94, 222, 47 },
		{ 94, 223, 47 },
		{ 94, 224, 47 },
		{ 94, 225, 47 },
		{ 94, 226, 47 },
		{ 94, 227, 47 },
		{ 94, 228, 47 },
		{ 94, 229, 47 },
		{ 94, 230, 47 },
		{ 94, 231, 47 },
		{ 94, 232, 47 },
		{ 94, 233, 47 },
		{ 94, 234, 47 },
		{ 94, 235, 47 },
		{ 94, 236, 47 },
		{ 94, 237, 47 },
		{ 94, 238, 47 },
		{ 94, 239, 47 },
		{ 94, 240, 47 },
		{ 94, 241, 47 },
		{ 94, 242, 47 },
		{ 94, 243, 47 },
		{ 94, 244, 47 },
		{ 94, 245, 47 },
		{ 94, 246, 47 },
		{ 94, 247, 47 },
		{ 94, 248, 47 },
		{ 94, 249, 47 },
		{ 94, 250, 47 },
		{ 94, 251, 47 },
		{ 94, 252, 47 },
		{ 94, 253, 47 },
		{ 94, 254, 47 },
		{ 94, 255, 47 },
		{ 94, 256, 47 },
		{ 95, 217, 47 },
		{ 95, 218, 47 },
		{ 95, 219, 47 },
		{ 95, 220, 47 },
		{ 95, 221, 47 },
		{ 95, 222, 47 },
		{ 95, 223, 47 },
		{ 95, 224, 47 },
		{ 95, 225, 47 },
		{ 95, 226, 47 },
		{ 95, 227, 47 },
		{ 95, 228, 47 },
		{ 95, 229, 47 },
		{ 95, 230, 47 },
		{ 95, 231, 47 },
		{ 95, 232, 47 },
		{ 95, 233, 47 },
		{ 95, 234, 47 },
		{ 95, 235, 47 },
		{ 95, 236, 47 },
		{ 95, 237, 47 },
		{ 95, 238, 47 },
		{ 95, 239, 47 },
		{ 95, 240, 47 },
		{ 95, 241, 47 },
		{ 95, 242, 47 },
		{ 95, 243, 47 },
		{ 95, 244, 47 },
		{ 95, 245, 47 },
		{ 95, 246, 47 },
		{ 95, 247, 47 },
		{ 95, 248, 47 },
		{ 95, 249, 47 },
		{ 95, 250, 47 },
		{ 95, 251, 47 },
		{ 95, 252, 47 },
		{ 95, 253, 47 },
		{ 95, 254, 47 },
		{ 95, 255, 47 },
		{ 95, 256, 47 },
		{ 96, 217, 47 },
		{ 96, 218, 47 },
		{ 96, 219, 47 },
		{ 96, 220, 47 },
		{ 96, 221, 47 },
		{ 96, 222, 47 },
		{ 96, 223, 47 },
		{ 96, 224, 47 },
		{ 96, 225, 47 },
		{ 96, 226, 47 },
		{ 96, 227, 47 },
		{ 96, 228, 47 },
		{ 96, 229, 47 },
		{ 96, 230, 47 },
		{ 96, 231, 47 },
		{ 96, 232, 47 },
		{ 96, 233, 47 },
		{ 96, 234, 47 },
		{ 96, 235, 47 },
		{ 96, 236, 47 },
		{ 96, 237, 47 },
		{ 96, 238, 47 },
		{ 96, 239, 47 },
		{ 96, 240, 47 },
		{ 96, 241, 47 },
		{ 96, 242, 47 },
		{ 96, 243, 47 },
		{ 96, 244, 47 },
		{ 96, 245, 47 },
		{ 96, 246, 47 },
		{ 96, 247, 47 },
		{ 96, 248, 47 },
		{ 96, 249, 47 },
		{ 96, 250, 47 },
		{ 96, 251, 47 },
		{ 96, 252, 47 },
		{ 96, 253, 47 },
		{ 96, 254, 47 },
		{ 96, 255, 47 },
		{ 96, 256, 47 },
		{ 97, 218, 47 },
		{ 97, 219, 47 },
		{ 97, 220, 47 },
		{ 97, 221, 47 },
		{ 97, 222, 47 },
		{ 97, 223, 47 },
		{ 97, 224, 47 },
		{ 97, 225, 47 },
		{ 97, 226, 47 },
		{ 97, 227, 47 },
		{ 97, 228, 47 },
		{ 97, 229, 47 },
		{ 97, 230, 47 },
		{ 97, 231, 47 },
		{ 97, 232, 47 },
		{ 97, 233, 47 },
		{ 97, 234, 47 },
		{ 97, 235, 47 },
		{ 97, 236, 47 },
		{ 97, 237, 47 },
		{ 97, 238, 47 },
		{ 97, 239, 47 },
		{ 97, 240, 47 },
		{ 97, 241, 47 },
		{ 97, 242, 47 },
		{ 97, 243, 47 },
		{ 97, 244, 47 },
		{ 97, 245, 47 },
		{ 97, 246, 47 },
		{ 97, 247, 47 },
		{ 97, 248, 47 },
		{ 97, 249, 47 },
		{ 97, 250, 47 },
		{ 97, 251, 47 },
		{ 97, 252, 47 },
		{ 97, 253, 47 },
		{ 97, 254, 47 },
		{ 97, 255, 47 },
		{ 97, 256, 47 },
		{ 98, 218, 47 },
		{ 98, 219, 47 },
		{ 98, 220, 47 },
		{ 98, 221, 47 },
		{ 98, 222, 47 },
		{ 98, 223, 47 },
		{ 98, 224, 47 },
		{ 98, 225, 47 },
		{ 98, 226, 47 },
		{ 98, 227, 47 },
		{ 98, 228, 47 },
		{ 98, 229, 47 },
		{ 98, 230, 47 },
		{ 98, 231, 47 },
		{ 98, 232, 47 },
		{ 98, 233, 47 },
		{ 98, 234, 47 },
		{ 98, 235, 47 },
		{ 98, 236, 47 },
		{ 98, 237, 47 },
		{ 98, 238, 47 },
		{ 98, 239, 47 },
		{ 98, 240, 47 },
		{ 98, 241, 47 },
		{ 98, 242, 47 },
		{ 98, 243, 47 },
		{ 98, 244, 47 },
		{ 98, 245, 47 },
		{ 98, 246, 47 },
		{ 98, 247, 47 },
		{ 98, 248, 47 },
		{ 98, 249, 47 },
		{ 98, 250, 47 },
		{ 98, 251, 47 },
		{ 98, 252, 47 },
		{ 98, 253, 47 },
		{ 98, 254, 47 },
		{ 98, 255, 47 },
		{ 98, 256, 47 },
		{ 99, 218, 47 },
		{ 99, 219, 47 },
		{ 99, 220, 47 },
		{ 99, 221, 47 },
		{ 99, 222, 47 },
		{ 99, 223, 47 },
		{ 99, 224, 47 },
		{ 99, 225, 47 },
		{ 99, 226, 47 },
		{ 99, 227, 47 },
		{ 99, 228, 47 },
		{ 99, 229, 47 },
		{ 99, 230, 47 },
		{ 99, 231, 47 },
		{ 99, 232, 47 },
		{ 99, 233, 47 },
		{ 99, 234, 47 },
		{ 99, 235, 47 },
		{ 99, 236, 47 },
		{ 99, 237, 47 },
		{ 99, 238, 47 },
		{ 99, 239, 47 },
		{ 99, 240, 47 },
		{ 99, 241, 47 },
		{ 99, 242, 47 },
		{ 99, 243, 47 },
		{ 99, 244, 47 },
		{ 99, 245, 47 },
		{ 99, 246, 47 },
		{ 99, 247, 47 },
		{ 99, 248, 47 },
		{ 99, 249, 47 },
		{ 99, 250, 47 },
		{ 99, 251, 47 },
		{ 99, 252, 47 },
		{ 99, 253, 47 },
		{ 99, 254, 47 },
		{ 99, 255, 47 },
		{ 99, 256, 47 },
		{ 100, 218, 47 },
		{ 100, 219, 47 },
		{ 100, 220, 47 },
		{ 100, 221, 47 },
		{ 100, 222, 47 },
		{ 100, 223, 47 },
		{ 100, 224, 47 },
		{ 100, 225, 47 },
		{ 100, 226, 47 },
		{ 100, 227, 47 },
		{ 100, 228, 47 },
		{ 100, 229, 47 },
		{ 100, 230, 47 },
		{ 100, 231, 47 },
		{ 100, 232, 47 },
		{ 100, 233, 47 },
		{ 100, 234, 47 },
		{ 100, 235, 47 },
		{ 100, 236, 47 },
		{ 100, 237, 47 },
		{ 100, 238, 47 },
		{ 100, 239, 47 },
		{ 100, 240, 47 },
		{ 100, 241, 47 },
		{ 100, 242, 47 },
		{ 100, 243, 47 },
		{ 100, 244, 47 },
		{ 100, 245, 47 },
		{ 100, 246, 47 },
		{ 100, 247, 47 },
		{ 100, 248, 47 },
		{ 100, 249, 47 },
		{ 100, 250, 47 },
		{ 100, 251, 47 },
		{ 100, 252, 47 },
		{ 100, 253, 47 },
		{ 100, 254, 47 },
		{ 100, 255, 47 },
		{ 100, 256, 47 },
		{ 101, 219, 47 },
		{ 101, 220, 47 },
		{ 101, 221, 47 },
		{ 101, 222, 47 },
		{ 101, 223, 47 },
		{ 101, 224, 47 },
		{ 101, 225, 47 },
		{ 101, 226, 47 },
		{ 101, 227, 47 },
		{ 101, 228, 47 },
		{ 101, 229, 47 },
		{ 101, 230, 47 },
		{ 101, 231, 47 },
		{ 101, 232, 47 },
		{ 101, 233, 47 },
		{ 101, 234, 47 },
		{ 101, 235, 47 },
		{ 101, 236, 47 },
		{ 101, 237, 47 },
		{ 101, 238, 47 },
		{ 101, 239, 47 },
		{ 101, 240, 47 },
		{ 101, 241, 47 },
		{ 101, 242, 47 },
		{ 101, 243, 47 },
		{ 101, 244, 47 },
		{ 101, 245, 47 },
		{ 101, 246, 47 },
		{ 101, 247, 47 },
		{ 101, 248, 47 },
		{ 101, 249, 47 },
		{ 101, 250, 47 },
		{ 101, 251, 47 },
		{ 101, 252, 47 },
		{ 101, 253, 47 },
		{ 101, 254, 47 },
		{ 101, 255, 47 },
		{ 101, 256, 47 },
		{ 102, 219, 47 },
		{ 102, 220, 47 },
		{ 102, 221, 47 },
		{ 102, 222, 47 },
		{ 102, 223, 47 },
		{ 102, 224, 47 },
		{ 102, 225, 47 },
		{ 102, 226, 47 },
		{ 102, 227, 47 },
		{ 102, 228, 47 },
		{ 102, 229, 47 },
		{ 102, 230, 47 },
		{ 102, 231, 47 },
		{ 102, 232, 47 },
		{ 102, 233, 47 },
		{ 102, 234, 47 },
		{ 102, 235, 47 },
		{ 102, 236, 47 },
		{ 102, 237, 47 },
		{ 102, 238, 47 },
		{ 102, 239, 47 },
		{ 102, 240, 47 },
		{ 102, 241, 47 },
		{ 102, 242, 47 },
		{ 102, 243, 47 },
		{ 102, 244, 47 },
		{ 102, 245, 47 },
		{ 102, 246, 47 },
		{ 102, 247, 47 },
		{ 102, 248, 47 },
		{ 102, 249, 47 },
		{ 102, 250, 47 },
		{ 102, 251, 47 },
		{ 102, 252, 47 },
		{ 102, 253, 47 },
		{ 102, 254, 47 },
		{ 102, 255, 47 },
		{ 102, 256, 47 },
		{ 103, 219, 47 },
		{ 103, 220, 47 },
		{ 103, 221, 47 },
		{ 103, 222, 47 },
		{ 103, 223, 47 },
		{ 103, 224, 47 },
		{ 103, 225, 47 },
		{ 103, 226, 47 },
		{ 103, 227, 47 },
		{ 103, 228, 47 },
		{ 103, 229, 47 },
		{ 103, 230, 47 },
		{ 103, 231, 47 },
		{ 103, 232, 47 },
		{ 103, 233, 47 },
		{ 103, 234, 47 },
		{ 103, 235, 47 },
		{ 103, 236, 47 },
		{ 103, 237, 47 },
		{ 103, 238, 47 },
		{ 103, 239, 47 },
		{ 103, 240, 47 },
		{ 103, 241, 47 },
		{ 103, 242, 47 },
		{ 103, 243, 47 },
		{ 103, 244, 47 },
		{ 103, 245, 47 },
		{ 103, 246, 47 },
		{ 103, 247, 47 },
		{ 103, 248, 47 },
		{ 103, 249, 47 },
		{ 103, 250, 47 },
		{ 103, 251, 47 },
		{ 103, 252, 47 },
		{ 103, 253, 47 },
		{ 103, 254, 47 },
		{ 103, 255, 47 },
		{ 103, 256, 47 },
		{ 104, 219, 47 },
		{ 104, 220, 47 },
		{ 104, 221, 47 },
		{ 104, 222, 47 },
		{ 104, 223, 47 },
		{ 104, 224, 47 },
		{ 104, 225, 47 },
		{ 104, 226, 47 },
		{ 104, 227, 47 },
		{ 104, 228, 47 },
		{ 104, 229, 47 },
		{ 104, 230, 47 },
		{ 104, 231, 47 },
		{ 104, 232, 47 },
		{ 104, 233, 47 },
		{ 104, 234, 47 },
		{ 104, 235, 47 },
		{ 104, 236, 47 },
		{ 104, 237, 47 },
		{ 104, 238, 47 },
		{ 104, 239, 47 },
		{ 104, 240, 47 },
		{ 104, 241, 47 },
		{ 104, 242, 47 },
		{ 104, 243, 47 },
		{ 104, 244, 47 },
		{ 104, 245, 47 },
		{ 104, 246, 47 },
		{ 104, 247, 47 },
		{ 104, 248, 47 },
		{ 104, 249, 47 },
		{ 104, 250, 47 },
		{ 104, 251, 47 },
		{ 104, 252, 47 },
		{ 104, 253, 47 },
		{ 104, 254, 47 },
		{ 104, 255, 47 },
		{ 104, 256, 47 },
		{ 105, 220, 47 },
		{ 105, 221, 47 },
		{ 105, 222, 47 },
		{ 105, 223, 47 },
		{ 105, 224, 47 },
		{ 105, 225, 47 },
		{ 105, 226, 47 },
		{ 105, 227, 47 },
		{ 105, 228, 47 },
		{ 105, 229, 47 },
		{ 105, 230, 47 },
		{ 105, 231, 47 },
		{ 105, 232, 47 },
		{ 105, 233, 47 },
		{ 105, 234, 47 },
		{ 105, 235, 47 },
		{ 105, 236, 47 },
		{ 105, 237, 47 },
		{ 105, 238, 47 },
		{ 105, 239, 47 },
		{ 105, 240, 47 },
		{ 105, 241, 47 },
		{ 105, 242, 47 },
		{ 105, 243, 47 },
		{ 105, 244, 47 },
		{ 105, 245, 47 },
		{ 105, 246, 47 },
		{ 105, 247, 47 },
		{ 105, 248, 47 },
		{ 105, 249, 47 },
		{ 105, 250, 47 },
		{ 105, 251, 47 },
		{ 105, 252, 47 },
		{ 105, 253, 47 },
		{ 105, 254, 47 },
		{ 105, 255, 47 },
		{ 105, 256, 47 },
		{ 106, 220, 47 },
		{ 106, 221, 47 },
		{ 106, 222, 47 },
		{ 106, 223, 47 },
		{ 106, 224, 47 },
		{ 106, 225, 47 },
		{ 106, 226, 47 },
		{ 106, 227, 47 },
		{ 106, 228, 47 },
		{ 106, 229, 47 },
		{ 106, 230, 47 },
		{ 106, 231, 47 },
		{ 106, 232, 47 },
		{ 106, 233, 47 },
		{ 106, 234, 47 },
		{ 106, 235, 47 },
		{ 106, 236, 47 },
		{ 106, 237, 47 },
		{ 106, 238, 47 },
		{ 106, 239, 47 },
		{ 106, 240, 47 },
		{ 106, 241, 47 },
		{ 106, 242, 47 },
		{ 106, 243, 47 },
		{ 106, 244, 47 },
		{ 106, 245, 47 },
		{ 106, 246, 47 },
		{ 106, 247, 47 },
		{ 106, 248, 47 },
		{ 106, 249, 47 },
		{ 106, 250, 47 },
		{ 106, 251, 47 },
		{ 106, 252, 47 },
		{ 106, 253, 47 },
		{ 106, 254, 47 },
		{ 106, 255, 47 },
		{ 106, 256, 47 },
		{ 107, 220, 47 },
		{ 107, 221, 47 },
		{ 107, 222, 47 },
		{ 107, 223, 47 },
		{ 107, 224, 47 },
		{ 107, 225, 47 },
		{ 107, 226, 47 },
		{ 107, 227, 47 },
		{ 107, 228, 47 },
		{ 107, 229, 47 },
		{ 107, 230, 47 },
		{ 107, 231, 47 },
		{ 107, 232, 47 },
		{ 107, 233, 47 },
		{ 107, 234, 47 },
		{ 107, 235, 47 },
		{ 107, 236, 47 },
		{ 107, 237, 47 },
		{ 107, 238, 47 },
		{ 107, 239, 47 },
		{ 107, 240, 47 },
		{ 107, 241, 47 },
		{ 107, 242, 47 },
		{ 107, 243, 47 },
		{ 107, 244, 47 },
		{ 107, 245, 47 },
		{ 107, 246, 47 },
		{ 107, 247, 47 },
		{ 107, 248, 47 },
		{ 107, 249, 47 },
		{ 107, 250, 47 },
		{ 107, 251, 47 },
		{ 107, 252, 47 },
		{ 107, 253, 47 },
		{ 107, 254, 47 },
		{ 107, 255, 47 },
		{ 107, 256, 47 },
		{ 108, 220, 47 },
		{ 108, 221, 47 },
		{ 108, 222, 47 },
		{ 108, 223, 47 },
		{ 108, 224, 47 },
		{ 108, 225, 47 },
		{ 108, 226, 47 },
		{ 108, 227, 47 },
		{ 108, 228, 47 },
		{ 108, 229, 47 },
		{ 108, 230, 47 },
		{ 108, 231, 47 },
		{ 108, 232, 47 },
		{ 108, 233, 47 },
		{ 108, 234, 47 },
		{ 108, 235, 47 },
		{ 108, 236, 47 },
		{ 108, 237, 47 },
		{ 108, 238, 47 },
		{ 108, 239, 47 },
		{ 108, 240, 47 },
		{ 108, 241, 47 },
		{ 108, 242, 47 },
		{ 108, 243, 47 },
		{ 108, 244, 47 },
		{ 108, 245, 47 },
		{ 108, 246, 47 },
		{ 108, 247, 47 },
		{ 108, 248, 47 },
		{ 108, 249, 47 },
		{ 108, 250, 47 },
		{ 108, 251, 47 },
		{ 108, 252, 47 },
		{ 108, 253, 47 },
		{ 108, 254, 47 },
		{ 108, 255, 47 },
		{ 108, 256, 47 },
		{ 109, 221, 47 },
		{ 109, 222, 47 },
		{ 109, 223, 47 },
		{ 109, 224, 47 },
		{ 109, 225, 47 },
		{ 109, 226, 47 },
		{ 109, 227, 47 },
		{ 109, 228, 47 },
		{ 109, 229, 47 },
		{ 109, 230, 47 },
		{ 109, 231, 47 },
		{ 109, 232, 47 },
		{ 109, 233, 47 },
		{ 109, 234, 47 },
		{ 109, 235, 47 },
		{ 109, 236, 47 },
		{ 109, 237, 47 },
		{ 109, 238, 47 },
		{ 109, 239, 47 },
		{ 109, 240, 47 },
		{ 109, 241, 47 },
		{ 109, 242, 47 },
		{ 109, 243, 47 },
		{ 109, 244, 47 },
		{ 109, 245, 47 },
		{ 109, 246, 47 },
		{ 109, 247, 47 },
		{ 109, 248, 47 },
		{ 109, 249, 47 },
		{ 109, 250, 47 },
		{ 109, 251, 47 },
		{ 109, 252, 47 },
		{ 109, 253, 47 },
		{ 109, 254, 47 },
		{ 109, 255, 47 },
		{ 109, 256, 47 },
		{ 110, 221, 47 },
		{ 110, 222, 47 },
		{ 110, 223, 47 },
		{ 110, 224, 47 },
		{ 110, 225, 47 },
		{ 110, 226, 47 },
		{ 110, 227, 47 },
		{ 110, 228, 47 },
		{ 110, 229, 47 },
		{ 110, 230, 47 },
		{ 110, 231, 47 },
		{ 110, 232, 47 },
		{ 110, 233, 47 },
		{ 110, 234, 47 },
		{ 110, 235, 47 },
		{ 110, 236, 47 },
		{ 110, 237, 47 },
		{ 110, 238, 47 },
		{ 110, 239, 47 },
		{ 110, 240, 47 },
		{ 110, 241, 47 },
		{ 110, 242, 47 },
		{ 110, 243, 47 },
		{ 110, 244, 47 },
		{ 110, 245, 47 },
		{ 110, 246, 47 },
		{ 110, 247, 47 },
		{ 110, 248, 47 },
		{ 110, 249, 47 },
		{ 110, 250, 47 },
		{ 110, 251, 47 },
		{ 110, 252, 47 },
		{ 110, 253, 47 },
		{ 110, 254, 47 },
		{ 110, 255, 47 },
		{ 110, 256, 47 },
		{ 111, 221, 47 },
		{ 111, 222, 47 },
		{ 111, 223, 47 },
		{ 111, 224, 47 },
		{ 111, 225, 47 },
		{ 111, 226, 47 },
		{ 111, 227, 47 },
		{ 111, 228, 47 },
		{ 111, 229, 47 },
		{ 111, 230, 47 },
		{ 111, 231, 47 },
		{ 111, 232, 47 },
		{ 111, 233, 47 },
		{ 111, 234, 47 },
		{ 111, 235, 47 },
		{ 111, 236, 47 },
		{ 111, 237, 47 },
		{ 111, 238, 47 },
		{ 111, 239, 47 },
		{ 111, 240, 47 },
		{ 111, 241, 47 },
		{ 111, 242, 47 },
		{ 111, 243, 47 },
		{ 111, 244, 47 },
		{ 111, 245, 47 },
		{ 111, 246, 47 },
		{ 111, 247, 47 },
		{ 111, 248, 47 },
		{ 111, 249, 47 },
		{ 111, 250, 47 },
		{ 111, 251, 47 },
		{ 111, 252, 47 },
		{ 111, 253, 47 },
		{ 111, 254, 47 },
		{ 111, 255, 47 },
		{ 111, 256, 47 },
		{ 112, 221, 47 },
		{ 112, 222, 47 },
		{ 112, 223, 47 },
		{ 112, 224, 47 },
		{ 112, 225, 47 },
		{ 112, 226, 47 },
		{ 112, 227, 47 },
		{ 112, 228, 47 },
		{ 112, 229, 47 },
		{ 112, 230, 47 },
		{ 112, 231, 47 },
		{ 112, 232, 47 },
		{ 112, 233, 47 },
		{ 112, 234, 47 },
		{ 112, 235, 47 },
		{ 112, 236, 47 },
		{ 112, 237, 47 },
		{ 112, 238, 47 },
		{ 112, 239, 47 },
		{ 112, 240, 47 },
		{ 112, 241, 47 },
		{ 112, 242, 47 },
		{ 112, 243, 47 },
		{ 112, 244, 47 },
		{ 112, 245, 47 },
		{ 112, 246, 47 },
		{ 112, 247, 47 },
		{ 112, 248, 47 },
		{ 112, 249, 47 },
		{ 112, 250, 47 },
		{ 112, 251, 47 },
		{ 112, 252, 47 },
		{ 112, 253, 47 },
		{ 112, 254, 47 },
		{ 112, 255, 47 },
		{ 112, 256, 47 },
		{ 113, 222, 47 },
		{ 113, 223, 47 },
		{ 113, 224, 47 },
		{ 113, 225, 47 },
		{ 113, 226, 47 },
		{ 113, 227, 47 },
		{ 113, 228, 47 },
		{ 113, 229, 47 },
		{ 113, 230, 47 },
		{ 113, 231, 47 },
		{ 113, 232, 47 },
		{ 113, 233, 47 },
		{ 113, 234, 47 },
		{ 113, 235, 47 },
		{ 113, 236, 47 },
		{ 113, 237, 47 },
		{ 113, 238, 47 },
		{ 113, 239, 47 },
		{ 113, 240, 47 },
		{ 113, 241, 47 },
		{ 113, 242, 47 },
		{ 113, 243, 47 },
		{ 113, 244, 47 },
		{ 113, 245, 47 },
		{ 113, 246, 47 },
		{ 113, 247, 47 },
		{ 113, 248, 47 },
		{ 113, 249, 47 },
		{ 113, 250, 47 },
		{ 113, 251, 47 },
		{ 113, 252, 47 },
		{ 113, 253, 47 },
		{ 113, 254, 47 },
		{ 113, 255, 47 },
		{ 113, 256, 47 },
		{ 114, 222, 47 },
		{ 114, 223, 47 },
		{ 114, 224, 47 },
		{ 114, 225, 47 },
		{ 114, 226, 47 },
		{ 114, 227, 47 },
		{ 114, 228, 47 },
		{ 114, 229, 47 },
		{ 114, 230, 47 },
		{ 114, 231, 47 },
		{ 114, 232, 47 },
		{ 114, 233, 47 },
		{ 114, 234, 47 },
		{ 114, 235, 47 },
		{ 114, 236, 47 },
		{ 114, 237, 47 },
		{ 114, 238, 47 },
		{ 114, 239, 47 },
		{ 114, 240, 47 },
		{ 114, 241, 47 },
		{ 114, 242, 47 },
		{ 114, 243, 47 },
		{ 114, 244, 47 },
		{ 114, 245, 47 },
		{ 114, 246, 47 },
		{ 114, 247, 47 },
		{ 114, 248, 47 },
		{ 114, 249, 47 },
		{ 114, 250, 47 },
		{ 114, 251, 47 },
		{ 114, 252, 47 },
		{ 114, 253, 47 },
		{ 114, 254, 47 },
		{ 114, 255, 47 },
		{ 114, 256, 47 },
		{ 115, 222, 47 },
		{ 115, 223, 47 },
		{ 115, 224, 47 },
		{ 115, 225, 47 },
		{ 115, 226, 47 },
		{ 115, 227, 47 },
		{ 115, 228, 47 },
		{ 115, 229, 47 },
		{ 115, 230, 47 },
		{ 115, 231, 47 },
		{ 115, 232, 47 },
		{ 115, 233, 47 },
		{ 115, 234, 47 },
		{ 115, 235, 47 },
		{ 115, 236, 47 },
		{ 115, 237, 47 },
		{ 115, 238, 47 },
		{ 115, 239, 47 },
		{ 115, 240, 47 },
		{ 115, 241, 47 },
		{ 115, 242, 47 },
		{ 115, 243, 47 },
		{ 115, 244, 47 },
		{ 115, 245, 47 },
		{ 115, 246, 47 },
		{ 115, 247, 47 },
		{ 115, 248, 47 },
		{ 115, 249, 47 },
		{ 115, 250, 47 },
		{ 115, 251, 47 },
		{ 115, 252, 47 },
		{ 115, 253, 47 },
		{ 115, 254, 47 },
		{ 115, 255, 47 },
		{ 115, 256, 47 },
		{ 116, 222, 47 },
		{ 116, 223, 47 },
		{ 116, 224, 47 },
		{ 116, 225, 47 },
		{ 116, 226, 47 },
		{ 116, 227, 47 },
		{ 116, 228, 47 },
		{ 116, 229, 47 },
		{ 116, 230, 47 },
		{ 116, 231, 47 },
		{ 116, 232, 47 },
		{ 116, 233, 47 },
		{ 116, 234, 47 },
		{ 116, 235, 47 },
		{ 116, 236, 47 },
		{ 116, 237, 47 },
		{ 116, 238, 47 },
		{ 116, 239, 47 },
		{ 116, 240, 47 },
		{ 116, 241, 47 },
		{ 116, 242, 47 },
		{ 116, 243, 47 },
		{ 116, 244, 47 },
		{ 116, 245, 47 },
		{ 116, 246, 47 },
		{ 116, 247, 47 },
		{ 116, 248, 47 },
		{ 116, 249, 47 },
		{ 116, 250, 47 },
		{ 116, 251, 47 },
		{ 116, 252, 47 },
		{ 116, 253, 47 },
		{ 116, 254, 47 },
		{ 116, 255, 47 },
		{ 116, 256, 47 },
		{ 117, 223, 47 },
		{ 117, 224, 47 },
		{ 117, 225, 47 },
		{ 117, 226, 47 },
		{ 117, 227, 47 },
		{ 117, 228, 47 },
		{ 117, 229, 47 },
		{ 117, 230, 47 },
		{ 117, 231, 47 },
		{ 117, 232, 47 },
		{ 117, 233, 47 },
		{ 117, 234, 47 },
		{ 117, 235, 47 },
		{ 117, 236, 47 },
		{ 117, 237, 47 },
		{ 117, 238, 47 },
		{ 117, 239, 47 },
		{ 117, 240, 47 },
		{ 117, 241, 47 },
		{ 117, 242, 47 },
		{ 117, 243, 47 },
		{ 117, 244, 47 },
		{ 117, 245, 47 },
		{ 117, 246, 47 },
		{ 117, 247, 47 },
		{ 117, 248, 47 },
		{ 117, 249, 47 },
		{ 117, 250, 47 },
		{ 117, 251, 47 },
		{ 117, 252, 47 },
		{ 117, 253, 47 },
		{ 117, 254, 47 },
		{ 117, 255, 47 },
		{ 117, 256, 47 },
		{ 118, 223, 47 },
		{ 118, 224, 47 },
		{ 118, 225, 47 },
		{ 118, 226, 47 },
		{ 118, 227, 47 },
		{ 118, 228, 47 },
		{ 118, 229, 47 },
		{ 118, 230, 47 },
		{ 118, 231, 47 },
		{ 118, 232, 47 },
		{ 118, 233, 47 },
		{ 118, 234, 47 },
		{ 118, 235, 47 },
		{ 118, 236, 47 },
		{ 118, 237, 47 },
		{ 118, 238, 47 },
		{ 118, 239, 47 },
		{ 118, 240, 47 },
		{ 118, 241, 47 },
		{ 118, 242, 47 },
		{ 118, 243, 47 },
		{ 118, 244, 47 },
		{ 118, 245, 47 },
		{ 118, 246, 47 },
		{ 118, 247, 47 },
		{ 118, 248, 47 },
		{ 118, 249, 47 },
		{ 118, 250, 47 },
		{ 118, 251, 47 },
		{ 118, 252, 47 },
		{ 118, 253, 47 },
		{ 118, 254, 47 },
		{ 118, 255, 47 },
		{ 118, 256, 47 },
		{ 119, 223, 47 },
		{ 119, 224, 47 },
		{ 119, 225, 47 },
		{ 119, 226, 47 },
		{ 119, 227, 47 },
		{ 119, 228, 47 },
		{ 119, 229, 47 },
		{ 119, 230, 47 },
		{ 119, 231, 47 },
		{ 119, 232, 47 },
		{ 119, 233, 47 },
		{ 119, 234, 47 },
		{ 119, 235, 47 },
		{ 119, 236, 47 },
		{ 119, 237, 47 },
		{ 119, 238, 47 },
		{ 119, 239, 47 },
		{ 119, 240, 47 },
		{ 119, 241, 47 },
		{ 119, 242, 47 },
		{ 119, 243, 47 },
		{ 119, 244, 47 },
		{ 119, 245, 47 },
		{ 119, 246, 47 },
		{ 119, 247, 47 },
		{ 119, 248, 47 },
		{ 119, 249, 47 },
		{ 119, 250, 47 },
		{ 119, 251, 47 },
		{ 119, 252, 47 },
		{ 119, 253, 47 },
		{ 119, 254, 47 },
		{ 119, 255, 47 },
		{ 119, 256, 47 },
		{ 120, 223, 47 },
		{ 120, 224, 47 },
		{ 120, 225, 47 },
		{ 120, 226, 47 },
		{ 120, 227, 47 },
		{ 120, 228, 47 },
		{ 120, 229, 47 },
		{ 120, 230, 47 },
		{ 120, 231, 47 },
		{ 120, 232, 47 },
		{ 120, 233, 47 },
		{ 120, 234, 47 },
		{ 120, 235, 47 },
		{ 120, 236, 47 },
		{ 120, 237, 47 },
		{ 120, 238, 47 },
		{ 120, 239, 47 },
		{ 120, 240, 47 },
		{ 120, 241, 47 },
		{ 120, 242, 47 },
		{ 120, 243, 47 },
		{ 120, 244, 47 },
		{ 120, 245, 47 },
		{ 120, 246, 47 },
		{ 120, 247, 47 },
		{ 120, 248, 47 },
		{ 120, 249, 47 },
		{ 120, 250, 47 },
		{ 120, 251, 47 },
		{ 120, 252, 47 },
		{ 120, 253, 47 },
		{ 120, 254, 47 },
		{ 120, 255, 47 },
		{ 120, 256, 47 },
		{ 121, 224, 47 },
		{ 121, 225, 47 },
		{ 121, 226, 47 },
		{ 121, 227, 47 },
		{ 121, 228, 47 },
		{ 121, 229, 47 },
		{ 121, 230, 47 },
		{ 121, 231, 47 },
		{ 121, 232, 47 },
		{ 121, 233, 47 },
		{ 121, 234, 47 },
		{ 121, 235, 47 },
		{ 121, 236, 47 },
		{ 121, 237, 47 },
		{ 121, 238, 47 },
		{ 121, 239, 47 },
		{ 121, 240, 47 },
		{ 121, 241, 47 },
		{ 121, 242, 47 },
		{ 121, 243, 47 },
		{ 121, 244, 47 },
		{ 121, 245, 47 },
		{ 121, 246, 47 },
		{ 121, 247, 47 },
		{ 121, 248, 47 },
		{ 121, 249, 47 },
		{ 121, 250, 47 },
		{ 121, 251, 47 },
		{ 121, 252, 47 },
		{ 121, 253, 47 },
		{ 121, 254, 47 },
		{ 121, 255, 47 },
		{ 121, 256, 47 },
		{ 122, 224, 47 },
		{ 122, 225, 47 },
		{ 122, 226, 47 },
		{ 122, 227, 47 },
		{ 122, 228, 47 },
		{ 122, 229, 47 },
		{ 122, 230, 47 },
		{ 122, 231, 47 },
		{ 122, 232, 47 },
		{ 122, 233, 47 },
		{ 122, 234, 47 },
		{ 122, 235, 47 },
		{ 122, 236, 47 },
		{ 122, 237, 47 },
		{ 122, 238, 47 },
		{ 122, 239, 47 },
		{ 122, 240, 47 },
		{ 122, 241, 47 },
		{ 122, 242, 47 },
		{ 122, 243, 47 },
		{ 122, 244, 47 },
		{ 122, 245, 47 },
		{ 122, 246, 47 },
		{ 122, 247, 47 },
		{ 122, 248, 47 },
		{ 122, 249, 47 },
		{ 122, 250, 47 },
		{ 122, 251, 47 },
		{ 122, 252, 47 },
		{ 122, 253, 47 },
		{ 122, 254, 47 },
		{ 122, 255, 47 },
		{ 122, 256, 47 },
		{ 123, 224, 47 },
		{ 123, 225, 47 },
		{ 123, 226, 47 },
		{ 123, 227, 47 },
		{ 123, 228, 47 },
		{ 123, 229, 47 },
		{ 123, 230, 47 },
		{ 123, 231, 47 },
		{ 123, 232, 47 },
		{ 123, 233, 47 },
		{ 123, 234, 47 },
		{ 123, 235, 47 },
		{ 123, 236, 47 },
		{ 123, 237, 47 },
		{ 123, 238, 47 },
		{ 123, 239, 47 },
		{ 123, 240, 47 },
		{ 123, 241, 47 },
		{ 123, 242, 47 },
		{ 123, 243, 47 },
		{ 123, 244, 47 },
		{ 123, 245, 47 },
		{ 123, 246, 47 },
		{ 123, 247, 47 },
		{ 123, 248, 47 },
		{ 123, 249, 47 },
		{ 123, 250, 47 },
		{ 123, 251, 47 },
		{ 123, 252, 47 },
		{ 123, 253, 47 },
		{ 123, 254, 47 },
		{ 123, 255, 47 },
		{ 123, 256, 47 },
		{ 124, 224, 47 },
		{ 124, 225, 47 },
		{ 124, 226, 47 },
		{ 124, 227, 47 },
		{ 124, 228, 47 },
		{ 124, 229, 47 },
		{ 124, 230, 47 },
		{ 124, 231, 47 },
		{ 124, 232, 47 },
		{ 124, 233, 47 },
		{ 124, 234, 47 },
		{ 124, 235, 47 },
		{ 124, 236, 47 },
		{ 124, 237, 47 },
		{ 124, 238, 47 },
		{ 124, 239, 47 },
		{ 124, 240, 47 },
		{ 124, 241, 47 },
		{ 124, 242, 47 },
		{ 124, 243, 47 },
		{ 124, 244, 47 },
		{ 124, 245, 47 },
		{ 124, 246, 47 },
		{ 124, 247, 47 },
		{ 124, 248, 47 },
		{ 124, 249, 47 },
		{ 124, 250, 47 },
		{ 124, 251, 47 },
		{ 124, 252, 47 },
		{ 124, 253, 47 },
		{ 124, 254, 47 },
		{ 124, 255, 47 },
		{ 124, 256, 47 },
		{ 125, 225, 47 },
		{ 125, 226, 47 },
		{ 125, 227, 47 },
		{ 125, 228, 47 },
		{ 125, 229, 47 },
		{ 125, 230, 47 },
		{ 125, 231, 47 },
		{ 125, 232, 47 },
		{ 125, 233, 47 },
		{ 125, 234, 47 },
		{ 125, 235, 47 },
		{ 125, 236, 47 },
		{ 125, 237, 47 },
		{ 125, 238, 47 },
		{ 125, 239, 47 },
		{ 125, 240, 47 },
		{ 125, 241, 47 },
		{ 125, 242, 47 },
		{ 125, 243, 47 },
		{ 125, 244, 47 },
		{ 125, 245, 47 },
		{ 125, 246, 47 },
		{ 125, 247, 47 },
		{ 125, 248, 47 },
		{ 125, 249, 47 },
		{ 125, 250, 47 },
		{ 125, 251, 47 },
		{ 125, 252, 47 },
		{ 125, 253, 47 },
		{ 125, 254, 47 },
		{ 125, 255, 47 },
		{ 125, 256, 47 },
		{ 126, 225, 47 },
		{ 126, 226, 47 },
		{ 126, 227, 47 },
		{ 126, 228, 47 },
		{ 126, 229, 47 },
		{ 126, 230, 47 },
		{ 126, 231, 47 },
		{ 126, 232, 47 },
		{ 126, 233, 47 },
		{ 126, 234, 47 },
		{ 126, 235, 47 },
		{ 126, 236, 47 },
		{ 126, 237, 47 },
		{ 126, 238, 47 },
		{ 126, 239, 47 },
		{ 126, 240, 47 },
		{ 126, 241, 47 },
		{ 126, 242, 47 },
		{ 126, 243, 47 },
		{ 126, 244, 47 },
		{ 126, 245, 47 },
		{ 126, 246, 47 },
		{ 126, 247, 47 },
		{ 126, 248, 47 },
		{ 126, 249, 47 },
		{ 126, 250, 47 },
		{ 126, 251, 47 },
		{ 126, 252, 47 },
		{ 126, 253, 47 },
		{ 126, 254, 47 },
		{ 126, 255, 47 },
		{ 126, 256, 47 },
		{ 127, 225, 47 },
		{ 127, 226, 47 },
		{ 127, 227, 47 },
		{ 127, 228, 47 },
		{ 127, 229, 47 },
		{ 127, 230, 47 },
		{ 127, 231, 47 },
		{ 127, 232, 47 },
		{ 127, 233, 47 },
		{ 127, 234, 47 },
		{ 127, 235, 47 },
		{ 127, 236, 47 },
		{ 127, 237, 47 },
		{ 127, 238, 47 },
		{ 127, 239, 47 },
		{ 127, 240, 47 },
		{ 127, 241, 47 },
		{ 127, 242, 47 },
		{ 127, 243, 47 },
		{ 127, 244, 47 },
		{ 127, 245, 47 },
		{ 127, 246, 47 },
		{ 127, 247, 47 },
		{ 127, 248, 47 },
		{ 127, 249, 47 },
		{ 127, 250, 47 },
		{ 127, 251, 47 },
		{ 127, 252, 47 },
		{ 127, 253, 47 },
		{ 127, 254, 47 },
		{ 127, 255, 47 },
		{ 127, 256, 47 },
		{ 128, 225, 47 },
		{ 128, 226, 47 },
		{ 128, 227, 47 },
		{ 128, 228, 47 },
		{ 128, 229, 47 },
		{ 128, 230, 47 },
		{ 128, 231, 47 },
		{ 128, 232, 47 },
		{ 128, 233, 47 },
		{ 128, 234, 47 },
		{ 128, 235, 47 },
		{ 128, 236, 47 },
		{ 128, 237, 47 },
		{ 128, 238, 47 },
		{ 128, 239, 47 },
		{ 128, 240, 47 },
		{ 128, 241, 47 },
		{ 128, 242, 47 },
		{ 128, 243, 47 },
		{ 128, 244, 47 },
		{ 128, 245, 47 },
		{ 128, 246, 47 },
		{ 128, 247, 47 },
		{ 128, 248, 47 },
		{ 128, 249, 47 },
		{ 128, 250, 47 },
		{ 128, 251, 47 },
		{ 128, 252, 47 },
		{ 128, 253, 47 },
		{ 128, 254, 47 },
		{ 128, 255, 47 },
		{ 128, 256, 47 },
		{ 129, 226, 47 },
		{ 129, 227, 47 },
		{ 129, 228, 47 },
		{ 129, 229, 47 },
		{ 129, 230, 47 },
		{ 129, 231, 47 },
		{ 129, 232, 47 },
		{ 129, 233, 47 },
		{ 129, 234, 47 },
		{ 129, 235, 47 },
		{ 129, 236, 47 },
		{ 129, 237, 47 },
		{ 129, 238, 47 },
		{ 129, 239, 47 },
		{ 129, 240, 47 },
		{ 129, 241, 47 },
		{ 129, 242, 47 },
		{ 129, 243, 47 },
		{ 129, 244, 47 },
		{ 129, 245, 47 },
		{ 129, 246, 47 },
		{ 129, 247, 47 },
		{ 129, 248, 47 },
		{ 129, 249, 47 },
		{ 129, 250, 47 },
		{ 129, 251, 47 },
		{ 129, 252, 47 },
		{ 129, 253, 47 },
		{ 129, 254, 47 },
		{ 129, 255, 47 },
		{ 129, 256, 47 },
		{ 130, 226, 47 },
		{ 130, 227, 47 },
		{ 130, 228, 47 },
		{ 130, 229, 47 },
		{ 130, 230, 47 },
		{ 130, 231, 47 },
		{ 130, 232, 47 },
		{ 130, 233, 47 },
		{ 130, 234, 47 },
		{ 130, 235, 47 },
		{ 130, 236, 47 },
		{ 130, 237, 47 },
		{ 130, 238, 47 },
		{ 130, 239, 47 },
		{ 130, 240, 47 },
		{ 130, 241, 47 },
		{ 130, 242, 47 },
		{ 130, 243, 47 },
		{ 130, 244, 47 },
		{ 130, 245, 47 },
		{ 130, 246, 47 },
		{ 130, 247, 47 },
		{ 130, 248, 47 },
		{ 130, 249, 47 },
		{ 130, 250, 47 },
		{ 130, 251, 47 },
		{ 130, 252, 47 },
		{ 130, 253, 47 },
		{ 130, 254, 47 },
		{ 130, 255, 47 },
		{ 130, 256, 47 },
		{ 131, 226, 47 },
		{ 131, 227, 47 },
		{ 131, 228, 47 },
		{ 131, 229, 47 },
		{ 131, 230, 47 },
		{ 131, 231, 47 },
		{ 131, 232, 47 },
		{ 131, 233, 47 },
		{ 131, 234, 47 },
		{ 131, 235, 47 },
		{ 131, 236, 47 },
		{ 131, 237, 47 },
		{ 131, 238, 47 },
		{ 131, 239, 47 },
		{ 131, 240, 47 },
		{ 131, 241, 47 },
		{ 131, 242, 47 },
		{ 131, 243, 47 },
		{ 131, 244, 47 },
		{ 131, 245, 47 },
		{ 131, 246, 47 },
		{ 131, 247, 47 },
		{ 131, 248, 47 },
		{ 131, 249, 47 },
		{ 131, 250, 47 },
		{ 131, 251, 47 },
		{ 131, 252, 47 },
		{ 131, 253, 47 },
		{ 131, 254, 47 },
		{ 131, 255, 47 },
		{ 131, 256, 47 },
		{ 132, 226, 47 },
		{ 132, 227, 47 },
		{ 132, 228, 47 },
		{ 132, 229, 47 },
		{ 132, 230, 47 },
		{ 132, 231, 47 },
		{ 132, 232, 47 },
		{ 132, 233, 47 },
		{ 132, 234, 47 },
		{ 132, 235, 47 },
		{ 132, 236, 47 },
		{ 132, 237, 47 },
		{ 132, 238, 47 },
		{ 132, 239, 47 },
		{ 132, 240, 47 },
		{ 132, 241, 47 },
		{ 132, 242, 47 },
		{ 132, 243, 47 },
		{ 132, 244, 47 },
		{ 132, 245, 47 },
		{ 132, 246, 47 },
		{ 132, 247, 47 },
		{ 132, 248, 47 },
		{ 132, 249, 47 },
		{ 132, 250, 47 },
		{ 132, 251, 47 },
		{ 132, 252, 47 },
		{ 132, 253, 47 },
		{ 132, 254, 47 },
		{ 132, 255, 47 },
		{ 132, 256, 47 },
		{ 133, 227, 47 },
		{ 133, 228, 47 },
		{ 133, 229, 47 },
		{ 133, 230, 47 },
		{ 133, 231, 47 },
		{ 133, 232, 47 },
		{ 133, 233, 47 },
		{ 133, 234, 47 },
		{ 133, 235, 47 },
		{ 133, 236, 47 },
		{ 133, 237, 47 },
		{ 133, 238, 47 },
		{ 133, 239, 47 },
		{ 133, 240, 47 },
		{ 133, 241, 47 },
		{ 133, 242, 47 },
		{ 133, 243, 47 },
		{ 133, 244, 47 },
		{ 133, 245, 47 },
		{ 133, 246, 47 },
		{ 133, 247, 47 },
		{ 133, 248, 47 },
		{ 133, 249, 47 },
		{ 133, 250, 47 },
		{ 133, 251, 47 },
		{ 133, 252, 47 },
		{ 133, 253, 47 },
		{ 133, 254, 47 },
		{ 133, 255, 47 },
		{ 133, 256, 47 },
		{ 134, 227, 47 },
		{ 134, 228, 47 },
		{ 134, 229, 47 },
		{ 134, 230, 47 },
		{ 134, 231, 47 },
		{ 134, 232, 47 },
		{ 134, 233, 47 },
		{ 134, 234, 47 },
		{ 134, 235, 47 },
		{ 134, 236, 47 },
		{ 134, 237, 47 },
		{ 134, 238, 47 },
		{ 134, 239, 47 },
		{ 134, 240, 47 },
		{ 134, 241, 47 },
		{ 134, 242, 47 },
		{ 134, 243, 47 },
		{ 134, 244, 47 },
		{ 134, 245, 47 },
		{ 134, 246, 47 },
		{ 134, 247, 47 },
		{ 134, 248, 47 },
		{ 134, 249, 47 },
		{ 134, 250, 47 },
		{ 134, 251, 47 },
		{ 134, 252, 47 },
		{ 134, 253, 47 },
		{ 134, 254, 47 },
		{ 134, 255, 47 },
		{ 134, 256, 47 },
		{ 135, 227, 47 },
		{ 135, 228, 47 },
		{ 135, 229, 47 },
		{ 135, 230, 47 },
		{ 135, 231, 47 },
		{ 135, 232, 47 },
		{ 135, 233, 47 },
		{ 135, 234, 47 },
		{ 135, 235, 47 },
		{ 135, 236, 47 },
		{ 135, 237, 47 },
		{ 135, 238, 47 },
		{ 135, 239, 47 },
		{ 135, 240, 47 },
		{ 135, 241, 47 },
		{ 135, 242, 47 },
		{ 135, 243, 47 },
		{ 135, 244, 47 },
		{ 135, 245, 47 },
		{ 135, 246, 47 },
		{ 135, 247, 47 },
		{ 135, 248, 47 },
		{ 135, 249, 47 },
		{ 135, 250, 47 },
		{ 135, 251, 47 },
		{ 135, 252, 47 },
		{ 135, 253, 47 },
		{ 135, 254, 47 },
		{ 135, 255, 47 },
		{ 135, 256, 47 },
		{ 136, 227, 47 },
		{ 136, 228, 47 },
		{ 136, 229, 47 },
		{ 136, 230, 47 },
		{ 136, 231, 47 },
		{ 136, 232, 47 },
		{ 136, 233, 47 },
		{ 136, 234, 47 },
		{ 136, 235, 47 },
		{ 136, 236, 47 },
		{ 136, 237, 47 },
		{ 136, 238, 47 },
		{ 136, 239, 47 },
		{ 136, 240, 47 },
		{ 136, 241, 47 },
		{ 136, 242, 47 },
		{ 136, 243, 47 },
		{ 136, 244, 47 },
		{ 136, 245, 47 },
		{ 136, 246, 47 },
		{ 136, 247, 47 },
		{ 136, 248, 47 },
		{ 136, 249, 47 },
		{ 136, 250, 47 },
		{ 136, 251, 47 },
		{ 136, 252, 47 },
		{ 136, 253, 47 },
		{ 136, 254, 47 },
		{ 136, 255, 47 },
		{ 136, 256, 47 },
		{ 137, 228, 47 },
		{ 137, 229, 47 },
		{ 137, 230, 47 },
		{ 137, 231, 47 },
		{ 137, 232, 47 },
		{ 137, 233, 47 },
		{ 137, 234, 47 },
		{ 137, 235, 47 },
		{ 137, 236, 47 },
		{ 137, 237, 47 },
		{ 137, 238, 47 },
		{ 137, 239, 47 },
		{ 137, 240, 47 },
		{ 137, 241, 47 },
		{ 137, 242, 47 },
		{ 137, 243, 47 },
		{ 137, 244, 47 },
		{ 137, 245, 47 },
		{ 137, 246, 47 },
		{ 137, 247, 47 },
		{ 137, 248, 47 },
		{ 137, 249, 47 },
		{ 137, 250, 47 },
		{ 137, 251, 47 },
		{ 137, 252, 47 },
		{ 137, 253, 47 },
		{ 137, 254, 47 },
		{ 137, 255, 47 },
		{ 137, 256, 47 },
		{ 138, 228, 47 },
		{ 138, 229, 47 },
		{ 138, 230, 47 },
		{ 138, 231, 47 },
		{ 138, 232, 47 },
		{ 138, 233, 47 },
		{ 138, 234, 47 },
		{ 138, 235, 47 },
		{ 138, 236, 47 },
		{ 138, 237, 47 },
		{ 138, 238, 47 },
		{ 138, 239, 47 },
		{ 138, 240, 47 },
		{ 138, 241, 47 },
		{ 138, 242, 47 },
		{ 138, 243, 47 },
		{ 138, 244, 47 },
		{ 138, 245, 47 },
		{ 138, 246, 47 },
		{ 138, 247, 47 },
		{ 138, 248, 47 },
		{ 138, 249, 47 },
		{ 138, 250, 47 },
		{ 138, 251, 47 },
		{ 138, 252, 47 },
		{ 138, 253, 47 },
		{ 138, 254, 47 },
		{ 138, 255, 47 },
		{ 138, 256, 47 },
		{ 139, 228, 47 },
		{ 139, 229, 47 },
		{ 139, 230, 47 },
		{ 139, 231, 47 },
		{ 139, 232, 47 },
		{ 139, 233, 47 },
		{ 139, 234, 47 },
		{ 139, 235, 47 },
		{ 139, 236, 47 },
		{ 139, 237, 47 },
		{ 139, 238, 47 },
		{ 139, 239, 47 },
		{ 139, 240, 47 },
		{ 139, 241, 47 },
		{ 139, 242, 47 },
		{ 139, 243, 47 },
		{ 139, 244, 47 },
		{ 139, 245, 47 },
		{ 139, 246, 47 },
		{ 139, 247, 47 },
		{ 139, 248, 47 },
		{ 139, 249, 47 },
		{ 139, 250, 47 },
		{ 139, 251, 47 },
		{ 139, 252, 47 },
		{ 139, 253, 47 },
		{ 139, 254, 47 },
		{ 139, 255, 47 },
		{ 139, 256, 47 },
		{ 140, 228, 47 },
		{ 140, 229, 47 },
		{ 140, 230, 47 },
		{ 140, 231, 47 },
		{ 140, 232, 47 },
		{ 140, 233, 47 },
		{ 140, 234, 47 },
		{ 140, 235, 47 },
		{ 140, 236, 47 },
		{ 140, 237, 47 },
		{ 140, 238, 47 },
		{ 140, 239, 47 },
		{ 140, 240, 47 },
		{ 140, 241, 47 },
		{ 140, 242, 47 },
		{ 140, 243, 47 },
		{ 140, 244, 47 },
		{ 140, 245, 47 },
		{ 140, 246, 47 },
		{ 140, 247, 47 },
		{ 140, 248, 47 },
		{ 140, 249, 47 },
		{ 140, 250, 47 },
		{ 140, 251, 47 },
		{ 140, 252, 47 },
		{ 140, 253, 47 },
		{ 140, 254, 47 },
		{ 140, 255, 47 },
		{ 140, 256, 47 },
		{ 141, 229, 47 },
		{ 141, 230, 47 },
		{ 141, 231, 47 },
		{ 141, 232, 47 },
		{ 141, 233, 47 },
		{ 141, 234, 47 },
		{ 141, 235, 47 },
		{ 141, 236, 47 },
		{ 141, 237, 47 },
		{ 141, 238, 47 },
		{ 141, 239, 47 },
		{ 141, 240, 47 },
		{ 141, 241, 47 },
		{ 141, 242, 47 },
		{ 141, 243, 47 },
		{ 141, 244, 47 },
		{ 141, 245, 47 },
		{ 141, 246, 47 },
		{ 141, 247, 47 },
		{ 141, 248, 47 },
		{ 141, 249, 47 },
		{ 141, 250, 47 },
		{ 141, 251, 47 },
		{ 141, 252, 47 },
		{ 141, 253, 47 },
		{ 141, 254, 47 },
		{ 141, 255, 47 },
		{ 141, 256, 47 },
		{ 142, 229, 47 },
		{ 142, 230, 47 },
		{ 142, 231, 47 },
		{ 142, 232, 47 },
		{ 142, 233, 47 },
		{ 142, 234, 47 },
		{ 142, 235, 47 },
		{ 142, 236, 47 },
		{ 142, 237, 47 },
		{ 142, 238, 47 },
		{ 142, 239, 47 },
		{ 142, 240, 47 },
		{ 142, 241, 47 },
		{ 142, 242, 47 },
		{ 142, 243, 47 },
		{ 142, 244, 47 },
		{ 142, 245, 47 },
		{ 142, 246, 47 },
		{ 142, 247, 47 },
		{ 142, 248, 47 },
		{ 142, 249, 47 },
		{ 142, 250, 47 },
		{ 142, 251, 47 },
		{ 142, 252, 47 },
		{ 142, 253, 47 },
		{ 142, 254, 47 },
		{ 142, 255, 47 },
		{ 142, 256, 47 },
		{ 143, 229, 47 },
		{ 143, 230, 47 },
		{ 143, 231, 47 },
		{ 143, 232, 47 },
		{ 143, 233, 47 },
		{ 143, 234, 47 },
		{ 143, 235, 47 },
		{ 143, 236, 47 },
		{ 143, 237, 47 },
		{ 143, 238, 47 },
		{ 143, 239, 47 },
		{ 143, 240, 47 },
		{ 143, 241, 47 },
		{ 143, 242, 47 },
		{ 143, 243, 47 },
		{ 143, 244, 47 },
		{ 143, 245, 47 },
		{ 143, 246, 47 },
		{ 143, 247, 47 },
		{ 143, 248, 47 },
		{ 143, 249, 47 },
		{ 143, 250, 47 },
		{ 143, 251, 47 },
		{ 143, 252, 47 },
		{ 143, 253, 47 },
		{ 143, 254, 47 },
		{ 143, 255, 47 },
		{ 143, 256, 47 },
		{ 144, 229, 47 },
		{ 144, 230, 47 },
		{ 144, 231, 47 },
		{ 144, 232, 47 },
		{ 144, 233, 47 },
		{ 144, 234, 47 },
		{ 144, 235, 47 },
		{ 144, 236, 47 },
		{ 144, 237, 47 },
		{ 144, 238, 47 },
		{ 144, 239, 47 },
		{ 144, 240, 47 },
		{ 144, 241, 47 },
		{ 144, 242, 47 },
		{ 144, 243, 47 },
		{ 144, 244, 47 },
		{ 144, 245, 47 },
		{ 144, 246, 47 },
		{ 144, 247, 47 },
		{ 144, 248, 47 },
		{ 144, 249, 47 },
		{ 144, 250, 47 },
		{ 144, 251, 47 },
		{ 144, 252, 47 },
		{ 144, 253, 47 },
		{ 144, 254, 47 },
		{ 144, 255, 47 },
		{ 144, 256, 47 },
		{ 145, 230, 47 },
		{ 145, 231, 47 },
		{ 145, 232, 47 },
		{ 145, 233, 47 },
		{ 145, 234, 47 },
		{ 145, 235, 47 },
		{ 145, 236, 47 },
		{ 145, 237, 47 },
		{ 145, 238, 47 },
		{ 145, 239, 47 },
		{ 145, 240, 47 },
		{ 145, 241, 47 },
		{ 145, 242, 47 },
		{ 145, 243, 47 },
		{ 145, 244, 47 },
		{ 145, 245, 47 },
		{ 145, 246, 47 },
		{ 145, 247, 47 },
		{ 145, 248, 47 },
		{ 145, 249, 47 },
		{ 145, 250, 47 },
		{ 145, 251, 47 },
		{ 145, 252, 47 },
		{ 145, 253, 47 },
		{ 145, 254, 47 },
		{ 145, 255, 47 },
		{ 145, 256, 47 },
		{ 146, 230, 47 },
		{ 146, 231, 47 },
		{ 146, 232, 47 },
		{ 146, 233, 47 },
		{ 146, 234, 47 },
		{ 146, 235, 47 },
		{ 146, 236, 47 },
		{ 146, 237, 47 },
		{ 146, 238, 47 },
		{ 146, 239, 47 },
		{ 146, 240, 47 },
		{ 146, 241, 47 },
		{ 146, 242, 47 },
		{ 146, 243, 47 },
		{ 146, 244, 47 },
		{ 146, 245, 47 },
		{ 146, 246, 47 },
		{ 146, 247, 47 },
		{ 146, 248, 47 },
		{ 146, 249, 47 },
		{ 146, 250, 47 },
		{ 146, 251, 47 },
		{ 146, 252, 47 },
		{ 146, 253, 47 },
		{ 146, 254, 47 },
		{ 146, 255, 47 },
		{ 146, 256, 47 },
		{ 147, 230, 47 },
		{ 147, 231, 47 },
		{ 147, 232, 47 },
		{ 147, 233, 47 },
		{ 147, 234, 47 },
		{ 147, 235, 47 },
		{ 147, 236, 47 },
		{ 147, 237, 47 },
		{ 147, 238, 47 },
		{ 147, 239, 47 },
		{ 147, 240, 47 },
		{ 147, 241, 47 },
		{ 147, 242, 47 },
		{ 147, 243, 47 },
		{ 147, 244, 47 },
		{ 147, 245, 47 },
		{ 147, 246, 47 },
		{ 147, 247, 47 },
		{ 147, 248, 47 },
		{ 147, 249, 47 },
		{ 147, 250, 47 },
		{ 147, 251, 47 },
		{ 147, 252, 47 },
		{ 147, 253, 47 },
		{ 147, 254, 47 },
		{ 147, 255, 47 },
		{ 147, 256, 47 },
		{ 148, 230, 47 },
		{ 148, 231, 47 },
		{ 148, 232, 47 },
		{ 148, 233, 47 },
		{ 148, 234, 47 },
		{ 148, 235, 47 },
		{ 148, 236, 47 },
		{ 148, 237, 47 },
		{ 148, 238, 47 },
		{ 148, 239, 47 },
		{ 148, 240, 47 },
		{ 148, 241, 47 },
		{ 148, 242, 47 },
		{ 148, 243, 47 },
		{ 148, 244, 47 },
		{ 148, 245, 47 },
		{ 148, 246, 47 },
		{ 148, 247, 47 },
		{ 148, 248, 47 },
		{ 148, 249, 47 },
		{ 148, 250, 47 },
		{ 148, 251, 47 },
		{ 148, 252, 47 },
		{ 148, 253, 47 },
		{ 148, 254, 47 },
		{ 148, 255, 47 },
		{ 148, 256, 47 },
		{ 149, 231, 47 },
		{ 149, 232, 47 },
		{ 149, 233, 47 },
		{ 149, 234, 47 },
		{ 149, 235, 47 },
		{ 149, 236, 47 },
		{ 149, 237, 47 },
		{ 149, 238, 47 },
		{ 149, 239, 47 },
		{ 149, 240, 47 },
		{ 149, 241, 47 },
		{ 149, 242, 47 },
		{ 149, 243, 47 },
		{ 149, 244, 47 },
		{ 149, 245, 47 },
		{ 149, 246, 47 },
		{ 149, 247, 47 },
		{ 149, 248, 47 },
		{ 149, 249, 47 },
		{ 149, 250, 47 },
		{ 149, 251, 47 },
		{ 149, 252, 47 },
		{ 149, 253, 47 },
		{ 149, 254, 47 },
		{ 149, 255, 47 },
		{ 149, 256, 47 },
		{ 150, 231, 47 },
		{ 150, 232, 47 },
		{ 150, 233, 47 },
		{ 150, 234, 47 },
		{ 150, 235, 47 },
		{ 150, 236, 47 },
		{ 150, 237, 47 },
		{ 150, 238, 47 },
		{ 150, 239, 47 },
		{ 150, 240, 47 },
		{ 150, 241, 47 },
		{ 150, 242, 47 },
		{ 150, 243, 47 },
		{ 150, 244, 47 },
		{ 150, 245, 47 },
		{ 150, 246, 47 },
		{ 150, 247, 47 },
		{ 150, 248, 47 },
		{ 150, 249, 47 },
		{ 150, 250, 47 },
		{ 150, 251, 47 },
		{ 150, 252, 47 },
		{ 150, 253, 47 },
		{ 150, 254, 47 },
		{ 150, 255, 47 },
		{ 150, 256, 47 },
		{ 151, 231, 47 },
		{ 151, 232, 47 },
		{ 151, 233, 47 },
		{ 151, 234, 47 },
		{ 151, 235, 47 },
		{ 151, 236, 47 },
		{ 151, 237, 47 },
		{ 151, 238, 47 },
		{ 151, 239, 47 },
		{ 151, 240, 47 },
		{ 151, 241, 47 },
		{ 151, 242, 47 },
		{ 151, 243, 47 },
		{ 151, 244, 47 },
		{ 151, 245, 47 },
		{ 151, 246, 47 },
		{ 151, 247, 47 },
		{ 151, 248, 47 },
		{ 151, 249, 47 },
		{ 151, 250, 47 },
		{ 151, 251, 47 },
		{ 151, 252, 47 },
		{ 151, 253, 47 },
		{ 151, 254, 47 },
		{ 151, 255, 47 },
		{ 151, 256, 47 },
		{ 152, 231, 47 },
		{ 152, 232, 47 },
		{ 152, 233, 47 },
		{ 152, 234, 47 },
		{ 152, 235, 47 },
		{ 152, 236, 47 },
		{ 152, 237, 47 },
		{ 152, 238, 47 },
		{ 152, 239, 47 },
		{ 152, 240, 47 },
		{ 152, 241, 47 },
		{ 152, 242, 47 },
		{ 152, 243, 47 },
		{ 152, 244, 47 },
		{ 152, 245, 47 },
		{ 152, 246, 47 },
		{ 152, 247, 47 },
		{ 152, 248, 47 },
		{ 152, 249, 47 },
		{ 152, 250, 47 },
		{ 152, 251, 47 },
		{ 152, 252, 47 },
		{ 152, 253, 47 },
		{ 152, 254, 47 },
		{ 152, 255, 47 },
		{ 152, 256, 47 },
		{ 153, 232, 47 },
		{ 153, 233, 47 },
		{ 153, 234, 47 },
		{ 153, 235, 47 },
		{ 153, 236, 47 },
		{ 153, 237, 47 },
		{ 153, 238, 47 },
		{ 153, 239, 47 },
		{ 153, 240, 47 },
		{ 153, 241, 47 },
		{ 153, 242, 47 },
		{ 153, 243, 47 },
		{ 153, 244, 47 },
		{ 153, 245, 47 },
		{ 153, 246, 47 },
		{ 153, 247, 47 },
		{ 153, 248, 47 },
		{ 153, 249, 47 },
		{ 153, 250, 47 },
		{ 153, 251, 47 },
		{ 153, 252, 47 },
		{ 153, 253, 47 },
		{ 153, 254, 47 },
		{ 153, 255, 47 },
		{ 153, 256, 47 },
		{ 154, 232, 47 },
		{ 154, 233, 47 },
		{ 154, 234, 47 },
		{ 154, 235, 47 },
		{ 154, 236, 47 },
		{ 154, 237, 47 },
		{ 154, 238, 47 },
		{ 154, 239, 47 },
		{ 154, 240, 47 },
		{ 154, 241, 47 },
		{ 154, 242, 47 },
		{ 154, 243, 47 },
		{ 154, 244, 47 },
		{ 154, 245, 47 },
		{ 154, 246, 47 },
		{ 154, 247, 47 },
		{ 154, 248, 47 },
		{ 154, 249, 47 },
		{ 154, 250, 47 },
		{ 154, 251, 47 },
		{ 154, 252, 47 },
		{ 154, 253, 47 },
		{ 154, 254, 47 },
		{ 154, 255, 47 },
		{ 154, 256, 47 },
		{ 155, 232, 47 },
		{ 155, 233, 47 },
		{ 155, 234, 47 },
		{ 155, 235, 47 },
		{ 155, 236, 47 },
		{ 155, 237, 47 },
		{ 155, 238, 47 },
		{ 155, 239, 47 },
		{ 155, 240, 47 },
		{ 155, 241, 47 },
		{ 155, 242, 47 },
		{ 155, 243, 47 },
		{ 155, 244, 47 },
		{ 155, 245, 47 },
		{ 155, 246, 47 },
		{ 155, 247, 47 },
		{ 155, 248, 47 },
		{ 155, 249, 47 },
		{ 155, 250, 47 },
		{ 155, 251, 47 },
		{ 155, 252, 47 },
		{ 155, 253, 47 },
		{ 155, 254, 47 },
		{ 155, 255, 47 },
		{ 155, 256, 47 },
		{ 156, 232, 47 },
		{ 156, 233, 47 },
		{ 156, 234, 47 },
		{ 156, 235, 47 },
		{ 156, 236, 47 },
		{ 156, 237, 47 },
		{ 156, 238, 47 },
		{ 156, 239, 47 },
		{ 156, 240, 47 },
		{ 156, 241, 47 },
		{ 156, 242, 47 },
		{ 156, 243, 47 },
		{ 156, 244, 47 },
		{ 156, 245, 47 },
		{ 156, 246, 47 },
		{ 156, 247, 47 },
		{ 156, 248, 47 },
		{ 156, 249, 47 },
		{ 156, 250, 47 },
		{ 156, 251, 47 },
		{ 156, 252, 47 },
		{ 156, 253, 47 },
		{ 156, 254, 47 },
		{ 156, 255, 47 },
		{ 156, 256, 47 },
		{ 157, 233, 47 },
		{ 157, 234, 47 },
		{ 157, 235, 47 },
		{ 157, 236, 47 },
		{ 157, 237, 47 },
		{ 157, 238, 47 },
		{ 157, 239, 47 },
		{ 157, 240, 47 },
		{ 157, 241, 47 },
		{ 157, 242, 47 },
		{ 157, 243, 47 },
		{ 157, 244, 47 },
		{ 157, 245, 47 },
		{ 157, 246, 47 },
		{ 157, 247, 47 },
		{ 157, 248, 47 },
		{ 157, 249, 47 },
		{ 157, 250, 47 },
		{ 157, 251, 47 },
		{ 157, 252, 47 },
		{ 157, 253, 47 },
		{ 157, 254, 47 },
		{ 157, 255, 47 },
		{ 157, 256, 47 },
		{ 158, 233, 47 },
		{ 158, 234, 47 },
		{ 158, 235, 47 },
		{ 158, 236, 47 },
		{ 158, 237, 47 },
		{ 158, 238, 47 },
		{ 158, 239, 47 },
		{ 158, 240, 47 },
		{ 158, 241, 47 },
		{ 158, 242, 47 },
		{ 158, 243, 47 },
		{ 158, 244, 47 },
		{ 158, 245, 47 },
		{ 158, 246, 47 },
		{ 158, 247, 47 },
		{ 158, 248, 47 },
		{ 158, 249, 47 },
		{ 158, 250, 47 },
		{ 158, 251, 47 },
		{ 158, 252, 47 },
		{ 158, 253, 47 },
		{ 158, 254, 47 },
		{ 158, 255, 47 },
		{ 158, 256, 47 },
		{ 159, 233, 47 },
		{ 159, 234, 47 },
		{ 159, 235, 47 },
		{ 159, 236, 47 },
		{ 159, 237, 47 },
		{ 159, 238, 47 },
		{ 159, 239, 47 },
		{ 159, 240, 47 },
		{ 159, 241, 47 },
		{ 159, 242, 47 },
		{ 159, 243, 47 },
		{ 159, 244, 47 },
		{ 159, 245, 47 },
		{ 159, 246, 47 },
		{ 159, 247, 47 },
		{ 159, 248, 47 },
		{ 159, 249, 47 },
		{ 159, 250, 47 },
		{ 159, 251, 47 },
		{ 159, 252, 47 },
		{ 159, 253, 47 },
		{ 159, 254, 47 },
		{ 159, 255, 47 },
		{ 159, 256, 47 },
		{ 160, 233, 47 },
		{ 160, 234, 47 },
		{ 160, 235, 47 },
		{ 160, 236, 47 },
		{ 160, 237, 47 },
		{ 160, 238, 47 },
		{ 160, 239, 47 },
		{ 160, 240, 47 },
		{ 160, 241, 47 },
		{ 160, 242, 47 },
		{ 160, 243, 47 },
		{ 160, 244, 47 },
		{ 160, 245, 47 },
		{ 160, 246, 47 },
		{ 160, 247, 47 },
		{ 160, 248, 47 },
		{ 160, 249, 47 },
		{ 160, 250, 47 },
		{ 160, 251, 47 },
		{ 160, 252, 47 },
		{ 160, 253, 47 },
		{ 160, 254, 47 },
		{ 160, 255, 47 },
		{ 160, 256, 47 },
		{ 161, 234, 47 },
		{ 161, 235, 47 },
		{ 161, 236, 47 },
		{ 161, 237, 47 },
		{ 161, 238, 47 },
		{ 161, 239, 47 },
		{ 161, 240, 47 },
		{ 161, 241, 47 },
		{ 161, 242, 47 },
		{ 161, 243, 47 },
		{ 161, 244, 47 },
		{ 161, 245, 47 },
		{ 161, 246, 47 },
		{ 161, 247, 47 },
		{ 161, 248, 47 },
		{ 161, 249, 47 },
		{ 161, 250, 47 },
		{ 161, 251, 47 },
		{ 161, 252, 47 },
		{ 161, 253, 47 },
		{ 161, 254, 47 },
		{ 161, 255, 47 },
		{ 161, 256, 47 },
		{ 162, 234, 47 },
		{ 162, 235, 47 },
		{ 162, 236, 47 },
		{ 162, 237, 47 },
		{ 162, 238, 47 },
		{ 162, 239, 47 },
		{ 162, 240, 47 },
		{ 162, 241, 47 },
		{ 162, 242, 47 },
		{ 162, 243, 47 },
		{ 162, 244, 47 },
		{ 162, 245, 47 },
		{ 162, 246, 47 },
		{ 162, 247, 47 },
		{ 162, 248, 47 },
		{ 162, 249, 47 },
		{ 162, 250, 47 },
		{ 162, 251, 47 },
		{ 162, 252, 47 },
		{ 162, 253, 47 },
		{ 162, 254, 47 },
		{ 162, 255, 47 },
		{ 162, 256, 47 },
		{ 163, 234, 47 },
		{ 163, 235, 47 },
		{ 163, 236, 47 },
		{ 163, 237, 47 },
		{ 163, 238, 47 },
		{ 163, 239, 47 },
		{ 163, 240, 47 },
		{ 163, 241, 47 },
		{ 163, 242, 47 },
		{ 163, 243, 47 },
		{ 163, 244, 47 },
		{ 163, 245, 47 },
		{ 163, 246, 47 },
		{ 163, 247, 47 },
		{ 163, 248, 47 },
		{ 163, 249, 47 },
		{ 163, 250, 47 },
		{ 163, 251, 47 },
		{ 163, 252, 47 },
		{ 163, 253, 47 },
		{ 163, 254, 47 },
		{ 163, 255, 47 },
		{ 163, 256, 47 },
		{ 164, 234, 47 },
		{ 164, 235, 47 },
		{ 164, 236, 47 },
		{ 164, 237, 47 },
		{ 164, 238, 47 },
		{ 164, 239, 47 },
		{ 164, 240, 47 },
		{ 164, 241, 47 },
		{ 164, 242, 47 },
		{ 164, 243, 47 },
		{ 164, 244, 47 },
		{ 164, 245, 47 },
		{ 164, 246, 47 },
		{ 164, 247, 47 },
		{ 164, 248, 47 },
		{ 164, 249, 47 },
		{ 164, 250, 47 },
		{ 164, 251, 47 },
		{ 164, 252, 47 },
		{ 164, 253, 47 },
		{ 164, 254, 47 },
		{ 164, 255, 47 },
		{ 164, 256, 47 },
		{ 165, 235, 47 },
		{ 165, 236, 47 },
		{ 165, 237, 47 },
		{ 165, 238, 47 },
		{ 165, 239, 47 },
		{ 165, 240, 47 },
		{ 165, 241, 47 },
		{ 165, 242, 47 },
		{ 165, 243, 47 },
		{ 165, 244, 47 },
		{ 165, 245, 47 },
		{ 165, 246, 47 },
		{ 165, 247, 47 },
		{ 165, 248, 47 },
		{ 165, 249, 47 },
		{ 165, 250, 47 },
		{ 165, 251, 47 },
		{ 165, 252, 47 },
		{ 165, 253, 47 },
		{ 165, 254, 47 },
		{ 165, 255, 47 },
		{ 165, 256, 47 },
		{ 166, 235, 47 },
		{ 166, 236, 47 },
		{ 166, 237, 47 },
		{ 166, 238, 47 },
		{ 166, 239, 47 },
		{ 166, 240, 47 },
		{ 166, 241, 47 },
		{ 166, 242, 47 },
		{ 166, 243, 47 },
		{ 166, 244, 47 },
		{ 166, 245, 47 },
		{ 166, 246, 47 },
		{ 166, 247, 47 },
		{ 166, 248, 47 },
		{ 166, 249, 47 },
		{ 166, 250, 47 },
		{ 166, 251, 47 },
		{ 166, 252, 47 },
		{ 166, 253, 47 },
		{ 166, 254, 47 },
		{ 166, 255, 47 },
		{ 166, 256, 47 },
		{ 167, 235, 47 },
		{ 167, 236, 47 },
		{ 167, 237, 47 },
		{ 167, 238, 47 },
		{ 167, 239, 47 },
		{ 167, 240, 47 },
		{ 167, 241, 47 },
		{ 167, 242, 47 },
		{ 167, 243, 47 },
		{ 167, 244, 47 },
		{ 167, 245, 47 },
		{ 167, 246, 47 },
		{ 167, 247, 47 },
		{ 167, 248, 47 },
		{ 167, 249, 47 },
		{ 167, 250, 47 },
		{ 167, 251, 47 },
		{ 167, 252, 47 },
		{ 167, 253, 47 },
		{ 167, 254, 47 },
		{ 167, 255, 47 },
		{ 167, 256, 47 },
		{ 168, 235, 47 },
		{ 168, 236, 47 },
		{ 168, 237, 47 },
		{ 168, 238, 47 },
		{ 168, 239, 47 },
		{ 168, 240, 47 },
		{ 168, 241, 47 },
		{ 168, 242, 47 },
		{ 168, 243, 47 },
		{ 168, 244, 47 },
		{ 168, 245, 47 },
		{ 168, 246, 47 },
		{ 168, 247, 47 },
		{ 168, 248, 47 },
		{ 168, 249, 47 },
		{ 168, 250, 47 },
		{ 168, 251, 47 },
		{ 168, 252, 47 },
		{ 168, 253, 47 },
		{ 168, 254, 47 },
		{ 168, 255, 47 },
		{ 168, 256, 47 },
		{ 169, 236, 47 },
		{ 169, 237, 47 },
		{ 169, 238, 47 },
		{ 169, 239, 47 },
		{ 169, 240, 47 },
		{ 169, 241, 47 },
		{ 169, 242, 47 },
		{ 169, 243, 47 },
		{ 169, 244, 47 },
		{ 169, 245, 47 },
		{ 169, 246, 47 },
		{ 169, 247, 47 },
		{ 169, 248, 47 },
		{ 169, 249, 47 },
		{ 169, 250, 47 },
		{ 169, 251, 47 },
		{ 169, 252, 47 },
		{ 169, 253, 47 },
		{ 169, 254, 47 },
		{ 169, 255, 47 },
		{ 169, 256, 47 },
		{ 170, 236, 47 },
		{ 170, 237, 47 },
		{ 170, 238, 47 },
		{ 170, 239, 47 },
		{ 170, 240, 47 },
		{ 170, 241, 47 },
		{ 170, 242, 47 },
		{ 170, 243, 47 },
		{ 170, 244, 47 },
		{ 170, 245, 47 },
		{ 170, 246, 47 },
		{ 170, 247, 47 },
		{ 170, 248, 47 },
		{ 170, 249, 47 },
		{ 170, 250, 47 },
		{ 170, 251, 47 },
		{ 170, 252, 47 },
		{ 170, 253, 47 },
		{ 170, 254, 47 },
		{ 170, 255, 47 },
		{ 170, 256, 47 },
		{ 171, 236, 47 },
		{ 171, 237, 47 },
		{ 171, 238, 47 },
		{ 171, 239, 47 },
		{ 171, 240, 47 },
		{ 171, 241, 47 },
		{ 171, 242, 47 },
		{ 171, 243, 47 },
		{ 171, 244, 47 },
		{ 171, 245, 47 },
		{ 171, 246, 47 },
		{ 171, 247, 47 },
		{ 171, 248, 47 },
		{ 171, 249, 47 },
		{ 171, 250, 47 },
		{ 171, 251, 47 },
		{ 171, 252, 47 },
		{ 171, 253, 47 },
		{ 171, 254, 47 },
		{ 171, 255, 47 },
		{ 171, 256, 47 },
		{ 172, 236, 47 },
		{ 172, 237, 47 },
		{ 172, 238, 47 },
		{ 172, 239, 47 },
		{ 172, 240, 47 },
		{ 172, 241, 47 },
		{ 172, 242, 47 },
		{ 172, 243, 47 },
		{ 172, 244, 47 },
		{ 172, 245, 47 },
		{ 172, 246, 47 },
		{ 172, 247, 47 },
		{ 172, 248, 47 },
		{ 172, 249, 47 },
		{ 172, 250, 47 },
		{ 172, 251, 47 },
		{ 172, 252, 47 },
		{ 172, 253, 47 },
		{ 172, 254, 47 },
		{ 172, 255, 47 },
		{ 172, 256, 47 },
		{ 173, 237, 47 },
		{ 173, 238, 47 },
		{ 173, 239, 47 },
		{ 173, 240, 47 },
		{ 173, 241, 47 },
		{ 173, 242, 47 },
		{ 173, 243, 47 },
		{ 173, 244, 47 },
		{ 173, 245, 47 },
		{ 173, 246, 47 },
		{ 173, 247, 47 },
		{ 173, 248, 47 },
		{ 173, 249, 47 },
		{ 173, 250, 47 },
		{ 173, 251, 47 },
		{ 173, 252, 47 },
		{ 173, 253, 47 },
		{ 173, 254, 47 },
		{ 173, 255, 47 },
		{ 173, 256, 47 },
		{ 174, 237, 47 },
		{ 174, 238, 47 },
		{ 174, 239, 47 },
		{ 174, 240, 47 },
		{ 174, 241, 47 },
		{ 174, 242, 47 },
		{ 174, 243, 47 },
		{ 174, 244, 47 },
		{ 174, 245, 47 },
		{ 174, 246, 47 },
		{ 174, 247, 47 },
		{ 174, 248, 47 },
		{ 174, 249, 47 },
		{ 174, 250, 47 },
		{ 174, 251, 47 },
		{ 174, 252, 47 },
		{ 174, 253, 47 },
		{ 174, 254, 47 },
		{ 174, 255, 47 },
		{ 174, 256, 47 },
		{ 175, 237, 47 },
		{ 175, 238, 47 },
		{ 175, 239, 47 },
		{ 175, 240, 47 },
		{ 175, 241, 47 },
		{ 175, 242, 47 },
		{ 175, 243, 47 },
		{ 175, 244, 47 },
		{ 175, 245, 47 },
		{ 175, 246, 47 },
		{ 175, 247, 47 },
		{ 175, 248, 47 },
		{ 175, 249, 47 },
		{ 175, 250, 47 },
		{ 175, 251, 47 },
		{ 175, 252, 47 },
		{ 175, 253, 47 },
		{ 175, 254, 47 },
		{ 175, 255, 47 },
		{ 175, 256, 47 },
		{ 176, 237, 47 },
		{ 176, 238, 47 },
		{ 176, 239, 47 },
		{ 176, 240, 47 },
		{ 176, 241, 47 },
		{ 176, 242, 47 },
		{ 176, 243, 47 },
		{ 176, 244, 47 },
		{ 176, 245, 47 },
		{ 176, 246, 47 },
		{ 176, 247, 47 },
		{ 176, 248, 47 },
		{ 176, 249, 47 },
		{ 176, 250, 47 },
		{ 176, 251, 47 },
		{ 176, 252, 47 },
		{ 176, 253, 47 },
		{ 176, 254, 47 },
		{ 176, 255, 47 },
		{ 176, 256, 47 },
		{ 177, 238, 47 },
		{ 177, 239, 47 },
		{ 177, 240, 47 },
		{ 177, 241, 47 },
		{ 177, 242, 47 },
		{ 177, 243, 47 },
		{ 177, 244, 47 },
		{ 177, 245, 47 },
		{ 177, 246, 47 },
		{ 177, 247, 47 },
		{ 177, 248, 47 },
		{ 177, 249, 47 },
		{ 177, 250, 47 },
		{ 177, 251, 47 },
		{ 177, 252, 47 },
		{ 177, 253, 47 },
		{ 177, 254, 47 },
		{ 177, 255, 47 },
		{ 177, 256, 47 },
		{ 178, 238, 47 },
		{ 178, 239, 47 },
		{ 178, 240, 47 },
		{ 178, 241, 47 },
		{ 178, 242, 47 },
		{ 178, 243, 47 },
		{ 178, 244, 47 },
		{ 178, 245, 47 },
		{ 178, 246, 47 },
		{ 178, 247, 47 },
		{ 178, 248, 47 },
		{ 178, 249, 47 },
		{ 178, 250, 47 },
		{ 178, 251, 47 },
		{ 178, 252, 47 },
		{ 178, 253, 47 },
		{ 178, 254, 47 },
		{ 178, 255, 47 },
		{ 178, 256, 47 },
		{ 179, 238, 47 },
		{ 179, 239, 47 },
		{ 179, 240, 47 },
		{ 179, 241, 47 },
		{ 179, 242, 47 },
		{ 179, 243, 47 },
		{ 179, 244, 47 },
		{ 179, 245, 47 },
		{ 179, 246, 47 },
		{ 179, 247, 47 },
		{ 179, 248, 47 },
		{ 179, 249, 47 },
		{ 179, 250, 47 },
		{ 179, 251, 47 },
		{ 179, 252, 47 },
		{ 179, 253, 47 },
		{ 179, 254, 47 },
		{ 179, 255, 47 },
		{ 179, 256, 47 },
		{ 180, 238, 47 },
		{ 180, 239, 47 },
		{ 180, 240, 47 },
		{ 180, 241, 47 },
		{ 180, 242, 47 },
		{ 180, 243, 47 },
		{ 180, 244, 47 },
		{ 180, 245, 47 },
		{ 180, 246, 47 },
		{ 180, 247, 47 },
		{ 180, 248, 47 },
		{ 180, 249, 47 },
		{ 180, 250, 47 },
		{ 180, 251, 47 },
		{ 180, 252, 47 },
		{ 180, 253, 47 },
		{ 180, 254, 47 },
		{ 180, 255, 47 },
		{ 180, 256, 47 },
		{ 181, 239, 47 },
		{ 181, 240, 47 },
		{ 181, 241, 47 },
		{ 181, 242, 47 },
		{ 181, 243, 47 },
		{ 181, 244, 47 },
		{ 181, 245, 47 },
		{ 181, 246, 47 },
		{ 181, 247, 47 },
		{ 181, 248, 47 },
		{ 181, 249, 47 },
		{ 181, 250, 47 },
		{ 181, 251, 47 },
		{ 181, 252, 47 },
		{ 181, 253, 47 },
		{ 181, 254, 47 },
		{ 181, 255, 47 },
		{ 181, 256, 47 },
		{ 182, 239, 47 },
		{ 182, 240, 47 },
		{ 182, 241, 47 },
		{ 182, 242, 47 },
		{ 182, 243, 47 },
		{ 182, 244, 47 },
		{ 182, 245, 47 },
		{ 182, 246, 47 },
		{ 182, 247, 47 },
		{ 182, 248, 47 },
		{ 182, 249, 47 },
		{ 182, 250, 47 },
		{ 182, 251, 47 },
		{ 182, 252, 47 },
		{ 182, 253, 47 },
		{ 182, 254, 47 },
		{ 182, 255, 47 },
		{ 182, 256, 47 },
		{ 183, 239, 47 },
		{ 183, 240, 47 },
		{ 183, 241, 47 },
		{ 183, 242, 47 },
		{ 183, 243, 47 },
		{ 183, 244, 47 },
		{ 183, 245, 47 },
		{ 183, 246, 47 },
		{ 183, 247, 47 },
		{ 183, 248, 47 },
		{ 183, 249, 47 },
		{ 183, 250, 47 },
		{ 183, 251, 47 },
		{ 183, 252, 47 },
		{ 183, 253, 47 },
		{ 183, 254, 47 },
		{ 183, 255, 47 },
		{ 183, 256, 47 },
		{ 184, 239, 47 },
		{ 184, 240, 47 },
		{ 184, 241, 47 },
		{ 184, 242, 47 },
		{ 184, 243, 47 },
		{ 184, 244, 47 },
		{ 184, 245, 47 },
		{ 184, 246, 47 },
		{ 184, 247, 47 },
		{ 184, 248, 47 },
		{ 184, 249, 47 },
		{ 184, 250, 47 },
		{ 184, 251, 47 },
		{ 184, 252, 47 },
		{ 184, 253, 47 },
		{ 184, 254, 47 },
		{ 184, 255, 47 },
		{ 184, 256, 47 },
		{ 185, 240, 47 },
		{ 185, 241, 47 },
		{ 185, 242, 47 },
		{ 185, 243, 47 },
		{ 185, 244, 47 },
		{ 185, 245, 47 },
		{ 185, 246, 47 },
		{ 185, 247, 47 },
		{ 185, 248, 47 },
		{ 185, 249, 47 },
		{ 185, 250, 47 },
		{ 185, 251, 47 },
		{ 185, 252, 47 },
		{ 185, 253, 47 },
		{ 185, 254, 47 },
		{ 185, 255, 47 },
		{ 185, 256, 47 },
		{ 186, 240, 47 },
		{ 186, 241, 47 },
		{ 186, 242, 47 },
		{ 186, 243, 47 },
		{ 186, 244, 47 },
		{ 186, 245, 47 },
		{ 186, 246, 47 },
		{ 186, 247, 47 },
		{ 186, 248, 47 },
		{ 186, 249, 47 },
		{ 186, 250, 47 },
		{ 186, 251, 47 },
		{ 186, 252, 47 },
		{ 186, 253, 47 },
		{ 186, 254, 47 },
		{ 186, 255, 47 },
		{ 186, 256, 47 },
		{ 187, 240, 47 },
		{ 187, 241, 47 },
		{ 187, 242, 47 },
		{ 187, 243, 47 },
		{ 187, 244, 47 },
		{ 187, 245, 47 },
		{ 187, 246, 47 },
		{ 187, 247, 47 },
		{ 187, 248, 47 },
		{ 187, 249, 47 },
		{ 187, 250, 47 },
		{ 187, 251, 47 },
		{ 187, 252, 47 },
		{ 187, 253, 47 },
		{ 187, 254, 47 },
		{ 187, 255, 47 },
		{ 187, 256, 47 },
		{ 188, 241, 47 },
		{ 188, 242, 47 },
		{ 188, 243, 47 },
		{ 188, 244, 47 },
		{ 188, 245, 47 },
		{ 188, 246, 47 },
		{ 188, 247, 47 },
		{ 188, 248, 47 },
		{ 188, 249, 47 },
		{ 188, 250, 47 },
		{ 188, 251, 47 },
		{ 188, 252, 47 },
		{ 188, 253, 47 },
		{ 188, 254, 47 },
		{ 188, 255, 47 },
		{ 188, 256, 47 },
		{ 189, 241, 47 },
		{ 189, 242, 47 },
		{ 189, 243, 47 },
		{ 189, 244, 47 },
		{ 189, 245, 47 },
		{ 189, 246, 47 },
		{ 189, 247, 47 },
		{ 189, 248, 47 },
		{ 189, 249, 47 },
		{ 189, 250, 47 },
		{ 189, 251, 47 },
		{ 189, 252, 47 },
		{ 189, 253, 47 },
		{ 189, 254, 47 },
		{ 189, 255, 47 },
		{ 189, 256, 47 },
		{ 190, 241, 47 },
		{ 190, 242, 47 },
		{ 190, 243, 47 },
		{ 190, 244, 47 },
		{ 190, 245, 47 },
		{ 190, 246, 47 },
		{ 190, 247, 47 },
		{ 190, 248, 47 },
		{ 190, 249, 47 },
		{ 190, 250, 47 },
		{ 190, 251, 47 },
		{ 190, 252, 47 },
		{ 190, 253, 47 },
		{ 190, 254, 47 },
		{ 190, 255, 47 },
		{ 190, 256, 47 },
		{ 191, 243, 47 },
		{ 191, 244, 47 },
		{ 191, 245, 47 },
		{ 191, 246, 47 },
		{ 191, 247, 47 },
		{ 191, 248, 47 },
		{ 191, 249, 47 },
		{ 191, 250, 47 },
		{ 191, 251, 47 },
		{ 191, 252, 47 },
		{ 191, 253, 47 },
		{ 191, 254, 47 },
		{ 191, 255, 47 },
		{ 323, 259, 47 },
		{ 323, 260, 47 },
		{ 323, 261, 47 },
		{ 323, 262, 47 },
		{ 323, 263, 47 },
		{ 323, 264, 47 },
		{ 323, 265, 47 },
		{ 323, 266, 47 },
		{ 323, 267, 47 },
		{ 323, 268, 47 },
		{ 323, 269, 47 },
		{ 323, 270, 47 },
		{ 323, 271, 47 },
		{ 324, 258, 47 },
		{ 324, 259, 47 },
		{ 324, 260, 47 },
		{ 324, 261, 47 },
		{ 324, 262, 47 },
		{ 324, 263, 47 },
		{ 324, 264, 47 },
		{ 324, 265, 47 },
		{ 324, 266, 47 },
		{ 324, 267, 47 },
		{ 324, 268, 47 },
		{ 324, 269, 47 },
		{ 324, 270, 47 },
		{ 324, 271, 47 },
		{ 324, 272, 47 },
		{ 324, 273, 47 },
		{ 325, 258, 47 },
		{ 325, 259, 47 },
		{ 325, 260, 47 },
		{ 325, 261, 47 },
		{ 325, 262, 47 },
		{ 325, 263, 47 },
		{ 325, 264, 47 },
		{ 325, 265, 47 },
		{ 325, 266, 47 },
		{ 325, 267, 47 },
		{ 325, 268, 47 },
		{ 325, 269, 47 },
		{ 325, 270, 47 },
		{ 325, 271, 47 },
		{ 325, 272, 47 },
		{ 325, 273, 47 },
		{ 326, 258, 47 },
		{ 326, 259, 47 },
		{ 326, 260, 47 },
		{ 326, 261, 47 },
		{ 326, 262, 47 },
		{ 326, 263, 47 },
		{ 326, 264, 47 },
		{ 326, 265, 47 },
		{ 326, 266, 47 },
		{ 326, 267, 47 },
		{ 326, 268, 47 },
		{ 326, 269, 47 },
		{ 326, 270, 47 },
		{ 326, 271, 47 },
		{ 326, 272, 47 },
		{ 326, 273, 47 },
		{ 327, 258, 47 },
		{ 327, 259, 47 },
		{ 327, 260, 47 },
		{ 327, 261, 47 },
		{ 327, 262, 47 },
		{ 327, 263, 47 },
		{ 327, 264, 47 },
		{ 327, 265, 47 },
		{ 327, 266, 47 },
		{ 327, 267, 47 },
		{ 327, 268, 47 },
		{ 327, 269, 47 },
		{ 327, 270, 47 },
		{ 327, 271, 47 },
		{ 327, 272, 47 },
		{ 327, 273, 47 },
		{ 327, 274, 47 },
		{ 328, 258, 47 },
		{ 328, 259, 47 },
		{ 328, 260, 47 },
		{ 328, 261, 47 },
		{ 328, 262, 47 },
		{ 328, 263, 47 },
		{ 328, 264, 47 },
		{ 328, 265, 47 },
		{ 328, 266, 47 },
		{ 328, 267, 47 },
		{ 328, 268, 47 },
		{ 328, 269, 47 },
		{ 328, 270, 47 },
		{ 328, 271, 47 },
		{ 328, 272, 47 },
		{ 328, 273, 47 },
		{ 328, 274, 47 },
		{ 329, 258, 47 },
		{ 329, 259, 47 },
		{ 329, 260, 47 },
		{ 329, 261, 47 },
		{ 329, 262, 47 },
		{ 329, 263, 47 },
		{ 329, 264, 47 },
		{ 329, 265, 47 },
		{ 329, 266, 47 },
		{ 329, 267, 47 },
		{ 329, 268, 47 },
		{ 329, 269, 47 },
		{ 329, 270, 47 },
		{ 329, 271, 47 },
		{ 329, 272, 47 },
		{ 329, 273, 47 },
		{ 329, 274, 47 },
		{ 330, 258, 47 },
		{ 330, 259, 47 },
		{ 330, 260, 47 },
		{ 330, 261, 47 },
		{ 330, 262, 47 },
		{ 330, 263, 47 },
		{ 330, 264, 47 },
		{ 330, 265, 47 },
		{ 330, 266, 47 },
		{ 330, 267, 47 },
		{ 330, 268, 47 },
		{ 330, 269, 47 },
		{ 330, 270, 47 },
		{ 330, 271, 47 },
		{ 330, 272, 47 },
		{ 330, 273, 47 },
		{ 330, 274, 47 },
		{ 330, 275, 47 },
		{ 331, 258, 47 },
		{ 331, 259, 47 },
		{ 331, 260, 47 },
		{ 331, 261, 47 },
		{ 331, 262, 47 },
		{ 331, 263, 47 },
		{ 331, 264, 47 },
		{ 331, 265, 47 },
		{ 331, 266, 47 },
		{ 331, 267, 47 },
		{ 331, 268, 47 },
		{ 331, 269, 47 },
		{ 331, 270, 47 },
		{ 331, 271, 47 },
		{ 331, 272, 47 },
		{ 331, 273, 47 },
		{ 331, 274, 47 },
		{ 331, 275, 47 },
		{ 332, 258, 47 },
		{ 332, 259, 47 },
		{ 332, 260, 47 },
		{ 332, 261, 47 },
		{ 332, 262, 47 },
		{ 332, 263, 47 },
		{ 332, 264, 47 },
		{ 332, 265, 47 },
		{ 332, 266, 47 },
		{ 332, 267, 47 },
		{ 332, 268, 47 },
		{ 332, 269, 47 },
		{ 332, 270, 47 },
		{ 332, 271, 47 },
		{ 332, 272, 47 },
		{ 332, 273, 47 },
		{ 332, 274, 47 },
		{ 332, 275, 47 },
		{ 333, 258, 47 },
		{ 333, 259, 47 },
		{ 333, 260, 47 },
		{ 333, 261, 47 },
		{ 333, 262, 47 },
		{ 333, 263, 47 },
		{ 333, 264, 47 },
		{ 333, 265, 47 },
		{ 333, 266, 47 },
		{ 333, 267, 47 },
		{ 333, 268, 47 },
		{ 333, 269, 47 },
		{ 333, 270, 47 },
		{ 333, 271, 47 },
		{ 333, 272, 47 },
		{ 333, 273, 47 },
		{ 333, 274, 47 },
		{ 333, 275, 47 },
		{ 334, 258, 47 },
		{ 334, 259, 47 },
		{ 334, 260, 47 },
		{ 334, 261, 47 },
		{ 334, 262, 47 },
		{ 334, 263, 47 },
		{ 334, 264, 47 },
		{ 334, 265, 47 },
		{ 334, 266, 47 },
		{ 334, 267, 47 },
		{ 334, 268, 47 },
		{ 334, 269, 47 },
		{ 334, 270, 47 },
		{ 334, 271, 47 },
		{ 334, 272, 47 },
		{ 334, 273, 47 },
		{ 334, 274, 47 },
		{ 334, 275, 47 },
		{ 334, 276, 47 },
		{ 335, 258, 47 },
		{ 335, 259, 47 },
		{ 335, 260, 47 },
		{ 335, 261, 47 },
		{ 335, 262, 47 },
		{ 335, 263, 47 },
		{ 335, 264, 47 },
		{ 335, 265, 47 },
		{ 335, 266, 47 },
		{ 335, 267, 47 },
		{ 335, 268, 47 },
		{ 335, 269, 47 },
		{ 335, 270, 47 },
		{ 335, 271, 47 },
		{ 335, 272, 47 },
		{ 335, 273, 47 },
		{ 335, 274, 47 },
		{ 335, 275, 47 },
		{ 335, 276, 47 },
		{ 336, 258, 47 },
		{ 336, 259, 47 },
		{ 336, 260, 47 },
		{ 336, 261, 47 },
		{ 336, 262, 47 },
		{ 336, 263, 47 },
		{ 336, 264, 47 },
		{ 336, 265, 47 },
		{ 336, 266, 47 },
		{ 336, 267, 47 },
		{ 336, 268, 47 },
		{ 336, 269, 47 },
		{ 336, 270, 47 },
		{ 336, 271, 47 },
		{ 336, 272, 47 },
		{ 336, 273, 47 },
		{ 336, 274, 47 },
		{ 336, 275, 47 },
		{ 336, 276, 47 },
		{ 337, 258, 47 },
		{ 337, 259, 47 },
		{ 337, 260, 47 },
		{ 337, 261, 47 },
		{ 337, 262, 47 },
		{ 337, 263, 47 },
		{ 337, 264, 47 },
		{ 337, 265, 47 },
		{ 337, 266, 47 },
		{ 337, 267, 47 },
		{ 337, 268, 47 },
		{ 337, 269, 47 },
		{ 337, 270, 47 },
		{ 337, 271, 47 },
		{ 337, 272, 47 },
		{ 337, 273, 47 },
		{ 337, 274, 47 },
		{ 337, 275, 47 },
		{ 337, 276, 47 },
		{ 338, 258, 47 },
		{ 338, 259, 47 },
		{ 338, 260, 47 },
		{ 338, 261, 47 },
		{ 338, 262, 47 },
		{ 338, 263, 47 },
		{ 338, 264, 47 },
		{ 338, 265, 47 },
		{ 338, 266, 47 },
		{ 338, 267, 47 },
		{ 338, 268, 47 },
		{ 338, 269, 47 },
		{ 338, 270, 47 },
		{ 338, 271, 47 },
		{ 338, 272, 47 },
		{ 338, 273, 47 },
		{ 338, 274, 47 },
		{ 338, 275, 47 },
		{ 338, 276, 47 },
		{ 338, 277, 47 },
		{ 339, 258, 47 },
		{ 339, 259, 47 },
		{ 339, 260, 47 },
		{ 339, 261, 47 },
		{ 339, 262, 47 },
		{ 339, 263, 47 },
		{ 339, 264, 47 },
		{ 339, 265, 47 },
		{ 339, 266, 47 },
		{ 339, 267, 47 },
		{ 339, 268, 47 },
		{ 339, 269, 47 },
		{ 339, 270, 47 },
		{ 339, 271, 47 },
		{ 339, 272, 47 },
		{ 339, 273, 47 },
		{ 339, 274, 47 },
		{ 339, 275, 47 },
		{ 339, 276, 47 },
		{ 339, 277, 47 },
		{ 340, 258, 47 },
		{ 340, 259, 47 },
		{ 340, 260, 47 },
		{ 340, 261, 47 },
		{ 340, 262, 47 },
		{ 340, 263, 47 },
		{ 340, 264, 47 },
		{ 340, 265, 47 },
		{ 340, 266, 47 },
		{ 340, 267, 47 },
		{ 340, 268, 47 },
		{ 340, 269, 47 },
		{ 340, 270, 47 },
		{ 340, 271, 47 },
		{ 340, 272, 47 },
		{ 340, 273, 47 },
		{ 340, 274, 47 },
		{ 340, 275, 47 },
		{ 340, 276, 47 },
		{ 340, 277, 47 },
		{ 341, 258, 47 },
		{ 341, 259, 47 },
		{ 341, 260, 47 },
		{ 341, 261, 47 },
		{ 341, 262, 47 },
		{ 341, 263, 47 },
		{ 341, 264, 47 },
		{ 341, 265, 47 },
		{ 341, 266, 47 },
		{ 341, 267, 47 },
		{ 341, 268, 47 },
		{ 341, 269, 47 },
		{ 341, 270, 47 },
		{ 341, 271, 47 },
		{ 341, 272, 47 },
		{ 341, 273, 47 },
		{ 341, 274, 47 },
		{ 341, 275, 47 },
		{ 341, 276, 47 },
		{ 341, 277, 47 },
		{ 342, 258, 47 },
		{ 342, 259, 47 },
		{ 342, 260, 47 },
		{ 342, 261, 47 },
		{ 342, 262, 47 },
		{ 342, 263, 47 },
		{ 342, 264, 47 },
		{ 342, 265, 47 },
		{ 342, 266, 47 },
		{ 342, 267, 47 },
		{ 342, 268, 47 },
		{ 342, 269, 47 },
		{ 342, 270, 47 },
		{ 342, 271, 47 },
		{ 342, 272, 47 },
		{ 342, 273, 47 },
		{ 342, 274, 47 },
		{ 342, 275, 47 },
		{ 342, 276, 47 },
		{ 342, 277, 47 },
		{ 342, 278, 47 },
		{ 343, 258, 47 },
		{ 343, 259, 47 },
		{ 343, 260, 47 },
		{ 343, 261, 47 },
		{ 343, 262, 47 },
		{ 343, 263, 47 },
		{ 343, 264, 47 },
		{ 343, 265, 47 },
		{ 343, 266, 47 },
		{ 343, 267, 47 },
		{ 343, 268, 47 },
		{ 343, 269, 47 },
		{ 343, 270, 47 },
		{ 343, 271, 47 },
		{ 343, 272, 47 },
		{ 343, 273, 47 },
		{ 343, 274, 47 },
		{ 343, 275, 47 },
		{ 343, 276, 47 },
		{ 343, 277, 47 },
		{ 343, 278, 47 },
		{ 344, 258, 47 },
		{ 344, 259, 47 },
		{ 344, 260, 47 },
		{ 344, 261, 47 },
		{ 344, 262, 47 },
		{ 344, 263, 47 },
		{ 344, 264, 47 },
		{ 344, 265, 47 },
		{ 344, 266, 47 },
		{ 344, 267, 47 },
		{ 344, 268, 47 },
		{ 344, 269, 47 },
		{ 344, 270, 47 },
		{ 344, 271, 47 },
		{ 344, 272, 47 },
		{ 344, 273, 47 },
		{ 344, 274, 47 },
		{ 344, 275, 47 },
		{ 344, 276, 47 },
		{ 344, 277, 47 },
		{ 344, 278, 47 },
		{ 345, 258, 47 },
		{ 345, 259, 47 },
		{ 345, 260, 47 },
		{ 345, 261, 47 },
		{ 345, 262, 47 },
		{ 345, 263, 47 },
		{ 345, 264, 47 },
		{ 345, 265, 47 },
		{ 345, 266, 47 },
		{ 345, 267, 47 },
		{ 345, 268, 47 },
		{ 345, 269, 47 },
		{ 345, 270, 47 },
		{ 345, 271, 47 },
		{ 345, 272, 47 },
		{ 345, 273, 47 },
		{ 345, 274, 47 },
		{ 345, 275, 47 },
		{ 345, 276, 47 },
		{ 345, 277, 47 },
		{ 345, 278, 47 },
		{ 346, 258, 47 },
		{ 346, 259, 47 },
		{ 346, 260, 47 },
		{ 346, 261, 47 },
		{ 346, 262, 47 },
		{ 346, 263, 47 },
		{ 346, 264, 47 },
		{ 346, 265, 47 },
		{ 346, 266, 47 },
		{ 346, 267, 47 },
		{ 346, 268, 47 },
		{ 346, 269, 47 },
		{ 346, 270, 47 },
		{ 346, 271, 47 },
		{ 346, 272, 47 },
		{ 346, 273, 47 },
		{ 346, 274, 47 },
		{ 346, 275, 47 },
		{ 346, 276, 47 },
		{ 346, 277, 47 },
		{ 346, 278, 47 },
		{ 346, 279, 47 },
		{ 347, 258, 47 },
		{ 347, 259, 47 },
		{ 347, 260, 47 },
		{ 347, 261, 47 },
		{ 347, 262, 47 },
		{ 347, 263, 47 },
		{ 347, 264, 47 },
		{ 347, 265, 47 },
		{ 347, 266, 47 },
		{ 347, 267, 47 },
		{ 347, 268, 47 },
		{ 347, 269, 47 },
		{ 347, 270, 47 },
		{ 347, 271, 47 },
		{ 347, 272, 47 },
		{ 347, 273, 47 },
		{ 347, 274, 47 },
		{ 347, 275, 47 },
		{ 347, 276, 47 },
		{ 347, 277, 47 },
		{ 347, 278, 47 },
		{ 347, 279, 47 },
		{ 348, 258, 47 },
		{ 348, 259, 47 },
		{ 348, 260, 47 },
		{ 348, 261, 47 },
		{ 348, 262, 47 },
		{ 348, 263, 47 },
		{ 348, 264, 47 },
		{ 348, 265, 47 },
		{ 348, 266, 47 },
		{ 348, 267, 47 },
		{ 348, 268, 47 },
		{ 348, 269, 47 },
		{ 348, 270, 47 },
		{ 348, 271, 47 },
		{ 348, 272, 47 },
		{ 348, 273, 47 },
		{ 348, 274, 47 },
		{ 348, 275, 47 },
		{ 348, 276, 47 },
		{ 348, 277, 47 },
		{ 348, 278, 47 },
		{ 348, 279, 47 },
		{ 349, 258, 47 },
		{ 349, 259, 47 },
		{ 349, 260, 47 },
		{ 349, 261, 47 },
		{ 349, 262, 47 },
		{ 349, 263, 47 },
		{ 349, 264, 47 },
		{ 349, 265, 47 },
		{ 349, 266, 47 },
		{ 349, 267, 47 },
		{ 349, 268, 47 },
		{ 349, 269, 47 },
		{ 349, 270, 47 },
		{ 349, 271, 47 },
		{ 349, 272, 47 },
		{ 349, 273, 47 },
		{ 349, 274, 47 },
		{ 349, 275, 47 },
		{ 349, 276, 47 },
		{ 349, 277, 47 },
		{ 349, 278, 47 },
		{ 349, 279, 47 },
		{ 350, 258, 47 },
		{ 350, 259, 47 },
		{ 350, 260, 47 },
		{ 350, 261, 47 },
		{ 350, 262, 47 },
		{ 350, 263, 47 },
		{ 350, 264, 47 },
		{ 350, 265, 47 },
		{ 350, 266, 47 },
		{ 350, 267, 47 },
		{ 350, 268, 47 },
		{ 350, 269, 47 },
		{ 350, 270, 47 },
		{ 350, 271, 47 },
		{ 350, 272, 47 },
		{ 350, 273, 47 },
		{ 350, 274, 47 },
		{ 350, 275, 47 },
		{ 350, 276, 47 },
		{ 350, 277, 47 },
		{ 350, 278, 47 },
		{ 350, 279, 47 },
		{ 350, 280, 47 },
		{ 351, 258, 47 },
		{ 351, 259, 47 },
		{ 351, 260, 47 },
		{ 351, 261, 47 },
		{ 351, 262, 47 },
		{ 351, 263, 47 },
		{ 351, 264, 47 },
		{ 351, 265, 47 },
		{ 351, 266, 47 },
		{ 351, 267, 47 },
		{ 351, 268, 47 },
		{ 351, 269, 47 },
		{ 351, 270, 47 },
		{ 351, 271, 47 },
		{ 351, 272, 47 },
		{ 351, 273, 47 },
		{ 351, 274, 47 },
		{ 351, 275, 47 },
		{ 351, 276, 47 },
		{ 351, 277, 47 },
		{ 351, 278, 47 },
		{ 351, 279, 47 },
		{ 351, 280, 47 },
		{ 352, 258, 47 },
		{ 352, 259, 47 },
		{ 352, 260, 47 },
		{ 352, 261, 47 },
		{ 352, 262, 47 },
		{ 352, 263, 47 },
		{ 352, 264, 47 },
		{ 352, 265, 47 },
		{ 352, 266, 47 },
		{ 352, 267, 47 },
		{ 352, 268, 47 },
		{ 352, 269, 47 },
		{ 352, 270, 47 },
		{ 352, 271, 47 },
		{ 352, 272, 47 },
		{ 352, 273, 47 },
		{ 352, 274, 47 },
		{ 352, 275, 47 },
		{ 352, 276, 47 },
		{ 352, 277, 47 },
		{ 352, 278, 47 },
		{ 352, 279, 47 },
		{ 352, 280, 47 },
		{ 353, 258, 47 },
		{ 353, 259, 47 },
		{ 353, 260, 47 },
		{ 353, 261, 47 },
		{ 353, 262, 47 },
		{ 353, 263, 47 },
		{ 353, 264, 47 },
		{ 353, 265, 47 },
		{ 353, 266, 47 },
		{ 353, 267, 47 },
		{ 353, 268, 47 },
		{ 353, 269, 47 },
		{ 353, 270, 47 },
		{ 353, 271, 47 },
		{ 353, 272, 47 },
		{ 353, 273, 47 },
		{ 353, 274, 47 },
		{ 353, 275, 47 },
		{ 353, 276, 47 },
		{ 353, 277, 47 },
		{ 353, 278, 47 },
		{ 353, 279, 47 },
		{ 353, 280, 47 },
		{ 354, 258, 47 },
		{ 354, 259, 47 },
		{ 354, 260, 47 },
		{ 354, 261, 47 },
		{ 354, 262, 47 },
		{ 354, 263, 47 },
		{ 354, 264, 47 },
		{ 354, 265, 47 },
		{ 354, 266, 47 },
		{ 354, 267, 47 },
		{ 354, 268, 47 },
		{ 354, 269, 47 },
		{ 354, 270, 47 },
		{ 354, 271, 47 },
		{ 354, 272, 47 },
		{ 354, 273, 47 },
		{ 354, 274, 47 },
		{ 354, 275, 47 },
		{ 354, 276, 47 },
		{ 354, 277, 47 },
		{ 354, 278, 47 },
		{ 354, 279, 47 },
		{ 354, 280, 47 },
		{ 354, 281, 47 },
		{ 355, 258, 47 },
		{ 355, 259, 47 },
		{ 355, 260, 47 },
		{ 355, 261, 47 },
		{ 355, 262, 47 },
		{ 355, 263, 47 },
		{ 355, 264, 47 },
		{ 355, 265, 47 },
		{ 355, 266, 47 },
		{ 355, 267, 47 },
		{ 355, 268, 47 },
		{ 355, 269, 47 },
		{ 355, 270, 47 },
		{ 355, 271, 47 },
		{ 355, 272, 47 },
		{ 355, 273, 47 },
		{ 355, 274, 47 },
		{ 355, 275, 47 },
		{ 355, 276, 47 },
		{ 355, 277, 47 },
		{ 355, 278, 47 },
		{ 355, 279, 47 },
		{ 355, 280, 47 },
		{ 355, 281, 47 },
		{ 356, 258, 47 },
		{ 356, 259, 47 },
		{ 356, 260, 47 },
		{ 356, 261, 47 },
		{ 356, 262, 47 },
		{ 356, 263, 47 },
		{ 356, 264, 47 },
		{ 356, 265, 47 },
		{ 356, 266, 47 },
		{ 356, 267, 47 },
		{ 356, 268, 47 },
		{ 356, 269, 47 },
		{ 356, 270, 47 },
		{ 356, 271, 47 },
		{ 356, 272, 47 },
		{ 356, 273, 47 },
		{ 356, 274, 47 },
		{ 356, 275, 47 },
		{ 356, 276, 47 },
		{ 356, 277, 47 },
		{ 356, 278, 47 },
		{ 356, 279, 47 },
		{ 356, 280, 47 },
		{ 356, 281, 47 },
		{ 357, 258, 47 },
		{ 357, 259, 47 },
		{ 357, 260, 47 },
		{ 357, 261, 47 },
		{ 357, 262, 47 },
		{ 357, 263, 47 },
		{ 357, 264, 47 },
		{ 357, 265, 47 },
		{ 357, 266, 47 },
		{ 357, 267, 47 },
		{ 357, 268, 47 },
		{ 357, 269, 47 },
		{ 357, 270, 47 },
		{ 357, 271, 47 },
		{ 357, 272, 47 },
		{ 357, 273, 47 },
		{ 357, 274, 47 },
		{ 357, 275, 47 },
		{ 357, 276, 47 },
		{ 357, 277, 47 },
		{ 357, 278, 47 },
		{ 357, 279, 47 },
		{ 357, 280, 47 },
		{ 357, 281, 47 },
		{ 358, 258, 47 },
		{ 358, 259, 47 },
		{ 358, 260, 47 },
		{ 358, 261, 47 },
		{ 358, 262, 47 },
		{ 358, 263, 47 },
		{ 358, 264, 47 },
		{ 358, 265, 47 },
		{ 358, 266, 47 },
		{ 358, 267, 47 },
		{ 358, 268, 47 },
		{ 358, 269, 47 },
		{ 358, 270, 47 },
		{ 358, 271, 47 },
		{ 358, 272, 47 },
		{ 358, 273, 47 },
		{ 358, 274, 47 },
		{ 358, 275, 47 },
		{ 358, 276, 47 },
		{ 358, 277, 47 },
		{ 358, 278, 47 },
		{ 358, 279, 47 },
		{ 358, 280, 47 },
		{ 358, 281, 47 },
		{ 358, 282, 47 },
		{ 359, 258, 47 },
		{ 359, 259, 47 },
		{ 359, 260, 47 },
		{ 359, 261, 47 },
		{ 359, 262, 47 },
		{ 359, 263, 47 },
		{ 359, 264, 47 },
		{ 359, 265, 47 },
		{ 359, 266, 47 },
		{ 359, 267, 47 },
		{ 359, 268, 47 },
		{ 359, 269, 47 },
		{ 359, 270, 47 },
		{ 359, 271, 47 },
		{ 359, 272, 47 },
		{ 359, 273, 47 },
		{ 359, 274, 47 },
		{ 359, 275, 47 },
		{ 359, 276, 47 },
		{ 359, 277, 47 },
		{ 359, 278, 47 },
		{ 359, 279, 47 },
		{ 359, 280, 47 },
		{ 359, 281, 47 },
		{ 359, 282, 47 },
		{ 360, 258, 47 },
		{ 360, 259, 47 },
		{ 360, 260, 47 },
		{ 360, 261, 47 },
		{ 360, 262, 47 },
		{ 360, 263, 47 },
		{ 360, 264, 47 },
		{ 360, 265, 47 },
		{ 360, 266, 47 },
		{ 360, 267, 47 },
		{ 360, 268, 47 },
		{ 360, 269, 47 },
		{ 360, 270, 47 },
		{ 360, 271, 47 },
		{ 360, 272, 47 },
		{ 360, 273, 47 },
		{ 360, 274, 47 },
		{ 360, 275, 47 },
		{ 360, 276, 47 },
		{ 360, 277, 47 },
		{ 360, 278, 47 },
		{ 360, 279, 47 },
		{ 360, 280, 47 },
		{ 360, 281, 47 },
		{ 360, 282, 47 },
		{ 361, 258, 47 },
		{ 361, 259, 47 },
		{ 361, 260, 47 },
		{ 361, 261, 47 },
		{ 361, 262, 47 },
		{ 361, 263, 47 },
		{ 361, 264, 47 },
		{ 361, 265, 47 },
		{ 361, 266, 47 },
		{ 361, 267, 47 },
		{ 361, 268, 47 },
		{ 361, 269, 47 },
		{ 361, 270, 47 },
		{ 361, 271, 47 },
		{ 361, 272, 47 },
		{ 361, 273, 47 },
		{ 361, 274, 47 },
		{ 361, 275, 47 },
		{ 361, 276, 47 },
		{ 361, 277, 47 },
		{ 361, 278, 47 },
		{ 361, 279, 47 },
		{ 361, 280, 47 },
		{ 361, 281, 47 },
		{ 361, 282, 47 },
		{ 362, 258, 47 },
		{ 362, 259, 47 },
		{ 362, 260, 47 },
		{ 362, 261, 47 },
		{ 362, 262, 47 },
		{ 362, 263, 47 },
		{ 362, 264, 47 },
		{ 362, 265, 47 },
		{ 362, 266, 47 },
		{ 362, 267, 47 },
		{ 362, 268, 47 },
		{ 362, 269, 47 },
		{ 362, 270, 47 },
		{ 362, 271, 47 },
		{ 362, 272, 47 },
		{ 362, 273, 47 },
		{ 362, 274, 47 },
		{ 362, 275, 47 },
		{ 362, 276, 47 },
		{ 362, 277, 47 },
		{ 362, 278, 47 },
		{ 362, 279, 47 },
		{ 362, 280, 47 },
		{ 362, 281, 47 },
		{ 362, 282, 47 },
		{ 362, 283, 47 },
		{ 363, 258, 47 },
		{ 363, 259, 47 },
		{ 363, 260, 47 },
		{ 363, 261, 47 },
		{ 363, 262, 47 },
		{ 363, 263, 47 },
		{ 363, 264, 47 },
		{ 363, 265, 47 },
		{ 363, 266, 47 },
		{ 363, 267, 47 },
		{ 363, 268, 47 },
		{ 363, 269, 47 },
		{ 363, 270, 47 },
		{ 363, 271, 47 },
		{ 363, 272, 47 },
		{ 363, 273, 47 },
		{ 363, 274, 47 },
		{ 363, 275, 47 },
		{ 363, 276, 47 },
		{ 363, 277, 47 },
		{ 363, 278, 47 },
		{ 363, 279, 47 },
		{ 363, 280, 47 },
		{ 363, 281, 47 },
		{ 363, 282, 47 },
		{ 363, 283, 47 },
		{ 364, 258, 47 },
		{ 364, 259, 47 },
		{ 364, 260, 47 },
		{ 364, 261, 47 },
		{ 364, 262, 47 },
		{ 364, 263, 47 },
		{ 364, 264, 47 },
		{ 364, 265, 47 },
		{ 364, 266, 47 },
		{ 364, 267, 47 },
		{ 364, 268, 47 },
		{ 364, 269, 47 },
		{ 364, 270, 47 },
		{ 364, 271, 47 },
		{ 364, 272, 47 },
		{ 364, 273, 47 },
		{ 364, 274, 47 },
		{ 364, 275, 47 },
		{ 364, 276, 47 },
		{ 364, 277, 47 },
		{ 364, 278, 47 },
		{ 364, 279, 47 },
		{ 364, 280, 47 },
		{ 364, 281, 47 },
		{ 364, 282, 47 },
		{ 364, 283, 47 },
		{ 365, 258, 47 },
		{ 365, 259, 47 },
		{ 365, 260, 47 },
		{ 365, 261, 47 },
		{ 365, 262, 47 },
		{ 365, 263, 47 },
		{ 365, 264, 47 },
		{ 365, 265, 47 },
		{ 365, 266, 47 },
		{ 365, 267, 47 },
		{ 365, 268, 47 },
		{ 365, 269, 47 },
		{ 365, 270, 47 },
		{ 365, 271, 47 },
		{ 365, 272, 47 },
		{ 365, 273, 47 },
		{ 365, 274, 47 },
		{ 365, 275, 47 },
		{ 365, 276, 47 },
		{ 365, 277, 47 },
		{ 365, 278, 47 },
		{ 365, 279, 47 },
		{ 365, 280, 47 },
		{ 365, 281, 47 },
		{ 365, 282, 47 },
		{ 365, 283, 47 },
		{ 366, 258, 47 },
		{ 366, 259, 47 },
		{ 366, 260, 47 },
		{ 366, 261, 47 },
		{ 366, 262, 47 },
		{ 366, 263, 47 },
		{ 366, 264, 47 },
		{ 366, 265, 47 },
		{ 366, 266, 47 },
		{ 366, 267, 47 },
		{ 366, 268, 47 },
		{ 366, 269, 47 },
		{ 366, 270, 47 },
		{ 366, 271, 47 },
		{ 366, 272, 47 },
		{ 366, 273, 47 },
		{ 366, 274, 47 },
		{ 366, 275, 47 },
		{ 366, 276, 47 },
		{ 366, 277, 47 },
		{ 366, 278, 47 },
		{ 366, 279, 47 },
		{ 366, 280, 47 },
		{ 366, 281, 47 },
		{ 366, 282, 47 },
		{ 366, 283, 47 },
		{ 366, 284, 47 },
		{ 367, 258, 47 },
		{ 367, 259, 47 },
		{ 367, 260, 47 },
		{ 367, 261, 47 },
		{ 367, 262, 47 },
		{ 367, 263, 47 },
		{ 367, 264, 47 },
		{ 367, 265, 47 },
		{ 367, 266, 47 },
		{ 367, 267, 47 },
		{ 367, 268, 47 },
		{ 367, 269, 47 },
		{ 367, 270, 47 },
		{ 367, 271, 47 },
		{ 367, 272, 47 },
		{ 367, 273, 47 },
		{ 367, 274, 47 },
		{ 367, 275, 47 },
		{ 367, 276, 47 },
		{ 367, 277, 47 },
		{ 367, 278, 47 },
		{ 367, 279, 47 },
		{ 367, 280, 47 },
		{ 367, 281, 47 },
		{ 367, 282, 47 },
		{ 367, 283, 47 },
		{ 367, 284, 47 },
		{ 368, 258, 47 },
		{ 368, 259, 47 },
		{ 368, 260, 47 },
		{ 368, 261, 47 },
		{ 368, 262, 47 },
		{ 368, 263, 47 },
		{ 368, 264, 47 },
		{ 368, 265, 47 },
		{ 368, 266, 47 },
		{ 368, 267, 47 },
		{ 368, 268, 47 },
		{ 368, 269, 47 },
		{ 368, 270, 47 },
		{ 368, 271, 47 },
		{ 368, 272, 47 },
		{ 368, 273, 47 },
		{ 368, 274, 47 },
		{ 368, 275, 47 },
		{ 368, 276, 47 },
		{ 368, 277, 47 },
		{ 368, 278, 47 },
		{ 368, 279, 47 },
		{ 368, 280, 47 },
		{ 368, 281, 47 },
		{ 368, 282, 47 },
		{ 368, 283, 47 },
		{ 368, 284, 47 },
		{ 369, 258, 47 },
		{ 369, 259, 47 },
		{ 369, 260, 47 },
		{ 369, 261, 47 },
		{ 369, 262, 47 },
		{ 369, 263, 47 },
		{ 369, 264, 47 },
		{ 369, 265, 47 },
		{ 369, 266, 47 },
		{ 369, 267, 47 },
		{ 369, 268, 47 },
		{ 369, 269, 47 },
		{ 369, 270, 47 },
		{ 369, 271, 47 },
		{ 369, 272, 47 },
		{ 369, 273, 47 },
		{ 369, 274, 47 },
		{ 369, 275, 47 },
		{ 369, 276, 47 },
		{ 369, 277, 47 },
		{ 369, 278, 47 },
		{ 369, 279, 47 },
		{ 369, 280, 47 },
		{ 369, 281, 47 },
		{ 369, 282, 47 },
		{ 369, 283, 47 },
		{ 369, 284, 47 },
		{ 370, 258, 47 },
		{ 370, 259, 47 },
		{ 370, 260, 47 },
		{ 370, 261, 47 },
		{ 370, 262, 47 },
		{ 370, 263, 47 },
		{ 370, 264, 47 },
		{ 370, 265, 47 },
		{ 370, 266, 47 },
		{ 370, 267, 47 },
		{ 370, 268, 47 },
		{ 370, 269, 47 },
		{ 370, 270, 47 },
		{ 370, 271, 47 },
		{ 370, 272, 47 },
		{ 370, 273, 47 },
		{ 370, 274, 47 },
		{ 370, 275, 47 },
		{ 370, 276, 47 },
		{ 370, 277, 47 },
		{ 370, 278, 47 },
		{ 370, 279, 47 },
		{ 370, 280, 47 },
		{ 370, 281, 47 },
		{ 370, 282, 47 },
		{ 370, 283, 47 },
		{ 370, 284, 47 },
		{ 370, 285, 47 },
		{ 371, 258, 47 },
		{ 371, 259, 47 },
		{ 371, 260, 47 },
		{ 371, 261, 47 },
		{ 371, 262, 47 },
		{ 371, 263, 47 },
		{ 371, 264, 47 },
		{ 371, 265, 47 },
		{ 371, 266, 47 },
		{ 371, 267, 47 },
		{ 371, 268, 47 },
		{ 371, 269, 47 },
		{ 371, 270, 47 },
		{ 371, 271, 47 },
		{ 371, 272, 47 },
		{ 371, 273, 47 },
		{ 371, 274, 47 },
		{ 371, 275, 47 },
		{ 371, 276, 47 },
		{ 371, 277, 47 },
		{ 371, 278, 47 },
		{ 371, 279, 47 },
		{ 371, 280, 47 },
		{ 371, 281, 47 },
		{ 371, 282, 47 },
		{ 371, 283, 47 },
		{ 371, 284, 47 },
		{ 371, 285, 47 },
		{ 372, 258, 47 },
		{ 372, 259, 47 },
		{ 372, 260, 47 },
		{ 372, 261, 47 },
		{ 372, 262, 47 },
		{ 372, 263, 47 },
		{ 372, 264, 47 },
		{ 372, 265, 47 },
		{ 372, 266, 47 },
		{ 372, 267, 47 },
		{ 372, 268, 47 },
		{ 372, 269, 47 },
		{ 372, 270, 47 },
		{ 372, 271, 47 },
		{ 372, 272, 47 },
		{ 372, 273, 47 },
		{ 372, 274, 47 },
		{ 372, 275, 47 },
		{ 372, 276, 47 },
		{ 372, 277, 47 },
		{ 372, 278, 47 },
		{ 372, 279, 47 },
		{ 372, 280, 47 },
		{ 372, 281, 47 },
		{ 372, 282, 47 },
		{ 372, 283, 47 },
		{ 372, 284, 47 },
		{ 372, 285, 47 },
		{ 373, 258, 47 },
		{ 373, 259, 47 },
		{ 373, 260, 47 },
		{ 373, 261, 47 },
		{ 373, 262, 47 },
		{ 373, 263, 47 },
		{ 373, 264, 47 },
		{ 373, 265, 47 },
		{ 373, 266, 47 },
		{ 373, 267, 47 },
		{ 373, 268, 47 },
		{ 373, 269, 47 },
		{ 373, 270, 47 },
		{ 373, 271, 47 },
		{ 373, 272, 47 },
		{ 373, 273, 47 },
		{ 373, 274, 47 },
		{ 373, 275, 47 },
		{ 373, 276, 47 },
		{ 373, 277, 47 },
		{ 373, 278, 47 },
		{ 373, 279, 47 },
		{ 373, 280, 47 },
		{ 373, 281, 47 },
		{ 373, 282, 47 },
		{ 373, 283, 47 },
		{ 373, 284, 47 },
		{ 373, 285, 47 },
		{ 374, 258, 47 },
		{ 374, 259, 47 },
		{ 374, 260, 47 },
		{ 374, 261, 47 },
		{ 374, 262, 47 },
		{ 374, 263, 47 },
		{ 374, 264, 47 },
		{ 374, 265, 47 },
		{ 374, 266, 47 },
		{ 374, 267, 47 },
		{ 374, 268, 47 },
		{ 374, 269, 47 },
		{ 374, 270, 47 },
		{ 374, 271, 47 },
		{ 374, 272, 47 },
		{ 374, 273, 47 },
		{ 374, 274, 47 },
		{ 374, 275, 47 },
		{ 374, 276, 47 },
		{ 374, 277, 47 },
		{ 374, 278, 47 },
		{ 374, 279, 47 },
		{ 374, 280, 47 },
		{ 374, 281, 47 },
		{ 374, 282, 47 },
		{ 374, 283, 47 },
		{ 374, 284, 47 },
		{ 374, 285, 47 },
		{ 374, 286, 47 },
		{ 375, 258, 47 },
		{ 375, 259, 47 },
		{ 375, 260, 47 },
		{ 375, 261, 47 },
		{ 375, 262, 47 },
		{ 375, 263, 47 },
		{ 375, 264, 47 },
		{ 375, 265, 47 },
		{ 375, 266, 47 },
		{ 375, 267, 47 },
		{ 375, 268, 47 },
		{ 375, 269, 47 },
		{ 375, 270, 47 },
		{ 375, 271, 47 },
		{ 375, 272, 47 },
		{ 375, 273, 47 },
		{ 375, 274, 47 },
		{ 375, 275, 47 },
		{ 375, 276, 47 },
		{ 375, 277, 47 },
		{ 375, 278, 47 },
		{ 375, 279, 47 },
		{ 375, 280, 47 },
		{ 375, 281, 47 },
		{ 375, 282, 47 },
		{ 375, 283, 47 },
		{ 375, 284, 47 },
		{ 375, 285, 47 },
		{ 375, 286, 47 },
		{ 376, 258, 47 },
		{ 376, 259, 47 },
		{ 376, 260, 47 },
		{ 376, 261, 47 },
		{ 376, 262, 47 },
		{ 376, 263, 47 },
		{ 376, 264, 47 },
		{ 376, 265, 47 },
		{ 376, 266, 47 },
		{ 376, 267, 47 },
		{ 376, 268, 47 },
		{ 376, 269, 47 },
		{ 376, 270, 47 },
		{ 376, 271, 47 },
		{ 376, 272, 47 },
		{ 376, 273, 47 },
		{ 376, 274, 47 },
		{ 376, 275, 47 },
		{ 376, 276, 47 },
		{ 376, 277, 47 },
		{ 376, 278, 47 },
		{ 376, 279, 47 },
		{ 376, 280, 47 },
		{ 376, 281, 47 },
		{ 376, 282, 47 },
		{ 376, 283, 47 },
		{ 376, 284, 47 },
		{ 376, 285, 47 },
		{ 376, 286, 47 },
		{ 377, 258, 47 },
		{ 377, 259, 47 },
		{ 377, 260, 47 },
		{ 377, 261, 47 },
		{ 377, 262, 47 },
		{ 377, 263, 47 },
		{ 377, 264, 47 },
		{ 377, 265, 47 },
		{ 377, 266, 47 },
		{ 377, 267, 47 },
		{ 377, 268, 47 },
		{ 377, 269, 47 },
		{ 377, 270, 47 },
		{ 377, 271, 47 },
		{ 377, 272, 47 },
		{ 377, 273, 47 },
		{ 377, 274, 47 },
		{ 377, 275, 47 },
		{ 377, 276, 47 },
		{ 377, 277, 47 },
		{ 377, 278, 47 },
		{ 377, 279, 47 },
		{ 377, 280, 47 },
		{ 377, 281, 47 },
		{ 377, 282, 47 },
		{ 377, 283, 47 },
		{ 377, 284, 47 },
		{ 377, 285, 47 },
		{ 377, 286, 47 },
		{ 378, 258, 47 },
		{ 378, 259, 47 },
		{ 378, 260, 47 },
		{ 378, 261, 47 },
		{ 378, 262, 47 },
		{ 378, 263, 47 },
		{ 378, 264, 47 },
		{ 378, 265, 47 },
		{ 378, 266, 47 },
		{ 378, 267, 47 },
		{ 378, 268, 47 },
		{ 378, 269, 47 },
		{ 378, 270, 47 },
		{ 378, 271, 47 },
		{ 378, 272, 47 },
		{ 378, 273, 47 },
		{ 378, 274, 47 },
		{ 378, 275, 47 },
		{ 378, 276, 47 },
		{ 378, 277, 47 },
		{ 378, 278, 47 },
		{ 378, 279, 47 },
		{ 378, 280, 47 },
		{ 378, 281, 47 },
		{ 378, 282, 47 },
		{ 378, 283, 47 },
		{ 378, 284, 47 },
		{ 378, 285, 47 },
		{ 378, 286, 47 },
		{ 378, 287, 47 },
		{ 379, 258, 47 },
		{ 379, 259, 47 },
		{ 379, 260, 47 },
		{ 379, 261, 47 },
		{ 379, 262, 47 },
		{ 379, 263, 47 },
		{ 379, 264, 47 },
		{ 379, 265, 47 },
		{ 379, 266, 47 },
		{ 379, 267, 47 },
		{ 379, 268, 47 },
		{ 379, 269, 47 },
		{ 379, 270, 47 },
		{ 379, 271, 47 },
		{ 379, 272, 47 },
		{ 379, 273, 47 },
		{ 379, 274, 47 },
		{ 379, 275, 47 },
		{ 379, 276, 47 },
		{ 379, 277, 47 },
		{ 379, 278, 47 },
		{ 379, 279, 47 },
		{ 379, 280, 47 },
		{ 379, 281, 47 },
		{ 379, 282, 47 },
		{ 379, 283, 47 },
		{ 379, 284, 47 },
		{ 379, 285, 47 },
		{ 379, 286, 47 },
		{ 379, 287, 47 },
		{ 380, 258, 47 },
		{ 380, 259, 47 },
		{ 380, 260, 47 },
		{ 380, 261, 47 },
		{ 380, 262, 47 },
		{ 380, 263, 47 },
		{ 380, 264, 47 },
		{ 380, 265, 47 },
		{ 380, 266, 47 },
		{ 380, 267, 47 },
		{ 380, 268, 47 },
		{ 380, 269, 47 },
		{ 380, 270, 47 },
		{ 380, 271, 47 },
		{ 380, 272, 47 },
		{ 380, 273, 47 },
		{ 380, 274, 47 },
		{ 380, 275, 47 },
		{ 380, 276, 47 },
		{ 380, 277, 47 },
		{ 380, 278, 47 },
		{ 380, 279, 47 },
		{ 380, 280, 47 },
		{ 380, 281, 47 },
		{ 380, 282, 47 },
		{ 380, 283, 47 },
		{ 380, 284, 47 },
		{ 380, 285, 47 },
		{ 380, 286, 47 },
		{ 380, 287, 47 },
		{ 381, 258, 47 },
		{ 381, 259, 47 },
		{ 381, 260, 47 },
		{ 381, 261, 47 },
		{ 381, 262, 47 },
		{ 381, 263, 47 },
		{ 381, 264, 47 },
		{ 381, 265, 47 },
		{ 381, 266, 47 },
		{ 381, 267, 47 },
		{ 381, 268, 47 },
		{ 381, 269, 47 },
		{ 381, 270, 47 },
		{ 381, 271, 47 },
		{ 381, 272, 47 },
		{ 381, 273, 47 },
		{ 381, 274, 47 },
		{ 381, 275, 47 },
		{ 381, 276, 47 },
		{ 381, 277, 47 },
		{ 381, 278, 47 },
		{ 381, 279, 47 },
		{ 381, 280, 47 },
		{ 381, 281, 47 },
		{ 381, 282, 47 },
		{ 381, 283, 47 },
		{ 381, 284, 47 },
		{ 381, 285, 47 },
		{ 381, 286, 47 },
		{ 381, 287, 47 },
		{ 382, 258, 47 },
		{ 382, 259, 47 },
		{ 382, 260, 47 },
		{ 382, 261, 47 },
		{ 382, 262, 47 },
		{ 382, 263, 47 },
		{ 382, 264, 47 },
		{ 382, 265, 47 },
		{ 382, 266, 47 },
		{ 382, 267, 47 },
		{ 382, 268, 47 },
		{ 382, 269, 47 },
		{ 382, 270, 47 },
		{ 382, 271, 47 },
		{ 382, 272, 47 },
		{ 382, 273, 47 },
		{ 382, 274, 47 },
		{ 382, 275, 47 },
		{ 382, 276, 47 },
		{ 382, 277, 47 },
		{ 382, 278, 47 },
		{ 382, 279, 47 },
		{ 382, 280, 47 },
		{ 382, 281, 47 },
		{ 382, 282, 47 },
		{ 382, 283, 47 },
		{ 382, 284, 47 },
		{ 382, 285, 47 },
		{ 382, 286, 47 },
		{ 382, 287, 47 },
		{ 382, 288, 47 },
		{ 383, 258, 47 },
		{ 383, 259, 47 },
		{ 383, 260, 47 },
		{ 383, 261, 47 },
		{ 383, 262, 47 },
		{ 383, 263, 47 },
		{ 383, 264, 47 },
		{ 383, 265, 47 },
		{ 383, 266, 47 },
		{ 383, 267, 47 },
		{ 383, 268, 47 },
		{ 383, 269, 47 },
		{ 383, 270, 47 },
		{ 383, 271, 47 },
		{ 383, 272, 47 },
		{ 383, 273, 47 },
		{ 383, 274, 47 },
		{ 383, 275, 47 },
		{ 383, 276, 47 },
		{ 383, 277, 47 },
		{ 383, 278, 47 },
		{ 383, 279, 47 },
		{ 383, 280, 47 },
		{ 383, 281, 47 },
		{ 383, 282, 47 },
		{ 383, 283, 47 },
		{ 383, 284, 47 },
		{ 383, 285, 47 },
		{ 383, 286, 47 },
		{ 383, 287, 47 },
		{ 383, 288, 47 },
		{ 384, 258, 47 },
		{ 384, 259, 47 },
		{ 384, 260, 47 },
		{ 384, 261, 47 },
		{ 384, 262, 47 },
		{ 384, 263, 47 },
		{ 384, 264, 47 },
		{ 384, 265, 47 },
		{ 384, 266, 47 },
		{ 384, 267, 47 },
		{ 384, 268, 47 },
		{ 384, 269, 47 },
		{ 384, 270, 47 },
		{ 384, 271, 47 },
		{ 384, 272, 47 },
		{ 384, 273, 47 },
		{ 384, 274, 47 },
		{ 384, 275, 47 },
		{ 384, 276, 47 },
		{ 384, 277, 47 },
		{ 384, 278, 47 },
		{ 384, 279, 47 },
		{ 384, 280, 47 },
		{ 384, 281, 47 },
		{ 384, 282, 47 },
		{ 384, 283, 47 },
		{ 384, 284, 47 },
		{ 384, 285, 47 },
		{ 384, 286, 47 },
		{ 384, 287, 47 },
		{ 384, 288, 47 },
		{ 385, 258, 47 },
		{ 385, 259, 47 },
		{ 385, 260, 47 },
		{ 385, 261, 47 },
		{ 385, 262, 47 },
		{ 385, 263, 47 },
		{ 385, 264, 47 },
		{ 385, 265, 47 },
		{ 385, 266, 47 },
		{ 385, 267, 47 },
		{ 385, 268, 47 },
		{ 385, 269, 47 },
		{ 385, 270, 47 },
		{ 385, 271, 47 },
		{ 385, 272, 47 },
		{ 385, 273, 47 },
		{ 385, 274, 47 },
		{ 385, 275, 47 },
		{ 385, 276, 47 },
		{ 385, 277, 47 },
		{ 385, 278, 47 },
		{ 385, 279, 47 },
		{ 385, 280, 47 },
		{ 385, 281, 47 },
		{ 385, 282, 47 },
		{ 385, 283, 47 },
		{ 385, 284, 47 },
		{ 385, 285, 47 },
		{ 385, 286, 47 },
		{ 385, 287, 47 },
		{ 385, 288, 47 },
		{ 386, 258, 47 },
		{ 386, 259, 47 },
		{ 386, 260, 47 },
		{ 386, 261, 47 },
		{ 386, 262, 47 },
		{ 386, 263, 47 },
		{ 386, 264, 47 },
		{ 386, 265, 47 },
		{ 386, 266, 47 },
		{ 386, 267, 47 },
		{ 386, 268, 47 },
		{ 386, 269, 47 },
		{ 386, 270, 47 },
		{ 386, 271, 47 },
		{ 386, 272, 47 },
		{ 386, 273, 47 },
		{ 386, 274, 47 },
		{ 386, 275, 47 },
		{ 386, 276, 47 },
		{ 386, 277, 47 },
		{ 386, 278, 47 },
		{ 386, 279, 47 },
		{ 386, 280, 47 },
		{ 386, 281, 47 },
		{ 386, 282, 47 },
		{ 386, 283, 47 },
		{ 386, 284, 47 },
		{ 386, 285, 47 },
		{ 386, 286, 47 },
		{ 386, 287, 47 },
		{ 386, 288, 47 },
		{ 386, 289, 47 },
		{ 387, 258, 47 },
		{ 387, 259, 47 },
		{ 387, 260, 47 },
		{ 387, 261, 47 },
		{ 387, 262, 47 },
		{ 387, 263, 47 },
		{ 387, 264, 47 },
		{ 387, 265, 47 },
		{ 387, 266, 47 },
		{ 387, 267, 47 },
		{ 387, 268, 47 },
		{ 387, 269, 47 },
		{ 387, 270, 47 },
		{ 387, 271, 47 },
		{ 387, 272, 47 },
		{ 387, 273, 47 },
		{ 387, 274, 47 },
		{ 387, 275, 47 },
		{ 387, 276, 47 },
		{ 387, 277, 47 },
		{ 387, 278, 47 },
		{ 387, 279, 47 },
		{ 387, 280, 47 },
		{ 387, 281, 47 },
		{ 387, 282, 47 },
		{ 387, 283, 47 },
		{ 387, 284, 47 },
		{ 387, 285, 47 },
		{ 387, 286, 47 },
		{ 387, 287, 47 },
		{ 387, 288, 47 },
		{ 387, 289, 47 },
		{ 388, 258, 47 },
		{ 388, 259, 47 },
		{ 388, 260, 47 },
		{ 388, 261, 47 },
		{ 388, 262, 47 },
		{ 388, 263, 47 },
		{ 388, 264, 47 },
		{ 388, 265, 47 },
		{ 388, 266, 47 },
		{ 388, 267, 47 },
		{ 388, 268, 47 },
		{ 388, 269, 47 },
		{ 388, 270, 47 },
		{ 388, 271, 47 },
		{ 388, 272, 47 },
		{ 388, 273, 47 },
		{ 388, 274, 47 },
		{ 388, 275, 47 },
		{ 388, 276, 47 },
		{ 388, 277, 47 },
		{ 388, 278, 47 },
		{ 388, 279, 47 },
		{ 388, 280, 47 },
		{ 388, 281, 47 },
		{ 388, 282, 47 },
		{ 388, 283, 47 },
		{ 388, 284, 47 },
		{ 388, 285, 47 },
		{ 388, 286, 47 },
		{ 388, 287, 47 },
		{ 388, 288, 47 },
		{ 388, 289, 47 },
		{ 389, 258, 47 },
		{ 389, 259, 47 },
		{ 389, 260, 47 },
		{ 389, 261, 47 },
		{ 389, 262, 47 },
		{ 389, 263, 47 },
		{ 389, 264, 47 },
		{ 389, 265, 47 },
		{ 389, 266, 47 },
		{ 389, 267, 47 },
		{ 389, 268, 47 },
		{ 389, 269, 47 },
		{ 389, 270, 47 },
		{ 389, 271, 47 },
		{ 389, 272, 47 },
		{ 389, 273, 47 },
		{ 389, 274, 47 },
		{ 389, 275, 47 },
		{ 389, 276, 47 },
		{ 389, 277, 47 },
		{ 389, 278, 47 },
		{ 389, 279, 47 },
		{ 389, 280, 47 },
		{ 389, 281, 47 },
		{ 389, 282, 47 },
		{ 389, 283, 47 },
		{ 389, 284, 47 },
		{ 389, 285, 47 },
		{ 389, 286, 47 },
		{ 389, 287, 47 },
		{ 389, 288, 47 },
		{ 389, 289, 47 },
		{ 390, 258, 47 },
		{ 390, 259, 47 },
		{ 390, 260, 47 },
		{ 390, 261, 47 },
		{ 390, 262, 47 },
		{ 390, 263, 47 },
		{ 390, 264, 47 },
		{ 390, 265, 47 },
		{ 390, 266, 47 },
		{ 390, 267, 47 },
		{ 390, 268, 47 },
		{ 390, 269, 47 },
		{ 390, 270, 47 },
		{ 390, 271, 47 },
		{ 390, 272, 47 },
		{ 390, 273, 47 },
		{ 390, 274, 47 },
		{ 390, 275, 47 },
		{ 390, 276, 47 },
		{ 390, 277, 47 },
		{ 390, 278, 47 },
		{ 390, 279, 47 },
		{ 390, 280, 47 },
		{ 390, 281, 47 },
		{ 390, 282, 47 },
		{ 390, 283, 47 },
		{ 390, 284, 47 },
		{ 390, 285, 47 },
		{ 390, 286, 47 },
		{ 390, 287, 47 },
		{ 390, 288, 47 },
		{ 390, 289, 47 },
		{ 390, 290, 47 },
		{ 391, 258, 47 },
		{ 391, 259, 47 },
		{ 391, 260, 47 },
		{ 391, 261, 47 },
		{ 391, 262, 47 },
		{ 391, 263, 47 },
		{ 391, 264, 47 },
		{ 391, 265, 47 },
		{ 391, 266, 47 },
		{ 391, 267, 47 },
		{ 391, 268, 47 },
		{ 391, 269, 47 },
		{ 391, 270, 47 },
		{ 391, 271, 47 },
		{ 391, 272, 47 },
		{ 391, 273, 47 },
		{ 391, 274, 47 },
		{ 391, 275, 47 },
		{ 391, 276, 47 },
		{ 391, 277, 47 },
		{ 391, 278, 47 },
		{ 391, 279, 47 },
		{ 391, 280, 47 },
		{ 391, 281, 47 },
		{ 391, 282, 47 },
		{ 391, 283, 47 },
		{ 391, 284, 47 },
		{ 391, 285, 47 },
		{ 391, 286, 47 },
		{ 391, 287, 47 },
		{ 391, 288, 47 },
		{ 391, 289, 47 },
		{ 391, 290, 47 },
		{ 392, 258, 47 },
		{ 392, 259, 47 },
		{ 392, 260, 47 },
		{ 392, 261, 47 },
		{ 392, 262, 47 },
		{ 392, 263, 47 },
		{ 392, 264, 47 },
		{ 392, 265, 47 },
		{ 392, 266, 47 },
		{ 392, 267, 47 },
		{ 392, 268, 47 },
		{ 392, 269, 47 },
		{ 392, 270, 47 },
		{ 392, 271, 47 },
		{ 392, 272, 47 },
		{ 392, 273, 47 },
		{ 392, 274, 47 },
		{ 392, 275, 47 },
		{ 392, 276, 47 },
		{ 392, 277, 47 },
		{ 392, 278, 47 },
		{ 392, 279, 47 },
		{ 392, 280, 47 },
		{ 392, 281, 47 },
		{ 392, 282, 47 },
		{ 392, 283, 47 },
		{ 392, 284, 47 },
		{ 392, 285, 47 },
		{ 392, 286, 47 },
		{ 392, 287, 47 },
		{ 392, 288, 47 },
		{ 392, 289, 47 },
		{ 392, 290, 47 },
		{ 393, 258, 47 },
		{ 393, 259, 47 },
		{ 393, 260, 47 },
		{ 393, 261, 47 },
		{ 393, 262, 47 },
		{ 393, 263, 47 },
		{ 393, 264, 47 },
		{ 393, 265, 47 },
		{ 393, 266, 47 },
		{ 393, 267, 47 },
		{ 393, 268, 47 },
		{ 393, 269, 47 },
		{ 393, 270, 47 },
		{ 393, 271, 47 },
		{ 393, 272, 47 },
		{ 393, 273, 47 },
		{ 393, 274, 47 },
		{ 393, 275, 47 },
		{ 393, 276, 47 },
		{ 393, 277, 47 },
		{ 393, 278, 47 },
		{ 393, 279, 47 },
		{ 393, 280, 47 },
		{ 393, 281, 47 },
		{ 393, 282, 47 },
		{ 393, 283, 47 },
		{ 393, 284, 47 },
		{ 393, 285, 47 },
		{ 393, 286, 47 },
		{ 393, 287, 47 },
		{ 393, 288, 47 },
		{ 393, 289, 47 },
		{ 393, 290, 47 },
		{ 394, 258, 47 },
		{ 394, 259, 47 },
		{ 394, 260, 47 },
		{ 394, 261, 47 },
		{ 394, 262, 47 },
		{ 394, 263, 47 },
		{ 394, 264, 47 },
		{ 394, 265, 47 },
		{ 394, 266, 47 },
		{ 394, 267, 47 },
		{ 394, 268, 47 },
		{ 394, 269, 47 },
		{ 394, 270, 47 },
		{ 394, 271, 47 },
		{ 394, 272, 47 },
		{ 394, 273, 47 },
		{ 394, 274, 47 },
		{ 394, 275, 47 },
		{ 394, 276, 47 },
		{ 394, 277, 47 },
		{ 394, 278, 47 },
		{ 394, 279, 47 },
		{ 394, 280, 47 },
		{ 394, 281, 47 },
		{ 394, 282, 47 },
		{ 394, 283, 47 },
		{ 394, 284, 47 },
		{ 394, 285, 47 },
		{ 394, 286, 47 },
		{ 394, 287, 47 },
		{ 394, 288, 47 },
		{ 394, 289, 47 },
		{ 394, 290, 47 },
		{ 394, 291, 47 },
		{ 395, 258, 47 },
		{ 395, 259, 47 },
		{ 395, 260, 47 },
		{ 395, 261, 47 },
		{ 395, 262, 47 },
		{ 395, 263, 47 },
		{ 395, 264, 47 },
		{ 395, 265, 47 },
		{ 395, 266, 47 },
		{ 395, 267, 47 },
		{ 395, 268, 47 },
		{ 395, 269, 47 },
		{ 395, 270, 47 },
		{ 395, 271, 47 },
		{ 395, 272, 47 },
		{ 395, 273, 47 },
		{ 395, 274, 47 },
		{ 395, 275, 47 },
		{ 395, 276, 47 },
		{ 395, 277, 47 },
		{ 395, 278, 47 },
		{ 395, 279, 47 },
		{ 395, 280, 47 },
		{ 395, 281, 47 },
		{ 395, 282, 47 },
		{ 395, 283, 47 },
		{ 395, 284, 47 },
		{ 395, 285, 47 },
		{ 395, 286, 47 },
		{ 395, 287, 47 },
		{ 395, 288, 47 },
		{ 395, 289, 47 },
		{ 395, 290, 47 },
		{ 395, 291, 47 },
		{ 396, 258, 47 },
		{ 396, 259, 47 },
		{ 396, 260, 47 },
		{ 396, 261, 47 },
		{ 396, 262, 47 },
		{ 396, 263, 47 },
		{ 396, 264, 47 },
		{ 396, 265, 47 },
		{ 396, 266, 47 },
		{ 396, 267, 47 },
		{ 396, 268, 47 },
		{ 396, 269, 47 },
		{ 396, 270, 47 },
		{ 396, 271, 47 },
		{ 396, 272, 47 },
		{ 396, 273, 47 },
		{ 396, 274, 47 },
		{ 396, 275, 47 },
		{ 396, 276, 47 },
		{ 396, 277, 47 },
		{ 396, 278, 47 },
		{ 396, 279, 47 },
		{ 396, 280, 47 },
		{ 396, 281, 47 },
		{ 396, 282, 47 },
		{ 396, 283, 47 },
		{ 396, 284, 47 },
		{ 396, 285, 47 },
		{ 396, 286, 47 },
		{ 396, 287, 47 },
		{ 396, 288, 47 },
		{ 396, 289, 47 },
		{ 396, 290, 47 },
		{ 396, 291, 47 },
		{ 397, 258, 47 },
		{ 397, 259, 47 },
		{ 397, 260, 47 },
		{ 397, 261, 47 },
		{ 397, 262, 47 },
		{ 397, 263, 47 },
		{ 397, 264, 47 },
		{ 397, 265, 47 },
		{ 397, 266, 47 },
		{ 397, 267, 47 },
		{ 397, 268, 47 },
		{ 397, 269, 47 },
		{ 397, 270, 47 },
		{ 397, 271, 47 },
		{ 397, 272, 47 },
		{ 397, 273, 47 },
		{ 397, 274, 47 },
		{ 397, 275, 47 },
		{ 397, 276, 47 },
		{ 397, 277, 47 },
		{ 397, 278, 47 },
		{ 397, 279, 47 },
		{ 397, 280, 47 },
		{ 397, 281, 47 },
		{ 397, 282, 47 },
		{ 397, 283, 47 },
		{ 397, 284, 47 },
		{ 397, 285, 47 },
		{ 397, 286, 47 },
		{ 397, 287, 47 },
		{ 397, 288, 47 },
		{ 397, 289, 47 },
		{ 397, 290, 47 },
		{ 397, 291, 47 },
		{ 398, 258, 47 },
		{ 398, 259, 47 },
		{ 398, 260, 47 },
		{ 398, 261, 47 },
		{ 398, 262, 47 },
		{ 398, 263, 47 },
		{ 398, 264, 47 },
		{ 398, 265, 47 },
		{ 398, 266, 47 },
		{ 398, 267, 47 },
		{ 398, 268, 47 },
		{ 398, 269, 47 },
		{ 398, 270, 47 },
		{ 398, 271, 47 },
		{ 398, 272, 47 },
		{ 398, 273, 47 },
		{ 398, 274, 47 },
		{ 398, 275, 47 },
		{ 398, 276, 47 },
		{ 398, 277, 47 },
		{ 398, 278, 47 },
		{ 398, 279, 47 },
		{ 398, 280, 47 },
		{ 398, 281, 47 },
		{ 398, 282, 47 },
		{ 398, 283, 47 },
		{ 398, 284, 47 },
		{ 398, 285, 47 },
		{ 398, 286, 47 },
		{ 398, 287, 47 },
		{ 398, 288, 47 },
		{ 398, 289, 47 },
		{ 398, 290, 47 },
		{ 398, 291, 47 },
		{ 398, 292, 47 },
		{ 399, 258, 47 },
		{ 399, 259, 47 },
		{ 399, 260, 47 },
		{ 399, 261, 47 },
		{ 399, 262, 47 },
		{ 399, 263, 47 },
		{ 399, 264, 47 },
		{ 399, 265, 47 },
		{ 399, 266, 47 },
		{ 399, 267, 47 },
		{ 399, 268, 47 },
		{ 399, 269, 47 },
		{ 399, 270, 47 },
		{ 399, 271, 47 },
		{ 399, 272, 47 },
		{ 399, 273, 47 },
		{ 399, 274, 47 },
		{ 399, 275, 47 },
		{ 399, 276, 47 },
		{ 399, 277, 47 },
		{ 399, 278, 47 },
		{ 399, 279, 47 },
		{ 399, 280, 47 },
		{ 399, 281, 47 },
		{ 399, 282, 47 },
		{ 399, 283, 47 },
		{ 399, 284, 47 },
		{ 399, 285, 47 },
		{ 399, 286, 47 },
		{ 399, 287, 47 },
		{ 399, 288, 47 },
		{ 399, 289, 47 },
		{ 399, 290, 47 },
		{ 399, 291, 47 },
		{ 399, 292, 47 },
		{ 400, 258, 47 },
		{ 400, 259, 47 },
		{ 400, 260, 47 },
		{ 400, 261, 47 },
		{ 400, 262, 47 },
		{ 400, 263, 47 },
		{ 400, 264, 47 },
		{ 400, 265, 47 },
		{ 400, 266, 47 },
		{ 400, 267, 47 },
		{ 400, 268, 47 },
		{ 400, 269, 47 },
		{ 400, 270, 47 },
		{ 400, 271, 47 },
		{ 400, 272, 47 },
		{ 400, 273, 47 },
		{ 400, 274, 47 },
		{ 400, 275, 47 },
		{ 400, 276, 47 },
		{ 400, 277, 47 },
		{ 400, 278, 47 },
		{ 400, 279, 47 },
		{ 400, 280, 47 },
		{ 400, 281, 47 },
		{ 400, 282, 47 },
		{ 400, 283, 47 },
		{ 400, 284, 47 },
		{ 400, 285, 47 },
		{ 400, 286, 47 },
		{ 400, 287, 47 },
		{ 400, 288, 47 },
		{ 400, 289, 47 },
		{ 400, 290, 47 },
		{ 400, 291, 47 },
		{ 400, 292, 47 },
		{ 401, 258, 47 },
		{ 401, 259, 47 },
		{ 401, 260, 47 },
		{ 401, 261, 47 },
		{ 401, 262, 47 },
		{ 401, 263, 47 },
		{ 401, 264, 47 },
		{ 401, 265, 47 },
		{ 401, 266, 47 },
		{ 401, 267, 47 },
		{ 401, 268, 47 },
		{ 401, 269, 47 },
		{ 401, 270, 47 },
		{ 401, 271, 47 },
		{ 401, 272, 47 },
		{ 401, 273, 47 },
		{ 401, 274, 47 },
		{ 401, 275, 47 },
		{ 401, 276, 47 },
		{ 401, 277, 47 },
		{ 401, 278, 47 },
		{ 401, 279, 47 },
		{ 401, 280, 47 },
		{ 401, 281, 47 },
		{ 401, 282, 47 },
		{ 401, 283, 47 },
		{ 401, 284, 47 },
		{ 401, 285, 47 },
		{ 401, 286, 47 },
		{ 401, 287, 47 },
		{ 401, 288, 47 },
		{ 401, 289, 47 },
		{ 401, 290, 47 },
		{ 401, 291, 47 },
		{ 401, 292, 47 },
		{ 402, 258, 47 },
		{ 402, 259, 47 },
		{ 402, 260, 47 },
		{ 402, 261, 47 },
		{ 402, 262, 47 },
		{ 402, 263, 47 },
		{ 402, 264, 47 },
		{ 402, 265, 47 },
		{ 402, 266, 47 },
		{ 402, 267, 47 },
		{ 402, 268, 47 },
		{ 402, 269, 47 },
		{ 402, 270, 47 },
		{ 402, 271, 47 },
		{ 402, 272, 47 },
		{ 402, 273, 47 },
		{ 402, 274, 47 },
		{ 402, 275, 47 },
		{ 402, 276, 47 },
		{ 402, 277, 47 },
		{ 402, 278, 47 },
		{ 402, 279, 47 },
		{ 402, 280, 47 },
		{ 402, 281, 47 },
		{ 402, 282, 47 },
		{ 402, 283, 47 },
		{ 402, 284, 47 },
		{ 402, 285, 47 },
		{ 402, 286, 47 },
		{ 402, 287, 47 },
		{ 402, 288, 47 },
		{ 402, 289, 47 },
		{ 402, 290, 47 },
		{ 402, 291, 47 },
		{ 402, 292, 47 },
		{ 402, 293, 47 },
		{ 403, 258, 47 },
		{ 403, 259, 47 },
		{ 403, 260, 47 },
		{ 403, 261, 47 },
		{ 403, 262, 47 },
		{ 403, 263, 47 },
		{ 403, 264, 47 },
		{ 403, 265, 47 },
		{ 403, 266, 47 },
		{ 403, 267, 47 },
		{ 403, 268, 47 },
		{ 403, 269, 47 },
		{ 403, 270, 47 },
		{ 403, 271, 47 },
		{ 403, 272, 47 },
		{ 403, 273, 47 },
		{ 403, 274, 47 },
		{ 403, 275, 47 },
		{ 403, 276, 47 },
		{ 403, 277, 47 },
		{ 403, 278, 47 },
		{ 403, 279, 47 },
		{ 403, 280, 47 },
		{ 403, 281, 47 },
		{ 403, 282, 47 },
		{ 403, 283, 47 },
		{ 403, 284, 47 },
		{ 403, 285, 47 },
		{ 403, 286, 47 },
		{ 403, 287, 47 },
		{ 403, 288, 47 },
		{ 403, 289, 47 },
		{ 403, 290, 47 },
		{ 403, 291, 47 },
		{ 403, 292, 47 },
		{ 403, 293, 47 },
		{ 404, 258, 47 },
		{ 404, 259, 47 },
		{ 404, 260, 47 },
		{ 404, 261, 47 },
		{ 404, 262, 47 },
		{ 404, 263, 47 },
		{ 404, 264, 47 },
		{ 404, 265, 47 },
		{ 404, 266, 47 },
		{ 404, 267, 47 },
		{ 404, 268, 47 },
		{ 404, 269, 47 },
		{ 404, 270, 47 },
		{ 404, 271, 47 },
		{ 404, 272, 47 },
		{ 404, 273, 47 },
		{ 404, 274, 47 },
		{ 404, 275, 47 },
		{ 404, 276, 47 },
		{ 404, 277, 47 },
		{ 404, 278, 47 },
		{ 404, 279, 47 },
		{ 404, 280, 47 },
		{ 404, 281, 47 },
		{ 404, 282, 47 },
		{ 404, 283, 47 },
		{ 404, 284, 47 },
		{ 404, 285, 47 },
		{ 404, 286, 47 },
		{ 404, 287, 47 },
		{ 404, 288, 47 },
		{ 404, 289, 47 },
		{ 404, 290, 47 },
		{ 404, 291, 47 },
		{ 404, 292, 47 },
		{ 404, 293, 47 },
		{ 405, 258, 47 },
		{ 405, 259, 47 },
		{ 405, 260, 47 },
		{ 405, 261, 47 },
		{ 405, 262, 47 },
		{ 405, 263, 47 },
		{ 405, 264, 47 },
		{ 405, 265, 47 },
		{ 405, 266, 47 },
		{ 405, 267, 47 },
		{ 405, 268, 47 },
		{ 405, 269, 47 },
		{ 405, 270, 47 },
		{ 405, 271, 47 },
		{ 405, 272, 47 },
		{ 405, 273, 47 },
		{ 405, 274, 47 },
		{ 405, 275, 47 },
		{ 405, 276, 47 },
		{ 405, 277, 47 },
		{ 405, 278, 47 },
		{ 405, 279, 47 },
		{ 405, 280, 47 },
		{ 405, 281, 47 },
		{ 405, 282, 47 },
		{ 405, 283, 47 },
		{ 405, 284, 47 },
		{ 405, 285, 47 },
		{ 405, 286, 47 },
		{ 405, 287, 47 },
		{ 405, 288, 47 },
		{ 405, 289, 47 },
		{ 405, 290, 47 },
		{ 405, 291, 47 },
		{ 405, 292, 47 },
		{ 405, 293, 47 },
		{ 406, 258, 47 },
		{ 406, 259, 47 },
		{ 406, 260, 47 },
		{ 406, 261, 47 },
		{ 406, 262, 47 },
		{ 406, 263, 47 },
		{ 406, 264, 47 },
		{ 406, 265, 47 },
		{ 406, 266, 47 },
		{ 406, 267, 47 },
		{ 406, 268, 47 },
		{ 406, 269, 47 },
		{ 406, 270, 47 },
		{ 406, 271, 47 },
		{ 406, 272, 47 },
		{ 406, 273, 47 },
		{ 406, 274, 47 },
		{ 406, 275, 47 },
		{ 406, 276, 47 },
		{ 406, 277, 47 },
		{ 406, 278, 47 },
		{ 406, 279, 47 },
		{ 406, 280, 47 },
		{ 406, 281, 47 },
		{ 406, 282, 47 },
		{ 406, 283, 47 },
		{ 406, 284, 47 },
		{ 406, 285, 47 },
		{ 406, 286, 47 },
		{ 406, 287, 47 },
		{ 406, 288, 47 },
		{ 406, 289, 47 },
		{ 406, 290, 47 },
		{ 406, 291, 47 },
		{ 406, 292, 47 },
		{ 406, 293, 47 },
		{ 406, 294, 47 },
		{ 407, 258, 47 },
		{ 407, 259, 47 },
		{ 407, 260, 47 },
		{ 407, 261, 47 },
		{ 407, 262, 47 },
		{ 407, 263, 47 },
		{ 407, 264, 47 },
		{ 407, 265, 47 },
		{ 407, 266, 47 },
		{ 407, 267, 47 },
		{ 407, 268, 47 },
		{ 407, 269, 47 },
		{ 407, 270, 47 },
		{ 407, 271, 47 },
		{ 407, 272, 47 },
		{ 407, 273, 47 },
		{ 407, 274, 47 },
		{ 407, 275, 47 },
		{ 407, 276, 47 },
		{ 407, 277, 47 },
		{ 407, 278, 47 },
		{ 407, 279, 47 },
		{ 407, 280, 47 },
		{ 407, 281, 47 },
		{ 407, 282, 47 },
		{ 407, 283, 47 },
		{ 407, 284, 47 },
		{ 407, 285, 47 },
		{ 407, 286, 47 },
		{ 407, 287, 47 },
		{ 407, 288, 47 },
		{ 407, 289, 47 },
		{ 407, 290, 47 },
		{ 407, 291, 47 },
		{ 407, 292, 47 },
		{ 407, 293, 47 },
		{ 407, 294, 47 },
		{ 408, 258, 47 },
		{ 408, 259, 47 },
		{ 408, 260, 47 },
		{ 408, 261, 47 },
		{ 408, 262, 47 },
		{ 408, 263, 47 },
		{ 408, 264, 47 },
		{ 408, 265, 47 },
		{ 408, 266, 47 },
		{ 408, 267, 47 },
		{ 408, 268, 47 },
		{ 408, 269, 47 },
		{ 408, 270, 47 },
		{ 408, 271, 47 },
		{ 408, 272, 47 },
		{ 408, 273, 47 },
		{ 408, 274, 47 },
		{ 408, 275, 47 },
		{ 408, 276, 47 },
		{ 408, 277, 47 },
		{ 408, 278, 47 },
		{ 408, 279, 47 },
		{ 408, 280, 47 },
		{ 408, 281, 47 },
		{ 408, 282, 47 },
		{ 408, 283, 47 },
		{ 408, 284, 47 },
		{ 408, 285, 47 },
		{ 408, 286, 47 },
		{ 408, 287, 47 },
		{ 408, 288, 47 },
		{ 408, 289, 47 },
		{ 408, 290, 47 },
		{ 408, 291, 47 },
		{ 408, 292, 47 },
		{ 408, 293, 47 },
		{ 408, 294, 47 },
		{ 409, 258, 47 },
		{ 409, 259, 47 },
		{ 409, 260, 47 },
		{ 409, 261, 47 },
		{ 409, 262, 47 },
		{ 409, 263, 47 },
		{ 409, 264, 47 },
		{ 409, 265, 47 },
		{ 409, 266, 47 },
		{ 409, 267, 47 },
		{ 409, 268, 47 },
		{ 409, 269, 47 },
		{ 409, 270, 47 },
		{ 409, 271, 47 },
		{ 409, 272, 47 },
		{ 409, 273, 47 },
		{ 409, 274, 47 },
		{ 409, 275, 47 },
		{ 409, 276, 47 },
		{ 409, 277, 47 },
		{ 409, 278, 47 },
		{ 409, 279, 47 },
		{ 409, 280, 47 },
		{ 409, 281, 47 },
		{ 409, 282, 47 },
		{ 409, 283, 47 },
		{ 409, 284, 47 },
		{ 409, 285, 47 },
		{ 409, 286, 47 },
		{ 409, 287, 47 },
		{ 409, 288, 47 },
		{ 409, 289, 47 },
		{ 409, 290, 47 },
		{ 409, 291, 47 },
		{ 409, 292, 47 },
		{ 409, 293, 47 },
		{ 409, 294, 47 },
		{ 410, 258, 47 },
		{ 410, 259, 47 },
		{ 410, 260, 47 },
		{ 410, 261, 47 },
		{ 410, 262, 47 },
		{ 410, 263, 47 },
		{ 410, 264, 47 },
		{ 410, 265, 47 },
		{ 410, 266, 47 },
		{ 410, 267, 47 },
		{ 410, 268, 47 },
		{ 410, 269, 47 },
		{ 410, 270, 47 },
		{ 410, 271, 47 },
		{ 410, 272, 47 },
		{ 410, 273, 47 },
		{ 410, 274, 47 },
		{ 410, 275, 47 },
		{ 410, 276, 47 },
		{ 410, 277, 47 },
		{ 410, 278, 47 },
		{ 410, 279, 47 },
		{ 410, 280, 47 },
		{ 410, 281, 47 },
		{ 410, 282, 47 },
		{ 410, 283, 47 },
		{ 410, 284, 47 },
		{ 410, 285, 47 },
		{ 410, 286, 47 },
		{ 410, 287, 47 },
		{ 410, 288, 47 },
		{ 410, 289, 47 },
		{ 410, 290, 47 },
		{ 410, 291, 47 },
		{ 410, 292, 47 },
		{ 410, 293, 47 },
		{ 410, 294, 47 },
		{ 410, 295, 47 },
		{ 411, 258, 47 },
		{ 411, 259, 47 },
		{ 411, 260, 47 },
		{ 411, 261, 47 },
		{ 411, 262, 47 },
		{ 411, 263, 47 },
		{ 411, 264, 47 },
		{ 411, 265, 47 },
		{ 411, 266, 47 },
		{ 411, 267, 47 },
		{ 411, 268, 47 },
		{ 411, 269, 47 },
		{ 411, 270, 47 },
		{ 411, 271, 47 },
		{ 411, 272, 47 },
		{ 411, 273, 47 },
		{ 411, 274, 47 },
		{ 411, 275, 47 },
		{ 411, 276, 47 },
		{ 411, 277, 47 },
		{ 411, 278, 47 },
		{ 411, 279, 47 },
		{ 411, 280, 47 },
		{ 411, 281, 47 },
		{ 411, 282, 47 },
		{ 411, 283, 47 },
		{ 411, 284, 47 },
		{ 411, 285, 47 },
		{ 411, 286, 47 },
		{ 411, 287, 47 },
		{ 411, 288, 47 },
		{ 411, 289, 47 },
		{ 411, 290, 47 },
		{ 411, 291, 47 },
		{ 411, 292, 47 },
		{ 411, 293, 47 },
		{ 411, 294, 47 },
		{ 411, 295, 47 },
		{ 412, 258, 47 },
		{ 412, 259, 47 },
		{ 412, 260, 47 },
		{ 412, 261, 47 },
		{ 412, 262, 47 },
		{ 412, 263, 47 },
		{ 412, 264, 47 },
		{ 412, 265, 47 },
		{ 412, 266, 47 },
		{ 412, 267, 47 },
		{ 412, 268, 47 },
		{ 412, 269, 47 },
		{ 412, 270, 47 },
		{ 412, 271, 47 },
		{ 412, 272, 47 },
		{ 412, 273, 47 },
		{ 412, 274, 47 },
		{ 412, 275, 47 },
		{ 412, 276, 47 },
		{ 412, 277, 47 },
		{ 412, 278, 47 },
		{ 412, 279, 47 },
		{ 412, 280, 47 },
		{ 412, 281, 47 },
		{ 412, 282, 47 },
		{ 412, 283, 47 },
		{ 412, 284, 47 },
		{ 412, 285, 47 },
		{ 412, 286, 47 },
		{ 412, 287, 47 },
		{ 412, 288, 47 },
		{ 412, 289, 47 },
		{ 412, 290, 47 },
		{ 412, 291, 47 },
		{ 412, 292, 47 },
		{ 412, 293, 47 },
		{ 412, 294, 47 },
		{ 412, 295, 47 },
		{ 413, 258, 47 },
		{ 413, 259, 47 },
		{ 413, 260, 47 },
		{ 413, 261, 47 },
		{ 413, 262, 47 },
		{ 413, 263, 47 },
		{ 413, 264, 47 },
		{ 413, 265, 47 },
		{ 413, 266, 47 },
		{ 413, 267, 47 },
		{ 413, 268, 47 },
		{ 413, 269, 47 },
		{ 413, 270, 47 },
		{ 413, 271, 47 },
		{ 413, 272, 47 },
		{ 413, 273, 47 },
		{ 413, 274, 47 },
		{ 413, 275, 47 },
		{ 413, 276, 47 },
		{ 413, 277, 47 },
		{ 413, 278, 47 },
		{ 413, 279, 47 },
		{ 413, 280, 47 },
		{ 413, 281, 47 },
		{ 413, 282, 47 },
		{ 413, 283, 47 },
		{ 413, 284, 47 },
		{ 413, 285, 47 },
		{ 413, 286, 47 },
		{ 413, 287, 47 },
		{ 413, 288, 47 },
		{ 413, 289, 47 },
		{ 413, 290, 47 },
		{ 413, 291, 47 },
		{ 413, 292, 47 },
		{ 413, 293, 47 },
		{ 413, 294, 47 },
		{ 413, 295, 47 },
		{ 414, 258, 47 },
		{ 414, 259, 47 },
		{ 414, 260, 47 },
		{ 414, 261, 47 },
		{ 414, 262, 47 },
		{ 414, 263, 47 },
		{ 414, 264, 47 },
		{ 414, 265, 47 },
		{ 414, 266, 47 },
		{ 414, 267, 47 },
		{ 414, 268, 47 },
		{ 414, 269, 47 },
		{ 414, 270, 47 },
		{ 414, 271, 47 },
		{ 414, 272, 47 },
		{ 414, 273, 47 },
		{ 414, 274, 47 },
		{ 414, 275, 47 },
		{ 414, 276, 47 },
		{ 414, 277, 47 },
		{ 414, 278, 47 },
		{ 414, 279, 47 },
		{ 414, 280, 47 },
		{ 414, 281, 47 },
		{ 414, 282, 47 },
		{ 414, 283, 47 },
		{ 414, 284, 47 },
		{ 414, 285, 47 },
		{ 414, 286, 47 },
		{ 414, 287, 47 },
		{ 414, 288, 47 },
		{ 414, 289, 47 },
		{ 414, 290, 47 },
		{ 414, 291, 47 },
		{ 414, 292, 47 },
		{ 414, 293, 47 },
		{ 414, 294, 47 },
		{ 414, 295, 47 },
		{ 414, 296, 47 },
		{ 415, 258, 47 },
		{ 415, 259, 47 },
		{ 415, 260, 47 },
		{ 415, 261, 47 },
		{ 415, 262, 47 },
		{ 415, 263, 47 },
		{ 415, 264, 47 },
		{ 415, 265, 47 },
		{ 415, 266, 47 },
		{ 415, 267, 47 },
		{ 415, 268, 47 },
		{ 415, 269, 47 },
		{ 415, 270, 47 },
		{ 415, 271, 47 },
		{ 415, 272, 47 },
		{ 415, 273, 47 },
		{ 415, 274, 47 },
		{ 415, 275, 47 },
		{ 415, 276, 47 },
		{ 415, 277, 47 },
		{ 415, 278, 47 },
		{ 415, 279, 47 },
		{ 415, 280, 47 },
		{ 415, 281, 47 },
		{ 415, 282, 47 },
		{ 415, 283, 47 },
		{ 415, 284, 47 },
		{ 415, 285, 47 },
		{ 415, 286, 47 },
		{ 415, 287, 47 },
		{ 415, 288, 47 },
		{ 415, 289, 47 },
		{ 415, 290, 47 },
		{ 415, 291, 47 },
		{ 415, 292, 47 },
		{ 415, 293, 47 },
		{ 415, 294, 47 },
		{ 415, 295, 47 },
		{ 415, 296, 47 },
		{ 416, 258, 47 },
		{ 416, 259, 47 },
		{ 416, 260, 47 },
		{ 416, 261, 47 },
		{ 416, 262, 47 },
		{ 416, 263, 47 },
		{ 416, 264, 47 },
		{ 416, 265, 47 },
		{ 416, 266, 47 },
		{ 416, 267, 47 },
		{ 416, 268, 47 },
		{ 416, 269, 47 },
		{ 416, 270, 47 },
		{ 416, 271, 47 },
		{ 416, 272, 47 },
		{ 416, 273, 47 },
		{ 416, 274, 47 },
		{ 416, 275, 47 },
		{ 416, 276, 47 },
		{ 416, 277, 47 },
		{ 416, 278, 47 },
		{ 416, 279, 47 },
		{ 416, 280, 47 },
		{ 416, 281, 47 },
		{ 416, 282, 47 },
		{ 416, 283, 47 },
		{ 416, 284, 47 },
		{ 416, 285, 47 },
		{ 416, 286, 47 },
		{ 416, 287, 47 },
		{ 416, 288, 47 },
		{ 416, 289, 47 },
		{ 416, 290, 47 },
		{ 416, 291, 47 },
		{ 416, 292, 47 },
		{ 416, 293, 47 },
		{ 416, 294, 47 },
		{ 416, 295, 47 },
		{ 416, 296, 47 },
		{ 417, 258, 47 },
		{ 417, 259, 47 },
		{ 417, 260, 47 },
		{ 417, 261, 47 },
		{ 417, 262, 47 },
		{ 417, 263, 47 },
		{ 417, 264, 47 },
		{ 417, 265, 47 },
		{ 417, 266, 47 },
		{ 417, 267, 47 },
		{ 417, 268, 47 },
		{ 417, 269, 47 },
		{ 417, 270, 47 },
		{ 417, 271, 47 },
		{ 417, 272, 47 },
		{ 417, 273, 47 },
		{ 417, 274, 47 },
		{ 417, 275, 47 },
		{ 417, 276, 47 },
		{ 417, 277, 47 },
		{ 417, 278, 47 },
		{ 417, 279, 47 },
		{ 417, 280, 47 },
		{ 417, 281, 47 },
		{ 417, 282, 47 },
		{ 417, 283, 47 },
		{ 417, 284, 47 },
		{ 417, 285, 47 },
		{ 417, 286, 47 },
		{ 417, 287, 47 },
		{ 417, 288, 47 },
		{ 417, 289, 47 },
		{ 417, 290, 47 },
		{ 417, 291, 47 },
		{ 417, 292, 47 },
		{ 417, 293, 47 },
		{ 417, 294, 47 },
		{ 417, 295, 47 },
		{ 417, 296, 47 },
		{ 418, 258, 47 },
		{ 418, 259, 47 },
		{ 418, 260, 47 },
		{ 418, 261, 47 },
		{ 418, 262, 47 },
		{ 418, 263, 47 },
		{ 418, 264, 47 },
		{ 418, 265, 47 },
		{ 418, 266, 47 },
		{ 418, 267, 47 },
		{ 418, 268, 47 },
		{ 418, 269, 47 },
		{ 418, 270, 47 },
		{ 418, 271, 47 },
		{ 418, 272, 47 },
		{ 418, 273, 47 },
		{ 418, 274, 47 },
		{ 418, 275, 47 },
		{ 418, 276, 47 },
		{ 418, 277, 47 },
		{ 418, 278, 47 },
		{ 418, 279, 47 },
		{ 418, 280, 47 },
		{ 418, 281, 47 },
		{ 418, 282, 47 },
		{ 418, 283, 47 },
		{ 418, 284, 47 },
		{ 418, 285, 47 },
		{ 418, 286, 47 },
		{ 418, 287, 47 },
		{ 418, 288, 47 },
		{ 418, 289, 47 },
		{ 418, 290, 47 },
		{ 418, 291, 47 },
		{ 418, 292, 47 },
		{ 418, 293, 47 },
		{ 418, 294, 47 },
		{ 418, 295, 47 },
		{ 418, 296, 47 },
		{ 418, 297, 47 },
		{ 419, 258, 47 },
		{ 419, 259, 47 },
		{ 419, 260, 47 },
		{ 419, 261, 47 },
		{ 419, 262, 47 },
		{ 419, 263, 47 },
		{ 419, 264, 47 },
		{ 419, 265, 47 },
		{ 419, 266, 47 },
		{ 419, 267, 47 },
		{ 419, 268, 47 },
		{ 419, 269, 47 },
		{ 419, 270, 47 },
		{ 419, 271, 47 },
		{ 419, 272, 47 },
		{ 419, 273, 47 },
		{ 419, 274, 47 },
		{ 419, 275, 47 },
		{ 419, 276, 47 },
		{ 419, 277, 47 },
		{ 419, 278, 47 },
		{ 419, 279, 47 },
		{ 419, 280, 47 },
		{ 419, 281, 47 },
		{ 419, 282, 47 },
		{ 419, 283, 47 },
		{ 419, 284, 47 },
		{ 419, 285, 47 },
		{ 419, 286, 47 },
		{ 419, 287, 47 },
		{ 419, 288, 47 },
		{ 419, 289, 47 },
		{ 419, 290, 47 },
		{ 419, 291, 47 },
		{ 419, 292, 47 },
		{ 419, 293, 47 },
		{ 419, 294, 47 },
		{ 419, 295, 47 },
		{ 419, 296, 47 },
		{ 419, 297, 47 },
		{ 420, 258, 47 },
		{ 420, 259, 47 },
		{ 420, 260, 47 },
		{ 420, 261, 47 },
		{ 420, 262, 47 },
		{ 420, 263, 47 },
		{ 420, 264, 47 },
		{ 420, 265, 47 },
		{ 420, 266, 47 },
		{ 420, 267, 47 },
		{ 420, 268, 47 },
		{ 420, 269, 47 },
		{ 420, 270, 47 },
		{ 420, 271, 47 },
		{ 420, 272, 47 },
		{ 420, 273, 47 },
		{ 420, 274, 47 },
		{ 420, 275, 47 },
		{ 420, 276, 47 },
		{ 420, 277, 47 },
		{ 420, 278, 47 },
		{ 420, 279, 47 },
		{ 420, 280, 47 },
		{ 420, 281, 47 },
		{ 420, 282, 47 },
		{ 420, 283, 47 },
		{ 420, 284, 47 },
		{ 420, 285, 47 },
		{ 420, 286, 47 },
		{ 420, 287, 47 },
		{ 420, 288, 47 },
		{ 420, 289, 47 },
		{ 420, 290, 47 },
		{ 420, 291, 47 },
		{ 420, 292, 47 },
		{ 420, 293, 47 },
		{ 420, 294, 47 },
		{ 420, 295, 47 },
		{ 420, 296, 47 },
		{ 420, 297, 47 },
		{ 421, 258, 47 },
		{ 421, 259, 47 },
		{ 421, 260, 47 },
		{ 421, 261, 47 },
		{ 421, 262, 47 },
		{ 421, 263, 47 },
		{ 421, 264, 47 },
		{ 421, 265, 47 },
		{ 421, 266, 47 },
		{ 421, 267, 47 },
		{ 421, 268, 47 },
		{ 421, 269, 47 },
		{ 421, 270, 47 },
		{ 421, 271, 47 },
		{ 421, 272, 47 },
		{ 421, 273, 47 },
		{ 421, 274, 47 },
		{ 421, 275, 47 },
		{ 421, 276, 47 },
		{ 421, 277, 47 },
		{ 421, 278, 47 },
		{ 421, 279, 47 },
		{ 421, 280, 47 },
		{ 421, 281, 47 },
		{ 421, 282, 47 },
		{ 421, 283, 47 },
		{ 421, 284, 47 },
		{ 421, 285, 47 },
		{ 421, 286, 47 },
		{ 421, 287, 47 },
		{ 421, 288, 47 },
		{ 421, 289, 47 },
		{ 421, 290, 47 },
		{ 421, 291, 47 },
		{ 421, 292, 47 },
		{ 421, 293, 47 },
		{ 421, 294, 47 },
		{ 421, 295, 47 },
		{ 421, 296, 47 },
		{ 421, 297, 47 },
		{ 422, 258, 47 },
		{ 422, 259, 47 },
		{ 422, 260, 47 },
		{ 422, 261, 47 },
		{ 422, 262, 47 },
		{ 422, 263, 47 },
		{ 422, 264, 47 },
		{ 422, 265, 47 },
		{ 422, 266, 47 },
		{ 422, 267, 47 },
		{ 422, 268, 47 },
		{ 422, 269, 47 },
		{ 422, 270, 47 },
		{ 422, 271, 47 },
		{ 422, 272, 47 },
		{ 422, 273, 47 },
		{ 422, 274, 47 },
		{ 422, 275, 47 },
		{ 422, 276, 47 },
		{ 422, 277, 47 },
		{ 422, 278, 47 },
		{ 422, 279, 47 },
		{ 422, 280, 47 },
		{ 422, 281, 47 },
		{ 422, 282, 47 },
		{ 422, 283, 47 },
		{ 422, 284, 47 },
		{ 422, 285, 47 },
		{ 422, 286, 47 },
		{ 422, 287, 47 },
		{ 422, 288, 47 },
		{ 422, 289, 47 },
		{ 422, 290, 47 },
		{ 422, 291, 47 },
		{ 422, 292, 47 },
		{ 422, 293, 47 },
		{ 422, 294, 47 },
		{ 422, 295, 47 },
		{ 422, 296, 47 },
		{ 422, 297, 47 },
		{ 422, 298, 47 },
		{ 423, 258, 47 },
		{ 423, 259, 47 },
		{ 423, 260, 47 },
		{ 423, 261, 47 },
		{ 423, 262, 47 },
		{ 423, 263, 47 },
		{ 423, 264, 47 },
		{ 423, 265, 47 },
		{ 423, 266, 47 },
		{ 423, 267, 47 },
		{ 423, 268, 47 },
		{ 423, 269, 47 },
		{ 423, 270, 47 },
		{ 423, 271, 47 },
		{ 423, 272, 47 },
		{ 423, 273, 47 },
		{ 423, 274, 47 },
		{ 423, 275, 47 },
		{ 423, 276, 47 },
		{ 423, 277, 47 },
		{ 423, 278, 47 },
		{ 423, 279, 47 },
		{ 423, 280, 47 },
		{ 423, 281, 47 },
		{ 423, 282, 47 },
		{ 423, 283, 47 },
		{ 423, 284, 47 },
		{ 423, 285, 47 },
		{ 423, 286, 47 },
		{ 423, 287, 47 },
		{ 423, 288, 47 },
		{ 423, 289, 47 },
		{ 423, 290, 47 },
		{ 423, 291, 47 },
		{ 423, 292, 47 },
		{ 423, 293, 47 },
		{ 423, 294, 47 },
		{ 423, 295, 47 },
		{ 423, 296, 47 },
		{ 423, 297, 47 },
		{ 423, 298, 47 },
		{ 424, 258, 47 },
		{ 424, 259, 47 },
		{ 424, 260, 47 },
		{ 424, 261, 47 },
		{ 424, 262, 47 },
		{ 424, 263, 47 },
		{ 424, 264, 47 },
		{ 424, 265, 47 },
		{ 424, 266, 47 },
		{ 424, 267, 47 },
		{ 424, 268, 47 },
		{ 424, 269, 47 },
		{ 424, 270, 47 },
		{ 424, 271, 47 },
		{ 424, 272, 47 },
		{ 424, 273, 47 },
		{ 424, 274, 47 },
		{ 424, 275, 47 },
		{ 424, 276, 47 },
		{ 424, 277, 47 },
		{ 424, 278, 47 },
		{ 424, 279, 47 },
		{ 424, 280, 47 },
		{ 424, 281, 47 },
		{ 424, 282, 47 },
		{ 424, 283, 47 },
		{ 424, 284, 47 },
		{ 424, 285, 47 },
		{ 424, 286, 47 },
		{ 424, 287, 47 },
		{ 424, 288, 47 },
		{ 424, 289, 47 },
		{ 424, 290, 47 },
		{ 424, 291, 47 },
		{ 424, 292, 47 },
		{ 424, 293, 47 },
		{ 424, 294, 47 },
		{ 424, 295, 47 },
		{ 424, 296, 47 },
		{ 424, 297, 47 },
		{ 424, 298, 47 },
		{ 425, 258, 47 },
		{ 425, 259, 47 },
		{ 425, 260, 47 },
		{ 425, 261, 47 },
		{ 425, 262, 47 },
		{ 425, 263, 47 },
		{ 425, 264, 47 },
		{ 425, 265, 47 },
		{ 425, 266, 47 },
		{ 425, 267, 47 },
		{ 425, 268, 47 },
		{ 425, 269, 47 },
		{ 425, 270, 47 },
		{ 425, 271, 47 },
		{ 425, 272, 47 },
		{ 425, 273, 47 },
		{ 425, 274, 47 },
		{ 425, 275, 47 },
		{ 425, 276, 47 },
		{ 425, 277, 47 },
		{ 425, 278, 47 },
		{ 425, 279, 47 },
		{ 425, 280, 47 },
		{ 425, 281, 47 },
		{ 425, 282, 47 },
		{ 425, 283, 47 },
		{ 425, 284, 47 },
		{ 425, 285, 47 },
		{ 425, 286, 47 },
		{ 425, 287, 47 },
		{ 425, 288, 47 },
		{ 425, 289, 47 },
		{ 425, 290, 47 },
		{ 425, 291, 47 },
		{ 425, 292, 47 },
		{ 425, 293, 47 },
		{ 425, 294, 47 },
		{ 425, 295, 47 },
		{ 425, 296, 47 },
		{ 425, 297, 47 },
		{ 425, 298, 47 },
		{ 426, 258, 47 },
		{ 426, 259, 47 },
		{ 426, 260, 47 },
		{ 426, 261, 47 },
		{ 426, 262, 47 },
		{ 426, 263, 47 },
		{ 426, 264, 47 },
		{ 426, 265, 47 },
		{ 426, 266, 47 },
		{ 426, 267, 47 },
		{ 426, 268, 47 },
		{ 426, 269, 47 },
		{ 426, 270, 47 },
		{ 426, 271, 47 },
		{ 426, 272, 47 },
		{ 426, 273, 47 },
		{ 426, 274, 47 },
		{ 426, 275, 47 },
		{ 426, 276, 47 },
		{ 426, 277, 47 },
		{ 426, 278, 47 },
		{ 426, 279, 47 },
		{ 426, 280, 47 },
		{ 426, 281, 47 },
		{ 426, 282, 47 },
		{ 426, 283, 47 },
		{ 426, 284, 47 },
		{ 426, 285, 47 },
		{ 426, 286, 47 },
		{ 426, 287, 47 },
		{ 426, 288, 47 },
		{ 426, 289, 47 },
		{ 426, 290, 47 },
		{ 426, 291, 47 },
		{ 426, 292, 47 },
		{ 426, 293, 47 },
		{ 426, 294, 47 },
		{ 426, 295, 47 },
		{ 426, 296, 47 },
		{ 426, 297, 47 },
		{ 426, 298, 47 },
		{ 426, 299, 47 },
		{ 427, 258, 47 },
		{ 427, 259, 47 },
		{ 427, 260, 47 },
		{ 427, 261, 47 },
		{ 427, 262, 47 },
		{ 427, 263, 47 },
		{ 427, 264, 47 },
		{ 427, 265, 47 },
		{ 427, 266, 47 },
		{ 427, 267, 47 },
		{ 427, 268, 47 },
		{ 427, 269, 47 },
		{ 427, 270, 47 },
		{ 427, 271, 47 },
		{ 427, 272, 47 },
		{ 427, 273, 47 },
		{ 427, 274, 47 },
		{ 427, 275, 47 },
		{ 427, 276, 47 },
		{ 427, 277, 47 },
		{ 427, 278, 47 },
		{ 427, 279, 47 },
		{ 427, 280, 47 },
		{ 427, 281, 47 },
		{ 427, 282, 47 },
		{ 427, 283, 47 },
		{ 427, 284, 47 },
		{ 427, 285, 47 },
		{ 427, 286, 47 },
		{ 427, 287, 47 },
		{ 427, 288, 47 },
		{ 427, 289, 47 },
		{ 427, 290, 47 },
		{ 427, 291, 47 },
		{ 427, 292, 47 },
		{ 427, 293, 47 },
		{ 427, 294, 47 },
		{ 427, 295, 47 },
		{ 427, 296, 47 },
		{ 427, 297, 47 },
		{ 427, 298, 47 },
		{ 427, 299, 47 },
		{ 428, 258, 47 },
		{ 428, 259, 47 },
		{ 428, 260, 47 },
		{ 428, 261, 47 },
		{ 428, 262, 47 },
		{ 428, 263, 47 },
		{ 428, 264, 47 },
		{ 428, 265, 47 },
		{ 428, 266, 47 },
		{ 428, 267, 47 },
		{ 428, 268, 47 },
		{ 428, 269, 47 },
		{ 428, 270, 47 },
		{ 428, 271, 47 },
		{ 428, 272, 47 },
		{ 428, 273, 47 },
		{ 428, 274, 47 },
		{ 428, 275, 47 },
		{ 428, 276, 47 },
		{ 428, 277, 47 },
		{ 428, 278, 47 },
		{ 428, 279, 47 },
		{ 428, 280, 47 },
		{ 428, 281, 47 },
		{ 428, 282, 47 },
		{ 428, 283, 47 },
		{ 428, 284, 47 },
		{ 428, 285, 47 },
		{ 428, 286, 47 },
		{ 428, 287, 47 },
		{ 428, 288, 47 },
		{ 428, 289, 47 },
		{ 428, 290, 47 },
		{ 428, 291, 47 },
		{ 428, 292, 47 },
		{ 428, 293, 47 },
		{ 428, 294, 47 },
		{ 428, 295, 47 },
		{ 428, 296, 47 },
		{ 428, 297, 47 },
		{ 428, 298, 47 },
		{ 428, 299, 47 },
		{ 429, 258, 47 },
		{ 429, 259, 47 },
		{ 429, 260, 47 },
		{ 429, 261, 47 },
		{ 429, 262, 47 },
		{ 429, 263, 47 },
		{ 429, 264, 47 },
		{ 429, 265, 47 },
		{ 429, 266, 47 },
		{ 429, 267, 47 },
		{ 429, 268, 47 },
		{ 429, 269, 47 },
		{ 429, 270, 47 },
		{ 429, 271, 47 },
		{ 429, 272, 47 },
		{ 429, 273, 47 },
		{ 429, 274, 47 },
		{ 429, 275, 47 },
		{ 429, 276, 47 },
		{ 429, 277, 47 },
		{ 429, 278, 47 },
		{ 429, 279, 47 },
		{ 429, 280, 47 },
		{ 429, 281, 47 },
		{ 429, 282, 47 },
		{ 429, 283, 47 },
		{ 429, 284, 47 },
		{ 429, 285, 47 },
		{ 429, 286, 47 },
		{ 429, 287, 47 },
		{ 429, 288, 47 },
		{ 429, 289, 47 },
		{ 429, 290, 47 },
		{ 429, 291, 47 },
		{ 429, 292, 47 },
		{ 429, 293, 47 },
		{ 429, 294, 47 },
		{ 429, 295, 47 },
		{ 429, 296, 47 },
		{ 429, 297, 47 },
		{ 429, 298, 47 },
		{ 429, 299, 47 },
		{ 430, 258, 47 },
		{ 430, 259, 47 },
		{ 430, 260, 47 },
		{ 430, 261, 47 },
		{ 430, 262, 47 },
		{ 430, 263, 47 },
		{ 430, 264, 47 },
		{ 430, 265, 47 },
		{ 430, 266, 47 },
		{ 430, 267, 47 },
		{ 430, 268, 47 },
		{ 430, 269, 47 },
		{ 430, 270, 47 },
		{ 430, 271, 47 },
		{ 430, 272, 47 },
		{ 430, 273, 47 },
		{ 430, 274, 47 },
		{ 430, 275, 47 },
		{ 430, 276, 47 },
		{ 430, 277, 47 },
		{ 430, 278, 47 },
		{ 430, 279, 47 },
		{ 430, 280, 47 },
		{ 430, 281, 47 },
		{ 430, 282, 47 },
		{ 430, 283, 47 },
		{ 430, 284, 47 },
		{ 430, 285, 47 },
		{ 430, 286, 47 },
		{ 430, 287, 47 },
		{ 430, 288, 47 },
		{ 430, 289, 47 },
		{ 430, 290, 47 },
		{ 430, 291, 47 },
		{ 430, 292, 47 },
		{ 430, 293, 47 },
		{ 430, 294, 47 },
		{ 430, 295, 47 },
		{ 430, 296, 47 },
		{ 430, 297, 47 },
		{ 430, 298, 47 },
		{ 430, 299, 47 },
		{ 430, 300, 47 },
		{ 431, 258, 47 },
		{ 431, 259, 47 },
		{ 431, 260, 47 },
		{ 431, 261, 47 },
		{ 431, 262, 47 },
		{ 431, 263, 47 },
		{ 431, 264, 47 },
		{ 431, 265, 47 },
		{ 431, 266, 47 },
		{ 431, 267, 47 },
		{ 431, 268, 47 },
		{ 431, 269, 47 },
		{ 431, 270, 47 },
		{ 431, 271, 47 },
		{ 431, 272, 47 },
		{ 431, 273, 47 },
		{ 431, 274, 47 },
		{ 431, 275, 47 },
		{ 431, 276, 47 },
		{ 431, 277, 47 },
		{ 431, 278, 47 },
		{ 431, 279, 47 },
		{ 431, 280, 47 },
		{ 431, 281, 47 },
		{ 431, 282, 47 },
		{ 431, 283, 47 },
		{ 431, 284, 47 },
		{ 431, 285, 47 },
		{ 431, 286, 47 },
		{ 431, 287, 47 },
		{ 431, 288, 47 },
		{ 431, 289, 47 },
		{ 431, 290, 47 },
		{ 431, 291, 47 },
		{ 431, 292, 47 },
		{ 431, 293, 47 },
		{ 431, 294, 47 },
		{ 431, 295, 47 },
		{ 431, 296, 47 },
		{ 431, 297, 47 },
		{ 431, 298, 47 },
		{ 431, 299, 47 },
		{ 431, 300, 47 },
		{ 432, 258, 47 },
		{ 432, 259, 47 },
		{ 432, 260, 47 },
		{ 432, 261, 47 },
		{ 432, 262, 47 },
		{ 432, 263, 47 },
		{ 432, 264, 47 },
		{ 432, 265, 47 },
		{ 432, 266, 47 },
		{ 432, 267, 47 },
		{ 432, 268, 47 },
		{ 432, 269, 47 },
		{ 432, 270, 47 },
		{ 432, 271, 47 },
		{ 432, 272, 47 },
		{ 432, 273, 47 },
		{ 432, 274, 47 },
		{ 432, 275, 47 },
		{ 432, 276, 47 },
		{ 432, 277, 47 },
		{ 432, 278, 47 },
		{ 432, 279, 47 },
		{ 432, 280, 47 },
		{ 432, 281, 47 },
		{ 432, 282, 47 },
		{ 432, 283, 47 },
		{ 432, 284, 47 },
		{ 432, 285, 47 },
		{ 432, 286, 47 },
		{ 432, 287, 47 },
		{ 432, 288, 47 },
		{ 432, 289, 47 },
		{ 432, 290, 47 },
		{ 432, 291, 47 },
		{ 432, 292, 47 },
		{ 432, 293, 47 },
		{ 432, 294, 47 },
		{ 432, 295, 47 },
		{ 432, 296, 47 },
		{ 432, 297, 47 },
		{ 432, 298, 47 },
		{ 432, 299, 47 },
		{ 432, 300, 47 },
		{ 433, 258, 47 },
		{ 433, 259, 47 },
		{ 433, 260, 47 },
		{ 433, 261, 47 },
		{ 433, 262, 47 },
		{ 433, 263, 47 },
		{ 433, 264, 47 },
		{ 433, 265, 47 },
		{ 433, 266, 47 },
		{ 433, 267, 47 },
		{ 433, 268, 47 },
		{ 433, 269, 47 },
		{ 433, 270, 47 },
		{ 433, 271, 47 },
		{ 433, 272, 47 },
		{ 433, 273, 47 },
		{ 433, 274, 47 },
		{ 433, 275, 47 },
		{ 433, 276, 47 },
		{ 433, 277, 47 },
		{ 433, 278, 47 },
		{ 433, 279, 47 },
		{ 433, 280, 47 },
		{ 433, 281, 47 },
		{ 433, 282, 47 },
		{ 433, 283, 47 },
		{ 433, 284, 47 },
		{ 433, 285, 47 },
		{ 433, 286, 47 },
		{ 433, 287, 47 },
		{ 433, 288, 47 },
		{ 433, 289, 47 },
		{ 433, 290, 47 },
		{ 433, 291, 47 },
		{ 433, 292, 47 },
		{ 433, 293, 47 },
		{ 433, 294, 47 },
		{ 433, 295, 47 },
		{ 433, 296, 47 },
		{ 433, 297, 47 },
		{ 433, 298, 47 },
		{ 433, 299, 47 },
		{ 433, 300, 47 },
		{ 434, 258, 47 },
		{ 434, 259, 47 },
		{ 434, 260, 47 },
		{ 434, 261, 47 },
		{ 434, 262, 47 },
		{ 434, 263, 47 },
		{ 434, 264, 47 },
		{ 434, 265, 47 },
		{ 434, 266, 47 },
		{ 434, 267, 47 },
		{ 434, 268, 47 },
		{ 434, 269, 47 },
		{ 434, 270, 47 },
		{ 434, 271, 47 },
		{ 434, 272, 47 },
		{ 434, 273, 47 },
		{ 434, 274, 47 },
		{ 434, 275, 47 },
		{ 434, 276, 47 },
		{ 434, 277, 47 },
		{ 434, 278, 47 },
		{ 434, 279, 47 },
		{ 434, 280, 47 },
		{ 434, 281, 47 },
		{ 434, 282, 47 },
		{ 434, 283, 47 },
		{ 434, 284, 47 },
		{ 434, 285, 47 },
		{ 434, 286, 47 },
		{ 434, 287, 47 },
		{ 434, 288, 47 },
		{ 434, 289, 47 },
		{ 434, 290, 47 },
		{ 434, 291, 47 },
		{ 434, 292, 47 },
		{ 434, 293, 47 },
		{ 434, 294, 47 },
		{ 434, 295, 47 },
		{ 434, 296, 47 },
		{ 434, 297, 47 },
		{ 434, 298, 47 },
		{ 434, 299, 47 },
		{ 434, 300, 47 },
		{ 434, 301, 47 },
		{ 435, 258, 47 },
		{ 435, 259, 47 },
		{ 435, 260, 47 },
		{ 435, 261, 47 },
		{ 435, 262, 47 },
		{ 435, 263, 47 },
		{ 435, 264, 47 },
		{ 435, 265, 47 },
		{ 435, 266, 47 },
		{ 435, 267, 47 },
		{ 435, 268, 47 },
		{ 435, 269, 47 },
		{ 435, 270, 47 },
		{ 435, 271, 47 },
		{ 435, 272, 47 },
		{ 435, 273, 47 },
		{ 435, 274, 47 },
		{ 435, 275, 47 },
		{ 435, 276, 47 },
		{ 435, 277, 47 },
		{ 435, 278, 47 },
		{ 435, 279, 47 },
		{ 435, 280, 47 },
		{ 435, 281, 47 },
		{ 435, 282, 47 },
		{ 435, 283, 47 },
		{ 435, 284, 47 },
		{ 435, 285, 47 },
		{ 435, 286, 47 },
		{ 435, 287, 47 },
		{ 435, 288, 47 },
		{ 435, 289, 47 },
		{ 435, 290, 47 },
		{ 435, 291, 47 },
		{ 435, 292, 47 },
		{ 435, 293, 47 },
		{ 435, 294, 47 },
		{ 435, 295, 47 },
		{ 435, 296, 47 },
		{ 435, 297, 47 },
		{ 435, 298, 47 },
		{ 435, 299, 47 },
		{ 435, 300, 47 },
		{ 435, 301, 47 },
		{ 436, 258, 47 },
		{ 436, 259, 47 },
		{ 436, 260, 47 },
		{ 436, 261, 47 },
		{ 436, 262, 47 },
		{ 436, 263, 47 },
		{ 436, 264, 47 },
		{ 436, 265, 47 },
		{ 436, 266, 47 },
		{ 436, 267, 47 },
		{ 436, 268, 47 },
		{ 436, 269, 47 },
		{ 436, 270, 47 },
		{ 436, 271, 47 },
		{ 436, 272, 47 },
		{ 436, 273, 47 },
		{ 436, 274, 47 },
		{ 436, 275, 47 },
		{ 436, 276, 47 },
		{ 436, 277, 47 },
		{ 436, 278, 47 },
		{ 436, 279, 47 },
		{ 436, 280, 47 },
		{ 436, 281, 47 },
		{ 436, 282, 47 },
		{ 436, 283, 47 },
		{ 436, 284, 47 },
		{ 436, 285, 47 },
		{ 436, 286, 47 },
		{ 436, 287, 47 },
		{ 436, 288, 47 },
		{ 436, 289, 47 },
		{ 436, 290, 47 },
		{ 436, 291, 47 },
		{ 436, 292, 47 },
		{ 436, 293, 47 },
		{ 436, 294, 47 },
		{ 436, 295, 47 },
		{ 436, 296, 47 },
		{ 436, 297, 47 },
		{ 436, 298, 47 },
		{ 436, 299, 47 },
		{ 436, 300, 47 },
		{ 436, 301, 47 },
		{ 437, 258, 47 },
		{ 437, 259, 47 },
		{ 437, 260, 47 },
		{ 437, 261, 47 },
		{ 437, 262, 47 },
		{ 437, 263, 47 },
		{ 437, 264, 47 },
		{ 437, 265, 47 },
		{ 437, 266, 47 },
		{ 437, 267, 47 },
		{ 437, 268, 47 },
		{ 437, 269, 47 },
		{ 437, 270, 47 },
		{ 437, 271, 47 },
		{ 437, 272, 47 },
		{ 437, 273, 47 },
		{ 437, 274, 47 },
		{ 437, 275, 47 },
		{ 437, 276, 47 },
		{ 437, 277, 47 },
		{ 437, 278, 47 },
		{ 437, 279, 47 },
		{ 437, 280, 47 },
		{ 437, 281, 47 },
		{ 437, 282, 47 },
		{ 437, 283, 47 },
		{ 437, 284, 47 },
		{ 437, 285, 47 },
		{ 437, 286, 47 },
		{ 437, 287, 47 },
		{ 437, 288, 47 },
		{ 437, 289, 47 },
		{ 437, 290, 47 },
		{ 437, 291, 47 },
		{ 437, 292, 47 },
		{ 437, 293, 47 },
		{ 437, 294, 47 },
		{ 437, 295, 47 },
		{ 437, 296, 47 },
		{ 437, 297, 47 },
		{ 437, 298, 47 },
		{ 437, 299, 47 },
		{ 437, 300, 47 },
		{ 437, 301, 47 },
		{ 438, 258, 47 },
		{ 438, 259, 47 },
		{ 438, 260, 47 },
		{ 438, 261, 47 },
		{ 438, 262, 47 },
		{ 438, 263, 47 },
		{ 438, 264, 47 },
		{ 438, 265, 47 },
		{ 438, 266, 47 },
		{ 438, 267, 47 },
		{ 438, 268, 47 },
		{ 438, 269, 47 },
		{ 438, 270, 47 },
		{ 438, 271, 47 },
		{ 438, 272, 47 },
		{ 438, 273, 47 },
		{ 438, 274, 47 },
		{ 438, 275, 47 },
		{ 438, 276, 47 },
		{ 438, 277, 47 },
		{ 438, 278, 47 },
		{ 438, 279, 47 },
		{ 438, 280, 47 },
		{ 438, 281, 47 },
		{ 438, 282, 47 },
		{ 438, 283, 47 },
		{ 438, 284, 47 },
		{ 438, 285, 47 },
		{ 438, 286, 47 },
		{ 438, 287, 47 },
		{ 438, 288, 47 },
		{ 438, 289, 47 },
		{ 438, 290, 47 },
		{ 438, 291, 47 },
		{ 438, 292, 47 },
		{ 438, 293, 47 },
		{ 438, 294, 47 },
		{ 438, 295, 47 },
		{ 438, 296, 47 },
		{ 438, 297, 47 },
		{ 438, 298, 47 },
		{ 438, 299, 47 },
		{ 438, 300, 47 },
		{ 438, 301, 47 },
		{ 438, 302, 47 },
		{ 439, 258, 47 },
		{ 439, 259, 47 },
		{ 439, 260, 47 },
		{ 439, 261, 47 },
		{ 439, 262, 47 },
		{ 439, 263, 47 },
		{ 439, 264, 47 },
		{ 439, 265, 47 },
		{ 439, 266, 47 },
		{ 439, 267, 47 },
		{ 439, 268, 47 },
		{ 439, 269, 47 },
		{ 439, 270, 47 },
		{ 439, 271, 47 },
		{ 439, 272, 47 },
		{ 439, 273, 47 },
		{ 439, 274, 47 },
		{ 439, 275, 47 },
		{ 439, 276, 47 },
		{ 439, 277, 47 },
		{ 439, 278, 47 },
		{ 439, 279, 47 },
		{ 439, 280, 47 },
		{ 439, 281, 47 },
		{ 439, 282, 47 },
		{ 439, 283, 47 },
		{ 439, 284, 47 },
		{ 439, 285, 47 },
		{ 439, 286, 47 },
		{ 439, 287, 47 },
		{ 439, 288, 47 },
		{ 439, 289, 47 },
		{ 439, 290, 47 },
		{ 439, 291, 47 },
		{ 439, 292, 47 },
		{ 439, 293, 47 },
		{ 439, 294, 47 },
		{ 439, 295, 47 },
		{ 439, 296, 47 },
		{ 439, 297, 47 },
		{ 439, 298, 47 },
		{ 439, 299, 47 },
		{ 439, 300, 47 },
		{ 439, 301, 47 },
		{ 439, 302, 47 },
		{ 440, 258, 47 },
		{ 440, 259, 47 },
		{ 440, 260, 47 },
		{ 440, 261, 47 },
		{ 440, 262, 47 },
		{ 440, 263, 47 },
		{ 440, 264, 47 },
		{ 440, 265, 47 },
		{ 440, 266, 47 },
		{ 440, 267, 47 },
		{ 440, 268, 47 },
		{ 440, 269, 47 },
		{ 440, 270, 47 },
		{ 440, 271, 47 },
		{ 440, 272, 47 },
		{ 440, 273, 47 },
		{ 440, 274, 47 },
		{ 440, 275, 47 },
		{ 440, 276, 47 },
		{ 440, 277, 47 },
		{ 440, 278, 47 },
		{ 440, 279, 47 },
		{ 440, 280, 47 },
		{ 440, 281, 47 },
		{ 440, 282, 47 },
		{ 440, 283, 47 },
		{ 440, 284, 47 },
		{ 440, 285, 47 },
		{ 440, 286, 47 },
		{ 440, 287, 47 },
		{ 440, 288, 47 },
		{ 440, 289, 47 },
		{ 440, 290, 47 },
		{ 440, 291, 47 },
		{ 440, 292, 47 },
		{ 440, 293, 47 },
		{ 440, 294, 47 },
		{ 440, 295, 47 },
		{ 440, 296, 47 },
		{ 440, 297, 47 },
		{ 440, 298, 47 },
		{ 440, 299, 47 },
		{ 440, 300, 47 },
		{ 440, 301, 47 },
		{ 440, 302, 47 },
		{ 441, 258, 47 },
		{ 441, 259, 47 },
		{ 441, 260, 47 },
		{ 441, 261, 47 },
		{ 441, 262, 47 },
		{ 441, 263, 47 },
		{ 441, 264, 47 },
		{ 441, 265, 47 },
		{ 441, 266, 47 },
		{ 441, 267, 47 },
		{ 441, 268, 47 },
		{ 441, 269, 47 },
		{ 441, 270, 47 },
		{ 441, 271, 47 },
		{ 441, 272, 47 },
		{ 441, 273, 47 },
		{ 441, 274, 47 },
		{ 441, 275, 47 },
		{ 441, 276, 47 },
		{ 441, 277, 47 },
		{ 441, 278, 47 },
		{ 441, 279, 47 },
		{ 441, 280, 47 },
		{ 441, 281, 47 },
		{ 441, 282, 47 },
		{ 441, 283, 47 },
		{ 441, 284, 47 },
		{ 441, 285, 47 },
		{ 441, 286, 47 },
		{ 441, 287, 47 },
		{ 441, 288, 47 },
		{ 441, 289, 47 },
		{ 441, 290, 47 },
		{ 441, 291, 47 },
		{ 441, 292, 47 },
		{ 441, 293, 47 },
		{ 441, 294, 47 },
		{ 441, 295, 47 },
		{ 441, 296, 47 },
		{ 441, 297, 47 },
		{ 441, 298, 47 },
		{ 441, 299, 47 },
		{ 441, 300, 47 },
		{ 441, 301, 47 },
		{ 441, 302, 47 },
		{ 442, 258, 47 },
		{ 442, 259, 47 },
		{ 442, 260, 47 },
		{ 442, 261, 47 },
		{ 442, 262, 47 },
		{ 442, 263, 47 },
		{ 442, 264, 47 },
		{ 442, 265, 47 },
		{ 442, 266, 47 },
		{ 442, 267, 47 },
		{ 442, 268, 47 },
		{ 442, 269, 47 },
		{ 442, 270, 47 },
		{ 442, 271, 47 },
		{ 442, 272, 47 },
		{ 442, 273, 47 },
		{ 442, 274, 47 },
		{ 442, 275, 47 },
		{ 442, 276, 47 },
		{ 442, 277, 47 },
		{ 442, 278, 47 },
		{ 442, 279, 47 },
		{ 442, 280, 47 },
		{ 442, 281, 47 },
		{ 442, 282, 47 },
		{ 442, 283, 47 },
		{ 442, 284, 47 },
		{ 442, 285, 47 },
		{ 442, 286, 47 },
		{ 442, 287, 47 },
		{ 442, 288, 47 },
		{ 442, 289, 47 },
		{ 442, 290, 47 },
		{ 442, 291, 47 },
		{ 442, 292, 47 },
		{ 442, 293, 47 },
		{ 442, 294, 47 },
		{ 442, 295, 47 },
		{ 442, 296, 47 },
		{ 442, 297, 47 },
		{ 442, 298, 47 },
		{ 442, 299, 47 },
		{ 442, 300, 47 },
		{ 442, 301, 47 },
		{ 442, 302, 47 },
		{ 442, 303, 47 },
		{ 443, 258, 47 },
		{ 443, 259, 47 },
		{ 443, 260, 47 },
		{ 443, 261, 47 },
		{ 443, 262, 47 },
		{ 443, 263, 47 },
		{ 443, 264, 47 },
		{ 443, 265, 47 },
		{ 443, 266, 47 },
		{ 443, 267, 47 },
		{ 443, 268, 47 },
		{ 443, 269, 47 },
		{ 443, 270, 47 },
		{ 443, 271, 47 },
		{ 443, 272, 47 },
		{ 443, 273, 47 },
		{ 443, 274, 47 },
		{ 443, 275, 47 },
		{ 443, 276, 47 },
		{ 443, 277, 47 },
		{ 443, 278, 47 },
		{ 443, 279, 47 },
		{ 443, 280, 47 },
		{ 443, 281, 47 },
		{ 443, 282, 47 },
		{ 443, 283, 47 },
		{ 443, 284, 47 },
		{ 443, 285, 47 },
		{ 443, 286, 47 },
		{ 443, 287, 47 },
		{ 443, 288, 47 },
		{ 443, 289, 47 },
		{ 443, 290, 47 },
		{ 443, 291, 47 },
		{ 443, 292, 47 },
		{ 443, 293, 47 },
		{ 443, 294, 47 },
		{ 443, 295, 47 },
		{ 443, 296, 47 },
		{ 443, 297, 47 },
		{ 443, 298, 47 },
		{ 443, 299, 47 },
		{ 443, 300, 47 },
		{ 443, 301, 47 },
		{ 443, 302, 47 },
		{ 443, 303, 47 },
		{ 444, 258, 47 },
		{ 444, 259, 47 },
		{ 444, 260, 47 },
		{ 444, 261, 47 },
		{ 444, 262, 47 },
		{ 444, 263, 47 },
		{ 444, 264, 47 },
		{ 444, 265, 47 },
		{ 444, 266, 47 },
		{ 444, 267, 47 },
		{ 444, 268, 47 },
		{ 444, 269, 47 },
		{ 444, 270, 47 },
		{ 444, 271, 47 },
		{ 444, 272, 47 },
		{ 444, 273, 47 },
		{ 444, 274, 47 },
		{ 444, 275, 47 },
		{ 444, 276, 47 },
		{ 444, 277, 47 },
		{ 444, 278, 47 },
		{ 444, 279, 47 },
		{ 444, 280, 47 },
		{ 444, 281, 47 },
		{ 444, 282, 47 },
		{ 444, 283, 47 },
		{ 444, 284, 47 },
		{ 444, 285, 47 },
		{ 444, 286, 47 },
		{ 444, 287, 47 },
		{ 444, 288, 47 },
		{ 444, 289, 47 },
		{ 444, 290, 47 },
		{ 444, 291, 47 },
		{ 444, 292, 47 },
		{ 444, 293, 47 },
		{ 444, 294, 47 },
		{ 444, 295, 47 },
		{ 444, 296, 47 },
		{ 444, 297, 47 },
		{ 444, 298, 47 },
		{ 444, 299, 47 },
		{ 444, 300, 47 },
		{ 444, 301, 47 },
		{ 444, 302, 47 },
		{ 444, 303, 47 },
		{ 445, 258, 47 },
		{ 445, 259, 47 },
		{ 445, 260, 47 },
		{ 445, 261, 47 },
		{ 445, 262, 47 },
		{ 445, 263, 47 },
		{ 445, 264, 47 },
		{ 445, 265, 47 },
		{ 445, 266, 47 },
		{ 445, 267, 47 },
		{ 445, 268, 47 },
		{ 445, 269, 47 },
		{ 445, 270, 47 },
		{ 445, 271, 47 },
		{ 445, 272, 47 },
		{ 445, 273, 47 },
		{ 445, 274, 47 },
		{ 445, 275, 47 },
		{ 445, 276, 47 },
		{ 445, 277, 47 },
		{ 445, 278, 47 },
		{ 445, 279, 47 },
		{ 445, 280, 47 },
		{ 445, 281, 47 },
		{ 445, 282, 47 },
		{ 445, 283, 47 },
		{ 445, 284, 47 },
		{ 445, 285, 47 },
		{ 445, 286, 47 },
		{ 445, 287, 47 },
		{ 445, 288, 47 },
		{ 445, 289, 47 },
		{ 445, 290, 47 },
		{ 445, 291, 47 },
		{ 445, 292, 47 },
		{ 445, 293, 47 },
		{ 445, 294, 47 },
		{ 445, 295, 47 },
		{ 445, 296, 47 },
		{ 445, 297, 47 },
		{ 445, 298, 47 },
		{ 445, 299, 47 },
		{ 445, 300, 47 },
		{ 445, 301, 47 },
		{ 445, 302, 47 },
		{ 445, 303, 47 },
		{ 446, 258, 47 },
		{ 446, 259, 47 },
		{ 446, 260, 47 },
		{ 446, 261, 47 },
		{ 446, 262, 47 },
		{ 446, 263, 47 },
		{ 446, 264, 47 },
		{ 446, 265, 47 },
		{ 446, 266, 47 },
		{ 446, 267, 47 },
		{ 446, 268, 47 },
		{ 446, 269, 47 },
		{ 446, 270, 47 },
		{ 446, 271, 47 },
		{ 446, 272, 47 },
		{ 446, 273, 47 },
		{ 446, 274, 47 },
		{ 446, 275, 47 },
		{ 446, 276, 47 },
		{ 446, 277, 47 },
		{ 446, 278, 47 },
		{ 446, 279, 47 },
		{ 446, 280, 47 },
		{ 446, 281, 47 },
		{ 446, 282, 47 },
		{ 446, 283, 47 },
		{ 446, 284, 47 },
		{ 446, 285, 47 },
		{ 446, 286, 47 },
		{ 446, 287, 47 },
		{ 446, 288, 47 },
		{ 446, 289, 47 },
		{ 446, 290, 47 },
		{ 446, 291, 47 },
		{ 446, 292, 47 },
		{ 446, 293, 47 },
		{ 446, 294, 47 },
		{ 446, 295, 47 },
		{ 446, 296, 47 },
		{ 446, 297, 47 },
		{ 446, 298, 47 },
		{ 446, 299, 47 },
		{ 446, 300, 47 },
		{ 446, 301, 47 },
		{ 446, 302, 47 },
		{ 446, 303, 47 },
		{ 446, 304, 47 },
		{ 447, 258, 47 },
		{ 447, 259, 47 },
		{ 447, 260, 47 },
		{ 447, 261, 47 },
		{ 447, 262, 47 },
		{ 447, 263, 47 },
		{ 447, 264, 47 },
		{ 447, 265, 47 },
		{ 447, 266, 47 },
		{ 447, 267, 47 },
		{ 447, 268, 47 },
		{ 447, 269, 47 },
		{ 447, 270, 47 },
		{ 447, 271, 47 },
		{ 447, 272, 47 },
		{ 447, 273, 47 },
		{ 447, 274, 47 },
		{ 447, 275, 47 },
		{ 447, 276, 47 },
		{ 447, 277, 47 },
		{ 447, 278, 47 },
		{ 447, 279, 47 },
		{ 447, 280, 47 },
		{ 447, 281, 47 },
		{ 447, 282, 47 },
		{ 447, 283, 47 },
		{ 447, 284, 47 },
		{ 447, 285, 47 },
		{ 447, 286, 47 },
		{ 447, 287, 47 },
		{ 447, 288, 47 },
		{ 447, 289, 47 },
		{ 447, 290, 47 },
		{ 447, 291, 47 },
		{ 447, 292, 47 },
		{ 447, 293, 47 },
		{ 447, 294, 47 },
		{ 447, 295, 47 },
		{ 447, 296, 47 },
		{ 447, 297, 47 },
		{ 447, 298, 47 },
		{ 447, 299, 47 },
		{ 447, 300, 47 },
		{ 447, 301, 47 },
		{ 447, 302, 47 },
		{ 447, 303, 47 },
		{ 447, 304, 47 },
		{ 448, 258, 47 },
		{ 448, 259, 47 },
		{ 448, 260, 47 },
		{ 448, 261, 47 },
		{ 448, 262, 47 },
		{ 448, 263, 47 },
		{ 448, 264, 47 },
		{ 448, 265, 47 },
		{ 448, 266, 47 },
		{ 448, 267, 47 },
		{ 448, 268, 47 },
		{ 448, 269, 47 },
		{ 448, 270, 47 },
		{ 448, 271, 47 },
		{ 448, 272, 47 },
		{ 448, 273, 47 },
		{ 448, 274, 47 },
		{ 448, 275, 47 },
		{ 448, 276, 47 },
		{ 448, 277, 47 },
		{ 448, 278, 47 },
		{ 448, 279, 47 },
		{ 448, 280, 47 },
		{ 448, 281, 47 },
		{ 448, 282, 47 },
		{ 448, 283, 47 },
		{ 448, 284, 47 },
		{ 448, 285, 47 },
		{ 448, 286, 47 },
		{ 448, 287, 47 },
		{ 448, 288, 47 },
		{ 448, 289, 47 },
		{ 448, 290, 47 },
		{ 448, 291, 47 },
		{ 448, 292, 47 },
		{ 448, 293, 47 },
		{ 448, 294, 47 },
		{ 448, 295, 47 },
		{ 448, 296, 47 },
		{ 448, 297, 47 },
		{ 448, 298, 47 },
		{ 448, 299, 47 },
		{ 448, 300, 47 },
		{ 448, 301, 47 },
		{ 448, 302, 47 },
		{ 448, 303, 47 },
		{ 448, 304, 47 },
		{ 449, 258, 47 },
		{ 449, 259, 47 },
		{ 449, 260, 47 },
		{ 449, 261, 47 },
		{ 449, 262, 47 },
		{ 449, 263, 47 },
		{ 449, 264, 47 },
		{ 449, 265, 47 },
		{ 449, 266, 47 },
		{ 449, 267, 47 },
		{ 449, 268, 47 },
		{ 449, 269, 47 },
		{ 449, 270, 47 },
		{ 449, 271, 47 },
		{ 449, 272, 47 },
		{ 449, 273, 47 },
		{ 449, 274, 47 },
		{ 449, 275, 47 },
		{ 449, 276, 47 },
		{ 449, 277, 47 },
		{ 449, 278, 47 },
		{ 449, 279, 47 },
		{ 449, 280, 47 },
		{ 449, 281, 47 },
		{ 449, 282, 47 },
		{ 449, 283, 47 },
		{ 449, 284, 47 },
		{ 449, 285, 47 },
		{ 449, 286, 47 },
		{ 449, 287, 47 },
		{ 449, 288, 47 },
		{ 449, 289, 47 },
		{ 449, 290, 47 },
		{ 449, 291, 47 },
		{ 449, 292, 47 },
		{ 449, 293, 47 },
		{ 449, 294, 47 },
		{ 449, 295, 47 },
		{ 449, 296, 47 },
		{ 449, 297, 47 },
		{ 449, 298, 47 },
		{ 449, 299, 47 },
		{ 449, 300, 47 },
		{ 449, 301, 47 },
		{ 449, 302, 47 },
		{ 449, 303, 47 },
		{ 449, 304, 47 },
		{ 450, 258, 47 },
		{ 450, 259, 47 },
		{ 450, 260, 47 },
		{ 450, 261, 47 },
		{ 450, 262, 47 },
		{ 450, 263, 47 },
		{ 450, 264, 47 },
		{ 450, 265, 47 },
		{ 450, 266, 47 },
		{ 450, 267, 47 },
		{ 450, 268, 47 },
		{ 450, 269, 47 },
		{ 450, 270, 47 },
		{ 450, 271, 47 },
		{ 450, 272, 47 },
		{ 450, 273, 47 },
		{ 450, 274, 47 },
		{ 450, 275, 47 },
		{ 450, 276, 47 },
		{ 450, 277, 47 },
		{ 450, 278, 47 },
		{ 450, 279, 47 },
		{ 450, 280, 47 },
		{ 450, 281, 47 },
		{ 450, 282, 47 },
		{ 450, 283, 47 },
		{ 450, 284, 47 },
		{ 450, 285, 47 },
		{ 450, 286, 47 },
		{ 450, 287, 47 },
		{ 450, 288, 47 },
		{ 450, 289, 47 },
		{ 450, 290, 47 },
		{ 450, 291, 47 },
		{ 450, 292, 47 },
		{ 450, 293, 47 },
		{ 450, 294, 47 },
		{ 450, 295, 47 },
		{ 450, 296, 47 },
		{ 450, 297, 47 },
		{ 450, 298, 47 },
		{ 450, 299, 47 },
		{ 450, 300, 47 },
		{ 450, 301, 47 },
		{ 450, 302, 47 },
		{ 450, 303, 47 },
		{ 450, 304, 47 },
		{ 450, 305, 47 },
		{ 451, 258, 47 },
		{ 451, 259, 47 },
		{ 451, 260, 47 },
		{ 451, 261, 47 },
		{ 451, 262, 47 },
		{ 451, 263, 47 },
		{ 451, 264, 47 },
		{ 451, 265, 47 },
		{ 451, 266, 47 },
		{ 451, 267, 47 },
		{ 451, 268, 47 },
		{ 451, 269, 47 },
		{ 451, 270, 47 },
		{ 451, 271, 47 },
		{ 451, 272, 47 },
		{ 451, 273, 47 },
		{ 451, 274, 47 },
		{ 451, 275, 47 },
		{ 451, 276, 47 },
		{ 451, 277, 47 },
		{ 451, 278, 47 },
		{ 451, 279, 47 },
		{ 451, 280, 47 },
		{ 451, 281, 47 },
		{ 451, 282, 47 },
		{ 451, 283, 47 },
		{ 451, 284, 47 },
		{ 451, 285, 47 },
		{ 451, 286, 47 },
		{ 451, 287, 47 },
		{ 451, 288, 47 },
		{ 451, 289, 47 },
		{ 451, 290, 47 },
		{ 451, 291, 47 },
		{ 451, 292, 47 },
		{ 451, 293, 47 },
		{ 451, 294, 47 },
		{ 451, 295, 47 },
		{ 451, 296, 47 },
		{ 451, 297, 47 },
		{ 451, 298, 47 },
		{ 451, 299, 47 },
		{ 451, 300, 47 },
		{ 451, 301, 47 },
		{ 451, 302, 47 },
		{ 451, 303, 47 },
		{ 451, 304, 47 },
		{ 451, 305, 47 },
		{ 452, 258, 47 },
		{ 452, 259, 47 },
		{ 452, 260, 47 },
		{ 452, 261, 47 },
		{ 452, 262, 47 },
		{ 452, 263, 47 },
		{ 452, 264, 47 },
		{ 452, 265, 47 },
		{ 452, 266, 47 },
		{ 452, 267, 47 },
		{ 452, 268, 47 },
		{ 452, 269, 47 },
		{ 452, 270, 47 },
		{ 452, 271, 47 },
		{ 452, 272, 47 },
		{ 452, 273, 47 },
		{ 452, 274, 47 },
		{ 452, 275, 47 },
		{ 452, 276, 47 },
		{ 452, 277, 47 },
		{ 452, 278, 47 },
		{ 452, 279, 47 },
		{ 452, 280, 47 },
		{ 452, 281, 47 },
		{ 452, 282, 47 },
		{ 452, 283, 47 },
		{ 452, 284, 47 },
		{ 452, 285, 47 },
		{ 452, 286, 47 },
		{ 452, 287, 47 },
		{ 452, 288, 47 },
		{ 452, 289, 47 },
		{ 452, 290, 47 },
		{ 452, 291, 47 },
		{ 452, 292, 47 },
		{ 452, 293, 47 },
		{ 452, 294, 47 },
		{ 452, 295, 47 },
		{ 452, 296, 47 },
		{ 452, 297, 47 },
		{ 452, 298, 47 },
		{ 452, 299, 47 },
		{ 452, 300, 47 },
		{ 452, 301, 47 },
		{ 452, 302, 47 },
		{ 452, 303, 47 },
		{ 452, 304, 47 },
		{ 452, 305, 47 },
		{ 453, 258, 47 },
		{ 453, 259, 47 },
		{ 453, 260, 47 },
		{ 453, 261, 47 },
		{ 453, 262, 47 },
		{ 453, 263, 47 },
		{ 453, 264, 47 },
		{ 453, 265, 47 },
		{ 453, 266, 47 },
		{ 453, 267, 47 },
		{ 453, 268, 47 },
		{ 453, 269, 47 },
		{ 453, 270, 47 },
		{ 453, 271, 47 },
		{ 453, 272, 47 },
		{ 453, 273, 47 },
		{ 453, 274, 47 },
		{ 453, 275, 47 },
		{ 453, 276, 47 },
		{ 453, 277, 47 },
		{ 453, 278, 47 },
		{ 453, 279, 47 },
		{ 453, 280, 47 },
		{ 453, 281, 47 },
		{ 453, 282, 47 },
		{ 453, 283, 47 },
		{ 453, 284, 47 },
		{ 453, 285, 47 },
		{ 453, 286, 47 },
		{ 453, 287, 47 },
		{ 453, 288, 47 },
		{ 453, 289, 47 },
		{ 453, 290, 47 },
		{ 453, 291, 47 },
		{ 453, 292, 47 },
		{ 453, 293, 47 },
		{ 453, 294, 47 },
		{ 453, 295, 47 },
		{ 453, 296, 47 },
		{ 453, 297, 47 },
		{ 453, 298, 47 },
		{ 453, 299, 47 },
		{ 453, 300, 47 },
		{ 453, 301, 47 },
		{ 453, 302, 47 },
		{ 453, 303, 47 },
		{ 453, 304, 47 },
		{ 453, 305, 47 },
		{ 454, 258, 47 },
		{ 454, 259, 47 },
		{ 454, 260, 47 },
		{ 454, 261, 47 },
		{ 454, 262, 47 },
		{ 454, 263, 47 },
		{ 454, 264, 47 },
		{ 454, 265, 47 },
		{ 454, 266, 47 },
		{ 454, 267, 47 },
		{ 454, 268, 47 },
		{ 454, 269, 47 },
		{ 454, 270, 47 },
		{ 454, 271, 47 },
		{ 454, 272, 47 },
		{ 454, 273, 47 },
		{ 454, 274, 47 },
		{ 454, 275, 47 },
		{ 454, 276, 47 },
		{ 454, 277, 47 },
		{ 454, 278, 47 },
		{ 454, 279, 47 },
		{ 454, 280, 47 },
		{ 454, 281, 47 },
		{ 454, 282, 47 },
		{ 454, 283, 47 },
		{ 454, 284, 47 },
		{ 454, 285, 47 },
		{ 454, 286, 47 },
		{ 454, 287, 47 },
		{ 454, 288, 47 },
		{ 454, 289, 47 },
		{ 454, 290, 47 },
		{ 454, 291, 47 },
		{ 454, 292, 47 },
		{ 454, 293, 47 },
		{ 454, 294, 47 },
		{ 454, 295, 47 },
		{ 454, 296, 47 },
		{ 454, 297, 47 },
		{ 454, 298, 47 },
		{ 454, 299, 47 },
		{ 454, 300, 47 },
		{ 454, 301, 47 },
		{ 454, 302, 47 },
		{ 454, 303, 47 },
		{ 454, 304, 47 },
		{ 454, 305, 47 },
		{ 454, 306, 47 },
		{ 455, 258, 47 },
		{ 455, 259, 47 },
		{ 455, 260, 47 },
		{ 455, 261, 47 },
		{ 455, 262, 47 },
		{ 455, 263, 47 },
		{ 455, 264, 47 },
		{ 455, 265, 47 },
		{ 455, 266, 47 },
		{ 455, 267, 47 },
		{ 455, 268, 47 },
		{ 455, 269, 47 },
		{ 455, 270, 47 },
		{ 455, 271, 47 },
		{ 455, 272, 47 },
		{ 455, 273, 47 },
		{ 455, 274, 47 },
		{ 455, 275, 47 },
		{ 455, 276, 47 },
		{ 455, 277, 47 },
		{ 455, 278, 47 },
		{ 455, 279, 47 },
		{ 455, 280, 47 },
		{ 455, 281, 47 },
		{ 455, 282, 47 },
		{ 455, 283, 47 },
		{ 455, 284, 47 },
		{ 455, 285, 47 },
		{ 455, 286, 47 },
		{ 455, 287, 47 },
		{ 455, 288, 47 },
		{ 455, 289, 47 },
		{ 455, 290, 47 },
		{ 455, 291, 47 },
		{ 455, 292, 47 },
		{ 455, 293, 47 },
		{ 455, 294, 47 },
		{ 455, 295, 47 },
		{ 455, 296, 47 },
		{ 455, 297, 47 },
		{ 455, 298, 47 },
		{ 455, 299, 47 },
		{ 455, 300, 47 },
		{ 455, 301, 47 },
		{ 455, 302, 47 },
		{ 455, 303, 47 },
		{ 455, 304, 47 },
		{ 455, 305, 47 },
		{ 455, 306, 47 },
		{ 456, 258, 47 },
		{ 456, 259, 47 },
		{ 456, 260, 47 },
		{ 456, 261, 47 },
		{ 456, 262, 47 },
		{ 456, 263, 47 },
		{ 456, 264, 47 },
		{ 456, 265, 47 },
		{ 456, 266, 47 },
		{ 456, 267, 47 },
		{ 456, 268, 47 },
		{ 456, 269, 47 },
		{ 456, 270, 47 },
		{ 456, 271, 47 },
		{ 456, 272, 47 },
		{ 456, 273, 47 },
		{ 456, 274, 47 },
		{ 456, 275, 47 },
		{ 456, 276, 47 },
		{ 456, 277, 47 },
		{ 456, 278, 47 },
		{ 456, 279, 47 },
		{ 456, 280, 47 },
		{ 456, 281, 47 },
		{ 456, 282, 47 },
		{ 456, 283, 47 },
		{ 456, 284, 47 },
		{ 456, 285, 47 },
		{ 456, 286, 47 },
		{ 456, 287, 47 },
		{ 456, 288, 47 },
		{ 456, 289, 47 },
		{ 456, 290, 47 },
		{ 456, 291, 47 },
		{ 456, 292, 47 },
		{ 456, 293, 47 },
		{ 456, 294, 47 },
		{ 456, 295, 47 },
		{ 456, 296, 47 },
		{ 456, 297, 47 },
		{ 456, 298, 47 },
		{ 456, 299, 47 },
		{ 456, 300, 47 },
		{ 456, 301, 47 },
		{ 456, 302, 47 },
		{ 456, 303, 47 },
		{ 456, 304, 47 },
		{ 456, 305, 47 },
		{ 456, 306, 47 },
		{ 457, 258, 47 },
		{ 457, 259, 47 },
		{ 457, 260, 47 },
		{ 457, 261, 47 },
		{ 457, 262, 47 },
		{ 457, 263, 47 },
		{ 457, 264, 47 },
		{ 457, 265, 47 },
		{ 457, 266, 47 },
		{ 457, 267, 47 },
		{ 457, 268, 47 },
		{ 457, 269, 47 },
		{ 457, 270, 47 },
		{ 457, 271, 47 },
		{ 457, 272, 47 },
		{ 457, 273, 47 },
		{ 457, 274, 47 },
		{ 457, 275, 47 },
		{ 457, 276, 47 },
		{ 457, 277, 47 },
		{ 457, 278, 47 },
		{ 457, 279, 47 },
		{ 457, 280, 47 },
		{ 457, 281, 47 },
		{ 457, 282, 47 },
		{ 457, 283, 47 },
		{ 457, 284, 47 },
		{ 457, 285, 47 },
		{ 457, 286, 47 },
		{ 457, 287, 47 },
		{ 457, 288, 47 },
		{ 457, 289, 47 },
		{ 457, 290, 47 },
		{ 457, 291, 47 },
		{ 457, 292, 47 },
		{ 457, 293, 47 },
		{ 457, 294, 47 },
		{ 457, 295, 47 },
		{ 457, 296, 47 },
		{ 457, 297, 47 },
		{ 457, 298, 47 },
		{ 457, 299, 47 },
		{ 457, 300, 47 },
		{ 457, 301, 47 },
		{ 457, 302, 47 },
		{ 457, 303, 47 },
		{ 457, 304, 47 },
		{ 457, 305, 47 },
		{ 457, 306, 47 },
		{ 458, 258, 47 },
		{ 458, 259, 47 },
		{ 458, 260, 47 },
		{ 458, 261, 47 },
		{ 458, 262, 47 },
		{ 458, 263, 47 },
		{ 458, 264, 47 },
		{ 458, 265, 47 },
		{ 458, 266, 47 },
		{ 458, 267, 47 },
		{ 458, 268, 47 },
		{ 458, 269, 47 },
		{ 458, 270, 47 },
		{ 458, 271, 47 },
		{ 458, 272, 47 },
		{ 458, 273, 47 },
		{ 458, 274, 47 },
		{ 458, 275, 47 },
		{ 458, 276, 47 },
		{ 458, 277, 47 },
		{ 458, 278, 47 },
		{ 458, 279, 47 },
		{ 458, 280, 47 },
		{ 458, 281, 47 },
		{ 458, 282, 47 },
		{ 458, 283, 47 },
		{ 458, 284, 47 },
		{ 458, 285, 47 },
		{ 458, 286, 47 },
		{ 458, 287, 47 },
		{ 458, 288, 47 },
		{ 458, 289, 47 },
		{ 458, 290, 47 },
		{ 458, 291, 47 },
		{ 458, 292, 47 },
		{ 458, 293, 47 },
		{ 458, 294, 47 },
		{ 458, 295, 47 },
		{ 458, 296, 47 },
		{ 458, 297, 47 },
		{ 458, 298, 47 },
		{ 458, 299, 47 },
		{ 458, 300, 47 },
		{ 458, 301, 47 },
		{ 458, 302, 47 },
		{ 458, 303, 47 },
		{ 458, 304, 47 },
		{ 458, 305, 47 },
		{ 458, 306, 47 },
		{ 458, 307, 47 },
		{ 459, 258, 47 },
		{ 459, 259, 47 },
		{ 459, 260, 47 },
		{ 459, 261, 47 },
		{ 459, 262, 47 },
		{ 459, 263, 47 },
		{ 459, 264, 47 },
		{ 459, 265, 47 },
		{ 459, 266, 47 },
		{ 459, 267, 47 },
		{ 459, 268, 47 },
		{ 459, 269, 47 },
		{ 459, 270, 47 },
		{ 459, 271, 47 },
		{ 459, 272, 47 },
		{ 459, 273, 47 },
		{ 459, 274, 47 },
		{ 459, 275, 47 },
		{ 459, 276, 47 },
		{ 459, 277, 47 },
		{ 459, 278, 47 },
		{ 459, 279, 47 },
		{ 459, 280, 47 },
		{ 459, 281, 47 },
		{ 459, 282, 47 },
		{ 459, 283, 47 },
		{ 459, 284, 47 },
		{ 459, 285, 47 },
		{ 459, 286, 47 },
		{ 459, 287, 47 },
		{ 459, 288, 47 },
		{ 459, 289, 47 },
		{ 459, 290, 47 },
		{ 459, 291, 47 },
		{ 459, 292, 47 },
		{ 459, 293, 47 },
		{ 459, 294, 47 },
		{ 459, 295, 47 },
		{ 459, 296, 47 },
		{ 459, 297, 47 },
		{ 459, 298, 47 },
		{ 459, 299, 47 },
		{ 459, 300, 47 },
		{ 459, 301, 47 },
		{ 459, 302, 47 },
		{ 459, 303, 47 },
		{ 459, 304, 47 },
		{ 459, 305, 47 },
		{ 459, 306, 47 },
		{ 459, 307, 47 },
		{ 460, 258, 47 },
		{ 460, 259, 47 },
		{ 460, 260, 47 },
		{ 460, 261, 47 },
		{ 460, 262, 47 },
		{ 460, 263, 47 },
		{ 460, 264, 47 },
		{ 460, 265, 47 },
		{ 460, 266, 47 },
		{ 460, 267, 47 },
		{ 460, 268, 47 },
		{ 460, 269, 47 },
		{ 460, 270, 47 },
		{ 460, 271, 47 },
		{ 460, 272, 47 },
		{ 460, 273, 47 },
		{ 460, 274, 47 },
		{ 460, 275, 47 },
		{ 460, 276, 47 },
		{ 460, 277, 47 },
		{ 460, 278, 47 },
		{ 460, 279, 47 },
		{ 460, 280, 47 },
		{ 460, 281, 47 },
		{ 460, 282, 47 },
		{ 460, 283, 47 },
		{ 460, 284, 47 },
		{ 460, 285, 47 },
		{ 460, 286, 47 },
		{ 460, 287, 47 },
		{ 460, 288, 47 },
		{ 460, 289, 47 },
		{ 460, 290, 47 },
		{ 460, 291, 47 },
		{ 460, 292, 47 },
		{ 460, 293, 47 },
		{ 460, 294, 47 },
		{ 460, 295, 47 },
		{ 460, 296, 47 },
		{ 460, 297, 47 },
		{ 460, 298, 47 },
		{ 460, 299, 47 },
		{ 460, 300, 47 },
		{ 460, 301, 47 },
		{ 460, 302, 47 },
		{ 460, 303, 47 },
		{ 460, 304, 47 },
		{ 460, 305, 47 },
		{ 460, 306, 47 },
		{ 460, 307, 47 },
		{ 461, 258, 47 },
		{ 461, 259, 47 },
		{ 461, 260, 47 },
		{ 461, 261, 47 },
		{ 461, 262, 47 },
		{ 461, 263, 47 },
		{ 461, 264, 47 },
		{ 461, 265, 47 },
		{ 461, 266, 47 },
		{ 461, 267, 47 },
		{ 461, 268, 47 },
		{ 461, 269, 47 },
		{ 461, 270, 47 },
		{ 461, 271, 47 },
		{ 461, 272, 47 },
		{ 461, 273, 47 },
		{ 461, 274, 47 },
		{ 461, 275, 47 },
		{ 461, 276, 47 },
		{ 461, 277, 47 },
		{ 461, 278, 47 },
		{ 461, 279, 47 },
		{ 461, 280, 47 },
		{ 461, 281, 47 },
		{ 461, 282, 47 },
		{ 461, 283, 47 },
		{ 461, 284, 47 },
		{ 461, 285, 47 },
		{ 461, 286, 47 },
		{ 461, 287, 47 },
		{ 461, 288, 47 },
		{ 461, 289, 47 },
		{ 461, 290, 47 },
		{ 461, 291, 47 },
		{ 461, 292, 47 },
		{ 461, 293, 47 },
		{ 461, 294, 47 },
		{ 461, 295, 47 },
		{ 461, 296, 47 },
		{ 461, 297, 47 },
		{ 461, 298, 47 },
		{ 461, 299, 47 },
		{ 461, 300, 47 },
		{ 461, 301, 47 },
		{ 461, 302, 47 },
		{ 461, 303, 47 },
		{ 461, 304, 47 },
		{ 461, 305, 47 },
		{ 461, 306, 47 },
		{ 461, 307, 47 },
		{ 462, 258, 47 },
		{ 462, 259, 47 },
		{ 462, 260, 47 },
		{ 462, 261, 47 },
		{ 462, 262, 47 },
		{ 462, 263, 47 },
		{ 462, 264, 47 },
		{ 462, 265, 47 },
		{ 462, 266, 47 },
		{ 462, 267, 47 },
		{ 462, 268, 47 },
		{ 462, 269, 47 },
		{ 462, 270, 47 },
		{ 462, 271, 47 },
		{ 462, 272, 47 },
		{ 462, 273, 47 },
		{ 462, 274, 47 },
		{ 462, 275, 47 },
		{ 462, 276, 47 },
		{ 462, 277, 47 },
		{ 462, 278, 47 },
		{ 462, 279, 47 },
		{ 462, 280, 47 },
		{ 462, 281, 47 },
		{ 462, 282, 47 },
		{ 462, 283, 47 },
		{ 462, 284, 47 },
		{ 462, 285, 47 },
		{ 462, 286, 47 },
		{ 462, 287, 47 },
		{ 462, 288, 47 },
		{ 462, 289, 47 },
		{ 462, 290, 47 },
		{ 462, 291, 47 },
		{ 462, 292, 47 },
		{ 462, 293, 47 },
		{ 462, 294, 47 },
		{ 462, 295, 47 },
		{ 462, 296, 47 },
		{ 462, 297, 47 },
		{ 462, 298, 47 },
		{ 462, 299, 47 },
		{ 462, 300, 47 },
		{ 462, 301, 47 },
		{ 462, 302, 47 },
		{ 462, 303, 47 },
		{ 462, 304, 47 },
		{ 462, 305, 47 },
		{ 462, 306, 47 },
		{ 462, 307, 47 },
		{ 462, 308, 47 },
		{ 463, 258, 47 },
		{ 463, 259, 47 },
		{ 463, 260, 47 },
		{ 463, 261, 47 },
		{ 463, 262, 47 },
		{ 463, 263, 47 },
		{ 463, 264, 47 },
		{ 463, 265, 47 },
		{ 463, 266, 47 },
		{ 463, 267, 47 },
		{ 463, 268, 47 },
		{ 463, 269, 47 },
		{ 463, 270, 47 },
		{ 463, 271, 47 },
		{ 463, 272, 47 },
		{ 463, 273, 47 },
		{ 463, 274, 47 },
		{ 463, 275, 47 },
		{ 463, 276, 47 },
		{ 463, 277, 47 },
		{ 463, 278, 47 },
		{ 463, 279, 47 },
		{ 463, 280, 47 },
		{ 463, 281, 47 },
		{ 463, 282, 47 },
		{ 463, 283, 47 },
		{ 463, 284, 47 },
		{ 463, 285, 47 },
		{ 463, 286, 47 },
		{ 463, 287, 47 },
		{ 463, 288, 47 },
		{ 463, 289, 47 },
		{ 463, 290, 47 },
		{ 463, 291, 47 },
		{ 463, 292, 47 },
		{ 463, 293, 47 },
		{ 463, 294, 47 },
		{ 463, 295, 47 },
		{ 463, 296, 47 },
		{ 463, 297, 47 },
		{ 463, 298, 47 },
		{ 463, 299, 47 },
		{ 463, 300, 47 },
		{ 463, 301, 47 },
		{ 463, 302, 47 },
		{ 463, 303, 47 },
		{ 463, 304, 47 },
		{ 463, 305, 47 },
		{ 463, 306, 47 },
		{ 463, 307, 47 },
		{ 463, 308, 47 },
		{ 464, 258, 47 },
		{ 464, 259, 47 },
		{ 464, 260, 47 },
		{ 464, 261, 47 },
		{ 464, 262, 47 },
		{ 464, 263, 47 },
		{ 464, 264, 47 },
		{ 464, 265, 47 },
		{ 464, 266, 47 },
		{ 464, 267, 47 },
		{ 464, 268, 47 },
		{ 464, 269, 47 },
		{ 464, 270, 47 },
		{ 464, 271, 47 },
		{ 464, 272, 47 },
		{ 464, 273, 47 },
		{ 464, 274, 47 },
		{ 464, 275, 47 },
		{ 464, 276, 47 },
		{ 464, 277, 47 },
		{ 464, 278, 47 },
		{ 464, 279, 47 },
		{ 464, 280, 47 },
		{ 464, 281, 47 },
		{ 464, 282, 47 },
		{ 464, 283, 47 },
		{ 464, 284, 47 },
		{ 464, 285, 47 },
		{ 464, 286, 47 },
		{ 464, 287, 47 },
		{ 464, 288, 47 },
		{ 464, 289, 47 },
		{ 464, 290, 47 },
		{ 464, 291, 47 },
		{ 464, 292, 47 },
		{ 464, 293, 47 },
		{ 464, 294, 47 },
		{ 464, 295, 47 },
		{ 464, 296, 47 },
		{ 464, 297, 47 },
		{ 464, 298, 47 },
		{ 464, 299, 47 },
		{ 464, 300, 47 },
		{ 464, 301, 47 },
		{ 464, 302, 47 },
		{ 464, 303, 47 },
		{ 464, 304, 47 },
		{ 464, 305, 47 },
		{ 464, 306, 47 },
		{ 464, 307, 47 },
		{ 464, 308, 47 },
		{ 465, 258, 47 },
		{ 465, 259, 47 },
		{ 465, 260, 47 },
		{ 465, 261, 47 },
		{ 465, 262, 47 },
		{ 465, 263, 47 },
		{ 465, 264, 47 },
		{ 465, 265, 47 },
		{ 465, 266, 47 },
		{ 465, 267, 47 },
		{ 465, 268, 47 },
		{ 465, 269, 47 },
		{ 465, 270, 47 },
		{ 465, 271, 47 },
		{ 465, 272, 47 },
		{ 465, 273, 47 },
		{ 465, 274, 47 },
		{ 465, 275, 47 },
		{ 465, 276, 47 },
		{ 465, 277, 47 },
		{ 465, 278, 47 },
		{ 465, 279, 47 },
		{ 465, 280, 47 },
		{ 465, 281, 47 },
		{ 465, 282, 47 },
		{ 465, 283, 47 },
		{ 465, 284, 47 },
		{ 465, 285, 47 },
		{ 465, 286, 47 },
		{ 465, 287, 47 },
		{ 465, 288, 47 },
		{ 465, 289, 47 },
		{ 465, 290, 47 },
		{ 465, 291, 47 },
		{ 465, 292, 47 },
		{ 465, 293, 47 },
		{ 465, 294, 47 },
		{ 465, 295, 47 },
		{ 465, 296, 47 },
		{ 465, 297, 47 },
		{ 465, 298, 47 },
		{ 465, 299, 47 },
		{ 465, 300, 47 },
		{ 465, 301, 47 },
		{ 465, 302, 47 },
		{ 465, 303, 47 },
		{ 465, 304, 47 },
		{ 465, 305, 47 },
		{ 465, 306, 47 },
		{ 465, 307, 47 },
		{ 465, 308, 47 },
		{ 466, 258, 47 },
		{ 466, 259, 47 },
		{ 466, 260, 47 },
		{ 466, 261, 47 },
		{ 466, 262, 47 },
		{ 466, 263, 47 },
		{ 466, 264, 47 },
		{ 466, 265, 47 },
		{ 466, 266, 47 },
		{ 466, 267, 47 },
		{ 466, 268, 47 },
		{ 466, 269, 47 },
		{ 466, 270, 47 },
		{ 466, 271, 47 },
		{ 466, 272, 47 },
		{ 466, 273, 47 },
		{ 466, 274, 47 },
		{ 466, 275, 47 },
		{ 466, 276, 47 },
		{ 466, 277, 47 },
		{ 466, 278, 47 },
		{ 466, 279, 47 },
		{ 466, 280, 47 },
		{ 466, 281, 47 },
		{ 466, 282, 47 },
		{ 466, 283, 47 },
		{ 466, 284, 47 },
		{ 466, 285, 47 },
		{ 466, 286, 47 },
		{ 466, 287, 47 },
		{ 466, 288, 47 },
		{ 466, 289, 47 },
		{ 466, 290, 47 },
		{ 466, 291, 47 },
		{ 466, 292, 47 },
		{ 466, 293, 47 },
		{ 466, 294, 47 },
		{ 466, 295, 47 },
		{ 466, 296, 47 },
		{ 466, 297, 47 },
		{ 466, 298, 47 },
		{ 466, 299, 47 },
		{ 466, 300, 47 },
		{ 466, 301, 47 },
		{ 466, 302, 47 },
		{ 466, 303, 47 },
		{ 466, 304, 47 },
		{ 466, 305, 47 },
		{ 466, 306, 47 },
		{ 466, 307, 47 },
		{ 466, 308, 47 },
		{ 466, 309, 47 },
		{ 467, 258, 47 },
		{ 467, 259, 47 },
		{ 467, 260, 47 },
		{ 467, 261, 47 },
		{ 467, 262, 47 },
		{ 467, 263, 47 },
		{ 467, 264, 47 },
		{ 467, 265, 47 },
		{ 467, 266, 47 },
		{ 467, 267, 47 },
		{ 467, 268, 47 },
		{ 467, 269, 47 },
		{ 467, 270, 47 },
		{ 467, 271, 47 },
		{ 467, 272, 47 },
		{ 467, 273, 47 },
		{ 467, 274, 47 },
		{ 467, 275, 47 },
		{ 467, 276, 47 },
		{ 467, 277, 47 },
		{ 467, 278, 47 },
		{ 467, 279, 47 },
		{ 467, 280, 47 },
		{ 467, 281, 47 },
		{ 467, 282, 47 },
		{ 467, 283, 47 },
		{ 467, 284, 47 },
		{ 467, 285, 47 },
		{ 467, 286, 47 },
		{ 467, 287, 47 },
		{ 467, 288, 47 },
		{ 467, 289, 47 },
		{ 467, 290, 47 },
		{ 467, 291, 47 },
		{ 467, 292, 47 },
		{ 467, 293, 47 },
		{ 467, 294, 47 },
		{ 467, 295, 47 },
		{ 467, 296, 47 },
		{ 467, 297, 47 },
		{ 467, 298, 47 },
		{ 467, 299, 47 },
		{ 467, 300, 47 },
		{ 467, 301, 47 },
		{ 467, 302, 47 },
		{ 467, 303, 47 },
		{ 467, 304, 47 },
		{ 467, 305, 47 },
		{ 467, 306, 47 },
		{ 467, 307, 47 },
		{ 467, 308, 47 },
		{ 467, 309, 47 },
		{ 468, 258, 47 },
		{ 468, 259, 47 },
		{ 468, 260, 47 },
		{ 468, 261, 47 },
		{ 468, 262, 47 },
		{ 468, 263, 47 },
		{ 468, 264, 47 },
		{ 468, 265, 47 },
		{ 468, 266, 47 },
		{ 468, 267, 47 },
		{ 468, 268, 47 },
		{ 468, 269, 47 },
		{ 468, 270, 47 },
		{ 468, 271, 47 },
		{ 468, 272, 47 },
		{ 468, 273, 47 },
		{ 468, 274, 47 },
		{ 468, 275, 47 },
		{ 468, 276, 47 },
		{ 468, 277, 47 },
		{ 468, 278, 47 },
		{ 468, 279, 47 },
		{ 468, 280, 47 },
		{ 468, 281, 47 },
		{ 468, 282, 47 },
		{ 468, 283, 47 },
		{ 468, 284, 47 },
		{ 468, 285, 47 },
		{ 468, 286, 47 },
		{ 468, 287, 47 },
		{ 468, 288, 47 },
		{ 468, 289, 47 },
		{ 468, 290, 47 },
		{ 468, 291, 47 },
		{ 468, 292, 47 },
		{ 468, 293, 47 },
		{ 468, 294, 47 },
		{ 468, 295, 47 },
		{ 468, 296, 47 },
		{ 468, 297, 47 },
		{ 468, 298, 47 },
		{ 468, 299, 47 },
		{ 468, 300, 47 },
		{ 468, 301, 47 },
		{ 468, 302, 47 },
		{ 468, 303, 47 },
		{ 468, 304, 47 },
		{ 468, 305, 47 },
		{ 468, 306, 47 },
		{ 468, 307, 47 },
		{ 468, 308, 47 },
		{ 468, 309, 47 },
		{ 469, 258, 47 },
		{ 469, 259, 47 },
		{ 469, 260, 47 },
		{ 469, 261, 47 },
		{ 469, 262, 47 },
		{ 469, 263, 47 },
		{ 469, 264, 47 },
		{ 469, 265, 47 },
		{ 469, 266, 47 },
		{ 469, 267, 47 },
		{ 469, 268, 47 },
		{ 469, 269, 47 },
		{ 469, 270, 47 },
		{ 469, 271, 47 },
		{ 469, 272, 47 },
		{ 469, 273, 47 },
		{ 469, 274, 47 },
		{ 469, 275, 47 },
		{ 469, 276, 47 },
		{ 469, 277, 47 },
		{ 469, 278, 47 },
		{ 469, 279, 47 },
		{ 469, 280, 47 },
		{ 469, 281, 47 },
		{ 469, 282, 47 },
		{ 469, 283, 47 },
		{ 469, 284, 47 },
		{ 469, 285, 47 },
		{ 469, 286, 47 },
		{ 469, 287, 47 },
		{ 469, 288, 47 },
		{ 469, 289, 47 },
		{ 469, 290, 47 },
		{ 469, 291, 47 },
		{ 469, 292, 47 },
		{ 469, 293, 47 },
		{ 469, 294, 47 },
		{ 469, 295, 47 },
		{ 469, 296, 47 },
		{ 469, 297, 47 },
		{ 469, 298, 47 },
		{ 469, 299, 47 },
		{ 469, 300, 47 },
		{ 469, 301, 47 },
		{ 469, 302, 47 },
		{ 469, 303, 47 },
		{ 469, 304, 47 },
		{ 469, 305, 47 },
		{ 469, 306, 47 },
		{ 469, 307, 47 },
		{ 469, 308, 47 },
		{ 469, 309, 47 },
		{ 470, 258, 47 },
		{ 470, 259, 47 },
		{ 470, 260, 47 },
		{ 470, 261, 47 },
		{ 470, 262, 47 },
		{ 470, 263, 47 },
		{ 470, 264, 47 },
		{ 470, 265, 47 },
		{ 470, 266, 47 },
		{ 470, 267, 47 },
		{ 470, 268, 47 },
		{ 470, 269, 47 },
		{ 470, 270, 47 },
		{ 470, 271, 47 },
		{ 470, 272, 47 },
		{ 470, 273, 47 },
		{ 470, 274, 47 },
		{ 470, 275, 47 },
		{ 470, 276, 47 },
		{ 470, 277, 47 },
		{ 470, 278, 47 },
		{ 470, 279, 47 },
		{ 470, 280, 47 },
		{ 470, 281, 47 },
		{ 470, 282, 47 },
		{ 470, 283, 47 },
		{ 470, 284, 47 },
		{ 470, 285, 47 },
		{ 470, 286, 47 },
		{ 470, 287, 47 },
		{ 470, 288, 47 },
		{ 470, 289, 47 },
		{ 470, 290, 47 },
		{ 470, 291, 47 },
		{ 470, 292, 47 },
		{ 470, 293, 47 },
		{ 470, 294, 47 },
		{ 470, 295, 47 },
		{ 470, 296, 47 },
		{ 470, 297, 47 },
		{ 470, 298, 47 },
		{ 470, 299, 47 },
		{ 470, 300, 47 },
		{ 470, 301, 47 },
		{ 470, 302, 47 },
		{ 470, 303, 47 },
		{ 470, 304, 47 },
		{ 470, 305, 47 },
		{ 470, 306, 47 },
		{ 470, 307, 47 },
		{ 470, 308, 47 },
		{ 470, 309, 47 },
		{ 470, 310, 47 },
		{ 471, 258, 47 },
		{ 471, 259, 47 },
		{ 471, 260, 47 },
		{ 471, 261, 47 },
		{ 471, 262, 47 },
		{ 471, 263, 47 },
		{ 471, 264, 47 },
		{ 471, 265, 47 },
		{ 471, 266, 47 },
		{ 471, 267, 47 },
		{ 471, 268, 47 },
		{ 471, 269, 47 },
		{ 471, 270, 47 },
		{ 471, 271, 47 },
		{ 471, 272, 47 },
		{ 471, 273, 47 },
		{ 471, 274, 47 },
		{ 471, 275, 47 },
		{ 471, 276, 47 },
		{ 471, 277, 47 },
		{ 471, 278, 47 },
		{ 471, 279, 47 },
		{ 471, 280, 47 },
		{ 471, 281, 47 },
		{ 471, 282, 47 },
		{ 471, 283, 47 },
		{ 471, 284, 47 },
		{ 471, 285, 47 },
		{ 471, 286, 47 },
		{ 471, 287, 47 },
		{ 471, 288, 47 },
		{ 471, 289, 47 },
		{ 471, 290, 47 },
		{ 471, 291, 47 },
		{ 471, 292, 47 },
		{ 471, 293, 47 },
		{ 471, 294, 47 },
		{ 471, 295, 47 },
		{ 471, 296, 47 },
		{ 471, 297, 47 },
		{ 471, 298, 47 },
		{ 471, 299, 47 },
		{ 471, 300, 47 },
		{ 471, 301, 47 },
		{ 471, 302, 47 },
		{ 471, 303, 47 },
		{ 471, 304, 47 },
		{ 471, 305, 47 },
		{ 471, 306, 47 },
		{ 471, 307, 47 },
		{ 471, 308, 47 },
		{ 471, 309, 47 },
		{ 471, 310, 47 },
		{ 472, 258, 47 },
		{ 472, 259, 47 },
		{ 472, 260, 47 },
		{ 472, 261, 47 },
		{ 472, 262, 47 },
		{ 472, 263, 47 },
		{ 472, 264, 47 },
		{ 472, 265, 47 },
		{ 472, 266, 47 },
		{ 472, 267, 47 },
		{ 472, 268, 47 },
		{ 472, 269, 47 },
		{ 472, 270, 47 },
		{ 472, 271, 47 },
		{ 472, 272, 47 },
		{ 472, 273, 47 },
		{ 472, 274, 47 },
		{ 472, 275, 47 },
		{ 472, 276, 47 },
		{ 472, 277, 47 },
		{ 472, 278, 47 },
		{ 472, 279, 47 },
		{ 472, 280, 47 },
		{ 472, 281, 47 },
		{ 472, 282, 47 },
		{ 472, 283, 47 },
		{ 472, 284, 47 },
		{ 472, 285, 47 },
		{ 472, 286, 47 },
		{ 472, 287, 47 },
		{ 472, 288, 47 },
		{ 472, 289, 47 },
		{ 472, 290, 47 },
		{ 472, 291, 47 },
		{ 472, 292, 47 },
		{ 472, 293, 47 },
		{ 472, 294, 47 },
		{ 472, 295, 47 },
		{ 472, 296, 47 },
		{ 472, 297, 47 },
		{ 472, 298, 47 },
		{ 472, 299, 47 },
		{ 472, 300, 47 },
		{ 472, 301, 47 },
		{ 472, 302, 47 },
		{ 472, 303, 47 },
		{ 472, 304, 47 },
		{ 472, 305, 47 },
		{ 472, 306, 47 },
		{ 472, 307, 47 },
		{ 472, 308, 47 },
		{ 472, 309, 47 },
		{ 472, 310, 47 },
		{ 473, 258, 47 },
		{ 473, 259, 47 },
		{ 473, 260, 47 },
		{ 473, 261, 47 },
		{ 473, 262, 47 },
		{ 473, 263, 47 },
		{ 473, 264, 47 },
		{ 473, 265, 47 },
		{ 473, 266, 47 },
		{ 473, 267, 47 },
		{ 473, 268, 47 },
		{ 473, 269, 47 },
		{ 473, 270, 47 },
		{ 473, 271, 47 },
		{ 473, 272, 47 },
		{ 473, 273, 47 },
		{ 473, 274, 47 },
		{ 473, 275, 47 },
		{ 473, 276, 47 },
		{ 473, 277, 47 },
		{ 473, 278, 47 },
		{ 473, 279, 47 },
		{ 473, 280, 47 },
		{ 473, 281, 47 },
		{ 473, 282, 47 },
		{ 473, 283, 47 },
		{ 473, 284, 47 },
		{ 473, 285, 47 },
		{ 473, 286, 47 },
		{ 473, 287, 47 },
		{ 473, 288, 47 },
		{ 473, 289, 47 },
		{ 473, 290, 47 },
		{ 473, 291, 47 },
		{ 473, 292, 47 },
		{ 473, 293, 47 },
		{ 473, 294, 47 },
		{ 473, 295, 47 },
		{ 473, 296, 47 },
		{ 473, 297, 47 },
		{ 473, 298, 47 },
		{ 473, 299, 47 },
		{ 473, 300, 47 },
		{ 473, 301, 47 },
		{ 473, 302, 47 },
		{ 473, 303, 47 },
		{ 473, 304, 47 },
		{ 473, 305, 47 },
		{ 473, 306, 47 },
		{ 473, 307, 47 },
		{ 473, 308, 47 },
		{ 473, 309, 47 },
		{ 473, 310, 47 },
		{ 474, 258, 47 },
		{ 474, 259, 47 },
		{ 474, 260, 47 },
		{ 474, 261, 47 },
		{ 474, 262, 47 },
		{ 474, 263, 47 },
		{ 474, 264, 47 },
		{ 474, 265, 47 },
		{ 474, 266, 47 },
		{ 474, 267, 47 },
		{ 474, 268, 47 },
		{ 474, 269, 47 },
		{ 474, 270, 47 },
		{ 474, 271, 47 },
		{ 474, 272, 47 },
		{ 474, 273, 47 },
		{ 474, 274, 47 },
		{ 474, 275, 47 },
		{ 474, 276, 47 },
		{ 474, 277, 47 },
		{ 474, 278, 47 },
		{ 474, 279, 47 },
		{ 474, 280, 47 },
		{ 474, 281, 47 },
		{ 474, 282, 47 },
		{ 474, 283, 47 },
		{ 474, 284, 47 },
		{ 474, 285, 47 },
		{ 474, 286, 47 },
		{ 474, 287, 47 },
		{ 474, 288, 47 },
		{ 474, 289, 47 },
		{ 474, 290, 47 },
		{ 474, 291, 47 },
		{ 474, 292, 47 },
		{ 474, 293, 47 },
		{ 474, 294, 47 },
		{ 474, 295, 47 },
		{ 474, 296, 47 },
		{ 474, 297, 47 },
		{ 474, 298, 47 },
		{ 474, 299, 47 },
		{ 474, 300, 47 },
		{ 474, 301, 47 },
		{ 474, 302, 47 },
		{ 474, 303, 47 },
		{ 474, 304, 47 },
		{ 474, 305, 47 },
		{ 474, 306, 47 },
		{ 474, 307, 47 },
		{ 474, 308, 47 },
		{ 474, 309, 47 },
		{ 474, 310, 47 },
		{ 474, 311, 47 },
		{ 475, 258, 47 },
		{ 475, 259, 47 },
		{ 475, 260, 47 },
		{ 475, 261, 47 },
		{ 475, 262, 47 },
		{ 475, 263, 47 },
		{ 475, 264, 47 },
		{ 475, 265, 47 },
		{ 475, 266, 47 },
		{ 475, 267, 47 },
		{ 475, 268, 47 },
		{ 475, 269, 47 },
		{ 475, 270, 47 },
		{ 475, 271, 47 },
		{ 475, 272, 47 },
		{ 475, 273, 47 },
		{ 475, 274, 47 },
		{ 475, 275, 47 },
		{ 475, 276, 47 },
		{ 475, 277, 47 },
		{ 475, 278, 47 },
		{ 475, 279, 47 },
		{ 475, 280, 47 },
		{ 475, 281, 47 },
		{ 475, 282, 47 },
		{ 475, 283, 47 },
		{ 475, 284, 47 },
		{ 475, 285, 47 },
		{ 475, 286, 47 },
		{ 475, 287, 47 },
		{ 475, 288, 47 },
		{ 475, 289, 47 },
		{ 475, 290, 47 },
		{ 475, 291, 47 },
		{ 475, 292, 47 },
		{ 475, 293, 47 },
		{ 475, 294, 47 },
		{ 475, 295, 47 },
		{ 475, 296, 47 },
		{ 475, 297, 47 },
		{ 475, 298, 47 },
		{ 475, 299, 47 },
		{ 475, 300, 47 },
		{ 475, 301, 47 },
		{ 475, 302, 47 },
		{ 475, 303, 47 },
		{ 475, 304, 47 },
		{ 475, 305, 47 },
		{ 475, 306, 47 },
		{ 475, 307, 47 },
		{ 475, 308, 47 },
		{ 475, 309, 47 },
		{ 475, 310, 47 },
		{ 475, 311, 47 },
		{ 476, 258, 47 },
		{ 476, 259, 47 },
		{ 476, 260, 47 },
		{ 476, 261, 47 },
		{ 476, 262, 47 },
		{ 476, 263, 47 },
		{ 476, 264, 47 },
		{ 476, 265, 47 },
		{ 476, 266, 47 },
		{ 476, 267, 47 },
		{ 476, 268, 47 },
		{ 476, 269, 47 },
		{ 476, 270, 47 },
		{ 476, 271, 47 },
		{ 476, 272, 47 },
		{ 476, 273, 47 },
		{ 476, 274, 47 },
		{ 476, 275, 47 },
		{ 476, 276, 47 },
		{ 476, 277, 47 },
		{ 476, 278, 47 },
		{ 476, 279, 47 },
		{ 476, 280, 47 },
		{ 476, 281, 47 },
		{ 476, 282, 47 },
		{ 476, 283, 47 },
		{ 476, 284, 47 },
		{ 476, 285, 47 },
		{ 476, 286, 47 },
		{ 476, 287, 47 },
		{ 476, 288, 47 },
		{ 476, 289, 47 },
		{ 476, 290, 47 },
		{ 476, 291, 47 },
		{ 476, 292, 47 },
		{ 476, 293, 47 },
		{ 476, 294, 47 },
		{ 476, 295, 47 },
		{ 476, 296, 47 },
		{ 476, 297, 47 },
		{ 476, 298, 47 },
		{ 476, 299, 47 },
		{ 476, 300, 47 },
		{ 476, 301, 47 },
		{ 476, 302, 47 },
		{ 476, 303, 47 },
		{ 476, 304, 47 },
		{ 476, 305, 47 },
		{ 476, 306, 47 },
		{ 476, 307, 47 },
		{ 476, 308, 47 },
		{ 476, 309, 47 },
		{ 476, 310, 47 },
		{ 476, 311, 47 },
		{ 477, 258, 47 },
		{ 477, 259, 47 },
		{ 477, 260, 47 },
		{ 477, 261, 47 },
		{ 477, 262, 47 },
		{ 477, 263, 47 },
		{ 477, 264, 47 },
		{ 477, 265, 47 },
		{ 477, 266, 47 },
		{ 477, 267, 47 },
		{ 477, 268, 47 },
		{ 477, 269, 47 },
		{ 477, 270, 47 },
		{ 477, 271, 47 },
		{ 477, 272, 47 },
		{ 477, 273, 47 },
		{ 477, 274, 47 },
		{ 477, 275, 47 },
		{ 477, 276, 47 },
		{ 477, 277, 47 },
		{ 477, 278, 47 },
		{ 477, 279, 47 },
		{ 477, 280, 47 },
		{ 477, 281, 47 },
		{ 477, 282, 47 },
		{ 477, 283, 47 },
		{ 477, 284, 47 },
		{ 477, 285, 47 },
		{ 477, 286, 47 },
		{ 477, 287, 47 },
		{ 477, 288, 47 },
		{ 477, 289, 47 },
		{ 477, 290, 47 },
		{ 477, 291, 47 },
		{ 477, 292, 47 },
		{ 477, 293, 47 },
		{ 477, 294, 47 },
		{ 477, 295, 47 },
		{ 477, 296, 47 },
		{ 477, 297, 47 },
		{ 477, 298, 47 },
		{ 477, 299, 47 },
		{ 477, 300, 47 },
		{ 477, 301, 47 },
		{ 477, 302, 47 },
		{ 477, 303, 47 },
		{ 477, 304, 47 },
		{ 477, 305, 47 },
		{ 477, 306, 47 },
		{ 477, 307, 47 },
		{ 477, 308, 47 },
		{ 477, 309, 47 },
		{ 477, 310, 47 },
		{ 477, 311, 47 },
		{ 478, 258, 47 },
		{ 478, 259, 47 },
		{ 478, 260, 47 },
		{ 478, 261, 47 },
		{ 478, 262, 47 },
		{ 478, 263, 47 },
		{ 478, 264, 47 },
		{ 478, 265, 47 },
		{ 478, 266, 47 },
		{ 478, 267, 47 },
		{ 478, 268, 47 },
		{ 478, 269, 47 },
		{ 478, 270, 47 },
		{ 478, 271, 47 },
		{ 478, 272, 47 },
		{ 478, 273, 47 },
		{ 478, 274, 47 },
		{ 478, 275, 47 },
		{ 478, 276, 47 },
		{ 478, 277, 47 },
		{ 478, 278, 47 },
		{ 478, 279, 47 },
		{ 478, 280, 47 },
		{ 478, 281, 47 },
		{ 478, 282, 47 },
		{ 478, 283, 47 },
		{ 478, 284, 47 },
		{ 478, 285, 47 },
		{ 478, 286, 47 },
		{ 478, 287, 47 },
		{ 478, 288, 47 },
		{ 478, 289, 47 },
		{ 478, 290, 47 },
		{ 478, 291, 47 },
		{ 478, 292, 47 },
		{ 478, 293, 47 },
		{ 478, 294, 47 },
		{ 478, 295, 47 },
		{ 478, 296, 47 },
		{ 478, 297, 47 },
		{ 478, 298, 47 },
		{ 478, 299, 47 },
		{ 478, 300, 47 },
		{ 478, 301, 47 },
		{ 478, 302, 47 },
		{ 478, 303, 47 },
		{ 478, 304, 47 },
		{ 478, 305, 47 },
		{ 478, 306, 47 },
		{ 478, 307, 47 },
		{ 478, 308, 47 },
		{ 478, 309, 47 },
		{ 478, 310, 47 },
		{ 478, 311, 47 },
		{ 478, 312, 47 },
		{ 479, 258, 47 },
		{ 479, 259, 47 },
		{ 479, 260, 47 },
		{ 479, 261, 47 },
		{ 479, 262, 47 },
		{ 479, 263, 47 },
		{ 479, 264, 47 },
		{ 479, 265, 47 },
		{ 479, 266, 47 },
		{ 479, 267, 47 },
		{ 479, 268, 47 },
		{ 479, 269, 47 },
		{ 479, 270, 47 },
		{ 479, 271, 47 },
		{ 479, 272, 47 },
		{ 479, 273, 47 },
		{ 479, 274, 47 },
		{ 479, 275, 47 },
		{ 479, 276, 47 },
		{ 479, 277, 47 },
		{ 479, 278, 47 },
		{ 479, 279, 47 },
		{ 479, 280, 47 },
		{ 479, 281, 47 },
		{ 479, 282, 47 },
		{ 479, 283, 47 },
		{ 479, 284, 47 },
		{ 479, 285, 47 },
		{ 479, 286, 47 },
		{ 479, 287, 47 },
		{ 479, 288, 47 },
		{ 479, 289, 47 },
		{ 479, 290, 47 },
		{ 479, 291, 47 },
		{ 479, 292, 47 },
		{ 479, 293, 47 },
		{ 479, 294, 47 },
		{ 479, 295, 47 },
		{ 479, 296, 47 },
		{ 479, 297, 47 },
		{ 479, 298, 47 },
		{ 479, 299, 47 },
		{ 479, 300, 47 },
		{ 479, 301, 47 },
		{ 479, 302, 47 },
		{ 479, 303, 47 },
		{ 479, 304, 47 },
		{ 479, 305, 47 },
		{ 479, 306, 47 },
		{ 479, 307, 47 },
		{ 479, 308, 47 },
		{ 479, 309, 47 },
		{ 479, 310, 47 },
		{ 479, 311, 47 },
		{ 479, 312, 47 },
		{ 480, 258, 47 },
		{ 480, 259, 47 },
		{ 480, 260, 47 },
		{ 480, 261, 47 },
		{ 480, 262, 47 },
		{ 480, 263, 47 },
		{ 480, 264, 47 },
		{ 480, 265, 47 },
		{ 480, 266, 47 },
		{ 480, 267, 47 },
		{ 480, 268, 47 },
		{ 480, 269, 47 },
		{ 480, 270, 47 },
		{ 480, 271, 47 },
		{ 480, 272, 47 },
		{ 480, 273, 47 },
		{ 480, 274, 47 },
		{ 480, 275, 47 },
		{ 480, 276, 47 },
		{ 480, 277, 47 },
		{ 480, 278, 47 },
		{ 480, 279, 47 },
		{ 480, 280, 47 },
		{ 480, 281, 47 },
		{ 480, 282, 47 },
		{ 480, 283, 47 },
		{ 480, 284, 47 },
		{ 480, 285, 47 },
		{ 480, 286, 47 },
		{ 480, 287, 47 },
		{ 480, 288, 47 },
		{ 480, 289, 47 },
		{ 480, 290, 47 },
		{ 480, 291, 47 },
		{ 480, 292, 47 },
		{ 480, 293, 47 },
		{ 480, 294, 47 },
		{ 480, 295, 47 },
		{ 480, 296, 47 },
		{ 480, 297, 47 },
		{ 480, 298, 47 },
		{ 480, 299, 47 },
		{ 480, 300, 47 },
		{ 480, 301, 47 },
		{ 480, 302, 47 },
		{ 480, 303, 47 },
		{ 480, 304, 47 },
		{ 480, 305, 47 },
		{ 480, 306, 47 },
		{ 480, 307, 47 },
		{ 480, 308, 47 },
		{ 480, 309, 47 },
		{ 480, 310, 47 },
		{ 480, 311, 47 },
		{ 480, 312, 47 },
		{ 481, 258, 47 },
		{ 481, 259, 47 },
		{ 481, 260, 47 },
		{ 481, 261, 47 },
		{ 481, 262, 47 },
		{ 481, 263, 47 },
		{ 481, 264, 47 },
		{ 481, 265, 47 },
		{ 481, 266, 47 },
		{ 481, 267, 47 },
		{ 481, 268, 47 },
		{ 481, 269, 47 },
		{ 481, 270, 47 },
		{ 481, 271, 47 },
		{ 481, 272, 47 },
		{ 481, 273, 47 },
		{ 481, 274, 47 },
		{ 481, 275, 47 },
		{ 481, 276, 47 },
		{ 481, 277, 47 },
		{ 481, 278, 47 },
		{ 481, 279, 47 },
		{ 481, 280, 47 },
		{ 481, 281, 47 },
		{ 481, 282, 47 },
		{ 481, 283, 47 },
		{ 481, 284, 47 },
		{ 481, 285, 47 },
		{ 481, 286, 47 },
		{ 481, 287, 47 },
		{ 481, 288, 47 },
		{ 481, 289, 47 },
		{ 481, 290, 47 },
		{ 481, 291, 47 },
		{ 481, 292, 47 },
		{ 481, 293, 47 },
		{ 481, 294, 47 },
		{ 481, 295, 47 },
		{ 481, 296, 47 },
		{ 481, 297, 47 },
		{ 481, 298, 47 },
		{ 481, 299, 47 },
		{ 481, 300, 47 },
		{ 481, 301, 47 },
		{ 481, 302, 47 },
		{ 481, 303, 47 },
		{ 481, 304, 47 },
		{ 481, 305, 47 },
		{ 481, 306, 47 },
		{ 481, 307, 47 },
		{ 481, 308, 47 },
		{ 481, 309, 47 },
		{ 481, 310, 47 },
		{ 481, 311, 47 },
		{ 481, 312, 47 },
		{ 482, 258, 47 },
		{ 482, 259, 47 },
		{ 482, 260, 47 },
		{ 482, 261, 47 },
		{ 482, 262, 47 },
		{ 482, 263, 47 },
		{ 482, 264, 47 },
		{ 482, 265, 47 },
		{ 482, 266, 47 },
		{ 482, 267, 47 },
		{ 482, 268, 47 },
		{ 482, 269, 47 },
		{ 482, 270, 47 },
		{ 482, 271, 47 },
		{ 482, 272, 47 },
		{ 482, 273, 47 },
		{ 482, 274, 47 },
		{ 482, 275, 47 },
		{ 482, 276, 47 },
		{ 482, 277, 47 },
		{ 482, 278, 47 },
		{ 482, 279, 47 },
		{ 482, 280, 47 },
		{ 482, 281, 47 },
		{ 482, 282, 47 },
		{ 482, 283, 47 },
		{ 482, 284, 47 },
		{ 482, 285, 47 },
		{ 482, 286, 47 },
		{ 482, 287, 47 },
		{ 482, 288, 47 },
		{ 482, 289, 47 },
		{ 482, 290, 47 },
		{ 482, 291, 47 },
		{ 482, 292, 47 },
		{ 482, 293, 47 },
		{ 482, 294, 47 },
		{ 482, 295, 47 },
		{ 482, 296, 47 },
		{ 482, 297, 47 },
		{ 482, 298, 47 },
		{ 482, 299, 47 },
		{ 482, 300, 47 },
		{ 482, 301, 47 },
		{ 482, 302, 47 },
		{ 482, 303, 47 },
		{ 482, 304, 47 },
		{ 482, 305, 47 },
		{ 482, 306, 47 },
		{ 482, 307, 47 },
		{ 482, 308, 47 },
		{ 482, 309, 47 },
		{ 482, 310, 47 },
		{ 482, 311, 47 },
		{ 482, 312, 47 },
		{ 482, 313, 47 },
		{ 483, 258, 47 },
		{ 483, 259, 47 },
		{ 483, 260, 47 },
		{ 483, 261, 47 },
		{ 483, 262, 47 },
		{ 483, 263, 47 },
		{ 483, 264, 47 },
		{ 483, 265, 47 },
		{ 483, 266, 47 },
		{ 483, 267, 47 },
		{ 483, 268, 47 },
		{ 483, 269, 47 },
		{ 483, 270, 47 },
		{ 483, 271, 47 },
		{ 483, 272, 47 },
		{ 483, 273, 47 },
		{ 483, 274, 47 },
		{ 483, 275, 47 },
		{ 483, 276, 47 },
		{ 483, 277, 47 },
		{ 483, 278, 47 },
		{ 483, 279, 47 },
		{ 483, 280, 47 },
		{ 483, 281, 47 },
		{ 483, 282, 47 },
		{ 483, 283, 47 },
		{ 483, 284, 47 },
		{ 483, 285, 47 },
		{ 483, 286, 47 },
		{ 483, 287, 47 },
		{ 483, 288, 47 },
		{ 483, 289, 47 },
		{ 483, 290, 47 },
		{ 483, 291, 47 },
		{ 483, 292, 47 },
		{ 483, 293, 47 },
		{ 483, 294, 47 },
		{ 483, 295, 47 },
		{ 483, 296, 47 },
		{ 483, 297, 47 },
		{ 483, 298, 47 },
		{ 483, 299, 47 },
		{ 483, 300, 47 },
		{ 483, 301, 47 },
		{ 483, 302, 47 },
		{ 483, 303, 47 },
		{ 483, 304, 47 },
		{ 483, 305, 47 },
		{ 483, 306, 47 },
		{ 483, 307, 47 },
		{ 483, 308, 47 },
		{ 483, 309, 47 },
		{ 483, 310, 47 },
		{ 483, 311, 47 },
		{ 483, 312, 47 },
		{ 483, 313, 47 },
		{ 484, 258, 47 },
		{ 484, 259, 47 },
		{ 484, 260, 47 },
		{ 484, 261, 47 },
		{ 484, 262, 47 },
		{ 484, 263, 47 },
		{ 484, 264, 47 },
		{ 484, 265, 47 },
		{ 484, 266, 47 },
		{ 484, 267, 47 },
		{ 484, 268, 47 },
		{ 484, 269, 47 },
		{ 484, 270, 47 },
		{ 484, 271, 47 },
		{ 484, 272, 47 },
		{ 484, 273, 47 },
		{ 484, 274, 47 },
		{ 484, 275, 47 },
		{ 484, 276, 47 },
		{ 484, 277, 47 },
		{ 484, 278, 47 },
		{ 484, 279, 47 },
		{ 484, 280, 47 },
		{ 484, 281, 47 },
		{ 484, 282, 47 },
		{ 484, 283, 47 },
		{ 484, 284, 47 },
		{ 484, 285, 47 },
		{ 484, 286, 47 },
		{ 484, 287, 47 },
		{ 484, 288, 47 },
		{ 484, 289, 47 },
		{ 484, 290, 47 },
		{ 484, 291, 47 },
		{ 484, 292, 47 },
		{ 484, 293, 47 },
		{ 484, 294, 47 },
		{ 484, 295, 47 },
		{ 484, 296, 47 },
		{ 484, 297, 47 },
		{ 484, 298, 47 },
		{ 484, 299, 47 },
		{ 484, 300, 47 },
		{ 484, 301, 47 },
		{ 484, 302, 47 },
		{ 484, 303, 47 },
		{ 484, 304, 47 },
		{ 484, 305, 47 },
		{ 484, 306, 47 },
		{ 484, 307, 47 },
		{ 484, 308, 47 },
		{ 484, 309, 47 },
		{ 484, 310, 47 },
		{ 484, 311, 47 },
		{ 484, 312, 47 },
		{ 484, 313, 47 },
		{ 485, 258, 47 },
		{ 485, 259, 47 },
		{ 485, 260, 47 },
		{ 485, 261, 47 },
		{ 485, 262, 47 },
		{ 485, 263, 47 },
		{ 485, 264, 47 },
		{ 485, 265, 47 },
		{ 485, 266, 47 },
		{ 485, 267, 47 },
		{ 485, 268, 47 },
		{ 485, 269, 47 },
		{ 485, 270, 47 },
		{ 485, 271, 47 },
		{ 485, 272, 47 },
		{ 485, 273, 47 },
		{ 485, 274, 47 },
		{ 485, 275, 47 },
		{ 485, 276, 47 },
		{ 485, 277, 47 },
		{ 485, 278, 47 },
		{ 485, 279, 47 },
		{ 485, 280, 47 },
		{ 485, 281, 47 },
		{ 485, 282, 47 },
		{ 485, 283, 47 },
		{ 485, 284, 47 },
		{ 485, 285, 47 },
		{ 485, 286, 47 },
		{ 485, 287, 47 },
		{ 485, 288, 47 },
		{ 485, 289, 47 },
		{ 485, 290, 47 },
		{ 485, 291, 47 },
		{ 485, 292, 47 },
		{ 485, 293, 47 },
		{ 485, 294, 47 },
		{ 485, 295, 47 },
		{ 485, 296, 47 },
		{ 485, 297, 47 },
		{ 485, 298, 47 },
		{ 485, 299, 47 },
		{ 485, 300, 47 },
		{ 485, 301, 47 },
		{ 485, 302, 47 },
		{ 485, 303, 47 },
		{ 485, 304, 47 },
		{ 485, 305, 47 },
		{ 485, 306, 47 },
		{ 485, 307, 47 },
		{ 485, 308, 47 },
		{ 485, 309, 47 },
		{ 485, 310, 47 },
		{ 485, 311, 47 },
		{ 485, 312, 47 },
		{ 485, 313, 47 },
		{ 486, 258, 47 },
		{ 486, 259, 47 },
		{ 486, 260, 47 },
		{ 486, 261, 47 },
		{ 486, 262, 47 },
		{ 486, 263, 47 },
		{ 486, 264, 47 },
		{ 486, 265, 47 },
		{ 486, 266, 47 },
		{ 486, 267, 47 },
		{ 486, 268, 47 },
		{ 486, 269, 47 },
		{ 486, 270, 47 },
		{ 486, 271, 47 },
		{ 486, 272, 47 },
		{ 486, 273, 47 },
		{ 486, 274, 47 },
		{ 486, 275, 47 },
		{ 486, 276, 47 },
		{ 486, 277, 47 },
		{ 486, 278, 47 },
		{ 486, 279, 47 },
		{ 486, 280, 47 },
		{ 486, 281, 47 },
		{ 486, 282, 47 },
		{ 486, 283, 47 },
		{ 486, 284, 47 },
		{ 486, 285, 47 },
		{ 486, 286, 47 },
		{ 486, 287, 47 },
		{ 486, 288, 47 },
		{ 486, 289, 47 },
		{ 486, 290, 47 },
		{ 486, 291, 47 },
		{ 486, 292, 47 },
		{ 486, 293, 47 },
		{ 486, 294, 47 },
		{ 486, 295, 47 },
		{ 486, 296, 47 },
		{ 486, 297, 47 },
		{ 486, 298, 47 },
		{ 486, 299, 47 },
		{ 486, 300, 47 },
		{ 486, 301, 47 },
		{ 486, 302, 47 },
		{ 486, 303, 47 },
		{ 486, 304, 47 },
		{ 486, 305, 47 },
		{ 486, 306, 47 },
		{ 486, 307, 47 },
		{ 486, 308, 47 },
		{ 486, 309, 47 },
		{ 486, 310, 47 },
		{ 486, 311, 47 },
		{ 486, 312, 47 },
		{ 486, 313, 47 },
		{ 486, 314, 47 },
		{ 487, 258, 47 },
		{ 487, 259, 47 },
		{ 487, 260, 47 },
		{ 487, 261, 47 },
		{ 487, 262, 47 },
		{ 487, 263, 47 },
		{ 487, 264, 47 },
		{ 487, 265, 47 },
		{ 487, 266, 47 },
		{ 487, 267, 47 },
		{ 487, 268, 47 },
		{ 487, 269, 47 },
		{ 487, 270, 47 },
		{ 487, 271, 47 },
		{ 487, 272, 47 },
		{ 487, 273, 47 },
		{ 487, 274, 47 },
		{ 487, 275, 47 },
		{ 487, 276, 47 },
		{ 487, 277, 47 },
		{ 487, 278, 47 },
		{ 487, 279, 47 },
		{ 487, 280, 47 },
		{ 487, 281, 47 },
		{ 487, 282, 47 },
		{ 487, 283, 47 },
		{ 487, 284, 47 },
		{ 487, 285, 47 },
		{ 487, 286, 47 },
		{ 487, 287, 47 },
		{ 487, 288, 47 },
		{ 487, 289, 47 },
		{ 487, 290, 47 },
		{ 487, 291, 47 },
		{ 487, 292, 47 },
		{ 487, 293, 47 },
		{ 487, 294, 47 },
		{ 487, 295, 47 },
		{ 487, 296, 47 },
		{ 487, 297, 47 },
		{ 487, 298, 47 },
		{ 487, 299, 47 },
		{ 487, 300, 47 },
		{ 487, 301, 47 },
		{ 487, 302, 47 },
		{ 487, 303, 47 },
		{ 487, 304, 47 },
		{ 487, 305, 47 },
		{ 487, 306, 47 },
		{ 487, 307, 47 },
		{ 487, 308, 47 },
		{ 487, 309, 47 },
		{ 487, 310, 47 },
		{ 487, 311, 47 },
		{ 487, 312, 47 },
		{ 487, 313, 47 },
		{ 487, 314, 47 },
		{ 488, 258, 47 },
		{ 488, 259, 47 },
		{ 488, 260, 47 },
		{ 488, 261, 47 },
		{ 488, 262, 47 },
		{ 488, 263, 47 },
		{ 488, 264, 47 },
		{ 488, 265, 47 },
		{ 488, 266, 47 },
		{ 488, 267, 47 },
		{ 488, 268, 47 },
		{ 488, 269, 47 },
		{ 488, 270, 47 },
		{ 488, 271, 47 },
		{ 488, 272, 47 },
		{ 488, 273, 47 },
		{ 488, 274, 47 },
		{ 488, 275, 47 },
		{ 488, 276, 47 },
		{ 488, 277, 47 },
		{ 488, 278, 47 },
		{ 488, 279, 47 },
		{ 488, 280, 47 },
		{ 488, 281, 47 },
		{ 488, 282, 47 },
		{ 488, 283, 47 },
		{ 488, 284, 47 },
		{ 488, 285, 47 },
		{ 488, 286, 47 },
		{ 488, 287, 47 },
		{ 488, 288, 47 },
		{ 488, 289, 47 },
		{ 488, 290, 47 },
		{ 488, 291, 47 },
		{ 488, 292, 47 },
		{ 488, 293, 47 },
		{ 488, 294, 47 },
		{ 488, 295, 47 },
		{ 488, 296, 47 },
		{ 488, 297, 47 },
		{ 488, 298, 47 },
		{ 488, 299, 47 },
		{ 488, 300, 47 },
		{ 488, 301, 47 },
		{ 488, 302, 47 },
		{ 488, 303, 47 },
		{ 488, 304, 47 },
		{ 488, 305, 47 },
		{ 488, 306, 47 },
		{ 488, 307, 47 },
		{ 488, 308, 47 },
		{ 488, 309, 47 },
		{ 488, 310, 47 },
		{ 488, 311, 47 },
		{ 488, 312, 47 },
		{ 488, 313, 47 },
		{ 488, 314, 47 },
		{ 489, 258, 47 },
		{ 489, 259, 47 },
		{ 489, 260, 47 },
		{ 489, 261, 47 },
		{ 489, 262, 47 },
		{ 489, 263, 47 },
		{ 489, 264, 47 },
		{ 489, 265, 47 },
		{ 489, 266, 47 },
		{ 489, 267, 47 },
		{ 489, 268, 47 },
		{ 489, 269, 47 },
		{ 489, 270, 47 },
		{ 489, 271, 47 },
		{ 489, 272, 47 },
		{ 489, 273, 47 },
		{ 489, 274, 47 },
		{ 489, 275, 47 },
		{ 489, 276, 47 },
		{ 489, 277, 47 },
		{ 489, 278, 47 },
		{ 489, 279, 47 },
		{ 489, 280, 47 },
		{ 489, 281, 47 },
		{ 489, 282, 47 },
		{ 489, 283, 47 },
		{ 489, 284, 47 },
		{ 489, 285, 47 },
		{ 489, 286, 47 },
		{ 489, 287, 47 },
		{ 489, 288, 47 },
		{ 489, 289, 47 },
		{ 489, 290, 47 },
		{ 489, 291, 47 },
		{ 489, 292, 47 },
		{ 489, 293, 47 },
		{ 489, 294, 47 },
		{ 489, 295, 47 },
		{ 489, 296, 47 },
		{ 489, 297, 47 },
		{ 489, 298, 47 },
		{ 489, 299, 47 },
		{ 489, 300, 47 },
		{ 489, 301, 47 },
		{ 489, 302, 47 },
		{ 489, 303, 47 },
		{ 489, 304, 47 },
		{ 489, 305, 47 },
		{ 489, 306, 47 },
		{ 489, 307, 47 },
		{ 489, 308, 47 },
		{ 489, 309, 47 },
		{ 489, 310, 47 },
		{ 489, 311, 47 },
		{ 489, 312, 47 },
		{ 489, 313, 47 },
		{ 489, 314, 47 },
		{ 490, 258, 47 },
		{ 490, 259, 47 },
		{ 490, 260, 47 },
		{ 490, 261, 47 },
		{ 490, 262, 47 },
		{ 490, 263, 47 },
		{ 490, 264, 47 },
		{ 490, 265, 47 },
		{ 490, 266, 47 },
		{ 490, 267, 47 },
		{ 490, 268, 47 },
		{ 490, 269, 47 },
		{ 490, 270, 47 },
		{ 490, 271, 47 },
		{ 490, 272, 47 },
		{ 490, 273, 47 },
		{ 490, 274, 47 },
		{ 490, 275, 47 },
		{ 490, 276, 47 },
		{ 490, 277, 47 },
		{ 490, 278, 47 },
		{ 490, 279, 47 },
		{ 490, 280, 47 },
		{ 490, 281, 47 },
		{ 490, 282, 47 },
		{ 490, 283, 47 },
		{ 490, 284, 47 },
		{ 490, 285, 47 },
		{ 490, 286, 47 },
		{ 490, 287, 47 },
		{ 490, 288, 47 },
		{ 490, 289, 47 },
		{ 490, 290, 47 },
		{ 490, 291, 47 },
		{ 490, 292, 47 },
		{ 490, 293, 47 },
		{ 490, 294, 47 },
		{ 490, 295, 47 },
		{ 490, 296, 47 },
		{ 490, 297, 47 },
		{ 490, 298, 47 },
		{ 490, 299, 47 },
		{ 490, 300, 47 },
		{ 490, 301, 47 },
		{ 490, 302, 47 },
		{ 490, 303, 47 },
		{ 490, 304, 47 },
		{ 490, 305, 47 },
		{ 490, 306, 47 },
		{ 490, 307, 47 },
		{ 490, 308, 47 },
		{ 490, 309, 47 },
		{ 490, 310, 47 },
		{ 490, 311, 47 },
		{ 490, 312, 47 },
		{ 490, 313, 47 },
		{ 490, 314, 47 },
		{ 490, 315, 47 },
		{ 491, 258, 47 },
		{ 491, 259, 47 },
		{ 491, 260, 47 },
		{ 491, 261, 47 },
		{ 491, 262, 47 },
		{ 491, 263, 47 },
		{ 491, 264, 47 },
		{ 491, 265, 47 },
		{ 491, 266, 47 },
		{ 491, 267, 47 },
		{ 491, 268, 47 },
		{ 491, 269, 47 },
		{ 491, 270, 47 },
		{ 491, 271, 47 },
		{ 491, 272, 47 },
		{ 491, 273, 47 },
		{ 491, 274, 47 },
		{ 491, 275, 47 },
		{ 491, 276, 47 },
		{ 491, 277, 47 },
		{ 491, 278, 47 },
		{ 491, 279, 47 },
		{ 491, 280, 47 },
		{ 491, 281, 47 },
		{ 491, 282, 47 },
		{ 491, 283, 47 },
		{ 491, 284, 47 },
		{ 491, 285, 47 },
		{ 491, 286, 47 },
		{ 491, 287, 47 },
		{ 491, 288, 47 },
		{ 491, 289, 47 },
		{ 491, 290, 47 },
		{ 491, 291, 47 },
		{ 491, 292, 47 },
		{ 491, 293, 47 },
		{ 491, 294, 47 },
		{ 491, 295, 47 },
		{ 491, 296, 47 },
		{ 491, 297, 47 },
		{ 491, 298, 47 },
		{ 491, 299, 47 },
		{ 491, 300, 47 },
		{ 491, 301, 47 },
		{ 491, 302, 47 },
		{ 491, 303, 47 },
		{ 491, 304, 47 },
		{ 491, 305, 47 },
		{ 491, 306, 47 },
		{ 491, 307, 47 },
		{ 491, 308, 47 },
		{ 491, 309, 47 },
		{ 491, 310, 47 },
		{ 491, 311, 47 },
		{ 491, 312, 47 },
		{ 491, 313, 47 },
		{ 491, 314, 47 },
		{ 491, 315, 47 },
		{ 492, 258, 47 },
		{ 492, 259, 47 },
		{ 492, 260, 47 },
		{ 492, 261, 47 },
		{ 492, 262, 47 },
		{ 492, 263, 47 },
		{ 492, 264, 47 },
		{ 492, 265, 47 },
		{ 492, 266, 47 },
		{ 492, 267, 47 },
		{ 492, 268, 47 },
		{ 492, 269, 47 },
		{ 492, 270, 47 },
		{ 492, 271, 47 },
		{ 492, 272, 47 },
		{ 492, 273, 47 },
		{ 492, 274, 47 },
		{ 492, 275, 47 },
		{ 492, 276, 47 },
		{ 492, 277, 47 },
		{ 492, 278, 47 },
		{ 492, 279, 47 },
		{ 492, 280, 47 },
		{ 492, 281, 47 },
		{ 492, 282, 47 },
		{ 492, 283, 47 },
		{ 492, 284, 47 },
		{ 492, 285, 47 },
		{ 492, 286, 47 },
		{ 492, 287, 47 },
		{ 492, 288, 47 },
		{ 492, 289, 47 },
		{ 492, 290, 47 },
		{ 492, 291, 47 },
		{ 492, 292, 47 },
		{ 492, 293, 47 },
		{ 492, 294, 47 },
		{ 492, 295, 47 },
		{ 492, 296, 47 },
		{ 492, 297, 47 },
		{ 492, 298, 47 },
		{ 492, 299, 47 },
		{ 492, 300, 47 },
		{ 492, 301, 47 },
		{ 492, 302, 47 },
		{ 492, 303, 47 },
		{ 492, 304, 47 },
		{ 492, 305, 47 },
		{ 492, 306, 47 },
		{ 492, 307, 47 },
		{ 492, 308, 47 },
		{ 492, 309, 47 },
		{ 492, 310, 47 },
		{ 492, 311, 47 },
		{ 492, 312, 47 },
		{ 492, 313, 47 },
		{ 492, 314, 47 },
		{ 492, 315, 47 },
		{ 493, 258, 47 },
		{ 493, 259, 47 },
		{ 493, 260, 47 },
		{ 493, 261, 47 },
		{ 493, 262, 47 },
		{ 493, 263, 47 },
		{ 493, 264, 47 },
		{ 493, 265, 47 },
		{ 493, 266, 47 },
		{ 493, 267, 47 },
		{ 493, 268, 47 },
		{ 493, 269, 47 },
		{ 493, 270, 47 },
		{ 493, 271, 47 },
		{ 493, 272, 47 },
		{ 493, 273, 47 },
		{ 493, 274, 47 },
		{ 493, 275, 47 },
		{ 493, 276, 47 },
		{ 493, 277, 47 },
		{ 493, 278, 47 },
		{ 493, 279, 47 },
		{ 493, 280, 47 },
		{ 493, 281, 47 },
		{ 493, 282, 47 },
		{ 493, 283, 47 },
		{ 493, 284, 47 },
		{ 493, 285, 47 },
		{ 493, 286, 47 },
		{ 493, 287, 47 },
		{ 493, 288, 47 },
		{ 493, 289, 47 },
		{ 493, 290, 47 },
		{ 493, 291, 47 },
		{ 493, 292, 47 },
		{ 493, 293, 47 },
		{ 493, 294, 47 },
		{ 493, 295, 47 },
		{ 493, 296, 47 },
		{ 493, 297, 47 },
		{ 493, 298, 47 },
		{ 493, 299, 47 },
		{ 493, 300, 47 },
		{ 493, 301, 47 },
		{ 493, 302, 47 },
		{ 493, 303, 47 },
		{ 493, 304, 47 },
		{ 493, 305, 47 },
		{ 493, 306, 47 },
		{ 493, 307, 47 },
		{ 493, 308, 47 },
		{ 493, 309, 47 },
		{ 493, 310, 47 },
		{ 493, 311, 47 },
		{ 493, 312, 47 },
		{ 493, 313, 47 },
		{ 493, 314, 47 },
		{ 493, 315, 47 },
		{ 494, 258, 47 },
		{ 494, 259, 47 },
		{ 494, 260, 47 },
		{ 494, 261, 47 },
		{ 494, 262, 47 },
		{ 494, 263, 47 },
		{ 494, 264, 47 },
		{ 494, 265, 47 },
		{ 494, 266, 47 },
		{ 494, 267, 47 },
		{ 494, 268, 47 },
		{ 494, 269, 47 },
		{ 494, 270, 47 },
		{ 494, 271, 47 },
		{ 494, 272, 47 },
		{ 494, 273, 47 },
		{ 494, 274, 47 },
		{ 494, 275, 47 },
		{ 494, 276, 47 },
		{ 494, 277, 47 },
		{ 494, 278, 47 },
		{ 494, 279, 47 },
		{ 494, 280, 47 },
		{ 494, 281, 47 },
		{ 494, 282, 47 },
		{ 494, 283, 47 },
		{ 494, 284, 47 },
		{ 494, 285, 47 },
		{ 494, 286, 47 },
		{ 494, 287, 47 },
		{ 494, 288, 47 },
		{ 494, 289, 47 },
		{ 494, 290, 47 },
		{ 494, 291, 47 },
		{ 494, 292, 47 },
		{ 494, 293, 47 },
		{ 494, 294, 47 },
		{ 494, 295, 47 },
		{ 494, 296, 47 },
		{ 494, 297, 47 },
		{ 494, 298, 47 },
		{ 494, 299, 47 },
		{ 494, 300, 47 },
		{ 494, 301, 47 },
		{ 494, 302, 47 },
		{ 494, 303, 47 },
		{ 494, 304, 47 },
		{ 494, 305, 47 },
		{ 494, 306, 47 },
		{ 494, 307, 47 },
		{ 494, 308, 47 },
		{ 494, 309, 47 },
		{ 494, 310, 47 },
		{ 494, 311, 47 },
		{ 494, 312, 47 },
		{ 494, 313, 47 },
		{ 494, 314, 47 },
		{ 494, 315, 47 },
		{ 494, 316, 47 },
		{ 495, 258, 47 },
		{ 495, 259, 47 },
		{ 495, 260, 47 },
		{ 495, 261, 47 },
		{ 495, 262, 47 },
		{ 495, 263, 47 },
		{ 495, 264, 47 },
		{ 495, 265, 47 },
		{ 495, 266, 47 },
		{ 495, 267, 47 },
		{ 495, 268, 47 },
		{ 495, 269, 47 },
		{ 495, 270, 47 },
		{ 495, 271, 47 },
		{ 495, 272, 47 },
		{ 495, 273, 47 },
		{ 495, 274, 47 },
		{ 495, 275, 47 },
		{ 495, 276, 47 },
		{ 495, 277, 47 },
		{ 495, 278, 47 },
		{ 495, 279, 47 },
		{ 495, 280, 47 },
		{ 495, 281, 47 },
		{ 495, 282, 47 },
		{ 495, 283, 47 },
		{ 495, 284, 47 },
		{ 495, 285, 47 },
		{ 495, 286, 47 },
		{ 495, 287, 47 },
		{ 495, 288, 47 },
		{ 495, 289, 47 },
		{ 495, 290, 47 },
		{ 495, 291, 47 },
		{ 495, 292, 47 },
		{ 495, 293, 47 },
		{ 495, 294, 47 },
		{ 495, 295, 47 },
		{ 495, 296, 47 },
		{ 495, 297, 47 },
		{ 495, 298, 47 },
		{ 495, 299, 47 },
		{ 495, 300, 47 },
		{ 495, 301, 47 },
		{ 495, 302, 47 },
		{ 495, 303, 47 },
		{ 495, 304, 47 },
		{ 495, 305, 47 },
		{ 495, 306, 47 },
		{ 495, 307, 47 },
		{ 495, 308, 47 },
		{ 495, 309, 47 },
		{ 495, 310, 47 },
		{ 495, 311, 47 },
		{ 495, 312, 47 },
		{ 495, 313, 47 },
		{ 495, 314, 47 },
		{ 495, 315, 47 },
		{ 495, 316, 47 },
		{ 496, 258, 47 },
		{ 496, 259, 47 },
		{ 496, 260, 47 },
		{ 496, 261, 47 },
		{ 496, 262, 47 },
		{ 496, 263, 47 },
		{ 496, 264, 47 },
		{ 496, 265, 47 },
		{ 496, 266, 47 },
		{ 496, 267, 47 },
		{ 496, 268, 47 },
		{ 496, 269, 47 },
		{ 496, 270, 47 },
		{ 496, 271, 47 },
		{ 496, 272, 47 },
		{ 496, 273, 47 },
		{ 496, 274, 47 },
		{ 496, 275, 47 },
		{ 496, 276, 47 },
		{ 496, 277, 47 },
		{ 496, 278, 47 },
		{ 496, 279, 47 },
		{ 496, 280, 47 },
		{ 496, 281, 47 },
		{ 496, 282, 47 },
		{ 496, 283, 47 },
		{ 496, 284, 47 },
		{ 496, 285, 47 },
		{ 496, 286, 47 },
		{ 496, 287, 47 },
		{ 496, 288, 47 },
		{ 496, 289, 47 },
		{ 496, 290, 47 },
		{ 496, 291, 47 },
		{ 496, 292, 47 },
		{ 496, 293, 47 },
		{ 496, 294, 47 },
		{ 496, 295, 47 },
		{ 496, 296, 47 },
		{ 496, 297, 47 },
		{ 496, 298, 47 },
		{ 496, 299, 47 },
		{ 496, 300, 47 },
		{ 496, 301, 47 },
		{ 496, 302, 47 },
		{ 496, 303, 47 },
		{ 496, 304, 47 },
		{ 496, 305, 47 },
		{ 496, 306, 47 },
		{ 496, 307, 47 },
		{ 496, 308, 47 },
		{ 496, 309, 47 },
		{ 496, 310, 47 },
		{ 496, 311, 47 },
		{ 496, 312, 47 },
		{ 496, 313, 47 },
		{ 496, 314, 47 },
		{ 496, 315, 47 },
		{ 496, 316, 47 },
		{ 497, 258, 47 },
		{ 497, 259, 47 },
		{ 497, 260, 47 },
		{ 497, 261, 47 },
		{ 497, 262, 47 },
		{ 497, 263, 47 },
		{ 497, 264, 47 },
		{ 497, 265, 47 },
		{ 497, 266, 47 },
		{ 497, 267, 47 },
		{ 497, 268, 47 },
		{ 497, 269, 47 },
		{ 497, 270, 47 },
		{ 497, 271, 47 },
		{ 497, 272, 47 },
		{ 497, 273, 47 },
		{ 497, 274, 47 },
		{ 497, 275, 47 },
		{ 497, 276, 47 },
		{ 497, 277, 47 },
		{ 497, 278, 47 },
		{ 497, 279, 47 },
		{ 497, 280, 47 },
		{ 497, 281, 47 },
		{ 497, 282, 47 },
		{ 497, 283, 47 },
		{ 497, 284, 47 },
		{ 497, 285, 47 },
		{ 497, 286, 47 },
		{ 497, 287, 47 },
		{ 497, 288, 47 },
		{ 497, 289, 47 },
		{ 497, 290, 47 },
		{ 497, 291, 47 },
		{ 497, 292, 47 },
		{ 497, 293, 47 },
		{ 497, 294, 47 },
		{ 497, 295, 47 },
		{ 497, 296, 47 },
		{ 497, 297, 47 },
		{ 497, 298, 47 },
		{ 497, 299, 47 },
		{ 497, 300, 47 },
		{ 497, 301, 47 },
		{ 497, 302, 47 },
		{ 497, 303, 47 },
		{ 497, 304, 47 },
		{ 497, 305, 47 },
		{ 497, 306, 47 },
		{ 497, 307, 47 },
		{ 497, 308, 47 },
		{ 497, 309, 47 },
		{ 497, 310, 47 },
		{ 497, 311, 47 },
		{ 497, 312, 47 },
		{ 497, 313, 47 },
		{ 497, 314, 47 },
		{ 497, 315, 47 },
		{ 497, 316, 47 },
		{ 498, 258, 47 },
		{ 498, 259, 47 },
		{ 498, 260, 47 },
		{ 498, 261, 47 },
		{ 498, 262, 47 },
		{ 498, 263, 47 },
		{ 498, 264, 47 },
		{ 498, 265, 47 },
		{ 498, 266, 47 },
		{ 498, 267, 47 },
		{ 498, 268, 47 },
		{ 498, 269, 47 },
		{ 498, 270, 47 },
		{ 498, 271, 47 },
		{ 498, 272, 47 },
		{ 498, 273, 47 },
		{ 498, 274, 47 },
		{ 498, 275, 47 },
		{ 498, 276, 47 },
		{ 498, 277, 47 },
		{ 498, 278, 47 },
		{ 498, 279, 47 },
		{ 498, 280, 47 },
		{ 498, 281, 47 },
		{ 498, 282, 47 },
		{ 498, 283, 47 },
		{ 498, 284, 47 },
		{ 498, 285, 47 },
		{ 498, 286, 47 },
		{ 498, 287, 47 },
		{ 498, 288, 47 },
		{ 498, 289, 47 },
		{ 498, 290, 47 },
		{ 498, 291, 47 },
		{ 498, 292, 47 },
		{ 498, 293, 47 },
		{ 498, 294, 47 },
		{ 498, 295, 47 },
		{ 498, 296, 47 },
		{ 498, 297, 47 },
		{ 498, 298, 47 },
		{ 498, 299, 47 },
		{ 498, 300, 47 },
		{ 498, 301, 47 },
		{ 498, 302, 47 },
		{ 498, 303, 47 },
		{ 498, 304, 47 },
		{ 498, 305, 47 },
		{ 498, 306, 47 },
		{ 498, 307, 47 },
		{ 498, 308, 47 },
		{ 498, 309, 47 },
		{ 498, 310, 47 },
		{ 498, 311, 47 },
		{ 498, 312, 47 },
		{ 498, 313, 47 },
		{ 498, 314, 47 },
		{ 498, 315, 47 },
		{ 498, 316, 47 },
		{ 498, 317, 47 },
		{ 499, 258, 47 },
		{ 499, 259, 47 },
		{ 499, 260, 47 },
		{ 499, 261, 47 },
		{ 499, 262, 47 },
		{ 499, 263, 47 },
		{ 499, 264, 47 },
		{ 499, 265, 47 },
		{ 499, 266, 47 },
		{ 499, 267, 47 },
		{ 499, 268, 47 },
		{ 499, 269, 47 },
		{ 499, 270, 47 },
		{ 499, 271, 47 },
		{ 499, 272, 47 },
		{ 499, 273, 47 },
		{ 499, 274, 47 },
		{ 499, 275, 47 },
		{ 499, 276, 47 },
		{ 499, 277, 47 },
		{ 499, 278, 47 },
		{ 499, 279, 47 },
		{ 499, 280, 47 },
		{ 499, 281, 47 },
		{ 499, 282, 47 },
		{ 499, 283, 47 },
		{ 499, 284, 47 },
		{ 499, 285, 47 },
		{ 499, 286, 47 },
		{ 499, 287, 47 },
		{ 499, 288, 47 },
		{ 499, 289, 47 },
		{ 499, 290, 47 },
		{ 499, 291, 47 },
		{ 499, 292, 47 },
		{ 499, 293, 47 },
		{ 499, 294, 47 },
		{ 499, 295, 47 },
		{ 499, 296, 47 },
		{ 499, 297, 47 },
		{ 499, 298, 47 },
		{ 499, 299, 47 },
		{ 499, 300, 47 },
		{ 499, 301, 47 },
		{ 499, 302, 47 },
		{ 499, 303, 47 },
		{ 499, 304, 47 },
		{ 499, 305, 47 },
		{ 499, 306, 47 },
		{ 499, 307, 47 },
		{ 499, 308, 47 },
		{ 499, 309, 47 },
		{ 499, 310, 47 },
		{ 499, 311, 47 },
		{ 499, 312, 47 },
		{ 499, 313, 47 },
		{ 499, 314, 47 },
		{ 499, 315, 47 },
		{ 499, 316, 47 },
		{ 499, 317, 47 },
		{ 500, 258, 47 },
		{ 500, 259, 47 },
		{ 500, 260, 47 },
		{ 500, 261, 47 },
		{ 500, 262, 47 },
		{ 500, 263, 47 },
		{ 500, 264, 47 },
		{ 500, 265, 47 },
		{ 500, 266, 47 },
		{ 500, 267, 47 },
		{ 500, 268, 47 },
		{ 500, 269, 47 },
		{ 500, 270, 47 },
		{ 500, 271, 47 },
		{ 500, 272, 47 },
		{ 500, 273, 47 },
		{ 500, 274, 47 },
		{ 500, 275, 47 },
		{ 500, 276, 47 },
		{ 500, 277, 47 },
		{ 500, 278, 47 },
		{ 500, 279, 47 },
		{ 500, 280, 47 },
		{ 500, 281, 47 },
		{ 500, 282, 47 },
		{ 500, 283, 47 },
		{ 500, 284, 47 },
		{ 500, 285, 47 },
		{ 500, 286, 47 },
		{ 500, 287, 47 },
		{ 500, 288, 47 },
		{ 500, 289, 47 },
		{ 500, 290, 47 },
		{ 500, 291, 47 },
		{ 500, 292, 47 },
		{ 500, 293, 47 },
		{ 500, 294, 47 },
		{ 500, 295, 47 },
		{ 500, 296, 47 },
		{ 500, 297, 47 },
		{ 500, 298, 47 },
		{ 500, 299, 47 },
		{ 500, 300, 47 },
		{ 500, 301, 47 },
		{ 500, 302, 47 },
		{ 500, 303, 47 },
		{ 500, 304, 47 },
		{ 500, 305, 47 },
		{ 500, 306, 47 },
		{ 500, 307, 47 },
		{ 500, 308, 47 },
		{ 500, 309, 47 },
		{ 500, 310, 47 },
		{ 500, 311, 47 },
		{ 500, 312, 47 },
		{ 500, 313, 47 },
		{ 500, 314, 47 },
		{ 500, 315, 47 },
		{ 500, 316, 47 },
		{ 500, 317, 47 },
		{ 501, 258, 47 },
		{ 501, 259, 47 },
		{ 501, 260, 47 },
		{ 501, 261, 47 },
		{ 501, 262, 47 },
		{ 501, 263, 47 },
		{ 501, 264, 47 },
		{ 501, 265, 47 },
		{ 501, 266, 47 },
		{ 501, 267, 47 },
		{ 501, 268, 47 },
		{ 501, 269, 47 },
		{ 501, 270, 47 },
		{ 501, 271, 47 },
		{ 501, 272, 47 },
		{ 501, 273, 47 },
		{ 501, 274, 47 },
		{ 501, 275, 47 },
		{ 501, 276, 47 },
		{ 501, 277, 47 },
		{ 501, 278, 47 },
		{ 501, 279, 47 },
		{ 501, 280, 47 },
		{ 501, 281, 47 },
		{ 501, 282, 47 },
		{ 501, 283, 47 },
		{ 501, 284, 47 },
		{ 501, 285, 47 },
		{ 501, 286, 47 },
		{ 501, 287, 47 },
		{ 501, 288, 47 },
		{ 501, 289, 47 },
		{ 501, 290, 47 },
		{ 501, 291, 47 },
		{ 501, 292, 47 },
		{ 501, 293, 47 },
		{ 501, 294, 47 },
		{ 501, 295, 47 },
		{ 501, 296, 47 },
		{ 501, 297, 47 },
		{ 501, 298, 47 },
		{ 501, 299, 47 },
		{ 501, 300, 47 },
		{ 501, 301, 47 },
		{ 501, 302, 47 },
		{ 501, 303, 47 },
		{ 501, 304, 47 },
		{ 501, 305, 47 },
		{ 501, 306, 47 },
		{ 501, 307, 47 },
		{ 501, 308, 47 },
		{ 501, 309, 47 },
		{ 501, 310, 47 },
		{ 501, 311, 47 },
		{ 501, 312, 47 },
		{ 501, 313, 47 },
		{ 501, 314, 47 },
		{ 501, 315, 47 },
		{ 501, 316, 47 },
		{ 501, 317, 47 },
		{ 502, 258, 47 },
		{ 502, 259, 47 },
		{ 502, 260, 47 },
		{ 502, 261, 47 },
		{ 502, 262, 47 },
		{ 502, 263, 47 },
		{ 502, 264, 47 },
		{ 502, 265, 47 },
		{ 502, 266, 47 },
		{ 502, 267, 47 },
		{ 502, 268, 47 },
		{ 502, 269, 47 },
		{ 502, 270, 47 },
		{ 502, 271, 47 },
		{ 502, 272, 47 },
		{ 502, 273, 47 },
		{ 502, 274, 47 },
		{ 502, 275, 47 },
		{ 502, 276, 47 },
		{ 502, 277, 47 },
		{ 502, 278, 47 },
		{ 502, 279, 47 },
		{ 502, 280, 47 },
		{ 502, 281, 47 },
		{ 502, 282, 47 },
		{ 502, 283, 47 },
		{ 502, 284, 47 },
		{ 502, 285, 47 },
		{ 502, 286, 47 },
		{ 502, 287, 47 },
		{ 502, 288, 47 },
		{ 502, 289, 47 },
		{ 502, 290, 47 },
		{ 502, 291, 47 },
		{ 502, 292, 47 },
		{ 502, 293, 47 },
		{ 502, 294, 47 },
		{ 502, 295, 47 },
		{ 502, 296, 47 },
		{ 502, 297, 47 },
		{ 502, 298, 47 },
		{ 502, 299, 47 },
		{ 502, 300, 47 },
		{ 502, 301, 47 },
		{ 502, 302, 47 },
		{ 502, 303, 47 },
		{ 502, 304, 47 },
		{ 502, 305, 47 },
		{ 502, 306, 47 },
		{ 502, 307, 47 },
		{ 502, 308, 47 },
		{ 502, 309, 47 },
		{ 502, 310, 47 },
		{ 502, 311, 47 },
		{ 502, 312, 47 },
		{ 502, 313, 47 },
		{ 502, 314, 47 },
		{ 502, 315, 47 },
		{ 502, 316, 47 },
		{ 502, 317, 47 },
		{ 502, 318, 47 },
		{ 503, 258, 47 },
		{ 503, 259, 47 },
		{ 503, 260, 47 },
		{ 503, 261, 47 },
		{ 503, 262, 47 },
		{ 503, 263, 47 },
		{ 503, 264, 47 },
		{ 503, 265, 47 },
		{ 503, 266, 47 },
		{ 503, 267, 47 },
		{ 503, 268, 47 },
		{ 503, 269, 47 },
		{ 503, 270, 47 },
		{ 503, 271, 47 },
		{ 503, 272, 47 },
		{ 503, 273, 47 },
		{ 503, 274, 47 },
		{ 503, 275, 47 },
		{ 503, 276, 47 },
		{ 503, 277, 47 },
		{ 503, 278, 47 },
		{ 503, 279, 47 },
		{ 503, 280, 47 },
		{ 503, 281, 47 },
		{ 503, 282, 47 },
		{ 503, 283, 47 },
		{ 503, 284, 47 },
		{ 503, 285, 47 },
		{ 503, 286, 47 },
		{ 503, 287, 47 },
		{ 503, 288, 47 },
		{ 503, 289, 47 },
		{ 503, 290, 47 },
		{ 503, 291, 47 },
		{ 503, 292, 47 },
		{ 503, 293, 47 },
		{ 503, 294, 47 },
		{ 503, 295, 47 },
		{ 503, 296, 47 },
		{ 503, 297, 47 },
		{ 503, 298, 47 },
		{ 503, 299, 47 },
		{ 503, 300, 47 },
		{ 503, 301, 47 },
		{ 503, 302, 47 },
		{ 503, 303, 47 },
		{ 503, 304, 47 },
		{ 503, 305, 47 },
		{ 503, 306, 47 },
		{ 503, 307, 47 },
		{ 503, 308, 47 },
		{ 503, 309, 47 },
		{ 503, 310, 47 },
		{ 503, 311, 47 },
		{ 503, 312, 47 },
		{ 503, 313, 47 },
		{ 503, 314, 47 },
		{ 503, 315, 47 },
		{ 503, 316, 47 },
		{ 503, 317, 47 },
		{ 503, 318, 47 },
		{ 504, 258, 47 },
		{ 504, 259, 47 },
		{ 504, 260, 47 },
		{ 504, 261, 47 },
		{ 504, 262, 47 },
		{ 504, 263, 47 },
		{ 504, 264, 47 },
		{ 504, 265, 47 },
		{ 504, 266, 47 },
		{ 504, 267, 47 },
		{ 504, 268, 47 },
		{ 504, 269, 47 },
		{ 504, 270, 47 },
		{ 504, 271, 47 },
		{ 504, 272, 47 },
		{ 504, 273, 47 },
		{ 504, 274, 47 },
		{ 504, 275, 47 },
		{ 504, 276, 47 },
		{ 504, 277, 47 },
		{ 504, 278, 47 },
		{ 504, 279, 47 },
		{ 504, 280, 47 },
		{ 504, 281, 47 },
		{ 504, 282, 47 },
		{ 504, 283, 47 },
		{ 504, 284, 47 },
		{ 504, 285, 47 },
		{ 504, 286, 47 },
		{ 504, 287, 47 },
		{ 504, 288, 47 },
		{ 504, 289, 47 },
		{ 504, 290, 47 },
		{ 504, 291, 47 },
		{ 504, 292, 47 },
		{ 504, 293, 47 },
		{ 504, 294, 47 },
		{ 504, 295, 47 },
		{ 504, 296, 47 },
		{ 504, 297, 47 },
		{ 504, 298, 47 },
		{ 504, 299, 47 },
		{ 504, 300, 47 },
		{ 504, 301, 47 },
		{ 504, 302, 47 },
		{ 504, 303, 47 },
		{ 504, 304, 47 },
		{ 504, 305, 47 },
		{ 504, 306, 47 },
		{ 504, 307, 47 },
		{ 504, 308, 47 },
		{ 504, 309, 47 },
		{ 504, 310, 47 },
		{ 504, 311, 47 },
		{ 504, 312, 47 },
		{ 504, 313, 47 },
		{ 504, 314, 47 },
		{ 504, 315, 47 },
		{ 504, 316, 47 },
		{ 504, 317, 47 },
		{ 504, 318, 47 },
		{ 505, 258, 47 },
		{ 505, 259, 47 },
		{ 505, 260, 47 },
		{ 505, 261, 47 },
		{ 505, 262, 47 },
		{ 505, 263, 47 },
		{ 505, 264, 47 },
		{ 505, 265, 47 },
		{ 505, 266, 47 },
		{ 505, 267, 47 },
		{ 505, 268, 47 },
		{ 505, 269, 47 },
		{ 505, 270, 47 },
		{ 505, 271, 47 },
		{ 505, 272, 47 },
		{ 505, 273, 47 },
		{ 505, 274, 47 },
		{ 505, 275, 47 },
		{ 505, 276, 47 },
		{ 505, 277, 47 },
		{ 505, 278, 47 },
		{ 505, 279, 47 },
		{ 505, 280, 47 },
		{ 505, 281, 47 },
		{ 505, 282, 47 },
		{ 505, 283, 47 },
		{ 505, 284, 47 },
		{ 505, 285, 47 },
		{ 505, 286, 47 },
		{ 505, 287, 47 },
		{ 505, 288, 47 },
		{ 505, 289, 47 },
		{ 505, 290, 47 },
		{ 505, 291, 47 },
		{ 505, 292, 47 },
		{ 505, 293, 47 },
		{ 505, 294, 47 },
		{ 505, 295, 47 },
		{ 505, 296, 47 },
		{ 505, 297, 47 },
		{ 505, 298, 47 },
		{ 505, 299, 47 },
		{ 505, 300, 47 },
		{ 505, 301, 47 },
		{ 505, 302, 47 },
		{ 505, 303, 47 },
		{ 505, 304, 47 },
		{ 505, 305, 47 },
		{ 505, 306, 47 },
		{ 505, 307, 47 },
		{ 505, 308, 47 },
		{ 505, 309, 47 },
		{ 505, 310, 47 },
		{ 505, 311, 47 },
		{ 505, 312, 47 },
		{ 505, 313, 47 },
		{ 505, 314, 47 },
		{ 505, 315, 47 },
		{ 505, 316, 47 },
		{ 505, 317, 47 },
		{ 505, 318, 47 },
		{ 506, 258, 47 },
		{ 506, 259, 47 },
		{ 506, 260, 47 },
		{ 506, 261, 47 },
		{ 506, 262, 47 },
		{ 506, 263, 47 },
		{ 506, 264, 47 },
		{ 506, 265, 47 },
		{ 506, 266, 47 },
		{ 506, 267, 47 },
		{ 506, 268, 47 },
		{ 506, 269, 47 },
		{ 506, 270, 47 },
		{ 506, 271, 47 },
		{ 506, 272, 47 },
		{ 506, 273, 47 },
		{ 506, 274, 47 },
		{ 506, 275, 47 },
		{ 506, 276, 47 },
		{ 506, 277, 47 },
		{ 506, 278, 47 },
		{ 506, 279, 47 },
		{ 506, 280, 47 },
		{ 506, 281, 47 },
		{ 506, 282, 47 },
		{ 506, 283, 47 },
		{ 506, 284, 47 },
		{ 506, 285, 47 },
		{ 506, 286, 47 },
		{ 506, 287, 47 },
		{ 506, 288, 47 },
		{ 506, 289, 47 },
		{ 506, 290, 47 },
		{ 506, 291, 47 },
		{ 506, 292, 47 },
		{ 506, 293, 47 },
		{ 506, 294, 47 },
		{ 506, 295, 47 },
		{ 506, 296, 47 },
		{ 506, 297, 47 },
		{ 506, 298, 47 },
		{ 506, 299, 47 },
		{ 506, 300, 47 },
		{ 506, 301, 47 },
		{ 506, 302, 47 },
		{ 506, 303, 47 },
		{ 506, 304, 47 },
		{ 506, 305, 47 },
		{ 506, 306, 47 },
		{ 506, 307, 47 },
		{ 506, 308, 47 },
		{ 506, 309, 47 },
		{ 506, 310, 47 },
		{ 506, 311, 47 },
		{ 506, 312, 47 },
		{ 506, 313, 47 },
		{ 506, 314, 47 },
		{ 506, 315, 47 },
		{ 506, 316, 47 },
		{ 506, 317, 47 },
		{ 506, 318, 47 },
		{ 506, 319, 47 },
		{ 507, 258, 47 },
		{ 507, 259, 47 },
		{ 507, 260, 47 },
		{ 507, 261, 47 },
		{ 507, 262, 47 },
		{ 507, 263, 47 },
		{ 507, 264, 47 },
		{ 507, 265, 47 },
		{ 507, 266, 47 },
		{ 507, 267, 47 },
		{ 507, 268, 47 },
		{ 507, 269, 47 },
		{ 507, 270, 47 },
		{ 507, 271, 47 },
		{ 507, 272, 47 },
		{ 507, 273, 47 },
		{ 507, 274, 47 },
		{ 507, 275, 47 },
		{ 507, 276, 47 },
		{ 507, 277, 47 },
		{ 507, 278, 47 },
		{ 507, 279, 47 },
		{ 507, 280, 47 },
		{ 507, 281, 47 },
		{ 507, 282, 47 },
		{ 507, 283, 47 },
		{ 507, 284, 47 },
		{ 507, 285, 47 },
		{ 507, 286, 47 },
		{ 507, 287, 47 },
		{ 507, 288, 47 },
		{ 507, 289, 47 },
		{ 507, 290, 47 },
		{ 507, 291, 47 },
		{ 507, 292, 47 },
		{ 507, 293, 47 },
		{ 507, 294, 47 },
		{ 507, 295, 47 },
		{ 507, 296, 47 },
		{ 507, 297, 47 },
		{ 507, 298, 47 },
		{ 507, 299, 47 },
		{ 507, 300, 47 },
		{ 507, 301, 47 },
		{ 507, 302, 47 },
		{ 507, 303, 47 },
		{ 507, 304, 47 },
		{ 507, 305, 47 },
		{ 507, 306, 47 },
		{ 507, 307, 47 },
		{ 507, 308, 47 },
		{ 507, 309, 47 },
		{ 507, 310, 47 },
		{ 507, 311, 47 },
		{ 507, 312, 47 },
		{ 507, 313, 47 },
		{ 507, 314, 47 },
		{ 507, 315, 47 },
		{ 507, 316, 47 },
		{ 507, 317, 47 },
		{ 507, 318, 47 },
		{ 507, 319, 47 },
		{ 508, 258, 47 },
		{ 508, 259, 47 },
		{ 508, 260, 47 },
		{ 508, 261, 47 },
		{ 508, 262, 47 },
		{ 508, 263, 47 },
		{ 508, 264, 47 },
		{ 508, 265, 47 },
		{ 508, 266, 47 },
		{ 508, 267, 47 },
		{ 508, 268, 47 },
		{ 508, 269, 47 },
		{ 508, 270, 47 },
		{ 508, 271, 47 },
		{ 508, 272, 47 },
		{ 508, 273, 47 },
		{ 508, 274, 47 },
		{ 508, 275, 47 },
		{ 508, 276, 47 },
		{ 508, 277, 47 },
		{ 508, 278, 47 },
		{ 508, 279, 47 },
		{ 508, 280, 47 },
		{ 508, 281, 47 },
		{ 508, 282, 47 },
		{ 508, 283, 47 },
		{ 508, 284, 47 },
		{ 508, 285, 47 },
		{ 508, 286, 47 },
		{ 508, 287, 47 },
		{ 508, 288, 47 },
		{ 508, 289, 47 },
		{ 508, 290, 47 },
		{ 508, 291, 47 },
		{ 508, 292, 47 },
		{ 508, 293, 47 },
		{ 508, 294, 47 },
		{ 508, 295, 47 },
		{ 508, 296, 47 },
		{ 508, 297, 47 },
		{ 508, 298, 47 },
		{ 508, 299, 47 },
		{ 508, 300, 47 },
		{ 508, 301, 47 },
		{ 508, 302, 47 },
		{ 508, 303, 47 },
		{ 508, 304, 47 },
		{ 508, 305, 47 },
		{ 508, 306, 47 },
		{ 508, 307, 47 },
		{ 508, 308, 47 },
		{ 508, 309, 47 },
		{ 508, 310, 47 },
		{ 508, 311, 47 },
		{ 508, 312, 47 },
		{ 508, 313, 47 },
		{ 508, 314, 47 },
		{ 508, 315, 47 },
		{ 508, 316, 47 },
		{ 508, 317, 47 },
		{ 508, 318, 47 },
		{ 508, 319, 47 },
		{ 509, 258, 47 },
		{ 509, 259, 47 },
		{ 509, 260, 47 },
		{ 509, 261, 47 },
		{ 509, 262, 47 },
		{ 509, 263, 47 },
		{ 509, 264, 47 },
		{ 509, 265, 47 },
		{ 509, 266, 47 },
		{ 509, 267, 47 },
		{ 509, 268, 47 },
		{ 509, 269, 47 },
		{ 509, 270, 47 },
		{ 509, 271, 47 },
		{ 509, 272, 47 },
		{ 509, 273, 47 },
		{ 509, 274, 47 },
		{ 509, 275, 47 },
		{ 509, 276, 47 },
		{ 509, 277, 47 },
		{ 509, 278, 47 },
		{ 509, 279, 47 },
		{ 509, 280, 47 },
		{ 509, 281, 47 },
		{ 509, 282, 47 },
		{ 509, 283, 47 },
		{ 509, 284, 47 },
		{ 509, 285, 47 },
		{ 509, 286, 47 },
		{ 509, 287, 47 },
		{ 509, 288, 47 },
		{ 509, 289, 47 },
		{ 509, 290, 47 },
		{ 509, 291, 47 },
		{ 509, 292, 47 },
		{ 509, 293, 47 },
		{ 509, 294, 47 },
		{ 509, 295, 47 },
		{ 509, 296, 47 },
		{ 509, 297, 47 },
		{ 509, 298, 47 },
		{ 509, 299, 47 },
		{ 509, 300, 47 },
		{ 509, 301, 47 },
		{ 509, 302, 47 },
		{ 509, 303, 47 },
		{ 509, 304, 47 },
		{ 509, 305, 47 },
		{ 509, 306, 47 },
		{ 509, 307, 47 },
		{ 509, 308, 47 },
		{ 509, 309, 47 },
		{ 509, 310, 47 },
		{ 509, 311, 47 },
		{ 509, 312, 47 },
		{ 509, 313, 47 },
		{ 509, 314, 47 },
		{ 509, 315, 47 },
		{ 509, 316, 47 },
		{ 509, 317, 47 },
		{ 509, 318, 47 },
		{ 509, 319, 47 },
		{ 510, 258, 47 },
		{ 510, 259, 47 },
		{ 510, 260, 47 },
		{ 510, 261, 47 },
		{ 510, 262, 47 },
		{ 510, 263, 47 },
		{ 510, 264, 47 },
		{ 510, 265, 47 },
		{ 510, 266, 47 },
		{ 510, 267, 47 },
		{ 510, 268, 47 },
		{ 510, 269, 47 },
		{ 510, 270, 47 },
		{ 510, 271, 47 },
		{ 510, 272, 47 },
		{ 510, 273, 47 },
		{ 510, 274, 47 },
		{ 510, 275, 47 },
		{ 510, 276, 47 },
		{ 510, 277, 47 },
		{ 510, 278, 47 },
		{ 510, 279, 47 },
		{ 510, 280, 47 },
		{ 510, 281, 47 },
		{ 510, 282, 47 },
		{ 510, 283, 47 },
		{ 510, 284, 47 },
		{ 510, 285, 47 },
		{ 510, 286, 47 },
		{ 510, 287, 47 },
		{ 510, 288, 47 },
		{ 510, 289, 47 },
		{ 510, 290, 47 },
		{ 510, 291, 47 },
		{ 510, 292, 47 },
		{ 510, 293, 47 },
		{ 510, 294, 47 },
		{ 510, 295, 47 },
		{ 510, 296, 47 },
		{ 510, 297, 47 },
		{ 510, 298, 47 },
		{ 510, 299, 47 },
		{ 510, 300, 47 },
		{ 510, 301, 47 },
		{ 510, 302, 47 },
		{ 510, 303, 47 },
		{ 510, 304, 47 },
		{ 510, 305, 47 },
		{ 510, 306, 47 },
		{ 510, 307, 47 },
		{ 510, 308, 47 },
		{ 510, 309, 47 },
		{ 510, 310, 47 },
		{ 510, 311, 47 },
		{ 510, 312, 47 },
		{ 510, 313, 47 },
		{ 510, 314, 47 },
		{ 510, 315, 47 },
		{ 510, 316, 47 },
		{ 510, 317, 47 },
		{ 510, 318, 47 },
		{ 510, 319, 47 },
		{ 510, 320, 47 },
		{ 511, 258, 47 },
		{ 511, 259, 47 },
		{ 511, 260, 47 },
		{ 511, 261, 47 },
		{ 511, 262, 47 },
		{ 511, 263, 47 },
		{ 511, 264, 47 },
		{ 511, 265, 47 },
		{ 511, 266, 47 },
		{ 511, 267, 47 },
		{ 511, 268, 47 },
		{ 511, 269, 47 },
		{ 511, 270, 47 },
		{ 511, 271, 47 },
		{ 511, 272, 47 },
		{ 511, 273, 47 },
		{ 511, 274, 47 },
		{ 511, 275, 47 },
		{ 511, 276, 47 },
		{ 511, 277, 47 },
		{ 511, 278, 47 },
		{ 511, 279, 47 },
		{ 511, 280, 47 },
		{ 511, 281, 47 },
		{ 511, 282, 47 },
		{ 511, 283, 47 },
		{ 511, 284, 47 },
		{ 511, 285, 47 },
		{ 511, 286, 47 },
		{ 511, 287, 47 },
		{ 511, 288, 47 },
		{ 511, 289, 47 },
		{ 511, 290, 47 },
		{ 511, 291, 47 },
		{ 511, 292, 47 },
		{ 511, 293, 47 },
		{ 511, 294, 47 },
		{ 511, 295, 47 },
		{ 511, 296, 47 },
		{ 511, 297, 47 },
		{ 511, 298, 47 },
		{ 511, 299, 47 },
		{ 511, 300, 47 },
		{ 511, 301, 47 },
		{ 511, 302, 47 },
		{ 511, 303, 47 },
		{ 511, 304, 47 },
		{ 511, 305, 47 },
		{ 511, 306, 47 },
		{ 511, 307, 47 },
		{ 511, 308, 47 },
		{ 511, 309, 47 },
		{ 511, 310, 47 },
		{ 511, 311, 47 },
		{ 511, 312, 47 },
		{ 511, 313, 47 },
		{ 511, 314, 47 },
		{ 511, 315, 47 },
		{ 511, 316, 47 },
		{ 511, 317, 47 },
		{ 511, 318, 47 },
		{ 511, 319, 47 },
		{ 511, 320, 47 },
		{ 512, 258, 47 },
		{ 512, 259, 47 },
		{ 512, 260, 47 },
		{ 512, 261, 47 },
		{ 512, 262, 47 },
		{ 512, 263, 47 },
		{ 512, 264, 47 },
		{ 512, 265, 47 },
		{ 512, 266, 47 },
		{ 512, 267, 47 },
		{ 512, 268, 47 },
		{ 512, 269, 47 },
		{ 512, 270, 47 },
		{ 512, 271, 47 },
		{ 512, 272, 47 },
		{ 512, 273, 47 },
		{ 512, 274, 47 },
		{ 512, 275, 47 },
		{ 512, 276, 47 },
		{ 512, 277, 47 },
		{ 512, 278, 47 },
		{ 512, 279, 47 },
		{ 512, 280, 47 },
		{ 512, 281, 47 },
		{ 512, 282, 47 },
		{ 512, 283, 47 },
		{ 512, 284, 47 },
		{ 512, 285, 47 },
		{ 512, 286, 47 },
		{ 512, 287, 47 },
		{ 512, 288, 47 },
		{ 512, 289, 47 },
		{ 512, 290, 47 },
		{ 512, 291, 47 },
		{ 512, 292, 47 },
		{ 512, 293, 47 },
		{ 512, 294, 47 },
		{ 512, 295, 47 },
		{ 512, 296, 47 },
		{ 512, 297, 47 },
		{ 512, 298, 47 },
		{ 512, 299, 47 },
		{ 512, 300, 47 },
		{ 512, 301, 47 },
		{ 512, 302, 47 },
		{ 512, 303, 47 },
		{ 512, 304, 47 },
		{ 512, 305, 47 },
		{ 512, 306, 47 },
		{ 512, 307, 47 },
		{ 512, 308, 47 },
		{ 512, 309, 47 },
		{ 512, 310, 47 },
		{ 512, 311, 47 },
		{ 512, 312, 47 },
		{ 512, 313, 47 },
		{ 512, 314, 47 },
		{ 512, 315, 47 },
		{ 512, 316, 47 },
		{ 512, 317, 47 },
		{ 512, 318, 47 },
		{ 512, 319, 47 },
		{ 512, 320, 47 },
		{ 1, 162, 48 },
		{ 1, 163, 48 },
		{ 1, 164, 48 },
		{ 1, 165, 48 },
		{ 1, 166, 48 },
		{ 1, 167, 48 },
		{ 1, 168, 48 },
		{ 1, 169, 48 },
		{ 1, 170, 48 },
		{ 1, 171, 48 },
		{ 1, 172, 48 },
		{ 1, 173, 48 },
		{ 1, 174, 48 },
		{ 1, 175, 48 },
		{ 1, 176, 48 },
		{ 1, 177, 48 },
		{ 1, 178, 48 },
		{ 1, 179, 48 },
		{ 1, 180, 48 },
		{ 1, 181, 48 },
		{ 1, 182, 48 },
		{ 1, 183, 48 },
		{ 1, 184, 48 },
		{ 1, 185, 48 },
		{ 1, 186, 48 },
		{ 1, 187, 48 },
		{ 1, 188, 48 },
		{ 1, 189, 48 },
		{ 1, 190, 48 },
		{ 1, 191, 48 },
		{ 1, 192, 48 },
		{ 1, 193, 48 },
		{ 1, 194, 48 },
		{ 1, 195, 48 },
		{ 1, 196, 48 },
		{ 1, 197, 48 },
		{ 1, 198, 48 },
		{ 1, 199, 48 },
		{ 1, 200, 48 },
		{ 1, 201, 48 },
		{ 1, 202, 48 },
		{ 1, 203, 48 },
		{ 1, 204, 48 },
		{ 1, 205, 48 },
		{ 1, 206, 48 },
		{ 1, 207, 48 },
		{ 1, 208, 48 },
		{ 1, 209, 48 },
		{ 1, 210, 48 },
		{ 1, 211, 48 },
		{ 1, 212, 48 },
		{ 1, 213, 48 },
		{ 1, 214, 48 },
		{ 1, 215, 48 },
		{ 1, 216, 48 },
		{ 1, 217, 48 },
		{ 1, 218, 48 },
		{ 1, 219, 48 },
		{ 1, 220, 48 },
		{ 1, 221, 48 },
		{ 1, 222, 48 },
		{ 1, 223, 48 },
		{ 1, 224, 48 },
		{ 2, 162, 48 },
		{ 2, 163, 48 },
		{ 2, 164, 48 },
		{ 2, 165, 48 },
		{ 2, 166, 48 },
		{ 2, 167, 48 },
		{ 2, 168, 48 },
		{ 2, 169, 48 },
		{ 2, 170, 48 },
		{ 2, 171, 48 },
		{ 2, 172, 48 },
		{ 2, 173, 48 },
		{ 2, 174, 48 },
		{ 2, 175, 48 },
		{ 2, 176, 48 },
		{ 2, 177, 48 },
		{ 2, 178, 48 },
		{ 2, 179, 48 },
		{ 2, 180, 48 },
		{ 2, 181, 48 },
		{ 2, 182, 48 },
		{ 2, 183, 48 },
		{ 2, 184, 48 },
		{ 2, 185, 48 },
		{ 2, 186, 48 },
		{ 2, 187, 48 },
		{ 2, 188, 48 },
		{ 2, 189, 48 },
		{ 2, 190, 48 },
		{ 2, 191, 48 },
		{ 2, 192, 48 },
		{ 2, 193, 48 },
		{ 2, 194, 48 },
		{ 2, 195, 48 },
		{ 2, 196, 48 },
		{ 2, 197, 48 },
		{ 2, 198, 48 },
		{ 2, 199, 48 },
		{ 2, 200, 48 },
		{ 2, 201, 48 },
		{ 2, 202, 48 },
		{ 2, 203, 48 },
		{ 2, 204, 48 },
		{ 2, 205, 48 },
		{ 2, 206, 48 },
		{ 2, 207, 48 },
		{ 2, 208, 48 },
		{ 2, 209, 48 },
		{ 2, 210, 48 },
		{ 2, 211, 48 },
		{ 2, 212, 48 },
		{ 2, 213, 48 },
		{ 2, 214, 48 },
		{ 2, 215, 48 },
		{ 2, 216, 48 },
		{ 2, 217, 48 },
		{ 2, 218, 48 },
		{ 2, 219, 48 },
		{ 2, 220, 48 },
		{ 2, 221, 48 },
		{ 2, 222, 48 },
		{ 2, 223, 48 },
		{ 2, 224, 48 },
		{ 2, 225, 48 },
		{ 3, 162, 48 },
		{ 3, 163, 48 },
		{ 3, 164, 48 },
		{ 3, 165, 48 },
		{ 3, 166, 48 },
		{ 3, 167, 48 },
		{ 3, 168, 48 },
		{ 3, 169, 48 },
		{ 3, 170, 48 },
		{ 3, 171, 48 },
		{ 3, 172, 48 },
		{ 3, 173, 48 },
		{ 3, 174, 48 },
		{ 3, 175, 48 },
		{ 3, 176, 48 },
		{ 3, 177, 48 },
		{ 3, 178, 48 },
		{ 3, 179, 48 },
		{ 3, 180, 48 },
		{ 3, 181, 48 },
		{ 3, 182, 48 },
		{ 3, 183, 48 },
		{ 3, 184, 48 },
		{ 3, 185, 48 },
		{ 3, 186, 48 },
		{ 3, 187, 48 },
		{ 3, 188, 48 },
		{ 3, 189, 48 },
		{ 3, 190, 48 },
		{ 3, 191, 48 },
		{ 3, 192, 48 },
		{ 3, 193, 48 },
		{ 3, 194, 48 },
		{ 3, 195, 48 },
		{ 3, 196, 48 },
		{ 3, 197, 48 },
		{ 3, 198, 48 },
		{ 3, 199, 48 },
		{ 3, 200, 48 },
		{ 3, 201, 48 },
		{ 3, 202, 48 },
		{ 3, 203, 48 },
		{ 3, 204, 48 },
		{ 3, 205, 48 },
		{ 3, 206, 48 },
		{ 3, 207, 48 },
		{ 3, 208, 48 },
		{ 3, 209, 48 },
		{ 3, 210, 48 },
		{ 3, 211, 48 },
		{ 3, 212, 48 },
		{ 3, 213, 48 },
		{ 3, 214, 48 },
		{ 3, 215, 48 },
		{ 3, 216, 48 },
		{ 3, 217, 48 },
		{ 3, 218, 48 },
		{ 3, 219, 48 },
		{ 3, 220, 48 },
		{ 3, 221, 48 },
		{ 3, 222, 48 },
		{ 3, 223, 48 },
		{ 3, 224, 48 },
		{ 3, 225, 48 },
		{ 4, 163, 48 },
		{ 4, 164, 48 },
		{ 4, 165, 48 },
		{ 4, 166, 48 },
		{ 4, 167, 48 },
		{ 4, 168, 48 },
		{ 4, 169, 48 },
		{ 4, 170, 48 },
		{ 4, 171, 48 },
		{ 4, 172, 48 },
		{ 4, 173, 48 },
		{ 4, 174, 48 },
		{ 4, 175, 48 },
		{ 4, 176, 48 },
		{ 4, 177, 48 },
		{ 4, 178, 48 },
		{ 4, 179, 48 },
		{ 4, 180, 48 },
		{ 4, 181, 48 },
		{ 4, 182, 48 },
		{ 4, 183, 48 },
		{ 4, 184, 48 },
		{ 4, 185, 48 },
		{ 4, 186, 48 },
		{ 4, 187, 48 },
		{ 4, 188, 48 },
		{ 4, 189, 48 },
		{ 4, 190, 48 },
		{ 4, 191, 48 },
		{ 4, 192, 48 },
		{ 4, 193, 48 },
		{ 4, 194, 48 },
		{ 4, 195, 48 },
		{ 4, 196, 48 },
		{ 4, 197, 48 },
		{ 4, 198, 48 },
		{ 4, 199, 48 },
		{ 4, 200, 48 },
		{ 4, 201, 48 },
		{ 4, 202, 48 },
		{ 4, 203, 48 },
		{ 4, 204, 48 },
		{ 4, 205, 48 },
		{ 4, 206, 48 },
		{ 4, 207, 48 },
		{ 4, 208, 48 },
		{ 4, 209, 48 },
		{ 4, 210, 48 },
		{ 4, 211, 48 },
		{ 4, 212, 48 },
		{ 4, 213, 48 },
		{ 4, 214, 48 },
		{ 4, 215, 48 },
		{ 4, 216, 48 },
		{ 4, 217, 48 },
		{ 4, 218, 48 },
		{ 4, 219, 48 },
		{ 4, 220, 48 },
		{ 4, 221, 48 },
		{ 4, 222, 48 },
		{ 4, 223, 48 },
		{ 4, 224, 48 },
		{ 4, 225, 48 },
		{ 5, 163, 48 },
		{ 5, 164, 48 },
		{ 5, 165, 48 },
		{ 5, 166, 48 },
		{ 5, 167, 48 },
		{ 5, 168, 48 },
		{ 5, 169, 48 },
		{ 5, 170, 48 },
		{ 5, 171, 48 },
		{ 5, 172, 48 },
		{ 5, 173, 48 },
		{ 5, 174, 48 },
		{ 5, 175, 48 },
		{ 5, 176, 48 },
		{ 5, 177, 48 },
		{ 5, 178, 48 },
		{ 5, 179, 48 },
		{ 5, 180, 48 },
		{ 5, 181, 48 },
		{ 5, 182, 48 },
		{ 5, 183, 48 },
		{ 5, 184, 48 },
		{ 5, 185, 48 },
		{ 5, 186, 48 },
		{ 5, 187, 48 },
		{ 5, 188, 48 },
		{ 5, 189, 48 },
		{ 5, 190, 48 },
		{ 5, 191, 48 },
		{ 5, 192, 48 },
		{ 5, 193, 48 },
		{ 5, 194, 48 },
		{ 5, 195, 48 },
		{ 5, 196, 48 },
		{ 5, 197, 48 },
		{ 5, 198, 48 },
		{ 5, 199, 48 },
		{ 5, 200, 48 },
		{ 5, 201, 48 },
		{ 5, 202, 48 },
		{ 5, 203, 48 },
		{ 5, 204, 48 },
		{ 5, 205, 48 },
		{ 5, 206, 48 },
		{ 5, 207, 48 },
		{ 5, 208, 48 },
		{ 5, 209, 48 },
		{ 5, 210, 48 },
		{ 5, 211, 48 },
		{ 5, 212, 48 },
		{ 5, 213, 48 },
		{ 5, 214, 48 },
		{ 5, 215, 48 },
		{ 5, 216, 48 },
		{ 5, 217, 48 },
		{ 5, 218, 48 },
		{ 5, 219, 48 },
		{ 5, 220, 48 },
		{ 5, 221, 48 },
		{ 5, 222, 48 },
		{ 5, 223, 48 },
		{ 5, 224, 48 },
		{ 5, 225, 48 },
		{ 6, 163, 48 },
		{ 6, 164, 48 },
		{ 6, 165, 48 },
		{ 6, 166, 48 },
		{ 6, 167, 48 },
		{ 6, 168, 48 },
		{ 6, 169, 48 },
		{ 6, 170, 48 },
		{ 6, 171, 48 },
		{ 6, 172, 48 },
		{ 6, 173, 48 },
		{ 6, 174, 48 },
		{ 6, 175, 48 },
		{ 6, 176, 48 },
		{ 6, 177, 48 },
		{ 6, 178, 48 },
		{ 6, 179, 48 },
		{ 6, 180, 48 },
		{ 6, 181, 48 },
		{ 6, 182, 48 },
		{ 6, 183, 48 },
		{ 6, 184, 48 },
		{ 6, 185, 48 },
		{ 6, 186, 48 },
		{ 6, 187, 48 },
		{ 6, 188, 48 },
		{ 6, 189, 48 },
		{ 6, 190, 48 },
		{ 6, 191, 48 },
		{ 6, 192, 48 },
		{ 6, 193, 48 },
		{ 6, 194, 48 },
		{ 6, 195, 48 },
		{ 6, 196, 48 },
		{ 6, 197, 48 },
		{ 6, 198, 48 },
		{ 6, 199, 48 },
		{ 6, 200, 48 },
		{ 6, 201, 48 },
		{ 6, 202, 48 },
		{ 6, 203, 48 },
		{ 6, 204, 48 },
		{ 6, 205, 48 },
		{ 6, 206, 48 },
		{ 6, 207, 48 },
		{ 6, 208, 48 },
		{ 6, 209, 48 },
		{ 6, 210, 48 },
		{ 6, 211, 48 },
		{ 6, 212, 48 },
		{ 6, 213, 48 },
		{ 6, 214, 48 },
		{ 6, 215, 48 },
		{ 6, 216, 48 },
		{ 6, 217, 48 },
		{ 6, 218, 48 },
		{ 6, 219, 48 },
		{ 6, 220, 48 },
		{ 6, 221, 48 },
		{ 6, 222, 48 },
		{ 6, 223, 48 },
		{ 6, 224, 48 },
		{ 6, 225, 48 },
		{ 7, 164, 48 },
		{ 7, 165, 48 },
		{ 7, 166, 48 },
		{ 7, 167, 48 },
		{ 7, 168, 48 },
		{ 7, 169, 48 },
		{ 7, 170, 48 },
		{ 7, 171, 48 },
		{ 7, 172, 48 },
		{ 7, 173, 48 },
		{ 7, 174, 48 },
		{ 7, 175, 48 },
		{ 7, 176, 48 },
		{ 7, 177, 48 },
		{ 7, 178, 48 },
		{ 7, 179, 48 },
		{ 7, 180, 48 },
		{ 7, 181, 48 },
		{ 7, 182, 48 },
		{ 7, 183, 48 },
		{ 7, 184, 48 },
		{ 7, 185, 48 },
		{ 7, 186, 48 },
		{ 7, 187, 48 },
		{ 7, 188, 48 },
		{ 7, 189, 48 },
		{ 7, 190, 48 },
		{ 7, 191, 48 },
		{ 7, 192, 48 },
		{ 7, 193, 48 },
		{ 7, 194, 48 },
		{ 7, 195, 48 },
		{ 7, 196, 48 },
		{ 7, 197, 48 },
		{ 7, 198, 48 },
		{ 7, 199, 48 },
		{ 7, 200, 48 },
		{ 7, 201, 48 },
		{ 7, 202, 48 },
		{ 7, 203, 48 },
		{ 7, 204, 48 },
		{ 7, 205, 48 },
		{ 7, 206, 48 },
		{ 7, 207, 48 },
		{ 7, 208, 48 },
		{ 7, 209, 48 },
		{ 7, 210, 48 },
		{ 7, 211, 48 },
		{ 7, 212, 48 },
		{ 7, 213, 48 },
		{ 7, 214, 48 },
		{ 7, 215, 48 },
		{ 7, 216, 48 },
		{ 7, 217, 48 },
		{ 7, 218, 48 },
		{ 7, 219, 48 },
		{ 7, 220, 48 },
		{ 7, 221, 48 },
		{ 7, 222, 48 },
		{ 7, 223, 48 },
		{ 7, 224, 48 },
		{ 7, 225, 48 },
		{ 8, 164, 48 },
		{ 8, 165, 48 },
		{ 8, 166, 48 },
		{ 8, 167, 48 },
		{ 8, 168, 48 },
		{ 8, 169, 48 },
		{ 8, 170, 48 },
		{ 8, 171, 48 },
		{ 8, 172, 48 },
		{ 8, 173, 48 },
		{ 8, 174, 48 },
		{ 8, 175, 48 },
		{ 8, 176, 48 },
		{ 8, 177, 48 },
		{ 8, 178, 48 },
		{ 8, 179, 48 },
		{ 8, 180, 48 },
		{ 8, 181, 48 },
		{ 8, 182, 48 },
		{ 8, 183, 48 },
		{ 8, 184, 48 },
		{ 8, 185, 48 },
		{ 8, 186, 48 },
		{ 8, 187, 48 },
		{ 8, 188, 48 },
		{ 8, 189, 48 },
		{ 8, 190, 48 },
		{ 8, 191, 48 },
		{ 8, 192, 48 },
		{ 8, 193, 48 },
		{ 8, 194, 48 },
		{ 8, 195, 48 },
		{ 8, 196, 48 },
		{ 8, 197, 48 },
		{ 8, 198, 48 },
		{ 8, 199, 48 },
		{ 8, 200, 48 },
		{ 8, 201, 48 },
		{ 8, 202, 48 },
		{ 8, 203, 48 },
		{ 8, 204, 48 },
		{ 8, 205, 48 },
		{ 8, 206, 48 },
		{ 8, 207, 48 },
		{ 8, 208, 48 },
		{ 8, 209, 48 },
		{ 8, 210, 48 },
		{ 8, 211, 48 },
		{ 8, 212, 48 },
		{ 8, 213, 48 },
		{ 8, 214, 48 },
		{ 8, 215, 48 },
		{ 8, 216, 48 },
		{ 8, 217, 48 },
		{ 8, 218, 48 },
		{ 8, 219, 48 },
		{ 8, 220, 48 },
		{ 8, 221, 48 },
		{ 8, 222, 48 },
		{ 8, 223, 48 },
		{ 8, 224, 48 },
		{ 8, 225, 48 },
		{ 9, 165, 48 },
		{ 9, 166, 48 },
		{ 9, 167, 48 },
		{ 9, 168, 48 },
		{ 9, 169, 48 },
		{ 9, 170, 48 },
		{ 9, 171, 48 },
		{ 9, 172, 48 },
		{ 9, 173, 48 },
		{ 9, 174, 48 },
		{ 9, 175, 48 },
		{ 9, 176, 48 },
		{ 9, 177, 48 },
		{ 9, 178, 48 },
		{ 9, 179, 48 },
		{ 9, 180, 48 },
		{ 9, 181, 48 },
		{ 9, 182, 48 },
		{ 9, 183, 48 },
		{ 9, 184, 48 },
		{ 9, 185, 48 },
		{ 9, 186, 48 },
		{ 9, 187, 48 },
		{ 9, 188, 48 },
		{ 9, 189, 48 },
		{ 9, 190, 48 },
		{ 9, 191, 48 },
		{ 9, 192, 48 },
		{ 9, 193, 48 },
		{ 9, 194, 48 },
		{ 9, 195, 48 },
		{ 9, 196, 48 },
		{ 9, 197, 48 },
		{ 9, 198, 48 },
		{ 9, 199, 48 },
		{ 9, 200, 48 },
		{ 9, 201, 48 },
		{ 9, 202, 48 },
		{ 9, 203, 48 },
		{ 9, 204, 48 },
		{ 9, 205, 48 },
		{ 9, 206, 48 },
		{ 9, 207, 48 },
		{ 9, 208, 48 },
		{ 9, 209, 48 },
		{ 9, 210, 48 },
		{ 9, 211, 48 },
		{ 9, 212, 48 },
		{ 9, 213, 48 },
		{ 9, 214, 48 },
		{ 9, 215, 48 },
		{ 9, 216, 48 },
		{ 9, 217, 48 },
		{ 9, 218, 48 },
		{ 9, 219, 48 },
		{ 9, 220, 48 },
		{ 9, 221, 48 },
		{ 9, 222, 48 },
		{ 9, 223, 48 },
		{ 9, 224, 48 },
		{ 9, 225, 48 },
		{ 10, 165, 48 },
		{ 10, 166, 48 },
		{ 10, 167, 48 },
		{ 10, 168, 48 },
		{ 10, 169, 48 },
		{ 10, 170, 48 },
		{ 10, 171, 48 },
		{ 10, 172, 48 },
		{ 10, 173, 48 },
		{ 10, 174, 48 },
		{ 10, 175, 48 },
		{ 10, 176, 48 },
		{ 10, 177, 48 },
		{ 10, 178, 48 },
		{ 10, 179, 48 },
		{ 10, 180, 48 },
		{ 10, 181, 48 },
		{ 10, 182, 48 },
		{ 10, 183, 48 },
		{ 10, 184, 48 },
		{ 10, 185, 48 },
		{ 10, 186, 48 },
		{ 10, 187, 48 },
		{ 10, 188, 48 },
		{ 10, 189, 48 },
		{ 10, 190, 48 },
		{ 10, 191, 48 },
		{ 10, 192, 48 },
		{ 10, 193, 48 },
		{ 10, 194, 48 },
		{ 10, 195, 48 },
		{ 10, 196, 48 },
		{ 10, 197, 48 },
		{ 10, 198, 48 },
		{ 10, 199, 48 },
		{ 10, 200, 48 },
		{ 10, 201, 48 },
		{ 10, 202, 48 },
		{ 10, 203, 48 },
		{ 10, 204, 48 },
		{ 10, 205, 48 },
		{ 10, 206, 48 },
		{ 10, 207, 48 },
		{ 10, 208, 48 },
		{ 10, 209, 48 },
		{ 10, 210, 48 },
		{ 10, 211, 48 },
		{ 10, 212, 48 },
		{ 10, 213, 48 },
		{ 10, 214, 48 },
		{ 10, 215, 48 },
		{ 10, 216, 48 },
		{ 10, 217, 48 },
		{ 10, 218, 48 },
		{ 10, 219, 48 },
		{ 10, 220, 48 },
		{ 10, 221, 48 },
		{ 10, 222, 48 },
		{ 10, 223, 48 },
		{ 10, 224, 48 },
		{ 10, 225, 48 },
		{ 10, 226, 48 },
		{ 11, 165, 48 },
		{ 11, 166, 48 },
		{ 11, 167, 48 },
		{ 11, 168, 48 },
		{ 11, 169, 48 },
		{ 11, 170, 48 },
		{ 11, 171, 48 },
		{ 11, 172, 48 },
		{ 11, 173, 48 },
		{ 11, 174, 48 },
		{ 11, 175, 48 },
		{ 11, 176, 48 },
		{ 11, 177, 48 },
		{ 11, 178, 48 },
		{ 11, 179, 48 },
		{ 11, 180, 48 },
		{ 11, 181, 48 },
		{ 11, 182, 48 },
		{ 11, 183, 48 },
		{ 11, 184, 48 },
		{ 11, 185, 48 },
		{ 11, 186, 48 },
		{ 11, 187, 48 },
		{ 11, 188, 48 },
		{ 11, 189, 48 },
		{ 11, 190, 48 },
		{ 11, 191, 48 },
		{ 11, 192, 48 },
		{ 11, 193, 48 },
		{ 11, 194, 48 },
		{ 11, 195, 48 },
		{ 11, 196, 48 },
		{ 11, 197, 48 },
		{ 11, 198, 48 },
		{ 11, 199, 48 },
		{ 11, 200, 48 },
		{ 11, 201, 48 },
		{ 11, 202, 48 },
		{ 11, 203, 48 },
		{ 11, 204, 48 },
		{ 11, 205, 48 },
		{ 11, 206, 48 },
		{ 11, 207, 48 },
		{ 11, 208, 48 },
		{ 11, 209, 48 },
		{ 11, 210, 48 },
		{ 11, 211, 48 },
		{ 11, 212, 48 },
		{ 11, 213, 48 },
		{ 11, 214, 48 },
		{ 11, 215, 48 },
		{ 11, 216, 48 },
		{ 11, 217, 48 },
		{ 11, 218, 48 },
		{ 11, 219, 48 },
		{ 11, 220, 48 },
		{ 11, 221, 48 },
		{ 11, 222, 48 },
		{ 11, 223, 48 },
		{ 11, 224, 48 },
		{ 11, 225, 48 },
		{ 11, 226, 48 },
		{ 12, 166, 48 },
		{ 12, 167, 48 },
		{ 12, 168, 48 },
		{ 12, 169, 48 },
		{ 12, 170, 48 },
		{ 12, 171, 48 },
		{ 12, 172, 48 },
		{ 12, 173, 48 },
		{ 12, 174, 48 },
		{ 12, 175, 48 },
		{ 12, 176, 48 },
		{ 12, 177, 48 },
		{ 12, 178, 48 },
		{ 12, 179, 48 },
		{ 12, 180, 48 },
		{ 12, 181, 48 },
		{ 12, 182, 48 },
		{ 12, 183, 48 },
		{ 12, 184, 48 },
		{ 12, 185, 48 },
		{ 12, 186, 48 },
		{ 12, 187, 48 },
		{ 12, 188, 48 },
		{ 12, 189, 48 },
		{ 12, 190, 48 },
		{ 12, 191, 48 },
		{ 12, 192, 48 },
		{ 12, 193, 48 },
		{ 12, 194, 48 },
		{ 12, 195, 48 },
		{ 12, 196, 48 },
		{ 12, 197, 48 },
		{ 12, 198, 48 },
		{ 12, 199, 48 },
		{ 12, 200, 48 },
		{ 12, 201, 48 },
		{ 12, 202, 48 },
		{ 12, 203, 48 },
		{ 12, 204, 48 },
		{ 12, 205, 48 },
		{ 12, 206, 48 },
		{ 12, 207, 48 },
		{ 12, 208, 48 },
		{ 12, 209, 48 },
		{ 12, 210, 48 },
		{ 12, 211, 48 },
		{ 12, 212, 48 },
		{ 12, 213, 48 },
		{ 12, 214, 48 },
		{ 12, 215, 48 },
		{ 12, 216, 48 },
		{ 12, 217, 48 },
		{ 12, 218, 48 },
		{ 12, 219, 48 },
		{ 12, 220, 48 },
		{ 12, 221, 48 },
		{ 12, 222, 48 },
		{ 12, 223, 48 },
		{ 12, 224, 48 },
		{ 12, 225, 48 },
		{ 12, 226, 48 },
		{ 13, 166, 48 },
		{ 13, 167, 48 },
		{ 13, 168, 48 },
		{ 13, 169, 48 },
		{ 13, 170, 48 },
		{ 13, 171, 48 },
		{ 13, 172, 48 },
		{ 13, 173, 48 },
		{ 13, 174, 48 },
		{ 13, 175, 48 },
		{ 13, 176, 48 },
		{ 13, 177, 48 },
		{ 13, 178, 48 },
		{ 13, 179, 48 },
		{ 13, 180, 48 },
		{ 13, 181, 48 },
		{ 13, 182, 48 },
		{ 13, 183, 48 },
		{ 13, 184, 48 },
		{ 13, 185, 48 },
		{ 13, 186, 48 },
		{ 13, 187, 48 },
		{ 13, 188, 48 },
		{ 13, 189, 48 },
		{ 13, 190, 48 },
		{ 13, 191, 48 },
		{ 13, 192, 48 },
		{ 13, 193, 48 },
		{ 13, 194, 48 },
		{ 13, 195, 48 },
		{ 13, 196, 48 },
		{ 13, 197, 48 },
		{ 13, 198, 48 },
		{ 13, 199, 48 },
		{ 13, 200, 48 },
		{ 13, 201, 48 },
		{ 13, 202, 48 },
		{ 13, 203, 48 },
		{ 13, 204, 48 },
		{ 13, 205, 48 },
		{ 13, 206, 48 },
		{ 13, 207, 48 },
		{ 13, 208, 48 },
		{ 13, 209, 48 },
		{ 13, 210, 48 },
		{ 13, 211, 48 },
		{ 13, 212, 48 },
		{ 13, 213, 48 },
		{ 13, 214, 48 },
		{ 13, 215, 48 },
		{ 13, 216, 48 },
		{ 13, 217, 48 },
		{ 13, 218, 48 },
		{ 13, 219, 48 },
		{ 13, 220, 48 },
		{ 13, 221, 48 },
		{ 13, 222, 48 },
		{ 13, 223, 48 },
		{ 13, 224, 48 },
		{ 13, 225, 48 },
		{ 13, 226, 48 },
		{ 14, 166, 48 },
		{ 14, 167, 48 },
		{ 14, 168, 48 },
		{ 14, 169, 48 },
		{ 14, 170, 48 },
		{ 14, 171, 48 },
		{ 14, 172, 48 },
		{ 14, 173, 48 },
		{ 14, 174, 48 },
		{ 14, 175, 48 },
		{ 14, 176, 48 },
		{ 14, 177, 48 },
		{ 14, 178, 48 },
		{ 14, 179, 48 },
		{ 14, 180, 48 },
		{ 14, 181, 48 },
		{ 14, 182, 48 },
		{ 14, 183, 48 },
		{ 14, 184, 48 },
		{ 14, 185, 48 },
		{ 14, 186, 48 },
		{ 14, 187, 48 },
		{ 14, 188, 48 },
		{ 14, 189, 48 },
		{ 14, 190, 48 },
		{ 14, 191, 48 },
		{ 14, 192, 48 },
		{ 14, 193, 48 },
		{ 14, 194, 48 },
		{ 14, 195, 48 },
		{ 14, 196, 48 },
		{ 14, 197, 48 },
		{ 14, 198, 48 },
		{ 14, 199, 48 },
		{ 14, 200, 48 },
		{ 14, 201, 48 },
		{ 14, 202, 48 },
		{ 14, 203, 48 },
		{ 14, 204, 48 },
		{ 14, 205, 48 },
		{ 14, 206, 48 },
		{ 14, 207, 48 },
		{ 14, 208, 48 },
		{ 14, 209, 48 },
		{ 14, 210, 48 },
		{ 14, 211, 48 },
		{ 14, 212, 48 },
		{ 14, 213, 48 },
		{ 14, 214, 48 },
		{ 14, 215, 48 },
		{ 14, 216, 48 },
		{ 14, 217, 48 },
		{ 14, 218, 48 },
		{ 14, 219, 48 },
		{ 14, 220, 48 },
		{ 14, 221, 48 },
		{ 14, 222, 48 },
		{ 14, 223, 48 },
		{ 14, 224, 48 },
		{ 14, 225, 48 },
		{ 14, 226, 48 },
		{ 15, 167, 48 },
		{ 15, 168, 48 },
		{ 15, 169, 48 },
		{ 15, 170, 48 },
		{ 15, 171, 48 },
		{ 15, 172, 48 },
		{ 15, 173, 48 },
		{ 15, 174, 48 },
		{ 15, 175, 48 },
		{ 15, 176, 48 },
		{ 15, 177, 48 },
		{ 15, 178, 48 },
		{ 15, 179, 48 },
		{ 15, 180, 48 },
		{ 15, 181, 48 },
		{ 15, 182, 48 },
		{ 15, 183, 48 },
		{ 15, 184, 48 },
		{ 15, 185, 48 },
		{ 15, 186, 48 },
		{ 15, 187, 48 },
		{ 15, 188, 48 },
		{ 15, 189, 48 },
		{ 15, 190, 48 },
		{ 15, 191, 48 },
		{ 15, 192, 48 },
		{ 15, 193, 48 },
		{ 15, 194, 48 },
		{ 15, 195, 48 },
		{ 15, 196, 48 },
		{ 15, 197, 48 },
		{ 15, 198, 48 },
		{ 15, 199, 48 },
		{ 15, 200, 48 },
		{ 15, 201, 48 },
		{ 15, 202, 48 },
		{ 15, 203, 48 },
		{ 15, 204, 48 },
		{ 15, 205, 48 },
		{ 15, 206, 48 },
		{ 15, 207, 48 },
		{ 15, 208, 48 },
		{ 15, 209, 48 },
		{ 15, 210, 48 },
		{ 15, 211, 48 },
		{ 15, 212, 48 },
		{ 15, 213, 48 },
		{ 15, 214, 48 },
		{ 15, 215, 48 },
		{ 15, 216, 48 },
		{ 15, 217, 48 },
		{ 15, 218, 48 },
		{ 15, 219, 48 },
		{ 15, 220, 48 },
		{ 15, 221, 48 },
		{ 15, 222, 48 },
		{ 15, 223, 48 },
		{ 15, 224, 48 },
		{ 15, 225, 48 },
		{ 15, 226, 48 },
		{ 16, 167, 48 },
		{ 16, 168, 48 },
		{ 16, 169, 48 },
		{ 16, 170, 48 },
		{ 16, 171, 48 },
		{ 16, 172, 48 },
		{ 16, 173, 48 },
		{ 16, 174, 48 },
		{ 16, 175, 48 },
		{ 16, 176, 48 },
		{ 16, 177, 48 },
		{ 16, 178, 48 },
		{ 16, 179, 48 },
		{ 16, 180, 48 },
		{ 16, 181, 48 },
		{ 16, 182, 48 },
		{ 16, 183, 48 },
		{ 16, 184, 48 },
		{ 16, 185, 48 },
		{ 16, 186, 48 },
		{ 16, 187, 48 },
		{ 16, 188, 48 },
		{ 16, 189, 48 },
		{ 16, 190, 48 },
		{ 16, 191, 48 },
		{ 16, 192, 48 },
		{ 16, 193, 48 },
		{ 16, 194, 48 },
		{ 16, 195, 48 },
		{ 16, 196, 48 },
		{ 16, 197, 48 },
		{ 16, 198, 48 },
		{ 16, 199, 48 },
		{ 16, 200, 48 },
		{ 16, 201, 48 },
		{ 16, 202, 48 },
		{ 16, 203, 48 },
		{ 16, 204, 48 },
		{ 16, 205, 48 },
		{ 16, 206, 48 },
		{ 16, 207, 48 },
		{ 16, 208, 48 },
		{ 16, 209, 48 },
		{ 16, 210, 48 },
		{ 16, 211, 48 },
		{ 16, 212, 48 },
		{ 16, 213, 48 },
		{ 16, 214, 48 },
		{ 16, 215, 48 },
		{ 16, 216, 48 },
		{ 16, 217, 48 },
		{ 16, 218, 48 },
		{ 16, 219, 48 },
		{ 16, 220, 48 },
		{ 16, 221, 48 },
		{ 16, 222, 48 },
		{ 16, 223, 48 },
		{ 16, 224, 48 },
		{ 16, 225, 48 },
		{ 16, 226, 48 },
		{ 17, 168, 48 },
		{ 17, 169, 48 },
		{ 17, 170, 48 },
		{ 17, 171, 48 },
		{ 17, 172, 48 },
		{ 17, 173, 48 },
		{ 17, 174, 48 },
		{ 17, 175, 48 },
		{ 17, 176, 48 },
		{ 17, 177, 48 },
		{ 17, 178, 48 },
		{ 17, 179, 48 },
		{ 17, 180, 48 },
		{ 17, 181, 48 },
		{ 17, 182, 48 },
		{ 17, 183, 48 },
		{ 17, 184, 48 },
		{ 17, 185, 48 },
		{ 17, 186, 48 },
		{ 17, 187, 48 },
		{ 17, 188, 48 },
		{ 17, 189, 48 },
		{ 17, 190, 48 },
		{ 17, 191, 48 },
		{ 17, 192, 48 },
		{ 17, 193, 48 },
		{ 17, 194, 48 },
		{ 17, 195, 48 },
		{ 17, 196, 48 },
		{ 17, 197, 48 },
		{ 17, 198, 48 },
		{ 17, 199, 48 },
		{ 17, 200, 48 },
		{ 17, 201, 48 },
		{ 17, 202, 48 },
		{ 17, 203, 48 },
		{ 17, 204, 48 },
		{ 17, 205, 48 },
		{ 17, 206, 48 },
		{ 17, 207, 48 },
		{ 17, 208, 48 },
		{ 17, 209, 48 },
		{ 17, 210, 48 },
		{ 17, 211, 48 },
		{ 17, 212, 48 },
		{ 17, 213, 48 },
		{ 17, 214, 48 },
		{ 17, 215, 48 },
		{ 17, 216, 48 },
		{ 17, 217, 48 },
		{ 17, 218, 48 },
		{ 17, 219, 48 },
		{ 17, 220, 48 },
		{ 17, 221, 48 },
		{ 17, 222, 48 },
		{ 17, 223, 48 },
		{ 17, 224, 48 },
		{ 17, 225, 48 },
		{ 17, 226, 48 },
		{ 18, 168, 48 },
		{ 18, 169, 48 },
		{ 18, 170, 48 },
		{ 18, 171, 48 },
		{ 18, 172, 48 },
		{ 18, 173, 48 },
		{ 18, 174, 48 },
		{ 18, 175, 48 },
		{ 18, 176, 48 },
		{ 18, 177, 48 },
		{ 18, 178, 48 },
		{ 18, 179, 48 },
		{ 18, 180, 48 },
		{ 18, 181, 48 },
		{ 18, 182, 48 },
		{ 18, 183, 48 },
		{ 18, 184, 48 },
		{ 18, 185, 48 },
		{ 18, 186, 48 },
		{ 18, 187, 48 },
		{ 18, 188, 48 },
		{ 18, 189, 48 },
		{ 18, 190, 48 },
		{ 18, 191, 48 },
		{ 18, 192, 48 },
		{ 18, 193, 48 },
		{ 18, 194, 48 },
		{ 18, 195, 48 },
		{ 18, 196, 48 },
		{ 18, 197, 48 },
		{ 18, 198, 48 },
		{ 18, 199, 48 },
		{ 18, 200, 48 },
		{ 18, 201, 48 },
		{ 18, 202, 48 },
		{ 18, 203, 48 },
		{ 18, 204, 48 },
		{ 18, 205, 48 },
		{ 18, 206, 48 },
		{ 18, 207, 48 },
		{ 18, 208, 48 },
		{ 18, 209, 48 },
		{ 18, 210, 48 },
		{ 18, 211, 48 },
		{ 18, 212, 48 },
		{ 18, 213, 48 },
		{ 18, 214, 48 },
		{ 18, 215, 48 },
		{ 18, 216, 48 },
		{ 18, 217, 48 },
		{ 18, 218, 48 },
		{ 18, 219, 48 },
		{ 18, 220, 48 },
		{ 18, 221, 48 },
		{ 18, 222, 48 },
		{ 18, 223, 48 },
		{ 18, 224, 48 },
		{ 18, 225, 48 },
		{ 18, 226, 48 },
		{ 18, 227, 48 },
		{ 19, 168, 48 },
		{ 19, 169, 48 },
		{ 19, 170, 48 },
		{ 19, 171, 48 },
		{ 19, 172, 48 },
		{ 19, 173, 48 },
		{ 19, 174, 48 },
		{ 19, 175, 48 },
		{ 19, 176, 48 },
		{ 19, 177, 48 },
		{ 19, 178, 48 },
		{ 19, 179, 48 },
		{ 19, 180, 48 },
		{ 19, 181, 48 },
		{ 19, 182, 48 },
		{ 19, 183, 48 },
		{ 19, 184, 48 },
		{ 19, 185, 48 },
		{ 19, 186, 48 },
		{ 19, 187, 48 },
		{ 19, 188, 48 },
		{ 19, 189, 48 },
		{ 19, 190, 48 },
		{ 19, 191, 48 },
		{ 19, 192, 48 },
		{ 19, 193, 48 },
		{ 19, 194, 48 },
		{ 19, 195, 48 },
		{ 19, 196, 48 },
		{ 19, 197, 48 },
		{ 19, 198, 48 },
		{ 19, 199, 48 },
		{ 19, 200, 48 },
		{ 19, 201, 48 },
		{ 19, 202, 48 },
		{ 19, 203, 48 },
		{ 19, 204, 48 },
		{ 19, 205, 48 },
		{ 19, 206, 48 },
		{ 19, 207, 48 },
		{ 19, 208, 48 },
		{ 19, 209, 48 },
		{ 19, 210, 48 },
		{ 19, 211, 48 },
		{ 19, 212, 48 },
		{ 19, 213, 48 },
		{ 19, 214, 48 },
		{ 19, 215, 48 },
		{ 19, 216, 48 },
		{ 19, 217, 48 },
		{ 19, 218, 48 },
		{ 19, 219, 48 },
		{ 19, 220, 48 },
		{ 19, 221, 48 },
		{ 19, 222, 48 },
		{ 19, 223, 48 },
		{ 19, 224, 48 },
		{ 19, 225, 48 },
		{ 19, 226, 48 },
		{ 19, 227, 48 },
		{ 20, 169, 48 },
		{ 20, 170, 48 },
		{ 20, 171, 48 },
		{ 20, 172, 48 },
		{ 20, 173, 48 },
		{ 20, 174, 48 },
		{ 20, 175, 48 },
		{ 20, 176, 48 },
		{ 20, 177, 48 },
		{ 20, 178, 48 },
		{ 20, 179, 48 },
		{ 20, 180, 48 },
		{ 20, 181, 48 },
		{ 20, 182, 48 },
		{ 20, 183, 48 },
		{ 20, 184, 48 },
		{ 20, 185, 48 },
		{ 20, 186, 48 },
		{ 20, 187, 48 },
		{ 20, 188, 48 },
		{ 20, 189, 48 },
		{ 20, 190, 48 },
		{ 20, 191, 48 },
		{ 20, 192, 48 },
		{ 20, 193, 48 },
		{ 20, 194, 48 },
		{ 20, 195, 48 },
		{ 20, 196, 48 },
		{ 20, 197, 48 },
		{ 20, 198, 48 },
		{ 20, 199, 48 },
		{ 20, 200, 48 },
		{ 20, 201, 48 },
		{ 20, 202, 48 },
		{ 20, 203, 48 },
		{ 20, 204, 48 },
		{ 20, 205, 48 },
		{ 20, 206, 48 },
		{ 20, 207, 48 },
		{ 20, 208, 48 },
		{ 20, 209, 48 },
		{ 20, 210, 48 },
		{ 20, 211, 48 },
		{ 20, 212, 48 },
		{ 20, 213, 48 },
		{ 20, 214, 48 },
		{ 20, 215, 48 },
		{ 20, 216, 48 },
		{ 20, 217, 48 },
		{ 20, 218, 48 },
		{ 20, 219, 48 },
		{ 20, 220, 48 },
		{ 20, 221, 48 },
		{ 20, 222, 48 },
		{ 20, 223, 48 },
		{ 20, 224, 48 },
		{ 20, 225, 48 },
		{ 20, 226, 48 },
		{ 20, 227, 48 },
		{ 21, 169, 48 },
		{ 21, 170, 48 },
		{ 21, 171, 48 },
		{ 21, 172, 48 },
		{ 21, 173, 48 },
		{ 21, 174, 48 },
		{ 21, 175, 48 },
		{ 21, 176, 48 },
		{ 21, 177, 48 },
		{ 21, 178, 48 },
		{ 21, 179, 48 },
		{ 21, 180, 48 },
		{ 21, 181, 48 },
		{ 21, 182, 48 },
		{ 21, 183, 48 },
		{ 21, 184, 48 },
		{ 21, 185, 48 },
		{ 21, 186, 48 },
		{ 21, 187, 48 },
		{ 21, 188, 48 },
		{ 21, 189, 48 },
		{ 21, 190, 48 },
		{ 21, 191, 48 },
		{ 21, 192, 48 },
		{ 21, 193, 48 },
		{ 21, 194, 48 },
		{ 21, 195, 48 },
		{ 21, 196, 48 },
		{ 21, 197, 48 },
		{ 21, 198, 48 },
		{ 21, 199, 48 },
		{ 21, 200, 48 },
		{ 21, 201, 48 },
		{ 21, 202, 48 },
		{ 21, 203, 48 },
		{ 21, 204, 48 },
		{ 21, 205, 48 },
		{ 21, 206, 48 },
		{ 21, 207, 48 },
		{ 21, 208, 48 },
		{ 21, 209, 48 },
		{ 21, 210, 48 },
		{ 21, 211, 48 },
		{ 21, 212, 48 },
		{ 21, 213, 48 },
		{ 21, 214, 48 },
		{ 21, 215, 48 },
		{ 21, 216, 48 },
		{ 21, 217, 48 },
		{ 21, 218, 48 },
		{ 21, 219, 48 },
		{ 21, 220, 48 },
		{ 21, 221, 48 },
		{ 21, 222, 48 },
		{ 21, 223, 48 },
		{ 21, 224, 48 },
		{ 21, 225, 48 },
		{ 21, 226, 48 },
		{ 21, 227, 48 },
		{ 22, 169, 48 },
		{ 22, 170, 48 },
		{ 22, 171, 48 },
		{ 22, 172, 48 },
		{ 22, 173, 48 },
		{ 22, 174, 48 },
		{ 22, 175, 48 },
		{ 22, 176, 48 },
		{ 22, 177, 48 },
		{ 22, 178, 48 },
		{ 22, 179, 48 },
		{ 22, 180, 48 },
		{ 22, 181, 48 },
		{ 22, 182, 48 },
		{ 22, 183, 48 },
		{ 22, 184, 48 },
		{ 22, 185, 48 },
		{ 22, 186, 48 },
		{ 22, 187, 48 },
		{ 22, 188, 48 },
		{ 22, 189, 48 },
		{ 22, 190, 48 },
		{ 22, 191, 48 },
		{ 22, 192, 48 },
		{ 22, 193, 48 },
		{ 22, 194, 48 },
		{ 22, 195, 48 },
		{ 22, 196, 48 },
		{ 22, 197, 48 },
		{ 22, 198, 48 },
		{ 22, 199, 48 },
		{ 22, 200, 48 },
		{ 22, 201, 48 },
		{ 22, 202, 48 },
		{ 22, 203, 48 },
		{ 22, 204, 48 },
		{ 22, 205, 48 },
		{ 22, 206, 48 },
		{ 22, 207, 48 },
		{ 22, 208, 48 },
		{ 22, 209, 48 },
		{ 22, 210, 48 },
		{ 22, 211, 48 },
		{ 22, 212, 48 },
		{ 22, 213, 48 },
		{ 22, 214, 48 },
		{ 22, 215, 48 },
		{ 22, 216, 48 },
		{ 22, 217, 48 },
		{ 22, 218, 48 },
		{ 22, 219, 48 },
		{ 22, 220, 48 },
		{ 22, 221, 48 },
		{ 22, 222, 48 },
		{ 22, 223, 48 },
		{ 22, 224, 48 },
		{ 22, 225, 48 },
		{ 22, 226, 48 },
		{ 22, 227, 48 },
		{ 23, 170, 48 },
		{ 23, 171, 48 },
		{ 23, 172, 48 },
		{ 23, 173, 48 },
		{ 23, 174, 48 },
		{ 23, 175, 48 },
		{ 23, 176, 48 },
		{ 23, 177, 48 },
		{ 23, 178, 48 },
		{ 23, 179, 48 },
		{ 23, 180, 48 },
		{ 23, 181, 48 },
		{ 23, 182, 48 },
		{ 23, 183, 48 },
		{ 23, 184, 48 },
		{ 23, 185, 48 },
		{ 23, 186, 48 },
		{ 23, 187, 48 },
		{ 23, 188, 48 },
		{ 23, 189, 48 },
		{ 23, 190, 48 },
		{ 23, 191, 48 },
		{ 23, 192, 48 },
		{ 23, 193, 48 },
		{ 23, 194, 48 },
		{ 23, 195, 48 },
		{ 23, 196, 48 },
		{ 23, 197, 48 },
		{ 23, 198, 48 },
		{ 23, 199, 48 },
		{ 23, 200, 48 },
		{ 23, 201, 48 },
		{ 23, 202, 48 },
		{ 23, 203, 48 },
		{ 23, 204, 48 },
		{ 23, 205, 48 },
		{ 23, 206, 48 },
		{ 23, 207, 48 },
		{ 23, 208, 48 },
		{ 23, 209, 48 },
		{ 23, 210, 48 },
		{ 23, 211, 48 },
		{ 23, 212, 48 },
		{ 23, 213, 48 },
		{ 23, 214, 48 },
		{ 23, 215, 48 },
		{ 23, 216, 48 },
		{ 23, 217, 48 },
		{ 23, 218, 48 },
		{ 23, 219, 48 },
		{ 23, 220, 48 },
		{ 23, 221, 48 },
		{ 23, 222, 48 },
		{ 23, 223, 48 },
		{ 23, 224, 48 },
		{ 23, 225, 48 },
		{ 23, 226, 48 },
		{ 23, 227, 48 },
		{ 24, 170, 48 },
		{ 24, 171, 48 },
		{ 24, 172, 48 },
		{ 24, 173, 48 },
		{ 24, 174, 48 },
		{ 24, 175, 48 },
		{ 24, 176, 48 },
		{ 24, 177, 48 },
		{ 24, 178, 48 },
		{ 24, 179, 48 },
		{ 24, 180, 48 },
		{ 24, 181, 48 },
		{ 24, 182, 48 },
		{ 24, 183, 48 },
		{ 24, 184, 48 },
		{ 24, 185, 48 },
		{ 24, 186, 48 },
		{ 24, 187, 48 },
		{ 24, 188, 48 },
		{ 24, 189, 48 },
		{ 24, 190, 48 },
		{ 24, 191, 48 },
		{ 24, 192, 48 },
		{ 24, 193, 48 },
		{ 24, 194, 48 },
		{ 24, 195, 48 },
		{ 24, 196, 48 },
		{ 24, 197, 48 },
		{ 24, 198, 48 },
		{ 24, 199, 48 },
		{ 24, 200, 48 },
		{ 24, 201, 48 },
		{ 24, 202, 48 },
		{ 24, 203, 48 },
		{ 24, 204, 48 },
		{ 24, 205, 48 },
		{ 24, 206, 48 },
		{ 24, 207, 48 },
		{ 24, 208, 48 },
		{ 24, 209, 48 },
		{ 24, 210, 48 },
		{ 24, 211, 48 },
		{ 24, 212, 48 },
		{ 24, 213, 48 },
		{ 24, 214, 48 },
		{ 24, 215, 48 },
		{ 24, 216, 48 },
		{ 24, 217, 48 },
		{ 24, 218, 48 },
		{ 24, 219, 48 },
		{ 24, 220, 48 },
		{ 24, 221, 48 },
		{ 24, 222, 48 },
		{ 24, 223, 48 },
		{ 24, 224, 48 },
		{ 24, 225, 48 },
		{ 24, 226, 48 },
		{ 24, 227, 48 },
		{ 25, 171, 48 },
		{ 25, 172, 48 },
		{ 25, 173, 48 },
		{ 25, 174, 48 },
		{ 25, 175, 48 },
		{ 25, 176, 48 },
		{ 25, 177, 48 },
		{ 25, 178, 48 },
		{ 25, 179, 48 },
		{ 25, 180, 48 },
		{ 25, 181, 48 },
		{ 25, 182, 48 },
		{ 25, 183, 48 },
		{ 25, 184, 48 },
		{ 25, 185, 48 },
		{ 25, 186, 48 },
		{ 25, 187, 48 },
		{ 25, 188, 48 },
		{ 25, 189, 48 },
		{ 25, 190, 48 },
		{ 25, 191, 48 },
		{ 25, 192, 48 },
		{ 25, 193, 48 },
		{ 25, 194, 48 },
		{ 25, 195, 48 },
		{ 25, 196, 48 },
		{ 25, 197, 48 },
		{ 25, 198, 48 },
		{ 25, 199, 48 },
		{ 25, 200, 48 },
		{ 25, 201, 48 },
		{ 25, 202, 48 },
		{ 25, 203, 48 },
		{ 25, 204, 48 },
		{ 25, 205, 48 },
		{ 25, 206, 48 },
		{ 25, 207, 48 },
		{ 25, 208, 48 },
		{ 25, 209, 48 },
		{ 25, 210, 48 },
		{ 25, 211, 48 },
		{ 25, 212, 48 },
		{ 25, 213, 48 },
		{ 25, 214, 48 },
		{ 25, 215, 48 },
		{ 25, 216, 48 },
		{ 25, 217, 48 },
		{ 25, 218, 48 },
		{ 25, 219, 48 },
		{ 25, 220, 48 },
		{ 25, 221, 48 },
		{ 25, 222, 48 },
		{ 25, 223, 48 },
		{ 25, 224, 48 },
		{ 25, 225, 48 },
		{ 25, 226, 48 },
		{ 25, 227, 48 },
		{ 26, 171, 48 },
		{ 26, 172, 48 },
		{ 26, 173, 48 },
		{ 26, 174, 48 },
		{ 26, 175, 48 },
		{ 26, 176, 48 },
		{ 26, 177, 48 },
		{ 26, 178, 48 },
		{ 26, 179, 48 },
		{ 26, 180, 48 },
		{ 26, 181, 48 },
		{ 26, 182, 48 },
		{ 26, 183, 48 },
		{ 26, 184, 48 },
		{ 26, 185, 48 },
		{ 26, 186, 48 },
		{ 26, 187, 48 },
		{ 26, 188, 48 },
		{ 26, 189, 48 },
		{ 26, 190, 48 },
		{ 26, 191, 48 },
		{ 26, 192, 48 },
		{ 26, 193, 48 },
		{ 26, 194, 48 },
		{ 26, 195, 48 },
		{ 26, 196, 48 },
		{ 26, 197, 48 },
		{ 26, 198, 48 },
		{ 26, 199, 48 },
		{ 26, 200, 48 },
		{ 26, 201, 48 },
		{ 26, 202, 48 },
		{ 26, 203, 48 },
		{ 26, 204, 48 },
		{ 26, 205, 48 },
		{ 26, 206, 48 },
		{ 26, 207, 48 },
		{ 26, 208, 48 },
		{ 26, 209, 48 },
		{ 26, 210, 48 },
		{ 26, 211, 48 },
		{ 26, 212, 48 },
		{ 26, 213, 48 },
		{ 26, 214, 48 },
		{ 26, 215, 48 },
		{ 26, 216, 48 },
		{ 26, 217, 48 },
		{ 26, 218, 48 },
		{ 26, 219, 48 },
		{ 26, 220, 48 },
		{ 26, 221, 48 },
		{ 26, 222, 48 },
		{ 26, 223, 48 },
		{ 26, 224, 48 },
		{ 26, 225, 48 },
		{ 26, 226, 48 },
		{ 26, 227, 48 },
		{ 26, 228, 48 },
		{ 27, 171, 48 },
		{ 27, 172, 48 },
		{ 27, 173, 48 },
		{ 27, 174, 48 },
		{ 27, 175, 48 },
		{ 27, 176, 48 },
		{ 27, 177, 48 },
		{ 27, 178, 48 },
		{ 27, 179, 48 },
		{ 27, 180, 48 },
		{ 27, 181, 48 },
		{ 27, 182, 48 },
		{ 27, 183, 48 },
		{ 27, 184, 48 },
		{ 27, 185, 48 },
		{ 27, 186, 48 },
		{ 27, 187, 48 },
		{ 27, 188, 48 },
		{ 27, 189, 48 },
		{ 27, 190, 48 },
		{ 27, 191, 48 },
		{ 27, 192, 48 },
		{ 27, 193, 48 },
		{ 27, 194, 48 },
		{ 27, 195, 48 },
		{ 27, 196, 48 },
		{ 27, 197, 48 },
		{ 27, 198, 48 },
		{ 27, 199, 48 },
		{ 27, 200, 48 },
		{ 27, 201, 48 },
		{ 27, 202, 48 },
		{ 27, 203, 48 },
		{ 27, 204, 48 },
		{ 27, 205, 48 },
		{ 27, 206, 48 },
		{ 27, 207, 48 },
		{ 27, 208, 48 },
		{ 27, 209, 48 },
		{ 27, 210, 48 },
		{ 27, 211, 48 },
		{ 27, 212, 48 },
		{ 27, 213, 48 },
		{ 27, 214, 48 },
		{ 27, 215, 48 },
		{ 27, 216, 48 },
		{ 27, 217, 48 },
		{ 27, 218, 48 },
		{ 27, 219, 48 },
		{ 27, 220, 48 },
		{ 27, 221, 48 },
		{ 27, 222, 48 },
		{ 27, 223, 48 },
		{ 27, 224, 48 },
		{ 27, 225, 48 },
		{ 27, 226, 48 },
		{ 27, 227, 48 },
		{ 27, 228, 48 },
		{ 28, 172, 48 },
		{ 28, 173, 48 },
		{ 28, 174, 48 },
		{ 28, 175, 48 },
		{ 28, 176, 48 },
		{ 28, 177, 48 },
		{ 28, 178, 48 },
		{ 28, 179, 48 },
		{ 28, 180, 48 },
		{ 28, 181, 48 },
		{ 28, 182, 48 },
		{ 28, 183, 48 },
		{ 28, 184, 48 },
		{ 28, 185, 48 },
		{ 28, 186, 48 },
		{ 28, 187, 48 },
		{ 28, 188, 48 },
		{ 28, 189, 48 },
		{ 28, 190, 48 },
		{ 28, 191, 48 },
		{ 28, 192, 48 },
		{ 28, 193, 48 },
		{ 28, 194, 48 },
		{ 28, 195, 48 },
		{ 28, 196, 48 },
		{ 28, 197, 48 },
		{ 28, 198, 48 },
		{ 28, 199, 48 },
		{ 28, 200, 48 },
		{ 28, 201, 48 },
		{ 28, 202, 48 },
		{ 28, 203, 48 },
		{ 28, 204, 48 },
		{ 28, 205, 48 },
		{ 28, 206, 48 },
		{ 28, 207, 48 },
		{ 28, 208, 48 },
		{ 28, 209, 48 },
		{ 28, 210, 48 },
		{ 28, 211, 48 },
		{ 28, 212, 48 },
		{ 28, 213, 48 },
		{ 28, 214, 48 },
		{ 28, 215, 48 },
		{ 28, 216, 48 },
		{ 28, 217, 48 },
		{ 28, 218, 48 },
		{ 28, 219, 48 },
		{ 28, 220, 48 },
		{ 28, 221, 48 },
		{ 28, 222, 48 },
		{ 28, 223, 48 },
		{ 28, 224, 48 },
		{ 28, 225, 48 },
		{ 28, 226, 48 },
		{ 28, 227, 48 },
		{ 28, 228, 48 },
		{ 29, 172, 48 },
		{ 29, 173, 48 },
		{ 29, 174, 48 },
		{ 29, 175, 48 },
		{ 29, 176, 48 },
		{ 29, 177, 48 },
		{ 29, 178, 48 },
		{ 29, 179, 48 },
		{ 29, 180, 48 },
		{ 29, 181, 48 },
		{ 29, 182, 48 },
		{ 29, 183, 48 },
		{ 29, 184, 48 },
		{ 29, 185, 48 },
		{ 29, 186, 48 },
		{ 29, 187, 48 },
		{ 29, 188, 48 },
		{ 29, 189, 48 },
		{ 29, 190, 48 },
		{ 29, 191, 48 },
		{ 29, 192, 48 },
		{ 29, 193, 48 },
		{ 29, 194, 48 },
		{ 29, 195, 48 },
		{ 29, 196, 48 },
		{ 29, 197, 48 },
		{ 29, 198, 48 },
		{ 29, 199, 48 },
		{ 29, 200, 48 },
		{ 29, 201, 48 },
		{ 29, 202, 48 },
		{ 29, 203, 48 },
		{ 29, 204, 48 },
		{ 29, 205, 48 },
		{ 29, 206, 48 },
		{ 29, 207, 48 },
		{ 29, 208, 48 },
		{ 29, 209, 48 },
		{ 29, 210, 48 },
		{ 29, 211, 48 },
		{ 29, 212, 48 },
		{ 29, 213, 48 },
		{ 29, 214, 48 },
		{ 29, 215, 48 },
		{ 29, 216, 48 },
		{ 29, 217, 48 },
		{ 29, 218, 48 },
		{ 29, 219, 48 },
		{ 29, 220, 48 },
		{ 29, 221, 48 },
		{ 29, 222, 48 },
		{ 29, 223, 48 },
		{ 29, 224, 48 },
		{ 29, 225, 48 },
		{ 29, 226, 48 },
		{ 29, 227, 48 },
		{ 29, 228, 48 },
		{ 30, 172, 48 },
		{ 30, 173, 48 },
		{ 30, 174, 48 },
		{ 30, 175, 48 },
		{ 30, 176, 48 },
		{ 30, 177, 48 },
		{ 30, 178, 48 },
		{ 30, 179, 48 },
		{ 30, 180, 48 },
		{ 30, 181, 48 },
		{ 30, 182, 48 },
		{ 30, 183, 48 },
		{ 30, 184, 48 },
		{ 30, 185, 48 },
		{ 30, 186, 48 },
		{ 30, 187, 48 },
		{ 30, 188, 48 },
		{ 30, 189, 48 },
		{ 30, 190, 48 },
		{ 30, 191, 48 },
		{ 30, 192, 48 },
		{ 30, 193, 48 },
		{ 30, 194, 48 },
		{ 30, 195, 48 },
		{ 30, 196, 48 },
		{ 30, 197, 48 },
		{ 30, 198, 48 },
		{ 30, 199, 48 },
		{ 30, 200, 48 },
		{ 30, 201, 48 },
		{ 30, 202, 48 },
		{ 30, 203, 48 },
		{ 30, 204, 48 },
		{ 30, 205, 48 },
		{ 30, 206, 48 },
		{ 30, 207, 48 },
		{ 30, 208, 48 },
		{ 30, 209, 48 },
		{ 30, 210, 48 },
		{ 30, 211, 48 },
		{ 30, 212, 48 },
		{ 30, 213, 48 },
		{ 30, 214, 48 },
		{ 30, 215, 48 },
		{ 30, 216, 48 },
		{ 30, 217, 48 },
		{ 30, 218, 48 },
		{ 30, 219, 48 },
		{ 30, 220, 48 },
		{ 30, 221, 48 },
		{ 30, 222, 48 },
		{ 30, 223, 48 },
		{ 30, 224, 48 },
		{ 30, 225, 48 },
		{ 30, 226, 48 },
		{ 30, 227, 48 },
		{ 30, 228, 48 },
		{ 31, 173, 48 },
		{ 31, 174, 48 },
		{ 31, 175, 48 },
		{ 31, 176, 48 },
		{ 31, 177, 48 },
		{ 31, 178, 48 },
		{ 31, 179, 48 },
		{ 31, 180, 48 },
		{ 31, 181, 48 },
		{ 31, 182, 48 },
		{ 31, 183, 48 },
		{ 31, 184, 48 },
		{ 31, 185, 48 },
		{ 31, 186, 48 },
		{ 31, 187, 48 },
		{ 31, 188, 48 },
		{ 31, 189, 48 },
		{ 31, 190, 48 },
		{ 31, 191, 48 },
		{ 31, 192, 48 },
		{ 31, 193, 48 },
		{ 31, 194, 48 },
		{ 31, 195, 48 },
		{ 31, 196, 48 },
		{ 31, 197, 48 },
		{ 31, 198, 48 },
		{ 31, 199, 48 },
		{ 31, 200, 48 },
		{ 31, 201, 48 },
		{ 31, 202, 48 },
		{ 31, 203, 48 },
		{ 31, 204, 48 },
		{ 31, 205, 48 },
		{ 31, 206, 48 },
		{ 31, 207, 48 },
		{ 31, 208, 48 },
		{ 31, 209, 48 },
		{ 31, 210, 48 },
		{ 31, 211, 48 },
		{ 31, 212, 48 },
		{ 31, 213, 48 },
		{ 31, 214, 48 },
		{ 31, 215, 48 },
		{ 31, 216, 48 },
		{ 31, 217, 48 },
		{ 31, 218, 48 },
		{ 31, 219, 48 },
		{ 31, 220, 48 },
		{ 31, 221, 48 },
		{ 31, 222, 48 },
		{ 31, 223, 48 },
		{ 31, 224, 48 },
		{ 31, 225, 48 },
		{ 31, 226, 48 },
		{ 31, 227, 48 },
		{ 31, 228, 48 },
		{ 32, 173, 48 },
		{ 32, 174, 48 },
		{ 32, 175, 48 },
		{ 32, 176, 48 },
		{ 32, 177, 48 },
		{ 32, 178, 48 },
		{ 32, 179, 48 },
		{ 32, 180, 48 },
		{ 32, 181, 48 },
		{ 32, 182, 48 },
		{ 32, 183, 48 },
		{ 32, 184, 48 },
		{ 32, 185, 48 },
		{ 32, 186, 48 },
		{ 32, 187, 48 },
		{ 32, 188, 48 },
		{ 32, 189, 48 },
		{ 32, 190, 48 },
		{ 32, 191, 48 },
		{ 32, 192, 48 },
		{ 32, 193, 48 },
		{ 32, 194, 48 },
		{ 32, 195, 48 },
		{ 32, 196, 48 },
		{ 32, 197, 48 },
		{ 32, 198, 48 },
		{ 32, 199, 48 },
		{ 32, 200, 48 },
		{ 32, 201, 48 },
		{ 32, 202, 48 },
		{ 32, 203, 48 },
		{ 32, 204, 48 },
		{ 32, 205, 48 },
		{ 32, 206, 48 },
		{ 32, 207, 48 },
		{ 32, 208, 48 },
		{ 32, 209, 48 },
		{ 32, 210, 48 },
		{ 32, 211, 48 },
		{ 32, 212, 48 },
		{ 32, 213, 48 },
		{ 32, 214, 48 },
		{ 32, 215, 48 },
		{ 32, 216, 48 },
		{ 32, 217, 48 },
		{ 32, 218, 48 },
		{ 32, 219, 48 },
		{ 32, 220, 48 },
		{ 32, 221, 48 },
		{ 32, 222, 48 },
		{ 32, 223, 48 },
		{ 32, 224, 48 },
		{ 32, 225, 48 },
		{ 32, 226, 48 },
		{ 32, 227, 48 },
		{ 32, 228, 48 },
		{ 33, 174, 48 },
		{ 33, 175, 48 },
		{ 33, 176, 48 },
		{ 33, 177, 48 },
		{ 33, 178, 48 },
		{ 33, 179, 48 },
		{ 33, 180, 48 },
		{ 33, 181, 48 },
		{ 33, 182, 48 },
		{ 33, 183, 48 },
		{ 33, 184, 48 },
		{ 33, 185, 48 },
		{ 33, 186, 48 },
		{ 33, 187, 48 },
		{ 33, 188, 48 },
		{ 33, 189, 48 },
		{ 33, 190, 48 },
		{ 33, 191, 48 },
		{ 33, 192, 48 },
		{ 33, 193, 48 },
		{ 33, 194, 48 },
		{ 33, 195, 48 },
		{ 33, 196, 48 },
		{ 33, 197, 48 },
		{ 33, 198, 48 },
		{ 33, 199, 48 },
		{ 33, 200, 48 },
		{ 33, 201, 48 },
		{ 33, 202, 48 },
		{ 33, 203, 48 },
		{ 33, 204, 48 },
		{ 33, 205, 48 },
		{ 33, 206, 48 },
		{ 33, 207, 48 },
		{ 33, 208, 48 },
		{ 33, 209, 48 },
		{ 33, 210, 48 },
		{ 33, 211, 48 },
		{ 33, 212, 48 },
		{ 33, 213, 48 },
		{ 33, 214, 48 },
		{ 33, 215, 48 },
		{ 33, 216, 48 },
		{ 33, 217, 48 },
		{ 33, 218, 48 },
		{ 33, 219, 48 },
		{ 33, 220, 48 },
		{ 33, 221, 48 },
		{ 33, 222, 48 },
		{ 33, 223, 48 },
		{ 33, 224, 48 },
		{ 33, 225, 48 },
		{ 33, 226, 48 },
		{ 33, 227, 48 },
		{ 33, 228, 48 },
		{ 34, 174, 48 },
		{ 34, 175, 48 },
		{ 34, 176, 48 },
		{ 34, 177, 48 },
		{ 34, 178, 48 },
		{ 34, 179, 48 },
		{ 34, 180, 48 },
		{ 34, 181, 48 },
		{ 34, 182, 48 },
		{ 34, 183, 48 },
		{ 34, 184, 48 },
		{ 34, 185, 48 },
		{ 34, 186, 48 },
		{ 34, 187, 48 },
		{ 34, 188, 48 },
		{ 34, 189, 48 },
		{ 34, 190, 48 },
		{ 34, 191, 48 },
		{ 34, 192, 48 },
		{ 34, 193, 48 },
		{ 34, 194, 48 },
		{ 34, 195, 48 },
		{ 34, 196, 48 },
		{ 34, 197, 48 },
		{ 34, 198, 48 },
		{ 34, 199, 48 },
		{ 34, 200, 48 },
		{ 34, 201, 48 },
		{ 34, 202, 48 },
		{ 34, 203, 48 },
		{ 34, 204, 48 },
		{ 34, 205, 48 },
		{ 34, 206, 48 },
		{ 34, 207, 48 },
		{ 34, 208, 48 },
		{ 34, 209, 48 },
		{ 34, 210, 48 },
		{ 34, 211, 48 },
		{ 34, 212, 48 },
		{ 34, 213, 48 },
		{ 34, 214, 48 },
		{ 34, 215, 48 },
		{ 34, 216, 48 },
		{ 34, 217, 48 },
		{ 34, 218, 48 },
		{ 34, 219, 48 },
		{ 34, 220, 48 },
		{ 34, 221, 48 },
		{ 34, 222, 48 },
		{ 34, 223, 48 },
		{ 34, 224, 48 },
		{ 34, 225, 48 },
		{ 34, 226, 48 },
		{ 34, 227, 48 },
		{ 34, 228, 48 },
		{ 34, 229, 48 },
		{ 35, 174, 48 },
		{ 35, 175, 48 },
		{ 35, 176, 48 },
		{ 35, 177, 48 },
		{ 35, 178, 48 },
		{ 35, 179, 48 },
		{ 35, 180, 48 },
		{ 35, 181, 48 },
		{ 35, 182, 48 },
		{ 35, 183, 48 },
		{ 35, 184, 48 },
		{ 35, 185, 48 },
		{ 35, 186, 48 },
		{ 35, 187, 48 },
		{ 35, 188, 48 },
		{ 35, 189, 48 },
		{ 35, 190, 48 },
		{ 35, 191, 48 },
		{ 35, 192, 48 },
		{ 35, 193, 48 },
		{ 35, 194, 48 },
		{ 35, 195, 48 },
		{ 35, 196, 48 },
		{ 35, 197, 48 },
		{ 35, 198, 48 },
		{ 35, 199, 48 },
		{ 35, 200, 48 },
		{ 35, 201, 48 },
		{ 35, 202, 48 },
		{ 35, 203, 48 },
		{ 35, 204, 48 },
		{ 35, 205, 48 },
		{ 35, 206, 48 },
		{ 35, 207, 48 },
		{ 35, 208, 48 },
		{ 35, 209, 48 },
		{ 35, 210, 48 },
		{ 35, 211, 48 },
		{ 35, 212, 48 },
		{ 35, 213, 48 },
		{ 35, 214, 48 },
		{ 35, 215, 48 },
		{ 35, 216, 48 },
		{ 35, 217, 48 },
		{ 35, 218, 48 },
		{ 35, 219, 48 },
		{ 35, 220, 48 },
		{ 35, 221, 48 },
		{ 35, 222, 48 },
		{ 35, 223, 48 },
		{ 35, 224, 48 },
		{ 35, 225, 48 },
		{ 35, 226, 48 },
		{ 35, 227, 48 },
		{ 35, 228, 48 },
		{ 35, 229, 48 },
		{ 36, 175, 48 },
		{ 36, 176, 48 },
		{ 36, 177, 48 },
		{ 36, 178, 48 },
		{ 36, 179, 48 },
		{ 36, 180, 48 },
		{ 36, 181, 48 },
		{ 36, 182, 48 },
		{ 36, 183, 48 },
		{ 36, 184, 48 },
		{ 36, 185, 48 },
		{ 36, 186, 48 },
		{ 36, 187, 48 },
		{ 36, 188, 48 },
		{ 36, 189, 48 },
		{ 36, 190, 48 },
		{ 36, 191, 48 },
		{ 36, 192, 48 },
		{ 36, 193, 48 },
		{ 36, 194, 48 },
		{ 36, 195, 48 },
		{ 36, 196, 48 },
		{ 36, 197, 48 },
		{ 36, 198, 48 },
		{ 36, 199, 48 },
		{ 36, 200, 48 },
		{ 36, 201, 48 },
		{ 36, 202, 48 },
		{ 36, 203, 48 },
		{ 36, 204, 48 },
		{ 36, 205, 48 },
		{ 36, 206, 48 },
		{ 36, 207, 48 },
		{ 36, 208, 48 },
		{ 36, 209, 48 },
		{ 36, 210, 48 },
		{ 36, 211, 48 },
		{ 36, 212, 48 },
		{ 36, 213, 48 },
		{ 36, 214, 48 },
		{ 36, 215, 48 },
		{ 36, 216, 48 },
		{ 36, 217, 48 },
		{ 36, 218, 48 },
		{ 36, 219, 48 },
		{ 36, 220, 48 },
		{ 36, 221, 48 },
		{ 36, 222, 48 },
		{ 36, 223, 48 },
		{ 36, 224, 48 },
		{ 36, 225, 48 },
		{ 36, 226, 48 },
		{ 36, 227, 48 },
		{ 36, 228, 48 },
		{ 36, 229, 48 },
		{ 37, 175, 48 },
		{ 37, 176, 48 },
		{ 37, 177, 48 },
		{ 37, 178, 48 },
		{ 37, 179, 48 },
		{ 37, 180, 48 },
		{ 37, 181, 48 },
		{ 37, 182, 48 },
		{ 37, 183, 48 },
		{ 37, 184, 48 },
		{ 37, 185, 48 },
		{ 37, 186, 48 },
		{ 37, 187, 48 },
		{ 37, 188, 48 },
		{ 37, 189, 48 },
		{ 37, 190, 48 },
		{ 37, 191, 48 },
		{ 37, 192, 48 },
		{ 37, 193, 48 },
		{ 37, 194, 48 },
		{ 37, 195, 48 },
		{ 37, 196, 48 },
		{ 37, 197, 48 },
		{ 37, 198, 48 },
		{ 37, 199, 48 },
		{ 37, 200, 48 },
		{ 37, 201, 48 },
		{ 37, 202, 48 },
		{ 37, 203, 48 },
		{ 37, 204, 48 },
		{ 37, 205, 48 },
		{ 37, 206, 48 },
		{ 37, 207, 48 },
		{ 37, 208, 48 },
		{ 37, 209, 48 },
		{ 37, 210, 48 },
		{ 37, 211, 48 },
		{ 37, 212, 48 },
		{ 37, 213, 48 },
		{ 37, 214, 48 },
		{ 37, 215, 48 },
		{ 37, 216, 48 },
		{ 37, 217, 48 },
		{ 37, 218, 48 },
		{ 37, 219, 48 },
		{ 37, 220, 48 },
		{ 37, 221, 48 },
		{ 37, 222, 48 },
		{ 37, 223, 48 },
		{ 37, 224, 48 },
		{ 37, 225, 48 },
		{ 37, 226, 48 },
		{ 37, 227, 48 },
		{ 37, 228, 48 },
		{ 37, 229, 48 },
		{ 38, 175, 48 },
		{ 38, 176, 48 },
		{ 38, 177, 48 },
		{ 38, 178, 48 },
		{ 38, 179, 48 },
		{ 38, 180, 48 },
		{ 38, 181, 48 },
		{ 38, 182, 48 },
		{ 38, 183, 48 },
		{ 38, 184, 48 },
		{ 38, 185, 48 },
		{ 38, 186, 48 },
		{ 38, 187, 48 },
		{ 38, 188, 48 },
		{ 38, 189, 48 },
		{ 38, 190, 48 },
		{ 38, 191, 48 },
		{ 38, 192, 48 },
		{ 38, 193, 48 },
		{ 38, 194, 48 },
		{ 38, 195, 48 },
		{ 38, 196, 48 },
		{ 38, 197, 48 },
		{ 38, 198, 48 },
		{ 38, 199, 48 },
		{ 38, 200, 48 },
		{ 38, 201, 48 },
		{ 38, 202, 48 },
		{ 38, 203, 48 },
		{ 38, 204, 48 },
		{ 38, 205, 48 },
		{ 38, 206, 48 },
		{ 38, 207, 48 },
		{ 38, 208, 48 },
		{ 38, 209, 48 },
		{ 38, 210, 48 },
		{ 38, 211, 48 },
		{ 38, 212, 48 },
		{ 38, 213, 48 },
		{ 38, 214, 48 },
		{ 38, 215, 48 },
		{ 38, 216, 48 },
		{ 38, 217, 48 },
		{ 38, 218, 48 },
		{ 38, 219, 48 },
		{ 38, 220, 48 },
		{ 38, 221, 48 },
		{ 38, 222, 48 },
		{ 38, 223, 48 },
		{ 38, 224, 48 },
		{ 38, 225, 48 },
		{ 38, 226, 48 },
		{ 38, 227, 48 },
		{ 38, 228, 48 },
		{ 38, 229, 48 },
		{ 39, 176, 48 },
		{ 39, 177, 48 },
		{ 39, 178, 48 },
		{ 39, 179, 48 },
		{ 39, 180, 48 },
		{ 39, 181, 48 },
		{ 39, 182, 48 },
		{ 39, 183, 48 },
		{ 39, 184, 48 },
		{ 39, 185, 48 },
		{ 39, 186, 48 },
		{ 39, 187, 48 },
		{ 39, 188, 48 },
		{ 39, 189, 48 },
		{ 39, 190, 48 },
		{ 39, 191, 48 },
		{ 39, 192, 48 },
		{ 39, 193, 48 },
		{ 39, 194, 48 },
		{ 39, 195, 48 },
		{ 39, 196, 48 },
		{ 39, 197, 48 },
		{ 39, 198, 48 },
		{ 39, 199, 48 },
		{ 39, 200, 48 },
		{ 39, 201, 48 },
		{ 39, 202, 48 },
		{ 39, 203, 48 },
		{ 39, 204, 48 },
		{ 39, 205, 48 },
		{ 39, 206, 48 },
		{ 39, 207, 48 },
		{ 39, 208, 48 },
		{ 39, 209, 48 },
		{ 39, 210, 48 },
		{ 39, 211, 48 },
		{ 39, 212, 48 },
		{ 39, 213, 48 },
		{ 39, 214, 48 },
		{ 39, 215, 48 },
		{ 39, 216, 48 },
		{ 39, 217, 48 },
		{ 39, 218, 48 },
		{ 39, 219, 48 },
		{ 39, 220, 48 },
		{ 39, 221, 48 },
		{ 39, 222, 48 },
		{ 39, 223, 48 },
		{ 39, 224, 48 },
		{ 39, 225, 48 },
		{ 39, 226, 48 },
		{ 39, 227, 48 },
		{ 39, 228, 48 },
		{ 39, 229, 48 },
		{ 40, 176, 48 },
		{ 40, 177, 48 },
		{ 40, 178, 48 },
		{ 40, 179, 48 },
		{ 40, 180, 48 },
		{ 40, 181, 48 },
		{ 40, 182, 48 },
		{ 40, 183, 48 },
		{ 40, 184, 48 },
		{ 40, 185, 48 },
		{ 40, 186, 48 },
		{ 40, 187, 48 },
		{ 40, 188, 48 },
		{ 40, 189, 48 },
		{ 40, 190, 48 },
		{ 40, 191, 48 },
		{ 40, 192, 48 },
		{ 40, 193, 48 },
		{ 40, 194, 48 },
		{ 40, 195, 48 },
		{ 40, 196, 48 },
		{ 40, 197, 48 },
		{ 40, 198, 48 },
		{ 40, 199, 48 },
		{ 40, 200, 48 },
		{ 40, 201, 48 },
		{ 40, 202, 48 },
		{ 40, 203, 48 },
		{ 40, 204, 48 },
		{ 40, 205, 48 },
		{ 40, 206, 48 },
		{ 40, 207, 48 },
		{ 40, 208, 48 },
		{ 40, 209, 48 },
		{ 40, 210, 48 },
		{ 40, 211, 48 },
		{ 40, 212, 48 },
		{ 40, 213, 48 },
		{ 40, 214, 48 },
		{ 40, 215, 48 },
		{ 40, 216, 48 },
		{ 40, 217, 48 },
		{ 40, 218, 48 },
		{ 40, 219, 48 },
		{ 40, 220, 48 },
		{ 40, 221, 48 },
		{ 40, 222, 48 },
		{ 40, 223, 48 },
		{ 40, 224, 48 },
		{ 40, 225, 48 },
		{ 40, 226, 48 },
		{ 40, 227, 48 },
		{ 40, 228, 48 },
		{ 40, 229, 48 },
		{ 41, 177, 48 },
		{ 41, 178, 48 },
		{ 41, 179, 48 },
		{ 41, 180, 48 },
		{ 41, 181, 48 },
		{ 41, 182, 48 },
		{ 41, 183, 48 },
		{ 41, 184, 48 },
		{ 41, 185, 48 },
		{ 41, 186, 48 },
		{ 41, 187, 48 },
		{ 41, 188, 48 },
		{ 41, 189, 48 },
		{ 41, 190, 48 },
		{ 41, 191, 48 },
		{ 41, 192, 48 },
		{ 41, 193, 48 },
		{ 41, 194, 48 },
		{ 41, 195, 48 },
		{ 41, 196, 48 },
		{ 41, 197, 48 },
		{ 41, 198, 48 },
		{ 41, 199, 48 },
		{ 41, 200, 48 },
		{ 41, 201, 48 },
		{ 41, 202, 48 },
		{ 41, 203, 48 },
		{ 41, 204, 48 },
		{ 41, 205, 48 },
		{ 41, 206, 48 },
		{ 41, 207, 48 },
		{ 41, 208, 48 },
		{ 41, 209, 48 },
		{ 41, 210, 48 },
		{ 41, 211, 48 },
		{ 41, 212, 48 },
		{ 41, 213, 48 },
		{ 41, 214, 48 },
		{ 41, 215, 48 },
		{ 41, 216, 48 },
		{ 41, 217, 48 },
		{ 41, 218, 48 },
		{ 41, 219, 48 },
		{ 41, 220, 48 },
		{ 41, 221, 48 },
		{ 41, 222, 48 },
		{ 41, 223, 48 },
		{ 41, 224, 48 },
		{ 41, 225, 48 },
		{ 41, 226, 48 },
		{ 41, 227, 48 },
		{ 41, 228, 48 },
		{ 41, 229, 48 },
		{ 42, 177, 48 },
		{ 42, 178, 48 },
		{ 42, 179, 48 },
		{ 42, 180, 48 },
		{ 42, 181, 48 },
		{ 42, 182, 48 },
		{ 42, 183, 48 },
		{ 42, 184, 48 },
		{ 42, 185, 48 },
		{ 42, 186, 48 },
		{ 42, 187, 48 },
		{ 42, 188, 48 },
		{ 42, 189, 48 },
		{ 42, 190, 48 },
		{ 42, 191, 48 },
		{ 42, 192, 48 },
		{ 42, 193, 48 },
		{ 42, 194, 48 },
		{ 42, 195, 48 },
		{ 42, 196, 48 },
		{ 42, 197, 48 },
		{ 42, 198, 48 },
		{ 42, 199, 48 },
		{ 42, 200, 48 },
		{ 42, 201, 48 },
		{ 42, 202, 48 },
		{ 42, 203, 48 },
		{ 42, 204, 48 },
		{ 42, 205, 48 },
		{ 42, 206, 48 },
		{ 42, 207, 48 },
		{ 42, 208, 48 },
		{ 42, 209, 48 },
		{ 42, 210, 48 },
		{ 42, 211, 48 },
		{ 42, 212, 48 },
		{ 42, 213, 48 },
		{ 42, 214, 48 },
		{ 42, 215, 48 },
		{ 42, 216, 48 },
		{ 42, 217, 48 },
		{ 42, 218, 48 },
		{ 42, 219, 48 },
		{ 42, 220, 48 },
		{ 42, 221, 48 },
		{ 42, 222, 48 },
		{ 42, 223, 48 },
		{ 42, 224, 48 },
		{ 42, 225, 48 },
		{ 42, 226, 48 },
		{ 42, 227, 48 },
		{ 42, 228, 48 },
		{ 42, 229, 48 },
		{ 42, 230, 48 },
		{ 43, 177, 48 },
		{ 43, 178, 48 },
		{ 43, 179, 48 },
		{ 43, 180, 48 },
		{ 43, 181, 48 },
		{ 43, 182, 48 },
		{ 43, 183, 48 },
		{ 43, 184, 48 },
		{ 43, 185, 48 },
		{ 43, 186, 48 },
		{ 43, 187, 48 },
		{ 43, 188, 48 },
		{ 43, 189, 48 },
		{ 43, 190, 48 },
		{ 43, 191, 48 },
		{ 43, 192, 48 },
		{ 43, 193, 48 },
		{ 43, 194, 48 },
		{ 43, 195, 48 },
		{ 43, 196, 48 },
		{ 43, 197, 48 },
		{ 43, 198, 48 },
		{ 43, 199, 48 },
		{ 43, 200, 48 },
		{ 43, 201, 48 },
		{ 43, 202, 48 },
		{ 43, 203, 48 },
		{ 43, 204, 48 },
		{ 43, 205, 48 },
		{ 43, 206, 48 },
		{ 43, 207, 48 },
		{ 43, 208, 48 },
		{ 43, 209, 48 },
		{ 43, 210, 48 },
		{ 43, 211, 48 },
		{ 43, 212, 48 },
		{ 43, 213, 48 },
		{ 43, 214, 48 },
		{ 43, 215, 48 },
		{ 43, 216, 48 },
		{ 43, 217, 48 },
		{ 43, 218, 48 },
		{ 43, 219, 48 },
		{ 43, 220, 48 },
		{ 43, 221, 48 },
		{ 43, 222, 48 },
		{ 43, 223, 48 },
		{ 43, 224, 48 },
		{ 43, 225, 48 },
		{ 43, 226, 48 },
		{ 43, 227, 48 },
		{ 43, 228, 48 },
		{ 43, 229, 48 },
		{ 43, 230, 48 },
		{ 44, 178, 48 },
		{ 44, 179, 48 },
		{ 44, 180, 48 },
		{ 44, 181, 48 },
		{ 44, 182, 48 },
		{ 44, 183, 48 },
		{ 44, 184, 48 },
		{ 44, 185, 48 },
		{ 44, 186, 48 },
		{ 44, 187, 48 },
		{ 44, 188, 48 },
		{ 44, 189, 48 },
		{ 44, 190, 48 },
		{ 44, 191, 48 },
		{ 44, 192, 48 },
		{ 44, 193, 48 },
		{ 44, 194, 48 },
		{ 44, 195, 48 },
		{ 44, 196, 48 },
		{ 44, 197, 48 },
		{ 44, 198, 48 },
		{ 44, 199, 48 },
		{ 44, 200, 48 },
		{ 44, 201, 48 },
		{ 44, 202, 48 },
		{ 44, 203, 48 },
		{ 44, 204, 48 },
		{ 44, 205, 48 },
		{ 44, 206, 48 },
		{ 44, 207, 48 },
		{ 44, 208, 48 },
		{ 44, 209, 48 },
		{ 44, 210, 48 },
		{ 44, 211, 48 },
		{ 44, 212, 48 },
		{ 44, 213, 48 },
		{ 44, 214, 48 },
		{ 44, 215, 48 },
		{ 44, 216, 48 },
		{ 44, 217, 48 },
		{ 44, 218, 48 },
		{ 44, 219, 48 },
		{ 44, 220, 48 },
		{ 44, 221, 48 },
		{ 44, 222, 48 },
		{ 44, 223, 48 },
		{ 44, 224, 48 },
		{ 44, 225, 48 },
		{ 44, 226, 48 },
		{ 44, 227, 48 },
		{ 44, 228, 48 },
		{ 44, 229, 48 },
		{ 44, 230, 48 },
		{ 45, 178, 48 },
		{ 45, 179, 48 },
		{ 45, 180, 48 },
		{ 45, 181, 48 },
		{ 45, 182, 48 },
		{ 45, 183, 48 },
		{ 45, 184, 48 },
		{ 45, 185, 48 },
		{ 45, 186, 48 },
		{ 45, 187, 48 },
		{ 45, 188, 48 },
		{ 45, 189, 48 },
		{ 45, 190, 48 },
		{ 45, 191, 48 },
		{ 45, 192, 48 },
		{ 45, 193, 48 },
		{ 45, 194, 48 },
		{ 45, 195, 48 },
		{ 45, 196, 48 },
		{ 45, 197, 48 },
		{ 45, 198, 48 },
		{ 45, 199, 48 },
		{ 45, 200, 48 },
		{ 45, 201, 48 },
		{ 45, 202, 48 },
		{ 45, 203, 48 },
		{ 45, 204, 48 },
		{ 45, 205, 48 },
		{ 45, 206, 48 },
		{ 45, 207, 48 },
		{ 45, 208, 48 },
		{ 45, 209, 48 },
		{ 45, 210, 48 },
		{ 45, 211, 48 },
		{ 45, 212, 48 },
		{ 45, 213, 48 },
		{ 45, 214, 48 },
		{ 45, 215, 48 },
		{ 45, 216, 48 },
		{ 45, 217, 48 },
		{ 45, 218, 48 },
		{ 45, 219, 48 },
		{ 45, 220, 48 },
		{ 45, 221, 48 },
		{ 45, 222, 48 },
		{ 45, 223, 48 },
		{ 45, 224, 48 },
		{ 45, 225, 48 },
		{ 45, 226, 48 },
		{ 45, 227, 48 },
		{ 45, 228, 48 },
		{ 45, 229, 48 },
		{ 45, 230, 48 },
		{ 46, 178, 48 },
		{ 46, 179, 48 },
		{ 46, 180, 48 },
		{ 46, 181, 48 },
		{ 46, 182, 48 },
		{ 46, 183, 48 },
		{ 46, 184, 48 },
		{ 46, 185, 48 },
		{ 46, 186, 48 },
		{ 46, 187, 48 },
		{ 46, 188, 48 },
		{ 46, 189, 48 },
		{ 46, 190, 48 },
		{ 46, 191, 48 },
		{ 46, 192, 48 },
		{ 46, 193, 48 },
		{ 46, 194, 48 },
		{ 46, 195, 48 },
		{ 46, 196, 48 },
		{ 46, 197, 48 },
		{ 46, 198, 48 },
		{ 46, 199, 48 },
		{ 46, 200, 48 },
		{ 46, 201, 48 },
		{ 46, 202, 48 },
		{ 46, 203, 48 },
		{ 46, 204, 48 },
		{ 46, 205, 48 },
		{ 46, 206, 48 },
		{ 46, 207, 48 },
		{ 46, 208, 48 },
		{ 46, 209, 48 },
		{ 46, 210, 48 },
		{ 46, 211, 48 },
		{ 46, 212, 48 },
		{ 46, 213, 48 },
		{ 46, 214, 48 },
		{ 46, 215, 48 },
		{ 46, 216, 48 },
		{ 46, 217, 48 },
		{ 46, 218, 48 },
		{ 46, 219, 48 },
		{ 46, 220, 48 },
		{ 46, 221, 48 },
		{ 46, 222, 48 },
		{ 46, 223, 48 },
		{ 46, 224, 48 },
		{ 46, 225, 48 },
		{ 46, 226, 48 },
		{ 46, 227, 48 },
		{ 46, 228, 48 },
		{ 46, 229, 48 },
		{ 46, 230, 48 },
		{ 47, 179, 48 },
		{ 47, 180, 48 },
		{ 47, 181, 48 },
		{ 47, 182, 48 },
		{ 47, 183, 48 },
		{ 47, 184, 48 },
		{ 47, 185, 48 },
		{ 47, 186, 48 },
		{ 47, 187, 48 },
		{ 47, 188, 48 },
		{ 47, 189, 48 },
		{ 47, 190, 48 },
		{ 47, 191, 48 },
		{ 47, 192, 48 },
		{ 47, 193, 48 },
		{ 47, 194, 48 },
		{ 47, 195, 48 },
		{ 47, 196, 48 },
		{ 47, 197, 48 },
		{ 47, 198, 48 },
		{ 47, 199, 48 },
		{ 47, 200, 48 },
		{ 47, 201, 48 },
		{ 47, 202, 48 },
		{ 47, 203, 48 },
		{ 47, 204, 48 },
		{ 47, 205, 48 },
		{ 47, 206, 48 },
		{ 47, 207, 48 },
		{ 47, 208, 48 },
		{ 47, 209, 48 },
		{ 47, 210, 48 },
		{ 47, 211, 48 },
		{ 47, 212, 48 },
		{ 47, 213, 48 },
		{ 47, 214, 48 },
		{ 47, 215, 48 },
		{ 47, 216, 48 },
		{ 47, 217, 48 },
		{ 47, 218, 48 },
		{ 47, 219, 48 },
		{ 47, 220, 48 },
		{ 47, 221, 48 },
		{ 47, 222, 48 },
		{ 47, 223, 48 },
		{ 47, 224, 48 },
		{ 47, 225, 48 },
		{ 47, 226, 48 },
		{ 47, 227, 48 },
		{ 47, 228, 48 },
		{ 47, 229, 48 },
		{ 47, 230, 48 },
		{ 48, 179, 48 },
		{ 48, 180, 48 },
		{ 48, 181, 48 },
		{ 48, 182, 48 },
		{ 48, 183, 48 },
		{ 48, 184, 48 },
		{ 48, 185, 48 },
		{ 48, 186, 48 },
		{ 48, 187, 48 },
		{ 48, 188, 48 },
		{ 48, 189, 48 },
		{ 48, 190, 48 },
		{ 48, 191, 48 },
		{ 48, 192, 48 },
		{ 48, 193, 48 },
		{ 48, 194, 48 },
		{ 48, 195, 48 },
		{ 48, 196, 48 },
		{ 48, 197, 48 },
		{ 48, 198, 48 },
		{ 48, 199, 48 },
		{ 48, 200, 48 },
		{ 48, 201, 48 },
		{ 48, 202, 48 },
		{ 48, 203, 48 },
		{ 48, 204, 48 },
		{ 48, 205, 48 },
		{ 48, 206, 48 },
		{ 48, 207, 48 },
		{ 48, 208, 48 },
		{ 48, 209, 48 },
		{ 48, 210, 48 },
		{ 48, 211, 48 },
		{ 48, 212, 48 },
		{ 48, 213, 48 },
		{ 48, 214, 48 },
		{ 48, 215, 48 },
		{ 48, 216, 48 },
		{ 48, 217, 48 },
		{ 48, 218, 48 },
		{ 48, 219, 48 },
		{ 48, 220, 48 },
		{ 48, 221, 48 },
		{ 48, 222, 48 },
		{ 48, 223, 48 },
		{ 48, 224, 48 },
		{ 48, 225, 48 },
		{ 48, 226, 48 },
		{ 48, 227, 48 },
		{ 48, 228, 48 },
		{ 48, 229, 48 },
		{ 48, 230, 48 },
		{ 49, 180, 48 },
		{ 49, 181, 48 },
		{ 49, 182, 48 },
		{ 49, 183, 48 },
		{ 49, 184, 48 },
		{ 49, 185, 48 },
		{ 49, 186, 48 },
		{ 49, 187, 48 },
		{ 49, 188, 48 },
		{ 49, 189, 48 },
		{ 49, 190, 48 },
		{ 49, 191, 48 },
		{ 49, 192, 48 },
		{ 49, 193, 48 },
		{ 49, 194, 48 },
		{ 49, 195, 48 },
		{ 49, 196, 48 },
		{ 49, 197, 48 },
		{ 49, 198, 48 },
		{ 49, 199, 48 },
		{ 49, 200, 48 },
		{ 49, 201, 48 },
		{ 49, 202, 48 },
		{ 49, 203, 48 },
		{ 49, 204, 48 },
		{ 49, 205, 48 },
		{ 49, 206, 48 },
		{ 49, 207, 48 },
		{ 49, 208, 48 },
		{ 49, 209, 48 },
		{ 49, 210, 48 },
		{ 49, 211, 48 },
		{ 49, 212, 48 },
		{ 49, 213, 48 },
		{ 49, 214, 48 },
		{ 49, 215, 48 },
		{ 49, 216, 48 },
		{ 49, 217, 48 },
		{ 49, 218, 48 },
		{ 49, 219, 48 },
		{ 49, 220, 48 },
		{ 49, 221, 48 },
		{ 49, 222, 48 },
		{ 49, 223, 48 },
		{ 49, 224, 48 },
		{ 49, 225, 48 },
		{ 49, 226, 48 },
		{ 49, 227, 48 },
		{ 49, 228, 48 },
		{ 49, 229, 48 },
		{ 49, 230, 48 },
		{ 50, 180, 48 },
		{ 50, 181, 48 },
		{ 50, 182, 48 },
		{ 50, 183, 48 },
		{ 50, 184, 48 },
		{ 50, 185, 48 },
		{ 50, 186, 48 },
		{ 50, 187, 48 },
		{ 50, 188, 48 },
		{ 50, 189, 48 },
		{ 50, 190, 48 },
		{ 50, 191, 48 },
		{ 50, 192, 48 },
		{ 50, 193, 48 },
		{ 50, 194, 48 },
		{ 50, 195, 48 },
		{ 50, 196, 48 },
		{ 50, 197, 48 },
		{ 50, 198, 48 },
		{ 50, 199, 48 },
		{ 50, 200, 48 },
		{ 50, 201, 48 },
		{ 50, 202, 48 },
		{ 50, 203, 48 },
		{ 50, 204, 48 },
		{ 50, 205, 48 },
		{ 50, 206, 48 },
		{ 50, 207, 48 },
		{ 50, 208, 48 },
		{ 50, 209, 48 },
		{ 50, 210, 48 },
		{ 50, 211, 48 },
		{ 50, 212, 48 },
		{ 50, 213, 48 },
		{ 50, 214, 48 },
		{ 50, 215, 48 },
		{ 50, 216, 48 },
		{ 50, 217, 48 },
		{ 50, 218, 48 },
		{ 50, 219, 48 },
		{ 50, 220, 48 },
		{ 50, 221, 48 },
		{ 50, 222, 48 },
		{ 50, 223, 48 },
		{ 50, 224, 48 },
		{ 50, 225, 48 },
		{ 50, 226, 48 },
		{ 50, 227, 48 },
		{ 50, 228, 48 },
		{ 50, 229, 48 },
		{ 50, 230, 48 },
		{ 50, 231, 48 },
		{ 51, 180, 48 },
		{ 51, 181, 48 },
		{ 51, 182, 48 },
		{ 51, 183, 48 },
		{ 51, 184, 48 },
		{ 51, 185, 48 },
		{ 51, 186, 48 },
		{ 51, 187, 48 },
		{ 51, 188, 48 },
		{ 51, 189, 48 },
		{ 51, 190, 48 },
		{ 51, 191, 48 },
		{ 51, 192, 48 },
		{ 51, 193, 48 },
		{ 51, 194, 48 },
		{ 51, 195, 48 },
		{ 51, 196, 48 },
		{ 51, 197, 48 },
		{ 51, 198, 48 },
		{ 51, 199, 48 },
		{ 51, 200, 48 },
		{ 51, 201, 48 },
		{ 51, 202, 48 },
		{ 51, 203, 48 },
		{ 51, 204, 48 },
		{ 51, 205, 48 },
		{ 51, 206, 48 },
		{ 51, 207, 48 },
		{ 51, 208, 48 },
		{ 51, 209, 48 },
		{ 51, 210, 48 },
		{ 51, 211, 48 },
		{ 51, 212, 48 },
		{ 51, 213, 48 },
		{ 51, 214, 48 },
		{ 51, 215, 48 },
		{ 51, 216, 48 },
		{ 51, 217, 48 },
		{ 51, 218, 48 },
		{ 51, 219, 48 },
		{ 51, 220, 48 },
		{ 51, 221, 48 },
		{ 51, 222, 48 },
		{ 51, 223, 48 },
		{ 51, 224, 48 },
		{ 51, 225, 48 },
		{ 51, 226, 48 },
		{ 51, 227, 48 },
		{ 51, 228, 48 },
		{ 51, 229, 48 },
		{ 51, 230, 48 },
		{ 51, 231, 48 },
		{ 52, 181, 48 },
		{ 52, 182, 48 },
		{ 52, 183, 48 },
		{ 52, 184, 48 },
		{ 52, 185, 48 },
		{ 52, 186, 48 },
		{ 52, 187, 48 },
		{ 52, 188, 48 },
		{ 52, 189, 48 },
		{ 52, 190, 48 },
		{ 52, 191, 48 },
		{ 52, 192, 48 },
		{ 52, 193, 48 },
		{ 52, 194, 48 },
		{ 52, 195, 48 },
		{ 52, 196, 48 },
		{ 52, 197, 48 },
		{ 52, 198, 48 },
		{ 52, 199, 48 },
		{ 52, 200, 48 },
		{ 52, 201, 48 },
		{ 52, 202, 48 },
		{ 52, 203, 48 },
		{ 52, 204, 48 },
		{ 52, 205, 48 },
		{ 52, 206, 48 },
		{ 52, 207, 48 },
		{ 52, 208, 48 },
		{ 52, 209, 48 },
		{ 52, 210, 48 },
		{ 52, 211, 48 },
		{ 52, 212, 48 },
		{ 52, 213, 48 },
		{ 52, 214, 48 },
		{ 52, 215, 48 },
		{ 52, 216, 48 },
		{ 52, 217, 48 },
		{ 52, 218, 48 },
		{ 52, 219, 48 },
		{ 52, 220, 48 },
		{ 52, 221, 48 },
		{ 52, 222, 48 },
		{ 52, 223, 48 },
		{ 52, 224, 48 },
		{ 52, 225, 48 },
		{ 52, 226, 48 },
		{ 52, 227, 48 },
		{ 52, 228, 48 },
		{ 52, 229, 48 },
		{ 52, 230, 48 },
		{ 52, 231, 48 },
		{ 53, 181, 48 },
		{ 53, 182, 48 },
		{ 53, 183, 48 },
		{ 53, 184, 48 },
		{ 53, 185, 48 },
		{ 53, 186, 48 },
		{ 53, 187, 48 },
		{ 53, 188, 48 },
		{ 53, 189, 48 },
		{ 53, 190, 48 },
		{ 53, 191, 48 },
		{ 53, 192, 48 },
		{ 53, 193, 48 },
		{ 53, 194, 48 },
		{ 53, 195, 48 },
		{ 53, 196, 48 },
		{ 53, 197, 48 },
		{ 53, 198, 48 },
		{ 53, 199, 48 },
		{ 53, 200, 48 },
		{ 53, 201, 48 },
		{ 53, 202, 48 },
		{ 53, 203, 48 },
		{ 53, 204, 48 },
		{ 53, 205, 48 },
		{ 53, 206, 48 },
		{ 53, 207, 48 },
		{ 53, 208, 48 },
		{ 53, 209, 48 },
		{ 53, 210, 48 },
		{ 53, 211, 48 },
		{ 53, 212, 48 },
		{ 53, 213, 48 },
		{ 53, 214, 48 },
		{ 53, 215, 48 },
		{ 53, 216, 48 },
		{ 53, 217, 48 },
		{ 53, 218, 48 },
		{ 53, 219, 48 },
		{ 53, 220, 48 },
		{ 53, 221, 48 },
		{ 53, 222, 48 },
		{ 53, 223, 48 },
		{ 53, 224, 48 },
		{ 53, 225, 48 },
		{ 53, 226, 48 },
		{ 53, 227, 48 },
		{ 53, 228, 48 },
		{ 53, 229, 48 },
		{ 53, 230, 48 },
		{ 53, 231, 48 },
		{ 54, 181, 48 },
		{ 54, 182, 48 },
		{ 54, 183, 48 },
		{ 54, 184, 48 },
		{ 54, 185, 48 },
		{ 54, 186, 48 },
		{ 54, 187, 48 },
		{ 54, 188, 48 },
		{ 54, 189, 48 },
		{ 54, 190, 48 },
		{ 54, 191, 48 },
		{ 54, 192, 48 },
		{ 54, 193, 48 },
		{ 54, 194, 48 },
		{ 54, 195, 48 },
		{ 54, 196, 48 },
		{ 54, 197, 48 },
		{ 54, 198, 48 },
		{ 54, 199, 48 },
		{ 54, 200, 48 },
		{ 54, 201, 48 },
		{ 54, 202, 48 },
		{ 54, 203, 48 },
		{ 54, 204, 48 },
		{ 54, 205, 48 },
		{ 54, 206, 48 },
		{ 54, 207, 48 },
		{ 54, 208, 48 },
		{ 54, 209, 48 },
		{ 54, 210, 48 },
		{ 54, 211, 48 },
		{ 54, 212, 48 },
		{ 54, 213, 48 },
		{ 54, 214, 48 },
		{ 54, 215, 48 },
		{ 54, 216, 48 },
		{ 54, 217, 48 },
		{ 54, 218, 48 },
		{ 54, 219, 48 },
		{ 54, 220, 48 },
		{ 54, 221, 48 },
		{ 54, 222, 48 },
		{ 54, 223, 48 },
		{ 54, 224, 48 },
		{ 54, 225, 48 },
		{ 54, 226, 48 },
		{ 54, 227, 48 },
		{ 54, 228, 48 },
		{ 54, 229, 48 },
		{ 54, 230, 48 },
		{ 54, 231, 48 },
		{ 55, 182, 48 },
		{ 55, 183, 48 },
		{ 55, 184, 48 },
		{ 55, 185, 48 },
		{ 55, 186, 48 },
		{ 55, 187, 48 },
		{ 55, 188, 48 },
		{ 55, 189, 48 },
		{ 55, 190, 48 },
		{ 55, 191, 48 },
		{ 55, 192, 48 },
		{ 55, 193, 48 },
		{ 55, 194, 48 },
		{ 55, 195, 48 },
		{ 55, 196, 48 },
		{ 55, 197, 48 },
		{ 55, 198, 48 },
		{ 55, 199, 48 },
		{ 55, 200, 48 },
		{ 55, 201, 48 },
		{ 55, 202, 48 },
		{ 55, 203, 48 },
		{ 55, 204, 48 },
		{ 55, 205, 48 },
		{ 55, 206, 48 },
		{ 55, 207, 48 },
		{ 55, 208, 48 },
		{ 55, 209, 48 },
		{ 55, 210, 48 },
		{ 55, 211, 48 },
		{ 55, 212, 48 },
		{ 55, 213, 48 },
		{ 55, 214, 48 },
		{ 55, 215, 48 },
		{ 55, 216, 48 },
		{ 55, 217, 48 },
		{ 55, 218, 48 },
		{ 55, 219, 48 },
		{ 55, 220, 48 },
		{ 55, 221, 48 },
		{ 55, 222, 48 },
		{ 55, 223, 48 },
		{ 55, 224, 48 },
		{ 55, 225, 48 },
		{ 55, 226, 48 },
		{ 55, 227, 48 },
		{ 55, 228, 48 },
		{ 55, 229, 48 },
		{ 55, 230, 48 },
		{ 55, 231, 48 },
		{ 56, 182, 48 },
		{ 56, 183, 48 },
		{ 56, 184, 48 },
		{ 56, 185, 48 },
		{ 56, 186, 48 },
		{ 56, 187, 48 },
		{ 56, 188, 48 },
		{ 56, 189, 48 },
		{ 56, 190, 48 },
		{ 56, 191, 48 },
		{ 56, 192, 48 },
		{ 56, 193, 48 },
		{ 56, 194, 48 },
		{ 56, 195, 48 },
		{ 56, 196, 48 },
		{ 56, 197, 48 },
		{ 56, 198, 48 },
		{ 56, 199, 48 },
		{ 56, 200, 48 },
		{ 56, 201, 48 },
		{ 56, 202, 48 },
		{ 56, 203, 48 },
		{ 56, 204, 48 },
		{ 56, 205, 48 },
		{ 56, 206, 48 },
		{ 56, 207, 48 },
		{ 56, 208, 48 },
		{ 56, 209, 48 },
		{ 56, 210, 48 },
		{ 56, 211, 48 },
		{ 56, 212, 48 },
		{ 56, 213, 48 },
		{ 56, 214, 48 },
		{ 56, 215, 48 },
		{ 56, 216, 48 },
		{ 56, 217, 48 },
		{ 56, 218, 48 },
		{ 56, 219, 48 },
		{ 56, 220, 48 },
		{ 56, 221, 48 },
		{ 56, 222, 48 },
		{ 56, 223, 48 },
		{ 56, 224, 48 },
		{ 56, 225, 48 },
		{ 56, 226, 48 },
		{ 56, 227, 48 },
		{ 56, 228, 48 },
		{ 56, 229, 48 },
		{ 56, 230, 48 },
		{ 56, 231, 48 },
		{ 57, 183, 48 },
		{ 57, 184, 48 },
		{ 57, 185, 48 },
		{ 57, 186, 48 },
		{ 57, 187, 48 },
		{ 57, 188, 48 },
		{ 57, 189, 48 },
		{ 57, 190, 48 },
		{ 57, 191, 48 },
		{ 57, 192, 48 },
		{ 57, 193, 48 },
		{ 57, 194, 48 },
		{ 57, 195, 48 },
		{ 57, 196, 48 },
		{ 57, 197, 48 },
		{ 57, 198, 48 },
		{ 57, 199, 48 },
		{ 57, 200, 48 },
		{ 57, 201, 48 },
		{ 57, 202, 48 },
		{ 57, 203, 48 },
		{ 57, 204, 48 },
		{ 57, 205, 48 },
		{ 57, 206, 48 },
		{ 57, 207, 48 },
		{ 57, 208, 48 },
		{ 57, 209, 48 },
		{ 57, 210, 48 },
		{ 57, 211, 48 },
		{ 57, 212, 48 },
		{ 57, 213, 48 },
		{ 57, 214, 48 },
		{ 57, 215, 48 },
		{ 57, 216, 48 },
		{ 57, 217, 48 },
		{ 57, 218, 48 },
		{ 57, 219, 48 },
		{ 57, 220, 48 },
		{ 57, 221, 48 },
		{ 57, 222, 48 },
		{ 57, 223, 48 },
		{ 57, 224, 48 },
		{ 57, 225, 48 },
		{ 57, 226, 48 },
		{ 57, 227, 48 },
		{ 57, 228, 48 },
		{ 57, 229, 48 },
		{ 57, 230, 48 },
		{ 57, 231, 48 },
		{ 58, 183, 48 },
		{ 58, 184, 48 },
		{ 58, 185, 48 },
		{ 58, 186, 48 },
		{ 58, 187, 48 },
		{ 58, 188, 48 },
		{ 58, 189, 48 },
		{ 58, 190, 48 },
		{ 58, 191, 48 },
		{ 58, 192, 48 },
		{ 58, 193, 48 },
		{ 58, 194, 48 },
		{ 58, 195, 48 },
		{ 58, 196, 48 },
		{ 58, 197, 48 },
		{ 58, 198, 48 },
		{ 58, 199, 48 },
		{ 58, 200, 48 },
		{ 58, 201, 48 },
		{ 58, 202, 48 },
		{ 58, 203, 48 },
		{ 58, 204, 48 },
		{ 58, 205, 48 },
		{ 58, 206, 48 },
		{ 58, 207, 48 },
		{ 58, 208, 48 },
		{ 58, 209, 48 },
		{ 58, 210, 48 },
		{ 58, 211, 48 },
		{ 58, 212, 48 },
		{ 58, 213, 48 },
		{ 58, 214, 48 },
		{ 58, 215, 48 },
		{ 58, 216, 48 },
		{ 58, 217, 48 },
		{ 58, 218, 48 },
		{ 58, 219, 48 },
		{ 58, 220, 48 },
		{ 58, 221, 48 },
		{ 58, 222, 48 },
		{ 58, 223, 48 },
		{ 58, 224, 48 },
		{ 58, 225, 48 },
		{ 58, 226, 48 },
		{ 58, 227, 48 },
		{ 58, 228, 48 },
		{ 58, 229, 48 },
		{ 58, 230, 48 },
		{ 58, 231, 48 },
		{ 58, 232, 48 },
		{ 59, 183, 48 },
		{ 59, 184, 48 },
		{ 59, 185, 48 },
		{ 59, 186, 48 },
		{ 59, 187, 48 },
		{ 59, 188, 48 },
		{ 59, 189, 48 },
		{ 59, 190, 48 },
		{ 59, 191, 48 },
		{ 59, 192, 48 },
		{ 59, 193, 48 },
		{ 59, 194, 48 },
		{ 59, 195, 48 },
		{ 59, 196, 48 },
		{ 59, 197, 48 },
		{ 59, 198, 48 },
		{ 59, 199, 48 },
		{ 59, 200, 48 },
		{ 59, 201, 48 },
		{ 59, 202, 48 },
		{ 59, 203, 48 },
		{ 59, 204, 48 },
		{ 59, 205, 48 },
		{ 59, 206, 48 },
		{ 59, 207, 48 },
		{ 59, 208, 48 },
		{ 59, 209, 48 },
		{ 59, 210, 48 },
		{ 59, 211, 48 },
		{ 59, 212, 48 },
		{ 59, 213, 48 },
		{ 59, 214, 48 },
		{ 59, 215, 48 },
		{ 59, 216, 48 },
		{ 59, 217, 48 },
		{ 59, 218, 48 },
		{ 59, 219, 48 },
		{ 59, 220, 48 },
		{ 59, 221, 48 },
		{ 59, 222, 48 },
		{ 59, 223, 48 },
		{ 59, 224, 48 },
		{ 59, 225, 48 },
		{ 59, 226, 48 },
		{ 59, 227, 48 },
		{ 59, 228, 48 },
		{ 59, 229, 48 },
		{ 59, 230, 48 },
		{ 59, 231, 48 },
		{ 59, 232, 48 },
		{ 60, 184, 48 },
		{ 60, 185, 48 },
		{ 60, 186, 48 },
		{ 60, 187, 48 },
		{ 60, 188, 48 },
		{ 60, 189, 48 },
		{ 60, 190, 48 },
		{ 60, 191, 48 },
		{ 60, 192, 48 },
		{ 60, 193, 48 },
		{ 60, 194, 48 },
		{ 60, 195, 48 },
		{ 60, 196, 48 },
		{ 60, 197, 48 },
		{ 60, 198, 48 },
		{ 60, 199, 48 },
		{ 60, 200, 48 },
		{ 60, 201, 48 },
		{ 60, 202, 48 },
		{ 60, 203, 48 },
		{ 60, 204, 48 },
		{ 60, 205, 48 },
		{ 60, 206, 48 },
		{ 60, 207, 48 },
		{ 60, 208, 48 },
		{ 60, 209, 48 },
		{ 60, 210, 48 },
		{ 60, 211, 48 },
		{ 60, 212, 48 },
		{ 60, 213, 48 },
		{ 60, 214, 48 },
		{ 60, 215, 48 },
		{ 60, 216, 48 },
		{ 60, 217, 48 },
		{ 60, 218, 48 },
		{ 60, 219, 48 },
		{ 60, 220, 48 },
		{ 60, 221, 48 },
		{ 60, 222, 48 },
		{ 60, 223, 48 },
		{ 60, 224, 48 },
		{ 60, 225, 48 },
		{ 60, 226, 48 },
		{ 60, 227, 48 },
		{ 60, 228, 48 },
		{ 60, 229, 48 },
		{ 60, 230, 48 },
		{ 60, 231, 48 },
		{ 60, 232, 48 },
		{ 61, 184, 48 },
		{ 61, 185, 48 },
		{ 61, 186, 48 },
		{ 61, 187, 48 },
		{ 61, 188, 48 },
		{ 61, 189, 48 },
		{ 61, 190, 48 },
		{ 61, 191, 48 },
		{ 61, 192, 48 },
		{ 61, 193, 48 },
		{ 61, 194, 48 },
		{ 61, 195, 48 },
		{ 61, 196, 48 },
		{ 61, 197, 48 },
		{ 61, 198, 48 },
		{ 61, 199, 48 },
		{ 61, 200, 48 },
		{ 61, 201, 48 },
		{ 61, 202, 48 },
		{ 61, 203, 48 },
		{ 61, 204, 48 },
		{ 61, 205, 48 },
		{ 61, 206, 48 },
		{ 61, 207, 48 },
		{ 61, 208, 48 },
		{ 61, 209, 48 },
		{ 61, 210, 48 },
		{ 61, 211, 48 },
		{ 61, 212, 48 },
		{ 61, 213, 48 },
		{ 61, 214, 48 },
		{ 61, 215, 48 },
		{ 61, 216, 48 },
		{ 61, 217, 48 },
		{ 61, 218, 48 },
		{ 61, 219, 48 },
		{ 61, 220, 48 },
		{ 61, 221, 48 },
		{ 61, 222, 48 },
		{ 61, 223, 48 },
		{ 61, 224, 48 },
		{ 61, 225, 48 },
		{ 61, 226, 48 },
		{ 61, 227, 48 },
		{ 61, 228, 48 },
		{ 61, 229, 48 },
		{ 61, 230, 48 },
		{ 61, 231, 48 },
		{ 61, 232, 48 },
		{ 62, 184, 48 },
		{ 62, 185, 48 },
		{ 62, 186, 48 },
		{ 62, 187, 48 },
		{ 62, 188, 48 },
		{ 62, 189, 48 },
		{ 62, 190, 48 },
		{ 62, 191, 48 },
		{ 62, 192, 48 },
		{ 62, 193, 48 },
		{ 62, 194, 48 },
		{ 62, 195, 48 },
		{ 62, 196, 48 },
		{ 62, 197, 48 },
		{ 62, 198, 48 },
		{ 62, 199, 48 },
		{ 62, 200, 48 },
		{ 62, 201, 48 },
		{ 62, 202, 48 },
		{ 62, 203, 48 },
		{ 62, 204, 48 },
		{ 62, 205, 48 },
		{ 62, 206, 48 },
		{ 62, 207, 48 },
		{ 62, 208, 48 },
		{ 62, 209, 48 },
		{ 62, 210, 48 },
		{ 62, 211, 48 },
		{ 62, 212, 48 },
		{ 62, 213, 48 },
		{ 62, 214, 48 },
		{ 62, 215, 48 },
		{ 62, 216, 48 },
		{ 62, 217, 48 },
		{ 62, 218, 48 },
		{ 62, 219, 48 },
		{ 62, 220, 48 },
		{ 62, 221, 48 },
		{ 62, 222, 48 },
		{ 62, 223, 48 },
		{ 62, 224, 48 },
		{ 62, 225, 48 },
		{ 62, 226, 48 },
		{ 62, 227, 48 },
		{ 62, 228, 48 },
		{ 62, 229, 48 },
		{ 62, 230, 48 },
		{ 62, 231, 48 },
		{ 62, 232, 48 },
		{ 63, 185, 48 },
		{ 63, 186, 48 },
		{ 63, 187, 48 },
		{ 63, 188, 48 },
		{ 63, 189, 48 },
		{ 63, 190, 48 },
		{ 63, 191, 48 },
		{ 63, 192, 48 },
		{ 63, 193, 48 },
		{ 63, 194, 48 },
		{ 63, 195, 48 },
		{ 63, 196, 48 },
		{ 63, 197, 48 },
		{ 63, 198, 48 },
		{ 63, 199, 48 },
		{ 63, 200, 48 },
		{ 63, 201, 48 },
		{ 63, 202, 48 },
		{ 63, 203, 48 },
		{ 63, 204, 48 },
		{ 63, 205, 48 },
		{ 63, 206, 48 },
		{ 63, 207, 48 },
		{ 63, 208, 48 },
		{ 63, 209, 48 },
		{ 63, 210, 48 },
		{ 63, 211, 48 },
		{ 63, 212, 48 },
		{ 63, 213, 48 },
		{ 63, 214, 48 },
		{ 63, 215, 48 },
		{ 63, 216, 48 },
		{ 63, 217, 48 },
		{ 63, 218, 48 },
		{ 63, 219, 48 },
		{ 63, 220, 48 },
		{ 63, 221, 48 },
		{ 63, 222, 48 },
		{ 63, 223, 48 },
		{ 63, 224, 48 },
		{ 63, 225, 48 },
		{ 63, 226, 48 },
		{ 63, 227, 48 },
		{ 63, 228, 48 },
		{ 63, 229, 48 },
		{ 63, 230, 48 },
		{ 63, 231, 48 },
		{ 63, 232, 48 },
		{ 64, 185, 48 },
		{ 64, 186, 48 },
		{ 64, 187, 48 },
		{ 64, 188, 48 },
		{ 64, 189, 48 },
		{ 64, 190, 48 },
		{ 64, 191, 48 },
		{ 64, 192, 48 },
		{ 64, 193, 48 },
		{ 64, 194, 48 },
		{ 64, 195, 48 },
		{ 64, 196, 48 },
		{ 64, 197, 48 },
		{ 64, 198, 48 },
		{ 64, 199, 48 },
		{ 64, 200, 48 },
		{ 64, 201, 48 },
		{ 64, 202, 48 },
		{ 64, 203, 48 },
		{ 64, 204, 48 },
		{ 64, 205, 48 },
		{ 64, 206, 48 },
		{ 64, 207, 48 },
		{ 64, 208, 48 },
		{ 64, 209, 48 },
		{ 64, 210, 48 },
		{ 64, 211, 48 },
		{ 64, 212, 48 },
		{ 64, 213, 48 },
		{ 64, 214, 48 },
		{ 64, 215, 48 },
		{ 64, 216, 48 },
		{ 64, 217, 48 },
		{ 64, 218, 48 },
		{ 64, 219, 48 },
		{ 64, 220, 48 },
		{ 64, 221, 48 },
		{ 64, 222, 48 },
		{ 64, 223, 48 },
		{ 64, 224, 48 },
		{ 64, 225, 48 },
		{ 64, 226, 48 },
		{ 64, 227, 48 },
		{ 64, 228, 48 },
		{ 64, 229, 48 },
		{ 64, 230, 48 },
		{ 64, 231, 48 },
		{ 64, 232, 48 },
		{ 65, 186, 48 },
		{ 65, 187, 48 },
		{ 65, 188, 48 },
		{ 65, 189, 48 },
		{ 65, 190, 48 },
		{ 65, 191, 48 },
		{ 65, 192, 48 },
		{ 65, 193, 48 },
		{ 65, 194, 48 },
		{ 65, 195, 48 },
		{ 65, 196, 48 },
		{ 65, 197, 48 },
		{ 65, 198, 48 },
		{ 65, 199, 48 },
		{ 65, 200, 48 },
		{ 65, 201, 48 },
		{ 65, 202, 48 },
		{ 65, 203, 48 },
		{ 65, 204, 48 },
		{ 65, 205, 48 },
		{ 65, 206, 48 },
		{ 65, 207, 48 },
		{ 65, 208, 48 },
		{ 65, 209, 48 },
		{ 65, 210, 48 },
		{ 65, 211, 48 },
		{ 65, 212, 48 },
		{ 65, 213, 48 },
		{ 65, 214, 48 },
		{ 65, 215, 48 },
		{ 65, 216, 48 },
		{ 65, 217, 48 },
		{ 65, 218, 48 },
		{ 65, 219, 48 },
		{ 65, 220, 48 },
		{ 65, 221, 48 },
		{ 65, 222, 48 },
		{ 65, 223, 48 },
		{ 65, 224, 48 },
		{ 65, 225, 48 },
		{ 65, 226, 48 },
		{ 65, 227, 48 },
		{ 65, 228, 48 },
		{ 65, 229, 48 },
		{ 65, 230, 48 },
		{ 65, 231, 48 },
		{ 65, 232, 48 },
		{ 66, 186, 48 },
		{ 66, 187, 48 },
		{ 66, 188, 48 },
		{ 66, 189, 48 },
		{ 66, 190, 48 },
		{ 66, 191, 48 },
		{ 66, 192, 48 },
		{ 66, 193, 48 },
		{ 66, 194, 48 },
		{ 66, 195, 48 },
		{ 66, 196, 48 },
		{ 66, 197, 48 },
		{ 66, 198, 48 },
		{ 66, 199, 48 },
		{ 66, 200, 48 },
		{ 66, 201, 48 },
		{ 66, 202, 48 },
		{ 66, 203, 48 },
		{ 66, 204, 48 },
		{ 66, 205, 48 },
		{ 66, 206, 48 },
		{ 66, 207, 48 },
		{ 66, 208, 48 },
		{ 66, 209, 48 },
		{ 66, 210, 48 },
		{ 66, 211, 48 },
		{ 66, 212, 48 },
		{ 66, 213, 48 },
		{ 66, 214, 48 },
		{ 66, 215, 48 },
		{ 66, 216, 48 },
		{ 66, 217, 48 },
		{ 66, 218, 48 },
		{ 66, 219, 48 },
		{ 66, 220, 48 },
		{ 66, 221, 48 },
		{ 66, 222, 48 },
		{ 66, 223, 48 },
		{ 66, 224, 48 },
		{ 66, 225, 48 },
		{ 66, 226, 48 },
		{ 66, 227, 48 },
		{ 66, 228, 48 },
		{ 66, 229, 48 },
		{ 66, 230, 48 },
		{ 66, 231, 48 },
		{ 66, 232, 48 },
		{ 66, 233, 48 },
		{ 67, 186, 48 },
		{ 67, 187, 48 },
		{ 67, 188, 48 },
		{ 67, 189, 48 },
		{ 67, 190, 48 },
		{ 67, 191, 48 },
		{ 67, 192, 48 },
		{ 67, 193, 48 },
		{ 67, 194, 48 },
		{ 67, 195, 48 },
		{ 67, 196, 48 },
		{ 67, 197, 48 },
		{ 67, 198, 48 },
		{ 67, 199, 48 },
		{ 67, 200, 48 },
		{ 67, 201, 48 },
		{ 67, 202, 48 },
		{ 67, 203, 48 },
		{ 67, 204, 48 },
		{ 67, 205, 48 },
		{ 67, 206, 48 },
		{ 67, 207, 48 },
		{ 67, 208, 48 },
		{ 67, 209, 48 },
		{ 67, 210, 48 },
		{ 67, 211, 48 },
		{ 67, 212, 48 },
		{ 67, 213, 48 },
		{ 67, 214, 48 },
		{ 67, 215, 48 },
		{ 67, 216, 48 },
		{ 67, 217, 48 },
		{ 67, 218, 48 },
		{ 67, 219, 48 },
		{ 67, 220, 48 },
		{ 67, 221, 48 },
		{ 67, 222, 48 },
		{ 67, 223, 48 },
		{ 67, 224, 48 },
		{ 67, 225, 48 },
		{ 67, 226, 48 },
		{ 67, 227, 48 },
		{ 67, 228, 48 },
		{ 67, 229, 48 },
		{ 67, 230, 48 },
		{ 67, 231, 48 },
		{ 67, 232, 48 },
		{ 67, 233, 48 },
		{ 68, 187, 48 },
		{ 68, 188, 48 },
		{ 68, 189, 48 },
		{ 68, 190, 48 },
		{ 68, 191, 48 },
		{ 68, 192, 48 },
		{ 68, 193, 48 },
		{ 68, 194, 48 },
		{ 68, 195, 48 },
		{ 68, 196, 48 },
		{ 68, 197, 48 },
		{ 68, 198, 48 },
		{ 68, 199, 48 },
		{ 68, 200, 48 },
		{ 68, 201, 48 },
		{ 68, 202, 48 },
		{ 68, 203, 48 },
		{ 68, 204, 48 },
		{ 68, 205, 48 },
		{ 68, 206, 48 },
		{ 68, 207, 48 },
		{ 68, 208, 48 },
		{ 68, 209, 48 },
		{ 68, 210, 48 },
		{ 68, 211, 48 },
		{ 68, 212, 48 },
		{ 68, 213, 48 },
		{ 68, 214, 48 },
		{ 68, 215, 48 },
		{ 68, 216, 48 },
		{ 68, 217, 48 },
		{ 68, 218, 48 },
		{ 68, 219, 48 },
		{ 68, 220, 48 },
		{ 68, 221, 48 },
		{ 68, 222, 48 },
		{ 68, 223, 48 },
		{ 68, 224, 48 },
		{ 68, 225, 48 },
		{ 68, 226, 48 },
		{ 68, 227, 48 },
		{ 68, 228, 48 },
		{ 68, 229, 48 },
		{ 68, 230, 48 },
		{ 68, 231, 48 },
		{ 68, 232, 48 },
		{ 68, 233, 48 },
		{ 69, 187, 48 },
		{ 69, 188, 48 },
		{ 69, 189, 48 },
		{ 69, 190, 48 },
		{ 69, 191, 48 },
		{ 69, 192, 48 },
		{ 69, 193, 48 },
		{ 69, 194, 48 },
		{ 69, 195, 48 },
		{ 69, 196, 48 },
		{ 69, 197, 48 },
		{ 69, 198, 48 },
		{ 69, 199, 48 },
		{ 69, 200, 48 },
		{ 69, 201, 48 },
		{ 69, 202, 48 },
		{ 69, 203, 48 },
		{ 69, 204, 48 },
		{ 69, 205, 48 },
		{ 69, 206, 48 },
		{ 69, 207, 48 },
		{ 69, 208, 48 },
		{ 69, 209, 48 },
		{ 69, 210, 48 },
		{ 69, 211, 48 },
		{ 69, 212, 48 },
		{ 69, 213, 48 },
		{ 69, 214, 48 },
		{ 69, 215, 48 },
		{ 69, 216, 48 },
		{ 69, 217, 48 },
		{ 69, 218, 48 },
		{ 69, 219, 48 },
		{ 69, 220, 48 },
		{ 69, 221, 48 },
		{ 69, 222, 48 },
		{ 69, 223, 48 },
		{ 69, 224, 48 },
		{ 69, 225, 48 },
		{ 69, 226, 48 },
		{ 69, 227, 48 },
		{ 69, 228, 48 },
		{ 69, 229, 48 },
		{ 69, 230, 48 },
		{ 69, 231, 48 },
		{ 69, 232, 48 },
		{ 69, 233, 48 },
		{ 70, 187, 48 },
		{ 70, 188, 48 },
		{ 70, 189, 48 },
		{ 70, 190, 48 },
		{ 70, 191, 48 },
		{ 70, 192, 48 },
		{ 70, 193, 48 },
		{ 70, 194, 48 },
		{ 70, 195, 48 },
		{ 70, 196, 48 },
		{ 70, 197, 48 },
		{ 70, 198, 48 },
		{ 70, 199, 48 },
		{ 70, 200, 48 },
		{ 70, 201, 48 },
		{ 70, 202, 48 },
		{ 70, 203, 48 },
		{ 70, 204, 48 },
		{ 70, 205, 48 },
		{ 70, 206, 48 },
		{ 70, 207, 48 },
		{ 70, 208, 48 },
		{ 70, 209, 48 },
		{ 70, 210, 48 },
		{ 70, 211, 48 },
		{ 70, 212, 48 },
		{ 70, 213, 48 },
		{ 70, 214, 48 },
		{ 70, 215, 48 },
		{ 70, 216, 48 },
		{ 70, 217, 48 },
		{ 70, 218, 48 },
		{ 70, 219, 48 },
		{ 70, 220, 48 },
		{ 70, 221, 48 },
		{ 70, 222, 48 },
		{ 70, 223, 48 },
		{ 70, 224, 48 },
		{ 70, 225, 48 },
		{ 70, 226, 48 },
		{ 70, 227, 48 },
		{ 70, 228, 48 },
		{ 70, 229, 48 },
		{ 70, 230, 48 },
		{ 70, 231, 48 },
		{ 70, 232, 48 },
		{ 70, 233, 48 },
		{ 71, 188, 48 },
		{ 71, 189, 48 },
		{ 71, 190, 48 },
		{ 71, 191, 48 },
		{ 71, 192, 48 },
		{ 71, 193, 48 },
		{ 71, 194, 48 },
		{ 71, 195, 48 },
		{ 71, 196, 48 },
		{ 71, 197, 48 },
		{ 71, 198, 48 },
		{ 71, 199, 48 },
		{ 71, 200, 48 },
		{ 71, 201, 48 },
		{ 71, 202, 48 },
		{ 71, 203, 48 },
		{ 71, 204, 48 },
		{ 71, 205, 48 },
		{ 71, 206, 48 },
		{ 71, 207, 48 },
		{ 71, 208, 48 },
		{ 71, 209, 48 },
		{ 71, 210, 48 },
		{ 71, 211, 48 },
		{ 71, 212, 48 },
		{ 71, 213, 48 },
		{ 71, 214, 48 },
		{ 71, 215, 48 },
		{ 71, 216, 48 },
		{ 71, 217, 48 },
		{ 71, 218, 48 },
		{ 71, 219, 48 },
		{ 71, 220, 48 },
		{ 71, 221, 48 },
		{ 71, 222, 48 },
		{ 71, 223, 48 },
		{ 71, 224, 48 },
		{ 71, 225, 48 },
		{ 71, 226, 48 },
		{ 71, 227, 48 },
		{ 71, 228, 48 },
		{ 71, 229, 48 },
		{ 71, 230, 48 },
		{ 71, 231, 48 },
		{ 71, 232, 48 },
		{ 71, 233, 48 },
		{ 72, 188, 48 },
		{ 72, 189, 48 },
		{ 72, 190, 48 },
		{ 72, 191, 48 },
		{ 72, 192, 48 },
		{ 72, 193, 48 },
		{ 72, 194, 48 },
		{ 72, 195, 48 },
		{ 72, 196, 48 },
		{ 72, 197, 48 },
		{ 72, 198, 48 },
		{ 72, 199, 48 },
		{ 72, 200, 48 },
		{ 72, 201, 48 },
		{ 72, 202, 48 },
		{ 72, 203, 48 },
		{ 72, 204, 48 },
		{ 72, 205, 48 },
		{ 72, 206, 48 },
		{ 72, 207, 48 },
		{ 72, 208, 48 },
		{ 72, 209, 48 },
		{ 72, 210, 48 },
		{ 72, 211, 48 },
		{ 72, 212, 48 },
		{ 72, 213, 48 },
		{ 72, 214, 48 },
		{ 72, 215, 48 },
		{ 72, 216, 48 },
		{ 72, 217, 48 },
		{ 72, 218, 48 },
		{ 72, 219, 48 },
		{ 72, 220, 48 },
		{ 72, 221, 48 },
		{ 72, 222, 48 },
		{ 72, 223, 48 },
		{ 72, 224, 48 },
		{ 72, 225, 48 },
		{ 72, 226, 48 },
		{ 72, 227, 48 },
		{ 72, 228, 48 },
		{ 72, 229, 48 },
		{ 72, 230, 48 },
		{ 72, 231, 48 },
		{ 72, 232, 48 },
		{ 72, 233, 48 },
		{ 73, 189, 48 },
		{ 73, 190, 48 },
		{ 73, 191, 48 },
		{ 73, 192, 48 },
		{ 73, 193, 48 },
		{ 73, 194, 48 },
		{ 73, 195, 48 },
		{ 73, 196, 48 },
		{ 73, 197, 48 },
		{ 73, 198, 48 },
		{ 73, 199, 48 },
		{ 73, 200, 48 },
		{ 73, 201, 48 },
		{ 73, 202, 48 },
		{ 73, 203, 48 },
		{ 73, 204, 48 },
		{ 73, 205, 48 },
		{ 73, 206, 48 },
		{ 73, 207, 48 },
		{ 73, 208, 48 },
		{ 73, 209, 48 },
		{ 73, 210, 48 },
		{ 73, 211, 48 },
		{ 73, 212, 48 },
		{ 73, 213, 48 },
		{ 73, 214, 48 },
		{ 73, 215, 48 },
		{ 73, 216, 48 },
		{ 73, 217, 48 },
		{ 73, 218, 48 },
		{ 73, 219, 48 },
		{ 73, 220, 48 },
		{ 73, 221, 48 },
		{ 73, 222, 48 },
		{ 73, 223, 48 },
		{ 73, 224, 48 },
		{ 73, 225, 48 },
		{ 73, 226, 48 },
		{ 73, 227, 48 },
		{ 73, 228, 48 },
		{ 73, 229, 48 },
		{ 73, 230, 48 },
		{ 73, 231, 48 },
		{ 73, 232, 48 },
		{ 73, 233, 48 },
		{ 74, 189, 48 },
		{ 74, 190, 48 },
		{ 74, 191, 48 },
		{ 74, 192, 48 },
		{ 74, 193, 48 },
		{ 74, 194, 48 },
		{ 74, 195, 48 },
		{ 74, 196, 48 },
		{ 74, 197, 48 },
		{ 74, 198, 48 },
		{ 74, 199, 48 },
		{ 74, 200, 48 },
		{ 74, 201, 48 },
		{ 74, 202, 48 },
		{ 74, 203, 48 },
		{ 74, 204, 48 },
		{ 74, 205, 48 },
		{ 74, 206, 48 },
		{ 74, 207, 48 },
		{ 74, 208, 48 },
		{ 74, 209, 48 },
		{ 74, 210, 48 },
		{ 74, 211, 48 },
		{ 74, 212, 48 },
		{ 74, 213, 48 },
		{ 74, 214, 48 },
		{ 74, 215, 48 },
		{ 74, 216, 48 },
		{ 74, 217, 48 },
		{ 74, 218, 48 },
		{ 74, 219, 48 },
		{ 74, 220, 48 },
		{ 74, 221, 48 },
		{ 74, 222, 48 },
		{ 74, 223, 48 },
		{ 74, 224, 48 },
		{ 74, 225, 48 },
		{ 74, 226, 48 },
		{ 74, 227, 48 },
		{ 74, 228, 48 },
		{ 74, 229, 48 },
		{ 74, 230, 48 },
		{ 74, 231, 48 },
		{ 74, 232, 48 },
		{ 74, 233, 48 },
		{ 74, 234, 48 },
		{ 75, 189, 48 },
		{ 75, 190, 48 },
		{ 75, 191, 48 },
		{ 75, 192, 48 },
		{ 75, 193, 48 },
		{ 75, 194, 48 },
		{ 75, 195, 48 },
		{ 75, 196, 48 },
		{ 75, 197, 48 },
		{ 75, 198, 48 },
		{ 75, 199, 48 },
		{ 75, 200, 48 },
		{ 75, 201, 48 },
		{ 75, 202, 48 },
		{ 75, 203, 48 },
		{ 75, 204, 48 },
		{ 75, 205, 48 },
		{ 75, 206, 48 },
		{ 75, 207, 48 },
		{ 75, 208, 48 },
		{ 75, 209, 48 },
		{ 75, 210, 48 },
		{ 75, 211, 48 },
		{ 75, 212, 48 },
		{ 75, 213, 48 },
		{ 75, 214, 48 },
		{ 75, 215, 48 },
		{ 75, 216, 48 },
		{ 75, 217, 48 },
		{ 75, 218, 48 },
		{ 75, 219, 48 },
		{ 75, 220, 48 },
		{ 75, 221, 48 },
		{ 75, 222, 48 },
		{ 75, 223, 48 },
		{ 75, 224, 48 },
		{ 75, 225, 48 },
		{ 75, 226, 48 },
		{ 75, 227, 48 },
		{ 75, 228, 48 },
		{ 75, 229, 48 },
		{ 75, 230, 48 },
		{ 75, 231, 48 },
		{ 75, 232, 48 },
		{ 75, 233, 48 },
		{ 75, 234, 48 },
		{ 76, 190, 48 },
		{ 76, 191, 48 },
		{ 76, 192, 48 },
		{ 76, 193, 48 },
		{ 76, 194, 48 },
		{ 76, 195, 48 },
		{ 76, 196, 48 },
		{ 76, 197, 48 },
		{ 76, 198, 48 },
		{ 76, 199, 48 },
		{ 76, 200, 48 },
		{ 76, 201, 48 },
		{ 76, 202, 48 },
		{ 76, 203, 48 },
		{ 76, 204, 48 },
		{ 76, 205, 48 },
		{ 76, 206, 48 },
		{ 76, 207, 48 },
		{ 76, 208, 48 },
		{ 76, 209, 48 },
		{ 76, 210, 48 },
		{ 76, 211, 48 },
		{ 76, 212, 48 },
		{ 76, 213, 48 },
		{ 76, 214, 48 },
		{ 76, 215, 48 },
		{ 76, 216, 48 },
		{ 76, 217, 48 },
		{ 76, 218, 48 },
		{ 76, 219, 48 },
		{ 76, 220, 48 },
		{ 76, 221, 48 },
		{ 76, 222, 48 },
		{ 76, 223, 48 },
		{ 76, 224, 48 },
		{ 76, 225, 48 },
		{ 76, 226, 48 },
		{ 76, 227, 48 },
		{ 76, 228, 48 },
		{ 76, 229, 48 },
		{ 76, 230, 48 },
		{ 76, 231, 48 },
		{ 76, 232, 48 },
		{ 76, 233, 48 },
		{ 76, 234, 48 },
		{ 77, 190, 48 },
		{ 77, 191, 48 },
		{ 77, 192, 48 },
		{ 77, 193, 48 },
		{ 77, 194, 48 },
		{ 77, 195, 48 },
		{ 77, 196, 48 },
		{ 77, 197, 48 },
		{ 77, 198, 48 },
		{ 77, 199, 48 },
		{ 77, 200, 48 },
		{ 77, 201, 48 },
		{ 77, 202, 48 },
		{ 77, 203, 48 },
		{ 77, 204, 48 },
		{ 77, 205, 48 },
		{ 77, 206, 48 },
		{ 77, 207, 48 },
		{ 77, 208, 48 },
		{ 77, 209, 48 },
		{ 77, 210, 48 },
		{ 77, 211, 48 },
		{ 77, 212, 48 },
		{ 77, 213, 48 },
		{ 77, 214, 48 },
		{ 77, 215, 48 },
		{ 77, 216, 48 },
		{ 77, 217, 48 },
		{ 77, 218, 48 },
		{ 77, 219, 48 },
		{ 77, 220, 48 },
		{ 77, 221, 48 },
		{ 77, 222, 48 },
		{ 77, 223, 48 },
		{ 77, 224, 48 },
		{ 77, 225, 48 },
		{ 77, 226, 48 },
		{ 77, 227, 48 },
		{ 77, 228, 48 },
		{ 77, 229, 48 },
		{ 77, 230, 48 },
		{ 77, 231, 48 },
		{ 77, 232, 48 },
		{ 77, 233, 48 },
		{ 77, 234, 48 },
		{ 78, 190, 48 },
		{ 78, 191, 48 },
		{ 78, 192, 48 },
		{ 78, 193, 48 },
		{ 78, 194, 48 },
		{ 78, 195, 48 },
		{ 78, 196, 48 },
		{ 78, 197, 48 },
		{ 78, 198, 48 },
		{ 78, 199, 48 },
		{ 78, 200, 48 },
		{ 78, 201, 48 },
		{ 78, 202, 48 },
		{ 78, 203, 48 },
		{ 78, 204, 48 },
		{ 78, 205, 48 },
		{ 78, 206, 48 },
		{ 78, 207, 48 },
		{ 78, 208, 48 },
		{ 78, 209, 48 },
		{ 78, 210, 48 },
		{ 78, 211, 48 },
		{ 78, 212, 48 },
		{ 78, 213, 48 },
		{ 78, 214, 48 },
		{ 78, 215, 48 },
		{ 78, 216, 48 },
		{ 78, 217, 48 },
		{ 78, 218, 48 },
		{ 78, 219, 48 },
		{ 78, 220, 48 },
		{ 78, 221, 48 },
		{ 78, 222, 48 },
		{ 78, 223, 48 },
		{ 78, 224, 48 },
		{ 78, 225, 48 },
		{ 78, 226, 48 },
		{ 78, 227, 48 },
		{ 78, 228, 48 },
		{ 78, 229, 48 },
		{ 78, 230, 48 },
		{ 78, 231, 48 },
		{ 78, 232, 48 },
		{ 78, 233, 48 },
		{ 78, 234, 48 },
		{ 79, 191, 48 },
		{ 79, 192, 48 },
		{ 79, 193, 48 },
		{ 79, 194, 48 },
		{ 79, 195, 48 },
		{ 79, 196, 48 },
		{ 79, 197, 48 },
		{ 79, 198, 48 },
		{ 79, 199, 48 },
		{ 79, 200, 48 },
		{ 79, 201, 48 },
		{ 79, 202, 48 },
		{ 79, 203, 48 },
		{ 79, 204, 48 },
		{ 79, 205, 48 },
		{ 79, 206, 48 },
		{ 79, 207, 48 },
		{ 79, 208, 48 },
		{ 79, 209, 48 },
		{ 79, 210, 48 },
		{ 79, 211, 48 },
		{ 79, 212, 48 },
		{ 79, 213, 48 },
		{ 79, 214, 48 },
		{ 79, 215, 48 },
		{ 79, 216, 48 },
		{ 79, 217, 48 },
		{ 79, 218, 48 },
		{ 79, 219, 48 },
		{ 79, 220, 48 },
		{ 79, 221, 48 },
		{ 79, 222, 48 },
		{ 79, 223, 48 },
		{ 79, 224, 48 },
		{ 79, 225, 48 },
		{ 79, 226, 48 },
		{ 79, 227, 48 },
		{ 79, 228, 48 },
		{ 79, 229, 48 },
		{ 79, 230, 48 },
		{ 79, 231, 48 },
		{ 79, 232, 48 },
		{ 79, 233, 48 },
		{ 79, 234, 48 },
		{ 80, 191, 48 },
		{ 80, 192, 48 },
		{ 80, 193, 48 },
		{ 80, 194, 48 },
		{ 80, 195, 48 },
		{ 80, 196, 48 },
		{ 80, 197, 48 },
		{ 80, 198, 48 },
		{ 80, 199, 48 },
		{ 80, 200, 48 },
		{ 80, 201, 48 },
		{ 80, 202, 48 },
		{ 80, 203, 48 },
		{ 80, 204, 48 },
		{ 80, 205, 48 },
		{ 80, 206, 48 },
		{ 80, 207, 48 },
		{ 80, 208, 48 },
		{ 80, 209, 48 },
		{ 80, 210, 48 },
		{ 80, 211, 48 },
		{ 80, 212, 48 },
		{ 80, 213, 48 },
		{ 80, 214, 48 },
		{ 80, 215, 48 },
		{ 80, 216, 48 },
		{ 80, 217, 48 },
		{ 80, 218, 48 },
		{ 80, 219, 48 },
		{ 80, 220, 48 },
		{ 80, 221, 48 },
		{ 80, 222, 48 },
		{ 80, 223, 48 },
		{ 80, 224, 48 },
		{ 80, 225, 48 },
		{ 80, 226, 48 },
		{ 80, 227, 48 },
		{ 80, 228, 48 },
		{ 80, 229, 48 },
		{ 80, 230, 48 },
		{ 80, 231, 48 },
		{ 80, 232, 48 },
		{ 80, 233, 48 },
		{ 80, 234, 48 },
		{ 81, 192, 48 },
		{ 81, 193, 48 },
		{ 81, 194, 48 },
		{ 81, 195, 48 },
		{ 81, 196, 48 },
		{ 81, 197, 48 },
		{ 81, 198, 48 },
		{ 81, 199, 48 },
		{ 81, 200, 48 },
		{ 81, 201, 48 },
		{ 81, 202, 48 },
		{ 81, 203, 48 },
		{ 81, 204, 48 },
		{ 81, 205, 48 },
		{ 81, 206, 48 },
		{ 81, 207, 48 },
		{ 81, 208, 48 },
		{ 81, 209, 48 },
		{ 81, 210, 48 },
		{ 81, 211, 48 },
		{ 81, 212, 48 },
		{ 81, 213, 48 },
		{ 81, 214, 48 },
		{ 81, 215, 48 },
		{ 81, 216, 48 },
		{ 81, 217, 48 },
		{ 81, 218, 48 },
		{ 81, 219, 48 },
		{ 81, 220, 48 },
		{ 81, 221, 48 },
		{ 81, 222, 48 },
		{ 81, 223, 48 },
		{ 81, 224, 48 },
		{ 81, 225, 48 },
		{ 81, 226, 48 },
		{ 81, 227, 48 },
		{ 81, 228, 48 },
		{ 81, 229, 48 },
		{ 81, 230, 48 },
		{ 81, 231, 48 },
		{ 81, 232, 48 },
		{ 81, 233, 48 },
		{ 81, 234, 48 },
		{ 82, 192, 48 },
		{ 82, 193, 48 },
		{ 82, 194, 48 },
		{ 82, 195, 48 },
		{ 82, 196, 48 },
		{ 82, 197, 48 },
		{ 82, 198, 48 },
		{ 82, 199, 48 },
		{ 82, 200, 48 },
		{ 82, 201, 48 },
		{ 82, 202, 48 },
		{ 82, 203, 48 },
		{ 82, 204, 48 },
		{ 82, 205, 48 },
		{ 82, 206, 48 },
		{ 82, 207, 48 },
		{ 82, 208, 48 },
		{ 82, 209, 48 },
		{ 82, 210, 48 },
		{ 82, 211, 48 },
		{ 82, 212, 48 },
		{ 82, 213, 48 },
		{ 82, 214, 48 },
		{ 82, 215, 48 },
		{ 82, 216, 48 },
		{ 82, 217, 48 },
		{ 82, 218, 48 },
		{ 82, 219, 48 },
		{ 82, 220, 48 },
		{ 82, 221, 48 },
		{ 82, 222, 48 },
		{ 82, 223, 48 },
		{ 82, 224, 48 },
		{ 82, 225, 48 },
		{ 82, 226, 48 },
		{ 82, 227, 48 },
		{ 82, 228, 48 },
		{ 82, 229, 48 },
		{ 82, 230, 48 },
		{ 82, 231, 48 },
		{ 82, 232, 48 },
		{ 82, 233, 48 },
		{ 82, 234, 48 },
		{ 82, 235, 48 },
		{ 83, 192, 48 },
		{ 83, 193, 48 },
		{ 83, 194, 48 },
		{ 83, 195, 48 },
		{ 83, 196, 48 },
		{ 83, 197, 48 },
		{ 83, 198, 48 },
		{ 83, 199, 48 },
		{ 83, 200, 48 },
		{ 83, 201, 48 },
		{ 83, 202, 48 },
		{ 83, 203, 48 },
		{ 83, 204, 48 },
		{ 83, 205, 48 },
		{ 83, 206, 48 },
		{ 83, 207, 48 },
		{ 83, 208, 48 },
		{ 83, 209, 48 },
		{ 83, 210, 48 },
		{ 83, 211, 48 },
		{ 83, 212, 48 },
		{ 83, 213, 48 },
		{ 83, 214, 48 },
		{ 83, 215, 48 },
		{ 83, 216, 48 },
		{ 83, 217, 48 },
		{ 83, 218, 48 },
		{ 83, 219, 48 },
		{ 83, 220, 48 },
		{ 83, 221, 48 },
		{ 83, 222, 48 },
		{ 83, 223, 48 },
		{ 83, 224, 48 },
		{ 83, 225, 48 },
		{ 83, 226, 48 },
		{ 83, 227, 48 },
		{ 83, 228, 48 },
		{ 83, 229, 48 },
		{ 83, 230, 48 },
		{ 83, 231, 48 },
		{ 83, 232, 48 },
		{ 83, 233, 48 },
		{ 83, 234, 48 },
		{ 83, 235, 48 },
		{ 84, 193, 48 },
		{ 84, 194, 48 },
		{ 84, 195, 48 },
		{ 84, 196, 48 },
		{ 84, 197, 48 },
		{ 84, 198, 48 },
		{ 84, 199, 48 },
		{ 84, 200, 48 },
		{ 84, 201, 48 },
		{ 84, 202, 48 },
		{ 84, 203, 48 },
		{ 84, 204, 48 },
		{ 84, 205, 48 },
		{ 84, 206, 48 },
		{ 84, 207, 48 },
		{ 84, 208, 48 },
		{ 84, 209, 48 },
		{ 84, 210, 48 },
		{ 84, 211, 48 },
		{ 84, 212, 48 },
		{ 84, 213, 48 },
		{ 84, 214, 48 },
		{ 84, 215, 48 },
		{ 84, 216, 48 },
		{ 84, 217, 48 },
		{ 84, 218, 48 },
		{ 84, 219, 48 },
		{ 84, 220, 48 },
		{ 84, 221, 48 },
		{ 84, 222, 48 },
		{ 84, 223, 48 },
		{ 84, 224, 48 },
		{ 84, 225, 48 },
		{ 84, 226, 48 },
		{ 84, 227, 48 },
		{ 84, 228, 48 },
		{ 84, 229, 48 },
		{ 84, 230, 48 },
		{ 84, 231, 48 },
		{ 84, 232, 48 },
		{ 84, 233, 48 },
		{ 84, 234, 48 },
		{ 84, 235, 48 },
		{ 85, 193, 48 },
		{ 85, 194, 48 },
		{ 85, 195, 48 },
		{ 85, 196, 48 },
		{ 85, 197, 48 },
		{ 85, 198, 48 },
		{ 85, 199, 48 },
		{ 85, 200, 48 },
		{ 85, 201, 48 },
		{ 85, 202, 48 },
		{ 85, 203, 48 },
		{ 85, 204, 48 },
		{ 85, 205, 48 },
		{ 85, 206, 48 },
		{ 85, 207, 48 },
		{ 85, 208, 48 },
		{ 85, 209, 48 },
		{ 85, 210, 48 },
		{ 85, 211, 48 },
		{ 85, 212, 48 },
		{ 85, 213, 48 },
		{ 85, 214, 48 },
		{ 85, 215, 48 },
		{ 85, 216, 48 },
		{ 85, 217, 48 },
		{ 85, 218, 48 },
		{ 85, 219, 48 },
		{ 85, 220, 48 },
		{ 85, 221, 48 },
		{ 85, 222, 48 },
		{ 85, 223, 48 },
		{ 85, 224, 48 },
		{ 85, 225, 48 },
		{ 85, 226, 48 },
		{ 85, 227, 48 },
		{ 85, 228, 48 },
		{ 85, 229, 48 },
		{ 85, 230, 48 },
		{ 85, 231, 48 },
		{ 85, 232, 48 },
		{ 85, 233, 48 },
		{ 85, 234, 48 },
		{ 85, 235, 48 },
		{ 86, 193, 48 },
		{ 86, 194, 48 },
		{ 86, 195, 48 },
		{ 86, 196, 48 },
		{ 86, 197, 48 },
		{ 86, 198, 48 },
		{ 86, 199, 48 },
		{ 86, 200, 48 },
		{ 86, 201, 48 },
		{ 86, 202, 48 },
		{ 86, 203, 48 },
		{ 86, 204, 48 },
		{ 86, 205, 48 },
		{ 86, 206, 48 },
		{ 86, 207, 48 },
		{ 86, 208, 48 },
		{ 86, 209, 48 },
		{ 86, 210, 48 },
		{ 86, 211, 48 },
		{ 86, 212, 48 },
		{ 86, 213, 48 },
		{ 86, 214, 48 },
		{ 86, 215, 48 },
		{ 86, 216, 48 },
		{ 86, 217, 48 },
		{ 86, 218, 48 },
		{ 86, 219, 48 },
		{ 86, 220, 48 },
		{ 86, 221, 48 },
		{ 86, 222, 48 },
		{ 86, 223, 48 },
		{ 86, 224, 48 },
		{ 86, 225, 48 },
		{ 86, 226, 48 },
		{ 86, 227, 48 },
		{ 86, 228, 48 },
		{ 86, 229, 48 },
		{ 86, 230, 48 },
		{ 86, 231, 48 },
		{ 86, 232, 48 },
		{ 86, 233, 48 },
		{ 86, 234, 48 },
		{ 86, 235, 48 },
		{ 87, 194, 48 },
		{ 87, 195, 48 },
		{ 87, 196, 48 },
		{ 87, 197, 48 },
		{ 87, 198, 48 },
		{ 87, 199, 48 },
		{ 87, 200, 48 },
		{ 87, 201, 48 },
		{ 87, 202, 48 },
		{ 87, 203, 48 },
		{ 87, 204, 48 },
		{ 87, 205, 48 },
		{ 87, 206, 48 },
		{ 87, 207, 48 },
		{ 87, 208, 48 },
		{ 87, 209, 48 },
		{ 87, 210, 48 },
		{ 87, 211, 48 },
		{ 87, 212, 48 },
		{ 87, 213, 48 },
		{ 87, 214, 48 },
		{ 87, 215, 48 },
		{ 87, 216, 48 },
		{ 87, 217, 48 },
		{ 87, 218, 48 },
		{ 87, 219, 48 },
		{ 87, 220, 48 },
		{ 87, 221, 48 },
		{ 87, 222, 48 },
		{ 87, 223, 48 },
		{ 87, 224, 48 },
		{ 87, 225, 48 },
		{ 87, 226, 48 },
		{ 87, 227, 48 },
		{ 87, 228, 48 },
		{ 87, 229, 48 },
		{ 87, 230, 48 },
		{ 87, 231, 48 },
		{ 87, 232, 48 },
		{ 87, 233, 48 },
		{ 87, 234, 48 },
		{ 87, 235, 48 },
		{ 88, 194, 48 },
		{ 88, 195, 48 },
		{ 88, 196, 48 },
		{ 88, 197, 48 },
		{ 88, 198, 48 },
		{ 88, 199, 48 },
		{ 88, 200, 48 },
		{ 88, 201, 48 },
		{ 88, 202, 48 },
		{ 88, 203, 48 },
		{ 88, 204, 48 },
		{ 88, 205, 48 },
		{ 88, 206, 48 },
		{ 88, 207, 48 },
		{ 88, 208, 48 },
		{ 88, 209, 48 },
		{ 88, 210, 48 },
		{ 88, 211, 48 },
		{ 88, 212, 48 },
		{ 88, 213, 48 },
		{ 88, 214, 48 },
		{ 88, 215, 48 },
		{ 88, 216, 48 },
		{ 88, 217, 48 },
		{ 88, 218, 48 },
		{ 88, 219, 48 },
		{ 88, 220, 48 },
		{ 88, 221, 48 },
		{ 88, 222, 48 },
		{ 88, 223, 48 },
		{ 88, 224, 48 },
		{ 88, 225, 48 },
		{ 88, 226, 48 },
		{ 88, 227, 48 },
		{ 88, 228, 48 },
		{ 88, 229, 48 },
		{ 88, 230, 48 },
		{ 88, 231, 48 },
		{ 88, 232, 48 },
		{ 88, 233, 48 },
		{ 88, 234, 48 },
		{ 88, 235, 48 },
		{ 89, 195, 48 },
		{ 89, 196, 48 },
		{ 89, 197, 48 },
		{ 89, 198, 48 },
		{ 89, 199, 48 },
		{ 89, 200, 48 },
		{ 89, 201, 48 },
		{ 89, 202, 48 },
		{ 89, 203, 48 },
		{ 89, 204, 48 },
		{ 89, 205, 48 },
		{ 89, 206, 48 },
		{ 89, 207, 48 },
		{ 89, 208, 48 },
		{ 89, 209, 48 },
		{ 89, 210, 48 },
		{ 89, 211, 48 },
		{ 89, 212, 48 },
		{ 89, 213, 48 },
		{ 89, 214, 48 },
		{ 89, 215, 48 },
		{ 89, 216, 48 },
		{ 89, 217, 48 },
		{ 89, 218, 48 },
		{ 89, 219, 48 },
		{ 89, 220, 48 },
		{ 89, 221, 48 },
		{ 89, 222, 48 },
		{ 89, 223, 48 },
		{ 89, 224, 48 },
		{ 89, 225, 48 },
		{ 89, 226, 48 },
		{ 89, 227, 48 },
		{ 89, 228, 48 },
		{ 89, 229, 48 },
		{ 89, 230, 48 },
		{ 89, 231, 48 },
		{ 89, 232, 48 },
		{ 89, 233, 48 },
		{ 89, 234, 48 },
		{ 89, 235, 48 },
		{ 90, 195, 48 },
		{ 90, 196, 48 },
		{ 90, 197, 48 },
		{ 90, 198, 48 },
		{ 90, 199, 48 },
		{ 90, 200, 48 },
		{ 90, 201, 48 },
		{ 90, 202, 48 },
		{ 90, 203, 48 },
		{ 90, 204, 48 },
		{ 90, 205, 48 },
		{ 90, 206, 48 },
		{ 90, 207, 48 },
		{ 90, 208, 48 },
		{ 90, 209, 48 },
		{ 90, 210, 48 },
		{ 90, 211, 48 },
		{ 90, 212, 48 },
		{ 90, 213, 48 },
		{ 90, 214, 48 },
		{ 90, 215, 48 },
		{ 90, 216, 48 },
		{ 90, 217, 48 },
		{ 90, 218, 48 },
		{ 90, 219, 48 },
		{ 90, 220, 48 },
		{ 90, 221, 48 },
		{ 90, 222, 48 },
		{ 90, 223, 48 },
		{ 90, 224, 48 },
		{ 90, 225, 48 },
		{ 90, 226, 48 },
		{ 90, 227, 48 },
		{ 90, 228, 48 },
		{ 90, 229, 48 },
		{ 90, 230, 48 },
		{ 90, 231, 48 },
		{ 90, 232, 48 },
		{ 90, 233, 48 },
		{ 90, 234, 48 },
		{ 90, 235, 48 },
		{ 90, 236, 48 },
		{ 91, 195, 48 },
		{ 91, 196, 48 },
		{ 91, 197, 48 },
		{ 91, 198, 48 },
		{ 91, 199, 48 },
		{ 91, 200, 48 },
		{ 91, 201, 48 },
		{ 91, 202, 48 },
		{ 91, 203, 48 },
		{ 91, 204, 48 },
		{ 91, 205, 48 },
		{ 91, 206, 48 },
		{ 91, 207, 48 },
		{ 91, 208, 48 },
		{ 91, 209, 48 },
		{ 91, 210, 48 },
		{ 91, 211, 48 },
		{ 91, 212, 48 },
		{ 91, 213, 48 },
		{ 91, 214, 48 },
		{ 91, 215, 48 },
		{ 91, 216, 48 },
		{ 91, 217, 48 },
		{ 91, 218, 48 },
		{ 91, 219, 48 },
		{ 91, 220, 48 },
		{ 91, 221, 48 },
		{ 91, 222, 48 },
		{ 91, 223, 48 },
		{ 91, 224, 48 },
		{ 91, 225, 48 },
		{ 91, 226, 48 },
		{ 91, 227, 48 },
		{ 91, 228, 48 },
		{ 91, 229, 48 },
		{ 91, 230, 48 },
		{ 91, 231, 48 },
		{ 91, 232, 48 },
		{ 91, 233, 48 },
		{ 91, 234, 48 },
		{ 91, 235, 48 },
		{ 91, 236, 48 },
		{ 92, 196, 48 },
		{ 92, 197, 48 },
		{ 92, 198, 48 },
		{ 92, 199, 48 },
		{ 92, 200, 48 },
		{ 92, 201, 48 },
		{ 92, 202, 48 },
		{ 92, 203, 48 },
		{ 92, 204, 48 },
		{ 92, 205, 48 },
		{ 92, 206, 48 },
		{ 92, 207, 48 },
		{ 92, 208, 48 },
		{ 92, 209, 48 },
		{ 92, 210, 48 },
		{ 92, 211, 48 },
		{ 92, 212, 48 },
		{ 92, 213, 48 },
		{ 92, 214, 48 },
		{ 92, 215, 48 },
		{ 92, 216, 48 },
		{ 92, 217, 48 },
		{ 92, 218, 48 },
		{ 92, 219, 48 },
		{ 92, 220, 48 },
		{ 92, 221, 48 },
		{ 92, 222, 48 },
		{ 92, 223, 48 },
		{ 92, 224, 48 },
		{ 92, 225, 48 },
		{ 92, 226, 48 },
		{ 92, 227, 48 },
		{ 92, 228, 48 },
		{ 92, 229, 48 },
		{ 92, 230, 48 },
		{ 92, 231, 48 },
		{ 92, 232, 48 },
		{ 92, 233, 48 },
		{ 92, 234, 48 },
		{ 92, 235, 48 },
		{ 92, 236, 48 },
		{ 93, 196, 48 },
		{ 93, 197, 48 },
		{ 93, 198, 48 },
		{ 93, 199, 48 },
		{ 93, 200, 48 },
		{ 93, 201, 48 },
		{ 93, 202, 48 },
		{ 93, 203, 48 },
		{ 93, 204, 48 },
		{ 93, 205, 48 },
		{ 93, 206, 48 },
		{ 93, 207, 48 },
		{ 93, 208, 48 },
		{ 93, 209, 48 },
		{ 93, 210, 48 },
		{ 93, 211, 48 },
		{ 93, 212, 48 },
		{ 93, 213, 48 },
		{ 93, 214, 48 },
		{ 93, 215, 48 },
		{ 93, 216, 48 },
		{ 93, 217, 48 },
		{ 93, 218, 48 },
		{ 93, 219, 48 },
		{ 93, 220, 48 },
		{ 93, 221, 48 },
		{ 93, 222, 48 },
		{ 93, 223, 48 },
		{ 93, 224, 48 },
		{ 93, 225, 48 },
		{ 93, 226, 48 },
		{ 93, 227, 48 },
		{ 93, 228, 48 },
		{ 93, 229, 48 },
		{ 93, 230, 48 },
		{ 93, 231, 48 },
		{ 93, 232, 48 },
		{ 93, 233, 48 },
		{ 93, 234, 48 },
		{ 93, 235, 48 },
		{ 93, 236, 48 },
		{ 94, 196, 48 },
		{ 94, 197, 48 },
		{ 94, 198, 48 },
		{ 94, 199, 48 },
		{ 94, 200, 48 },
		{ 94, 201, 48 },
		{ 94, 202, 48 },
		{ 94, 203, 48 },
		{ 94, 204, 48 },
		{ 94, 205, 48 },
		{ 94, 206, 48 },
		{ 94, 207, 48 },
		{ 94, 208, 48 },
		{ 94, 209, 48 },
		{ 94, 210, 48 },
		{ 94, 211, 48 },
		{ 94, 212, 48 },
		{ 94, 213, 48 },
		{ 94, 214, 48 },
		{ 94, 215, 48 },
		{ 94, 216, 48 },
		{ 94, 217, 48 },
		{ 94, 218, 48 },
		{ 94, 219, 48 },
		{ 94, 220, 48 },
		{ 94, 221, 48 },
		{ 94, 222, 48 },
		{ 94, 223, 48 },
		{ 94, 224, 48 },
		{ 94, 225, 48 },
		{ 94, 226, 48 },
		{ 94, 227, 48 },
		{ 94, 228, 48 },
		{ 94, 229, 48 },
		{ 94, 230, 48 },
		{ 94, 231, 48 },
		{ 94, 232, 48 },
		{ 94, 233, 48 },
		{ 94, 234, 48 },
		{ 94, 235, 48 },
		{ 94, 236, 48 },
		{ 95, 197, 48 },
		{ 95, 198, 48 },
		{ 95, 199, 48 },
		{ 95, 200, 48 },
		{ 95, 201, 48 },
		{ 95, 202, 48 },
		{ 95, 203, 48 },
		{ 95, 204, 48 },
		{ 95, 205, 48 },
		{ 95, 206, 48 },
		{ 95, 207, 48 },
		{ 95, 208, 48 },
		{ 95, 209, 48 },
		{ 95, 210, 48 },
		{ 95, 211, 48 },
		{ 95, 212, 48 },
		{ 95, 213, 48 },
		{ 95, 214, 48 },
		{ 95, 215, 48 },
		{ 95, 216, 48 },
		{ 95, 217, 48 },
		{ 95, 218, 48 },
		{ 95, 219, 48 },
		{ 95, 220, 48 },
		{ 95, 221, 48 },
		{ 95, 222, 48 },
		{ 95, 223, 48 },
		{ 95, 224, 48 },
		{ 95, 225, 48 },
		{ 95, 226, 48 },
		{ 95, 227, 48 },
		{ 95, 228, 48 },
		{ 95, 229, 48 },
		{ 95, 230, 48 },
		{ 95, 231, 48 },
		{ 95, 232, 48 },
		{ 95, 233, 48 },
		{ 95, 234, 48 },
		{ 95, 235, 48 },
		{ 95, 236, 48 },
		{ 96, 197, 48 },
		{ 96, 198, 48 },
		{ 96, 199, 48 },
		{ 96, 200, 48 },
		{ 96, 201, 48 },
		{ 96, 202, 48 },
		{ 96, 203, 48 },
		{ 96, 204, 48 },
		{ 96, 205, 48 },
		{ 96, 206, 48 },
		{ 96, 207, 48 },
		{ 96, 208, 48 },
		{ 96, 209, 48 },
		{ 96, 210, 48 },
		{ 96, 211, 48 },
		{ 96, 212, 48 },
		{ 96, 213, 48 },
		{ 96, 214, 48 },
		{ 96, 215, 48 },
		{ 96, 216, 48 },
		{ 96, 217, 48 },
		{ 96, 218, 48 },
		{ 96, 219, 48 },
		{ 96, 220, 48 },
		{ 96, 221, 48 },
		{ 96, 222, 48 },
		{ 96, 223, 48 },
		{ 96, 224, 48 },
		{ 96, 225, 48 },
		{ 96, 226, 48 },
		{ 96, 227, 48 },
		{ 96, 228, 48 },
		{ 96, 229, 48 },
		{ 96, 230, 48 },
		{ 96, 231, 48 },
		{ 96, 232, 48 },
		{ 96, 233, 48 },
		{ 96, 234, 48 },
		{ 96, 235, 48 },
		{ 96, 236, 48 },
		{ 97, 198, 48 },
		{ 97, 199, 48 },
		{ 97, 200, 48 },
		{ 97, 201, 48 },
		{ 97, 202, 48 },
		{ 97, 203, 48 },
		{ 97, 204, 48 },
		{ 97, 205, 48 },
		{ 97, 206, 48 },
		{ 97, 207, 48 },
		{ 97, 208, 48 },
		{ 97, 209, 48 },
		{ 97, 210, 48 },
		{ 97, 211, 48 },
		{ 97, 212, 48 },
		{ 97, 213, 48 },
		{ 97, 214, 48 },
		{ 97, 215, 48 },
		{ 97, 216, 48 },
		{ 97, 217, 48 },
		{ 97, 218, 48 },
		{ 97, 219, 48 },
		{ 97, 220, 48 },
		{ 97, 221, 48 },
		{ 97, 222, 48 },
		{ 97, 223, 48 },
		{ 97, 224, 48 },
		{ 97, 225, 48 },
		{ 97, 226, 48 },
		{ 97, 227, 48 },
		{ 97, 228, 48 },
		{ 97, 229, 48 },
		{ 97, 230, 48 },
		{ 97, 231, 48 },
		{ 97, 232, 48 },
		{ 97, 233, 48 },
		{ 97, 234, 48 },
		{ 97, 235, 48 },
		{ 97, 236, 48 },
		{ 98, 198, 48 },
		{ 98, 199, 48 },
		{ 98, 200, 48 },
		{ 98, 201, 48 },
		{ 98, 202, 48 },
		{ 98, 203, 48 },
		{ 98, 204, 48 },
		{ 98, 205, 48 },
		{ 98, 206, 48 },
		{ 98, 207, 48 },
		{ 98, 208, 48 },
		{ 98, 209, 48 },
		{ 98, 210, 48 },
		{ 98, 211, 48 },
		{ 98, 212, 48 },
		{ 98, 213, 48 },
		{ 98, 214, 48 },
		{ 98, 215, 48 },
		{ 98, 216, 48 },
		{ 98, 217, 48 },
		{ 98, 218, 48 },
		{ 98, 219, 48 },
		{ 98, 220, 48 },
		{ 98, 221, 48 },
		{ 98, 222, 48 },
		{ 98, 223, 48 },
		{ 98, 224, 48 },
		{ 98, 225, 48 },
		{ 98, 226, 48 },
		{ 98, 227, 48 },
		{ 98, 228, 48 },
		{ 98, 229, 48 },
		{ 98, 230, 48 },
		{ 98, 231, 48 },
		{ 98, 232, 48 },
		{ 98, 233, 48 },
		{ 98, 234, 48 },
		{ 98, 235, 48 },
		{ 98, 236, 48 },
		{ 98, 237, 48 },
		{ 99, 198, 48 },
		{ 99, 199, 48 },
		{ 99, 200, 48 },
		{ 99, 201, 48 },
		{ 99, 202, 48 },
		{ 99, 203, 48 },
		{ 99, 204, 48 },
		{ 99, 205, 48 },
		{ 99, 206, 48 },
		{ 99, 207, 48 },
		{ 99, 208, 48 },
		{ 99, 209, 48 },
		{ 99, 210, 48 },
		{ 99, 211, 48 },
		{ 99, 212, 48 },
		{ 99, 213, 48 },
		{ 99, 214, 48 },
		{ 99, 215, 48 },
		{ 99, 216, 48 },
		{ 99, 217, 48 },
		{ 99, 218, 48 },
		{ 99, 219, 48 },
		{ 99, 220, 48 },
		{ 99, 221, 48 },
		{ 99, 222, 48 },
		{ 99, 223, 48 },
		{ 99, 224, 48 },
		{ 99, 225, 48 },
		{ 99, 226, 48 },
		{ 99, 227, 48 },
		{ 99, 228, 48 },
		{ 99, 229, 48 },
		{ 99, 230, 48 },
		{ 99, 231, 48 },
		{ 99, 232, 48 },
		{ 99, 233, 48 },
		{ 99, 234, 48 },
		{ 99, 235, 48 },
		{ 99, 236, 48 },
		{ 99, 237, 48 },
		{ 100, 199, 48 },
		{ 100, 200, 48 },
		{ 100, 201, 48 },
		{ 100, 202, 48 },
		{ 100, 203, 48 },
		{ 100, 204, 48 },
		{ 100, 205, 48 },
		{ 100, 206, 48 },
		{ 100, 207, 48 },
		{ 100, 208, 48 },
		{ 100, 209, 48 },
		{ 100, 210, 48 },
		{ 100, 211, 48 },
		{ 100, 212, 48 },
		{ 100, 213, 48 },
		{ 100, 214, 48 },
		{ 100, 215, 48 },
		{ 100, 216, 48 },
		{ 100, 217, 48 },
		{ 100, 218, 48 },
		{ 100, 219, 48 },
		{ 100, 220, 48 },
		{ 100, 221, 48 },
		{ 100, 222, 48 },
		{ 100, 223, 48 },
		{ 100, 224, 48 },
		{ 100, 225, 48 },
		{ 100, 226, 48 },
		{ 100, 227, 48 },
		{ 100, 228, 48 },
		{ 100, 229, 48 },
		{ 100, 230, 48 },
		{ 100, 231, 48 },
		{ 100, 232, 48 },
		{ 100, 233, 48 },
		{ 100, 234, 48 },
		{ 100, 235, 48 },
		{ 100, 236, 48 },
		{ 100, 237, 48 },
		{ 101, 199, 48 },
		{ 101, 200, 48 },
		{ 101, 201, 48 },
		{ 101, 202, 48 },
		{ 101, 203, 48 },
		{ 101, 204, 48 },
		{ 101, 205, 48 },
		{ 101, 206, 48 },
		{ 101, 207, 48 },
		{ 101, 208, 48 },
		{ 101, 209, 48 },
		{ 101, 210, 48 },
		{ 101, 211, 48 },
		{ 101, 212, 48 },
		{ 101, 213, 48 },
		{ 101, 214, 48 },
		{ 101, 215, 48 },
		{ 101, 216, 48 },
		{ 101, 217, 48 },
		{ 101, 218, 48 },
		{ 101, 219, 48 },
		{ 101, 220, 48 },
		{ 101, 221, 48 },
		{ 101, 222, 48 },
		{ 101, 223, 48 },
		{ 101, 224, 48 },
		{ 101, 225, 48 },
		{ 101, 226, 48 },
		{ 101, 227, 48 },
		{ 101, 228, 48 },
		{ 101, 229, 48 },
		{ 101, 230, 48 },
		{ 101, 231, 48 },
		{ 101, 232, 48 },
		{ 101, 233, 48 },
		{ 101, 234, 48 },
		{ 101, 235, 48 },
		{ 101, 236, 48 },
		{ 101, 237, 48 },
		{ 102, 199, 48 },
		{ 102, 200, 48 },
		{ 102, 201, 48 },
		{ 102, 202, 48 },
		{ 102, 203, 48 },
		{ 102, 204, 48 },
		{ 102, 205, 48 },
		{ 102, 206, 48 },
		{ 102, 207, 48 },
		{ 102, 208, 48 },
		{ 102, 209, 48 },
		{ 102, 210, 48 },
		{ 102, 211, 48 },
		{ 102, 212, 48 },
		{ 102, 213, 48 },
		{ 102, 214, 48 },
		{ 102, 215, 48 },
		{ 102, 216, 48 },
		{ 102, 217, 48 },
		{ 102, 218, 48 },
		{ 102, 219, 48 },
		{ 102, 220, 48 },
		{ 102, 221, 48 },
		{ 102, 222, 48 },
		{ 102, 223, 48 },
		{ 102, 224, 48 },
		{ 102, 225, 48 },
		{ 102, 226, 48 },
		{ 102, 227, 48 },
		{ 102, 228, 48 },
		{ 102, 229, 48 },
		{ 102, 230, 48 },
		{ 102, 231, 48 },
		{ 102, 232, 48 },
		{ 102, 233, 48 },
		{ 102, 234, 48 },
		{ 102, 235, 48 },
		{ 102, 236, 48 },
		{ 102, 237, 48 },
		{ 103, 200, 48 },
		{ 103, 201, 48 },
		{ 103, 202, 48 },
		{ 103, 203, 48 },
		{ 103, 204, 48 },
		{ 103, 205, 48 },
		{ 103, 206, 48 },
		{ 103, 207, 48 },
		{ 103, 208, 48 },
		{ 103, 209, 48 },
		{ 103, 210, 48 },
		{ 103, 211, 48 },
		{ 103, 212, 48 },
		{ 103, 213, 48 },
		{ 103, 214, 48 },
		{ 103, 215, 48 },
		{ 103, 216, 48 },
		{ 103, 217, 48 },
		{ 103, 218, 48 },
		{ 103, 219, 48 },
		{ 103, 220, 48 },
		{ 103, 221, 48 },
		{ 103, 222, 48 },
		{ 103, 223, 48 },
		{ 103, 224, 48 },
		{ 103, 225, 48 },
		{ 103, 226, 48 },
		{ 103, 227, 48 },
		{ 103, 228, 48 },
		{ 103, 229, 48 },
		{ 103, 230, 48 },
		{ 103, 231, 48 },
		{ 103, 232, 48 },
		{ 103, 233, 48 },
		{ 103, 234, 48 },
		{ 103, 235, 48 },
		{ 103, 236, 48 },
		{ 103, 237, 48 },
		{ 104, 200, 48 },
		{ 104, 201, 48 },
		{ 104, 202, 48 },
		{ 104, 203, 48 },
		{ 104, 204, 48 },
		{ 104, 205, 48 },
		{ 104, 206, 48 },
		{ 104, 207, 48 },
		{ 104, 208, 48 },
		{ 104, 209, 48 },
		{ 104, 210, 48 },
		{ 104, 211, 48 },
		{ 104, 212, 48 },
		{ 104, 213, 48 },
		{ 104, 214, 48 },
		{ 104, 215, 48 },
		{ 104, 216, 48 },
		{ 104, 217, 48 },
		{ 104, 218, 48 },
		{ 104, 219, 48 },
		{ 104, 220, 48 },
		{ 104, 221, 48 },
		{ 104, 222, 48 },
		{ 104, 223, 48 },
		{ 104, 224, 48 },
		{ 104, 225, 48 },
		{ 104, 226, 48 },
		{ 104, 227, 48 },
		{ 104, 228, 48 },
		{ 104, 229, 48 },
		{ 104, 230, 48 },
		{ 104, 231, 48 },
		{ 104, 232, 48 },
		{ 104, 233, 48 },
		{ 104, 234, 48 },
		{ 104, 235, 48 },
		{ 104, 236, 48 },
		{ 104, 237, 48 },
		{ 105, 201, 48 },
		{ 105, 202, 48 },
		{ 105, 203, 48 },
		{ 105, 204, 48 },
		{ 105, 205, 48 },
		{ 105, 206, 48 },
		{ 105, 207, 48 },
		{ 105, 208, 48 },
		{ 105, 209, 48 },
		{ 105, 210, 48 },
		{ 105, 211, 48 },
		{ 105, 212, 48 },
		{ 105, 213, 48 },
		{ 105, 214, 48 },
		{ 105, 215, 48 },
		{ 105, 216, 48 },
		{ 105, 217, 48 },
		{ 105, 218, 48 },
		{ 105, 219, 48 },
		{ 105, 220, 48 },
		{ 105, 221, 48 },
		{ 105, 222, 48 },
		{ 105, 223, 48 },
		{ 105, 224, 48 },
		{ 105, 225, 48 },
		{ 105, 226, 48 },
		{ 105, 227, 48 },
		{ 105, 228, 48 },
		{ 105, 229, 48 },
		{ 105, 230, 48 },
		{ 105, 231, 48 },
		{ 105, 232, 48 },
		{ 105, 233, 48 },
		{ 105, 234, 48 },
		{ 105, 235, 48 },
		{ 105, 236, 48 },
		{ 105, 237, 48 },
		{ 106, 201, 48 },
		{ 106, 202, 48 },
		{ 106, 203, 48 },
		{ 106, 204, 48 },
		{ 106, 205, 48 },
		{ 106, 206, 48 },
		{ 106, 207, 48 },
		{ 106, 208, 48 },
		{ 106, 209, 48 },
		{ 106, 210, 48 },
		{ 106, 211, 48 },
		{ 106, 212, 48 },
		{ 106, 213, 48 },
		{ 106, 214, 48 },
		{ 106, 215, 48 },
		{ 106, 216, 48 },
		{ 106, 217, 48 },
		{ 106, 218, 48 },
		{ 106, 219, 48 },
		{ 106, 220, 48 },
		{ 106, 221, 48 },
		{ 106, 222, 48 },
		{ 106, 223, 48 },
		{ 106, 224, 48 },
		{ 106, 225, 48 },
		{ 106, 226, 48 },
		{ 106, 227, 48 },
		{ 106, 228, 48 },
		{ 106, 229, 48 },
		{ 106, 230, 48 },
		{ 106, 231, 48 },
		{ 106, 232, 48 },
		{ 106, 233, 48 },
		{ 106, 234, 48 },
		{ 106, 235, 48 },
		{ 106, 236, 48 },
		{ 106, 237, 48 },
		{ 106, 238, 48 },
		{ 107, 201, 48 },
		{ 107, 202, 48 },
		{ 107, 203, 48 },
		{ 107, 204, 48 },
		{ 107, 205, 48 },
		{ 107, 206, 48 },
		{ 107, 207, 48 },
		{ 107, 208, 48 },
		{ 107, 209, 48 },
		{ 107, 210, 48 },
		{ 107, 211, 48 },
		{ 107, 212, 48 },
		{ 107, 213, 48 },
		{ 107, 214, 48 },
		{ 107, 215, 48 },
		{ 107, 216, 48 },
		{ 107, 217, 48 },
		{ 107, 218, 48 },
		{ 107, 219, 48 },
		{ 107, 220, 48 },
		{ 107, 221, 48 },
		{ 107, 222, 48 },
		{ 107, 223, 48 },
		{ 107, 224, 48 },
		{ 107, 225, 48 },
		{ 107, 226, 48 },
		{ 107, 227, 48 },
		{ 107, 228, 48 },
		{ 107, 229, 48 },
		{ 107, 230, 48 },
		{ 107, 231, 48 },
		{ 107, 232, 48 },
		{ 107, 233, 48 },
		{ 107, 234, 48 },
		{ 107, 235, 48 },
		{ 107, 236, 48 },
		{ 107, 237, 48 },
		{ 107, 238, 48 },
		{ 108, 202, 48 },
		{ 108, 203, 48 },
		{ 108, 204, 48 },
		{ 108, 205, 48 },
		{ 108, 206, 48 },
		{ 108, 207, 48 },
		{ 108, 208, 48 },
		{ 108, 209, 48 },
		{ 108, 210, 48 },
		{ 108, 211, 48 },
		{ 108, 212, 48 },
		{ 108, 213, 48 },
		{ 108, 214, 48 },
		{ 108, 215, 48 },
		{ 108, 216, 48 },
		{ 108, 217, 48 },
		{ 108, 218, 48 },
		{ 108, 219, 48 },
		{ 108, 220, 48 },
		{ 108, 221, 48 },
		{ 108, 222, 48 },
		{ 108, 223, 48 },
		{ 108, 224, 48 },
		{ 108, 225, 48 },
		{ 108, 226, 48 },
		{ 108, 227, 48 },
		{ 108, 228, 48 },
		{ 108, 229, 48 },
		{ 108, 230, 48 },
		{ 108, 231, 48 },
		{ 108, 232, 48 },
		{ 108, 233, 48 },
		{ 108, 234, 48 },
		{ 108, 235, 48 },
		{ 108, 236, 48 },
		{ 108, 237, 48 },
		{ 108, 238, 48 },
		{ 109, 202, 48 },
		{ 109, 203, 48 },
		{ 109, 204, 48 },
		{ 109, 205, 48 },
		{ 109, 206, 48 },
		{ 109, 207, 48 },
		{ 109, 208, 48 },
		{ 109, 209, 48 },
		{ 109, 210, 48 },
		{ 109, 211, 48 },
		{ 109, 212, 48 },
		{ 109, 213, 48 },
		{ 109, 214, 48 },
		{ 109, 215, 48 },
		{ 109, 216, 48 },
		{ 109, 217, 48 },
		{ 109, 218, 48 },
		{ 109, 219, 48 },
		{ 109, 220, 48 },
		{ 109, 221, 48 },
		{ 109, 222, 48 },
		{ 109, 223, 48 },
		{ 109, 224, 48 },
		{ 109, 225, 48 },
		{ 109, 226, 48 },
		{ 109, 227, 48 },
		{ 109, 228, 48 },
		{ 109, 229, 48 },
		{ 109, 230, 48 },
		{ 109, 231, 48 },
		{ 109, 232, 48 },
		{ 109, 233, 48 },
		{ 109, 234, 48 },
		{ 109, 235, 48 },
		{ 109, 236, 48 },
		{ 109, 237, 48 },
		{ 109, 238, 48 },
		{ 110, 202, 48 },
		{ 110, 203, 48 },
		{ 110, 204, 48 },
		{ 110, 205, 48 },
		{ 110, 206, 48 },
		{ 110, 207, 48 },
		{ 110, 208, 48 },
		{ 110, 209, 48 },
		{ 110, 210, 48 },
		{ 110, 211, 48 },
		{ 110, 212, 48 },
		{ 110, 213, 48 },
		{ 110, 214, 48 },
		{ 110, 215, 48 },
		{ 110, 216, 48 },
		{ 110, 217, 48 },
		{ 110, 218, 48 },
		{ 110, 219, 48 },
		{ 110, 220, 48 },
		{ 110, 221, 48 },
		{ 110, 222, 48 },
		{ 110, 223, 48 },
		{ 110, 224, 48 },
		{ 110, 225, 48 },
		{ 110, 226, 48 },
		{ 110, 227, 48 },
		{ 110, 228, 48 },
		{ 110, 229, 48 },
		{ 110, 230, 48 },
		{ 110, 231, 48 },
		{ 110, 232, 48 },
		{ 110, 233, 48 },
		{ 110, 234, 48 },
		{ 110, 235, 48 },
		{ 110, 236, 48 },
		{ 110, 237, 48 },
		{ 110, 238, 48 },
		{ 111, 203, 48 },
		{ 111, 204, 48 },
		{ 111, 205, 48 },
		{ 111, 206, 48 },
		{ 111, 207, 48 },
		{ 111, 208, 48 },
		{ 111, 209, 48 },
		{ 111, 210, 48 },
		{ 111, 211, 48 },
		{ 111, 212, 48 },
		{ 111, 213, 48 },
		{ 111, 214, 48 },
		{ 111, 215, 48 },
		{ 111, 216, 48 },
		{ 111, 217, 48 },
		{ 111, 218, 48 },
		{ 111, 219, 48 },
		{ 111, 220, 48 },
		{ 111, 221, 48 },
		{ 111, 222, 48 },
		{ 111, 223, 48 },
		{ 111, 224, 48 },
		{ 111, 225, 48 },
		{ 111, 226, 48 },
		{ 111, 227, 48 },
		{ 111, 228, 48 },
		{ 111, 229, 48 },
		{ 111, 230, 48 },
		{ 111, 231, 48 },
		{ 111, 232, 48 },
		{ 111, 233, 48 },
		{ 111, 234, 48 },
		{ 111, 235, 48 },
		{ 111, 236, 48 },
		{ 111, 237, 48 },
		{ 111, 238, 48 },
		{ 112, 203, 48 },
		{ 112, 204, 48 },
		{ 112, 205, 48 },
		{ 112, 206, 48 },
		{ 112, 207, 48 },
		{ 112, 208, 48 },
		{ 112, 209, 48 },
		{ 112, 210, 48 },
		{ 112, 211, 48 },
		{ 112, 212, 48 },
		{ 112, 213, 48 },
		{ 112, 214, 48 },
		{ 112, 215, 48 },
		{ 112, 216, 48 },
		{ 112, 217, 48 },
		{ 112, 218, 48 },
		{ 112, 219, 48 },
		{ 112, 220, 48 },
		{ 112, 221, 48 },
		{ 112, 222, 48 },
		{ 112, 223, 48 },
		{ 112, 224, 48 },
		{ 112, 225, 48 },
		{ 112, 226, 48 },
		{ 112, 227, 48 },
		{ 112, 228, 48 },
		{ 112, 229, 48 },
		{ 112, 230, 48 },
		{ 112, 231, 48 },
		{ 112, 232, 48 },
		{ 112, 233, 48 },
		{ 112, 234, 48 },
		{ 112, 235, 48 },
		{ 112, 236, 48 },
		{ 112, 237, 48 },
		{ 112, 238, 48 },
		{ 113, 204, 48 },
		{ 113, 205, 48 },
		{ 113, 206, 48 },
		{ 113, 207, 48 },
		{ 113, 208, 48 },
		{ 113, 209, 48 },
		{ 113, 210, 48 },
		{ 113, 211, 48 },
		{ 113, 212, 48 },
		{ 113, 213, 48 },
		{ 113, 214, 48 },
		{ 113, 215, 48 },
		{ 113, 216, 48 },
		{ 113, 217, 48 },
		{ 113, 218, 48 },
		{ 113, 219, 48 },
		{ 113, 220, 48 },
		{ 113, 221, 48 },
		{ 113, 222, 48 },
		{ 113, 223, 48 },
		{ 113, 224, 48 },
		{ 113, 225, 48 },
		{ 113, 226, 48 },
		{ 113, 227, 48 },
		{ 113, 228, 48 },
		{ 113, 229, 48 },
		{ 113, 230, 48 },
		{ 113, 231, 48 },
		{ 113, 232, 48 },
		{ 113, 233, 48 },
		{ 113, 234, 48 },
		{ 113, 235, 48 },
		{ 113, 236, 48 },
		{ 113, 237, 48 },
		{ 113, 238, 48 },
		{ 114, 204, 48 },
		{ 114, 205, 48 },
		{ 114, 206, 48 },
		{ 114, 207, 48 },
		{ 114, 208, 48 },
		{ 114, 209, 48 },
		{ 114, 210, 48 },
		{ 114, 211, 48 },
		{ 114, 212, 48 },
		{ 114, 213, 48 },
		{ 114, 214, 48 },
		{ 114, 215, 48 },
		{ 114, 216, 48 },
		{ 114, 217, 48 },
		{ 114, 218, 48 },
		{ 114, 219, 48 },
		{ 114, 220, 48 },
		{ 114, 221, 48 },
		{ 114, 222, 48 },
		{ 114, 223, 48 },
		{ 114, 224, 48 },
		{ 114, 225, 48 },
		{ 114, 226, 48 },
		{ 114, 227, 48 },
		{ 114, 228, 48 },
		{ 114, 229, 48 },
		{ 114, 230, 48 },
		{ 114, 231, 48 },
		{ 114, 232, 48 },
		{ 114, 233, 48 },
		{ 114, 234, 48 },
		{ 114, 235, 48 },
		{ 114, 236, 48 },
		{ 114, 237, 48 },
		{ 114, 238, 48 },
		{ 114, 239, 48 },
		{ 115, 204, 48 },
		{ 115, 205, 48 },
		{ 115, 206, 48 },
		{ 115, 207, 48 },
		{ 115, 208, 48 },
		{ 115, 209, 48 },
		{ 115, 210, 48 },
		{ 115, 211, 48 },
		{ 115, 212, 48 },
		{ 115, 213, 48 },
		{ 115, 214, 48 },
		{ 115, 215, 48 },
		{ 115, 216, 48 },
		{ 115, 217, 48 },
		{ 115, 218, 48 },
		{ 115, 219, 48 },
		{ 115, 220, 48 },
		{ 115, 221, 48 },
		{ 115, 222, 48 },
		{ 115, 223, 48 },
		{ 115, 224, 48 },
		{ 115, 225, 48 },
		{ 115, 226, 48 },
		{ 115, 227, 48 },
		{ 115, 228, 48 },
		{ 115, 229, 48 },
		{ 115, 230, 48 },
		{ 115, 231, 48 },
		{ 115, 232, 48 },
		{ 115, 233, 48 },
		{ 115, 234, 48 },
		{ 115, 235, 48 },
		{ 115, 236, 48 },
		{ 115, 237, 48 },
		{ 115, 238, 48 },
		{ 115, 239, 48 },
		{ 116, 205, 48 },
		{ 116, 206, 48 },
		{ 116, 207, 48 },
		{ 116, 208, 48 },
		{ 116, 209, 48 },
		{ 116, 210, 48 },
		{ 116, 211, 48 },
		{ 116, 212, 48 },
		{ 116, 213, 48 },
		{ 116, 214, 48 },
		{ 116, 215, 48 },
		{ 116, 216, 48 },
		{ 116, 217, 48 },
		{ 116, 218, 48 },
		{ 116, 219, 48 },
		{ 116, 220, 48 },
		{ 116, 221, 48 },
		{ 116, 222, 48 },
		{ 116, 223, 48 },
		{ 116, 224, 48 },
		{ 116, 225, 48 },
		{ 116, 226, 48 },
		{ 116, 227, 48 },
		{ 116, 228, 48 },
		{ 116, 229, 48 },
		{ 116, 230, 48 },
		{ 116, 231, 48 },
		{ 116, 232, 48 },
		{ 116, 233, 48 },
		{ 116, 234, 48 },
		{ 116, 235, 48 },
		{ 116, 236, 48 },
		{ 116, 237, 48 },
		{ 116, 238, 48 },
		{ 116, 239, 48 },
		{ 117, 205, 48 },
		{ 117, 206, 48 },
		{ 117, 207, 48 },
		{ 117, 208, 48 },
		{ 117, 209, 48 },
		{ 117, 210, 48 },
		{ 117, 211, 48 },
		{ 117, 212, 48 },
		{ 117, 213, 48 },
		{ 117, 214, 48 },
		{ 117, 215, 48 },
		{ 117, 216, 48 },
		{ 117, 217, 48 },
		{ 117, 218, 48 },
		{ 117, 219, 48 },
		{ 117, 220, 48 },
		{ 117, 221, 48 },
		{ 117, 222, 48 },
		{ 117, 223, 48 },
		{ 117, 224, 48 },
		{ 117, 225, 48 },
		{ 117, 226, 48 },
		{ 117, 227, 48 },
		{ 117, 228, 48 },
		{ 117, 229, 48 },
		{ 117, 230, 48 },
		{ 117, 231, 48 },
		{ 117, 232, 48 },
		{ 117, 233, 48 },
		{ 117, 234, 48 },
		{ 117, 235, 48 },
		{ 117, 236, 48 },
		{ 117, 237, 48 },
		{ 117, 238, 48 },
		{ 117, 239, 48 },
		{ 118, 205, 48 },
		{ 118, 206, 48 },
		{ 118, 207, 48 },
		{ 118, 208, 48 },
		{ 118, 209, 48 },
		{ 118, 210, 48 },
		{ 118, 211, 48 },
		{ 118, 212, 48 },
		{ 118, 213, 48 },
		{ 118, 214, 48 },
		{ 118, 215, 48 },
		{ 118, 216, 48 },
		{ 118, 217, 48 },
		{ 118, 218, 48 },
		{ 118, 219, 48 },
		{ 118, 220, 48 },
		{ 118, 221, 48 },
		{ 118, 222, 48 },
		{ 118, 223, 48 },
		{ 118, 224, 48 },
		{ 118, 225, 48 },
		{ 118, 226, 48 },
		{ 118, 227, 48 },
		{ 118, 228, 48 },
		{ 118, 229, 48 },
		{ 118, 230, 48 },
		{ 118, 231, 48 },
		{ 118, 232, 48 },
		{ 118, 233, 48 },
		{ 118, 234, 48 },
		{ 118, 235, 48 },
		{ 118, 236, 48 },
		{ 118, 237, 48 },
		{ 118, 238, 48 },
		{ 118, 239, 48 },
		{ 119, 206, 48 },
		{ 119, 207, 48 },
		{ 119, 208, 48 },
		{ 119, 209, 48 },
		{ 119, 210, 48 },
		{ 119, 211, 48 },
		{ 119, 212, 48 },
		{ 119, 213, 48 },
		{ 119, 214, 48 },
		{ 119, 215, 48 },
		{ 119, 216, 48 },
		{ 119, 217, 48 },
		{ 119, 218, 48 },
		{ 119, 219, 48 },
		{ 119, 220, 48 },
		{ 119, 221, 48 },
		{ 119, 222, 48 },
		{ 119, 223, 48 },
		{ 119, 224, 48 },
		{ 119, 225, 48 },
		{ 119, 226, 48 },
		{ 119, 227, 48 },
		{ 119, 228, 48 },
		{ 119, 229, 48 },
		{ 119, 230, 48 },
		{ 119, 231, 48 },
		{ 119, 232, 48 },
		{ 119, 233, 48 },
		{ 119, 234, 48 },
		{ 119, 235, 48 },
		{ 119, 236, 48 },
		{ 119, 237, 48 },
		{ 119, 238, 48 },
		{ 119, 239, 48 },
		{ 120, 206, 48 },
		{ 120, 207, 48 },
		{ 120, 208, 48 },
		{ 120, 209, 48 },
		{ 120, 210, 48 },
		{ 120, 211, 48 },
		{ 120, 212, 48 },
		{ 120, 213, 48 },
		{ 120, 214, 48 },
		{ 120, 215, 48 },
		{ 120, 216, 48 },
		{ 120, 217, 48 },
		{ 120, 218, 48 },
		{ 120, 219, 48 },
		{ 120, 220, 48 },
		{ 120, 221, 48 },
		{ 120, 222, 48 },
		{ 120, 223, 48 },
		{ 120, 224, 48 },
		{ 120, 225, 48 },
		{ 120, 226, 48 },
		{ 120, 227, 48 },
		{ 120, 228, 48 },
		{ 120, 229, 48 },
		{ 120, 230, 48 },
		{ 120, 231, 48 },
		{ 120, 232, 48 },
		{ 120, 233, 48 },
		{ 120, 234, 48 },
		{ 120, 235, 48 },
		{ 120, 236, 48 },
		{ 120, 237, 48 },
		{ 120, 238, 48 },
		{ 120, 239, 48 },
		{ 121, 207, 48 },
		{ 121, 208, 48 },
		{ 121, 209, 48 },
		{ 121, 210, 48 },
		{ 121, 211, 48 },
		{ 121, 212, 48 },
		{ 121, 213, 48 },
		{ 121, 214, 48 },
		{ 121, 215, 48 },
		{ 121, 216, 48 },
		{ 121, 217, 48 },
		{ 121, 218, 48 },
		{ 121, 219, 48 },
		{ 121, 220, 48 },
		{ 121, 221, 48 },
		{ 121, 222, 48 },
		{ 121, 223, 48 },
		{ 121, 224, 48 },
		{ 121, 225, 48 },
		{ 121, 226, 48 },
		{ 121, 227, 48 },
		{ 121, 228, 48 },
		{ 121, 229, 48 },
		{ 121, 230, 48 },
		{ 121, 231, 48 },
		{ 121, 232, 48 },
		{ 121, 233, 48 },
		{ 121, 234, 48 },
		{ 121, 235, 48 },
		{ 121, 236, 48 },
		{ 121, 237, 48 },
		{ 121, 238, 48 },
		{ 121, 239, 48 },
		{ 122, 207, 48 },
		{ 122, 208, 48 },
		{ 122, 209, 48 },
		{ 122, 210, 48 },
		{ 122, 211, 48 },
		{ 122, 212, 48 },
		{ 122, 213, 48 },
		{ 122, 214, 48 },
		{ 122, 215, 48 },
		{ 122, 216, 48 },
		{ 122, 217, 48 },
		{ 122, 218, 48 },
		{ 122, 219, 48 },
		{ 122, 220, 48 },
		{ 122, 221, 48 },
		{ 122, 222, 48 },
		{ 122, 223, 48 },
		{ 122, 224, 48 },
		{ 122, 225, 48 },
		{ 122, 226, 48 },
		{ 122, 227, 48 },
		{ 122, 228, 48 },
		{ 122, 229, 48 },
		{ 122, 230, 48 },
		{ 122, 231, 48 },
		{ 122, 232, 48 },
		{ 122, 233, 48 },
		{ 122, 234, 48 },
		{ 122, 235, 48 },
		{ 122, 236, 48 },
		{ 122, 237, 48 },
		{ 122, 238, 48 },
		{ 122, 239, 48 },
		{ 122, 240, 48 },
		{ 123, 207, 48 },
		{ 123, 208, 48 },
		{ 123, 209, 48 },
		{ 123, 210, 48 },
		{ 123, 211, 48 },
		{ 123, 212, 48 },
		{ 123, 213, 48 },
		{ 123, 214, 48 },
		{ 123, 215, 48 },
		{ 123, 216, 48 },
		{ 123, 217, 48 },
		{ 123, 218, 48 },
		{ 123, 219, 48 },
		{ 123, 220, 48 },
		{ 123, 221, 48 },
		{ 123, 222, 48 },
		{ 123, 223, 48 },
		{ 123, 224, 48 },
		{ 123, 225, 48 },
		{ 123, 226, 48 },
		{ 123, 227, 48 },
		{ 123, 228, 48 },
		{ 123, 229, 48 },
		{ 123, 230, 48 },
		{ 123, 231, 48 },
		{ 123, 232, 48 },
		{ 123, 233, 48 },
		{ 123, 234, 48 },
		{ 123, 235, 48 },
		{ 123, 236, 48 },
		{ 123, 237, 48 },
		{ 123, 238, 48 },
		{ 123, 239, 48 },
		{ 123, 240, 48 },
		{ 124, 208, 48 },
		{ 124, 209, 48 },
		{ 124, 210, 48 },
		{ 124, 211, 48 },
		{ 124, 212, 48 },
		{ 124, 213, 48 },
		{ 124, 214, 48 },
		{ 124, 215, 48 },
		{ 124, 216, 48 },
		{ 124, 217, 48 },
		{ 124, 218, 48 },
		{ 124, 219, 48 },
		{ 124, 220, 48 },
		{ 124, 221, 48 },
		{ 124, 222, 48 },
		{ 124, 223, 48 },
		{ 124, 224, 48 },
		{ 124, 225, 48 },
		{ 124, 226, 48 },
		{ 124, 227, 48 },
		{ 124, 228, 48 },
		{ 124, 229, 48 },
		{ 124, 230, 48 },
		{ 124, 231, 48 },
		{ 124, 232, 48 },
		{ 124, 233, 48 },
		{ 124, 234, 48 },
		{ 124, 235, 48 },
		{ 124, 236, 48 },
		{ 124, 237, 48 },
		{ 124, 238, 48 },
		{ 124, 239, 48 },
		{ 124, 240, 48 },
		{ 125, 208, 48 },
		{ 125, 209, 48 },
		{ 125, 210, 48 },
		{ 125, 211, 48 },
		{ 125, 212, 48 },
		{ 125, 213, 48 },
		{ 125, 214, 48 },
		{ 125, 215, 48 },
		{ 125, 216, 48 },
		{ 125, 217, 48 },
		{ 125, 218, 48 },
		{ 125, 219, 48 },
		{ 125, 220, 48 },
		{ 125, 221, 48 },
		{ 125, 222, 48 },
		{ 125, 223, 48 },
		{ 125, 224, 48 },
		{ 125, 225, 48 },
		{ 125, 226, 48 },
		{ 125, 227, 48 },
		{ 125, 228, 48 },
		{ 125, 229, 48 },
		{ 125, 230, 48 },
		{ 125, 231, 48 },
		{ 125, 232, 48 },
		{ 125, 233, 48 },
		{ 125, 234, 48 },
		{ 125, 235, 48 },
		{ 125, 236, 48 },
		{ 125, 237, 48 },
		{ 125, 238, 48 },
		{ 125, 239, 48 },
		{ 125, 240, 48 },
		{ 126, 208, 48 },
		{ 126, 209, 48 },
		{ 126, 210, 48 },
		{ 126, 211, 48 },
		{ 126, 212, 48 },
		{ 126, 213, 48 },
		{ 126, 214, 48 },
		{ 126, 215, 48 },
		{ 126, 216, 48 },
		{ 126, 217, 48 },
		{ 126, 218, 48 },
		{ 126, 219, 48 },
		{ 126, 220, 48 },
		{ 126, 221, 48 },
		{ 126, 222, 48 },
		{ 126, 223, 48 },
		{ 126, 224, 48 },
		{ 126, 225, 48 },
		{ 126, 226, 48 },
		{ 126, 227, 48 },
		{ 126, 228, 48 },
		{ 126, 229, 48 },
		{ 126, 230, 48 },
		{ 126, 231, 48 },
		{ 126, 232, 48 },
		{ 126, 233, 48 },
		{ 126, 234, 48 },
		{ 126, 235, 48 },
		{ 126, 236, 48 },
		{ 126, 237, 48 },
		{ 126, 238, 48 },
		{ 126, 239, 48 },
		{ 126, 240, 48 },
		{ 127, 209, 48 },
		{ 127, 210, 48 },
		{ 127, 211, 48 },
		{ 127, 212, 48 },
		{ 127, 213, 48 },
		{ 127, 214, 48 },
		{ 127, 215, 48 },
		{ 127, 216, 48 },
		{ 127, 217, 48 },
		{ 127, 218, 48 },
		{ 127, 219, 48 },
		{ 127, 220, 48 },
		{ 127, 221, 48 },
		{ 127, 222, 48 },
		{ 127, 223, 48 },
		{ 127, 224, 48 },
		{ 127, 225, 48 },
		{ 127, 226, 48 },
		{ 127, 227, 48 },
		{ 127, 228, 48 },
		{ 127, 229, 48 },
		{ 127, 230, 48 },
		{ 127, 231, 48 },
		{ 127, 232, 48 },
		{ 127, 233, 48 },
		{ 127, 234, 48 },
		{ 127, 235, 48 },
		{ 127, 236, 48 },
		{ 127, 237, 48 },
		{ 127, 238, 48 },
		{ 127, 239, 48 },
		{ 127, 240, 48 },
		{ 128, 209, 48 },
		{ 128, 210, 48 },
		{ 128, 211, 48 },
		{ 128, 212, 48 },
		{ 128, 213, 48 },
		{ 128, 214, 48 },
		{ 128, 215, 48 },
		{ 128, 216, 48 },
		{ 128, 217, 48 },
		{ 128, 218, 48 },
		{ 128, 219, 48 },
		{ 128, 220, 48 },
		{ 128, 221, 48 },
		{ 128, 222, 48 },
		{ 128, 223, 48 },
		{ 128, 224, 48 },
		{ 128, 225, 48 },
		{ 128, 226, 48 },
		{ 128, 227, 48 },
		{ 128, 228, 48 },
		{ 128, 229, 48 },
		{ 128, 230, 48 },
		{ 128, 231, 48 },
		{ 128, 232, 48 },
		{ 128, 233, 48 },
		{ 128, 234, 48 },
		{ 128, 235, 48 },
		{ 128, 236, 48 },
		{ 128, 237, 48 },
		{ 128, 238, 48 },
		{ 128, 239, 48 },
		{ 128, 240, 48 },
		{ 129, 210, 48 },
		{ 129, 211, 48 },
		{ 129, 212, 48 },
		{ 129, 213, 48 },
		{ 129, 214, 48 },
		{ 129, 215, 48 },
		{ 129, 216, 48 },
		{ 129, 217, 48 },
		{ 129, 218, 48 },
		{ 129, 219, 48 },
		{ 129, 220, 48 },
		{ 129, 221, 48 },
		{ 129, 222, 48 },
		{ 129, 223, 48 },
		{ 129, 224, 48 },
		{ 129, 225, 48 },
		{ 129, 226, 48 },
		{ 129, 227, 48 },
		{ 129, 228, 48 },
		{ 129, 229, 48 },
		{ 129, 230, 48 },
		{ 129, 231, 48 },
		{ 129, 232, 48 },
		{ 129, 233, 48 },
		{ 129, 234, 48 },
		{ 129, 235, 48 },
		{ 129, 236, 48 },
		{ 129, 237, 48 },
		{ 129, 238, 48 },
		{ 129, 239, 48 },
		{ 129, 240, 48 },
		{ 130, 210, 48 },
		{ 130, 211, 48 },
		{ 130, 212, 48 },
		{ 130, 213, 48 },
		{ 130, 214, 48 },
		{ 130, 215, 48 },
		{ 130, 216, 48 },
		{ 130, 217, 48 },
		{ 130, 218, 48 },
		{ 130, 219, 48 },
		{ 130, 220, 48 },
		{ 130, 221, 48 },
		{ 130, 222, 48 },
		{ 130, 223, 48 },
		{ 130, 224, 48 },
		{ 130, 225, 48 },
		{ 130, 226, 48 },
		{ 130, 227, 48 },
		{ 130, 228, 48 },
		{ 130, 229, 48 },
		{ 130, 230, 48 },
		{ 130, 231, 48 },
		{ 130, 232, 48 },
		{ 130, 233, 48 },
		{ 130, 234, 48 },
		{ 130, 235, 48 },
		{ 130, 236, 48 },
		{ 130, 237, 48 },
		{ 130, 238, 48 },
		{ 130, 239, 48 },
		{ 130, 240, 48 },
		{ 130, 241, 48 },
		{ 131, 210, 48 },
		{ 131, 211, 48 },
		{ 131, 212, 48 },
		{ 131, 213, 48 },
		{ 131, 214, 48 },
		{ 131, 215, 48 },
		{ 131, 216, 48 },
		{ 131, 217, 48 },
		{ 131, 218, 48 },
		{ 131, 219, 48 },
		{ 131, 220, 48 },
		{ 131, 221, 48 },
		{ 131, 222, 48 },
		{ 131, 223, 48 },
		{ 131, 224, 48 },
		{ 131, 225, 48 },
		{ 131, 226, 48 },
		{ 131, 227, 48 },
		{ 131, 228, 48 },
		{ 131, 229, 48 },
		{ 131, 230, 48 },
		{ 131, 231, 48 },
		{ 131, 232, 48 },
		{ 131, 233, 48 },
		{ 131, 234, 48 },
		{ 131, 235, 48 },
		{ 131, 236, 48 },
		{ 131, 237, 48 },
		{ 131, 238, 48 },
		{ 131, 239, 48 },
		{ 131, 240, 48 },
		{ 131, 241, 48 },
		{ 132, 211, 48 },
		{ 132, 212, 48 },
		{ 132, 213, 48 },
		{ 132, 214, 48 },
		{ 132, 215, 48 },
		{ 132, 216, 48 },
		{ 132, 217, 48 },
		{ 132, 218, 48 },
		{ 132, 219, 48 },
		{ 132, 220, 48 },
		{ 132, 221, 48 },
		{ 132, 222, 48 },
		{ 132, 223, 48 },
		{ 132, 224, 48 },
		{ 132, 225, 48 },
		{ 132, 226, 48 },
		{ 132, 227, 48 },
		{ 132, 228, 48 },
		{ 132, 229, 48 },
		{ 132, 230, 48 },
		{ 132, 231, 48 },
		{ 132, 232, 48 },
		{ 132, 233, 48 },
		{ 132, 234, 48 },
		{ 132, 235, 48 },
		{ 132, 236, 48 },
		{ 132, 237, 48 },
		{ 132, 238, 48 },
		{ 132, 239, 48 },
		{ 132, 240, 48 },
		{ 132, 241, 48 },
		{ 133, 211, 48 },
		{ 133, 212, 48 },
		{ 133, 213, 48 },
		{ 133, 214, 48 },
		{ 133, 215, 48 },
		{ 133, 216, 48 },
		{ 133, 217, 48 },
		{ 133, 218, 48 },
		{ 133, 219, 48 },
		{ 133, 220, 48 },
		{ 133, 221, 48 },
		{ 133, 222, 48 },
		{ 133, 223, 48 },
		{ 133, 224, 48 },
		{ 133, 225, 48 },
		{ 133, 226, 48 },
		{ 133, 227, 48 },
		{ 133, 228, 48 },
		{ 133, 229, 48 },
		{ 133, 230, 48 },
		{ 133, 231, 48 },
		{ 133, 232, 48 },
		{ 133, 233, 48 },
		{ 133, 234, 48 },
		{ 133, 235, 48 },
		{ 133, 236, 48 },
		{ 133, 237, 48 },
		{ 133, 238, 48 },
		{ 133, 239, 48 },
		{ 133, 240, 48 },
		{ 133, 241, 48 },
		{ 134, 211, 48 },
		{ 134, 212, 48 },
		{ 134, 213, 48 },
		{ 134, 214, 48 },
		{ 134, 215, 48 },
		{ 134, 216, 48 },
		{ 134, 217, 48 },
		{ 134, 218, 48 },
		{ 134, 219, 48 },
		{ 134, 220, 48 },
		{ 134, 221, 48 },
		{ 134, 222, 48 },
		{ 134, 223, 48 },
		{ 134, 224, 48 },
		{ 134, 225, 48 },
		{ 134, 226, 48 },
		{ 134, 227, 48 },
		{ 134, 228, 48 },
		{ 134, 229, 48 },
		{ 134, 230, 48 },
		{ 134, 231, 48 },
		{ 134, 232, 48 },
		{ 134, 233, 48 },
		{ 134, 234, 48 },
		{ 134, 235, 48 },
		{ 134, 236, 48 },
		{ 134, 237, 48 },
		{ 134, 238, 48 },
		{ 134, 239, 48 },
		{ 134, 240, 48 },
		{ 134, 241, 48 },
		{ 135, 212, 48 },
		{ 135, 213, 48 },
		{ 135, 214, 48 },
		{ 135, 215, 48 },
		{ 135, 216, 48 },
		{ 135, 217, 48 },
		{ 135, 218, 48 },
		{ 135, 219, 48 },
		{ 135, 220, 48 },
		{ 135, 221, 48 },
		{ 135, 222, 48 },
		{ 135, 223, 48 },
		{ 135, 224, 48 },
		{ 135, 225, 48 },
		{ 135, 226, 48 },
		{ 135, 227, 48 },
		{ 135, 228, 48 },
		{ 135, 229, 48 },
		{ 135, 230, 48 },
		{ 135, 231, 48 },
		{ 135, 232, 48 },
		{ 135, 233, 48 },
		{ 135, 234, 48 },
		{ 135, 235, 48 },
		{ 135, 236, 48 },
		{ 135, 237, 48 },
		{ 135, 238, 48 },
		{ 135, 239, 48 },
		{ 135, 240, 48 },
		{ 135, 241, 48 },
		{ 136, 212, 48 },
		{ 136, 213, 48 },
		{ 136, 214, 48 },
		{ 136, 215, 48 },
		{ 136, 216, 48 },
		{ 136, 217, 48 },
		{ 136, 218, 48 },
		{ 136, 219, 48 },
		{ 136, 220, 48 },
		{ 136, 221, 48 },
		{ 136, 222, 48 },
		{ 136, 223, 48 },
		{ 136, 224, 48 },
		{ 136, 225, 48 },
		{ 136, 226, 48 },
		{ 136, 227, 48 },
		{ 136, 228, 48 },
		{ 136, 229, 48 },
		{ 136, 230, 48 },
		{ 136, 231, 48 },
		{ 136, 232, 48 },
		{ 136, 233, 48 },
		{ 136, 234, 48 },
		{ 136, 235, 48 },
		{ 136, 236, 48 },
		{ 136, 237, 48 },
		{ 136, 238, 48 },
		{ 136, 239, 48 },
		{ 136, 240, 48 },
		{ 136, 241, 48 },
		{ 137, 213, 48 },
		{ 137, 214, 48 },
		{ 137, 215, 48 },
		{ 137, 216, 48 },
		{ 137, 217, 48 },
		{ 137, 218, 48 },
		{ 137, 219, 48 },
		{ 137, 220, 48 },
		{ 137, 221, 48 },
		{ 137, 222, 48 },
		{ 137, 223, 48 },
		{ 137, 224, 48 },
		{ 137, 225, 48 },
		{ 137, 226, 48 },
		{ 137, 227, 48 },
		{ 137, 228, 48 },
		{ 137, 229, 48 },
		{ 137, 230, 48 },
		{ 137, 231, 48 },
		{ 137, 232, 48 },
		{ 137, 233, 48 },
		{ 137, 234, 48 },
		{ 137, 235, 48 },
		{ 137, 236, 48 },
		{ 137, 237, 48 },
		{ 137, 238, 48 },
		{ 137, 239, 48 },
		{ 137, 240, 48 },
		{ 137, 241, 48 },
		{ 138, 213, 48 },
		{ 138, 214, 48 },
		{ 138, 215, 48 },
		{ 138, 216, 48 },
		{ 138, 217, 48 },
		{ 138, 218, 48 },
		{ 138, 219, 48 },
		{ 138, 220, 48 },
		{ 138, 221, 48 },
		{ 138, 222, 48 },
		{ 138, 223, 48 },
		{ 138, 224, 48 },
		{ 138, 225, 48 },
		{ 138, 226, 48 },
		{ 138, 227, 48 },
		{ 138, 228, 48 },
		{ 138, 229, 48 },
		{ 138, 230, 48 },
		{ 138, 231, 48 },
		{ 138, 232, 48 },
		{ 138, 233, 48 },
		{ 138, 234, 48 },
		{ 138, 235, 48 },
		{ 138, 236, 48 },
		{ 138, 237, 48 },
		{ 138, 238, 48 },
		{ 138, 239, 48 },
		{ 138, 240, 48 },
		{ 138, 241, 48 },
		{ 138, 242, 48 },
		{ 139, 213, 48 },
		{ 139, 214, 48 },
		{ 139, 215, 48 },
		{ 139, 216, 48 },
		{ 139, 217, 48 },
		{ 139, 218, 48 },
		{ 139, 219, 48 },
		{ 139, 220, 48 },
		{ 139, 221, 48 },
		{ 139, 222, 48 },
		{ 139, 223, 48 },
		{ 139, 224, 48 },
		{ 139, 225, 48 },
		{ 139, 226, 48 },
		{ 139, 227, 48 },
		{ 139, 228, 48 },
		{ 139, 229, 48 },
		{ 139, 230, 48 },
		{ 139, 231, 48 },
		{ 139, 232, 48 },
		{ 139, 233, 48 },
		{ 139, 234, 48 },
		{ 139, 235, 48 },
		{ 139, 236, 48 },
		{ 139, 237, 48 },
		{ 139, 238, 48 },
		{ 139, 239, 48 },
		{ 139, 240, 48 },
		{ 139, 241, 48 },
		{ 139, 242, 48 },
		{ 140, 214, 48 },
		{ 140, 215, 48 },
		{ 140, 216, 48 },
		{ 140, 217, 48 },
		{ 140, 218, 48 },
		{ 140, 219, 48 },
		{ 140, 220, 48 },
		{ 140, 221, 48 },
		{ 140, 222, 48 },
		{ 140, 223, 48 },
		{ 140, 224, 48 },
		{ 140, 225, 48 },
		{ 140, 226, 48 },
		{ 140, 227, 48 },
		{ 140, 228, 48 },
		{ 140, 229, 48 },
		{ 140, 230, 48 },
		{ 140, 231, 48 },
		{ 140, 232, 48 },
		{ 140, 233, 48 },
		{ 140, 234, 48 },
		{ 140, 235, 48 },
		{ 140, 236, 48 },
		{ 140, 237, 48 },
		{ 140, 238, 48 },
		{ 140, 239, 48 },
		{ 140, 240, 48 },
		{ 140, 241, 48 },
		{ 140, 242, 48 },
		{ 141, 214, 48 },
		{ 141, 215, 48 },
		{ 141, 216, 48 },
		{ 141, 217, 48 },
		{ 141, 218, 48 },
		{ 141, 219, 48 },
		{ 141, 220, 48 },
		{ 141, 221, 48 },
		{ 141, 222, 48 },
		{ 141, 223, 48 },
		{ 141, 224, 48 },
		{ 141, 225, 48 },
		{ 141, 226, 48 },
		{ 141, 227, 48 },
		{ 141, 228, 48 },
		{ 141, 229, 48 },
		{ 141, 230, 48 },
		{ 141, 231, 48 },
		{ 141, 232, 48 },
		{ 141, 233, 48 },
		{ 141, 234, 48 },
		{ 141, 235, 48 },
		{ 141, 236, 48 },
		{ 141, 237, 48 },
		{ 141, 238, 48 },
		{ 141, 239, 48 },
		{ 141, 240, 48 },
		{ 141, 241, 48 },
		{ 141, 242, 48 },
		{ 142, 214, 48 },
		{ 142, 215, 48 },
		{ 142, 216, 48 },
		{ 142, 217, 48 },
		{ 142, 218, 48 },
		{ 142, 219, 48 },
		{ 142, 220, 48 },
		{ 142, 221, 48 },
		{ 142, 222, 48 },
		{ 142, 223, 48 },
		{ 142, 224, 48 },
		{ 142, 225, 48 },
		{ 142, 226, 48 },
		{ 142, 227, 48 },
		{ 142, 228, 48 },
		{ 142, 229, 48 },
		{ 142, 230, 48 },
		{ 142, 231, 48 },
		{ 142, 232, 48 },
		{ 142, 233, 48 },
		{ 142, 234, 48 },
		{ 142, 235, 48 },
		{ 142, 236, 48 },
		{ 142, 237, 48 },
		{ 142, 238, 48 },
		{ 142, 239, 48 },
		{ 142, 240, 48 },
		{ 142, 241, 48 },
		{ 142, 242, 48 },
		{ 143, 215, 48 },
		{ 143, 216, 48 },
		{ 143, 217, 48 },
		{ 143, 218, 48 },
		{ 143, 219, 48 },
		{ 143, 220, 48 },
		{ 143, 221, 48 },
		{ 143, 222, 48 },
		{ 143, 223, 48 },
		{ 143, 224, 48 },
		{ 143, 225, 48 },
		{ 143, 226, 48 },
		{ 143, 227, 48 },
		{ 143, 228, 48 },
		{ 143, 229, 48 },
		{ 143, 230, 48 },
		{ 143, 231, 48 },
		{ 143, 232, 48 },
		{ 143, 233, 48 },
		{ 143, 234, 48 },
		{ 143, 235, 48 },
		{ 143, 236, 48 },
		{ 143, 237, 48 },
		{ 143, 238, 48 },
		{ 143, 239, 48 },
		{ 143, 240, 48 },
		{ 143, 241, 48 },
		{ 143, 242, 48 },
		{ 144, 215, 48 },
		{ 144, 216, 48 },
		{ 144, 217, 48 },
		{ 144, 218, 48 },
		{ 144, 219, 48 },
		{ 144, 220, 48 },
		{ 144, 221, 48 },
		{ 144, 222, 48 },
		{ 144, 223, 48 },
		{ 144, 224, 48 },
		{ 144, 225, 48 },
		{ 144, 226, 48 },
		{ 144, 227, 48 },
		{ 144, 228, 48 },
		{ 144, 229, 48 },
		{ 144, 230, 48 },
		{ 144, 231, 48 },
		{ 144, 232, 48 },
		{ 144, 233, 48 },
		{ 144, 234, 48 },
		{ 144, 235, 48 },
		{ 144, 236, 48 },
		{ 144, 237, 48 },
		{ 144, 238, 48 },
		{ 144, 239, 48 },
		{ 144, 240, 48 },
		{ 144, 241, 48 },
		{ 144, 242, 48 },
		{ 145, 216, 48 },
		{ 145, 217, 48 },
		{ 145, 218, 48 },
		{ 145, 219, 48 },
		{ 145, 220, 48 },
		{ 145, 221, 48 },
		{ 145, 222, 48 },
		{ 145, 223, 48 },
		{ 145, 224, 48 },
		{ 145, 225, 48 },
		{ 145, 226, 48 },
		{ 145, 227, 48 },
		{ 145, 228, 48 },
		{ 145, 229, 48 },
		{ 145, 230, 48 },
		{ 145, 231, 48 },
		{ 145, 232, 48 },
		{ 145, 233, 48 },
		{ 145, 234, 48 },
		{ 145, 235, 48 },
		{ 145, 236, 48 },
		{ 145, 237, 48 },
		{ 145, 238, 48 },
		{ 145, 239, 48 },
		{ 145, 240, 48 },
		{ 145, 241, 48 },
		{ 145, 242, 48 },
		{ 146, 216, 48 },
		{ 146, 217, 48 },
		{ 146, 218, 48 },
		{ 146, 219, 48 },
		{ 146, 220, 48 },
		{ 146, 221, 48 },
		{ 146, 222, 48 },
		{ 146, 223, 48 },
		{ 146, 224, 48 },
		{ 146, 225, 48 },
		{ 146, 226, 48 },
		{ 146, 227, 48 },
		{ 146, 228, 48 },
		{ 146, 229, 48 },
		{ 146, 230, 48 },
		{ 146, 231, 48 },
		{ 146, 232, 48 },
		{ 146, 233, 48 },
		{ 146, 234, 48 },
		{ 146, 235, 48 },
		{ 146, 236, 48 },
		{ 146, 237, 48 },
		{ 146, 238, 48 },
		{ 146, 239, 48 },
		{ 146, 240, 48 },
		{ 146, 241, 48 },
		{ 146, 242, 48 },
		{ 146, 243, 48 },
		{ 147, 216, 48 },
		{ 147, 217, 48 },
		{ 147, 218, 48 },
		{ 147, 219, 48 },
		{ 147, 220, 48 },
		{ 147, 221, 48 },
		{ 147, 222, 48 },
		{ 147, 223, 48 },
		{ 147, 224, 48 },
		{ 147, 225, 48 },
		{ 147, 226, 48 },
		{ 147, 227, 48 },
		{ 147, 228, 48 },
		{ 147, 229, 48 },
		{ 147, 230, 48 },
		{ 147, 231, 48 },
		{ 147, 232, 48 },
		{ 147, 233, 48 },
		{ 147, 234, 48 },
		{ 147, 235, 48 },
		{ 147, 236, 48 },
		{ 147, 237, 48 },
		{ 147, 238, 48 },
		{ 147, 239, 48 },
		{ 147, 240, 48 },
		{ 147, 241, 48 },
		{ 147, 242, 48 },
		{ 147, 243, 48 },
		{ 148, 217, 48 },
		{ 148, 218, 48 },
		{ 148, 219, 48 },
		{ 148, 220, 48 },
		{ 148, 221, 48 },
		{ 148, 222, 48 },
		{ 148, 223, 48 },
		{ 148, 224, 48 },
		{ 148, 225, 48 },
		{ 148, 226, 48 },
		{ 148, 227, 48 },
		{ 148, 228, 48 },
		{ 148, 229, 48 },
		{ 148, 230, 48 },
		{ 148, 231, 48 },
		{ 148, 232, 48 },
		{ 148, 233, 48 },
		{ 148, 234, 48 },
		{ 148, 235, 48 },
		{ 148, 236, 48 },
		{ 148, 237, 48 },
		{ 148, 238, 48 },
		{ 148, 239, 48 },
		{ 148, 240, 48 },
		{ 148, 241, 48 },
		{ 148, 242, 48 },
		{ 148, 243, 48 },
		{ 149, 217, 48 },
		{ 149, 218, 48 },
		{ 149, 219, 48 },
		{ 149, 220, 48 },
		{ 149, 221, 48 },
		{ 149, 222, 48 },
		{ 149, 223, 48 },
		{ 149, 224, 48 },
		{ 149, 225, 48 },
		{ 149, 226, 48 },
		{ 149, 227, 48 },
		{ 149, 228, 48 },
		{ 149, 229, 48 },
		{ 149, 230, 48 },
		{ 149, 231, 48 },
		{ 149, 232, 48 },
		{ 149, 233, 48 },
		{ 149, 234, 48 },
		{ 149, 235, 48 },
		{ 149, 236, 48 },
		{ 149, 237, 48 },
		{ 149, 238, 48 },
		{ 149, 239, 48 },
		{ 149, 240, 48 },
		{ 149, 241, 48 },
		{ 149, 242, 48 },
		{ 149, 243, 48 },
		{ 150, 217, 48 },
		{ 150, 218, 48 },
		{ 150, 219, 48 },
		{ 150, 220, 48 },
		{ 150, 221, 48 },
		{ 150, 222, 48 },
		{ 150, 223, 48 },
		{ 150, 224, 48 },
		{ 150, 225, 48 },
		{ 150, 226, 48 },
		{ 150, 227, 48 },
		{ 150, 228, 48 },
		{ 150, 229, 48 },
		{ 150, 230, 48 },
		{ 150, 231, 48 },
		{ 150, 232, 48 },
		{ 150, 233, 48 },
		{ 150, 234, 48 },
		{ 150, 235, 48 },
		{ 150, 236, 48 },
		{ 150, 237, 48 },
		{ 150, 238, 48 },
		{ 150, 239, 48 },
		{ 150, 240, 48 },
		{ 150, 241, 48 },
		{ 150, 242, 48 },
		{ 150, 243, 48 },
		{ 151, 218, 48 },
		{ 151, 219, 48 },
		{ 151, 220, 48 },
		{ 151, 221, 48 },
		{ 151, 222, 48 },
		{ 151, 223, 48 },
		{ 151, 224, 48 },
		{ 151, 225, 48 },
		{ 151, 226, 48 },
		{ 151, 227, 48 },
		{ 151, 228, 48 },
		{ 151, 229, 48 },
		{ 151, 230, 48 },
		{ 151, 231, 48 },
		{ 151, 232, 48 },
		{ 151, 233, 48 },
		{ 151, 234, 48 },
		{ 151, 235, 48 },
		{ 151, 236, 48 },
		{ 151, 237, 48 },
		{ 151, 238, 48 },
		{ 151, 239, 48 },
		{ 151, 240, 48 },
		{ 151, 241, 48 },
		{ 151, 242, 48 },
		{ 151, 243, 48 },
		{ 152, 218, 48 },
		{ 152, 219, 48 },
		{ 152, 220, 48 },
		{ 152, 221, 48 },
		{ 152, 222, 48 },
		{ 152, 223, 48 },
		{ 152, 224, 48 },
		{ 152, 225, 48 },
		{ 152, 226, 48 },
		{ 152, 227, 48 },
		{ 152, 228, 48 },
		{ 152, 229, 48 },
		{ 152, 230, 48 },
		{ 152, 231, 48 },
		{ 152, 232, 48 },
		{ 152, 233, 48 },
		{ 152, 234, 48 },
		{ 152, 235, 48 },
		{ 152, 236, 48 },
		{ 152, 237, 48 },
		{ 152, 238, 48 },
		{ 152, 239, 48 },
		{ 152, 240, 48 },
		{ 152, 241, 48 },
		{ 152, 242, 48 },
		{ 152, 243, 48 },
		{ 153, 219, 48 },
		{ 153, 220, 48 },
		{ 153, 221, 48 },
		{ 153, 222, 48 },
		{ 153, 223, 48 },
		{ 153, 224, 48 },
		{ 153, 225, 48 },
		{ 153, 226, 48 },
		{ 153, 227, 48 },
		{ 153, 228, 48 },
		{ 153, 229, 48 },
		{ 153, 230, 48 },
		{ 153, 231, 48 },
		{ 153, 232, 48 },
		{ 153, 233, 48 },
		{ 153, 234, 48 },
		{ 153, 235, 48 },
		{ 153, 236, 48 },
		{ 153, 237, 48 },
		{ 153, 238, 48 },
		{ 153, 239, 48 },
		{ 153, 240, 48 },
		{ 153, 241, 48 },
		{ 153, 242, 48 },
		{ 153, 243, 48 },
		{ 154, 219, 48 },
		{ 154, 220, 48 },
		{ 154, 221, 48 },
		{ 154, 222, 48 },
		{ 154, 223, 48 },
		{ 154, 224, 48 },
		{ 154, 225, 48 },
		{ 154, 226, 48 },
		{ 154, 227, 48 },
		{ 154, 228, 48 },
		{ 154, 229, 48 },
		{ 154, 230, 48 },
		{ 154, 231, 48 },
		{ 154, 232, 48 },
		{ 154, 233, 48 },
		{ 154, 234, 48 },
		{ 154, 235, 48 },
		{ 154, 236, 48 },
		{ 154, 237, 48 },
		{ 154, 238, 48 },
		{ 154, 239, 48 },
		{ 154, 240, 48 },
		{ 154, 241, 48 },
		{ 154, 242, 48 },
		{ 154, 243, 48 },
		{ 154, 244, 48 },
		{ 155, 219, 48 },
		{ 155, 220, 48 },
		{ 155, 221, 48 },
		{ 155, 222, 48 },
		{ 155, 223, 48 },
		{ 155, 224, 48 },
		{ 155, 225, 48 },
		{ 155, 226, 48 },
		{ 155, 227, 48 },
		{ 155, 228, 48 },
		{ 155, 229, 48 },
		{ 155, 230, 48 },
		{ 155, 231, 48 },
		{ 155, 232, 48 },
		{ 155, 233, 48 },
		{ 155, 234, 48 },
		{ 155, 235, 48 },
		{ 155, 236, 48 },
		{ 155, 237, 48 },
		{ 155, 238, 48 },
		{ 155, 239, 48 },
		{ 155, 240, 48 },
		{ 155, 241, 48 },
		{ 155, 242, 48 },
		{ 155, 243, 48 },
		{ 155, 244, 48 },
		{ 156, 220, 48 },
		{ 156, 221, 48 },
		{ 156, 222, 48 },
		{ 156, 223, 48 },
		{ 156, 224, 48 },
		{ 156, 225, 48 },
		{ 156, 226, 48 },
		{ 156, 227, 48 },
		{ 156, 228, 48 },
		{ 156, 229, 48 },
		{ 156, 230, 48 },
		{ 156, 231, 48 },
		{ 156, 232, 48 },
		{ 156, 233, 48 },
		{ 156, 234, 48 },
		{ 156, 235, 48 },
		{ 156, 236, 48 },
		{ 156, 237, 48 },
		{ 156, 238, 48 },
		{ 156, 239, 48 },
		{ 156, 240, 48 },
		{ 156, 241, 48 },
		{ 156, 242, 48 },
		{ 156, 243, 48 },
		{ 156, 244, 48 },
		{ 157, 220, 48 },
		{ 157, 221, 48 },
		{ 157, 222, 48 },
		{ 157, 223, 48 },
		{ 157, 224, 48 },
		{ 157, 225, 48 },
		{ 157, 226, 48 },
		{ 157, 227, 48 },
		{ 157, 228, 48 },
		{ 157, 229, 48 },
		{ 157, 230, 48 },
		{ 157, 231, 48 },
		{ 157, 232, 48 },
		{ 157, 233, 48 },
		{ 157, 234, 48 },
		{ 157, 235, 48 },
		{ 157, 236, 48 },
		{ 157, 237, 48 },
		{ 157, 238, 48 },
		{ 157, 239, 48 },
		{ 157, 240, 48 },
		{ 157, 241, 48 },
		{ 157, 242, 48 },
		{ 157, 243, 48 },
		{ 157, 244, 48 },
		{ 158, 220, 48 },
		{ 158, 221, 48 },
		{ 158, 222, 48 },
		{ 158, 223, 48 },
		{ 158, 224, 48 },
		{ 158, 225, 48 },
		{ 158, 226, 48 },
		{ 158, 227, 48 },
		{ 158, 228, 48 },
		{ 158, 229, 48 },
		{ 158, 230, 48 },
		{ 158, 231, 48 },
		{ 158, 232, 48 },
		{ 158, 233, 48 },
		{ 158, 234, 48 },
		{ 158, 235, 48 },
		{ 158, 236, 48 },
		{ 158, 237, 48 },
		{ 158, 238, 48 },
		{ 158, 239, 48 },
		{ 158, 240, 48 },
		{ 158, 241, 48 },
		{ 158, 242, 48 },
		{ 158, 243, 48 },
		{ 158, 244, 48 },
		{ 159, 221, 48 },
		{ 159, 222, 48 },
		{ 159, 223, 48 },
		{ 159, 224, 48 },
		{ 159, 225, 48 },
		{ 159, 226, 48 },
		{ 159, 227, 48 },
		{ 159, 228, 48 },
		{ 159, 229, 48 },
		{ 159, 230, 48 },
		{ 159, 231, 48 },
		{ 159, 232, 48 },
		{ 159, 233, 48 },
		{ 159, 234, 48 },
		{ 159, 235, 48 },
		{ 159, 236, 48 },
		{ 159, 237, 48 },
		{ 159, 238, 48 },
		{ 159, 239, 48 },
		{ 159, 240, 48 },
		{ 159, 241, 48 },
		{ 159, 242, 48 },
		{ 159, 243, 48 },
		{ 159, 244, 48 },
		{ 160, 221, 48 },
		{ 160, 222, 48 },
		{ 160, 223, 48 },
		{ 160, 224, 48 },
		{ 160, 225, 48 },
		{ 160, 226, 48 },
		{ 160, 227, 48 },
		{ 160, 228, 48 },
		{ 160, 229, 48 },
		{ 160, 230, 48 },
		{ 160, 231, 48 },
		{ 160, 232, 48 },
		{ 160, 233, 48 },
		{ 160, 234, 48 },
		{ 160, 235, 48 },
		{ 160, 236, 48 },
		{ 160, 237, 48 },
		{ 160, 238, 48 },
		{ 160, 239, 48 },
		{ 160, 240, 48 },
		{ 160, 241, 48 },
		{ 160, 242, 48 },
		{ 160, 243, 48 },
		{ 160, 244, 48 },
		{ 161, 222, 48 },
		{ 161, 223, 48 },
		{ 161, 224, 48 },
		{ 161, 225, 48 },
		{ 161, 226, 48 },
		{ 161, 227, 48 },
		{ 161, 228, 48 },
		{ 161, 229, 48 },
		{ 161, 230, 48 },
		{ 161, 231, 48 },
		{ 161, 232, 48 },
		{ 161, 233, 48 },
		{ 161, 234, 48 },
		{ 161, 235, 48 },
		{ 161, 236, 48 },
		{ 161, 237, 48 },
		{ 161, 238, 48 },
		{ 161, 239, 48 },
		{ 161, 240, 48 },
		{ 161, 241, 48 },
		{ 161, 242, 48 },
		{ 161, 243, 48 },
		{ 161, 244, 48 },
		{ 162, 222, 48 },
		{ 162, 223, 48 },
		{ 162, 224, 48 },
		{ 162, 225, 48 },
		{ 162, 226, 48 },
		{ 162, 227, 48 },
		{ 162, 228, 48 },
		{ 162, 229, 48 },
		{ 162, 230, 48 },
		{ 162, 231, 48 },
		{ 162, 232, 48 },
		{ 162, 233, 48 },
		{ 162, 234, 48 },
		{ 162, 235, 48 },
		{ 162, 236, 48 },
		{ 162, 237, 48 },
		{ 162, 238, 48 },
		{ 162, 239, 48 },
		{ 162, 240, 48 },
		{ 162, 241, 48 },
		{ 162, 242, 48 },
		{ 162, 243, 48 },
		{ 162, 244, 48 },
		{ 162, 245, 48 },
		{ 163, 222, 48 },
		{ 163, 223, 48 },
		{ 163, 224, 48 },
		{ 163, 225, 48 },
		{ 163, 226, 48 },
		{ 163, 227, 48 },
		{ 163, 228, 48 },
		{ 163, 229, 48 },
		{ 163, 230, 48 },
		{ 163, 231, 48 },
		{ 163, 232, 48 },
		{ 163, 233, 48 },
		{ 163, 234, 48 },
		{ 163, 235, 48 },
		{ 163, 236, 48 },
		{ 163, 237, 48 },
		{ 163, 238, 48 },
		{ 163, 239, 48 },
		{ 163, 240, 48 },
		{ 163, 241, 48 },
		{ 163, 242, 48 },
		{ 163, 243, 48 },
		{ 163, 244, 48 },
		{ 163, 245, 48 },
		{ 164, 223, 48 },
		{ 164, 224, 48 },
		{ 164, 225, 48 },
		{ 164, 226, 48 },
		{ 164, 227, 48 },
		{ 164, 228, 48 },
		{ 164, 229, 48 },
		{ 164, 230, 48 },
		{ 164, 231, 48 },
		{ 164, 232, 48 },
		{ 164, 233, 48 },
		{ 164, 234, 48 },
		{ 164, 235, 48 },
		{ 164, 236, 48 },
		{ 164, 237, 48 },
		{ 164, 238, 48 },
		{ 164, 239, 48 },
		{ 164, 240, 48 },
		{ 164, 241, 48 },
		{ 164, 242, 48 },
		{ 164, 243, 48 },
		{ 164, 244, 48 },
		{ 164, 245, 48 },
		{ 165, 223, 48 },
		{ 165, 224, 48 },
		{ 165, 225, 48 },
		{ 165, 226, 48 },
		{ 165, 227, 48 },
		{ 165, 228, 48 },
		{ 165, 229, 48 },
		{ 165, 230, 48 },
		{ 165, 231, 48 },
		{ 165, 232, 48 },
		{ 165, 233, 48 },
		{ 165, 234, 48 },
		{ 165, 235, 48 },
		{ 165, 236, 48 },
		{ 165, 237, 48 },
		{ 165, 238, 48 },
		{ 165, 239, 48 },
		{ 165, 240, 48 },
		{ 165, 241, 48 },
		{ 165, 242, 48 },
		{ 165, 243, 48 },
		{ 165, 244, 48 },
		{ 165, 245, 48 },
		{ 166, 223, 48 },
		{ 166, 224, 48 },
		{ 166, 225, 48 },
		{ 166, 226, 48 },
		{ 166, 227, 48 },
		{ 166, 228, 48 },
		{ 166, 229, 48 },
		{ 166, 230, 48 },
		{ 166, 231, 48 },
		{ 166, 232, 48 },
		{ 166, 233, 48 },
		{ 166, 234, 48 },
		{ 166, 235, 48 },
		{ 166, 236, 48 },
		{ 166, 237, 48 },
		{ 166, 238, 48 },
		{ 166, 239, 48 },
		{ 166, 240, 48 },
		{ 166, 241, 48 },
		{ 166, 242, 48 },
		{ 166, 243, 48 },
		{ 166, 244, 48 },
		{ 166, 245, 48 },
		{ 167, 224, 48 },
		{ 167, 225, 48 },
		{ 167, 226, 48 },
		{ 167, 227, 48 },
		{ 167, 228, 48 },
		{ 167, 229, 48 },
		{ 167, 230, 48 },
		{ 167, 231, 48 },
		{ 167, 232, 48 },
		{ 167, 233, 48 },
		{ 167, 234, 48 },
		{ 167, 235, 48 },
		{ 167, 236, 48 },
		{ 167, 237, 48 },
		{ 167, 238, 48 },
		{ 167, 239, 48 },
		{ 167, 240, 48 },
		{ 167, 241, 48 },
		{ 167, 242, 48 },
		{ 167, 243, 48 },
		{ 167, 244, 48 },
		{ 167, 245, 48 },
		{ 168, 224, 48 },
		{ 168, 225, 48 },
		{ 168, 226, 48 },
		{ 168, 227, 48 },
		{ 168, 228, 48 },
		{ 168, 229, 48 },
		{ 168, 230, 48 },
		{ 168, 231, 48 },
		{ 168, 232, 48 },
		{ 168, 233, 48 },
		{ 168, 234, 48 },
		{ 168, 235, 48 },
		{ 168, 236, 48 },
		{ 168, 237, 48 },
		{ 168, 238, 48 },
		{ 168, 239, 48 },
		{ 168, 240, 48 },
		{ 168, 241, 48 },
		{ 168, 242, 48 },
		{ 168, 243, 48 },
		{ 168, 244, 48 },
		{ 168, 245, 48 },
		{ 169, 225, 48 },
		{ 169, 226, 48 },
		{ 169, 227, 48 },
		{ 169, 228, 48 },
		{ 169, 229, 48 },
		{ 169, 230, 48 },
		{ 169, 231, 48 },
		{ 169, 232, 48 },
		{ 169, 233, 48 },
		{ 169, 234, 48 },
		{ 169, 235, 48 },
		{ 169, 236, 48 },
		{ 169, 237, 48 },
		{ 169, 238, 48 },
		{ 169, 239, 48 },
		{ 169, 240, 48 },
		{ 169, 241, 48 },
		{ 169, 242, 48 },
		{ 169, 243, 48 },
		{ 169, 244, 48 },
		{ 169, 245, 48 },
		{ 170, 225, 48 },
		{ 170, 226, 48 },
		{ 170, 227, 48 },
		{ 170, 228, 48 },
		{ 170, 229, 48 },
		{ 170, 230, 48 },
		{ 170, 231, 48 },
		{ 170, 232, 48 },
		{ 170, 233, 48 },
		{ 170, 234, 48 },
		{ 170, 235, 48 },
		{ 170, 236, 48 },
		{ 170, 237, 48 },
		{ 170, 238, 48 },
		{ 170, 239, 48 },
		{ 170, 240, 48 },
		{ 170, 241, 48 },
		{ 170, 242, 48 },
		{ 170, 243, 48 },
		{ 170, 244, 48 },
		{ 170, 245, 48 },
		{ 170, 246, 48 },
		{ 171, 225, 48 },
		{ 171, 226, 48 },
		{ 171, 227, 48 },
		{ 171, 228, 48 },
		{ 171, 229, 48 },
		{ 171, 230, 48 },
		{ 171, 231, 48 },
		{ 171, 232, 48 },
		{ 171, 233, 48 },
		{ 171, 234, 48 },
		{ 171, 235, 48 },
		{ 171, 236, 48 },
		{ 171, 237, 48 },
		{ 171, 238, 48 },
		{ 171, 239, 48 },
		{ 171, 240, 48 },
		{ 171, 241, 48 },
		{ 171, 242, 48 },
		{ 171, 243, 48 },
		{ 171, 244, 48 },
		{ 171, 245, 48 },
		{ 171, 246, 48 },
		{ 172, 226, 48 },
		{ 172, 227, 48 },
		{ 172, 228, 48 },
		{ 172, 229, 48 },
		{ 172, 230, 48 },
		{ 172, 231, 48 },
		{ 172, 232, 48 },
		{ 172, 233, 48 },
		{ 172, 234, 48 },
		{ 172, 235, 48 },
		{ 172, 236, 48 },
		{ 172, 237, 48 },
		{ 172, 238, 48 },
		{ 172, 239, 48 },
		{ 172, 240, 48 },
		{ 172, 241, 48 },
		{ 172, 242, 48 },
		{ 172, 243, 48 },
		{ 172, 244, 48 },
		{ 172, 245, 48 },
		{ 172, 246, 48 },
		{ 173, 226, 48 },
		{ 173, 227, 48 },
		{ 173, 228, 48 },
		{ 173, 229, 48 },
		{ 173, 230, 48 },
		{ 173, 231, 48 },
		{ 173, 232, 48 },
		{ 173, 233, 48 },
		{ 173, 234, 48 },
		{ 173, 235, 48 },
		{ 173, 236, 48 },
		{ 173, 237, 48 },
		{ 173, 238, 48 },
		{ 173, 239, 48 },
		{ 173, 240, 48 },
		{ 173, 241, 48 },
		{ 173, 242, 48 },
		{ 173, 243, 48 },
		{ 173, 244, 48 },
		{ 173, 245, 48 },
		{ 173, 246, 48 },
		{ 174, 226, 48 },
		{ 174, 227, 48 },
		{ 174, 228, 48 },
		{ 174, 229, 48 },
		{ 174, 230, 48 },
		{ 174, 231, 48 },
		{ 174, 232, 48 },
		{ 174, 233, 48 },
		{ 174, 234, 48 },
		{ 174, 235, 48 },
		{ 174, 236, 48 },
		{ 174, 237, 48 },
		{ 174, 238, 48 },
		{ 174, 239, 48 },
		{ 174, 240, 48 },
		{ 174, 241, 48 },
		{ 174, 242, 48 },
		{ 174, 243, 48 },
		{ 174, 244, 48 },
		{ 174, 245, 48 },
		{ 174, 246, 48 },
		{ 175, 227, 48 },
		{ 175, 228, 48 },
		{ 175, 229, 48 },
		{ 175, 230, 48 },
		{ 175, 231, 48 },
		{ 175, 232, 48 },
		{ 175, 233, 48 },
		{ 175, 234, 48 },
		{ 175, 235, 48 },
		{ 175, 236, 48 },
		{ 175, 237, 48 },
		{ 175, 238, 48 },
		{ 175, 239, 48 },
		{ 175, 240, 48 },
		{ 175, 241, 48 },
		{ 175, 242, 48 },
		{ 175, 243, 48 },
		{ 175, 244, 48 },
		{ 175, 245, 48 },
		{ 175, 246, 48 },
		{ 176, 227, 48 },
		{ 176, 228, 48 },
		{ 176, 229, 48 },
		{ 176, 230, 48 },
		{ 176, 231, 48 },
		{ 176, 232, 48 },
		{ 176, 233, 48 },
		{ 176, 234, 48 },
		{ 176, 235, 48 },
		{ 176, 236, 48 },
		{ 176, 237, 48 },
		{ 176, 238, 48 },
		{ 176, 239, 48 },
		{ 176, 240, 48 },
		{ 176, 241, 48 },
		{ 176, 242, 48 },
		{ 176, 243, 48 },
		{ 176, 244, 48 },
		{ 176, 245, 48 },
		{ 176, 246, 48 },
		{ 177, 228, 48 },
		{ 177, 229, 48 },
		{ 177, 230, 48 },
		{ 177, 231, 48 },
		{ 177, 232, 48 },
		{ 177, 233, 48 },
		{ 177, 234, 48 },
		{ 177, 235, 48 },
		{ 177, 236, 48 },
		{ 177, 237, 48 },
		{ 177, 238, 48 },
		{ 177, 239, 48 },
		{ 177, 240, 48 },
		{ 177, 241, 48 },
		{ 177, 242, 48 },
		{ 177, 243, 48 },
		{ 177, 244, 48 },
		{ 177, 245, 48 },
		{ 177, 246, 48 },
		{ 178, 228, 48 },
		{ 178, 229, 48 },
		{ 178, 230, 48 },
		{ 178, 231, 48 },
		{ 178, 232, 48 },
		{ 178, 233, 48 },
		{ 178, 234, 48 },
		{ 178, 235, 48 },
		{ 178, 236, 48 },
		{ 178, 237, 48 },
		{ 178, 238, 48 },
		{ 178, 239, 48 },
		{ 178, 240, 48 },
		{ 178, 241, 48 },
		{ 178, 242, 48 },
		{ 178, 243, 48 },
		{ 178, 244, 48 },
		{ 178, 245, 48 },
		{ 178, 246, 48 },
		{ 178, 247, 48 },
		{ 179, 228, 48 },
		{ 179, 229, 48 },
		{ 179, 230, 48 },
		{ 179, 231, 48 },
		{ 179, 232, 48 },
		{ 179, 233, 48 },
		{ 179, 234, 48 },
		{ 179, 235, 48 },
		{ 179, 236, 48 },
		{ 179, 237, 48 },
		{ 179, 238, 48 },
		{ 179, 239, 48 },
		{ 179, 240, 48 },
		{ 179, 241, 48 },
		{ 179, 242, 48 },
		{ 179, 243, 48 },
		{ 179, 244, 48 },
		{ 179, 245, 48 },
		{ 179, 246, 48 },
		{ 179, 247, 48 },
		{ 180, 229, 48 },
		{ 180, 230, 48 },
		{ 180, 231, 48 },
		{ 180, 232, 48 },
		{ 180, 233, 48 },
		{ 180, 234, 48 },
		{ 180, 235, 48 },
		{ 180, 236, 48 },
		{ 180, 237, 48 },
		{ 180, 238, 48 },
		{ 180, 239, 48 },
		{ 180, 240, 48 },
		{ 180, 241, 48 },
		{ 180, 242, 48 },
		{ 180, 243, 48 },
		{ 180, 244, 48 },
		{ 180, 245, 48 },
		{ 180, 246, 48 },
		{ 180, 247, 48 },
		{ 181, 229, 48 },
		{ 181, 230, 48 },
		{ 181, 231, 48 },
		{ 181, 232, 48 },
		{ 181, 233, 48 },
		{ 181, 234, 48 },
		{ 181, 235, 48 },
		{ 181, 236, 48 },
		{ 181, 237, 48 },
		{ 181, 238, 48 },
		{ 181, 239, 48 },
		{ 181, 240, 48 },
		{ 181, 241, 48 },
		{ 181, 242, 48 },
		{ 181, 243, 48 },
		{ 181, 244, 48 },
		{ 181, 245, 48 },
		{ 181, 246, 48 },
		{ 181, 247, 48 },
		{ 182, 229, 48 },
		{ 182, 230, 48 },
		{ 182, 231, 48 },
		{ 182, 232, 48 },
		{ 182, 233, 48 },
		{ 182, 234, 48 },
		{ 182, 235, 48 },
		{ 182, 236, 48 },
		{ 182, 237, 48 },
		{ 182, 238, 48 },
		{ 182, 239, 48 },
		{ 182, 240, 48 },
		{ 182, 241, 48 },
		{ 182, 242, 48 },
		{ 182, 243, 48 },
		{ 182, 244, 48 },
		{ 182, 245, 48 },
		{ 182, 246, 48 },
		{ 182, 247, 48 },
		{ 183, 230, 48 },
		{ 183, 231, 48 },
		{ 183, 232, 48 },
		{ 183, 233, 48 },
		{ 183, 234, 48 },
		{ 183, 235, 48 },
		{ 183, 236, 48 },
		{ 183, 237, 48 },
		{ 183, 238, 48 },
		{ 183, 239, 48 },
		{ 183, 240, 48 },
		{ 183, 241, 48 },
		{ 183, 242, 48 },
		{ 183, 243, 48 },
		{ 183, 244, 48 },
		{ 183, 245, 48 },
		{ 183, 246, 48 },
		{ 183, 247, 48 },
		{ 184, 230, 48 },
		{ 184, 231, 48 },
		{ 184, 232, 48 },
		{ 184, 233, 48 },
		{ 184, 234, 48 },
		{ 184, 235, 48 },
		{ 184, 236, 48 },
		{ 184, 237, 48 },
		{ 184, 238, 48 },
		{ 184, 239, 48 },
		{ 184, 240, 48 },
		{ 184, 241, 48 },
		{ 184, 242, 48 },
		{ 184, 243, 48 },
		{ 184, 244, 48 },
		{ 184, 245, 48 },
		{ 184, 246, 48 },
		{ 184, 247, 48 },
		{ 185, 231, 48 },
		{ 185, 232, 48 },
		{ 185, 233, 48 },
		{ 185, 234, 48 },
		{ 185, 235, 48 },
		{ 185, 236, 48 },
		{ 185, 237, 48 },
		{ 185, 238, 48 },
		{ 185, 239, 48 },
		{ 185, 240, 48 },
		{ 185, 241, 48 },
		{ 185, 242, 48 },
		{ 185, 243, 48 },
		{ 185, 244, 48 },
		{ 185, 245, 48 },
		{ 185, 246, 48 },
		{ 185, 247, 48 },
		{ 186, 231, 48 },
		{ 186, 232, 48 },
		{ 186, 233, 48 },
		{ 186, 234, 48 },
		{ 186, 235, 48 },
		{ 186, 236, 48 },
		{ 186, 237, 48 },
		{ 186, 238, 48 },
		{ 186, 239, 48 },
		{ 186, 240, 48 },
		{ 186, 241, 48 },
		{ 186, 242, 48 },
		{ 186, 243, 48 },
		{ 186, 244, 48 },
		{ 186, 245, 48 },
		{ 186, 246, 48 },
		{ 186, 247, 48 },
		{ 187, 231, 48 },
		{ 187, 232, 48 },
		{ 187, 233, 48 },
		{ 187, 234, 48 },
		{ 187, 235, 48 },
		{ 187, 236, 48 },
		{ 187, 237, 48 },
		{ 187, 238, 48 },
		{ 187, 239, 48 },
		{ 187, 240, 48 },
		{ 187, 241, 48 },
		{ 187, 242, 48 },
		{ 187, 243, 48 },
		{ 187, 244, 48 },
		{ 187, 245, 48 },
		{ 187, 246, 48 },
		{ 187, 247, 48 },
		{ 187, 248, 48 },
		{ 188, 232, 48 },
		{ 188, 233, 48 },
		{ 188, 234, 48 },
		{ 188, 235, 48 },
		{ 188, 236, 48 },
		{ 188, 237, 48 },
		{ 188, 238, 48 },
		{ 188, 239, 48 },
		{ 188, 240, 48 },
		{ 188, 241, 48 },
		{ 188, 242, 48 },
		{ 188, 243, 48 },
		{ 188, 244, 48 },
		{ 188, 245, 48 },
		{ 188, 246, 48 },
		{ 188, 247, 48 },
		{ 188, 248, 48 },
		{ 189, 232, 48 },
		{ 189, 233, 48 },
		{ 189, 234, 48 },
		{ 189, 235, 48 },
		{ 189, 236, 48 },
		{ 189, 237, 48 },
		{ 189, 238, 48 },
		{ 189, 239, 48 },
		{ 189, 240, 48 },
		{ 189, 241, 48 },
		{ 189, 242, 48 },
		{ 189, 243, 48 },
		{ 189, 244, 48 },
		{ 189, 245, 48 },
		{ 189, 246, 48 },
		{ 189, 247, 48 },
		{ 189, 248, 48 },
		{ 190, 233, 48 },
		{ 190, 234, 48 },
		{ 190, 235, 48 },
		{ 190, 236, 48 },
		{ 190, 237, 48 },
		{ 190, 238, 48 },
		{ 190, 239, 48 },
		{ 190, 240, 48 },
		{ 190, 241, 48 },
		{ 190, 242, 48 },
		{ 190, 243, 48 },
		{ 190, 244, 48 },
		{ 190, 245, 48 },
		{ 190, 246, 48 },
		{ 190, 247, 48 },
		{ 191, 235, 48 },
		{ 191, 236, 48 },
		{ 191, 237, 48 },
		{ 191, 238, 48 },
		{ 191, 239, 48 },
		{ 191, 240, 48 },
		{ 191, 241, 48 },
		{ 191, 242, 48 },
		{ 191, 243, 48 },
		{ 191, 244, 48 },
		{ 191, 245, 48 },
		{ 191, 246, 48 },
		{ 323, 268, 48 },
		{ 323, 269, 48 },
		{ 323, 270, 48 },
		{ 323, 271, 48 },
		{ 323, 272, 48 },
		{ 323, 273, 48 },
		{ 323, 274, 48 },
		{ 323, 275, 48 },
		{ 323, 276, 48 },
		{ 323, 277, 48 },
		{ 323, 278, 48 },
		{ 323, 279, 48 },
		{ 324, 267, 48 },
		{ 324, 268, 48 },
		{ 324, 269, 48 },
		{ 324, 270, 48 },
		{ 324, 271, 48 },
		{ 324, 272, 48 },
		{ 324, 273, 48 },
		{ 324, 274, 48 },
		{ 324, 275, 48 },
		{ 324, 276, 48 },
		{ 324, 277, 48 },
		{ 324, 278, 48 },
		{ 324, 279, 48 },
		{ 324, 280, 48 },
		{ 324, 281, 48 },
		{ 325, 266, 48 },
		{ 325, 267, 48 },
		{ 325, 268, 48 },
		{ 325, 269, 48 },
		{ 325, 270, 48 },
		{ 325, 271, 48 },
		{ 325, 272, 48 },
		{ 325, 273, 48 },
		{ 325, 274, 48 },
		{ 325, 275, 48 },
		{ 325, 276, 48 },
		{ 325, 277, 48 },
		{ 325, 278, 48 },
		{ 325, 279, 48 },
		{ 325, 280, 48 },
		{ 325, 281, 48 },
		{ 325, 282, 48 },
		{ 326, 266, 48 },
		{ 326, 267, 48 },
		{ 326, 268, 48 },
		{ 326, 269, 48 },
		{ 326, 270, 48 },
		{ 326, 271, 48 },
		{ 326, 272, 48 },
		{ 326, 273, 48 },
		{ 326, 274, 48 },
		{ 326, 275, 48 },
		{ 326, 276, 48 },
		{ 326, 277, 48 },
		{ 326, 278, 48 },
		{ 326, 279, 48 },
		{ 326, 280, 48 },
		{ 326, 281, 48 },
		{ 326, 282, 48 },
		{ 327, 266, 48 },
		{ 327, 267, 48 },
		{ 327, 268, 48 },
		{ 327, 269, 48 },
		{ 327, 270, 48 },
		{ 327, 271, 48 },
		{ 327, 272, 48 },
		{ 327, 273, 48 },
		{ 327, 274, 48 },
		{ 327, 275, 48 },
		{ 327, 276, 48 },
		{ 327, 277, 48 },
		{ 327, 278, 48 },
		{ 327, 279, 48 },
		{ 327, 280, 48 },
		{ 327, 281, 48 },
		{ 327, 282, 48 },
		{ 327, 283, 48 },
		{ 328, 267, 48 },
		{ 328, 268, 48 },
		{ 328, 269, 48 },
		{ 328, 270, 48 },
		{ 328, 271, 48 },
		{ 328, 272, 48 },
		{ 328, 273, 48 },
		{ 328, 274, 48 },
		{ 328, 275, 48 },
		{ 328, 276, 48 },
		{ 328, 277, 48 },
		{ 328, 278, 48 },
		{ 328, 279, 48 },
		{ 328, 280, 48 },
		{ 328, 281, 48 },
		{ 328, 282, 48 },
		{ 328, 283, 48 },
		{ 329, 267, 48 },
		{ 329, 268, 48 },
		{ 329, 269, 48 },
		{ 329, 270, 48 },
		{ 329, 271, 48 },
		{ 329, 272, 48 },
		{ 329, 273, 48 },
		{ 329, 274, 48 },
		{ 329, 275, 48 },
		{ 329, 276, 48 },
		{ 329, 277, 48 },
		{ 329, 278, 48 },
		{ 329, 279, 48 },
		{ 329, 280, 48 },
		{ 329, 281, 48 },
		{ 329, 282, 48 },
		{ 329, 283, 48 },
		{ 330, 267, 48 },
		{ 330, 268, 48 },
		{ 330, 269, 48 },
		{ 330, 270, 48 },
		{ 330, 271, 48 },
		{ 330, 272, 48 },
		{ 330, 273, 48 },
		{ 330, 274, 48 },
		{ 330, 275, 48 },
		{ 330, 276, 48 },
		{ 330, 277, 48 },
		{ 330, 278, 48 },
		{ 330, 279, 48 },
		{ 330, 280, 48 },
		{ 330, 281, 48 },
		{ 330, 282, 48 },
		{ 330, 283, 48 },
		{ 330, 284, 48 },
		{ 331, 267, 48 },
		{ 331, 268, 48 },
		{ 331, 269, 48 },
		{ 331, 270, 48 },
		{ 331, 271, 48 },
		{ 331, 272, 48 },
		{ 331, 273, 48 },
		{ 331, 274, 48 },
		{ 331, 275, 48 },
		{ 331, 276, 48 },
		{ 331, 277, 48 },
		{ 331, 278, 48 },
		{ 331, 279, 48 },
		{ 331, 280, 48 },
		{ 331, 281, 48 },
		{ 331, 282, 48 },
		{ 331, 283, 48 },
		{ 331, 284, 48 },
		{ 332, 267, 48 },
		{ 332, 268, 48 },
		{ 332, 269, 48 },
		{ 332, 270, 48 },
		{ 332, 271, 48 },
		{ 332, 272, 48 },
		{ 332, 273, 48 },
		{ 332, 274, 48 },
		{ 332, 275, 48 },
		{ 332, 276, 48 },
		{ 332, 277, 48 },
		{ 332, 278, 48 },
		{ 332, 279, 48 },
		{ 332, 280, 48 },
		{ 332, 281, 48 },
		{ 332, 282, 48 },
		{ 332, 283, 48 },
		{ 332, 284, 48 },
		{ 332, 285, 48 },
		{ 333, 267, 48 },
		{ 333, 268, 48 },
		{ 333, 269, 48 },
		{ 333, 270, 48 },
		{ 333, 271, 48 },
		{ 333, 272, 48 },
		{ 333, 273, 48 },
		{ 333, 274, 48 },
		{ 333, 275, 48 },
		{ 333, 276, 48 },
		{ 333, 277, 48 },
		{ 333, 278, 48 },
		{ 333, 279, 48 },
		{ 333, 280, 48 },
		{ 333, 281, 48 },
		{ 333, 282, 48 },
		{ 333, 283, 48 },
		{ 333, 284, 48 },
		{ 333, 285, 48 },
		{ 334, 267, 48 },
		{ 334, 268, 48 },
		{ 334, 269, 48 },
		{ 334, 270, 48 },
		{ 334, 271, 48 },
		{ 334, 272, 48 },
		{ 334, 273, 48 },
		{ 334, 274, 48 },
		{ 334, 275, 48 },
		{ 334, 276, 48 },
		{ 334, 277, 48 },
		{ 334, 278, 48 },
		{ 334, 279, 48 },
		{ 334, 280, 48 },
		{ 334, 281, 48 },
		{ 334, 282, 48 },
		{ 334, 283, 48 },
		{ 334, 284, 48 },
		{ 334, 285, 48 },
		{ 335, 267, 48 },
		{ 335, 268, 48 },
		{ 335, 269, 48 },
		{ 335, 270, 48 },
		{ 335, 271, 48 },
		{ 335, 272, 48 },
		{ 335, 273, 48 },
		{ 335, 274, 48 },
		{ 335, 275, 48 },
		{ 335, 276, 48 },
		{ 335, 277, 48 },
		{ 335, 278, 48 },
		{ 335, 279, 48 },
		{ 335, 280, 48 },
		{ 335, 281, 48 },
		{ 335, 282, 48 },
		{ 335, 283, 48 },
		{ 335, 284, 48 },
		{ 335, 285, 48 },
		{ 335, 286, 48 },
		{ 336, 267, 48 },
		{ 336, 268, 48 },
		{ 336, 269, 48 },
		{ 336, 270, 48 },
		{ 336, 271, 48 },
		{ 336, 272, 48 },
		{ 336, 273, 48 },
		{ 336, 274, 48 },
		{ 336, 275, 48 },
		{ 336, 276, 48 },
		{ 336, 277, 48 },
		{ 336, 278, 48 },
		{ 336, 279, 48 },
		{ 336, 280, 48 },
		{ 336, 281, 48 },
		{ 336, 282, 48 },
		{ 336, 283, 48 },
		{ 336, 284, 48 },
		{ 336, 285, 48 },
		{ 336, 286, 48 },
		{ 337, 268, 48 },
		{ 337, 269, 48 },
		{ 337, 270, 48 },
		{ 337, 271, 48 },
		{ 337, 272, 48 },
		{ 337, 273, 48 },
		{ 337, 274, 48 },
		{ 337, 275, 48 },
		{ 337, 276, 48 },
		{ 337, 277, 48 },
		{ 337, 278, 48 },
		{ 337, 279, 48 },
		{ 337, 280, 48 },
		{ 337, 281, 48 },
		{ 337, 282, 48 },
		{ 337, 283, 48 },
		{ 337, 284, 48 },
		{ 337, 285, 48 },
		{ 337, 286, 48 },
		{ 338, 268, 48 },
		{ 338, 269, 48 },
		{ 338, 270, 48 },
		{ 338, 271, 48 },
		{ 338, 272, 48 },
		{ 338, 273, 48 },
		{ 338, 274, 48 },
		{ 338, 275, 48 },
		{ 338, 276, 48 },
		{ 338, 277, 48 },
		{ 338, 278, 48 },
		{ 338, 279, 48 },
		{ 338, 280, 48 },
		{ 338, 281, 48 },
		{ 338, 282, 48 },
		{ 338, 283, 48 },
		{ 338, 284, 48 },
		{ 338, 285, 48 },
		{ 338, 286, 48 },
		{ 338, 287, 48 },
		{ 339, 268, 48 },
		{ 339, 269, 48 },
		{ 339, 270, 48 },
		{ 339, 271, 48 },
		{ 339, 272, 48 },
		{ 339, 273, 48 },
		{ 339, 274, 48 },
		{ 339, 275, 48 },
		{ 339, 276, 48 },
		{ 339, 277, 48 },
		{ 339, 278, 48 },
		{ 339, 279, 48 },
		{ 339, 280, 48 },
		{ 339, 281, 48 },
		{ 339, 282, 48 },
		{ 339, 283, 48 },
		{ 339, 284, 48 },
		{ 339, 285, 48 },
		{ 339, 286, 48 },
		{ 339, 287, 48 },
		{ 340, 268, 48 },
		{ 340, 269, 48 },
		{ 340, 270, 48 },
		{ 340, 271, 48 },
		{ 340, 272, 48 },
		{ 340, 273, 48 },
		{ 340, 274, 48 },
		{ 340, 275, 48 },
		{ 340, 276, 48 },
		{ 340, 277, 48 },
		{ 340, 278, 48 },
		{ 340, 279, 48 },
		{ 340, 280, 48 },
		{ 340, 281, 48 },
		{ 340, 282, 48 },
		{ 340, 283, 48 },
		{ 340, 284, 48 },
		{ 340, 285, 48 },
		{ 340, 286, 48 },
		{ 340, 287, 48 },
		{ 340, 288, 48 },
		{ 341, 268, 48 },
		{ 341, 269, 48 },
		{ 341, 270, 48 },
		{ 341, 271, 48 },
		{ 341, 272, 48 },
		{ 341, 273, 48 },
		{ 341, 274, 48 },
		{ 341, 275, 48 },
		{ 341, 276, 48 },
		{ 341, 277, 48 },
		{ 341, 278, 48 },
		{ 341, 279, 48 },
		{ 341, 280, 48 },
		{ 341, 281, 48 },
		{ 341, 282, 48 },
		{ 341, 283, 48 },
		{ 341, 284, 48 },
		{ 341, 285, 48 },
		{ 341, 286, 48 },
		{ 341, 287, 48 },
		{ 341, 288, 48 },
		{ 342, 268, 48 },
		{ 342, 269, 48 },
		{ 342, 270, 48 },
		{ 342, 271, 48 },
		{ 342, 272, 48 },
		{ 342, 273, 48 },
		{ 342, 274, 48 },
		{ 342, 275, 48 },
		{ 342, 276, 48 },
		{ 342, 277, 48 },
		{ 342, 278, 48 },
		{ 342, 279, 48 },
		{ 342, 280, 48 },
		{ 342, 281, 48 },
		{ 342, 282, 48 },
		{ 342, 283, 48 },
		{ 342, 284, 48 },
		{ 342, 285, 48 },
		{ 342, 286, 48 },
		{ 342, 287, 48 },
		{ 342, 288, 48 },
		{ 343, 268, 48 },
		{ 343, 269, 48 },
		{ 343, 270, 48 },
		{ 343, 271, 48 },
		{ 343, 272, 48 },
		{ 343, 273, 48 },
		{ 343, 274, 48 },
		{ 343, 275, 48 },
		{ 343, 276, 48 },
		{ 343, 277, 48 },
		{ 343, 278, 48 },
		{ 343, 279, 48 },
		{ 343, 280, 48 },
		{ 343, 281, 48 },
		{ 343, 282, 48 },
		{ 343, 283, 48 },
		{ 343, 284, 48 },
		{ 343, 285, 48 },
		{ 343, 286, 48 },
		{ 343, 287, 48 },
		{ 343, 288, 48 },
		{ 343, 289, 48 },
		{ 344, 268, 48 },
		{ 344, 269, 48 },
		{ 344, 270, 48 },
		{ 344, 271, 48 },
		{ 344, 272, 48 },
		{ 344, 273, 48 },
		{ 344, 274, 48 },
		{ 344, 275, 48 },
		{ 344, 276, 48 },
		{ 344, 277, 48 },
		{ 344, 278, 48 },
		{ 344, 279, 48 },
		{ 344, 280, 48 },
		{ 344, 281, 48 },
		{ 344, 282, 48 },
		{ 344, 283, 48 },
		{ 344, 284, 48 },
		{ 344, 285, 48 },
		{ 344, 286, 48 },
		{ 344, 287, 48 },
		{ 344, 288, 48 },
		{ 344, 289, 48 },
		{ 345, 269, 48 },
		{ 345, 270, 48 },
		{ 345, 271, 48 },
		{ 345, 272, 48 },
		{ 345, 273, 48 },
		{ 345, 274, 48 },
		{ 345, 275, 48 },
		{ 345, 276, 48 },
		{ 345, 277, 48 },
		{ 345, 278, 48 },
		{ 345, 279, 48 },
		{ 345, 280, 48 },
		{ 345, 281, 48 },
		{ 345, 282, 48 },
		{ 345, 283, 48 },
		{ 345, 284, 48 },
		{ 345, 285, 48 },
		{ 345, 286, 48 },
		{ 345, 287, 48 },
		{ 345, 288, 48 },
		{ 345, 289, 48 },
		{ 346, 269, 48 },
		{ 346, 270, 48 },
		{ 346, 271, 48 },
		{ 346, 272, 48 },
		{ 346, 273, 48 },
		{ 346, 274, 48 },
		{ 346, 275, 48 },
		{ 346, 276, 48 },
		{ 346, 277, 48 },
		{ 346, 278, 48 },
		{ 346, 279, 48 },
		{ 346, 280, 48 },
		{ 346, 281, 48 },
		{ 346, 282, 48 },
		{ 346, 283, 48 },
		{ 346, 284, 48 },
		{ 346, 285, 48 },
		{ 346, 286, 48 },
		{ 346, 287, 48 },
		{ 346, 288, 48 },
		{ 346, 289, 48 },
		{ 346, 290, 48 },
		{ 347, 269, 48 },
		{ 347, 270, 48 },
		{ 347, 271, 48 },
		{ 347, 272, 48 },
		{ 347, 273, 48 },
		{ 347, 274, 48 },
		{ 347, 275, 48 },
		{ 347, 276, 48 },
		{ 347, 277, 48 },
		{ 347, 278, 48 },
		{ 347, 279, 48 },
		{ 347, 280, 48 },
		{ 347, 281, 48 },
		{ 347, 282, 48 },
		{ 347, 283, 48 },
		{ 347, 284, 48 },
		{ 347, 285, 48 },
		{ 347, 286, 48 },
		{ 347, 287, 48 },
		{ 347, 288, 48 },
		{ 347, 289, 48 },
		{ 347, 290, 48 },
		{ 348, 269, 48 },
		{ 348, 270, 48 },
		{ 348, 271, 48 },
		{ 348, 272, 48 },
		{ 348, 273, 48 },
		{ 348, 274, 48 },
		{ 348, 275, 48 },
		{ 348, 276, 48 },
		{ 348, 277, 48 },
		{ 348, 278, 48 },
		{ 348, 279, 48 },
		{ 348, 280, 48 },
		{ 348, 281, 48 },
		{ 348, 282, 48 },
		{ 348, 283, 48 },
		{ 348, 284, 48 },
		{ 348, 285, 48 },
		{ 348, 286, 48 },
		{ 348, 287, 48 },
		{ 348, 288, 48 },
		{ 348, 289, 48 },
		{ 348, 290, 48 },
		{ 348, 291, 48 },
		{ 349, 269, 48 },
		{ 349, 270, 48 },
		{ 349, 271, 48 },
		{ 349, 272, 48 },
		{ 349, 273, 48 },
		{ 349, 274, 48 },
		{ 349, 275, 48 },
		{ 349, 276, 48 },
		{ 349, 277, 48 },
		{ 349, 278, 48 },
		{ 349, 279, 48 },
		{ 349, 280, 48 },
		{ 349, 281, 48 },
		{ 349, 282, 48 },
		{ 349, 283, 48 },
		{ 349, 284, 48 },
		{ 349, 285, 48 },
		{ 349, 286, 48 },
		{ 349, 287, 48 },
		{ 349, 288, 48 },
		{ 349, 289, 48 },
		{ 349, 290, 48 },
		{ 349, 291, 48 },
		{ 350, 269, 48 },
		{ 350, 270, 48 },
		{ 350, 271, 48 },
		{ 350, 272, 48 },
		{ 350, 273, 48 },
		{ 350, 274, 48 },
		{ 350, 275, 48 },
		{ 350, 276, 48 },
		{ 350, 277, 48 },
		{ 350, 278, 48 },
		{ 350, 279, 48 },
		{ 350, 280, 48 },
		{ 350, 281, 48 },
		{ 350, 282, 48 },
		{ 350, 283, 48 },
		{ 350, 284, 48 },
		{ 350, 285, 48 },
		{ 350, 286, 48 },
		{ 350, 287, 48 },
		{ 350, 288, 48 },
		{ 350, 289, 48 },
		{ 350, 290, 48 },
		{ 350, 291, 48 },
		{ 351, 269, 48 },
		{ 351, 270, 48 },
		{ 351, 271, 48 },
		{ 351, 272, 48 },
		{ 351, 273, 48 },
		{ 351, 274, 48 },
		{ 351, 275, 48 },
		{ 351, 276, 48 },
		{ 351, 277, 48 },
		{ 351, 278, 48 },
		{ 351, 279, 48 },
		{ 351, 280, 48 },
		{ 351, 281, 48 },
		{ 351, 282, 48 },
		{ 351, 283, 48 },
		{ 351, 284, 48 },
		{ 351, 285, 48 },
		{ 351, 286, 48 },
		{ 351, 287, 48 },
		{ 351, 288, 48 },
		{ 351, 289, 48 },
		{ 351, 290, 48 },
		{ 351, 291, 48 },
		{ 351, 292, 48 },
		{ 352, 269, 48 },
		{ 352, 270, 48 },
		{ 352, 271, 48 },
		{ 352, 272, 48 },
		{ 352, 273, 48 },
		{ 352, 274, 48 },
		{ 352, 275, 48 },
		{ 352, 276, 48 },
		{ 352, 277, 48 },
		{ 352, 278, 48 },
		{ 352, 279, 48 },
		{ 352, 280, 48 },
		{ 352, 281, 48 },
		{ 352, 282, 48 },
		{ 352, 283, 48 },
		{ 352, 284, 48 },
		{ 352, 285, 48 },
		{ 352, 286, 48 },
		{ 352, 287, 48 },
		{ 352, 288, 48 },
		{ 352, 289, 48 },
		{ 352, 290, 48 },
		{ 352, 291, 48 },
		{ 352, 292, 48 },
		{ 353, 270, 48 },
		{ 353, 271, 48 },
		{ 353, 272, 48 },
		{ 353, 273, 48 },
		{ 353, 274, 48 },
		{ 353, 275, 48 },
		{ 353, 276, 48 },
		{ 353, 277, 48 },
		{ 353, 278, 48 },
		{ 353, 279, 48 },
		{ 353, 280, 48 },
		{ 353, 281, 48 },
		{ 353, 282, 48 },
		{ 353, 283, 48 },
		{ 353, 284, 48 },
		{ 353, 285, 48 },
		{ 353, 286, 48 },
		{ 353, 287, 48 },
		{ 353, 288, 48 },
		{ 353, 289, 48 },
		{ 353, 290, 48 },
		{ 353, 291, 48 },
		{ 353, 292, 48 },
		{ 354, 270, 48 },
		{ 354, 271, 48 },
		{ 354, 272, 48 },
		{ 354, 273, 48 },
		{ 354, 274, 48 },
		{ 354, 275, 48 },
		{ 354, 276, 48 },
		{ 354, 277, 48 },
		{ 354, 278, 48 },
		{ 354, 279, 48 },
		{ 354, 280, 48 },
		{ 354, 281, 48 },
		{ 354, 282, 48 },
		{ 354, 283, 48 },
		{ 354, 284, 48 },
		{ 354, 285, 48 },
		{ 354, 286, 48 },
		{ 354, 287, 48 },
		{ 354, 288, 48 },
		{ 354, 289, 48 },
		{ 354, 290, 48 },
		{ 354, 291, 48 },
		{ 354, 292, 48 },
		{ 354, 293, 48 },
		{ 355, 270, 48 },
		{ 355, 271, 48 },
		{ 355, 272, 48 },
		{ 355, 273, 48 },
		{ 355, 274, 48 },
		{ 355, 275, 48 },
		{ 355, 276, 48 },
		{ 355, 277, 48 },
		{ 355, 278, 48 },
		{ 355, 279, 48 },
		{ 355, 280, 48 },
		{ 355, 281, 48 },
		{ 355, 282, 48 },
		{ 355, 283, 48 },
		{ 355, 284, 48 },
		{ 355, 285, 48 },
		{ 355, 286, 48 },
		{ 355, 287, 48 },
		{ 355, 288, 48 },
		{ 355, 289, 48 },
		{ 355, 290, 48 },
		{ 355, 291, 48 },
		{ 355, 292, 48 },
		{ 355, 293, 48 },
		{ 356, 270, 48 },
		{ 356, 271, 48 },
		{ 356, 272, 48 },
		{ 356, 273, 48 },
		{ 356, 274, 48 },
		{ 356, 275, 48 },
		{ 356, 276, 48 },
		{ 356, 277, 48 },
		{ 356, 278, 48 },
		{ 356, 279, 48 },
		{ 356, 280, 48 },
		{ 356, 281, 48 },
		{ 356, 282, 48 },
		{ 356, 283, 48 },
		{ 356, 284, 48 },
		{ 356, 285, 48 },
		{ 356, 286, 48 },
		{ 356, 287, 48 },
		{ 356, 288, 48 },
		{ 356, 289, 48 },
		{ 356, 290, 48 },
		{ 356, 291, 48 },
		{ 356, 292, 48 },
		{ 356, 293, 48 },
		{ 356, 294, 48 },
		{ 357, 270, 48 },
		{ 357, 271, 48 },
		{ 357, 272, 48 },
		{ 357, 273, 48 },
		{ 357, 274, 48 },
		{ 357, 275, 48 },
		{ 357, 276, 48 },
		{ 357, 277, 48 },
		{ 357, 278, 48 },
		{ 357, 279, 48 },
		{ 357, 280, 48 },
		{ 357, 281, 48 },
		{ 357, 282, 48 },
		{ 357, 283, 48 },
		{ 357, 284, 48 },
		{ 357, 285, 48 },
		{ 357, 286, 48 },
		{ 357, 287, 48 },
		{ 357, 288, 48 },
		{ 357, 289, 48 },
		{ 357, 290, 48 },
		{ 357, 291, 48 },
		{ 357, 292, 48 },
		{ 357, 293, 48 },
		{ 357, 294, 48 },
		{ 358, 270, 48 },
		{ 358, 271, 48 },
		{ 358, 272, 48 },
		{ 358, 273, 48 },
		{ 358, 274, 48 },
		{ 358, 275, 48 },
		{ 358, 276, 48 },
		{ 358, 277, 48 },
		{ 358, 278, 48 },
		{ 358, 279, 48 },
		{ 358, 280, 48 },
		{ 358, 281, 48 },
		{ 358, 282, 48 },
		{ 358, 283, 48 },
		{ 358, 284, 48 },
		{ 358, 285, 48 },
		{ 358, 286, 48 },
		{ 358, 287, 48 },
		{ 358, 288, 48 },
		{ 358, 289, 48 },
		{ 358, 290, 48 },
		{ 358, 291, 48 },
		{ 358, 292, 48 },
		{ 358, 293, 48 },
		{ 358, 294, 48 },
		{ 359, 270, 48 },
		{ 359, 271, 48 },
		{ 359, 272, 48 },
		{ 359, 273, 48 },
		{ 359, 274, 48 },
		{ 359, 275, 48 },
		{ 359, 276, 48 },
		{ 359, 277, 48 },
		{ 359, 278, 48 },
		{ 359, 279, 48 },
		{ 359, 280, 48 },
		{ 359, 281, 48 },
		{ 359, 282, 48 },
		{ 359, 283, 48 },
		{ 359, 284, 48 },
		{ 359, 285, 48 },
		{ 359, 286, 48 },
		{ 359, 287, 48 },
		{ 359, 288, 48 },
		{ 359, 289, 48 },
		{ 359, 290, 48 },
		{ 359, 291, 48 },
		{ 359, 292, 48 },
		{ 359, 293, 48 },
		{ 359, 294, 48 },
		{ 359, 295, 48 },
		{ 360, 270, 48 },
		{ 360, 271, 48 },
		{ 360, 272, 48 },
		{ 360, 273, 48 },
		{ 360, 274, 48 },
		{ 360, 275, 48 },
		{ 360, 276, 48 },
		{ 360, 277, 48 },
		{ 360, 278, 48 },
		{ 360, 279, 48 },
		{ 360, 280, 48 },
		{ 360, 281, 48 },
		{ 360, 282, 48 },
		{ 360, 283, 48 },
		{ 360, 284, 48 },
		{ 360, 285, 48 },
		{ 360, 286, 48 },
		{ 360, 287, 48 },
		{ 360, 288, 48 },
		{ 360, 289, 48 },
		{ 360, 290, 48 },
		{ 360, 291, 48 },
		{ 360, 292, 48 },
		{ 360, 293, 48 },
		{ 360, 294, 48 },
		{ 360, 295, 48 },
		{ 361, 271, 48 },
		{ 361, 272, 48 },
		{ 361, 273, 48 },
		{ 361, 274, 48 },
		{ 361, 275, 48 },
		{ 361, 276, 48 },
		{ 361, 277, 48 },
		{ 361, 278, 48 },
		{ 361, 279, 48 },
		{ 361, 280, 48 },
		{ 361, 281, 48 },
		{ 361, 282, 48 },
		{ 361, 283, 48 },
		{ 361, 284, 48 },
		{ 361, 285, 48 },
		{ 361, 286, 48 },
		{ 361, 287, 48 },
		{ 361, 288, 48 },
		{ 361, 289, 48 },
		{ 361, 290, 48 },
		{ 361, 291, 48 },
		{ 361, 292, 48 },
		{ 361, 293, 48 },
		{ 361, 294, 48 },
		{ 361, 295, 48 },
		{ 362, 271, 48 },
		{ 362, 272, 48 },
		{ 362, 273, 48 },
		{ 362, 274, 48 },
		{ 362, 275, 48 },
		{ 362, 276, 48 },
		{ 362, 277, 48 },
		{ 362, 278, 48 },
		{ 362, 279, 48 },
		{ 362, 280, 48 },
		{ 362, 281, 48 },
		{ 362, 282, 48 },
		{ 362, 283, 48 },
		{ 362, 284, 48 },
		{ 362, 285, 48 },
		{ 362, 286, 48 },
		{ 362, 287, 48 },
		{ 362, 288, 48 },
		{ 362, 289, 48 },
		{ 362, 290, 48 },
		{ 362, 291, 48 },
		{ 362, 292, 48 },
		{ 362, 293, 48 },
		{ 362, 294, 48 },
		{ 362, 295, 48 },
		{ 362, 296, 48 },
		{ 363, 271, 48 },
		{ 363, 272, 48 },
		{ 363, 273, 48 },
		{ 363, 274, 48 },
		{ 363, 275, 48 },
		{ 363, 276, 48 },
		{ 363, 277, 48 },
		{ 363, 278, 48 },
		{ 363, 279, 48 },
		{ 363, 280, 48 },
		{ 363, 281, 48 },
		{ 363, 282, 48 },
		{ 363, 283, 48 },
		{ 363, 284, 48 },
		{ 363, 285, 48 },
		{ 363, 286, 48 },
		{ 363, 287, 48 },
		{ 363, 288, 48 },
		{ 363, 289, 48 },
		{ 363, 290, 48 },
		{ 363, 291, 48 },
		{ 363, 292, 48 },
		{ 363, 293, 48 },
		{ 363, 294, 48 },
		{ 363, 295, 48 },
		{ 363, 296, 48 },
		{ 364, 271, 48 },
		{ 364, 272, 48 },
		{ 364, 273, 48 },
		{ 364, 274, 48 },
		{ 364, 275, 48 },
		{ 364, 276, 48 },
		{ 364, 277, 48 },
		{ 364, 278, 48 },
		{ 364, 279, 48 },
		{ 364, 280, 48 },
		{ 364, 281, 48 },
		{ 364, 282, 48 },
		{ 364, 283, 48 },
		{ 364, 284, 48 },
		{ 364, 285, 48 },
		{ 364, 286, 48 },
		{ 364, 287, 48 },
		{ 364, 288, 48 },
		{ 364, 289, 48 },
		{ 364, 290, 48 },
		{ 364, 291, 48 },
		{ 364, 292, 48 },
		{ 364, 293, 48 },
		{ 364, 294, 48 },
		{ 364, 295, 48 },
		{ 364, 296, 48 },
		{ 364, 297, 48 },
		{ 365, 271, 48 },
		{ 365, 272, 48 },
		{ 365, 273, 48 },
		{ 365, 274, 48 },
		{ 365, 275, 48 },
		{ 365, 276, 48 },
		{ 365, 277, 48 },
		{ 365, 278, 48 },
		{ 365, 279, 48 },
		{ 365, 280, 48 },
		{ 365, 281, 48 },
		{ 365, 282, 48 },
		{ 365, 283, 48 },
		{ 365, 284, 48 },
		{ 365, 285, 48 },
		{ 365, 286, 48 },
		{ 365, 287, 48 },
		{ 365, 288, 48 },
		{ 365, 289, 48 },
		{ 365, 290, 48 },
		{ 365, 291, 48 },
		{ 365, 292, 48 },
		{ 365, 293, 48 },
		{ 365, 294, 48 },
		{ 365, 295, 48 },
		{ 365, 296, 48 },
		{ 365, 297, 48 },
		{ 366, 271, 48 },
		{ 366, 272, 48 },
		{ 366, 273, 48 },
		{ 366, 274, 48 },
		{ 366, 275, 48 },
		{ 366, 276, 48 },
		{ 366, 277, 48 },
		{ 366, 278, 48 },
		{ 366, 279, 48 },
		{ 366, 280, 48 },
		{ 366, 281, 48 },
		{ 366, 282, 48 },
		{ 366, 283, 48 },
		{ 366, 284, 48 },
		{ 366, 285, 48 },
		{ 366, 286, 48 },
		{ 366, 287, 48 },
		{ 366, 288, 48 },
		{ 366, 289, 48 },
		{ 366, 290, 48 },
		{ 366, 291, 48 },
		{ 366, 292, 48 },
		{ 366, 293, 48 },
		{ 366, 294, 48 },
		{ 366, 295, 48 },
		{ 366, 296, 48 },
		{ 366, 297, 48 },
		{ 367, 271, 48 },
		{ 367, 272, 48 },
		{ 367, 273, 48 },
		{ 367, 274, 48 },
		{ 367, 275, 48 },
		{ 367, 276, 48 },
		{ 367, 277, 48 },
		{ 367, 278, 48 },
		{ 367, 279, 48 },
		{ 367, 280, 48 },
		{ 367, 281, 48 },
		{ 367, 282, 48 },
		{ 367, 283, 48 },
		{ 367, 284, 48 },
		{ 367, 285, 48 },
		{ 367, 286, 48 },
		{ 367, 287, 48 },
		{ 367, 288, 48 },
		{ 367, 289, 48 },
		{ 367, 290, 48 },
		{ 367, 291, 48 },
		{ 367, 292, 48 },
		{ 367, 293, 48 },
		{ 367, 294, 48 },
		{ 367, 295, 48 },
		{ 367, 296, 48 },
		{ 367, 297, 48 },
		{ 367, 298, 48 },
		{ 368, 271, 48 },
		{ 368, 272, 48 },
		{ 368, 273, 48 },
		{ 368, 274, 48 },
		{ 368, 275, 48 },
		{ 368, 276, 48 },
		{ 368, 277, 48 },
		{ 368, 278, 48 },
		{ 368, 279, 48 },
		{ 368, 280, 48 },
		{ 368, 281, 48 },
		{ 368, 282, 48 },
		{ 368, 283, 48 },
		{ 368, 284, 48 },
		{ 368, 285, 48 },
		{ 368, 286, 48 },
		{ 368, 287, 48 },
		{ 368, 288, 48 },
		{ 368, 289, 48 },
		{ 368, 290, 48 },
		{ 368, 291, 48 },
		{ 368, 292, 48 },
		{ 368, 293, 48 },
		{ 368, 294, 48 },
		{ 368, 295, 48 },
		{ 368, 296, 48 },
		{ 368, 297, 48 },
		{ 368, 298, 48 },
		{ 369, 272, 48 },
		{ 369, 273, 48 },
		{ 369, 274, 48 },
		{ 369, 275, 48 },
		{ 369, 276, 48 },
		{ 369, 277, 48 },
		{ 369, 278, 48 },
		{ 369, 279, 48 },
		{ 369, 280, 48 },
		{ 369, 281, 48 },
		{ 369, 282, 48 },
		{ 369, 283, 48 },
		{ 369, 284, 48 },
		{ 369, 285, 48 },
		{ 369, 286, 48 },
		{ 369, 287, 48 },
		{ 369, 288, 48 },
		{ 369, 289, 48 },
		{ 369, 290, 48 },
		{ 369, 291, 48 },
		{ 369, 292, 48 },
		{ 369, 293, 48 },
		{ 369, 294, 48 },
		{ 369, 295, 48 },
		{ 369, 296, 48 },
		{ 369, 297, 48 },
		{ 369, 298, 48 },
		{ 370, 272, 48 },
		{ 370, 273, 48 },
		{ 370, 274, 48 },
		{ 370, 275, 48 },
		{ 370, 276, 48 },
		{ 370, 277, 48 },
		{ 370, 278, 48 },
		{ 370, 279, 48 },
		{ 370, 280, 48 },
		{ 370, 281, 48 },
		{ 370, 282, 48 },
		{ 370, 283, 48 },
		{ 370, 284, 48 },
		{ 370, 285, 48 },
		{ 370, 286, 48 },
		{ 370, 287, 48 },
		{ 370, 288, 48 },
		{ 370, 289, 48 },
		{ 370, 290, 48 },
		{ 370, 291, 48 },
		{ 370, 292, 48 },
		{ 370, 293, 48 },
		{ 370, 294, 48 },
		{ 370, 295, 48 },
		{ 370, 296, 48 },
		{ 370, 297, 48 },
		{ 370, 298, 48 },
		{ 370, 299, 48 },
		{ 371, 272, 48 },
		{ 371, 273, 48 },
		{ 371, 274, 48 },
		{ 371, 275, 48 },
		{ 371, 276, 48 },
		{ 371, 277, 48 },
		{ 371, 278, 48 },
		{ 371, 279, 48 },
		{ 371, 280, 48 },
		{ 371, 281, 48 },
		{ 371, 282, 48 },
		{ 371, 283, 48 },
		{ 371, 284, 48 },
		{ 371, 285, 48 },
		{ 371, 286, 48 },
		{ 371, 287, 48 },
		{ 371, 288, 48 },
		{ 371, 289, 48 },
		{ 371, 290, 48 },
		{ 371, 291, 48 },
		{ 371, 292, 48 },
		{ 371, 293, 48 },
		{ 371, 294, 48 },
		{ 371, 295, 48 },
		{ 371, 296, 48 },
		{ 371, 297, 48 },
		{ 371, 298, 48 },
		{ 371, 299, 48 },
		{ 372, 272, 48 },
		{ 372, 273, 48 },
		{ 372, 274, 48 },
		{ 372, 275, 48 },
		{ 372, 276, 48 },
		{ 372, 277, 48 },
		{ 372, 278, 48 },
		{ 372, 279, 48 },
		{ 372, 280, 48 },
		{ 372, 281, 48 },
		{ 372, 282, 48 },
		{ 372, 283, 48 },
		{ 372, 284, 48 },
		{ 372, 285, 48 },
		{ 372, 286, 48 },
		{ 372, 287, 48 },
		{ 372, 288, 48 },
		{ 372, 289, 48 },
		{ 372, 290, 48 },
		{ 372, 291, 48 },
		{ 372, 292, 48 },
		{ 372, 293, 48 },
		{ 372, 294, 48 },
		{ 372, 295, 48 },
		{ 372, 296, 48 },
		{ 372, 297, 48 },
		{ 372, 298, 48 },
		{ 372, 299, 48 },
		{ 372, 300, 48 },
		{ 373, 272, 48 },
		{ 373, 273, 48 },
		{ 373, 274, 48 },
		{ 373, 275, 48 },
		{ 373, 276, 48 },
		{ 373, 277, 48 },
		{ 373, 278, 48 },
		{ 373, 279, 48 },
		{ 373, 280, 48 },
		{ 373, 281, 48 },
		{ 373, 282, 48 },
		{ 373, 283, 48 },
		{ 373, 284, 48 },
		{ 373, 285, 48 },
		{ 373, 286, 48 },
		{ 373, 287, 48 },
		{ 373, 288, 48 },
		{ 373, 289, 48 },
		{ 373, 290, 48 },
		{ 373, 291, 48 },
		{ 373, 292, 48 },
		{ 373, 293, 48 },
		{ 373, 294, 48 },
		{ 373, 295, 48 },
		{ 373, 296, 48 },
		{ 373, 297, 48 },
		{ 373, 298, 48 },
		{ 373, 299, 48 },
		{ 373, 300, 48 },
		{ 374, 272, 48 },
		{ 374, 273, 48 },
		{ 374, 274, 48 },
		{ 374, 275, 48 },
		{ 374, 276, 48 },
		{ 374, 277, 48 },
		{ 374, 278, 48 },
		{ 374, 279, 48 },
		{ 374, 280, 48 },
		{ 374, 281, 48 },
		{ 374, 282, 48 },
		{ 374, 283, 48 },
		{ 374, 284, 48 },
		{ 374, 285, 48 },
		{ 374, 286, 48 },
		{ 374, 287, 48 },
		{ 374, 288, 48 },
		{ 374, 289, 48 },
		{ 374, 290, 48 },
		{ 374, 291, 48 },
		{ 374, 292, 48 },
		{ 374, 293, 48 },
		{ 374, 294, 48 },
		{ 374, 295, 48 },
		{ 374, 296, 48 },
		{ 374, 297, 48 },
		{ 374, 298, 48 },
		{ 374, 299, 48 },
		{ 374, 300, 48 },
		{ 375, 272, 48 },
		{ 375, 273, 48 },
		{ 375, 274, 48 },
		{ 375, 275, 48 },
		{ 375, 276, 48 },
		{ 375, 277, 48 },
		{ 375, 278, 48 },
		{ 375, 279, 48 },
		{ 375, 280, 48 },
		{ 375, 281, 48 },
		{ 375, 282, 48 },
		{ 375, 283, 48 },
		{ 375, 284, 48 },
		{ 375, 285, 48 },
		{ 375, 286, 48 },
		{ 375, 287, 48 },
		{ 375, 288, 48 },
		{ 375, 289, 48 },
		{ 375, 290, 48 },
		{ 375, 291, 48 },
		{ 375, 292, 48 },
		{ 375, 293, 48 },
		{ 375, 294, 48 },
		{ 375, 295, 48 },
		{ 375, 296, 48 },
		{ 375, 297, 48 },
		{ 375, 298, 48 },
		{ 375, 299, 48 },
		{ 375, 300, 48 },
		{ 375, 301, 48 },
		{ 376, 272, 48 },
		{ 376, 273, 48 },
		{ 376, 274, 48 },
		{ 376, 275, 48 },
		{ 376, 276, 48 },
		{ 376, 277, 48 },
		{ 376, 278, 48 },
		{ 376, 279, 48 },
		{ 376, 280, 48 },
		{ 376, 281, 48 },
		{ 376, 282, 48 },
		{ 376, 283, 48 },
		{ 376, 284, 48 },
		{ 376, 285, 48 },
		{ 376, 286, 48 },
		{ 376, 287, 48 },
		{ 376, 288, 48 },
		{ 376, 289, 48 },
		{ 376, 290, 48 },
		{ 376, 291, 48 },
		{ 376, 292, 48 },
		{ 376, 293, 48 },
		{ 376, 294, 48 },
		{ 376, 295, 48 },
		{ 376, 296, 48 },
		{ 376, 297, 48 },
		{ 376, 298, 48 },
		{ 376, 299, 48 },
		{ 376, 300, 48 },
		{ 376, 301, 48 },
		{ 377, 273, 48 },
		{ 377, 274, 48 },
		{ 377, 275, 48 },
		{ 377, 276, 48 },
		{ 377, 277, 48 },
		{ 377, 278, 48 },
		{ 377, 279, 48 },
		{ 377, 280, 48 },
		{ 377, 281, 48 },
		{ 377, 282, 48 },
		{ 377, 283, 48 },
		{ 377, 284, 48 },
		{ 377, 285, 48 },
		{ 377, 286, 48 },
		{ 377, 287, 48 },
		{ 377, 288, 48 },
		{ 377, 289, 48 },
		{ 377, 290, 48 },
		{ 377, 291, 48 },
		{ 377, 292, 48 },
		{ 377, 293, 48 },
		{ 377, 294, 48 },
		{ 377, 295, 48 },
		{ 377, 296, 48 },
		{ 377, 297, 48 },
		{ 377, 298, 48 },
		{ 377, 299, 48 },
		{ 377, 300, 48 },
		{ 377, 301, 48 },
		{ 378, 273, 48 },
		{ 378, 274, 48 },
		{ 378, 275, 48 },
		{ 378, 276, 48 },
		{ 378, 277, 48 },
		{ 378, 278, 48 },
		{ 378, 279, 48 },
		{ 378, 280, 48 },
		{ 378, 281, 48 },
		{ 378, 282, 48 },
		{ 378, 283, 48 },
		{ 378, 284, 48 },
		{ 378, 285, 48 },
		{ 378, 286, 48 },
		{ 378, 287, 48 },
		{ 378, 288, 48 },
		{ 378, 289, 48 },
		{ 378, 290, 48 },
		{ 378, 291, 48 },
		{ 378, 292, 48 },
		{ 378, 293, 48 },
		{ 378, 294, 48 },
		{ 378, 295, 48 },
		{ 378, 296, 48 },
		{ 378, 297, 48 },
		{ 378, 298, 48 },
		{ 378, 299, 48 },
		{ 378, 300, 48 },
		{ 378, 301, 48 },
		{ 378, 302, 48 },
		{ 379, 273, 48 },
		{ 379, 274, 48 },
		{ 379, 275, 48 },
		{ 379, 276, 48 },
		{ 379, 277, 48 },
		{ 379, 278, 48 },
		{ 379, 279, 48 },
		{ 379, 280, 48 },
		{ 379, 281, 48 },
		{ 379, 282, 48 },
		{ 379, 283, 48 },
		{ 379, 284, 48 },
		{ 379, 285, 48 },
		{ 379, 286, 48 },
		{ 379, 287, 48 },
		{ 379, 288, 48 },
		{ 379, 289, 48 },
		{ 379, 290, 48 },
		{ 379, 291, 48 },
		{ 379, 292, 48 },
		{ 379, 293, 48 },
		{ 379, 294, 48 },
		{ 379, 295, 48 },
		{ 379, 296, 48 },
		{ 379, 297, 48 },
		{ 379, 298, 48 },
		{ 379, 299, 48 },
		{ 379, 300, 48 },
		{ 379, 301, 48 },
		{ 379, 302, 48 },
		{ 380, 273, 48 },
		{ 380, 274, 48 },
		{ 380, 275, 48 },
		{ 380, 276, 48 },
		{ 380, 277, 48 },
		{ 380, 278, 48 },
		{ 380, 279, 48 },
		{ 380, 280, 48 },
		{ 380, 281, 48 },
		{ 380, 282, 48 },
		{ 380, 283, 48 },
		{ 380, 284, 48 },
		{ 380, 285, 48 },
		{ 380, 286, 48 },
		{ 380, 287, 48 },
		{ 380, 288, 48 },
		{ 380, 289, 48 },
		{ 380, 290, 48 },
		{ 380, 291, 48 },
		{ 380, 292, 48 },
		{ 380, 293, 48 },
		{ 380, 294, 48 },
		{ 380, 295, 48 },
		{ 380, 296, 48 },
		{ 380, 297, 48 },
		{ 380, 298, 48 },
		{ 380, 299, 48 },
		{ 380, 300, 48 },
		{ 380, 301, 48 },
		{ 380, 302, 48 },
		{ 380, 303, 48 },
		{ 381, 273, 48 },
		{ 381, 274, 48 },
		{ 381, 275, 48 },
		{ 381, 276, 48 },
		{ 381, 277, 48 },
		{ 381, 278, 48 },
		{ 381, 279, 48 },
		{ 381, 280, 48 },
		{ 381, 281, 48 },
		{ 381, 282, 48 },
		{ 381, 283, 48 },
		{ 381, 284, 48 },
		{ 381, 285, 48 },
		{ 381, 286, 48 },
		{ 381, 287, 48 },
		{ 381, 288, 48 },
		{ 381, 289, 48 },
		{ 381, 290, 48 },
		{ 381, 291, 48 },
		{ 381, 292, 48 },
		{ 381, 293, 48 },
		{ 381, 294, 48 },
		{ 381, 295, 48 },
		{ 381, 296, 48 },
		{ 381, 297, 48 },
		{ 381, 298, 48 },
		{ 381, 299, 48 },
		{ 381, 300, 48 },
		{ 381, 301, 48 },
		{ 381, 302, 48 },
		{ 381, 303, 48 },
		{ 382, 273, 48 },
		{ 382, 274, 48 },
		{ 382, 275, 48 },
		{ 382, 276, 48 },
		{ 382, 277, 48 },
		{ 382, 278, 48 },
		{ 382, 279, 48 },
		{ 382, 280, 48 },
		{ 382, 281, 48 },
		{ 382, 282, 48 },
		{ 382, 283, 48 },
		{ 382, 284, 48 },
		{ 382, 285, 48 },
		{ 382, 286, 48 },
		{ 382, 287, 48 },
		{ 382, 288, 48 },
		{ 382, 289, 48 },
		{ 382, 290, 48 },
		{ 382, 291, 48 },
		{ 382, 292, 48 },
		{ 382, 293, 48 },
		{ 382, 294, 48 },
		{ 382, 295, 48 },
		{ 382, 296, 48 },
		{ 382, 297, 48 },
		{ 382, 298, 48 },
		{ 382, 299, 48 },
		{ 382, 300, 48 },
		{ 382, 301, 48 },
		{ 382, 302, 48 },
		{ 382, 303, 48 },
		{ 383, 273, 48 },
		{ 383, 274, 48 },
		{ 383, 275, 48 },
		{ 383, 276, 48 },
		{ 383, 277, 48 },
		{ 383, 278, 48 },
		{ 383, 279, 48 },
		{ 383, 280, 48 },
		{ 383, 281, 48 },
		{ 383, 282, 48 },
		{ 383, 283, 48 },
		{ 383, 284, 48 },
		{ 383, 285, 48 },
		{ 383, 286, 48 },
		{ 383, 287, 48 },
		{ 383, 288, 48 },
		{ 383, 289, 48 },
		{ 383, 290, 48 },
		{ 383, 291, 48 },
		{ 383, 292, 48 },
		{ 383, 293, 48 },
		{ 383, 294, 48 },
		{ 383, 295, 48 },
		{ 383, 296, 48 },
		{ 383, 297, 48 },
		{ 383, 298, 48 },
		{ 383, 299, 48 },
		{ 383, 300, 48 },
		{ 383, 301, 48 },
		{ 383, 302, 48 },
		{ 383, 303, 48 },
		{ 383, 304, 48 },
		{ 384, 273, 48 },
		{ 384, 274, 48 },
		{ 384, 275, 48 },
		{ 384, 276, 48 },
		{ 384, 277, 48 },
		{ 384, 278, 48 },
		{ 384, 279, 48 },
		{ 384, 280, 48 },
		{ 384, 281, 48 },
		{ 384, 282, 48 },
		{ 384, 283, 48 },
		{ 384, 284, 48 },
		{ 384, 285, 48 },
		{ 384, 286, 48 },
		{ 384, 287, 48 },
		{ 384, 288, 48 },
		{ 384, 289, 48 },
		{ 384, 290, 48 },
		{ 384, 291, 48 },
		{ 384, 292, 48 },
		{ 384, 293, 48 },
		{ 384, 294, 48 },
		{ 384, 295, 48 },
		{ 384, 296, 48 },
		{ 384, 297, 48 },
		{ 384, 298, 48 },
		{ 384, 299, 48 },
		{ 384, 300, 48 },
		{ 384, 301, 48 },
		{ 384, 302, 48 },
		{ 384, 303, 48 },
		{ 384, 304, 48 },
		{ 385, 274, 48 },
		{ 385, 275, 48 },
		{ 385, 276, 48 },
		{ 385, 277, 48 },
		{ 385, 278, 48 },
		{ 385, 279, 48 },
		{ 385, 280, 48 },
		{ 385, 281, 48 },
		{ 385, 282, 48 },
		{ 385, 283, 48 },
		{ 385, 284, 48 },
		{ 385, 285, 48 },
		{ 385, 286, 48 },
		{ 385, 287, 48 },
		{ 385, 288, 48 },
		{ 385, 289, 48 },
		{ 385, 290, 48 },
		{ 385, 291, 48 },
		{ 385, 292, 48 },
		{ 385, 293, 48 },
		{ 385, 294, 48 },
		{ 385, 295, 48 },
		{ 385, 296, 48 },
		{ 385, 297, 48 },
		{ 385, 298, 48 },
		{ 385, 299, 48 },
		{ 385, 300, 48 },
		{ 385, 301, 48 },
		{ 385, 302, 48 },
		{ 385, 303, 48 },
		{ 385, 304, 48 },
		{ 386, 274, 48 },
		{ 386, 275, 48 },
		{ 386, 276, 48 },
		{ 386, 277, 48 },
		{ 386, 278, 48 },
		{ 386, 279, 48 },
		{ 386, 280, 48 },
		{ 386, 281, 48 },
		{ 386, 282, 48 },
		{ 386, 283, 48 },
		{ 386, 284, 48 },
		{ 386, 285, 48 },
		{ 386, 286, 48 },
		{ 386, 287, 48 },
		{ 386, 288, 48 },
		{ 386, 289, 48 },
		{ 386, 290, 48 },
		{ 386, 291, 48 },
		{ 386, 292, 48 },
		{ 386, 293, 48 },
		{ 386, 294, 48 },
		{ 386, 295, 48 },
		{ 386, 296, 48 },
		{ 386, 297, 48 },
		{ 386, 298, 48 },
		{ 386, 299, 48 },
		{ 386, 300, 48 },
		{ 386, 301, 48 },
		{ 386, 302, 48 },
		{ 386, 303, 48 },
		{ 386, 304, 48 },
		{ 386, 305, 48 },
		{ 387, 274, 48 },
		{ 387, 275, 48 },
		{ 387, 276, 48 },
		{ 387, 277, 48 },
		{ 387, 278, 48 },
		{ 387, 279, 48 },
		{ 387, 280, 48 },
		{ 387, 281, 48 },
		{ 387, 282, 48 },
		{ 387, 283, 48 },
		{ 387, 284, 48 },
		{ 387, 285, 48 },
		{ 387, 286, 48 },
		{ 387, 287, 48 },
		{ 387, 288, 48 },
		{ 387, 289, 48 },
		{ 387, 290, 48 },
		{ 387, 291, 48 },
		{ 387, 292, 48 },
		{ 387, 293, 48 },
		{ 387, 294, 48 },
		{ 387, 295, 48 },
		{ 387, 296, 48 },
		{ 387, 297, 48 },
		{ 387, 298, 48 },
		{ 387, 299, 48 },
		{ 387, 300, 48 },
		{ 387, 301, 48 },
		{ 387, 302, 48 },
		{ 387, 303, 48 },
		{ 387, 304, 48 },
		{ 387, 305, 48 },
		{ 388, 274, 48 },
		{ 388, 275, 48 },
		{ 388, 276, 48 },
		{ 388, 277, 48 },
		{ 388, 278, 48 },
		{ 388, 279, 48 },
		{ 388, 280, 48 },
		{ 388, 281, 48 },
		{ 388, 282, 48 },
		{ 388, 283, 48 },
		{ 388, 284, 48 },
		{ 388, 285, 48 },
		{ 388, 286, 48 },
		{ 388, 287, 48 },
		{ 388, 288, 48 },
		{ 388, 289, 48 },
		{ 388, 290, 48 },
		{ 388, 291, 48 },
		{ 388, 292, 48 },
		{ 388, 293, 48 },
		{ 388, 294, 48 },
		{ 388, 295, 48 },
		{ 388, 296, 48 },
		{ 388, 297, 48 },
		{ 388, 298, 48 },
		{ 388, 299, 48 },
		{ 388, 300, 48 },
		{ 388, 301, 48 },
		{ 388, 302, 48 },
		{ 388, 303, 48 },
		{ 388, 304, 48 },
		{ 388, 305, 48 },
		{ 388, 306, 48 },
		{ 389, 274, 48 },
		{ 389, 275, 48 },
		{ 389, 276, 48 },
		{ 389, 277, 48 },
		{ 389, 278, 48 },
		{ 389, 279, 48 },
		{ 389, 280, 48 },
		{ 389, 281, 48 },
		{ 389, 282, 48 },
		{ 389, 283, 48 },
		{ 389, 284, 48 },
		{ 389, 285, 48 },
		{ 389, 286, 48 },
		{ 389, 287, 48 },
		{ 389, 288, 48 },
		{ 389, 289, 48 },
		{ 389, 290, 48 },
		{ 389, 291, 48 },
		{ 389, 292, 48 },
		{ 389, 293, 48 },
		{ 389, 294, 48 },
		{ 389, 295, 48 },
		{ 389, 296, 48 },
		{ 389, 297, 48 },
		{ 389, 298, 48 },
		{ 389, 299, 48 },
		{ 389, 300, 48 },
		{ 389, 301, 48 },
		{ 389, 302, 48 },
		{ 389, 303, 48 },
		{ 389, 304, 48 },
		{ 389, 305, 48 },
		{ 389, 306, 48 },
		{ 390, 274, 48 },
		{ 390, 275, 48 },
		{ 390, 276, 48 },
		{ 390, 277, 48 },
		{ 390, 278, 48 },
		{ 390, 279, 48 },
		{ 390, 280, 48 },
		{ 390, 281, 48 },
		{ 390, 282, 48 },
		{ 390, 283, 48 },
		{ 390, 284, 48 },
		{ 390, 285, 48 },
		{ 390, 286, 48 },
		{ 390, 287, 48 },
		{ 390, 288, 48 },
		{ 390, 289, 48 },
		{ 390, 290, 48 },
		{ 390, 291, 48 },
		{ 390, 292, 48 },
		{ 390, 293, 48 },
		{ 390, 294, 48 },
		{ 390, 295, 48 },
		{ 390, 296, 48 },
		{ 390, 297, 48 },
		{ 390, 298, 48 },
		{ 390, 299, 48 },
		{ 390, 300, 48 },
		{ 390, 301, 48 },
		{ 390, 302, 48 },
		{ 390, 303, 48 },
		{ 390, 304, 48 },
		{ 390, 305, 48 },
		{ 390, 306, 48 },
		{ 391, 274, 48 },
		{ 391, 275, 48 },
		{ 391, 276, 48 },
		{ 391, 277, 48 },
		{ 391, 278, 48 },
		{ 391, 279, 48 },
		{ 391, 280, 48 },
		{ 391, 281, 48 },
		{ 391, 282, 48 },
		{ 391, 283, 48 },
		{ 391, 284, 48 },
		{ 391, 285, 48 },
		{ 391, 286, 48 },
		{ 391, 287, 48 },
		{ 391, 288, 48 },
		{ 391, 289, 48 },
		{ 391, 290, 48 },
		{ 391, 291, 48 },
		{ 391, 292, 48 },
		{ 391, 293, 48 },
		{ 391, 294, 48 },
		{ 391, 295, 48 },
		{ 391, 296, 48 },
		{ 391, 297, 48 },
		{ 391, 298, 48 },
		{ 391, 299, 48 },
		{ 391, 300, 48 },
		{ 391, 301, 48 },
		{ 391, 302, 48 },
		{ 391, 303, 48 },
		{ 391, 304, 48 },
		{ 391, 305, 48 },
		{ 391, 306, 48 },
		{ 391, 307, 48 },
		{ 392, 274, 48 },
		{ 392, 275, 48 },
		{ 392, 276, 48 },
		{ 392, 277, 48 },
		{ 392, 278, 48 },
		{ 392, 279, 48 },
		{ 392, 280, 48 },
		{ 392, 281, 48 },
		{ 392, 282, 48 },
		{ 392, 283, 48 },
		{ 392, 284, 48 },
		{ 392, 285, 48 },
		{ 392, 286, 48 },
		{ 392, 287, 48 },
		{ 392, 288, 48 },
		{ 392, 289, 48 },
		{ 392, 290, 48 },
		{ 392, 291, 48 },
		{ 392, 292, 48 },
		{ 392, 293, 48 },
		{ 392, 294, 48 },
		{ 392, 295, 48 },
		{ 392, 296, 48 },
		{ 392, 297, 48 },
		{ 392, 298, 48 },
		{ 392, 299, 48 },
		{ 392, 300, 48 },
		{ 392, 301, 48 },
		{ 392, 302, 48 },
		{ 392, 303, 48 },
		{ 392, 304, 48 },
		{ 392, 305, 48 },
		{ 392, 306, 48 },
		{ 392, 307, 48 },
		{ 393, 275, 48 },
		{ 393, 276, 48 },
		{ 393, 277, 48 },
		{ 393, 278, 48 },
		{ 393, 279, 48 },
		{ 393, 280, 48 },
		{ 393, 281, 48 },
		{ 393, 282, 48 },
		{ 393, 283, 48 },
		{ 393, 284, 48 },
		{ 393, 285, 48 },
		{ 393, 286, 48 },
		{ 393, 287, 48 },
		{ 393, 288, 48 },
		{ 393, 289, 48 },
		{ 393, 290, 48 },
		{ 393, 291, 48 },
		{ 393, 292, 48 },
		{ 393, 293, 48 },
		{ 393, 294, 48 },
		{ 393, 295, 48 },
		{ 393, 296, 48 },
		{ 393, 297, 48 },
		{ 393, 298, 48 },
		{ 393, 299, 48 },
		{ 393, 300, 48 },
		{ 393, 301, 48 },
		{ 393, 302, 48 },
		{ 393, 303, 48 },
		{ 393, 304, 48 },
		{ 393, 305, 48 },
		{ 393, 306, 48 },
		{ 393, 307, 48 },
		{ 394, 275, 48 },
		{ 394, 276, 48 },
		{ 394, 277, 48 },
		{ 394, 278, 48 },
		{ 394, 279, 48 },
		{ 394, 280, 48 },
		{ 394, 281, 48 },
		{ 394, 282, 48 },
		{ 394, 283, 48 },
		{ 394, 284, 48 },
		{ 394, 285, 48 },
		{ 394, 286, 48 },
		{ 394, 287, 48 },
		{ 394, 288, 48 },
		{ 394, 289, 48 },
		{ 394, 290, 48 },
		{ 394, 291, 48 },
		{ 394, 292, 48 },
		{ 394, 293, 48 },
		{ 394, 294, 48 },
		{ 394, 295, 48 },
		{ 394, 296, 48 },
		{ 394, 297, 48 },
		{ 394, 298, 48 },
		{ 394, 299, 48 },
		{ 394, 300, 48 },
		{ 394, 301, 48 },
		{ 394, 302, 48 },
		{ 394, 303, 48 },
		{ 394, 304, 48 },
		{ 394, 305, 48 },
		{ 394, 306, 48 },
		{ 394, 307, 48 },
		{ 394, 308, 48 },
		{ 395, 275, 48 },
		{ 395, 276, 48 },
		{ 395, 277, 48 },
		{ 395, 278, 48 },
		{ 395, 279, 48 },
		{ 395, 280, 48 },
		{ 395, 281, 48 },
		{ 395, 282, 48 },
		{ 395, 283, 48 },
		{ 395, 284, 48 },
		{ 395, 285, 48 },
		{ 395, 286, 48 },
		{ 395, 287, 48 },
		{ 395, 288, 48 },
		{ 395, 289, 48 },
		{ 395, 290, 48 },
		{ 395, 291, 48 },
		{ 395, 292, 48 },
		{ 395, 293, 48 },
		{ 395, 294, 48 },
		{ 395, 295, 48 },
		{ 395, 296, 48 },
		{ 395, 297, 48 },
		{ 395, 298, 48 },
		{ 395, 299, 48 },
		{ 395, 300, 48 },
		{ 395, 301, 48 },
		{ 395, 302, 48 },
		{ 395, 303, 48 },
		{ 395, 304, 48 },
		{ 395, 305, 48 },
		{ 395, 306, 48 },
		{ 395, 307, 48 },
		{ 395, 308, 48 },
		{ 396, 275, 48 },
		{ 396, 276, 48 },
		{ 396, 277, 48 },
		{ 396, 278, 48 },
		{ 396, 279, 48 },
		{ 396, 280, 48 },
		{ 396, 281, 48 },
		{ 396, 282, 48 },
		{ 396, 283, 48 },
		{ 396, 284, 48 },
		{ 396, 285, 48 },
		{ 396, 286, 48 },
		{ 396, 287, 48 },
		{ 396, 288, 48 },
		{ 396, 289, 48 },
		{ 396, 290, 48 },
		{ 396, 291, 48 },
		{ 396, 292, 48 },
		{ 396, 293, 48 },
		{ 396, 294, 48 },
		{ 396, 295, 48 },
		{ 396, 296, 48 },
		{ 396, 297, 48 },
		{ 396, 298, 48 },
		{ 396, 299, 48 },
		{ 396, 300, 48 },
		{ 396, 301, 48 },
		{ 396, 302, 48 },
		{ 396, 303, 48 },
		{ 396, 304, 48 },
		{ 396, 305, 48 },
		{ 396, 306, 48 },
		{ 396, 307, 48 },
		{ 396, 308, 48 },
		{ 396, 309, 48 },
		{ 397, 275, 48 },
		{ 397, 276, 48 },
		{ 397, 277, 48 },
		{ 397, 278, 48 },
		{ 397, 279, 48 },
		{ 397, 280, 48 },
		{ 397, 281, 48 },
		{ 397, 282, 48 },
		{ 397, 283, 48 },
		{ 397, 284, 48 },
		{ 397, 285, 48 },
		{ 397, 286, 48 },
		{ 397, 287, 48 },
		{ 397, 288, 48 },
		{ 397, 289, 48 },
		{ 397, 290, 48 },
		{ 397, 291, 48 },
		{ 397, 292, 48 },
		{ 397, 293, 48 },
		{ 397, 294, 48 },
		{ 397, 295, 48 },
		{ 397, 296, 48 },
		{ 397, 297, 48 },
		{ 397, 298, 48 },
		{ 397, 299, 48 },
		{ 397, 300, 48 },
		{ 397, 301, 48 },
		{ 397, 302, 48 },
		{ 397, 303, 48 },
		{ 397, 304, 48 },
		{ 397, 305, 48 },
		{ 397, 306, 48 },
		{ 397, 307, 48 },
		{ 397, 308, 48 },
		{ 397, 309, 48 },
		{ 398, 275, 48 },
		{ 398, 276, 48 },
		{ 398, 277, 48 },
		{ 398, 278, 48 },
		{ 398, 279, 48 },
		{ 398, 280, 48 },
		{ 398, 281, 48 },
		{ 398, 282, 48 },
		{ 398, 283, 48 },
		{ 398, 284, 48 },
		{ 398, 285, 48 },
		{ 398, 286, 48 },
		{ 398, 287, 48 },
		{ 398, 288, 48 },
		{ 398, 289, 48 },
		{ 398, 290, 48 },
		{ 398, 291, 48 },
		{ 398, 292, 48 },
		{ 398, 293, 48 },
		{ 398, 294, 48 },
		{ 398, 295, 48 },
		{ 398, 296, 48 },
		{ 398, 297, 48 },
		{ 398, 298, 48 },
		{ 398, 299, 48 },
		{ 398, 300, 48 },
		{ 398, 301, 48 },
		{ 398, 302, 48 },
		{ 398, 303, 48 },
		{ 398, 304, 48 },
		{ 398, 305, 48 },
		{ 398, 306, 48 },
		{ 398, 307, 48 },
		{ 398, 308, 48 },
		{ 398, 309, 48 },
		{ 399, 275, 48 },
		{ 399, 276, 48 },
		{ 399, 277, 48 },
		{ 399, 278, 48 },
		{ 399, 279, 48 },
		{ 399, 280, 48 },
		{ 399, 281, 48 },
		{ 399, 282, 48 },
		{ 399, 283, 48 },
		{ 399, 284, 48 },
		{ 399, 285, 48 },
		{ 399, 286, 48 },
		{ 399, 287, 48 },
		{ 399, 288, 48 },
		{ 399, 289, 48 },
		{ 399, 290, 48 },
		{ 399, 291, 48 },
		{ 399, 292, 48 },
		{ 399, 293, 48 },
		{ 399, 294, 48 },
		{ 399, 295, 48 },
		{ 399, 296, 48 },
		{ 399, 297, 48 },
		{ 399, 298, 48 },
		{ 399, 299, 48 },
		{ 399, 300, 48 },
		{ 399, 301, 48 },
		{ 399, 302, 48 },
		{ 399, 303, 48 },
		{ 399, 304, 48 },
		{ 399, 305, 48 },
		{ 399, 306, 48 },
		{ 399, 307, 48 },
		{ 399, 308, 48 },
		{ 399, 309, 48 },
		{ 399, 310, 48 },
		{ 400, 275, 48 },
		{ 400, 276, 48 },
		{ 400, 277, 48 },
		{ 400, 278, 48 },
		{ 400, 279, 48 },
		{ 400, 280, 48 },
		{ 400, 281, 48 },
		{ 400, 282, 48 },
		{ 400, 283, 48 },
		{ 400, 284, 48 },
		{ 400, 285, 48 },
		{ 400, 286, 48 },
		{ 400, 287, 48 },
		{ 400, 288, 48 },
		{ 400, 289, 48 },
		{ 400, 290, 48 },
		{ 400, 291, 48 },
		{ 400, 292, 48 },
		{ 400, 293, 48 },
		{ 400, 294, 48 },
		{ 400, 295, 48 },
		{ 400, 296, 48 },
		{ 400, 297, 48 },
		{ 400, 298, 48 },
		{ 400, 299, 48 },
		{ 400, 300, 48 },
		{ 400, 301, 48 },
		{ 400, 302, 48 },
		{ 400, 303, 48 },
		{ 400, 304, 48 },
		{ 400, 305, 48 },
		{ 400, 306, 48 },
		{ 400, 307, 48 },
		{ 400, 308, 48 },
		{ 400, 309, 48 },
		{ 400, 310, 48 },
		{ 401, 276, 48 },
		{ 401, 277, 48 },
		{ 401, 278, 48 },
		{ 401, 279, 48 },
		{ 401, 280, 48 },
		{ 401, 281, 48 },
		{ 401, 282, 48 },
		{ 401, 283, 48 },
		{ 401, 284, 48 },
		{ 401, 285, 48 },
		{ 401, 286, 48 },
		{ 401, 287, 48 },
		{ 401, 288, 48 },
		{ 401, 289, 48 },
		{ 401, 290, 48 },
		{ 401, 291, 48 },
		{ 401, 292, 48 },
		{ 401, 293, 48 },
		{ 401, 294, 48 },
		{ 401, 295, 48 },
		{ 401, 296, 48 },
		{ 401, 297, 48 },
		{ 401, 298, 48 },
		{ 401, 299, 48 },
		{ 401, 300, 48 },
		{ 401, 301, 48 },
		{ 401, 302, 48 },
		{ 401, 303, 48 },
		{ 401, 304, 48 },
		{ 401, 305, 48 },
		{ 401, 306, 48 },
		{ 401, 307, 48 },
		{ 401, 308, 48 },
		{ 401, 309, 48 },
		{ 401, 310, 48 },
		{ 402, 276, 48 },
		{ 402, 277, 48 },
		{ 402, 278, 48 },
		{ 402, 279, 48 },
		{ 402, 280, 48 },
		{ 402, 281, 48 },
		{ 402, 282, 48 },
		{ 402, 283, 48 },
		{ 402, 284, 48 },
		{ 402, 285, 48 },
		{ 402, 286, 48 },
		{ 402, 287, 48 },
		{ 402, 288, 48 },
		{ 402, 289, 48 },
		{ 402, 290, 48 },
		{ 402, 291, 48 },
		{ 402, 292, 48 },
		{ 402, 293, 48 },
		{ 402, 294, 48 },
		{ 402, 295, 48 },
		{ 402, 296, 48 },
		{ 402, 297, 48 },
		{ 402, 298, 48 },
		{ 402, 299, 48 },
		{ 402, 300, 48 },
		{ 402, 301, 48 },
		{ 402, 302, 48 },
		{ 402, 303, 48 },
		{ 402, 304, 48 },
		{ 402, 305, 48 },
		{ 402, 306, 48 },
		{ 402, 307, 48 },
		{ 402, 308, 48 },
		{ 402, 309, 48 },
		{ 402, 310, 48 },
		{ 402, 311, 48 },
		{ 403, 276, 48 },
		{ 403, 277, 48 },
		{ 403, 278, 48 },
		{ 403, 279, 48 },
		{ 403, 280, 48 },
		{ 403, 281, 48 },
		{ 403, 282, 48 },
		{ 403, 283, 48 },
		{ 403, 284, 48 },
		{ 403, 285, 48 },
		{ 403, 286, 48 },
		{ 403, 287, 48 },
		{ 403, 288, 48 },
		{ 403, 289, 48 },
		{ 403, 290, 48 },
		{ 403, 291, 48 },
		{ 403, 292, 48 },
		{ 403, 293, 48 },
		{ 403, 294, 48 },
		{ 403, 295, 48 },
		{ 403, 296, 48 },
		{ 403, 297, 48 },
		{ 403, 298, 48 },
		{ 403, 299, 48 },
		{ 403, 300, 48 },
		{ 403, 301, 48 },
		{ 403, 302, 48 },
		{ 403, 303, 48 },
		{ 403, 304, 48 },
		{ 403, 305, 48 },
		{ 403, 306, 48 },
		{ 403, 307, 48 },
		{ 403, 308, 48 },
		{ 403, 309, 48 },
		{ 403, 310, 48 },
		{ 403, 311, 48 },
		{ 404, 276, 48 },
		{ 404, 277, 48 },
		{ 404, 278, 48 },
		{ 404, 279, 48 },
		{ 404, 280, 48 },
		{ 404, 281, 48 },
		{ 404, 282, 48 },
		{ 404, 283, 48 },
		{ 404, 284, 48 },
		{ 404, 285, 48 },
		{ 404, 286, 48 },
		{ 404, 287, 48 },
		{ 404, 288, 48 },
		{ 404, 289, 48 },
		{ 404, 290, 48 },
		{ 404, 291, 48 },
		{ 404, 292, 48 },
		{ 404, 293, 48 },
		{ 404, 294, 48 },
		{ 404, 295, 48 },
		{ 404, 296, 48 },
		{ 404, 297, 48 },
		{ 404, 298, 48 },
		{ 404, 299, 48 },
		{ 404, 300, 48 },
		{ 404, 301, 48 },
		{ 404, 302, 48 },
		{ 404, 303, 48 },
		{ 404, 304, 48 },
		{ 404, 305, 48 },
		{ 404, 306, 48 },
		{ 404, 307, 48 },
		{ 404, 308, 48 },
		{ 404, 309, 48 },
		{ 404, 310, 48 },
		{ 404, 311, 48 },
		{ 404, 312, 48 },
		{ 405, 276, 48 },
		{ 405, 277, 48 },
		{ 405, 278, 48 },
		{ 405, 279, 48 },
		{ 405, 280, 48 },
		{ 405, 281, 48 },
		{ 405, 282, 48 },
		{ 405, 283, 48 },
		{ 405, 284, 48 },
		{ 405, 285, 48 },
		{ 405, 286, 48 },
		{ 405, 287, 48 },
		{ 405, 288, 48 },
		{ 405, 289, 48 },
		{ 405, 290, 48 },
		{ 405, 291, 48 },
		{ 405, 292, 48 },
		{ 405, 293, 48 },
		{ 405, 294, 48 },
		{ 405, 295, 48 },
		{ 405, 296, 48 },
		{ 405, 297, 48 },
		{ 405, 298, 48 },
		{ 405, 299, 48 },
		{ 405, 300, 48 },
		{ 405, 301, 48 },
		{ 405, 302, 48 },
		{ 405, 303, 48 },
		{ 405, 304, 48 },
		{ 405, 305, 48 },
		{ 405, 306, 48 },
		{ 405, 307, 48 },
		{ 405, 308, 48 },
		{ 405, 309, 48 },
		{ 405, 310, 48 },
		{ 405, 311, 48 },
		{ 405, 312, 48 },
		{ 406, 276, 48 },
		{ 406, 277, 48 },
		{ 406, 278, 48 },
		{ 406, 279, 48 },
		{ 406, 280, 48 },
		{ 406, 281, 48 },
		{ 406, 282, 48 },
		{ 406, 283, 48 },
		{ 406, 284, 48 },
		{ 406, 285, 48 },
		{ 406, 286, 48 },
		{ 406, 287, 48 },
		{ 406, 288, 48 },
		{ 406, 289, 48 },
		{ 406, 290, 48 },
		{ 406, 291, 48 },
		{ 406, 292, 48 },
		{ 406, 293, 48 },
		{ 406, 294, 48 },
		{ 406, 295, 48 },
		{ 406, 296, 48 },
		{ 406, 297, 48 },
		{ 406, 298, 48 },
		{ 406, 299, 48 },
		{ 406, 300, 48 },
		{ 406, 301, 48 },
		{ 406, 302, 48 },
		{ 406, 303, 48 },
		{ 406, 304, 48 },
		{ 406, 305, 48 },
		{ 406, 306, 48 },
		{ 406, 307, 48 },
		{ 406, 308, 48 },
		{ 406, 309, 48 },
		{ 406, 310, 48 },
		{ 406, 311, 48 },
		{ 406, 312, 48 },
		{ 407, 276, 48 },
		{ 407, 277, 48 },
		{ 407, 278, 48 },
		{ 407, 279, 48 },
		{ 407, 280, 48 },
		{ 407, 281, 48 },
		{ 407, 282, 48 },
		{ 407, 283, 48 },
		{ 407, 284, 48 },
		{ 407, 285, 48 },
		{ 407, 286, 48 },
		{ 407, 287, 48 },
		{ 407, 288, 48 },
		{ 407, 289, 48 },
		{ 407, 290, 48 },
		{ 407, 291, 48 },
		{ 407, 292, 48 },
		{ 407, 293, 48 },
		{ 407, 294, 48 },
		{ 407, 295, 48 },
		{ 407, 296, 48 },
		{ 407, 297, 48 },
		{ 407, 298, 48 },
		{ 407, 299, 48 },
		{ 407, 300, 48 },
		{ 407, 301, 48 },
		{ 407, 302, 48 },
		{ 407, 303, 48 },
		{ 407, 304, 48 },
		{ 407, 305, 48 },
		{ 407, 306, 48 },
		{ 407, 307, 48 },
		{ 407, 308, 48 },
		{ 407, 309, 48 },
		{ 407, 310, 48 },
		{ 407, 311, 48 },
		{ 407, 312, 48 },
		{ 407, 313, 48 },
		{ 408, 276, 48 },
		{ 408, 277, 48 },
		{ 408, 278, 48 },
		{ 408, 279, 48 },
		{ 408, 280, 48 },
		{ 408, 281, 48 },
		{ 408, 282, 48 },
		{ 408, 283, 48 },
		{ 408, 284, 48 },
		{ 408, 285, 48 },
		{ 408, 286, 48 },
		{ 408, 287, 48 },
		{ 408, 288, 48 },
		{ 408, 289, 48 },
		{ 408, 290, 48 },
		{ 408, 291, 48 },
		{ 408, 292, 48 },
		{ 408, 293, 48 },
		{ 408, 294, 48 },
		{ 408, 295, 48 },
		{ 408, 296, 48 },
		{ 408, 297, 48 },
		{ 408, 298, 48 },
		{ 408, 299, 48 },
		{ 408, 300, 48 },
		{ 408, 301, 48 },
		{ 408, 302, 48 },
		{ 408, 303, 48 },
		{ 408, 304, 48 },
		{ 408, 305, 48 },
		{ 408, 306, 48 },
		{ 408, 307, 48 },
		{ 408, 308, 48 },
		{ 408, 309, 48 },
		{ 408, 310, 48 },
		{ 408, 311, 48 },
		{ 408, 312, 48 },
		{ 408, 313, 48 },
		{ 409, 277, 48 },
		{ 409, 278, 48 },
		{ 409, 279, 48 },
		{ 409, 280, 48 },
		{ 409, 281, 48 },
		{ 409, 282, 48 },
		{ 409, 283, 48 },
		{ 409, 284, 48 },
		{ 409, 285, 48 },
		{ 409, 286, 48 },
		{ 409, 287, 48 },
		{ 409, 288, 48 },
		{ 409, 289, 48 },
		{ 409, 290, 48 },
		{ 409, 291, 48 },
		{ 409, 292, 48 },
		{ 409, 293, 48 },
		{ 409, 294, 48 },
		{ 409, 295, 48 },
		{ 409, 296, 48 },
		{ 409, 297, 48 },
		{ 409, 298, 48 },
		{ 409, 299, 48 },
		{ 409, 300, 48 },
		{ 409, 301, 48 },
		{ 409, 302, 48 },
		{ 409, 303, 48 },
		{ 409, 304, 48 },
		{ 409, 305, 48 },
		{ 409, 306, 48 },
		{ 409, 307, 48 },
		{ 409, 308, 48 },
		{ 409, 309, 48 },
		{ 409, 310, 48 },
		{ 409, 311, 48 },
		{ 409, 312, 48 },
		{ 409, 313, 48 },
		{ 410, 277, 48 },
		{ 410, 278, 48 },
		{ 410, 279, 48 },
		{ 410, 280, 48 },
		{ 410, 281, 48 },
		{ 410, 282, 48 },
		{ 410, 283, 48 },
		{ 410, 284, 48 },
		{ 410, 285, 48 },
		{ 410, 286, 48 },
		{ 410, 287, 48 },
		{ 410, 288, 48 },
		{ 410, 289, 48 },
		{ 410, 290, 48 },
		{ 410, 291, 48 },
		{ 410, 292, 48 },
		{ 410, 293, 48 },
		{ 410, 294, 48 },
		{ 410, 295, 48 },
		{ 410, 296, 48 },
		{ 410, 297, 48 },
		{ 410, 298, 48 },
		{ 410, 299, 48 },
		{ 410, 300, 48 },
		{ 410, 301, 48 },
		{ 410, 302, 48 },
		{ 410, 303, 48 },
		{ 410, 304, 48 },
		{ 410, 305, 48 },
		{ 410, 306, 48 },
		{ 410, 307, 48 },
		{ 410, 308, 48 },
		{ 410, 309, 48 },
		{ 410, 310, 48 },
		{ 410, 311, 48 },
		{ 410, 312, 48 },
		{ 410, 313, 48 },
		{ 410, 314, 48 },
		{ 411, 277, 48 },
		{ 411, 278, 48 },
		{ 411, 279, 48 },
		{ 411, 280, 48 },
		{ 411, 281, 48 },
		{ 411, 282, 48 },
		{ 411, 283, 48 },
		{ 411, 284, 48 },
		{ 411, 285, 48 },
		{ 411, 286, 48 },
		{ 411, 287, 48 },
		{ 411, 288, 48 },
		{ 411, 289, 48 },
		{ 411, 290, 48 },
		{ 411, 291, 48 },
		{ 411, 292, 48 },
		{ 411, 293, 48 },
		{ 411, 294, 48 },
		{ 411, 295, 48 },
		{ 411, 296, 48 },
		{ 411, 297, 48 },
		{ 411, 298, 48 },
		{ 411, 299, 48 },
		{ 411, 300, 48 },
		{ 411, 301, 48 },
		{ 411, 302, 48 },
		{ 411, 303, 48 },
		{ 411, 304, 48 },
		{ 411, 305, 48 },
		{ 411, 306, 48 },
		{ 411, 307, 48 },
		{ 411, 308, 48 },
		{ 411, 309, 48 },
		{ 411, 310, 48 },
		{ 411, 311, 48 },
		{ 411, 312, 48 },
		{ 411, 313, 48 },
		{ 411, 314, 48 },
		{ 412, 277, 48 },
		{ 412, 278, 48 },
		{ 412, 279, 48 },
		{ 412, 280, 48 },
		{ 412, 281, 48 },
		{ 412, 282, 48 },
		{ 412, 283, 48 },
		{ 412, 284, 48 },
		{ 412, 285, 48 },
		{ 412, 286, 48 },
		{ 412, 287, 48 },
		{ 412, 288, 48 },
		{ 412, 289, 48 },
		{ 412, 290, 48 },
		{ 412, 291, 48 },
		{ 412, 292, 48 },
		{ 412, 293, 48 },
		{ 412, 294, 48 },
		{ 412, 295, 48 },
		{ 412, 296, 48 },
		{ 412, 297, 48 },
		{ 412, 298, 48 },
		{ 412, 299, 48 },
		{ 412, 300, 48 },
		{ 412, 301, 48 },
		{ 412, 302, 48 },
		{ 412, 303, 48 },
		{ 412, 304, 48 },
		{ 412, 305, 48 },
		{ 412, 306, 48 },
		{ 412, 307, 48 },
		{ 412, 308, 48 },
		{ 412, 309, 48 },
		{ 412, 310, 48 },
		{ 412, 311, 48 },
		{ 412, 312, 48 },
		{ 412, 313, 48 },
		{ 412, 314, 48 },
		{ 412, 315, 48 },
		{ 413, 277, 48 },
		{ 413, 278, 48 },
		{ 413, 279, 48 },
		{ 413, 280, 48 },
		{ 413, 281, 48 },
		{ 413, 282, 48 },
		{ 413, 283, 48 },
		{ 413, 284, 48 },
		{ 413, 285, 48 },
		{ 413, 286, 48 },
		{ 413, 287, 48 },
		{ 413, 288, 48 },
		{ 413, 289, 48 },
		{ 413, 290, 48 },
		{ 413, 291, 48 },
		{ 413, 292, 48 },
		{ 413, 293, 48 },
		{ 413, 294, 48 },
		{ 413, 295, 48 },
		{ 413, 296, 48 },
		{ 413, 297, 48 },
		{ 413, 298, 48 },
		{ 413, 299, 48 },
		{ 413, 300, 48 },
		{ 413, 301, 48 },
		{ 413, 302, 48 },
		{ 413, 303, 48 },
		{ 413, 304, 48 },
		{ 413, 305, 48 },
		{ 413, 306, 48 },
		{ 413, 307, 48 },
		{ 413, 308, 48 },
		{ 413, 309, 48 },
		{ 413, 310, 48 },
		{ 413, 311, 48 },
		{ 413, 312, 48 },
		{ 413, 313, 48 },
		{ 413, 314, 48 },
		{ 413, 315, 48 },
		{ 414, 277, 48 },
		{ 414, 278, 48 },
		{ 414, 279, 48 },
		{ 414, 280, 48 },
		{ 414, 281, 48 },
		{ 414, 282, 48 },
		{ 414, 283, 48 },
		{ 414, 284, 48 },
		{ 414, 285, 48 },
		{ 414, 286, 48 },
		{ 414, 287, 48 },
		{ 414, 288, 48 },
		{ 414, 289, 48 },
		{ 414, 290, 48 },
		{ 414, 291, 48 },
		{ 414, 292, 48 },
		{ 414, 293, 48 },
		{ 414, 294, 48 },
		{ 414, 295, 48 },
		{ 414, 296, 48 },
		{ 414, 297, 48 },
		{ 414, 298, 48 },
		{ 414, 299, 48 },
		{ 414, 300, 48 },
		{ 414, 301, 48 },
		{ 414, 302, 48 },
		{ 414, 303, 48 },
		{ 414, 304, 48 },
		{ 414, 305, 48 },
		{ 414, 306, 48 },
		{ 414, 307, 48 },
		{ 414, 308, 48 },
		{ 414, 309, 48 },
		{ 414, 310, 48 },
		{ 414, 311, 48 },
		{ 414, 312, 48 },
		{ 414, 313, 48 },
		{ 414, 314, 48 },
		{ 414, 315, 48 },
		{ 415, 277, 48 },
		{ 415, 278, 48 },
		{ 415, 279, 48 },
		{ 415, 280, 48 },
		{ 415, 281, 48 },
		{ 415, 282, 48 },
		{ 415, 283, 48 },
		{ 415, 284, 48 },
		{ 415, 285, 48 },
		{ 415, 286, 48 },
		{ 415, 287, 48 },
		{ 415, 288, 48 },
		{ 415, 289, 48 },
		{ 415, 290, 48 },
		{ 415, 291, 48 },
		{ 415, 292, 48 },
		{ 415, 293, 48 },
		{ 415, 294, 48 },
		{ 415, 295, 48 },
		{ 415, 296, 48 },
		{ 415, 297, 48 },
		{ 415, 298, 48 },
		{ 415, 299, 48 },
		{ 415, 300, 48 },
		{ 415, 301, 48 },
		{ 415, 302, 48 },
		{ 415, 303, 48 },
		{ 415, 304, 48 },
		{ 415, 305, 48 },
		{ 415, 306, 48 },
		{ 415, 307, 48 },
		{ 415, 308, 48 },
		{ 415, 309, 48 },
		{ 415, 310, 48 },
		{ 415, 311, 48 },
		{ 415, 312, 48 },
		{ 415, 313, 48 },
		{ 415, 314, 48 },
		{ 415, 315, 48 },
		{ 415, 316, 48 },
		{ 416, 277, 48 },
		{ 416, 278, 48 },
		{ 416, 279, 48 },
		{ 416, 280, 48 },
		{ 416, 281, 48 },
		{ 416, 282, 48 },
		{ 416, 283, 48 },
		{ 416, 284, 48 },
		{ 416, 285, 48 },
		{ 416, 286, 48 },
		{ 416, 287, 48 },
		{ 416, 288, 48 },
		{ 416, 289, 48 },
		{ 416, 290, 48 },
		{ 416, 291, 48 },
		{ 416, 292, 48 },
		{ 416, 293, 48 },
		{ 416, 294, 48 },
		{ 416, 295, 48 },
		{ 416, 296, 48 },
		{ 416, 297, 48 },
		{ 416, 298, 48 },
		{ 416, 299, 48 },
		{ 416, 300, 48 },
		{ 416, 301, 48 },
		{ 416, 302, 48 },
		{ 416, 303, 48 },
		{ 416, 304, 48 },
		{ 416, 305, 48 },
		{ 416, 306, 48 },
		{ 416, 307, 48 },
		{ 416, 308, 48 },
		{ 416, 309, 48 },
		{ 416, 310, 48 },
		{ 416, 311, 48 },
		{ 416, 312, 48 },
		{ 416, 313, 48 },
		{ 416, 314, 48 },
		{ 416, 315, 48 },
		{ 416, 316, 48 },
		{ 417, 278, 48 },
		{ 417, 279, 48 },
		{ 417, 280, 48 },
		{ 417, 281, 48 },
		{ 417, 282, 48 },
		{ 417, 283, 48 },
		{ 417, 284, 48 },
		{ 417, 285, 48 },
		{ 417, 286, 48 },
		{ 417, 287, 48 },
		{ 417, 288, 48 },
		{ 417, 289, 48 },
		{ 417, 290, 48 },
		{ 417, 291, 48 },
		{ 417, 292, 48 },
		{ 417, 293, 48 },
		{ 417, 294, 48 },
		{ 417, 295, 48 },
		{ 417, 296, 48 },
		{ 417, 297, 48 },
		{ 417, 298, 48 },
		{ 417, 299, 48 },
		{ 417, 300, 48 },
		{ 417, 301, 48 },
		{ 417, 302, 48 },
		{ 417, 303, 48 },
		{ 417, 304, 48 },
		{ 417, 305, 48 },
		{ 417, 306, 48 },
		{ 417, 307, 48 },
		{ 417, 308, 48 },
		{ 417, 309, 48 },
		{ 417, 310, 48 },
		{ 417, 311, 48 },
		{ 417, 312, 48 },
		{ 417, 313, 48 },
		{ 417, 314, 48 },
		{ 417, 315, 48 },
		{ 417, 316, 48 },
		{ 418, 278, 48 },
		{ 418, 279, 48 },
		{ 418, 280, 48 },
		{ 418, 281, 48 },
		{ 418, 282, 48 },
		{ 418, 283, 48 },
		{ 418, 284, 48 },
		{ 418, 285, 48 },
		{ 418, 286, 48 },
		{ 418, 287, 48 },
		{ 418, 288, 48 },
		{ 418, 289, 48 },
		{ 418, 290, 48 },
		{ 418, 291, 48 },
		{ 418, 292, 48 },
		{ 418, 293, 48 },
		{ 418, 294, 48 },
		{ 418, 295, 48 },
		{ 418, 296, 48 },
		{ 418, 297, 48 },
		{ 418, 298, 48 },
		{ 418, 299, 48 },
		{ 418, 300, 48 },
		{ 418, 301, 48 },
		{ 418, 302, 48 },
		{ 418, 303, 48 },
		{ 418, 304, 48 },
		{ 418, 305, 48 },
		{ 418, 306, 48 },
		{ 418, 307, 48 },
		{ 418, 308, 48 },
		{ 418, 309, 48 },
		{ 418, 310, 48 },
		{ 418, 311, 48 },
		{ 418, 312, 48 },
		{ 418, 313, 48 },
		{ 418, 314, 48 },
		{ 418, 315, 48 },
		{ 418, 316, 48 },
		{ 418, 317, 48 },
		{ 419, 278, 48 },
		{ 419, 279, 48 },
		{ 419, 280, 48 },
		{ 419, 281, 48 },
		{ 419, 282, 48 },
		{ 419, 283, 48 },
		{ 419, 284, 48 },
		{ 419, 285, 48 },
		{ 419, 286, 48 },
		{ 419, 287, 48 },
		{ 419, 288, 48 },
		{ 419, 289, 48 },
		{ 419, 290, 48 },
		{ 419, 291, 48 },
		{ 419, 292, 48 },
		{ 419, 293, 48 },
		{ 419, 294, 48 },
		{ 419, 295, 48 },
		{ 419, 296, 48 },
		{ 419, 297, 48 },
		{ 419, 298, 48 },
		{ 419, 299, 48 },
		{ 419, 300, 48 },
		{ 419, 301, 48 },
		{ 419, 302, 48 },
		{ 419, 303, 48 },
		{ 419, 304, 48 },
		{ 419, 305, 48 },
		{ 419, 306, 48 },
		{ 419, 307, 48 },
		{ 419, 308, 48 },
		{ 419, 309, 48 },
		{ 419, 310, 48 },
		{ 419, 311, 48 },
		{ 419, 312, 48 },
		{ 419, 313, 48 },
		{ 419, 314, 48 },
		{ 419, 315, 48 },
		{ 419, 316, 48 },
		{ 419, 317, 48 },
		{ 420, 278, 48 },
		{ 420, 279, 48 },
		{ 420, 280, 48 },
		{ 420, 281, 48 },
		{ 420, 282, 48 },
		{ 420, 283, 48 },
		{ 420, 284, 48 },
		{ 420, 285, 48 },
		{ 420, 286, 48 },
		{ 420, 287, 48 },
		{ 420, 288, 48 },
		{ 420, 289, 48 },
		{ 420, 290, 48 },
		{ 420, 291, 48 },
		{ 420, 292, 48 },
		{ 420, 293, 48 },
		{ 420, 294, 48 },
		{ 420, 295, 48 },
		{ 420, 296, 48 },
		{ 420, 297, 48 },
		{ 420, 298, 48 },
		{ 420, 299, 48 },
		{ 420, 300, 48 },
		{ 420, 301, 48 },
		{ 420, 302, 48 },
		{ 420, 303, 48 },
		{ 420, 304, 48 },
		{ 420, 305, 48 },
		{ 420, 306, 48 },
		{ 420, 307, 48 },
		{ 420, 308, 48 },
		{ 420, 309, 48 },
		{ 420, 310, 48 },
		{ 420, 311, 48 },
		{ 420, 312, 48 },
		{ 420, 313, 48 },
		{ 420, 314, 48 },
		{ 420, 315, 48 },
		{ 420, 316, 48 },
		{ 420, 317, 48 },
		{ 420, 318, 48 },
		{ 421, 278, 48 },
		{ 421, 279, 48 },
		{ 421, 280, 48 },
		{ 421, 281, 48 },
		{ 421, 282, 48 },
		{ 421, 283, 48 },
		{ 421, 284, 48 },
		{ 421, 285, 48 },
		{ 421, 286, 48 },
		{ 421, 287, 48 },
		{ 421, 288, 48 },
		{ 421, 289, 48 },
		{ 421, 290, 48 },
		{ 421, 291, 48 },
		{ 421, 292, 48 },
		{ 421, 293, 48 },
		{ 421, 294, 48 },
		{ 421, 295, 48 },
		{ 421, 296, 48 },
		{ 421, 297, 48 },
		{ 421, 298, 48 },
		{ 421, 299, 48 },
		{ 421, 300, 48 },
		{ 421, 301, 48 },
		{ 421, 302, 48 },
		{ 421, 303, 48 },
		{ 421, 304, 48 },
		{ 421, 305, 48 },
		{ 421, 306, 48 },
		{ 421, 307, 48 },
		{ 421, 308, 48 },
		{ 421, 309, 48 },
		{ 421, 310, 48 },
		{ 421, 311, 48 },
		{ 421, 312, 48 },
		{ 421, 313, 48 },
		{ 421, 314, 48 },
		{ 421, 315, 48 },
		{ 421, 316, 48 },
		{ 421, 317, 48 },
		{ 421, 318, 48 },
		{ 422, 278, 48 },
		{ 422, 279, 48 },
		{ 422, 280, 48 },
		{ 422, 281, 48 },
		{ 422, 282, 48 },
		{ 422, 283, 48 },
		{ 422, 284, 48 },
		{ 422, 285, 48 },
		{ 422, 286, 48 },
		{ 422, 287, 48 },
		{ 422, 288, 48 },
		{ 422, 289, 48 },
		{ 422, 290, 48 },
		{ 422, 291, 48 },
		{ 422, 292, 48 },
		{ 422, 293, 48 },
		{ 422, 294, 48 },
		{ 422, 295, 48 },
		{ 422, 296, 48 },
		{ 422, 297, 48 },
		{ 422, 298, 48 },
		{ 422, 299, 48 },
		{ 422, 300, 48 },
		{ 422, 301, 48 },
		{ 422, 302, 48 },
		{ 422, 303, 48 },
		{ 422, 304, 48 },
		{ 422, 305, 48 },
		{ 422, 306, 48 },
		{ 422, 307, 48 },
		{ 422, 308, 48 },
		{ 422, 309, 48 },
		{ 422, 310, 48 },
		{ 422, 311, 48 },
		{ 422, 312, 48 },
		{ 422, 313, 48 },
		{ 422, 314, 48 },
		{ 422, 315, 48 },
		{ 422, 316, 48 },
		{ 422, 317, 48 },
		{ 422, 318, 48 },
		{ 423, 278, 48 },
		{ 423, 279, 48 },
		{ 423, 280, 48 },
		{ 423, 281, 48 },
		{ 423, 282, 48 },
		{ 423, 283, 48 },
		{ 423, 284, 48 },
		{ 423, 285, 48 },
		{ 423, 286, 48 },
		{ 423, 287, 48 },
		{ 423, 288, 48 },
		{ 423, 289, 48 },
		{ 423, 290, 48 },
		{ 423, 291, 48 },
		{ 423, 292, 48 },
		{ 423, 293, 48 },
		{ 423, 294, 48 },
		{ 423, 295, 48 },
		{ 423, 296, 48 },
		{ 423, 297, 48 },
		{ 423, 298, 48 },
		{ 423, 299, 48 },
		{ 423, 300, 48 },
		{ 423, 301, 48 },
		{ 423, 302, 48 },
		{ 423, 303, 48 },
		{ 423, 304, 48 },
		{ 423, 305, 48 },
		{ 423, 306, 48 },
		{ 423, 307, 48 },
		{ 423, 308, 48 },
		{ 423, 309, 48 },
		{ 423, 310, 48 },
		{ 423, 311, 48 },
		{ 423, 312, 48 },
		{ 423, 313, 48 },
		{ 423, 314, 48 },
		{ 423, 315, 48 },
		{ 423, 316, 48 },
		{ 423, 317, 48 },
		{ 423, 318, 48 },
		{ 423, 319, 48 },
		{ 424, 278, 48 },
		{ 424, 279, 48 },
		{ 424, 280, 48 },
		{ 424, 281, 48 },
		{ 424, 282, 48 },
		{ 424, 283, 48 },
		{ 424, 284, 48 },
		{ 424, 285, 48 },
		{ 424, 286, 48 },
		{ 424, 287, 48 },
		{ 424, 288, 48 },
		{ 424, 289, 48 },
		{ 424, 290, 48 },
		{ 424, 291, 48 },
		{ 424, 292, 48 },
		{ 424, 293, 48 },
		{ 424, 294, 48 },
		{ 424, 295, 48 },
		{ 424, 296, 48 },
		{ 424, 297, 48 },
		{ 424, 298, 48 },
		{ 424, 299, 48 },
		{ 424, 300, 48 },
		{ 424, 301, 48 },
		{ 424, 302, 48 },
		{ 424, 303, 48 },
		{ 424, 304, 48 },
		{ 424, 305, 48 },
		{ 424, 306, 48 },
		{ 424, 307, 48 },
		{ 424, 308, 48 },
		{ 424, 309, 48 },
		{ 424, 310, 48 },
		{ 424, 311, 48 },
		{ 424, 312, 48 },
		{ 424, 313, 48 },
		{ 424, 314, 48 },
		{ 424, 315, 48 },
		{ 424, 316, 48 },
		{ 424, 317, 48 },
		{ 424, 318, 48 },
		{ 424, 319, 48 },
		{ 425, 279, 48 },
		{ 425, 280, 48 },
		{ 425, 281, 48 },
		{ 425, 282, 48 },
		{ 425, 283, 48 },
		{ 425, 284, 48 },
		{ 425, 285, 48 },
		{ 425, 286, 48 },
		{ 425, 287, 48 },
		{ 425, 288, 48 },
		{ 425, 289, 48 },
		{ 425, 290, 48 },
		{ 425, 291, 48 },
		{ 425, 292, 48 },
		{ 425, 293, 48 },
		{ 425, 294, 48 },
		{ 425, 295, 48 },
		{ 425, 296, 48 },
		{ 425, 297, 48 },
		{ 425, 298, 48 },
		{ 425, 299, 48 },
		{ 425, 300, 48 },
		{ 425, 301, 48 },
		{ 425, 302, 48 },
		{ 425, 303, 48 },
		{ 425, 304, 48 },
		{ 425, 305, 48 },
		{ 425, 306, 48 },
		{ 425, 307, 48 },
		{ 425, 308, 48 },
		{ 425, 309, 48 },
		{ 425, 310, 48 },
		{ 425, 311, 48 },
		{ 425, 312, 48 },
		{ 425, 313, 48 },
		{ 425, 314, 48 },
		{ 425, 315, 48 },
		{ 425, 316, 48 },
		{ 425, 317, 48 },
		{ 425, 318, 48 },
		{ 425, 319, 48 },
		{ 426, 279, 48 },
		{ 426, 280, 48 },
		{ 426, 281, 48 },
		{ 426, 282, 48 },
		{ 426, 283, 48 },
		{ 426, 284, 48 },
		{ 426, 285, 48 },
		{ 426, 286, 48 },
		{ 426, 287, 48 },
		{ 426, 288, 48 },
		{ 426, 289, 48 },
		{ 426, 290, 48 },
		{ 426, 291, 48 },
		{ 426, 292, 48 },
		{ 426, 293, 48 },
		{ 426, 294, 48 },
		{ 426, 295, 48 },
		{ 426, 296, 48 },
		{ 426, 297, 48 },
		{ 426, 298, 48 },
		{ 426, 299, 48 },
		{ 426, 300, 48 },
		{ 426, 301, 48 },
		{ 426, 302, 48 },
		{ 426, 303, 48 },
		{ 426, 304, 48 },
		{ 426, 305, 48 },
		{ 426, 306, 48 },
		{ 426, 307, 48 },
		{ 426, 308, 48 },
		{ 426, 309, 48 },
		{ 426, 310, 48 },
		{ 426, 311, 48 },
		{ 426, 312, 48 },
		{ 426, 313, 48 },
		{ 426, 314, 48 },
		{ 426, 315, 48 },
		{ 426, 316, 48 },
		{ 426, 317, 48 },
		{ 426, 318, 48 },
		{ 426, 319, 48 },
		{ 426, 320, 48 },
		{ 427, 279, 48 },
		{ 427, 280, 48 },
		{ 427, 281, 48 },
		{ 427, 282, 48 },
		{ 427, 283, 48 },
		{ 427, 284, 48 },
		{ 427, 285, 48 },
		{ 427, 286, 48 },
		{ 427, 287, 48 },
		{ 427, 288, 48 },
		{ 427, 289, 48 },
		{ 427, 290, 48 },
		{ 427, 291, 48 },
		{ 427, 292, 48 },
		{ 427, 293, 48 },
		{ 427, 294, 48 },
		{ 427, 295, 48 },
		{ 427, 296, 48 },
		{ 427, 297, 48 },
		{ 427, 298, 48 },
		{ 427, 299, 48 },
		{ 427, 300, 48 },
		{ 427, 301, 48 },
		{ 427, 302, 48 },
		{ 427, 303, 48 },
		{ 427, 304, 48 },
		{ 427, 305, 48 },
		{ 427, 306, 48 },
		{ 427, 307, 48 },
		{ 427, 308, 48 },
		{ 427, 309, 48 },
		{ 427, 310, 48 },
		{ 427, 311, 48 },
		{ 427, 312, 48 },
		{ 427, 313, 48 },
		{ 427, 314, 48 },
		{ 427, 315, 48 },
		{ 427, 316, 48 },
		{ 427, 317, 48 },
		{ 427, 318, 48 },
		{ 427, 319, 48 },
		{ 427, 320, 48 },
		{ 428, 279, 48 },
		{ 428, 280, 48 },
		{ 428, 281, 48 },
		{ 428, 282, 48 },
		{ 428, 283, 48 },
		{ 428, 284, 48 },
		{ 428, 285, 48 },
		{ 428, 286, 48 },
		{ 428, 287, 48 },
		{ 428, 288, 48 },
		{ 428, 289, 48 },
		{ 428, 290, 48 },
		{ 428, 291, 48 },
		{ 428, 292, 48 },
		{ 428, 293, 48 },
		{ 428, 294, 48 },
		{ 428, 295, 48 },
		{ 428, 296, 48 },
		{ 428, 297, 48 },
		{ 428, 298, 48 },
		{ 428, 299, 48 },
		{ 428, 300, 48 },
		{ 428, 301, 48 },
		{ 428, 302, 48 },
		{ 428, 303, 48 },
		{ 428, 304, 48 },
		{ 428, 305, 48 },
		{ 428, 306, 48 },
		{ 428, 307, 48 },
		{ 428, 308, 48 },
		{ 428, 309, 48 },
		{ 428, 310, 48 },
		{ 428, 311, 48 },
		{ 428, 312, 48 },
		{ 428, 313, 48 },
		{ 428, 314, 48 },
		{ 428, 315, 48 },
		{ 428, 316, 48 },
		{ 428, 317, 48 },
		{ 428, 318, 48 },
		{ 428, 319, 48 },
		{ 428, 320, 48 },
		{ 428, 321, 48 },
		{ 429, 279, 48 },
		{ 429, 280, 48 },
		{ 429, 281, 48 },
		{ 429, 282, 48 },
		{ 429, 283, 48 },
		{ 429, 284, 48 },
		{ 429, 285, 48 },
		{ 429, 286, 48 },
		{ 429, 287, 48 },
		{ 429, 288, 48 },
		{ 429, 289, 48 },
		{ 429, 290, 48 },
		{ 429, 291, 48 },
		{ 429, 292, 48 },
		{ 429, 293, 48 },
		{ 429, 294, 48 },
		{ 429, 295, 48 },
		{ 429, 296, 48 },
		{ 429, 297, 48 },
		{ 429, 298, 48 },
		{ 429, 299, 48 },
		{ 429, 300, 48 },
		{ 429, 301, 48 },
		{ 429, 302, 48 },
		{ 429, 303, 48 },
		{ 429, 304, 48 },
		{ 429, 305, 48 },
		{ 429, 306, 48 },
		{ 429, 307, 48 },
		{ 429, 308, 48 },
		{ 429, 309, 48 },
		{ 429, 310, 48 },
		{ 429, 311, 48 },
		{ 429, 312, 48 },
		{ 429, 313, 48 },
		{ 429, 314, 48 },
		{ 429, 315, 48 },
		{ 429, 316, 48 },
		{ 429, 317, 48 },
		{ 429, 318, 48 },
		{ 429, 319, 48 },
		{ 429, 320, 48 },
		{ 429, 321, 48 },
		{ 430, 279, 48 },
		{ 430, 280, 48 },
		{ 430, 281, 48 },
		{ 430, 282, 48 },
		{ 430, 283, 48 },
		{ 430, 284, 48 },
		{ 430, 285, 48 },
		{ 430, 286, 48 },
		{ 430, 287, 48 },
		{ 430, 288, 48 },
		{ 430, 289, 48 },
		{ 430, 290, 48 },
		{ 430, 291, 48 },
		{ 430, 292, 48 },
		{ 430, 293, 48 },
		{ 430, 294, 48 },
		{ 430, 295, 48 },
		{ 430, 296, 48 },
		{ 430, 297, 48 },
		{ 430, 298, 48 },
		{ 430, 299, 48 },
		{ 430, 300, 48 },
		{ 430, 301, 48 },
		{ 430, 302, 48 },
		{ 430, 303, 48 },
		{ 430, 304, 48 },
		{ 430, 305, 48 },
		{ 430, 306, 48 },
		{ 430, 307, 48 },
		{ 430, 308, 48 },
		{ 430, 309, 48 },
		{ 430, 310, 48 },
		{ 430, 311, 48 },
		{ 430, 312, 48 },
		{ 430, 313, 48 },
		{ 430, 314, 48 },
		{ 430, 315, 48 },
		{ 430, 316, 48 },
		{ 430, 317, 48 },
		{ 430, 318, 48 },
		{ 430, 319, 48 },
		{ 430, 320, 48 },
		{ 430, 321, 48 },
		{ 431, 279, 48 },
		{ 431, 280, 48 },
		{ 431, 281, 48 },
		{ 431, 282, 48 },
		{ 431, 283, 48 },
		{ 431, 284, 48 },
		{ 431, 285, 48 },
		{ 431, 286, 48 },
		{ 431, 287, 48 },
		{ 431, 288, 48 },
		{ 431, 289, 48 },
		{ 431, 290, 48 },
		{ 431, 291, 48 },
		{ 431, 292, 48 },
		{ 431, 293, 48 },
		{ 431, 294, 48 },
		{ 431, 295, 48 },
		{ 431, 296, 48 },
		{ 431, 297, 48 },
		{ 431, 298, 48 },
		{ 431, 299, 48 },
		{ 431, 300, 48 },
		{ 431, 301, 48 },
		{ 431, 302, 48 },
		{ 431, 303, 48 },
		{ 431, 304, 48 },
		{ 431, 305, 48 },
		{ 431, 306, 48 },
		{ 431, 307, 48 },
		{ 431, 308, 48 },
		{ 431, 309, 48 },
		{ 431, 310, 48 },
		{ 431, 311, 48 },
		{ 431, 312, 48 },
		{ 431, 313, 48 },
		{ 431, 314, 48 },
		{ 431, 315, 48 },
		{ 431, 316, 48 },
		{ 431, 317, 48 },
		{ 431, 318, 48 },
		{ 431, 319, 48 },
		{ 431, 320, 48 },
		{ 431, 321, 48 },
		{ 431, 322, 48 },
		{ 432, 279, 48 },
		{ 432, 280, 48 },
		{ 432, 281, 48 },
		{ 432, 282, 48 },
		{ 432, 283, 48 },
		{ 432, 284, 48 },
		{ 432, 285, 48 },
		{ 432, 286, 48 },
		{ 432, 287, 48 },
		{ 432, 288, 48 },
		{ 432, 289, 48 },
		{ 432, 290, 48 },
		{ 432, 291, 48 },
		{ 432, 292, 48 },
		{ 432, 293, 48 },
		{ 432, 294, 48 },
		{ 432, 295, 48 },
		{ 432, 296, 48 },
		{ 432, 297, 48 },
		{ 432, 298, 48 },
		{ 432, 299, 48 },
		{ 432, 300, 48 },
		{ 432, 301, 48 },
		{ 432, 302, 48 },
		{ 432, 303, 48 },
		{ 432, 304, 48 },
		{ 432, 305, 48 },
		{ 432, 306, 48 },
		{ 432, 307, 48 },
		{ 432, 308, 48 },
		{ 432, 309, 48 },
		{ 432, 310, 48 },
		{ 432, 311, 48 },
		{ 432, 312, 48 },
		{ 432, 313, 48 },
		{ 432, 314, 48 },
		{ 432, 315, 48 },
		{ 432, 316, 48 },
		{ 432, 317, 48 },
		{ 432, 318, 48 },
		{ 432, 319, 48 },
		{ 432, 320, 48 },
		{ 432, 321, 48 },
		{ 432, 322, 48 },
		{ 433, 280, 48 },
		{ 433, 281, 48 },
		{ 433, 282, 48 },
		{ 433, 283, 48 },
		{ 433, 284, 48 },
		{ 433, 285, 48 },
		{ 433, 286, 48 },
		{ 433, 287, 48 },
		{ 433, 288, 48 },
		{ 433, 289, 48 },
		{ 433, 290, 48 },
		{ 433, 291, 48 },
		{ 433, 292, 48 },
		{ 433, 293, 48 },
		{ 433, 294, 48 },
		{ 433, 295, 48 },
		{ 433, 296, 48 },
		{ 433, 297, 48 },
		{ 433, 298, 48 },
		{ 433, 299, 48 },
		{ 433, 300, 48 },
		{ 433, 301, 48 },
		{ 433, 302, 48 },
		{ 433, 303, 48 },
		{ 433, 304, 48 },
		{ 433, 305, 48 },
		{ 433, 306, 48 },
		{ 433, 307, 48 },
		{ 433, 308, 48 },
		{ 433, 309, 48 },
		{ 433, 310, 48 },
		{ 433, 311, 48 },
		{ 433, 312, 48 },
		{ 433, 313, 48 },
		{ 433, 314, 48 },
		{ 433, 315, 48 },
		{ 433, 316, 48 },
		{ 433, 317, 48 },
		{ 433, 318, 48 },
		{ 433, 319, 48 },
		{ 433, 320, 48 },
		{ 433, 321, 48 },
		{ 433, 322, 48 },
		{ 434, 280, 48 },
		{ 434, 281, 48 },
		{ 434, 282, 48 },
		{ 434, 283, 48 },
		{ 434, 284, 48 },
		{ 434, 285, 48 },
		{ 434, 286, 48 },
		{ 434, 287, 48 },
		{ 434, 288, 48 },
		{ 434, 289, 48 },
		{ 434, 290, 48 },
		{ 434, 291, 48 },
		{ 434, 292, 48 },
		{ 434, 293, 48 },
		{ 434, 294, 48 },
		{ 434, 295, 48 },
		{ 434, 296, 48 },
		{ 434, 297, 48 },
		{ 434, 298, 48 },
		{ 434, 299, 48 },
		{ 434, 300, 48 },
		{ 434, 301, 48 },
		{ 434, 302, 48 },
		{ 434, 303, 48 },
		{ 434, 304, 48 },
		{ 434, 305, 48 },
		{ 434, 306, 48 },
		{ 434, 307, 48 },
		{ 434, 308, 48 },
		{ 434, 309, 48 },
		{ 434, 310, 48 },
		{ 434, 311, 48 },
		{ 434, 312, 48 },
		{ 434, 313, 48 },
		{ 434, 314, 48 },
		{ 434, 315, 48 },
		{ 434, 316, 48 },
		{ 434, 317, 48 },
		{ 434, 318, 48 },
		{ 434, 319, 48 },
		{ 434, 320, 48 },
		{ 434, 321, 48 },
		{ 434, 322, 48 },
		{ 434, 323, 48 },
		{ 435, 280, 48 },
		{ 435, 281, 48 },
		{ 435, 282, 48 },
		{ 435, 283, 48 },
		{ 435, 284, 48 },
		{ 435, 285, 48 },
		{ 435, 286, 48 },
		{ 435, 287, 48 },
		{ 435, 288, 48 },
		{ 435, 289, 48 },
		{ 435, 290, 48 },
		{ 435, 291, 48 },
		{ 435, 292, 48 },
		{ 435, 293, 48 },
		{ 435, 294, 48 },
		{ 435, 295, 48 },
		{ 435, 296, 48 },
		{ 435, 297, 48 },
		{ 435, 298, 48 },
		{ 435, 299, 48 },
		{ 435, 300, 48 },
		{ 435, 301, 48 },
		{ 435, 302, 48 },
		{ 435, 303, 48 },
		{ 435, 304, 48 },
		{ 435, 305, 48 },
		{ 435, 306, 48 },
		{ 435, 307, 48 },
		{ 435, 308, 48 },
		{ 435, 309, 48 },
		{ 435, 310, 48 },
		{ 435, 311, 48 },
		{ 435, 312, 48 },
		{ 435, 313, 48 },
		{ 435, 314, 48 },
		{ 435, 315, 48 },
		{ 435, 316, 48 },
		{ 435, 317, 48 },
		{ 435, 318, 48 },
		{ 435, 319, 48 },
		{ 435, 320, 48 },
		{ 435, 321, 48 },
		{ 435, 322, 48 },
		{ 435, 323, 48 },
		{ 436, 280, 48 },
		{ 436, 281, 48 },
		{ 436, 282, 48 },
		{ 436, 283, 48 },
		{ 436, 284, 48 },
		{ 436, 285, 48 },
		{ 436, 286, 48 },
		{ 436, 287, 48 },
		{ 436, 288, 48 },
		{ 436, 289, 48 },
		{ 436, 290, 48 },
		{ 436, 291, 48 },
		{ 436, 292, 48 },
		{ 436, 293, 48 },
		{ 436, 294, 48 },
		{ 436, 295, 48 },
		{ 436, 296, 48 },
		{ 436, 297, 48 },
		{ 436, 298, 48 },
		{ 436, 299, 48 },
		{ 436, 300, 48 },
		{ 436, 301, 48 },
		{ 436, 302, 48 },
		{ 436, 303, 48 },
		{ 436, 304, 48 },
		{ 436, 305, 48 },
		{ 436, 306, 48 },
		{ 436, 307, 48 },
		{ 436, 308, 48 },
		{ 436, 309, 48 },
		{ 436, 310, 48 },
		{ 436, 311, 48 },
		{ 436, 312, 48 },
		{ 436, 313, 48 },
		{ 436, 314, 48 },
		{ 436, 315, 48 },
		{ 436, 316, 48 },
		{ 436, 317, 48 },
		{ 436, 318, 48 },
		{ 436, 319, 48 },
		{ 436, 320, 48 },
		{ 436, 321, 48 },
		{ 436, 322, 48 },
		{ 436, 323, 48 },
		{ 436, 324, 48 },
		{ 437, 280, 48 },
		{ 437, 281, 48 },
		{ 437, 282, 48 },
		{ 437, 283, 48 },
		{ 437, 284, 48 },
		{ 437, 285, 48 },
		{ 437, 286, 48 },
		{ 437, 287, 48 },
		{ 437, 288, 48 },
		{ 437, 289, 48 },
		{ 437, 290, 48 },
		{ 437, 291, 48 },
		{ 437, 292, 48 },
		{ 437, 293, 48 },
		{ 437, 294, 48 },
		{ 437, 295, 48 },
		{ 437, 296, 48 },
		{ 437, 297, 48 },
		{ 437, 298, 48 },
		{ 437, 299, 48 },
		{ 437, 300, 48 },
		{ 437, 301, 48 },
		{ 437, 302, 48 },
		{ 437, 303, 48 },
		{ 437, 304, 48 },
		{ 437, 305, 48 },
		{ 437, 306, 48 },
		{ 437, 307, 48 },
		{ 437, 308, 48 },
		{ 437, 309, 48 },
		{ 437, 310, 48 },
		{ 437, 311, 48 },
		{ 437, 312, 48 },
		{ 437, 313, 48 },
		{ 437, 314, 48 },
		{ 437, 315, 48 },
		{ 437, 316, 48 },
		{ 437, 317, 48 },
		{ 437, 318, 48 },
		{ 437, 319, 48 },
		{ 437, 320, 48 },
		{ 437, 321, 48 },
		{ 437, 322, 48 },
		{ 437, 323, 48 },
		{ 437, 324, 48 },
		{ 438, 280, 48 },
		{ 438, 281, 48 },
		{ 438, 282, 48 },
		{ 438, 283, 48 },
		{ 438, 284, 48 },
		{ 438, 285, 48 },
		{ 438, 286, 48 },
		{ 438, 287, 48 },
		{ 438, 288, 48 },
		{ 438, 289, 48 },
		{ 438, 290, 48 },
		{ 438, 291, 48 },
		{ 438, 292, 48 },
		{ 438, 293, 48 },
		{ 438, 294, 48 },
		{ 438, 295, 48 },
		{ 438, 296, 48 },
		{ 438, 297, 48 },
		{ 438, 298, 48 },
		{ 438, 299, 48 },
		{ 438, 300, 48 },
		{ 438, 301, 48 },
		{ 438, 302, 48 },
		{ 438, 303, 48 },
		{ 438, 304, 48 },
		{ 438, 305, 48 },
		{ 438, 306, 48 },
		{ 438, 307, 48 },
		{ 438, 308, 48 },
		{ 438, 309, 48 },
		{ 438, 310, 48 },
		{ 438, 311, 48 },
		{ 438, 312, 48 },
		{ 438, 313, 48 },
		{ 438, 314, 48 },
		{ 438, 315, 48 },
		{ 438, 316, 48 },
		{ 438, 317, 48 },
		{ 438, 318, 48 },
		{ 438, 319, 48 },
		{ 438, 320, 48 },
		{ 438, 321, 48 },
		{ 438, 322, 48 },
		{ 438, 323, 48 },
		{ 438, 324, 48 },
		{ 439, 280, 48 },
		{ 439, 281, 48 },
		{ 439, 282, 48 },
		{ 439, 283, 48 },
		{ 439, 284, 48 },
		{ 439, 285, 48 },
		{ 439, 286, 48 },
		{ 439, 287, 48 },
		{ 439, 288, 48 },
		{ 439, 289, 48 },
		{ 439, 290, 48 },
		{ 439, 291, 48 },
		{ 439, 292, 48 },
		{ 439, 293, 48 },
		{ 439, 294, 48 },
		{ 439, 295, 48 },
		{ 439, 296, 48 },
		{ 439, 297, 48 },
		{ 439, 298, 48 },
		{ 439, 299, 48 },
		{ 439, 300, 48 },
		{ 439, 301, 48 },
		{ 439, 302, 48 },
		{ 439, 303, 48 },
		{ 439, 304, 48 },
		{ 439, 305, 48 },
		{ 439, 306, 48 },
		{ 439, 307, 48 },
		{ 439, 308, 48 },
		{ 439, 309, 48 },
		{ 439, 310, 48 },
		{ 439, 311, 48 },
		{ 439, 312, 48 },
		{ 439, 313, 48 },
		{ 439, 314, 48 },
		{ 439, 315, 48 },
		{ 439, 316, 48 },
		{ 439, 317, 48 },
		{ 439, 318, 48 },
		{ 439, 319, 48 },
		{ 439, 320, 48 },
		{ 439, 321, 48 },
		{ 439, 322, 48 },
		{ 439, 323, 48 },
		{ 439, 324, 48 },
		{ 439, 325, 48 },
		{ 440, 280, 48 },
		{ 440, 281, 48 },
		{ 440, 282, 48 },
		{ 440, 283, 48 },
		{ 440, 284, 48 },
		{ 440, 285, 48 },
		{ 440, 286, 48 },
		{ 440, 287, 48 },
		{ 440, 288, 48 },
		{ 440, 289, 48 },
		{ 440, 290, 48 },
		{ 440, 291, 48 },
		{ 440, 292, 48 },
		{ 440, 293, 48 },
		{ 440, 294, 48 },
		{ 440, 295, 48 },
		{ 440, 296, 48 },
		{ 440, 297, 48 },
		{ 440, 298, 48 },
		{ 440, 299, 48 },
		{ 440, 300, 48 },
		{ 440, 301, 48 },
		{ 440, 302, 48 },
		{ 440, 303, 48 },
		{ 440, 304, 48 },
		{ 440, 305, 48 },
		{ 440, 306, 48 },
		{ 440, 307, 48 },
		{ 440, 308, 48 },
		{ 440, 309, 48 },
		{ 440, 310, 48 },
		{ 440, 311, 48 },
		{ 440, 312, 48 },
		{ 440, 313, 48 },
		{ 440, 314, 48 },
		{ 440, 315, 48 },
		{ 440, 316, 48 },
		{ 440, 317, 48 },
		{ 440, 318, 48 },
		{ 440, 319, 48 },
		{ 440, 320, 48 },
		{ 440, 321, 48 },
		{ 440, 322, 48 },
		{ 440, 323, 48 },
		{ 440, 324, 48 },
		{ 440, 325, 48 },
		{ 441, 281, 48 },
		{ 441, 282, 48 },
		{ 441, 283, 48 },
		{ 441, 284, 48 },
		{ 441, 285, 48 },
		{ 441, 286, 48 },
		{ 441, 287, 48 },
		{ 441, 288, 48 },
		{ 441, 289, 48 },
		{ 441, 290, 48 },
		{ 441, 291, 48 },
		{ 441, 292, 48 },
		{ 441, 293, 48 },
		{ 441, 294, 48 },
		{ 441, 295, 48 },
		{ 441, 296, 48 },
		{ 441, 297, 48 },
		{ 441, 298, 48 },
		{ 441, 299, 48 },
		{ 441, 300, 48 },
		{ 441, 301, 48 },
		{ 441, 302, 48 },
		{ 441, 303, 48 },
		{ 441, 304, 48 },
		{ 441, 305, 48 },
		{ 441, 306, 48 },
		{ 441, 307, 48 },
		{ 441, 308, 48 },
		{ 441, 309, 48 },
		{ 441, 310, 48 },
		{ 441, 311, 48 },
		{ 441, 312, 48 },
		{ 441, 313, 48 },
		{ 441, 314, 48 },
		{ 441, 315, 48 },
		{ 441, 316, 48 },
		{ 441, 317, 48 },
		{ 441, 318, 48 },
		{ 441, 319, 48 },
		{ 441, 320, 48 },
		{ 441, 321, 48 },
		{ 441, 322, 48 },
		{ 441, 323, 48 },
		{ 441, 324, 48 },
		{ 441, 325, 48 },
		{ 442, 281, 48 },
		{ 442, 282, 48 },
		{ 442, 283, 48 },
		{ 442, 284, 48 },
		{ 442, 285, 48 },
		{ 442, 286, 48 },
		{ 442, 287, 48 },
		{ 442, 288, 48 },
		{ 442, 289, 48 },
		{ 442, 290, 48 },
		{ 442, 291, 48 },
		{ 442, 292, 48 },
		{ 442, 293, 48 },
		{ 442, 294, 48 },
		{ 442, 295, 48 },
		{ 442, 296, 48 },
		{ 442, 297, 48 },
		{ 442, 298, 48 },
		{ 442, 299, 48 },
		{ 442, 300, 48 },
		{ 442, 301, 48 },
		{ 442, 302, 48 },
		{ 442, 303, 48 },
		{ 442, 304, 48 },
		{ 442, 305, 48 },
		{ 442, 306, 48 },
		{ 442, 307, 48 },
		{ 442, 308, 48 },
		{ 442, 309, 48 },
		{ 442, 310, 48 },
		{ 442, 311, 48 },
		{ 442, 312, 48 },
		{ 442, 313, 48 },
		{ 442, 314, 48 },
		{ 442, 315, 48 },
		{ 442, 316, 48 },
		{ 442, 317, 48 },
		{ 442, 318, 48 },
		{ 442, 319, 48 },
		{ 442, 320, 48 },
		{ 442, 321, 48 },
		{ 442, 322, 48 },
		{ 442, 323, 48 },
		{ 442, 324, 48 },
		{ 442, 325, 48 },
		{ 442, 326, 48 },
		{ 443, 281, 48 },
		{ 443, 282, 48 },
		{ 443, 283, 48 },
		{ 443, 284, 48 },
		{ 443, 285, 48 },
		{ 443, 286, 48 },
		{ 443, 287, 48 },
		{ 443, 288, 48 },
		{ 443, 289, 48 },
		{ 443, 290, 48 },
		{ 443, 291, 48 },
		{ 443, 292, 48 },
		{ 443, 293, 48 },
		{ 443, 294, 48 },
		{ 443, 295, 48 },
		{ 443, 296, 48 },
		{ 443, 297, 48 },
		{ 443, 298, 48 },
		{ 443, 299, 48 },
		{ 443, 300, 48 },
		{ 443, 301, 48 },
		{ 443, 302, 48 },
		{ 443, 303, 48 },
		{ 443, 304, 48 },
		{ 443, 305, 48 },
		{ 443, 306, 48 },
		{ 443, 307, 48 },
		{ 443, 308, 48 },
		{ 443, 309, 48 },
		{ 443, 310, 48 },
		{ 443, 311, 48 },
		{ 443, 312, 48 },
		{ 443, 313, 48 },
		{ 443, 314, 48 },
		{ 443, 315, 48 },
		{ 443, 316, 48 },
		{ 443, 317, 48 },
		{ 443, 318, 48 },
		{ 443, 319, 48 },
		{ 443, 320, 48 },
		{ 443, 321, 48 },
		{ 443, 322, 48 },
		{ 443, 323, 48 },
		{ 443, 324, 48 },
		{ 443, 325, 48 },
		{ 443, 326, 48 },
		{ 444, 281, 48 },
		{ 444, 282, 48 },
		{ 444, 283, 48 },
		{ 444, 284, 48 },
		{ 444, 285, 48 },
		{ 444, 286, 48 },
		{ 444, 287, 48 },
		{ 444, 288, 48 },
		{ 444, 289, 48 },
		{ 444, 290, 48 },
		{ 444, 291, 48 },
		{ 444, 292, 48 },
		{ 444, 293, 48 },
		{ 444, 294, 48 },
		{ 444, 295, 48 },
		{ 444, 296, 48 },
		{ 444, 297, 48 },
		{ 444, 298, 48 },
		{ 444, 299, 48 },
		{ 444, 300, 48 },
		{ 444, 301, 48 },
		{ 444, 302, 48 },
		{ 444, 303, 48 },
		{ 444, 304, 48 },
		{ 444, 305, 48 },
		{ 444, 306, 48 },
		{ 444, 307, 48 },
		{ 444, 308, 48 },
		{ 444, 309, 48 },
		{ 444, 310, 48 },
		{ 444, 311, 48 },
		{ 444, 312, 48 },
		{ 444, 313, 48 },
		{ 444, 314, 48 },
		{ 444, 315, 48 },
		{ 444, 316, 48 },
		{ 444, 317, 48 },
		{ 444, 318, 48 },
		{ 444, 319, 48 },
		{ 444, 320, 48 },
		{ 444, 321, 48 },
		{ 444, 322, 48 },
		{ 444, 323, 48 },
		{ 444, 324, 48 },
		{ 444, 325, 48 },
		{ 444, 326, 48 },
		{ 444, 327, 48 },
		{ 445, 281, 48 },
		{ 445, 282, 48 },
		{ 445, 283, 48 },
		{ 445, 284, 48 },
		{ 445, 285, 48 },
		{ 445, 286, 48 },
		{ 445, 287, 48 },
		{ 445, 288, 48 },
		{ 445, 289, 48 },
		{ 445, 290, 48 },
		{ 445, 291, 48 },
		{ 445, 292, 48 },
		{ 445, 293, 48 },
		{ 445, 294, 48 },
		{ 445, 295, 48 },
		{ 445, 296, 48 },
		{ 445, 297, 48 },
		{ 445, 298, 48 },
		{ 445, 299, 48 },
		{ 445, 300, 48 },
		{ 445, 301, 48 },
		{ 445, 302, 48 },
		{ 445, 303, 48 },
		{ 445, 304, 48 },
		{ 445, 305, 48 },
		{ 445, 306, 48 },
		{ 445, 307, 48 },
		{ 445, 308, 48 },
		{ 445, 309, 48 },
		{ 445, 310, 48 },
		{ 445, 311, 48 },
		{ 445, 312, 48 },
		{ 445, 313, 48 },
		{ 445, 314, 48 },
		{ 445, 315, 48 },
		{ 445, 316, 48 },
		{ 445, 317, 48 },
		{ 445, 318, 48 },
		{ 445, 319, 48 },
		{ 445, 320, 48 },
		{ 445, 321, 48 },
		{ 445, 322, 48 },
		{ 445, 323, 48 },
		{ 445, 324, 48 },
		{ 445, 325, 48 },
		{ 445, 326, 48 },
		{ 445, 327, 48 },
		{ 446, 281, 48 },
		{ 446, 282, 48 },
		{ 446, 283, 48 },
		{ 446, 284, 48 },
		{ 446, 285, 48 },
		{ 446, 286, 48 },
		{ 446, 287, 48 },
		{ 446, 288, 48 },
		{ 446, 289, 48 },
		{ 446, 290, 48 },
		{ 446, 291, 48 },
		{ 446, 292, 48 },
		{ 446, 293, 48 },
		{ 446, 294, 48 },
		{ 446, 295, 48 },
		{ 446, 296, 48 },
		{ 446, 297, 48 },
		{ 446, 298, 48 },
		{ 446, 299, 48 },
		{ 446, 300, 48 },
		{ 446, 301, 48 },
		{ 446, 302, 48 },
		{ 446, 303, 48 },
		{ 446, 304, 48 },
		{ 446, 305, 48 },
		{ 446, 306, 48 },
		{ 446, 307, 48 },
		{ 446, 308, 48 },
		{ 446, 309, 48 },
		{ 446, 310, 48 },
		{ 446, 311, 48 },
		{ 446, 312, 48 },
		{ 446, 313, 48 },
		{ 446, 314, 48 },
		{ 446, 315, 48 },
		{ 446, 316, 48 },
		{ 446, 317, 48 },
		{ 446, 318, 48 },
		{ 446, 319, 48 },
		{ 446, 320, 48 },
		{ 446, 321, 48 },
		{ 446, 322, 48 },
		{ 446, 323, 48 },
		{ 446, 324, 48 },
		{ 446, 325, 48 },
		{ 446, 326, 48 },
		{ 446, 327, 48 },
		{ 447, 281, 48 },
		{ 447, 282, 48 },
		{ 447, 283, 48 },
		{ 447, 284, 48 },
		{ 447, 285, 48 },
		{ 447, 286, 48 },
		{ 447, 287, 48 },
		{ 447, 288, 48 },
		{ 447, 289, 48 },
		{ 447, 290, 48 },
		{ 447, 291, 48 },
		{ 447, 292, 48 },
		{ 447, 293, 48 },
		{ 447, 294, 48 },
		{ 447, 295, 48 },
		{ 447, 296, 48 },
		{ 447, 297, 48 },
		{ 447, 298, 48 },
		{ 447, 299, 48 },
		{ 447, 300, 48 },
		{ 447, 301, 48 },
		{ 447, 302, 48 },
		{ 447, 303, 48 },
		{ 447, 304, 48 },
		{ 447, 305, 48 },
		{ 447, 306, 48 },
		{ 447, 307, 48 },
		{ 447, 308, 48 },
		{ 447, 309, 48 },
		{ 447, 310, 48 },
		{ 447, 311, 48 },
		{ 447, 312, 48 },
		{ 447, 313, 48 },
		{ 447, 314, 48 },
		{ 447, 315, 48 },
		{ 447, 316, 48 },
		{ 447, 317, 48 },
		{ 447, 318, 48 },
		{ 447, 319, 48 },
		{ 447, 320, 48 },
		{ 447, 321, 48 },
		{ 447, 322, 48 },
		{ 447, 323, 48 },
		{ 447, 324, 48 },
		{ 447, 325, 48 },
		{ 447, 326, 48 },
		{ 447, 327, 48 },
		{ 447, 328, 48 },
		{ 448, 281, 48 },
		{ 448, 282, 48 },
		{ 448, 283, 48 },
		{ 448, 284, 48 },
		{ 448, 285, 48 },
		{ 448, 286, 48 },
		{ 448, 287, 48 },
		{ 448, 288, 48 },
		{ 448, 289, 48 },
		{ 448, 290, 48 },
		{ 448, 291, 48 },
		{ 448, 292, 48 },
		{ 448, 293, 48 },
		{ 448, 294, 48 },
		{ 448, 295, 48 },
		{ 448, 296, 48 },
		{ 448, 297, 48 },
		{ 448, 298, 48 },
		{ 448, 299, 48 },
		{ 448, 300, 48 },
		{ 448, 301, 48 },
		{ 448, 302, 48 },
		{ 448, 303, 48 },
		{ 448, 304, 48 },
		{ 448, 305, 48 },
		{ 448, 306, 48 },
		{ 448, 307, 48 },
		{ 448, 308, 48 },
		{ 448, 309, 48 },
		{ 448, 310, 48 },
		{ 448, 311, 48 },
		{ 448, 312, 48 },
		{ 448, 313, 48 },
		{ 448, 314, 48 },
		{ 448, 315, 48 },
		{ 448, 316, 48 },
		{ 448, 317, 48 },
		{ 448, 318, 48 },
		{ 448, 319, 48 },
		{ 448, 320, 48 },
		{ 448, 321, 48 },
		{ 448, 322, 48 },
		{ 448, 323, 48 },
		{ 448, 324, 48 },
		{ 448, 325, 48 },
		{ 448, 326, 48 },
		{ 448, 327, 48 },
		{ 448, 328, 48 },
		{ 449, 282, 48 },
		{ 449, 283, 48 },
		{ 449, 284, 48 },
		{ 449, 285, 48 },
		{ 449, 286, 48 },
		{ 449, 287, 48 },
		{ 449, 288, 48 },
		{ 449, 289, 48 },
		{ 449, 290, 48 },
		{ 449, 291, 48 },
		{ 449, 292, 48 },
		{ 449, 293, 48 },
		{ 449, 294, 48 },
		{ 449, 295, 48 },
		{ 449, 296, 48 },
		{ 449, 297, 48 },
		{ 449, 298, 48 },
		{ 449, 299, 48 },
		{ 449, 300, 48 },
		{ 449, 301, 48 },
		{ 449, 302, 48 },
		{ 449, 303, 48 },
		{ 449, 304, 48 },
		{ 449, 305, 48 },
		{ 449, 306, 48 },
		{ 449, 307, 48 },
		{ 449, 308, 48 },
		{ 449, 309, 48 },
		{ 449, 310, 48 },
		{ 449, 311, 48 },
		{ 449, 312, 48 },
		{ 449, 313, 48 },
		{ 449, 314, 48 },
		{ 449, 315, 48 },
		{ 449, 316, 48 },
		{ 449, 317, 48 },
		{ 449, 318, 48 },
		{ 449, 319, 48 },
		{ 449, 320, 48 },
		{ 449, 321, 48 },
		{ 449, 322, 48 },
		{ 449, 323, 48 },
		{ 449, 324, 48 },
		{ 449, 325, 48 },
		{ 449, 326, 48 },
		{ 449, 327, 48 },
		{ 449, 328, 48 },
		{ 450, 282, 48 },
		{ 450, 283, 48 },
		{ 450, 284, 48 },
		{ 450, 285, 48 },
		{ 450, 286, 48 },
		{ 450, 287, 48 },
		{ 450, 288, 48 },
		{ 450, 289, 48 },
		{ 450, 290, 48 },
		{ 450, 291, 48 },
		{ 450, 292, 48 },
		{ 450, 293, 48 },
		{ 450, 294, 48 },
		{ 450, 295, 48 },
		{ 450, 296, 48 },
		{ 450, 297, 48 },
		{ 450, 298, 48 },
		{ 450, 299, 48 },
		{ 450, 300, 48 },
		{ 450, 301, 48 },
		{ 450, 302, 48 },
		{ 450, 303, 48 },
		{ 450, 304, 48 },
		{ 450, 305, 48 },
		{ 450, 306, 48 },
		{ 450, 307, 48 },
		{ 450, 308, 48 },
		{ 450, 309, 48 },
		{ 450, 310, 48 },
		{ 450, 311, 48 },
		{ 450, 312, 48 },
		{ 450, 313, 48 },
		{ 450, 314, 48 },
		{ 450, 315, 48 },
		{ 450, 316, 48 },
		{ 450, 317, 48 },
		{ 450, 318, 48 },
		{ 450, 319, 48 },
		{ 450, 320, 48 },
		{ 450, 321, 48 },
		{ 450, 322, 48 },
		{ 450, 323, 48 },
		{ 450, 324, 48 },
		{ 450, 325, 48 },
		{ 450, 326, 48 },
		{ 450, 327, 48 },
		{ 450, 328, 48 },
		{ 450, 329, 48 },
		{ 451, 282, 48 },
		{ 451, 283, 48 },
		{ 451, 284, 48 },
		{ 451, 285, 48 },
		{ 451, 286, 48 },
		{ 451, 287, 48 },
		{ 451, 288, 48 },
		{ 451, 289, 48 },
		{ 451, 290, 48 },
		{ 451, 291, 48 },
		{ 451, 292, 48 },
		{ 451, 293, 48 },
		{ 451, 294, 48 },
		{ 451, 295, 48 },
		{ 451, 296, 48 },
		{ 451, 297, 48 },
		{ 451, 298, 48 },
		{ 451, 299, 48 },
		{ 451, 300, 48 },
		{ 451, 301, 48 },
		{ 451, 302, 48 },
		{ 451, 303, 48 },
		{ 451, 304, 48 },
		{ 451, 305, 48 },
		{ 451, 306, 48 },
		{ 451, 307, 48 },
		{ 451, 308, 48 },
		{ 451, 309, 48 },
		{ 451, 310, 48 },
		{ 451, 311, 48 },
		{ 451, 312, 48 },
		{ 451, 313, 48 },
		{ 451, 314, 48 },
		{ 451, 315, 48 },
		{ 451, 316, 48 },
		{ 451, 317, 48 },
		{ 451, 318, 48 },
		{ 451, 319, 48 },
		{ 451, 320, 48 },
		{ 451, 321, 48 },
		{ 451, 322, 48 },
		{ 451, 323, 48 },
		{ 451, 324, 48 },
		{ 451, 325, 48 },
		{ 451, 326, 48 },
		{ 451, 327, 48 },
		{ 451, 328, 48 },
		{ 451, 329, 48 },
		{ 452, 282, 48 },
		{ 452, 283, 48 },
		{ 452, 284, 48 },
		{ 452, 285, 48 },
		{ 452, 286, 48 },
		{ 452, 287, 48 },
		{ 452, 288, 48 },
		{ 452, 289, 48 },
		{ 452, 290, 48 },
		{ 452, 291, 48 },
		{ 452, 292, 48 },
		{ 452, 293, 48 },
		{ 452, 294, 48 },
		{ 452, 295, 48 },
		{ 452, 296, 48 },
		{ 452, 297, 48 },
		{ 452, 298, 48 },
		{ 452, 299, 48 },
		{ 452, 300, 48 },
		{ 452, 301, 48 },
		{ 452, 302, 48 },
		{ 452, 303, 48 },
		{ 452, 304, 48 },
		{ 452, 305, 48 },
		{ 452, 306, 48 },
		{ 452, 307, 48 },
		{ 452, 308, 48 },
		{ 452, 309, 48 },
		{ 452, 310, 48 },
		{ 452, 311, 48 },
		{ 452, 312, 48 },
		{ 452, 313, 48 },
		{ 452, 314, 48 },
		{ 452, 315, 48 },
		{ 452, 316, 48 },
		{ 452, 317, 48 },
		{ 452, 318, 48 },
		{ 452, 319, 48 },
		{ 452, 320, 48 },
		{ 452, 321, 48 },
		{ 452, 322, 48 },
		{ 452, 323, 48 },
		{ 452, 324, 48 },
		{ 452, 325, 48 },
		{ 452, 326, 48 },
		{ 452, 327, 48 },
		{ 452, 328, 48 },
		{ 452, 329, 48 },
		{ 452, 330, 48 },
		{ 453, 282, 48 },
		{ 453, 283, 48 },
		{ 453, 284, 48 },
		{ 453, 285, 48 },
		{ 453, 286, 48 },
		{ 453, 287, 48 },
		{ 453, 288, 48 },
		{ 453, 289, 48 },
		{ 453, 290, 48 },
		{ 453, 291, 48 },
		{ 453, 292, 48 },
		{ 453, 293, 48 },
		{ 453, 294, 48 },
		{ 453, 295, 48 },
		{ 453, 296, 48 },
		{ 453, 297, 48 },
		{ 453, 298, 48 },
		{ 453, 299, 48 },
		{ 453, 300, 48 },
		{ 453, 301, 48 },
		{ 453, 302, 48 },
		{ 453, 303, 48 },
		{ 453, 304, 48 },
		{ 453, 305, 48 },
		{ 453, 306, 48 },
		{ 453, 307, 48 },
		{ 453, 308, 48 },
		{ 453, 309, 48 },
		{ 453, 310, 48 },
		{ 453, 311, 48 },
		{ 453, 312, 48 },
		{ 453, 313, 48 },
		{ 453, 314, 48 },
		{ 453, 315, 48 },
		{ 453, 316, 48 },
		{ 453, 317, 48 },
		{ 453, 318, 48 },
		{ 453, 319, 48 },
		{ 453, 320, 48 },
		{ 453, 321, 48 },
		{ 453, 322, 48 },
		{ 453, 323, 48 },
		{ 453, 324, 48 },
		{ 453, 325, 48 },
		{ 453, 326, 48 },
		{ 453, 327, 48 },
		{ 453, 328, 48 },
		{ 453, 329, 48 },
		{ 453, 330, 48 },
		{ 454, 282, 48 },
		{ 454, 283, 48 },
		{ 454, 284, 48 },
		{ 454, 285, 48 },
		{ 454, 286, 48 },
		{ 454, 287, 48 },
		{ 454, 288, 48 },
		{ 454, 289, 48 },
		{ 454, 290, 48 },
		{ 454, 291, 48 },
		{ 454, 292, 48 },
		{ 454, 293, 48 },
		{ 454, 294, 48 },
		{ 454, 295, 48 },
		{ 454, 296, 48 },
		{ 454, 297, 48 },
		{ 454, 298, 48 },
		{ 454, 299, 48 },
		{ 454, 300, 48 },
		{ 454, 301, 48 },
		{ 454, 302, 48 },
		{ 454, 303, 48 },
		{ 454, 304, 48 },
		{ 454, 305, 48 },
		{ 454, 306, 48 },
		{ 454, 307, 48 },
		{ 454, 308, 48 },
		{ 454, 309, 48 },
		{ 454, 310, 48 },
		{ 454, 311, 48 },
		{ 454, 312, 48 },
		{ 454, 313, 48 },
		{ 454, 314, 48 },
		{ 454, 315, 48 },
		{ 454, 316, 48 },
		{ 454, 317, 48 },
		{ 454, 318, 48 },
		{ 454, 319, 48 },
		{ 454, 320, 48 },
		{ 454, 321, 48 },
		{ 454, 322, 48 },
		{ 454, 323, 48 },
		{ 454, 324, 48 },
		{ 454, 325, 48 },
		{ 454, 326, 48 },
		{ 454, 327, 48 },
		{ 454, 328, 48 },
		{ 454, 329, 48 },
		{ 454, 330, 48 },
		{ 455, 282, 48 },
		{ 455, 283, 48 },
		{ 455, 284, 48 },
		{ 455, 285, 48 },
		{ 455, 286, 48 },
		{ 455, 287, 48 },
		{ 455, 288, 48 },
		{ 455, 289, 48 },
		{ 455, 290, 48 },
		{ 455, 291, 48 },
		{ 455, 292, 48 },
		{ 455, 293, 48 },
		{ 455, 294, 48 },
		{ 455, 295, 48 },
		{ 455, 296, 48 },
		{ 455, 297, 48 },
		{ 455, 298, 48 },
		{ 455, 299, 48 },
		{ 455, 300, 48 },
		{ 455, 301, 48 },
		{ 455, 302, 48 },
		{ 455, 303, 48 },
		{ 455, 304, 48 },
		{ 455, 305, 48 },
		{ 455, 306, 48 },
		{ 455, 307, 48 },
		{ 455, 308, 48 },
		{ 455, 309, 48 },
		{ 455, 310, 48 },
		{ 455, 311, 48 },
		{ 455, 312, 48 },
		{ 455, 313, 48 },
		{ 455, 314, 48 },
		{ 455, 315, 48 },
		{ 455, 316, 48 },
		{ 455, 317, 48 },
		{ 455, 318, 48 },
		{ 455, 319, 48 },
		{ 455, 320, 48 },
		{ 455, 321, 48 },
		{ 455, 322, 48 },
		{ 455, 323, 48 },
		{ 455, 324, 48 },
		{ 455, 325, 48 },
		{ 455, 326, 48 },
		{ 455, 327, 48 },
		{ 455, 328, 48 },
		{ 455, 329, 48 },
		{ 455, 330, 48 },
		{ 455, 331, 48 },
		{ 456, 282, 48 },
		{ 456, 283, 48 },
		{ 456, 284, 48 },
		{ 456, 285, 48 },
		{ 456, 286, 48 },
		{ 456, 287, 48 },
		{ 456, 288, 48 },
		{ 456, 289, 48 },
		{ 456, 290, 48 },
		{ 456, 291, 48 },
		{ 456, 292, 48 },
		{ 456, 293, 48 },
		{ 456, 294, 48 },
		{ 456, 295, 48 },
		{ 456, 296, 48 },
		{ 456, 297, 48 },
		{ 456, 298, 48 },
		{ 456, 299, 48 },
		{ 456, 300, 48 },
		{ 456, 301, 48 },
		{ 456, 302, 48 },
		{ 456, 303, 48 },
		{ 456, 304, 48 },
		{ 456, 305, 48 },
		{ 456, 306, 48 },
		{ 456, 307, 48 },
		{ 456, 308, 48 },
		{ 456, 309, 48 },
		{ 456, 310, 48 },
		{ 456, 311, 48 },
		{ 456, 312, 48 },
		{ 456, 313, 48 },
		{ 456, 314, 48 },
		{ 456, 315, 48 },
		{ 456, 316, 48 },
		{ 456, 317, 48 },
		{ 456, 318, 48 },
		{ 456, 319, 48 },
		{ 456, 320, 48 },
		{ 456, 321, 48 },
		{ 456, 322, 48 },
		{ 456, 323, 48 },
		{ 456, 324, 48 },
		{ 456, 325, 48 },
		{ 456, 326, 48 },
		{ 456, 327, 48 },
		{ 456, 328, 48 },
		{ 456, 329, 48 },
		{ 456, 330, 48 },
		{ 456, 331, 48 },
		{ 457, 283, 48 },
		{ 457, 284, 48 },
		{ 457, 285, 48 },
		{ 457, 286, 48 },
		{ 457, 287, 48 },
		{ 457, 288, 48 },
		{ 457, 289, 48 },
		{ 457, 290, 48 },
		{ 457, 291, 48 },
		{ 457, 292, 48 },
		{ 457, 293, 48 },
		{ 457, 294, 48 },
		{ 457, 295, 48 },
		{ 457, 296, 48 },
		{ 457, 297, 48 },
		{ 457, 298, 48 },
		{ 457, 299, 48 },
		{ 457, 300, 48 },
		{ 457, 301, 48 },
		{ 457, 302, 48 },
		{ 457, 303, 48 },
		{ 457, 304, 48 },
		{ 457, 305, 48 },
		{ 457, 306, 48 },
		{ 457, 307, 48 },
		{ 457, 308, 48 },
		{ 457, 309, 48 },
		{ 457, 310, 48 },
		{ 457, 311, 48 },
		{ 457, 312, 48 },
		{ 457, 313, 48 },
		{ 457, 314, 48 },
		{ 457, 315, 48 },
		{ 457, 316, 48 },
		{ 457, 317, 48 },
		{ 457, 318, 48 },
		{ 457, 319, 48 },
		{ 457, 320, 48 },
		{ 457, 321, 48 },
		{ 457, 322, 48 },
		{ 457, 323, 48 },
		{ 457, 324, 48 },
		{ 457, 325, 48 },
		{ 457, 326, 48 },
		{ 457, 327, 48 },
		{ 457, 328, 48 },
		{ 457, 329, 48 },
		{ 457, 330, 48 },
		{ 457, 331, 48 },
		{ 458, 283, 48 },
		{ 458, 284, 48 },
		{ 458, 285, 48 },
		{ 458, 286, 48 },
		{ 458, 287, 48 },
		{ 458, 288, 48 },
		{ 458, 289, 48 },
		{ 458, 290, 48 },
		{ 458, 291, 48 },
		{ 458, 292, 48 },
		{ 458, 293, 48 },
		{ 458, 294, 48 },
		{ 458, 295, 48 },
		{ 458, 296, 48 },
		{ 458, 297, 48 },
		{ 458, 298, 48 },
		{ 458, 299, 48 },
		{ 458, 300, 48 },
		{ 458, 301, 48 },
		{ 458, 302, 48 },
		{ 458, 303, 48 },
		{ 458, 304, 48 },
		{ 458, 305, 48 },
		{ 458, 306, 48 },
		{ 458, 307, 48 },
		{ 458, 308, 48 },
		{ 458, 309, 48 },
		{ 458, 310, 48 },
		{ 458, 311, 48 },
		{ 458, 312, 48 },
		{ 458, 313, 48 },
		{ 458, 314, 48 },
		{ 458, 315, 48 },
		{ 458, 316, 48 },
		{ 458, 317, 48 },
		{ 458, 318, 48 },
		{ 458, 319, 48 },
		{ 458, 320, 48 },
		{ 458, 321, 48 },
		{ 458, 322, 48 },
		{ 458, 323, 48 },
		{ 458, 324, 48 },
		{ 458, 325, 48 },
		{ 458, 326, 48 },
		{ 458, 327, 48 },
		{ 458, 328, 48 },
		{ 458, 329, 48 },
		{ 458, 330, 48 },
		{ 458, 331, 48 },
		{ 458, 332, 48 },
		{ 459, 283, 48 },
		{ 459, 284, 48 },
		{ 459, 285, 48 },
		{ 459, 286, 48 },
		{ 459, 287, 48 },
		{ 459, 288, 48 },
		{ 459, 289, 48 },
		{ 459, 290, 48 },
		{ 459, 291, 48 },
		{ 459, 292, 48 },
		{ 459, 293, 48 },
		{ 459, 294, 48 },
		{ 459, 295, 48 },
		{ 459, 296, 48 },
		{ 459, 297, 48 },
		{ 459, 298, 48 },
		{ 459, 299, 48 },
		{ 459, 300, 48 },
		{ 459, 301, 48 },
		{ 459, 302, 48 },
		{ 459, 303, 48 },
		{ 459, 304, 48 },
		{ 459, 305, 48 },
		{ 459, 306, 48 },
		{ 459, 307, 48 },
		{ 459, 308, 48 },
		{ 459, 309, 48 },
		{ 459, 310, 48 },
		{ 459, 311, 48 },
		{ 459, 312, 48 },
		{ 459, 313, 48 },
		{ 459, 314, 48 },
		{ 459, 315, 48 },
		{ 459, 316, 48 },
		{ 459, 317, 48 },
		{ 459, 318, 48 },
		{ 459, 319, 48 },
		{ 459, 320, 48 },
		{ 459, 321, 48 },
		{ 459, 322, 48 },
		{ 459, 323, 48 },
		{ 459, 324, 48 },
		{ 459, 325, 48 },
		{ 459, 326, 48 },
		{ 459, 327, 48 },
		{ 459, 328, 48 },
		{ 459, 329, 48 },
		{ 459, 330, 48 },
		{ 459, 331, 48 },
		{ 459, 332, 48 },
		{ 460, 283, 48 },
		{ 460, 284, 48 },
		{ 460, 285, 48 },
		{ 460, 286, 48 },
		{ 460, 287, 48 },
		{ 460, 288, 48 },
		{ 460, 289, 48 },
		{ 460, 290, 48 },
		{ 460, 291, 48 },
		{ 460, 292, 48 },
		{ 460, 293, 48 },
		{ 460, 294, 48 },
		{ 460, 295, 48 },
		{ 460, 296, 48 },
		{ 460, 297, 48 },
		{ 460, 298, 48 },
		{ 460, 299, 48 },
		{ 460, 300, 48 },
		{ 460, 301, 48 },
		{ 460, 302, 48 },
		{ 460, 303, 48 },
		{ 460, 304, 48 },
		{ 460, 305, 48 },
		{ 460, 306, 48 },
		{ 460, 307, 48 },
		{ 460, 308, 48 },
		{ 460, 309, 48 },
		{ 460, 310, 48 },
		{ 460, 311, 48 },
		{ 460, 312, 48 },
		{ 460, 313, 48 },
		{ 460, 314, 48 },
		{ 460, 315, 48 },
		{ 460, 316, 48 },
		{ 460, 317, 48 },
		{ 460, 318, 48 },
		{ 460, 319, 48 },
		{ 460, 320, 48 },
		{ 460, 321, 48 },
		{ 460, 322, 48 },
		{ 460, 323, 48 },
		{ 460, 324, 48 },
		{ 460, 325, 48 },
		{ 460, 326, 48 },
		{ 460, 327, 48 },
		{ 460, 328, 48 },
		{ 460, 329, 48 },
		{ 460, 330, 48 },
		{ 460, 331, 48 },
		{ 460, 332, 48 },
		{ 460, 333, 48 },
		{ 461, 283, 48 },
		{ 461, 284, 48 },
		{ 461, 285, 48 },
		{ 461, 286, 48 },
		{ 461, 287, 48 },
		{ 461, 288, 48 },
		{ 461, 289, 48 },
		{ 461, 290, 48 },
		{ 461, 291, 48 },
		{ 461, 292, 48 },
		{ 461, 293, 48 },
		{ 461, 294, 48 },
		{ 461, 295, 48 },
		{ 461, 296, 48 },
		{ 461, 297, 48 },
		{ 461, 298, 48 },
		{ 461, 299, 48 },
		{ 461, 300, 48 },
		{ 461, 301, 48 },
		{ 461, 302, 48 },
		{ 461, 303, 48 },
		{ 461, 304, 48 },
		{ 461, 305, 48 },
		{ 461, 306, 48 },
		{ 461, 307, 48 },
		{ 461, 308, 48 },
		{ 461, 309, 48 },
		{ 461, 310, 48 },
		{ 461, 311, 48 },
		{ 461, 312, 48 },
		{ 461, 313, 48 },
		{ 461, 314, 48 },
		{ 461, 315, 48 },
		{ 461, 316, 48 },
		{ 461, 317, 48 },
		{ 461, 318, 48 },
		{ 461, 319, 48 },
		{ 461, 320, 48 },
		{ 461, 321, 48 },
		{ 461, 322, 48 },
		{ 461, 323, 48 },
		{ 461, 324, 48 },
		{ 461, 325, 48 },
		{ 461, 326, 48 },
		{ 461, 327, 48 },
		{ 461, 328, 48 },
		{ 461, 329, 48 },
		{ 461, 330, 48 },
		{ 461, 331, 48 },
		{ 461, 332, 48 },
		{ 461, 333, 48 },
		{ 462, 283, 48 },
		{ 462, 284, 48 },
		{ 462, 285, 48 },
		{ 462, 286, 48 },
		{ 462, 287, 48 },
		{ 462, 288, 48 },
		{ 462, 289, 48 },
		{ 462, 290, 48 },
		{ 462, 291, 48 },
		{ 462, 292, 48 },
		{ 462, 293, 48 },
		{ 462, 294, 48 },
		{ 462, 295, 48 },
		{ 462, 296, 48 },
		{ 462, 297, 48 },
		{ 462, 298, 48 },
		{ 462, 299, 48 },
		{ 462, 300, 48 },
		{ 462, 301, 48 },
		{ 462, 302, 48 },
		{ 462, 303, 48 },
		{ 462, 304, 48 },
		{ 462, 305, 48 },
		{ 462, 306, 48 },
		{ 462, 307, 48 },
		{ 462, 308, 48 },
		{ 462, 309, 48 },
		{ 462, 310, 48 },
		{ 462, 311, 48 },
		{ 462, 312, 48 },
		{ 462, 313, 48 },
		{ 462, 314, 48 },
		{ 462, 315, 48 },
		{ 462, 316, 48 },
		{ 462, 317, 48 },
		{ 462, 318, 48 },
		{ 462, 319, 48 },
		{ 462, 320, 48 },
		{ 462, 321, 48 },
		{ 462, 322, 48 },
		{ 462, 323, 48 },
		{ 462, 324, 48 },
		{ 462, 325, 48 },
		{ 462, 326, 48 },
		{ 462, 327, 48 },
		{ 462, 328, 48 },
		{ 462, 329, 48 },
		{ 462, 330, 48 },
		{ 462, 331, 48 },
		{ 462, 332, 48 },
		{ 462, 333, 48 },
		{ 463, 283, 48 },
		{ 463, 284, 48 },
		{ 463, 285, 48 },
		{ 463, 286, 48 },
		{ 463, 287, 48 },
		{ 463, 288, 48 },
		{ 463, 289, 48 },
		{ 463, 290, 48 },
		{ 463, 291, 48 },
		{ 463, 292, 48 },
		{ 463, 293, 48 },
		{ 463, 294, 48 },
		{ 463, 295, 48 },
		{ 463, 296, 48 },
		{ 463, 297, 48 },
		{ 463, 298, 48 },
		{ 463, 299, 48 },
		{ 463, 300, 48 },
		{ 463, 301, 48 },
		{ 463, 302, 48 },
		{ 463, 303, 48 },
		{ 463, 304, 48 },
		{ 463, 305, 48 },
		{ 463, 306, 48 },
		{ 463, 307, 48 },
		{ 463, 308, 48 },
		{ 463, 309, 48 },
		{ 463, 310, 48 },
		{ 463, 311, 48 },
		{ 463, 312, 48 },
		{ 463, 313, 48 },
		{ 463, 314, 48 },
		{ 463, 315, 48 },
		{ 463, 316, 48 },
		{ 463, 317, 48 },
		{ 463, 318, 48 },
		{ 463, 319, 48 },
		{ 463, 320, 48 },
		{ 463, 321, 48 },
		{ 463, 322, 48 },
		{ 463, 323, 48 },
		{ 463, 324, 48 },
		{ 463, 325, 48 },
		{ 463, 326, 48 },
		{ 463, 327, 48 },
		{ 463, 328, 48 },
		{ 463, 329, 48 },
		{ 463, 330, 48 },
		{ 463, 331, 48 },
		{ 463, 332, 48 },
		{ 463, 333, 48 },
		{ 463, 334, 48 },
		{ 464, 283, 48 },
		{ 464, 284, 48 },
		{ 464, 285, 48 },
		{ 464, 286, 48 },
		{ 464, 287, 48 },
		{ 464, 288, 48 },
		{ 464, 289, 48 },
		{ 464, 290, 48 },
		{ 464, 291, 48 },
		{ 464, 292, 48 },
		{ 464, 293, 48 },
		{ 464, 294, 48 },
		{ 464, 295, 48 },
		{ 464, 296, 48 },
		{ 464, 297, 48 },
		{ 464, 298, 48 },
		{ 464, 299, 48 },
		{ 464, 300, 48 },
		{ 464, 301, 48 },
		{ 464, 302, 48 },
		{ 464, 303, 48 },
		{ 464, 304, 48 },
		{ 464, 305, 48 },
		{ 464, 306, 48 },
		{ 464, 307, 48 },
		{ 464, 308, 48 },
		{ 464, 309, 48 },
		{ 464, 310, 48 },
		{ 464, 311, 48 },
		{ 464, 312, 48 },
		{ 464, 313, 48 },
		{ 464, 314, 48 },
		{ 464, 315, 48 },
		{ 464, 316, 48 },
		{ 464, 317, 48 },
		{ 464, 318, 48 },
		{ 464, 319, 48 },
		{ 464, 320, 48 },
		{ 464, 321, 48 },
		{ 464, 322, 48 },
		{ 464, 323, 48 },
		{ 464, 324, 48 },
		{ 464, 325, 48 },
		{ 464, 326, 48 },
		{ 464, 327, 48 },
		{ 464, 328, 48 },
		{ 464, 329, 48 },
		{ 464, 330, 48 },
		{ 464, 331, 48 },
		{ 464, 332, 48 },
		{ 464, 333, 48 },
		{ 464, 334, 48 },
		{ 465, 284, 48 },
		{ 465, 285, 48 },
		{ 465, 286, 48 },
		{ 465, 287, 48 },
		{ 465, 288, 48 },
		{ 465, 289, 48 },
		{ 465, 290, 48 },
		{ 465, 291, 48 },
		{ 465, 292, 48 },
		{ 465, 293, 48 },
		{ 465, 294, 48 },
		{ 465, 295, 48 },
		{ 465, 296, 48 },
		{ 465, 297, 48 },
		{ 465, 298, 48 },
		{ 465, 299, 48 },
		{ 465, 300, 48 },
		{ 465, 301, 48 },
		{ 465, 302, 48 },
		{ 465, 303, 48 },
		{ 465, 304, 48 },
		{ 465, 305, 48 },
		{ 465, 306, 48 },
		{ 465, 307, 48 },
		{ 465, 308, 48 },
		{ 465, 309, 48 },
		{ 465, 310, 48 },
		{ 465, 311, 48 },
		{ 465, 312, 48 },
		{ 465, 313, 48 },
		{ 465, 314, 48 },
		{ 465, 315, 48 },
		{ 465, 316, 48 },
		{ 465, 317, 48 },
		{ 465, 318, 48 },
		{ 465, 319, 48 },
		{ 465, 320, 48 },
		{ 465, 321, 48 },
		{ 465, 322, 48 },
		{ 465, 323, 48 },
		{ 465, 324, 48 },
		{ 465, 325, 48 },
		{ 465, 326, 48 },
		{ 465, 327, 48 },
		{ 465, 328, 48 },
		{ 465, 329, 48 },
		{ 465, 330, 48 },
		{ 465, 331, 48 },
		{ 465, 332, 48 },
		{ 465, 333, 48 },
		{ 465, 334, 48 },
		{ 466, 284, 48 },
		{ 466, 285, 48 },
		{ 466, 286, 48 },
		{ 466, 287, 48 },
		{ 466, 288, 48 },
		{ 466, 289, 48 },
		{ 466, 290, 48 },
		{ 466, 291, 48 },
		{ 466, 292, 48 },
		{ 466, 293, 48 },
		{ 466, 294, 48 },
		{ 466, 295, 48 },
		{ 466, 296, 48 },
		{ 466, 297, 48 },
		{ 466, 298, 48 },
		{ 466, 299, 48 },
		{ 466, 300, 48 },
		{ 466, 301, 48 },
		{ 466, 302, 48 },
		{ 466, 303, 48 },
		{ 466, 304, 48 },
		{ 466, 305, 48 },
		{ 466, 306, 48 },
		{ 466, 307, 48 },
		{ 466, 308, 48 },
		{ 466, 309, 48 },
		{ 466, 310, 48 },
		{ 466, 311, 48 },
		{ 466, 312, 48 },
		{ 466, 313, 48 },
		{ 466, 314, 48 },
		{ 466, 315, 48 },
		{ 466, 316, 48 },
		{ 466, 317, 48 },
		{ 466, 318, 48 },
		{ 466, 319, 48 },
		{ 466, 320, 48 },
		{ 466, 321, 48 },
		{ 466, 322, 48 },
		{ 466, 323, 48 },
		{ 466, 324, 48 },
		{ 466, 325, 48 },
		{ 466, 326, 48 },
		{ 466, 327, 48 },
		{ 466, 328, 48 },
		{ 466, 329, 48 },
		{ 466, 330, 48 },
		{ 466, 331, 48 },
		{ 466, 332, 48 },
		{ 466, 333, 48 },
		{ 466, 334, 48 },
		{ 466, 335, 48 },
		{ 467, 284, 48 },
		{ 467, 285, 48 },
		{ 467, 286, 48 },
		{ 467, 287, 48 },
		{ 467, 288, 48 },
		{ 467, 289, 48 },
		{ 467, 290, 48 },
		{ 467, 291, 48 },
		{ 467, 292, 48 },
		{ 467, 293, 48 },
		{ 467, 294, 48 },
		{ 467, 295, 48 },
		{ 467, 296, 48 },
		{ 467, 297, 48 },
		{ 467, 298, 48 },
		{ 467, 299, 48 },
		{ 467, 300, 48 },
		{ 467, 301, 48 },
		{ 467, 302, 48 },
		{ 467, 303, 48 },
		{ 467, 304, 48 },
		{ 467, 305, 48 },
		{ 467, 306, 48 },
		{ 467, 307, 48 },
		{ 467, 308, 48 },
		{ 467, 309, 48 },
		{ 467, 310, 48 },
		{ 467, 311, 48 },
		{ 467, 312, 48 },
		{ 467, 313, 48 },
		{ 467, 314, 48 },
		{ 467, 315, 48 },
		{ 467, 316, 48 },
		{ 467, 317, 48 },
		{ 467, 318, 48 },
		{ 467, 319, 48 },
		{ 467, 320, 48 },
		{ 467, 321, 48 },
		{ 467, 322, 48 },
		{ 467, 323, 48 },
		{ 467, 324, 48 },
		{ 467, 325, 48 },
		{ 467, 326, 48 },
		{ 467, 327, 48 },
		{ 467, 328, 48 },
		{ 467, 329, 48 },
		{ 467, 330, 48 },
		{ 467, 331, 48 },
		{ 467, 332, 48 },
		{ 467, 333, 48 },
		{ 467, 334, 48 },
		{ 467, 335, 48 },
		{ 468, 284, 48 },
		{ 468, 285, 48 },
		{ 468, 286, 48 },
		{ 468, 287, 48 },
		{ 468, 288, 48 },
		{ 468, 289, 48 },
		{ 468, 290, 48 },
		{ 468, 291, 48 },
		{ 468, 292, 48 },
		{ 468, 293, 48 },
		{ 468, 294, 48 },
		{ 468, 295, 48 },
		{ 468, 296, 48 },
		{ 468, 297, 48 },
		{ 468, 298, 48 },
		{ 468, 299, 48 },
		{ 468, 300, 48 },
		{ 468, 301, 48 },
		{ 468, 302, 48 },
		{ 468, 303, 48 },
		{ 468, 304, 48 },
		{ 468, 305, 48 },
		{ 468, 306, 48 },
		{ 468, 307, 48 },
		{ 468, 308, 48 },
		{ 468, 309, 48 },
		{ 468, 310, 48 },
		{ 468, 311, 48 },
		{ 468, 312, 48 },
		{ 468, 313, 48 },
		{ 468, 314, 48 },
		{ 468, 315, 48 },
		{ 468, 316, 48 },
		{ 468, 317, 48 },
		{ 468, 318, 48 },
		{ 468, 319, 48 },
		{ 468, 320, 48 },
		{ 468, 321, 48 },
		{ 468, 322, 48 },
		{ 468, 323, 48 },
		{ 468, 324, 48 },
		{ 468, 325, 48 },
		{ 468, 326, 48 },
		{ 468, 327, 48 },
		{ 468, 328, 48 },
		{ 468, 329, 48 },
		{ 468, 330, 48 },
		{ 468, 331, 48 },
		{ 468, 332, 48 },
		{ 468, 333, 48 },
		{ 468, 334, 48 },
		{ 468, 335, 48 },
		{ 468, 336, 48 },
		{ 469, 284, 48 },
		{ 469, 285, 48 },
		{ 469, 286, 48 },
		{ 469, 287, 48 },
		{ 469, 288, 48 },
		{ 469, 289, 48 },
		{ 469, 290, 48 },
		{ 469, 291, 48 },
		{ 469, 292, 48 },
		{ 469, 293, 48 },
		{ 469, 294, 48 },
		{ 469, 295, 48 },
		{ 469, 296, 48 },
		{ 469, 297, 48 },
		{ 469, 298, 48 },
		{ 469, 299, 48 },
		{ 469, 300, 48 },
		{ 469, 301, 48 },
		{ 469, 302, 48 },
		{ 469, 303, 48 },
		{ 469, 304, 48 },
		{ 469, 305, 48 },
		{ 469, 306, 48 },
		{ 469, 307, 48 },
		{ 469, 308, 48 },
		{ 469, 309, 48 },
		{ 469, 310, 48 },
		{ 469, 311, 48 },
		{ 469, 312, 48 },
		{ 469, 313, 48 },
		{ 469, 314, 48 },
		{ 469, 315, 48 },
		{ 469, 316, 48 },
		{ 469, 317, 48 },
		{ 469, 318, 48 },
		{ 469, 319, 48 },
		{ 469, 320, 48 },
		{ 469, 321, 48 },
		{ 469, 322, 48 },
		{ 469, 323, 48 },
		{ 469, 324, 48 },
		{ 469, 325, 48 },
		{ 469, 326, 48 },
		{ 469, 327, 48 },
		{ 469, 328, 48 },
		{ 469, 329, 48 },
		{ 469, 330, 48 },
		{ 469, 331, 48 },
		{ 469, 332, 48 },
		{ 469, 333, 48 },
		{ 469, 334, 48 },
		{ 469, 335, 48 },
		{ 469, 336, 48 },
		{ 470, 284, 48 },
		{ 470, 285, 48 },
		{ 470, 286, 48 },
		{ 470, 287, 48 },
		{ 470, 288, 48 },
		{ 470, 289, 48 },
		{ 470, 290, 48 },
		{ 470, 291, 48 },
		{ 470, 292, 48 },
		{ 470, 293, 48 },
		{ 470, 294, 48 },
		{ 470, 295, 48 },
		{ 470, 296, 48 },
		{ 470, 297, 48 },
		{ 470, 298, 48 },
		{ 470, 299, 48 },
		{ 470, 300, 48 },
		{ 470, 301, 48 },
		{ 470, 302, 48 },
		{ 470, 303, 48 },
		{ 470, 304, 48 },
		{ 470, 305, 48 },
		{ 470, 306, 48 },
		{ 470, 307, 48 },
		{ 470, 308, 48 },
		{ 470, 309, 48 },
		{ 470, 310, 48 },
		{ 470, 311, 48 },
		{ 470, 312, 48 },
		{ 470, 313, 48 },
		{ 470, 314, 48 },
		{ 470, 315, 48 },
		{ 470, 316, 48 },
		{ 470, 317, 48 },
		{ 470, 318, 48 },
		{ 470, 319, 48 },
		{ 470, 320, 48 },
		{ 470, 321, 48 },
		{ 470, 322, 48 },
		{ 470, 323, 48 },
		{ 470, 324, 48 },
		{ 470, 325, 48 },
		{ 470, 326, 48 },
		{ 470, 327, 48 },
		{ 470, 328, 48 },
		{ 470, 329, 48 },
		{ 470, 330, 48 },
		{ 470, 331, 48 },
		{ 470, 332, 48 },
		{ 470, 333, 48 },
		{ 470, 334, 48 },
		{ 470, 335, 48 },
		{ 470, 336, 48 },
		{ 471, 284, 48 },
		{ 471, 285, 48 },
		{ 471, 286, 48 },
		{ 471, 287, 48 },
		{ 471, 288, 48 },
		{ 471, 289, 48 },
		{ 471, 290, 48 },
		{ 471, 291, 48 },
		{ 471, 292, 48 },
		{ 471, 293, 48 },
		{ 471, 294, 48 },
		{ 471, 295, 48 },
		{ 471, 296, 48 },
		{ 471, 297, 48 },
		{ 471, 298, 48 },
		{ 471, 299, 48 },
		{ 471, 300, 48 },
		{ 471, 301, 48 },
		{ 471, 302, 48 },
		{ 471, 303, 48 },
		{ 471, 304, 48 },
		{ 471, 305, 48 },
		{ 471, 306, 48 },
		{ 471, 307, 48 },
		{ 471, 308, 48 },
		{ 471, 309, 48 },
		{ 471, 310, 48 },
		{ 471, 311, 48 },
		{ 471, 312, 48 },
		{ 471, 313, 48 },
		{ 471, 314, 48 },
		{ 471, 315, 48 },
		{ 471, 316, 48 },
		{ 471, 317, 48 },
		{ 471, 318, 48 },
		{ 471, 319, 48 },
		{ 471, 320, 48 },
		{ 471, 321, 48 },
		{ 471, 322, 48 },
		{ 471, 323, 48 },
		{ 471, 324, 48 },
		{ 471, 325, 48 },
		{ 471, 326, 48 },
		{ 471, 327, 48 },
		{ 471, 328, 48 },
		{ 471, 329, 48 },
		{ 471, 330, 48 },
		{ 471, 331, 48 },
		{ 471, 332, 48 },
		{ 471, 333, 48 },
		{ 471, 334, 48 },
		{ 471, 335, 48 },
		{ 471, 336, 48 },
		{ 471, 337, 48 },
		{ 472, 284, 48 },
		{ 472, 285, 48 },
		{ 472, 286, 48 },
		{ 472, 287, 48 },
		{ 472, 288, 48 },
		{ 472, 289, 48 },
		{ 472, 290, 48 },
		{ 472, 291, 48 },
		{ 472, 292, 48 },
		{ 472, 293, 48 },
		{ 472, 294, 48 },
		{ 472, 295, 48 },
		{ 472, 296, 48 },
		{ 472, 297, 48 },
		{ 472, 298, 48 },
		{ 472, 299, 48 },
		{ 472, 300, 48 },
		{ 472, 301, 48 },
		{ 472, 302, 48 },
		{ 472, 303, 48 },
		{ 472, 304, 48 },
		{ 472, 305, 48 },
		{ 472, 306, 48 },
		{ 472, 307, 48 },
		{ 472, 308, 48 },
		{ 472, 309, 48 },
		{ 472, 310, 48 },
		{ 472, 311, 48 },
		{ 472, 312, 48 },
		{ 472, 313, 48 },
		{ 472, 314, 48 },
		{ 472, 315, 48 },
		{ 472, 316, 48 },
		{ 472, 317, 48 },
		{ 472, 318, 48 },
		{ 472, 319, 48 },
		{ 472, 320, 48 },
		{ 472, 321, 48 },
		{ 472, 322, 48 },
		{ 472, 323, 48 },
		{ 472, 324, 48 },
		{ 472, 325, 48 },
		{ 472, 326, 48 },
		{ 472, 327, 48 },
		{ 472, 328, 48 },
		{ 472, 329, 48 },
		{ 472, 330, 48 },
		{ 472, 331, 48 },
		{ 472, 332, 48 },
		{ 472, 333, 48 },
		{ 472, 334, 48 },
		{ 472, 335, 48 },
		{ 472, 336, 48 },
		{ 472, 337, 48 },
		{ 473, 285, 48 },
		{ 473, 286, 48 },
		{ 473, 287, 48 },
		{ 473, 288, 48 },
		{ 473, 289, 48 },
		{ 473, 290, 48 },
		{ 473, 291, 48 },
		{ 473, 292, 48 },
		{ 473, 293, 48 },
		{ 473, 294, 48 },
		{ 473, 295, 48 },
		{ 473, 296, 48 },
		{ 473, 297, 48 },
		{ 473, 298, 48 },
		{ 473, 299, 48 },
		{ 473, 300, 48 },
		{ 473, 301, 48 },
		{ 473, 302, 48 },
		{ 473, 303, 48 },
		{ 473, 304, 48 },
		{ 473, 305, 48 },
		{ 473, 306, 48 },
		{ 473, 307, 48 },
		{ 473, 308, 48 },
		{ 473, 309, 48 },
		{ 473, 310, 48 },
		{ 473, 311, 48 },
		{ 473, 312, 48 },
		{ 473, 313, 48 },
		{ 473, 314, 48 },
		{ 473, 315, 48 },
		{ 473, 316, 48 },
		{ 473, 317, 48 },
		{ 473, 318, 48 },
		{ 473, 319, 48 },
		{ 473, 320, 48 },
		{ 473, 321, 48 },
		{ 473, 322, 48 },
		{ 473, 323, 48 },
		{ 473, 324, 48 },
		{ 473, 325, 48 },
		{ 473, 326, 48 },
		{ 473, 327, 48 },
		{ 473, 328, 48 },
		{ 473, 329, 48 },
		{ 473, 330, 48 },
		{ 473, 331, 48 },
		{ 473, 332, 48 },
		{ 473, 333, 48 },
		{ 473, 334, 48 },
		{ 473, 335, 48 },
		{ 473, 336, 48 },
		{ 473, 337, 48 },
		{ 474, 285, 48 },
		{ 474, 286, 48 },
		{ 474, 287, 48 },
		{ 474, 288, 48 },
		{ 474, 289, 48 },
		{ 474, 290, 48 },
		{ 474, 291, 48 },
		{ 474, 292, 48 },
		{ 474, 293, 48 },
		{ 474, 294, 48 },
		{ 474, 295, 48 },
		{ 474, 296, 48 },
		{ 474, 297, 48 },
		{ 474, 298, 48 },
		{ 474, 299, 48 },
		{ 474, 300, 48 },
		{ 474, 301, 48 },
		{ 474, 302, 48 },
		{ 474, 303, 48 },
		{ 474, 304, 48 },
		{ 474, 305, 48 },
		{ 474, 306, 48 },
		{ 474, 307, 48 },
		{ 474, 308, 48 },
		{ 474, 309, 48 },
		{ 474, 310, 48 },
		{ 474, 311, 48 },
		{ 474, 312, 48 },
		{ 474, 313, 48 },
		{ 474, 314, 48 },
		{ 474, 315, 48 },
		{ 474, 316, 48 },
		{ 474, 317, 48 },
		{ 474, 318, 48 },
		{ 474, 319, 48 },
		{ 474, 320, 48 },
		{ 474, 321, 48 },
		{ 474, 322, 48 },
		{ 474, 323, 48 },
		{ 474, 324, 48 },
		{ 474, 325, 48 },
		{ 474, 326, 48 },
		{ 474, 327, 48 },
		{ 474, 328, 48 },
		{ 474, 329, 48 },
		{ 474, 330, 48 },
		{ 474, 331, 48 },
		{ 474, 332, 48 },
		{ 474, 333, 48 },
		{ 474, 334, 48 },
		{ 474, 335, 48 },
		{ 474, 336, 48 },
		{ 474, 337, 48 },
		{ 474, 338, 48 },
		{ 475, 285, 48 },
		{ 475, 286, 48 },
		{ 475, 287, 48 },
		{ 475, 288, 48 },
		{ 475, 289, 48 },
		{ 475, 290, 48 },
		{ 475, 291, 48 },
		{ 475, 292, 48 },
		{ 475, 293, 48 },
		{ 475, 294, 48 },
		{ 475, 295, 48 },
		{ 475, 296, 48 },
		{ 475, 297, 48 },
		{ 475, 298, 48 },
		{ 475, 299, 48 },
		{ 475, 300, 48 },
		{ 475, 301, 48 },
		{ 475, 302, 48 },
		{ 475, 303, 48 },
		{ 475, 304, 48 },
		{ 475, 305, 48 },
		{ 475, 306, 48 },
		{ 475, 307, 48 },
		{ 475, 308, 48 },
		{ 475, 309, 48 },
		{ 475, 310, 48 },
		{ 475, 311, 48 },
		{ 475, 312, 48 },
		{ 475, 313, 48 },
		{ 475, 314, 48 },
		{ 475, 315, 48 },
		{ 475, 316, 48 },
		{ 475, 317, 48 },
		{ 475, 318, 48 },
		{ 475, 319, 48 },
		{ 475, 320, 48 },
		{ 475, 321, 48 },
		{ 475, 322, 48 },
		{ 475, 323, 48 },
		{ 475, 324, 48 },
		{ 475, 325, 48 },
		{ 475, 326, 48 },
		{ 475, 327, 48 },
		{ 475, 328, 48 },
		{ 475, 329, 48 },
		{ 475, 330, 48 },
		{ 475, 331, 48 },
		{ 475, 332, 48 },
		{ 475, 333, 48 },
		{ 475, 334, 48 },
		{ 475, 335, 48 },
		{ 475, 336, 48 },
		{ 475, 337, 48 },
		{ 475, 338, 48 },
		{ 476, 285, 48 },
		{ 476, 286, 48 },
		{ 476, 287, 48 },
		{ 476, 288, 48 },
		{ 476, 289, 48 },
		{ 476, 290, 48 },
		{ 476, 291, 48 },
		{ 476, 292, 48 },
		{ 476, 293, 48 },
		{ 476, 294, 48 },
		{ 476, 295, 48 },
		{ 476, 296, 48 },
		{ 476, 297, 48 },
		{ 476, 298, 48 },
		{ 476, 299, 48 },
		{ 476, 300, 48 },
		{ 476, 301, 48 },
		{ 476, 302, 48 },
		{ 476, 303, 48 },
		{ 476, 304, 48 },
		{ 476, 305, 48 },
		{ 476, 306, 48 },
		{ 476, 307, 48 },
		{ 476, 308, 48 },
		{ 476, 309, 48 },
		{ 476, 310, 48 },
		{ 476, 311, 48 },
		{ 476, 312, 48 },
		{ 476, 313, 48 },
		{ 476, 314, 48 },
		{ 476, 315, 48 },
		{ 476, 316, 48 },
		{ 476, 317, 48 },
		{ 476, 318, 48 },
		{ 476, 319, 48 },
		{ 476, 320, 48 },
		{ 476, 321, 48 },
		{ 476, 322, 48 },
		{ 476, 323, 48 },
		{ 476, 324, 48 },
		{ 476, 325, 48 },
		{ 476, 326, 48 },
		{ 476, 327, 48 },
		{ 476, 328, 48 },
		{ 476, 329, 48 },
		{ 476, 330, 48 },
		{ 476, 331, 48 },
		{ 476, 332, 48 },
		{ 476, 333, 48 },
		{ 476, 334, 48 },
		{ 476, 335, 48 },
		{ 476, 336, 48 },
		{ 476, 337, 48 },
		{ 476, 338, 48 },
		{ 476, 339, 48 },
		{ 477, 285, 48 },
		{ 477, 286, 48 },
		{ 477, 287, 48 },
		{ 477, 288, 48 },
		{ 477, 289, 48 },
		{ 477, 290, 48 },
		{ 477, 291, 48 },
		{ 477, 292, 48 },
		{ 477, 293, 48 },
		{ 477, 294, 48 },
		{ 477, 295, 48 },
		{ 477, 296, 48 },
		{ 477, 297, 48 },
		{ 477, 298, 48 },
		{ 477, 299, 48 },
		{ 477, 300, 48 },
		{ 477, 301, 48 },
		{ 477, 302, 48 },
		{ 477, 303, 48 },
		{ 477, 304, 48 },
		{ 477, 305, 48 },
		{ 477, 306, 48 },
		{ 477, 307, 48 },
		{ 477, 308, 48 },
		{ 477, 309, 48 },
		{ 477, 310, 48 },
		{ 477, 311, 48 },
		{ 477, 312, 48 },
		{ 477, 313, 48 },
		{ 477, 314, 48 },
		{ 477, 315, 48 },
		{ 477, 316, 48 },
		{ 477, 317, 48 },
		{ 477, 318, 48 },
		{ 477, 319, 48 },
		{ 477, 320, 48 },
		{ 477, 321, 48 },
		{ 477, 322, 48 },
		{ 477, 323, 48 },
		{ 477, 324, 48 },
		{ 477, 325, 48 },
		{ 477, 326, 48 },
		{ 477, 327, 48 },
		{ 477, 328, 48 },
		{ 477, 329, 48 },
		{ 477, 330, 48 },
		{ 477, 331, 48 },
		{ 477, 332, 48 },
		{ 477, 333, 48 },
		{ 477, 334, 48 },
		{ 477, 335, 48 },
		{ 477, 336, 48 },
		{ 477, 337, 48 },
		{ 477, 338, 48 },
		{ 477, 339, 48 },
		{ 478, 285, 48 },
		{ 478, 286, 48 },
		{ 478, 287, 48 },
		{ 478, 288, 48 },
		{ 478, 289, 48 },
		{ 478, 290, 48 },
		{ 478, 291, 48 },
		{ 478, 292, 48 },
		{ 478, 293, 48 },
		{ 478, 294, 48 },
		{ 478, 295, 48 },
		{ 478, 296, 48 },
		{ 478, 297, 48 },
		{ 478, 298, 48 },
		{ 478, 299, 48 },
		{ 478, 300, 48 },
		{ 478, 301, 48 },
		{ 478, 302, 48 },
		{ 478, 303, 48 },
		{ 478, 304, 48 },
		{ 478, 305, 48 },
		{ 478, 306, 48 },
		{ 478, 307, 48 },
		{ 478, 308, 48 },
		{ 478, 309, 48 },
		{ 478, 310, 48 },
		{ 478, 311, 48 },
		{ 478, 312, 48 },
		{ 478, 313, 48 },
		{ 478, 314, 48 },
		{ 478, 315, 48 },
		{ 478, 316, 48 },
		{ 478, 317, 48 },
		{ 478, 318, 48 },
		{ 478, 319, 48 },
		{ 478, 320, 48 },
		{ 478, 321, 48 },
		{ 478, 322, 48 },
		{ 478, 323, 48 },
		{ 478, 324, 48 },
		{ 478, 325, 48 },
		{ 478, 326, 48 },
		{ 478, 327, 48 },
		{ 478, 328, 48 },
		{ 478, 329, 48 },
		{ 478, 330, 48 },
		{ 478, 331, 48 },
		{ 478, 332, 48 },
		{ 478, 333, 48 },
		{ 478, 334, 48 },
		{ 478, 335, 48 },
		{ 478, 336, 48 },
		{ 478, 337, 48 },
		{ 478, 338, 48 },
		{ 478, 339, 48 },
		{ 479, 285, 48 },
		{ 479, 286, 48 },
		{ 479, 287, 48 },
		{ 479, 288, 48 },
		{ 479, 289, 48 },
		{ 479, 290, 48 },
		{ 479, 291, 48 },
		{ 479, 292, 48 },
		{ 479, 293, 48 },
		{ 479, 294, 48 },
		{ 479, 295, 48 },
		{ 479, 296, 48 },
		{ 479, 297, 48 },
		{ 479, 298, 48 },
		{ 479, 299, 48 },
		{ 479, 300, 48 },
		{ 479, 301, 48 },
		{ 479, 302, 48 },
		{ 479, 303, 48 },
		{ 479, 304, 48 },
		{ 479, 305, 48 },
		{ 479, 306, 48 },
		{ 479, 307, 48 },
		{ 479, 308, 48 },
		{ 479, 309, 48 },
		{ 479, 310, 48 },
		{ 479, 311, 48 },
		{ 479, 312, 48 },
		{ 479, 313, 48 },
		{ 479, 314, 48 },
		{ 479, 315, 48 },
		{ 479, 316, 48 },
		{ 479, 317, 48 },
		{ 479, 318, 48 },
		{ 479, 319, 48 },
		{ 479, 320, 48 },
		{ 479, 321, 48 },
		{ 479, 322, 48 },
		{ 479, 323, 48 },
		{ 479, 324, 48 },
		{ 479, 325, 48 },
		{ 479, 326, 48 },
		{ 479, 327, 48 },
		{ 479, 328, 48 },
		{ 479, 329, 48 },
		{ 479, 330, 48 },
		{ 479, 331, 48 },
		{ 479, 332, 48 },
		{ 479, 333, 48 },
		{ 479, 334, 48 },
		{ 479, 335, 48 },
		{ 479, 336, 48 },
		{ 479, 337, 48 },
		{ 479, 338, 48 },
		{ 479, 339, 48 },
		{ 479, 340, 48 },
		{ 480, 285, 48 },
		{ 480, 286, 48 },
		{ 480, 287, 48 },
		{ 480, 288, 48 },
		{ 480, 289, 48 },
		{ 480, 290, 48 },
		{ 480, 291, 48 },
		{ 480, 292, 48 },
		{ 480, 293, 48 },
		{ 480, 294, 48 },
		{ 480, 295, 48 },
		{ 480, 296, 48 },
		{ 480, 297, 48 },
		{ 480, 298, 48 },
		{ 480, 299, 48 },
		{ 480, 300, 48 },
		{ 480, 301, 48 },
		{ 480, 302, 48 },
		{ 480, 303, 48 },
		{ 480, 304, 48 },
		{ 480, 305, 48 },
		{ 480, 306, 48 },
		{ 480, 307, 48 },
		{ 480, 308, 48 },
		{ 480, 309, 48 },
		{ 480, 310, 48 },
		{ 480, 311, 48 },
		{ 480, 312, 48 },
		{ 480, 313, 48 },
		{ 480, 314, 48 },
		{ 480, 315, 48 },
		{ 480, 316, 48 },
		{ 480, 317, 48 },
		{ 480, 318, 48 },
		{ 480, 319, 48 },
		{ 480, 320, 48 },
		{ 480, 321, 48 },
		{ 480, 322, 48 },
		{ 480, 323, 48 },
		{ 480, 324, 48 },
		{ 480, 325, 48 },
		{ 480, 326, 48 },
		{ 480, 327, 48 },
		{ 480, 328, 48 },
		{ 480, 329, 48 },
		{ 480, 330, 48 },
		{ 480, 331, 48 },
		{ 480, 332, 48 },
		{ 480, 333, 48 },
		{ 480, 334, 48 },
		{ 480, 335, 48 },
		{ 480, 336, 48 },
		{ 480, 337, 48 },
		{ 480, 338, 48 },
		{ 480, 339, 48 },
		{ 480, 340, 48 },
		{ 481, 286, 48 },
		{ 481, 287, 48 },
		{ 481, 288, 48 },
		{ 481, 289, 48 },
		{ 481, 290, 48 },
		{ 481, 291, 48 },
		{ 481, 292, 48 },
		{ 481, 293, 48 },
		{ 481, 294, 48 },
		{ 481, 295, 48 },
		{ 481, 296, 48 },
		{ 481, 297, 48 },
		{ 481, 298, 48 },
		{ 481, 299, 48 },
		{ 481, 300, 48 },
		{ 481, 301, 48 },
		{ 481, 302, 48 },
		{ 481, 303, 48 },
		{ 481, 304, 48 },
		{ 481, 305, 48 },
		{ 481, 306, 48 },
		{ 481, 307, 48 },
		{ 481, 308, 48 },
		{ 481, 309, 48 },
		{ 481, 310, 48 },
		{ 481, 311, 48 },
		{ 481, 312, 48 },
		{ 481, 313, 48 },
		{ 481, 314, 48 },
		{ 481, 315, 48 },
		{ 481, 316, 48 },
		{ 481, 317, 48 },
		{ 481, 318, 48 },
		{ 481, 319, 48 },
		{ 481, 320, 48 },
		{ 481, 321, 48 },
		{ 481, 322, 48 },
		{ 481, 323, 48 },
		{ 481, 324, 48 },
		{ 481, 325, 48 },
		{ 481, 326, 48 },
		{ 481, 327, 48 },
		{ 481, 328, 48 },
		{ 481, 329, 48 },
		{ 481, 330, 48 },
		{ 481, 331, 48 },
		{ 481, 332, 48 },
		{ 481, 333, 48 },
		{ 481, 334, 48 },
		{ 481, 335, 48 },
		{ 481, 336, 48 },
		{ 481, 337, 48 },
		{ 481, 338, 48 },
		{ 481, 339, 48 },
		{ 481, 340, 48 },
		{ 482, 286, 48 },
		{ 482, 287, 48 },
		{ 482, 288, 48 },
		{ 482, 289, 48 },
		{ 482, 290, 48 },
		{ 482, 291, 48 },
		{ 482, 292, 48 },
		{ 482, 293, 48 },
		{ 482, 294, 48 },
		{ 482, 295, 48 },
		{ 482, 296, 48 },
		{ 482, 297, 48 },
		{ 482, 298, 48 },
		{ 482, 299, 48 },
		{ 482, 300, 48 },
		{ 482, 301, 48 },
		{ 482, 302, 48 },
		{ 482, 303, 48 },
		{ 482, 304, 48 },
		{ 482, 305, 48 },
		{ 482, 306, 48 },
		{ 482, 307, 48 },
		{ 482, 308, 48 },
		{ 482, 309, 48 },
		{ 482, 310, 48 },
		{ 482, 311, 48 },
		{ 482, 312, 48 },
		{ 482, 313, 48 },
		{ 482, 314, 48 },
		{ 482, 315, 48 },
		{ 482, 316, 48 },
		{ 482, 317, 48 },
		{ 482, 318, 48 },
		{ 482, 319, 48 },
		{ 482, 320, 48 },
		{ 482, 321, 48 },
		{ 482, 322, 48 },
		{ 482, 323, 48 },
		{ 482, 324, 48 },
		{ 482, 325, 48 },
		{ 482, 326, 48 },
		{ 482, 327, 48 },
		{ 482, 328, 48 },
		{ 482, 329, 48 },
		{ 482, 330, 48 },
		{ 482, 331, 48 },
		{ 482, 332, 48 },
		{ 482, 333, 48 },
		{ 482, 334, 48 },
		{ 482, 335, 48 },
		{ 482, 336, 48 },
		{ 482, 337, 48 },
		{ 482, 338, 48 },
		{ 482, 339, 48 },
		{ 482, 340, 48 },
		{ 482, 341, 48 },
		{ 483, 286, 48 },
		{ 483, 287, 48 },
		{ 483, 288, 48 },
		{ 483, 289, 48 },
		{ 483, 290, 48 },
		{ 483, 291, 48 },
		{ 483, 292, 48 },
		{ 483, 293, 48 },
		{ 483, 294, 48 },
		{ 483, 295, 48 },
		{ 483, 296, 48 },
		{ 483, 297, 48 },
		{ 483, 298, 48 },
		{ 483, 299, 48 },
		{ 483, 300, 48 },
		{ 483, 301, 48 },
		{ 483, 302, 48 },
		{ 483, 303, 48 },
		{ 483, 304, 48 },
		{ 483, 305, 48 },
		{ 483, 306, 48 },
		{ 483, 307, 48 },
		{ 483, 308, 48 },
		{ 483, 309, 48 },
		{ 483, 310, 48 },
		{ 483, 311, 48 },
		{ 483, 312, 48 },
		{ 483, 313, 48 },
		{ 483, 314, 48 },
		{ 483, 315, 48 },
		{ 483, 316, 48 },
		{ 483, 317, 48 },
		{ 483, 318, 48 },
		{ 483, 319, 48 },
		{ 483, 320, 48 },
		{ 483, 321, 48 },
		{ 483, 322, 48 },
		{ 483, 323, 48 },
		{ 483, 324, 48 },
		{ 483, 325, 48 },
		{ 483, 326, 48 },
		{ 483, 327, 48 },
		{ 483, 328, 48 },
		{ 483, 329, 48 },
		{ 483, 330, 48 },
		{ 483, 331, 48 },
		{ 483, 332, 48 },
		{ 483, 333, 48 },
		{ 483, 334, 48 },
		{ 483, 335, 48 },
		{ 483, 336, 48 },
		{ 483, 337, 48 },
		{ 483, 338, 48 },
		{ 483, 339, 48 },
		{ 483, 340, 48 },
		{ 483, 341, 48 },
		{ 484, 286, 48 },
		{ 484, 287, 48 },
		{ 484, 288, 48 },
		{ 484, 289, 48 },
		{ 484, 290, 48 },
		{ 484, 291, 48 },
		{ 484, 292, 48 },
		{ 484, 293, 48 },
		{ 484, 294, 48 },
		{ 484, 295, 48 },
		{ 484, 296, 48 },
		{ 484, 297, 48 },
		{ 484, 298, 48 },
		{ 484, 299, 48 },
		{ 484, 300, 48 },
		{ 484, 301, 48 },
		{ 484, 302, 48 },
		{ 484, 303, 48 },
		{ 484, 304, 48 },
		{ 484, 305, 48 },
		{ 484, 306, 48 },
		{ 484, 307, 48 },
		{ 484, 308, 48 },
		{ 484, 309, 48 },
		{ 484, 310, 48 },
		{ 484, 311, 48 },
		{ 484, 312, 48 },
		{ 484, 313, 48 },
		{ 484, 314, 48 },
		{ 484, 315, 48 },
		{ 484, 316, 48 },
		{ 484, 317, 48 },
		{ 484, 318, 48 },
		{ 484, 319, 48 },
		{ 484, 320, 48 },
		{ 484, 321, 48 },
		{ 484, 322, 48 },
		{ 484, 323, 48 },
		{ 484, 324, 48 },
		{ 484, 325, 48 },
		{ 484, 326, 48 },
		{ 484, 327, 48 },
		{ 484, 328, 48 },
		{ 484, 329, 48 },
		{ 484, 330, 48 },
		{ 484, 331, 48 },
		{ 484, 332, 48 },
		{ 484, 333, 48 },
		{ 484, 334, 48 },
		{ 484, 335, 48 },
		{ 484, 336, 48 },
		{ 484, 337, 48 },
		{ 484, 338, 48 },
		{ 484, 339, 48 },
		{ 484, 340, 48 },
		{ 484, 341, 48 },
		{ 484, 342, 48 },
		{ 485, 286, 48 },
		{ 485, 287, 48 },
		{ 485, 288, 48 },
		{ 485, 289, 48 },
		{ 485, 290, 48 },
		{ 485, 291, 48 },
		{ 485, 292, 48 },
		{ 485, 293, 48 },
		{ 485, 294, 48 },
		{ 485, 295, 48 },
		{ 485, 296, 48 },
		{ 485, 297, 48 },
		{ 485, 298, 48 },
		{ 485, 299, 48 },
		{ 485, 300, 48 },
		{ 485, 301, 48 },
		{ 485, 302, 48 },
		{ 485, 303, 48 },
		{ 485, 304, 48 },
		{ 485, 305, 48 },
		{ 485, 306, 48 },
		{ 485, 307, 48 },
		{ 485, 308, 48 },
		{ 485, 309, 48 },
		{ 485, 310, 48 },
		{ 485, 311, 48 },
		{ 485, 312, 48 },
		{ 485, 313, 48 },
		{ 485, 314, 48 },
		{ 485, 315, 48 },
		{ 485, 316, 48 },
		{ 485, 317, 48 },
		{ 485, 318, 48 },
		{ 485, 319, 48 },
		{ 485, 320, 48 },
		{ 485, 321, 48 },
		{ 485, 322, 48 },
		{ 485, 323, 48 },
		{ 485, 324, 48 },
		{ 485, 325, 48 },
		{ 485, 326, 48 },
		{ 485, 327, 48 },
		{ 485, 328, 48 },
		{ 485, 329, 48 },
		{ 485, 330, 48 },
		{ 485, 331, 48 },
		{ 485, 332, 48 },
		{ 485, 333, 48 },
		{ 485, 334, 48 },
		{ 485, 335, 48 },
		{ 485, 336, 48 },
		{ 485, 337, 48 },
		{ 485, 338, 48 },
		{ 485, 339, 48 },
		{ 485, 340, 48 },
		{ 485, 341, 48 },
		{ 485, 342, 48 },
		{ 486, 286, 48 },
		{ 486, 287, 48 },
		{ 486, 288, 48 },
		{ 486, 289, 48 },
		{ 486, 290, 48 },
		{ 486, 291, 48 },
		{ 486, 292, 48 },
		{ 486, 293, 48 },
		{ 486, 294, 48 },
		{ 486, 295, 48 },
		{ 486, 296, 48 },
		{ 486, 297, 48 },
		{ 486, 298, 48 },
		{ 486, 299, 48 },
		{ 486, 300, 48 },
		{ 486, 301, 48 },
		{ 486, 302, 48 },
		{ 486, 303, 48 },
		{ 486, 304, 48 },
		{ 486, 305, 48 },
		{ 486, 306, 48 },
		{ 486, 307, 48 },
		{ 486, 308, 48 },
		{ 486, 309, 48 },
		{ 486, 310, 48 },
		{ 486, 311, 48 },
		{ 486, 312, 48 },
		{ 486, 313, 48 },
		{ 486, 314, 48 },
		{ 486, 315, 48 },
		{ 486, 316, 48 },
		{ 486, 317, 48 },
		{ 486, 318, 48 },
		{ 486, 319, 48 },
		{ 486, 320, 48 },
		{ 486, 321, 48 },
		{ 486, 322, 48 },
		{ 486, 323, 48 },
		{ 486, 324, 48 },
		{ 486, 325, 48 },
		{ 486, 326, 48 },
		{ 486, 327, 48 },
		{ 486, 328, 48 },
		{ 486, 329, 48 },
		{ 486, 330, 48 },
		{ 486, 331, 48 },
		{ 486, 332, 48 },
		{ 486, 333, 48 },
		{ 486, 334, 48 },
		{ 486, 335, 48 },
		{ 486, 336, 48 },
		{ 486, 337, 48 },
		{ 486, 338, 48 },
		{ 486, 339, 48 },
		{ 486, 340, 48 },
		{ 486, 341, 48 },
		{ 486, 342, 48 },
		{ 487, 286, 48 },
		{ 487, 287, 48 },
		{ 487, 288, 48 },
		{ 487, 289, 48 },
		{ 487, 290, 48 },
		{ 487, 291, 48 },
		{ 487, 292, 48 },
		{ 487, 293, 48 },
		{ 487, 294, 48 },
		{ 487, 295, 48 },
		{ 487, 296, 48 },
		{ 487, 297, 48 },
		{ 487, 298, 48 },
		{ 487, 299, 48 },
		{ 487, 300, 48 },
		{ 487, 301, 48 },
		{ 487, 302, 48 },
		{ 487, 303, 48 },
		{ 487, 304, 48 },
		{ 487, 305, 48 },
		{ 487, 306, 48 },
		{ 487, 307, 48 },
		{ 487, 308, 48 },
		{ 487, 309, 48 },
		{ 487, 310, 48 },
		{ 487, 311, 48 },
		{ 487, 312, 48 },
		{ 487, 313, 48 },
		{ 487, 314, 48 },
		{ 487, 315, 48 },
		{ 487, 316, 48 },
		{ 487, 317, 48 },
		{ 487, 318, 48 },
		{ 487, 319, 48 },
		{ 487, 320, 48 },
		{ 487, 321, 48 },
		{ 487, 322, 48 },
		{ 487, 323, 48 },
		{ 487, 324, 48 },
		{ 487, 325, 48 },
		{ 487, 326, 48 },
		{ 487, 327, 48 },
		{ 487, 328, 48 },
		{ 487, 329, 48 },
		{ 487, 330, 48 },
		{ 487, 331, 48 },
		{ 487, 332, 48 },
		{ 487, 333, 48 },
		{ 487, 334, 48 },
		{ 487, 335, 48 },
		{ 487, 336, 48 },
		{ 487, 337, 48 },
		{ 487, 338, 48 },
		{ 487, 339, 48 },
		{ 487, 340, 48 },
		{ 487, 341, 48 },
		{ 487, 342, 48 },
		{ 487, 343, 48 },
		{ 488, 286, 48 },
		{ 488, 287, 48 },
		{ 488, 288, 48 },
		{ 488, 289, 48 },
		{ 488, 290, 48 },
		{ 488, 291, 48 },
		{ 488, 292, 48 },
		{ 488, 293, 48 },
		{ 488, 294, 48 },
		{ 488, 295, 48 },
		{ 488, 296, 48 },
		{ 488, 297, 48 },
		{ 488, 298, 48 },
		{ 488, 299, 48 },
		{ 488, 300, 48 },
		{ 488, 301, 48 },
		{ 488, 302, 48 },
		{ 488, 303, 48 },
		{ 488, 304, 48 },
		{ 488, 305, 48 },
		{ 488, 306, 48 },
		{ 488, 307, 48 },
		{ 488, 308, 48 },
		{ 488, 309, 48 },
		{ 488, 310, 48 },
		{ 488, 311, 48 },
		{ 488, 312, 48 },
		{ 488, 313, 48 },
		{ 488, 314, 48 },
		{ 488, 315, 48 },
		{ 488, 316, 48 },
		{ 488, 317, 48 },
		{ 488, 318, 48 },
		{ 488, 319, 48 },
		{ 488, 320, 48 },
		{ 488, 321, 48 },
		{ 488, 322, 48 },
		{ 488, 323, 48 },
		{ 488, 324, 48 },
		{ 488, 325, 48 },
		{ 488, 326, 48 },
		{ 488, 327, 48 },
		{ 488, 328, 48 },
		{ 488, 329, 48 },
		{ 488, 330, 48 },
		{ 488, 331, 48 },
		{ 488, 332, 48 },
		{ 488, 333, 48 },
		{ 488, 334, 48 },
		{ 488, 335, 48 },
		{ 488, 336, 48 },
		{ 488, 337, 48 },
		{ 488, 338, 48 },
		{ 488, 339, 48 },
		{ 488, 340, 48 },
		{ 488, 341, 48 },
		{ 488, 342, 48 },
		{ 488, 343, 48 },
		{ 489, 287, 48 },
		{ 489, 288, 48 },
		{ 489, 289, 48 },
		{ 489, 290, 48 },
		{ 489, 291, 48 },
		{ 489, 292, 48 },
		{ 489, 293, 48 },
		{ 489, 294, 48 },
		{ 489, 295, 48 },
		{ 489, 296, 48 },
		{ 489, 297, 48 },
		{ 489, 298, 48 },
		{ 489, 299, 48 },
		{ 489, 300, 48 },
		{ 489, 301, 48 },
		{ 489, 302, 48 },
		{ 489, 303, 48 },
		{ 489, 304, 48 },
		{ 489, 305, 48 },
		{ 489, 306, 48 },
		{ 489, 307, 48 },
		{ 489, 308, 48 },
		{ 489, 309, 48 },
		{ 489, 310, 48 },
		{ 489, 311, 48 },
		{ 489, 312, 48 },
		{ 489, 313, 48 },
		{ 489, 314, 48 },
		{ 489, 315, 48 },
		{ 489, 316, 48 },
		{ 489, 317, 48 },
		{ 489, 318, 48 },
		{ 489, 319, 48 },
		{ 489, 320, 48 },
		{ 489, 321, 48 },
		{ 489, 322, 48 },
		{ 489, 323, 48 },
		{ 489, 324, 48 },
		{ 489, 325, 48 },
		{ 489, 326, 48 },
		{ 489, 327, 48 },
		{ 489, 328, 48 },
		{ 489, 329, 48 },
		{ 489, 330, 48 },
		{ 489, 331, 48 },
		{ 489, 332, 48 },
		{ 489, 333, 48 },
		{ 489, 334, 48 },
		{ 489, 335, 48 },
		{ 489, 336, 48 },
		{ 489, 337, 48 },
		{ 489, 338, 48 },
		{ 489, 339, 48 },
		{ 489, 340, 48 },
		{ 489, 341, 48 },
		{ 489, 342, 48 },
		{ 489, 343, 48 },
		{ 490, 287, 48 },
		{ 490, 288, 48 },
		{ 490, 289, 48 },
		{ 490, 290, 48 },
		{ 490, 291, 48 },
		{ 490, 292, 48 },
		{ 490, 293, 48 },
		{ 490, 294, 48 },
		{ 490, 295, 48 },
		{ 490, 296, 48 },
		{ 490, 297, 48 },
		{ 490, 298, 48 },
		{ 490, 299, 48 },
		{ 490, 300, 48 },
		{ 490, 301, 48 },
		{ 490, 302, 48 },
		{ 490, 303, 48 },
		{ 490, 304, 48 },
		{ 490, 305, 48 },
		{ 490, 306, 48 },
		{ 490, 307, 48 },
		{ 490, 308, 48 },
		{ 490, 309, 48 },
		{ 490, 310, 48 },
		{ 490, 311, 48 },
		{ 490, 312, 48 },
		{ 490, 313, 48 },
		{ 490, 314, 48 },
		{ 490, 315, 48 },
		{ 490, 316, 48 },
		{ 490, 317, 48 },
		{ 490, 318, 48 },
		{ 490, 319, 48 },
		{ 490, 320, 48 },
		{ 490, 321, 48 },
		{ 490, 322, 48 },
		{ 490, 323, 48 },
		{ 490, 324, 48 },
		{ 490, 325, 48 },
		{ 490, 326, 48 },
		{ 490, 327, 48 },
		{ 490, 328, 48 },
		{ 490, 329, 48 },
		{ 490, 330, 48 },
		{ 490, 331, 48 },
		{ 490, 332, 48 },
		{ 490, 333, 48 },
		{ 490, 334, 48 },
		{ 490, 335, 48 },
		{ 490, 336, 48 },
		{ 490, 337, 48 },
		{ 490, 338, 48 },
		{ 490, 339, 48 },
		{ 490, 340, 48 },
		{ 490, 341, 48 },
		{ 490, 342, 48 },
		{ 490, 343, 48 },
		{ 490, 344, 48 },
		{ 491, 287, 48 },
		{ 491, 288, 48 },
		{ 491, 289, 48 },
		{ 491, 290, 48 },
		{ 491, 291, 48 },
		{ 491, 292, 48 },
		{ 491, 293, 48 },
		{ 491, 294, 48 },
		{ 491, 295, 48 },
		{ 491, 296, 48 },
		{ 491, 297, 48 },
		{ 491, 298, 48 },
		{ 491, 299, 48 },
		{ 491, 300, 48 },
		{ 491, 301, 48 },
		{ 491, 302, 48 },
		{ 491, 303, 48 },
		{ 491, 304, 48 },
		{ 491, 305, 48 },
		{ 491, 306, 48 },
		{ 491, 307, 48 },
		{ 491, 308, 48 },
		{ 491, 309, 48 },
		{ 491, 310, 48 },
		{ 491, 311, 48 },
		{ 491, 312, 48 },
		{ 491, 313, 48 },
		{ 491, 314, 48 },
		{ 491, 315, 48 },
		{ 491, 316, 48 },
		{ 491, 317, 48 },
		{ 491, 318, 48 },
		{ 491, 319, 48 },
		{ 491, 320, 48 },
		{ 491, 321, 48 },
		{ 491, 322, 48 },
		{ 491, 323, 48 },
		{ 491, 324, 48 },
		{ 491, 325, 48 },
		{ 491, 326, 48 },
		{ 491, 327, 48 },
		{ 491, 328, 48 },
		{ 491, 329, 48 },
		{ 491, 330, 48 },
		{ 491, 331, 48 },
		{ 491, 332, 48 },
		{ 491, 333, 48 },
		{ 491, 334, 48 },
		{ 491, 335, 48 },
		{ 491, 336, 48 },
		{ 491, 337, 48 },
		{ 491, 338, 48 },
		{ 491, 339, 48 },
		{ 491, 340, 48 },
		{ 491, 341, 48 },
		{ 491, 342, 48 },
		{ 491, 343, 48 },
		{ 491, 344, 48 },
		{ 492, 287, 48 },
		{ 492, 288, 48 },
		{ 492, 289, 48 },
		{ 492, 290, 48 },
		{ 492, 291, 48 },
		{ 492, 292, 48 },
		{ 492, 293, 48 },
		{ 492, 294, 48 },
		{ 492, 295, 48 },
		{ 492, 296, 48 },
		{ 492, 297, 48 },
		{ 492, 298, 48 },
		{ 492, 299, 48 },
		{ 492, 300, 48 },
		{ 492, 301, 48 },
		{ 492, 302, 48 },
		{ 492, 303, 48 },
		{ 492, 304, 48 },
		{ 492, 305, 48 },
		{ 492, 306, 48 },
		{ 492, 307, 48 },
		{ 492, 308, 48 },
		{ 492, 309, 48 },
		{ 492, 310, 48 },
		{ 492, 311, 48 },
		{ 492, 312, 48 },
		{ 492, 313, 48 },
		{ 492, 314, 48 },
		{ 492, 315, 48 },
		{ 492, 316, 48 },
		{ 492, 317, 48 },
		{ 492, 318, 48 },
		{ 492, 319, 48 },
		{ 492, 320, 48 },
		{ 492, 321, 48 },
		{ 492, 322, 48 },
		{ 492, 323, 48 },
		{ 492, 324, 48 },
		{ 492, 325, 48 },
		{ 492, 326, 48 },
		{ 492, 327, 48 },
		{ 492, 328, 48 },
		{ 492, 329, 48 },
		{ 492, 330, 48 },
		{ 492, 331, 48 },
		{ 492, 332, 48 },
		{ 492, 333, 48 },
		{ 492, 334, 48 },
		{ 492, 335, 48 },
		{ 492, 336, 48 },
		{ 492, 337, 48 },
		{ 492, 338, 48 },
		{ 492, 339, 48 },
		{ 492, 340, 48 },
		{ 492, 341, 48 },
		{ 492, 342, 48 },
		{ 492, 343, 48 },
		{ 492, 344, 48 },
		{ 492, 345, 48 },
		{ 493, 287, 48 },
		{ 493, 288, 48 },
		{ 493, 289, 48 },
		{ 493, 290, 48 },
		{ 493, 291, 48 },
		{ 493, 292, 48 },
		{ 493, 293, 48 },
		{ 493, 294, 48 },
		{ 493, 295, 48 },
		{ 493, 296, 48 },
		{ 493, 297, 48 },
		{ 493, 298, 48 },
		{ 493, 299, 48 },
		{ 493, 300, 48 },
		{ 493, 301, 48 },
		{ 493, 302, 48 },
		{ 493, 303, 48 },
		{ 493, 304, 48 },
		{ 493, 305, 48 },
		{ 493, 306, 48 },
		{ 493, 307, 48 },
		{ 493, 308, 48 },
		{ 493, 309, 48 },
		{ 493, 310, 48 },
		{ 493, 311, 48 },
		{ 493, 312, 48 },
		{ 493, 313, 48 },
		{ 493, 314, 48 },
		{ 493, 315, 48 },
		{ 493, 316, 48 },
		{ 493, 317, 48 },
		{ 493, 318, 48 },
		{ 493, 319, 48 },
		{ 493, 320, 48 },
		{ 493, 321, 48 },
		{ 493, 322, 48 },
		{ 493, 323, 48 },
		{ 493, 324, 48 },
		{ 493, 325, 48 },
		{ 493, 326, 48 },
		{ 493, 327, 48 },
		{ 493, 328, 48 },
		{ 493, 329, 48 },
		{ 493, 330, 48 },
		{ 493, 331, 48 },
		{ 493, 332, 48 },
		{ 493, 333, 48 },
		{ 493, 334, 48 },
		{ 493, 335, 48 },
		{ 493, 336, 48 },
		{ 493, 337, 48 },
		{ 493, 338, 48 },
		{ 493, 339, 48 },
		{ 493, 340, 48 },
		{ 493, 341, 48 },
		{ 493, 342, 48 },
		{ 493, 343, 48 },
		{ 493, 344, 48 },
		{ 493, 345, 48 },
		{ 494, 287, 48 },
		{ 494, 288, 48 },
		{ 494, 289, 48 },
		{ 494, 290, 48 },
		{ 494, 291, 48 },
		{ 494, 292, 48 },
		{ 494, 293, 48 },
		{ 494, 294, 48 },
		{ 494, 295, 48 },
		{ 494, 296, 48 },
		{ 494, 297, 48 },
		{ 494, 298, 48 },
		{ 494, 299, 48 },
		{ 494, 300, 48 },
		{ 494, 301, 48 },
		{ 494, 302, 48 },
		{ 494, 303, 48 },
		{ 494, 304, 48 },
		{ 494, 305, 48 },
		{ 494, 306, 48 },
		{ 494, 307, 48 },
		{ 494, 308, 48 },
		{ 494, 309, 48 },
		{ 494, 310, 48 },
		{ 494, 311, 48 },
		{ 494, 312, 48 },
		{ 494, 313, 48 },
		{ 494, 314, 48 },
		{ 494, 315, 48 },
		{ 494, 316, 48 },
		{ 494, 317, 48 },
		{ 494, 318, 48 },
		{ 494, 319, 48 },
		{ 494, 320, 48 },
		{ 494, 321, 48 },
		{ 494, 322, 48 },
		{ 494, 323, 48 },
		{ 494, 324, 48 },
		{ 494, 325, 48 },
		{ 494, 326, 48 },
		{ 494, 327, 48 },
		{ 494, 328, 48 },
		{ 494, 329, 48 },
		{ 494, 330, 48 },
		{ 494, 331, 48 },
		{ 494, 332, 48 },
		{ 494, 333, 48 },
		{ 494, 334, 48 },
		{ 494, 335, 48 },
		{ 494, 336, 48 },
		{ 494, 337, 48 },
		{ 494, 338, 48 },
		{ 494, 339, 48 },
		{ 494, 340, 48 },
		{ 494, 341, 48 },
		{ 494, 342, 48 },
		{ 494, 343, 48 },
		{ 494, 344, 48 },
		{ 494, 345, 48 },
		{ 495, 287, 48 },
		{ 495, 288, 48 },
		{ 495, 289, 48 },
		{ 495, 290, 48 },
		{ 495, 291, 48 },
		{ 495, 292, 48 },
		{ 495, 293, 48 },
		{ 495, 294, 48 },
		{ 495, 295, 48 },
		{ 495, 296, 48 },
		{ 495, 297, 48 },
		{ 495, 298, 48 },
		{ 495, 299, 48 },
		{ 495, 300, 48 },
		{ 495, 301, 48 },
		{ 495, 302, 48 },
		{ 495, 303, 48 },
		{ 495, 304, 48 },
		{ 495, 305, 48 },
		{ 495, 306, 48 },
		{ 495, 307, 48 },
		{ 495, 308, 48 },
		{ 495, 309, 48 },
		{ 495, 310, 48 },
		{ 495, 311, 48 },
		{ 495, 312, 48 },
		{ 495, 313, 48 },
		{ 495, 314, 48 },
		{ 495, 315, 48 },
		{ 495, 316, 48 },
		{ 495, 317, 48 },
		{ 495, 318, 48 },
		{ 495, 319, 48 },
		{ 495, 320, 48 },
		{ 495, 321, 48 },
		{ 495, 322, 48 },
		{ 495, 323, 48 },
		{ 495, 324, 48 },
		{ 495, 325, 48 },
		{ 495, 326, 48 },
		{ 495, 327, 48 },
		{ 495, 328, 48 },
		{ 495, 329, 48 },
		{ 495, 330, 48 },
		{ 495, 331, 48 },
		{ 495, 332, 48 },
		{ 495, 333, 48 },
		{ 495, 334, 48 },
		{ 495, 335, 48 },
		{ 495, 336, 48 },
		{ 495, 337, 48 },
		{ 495, 338, 48 },
		{ 495, 339, 48 },
		{ 495, 340, 48 },
		{ 495, 341, 48 },
		{ 495, 342, 48 },
		{ 495, 343, 48 },
		{ 495, 344, 48 },
		{ 495, 345, 48 },
		{ 495, 346, 48 },
		{ 496, 287, 48 },
		{ 496, 288, 48 },
		{ 496, 289, 48 },
		{ 496, 290, 48 },
		{ 496, 291, 48 },
		{ 496, 292, 48 },
		{ 496, 293, 48 },
		{ 496, 294, 48 },
		{ 496, 295, 48 },
		{ 496, 296, 48 },
		{ 496, 297, 48 },
		{ 496, 298, 48 },
		{ 496, 299, 48 },
		{ 496, 300, 48 },
		{ 496, 301, 48 },
		{ 496, 302, 48 },
		{ 496, 303, 48 },
		{ 496, 304, 48 },
		{ 496, 305, 48 },
		{ 496, 306, 48 },
		{ 496, 307, 48 },
		{ 496, 308, 48 },
		{ 496, 309, 48 },
		{ 496, 310, 48 },
		{ 496, 311, 48 },
		{ 496, 312, 48 },
		{ 496, 313, 48 },
		{ 496, 314, 48 },
		{ 496, 315, 48 },
		{ 496, 316, 48 },
		{ 496, 317, 48 },
		{ 496, 318, 48 },
		{ 496, 319, 48 },
		{ 496, 320, 48 },
		{ 496, 321, 48 },
		{ 496, 322, 48 },
		{ 496, 323, 48 },
		{ 496, 324, 48 },
		{ 496, 325, 48 },
		{ 496, 326, 48 },
		{ 496, 327, 48 },
		{ 496, 328, 48 },
		{ 496, 329, 48 },
		{ 496, 330, 48 },
		{ 496, 331, 48 },
		{ 496, 332, 48 },
		{ 496, 333, 48 },
		{ 496, 334, 48 },
		{ 496, 335, 48 },
		{ 496, 336, 48 },
		{ 496, 337, 48 },
		{ 496, 338, 48 },
		{ 496, 339, 48 },
		{ 496, 340, 48 },
		{ 496, 341, 48 },
		{ 496, 342, 48 },
		{ 496, 343, 48 },
		{ 496, 344, 48 },
		{ 496, 345, 48 },
		{ 496, 346, 48 },
		{ 497, 288, 48 },
		{ 497, 289, 48 },
		{ 497, 290, 48 },
		{ 497, 291, 48 },
		{ 497, 292, 48 },
		{ 497, 293, 48 },
		{ 497, 294, 48 },
		{ 497, 295, 48 },
		{ 497, 296, 48 },
		{ 497, 297, 48 },
		{ 497, 298, 48 },
		{ 497, 299, 48 },
		{ 497, 300, 48 },
		{ 497, 301, 48 },
		{ 497, 302, 48 },
		{ 497, 303, 48 },
		{ 497, 304, 48 },
		{ 497, 305, 48 },
		{ 497, 306, 48 },
		{ 497, 307, 48 },
		{ 497, 308, 48 },
		{ 497, 309, 48 },
		{ 497, 310, 48 },
		{ 497, 311, 48 },
		{ 497, 312, 48 },
		{ 497, 313, 48 },
		{ 497, 314, 48 },
		{ 497, 315, 48 },
		{ 497, 316, 48 },
		{ 497, 317, 48 },
		{ 497, 318, 48 },
		{ 497, 319, 48 },
		{ 497, 320, 48 },
		{ 497, 321, 48 },
		{ 497, 322, 48 },
		{ 497, 323, 48 },
		{ 497, 324, 48 },
		{ 497, 325, 48 },
		{ 497, 326, 48 },
		{ 497, 327, 48 },
		{ 497, 328, 48 },
		{ 497, 329, 48 },
		{ 497, 330, 48 },
		{ 497, 331, 48 },
		{ 497, 332, 48 },
		{ 497, 333, 48 },
		{ 497, 334, 48 },
		{ 497, 335, 48 },
		{ 497, 336, 48 },
		{ 497, 337, 48 },
		{ 497, 338, 48 },
		{ 497, 339, 48 },
		{ 497, 340, 48 },
		{ 497, 341, 48 },
		{ 497, 342, 48 },
		{ 497, 343, 48 },
		{ 497, 344, 48 },
		{ 497, 345, 48 },
		{ 497, 346, 48 },
		{ 498, 288, 48 },
		{ 498, 289, 48 },
		{ 498, 290, 48 },
		{ 498, 291, 48 },
		{ 498, 292, 48 },
		{ 498, 293, 48 },
		{ 498, 294, 48 },
		{ 498, 295, 48 },
		{ 498, 296, 48 },
		{ 498, 297, 48 },
		{ 498, 298, 48 },
		{ 498, 299, 48 },
		{ 498, 300, 48 },
		{ 498, 301, 48 },
		{ 498, 302, 48 },
		{ 498, 303, 48 },
		{ 498, 304, 48 },
		{ 498, 305, 48 },
		{ 498, 306, 48 },
		{ 498, 307, 48 },
		{ 498, 308, 48 },
		{ 498, 309, 48 },
		{ 498, 310, 48 },
		{ 498, 311, 48 },
		{ 498, 312, 48 },
		{ 498, 313, 48 },
		{ 498, 314, 48 },
		{ 498, 315, 48 },
		{ 498, 316, 48 },
		{ 498, 317, 48 },
		{ 498, 318, 48 },
		{ 498, 319, 48 },
		{ 498, 320, 48 },
		{ 498, 321, 48 },
		{ 498, 322, 48 },
		{ 498, 323, 48 },
		{ 498, 324, 48 },
		{ 498, 325, 48 },
		{ 498, 326, 48 },
		{ 498, 327, 48 },
		{ 498, 328, 48 },
		{ 498, 329, 48 },
		{ 498, 330, 48 },
		{ 498, 331, 48 },
		{ 498, 332, 48 },
		{ 498, 333, 48 },
		{ 498, 334, 48 },
		{ 498, 335, 48 },
		{ 498, 336, 48 },
		{ 498, 337, 48 },
		{ 498, 338, 48 },
		{ 498, 339, 48 },
		{ 498, 340, 48 },
		{ 498, 341, 48 },
		{ 498, 342, 48 },
		{ 498, 343, 48 },
		{ 498, 344, 48 },
		{ 498, 345, 48 },
		{ 498, 346, 48 },
		{ 498, 347, 48 },
		{ 499, 288, 48 },
		{ 499, 289, 48 },
		{ 499, 290, 48 },
		{ 499, 291, 48 },
		{ 499, 292, 48 },
		{ 499, 293, 48 },
		{ 499, 294, 48 },
		{ 499, 295, 48 },
		{ 499, 296, 48 },
		{ 499, 297, 48 },
		{ 499, 298, 48 },
		{ 499, 299, 48 },
		{ 499, 300, 48 },
		{ 499, 301, 48 },
		{ 499, 302, 48 },
		{ 499, 303, 48 },
		{ 499, 304, 48 },
		{ 499, 305, 48 },
		{ 499, 306, 48 },
		{ 499, 307, 48 },
		{ 499, 308, 48 },
		{ 499, 309, 48 },
		{ 499, 310, 48 },
		{ 499, 311, 48 },
		{ 499, 312, 48 },
		{ 499, 313, 48 },
		{ 499, 314, 48 },
		{ 499, 315, 48 },
		{ 499, 316, 48 },
		{ 499, 317, 48 },
		{ 499, 318, 48 },
		{ 499, 319, 48 },
		{ 499, 320, 48 },
		{ 499, 321, 48 },
		{ 499, 322, 48 },
		{ 499, 323, 48 },
		{ 499, 324, 48 },
		{ 499, 325, 48 },
		{ 499, 326, 48 },
		{ 499, 327, 48 },
		{ 499, 328, 48 },
		{ 499, 329, 48 },
		{ 499, 330, 48 },
		{ 499, 331, 48 },
		{ 499, 332, 48 },
		{ 499, 333, 48 },
		{ 499, 334, 48 },
		{ 499, 335, 48 },
		{ 499, 336, 48 },
		{ 499, 337, 48 },
		{ 499, 338, 48 },
		{ 499, 339, 48 },
		{ 499, 340, 48 },
		{ 499, 341, 48 },
		{ 499, 342, 48 },
		{ 499, 343, 48 },
		{ 499, 344, 48 },
		{ 499, 345, 48 },
		{ 499, 346, 48 },
		{ 499, 347, 48 },
		{ 500, 288, 48 },
		{ 500, 289, 48 },
		{ 500, 290, 48 },
		{ 500, 291, 48 },
		{ 500, 292, 48 },
		{ 500, 293, 48 },
		{ 500, 294, 48 },
		{ 500, 295, 48 },
		{ 500, 296, 48 },
		{ 500, 297, 48 },
		{ 500, 298, 48 },
		{ 500, 299, 48 },
		{ 500, 300, 48 },
		{ 500, 301, 48 },
		{ 500, 302, 48 },
		{ 500, 303, 48 },
		{ 500, 304, 48 },
		{ 500, 305, 48 },
		{ 500, 306, 48 },
		{ 500, 307, 48 },
		{ 500, 308, 48 },
		{ 500, 309, 48 },
		{ 500, 310, 48 },
		{ 500, 311, 48 },
		{ 500, 312, 48 },
		{ 500, 313, 48 },
		{ 500, 314, 48 },
		{ 500, 315, 48 },
		{ 500, 316, 48 },
		{ 500, 317, 48 },
		{ 500, 318, 48 },
		{ 500, 319, 48 },
		{ 500, 320, 48 },
		{ 500, 321, 48 },
		{ 500, 322, 48 },
		{ 500, 323, 48 },
		{ 500, 324, 48 },
		{ 500, 325, 48 },
		{ 500, 326, 48 },
		{ 500, 327, 48 },
		{ 500, 328, 48 },
		{ 500, 329, 48 },
		{ 500, 330, 48 },
		{ 500, 331, 48 },
		{ 500, 332, 48 },
		{ 500, 333, 48 },
		{ 500, 334, 48 },
		{ 500, 335, 48 },
		{ 500, 336, 48 },
		{ 500, 337, 48 },
		{ 500, 338, 48 },
		{ 500, 339, 48 },
		{ 500, 340, 48 },
		{ 500, 341, 48 },
		{ 500, 342, 48 },
		{ 500, 343, 48 },
		{ 500, 344, 48 },
		{ 500, 345, 48 },
		{ 500, 346, 48 },
		{ 500, 347, 48 },
		{ 500, 348, 48 },
		{ 501, 288, 48 },
		{ 501, 289, 48 },
		{ 501, 290, 48 },
		{ 501, 291, 48 },
		{ 501, 292, 48 },
		{ 501, 293, 48 },
		{ 501, 294, 48 },
		{ 501, 295, 48 },
		{ 501, 296, 48 },
		{ 501, 297, 48 },
		{ 501, 298, 48 },
		{ 501, 299, 48 },
		{ 501, 300, 48 },
		{ 501, 301, 48 },
		{ 501, 302, 48 },
		{ 501, 303, 48 },
		{ 501, 304, 48 },
		{ 501, 305, 48 },
		{ 501, 306, 48 },
		{ 501, 307, 48 },
		{ 501, 308, 48 },
		{ 501, 309, 48 },
		{ 501, 310, 48 },
		{ 501, 311, 48 },
		{ 501, 312, 48 },
		{ 501, 313, 48 },
		{ 501, 314, 48 },
		{ 501, 315, 48 },
		{ 501, 316, 48 },
		{ 501, 317, 48 },
		{ 501, 318, 48 },
		{ 501, 319, 48 },
		{ 501, 320, 48 },
		{ 501, 321, 48 },
		{ 501, 322, 48 },
		{ 501, 323, 48 },
		{ 501, 324, 48 },
		{ 501, 325, 48 },
		{ 501, 326, 48 },
		{ 501, 327, 48 },
		{ 501, 328, 48 },
		{ 501, 329, 48 },
		{ 501, 330, 48 },
		{ 501, 331, 48 },
		{ 501, 332, 48 },
		{ 501, 333, 48 },
		{ 501, 334, 48 },
		{ 501, 335, 48 },
		{ 501, 336, 48 },
		{ 501, 337, 48 },
		{ 501, 338, 48 },
		{ 501, 339, 48 },
		{ 501, 340, 48 },
		{ 501, 341, 48 },
		{ 501, 342, 48 },
		{ 501, 343, 48 },
		{ 501, 344, 48 },
		{ 501, 345, 48 },
		{ 501, 346, 48 },
		{ 501, 347, 48 },
		{ 501, 348, 48 },
		{ 502, 288, 48 },
		{ 502, 289, 48 },
		{ 502, 290, 48 },
		{ 502, 291, 48 },
		{ 502, 292, 48 },
		{ 502, 293, 48 },
		{ 502, 294, 48 },
		{ 502, 295, 48 },
		{ 502, 296, 48 },
		{ 502, 297, 48 },
		{ 502, 298, 48 },
		{ 502, 299, 48 },
		{ 502, 300, 48 },
		{ 502, 301, 48 },
		{ 502, 302, 48 },
		{ 502, 303, 48 },
		{ 502, 304, 48 },
		{ 502, 305, 48 },
		{ 502, 306, 48 },
		{ 502, 307, 48 },
		{ 502, 308, 48 },
		{ 502, 309, 48 },
		{ 502, 310, 48 },
		{ 502, 311, 48 },
		{ 502, 312, 48 },
		{ 502, 313, 48 },
		{ 502, 314, 48 },
		{ 502, 315, 48 },
		{ 502, 316, 48 },
		{ 502, 317, 48 },
		{ 502, 318, 48 },
		{ 502, 319, 48 },
		{ 502, 320, 48 },
		{ 502, 321, 48 },
		{ 502, 322, 48 },
		{ 502, 323, 48 },
		{ 502, 324, 48 },
		{ 502, 325, 48 },
		{ 502, 326, 48 },
		{ 502, 327, 48 },
		{ 502, 328, 48 },
		{ 502, 329, 48 },
		{ 502, 330, 48 },
		{ 502, 331, 48 },
		{ 502, 332, 48 },
		{ 502, 333, 48 },
		{ 502, 334, 48 },
		{ 502, 335, 48 },
		{ 502, 336, 48 },
		{ 502, 337, 48 },
		{ 502, 338, 48 },
		{ 502, 339, 48 },
		{ 502, 340, 48 },
		{ 502, 341, 48 },
		{ 502, 342, 48 },
		{ 502, 343, 48 },
		{ 502, 344, 48 },
		{ 502, 345, 48 },
		{ 502, 346, 48 },
		{ 502, 347, 48 },
		{ 502, 348, 48 },
		{ 503, 288, 48 },
		{ 503, 289, 48 },
		{ 503, 290, 48 },
		{ 503, 291, 48 },
		{ 503, 292, 48 },
		{ 503, 293, 48 },
		{ 503, 294, 48 },
		{ 503, 295, 48 },
		{ 503, 296, 48 },
		{ 503, 297, 48 },
		{ 503, 298, 48 },
		{ 503, 299, 48 },
		{ 503, 300, 48 },
		{ 503, 301, 48 },
		{ 503, 302, 48 },
		{ 503, 303, 48 },
		{ 503, 304, 48 },
		{ 503, 305, 48 },
		{ 503, 306, 48 },
		{ 503, 307, 48 },
		{ 503, 308, 48 },
		{ 503, 309, 48 },
		{ 503, 310, 48 },
		{ 503, 311, 48 },
		{ 503, 312, 48 },
		{ 503, 313, 48 },
		{ 503, 314, 48 },
		{ 503, 315, 48 },
		{ 503, 316, 48 },
		{ 503, 317, 48 },
		{ 503, 318, 48 },
		{ 503, 319, 48 },
		{ 503, 320, 48 },
		{ 503, 321, 48 },
		{ 503, 322, 48 },
		{ 503, 323, 48 },
		{ 503, 324, 48 },
		{ 503, 325, 48 },
		{ 503, 326, 48 },
		{ 503, 327, 48 },
		{ 503, 328, 48 },
		{ 503, 329, 48 },
		{ 503, 330, 48 },
		{ 503, 331, 48 },
		{ 503, 332, 48 },
		{ 503, 333, 48 },
		{ 503, 334, 48 },
		{ 503, 335, 48 },
		{ 503, 336, 48 },
		{ 503, 337, 48 },
		{ 503, 338, 48 },
		{ 503, 339, 48 },
		{ 503, 340, 48 },
		{ 503, 341, 48 },
		{ 503, 342, 48 },
		{ 503, 343, 48 },
		{ 503, 344, 48 },
		{ 503, 345, 48 },
		{ 503, 346, 48 },
		{ 503, 347, 48 },
		{ 503, 348, 48 },
		{ 503, 349, 48 },
		{ 504, 288, 48 },
		{ 504, 289, 48 },
		{ 504, 290, 48 },
		{ 504, 291, 48 },
		{ 504, 292, 48 },
		{ 504, 293, 48 },
		{ 504, 294, 48 },
		{ 504, 295, 48 },
		{ 504, 296, 48 },
		{ 504, 297, 48 },
		{ 504, 298, 48 },
		{ 504, 299, 48 },
		{ 504, 300, 48 },
		{ 504, 301, 48 },
		{ 504, 302, 48 },
		{ 504, 303, 48 },
		{ 504, 304, 48 },
		{ 504, 305, 48 },
		{ 504, 306, 48 },
		{ 504, 307, 48 },
		{ 504, 308, 48 },
		{ 504, 309, 48 },
		{ 504, 310, 48 },
		{ 504, 311, 48 },
		{ 504, 312, 48 },
		{ 504, 313, 48 },
		{ 504, 314, 48 },
		{ 504, 315, 48 },
		{ 504, 316, 48 },
		{ 504, 317, 48 },
		{ 504, 318, 48 },
		{ 504, 319, 48 },
		{ 504, 320, 48 },
		{ 504, 321, 48 },
		{ 504, 322, 48 },
		{ 504, 323, 48 },
		{ 504, 324, 48 },
		{ 504, 325, 48 },
		{ 504, 326, 48 },
		{ 504, 327, 48 },
		{ 504, 328, 48 },
		{ 504, 329, 48 },
		{ 504, 330, 48 },
		{ 504, 331, 48 },
		{ 504, 332, 48 },
		{ 504, 333, 48 },
		{ 504, 334, 48 },
		{ 504, 335, 48 },
		{ 504, 336, 48 },
		{ 504, 337, 48 },
		{ 504, 338, 48 },
		{ 504, 339, 48 },
		{ 504, 340, 48 },
		{ 504, 341, 48 },
		{ 504, 342, 48 },
		{ 504, 343, 48 },
		{ 504, 344, 48 },
		{ 504, 345, 48 },
		{ 504, 346, 48 },
		{ 504, 347, 48 },
		{ 504, 348, 48 },
		{ 504, 349, 48 },
		{ 505, 289, 48 },
		{ 505, 290, 48 },
		{ 505, 291, 48 },
		{ 505, 292, 48 },
		{ 505, 293, 48 },
		{ 505, 294, 48 },
		{ 505, 295, 48 },
		{ 505, 296, 48 },
		{ 505, 297, 48 },
		{ 505, 298, 48 },
		{ 505, 299, 48 },
		{ 505, 300, 48 },
		{ 505, 301, 48 },
		{ 505, 302, 48 },
		{ 505, 303, 48 },
		{ 505, 304, 48 },
		{ 505, 305, 48 },
		{ 505, 306, 48 },
		{ 505, 307, 48 },
		{ 505, 308, 48 },
		{ 505, 309, 48 },
		{ 505, 310, 48 },
		{ 505, 311, 48 },
		{ 505, 312, 48 },
		{ 505, 313, 48 },
		{ 505, 314, 48 },
		{ 505, 315, 48 },
		{ 505, 316, 48 },
		{ 505, 317, 48 },
		{ 505, 318, 48 },
		{ 505, 319, 48 },
		{ 505, 320, 48 },
		{ 505, 321, 48 },
		{ 505, 322, 48 },
		{ 505, 323, 48 },
		{ 505, 324, 48 },
		{ 505, 325, 48 },
		{ 505, 326, 48 },
		{ 505, 327, 48 },
		{ 505, 328, 48 },
		{ 505, 329, 48 },
		{ 505, 330, 48 },
		{ 505, 331, 48 },
		{ 505, 332, 48 },
		{ 505, 333, 48 },
		{ 505, 334, 48 },
		{ 505, 335, 48 },
		{ 505, 336, 48 },
		{ 505, 337, 48 },
		{ 505, 338, 48 },
		{ 505, 339, 48 },
		{ 505, 340, 48 },
		{ 505, 341, 48 },
		{ 505, 342, 48 },
		{ 505, 343, 48 },
		{ 505, 344, 48 },
		{ 505, 345, 48 },
		{ 505, 346, 48 },
		{ 505, 347, 48 },
		{ 505, 348, 48 },
		{ 505, 349, 48 },
		{ 506, 289, 48 },
		{ 506, 290, 48 },
		{ 506, 291, 48 },
		{ 506, 292, 48 },
		{ 506, 293, 48 },
		{ 506, 294, 48 },
		{ 506, 295, 48 },
		{ 506, 296, 48 },
		{ 506, 297, 48 },
		{ 506, 298, 48 },
		{ 506, 299, 48 },
		{ 506, 300, 48 },
		{ 506, 301, 48 },
		{ 506, 302, 48 },
		{ 506, 303, 48 },
		{ 506, 304, 48 },
		{ 506, 305, 48 },
		{ 506, 306, 48 },
		{ 506, 307, 48 },
		{ 506, 308, 48 },
		{ 506, 309, 48 },
		{ 506, 310, 48 },
		{ 506, 311, 48 },
		{ 506, 312, 48 },
		{ 506, 313, 48 },
		{ 506, 314, 48 },
		{ 506, 315, 48 },
		{ 506, 316, 48 },
		{ 506, 317, 48 },
		{ 506, 318, 48 },
		{ 506, 319, 48 },
		{ 506, 320, 48 },
		{ 506, 321, 48 },
		{ 506, 322, 48 },
		{ 506, 323, 48 },
		{ 506, 324, 48 },
		{ 506, 325, 48 },
		{ 506, 326, 48 },
		{ 506, 327, 48 },
		{ 506, 328, 48 },
		{ 506, 329, 48 },
		{ 506, 330, 48 },
		{ 506, 331, 48 },
		{ 506, 332, 48 },
		{ 506, 333, 48 },
		{ 506, 334, 48 },
		{ 506, 335, 48 },
		{ 506, 336, 48 },
		{ 506, 337, 48 },
		{ 506, 338, 48 },
		{ 506, 339, 48 },
		{ 506, 340, 48 },
		{ 506, 341, 48 },
		{ 506, 342, 48 },
		{ 506, 343, 48 },
		{ 506, 344, 48 },
		{ 506, 345, 48 },
		{ 506, 346, 48 },
		{ 506, 347, 48 },
		{ 506, 348, 48 },
		{ 506, 349, 48 },
		{ 506, 350, 48 },
		{ 507, 289, 48 },
		{ 507, 290, 48 },
		{ 507, 291, 48 },
		{ 507, 292, 48 },
		{ 507, 293, 48 },
		{ 507, 294, 48 },
		{ 507, 295, 48 },
		{ 507, 296, 48 },
		{ 507, 297, 48 },
		{ 507, 298, 48 },
		{ 507, 299, 48 },
		{ 507, 300, 48 },
		{ 507, 301, 48 },
		{ 507, 302, 48 },
		{ 507, 303, 48 },
		{ 507, 304, 48 },
		{ 507, 305, 48 },
		{ 507, 306, 48 },
		{ 507, 307, 48 },
		{ 507, 308, 48 },
		{ 507, 309, 48 },
		{ 507, 310, 48 },
		{ 507, 311, 48 },
		{ 507, 312, 48 },
		{ 507, 313, 48 },
		{ 507, 314, 48 },
		{ 507, 315, 48 },
		{ 507, 316, 48 },
		{ 507, 317, 48 },
		{ 507, 318, 48 },
		{ 507, 319, 48 },
		{ 507, 320, 48 },
		{ 507, 321, 48 },
		{ 507, 322, 48 },
		{ 507, 323, 48 },
		{ 507, 324, 48 },
		{ 507, 325, 48 },
		{ 507, 326, 48 },
		{ 507, 327, 48 },
		{ 507, 328, 48 },
		{ 507, 329, 48 },
		{ 507, 330, 48 },
		{ 507, 331, 48 },
		{ 507, 332, 48 },
		{ 507, 333, 48 },
		{ 507, 334, 48 },
		{ 507, 335, 48 },
		{ 507, 336, 48 },
		{ 507, 337, 48 },
		{ 507, 338, 48 },
		{ 507, 339, 48 },
		{ 507, 340, 48 },
		{ 507, 341, 48 },
		{ 507, 342, 48 },
		{ 507, 343, 48 },
		{ 507, 344, 48 },
		{ 507, 345, 48 },
		{ 507, 346, 48 },
		{ 507, 347, 48 },
		{ 507, 348, 48 },
		{ 507, 349, 48 },
		{ 507, 350, 48 },
		{ 508, 289, 48 },
		{ 508, 290, 48 },
		{ 508, 291, 48 },
		{ 508, 292, 48 },
		{ 508, 293, 48 },
		{ 508, 294, 48 },
		{ 508, 295, 48 },
		{ 508, 296, 48 },
		{ 508, 297, 48 },
		{ 508, 298, 48 },
		{ 508, 299, 48 },
		{ 508, 300, 48 },
		{ 508, 301, 48 },
		{ 508, 302, 48 },
		{ 508, 303, 48 },
		{ 508, 304, 48 },
		{ 508, 305, 48 },
		{ 508, 306, 48 },
		{ 508, 307, 48 },
		{ 508, 308, 48 },
		{ 508, 309, 48 },
		{ 508, 310, 48 },
		{ 508, 311, 48 },
		{ 508, 312, 48 },
		{ 508, 313, 48 },
		{ 508, 314, 48 },
		{ 508, 315, 48 },
		{ 508, 316, 48 },
		{ 508, 317, 48 },
		{ 508, 318, 48 },
		{ 508, 319, 48 },
		{ 508, 320, 48 },
		{ 508, 321, 48 },
		{ 508, 322, 48 },
		{ 508, 323, 48 },
		{ 508, 324, 48 },
		{ 508, 325, 48 },
		{ 508, 326, 48 },
		{ 508, 327, 48 },
		{ 508, 328, 48 },
		{ 508, 329, 48 },
		{ 508, 330, 48 },
		{ 508, 331, 48 },
		{ 508, 332, 48 },
		{ 508, 333, 48 },
		{ 508, 334, 48 },
		{ 508, 335, 48 },
		{ 508, 336, 48 },
		{ 508, 337, 48 },
		{ 508, 338, 48 },
		{ 508, 339, 48 },
		{ 508, 340, 48 },
		{ 508, 341, 48 },
		{ 508, 342, 48 },
		{ 508, 343, 48 },
		{ 508, 344, 48 },
		{ 508, 345, 48 },
		{ 508, 346, 48 },
		{ 508, 347, 48 },
		{ 508, 348, 48 },
		{ 508, 349, 48 },
		{ 508, 350, 48 },
		{ 508, 351, 48 },
		{ 509, 289, 48 },
		{ 509, 290, 48 },
		{ 509, 291, 48 },
		{ 509, 292, 48 },
		{ 509, 293, 48 },
		{ 509, 294, 48 },
		{ 509, 295, 48 },
		{ 509, 296, 48 },
		{ 509, 297, 48 },
		{ 509, 298, 48 },
		{ 509, 299, 48 },
		{ 509, 300, 48 },
		{ 509, 301, 48 },
		{ 509, 302, 48 },
		{ 509, 303, 48 },
		{ 509, 304, 48 },
		{ 509, 305, 48 },
		{ 509, 306, 48 },
		{ 509, 307, 48 },
		{ 509, 308, 48 },
		{ 509, 309, 48 },
		{ 509, 310, 48 },
		{ 509, 311, 48 },
		{ 509, 312, 48 },
		{ 509, 313, 48 },
		{ 509, 314, 48 },
		{ 509, 315, 48 },
		{ 509, 316, 48 },
		{ 509, 317, 48 },
		{ 509, 318, 48 },
		{ 509, 319, 48 },
		{ 509, 320, 48 },
		{ 509, 321, 48 },
		{ 509, 322, 48 },
		{ 509, 323, 48 },
		{ 509, 324, 48 },
		{ 509, 325, 48 },
		{ 509, 326, 48 },
		{ 509, 327, 48 },
		{ 509, 328, 48 },
		{ 509, 329, 48 },
		{ 509, 330, 48 },
		{ 509, 331, 48 },
		{ 509, 332, 48 },
		{ 509, 333, 48 },
		{ 509, 334, 48 },
		{ 509, 335, 48 },
		{ 509, 336, 48 },
		{ 509, 337, 48 },
		{ 509, 338, 48 },
		{ 509, 339, 48 },
		{ 509, 340, 48 },
		{ 509, 341, 48 },
		{ 509, 342, 48 },
		{ 509, 343, 48 },
		{ 509, 344, 48 },
		{ 509, 345, 48 },
		{ 509, 346, 48 },
		{ 509, 347, 48 },
		{ 509, 348, 48 },
		{ 509, 349, 48 },
		{ 509, 350, 48 },
		{ 509, 351, 48 },
		{ 510, 289, 48 },
		{ 510, 290, 48 },
		{ 510, 291, 48 },
		{ 510, 292, 48 },
		{ 510, 293, 48 },
		{ 510, 294, 48 },
		{ 510, 295, 48 },
		{ 510, 296, 48 },
		{ 510, 297, 48 },
		{ 510, 298, 48 },
		{ 510, 299, 48 },
		{ 510, 300, 48 },
		{ 510, 301, 48 },
		{ 510, 302, 48 },
		{ 510, 303, 48 },
		{ 510, 304, 48 },
		{ 510, 305, 48 },
		{ 510, 306, 48 },
		{ 510, 307, 48 },
		{ 510, 308, 48 },
		{ 510, 309, 48 },
		{ 510, 310, 48 },
		{ 510, 311, 48 },
		{ 510, 312, 48 },
		{ 510, 313, 48 },
		{ 510, 314, 48 },
		{ 510, 315, 48 },
		{ 510, 316, 48 },
		{ 510, 317, 48 },
		{ 510, 318, 48 },
		{ 510, 319, 48 },
		{ 510, 320, 48 },
		{ 510, 321, 48 },
		{ 510, 322, 48 },
		{ 510, 323, 48 },
		{ 510, 324, 48 },
		{ 510, 325, 48 },
		{ 510, 326, 48 },
		{ 510, 327, 48 },
		{ 510, 328, 48 },
		{ 510, 329, 48 },
		{ 510, 330, 48 },
		{ 510, 331, 48 },
		{ 510, 332, 48 },
		{ 510, 333, 48 },
		{ 510, 334, 48 },
		{ 510, 335, 48 },
		{ 510, 336, 48 },
		{ 510, 337, 48 },
		{ 510, 338, 48 },
		{ 510, 339, 48 },
		{ 510, 340, 48 },
		{ 510, 341, 48 },
		{ 510, 342, 48 },
		{ 510, 343, 48 },
		{ 510, 344, 48 },
		{ 510, 345, 48 },
		{ 510, 346, 48 },
		{ 510, 347, 48 },
		{ 510, 348, 48 },
		{ 510, 349, 48 },
		{ 510, 350, 48 },
		{ 510, 351, 48 },
		{ 511, 289, 48 },
		{ 511, 290, 48 },
		{ 511, 291, 48 },
		{ 511, 292, 48 },
		{ 511, 293, 48 },
		{ 511, 294, 48 },
		{ 511, 295, 48 },
		{ 511, 296, 48 },
		{ 511, 297, 48 },
		{ 511, 298, 48 },
		{ 511, 299, 48 },
		{ 511, 300, 48 },
		{ 511, 301, 48 },
		{ 511, 302, 48 },
		{ 511, 303, 48 },
		{ 511, 304, 48 },
		{ 511, 305, 48 },
		{ 511, 306, 48 },
		{ 511, 307, 48 },
		{ 511, 308, 48 },
		{ 511, 309, 48 },
		{ 511, 310, 48 },
		{ 511, 311, 48 },
		{ 511, 312, 48 },
		{ 511, 313, 48 },
		{ 511, 314, 48 },
		{ 511, 315, 48 },
		{ 511, 316, 48 },
		{ 511, 317, 48 },
		{ 511, 318, 48 },
		{ 511, 319, 48 },
		{ 511, 320, 48 },
		{ 511, 321, 48 },
		{ 511, 322, 48 },
		{ 511, 323, 48 },
		{ 511, 324, 48 },
		{ 511, 325, 48 },
		{ 511, 326, 48 },
		{ 511, 327, 48 },
		{ 511, 328, 48 },
		{ 511, 329, 48 },
		{ 511, 330, 48 },
		{ 511, 331, 48 },
		{ 511, 332, 48 },
		{ 511, 333, 48 },
		{ 511, 334, 48 },
		{ 511, 335, 48 },
		{ 511, 336, 48 },
		{ 511, 337, 48 },
		{ 511, 338, 48 },
		{ 511, 339, 48 },
		{ 511, 340, 48 },
		{ 511, 341, 48 },
		{ 511, 342, 48 },
		{ 511, 343, 48 },
		{ 511, 344, 48 },
		{ 511, 345, 48 },
		{ 511, 346, 48 },
		{ 511, 347, 48 },
		{ 511, 348, 48 },
		{ 511, 349, 48 },
		{ 511, 350, 48 },
		{ 511, 351, 48 },
		{ 511, 352, 48 },
		{ 512, 289, 48 },
		{ 512, 290, 48 },
		{ 512, 291, 48 },
		{ 512, 292, 48 },
		{ 512, 293, 48 },
		{ 512, 294, 48 },
		{ 512, 295, 48 },
		{ 512, 296, 48 },
		{ 512, 297, 48 },
		{ 512, 298, 48 },
		{ 512, 299, 48 },
		{ 512, 300, 48 },
		{ 512, 301, 48 },
		{ 512, 302, 48 },
		{ 512, 303, 48 },
		{ 512, 304, 48 },
		{ 512, 305, 48 },
		{ 512, 306, 48 },
		{ 512, 307, 48 },
		{ 512, 308, 48 },
		{ 512, 309, 48 },
		{ 512, 310, 48 },
		{ 512, 311, 48 },
		{ 512, 312, 48 },
		{ 512, 313, 48 },
		{ 512, 314, 48 },
		{ 512, 315, 48 },
		{ 512, 316, 48 },
		{ 512, 317, 48 },
		{ 512, 318, 48 },
		{ 512, 319, 48 },
		{ 512, 320, 48 },
		{ 512, 321, 48 },
		{ 512, 322, 48 },
		{ 512, 323, 48 },
		{ 512, 324, 48 },
		{ 512, 325, 48 },
		{ 512, 326, 48 },
		{ 512, 327, 48 },
		{ 512, 328, 48 },
		{ 512, 329, 48 },
		{ 512, 330, 48 },
		{ 512, 331, 48 },
		{ 512, 332, 48 },
		{ 512, 333, 48 },
		{ 512, 334, 48 },
		{ 512, 335, 48 },
		{ 512, 336, 48 },
		{ 512, 337, 48 },
		{ 512, 338, 48 },
		{ 512, 339, 48 },
		{ 512, 340, 48 },
		{ 512, 341, 48 },
		{ 512, 342, 48 },
		{ 512, 343, 48 },
		{ 512, 344, 48 },
		{ 512, 345, 48 },
		{ 512, 346, 48 },
		{ 512, 347, 48 },
		{ 512, 348, 48 },
		{ 512, 349, 48 },
		{ 512, 350, 48 },
		{ 512, 351, 48 },
		{ 512, 352, 48 },
		{ 1, 130, 49 },
		{ 1, 131, 49 },
		{ 1, 132, 49 },
		{ 1, 133, 49 },
		{ 1, 134, 49 },
		{ 1, 135, 49 },
		{ 1, 136, 49 },
		{ 1, 137, 49 },
		{ 1, 138, 49 },
		{ 1, 139, 49 },
		{ 1, 140, 49 },
		{ 1, 141, 49 },
		{ 1, 142, 49 },
		{ 1, 143, 49 },
		{ 1, 144, 49 },
		{ 1, 145, 49 },
		{ 1, 146, 49 },
		{ 1, 147, 49 },
		{ 1, 148, 49 },
		{ 1, 149, 49 },
		{ 1, 150, 49 },
		{ 1, 151, 49 },
		{ 1, 152, 49 },
		{ 1, 153, 49 },
		{ 1, 154, 49 },
		{ 1, 155, 49 },
		{ 1, 156, 49 },
		{ 1, 157, 49 },
		{ 1, 158, 49 },
		{ 1, 159, 49 },
		{ 1, 160, 49 },
		{ 1, 161, 49 },
		{ 1, 162, 49 },
		{ 1, 163, 49 },
		{ 1, 164, 49 },
		{ 1, 165, 49 },
		{ 1, 166, 49 },
		{ 1, 167, 49 },
		{ 1, 168, 49 },
		{ 1, 169, 49 },
		{ 1, 170, 49 },
		{ 1, 171, 49 },
		{ 1, 172, 49 },
		{ 1, 173, 49 },
		{ 1, 174, 49 },
		{ 1, 175, 49 },
		{ 1, 176, 49 },
		{ 1, 177, 49 },
		{ 1, 178, 49 },
		{ 1, 179, 49 },
		{ 1, 180, 49 },
		{ 1, 181, 49 },
		{ 1, 182, 49 },
		{ 1, 183, 49 },
		{ 1, 184, 49 },
		{ 1, 185, 49 },
		{ 1, 186, 49 },
		{ 1, 187, 49 },
		{ 1, 188, 49 },
		{ 1, 189, 49 },
		{ 1, 190, 49 },
		{ 1, 191, 49 },
		{ 1, 192, 49 },
		{ 2, 130, 49 },
		{ 2, 131, 49 },
		{ 2, 132, 49 },
		{ 2, 133, 49 },
		{ 2, 134, 49 },
		{ 2, 135, 49 },
		{ 2, 136, 49 },
		{ 2, 137, 49 },
		{ 2, 138, 49 },
		{ 2, 139, 49 },
		{ 2, 140, 49 },
		{ 2, 141, 49 },
		{ 2, 142, 49 },
		{ 2, 143, 49 },
		{ 2, 144, 49 },
		{ 2, 145, 49 },
		{ 2, 146, 49 },
		{ 2, 147, 49 },
		{ 2, 148, 49 },
		{ 2, 149, 49 },
		{ 2, 150, 49 },
		{ 2, 151, 49 },
		{ 2, 152, 49 },
		{ 2, 153, 49 },
		{ 2, 154, 49 },
		{ 2, 155, 49 },
		{ 2, 156, 49 },
		{ 2, 157, 49 },
		{ 2, 158, 49 },
		{ 2, 159, 49 },
		{ 2, 160, 49 },
		{ 2, 161, 49 },
		{ 2, 162, 49 },
		{ 2, 163, 49 },
		{ 2, 164, 49 },
		{ 2, 165, 49 },
		{ 2, 166, 49 },
		{ 2, 167, 49 },
		{ 2, 168, 49 },
		{ 2, 169, 49 },
		{ 2, 170, 49 },
		{ 2, 171, 49 },
		{ 2, 172, 49 },
		{ 2, 173, 49 },
		{ 2, 174, 49 },
		{ 2, 175, 49 },
		{ 2, 176, 49 },
		{ 2, 177, 49 },
		{ 2, 178, 49 },
		{ 2, 179, 49 },
		{ 2, 180, 49 },
		{ 2, 181, 49 },
		{ 2, 182, 49 },
		{ 2, 183, 49 },
		{ 2, 184, 49 },
		{ 2, 185, 49 },
		{ 2, 186, 49 },
		{ 2, 187, 49 },
		{ 2, 188, 49 },
		{ 2, 189, 49 },
		{ 2, 190, 49 },
		{ 2, 191, 49 },
		{ 2, 192, 49 },
		{ 2, 193, 49 },
		{ 3, 131, 49 },
		{ 3, 132, 49 },
		{ 3, 133, 49 },
		{ 3, 134, 49 },
		{ 3, 135, 49 },
		{ 3, 136, 49 },
		{ 3, 137, 49 },
		{ 3, 138, 49 },
		{ 3, 139, 49 },
		{ 3, 140, 49 },
		{ 3, 141, 49 },
		{ 3, 142, 49 },
		{ 3, 143, 49 },
		{ 3, 144, 49 },
		{ 3, 145, 49 },
		{ 3, 146, 49 },
		{ 3, 147, 49 },
		{ 3, 148, 49 },
		{ 3, 149, 49 },
		{ 3, 150, 49 },
		{ 3, 151, 49 },
		{ 3, 152, 49 },
		{ 3, 153, 49 },
		{ 3, 154, 49 },
		{ 3, 155, 49 },
		{ 3, 156, 49 },
		{ 3, 157, 49 },
		{ 3, 158, 49 },
		{ 3, 159, 49 },
		{ 3, 160, 49 },
		{ 3, 161, 49 },
		{ 3, 162, 49 },
		{ 3, 163, 49 },
		{ 3, 164, 49 },
		{ 3, 165, 49 },
		{ 3, 166, 49 },
		{ 3, 167, 49 },
		{ 3, 168, 49 },
		{ 3, 169, 49 },
		{ 3, 170, 49 },
		{ 3, 171, 49 },
		{ 3, 172, 49 },
		{ 3, 173, 49 },
		{ 3, 174, 49 },
		{ 3, 175, 49 },
		{ 3, 176, 49 },
		{ 3, 177, 49 },
		{ 3, 178, 49 },
		{ 3, 179, 49 },
		{ 3, 180, 49 },
		{ 3, 181, 49 },
		{ 3, 182, 49 },
		{ 3, 183, 49 },
		{ 3, 184, 49 },
		{ 3, 185, 49 },
		{ 3, 186, 49 },
		{ 3, 187, 49 },
		{ 3, 188, 49 },
		{ 3, 189, 49 },
		{ 3, 190, 49 },
		{ 3, 191, 49 },
		{ 3, 192, 49 },
		{ 3, 193, 49 },
		{ 4, 131, 49 },
		{ 4, 132, 49 },
		{ 4, 133, 49 },
		{ 4, 134, 49 },
		{ 4, 135, 49 },
		{ 4, 136, 49 },
		{ 4, 137, 49 },
		{ 4, 138, 49 },
		{ 4, 139, 49 },
		{ 4, 140, 49 },
		{ 4, 141, 49 },
		{ 4, 142, 49 },
		{ 4, 143, 49 },
		{ 4, 144, 49 },
		{ 4, 145, 49 },
		{ 4, 146, 49 },
		{ 4, 147, 49 },
		{ 4, 148, 49 },
		{ 4, 149, 49 },
		{ 4, 150, 49 },
		{ 4, 151, 49 },
		{ 4, 152, 49 },
		{ 4, 153, 49 },
		{ 4, 154, 49 },
		{ 4, 155, 49 },
		{ 4, 156, 49 },
		{ 4, 157, 49 },
		{ 4, 158, 49 },
		{ 4, 159, 49 },
		{ 4, 160, 49 },
		{ 4, 161, 49 },
		{ 4, 162, 49 },
		{ 4, 163, 49 },
		{ 4, 164, 49 },
		{ 4, 165, 49 },
		{ 4, 166, 49 },
		{ 4, 167, 49 },
		{ 4, 168, 49 },
		{ 4, 169, 49 },
		{ 4, 170, 49 },
		{ 4, 171, 49 },
		{ 4, 172, 49 },
		{ 4, 173, 49 },
		{ 4, 174, 49 },
		{ 4, 175, 49 },
		{ 4, 176, 49 },
		{ 4, 177, 49 },
		{ 4, 178, 49 },
		{ 4, 179, 49 },
		{ 4, 180, 49 },
		{ 4, 181, 49 },
		{ 4, 182, 49 },
		{ 4, 183, 49 },
		{ 4, 184, 49 },
		{ 4, 185, 49 },
		{ 4, 186, 49 },
		{ 4, 187, 49 },
		{ 4, 188, 49 },
		{ 4, 189, 49 },
		{ 4, 190, 49 },
		{ 4, 191, 49 },
		{ 4, 192, 49 },
		{ 4, 193, 49 },
		{ 5, 132, 49 },
		{ 5, 133, 49 },
		{ 5, 134, 49 },
		{ 5, 135, 49 },
		{ 5, 136, 49 },
		{ 5, 137, 49 },
		{ 5, 138, 49 },
		{ 5, 139, 49 },
		{ 5, 140, 49 },
		{ 5, 141, 49 },
		{ 5, 142, 49 },
		{ 5, 143, 49 },
		{ 5, 144, 49 },
		{ 5, 145, 49 },
		{ 5, 146, 49 },
		{ 5, 147, 49 },
		{ 5, 148, 49 },
		{ 5, 149, 49 },
		{ 5, 150, 49 },
		{ 5, 151, 49 },
		{ 5, 152, 49 },
		{ 5, 153, 49 },
		{ 5, 154, 49 },
		{ 5, 155, 49 },
		{ 5, 156, 49 },
		{ 5, 157, 49 },
		{ 5, 158, 49 },
		{ 5, 159, 49 },
		{ 5, 160, 49 },
		{ 5, 161, 49 },
		{ 5, 162, 49 },
		{ 5, 163, 49 },
		{ 5, 164, 49 },
		{ 5, 165, 49 },
		{ 5, 166, 49 },
		{ 5, 167, 49 },
		{ 5, 168, 49 },
		{ 5, 169, 49 },
		{ 5, 170, 49 },
		{ 5, 171, 49 },
		{ 5, 172, 49 },
		{ 5, 173, 49 },
		{ 5, 174, 49 },
		{ 5, 175, 49 },
		{ 5, 176, 49 },
		{ 5, 177, 49 },
		{ 5, 178, 49 },
		{ 5, 179, 49 },
		{ 5, 180, 49 },
		{ 5, 181, 49 },
		{ 5, 182, 49 },
		{ 5, 183, 49 },
		{ 5, 184, 49 },
		{ 5, 185, 49 },
		{ 5, 186, 49 },
		{ 5, 187, 49 },
		{ 5, 188, 49 },
		{ 5, 189, 49 },
		{ 5, 190, 49 },
		{ 5, 191, 49 },
		{ 5, 192, 49 },
		{ 5, 193, 49 },
		{ 6, 132, 49 },
		{ 6, 133, 49 },
		{ 6, 134, 49 },
		{ 6, 135, 49 },
		{ 6, 136, 49 },
		{ 6, 137, 49 },
		{ 6, 138, 49 },
		{ 6, 139, 49 },
		{ 6, 140, 49 },
		{ 6, 141, 49 },
		{ 6, 142, 49 },
		{ 6, 143, 49 },
		{ 6, 144, 49 },
		{ 6, 145, 49 },
		{ 6, 146, 49 },
		{ 6, 147, 49 },
		{ 6, 148, 49 },
		{ 6, 149, 49 },
		{ 6, 150, 49 },
		{ 6, 151, 49 },
		{ 6, 152, 49 },
		{ 6, 153, 49 },
		{ 6, 154, 49 },
		{ 6, 155, 49 },
		{ 6, 156, 49 },
		{ 6, 157, 49 },
		{ 6, 158, 49 },
		{ 6, 159, 49 },
		{ 6, 160, 49 },
		{ 6, 161, 49 },
		{ 6, 162, 49 },
		{ 6, 163, 49 },
		{ 6, 164, 49 },
		{ 6, 165, 49 },
		{ 6, 166, 49 },
		{ 6, 167, 49 },
		{ 6, 168, 49 },
		{ 6, 169, 49 },
		{ 6, 170, 49 },
		{ 6, 171, 49 },
		{ 6, 172, 49 },
		{ 6, 173, 49 },
		{ 6, 174, 49 },
		{ 6, 175, 49 },
		{ 6, 176, 49 },
		{ 6, 177, 49 },
		{ 6, 178, 49 },
		{ 6, 179, 49 },
		{ 6, 180, 49 },
		{ 6, 181, 49 },
		{ 6, 182, 49 },
		{ 6, 183, 49 },
		{ 6, 184, 49 },
		{ 6, 185, 49 },
		{ 6, 186, 49 },
		{ 6, 187, 49 },
		{ 6, 188, 49 },
		{ 6, 189, 49 },
		{ 6, 190, 49 },
		{ 6, 191, 49 },
		{ 6, 192, 49 },
		{ 6, 193, 49 },
		{ 6, 194, 49 },
		{ 7, 133, 49 },
		{ 7, 134, 49 },
		{ 7, 135, 49 },
		{ 7, 136, 49 },
		{ 7, 137, 49 },
		{ 7, 138, 49 },
		{ 7, 139, 49 },
		{ 7, 140, 49 },
		{ 7, 141, 49 },
		{ 7, 142, 49 },
		{ 7, 143, 49 },
		{ 7, 144, 49 },
		{ 7, 145, 49 },
		{ 7, 146, 49 },
		{ 7, 147, 49 },
		{ 7, 148, 49 },
		{ 7, 149, 49 },
		{ 7, 150, 49 },
		{ 7, 151, 49 },
		{ 7, 152, 49 },
		{ 7, 153, 49 },
		{ 7, 154, 49 },
		{ 7, 155, 49 },
		{ 7, 156, 49 },
		{ 7, 157, 49 },
		{ 7, 158, 49 },
		{ 7, 159, 49 },
		{ 7, 160, 49 },
		{ 7, 161, 49 },
		{ 7, 162, 49 },
		{ 7, 163, 49 },
		{ 7, 164, 49 },
		{ 7, 165, 49 },
		{ 7, 166, 49 },
		{ 7, 167, 49 },
		{ 7, 168, 49 },
		{ 7, 169, 49 },
		{ 7, 170, 49 },
		{ 7, 171, 49 },
		{ 7, 172, 49 },
		{ 7, 173, 49 },
		{ 7, 174, 49 },
		{ 7, 175, 49 },
		{ 7, 176, 49 },
		{ 7, 177, 49 },
		{ 7, 178, 49 },
		{ 7, 179, 49 },
		{ 7, 180, 49 },
		{ 7, 181, 49 },
		{ 7, 182, 49 },
		{ 7, 183, 49 },
		{ 7, 184, 49 },
		{ 7, 185, 49 },
		{ 7, 186, 49 },
		{ 7, 187, 49 },
		{ 7, 188, 49 },
		{ 7, 189, 49 },
		{ 7, 190, 49 },
		{ 7, 191, 49 },
		{ 7, 192, 49 },
		{ 7, 193, 49 },
		{ 7, 194, 49 },
		{ 8, 133, 49 },
		{ 8, 134, 49 },
		{ 8, 135, 49 },
		{ 8, 136, 49 },
		{ 8, 137, 49 },
		{ 8, 138, 49 },
		{ 8, 139, 49 },
		{ 8, 140, 49 },
		{ 8, 141, 49 },
		{ 8, 142, 49 },
		{ 8, 143, 49 },
		{ 8, 144, 49 },
		{ 8, 145, 49 },
		{ 8, 146, 49 },
		{ 8, 147, 49 },
		{ 8, 148, 49 },
		{ 8, 149, 49 },
		{ 8, 150, 49 },
		{ 8, 151, 49 },
		{ 8, 152, 49 },
		{ 8, 153, 49 },
		{ 8, 154, 49 },
		{ 8, 155, 49 },
		{ 8, 156, 49 },
		{ 8, 157, 49 },
		{ 8, 158, 49 },
		{ 8, 159, 49 },
		{ 8, 160, 49 },
		{ 8, 161, 49 },
		{ 8, 162, 49 },
		{ 8, 163, 49 },
		{ 8, 164, 49 },
		{ 8, 165, 49 },
		{ 8, 166, 49 },
		{ 8, 167, 49 },
		{ 8, 168, 49 },
		{ 8, 169, 49 },
		{ 8, 170, 49 },
		{ 8, 171, 49 },
		{ 8, 172, 49 },
		{ 8, 173, 49 },
		{ 8, 174, 49 },
		{ 8, 175, 49 },
		{ 8, 176, 49 },
		{ 8, 177, 49 },
		{ 8, 178, 49 },
		{ 8, 179, 49 },
		{ 8, 180, 49 },
		{ 8, 181, 49 },
		{ 8, 182, 49 },
		{ 8, 183, 49 },
		{ 8, 184, 49 },
		{ 8, 185, 49 },
		{ 8, 186, 49 },
		{ 8, 187, 49 },
		{ 8, 188, 49 },
		{ 8, 189, 49 },
		{ 8, 190, 49 },
		{ 8, 191, 49 },
		{ 8, 192, 49 },
		{ 8, 193, 49 },
		{ 8, 194, 49 },
		{ 9, 134, 49 },
		{ 9, 135, 49 },
		{ 9, 136, 49 },
		{ 9, 137, 49 },
		{ 9, 138, 49 },
		{ 9, 139, 49 },
		{ 9, 140, 49 },
		{ 9, 141, 49 },
		{ 9, 142, 49 },
		{ 9, 143, 49 },
		{ 9, 144, 49 },
		{ 9, 145, 49 },
		{ 9, 146, 49 },
		{ 9, 147, 49 },
		{ 9, 148, 49 },
		{ 9, 149, 49 },
		{ 9, 150, 49 },
		{ 9, 151, 49 },
		{ 9, 152, 49 },
		{ 9, 153, 49 },
		{ 9, 154, 49 },
		{ 9, 155, 49 },
		{ 9, 156, 49 },
		{ 9, 157, 49 },
		{ 9, 158, 49 },
		{ 9, 159, 49 },
		{ 9, 160, 49 },
		{ 9, 161, 49 },
		{ 9, 162, 49 },
		{ 9, 163, 49 },
		{ 9, 164, 49 },
		{ 9, 165, 49 },
		{ 9, 166, 49 },
		{ 9, 167, 49 },
		{ 9, 168, 49 },
		{ 9, 169, 49 },
		{ 9, 170, 49 },
		{ 9, 171, 49 },
		{ 9, 172, 49 },
		{ 9, 173, 49 },
		{ 9, 174, 49 },
		{ 9, 175, 49 },
		{ 9, 176, 49 },
		{ 9, 177, 49 },
		{ 9, 178, 49 },
		{ 9, 179, 49 },
		{ 9, 180, 49 },
		{ 9, 181, 49 },
		{ 9, 182, 49 },
		{ 9, 183, 49 },
		{ 9, 184, 49 },
		{ 9, 185, 49 },
		{ 9, 186, 49 },
		{ 9, 187, 49 },
		{ 9, 188, 49 },
		{ 9, 189, 49 },
		{ 9, 190, 49 },
		{ 9, 191, 49 },
		{ 9, 192, 49 },
		{ 9, 193, 49 },
		{ 9, 194, 49 },
		{ 10, 134, 49 },
		{ 10, 135, 49 },
		{ 10, 136, 49 },
		{ 10, 137, 49 },
		{ 10, 138, 49 },
		{ 10, 139, 49 },
		{ 10, 140, 49 },
		{ 10, 141, 49 },
		{ 10, 142, 49 },
		{ 10, 143, 49 },
		{ 10, 144, 49 },
		{ 10, 145, 49 },
		{ 10, 146, 49 },
		{ 10, 147, 49 },
		{ 10, 148, 49 },
		{ 10, 149, 49 },
		{ 10, 150, 49 },
		{ 10, 151, 49 },
		{ 10, 152, 49 },
		{ 10, 153, 49 },
		{ 10, 154, 49 },
		{ 10, 155, 49 },
		{ 10, 156, 49 },
		{ 10, 157, 49 },
		{ 10, 158, 49 },
		{ 10, 159, 49 },
		{ 10, 160, 49 },
		{ 10, 161, 49 },
		{ 10, 162, 49 },
		{ 10, 163, 49 },
		{ 10, 164, 49 },
		{ 10, 165, 49 },
		{ 10, 166, 49 },
		{ 10, 167, 49 },
		{ 10, 168, 49 },
		{ 10, 169, 49 },
		{ 10, 170, 49 },
		{ 10, 171, 49 },
		{ 10, 172, 49 },
		{ 10, 173, 49 },
		{ 10, 174, 49 },
		{ 10, 175, 49 },
		{ 10, 176, 49 },
		{ 10, 177, 49 },
		{ 10, 178, 49 },
		{ 10, 179, 49 },
		{ 10, 180, 49 },
		{ 10, 181, 49 },
		{ 10, 182, 49 },
		{ 10, 183, 49 },
		{ 10, 184, 49 },
		{ 10, 185, 49 },
		{ 10, 186, 49 },
		{ 10, 187, 49 },
		{ 10, 188, 49 },
		{ 10, 189, 49 },
		{ 10, 190, 49 },
		{ 10, 191, 49 },
		{ 10, 192, 49 },
		{ 10, 193, 49 },
		{ 10, 194, 49 },
		{ 10, 195, 49 },
		{ 11, 135, 49 },
		{ 11, 136, 49 },
		{ 11, 137, 49 },
		{ 11, 138, 49 },
		{ 11, 139, 49 },
		{ 11, 140, 49 },
		{ 11, 141, 49 },
		{ 11, 142, 49 },
		{ 11, 143, 49 },
		{ 11, 144, 49 },
		{ 11, 145, 49 },
		{ 11, 146, 49 },
		{ 11, 147, 49 },
		{ 11, 148, 49 },
		{ 11, 149, 49 },
		{ 11, 150, 49 },
		{ 11, 151, 49 },
		{ 11, 152, 49 },
		{ 11, 153, 49 },
		{ 11, 154, 49 },
		{ 11, 155, 49 },
		{ 11, 156, 49 },
		{ 11, 157, 49 },
		{ 11, 158, 49 },
		{ 11, 159, 49 },
		{ 11, 160, 49 },
		{ 11, 161, 49 },
		{ 11, 162, 49 },
		{ 11, 163, 49 },
		{ 11, 164, 49 },
		{ 11, 165, 49 },
		{ 11, 166, 49 },
		{ 11, 167, 49 },
		{ 11, 168, 49 },
		{ 11, 169, 49 },
		{ 11, 170, 49 },
		{ 11, 171, 49 },
		{ 11, 172, 49 },
		{ 11, 173, 49 },
		{ 11, 174, 49 },
		{ 11, 175, 49 },
		{ 11, 176, 49 },
		{ 11, 177, 49 },
		{ 11, 178, 49 },
		{ 11, 179, 49 },
		{ 11, 180, 49 },
		{ 11, 181, 49 },
		{ 11, 182, 49 },
		{ 11, 183, 49 },
		{ 11, 184, 49 },
		{ 11, 185, 49 },
		{ 11, 186, 49 },
		{ 11, 187, 49 },
		{ 11, 188, 49 },
		{ 11, 189, 49 },
		{ 11, 190, 49 },
		{ 11, 191, 49 },
		{ 11, 192, 49 },
		{ 11, 193, 49 },
		{ 11, 194, 49 },
		{ 11, 195, 49 },
		{ 12, 135, 49 },
		{ 12, 136, 49 },
		{ 12, 137, 49 },
		{ 12, 138, 49 },
		{ 12, 139, 49 },
		{ 12, 140, 49 },
		{ 12, 141, 49 },
		{ 12, 142, 49 },
		{ 12, 143, 49 },
		{ 12, 144, 49 },
		{ 12, 145, 49 },
		{ 12, 146, 49 },
		{ 12, 147, 49 },
		{ 12, 148, 49 },
		{ 12, 149, 49 },
		{ 12, 150, 49 },
		{ 12, 151, 49 },
		{ 12, 152, 49 },
		{ 12, 153, 49 },
		{ 12, 154, 49 },
		{ 12, 155, 49 },
		{ 12, 156, 49 },
		{ 12, 157, 49 },
		{ 12, 158, 49 },
		{ 12, 159, 49 },
		{ 12, 160, 49 },
		{ 12, 161, 49 },
		{ 12, 162, 49 },
		{ 12, 163, 49 },
		{ 12, 164, 49 },
		{ 12, 165, 49 },
		{ 12, 166, 49 },
		{ 12, 167, 49 },
		{ 12, 168, 49 },
		{ 12, 169, 49 },
		{ 12, 170, 49 },
		{ 12, 171, 49 },
		{ 12, 172, 49 },
		{ 12, 173, 49 },
		{ 12, 174, 49 },
		{ 12, 175, 49 },
		{ 12, 176, 49 },
		{ 12, 177, 49 },
		{ 12, 178, 49 },
		{ 12, 179, 49 },
		{ 12, 180, 49 },
		{ 12, 181, 49 },
		{ 12, 182, 49 },
		{ 12, 183, 49 },
		{ 12, 184, 49 },
		{ 12, 185, 49 },
		{ 12, 186, 49 },
		{ 12, 187, 49 },
		{ 12, 188, 49 },
		{ 12, 189, 49 },
		{ 12, 190, 49 },
		{ 12, 191, 49 },
		{ 12, 192, 49 },
		{ 12, 193, 49 },
		{ 12, 194, 49 },
		{ 12, 195, 49 },
		{ 13, 136, 49 },
		{ 13, 137, 49 },
		{ 13, 138, 49 },
		{ 13, 139, 49 },
		{ 13, 140, 49 },
		{ 13, 141, 49 },
		{ 13, 142, 49 },
		{ 13, 143, 49 },
		{ 13, 144, 49 },
		{ 13, 145, 49 },
		{ 13, 146, 49 },
		{ 13, 147, 49 },
		{ 13, 148, 49 },
		{ 13, 149, 49 },
		{ 13, 150, 49 },
		{ 13, 151, 49 },
		{ 13, 152, 49 },
		{ 13, 153, 49 },
		{ 13, 154, 49 },
		{ 13, 155, 49 },
		{ 13, 156, 49 },
		{ 13, 157, 49 },
		{ 13, 158, 49 },
		{ 13, 159, 49 },
		{ 13, 160, 49 },
		{ 13, 161, 49 },
		{ 13, 162, 49 },
		{ 13, 163, 49 },
		{ 13, 164, 49 },
		{ 13, 165, 49 },
		{ 13, 166, 49 },
		{ 13, 167, 49 },
		{ 13, 168, 49 },
		{ 13, 169, 49 },
		{ 13, 170, 49 },
		{ 13, 171, 49 },
		{ 13, 172, 49 },
		{ 13, 173, 49 },
		{ 13, 174, 49 },
		{ 13, 175, 49 },
		{ 13, 176, 49 },
		{ 13, 177, 49 },
		{ 13, 178, 49 },
		{ 13, 179, 49 },
		{ 13, 180, 49 },
		{ 13, 181, 49 },
		{ 13, 182, 49 },
		{ 13, 183, 49 },
		{ 13, 184, 49 },
		{ 13, 185, 49 },
		{ 13, 186, 49 },
		{ 13, 187, 49 },
		{ 13, 188, 49 },
		{ 13, 189, 49 },
		{ 13, 190, 49 },
		{ 13, 191, 49 },
		{ 13, 192, 49 },
		{ 13, 193, 49 },
		{ 13, 194, 49 },
		{ 13, 195, 49 },
		{ 14, 136, 49 },
		{ 14, 137, 49 },
		{ 14, 138, 49 },
		{ 14, 139, 49 },
		{ 14, 140, 49 },
		{ 14, 141, 49 },
		{ 14, 142, 49 },
		{ 14, 143, 49 },
		{ 14, 144, 49 },
		{ 14, 145, 49 },
		{ 14, 146, 49 },
		{ 14, 147, 49 },
		{ 14, 148, 49 },
		{ 14, 149, 49 },
		{ 14, 150, 49 },
		{ 14, 151, 49 },
		{ 14, 152, 49 },
		{ 14, 153, 49 },
		{ 14, 154, 49 },
		{ 14, 155, 49 },
		{ 14, 156, 49 },
		{ 14, 157, 49 },
		{ 14, 158, 49 },
		{ 14, 159, 49 },
		{ 14, 160, 49 },
		{ 14, 161, 49 },
		{ 14, 162, 49 },
		{ 14, 163, 49 },
		{ 14, 164, 49 },
		{ 14, 165, 49 },
		{ 14, 166, 49 },
		{ 14, 167, 49 },
		{ 14, 168, 49 },
		{ 14, 169, 49 },
		{ 14, 170, 49 },
		{ 14, 171, 49 },
		{ 14, 172, 49 },
		{ 14, 173, 49 },
		{ 14, 174, 49 },
		{ 14, 175, 49 },
		{ 14, 176, 49 },
		{ 14, 177, 49 },
		{ 14, 178, 49 },
		{ 14, 179, 49 },
		{ 14, 180, 49 },
		{ 14, 181, 49 },
		{ 14, 182, 49 },
		{ 14, 183, 49 },
		{ 14, 184, 49 },
		{ 14, 185, 49 },
		{ 14, 186, 49 },
		{ 14, 187, 49 },
		{ 14, 188, 49 },
		{ 14, 189, 49 },
		{ 14, 190, 49 },
		{ 14, 191, 49 },
		{ 14, 192, 49 },
		{ 14, 193, 49 },
		{ 14, 194, 49 },
		{ 14, 195, 49 },
		{ 14, 196, 49 },
		{ 15, 137, 49 },
		{ 15, 138, 49 },
		{ 15, 139, 49 },
		{ 15, 140, 49 },
		{ 15, 141, 49 },
		{ 15, 142, 49 },
		{ 15, 143, 49 },
		{ 15, 144, 49 },
		{ 15, 145, 49 },
		{ 15, 146, 49 },
		{ 15, 147, 49 },
		{ 15, 148, 49 },
		{ 15, 149, 49 },
		{ 15, 150, 49 },
		{ 15, 151, 49 },
		{ 15, 152, 49 },
		{ 15, 153, 49 },
		{ 15, 154, 49 },
		{ 15, 155, 49 },
		{ 15, 156, 49 },
		{ 15, 157, 49 },
		{ 15, 158, 49 },
		{ 15, 159, 49 },
		{ 15, 160, 49 },
		{ 15, 161, 49 },
		{ 15, 162, 49 },
		{ 15, 163, 49 },
		{ 15, 164, 49 },
		{ 15, 165, 49 },
		{ 15, 166, 49 },
		{ 15, 167, 49 },
		{ 15, 168, 49 },
		{ 15, 169, 49 },
		{ 15, 170, 49 },
		{ 15, 171, 49 },
		{ 15, 172, 49 },
		{ 15, 173, 49 },
		{ 15, 174, 49 },
		{ 15, 175, 49 },
		{ 15, 176, 49 },
		{ 15, 177, 49 },
		{ 15, 178, 49 },
		{ 15, 179, 49 },
		{ 15, 180, 49 },
		{ 15, 181, 49 },
		{ 15, 182, 49 },
		{ 15, 183, 49 },
		{ 15, 184, 49 },
		{ 15, 185, 49 },
		{ 15, 186, 49 },
		{ 15, 187, 49 },
		{ 15, 188, 49 },
		{ 15, 189, 49 },
		{ 15, 190, 49 },
		{ 15, 191, 49 },
		{ 15, 192, 49 },
		{ 15, 193, 49 },
		{ 15, 194, 49 },
		{ 15, 195, 49 },
		{ 15, 196, 49 },
		{ 16, 137, 49 },
		{ 16, 138, 49 },
		{ 16, 139, 49 },
		{ 16, 140, 49 },
		{ 16, 141, 49 },
		{ 16, 142, 49 },
		{ 16, 143, 49 },
		{ 16, 144, 49 },
		{ 16, 145, 49 },
		{ 16, 146, 49 },
		{ 16, 147, 49 },
		{ 16, 148, 49 },
		{ 16, 149, 49 },
		{ 16, 150, 49 },
		{ 16, 151, 49 },
		{ 16, 152, 49 },
		{ 16, 153, 49 },
		{ 16, 154, 49 },
		{ 16, 155, 49 },
		{ 16, 156, 49 },
		{ 16, 157, 49 },
		{ 16, 158, 49 },
		{ 16, 159, 49 },
		{ 16, 160, 49 },
		{ 16, 161, 49 },
		{ 16, 162, 49 },
		{ 16, 163, 49 },
		{ 16, 164, 49 },
		{ 16, 165, 49 },
		{ 16, 166, 49 },
		{ 16, 167, 49 },
		{ 16, 168, 49 },
		{ 16, 169, 49 },
		{ 16, 170, 49 },
		{ 16, 171, 49 },
		{ 16, 172, 49 },
		{ 16, 173, 49 },
		{ 16, 174, 49 },
		{ 16, 175, 49 },
		{ 16, 176, 49 },
		{ 16, 177, 49 },
		{ 16, 178, 49 },
		{ 16, 179, 49 },
		{ 16, 180, 49 },
		{ 16, 181, 49 },
		{ 16, 182, 49 },
		{ 16, 183, 49 },
		{ 16, 184, 49 },
		{ 16, 185, 49 },
		{ 16, 186, 49 },
		{ 16, 187, 49 },
		{ 16, 188, 49 },
		{ 16, 189, 49 },
		{ 16, 190, 49 },
		{ 16, 191, 49 },
		{ 16, 192, 49 },
		{ 16, 193, 49 },
		{ 16, 194, 49 },
		{ 16, 195, 49 },
		{ 16, 196, 49 },
		{ 17, 138, 49 },
		{ 17, 139, 49 },
		{ 17, 140, 49 },
		{ 17, 141, 49 },
		{ 17, 142, 49 },
		{ 17, 143, 49 },
		{ 17, 144, 49 },
		{ 17, 145, 49 },
		{ 17, 146, 49 },
		{ 17, 147, 49 },
		{ 17, 148, 49 },
		{ 17, 149, 49 },
		{ 17, 150, 49 },
		{ 17, 151, 49 },
		{ 17, 152, 49 },
		{ 17, 153, 49 },
		{ 17, 154, 49 },
		{ 17, 155, 49 },
		{ 17, 156, 49 },
		{ 17, 157, 49 },
		{ 17, 158, 49 },
		{ 17, 159, 49 },
		{ 17, 160, 49 },
		{ 17, 161, 49 },
		{ 17, 162, 49 },
		{ 17, 163, 49 },
		{ 17, 164, 49 },
		{ 17, 165, 49 },
		{ 17, 166, 49 },
		{ 17, 167, 49 },
		{ 17, 168, 49 },
		{ 17, 169, 49 },
		{ 17, 170, 49 },
		{ 17, 171, 49 },
		{ 17, 172, 49 },
		{ 17, 173, 49 },
		{ 17, 174, 49 },
		{ 17, 175, 49 },
		{ 17, 176, 49 },
		{ 17, 177, 49 },
		{ 17, 178, 49 },
		{ 17, 179, 49 },
		{ 17, 180, 49 },
		{ 17, 181, 49 },
		{ 17, 182, 49 },
		{ 17, 183, 49 },
		{ 17, 184, 49 },
		{ 17, 185, 49 },
		{ 17, 186, 49 },
		{ 17, 187, 49 },
		{ 17, 188, 49 },
		{ 17, 189, 49 },
		{ 17, 190, 49 },
		{ 17, 191, 49 },
		{ 17, 192, 49 },
		{ 17, 193, 49 },
		{ 17, 194, 49 },
		{ 17, 195, 49 },
		{ 17, 196, 49 },
		{ 18, 138, 49 },
		{ 18, 139, 49 },
		{ 18, 140, 49 },
		{ 18, 141, 49 },
		{ 18, 142, 49 },
		{ 18, 143, 49 },
		{ 18, 144, 49 },
		{ 18, 145, 49 },
		{ 18, 146, 49 },
		{ 18, 147, 49 },
		{ 18, 148, 49 },
		{ 18, 149, 49 },
		{ 18, 150, 49 },
		{ 18, 151, 49 },
		{ 18, 152, 49 },
		{ 18, 153, 49 },
		{ 18, 154, 49 },
		{ 18, 155, 49 },
		{ 18, 156, 49 },
		{ 18, 157, 49 },
		{ 18, 158, 49 },
		{ 18, 159, 49 },
		{ 18, 160, 49 },
		{ 18, 161, 49 },
		{ 18, 162, 49 },
		{ 18, 163, 49 },
		{ 18, 164, 49 },
		{ 18, 165, 49 },
		{ 18, 166, 49 },
		{ 18, 167, 49 },
		{ 18, 168, 49 },
		{ 18, 169, 49 },
		{ 18, 170, 49 },
		{ 18, 171, 49 },
		{ 18, 172, 49 },
		{ 18, 173, 49 },
		{ 18, 174, 49 },
		{ 18, 175, 49 },
		{ 18, 176, 49 },
		{ 18, 177, 49 },
		{ 18, 178, 49 },
		{ 18, 179, 49 },
		{ 18, 180, 49 },
		{ 18, 181, 49 },
		{ 18, 182, 49 },
		{ 18, 183, 49 },
		{ 18, 184, 49 },
		{ 18, 185, 49 },
		{ 18, 186, 49 },
		{ 18, 187, 49 },
		{ 18, 188, 49 },
		{ 18, 189, 49 },
		{ 18, 190, 49 },
		{ 18, 191, 49 },
		{ 18, 192, 49 },
		{ 18, 193, 49 },
		{ 18, 194, 49 },
		{ 18, 195, 49 },
		{ 18, 196, 49 },
		{ 18, 197, 49 },
		{ 19, 139, 49 },
		{ 19, 140, 49 },
		{ 19, 141, 49 },
		{ 19, 142, 49 },
		{ 19, 143, 49 },
		{ 19, 144, 49 },
		{ 19, 145, 49 },
		{ 19, 146, 49 },
		{ 19, 147, 49 },
		{ 19, 148, 49 },
		{ 19, 149, 49 },
		{ 19, 150, 49 },
		{ 19, 151, 49 },
		{ 19, 152, 49 },
		{ 19, 153, 49 },
		{ 19, 154, 49 },
		{ 19, 155, 49 },
		{ 19, 156, 49 },
		{ 19, 157, 49 },
		{ 19, 158, 49 },
		{ 19, 159, 49 },
		{ 19, 160, 49 },
		{ 19, 161, 49 },
		{ 19, 162, 49 },
		{ 19, 163, 49 },
		{ 19, 164, 49 },
		{ 19, 165, 49 },
		{ 19, 166, 49 },
		{ 19, 167, 49 },
		{ 19, 168, 49 },
		{ 19, 169, 49 },
		{ 19, 170, 49 },
		{ 19, 171, 49 },
		{ 19, 172, 49 },
		{ 19, 173, 49 },
		{ 19, 174, 49 },
		{ 19, 175, 49 },
		{ 19, 176, 49 },
		{ 19, 177, 49 },
		{ 19, 178, 49 },
		{ 19, 179, 49 },
		{ 19, 180, 49 },
		{ 19, 181, 49 },
		{ 19, 182, 49 },
		{ 19, 183, 49 },
		{ 19, 184, 49 },
		{ 19, 185, 49 },
		{ 19, 186, 49 },
		{ 19, 187, 49 },
		{ 19, 188, 49 },
		{ 19, 189, 49 },
		{ 19, 190, 49 },
		{ 19, 191, 49 },
		{ 19, 192, 49 },
		{ 19, 193, 49 },
		{ 19, 194, 49 },
		{ 19, 195, 49 },
		{ 19, 196, 49 },
		{ 19, 197, 49 },
		{ 20, 139, 49 },
		{ 20, 140, 49 },
		{ 20, 141, 49 },
		{ 20, 142, 49 },
		{ 20, 143, 49 },
		{ 20, 144, 49 },
		{ 20, 145, 49 },
		{ 20, 146, 49 },
		{ 20, 147, 49 },
		{ 20, 148, 49 },
		{ 20, 149, 49 },
		{ 20, 150, 49 },
		{ 20, 151, 49 },
		{ 20, 152, 49 },
		{ 20, 153, 49 },
		{ 20, 154, 49 },
		{ 20, 155, 49 },
		{ 20, 156, 49 },
		{ 20, 157, 49 },
		{ 20, 158, 49 },
		{ 20, 159, 49 },
		{ 20, 160, 49 },
		{ 20, 161, 49 },
		{ 20, 162, 49 },
		{ 20, 163, 49 },
		{ 20, 164, 49 },
		{ 20, 165, 49 },
		{ 20, 166, 49 },
		{ 20, 167, 49 },
		{ 20, 168, 49 },
		{ 20, 169, 49 },
		{ 20, 170, 49 },
		{ 20, 171, 49 },
		{ 20, 172, 49 },
		{ 20, 173, 49 },
		{ 20, 174, 49 },
		{ 20, 175, 49 },
		{ 20, 176, 49 },
		{ 20, 177, 49 },
		{ 20, 178, 49 },
		{ 20, 179, 49 },
		{ 20, 180, 49 },
		{ 20, 181, 49 },
		{ 20, 182, 49 },
		{ 20, 183, 49 },
		{ 20, 184, 49 },
		{ 20, 185, 49 },
		{ 20, 186, 49 },
		{ 20, 187, 49 },
		{ 20, 188, 49 },
		{ 20, 189, 49 },
		{ 20, 190, 49 },
		{ 20, 191, 49 },
		{ 20, 192, 49 },
		{ 20, 193, 49 },
		{ 20, 194, 49 },
		{ 20, 195, 49 },
		{ 20, 196, 49 },
		{ 20, 197, 49 },
		{ 21, 140, 49 },
		{ 21, 141, 49 },
		{ 21, 142, 49 },
		{ 21, 143, 49 },
		{ 21, 144, 49 },
		{ 21, 145, 49 },
		{ 21, 146, 49 },
		{ 21, 147, 49 },
		{ 21, 148, 49 },
		{ 21, 149, 49 },
		{ 21, 150, 49 },
		{ 21, 151, 49 },
		{ 21, 152, 49 },
		{ 21, 153, 49 },
		{ 21, 154, 49 },
		{ 21, 155, 49 },
		{ 21, 156, 49 },
		{ 21, 157, 49 },
		{ 21, 158, 49 },
		{ 21, 159, 49 },
		{ 21, 160, 49 },
		{ 21, 161, 49 },
		{ 21, 162, 49 },
		{ 21, 163, 49 },
		{ 21, 164, 49 },
		{ 21, 165, 49 },
		{ 21, 166, 49 },
		{ 21, 167, 49 },
		{ 21, 168, 49 },
		{ 21, 169, 49 },
		{ 21, 170, 49 },
		{ 21, 171, 49 },
		{ 21, 172, 49 },
		{ 21, 173, 49 },
		{ 21, 174, 49 },
		{ 21, 175, 49 },
		{ 21, 176, 49 },
		{ 21, 177, 49 },
		{ 21, 178, 49 },
		{ 21, 179, 49 },
		{ 21, 180, 49 },
		{ 21, 181, 49 },
		{ 21, 182, 49 },
		{ 21, 183, 49 },
		{ 21, 184, 49 },
		{ 21, 185, 49 },
		{ 21, 186, 49 },
		{ 21, 187, 49 },
		{ 21, 188, 49 },
		{ 21, 189, 49 },
		{ 21, 190, 49 },
		{ 21, 191, 49 },
		{ 21, 192, 49 },
		{ 21, 193, 49 },
		{ 21, 194, 49 },
		{ 21, 195, 49 },
		{ 21, 196, 49 },
		{ 21, 197, 49 },
		{ 22, 140, 49 },
		{ 22, 141, 49 },
		{ 22, 142, 49 },
		{ 22, 143, 49 },
		{ 22, 144, 49 },
		{ 22, 145, 49 },
		{ 22, 146, 49 },
		{ 22, 147, 49 },
		{ 22, 148, 49 },
		{ 22, 149, 49 },
		{ 22, 150, 49 },
		{ 22, 151, 49 },
		{ 22, 152, 49 },
		{ 22, 153, 49 },
		{ 22, 154, 49 },
		{ 22, 155, 49 },
		{ 22, 156, 49 },
		{ 22, 157, 49 },
		{ 22, 158, 49 },
		{ 22, 159, 49 },
		{ 22, 160, 49 },
		{ 22, 161, 49 },
		{ 22, 162, 49 },
		{ 22, 163, 49 },
		{ 22, 164, 49 },
		{ 22, 165, 49 },
		{ 22, 166, 49 },
		{ 22, 167, 49 },
		{ 22, 168, 49 },
		{ 22, 169, 49 },
		{ 22, 170, 49 },
		{ 22, 171, 49 },
		{ 22, 172, 49 },
		{ 22, 173, 49 },
		{ 22, 174, 49 },
		{ 22, 175, 49 },
		{ 22, 176, 49 },
		{ 22, 177, 49 },
		{ 22, 178, 49 },
		{ 22, 179, 49 },
		{ 22, 180, 49 },
		{ 22, 181, 49 },
		{ 22, 182, 49 },
		{ 22, 183, 49 },
		{ 22, 184, 49 },
		{ 22, 185, 49 },
		{ 22, 186, 49 },
		{ 22, 187, 49 },
		{ 22, 188, 49 },
		{ 22, 189, 49 },
		{ 22, 190, 49 },
		{ 22, 191, 49 },
		{ 22, 192, 49 },
		{ 22, 193, 49 },
		{ 22, 194, 49 },
		{ 22, 195, 49 },
		{ 22, 196, 49 },
		{ 22, 197, 49 },
		{ 22, 198, 49 },
		{ 23, 141, 49 },
		{ 23, 142, 49 },
		{ 23, 143, 49 },
		{ 23, 144, 49 },
		{ 23, 145, 49 },
		{ 23, 146, 49 },
		{ 23, 147, 49 },
		{ 23, 148, 49 },
		{ 23, 149, 49 },
		{ 23, 150, 49 },
		{ 23, 151, 49 },
		{ 23, 152, 49 },
		{ 23, 153, 49 },
		{ 23, 154, 49 },
		{ 23, 155, 49 },
		{ 23, 156, 49 },
		{ 23, 157, 49 },
		{ 23, 158, 49 },
		{ 23, 159, 49 },
		{ 23, 160, 49 },
		{ 23, 161, 49 },
		{ 23, 162, 49 },
		{ 23, 163, 49 },
		{ 23, 164, 49 },
		{ 23, 165, 49 },
		{ 23, 166, 49 },
		{ 23, 167, 49 },
		{ 23, 168, 49 },
		{ 23, 169, 49 },
		{ 23, 170, 49 },
		{ 23, 171, 49 },
		{ 23, 172, 49 },
		{ 23, 173, 49 },
		{ 23, 174, 49 },
		{ 23, 175, 49 },
		{ 23, 176, 49 },
		{ 23, 177, 49 },
		{ 23, 178, 49 },
		{ 23, 179, 49 },
		{ 23, 180, 49 },
		{ 23, 181, 49 },
		{ 23, 182, 49 },
		{ 23, 183, 49 },
		{ 23, 184, 49 },
		{ 23, 185, 49 },
		{ 23, 186, 49 },
		{ 23, 187, 49 },
		{ 23, 188, 49 },
		{ 23, 189, 49 },
		{ 23, 190, 49 },
		{ 23, 191, 49 },
		{ 23, 192, 49 },
		{ 23, 193, 49 },
		{ 23, 194, 49 },
		{ 23, 195, 49 },
		{ 23, 196, 49 },
		{ 23, 197, 49 },
		{ 23, 198, 49 },
		{ 24, 141, 49 },
		{ 24, 142, 49 },
		{ 24, 143, 49 },
		{ 24, 144, 49 },
		{ 24, 145, 49 },
		{ 24, 146, 49 },
		{ 24, 147, 49 },
		{ 24, 148, 49 },
		{ 24, 149, 49 },
		{ 24, 150, 49 },
		{ 24, 151, 49 },
		{ 24, 152, 49 },
		{ 24, 153, 49 },
		{ 24, 154, 49 },
		{ 24, 155, 49 },
		{ 24, 156, 49 },
		{ 24, 157, 49 },
		{ 24, 158, 49 },
		{ 24, 159, 49 },
		{ 24, 160, 49 },
		{ 24, 161, 49 },
		{ 24, 162, 49 },
		{ 24, 163, 49 },
		{ 24, 164, 49 },
		{ 24, 165, 49 },
		{ 24, 166, 49 },
		{ 24, 167, 49 },
		{ 24, 168, 49 },
		{ 24, 169, 49 },
		{ 24, 170, 49 },
		{ 24, 171, 49 },
		{ 24, 172, 49 },
		{ 24, 173, 49 },
		{ 24, 174, 49 },
		{ 24, 175, 49 },
		{ 24, 176, 49 },
		{ 24, 177, 49 },
		{ 24, 178, 49 },
		{ 24, 179, 49 },
		{ 24, 180, 49 },
		{ 24, 181, 49 },
		{ 24, 182, 49 },
		{ 24, 183, 49 },
		{ 24, 184, 49 },
		{ 24, 185, 49 },
		{ 24, 186, 49 },
		{ 24, 187, 49 },
		{ 24, 188, 49 },
		{ 24, 189, 49 },
		{ 24, 190, 49 },
		{ 24, 191, 49 },
		{ 24, 192, 49 },
		{ 24, 193, 49 },
		{ 24, 194, 49 },
		{ 24, 195, 49 },
		{ 24, 196, 49 },
		{ 24, 197, 49 },
		{ 24, 198, 49 },
		{ 25, 142, 49 },
		{ 25, 143, 49 },
		{ 25, 144, 49 },
		{ 25, 145, 49 },
		{ 25, 146, 49 },
		{ 25, 147, 49 },
		{ 25, 148, 49 },
		{ 25, 149, 49 },
		{ 25, 150, 49 },
		{ 25, 151, 49 },
		{ 25, 152, 49 },
		{ 25, 153, 49 },
		{ 25, 154, 49 },
		{ 25, 155, 49 },
		{ 25, 156, 49 },
		{ 25, 157, 49 },
		{ 25, 158, 49 },
		{ 25, 159, 49 },
		{ 25, 160, 49 },
		{ 25, 161, 49 },
		{ 25, 162, 49 },
		{ 25, 163, 49 },
		{ 25, 164, 49 },
		{ 25, 165, 49 },
		{ 25, 166, 49 },
		{ 25, 167, 49 },
		{ 25, 168, 49 },
		{ 25, 169, 49 },
		{ 25, 170, 49 },
		{ 25, 171, 49 },
		{ 25, 172, 49 },
		{ 25, 173, 49 },
		{ 25, 174, 49 },
		{ 25, 175, 49 },
		{ 25, 176, 49 },
		{ 25, 177, 49 },
		{ 25, 178, 49 },
		{ 25, 179, 49 },
		{ 25, 180, 49 },
		{ 25, 181, 49 },
		{ 25, 182, 49 },
		{ 25, 183, 49 },
		{ 25, 184, 49 },
		{ 25, 185, 49 },
		{ 25, 186, 49 },
		{ 25, 187, 49 },
		{ 25, 188, 49 },
		{ 25, 189, 49 },
		{ 25, 190, 49 },
		{ 25, 191, 49 },
		{ 25, 192, 49 },
		{ 25, 193, 49 },
		{ 25, 194, 49 },
		{ 25, 195, 49 },
		{ 25, 196, 49 },
		{ 25, 197, 49 },
		{ 25, 198, 49 },
		{ 26, 142, 49 },
		{ 26, 143, 49 },
		{ 26, 144, 49 },
		{ 26, 145, 49 },
		{ 26, 146, 49 },
		{ 26, 147, 49 },
		{ 26, 148, 49 },
		{ 26, 149, 49 },
		{ 26, 150, 49 },
		{ 26, 151, 49 },
		{ 26, 152, 49 },
		{ 26, 153, 49 },
		{ 26, 154, 49 },
		{ 26, 155, 49 },
		{ 26, 156, 49 },
		{ 26, 157, 49 },
		{ 26, 158, 49 },
		{ 26, 159, 49 },
		{ 26, 160, 49 },
		{ 26, 161, 49 },
		{ 26, 162, 49 },
		{ 26, 163, 49 },
		{ 26, 164, 49 },
		{ 26, 165, 49 },
		{ 26, 166, 49 },
		{ 26, 167, 49 },
		{ 26, 168, 49 },
		{ 26, 169, 49 },
		{ 26, 170, 49 },
		{ 26, 171, 49 },
		{ 26, 172, 49 },
		{ 26, 173, 49 },
		{ 26, 174, 49 },
		{ 26, 175, 49 },
		{ 26, 176, 49 },
		{ 26, 177, 49 },
		{ 26, 178, 49 },
		{ 26, 179, 49 },
		{ 26, 180, 49 },
		{ 26, 181, 49 },
		{ 26, 182, 49 },
		{ 26, 183, 49 },
		{ 26, 184, 49 },
		{ 26, 185, 49 },
		{ 26, 186, 49 },
		{ 26, 187, 49 },
		{ 26, 188, 49 },
		{ 26, 189, 49 },
		{ 26, 190, 49 },
		{ 26, 191, 49 },
		{ 26, 192, 49 },
		{ 26, 193, 49 },
		{ 26, 194, 49 },
		{ 26, 195, 49 },
		{ 26, 196, 49 },
		{ 26, 197, 49 },
		{ 26, 198, 49 },
		{ 26, 199, 49 },
		{ 27, 143, 49 },
		{ 27, 144, 49 },
		{ 27, 145, 49 },
		{ 27, 146, 49 },
		{ 27, 147, 49 },
		{ 27, 148, 49 },
		{ 27, 149, 49 },
		{ 27, 150, 49 },
		{ 27, 151, 49 },
		{ 27, 152, 49 },
		{ 27, 153, 49 },
		{ 27, 154, 49 },
		{ 27, 155, 49 },
		{ 27, 156, 49 },
		{ 27, 157, 49 },
		{ 27, 158, 49 },
		{ 27, 159, 49 },
		{ 27, 160, 49 },
		{ 27, 161, 49 },
		{ 27, 162, 49 },
		{ 27, 163, 49 },
		{ 27, 164, 49 },
		{ 27, 165, 49 },
		{ 27, 166, 49 },
		{ 27, 167, 49 },
		{ 27, 168, 49 },
		{ 27, 169, 49 },
		{ 27, 170, 49 },
		{ 27, 171, 49 },
		{ 27, 172, 49 },
		{ 27, 173, 49 },
		{ 27, 174, 49 },
		{ 27, 175, 49 },
		{ 27, 176, 49 },
		{ 27, 177, 49 },
		{ 27, 178, 49 },
		{ 27, 179, 49 },
		{ 27, 180, 49 },
		{ 27, 181, 49 },
		{ 27, 182, 49 },
		{ 27, 183, 49 },
		{ 27, 184, 49 },
		{ 27, 185, 49 },
		{ 27, 186, 49 },
		{ 27, 187, 49 },
		{ 27, 188, 49 },
		{ 27, 189, 49 },
		{ 27, 190, 49 },
		{ 27, 191, 49 },
		{ 27, 192, 49 },
		{ 27, 193, 49 },
		{ 27, 194, 49 },
		{ 27, 195, 49 },
		{ 27, 196, 49 },
		{ 27, 197, 49 },
		{ 27, 198, 49 },
		{ 27, 199, 49 },
		{ 28, 143, 49 },
		{ 28, 144, 49 },
		{ 28, 145, 49 },
		{ 28, 146, 49 },
		{ 28, 147, 49 },
		{ 28, 148, 49 },
		{ 28, 149, 49 },
		{ 28, 150, 49 },
		{ 28, 151, 49 },
		{ 28, 152, 49 },
		{ 28, 153, 49 },
		{ 28, 154, 49 },
		{ 28, 155, 49 },
		{ 28, 156, 49 },
		{ 28, 157, 49 },
		{ 28, 158, 49 },
		{ 28, 159, 49 },
		{ 28, 160, 49 },
		{ 28, 161, 49 },
		{ 28, 162, 49 },
		{ 28, 163, 49 },
		{ 28, 164, 49 },
		{ 28, 165, 49 },
		{ 28, 166, 49 },
		{ 28, 167, 49 },
		{ 28, 168, 49 },
		{ 28, 169, 49 },
		{ 28, 170, 49 },
		{ 28, 171, 49 },
		{ 28, 172, 49 },
		{ 28, 173, 49 },
		{ 28, 174, 49 },
		{ 28, 175, 49 },
		{ 28, 176, 49 },
		{ 28, 177, 49 },
		{ 28, 178, 49 },
		{ 28, 179, 49 },
		{ 28, 180, 49 },
		{ 28, 181, 49 },
		{ 28, 182, 49 },
		{ 28, 183, 49 },
		{ 28, 184, 49 },
		{ 28, 185, 49 },
		{ 28, 186, 49 },
		{ 28, 187, 49 },
		{ 28, 188, 49 },
		{ 28, 189, 49 },
		{ 28, 190, 49 },
		{ 28, 191, 49 },
		{ 28, 192, 49 },
		{ 28, 193, 49 },
		{ 28, 194, 49 },
		{ 28, 195, 49 },
		{ 28, 196, 49 },
		{ 28, 197, 49 },
		{ 28, 198, 49 },
		{ 28, 199, 49 },
		{ 29, 144, 49 },
		{ 29, 145, 49 },
		{ 29, 146, 49 },
		{ 29, 147, 49 },
		{ 29, 148, 49 },
		{ 29, 149, 49 },
		{ 29, 150, 49 },
		{ 29, 151, 49 },
		{ 29, 152, 49 },
		{ 29, 153, 49 },
		{ 29, 154, 49 },
		{ 29, 155, 49 },
		{ 29, 156, 49 },
		{ 29, 157, 49 },
		{ 29, 158, 49 },
		{ 29, 159, 49 },
		{ 29, 160, 49 },
		{ 29, 161, 49 },
		{ 29, 162, 49 },
		{ 29, 163, 49 },
		{ 29, 164, 49 },
		{ 29, 165, 49 },
		{ 29, 166, 49 },
		{ 29, 167, 49 },
		{ 29, 168, 49 },
		{ 29, 169, 49 },
		{ 29, 170, 49 },
		{ 29, 171, 49 },
		{ 29, 172, 49 },
		{ 29, 173, 49 },
		{ 29, 174, 49 },
		{ 29, 175, 49 },
		{ 29, 176, 49 },
		{ 29, 177, 49 },
		{ 29, 178, 49 },
		{ 29, 179, 49 },
		{ 29, 180, 49 },
		{ 29, 181, 49 },
		{ 29, 182, 49 },
		{ 29, 183, 49 },
		{ 29, 184, 49 },
		{ 29, 185, 49 },
		{ 29, 186, 49 },
		{ 29, 187, 49 },
		{ 29, 188, 49 },
		{ 29, 189, 49 },
		{ 29, 190, 49 },
		{ 29, 191, 49 },
		{ 29, 192, 49 },
		{ 29, 193, 49 },
		{ 29, 194, 49 },
		{ 29, 195, 49 },
		{ 29, 196, 49 },
		{ 29, 197, 49 },
		{ 29, 198, 49 },
		{ 29, 199, 49 },
		{ 30, 144, 49 },
		{ 30, 145, 49 },
		{ 30, 146, 49 },
		{ 30, 147, 49 },
		{ 30, 148, 49 },
		{ 30, 149, 49 },
		{ 30, 150, 49 },
		{ 30, 151, 49 },
		{ 30, 152, 49 },
		{ 30, 153, 49 },
		{ 30, 154, 49 },
		{ 30, 155, 49 },
		{ 30, 156, 49 },
		{ 30, 157, 49 },
		{ 30, 158, 49 },
		{ 30, 159, 49 },
		{ 30, 160, 49 },
		{ 30, 161, 49 },
		{ 30, 162, 49 },
		{ 30, 163, 49 },
		{ 30, 164, 49 },
		{ 30, 165, 49 },
		{ 30, 166, 49 },
		{ 30, 167, 49 },
		{ 30, 168, 49 },
		{ 30, 169, 49 },
		{ 30, 170, 49 },
		{ 30, 171, 49 },
		{ 30, 172, 49 },
		{ 30, 173, 49 },
		{ 30, 174, 49 },
		{ 30, 175, 49 },
		{ 30, 176, 49 },
		{ 30, 177, 49 },
		{ 30, 178, 49 },
		{ 30, 179, 49 },
		{ 30, 180, 49 },
		{ 30, 181, 49 },
		{ 30, 182, 49 },
		{ 30, 183, 49 },
		{ 30, 184, 49 },
		{ 30, 185, 49 },
		{ 30, 186, 49 },
		{ 30, 187, 49 },
		{ 30, 188, 49 },
		{ 30, 189, 49 },
		{ 30, 190, 49 },
		{ 30, 191, 49 },
		{ 30, 192, 49 },
		{ 30, 193, 49 },
		{ 30, 194, 49 },
		{ 30, 195, 49 },
		{ 30, 196, 49 },
		{ 30, 197, 49 },
		{ 30, 198, 49 },
		{ 30, 199, 49 },
		{ 30, 200, 49 },
		{ 31, 145, 49 },
		{ 31, 146, 49 },
		{ 31, 147, 49 },
		{ 31, 148, 49 },
		{ 31, 149, 49 },
		{ 31, 150, 49 },
		{ 31, 151, 49 },
		{ 31, 152, 49 },
		{ 31, 153, 49 },
		{ 31, 154, 49 },
		{ 31, 155, 49 },
		{ 31, 156, 49 },
		{ 31, 157, 49 },
		{ 31, 158, 49 },
		{ 31, 159, 49 },
		{ 31, 160, 49 },
		{ 31, 161, 49 },
		{ 31, 162, 49 },
		{ 31, 163, 49 },
		{ 31, 164, 49 },
		{ 31, 165, 49 },
		{ 31, 166, 49 },
		{ 31, 167, 49 },
		{ 31, 168, 49 },
		{ 31, 169, 49 },
		{ 31, 170, 49 },
		{ 31, 171, 49 },
		{ 31, 172, 49 },
		{ 31, 173, 49 },
		{ 31, 174, 49 },
		{ 31, 175, 49 },
		{ 31, 176, 49 },
		{ 31, 177, 49 },
		{ 31, 178, 49 },
		{ 31, 179, 49 },
		{ 31, 180, 49 },
		{ 31, 181, 49 },
		{ 31, 182, 49 },
		{ 31, 183, 49 },
		{ 31, 184, 49 },
		{ 31, 185, 49 },
		{ 31, 186, 49 },
		{ 31, 187, 49 },
		{ 31, 188, 49 },
		{ 31, 189, 49 },
		{ 31, 190, 49 },
		{ 31, 191, 49 },
		{ 31, 192, 49 },
		{ 31, 193, 49 },
		{ 31, 194, 49 },
		{ 31, 195, 49 },
		{ 31, 196, 49 },
		{ 31, 197, 49 },
		{ 31, 198, 49 },
		{ 31, 199, 49 },
		{ 31, 200, 49 },
		{ 32, 145, 49 },
		{ 32, 146, 49 },
		{ 32, 147, 49 },
		{ 32, 148, 49 },
		{ 32, 149, 49 },
		{ 32, 150, 49 },
		{ 32, 151, 49 },
		{ 32, 152, 49 },
		{ 32, 153, 49 },
		{ 32, 154, 49 },
		{ 32, 155, 49 },
		{ 32, 156, 49 },
		{ 32, 157, 49 },
		{ 32, 158, 49 },
		{ 32, 159, 49 },
		{ 32, 160, 49 },
		{ 32, 161, 49 },
		{ 32, 162, 49 },
		{ 32, 163, 49 },
		{ 32, 164, 49 },
		{ 32, 165, 49 },
		{ 32, 166, 49 },
		{ 32, 167, 49 },
		{ 32, 168, 49 },
		{ 32, 169, 49 },
		{ 32, 170, 49 },
		{ 32, 171, 49 },
		{ 32, 172, 49 },
		{ 32, 173, 49 },
		{ 32, 174, 49 },
		{ 32, 175, 49 },
		{ 32, 176, 49 },
		{ 32, 177, 49 },
		{ 32, 178, 49 },
		{ 32, 179, 49 },
		{ 32, 180, 49 },
		{ 32, 181, 49 },
		{ 32, 182, 49 },
		{ 32, 183, 49 },
		{ 32, 184, 49 },
		{ 32, 185, 49 },
		{ 32, 186, 49 },
		{ 32, 187, 49 },
		{ 32, 188, 49 },
		{ 32, 189, 49 },
		{ 32, 190, 49 },
		{ 32, 191, 49 },
		{ 32, 192, 49 },
		{ 32, 193, 49 },
		{ 32, 194, 49 },
		{ 32, 195, 49 },
		{ 32, 196, 49 },
		{ 32, 197, 49 },
		{ 32, 198, 49 },
		{ 32, 199, 49 },
		{ 32, 200, 49 },
		{ 33, 146, 49 },
		{ 33, 147, 49 },
		{ 33, 148, 49 },
		{ 33, 149, 49 },
		{ 33, 150, 49 },
		{ 33, 151, 49 },
		{ 33, 152, 49 },
		{ 33, 153, 49 },
		{ 33, 154, 49 },
		{ 33, 155, 49 },
		{ 33, 156, 49 },
		{ 33, 157, 49 },
		{ 33, 158, 49 },
		{ 33, 159, 49 },
		{ 33, 160, 49 },
		{ 33, 161, 49 },
		{ 33, 162, 49 },
		{ 33, 163, 49 },
		{ 33, 164, 49 },
		{ 33, 165, 49 },
		{ 33, 166, 49 },
		{ 33, 167, 49 },
		{ 33, 168, 49 },
		{ 33, 169, 49 },
		{ 33, 170, 49 },
		{ 33, 171, 49 },
		{ 33, 172, 49 },
		{ 33, 173, 49 },
		{ 33, 174, 49 },
		{ 33, 175, 49 },
		{ 33, 176, 49 },
		{ 33, 177, 49 },
		{ 33, 178, 49 },
		{ 33, 179, 49 },
		{ 33, 180, 49 },
		{ 33, 181, 49 },
		{ 33, 182, 49 },
		{ 33, 183, 49 },
		{ 33, 184, 49 },
		{ 33, 185, 49 },
		{ 33, 186, 49 },
		{ 33, 187, 49 },
		{ 33, 188, 49 },
		{ 33, 189, 49 },
		{ 33, 190, 49 },
		{ 33, 191, 49 },
		{ 33, 192, 49 },
		{ 33, 193, 49 },
		{ 33, 194, 49 },
		{ 33, 195, 49 },
		{ 33, 196, 49 },
		{ 33, 197, 49 },
		{ 33, 198, 49 },
		{ 33, 199, 49 },
		{ 33, 200, 49 },
		{ 34, 146, 49 },
		{ 34, 147, 49 },
		{ 34, 148, 49 },
		{ 34, 149, 49 },
		{ 34, 150, 49 },
		{ 34, 151, 49 },
		{ 34, 152, 49 },
		{ 34, 153, 49 },
		{ 34, 154, 49 },
		{ 34, 155, 49 },
		{ 34, 156, 49 },
		{ 34, 157, 49 },
		{ 34, 158, 49 },
		{ 34, 159, 49 },
		{ 34, 160, 49 },
		{ 34, 161, 49 },
		{ 34, 162, 49 },
		{ 34, 163, 49 },
		{ 34, 164, 49 },
		{ 34, 165, 49 },
		{ 34, 166, 49 },
		{ 34, 167, 49 },
		{ 34, 168, 49 },
		{ 34, 169, 49 },
		{ 34, 170, 49 },
		{ 34, 171, 49 },
		{ 34, 172, 49 },
		{ 34, 173, 49 },
		{ 34, 174, 49 },
		{ 34, 175, 49 },
		{ 34, 176, 49 },
		{ 34, 177, 49 },
		{ 34, 178, 49 },
		{ 34, 179, 49 },
		{ 34, 180, 49 },
		{ 34, 181, 49 },
		{ 34, 182, 49 },
		{ 34, 183, 49 },
		{ 34, 184, 49 },
		{ 34, 185, 49 },
		{ 34, 186, 49 },
		{ 34, 187, 49 },
		{ 34, 188, 49 },
		{ 34, 189, 49 },
		{ 34, 190, 49 },
		{ 34, 191, 49 },
		{ 34, 192, 49 },
		{ 34, 193, 49 },
		{ 34, 194, 49 },
		{ 34, 195, 49 },
		{ 34, 196, 49 },
		{ 34, 197, 49 },
		{ 34, 198, 49 },
		{ 34, 199, 49 },
		{ 34, 200, 49 },
		{ 34, 201, 49 },
		{ 35, 147, 49 },
		{ 35, 148, 49 },
		{ 35, 149, 49 },
		{ 35, 150, 49 },
		{ 35, 151, 49 },
		{ 35, 152, 49 },
		{ 35, 153, 49 },
		{ 35, 154, 49 },
		{ 35, 155, 49 },
		{ 35, 156, 49 },
		{ 35, 157, 49 },
		{ 35, 158, 49 },
		{ 35, 159, 49 },
		{ 35, 160, 49 },
		{ 35, 161, 49 },
		{ 35, 162, 49 },
		{ 35, 163, 49 },
		{ 35, 164, 49 },
		{ 35, 165, 49 },
		{ 35, 166, 49 },
		{ 35, 167, 49 },
		{ 35, 168, 49 },
		{ 35, 169, 49 },
		{ 35, 170, 49 },
		{ 35, 171, 49 },
		{ 35, 172, 49 },
		{ 35, 173, 49 },
		{ 35, 174, 49 },
		{ 35, 175, 49 },
		{ 35, 176, 49 },
		{ 35, 177, 49 },
		{ 35, 178, 49 },
		{ 35, 179, 49 },
		{ 35, 180, 49 },
		{ 35, 181, 49 },
		{ 35, 182, 49 },
		{ 35, 183, 49 },
		{ 35, 184, 49 },
		{ 35, 185, 49 },
		{ 35, 186, 49 },
		{ 35, 187, 49 },
		{ 35, 188, 49 },
		{ 35, 189, 49 },
		{ 35, 190, 49 },
		{ 35, 191, 49 },
		{ 35, 192, 49 },
		{ 35, 193, 49 },
		{ 35, 194, 49 },
		{ 35, 195, 49 },
		{ 35, 196, 49 },
		{ 35, 197, 49 },
		{ 35, 198, 49 },
		{ 35, 199, 49 },
		{ 35, 200, 49 },
		{ 35, 201, 49 },
		{ 36, 147, 49 },
		{ 36, 148, 49 },
		{ 36, 149, 49 },
		{ 36, 150, 49 },
		{ 36, 151, 49 },
		{ 36, 152, 49 },
		{ 36, 153, 49 },
		{ 36, 154, 49 },
		{ 36, 155, 49 },
		{ 36, 156, 49 },
		{ 36, 157, 49 },
		{ 36, 158, 49 },
		{ 36, 159, 49 },
		{ 36, 160, 49 },
		{ 36, 161, 49 },
		{ 36, 162, 49 },
		{ 36, 163, 49 },
		{ 36, 164, 49 },
		{ 36, 165, 49 },
		{ 36, 166, 49 },
		{ 36, 167, 49 },
		{ 36, 168, 49 },
		{ 36, 169, 49 },
		{ 36, 170, 49 },
		{ 36, 171, 49 },
		{ 36, 172, 49 },
		{ 36, 173, 49 },
		{ 36, 174, 49 },
		{ 36, 175, 49 },
		{ 36, 176, 49 },
		{ 36, 177, 49 },
		{ 36, 178, 49 },
		{ 36, 179, 49 },
		{ 36, 180, 49 },
		{ 36, 181, 49 },
		{ 36, 182, 49 },
		{ 36, 183, 49 },
		{ 36, 184, 49 },
		{ 36, 185, 49 },
		{ 36, 186, 49 },
		{ 36, 187, 49 },
		{ 36, 188, 49 },
		{ 36, 189, 49 },
		{ 36, 190, 49 },
		{ 36, 191, 49 },
		{ 36, 192, 49 },
		{ 36, 193, 49 },
		{ 36, 194, 49 },
		{ 36, 195, 49 },
		{ 36, 196, 49 },
		{ 36, 197, 49 },
		{ 36, 198, 49 },
		{ 36, 199, 49 },
		{ 36, 200, 49 },
		{ 36, 201, 49 },
		{ 37, 148, 49 },
		{ 37, 149, 49 },
		{ 37, 150, 49 },
		{ 37, 151, 49 },
		{ 37, 152, 49 },
		{ 37, 153, 49 },
		{ 37, 154, 49 },
		{ 37, 155, 49 },
		{ 37, 156, 49 },
		{ 37, 157, 49 },
		{ 37, 158, 49 },
		{ 37, 159, 49 },
		{ 37, 160, 49 },
		{ 37, 161, 49 },
		{ 37, 162, 49 },
		{ 37, 163, 49 },
		{ 37, 164, 49 },
		{ 37, 165, 49 },
		{ 37, 166, 49 },
		{ 37, 167, 49 },
		{ 37, 168, 49 },
		{ 37, 169, 49 },
		{ 37, 170, 49 },
		{ 37, 171, 49 },
		{ 37, 172, 49 },
		{ 37, 173, 49 },
		{ 37, 174, 49 },
		{ 37, 175, 49 },
		{ 37, 176, 49 },
		{ 37, 177, 49 },
		{ 37, 178, 49 },
		{ 37, 179, 49 },
		{ 37, 180, 49 },
		{ 37, 181, 49 },
		{ 37, 182, 49 },
		{ 37, 183, 49 },
		{ 37, 184, 49 },
		{ 37, 185, 49 },
		{ 37, 186, 49 },
		{ 37, 187, 49 },
		{ 37, 188, 49 },
		{ 37, 189, 49 },
		{ 37, 190, 49 },
		{ 37, 191, 49 },
		{ 37, 192, 49 },
		{ 37, 193, 49 },
		{ 37, 194, 49 },
		{ 37, 195, 49 },
		{ 37, 196, 49 },
		{ 37, 197, 49 },
		{ 37, 198, 49 },
		{ 37, 199, 49 },
		{ 37, 200, 49 },
		{ 37, 201, 49 },
		{ 38, 148, 49 },
		{ 38, 149, 49 },
		{ 38, 150, 49 },
		{ 38, 151, 49 },
		{ 38, 152, 49 },
		{ 38, 153, 49 },
		{ 38, 154, 49 },
		{ 38, 155, 49 },
		{ 38, 156, 49 },
		{ 38, 157, 49 },
		{ 38, 158, 49 },
		{ 38, 159, 49 },
		{ 38, 160, 49 },
		{ 38, 161, 49 },
		{ 38, 162, 49 },
		{ 38, 163, 49 },
		{ 38, 164, 49 },
		{ 38, 165, 49 },
		{ 38, 166, 49 },
		{ 38, 167, 49 },
		{ 38, 168, 49 },
		{ 38, 169, 49 },
		{ 38, 170, 49 },
		{ 38, 171, 49 },
		{ 38, 172, 49 },
		{ 38, 173, 49 },
		{ 38, 174, 49 },
		{ 38, 175, 49 },
		{ 38, 176, 49 },
		{ 38, 177, 49 },
		{ 38, 178, 49 },
		{ 38, 179, 49 },
		{ 38, 180, 49 },
		{ 38, 181, 49 },
		{ 38, 182, 49 },
		{ 38, 183, 49 },
		{ 38, 184, 49 },
		{ 38, 185, 49 },
		{ 38, 186, 49 },
		{ 38, 187, 49 },
		{ 38, 188, 49 },
		{ 38, 189, 49 },
		{ 38, 190, 49 },
		{ 38, 191, 49 },
		{ 38, 192, 49 },
		{ 38, 193, 49 },
		{ 38, 194, 49 },
		{ 38, 195, 49 },
		{ 38, 196, 49 },
		{ 38, 197, 49 },
		{ 38, 198, 49 },
		{ 38, 199, 49 },
		{ 38, 200, 49 },
		{ 38, 201, 49 },
		{ 38, 202, 49 },
		{ 39, 149, 49 },
		{ 39, 150, 49 },
		{ 39, 151, 49 },
		{ 39, 152, 49 },
		{ 39, 153, 49 },
		{ 39, 154, 49 },
		{ 39, 155, 49 },
		{ 39, 156, 49 },
		{ 39, 157, 49 },
		{ 39, 158, 49 },
		{ 39, 159, 49 },
		{ 39, 160, 49 },
		{ 39, 161, 49 },
		{ 39, 162, 49 },
		{ 39, 163, 49 },
		{ 39, 164, 49 },
		{ 39, 165, 49 },
		{ 39, 166, 49 },
		{ 39, 167, 49 },
		{ 39, 168, 49 },
		{ 39, 169, 49 },
		{ 39, 170, 49 },
		{ 39, 171, 49 },
		{ 39, 172, 49 },
		{ 39, 173, 49 },
		{ 39, 174, 49 },
		{ 39, 175, 49 },
		{ 39, 176, 49 },
		{ 39, 177, 49 },
		{ 39, 178, 49 },
		{ 39, 179, 49 },
		{ 39, 180, 49 },
		{ 39, 181, 49 },
		{ 39, 182, 49 },
		{ 39, 183, 49 },
		{ 39, 184, 49 },
		{ 39, 185, 49 },
		{ 39, 186, 49 },
		{ 39, 187, 49 },
		{ 39, 188, 49 },
		{ 39, 189, 49 },
		{ 39, 190, 49 },
		{ 39, 191, 49 },
		{ 39, 192, 49 },
		{ 39, 193, 49 },
		{ 39, 194, 49 },
		{ 39, 195, 49 },
		{ 39, 196, 49 },
		{ 39, 197, 49 },
		{ 39, 198, 49 },
		{ 39, 199, 49 },
		{ 39, 200, 49 },
		{ 39, 201, 49 },
		{ 39, 202, 49 },
		{ 40, 149, 49 },
		{ 40, 150, 49 },
		{ 40, 151, 49 },
		{ 40, 152, 49 },
		{ 40, 153, 49 },
		{ 40, 154, 49 },
		{ 40, 155, 49 },
		{ 40, 156, 49 },
		{ 40, 157, 49 },
		{ 40, 158, 49 },
		{ 40, 159, 49 },
		{ 40, 160, 49 },
		{ 40, 161, 49 },
		{ 40, 162, 49 },
		{ 40, 163, 49 },
		{ 40, 164, 49 },
		{ 40, 165, 49 },
		{ 40, 166, 49 },
		{ 40, 167, 49 },
		{ 40, 168, 49 },
		{ 40, 169, 49 },
		{ 40, 170, 49 },
		{ 40, 171, 49 },
		{ 40, 172, 49 },
		{ 40, 173, 49 },
		{ 40, 174, 49 },
		{ 40, 175, 49 },
		{ 40, 176, 49 },
		{ 40, 177, 49 },
		{ 40, 178, 49 },
		{ 40, 179, 49 },
		{ 40, 180, 49 },
		{ 40, 181, 49 },
		{ 40, 182, 49 },
		{ 40, 183, 49 },
		{ 40, 184, 49 },
		{ 40, 185, 49 },
		{ 40, 186, 49 },
		{ 40, 187, 49 },
		{ 40, 188, 49 },
		{ 40, 189, 49 },
		{ 40, 190, 49 },
		{ 40, 191, 49 },
		{ 40, 192, 49 },
		{ 40, 193, 49 },
		{ 40, 194, 49 },
		{ 40, 195, 49 },
		{ 40, 196, 49 },
		{ 40, 197, 49 },
		{ 40, 198, 49 },
		{ 40, 199, 49 },
		{ 40, 200, 49 },
		{ 40, 201, 49 },
		{ 40, 202, 49 },
		{ 41, 150, 49 },
		{ 41, 151, 49 },
		{ 41, 152, 49 },
		{ 41, 153, 49 },
		{ 41, 154, 49 },
		{ 41, 155, 49 },
		{ 41, 156, 49 },
		{ 41, 157, 49 },
		{ 41, 158, 49 },
		{ 41, 159, 49 },
		{ 41, 160, 49 },
		{ 41, 161, 49 },
		{ 41, 162, 49 },
		{ 41, 163, 49 },
		{ 41, 164, 49 },
		{ 41, 165, 49 },
		{ 41, 166, 49 },
		{ 41, 167, 49 },
		{ 41, 168, 49 },
		{ 41, 169, 49 },
		{ 41, 170, 49 },
		{ 41, 171, 49 },
		{ 41, 172, 49 },
		{ 41, 173, 49 },
		{ 41, 174, 49 },
		{ 41, 175, 49 },
		{ 41, 176, 49 },
		{ 41, 177, 49 },
		{ 41, 178, 49 },
		{ 41, 179, 49 },
		{ 41, 180, 49 },
		{ 41, 181, 49 },
		{ 41, 182, 49 },
		{ 41, 183, 49 },
		{ 41, 184, 49 },
		{ 41, 185, 49 },
		{ 41, 186, 49 },
		{ 41, 187, 49 },
		{ 41, 188, 49 },
		{ 41, 189, 49 },
		{ 41, 190, 49 },
		{ 41, 191, 49 },
		{ 41, 192, 49 },
		{ 41, 193, 49 },
		{ 41, 194, 49 },
		{ 41, 195, 49 },
		{ 41, 196, 49 },
		{ 41, 197, 49 },
		{ 41, 198, 49 },
		{ 41, 199, 49 },
		{ 41, 200, 49 },
		{ 41, 201, 49 },
		{ 41, 202, 49 },
		{ 42, 150, 49 },
		{ 42, 151, 49 },
		{ 42, 152, 49 },
		{ 42, 153, 49 },
		{ 42, 154, 49 },
		{ 42, 155, 49 },
		{ 42, 156, 49 },
		{ 42, 157, 49 },
		{ 42, 158, 49 },
		{ 42, 159, 49 },
		{ 42, 160, 49 },
		{ 42, 161, 49 },
		{ 42, 162, 49 },
		{ 42, 163, 49 },
		{ 42, 164, 49 },
		{ 42, 165, 49 },
		{ 42, 166, 49 },
		{ 42, 167, 49 },
		{ 42, 168, 49 },
		{ 42, 169, 49 },
		{ 42, 170, 49 },
		{ 42, 171, 49 },
		{ 42, 172, 49 },
		{ 42, 173, 49 },
		{ 42, 174, 49 },
		{ 42, 175, 49 },
		{ 42, 176, 49 },
		{ 42, 177, 49 },
		{ 42, 178, 49 },
		{ 42, 179, 49 },
		{ 42, 180, 49 },
		{ 42, 181, 49 },
		{ 42, 182, 49 },
		{ 42, 183, 49 },
		{ 42, 184, 49 },
		{ 42, 185, 49 },
		{ 42, 186, 49 },
		{ 42, 187, 49 },
		{ 42, 188, 49 },
		{ 42, 189, 49 },
		{ 42, 190, 49 },
		{ 42, 191, 49 },
		{ 42, 192, 49 },
		{ 42, 193, 49 },
		{ 42, 194, 49 },
		{ 42, 195, 49 },
		{ 42, 196, 49 },
		{ 42, 197, 49 },
		{ 42, 198, 49 },
		{ 42, 199, 49 },
		{ 42, 200, 49 },
		{ 42, 201, 49 },
		{ 42, 202, 49 },
		{ 42, 203, 49 },
		{ 43, 151, 49 },
		{ 43, 152, 49 },
		{ 43, 153, 49 },
		{ 43, 154, 49 },
		{ 43, 155, 49 },
		{ 43, 156, 49 },
		{ 43, 157, 49 },
		{ 43, 158, 49 },
		{ 43, 159, 49 },
		{ 43, 160, 49 },
		{ 43, 161, 49 },
		{ 43, 162, 49 },
		{ 43, 163, 49 },
		{ 43, 164, 49 },
		{ 43, 165, 49 },
		{ 43, 166, 49 },
		{ 43, 167, 49 },
		{ 43, 168, 49 },
		{ 43, 169, 49 },
		{ 43, 170, 49 },
		{ 43, 171, 49 },
		{ 43, 172, 49 },
		{ 43, 173, 49 },
		{ 43, 174, 49 },
		{ 43, 175, 49 },
		{ 43, 176, 49 },
		{ 43, 177, 49 },
		{ 43, 178, 49 },
		{ 43, 179, 49 },
		{ 43, 180, 49 },
		{ 43, 181, 49 },
		{ 43, 182, 49 },
		{ 43, 183, 49 },
		{ 43, 184, 49 },
		{ 43, 185, 49 },
		{ 43, 186, 49 },
		{ 43, 187, 49 },
		{ 43, 188, 49 },
		{ 43, 189, 49 },
		{ 43, 190, 49 },
		{ 43, 191, 49 },
		{ 43, 192, 49 },
		{ 43, 193, 49 },
		{ 43, 194, 49 },
		{ 43, 195, 49 },
		{ 43, 196, 49 },
		{ 43, 197, 49 },
		{ 43, 198, 49 },
		{ 43, 199, 49 },
		{ 43, 200, 49 },
		{ 43, 201, 49 },
		{ 43, 202, 49 },
		{ 43, 203, 49 },
		{ 44, 151, 49 },
		{ 44, 152, 49 },
		{ 44, 153, 49 },
		{ 44, 154, 49 },
		{ 44, 155, 49 },
		{ 44, 156, 49 },
		{ 44, 157, 49 },
		{ 44, 158, 49 },
		{ 44, 159, 49 },
		{ 44, 160, 49 },
		{ 44, 161, 49 },
		{ 44, 162, 49 },
		{ 44, 163, 49 },
		{ 44, 164, 49 },
		{ 44, 165, 49 },
		{ 44, 166, 49 },
		{ 44, 167, 49 },
		{ 44, 168, 49 },
		{ 44, 169, 49 },
		{ 44, 170, 49 },
		{ 44, 171, 49 },
		{ 44, 172, 49 },
		{ 44, 173, 49 },
		{ 44, 174, 49 },
		{ 44, 175, 49 },
		{ 44, 176, 49 },
		{ 44, 177, 49 },
		{ 44, 178, 49 },
		{ 44, 179, 49 },
		{ 44, 180, 49 },
		{ 44, 181, 49 },
		{ 44, 182, 49 },
		{ 44, 183, 49 },
		{ 44, 184, 49 },
		{ 44, 185, 49 },
		{ 44, 186, 49 },
		{ 44, 187, 49 },
		{ 44, 188, 49 },
		{ 44, 189, 49 },
		{ 44, 190, 49 },
		{ 44, 191, 49 },
		{ 44, 192, 49 },
		{ 44, 193, 49 },
		{ 44, 194, 49 },
		{ 44, 195, 49 },
		{ 44, 196, 49 },
		{ 44, 197, 49 },
		{ 44, 198, 49 },
		{ 44, 199, 49 },
		{ 44, 200, 49 },
		{ 44, 201, 49 },
		{ 44, 202, 49 },
		{ 44, 203, 49 },
		{ 45, 152, 49 },
		{ 45, 153, 49 },
		{ 45, 154, 49 },
		{ 45, 155, 49 },
		{ 45, 156, 49 },
		{ 45, 157, 49 },
		{ 45, 158, 49 },
		{ 45, 159, 49 },
		{ 45, 160, 49 },
		{ 45, 161, 49 },
		{ 45, 162, 49 },
		{ 45, 163, 49 },
		{ 45, 164, 49 },
		{ 45, 165, 49 },
		{ 45, 166, 49 },
		{ 45, 167, 49 },
		{ 45, 168, 49 },
		{ 45, 169, 49 },
		{ 45, 170, 49 },
		{ 45, 171, 49 },
		{ 45, 172, 49 },
		{ 45, 173, 49 },
		{ 45, 174, 49 },
		{ 45, 175, 49 },
		{ 45, 176, 49 },
		{ 45, 177, 49 },
		{ 45, 178, 49 },
		{ 45, 179, 49 },
		{ 45, 180, 49 },
		{ 45, 181, 49 },
		{ 45, 182, 49 },
		{ 45, 183, 49 },
		{ 45, 184, 49 },
		{ 45, 185, 49 },
		{ 45, 186, 49 },
		{ 45, 187, 49 },
		{ 45, 188, 49 },
		{ 45, 189, 49 },
		{ 45, 190, 49 },
		{ 45, 191, 49 },
		{ 45, 192, 49 },
		{ 45, 193, 49 },
		{ 45, 194, 49 },
		{ 45, 195, 49 },
		{ 45, 196, 49 },
		{ 45, 197, 49 },
		{ 45, 198, 49 },
		{ 45, 199, 49 },
		{ 45, 200, 49 },
		{ 45, 201, 49 },
		{ 45, 202, 49 },
		{ 45, 203, 49 },
		{ 46, 152, 49 },
		{ 46, 153, 49 },
		{ 46, 154, 49 },
		{ 46, 155, 49 },
		{ 46, 156, 49 },
		{ 46, 157, 49 },
		{ 46, 158, 49 },
		{ 46, 159, 49 },
		{ 46, 160, 49 },
		{ 46, 161, 49 },
		{ 46, 162, 49 },
		{ 46, 163, 49 },
		{ 46, 164, 49 },
		{ 46, 165, 49 },
		{ 46, 166, 49 },
		{ 46, 167, 49 },
		{ 46, 168, 49 },
		{ 46, 169, 49 },
		{ 46, 170, 49 },
		{ 46, 171, 49 },
		{ 46, 172, 49 },
		{ 46, 173, 49 },
		{ 46, 174, 49 },
		{ 46, 175, 49 },
		{ 46, 176, 49 },
		{ 46, 177, 49 },
		{ 46, 178, 49 },
		{ 46, 179, 49 },
		{ 46, 180, 49 },
		{ 46, 181, 49 },
		{ 46, 182, 49 },
		{ 46, 183, 49 },
		{ 46, 184, 49 },
		{ 46, 185, 49 },
		{ 46, 186, 49 },
		{ 46, 187, 49 },
		{ 46, 188, 49 },
		{ 46, 189, 49 },
		{ 46, 190, 49 },
		{ 46, 191, 49 },
		{ 46, 192, 49 },
		{ 46, 193, 49 },
		{ 46, 194, 49 },
		{ 46, 195, 49 },
		{ 46, 196, 49 },
		{ 46, 197, 49 },
		{ 46, 198, 49 },
		{ 46, 199, 49 },
		{ 46, 200, 49 },
		{ 46, 201, 49 },
		{ 46, 202, 49 },
		{ 46, 203, 49 },
		{ 46, 204, 49 },
		{ 47, 153, 49 },
		{ 47, 154, 49 },
		{ 47, 155, 49 },
		{ 47, 156, 49 },
		{ 47, 157, 49 },
		{ 47, 158, 49 },
		{ 47, 159, 49 },
		{ 47, 160, 49 },
		{ 47, 161, 49 },
		{ 47, 162, 49 },
		{ 47, 163, 49 },
		{ 47, 164, 49 },
		{ 47, 165, 49 },
		{ 47, 166, 49 },
		{ 47, 167, 49 },
		{ 47, 168, 49 },
		{ 47, 169, 49 },
		{ 47, 170, 49 },
		{ 47, 171, 49 },
		{ 47, 172, 49 },
		{ 47, 173, 49 },
		{ 47, 174, 49 },
		{ 47, 175, 49 },
		{ 47, 176, 49 },
		{ 47, 177, 49 },
		{ 47, 178, 49 },
		{ 47, 179, 49 },
		{ 47, 180, 49 },
		{ 47, 181, 49 },
		{ 47, 182, 49 },
		{ 47, 183, 49 },
		{ 47, 184, 49 },
		{ 47, 185, 49 },
		{ 47, 186, 49 },
		{ 47, 187, 49 },
		{ 47, 188, 49 },
		{ 47, 189, 49 },
		{ 47, 190, 49 },
		{ 47, 191, 49 },
		{ 47, 192, 49 },
		{ 47, 193, 49 },
		{ 47, 194, 49 },
		{ 47, 195, 49 },
		{ 47, 196, 49 },
		{ 47, 197, 49 },
		{ 47, 198, 49 },
		{ 47, 199, 49 },
		{ 47, 200, 49 },
		{ 47, 201, 49 },
		{ 47, 202, 49 },
		{ 47, 203, 49 },
		{ 47, 204, 49 },
		{ 48, 153, 49 },
		{ 48, 154, 49 },
		{ 48, 155, 49 },
		{ 48, 156, 49 },
		{ 48, 157, 49 },
		{ 48, 158, 49 },
		{ 48, 159, 49 },
		{ 48, 160, 49 },
		{ 48, 161, 49 },
		{ 48, 162, 49 },
		{ 48, 163, 49 },
		{ 48, 164, 49 },
		{ 48, 165, 49 },
		{ 48, 166, 49 },
		{ 48, 167, 49 },
		{ 48, 168, 49 },
		{ 48, 169, 49 },
		{ 48, 170, 49 },
		{ 48, 171, 49 },
		{ 48, 172, 49 },
		{ 48, 173, 49 },
		{ 48, 174, 49 },
		{ 48, 175, 49 },
		{ 48, 176, 49 },
		{ 48, 177, 49 },
		{ 48, 178, 49 },
		{ 48, 179, 49 },
		{ 48, 180, 49 },
		{ 48, 181, 49 },
		{ 48, 182, 49 },
		{ 48, 183, 49 },
		{ 48, 184, 49 },
		{ 48, 185, 49 },
		{ 48, 186, 49 },
		{ 48, 187, 49 },
		{ 48, 188, 49 },
		{ 48, 189, 49 },
		{ 48, 190, 49 },
		{ 48, 191, 49 },
		{ 48, 192, 49 },
		{ 48, 193, 49 },
		{ 48, 194, 49 },
		{ 48, 195, 49 },
		{ 48, 196, 49 },
		{ 48, 197, 49 },
		{ 48, 198, 49 },
		{ 48, 199, 49 },
		{ 48, 200, 49 },
		{ 48, 201, 49 },
		{ 48, 202, 49 },
		{ 48, 203, 49 },
		{ 48, 204, 49 },
		{ 49, 154, 49 },
		{ 49, 155, 49 },
		{ 49, 156, 49 },
		{ 49, 157, 49 },
		{ 49, 158, 49 },
		{ 49, 159, 49 },
		{ 49, 160, 49 },
		{ 49, 161, 49 },
		{ 49, 162, 49 },
		{ 49, 163, 49 },
		{ 49, 164, 49 },
		{ 49, 165, 49 },
		{ 49, 166, 49 },
		{ 49, 167, 49 },
		{ 49, 168, 49 },
		{ 49, 169, 49 },
		{ 49, 170, 49 },
		{ 49, 171, 49 },
		{ 49, 172, 49 },
		{ 49, 173, 49 },
		{ 49, 174, 49 },
		{ 49, 175, 49 },
		{ 49, 176, 49 },
		{ 49, 177, 49 },
		{ 49, 178, 49 },
		{ 49, 179, 49 },
		{ 49, 180, 49 },
		{ 49, 181, 49 },
		{ 49, 182, 49 },
		{ 49, 183, 49 },
		{ 49, 184, 49 },
		{ 49, 185, 49 },
		{ 49, 186, 49 },
		{ 49, 187, 49 },
		{ 49, 188, 49 },
		{ 49, 189, 49 },
		{ 49, 190, 49 },
		{ 49, 191, 49 },
		{ 49, 192, 49 },
		{ 49, 193, 49 },
		{ 49, 194, 49 },
		{ 49, 195, 49 },
		{ 49, 196, 49 },
		{ 49, 197, 49 },
		{ 49, 198, 49 },
		{ 49, 199, 49 },
		{ 49, 200, 49 },
		{ 49, 201, 49 },
		{ 49, 202, 49 },
		{ 49, 203, 49 },
		{ 49, 204, 49 },
		{ 50, 154, 49 },
		{ 50, 155, 49 },
		{ 50, 156, 49 },
		{ 50, 157, 49 },
		{ 50, 158, 49 },
		{ 50, 159, 49 },
		{ 50, 160, 49 },
		{ 50, 161, 49 },
		{ 50, 162, 49 },
		{ 50, 163, 49 },
		{ 50, 164, 49 },
		{ 50, 165, 49 },
		{ 50, 166, 49 },
		{ 50, 167, 49 },
		{ 50, 168, 49 },
		{ 50, 169, 49 },
		{ 50, 170, 49 },
		{ 50, 171, 49 },
		{ 50, 172, 49 },
		{ 50, 173, 49 },
		{ 50, 174, 49 },
		{ 50, 175, 49 },
		{ 50, 176, 49 },
		{ 50, 177, 49 },
		{ 50, 178, 49 },
		{ 50, 179, 49 },
		{ 50, 180, 49 },
		{ 50, 181, 49 },
		{ 50, 182, 49 },
		{ 50, 183, 49 },
		{ 50, 184, 49 },
		{ 50, 185, 49 },
		{ 50, 186, 49 },
		{ 50, 187, 49 },
		{ 50, 188, 49 },
		{ 50, 189, 49 },
		{ 50, 190, 49 },
		{ 50, 191, 49 },
		{ 50, 192, 49 },
		{ 50, 193, 49 },
		{ 50, 194, 49 },
		{ 50, 195, 49 },
		{ 50, 196, 49 },
		{ 50, 197, 49 },
		{ 50, 198, 49 },
		{ 50, 199, 49 },
		{ 50, 200, 49 },
		{ 50, 201, 49 },
		{ 50, 202, 49 },
		{ 50, 203, 49 },
		{ 50, 204, 49 },
		{ 50, 205, 49 },
		{ 51, 155, 49 },
		{ 51, 156, 49 },
		{ 51, 157, 49 },
		{ 51, 158, 49 },
		{ 51, 159, 49 },
		{ 51, 160, 49 },
		{ 51, 161, 49 },
		{ 51, 162, 49 },
		{ 51, 163, 49 },
		{ 51, 164, 49 },
		{ 51, 165, 49 },
		{ 51, 166, 49 },
		{ 51, 167, 49 },
		{ 51, 168, 49 },
		{ 51, 169, 49 },
		{ 51, 170, 49 },
		{ 51, 171, 49 },
		{ 51, 172, 49 },
		{ 51, 173, 49 },
		{ 51, 174, 49 },
		{ 51, 175, 49 },
		{ 51, 176, 49 },
		{ 51, 177, 49 },
		{ 51, 178, 49 },
		{ 51, 179, 49 },
		{ 51, 180, 49 },
		{ 51, 181, 49 },
		{ 51, 182, 49 },
		{ 51, 183, 49 },
		{ 51, 184, 49 },
		{ 51, 185, 49 },
		{ 51, 186, 49 },
		{ 51, 187, 49 },
		{ 51, 188, 49 },
		{ 51, 189, 49 },
		{ 51, 190, 49 },
		{ 51, 191, 49 },
		{ 51, 192, 49 },
		{ 51, 193, 49 },
		{ 51, 194, 49 },
		{ 51, 195, 49 },
		{ 51, 196, 49 },
		{ 51, 197, 49 },
		{ 51, 198, 49 },
		{ 51, 199, 49 },
		{ 51, 200, 49 },
		{ 51, 201, 49 },
		{ 51, 202, 49 },
		{ 51, 203, 49 },
		{ 51, 204, 49 },
		{ 51, 205, 49 },
		{ 52, 155, 49 },
		{ 52, 156, 49 },
		{ 52, 157, 49 },
		{ 52, 158, 49 },
		{ 52, 159, 49 },
		{ 52, 160, 49 },
		{ 52, 161, 49 },
		{ 52, 162, 49 },
		{ 52, 163, 49 },
		{ 52, 164, 49 },
		{ 52, 165, 49 },
		{ 52, 166, 49 },
		{ 52, 167, 49 },
		{ 52, 168, 49 },
		{ 52, 169, 49 },
		{ 52, 170, 49 },
		{ 52, 171, 49 },
		{ 52, 172, 49 },
		{ 52, 173, 49 },
		{ 52, 174, 49 },
		{ 52, 175, 49 },
		{ 52, 176, 49 },
		{ 52, 177, 49 },
		{ 52, 178, 49 },
		{ 52, 179, 49 },
		{ 52, 180, 49 },
		{ 52, 181, 49 },
		{ 52, 182, 49 },
		{ 52, 183, 49 },
		{ 52, 184, 49 },
		{ 52, 185, 49 },
		{ 52, 186, 49 },
		{ 52, 187, 49 },
		{ 52, 188, 49 },
		{ 52, 189, 49 },
		{ 52, 190, 49 },
		{ 52, 191, 49 },
		{ 52, 192, 49 },
		{ 52, 193, 49 },
		{ 52, 194, 49 },
		{ 52, 195, 49 },
		{ 52, 196, 49 },
		{ 52, 197, 49 },
		{ 52, 198, 49 },
		{ 52, 199, 49 },
		{ 52, 200, 49 },
		{ 52, 201, 49 },
		{ 52, 202, 49 },
		{ 52, 203, 49 },
		{ 52, 204, 49 },
		{ 52, 205, 49 },
		{ 53, 156, 49 },
		{ 53, 157, 49 },
		{ 53, 158, 49 },
		{ 53, 159, 49 },
		{ 53, 160, 49 },
		{ 53, 161, 49 },
		{ 53, 162, 49 },
		{ 53, 163, 49 },
		{ 53, 164, 49 },
		{ 53, 165, 49 },
		{ 53, 166, 49 },
		{ 53, 167, 49 },
		{ 53, 168, 49 },
		{ 53, 169, 49 },
		{ 53, 170, 49 },
		{ 53, 171, 49 },
		{ 53, 172, 49 },
		{ 53, 173, 49 },
		{ 53, 174, 49 },
		{ 53, 175, 49 },
		{ 53, 176, 49 },
		{ 53, 177, 49 },
		{ 53, 178, 49 },
		{ 53, 179, 49 },
		{ 53, 180, 49 },
		{ 53, 181, 49 },
		{ 53, 182, 49 },
		{ 53, 183, 49 },
		{ 53, 184, 49 },
		{ 53, 185, 49 },
		{ 53, 186, 49 },
		{ 53, 187, 49 },
		{ 53, 188, 49 },
		{ 53, 189, 49 },
		{ 53, 190, 49 },
		{ 53, 191, 49 },
		{ 53, 192, 49 },
		{ 53, 193, 49 },
		{ 53, 194, 49 },
		{ 53, 195, 49 },
		{ 53, 196, 49 },
		{ 53, 197, 49 },
		{ 53, 198, 49 },
		{ 53, 199, 49 },
		{ 53, 200, 49 },
		{ 53, 201, 49 },
		{ 53, 202, 49 },
		{ 53, 203, 49 },
		{ 53, 204, 49 },
		{ 53, 205, 49 },
		{ 54, 156, 49 },
		{ 54, 157, 49 },
		{ 54, 158, 49 },
		{ 54, 159, 49 },
		{ 54, 160, 49 },
		{ 54, 161, 49 },
		{ 54, 162, 49 },
		{ 54, 163, 49 },
		{ 54, 164, 49 },
		{ 54, 165, 49 },
		{ 54, 166, 49 },
		{ 54, 167, 49 },
		{ 54, 168, 49 },
		{ 54, 169, 49 },
		{ 54, 170, 49 },
		{ 54, 171, 49 },
		{ 54, 172, 49 },
		{ 54, 173, 49 },
		{ 54, 174, 49 },
		{ 54, 175, 49 },
		{ 54, 176, 49 },
		{ 54, 177, 49 },
		{ 54, 178, 49 },
		{ 54, 179, 49 },
		{ 54, 180, 49 },
		{ 54, 181, 49 },
		{ 54, 182, 49 },
		{ 54, 183, 49 },
		{ 54, 184, 49 },
		{ 54, 185, 49 },
		{ 54, 186, 49 },
		{ 54, 187, 49 },
		{ 54, 188, 49 },
		{ 54, 189, 49 },
		{ 54, 190, 49 },
		{ 54, 191, 49 },
		{ 54, 192, 49 },
		{ 54, 193, 49 },
		{ 54, 194, 49 },
		{ 54, 195, 49 },
		{ 54, 196, 49 },
		{ 54, 197, 49 },
		{ 54, 198, 49 },
		{ 54, 199, 49 },
		{ 54, 200, 49 },
		{ 54, 201, 49 },
		{ 54, 202, 49 },
		{ 54, 203, 49 },
		{ 54, 204, 49 },
		{ 54, 205, 49 },
		{ 54, 206, 49 },
		{ 55, 157, 49 },
		{ 55, 158, 49 },
		{ 55, 159, 49 },
		{ 55, 160, 49 },
		{ 55, 161, 49 },
		{ 55, 162, 49 },
		{ 55, 163, 49 },
		{ 55, 164, 49 },
		{ 55, 165, 49 },
		{ 55, 166, 49 },
		{ 55, 167, 49 },
		{ 55, 168, 49 },
		{ 55, 169, 49 },
		{ 55, 170, 49 },
		{ 55, 171, 49 },
		{ 55, 172, 49 },
		{ 55, 173, 49 },
		{ 55, 174, 49 },
		{ 55, 175, 49 },
		{ 55, 176, 49 },
		{ 55, 177, 49 },
		{ 55, 178, 49 },
		{ 55, 179, 49 },
		{ 55, 180, 49 },
		{ 55, 181, 49 },
		{ 55, 182, 49 },
		{ 55, 183, 49 },
		{ 55, 184, 49 },
		{ 55, 185, 49 },
		{ 55, 186, 49 },
		{ 55, 187, 49 },
		{ 55, 188, 49 },
		{ 55, 189, 49 },
		{ 55, 190, 49 },
		{ 55, 191, 49 },
		{ 55, 192, 49 },
		{ 55, 193, 49 },
		{ 55, 194, 49 },
		{ 55, 195, 49 },
		{ 55, 196, 49 },
		{ 55, 197, 49 },
		{ 55, 198, 49 },
		{ 55, 199, 49 },
		{ 55, 200, 49 },
		{ 55, 201, 49 },
		{ 55, 202, 49 },
		{ 55, 203, 49 },
		{ 55, 204, 49 },
		{ 55, 205, 49 },
		{ 55, 206, 49 },
		{ 56, 157, 49 },
		{ 56, 158, 49 },
		{ 56, 159, 49 },
		{ 56, 160, 49 },
		{ 56, 161, 49 },
		{ 56, 162, 49 },
		{ 56, 163, 49 },
		{ 56, 164, 49 },
		{ 56, 165, 49 },
		{ 56, 166, 49 },
		{ 56, 167, 49 },
		{ 56, 168, 49 },
		{ 56, 169, 49 },
		{ 56, 170, 49 },
		{ 56, 171, 49 },
		{ 56, 172, 49 },
		{ 56, 173, 49 },
		{ 56, 174, 49 },
		{ 56, 175, 49 },
		{ 56, 176, 49 },
		{ 56, 177, 49 },
		{ 56, 178, 49 },
		{ 56, 179, 49 },
		{ 56, 180, 49 },
		{ 56, 181, 49 },
		{ 56, 182, 49 },
		{ 56, 183, 49 },
		{ 56, 184, 49 },
		{ 56, 185, 49 },
		{ 56, 186, 49 },
		{ 56, 187, 49 },
		{ 56, 188, 49 },
		{ 56, 189, 49 },
		{ 56, 190, 49 },
		{ 56, 191, 49 },
		{ 56, 192, 49 },
		{ 56, 193, 49 },
		{ 56, 194, 49 },
		{ 56, 195, 49 },
		{ 56, 196, 49 },
		{ 56, 197, 49 },
		{ 56, 198, 49 },
		{ 56, 199, 49 },
		{ 56, 200, 49 },
		{ 56, 201, 49 },
		{ 56, 202, 49 },
		{ 56, 203, 49 },
		{ 56, 204, 49 },
		{ 56, 205, 49 },
		{ 56, 206, 49 },
		{ 57, 158, 49 },
		{ 57, 159, 49 },
		{ 57, 160, 49 },
		{ 57, 161, 49 },
		{ 57, 162, 49 },
		{ 57, 163, 49 },
		{ 57, 164, 49 },
		{ 57, 165, 49 },
		{ 57, 166, 49 },
		{ 57, 167, 49 },
		{ 57, 168, 49 },
		{ 57, 169, 49 },
		{ 57, 170, 49 },
		{ 57, 171, 49 },
		{ 57, 172, 49 },
		{ 57, 173, 49 },
		{ 57, 174, 49 },
		{ 57, 175, 49 },
		{ 57, 176, 49 },
		{ 57, 177, 49 },
		{ 57, 178, 49 },
		{ 57, 179, 49 },
		{ 57, 180, 49 },
		{ 57, 181, 49 },
		{ 57, 182, 49 },
		{ 57, 183, 49 },
		{ 57, 184, 49 },
		{ 57, 185, 49 },
		{ 57, 186, 49 },
		{ 57, 187, 49 },
		{ 57, 188, 49 },
		{ 57, 189, 49 },
		{ 57, 190, 49 },
		{ 57, 191, 49 },
		{ 57, 192, 49 },
		{ 57, 193, 49 },
		{ 57, 194, 49 },
		{ 57, 195, 49 },
		{ 57, 196, 49 },
		{ 57, 197, 49 },
		{ 57, 198, 49 },
		{ 57, 199, 49 },
		{ 57, 200, 49 },
		{ 57, 201, 49 },
		{ 57, 202, 49 },
		{ 57, 203, 49 },
		{ 57, 204, 49 },
		{ 57, 205, 49 },
		{ 57, 206, 49 },
		{ 58, 158, 49 },
		{ 58, 159, 49 },
		{ 58, 160, 49 },
		{ 58, 161, 49 },
		{ 58, 162, 49 },
		{ 58, 163, 49 },
		{ 58, 164, 49 },
		{ 58, 165, 49 },
		{ 58, 166, 49 },
		{ 58, 167, 49 },
		{ 58, 168, 49 },
		{ 58, 169, 49 },
		{ 58, 170, 49 },
		{ 58, 171, 49 },
		{ 58, 172, 49 },
		{ 58, 173, 49 },
		{ 58, 174, 49 },
		{ 58, 175, 49 },
		{ 58, 176, 49 },
		{ 58, 177, 49 },
		{ 58, 178, 49 },
		{ 58, 179, 49 },
		{ 58, 180, 49 },
		{ 58, 181, 49 },
		{ 58, 182, 49 },
		{ 58, 183, 49 },
		{ 58, 184, 49 },
		{ 58, 185, 49 },
		{ 58, 186, 49 },
		{ 58, 187, 49 },
		{ 58, 188, 49 },
		{ 58, 189, 49 },
		{ 58, 190, 49 },
		{ 58, 191, 49 },
		{ 58, 192, 49 },
		{ 58, 193, 49 },
		{ 58, 194, 49 },
		{ 58, 195, 49 },
		{ 58, 196, 49 },
		{ 58, 197, 49 },
		{ 58, 198, 49 },
		{ 58, 199, 49 },
		{ 58, 200, 49 },
		{ 58, 201, 49 },
		{ 58, 202, 49 },
		{ 58, 203, 49 },
		{ 58, 204, 49 },
		{ 58, 205, 49 },
		{ 58, 206, 49 },
		{ 58, 207, 49 },
		{ 59, 159, 49 },
		{ 59, 160, 49 },
		{ 59, 161, 49 },
		{ 59, 162, 49 },
		{ 59, 163, 49 },
		{ 59, 164, 49 },
		{ 59, 165, 49 },
		{ 59, 166, 49 },
		{ 59, 167, 49 },
		{ 59, 168, 49 },
		{ 59, 169, 49 },
		{ 59, 170, 49 },
		{ 59, 171, 49 },
		{ 59, 172, 49 },
		{ 59, 173, 49 },
		{ 59, 174, 49 },
		{ 59, 175, 49 },
		{ 59, 176, 49 },
		{ 59, 177, 49 },
		{ 59, 178, 49 },
		{ 59, 179, 49 },
		{ 59, 180, 49 },
		{ 59, 181, 49 },
		{ 59, 182, 49 },
		{ 59, 183, 49 },
		{ 59, 184, 49 },
		{ 59, 185, 49 },
		{ 59, 186, 49 },
		{ 59, 187, 49 },
		{ 59, 188, 49 },
		{ 59, 189, 49 },
		{ 59, 190, 49 },
		{ 59, 191, 49 },
		{ 59, 192, 49 },
		{ 59, 193, 49 },
		{ 59, 194, 49 },
		{ 59, 195, 49 },
		{ 59, 196, 49 },
		{ 59, 197, 49 },
		{ 59, 198, 49 },
		{ 59, 199, 49 },
		{ 59, 200, 49 },
		{ 59, 201, 49 },
		{ 59, 202, 49 },
		{ 59, 203, 49 },
		{ 59, 204, 49 },
		{ 59, 205, 49 },
		{ 59, 206, 49 },
		{ 59, 207, 49 },
		{ 60, 159, 49 },
		{ 60, 160, 49 },
		{ 60, 161, 49 },
		{ 60, 162, 49 },
		{ 60, 163, 49 },
		{ 60, 164, 49 },
		{ 60, 165, 49 },
		{ 60, 166, 49 },
		{ 60, 167, 49 },
		{ 60, 168, 49 },
		{ 60, 169, 49 },
		{ 60, 170, 49 },
		{ 60, 171, 49 },
		{ 60, 172, 49 },
		{ 60, 173, 49 },
		{ 60, 174, 49 },
		{ 60, 175, 49 },
		{ 60, 176, 49 },
		{ 60, 177, 49 },
		{ 60, 178, 49 },
		{ 60, 179, 49 },
		{ 60, 180, 49 },
		{ 60, 181, 49 },
		{ 60, 182, 49 },
		{ 60, 183, 49 },
		{ 60, 184, 49 },
		{ 60, 185, 49 },
		{ 60, 186, 49 },
		{ 60, 187, 49 },
		{ 60, 188, 49 },
		{ 60, 189, 49 },
		{ 60, 190, 49 },
		{ 60, 191, 49 },
		{ 60, 192, 49 },
		{ 60, 193, 49 },
		{ 60, 194, 49 },
		{ 60, 195, 49 },
		{ 60, 196, 49 },
		{ 60, 197, 49 },
		{ 60, 198, 49 },
		{ 60, 199, 49 },
		{ 60, 200, 49 },
		{ 60, 201, 49 },
		{ 60, 202, 49 },
		{ 60, 203, 49 },
		{ 60, 204, 49 },
		{ 60, 205, 49 },
		{ 60, 206, 49 },
		{ 60, 207, 49 },
		{ 61, 160, 49 },
		{ 61, 161, 49 },
		{ 61, 162, 49 },
		{ 61, 163, 49 },
		{ 61, 164, 49 },
		{ 61, 165, 49 },
		{ 61, 166, 49 },
		{ 61, 167, 49 },
		{ 61, 168, 49 },
		{ 61, 169, 49 },
		{ 61, 170, 49 },
		{ 61, 171, 49 },
		{ 61, 172, 49 },
		{ 61, 173, 49 },
		{ 61, 174, 49 },
		{ 61, 175, 49 },
		{ 61, 176, 49 },
		{ 61, 177, 49 },
		{ 61, 178, 49 },
		{ 61, 179, 49 },
		{ 61, 180, 49 },
		{ 61, 181, 49 },
		{ 61, 182, 49 },
		{ 61, 183, 49 },
		{ 61, 184, 49 },
		{ 61, 185, 49 },
		{ 61, 186, 49 },
		{ 61, 187, 49 },
		{ 61, 188, 49 },
		{ 61, 189, 49 },
		{ 61, 190, 49 },
		{ 61, 191, 49 },
		{ 61, 192, 49 },
		{ 61, 193, 49 },
		{ 61, 194, 49 },
		{ 61, 195, 49 },
		{ 61, 196, 49 },
		{ 61, 197, 49 },
		{ 61, 198, 49 },
		{ 61, 199, 49 },
		{ 61, 200, 49 },
		{ 61, 201, 49 },
		{ 61, 202, 49 },
		{ 61, 203, 49 },
		{ 61, 204, 49 },
		{ 61, 205, 49 },
		{ 61, 206, 49 },
		{ 61, 207, 49 },
		{ 62, 160, 49 },
		{ 62, 161, 49 },
		{ 62, 162, 49 },
		{ 62, 163, 49 },
		{ 62, 164, 49 },
		{ 62, 165, 49 },
		{ 62, 166, 49 },
		{ 62, 167, 49 },
		{ 62, 168, 49 },
		{ 62, 169, 49 },
		{ 62, 170, 49 },
		{ 62, 171, 49 },
		{ 62, 172, 49 },
		{ 62, 173, 49 },
		{ 62, 174, 49 },
		{ 62, 175, 49 },
		{ 62, 176, 49 },
		{ 62, 177, 49 },
		{ 62, 178, 49 },
		{ 62, 179, 49 },
		{ 62, 180, 49 },
		{ 62, 181, 49 },
		{ 62, 182, 49 },
		{ 62, 183, 49 },
		{ 62, 184, 49 },
		{ 62, 185, 49 },
		{ 62, 186, 49 },
		{ 62, 187, 49 },
		{ 62, 188, 49 },
		{ 62, 189, 49 },
		{ 62, 190, 49 },
		{ 62, 191, 49 },
		{ 62, 192, 49 },
		{ 62, 193, 49 },
		{ 62, 194, 49 },
		{ 62, 195, 49 },
		{ 62, 196, 49 },
		{ 62, 197, 49 },
		{ 62, 198, 49 },
		{ 62, 199, 49 },
		{ 62, 200, 49 },
		{ 62, 201, 49 },
		{ 62, 202, 49 },
		{ 62, 203, 49 },
		{ 62, 204, 49 },
		{ 62, 205, 49 },
		{ 62, 206, 49 },
		{ 62, 207, 49 },
		{ 62, 208, 49 },
		{ 63, 161, 49 },
		{ 63, 162, 49 },
		{ 63, 163, 49 },
		{ 63, 164, 49 },
		{ 63, 165, 49 },
		{ 63, 166, 49 },
		{ 63, 167, 49 },
		{ 63, 168, 49 },
		{ 63, 169, 49 },
		{ 63, 170, 49 },
		{ 63, 171, 49 },
		{ 63, 172, 49 },
		{ 63, 173, 49 },
		{ 63, 174, 49 },
		{ 63, 175, 49 },
		{ 63, 176, 49 },
		{ 63, 177, 49 },
		{ 63, 178, 49 },
		{ 63, 179, 49 },
		{ 63, 180, 49 },
		{ 63, 181, 49 },
		{ 63, 182, 49 },
		{ 63, 183, 49 },
		{ 63, 184, 49 },
		{ 63, 185, 49 },
		{ 63, 186, 49 },
		{ 63, 187, 49 },
		{ 63, 188, 49 },
		{ 63, 189, 49 },
		{ 63, 190, 49 },
		{ 63, 191, 49 },
		{ 63, 192, 49 },
		{ 63, 193, 49 },
		{ 63, 194, 49 },
		{ 63, 195, 49 },
		{ 63, 196, 49 },
		{ 63, 197, 49 },
		{ 63, 198, 49 },
		{ 63, 199, 49 },
		{ 63, 200, 49 },
		{ 63, 201, 49 },
		{ 63, 202, 49 },
		{ 63, 203, 49 },
		{ 63, 204, 49 },
		{ 63, 205, 49 },
		{ 63, 206, 49 },
		{ 63, 207, 49 },
		{ 63, 208, 49 },
		{ 64, 161, 49 },
		{ 64, 162, 49 },
		{ 64, 163, 49 },
		{ 64, 164, 49 },
		{ 64, 165, 49 },
		{ 64, 166, 49 },
		{ 64, 167, 49 },
		{ 64, 168, 49 },
		{ 64, 169, 49 },
		{ 64, 170, 49 },
		{ 64, 171, 49 },
		{ 64, 172, 49 },
		{ 64, 173, 49 },
		{ 64, 174, 49 },
		{ 64, 175, 49 },
		{ 64, 176, 49 },
		{ 64, 177, 49 },
		{ 64, 178, 49 },
		{ 64, 179, 49 },
		{ 64, 180, 49 },
		{ 64, 181, 49 },
		{ 64, 182, 49 },
		{ 64, 183, 49 },
		{ 64, 184, 49 },
		{ 64, 185, 49 },
		{ 64, 186, 49 },
		{ 64, 187, 49 },
		{ 64, 188, 49 },
		{ 64, 189, 49 },
		{ 64, 190, 49 },
		{ 64, 191, 49 },
		{ 64, 192, 49 },
		{ 64, 193, 49 },
		{ 64, 194, 49 },
		{ 64, 195, 49 },
		{ 64, 196, 49 },
		{ 64, 197, 49 },
		{ 64, 198, 49 },
		{ 64, 199, 49 },
		{ 64, 200, 49 },
		{ 64, 201, 49 },
		{ 64, 202, 49 },
		{ 64, 203, 49 },
		{ 64, 204, 49 },
		{ 64, 205, 49 },
		{ 64, 206, 49 },
		{ 64, 207, 49 },
		{ 64, 208, 49 },
		{ 65, 162, 49 },
		{ 65, 163, 49 },
		{ 65, 164, 49 },
		{ 65, 165, 49 },
		{ 65, 166, 49 },
		{ 65, 167, 49 },
		{ 65, 168, 49 },
		{ 65, 169, 49 },
		{ 65, 170, 49 },
		{ 65, 171, 49 },
		{ 65, 172, 49 },
		{ 65, 173, 49 },
		{ 65, 174, 49 },
		{ 65, 175, 49 },
		{ 65, 176, 49 },
		{ 65, 177, 49 },
		{ 65, 178, 49 },
		{ 65, 179, 49 },
		{ 65, 180, 49 },
		{ 65, 181, 49 },
		{ 65, 182, 49 },
		{ 65, 183, 49 },
		{ 65, 184, 49 },
		{ 65, 185, 49 },
		{ 65, 186, 49 },
		{ 65, 187, 49 },
		{ 65, 188, 49 },
		{ 65, 189, 49 },
		{ 65, 190, 49 },
		{ 65, 191, 49 },
		{ 65, 192, 49 },
		{ 65, 193, 49 },
		{ 65, 194, 49 },
		{ 65, 195, 49 },
		{ 65, 196, 49 },
		{ 65, 197, 49 },
		{ 65, 198, 49 },
		{ 65, 199, 49 },
		{ 65, 200, 49 },
		{ 65, 201, 49 },
		{ 65, 202, 49 },
		{ 65, 203, 49 },
		{ 65, 204, 49 },
		{ 65, 205, 49 },
		{ 65, 206, 49 },
		{ 65, 207, 49 },
		{ 65, 208, 49 },
		{ 66, 162, 49 },
		{ 66, 163, 49 },
		{ 66, 164, 49 },
		{ 66, 165, 49 },
		{ 66, 166, 49 },
		{ 66, 167, 49 },
		{ 66, 168, 49 },
		{ 66, 169, 49 },
		{ 66, 170, 49 },
		{ 66, 171, 49 },
		{ 66, 172, 49 },
		{ 66, 173, 49 },
		{ 66, 174, 49 },
		{ 66, 175, 49 },
		{ 66, 176, 49 },
		{ 66, 177, 49 },
		{ 66, 178, 49 },
		{ 66, 179, 49 },
		{ 66, 180, 49 },
		{ 66, 181, 49 },
		{ 66, 182, 49 },
		{ 66, 183, 49 },
		{ 66, 184, 49 },
		{ 66, 185, 49 },
		{ 66, 186, 49 },
		{ 66, 187, 49 },
		{ 66, 188, 49 },
		{ 66, 189, 49 },
		{ 66, 190, 49 },
		{ 66, 191, 49 },
		{ 66, 192, 49 },
		{ 66, 193, 49 },
		{ 66, 194, 49 },
		{ 66, 195, 49 },
		{ 66, 196, 49 },
		{ 66, 197, 49 },
		{ 66, 198, 49 },
		{ 66, 199, 49 },
		{ 66, 200, 49 },
		{ 66, 201, 49 },
		{ 66, 202, 49 },
		{ 66, 203, 49 },
		{ 66, 204, 49 },
		{ 66, 205, 49 },
		{ 66, 206, 49 },
		{ 66, 207, 49 },
		{ 66, 208, 49 },
		{ 66, 209, 49 },
		{ 67, 163, 49 },
		{ 67, 164, 49 },
		{ 67, 165, 49 },
		{ 67, 166, 49 },
		{ 67, 167, 49 },
		{ 67, 168, 49 },
		{ 67, 169, 49 },
		{ 67, 170, 49 },
		{ 67, 171, 49 },
		{ 67, 172, 49 },
		{ 67, 173, 49 },
		{ 67, 174, 49 },
		{ 67, 175, 49 },
		{ 67, 176, 49 },
		{ 67, 177, 49 },
		{ 67, 178, 49 },
		{ 67, 179, 49 },
		{ 67, 180, 49 },
		{ 67, 181, 49 },
		{ 67, 182, 49 },
		{ 67, 183, 49 },
		{ 67, 184, 49 },
		{ 67, 185, 49 },
		{ 67, 186, 49 },
		{ 67, 187, 49 },
		{ 67, 188, 49 },
		{ 67, 189, 49 },
		{ 67, 190, 49 },
		{ 67, 191, 49 },
		{ 67, 192, 49 },
		{ 67, 193, 49 },
		{ 67, 194, 49 },
		{ 67, 195, 49 },
		{ 67, 196, 49 },
		{ 67, 197, 49 },
		{ 67, 198, 49 },
		{ 67, 199, 49 },
		{ 67, 200, 49 },
		{ 67, 201, 49 },
		{ 67, 202, 49 },
		{ 67, 203, 49 },
		{ 67, 204, 49 },
		{ 67, 205, 49 },
		{ 67, 206, 49 },
		{ 67, 207, 49 },
		{ 67, 208, 49 },
		{ 67, 209, 49 },
		{ 68, 163, 49 },
		{ 68, 164, 49 },
		{ 68, 165, 49 },
		{ 68, 166, 49 },
		{ 68, 167, 49 },
		{ 68, 168, 49 },
		{ 68, 169, 49 },
		{ 68, 170, 49 },
		{ 68, 171, 49 },
		{ 68, 172, 49 },
		{ 68, 173, 49 },
		{ 68, 174, 49 },
		{ 68, 175, 49 },
		{ 68, 176, 49 },
		{ 68, 177, 49 },
		{ 68, 178, 49 },
		{ 68, 179, 49 },
		{ 68, 180, 49 },
		{ 68, 181, 49 },
		{ 68, 182, 49 },
		{ 68, 183, 49 },
		{ 68, 184, 49 },
		{ 68, 185, 49 },
		{ 68, 186, 49 },
		{ 68, 187, 49 },
		{ 68, 188, 49 },
		{ 68, 189, 49 },
		{ 68, 190, 49 },
		{ 68, 191, 49 },
		{ 68, 192, 49 },
		{ 68, 193, 49 },
		{ 68, 194, 49 },
		{ 68, 195, 49 },
		{ 68, 196, 49 },
		{ 68, 197, 49 },
		{ 68, 198, 49 },
		{ 68, 199, 49 },
		{ 68, 200, 49 },
		{ 68, 201, 49 },
		{ 68, 202, 49 },
		{ 68, 203, 49 },
		{ 68, 204, 49 },
		{ 68, 205, 49 },
		{ 68, 206, 49 },
		{ 68, 207, 49 },
		{ 68, 208, 49 },
		{ 68, 209, 49 },
		{ 69, 164, 49 },
		{ 69, 165, 49 },
		{ 69, 166, 49 },
		{ 69, 167, 49 },
		{ 69, 168, 49 },
		{ 69, 169, 49 },
		{ 69, 170, 49 },
		{ 69, 171, 49 },
		{ 69, 172, 49 },
		{ 69, 173, 49 },
		{ 69, 174, 49 },
		{ 69, 175, 49 },
		{ 69, 176, 49 },
		{ 69, 177, 49 },
		{ 69, 178, 49 },
		{ 69, 179, 49 },
		{ 69, 180, 49 },
		{ 69, 181, 49 },
		{ 69, 182, 49 },
		{ 69, 183, 49 },
		{ 69, 184, 49 },
		{ 69, 185, 49 },
		{ 69, 186, 49 },
		{ 69, 187, 49 },
		{ 69, 188, 49 },
		{ 69, 189, 49 },
		{ 69, 190, 49 },
		{ 69, 191, 49 },
		{ 69, 192, 49 },
		{ 69, 193, 49 },
		{ 69, 194, 49 },
		{ 69, 195, 49 },
		{ 69, 196, 49 },
		{ 69, 197, 49 },
		{ 69, 198, 49 },
		{ 69, 199, 49 },
		{ 69, 200, 49 },
		{ 69, 201, 49 },
		{ 69, 202, 49 },
		{ 69, 203, 49 },
		{ 69, 204, 49 },
		{ 69, 205, 49 },
		{ 69, 206, 49 },
		{ 69, 207, 49 },
		{ 69, 208, 49 },
		{ 69, 209, 49 },
		{ 70, 164, 49 },
		{ 70, 165, 49 },
		{ 70, 166, 49 },
		{ 70, 167, 49 },
		{ 70, 168, 49 },
		{ 70, 169, 49 },
		{ 70, 170, 49 },
		{ 70, 171, 49 },
		{ 70, 172, 49 },
		{ 70, 173, 49 },
		{ 70, 174, 49 },
		{ 70, 175, 49 },
		{ 70, 176, 49 },
		{ 70, 177, 49 },
		{ 70, 178, 49 },
		{ 70, 179, 49 },
		{ 70, 180, 49 },
		{ 70, 181, 49 },
		{ 70, 182, 49 },
		{ 70, 183, 49 },
		{ 70, 184, 49 },
		{ 70, 185, 49 },
		{ 70, 186, 49 },
		{ 70, 187, 49 },
		{ 70, 188, 49 },
		{ 70, 189, 49 },
		{ 70, 190, 49 },
		{ 70, 191, 49 },
		{ 70, 192, 49 },
		{ 70, 193, 49 },
		{ 70, 194, 49 },
		{ 70, 195, 49 },
		{ 70, 196, 49 },
		{ 70, 197, 49 },
		{ 70, 198, 49 },
		{ 70, 199, 49 },
		{ 70, 200, 49 },
		{ 70, 201, 49 },
		{ 70, 202, 49 },
		{ 70, 203, 49 },
		{ 70, 204, 49 },
		{ 70, 205, 49 },
		{ 70, 206, 49 },
		{ 70, 207, 49 },
		{ 70, 208, 49 },
		{ 70, 209, 49 },
		{ 70, 210, 49 },
		{ 71, 165, 49 },
		{ 71, 166, 49 },
		{ 71, 167, 49 },
		{ 71, 168, 49 },
		{ 71, 169, 49 },
		{ 71, 170, 49 },
		{ 71, 171, 49 },
		{ 71, 172, 49 },
		{ 71, 173, 49 },
		{ 71, 174, 49 },
		{ 71, 175, 49 },
		{ 71, 176, 49 },
		{ 71, 177, 49 },
		{ 71, 178, 49 },
		{ 71, 179, 49 },
		{ 71, 180, 49 },
		{ 71, 181, 49 },
		{ 71, 182, 49 },
		{ 71, 183, 49 },
		{ 71, 184, 49 },
		{ 71, 185, 49 },
		{ 71, 186, 49 },
		{ 71, 187, 49 },
		{ 71, 188, 49 },
		{ 71, 189, 49 },
		{ 71, 190, 49 },
		{ 71, 191, 49 },
		{ 71, 192, 49 },
		{ 71, 193, 49 },
		{ 71, 194, 49 },
		{ 71, 195, 49 },
		{ 71, 196, 49 },
		{ 71, 197, 49 },
		{ 71, 198, 49 },
		{ 71, 199, 49 },
		{ 71, 200, 49 },
		{ 71, 201, 49 },
		{ 71, 202, 49 },
		{ 71, 203, 49 },
		{ 71, 204, 49 },
		{ 71, 205, 49 },
		{ 71, 206, 49 },
		{ 71, 207, 49 },
		{ 71, 208, 49 },
		{ 71, 209, 49 },
		{ 71, 210, 49 },
		{ 72, 165, 49 },
		{ 72, 166, 49 },
		{ 72, 167, 49 },
		{ 72, 168, 49 },
		{ 72, 169, 49 },
		{ 72, 170, 49 },
		{ 72, 171, 49 },
		{ 72, 172, 49 },
		{ 72, 173, 49 },
		{ 72, 174, 49 },
		{ 72, 175, 49 },
		{ 72, 176, 49 },
		{ 72, 177, 49 },
		{ 72, 178, 49 },
		{ 72, 179, 49 },
		{ 72, 180, 49 },
		{ 72, 181, 49 },
		{ 72, 182, 49 },
		{ 72, 183, 49 },
		{ 72, 184, 49 },
		{ 72, 185, 49 },
		{ 72, 186, 49 },
		{ 72, 187, 49 },
		{ 72, 188, 49 },
		{ 72, 189, 49 },
		{ 72, 190, 49 },
		{ 72, 191, 49 },
		{ 72, 192, 49 },
		{ 72, 193, 49 },
		{ 72, 194, 49 },
		{ 72, 195, 49 },
		{ 72, 196, 49 },
		{ 72, 197, 49 },
		{ 72, 198, 49 },
		{ 72, 199, 49 },
		{ 72, 200, 49 },
		{ 72, 201, 49 },
		{ 72, 202, 49 },
		{ 72, 203, 49 },
		{ 72, 204, 49 },
		{ 72, 205, 49 },
		{ 72, 206, 49 },
		{ 72, 207, 49 },
		{ 72, 208, 49 },
		{ 72, 209, 49 },
		{ 72, 210, 49 },
		{ 73, 166, 49 },
		{ 73, 167, 49 },
		{ 73, 168, 49 },
		{ 73, 169, 49 },
		{ 73, 170, 49 },
		{ 73, 171, 49 },
		{ 73, 172, 49 },
		{ 73, 173, 49 },
		{ 73, 174, 49 },
		{ 73, 175, 49 },
		{ 73, 176, 49 },
		{ 73, 177, 49 },
		{ 73, 178, 49 },
		{ 73, 179, 49 },
		{ 73, 180, 49 },
		{ 73, 181, 49 },
		{ 73, 182, 49 },
		{ 73, 183, 49 },
		{ 73, 184, 49 },
		{ 73, 185, 49 },
		{ 73, 186, 49 },
		{ 73, 187, 49 },
		{ 73, 188, 49 },
		{ 73, 189, 49 },
		{ 73, 190, 49 },
		{ 73, 191, 49 },
		{ 73, 192, 49 },
		{ 73, 193, 49 },
		{ 73, 194, 49 },
		{ 73, 195, 49 },
		{ 73, 196, 49 },
		{ 73, 197, 49 },
		{ 73, 198, 49 },
		{ 73, 199, 49 },
		{ 73, 200, 49 },
		{ 73, 201, 49 },
		{ 73, 202, 49 },
		{ 73, 203, 49 },
		{ 73, 204, 49 },
		{ 73, 205, 49 },
		{ 73, 206, 49 },
		{ 73, 207, 49 },
		{ 73, 208, 49 },
		{ 73, 209, 49 },
		{ 73, 210, 49 },
		{ 74, 166, 49 },
		{ 74, 167, 49 },
		{ 74, 168, 49 },
		{ 74, 169, 49 },
		{ 74, 170, 49 },
		{ 74, 171, 49 },
		{ 74, 172, 49 },
		{ 74, 173, 49 },
		{ 74, 174, 49 },
		{ 74, 175, 49 },
		{ 74, 176, 49 },
		{ 74, 177, 49 },
		{ 74, 178, 49 },
		{ 74, 179, 49 },
		{ 74, 180, 49 },
		{ 74, 181, 49 },
		{ 74, 182, 49 },
		{ 74, 183, 49 },
		{ 74, 184, 49 },
		{ 74, 185, 49 },
		{ 74, 186, 49 },
		{ 74, 187, 49 },
		{ 74, 188, 49 },
		{ 74, 189, 49 },
		{ 74, 190, 49 },
		{ 74, 191, 49 },
		{ 74, 192, 49 },
		{ 74, 193, 49 },
		{ 74, 194, 49 },
		{ 74, 195, 49 },
		{ 74, 196, 49 },
		{ 74, 197, 49 },
		{ 74, 198, 49 },
		{ 74, 199, 49 },
		{ 74, 200, 49 },
		{ 74, 201, 49 },
		{ 74, 202, 49 },
		{ 74, 203, 49 },
		{ 74, 204, 49 },
		{ 74, 205, 49 },
		{ 74, 206, 49 },
		{ 74, 207, 49 },
		{ 74, 208, 49 },
		{ 74, 209, 49 },
		{ 74, 210, 49 },
		{ 74, 211, 49 },
		{ 75, 167, 49 },
		{ 75, 168, 49 },
		{ 75, 169, 49 },
		{ 75, 170, 49 },
		{ 75, 171, 49 },
		{ 75, 172, 49 },
		{ 75, 173, 49 },
		{ 75, 174, 49 },
		{ 75, 175, 49 },
		{ 75, 176, 49 },
		{ 75, 177, 49 },
		{ 75, 178, 49 },
		{ 75, 179, 49 },
		{ 75, 180, 49 },
		{ 75, 181, 49 },
		{ 75, 182, 49 },
		{ 75, 183, 49 },
		{ 75, 184, 49 },
		{ 75, 185, 49 },
		{ 75, 186, 49 },
		{ 75, 187, 49 },
		{ 75, 188, 49 },
		{ 75, 189, 49 },
		{ 75, 190, 49 },
		{ 75, 191, 49 },
		{ 75, 192, 49 },
		{ 75, 193, 49 },
		{ 75, 194, 49 },
		{ 75, 195, 49 },
		{ 75, 196, 49 },
		{ 75, 197, 49 },
		{ 75, 198, 49 },
		{ 75, 199, 49 },
		{ 75, 200, 49 },
		{ 75, 201, 49 },
		{ 75, 202, 49 },
		{ 75, 203, 49 },
		{ 75, 204, 49 },
		{ 75, 205, 49 },
		{ 75, 206, 49 },
		{ 75, 207, 49 },
		{ 75, 208, 49 },
		{ 75, 209, 49 },
		{ 75, 210, 49 },
		{ 75, 211, 49 },
		{ 76, 167, 49 },
		{ 76, 168, 49 },
		{ 76, 169, 49 },
		{ 76, 170, 49 },
		{ 76, 171, 49 },
		{ 76, 172, 49 },
		{ 76, 173, 49 },
		{ 76, 174, 49 },
		{ 76, 175, 49 },
		{ 76, 176, 49 },
		{ 76, 177, 49 },
		{ 76, 178, 49 },
		{ 76, 179, 49 },
		{ 76, 180, 49 },
		{ 76, 181, 49 },
		{ 76, 182, 49 },
		{ 76, 183, 49 },
		{ 76, 184, 49 },
		{ 76, 185, 49 },
		{ 76, 186, 49 },
		{ 76, 187, 49 },
		{ 76, 188, 49 },
		{ 76, 189, 49 },
		{ 76, 190, 49 },
		{ 76, 191, 49 },
		{ 76, 192, 49 },
		{ 76, 193, 49 },
		{ 76, 194, 49 },
		{ 76, 195, 49 },
		{ 76, 196, 49 },
		{ 76, 197, 49 },
		{ 76, 198, 49 },
		{ 76, 199, 49 },
		{ 76, 200, 49 },
		{ 76, 201, 49 },
		{ 76, 202, 49 },
		{ 76, 203, 49 },
		{ 76, 204, 49 },
		{ 76, 205, 49 },
		{ 76, 206, 49 },
		{ 76, 207, 49 },
		{ 76, 208, 49 },
		{ 76, 209, 49 },
		{ 76, 210, 49 },
		{ 76, 211, 49 },
		{ 77, 168, 49 },
		{ 77, 169, 49 },
		{ 77, 170, 49 },
		{ 77, 171, 49 },
		{ 77, 172, 49 },
		{ 77, 173, 49 },
		{ 77, 174, 49 },
		{ 77, 175, 49 },
		{ 77, 176, 49 },
		{ 77, 177, 49 },
		{ 77, 178, 49 },
		{ 77, 179, 49 },
		{ 77, 180, 49 },
		{ 77, 181, 49 },
		{ 77, 182, 49 },
		{ 77, 183, 49 },
		{ 77, 184, 49 },
		{ 77, 185, 49 },
		{ 77, 186, 49 },
		{ 77, 187, 49 },
		{ 77, 188, 49 },
		{ 77, 189, 49 },
		{ 77, 190, 49 },
		{ 77, 191, 49 },
		{ 77, 192, 49 },
		{ 77, 193, 49 },
		{ 77, 194, 49 },
		{ 77, 195, 49 },
		{ 77, 196, 49 },
		{ 77, 197, 49 },
		{ 77, 198, 49 },
		{ 77, 199, 49 },
		{ 77, 200, 49 },
		{ 77, 201, 49 },
		{ 77, 202, 49 },
		{ 77, 203, 49 },
		{ 77, 204, 49 },
		{ 77, 205, 49 },
		{ 77, 206, 49 },
		{ 77, 207, 49 },
		{ 77, 208, 49 },
		{ 77, 209, 49 },
		{ 77, 210, 49 },
		{ 77, 211, 49 },
		{ 78, 168, 49 },
		{ 78, 169, 49 },
		{ 78, 170, 49 },
		{ 78, 171, 49 },
		{ 78, 172, 49 },
		{ 78, 173, 49 },
		{ 78, 174, 49 },
		{ 78, 175, 49 },
		{ 78, 176, 49 },
		{ 78, 177, 49 },
		{ 78, 178, 49 },
		{ 78, 179, 49 },
		{ 78, 180, 49 },
		{ 78, 181, 49 },
		{ 78, 182, 49 },
		{ 78, 183, 49 },
		{ 78, 184, 49 },
		{ 78, 185, 49 },
		{ 78, 186, 49 },
		{ 78, 187, 49 },
		{ 78, 188, 49 },
		{ 78, 189, 49 },
		{ 78, 190, 49 },
		{ 78, 191, 49 },
		{ 78, 192, 49 },
		{ 78, 193, 49 },
		{ 78, 194, 49 },
		{ 78, 195, 49 },
		{ 78, 196, 49 },
		{ 78, 197, 49 },
		{ 78, 198, 49 },
		{ 78, 199, 49 },
		{ 78, 200, 49 },
		{ 78, 201, 49 },
		{ 78, 202, 49 },
		{ 78, 203, 49 },
		{ 78, 204, 49 },
		{ 78, 205, 49 },
		{ 78, 206, 49 },
		{ 78, 207, 49 },
		{ 78, 208, 49 },
		{ 78, 209, 49 },
		{ 78, 210, 49 },
		{ 78, 211, 49 },
		{ 78, 212, 49 },
		{ 79, 169, 49 },
		{ 79, 170, 49 },
		{ 79, 171, 49 },
		{ 79, 172, 49 },
		{ 79, 173, 49 },
		{ 79, 174, 49 },
		{ 79, 175, 49 },
		{ 79, 176, 49 },
		{ 79, 177, 49 },
		{ 79, 178, 49 },
		{ 79, 179, 49 },
		{ 79, 180, 49 },
		{ 79, 181, 49 },
		{ 79, 182, 49 },
		{ 79, 183, 49 },
		{ 79, 184, 49 },
		{ 79, 185, 49 },
		{ 79, 186, 49 },
		{ 79, 187, 49 },
		{ 79, 188, 49 },
		{ 79, 189, 49 },
		{ 79, 190, 49 },
		{ 79, 191, 49 },
		{ 79, 192, 49 },
		{ 79, 193, 49 },
		{ 79, 194, 49 },
		{ 79, 195, 49 },
		{ 79, 196, 49 },
		{ 79, 197, 49 },
		{ 79, 198, 49 },
		{ 79, 199, 49 },
		{ 79, 200, 49 },
		{ 79, 201, 49 },
		{ 79, 202, 49 },
		{ 79, 203, 49 },
		{ 79, 204, 49 },
		{ 79, 205, 49 },
		{ 79, 206, 49 },
		{ 79, 207, 49 },
		{ 79, 208, 49 },
		{ 79, 209, 49 },
		{ 79, 210, 49 },
		{ 79, 211, 49 },
		{ 79, 212, 49 },
		{ 80, 169, 49 },
		{ 80, 170, 49 },
		{ 80, 171, 49 },
		{ 80, 172, 49 },
		{ 80, 173, 49 },
		{ 80, 174, 49 },
		{ 80, 175, 49 },
		{ 80, 176, 49 },
		{ 80, 177, 49 },
		{ 80, 178, 49 },
		{ 80, 179, 49 },
		{ 80, 180, 49 },
		{ 80, 181, 49 },
		{ 80, 182, 49 },
		{ 80, 183, 49 },
		{ 80, 184, 49 },
		{ 80, 185, 49 },
		{ 80, 186, 49 },
		{ 80, 187, 49 },
		{ 80, 188, 49 },
		{ 80, 189, 49 },
		{ 80, 190, 49 },
		{ 80, 191, 49 },
		{ 80, 192, 49 },
		{ 80, 193, 49 },
		{ 80, 194, 49 },
		{ 80, 195, 49 },
		{ 80, 196, 49 },
		{ 80, 197, 49 },
		{ 80, 198, 49 },
		{ 80, 199, 49 },
		{ 80, 200, 49 },
		{ 80, 201, 49 },
		{ 80, 202, 49 },
		{ 80, 203, 49 },
		{ 80, 204, 49 },
		{ 80, 205, 49 },
		{ 80, 206, 49 },
		{ 80, 207, 49 },
		{ 80, 208, 49 },
		{ 80, 209, 49 },
		{ 80, 210, 49 },
		{ 80, 211, 49 },
		{ 80, 212, 49 },
		{ 81, 170, 49 },
		{ 81, 171, 49 },
		{ 81, 172, 49 },
		{ 81, 173, 49 },
		{ 81, 174, 49 },
		{ 81, 175, 49 },
		{ 81, 176, 49 },
		{ 81, 177, 49 },
		{ 81, 178, 49 },
		{ 81, 179, 49 },
		{ 81, 180, 49 },
		{ 81, 181, 49 },
		{ 81, 182, 49 },
		{ 81, 183, 49 },
		{ 81, 184, 49 },
		{ 81, 185, 49 },
		{ 81, 186, 49 },
		{ 81, 187, 49 },
		{ 81, 188, 49 },
		{ 81, 189, 49 },
		{ 81, 190, 49 },
		{ 81, 191, 49 },
		{ 81, 192, 49 },
		{ 81, 193, 49 },
		{ 81, 194, 49 },
		{ 81, 195, 49 },
		{ 81, 196, 49 },
		{ 81, 197, 49 },
		{ 81, 198, 49 },
		{ 81, 199, 49 },
		{ 81, 200, 49 },
		{ 81, 201, 49 },
		{ 81, 202, 49 },
		{ 81, 203, 49 },
		{ 81, 204, 49 },
		{ 81, 205, 49 },
		{ 81, 206, 49 },
		{ 81, 207, 49 },
		{ 81, 208, 49 },
		{ 81, 209, 49 },
		{ 81, 210, 49 },
		{ 81, 211, 49 },
		{ 81, 212, 49 },
		{ 82, 170, 49 },
		{ 82, 171, 49 },
		{ 82, 172, 49 },
		{ 82, 173, 49 },
		{ 82, 174, 49 },
		{ 82, 175, 49 },
		{ 82, 176, 49 },
		{ 82, 177, 49 },
		{ 82, 178, 49 },
		{ 82, 179, 49 },
		{ 82, 180, 49 },
		{ 82, 181, 49 },
		{ 82, 182, 49 },
		{ 82, 183, 49 },
		{ 82, 184, 49 },
		{ 82, 185, 49 },
		{ 82, 186, 49 },
		{ 82, 187, 49 },
		{ 82, 188, 49 },
		{ 82, 189, 49 },
		{ 82, 190, 49 },
		{ 82, 191, 49 },
		{ 82, 192, 49 },
		{ 82, 193, 49 },
		{ 82, 194, 49 },
		{ 82, 195, 49 },
		{ 82, 196, 49 },
		{ 82, 197, 49 },
		{ 82, 198, 49 },
		{ 82, 199, 49 },
		{ 82, 200, 49 },
		{ 82, 201, 49 },
		{ 82, 202, 49 },
		{ 82, 203, 49 },
		{ 82, 204, 49 },
		{ 82, 205, 49 },
		{ 82, 206, 49 },
		{ 82, 207, 49 },
		{ 82, 208, 49 },
		{ 82, 209, 49 },
		{ 82, 210, 49 },
		{ 82, 211, 49 },
		{ 82, 212, 49 },
		{ 82, 213, 49 },
		{ 83, 171, 49 },
		{ 83, 172, 49 },
		{ 83, 173, 49 },
		{ 83, 174, 49 },
		{ 83, 175, 49 },
		{ 83, 176, 49 },
		{ 83, 177, 49 },
		{ 83, 178, 49 },
		{ 83, 179, 49 },
		{ 83, 180, 49 },
		{ 83, 181, 49 },
		{ 83, 182, 49 },
		{ 83, 183, 49 },
		{ 83, 184, 49 },
		{ 83, 185, 49 },
		{ 83, 186, 49 },
		{ 83, 187, 49 },
		{ 83, 188, 49 },
		{ 83, 189, 49 },
		{ 83, 190, 49 },
		{ 83, 191, 49 },
		{ 83, 192, 49 },
		{ 83, 193, 49 },
		{ 83, 194, 49 },
		{ 83, 195, 49 },
		{ 83, 196, 49 },
		{ 83, 197, 49 },
		{ 83, 198, 49 },
		{ 83, 199, 49 },
		{ 83, 200, 49 },
		{ 83, 201, 49 },
		{ 83, 202, 49 },
		{ 83, 203, 49 },
		{ 83, 204, 49 },
		{ 83, 205, 49 },
		{ 83, 206, 49 },
		{ 83, 207, 49 },
		{ 83, 208, 49 },
		{ 83, 209, 49 },
		{ 83, 210, 49 },
		{ 83, 211, 49 },
		{ 83, 212, 49 },
		{ 83, 213, 49 },
		{ 84, 171, 49 },
		{ 84, 172, 49 },
		{ 84, 173, 49 },
		{ 84, 174, 49 },
		{ 84, 175, 49 },
		{ 84, 176, 49 },
		{ 84, 177, 49 },
		{ 84, 178, 49 },
		{ 84, 179, 49 },
		{ 84, 180, 49 },
		{ 84, 181, 49 },
		{ 84, 182, 49 },
		{ 84, 183, 49 },
		{ 84, 184, 49 },
		{ 84, 185, 49 },
		{ 84, 186, 49 },
		{ 84, 187, 49 },
		{ 84, 188, 49 },
		{ 84, 189, 49 },
		{ 84, 190, 49 },
		{ 84, 191, 49 },
		{ 84, 192, 49 },
		{ 84, 193, 49 },
		{ 84, 194, 49 },
		{ 84, 195, 49 },
		{ 84, 196, 49 },
		{ 84, 197, 49 },
		{ 84, 198, 49 },
		{ 84, 199, 49 },
		{ 84, 200, 49 },
		{ 84, 201, 49 },
		{ 84, 202, 49 },
		{ 84, 203, 49 },
		{ 84, 204, 49 },
		{ 84, 205, 49 },
		{ 84, 206, 49 },
		{ 84, 207, 49 },
		{ 84, 208, 49 },
		{ 84, 209, 49 },
		{ 84, 210, 49 },
		{ 84, 211, 49 },
		{ 84, 212, 49 },
		{ 84, 213, 49 },
		{ 85, 172, 49 },
		{ 85, 173, 49 },
		{ 85, 174, 49 },
		{ 85, 175, 49 },
		{ 85, 176, 49 },
		{ 85, 177, 49 },
		{ 85, 178, 49 },
		{ 85, 179, 49 },
		{ 85, 180, 49 },
		{ 85, 181, 49 },
		{ 85, 182, 49 },
		{ 85, 183, 49 },
		{ 85, 184, 49 },
		{ 85, 185, 49 },
		{ 85, 186, 49 },
		{ 85, 187, 49 },
		{ 85, 188, 49 },
		{ 85, 189, 49 },
		{ 85, 190, 49 },
		{ 85, 191, 49 },
		{ 85, 192, 49 },
		{ 85, 193, 49 },
		{ 85, 194, 49 },
		{ 85, 195, 49 },
		{ 85, 196, 49 },
		{ 85, 197, 49 },
		{ 85, 198, 49 },
		{ 85, 199, 49 },
		{ 85, 200, 49 },
		{ 85, 201, 49 },
		{ 85, 202, 49 },
		{ 85, 203, 49 },
		{ 85, 204, 49 },
		{ 85, 205, 49 },
		{ 85, 206, 49 },
		{ 85, 207, 49 },
		{ 85, 208, 49 },
		{ 85, 209, 49 },
		{ 85, 210, 49 },
		{ 85, 211, 49 },
		{ 85, 212, 49 },
		{ 85, 213, 49 },
		{ 86, 172, 49 },
		{ 86, 173, 49 },
		{ 86, 174, 49 },
		{ 86, 175, 49 },
		{ 86, 176, 49 },
		{ 86, 177, 49 },
		{ 86, 178, 49 },
		{ 86, 179, 49 },
		{ 86, 180, 49 },
		{ 86, 181, 49 },
		{ 86, 182, 49 },
		{ 86, 183, 49 },
		{ 86, 184, 49 },
		{ 86, 185, 49 },
		{ 86, 186, 49 },
		{ 86, 187, 49 },
		{ 86, 188, 49 },
		{ 86, 189, 49 },
		{ 86, 190, 49 },
		{ 86, 191, 49 },
		{ 86, 192, 49 },
		{ 86, 193, 49 },
		{ 86, 194, 49 },
		{ 86, 195, 49 },
		{ 86, 196, 49 },
		{ 86, 197, 49 },
		{ 86, 198, 49 },
		{ 86, 199, 49 },
		{ 86, 200, 49 },
		{ 86, 201, 49 },
		{ 86, 202, 49 },
		{ 86, 203, 49 },
		{ 86, 204, 49 },
		{ 86, 205, 49 },
		{ 86, 206, 49 },
		{ 86, 207, 49 },
		{ 86, 208, 49 },
		{ 86, 209, 49 },
		{ 86, 210, 49 },
		{ 86, 211, 49 },
		{ 86, 212, 49 },
		{ 86, 213, 49 },
		{ 86, 214, 49 },
		{ 87, 173, 49 },
		{ 87, 174, 49 },
		{ 87, 175, 49 },
		{ 87, 176, 49 },
		{ 87, 177, 49 },
		{ 87, 178, 49 },
		{ 87, 179, 49 },
		{ 87, 180, 49 },
		{ 87, 181, 49 },
		{ 87, 182, 49 },
		{ 87, 183, 49 },
		{ 87, 184, 49 },
		{ 87, 185, 49 },
		{ 87, 186, 49 },
		{ 87, 187, 49 },
		{ 87, 188, 49 },
		{ 87, 189, 49 },
		{ 87, 190, 49 },
		{ 87, 191, 49 },
		{ 87, 192, 49 },
		{ 87, 193, 49 },
		{ 87, 194, 49 },
		{ 87, 195, 49 },
		{ 87, 196, 49 },
		{ 87, 197, 49 },
		{ 87, 198, 49 },
		{ 87, 199, 49 },
		{ 87, 200, 49 },
		{ 87, 201, 49 },
		{ 87, 202, 49 },
		{ 87, 203, 49 },
		{ 87, 204, 49 },
		{ 87, 205, 49 },
		{ 87, 206, 49 },
		{ 87, 207, 49 },
		{ 87, 208, 49 },
		{ 87, 209, 49 },
		{ 87, 210, 49 },
		{ 87, 211, 49 },
		{ 87, 212, 49 },
		{ 87, 213, 49 },
		{ 87, 214, 49 },
		{ 88, 173, 49 },
		{ 88, 174, 49 },
		{ 88, 175, 49 },
		{ 88, 176, 49 },
		{ 88, 177, 49 },
		{ 88, 178, 49 },
		{ 88, 179, 49 },
		{ 88, 180, 49 },
		{ 88, 181, 49 },
		{ 88, 182, 49 },
		{ 88, 183, 49 },
		{ 88, 184, 49 },
		{ 88, 185, 49 },
		{ 88, 186, 49 },
		{ 88, 187, 49 },
		{ 88, 188, 49 },
		{ 88, 189, 49 },
		{ 88, 190, 49 },
		{ 88, 191, 49 },
		{ 88, 192, 49 },
		{ 88, 193, 49 },
		{ 88, 194, 49 },
		{ 88, 195, 49 },
		{ 88, 196, 49 },
		{ 88, 197, 49 },
		{ 88, 198, 49 },
		{ 88, 199, 49 },
		{ 88, 200, 49 },
		{ 88, 201, 49 },
		{ 88, 202, 49 },
		{ 88, 203, 49 },
		{ 88, 204, 49 },
		{ 88, 205, 49 },
		{ 88, 206, 49 },
		{ 88, 207, 49 },
		{ 88, 208, 49 },
		{ 88, 209, 49 },
		{ 88, 210, 49 },
		{ 88, 211, 49 },
		{ 88, 212, 49 },
		{ 88, 213, 49 },
		{ 88, 214, 49 },
		{ 89, 174, 49 },
		{ 89, 175, 49 },
		{ 89, 176, 49 },
		{ 89, 177, 49 },
		{ 89, 178, 49 },
		{ 89, 179, 49 },
		{ 89, 180, 49 },
		{ 89, 181, 49 },
		{ 89, 182, 49 },
		{ 89, 183, 49 },
		{ 89, 184, 49 },
		{ 89, 185, 49 },
		{ 89, 186, 49 },
		{ 89, 187, 49 },
		{ 89, 188, 49 },
		{ 89, 189, 49 },
		{ 89, 190, 49 },
		{ 89, 191, 49 },
		{ 89, 192, 49 },
		{ 89, 193, 49 },
		{ 89, 194, 49 },
		{ 89, 195, 49 },
		{ 89, 196, 49 },
		{ 89, 197, 49 },
		{ 89, 198, 49 },
		{ 89, 199, 49 },
		{ 89, 200, 49 },
		{ 89, 201, 49 },
		{ 89, 202, 49 },
		{ 89, 203, 49 },
		{ 89, 204, 49 },
		{ 89, 205, 49 },
		{ 89, 206, 49 },
		{ 89, 207, 49 },
		{ 89, 208, 49 },
		{ 89, 209, 49 },
		{ 89, 210, 49 },
		{ 89, 211, 49 },
		{ 89, 212, 49 },
		{ 89, 213, 49 },
		{ 89, 214, 49 },
		{ 90, 174, 49 },
		{ 90, 175, 49 },
		{ 90, 176, 49 },
		{ 90, 177, 49 },
		{ 90, 178, 49 },
		{ 90, 179, 49 },
		{ 90, 180, 49 },
		{ 90, 181, 49 },
		{ 90, 182, 49 },
		{ 90, 183, 49 },
		{ 90, 184, 49 },
		{ 90, 185, 49 },
		{ 90, 186, 49 },
		{ 90, 187, 49 },
		{ 90, 188, 49 },
		{ 90, 189, 49 },
		{ 90, 190, 49 },
		{ 90, 191, 49 },
		{ 90, 192, 49 },
		{ 90, 193, 49 },
		{ 90, 194, 49 },
		{ 90, 195, 49 },
		{ 90, 196, 49 },
		{ 90, 197, 49 },
		{ 90, 198, 49 },
		{ 90, 199, 49 },
		{ 90, 200, 49 },
		{ 90, 201, 49 },
		{ 90, 202, 49 },
		{ 90, 203, 49 },
		{ 90, 204, 49 },
		{ 90, 205, 49 },
		{ 90, 206, 49 },
		{ 90, 207, 49 },
		{ 90, 208, 49 },
		{ 90, 209, 49 },
		{ 90, 210, 49 },
		{ 90, 211, 49 },
		{ 90, 212, 49 },
		{ 90, 213, 49 },
		{ 90, 214, 49 },
		{ 90, 215, 49 },
		{ 91, 175, 49 },
		{ 91, 176, 49 },
		{ 91, 177, 49 },
		{ 91, 178, 49 },
		{ 91, 179, 49 },
		{ 91, 180, 49 },
		{ 91, 181, 49 },
		{ 91, 182, 49 },
		{ 91, 183, 49 },
		{ 91, 184, 49 },
		{ 91, 185, 49 },
		{ 91, 186, 49 },
		{ 91, 187, 49 },
		{ 91, 188, 49 },
		{ 91, 189, 49 },
		{ 91, 190, 49 },
		{ 91, 191, 49 },
		{ 91, 192, 49 },
		{ 91, 193, 49 },
		{ 91, 194, 49 },
		{ 91, 195, 49 },
		{ 91, 196, 49 },
		{ 91, 197, 49 },
		{ 91, 198, 49 },
		{ 91, 199, 49 },
		{ 91, 200, 49 },
		{ 91, 201, 49 },
		{ 91, 202, 49 },
		{ 91, 203, 49 },
		{ 91, 204, 49 },
		{ 91, 205, 49 },
		{ 91, 206, 49 },
		{ 91, 207, 49 },
		{ 91, 208, 49 },
		{ 91, 209, 49 },
		{ 91, 210, 49 },
		{ 91, 211, 49 },
		{ 91, 212, 49 },
		{ 91, 213, 49 },
		{ 91, 214, 49 },
		{ 91, 215, 49 },
		{ 92, 175, 49 },
		{ 92, 176, 49 },
		{ 92, 177, 49 },
		{ 92, 178, 49 },
		{ 92, 179, 49 },
		{ 92, 180, 49 },
		{ 92, 181, 49 },
		{ 92, 182, 49 },
		{ 92, 183, 49 },
		{ 92, 184, 49 },
		{ 92, 185, 49 },
		{ 92, 186, 49 },
		{ 92, 187, 49 },
		{ 92, 188, 49 },
		{ 92, 189, 49 },
		{ 92, 190, 49 },
		{ 92, 191, 49 },
		{ 92, 192, 49 },
		{ 92, 193, 49 },
		{ 92, 194, 49 },
		{ 92, 195, 49 },
		{ 92, 196, 49 },
		{ 92, 197, 49 },
		{ 92, 198, 49 },
		{ 92, 199, 49 },
		{ 92, 200, 49 },
		{ 92, 201, 49 },
		{ 92, 202, 49 },
		{ 92, 203, 49 },
		{ 92, 204, 49 },
		{ 92, 205, 49 },
		{ 92, 206, 49 },
		{ 92, 207, 49 },
		{ 92, 208, 49 },
		{ 92, 209, 49 },
		{ 92, 210, 49 },
		{ 92, 211, 49 },
		{ 92, 212, 49 },
		{ 92, 213, 49 },
		{ 92, 214, 49 },
		{ 92, 215, 49 },
		{ 93, 176, 49 },
		{ 93, 177, 49 },
		{ 93, 178, 49 },
		{ 93, 179, 49 },
		{ 93, 180, 49 },
		{ 93, 181, 49 },
		{ 93, 182, 49 },
		{ 93, 183, 49 },
		{ 93, 184, 49 },
		{ 93, 185, 49 },
		{ 93, 186, 49 },
		{ 93, 187, 49 },
		{ 93, 188, 49 },
		{ 93, 189, 49 },
		{ 93, 190, 49 },
		{ 93, 191, 49 },
		{ 93, 192, 49 },
		{ 93, 193, 49 },
		{ 93, 194, 49 },
		{ 93, 195, 49 },
		{ 93, 196, 49 },
		{ 93, 197, 49 },
		{ 93, 198, 49 },
		{ 93, 199, 49 },
		{ 93, 200, 49 },
		{ 93, 201, 49 },
		{ 93, 202, 49 },
		{ 93, 203, 49 },
		{ 93, 204, 49 },
		{ 93, 205, 49 },
		{ 93, 206, 49 },
		{ 93, 207, 49 },
		{ 93, 208, 49 },
		{ 93, 209, 49 },
		{ 93, 210, 49 },
		{ 93, 211, 49 },
		{ 93, 212, 49 },
		{ 93, 213, 49 },
		{ 93, 214, 49 },
		{ 93, 215, 49 },
		{ 94, 176, 49 },
		{ 94, 177, 49 },
		{ 94, 178, 49 },
		{ 94, 179, 49 },
		{ 94, 180, 49 },
		{ 94, 181, 49 },
		{ 94, 182, 49 },
		{ 94, 183, 49 },
		{ 94, 184, 49 },
		{ 94, 185, 49 },
		{ 94, 186, 49 },
		{ 94, 187, 49 },
		{ 94, 188, 49 },
		{ 94, 189, 49 },
		{ 94, 190, 49 },
		{ 94, 191, 49 },
		{ 94, 192, 49 },
		{ 94, 193, 49 },
		{ 94, 194, 49 },
		{ 94, 195, 49 },
		{ 94, 196, 49 },
		{ 94, 197, 49 },
		{ 94, 198, 49 },
		{ 94, 199, 49 },
		{ 94, 200, 49 },
		{ 94, 201, 49 },
		{ 94, 202, 49 },
		{ 94, 203, 49 },
		{ 94, 204, 49 },
		{ 94, 205, 49 },
		{ 94, 206, 49 },
		{ 94, 207, 49 },
		{ 94, 208, 49 },
		{ 94, 209, 49 },
		{ 94, 210, 49 },
		{ 94, 211, 49 },
		{ 94, 212, 49 },
		{ 94, 213, 49 },
		{ 94, 214, 49 },
		{ 94, 215, 49 },
		{ 94, 216, 49 },
		{ 95, 177, 49 },
		{ 95, 178, 49 },
		{ 95, 179, 49 },
		{ 95, 180, 49 },
		{ 95, 181, 49 },
		{ 95, 182, 49 },
		{ 95, 183, 49 },
		{ 95, 184, 49 },
		{ 95, 185, 49 },
		{ 95, 186, 49 },
		{ 95, 187, 49 },
		{ 95, 188, 49 },
		{ 95, 189, 49 },
		{ 95, 190, 49 },
		{ 95, 191, 49 },
		{ 95, 192, 49 },
		{ 95, 193, 49 },
		{ 95, 194, 49 },
		{ 95, 195, 49 },
		{ 95, 196, 49 },
		{ 95, 197, 49 },
		{ 95, 198, 49 },
		{ 95, 199, 49 },
		{ 95, 200, 49 },
		{ 95, 201, 49 },
		{ 95, 202, 49 },
		{ 95, 203, 49 },
		{ 95, 204, 49 },
		{ 95, 205, 49 },
		{ 95, 206, 49 },
		{ 95, 207, 49 },
		{ 95, 208, 49 },
		{ 95, 209, 49 },
		{ 95, 210, 49 },
		{ 95, 211, 49 },
		{ 95, 212, 49 },
		{ 95, 213, 49 },
		{ 95, 214, 49 },
		{ 95, 215, 49 },
		{ 95, 216, 49 },
		{ 96, 177, 49 },
		{ 96, 178, 49 },
		{ 96, 179, 49 },
		{ 96, 180, 49 },
		{ 96, 181, 49 },
		{ 96, 182, 49 },
		{ 96, 183, 49 },
		{ 96, 184, 49 },
		{ 96, 185, 49 },
		{ 96, 186, 49 },
		{ 96, 187, 49 },
		{ 96, 188, 49 },
		{ 96, 189, 49 },
		{ 96, 190, 49 },
		{ 96, 191, 49 },
		{ 96, 192, 49 },
		{ 96, 193, 49 },
		{ 96, 194, 49 },
		{ 96, 195, 49 },
		{ 96, 196, 49 },
		{ 96, 197, 49 },
		{ 96, 198, 49 },
		{ 96, 199, 49 },
		{ 96, 200, 49 },
		{ 96, 201, 49 },
		{ 96, 202, 49 },
		{ 96, 203, 49 },
		{ 96, 204, 49 },
		{ 96, 205, 49 },
		{ 96, 206, 49 },
		{ 96, 207, 49 },
		{ 96, 208, 49 },
		{ 96, 209, 49 },
		{ 96, 210, 49 },
		{ 96, 211, 49 },
		{ 96, 212, 49 },
		{ 96, 213, 49 },
		{ 96, 214, 49 },
		{ 96, 215, 49 },
		{ 96, 216, 49 },
		{ 97, 178, 49 },
		{ 97, 179, 49 },
		{ 97, 180, 49 },
		{ 97, 181, 49 },
		{ 97, 182, 49 },
		{ 97, 183, 49 },
		{ 97, 184, 49 },
		{ 97, 185, 49 },
		{ 97, 186, 49 },
		{ 97, 187, 49 },
		{ 97, 188, 49 },
		{ 97, 189, 49 },
		{ 97, 190, 49 },
		{ 97, 191, 49 },
		{ 97, 192, 49 },
		{ 97, 193, 49 },
		{ 97, 194, 49 },
		{ 97, 195, 49 },
		{ 97, 196, 49 },
		{ 97, 197, 49 },
		{ 97, 198, 49 },
		{ 97, 199, 49 },
		{ 97, 200, 49 },
		{ 97, 201, 49 },
		{ 97, 202, 49 },
		{ 97, 203, 49 },
		{ 97, 204, 49 },
		{ 97, 205, 49 },
		{ 97, 206, 49 },
		{ 97, 207, 49 },
		{ 97, 208, 49 },
		{ 97, 209, 49 },
		{ 97, 210, 49 },
		{ 97, 211, 49 },
		{ 97, 212, 49 },
		{ 97, 213, 49 },
		{ 97, 214, 49 },
		{ 97, 215, 49 },
		{ 97, 216, 49 },
		{ 98, 178, 49 },
		{ 98, 179, 49 },
		{ 98, 180, 49 },
		{ 98, 181, 49 },
		{ 98, 182, 49 },
		{ 98, 183, 49 },
		{ 98, 184, 49 },
		{ 98, 185, 49 },
		{ 98, 186, 49 },
		{ 98, 187, 49 },
		{ 98, 188, 49 },
		{ 98, 189, 49 },
		{ 98, 190, 49 },
		{ 98, 191, 49 },
		{ 98, 192, 49 },
		{ 98, 193, 49 },
		{ 98, 194, 49 },
		{ 98, 195, 49 },
		{ 98, 196, 49 },
		{ 98, 197, 49 },
		{ 98, 198, 49 },
		{ 98, 199, 49 },
		{ 98, 200, 49 },
		{ 98, 201, 49 },
		{ 98, 202, 49 },
		{ 98, 203, 49 },
		{ 98, 204, 49 },
		{ 98, 205, 49 },
		{ 98, 206, 49 },
		{ 98, 207, 49 },
		{ 98, 208, 49 },
		{ 98, 209, 49 },
		{ 98, 210, 49 },
		{ 98, 211, 49 },
		{ 98, 212, 49 },
		{ 98, 213, 49 },
		{ 98, 214, 49 },
		{ 98, 215, 49 },
		{ 98, 216, 49 },
		{ 98, 217, 49 },
		{ 99, 179, 49 },
		{ 99, 180, 49 },
		{ 99, 181, 49 },
		{ 99, 182, 49 },
		{ 99, 183, 49 },
		{ 99, 184, 49 },
		{ 99, 185, 49 },
		{ 99, 186, 49 },
		{ 99, 187, 49 },
		{ 99, 188, 49 },
		{ 99, 189, 49 },
		{ 99, 190, 49 },
		{ 99, 191, 49 },
		{ 99, 192, 49 },
		{ 99, 193, 49 },
		{ 99, 194, 49 },
		{ 99, 195, 49 },
		{ 99, 196, 49 },
		{ 99, 197, 49 },
		{ 99, 198, 49 },
		{ 99, 199, 49 },
		{ 99, 200, 49 },
		{ 99, 201, 49 },
		{ 99, 202, 49 },
		{ 99, 203, 49 },
		{ 99, 204, 49 },
		{ 99, 205, 49 },
		{ 99, 206, 49 },
		{ 99, 207, 49 },
		{ 99, 208, 49 },
		{ 99, 209, 49 },
		{ 99, 210, 49 },
		{ 99, 211, 49 },
		{ 99, 212, 49 },
		{ 99, 213, 49 },
		{ 99, 214, 49 },
		{ 99, 215, 49 },
		{ 99, 216, 49 },
		{ 99, 217, 49 },
		{ 100, 179, 49 },
		{ 100, 180, 49 },
		{ 100, 181, 49 },
		{ 100, 182, 49 },
		{ 100, 183, 49 },
		{ 100, 184, 49 },
		{ 100, 185, 49 },
		{ 100, 186, 49 },
		{ 100, 187, 49 },
		{ 100, 188, 49 },
		{ 100, 189, 49 },
		{ 100, 190, 49 },
		{ 100, 191, 49 },
		{ 100, 192, 49 },
		{ 100, 193, 49 },
		{ 100, 194, 49 },
		{ 100, 195, 49 },
		{ 100, 196, 49 },
		{ 100, 197, 49 },
		{ 100, 198, 49 },
		{ 100, 199, 49 },
		{ 100, 200, 49 },
		{ 100, 201, 49 },
		{ 100, 202, 49 },
		{ 100, 203, 49 },
		{ 100, 204, 49 },
		{ 100, 205, 49 },
		{ 100, 206, 49 },
		{ 100, 207, 49 },
		{ 100, 208, 49 },
		{ 100, 209, 49 },
		{ 100, 210, 49 },
		{ 100, 211, 49 },
		{ 100, 212, 49 },
		{ 100, 213, 49 },
		{ 100, 214, 49 },
		{ 100, 215, 49 },
		{ 100, 216, 49 },
		{ 100, 217, 49 },
		{ 101, 180, 49 },
		{ 101, 181, 49 },
		{ 101, 182, 49 },
		{ 101, 183, 49 },
		{ 101, 184, 49 },
		{ 101, 185, 49 },
		{ 101, 186, 49 },
		{ 101, 187, 49 },
		{ 101, 188, 49 },
		{ 101, 189, 49 },
		{ 101, 190, 49 },
		{ 101, 191, 49 },
		{ 101, 192, 49 },
		{ 101, 193, 49 },
		{ 101, 194, 49 },
		{ 101, 195, 49 },
		{ 101, 196, 49 },
		{ 101, 197, 49 },
		{ 101, 198, 49 },
		{ 101, 199, 49 },
		{ 101, 200, 49 },
		{ 101, 201, 49 },
		{ 101, 202, 49 },
		{ 101, 203, 49 },
		{ 101, 204, 49 },
		{ 101, 205, 49 },
		{ 101, 206, 49 },
		{ 101, 207, 49 },
		{ 101, 208, 49 },
		{ 101, 209, 49 },
		{ 101, 210, 49 },
		{ 101, 211, 49 },
		{ 101, 212, 49 },
		{ 101, 213, 49 },
		{ 101, 214, 49 },
		{ 101, 215, 49 },
		{ 101, 216, 49 },
		{ 101, 217, 49 },
		{ 102, 180, 49 },
		{ 102, 181, 49 },
		{ 102, 182, 49 },
		{ 102, 183, 49 },
		{ 102, 184, 49 },
		{ 102, 185, 49 },
		{ 102, 186, 49 },
		{ 102, 187, 49 },
		{ 102, 188, 49 },
		{ 102, 189, 49 },
		{ 102, 190, 49 },
		{ 102, 191, 49 },
		{ 102, 192, 49 },
		{ 102, 193, 49 },
		{ 102, 194, 49 },
		{ 102, 195, 49 },
		{ 102, 196, 49 },
		{ 102, 197, 49 },
		{ 102, 198, 49 },
		{ 102, 199, 49 },
		{ 102, 200, 49 },
		{ 102, 201, 49 },
		{ 102, 202, 49 },
		{ 102, 203, 49 },
		{ 102, 204, 49 },
		{ 102, 205, 49 },
		{ 102, 206, 49 },
		{ 102, 207, 49 },
		{ 102, 208, 49 },
		{ 102, 209, 49 },
		{ 102, 210, 49 },
		{ 102, 211, 49 },
		{ 102, 212, 49 },
		{ 102, 213, 49 },
		{ 102, 214, 49 },
		{ 102, 215, 49 },
		{ 102, 216, 49 },
		{ 102, 217, 49 },
		{ 102, 218, 49 },
		{ 103, 181, 49 },
		{ 103, 182, 49 },
		{ 103, 183, 49 },
		{ 103, 184, 49 },
		{ 103, 185, 49 },
		{ 103, 186, 49 },
		{ 103, 187, 49 },
		{ 103, 188, 49 },
		{ 103, 189, 49 },
		{ 103, 190, 49 },
		{ 103, 191, 49 },
		{ 103, 192, 49 },
		{ 103, 193, 49 },
		{ 103, 194, 49 },
		{ 103, 195, 49 },
		{ 103, 196, 49 },
		{ 103, 197, 49 },
		{ 103, 198, 49 },
		{ 103, 199, 49 },
		{ 103, 200, 49 },
		{ 103, 201, 49 },
		{ 103, 202, 49 },
		{ 103, 203, 49 },
		{ 103, 204, 49 },
		{ 103, 205, 49 },
		{ 103, 206, 49 },
		{ 103, 207, 49 },
		{ 103, 208, 49 },
		{ 103, 209, 49 },
		{ 103, 210, 49 },
		{ 103, 211, 49 },
		{ 103, 212, 49 },
		{ 103, 213, 49 },
		{ 103, 214, 49 },
		{ 103, 215, 49 },
		{ 103, 216, 49 },
		{ 103, 217, 49 },
		{ 103, 218, 49 },
		{ 104, 181, 49 },
		{ 104, 182, 49 },
		{ 104, 183, 49 },
		{ 104, 184, 49 },
		{ 104, 185, 49 },
		{ 104, 186, 49 },
		{ 104, 187, 49 },
		{ 104, 188, 49 },
		{ 104, 189, 49 },
		{ 104, 190, 49 },
		{ 104, 191, 49 },
		{ 104, 192, 49 },
		{ 104, 193, 49 },
		{ 104, 194, 49 },
		{ 104, 195, 49 },
		{ 104, 196, 49 },
		{ 104, 197, 49 },
		{ 104, 198, 49 },
		{ 104, 199, 49 },
		{ 104, 200, 49 },
		{ 104, 201, 49 },
		{ 104, 202, 49 },
		{ 104, 203, 49 },
		{ 104, 204, 49 },
		{ 104, 205, 49 },
		{ 104, 206, 49 },
		{ 104, 207, 49 },
		{ 104, 208, 49 },
		{ 104, 209, 49 },
		{ 104, 210, 49 },
		{ 104, 211, 49 },
		{ 104, 212, 49 },
		{ 104, 213, 49 },
		{ 104, 214, 49 },
		{ 104, 215, 49 },
		{ 104, 216, 49 },
		{ 104, 217, 49 },
		{ 104, 218, 49 },
		{ 105, 182, 49 },
		{ 105, 183, 49 },
		{ 105, 184, 49 },
		{ 105, 185, 49 },
		{ 105, 186, 49 },
		{ 105, 187, 49 },
		{ 105, 188, 49 },
		{ 105, 189, 49 },
		{ 105, 190, 49 },
		{ 105, 191, 49 },
		{ 105, 192, 49 },
		{ 105, 193, 49 },
		{ 105, 194, 49 },
		{ 105, 195, 49 },
		{ 105, 196, 49 },
		{ 105, 197, 49 },
		{ 105, 198, 49 },
		{ 105, 199, 49 },
		{ 105, 200, 49 },
		{ 105, 201, 49 },
		{ 105, 202, 49 },
		{ 105, 203, 49 },
		{ 105, 204, 49 },
		{ 105, 205, 49 },
		{ 105, 206, 49 },
		{ 105, 207, 49 },
		{ 105, 208, 49 },
		{ 105, 209, 49 },
		{ 105, 210, 49 },
		{ 105, 211, 49 },
		{ 105, 212, 49 },
		{ 105, 213, 49 },
		{ 105, 214, 49 },
		{ 105, 215, 49 },
		{ 105, 216, 49 },
		{ 105, 217, 49 },
		{ 105, 218, 49 },
		{ 106, 182, 49 },
		{ 106, 183, 49 },
		{ 106, 184, 49 },
		{ 106, 185, 49 },
		{ 106, 186, 49 },
		{ 106, 187, 49 },
		{ 106, 188, 49 },
		{ 106, 189, 49 },
		{ 106, 190, 49 },
		{ 106, 191, 49 },
		{ 106, 192, 49 },
		{ 106, 193, 49 },
		{ 106, 194, 49 },
		{ 106, 195, 49 },
		{ 106, 196, 49 },
		{ 106, 197, 49 },
		{ 106, 198, 49 },
		{ 106, 199, 49 },
		{ 106, 200, 49 },
		{ 106, 201, 49 },
		{ 106, 202, 49 },
		{ 106, 203, 49 },
		{ 106, 204, 49 },
		{ 106, 205, 49 },
		{ 106, 206, 49 },
		{ 106, 207, 49 },
		{ 106, 208, 49 },
		{ 106, 209, 49 },
		{ 106, 210, 49 },
		{ 106, 211, 49 },
		{ 106, 212, 49 },
		{ 106, 213, 49 },
		{ 106, 214, 49 },
		{ 106, 215, 49 },
		{ 106, 216, 49 },
		{ 106, 217, 49 },
		{ 106, 218, 49 },
		{ 106, 219, 49 },
		{ 107, 183, 49 },
		{ 107, 184, 49 },
		{ 107, 185, 49 },
		{ 107, 186, 49 },
		{ 107, 187, 49 },
		{ 107, 188, 49 },
		{ 107, 189, 49 },
		{ 107, 190, 49 },
		{ 107, 191, 49 },
		{ 107, 192, 49 },
		{ 107, 193, 49 },
		{ 107, 194, 49 },
		{ 107, 195, 49 },
		{ 107, 196, 49 },
		{ 107, 197, 49 },
		{ 107, 198, 49 },
		{ 107, 199, 49 },
		{ 107, 200, 49 },
		{ 107, 201, 49 },
		{ 107, 202, 49 },
		{ 107, 203, 49 },
		{ 107, 204, 49 },
		{ 107, 205, 49 },
		{ 107, 206, 49 },
		{ 107, 207, 49 },
		{ 107, 208, 49 },
		{ 107, 209, 49 },
		{ 107, 210, 49 },
		{ 107, 211, 49 },
		{ 107, 212, 49 },
		{ 107, 213, 49 },
		{ 107, 214, 49 },
		{ 107, 215, 49 },
		{ 107, 216, 49 },
		{ 107, 217, 49 },
		{ 107, 218, 49 },
		{ 107, 219, 49 },
		{ 108, 183, 49 },
		{ 108, 184, 49 },
		{ 108, 185, 49 },
		{ 108, 186, 49 },
		{ 108, 187, 49 },
		{ 108, 188, 49 },
		{ 108, 189, 49 },
		{ 108, 190, 49 },
		{ 108, 191, 49 },
		{ 108, 192, 49 },
		{ 108, 193, 49 },
		{ 108, 194, 49 },
		{ 108, 195, 49 },
		{ 108, 196, 49 },
		{ 108, 197, 49 },
		{ 108, 198, 49 },
		{ 108, 199, 49 },
		{ 108, 200, 49 },
		{ 108, 201, 49 },
		{ 108, 202, 49 },
		{ 108, 203, 49 },
		{ 108, 204, 49 },
		{ 108, 205, 49 },
		{ 108, 206, 49 },
		{ 108, 207, 49 },
		{ 108, 208, 49 },
		{ 108, 209, 49 },
		{ 108, 210, 49 },
		{ 108, 211, 49 },
		{ 108, 212, 49 },
		{ 108, 213, 49 },
		{ 108, 214, 49 },
		{ 108, 215, 49 },
		{ 108, 216, 49 },
		{ 108, 217, 49 },
		{ 108, 218, 49 },
		{ 108, 219, 49 },
		{ 109, 184, 49 },
		{ 109, 185, 49 },
		{ 109, 186, 49 },
		{ 109, 187, 49 },
		{ 109, 188, 49 },
		{ 109, 189, 49 },
		{ 109, 190, 49 },
		{ 109, 191, 49 },
		{ 109, 192, 49 },
		{ 109, 193, 49 },
		{ 109, 194, 49 },
		{ 109, 195, 49 },
		{ 109, 196, 49 },
		{ 109, 197, 49 },
		{ 109, 198, 49 },
		{ 109, 199, 49 },
		{ 109, 200, 49 },
		{ 109, 201, 49 },
		{ 109, 202, 49 },
		{ 109, 203, 49 },
		{ 109, 204, 49 },
		{ 109, 205, 49 },
		{ 109, 206, 49 },
		{ 109, 207, 49 },
		{ 109, 208, 49 },
		{ 109, 209, 49 },
		{ 109, 210, 49 },
		{ 109, 211, 49 },
		{ 109, 212, 49 },
		{ 109, 213, 49 },
		{ 109, 214, 49 },
		{ 109, 215, 49 },
		{ 109, 216, 49 },
		{ 109, 217, 49 },
		{ 109, 218, 49 },
		{ 109, 219, 49 },
		{ 110, 184, 49 },
		{ 110, 185, 49 },
		{ 110, 186, 49 },
		{ 110, 187, 49 },
		{ 110, 188, 49 },
		{ 110, 189, 49 },
		{ 110, 190, 49 },
		{ 110, 191, 49 },
		{ 110, 192, 49 },
		{ 110, 193, 49 },
		{ 110, 194, 49 },
		{ 110, 195, 49 },
		{ 110, 196, 49 },
		{ 110, 197, 49 },
		{ 110, 198, 49 },
		{ 110, 199, 49 },
		{ 110, 200, 49 },
		{ 110, 201, 49 },
		{ 110, 202, 49 },
		{ 110, 203, 49 },
		{ 110, 204, 49 },
		{ 110, 205, 49 },
		{ 110, 206, 49 },
		{ 110, 207, 49 },
		{ 110, 208, 49 },
		{ 110, 209, 49 },
		{ 110, 210, 49 },
		{ 110, 211, 49 },
		{ 110, 212, 49 },
		{ 110, 213, 49 },
		{ 110, 214, 49 },
		{ 110, 215, 49 },
		{ 110, 216, 49 },
		{ 110, 217, 49 },
		{ 110, 218, 49 },
		{ 110, 219, 49 },
		{ 110, 220, 49 },
		{ 111, 185, 49 },
		{ 111, 186, 49 },
		{ 111, 187, 49 },
		{ 111, 188, 49 },
		{ 111, 189, 49 },
		{ 111, 190, 49 },
		{ 111, 191, 49 },
		{ 111, 192, 49 },
		{ 111, 193, 49 },
		{ 111, 194, 49 },
		{ 111, 195, 49 },
		{ 111, 196, 49 },
		{ 111, 197, 49 },
		{ 111, 198, 49 },
		{ 111, 199, 49 },
		{ 111, 200, 49 },
		{ 111, 201, 49 },
		{ 111, 202, 49 },
		{ 111, 203, 49 },
		{ 111, 204, 49 },
		{ 111, 205, 49 },
		{ 111, 206, 49 },
		{ 111, 207, 49 },
		{ 111, 208, 49 },
		{ 111, 209, 49 },
		{ 111, 210, 49 },
		{ 111, 211, 49 },
		{ 111, 212, 49 },
		{ 111, 213, 49 },
		{ 111, 214, 49 },
		{ 111, 215, 49 },
		{ 111, 216, 49 },
		{ 111, 217, 49 },
		{ 111, 218, 49 },
		{ 111, 219, 49 },
		{ 111, 220, 49 },
		{ 112, 185, 49 },
		{ 112, 186, 49 },
		{ 112, 187, 49 },
		{ 112, 188, 49 },
		{ 112, 189, 49 },
		{ 112, 190, 49 },
		{ 112, 191, 49 },
		{ 112, 192, 49 },
		{ 112, 193, 49 },
		{ 112, 194, 49 },
		{ 112, 195, 49 },
		{ 112, 196, 49 },
		{ 112, 197, 49 },
		{ 112, 198, 49 },
		{ 112, 199, 49 },
		{ 112, 200, 49 },
		{ 112, 201, 49 },
		{ 112, 202, 49 },
		{ 112, 203, 49 },
		{ 112, 204, 49 },
		{ 112, 205, 49 },
		{ 112, 206, 49 },
		{ 112, 207, 49 },
		{ 112, 208, 49 },
		{ 112, 209, 49 },
		{ 112, 210, 49 },
		{ 112, 211, 49 },
		{ 112, 212, 49 },
		{ 112, 213, 49 },
		{ 112, 214, 49 },
		{ 112, 215, 49 },
		{ 112, 216, 49 },
		{ 112, 217, 49 },
		{ 112, 218, 49 },
		{ 112, 219, 49 },
		{ 112, 220, 49 },
		{ 113, 186, 49 },
		{ 113, 187, 49 },
		{ 113, 188, 49 },
		{ 113, 189, 49 },
		{ 113, 190, 49 },
		{ 113, 191, 49 },
		{ 113, 192, 49 },
		{ 113, 193, 49 },
		{ 113, 194, 49 },
		{ 113, 195, 49 },
		{ 113, 196, 49 },
		{ 113, 197, 49 },
		{ 113, 198, 49 },
		{ 113, 199, 49 },
		{ 113, 200, 49 },
		{ 113, 201, 49 },
		{ 113, 202, 49 },
		{ 113, 203, 49 },
		{ 113, 204, 49 },
		{ 113, 205, 49 },
		{ 113, 206, 49 },
		{ 113, 207, 49 },
		{ 113, 208, 49 },
		{ 113, 209, 49 },
		{ 113, 210, 49 },
		{ 113, 211, 49 },
		{ 113, 212, 49 },
		{ 113, 213, 49 },
		{ 113, 214, 49 },
		{ 113, 215, 49 },
		{ 113, 216, 49 },
		{ 113, 217, 49 },
		{ 113, 218, 49 },
		{ 113, 219, 49 },
		{ 113, 220, 49 },
		{ 114, 186, 49 },
		{ 114, 187, 49 },
		{ 114, 188, 49 },
		{ 114, 189, 49 },
		{ 114, 190, 49 },
		{ 114, 191, 49 },
		{ 114, 192, 49 },
		{ 114, 193, 49 },
		{ 114, 194, 49 },
		{ 114, 195, 49 },
		{ 114, 196, 49 },
		{ 114, 197, 49 },
		{ 114, 198, 49 },
		{ 114, 199, 49 },
		{ 114, 200, 49 },
		{ 114, 201, 49 },
		{ 114, 202, 49 },
		{ 114, 203, 49 },
		{ 114, 204, 49 },
		{ 114, 205, 49 },
		{ 114, 206, 49 },
		{ 114, 207, 49 },
		{ 114, 208, 49 },
		{ 114, 209, 49 },
		{ 114, 210, 49 },
		{ 114, 211, 49 },
		{ 114, 212, 49 },
		{ 114, 213, 49 },
		{ 114, 214, 49 },
		{ 114, 215, 49 },
		{ 114, 216, 49 },
		{ 114, 217, 49 },
		{ 114, 218, 49 },
		{ 114, 219, 49 },
		{ 114, 220, 49 },
		{ 114, 221, 49 },
		{ 115, 187, 49 },
		{ 115, 188, 49 },
		{ 115, 189, 49 },
		{ 115, 190, 49 },
		{ 115, 191, 49 },
		{ 115, 192, 49 },
		{ 115, 193, 49 },
		{ 115, 194, 49 },
		{ 115, 195, 49 },
		{ 115, 196, 49 },
		{ 115, 197, 49 },
		{ 115, 198, 49 },
		{ 115, 199, 49 },
		{ 115, 200, 49 },
		{ 115, 201, 49 },
		{ 115, 202, 49 },
		{ 115, 203, 49 },
		{ 115, 204, 49 },
		{ 115, 205, 49 },
		{ 115, 206, 49 },
		{ 115, 207, 49 },
		{ 115, 208, 49 },
		{ 115, 209, 49 },
		{ 115, 210, 49 },
		{ 115, 211, 49 },
		{ 115, 212, 49 },
		{ 115, 213, 49 },
		{ 115, 214, 49 },
		{ 115, 215, 49 },
		{ 115, 216, 49 },
		{ 115, 217, 49 },
		{ 115, 218, 49 },
		{ 115, 219, 49 },
		{ 115, 220, 49 },
		{ 115, 221, 49 },
		{ 116, 187, 49 },
		{ 116, 188, 49 },
		{ 116, 189, 49 },
		{ 116, 190, 49 },
		{ 116, 191, 49 },
		{ 116, 192, 49 },
		{ 116, 193, 49 },
		{ 116, 194, 49 },
		{ 116, 195, 49 },
		{ 116, 196, 49 },
		{ 116, 197, 49 },
		{ 116, 198, 49 },
		{ 116, 199, 49 },
		{ 116, 200, 49 },
		{ 116, 201, 49 },
		{ 116, 202, 49 },
		{ 116, 203, 49 },
		{ 116, 204, 49 },
		{ 116, 205, 49 },
		{ 116, 206, 49 },
		{ 116, 207, 49 },
		{ 116, 208, 49 },
		{ 116, 209, 49 },
		{ 116, 210, 49 },
		{ 116, 211, 49 },
		{ 116, 212, 49 },
		{ 116, 213, 49 },
		{ 116, 214, 49 },
		{ 116, 215, 49 },
		{ 116, 216, 49 },
		{ 116, 217, 49 },
		{ 116, 218, 49 },
		{ 116, 219, 49 },
		{ 116, 220, 49 },
		{ 116, 221, 49 },
		{ 117, 188, 49 },
		{ 117, 189, 49 },
		{ 117, 190, 49 },
		{ 117, 191, 49 },
		{ 117, 192, 49 },
		{ 117, 193, 49 },
		{ 117, 194, 49 },
		{ 117, 195, 49 },
		{ 117, 196, 49 },
		{ 117, 197, 49 },
		{ 117, 198, 49 },
		{ 117, 199, 49 },
		{ 117, 200, 49 },
		{ 117, 201, 49 },
		{ 117, 202, 49 },
		{ 117, 203, 49 },
		{ 117, 204, 49 },
		{ 117, 205, 49 },
		{ 117, 206, 49 },
		{ 117, 207, 49 },
		{ 117, 208, 49 },
		{ 117, 209, 49 },
		{ 117, 210, 49 },
		{ 117, 211, 49 },
		{ 117, 212, 49 },
		{ 117, 213, 49 },
		{ 117, 214, 49 },
		{ 117, 215, 49 },
		{ 117, 216, 49 },
		{ 117, 217, 49 },
		{ 117, 218, 49 },
		{ 117, 219, 49 },
		{ 117, 220, 49 },
		{ 117, 221, 49 },
		{ 118, 188, 49 },
		{ 118, 189, 49 },
		{ 118, 190, 49 },
		{ 118, 191, 49 },
		{ 118, 192, 49 },
		{ 118, 193, 49 },
		{ 118, 194, 49 },
		{ 118, 195, 49 },
		{ 118, 196, 49 },
		{ 118, 197, 49 },
		{ 118, 198, 49 },
		{ 118, 199, 49 },
		{ 118, 200, 49 },
		{ 118, 201, 49 },
		{ 118, 202, 49 },
		{ 118, 203, 49 },
		{ 118, 204, 49 },
		{ 118, 205, 49 },
		{ 118, 206, 49 },
		{ 118, 207, 49 },
		{ 118, 208, 49 },
		{ 118, 209, 49 },
		{ 118, 210, 49 },
		{ 118, 211, 49 },
		{ 118, 212, 49 },
		{ 118, 213, 49 },
		{ 118, 214, 49 },
		{ 118, 215, 49 },
		{ 118, 216, 49 },
		{ 118, 217, 49 },
		{ 118, 218, 49 },
		{ 118, 219, 49 },
		{ 118, 220, 49 },
		{ 118, 221, 49 },
		{ 118, 222, 49 },
		{ 119, 189, 49 },
		{ 119, 190, 49 },
		{ 119, 191, 49 },
		{ 119, 192, 49 },
		{ 119, 193, 49 },
		{ 119, 194, 49 },
		{ 119, 195, 49 },
		{ 119, 196, 49 },
		{ 119, 197, 49 },
		{ 119, 198, 49 },
		{ 119, 199, 49 },
		{ 119, 200, 49 },
		{ 119, 201, 49 },
		{ 119, 202, 49 },
		{ 119, 203, 49 },
		{ 119, 204, 49 },
		{ 119, 205, 49 },
		{ 119, 206, 49 },
		{ 119, 207, 49 },
		{ 119, 208, 49 },
		{ 119, 209, 49 },
		{ 119, 210, 49 },
		{ 119, 211, 49 },
		{ 119, 212, 49 },
		{ 119, 213, 49 },
		{ 119, 214, 49 },
		{ 119, 215, 49 },
		{ 119, 216, 49 },
		{ 119, 217, 49 },
		{ 119, 218, 49 },
		{ 119, 219, 49 },
		{ 119, 220, 49 },
		{ 119, 221, 49 },
		{ 119, 222, 49 },
		{ 120, 189, 49 },
		{ 120, 190, 49 },
		{ 120, 191, 49 },
		{ 120, 192, 49 },
		{ 120, 193, 49 },
		{ 120, 194, 49 },
		{ 120, 195, 49 },
		{ 120, 196, 49 },
		{ 120, 197, 49 },
		{ 120, 198, 49 },
		{ 120, 199, 49 },
		{ 120, 200, 49 },
		{ 120, 201, 49 },
		{ 120, 202, 49 },
		{ 120, 203, 49 },
		{ 120, 204, 49 },
		{ 120, 205, 49 },
		{ 120, 206, 49 },
		{ 120, 207, 49 },
		{ 120, 208, 49 },
		{ 120, 209, 49 },
		{ 120, 210, 49 },
		{ 120, 211, 49 },
		{ 120, 212, 49 },
		{ 120, 213, 49 },
		{ 120, 214, 49 },
		{ 120, 215, 49 },
		{ 120, 216, 49 },
		{ 120, 217, 49 },
		{ 120, 218, 49 },
		{ 120, 219, 49 },
		{ 120, 220, 49 },
		{ 120, 221, 49 },
		{ 120, 222, 49 },
		{ 121, 190, 49 },
		{ 121, 191, 49 },
		{ 121, 192, 49 },
		{ 121, 193, 49 },
		{ 121, 194, 49 },
		{ 121, 195, 49 },
		{ 121, 196, 49 },
		{ 121, 197, 49 },
		{ 121, 198, 49 },
		{ 121, 199, 49 },
		{ 121, 200, 49 },
		{ 121, 201, 49 },
		{ 121, 202, 49 },
		{ 121, 203, 49 },
		{ 121, 204, 49 },
		{ 121, 205, 49 },
		{ 121, 206, 49 },
		{ 121, 207, 49 },
		{ 121, 208, 49 },
		{ 121, 209, 49 },
		{ 121, 210, 49 },
		{ 121, 211, 49 },
		{ 121, 212, 49 },
		{ 121, 213, 49 },
		{ 121, 214, 49 },
		{ 121, 215, 49 },
		{ 121, 216, 49 },
		{ 121, 217, 49 },
		{ 121, 218, 49 },
		{ 121, 219, 49 },
		{ 121, 220, 49 },
		{ 121, 221, 49 },
		{ 121, 222, 49 },
		{ 122, 190, 49 },
		{ 122, 191, 49 },
		{ 122, 192, 49 },
		{ 122, 193, 49 },
		{ 122, 194, 49 },
		{ 122, 195, 49 },
		{ 122, 196, 49 },
		{ 122, 197, 49 },
		{ 122, 198, 49 },
		{ 122, 199, 49 },
		{ 122, 200, 49 },
		{ 122, 201, 49 },
		{ 122, 202, 49 },
		{ 122, 203, 49 },
		{ 122, 204, 49 },
		{ 122, 205, 49 },
		{ 122, 206, 49 },
		{ 122, 207, 49 },
		{ 122, 208, 49 },
		{ 122, 209, 49 },
		{ 122, 210, 49 },
		{ 122, 211, 49 },
		{ 122, 212, 49 },
		{ 122, 213, 49 },
		{ 122, 214, 49 },
		{ 122, 215, 49 },
		{ 122, 216, 49 },
		{ 122, 217, 49 },
		{ 122, 218, 49 },
		{ 122, 219, 49 },
		{ 122, 220, 49 },
		{ 122, 221, 49 },
		{ 122, 222, 49 },
		{ 122, 223, 49 },
		{ 123, 191, 49 },
		{ 123, 192, 49 },
		{ 123, 193, 49 },
		{ 123, 194, 49 },
		{ 123, 195, 49 },
		{ 123, 196, 49 },
		{ 123, 197, 49 },
		{ 123, 198, 49 },
		{ 123, 199, 49 },
		{ 123, 200, 49 },
		{ 123, 201, 49 },
		{ 123, 202, 49 },
		{ 123, 203, 49 },
		{ 123, 204, 49 },
		{ 123, 205, 49 },
		{ 123, 206, 49 },
		{ 123, 207, 49 },
		{ 123, 208, 49 },
		{ 123, 209, 49 },
		{ 123, 210, 49 },
		{ 123, 211, 49 },
		{ 123, 212, 49 },
		{ 123, 213, 49 },
		{ 123, 214, 49 },
		{ 123, 215, 49 },
		{ 123, 216, 49 },
		{ 123, 217, 49 },
		{ 123, 218, 49 },
		{ 123, 219, 49 },
		{ 123, 220, 49 },
		{ 123, 221, 49 },
		{ 123, 222, 49 },
		{ 123, 223, 49 },
		{ 124, 191, 49 },
		{ 124, 192, 49 },
		{ 124, 193, 49 },
		{ 124, 194, 49 },
		{ 124, 195, 49 },
		{ 124, 196, 49 },
		{ 124, 197, 49 },
		{ 124, 198, 49 },
		{ 124, 199, 49 },
		{ 124, 200, 49 },
		{ 124, 201, 49 },
		{ 124, 202, 49 },
		{ 124, 203, 49 },
		{ 124, 204, 49 },
		{ 124, 205, 49 },
		{ 124, 206, 49 },
		{ 124, 207, 49 },
		{ 124, 208, 49 },
		{ 124, 209, 49 },
		{ 124, 210, 49 },
		{ 124, 211, 49 },
		{ 124, 212, 49 },
		{ 124, 213, 49 },
		{ 124, 214, 49 },
		{ 124, 215, 49 },
		{ 124, 216, 49 },
		{ 124, 217, 49 },
		{ 124, 218, 49 },
		{ 124, 219, 49 },
		{ 124, 220, 49 },
		{ 124, 221, 49 },
		{ 124, 222, 49 },
		{ 124, 223, 49 },
		{ 125, 192, 49 },
		{ 125, 193, 49 },
		{ 125, 194, 49 },
		{ 125, 195, 49 },
		{ 125, 196, 49 },
		{ 125, 197, 49 },
		{ 125, 198, 49 },
		{ 125, 199, 49 },
		{ 125, 200, 49 },
		{ 125, 201, 49 },
		{ 125, 202, 49 },
		{ 125, 203, 49 },
		{ 125, 204, 49 },
		{ 125, 205, 49 },
		{ 125, 206, 49 },
		{ 125, 207, 49 },
		{ 125, 208, 49 },
		{ 125, 209, 49 },
		{ 125, 210, 49 },
		{ 125, 211, 49 },
		{ 125, 212, 49 },
		{ 125, 213, 49 },
		{ 125, 214, 49 },
		{ 125, 215, 49 },
		{ 125, 216, 49 },
		{ 125, 217, 49 },
		{ 125, 218, 49 },
		{ 125, 219, 49 },
		{ 125, 220, 49 },
		{ 125, 221, 49 },
		{ 125, 222, 49 },
		{ 125, 223, 49 },
		{ 126, 192, 49 },
		{ 126, 193, 49 },
		{ 126, 194, 49 },
		{ 126, 195, 49 },
		{ 126, 196, 49 },
		{ 126, 197, 49 },
		{ 126, 198, 49 },
		{ 126, 199, 49 },
		{ 126, 200, 49 },
		{ 126, 201, 49 },
		{ 126, 202, 49 },
		{ 126, 203, 49 },
		{ 126, 204, 49 },
		{ 126, 205, 49 },
		{ 126, 206, 49 },
		{ 126, 207, 49 },
		{ 126, 208, 49 },
		{ 126, 209, 49 },
		{ 126, 210, 49 },
		{ 126, 211, 49 },
		{ 126, 212, 49 },
		{ 126, 213, 49 },
		{ 126, 214, 49 },
		{ 126, 215, 49 },
		{ 126, 216, 49 },
		{ 126, 217, 49 },
		{ 126, 218, 49 },
		{ 126, 219, 49 },
		{ 126, 220, 49 },
		{ 126, 221, 49 },
		{ 126, 222, 49 },
		{ 126, 223, 49 },
		{ 126, 224, 49 },
		{ 127, 193, 49 },
		{ 127, 194, 49 },
		{ 127, 195, 49 },
		{ 127, 196, 49 },
		{ 127, 197, 49 },
		{ 127, 198, 49 },
		{ 127, 199, 49 },
		{ 127, 200, 49 },
		{ 127, 201, 49 },
		{ 127, 202, 49 },
		{ 127, 203, 49 },
		{ 127, 204, 49 },
		{ 127, 205, 49 },
		{ 127, 206, 49 },
		{ 127, 207, 49 },
		{ 127, 208, 49 },
		{ 127, 209, 49 },
		{ 127, 210, 49 },
		{ 127, 211, 49 },
		{ 127, 212, 49 },
		{ 127, 213, 49 },
		{ 127, 214, 49 },
		{ 127, 215, 49 },
		{ 127, 216, 49 },
		{ 127, 217, 49 },
		{ 127, 218, 49 },
		{ 127, 219, 49 },
		{ 127, 220, 49 },
		{ 127, 221, 49 },
		{ 127, 222, 49 },
		{ 127, 223, 49 },
		{ 127, 224, 49 },
		{ 128, 193, 49 },
		{ 128, 194, 49 },
		{ 128, 195, 49 },
		{ 128, 196, 49 },
		{ 128, 197, 49 },
		{ 128, 198, 49 },
		{ 128, 199, 49 },
		{ 128, 200, 49 },
		{ 128, 201, 49 },
		{ 128, 202, 49 },
		{ 128, 203, 49 },
		{ 128, 204, 49 },
		{ 128, 205, 49 },
		{ 128, 206, 49 },
		{ 128, 207, 49 },
		{ 128, 208, 49 },
		{ 128, 209, 49 },
		{ 128, 210, 49 },
		{ 128, 211, 49 },
		{ 128, 212, 49 },
		{ 128, 213, 49 },
		{ 128, 214, 49 },
		{ 128, 215, 49 },
		{ 128, 216, 49 },
		{ 128, 217, 49 },
		{ 128, 218, 49 },
		{ 128, 219, 49 },
		{ 128, 220, 49 },
		{ 128, 221, 49 },
		{ 128, 222, 49 },
		{ 128, 223, 49 },
		{ 128, 224, 49 },
		{ 129, 194, 49 },
		{ 129, 195, 49 },
		{ 129, 196, 49 },
		{ 129, 197, 49 },
		{ 129, 198, 49 },
		{ 129, 199, 49 },
		{ 129, 200, 49 },
		{ 129, 201, 49 },
		{ 129, 202, 49 },
		{ 129, 203, 49 },
		{ 129, 204, 49 },
		{ 129, 205, 49 },
		{ 129, 206, 49 },
		{ 129, 207, 49 },
		{ 129, 208, 49 },
		{ 129, 209, 49 },
		{ 129, 210, 49 },
		{ 129, 211, 49 },
		{ 129, 212, 49 },
		{ 129, 213, 49 },
		{ 129, 214, 49 },
		{ 129, 215, 49 },
		{ 129, 216, 49 },
		{ 129, 217, 49 },
		{ 129, 218, 49 },
		{ 129, 219, 49 },
		{ 129, 220, 49 },
		{ 129, 221, 49 },
		{ 129, 222, 49 },
		{ 129, 223, 49 },
		{ 129, 224, 49 },
		{ 130, 194, 49 },
		{ 130, 195, 49 },
		{ 130, 196, 49 },
		{ 130, 197, 49 },
		{ 130, 198, 49 },
		{ 130, 199, 49 },
		{ 130, 200, 49 },
		{ 130, 201, 49 },
		{ 130, 202, 49 },
		{ 130, 203, 49 },
		{ 130, 204, 49 },
		{ 130, 205, 49 },
		{ 130, 206, 49 },
		{ 130, 207, 49 },
		{ 130, 208, 49 },
		{ 130, 209, 49 },
		{ 130, 210, 49 },
		{ 130, 211, 49 },
		{ 130, 212, 49 },
		{ 130, 213, 49 },
		{ 130, 214, 49 },
		{ 130, 215, 49 },
		{ 130, 216, 49 },
		{ 130, 217, 49 },
		{ 130, 218, 49 },
		{ 130, 219, 49 },
		{ 130, 220, 49 },
		{ 130, 221, 49 },
		{ 130, 222, 49 },
		{ 130, 223, 49 },
		{ 130, 224, 49 },
		{ 130, 225, 49 },
		{ 131, 195, 49 },
		{ 131, 196, 49 },
		{ 131, 197, 49 },
		{ 131, 198, 49 },
		{ 131, 199, 49 },
		{ 131, 200, 49 },
		{ 131, 201, 49 },
		{ 131, 202, 49 },
		{ 131, 203, 49 },
		{ 131, 204, 49 },
		{ 131, 205, 49 },
		{ 131, 206, 49 },
		{ 131, 207, 49 },
		{ 131, 208, 49 },
		{ 131, 209, 49 },
		{ 131, 210, 49 },
		{ 131, 211, 49 },
		{ 131, 212, 49 },
		{ 131, 213, 49 },
		{ 131, 214, 49 },
		{ 131, 215, 49 },
		{ 131, 216, 49 },
		{ 131, 217, 49 },
		{ 131, 218, 49 },
		{ 131, 219, 49 },
		{ 131, 220, 49 },
		{ 131, 221, 49 },
		{ 131, 222, 49 },
		{ 131, 223, 49 },
		{ 131, 224, 49 },
		{ 131, 225, 49 },
		{ 132, 195, 49 },
		{ 132, 196, 49 },
		{ 132, 197, 49 },
		{ 132, 198, 49 },
		{ 132, 199, 49 },
		{ 132, 200, 49 },
		{ 132, 201, 49 },
		{ 132, 202, 49 },
		{ 132, 203, 49 },
		{ 132, 204, 49 },
		{ 132, 205, 49 },
		{ 132, 206, 49 },
		{ 132, 207, 49 },
		{ 132, 208, 49 },
		{ 132, 209, 49 },
		{ 132, 210, 49 },
		{ 132, 211, 49 },
		{ 132, 212, 49 },
		{ 132, 213, 49 },
		{ 132, 214, 49 },
		{ 132, 215, 49 },
		{ 132, 216, 49 },
		{ 132, 217, 49 },
		{ 132, 218, 49 },
		{ 132, 219, 49 },
		{ 132, 220, 49 },
		{ 132, 221, 49 },
		{ 132, 222, 49 },
		{ 132, 223, 49 },
		{ 132, 224, 49 },
		{ 132, 225, 49 },
		{ 133, 196, 49 },
		{ 133, 197, 49 },
		{ 133, 198, 49 },
		{ 133, 199, 49 },
		{ 133, 200, 49 },
		{ 133, 201, 49 },
		{ 133, 202, 49 },
		{ 133, 203, 49 },
		{ 133, 204, 49 },
		{ 133, 205, 49 },
		{ 133, 206, 49 },
		{ 133, 207, 49 },
		{ 133, 208, 49 },
		{ 133, 209, 49 },
		{ 133, 210, 49 },
		{ 133, 211, 49 },
		{ 133, 212, 49 },
		{ 133, 213, 49 },
		{ 133, 214, 49 },
		{ 133, 215, 49 },
		{ 133, 216, 49 },
		{ 133, 217, 49 },
		{ 133, 218, 49 },
		{ 133, 219, 49 },
		{ 133, 220, 49 },
		{ 133, 221, 49 },
		{ 133, 222, 49 },
		{ 133, 223, 49 },
		{ 133, 224, 49 },
		{ 133, 225, 49 },
		{ 134, 196, 49 },
		{ 134, 197, 49 },
		{ 134, 198, 49 },
		{ 134, 199, 49 },
		{ 134, 200, 49 },
		{ 134, 201, 49 },
		{ 134, 202, 49 },
		{ 134, 203, 49 },
		{ 134, 204, 49 },
		{ 134, 205, 49 },
		{ 134, 206, 49 },
		{ 134, 207, 49 },
		{ 134, 208, 49 },
		{ 134, 209, 49 },
		{ 134, 210, 49 },
		{ 134, 211, 49 },
		{ 134, 212, 49 },
		{ 134, 213, 49 },
		{ 134, 214, 49 },
		{ 134, 215, 49 },
		{ 134, 216, 49 },
		{ 134, 217, 49 },
		{ 134, 218, 49 },
		{ 134, 219, 49 },
		{ 134, 220, 49 },
		{ 134, 221, 49 },
		{ 134, 222, 49 },
		{ 134, 223, 49 },
		{ 134, 224, 49 },
		{ 134, 225, 49 },
		{ 134, 226, 49 },
		{ 135, 197, 49 },
		{ 135, 198, 49 },
		{ 135, 199, 49 },
		{ 135, 200, 49 },
		{ 135, 201, 49 },
		{ 135, 202, 49 },
		{ 135, 203, 49 },
		{ 135, 204, 49 },
		{ 135, 205, 49 },
		{ 135, 206, 49 },
		{ 135, 207, 49 },
		{ 135, 208, 49 },
		{ 135, 209, 49 },
		{ 135, 210, 49 },
		{ 135, 211, 49 },
		{ 135, 212, 49 },
		{ 135, 213, 49 },
		{ 135, 214, 49 },
		{ 135, 215, 49 },
		{ 135, 216, 49 },
		{ 135, 217, 49 },
		{ 135, 218, 49 },
		{ 135, 219, 49 },
		{ 135, 220, 49 },
		{ 135, 221, 49 },
		{ 135, 222, 49 },
		{ 135, 223, 49 },
		{ 135, 224, 49 },
		{ 135, 225, 49 },
		{ 135, 226, 49 },
		{ 136, 197, 49 },
		{ 136, 198, 49 },
		{ 136, 199, 49 },
		{ 136, 200, 49 },
		{ 136, 201, 49 },
		{ 136, 202, 49 },
		{ 136, 203, 49 },
		{ 136, 204, 49 },
		{ 136, 205, 49 },
		{ 136, 206, 49 },
		{ 136, 207, 49 },
		{ 136, 208, 49 },
		{ 136, 209, 49 },
		{ 136, 210, 49 },
		{ 136, 211, 49 },
		{ 136, 212, 49 },
		{ 136, 213, 49 },
		{ 136, 214, 49 },
		{ 136, 215, 49 },
		{ 136, 216, 49 },
		{ 136, 217, 49 },
		{ 136, 218, 49 },
		{ 136, 219, 49 },
		{ 136, 220, 49 },
		{ 136, 221, 49 },
		{ 136, 222, 49 },
		{ 136, 223, 49 },
		{ 136, 224, 49 },
		{ 136, 225, 49 },
		{ 136, 226, 49 },
		{ 137, 198, 49 },
		{ 137, 199, 49 },
		{ 137, 200, 49 },
		{ 137, 201, 49 },
		{ 137, 202, 49 },
		{ 137, 203, 49 },
		{ 137, 204, 49 },
		{ 137, 205, 49 },
		{ 137, 206, 49 },
		{ 137, 207, 49 },
		{ 137, 208, 49 },
		{ 137, 209, 49 },
		{ 137, 210, 49 },
		{ 137, 211, 49 },
		{ 137, 212, 49 },
		{ 137, 213, 49 },
		{ 137, 214, 49 },
		{ 137, 215, 49 },
		{ 137, 216, 49 },
		{ 137, 217, 49 },
		{ 137, 218, 49 },
		{ 137, 219, 49 },
		{ 137, 220, 49 },
		{ 137, 221, 49 },
		{ 137, 222, 49 },
		{ 137, 223, 49 },
		{ 137, 224, 49 },
		{ 137, 225, 49 },
		{ 137, 226, 49 },
		{ 138, 198, 49 },
		{ 138, 199, 49 },
		{ 138, 200, 49 },
		{ 138, 201, 49 },
		{ 138, 202, 49 },
		{ 138, 203, 49 },
		{ 138, 204, 49 },
		{ 138, 205, 49 },
		{ 138, 206, 49 },
		{ 138, 207, 49 },
		{ 138, 208, 49 },
		{ 138, 209, 49 },
		{ 138, 210, 49 },
		{ 138, 211, 49 },
		{ 138, 212, 49 },
		{ 138, 213, 49 },
		{ 138, 214, 49 },
		{ 138, 215, 49 },
		{ 138, 216, 49 },
		{ 138, 217, 49 },
		{ 138, 218, 49 },
		{ 138, 219, 49 },
		{ 138, 220, 49 },
		{ 138, 221, 49 },
		{ 138, 222, 49 },
		{ 138, 223, 49 },
		{ 138, 224, 49 },
		{ 138, 225, 49 },
		{ 138, 226, 49 },
		{ 138, 227, 49 },
		{ 139, 199, 49 },
		{ 139, 200, 49 },
		{ 139, 201, 49 },
		{ 139, 202, 49 },
		{ 139, 203, 49 },
		{ 139, 204, 49 },
		{ 139, 205, 49 },
		{ 139, 206, 49 },
		{ 139, 207, 49 },
		{ 139, 208, 49 },
		{ 139, 209, 49 },
		{ 139, 210, 49 },
		{ 139, 211, 49 },
		{ 139, 212, 49 },
		{ 139, 213, 49 },
		{ 139, 214, 49 },
		{ 139, 215, 49 },
		{ 139, 216, 49 },
		{ 139, 217, 49 },
		{ 139, 218, 49 },
		{ 139, 219, 49 },
		{ 139, 220, 49 },
		{ 139, 221, 49 },
		{ 139, 222, 49 },
		{ 139, 223, 49 },
		{ 139, 224, 49 },
		{ 139, 225, 49 },
		{ 139, 226, 49 },
		{ 139, 227, 49 },
		{ 140, 199, 49 },
		{ 140, 200, 49 },
		{ 140, 201, 49 },
		{ 140, 202, 49 },
		{ 140, 203, 49 },
		{ 140, 204, 49 },
		{ 140, 205, 49 },
		{ 140, 206, 49 },
		{ 140, 207, 49 },
		{ 140, 208, 49 },
		{ 140, 209, 49 },
		{ 140, 210, 49 },
		{ 140, 211, 49 },
		{ 140, 212, 49 },
		{ 140, 213, 49 },
		{ 140, 214, 49 },
		{ 140, 215, 49 },
		{ 140, 216, 49 },
		{ 140, 217, 49 },
		{ 140, 218, 49 },
		{ 140, 219, 49 },
		{ 140, 220, 49 },
		{ 140, 221, 49 },
		{ 140, 222, 49 },
		{ 140, 223, 49 },
		{ 140, 224, 49 },
		{ 140, 225, 49 },
		{ 140, 226, 49 },
		{ 140, 227, 49 },
		{ 141, 200, 49 },
		{ 141, 201, 49 },
		{ 141, 202, 49 },
		{ 141, 203, 49 },
		{ 141, 204, 49 },
		{ 141, 205, 49 },
		{ 141, 206, 49 },
		{ 141, 207, 49 },
		{ 141, 208, 49 },
		{ 141, 209, 49 },
		{ 141, 210, 49 },
		{ 141, 211, 49 },
		{ 141, 212, 49 },
		{ 141, 213, 49 },
		{ 141, 214, 49 },
		{ 141, 215, 49 },
		{ 141, 216, 49 },
		{ 141, 217, 49 },
		{ 141, 218, 49 },
		{ 141, 219, 49 },
		{ 141, 220, 49 },
		{ 141, 221, 49 },
		{ 141, 222, 49 },
		{ 141, 223, 49 },
		{ 141, 224, 49 },
		{ 141, 225, 49 },
		{ 141, 226, 49 },
		{ 141, 227, 49 },
		{ 142, 200, 49 },
		{ 142, 201, 49 },
		{ 142, 202, 49 },
		{ 142, 203, 49 },
		{ 142, 204, 49 },
		{ 142, 205, 49 },
		{ 142, 206, 49 },
		{ 142, 207, 49 },
		{ 142, 208, 49 },
		{ 142, 209, 49 },
		{ 142, 210, 49 },
		{ 142, 211, 49 },
		{ 142, 212, 49 },
		{ 142, 213, 49 },
		{ 142, 214, 49 },
		{ 142, 215, 49 },
		{ 142, 216, 49 },
		{ 142, 217, 49 },
		{ 142, 218, 49 },
		{ 142, 219, 49 },
		{ 142, 220, 49 },
		{ 142, 221, 49 },
		{ 142, 222, 49 },
		{ 142, 223, 49 },
		{ 142, 224, 49 },
		{ 142, 225, 49 },
		{ 142, 226, 49 },
		{ 142, 227, 49 },
		{ 142, 228, 49 },
		{ 143, 201, 49 },
		{ 143, 202, 49 },
		{ 143, 203, 49 },
		{ 143, 204, 49 },
		{ 143, 205, 49 },
		{ 143, 206, 49 },
		{ 143, 207, 49 },
		{ 143, 208, 49 },
		{ 143, 209, 49 },
		{ 143, 210, 49 },
		{ 143, 211, 49 },
		{ 143, 212, 49 },
		{ 143, 213, 49 },
		{ 143, 214, 49 },
		{ 143, 215, 49 },
		{ 143, 216, 49 },
		{ 143, 217, 49 },
		{ 143, 218, 49 },
		{ 143, 219, 49 },
		{ 143, 220, 49 },
		{ 143, 221, 49 },
		{ 143, 222, 49 },
		{ 143, 223, 49 },
		{ 143, 224, 49 },
		{ 143, 225, 49 },
		{ 143, 226, 49 },
		{ 143, 227, 49 },
		{ 143, 228, 49 },
		{ 144, 201, 49 },
		{ 144, 202, 49 },
		{ 144, 203, 49 },
		{ 144, 204, 49 },
		{ 144, 205, 49 },
		{ 144, 206, 49 },
		{ 144, 207, 49 },
		{ 144, 208, 49 },
		{ 144, 209, 49 },
		{ 144, 210, 49 },
		{ 144, 211, 49 },
		{ 144, 212, 49 },
		{ 144, 213, 49 },
		{ 144, 214, 49 },
		{ 144, 215, 49 },
		{ 144, 216, 49 },
		{ 144, 217, 49 },
		{ 144, 218, 49 },
		{ 144, 219, 49 },
		{ 144, 220, 49 },
		{ 144, 221, 49 },
		{ 144, 222, 49 },
		{ 144, 223, 49 },
		{ 144, 224, 49 },
		{ 144, 225, 49 },
		{ 144, 226, 49 },
		{ 144, 227, 49 },
		{ 144, 228, 49 },
		{ 145, 202, 49 },
		{ 145, 203, 49 },
		{ 145, 204, 49 },
		{ 145, 205, 49 },
		{ 145, 206, 49 },
		{ 145, 207, 49 },
		{ 145, 208, 49 },
		{ 145, 209, 49 },
		{ 145, 210, 49 },
		{ 145, 211, 49 },
		{ 145, 212, 49 },
		{ 145, 213, 49 },
		{ 145, 214, 49 },
		{ 145, 215, 49 },
		{ 145, 216, 49 },
		{ 145, 217, 49 },
		{ 145, 218, 49 },
		{ 145, 219, 49 },
		{ 145, 220, 49 },
		{ 145, 221, 49 },
		{ 145, 222, 49 },
		{ 145, 223, 49 },
		{ 145, 224, 49 },
		{ 145, 225, 49 },
		{ 145, 226, 49 },
		{ 145, 227, 49 },
		{ 145, 228, 49 },
		{ 146, 202, 49 },
		{ 146, 203, 49 },
		{ 146, 204, 49 },
		{ 146, 205, 49 },
		{ 146, 206, 49 },
		{ 146, 207, 49 },
		{ 146, 208, 49 },
		{ 146, 209, 49 },
		{ 146, 210, 49 },
		{ 146, 211, 49 },
		{ 146, 212, 49 },
		{ 146, 213, 49 },
		{ 146, 214, 49 },
		{ 146, 215, 49 },
		{ 146, 216, 49 },
		{ 146, 217, 49 },
		{ 146, 218, 49 },
		{ 146, 219, 49 },
		{ 146, 220, 49 },
		{ 146, 221, 49 },
		{ 146, 222, 49 },
		{ 146, 223, 49 },
		{ 146, 224, 49 },
		{ 146, 225, 49 },
		{ 146, 226, 49 },
		{ 146, 227, 49 },
		{ 146, 228, 49 },
		{ 146, 229, 49 },
		{ 147, 203, 49 },
		{ 147, 204, 49 },
		{ 147, 205, 49 },
		{ 147, 206, 49 },
		{ 147, 207, 49 },
		{ 147, 208, 49 },
		{ 147, 209, 49 },
		{ 147, 210, 49 },
		{ 147, 211, 49 },
		{ 147, 212, 49 },
		{ 147, 213, 49 },
		{ 147, 214, 49 },
		{ 147, 215, 49 },
		{ 147, 216, 49 },
		{ 147, 217, 49 },
		{ 147, 218, 49 },
		{ 147, 219, 49 },
		{ 147, 220, 49 },
		{ 147, 221, 49 },
		{ 147, 222, 49 },
		{ 147, 223, 49 },
		{ 147, 224, 49 },
		{ 147, 225, 49 },
		{ 147, 226, 49 },
		{ 147, 227, 49 },
		{ 147, 228, 49 },
		{ 147, 229, 49 },
		{ 148, 203, 49 },
		{ 148, 204, 49 },
		{ 148, 205, 49 },
		{ 148, 206, 49 },
		{ 148, 207, 49 },
		{ 148, 208, 49 },
		{ 148, 209, 49 },
		{ 148, 210, 49 },
		{ 148, 211, 49 },
		{ 148, 212, 49 },
		{ 148, 213, 49 },
		{ 148, 214, 49 },
		{ 148, 215, 49 },
		{ 148, 216, 49 },
		{ 148, 217, 49 },
		{ 148, 218, 49 },
		{ 148, 219, 49 },
		{ 148, 220, 49 },
		{ 148, 221, 49 },
		{ 148, 222, 49 },
		{ 148, 223, 49 },
		{ 148, 224, 49 },
		{ 148, 225, 49 },
		{ 148, 226, 49 },
		{ 148, 227, 49 },
		{ 148, 228, 49 },
		{ 148, 229, 49 },
		{ 149, 204, 49 },
		{ 149, 205, 49 },
		{ 149, 206, 49 },
		{ 149, 207, 49 },
		{ 149, 208, 49 },
		{ 149, 209, 49 },
		{ 149, 210, 49 },
		{ 149, 211, 49 },
		{ 149, 212, 49 },
		{ 149, 213, 49 },
		{ 149, 214, 49 },
		{ 149, 215, 49 },
		{ 149, 216, 49 },
		{ 149, 217, 49 },
		{ 149, 218, 49 },
		{ 149, 219, 49 },
		{ 149, 220, 49 },
		{ 149, 221, 49 },
		{ 149, 222, 49 },
		{ 149, 223, 49 },
		{ 149, 224, 49 },
		{ 149, 225, 49 },
		{ 149, 226, 49 },
		{ 149, 227, 49 },
		{ 149, 228, 49 },
		{ 149, 229, 49 },
		{ 150, 204, 49 },
		{ 150, 205, 49 },
		{ 150, 206, 49 },
		{ 150, 207, 49 },
		{ 150, 208, 49 },
		{ 150, 209, 49 },
		{ 150, 210, 49 },
		{ 150, 211, 49 },
		{ 150, 212, 49 },
		{ 150, 213, 49 },
		{ 150, 214, 49 },
		{ 150, 215, 49 },
		{ 150, 216, 49 },
		{ 150, 217, 49 },
		{ 150, 218, 49 },
		{ 150, 219, 49 },
		{ 150, 220, 49 },
		{ 150, 221, 49 },
		{ 150, 222, 49 },
		{ 150, 223, 49 },
		{ 150, 224, 49 },
		{ 150, 225, 49 },
		{ 150, 226, 49 },
		{ 150, 227, 49 },
		{ 150, 228, 49 },
		{ 150, 229, 49 },
		{ 150, 230, 49 },
		{ 151, 205, 49 },
		{ 151, 206, 49 },
		{ 151, 207, 49 },
		{ 151, 208, 49 },
		{ 151, 209, 49 },
		{ 151, 210, 49 },
		{ 151, 211, 49 },
		{ 151, 212, 49 },
		{ 151, 213, 49 },
		{ 151, 214, 49 },
		{ 151, 215, 49 },
		{ 151, 216, 49 },
		{ 151, 217, 49 },
		{ 151, 218, 49 },
		{ 151, 219, 49 },
		{ 151, 220, 49 },
		{ 151, 221, 49 },
		{ 151, 222, 49 },
		{ 151, 223, 49 },
		{ 151, 224, 49 },
		{ 151, 225, 49 },
		{ 151, 226, 49 },
		{ 151, 227, 49 },
		{ 151, 228, 49 },
		{ 151, 229, 49 },
		{ 151, 230, 49 },
		{ 152, 205, 49 },
		{ 152, 206, 49 },
		{ 152, 207, 49 },
		{ 152, 208, 49 },
		{ 152, 209, 49 },
		{ 152, 210, 49 },
		{ 152, 211, 49 },
		{ 152, 212, 49 },
		{ 152, 213, 49 },
		{ 152, 214, 49 },
		{ 152, 215, 49 },
		{ 152, 216, 49 },
		{ 152, 217, 49 },
		{ 152, 218, 49 },
		{ 152, 219, 49 },
		{ 152, 220, 49 },
		{ 152, 221, 49 },
		{ 152, 222, 49 },
		{ 152, 223, 49 },
		{ 152, 224, 49 },
		{ 152, 225, 49 },
		{ 152, 226, 49 },
		{ 152, 227, 49 },
		{ 152, 228, 49 },
		{ 152, 229, 49 },
		{ 152, 230, 49 },
		{ 153, 206, 49 },
		{ 153, 207, 49 },
		{ 153, 208, 49 },
		{ 153, 209, 49 },
		{ 153, 210, 49 },
		{ 153, 211, 49 },
		{ 153, 212, 49 },
		{ 153, 213, 49 },
		{ 153, 214, 49 },
		{ 153, 215, 49 },
		{ 153, 216, 49 },
		{ 153, 217, 49 },
		{ 153, 218, 49 },
		{ 153, 219, 49 },
		{ 153, 220, 49 },
		{ 153, 221, 49 },
		{ 153, 222, 49 },
		{ 153, 223, 49 },
		{ 153, 224, 49 },
		{ 153, 225, 49 },
		{ 153, 226, 49 },
		{ 153, 227, 49 },
		{ 153, 228, 49 },
		{ 153, 229, 49 },
		{ 153, 230, 49 },
		{ 154, 206, 49 },
		{ 154, 207, 49 },
		{ 154, 208, 49 },
		{ 154, 209, 49 },
		{ 154, 210, 49 },
		{ 154, 211, 49 },
		{ 154, 212, 49 },
		{ 154, 213, 49 },
		{ 154, 214, 49 },
		{ 154, 215, 49 },
		{ 154, 216, 49 },
		{ 154, 217, 49 },
		{ 154, 218, 49 },
		{ 154, 219, 49 },
		{ 154, 220, 49 },
		{ 154, 221, 49 },
		{ 154, 222, 49 },
		{ 154, 223, 49 },
		{ 154, 224, 49 },
		{ 154, 225, 49 },
		{ 154, 226, 49 },
		{ 154, 227, 49 },
		{ 154, 228, 49 },
		{ 154, 229, 49 },
		{ 154, 230, 49 },
		{ 154, 231, 49 },
		{ 155, 207, 49 },
		{ 155, 208, 49 },
		{ 155, 209, 49 },
		{ 155, 210, 49 },
		{ 155, 211, 49 },
		{ 155, 212, 49 },
		{ 155, 213, 49 },
		{ 155, 214, 49 },
		{ 155, 215, 49 },
		{ 155, 216, 49 },
		{ 155, 217, 49 },
		{ 155, 218, 49 },
		{ 155, 219, 49 },
		{ 155, 220, 49 },
		{ 155, 221, 49 },
		{ 155, 222, 49 },
		{ 155, 223, 49 },
		{ 155, 224, 49 },
		{ 155, 225, 49 },
		{ 155, 226, 49 },
		{ 155, 227, 49 },
		{ 155, 228, 49 },
		{ 155, 229, 49 },
		{ 155, 230, 49 },
		{ 155, 231, 49 },
		{ 156, 207, 49 },
		{ 156, 208, 49 },
		{ 156, 209, 49 },
		{ 156, 210, 49 },
		{ 156, 211, 49 },
		{ 156, 212, 49 },
		{ 156, 213, 49 },
		{ 156, 214, 49 },
		{ 156, 215, 49 },
		{ 156, 216, 49 },
		{ 156, 217, 49 },
		{ 156, 218, 49 },
		{ 156, 219, 49 },
		{ 156, 220, 49 },
		{ 156, 221, 49 },
		{ 156, 222, 49 },
		{ 156, 223, 49 },
		{ 156, 224, 49 },
		{ 156, 225, 49 },
		{ 156, 226, 49 },
		{ 156, 227, 49 },
		{ 156, 228, 49 },
		{ 156, 229, 49 },
		{ 156, 230, 49 },
		{ 156, 231, 49 },
		{ 157, 208, 49 },
		{ 157, 209, 49 },
		{ 157, 210, 49 },
		{ 157, 211, 49 },
		{ 157, 212, 49 },
		{ 157, 213, 49 },
		{ 157, 214, 49 },
		{ 157, 215, 49 },
		{ 157, 216, 49 },
		{ 157, 217, 49 },
		{ 157, 218, 49 },
		{ 157, 219, 49 },
		{ 157, 220, 49 },
		{ 157, 221, 49 },
		{ 157, 222, 49 },
		{ 157, 223, 49 },
		{ 157, 224, 49 },
		{ 157, 225, 49 },
		{ 157, 226, 49 },
		{ 157, 227, 49 },
		{ 157, 228, 49 },
		{ 157, 229, 49 },
		{ 157, 230, 49 },
		{ 157, 231, 49 },
		{ 158, 208, 49 },
		{ 158, 209, 49 },
		{ 158, 210, 49 },
		{ 158, 211, 49 },
		{ 158, 212, 49 },
		{ 158, 213, 49 },
		{ 158, 214, 49 },
		{ 158, 215, 49 },
		{ 158, 216, 49 },
		{ 158, 217, 49 },
		{ 158, 218, 49 },
		{ 158, 219, 49 },
		{ 158, 220, 49 },
		{ 158, 221, 49 },
		{ 158, 222, 49 },
		{ 158, 223, 49 },
		{ 158, 224, 49 },
		{ 158, 225, 49 },
		{ 158, 226, 49 },
		{ 158, 227, 49 },
		{ 158, 228, 49 },
		{ 158, 229, 49 },
		{ 158, 230, 49 },
		{ 158, 231, 49 },
		{ 158, 232, 49 },
		{ 159, 209, 49 },
		{ 159, 210, 49 },
		{ 159, 211, 49 },
		{ 159, 212, 49 },
		{ 159, 213, 49 },
		{ 159, 214, 49 },
		{ 159, 215, 49 },
		{ 159, 216, 49 },
		{ 159, 217, 49 },
		{ 159, 218, 49 },
		{ 159, 219, 49 },
		{ 159, 220, 49 },
		{ 159, 221, 49 },
		{ 159, 222, 49 },
		{ 159, 223, 49 },
		{ 159, 224, 49 },
		{ 159, 225, 49 },
		{ 159, 226, 49 },
		{ 159, 227, 49 },
		{ 159, 228, 49 },
		{ 159, 229, 49 },
		{ 159, 230, 49 },
		{ 159, 231, 49 },
		{ 159, 232, 49 },
		{ 160, 209, 49 },
		{ 160, 210, 49 },
		{ 160, 211, 49 },
		{ 160, 212, 49 },
		{ 160, 213, 49 },
		{ 160, 214, 49 },
		{ 160, 215, 49 },
		{ 160, 216, 49 },
		{ 160, 217, 49 },
		{ 160, 218, 49 },
		{ 160, 219, 49 },
		{ 160, 220, 49 },
		{ 160, 221, 49 },
		{ 160, 222, 49 },
		{ 160, 223, 49 },
		{ 160, 224, 49 },
		{ 160, 225, 49 },
		{ 160, 226, 49 },
		{ 160, 227, 49 },
		{ 160, 228, 49 },
		{ 160, 229, 49 },
		{ 160, 230, 49 },
		{ 160, 231, 49 },
		{ 160, 232, 49 },
		{ 161, 210, 49 },
		{ 161, 211, 49 },
		{ 161, 212, 49 },
		{ 161, 213, 49 },
		{ 161, 214, 49 },
		{ 161, 215, 49 },
		{ 161, 216, 49 },
		{ 161, 217, 49 },
		{ 161, 218, 49 },
		{ 161, 219, 49 },
		{ 161, 220, 49 },
		{ 161, 221, 49 },
		{ 161, 222, 49 },
		{ 161, 223, 49 },
		{ 161, 224, 49 },
		{ 161, 225, 49 },
		{ 161, 226, 49 },
		{ 161, 227, 49 },
		{ 161, 228, 49 },
		{ 161, 229, 49 },
		{ 161, 230, 49 },
		{ 161, 231, 49 },
		{ 161, 232, 49 },
		{ 162, 210, 49 },
		{ 162, 211, 49 },
		{ 162, 212, 49 },
		{ 162, 213, 49 },
		{ 162, 214, 49 },
		{ 162, 215, 49 },
		{ 162, 216, 49 },
		{ 162, 217, 49 },
		{ 162, 218, 49 },
		{ 162, 219, 49 },
		{ 162, 220, 49 },
		{ 162, 221, 49 },
		{ 162, 222, 49 },
		{ 162, 223, 49 },
		{ 162, 224, 49 },
		{ 162, 225, 49 },
		{ 162, 226, 49 },
		{ 162, 227, 49 },
		{ 162, 228, 49 },
		{ 162, 229, 49 },
		{ 162, 230, 49 },
		{ 162, 231, 49 },
		{ 162, 232, 49 },
		{ 162, 233, 49 },
		{ 163, 211, 49 },
		{ 163, 212, 49 },
		{ 163, 213, 49 },
		{ 163, 214, 49 },
		{ 163, 215, 49 },
		{ 163, 216, 49 },
		{ 163, 217, 49 },
		{ 163, 218, 49 },
		{ 163, 219, 49 },
		{ 163, 220, 49 },
		{ 163, 221, 49 },
		{ 163, 222, 49 },
		{ 163, 223, 49 },
		{ 163, 224, 49 },
		{ 163, 225, 49 },
		{ 163, 226, 49 },
		{ 163, 227, 49 },
		{ 163, 228, 49 },
		{ 163, 229, 49 },
		{ 163, 230, 49 },
		{ 163, 231, 49 },
		{ 163, 232, 49 },
		{ 163, 233, 49 },
		{ 164, 211, 49 },
		{ 164, 212, 49 },
		{ 164, 213, 49 },
		{ 164, 214, 49 },
		{ 164, 215, 49 },
		{ 164, 216, 49 },
		{ 164, 217, 49 },
		{ 164, 218, 49 },
		{ 164, 219, 49 },
		{ 164, 220, 49 },
		{ 164, 221, 49 },
		{ 164, 222, 49 },
		{ 164, 223, 49 },
		{ 164, 224, 49 },
		{ 164, 225, 49 },
		{ 164, 226, 49 },
		{ 164, 227, 49 },
		{ 164, 228, 49 },
		{ 164, 229, 49 },
		{ 164, 230, 49 },
		{ 164, 231, 49 },
		{ 164, 232, 49 },
		{ 164, 233, 49 },
		{ 165, 212, 49 },
		{ 165, 213, 49 },
		{ 165, 214, 49 },
		{ 165, 215, 49 },
		{ 165, 216, 49 },
		{ 165, 217, 49 },
		{ 165, 218, 49 },
		{ 165, 219, 49 },
		{ 165, 220, 49 },
		{ 165, 221, 49 },
		{ 165, 222, 49 },
		{ 165, 223, 49 },
		{ 165, 224, 49 },
		{ 165, 225, 49 },
		{ 165, 226, 49 },
		{ 165, 227, 49 },
		{ 165, 228, 49 },
		{ 165, 229, 49 },
		{ 165, 230, 49 },
		{ 165, 231, 49 },
		{ 165, 232, 49 },
		{ 165, 233, 49 },
		{ 166, 212, 49 },
		{ 166, 213, 49 },
		{ 166, 214, 49 },
		{ 166, 215, 49 },
		{ 166, 216, 49 },
		{ 166, 217, 49 },
		{ 166, 218, 49 },
		{ 166, 219, 49 },
		{ 166, 220, 49 },
		{ 166, 221, 49 },
		{ 166, 222, 49 },
		{ 166, 223, 49 },
		{ 166, 224, 49 },
		{ 166, 225, 49 },
		{ 166, 226, 49 },
		{ 166, 227, 49 },
		{ 166, 228, 49 },
		{ 166, 229, 49 },
		{ 166, 230, 49 },
		{ 166, 231, 49 },
		{ 166, 232, 49 },
		{ 166, 233, 49 },
		{ 166, 234, 49 },
		{ 167, 213, 49 },
		{ 167, 214, 49 },
		{ 167, 215, 49 },
		{ 167, 216, 49 },
		{ 167, 217, 49 },
		{ 167, 218, 49 },
		{ 167, 219, 49 },
		{ 167, 220, 49 },
		{ 167, 221, 49 },
		{ 167, 222, 49 },
		{ 167, 223, 49 },
		{ 167, 224, 49 },
		{ 167, 225, 49 },
		{ 167, 226, 49 },
		{ 167, 227, 49 },
		{ 167, 228, 49 },
		{ 167, 229, 49 },
		{ 167, 230, 49 },
		{ 167, 231, 49 },
		{ 167, 232, 49 },
		{ 167, 233, 49 },
		{ 167, 234, 49 },
		{ 168, 213, 49 },
		{ 168, 214, 49 },
		{ 168, 215, 49 },
		{ 168, 216, 49 },
		{ 168, 217, 49 },
		{ 168, 218, 49 },
		{ 168, 219, 49 },
		{ 168, 220, 49 },
		{ 168, 221, 49 },
		{ 168, 222, 49 },
		{ 168, 223, 49 },
		{ 168, 224, 49 },
		{ 168, 225, 49 },
		{ 168, 226, 49 },
		{ 168, 227, 49 },
		{ 168, 228, 49 },
		{ 168, 229, 49 },
		{ 168, 230, 49 },
		{ 168, 231, 49 },
		{ 168, 232, 49 },
		{ 168, 233, 49 },
		{ 168, 234, 49 },
		{ 169, 214, 49 },
		{ 169, 215, 49 },
		{ 169, 216, 49 },
		{ 169, 217, 49 },
		{ 169, 218, 49 },
		{ 169, 219, 49 },
		{ 169, 220, 49 },
		{ 169, 221, 49 },
		{ 169, 222, 49 },
		{ 169, 223, 49 },
		{ 169, 224, 49 },
		{ 169, 225, 49 },
		{ 169, 226, 49 },
		{ 169, 227, 49 },
		{ 169, 228, 49 },
		{ 169, 229, 49 },
		{ 169, 230, 49 },
		{ 169, 231, 49 },
		{ 169, 232, 49 },
		{ 169, 233, 49 },
		{ 169, 234, 49 },
		{ 170, 214, 49 },
		{ 170, 215, 49 },
		{ 170, 216, 49 },
		{ 170, 217, 49 },
		{ 170, 218, 49 },
		{ 170, 219, 49 },
		{ 170, 220, 49 },
		{ 170, 221, 49 },
		{ 170, 222, 49 },
		{ 170, 223, 49 },
		{ 170, 224, 49 },
		{ 170, 225, 49 },
		{ 170, 226, 49 },
		{ 170, 227, 49 },
		{ 170, 228, 49 },
		{ 170, 229, 49 },
		{ 170, 230, 49 },
		{ 170, 231, 49 },
		{ 170, 232, 49 },
		{ 170, 233, 49 },
		{ 170, 234, 49 },
		{ 170, 235, 49 },
		{ 171, 215, 49 },
		{ 171, 216, 49 },
		{ 171, 217, 49 },
		{ 171, 218, 49 },
		{ 171, 219, 49 },
		{ 171, 220, 49 },
		{ 171, 221, 49 },
		{ 171, 222, 49 },
		{ 171, 223, 49 },
		{ 171, 224, 49 },
		{ 171, 225, 49 },
		{ 171, 226, 49 },
		{ 171, 227, 49 },
		{ 171, 228, 49 },
		{ 171, 229, 49 },
		{ 171, 230, 49 },
		{ 171, 231, 49 },
		{ 171, 232, 49 },
		{ 171, 233, 49 },
		{ 171, 234, 49 },
		{ 171, 235, 49 },
		{ 172, 215, 49 },
		{ 172, 216, 49 },
		{ 172, 217, 49 },
		{ 172, 218, 49 },
		{ 172, 219, 49 },
		{ 172, 220, 49 },
		{ 172, 221, 49 },
		{ 172, 222, 49 },
		{ 172, 223, 49 },
		{ 172, 224, 49 },
		{ 172, 225, 49 },
		{ 172, 226, 49 },
		{ 172, 227, 49 },
		{ 172, 228, 49 },
		{ 172, 229, 49 },
		{ 172, 230, 49 },
		{ 172, 231, 49 },
		{ 172, 232, 49 },
		{ 172, 233, 49 },
		{ 172, 234, 49 },
		{ 172, 235, 49 },
		{ 173, 216, 49 },
		{ 173, 217, 49 },
		{ 173, 218, 49 },
		{ 173, 219, 49 },
		{ 173, 220, 49 },
		{ 173, 221, 49 },
		{ 173, 222, 49 },
		{ 173, 223, 49 },
		{ 173, 224, 49 },
		{ 173, 225, 49 },
		{ 173, 226, 49 },
		{ 173, 227, 49 },
		{ 173, 228, 49 },
		{ 173, 229, 49 },
		{ 173, 230, 49 },
		{ 173, 231, 49 },
		{ 173, 232, 49 },
		{ 173, 233, 49 },
		{ 173, 234, 49 },
		{ 173, 235, 49 },
		{ 174, 216, 49 },
		{ 174, 217, 49 },
		{ 174, 218, 49 },
		{ 174, 219, 49 },
		{ 174, 220, 49 },
		{ 174, 221, 49 },
		{ 174, 222, 49 },
		{ 174, 223, 49 },
		{ 174, 224, 49 },
		{ 174, 225, 49 },
		{ 174, 226, 49 },
		{ 174, 227, 49 },
		{ 174, 228, 49 },
		{ 174, 229, 49 },
		{ 174, 230, 49 },
		{ 174, 231, 49 },
		{ 174, 232, 49 },
		{ 174, 233, 49 },
		{ 174, 234, 49 },
		{ 174, 235, 49 },
		{ 174, 236, 49 },
		{ 175, 217, 49 },
		{ 175, 218, 49 },
		{ 175, 219, 49 },
		{ 175, 220, 49 },
		{ 175, 221, 49 },
		{ 175, 222, 49 },
		{ 175, 223, 49 },
		{ 175, 224, 49 },
		{ 175, 225, 49 },
		{ 175, 226, 49 },
		{ 175, 227, 49 },
		{ 175, 228, 49 },
		{ 175, 229, 49 },
		{ 175, 230, 49 },
		{ 175, 231, 49 },
		{ 175, 232, 49 },
		{ 175, 233, 49 },
		{ 175, 234, 49 },
		{ 175, 235, 49 },
		{ 175, 236, 49 },
		{ 176, 217, 49 },
		{ 176, 218, 49 },
		{ 176, 219, 49 },
		{ 176, 220, 49 },
		{ 176, 221, 49 },
		{ 176, 222, 49 },
		{ 176, 223, 49 },
		{ 176, 224, 49 },
		{ 176, 225, 49 },
		{ 176, 226, 49 },
		{ 176, 227, 49 },
		{ 176, 228, 49 },
		{ 176, 229, 49 },
		{ 176, 230, 49 },
		{ 176, 231, 49 },
		{ 176, 232, 49 },
		{ 176, 233, 49 },
		{ 176, 234, 49 },
		{ 176, 235, 49 },
		{ 176, 236, 49 },
		{ 177, 218, 49 },
		{ 177, 219, 49 },
		{ 177, 220, 49 },
		{ 177, 221, 49 },
		{ 177, 222, 49 },
		{ 177, 223, 49 },
		{ 177, 224, 49 },
		{ 177, 225, 49 },
		{ 177, 226, 49 },
		{ 177, 227, 49 },
		{ 177, 228, 49 },
		{ 177, 229, 49 },
		{ 177, 230, 49 },
		{ 177, 231, 49 },
		{ 177, 232, 49 },
		{ 177, 233, 49 },
		{ 177, 234, 49 },
		{ 177, 235, 49 },
		{ 177, 236, 49 },
		{ 178, 218, 49 },
		{ 178, 219, 49 },
		{ 178, 220, 49 },
		{ 178, 221, 49 },
		{ 178, 222, 49 },
		{ 178, 223, 49 },
		{ 178, 224, 49 },
		{ 178, 225, 49 },
		{ 178, 226, 49 },
		{ 178, 227, 49 },
		{ 178, 228, 49 },
		{ 178, 229, 49 },
		{ 178, 230, 49 },
		{ 178, 231, 49 },
		{ 178, 232, 49 },
		{ 178, 233, 49 },
		{ 178, 234, 49 },
		{ 178, 235, 49 },
		{ 178, 236, 49 },
		{ 178, 237, 49 },
		{ 179, 219, 49 },
		{ 179, 220, 49 },
		{ 179, 221, 49 },
		{ 179, 222, 49 },
		{ 179, 223, 49 },
		{ 179, 224, 49 },
		{ 179, 225, 49 },
		{ 179, 226, 49 },
		{ 179, 227, 49 },
		{ 179, 228, 49 },
		{ 179, 229, 49 },
		{ 179, 230, 49 },
		{ 179, 231, 49 },
		{ 179, 232, 49 },
		{ 179, 233, 49 },
		{ 179, 234, 49 },
		{ 179, 235, 49 },
		{ 179, 236, 49 },
		{ 179, 237, 49 },
		{ 180, 219, 49 },
		{ 180, 220, 49 },
		{ 180, 221, 49 },
		{ 180, 222, 49 },
		{ 180, 223, 49 },
		{ 180, 224, 49 },
		{ 180, 225, 49 },
		{ 180, 226, 49 },
		{ 180, 227, 49 },
		{ 180, 228, 49 },
		{ 180, 229, 49 },
		{ 180, 230, 49 },
		{ 180, 231, 49 },
		{ 180, 232, 49 },
		{ 180, 233, 49 },
		{ 180, 234, 49 },
		{ 180, 235, 49 },
		{ 180, 236, 49 },
		{ 180, 237, 49 },
		{ 181, 220, 49 },
		{ 181, 221, 49 },
		{ 181, 222, 49 },
		{ 181, 223, 49 },
		{ 181, 224, 49 },
		{ 181, 225, 49 },
		{ 181, 226, 49 },
		{ 181, 227, 49 },
		{ 181, 228, 49 },
		{ 181, 229, 49 },
		{ 181, 230, 49 },
		{ 181, 231, 49 },
		{ 181, 232, 49 },
		{ 181, 233, 49 },
		{ 181, 234, 49 },
		{ 181, 235, 49 },
		{ 181, 236, 49 },
		{ 181, 237, 49 },
		{ 182, 220, 49 },
		{ 182, 221, 49 },
		{ 182, 222, 49 },
		{ 182, 223, 49 },
		{ 182, 224, 49 },
		{ 182, 225, 49 },
		{ 182, 226, 49 },
		{ 182, 227, 49 },
		{ 182, 228, 49 },
		{ 182, 229, 49 },
		{ 182, 230, 49 },
		{ 182, 231, 49 },
		{ 182, 232, 49 },
		{ 182, 233, 49 },
		{ 182, 234, 49 },
		{ 182, 235, 49 },
		{ 182, 236, 49 },
		{ 182, 237, 49 },
		{ 182, 238, 49 },
		{ 183, 221, 49 },
		{ 183, 222, 49 },
		{ 183, 223, 49 },
		{ 183, 224, 49 },
		{ 183, 225, 49 },
		{ 183, 226, 49 },
		{ 183, 227, 49 },
		{ 183, 228, 49 },
		{ 183, 229, 49 },
		{ 183, 230, 49 },
		{ 183, 231, 49 },
		{ 183, 232, 49 },
		{ 183, 233, 49 },
		{ 183, 234, 49 },
		{ 183, 235, 49 },
		{ 183, 236, 49 },
		{ 183, 237, 49 },
		{ 183, 238, 49 },
		{ 184, 221, 49 },
		{ 184, 222, 49 },
		{ 184, 223, 49 },
		{ 184, 224, 49 },
		{ 184, 225, 49 },
		{ 184, 226, 49 },
		{ 184, 227, 49 },
		{ 184, 228, 49 },
		{ 184, 229, 49 },
		{ 184, 230, 49 },
		{ 184, 231, 49 },
		{ 184, 232, 49 },
		{ 184, 233, 49 },
		{ 184, 234, 49 },
		{ 184, 235, 49 },
		{ 184, 236, 49 },
		{ 184, 237, 49 },
		{ 184, 238, 49 },
		{ 185, 222, 49 },
		{ 185, 223, 49 },
		{ 185, 224, 49 },
		{ 185, 225, 49 },
		{ 185, 226, 49 },
		{ 185, 227, 49 },
		{ 185, 228, 49 },
		{ 185, 229, 49 },
		{ 185, 230, 49 },
		{ 185, 231, 49 },
		{ 185, 232, 49 },
		{ 185, 233, 49 },
		{ 185, 234, 49 },
		{ 185, 235, 49 },
		{ 185, 236, 49 },
		{ 185, 237, 49 },
		{ 185, 238, 49 },
		{ 186, 222, 49 },
		{ 186, 223, 49 },
		{ 186, 224, 49 },
		{ 186, 225, 49 },
		{ 186, 226, 49 },
		{ 186, 227, 49 },
		{ 186, 228, 49 },
		{ 186, 229, 49 },
		{ 186, 230, 49 },
		{ 186, 231, 49 },
		{ 186, 232, 49 },
		{ 186, 233, 49 },
		{ 186, 234, 49 },
		{ 186, 235, 49 },
		{ 186, 236, 49 },
		{ 186, 237, 49 },
		{ 186, 238, 49 },
		{ 186, 239, 49 },
		{ 187, 223, 49 },
		{ 187, 224, 49 },
		{ 187, 225, 49 },
		{ 187, 226, 49 },
		{ 187, 227, 49 },
		{ 187, 228, 49 },
		{ 187, 229, 49 },
		{ 187, 230, 49 },
		{ 187, 231, 49 },
		{ 187, 232, 49 },
		{ 187, 233, 49 },
		{ 187, 234, 49 },
		{ 187, 235, 49 },
		{ 187, 236, 49 },
		{ 187, 237, 49 },
		{ 187, 238, 49 },
		{ 187, 239, 49 },
		{ 188, 223, 49 },
		{ 188, 224, 49 },
		{ 188, 225, 49 },
		{ 188, 226, 49 },
		{ 188, 227, 49 },
		{ 188, 228, 49 },
		{ 188, 229, 49 },
		{ 188, 230, 49 },
		{ 188, 231, 49 },
		{ 188, 232, 49 },
		{ 188, 233, 49 },
		{ 188, 234, 49 },
		{ 188, 235, 49 },
		{ 188, 236, 49 },
		{ 188, 237, 49 },
		{ 188, 238, 49 },
		{ 188, 239, 49 },
		{ 189, 224, 49 },
		{ 189, 225, 49 },
		{ 189, 226, 49 },
		{ 189, 227, 49 },
		{ 189, 228, 49 },
		{ 189, 229, 49 },
		{ 189, 230, 49 },
		{ 189, 231, 49 },
		{ 189, 232, 49 },
		{ 189, 233, 49 },
		{ 189, 234, 49 },
		{ 189, 235, 49 },
		{ 189, 236, 49 },
		{ 189, 237, 49 },
		{ 189, 238, 49 },
		{ 189, 239, 49 },
		{ 190, 225, 49 },
		{ 190, 226, 49 },
		{ 190, 227, 49 },
		{ 190, 228, 49 },
		{ 190, 229, 49 },
		{ 190, 230, 49 },
		{ 190, 231, 49 },
		{ 190, 232, 49 },
		{ 190, 233, 49 },
		{ 190, 234, 49 },
		{ 190, 235, 49 },
		{ 190, 236, 49 },
		{ 190, 237, 49 },
		{ 190, 238, 49 },
		{ 190, 239, 49 },
		{ 191, 226, 49 },
		{ 191, 227, 49 },
		{ 191, 228, 49 },
		{ 191, 229, 49 },
		{ 191, 230, 49 },
		{ 191, 231, 49 },
		{ 191, 232, 49 },
		{ 191, 233, 49 },
		{ 191, 234, 49 },
		{ 191, 235, 49 },
		{ 191, 236, 49 },
		{ 191, 237, 49 },
		{ 191, 238, 49 },
		{ 323, 276, 49 },
		{ 323, 277, 49 },
		{ 323, 278, 49 },
		{ 323, 279, 49 },
		{ 323, 280, 49 },
		{ 323, 281, 49 },
		{ 323, 282, 49 },
		{ 323, 283, 49 },
		{ 323, 284, 49 },
		{ 323, 285, 49 },
		{ 323, 286, 49 },
		{ 323, 287, 49 },
		{ 323, 288, 49 },
		{ 324, 275, 49 },
		{ 324, 276, 49 },
		{ 324, 277, 49 },
		{ 324, 278, 49 },
		{ 324, 279, 49 },
		{ 324, 280, 49 },
		{ 324, 281, 49 },
		{ 324, 282, 49 },
		{ 324, 283, 49 },
		{ 324, 284, 49 },
		{ 324, 285, 49 },
		{ 324, 286, 49 },
		{ 324, 287, 49 },
		{ 324, 288, 49 },
		{ 324, 289, 49 },
		{ 325, 275, 49 },
		{ 325, 276, 49 },
		{ 325, 277, 49 },
		{ 325, 278, 49 },
		{ 325, 279, 49 },
		{ 325, 280, 49 },
		{ 325, 281, 49 },
		{ 325, 282, 49 },
		{ 325, 283, 49 },
		{ 325, 284, 49 },
		{ 325, 285, 49 },
		{ 325, 286, 49 },
		{ 325, 287, 49 },
		{ 325, 288, 49 },
		{ 325, 289, 49 },
		{ 325, 290, 49 },
		{ 326, 275, 49 },
		{ 326, 276, 49 },
		{ 326, 277, 49 },
		{ 326, 278, 49 },
		{ 326, 279, 49 },
		{ 326, 280, 49 },
		{ 326, 281, 49 },
		{ 326, 282, 49 },
		{ 326, 283, 49 },
		{ 326, 284, 49 },
		{ 326, 285, 49 },
		{ 326, 286, 49 },
		{ 326, 287, 49 },
		{ 326, 288, 49 },
		{ 326, 289, 49 },
		{ 326, 290, 49 },
		{ 326, 291, 49 },
		{ 327, 275, 49 },
		{ 327, 276, 49 },
		{ 327, 277, 49 },
		{ 327, 278, 49 },
		{ 327, 279, 49 },
		{ 327, 280, 49 },
		{ 327, 281, 49 },
		{ 327, 282, 49 },
		{ 327, 283, 49 },
		{ 327, 284, 49 },
		{ 327, 285, 49 },
		{ 327, 286, 49 },
		{ 327, 287, 49 },
		{ 327, 288, 49 },
		{ 327, 289, 49 },
		{ 327, 290, 49 },
		{ 327, 291, 49 },
		{ 328, 275, 49 },
		{ 328, 276, 49 },
		{ 328, 277, 49 },
		{ 328, 278, 49 },
		{ 328, 279, 49 },
		{ 328, 280, 49 },
		{ 328, 281, 49 },
		{ 328, 282, 49 },
		{ 328, 283, 49 },
		{ 328, 284, 49 },
		{ 328, 285, 49 },
		{ 328, 286, 49 },
		{ 328, 287, 49 },
		{ 328, 288, 49 },
		{ 328, 289, 49 },
		{ 328, 290, 49 },
		{ 328, 291, 49 },
		{ 328, 292, 49 },
		{ 329, 276, 49 },
		{ 329, 277, 49 },
		{ 329, 278, 49 },
		{ 329, 279, 49 },
		{ 329, 280, 49 },
		{ 329, 281, 49 },
		{ 329, 282, 49 },
		{ 329, 283, 49 },
		{ 329, 284, 49 },
		{ 329, 285, 49 },
		{ 329, 286, 49 },
		{ 329, 287, 49 },
		{ 329, 288, 49 },
		{ 329, 289, 49 },
		{ 329, 290, 49 },
		{ 329, 291, 49 },
		{ 329, 292, 49 },
		{ 330, 276, 49 },
		{ 330, 277, 49 },
		{ 330, 278, 49 },
		{ 330, 279, 49 },
		{ 330, 280, 49 },
		{ 330, 281, 49 },
		{ 330, 282, 49 },
		{ 330, 283, 49 },
		{ 330, 284, 49 },
		{ 330, 285, 49 },
		{ 330, 286, 49 },
		{ 330, 287, 49 },
		{ 330, 288, 49 },
		{ 330, 289, 49 },
		{ 330, 290, 49 },
		{ 330, 291, 49 },
		{ 330, 292, 49 },
		{ 330, 293, 49 },
		{ 331, 276, 49 },
		{ 331, 277, 49 },
		{ 331, 278, 49 },
		{ 331, 279, 49 },
		{ 331, 280, 49 },
		{ 331, 281, 49 },
		{ 331, 282, 49 },
		{ 331, 283, 49 },
		{ 331, 284, 49 },
		{ 331, 285, 49 },
		{ 331, 286, 49 },
		{ 331, 287, 49 },
		{ 331, 288, 49 },
		{ 331, 289, 49 },
		{ 331, 290, 49 },
		{ 331, 291, 49 },
		{ 331, 292, 49 },
		{ 331, 293, 49 },
		{ 332, 276, 49 },
		{ 332, 277, 49 },
		{ 332, 278, 49 },
		{ 332, 279, 49 },
		{ 332, 280, 49 },
		{ 332, 281, 49 },
		{ 332, 282, 49 },
		{ 332, 283, 49 },
		{ 332, 284, 49 },
		{ 332, 285, 49 },
		{ 332, 286, 49 },
		{ 332, 287, 49 },
		{ 332, 288, 49 },
		{ 332, 289, 49 },
		{ 332, 290, 49 },
		{ 332, 291, 49 },
		{ 332, 292, 49 },
		{ 332, 293, 49 },
		{ 332, 294, 49 },
		{ 333, 277, 49 },
		{ 333, 278, 49 },
		{ 333, 279, 49 },
		{ 333, 280, 49 },
		{ 333, 281, 49 },
		{ 333, 282, 49 },
		{ 333, 283, 49 },
		{ 333, 284, 49 },
		{ 333, 285, 49 },
		{ 333, 286, 49 },
		{ 333, 287, 49 },
		{ 333, 288, 49 },
		{ 333, 289, 49 },
		{ 333, 290, 49 },
		{ 333, 291, 49 },
		{ 333, 292, 49 },
		{ 333, 293, 49 },
		{ 333, 294, 49 },
		{ 334, 277, 49 },
		{ 334, 278, 49 },
		{ 334, 279, 49 },
		{ 334, 280, 49 },
		{ 334, 281, 49 },
		{ 334, 282, 49 },
		{ 334, 283, 49 },
		{ 334, 284, 49 },
		{ 334, 285, 49 },
		{ 334, 286, 49 },
		{ 334, 287, 49 },
		{ 334, 288, 49 },
		{ 334, 289, 49 },
		{ 334, 290, 49 },
		{ 334, 291, 49 },
		{ 334, 292, 49 },
		{ 334, 293, 49 },
		{ 334, 294, 49 },
		{ 334, 295, 49 },
		{ 335, 277, 49 },
		{ 335, 278, 49 },
		{ 335, 279, 49 },
		{ 335, 280, 49 },
		{ 335, 281, 49 },
		{ 335, 282, 49 },
		{ 335, 283, 49 },
		{ 335, 284, 49 },
		{ 335, 285, 49 },
		{ 335, 286, 49 },
		{ 335, 287, 49 },
		{ 335, 288, 49 },
		{ 335, 289, 49 },
		{ 335, 290, 49 },
		{ 335, 291, 49 },
		{ 335, 292, 49 },
		{ 335, 293, 49 },
		{ 335, 294, 49 },
		{ 335, 295, 49 },
		{ 336, 277, 49 },
		{ 336, 278, 49 },
		{ 336, 279, 49 },
		{ 336, 280, 49 },
		{ 336, 281, 49 },
		{ 336, 282, 49 },
		{ 336, 283, 49 },
		{ 336, 284, 49 },
		{ 336, 285, 49 },
		{ 336, 286, 49 },
		{ 336, 287, 49 },
		{ 336, 288, 49 },
		{ 336, 289, 49 },
		{ 336, 290, 49 },
		{ 336, 291, 49 },
		{ 336, 292, 49 },
		{ 336, 293, 49 },
		{ 336, 294, 49 },
		{ 336, 295, 49 },
		{ 336, 296, 49 },
		{ 337, 278, 49 },
		{ 337, 279, 49 },
		{ 337, 280, 49 },
		{ 337, 281, 49 },
		{ 337, 282, 49 },
		{ 337, 283, 49 },
		{ 337, 284, 49 },
		{ 337, 285, 49 },
		{ 337, 286, 49 },
		{ 337, 287, 49 },
		{ 337, 288, 49 },
		{ 337, 289, 49 },
		{ 337, 290, 49 },
		{ 337, 291, 49 },
		{ 337, 292, 49 },
		{ 337, 293, 49 },
		{ 337, 294, 49 },
		{ 337, 295, 49 },
		{ 337, 296, 49 },
		{ 338, 278, 49 },
		{ 338, 279, 49 },
		{ 338, 280, 49 },
		{ 338, 281, 49 },
		{ 338, 282, 49 },
		{ 338, 283, 49 },
		{ 338, 284, 49 },
		{ 338, 285, 49 },
		{ 338, 286, 49 },
		{ 338, 287, 49 },
		{ 338, 288, 49 },
		{ 338, 289, 49 },
		{ 338, 290, 49 },
		{ 338, 291, 49 },
		{ 338, 292, 49 },
		{ 338, 293, 49 },
		{ 338, 294, 49 },
		{ 338, 295, 49 },
		{ 338, 296, 49 },
		{ 338, 297, 49 },
		{ 339, 278, 49 },
		{ 339, 279, 49 },
		{ 339, 280, 49 },
		{ 339, 281, 49 },
		{ 339, 282, 49 },
		{ 339, 283, 49 },
		{ 339, 284, 49 },
		{ 339, 285, 49 },
		{ 339, 286, 49 },
		{ 339, 287, 49 },
		{ 339, 288, 49 },
		{ 339, 289, 49 },
		{ 339, 290, 49 },
		{ 339, 291, 49 },
		{ 339, 292, 49 },
		{ 339, 293, 49 },
		{ 339, 294, 49 },
		{ 339, 295, 49 },
		{ 339, 296, 49 },
		{ 339, 297, 49 },
		{ 340, 278, 49 },
		{ 340, 279, 49 },
		{ 340, 280, 49 },
		{ 340, 281, 49 },
		{ 340, 282, 49 },
		{ 340, 283, 49 },
		{ 340, 284, 49 },
		{ 340, 285, 49 },
		{ 340, 286, 49 },
		{ 340, 287, 49 },
		{ 340, 288, 49 },
		{ 340, 289, 49 },
		{ 340, 290, 49 },
		{ 340, 291, 49 },
		{ 340, 292, 49 },
		{ 340, 293, 49 },
		{ 340, 294, 49 },
		{ 340, 295, 49 },
		{ 340, 296, 49 },
		{ 340, 297, 49 },
		{ 340, 298, 49 },
		{ 341, 279, 49 },
		{ 341, 280, 49 },
		{ 341, 281, 49 },
		{ 341, 282, 49 },
		{ 341, 283, 49 },
		{ 341, 284, 49 },
		{ 341, 285, 49 },
		{ 341, 286, 49 },
		{ 341, 287, 49 },
		{ 341, 288, 49 },
		{ 341, 289, 49 },
		{ 341, 290, 49 },
		{ 341, 291, 49 },
		{ 341, 292, 49 },
		{ 341, 293, 49 },
		{ 341, 294, 49 },
		{ 341, 295, 49 },
		{ 341, 296, 49 },
		{ 341, 297, 49 },
		{ 341, 298, 49 },
		{ 342, 279, 49 },
		{ 342, 280, 49 },
		{ 342, 281, 49 },
		{ 342, 282, 49 },
		{ 342, 283, 49 },
		{ 342, 284, 49 },
		{ 342, 285, 49 },
		{ 342, 286, 49 },
		{ 342, 287, 49 },
		{ 342, 288, 49 },
		{ 342, 289, 49 },
		{ 342, 290, 49 },
		{ 342, 291, 49 },
		{ 342, 292, 49 },
		{ 342, 293, 49 },
		{ 342, 294, 49 },
		{ 342, 295, 49 },
		{ 342, 296, 49 },
		{ 342, 297, 49 },
		{ 342, 298, 49 },
		{ 342, 299, 49 },
		{ 343, 279, 49 },
		{ 343, 280, 49 },
		{ 343, 281, 49 },
		{ 343, 282, 49 },
		{ 343, 283, 49 },
		{ 343, 284, 49 },
		{ 343, 285, 49 },
		{ 343, 286, 49 },
		{ 343, 287, 49 },
		{ 343, 288, 49 },
		{ 343, 289, 49 },
		{ 343, 290, 49 },
		{ 343, 291, 49 },
		{ 343, 292, 49 },
		{ 343, 293, 49 },
		{ 343, 294, 49 },
		{ 343, 295, 49 },
		{ 343, 296, 49 },
		{ 343, 297, 49 },
		{ 343, 298, 49 },
		{ 343, 299, 49 },
		{ 344, 279, 49 },
		{ 344, 280, 49 },
		{ 344, 281, 49 },
		{ 344, 282, 49 },
		{ 344, 283, 49 },
		{ 344, 284, 49 },
		{ 344, 285, 49 },
		{ 344, 286, 49 },
		{ 344, 287, 49 },
		{ 344, 288, 49 },
		{ 344, 289, 49 },
		{ 344, 290, 49 },
		{ 344, 291, 49 },
		{ 344, 292, 49 },
		{ 344, 293, 49 },
		{ 344, 294, 49 },
		{ 344, 295, 49 },
		{ 344, 296, 49 },
		{ 344, 297, 49 },
		{ 344, 298, 49 },
		{ 344, 299, 49 },
		{ 344, 300, 49 },
		{ 345, 280, 49 },
		{ 345, 281, 49 },
		{ 345, 282, 49 },
		{ 345, 283, 49 },
		{ 345, 284, 49 },
		{ 345, 285, 49 },
		{ 345, 286, 49 },
		{ 345, 287, 49 },
		{ 345, 288, 49 },
		{ 345, 289, 49 },
		{ 345, 290, 49 },
		{ 345, 291, 49 },
		{ 345, 292, 49 },
		{ 345, 293, 49 },
		{ 345, 294, 49 },
		{ 345, 295, 49 },
		{ 345, 296, 49 },
		{ 345, 297, 49 },
		{ 345, 298, 49 },
		{ 345, 299, 49 },
		{ 345, 300, 49 },
		{ 346, 280, 49 },
		{ 346, 281, 49 },
		{ 346, 282, 49 },
		{ 346, 283, 49 },
		{ 346, 284, 49 },
		{ 346, 285, 49 },
		{ 346, 286, 49 },
		{ 346, 287, 49 },
		{ 346, 288, 49 },
		{ 346, 289, 49 },
		{ 346, 290, 49 },
		{ 346, 291, 49 },
		{ 346, 292, 49 },
		{ 346, 293, 49 },
		{ 346, 294, 49 },
		{ 346, 295, 49 },
		{ 346, 296, 49 },
		{ 346, 297, 49 },
		{ 346, 298, 49 },
		{ 346, 299, 49 },
		{ 346, 300, 49 },
		{ 346, 301, 49 },
		{ 347, 280, 49 },
		{ 347, 281, 49 },
		{ 347, 282, 49 },
		{ 347, 283, 49 },
		{ 347, 284, 49 },
		{ 347, 285, 49 },
		{ 347, 286, 49 },
		{ 347, 287, 49 },
		{ 347, 288, 49 },
		{ 347, 289, 49 },
		{ 347, 290, 49 },
		{ 347, 291, 49 },
		{ 347, 292, 49 },
		{ 347, 293, 49 },
		{ 347, 294, 49 },
		{ 347, 295, 49 },
		{ 347, 296, 49 },
		{ 347, 297, 49 },
		{ 347, 298, 49 },
		{ 347, 299, 49 },
		{ 347, 300, 49 },
		{ 347, 301, 49 },
		{ 348, 280, 49 },
		{ 348, 281, 49 },
		{ 348, 282, 49 },
		{ 348, 283, 49 },
		{ 348, 284, 49 },
		{ 348, 285, 49 },
		{ 348, 286, 49 },
		{ 348, 287, 49 },
		{ 348, 288, 49 },
		{ 348, 289, 49 },
		{ 348, 290, 49 },
		{ 348, 291, 49 },
		{ 348, 292, 49 },
		{ 348, 293, 49 },
		{ 348, 294, 49 },
		{ 348, 295, 49 },
		{ 348, 296, 49 },
		{ 348, 297, 49 },
		{ 348, 298, 49 },
		{ 348, 299, 49 },
		{ 348, 300, 49 },
		{ 348, 301, 49 },
		{ 348, 302, 49 },
		{ 349, 281, 49 },
		{ 349, 282, 49 },
		{ 349, 283, 49 },
		{ 349, 284, 49 },
		{ 349, 285, 49 },
		{ 349, 286, 49 },
		{ 349, 287, 49 },
		{ 349, 288, 49 },
		{ 349, 289, 49 },
		{ 349, 290, 49 },
		{ 349, 291, 49 },
		{ 349, 292, 49 },
		{ 349, 293, 49 },
		{ 349, 294, 49 },
		{ 349, 295, 49 },
		{ 349, 296, 49 },
		{ 349, 297, 49 },
		{ 349, 298, 49 },
		{ 349, 299, 49 },
		{ 349, 300, 49 },
		{ 349, 301, 49 },
		{ 349, 302, 49 },
		{ 350, 281, 49 },
		{ 350, 282, 49 },
		{ 350, 283, 49 },
		{ 350, 284, 49 },
		{ 350, 285, 49 },
		{ 350, 286, 49 },
		{ 350, 287, 49 },
		{ 350, 288, 49 },
		{ 350, 289, 49 },
		{ 350, 290, 49 },
		{ 350, 291, 49 },
		{ 350, 292, 49 },
		{ 350, 293, 49 },
		{ 350, 294, 49 },
		{ 350, 295, 49 },
		{ 350, 296, 49 },
		{ 350, 297, 49 },
		{ 350, 298, 49 },
		{ 350, 299, 49 },
		{ 350, 300, 49 },
		{ 350, 301, 49 },
		{ 350, 302, 49 },
		{ 350, 303, 49 },
		{ 351, 281, 49 },
		{ 351, 282, 49 },
		{ 351, 283, 49 },
		{ 351, 284, 49 },
		{ 351, 285, 49 },
		{ 351, 286, 49 },
		{ 351, 287, 49 },
		{ 351, 288, 49 },
		{ 351, 289, 49 },
		{ 351, 290, 49 },
		{ 351, 291, 49 },
		{ 351, 292, 49 },
		{ 351, 293, 49 },
		{ 351, 294, 49 },
		{ 351, 295, 49 },
		{ 351, 296, 49 },
		{ 351, 297, 49 },
		{ 351, 298, 49 },
		{ 351, 299, 49 },
		{ 351, 300, 49 },
		{ 351, 301, 49 },
		{ 351, 302, 49 },
		{ 351, 303, 49 },
		{ 352, 281, 49 },
		{ 352, 282, 49 },
		{ 352, 283, 49 },
		{ 352, 284, 49 },
		{ 352, 285, 49 },
		{ 352, 286, 49 },
		{ 352, 287, 49 },
		{ 352, 288, 49 },
		{ 352, 289, 49 },
		{ 352, 290, 49 },
		{ 352, 291, 49 },
		{ 352, 292, 49 },
		{ 352, 293, 49 },
		{ 352, 294, 49 },
		{ 352, 295, 49 },
		{ 352, 296, 49 },
		{ 352, 297, 49 },
		{ 352, 298, 49 },
		{ 352, 299, 49 },
		{ 352, 300, 49 },
		{ 352, 301, 49 },
		{ 352, 302, 49 },
		{ 352, 303, 49 },
		{ 352, 304, 49 },
		{ 353, 282, 49 },
		{ 353, 283, 49 },
		{ 353, 284, 49 },
		{ 353, 285, 49 },
		{ 353, 286, 49 },
		{ 353, 287, 49 },
		{ 353, 288, 49 },
		{ 353, 289, 49 },
		{ 353, 290, 49 },
		{ 353, 291, 49 },
		{ 353, 292, 49 },
		{ 353, 293, 49 },
		{ 353, 294, 49 },
		{ 353, 295, 49 },
		{ 353, 296, 49 },
		{ 353, 297, 49 },
		{ 353, 298, 49 },
		{ 353, 299, 49 },
		{ 353, 300, 49 },
		{ 353, 301, 49 },
		{ 353, 302, 49 },
		{ 353, 303, 49 },
		{ 353, 304, 49 },
		{ 354, 282, 49 },
		{ 354, 283, 49 },
		{ 354, 284, 49 },
		{ 354, 285, 49 },
		{ 354, 286, 49 },
		{ 354, 287, 49 },
		{ 354, 288, 49 },
		{ 354, 289, 49 },
		{ 354, 290, 49 },
		{ 354, 291, 49 },
		{ 354, 292, 49 },
		{ 354, 293, 49 },
		{ 354, 294, 49 },
		{ 354, 295, 49 },
		{ 354, 296, 49 },
		{ 354, 297, 49 },
		{ 354, 298, 49 },
		{ 354, 299, 49 },
		{ 354, 300, 49 },
		{ 354, 301, 49 },
		{ 354, 302, 49 },
		{ 354, 303, 49 },
		{ 354, 304, 49 },
		{ 354, 305, 49 },
		{ 355, 282, 49 },
		{ 355, 283, 49 },
		{ 355, 284, 49 },
		{ 355, 285, 49 },
		{ 355, 286, 49 },
		{ 355, 287, 49 },
		{ 355, 288, 49 },
		{ 355, 289, 49 },
		{ 355, 290, 49 },
		{ 355, 291, 49 },
		{ 355, 292, 49 },
		{ 355, 293, 49 },
		{ 355, 294, 49 },
		{ 355, 295, 49 },
		{ 355, 296, 49 },
		{ 355, 297, 49 },
		{ 355, 298, 49 },
		{ 355, 299, 49 },
		{ 355, 300, 49 },
		{ 355, 301, 49 },
		{ 355, 302, 49 },
		{ 355, 303, 49 },
		{ 355, 304, 49 },
		{ 355, 305, 49 },
		{ 356, 282, 49 },
		{ 356, 283, 49 },
		{ 356, 284, 49 },
		{ 356, 285, 49 },
		{ 356, 286, 49 },
		{ 356, 287, 49 },
		{ 356, 288, 49 },
		{ 356, 289, 49 },
		{ 356, 290, 49 },
		{ 356, 291, 49 },
		{ 356, 292, 49 },
		{ 356, 293, 49 },
		{ 356, 294, 49 },
		{ 356, 295, 49 },
		{ 356, 296, 49 },
		{ 356, 297, 49 },
		{ 356, 298, 49 },
		{ 356, 299, 49 },
		{ 356, 300, 49 },
		{ 356, 301, 49 },
		{ 356, 302, 49 },
		{ 356, 303, 49 },
		{ 356, 304, 49 },
		{ 356, 305, 49 },
		{ 356, 306, 49 },
		{ 357, 283, 49 },
		{ 357, 284, 49 },
		{ 357, 285, 49 },
		{ 357, 286, 49 },
		{ 357, 287, 49 },
		{ 357, 288, 49 },
		{ 357, 289, 49 },
		{ 357, 290, 49 },
		{ 357, 291, 49 },
		{ 357, 292, 49 },
		{ 357, 293, 49 },
		{ 357, 294, 49 },
		{ 357, 295, 49 },
		{ 357, 296, 49 },
		{ 357, 297, 49 },
		{ 357, 298, 49 },
		{ 357, 299, 49 },
		{ 357, 300, 49 },
		{ 357, 301, 49 },
		{ 357, 302, 49 },
		{ 357, 303, 49 },
		{ 357, 304, 49 },
		{ 357, 305, 49 },
		{ 357, 306, 49 },
		{ 358, 283, 49 },
		{ 358, 284, 49 },
		{ 358, 285, 49 },
		{ 358, 286, 49 },
		{ 358, 287, 49 },
		{ 358, 288, 49 },
		{ 358, 289, 49 },
		{ 358, 290, 49 },
		{ 358, 291, 49 },
		{ 358, 292, 49 },
		{ 358, 293, 49 },
		{ 358, 294, 49 },
		{ 358, 295, 49 },
		{ 358, 296, 49 },
		{ 358, 297, 49 },
		{ 358, 298, 49 },
		{ 358, 299, 49 },
		{ 358, 300, 49 },
		{ 358, 301, 49 },
		{ 358, 302, 49 },
		{ 358, 303, 49 },
		{ 358, 304, 49 },
		{ 358, 305, 49 },
		{ 358, 306, 49 },
		{ 358, 307, 49 },
		{ 359, 283, 49 },
		{ 359, 284, 49 },
		{ 359, 285, 49 },
		{ 359, 286, 49 },
		{ 359, 287, 49 },
		{ 359, 288, 49 },
		{ 359, 289, 49 },
		{ 359, 290, 49 },
		{ 359, 291, 49 },
		{ 359, 292, 49 },
		{ 359, 293, 49 },
		{ 359, 294, 49 },
		{ 359, 295, 49 },
		{ 359, 296, 49 },
		{ 359, 297, 49 },
		{ 359, 298, 49 },
		{ 359, 299, 49 },
		{ 359, 300, 49 },
		{ 359, 301, 49 },
		{ 359, 302, 49 },
		{ 359, 303, 49 },
		{ 359, 304, 49 },
		{ 359, 305, 49 },
		{ 359, 306, 49 },
		{ 359, 307, 49 },
		{ 360, 283, 49 },
		{ 360, 284, 49 },
		{ 360, 285, 49 },
		{ 360, 286, 49 },
		{ 360, 287, 49 },
		{ 360, 288, 49 },
		{ 360, 289, 49 },
		{ 360, 290, 49 },
		{ 360, 291, 49 },
		{ 360, 292, 49 },
		{ 360, 293, 49 },
		{ 360, 294, 49 },
		{ 360, 295, 49 },
		{ 360, 296, 49 },
		{ 360, 297, 49 },
		{ 360, 298, 49 },
		{ 360, 299, 49 },
		{ 360, 300, 49 },
		{ 360, 301, 49 },
		{ 360, 302, 49 },
		{ 360, 303, 49 },
		{ 360, 304, 49 },
		{ 360, 305, 49 },
		{ 360, 306, 49 },
		{ 360, 307, 49 },
		{ 360, 308, 49 },
		{ 361, 284, 49 },
		{ 361, 285, 49 },
		{ 361, 286, 49 },
		{ 361, 287, 49 },
		{ 361, 288, 49 },
		{ 361, 289, 49 },
		{ 361, 290, 49 },
		{ 361, 291, 49 },
		{ 361, 292, 49 },
		{ 361, 293, 49 },
		{ 361, 294, 49 },
		{ 361, 295, 49 },
		{ 361, 296, 49 },
		{ 361, 297, 49 },
		{ 361, 298, 49 },
		{ 361, 299, 49 },
		{ 361, 300, 49 },
		{ 361, 301, 49 },
		{ 361, 302, 49 },
		{ 361, 303, 49 },
		{ 361, 304, 49 },
		{ 361, 305, 49 },
		{ 361, 306, 49 },
		{ 361, 307, 49 },
		{ 361, 308, 49 },
		{ 362, 284, 49 },
		{ 362, 285, 49 },
		{ 362, 286, 49 },
		{ 362, 287, 49 },
		{ 362, 288, 49 },
		{ 362, 289, 49 },
		{ 362, 290, 49 },
		{ 362, 291, 49 },
		{ 362, 292, 49 },
		{ 362, 293, 49 },
		{ 362, 294, 49 },
		{ 362, 295, 49 },
		{ 362, 296, 49 },
		{ 362, 297, 49 },
		{ 362, 298, 49 },
		{ 362, 299, 49 },
		{ 362, 300, 49 },
		{ 362, 301, 49 },
		{ 362, 302, 49 },
		{ 362, 303, 49 },
		{ 362, 304, 49 },
		{ 362, 305, 49 },
		{ 362, 306, 49 },
		{ 362, 307, 49 },
		{ 362, 308, 49 },
		{ 362, 309, 49 },
		{ 363, 284, 49 },
		{ 363, 285, 49 },
		{ 363, 286, 49 },
		{ 363, 287, 49 },
		{ 363, 288, 49 },
		{ 363, 289, 49 },
		{ 363, 290, 49 },
		{ 363, 291, 49 },
		{ 363, 292, 49 },
		{ 363, 293, 49 },
		{ 363, 294, 49 },
		{ 363, 295, 49 },
		{ 363, 296, 49 },
		{ 363, 297, 49 },
		{ 363, 298, 49 },
		{ 363, 299, 49 },
		{ 363, 300, 49 },
		{ 363, 301, 49 },
		{ 363, 302, 49 },
		{ 363, 303, 49 },
		{ 363, 304, 49 },
		{ 363, 305, 49 },
		{ 363, 306, 49 },
		{ 363, 307, 49 },
		{ 363, 308, 49 },
		{ 363, 309, 49 },
		{ 364, 284, 49 },
		{ 364, 285, 49 },
		{ 364, 286, 49 },
		{ 364, 287, 49 },
		{ 364, 288, 49 },
		{ 364, 289, 49 },
		{ 364, 290, 49 },
		{ 364, 291, 49 },
		{ 364, 292, 49 },
		{ 364, 293, 49 },
		{ 364, 294, 49 },
		{ 364, 295, 49 },
		{ 364, 296, 49 },
		{ 364, 297, 49 },
		{ 364, 298, 49 },
		{ 364, 299, 49 },
		{ 364, 300, 49 },
		{ 364, 301, 49 },
		{ 364, 302, 49 },
		{ 364, 303, 49 },
		{ 364, 304, 49 },
		{ 364, 305, 49 },
		{ 364, 306, 49 },
		{ 364, 307, 49 },
		{ 364, 308, 49 },
		{ 364, 309, 49 },
		{ 364, 310, 49 },
		{ 365, 285, 49 },
		{ 365, 286, 49 },
		{ 365, 287, 49 },
		{ 365, 288, 49 },
		{ 365, 289, 49 },
		{ 365, 290, 49 },
		{ 365, 291, 49 },
		{ 365, 292, 49 },
		{ 365, 293, 49 },
		{ 365, 294, 49 },
		{ 365, 295, 49 },
		{ 365, 296, 49 },
		{ 365, 297, 49 },
		{ 365, 298, 49 },
		{ 365, 299, 49 },
		{ 365, 300, 49 },
		{ 365, 301, 49 },
		{ 365, 302, 49 },
		{ 365, 303, 49 },
		{ 365, 304, 49 },
		{ 365, 305, 49 },
		{ 365, 306, 49 },
		{ 365, 307, 49 },
		{ 365, 308, 49 },
		{ 365, 309, 49 },
		{ 365, 310, 49 },
		{ 366, 285, 49 },
		{ 366, 286, 49 },
		{ 366, 287, 49 },
		{ 366, 288, 49 },
		{ 366, 289, 49 },
		{ 366, 290, 49 },
		{ 366, 291, 49 },
		{ 366, 292, 49 },
		{ 366, 293, 49 },
		{ 366, 294, 49 },
		{ 366, 295, 49 },
		{ 366, 296, 49 },
		{ 366, 297, 49 },
		{ 366, 298, 49 },
		{ 366, 299, 49 },
		{ 366, 300, 49 },
		{ 366, 301, 49 },
		{ 366, 302, 49 },
		{ 366, 303, 49 },
		{ 366, 304, 49 },
		{ 366, 305, 49 },
		{ 366, 306, 49 },
		{ 366, 307, 49 },
		{ 366, 308, 49 },
		{ 366, 309, 49 },
		{ 366, 310, 49 },
		{ 366, 311, 49 },
		{ 367, 285, 49 },
		{ 367, 286, 49 },
		{ 367, 287, 49 },
		{ 367, 288, 49 },
		{ 367, 289, 49 },
		{ 367, 290, 49 },
		{ 367, 291, 49 },
		{ 367, 292, 49 },
		{ 367, 293, 49 },
		{ 367, 294, 49 },
		{ 367, 295, 49 },
		{ 367, 296, 49 },
		{ 367, 297, 49 },
		{ 367, 298, 49 },
		{ 367, 299, 49 },
		{ 367, 300, 49 },
		{ 367, 301, 49 },
		{ 367, 302, 49 },
		{ 367, 303, 49 },
		{ 367, 304, 49 },
		{ 367, 305, 49 },
		{ 367, 306, 49 },
		{ 367, 307, 49 },
		{ 367, 308, 49 },
		{ 367, 309, 49 },
		{ 367, 310, 49 },
		{ 367, 311, 49 },
		{ 368, 285, 49 },
		{ 368, 286, 49 },
		{ 368, 287, 49 },
		{ 368, 288, 49 },
		{ 368, 289, 49 },
		{ 368, 290, 49 },
		{ 368, 291, 49 },
		{ 368, 292, 49 },
		{ 368, 293, 49 },
		{ 368, 294, 49 },
		{ 368, 295, 49 },
		{ 368, 296, 49 },
		{ 368, 297, 49 },
		{ 368, 298, 49 },
		{ 368, 299, 49 },
		{ 368, 300, 49 },
		{ 368, 301, 49 },
		{ 368, 302, 49 },
		{ 368, 303, 49 },
		{ 368, 304, 49 },
		{ 368, 305, 49 },
		{ 368, 306, 49 },
		{ 368, 307, 49 },
		{ 368, 308, 49 },
		{ 368, 309, 49 },
		{ 368, 310, 49 },
		{ 368, 311, 49 },
		{ 368, 312, 49 },
		{ 369, 286, 49 },
		{ 369, 287, 49 },
		{ 369, 288, 49 },
		{ 369, 289, 49 },
		{ 369, 290, 49 },
		{ 369, 291, 49 },
		{ 369, 292, 49 },
		{ 369, 293, 49 },
		{ 369, 294, 49 },
		{ 369, 295, 49 },
		{ 369, 296, 49 },
		{ 369, 297, 49 },
		{ 369, 298, 49 },
		{ 369, 299, 49 },
		{ 369, 300, 49 },
		{ 369, 301, 49 },
		{ 369, 302, 49 },
		{ 369, 303, 49 },
		{ 369, 304, 49 },
		{ 369, 305, 49 },
		{ 369, 306, 49 },
		{ 369, 307, 49 },
		{ 369, 308, 49 },
		{ 369, 309, 49 },
		{ 369, 310, 49 },
		{ 369, 311, 49 },
		{ 369, 312, 49 },
		{ 370, 286, 49 },
		{ 370, 287, 49 },
		{ 370, 288, 49 },
		{ 370, 289, 49 },
		{ 370, 290, 49 },
		{ 370, 291, 49 },
		{ 370, 292, 49 },
		{ 370, 293, 49 },
		{ 370, 294, 49 },
		{ 370, 295, 49 },
		{ 370, 296, 49 },
		{ 370, 297, 49 },
		{ 370, 298, 49 },
		{ 370, 299, 49 },
		{ 370, 300, 49 },
		{ 370, 301, 49 },
		{ 370, 302, 49 },
		{ 370, 303, 49 },
		{ 370, 304, 49 },
		{ 370, 305, 49 },
		{ 370, 306, 49 },
		{ 370, 307, 49 },
		{ 370, 308, 49 },
		{ 370, 309, 49 },
		{ 370, 310, 49 },
		{ 370, 311, 49 },
		{ 370, 312, 49 },
		{ 370, 313, 49 },
		{ 371, 286, 49 },
		{ 371, 287, 49 },
		{ 371, 288, 49 },
		{ 371, 289, 49 },
		{ 371, 290, 49 },
		{ 371, 291, 49 },
		{ 371, 292, 49 },
		{ 371, 293, 49 },
		{ 371, 294, 49 },
		{ 371, 295, 49 },
		{ 371, 296, 49 },
		{ 371, 297, 49 },
		{ 371, 298, 49 },
		{ 371, 299, 49 },
		{ 371, 300, 49 },
		{ 371, 301, 49 },
		{ 371, 302, 49 },
		{ 371, 303, 49 },
		{ 371, 304, 49 },
		{ 371, 305, 49 },
		{ 371, 306, 49 },
		{ 371, 307, 49 },
		{ 371, 308, 49 },
		{ 371, 309, 49 },
		{ 371, 310, 49 },
		{ 371, 311, 49 },
		{ 371, 312, 49 },
		{ 371, 313, 49 },
		{ 372, 286, 49 },
		{ 372, 287, 49 },
		{ 372, 288, 49 },
		{ 372, 289, 49 },
		{ 372, 290, 49 },
		{ 372, 291, 49 },
		{ 372, 292, 49 },
		{ 372, 293, 49 },
		{ 372, 294, 49 },
		{ 372, 295, 49 },
		{ 372, 296, 49 },
		{ 372, 297, 49 },
		{ 372, 298, 49 },
		{ 372, 299, 49 },
		{ 372, 300, 49 },
		{ 372, 301, 49 },
		{ 372, 302, 49 },
		{ 372, 303, 49 },
		{ 372, 304, 49 },
		{ 372, 305, 49 },
		{ 372, 306, 49 },
		{ 372, 307, 49 },
		{ 372, 308, 49 },
		{ 372, 309, 49 },
		{ 372, 310, 49 },
		{ 372, 311, 49 },
		{ 372, 312, 49 },
		{ 372, 313, 49 },
		{ 372, 314, 49 },
		{ 373, 287, 49 },
		{ 373, 288, 49 },
		{ 373, 289, 49 },
		{ 373, 290, 49 },
		{ 373, 291, 49 },
		{ 373, 292, 49 },
		{ 373, 293, 49 },
		{ 373, 294, 49 },
		{ 373, 295, 49 },
		{ 373, 296, 49 },
		{ 373, 297, 49 },
		{ 373, 298, 49 },
		{ 373, 299, 49 },
		{ 373, 300, 49 },
		{ 373, 301, 49 },
		{ 373, 302, 49 },
		{ 373, 303, 49 },
		{ 373, 304, 49 },
		{ 373, 305, 49 },
		{ 373, 306, 49 },
		{ 373, 307, 49 },
		{ 373, 308, 49 },
		{ 373, 309, 49 },
		{ 373, 310, 49 },
		{ 373, 311, 49 },
		{ 373, 312, 49 },
		{ 373, 313, 49 },
		{ 373, 314, 49 },
		{ 374, 287, 49 },
		{ 374, 288, 49 },
		{ 374, 289, 49 },
		{ 374, 290, 49 },
		{ 374, 291, 49 },
		{ 374, 292, 49 },
		{ 374, 293, 49 },
		{ 374, 294, 49 },
		{ 374, 295, 49 },
		{ 374, 296, 49 },
		{ 374, 297, 49 },
		{ 374, 298, 49 },
		{ 374, 299, 49 },
		{ 374, 300, 49 },
		{ 374, 301, 49 },
		{ 374, 302, 49 },
		{ 374, 303, 49 },
		{ 374, 304, 49 },
		{ 374, 305, 49 },
		{ 374, 306, 49 },
		{ 374, 307, 49 },
		{ 374, 308, 49 },
		{ 374, 309, 49 },
		{ 374, 310, 49 },
		{ 374, 311, 49 },
		{ 374, 312, 49 },
		{ 374, 313, 49 },
		{ 374, 314, 49 },
		{ 374, 315, 49 },
		{ 375, 287, 49 },
		{ 375, 288, 49 },
		{ 375, 289, 49 },
		{ 375, 290, 49 },
		{ 375, 291, 49 },
		{ 375, 292, 49 },
		{ 375, 293, 49 },
		{ 375, 294, 49 },
		{ 375, 295, 49 },
		{ 375, 296, 49 },
		{ 375, 297, 49 },
		{ 375, 298, 49 },
		{ 375, 299, 49 },
		{ 375, 300, 49 },
		{ 375, 301, 49 },
		{ 375, 302, 49 },
		{ 375, 303, 49 },
		{ 375, 304, 49 },
		{ 375, 305, 49 },
		{ 375, 306, 49 },
		{ 375, 307, 49 },
		{ 375, 308, 49 },
		{ 375, 309, 49 },
		{ 375, 310, 49 },
		{ 375, 311, 49 },
		{ 375, 312, 49 },
		{ 375, 313, 49 },
		{ 375, 314, 49 },
		{ 375, 315, 49 },
		{ 376, 287, 49 },
		{ 376, 288, 49 },
		{ 376, 289, 49 },
		{ 376, 290, 49 },
		{ 376, 291, 49 },
		{ 376, 292, 49 },
		{ 376, 293, 49 },
		{ 376, 294, 49 },
		{ 376, 295, 49 },
		{ 376, 296, 49 },
		{ 376, 297, 49 },
		{ 376, 298, 49 },
		{ 376, 299, 49 },
		{ 376, 300, 49 },
		{ 376, 301, 49 },
		{ 376, 302, 49 },
		{ 376, 303, 49 },
		{ 376, 304, 49 },
		{ 376, 305, 49 },
		{ 376, 306, 49 },
		{ 376, 307, 49 },
		{ 376, 308, 49 },
		{ 376, 309, 49 },
		{ 376, 310, 49 },
		{ 376, 311, 49 },
		{ 376, 312, 49 },
		{ 376, 313, 49 },
		{ 376, 314, 49 },
		{ 376, 315, 49 },
		{ 376, 316, 49 },
		{ 377, 288, 49 },
		{ 377, 289, 49 },
		{ 377, 290, 49 },
		{ 377, 291, 49 },
		{ 377, 292, 49 },
		{ 377, 293, 49 },
		{ 377, 294, 49 },
		{ 377, 295, 49 },
		{ 377, 296, 49 },
		{ 377, 297, 49 },
		{ 377, 298, 49 },
		{ 377, 299, 49 },
		{ 377, 300, 49 },
		{ 377, 301, 49 },
		{ 377, 302, 49 },
		{ 377, 303, 49 },
		{ 377, 304, 49 },
		{ 377, 305, 49 },
		{ 377, 306, 49 },
		{ 377, 307, 49 },
		{ 377, 308, 49 },
		{ 377, 309, 49 },
		{ 377, 310, 49 },
		{ 377, 311, 49 },
		{ 377, 312, 49 },
		{ 377, 313, 49 },
		{ 377, 314, 49 },
		{ 377, 315, 49 },
		{ 377, 316, 49 },
		{ 378, 288, 49 },
		{ 378, 289, 49 },
		{ 378, 290, 49 },
		{ 378, 291, 49 },
		{ 378, 292, 49 },
		{ 378, 293, 49 },
		{ 378, 294, 49 },
		{ 378, 295, 49 },
		{ 378, 296, 49 },
		{ 378, 297, 49 },
		{ 378, 298, 49 },
		{ 378, 299, 49 },
		{ 378, 300, 49 },
		{ 378, 301, 49 },
		{ 378, 302, 49 },
		{ 378, 303, 49 },
		{ 378, 304, 49 },
		{ 378, 305, 49 },
		{ 378, 306, 49 },
		{ 378, 307, 49 },
		{ 378, 308, 49 },
		{ 378, 309, 49 },
		{ 378, 310, 49 },
		{ 378, 311, 49 },
		{ 378, 312, 49 },
		{ 378, 313, 49 },
		{ 378, 314, 49 },
		{ 378, 315, 49 },
		{ 378, 316, 49 },
		{ 378, 317, 49 },
		{ 379, 288, 49 },
		{ 379, 289, 49 },
		{ 379, 290, 49 },
		{ 379, 291, 49 },
		{ 379, 292, 49 },
		{ 379, 293, 49 },
		{ 379, 294, 49 },
		{ 379, 295, 49 },
		{ 379, 296, 49 },
		{ 379, 297, 49 },
		{ 379, 298, 49 },
		{ 379, 299, 49 },
		{ 379, 300, 49 },
		{ 379, 301, 49 },
		{ 379, 302, 49 },
		{ 379, 303, 49 },
		{ 379, 304, 49 },
		{ 379, 305, 49 },
		{ 379, 306, 49 },
		{ 379, 307, 49 },
		{ 379, 308, 49 },
		{ 379, 309, 49 },
		{ 379, 310, 49 },
		{ 379, 311, 49 },
		{ 379, 312, 49 },
		{ 379, 313, 49 },
		{ 379, 314, 49 },
		{ 379, 315, 49 },
		{ 379, 316, 49 },
		{ 379, 317, 49 },
		{ 380, 288, 49 },
		{ 380, 289, 49 },
		{ 380, 290, 49 },
		{ 380, 291, 49 },
		{ 380, 292, 49 },
		{ 380, 293, 49 },
		{ 380, 294, 49 },
		{ 380, 295, 49 },
		{ 380, 296, 49 },
		{ 380, 297, 49 },
		{ 380, 298, 49 },
		{ 380, 299, 49 },
		{ 380, 300, 49 },
		{ 380, 301, 49 },
		{ 380, 302, 49 },
		{ 380, 303, 49 },
		{ 380, 304, 49 },
		{ 380, 305, 49 },
		{ 380, 306, 49 },
		{ 380, 307, 49 },
		{ 380, 308, 49 },
		{ 380, 309, 49 },
		{ 380, 310, 49 },
		{ 380, 311, 49 },
		{ 380, 312, 49 },
		{ 380, 313, 49 },
		{ 380, 314, 49 },
		{ 380, 315, 49 },
		{ 380, 316, 49 },
		{ 380, 317, 49 },
		{ 380, 318, 49 },
		{ 381, 289, 49 },
		{ 381, 290, 49 },
		{ 381, 291, 49 },
		{ 381, 292, 49 },
		{ 381, 293, 49 },
		{ 381, 294, 49 },
		{ 381, 295, 49 },
		{ 381, 296, 49 },
		{ 381, 297, 49 },
		{ 381, 298, 49 },
		{ 381, 299, 49 },
		{ 381, 300, 49 },
		{ 381, 301, 49 },
		{ 381, 302, 49 },
		{ 381, 303, 49 },
		{ 381, 304, 49 },
		{ 381, 305, 49 },
		{ 381, 306, 49 },
		{ 381, 307, 49 },
		{ 381, 308, 49 },
		{ 381, 309, 49 },
		{ 381, 310, 49 },
		{ 381, 311, 49 },
		{ 381, 312, 49 },
		{ 381, 313, 49 },
		{ 381, 314, 49 },
		{ 381, 315, 49 },
		{ 381, 316, 49 },
		{ 381, 317, 49 },
		{ 381, 318, 49 },
		{ 382, 289, 49 },
		{ 382, 290, 49 },
		{ 382, 291, 49 },
		{ 382, 292, 49 },
		{ 382, 293, 49 },
		{ 382, 294, 49 },
		{ 382, 295, 49 },
		{ 382, 296, 49 },
		{ 382, 297, 49 },
		{ 382, 298, 49 },
		{ 382, 299, 49 },
		{ 382, 300, 49 },
		{ 382, 301, 49 },
		{ 382, 302, 49 },
		{ 382, 303, 49 },
		{ 382, 304, 49 },
		{ 382, 305, 49 },
		{ 382, 306, 49 },
		{ 382, 307, 49 },
		{ 382, 308, 49 },
		{ 382, 309, 49 },
		{ 382, 310, 49 },
		{ 382, 311, 49 },
		{ 382, 312, 49 },
		{ 382, 313, 49 },
		{ 382, 314, 49 },
		{ 382, 315, 49 },
		{ 382, 316, 49 },
		{ 382, 317, 49 },
		{ 382, 318, 49 },
		{ 382, 319, 49 },
		{ 383, 289, 49 },
		{ 383, 290, 49 },
		{ 383, 291, 49 },
		{ 383, 292, 49 },
		{ 383, 293, 49 },
		{ 383, 294, 49 },
		{ 383, 295, 49 },
		{ 383, 296, 49 },
		{ 383, 297, 49 },
		{ 383, 298, 49 },
		{ 383, 299, 49 },
		{ 383, 300, 49 },
		{ 383, 301, 49 },
		{ 383, 302, 49 },
		{ 383, 303, 49 },
		{ 383, 304, 49 },
		{ 383, 305, 49 },
		{ 383, 306, 49 },
		{ 383, 307, 49 },
		{ 383, 308, 49 },
		{ 383, 309, 49 },
		{ 383, 310, 49 },
		{ 383, 311, 49 },
		{ 383, 312, 49 },
		{ 383, 313, 49 },
		{ 383, 314, 49 },
		{ 383, 315, 49 },
		{ 383, 316, 49 },
		{ 383, 317, 49 },
		{ 383, 318, 49 },
		{ 383, 319, 49 },
		{ 384, 289, 49 },
		{ 384, 290, 49 },
		{ 384, 291, 49 },
		{ 384, 292, 49 },
		{ 384, 293, 49 },
		{ 384, 294, 49 },
		{ 384, 295, 49 },
		{ 384, 296, 49 },
		{ 384, 297, 49 },
		{ 384, 298, 49 },
		{ 384, 299, 49 },
		{ 384, 300, 49 },
		{ 384, 301, 49 },
		{ 384, 302, 49 },
		{ 384, 303, 49 },
		{ 384, 304, 49 },
		{ 384, 305, 49 },
		{ 384, 306, 49 },
		{ 384, 307, 49 },
		{ 384, 308, 49 },
		{ 384, 309, 49 },
		{ 384, 310, 49 },
		{ 384, 311, 49 },
		{ 384, 312, 49 },
		{ 384, 313, 49 },
		{ 384, 314, 49 },
		{ 384, 315, 49 },
		{ 384, 316, 49 },
		{ 384, 317, 49 },
		{ 384, 318, 49 },
		{ 384, 319, 49 },
		{ 384, 320, 49 },
		{ 385, 290, 49 },
		{ 385, 291, 49 },
		{ 385, 292, 49 },
		{ 385, 293, 49 },
		{ 385, 294, 49 },
		{ 385, 295, 49 },
		{ 385, 296, 49 },
		{ 385, 297, 49 },
		{ 385, 298, 49 },
		{ 385, 299, 49 },
		{ 385, 300, 49 },
		{ 385, 301, 49 },
		{ 385, 302, 49 },
		{ 385, 303, 49 },
		{ 385, 304, 49 },
		{ 385, 305, 49 },
		{ 385, 306, 49 },
		{ 385, 307, 49 },
		{ 385, 308, 49 },
		{ 385, 309, 49 },
		{ 385, 310, 49 },
		{ 385, 311, 49 },
		{ 385, 312, 49 },
		{ 385, 313, 49 },
		{ 385, 314, 49 },
		{ 385, 315, 49 },
		{ 385, 316, 49 },
		{ 385, 317, 49 },
		{ 385, 318, 49 },
		{ 385, 319, 49 },
		{ 385, 320, 49 },
		{ 386, 290, 49 },
		{ 386, 291, 49 },
		{ 386, 292, 49 },
		{ 386, 293, 49 },
		{ 386, 294, 49 },
		{ 386, 295, 49 },
		{ 386, 296, 49 },
		{ 386, 297, 49 },
		{ 386, 298, 49 },
		{ 386, 299, 49 },
		{ 386, 300, 49 },
		{ 386, 301, 49 },
		{ 386, 302, 49 },
		{ 386, 303, 49 },
		{ 386, 304, 49 },
		{ 386, 305, 49 },
		{ 386, 306, 49 },
		{ 386, 307, 49 },
		{ 386, 308, 49 },
		{ 386, 309, 49 },
		{ 386, 310, 49 },
		{ 386, 311, 49 },
		{ 386, 312, 49 },
		{ 386, 313, 49 },
		{ 386, 314, 49 },
		{ 386, 315, 49 },
		{ 386, 316, 49 },
		{ 386, 317, 49 },
		{ 386, 318, 49 },
		{ 386, 319, 49 },
		{ 386, 320, 49 },
		{ 386, 321, 49 },
		{ 387, 290, 49 },
		{ 387, 291, 49 },
		{ 387, 292, 49 },
		{ 387, 293, 49 },
		{ 387, 294, 49 },
		{ 387, 295, 49 },
		{ 387, 296, 49 },
		{ 387, 297, 49 },
		{ 387, 298, 49 },
		{ 387, 299, 49 },
		{ 387, 300, 49 },
		{ 387, 301, 49 },
		{ 387, 302, 49 },
		{ 387, 303, 49 },
		{ 387, 304, 49 },
		{ 387, 305, 49 },
		{ 387, 306, 49 },
		{ 387, 307, 49 },
		{ 387, 308, 49 },
		{ 387, 309, 49 },
		{ 387, 310, 49 },
		{ 387, 311, 49 },
		{ 387, 312, 49 },
		{ 387, 313, 49 },
		{ 387, 314, 49 },
		{ 387, 315, 49 },
		{ 387, 316, 49 },
		{ 387, 317, 49 },
		{ 387, 318, 49 },
		{ 387, 319, 49 },
		{ 387, 320, 49 },
		{ 387, 321, 49 },
		{ 388, 290, 49 },
		{ 388, 291, 49 },
		{ 388, 292, 49 },
		{ 388, 293, 49 },
		{ 388, 294, 49 },
		{ 388, 295, 49 },
		{ 388, 296, 49 },
		{ 388, 297, 49 },
		{ 388, 298, 49 },
		{ 388, 299, 49 },
		{ 388, 300, 49 },
		{ 388, 301, 49 },
		{ 388, 302, 49 },
		{ 388, 303, 49 },
		{ 388, 304, 49 },
		{ 388, 305, 49 },
		{ 388, 306, 49 },
		{ 388, 307, 49 },
		{ 388, 308, 49 },
		{ 388, 309, 49 },
		{ 388, 310, 49 },
		{ 388, 311, 49 },
		{ 388, 312, 49 },
		{ 388, 313, 49 },
		{ 388, 314, 49 },
		{ 388, 315, 49 },
		{ 388, 316, 49 },
		{ 388, 317, 49 },
		{ 388, 318, 49 },
		{ 388, 319, 49 },
		{ 388, 320, 49 },
		{ 388, 321, 49 },
		{ 388, 322, 49 },
		{ 389, 291, 49 },
		{ 389, 292, 49 },
		{ 389, 293, 49 },
		{ 389, 294, 49 },
		{ 389, 295, 49 },
		{ 389, 296, 49 },
		{ 389, 297, 49 },
		{ 389, 298, 49 },
		{ 389, 299, 49 },
		{ 389, 300, 49 },
		{ 389, 301, 49 },
		{ 389, 302, 49 },
		{ 389, 303, 49 },
		{ 389, 304, 49 },
		{ 389, 305, 49 },
		{ 389, 306, 49 },
		{ 389, 307, 49 },
		{ 389, 308, 49 },
		{ 389, 309, 49 },
		{ 389, 310, 49 },
		{ 389, 311, 49 },
		{ 389, 312, 49 },
		{ 389, 313, 49 },
		{ 389, 314, 49 },
		{ 389, 315, 49 },
		{ 389, 316, 49 },
		{ 389, 317, 49 },
		{ 389, 318, 49 },
		{ 389, 319, 49 },
		{ 389, 320, 49 },
		{ 389, 321, 49 },
		{ 389, 322, 49 },
		{ 390, 291, 49 },
		{ 390, 292, 49 },
		{ 390, 293, 49 },
		{ 390, 294, 49 },
		{ 390, 295, 49 },
		{ 390, 296, 49 },
		{ 390, 297, 49 },
		{ 390, 298, 49 },
		{ 390, 299, 49 },
		{ 390, 300, 49 },
		{ 390, 301, 49 },
		{ 390, 302, 49 },
		{ 390, 303, 49 },
		{ 390, 304, 49 },
		{ 390, 305, 49 },
		{ 390, 306, 49 },
		{ 390, 307, 49 },
		{ 390, 308, 49 },
		{ 390, 309, 49 },
		{ 390, 310, 49 },
		{ 390, 311, 49 },
		{ 390, 312, 49 },
		{ 390, 313, 49 },
		{ 390, 314, 49 },
		{ 390, 315, 49 },
		{ 390, 316, 49 },
		{ 390, 317, 49 },
		{ 390, 318, 49 },
		{ 390, 319, 49 },
		{ 390, 320, 49 },
		{ 390, 321, 49 },
		{ 390, 322, 49 },
		{ 390, 323, 49 },
		{ 391, 291, 49 },
		{ 391, 292, 49 },
		{ 391, 293, 49 },
		{ 391, 294, 49 },
		{ 391, 295, 49 },
		{ 391, 296, 49 },
		{ 391, 297, 49 },
		{ 391, 298, 49 },
		{ 391, 299, 49 },
		{ 391, 300, 49 },
		{ 391, 301, 49 },
		{ 391, 302, 49 },
		{ 391, 303, 49 },
		{ 391, 304, 49 },
		{ 391, 305, 49 },
		{ 391, 306, 49 },
		{ 391, 307, 49 },
		{ 391, 308, 49 },
		{ 391, 309, 49 },
		{ 391, 310, 49 },
		{ 391, 311, 49 },
		{ 391, 312, 49 },
		{ 391, 313, 49 },
		{ 391, 314, 49 },
		{ 391, 315, 49 },
		{ 391, 316, 49 },
		{ 391, 317, 49 },
		{ 391, 318, 49 },
		{ 391, 319, 49 },
		{ 391, 320, 49 },
		{ 391, 321, 49 },
		{ 391, 322, 49 },
		{ 391, 323, 49 },
		{ 392, 291, 49 },
		{ 392, 292, 49 },
		{ 392, 293, 49 },
		{ 392, 294, 49 },
		{ 392, 295, 49 },
		{ 392, 296, 49 },
		{ 392, 297, 49 },
		{ 392, 298, 49 },
		{ 392, 299, 49 },
		{ 392, 300, 49 },
		{ 392, 301, 49 },
		{ 392, 302, 49 },
		{ 392, 303, 49 },
		{ 392, 304, 49 },
		{ 392, 305, 49 },
		{ 392, 306, 49 },
		{ 392, 307, 49 },
		{ 392, 308, 49 },
		{ 392, 309, 49 },
		{ 392, 310, 49 },
		{ 392, 311, 49 },
		{ 392, 312, 49 },
		{ 392, 313, 49 },
		{ 392, 314, 49 },
		{ 392, 315, 49 },
		{ 392, 316, 49 },
		{ 392, 317, 49 },
		{ 392, 318, 49 },
		{ 392, 319, 49 },
		{ 392, 320, 49 },
		{ 392, 321, 49 },
		{ 392, 322, 49 },
		{ 392, 323, 49 },
		{ 392, 324, 49 },
		{ 393, 292, 49 },
		{ 393, 293, 49 },
		{ 393, 294, 49 },
		{ 393, 295, 49 },
		{ 393, 296, 49 },
		{ 393, 297, 49 },
		{ 393, 298, 49 },
		{ 393, 299, 49 },
		{ 393, 300, 49 },
		{ 393, 301, 49 },
		{ 393, 302, 49 },
		{ 393, 303, 49 },
		{ 393, 304, 49 },
		{ 393, 305, 49 },
		{ 393, 306, 49 },
		{ 393, 307, 49 },
		{ 393, 308, 49 },
		{ 393, 309, 49 },
		{ 393, 310, 49 },
		{ 393, 311, 49 },
		{ 393, 312, 49 },
		{ 393, 313, 49 },
		{ 393, 314, 49 },
		{ 393, 315, 49 },
		{ 393, 316, 49 },
		{ 393, 317, 49 },
		{ 393, 318, 49 },
		{ 393, 319, 49 },
		{ 393, 320, 49 },
		{ 393, 321, 49 },
		{ 393, 322, 49 },
		{ 393, 323, 49 },
		{ 393, 324, 49 },
		{ 394, 292, 49 },
		{ 394, 293, 49 },
		{ 394, 294, 49 },
		{ 394, 295, 49 },
		{ 394, 296, 49 },
		{ 394, 297, 49 },
		{ 394, 298, 49 },
		{ 394, 299, 49 },
		{ 394, 300, 49 },
		{ 394, 301, 49 },
		{ 394, 302, 49 },
		{ 394, 303, 49 },
		{ 394, 304, 49 },
		{ 394, 305, 49 },
		{ 394, 306, 49 },
		{ 394, 307, 49 },
		{ 394, 308, 49 },
		{ 394, 309, 49 },
		{ 394, 310, 49 },
		{ 394, 311, 49 },
		{ 394, 312, 49 },
		{ 394, 313, 49 },
		{ 394, 314, 49 },
		{ 394, 315, 49 },
		{ 394, 316, 49 },
		{ 394, 317, 49 },
		{ 394, 318, 49 },
		{ 394, 319, 49 },
		{ 394, 320, 49 },
		{ 394, 321, 49 },
		{ 394, 322, 49 },
		{ 394, 323, 49 },
		{ 394, 324, 49 },
		{ 394, 325, 49 },
		{ 395, 292, 49 },
		{ 395, 293, 49 },
		{ 395, 294, 49 },
		{ 395, 295, 49 },
		{ 395, 296, 49 },
		{ 395, 297, 49 },
		{ 395, 298, 49 },
		{ 395, 299, 49 },
		{ 395, 300, 49 },
		{ 395, 301, 49 },
		{ 395, 302, 49 },
		{ 395, 303, 49 },
		{ 395, 304, 49 },
		{ 395, 305, 49 },
		{ 395, 306, 49 },
		{ 395, 307, 49 },
		{ 395, 308, 49 },
		{ 395, 309, 49 },
		{ 395, 310, 49 },
		{ 395, 311, 49 },
		{ 395, 312, 49 },
		{ 395, 313, 49 },
		{ 395, 314, 49 },
		{ 395, 315, 49 },
		{ 395, 316, 49 },
		{ 395, 317, 49 },
		{ 395, 318, 49 },
		{ 395, 319, 49 },
		{ 395, 320, 49 },
		{ 395, 321, 49 },
		{ 395, 322, 49 },
		{ 395, 323, 49 },
		{ 395, 324, 49 },
		{ 395, 325, 49 },
		{ 396, 292, 49 },
		{ 396, 293, 49 },
		{ 396, 294, 49 },
		{ 396, 295, 49 },
		{ 396, 296, 49 },
		{ 396, 297, 49 },
		{ 396, 298, 49 },
		{ 396, 299, 49 },
		{ 396, 300, 49 },
		{ 396, 301, 49 },
		{ 396, 302, 49 },
		{ 396, 303, 49 },
		{ 396, 304, 49 },
		{ 396, 305, 49 },
		{ 396, 306, 49 },
		{ 396, 307, 49 },
		{ 396, 308, 49 },
		{ 396, 309, 49 },
		{ 396, 310, 49 },
		{ 396, 311, 49 },
		{ 396, 312, 49 },
		{ 396, 313, 49 },
		{ 396, 314, 49 },
		{ 396, 315, 49 },
		{ 396, 316, 49 },
		{ 396, 317, 49 },
		{ 396, 318, 49 },
		{ 396, 319, 49 },
		{ 396, 320, 49 },
		{ 396, 321, 49 },
		{ 396, 322, 49 },
		{ 396, 323, 49 },
		{ 396, 324, 49 },
		{ 396, 325, 49 },
		{ 396, 326, 49 },
		{ 397, 293, 49 },
		{ 397, 294, 49 },
		{ 397, 295, 49 },
		{ 397, 296, 49 },
		{ 397, 297, 49 },
		{ 397, 298, 49 },
		{ 397, 299, 49 },
		{ 397, 300, 49 },
		{ 397, 301, 49 },
		{ 397, 302, 49 },
		{ 397, 303, 49 },
		{ 397, 304, 49 },
		{ 397, 305, 49 },
		{ 397, 306, 49 },
		{ 397, 307, 49 },
		{ 397, 308, 49 },
		{ 397, 309, 49 },
		{ 397, 310, 49 },
		{ 397, 311, 49 },
		{ 397, 312, 49 },
		{ 397, 313, 49 },
		{ 397, 314, 49 },
		{ 397, 315, 49 },
		{ 397, 316, 49 },
		{ 397, 317, 49 },
		{ 397, 318, 49 },
		{ 397, 319, 49 },
		{ 397, 320, 49 },
		{ 397, 321, 49 },
		{ 397, 322, 49 },
		{ 397, 323, 49 },
		{ 397, 324, 49 },
		{ 397, 325, 49 },
		{ 397, 326, 49 },
		{ 398, 293, 49 },
		{ 398, 294, 49 },
		{ 398, 295, 49 },
		{ 398, 296, 49 },
		{ 398, 297, 49 },
		{ 398, 298, 49 },
		{ 398, 299, 49 },
		{ 398, 300, 49 },
		{ 398, 301, 49 },
		{ 398, 302, 49 },
		{ 398, 303, 49 },
		{ 398, 304, 49 },
		{ 398, 305, 49 },
		{ 398, 306, 49 },
		{ 398, 307, 49 },
		{ 398, 308, 49 },
		{ 398, 309, 49 },
		{ 398, 310, 49 },
		{ 398, 311, 49 },
		{ 398, 312, 49 },
		{ 398, 313, 49 },
		{ 398, 314, 49 },
		{ 398, 315, 49 },
		{ 398, 316, 49 },
		{ 398, 317, 49 },
		{ 398, 318, 49 },
		{ 398, 319, 49 },
		{ 398, 320, 49 },
		{ 398, 321, 49 },
		{ 398, 322, 49 },
		{ 398, 323, 49 },
		{ 398, 324, 49 },
		{ 398, 325, 49 },
		{ 398, 326, 49 },
		{ 398, 327, 49 },
		{ 399, 293, 49 },
		{ 399, 294, 49 },
		{ 399, 295, 49 },
		{ 399, 296, 49 },
		{ 399, 297, 49 },
		{ 399, 298, 49 },
		{ 399, 299, 49 },
		{ 399, 300, 49 },
		{ 399, 301, 49 },
		{ 399, 302, 49 },
		{ 399, 303, 49 },
		{ 399, 304, 49 },
		{ 399, 305, 49 },
		{ 399, 306, 49 },
		{ 399, 307, 49 },
		{ 399, 308, 49 },
		{ 399, 309, 49 },
		{ 399, 310, 49 },
		{ 399, 311, 49 },
		{ 399, 312, 49 },
		{ 399, 313, 49 },
		{ 399, 314, 49 },
		{ 399, 315, 49 },
		{ 399, 316, 49 },
		{ 399, 317, 49 },
		{ 399, 318, 49 },
		{ 399, 319, 49 },
		{ 399, 320, 49 },
		{ 399, 321, 49 },
		{ 399, 322, 49 },
		{ 399, 323, 49 },
		{ 399, 324, 49 },
		{ 399, 325, 49 },
		{ 399, 326, 49 },
		{ 399, 327, 49 },
		{ 400, 293, 49 },
		{ 400, 294, 49 },
		{ 400, 295, 49 },
		{ 400, 296, 49 },
		{ 400, 297, 49 },
		{ 400, 298, 49 },
		{ 400, 299, 49 },
		{ 400, 300, 49 },
		{ 400, 301, 49 },
		{ 400, 302, 49 },
		{ 400, 303, 49 },
		{ 400, 304, 49 },
		{ 400, 305, 49 },
		{ 400, 306, 49 },
		{ 400, 307, 49 },
		{ 400, 308, 49 },
		{ 400, 309, 49 },
		{ 400, 310, 49 },
		{ 400, 311, 49 },
		{ 400, 312, 49 },
		{ 400, 313, 49 },
		{ 400, 314, 49 },
		{ 400, 315, 49 },
		{ 400, 316, 49 },
		{ 400, 317, 49 },
		{ 400, 318, 49 },
		{ 400, 319, 49 },
		{ 400, 320, 49 },
		{ 400, 321, 49 },
		{ 400, 322, 49 },
		{ 400, 323, 49 },
		{ 400, 324, 49 },
		{ 400, 325, 49 },
		{ 400, 326, 49 },
		{ 400, 327, 49 },
		{ 400, 328, 49 },
		{ 401, 294, 49 },
		{ 401, 295, 49 },
		{ 401, 296, 49 },
		{ 401, 297, 49 },
		{ 401, 298, 49 },
		{ 401, 299, 49 },
		{ 401, 300, 49 },
		{ 401, 301, 49 },
		{ 401, 302, 49 },
		{ 401, 303, 49 },
		{ 401, 304, 49 },
		{ 401, 305, 49 },
		{ 401, 306, 49 },
		{ 401, 307, 49 },
		{ 401, 308, 49 },
		{ 401, 309, 49 },
		{ 401, 310, 49 },
		{ 401, 311, 49 },
		{ 401, 312, 49 },
		{ 401, 313, 49 },
		{ 401, 314, 49 },
		{ 401, 315, 49 },
		{ 401, 316, 49 },
		{ 401, 317, 49 },
		{ 401, 318, 49 },
		{ 401, 319, 49 },
		{ 401, 320, 49 },
		{ 401, 321, 49 },
		{ 401, 322, 49 },
		{ 401, 323, 49 },
		{ 401, 324, 49 },
		{ 401, 325, 49 },
		{ 401, 326, 49 },
		{ 401, 327, 49 },
		{ 401, 328, 49 },
		{ 402, 294, 49 },
		{ 402, 295, 49 },
		{ 402, 296, 49 },
		{ 402, 297, 49 },
		{ 402, 298, 49 },
		{ 402, 299, 49 },
		{ 402, 300, 49 },
		{ 402, 301, 49 },
		{ 402, 302, 49 },
		{ 402, 303, 49 },
		{ 402, 304, 49 },
		{ 402, 305, 49 },
		{ 402, 306, 49 },
		{ 402, 307, 49 },
		{ 402, 308, 49 },
		{ 402, 309, 49 },
		{ 402, 310, 49 },
		{ 402, 311, 49 },
		{ 402, 312, 49 },
		{ 402, 313, 49 },
		{ 402, 314, 49 },
		{ 402, 315, 49 },
		{ 402, 316, 49 },
		{ 402, 317, 49 },
		{ 402, 318, 49 },
		{ 402, 319, 49 },
		{ 402, 320, 49 },
		{ 402, 321, 49 },
		{ 402, 322, 49 },
		{ 402, 323, 49 },
		{ 402, 324, 49 },
		{ 402, 325, 49 },
		{ 402, 326, 49 },
		{ 402, 327, 49 },
		{ 402, 328, 49 },
		{ 402, 329, 49 },
		{ 403, 294, 49 },
		{ 403, 295, 49 },
		{ 403, 296, 49 },
		{ 403, 297, 49 },
		{ 403, 298, 49 },
		{ 403, 299, 49 },
		{ 403, 300, 49 },
		{ 403, 301, 49 },
		{ 403, 302, 49 },
		{ 403, 303, 49 },
		{ 403, 304, 49 },
		{ 403, 305, 49 },
		{ 403, 306, 49 },
		{ 403, 307, 49 },
		{ 403, 308, 49 },
		{ 403, 309, 49 },
		{ 403, 310, 49 },
		{ 403, 311, 49 },
		{ 403, 312, 49 },
		{ 403, 313, 49 },
		{ 403, 314, 49 },
		{ 403, 315, 49 },
		{ 403, 316, 49 },
		{ 403, 317, 49 },
		{ 403, 318, 49 },
		{ 403, 319, 49 },
		{ 403, 320, 49 },
		{ 403, 321, 49 },
		{ 403, 322, 49 },
		{ 403, 323, 49 },
		{ 403, 324, 49 },
		{ 403, 325, 49 },
		{ 403, 326, 49 },
		{ 403, 327, 49 },
		{ 403, 328, 49 },
		{ 403, 329, 49 },
		{ 404, 294, 49 },
		{ 404, 295, 49 },
		{ 404, 296, 49 },
		{ 404, 297, 49 },
		{ 404, 298, 49 },
		{ 404, 299, 49 },
		{ 404, 300, 49 },
		{ 404, 301, 49 },
		{ 404, 302, 49 },
		{ 404, 303, 49 },
		{ 404, 304, 49 },
		{ 404, 305, 49 },
		{ 404, 306, 49 },
		{ 404, 307, 49 },
		{ 404, 308, 49 },
		{ 404, 309, 49 },
		{ 404, 310, 49 },
		{ 404, 311, 49 },
		{ 404, 312, 49 },
		{ 404, 313, 49 },
		{ 404, 314, 49 },
		{ 404, 315, 49 },
		{ 404, 316, 49 },
		{ 404, 317, 49 },
		{ 404, 318, 49 },
		{ 404, 319, 49 },
		{ 404, 320, 49 },
		{ 404, 321, 49 },
		{ 404, 322, 49 },
		{ 404, 323, 49 },
		{ 404, 324, 49 },
		{ 404, 325, 49 },
		{ 404, 326, 49 },
		{ 404, 327, 49 },
		{ 404, 328, 49 },
		{ 404, 329, 49 },
		{ 404, 330, 49 },
		{ 405, 295, 49 },
		{ 405, 296, 49 },
		{ 405, 297, 49 },
		{ 405, 298, 49 },
		{ 405, 299, 49 },
		{ 405, 300, 49 },
		{ 405, 301, 49 },
		{ 405, 302, 49 },
		{ 405, 303, 49 },
		{ 405, 304, 49 },
		{ 405, 305, 49 },
		{ 405, 306, 49 },
		{ 405, 307, 49 },
		{ 405, 308, 49 },
		{ 405, 309, 49 },
		{ 405, 310, 49 },
		{ 405, 311, 49 },
		{ 405, 312, 49 },
		{ 405, 313, 49 },
		{ 405, 314, 49 },
		{ 405, 315, 49 },
		{ 405, 316, 49 },
		{ 405, 317, 49 },
		{ 405, 318, 49 },
		{ 405, 319, 49 },
		{ 405, 320, 49 },
		{ 405, 321, 49 },
		{ 405, 322, 49 },
		{ 405, 323, 49 },
		{ 405, 324, 49 },
		{ 405, 325, 49 },
		{ 405, 326, 49 },
		{ 405, 327, 49 },
		{ 405, 328, 49 },
		{ 405, 329, 49 },
		{ 405, 330, 49 },
		{ 406, 295, 49 },
		{ 406, 296, 49 },
		{ 406, 297, 49 },
		{ 406, 298, 49 },
		{ 406, 299, 49 },
		{ 406, 300, 49 },
		{ 406, 301, 49 },
		{ 406, 302, 49 },
		{ 406, 303, 49 },
		{ 406, 304, 49 },
		{ 406, 305, 49 },
		{ 406, 306, 49 },
		{ 406, 307, 49 },
		{ 406, 308, 49 },
		{ 406, 309, 49 },
		{ 406, 310, 49 },
		{ 406, 311, 49 },
		{ 406, 312, 49 },
		{ 406, 313, 49 },
		{ 406, 314, 49 },
		{ 406, 315, 49 },
		{ 406, 316, 49 },
		{ 406, 317, 49 },
		{ 406, 318, 49 },
		{ 406, 319, 49 },
		{ 406, 320, 49 },
		{ 406, 321, 49 },
		{ 406, 322, 49 },
		{ 406, 323, 49 },
		{ 406, 324, 49 },
		{ 406, 325, 49 },
		{ 406, 326, 49 },
		{ 406, 327, 49 },
		{ 406, 328, 49 },
		{ 406, 329, 49 },
		{ 406, 330, 49 },
		{ 406, 331, 49 },
		{ 407, 295, 49 },
		{ 407, 296, 49 },
		{ 407, 297, 49 },
		{ 407, 298, 49 },
		{ 407, 299, 49 },
		{ 407, 300, 49 },
		{ 407, 301, 49 },
		{ 407, 302, 49 },
		{ 407, 303, 49 },
		{ 407, 304, 49 },
		{ 407, 305, 49 },
		{ 407, 306, 49 },
		{ 407, 307, 49 },
		{ 407, 308, 49 },
		{ 407, 309, 49 },
		{ 407, 310, 49 },
		{ 407, 311, 49 },
		{ 407, 312, 49 },
		{ 407, 313, 49 },
		{ 407, 314, 49 },
		{ 407, 315, 49 },
		{ 407, 316, 49 },
		{ 407, 317, 49 },
		{ 407, 318, 49 },
		{ 407, 319, 49 },
		{ 407, 320, 49 },
		{ 407, 321, 49 },
		{ 407, 322, 49 },
		{ 407, 323, 49 },
		{ 407, 324, 49 },
		{ 407, 325, 49 },
		{ 407, 326, 49 },
		{ 407, 327, 49 },
		{ 407, 328, 49 },
		{ 407, 329, 49 },
		{ 407, 330, 49 },
		{ 407, 331, 49 },
		{ 408, 295, 49 },
		{ 408, 296, 49 },
		{ 408, 297, 49 },
		{ 408, 298, 49 },
		{ 408, 299, 49 },
		{ 408, 300, 49 },
		{ 408, 301, 49 },
		{ 408, 302, 49 },
		{ 408, 303, 49 },
		{ 408, 304, 49 },
		{ 408, 305, 49 },
		{ 408, 306, 49 },
		{ 408, 307, 49 },
		{ 408, 308, 49 },
		{ 408, 309, 49 },
		{ 408, 310, 49 },
		{ 408, 311, 49 },
		{ 408, 312, 49 },
		{ 408, 313, 49 },
		{ 408, 314, 49 },
		{ 408, 315, 49 },
		{ 408, 316, 49 },
		{ 408, 317, 49 },
		{ 408, 318, 49 },
		{ 408, 319, 49 },
		{ 408, 320, 49 },
		{ 408, 321, 49 },
		{ 408, 322, 49 },
		{ 408, 323, 49 },
		{ 408, 324, 49 },
		{ 408, 325, 49 },
		{ 408, 326, 49 },
		{ 408, 327, 49 },
		{ 408, 328, 49 },
		{ 408, 329, 49 },
		{ 408, 330, 49 },
		{ 408, 331, 49 },
		{ 408, 332, 49 },
		{ 409, 296, 49 },
		{ 409, 297, 49 },
		{ 409, 298, 49 },
		{ 409, 299, 49 },
		{ 409, 300, 49 },
		{ 409, 301, 49 },
		{ 409, 302, 49 },
		{ 409, 303, 49 },
		{ 409, 304, 49 },
		{ 409, 305, 49 },
		{ 409, 306, 49 },
		{ 409, 307, 49 },
		{ 409, 308, 49 },
		{ 409, 309, 49 },
		{ 409, 310, 49 },
		{ 409, 311, 49 },
		{ 409, 312, 49 },
		{ 409, 313, 49 },
		{ 409, 314, 49 },
		{ 409, 315, 49 },
		{ 409, 316, 49 },
		{ 409, 317, 49 },
		{ 409, 318, 49 },
		{ 409, 319, 49 },
		{ 409, 320, 49 },
		{ 409, 321, 49 },
		{ 409, 322, 49 },
		{ 409, 323, 49 },
		{ 409, 324, 49 },
		{ 409, 325, 49 },
		{ 409, 326, 49 },
		{ 409, 327, 49 },
		{ 409, 328, 49 },
		{ 409, 329, 49 },
		{ 409, 330, 49 },
		{ 409, 331, 49 },
		{ 409, 332, 49 },
		{ 410, 296, 49 },
		{ 410, 297, 49 },
		{ 410, 298, 49 },
		{ 410, 299, 49 },
		{ 410, 300, 49 },
		{ 410, 301, 49 },
		{ 410, 302, 49 },
		{ 410, 303, 49 },
		{ 410, 304, 49 },
		{ 410, 305, 49 },
		{ 410, 306, 49 },
		{ 410, 307, 49 },
		{ 410, 308, 49 },
		{ 410, 309, 49 },
		{ 410, 310, 49 },
		{ 410, 311, 49 },
		{ 410, 312, 49 },
		{ 410, 313, 49 },
		{ 410, 314, 49 },
		{ 410, 315, 49 },
		{ 410, 316, 49 },
		{ 410, 317, 49 },
		{ 410, 318, 49 },
		{ 410, 319, 49 },
		{ 410, 320, 49 },
		{ 410, 321, 49 },
		{ 410, 322, 49 },
		{ 410, 323, 49 },
		{ 410, 324, 49 },
		{ 410, 325, 49 },
		{ 410, 326, 49 },
		{ 410, 327, 49 },
		{ 410, 328, 49 },
		{ 410, 329, 49 },
		{ 410, 330, 49 },
		{ 410, 331, 49 },
		{ 410, 332, 49 },
		{ 410, 333, 49 },
		{ 411, 296, 49 },
		{ 411, 297, 49 },
		{ 411, 298, 49 },
		{ 411, 299, 49 },
		{ 411, 300, 49 },
		{ 411, 301, 49 },
		{ 411, 302, 49 },
		{ 411, 303, 49 },
		{ 411, 304, 49 },
		{ 411, 305, 49 },
		{ 411, 306, 49 },
		{ 411, 307, 49 },
		{ 411, 308, 49 },
		{ 411, 309, 49 },
		{ 411, 310, 49 },
		{ 411, 311, 49 },
		{ 411, 312, 49 },
		{ 411, 313, 49 },
		{ 411, 314, 49 },
		{ 411, 315, 49 },
		{ 411, 316, 49 },
		{ 411, 317, 49 },
		{ 411, 318, 49 },
		{ 411, 319, 49 },
		{ 411, 320, 49 },
		{ 411, 321, 49 },
		{ 411, 322, 49 },
		{ 411, 323, 49 },
		{ 411, 324, 49 },
		{ 411, 325, 49 },
		{ 411, 326, 49 },
		{ 411, 327, 49 },
		{ 411, 328, 49 },
		{ 411, 329, 49 },
		{ 411, 330, 49 },
		{ 411, 331, 49 },
		{ 411, 332, 49 },
		{ 411, 333, 49 },
		{ 412, 296, 49 },
		{ 412, 297, 49 },
		{ 412, 298, 49 },
		{ 412, 299, 49 },
		{ 412, 300, 49 },
		{ 412, 301, 49 },
		{ 412, 302, 49 },
		{ 412, 303, 49 },
		{ 412, 304, 49 },
		{ 412, 305, 49 },
		{ 412, 306, 49 },
		{ 412, 307, 49 },
		{ 412, 308, 49 },
		{ 412, 309, 49 },
		{ 412, 310, 49 },
		{ 412, 311, 49 },
		{ 412, 312, 49 },
		{ 412, 313, 49 },
		{ 412, 314, 49 },
		{ 412, 315, 49 },
		{ 412, 316, 49 },
		{ 412, 317, 49 },
		{ 412, 318, 49 },
		{ 412, 319, 49 },
		{ 412, 320, 49 },
		{ 412, 321, 49 },
		{ 412, 322, 49 },
		{ 412, 323, 49 },
		{ 412, 324, 49 },
		{ 412, 325, 49 },
		{ 412, 326, 49 },
		{ 412, 327, 49 },
		{ 412, 328, 49 },
		{ 412, 329, 49 },
		{ 412, 330, 49 },
		{ 412, 331, 49 },
		{ 412, 332, 49 },
		{ 412, 333, 49 },
		{ 412, 334, 49 },
		{ 413, 297, 49 },
		{ 413, 298, 49 },
		{ 413, 299, 49 },
		{ 413, 300, 49 },
		{ 413, 301, 49 },
		{ 413, 302, 49 },
		{ 413, 303, 49 },
		{ 413, 304, 49 },
		{ 413, 305, 49 },
		{ 413, 306, 49 },
		{ 413, 307, 49 },
		{ 413, 308, 49 },
		{ 413, 309, 49 },
		{ 413, 310, 49 },
		{ 413, 311, 49 },
		{ 413, 312, 49 },
		{ 413, 313, 49 },
		{ 413, 314, 49 },
		{ 413, 315, 49 },
		{ 413, 316, 49 },
		{ 413, 317, 49 },
		{ 413, 318, 49 },
		{ 413, 319, 49 },
		{ 413, 320, 49 },
		{ 413, 321, 49 },
		{ 413, 322, 49 },
		{ 413, 323, 49 },
		{ 413, 324, 49 },
		{ 413, 325, 49 },
		{ 413, 326, 49 },
		{ 413, 327, 49 },
		{ 413, 328, 49 },
		{ 413, 329, 49 },
		{ 413, 330, 49 },
		{ 413, 331, 49 },
		{ 413, 332, 49 },
		{ 413, 333, 49 },
		{ 413, 334, 49 },
		{ 414, 297, 49 },
		{ 414, 298, 49 },
		{ 414, 299, 49 },
		{ 414, 300, 49 },
		{ 414, 301, 49 },
		{ 414, 302, 49 },
		{ 414, 303, 49 },
		{ 414, 304, 49 },
		{ 414, 305, 49 },
		{ 414, 306, 49 },
		{ 414, 307, 49 },
		{ 414, 308, 49 },
		{ 414, 309, 49 },
		{ 414, 310, 49 },
		{ 414, 311, 49 },
		{ 414, 312, 49 },
		{ 414, 313, 49 },
		{ 414, 314, 49 },
		{ 414, 315, 49 },
		{ 414, 316, 49 },
		{ 414, 317, 49 },
		{ 414, 318, 49 },
		{ 414, 319, 49 },
		{ 414, 320, 49 },
		{ 414, 321, 49 },
		{ 414, 322, 49 },
		{ 414, 323, 49 },
		{ 414, 324, 49 },
		{ 414, 325, 49 },
		{ 414, 326, 49 },
		{ 414, 327, 49 },
		{ 414, 328, 49 },
		{ 414, 329, 49 },
		{ 414, 330, 49 },
		{ 414, 331, 49 },
		{ 414, 332, 49 },
		{ 414, 333, 49 },
		{ 414, 334, 49 },
		{ 414, 335, 49 },
		{ 415, 297, 49 },
		{ 415, 298, 49 },
		{ 415, 299, 49 },
		{ 415, 300, 49 },
		{ 415, 301, 49 },
		{ 415, 302, 49 },
		{ 415, 303, 49 },
		{ 415, 304, 49 },
		{ 415, 305, 49 },
		{ 415, 306, 49 },
		{ 415, 307, 49 },
		{ 415, 308, 49 },
		{ 415, 309, 49 },
		{ 415, 310, 49 },
		{ 415, 311, 49 },
		{ 415, 312, 49 },
		{ 415, 313, 49 },
		{ 415, 314, 49 },
		{ 415, 315, 49 },
		{ 415, 316, 49 },
		{ 415, 317, 49 },
		{ 415, 318, 49 },
		{ 415, 319, 49 },
		{ 415, 320, 49 },
		{ 415, 321, 49 },
		{ 415, 322, 49 },
		{ 415, 323, 49 },
		{ 415, 324, 49 },
		{ 415, 325, 49 },
		{ 415, 326, 49 },
		{ 415, 327, 49 },
		{ 415, 328, 49 },
		{ 415, 329, 49 },
		{ 415, 330, 49 },
		{ 415, 331, 49 },
		{ 415, 332, 49 },
		{ 415, 333, 49 },
		{ 415, 334, 49 },
		{ 415, 335, 49 },
		{ 416, 297, 49 },
		{ 416, 298, 49 },
		{ 416, 299, 49 },
		{ 416, 300, 49 },
		{ 416, 301, 49 },
		{ 416, 302, 49 },
		{ 416, 303, 49 },
		{ 416, 304, 49 },
		{ 416, 305, 49 },
		{ 416, 306, 49 },
		{ 416, 307, 49 },
		{ 416, 308, 49 },
		{ 416, 309, 49 },
		{ 416, 310, 49 },
		{ 416, 311, 49 },
		{ 416, 312, 49 },
		{ 416, 313, 49 },
		{ 416, 314, 49 },
		{ 416, 315, 49 },
		{ 416, 316, 49 },
		{ 416, 317, 49 },
		{ 416, 318, 49 },
		{ 416, 319, 49 },
		{ 416, 320, 49 },
		{ 416, 321, 49 },
		{ 416, 322, 49 },
		{ 416, 323, 49 },
		{ 416, 324, 49 },
		{ 416, 325, 49 },
		{ 416, 326, 49 },
		{ 416, 327, 49 },
		{ 416, 328, 49 },
		{ 416, 329, 49 },
		{ 416, 330, 49 },
		{ 416, 331, 49 },
		{ 416, 332, 49 },
		{ 416, 333, 49 },
		{ 416, 334, 49 },
		{ 416, 335, 49 },
		{ 416, 336, 49 },
		{ 417, 298, 49 },
		{ 417, 299, 49 },
		{ 417, 300, 49 },
		{ 417, 301, 49 },
		{ 417, 302, 49 },
		{ 417, 303, 49 },
		{ 417, 304, 49 },
		{ 417, 305, 49 },
		{ 417, 306, 49 },
		{ 417, 307, 49 },
		{ 417, 308, 49 },
		{ 417, 309, 49 },
		{ 417, 310, 49 },
		{ 417, 311, 49 },
		{ 417, 312, 49 },
		{ 417, 313, 49 },
		{ 417, 314, 49 },
		{ 417, 315, 49 },
		{ 417, 316, 49 },
		{ 417, 317, 49 },
		{ 417, 318, 49 },
		{ 417, 319, 49 },
		{ 417, 320, 49 },
		{ 417, 321, 49 },
		{ 417, 322, 49 },
		{ 417, 323, 49 },
		{ 417, 324, 49 },
		{ 417, 325, 49 },
		{ 417, 326, 49 },
		{ 417, 327, 49 },
		{ 417, 328, 49 },
		{ 417, 329, 49 },
		{ 417, 330, 49 },
		{ 417, 331, 49 },
		{ 417, 332, 49 },
		{ 417, 333, 49 },
		{ 417, 334, 49 },
		{ 417, 335, 49 },
		{ 417, 336, 49 },
		{ 418, 298, 49 },
		{ 418, 299, 49 },
		{ 418, 300, 49 },
		{ 418, 301, 49 },
		{ 418, 302, 49 },
		{ 418, 303, 49 },
		{ 418, 304, 49 },
		{ 418, 305, 49 },
		{ 418, 306, 49 },
		{ 418, 307, 49 },
		{ 418, 308, 49 },
		{ 418, 309, 49 },
		{ 418, 310, 49 },
		{ 418, 311, 49 },
		{ 418, 312, 49 },
		{ 418, 313, 49 },
		{ 418, 314, 49 },
		{ 418, 315, 49 },
		{ 418, 316, 49 },
		{ 418, 317, 49 },
		{ 418, 318, 49 },
		{ 418, 319, 49 },
		{ 418, 320, 49 },
		{ 418, 321, 49 },
		{ 418, 322, 49 },
		{ 418, 323, 49 },
		{ 418, 324, 49 },
		{ 418, 325, 49 },
		{ 418, 326, 49 },
		{ 418, 327, 49 },
		{ 418, 328, 49 },
		{ 418, 329, 49 },
		{ 418, 330, 49 },
		{ 418, 331, 49 },
		{ 418, 332, 49 },
		{ 418, 333, 49 },
		{ 418, 334, 49 },
		{ 418, 335, 49 },
		{ 418, 336, 49 },
		{ 418, 337, 49 },
		{ 419, 298, 49 },
		{ 419, 299, 49 },
		{ 419, 300, 49 },
		{ 419, 301, 49 },
		{ 419, 302, 49 },
		{ 419, 303, 49 },
		{ 419, 304, 49 },
		{ 419, 305, 49 },
		{ 419, 306, 49 },
		{ 419, 307, 49 },
		{ 419, 308, 49 },
		{ 419, 309, 49 },
		{ 419, 310, 49 },
		{ 419, 311, 49 },
		{ 419, 312, 49 },
		{ 419, 313, 49 },
		{ 419, 314, 49 },
		{ 419, 315, 49 },
		{ 419, 316, 49 },
		{ 419, 317, 49 },
		{ 419, 318, 49 },
		{ 419, 319, 49 },
		{ 419, 320, 49 },
		{ 419, 321, 49 },
		{ 419, 322, 49 },
		{ 419, 323, 49 },
		{ 419, 324, 49 },
		{ 419, 325, 49 },
		{ 419, 326, 49 },
		{ 419, 327, 49 },
		{ 419, 328, 49 },
		{ 419, 329, 49 },
		{ 419, 330, 49 },
		{ 419, 331, 49 },
		{ 419, 332, 49 },
		{ 419, 333, 49 },
		{ 419, 334, 49 },
		{ 419, 335, 49 },
		{ 419, 336, 49 },
		{ 419, 337, 49 },
		{ 420, 298, 49 },
		{ 420, 299, 49 },
		{ 420, 300, 49 },
		{ 420, 301, 49 },
		{ 420, 302, 49 },
		{ 420, 303, 49 },
		{ 420, 304, 49 },
		{ 420, 305, 49 },
		{ 420, 306, 49 },
		{ 420, 307, 49 },
		{ 420, 308, 49 },
		{ 420, 309, 49 },
		{ 420, 310, 49 },
		{ 420, 311, 49 },
		{ 420, 312, 49 },
		{ 420, 313, 49 },
		{ 420, 314, 49 },
		{ 420, 315, 49 },
		{ 420, 316, 49 },
		{ 420, 317, 49 },
		{ 420, 318, 49 },
		{ 420, 319, 49 },
		{ 420, 320, 49 },
		{ 420, 321, 49 },
		{ 420, 322, 49 },
		{ 420, 323, 49 },
		{ 420, 324, 49 },
		{ 420, 325, 49 },
		{ 420, 326, 49 },
		{ 420, 327, 49 },
		{ 420, 328, 49 },
		{ 420, 329, 49 },
		{ 420, 330, 49 },
		{ 420, 331, 49 },
		{ 420, 332, 49 },
		{ 420, 333, 49 },
		{ 420, 334, 49 },
		{ 420, 335, 49 },
		{ 420, 336, 49 },
		{ 420, 337, 49 },
		{ 420, 338, 49 },
		{ 421, 299, 49 },
		{ 421, 300, 49 },
		{ 421, 301, 49 },
		{ 421, 302, 49 },
		{ 421, 303, 49 },
		{ 421, 304, 49 },
		{ 421, 305, 49 },
		{ 421, 306, 49 },
		{ 421, 307, 49 },
		{ 421, 308, 49 },
		{ 421, 309, 49 },
		{ 421, 310, 49 },
		{ 421, 311, 49 },
		{ 421, 312, 49 },
		{ 421, 313, 49 },
		{ 421, 314, 49 },
		{ 421, 315, 49 },
		{ 421, 316, 49 },
		{ 421, 317, 49 },
		{ 421, 318, 49 },
		{ 421, 319, 49 },
		{ 421, 320, 49 },
		{ 421, 321, 49 },
		{ 421, 322, 49 },
		{ 421, 323, 49 },
		{ 421, 324, 49 },
		{ 421, 325, 49 },
		{ 421, 326, 49 },
		{ 421, 327, 49 },
		{ 421, 328, 49 },
		{ 421, 329, 49 },
		{ 421, 330, 49 },
		{ 421, 331, 49 },
		{ 421, 332, 49 },
		{ 421, 333, 49 },
		{ 421, 334, 49 },
		{ 421, 335, 49 },
		{ 421, 336, 49 },
		{ 421, 337, 49 },
		{ 421, 338, 49 },
		{ 422, 299, 49 },
		{ 422, 300, 49 },
		{ 422, 301, 49 },
		{ 422, 302, 49 },
		{ 422, 303, 49 },
		{ 422, 304, 49 },
		{ 422, 305, 49 },
		{ 422, 306, 49 },
		{ 422, 307, 49 },
		{ 422, 308, 49 },
		{ 422, 309, 49 },
		{ 422, 310, 49 },
		{ 422, 311, 49 },
		{ 422, 312, 49 },
		{ 422, 313, 49 },
		{ 422, 314, 49 },
		{ 422, 315, 49 },
		{ 422, 316, 49 },
		{ 422, 317, 49 },
		{ 422, 318, 49 },
		{ 422, 319, 49 },
		{ 422, 320, 49 },
		{ 422, 321, 49 },
		{ 422, 322, 49 },
		{ 422, 323, 49 },
		{ 422, 324, 49 },
		{ 422, 325, 49 },
		{ 422, 326, 49 },
		{ 422, 327, 49 },
		{ 422, 328, 49 },
		{ 422, 329, 49 },
		{ 422, 330, 49 },
		{ 422, 331, 49 },
		{ 422, 332, 49 },
		{ 422, 333, 49 },
		{ 422, 334, 49 },
		{ 422, 335, 49 },
		{ 422, 336, 49 },
		{ 422, 337, 49 },
		{ 422, 338, 49 },
		{ 422, 339, 49 },
		{ 423, 299, 49 },
		{ 423, 300, 49 },
		{ 423, 301, 49 },
		{ 423, 302, 49 },
		{ 423, 303, 49 },
		{ 423, 304, 49 },
		{ 423, 305, 49 },
		{ 423, 306, 49 },
		{ 423, 307, 49 },
		{ 423, 308, 49 },
		{ 423, 309, 49 },
		{ 423, 310, 49 },
		{ 423, 311, 49 },
		{ 423, 312, 49 },
		{ 423, 313, 49 },
		{ 423, 314, 49 },
		{ 423, 315, 49 },
		{ 423, 316, 49 },
		{ 423, 317, 49 },
		{ 423, 318, 49 },
		{ 423, 319, 49 },
		{ 423, 320, 49 },
		{ 423, 321, 49 },
		{ 423, 322, 49 },
		{ 423, 323, 49 },
		{ 423, 324, 49 },
		{ 423, 325, 49 },
		{ 423, 326, 49 },
		{ 423, 327, 49 },
		{ 423, 328, 49 },
		{ 423, 329, 49 },
		{ 423, 330, 49 },
		{ 423, 331, 49 },
		{ 423, 332, 49 },
		{ 423, 333, 49 },
		{ 423, 334, 49 },
		{ 423, 335, 49 },
		{ 423, 336, 49 },
		{ 423, 337, 49 },
		{ 423, 338, 49 },
		{ 423, 339, 49 },
		{ 424, 299, 49 },
		{ 424, 300, 49 },
		{ 424, 301, 49 },
		{ 424, 302, 49 },
		{ 424, 303, 49 },
		{ 424, 304, 49 },
		{ 424, 305, 49 },
		{ 424, 306, 49 },
		{ 424, 307, 49 },
		{ 424, 308, 49 },
		{ 424, 309, 49 },
		{ 424, 310, 49 },
		{ 424, 311, 49 },
		{ 424, 312, 49 },
		{ 424, 313, 49 },
		{ 424, 314, 49 },
		{ 424, 315, 49 },
		{ 424, 316, 49 },
		{ 424, 317, 49 },
		{ 424, 318, 49 },
		{ 424, 319, 49 },
		{ 424, 320, 49 },
		{ 424, 321, 49 },
		{ 424, 322, 49 },
		{ 424, 323, 49 },
		{ 424, 324, 49 },
		{ 424, 325, 49 },
		{ 424, 326, 49 },
		{ 424, 327, 49 },
		{ 424, 328, 49 },
		{ 424, 329, 49 },
		{ 424, 330, 49 },
		{ 424, 331, 49 },
		{ 424, 332, 49 },
		{ 424, 333, 49 },
		{ 424, 334, 49 },
		{ 424, 335, 49 },
		{ 424, 336, 49 },
		{ 424, 337, 49 },
		{ 424, 338, 49 },
		{ 424, 339, 49 },
		{ 424, 340, 49 },
		{ 425, 300, 49 },
		{ 425, 301, 49 },
		{ 425, 302, 49 },
		{ 425, 303, 49 },
		{ 425, 304, 49 },
		{ 425, 305, 49 },
		{ 425, 306, 49 },
		{ 425, 307, 49 },
		{ 425, 308, 49 },
		{ 425, 309, 49 },
		{ 425, 310, 49 },
		{ 425, 311, 49 },
		{ 425, 312, 49 },
		{ 425, 313, 49 },
		{ 425, 314, 49 },
		{ 425, 315, 49 },
		{ 425, 316, 49 },
		{ 425, 317, 49 },
		{ 425, 318, 49 },
		{ 425, 319, 49 },
		{ 425, 320, 49 },
		{ 425, 321, 49 },
		{ 425, 322, 49 },
		{ 425, 323, 49 },
		{ 425, 324, 49 },
		{ 425, 325, 49 },
		{ 425, 326, 49 },
		{ 425, 327, 49 },
		{ 425, 328, 49 },
		{ 425, 329, 49 },
		{ 425, 330, 49 },
		{ 425, 331, 49 },
		{ 425, 332, 49 },
		{ 425, 333, 49 },
		{ 425, 334, 49 },
		{ 425, 335, 49 },
		{ 425, 336, 49 },
		{ 425, 337, 49 },
		{ 425, 338, 49 },
		{ 425, 339, 49 },
		{ 425, 340, 49 },
		{ 426, 300, 49 },
		{ 426, 301, 49 },
		{ 426, 302, 49 },
		{ 426, 303, 49 },
		{ 426, 304, 49 },
		{ 426, 305, 49 },
		{ 426, 306, 49 },
		{ 426, 307, 49 },
		{ 426, 308, 49 },
		{ 426, 309, 49 },
		{ 426, 310, 49 },
		{ 426, 311, 49 },
		{ 426, 312, 49 },
		{ 426, 313, 49 },
		{ 426, 314, 49 },
		{ 426, 315, 49 },
		{ 426, 316, 49 },
		{ 426, 317, 49 },
		{ 426, 318, 49 },
		{ 426, 319, 49 },
		{ 426, 320, 49 },
		{ 426, 321, 49 },
		{ 426, 322, 49 },
		{ 426, 323, 49 },
		{ 426, 324, 49 },
		{ 426, 325, 49 },
		{ 426, 326, 49 },
		{ 426, 327, 49 },
		{ 426, 328, 49 },
		{ 426, 329, 49 },
		{ 426, 330, 49 },
		{ 426, 331, 49 },
		{ 426, 332, 49 },
		{ 426, 333, 49 },
		{ 426, 334, 49 },
		{ 426, 335, 49 },
		{ 426, 336, 49 },
		{ 426, 337, 49 },
		{ 426, 338, 49 },
		{ 426, 339, 49 },
		{ 426, 340, 49 },
		{ 426, 341, 49 },
		{ 427, 300, 49 },
		{ 427, 301, 49 },
		{ 427, 302, 49 },
		{ 427, 303, 49 },
		{ 427, 304, 49 },
		{ 427, 305, 49 },
		{ 427, 306, 49 },
		{ 427, 307, 49 },
		{ 427, 308, 49 },
		{ 427, 309, 49 },
		{ 427, 310, 49 },
		{ 427, 311, 49 },
		{ 427, 312, 49 },
		{ 427, 313, 49 },
		{ 427, 314, 49 },
		{ 427, 315, 49 },
		{ 427, 316, 49 },
		{ 427, 317, 49 },
		{ 427, 318, 49 },
		{ 427, 319, 49 },
		{ 427, 320, 49 },
		{ 427, 321, 49 },
		{ 427, 322, 49 },
		{ 427, 323, 49 },
		{ 427, 324, 49 },
		{ 427, 325, 49 },
		{ 427, 326, 49 },
		{ 427, 327, 49 },
		{ 427, 328, 49 },
		{ 427, 329, 49 },
		{ 427, 330, 49 },
		{ 427, 331, 49 },
		{ 427, 332, 49 },
		{ 427, 333, 49 },
		{ 427, 334, 49 },
		{ 427, 335, 49 },
		{ 427, 336, 49 },
		{ 427, 337, 49 },
		{ 427, 338, 49 },
		{ 427, 339, 49 },
		{ 427, 340, 49 },
		{ 427, 341, 49 },
		{ 428, 300, 49 },
		{ 428, 301, 49 },
		{ 428, 302, 49 },
		{ 428, 303, 49 },
		{ 428, 304, 49 },
		{ 428, 305, 49 },
		{ 428, 306, 49 },
		{ 428, 307, 49 },
		{ 428, 308, 49 },
		{ 428, 309, 49 },
		{ 428, 310, 49 },
		{ 428, 311, 49 },
		{ 428, 312, 49 },
		{ 428, 313, 49 },
		{ 428, 314, 49 },
		{ 428, 315, 49 },
		{ 428, 316, 49 },
		{ 428, 317, 49 },
		{ 428, 318, 49 },
		{ 428, 319, 49 },
		{ 428, 320, 49 },
		{ 428, 321, 49 },
		{ 428, 322, 49 },
		{ 428, 323, 49 },
		{ 428, 324, 49 },
		{ 428, 325, 49 },
		{ 428, 326, 49 },
		{ 428, 327, 49 },
		{ 428, 328, 49 },
		{ 428, 329, 49 },
		{ 428, 330, 49 },
		{ 428, 331, 49 },
		{ 428, 332, 49 },
		{ 428, 333, 49 },
		{ 428, 334, 49 },
		{ 428, 335, 49 },
		{ 428, 336, 49 },
		{ 428, 337, 49 },
		{ 428, 338, 49 },
		{ 428, 339, 49 },
		{ 428, 340, 49 },
		{ 428, 341, 49 },
		{ 428, 342, 49 },
		{ 429, 301, 49 },
		{ 429, 302, 49 },
		{ 429, 303, 49 },
		{ 429, 304, 49 },
		{ 429, 305, 49 },
		{ 429, 306, 49 },
		{ 429, 307, 49 },
		{ 429, 308, 49 },
		{ 429, 309, 49 },
		{ 429, 310, 49 },
		{ 429, 311, 49 },
		{ 429, 312, 49 },
		{ 429, 313, 49 },
		{ 429, 314, 49 },
		{ 429, 315, 49 },
		{ 429, 316, 49 },
		{ 429, 317, 49 },
		{ 429, 318, 49 },
		{ 429, 319, 49 },
		{ 429, 320, 49 },
		{ 429, 321, 49 },
		{ 429, 322, 49 },
		{ 429, 323, 49 },
		{ 429, 324, 49 },
		{ 429, 325, 49 },
		{ 429, 326, 49 },
		{ 429, 327, 49 },
		{ 429, 328, 49 },
		{ 429, 329, 49 },
		{ 429, 330, 49 },
		{ 429, 331, 49 },
		{ 429, 332, 49 },
		{ 429, 333, 49 },
		{ 429, 334, 49 },
		{ 429, 335, 49 },
		{ 429, 336, 49 },
		{ 429, 337, 49 },
		{ 429, 338, 49 },
		{ 429, 339, 49 },
		{ 429, 340, 49 },
		{ 429, 341, 49 },
		{ 429, 342, 49 },
		{ 430, 301, 49 },
		{ 430, 302, 49 },
		{ 430, 303, 49 },
		{ 430, 304, 49 },
		{ 430, 305, 49 },
		{ 430, 306, 49 },
		{ 430, 307, 49 },
		{ 430, 308, 49 },
		{ 430, 309, 49 },
		{ 430, 310, 49 },
		{ 430, 311, 49 },
		{ 430, 312, 49 },
		{ 430, 313, 49 },
		{ 430, 314, 49 },
		{ 430, 315, 49 },
		{ 430, 316, 49 },
		{ 430, 317, 49 },
		{ 430, 318, 49 },
		{ 430, 319, 49 },
		{ 430, 320, 49 },
		{ 430, 321, 49 },
		{ 430, 322, 49 },
		{ 430, 323, 49 },
		{ 430, 324, 49 },
		{ 430, 325, 49 },
		{ 430, 326, 49 },
		{ 430, 327, 49 },
		{ 430, 328, 49 },
		{ 430, 329, 49 },
		{ 430, 330, 49 },
		{ 430, 331, 49 },
		{ 430, 332, 49 },
		{ 430, 333, 49 },
		{ 430, 334, 49 },
		{ 430, 335, 49 },
		{ 430, 336, 49 },
		{ 430, 337, 49 },
		{ 430, 338, 49 },
		{ 430, 339, 49 },
		{ 430, 340, 49 },
		{ 430, 341, 49 },
		{ 430, 342, 49 },
		{ 430, 343, 49 },
		{ 431, 301, 49 },
		{ 431, 302, 49 },
		{ 431, 303, 49 },
		{ 431, 304, 49 },
		{ 431, 305, 49 },
		{ 431, 306, 49 },
		{ 431, 307, 49 },
		{ 431, 308, 49 },
		{ 431, 309, 49 },
		{ 431, 310, 49 },
		{ 431, 311, 49 },
		{ 431, 312, 49 },
		{ 431, 313, 49 },
		{ 431, 314, 49 },
		{ 431, 315, 49 },
		{ 431, 316, 49 },
		{ 431, 317, 49 },
		{ 431, 318, 49 },
		{ 431, 319, 49 },
		{ 431, 320, 49 },
		{ 431, 321, 49 },
		{ 431, 322, 49 },
		{ 431, 323, 49 },
		{ 431, 324, 49 },
		{ 431, 325, 49 },
		{ 431, 326, 49 },
		{ 431, 327, 49 },
		{ 431, 328, 49 },
		{ 431, 329, 49 },
		{ 431, 330, 49 },
		{ 431, 331, 49 },
		{ 431, 332, 49 },
		{ 431, 333, 49 },
		{ 431, 334, 49 },
		{ 431, 335, 49 },
		{ 431, 336, 49 },
		{ 431, 337, 49 },
		{ 431, 338, 49 },
		{ 431, 339, 49 },
		{ 431, 340, 49 },
		{ 431, 341, 49 },
		{ 431, 342, 49 },
		{ 431, 343, 49 },
		{ 432, 301, 49 },
		{ 432, 302, 49 },
		{ 432, 303, 49 },
		{ 432, 304, 49 },
		{ 432, 305, 49 },
		{ 432, 306, 49 },
		{ 432, 307, 49 },
		{ 432, 308, 49 },
		{ 432, 309, 49 },
		{ 432, 310, 49 },
		{ 432, 311, 49 },
		{ 432, 312, 49 },
		{ 432, 313, 49 },
		{ 432, 314, 49 },
		{ 432, 315, 49 },
		{ 432, 316, 49 },
		{ 432, 317, 49 },
		{ 432, 318, 49 },
		{ 432, 319, 49 },
		{ 432, 320, 49 },
		{ 432, 321, 49 },
		{ 432, 322, 49 },
		{ 432, 323, 49 },
		{ 432, 324, 49 },
		{ 432, 325, 49 },
		{ 432, 326, 49 },
		{ 432, 327, 49 },
		{ 432, 328, 49 },
		{ 432, 329, 49 },
		{ 432, 330, 49 },
		{ 432, 331, 49 },
		{ 432, 332, 49 },
		{ 432, 333, 49 },
		{ 432, 334, 49 },
		{ 432, 335, 49 },
		{ 432, 336, 49 },
		{ 432, 337, 49 },
		{ 432, 338, 49 },
		{ 432, 339, 49 },
		{ 432, 340, 49 },
		{ 432, 341, 49 },
		{ 432, 342, 49 },
		{ 432, 343, 49 },
		{ 432, 344, 49 },
		{ 433, 302, 49 },
		{ 433, 303, 49 },
		{ 433, 304, 49 },
		{ 433, 305, 49 },
		{ 433, 306, 49 },
		{ 433, 307, 49 },
		{ 433, 308, 49 },
		{ 433, 309, 49 },
		{ 433, 310, 49 },
		{ 433, 311, 49 },
		{ 433, 312, 49 },
		{ 433, 313, 49 },
		{ 433, 314, 49 },
		{ 433, 315, 49 },
		{ 433, 316, 49 },
		{ 433, 317, 49 },
		{ 433, 318, 49 },
		{ 433, 319, 49 },
		{ 433, 320, 49 },
		{ 433, 321, 49 },
		{ 433, 322, 49 },
		{ 433, 323, 49 },
		{ 433, 324, 49 },
		{ 433, 325, 49 },
		{ 433, 326, 49 },
		{ 433, 327, 49 },
		{ 433, 328, 49 },
		{ 433, 329, 49 },
		{ 433, 330, 49 },
		{ 433, 331, 49 },
		{ 433, 332, 49 },
		{ 433, 333, 49 },
		{ 433, 334, 49 },
		{ 433, 335, 49 },
		{ 433, 336, 49 },
		{ 433, 337, 49 },
		{ 433, 338, 49 },
		{ 433, 339, 49 },
		{ 433, 340, 49 },
		{ 433, 341, 49 },
		{ 433, 342, 49 },
		{ 433, 343, 49 },
		{ 433, 344, 49 },
		{ 434, 302, 49 },
		{ 434, 303, 49 },
		{ 434, 304, 49 },
		{ 434, 305, 49 },
		{ 434, 306, 49 },
		{ 434, 307, 49 },
		{ 434, 308, 49 },
		{ 434, 309, 49 },
		{ 434, 310, 49 },
		{ 434, 311, 49 },
		{ 434, 312, 49 },
		{ 434, 313, 49 },
		{ 434, 314, 49 },
		{ 434, 315, 49 },
		{ 434, 316, 49 },
		{ 434, 317, 49 },
		{ 434, 318, 49 },
		{ 434, 319, 49 },
		{ 434, 320, 49 },
		{ 434, 321, 49 },
		{ 434, 322, 49 },
		{ 434, 323, 49 },
		{ 434, 324, 49 },
		{ 434, 325, 49 },
		{ 434, 326, 49 },
		{ 434, 327, 49 },
		{ 434, 328, 49 },
		{ 434, 329, 49 },
		{ 434, 330, 49 },
		{ 434, 331, 49 },
		{ 434, 332, 49 },
		{ 434, 333, 49 },
		{ 434, 334, 49 },
		{ 434, 335, 49 },
		{ 434, 336, 49 },
		{ 434, 337, 49 },
		{ 434, 338, 49 },
		{ 434, 339, 49 },
		{ 434, 340, 49 },
		{ 434, 341, 49 },
		{ 434, 342, 49 },
		{ 434, 343, 49 },
		{ 434, 344, 49 },
		{ 434, 345, 49 },
		{ 435, 302, 49 },
		{ 435, 303, 49 },
		{ 435, 304, 49 },
		{ 435, 305, 49 },
		{ 435, 306, 49 },
		{ 435, 307, 49 },
		{ 435, 308, 49 },
		{ 435, 309, 49 },
		{ 435, 310, 49 },
		{ 435, 311, 49 },
		{ 435, 312, 49 },
		{ 435, 313, 49 },
		{ 435, 314, 49 },
		{ 435, 315, 49 },
		{ 435, 316, 49 },
		{ 435, 317, 49 },
		{ 435, 318, 49 },
		{ 435, 319, 49 },
		{ 435, 320, 49 },
		{ 435, 321, 49 },
		{ 435, 322, 49 },
		{ 435, 323, 49 },
		{ 435, 324, 49 },
		{ 435, 325, 49 },
		{ 435, 326, 49 },
		{ 435, 327, 49 },
		{ 435, 328, 49 },
		{ 435, 329, 49 },
		{ 435, 330, 49 },
		{ 435, 331, 49 },
		{ 435, 332, 49 },
		{ 435, 333, 49 },
		{ 435, 334, 49 },
		{ 435, 335, 49 },
		{ 435, 336, 49 },
		{ 435, 337, 49 },
		{ 435, 338, 49 },
		{ 435, 339, 49 },
		{ 435, 340, 49 },
		{ 435, 341, 49 },
		{ 435, 342, 49 },
		{ 435, 343, 49 },
		{ 435, 344, 49 },
		{ 435, 345, 49 },
		{ 436, 302, 49 },
		{ 436, 303, 49 },
		{ 436, 304, 49 },
		{ 436, 305, 49 },
		{ 436, 306, 49 },
		{ 436, 307, 49 },
		{ 436, 308, 49 },
		{ 436, 309, 49 },
		{ 436, 310, 49 },
		{ 436, 311, 49 },
		{ 436, 312, 49 },
		{ 436, 313, 49 },
		{ 436, 314, 49 },
		{ 436, 315, 49 },
		{ 436, 316, 49 },
		{ 436, 317, 49 },
		{ 436, 318, 49 },
		{ 436, 319, 49 },
		{ 436, 320, 49 },
		{ 436, 321, 49 },
		{ 436, 322, 49 },
		{ 436, 323, 49 },
		{ 436, 324, 49 },
		{ 436, 325, 49 },
		{ 436, 326, 49 },
		{ 436, 327, 49 },
		{ 436, 328, 49 },
		{ 436, 329, 49 },
		{ 436, 330, 49 },
		{ 436, 331, 49 },
		{ 436, 332, 49 },
		{ 436, 333, 49 },
		{ 436, 334, 49 },
		{ 436, 335, 49 },
		{ 436, 336, 49 },
		{ 436, 337, 49 },
		{ 436, 338, 49 },
		{ 436, 339, 49 },
		{ 436, 340, 49 },
		{ 436, 341, 49 },
		{ 436, 342, 49 },
		{ 436, 343, 49 },
		{ 436, 344, 49 },
		{ 436, 345, 49 },
		{ 436, 346, 49 },
		{ 437, 303, 49 },
		{ 437, 304, 49 },
		{ 437, 305, 49 },
		{ 437, 306, 49 },
		{ 437, 307, 49 },
		{ 437, 308, 49 },
		{ 437, 309, 49 },
		{ 437, 310, 49 },
		{ 437, 311, 49 },
		{ 437, 312, 49 },
		{ 437, 313, 49 },
		{ 437, 314, 49 },
		{ 437, 315, 49 },
		{ 437, 316, 49 },
		{ 437, 317, 49 },
		{ 437, 318, 49 },
		{ 437, 319, 49 },
		{ 437, 320, 49 },
		{ 437, 321, 49 },
		{ 437, 322, 49 },
		{ 437, 323, 49 },
		{ 437, 324, 49 },
		{ 437, 325, 49 },
		{ 437, 326, 49 },
		{ 437, 327, 49 },
		{ 437, 328, 49 },
		{ 437, 329, 49 },
		{ 437, 330, 49 },
		{ 437, 331, 49 },
		{ 437, 332, 49 },
		{ 437, 333, 49 },
		{ 437, 334, 49 },
		{ 437, 335, 49 },
		{ 437, 336, 49 },
		{ 437, 337, 49 },
		{ 437, 338, 49 },
		{ 437, 339, 49 },
		{ 437, 340, 49 },
		{ 437, 341, 49 },
		{ 437, 342, 49 },
		{ 437, 343, 49 },
		{ 437, 344, 49 },
		{ 437, 345, 49 },
		{ 437, 346, 49 },
		{ 438, 303, 49 },
		{ 438, 304, 49 },
		{ 438, 305, 49 },
		{ 438, 306, 49 },
		{ 438, 307, 49 },
		{ 438, 308, 49 },
		{ 438, 309, 49 },
		{ 438, 310, 49 },
		{ 438, 311, 49 },
		{ 438, 312, 49 },
		{ 438, 313, 49 },
		{ 438, 314, 49 },
		{ 438, 315, 49 },
		{ 438, 316, 49 },
		{ 438, 317, 49 },
		{ 438, 318, 49 },
		{ 438, 319, 49 },
		{ 438, 320, 49 },
		{ 438, 321, 49 },
		{ 438, 322, 49 },
		{ 438, 323, 49 },
		{ 438, 324, 49 },
		{ 438, 325, 49 },
		{ 438, 326, 49 },
		{ 438, 327, 49 },
		{ 438, 328, 49 },
		{ 438, 329, 49 },
		{ 438, 330, 49 },
		{ 438, 331, 49 },
		{ 438, 332, 49 },
		{ 438, 333, 49 },
		{ 438, 334, 49 },
		{ 438, 335, 49 },
		{ 438, 336, 49 },
		{ 438, 337, 49 },
		{ 438, 338, 49 },
		{ 438, 339, 49 },
		{ 438, 340, 49 },
		{ 438, 341, 49 },
		{ 438, 342, 49 },
		{ 438, 343, 49 },
		{ 438, 344, 49 },
		{ 438, 345, 49 },
		{ 438, 346, 49 },
		{ 438, 347, 49 },
		{ 439, 303, 49 },
		{ 439, 304, 49 },
		{ 439, 305, 49 },
		{ 439, 306, 49 },
		{ 439, 307, 49 },
		{ 439, 308, 49 },
		{ 439, 309, 49 },
		{ 439, 310, 49 },
		{ 439, 311, 49 },
		{ 439, 312, 49 },
		{ 439, 313, 49 },
		{ 439, 314, 49 },
		{ 439, 315, 49 },
		{ 439, 316, 49 },
		{ 439, 317, 49 },
		{ 439, 318, 49 },
		{ 439, 319, 49 },
		{ 439, 320, 49 },
		{ 439, 321, 49 },
		{ 439, 322, 49 },
		{ 439, 323, 49 },
		{ 439, 324, 49 },
		{ 439, 325, 49 },
		{ 439, 326, 49 },
		{ 439, 327, 49 },
		{ 439, 328, 49 },
		{ 439, 329, 49 },
		{ 439, 330, 49 },
		{ 439, 331, 49 },
		{ 439, 332, 49 },
		{ 439, 333, 49 },
		{ 439, 334, 49 },
		{ 439, 335, 49 },
		{ 439, 336, 49 },
		{ 439, 337, 49 },
		{ 439, 338, 49 },
		{ 439, 339, 49 },
		{ 439, 340, 49 },
		{ 439, 341, 49 },
		{ 439, 342, 49 },
		{ 439, 343, 49 },
		{ 439, 344, 49 },
		{ 439, 345, 49 },
		{ 439, 346, 49 },
		{ 439, 347, 49 },
		{ 440, 303, 49 },
		{ 440, 304, 49 },
		{ 440, 305, 49 },
		{ 440, 306, 49 },
		{ 440, 307, 49 },
		{ 440, 308, 49 },
		{ 440, 309, 49 },
		{ 440, 310, 49 },
		{ 440, 311, 49 },
		{ 440, 312, 49 },
		{ 440, 313, 49 },
		{ 440, 314, 49 },
		{ 440, 315, 49 },
		{ 440, 316, 49 },
		{ 440, 317, 49 },
		{ 440, 318, 49 },
		{ 440, 319, 49 },
		{ 440, 320, 49 },
		{ 440, 321, 49 },
		{ 440, 322, 49 },
		{ 440, 323, 49 },
		{ 440, 324, 49 },
		{ 440, 325, 49 },
		{ 440, 326, 49 },
		{ 440, 327, 49 },
		{ 440, 328, 49 },
		{ 440, 329, 49 },
		{ 440, 330, 49 },
		{ 440, 331, 49 },
		{ 440, 332, 49 },
		{ 440, 333, 49 },
		{ 440, 334, 49 },
		{ 440, 335, 49 },
		{ 440, 336, 49 },
		{ 440, 337, 49 },
		{ 440, 338, 49 },
		{ 440, 339, 49 },
		{ 440, 340, 49 },
		{ 440, 341, 49 },
		{ 440, 342, 49 },
		{ 440, 343, 49 },
		{ 440, 344, 49 },
		{ 440, 345, 49 },
		{ 440, 346, 49 },
		{ 440, 347, 49 },
		{ 440, 348, 49 },
		{ 441, 304, 49 },
		{ 441, 305, 49 },
		{ 441, 306, 49 },
		{ 441, 307, 49 },
		{ 441, 308, 49 },
		{ 441, 309, 49 },
		{ 441, 310, 49 },
		{ 441, 311, 49 },
		{ 441, 312, 49 },
		{ 441, 313, 49 },
		{ 441, 314, 49 },
		{ 441, 315, 49 },
		{ 441, 316, 49 },
		{ 441, 317, 49 },
		{ 441, 318, 49 },
		{ 441, 319, 49 },
		{ 441, 320, 49 },
		{ 441, 321, 49 },
		{ 441, 322, 49 },
		{ 441, 323, 49 },
		{ 441, 324, 49 },
		{ 441, 325, 49 },
		{ 441, 326, 49 },
		{ 441, 327, 49 },
		{ 441, 328, 49 },
		{ 441, 329, 49 },
		{ 441, 330, 49 },
		{ 441, 331, 49 },
		{ 441, 332, 49 },
		{ 441, 333, 49 },
		{ 441, 334, 49 },
		{ 441, 335, 49 },
		{ 441, 336, 49 },
		{ 441, 337, 49 },
		{ 441, 338, 49 },
		{ 441, 339, 49 },
		{ 441, 340, 49 },
		{ 441, 341, 49 },
		{ 441, 342, 49 },
		{ 441, 343, 49 },
		{ 441, 344, 49 },
		{ 441, 345, 49 },
		{ 441, 346, 49 },
		{ 441, 347, 49 },
		{ 441, 348, 49 },
		{ 442, 304, 49 },
		{ 442, 305, 49 },
		{ 442, 306, 49 },
		{ 442, 307, 49 },
		{ 442, 308, 49 },
		{ 442, 309, 49 },
		{ 442, 310, 49 },
		{ 442, 311, 49 },
		{ 442, 312, 49 },
		{ 442, 313, 49 },
		{ 442, 314, 49 },
		{ 442, 315, 49 },
		{ 442, 316, 49 },
		{ 442, 317, 49 },
		{ 442, 318, 49 },
		{ 442, 319, 49 },
		{ 442, 320, 49 },
		{ 442, 321, 49 },
		{ 442, 322, 49 },
		{ 442, 323, 49 },
		{ 442, 324, 49 },
		{ 442, 325, 49 },
		{ 442, 326, 49 },
		{ 442, 327, 49 },
		{ 442, 328, 49 },
		{ 442, 329, 49 },
		{ 442, 330, 49 },
		{ 442, 331, 49 },
		{ 442, 332, 49 },
		{ 442, 333, 49 },
		{ 442, 334, 49 },
		{ 442, 335, 49 },
		{ 442, 336, 49 },
		{ 442, 337, 49 },
		{ 442, 338, 49 },
		{ 442, 339, 49 },
		{ 442, 340, 49 },
		{ 442, 341, 49 },
		{ 442, 342, 49 },
		{ 442, 343, 49 },
		{ 442, 344, 49 },
		{ 442, 345, 49 },
		{ 442, 346, 49 },
		{ 442, 347, 49 },
		{ 442, 348, 49 },
		{ 442, 349, 49 },
		{ 443, 304, 49 },
		{ 443, 305, 49 },
		{ 443, 306, 49 },
		{ 443, 307, 49 },
		{ 443, 308, 49 },
		{ 443, 309, 49 },
		{ 443, 310, 49 },
		{ 443, 311, 49 },
		{ 443, 312, 49 },
		{ 443, 313, 49 },
		{ 443, 314, 49 },
		{ 443, 315, 49 },
		{ 443, 316, 49 },
		{ 443, 317, 49 },
		{ 443, 318, 49 },
		{ 443, 319, 49 },
		{ 443, 320, 49 },
		{ 443, 321, 49 },
		{ 443, 322, 49 },
		{ 443, 323, 49 },
		{ 443, 324, 49 },
		{ 443, 325, 49 },
		{ 443, 326, 49 },
		{ 443, 327, 49 },
		{ 443, 328, 49 },
		{ 443, 329, 49 },
		{ 443, 330, 49 },
		{ 443, 331, 49 },
		{ 443, 332, 49 },
		{ 443, 333, 49 },
		{ 443, 334, 49 },
		{ 443, 335, 49 },
		{ 443, 336, 49 },
		{ 443, 337, 49 },
		{ 443, 338, 49 },
		{ 443, 339, 49 },
		{ 443, 340, 49 },
		{ 443, 341, 49 },
		{ 443, 342, 49 },
		{ 443, 343, 49 },
		{ 443, 344, 49 },
		{ 443, 345, 49 },
		{ 443, 346, 49 },
		{ 443, 347, 49 },
		{ 443, 348, 49 },
		{ 443, 349, 49 },
		{ 444, 304, 49 },
		{ 444, 305, 49 },
		{ 444, 306, 49 },
		{ 444, 307, 49 },
		{ 444, 308, 49 },
		{ 444, 309, 49 },
		{ 444, 310, 49 },
		{ 444, 311, 49 },
		{ 444, 312, 49 },
		{ 444, 313, 49 },
		{ 444, 314, 49 },
		{ 444, 315, 49 },
		{ 444, 316, 49 },
		{ 444, 317, 49 },
		{ 444, 318, 49 },
		{ 444, 319, 49 },
		{ 444, 320, 49 },
		{ 444, 321, 49 },
		{ 444, 322, 49 },
		{ 444, 323, 49 },
		{ 444, 324, 49 },
		{ 444, 325, 49 },
		{ 444, 326, 49 },
		{ 444, 327, 49 },
		{ 444, 328, 49 },
		{ 444, 329, 49 },
		{ 444, 330, 49 },
		{ 444, 331, 49 },
		{ 444, 332, 49 },
		{ 444, 333, 49 },
		{ 444, 334, 49 },
		{ 444, 335, 49 },
		{ 444, 336, 49 },
		{ 444, 337, 49 },
		{ 444, 338, 49 },
		{ 444, 339, 49 },
		{ 444, 340, 49 },
		{ 444, 341, 49 },
		{ 444, 342, 49 },
		{ 444, 343, 49 },
		{ 444, 344, 49 },
		{ 444, 345, 49 },
		{ 444, 346, 49 },
		{ 444, 347, 49 },
		{ 444, 348, 49 },
		{ 444, 349, 49 },
		{ 444, 350, 49 },
		{ 445, 305, 49 },
		{ 445, 306, 49 },
		{ 445, 307, 49 },
		{ 445, 308, 49 },
		{ 445, 309, 49 },
		{ 445, 310, 49 },
		{ 445, 311, 49 },
		{ 445, 312, 49 },
		{ 445, 313, 49 },
		{ 445, 314, 49 },
		{ 445, 315, 49 },
		{ 445, 316, 49 },
		{ 445, 317, 49 },
		{ 445, 318, 49 },
		{ 445, 319, 49 },
		{ 445, 320, 49 },
		{ 445, 321, 49 },
		{ 445, 322, 49 },
		{ 445, 323, 49 },
		{ 445, 324, 49 },
		{ 445, 325, 49 },
		{ 445, 326, 49 },
		{ 445, 327, 49 },
		{ 445, 328, 49 },
		{ 445, 329, 49 },
		{ 445, 330, 49 },
		{ 445, 331, 49 },
		{ 445, 332, 49 },
		{ 445, 333, 49 },
		{ 445, 334, 49 },
		{ 445, 335, 49 },
		{ 445, 336, 49 },
		{ 445, 337, 49 },
		{ 445, 338, 49 },
		{ 445, 339, 49 },
		{ 445, 340, 49 },
		{ 445, 341, 49 },
		{ 445, 342, 49 },
		{ 445, 343, 49 },
		{ 445, 344, 49 },
		{ 445, 345, 49 },
		{ 445, 346, 49 },
		{ 445, 347, 49 },
		{ 445, 348, 49 },
		{ 445, 349, 49 },
		{ 445, 350, 49 },
		{ 446, 305, 49 },
		{ 446, 306, 49 },
		{ 446, 307, 49 },
		{ 446, 308, 49 },
		{ 446, 309, 49 },
		{ 446, 310, 49 },
		{ 446, 311, 49 },
		{ 446, 312, 49 },
		{ 446, 313, 49 },
		{ 446, 314, 49 },
		{ 446, 315, 49 },
		{ 446, 316, 49 },
		{ 446, 317, 49 },
		{ 446, 318, 49 },
		{ 446, 319, 49 },
		{ 446, 320, 49 },
		{ 446, 321, 49 },
		{ 446, 322, 49 },
		{ 446, 323, 49 },
		{ 446, 324, 49 },
		{ 446, 325, 49 },
		{ 446, 326, 49 },
		{ 446, 327, 49 },
		{ 446, 328, 49 },
		{ 446, 329, 49 },
		{ 446, 330, 49 },
		{ 446, 331, 49 },
		{ 446, 332, 49 },
		{ 446, 333, 49 },
		{ 446, 334, 49 },
		{ 446, 335, 49 },
		{ 446, 336, 49 },
		{ 446, 337, 49 },
		{ 446, 338, 49 },
		{ 446, 339, 49 },
		{ 446, 340, 49 },
		{ 446, 341, 49 },
		{ 446, 342, 49 },
		{ 446, 343, 49 },
		{ 446, 344, 49 },
		{ 446, 345, 49 },
		{ 446, 346, 49 },
		{ 446, 347, 49 },
		{ 446, 348, 49 },
		{ 446, 349, 49 },
		{ 446, 350, 49 },
		{ 446, 351, 49 },
		{ 447, 305, 49 },
		{ 447, 306, 49 },
		{ 447, 307, 49 },
		{ 447, 308, 49 },
		{ 447, 309, 49 },
		{ 447, 310, 49 },
		{ 447, 311, 49 },
		{ 447, 312, 49 },
		{ 447, 313, 49 },
		{ 447, 314, 49 },
		{ 447, 315, 49 },
		{ 447, 316, 49 },
		{ 447, 317, 49 },
		{ 447, 318, 49 },
		{ 447, 319, 49 },
		{ 447, 320, 49 },
		{ 447, 321, 49 },
		{ 447, 322, 49 },
		{ 447, 323, 49 },
		{ 447, 324, 49 },
		{ 447, 325, 49 },
		{ 447, 326, 49 },
		{ 447, 327, 49 },
		{ 447, 328, 49 },
		{ 447, 329, 49 },
		{ 447, 330, 49 },
		{ 447, 331, 49 },
		{ 447, 332, 49 },
		{ 447, 333, 49 },
		{ 447, 334, 49 },
		{ 447, 335, 49 },
		{ 447, 336, 49 },
		{ 447, 337, 49 },
		{ 447, 338, 49 },
		{ 447, 339, 49 },
		{ 447, 340, 49 },
		{ 447, 341, 49 },
		{ 447, 342, 49 },
		{ 447, 343, 49 },
		{ 447, 344, 49 },
		{ 447, 345, 49 },
		{ 447, 346, 49 },
		{ 447, 347, 49 },
		{ 447, 348, 49 },
		{ 447, 349, 49 },
		{ 447, 350, 49 },
		{ 447, 351, 49 },
		{ 448, 305, 49 },
		{ 448, 306, 49 },
		{ 448, 307, 49 },
		{ 448, 308, 49 },
		{ 448, 309, 49 },
		{ 448, 310, 49 },
		{ 448, 311, 49 },
		{ 448, 312, 49 },
		{ 448, 313, 49 },
		{ 448, 314, 49 },
		{ 448, 315, 49 },
		{ 448, 316, 49 },
		{ 448, 317, 49 },
		{ 448, 318, 49 },
		{ 448, 319, 49 },
		{ 448, 320, 49 },
		{ 448, 321, 49 },
		{ 448, 322, 49 },
		{ 448, 323, 49 },
		{ 448, 324, 49 },
		{ 448, 325, 49 },
		{ 448, 326, 49 },
		{ 448, 327, 49 },
		{ 448, 328, 49 },
		{ 448, 329, 49 },
		{ 448, 330, 49 },
		{ 448, 331, 49 },
		{ 448, 332, 49 },
		{ 448, 333, 49 },
		{ 448, 334, 49 },
		{ 448, 335, 49 },
		{ 448, 336, 49 },
		{ 448, 337, 49 },
		{ 448, 338, 49 },
		{ 448, 339, 49 },
		{ 448, 340, 49 },
		{ 448, 341, 49 },
		{ 448, 342, 49 },
		{ 448, 343, 49 },
		{ 448, 344, 49 },
		{ 448, 345, 49 },
		{ 448, 346, 49 },
		{ 448, 347, 49 },
		{ 448, 348, 49 },
		{ 448, 349, 49 },
		{ 448, 350, 49 },
		{ 448, 351, 49 },
		{ 448, 352, 49 },
		{ 449, 306, 49 },
		{ 449, 307, 49 },
		{ 449, 308, 49 },
		{ 449, 309, 49 },
		{ 449, 310, 49 },
		{ 449, 311, 49 },
		{ 449, 312, 49 },
		{ 449, 313, 49 },
		{ 449, 314, 49 },
		{ 449, 315, 49 },
		{ 449, 316, 49 },
		{ 449, 317, 49 },
		{ 449, 318, 49 },
		{ 449, 319, 49 },
		{ 449, 320, 49 },
		{ 449, 321, 49 },
		{ 449, 322, 49 },
		{ 449, 323, 49 },
		{ 449, 324, 49 },
		{ 449, 325, 49 },
		{ 449, 326, 49 },
		{ 449, 327, 49 },
		{ 449, 328, 49 },
		{ 449, 329, 49 },
		{ 449, 330, 49 },
		{ 449, 331, 49 },
		{ 449, 332, 49 },
		{ 449, 333, 49 },
		{ 449, 334, 49 },
		{ 449, 335, 49 },
		{ 449, 336, 49 },
		{ 449, 337, 49 },
		{ 449, 338, 49 },
		{ 449, 339, 49 },
		{ 449, 340, 49 },
		{ 449, 341, 49 },
		{ 449, 342, 49 },
		{ 449, 343, 49 },
		{ 449, 344, 49 },
		{ 449, 345, 49 },
		{ 449, 346, 49 },
		{ 449, 347, 49 },
		{ 449, 348, 49 },
		{ 449, 349, 49 },
		{ 449, 350, 49 },
		{ 449, 351, 49 },
		{ 449, 352, 49 },
		{ 450, 306, 49 },
		{ 450, 307, 49 },
		{ 450, 308, 49 },
		{ 450, 309, 49 },
		{ 450, 310, 49 },
		{ 450, 311, 49 },
		{ 450, 312, 49 },
		{ 450, 313, 49 },
		{ 450, 314, 49 },
		{ 450, 315, 49 },
		{ 450, 316, 49 },
		{ 450, 317, 49 },
		{ 450, 318, 49 },
		{ 450, 319, 49 },
		{ 450, 320, 49 },
		{ 450, 321, 49 },
		{ 450, 322, 49 },
		{ 450, 323, 49 },
		{ 450, 324, 49 },
		{ 450, 325, 49 },
		{ 450, 326, 49 },
		{ 450, 327, 49 },
		{ 450, 328, 49 },
		{ 450, 329, 49 },
		{ 450, 330, 49 },
		{ 450, 331, 49 },
		{ 450, 332, 49 },
		{ 450, 333, 49 },
		{ 450, 334, 49 },
		{ 450, 335, 49 },
		{ 450, 336, 49 },
		{ 450, 337, 49 },
		{ 450, 338, 49 },
		{ 450, 339, 49 },
		{ 450, 340, 49 },
		{ 450, 341, 49 },
		{ 450, 342, 49 },
		{ 450, 343, 49 },
		{ 450, 344, 49 },
		{ 450, 345, 49 },
		{ 450, 346, 49 },
		{ 450, 347, 49 },
		{ 450, 348, 49 },
		{ 450, 349, 49 },
		{ 450, 350, 49 },
		{ 450, 351, 49 },
		{ 450, 352, 49 },
		{ 450, 353, 49 },
		{ 451, 306, 49 },
		{ 451, 307, 49 },
		{ 451, 308, 49 },
		{ 451, 309, 49 },
		{ 451, 310, 49 },
		{ 451, 311, 49 },
		{ 451, 312, 49 },
		{ 451, 313, 49 },
		{ 451, 314, 49 },
		{ 451, 315, 49 },
		{ 451, 316, 49 },
		{ 451, 317, 49 },
		{ 451, 318, 49 },
		{ 451, 319, 49 },
		{ 451, 320, 49 },
		{ 451, 321, 49 },
		{ 451, 322, 49 },
		{ 451, 323, 49 },
		{ 451, 324, 49 },
		{ 451, 325, 49 },
		{ 451, 326, 49 },
		{ 451, 327, 49 },
		{ 451, 328, 49 },
		{ 451, 329, 49 },
		{ 451, 330, 49 },
		{ 451, 331, 49 },
		{ 451, 332, 49 },
		{ 451, 333, 49 },
		{ 451, 334, 49 },
		{ 451, 335, 49 },
		{ 451, 336, 49 },
		{ 451, 337, 49 },
		{ 451, 338, 49 },
		{ 451, 339, 49 },
		{ 451, 340, 49 },
		{ 451, 341, 49 },
		{ 451, 342, 49 },
		{ 451, 343, 49 },
		{ 451, 344, 49 },
		{ 451, 345, 49 },
		{ 451, 346, 49 },
		{ 451, 347, 49 },
		{ 451, 348, 49 },
		{ 451, 349, 49 },
		{ 451, 350, 49 },
		{ 451, 351, 49 },
		{ 451, 352, 49 },
		{ 451, 353, 49 },
		{ 452, 306, 49 },
		{ 452, 307, 49 },
		{ 452, 308, 49 },
		{ 452, 309, 49 },
		{ 452, 310, 49 },
		{ 452, 311, 49 },
		{ 452, 312, 49 },
		{ 452, 313, 49 },
		{ 452, 314, 49 },
		{ 452, 315, 49 },
		{ 452, 316, 49 },
		{ 452, 317, 49 },
		{ 452, 318, 49 },
		{ 452, 319, 49 },
		{ 452, 320, 49 },
		{ 452, 321, 49 },
		{ 452, 322, 49 },
		{ 452, 323, 49 },
		{ 452, 324, 49 },
		{ 452, 325, 49 },
		{ 452, 326, 49 },
		{ 452, 327, 49 },
		{ 452, 328, 49 },
		{ 452, 329, 49 },
		{ 452, 330, 49 },
		{ 452, 331, 49 },
		{ 452, 332, 49 },
		{ 452, 333, 49 },
		{ 452, 334, 49 },
		{ 452, 335, 49 },
		{ 452, 336, 49 },
		{ 452, 337, 49 },
		{ 452, 338, 49 },
		{ 452, 339, 49 },
		{ 452, 340, 49 },
		{ 452, 341, 49 },
		{ 452, 342, 49 },
		{ 452, 343, 49 },
		{ 452, 344, 49 },
		{ 452, 345, 49 },
		{ 452, 346, 49 },
		{ 452, 347, 49 },
		{ 452, 348, 49 },
		{ 452, 349, 49 },
		{ 452, 350, 49 },
		{ 452, 351, 49 },
		{ 452, 352, 49 },
		{ 452, 353, 49 },
		{ 452, 354, 49 },
		{ 453, 307, 49 },
		{ 453, 308, 49 },
		{ 453, 309, 49 },
		{ 453, 310, 49 },
		{ 453, 311, 49 },
		{ 453, 312, 49 },
		{ 453, 313, 49 },
		{ 453, 314, 49 },
		{ 453, 315, 49 },
		{ 453, 316, 49 },
		{ 453, 317, 49 },
		{ 453, 318, 49 },
		{ 453, 319, 49 },
		{ 453, 320, 49 },
		{ 453, 321, 49 },
		{ 453, 322, 49 },
		{ 453, 323, 49 },
		{ 453, 324, 49 },
		{ 453, 325, 49 },
		{ 453, 326, 49 },
		{ 453, 327, 49 },
		{ 453, 328, 49 },
		{ 453, 329, 49 },
		{ 453, 330, 49 },
		{ 453, 331, 49 },
		{ 453, 332, 49 },
		{ 453, 333, 49 },
		{ 453, 334, 49 },
		{ 453, 335, 49 },
		{ 453, 336, 49 },
		{ 453, 337, 49 },
		{ 453, 338, 49 },
		{ 453, 339, 49 },
		{ 453, 340, 49 },
		{ 453, 341, 49 },
		{ 453, 342, 49 },
		{ 453, 343, 49 },
		{ 453, 344, 49 },
		{ 453, 345, 49 },
		{ 453, 346, 49 },
		{ 453, 347, 49 },
		{ 453, 348, 49 },
		{ 453, 349, 49 },
		{ 453, 350, 49 },
		{ 453, 351, 49 },
		{ 453, 352, 49 },
		{ 453, 353, 49 },
		{ 453, 354, 49 },
		{ 454, 307, 49 },
		{ 454, 308, 49 },
		{ 454, 309, 49 },
		{ 454, 310, 49 },
		{ 454, 311, 49 },
		{ 454, 312, 49 },
		{ 454, 313, 49 },
		{ 454, 314, 49 },
		{ 454, 315, 49 },
		{ 454, 316, 49 },
		{ 454, 317, 49 },
		{ 454, 318, 49 },
		{ 454, 319, 49 },
		{ 454, 320, 49 },
		{ 454, 321, 49 },
		{ 454, 322, 49 },
		{ 454, 323, 49 },
		{ 454, 324, 49 },
		{ 454, 325, 49 },
		{ 454, 326, 49 },
		{ 454, 327, 49 },
		{ 454, 328, 49 },
		{ 454, 329, 49 },
		{ 454, 330, 49 },
		{ 454, 331, 49 },
		{ 454, 332, 49 },
		{ 454, 333, 49 },
		{ 454, 334, 49 },
		{ 454, 335, 49 },
		{ 454, 336, 49 },
		{ 454, 337, 49 },
		{ 454, 338, 49 },
		{ 454, 339, 49 },
		{ 454, 340, 49 },
		{ 454, 341, 49 },
		{ 454, 342, 49 },
		{ 454, 343, 49 },
		{ 454, 344, 49 },
		{ 454, 345, 49 },
		{ 454, 346, 49 },
		{ 454, 347, 49 },
		{ 454, 348, 49 },
		{ 454, 349, 49 },
		{ 454, 350, 49 },
		{ 454, 351, 49 },
		{ 454, 352, 49 },
		{ 454, 353, 49 },
		{ 454, 354, 49 },
		{ 454, 355, 49 },
		{ 455, 307, 49 },
		{ 455, 308, 49 },
		{ 455, 309, 49 },
		{ 455, 310, 49 },
		{ 455, 311, 49 },
		{ 455, 312, 49 },
		{ 455, 313, 49 },
		{ 455, 314, 49 },
		{ 455, 315, 49 },
		{ 455, 316, 49 },
		{ 455, 317, 49 },
		{ 455, 318, 49 },
		{ 455, 319, 49 },
		{ 455, 320, 49 },
		{ 455, 321, 49 },
		{ 455, 322, 49 },
		{ 455, 323, 49 },
		{ 455, 324, 49 },
		{ 455, 325, 49 },
		{ 455, 326, 49 },
		{ 455, 327, 49 },
		{ 455, 328, 49 },
		{ 455, 329, 49 },
		{ 455, 330, 49 },
		{ 455, 331, 49 },
		{ 455, 332, 49 },
		{ 455, 333, 49 },
		{ 455, 334, 49 },
		{ 455, 335, 49 },
		{ 455, 336, 49 },
		{ 455, 337, 49 },
		{ 455, 338, 49 },
		{ 455, 339, 49 },
		{ 455, 340, 49 },
		{ 455, 341, 49 },
		{ 455, 342, 49 },
		{ 455, 343, 49 },
		{ 455, 344, 49 },
		{ 455, 345, 49 },
		{ 455, 346, 49 },
		{ 455, 347, 49 },
		{ 455, 348, 49 },
		{ 455, 349, 49 },
		{ 455, 350, 49 },
		{ 455, 351, 49 },
		{ 455, 352, 49 },
		{ 455, 353, 49 },
		{ 455, 354, 49 },
		{ 455, 355, 49 },
		{ 456, 307, 49 },
		{ 456, 308, 49 },
		{ 456, 309, 49 },
		{ 456, 310, 49 },
		{ 456, 311, 49 },
		{ 456, 312, 49 },
		{ 456, 313, 49 },
		{ 456, 314, 49 },
		{ 456, 315, 49 },
		{ 456, 316, 49 },
		{ 456, 317, 49 },
		{ 456, 318, 49 },
		{ 456, 319, 49 },
		{ 456, 320, 49 },
		{ 456, 321, 49 },
		{ 456, 322, 49 },
		{ 456, 323, 49 },
		{ 456, 324, 49 },
		{ 456, 325, 49 },
		{ 456, 326, 49 },
		{ 456, 327, 49 },
		{ 456, 328, 49 },
		{ 456, 329, 49 },
		{ 456, 330, 49 },
		{ 456, 331, 49 },
		{ 456, 332, 49 },
		{ 456, 333, 49 },
		{ 456, 334, 49 },
		{ 456, 335, 49 },
		{ 456, 336, 49 },
		{ 456, 337, 49 },
		{ 456, 338, 49 },
		{ 456, 339, 49 },
		{ 456, 340, 49 },
		{ 456, 341, 49 },
		{ 456, 342, 49 },
		{ 456, 343, 49 },
		{ 456, 344, 49 },
		{ 456, 345, 49 },
		{ 456, 346, 49 },
		{ 456, 347, 49 },
		{ 456, 348, 49 },
		{ 456, 349, 49 },
		{ 456, 350, 49 },
		{ 456, 351, 49 },
		{ 456, 352, 49 },
		{ 456, 353, 49 },
		{ 456, 354, 49 },
		{ 456, 355, 49 },
		{ 456, 356, 49 },
		{ 457, 308, 49 },
		{ 457, 309, 49 },
		{ 457, 310, 49 },
		{ 457, 311, 49 },
		{ 457, 312, 49 },
		{ 457, 313, 49 },
		{ 457, 314, 49 },
		{ 457, 315, 49 },
		{ 457, 316, 49 },
		{ 457, 317, 49 },
		{ 457, 318, 49 },
		{ 457, 319, 49 },
		{ 457, 320, 49 },
		{ 457, 321, 49 },
		{ 457, 322, 49 },
		{ 457, 323, 49 },
		{ 457, 324, 49 },
		{ 457, 325, 49 },
		{ 457, 326, 49 },
		{ 457, 327, 49 },
		{ 457, 328, 49 },
		{ 457, 329, 49 },
		{ 457, 330, 49 },
		{ 457, 331, 49 },
		{ 457, 332, 49 },
		{ 457, 333, 49 },
		{ 457, 334, 49 },
		{ 457, 335, 49 },
		{ 457, 336, 49 },
		{ 457, 337, 49 },
		{ 457, 338, 49 },
		{ 457, 339, 49 },
		{ 457, 340, 49 },
		{ 457, 341, 49 },
		{ 457, 342, 49 },
		{ 457, 343, 49 },
		{ 457, 344, 49 },
		{ 457, 345, 49 },
		{ 457, 346, 49 },
		{ 457, 347, 49 },
		{ 457, 348, 49 },
		{ 457, 349, 49 },
		{ 457, 350, 49 },
		{ 457, 351, 49 },
		{ 457, 352, 49 },
		{ 457, 353, 49 },
		{ 457, 354, 49 },
		{ 457, 355, 49 },
		{ 457, 356, 49 },
		{ 458, 308, 49 },
		{ 458, 309, 49 },
		{ 458, 310, 49 },
		{ 458, 311, 49 },
		{ 458, 312, 49 },
		{ 458, 313, 49 },
		{ 458, 314, 49 },
		{ 458, 315, 49 },
		{ 458, 316, 49 },
		{ 458, 317, 49 },
		{ 458, 318, 49 },
		{ 458, 319, 49 },
		{ 458, 320, 49 },
		{ 458, 321, 49 },
		{ 458, 322, 49 },
		{ 458, 323, 49 },
		{ 458, 324, 49 },
		{ 458, 325, 49 },
		{ 458, 326, 49 },
		{ 458, 327, 49 },
		{ 458, 328, 49 },
		{ 458, 329, 49 },
		{ 458, 330, 49 },
		{ 458, 331, 49 },
		{ 458, 332, 49 },
		{ 458, 333, 49 },
		{ 458, 334, 49 },
		{ 458, 335, 49 },
		{ 458, 336, 49 },
		{ 458, 337, 49 },
		{ 458, 338, 49 },
		{ 458, 339, 49 },
		{ 458, 340, 49 },
		{ 458, 341, 49 },
		{ 458, 342, 49 },
		{ 458, 343, 49 },
		{ 458, 344, 49 },
		{ 458, 345, 49 },
		{ 458, 346, 49 },
		{ 458, 347, 49 },
		{ 458, 348, 49 },
		{ 458, 349, 49 },
		{ 458, 350, 49 },
		{ 458, 351, 49 },
		{ 458, 352, 49 },
		{ 458, 353, 49 },
		{ 458, 354, 49 },
		{ 458, 355, 49 },
		{ 458, 356, 49 },
		{ 458, 357, 49 },
		{ 459, 308, 49 },
		{ 459, 309, 49 },
		{ 459, 310, 49 },
		{ 459, 311, 49 },
		{ 459, 312, 49 },
		{ 459, 313, 49 },
		{ 459, 314, 49 },
		{ 459, 315, 49 },
		{ 459, 316, 49 },
		{ 459, 317, 49 },
		{ 459, 318, 49 },
		{ 459, 319, 49 },
		{ 459, 320, 49 },
		{ 459, 321, 49 },
		{ 459, 322, 49 },
		{ 459, 323, 49 },
		{ 459, 324, 49 },
		{ 459, 325, 49 },
		{ 459, 326, 49 },
		{ 459, 327, 49 },
		{ 459, 328, 49 },
		{ 459, 329, 49 },
		{ 459, 330, 49 },
		{ 459, 331, 49 },
		{ 459, 332, 49 },
		{ 459, 333, 49 },
		{ 459, 334, 49 },
		{ 459, 335, 49 },
		{ 459, 336, 49 },
		{ 459, 337, 49 },
		{ 459, 338, 49 },
		{ 459, 339, 49 },
		{ 459, 340, 49 },
		{ 459, 341, 49 },
		{ 459, 342, 49 },
		{ 459, 343, 49 },
		{ 459, 344, 49 },
		{ 459, 345, 49 },
		{ 459, 346, 49 },
		{ 459, 347, 49 },
		{ 459, 348, 49 },
		{ 459, 349, 49 },
		{ 459, 350, 49 },
		{ 459, 351, 49 },
		{ 459, 352, 49 },
		{ 459, 353, 49 },
		{ 459, 354, 49 },
		{ 459, 355, 49 },
		{ 459, 356, 49 },
		{ 459, 357, 49 },
		{ 460, 308, 49 },
		{ 460, 309, 49 },
		{ 460, 310, 49 },
		{ 460, 311, 49 },
		{ 460, 312, 49 },
		{ 460, 313, 49 },
		{ 460, 314, 49 },
		{ 460, 315, 49 },
		{ 460, 316, 49 },
		{ 460, 317, 49 },
		{ 460, 318, 49 },
		{ 460, 319, 49 },
		{ 460, 320, 49 },
		{ 460, 321, 49 },
		{ 460, 322, 49 },
		{ 460, 323, 49 },
		{ 460, 324, 49 },
		{ 460, 325, 49 },
		{ 460, 326, 49 },
		{ 460, 327, 49 },
		{ 460, 328, 49 },
		{ 460, 329, 49 },
		{ 460, 330, 49 },
		{ 460, 331, 49 },
		{ 460, 332, 49 },
		{ 460, 333, 49 },
		{ 460, 334, 49 },
		{ 460, 335, 49 },
		{ 460, 336, 49 },
		{ 460, 337, 49 },
		{ 460, 338, 49 },
		{ 460, 339, 49 },
		{ 460, 340, 49 },
		{ 460, 341, 49 },
		{ 460, 342, 49 },
		{ 460, 343, 49 },
		{ 460, 344, 49 },
		{ 460, 345, 49 },
		{ 460, 346, 49 },
		{ 460, 347, 49 },
		{ 460, 348, 49 },
		{ 460, 349, 49 },
		{ 460, 350, 49 },
		{ 460, 351, 49 },
		{ 460, 352, 49 },
		{ 460, 353, 49 },
		{ 460, 354, 49 },
		{ 460, 355, 49 },
		{ 460, 356, 49 },
		{ 460, 357, 49 },
		{ 460, 358, 49 },
		{ 461, 309, 49 },
		{ 461, 310, 49 },
		{ 461, 311, 49 },
		{ 461, 312, 49 },
		{ 461, 313, 49 },
		{ 461, 314, 49 },
		{ 461, 315, 49 },
		{ 461, 316, 49 },
		{ 461, 317, 49 },
		{ 461, 318, 49 },
		{ 461, 319, 49 },
		{ 461, 320, 49 },
		{ 461, 321, 49 },
		{ 461, 322, 49 },
		{ 461, 323, 49 },
		{ 461, 324, 49 },
		{ 461, 325, 49 },
		{ 461, 326, 49 },
		{ 461, 327, 49 },
		{ 461, 328, 49 },
		{ 461, 329, 49 },
		{ 461, 330, 49 },
		{ 461, 331, 49 },
		{ 461, 332, 49 },
		{ 461, 333, 49 },
		{ 461, 334, 49 },
		{ 461, 335, 49 },
		{ 461, 336, 49 },
		{ 461, 337, 49 },
		{ 461, 338, 49 },
		{ 461, 339, 49 },
		{ 461, 340, 49 },
		{ 461, 341, 49 },
		{ 461, 342, 49 },
		{ 461, 343, 49 },
		{ 461, 344, 49 },
		{ 461, 345, 49 },
		{ 461, 346, 49 },
		{ 461, 347, 49 },
		{ 461, 348, 49 },
		{ 461, 349, 49 },
		{ 461, 350, 49 },
		{ 461, 351, 49 },
		{ 461, 352, 49 },
		{ 461, 353, 49 },
		{ 461, 354, 49 },
		{ 461, 355, 49 },
		{ 461, 356, 49 },
		{ 461, 357, 49 },
		{ 461, 358, 49 },
		{ 462, 309, 49 },
		{ 462, 310, 49 },
		{ 462, 311, 49 },
		{ 462, 312, 49 },
		{ 462, 313, 49 },
		{ 462, 314, 49 },
		{ 462, 315, 49 },
		{ 462, 316, 49 },
		{ 462, 317, 49 },
		{ 462, 318, 49 },
		{ 462, 319, 49 },
		{ 462, 320, 49 },
		{ 462, 321, 49 },
		{ 462, 322, 49 },
		{ 462, 323, 49 },
		{ 462, 324, 49 },
		{ 462, 325, 49 },
		{ 462, 326, 49 },
		{ 462, 327, 49 },
		{ 462, 328, 49 },
		{ 462, 329, 49 },
		{ 462, 330, 49 },
		{ 462, 331, 49 },
		{ 462, 332, 49 },
		{ 462, 333, 49 },
		{ 462, 334, 49 },
		{ 462, 335, 49 },
		{ 462, 336, 49 },
		{ 462, 337, 49 },
		{ 462, 338, 49 },
		{ 462, 339, 49 },
		{ 462, 340, 49 },
		{ 462, 341, 49 },
		{ 462, 342, 49 },
		{ 462, 343, 49 },
		{ 462, 344, 49 },
		{ 462, 345, 49 },
		{ 462, 346, 49 },
		{ 462, 347, 49 },
		{ 462, 348, 49 },
		{ 462, 349, 49 },
		{ 462, 350, 49 },
		{ 462, 351, 49 },
		{ 462, 352, 49 },
		{ 462, 353, 49 },
		{ 462, 354, 49 },
		{ 462, 355, 49 },
		{ 462, 356, 49 },
		{ 462, 357, 49 },
		{ 462, 358, 49 },
		{ 462, 359, 49 },
		{ 463, 309, 49 },
		{ 463, 310, 49 },
		{ 463, 311, 49 },
		{ 463, 312, 49 },
		{ 463, 313, 49 },
		{ 463, 314, 49 },
		{ 463, 315, 49 },
		{ 463, 316, 49 },
		{ 463, 317, 49 },
		{ 463, 318, 49 },
		{ 463, 319, 49 },
		{ 463, 320, 49 },
		{ 463, 321, 49 },
		{ 463, 322, 49 },
		{ 463, 323, 49 },
		{ 463, 324, 49 },
		{ 463, 325, 49 },
		{ 463, 326, 49 },
		{ 463, 327, 49 },
		{ 463, 328, 49 },
		{ 463, 329, 49 },
		{ 463, 330, 49 },
		{ 463, 331, 49 },
		{ 463, 332, 49 },
		{ 463, 333, 49 },
		{ 463, 334, 49 },
		{ 463, 335, 49 },
		{ 463, 336, 49 },
		{ 463, 337, 49 },
		{ 463, 338, 49 },
		{ 463, 339, 49 },
		{ 463, 340, 49 },
		{ 463, 341, 49 },
		{ 463, 342, 49 },
		{ 463, 343, 49 },
		{ 463, 344, 49 },
		{ 463, 345, 49 },
		{ 463, 346, 49 },
		{ 463, 347, 49 },
		{ 463, 348, 49 },
		{ 463, 349, 49 },
		{ 463, 350, 49 },
		{ 463, 351, 49 },
		{ 463, 352, 49 },
		{ 463, 353, 49 },
		{ 463, 354, 49 },
		{ 463, 355, 49 },
		{ 463, 356, 49 },
		{ 463, 357, 49 },
		{ 463, 358, 49 },
		{ 463, 359, 49 },
		{ 464, 309, 49 },
		{ 464, 310, 49 },
		{ 464, 311, 49 },
		{ 464, 312, 49 },
		{ 464, 313, 49 },
		{ 464, 314, 49 },
		{ 464, 315, 49 },
		{ 464, 316, 49 },
		{ 464, 317, 49 },
		{ 464, 318, 49 },
		{ 464, 319, 49 },
		{ 464, 320, 49 },
		{ 464, 321, 49 },
		{ 464, 322, 49 },
		{ 464, 323, 49 },
		{ 464, 324, 49 },
		{ 464, 325, 49 },
		{ 464, 326, 49 },
		{ 464, 327, 49 },
		{ 464, 328, 49 },
		{ 464, 329, 49 },
		{ 464, 330, 49 },
		{ 464, 331, 49 },
		{ 464, 332, 49 },
		{ 464, 333, 49 },
		{ 464, 334, 49 },
		{ 464, 335, 49 },
		{ 464, 336, 49 },
		{ 464, 337, 49 },
		{ 464, 338, 49 },
		{ 464, 339, 49 },
		{ 464, 340, 49 },
		{ 464, 341, 49 },
		{ 464, 342, 49 },
		{ 464, 343, 49 },
		{ 464, 344, 49 },
		{ 464, 345, 49 },
		{ 464, 346, 49 },
		{ 464, 347, 49 },
		{ 464, 348, 49 },
		{ 464, 349, 49 },
		{ 464, 350, 49 },
		{ 464, 351, 49 },
		{ 464, 352, 49 },
		{ 464, 353, 49 },
		{ 464, 354, 49 },
		{ 464, 355, 49 },
		{ 464, 356, 49 },
		{ 464, 357, 49 },
		{ 464, 358, 49 },
		{ 464, 359, 49 },
		{ 464, 360, 49 },
		{ 465, 310, 49 },
		{ 465, 311, 49 },
		{ 465, 312, 49 },
		{ 465, 313, 49 },
		{ 465, 314, 49 },
		{ 465, 315, 49 },
		{ 465, 316, 49 },
		{ 465, 317, 49 },
		{ 465, 318, 49 },
		{ 465, 319, 49 },
		{ 465, 320, 49 },
		{ 465, 321, 49 },
		{ 465, 322, 49 },
		{ 465, 323, 49 },
		{ 465, 324, 49 },
		{ 465, 325, 49 },
		{ 465, 326, 49 },
		{ 465, 327, 49 },
		{ 465, 328, 49 },
		{ 465, 329, 49 },
		{ 465, 330, 49 },
		{ 465, 331, 49 },
		{ 465, 332, 49 },
		{ 465, 333, 49 },
		{ 465, 334, 49 },
		{ 465, 335, 49 },
		{ 465, 336, 49 },
		{ 465, 337, 49 },
		{ 465, 338, 49 },
		{ 465, 339, 49 },
		{ 465, 340, 49 },
		{ 465, 341, 49 },
		{ 465, 342, 49 },
		{ 465, 343, 49 },
		{ 465, 344, 49 },
		{ 465, 345, 49 },
		{ 465, 346, 49 },
		{ 465, 347, 49 },
		{ 465, 348, 49 },
		{ 465, 349, 49 },
		{ 465, 350, 49 },
		{ 465, 351, 49 },
		{ 465, 352, 49 },
		{ 465, 353, 49 },
		{ 465, 354, 49 },
		{ 465, 355, 49 },
		{ 465, 356, 49 },
		{ 465, 357, 49 },
		{ 465, 358, 49 },
		{ 465, 359, 49 },
		{ 465, 360, 49 },
		{ 466, 310, 49 },
		{ 466, 311, 49 },
		{ 466, 312, 49 },
		{ 466, 313, 49 },
		{ 466, 314, 49 },
		{ 466, 315, 49 },
		{ 466, 316, 49 },
		{ 466, 317, 49 },
		{ 466, 318, 49 },
		{ 466, 319, 49 },
		{ 466, 320, 49 },
		{ 466, 321, 49 },
		{ 466, 322, 49 },
		{ 466, 323, 49 },
		{ 466, 324, 49 },
		{ 466, 325, 49 },
		{ 466, 326, 49 },
		{ 466, 327, 49 },
		{ 466, 328, 49 },
		{ 466, 329, 49 },
		{ 466, 330, 49 },
		{ 466, 331, 49 },
		{ 466, 332, 49 },
		{ 466, 333, 49 },
		{ 466, 334, 49 },
		{ 466, 335, 49 },
		{ 466, 336, 49 },
		{ 466, 337, 49 },
		{ 466, 338, 49 },
		{ 466, 339, 49 },
		{ 466, 340, 49 },
		{ 466, 341, 49 },
		{ 466, 342, 49 },
		{ 466, 343, 49 },
		{ 466, 344, 49 },
		{ 466, 345, 49 },
		{ 466, 346, 49 },
		{ 466, 347, 49 },
		{ 466, 348, 49 },
		{ 466, 349, 49 },
		{ 466, 350, 49 },
		{ 466, 351, 49 },
		{ 466, 352, 49 },
		{ 466, 353, 49 },
		{ 466, 354, 49 },
		{ 466, 355, 49 },
		{ 466, 356, 49 },
		{ 466, 357, 49 },
		{ 466, 358, 49 },
		{ 466, 359, 49 },
		{ 466, 360, 49 },
		{ 466, 361, 49 },
		{ 467, 310, 49 },
		{ 467, 311, 49 },
		{ 467, 312, 49 },
		{ 467, 313, 49 },
		{ 467, 314, 49 },
		{ 467, 315, 49 },
		{ 467, 316, 49 },
		{ 467, 317, 49 },
		{ 467, 318, 49 },
		{ 467, 319, 49 },
		{ 467, 320, 49 },
		{ 467, 321, 49 },
		{ 467, 322, 49 },
		{ 467, 323, 49 },
		{ 467, 324, 49 },
		{ 467, 325, 49 },
		{ 467, 326, 49 },
		{ 467, 327, 49 },
		{ 467, 328, 49 },
		{ 467, 329, 49 },
		{ 467, 330, 49 },
		{ 467, 331, 49 },
		{ 467, 332, 49 },
		{ 467, 333, 49 },
		{ 467, 334, 49 },
		{ 467, 335, 49 },
		{ 467, 336, 49 },
		{ 467, 337, 49 },
		{ 467, 338, 49 },
		{ 467, 339, 49 },
		{ 467, 340, 49 },
		{ 467, 341, 49 },
		{ 467, 342, 49 },
		{ 467, 343, 49 },
		{ 467, 344, 49 },
		{ 467, 345, 49 },
		{ 467, 346, 49 },
		{ 467, 347, 49 },
		{ 467, 348, 49 },
		{ 467, 349, 49 },
		{ 467, 350, 49 },
		{ 467, 351, 49 },
		{ 467, 352, 49 },
		{ 467, 353, 49 },
		{ 467, 354, 49 },
		{ 467, 355, 49 },
		{ 467, 356, 49 },
		{ 467, 357, 49 },
		{ 467, 358, 49 },
		{ 467, 359, 49 },
		{ 467, 360, 49 },
		{ 467, 361, 49 },
		{ 468, 310, 49 },
		{ 468, 311, 49 },
		{ 468, 312, 49 },
		{ 468, 313, 49 },
		{ 468, 314, 49 },
		{ 468, 315, 49 },
		{ 468, 316, 49 },
		{ 468, 317, 49 },
		{ 468, 318, 49 },
		{ 468, 319, 49 },
		{ 468, 320, 49 },
		{ 468, 321, 49 },
		{ 468, 322, 49 },
		{ 468, 323, 49 },
		{ 468, 324, 49 },
		{ 468, 325, 49 },
		{ 468, 326, 49 },
		{ 468, 327, 49 },
		{ 468, 328, 49 },
		{ 468, 329, 49 },
		{ 468, 330, 49 },
		{ 468, 331, 49 },
		{ 468, 332, 49 },
		{ 468, 333, 49 },
		{ 468, 334, 49 },
		{ 468, 335, 49 },
		{ 468, 336, 49 },
		{ 468, 337, 49 },
		{ 468, 338, 49 },
		{ 468, 339, 49 },
		{ 468, 340, 49 },
		{ 468, 341, 49 },
		{ 468, 342, 49 },
		{ 468, 343, 49 },
		{ 468, 344, 49 },
		{ 468, 345, 49 },
		{ 468, 346, 49 },
		{ 468, 347, 49 },
		{ 468, 348, 49 },
		{ 468, 349, 49 },
		{ 468, 350, 49 },
		{ 468, 351, 49 },
		{ 468, 352, 49 },
		{ 468, 353, 49 },
		{ 468, 354, 49 },
		{ 468, 355, 49 },
		{ 468, 356, 49 },
		{ 468, 357, 49 },
		{ 468, 358, 49 },
		{ 468, 359, 49 },
		{ 468, 360, 49 },
		{ 468, 361, 49 },
		{ 468, 362, 49 },
		{ 469, 311, 49 },
		{ 469, 312, 49 },
		{ 469, 313, 49 },
		{ 469, 314, 49 },
		{ 469, 315, 49 },
		{ 469, 316, 49 },
		{ 469, 317, 49 },
		{ 469, 318, 49 },
		{ 469, 319, 49 },
		{ 469, 320, 49 },
		{ 469, 321, 49 },
		{ 469, 322, 49 },
		{ 469, 323, 49 },
		{ 469, 324, 49 },
		{ 469, 325, 49 },
		{ 469, 326, 49 },
		{ 469, 327, 49 },
		{ 469, 328, 49 },
		{ 469, 329, 49 },
		{ 469, 330, 49 },
		{ 469, 331, 49 },
		{ 469, 332, 49 },
		{ 469, 333, 49 },
		{ 469, 334, 49 },
		{ 469, 335, 49 },
		{ 469, 336, 49 },
		{ 469, 337, 49 },
		{ 469, 338, 49 },
		{ 469, 339, 49 },
		{ 469, 340, 49 },
		{ 469, 341, 49 },
		{ 469, 342, 49 },
		{ 469, 343, 49 },
		{ 469, 344, 49 },
		{ 469, 345, 49 },
		{ 469, 346, 49 },
		{ 469, 347, 49 },
		{ 469, 348, 49 },
		{ 469, 349, 49 },
		{ 469, 350, 49 },
		{ 469, 351, 49 },
		{ 469, 352, 49 },
		{ 469, 353, 49 },
		{ 469, 354, 49 },
		{ 469, 355, 49 },
		{ 469, 356, 49 },
		{ 469, 357, 49 },
		{ 469, 358, 49 },
		{ 469, 359, 49 },
		{ 469, 360, 49 },
		{ 469, 361, 49 },
		{ 469, 362, 49 },
		{ 470, 311, 49 },
		{ 470, 312, 49 },
		{ 470, 313, 49 },
		{ 470, 314, 49 },
		{ 470, 315, 49 },
		{ 470, 316, 49 },
		{ 470, 317, 49 },
		{ 470, 318, 49 },
		{ 470, 319, 49 },
		{ 470, 320, 49 },
		{ 470, 321, 49 },
		{ 470, 322, 49 },
		{ 470, 323, 49 },
		{ 470, 324, 49 },
		{ 470, 325, 49 },
		{ 470, 326, 49 },
		{ 470, 327, 49 },
		{ 470, 328, 49 },
		{ 470, 329, 49 },
		{ 470, 330, 49 },
		{ 470, 331, 49 },
		{ 470, 332, 49 },
		{ 470, 333, 49 },
		{ 470, 334, 49 },
		{ 470, 335, 49 },
		{ 470, 336, 49 },
		{ 470, 337, 49 },
		{ 470, 338, 49 },
		{ 470, 339, 49 },
		{ 470, 340, 49 },
		{ 470, 341, 49 },
		{ 470, 342, 49 },
		{ 470, 343, 49 },
		{ 470, 344, 49 },
		{ 470, 345, 49 },
		{ 470, 346, 49 },
		{ 470, 347, 49 },
		{ 470, 348, 49 },
		{ 470, 349, 49 },
		{ 470, 350, 49 },
		{ 470, 351, 49 },
		{ 470, 352, 49 },
		{ 470, 353, 49 },
		{ 470, 354, 49 },
		{ 470, 355, 49 },
		{ 470, 356, 49 },
		{ 470, 357, 49 },
		{ 470, 358, 49 },
		{ 470, 359, 49 },
		{ 470, 360, 49 },
		{ 470, 361, 49 },
		{ 470, 362, 49 },
		{ 470, 363, 49 },
		{ 471, 311, 49 },
		{ 471, 312, 49 },
		{ 471, 313, 49 },
		{ 471, 314, 49 },
		{ 471, 315, 49 },
		{ 471, 316, 49 },
		{ 471, 317, 49 },
		{ 471, 318, 49 },
		{ 471, 319, 49 },
		{ 471, 320, 49 },
		{ 471, 321, 49 },
		{ 471, 322, 49 },
		{ 471, 323, 49 },
		{ 471, 324, 49 },
		{ 471, 325, 49 },
		{ 471, 326, 49 },
		{ 471, 327, 49 },
		{ 471, 328, 49 },
		{ 471, 329, 49 },
		{ 471, 330, 49 },
		{ 471, 331, 49 },
		{ 471, 332, 49 },
		{ 471, 333, 49 },
		{ 471, 334, 49 },
		{ 471, 335, 49 },
		{ 471, 336, 49 },
		{ 471, 337, 49 },
		{ 471, 338, 49 },
		{ 471, 339, 49 },
		{ 471, 340, 49 },
		{ 471, 341, 49 },
		{ 471, 342, 49 },
		{ 471, 343, 49 },
		{ 471, 344, 49 },
		{ 471, 345, 49 },
		{ 471, 346, 49 },
		{ 471, 347, 49 },
		{ 471, 348, 49 },
		{ 471, 349, 49 },
		{ 471, 350, 49 },
		{ 471, 351, 49 },
		{ 471, 352, 49 },
		{ 471, 353, 49 },
		{ 471, 354, 49 },
		{ 471, 355, 49 },
		{ 471, 356, 49 },
		{ 471, 357, 49 },
		{ 471, 358, 49 },
		{ 471, 359, 49 },
		{ 471, 360, 49 },
		{ 471, 361, 49 },
		{ 471, 362, 49 },
		{ 471, 363, 49 },
		{ 472, 311, 49 },
		{ 472, 312, 49 },
		{ 472, 313, 49 },
		{ 472, 314, 49 },
		{ 472, 315, 49 },
		{ 472, 316, 49 },
		{ 472, 317, 49 },
		{ 472, 318, 49 },
		{ 472, 319, 49 },
		{ 472, 320, 49 },
		{ 472, 321, 49 },
		{ 472, 322, 49 },
		{ 472, 323, 49 },
		{ 472, 324, 49 },
		{ 472, 325, 49 },
		{ 472, 326, 49 },
		{ 472, 327, 49 },
		{ 472, 328, 49 },
		{ 472, 329, 49 },
		{ 472, 330, 49 },
		{ 472, 331, 49 },
		{ 472, 332, 49 },
		{ 472, 333, 49 },
		{ 472, 334, 49 },
		{ 472, 335, 49 },
		{ 472, 336, 49 },
		{ 472, 337, 49 },
		{ 472, 338, 49 },
		{ 472, 339, 49 },
		{ 472, 340, 49 },
		{ 472, 341, 49 },
		{ 472, 342, 49 },
		{ 472, 343, 49 },
		{ 472, 344, 49 },
		{ 472, 345, 49 },
		{ 472, 346, 49 },
		{ 472, 347, 49 },
		{ 472, 348, 49 },
		{ 472, 349, 49 },
		{ 472, 350, 49 },
		{ 472, 351, 49 },
		{ 472, 352, 49 },
		{ 472, 353, 49 },
		{ 472, 354, 49 },
		{ 472, 355, 49 },
		{ 472, 356, 49 },
		{ 472, 357, 49 },
		{ 472, 358, 49 },
		{ 472, 359, 49 },
		{ 472, 360, 49 },
		{ 472, 361, 49 },
		{ 472, 362, 49 },
		{ 472, 363, 49 },
		{ 472, 364, 49 },
		{ 473, 312, 49 },
		{ 473, 313, 49 },
		{ 473, 314, 49 },
		{ 473, 315, 49 },
		{ 473, 316, 49 },
		{ 473, 317, 49 },
		{ 473, 318, 49 },
		{ 473, 319, 49 },
		{ 473, 320, 49 },
		{ 473, 321, 49 },
		{ 473, 322, 49 },
		{ 473, 323, 49 },
		{ 473, 324, 49 },
		{ 473, 325, 49 },
		{ 473, 326, 49 },
		{ 473, 327, 49 },
		{ 473, 328, 49 },
		{ 473, 329, 49 },
		{ 473, 330, 49 },
		{ 473, 331, 49 },
		{ 473, 332, 49 },
		{ 473, 333, 49 },
		{ 473, 334, 49 },
		{ 473, 335, 49 },
		{ 473, 336, 49 },
		{ 473, 337, 49 },
		{ 473, 338, 49 },
		{ 473, 339, 49 },
		{ 473, 340, 49 },
		{ 473, 341, 49 },
		{ 473, 342, 49 },
		{ 473, 343, 49 },
		{ 473, 344, 49 },
		{ 473, 345, 49 },
		{ 473, 346, 49 },
		{ 473, 347, 49 },
		{ 473, 348, 49 },
		{ 473, 349, 49 },
		{ 473, 350, 49 },
		{ 473, 351, 49 },
		{ 473, 352, 49 },
		{ 473, 353, 49 },
		{ 473, 354, 49 },
		{ 473, 355, 49 },
		{ 473, 356, 49 },
		{ 473, 357, 49 },
		{ 473, 358, 49 },
		{ 473, 359, 49 },
		{ 473, 360, 49 },
		{ 473, 361, 49 },
		{ 473, 362, 49 },
		{ 473, 363, 49 },
		{ 473, 364, 49 },
		{ 474, 312, 49 },
		{ 474, 313, 49 },
		{ 474, 314, 49 },
		{ 474, 315, 49 },
		{ 474, 316, 49 },
		{ 474, 317, 49 },
		{ 474, 318, 49 },
		{ 474, 319, 49 },
		{ 474, 320, 49 },
		{ 474, 321, 49 },
		{ 474, 322, 49 },
		{ 474, 323, 49 },
		{ 474, 324, 49 },
		{ 474, 325, 49 },
		{ 474, 326, 49 },
		{ 474, 327, 49 },
		{ 474, 328, 49 },
		{ 474, 329, 49 },
		{ 474, 330, 49 },
		{ 474, 331, 49 },
		{ 474, 332, 49 },
		{ 474, 333, 49 },
		{ 474, 334, 49 },
		{ 474, 335, 49 },
		{ 474, 336, 49 },
		{ 474, 337, 49 },
		{ 474, 338, 49 },
		{ 474, 339, 49 },
		{ 474, 340, 49 },
		{ 474, 341, 49 },
		{ 474, 342, 49 },
		{ 474, 343, 49 },
		{ 474, 344, 49 },
		{ 474, 345, 49 },
		{ 474, 346, 49 },
		{ 474, 347, 49 },
		{ 474, 348, 49 },
		{ 474, 349, 49 },
		{ 474, 350, 49 },
		{ 474, 351, 49 },
		{ 474, 352, 49 },
		{ 474, 353, 49 },
		{ 474, 354, 49 },
		{ 474, 355, 49 },
		{ 474, 356, 49 },
		{ 474, 357, 49 },
		{ 474, 358, 49 },
		{ 474, 359, 49 },
		{ 474, 360, 49 },
		{ 474, 361, 49 },
		{ 474, 362, 49 },
		{ 474, 363, 49 },
		{ 474, 364, 49 },
		{ 474, 365, 49 },
		{ 475, 312, 49 },
		{ 475, 313, 49 },
		{ 475, 314, 49 },
		{ 475, 315, 49 },
		{ 475, 316, 49 },
		{ 475, 317, 49 },
		{ 475, 318, 49 },
		{ 475, 319, 49 },
		{ 475, 320, 49 },
		{ 475, 321, 49 },
		{ 475, 322, 49 },
		{ 475, 323, 49 },
		{ 475, 324, 49 },
		{ 475, 325, 49 },
		{ 475, 326, 49 },
		{ 475, 327, 49 },
		{ 475, 328, 49 },
		{ 475, 329, 49 },
		{ 475, 330, 49 },
		{ 475, 331, 49 },
		{ 475, 332, 49 },
		{ 475, 333, 49 },
		{ 475, 334, 49 },
		{ 475, 335, 49 },
		{ 475, 336, 49 },
		{ 475, 337, 49 },
		{ 475, 338, 49 },
		{ 475, 339, 49 },
		{ 475, 340, 49 },
		{ 475, 341, 49 },
		{ 475, 342, 49 },
		{ 475, 343, 49 },
		{ 475, 344, 49 },
		{ 475, 345, 49 },
		{ 475, 346, 49 },
		{ 475, 347, 49 },
		{ 475, 348, 49 },
		{ 475, 349, 49 },
		{ 475, 350, 49 },
		{ 475, 351, 49 },
		{ 475, 352, 49 },
		{ 475, 353, 49 },
		{ 475, 354, 49 },
		{ 475, 355, 49 },
		{ 475, 356, 49 },
		{ 475, 357, 49 },
		{ 475, 358, 49 },
		{ 475, 359, 49 },
		{ 475, 360, 49 },
		{ 475, 361, 49 },
		{ 475, 362, 49 },
		{ 475, 363, 49 },
		{ 475, 364, 49 },
		{ 475, 365, 49 },
		{ 476, 312, 49 },
		{ 476, 313, 49 },
		{ 476, 314, 49 },
		{ 476, 315, 49 },
		{ 476, 316, 49 },
		{ 476, 317, 49 },
		{ 476, 318, 49 },
		{ 476, 319, 49 },
		{ 476, 320, 49 },
		{ 476, 321, 49 },
		{ 476, 322, 49 },
		{ 476, 323, 49 },
		{ 476, 324, 49 },
		{ 476, 325, 49 },
		{ 476, 326, 49 },
		{ 476, 327, 49 },
		{ 476, 328, 49 },
		{ 476, 329, 49 },
		{ 476, 330, 49 },
		{ 476, 331, 49 },
		{ 476, 332, 49 },
		{ 476, 333, 49 },
		{ 476, 334, 49 },
		{ 476, 335, 49 },
		{ 476, 336, 49 },
		{ 476, 337, 49 },
		{ 476, 338, 49 },
		{ 476, 339, 49 },
		{ 476, 340, 49 },
		{ 476, 341, 49 },
		{ 476, 342, 49 },
		{ 476, 343, 49 },
		{ 476, 344, 49 },
		{ 476, 345, 49 },
		{ 476, 346, 49 },
		{ 476, 347, 49 },
		{ 476, 348, 49 },
		{ 476, 349, 49 },
		{ 476, 350, 49 },
		{ 476, 351, 49 },
		{ 476, 352, 49 },
		{ 476, 353, 49 },
		{ 476, 354, 49 },
		{ 476, 355, 49 },
		{ 476, 356, 49 },
		{ 476, 357, 49 },
		{ 476, 358, 49 },
		{ 476, 359, 49 },
		{ 476, 360, 49 },
		{ 476, 361, 49 },
		{ 476, 362, 49 },
		{ 476, 363, 49 },
		{ 476, 364, 49 },
		{ 476, 365, 49 },
		{ 476, 366, 49 },
		{ 477, 313, 49 },
		{ 477, 314, 49 },
		{ 477, 315, 49 },
		{ 477, 316, 49 },
		{ 477, 317, 49 },
		{ 477, 318, 49 },
		{ 477, 319, 49 },
		{ 477, 320, 49 },
		{ 477, 321, 49 },
		{ 477, 322, 49 },
		{ 477, 323, 49 },
		{ 477, 324, 49 },
		{ 477, 325, 49 },
		{ 477, 326, 49 },
		{ 477, 327, 49 },
		{ 477, 328, 49 },
		{ 477, 329, 49 },
		{ 477, 330, 49 },
		{ 477, 331, 49 },
		{ 477, 332, 49 },
		{ 477, 333, 49 },
		{ 477, 334, 49 },
		{ 477, 335, 49 },
		{ 477, 336, 49 },
		{ 477, 337, 49 },
		{ 477, 338, 49 },
		{ 477, 339, 49 },
		{ 477, 340, 49 },
		{ 477, 341, 49 },
		{ 477, 342, 49 },
		{ 477, 343, 49 },
		{ 477, 344, 49 },
		{ 477, 345, 49 },
		{ 477, 346, 49 },
		{ 477, 347, 49 },
		{ 477, 348, 49 },
		{ 477, 349, 49 },
		{ 477, 350, 49 },
		{ 477, 351, 49 },
		{ 477, 352, 49 },
		{ 477, 353, 49 },
		{ 477, 354, 49 },
		{ 477, 355, 49 },
		{ 477, 356, 49 },
		{ 477, 357, 49 },
		{ 477, 358, 49 },
		{ 477, 359, 49 },
		{ 477, 360, 49 },
		{ 477, 361, 49 },
		{ 477, 362, 49 },
		{ 477, 363, 49 },
		{ 477, 364, 49 },
		{ 477, 365, 49 },
		{ 477, 366, 49 },
		{ 478, 313, 49 },
		{ 478, 314, 49 },
		{ 478, 315, 49 },
		{ 478, 316, 49 },
		{ 478, 317, 49 },
		{ 478, 318, 49 },
		{ 478, 319, 49 },
		{ 478, 320, 49 },
		{ 478, 321, 49 },
		{ 478, 322, 49 },
		{ 478, 323, 49 },
		{ 478, 324, 49 },
		{ 478, 325, 49 },
		{ 478, 326, 49 },
		{ 478, 327, 49 },
		{ 478, 328, 49 },
		{ 478, 329, 49 },
		{ 478, 330, 49 },
		{ 478, 331, 49 },
		{ 478, 332, 49 },
		{ 478, 333, 49 },
		{ 478, 334, 49 },
		{ 478, 335, 49 },
		{ 478, 336, 49 },
		{ 478, 337, 49 },
		{ 478, 338, 49 },
		{ 478, 339, 49 },
		{ 478, 340, 49 },
		{ 478, 341, 49 },
		{ 478, 342, 49 },
		{ 478, 343, 49 },
		{ 478, 344, 49 },
		{ 478, 345, 49 },
		{ 478, 346, 49 },
		{ 478, 347, 49 },
		{ 478, 348, 49 },
		{ 478, 349, 49 },
		{ 478, 350, 49 },
		{ 478, 351, 49 },
		{ 478, 352, 49 },
		{ 478, 353, 49 },
		{ 478, 354, 49 },
		{ 478, 355, 49 },
		{ 478, 356, 49 },
		{ 478, 357, 49 },
		{ 478, 358, 49 },
		{ 478, 359, 49 },
		{ 478, 360, 49 },
		{ 478, 361, 49 },
		{ 478, 362, 49 },
		{ 478, 363, 49 },
		{ 478, 364, 49 },
		{ 478, 365, 49 },
		{ 478, 366, 49 },
		{ 478, 367, 49 },
		{ 479, 313, 49 },
		{ 479, 314, 49 },
		{ 479, 315, 49 },
		{ 479, 316, 49 },
		{ 479, 317, 49 },
		{ 479, 318, 49 },
		{ 479, 319, 49 },
		{ 479, 320, 49 },
		{ 479, 321, 49 },
		{ 479, 322, 49 },
		{ 479, 323, 49 },
		{ 479, 324, 49 },
		{ 479, 325, 49 },
		{ 479, 326, 49 },
		{ 479, 327, 49 },
		{ 479, 328, 49 },
		{ 479, 329, 49 },
		{ 479, 330, 49 },
		{ 479, 331, 49 },
		{ 479, 332, 49 },
		{ 479, 333, 49 },
		{ 479, 334, 49 },
		{ 479, 335, 49 },
		{ 479, 336, 49 },
		{ 479, 337, 49 },
		{ 479, 338, 49 },
		{ 479, 339, 49 },
		{ 479, 340, 49 },
		{ 479, 341, 49 },
		{ 479, 342, 49 },
		{ 479, 343, 49 },
		{ 479, 344, 49 },
		{ 479, 345, 49 },
		{ 479, 346, 49 },
		{ 479, 347, 49 },
		{ 479, 348, 49 },
		{ 479, 349, 49 },
		{ 479, 350, 49 },
		{ 479, 351, 49 },
		{ 479, 352, 49 },
		{ 479, 353, 49 },
		{ 479, 354, 49 },
		{ 479, 355, 49 },
		{ 479, 356, 49 },
		{ 479, 357, 49 },
		{ 479, 358, 49 },
		{ 479, 359, 49 },
		{ 479, 360, 49 },
		{ 479, 361, 49 },
		{ 479, 362, 49 },
		{ 479, 363, 49 },
		{ 479, 364, 49 },
		{ 479, 365, 49 },
		{ 479, 366, 49 },
		{ 479, 367, 49 },
		{ 480, 313, 49 },
		{ 480, 314, 49 },
		{ 480, 315, 49 },
		{ 480, 316, 49 },
		{ 480, 317, 49 },
		{ 480, 318, 49 },
		{ 480, 319, 49 },
		{ 480, 320, 49 },
		{ 480, 321, 49 },
		{ 480, 322, 49 },
		{ 480, 323, 49 },
		{ 480, 324, 49 },
		{ 480, 325, 49 },
		{ 480, 326, 49 },
		{ 480, 327, 49 },
		{ 480, 328, 49 },
		{ 480, 329, 49 },
		{ 480, 330, 49 },
		{ 480, 331, 49 },
		{ 480, 332, 49 },
		{ 480, 333, 49 },
		{ 480, 334, 49 },
		{ 480, 335, 49 },
		{ 480, 336, 49 },
		{ 480, 337, 49 },
		{ 480, 338, 49 },
		{ 480, 339, 49 },
		{ 480, 340, 49 },
		{ 480, 341, 49 },
		{ 480, 342, 49 },
		{ 480, 343, 49 },
		{ 480, 344, 49 },
		{ 480, 345, 49 },
		{ 480, 346, 49 },
		{ 480, 347, 49 },
		{ 480, 348, 49 },
		{ 480, 349, 49 },
		{ 480, 350, 49 },
		{ 480, 351, 49 },
		{ 480, 352, 49 },
		{ 480, 353, 49 },
		{ 480, 354, 49 },
		{ 480, 355, 49 },
		{ 480, 356, 49 },
		{ 480, 357, 49 },
		{ 480, 358, 49 },
		{ 480, 359, 49 },
		{ 480, 360, 49 },
		{ 480, 361, 49 },
		{ 480, 362, 49 },
		{ 480, 363, 49 },
		{ 480, 364, 49 },
		{ 480, 365, 49 },
		{ 480, 366, 49 },
		{ 480, 367, 49 },
		{ 480, 368, 49 },
		{ 481, 314, 49 },
		{ 481, 315, 49 },
		{ 481, 316, 49 },
		{ 481, 317, 49 },
		{ 481, 318, 49 },
		{ 481, 319, 49 },
		{ 481, 320, 49 },
		{ 481, 321, 49 },
		{ 481, 322, 49 },
		{ 481, 323, 49 },
		{ 481, 324, 49 },
		{ 481, 325, 49 },
		{ 481, 326, 49 },
		{ 481, 327, 49 },
		{ 481, 328, 49 },
		{ 481, 329, 49 },
		{ 481, 330, 49 },
		{ 481, 331, 49 },
		{ 481, 332, 49 },
		{ 481, 333, 49 },
		{ 481, 334, 49 },
		{ 481, 335, 49 },
		{ 481, 336, 49 },
		{ 481, 337, 49 },
		{ 481, 338, 49 },
		{ 481, 339, 49 },
		{ 481, 340, 49 },
		{ 481, 341, 49 },
		{ 481, 342, 49 },
		{ 481, 343, 49 },
		{ 481, 344, 49 },
		{ 481, 345, 49 },
		{ 481, 346, 49 },
		{ 481, 347, 49 },
		{ 481, 348, 49 },
		{ 481, 349, 49 },
		{ 481, 350, 49 },
		{ 481, 351, 49 },
		{ 481, 352, 49 },
		{ 481, 353, 49 },
		{ 481, 354, 49 },
		{ 481, 355, 49 },
		{ 481, 356, 49 },
		{ 481, 357, 49 },
		{ 481, 358, 49 },
		{ 481, 359, 49 },
		{ 481, 360, 49 },
		{ 481, 361, 49 },
		{ 481, 362, 49 },
		{ 481, 363, 49 },
		{ 481, 364, 49 },
		{ 481, 365, 49 },
		{ 481, 366, 49 },
		{ 481, 367, 49 },
		{ 481, 368, 49 },
		{ 482, 314, 49 },
		{ 482, 315, 49 },
		{ 482, 316, 49 },
		{ 482, 317, 49 },
		{ 482, 318, 49 },
		{ 482, 319, 49 },
		{ 482, 320, 49 },
		{ 482, 321, 49 },
		{ 482, 322, 49 },
		{ 482, 323, 49 },
		{ 482, 324, 49 },
		{ 482, 325, 49 },
		{ 482, 326, 49 },
		{ 482, 327, 49 },
		{ 482, 328, 49 },
		{ 482, 329, 49 },
		{ 482, 330, 49 },
		{ 482, 331, 49 },
		{ 482, 332, 49 },
		{ 482, 333, 49 },
		{ 482, 334, 49 },
		{ 482, 335, 49 },
		{ 482, 336, 49 },
		{ 482, 337, 49 },
		{ 482, 338, 49 },
		{ 482, 339, 49 },
		{ 482, 340, 49 },
		{ 482, 341, 49 },
		{ 482, 342, 49 },
		{ 482, 343, 49 },
		{ 482, 344, 49 },
		{ 482, 345, 49 },
		{ 482, 346, 49 },
		{ 482, 347, 49 },
		{ 482, 348, 49 },
		{ 482, 349, 49 },
		{ 482, 350, 49 },
		{ 482, 351, 49 },
		{ 482, 352, 49 },
		{ 482, 353, 49 },
		{ 482, 354, 49 },
		{ 482, 355, 49 },
		{ 482, 356, 49 },
		{ 482, 357, 49 },
		{ 482, 358, 49 },
		{ 482, 359, 49 },
		{ 482, 360, 49 },
		{ 482, 361, 49 },
		{ 482, 362, 49 },
		{ 482, 363, 49 },
		{ 482, 364, 49 },
		{ 482, 365, 49 },
		{ 482, 366, 49 },
		{ 482, 367, 49 },
		{ 482, 368, 49 },
		{ 482, 369, 49 },
		{ 483, 314, 49 },
		{ 483, 315, 49 },
		{ 483, 316, 49 },
		{ 483, 317, 49 },
		{ 483, 318, 49 },
		{ 483, 319, 49 },
		{ 483, 320, 49 },
		{ 483, 321, 49 },
		{ 483, 322, 49 },
		{ 483, 323, 49 },
		{ 483, 324, 49 },
		{ 483, 325, 49 },
		{ 483, 326, 49 },
		{ 483, 327, 49 },
		{ 483, 328, 49 },
		{ 483, 329, 49 },
		{ 483, 330, 49 },
		{ 483, 331, 49 },
		{ 483, 332, 49 },
		{ 483, 333, 49 },
		{ 483, 334, 49 },
		{ 483, 335, 49 },
		{ 483, 336, 49 },
		{ 483, 337, 49 },
		{ 483, 338, 49 },
		{ 483, 339, 49 },
		{ 483, 340, 49 },
		{ 483, 341, 49 },
		{ 483, 342, 49 },
		{ 483, 343, 49 },
		{ 483, 344, 49 },
		{ 483, 345, 49 },
		{ 483, 346, 49 },
		{ 483, 347, 49 },
		{ 483, 348, 49 },
		{ 483, 349, 49 },
		{ 483, 350, 49 },
		{ 483, 351, 49 },
		{ 483, 352, 49 },
		{ 483, 353, 49 },
		{ 483, 354, 49 },
		{ 483, 355, 49 },
		{ 483, 356, 49 },
		{ 483, 357, 49 },
		{ 483, 358, 49 },
		{ 483, 359, 49 },
		{ 483, 360, 49 },
		{ 483, 361, 49 },
		{ 483, 362, 49 },
		{ 483, 363, 49 },
		{ 483, 364, 49 },
		{ 483, 365, 49 },
		{ 483, 366, 49 },
		{ 483, 367, 49 },
		{ 483, 368, 49 },
		{ 483, 369, 49 },
		{ 484, 314, 49 },
		{ 484, 315, 49 },
		{ 484, 316, 49 },
		{ 484, 317, 49 },
		{ 484, 318, 49 },
		{ 484, 319, 49 },
		{ 484, 320, 49 },
		{ 484, 321, 49 },
		{ 484, 322, 49 },
		{ 484, 323, 49 },
		{ 484, 324, 49 },
		{ 484, 325, 49 },
		{ 484, 326, 49 },
		{ 484, 327, 49 },
		{ 484, 328, 49 },
		{ 484, 329, 49 },
		{ 484, 330, 49 },
		{ 484, 331, 49 },
		{ 484, 332, 49 },
		{ 484, 333, 49 },
		{ 484, 334, 49 },
		{ 484, 335, 49 },
		{ 484, 336, 49 },
		{ 484, 337, 49 },
		{ 484, 338, 49 },
		{ 484, 339, 49 },
		{ 484, 340, 49 },
		{ 484, 341, 49 },
		{ 484, 342, 49 },
		{ 484, 343, 49 },
		{ 484, 344, 49 },
		{ 484, 345, 49 },
		{ 484, 346, 49 },
		{ 484, 347, 49 },
		{ 484, 348, 49 },
		{ 484, 349, 49 },
		{ 484, 350, 49 },
		{ 484, 351, 49 },
		{ 484, 352, 49 },
		{ 484, 353, 49 },
		{ 484, 354, 49 },
		{ 484, 355, 49 },
		{ 484, 356, 49 },
		{ 484, 357, 49 },
		{ 484, 358, 49 },
		{ 484, 359, 49 },
		{ 484, 360, 49 },
		{ 484, 361, 49 },
		{ 484, 362, 49 },
		{ 484, 363, 49 },
		{ 484, 364, 49 },
		{ 484, 365, 49 },
		{ 484, 366, 49 },
		{ 484, 367, 49 },
		{ 484, 368, 49 },
		{ 484, 369, 49 },
		{ 484, 370, 49 },
		{ 485, 315, 49 },
		{ 485, 316, 49 },
		{ 485, 317, 49 },
		{ 485, 318, 49 },
		{ 485, 319, 49 },
		{ 485, 320, 49 },
		{ 485, 321, 49 },
		{ 485, 322, 49 },
		{ 485, 323, 49 },
		{ 485, 324, 49 },
		{ 485, 325, 49 },
		{ 485, 326, 49 },
		{ 485, 327, 49 },
		{ 485, 328, 49 },
		{ 485, 329, 49 },
		{ 485, 330, 49 },
		{ 485, 331, 49 },
		{ 485, 332, 49 },
		{ 485, 333, 49 },
		{ 485, 334, 49 },
		{ 485, 335, 49 },
		{ 485, 336, 49 },
		{ 485, 337, 49 },
		{ 485, 338, 49 },
		{ 485, 339, 49 },
		{ 485, 340, 49 },
		{ 485, 341, 49 },
		{ 485, 342, 49 },
		{ 485, 343, 49 },
		{ 485, 344, 49 },
		{ 485, 345, 49 },
		{ 485, 346, 49 },
		{ 485, 347, 49 },
		{ 485, 348, 49 },
		{ 485, 349, 49 },
		{ 485, 350, 49 },
		{ 485, 351, 49 },
		{ 485, 352, 49 },
		{ 485, 353, 49 },
		{ 485, 354, 49 },
		{ 485, 355, 49 },
		{ 485, 356, 49 },
		{ 485, 357, 49 },
		{ 485, 358, 49 },
		{ 485, 359, 49 },
		{ 485, 360, 49 },
		{ 485, 361, 49 },
		{ 485, 362, 49 },
		{ 485, 363, 49 },
		{ 485, 364, 49 },
		{ 485, 365, 49 },
		{ 485, 366, 49 },
		{ 485, 367, 49 },
		{ 485, 368, 49 },
		{ 485, 369, 49 },
		{ 485, 370, 49 },
		{ 486, 315, 49 },
		{ 486, 316, 49 },
		{ 486, 317, 49 },
		{ 486, 318, 49 },
		{ 486, 319, 49 },
		{ 486, 320, 49 },
		{ 486, 321, 49 },
		{ 486, 322, 49 },
		{ 486, 323, 49 },
		{ 486, 324, 49 },
		{ 486, 325, 49 },
		{ 486, 326, 49 },
		{ 486, 327, 49 },
		{ 486, 328, 49 },
		{ 486, 329, 49 },
		{ 486, 330, 49 },
		{ 486, 331, 49 },
		{ 486, 332, 49 },
		{ 486, 333, 49 },
		{ 486, 334, 49 },
		{ 486, 335, 49 },
		{ 486, 336, 49 },
		{ 486, 337, 49 },
		{ 486, 338, 49 },
		{ 486, 339, 49 },
		{ 486, 340, 49 },
		{ 486, 341, 49 },
		{ 486, 342, 49 },
		{ 486, 343, 49 },
		{ 486, 344, 49 },
		{ 486, 345, 49 },
		{ 486, 346, 49 },
		{ 486, 347, 49 },
		{ 486, 348, 49 },
		{ 486, 349, 49 },
		{ 486, 350, 49 },
		{ 486, 351, 49 },
		{ 486, 352, 49 },
		{ 486, 353, 49 },
		{ 486, 354, 49 },
		{ 486, 355, 49 },
		{ 486, 356, 49 },
		{ 486, 357, 49 },
		{ 486, 358, 49 },
		{ 486, 359, 49 },
		{ 486, 360, 49 },
		{ 486, 361, 49 },
		{ 486, 362, 49 },
		{ 486, 363, 49 },
		{ 486, 364, 49 },
		{ 486, 365, 49 },
		{ 486, 366, 49 },
		{ 486, 367, 49 },
		{ 486, 368, 49 },
		{ 486, 369, 49 },
		{ 486, 370, 49 },
		{ 486, 371, 49 },
		{ 487, 315, 49 },
		{ 487, 316, 49 },
		{ 487, 317, 49 },
		{ 487, 318, 49 },
		{ 487, 319, 49 },
		{ 487, 320, 49 },
		{ 487, 321, 49 },
		{ 487, 322, 49 },
		{ 487, 323, 49 },
		{ 487, 324, 49 },
		{ 487, 325, 49 },
		{ 487, 326, 49 },
		{ 487, 327, 49 },
		{ 487, 328, 49 },
		{ 487, 329, 49 },
		{ 487, 330, 49 },
		{ 487, 331, 49 },
		{ 487, 332, 49 },
		{ 487, 333, 49 },
		{ 487, 334, 49 },
		{ 487, 335, 49 },
		{ 487, 336, 49 },
		{ 487, 337, 49 },
		{ 487, 338, 49 },
		{ 487, 339, 49 },
		{ 487, 340, 49 },
		{ 487, 341, 49 },
		{ 487, 342, 49 },
		{ 487, 343, 49 },
		{ 487, 344, 49 },
		{ 487, 345, 49 },
		{ 487, 346, 49 },
		{ 487, 347, 49 },
		{ 487, 348, 49 },
		{ 487, 349, 49 },
		{ 487, 350, 49 },
		{ 487, 351, 49 },
		{ 487, 352, 49 },
		{ 487, 353, 49 },
		{ 487, 354, 49 },
		{ 487, 355, 49 },
		{ 487, 356, 49 },
		{ 487, 357, 49 },
		{ 487, 358, 49 },
		{ 487, 359, 49 },
		{ 487, 360, 49 },
		{ 487, 361, 49 },
		{ 487, 362, 49 },
		{ 487, 363, 49 },
		{ 487, 364, 49 },
		{ 487, 365, 49 },
		{ 487, 366, 49 },
		{ 487, 367, 49 },
		{ 487, 368, 49 },
		{ 487, 369, 49 },
		{ 487, 370, 49 },
		{ 487, 371, 49 },
		{ 488, 315, 49 },
		{ 488, 316, 49 },
		{ 488, 317, 49 },
		{ 488, 318, 49 },
		{ 488, 319, 49 },
		{ 488, 320, 49 },
		{ 488, 321, 49 },
		{ 488, 322, 49 },
		{ 488, 323, 49 },
		{ 488, 324, 49 },
		{ 488, 325, 49 },
		{ 488, 326, 49 },
		{ 488, 327, 49 },
		{ 488, 328, 49 },
		{ 488, 329, 49 },
		{ 488, 330, 49 },
		{ 488, 331, 49 },
		{ 488, 332, 49 },
		{ 488, 333, 49 },
		{ 488, 334, 49 },
		{ 488, 335, 49 },
		{ 488, 336, 49 },
		{ 488, 337, 49 },
		{ 488, 338, 49 },
		{ 488, 339, 49 },
		{ 488, 340, 49 },
		{ 488, 341, 49 },
		{ 488, 342, 49 },
		{ 488, 343, 49 },
		{ 488, 344, 49 },
		{ 488, 345, 49 },
		{ 488, 346, 49 },
		{ 488, 347, 49 },
		{ 488, 348, 49 },
		{ 488, 349, 49 },
		{ 488, 350, 49 },
		{ 488, 351, 49 },
		{ 488, 352, 49 },
		{ 488, 353, 49 },
		{ 488, 354, 49 },
		{ 488, 355, 49 },
		{ 488, 356, 49 },
		{ 488, 357, 49 },
		{ 488, 358, 49 },
		{ 488, 359, 49 },
		{ 488, 360, 49 },
		{ 488, 361, 49 },
		{ 488, 362, 49 },
		{ 488, 363, 49 },
		{ 488, 364, 49 },
		{ 488, 365, 49 },
		{ 488, 366, 49 },
		{ 488, 367, 49 },
		{ 488, 368, 49 },
		{ 488, 369, 49 },
		{ 488, 370, 49 },
		{ 488, 371, 49 },
		{ 488, 372, 49 },
		{ 489, 316, 49 },
		{ 489, 317, 49 },
		{ 489, 318, 49 },
		{ 489, 319, 49 },
		{ 489, 320, 49 },
		{ 489, 321, 49 },
		{ 489, 322, 49 },
		{ 489, 323, 49 },
		{ 489, 324, 49 },
		{ 489, 325, 49 },
		{ 489, 326, 49 },
		{ 489, 327, 49 },
		{ 489, 328, 49 },
		{ 489, 329, 49 },
		{ 489, 330, 49 },
		{ 489, 331, 49 },
		{ 489, 332, 49 },
		{ 489, 333, 49 },
		{ 489, 334, 49 },
		{ 489, 335, 49 },
		{ 489, 336, 49 },
		{ 489, 337, 49 },
		{ 489, 338, 49 },
		{ 489, 339, 49 },
		{ 489, 340, 49 },
		{ 489, 341, 49 },
		{ 489, 342, 49 },
		{ 489, 343, 49 },
		{ 489, 344, 49 },
		{ 489, 345, 49 },
		{ 489, 346, 49 },
		{ 489, 347, 49 },
		{ 489, 348, 49 },
		{ 489, 349, 49 },
		{ 489, 350, 49 },
		{ 489, 351, 49 },
		{ 489, 352, 49 },
		{ 489, 353, 49 },
		{ 489, 354, 49 },
		{ 489, 355, 49 },
		{ 489, 356, 49 },
		{ 489, 357, 49 },
		{ 489, 358, 49 },
		{ 489, 359, 49 },
		{ 489, 360, 49 },
		{ 489, 361, 49 },
		{ 489, 362, 49 },
		{ 489, 363, 49 },
		{ 489, 364, 49 },
		{ 489, 365, 49 },
		{ 489, 366, 49 },
		{ 489, 367, 49 },
		{ 489, 368, 49 },
		{ 489, 369, 49 },
		{ 489, 370, 49 },
		{ 489, 371, 49 },
		{ 489, 372, 49 },
		{ 490, 316, 49 },
		{ 490, 317, 49 },
		{ 490, 318, 49 },
		{ 490, 319, 49 },
		{ 490, 320, 49 },
		{ 490, 321, 49 },
		{ 490, 322, 49 },
		{ 490, 323, 49 },
		{ 490, 324, 49 },
		{ 490, 325, 49 },
		{ 490, 326, 49 },
		{ 490, 327, 49 },
		{ 490, 328, 49 },
		{ 490, 329, 49 },
		{ 490, 330, 49 },
		{ 490, 331, 49 },
		{ 490, 332, 49 },
		{ 490, 333, 49 },
		{ 490, 334, 49 },
		{ 490, 335, 49 },
		{ 490, 336, 49 },
		{ 490, 337, 49 },
		{ 490, 338, 49 },
		{ 490, 339, 49 },
		{ 490, 340, 49 },
		{ 490, 341, 49 },
		{ 490, 342, 49 },
		{ 490, 343, 49 },
		{ 490, 344, 49 },
		{ 490, 345, 49 },
		{ 490, 346, 49 },
		{ 490, 347, 49 },
		{ 490, 348, 49 },
		{ 490, 349, 49 },
		{ 490, 350, 49 },
		{ 490, 351, 49 },
		{ 490, 352, 49 },
		{ 490, 353, 49 },
		{ 490, 354, 49 },
		{ 490, 355, 49 },
		{ 490, 356, 49 },
		{ 490, 357, 49 },
		{ 490, 358, 49 },
		{ 490, 359, 49 },
		{ 490, 360, 49 },
		{ 490, 361, 49 },
		{ 490, 362, 49 },
		{ 490, 363, 49 },
		{ 490, 364, 49 },
		{ 490, 365, 49 },
		{ 490, 366, 49 },
		{ 490, 367, 49 },
		{ 490, 368, 49 },
		{ 490, 369, 49 },
		{ 490, 370, 49 },
		{ 490, 371, 49 },
		{ 490, 372, 49 },
		{ 490, 373, 49 },
		{ 491, 316, 49 },
		{ 491, 317, 49 },
		{ 491, 318, 49 },
		{ 491, 319, 49 },
		{ 491, 320, 49 },
		{ 491, 321, 49 },
		{ 491, 322, 49 },
		{ 491, 323, 49 },
		{ 491, 324, 49 },
		{ 491, 325, 49 },
		{ 491, 326, 49 },
		{ 491, 327, 49 },
		{ 491, 328, 49 },
		{ 491, 329, 49 },
		{ 491, 330, 49 },
		{ 491, 331, 49 },
		{ 491, 332, 49 },
		{ 491, 333, 49 },
		{ 491, 334, 49 },
		{ 491, 335, 49 },
		{ 491, 336, 49 },
		{ 491, 337, 49 },
		{ 491, 338, 49 },
		{ 491, 339, 49 },
		{ 491, 340, 49 },
		{ 491, 341, 49 },
		{ 491, 342, 49 },
		{ 491, 343, 49 },
		{ 491, 344, 49 },
		{ 491, 345, 49 },
		{ 491, 346, 49 },
		{ 491, 347, 49 },
		{ 491, 348, 49 },
		{ 491, 349, 49 },
		{ 491, 350, 49 },
		{ 491, 351, 49 },
		{ 491, 352, 49 },
		{ 491, 353, 49 },
		{ 491, 354, 49 },
		{ 491, 355, 49 },
		{ 491, 356, 49 },
		{ 491, 357, 49 },
		{ 491, 358, 49 },
		{ 491, 359, 49 },
		{ 491, 360, 49 },
		{ 491, 361, 49 },
		{ 491, 362, 49 },
		{ 491, 363, 49 },
		{ 491, 364, 49 },
		{ 491, 365, 49 },
		{ 491, 366, 49 },
		{ 491, 367, 49 },
		{ 491, 368, 49 },
		{ 491, 369, 49 },
		{ 491, 370, 49 },
		{ 491, 371, 49 },
		{ 491, 372, 49 },
		{ 491, 373, 49 },
		{ 492, 316, 49 },
		{ 492, 317, 49 },
		{ 492, 318, 49 },
		{ 492, 319, 49 },
		{ 492, 320, 49 },
		{ 492, 321, 49 },
		{ 492, 322, 49 },
		{ 492, 323, 49 },
		{ 492, 324, 49 },
		{ 492, 325, 49 },
		{ 492, 326, 49 },
		{ 492, 327, 49 },
		{ 492, 328, 49 },
		{ 492, 329, 49 },
		{ 492, 330, 49 },
		{ 492, 331, 49 },
		{ 492, 332, 49 },
		{ 492, 333, 49 },
		{ 492, 334, 49 },
		{ 492, 335, 49 },
		{ 492, 336, 49 },
		{ 492, 337, 49 },
		{ 492, 338, 49 },
		{ 492, 339, 49 },
		{ 492, 340, 49 },
		{ 492, 341, 49 },
		{ 492, 342, 49 },
		{ 492, 343, 49 },
		{ 492, 344, 49 },
		{ 492, 345, 49 },
		{ 492, 346, 49 },
		{ 492, 347, 49 },
		{ 492, 348, 49 },
		{ 492, 349, 49 },
		{ 492, 350, 49 },
		{ 492, 351, 49 },
		{ 492, 352, 49 },
		{ 492, 353, 49 },
		{ 492, 354, 49 },
		{ 492, 355, 49 },
		{ 492, 356, 49 },
		{ 492, 357, 49 },
		{ 492, 358, 49 },
		{ 492, 359, 49 },
		{ 492, 360, 49 },
		{ 492, 361, 49 },
		{ 492, 362, 49 },
		{ 492, 363, 49 },
		{ 492, 364, 49 },
		{ 492, 365, 49 },
		{ 492, 366, 49 },
		{ 492, 367, 49 },
		{ 492, 368, 49 },
		{ 492, 369, 49 },
		{ 492, 370, 49 },
		{ 492, 371, 49 },
		{ 492, 372, 49 },
		{ 492, 373, 49 },
		{ 492, 374, 49 },
		{ 493, 317, 49 },
		{ 493, 318, 49 },
		{ 493, 319, 49 },
		{ 493, 320, 49 },
		{ 493, 321, 49 },
		{ 493, 322, 49 },
		{ 493, 323, 49 },
		{ 493, 324, 49 },
		{ 493, 325, 49 },
		{ 493, 326, 49 },
		{ 493, 327, 49 },
		{ 493, 328, 49 },
		{ 493, 329, 49 },
		{ 493, 330, 49 },
		{ 493, 331, 49 },
		{ 493, 332, 49 },
		{ 493, 333, 49 },
		{ 493, 334, 49 },
		{ 493, 335, 49 },
		{ 493, 336, 49 },
		{ 493, 337, 49 },
		{ 493, 338, 49 },
		{ 493, 339, 49 },
		{ 493, 340, 49 },
		{ 493, 341, 49 },
		{ 493, 342, 49 },
		{ 493, 343, 49 },
		{ 493, 344, 49 },
		{ 493, 345, 49 },
		{ 493, 346, 49 },
		{ 493, 347, 49 },
		{ 493, 348, 49 },
		{ 493, 349, 49 },
		{ 493, 350, 49 },
		{ 493, 351, 49 },
		{ 493, 352, 49 },
		{ 493, 353, 49 },
		{ 493, 354, 49 },
		{ 493, 355, 49 },
		{ 493, 356, 49 },
		{ 493, 357, 49 },
		{ 493, 358, 49 },
		{ 493, 359, 49 },
		{ 493, 360, 49 },
		{ 493, 361, 49 },
		{ 493, 362, 49 },
		{ 493, 363, 49 },
		{ 493, 364, 49 },
		{ 493, 365, 49 },
		{ 493, 366, 49 },
		{ 493, 367, 49 },
		{ 493, 368, 49 },
		{ 493, 369, 49 },
		{ 493, 370, 49 },
		{ 493, 371, 49 },
		{ 493, 372, 49 },
		{ 493, 373, 49 },
		{ 493, 374, 49 },
		{ 494, 317, 49 },
		{ 494, 318, 49 },
		{ 494, 319, 49 },
		{ 494, 320, 49 },
		{ 494, 321, 49 },
		{ 494, 322, 49 },
		{ 494, 323, 49 },
		{ 494, 324, 49 },
		{ 494, 325, 49 },
		{ 494, 326, 49 },
		{ 494, 327, 49 },
		{ 494, 328, 49 },
		{ 494, 329, 49 },
		{ 494, 330, 49 },
		{ 494, 331, 49 },
		{ 494, 332, 49 },
		{ 494, 333, 49 },
		{ 494, 334, 49 },
		{ 494, 335, 49 },
		{ 494, 336, 49 },
		{ 494, 337, 49 },
		{ 494, 338, 49 },
		{ 494, 339, 49 },
		{ 494, 340, 49 },
		{ 494, 341, 49 },
		{ 494, 342, 49 },
		{ 494, 343, 49 },
		{ 494, 344, 49 },
		{ 494, 345, 49 },
		{ 494, 346, 49 },
		{ 494, 347, 49 },
		{ 494, 348, 49 },
		{ 494, 349, 49 },
		{ 494, 350, 49 },
		{ 494, 351, 49 },
		{ 494, 352, 49 },
		{ 494, 353, 49 },
		{ 494, 354, 49 },
		{ 494, 355, 49 },
		{ 494, 356, 49 },
		{ 494, 357, 49 },
		{ 494, 358, 49 },
		{ 494, 359, 49 },
		{ 494, 360, 49 },
		{ 494, 361, 49 },
		{ 494, 362, 49 },
		{ 494, 363, 49 },
		{ 494, 364, 49 },
		{ 494, 365, 49 },
		{ 494, 366, 49 },
		{ 494, 367, 49 },
		{ 494, 368, 49 },
		{ 494, 369, 49 },
		{ 494, 370, 49 },
		{ 494, 371, 49 },
		{ 494, 372, 49 },
		{ 494, 373, 49 },
		{ 494, 374, 49 },
		{ 494, 375, 49 },
		{ 495, 317, 49 },
		{ 495, 318, 49 },
		{ 495, 319, 49 },
		{ 495, 320, 49 },
		{ 495, 321, 49 },
		{ 495, 322, 49 },
		{ 495, 323, 49 },
		{ 495, 324, 49 },
		{ 495, 325, 49 },
		{ 495, 326, 49 },
		{ 495, 327, 49 },
		{ 495, 328, 49 },
		{ 495, 329, 49 },
		{ 495, 330, 49 },
		{ 495, 331, 49 },
		{ 495, 332, 49 },
		{ 495, 333, 49 },
		{ 495, 334, 49 },
		{ 495, 335, 49 },
		{ 495, 336, 49 },
		{ 495, 337, 49 },
		{ 495, 338, 49 },
		{ 495, 339, 49 },
		{ 495, 340, 49 },
		{ 495, 341, 49 },
		{ 495, 342, 49 },
		{ 495, 343, 49 },
		{ 495, 344, 49 },
		{ 495, 345, 49 },
		{ 495, 346, 49 },
		{ 495, 347, 49 },
		{ 495, 348, 49 },
		{ 495, 349, 49 },
		{ 495, 350, 49 },
		{ 495, 351, 49 },
		{ 495, 352, 49 },
		{ 495, 353, 49 },
		{ 495, 354, 49 },
		{ 495, 355, 49 },
		{ 495, 356, 49 },
		{ 495, 357, 49 },
		{ 495, 358, 49 },
		{ 495, 359, 49 },
		{ 495, 360, 49 },
		{ 495, 361, 49 },
		{ 495, 362, 49 },
		{ 495, 363, 49 },
		{ 495, 364, 49 },
		{ 495, 365, 49 },
		{ 495, 366, 49 },
		{ 495, 367, 49 },
		{ 495, 368, 49 },
		{ 495, 369, 49 },
		{ 495, 370, 49 },
		{ 495, 371, 49 },
		{ 495, 372, 49 },
		{ 495, 373, 49 },
		{ 495, 374, 49 },
		{ 495, 375, 49 },
		{ 496, 317, 49 },
		{ 496, 318, 49 },
		{ 496, 319, 49 },
		{ 496, 320, 49 },
		{ 496, 321, 49 },
		{ 496, 322, 49 },
		{ 496, 323, 49 },
		{ 496, 324, 49 },
		{ 496, 325, 49 },
		{ 496, 326, 49 },
		{ 496, 327, 49 },
		{ 496, 328, 49 },
		{ 496, 329, 49 },
		{ 496, 330, 49 },
		{ 496, 331, 49 },
		{ 496, 332, 49 },
		{ 496, 333, 49 },
		{ 496, 334, 49 },
		{ 496, 335, 49 },
		{ 496, 336, 49 },
		{ 496, 337, 49 },
		{ 496, 338, 49 },
		{ 496, 339, 49 },
		{ 496, 340, 49 },
		{ 496, 341, 49 },
		{ 496, 342, 49 },
		{ 496, 343, 49 },
		{ 496, 344, 49 },
		{ 496, 345, 49 },
		{ 496, 346, 49 },
		{ 496, 347, 49 },
		{ 496, 348, 49 },
		{ 496, 349, 49 },
		{ 496, 350, 49 },
		{ 496, 351, 49 },
		{ 496, 352, 49 },
		{ 496, 353, 49 },
		{ 496, 354, 49 },
		{ 496, 355, 49 },
		{ 496, 356, 49 },
		{ 496, 357, 49 },
		{ 496, 358, 49 },
		{ 496, 359, 49 },
		{ 496, 360, 49 },
		{ 496, 361, 49 },
		{ 496, 362, 49 },
		{ 496, 363, 49 },
		{ 496, 364, 49 },
		{ 496, 365, 49 },
		{ 496, 366, 49 },
		{ 496, 367, 49 },
		{ 496, 368, 49 },
		{ 496, 369, 49 },
		{ 496, 370, 49 },
		{ 496, 371, 49 },
		{ 496, 372, 49 },
		{ 496, 373, 49 },
		{ 496, 374, 49 },
		{ 496, 375, 49 },
		{ 496, 376, 49 },
		{ 497, 318, 49 },
		{ 497, 319, 49 },
		{ 497, 320, 49 },
		{ 497, 321, 49 },
		{ 497, 322, 49 },
		{ 497, 323, 49 },
		{ 497, 324, 49 },
		{ 497, 325, 49 },
		{ 497, 326, 49 },
		{ 497, 327, 49 },
		{ 497, 328, 49 },
		{ 497, 329, 49 },
		{ 497, 330, 49 },
		{ 497, 331, 49 },
		{ 497, 332, 49 },
		{ 497, 333, 49 },
		{ 497, 334, 49 },
		{ 497, 335, 49 },
		{ 497, 336, 49 },
		{ 497, 337, 49 },
		{ 497, 338, 49 },
		{ 497, 339, 49 },
		{ 497, 340, 49 },
		{ 497, 341, 49 },
		{ 497, 342, 49 },
		{ 497, 343, 49 },
		{ 497, 344, 49 },
		{ 497, 345, 49 },
		{ 497, 346, 49 },
		{ 497, 347, 49 },
		{ 497, 348, 49 },
		{ 497, 349, 49 },
		{ 497, 350, 49 },
		{ 497, 351, 49 },
		{ 497, 352, 49 },
		{ 497, 353, 49 },
		{ 497, 354, 49 },
		{ 497, 355, 49 },
		{ 497, 356, 49 },
		{ 497, 357, 49 },
		{ 497, 358, 49 },
		{ 497, 359, 49 },
		{ 497, 360, 49 },
		{ 497, 361, 49 },
		{ 497, 362, 49 },
		{ 497, 363, 49 },
		{ 497, 364, 49 },
		{ 497, 365, 49 },
		{ 497, 366, 49 },
		{ 497, 367, 49 },
		{ 497, 368, 49 },
		{ 497, 369, 49 },
		{ 497, 370, 49 },
		{ 497, 371, 49 },
		{ 497, 372, 49 },
		{ 497, 373, 49 },
		{ 497, 374, 49 },
		{ 497, 375, 49 },
		{ 497, 376, 49 },
		{ 498, 318, 49 },
		{ 498, 319, 49 },
		{ 498, 320, 49 },
		{ 498, 321, 49 },
		{ 498, 322, 49 },
		{ 498, 323, 49 },
		{ 498, 324, 49 },
		{ 498, 325, 49 },
		{ 498, 326, 49 },
		{ 498, 327, 49 },
		{ 498, 328, 49 },
		{ 498, 329, 49 },
		{ 498, 330, 49 },
		{ 498, 331, 49 },
		{ 498, 332, 49 },
		{ 498, 333, 49 },
		{ 498, 334, 49 },
		{ 498, 335, 49 },
		{ 498, 336, 49 },
		{ 498, 337, 49 },
		{ 498, 338, 49 },
		{ 498, 339, 49 },
		{ 498, 340, 49 },
		{ 498, 341, 49 },
		{ 498, 342, 49 },
		{ 498, 343, 49 },
		{ 498, 344, 49 },
		{ 498, 345, 49 },
		{ 498, 346, 49 },
		{ 498, 347, 49 },
		{ 498, 348, 49 },
		{ 498, 349, 49 },
		{ 498, 350, 49 },
		{ 498, 351, 49 },
		{ 498, 352, 49 },
		{ 498, 353, 49 },
		{ 498, 354, 49 },
		{ 498, 355, 49 },
		{ 498, 356, 49 },
		{ 498, 357, 49 },
		{ 498, 358, 49 },
		{ 498, 359, 49 },
		{ 498, 360, 49 },
		{ 498, 361, 49 },
		{ 498, 362, 49 },
		{ 498, 363, 49 },
		{ 498, 364, 49 },
		{ 498, 365, 49 },
		{ 498, 366, 49 },
		{ 498, 367, 49 },
		{ 498, 368, 49 },
		{ 498, 369, 49 },
		{ 498, 370, 49 },
		{ 498, 371, 49 },
		{ 498, 372, 49 },
		{ 498, 373, 49 },
		{ 498, 374, 49 },
		{ 498, 375, 49 },
		{ 498, 376, 49 },
		{ 498, 377, 49 },
		{ 499, 318, 49 },
		{ 499, 319, 49 },
		{ 499, 320, 49 },
		{ 499, 321, 49 },
		{ 499, 322, 49 },
		{ 499, 323, 49 },
		{ 499, 324, 49 },
		{ 499, 325, 49 },
		{ 499, 326, 49 },
		{ 499, 327, 49 },
		{ 499, 328, 49 },
		{ 499, 329, 49 },
		{ 499, 330, 49 },
		{ 499, 331, 49 },
		{ 499, 332, 49 },
		{ 499, 333, 49 },
		{ 499, 334, 49 },
		{ 499, 335, 49 },
		{ 499, 336, 49 },
		{ 499, 337, 49 },
		{ 499, 338, 49 },
		{ 499, 339, 49 },
		{ 499, 340, 49 },
		{ 499, 341, 49 },
		{ 499, 342, 49 },
		{ 499, 343, 49 },
		{ 499, 344, 49 },
		{ 499, 345, 49 },
		{ 499, 346, 49 },
		{ 499, 347, 49 },
		{ 499, 348, 49 },
		{ 499, 349, 49 },
		{ 499, 350, 49 },
		{ 499, 351, 49 },
		{ 499, 352, 49 },
		{ 499, 353, 49 },
		{ 499, 354, 49 },
		{ 499, 355, 49 },
		{ 499, 356, 49 },
		{ 499, 357, 49 },
		{ 499, 358, 49 },
		{ 499, 359, 49 },
		{ 499, 360, 49 },
		{ 499, 361, 49 },
		{ 499, 362, 49 },
		{ 499, 363, 49 },
		{ 499, 364, 49 },
		{ 499, 365, 49 },
		{ 499, 366, 49 },
		{ 499, 367, 49 },
		{ 499, 368, 49 },
		{ 499, 369, 49 },
		{ 499, 370, 49 },
		{ 499, 371, 49 },
		{ 499, 372, 49 },
		{ 499, 373, 49 },
		{ 499, 374, 49 },
		{ 499, 375, 49 },
		{ 499, 376, 49 },
		{ 499, 377, 49 },
		{ 500, 318, 49 },
		{ 500, 319, 49 },
		{ 500, 320, 49 },
		{ 500, 321, 49 },
		{ 500, 322, 49 },
		{ 500, 323, 49 },
		{ 500, 324, 49 },
		{ 500, 325, 49 },
		{ 500, 326, 49 },
		{ 500, 327, 49 },
		{ 500, 328, 49 },
		{ 500, 329, 49 },
		{ 500, 330, 49 },
		{ 500, 331, 49 },
		{ 500, 332, 49 },
		{ 500, 333, 49 },
		{ 500, 334, 49 },
		{ 500, 335, 49 },
		{ 500, 336, 49 },
		{ 500, 337, 49 },
		{ 500, 338, 49 },
		{ 500, 339, 49 },
		{ 500, 340, 49 },
		{ 500, 341, 49 },
		{ 500, 342, 49 },
		{ 500, 343, 49 },
		{ 500, 344, 49 },
		{ 500, 345, 49 },
		{ 500, 346, 49 },
		{ 500, 347, 49 },
		{ 500, 348, 49 },
		{ 500, 349, 49 },
		{ 500, 350, 49 },
		{ 500, 351, 49 },
		{ 500, 352, 49 },
		{ 500, 353, 49 },
		{ 500, 354, 49 },
		{ 500, 355, 49 },
		{ 500, 356, 49 },
		{ 500, 357, 49 },
		{ 500, 358, 49 },
		{ 500, 359, 49 },
		{ 500, 360, 49 },
		{ 500, 361, 49 },
		{ 500, 362, 49 },
		{ 500, 363, 49 },
		{ 500, 364, 49 },
		{ 500, 365, 49 },
		{ 500, 366, 49 },
		{ 500, 367, 49 },
		{ 500, 368, 49 },
		{ 500, 369, 49 },
		{ 500, 370, 49 },
		{ 500, 371, 49 },
		{ 500, 372, 49 },
		{ 500, 373, 49 },
		{ 500, 374, 49 },
		{ 500, 375, 49 },
		{ 500, 376, 49 },
		{ 500, 377, 49 },
		{ 500, 378, 49 },
		{ 501, 319, 49 },
		{ 501, 320, 49 },
		{ 501, 321, 49 },
		{ 501, 322, 49 },
		{ 501, 323, 49 },
		{ 501, 324, 49 },
		{ 501, 325, 49 },
		{ 501, 326, 49 },
		{ 501, 327, 49 },
		{ 501, 328, 49 },
		{ 501, 329, 49 },
		{ 501, 330, 49 },
		{ 501, 331, 49 },
		{ 501, 332, 49 },
		{ 501, 333, 49 },
		{ 501, 334, 49 },
		{ 501, 335, 49 },
		{ 501, 336, 49 },
		{ 501, 337, 49 },
		{ 501, 338, 49 },
		{ 501, 339, 49 },
		{ 501, 340, 49 },
		{ 501, 341, 49 },
		{ 501, 342, 49 },
		{ 501, 343, 49 },
		{ 501, 344, 49 },
		{ 501, 345, 49 },
		{ 501, 346, 49 },
		{ 501, 347, 49 },
		{ 501, 348, 49 },
		{ 501, 349, 49 },
		{ 501, 350, 49 },
		{ 501, 351, 49 },
		{ 501, 352, 49 },
		{ 501, 353, 49 },
		{ 501, 354, 49 },
		{ 501, 355, 49 },
		{ 501, 356, 49 },
		{ 501, 357, 49 },
		{ 501, 358, 49 },
		{ 501, 359, 49 },
		{ 501, 360, 49 },
		{ 501, 361, 49 },
		{ 501, 362, 49 },
		{ 501, 363, 49 },
		{ 501, 364, 49 },
		{ 501, 365, 49 },
		{ 501, 366, 49 },
		{ 501, 367, 49 },
		{ 501, 368, 49 },
		{ 501, 369, 49 },
		{ 501, 370, 49 },
		{ 501, 371, 49 },
		{ 501, 372, 49 },
		{ 501, 373, 49 },
		{ 501, 374, 49 },
		{ 501, 375, 49 },
		{ 501, 376, 49 },
		{ 501, 377, 49 },
		{ 501, 378, 49 },
		{ 502, 319, 49 },
		{ 502, 320, 49 },
		{ 502, 321, 49 },
		{ 502, 322, 49 },
		{ 502, 323, 49 },
		{ 502, 324, 49 },
		{ 502, 325, 49 },
		{ 502, 326, 49 },
		{ 502, 327, 49 },
		{ 502, 328, 49 },
		{ 502, 329, 49 },
		{ 502, 330, 49 },
		{ 502, 331, 49 },
		{ 502, 332, 49 },
		{ 502, 333, 49 },
		{ 502, 334, 49 },
		{ 502, 335, 49 },
		{ 502, 336, 49 },
		{ 502, 337, 49 },
		{ 502, 338, 49 },
		{ 502, 339, 49 },
		{ 502, 340, 49 },
		{ 502, 341, 49 },
		{ 502, 342, 49 },
		{ 502, 343, 49 },
		{ 502, 344, 49 },
		{ 502, 345, 49 },
		{ 502, 346, 49 },
		{ 502, 347, 49 },
		{ 502, 348, 49 },
		{ 502, 349, 49 },
		{ 502, 350, 49 },
		{ 502, 351, 49 },
		{ 502, 352, 49 },
		{ 502, 353, 49 },
		{ 502, 354, 49 },
		{ 502, 355, 49 },
		{ 502, 356, 49 },
		{ 502, 357, 49 },
		{ 502, 358, 49 },
		{ 502, 359, 49 },
		{ 502, 360, 49 },
		{ 502, 361, 49 },
		{ 502, 362, 49 },
		{ 502, 363, 49 },
		{ 502, 364, 49 },
		{ 502, 365, 49 },
		{ 502, 366, 49 },
		{ 502, 367, 49 },
		{ 502, 368, 49 },
		{ 502, 369, 49 },
		{ 502, 370, 49 },
		{ 502, 371, 49 },
		{ 502, 372, 49 },
		{ 502, 373, 49 },
		{ 502, 374, 49 },
		{ 502, 375, 49 },
		{ 502, 376, 49 },
		{ 502, 377, 49 },
		{ 502, 378, 49 },
		{ 502, 379, 49 },
		{ 503, 319, 49 },
		{ 503, 320, 49 },
		{ 503, 321, 49 },
		{ 503, 322, 49 },
		{ 503, 323, 49 },
		{ 503, 324, 49 },
		{ 503, 325, 49 },
		{ 503, 326, 49 },
		{ 503, 327, 49 },
		{ 503, 328, 49 },
		{ 503, 329, 49 },
		{ 503, 330, 49 },
		{ 503, 331, 49 },
		{ 503, 332, 49 },
		{ 503, 333, 49 },
		{ 503, 334, 49 },
		{ 503, 335, 49 },
		{ 503, 336, 49 },
		{ 503, 337, 49 },
		{ 503, 338, 49 },
		{ 503, 339, 49 },
		{ 503, 340, 49 },
		{ 503, 341, 49 },
		{ 503, 342, 49 },
		{ 503, 343, 49 },
		{ 503, 344, 49 },
		{ 503, 345, 49 },
		{ 503, 346, 49 },
		{ 503, 347, 49 },
		{ 503, 348, 49 },
		{ 503, 349, 49 },
		{ 503, 350, 49 },
		{ 503, 351, 49 },
		{ 503, 352, 49 },
		{ 503, 353, 49 },
		{ 503, 354, 49 },
		{ 503, 355, 49 },
		{ 503, 356, 49 },
		{ 503, 357, 49 },
		{ 503, 358, 49 },
		{ 503, 359, 49 },
		{ 503, 360, 49 },
		{ 503, 361, 49 },
		{ 503, 362, 49 },
		{ 503, 363, 49 },
		{ 503, 364, 49 },
		{ 503, 365, 49 },
		{ 503, 366, 49 },
		{ 503, 367, 49 },
		{ 503, 368, 49 },
		{ 503, 369, 49 },
		{ 503, 370, 49 },
		{ 503, 371, 49 },
		{ 503, 372, 49 },
		{ 503, 373, 49 },
		{ 503, 374, 49 },
		{ 503, 375, 49 },
		{ 503, 376, 49 },
		{ 503, 377, 49 },
		{ 503, 378, 49 },
		{ 503, 379, 49 },
		{ 504, 319, 49 },
		{ 504, 320, 49 },
		{ 504, 321, 49 },
		{ 504, 322, 49 },
		{ 504, 323, 49 },
		{ 504, 324, 49 },
		{ 504, 325, 49 },
		{ 504, 326, 49 },
		{ 504, 327, 49 },
		{ 504, 328, 49 },
		{ 504, 329, 49 },
		{ 504, 330, 49 },
		{ 504, 331, 49 },
		{ 504, 332, 49 },
		{ 504, 333, 49 },
		{ 504, 334, 49 },
		{ 504, 335, 49 },
		{ 504, 336, 49 },
		{ 504, 337, 49 },
		{ 504, 338, 49 },
		{ 504, 339, 49 },
		{ 504, 340, 49 },
		{ 504, 341, 49 },
		{ 504, 342, 49 },
		{ 504, 343, 49 },
		{ 504, 344, 49 },
		{ 504, 345, 49 },
		{ 504, 346, 49 },
		{ 504, 347, 49 },
		{ 504, 348, 49 },
		{ 504, 349, 49 },
		{ 504, 350, 49 },
		{ 504, 351, 49 },
		{ 504, 352, 49 },
		{ 504, 353, 49 },
		{ 504, 354, 49 },
		{ 504, 355, 49 },
		{ 504, 356, 49 },
		{ 504, 357, 49 },
		{ 504, 358, 49 },
		{ 504, 359, 49 },
		{ 504, 360, 49 },
		{ 504, 361, 49 },
		{ 504, 362, 49 },
		{ 504, 363, 49 },
		{ 504, 364, 49 },
		{ 504, 365, 49 },
		{ 504, 366, 49 },
		{ 504, 367, 49 },
		{ 504, 368, 49 },
		{ 504, 369, 49 },
		{ 504, 370, 49 },
		{ 504, 371, 49 },
		{ 504, 372, 49 },
		{ 504, 373, 49 },
		{ 504, 374, 49 },
		{ 504, 375, 49 },
		{ 504, 376, 49 },
		{ 504, 377, 49 },
		{ 504, 378, 49 },
		{ 504, 379, 49 },
		{ 504, 380, 49 },
		{ 505, 320, 49 },
		{ 505, 321, 49 },
		{ 505, 322, 49 },
		{ 505, 323, 49 },
		{ 505, 324, 49 },
		{ 505, 325, 49 },
		{ 505, 326, 49 },
		{ 505, 327, 49 },
		{ 505, 328, 49 },
		{ 505, 329, 49 },
		{ 505, 330, 49 },
		{ 505, 331, 49 },
		{ 505, 332, 49 },
		{ 505, 333, 49 },
		{ 505, 334, 49 },
		{ 505, 335, 49 },
		{ 505, 336, 49 },
		{ 505, 337, 49 },
		{ 505, 338, 49 },
		{ 505, 339, 49 },
		{ 505, 340, 49 },
		{ 505, 341, 49 },
		{ 505, 342, 49 },
		{ 505, 343, 49 },
		{ 505, 344, 49 },
		{ 505, 345, 49 },
		{ 505, 346, 49 },
		{ 505, 347, 49 },
		{ 505, 348, 49 },
		{ 505, 349, 49 },
		{ 505, 350, 49 },
		{ 505, 351, 49 },
		{ 505, 352, 49 },
		{ 505, 353, 49 },
		{ 505, 354, 49 },
		{ 505, 355, 49 },
		{ 505, 356, 49 },
		{ 505, 357, 49 },
		{ 505, 358, 49 },
		{ 505, 359, 49 },
		{ 505, 360, 49 },
		{ 505, 361, 49 },
		{ 505, 362, 49 },
		{ 505, 363, 49 },
		{ 505, 364, 49 },
		{ 505, 365, 49 },
		{ 505, 366, 49 },
		{ 505, 367, 49 },
		{ 505, 368, 49 },
		{ 505, 369, 49 },
		{ 505, 370, 49 },
		{ 505, 371, 49 },
		{ 505, 372, 49 },
		{ 505, 373, 49 },
		{ 505, 374, 49 },
		{ 505, 375, 49 },
		{ 505, 376, 49 },
		{ 505, 377, 49 },
		{ 505, 378, 49 },
		{ 505, 379, 49 },
		{ 505, 380, 49 },
		{ 506, 320, 49 },
		{ 506, 321, 49 },
		{ 506, 322, 49 },
		{ 506, 323, 49 },
		{ 506, 324, 49 },
		{ 506, 325, 49 },
		{ 506, 326, 49 },
		{ 506, 327, 49 },
		{ 506, 328, 49 },
		{ 506, 329, 49 },
		{ 506, 330, 49 },
		{ 506, 331, 49 },
		{ 506, 332, 49 },
		{ 506, 333, 49 },
		{ 506, 334, 49 },
		{ 506, 335, 49 },
		{ 506, 336, 49 },
		{ 506, 337, 49 },
		{ 506, 338, 49 },
		{ 506, 339, 49 },
		{ 506, 340, 49 },
		{ 506, 341, 49 },
		{ 506, 342, 49 },
		{ 506, 343, 49 },
		{ 506, 344, 49 },
		{ 506, 345, 49 },
		{ 506, 346, 49 },
		{ 506, 347, 49 },
		{ 506, 348, 49 },
		{ 506, 349, 49 },
		{ 506, 350, 49 },
		{ 506, 351, 49 },
		{ 506, 352, 49 },
		{ 506, 353, 49 },
		{ 506, 354, 49 },
		{ 506, 355, 49 },
		{ 506, 356, 49 },
		{ 506, 357, 49 },
		{ 506, 358, 49 },
		{ 506, 359, 49 },
		{ 506, 360, 49 },
		{ 506, 361, 49 },
		{ 506, 362, 49 },
		{ 506, 363, 49 },
		{ 506, 364, 49 },
		{ 506, 365, 49 },
		{ 506, 366, 49 },
		{ 506, 367, 49 },
		{ 506, 368, 49 },
		{ 506, 369, 49 },
		{ 506, 370, 49 },
		{ 506, 371, 49 },
		{ 506, 372, 49 },
		{ 506, 373, 49 },
		{ 506, 374, 49 },
		{ 506, 375, 49 },
		{ 506, 376, 49 },
		{ 506, 377, 49 },
		{ 506, 378, 49 },
		{ 506, 379, 49 },
		{ 506, 380, 49 },
		{ 506, 381, 49 },
		{ 507, 320, 49 },
		{ 507, 321, 49 },
		{ 507, 322, 49 },
		{ 507, 323, 49 },
		{ 507, 324, 49 },
		{ 507, 325, 49 },
		{ 507, 326, 49 },
		{ 507, 327, 49 },
		{ 507, 328, 49 },
		{ 507, 329, 49 },
		{ 507, 330, 49 },
		{ 507, 331, 49 },
		{ 507, 332, 49 },
		{ 507, 333, 49 },
		{ 507, 334, 49 },
		{ 507, 335, 49 },
		{ 507, 336, 49 },
		{ 507, 337, 49 },
		{ 507, 338, 49 },
		{ 507, 339, 49 },
		{ 507, 340, 49 },
		{ 507, 341, 49 },
		{ 507, 342, 49 },
		{ 507, 343, 49 },
		{ 507, 344, 49 },
		{ 507, 345, 49 },
		{ 507, 346, 49 },
		{ 507, 347, 49 },
		{ 507, 348, 49 },
		{ 507, 349, 49 },
		{ 507, 350, 49 },
		{ 507, 351, 49 },
		{ 507, 352, 49 },
		{ 507, 353, 49 },
		{ 507, 354, 49 },
		{ 507, 355, 49 },
		{ 507, 356, 49 },
		{ 507, 357, 49 },
		{ 507, 358, 49 },
		{ 507, 359, 49 },
		{ 507, 360, 49 },
		{ 507, 361, 49 },
		{ 507, 362, 49 },
		{ 507, 363, 49 },
		{ 507, 364, 49 },
		{ 507, 365, 49 },
		{ 507, 366, 49 },
		{ 507, 367, 49 },
		{ 507, 368, 49 },
		{ 507, 369, 49 },
		{ 507, 370, 49 },
		{ 507, 371, 49 },
		{ 507, 372, 49 },
		{ 507, 373, 49 },
		{ 507, 374, 49 },
		{ 507, 375, 49 },
		{ 507, 376, 49 },
		{ 507, 377, 49 },
		{ 507, 378, 49 },
		{ 507, 379, 49 },
		{ 507, 380, 49 },
		{ 507, 381, 49 },
		{ 508, 320, 49 },
		{ 508, 321, 49 },
		{ 508, 322, 49 },
		{ 508, 323, 49 },
		{ 508, 324, 49 },
		{ 508, 325, 49 },
		{ 508, 326, 49 },
		{ 508, 327, 49 },
		{ 508, 328, 49 },
		{ 508, 329, 49 },
		{ 508, 330, 49 },
		{ 508, 331, 49 },
		{ 508, 332, 49 },
		{ 508, 333, 49 },
		{ 508, 334, 49 },
		{ 508, 335, 49 },
		{ 508, 336, 49 },
		{ 508, 337, 49 },
		{ 508, 338, 49 },
		{ 508, 339, 49 },
		{ 508, 340, 49 },
		{ 508, 341, 49 },
		{ 508, 342, 49 },
		{ 508, 343, 49 },
		{ 508, 344, 49 },
		{ 508, 345, 49 },
		{ 508, 346, 49 },
		{ 508, 347, 49 },
		{ 508, 348, 49 },
		{ 508, 349, 49 },
		{ 508, 350, 49 },
		{ 508, 351, 49 },
		{ 508, 352, 49 },
		{ 508, 353, 49 },
		{ 508, 354, 49 },
		{ 508, 355, 49 },
		{ 508, 356, 49 },
		{ 508, 357, 49 },
		{ 508, 358, 49 },
		{ 508, 359, 49 },
		{ 508, 360, 49 },
		{ 508, 361, 49 },
		{ 508, 362, 49 },
		{ 508, 363, 49 },
		{ 508, 364, 49 },
		{ 508, 365, 49 },
		{ 508, 366, 49 },
		{ 508, 367, 49 },
		{ 508, 368, 49 },
		{ 508, 369, 49 },
		{ 508, 370, 49 },
		{ 508, 371, 49 },
		{ 508, 372, 49 },
		{ 508, 373, 49 },
		{ 508, 374, 49 },
		{ 508, 375, 49 },
		{ 508, 376, 49 },
		{ 508, 377, 49 },
		{ 508, 378, 49 },
		{ 508, 379, 49 },
		{ 508, 380, 49 },
		{ 508, 381, 49 },
		{ 508, 382, 49 },
		{ 509, 321, 49 },
		{ 509, 322, 49 },
		{ 509, 323, 49 },
		{ 509, 324, 49 },
		{ 509, 325, 49 },
		{ 509, 326, 49 },
		{ 509, 327, 49 },
		{ 509, 328, 49 },
		{ 509, 329, 49 },
		{ 509, 330, 49 },
		{ 509, 331, 49 },
		{ 509, 332, 49 },
		{ 509, 333, 49 },
		{ 509, 334, 49 },
		{ 509, 335, 49 },
		{ 509, 336, 49 },
		{ 509, 337, 49 },
		{ 509, 338, 49 },
		{ 509, 339, 49 },
		{ 509, 340, 49 },
		{ 509, 341, 49 },
		{ 509, 342, 49 },
		{ 509, 343, 49 },
		{ 509, 344, 49 },
		{ 509, 345, 49 },
		{ 509, 346, 49 },
		{ 509, 347, 49 },
		{ 509, 348, 49 },
		{ 509, 349, 49 },
		{ 509, 350, 49 },
		{ 509, 351, 49 },
		{ 509, 352, 49 },
		{ 509, 353, 49 },
		{ 509, 354, 49 },
		{ 509, 355, 49 },
		{ 509, 356, 49 },
		{ 509, 357, 49 },
		{ 509, 358, 49 },
		{ 509, 359, 49 },
		{ 509, 360, 49 },
		{ 509, 361, 49 },
		{ 509, 362, 49 },
		{ 509, 363, 49 },
		{ 509, 364, 49 },
		{ 509, 365, 49 },
		{ 509, 366, 49 },
		{ 509, 367, 49 },
		{ 509, 368, 49 },
		{ 509, 369, 49 },
		{ 509, 370, 49 },
		{ 509, 371, 49 },
		{ 509, 372, 49 },
		{ 509, 373, 49 },
		{ 509, 374, 49 },
		{ 509, 375, 49 },
		{ 509, 376, 49 },
		{ 509, 377, 49 },
		{ 509, 378, 49 },
		{ 509, 379, 49 },
		{ 509, 380, 49 },
		{ 509, 381, 49 },
		{ 509, 382, 49 },
		{ 510, 321, 49 },
		{ 510, 322, 49 },
		{ 510, 323, 49 },
		{ 510, 324, 49 },
		{ 510, 325, 49 },
		{ 510, 326, 49 },
		{ 510, 327, 49 },
		{ 510, 328, 49 },
		{ 510, 329, 49 },
		{ 510, 330, 49 },
		{ 510, 331, 49 },
		{ 510, 332, 49 },
		{ 510, 333, 49 },
		{ 510, 334, 49 },
		{ 510, 335, 49 },
		{ 510, 336, 49 },
		{ 510, 337, 49 },
		{ 510, 338, 49 },
		{ 510, 339, 49 },
		{ 510, 340, 49 },
		{ 510, 341, 49 },
		{ 510, 342, 49 },
		{ 510, 343, 49 },
		{ 510, 344, 49 },
		{ 510, 345, 49 },
		{ 510, 346, 49 },
		{ 510, 347, 49 },
		{ 510, 348, 49 },
		{ 510, 349, 49 },
		{ 510, 350, 49 },
		{ 510, 351, 49 },
		{ 510, 352, 49 },
		{ 510, 353, 49 },
		{ 510, 354, 49 },
		{ 510, 355, 49 },
		{ 510, 356, 49 },
		{ 510, 357, 49 },
		{ 510, 358, 49 },
		{ 510, 359, 49 },
		{ 510, 360, 49 },
		{ 510, 361, 49 },
		{ 510, 362, 49 },
		{ 510, 363, 49 },
		{ 510, 364, 49 },
		{ 510, 365, 49 },
		{ 510, 366, 49 },
		{ 510, 367, 49 },
		{ 510, 368, 49 },
		{ 510, 369, 49 },
		{ 510, 370, 49 },
		{ 510, 371, 49 },
		{ 510, 372, 49 },
		{ 510, 373, 49 },
		{ 510, 374, 49 },
		{ 510, 375, 49 },
		{ 510, 376, 49 },
		{ 510, 377, 49 },
		{ 510, 378, 49 },
		{ 510, 379, 49 },
		{ 510, 380, 49 },
		{ 510, 381, 49 },
		{ 510, 382, 49 },
		{ 510, 383, 49 },
		{ 511, 321, 49 },
		{ 511, 322, 49 },
		{ 511, 323, 49 },
		{ 511, 324, 49 },
		{ 511, 325, 49 },
		{ 511, 326, 49 },
		{ 511, 327, 49 },
		{ 511, 328, 49 },
		{ 511, 329, 49 },
		{ 511, 330, 49 },
		{ 511, 331, 49 },
		{ 511, 332, 49 },
		{ 511, 333, 49 },
		{ 511, 334, 49 },
		{ 511, 335, 49 },
		{ 511, 336, 49 },
		{ 511, 337, 49 },
		{ 511, 338, 49 },
		{ 511, 339, 49 },
		{ 511, 340, 49 },
		{ 511, 341, 49 },
		{ 511, 342, 49 },
		{ 511, 343, 49 },
		{ 511, 344, 49 },
		{ 511, 345, 49 },
		{ 511, 346, 49 },
		{ 511, 347, 49 },
		{ 511, 348, 49 },
		{ 511, 349, 49 },
		{ 511, 350, 49 },
		{ 511, 351, 49 },
		{ 511, 352, 49 },
		{ 511, 353, 49 },
		{ 511, 354, 49 },
		{ 511, 355, 49 },
		{ 511, 356, 49 },
		{ 511, 357, 49 },
		{ 511, 358, 49 },
		{ 511, 359, 49 },
		{ 511, 360, 49 },
		{ 511, 361, 49 },
		{ 511, 362, 49 },
		{ 511, 363, 49 },
		{ 511, 364, 49 },
		{ 511, 365, 49 },
		{ 511, 366, 49 },
		{ 511, 367, 49 },
		{ 511, 368, 49 },
		{ 511, 369, 49 },
		{ 511, 370, 49 },
		{ 511, 371, 49 },
		{ 511, 372, 49 },
		{ 511, 373, 49 },
		{ 511, 374, 49 },
		{ 511, 375, 49 },
		{ 511, 376, 49 },
		{ 511, 377, 49 },
		{ 511, 378, 49 },
		{ 511, 379, 49 },
		{ 511, 380, 49 },
		{ 511, 381, 49 },
		{ 511, 382, 49 },
		{ 511, 383, 49 },
		{ 512, 321, 49 },
		{ 512, 322, 49 },
		{ 512, 323, 49 },
		{ 512, 324, 49 },
		{ 512, 325, 49 },
		{ 512, 326, 49 },
		{ 512, 327, 49 },
		{ 512, 328, 49 },
		{ 512, 329, 49 },
		{ 512, 330, 49 },
		{ 512, 331, 49 },
		{ 512, 332, 49 },
		{ 512, 333, 49 },
		{ 512, 334, 49 },
		{ 512, 335, 49 },
		{ 512, 336, 49 },
		{ 512, 337, 49 },
		{ 512, 338, 49 },
		{ 512, 339, 49 },
		{ 512, 340, 49 },
		{ 512, 341, 49 },
		{ 512, 342, 49 },
		{ 512, 343, 49 },
		{ 512, 344, 49 },
		{ 512, 345, 49 },
		{ 512, 346, 49 },
		{ 512, 347, 49 },
		{ 512, 348, 49 },
		{ 512, 349, 49 },
		{ 512, 350, 49 },
		{ 512, 351, 49 },
		{ 512, 352, 49 },
		{ 512, 353, 49 },
		{ 512, 354, 49 },
		{ 512, 355, 49 },
		{ 512, 356, 49 },
		{ 512, 357, 49 },
		{ 512, 358, 49 },
		{ 512, 359, 49 },
		{ 512, 360, 49 },
		{ 512, 361, 49 },
		{ 512, 362, 49 },
		{ 512, 363, 49 },
		{ 512, 364, 49 },
		{ 512, 365, 49 },
		{ 512, 366, 49 },
		{ 512, 367, 49 },
		{ 512, 368, 49 },
		{ 512, 369, 49 },
		{ 512, 370, 49 },
		{ 512, 371, 49 },
		{ 512, 372, 49 },
		{ 512, 373, 49 },
		{ 512, 374, 49 },
		{ 512, 375, 49 },
		{ 512, 376, 49 },
		{ 512, 377, 49 },
		{ 512, 378, 49 },
		{ 512, 379, 49 },
		{ 512, 380, 49 },
		{ 512, 381, 49 },
		{ 512, 382, 49 },
		{ 512, 383, 49 },
		{ 512, 384, 49 },
		{ 1, 98, 50 },
		{ 1, 99, 50 },
		{ 1, 100, 50 },
		{ 1, 101, 50 },
		{ 1, 102, 50 },
		{ 1, 103, 50 },
		{ 1, 104, 50 },
		{ 1, 105, 50 },
		{ 1, 106, 50 },
		{ 1, 107, 50 },
		{ 1, 108, 50 },
		{ 1, 109, 50 },
		{ 1, 110, 50 },
		{ 1, 111, 50 },
		{ 1, 112, 50 },
		{ 1, 113, 50 },
		{ 1, 114, 50 },
		{ 1, 115, 50 },
		{ 1, 116, 50 },
		{ 1, 117, 50 },
		{ 1, 118, 50 },
		{ 1, 119, 50 },
		{ 1, 120, 50 },
		{ 1, 121, 50 },
		{ 1, 122, 50 },
		{ 1, 123, 50 },
		{ 1, 124, 50 },
		{ 1, 125, 50 },
		{ 1, 126, 50 },
		{ 1, 127, 50 },
		{ 1, 128, 50 },
		{ 1, 129, 50 },
		{ 1, 130, 50 },
		{ 1, 131, 50 },
		{ 1, 132, 50 },
		{ 1, 133, 50 },
		{ 1, 134, 50 },
		{ 1, 135, 50 },
		{ 1, 136, 50 },
		{ 1, 137, 50 },
		{ 1, 138, 50 },
		{ 1, 139, 50 },
		{ 1, 140, 50 },
		{ 1, 141, 50 },
		{ 1, 142, 50 },
		{ 1, 143, 50 },
		{ 1, 144, 50 },
		{ 1, 145, 50 },
		{ 1, 146, 50 },
		{ 1, 147, 50 },
		{ 1, 148, 50 },
		{ 1, 149, 50 },
		{ 1, 150, 50 },
		{ 1, 151, 50 },
		{ 1, 152, 50 },
		{ 1, 153, 50 },
		{ 1, 154, 50 },
		{ 1, 155, 50 },
		{ 1, 156, 50 },
		{ 1, 157, 50 },
		{ 1, 158, 50 },
		{ 1, 159, 50 },
		{ 1, 160, 50 },
		{ 2, 98, 50 },
		{ 2, 99, 50 },
		{ 2, 100, 50 },
		{ 2, 101, 50 },
		{ 2, 102, 50 },
		{ 2, 103, 50 },
		{ 2, 104, 50 },
		{ 2, 105, 50 },
		{ 2, 106, 50 },
		{ 2, 107, 50 },
		{ 2, 108, 50 },
		{ 2, 109, 50 },
		{ 2, 110, 50 },
		{ 2, 111, 50 },
		{ 2, 112, 50 },
		{ 2, 113, 50 },
		{ 2, 114, 50 },
		{ 2, 115, 50 },
		{ 2, 116, 50 },
		{ 2, 117, 50 },
		{ 2, 118, 50 },
		{ 2, 119, 50 },
		{ 2, 120, 50 },
		{ 2, 121, 50 },
		{ 2, 122, 50 },
		{ 2, 123, 50 },
		{ 2, 124, 50 },
		{ 2, 125, 50 },
		{ 2, 126, 50 },
		{ 2, 127, 50 },
		{ 2, 128, 50 },
		{ 2, 129, 50 },
		{ 2, 130, 50 },
		{ 2, 131, 50 },
		{ 2, 132, 50 },
		{ 2, 133, 50 },
		{ 2, 134, 50 },
		{ 2, 135, 50 },
		{ 2, 136, 50 },
		{ 2, 137, 50 },
		{ 2, 138, 50 },
		{ 2, 139, 50 },
		{ 2, 140, 50 },
		{ 2, 141, 50 },
		{ 2, 142, 50 },
		{ 2, 143, 50 },
		{ 2, 144, 50 },
		{ 2, 145, 50 },
		{ 2, 146, 50 },
		{ 2, 147, 50 },
		{ 2, 148, 50 },
		{ 2, 149, 50 },
		{ 2, 150, 50 },
		{ 2, 151, 50 },
		{ 2, 152, 50 },
		{ 2, 153, 50 },
		{ 2, 154, 50 },
		{ 2, 155, 50 },
		{ 2, 156, 50 },
		{ 2, 157, 50 },
		{ 2, 158, 50 },
		{ 2, 159, 50 },
		{ 2, 160, 50 },
		{ 2, 161, 50 },
		{ 3, 99, 50 },
		{ 3, 100, 50 },
		{ 3, 101, 50 },
		{ 3, 102, 50 },
		{ 3, 103, 50 },
		{ 3, 104, 50 },
		{ 3, 105, 50 },
		{ 3, 106, 50 },
		{ 3, 107, 50 },
		{ 3, 108, 50 },
		{ 3, 109, 50 },
		{ 3, 110, 50 },
		{ 3, 111, 50 },
		{ 3, 112, 50 },
		{ 3, 113, 50 },
		{ 3, 114, 50 },
		{ 3, 115, 50 },
		{ 3, 116, 50 },
		{ 3, 117, 50 },
		{ 3, 118, 50 },
		{ 3, 119, 50 },
		{ 3, 120, 50 },
		{ 3, 121, 50 },
		{ 3, 122, 50 },
		{ 3, 123, 50 },
		{ 3, 124, 50 },
		{ 3, 125, 50 },
		{ 3, 126, 50 },
		{ 3, 127, 50 },
		{ 3, 128, 50 },
		{ 3, 129, 50 },
		{ 3, 130, 50 },
		{ 3, 131, 50 },
		{ 3, 132, 50 },
		{ 3, 133, 50 },
		{ 3, 134, 50 },
		{ 3, 135, 50 },
		{ 3, 136, 50 },
		{ 3, 137, 50 },
		{ 3, 138, 50 },
		{ 3, 139, 50 },
		{ 3, 140, 50 },
		{ 3, 141, 50 },
		{ 3, 142, 50 },
		{ 3, 143, 50 },
		{ 3, 144, 50 },
		{ 3, 145, 50 },
		{ 3, 146, 50 },
		{ 3, 147, 50 },
		{ 3, 148, 50 },
		{ 3, 149, 50 },
		{ 3, 150, 50 },
		{ 3, 151, 50 },
		{ 3, 152, 50 },
		{ 3, 153, 50 },
		{ 3, 154, 50 },
		{ 3, 155, 50 },
		{ 3, 156, 50 },
		{ 3, 157, 50 },
		{ 3, 158, 50 },
		{ 3, 159, 50 },
		{ 3, 160, 50 },
		{ 3, 161, 50 },
		{ 4, 99, 50 },
		{ 4, 100, 50 },
		{ 4, 101, 50 },
		{ 4, 102, 50 },
		{ 4, 103, 50 },
		{ 4, 104, 50 },
		{ 4, 105, 50 },
		{ 4, 106, 50 },
		{ 4, 107, 50 },
		{ 4, 108, 50 },
		{ 4, 109, 50 },
		{ 4, 110, 50 },
		{ 4, 111, 50 },
		{ 4, 112, 50 },
		{ 4, 113, 50 },
		{ 4, 114, 50 },
		{ 4, 115, 50 },
		{ 4, 116, 50 },
		{ 4, 117, 50 },
		{ 4, 118, 50 },
		{ 4, 119, 50 },
		{ 4, 120, 50 },
		{ 4, 121, 50 },
		{ 4, 122, 50 },
		{ 4, 123, 50 },
		{ 4, 124, 50 },
		{ 4, 125, 50 },
		{ 4, 126, 50 },
		{ 4, 127, 50 },
		{ 4, 128, 50 },
		{ 4, 129, 50 },
		{ 4, 130, 50 },
		{ 4, 131, 50 },
		{ 4, 132, 50 },
		{ 4, 133, 50 },
		{ 4, 134, 50 },
		{ 4, 135, 50 },
		{ 4, 136, 50 },
		{ 4, 137, 50 },
		{ 4, 138, 50 },
		{ 4, 139, 50 },
		{ 4, 140, 50 },
		{ 4, 141, 50 },
		{ 4, 142, 50 },
		{ 4, 143, 50 },
		{ 4, 144, 50 },
		{ 4, 145, 50 },
		{ 4, 146, 50 },
		{ 4, 147, 50 },
		{ 4, 148, 50 },
		{ 4, 149, 50 },
		{ 4, 150, 50 },
		{ 4, 151, 50 },
		{ 4, 152, 50 },
		{ 4, 153, 50 },
		{ 4, 154, 50 },
		{ 4, 155, 50 },
		{ 4, 156, 50 },
		{ 4, 157, 50 },
		{ 4, 158, 50 },
		{ 4, 159, 50 },
		{ 4, 160, 50 },
		{ 4, 161, 50 },
		{ 4, 162, 50 },
		{ 5, 100, 50 },
		{ 5, 101, 50 },
		{ 5, 102, 50 },
		{ 5, 103, 50 },
		{ 5, 104, 50 },
		{ 5, 105, 50 },
		{ 5, 106, 50 },
		{ 5, 107, 50 },
		{ 5, 108, 50 },
		{ 5, 109, 50 },
		{ 5, 110, 50 },
		{ 5, 111, 50 },
		{ 5, 112, 50 },
		{ 5, 113, 50 },
		{ 5, 114, 50 },
		{ 5, 115, 50 },
		{ 5, 116, 50 },
		{ 5, 117, 50 },
		{ 5, 118, 50 },
		{ 5, 119, 50 },
		{ 5, 120, 50 },
		{ 5, 121, 50 },
		{ 5, 122, 50 },
		{ 5, 123, 50 },
		{ 5, 124, 50 },
		{ 5, 125, 50 },
		{ 5, 126, 50 },
		{ 5, 127, 50 },
		{ 5, 128, 50 },
		{ 5, 129, 50 },
		{ 5, 130, 50 },
		{ 5, 131, 50 },
		{ 5, 132, 50 },
		{ 5, 133, 50 },
		{ 5, 134, 50 },
		{ 5, 135, 50 },
		{ 5, 136, 50 },
		{ 5, 137, 50 },
		{ 5, 138, 50 },
		{ 5, 139, 50 },
		{ 5, 140, 50 },
		{ 5, 141, 50 },
		{ 5, 142, 50 },
		{ 5, 143, 50 },
		{ 5, 144, 50 },
		{ 5, 145, 50 },
		{ 5, 146, 50 },
		{ 5, 147, 50 },
		{ 5, 148, 50 },
		{ 5, 149, 50 },
		{ 5, 150, 50 },
		{ 5, 151, 50 },
		{ 5, 152, 50 },
		{ 5, 153, 50 },
		{ 5, 154, 50 },
		{ 5, 155, 50 },
		{ 5, 156, 50 },
		{ 5, 157, 50 },
		{ 5, 158, 50 },
		{ 5, 159, 50 },
		{ 5, 160, 50 },
		{ 5, 161, 50 },
		{ 5, 162, 50 },
		{ 6, 101, 50 },
		{ 6, 102, 50 },
		{ 6, 103, 50 },
		{ 6, 104, 50 },
		{ 6, 105, 50 },
		{ 6, 106, 50 },
		{ 6, 107, 50 },
		{ 6, 108, 50 },
		{ 6, 109, 50 },
		{ 6, 110, 50 },
		{ 6, 111, 50 },
		{ 6, 112, 50 },
		{ 6, 113, 50 },
		{ 6, 114, 50 },
		{ 6, 115, 50 },
		{ 6, 116, 50 },
		{ 6, 117, 50 },
		{ 6, 118, 50 },
		{ 6, 119, 50 },
		{ 6, 120, 50 },
		{ 6, 121, 50 },
		{ 6, 122, 50 },
		{ 6, 123, 50 },
		{ 6, 124, 50 },
		{ 6, 125, 50 },
		{ 6, 126, 50 },
		{ 6, 127, 50 },
		{ 6, 128, 50 },
		{ 6, 129, 50 },
		{ 6, 130, 50 },
		{ 6, 131, 50 },
		{ 6, 132, 50 },
		{ 6, 133, 50 },
		{ 6, 134, 50 },
		{ 6, 135, 50 },
		{ 6, 136, 50 },
		{ 6, 137, 50 },
		{ 6, 138, 50 },
		{ 6, 139, 50 },
		{ 6, 140, 50 },
		{ 6, 141, 50 },
		{ 6, 142, 50 },
		{ 6, 143, 50 },
		{ 6, 144, 50 },
		{ 6, 145, 50 },
		{ 6, 146, 50 },
		{ 6, 147, 50 },
		{ 6, 148, 50 },
		{ 6, 149, 50 },
		{ 6, 150, 50 },
		{ 6, 151, 50 },
		{ 6, 152, 50 },
		{ 6, 153, 50 },
		{ 6, 154, 50 },
		{ 6, 155, 50 },
		{ 6, 156, 50 },
		{ 6, 157, 50 },
		{ 6, 158, 50 },
		{ 6, 159, 50 },
		{ 6, 160, 50 },
		{ 6, 161, 50 },
		{ 6, 162, 50 },
		{ 7, 101, 50 },
		{ 7, 102, 50 },
		{ 7, 103, 50 },
		{ 7, 104, 50 },
		{ 7, 105, 50 },
		{ 7, 106, 50 },
		{ 7, 107, 50 },
		{ 7, 108, 50 },
		{ 7, 109, 50 },
		{ 7, 110, 50 },
		{ 7, 111, 50 },
		{ 7, 112, 50 },
		{ 7, 113, 50 },
		{ 7, 114, 50 },
		{ 7, 115, 50 },
		{ 7, 116, 50 },
		{ 7, 117, 50 },
		{ 7, 118, 50 },
		{ 7, 119, 50 },
		{ 7, 120, 50 },
		{ 7, 121, 50 },
		{ 7, 122, 50 },
		{ 7, 123, 50 },
		{ 7, 124, 50 },
		{ 7, 125, 50 },
		{ 7, 126, 50 },
		{ 7, 127, 50 },
		{ 7, 128, 50 },
		{ 7, 129, 50 },
		{ 7, 130, 50 },
		{ 7, 131, 50 },
		{ 7, 132, 50 },
		{ 7, 133, 50 },
		{ 7, 134, 50 },
		{ 7, 135, 50 },
		{ 7, 136, 50 },
		{ 7, 137, 50 },
		{ 7, 138, 50 },
		{ 7, 139, 50 },
		{ 7, 140, 50 },
		{ 7, 141, 50 },
		{ 7, 142, 50 },
		{ 7, 143, 50 },
		{ 7, 144, 50 },
		{ 7, 145, 50 },
		{ 7, 146, 50 },
		{ 7, 147, 50 },
		{ 7, 148, 50 },
		{ 7, 149, 50 },
		{ 7, 150, 50 },
		{ 7, 151, 50 },
		{ 7, 152, 50 },
		{ 7, 153, 50 },
		{ 7, 154, 50 },
		{ 7, 155, 50 },
		{ 7, 156, 50 },
		{ 7, 157, 50 },
		{ 7, 158, 50 },
		{ 7, 159, 50 },
		{ 7, 160, 50 },
		{ 7, 161, 50 },
		{ 7, 162, 50 },
		{ 7, 163, 50 },
		{ 8, 102, 50 },
		{ 8, 103, 50 },
		{ 8, 104, 50 },
		{ 8, 105, 50 },
		{ 8, 106, 50 },
		{ 8, 107, 50 },
		{ 8, 108, 50 },
		{ 8, 109, 50 },
		{ 8, 110, 50 },
		{ 8, 111, 50 },
		{ 8, 112, 50 },
		{ 8, 113, 50 },
		{ 8, 114, 50 },
		{ 8, 115, 50 },
		{ 8, 116, 50 },
		{ 8, 117, 50 },
		{ 8, 118, 50 },
		{ 8, 119, 50 },
		{ 8, 120, 50 },
		{ 8, 121, 50 },
		{ 8, 122, 50 },
		{ 8, 123, 50 },
		{ 8, 124, 50 },
		{ 8, 125, 50 },
		{ 8, 126, 50 },
		{ 8, 127, 50 },
		{ 8, 128, 50 },
		{ 8, 129, 50 },
		{ 8, 130, 50 },
		{ 8, 131, 50 },
		{ 8, 132, 50 },
		{ 8, 133, 50 },
		{ 8, 134, 50 },
		{ 8, 135, 50 },
		{ 8, 136, 50 },
		{ 8, 137, 50 },
		{ 8, 138, 50 },
		{ 8, 139, 50 },
		{ 8, 140, 50 },
		{ 8, 141, 50 },
		{ 8, 142, 50 },
		{ 8, 143, 50 },
		{ 8, 144, 50 },
		{ 8, 145, 50 },
		{ 8, 146, 50 },
		{ 8, 147, 50 },
		{ 8, 148, 50 },
		{ 8, 149, 50 },
		{ 8, 150, 50 },
		{ 8, 151, 50 },
		{ 8, 152, 50 },
		{ 8, 153, 50 },
		{ 8, 154, 50 },
		{ 8, 155, 50 },
		{ 8, 156, 50 },
		{ 8, 157, 50 },
		{ 8, 158, 50 },
		{ 8, 159, 50 },
		{ 8, 160, 50 },
		{ 8, 161, 50 },
		{ 8, 162, 50 },
		{ 8, 163, 50 },
		{ 9, 103, 50 },
		{ 9, 104, 50 },
		{ 9, 105, 50 },
		{ 9, 106, 50 },
		{ 9, 107, 50 },
		{ 9, 108, 50 },
		{ 9, 109, 50 },
		{ 9, 110, 50 },
		{ 9, 111, 50 },
		{ 9, 112, 50 },
		{ 9, 113, 50 },
		{ 9, 114, 50 },
		{ 9, 115, 50 },
		{ 9, 116, 50 },
		{ 9, 117, 50 },
		{ 9, 118, 50 },
		{ 9, 119, 50 },
		{ 9, 120, 50 },
		{ 9, 121, 50 },
		{ 9, 122, 50 },
		{ 9, 123, 50 },
		{ 9, 124, 50 },
		{ 9, 125, 50 },
		{ 9, 126, 50 },
		{ 9, 127, 50 },
		{ 9, 128, 50 },
		{ 9, 129, 50 },
		{ 9, 130, 50 },
		{ 9, 131, 50 },
		{ 9, 132, 50 },
		{ 9, 133, 50 },
		{ 9, 134, 50 },
		{ 9, 135, 50 },
		{ 9, 136, 50 },
		{ 9, 137, 50 },
		{ 9, 138, 50 },
		{ 9, 139, 50 },
		{ 9, 140, 50 },
		{ 9, 141, 50 },
		{ 9, 142, 50 },
		{ 9, 143, 50 },
		{ 9, 144, 50 },
		{ 9, 145, 50 },
		{ 9, 146, 50 },
		{ 9, 147, 50 },
		{ 9, 148, 50 },
		{ 9, 149, 50 },
		{ 9, 150, 50 },
		{ 9, 151, 50 },
		{ 9, 152, 50 },
		{ 9, 153, 50 },
		{ 9, 154, 50 },
		{ 9, 155, 50 },
		{ 9, 156, 50 },
		{ 9, 157, 50 },
		{ 9, 158, 50 },
		{ 9, 159, 50 },
		{ 9, 160, 50 },
		{ 9, 161, 50 },
		{ 9, 162, 50 },
		{ 9, 163, 50 },
		{ 10, 103, 50 },
		{ 10, 104, 50 },
		{ 10, 105, 50 },
		{ 10, 106, 50 },
		{ 10, 107, 50 },
		{ 10, 108, 50 },
		{ 10, 109, 50 },
		{ 10, 110, 50 },
		{ 10, 111, 50 },
		{ 10, 112, 50 },
		{ 10, 113, 50 },
		{ 10, 114, 50 },
		{ 10, 115, 50 },
		{ 10, 116, 50 },
		{ 10, 117, 50 },
		{ 10, 118, 50 },
		{ 10, 119, 50 },
		{ 10, 120, 50 },
		{ 10, 121, 50 },
		{ 10, 122, 50 },
		{ 10, 123, 50 },
		{ 10, 124, 50 },
		{ 10, 125, 50 },
		{ 10, 126, 50 },
		{ 10, 127, 50 },
		{ 10, 128, 50 },
		{ 10, 129, 50 },
		{ 10, 130, 50 },
		{ 10, 131, 50 },
		{ 10, 132, 50 },
		{ 10, 133, 50 },
		{ 10, 134, 50 },
		{ 10, 135, 50 },
		{ 10, 136, 50 },
		{ 10, 137, 50 },
		{ 10, 138, 50 },
		{ 10, 139, 50 },
		{ 10, 140, 50 },
		{ 10, 141, 50 },
		{ 10, 142, 50 },
		{ 10, 143, 50 },
		{ 10, 144, 50 },
		{ 10, 145, 50 },
		{ 10, 146, 50 },
		{ 10, 147, 50 },
		{ 10, 148, 50 },
		{ 10, 149, 50 },
		{ 10, 150, 50 },
		{ 10, 151, 50 },
		{ 10, 152, 50 },
		{ 10, 153, 50 },
		{ 10, 154, 50 },
		{ 10, 155, 50 },
		{ 10, 156, 50 },
		{ 10, 157, 50 },
		{ 10, 158, 50 },
		{ 10, 159, 50 },
		{ 10, 160, 50 },
		{ 10, 161, 50 },
		{ 10, 162, 50 },
		{ 10, 163, 50 },
		{ 10, 164, 50 },
		{ 11, 104, 50 },
		{ 11, 105, 50 },
		{ 11, 106, 50 },
		{ 11, 107, 50 },
		{ 11, 108, 50 },
		{ 11, 109, 50 },
		{ 11, 110, 50 },
		{ 11, 111, 50 },
		{ 11, 112, 50 },
		{ 11, 113, 50 },
		{ 11, 114, 50 },
		{ 11, 115, 50 },
		{ 11, 116, 50 },
		{ 11, 117, 50 },
		{ 11, 118, 50 },
		{ 11, 119, 50 },
		{ 11, 120, 50 },
		{ 11, 121, 50 },
		{ 11, 122, 50 },
		{ 11, 123, 50 },
		{ 11, 124, 50 },
		{ 11, 125, 50 },
		{ 11, 126, 50 },
		{ 11, 127, 50 },
		{ 11, 128, 50 },
		{ 11, 129, 50 },
		{ 11, 130, 50 },
		{ 11, 131, 50 },
		{ 11, 132, 50 },
		{ 11, 133, 50 },
		{ 11, 134, 50 },
		{ 11, 135, 50 },
		{ 11, 136, 50 },
		{ 11, 137, 50 },
		{ 11, 138, 50 },
		{ 11, 139, 50 },
		{ 11, 140, 50 },
		{ 11, 141, 50 },
		{ 11, 142, 50 },
		{ 11, 143, 50 },
		{ 11, 144, 50 },
		{ 11, 145, 50 },
		{ 11, 146, 50 },
		{ 11, 147, 50 },
		{ 11, 148, 50 },
		{ 11, 149, 50 },
		{ 11, 150, 50 },
		{ 11, 151, 50 },
		{ 11, 152, 50 },
		{ 11, 153, 50 },
		{ 11, 154, 50 },
		{ 11, 155, 50 },
		{ 11, 156, 50 },
		{ 11, 157, 50 },
		{ 11, 158, 50 },
		{ 11, 159, 50 },
		{ 11, 160, 50 },
		{ 11, 161, 50 },
		{ 11, 162, 50 },
		{ 11, 163, 50 },
		{ 11, 164, 50 },
		{ 12, 104, 50 },
		{ 12, 105, 50 },
		{ 12, 106, 50 },
		{ 12, 107, 50 },
		{ 12, 108, 50 },
		{ 12, 109, 50 },
		{ 12, 110, 50 },
		{ 12, 111, 50 },
		{ 12, 112, 50 },
		{ 12, 113, 50 },
		{ 12, 114, 50 },
		{ 12, 115, 50 },
		{ 12, 116, 50 },
		{ 12, 117, 50 },
		{ 12, 118, 50 },
		{ 12, 119, 50 },
		{ 12, 120, 50 },
		{ 12, 121, 50 },
		{ 12, 122, 50 },
		{ 12, 123, 50 },
		{ 12, 124, 50 },
		{ 12, 125, 50 },
		{ 12, 126, 50 },
		{ 12, 127, 50 },
		{ 12, 128, 50 },
		{ 12, 129, 50 },
		{ 12, 130, 50 },
		{ 12, 131, 50 },
		{ 12, 132, 50 },
		{ 12, 133, 50 },
		{ 12, 134, 50 },
		{ 12, 135, 50 },
		{ 12, 136, 50 },
		{ 12, 137, 50 },
		{ 12, 138, 50 },
		{ 12, 139, 50 },
		{ 12, 140, 50 },
		{ 12, 141, 50 },
		{ 12, 142, 50 },
		{ 12, 143, 50 },
		{ 12, 144, 50 },
		{ 12, 145, 50 },
		{ 12, 146, 50 },
		{ 12, 147, 50 },
		{ 12, 148, 50 },
		{ 12, 149, 50 },
		{ 12, 150, 50 },
		{ 12, 151, 50 },
		{ 12, 152, 50 },
		{ 12, 153, 50 },
		{ 12, 154, 50 },
		{ 12, 155, 50 },
		{ 12, 156, 50 },
		{ 12, 157, 50 },
		{ 12, 158, 50 },
		{ 12, 159, 50 },
		{ 12, 160, 50 },
		{ 12, 161, 50 },
		{ 12, 162, 50 },
		{ 12, 163, 50 },
		{ 12, 164, 50 },
		{ 12, 165, 50 },
		{ 13, 105, 50 },
		{ 13, 106, 50 },
		{ 13, 107, 50 },
		{ 13, 108, 50 },
		{ 13, 109, 50 },
		{ 13, 110, 50 },
		{ 13, 111, 50 },
		{ 13, 112, 50 },
		{ 13, 113, 50 },
		{ 13, 114, 50 },
		{ 13, 115, 50 },
		{ 13, 116, 50 },
		{ 13, 117, 50 },
		{ 13, 118, 50 },
		{ 13, 119, 50 },
		{ 13, 120, 50 },
		{ 13, 121, 50 },
		{ 13, 122, 50 },
		{ 13, 123, 50 },
		{ 13, 124, 50 },
		{ 13, 125, 50 },
		{ 13, 126, 50 },
		{ 13, 127, 50 },
		{ 13, 128, 50 },
		{ 13, 129, 50 },
		{ 13, 130, 50 },
		{ 13, 131, 50 },
		{ 13, 132, 50 },
		{ 13, 133, 50 },
		{ 13, 134, 50 },
		{ 13, 135, 50 },
		{ 13, 136, 50 },
		{ 13, 137, 50 },
		{ 13, 138, 50 },
		{ 13, 139, 50 },
		{ 13, 140, 50 },
		{ 13, 141, 50 },
		{ 13, 142, 50 },
		{ 13, 143, 50 },
		{ 13, 144, 50 },
		{ 13, 145, 50 },
		{ 13, 146, 50 },
		{ 13, 147, 50 },
		{ 13, 148, 50 },
		{ 13, 149, 50 },
		{ 13, 150, 50 },
		{ 13, 151, 50 },
		{ 13, 152, 50 },
		{ 13, 153, 50 },
		{ 13, 154, 50 },
		{ 13, 155, 50 },
		{ 13, 156, 50 },
		{ 13, 157, 50 },
		{ 13, 158, 50 },
		{ 13, 159, 50 },
		{ 13, 160, 50 },
		{ 13, 161, 50 },
		{ 13, 162, 50 },
		{ 13, 163, 50 },
		{ 13, 164, 50 },
		{ 13, 165, 50 },
		{ 14, 106, 50 },
		{ 14, 107, 50 },
		{ 14, 108, 50 },
		{ 14, 109, 50 },
		{ 14, 110, 50 },
		{ 14, 111, 50 },
		{ 14, 112, 50 },
		{ 14, 113, 50 },
		{ 14, 114, 50 },
		{ 14, 115, 50 },
		{ 14, 116, 50 },
		{ 14, 117, 50 },
		{ 14, 118, 50 },
		{ 14, 119, 50 },
		{ 14, 120, 50 },
		{ 14, 121, 50 },
		{ 14, 122, 50 },
		{ 14, 123, 50 },
		{ 14, 124, 50 },
		{ 14, 125, 50 },
		{ 14, 126, 50 },
		{ 14, 127, 50 },
		{ 14, 128, 50 },
		{ 14, 129, 50 },
		{ 14, 130, 50 },
		{ 14, 131, 50 },
		{ 14, 132, 50 },
		{ 14, 133, 50 },
		{ 14, 134, 50 },
		{ 14, 135, 50 },
		{ 14, 136, 50 },
		{ 14, 137, 50 },
		{ 14, 138, 50 },
		{ 14, 139, 50 },
		{ 14, 140, 50 },
		{ 14, 141, 50 },
		{ 14, 142, 50 },
		{ 14, 143, 50 },
		{ 14, 144, 50 },
		{ 14, 145, 50 },
		{ 14, 146, 50 },
		{ 14, 147, 50 },
		{ 14, 148, 50 },
		{ 14, 149, 50 },
		{ 14, 150, 50 },
		{ 14, 151, 50 },
		{ 14, 152, 50 },
		{ 14, 153, 50 },
		{ 14, 154, 50 },
		{ 14, 155, 50 },
		{ 14, 156, 50 },
		{ 14, 157, 50 },
		{ 14, 158, 50 },
		{ 14, 159, 50 },
		{ 14, 160, 50 },
		{ 14, 161, 50 },
		{ 14, 162, 50 },
		{ 14, 163, 50 },
		{ 14, 164, 50 },
		{ 14, 165, 50 },
		{ 15, 106, 50 },
		{ 15, 107, 50 },
		{ 15, 108, 50 },
		{ 15, 109, 50 },
		{ 15, 110, 50 },
		{ 15, 111, 50 },
		{ 15, 112, 50 },
		{ 15, 113, 50 },
		{ 15, 114, 50 },
		{ 15, 115, 50 },
		{ 15, 116, 50 },
		{ 15, 117, 50 },
		{ 15, 118, 50 },
		{ 15, 119, 50 },
		{ 15, 120, 50 },
		{ 15, 121, 50 },
		{ 15, 122, 50 },
		{ 15, 123, 50 },
		{ 15, 124, 50 },
		{ 15, 125, 50 },
		{ 15, 126, 50 },
		{ 15, 127, 50 },
		{ 15, 128, 50 },
		{ 15, 129, 50 },
		{ 15, 130, 50 },
		{ 15, 131, 50 },
		{ 15, 132, 50 },
		{ 15, 133, 50 },
		{ 15, 134, 50 },
		{ 15, 135, 50 },
		{ 15, 136, 50 },
		{ 15, 137, 50 },
		{ 15, 138, 50 },
		{ 15, 139, 50 },
		{ 15, 140, 50 },
		{ 15, 141, 50 },
		{ 15, 142, 50 },
		{ 15, 143, 50 },
		{ 15, 144, 50 },
		{ 15, 145, 50 },
		{ 15, 146, 50 },
		{ 15, 147, 50 },
		{ 15, 148, 50 },
		{ 15, 149, 50 },
		{ 15, 150, 50 },
		{ 15, 151, 50 },
		{ 15, 152, 50 },
		{ 15, 153, 50 },
		{ 15, 154, 50 },
		{ 15, 155, 50 },
		{ 15, 156, 50 },
		{ 15, 157, 50 },
		{ 15, 158, 50 },
		{ 15, 159, 50 },
		{ 15, 160, 50 },
		{ 15, 161, 50 },
		{ 15, 162, 50 },
		{ 15, 163, 50 },
		{ 15, 164, 50 },
		{ 15, 165, 50 },
		{ 15, 166, 50 },
		{ 16, 107, 50 },
		{ 16, 108, 50 },
		{ 16, 109, 50 },
		{ 16, 110, 50 },
		{ 16, 111, 50 },
		{ 16, 112, 50 },
		{ 16, 113, 50 },
		{ 16, 114, 50 },
		{ 16, 115, 50 },
		{ 16, 116, 50 },
		{ 16, 117, 50 },
		{ 16, 118, 50 },
		{ 16, 119, 50 },
		{ 16, 120, 50 },
		{ 16, 121, 50 },
		{ 16, 122, 50 },
		{ 16, 123, 50 },
		{ 16, 124, 50 },
		{ 16, 125, 50 },
		{ 16, 126, 50 },
		{ 16, 127, 50 },
		{ 16, 128, 50 },
		{ 16, 129, 50 },
		{ 16, 130, 50 },
		{ 16, 131, 50 },
		{ 16, 132, 50 },
		{ 16, 133, 50 },
		{ 16, 134, 50 },
		{ 16, 135, 50 },
		{ 16, 136, 50 },
		{ 16, 137, 50 },
		{ 16, 138, 50 },
		{ 16, 139, 50 },
		{ 16, 140, 50 },
		{ 16, 141, 50 },
		{ 16, 142, 50 },
		{ 16, 143, 50 },
		{ 16, 144, 50 },
		{ 16, 145, 50 },
		{ 16, 146, 50 },
		{ 16, 147, 50 },
		{ 16, 148, 50 },
		{ 16, 149, 50 },
		{ 16, 150, 50 },
		{ 16, 151, 50 },
		{ 16, 152, 50 },
		{ 16, 153, 50 },
		{ 16, 154, 50 },
		{ 16, 155, 50 },
		{ 16, 156, 50 },
		{ 16, 157, 50 },
		{ 16, 158, 50 },
		{ 16, 159, 50 },
		{ 16, 160, 50 },
		{ 16, 161, 50 },
		{ 16, 162, 50 },
		{ 16, 163, 50 },
		{ 16, 164, 50 },
		{ 16, 165, 50 },
		{ 16, 166, 50 },
		{ 17, 108, 50 },
		{ 17, 109, 50 },
		{ 17, 110, 50 },
		{ 17, 111, 50 },
		{ 17, 112, 50 },
		{ 17, 113, 50 },
		{ 17, 114, 50 },
		{ 17, 115, 50 },
		{ 17, 116, 50 },
		{ 17, 117, 50 },
		{ 17, 118, 50 },
		{ 17, 119, 50 },
		{ 17, 120, 50 },
		{ 17, 121, 50 },
		{ 17, 122, 50 },
		{ 17, 123, 50 },
		{ 17, 124, 50 },
		{ 17, 125, 50 },
		{ 17, 126, 50 },
		{ 17, 127, 50 },
		{ 17, 128, 50 },
		{ 17, 129, 50 },
		{ 17, 130, 50 },
		{ 17, 131, 50 },
		{ 17, 132, 50 },
		{ 17, 133, 50 },
		{ 17, 134, 50 },
		{ 17, 135, 50 },
		{ 17, 136, 50 },
		{ 17, 137, 50 },
		{ 17, 138, 50 },
		{ 17, 139, 50 },
		{ 17, 140, 50 },
		{ 17, 141, 50 },
		{ 17, 142, 50 },
		{ 17, 143, 50 },
		{ 17, 144, 50 },
		{ 17, 145, 50 },
		{ 17, 146, 50 },
		{ 17, 147, 50 },
		{ 17, 148, 50 },
		{ 17, 149, 50 },
		{ 17, 150, 50 },
		{ 17, 151, 50 },
		{ 17, 152, 50 },
		{ 17, 153, 50 },
		{ 17, 154, 50 },
		{ 17, 155, 50 },
		{ 17, 156, 50 },
		{ 17, 157, 50 },
		{ 17, 158, 50 },
		{ 17, 159, 50 },
		{ 17, 160, 50 },
		{ 17, 161, 50 },
		{ 17, 162, 50 },
		{ 17, 163, 50 },
		{ 17, 164, 50 },
		{ 17, 165, 50 },
		{ 17, 166, 50 },
		{ 18, 108, 50 },
		{ 18, 109, 50 },
		{ 18, 110, 50 },
		{ 18, 111, 50 },
		{ 18, 112, 50 },
		{ 18, 113, 50 },
		{ 18, 114, 50 },
		{ 18, 115, 50 },
		{ 18, 116, 50 },
		{ 18, 117, 50 },
		{ 18, 118, 50 },
		{ 18, 119, 50 },
		{ 18, 120, 50 },
		{ 18, 121, 50 },
		{ 18, 122, 50 },
		{ 18, 123, 50 },
		{ 18, 124, 50 },
		{ 18, 125, 50 },
		{ 18, 126, 50 },
		{ 18, 127, 50 },
		{ 18, 128, 50 },
		{ 18, 129, 50 },
		{ 18, 130, 50 },
		{ 18, 131, 50 },
		{ 18, 132, 50 },
		{ 18, 133, 50 },
		{ 18, 134, 50 },
		{ 18, 135, 50 },
		{ 18, 136, 50 },
		{ 18, 137, 50 },
		{ 18, 138, 50 },
		{ 18, 139, 50 },
		{ 18, 140, 50 },
		{ 18, 141, 50 },
		{ 18, 142, 50 },
		{ 18, 143, 50 },
		{ 18, 144, 50 },
		{ 18, 145, 50 },
		{ 18, 146, 50 },
		{ 18, 147, 50 },
		{ 18, 148, 50 },
		{ 18, 149, 50 },
		{ 18, 150, 50 },
		{ 18, 151, 50 },
		{ 18, 152, 50 },
		{ 18, 153, 50 },
		{ 18, 154, 50 },
		{ 18, 155, 50 },
		{ 18, 156, 50 },
		{ 18, 157, 50 },
		{ 18, 158, 50 },
		{ 18, 159, 50 },
		{ 18, 160, 50 },
		{ 18, 161, 50 },
		{ 18, 162, 50 },
		{ 18, 163, 50 },
		{ 18, 164, 50 },
		{ 18, 165, 50 },
		{ 18, 166, 50 },
		{ 18, 167, 50 },
		{ 19, 109, 50 },
		{ 19, 110, 50 },
		{ 19, 111, 50 },
		{ 19, 112, 50 },
		{ 19, 113, 50 },
		{ 19, 114, 50 },
		{ 19, 115, 50 },
		{ 19, 116, 50 },
		{ 19, 117, 50 },
		{ 19, 118, 50 },
		{ 19, 119, 50 },
		{ 19, 120, 50 },
		{ 19, 121, 50 },
		{ 19, 122, 50 },
		{ 19, 123, 50 },
		{ 19, 124, 50 },
		{ 19, 125, 50 },
		{ 19, 126, 50 },
		{ 19, 127, 50 },
		{ 19, 128, 50 },
		{ 19, 129, 50 },
		{ 19, 130, 50 },
		{ 19, 131, 50 },
		{ 19, 132, 50 },
		{ 19, 133, 50 },
		{ 19, 134, 50 },
		{ 19, 135, 50 },
		{ 19, 136, 50 },
		{ 19, 137, 50 },
		{ 19, 138, 50 },
		{ 19, 139, 50 },
		{ 19, 140, 50 },
		{ 19, 141, 50 },
		{ 19, 142, 50 },
		{ 19, 143, 50 },
		{ 19, 144, 50 },
		{ 19, 145, 50 },
		{ 19, 146, 50 },
		{ 19, 147, 50 },
		{ 19, 148, 50 },
		{ 19, 149, 50 },
		{ 19, 150, 50 },
		{ 19, 151, 50 },
		{ 19, 152, 50 },
		{ 19, 153, 50 },
		{ 19, 154, 50 },
		{ 19, 155, 50 },
		{ 19, 156, 50 },
		{ 19, 157, 50 },
		{ 19, 158, 50 },
		{ 19, 159, 50 },
		{ 19, 160, 50 },
		{ 19, 161, 50 },
		{ 19, 162, 50 },
		{ 19, 163, 50 },
		{ 19, 164, 50 },
		{ 19, 165, 50 },
		{ 19, 166, 50 },
		{ 19, 167, 50 },
		{ 20, 109, 50 },
		{ 20, 110, 50 },
		{ 20, 111, 50 },
		{ 20, 112, 50 },
		{ 20, 113, 50 },
		{ 20, 114, 50 },
		{ 20, 115, 50 },
		{ 20, 116, 50 },
		{ 20, 117, 50 },
		{ 20, 118, 50 },
		{ 20, 119, 50 },
		{ 20, 120, 50 },
		{ 20, 121, 50 },
		{ 20, 122, 50 },
		{ 20, 123, 50 },
		{ 20, 124, 50 },
		{ 20, 125, 50 },
		{ 20, 126, 50 },
		{ 20, 127, 50 },
		{ 20, 128, 50 },
		{ 20, 129, 50 },
		{ 20, 130, 50 },
		{ 20, 131, 50 },
		{ 20, 132, 50 },
		{ 20, 133, 50 },
		{ 20, 134, 50 },
		{ 20, 135, 50 },
		{ 20, 136, 50 },
		{ 20, 137, 50 },
		{ 20, 138, 50 },
		{ 20, 139, 50 },
		{ 20, 140, 50 },
		{ 20, 141, 50 },
		{ 20, 142, 50 },
		{ 20, 143, 50 },
		{ 20, 144, 50 },
		{ 20, 145, 50 },
		{ 20, 146, 50 },
		{ 20, 147, 50 },
		{ 20, 148, 50 },
		{ 20, 149, 50 },
		{ 20, 150, 50 },
		{ 20, 151, 50 },
		{ 20, 152, 50 },
		{ 20, 153, 50 },
		{ 20, 154, 50 },
		{ 20, 155, 50 },
		{ 20, 156, 50 },
		{ 20, 157, 50 },
		{ 20, 158, 50 },
		{ 20, 159, 50 },
		{ 20, 160, 50 },
		{ 20, 161, 50 },
		{ 20, 162, 50 },
		{ 20, 163, 50 },
		{ 20, 164, 50 },
		{ 20, 165, 50 },
		{ 20, 166, 50 },
		{ 20, 167, 50 },
		{ 20, 168, 50 },
		{ 21, 110, 50 },
		{ 21, 111, 50 },
		{ 21, 112, 50 },
		{ 21, 113, 50 },
		{ 21, 114, 50 },
		{ 21, 115, 50 },
		{ 21, 116, 50 },
		{ 21, 117, 50 },
		{ 21, 118, 50 },
		{ 21, 119, 50 },
		{ 21, 120, 50 },
		{ 21, 121, 50 },
		{ 21, 122, 50 },
		{ 21, 123, 50 },
		{ 21, 124, 50 },
		{ 21, 125, 50 },
		{ 21, 126, 50 },
		{ 21, 127, 50 },
		{ 21, 128, 50 },
		{ 21, 129, 50 },
		{ 21, 130, 50 },
		{ 21, 131, 50 },
		{ 21, 132, 50 },
		{ 21, 133, 50 },
		{ 21, 134, 50 },
		{ 21, 135, 50 },
		{ 21, 136, 50 },
		{ 21, 137, 50 },
		{ 21, 138, 50 },
		{ 21, 139, 50 },
		{ 21, 140, 50 },
		{ 21, 141, 50 },
		{ 21, 142, 50 },
		{ 21, 143, 50 },
		{ 21, 144, 50 },
		{ 21, 145, 50 },
		{ 21, 146, 50 },
		{ 21, 147, 50 },
		{ 21, 148, 50 },
		{ 21, 149, 50 },
		{ 21, 150, 50 },
		{ 21, 151, 50 },
		{ 21, 152, 50 },
		{ 21, 153, 50 },
		{ 21, 154, 50 },
		{ 21, 155, 50 },
		{ 21, 156, 50 },
		{ 21, 157, 50 },
		{ 21, 158, 50 },
		{ 21, 159, 50 },
		{ 21, 160, 50 },
		{ 21, 161, 50 },
		{ 21, 162, 50 },
		{ 21, 163, 50 },
		{ 21, 164, 50 },
		{ 21, 165, 50 },
		{ 21, 166, 50 },
		{ 21, 167, 50 },
		{ 21, 168, 50 },
		{ 22, 111, 50 },
		{ 22, 112, 50 },
		{ 22, 113, 50 },
		{ 22, 114, 50 },
		{ 22, 115, 50 },
		{ 22, 116, 50 },
		{ 22, 117, 50 },
		{ 22, 118, 50 },
		{ 22, 119, 50 },
		{ 22, 120, 50 },
		{ 22, 121, 50 },
		{ 22, 122, 50 },
		{ 22, 123, 50 },
		{ 22, 124, 50 },
		{ 22, 125, 50 },
		{ 22, 126, 50 },
		{ 22, 127, 50 },
		{ 22, 128, 50 },
		{ 22, 129, 50 },
		{ 22, 130, 50 },
		{ 22, 131, 50 },
		{ 22, 132, 50 },
		{ 22, 133, 50 },
		{ 22, 134, 50 },
		{ 22, 135, 50 },
		{ 22, 136, 50 },
		{ 22, 137, 50 },
		{ 22, 138, 50 },
		{ 22, 139, 50 },
		{ 22, 140, 50 },
		{ 22, 141, 50 },
		{ 22, 142, 50 },
		{ 22, 143, 50 },
		{ 22, 144, 50 },
		{ 22, 145, 50 },
		{ 22, 146, 50 },
		{ 22, 147, 50 },
		{ 22, 148, 50 },
		{ 22, 149, 50 },
		{ 22, 150, 50 },
		{ 22, 151, 50 },
		{ 22, 152, 50 },
		{ 22, 153, 50 },
		{ 22, 154, 50 },
		{ 22, 155, 50 },
		{ 22, 156, 50 },
		{ 22, 157, 50 },
		{ 22, 158, 50 },
		{ 22, 159, 50 },
		{ 22, 160, 50 },
		{ 22, 161, 50 },
		{ 22, 162, 50 },
		{ 22, 163, 50 },
		{ 22, 164, 50 },
		{ 22, 165, 50 },
		{ 22, 166, 50 },
		{ 22, 167, 50 },
		{ 22, 168, 50 },
		{ 23, 111, 50 },
		{ 23, 112, 50 },
		{ 23, 113, 50 },
		{ 23, 114, 50 },
		{ 23, 115, 50 },
		{ 23, 116, 50 },
		{ 23, 117, 50 },
		{ 23, 118, 50 },
		{ 23, 119, 50 },
		{ 23, 120, 50 },
		{ 23, 121, 50 },
		{ 23, 122, 50 },
		{ 23, 123, 50 },
		{ 23, 124, 50 },
		{ 23, 125, 50 },
		{ 23, 126, 50 },
		{ 23, 127, 50 },
		{ 23, 128, 50 },
		{ 23, 129, 50 },
		{ 23, 130, 50 },
		{ 23, 131, 50 },
		{ 23, 132, 50 },
		{ 23, 133, 50 },
		{ 23, 134, 50 },
		{ 23, 135, 50 },
		{ 23, 136, 50 },
		{ 23, 137, 50 },
		{ 23, 138, 50 },
		{ 23, 139, 50 },
		{ 23, 140, 50 },
		{ 23, 141, 50 },
		{ 23, 142, 50 },
		{ 23, 143, 50 },
		{ 23, 144, 50 },
		{ 23, 145, 50 },
		{ 23, 146, 50 },
		{ 23, 147, 50 },
		{ 23, 148, 50 },
		{ 23, 149, 50 },
		{ 23, 150, 50 },
		{ 23, 151, 50 },
		{ 23, 152, 50 },
		{ 23, 153, 50 },
		{ 23, 154, 50 },
		{ 23, 155, 50 },
		{ 23, 156, 50 },
		{ 23, 157, 50 },
		{ 23, 158, 50 },
		{ 23, 159, 50 },
		{ 23, 160, 50 },
		{ 23, 161, 50 },
		{ 23, 162, 50 },
		{ 23, 163, 50 },
		{ 23, 164, 50 },
		{ 23, 165, 50 },
		{ 23, 166, 50 },
		{ 23, 167, 50 },
		{ 23, 168, 50 },
		{ 23, 169, 50 },
		{ 24, 112, 50 },
		{ 24, 113, 50 },
		{ 24, 114, 50 },
		{ 24, 115, 50 },
		{ 24, 116, 50 },
		{ 24, 117, 50 },
		{ 24, 118, 50 },
		{ 24, 119, 50 },
		{ 24, 120, 50 },
		{ 24, 121, 50 },
		{ 24, 122, 50 },
		{ 24, 123, 50 },
		{ 24, 124, 50 },
		{ 24, 125, 50 },
		{ 24, 126, 50 },
		{ 24, 127, 50 },
		{ 24, 128, 50 },
		{ 24, 129, 50 },
		{ 24, 130, 50 },
		{ 24, 131, 50 },
		{ 24, 132, 50 },
		{ 24, 133, 50 },
		{ 24, 134, 50 },
		{ 24, 135, 50 },
		{ 24, 136, 50 },
		{ 24, 137, 50 },
		{ 24, 138, 50 },
		{ 24, 139, 50 },
		{ 24, 140, 50 },
		{ 24, 141, 50 },
		{ 24, 142, 50 },
		{ 24, 143, 50 },
		{ 24, 144, 50 },
		{ 24, 145, 50 },
		{ 24, 146, 50 },
		{ 24, 147, 50 },
		{ 24, 148, 50 },
		{ 24, 149, 50 },
		{ 24, 150, 50 },
		{ 24, 151, 50 },
		{ 24, 152, 50 },
		{ 24, 153, 50 },
		{ 24, 154, 50 },
		{ 24, 155, 50 },
		{ 24, 156, 50 },
		{ 24, 157, 50 },
		{ 24, 158, 50 },
		{ 24, 159, 50 },
		{ 24, 160, 50 },
		{ 24, 161, 50 },
		{ 24, 162, 50 },
		{ 24, 163, 50 },
		{ 24, 164, 50 },
		{ 24, 165, 50 },
		{ 24, 166, 50 },
		{ 24, 167, 50 },
		{ 24, 168, 50 },
		{ 24, 169, 50 },
		{ 25, 113, 50 },
		{ 25, 114, 50 },
		{ 25, 115, 50 },
		{ 25, 116, 50 },
		{ 25, 117, 50 },
		{ 25, 118, 50 },
		{ 25, 119, 50 },
		{ 25, 120, 50 },
		{ 25, 121, 50 },
		{ 25, 122, 50 },
		{ 25, 123, 50 },
		{ 25, 124, 50 },
		{ 25, 125, 50 },
		{ 25, 126, 50 },
		{ 25, 127, 50 },
		{ 25, 128, 50 },
		{ 25, 129, 50 },
		{ 25, 130, 50 },
		{ 25, 131, 50 },
		{ 25, 132, 50 },
		{ 25, 133, 50 },
		{ 25, 134, 50 },
		{ 25, 135, 50 },
		{ 25, 136, 50 },
		{ 25, 137, 50 },
		{ 25, 138, 50 },
		{ 25, 139, 50 },
		{ 25, 140, 50 },
		{ 25, 141, 50 },
		{ 25, 142, 50 },
		{ 25, 143, 50 },
		{ 25, 144, 50 },
		{ 25, 145, 50 },
		{ 25, 146, 50 },
		{ 25, 147, 50 },
		{ 25, 148, 50 },
		{ 25, 149, 50 },
		{ 25, 150, 50 },
		{ 25, 151, 50 },
		{ 25, 152, 50 },
		{ 25, 153, 50 },
		{ 25, 154, 50 },
		{ 25, 155, 50 },
		{ 25, 156, 50 },
		{ 25, 157, 50 },
		{ 25, 158, 50 },
		{ 25, 159, 50 },
		{ 25, 160, 50 },
		{ 25, 161, 50 },
		{ 25, 162, 50 },
		{ 25, 163, 50 },
		{ 25, 164, 50 },
		{ 25, 165, 50 },
		{ 25, 166, 50 },
		{ 25, 167, 50 },
		{ 25, 168, 50 },
		{ 25, 169, 50 },
		{ 26, 113, 50 },
		{ 26, 114, 50 },
		{ 26, 115, 50 },
		{ 26, 116, 50 },
		{ 26, 117, 50 },
		{ 26, 118, 50 },
		{ 26, 119, 50 },
		{ 26, 120, 50 },
		{ 26, 121, 50 },
		{ 26, 122, 50 },
		{ 26, 123, 50 },
		{ 26, 124, 50 },
		{ 26, 125, 50 },
		{ 26, 126, 50 },
		{ 26, 127, 50 },
		{ 26, 128, 50 },
		{ 26, 129, 50 },
		{ 26, 130, 50 },
		{ 26, 131, 50 },
		{ 26, 132, 50 },
		{ 26, 133, 50 },
		{ 26, 134, 50 },
		{ 26, 135, 50 },
		{ 26, 136, 50 },
		{ 26, 137, 50 },
		{ 26, 138, 50 },
		{ 26, 139, 50 },
		{ 26, 140, 50 },
		{ 26, 141, 50 },
		{ 26, 142, 50 },
		{ 26, 143, 50 },
		{ 26, 144, 50 },
		{ 26, 145, 50 },
		{ 26, 146, 50 },
		{ 26, 147, 50 },
		{ 26, 148, 50 },
		{ 26, 149, 50 },
		{ 26, 150, 50 },
		{ 26, 151, 50 },
		{ 26, 152, 50 },
		{ 26, 153, 50 },
		{ 26, 154, 50 },
		{ 26, 155, 50 },
		{ 26, 156, 50 },
		{ 26, 157, 50 },
		{ 26, 158, 50 },
		{ 26, 159, 50 },
		{ 26, 160, 50 },
		{ 26, 161, 50 },
		{ 26, 162, 50 },
		{ 26, 163, 50 },
		{ 26, 164, 50 },
		{ 26, 165, 50 },
		{ 26, 166, 50 },
		{ 26, 167, 50 },
		{ 26, 168, 50 },
		{ 26, 169, 50 },
		{ 26, 170, 50 },
		{ 27, 114, 50 },
		{ 27, 115, 50 },
		{ 27, 116, 50 },
		{ 27, 117, 50 },
		{ 27, 118, 50 },
		{ 27, 119, 50 },
		{ 27, 120, 50 },
		{ 27, 121, 50 },
		{ 27, 122, 50 },
		{ 27, 123, 50 },
		{ 27, 124, 50 },
		{ 27, 125, 50 },
		{ 27, 126, 50 },
		{ 27, 127, 50 },
		{ 27, 128, 50 },
		{ 27, 129, 50 },
		{ 27, 130, 50 },
		{ 27, 131, 50 },
		{ 27, 132, 50 },
		{ 27, 133, 50 },
		{ 27, 134, 50 },
		{ 27, 135, 50 },
		{ 27, 136, 50 },
		{ 27, 137, 50 },
		{ 27, 138, 50 },
		{ 27, 139, 50 },
		{ 27, 140, 50 },
		{ 27, 141, 50 },
		{ 27, 142, 50 },
		{ 27, 143, 50 },
		{ 27, 144, 50 },
		{ 27, 145, 50 },
		{ 27, 146, 50 },
		{ 27, 147, 50 },
		{ 27, 148, 50 },
		{ 27, 149, 50 },
		{ 27, 150, 50 },
		{ 27, 151, 50 },
		{ 27, 152, 50 },
		{ 27, 153, 50 },
		{ 27, 154, 50 },
		{ 27, 155, 50 },
		{ 27, 156, 50 },
		{ 27, 157, 50 },
		{ 27, 158, 50 },
		{ 27, 159, 50 },
		{ 27, 160, 50 },
		{ 27, 161, 50 },
		{ 27, 162, 50 },
		{ 27, 163, 50 },
		{ 27, 164, 50 },
		{ 27, 165, 50 },
		{ 27, 166, 50 },
		{ 27, 167, 50 },
		{ 27, 168, 50 },
		{ 27, 169, 50 },
		{ 27, 170, 50 },
		{ 28, 114, 50 },
		{ 28, 115, 50 },
		{ 28, 116, 50 },
		{ 28, 117, 50 },
		{ 28, 118, 50 },
		{ 28, 119, 50 },
		{ 28, 120, 50 },
		{ 28, 121, 50 },
		{ 28, 122, 50 },
		{ 28, 123, 50 },
		{ 28, 124, 50 },
		{ 28, 125, 50 },
		{ 28, 126, 50 },
		{ 28, 127, 50 },
		{ 28, 128, 50 },
		{ 28, 129, 50 },
		{ 28, 130, 50 },
		{ 28, 131, 50 },
		{ 28, 132, 50 },
		{ 28, 133, 50 },
		{ 28, 134, 50 },
		{ 28, 135, 50 },
		{ 28, 136, 50 },
		{ 28, 137, 50 },
		{ 28, 138, 50 },
		{ 28, 139, 50 },
		{ 28, 140, 50 },
		{ 28, 141, 50 },
		{ 28, 142, 50 },
		{ 28, 143, 50 },
		{ 28, 144, 50 },
		{ 28, 145, 50 },
		{ 28, 146, 50 },
		{ 28, 147, 50 },
		{ 28, 148, 50 },
		{ 28, 149, 50 },
		{ 28, 150, 50 },
		{ 28, 151, 50 },
		{ 28, 152, 50 },
		{ 28, 153, 50 },
		{ 28, 154, 50 },
		{ 28, 155, 50 },
		{ 28, 156, 50 },
		{ 28, 157, 50 },
		{ 28, 158, 50 },
		{ 28, 159, 50 },
		{ 28, 160, 50 },
		{ 28, 161, 50 },
		{ 28, 162, 50 },
		{ 28, 163, 50 },
		{ 28, 164, 50 },
		{ 28, 165, 50 },
		{ 28, 166, 50 },
		{ 28, 167, 50 },
		{ 28, 168, 50 },
		{ 28, 169, 50 },
		{ 28, 170, 50 },
		{ 28, 171, 50 },
		{ 29, 115, 50 },
		{ 29, 116, 50 },
		{ 29, 117, 50 },
		{ 29, 118, 50 },
		{ 29, 119, 50 },
		{ 29, 120, 50 },
		{ 29, 121, 50 },
		{ 29, 122, 50 },
		{ 29, 123, 50 },
		{ 29, 124, 50 },
		{ 29, 125, 50 },
		{ 29, 126, 50 },
		{ 29, 127, 50 },
		{ 29, 128, 50 },
		{ 29, 129, 50 },
		{ 29, 130, 50 },
		{ 29, 131, 50 },
		{ 29, 132, 50 },
		{ 29, 133, 50 },
		{ 29, 134, 50 },
		{ 29, 135, 50 },
		{ 29, 136, 50 },
		{ 29, 137, 50 },
		{ 29, 138, 50 },
		{ 29, 139, 50 },
		{ 29, 140, 50 },
		{ 29, 141, 50 },
		{ 29, 142, 50 },
		{ 29, 143, 50 },
		{ 29, 144, 50 },
		{ 29, 145, 50 },
		{ 29, 146, 50 },
		{ 29, 147, 50 },
		{ 29, 148, 50 },
		{ 29, 149, 50 },
		{ 29, 150, 50 },
		{ 29, 151, 50 },
		{ 29, 152, 50 },
		{ 29, 153, 50 },
		{ 29, 154, 50 },
		{ 29, 155, 50 },
		{ 29, 156, 50 },
		{ 29, 157, 50 },
		{ 29, 158, 50 },
		{ 29, 159, 50 },
		{ 29, 160, 50 },
		{ 29, 161, 50 },
		{ 29, 162, 50 },
		{ 29, 163, 50 },
		{ 29, 164, 50 },
		{ 29, 165, 50 },
		{ 29, 166, 50 },
		{ 29, 167, 50 },
		{ 29, 168, 50 },
		{ 29, 169, 50 },
		{ 29, 170, 50 },
		{ 29, 171, 50 },
		{ 30, 116, 50 },
		{ 30, 117, 50 },
		{ 30, 118, 50 },
		{ 30, 119, 50 },
		{ 30, 120, 50 },
		{ 30, 121, 50 },
		{ 30, 122, 50 },
		{ 30, 123, 50 },
		{ 30, 124, 50 },
		{ 30, 125, 50 },
		{ 30, 126, 50 },
		{ 30, 127, 50 },
		{ 30, 128, 50 },
		{ 30, 129, 50 },
		{ 30, 130, 50 },
		{ 30, 131, 50 },
		{ 30, 132, 50 },
		{ 30, 133, 50 },
		{ 30, 134, 50 },
		{ 30, 135, 50 },
		{ 30, 136, 50 },
		{ 30, 137, 50 },
		{ 30, 138, 50 },
		{ 30, 139, 50 },
		{ 30, 140, 50 },
		{ 30, 141, 50 },
		{ 30, 142, 50 },
		{ 30, 143, 50 },
		{ 30, 144, 50 },
		{ 30, 145, 50 },
		{ 30, 146, 50 },
		{ 30, 147, 50 },
		{ 30, 148, 50 },
		{ 30, 149, 50 },
		{ 30, 150, 50 },
		{ 30, 151, 50 },
		{ 30, 152, 50 },
		{ 30, 153, 50 },
		{ 30, 154, 50 },
		{ 30, 155, 50 },
		{ 30, 156, 50 },
		{ 30, 157, 50 },
		{ 30, 158, 50 },
		{ 30, 159, 50 },
		{ 30, 160, 50 },
		{ 30, 161, 50 },
		{ 30, 162, 50 },
		{ 30, 163, 50 },
		{ 30, 164, 50 },
		{ 30, 165, 50 },
		{ 30, 166, 50 },
		{ 30, 167, 50 },
		{ 30, 168, 50 },
		{ 30, 169, 50 },
		{ 30, 170, 50 },
		{ 30, 171, 50 },
		{ 31, 116, 50 },
		{ 31, 117, 50 },
		{ 31, 118, 50 },
		{ 31, 119, 50 },
		{ 31, 120, 50 },
		{ 31, 121, 50 },
		{ 31, 122, 50 },
		{ 31, 123, 50 },
		{ 31, 124, 50 },
		{ 31, 125, 50 },
		{ 31, 126, 50 },
		{ 31, 127, 50 },
		{ 31, 128, 50 },
		{ 31, 129, 50 },
		{ 31, 130, 50 },
		{ 31, 131, 50 },
		{ 31, 132, 50 },
		{ 31, 133, 50 },
		{ 31, 134, 50 },
		{ 31, 135, 50 },
		{ 31, 136, 50 },
		{ 31, 137, 50 },
		{ 31, 138, 50 },
		{ 31, 139, 50 },
		{ 31, 140, 50 },
		{ 31, 141, 50 },
		{ 31, 142, 50 },
		{ 31, 143, 50 },
		{ 31, 144, 50 },
		{ 31, 145, 50 },
		{ 31, 146, 50 },
		{ 31, 147, 50 },
		{ 31, 148, 50 },
		{ 31, 149, 50 },
		{ 31, 150, 50 },
		{ 31, 151, 50 },
		{ 31, 152, 50 },
		{ 31, 153, 50 },
		{ 31, 154, 50 },
		{ 31, 155, 50 },
		{ 31, 156, 50 },
		{ 31, 157, 50 },
		{ 31, 158, 50 },
		{ 31, 159, 50 },
		{ 31, 160, 50 },
		{ 31, 161, 50 },
		{ 31, 162, 50 },
		{ 31, 163, 50 },
		{ 31, 164, 50 },
		{ 31, 165, 50 },
		{ 31, 166, 50 },
		{ 31, 167, 50 },
		{ 31, 168, 50 },
		{ 31, 169, 50 },
		{ 31, 170, 50 },
		{ 31, 171, 50 },
		{ 31, 172, 50 },
		{ 32, 117, 50 },
		{ 32, 118, 50 },
		{ 32, 119, 50 },
		{ 32, 120, 50 },
		{ 32, 121, 50 },
		{ 32, 122, 50 },
		{ 32, 123, 50 },
		{ 32, 124, 50 },
		{ 32, 125, 50 },
		{ 32, 126, 50 },
		{ 32, 127, 50 },
		{ 32, 128, 50 },
		{ 32, 129, 50 },
		{ 32, 130, 50 },
		{ 32, 131, 50 },
		{ 32, 132, 50 },
		{ 32, 133, 50 },
		{ 32, 134, 50 },
		{ 32, 135, 50 },
		{ 32, 136, 50 },
		{ 32, 137, 50 },
		{ 32, 138, 50 },
		{ 32, 139, 50 },
		{ 32, 140, 50 },
		{ 32, 141, 50 },
		{ 32, 142, 50 },
		{ 32, 143, 50 },
		{ 32, 144, 50 },
		{ 32, 145, 50 },
		{ 32, 146, 50 },
		{ 32, 147, 50 },
		{ 32, 148, 50 },
		{ 32, 149, 50 },
		{ 32, 150, 50 },
		{ 32, 151, 50 },
		{ 32, 152, 50 },
		{ 32, 153, 50 },
		{ 32, 154, 50 },
		{ 32, 155, 50 },
		{ 32, 156, 50 },
		{ 32, 157, 50 },
		{ 32, 158, 50 },
		{ 32, 159, 50 },
		{ 32, 160, 50 },
		{ 32, 161, 50 },
		{ 32, 162, 50 },
		{ 32, 163, 50 },
		{ 32, 164, 50 },
		{ 32, 165, 50 },
		{ 32, 166, 50 },
		{ 32, 167, 50 },
		{ 32, 168, 50 },
		{ 32, 169, 50 },
		{ 32, 170, 50 },
		{ 32, 171, 50 },
		{ 32, 172, 50 },
		{ 33, 118, 50 },
		{ 33, 119, 50 },
		{ 33, 120, 50 },
		{ 33, 121, 50 },
		{ 33, 122, 50 },
		{ 33, 123, 50 },
		{ 33, 124, 50 },
		{ 33, 125, 50 },
		{ 33, 126, 50 },
		{ 33, 127, 50 },
		{ 33, 128, 50 },
		{ 33, 129, 50 },
		{ 33, 130, 50 },
		{ 33, 131, 50 },
		{ 33, 132, 50 },
		{ 33, 133, 50 },
		{ 33, 134, 50 },
		{ 33, 135, 50 },
		{ 33, 136, 50 },
		{ 33, 137, 50 },
		{ 33, 138, 50 },
		{ 33, 139, 50 },
		{ 33, 140, 50 },
		{ 33, 141, 50 },
		{ 33, 142, 50 },
		{ 33, 143, 50 },
		{ 33, 144, 50 },
		{ 33, 145, 50 },
		{ 33, 146, 50 },
		{ 33, 147, 50 },
		{ 33, 148, 50 },
		{ 33, 149, 50 },
		{ 33, 150, 50 },
		{ 33, 151, 50 },
		{ 33, 152, 50 },
		{ 33, 153, 50 },
		{ 33, 154, 50 },
		{ 33, 155, 50 },
		{ 33, 156, 50 },
		{ 33, 157, 50 },
		{ 33, 158, 50 },
		{ 33, 159, 50 },
		{ 33, 160, 50 },
		{ 33, 161, 50 },
		{ 33, 162, 50 },
		{ 33, 163, 50 },
		{ 33, 164, 50 },
		{ 33, 165, 50 },
		{ 33, 166, 50 },
		{ 33, 167, 50 },
		{ 33, 168, 50 },
		{ 33, 169, 50 },
		{ 33, 170, 50 },
		{ 33, 171, 50 },
		{ 33, 172, 50 },
		{ 34, 118, 50 },
		{ 34, 119, 50 },
		{ 34, 120, 50 },
		{ 34, 121, 50 },
		{ 34, 122, 50 },
		{ 34, 123, 50 },
		{ 34, 124, 50 },
		{ 34, 125, 50 },
		{ 34, 126, 50 },
		{ 34, 127, 50 },
		{ 34, 128, 50 },
		{ 34, 129, 50 },
		{ 34, 130, 50 },
		{ 34, 131, 50 },
		{ 34, 132, 50 },
		{ 34, 133, 50 },
		{ 34, 134, 50 },
		{ 34, 135, 50 },
		{ 34, 136, 50 },
		{ 34, 137, 50 },
		{ 34, 138, 50 },
		{ 34, 139, 50 },
		{ 34, 140, 50 },
		{ 34, 141, 50 },
		{ 34, 142, 50 },
		{ 34, 143, 50 },
		{ 34, 144, 50 },
		{ 34, 145, 50 },
		{ 34, 146, 50 },
		{ 34, 147, 50 },
		{ 34, 148, 50 },
		{ 34, 149, 50 },
		{ 34, 150, 50 },
		{ 34, 151, 50 },
		{ 34, 152, 50 },
		{ 34, 153, 50 },
		{ 34, 154, 50 },
		{ 34, 155, 50 },
		{ 34, 156, 50 },
		{ 34, 157, 50 },
		{ 34, 158, 50 },
		{ 34, 159, 50 },
		{ 34, 160, 50 },
		{ 34, 161, 50 },
		{ 34, 162, 50 },
		{ 34, 163, 50 },
		{ 34, 164, 50 },
		{ 34, 165, 50 },
		{ 34, 166, 50 },
		{ 34, 167, 50 },
		{ 34, 168, 50 },
		{ 34, 169, 50 },
		{ 34, 170, 50 },
		{ 34, 171, 50 },
		{ 34, 172, 50 },
		{ 34, 173, 50 },
		{ 35, 119, 50 },
		{ 35, 120, 50 },
		{ 35, 121, 50 },
		{ 35, 122, 50 },
		{ 35, 123, 50 },
		{ 35, 124, 50 },
		{ 35, 125, 50 },
		{ 35, 126, 50 },
		{ 35, 127, 50 },
		{ 35, 128, 50 },
		{ 35, 129, 50 },
		{ 35, 130, 50 },
		{ 35, 131, 50 },
		{ 35, 132, 50 },
		{ 35, 133, 50 },
		{ 35, 134, 50 },
		{ 35, 135, 50 },
		{ 35, 136, 50 },
		{ 35, 137, 50 },
		{ 35, 138, 50 },
		{ 35, 139, 50 },
		{ 35, 140, 50 },
		{ 35, 141, 50 },
		{ 35, 142, 50 },
		{ 35, 143, 50 },
		{ 35, 144, 50 },
		{ 35, 145, 50 },
		{ 35, 146, 50 },
		{ 35, 147, 50 },
		{ 35, 148, 50 },
		{ 35, 149, 50 },
		{ 35, 150, 50 },
		{ 35, 151, 50 },
		{ 35, 152, 50 },
		{ 35, 153, 50 },
		{ 35, 154, 50 },
		{ 35, 155, 50 },
		{ 35, 156, 50 },
		{ 35, 157, 50 },
		{ 35, 158, 50 },
		{ 35, 159, 50 },
		{ 35, 160, 50 },
		{ 35, 161, 50 },
		{ 35, 162, 50 },
		{ 35, 163, 50 },
		{ 35, 164, 50 },
		{ 35, 165, 50 },
		{ 35, 166, 50 },
		{ 35, 167, 50 },
		{ 35, 168, 50 },
		{ 35, 169, 50 },
		{ 35, 170, 50 },
		{ 35, 171, 50 },
		{ 35, 172, 50 },
		{ 35, 173, 50 },
		{ 36, 119, 50 },
		{ 36, 120, 50 },
		{ 36, 121, 50 },
		{ 36, 122, 50 },
		{ 36, 123, 50 },
		{ 36, 124, 50 },
		{ 36, 125, 50 },
		{ 36, 126, 50 },
		{ 36, 127, 50 },
		{ 36, 128, 50 },
		{ 36, 129, 50 },
		{ 36, 130, 50 },
		{ 36, 131, 50 },
		{ 36, 132, 50 },
		{ 36, 133, 50 },
		{ 36, 134, 50 },
		{ 36, 135, 50 },
		{ 36, 136, 50 },
		{ 36, 137, 50 },
		{ 36, 138, 50 },
		{ 36, 139, 50 },
		{ 36, 140, 50 },
		{ 36, 141, 50 },
		{ 36, 142, 50 },
		{ 36, 143, 50 },
		{ 36, 144, 50 },
		{ 36, 145, 50 },
		{ 36, 146, 50 },
		{ 36, 147, 50 },
		{ 36, 148, 50 },
		{ 36, 149, 50 },
		{ 36, 150, 50 },
		{ 36, 151, 50 },
		{ 36, 152, 50 },
		{ 36, 153, 50 },
		{ 36, 154, 50 },
		{ 36, 155, 50 },
		{ 36, 156, 50 },
		{ 36, 157, 50 },
		{ 36, 158, 50 },
		{ 36, 159, 50 },
		{ 36, 160, 50 },
		{ 36, 161, 50 },
		{ 36, 162, 50 },
		{ 36, 163, 50 },
		{ 36, 164, 50 },
		{ 36, 165, 50 },
		{ 36, 166, 50 },
		{ 36, 167, 50 },
		{ 36, 168, 50 },
		{ 36, 169, 50 },
		{ 36, 170, 50 },
		{ 36, 171, 50 },
		{ 36, 172, 50 },
		{ 36, 173, 50 },
		{ 36, 174, 50 },
		{ 37, 120, 50 },
		{ 37, 121, 50 },
		{ 37, 122, 50 },
		{ 37, 123, 50 },
		{ 37, 124, 50 },
		{ 37, 125, 50 },
		{ 37, 126, 50 },
		{ 37, 127, 50 },
		{ 37, 128, 50 },
		{ 37, 129, 50 },
		{ 37, 130, 50 },
		{ 37, 131, 50 },
		{ 37, 132, 50 },
		{ 37, 133, 50 },
		{ 37, 134, 50 },
		{ 37, 135, 50 },
		{ 37, 136, 50 },
		{ 37, 137, 50 },
		{ 37, 138, 50 },
		{ 37, 139, 50 },
		{ 37, 140, 50 },
		{ 37, 141, 50 },
		{ 37, 142, 50 },
		{ 37, 143, 50 },
		{ 37, 144, 50 },
		{ 37, 145, 50 },
		{ 37, 146, 50 },
		{ 37, 147, 50 },
		{ 37, 148, 50 },
		{ 37, 149, 50 },
		{ 37, 150, 50 },
		{ 37, 151, 50 },
		{ 37, 152, 50 },
		{ 37, 153, 50 },
		{ 37, 154, 50 },
		{ 37, 155, 50 },
		{ 37, 156, 50 },
		{ 37, 157, 50 },
		{ 37, 158, 50 },
		{ 37, 159, 50 },
		{ 37, 160, 50 },
		{ 37, 161, 50 },
		{ 37, 162, 50 },
		{ 37, 163, 50 },
		{ 37, 164, 50 },
		{ 37, 165, 50 },
		{ 37, 166, 50 },
		{ 37, 167, 50 },
		{ 37, 168, 50 },
		{ 37, 169, 50 },
		{ 37, 170, 50 },
		{ 37, 171, 50 },
		{ 37, 172, 50 },
		{ 37, 173, 50 },
		{ 37, 174, 50 },
		{ 38, 121, 50 },
		{ 38, 122, 50 },
		{ 38, 123, 50 },
		{ 38, 124, 50 },
		{ 38, 125, 50 },
		{ 38, 126, 50 },
		{ 38, 127, 50 },
		{ 38, 128, 50 },
		{ 38, 129, 50 },
		{ 38, 130, 50 },
		{ 38, 131, 50 },
		{ 38, 132, 50 },
		{ 38, 133, 50 },
		{ 38, 134, 50 },
		{ 38, 135, 50 },
		{ 38, 136, 50 },
		{ 38, 137, 50 },
		{ 38, 138, 50 },
		{ 38, 139, 50 },
		{ 38, 140, 50 },
		{ 38, 141, 50 },
		{ 38, 142, 50 },
		{ 38, 143, 50 },
		{ 38, 144, 50 },
		{ 38, 145, 50 },
		{ 38, 146, 50 },
		{ 38, 147, 50 },
		{ 38, 148, 50 },
		{ 38, 149, 50 },
		{ 38, 150, 50 },
		{ 38, 151, 50 },
		{ 38, 152, 50 },
		{ 38, 153, 50 },
		{ 38, 154, 50 },
		{ 38, 155, 50 },
		{ 38, 156, 50 },
		{ 38, 157, 50 },
		{ 38, 158, 50 },
		{ 38, 159, 50 },
		{ 38, 160, 50 },
		{ 38, 161, 50 },
		{ 38, 162, 50 },
		{ 38, 163, 50 },
		{ 38, 164, 50 },
		{ 38, 165, 50 },
		{ 38, 166, 50 },
		{ 38, 167, 50 },
		{ 38, 168, 50 },
		{ 38, 169, 50 },
		{ 38, 170, 50 },
		{ 38, 171, 50 },
		{ 38, 172, 50 },
		{ 38, 173, 50 },
		{ 38, 174, 50 },
		{ 39, 121, 50 },
		{ 39, 122, 50 },
		{ 39, 123, 50 },
		{ 39, 124, 50 },
		{ 39, 125, 50 },
		{ 39, 126, 50 },
		{ 39, 127, 50 },
		{ 39, 128, 50 },
		{ 39, 129, 50 },
		{ 39, 130, 50 },
		{ 39, 131, 50 },
		{ 39, 132, 50 },
		{ 39, 133, 50 },
		{ 39, 134, 50 },
		{ 39, 135, 50 },
		{ 39, 136, 50 },
		{ 39, 137, 50 },
		{ 39, 138, 50 },
		{ 39, 139, 50 },
		{ 39, 140, 50 },
		{ 39, 141, 50 },
		{ 39, 142, 50 },
		{ 39, 143, 50 },
		{ 39, 144, 50 },
		{ 39, 145, 50 },
		{ 39, 146, 50 },
		{ 39, 147, 50 },
		{ 39, 148, 50 },
		{ 39, 149, 50 },
		{ 39, 150, 50 },
		{ 39, 151, 50 },
		{ 39, 152, 50 },
		{ 39, 153, 50 },
		{ 39, 154, 50 },
		{ 39, 155, 50 },
		{ 39, 156, 50 },
		{ 39, 157, 50 },
		{ 39, 158, 50 },
		{ 39, 159, 50 },
		{ 39, 160, 50 },
		{ 39, 161, 50 },
		{ 39, 162, 50 },
		{ 39, 163, 50 },
		{ 39, 164, 50 },
		{ 39, 165, 50 },
		{ 39, 166, 50 },
		{ 39, 167, 50 },
		{ 39, 168, 50 },
		{ 39, 169, 50 },
		{ 39, 170, 50 },
		{ 39, 171, 50 },
		{ 39, 172, 50 },
		{ 39, 173, 50 },
		{ 39, 174, 50 },
		{ 39, 175, 50 },
		{ 40, 122, 50 },
		{ 40, 123, 50 },
		{ 40, 124, 50 },
		{ 40, 125, 50 },
		{ 40, 126, 50 },
		{ 40, 127, 50 },
		{ 40, 128, 50 },
		{ 40, 129, 50 },
		{ 40, 130, 50 },
		{ 40, 131, 50 },
		{ 40, 132, 50 },
		{ 40, 133, 50 },
		{ 40, 134, 50 },
		{ 40, 135, 50 },
		{ 40, 136, 50 },
		{ 40, 137, 50 },
		{ 40, 138, 50 },
		{ 40, 139, 50 },
		{ 40, 140, 50 },
		{ 40, 141, 50 },
		{ 40, 142, 50 },
		{ 40, 143, 50 },
		{ 40, 144, 50 },
		{ 40, 145, 50 },
		{ 40, 146, 50 },
		{ 40, 147, 50 },
		{ 40, 148, 50 },
		{ 40, 149, 50 },
		{ 40, 150, 50 },
		{ 40, 151, 50 },
		{ 40, 152, 50 },
		{ 40, 153, 50 },
		{ 40, 154, 50 },
		{ 40, 155, 50 },
		{ 40, 156, 50 },
		{ 40, 157, 50 },
		{ 40, 158, 50 },
		{ 40, 159, 50 },
		{ 40, 160, 50 },
		{ 40, 161, 50 },
		{ 40, 162, 50 },
		{ 40, 163, 50 },
		{ 40, 164, 50 },
		{ 40, 165, 50 },
		{ 40, 166, 50 },
		{ 40, 167, 50 },
		{ 40, 168, 50 },
		{ 40, 169, 50 },
		{ 40, 170, 50 },
		{ 40, 171, 50 },
		{ 40, 172, 50 },
		{ 40, 173, 50 },
		{ 40, 174, 50 },
		{ 40, 175, 50 },
		{ 41, 123, 50 },
		{ 41, 124, 50 },
		{ 41, 125, 50 },
		{ 41, 126, 50 },
		{ 41, 127, 50 },
		{ 41, 128, 50 },
		{ 41, 129, 50 },
		{ 41, 130, 50 },
		{ 41, 131, 50 },
		{ 41, 132, 50 },
		{ 41, 133, 50 },
		{ 41, 134, 50 },
		{ 41, 135, 50 },
		{ 41, 136, 50 },
		{ 41, 137, 50 },
		{ 41, 138, 50 },
		{ 41, 139, 50 },
		{ 41, 140, 50 },
		{ 41, 141, 50 },
		{ 41, 142, 50 },
		{ 41, 143, 50 },
		{ 41, 144, 50 },
		{ 41, 145, 50 },
		{ 41, 146, 50 },
		{ 41, 147, 50 },
		{ 41, 148, 50 },
		{ 41, 149, 50 },
		{ 41, 150, 50 },
		{ 41, 151, 50 },
		{ 41, 152, 50 },
		{ 41, 153, 50 },
		{ 41, 154, 50 },
		{ 41, 155, 50 },
		{ 41, 156, 50 },
		{ 41, 157, 50 },
		{ 41, 158, 50 },
		{ 41, 159, 50 },
		{ 41, 160, 50 },
		{ 41, 161, 50 },
		{ 41, 162, 50 },
		{ 41, 163, 50 },
		{ 41, 164, 50 },
		{ 41, 165, 50 },
		{ 41, 166, 50 },
		{ 41, 167, 50 },
		{ 41, 168, 50 },
		{ 41, 169, 50 },
		{ 41, 170, 50 },
		{ 41, 171, 50 },
		{ 41, 172, 50 },
		{ 41, 173, 50 },
		{ 41, 174, 50 },
		{ 41, 175, 50 },
		{ 42, 123, 50 },
		{ 42, 124, 50 },
		{ 42, 125, 50 },
		{ 42, 126, 50 },
		{ 42, 127, 50 },
		{ 42, 128, 50 },
		{ 42, 129, 50 },
		{ 42, 130, 50 },
		{ 42, 131, 50 },
		{ 42, 132, 50 },
		{ 42, 133, 50 },
		{ 42, 134, 50 },
		{ 42, 135, 50 },
		{ 42, 136, 50 },
		{ 42, 137, 50 },
		{ 42, 138, 50 },
		{ 42, 139, 50 },
		{ 42, 140, 50 },
		{ 42, 141, 50 },
		{ 42, 142, 50 },
		{ 42, 143, 50 },
		{ 42, 144, 50 },
		{ 42, 145, 50 },
		{ 42, 146, 50 },
		{ 42, 147, 50 },
		{ 42, 148, 50 },
		{ 42, 149, 50 },
		{ 42, 150, 50 },
		{ 42, 151, 50 },
		{ 42, 152, 50 },
		{ 42, 153, 50 },
		{ 42, 154, 50 },
		{ 42, 155, 50 },
		{ 42, 156, 50 },
		{ 42, 157, 50 },
		{ 42, 158, 50 },
		{ 42, 159, 50 },
		{ 42, 160, 50 },
		{ 42, 161, 50 },
		{ 42, 162, 50 },
		{ 42, 163, 50 },
		{ 42, 164, 50 },
		{ 42, 165, 50 },
		{ 42, 166, 50 },
		{ 42, 167, 50 },
		{ 42, 168, 50 },
		{ 42, 169, 50 },
		{ 42, 170, 50 },
		{ 42, 171, 50 },
		{ 42, 172, 50 },
		{ 42, 173, 50 },
		{ 42, 174, 50 },
		{ 42, 175, 50 },
		{ 42, 176, 50 },
		{ 43, 124, 50 },
		{ 43, 125, 50 },
		{ 43, 126, 50 },
		{ 43, 127, 50 },
		{ 43, 128, 50 },
		{ 43, 129, 50 },
		{ 43, 130, 50 },
		{ 43, 131, 50 },
		{ 43, 132, 50 },
		{ 43, 133, 50 },
		{ 43, 134, 50 },
		{ 43, 135, 50 },
		{ 43, 136, 50 },
		{ 43, 137, 50 },
		{ 43, 138, 50 },
		{ 43, 139, 50 },
		{ 43, 140, 50 },
		{ 43, 141, 50 },
		{ 43, 142, 50 },
		{ 43, 143, 50 },
		{ 43, 144, 50 },
		{ 43, 145, 50 },
		{ 43, 146, 50 },
		{ 43, 147, 50 },
		{ 43, 148, 50 },
		{ 43, 149, 50 },
		{ 43, 150, 50 },
		{ 43, 151, 50 },
		{ 43, 152, 50 },
		{ 43, 153, 50 },
		{ 43, 154, 50 },
		{ 43, 155, 50 },
		{ 43, 156, 50 },
		{ 43, 157, 50 },
		{ 43, 158, 50 },
		{ 43, 159, 50 },
		{ 43, 160, 50 },
		{ 43, 161, 50 },
		{ 43, 162, 50 },
		{ 43, 163, 50 },
		{ 43, 164, 50 },
		{ 43, 165, 50 },
		{ 43, 166, 50 },
		{ 43, 167, 50 },
		{ 43, 168, 50 },
		{ 43, 169, 50 },
		{ 43, 170, 50 },
		{ 43, 171, 50 },
		{ 43, 172, 50 },
		{ 43, 173, 50 },
		{ 43, 174, 50 },
		{ 43, 175, 50 },
		{ 43, 176, 50 },
		{ 44, 124, 50 },
		{ 44, 125, 50 },
		{ 44, 126, 50 },
		{ 44, 127, 50 },
		{ 44, 128, 50 },
		{ 44, 129, 50 },
		{ 44, 130, 50 },
		{ 44, 131, 50 },
		{ 44, 132, 50 },
		{ 44, 133, 50 },
		{ 44, 134, 50 },
		{ 44, 135, 50 },
		{ 44, 136, 50 },
		{ 44, 137, 50 },
		{ 44, 138, 50 },
		{ 44, 139, 50 },
		{ 44, 140, 50 },
		{ 44, 141, 50 },
		{ 44, 142, 50 },
		{ 44, 143, 50 },
		{ 44, 144, 50 },
		{ 44, 145, 50 },
		{ 44, 146, 50 },
		{ 44, 147, 50 },
		{ 44, 148, 50 },
		{ 44, 149, 50 },
		{ 44, 150, 50 },
		{ 44, 151, 50 },
		{ 44, 152, 50 },
		{ 44, 153, 50 },
		{ 44, 154, 50 },
		{ 44, 155, 50 },
		{ 44, 156, 50 },
		{ 44, 157, 50 },
		{ 44, 158, 50 },
		{ 44, 159, 50 },
		{ 44, 160, 50 },
		{ 44, 161, 50 },
		{ 44, 162, 50 },
		{ 44, 163, 50 },
		{ 44, 164, 50 },
		{ 44, 165, 50 },
		{ 44, 166, 50 },
		{ 44, 167, 50 },
		{ 44, 168, 50 },
		{ 44, 169, 50 },
		{ 44, 170, 50 },
		{ 44, 171, 50 },
		{ 44, 172, 50 },
		{ 44, 173, 50 },
		{ 44, 174, 50 },
		{ 44, 175, 50 },
		{ 44, 176, 50 },
		{ 44, 177, 50 },
		{ 45, 125, 50 },
		{ 45, 126, 50 },
		{ 45, 127, 50 },
		{ 45, 128, 50 },
		{ 45, 129, 50 },
		{ 45, 130, 50 },
		{ 45, 131, 50 },
		{ 45, 132, 50 },
		{ 45, 133, 50 },
		{ 45, 134, 50 },
		{ 45, 135, 50 },
		{ 45, 136, 50 },
		{ 45, 137, 50 },
		{ 45, 138, 50 },
		{ 45, 139, 50 },
		{ 45, 140, 50 },
		{ 45, 141, 50 },
		{ 45, 142, 50 },
		{ 45, 143, 50 },
		{ 45, 144, 50 },
		{ 45, 145, 50 },
		{ 45, 146, 50 },
		{ 45, 147, 50 },
		{ 45, 148, 50 },
		{ 45, 149, 50 },
		{ 45, 150, 50 },
		{ 45, 151, 50 },
		{ 45, 152, 50 },
		{ 45, 153, 50 },
		{ 45, 154, 50 },
		{ 45, 155, 50 },
		{ 45, 156, 50 },
		{ 45, 157, 50 },
		{ 45, 158, 50 },
		{ 45, 159, 50 },
		{ 45, 160, 50 },
		{ 45, 161, 50 },
		{ 45, 162, 50 },
		{ 45, 163, 50 },
		{ 45, 164, 50 },
		{ 45, 165, 50 },
		{ 45, 166, 50 },
		{ 45, 167, 50 },
		{ 45, 168, 50 },
		{ 45, 169, 50 },
		{ 45, 170, 50 },
		{ 45, 171, 50 },
		{ 45, 172, 50 },
		{ 45, 173, 50 },
		{ 45, 174, 50 },
		{ 45, 175, 50 },
		{ 45, 176, 50 },
		{ 45, 177, 50 },
		{ 46, 126, 50 },
		{ 46, 127, 50 },
		{ 46, 128, 50 },
		{ 46, 129, 50 },
		{ 46, 130, 50 },
		{ 46, 131, 50 },
		{ 46, 132, 50 },
		{ 46, 133, 50 },
		{ 46, 134, 50 },
		{ 46, 135, 50 },
		{ 46, 136, 50 },
		{ 46, 137, 50 },
		{ 46, 138, 50 },
		{ 46, 139, 50 },
		{ 46, 140, 50 },
		{ 46, 141, 50 },
		{ 46, 142, 50 },
		{ 46, 143, 50 },
		{ 46, 144, 50 },
		{ 46, 145, 50 },
		{ 46, 146, 50 },
		{ 46, 147, 50 },
		{ 46, 148, 50 },
		{ 46, 149, 50 },
		{ 46, 150, 50 },
		{ 46, 151, 50 },
		{ 46, 152, 50 },
		{ 46, 153, 50 },
		{ 46, 154, 50 },
		{ 46, 155, 50 },
		{ 46, 156, 50 },
		{ 46, 157, 50 },
		{ 46, 158, 50 },
		{ 46, 159, 50 },
		{ 46, 160, 50 },
		{ 46, 161, 50 },
		{ 46, 162, 50 },
		{ 46, 163, 50 },
		{ 46, 164, 50 },
		{ 46, 165, 50 },
		{ 46, 166, 50 },
		{ 46, 167, 50 },
		{ 46, 168, 50 },
		{ 46, 169, 50 },
		{ 46, 170, 50 },
		{ 46, 171, 50 },
		{ 46, 172, 50 },
		{ 46, 173, 50 },
		{ 46, 174, 50 },
		{ 46, 175, 50 },
		{ 46, 176, 50 },
		{ 46, 177, 50 },
		{ 47, 126, 50 },
		{ 47, 127, 50 },
		{ 47, 128, 50 },
		{ 47, 129, 50 },
		{ 47, 130, 50 },
		{ 47, 131, 50 },
		{ 47, 132, 50 },
		{ 47, 133, 50 },
		{ 47, 134, 50 },
		{ 47, 135, 50 },
		{ 47, 136, 50 },
		{ 47, 137, 50 },
		{ 47, 138, 50 },
		{ 47, 139, 50 },
		{ 47, 140, 50 },
		{ 47, 141, 50 },
		{ 47, 142, 50 },
		{ 47, 143, 50 },
		{ 47, 144, 50 },
		{ 47, 145, 50 },
		{ 47, 146, 50 },
		{ 47, 147, 50 },
		{ 47, 148, 50 },
		{ 47, 149, 50 },
		{ 47, 150, 50 },
		{ 47, 151, 50 },
		{ 47, 152, 50 },
		{ 47, 153, 50 },
		{ 47, 154, 50 },
		{ 47, 155, 50 },
		{ 47, 156, 50 },
		{ 47, 157, 50 },
		{ 47, 158, 50 },
		{ 47, 159, 50 },
		{ 47, 160, 50 },
		{ 47, 161, 50 },
		{ 47, 162, 50 },
		{ 47, 163, 50 },
		{ 47, 164, 50 },
		{ 47, 165, 50 },
		{ 47, 166, 50 },
		{ 47, 167, 50 },
		{ 47, 168, 50 },
		{ 47, 169, 50 },
		{ 47, 170, 50 },
		{ 47, 171, 50 },
		{ 47, 172, 50 },
		{ 47, 173, 50 },
		{ 47, 174, 50 },
		{ 47, 175, 50 },
		{ 47, 176, 50 },
		{ 47, 177, 50 },
		{ 47, 178, 50 },
		{ 48, 127, 50 },
		{ 48, 128, 50 },
		{ 48, 129, 50 },
		{ 48, 130, 50 },
		{ 48, 131, 50 },
		{ 48, 132, 50 },
		{ 48, 133, 50 },
		{ 48, 134, 50 },
		{ 48, 135, 50 },
		{ 48, 136, 50 },
		{ 48, 137, 50 },
		{ 48, 138, 50 },
		{ 48, 139, 50 },
		{ 48, 140, 50 },
		{ 48, 141, 50 },
		{ 48, 142, 50 },
		{ 48, 143, 50 },
		{ 48, 144, 50 },
		{ 48, 145, 50 },
		{ 48, 146, 50 },
		{ 48, 147, 50 },
		{ 48, 148, 50 },
		{ 48, 149, 50 },
		{ 48, 150, 50 },
		{ 48, 151, 50 },
		{ 48, 152, 50 },
		{ 48, 153, 50 },
		{ 48, 154, 50 },
		{ 48, 155, 50 },
		{ 48, 156, 50 },
		{ 48, 157, 50 },
		{ 48, 158, 50 },
		{ 48, 159, 50 },
		{ 48, 160, 50 },
		{ 48, 161, 50 },
		{ 48, 162, 50 },
		{ 48, 163, 50 },
		{ 48, 164, 50 },
		{ 48, 165, 50 },
		{ 48, 166, 50 },
		{ 48, 167, 50 },
		{ 48, 168, 50 },
		{ 48, 169, 50 },
		{ 48, 170, 50 },
		{ 48, 171, 50 },
		{ 48, 172, 50 },
		{ 48, 173, 50 },
		{ 48, 174, 50 },
		{ 48, 175, 50 },
		{ 48, 176, 50 },
		{ 48, 177, 50 },
		{ 48, 178, 50 },
		{ 49, 128, 50 },
		{ 49, 129, 50 },
		{ 49, 130, 50 },
		{ 49, 131, 50 },
		{ 49, 132, 50 },
		{ 49, 133, 50 },
		{ 49, 134, 50 },
		{ 49, 135, 50 },
		{ 49, 136, 50 },
		{ 49, 137, 50 },
		{ 49, 138, 50 },
		{ 49, 139, 50 },
		{ 49, 140, 50 },
		{ 49, 141, 50 },
		{ 49, 142, 50 },
		{ 49, 143, 50 },
		{ 49, 144, 50 },
		{ 49, 145, 50 },
		{ 49, 146, 50 },
		{ 49, 147, 50 },
		{ 49, 148, 50 },
		{ 49, 149, 50 },
		{ 49, 150, 50 },
		{ 49, 151, 50 },
		{ 49, 152, 50 },
		{ 49, 153, 50 },
		{ 49, 154, 50 },
		{ 49, 155, 50 },
		{ 49, 156, 50 },
		{ 49, 157, 50 },
		{ 49, 158, 50 },
		{ 49, 159, 50 },
		{ 49, 160, 50 },
		{ 49, 161, 50 },
		{ 49, 162, 50 },
		{ 49, 163, 50 },
		{ 49, 164, 50 },
		{ 49, 165, 50 },
		{ 49, 166, 50 },
		{ 49, 167, 50 },
		{ 49, 168, 50 },
		{ 49, 169, 50 },
		{ 49, 170, 50 },
		{ 49, 171, 50 },
		{ 49, 172, 50 },
		{ 49, 173, 50 },
		{ 49, 174, 50 },
		{ 49, 175, 50 },
		{ 49, 176, 50 },
		{ 49, 177, 50 },
		{ 49, 178, 50 },
		{ 50, 128, 50 },
		{ 50, 129, 50 },
		{ 50, 130, 50 },
		{ 50, 131, 50 },
		{ 50, 132, 50 },
		{ 50, 133, 50 },
		{ 50, 134, 50 },
		{ 50, 135, 50 },
		{ 50, 136, 50 },
		{ 50, 137, 50 },
		{ 50, 138, 50 },
		{ 50, 139, 50 },
		{ 50, 140, 50 },
		{ 50, 141, 50 },
		{ 50, 142, 50 },
		{ 50, 143, 50 },
		{ 50, 144, 50 },
		{ 50, 145, 50 },
		{ 50, 146, 50 },
		{ 50, 147, 50 },
		{ 50, 148, 50 },
		{ 50, 149, 50 },
		{ 50, 150, 50 },
		{ 50, 151, 50 },
		{ 50, 152, 50 },
		{ 50, 153, 50 },
		{ 50, 154, 50 },
		{ 50, 155, 50 },
		{ 50, 156, 50 },
		{ 50, 157, 50 },
		{ 50, 158, 50 },
		{ 50, 159, 50 },
		{ 50, 160, 50 },
		{ 50, 161, 50 },
		{ 50, 162, 50 },
		{ 50, 163, 50 },
		{ 50, 164, 50 },
		{ 50, 165, 50 },
		{ 50, 166, 50 },
		{ 50, 167, 50 },
		{ 50, 168, 50 },
		{ 50, 169, 50 },
		{ 50, 170, 50 },
		{ 50, 171, 50 },
		{ 50, 172, 50 },
		{ 50, 173, 50 },
		{ 50, 174, 50 },
		{ 50, 175, 50 },
		{ 50, 176, 50 },
		{ 50, 177, 50 },
		{ 50, 178, 50 },
		{ 50, 179, 50 },
		{ 51, 129, 50 },
		{ 51, 130, 50 },
		{ 51, 131, 50 },
		{ 51, 132, 50 },
		{ 51, 133, 50 },
		{ 51, 134, 50 },
		{ 51, 135, 50 },
		{ 51, 136, 50 },
		{ 51, 137, 50 },
		{ 51, 138, 50 },
		{ 51, 139, 50 },
		{ 51, 140, 50 },
		{ 51, 141, 50 },
		{ 51, 142, 50 },
		{ 51, 143, 50 },
		{ 51, 144, 50 },
		{ 51, 145, 50 },
		{ 51, 146, 50 },
		{ 51, 147, 50 },
		{ 51, 148, 50 },
		{ 51, 149, 50 },
		{ 51, 150, 50 },
		{ 51, 151, 50 },
		{ 51, 152, 50 },
		{ 51, 153, 50 },
		{ 51, 154, 50 },
		{ 51, 155, 50 },
		{ 51, 156, 50 },
		{ 51, 157, 50 },
		{ 51, 158, 50 },
		{ 51, 159, 50 },
		{ 51, 160, 50 },
		{ 51, 161, 50 },
		{ 51, 162, 50 },
		{ 51, 163, 50 },
		{ 51, 164, 50 },
		{ 51, 165, 50 },
		{ 51, 166, 50 },
		{ 51, 167, 50 },
		{ 51, 168, 50 },
		{ 51, 169, 50 },
		{ 51, 170, 50 },
		{ 51, 171, 50 },
		{ 51, 172, 50 },
		{ 51, 173, 50 },
		{ 51, 174, 50 },
		{ 51, 175, 50 },
		{ 51, 176, 50 },
		{ 51, 177, 50 },
		{ 51, 178, 50 },
		{ 51, 179, 50 },
		{ 52, 129, 50 },
		{ 52, 130, 50 },
		{ 52, 131, 50 },
		{ 52, 132, 50 },
		{ 52, 133, 50 },
		{ 52, 134, 50 },
		{ 52, 135, 50 },
		{ 52, 136, 50 },
		{ 52, 137, 50 },
		{ 52, 138, 50 },
		{ 52, 139, 50 },
		{ 52, 140, 50 },
		{ 52, 141, 50 },
		{ 52, 142, 50 },
		{ 52, 143, 50 },
		{ 52, 144, 50 },
		{ 52, 145, 50 },
		{ 52, 146, 50 },
		{ 52, 147, 50 },
		{ 52, 148, 50 },
		{ 52, 149, 50 },
		{ 52, 150, 50 },
		{ 52, 151, 50 },
		{ 52, 152, 50 },
		{ 52, 153, 50 },
		{ 52, 154, 50 },
		{ 52, 155, 50 },
		{ 52, 156, 50 },
		{ 52, 157, 50 },
		{ 52, 158, 50 },
		{ 52, 159, 50 },
		{ 52, 160, 50 },
		{ 52, 161, 50 },
		{ 52, 162, 50 },
		{ 52, 163, 50 },
		{ 52, 164, 50 },
		{ 52, 165, 50 },
		{ 52, 166, 50 },
		{ 52, 167, 50 },
		{ 52, 168, 50 },
		{ 52, 169, 50 },
		{ 52, 170, 50 },
		{ 52, 171, 50 },
		{ 52, 172, 50 },
		{ 52, 173, 50 },
		{ 52, 174, 50 },
		{ 52, 175, 50 },
		{ 52, 176, 50 },
		{ 52, 177, 50 },
		{ 52, 178, 50 },
		{ 52, 179, 50 },
		{ 52, 180, 50 },
		{ 53, 130, 50 },
		{ 53, 131, 50 },
		{ 53, 132, 50 },
		{ 53, 133, 50 },
		{ 53, 134, 50 },
		{ 53, 135, 50 },
		{ 53, 136, 50 },
		{ 53, 137, 50 },
		{ 53, 138, 50 },
		{ 53, 139, 50 },
		{ 53, 140, 50 },
		{ 53, 141, 50 },
		{ 53, 142, 50 },
		{ 53, 143, 50 },
		{ 53, 144, 50 },
		{ 53, 145, 50 },
		{ 53, 146, 50 },
		{ 53, 147, 50 },
		{ 53, 148, 50 },
		{ 53, 149, 50 },
		{ 53, 150, 50 },
		{ 53, 151, 50 },
		{ 53, 152, 50 },
		{ 53, 153, 50 },
		{ 53, 154, 50 },
		{ 53, 155, 50 },
		{ 53, 156, 50 },
		{ 53, 157, 50 },
		{ 53, 158, 50 },
		{ 53, 159, 50 },
		{ 53, 160, 50 },
		{ 53, 161, 50 },
		{ 53, 162, 50 },
		{ 53, 163, 50 },
		{ 53, 164, 50 },
		{ 53, 165, 50 },
		{ 53, 166, 50 },
		{ 53, 167, 50 },
		{ 53, 168, 50 },
		{ 53, 169, 50 },
		{ 53, 170, 50 },
		{ 53, 171, 50 },
		{ 53, 172, 50 },
		{ 53, 173, 50 },
		{ 53, 174, 50 },
		{ 53, 175, 50 },
		{ 53, 176, 50 },
		{ 53, 177, 50 },
		{ 53, 178, 50 },
		{ 53, 179, 50 },
		{ 53, 180, 50 },
		{ 54, 131, 50 },
		{ 54, 132, 50 },
		{ 54, 133, 50 },
		{ 54, 134, 50 },
		{ 54, 135, 50 },
		{ 54, 136, 50 },
		{ 54, 137, 50 },
		{ 54, 138, 50 },
		{ 54, 139, 50 },
		{ 54, 140, 50 },
		{ 54, 141, 50 },
		{ 54, 142, 50 },
		{ 54, 143, 50 },
		{ 54, 144, 50 },
		{ 54, 145, 50 },
		{ 54, 146, 50 },
		{ 54, 147, 50 },
		{ 54, 148, 50 },
		{ 54, 149, 50 },
		{ 54, 150, 50 },
		{ 54, 151, 50 },
		{ 54, 152, 50 },
		{ 54, 153, 50 },
		{ 54, 154, 50 },
		{ 54, 155, 50 },
		{ 54, 156, 50 },
		{ 54, 157, 50 },
		{ 54, 158, 50 },
		{ 54, 159, 50 },
		{ 54, 160, 50 },
		{ 54, 161, 50 },
		{ 54, 162, 50 },
		{ 54, 163, 50 },
		{ 54, 164, 50 },
		{ 54, 165, 50 },
		{ 54, 166, 50 },
		{ 54, 167, 50 },
		{ 54, 168, 50 },
		{ 54, 169, 50 },
		{ 54, 170, 50 },
		{ 54, 171, 50 },
		{ 54, 172, 50 },
		{ 54, 173, 50 },
		{ 54, 174, 50 },
		{ 54, 175, 50 },
		{ 54, 176, 50 },
		{ 54, 177, 50 },
		{ 54, 178, 50 },
		{ 54, 179, 50 },
		{ 54, 180, 50 },
		{ 55, 131, 50 },
		{ 55, 132, 50 },
		{ 55, 133, 50 },
		{ 55, 134, 50 },
		{ 55, 135, 50 },
		{ 55, 136, 50 },
		{ 55, 137, 50 },
		{ 55, 138, 50 },
		{ 55, 139, 50 },
		{ 55, 140, 50 },
		{ 55, 141, 50 },
		{ 55, 142, 50 },
		{ 55, 143, 50 },
		{ 55, 144, 50 },
		{ 55, 145, 50 },
		{ 55, 146, 50 },
		{ 55, 147, 50 },
		{ 55, 148, 50 },
		{ 55, 149, 50 },
		{ 55, 150, 50 },
		{ 55, 151, 50 },
		{ 55, 152, 50 },
		{ 55, 153, 50 },
		{ 55, 154, 50 },
		{ 55, 155, 50 },
		{ 55, 156, 50 },
		{ 55, 157, 50 },
		{ 55, 158, 50 },
		{ 55, 159, 50 },
		{ 55, 160, 50 },
		{ 55, 161, 50 },
		{ 55, 162, 50 },
		{ 55, 163, 50 },
		{ 55, 164, 50 },
		{ 55, 165, 50 },
		{ 55, 166, 50 },
		{ 55, 167, 50 },
		{ 55, 168, 50 },
		{ 55, 169, 50 },
		{ 55, 170, 50 },
		{ 55, 171, 50 },
		{ 55, 172, 50 },
		{ 55, 173, 50 },
		{ 55, 174, 50 },
		{ 55, 175, 50 },
		{ 55, 176, 50 },
		{ 55, 177, 50 },
		{ 55, 178, 50 },
		{ 55, 179, 50 },
		{ 55, 180, 50 },
		{ 55, 181, 50 },
		{ 56, 132, 50 },
		{ 56, 133, 50 },
		{ 56, 134, 50 },
		{ 56, 135, 50 },
		{ 56, 136, 50 },
		{ 56, 137, 50 },
		{ 56, 138, 50 },
		{ 56, 139, 50 },
		{ 56, 140, 50 },
		{ 56, 141, 50 },
		{ 56, 142, 50 },
		{ 56, 143, 50 },
		{ 56, 144, 50 },
		{ 56, 145, 50 },
		{ 56, 146, 50 },
		{ 56, 147, 50 },
		{ 56, 148, 50 },
		{ 56, 149, 50 },
		{ 56, 150, 50 },
		{ 56, 151, 50 },
		{ 56, 152, 50 },
		{ 56, 153, 50 },
		{ 56, 154, 50 },
		{ 56, 155, 50 },
		{ 56, 156, 50 },
		{ 56, 157, 50 },
		{ 56, 158, 50 },
		{ 56, 159, 50 },
		{ 56, 160, 50 },
		{ 56, 161, 50 },
		{ 56, 162, 50 },
		{ 56, 163, 50 },
		{ 56, 164, 50 },
		{ 56, 165, 50 },
		{ 56, 166, 50 },
		{ 56, 167, 50 },
		{ 56, 168, 50 },
		{ 56, 169, 50 },
		{ 56, 170, 50 },
		{ 56, 171, 50 },
		{ 56, 172, 50 },
		{ 56, 173, 50 },
		{ 56, 174, 50 },
		{ 56, 175, 50 },
		{ 56, 176, 50 },
		{ 56, 177, 50 },
		{ 56, 178, 50 },
		{ 56, 179, 50 },
		{ 56, 180, 50 },
		{ 56, 181, 50 },
		{ 57, 133, 50 },
		{ 57, 134, 50 },
		{ 57, 135, 50 },
		{ 57, 136, 50 },
		{ 57, 137, 50 },
		{ 57, 138, 50 },
		{ 57, 139, 50 },
		{ 57, 140, 50 },
		{ 57, 141, 50 },
		{ 57, 142, 50 },
		{ 57, 143, 50 },
		{ 57, 144, 50 },
		{ 57, 145, 50 },
		{ 57, 146, 50 },
		{ 57, 147, 50 },
		{ 57, 148, 50 },
		{ 57, 149, 50 },
		{ 57, 150, 50 },
		{ 57, 151, 50 },
		{ 57, 152, 50 },
		{ 57, 153, 50 },
		{ 57, 154, 50 },
		{ 57, 155, 50 },
		{ 57, 156, 50 },
		{ 57, 157, 50 },
		{ 57, 158, 50 },
		{ 57, 159, 50 },
		{ 57, 160, 50 },
		{ 57, 161, 50 },
		{ 57, 162, 50 },
		{ 57, 163, 50 },
		{ 57, 164, 50 },
		{ 57, 165, 50 },
		{ 57, 166, 50 },
		{ 57, 167, 50 },
		{ 57, 168, 50 },
		{ 57, 169, 50 },
		{ 57, 170, 50 },
		{ 57, 171, 50 },
		{ 57, 172, 50 },
		{ 57, 173, 50 },
		{ 57, 174, 50 },
		{ 57, 175, 50 },
		{ 57, 176, 50 },
		{ 57, 177, 50 },
		{ 57, 178, 50 },
		{ 57, 179, 50 },
		{ 57, 180, 50 },
		{ 57, 181, 50 },
		{ 58, 133, 50 },
		{ 58, 134, 50 },
		{ 58, 135, 50 },
		{ 58, 136, 50 },
		{ 58, 137, 50 },
		{ 58, 138, 50 },
		{ 58, 139, 50 },
		{ 58, 140, 50 },
		{ 58, 141, 50 },
		{ 58, 142, 50 },
		{ 58, 143, 50 },
		{ 58, 144, 50 },
		{ 58, 145, 50 },
		{ 58, 146, 50 },
		{ 58, 147, 50 },
		{ 58, 148, 50 },
		{ 58, 149, 50 },
		{ 58, 150, 50 },
		{ 58, 151, 50 },
		{ 58, 152, 50 },
		{ 58, 153, 50 },
		{ 58, 154, 50 },
		{ 58, 155, 50 },
		{ 58, 156, 50 },
		{ 58, 157, 50 },
		{ 58, 158, 50 },
		{ 58, 159, 50 },
		{ 58, 160, 50 },
		{ 58, 161, 50 },
		{ 58, 162, 50 },
		{ 58, 163, 50 },
		{ 58, 164, 50 },
		{ 58, 165, 50 },
		{ 58, 166, 50 },
		{ 58, 167, 50 },
		{ 58, 168, 50 },
		{ 58, 169, 50 },
		{ 58, 170, 50 },
		{ 58, 171, 50 },
		{ 58, 172, 50 },
		{ 58, 173, 50 },
		{ 58, 174, 50 },
		{ 58, 175, 50 },
		{ 58, 176, 50 },
		{ 58, 177, 50 },
		{ 58, 178, 50 },
		{ 58, 179, 50 },
		{ 58, 180, 50 },
		{ 58, 181, 50 },
		{ 58, 182, 50 },
		{ 59, 134, 50 },
		{ 59, 135, 50 },
		{ 59, 136, 50 },
		{ 59, 137, 50 },
		{ 59, 138, 50 },
		{ 59, 139, 50 },
		{ 59, 140, 50 },
		{ 59, 141, 50 },
		{ 59, 142, 50 },
		{ 59, 143, 50 },
		{ 59, 144, 50 },
		{ 59, 145, 50 },
		{ 59, 146, 50 },
		{ 59, 147, 50 },
		{ 59, 148, 50 },
		{ 59, 149, 50 },
		{ 59, 150, 50 },
		{ 59, 151, 50 },
		{ 59, 152, 50 },
		{ 59, 153, 50 },
		{ 59, 154, 50 },
		{ 59, 155, 50 },
		{ 59, 156, 50 },
		{ 59, 157, 50 },
		{ 59, 158, 50 },
		{ 59, 159, 50 },
		{ 59, 160, 50 },
		{ 59, 161, 50 },
		{ 59, 162, 50 },
		{ 59, 163, 50 },
		{ 59, 164, 50 },
		{ 59, 165, 50 },
		{ 59, 166, 50 },
		{ 59, 167, 50 },
		{ 59, 168, 50 },
		{ 59, 169, 50 },
		{ 59, 170, 50 },
		{ 59, 171, 50 },
		{ 59, 172, 50 },
		{ 59, 173, 50 },
		{ 59, 174, 50 },
		{ 59, 175, 50 },
		{ 59, 176, 50 },
		{ 59, 177, 50 },
		{ 59, 178, 50 },
		{ 59, 179, 50 },
		{ 59, 180, 50 },
		{ 59, 181, 50 },
		{ 59, 182, 50 },
		{ 60, 134, 50 },
		{ 60, 135, 50 },
		{ 60, 136, 50 },
		{ 60, 137, 50 },
		{ 60, 138, 50 },
		{ 60, 139, 50 },
		{ 60, 140, 50 },
		{ 60, 141, 50 },
		{ 60, 142, 50 },
		{ 60, 143, 50 },
		{ 60, 144, 50 },
		{ 60, 145, 50 },
		{ 60, 146, 50 },
		{ 60, 147, 50 },
		{ 60, 148, 50 },
		{ 60, 149, 50 },
		{ 60, 150, 50 },
		{ 60, 151, 50 },
		{ 60, 152, 50 },
		{ 60, 153, 50 },
		{ 60, 154, 50 },
		{ 60, 155, 50 },
		{ 60, 156, 50 },
		{ 60, 157, 50 },
		{ 60, 158, 50 },
		{ 60, 159, 50 },
		{ 60, 160, 50 },
		{ 60, 161, 50 },
		{ 60, 162, 50 },
		{ 60, 163, 50 },
		{ 60, 164, 50 },
		{ 60, 165, 50 },
		{ 60, 166, 50 },
		{ 60, 167, 50 },
		{ 60, 168, 50 },
		{ 60, 169, 50 },
		{ 60, 170, 50 },
		{ 60, 171, 50 },
		{ 60, 172, 50 },
		{ 60, 173, 50 },
		{ 60, 174, 50 },
		{ 60, 175, 50 },
		{ 60, 176, 50 },
		{ 60, 177, 50 },
		{ 60, 178, 50 },
		{ 60, 179, 50 },
		{ 60, 180, 50 },
		{ 60, 181, 50 },
		{ 60, 182, 50 },
		{ 60, 183, 50 },
		{ 61, 135, 50 },
		{ 61, 136, 50 },
		{ 61, 137, 50 },
		{ 61, 138, 50 },
		{ 61, 139, 50 },
		{ 61, 140, 50 },
		{ 61, 141, 50 },
		{ 61, 142, 50 },
		{ 61, 143, 50 },
		{ 61, 144, 50 },
		{ 61, 145, 50 },
		{ 61, 146, 50 },
		{ 61, 147, 50 },
		{ 61, 148, 50 },
		{ 61, 149, 50 },
		{ 61, 150, 50 },
		{ 61, 151, 50 },
		{ 61, 152, 50 },
		{ 61, 153, 50 },
		{ 61, 154, 50 },
		{ 61, 155, 50 },
		{ 61, 156, 50 },
		{ 61, 157, 50 },
		{ 61, 158, 50 },
		{ 61, 159, 50 },
		{ 61, 160, 50 },
		{ 61, 161, 50 },
		{ 61, 162, 50 },
		{ 61, 163, 50 },
		{ 61, 164, 50 },
		{ 61, 165, 50 },
		{ 61, 166, 50 },
		{ 61, 167, 50 },
		{ 61, 168, 50 },
		{ 61, 169, 50 },
		{ 61, 170, 50 },
		{ 61, 171, 50 },
		{ 61, 172, 50 },
		{ 61, 173, 50 },
		{ 61, 174, 50 },
		{ 61, 175, 50 },
		{ 61, 176, 50 },
		{ 61, 177, 50 },
		{ 61, 178, 50 },
		{ 61, 179, 50 },
		{ 61, 180, 50 },
		{ 61, 181, 50 },
		{ 61, 182, 50 },
		{ 61, 183, 50 },
		{ 62, 136, 50 },
		{ 62, 137, 50 },
		{ 62, 138, 50 },
		{ 62, 139, 50 },
		{ 62, 140, 50 },
		{ 62, 141, 50 },
		{ 62, 142, 50 },
		{ 62, 143, 50 },
		{ 62, 144, 50 },
		{ 62, 145, 50 },
		{ 62, 146, 50 },
		{ 62, 147, 50 },
		{ 62, 148, 50 },
		{ 62, 149, 50 },
		{ 62, 150, 50 },
		{ 62, 151, 50 },
		{ 62, 152, 50 },
		{ 62, 153, 50 },
		{ 62, 154, 50 },
		{ 62, 155, 50 },
		{ 62, 156, 50 },
		{ 62, 157, 50 },
		{ 62, 158, 50 },
		{ 62, 159, 50 },
		{ 62, 160, 50 },
		{ 62, 161, 50 },
		{ 62, 162, 50 },
		{ 62, 163, 50 },
		{ 62, 164, 50 },
		{ 62, 165, 50 },
		{ 62, 166, 50 },
		{ 62, 167, 50 },
		{ 62, 168, 50 },
		{ 62, 169, 50 },
		{ 62, 170, 50 },
		{ 62, 171, 50 },
		{ 62, 172, 50 },
		{ 62, 173, 50 },
		{ 62, 174, 50 },
		{ 62, 175, 50 },
		{ 62, 176, 50 },
		{ 62, 177, 50 },
		{ 62, 178, 50 },
		{ 62, 179, 50 },
		{ 62, 180, 50 },
		{ 62, 181, 50 },
		{ 62, 182, 50 },
		{ 62, 183, 50 },
		{ 63, 136, 50 },
		{ 63, 137, 50 },
		{ 63, 138, 50 },
		{ 63, 139, 50 },
		{ 63, 140, 50 },
		{ 63, 141, 50 },
		{ 63, 142, 50 },
		{ 63, 143, 50 },
		{ 63, 144, 50 },
		{ 63, 145, 50 },
		{ 63, 146, 50 },
		{ 63, 147, 50 },
		{ 63, 148, 50 },
		{ 63, 149, 50 },
		{ 63, 150, 50 },
		{ 63, 151, 50 },
		{ 63, 152, 50 },
		{ 63, 153, 50 },
		{ 63, 154, 50 },
		{ 63, 155, 50 },
		{ 63, 156, 50 },
		{ 63, 157, 50 },
		{ 63, 158, 50 },
		{ 63, 159, 50 },
		{ 63, 160, 50 },
		{ 63, 161, 50 },
		{ 63, 162, 50 },
		{ 63, 163, 50 },
		{ 63, 164, 50 },
		{ 63, 165, 50 },
		{ 63, 166, 50 },
		{ 63, 167, 50 },
		{ 63, 168, 50 },
		{ 63, 169, 50 },
		{ 63, 170, 50 },
		{ 63, 171, 50 },
		{ 63, 172, 50 },
		{ 63, 173, 50 },
		{ 63, 174, 50 },
		{ 63, 175, 50 },
		{ 63, 176, 50 },
		{ 63, 177, 50 },
		{ 63, 178, 50 },
		{ 63, 179, 50 },
		{ 63, 180, 50 },
		{ 63, 181, 50 },
		{ 63, 182, 50 },
		{ 63, 183, 50 },
		{ 63, 184, 50 },
		{ 64, 137, 50 },
		{ 64, 138, 50 },
		{ 64, 139, 50 },
		{ 64, 140, 50 },
		{ 64, 141, 50 },
		{ 64, 142, 50 },
		{ 64, 143, 50 },
		{ 64, 144, 50 },
		{ 64, 145, 50 },
		{ 64, 146, 50 },
		{ 64, 147, 50 },
		{ 64, 148, 50 },
		{ 64, 149, 50 },
		{ 64, 150, 50 },
		{ 64, 151, 50 },
		{ 64, 152, 50 },
		{ 64, 153, 50 },
		{ 64, 154, 50 },
		{ 64, 155, 50 },
		{ 64, 156, 50 },
		{ 64, 157, 50 },
		{ 64, 158, 50 },
		{ 64, 159, 50 },
		{ 64, 160, 50 },
		{ 64, 161, 50 },
		{ 64, 162, 50 },
		{ 64, 163, 50 },
		{ 64, 164, 50 },
		{ 64, 165, 50 },
		{ 64, 166, 50 },
		{ 64, 167, 50 },
		{ 64, 168, 50 },
		{ 64, 169, 50 },
		{ 64, 170, 50 },
		{ 64, 171, 50 },
		{ 64, 172, 50 },
		{ 64, 173, 50 },
		{ 64, 174, 50 },
		{ 64, 175, 50 },
		{ 64, 176, 50 },
		{ 64, 177, 50 },
		{ 64, 178, 50 },
		{ 64, 179, 50 },
		{ 64, 180, 50 },
		{ 64, 181, 50 },
		{ 64, 182, 50 },
		{ 64, 183, 50 },
		{ 64, 184, 50 },
		{ 65, 138, 50 },
		{ 65, 139, 50 },
		{ 65, 140, 50 },
		{ 65, 141, 50 },
		{ 65, 142, 50 },
		{ 65, 143, 50 },
		{ 65, 144, 50 },
		{ 65, 145, 50 },
		{ 65, 146, 50 },
		{ 65, 147, 50 },
		{ 65, 148, 50 },
		{ 65, 149, 50 },
		{ 65, 150, 50 },
		{ 65, 151, 50 },
		{ 65, 152, 50 },
		{ 65, 153, 50 },
		{ 65, 154, 50 },
		{ 65, 155, 50 },
		{ 65, 156, 50 },
		{ 65, 157, 50 },
		{ 65, 158, 50 },
		{ 65, 159, 50 },
		{ 65, 160, 50 },
		{ 65, 161, 50 },
		{ 65, 162, 50 },
		{ 65, 163, 50 },
		{ 65, 164, 50 },
		{ 65, 165, 50 },
		{ 65, 166, 50 },
		{ 65, 167, 50 },
		{ 65, 168, 50 },
		{ 65, 169, 50 },
		{ 65, 170, 50 },
		{ 65, 171, 50 },
		{ 65, 172, 50 },
		{ 65, 173, 50 },
		{ 65, 174, 50 },
		{ 65, 175, 50 },
		{ 65, 176, 50 },
		{ 65, 177, 50 },
		{ 65, 178, 50 },
		{ 65, 179, 50 },
		{ 65, 180, 50 },
		{ 65, 181, 50 },
		{ 65, 182, 50 },
		{ 65, 183, 50 },
		{ 65, 184, 50 },
		{ 66, 138, 50 },
		{ 66, 139, 50 },
		{ 66, 140, 50 },
		{ 66, 141, 50 },
		{ 66, 142, 50 },
		{ 66, 143, 50 },
		{ 66, 144, 50 },
		{ 66, 145, 50 },
		{ 66, 146, 50 },
		{ 66, 147, 50 },
		{ 66, 148, 50 },
		{ 66, 149, 50 },
		{ 66, 150, 50 },
		{ 66, 151, 50 },
		{ 66, 152, 50 },
		{ 66, 153, 50 },
		{ 66, 154, 50 },
		{ 66, 155, 50 },
		{ 66, 156, 50 },
		{ 66, 157, 50 },
		{ 66, 158, 50 },
		{ 66, 159, 50 },
		{ 66, 160, 50 },
		{ 66, 161, 50 },
		{ 66, 162, 50 },
		{ 66, 163, 50 },
		{ 66, 164, 50 },
		{ 66, 165, 50 },
		{ 66, 166, 50 },
		{ 66, 167, 50 },
		{ 66, 168, 50 },
		{ 66, 169, 50 },
		{ 66, 170, 50 },
		{ 66, 171, 50 },
		{ 66, 172, 50 },
		{ 66, 173, 50 },
		{ 66, 174, 50 },
		{ 66, 175, 50 },
		{ 66, 176, 50 },
		{ 66, 177, 50 },
		{ 66, 178, 50 },
		{ 66, 179, 50 },
		{ 66, 180, 50 },
		{ 66, 181, 50 },
		{ 66, 182, 50 },
		{ 66, 183, 50 },
		{ 66, 184, 50 },
		{ 66, 185, 50 },
		{ 67, 139, 50 },
		{ 67, 140, 50 },
		{ 67, 141, 50 },
		{ 67, 142, 50 },
		{ 67, 143, 50 },
		{ 67, 144, 50 },
		{ 67, 145, 50 },
		{ 67, 146, 50 },
		{ 67, 147, 50 },
		{ 67, 148, 50 },
		{ 67, 149, 50 },
		{ 67, 150, 50 },
		{ 67, 151, 50 },
		{ 67, 152, 50 },
		{ 67, 153, 50 },
		{ 67, 154, 50 },
		{ 67, 155, 50 },
		{ 67, 156, 50 },
		{ 67, 157, 50 },
		{ 67, 158, 50 },
		{ 67, 159, 50 },
		{ 67, 160, 50 },
		{ 67, 161, 50 },
		{ 67, 162, 50 },
		{ 67, 163, 50 },
		{ 67, 164, 50 },
		{ 67, 165, 50 },
		{ 67, 166, 50 },
		{ 67, 167, 50 },
		{ 67, 168, 50 },
		{ 67, 169, 50 },
		{ 67, 170, 50 },
		{ 67, 171, 50 },
		{ 67, 172, 50 },
		{ 67, 173, 50 },
		{ 67, 174, 50 },
		{ 67, 175, 50 },
		{ 67, 176, 50 },
		{ 67, 177, 50 },
		{ 67, 178, 50 },
		{ 67, 179, 50 },
		{ 67, 180, 50 },
		{ 67, 181, 50 },
		{ 67, 182, 50 },
		{ 67, 183, 50 },
		{ 67, 184, 50 },
		{ 67, 185, 50 },
		{ 68, 139, 50 },
		{ 68, 140, 50 },
		{ 68, 141, 50 },
		{ 68, 142, 50 },
		{ 68, 143, 50 },
		{ 68, 144, 50 },
		{ 68, 145, 50 },
		{ 68, 146, 50 },
		{ 68, 147, 50 },
		{ 68, 148, 50 },
		{ 68, 149, 50 },
		{ 68, 150, 50 },
		{ 68, 151, 50 },
		{ 68, 152, 50 },
		{ 68, 153, 50 },
		{ 68, 154, 50 },
		{ 68, 155, 50 },
		{ 68, 156, 50 },
		{ 68, 157, 50 },
		{ 68, 158, 50 },
		{ 68, 159, 50 },
		{ 68, 160, 50 },
		{ 68, 161, 50 },
		{ 68, 162, 50 },
		{ 68, 163, 50 },
		{ 68, 164, 50 },
		{ 68, 165, 50 },
		{ 68, 166, 50 },
		{ 68, 167, 50 },
		{ 68, 168, 50 },
		{ 68, 169, 50 },
		{ 68, 170, 50 },
		{ 68, 171, 50 },
		{ 68, 172, 50 },
		{ 68, 173, 50 },
		{ 68, 174, 50 },
		{ 68, 175, 50 },
		{ 68, 176, 50 },
		{ 68, 177, 50 },
		{ 68, 178, 50 },
		{ 68, 179, 50 },
		{ 68, 180, 50 },
		{ 68, 181, 50 },
		{ 68, 182, 50 },
		{ 68, 183, 50 },
		{ 68, 184, 50 },
		{ 68, 185, 50 },
		{ 68, 186, 50 },
		{ 69, 140, 50 },
		{ 69, 141, 50 },
		{ 69, 142, 50 },
		{ 69, 143, 50 },
		{ 69, 144, 50 },
		{ 69, 145, 50 },
		{ 69, 146, 50 },
		{ 69, 147, 50 },
		{ 69, 148, 50 },
		{ 69, 149, 50 },
		{ 69, 150, 50 },
		{ 69, 151, 50 },
		{ 69, 152, 50 },
		{ 69, 153, 50 },
		{ 69, 154, 50 },
		{ 69, 155, 50 },
		{ 69, 156, 50 },
		{ 69, 157, 50 },
		{ 69, 158, 50 },
		{ 69, 159, 50 },
		{ 69, 160, 50 },
		{ 69, 161, 50 },
		{ 69, 162, 50 },
		{ 69, 163, 50 },
		{ 69, 164, 50 },
		{ 69, 165, 50 },
		{ 69, 166, 50 },
		{ 69, 167, 50 },
		{ 69, 168, 50 },
		{ 69, 169, 50 },
		{ 69, 170, 50 },
		{ 69, 171, 50 },
		{ 69, 172, 50 },
		{ 69, 173, 50 },
		{ 69, 174, 50 },
		{ 69, 175, 50 },
		{ 69, 176, 50 },
		{ 69, 177, 50 },
		{ 69, 178, 50 },
		{ 69, 179, 50 },
		{ 69, 180, 50 },
		{ 69, 181, 50 },
		{ 69, 182, 50 },
		{ 69, 183, 50 },
		{ 69, 184, 50 },
		{ 69, 185, 50 },
		{ 69, 186, 50 },
		{ 70, 141, 50 },
		{ 70, 142, 50 },
		{ 70, 143, 50 },
		{ 70, 144, 50 },
		{ 70, 145, 50 },
		{ 70, 146, 50 },
		{ 70, 147, 50 },
		{ 70, 148, 50 },
		{ 70, 149, 50 },
		{ 70, 150, 50 },
		{ 70, 151, 50 },
		{ 70, 152, 50 },
		{ 70, 153, 50 },
		{ 70, 154, 50 },
		{ 70, 155, 50 },
		{ 70, 156, 50 },
		{ 70, 157, 50 },
		{ 70, 158, 50 },
		{ 70, 159, 50 },
		{ 70, 160, 50 },
		{ 70, 161, 50 },
		{ 70, 162, 50 },
		{ 70, 163, 50 },
		{ 70, 164, 50 },
		{ 70, 165, 50 },
		{ 70, 166, 50 },
		{ 70, 167, 50 },
		{ 70, 168, 50 },
		{ 70, 169, 50 },
		{ 70, 170, 50 },
		{ 70, 171, 50 },
		{ 70, 172, 50 },
		{ 70, 173, 50 },
		{ 70, 174, 50 },
		{ 70, 175, 50 },
		{ 70, 176, 50 },
		{ 70, 177, 50 },
		{ 70, 178, 50 },
		{ 70, 179, 50 },
		{ 70, 180, 50 },
		{ 70, 181, 50 },
		{ 70, 182, 50 },
		{ 70, 183, 50 },
		{ 70, 184, 50 },
		{ 70, 185, 50 },
		{ 70, 186, 50 },
		{ 71, 141, 50 },
		{ 71, 142, 50 },
		{ 71, 143, 50 },
		{ 71, 144, 50 },
		{ 71, 145, 50 },
		{ 71, 146, 50 },
		{ 71, 147, 50 },
		{ 71, 148, 50 },
		{ 71, 149, 50 },
		{ 71, 150, 50 },
		{ 71, 151, 50 },
		{ 71, 152, 50 },
		{ 71, 153, 50 },
		{ 71, 154, 50 },
		{ 71, 155, 50 },
		{ 71, 156, 50 },
		{ 71, 157, 50 },
		{ 71, 158, 50 },
		{ 71, 159, 50 },
		{ 71, 160, 50 },
		{ 71, 161, 50 },
		{ 71, 162, 50 },
		{ 71, 163, 50 },
		{ 71, 164, 50 },
		{ 71, 165, 50 },
		{ 71, 166, 50 },
		{ 71, 167, 50 },
		{ 71, 168, 50 },
		{ 71, 169, 50 },
		{ 71, 170, 50 },
		{ 71, 171, 50 },
		{ 71, 172, 50 },
		{ 71, 173, 50 },
		{ 71, 174, 50 },
		{ 71, 175, 50 },
		{ 71, 176, 50 },
		{ 71, 177, 50 },
		{ 71, 178, 50 },
		{ 71, 179, 50 },
		{ 71, 180, 50 },
		{ 71, 181, 50 },
		{ 71, 182, 50 },
		{ 71, 183, 50 },
		{ 71, 184, 50 },
		{ 71, 185, 50 },
		{ 71, 186, 50 },
		{ 71, 187, 50 },
		{ 72, 142, 50 },
		{ 72, 143, 50 },
		{ 72, 144, 50 },
		{ 72, 145, 50 },
		{ 72, 146, 50 },
		{ 72, 147, 50 },
		{ 72, 148, 50 },
		{ 72, 149, 50 },
		{ 72, 150, 50 },
		{ 72, 151, 50 },
		{ 72, 152, 50 },
		{ 72, 153, 50 },
		{ 72, 154, 50 },
		{ 72, 155, 50 },
		{ 72, 156, 50 },
		{ 72, 157, 50 },
		{ 72, 158, 50 },
		{ 72, 159, 50 },
		{ 72, 160, 50 },
		{ 72, 161, 50 },
		{ 72, 162, 50 },
		{ 72, 163, 50 },
		{ 72, 164, 50 },
		{ 72, 165, 50 },
		{ 72, 166, 50 },
		{ 72, 167, 50 },
		{ 72, 168, 50 },
		{ 72, 169, 50 },
		{ 72, 170, 50 },
		{ 72, 171, 50 },
		{ 72, 172, 50 },
		{ 72, 173, 50 },
		{ 72, 174, 50 },
		{ 72, 175, 50 },
		{ 72, 176, 50 },
		{ 72, 177, 50 },
		{ 72, 178, 50 },
		{ 72, 179, 50 },
		{ 72, 180, 50 },
		{ 72, 181, 50 },
		{ 72, 182, 50 },
		{ 72, 183, 50 },
		{ 72, 184, 50 },
		{ 72, 185, 50 },
		{ 72, 186, 50 },
		{ 72, 187, 50 },
		{ 73, 143, 50 },
		{ 73, 144, 50 },
		{ 73, 145, 50 },
		{ 73, 146, 50 },
		{ 73, 147, 50 },
		{ 73, 148, 50 },
		{ 73, 149, 50 },
		{ 73, 150, 50 },
		{ 73, 151, 50 },
		{ 73, 152, 50 },
		{ 73, 153, 50 },
		{ 73, 154, 50 },
		{ 73, 155, 50 },
		{ 73, 156, 50 },
		{ 73, 157, 50 },
		{ 73, 158, 50 },
		{ 73, 159, 50 },
		{ 73, 160, 50 },
		{ 73, 161, 50 },
		{ 73, 162, 50 },
		{ 73, 163, 50 },
		{ 73, 164, 50 },
		{ 73, 165, 50 },
		{ 73, 166, 50 },
		{ 73, 167, 50 },
		{ 73, 168, 50 },
		{ 73, 169, 50 },
		{ 73, 170, 50 },
		{ 73, 171, 50 },
		{ 73, 172, 50 },
		{ 73, 173, 50 },
		{ 73, 174, 50 },
		{ 73, 175, 50 },
		{ 73, 176, 50 },
		{ 73, 177, 50 },
		{ 73, 178, 50 },
		{ 73, 179, 50 },
		{ 73, 180, 50 },
		{ 73, 181, 50 },
		{ 73, 182, 50 },
		{ 73, 183, 50 },
		{ 73, 184, 50 },
		{ 73, 185, 50 },
		{ 73, 186, 50 },
		{ 73, 187, 50 },
		{ 74, 143, 50 },
		{ 74, 144, 50 },
		{ 74, 145, 50 },
		{ 74, 146, 50 },
		{ 74, 147, 50 },
		{ 74, 148, 50 },
		{ 74, 149, 50 },
		{ 74, 150, 50 },
		{ 74, 151, 50 },
		{ 74, 152, 50 },
		{ 74, 153, 50 },
		{ 74, 154, 50 },
		{ 74, 155, 50 },
		{ 74, 156, 50 },
		{ 74, 157, 50 },
		{ 74, 158, 50 },
		{ 74, 159, 50 },
		{ 74, 160, 50 },
		{ 74, 161, 50 },
		{ 74, 162, 50 },
		{ 74, 163, 50 },
		{ 74, 164, 50 },
		{ 74, 165, 50 },
		{ 74, 166, 50 },
		{ 74, 167, 50 },
		{ 74, 168, 50 },
		{ 74, 169, 50 },
		{ 74, 170, 50 },
		{ 74, 171, 50 },
		{ 74, 172, 50 },
		{ 74, 173, 50 },
		{ 74, 174, 50 },
		{ 74, 175, 50 },
		{ 74, 176, 50 },
		{ 74, 177, 50 },
		{ 74, 178, 50 },
		{ 74, 179, 50 },
		{ 74, 180, 50 },
		{ 74, 181, 50 },
		{ 74, 182, 50 },
		{ 74, 183, 50 },
		{ 74, 184, 50 },
		{ 74, 185, 50 },
		{ 74, 186, 50 },
		{ 74, 187, 50 },
		{ 74, 188, 50 },
		{ 75, 144, 50 },
		{ 75, 145, 50 },
		{ 75, 146, 50 },
		{ 75, 147, 50 },
		{ 75, 148, 50 },
		{ 75, 149, 50 },
		{ 75, 150, 50 },
		{ 75, 151, 50 },
		{ 75, 152, 50 },
		{ 75, 153, 50 },
		{ 75, 154, 50 },
		{ 75, 155, 50 },
		{ 75, 156, 50 },
		{ 75, 157, 50 },
		{ 75, 158, 50 },
		{ 75, 159, 50 },
		{ 75, 160, 50 },
		{ 75, 161, 50 },
		{ 75, 162, 50 },
		{ 75, 163, 50 },
		{ 75, 164, 50 },
		{ 75, 165, 50 },
		{ 75, 166, 50 },
		{ 75, 167, 50 },
		{ 75, 168, 50 },
		{ 75, 169, 50 },
		{ 75, 170, 50 },
		{ 75, 171, 50 },
		{ 75, 172, 50 },
		{ 75, 173, 50 },
		{ 75, 174, 50 },
		{ 75, 175, 50 },
		{ 75, 176, 50 },
		{ 75, 177, 50 },
		{ 75, 178, 50 },
		{ 75, 179, 50 },
		{ 75, 180, 50 },
		{ 75, 181, 50 },
		{ 75, 182, 50 },
		{ 75, 183, 50 },
		{ 75, 184, 50 },
		{ 75, 185, 50 },
		{ 75, 186, 50 },
		{ 75, 187, 50 },
		{ 75, 188, 50 },
		{ 76, 144, 50 },
		{ 76, 145, 50 },
		{ 76, 146, 50 },
		{ 76, 147, 50 },
		{ 76, 148, 50 },
		{ 76, 149, 50 },
		{ 76, 150, 50 },
		{ 76, 151, 50 },
		{ 76, 152, 50 },
		{ 76, 153, 50 },
		{ 76, 154, 50 },
		{ 76, 155, 50 },
		{ 76, 156, 50 },
		{ 76, 157, 50 },
		{ 76, 158, 50 },
		{ 76, 159, 50 },
		{ 76, 160, 50 },
		{ 76, 161, 50 },
		{ 76, 162, 50 },
		{ 76, 163, 50 },
		{ 76, 164, 50 },
		{ 76, 165, 50 },
		{ 76, 166, 50 },
		{ 76, 167, 50 },
		{ 76, 168, 50 },
		{ 76, 169, 50 },
		{ 76, 170, 50 },
		{ 76, 171, 50 },
		{ 76, 172, 50 },
		{ 76, 173, 50 },
		{ 76, 174, 50 },
		{ 76, 175, 50 },
		{ 76, 176, 50 },
		{ 76, 177, 50 },
		{ 76, 178, 50 },
		{ 76, 179, 50 },
		{ 76, 180, 50 },
		{ 76, 181, 50 },
		{ 76, 182, 50 },
		{ 76, 183, 50 },
		{ 76, 184, 50 },
		{ 76, 185, 50 },
		{ 76, 186, 50 },
		{ 76, 187, 50 },
		{ 76, 188, 50 },
		{ 76, 189, 50 },
		{ 77, 145, 50 },
		{ 77, 146, 50 },
		{ 77, 147, 50 },
		{ 77, 148, 50 },
		{ 77, 149, 50 },
		{ 77, 150, 50 },
		{ 77, 151, 50 },
		{ 77, 152, 50 },
		{ 77, 153, 50 },
		{ 77, 154, 50 },
		{ 77, 155, 50 },
		{ 77, 156, 50 },
		{ 77, 157, 50 },
		{ 77, 158, 50 },
		{ 77, 159, 50 },
		{ 77, 160, 50 },
		{ 77, 161, 50 },
		{ 77, 162, 50 },
		{ 77, 163, 50 },
		{ 77, 164, 50 },
		{ 77, 165, 50 },
		{ 77, 166, 50 },
		{ 77, 167, 50 },
		{ 77, 168, 50 },
		{ 77, 169, 50 },
		{ 77, 170, 50 },
		{ 77, 171, 50 },
		{ 77, 172, 50 },
		{ 77, 173, 50 },
		{ 77, 174, 50 },
		{ 77, 175, 50 },
		{ 77, 176, 50 },
		{ 77, 177, 50 },
		{ 77, 178, 50 },
		{ 77, 179, 50 },
		{ 77, 180, 50 },
		{ 77, 181, 50 },
		{ 77, 182, 50 },
		{ 77, 183, 50 },
		{ 77, 184, 50 },
		{ 77, 185, 50 },
		{ 77, 186, 50 },
		{ 77, 187, 50 },
		{ 77, 188, 50 },
		{ 77, 189, 50 },
		{ 78, 146, 50 },
		{ 78, 147, 50 },
		{ 78, 148, 50 },
		{ 78, 149, 50 },
		{ 78, 150, 50 },
		{ 78, 151, 50 },
		{ 78, 152, 50 },
		{ 78, 153, 50 },
		{ 78, 154, 50 },
		{ 78, 155, 50 },
		{ 78, 156, 50 },
		{ 78, 157, 50 },
		{ 78, 158, 50 },
		{ 78, 159, 50 },
		{ 78, 160, 50 },
		{ 78, 161, 50 },
		{ 78, 162, 50 },
		{ 78, 163, 50 },
		{ 78, 164, 50 },
		{ 78, 165, 50 },
		{ 78, 166, 50 },
		{ 78, 167, 50 },
		{ 78, 168, 50 },
		{ 78, 169, 50 },
		{ 78, 170, 50 },
		{ 78, 171, 50 },
		{ 78, 172, 50 },
		{ 78, 173, 50 },
		{ 78, 174, 50 },
		{ 78, 175, 50 },
		{ 78, 176, 50 },
		{ 78, 177, 50 },
		{ 78, 178, 50 },
		{ 78, 179, 50 },
		{ 78, 180, 50 },
		{ 78, 181, 50 },
		{ 78, 182, 50 },
		{ 78, 183, 50 },
		{ 78, 184, 50 },
		{ 78, 185, 50 },
		{ 78, 186, 50 },
		{ 78, 187, 50 },
		{ 78, 188, 50 },
		{ 78, 189, 50 },
		{ 79, 146, 50 },
		{ 79, 147, 50 },
		{ 79, 148, 50 },
		{ 79, 149, 50 },
		{ 79, 150, 50 },
		{ 79, 151, 50 },
		{ 79, 152, 50 },
		{ 79, 153, 50 },
		{ 79, 154, 50 },
		{ 79, 155, 50 },
		{ 79, 156, 50 },
		{ 79, 157, 50 },
		{ 79, 158, 50 },
		{ 79, 159, 50 },
		{ 79, 160, 50 },
		{ 79, 161, 50 },
		{ 79, 162, 50 },
		{ 79, 163, 50 },
		{ 79, 164, 50 },
		{ 79, 165, 50 },
		{ 79, 166, 50 },
		{ 79, 167, 50 },
		{ 79, 168, 50 },
		{ 79, 169, 50 },
		{ 79, 170, 50 },
		{ 79, 171, 50 },
		{ 79, 172, 50 },
		{ 79, 173, 50 },
		{ 79, 174, 50 },
		{ 79, 175, 50 },
		{ 79, 176, 50 },
		{ 79, 177, 50 },
		{ 79, 178, 50 },
		{ 79, 179, 50 },
		{ 79, 180, 50 },
		{ 79, 181, 50 },
		{ 79, 182, 50 },
		{ 79, 183, 50 },
		{ 79, 184, 50 },
		{ 79, 185, 50 },
		{ 79, 186, 50 },
		{ 79, 187, 50 },
		{ 79, 188, 50 },
		{ 79, 189, 50 },
		{ 79, 190, 50 },
		{ 80, 147, 50 },
		{ 80, 148, 50 },
		{ 80, 149, 50 },
		{ 80, 150, 50 },
		{ 80, 151, 50 },
		{ 80, 152, 50 },
		{ 80, 153, 50 },
		{ 80, 154, 50 },
		{ 80, 155, 50 },
		{ 80, 156, 50 },
		{ 80, 157, 50 },
		{ 80, 158, 50 },
		{ 80, 159, 50 },
		{ 80, 160, 50 },
		{ 80, 161, 50 },
		{ 80, 162, 50 },
		{ 80, 163, 50 },
		{ 80, 164, 50 },
		{ 80, 165, 50 },
		{ 80, 166, 50 },
		{ 80, 167, 50 },
		{ 80, 168, 50 },
		{ 80, 169, 50 },
		{ 80, 170, 50 },
		{ 80, 171, 50 },
		{ 80, 172, 50 },
		{ 80, 173, 50 },
		{ 80, 174, 50 },
		{ 80, 175, 50 },
		{ 80, 176, 50 },
		{ 80, 177, 50 },
		{ 80, 178, 50 },
		{ 80, 179, 50 },
		{ 80, 180, 50 },
		{ 80, 181, 50 },
		{ 80, 182, 50 },
		{ 80, 183, 50 },
		{ 80, 184, 50 },
		{ 80, 185, 50 },
		{ 80, 186, 50 },
		{ 80, 187, 50 },
		{ 80, 188, 50 },
		{ 80, 189, 50 },
		{ 80, 190, 50 },
		{ 81, 148, 50 },
		{ 81, 149, 50 },
		{ 81, 150, 50 },
		{ 81, 151, 50 },
		{ 81, 152, 50 },
		{ 81, 153, 50 },
		{ 81, 154, 50 },
		{ 81, 155, 50 },
		{ 81, 156, 50 },
		{ 81, 157, 50 },
		{ 81, 158, 50 },
		{ 81, 159, 50 },
		{ 81, 160, 50 },
		{ 81, 161, 50 },
		{ 81, 162, 50 },
		{ 81, 163, 50 },
		{ 81, 164, 50 },
		{ 81, 165, 50 },
		{ 81, 166, 50 },
		{ 81, 167, 50 },
		{ 81, 168, 50 },
		{ 81, 169, 50 },
		{ 81, 170, 50 },
		{ 81, 171, 50 },
		{ 81, 172, 50 },
		{ 81, 173, 50 },
		{ 81, 174, 50 },
		{ 81, 175, 50 },
		{ 81, 176, 50 },
		{ 81, 177, 50 },
		{ 81, 178, 50 },
		{ 81, 179, 50 },
		{ 81, 180, 50 },
		{ 81, 181, 50 },
		{ 81, 182, 50 },
		{ 81, 183, 50 },
		{ 81, 184, 50 },
		{ 81, 185, 50 },
		{ 81, 186, 50 },
		{ 81, 187, 50 },
		{ 81, 188, 50 },
		{ 81, 189, 50 },
		{ 81, 190, 50 },
		{ 82, 148, 50 },
		{ 82, 149, 50 },
		{ 82, 150, 50 },
		{ 82, 151, 50 },
		{ 82, 152, 50 },
		{ 82, 153, 50 },
		{ 82, 154, 50 },
		{ 82, 155, 50 },
		{ 82, 156, 50 },
		{ 82, 157, 50 },
		{ 82, 158, 50 },
		{ 82, 159, 50 },
		{ 82, 160, 50 },
		{ 82, 161, 50 },
		{ 82, 162, 50 },
		{ 82, 163, 50 },
		{ 82, 164, 50 },
		{ 82, 165, 50 },
		{ 82, 166, 50 },
		{ 82, 167, 50 },
		{ 82, 168, 50 },
		{ 82, 169, 50 },
		{ 82, 170, 50 },
		{ 82, 171, 50 },
		{ 82, 172, 50 },
		{ 82, 173, 50 },
		{ 82, 174, 50 },
		{ 82, 175, 50 },
		{ 82, 176, 50 },
		{ 82, 177, 50 },
		{ 82, 178, 50 },
		{ 82, 179, 50 },
		{ 82, 180, 50 },
		{ 82, 181, 50 },
		{ 82, 182, 50 },
		{ 82, 183, 50 },
		{ 82, 184, 50 },
		{ 82, 185, 50 },
		{ 82, 186, 50 },
		{ 82, 187, 50 },
		{ 82, 188, 50 },
		{ 82, 189, 50 },
		{ 82, 190, 50 },
		{ 82, 191, 50 },
		{ 83, 149, 50 },
		{ 83, 150, 50 },
		{ 83, 151, 50 },
		{ 83, 152, 50 },
		{ 83, 153, 50 },
		{ 83, 154, 50 },
		{ 83, 155, 50 },
		{ 83, 156, 50 },
		{ 83, 157, 50 },
		{ 83, 158, 50 },
		{ 83, 159, 50 },
		{ 83, 160, 50 },
		{ 83, 161, 50 },
		{ 83, 162, 50 },
		{ 83, 163, 50 },
		{ 83, 164, 50 },
		{ 83, 165, 50 },
		{ 83, 166, 50 },
		{ 83, 167, 50 },
		{ 83, 168, 50 },
		{ 83, 169, 50 },
		{ 83, 170, 50 },
		{ 83, 171, 50 },
		{ 83, 172, 50 },
		{ 83, 173, 50 },
		{ 83, 174, 50 },
		{ 83, 175, 50 },
		{ 83, 176, 50 },
		{ 83, 177, 50 },
		{ 83, 178, 50 },
		{ 83, 179, 50 },
		{ 83, 180, 50 },
		{ 83, 181, 50 },
		{ 83, 182, 50 },
		{ 83, 183, 50 },
		{ 83, 184, 50 },
		{ 83, 185, 50 },
		{ 83, 186, 50 },
		{ 83, 187, 50 },
		{ 83, 188, 50 },
		{ 83, 189, 50 },
		{ 83, 190, 50 },
		{ 83, 191, 50 },
		{ 84, 149, 50 },
		{ 84, 150, 50 },
		{ 84, 151, 50 },
		{ 84, 152, 50 },
		{ 84, 153, 50 },
		{ 84, 154, 50 },
		{ 84, 155, 50 },
		{ 84, 156, 50 },
		{ 84, 157, 50 },
		{ 84, 158, 50 },
		{ 84, 159, 50 },
		{ 84, 160, 50 },
		{ 84, 161, 50 },
		{ 84, 162, 50 },
		{ 84, 163, 50 },
		{ 84, 164, 50 },
		{ 84, 165, 50 },
		{ 84, 166, 50 },
		{ 84, 167, 50 },
		{ 84, 168, 50 },
		{ 84, 169, 50 },
		{ 84, 170, 50 },
		{ 84, 171, 50 },
		{ 84, 172, 50 },
		{ 84, 173, 50 },
		{ 84, 174, 50 },
		{ 84, 175, 50 },
		{ 84, 176, 50 },
		{ 84, 177, 50 },
		{ 84, 178, 50 },
		{ 84, 179, 50 },
		{ 84, 180, 50 },
		{ 84, 181, 50 },
		{ 84, 182, 50 },
		{ 84, 183, 50 },
		{ 84, 184, 50 },
		{ 84, 185, 50 },
		{ 84, 186, 50 },
		{ 84, 187, 50 },
		{ 84, 188, 50 },
		{ 84, 189, 50 },
		{ 84, 190, 50 },
		{ 84, 191, 50 },
		{ 84, 192, 50 },
		{ 85, 150, 50 },
		{ 85, 151, 50 },
		{ 85, 152, 50 },
		{ 85, 153, 50 },
		{ 85, 154, 50 },
		{ 85, 155, 50 },
		{ 85, 156, 50 },
		{ 85, 157, 50 },
		{ 85, 158, 50 },
		{ 85, 159, 50 },
		{ 85, 160, 50 },
		{ 85, 161, 50 },
		{ 85, 162, 50 },
		{ 85, 163, 50 },
		{ 85, 164, 50 },
		{ 85, 165, 50 },
		{ 85, 166, 50 },
		{ 85, 167, 50 },
		{ 85, 168, 50 },
		{ 85, 169, 50 },
		{ 85, 170, 50 },
		{ 85, 171, 50 },
		{ 85, 172, 50 },
		{ 85, 173, 50 },
		{ 85, 174, 50 },
		{ 85, 175, 50 },
		{ 85, 176, 50 },
		{ 85, 177, 50 },
		{ 85, 178, 50 },
		{ 85, 179, 50 },
		{ 85, 180, 50 },
		{ 85, 181, 50 },
		{ 85, 182, 50 },
		{ 85, 183, 50 },
		{ 85, 184, 50 },
		{ 85, 185, 50 },
		{ 85, 186, 50 },
		{ 85, 187, 50 },
		{ 85, 188, 50 },
		{ 85, 189, 50 },
		{ 85, 190, 50 },
		{ 85, 191, 50 },
		{ 85, 192, 50 },
		{ 86, 151, 50 },
		{ 86, 152, 50 },
		{ 86, 153, 50 },
		{ 86, 154, 50 },
		{ 86, 155, 50 },
		{ 86, 156, 50 },
		{ 86, 157, 50 },
		{ 86, 158, 50 },
		{ 86, 159, 50 },
		{ 86, 160, 50 },
		{ 86, 161, 50 },
		{ 86, 162, 50 },
		{ 86, 163, 50 },
		{ 86, 164, 50 },
		{ 86, 165, 50 },
		{ 86, 166, 50 },
		{ 86, 167, 50 },
		{ 86, 168, 50 },
		{ 86, 169, 50 },
		{ 86, 170, 50 },
		{ 86, 171, 50 },
		{ 86, 172, 50 },
		{ 86, 173, 50 },
		{ 86, 174, 50 },
		{ 86, 175, 50 },
		{ 86, 176, 50 },
		{ 86, 177, 50 },
		{ 86, 178, 50 },
		{ 86, 179, 50 },
		{ 86, 180, 50 },
		{ 86, 181, 50 },
		{ 86, 182, 50 },
		{ 86, 183, 50 },
		{ 86, 184, 50 },
		{ 86, 185, 50 },
		{ 86, 186, 50 },
		{ 86, 187, 50 },
		{ 86, 188, 50 },
		{ 86, 189, 50 },
		{ 86, 190, 50 },
		{ 86, 191, 50 },
		{ 86, 192, 50 },
		{ 87, 151, 50 },
		{ 87, 152, 50 },
		{ 87, 153, 50 },
		{ 87, 154, 50 },
		{ 87, 155, 50 },
		{ 87, 156, 50 },
		{ 87, 157, 50 },
		{ 87, 158, 50 },
		{ 87, 159, 50 },
		{ 87, 160, 50 },
		{ 87, 161, 50 },
		{ 87, 162, 50 },
		{ 87, 163, 50 },
		{ 87, 164, 50 },
		{ 87, 165, 50 },
		{ 87, 166, 50 },
		{ 87, 167, 50 },
		{ 87, 168, 50 },
		{ 87, 169, 50 },
		{ 87, 170, 50 },
		{ 87, 171, 50 },
		{ 87, 172, 50 },
		{ 87, 173, 50 },
		{ 87, 174, 50 },
		{ 87, 175, 50 },
		{ 87, 176, 50 },
		{ 87, 177, 50 },
		{ 87, 178, 50 },
		{ 87, 179, 50 },
		{ 87, 180, 50 },
		{ 87, 181, 50 },
		{ 87, 182, 50 },
		{ 87, 183, 50 },
		{ 87, 184, 50 },
		{ 87, 185, 50 },
		{ 87, 186, 50 },
		{ 87, 187, 50 },
		{ 87, 188, 50 },
		{ 87, 189, 50 },
		{ 87, 190, 50 },
		{ 87, 191, 50 },
		{ 87, 192, 50 },
		{ 87, 193, 50 },
		{ 88, 152, 50 },
		{ 88, 153, 50 },
		{ 88, 154, 50 },
		{ 88, 155, 50 },
		{ 88, 156, 50 },
		{ 88, 157, 50 },
		{ 88, 158, 50 },
		{ 88, 159, 50 },
		{ 88, 160, 50 },
		{ 88, 161, 50 },
		{ 88, 162, 50 },
		{ 88, 163, 50 },
		{ 88, 164, 50 },
		{ 88, 165, 50 },
		{ 88, 166, 50 },
		{ 88, 167, 50 },
		{ 88, 168, 50 },
		{ 88, 169, 50 },
		{ 88, 170, 50 },
		{ 88, 171, 50 },
		{ 88, 172, 50 },
		{ 88, 173, 50 },
		{ 88, 174, 50 },
		{ 88, 175, 50 },
		{ 88, 176, 50 },
		{ 88, 177, 50 },
		{ 88, 178, 50 },
		{ 88, 179, 50 },
		{ 88, 180, 50 },
		{ 88, 181, 50 },
		{ 88, 182, 50 },
		{ 88, 183, 50 },
		{ 88, 184, 50 },
		{ 88, 185, 50 },
		{ 88, 186, 50 },
		{ 88, 187, 50 },
		{ 88, 188, 50 },
		{ 88, 189, 50 },
		{ 88, 190, 50 },
		{ 88, 191, 50 },
		{ 88, 192, 50 },
		{ 88, 193, 50 },
		{ 89, 153, 50 },
		{ 89, 154, 50 },
		{ 89, 155, 50 },
		{ 89, 156, 50 },
		{ 89, 157, 50 },
		{ 89, 158, 50 },
		{ 89, 159, 50 },
		{ 89, 160, 50 },
		{ 89, 161, 50 },
		{ 89, 162, 50 },
		{ 89, 163, 50 },
		{ 89, 164, 50 },
		{ 89, 165, 50 },
		{ 89, 166, 50 },
		{ 89, 167, 50 },
		{ 89, 168, 50 },
		{ 89, 169, 50 },
		{ 89, 170, 50 },
		{ 89, 171, 50 },
		{ 89, 172, 50 },
		{ 89, 173, 50 },
		{ 89, 174, 50 },
		{ 89, 175, 50 },
		{ 89, 176, 50 },
		{ 89, 177, 50 },
		{ 89, 178, 50 },
		{ 89, 179, 50 },
		{ 89, 180, 50 },
		{ 89, 181, 50 },
		{ 89, 182, 50 },
		{ 89, 183, 50 },
		{ 89, 184, 50 },
		{ 89, 185, 50 },
		{ 89, 186, 50 },
		{ 89, 187, 50 },
		{ 89, 188, 50 },
		{ 89, 189, 50 },
		{ 89, 190, 50 },
		{ 89, 191, 50 },
		{ 89, 192, 50 },
		{ 89, 193, 50 },
		{ 90, 153, 50 },
		{ 90, 154, 50 },
		{ 90, 155, 50 },
		{ 90, 156, 50 },
		{ 90, 157, 50 },
		{ 90, 158, 50 },
		{ 90, 159, 50 },
		{ 90, 160, 50 },
		{ 90, 161, 50 },
		{ 90, 162, 50 },
		{ 90, 163, 50 },
		{ 90, 164, 50 },
		{ 90, 165, 50 },
		{ 90, 166, 50 },
		{ 90, 167, 50 },
		{ 90, 168, 50 },
		{ 90, 169, 50 },
		{ 90, 170, 50 },
		{ 90, 171, 50 },
		{ 90, 172, 50 },
		{ 90, 173, 50 },
		{ 90, 174, 50 },
		{ 90, 175, 50 },
		{ 90, 176, 50 },
		{ 90, 177, 50 },
		{ 90, 178, 50 },
		{ 90, 179, 50 },
		{ 90, 180, 50 },
		{ 90, 181, 50 },
		{ 90, 182, 50 },
		{ 90, 183, 50 },
		{ 90, 184, 50 },
		{ 90, 185, 50 },
		{ 90, 186, 50 },
		{ 90, 187, 50 },
		{ 90, 188, 50 },
		{ 90, 189, 50 },
		{ 90, 190, 50 },
		{ 90, 191, 50 },
		{ 90, 192, 50 },
		{ 90, 193, 50 },
		{ 90, 194, 50 },
		{ 91, 154, 50 },
		{ 91, 155, 50 },
		{ 91, 156, 50 },
		{ 91, 157, 50 },
		{ 91, 158, 50 },
		{ 91, 159, 50 },
		{ 91, 160, 50 },
		{ 91, 161, 50 },
		{ 91, 162, 50 },
		{ 91, 163, 50 },
		{ 91, 164, 50 },
		{ 91, 165, 50 },
		{ 91, 166, 50 },
		{ 91, 167, 50 },
		{ 91, 168, 50 },
		{ 91, 169, 50 },
		{ 91, 170, 50 },
		{ 91, 171, 50 },
		{ 91, 172, 50 },
		{ 91, 173, 50 },
		{ 91, 174, 50 },
		{ 91, 175, 50 },
		{ 91, 176, 50 },
		{ 91, 177, 50 },
		{ 91, 178, 50 },
		{ 91, 179, 50 },
		{ 91, 180, 50 },
		{ 91, 181, 50 },
		{ 91, 182, 50 },
		{ 91, 183, 50 },
		{ 91, 184, 50 },
		{ 91, 185, 50 },
		{ 91, 186, 50 },
		{ 91, 187, 50 },
		{ 91, 188, 50 },
		{ 91, 189, 50 },
		{ 91, 190, 50 },
		{ 91, 191, 50 },
		{ 91, 192, 50 },
		{ 91, 193, 50 },
		{ 91, 194, 50 },
		{ 92, 154, 50 },
		{ 92, 155, 50 },
		{ 92, 156, 50 },
		{ 92, 157, 50 },
		{ 92, 158, 50 },
		{ 92, 159, 50 },
		{ 92, 160, 50 },
		{ 92, 161, 50 },
		{ 92, 162, 50 },
		{ 92, 163, 50 },
		{ 92, 164, 50 },
		{ 92, 165, 50 },
		{ 92, 166, 50 },
		{ 92, 167, 50 },
		{ 92, 168, 50 },
		{ 92, 169, 50 },
		{ 92, 170, 50 },
		{ 92, 171, 50 },
		{ 92, 172, 50 },
		{ 92, 173, 50 },
		{ 92, 174, 50 },
		{ 92, 175, 50 },
		{ 92, 176, 50 },
		{ 92, 177, 50 },
		{ 92, 178, 50 },
		{ 92, 179, 50 },
		{ 92, 180, 50 },
		{ 92, 181, 50 },
		{ 92, 182, 50 },
		{ 92, 183, 50 },
		{ 92, 184, 50 },
		{ 92, 185, 50 },
		{ 92, 186, 50 },
		{ 92, 187, 50 },
		{ 92, 188, 50 },
		{ 92, 189, 50 },
		{ 92, 190, 50 },
		{ 92, 191, 50 },
		{ 92, 192, 50 },
		{ 92, 193, 50 },
		{ 92, 194, 50 },
		{ 92, 195, 50 },
		{ 93, 155, 50 },
		{ 93, 156, 50 },
		{ 93, 157, 50 },
		{ 93, 158, 50 },
		{ 93, 159, 50 },
		{ 93, 160, 50 },
		{ 93, 161, 50 },
		{ 93, 162, 50 },
		{ 93, 163, 50 },
		{ 93, 164, 50 },
		{ 93, 165, 50 },
		{ 93, 166, 50 },
		{ 93, 167, 50 },
		{ 93, 168, 50 },
		{ 93, 169, 50 },
		{ 93, 170, 50 },
		{ 93, 171, 50 },
		{ 93, 172, 50 },
		{ 93, 173, 50 },
		{ 93, 174, 50 },
		{ 93, 175, 50 },
		{ 93, 176, 50 },
		{ 93, 177, 50 },
		{ 93, 178, 50 },
		{ 93, 179, 50 },
		{ 93, 180, 50 },
		{ 93, 181, 50 },
		{ 93, 182, 50 },
		{ 93, 183, 50 },
		{ 93, 184, 50 },
		{ 93, 185, 50 },
		{ 93, 186, 50 },
		{ 93, 187, 50 },
		{ 93, 188, 50 },
		{ 93, 189, 50 },
		{ 93, 190, 50 },
		{ 93, 191, 50 },
		{ 93, 192, 50 },
		{ 93, 193, 50 },
		{ 93, 194, 50 },
		{ 93, 195, 50 },
		{ 94, 156, 50 },
		{ 94, 157, 50 },
		{ 94, 158, 50 },
		{ 94, 159, 50 },
		{ 94, 160, 50 },
		{ 94, 161, 50 },
		{ 94, 162, 50 },
		{ 94, 163, 50 },
		{ 94, 164, 50 },
		{ 94, 165, 50 },
		{ 94, 166, 50 },
		{ 94, 167, 50 },
		{ 94, 168, 50 },
		{ 94, 169, 50 },
		{ 94, 170, 50 },
		{ 94, 171, 50 },
		{ 94, 172, 50 },
		{ 94, 173, 50 },
		{ 94, 174, 50 },
		{ 94, 175, 50 },
		{ 94, 176, 50 },
		{ 94, 177, 50 },
		{ 94, 178, 50 },
		{ 94, 179, 50 },
		{ 94, 180, 50 },
		{ 94, 181, 50 },
		{ 94, 182, 50 },
		{ 94, 183, 50 },
		{ 94, 184, 50 },
		{ 94, 185, 50 },
		{ 94, 186, 50 },
		{ 94, 187, 50 },
		{ 94, 188, 50 },
		{ 94, 189, 50 },
		{ 94, 190, 50 },
		{ 94, 191, 50 },
		{ 94, 192, 50 },
		{ 94, 193, 50 },
		{ 94, 194, 50 },
		{ 94, 195, 50 },
		{ 95, 156, 50 },
		{ 95, 157, 50 },
		{ 95, 158, 50 },
		{ 95, 159, 50 },
		{ 95, 160, 50 },
		{ 95, 161, 50 },
		{ 95, 162, 50 },
		{ 95, 163, 50 },
		{ 95, 164, 50 },
		{ 95, 165, 50 },
		{ 95, 166, 50 },
		{ 95, 167, 50 },
		{ 95, 168, 50 },
		{ 95, 169, 50 },
		{ 95, 170, 50 },
		{ 95, 171, 50 },
		{ 95, 172, 50 },
		{ 95, 173, 50 },
		{ 95, 174, 50 },
		{ 95, 175, 50 },
		{ 95, 176, 50 },
		{ 95, 177, 50 },
		{ 95, 178, 50 },
		{ 95, 179, 50 },
		{ 95, 180, 50 },
		{ 95, 181, 50 },
		{ 95, 182, 50 },
		{ 95, 183, 50 },
		{ 95, 184, 50 },
		{ 95, 185, 50 },
		{ 95, 186, 50 },
		{ 95, 187, 50 },
		{ 95, 188, 50 },
		{ 95, 189, 50 },
		{ 95, 190, 50 },
		{ 95, 191, 50 },
		{ 95, 192, 50 },
		{ 95, 193, 50 },
		{ 95, 194, 50 },
		{ 95, 195, 50 },
		{ 95, 196, 50 },
		{ 96, 157, 50 },
		{ 96, 158, 50 },
		{ 96, 159, 50 },
		{ 96, 160, 50 },
		{ 96, 161, 50 },
		{ 96, 162, 50 },
		{ 96, 163, 50 },
		{ 96, 164, 50 },
		{ 96, 165, 50 },
		{ 96, 166, 50 },
		{ 96, 167, 50 },
		{ 96, 168, 50 },
		{ 96, 169, 50 },
		{ 96, 170, 50 },
		{ 96, 171, 50 },
		{ 96, 172, 50 },
		{ 96, 173, 50 },
		{ 96, 174, 50 },
		{ 96, 175, 50 },
		{ 96, 176, 50 },
		{ 96, 177, 50 },
		{ 96, 178, 50 },
		{ 96, 179, 50 },
		{ 96, 180, 50 },
		{ 96, 181, 50 },
		{ 96, 182, 50 },
		{ 96, 183, 50 },
		{ 96, 184, 50 },
		{ 96, 185, 50 },
		{ 96, 186, 50 },
		{ 96, 187, 50 },
		{ 96, 188, 50 },
		{ 96, 189, 50 },
		{ 96, 190, 50 },
		{ 96, 191, 50 },
		{ 96, 192, 50 },
		{ 96, 193, 50 },
		{ 96, 194, 50 },
		{ 96, 195, 50 },
		{ 96, 196, 50 },
		{ 97, 158, 50 },
		{ 97, 159, 50 },
		{ 97, 160, 50 },
		{ 97, 161, 50 },
		{ 97, 162, 50 },
		{ 97, 163, 50 },
		{ 97, 164, 50 },
		{ 97, 165, 50 },
		{ 97, 166, 50 },
		{ 97, 167, 50 },
		{ 97, 168, 50 },
		{ 97, 169, 50 },
		{ 97, 170, 50 },
		{ 97, 171, 50 },
		{ 97, 172, 50 },
		{ 97, 173, 50 },
		{ 97, 174, 50 },
		{ 97, 175, 50 },
		{ 97, 176, 50 },
		{ 97, 177, 50 },
		{ 97, 178, 50 },
		{ 97, 179, 50 },
		{ 97, 180, 50 },
		{ 97, 181, 50 },
		{ 97, 182, 50 },
		{ 97, 183, 50 },
		{ 97, 184, 50 },
		{ 97, 185, 50 },
		{ 97, 186, 50 },
		{ 97, 187, 50 },
		{ 97, 188, 50 },
		{ 97, 189, 50 },
		{ 97, 190, 50 },
		{ 97, 191, 50 },
		{ 97, 192, 50 },
		{ 97, 193, 50 },
		{ 97, 194, 50 },
		{ 97, 195, 50 },
		{ 97, 196, 50 },
		{ 98, 158, 50 },
		{ 98, 159, 50 },
		{ 98, 160, 50 },
		{ 98, 161, 50 },
		{ 98, 162, 50 },
		{ 98, 163, 50 },
		{ 98, 164, 50 },
		{ 98, 165, 50 },
		{ 98, 166, 50 },
		{ 98, 167, 50 },
		{ 98, 168, 50 },
		{ 98, 169, 50 },
		{ 98, 170, 50 },
		{ 98, 171, 50 },
		{ 98, 172, 50 },
		{ 98, 173, 50 },
		{ 98, 174, 50 },
		{ 98, 175, 50 },
		{ 98, 176, 50 },
		{ 98, 177, 50 },
		{ 98, 178, 50 },
		{ 98, 179, 50 },
		{ 98, 180, 50 },
		{ 98, 181, 50 },
		{ 98, 182, 50 },
		{ 98, 183, 50 },
		{ 98, 184, 50 },
		{ 98, 185, 50 },
		{ 98, 186, 50 },
		{ 98, 187, 50 },
		{ 98, 188, 50 },
		{ 98, 189, 50 },
		{ 98, 190, 50 },
		{ 98, 191, 50 },
		{ 98, 192, 50 },
		{ 98, 193, 50 },
		{ 98, 194, 50 },
		{ 98, 195, 50 },
		{ 98, 196, 50 },
		{ 98, 197, 50 },
		{ 99, 159, 50 },
		{ 99, 160, 50 },
		{ 99, 161, 50 },
		{ 99, 162, 50 },
		{ 99, 163, 50 },
		{ 99, 164, 50 },
		{ 99, 165, 50 },
		{ 99, 166, 50 },
		{ 99, 167, 50 },
		{ 99, 168, 50 },
		{ 99, 169, 50 },
		{ 99, 170, 50 },
		{ 99, 171, 50 },
		{ 99, 172, 50 },
		{ 99, 173, 50 },
		{ 99, 174, 50 },
		{ 99, 175, 50 },
		{ 99, 176, 50 },
		{ 99, 177, 50 },
		{ 99, 178, 50 },
		{ 99, 179, 50 },
		{ 99, 180, 50 },
		{ 99, 181, 50 },
		{ 99, 182, 50 },
		{ 99, 183, 50 },
		{ 99, 184, 50 },
		{ 99, 185, 50 },
		{ 99, 186, 50 },
		{ 99, 187, 50 },
		{ 99, 188, 50 },
		{ 99, 189, 50 },
		{ 99, 190, 50 },
		{ 99, 191, 50 },
		{ 99, 192, 50 },
		{ 99, 193, 50 },
		{ 99, 194, 50 },
		{ 99, 195, 50 },
		{ 99, 196, 50 },
		{ 99, 197, 50 },
		{ 100, 159, 50 },
		{ 100, 160, 50 },
		{ 100, 161, 50 },
		{ 100, 162, 50 },
		{ 100, 163, 50 },
		{ 100, 164, 50 },
		{ 100, 165, 50 },
		{ 100, 166, 50 },
		{ 100, 167, 50 },
		{ 100, 168, 50 },
		{ 100, 169, 50 },
		{ 100, 170, 50 },
		{ 100, 171, 50 },
		{ 100, 172, 50 },
		{ 100, 173, 50 },
		{ 100, 174, 50 },
		{ 100, 175, 50 },
		{ 100, 176, 50 },
		{ 100, 177, 50 },
		{ 100, 178, 50 },
		{ 100, 179, 50 },
		{ 100, 180, 50 },
		{ 100, 181, 50 },
		{ 100, 182, 50 },
		{ 100, 183, 50 },
		{ 100, 184, 50 },
		{ 100, 185, 50 },
		{ 100, 186, 50 },
		{ 100, 187, 50 },
		{ 100, 188, 50 },
		{ 100, 189, 50 },
		{ 100, 190, 50 },
		{ 100, 191, 50 },
		{ 100, 192, 50 },
		{ 100, 193, 50 },
		{ 100, 194, 50 },
		{ 100, 195, 50 },
		{ 100, 196, 50 },
		{ 100, 197, 50 },
		{ 100, 198, 50 },
		{ 101, 160, 50 },
		{ 101, 161, 50 },
		{ 101, 162, 50 },
		{ 101, 163, 50 },
		{ 101, 164, 50 },
		{ 101, 165, 50 },
		{ 101, 166, 50 },
		{ 101, 167, 50 },
		{ 101, 168, 50 },
		{ 101, 169, 50 },
		{ 101, 170, 50 },
		{ 101, 171, 50 },
		{ 101, 172, 50 },
		{ 101, 173, 50 },
		{ 101, 174, 50 },
		{ 101, 175, 50 },
		{ 101, 176, 50 },
		{ 101, 177, 50 },
		{ 101, 178, 50 },
		{ 101, 179, 50 },
		{ 101, 180, 50 },
		{ 101, 181, 50 },
		{ 101, 182, 50 },
		{ 101, 183, 50 },
		{ 101, 184, 50 },
		{ 101, 185, 50 },
		{ 101, 186, 50 },
		{ 101, 187, 50 },
		{ 101, 188, 50 },
		{ 101, 189, 50 },
		{ 101, 190, 50 },
		{ 101, 191, 50 },
		{ 101, 192, 50 },
		{ 101, 193, 50 },
		{ 101, 194, 50 },
		{ 101, 195, 50 },
		{ 101, 196, 50 },
		{ 101, 197, 50 },
		{ 101, 198, 50 },
		{ 102, 161, 50 },
		{ 102, 162, 50 },
		{ 102, 163, 50 },
		{ 102, 164, 50 },
		{ 102, 165, 50 },
		{ 102, 166, 50 },
		{ 102, 167, 50 },
		{ 102, 168, 50 },
		{ 102, 169, 50 },
		{ 102, 170, 50 },
		{ 102, 171, 50 },
		{ 102, 172, 50 },
		{ 102, 173, 50 },
		{ 102, 174, 50 },
		{ 102, 175, 50 },
		{ 102, 176, 50 },
		{ 102, 177, 50 },
		{ 102, 178, 50 },
		{ 102, 179, 50 },
		{ 102, 180, 50 },
		{ 102, 181, 50 },
		{ 102, 182, 50 },
		{ 102, 183, 50 },
		{ 102, 184, 50 },
		{ 102, 185, 50 },
		{ 102, 186, 50 },
		{ 102, 187, 50 },
		{ 102, 188, 50 },
		{ 102, 189, 50 },
		{ 102, 190, 50 },
		{ 102, 191, 50 },
		{ 102, 192, 50 },
		{ 102, 193, 50 },
		{ 102, 194, 50 },
		{ 102, 195, 50 },
		{ 102, 196, 50 },
		{ 102, 197, 50 },
		{ 102, 198, 50 },
		{ 103, 161, 50 },
		{ 103, 162, 50 },
		{ 103, 163, 50 },
		{ 103, 164, 50 },
		{ 103, 165, 50 },
		{ 103, 166, 50 },
		{ 103, 167, 50 },
		{ 103, 168, 50 },
		{ 103, 169, 50 },
		{ 103, 170, 50 },
		{ 103, 171, 50 },
		{ 103, 172, 50 },
		{ 103, 173, 50 },
		{ 103, 174, 50 },
		{ 103, 175, 50 },
		{ 103, 176, 50 },
		{ 103, 177, 50 },
		{ 103, 178, 50 },
		{ 103, 179, 50 },
		{ 103, 180, 50 },
		{ 103, 181, 50 },
		{ 103, 182, 50 },
		{ 103, 183, 50 },
		{ 103, 184, 50 },
		{ 103, 185, 50 },
		{ 103, 186, 50 },
		{ 103, 187, 50 },
		{ 103, 188, 50 },
		{ 103, 189, 50 },
		{ 103, 190, 50 },
		{ 103, 191, 50 },
		{ 103, 192, 50 },
		{ 103, 193, 50 },
		{ 103, 194, 50 },
		{ 103, 195, 50 },
		{ 103, 196, 50 },
		{ 103, 197, 50 },
		{ 103, 198, 50 },
		{ 103, 199, 50 },
		{ 104, 162, 50 },
		{ 104, 163, 50 },
		{ 104, 164, 50 },
		{ 104, 165, 50 },
		{ 104, 166, 50 },
		{ 104, 167, 50 },
		{ 104, 168, 50 },
		{ 104, 169, 50 },
		{ 104, 170, 50 },
		{ 104, 171, 50 },
		{ 104, 172, 50 },
		{ 104, 173, 50 },
		{ 104, 174, 50 },
		{ 104, 175, 50 },
		{ 104, 176, 50 },
		{ 104, 177, 50 },
		{ 104, 178, 50 },
		{ 104, 179, 50 },
		{ 104, 180, 50 },
		{ 104, 181, 50 },
		{ 104, 182, 50 },
		{ 104, 183, 50 },
		{ 104, 184, 50 },
		{ 104, 185, 50 },
		{ 104, 186, 50 },
		{ 104, 187, 50 },
		{ 104, 188, 50 },
		{ 104, 189, 50 },
		{ 104, 190, 50 },
		{ 104, 191, 50 },
		{ 104, 192, 50 },
		{ 104, 193, 50 },
		{ 104, 194, 50 },
		{ 104, 195, 50 },
		{ 104, 196, 50 },
		{ 104, 197, 50 },
		{ 104, 198, 50 },
		{ 104, 199, 50 },
		{ 105, 163, 50 },
		{ 105, 164, 50 },
		{ 105, 165, 50 },
		{ 105, 166, 50 },
		{ 105, 167, 50 },
		{ 105, 168, 50 },
		{ 105, 169, 50 },
		{ 105, 170, 50 },
		{ 105, 171, 50 },
		{ 105, 172, 50 },
		{ 105, 173, 50 },
		{ 105, 174, 50 },
		{ 105, 175, 50 },
		{ 105, 176, 50 },
		{ 105, 177, 50 },
		{ 105, 178, 50 },
		{ 105, 179, 50 },
		{ 105, 180, 50 },
		{ 105, 181, 50 },
		{ 105, 182, 50 },
		{ 105, 183, 50 },
		{ 105, 184, 50 },
		{ 105, 185, 50 },
		{ 105, 186, 50 },
		{ 105, 187, 50 },
		{ 105, 188, 50 },
		{ 105, 189, 50 },
		{ 105, 190, 50 },
		{ 105, 191, 50 },
		{ 105, 192, 50 },
		{ 105, 193, 50 },
		{ 105, 194, 50 },
		{ 105, 195, 50 },
		{ 105, 196, 50 },
		{ 105, 197, 50 },
		{ 105, 198, 50 },
		{ 105, 199, 50 },
		{ 106, 163, 50 },
		{ 106, 164, 50 },
		{ 106, 165, 50 },
		{ 106, 166, 50 },
		{ 106, 167, 50 },
		{ 106, 168, 50 },
		{ 106, 169, 50 },
		{ 106, 170, 50 },
		{ 106, 171, 50 },
		{ 106, 172, 50 },
		{ 106, 173, 50 },
		{ 106, 174, 50 },
		{ 106, 175, 50 },
		{ 106, 176, 50 },
		{ 106, 177, 50 },
		{ 106, 178, 50 },
		{ 106, 179, 50 },
		{ 106, 180, 50 },
		{ 106, 181, 50 },
		{ 106, 182, 50 },
		{ 106, 183, 50 },
		{ 106, 184, 50 },
		{ 106, 185, 50 },
		{ 106, 186, 50 },
		{ 106, 187, 50 },
		{ 106, 188, 50 },
		{ 106, 189, 50 },
		{ 106, 190, 50 },
		{ 106, 191, 50 },
		{ 106, 192, 50 },
		{ 106, 193, 50 },
		{ 106, 194, 50 },
		{ 106, 195, 50 },
		{ 106, 196, 50 },
		{ 106, 197, 50 },
		{ 106, 198, 50 },
		{ 106, 199, 50 },
		{ 106, 200, 50 },
		{ 107, 164, 50 },
		{ 107, 165, 50 },
		{ 107, 166, 50 },
		{ 107, 167, 50 },
		{ 107, 168, 50 },
		{ 107, 169, 50 },
		{ 107, 170, 50 },
		{ 107, 171, 50 },
		{ 107, 172, 50 },
		{ 107, 173, 50 },
		{ 107, 174, 50 },
		{ 107, 175, 50 },
		{ 107, 176, 50 },
		{ 107, 177, 50 },
		{ 107, 178, 50 },
		{ 107, 179, 50 },
		{ 107, 180, 50 },
		{ 107, 181, 50 },
		{ 107, 182, 50 },
		{ 107, 183, 50 },
		{ 107, 184, 50 },
		{ 107, 185, 50 },
		{ 107, 186, 50 },
		{ 107, 187, 50 },
		{ 107, 188, 50 },
		{ 107, 189, 50 },
		{ 107, 190, 50 },
		{ 107, 191, 50 },
		{ 107, 192, 50 },
		{ 107, 193, 50 },
		{ 107, 194, 50 },
		{ 107, 195, 50 },
		{ 107, 196, 50 },
		{ 107, 197, 50 },
		{ 107, 198, 50 },
		{ 107, 199, 50 },
		{ 107, 200, 50 },
		{ 108, 164, 50 },
		{ 108, 165, 50 },
		{ 108, 166, 50 },
		{ 108, 167, 50 },
		{ 108, 168, 50 },
		{ 108, 169, 50 },
		{ 108, 170, 50 },
		{ 108, 171, 50 },
		{ 108, 172, 50 },
		{ 108, 173, 50 },
		{ 108, 174, 50 },
		{ 108, 175, 50 },
		{ 108, 176, 50 },
		{ 108, 177, 50 },
		{ 108, 178, 50 },
		{ 108, 179, 50 },
		{ 108, 180, 50 },
		{ 108, 181, 50 },
		{ 108, 182, 50 },
		{ 108, 183, 50 },
		{ 108, 184, 50 },
		{ 108, 185, 50 },
		{ 108, 186, 50 },
		{ 108, 187, 50 },
		{ 108, 188, 50 },
		{ 108, 189, 50 },
		{ 108, 190, 50 },
		{ 108, 191, 50 },
		{ 108, 192, 50 },
		{ 108, 193, 50 },
		{ 108, 194, 50 },
		{ 108, 195, 50 },
		{ 108, 196, 50 },
		{ 108, 197, 50 },
		{ 108, 198, 50 },
		{ 108, 199, 50 },
		{ 108, 200, 50 },
		{ 108, 201, 50 },
		{ 109, 165, 50 },
		{ 109, 166, 50 },
		{ 109, 167, 50 },
		{ 109, 168, 50 },
		{ 109, 169, 50 },
		{ 109, 170, 50 },
		{ 109, 171, 50 },
		{ 109, 172, 50 },
		{ 109, 173, 50 },
		{ 109, 174, 50 },
		{ 109, 175, 50 },
		{ 109, 176, 50 },
		{ 109, 177, 50 },
		{ 109, 178, 50 },
		{ 109, 179, 50 },
		{ 109, 180, 50 },
		{ 109, 181, 50 },
		{ 109, 182, 50 },
		{ 109, 183, 50 },
		{ 109, 184, 50 },
		{ 109, 185, 50 },
		{ 109, 186, 50 },
		{ 109, 187, 50 },
		{ 109, 188, 50 },
		{ 109, 189, 50 },
		{ 109, 190, 50 },
		{ 109, 191, 50 },
		{ 109, 192, 50 },
		{ 109, 193, 50 },
		{ 109, 194, 50 },
		{ 109, 195, 50 },
		{ 109, 196, 50 },
		{ 109, 197, 50 },
		{ 109, 198, 50 },
		{ 109, 199, 50 },
		{ 109, 200, 50 },
		{ 109, 201, 50 },
		{ 110, 166, 50 },
		{ 110, 167, 50 },
		{ 110, 168, 50 },
		{ 110, 169, 50 },
		{ 110, 170, 50 },
		{ 110, 171, 50 },
		{ 110, 172, 50 },
		{ 110, 173, 50 },
		{ 110, 174, 50 },
		{ 110, 175, 50 },
		{ 110, 176, 50 },
		{ 110, 177, 50 },
		{ 110, 178, 50 },
		{ 110, 179, 50 },
		{ 110, 180, 50 },
		{ 110, 181, 50 },
		{ 110, 182, 50 },
		{ 110, 183, 50 },
		{ 110, 184, 50 },
		{ 110, 185, 50 },
		{ 110, 186, 50 },
		{ 110, 187, 50 },
		{ 110, 188, 50 },
		{ 110, 189, 50 },
		{ 110, 190, 50 },
		{ 110, 191, 50 },
		{ 110, 192, 50 },
		{ 110, 193, 50 },
		{ 110, 194, 50 },
		{ 110, 195, 50 },
		{ 110, 196, 50 },
		{ 110, 197, 50 },
		{ 110, 198, 50 },
		{ 110, 199, 50 },
		{ 110, 200, 50 },
		{ 110, 201, 50 },
		{ 111, 166, 50 },
		{ 111, 167, 50 },
		{ 111, 168, 50 },
		{ 111, 169, 50 },
		{ 111, 170, 50 },
		{ 111, 171, 50 },
		{ 111, 172, 50 },
		{ 111, 173, 50 },
		{ 111, 174, 50 },
		{ 111, 175, 50 },
		{ 111, 176, 50 },
		{ 111, 177, 50 },
		{ 111, 178, 50 },
		{ 111, 179, 50 },
		{ 111, 180, 50 },
		{ 111, 181, 50 },
		{ 111, 182, 50 },
		{ 111, 183, 50 },
		{ 111, 184, 50 },
		{ 111, 185, 50 },
		{ 111, 186, 50 },
		{ 111, 187, 50 },
		{ 111, 188, 50 },
		{ 111, 189, 50 },
		{ 111, 190, 50 },
		{ 111, 191, 50 },
		{ 111, 192, 50 },
		{ 111, 193, 50 },
		{ 111, 194, 50 },
		{ 111, 195, 50 },
		{ 111, 196, 50 },
		{ 111, 197, 50 },
		{ 111, 198, 50 },
		{ 111, 199, 50 },
		{ 111, 200, 50 },
		{ 111, 201, 50 },
		{ 111, 202, 50 },
		{ 112, 167, 50 },
		{ 112, 168, 50 },
		{ 112, 169, 50 },
		{ 112, 170, 50 },
		{ 112, 171, 50 },
		{ 112, 172, 50 },
		{ 112, 173, 50 },
		{ 112, 174, 50 },
		{ 112, 175, 50 },
		{ 112, 176, 50 },
		{ 112, 177, 50 },
		{ 112, 178, 50 },
		{ 112, 179, 50 },
		{ 112, 180, 50 },
		{ 112, 181, 50 },
		{ 112, 182, 50 },
		{ 112, 183, 50 },
		{ 112, 184, 50 },
		{ 112, 185, 50 },
		{ 112, 186, 50 },
		{ 112, 187, 50 },
		{ 112, 188, 50 },
		{ 112, 189, 50 },
		{ 112, 190, 50 },
		{ 112, 191, 50 },
		{ 112, 192, 50 },
		{ 112, 193, 50 },
		{ 112, 194, 50 },
		{ 112, 195, 50 },
		{ 112, 196, 50 },
		{ 112, 197, 50 },
		{ 112, 198, 50 },
		{ 112, 199, 50 },
		{ 112, 200, 50 },
		{ 112, 201, 50 },
		{ 112, 202, 50 },
		{ 113, 168, 50 },
		{ 113, 169, 50 },
		{ 113, 170, 50 },
		{ 113, 171, 50 },
		{ 113, 172, 50 },
		{ 113, 173, 50 },
		{ 113, 174, 50 },
		{ 113, 175, 50 },
		{ 113, 176, 50 },
		{ 113, 177, 50 },
		{ 113, 178, 50 },
		{ 113, 179, 50 },
		{ 113, 180, 50 },
		{ 113, 181, 50 },
		{ 113, 182, 50 },
		{ 113, 183, 50 },
		{ 113, 184, 50 },
		{ 113, 185, 50 },
		{ 113, 186, 50 },
		{ 113, 187, 50 },
		{ 113, 188, 50 },
		{ 113, 189, 50 },
		{ 113, 190, 50 },
		{ 113, 191, 50 },
		{ 113, 192, 50 },
		{ 113, 193, 50 },
		{ 113, 194, 50 },
		{ 113, 195, 50 },
		{ 113, 196, 50 },
		{ 113, 197, 50 },
		{ 113, 198, 50 },
		{ 113, 199, 50 },
		{ 113, 200, 50 },
		{ 113, 201, 50 },
		{ 113, 202, 50 },
		{ 114, 168, 50 },
		{ 114, 169, 50 },
		{ 114, 170, 50 },
		{ 114, 171, 50 },
		{ 114, 172, 50 },
		{ 114, 173, 50 },
		{ 114, 174, 50 },
		{ 114, 175, 50 },
		{ 114, 176, 50 },
		{ 114, 177, 50 },
		{ 114, 178, 50 },
		{ 114, 179, 50 },
		{ 114, 180, 50 },
		{ 114, 181, 50 },
		{ 114, 182, 50 },
		{ 114, 183, 50 },
		{ 114, 184, 50 },
		{ 114, 185, 50 },
		{ 114, 186, 50 },
		{ 114, 187, 50 },
		{ 114, 188, 50 },
		{ 114, 189, 50 },
		{ 114, 190, 50 },
		{ 114, 191, 50 },
		{ 114, 192, 50 },
		{ 114, 193, 50 },
		{ 114, 194, 50 },
		{ 114, 195, 50 },
		{ 114, 196, 50 },
		{ 114, 197, 50 },
		{ 114, 198, 50 },
		{ 114, 199, 50 },
		{ 114, 200, 50 },
		{ 114, 201, 50 },
		{ 114, 202, 50 },
		{ 114, 203, 50 },
		{ 115, 169, 50 },
		{ 115, 170, 50 },
		{ 115, 171, 50 },
		{ 115, 172, 50 },
		{ 115, 173, 50 },
		{ 115, 174, 50 },
		{ 115, 175, 50 },
		{ 115, 176, 50 },
		{ 115, 177, 50 },
		{ 115, 178, 50 },
		{ 115, 179, 50 },
		{ 115, 180, 50 },
		{ 115, 181, 50 },
		{ 115, 182, 50 },
		{ 115, 183, 50 },
		{ 115, 184, 50 },
		{ 115, 185, 50 },
		{ 115, 186, 50 },
		{ 115, 187, 50 },
		{ 115, 188, 50 },
		{ 115, 189, 50 },
		{ 115, 190, 50 },
		{ 115, 191, 50 },
		{ 115, 192, 50 },
		{ 115, 193, 50 },
		{ 115, 194, 50 },
		{ 115, 195, 50 },
		{ 115, 196, 50 },
		{ 115, 197, 50 },
		{ 115, 198, 50 },
		{ 115, 199, 50 },
		{ 115, 200, 50 },
		{ 115, 201, 50 },
		{ 115, 202, 50 },
		{ 115, 203, 50 },
		{ 116, 169, 50 },
		{ 116, 170, 50 },
		{ 116, 171, 50 },
		{ 116, 172, 50 },
		{ 116, 173, 50 },
		{ 116, 174, 50 },
		{ 116, 175, 50 },
		{ 116, 176, 50 },
		{ 116, 177, 50 },
		{ 116, 178, 50 },
		{ 116, 179, 50 },
		{ 116, 180, 50 },
		{ 116, 181, 50 },
		{ 116, 182, 50 },
		{ 116, 183, 50 },
		{ 116, 184, 50 },
		{ 116, 185, 50 },
		{ 116, 186, 50 },
		{ 116, 187, 50 },
		{ 116, 188, 50 },
		{ 116, 189, 50 },
		{ 116, 190, 50 },
		{ 116, 191, 50 },
		{ 116, 192, 50 },
		{ 116, 193, 50 },
		{ 116, 194, 50 },
		{ 116, 195, 50 },
		{ 116, 196, 50 },
		{ 116, 197, 50 },
		{ 116, 198, 50 },
		{ 116, 199, 50 },
		{ 116, 200, 50 },
		{ 116, 201, 50 },
		{ 116, 202, 50 },
		{ 116, 203, 50 },
		{ 116, 204, 50 },
		{ 117, 170, 50 },
		{ 117, 171, 50 },
		{ 117, 172, 50 },
		{ 117, 173, 50 },
		{ 117, 174, 50 },
		{ 117, 175, 50 },
		{ 117, 176, 50 },
		{ 117, 177, 50 },
		{ 117, 178, 50 },
		{ 117, 179, 50 },
		{ 117, 180, 50 },
		{ 117, 181, 50 },
		{ 117, 182, 50 },
		{ 117, 183, 50 },
		{ 117, 184, 50 },
		{ 117, 185, 50 },
		{ 117, 186, 50 },
		{ 117, 187, 50 },
		{ 117, 188, 50 },
		{ 117, 189, 50 },
		{ 117, 190, 50 },
		{ 117, 191, 50 },
		{ 117, 192, 50 },
		{ 117, 193, 50 },
		{ 117, 194, 50 },
		{ 117, 195, 50 },
		{ 117, 196, 50 },
		{ 117, 197, 50 },
		{ 117, 198, 50 },
		{ 117, 199, 50 },
		{ 117, 200, 50 },
		{ 117, 201, 50 },
		{ 117, 202, 50 },
		{ 117, 203, 50 },
		{ 117, 204, 50 },
		{ 118, 171, 50 },
		{ 118, 172, 50 },
		{ 118, 173, 50 },
		{ 118, 174, 50 },
		{ 118, 175, 50 },
		{ 118, 176, 50 },
		{ 118, 177, 50 },
		{ 118, 178, 50 },
		{ 118, 179, 50 },
		{ 118, 180, 50 },
		{ 118, 181, 50 },
		{ 118, 182, 50 },
		{ 118, 183, 50 },
		{ 118, 184, 50 },
		{ 118, 185, 50 },
		{ 118, 186, 50 },
		{ 118, 187, 50 },
		{ 118, 188, 50 },
		{ 118, 189, 50 },
		{ 118, 190, 50 },
		{ 118, 191, 50 },
		{ 118, 192, 50 },
		{ 118, 193, 50 },
		{ 118, 194, 50 },
		{ 118, 195, 50 },
		{ 118, 196, 50 },
		{ 118, 197, 50 },
		{ 118, 198, 50 },
		{ 118, 199, 50 },
		{ 118, 200, 50 },
		{ 118, 201, 50 },
		{ 118, 202, 50 },
		{ 118, 203, 50 },
		{ 118, 204, 50 },
		{ 119, 171, 50 },
		{ 119, 172, 50 },
		{ 119, 173, 50 },
		{ 119, 174, 50 },
		{ 119, 175, 50 },
		{ 119, 176, 50 },
		{ 119, 177, 50 },
		{ 119, 178, 50 },
		{ 119, 179, 50 },
		{ 119, 180, 50 },
		{ 119, 181, 50 },
		{ 119, 182, 50 },
		{ 119, 183, 50 },
		{ 119, 184, 50 },
		{ 119, 185, 50 },
		{ 119, 186, 50 },
		{ 119, 187, 50 },
		{ 119, 188, 50 },
		{ 119, 189, 50 },
		{ 119, 190, 50 },
		{ 119, 191, 50 },
		{ 119, 192, 50 },
		{ 119, 193, 50 },
		{ 119, 194, 50 },
		{ 119, 195, 50 },
		{ 119, 196, 50 },
		{ 119, 197, 50 },
		{ 119, 198, 50 },
		{ 119, 199, 50 },
		{ 119, 200, 50 },
		{ 119, 201, 50 },
		{ 119, 202, 50 },
		{ 119, 203, 50 },
		{ 119, 204, 50 },
		{ 119, 205, 50 },
		{ 120, 172, 50 },
		{ 120, 173, 50 },
		{ 120, 174, 50 },
		{ 120, 175, 50 },
		{ 120, 176, 50 },
		{ 120, 177, 50 },
		{ 120, 178, 50 },
		{ 120, 179, 50 },
		{ 120, 180, 50 },
		{ 120, 181, 50 },
		{ 120, 182, 50 },
		{ 120, 183, 50 },
		{ 120, 184, 50 },
		{ 120, 185, 50 },
		{ 120, 186, 50 },
		{ 120, 187, 50 },
		{ 120, 188, 50 },
		{ 120, 189, 50 },
		{ 120, 190, 50 },
		{ 120, 191, 50 },
		{ 120, 192, 50 },
		{ 120, 193, 50 },
		{ 120, 194, 50 },
		{ 120, 195, 50 },
		{ 120, 196, 50 },
		{ 120, 197, 50 },
		{ 120, 198, 50 },
		{ 120, 199, 50 },
		{ 120, 200, 50 },
		{ 120, 201, 50 },
		{ 120, 202, 50 },
		{ 120, 203, 50 },
		{ 120, 204, 50 },
		{ 120, 205, 50 },
		{ 121, 173, 50 },
		{ 121, 174, 50 },
		{ 121, 175, 50 },
		{ 121, 176, 50 },
		{ 121, 177, 50 },
		{ 121, 178, 50 },
		{ 121, 179, 50 },
		{ 121, 180, 50 },
		{ 121, 181, 50 },
		{ 121, 182, 50 },
		{ 121, 183, 50 },
		{ 121, 184, 50 },
		{ 121, 185, 50 },
		{ 121, 186, 50 },
		{ 121, 187, 50 },
		{ 121, 188, 50 },
		{ 121, 189, 50 },
		{ 121, 190, 50 },
		{ 121, 191, 50 },
		{ 121, 192, 50 },
		{ 121, 193, 50 },
		{ 121, 194, 50 },
		{ 121, 195, 50 },
		{ 121, 196, 50 },
		{ 121, 197, 50 },
		{ 121, 198, 50 },
		{ 121, 199, 50 },
		{ 121, 200, 50 },
		{ 121, 201, 50 },
		{ 121, 202, 50 },
		{ 121, 203, 50 },
		{ 121, 204, 50 },
		{ 121, 205, 50 },
		{ 122, 173, 50 },
		{ 122, 174, 50 },
		{ 122, 175, 50 },
		{ 122, 176, 50 },
		{ 122, 177, 50 },
		{ 122, 178, 50 },
		{ 122, 179, 50 },
		{ 122, 180, 50 },
		{ 122, 181, 50 },
		{ 122, 182, 50 },
		{ 122, 183, 50 },
		{ 122, 184, 50 },
		{ 122, 185, 50 },
		{ 122, 186, 50 },
		{ 122, 187, 50 },
		{ 122, 188, 50 },
		{ 122, 189, 50 },
		{ 122, 190, 50 },
		{ 122, 191, 50 },
		{ 122, 192, 50 },
		{ 122, 193, 50 },
		{ 122, 194, 50 },
		{ 122, 195, 50 },
		{ 122, 196, 50 },
		{ 122, 197, 50 },
		{ 122, 198, 50 },
		{ 122, 199, 50 },
		{ 122, 200, 50 },
		{ 122, 201, 50 },
		{ 122, 202, 50 },
		{ 122, 203, 50 },
		{ 122, 204, 50 },
		{ 122, 205, 50 },
		{ 122, 206, 50 },
		{ 123, 174, 50 },
		{ 123, 175, 50 },
		{ 123, 176, 50 },
		{ 123, 177, 50 },
		{ 123, 178, 50 },
		{ 123, 179, 50 },
		{ 123, 180, 50 },
		{ 123, 181, 50 },
		{ 123, 182, 50 },
		{ 123, 183, 50 },
		{ 123, 184, 50 },
		{ 123, 185, 50 },
		{ 123, 186, 50 },
		{ 123, 187, 50 },
		{ 123, 188, 50 },
		{ 123, 189, 50 },
		{ 123, 190, 50 },
		{ 123, 191, 50 },
		{ 123, 192, 50 },
		{ 123, 193, 50 },
		{ 123, 194, 50 },
		{ 123, 195, 50 },
		{ 123, 196, 50 },
		{ 123, 197, 50 },
		{ 123, 198, 50 },
		{ 123, 199, 50 },
		{ 123, 200, 50 },
		{ 123, 201, 50 },
		{ 123, 202, 50 },
		{ 123, 203, 50 },
		{ 123, 204, 50 },
		{ 123, 205, 50 },
		{ 123, 206, 50 },
		{ 124, 174, 50 },
		{ 124, 175, 50 },
		{ 124, 176, 50 },
		{ 124, 177, 50 },
		{ 124, 178, 50 },
		{ 124, 179, 50 },
		{ 124, 180, 50 },
		{ 124, 181, 50 },
		{ 124, 182, 50 },
		{ 124, 183, 50 },
		{ 124, 184, 50 },
		{ 124, 185, 50 },
		{ 124, 186, 50 },
		{ 124, 187, 50 },
		{ 124, 188, 50 },
		{ 124, 189, 50 },
		{ 124, 190, 50 },
		{ 124, 191, 50 },
		{ 124, 192, 50 },
		{ 124, 193, 50 },
		{ 124, 194, 50 },
		{ 124, 195, 50 },
		{ 124, 196, 50 },
		{ 124, 197, 50 },
		{ 124, 198, 50 },
		{ 124, 199, 50 },
		{ 124, 200, 50 },
		{ 124, 201, 50 },
		{ 124, 202, 50 },
		{ 124, 203, 50 },
		{ 124, 204, 50 },
		{ 124, 205, 50 },
		{ 124, 206, 50 },
		{ 124, 207, 50 },
		{ 125, 175, 50 },
		{ 125, 176, 50 },
		{ 125, 177, 50 },
		{ 125, 178, 50 },
		{ 125, 179, 50 },
		{ 125, 180, 50 },
		{ 125, 181, 50 },
		{ 125, 182, 50 },
		{ 125, 183, 50 },
		{ 125, 184, 50 },
		{ 125, 185, 50 },
		{ 125, 186, 50 },
		{ 125, 187, 50 },
		{ 125, 188, 50 },
		{ 125, 189, 50 },
		{ 125, 190, 50 },
		{ 125, 191, 50 },
		{ 125, 192, 50 },
		{ 125, 193, 50 },
		{ 125, 194, 50 },
		{ 125, 195, 50 },
		{ 125, 196, 50 },
		{ 125, 197, 50 },
		{ 125, 198, 50 },
		{ 125, 199, 50 },
		{ 125, 200, 50 },
		{ 125, 201, 50 },
		{ 125, 202, 50 },
		{ 125, 203, 50 },
		{ 125, 204, 50 },
		{ 125, 205, 50 },
		{ 125, 206, 50 },
		{ 125, 207, 50 },
		{ 126, 176, 50 },
		{ 126, 177, 50 },
		{ 126, 178, 50 },
		{ 126, 179, 50 },
		{ 126, 180, 50 },
		{ 126, 181, 50 },
		{ 126, 182, 50 },
		{ 126, 183, 50 },
		{ 126, 184, 50 },
		{ 126, 185, 50 },
		{ 126, 186, 50 },
		{ 126, 187, 50 },
		{ 126, 188, 50 },
		{ 126, 189, 50 },
		{ 126, 190, 50 },
		{ 126, 191, 50 },
		{ 126, 192, 50 },
		{ 126, 193, 50 },
		{ 126, 194, 50 },
		{ 126, 195, 50 },
		{ 126, 196, 50 },
		{ 126, 197, 50 },
		{ 126, 198, 50 },
		{ 126, 199, 50 },
		{ 126, 200, 50 },
		{ 126, 201, 50 },
		{ 126, 202, 50 },
		{ 126, 203, 50 },
		{ 126, 204, 50 },
		{ 126, 205, 50 },
		{ 126, 206, 50 },
		{ 126, 207, 50 },
		{ 127, 176, 50 },
		{ 127, 177, 50 },
		{ 127, 178, 50 },
		{ 127, 179, 50 },
		{ 127, 180, 50 },
		{ 127, 181, 50 },
		{ 127, 182, 50 },
		{ 127, 183, 50 },
		{ 127, 184, 50 },
		{ 127, 185, 50 },
		{ 127, 186, 50 },
		{ 127, 187, 50 },
		{ 127, 188, 50 },
		{ 127, 189, 50 },
		{ 127, 190, 50 },
		{ 127, 191, 50 },
		{ 127, 192, 50 },
		{ 127, 193, 50 },
		{ 127, 194, 50 },
		{ 127, 195, 50 },
		{ 127, 196, 50 },
		{ 127, 197, 50 },
		{ 127, 198, 50 },
		{ 127, 199, 50 },
		{ 127, 200, 50 },
		{ 127, 201, 50 },
		{ 127, 202, 50 },
		{ 127, 203, 50 },
		{ 127, 204, 50 },
		{ 127, 205, 50 },
		{ 127, 206, 50 },
		{ 127, 207, 50 },
		{ 127, 208, 50 },
		{ 128, 177, 50 },
		{ 128, 178, 50 },
		{ 128, 179, 50 },
		{ 128, 180, 50 },
		{ 128, 181, 50 },
		{ 128, 182, 50 },
		{ 128, 183, 50 },
		{ 128, 184, 50 },
		{ 128, 185, 50 },
		{ 128, 186, 50 },
		{ 128, 187, 50 },
		{ 128, 188, 50 },
		{ 128, 189, 50 },
		{ 128, 190, 50 },
		{ 128, 191, 50 },
		{ 128, 192, 50 },
		{ 128, 193, 50 },
		{ 128, 194, 50 },
		{ 128, 195, 50 },
		{ 128, 196, 50 },
		{ 128, 197, 50 },
		{ 128, 198, 50 },
		{ 128, 199, 50 },
		{ 128, 200, 50 },
		{ 128, 201, 50 },
		{ 128, 202, 50 },
		{ 128, 203, 50 },
		{ 128, 204, 50 },
		{ 128, 205, 50 },
		{ 128, 206, 50 },
		{ 128, 207, 50 },
		{ 128, 208, 50 },
		{ 129, 178, 50 },
		{ 129, 179, 50 },
		{ 129, 180, 50 },
		{ 129, 181, 50 },
		{ 129, 182, 50 },
		{ 129, 183, 50 },
		{ 129, 184, 50 },
		{ 129, 185, 50 },
		{ 129, 186, 50 },
		{ 129, 187, 50 },
		{ 129, 188, 50 },
		{ 129, 189, 50 },
		{ 129, 190, 50 },
		{ 129, 191, 50 },
		{ 129, 192, 50 },
		{ 129, 193, 50 },
		{ 129, 194, 50 },
		{ 129, 195, 50 },
		{ 129, 196, 50 },
		{ 129, 197, 50 },
		{ 129, 198, 50 },
		{ 129, 199, 50 },
		{ 129, 200, 50 },
		{ 129, 201, 50 },
		{ 129, 202, 50 },
		{ 129, 203, 50 },
		{ 129, 204, 50 },
		{ 129, 205, 50 },
		{ 129, 206, 50 },
		{ 129, 207, 50 },
		{ 129, 208, 50 },
		{ 130, 178, 50 },
		{ 130, 179, 50 },
		{ 130, 180, 50 },
		{ 130, 181, 50 },
		{ 130, 182, 50 },
		{ 130, 183, 50 },
		{ 130, 184, 50 },
		{ 130, 185, 50 },
		{ 130, 186, 50 },
		{ 130, 187, 50 },
		{ 130, 188, 50 },
		{ 130, 189, 50 },
		{ 130, 190, 50 },
		{ 130, 191, 50 },
		{ 130, 192, 50 },
		{ 130, 193, 50 },
		{ 130, 194, 50 },
		{ 130, 195, 50 },
		{ 130, 196, 50 },
		{ 130, 197, 50 },
		{ 130, 198, 50 },
		{ 130, 199, 50 },
		{ 130, 200, 50 },
		{ 130, 201, 50 },
		{ 130, 202, 50 },
		{ 130, 203, 50 },
		{ 130, 204, 50 },
		{ 130, 205, 50 },
		{ 130, 206, 50 },
		{ 130, 207, 50 },
		{ 130, 208, 50 },
		{ 130, 209, 50 },
		{ 131, 179, 50 },
		{ 131, 180, 50 },
		{ 131, 181, 50 },
		{ 131, 182, 50 },
		{ 131, 183, 50 },
		{ 131, 184, 50 },
		{ 131, 185, 50 },
		{ 131, 186, 50 },
		{ 131, 187, 50 },
		{ 131, 188, 50 },
		{ 131, 189, 50 },
		{ 131, 190, 50 },
		{ 131, 191, 50 },
		{ 131, 192, 50 },
		{ 131, 193, 50 },
		{ 131, 194, 50 },
		{ 131, 195, 50 },
		{ 131, 196, 50 },
		{ 131, 197, 50 },
		{ 131, 198, 50 },
		{ 131, 199, 50 },
		{ 131, 200, 50 },
		{ 131, 201, 50 },
		{ 131, 202, 50 },
		{ 131, 203, 50 },
		{ 131, 204, 50 },
		{ 131, 205, 50 },
		{ 131, 206, 50 },
		{ 131, 207, 50 },
		{ 131, 208, 50 },
		{ 131, 209, 50 },
		{ 132, 179, 50 },
		{ 132, 180, 50 },
		{ 132, 181, 50 },
		{ 132, 182, 50 },
		{ 132, 183, 50 },
		{ 132, 184, 50 },
		{ 132, 185, 50 },
		{ 132, 186, 50 },
		{ 132, 187, 50 },
		{ 132, 188, 50 },
		{ 132, 189, 50 },
		{ 132, 190, 50 },
		{ 132, 191, 50 },
		{ 132, 192, 50 },
		{ 132, 193, 50 },
		{ 132, 194, 50 },
		{ 132, 195, 50 },
		{ 132, 196, 50 },
		{ 132, 197, 50 },
		{ 132, 198, 50 },
		{ 132, 199, 50 },
		{ 132, 200, 50 },
		{ 132, 201, 50 },
		{ 132, 202, 50 },
		{ 132, 203, 50 },
		{ 132, 204, 50 },
		{ 132, 205, 50 },
		{ 132, 206, 50 },
		{ 132, 207, 50 },
		{ 132, 208, 50 },
		{ 132, 209, 50 },
		{ 132, 210, 50 },
		{ 133, 180, 50 },
		{ 133, 181, 50 },
		{ 133, 182, 50 },
		{ 133, 183, 50 },
		{ 133, 184, 50 },
		{ 133, 185, 50 },
		{ 133, 186, 50 },
		{ 133, 187, 50 },
		{ 133, 188, 50 },
		{ 133, 189, 50 },
		{ 133, 190, 50 },
		{ 133, 191, 50 },
		{ 133, 192, 50 },
		{ 133, 193, 50 },
		{ 133, 194, 50 },
		{ 133, 195, 50 },
		{ 133, 196, 50 },
		{ 133, 197, 50 },
		{ 133, 198, 50 },
		{ 133, 199, 50 },
		{ 133, 200, 50 },
		{ 133, 201, 50 },
		{ 133, 202, 50 },
		{ 133, 203, 50 },
		{ 133, 204, 50 },
		{ 133, 205, 50 },
		{ 133, 206, 50 },
		{ 133, 207, 50 },
		{ 133, 208, 50 },
		{ 133, 209, 50 },
		{ 133, 210, 50 },
		{ 134, 181, 50 },
		{ 134, 182, 50 },
		{ 134, 183, 50 },
		{ 134, 184, 50 },
		{ 134, 185, 50 },
		{ 134, 186, 50 },
		{ 134, 187, 50 },
		{ 134, 188, 50 },
		{ 134, 189, 50 },
		{ 134, 190, 50 },
		{ 134, 191, 50 },
		{ 134, 192, 50 },
		{ 134, 193, 50 },
		{ 134, 194, 50 },
		{ 134, 195, 50 },
		{ 134, 196, 50 },
		{ 134, 197, 50 },
		{ 134, 198, 50 },
		{ 134, 199, 50 },
		{ 134, 200, 50 },
		{ 134, 201, 50 },
		{ 134, 202, 50 },
		{ 134, 203, 50 },
		{ 134, 204, 50 },
		{ 134, 205, 50 },
		{ 134, 206, 50 },
		{ 134, 207, 50 },
		{ 134, 208, 50 },
		{ 134, 209, 50 },
		{ 134, 210, 50 },
		{ 135, 181, 50 },
		{ 135, 182, 50 },
		{ 135, 183, 50 },
		{ 135, 184, 50 },
		{ 135, 185, 50 },
		{ 135, 186, 50 },
		{ 135, 187, 50 },
		{ 135, 188, 50 },
		{ 135, 189, 50 },
		{ 135, 190, 50 },
		{ 135, 191, 50 },
		{ 135, 192, 50 },
		{ 135, 193, 50 },
		{ 135, 194, 50 },
		{ 135, 195, 50 },
		{ 135, 196, 50 },
		{ 135, 197, 50 },
		{ 135, 198, 50 },
		{ 135, 199, 50 },
		{ 135, 200, 50 },
		{ 135, 201, 50 },
		{ 135, 202, 50 },
		{ 135, 203, 50 },
		{ 135, 204, 50 },
		{ 135, 205, 50 },
		{ 135, 206, 50 },
		{ 135, 207, 50 },
		{ 135, 208, 50 },
		{ 135, 209, 50 },
		{ 135, 210, 50 },
		{ 135, 211, 50 },
		{ 136, 182, 50 },
		{ 136, 183, 50 },
		{ 136, 184, 50 },
		{ 136, 185, 50 },
		{ 136, 186, 50 },
		{ 136, 187, 50 },
		{ 136, 188, 50 },
		{ 136, 189, 50 },
		{ 136, 190, 50 },
		{ 136, 191, 50 },
		{ 136, 192, 50 },
		{ 136, 193, 50 },
		{ 136, 194, 50 },
		{ 136, 195, 50 },
		{ 136, 196, 50 },
		{ 136, 197, 50 },
		{ 136, 198, 50 },
		{ 136, 199, 50 },
		{ 136, 200, 50 },
		{ 136, 201, 50 },
		{ 136, 202, 50 },
		{ 136, 203, 50 },
		{ 136, 204, 50 },
		{ 136, 205, 50 },
		{ 136, 206, 50 },
		{ 136, 207, 50 },
		{ 136, 208, 50 },
		{ 136, 209, 50 },
		{ 136, 210, 50 },
		{ 136, 211, 50 },
		{ 137, 183, 50 },
		{ 137, 184, 50 },
		{ 137, 185, 50 },
		{ 137, 186, 50 },
		{ 137, 187, 50 },
		{ 137, 188, 50 },
		{ 137, 189, 50 },
		{ 137, 190, 50 },
		{ 137, 191, 50 },
		{ 137, 192, 50 },
		{ 137, 193, 50 },
		{ 137, 194, 50 },
		{ 137, 195, 50 },
		{ 137, 196, 50 },
		{ 137, 197, 50 },
		{ 137, 198, 50 },
		{ 137, 199, 50 },
		{ 137, 200, 50 },
		{ 137, 201, 50 },
		{ 137, 202, 50 },
		{ 137, 203, 50 },
		{ 137, 204, 50 },
		{ 137, 205, 50 },
		{ 137, 206, 50 },
		{ 137, 207, 50 },
		{ 137, 208, 50 },
		{ 137, 209, 50 },
		{ 137, 210, 50 },
		{ 137, 211, 50 },
		{ 138, 183, 50 },
		{ 138, 184, 50 },
		{ 138, 185, 50 },
		{ 138, 186, 50 },
		{ 138, 187, 50 },
		{ 138, 188, 50 },
		{ 138, 189, 50 },
		{ 138, 190, 50 },
		{ 138, 191, 50 },
		{ 138, 192, 50 },
		{ 138, 193, 50 },
		{ 138, 194, 50 },
		{ 138, 195, 50 },
		{ 138, 196, 50 },
		{ 138, 197, 50 },
		{ 138, 198, 50 },
		{ 138, 199, 50 },
		{ 138, 200, 50 },
		{ 138, 201, 50 },
		{ 138, 202, 50 },
		{ 138, 203, 50 },
		{ 138, 204, 50 },
		{ 138, 205, 50 },
		{ 138, 206, 50 },
		{ 138, 207, 50 },
		{ 138, 208, 50 },
		{ 138, 209, 50 },
		{ 138, 210, 50 },
		{ 138, 211, 50 },
		{ 138, 212, 50 },
		{ 139, 184, 50 },
		{ 139, 185, 50 },
		{ 139, 186, 50 },
		{ 139, 187, 50 },
		{ 139, 188, 50 },
		{ 139, 189, 50 },
		{ 139, 190, 50 },
		{ 139, 191, 50 },
		{ 139, 192, 50 },
		{ 139, 193, 50 },
		{ 139, 194, 50 },
		{ 139, 195, 50 },
		{ 139, 196, 50 },
		{ 139, 197, 50 },
		{ 139, 198, 50 },
		{ 139, 199, 50 },
		{ 139, 200, 50 },
		{ 139, 201, 50 },
		{ 139, 202, 50 },
		{ 139, 203, 50 },
		{ 139, 204, 50 },
		{ 139, 205, 50 },
		{ 139, 206, 50 },
		{ 139, 207, 50 },
		{ 139, 208, 50 },
		{ 139, 209, 50 },
		{ 139, 210, 50 },
		{ 139, 211, 50 },
		{ 139, 212, 50 },
		{ 140, 184, 50 },
		{ 140, 185, 50 },
		{ 140, 186, 50 },
		{ 140, 187, 50 },
		{ 140, 188, 50 },
		{ 140, 189, 50 },
		{ 140, 190, 50 },
		{ 140, 191, 50 },
		{ 140, 192, 50 },
		{ 140, 193, 50 },
		{ 140, 194, 50 },
		{ 140, 195, 50 },
		{ 140, 196, 50 },
		{ 140, 197, 50 },
		{ 140, 198, 50 },
		{ 140, 199, 50 },
		{ 140, 200, 50 },
		{ 140, 201, 50 },
		{ 140, 202, 50 },
		{ 140, 203, 50 },
		{ 140, 204, 50 },
		{ 140, 205, 50 },
		{ 140, 206, 50 },
		{ 140, 207, 50 },
		{ 140, 208, 50 },
		{ 140, 209, 50 },
		{ 140, 210, 50 },
		{ 140, 211, 50 },
		{ 140, 212, 50 },
		{ 140, 213, 50 },
		{ 141, 185, 50 },
		{ 141, 186, 50 },
		{ 141, 187, 50 },
		{ 141, 188, 50 },
		{ 141, 189, 50 },
		{ 141, 190, 50 },
		{ 141, 191, 50 },
		{ 141, 192, 50 },
		{ 141, 193, 50 },
		{ 141, 194, 50 },
		{ 141, 195, 50 },
		{ 141, 196, 50 },
		{ 141, 197, 50 },
		{ 141, 198, 50 },
		{ 141, 199, 50 },
		{ 141, 200, 50 },
		{ 141, 201, 50 },
		{ 141, 202, 50 },
		{ 141, 203, 50 },
		{ 141, 204, 50 },
		{ 141, 205, 50 },
		{ 141, 206, 50 },
		{ 141, 207, 50 },
		{ 141, 208, 50 },
		{ 141, 209, 50 },
		{ 141, 210, 50 },
		{ 141, 211, 50 },
		{ 141, 212, 50 },
		{ 141, 213, 50 },
		{ 142, 186, 50 },
		{ 142, 187, 50 },
		{ 142, 188, 50 },
		{ 142, 189, 50 },
		{ 142, 190, 50 },
		{ 142, 191, 50 },
		{ 142, 192, 50 },
		{ 142, 193, 50 },
		{ 142, 194, 50 },
		{ 142, 195, 50 },
		{ 142, 196, 50 },
		{ 142, 197, 50 },
		{ 142, 198, 50 },
		{ 142, 199, 50 },
		{ 142, 200, 50 },
		{ 142, 201, 50 },
		{ 142, 202, 50 },
		{ 142, 203, 50 },
		{ 142, 204, 50 },
		{ 142, 205, 50 },
		{ 142, 206, 50 },
		{ 142, 207, 50 },
		{ 142, 208, 50 },
		{ 142, 209, 50 },
		{ 142, 210, 50 },
		{ 142, 211, 50 },
		{ 142, 212, 50 },
		{ 142, 213, 50 },
		{ 143, 186, 50 },
		{ 143, 187, 50 },
		{ 143, 188, 50 },
		{ 143, 189, 50 },
		{ 143, 190, 50 },
		{ 143, 191, 50 },
		{ 143, 192, 50 },
		{ 143, 193, 50 },
		{ 143, 194, 50 },
		{ 143, 195, 50 },
		{ 143, 196, 50 },
		{ 143, 197, 50 },
		{ 143, 198, 50 },
		{ 143, 199, 50 },
		{ 143, 200, 50 },
		{ 143, 201, 50 },
		{ 143, 202, 50 },
		{ 143, 203, 50 },
		{ 143, 204, 50 },
		{ 143, 205, 50 },
		{ 143, 206, 50 },
		{ 143, 207, 50 },
		{ 143, 208, 50 },
		{ 143, 209, 50 },
		{ 143, 210, 50 },
		{ 143, 211, 50 },
		{ 143, 212, 50 },
		{ 143, 213, 50 },
		{ 143, 214, 50 },
		{ 144, 187, 50 },
		{ 144, 188, 50 },
		{ 144, 189, 50 },
		{ 144, 190, 50 },
		{ 144, 191, 50 },
		{ 144, 192, 50 },
		{ 144, 193, 50 },
		{ 144, 194, 50 },
		{ 144, 195, 50 },
		{ 144, 196, 50 },
		{ 144, 197, 50 },
		{ 144, 198, 50 },
		{ 144, 199, 50 },
		{ 144, 200, 50 },
		{ 144, 201, 50 },
		{ 144, 202, 50 },
		{ 144, 203, 50 },
		{ 144, 204, 50 },
		{ 144, 205, 50 },
		{ 144, 206, 50 },
		{ 144, 207, 50 },
		{ 144, 208, 50 },
		{ 144, 209, 50 },
		{ 144, 210, 50 },
		{ 144, 211, 50 },
		{ 144, 212, 50 },
		{ 144, 213, 50 },
		{ 144, 214, 50 },
		{ 145, 188, 50 },
		{ 145, 189, 50 },
		{ 145, 190, 50 },
		{ 145, 191, 50 },
		{ 145, 192, 50 },
		{ 145, 193, 50 },
		{ 145, 194, 50 },
		{ 145, 195, 50 },
		{ 145, 196, 50 },
		{ 145, 197, 50 },
		{ 145, 198, 50 },
		{ 145, 199, 50 },
		{ 145, 200, 50 },
		{ 145, 201, 50 },
		{ 145, 202, 50 },
		{ 145, 203, 50 },
		{ 145, 204, 50 },
		{ 145, 205, 50 },
		{ 145, 206, 50 },
		{ 145, 207, 50 },
		{ 145, 208, 50 },
		{ 145, 209, 50 },
		{ 145, 210, 50 },
		{ 145, 211, 50 },
		{ 145, 212, 50 },
		{ 145, 213, 50 },
		{ 145, 214, 50 },
		{ 146, 188, 50 },
		{ 146, 189, 50 },
		{ 146, 190, 50 },
		{ 146, 191, 50 },
		{ 146, 192, 50 },
		{ 146, 193, 50 },
		{ 146, 194, 50 },
		{ 146, 195, 50 },
		{ 146, 196, 50 },
		{ 146, 197, 50 },
		{ 146, 198, 50 },
		{ 146, 199, 50 },
		{ 146, 200, 50 },
		{ 146, 201, 50 },
		{ 146, 202, 50 },
		{ 146, 203, 50 },
		{ 146, 204, 50 },
		{ 146, 205, 50 },
		{ 146, 206, 50 },
		{ 146, 207, 50 },
		{ 146, 208, 50 },
		{ 146, 209, 50 },
		{ 146, 210, 50 },
		{ 146, 211, 50 },
		{ 146, 212, 50 },
		{ 146, 213, 50 },
		{ 146, 214, 50 },
		{ 146, 215, 50 },
		{ 147, 189, 50 },
		{ 147, 190, 50 },
		{ 147, 191, 50 },
		{ 147, 192, 50 },
		{ 147, 193, 50 },
		{ 147, 194, 50 },
		{ 147, 195, 50 },
		{ 147, 196, 50 },
		{ 147, 197, 50 },
		{ 147, 198, 50 },
		{ 147, 199, 50 },
		{ 147, 200, 50 },
		{ 147, 201, 50 },
		{ 147, 202, 50 },
		{ 147, 203, 50 },
		{ 147, 204, 50 },
		{ 147, 205, 50 },
		{ 147, 206, 50 },
		{ 147, 207, 50 },
		{ 147, 208, 50 },
		{ 147, 209, 50 },
		{ 147, 210, 50 },
		{ 147, 211, 50 },
		{ 147, 212, 50 },
		{ 147, 213, 50 },
		{ 147, 214, 50 },
		{ 147, 215, 50 },
		{ 148, 189, 50 },
		{ 148, 190, 50 },
		{ 148, 191, 50 },
		{ 148, 192, 50 },
		{ 148, 193, 50 },
		{ 148, 194, 50 },
		{ 148, 195, 50 },
		{ 148, 196, 50 },
		{ 148, 197, 50 },
		{ 148, 198, 50 },
		{ 148, 199, 50 },
		{ 148, 200, 50 },
		{ 148, 201, 50 },
		{ 148, 202, 50 },
		{ 148, 203, 50 },
		{ 148, 204, 50 },
		{ 148, 205, 50 },
		{ 148, 206, 50 },
		{ 148, 207, 50 },
		{ 148, 208, 50 },
		{ 148, 209, 50 },
		{ 148, 210, 50 },
		{ 148, 211, 50 },
		{ 148, 212, 50 },
		{ 148, 213, 50 },
		{ 148, 214, 50 },
		{ 148, 215, 50 },
		{ 148, 216, 50 },
		{ 149, 190, 50 },
		{ 149, 191, 50 },
		{ 149, 192, 50 },
		{ 149, 193, 50 },
		{ 149, 194, 50 },
		{ 149, 195, 50 },
		{ 149, 196, 50 },
		{ 149, 197, 50 },
		{ 149, 198, 50 },
		{ 149, 199, 50 },
		{ 149, 200, 50 },
		{ 149, 201, 50 },
		{ 149, 202, 50 },
		{ 149, 203, 50 },
		{ 149, 204, 50 },
		{ 149, 205, 50 },
		{ 149, 206, 50 },
		{ 149, 207, 50 },
		{ 149, 208, 50 },
		{ 149, 209, 50 },
		{ 149, 210, 50 },
		{ 149, 211, 50 },
		{ 149, 212, 50 },
		{ 149, 213, 50 },
		{ 149, 214, 50 },
		{ 149, 215, 50 },
		{ 149, 216, 50 },
		{ 150, 191, 50 },
		{ 150, 192, 50 },
		{ 150, 193, 50 },
		{ 150, 194, 50 },
		{ 150, 195, 50 },
		{ 150, 196, 50 },
		{ 150, 197, 50 },
		{ 150, 198, 50 },
		{ 150, 199, 50 },
		{ 150, 200, 50 },
		{ 150, 201, 50 },
		{ 150, 202, 50 },
		{ 150, 203, 50 },
		{ 150, 204, 50 },
		{ 150, 205, 50 },
		{ 150, 206, 50 },
		{ 150, 207, 50 },
		{ 150, 208, 50 },
		{ 150, 209, 50 },
		{ 150, 210, 50 },
		{ 150, 211, 50 },
		{ 150, 212, 50 },
		{ 150, 213, 50 },
		{ 150, 214, 50 },
		{ 150, 215, 50 },
		{ 150, 216, 50 },
		{ 151, 191, 50 },
		{ 151, 192, 50 },
		{ 151, 193, 50 },
		{ 151, 194, 50 },
		{ 151, 195, 50 },
		{ 151, 196, 50 },
		{ 151, 197, 50 },
		{ 151, 198, 50 },
		{ 151, 199, 50 },
		{ 151, 200, 50 },
		{ 151, 201, 50 },
		{ 151, 202, 50 },
		{ 151, 203, 50 },
		{ 151, 204, 50 },
		{ 151, 205, 50 },
		{ 151, 206, 50 },
		{ 151, 207, 50 },
		{ 151, 208, 50 },
		{ 151, 209, 50 },
		{ 151, 210, 50 },
		{ 151, 211, 50 },
		{ 151, 212, 50 },
		{ 151, 213, 50 },
		{ 151, 214, 50 },
		{ 151, 215, 50 },
		{ 151, 216, 50 },
		{ 151, 217, 50 },
		{ 152, 192, 50 },
		{ 152, 193, 50 },
		{ 152, 194, 50 },
		{ 152, 195, 50 },
		{ 152, 196, 50 },
		{ 152, 197, 50 },
		{ 152, 198, 50 },
		{ 152, 199, 50 },
		{ 152, 200, 50 },
		{ 152, 201, 50 },
		{ 152, 202, 50 },
		{ 152, 203, 50 },
		{ 152, 204, 50 },
		{ 152, 205, 50 },
		{ 152, 206, 50 },
		{ 152, 207, 50 },
		{ 152, 208, 50 },
		{ 152, 209, 50 },
		{ 152, 210, 50 },
		{ 152, 211, 50 },
		{ 152, 212, 50 },
		{ 152, 213, 50 },
		{ 152, 214, 50 },
		{ 152, 215, 50 },
		{ 152, 216, 50 },
		{ 152, 217, 50 },
		{ 153, 193, 50 },
		{ 153, 194, 50 },
		{ 153, 195, 50 },
		{ 153, 196, 50 },
		{ 153, 197, 50 },
		{ 153, 198, 50 },
		{ 153, 199, 50 },
		{ 153, 200, 50 },
		{ 153, 201, 50 },
		{ 153, 202, 50 },
		{ 153, 203, 50 },
		{ 153, 204, 50 },
		{ 153, 205, 50 },
		{ 153, 206, 50 },
		{ 153, 207, 50 },
		{ 153, 208, 50 },
		{ 153, 209, 50 },
		{ 153, 210, 50 },
		{ 153, 211, 50 },
		{ 153, 212, 50 },
		{ 153, 213, 50 },
		{ 153, 214, 50 },
		{ 153, 215, 50 },
		{ 153, 216, 50 },
		{ 153, 217, 50 },
		{ 154, 193, 50 },
		{ 154, 194, 50 },
		{ 154, 195, 50 },
		{ 154, 196, 50 },
		{ 154, 197, 50 },
		{ 154, 198, 50 },
		{ 154, 199, 50 },
		{ 154, 200, 50 },
		{ 154, 201, 50 },
		{ 154, 202, 50 },
		{ 154, 203, 50 },
		{ 154, 204, 50 },
		{ 154, 205, 50 },
		{ 154, 206, 50 },
		{ 154, 207, 50 },
		{ 154, 208, 50 },
		{ 154, 209, 50 },
		{ 154, 210, 50 },
		{ 154, 211, 50 },
		{ 154, 212, 50 },
		{ 154, 213, 50 },
		{ 154, 214, 50 },
		{ 154, 215, 50 },
		{ 154, 216, 50 },
		{ 154, 217, 50 },
		{ 154, 218, 50 },
		{ 155, 194, 50 },
		{ 155, 195, 50 },
		{ 155, 196, 50 },
		{ 155, 197, 50 },
		{ 155, 198, 50 },
		{ 155, 199, 50 },
		{ 155, 200, 50 },
		{ 155, 201, 50 },
		{ 155, 202, 50 },
		{ 155, 203, 50 },
		{ 155, 204, 50 },
		{ 155, 205, 50 },
		{ 155, 206, 50 },
		{ 155, 207, 50 },
		{ 155, 208, 50 },
		{ 155, 209, 50 },
		{ 155, 210, 50 },
		{ 155, 211, 50 },
		{ 155, 212, 50 },
		{ 155, 213, 50 },
		{ 155, 214, 50 },
		{ 155, 215, 50 },
		{ 155, 216, 50 },
		{ 155, 217, 50 },
		{ 155, 218, 50 },
		{ 156, 194, 50 },
		{ 156, 195, 50 },
		{ 156, 196, 50 },
		{ 156, 197, 50 },
		{ 156, 198, 50 },
		{ 156, 199, 50 },
		{ 156, 200, 50 },
		{ 156, 201, 50 },
		{ 156, 202, 50 },
		{ 156, 203, 50 },
		{ 156, 204, 50 },
		{ 156, 205, 50 },
		{ 156, 206, 50 },
		{ 156, 207, 50 },
		{ 156, 208, 50 },
		{ 156, 209, 50 },
		{ 156, 210, 50 },
		{ 156, 211, 50 },
		{ 156, 212, 50 },
		{ 156, 213, 50 },
		{ 156, 214, 50 },
		{ 156, 215, 50 },
		{ 156, 216, 50 },
		{ 156, 217, 50 },
		{ 156, 218, 50 },
		{ 156, 219, 50 },
		{ 157, 195, 50 },
		{ 157, 196, 50 },
		{ 157, 197, 50 },
		{ 157, 198, 50 },
		{ 157, 199, 50 },
		{ 157, 200, 50 },
		{ 157, 201, 50 },
		{ 157, 202, 50 },
		{ 157, 203, 50 },
		{ 157, 204, 50 },
		{ 157, 205, 50 },
		{ 157, 206, 50 },
		{ 157, 207, 50 },
		{ 157, 208, 50 },
		{ 157, 209, 50 },
		{ 157, 210, 50 },
		{ 157, 211, 50 },
		{ 157, 212, 50 },
		{ 157, 213, 50 },
		{ 157, 214, 50 },
		{ 157, 215, 50 },
		{ 157, 216, 50 },
		{ 157, 217, 50 },
		{ 157, 218, 50 },
		{ 157, 219, 50 },
		{ 158, 196, 50 },
		{ 158, 197, 50 },
		{ 158, 198, 50 },
		{ 158, 199, 50 },
		{ 158, 200, 50 },
		{ 158, 201, 50 },
		{ 158, 202, 50 },
		{ 158, 203, 50 },
		{ 158, 204, 50 },
		{ 158, 205, 50 },
		{ 158, 206, 50 },
		{ 158, 207, 50 },
		{ 158, 208, 50 },
		{ 158, 209, 50 },
		{ 158, 210, 50 },
		{ 158, 211, 50 },
		{ 158, 212, 50 },
		{ 158, 213, 50 },
		{ 158, 214, 50 },
		{ 158, 215, 50 },
		{ 158, 216, 50 },
		{ 158, 217, 50 },
		{ 158, 218, 50 },
		{ 158, 219, 50 },
		{ 159, 196, 50 },
		{ 159, 197, 50 },
		{ 159, 198, 50 },
		{ 159, 199, 50 },
		{ 159, 200, 50 },
		{ 159, 201, 50 },
		{ 159, 202, 50 },
		{ 159, 203, 50 },
		{ 159, 204, 50 },
		{ 159, 205, 50 },
		{ 159, 206, 50 },
		{ 159, 207, 50 },
		{ 159, 208, 50 },
		{ 159, 209, 50 },
		{ 159, 210, 50 },
		{ 159, 211, 50 },
		{ 159, 212, 50 },
		{ 159, 213, 50 },
		{ 159, 214, 50 },
		{ 159, 215, 50 },
		{ 159, 216, 50 },
		{ 159, 217, 50 },
		{ 159, 218, 50 },
		{ 159, 219, 50 },
		{ 159, 220, 50 },
		{ 160, 197, 50 },
		{ 160, 198, 50 },
		{ 160, 199, 50 },
		{ 160, 200, 50 },
		{ 160, 201, 50 },
		{ 160, 202, 50 },
		{ 160, 203, 50 },
		{ 160, 204, 50 },
		{ 160, 205, 50 },
		{ 160, 206, 50 },
		{ 160, 207, 50 },
		{ 160, 208, 50 },
		{ 160, 209, 50 },
		{ 160, 210, 50 },
		{ 160, 211, 50 },
		{ 160, 212, 50 },
		{ 160, 213, 50 },
		{ 160, 214, 50 },
		{ 160, 215, 50 },
		{ 160, 216, 50 },
		{ 160, 217, 50 },
		{ 160, 218, 50 },
		{ 160, 219, 50 },
		{ 160, 220, 50 },
		{ 161, 198, 50 },
		{ 161, 199, 50 },
		{ 161, 200, 50 },
		{ 161, 201, 50 },
		{ 161, 202, 50 },
		{ 161, 203, 50 },
		{ 161, 204, 50 },
		{ 161, 205, 50 },
		{ 161, 206, 50 },
		{ 161, 207, 50 },
		{ 161, 208, 50 },
		{ 161, 209, 50 },
		{ 161, 210, 50 },
		{ 161, 211, 50 },
		{ 161, 212, 50 },
		{ 161, 213, 50 },
		{ 161, 214, 50 },
		{ 161, 215, 50 },
		{ 161, 216, 50 },
		{ 161, 217, 50 },
		{ 161, 218, 50 },
		{ 161, 219, 50 },
		{ 161, 220, 50 },
		{ 162, 198, 50 },
		{ 162, 199, 50 },
		{ 162, 200, 50 },
		{ 162, 201, 50 },
		{ 162, 202, 50 },
		{ 162, 203, 50 },
		{ 162, 204, 50 },
		{ 162, 205, 50 },
		{ 162, 206, 50 },
		{ 162, 207, 50 },
		{ 162, 208, 50 },
		{ 162, 209, 50 },
		{ 162, 210, 50 },
		{ 162, 211, 50 },
		{ 162, 212, 50 },
		{ 162, 213, 50 },
		{ 162, 214, 50 },
		{ 162, 215, 50 },
		{ 162, 216, 50 },
		{ 162, 217, 50 },
		{ 162, 218, 50 },
		{ 162, 219, 50 },
		{ 162, 220, 50 },
		{ 162, 221, 50 },
		{ 163, 199, 50 },
		{ 163, 200, 50 },
		{ 163, 201, 50 },
		{ 163, 202, 50 },
		{ 163, 203, 50 },
		{ 163, 204, 50 },
		{ 163, 205, 50 },
		{ 163, 206, 50 },
		{ 163, 207, 50 },
		{ 163, 208, 50 },
		{ 163, 209, 50 },
		{ 163, 210, 50 },
		{ 163, 211, 50 },
		{ 163, 212, 50 },
		{ 163, 213, 50 },
		{ 163, 214, 50 },
		{ 163, 215, 50 },
		{ 163, 216, 50 },
		{ 163, 217, 50 },
		{ 163, 218, 50 },
		{ 163, 219, 50 },
		{ 163, 220, 50 },
		{ 163, 221, 50 },
		{ 164, 199, 50 },
		{ 164, 200, 50 },
		{ 164, 201, 50 },
		{ 164, 202, 50 },
		{ 164, 203, 50 },
		{ 164, 204, 50 },
		{ 164, 205, 50 },
		{ 164, 206, 50 },
		{ 164, 207, 50 },
		{ 164, 208, 50 },
		{ 164, 209, 50 },
		{ 164, 210, 50 },
		{ 164, 211, 50 },
		{ 164, 212, 50 },
		{ 164, 213, 50 },
		{ 164, 214, 50 },
		{ 164, 215, 50 },
		{ 164, 216, 50 },
		{ 164, 217, 50 },
		{ 164, 218, 50 },
		{ 164, 219, 50 },
		{ 164, 220, 50 },
		{ 164, 221, 50 },
		{ 164, 222, 50 },
		{ 165, 200, 50 },
		{ 165, 201, 50 },
		{ 165, 202, 50 },
		{ 165, 203, 50 },
		{ 165, 204, 50 },
		{ 165, 205, 50 },
		{ 165, 206, 50 },
		{ 165, 207, 50 },
		{ 165, 208, 50 },
		{ 165, 209, 50 },
		{ 165, 210, 50 },
		{ 165, 211, 50 },
		{ 165, 212, 50 },
		{ 165, 213, 50 },
		{ 165, 214, 50 },
		{ 165, 215, 50 },
		{ 165, 216, 50 },
		{ 165, 217, 50 },
		{ 165, 218, 50 },
		{ 165, 219, 50 },
		{ 165, 220, 50 },
		{ 165, 221, 50 },
		{ 165, 222, 50 },
		{ 166, 201, 50 },
		{ 166, 202, 50 },
		{ 166, 203, 50 },
		{ 166, 204, 50 },
		{ 166, 205, 50 },
		{ 166, 206, 50 },
		{ 166, 207, 50 },
		{ 166, 208, 50 },
		{ 166, 209, 50 },
		{ 166, 210, 50 },
		{ 166, 211, 50 },
		{ 166, 212, 50 },
		{ 166, 213, 50 },
		{ 166, 214, 50 },
		{ 166, 215, 50 },
		{ 166, 216, 50 },
		{ 166, 217, 50 },
		{ 166, 218, 50 },
		{ 166, 219, 50 },
		{ 166, 220, 50 },
		{ 166, 221, 50 },
		{ 166, 222, 50 },
		{ 167, 201, 50 },
		{ 167, 202, 50 },
		{ 167, 203, 50 },
		{ 167, 204, 50 },
		{ 167, 205, 50 },
		{ 167, 206, 50 },
		{ 167, 207, 50 },
		{ 167, 208, 50 },
		{ 167, 209, 50 },
		{ 167, 210, 50 },
		{ 167, 211, 50 },
		{ 167, 212, 50 },
		{ 167, 213, 50 },
		{ 167, 214, 50 },
		{ 167, 215, 50 },
		{ 167, 216, 50 },
		{ 167, 217, 50 },
		{ 167, 218, 50 },
		{ 167, 219, 50 },
		{ 167, 220, 50 },
		{ 167, 221, 50 },
		{ 167, 222, 50 },
		{ 167, 223, 50 },
		{ 168, 202, 50 },
		{ 168, 203, 50 },
		{ 168, 204, 50 },
		{ 168, 205, 50 },
		{ 168, 206, 50 },
		{ 168, 207, 50 },
		{ 168, 208, 50 },
		{ 168, 209, 50 },
		{ 168, 210, 50 },
		{ 168, 211, 50 },
		{ 168, 212, 50 },
		{ 168, 213, 50 },
		{ 168, 214, 50 },
		{ 168, 215, 50 },
		{ 168, 216, 50 },
		{ 168, 217, 50 },
		{ 168, 218, 50 },
		{ 168, 219, 50 },
		{ 168, 220, 50 },
		{ 168, 221, 50 },
		{ 168, 222, 50 },
		{ 168, 223, 50 },
		{ 169, 203, 50 },
		{ 169, 204, 50 },
		{ 169, 205, 50 },
		{ 169, 206, 50 },
		{ 169, 207, 50 },
		{ 169, 208, 50 },
		{ 169, 209, 50 },
		{ 169, 210, 50 },
		{ 169, 211, 50 },
		{ 169, 212, 50 },
		{ 169, 213, 50 },
		{ 169, 214, 50 },
		{ 169, 215, 50 },
		{ 169, 216, 50 },
		{ 169, 217, 50 },
		{ 169, 218, 50 },
		{ 169, 219, 50 },
		{ 169, 220, 50 },
		{ 169, 221, 50 },
		{ 169, 222, 50 },
		{ 169, 223, 50 },
		{ 170, 203, 50 },
		{ 170, 204, 50 },
		{ 170, 205, 50 },
		{ 170, 206, 50 },
		{ 170, 207, 50 },
		{ 170, 208, 50 },
		{ 170, 209, 50 },
		{ 170, 210, 50 },
		{ 170, 211, 50 },
		{ 170, 212, 50 },
		{ 170, 213, 50 },
		{ 170, 214, 50 },
		{ 170, 215, 50 },
		{ 170, 216, 50 },
		{ 170, 217, 50 },
		{ 170, 218, 50 },
		{ 170, 219, 50 },
		{ 170, 220, 50 },
		{ 170, 221, 50 },
		{ 170, 222, 50 },
		{ 170, 223, 50 },
		{ 170, 224, 50 },
		{ 171, 204, 50 },
		{ 171, 205, 50 },
		{ 171, 206, 50 },
		{ 171, 207, 50 },
		{ 171, 208, 50 },
		{ 171, 209, 50 },
		{ 171, 210, 50 },
		{ 171, 211, 50 },
		{ 171, 212, 50 },
		{ 171, 213, 50 },
		{ 171, 214, 50 },
		{ 171, 215, 50 },
		{ 171, 216, 50 },
		{ 171, 217, 50 },
		{ 171, 218, 50 },
		{ 171, 219, 50 },
		{ 171, 220, 50 },
		{ 171, 221, 50 },
		{ 171, 222, 50 },
		{ 171, 223, 50 },
		{ 171, 224, 50 },
		{ 172, 204, 50 },
		{ 172, 205, 50 },
		{ 172, 206, 50 },
		{ 172, 207, 50 },
		{ 172, 208, 50 },
		{ 172, 209, 50 },
		{ 172, 210, 50 },
		{ 172, 211, 50 },
		{ 172, 212, 50 },
		{ 172, 213, 50 },
		{ 172, 214, 50 },
		{ 172, 215, 50 },
		{ 172, 216, 50 },
		{ 172, 217, 50 },
		{ 172, 218, 50 },
		{ 172, 219, 50 },
		{ 172, 220, 50 },
		{ 172, 221, 50 },
		{ 172, 222, 50 },
		{ 172, 223, 50 },
		{ 172, 224, 50 },
		{ 172, 225, 50 },
		{ 173, 205, 50 },
		{ 173, 206, 50 },
		{ 173, 207, 50 },
		{ 173, 208, 50 },
		{ 173, 209, 50 },
		{ 173, 210, 50 },
		{ 173, 211, 50 },
		{ 173, 212, 50 },
		{ 173, 213, 50 },
		{ 173, 214, 50 },
		{ 173, 215, 50 },
		{ 173, 216, 50 },
		{ 173, 217, 50 },
		{ 173, 218, 50 },
		{ 173, 219, 50 },
		{ 173, 220, 50 },
		{ 173, 221, 50 },
		{ 173, 222, 50 },
		{ 173, 223, 50 },
		{ 173, 224, 50 },
		{ 173, 225, 50 },
		{ 174, 206, 50 },
		{ 174, 207, 50 },
		{ 174, 208, 50 },
		{ 174, 209, 50 },
		{ 174, 210, 50 },
		{ 174, 211, 50 },
		{ 174, 212, 50 },
		{ 174, 213, 50 },
		{ 174, 214, 50 },
		{ 174, 215, 50 },
		{ 174, 216, 50 },
		{ 174, 217, 50 },
		{ 174, 218, 50 },
		{ 174, 219, 50 },
		{ 174, 220, 50 },
		{ 174, 221, 50 },
		{ 174, 222, 50 },
		{ 174, 223, 50 },
		{ 174, 224, 50 },
		{ 174, 225, 50 },
		{ 175, 206, 50 },
		{ 175, 207, 50 },
		{ 175, 208, 50 },
		{ 175, 209, 50 },
		{ 175, 210, 50 },
		{ 175, 211, 50 },
		{ 175, 212, 50 },
		{ 175, 213, 50 },
		{ 175, 214, 50 },
		{ 175, 215, 50 },
		{ 175, 216, 50 },
		{ 175, 217, 50 },
		{ 175, 218, 50 },
		{ 175, 219, 50 },
		{ 175, 220, 50 },
		{ 175, 221, 50 },
		{ 175, 222, 50 },
		{ 175, 223, 50 },
		{ 175, 224, 50 },
		{ 175, 225, 50 },
		{ 175, 226, 50 },
		{ 176, 207, 50 },
		{ 176, 208, 50 },
		{ 176, 209, 50 },
		{ 176, 210, 50 },
		{ 176, 211, 50 },
		{ 176, 212, 50 },
		{ 176, 213, 50 },
		{ 176, 214, 50 },
		{ 176, 215, 50 },
		{ 176, 216, 50 },
		{ 176, 217, 50 },
		{ 176, 218, 50 },
		{ 176, 219, 50 },
		{ 176, 220, 50 },
		{ 176, 221, 50 },
		{ 176, 222, 50 },
		{ 176, 223, 50 },
		{ 176, 224, 50 },
		{ 176, 225, 50 },
		{ 176, 226, 50 },
		{ 177, 208, 50 },
		{ 177, 209, 50 },
		{ 177, 210, 50 },
		{ 177, 211, 50 },
		{ 177, 212, 50 },
		{ 177, 213, 50 },
		{ 177, 214, 50 },
		{ 177, 215, 50 },
		{ 177, 216, 50 },
		{ 177, 217, 50 },
		{ 177, 218, 50 },
		{ 177, 219, 50 },
		{ 177, 220, 50 },
		{ 177, 221, 50 },
		{ 177, 222, 50 },
		{ 177, 223, 50 },
		{ 177, 224, 50 },
		{ 177, 225, 50 },
		{ 177, 226, 50 },
		{ 178, 208, 50 },
		{ 178, 209, 50 },
		{ 178, 210, 50 },
		{ 178, 211, 50 },
		{ 178, 212, 50 },
		{ 178, 213, 50 },
		{ 178, 214, 50 },
		{ 178, 215, 50 },
		{ 178, 216, 50 },
		{ 178, 217, 50 },
		{ 178, 218, 50 },
		{ 178, 219, 50 },
		{ 178, 220, 50 },
		{ 178, 221, 50 },
		{ 178, 222, 50 },
		{ 178, 223, 50 },
		{ 178, 224, 50 },
		{ 178, 225, 50 },
		{ 178, 226, 50 },
		{ 178, 227, 50 },
		{ 179, 209, 50 },
		{ 179, 210, 50 },
		{ 179, 211, 50 },
		{ 179, 212, 50 },
		{ 179, 213, 50 },
		{ 179, 214, 50 },
		{ 179, 215, 50 },
		{ 179, 216, 50 },
		{ 179, 217, 50 },
		{ 179, 218, 50 },
		{ 179, 219, 50 },
		{ 179, 220, 50 },
		{ 179, 221, 50 },
		{ 179, 222, 50 },
		{ 179, 223, 50 },
		{ 179, 224, 50 },
		{ 179, 225, 50 },
		{ 179, 226, 50 },
		{ 179, 227, 50 },
		{ 180, 209, 50 },
		{ 180, 210, 50 },
		{ 180, 211, 50 },
		{ 180, 212, 50 },
		{ 180, 213, 50 },
		{ 180, 214, 50 },
		{ 180, 215, 50 },
		{ 180, 216, 50 },
		{ 180, 217, 50 },
		{ 180, 218, 50 },
		{ 180, 219, 50 },
		{ 180, 220, 50 },
		{ 180, 221, 50 },
		{ 180, 222, 50 },
		{ 180, 223, 50 },
		{ 180, 224, 50 },
		{ 180, 225, 50 },
		{ 180, 226, 50 },
		{ 180, 227, 50 },
		{ 180, 228, 50 },
		{ 181, 210, 50 },
		{ 181, 211, 50 },
		{ 181, 212, 50 },
		{ 181, 213, 50 },
		{ 181, 214, 50 },
		{ 181, 215, 50 },
		{ 181, 216, 50 },
		{ 181, 217, 50 },
		{ 181, 218, 50 },
		{ 181, 219, 50 },
		{ 181, 220, 50 },
		{ 181, 221, 50 },
		{ 181, 222, 50 },
		{ 181, 223, 50 },
		{ 181, 224, 50 },
		{ 181, 225, 50 },
		{ 181, 226, 50 },
		{ 181, 227, 50 },
		{ 181, 228, 50 },
		{ 182, 211, 50 },
		{ 182, 212, 50 },
		{ 182, 213, 50 },
		{ 182, 214, 50 },
		{ 182, 215, 50 },
		{ 182, 216, 50 },
		{ 182, 217, 50 },
		{ 182, 218, 50 },
		{ 182, 219, 50 },
		{ 182, 220, 50 },
		{ 182, 221, 50 },
		{ 182, 222, 50 },
		{ 182, 223, 50 },
		{ 182, 224, 50 },
		{ 182, 225, 50 },
		{ 182, 226, 50 },
		{ 182, 227, 50 },
		{ 182, 228, 50 },
		{ 183, 211, 50 },
		{ 183, 212, 50 },
		{ 183, 213, 50 },
		{ 183, 214, 50 },
		{ 183, 215, 50 },
		{ 183, 216, 50 },
		{ 183, 217, 50 },
		{ 183, 218, 50 },
		{ 183, 219, 50 },
		{ 183, 220, 50 },
		{ 183, 221, 50 },
		{ 183, 222, 50 },
		{ 183, 223, 50 },
		{ 183, 224, 50 },
		{ 183, 225, 50 },
		{ 183, 226, 50 },
		{ 183, 227, 50 },
		{ 183, 228, 50 },
		{ 183, 229, 50 },
		{ 184, 212, 50 },
		{ 184, 213, 50 },
		{ 184, 214, 50 },
		{ 184, 215, 50 },
		{ 184, 216, 50 },
		{ 184, 217, 50 },
		{ 184, 218, 50 },
		{ 184, 219, 50 },
		{ 184, 220, 50 },
		{ 184, 221, 50 },
		{ 184, 222, 50 },
		{ 184, 223, 50 },
		{ 184, 224, 50 },
		{ 184, 225, 50 },
		{ 184, 226, 50 },
		{ 184, 227, 50 },
		{ 184, 228, 50 },
		{ 184, 229, 50 },
		{ 185, 213, 50 },
		{ 185, 214, 50 },
		{ 185, 215, 50 },
		{ 185, 216, 50 },
		{ 185, 217, 50 },
		{ 185, 218, 50 },
		{ 185, 219, 50 },
		{ 185, 220, 50 },
		{ 185, 221, 50 },
		{ 185, 222, 50 },
		{ 185, 223, 50 },
		{ 185, 224, 50 },
		{ 185, 225, 50 },
		{ 185, 226, 50 },
		{ 185, 227, 50 },
		{ 185, 228, 50 },
		{ 185, 229, 50 },
		{ 186, 213, 50 },
		{ 186, 214, 50 },
		{ 186, 215, 50 },
		{ 186, 216, 50 },
		{ 186, 217, 50 },
		{ 186, 218, 50 },
		{ 186, 219, 50 },
		{ 186, 220, 50 },
		{ 186, 221, 50 },
		{ 186, 222, 50 },
		{ 186, 223, 50 },
		{ 186, 224, 50 },
		{ 186, 225, 50 },
		{ 186, 226, 50 },
		{ 186, 227, 50 },
		{ 186, 228, 50 },
		{ 186, 229, 50 },
		{ 186, 230, 50 },
		{ 187, 214, 50 },
		{ 187, 215, 50 },
		{ 187, 216, 50 },
		{ 187, 217, 50 },
		{ 187, 218, 50 },
		{ 187, 219, 50 },
		{ 187, 220, 50 },
		{ 187, 221, 50 },
		{ 187, 222, 50 },
		{ 187, 223, 50 },
		{ 187, 224, 50 },
		{ 187, 225, 50 },
		{ 187, 226, 50 },
		{ 187, 227, 50 },
		{ 187, 228, 50 },
		{ 187, 229, 50 },
		{ 187, 230, 50 },
		{ 188, 215, 50 },
		{ 188, 216, 50 },
		{ 188, 217, 50 },
		{ 188, 218, 50 },
		{ 188, 219, 50 },
		{ 188, 220, 50 },
		{ 188, 221, 50 },
		{ 188, 222, 50 },
		{ 188, 223, 50 },
		{ 188, 224, 50 },
		{ 188, 225, 50 },
		{ 188, 226, 50 },
		{ 188, 227, 50 },
		{ 188, 228, 50 },
		{ 188, 229, 50 },
		{ 188, 230, 50 },
		{ 189, 215, 50 },
		{ 189, 216, 50 },
		{ 189, 217, 50 },
		{ 189, 218, 50 },
		{ 189, 219, 50 },
		{ 189, 220, 50 },
		{ 189, 221, 50 },
		{ 189, 222, 50 },
		{ 189, 223, 50 },
		{ 189, 224, 50 },
		{ 189, 225, 50 },
		{ 189, 226, 50 },
		{ 189, 227, 50 },
		{ 189, 228, 50 },
		{ 189, 229, 50 },
		{ 189, 230, 50 },
		{ 189, 231, 50 },
		{ 190, 216, 50 },
		{ 190, 217, 50 },
		{ 190, 218, 50 },
		{ 190, 219, 50 },
		{ 190, 220, 50 },
		{ 190, 221, 50 },
		{ 190, 222, 50 },
		{ 190, 223, 50 },
		{ 190, 224, 50 },
		{ 190, 225, 50 },
		{ 190, 226, 50 },
		{ 190, 227, 50 },
		{ 190, 228, 50 },
		{ 190, 229, 50 },
		{ 190, 230, 50 },
		{ 190, 231, 50 },
		{ 191, 218, 50 },
		{ 191, 219, 50 },
		{ 191, 220, 50 },
		{ 191, 221, 50 },
		{ 191, 222, 50 },
		{ 191, 223, 50 },
		{ 191, 224, 50 },
		{ 191, 225, 50 },
		{ 191, 226, 50 },
		{ 191, 227, 50 },
		{ 191, 228, 50 },
		{ 191, 229, 50 },
		{ 191, 230, 50 },
		{ 323, 284, 50 },
		{ 323, 285, 50 },
		{ 323, 286, 50 },
		{ 323, 287, 50 },
		{ 323, 288, 50 },
		{ 323, 289, 50 },
		{ 323, 290, 50 },
		{ 323, 291, 50 },
		{ 323, 292, 50 },
		{ 323, 293, 50 },
		{ 323, 294, 50 },
		{ 323, 295, 50 },
		{ 323, 296, 50 },
		{ 324, 283, 50 },
		{ 324, 284, 50 },
		{ 324, 285, 50 },
		{ 324, 286, 50 },
		{ 324, 287, 50 },
		{ 324, 288, 50 },
		{ 324, 289, 50 },
		{ 324, 290, 50 },
		{ 324, 291, 50 },
		{ 324, 292, 50 },
		{ 324, 293, 50 },
		{ 324, 294, 50 },
		{ 324, 295, 50 },
		{ 324, 296, 50 },
		{ 324, 297, 50 },
		{ 324, 298, 50 },
		{ 325, 283, 50 },
		{ 325, 284, 50 },
		{ 325, 285, 50 },
		{ 325, 286, 50 },
		{ 325, 287, 50 },
		{ 325, 288, 50 },
		{ 325, 289, 50 },
		{ 325, 290, 50 },
		{ 325, 291, 50 },
		{ 325, 292, 50 },
		{ 325, 293, 50 },
		{ 325, 294, 50 },
		{ 325, 295, 50 },
		{ 325, 296, 50 },
		{ 325, 297, 50 },
		{ 325, 298, 50 },
		{ 325, 299, 50 },
		{ 326, 284, 50 },
		{ 326, 285, 50 },
		{ 326, 286, 50 },
		{ 326, 287, 50 },
		{ 326, 288, 50 },
		{ 326, 289, 50 },
		{ 326, 290, 50 },
		{ 326, 291, 50 },
		{ 326, 292, 50 },
		{ 326, 293, 50 },
		{ 326, 294, 50 },
		{ 326, 295, 50 },
		{ 326, 296, 50 },
		{ 326, 297, 50 },
		{ 326, 298, 50 },
		{ 326, 299, 50 },
		{ 327, 284, 50 },
		{ 327, 285, 50 },
		{ 327, 286, 50 },
		{ 327, 287, 50 },
		{ 327, 288, 50 },
		{ 327, 289, 50 },
		{ 327, 290, 50 },
		{ 327, 291, 50 },
		{ 327, 292, 50 },
		{ 327, 293, 50 },
		{ 327, 294, 50 },
		{ 327, 295, 50 },
		{ 327, 296, 50 },
		{ 327, 297, 50 },
		{ 327, 298, 50 },
		{ 327, 299, 50 },
		{ 327, 300, 50 },
		{ 328, 284, 50 },
		{ 328, 285, 50 },
		{ 328, 286, 50 },
		{ 328, 287, 50 },
		{ 328, 288, 50 },
		{ 328, 289, 50 },
		{ 328, 290, 50 },
		{ 328, 291, 50 },
		{ 328, 292, 50 },
		{ 328, 293, 50 },
		{ 328, 294, 50 },
		{ 328, 295, 50 },
		{ 328, 296, 50 },
		{ 328, 297, 50 },
		{ 328, 298, 50 },
		{ 328, 299, 50 },
		{ 328, 300, 50 },
		{ 328, 301, 50 },
		{ 329, 285, 50 },
		{ 329, 286, 50 },
		{ 329, 287, 50 },
		{ 329, 288, 50 },
		{ 329, 289, 50 },
		{ 329, 290, 50 },
		{ 329, 291, 50 },
		{ 329, 292, 50 },
		{ 329, 293, 50 },
		{ 329, 294, 50 },
		{ 329, 295, 50 },
		{ 329, 296, 50 },
		{ 329, 297, 50 },
		{ 329, 298, 50 },
		{ 329, 299, 50 },
		{ 329, 300, 50 },
		{ 329, 301, 50 },
		{ 330, 285, 50 },
		{ 330, 286, 50 },
		{ 330, 287, 50 },
		{ 330, 288, 50 },
		{ 330, 289, 50 },
		{ 330, 290, 50 },
		{ 330, 291, 50 },
		{ 330, 292, 50 },
		{ 330, 293, 50 },
		{ 330, 294, 50 },
		{ 330, 295, 50 },
		{ 330, 296, 50 },
		{ 330, 297, 50 },
		{ 330, 298, 50 },
		{ 330, 299, 50 },
		{ 330, 300, 50 },
		{ 330, 301, 50 },
		{ 330, 302, 50 },
		{ 331, 285, 50 },
		{ 331, 286, 50 },
		{ 331, 287, 50 },
		{ 331, 288, 50 },
		{ 331, 289, 50 },
		{ 331, 290, 50 },
		{ 331, 291, 50 },
		{ 331, 292, 50 },
		{ 331, 293, 50 },
		{ 331, 294, 50 },
		{ 331, 295, 50 },
		{ 331, 296, 50 },
		{ 331, 297, 50 },
		{ 331, 298, 50 },
		{ 331, 299, 50 },
		{ 331, 300, 50 },
		{ 331, 301, 50 },
		{ 331, 302, 50 },
		{ 331, 303, 50 },
		{ 332, 286, 50 },
		{ 332, 287, 50 },
		{ 332, 288, 50 },
		{ 332, 289, 50 },
		{ 332, 290, 50 },
		{ 332, 291, 50 },
		{ 332, 292, 50 },
		{ 332, 293, 50 },
		{ 332, 294, 50 },
		{ 332, 295, 50 },
		{ 332, 296, 50 },
		{ 332, 297, 50 },
		{ 332, 298, 50 },
		{ 332, 299, 50 },
		{ 332, 300, 50 },
		{ 332, 301, 50 },
		{ 332, 302, 50 },
		{ 332, 303, 50 },
		{ 333, 286, 50 },
		{ 333, 287, 50 },
		{ 333, 288, 50 },
		{ 333, 289, 50 },
		{ 333, 290, 50 },
		{ 333, 291, 50 },
		{ 333, 292, 50 },
		{ 333, 293, 50 },
		{ 333, 294, 50 },
		{ 333, 295, 50 },
		{ 333, 296, 50 },
		{ 333, 297, 50 },
		{ 333, 298, 50 },
		{ 333, 299, 50 },
		{ 333, 300, 50 },
		{ 333, 301, 50 },
		{ 333, 302, 50 },
		{ 333, 303, 50 },
		{ 333, 304, 50 },
		{ 334, 286, 50 },
		{ 334, 287, 50 },
		{ 334, 288, 50 },
		{ 334, 289, 50 },
		{ 334, 290, 50 },
		{ 334, 291, 50 },
		{ 334, 292, 50 },
		{ 334, 293, 50 },
		{ 334, 294, 50 },
		{ 334, 295, 50 },
		{ 334, 296, 50 },
		{ 334, 297, 50 },
		{ 334, 298, 50 },
		{ 334, 299, 50 },
		{ 334, 300, 50 },
		{ 334, 301, 50 },
		{ 334, 302, 50 },
		{ 334, 303, 50 },
		{ 334, 304, 50 },
		{ 334, 305, 50 },
		{ 335, 287, 50 },
		{ 335, 288, 50 },
		{ 335, 289, 50 },
		{ 335, 290, 50 },
		{ 335, 291, 50 },
		{ 335, 292, 50 },
		{ 335, 293, 50 },
		{ 335, 294, 50 },
		{ 335, 295, 50 },
		{ 335, 296, 50 },
		{ 335, 297, 50 },
		{ 335, 298, 50 },
		{ 335, 299, 50 },
		{ 335, 300, 50 },
		{ 335, 301, 50 },
		{ 335, 302, 50 },
		{ 335, 303, 50 },
		{ 335, 304, 50 },
		{ 335, 305, 50 },
		{ 336, 287, 50 },
		{ 336, 288, 50 },
		{ 336, 289, 50 },
		{ 336, 290, 50 },
		{ 336, 291, 50 },
		{ 336, 292, 50 },
		{ 336, 293, 50 },
		{ 336, 294, 50 },
		{ 336, 295, 50 },
		{ 336, 296, 50 },
		{ 336, 297, 50 },
		{ 336, 298, 50 },
		{ 336, 299, 50 },
		{ 336, 300, 50 },
		{ 336, 301, 50 },
		{ 336, 302, 50 },
		{ 336, 303, 50 },
		{ 336, 304, 50 },
		{ 336, 305, 50 },
		{ 336, 306, 50 },
		{ 337, 288, 50 },
		{ 337, 289, 50 },
		{ 337, 290, 50 },
		{ 337, 291, 50 },
		{ 337, 292, 50 },
		{ 337, 293, 50 },
		{ 337, 294, 50 },
		{ 337, 295, 50 },
		{ 337, 296, 50 },
		{ 337, 297, 50 },
		{ 337, 298, 50 },
		{ 337, 299, 50 },
		{ 337, 300, 50 },
		{ 337, 301, 50 },
		{ 337, 302, 50 },
		{ 337, 303, 50 },
		{ 337, 304, 50 },
		{ 337, 305, 50 },
		{ 337, 306, 50 },
		{ 338, 288, 50 },
		{ 338, 289, 50 },
		{ 338, 290, 50 },
		{ 338, 291, 50 },
		{ 338, 292, 50 },
		{ 338, 293, 50 },
		{ 338, 294, 50 },
		{ 338, 295, 50 },
		{ 338, 296, 50 },
		{ 338, 297, 50 },
		{ 338, 298, 50 },
		{ 338, 299, 50 },
		{ 338, 300, 50 },
		{ 338, 301, 50 },
		{ 338, 302, 50 },
		{ 338, 303, 50 },
		{ 338, 304, 50 },
		{ 338, 305, 50 },
		{ 338, 306, 50 },
		{ 338, 307, 50 },
		{ 339, 288, 50 },
		{ 339, 289, 50 },
		{ 339, 290, 50 },
		{ 339, 291, 50 },
		{ 339, 292, 50 },
		{ 339, 293, 50 },
		{ 339, 294, 50 },
		{ 339, 295, 50 },
		{ 339, 296, 50 },
		{ 339, 297, 50 },
		{ 339, 298, 50 },
		{ 339, 299, 50 },
		{ 339, 300, 50 },
		{ 339, 301, 50 },
		{ 339, 302, 50 },
		{ 339, 303, 50 },
		{ 339, 304, 50 },
		{ 339, 305, 50 },
		{ 339, 306, 50 },
		{ 339, 307, 50 },
		{ 339, 308, 50 },
		{ 340, 289, 50 },
		{ 340, 290, 50 },
		{ 340, 291, 50 },
		{ 340, 292, 50 },
		{ 340, 293, 50 },
		{ 340, 294, 50 },
		{ 340, 295, 50 },
		{ 340, 296, 50 },
		{ 340, 297, 50 },
		{ 340, 298, 50 },
		{ 340, 299, 50 },
		{ 340, 300, 50 },
		{ 340, 301, 50 },
		{ 340, 302, 50 },
		{ 340, 303, 50 },
		{ 340, 304, 50 },
		{ 340, 305, 50 },
		{ 340, 306, 50 },
		{ 340, 307, 50 },
		{ 340, 308, 50 },
		{ 341, 289, 50 },
		{ 341, 290, 50 },
		{ 341, 291, 50 },
		{ 341, 292, 50 },
		{ 341, 293, 50 },
		{ 341, 294, 50 },
		{ 341, 295, 50 },
		{ 341, 296, 50 },
		{ 341, 297, 50 },
		{ 341, 298, 50 },
		{ 341, 299, 50 },
		{ 341, 300, 50 },
		{ 341, 301, 50 },
		{ 341, 302, 50 },
		{ 341, 303, 50 },
		{ 341, 304, 50 },
		{ 341, 305, 50 },
		{ 341, 306, 50 },
		{ 341, 307, 50 },
		{ 341, 308, 50 },
		{ 341, 309, 50 },
		{ 342, 289, 50 },
		{ 342, 290, 50 },
		{ 342, 291, 50 },
		{ 342, 292, 50 },
		{ 342, 293, 50 },
		{ 342, 294, 50 },
		{ 342, 295, 50 },
		{ 342, 296, 50 },
		{ 342, 297, 50 },
		{ 342, 298, 50 },
		{ 342, 299, 50 },
		{ 342, 300, 50 },
		{ 342, 301, 50 },
		{ 342, 302, 50 },
		{ 342, 303, 50 },
		{ 342, 304, 50 },
		{ 342, 305, 50 },
		{ 342, 306, 50 },
		{ 342, 307, 50 },
		{ 342, 308, 50 },
		{ 342, 309, 50 },
		{ 342, 310, 50 },
		{ 343, 290, 50 },
		{ 343, 291, 50 },
		{ 343, 292, 50 },
		{ 343, 293, 50 },
		{ 343, 294, 50 },
		{ 343, 295, 50 },
		{ 343, 296, 50 },
		{ 343, 297, 50 },
		{ 343, 298, 50 },
		{ 343, 299, 50 },
		{ 343, 300, 50 },
		{ 343, 301, 50 },
		{ 343, 302, 50 },
		{ 343, 303, 50 },
		{ 343, 304, 50 },
		{ 343, 305, 50 },
		{ 343, 306, 50 },
		{ 343, 307, 50 },
		{ 343, 308, 50 },
		{ 343, 309, 50 },
		{ 343, 310, 50 },
		{ 344, 290, 50 },
		{ 344, 291, 50 },
		{ 344, 292, 50 },
		{ 344, 293, 50 },
		{ 344, 294, 50 },
		{ 344, 295, 50 },
		{ 344, 296, 50 },
		{ 344, 297, 50 },
		{ 344, 298, 50 },
		{ 344, 299, 50 },
		{ 344, 300, 50 },
		{ 344, 301, 50 },
		{ 344, 302, 50 },
		{ 344, 303, 50 },
		{ 344, 304, 50 },
		{ 344, 305, 50 },
		{ 344, 306, 50 },
		{ 344, 307, 50 },
		{ 344, 308, 50 },
		{ 344, 309, 50 },
		{ 344, 310, 50 },
		{ 344, 311, 50 },
		{ 345, 291, 50 },
		{ 345, 292, 50 },
		{ 345, 293, 50 },
		{ 345, 294, 50 },
		{ 345, 295, 50 },
		{ 345, 296, 50 },
		{ 345, 297, 50 },
		{ 345, 298, 50 },
		{ 345, 299, 50 },
		{ 345, 300, 50 },
		{ 345, 301, 50 },
		{ 345, 302, 50 },
		{ 345, 303, 50 },
		{ 345, 304, 50 },
		{ 345, 305, 50 },
		{ 345, 306, 50 },
		{ 345, 307, 50 },
		{ 345, 308, 50 },
		{ 345, 309, 50 },
		{ 345, 310, 50 },
		{ 345, 311, 50 },
		{ 346, 291, 50 },
		{ 346, 292, 50 },
		{ 346, 293, 50 },
		{ 346, 294, 50 },
		{ 346, 295, 50 },
		{ 346, 296, 50 },
		{ 346, 297, 50 },
		{ 346, 298, 50 },
		{ 346, 299, 50 },
		{ 346, 300, 50 },
		{ 346, 301, 50 },
		{ 346, 302, 50 },
		{ 346, 303, 50 },
		{ 346, 304, 50 },
		{ 346, 305, 50 },
		{ 346, 306, 50 },
		{ 346, 307, 50 },
		{ 346, 308, 50 },
		{ 346, 309, 50 },
		{ 346, 310, 50 },
		{ 346, 311, 50 },
		{ 346, 312, 50 },
		{ 347, 291, 50 },
		{ 347, 292, 50 },
		{ 347, 293, 50 },
		{ 347, 294, 50 },
		{ 347, 295, 50 },
		{ 347, 296, 50 },
		{ 347, 297, 50 },
		{ 347, 298, 50 },
		{ 347, 299, 50 },
		{ 347, 300, 50 },
		{ 347, 301, 50 },
		{ 347, 302, 50 },
		{ 347, 303, 50 },
		{ 347, 304, 50 },
		{ 347, 305, 50 },
		{ 347, 306, 50 },
		{ 347, 307, 50 },
		{ 347, 308, 50 },
		{ 347, 309, 50 },
		{ 347, 310, 50 },
		{ 347, 311, 50 },
		{ 347, 312, 50 },
		{ 347, 313, 50 },
		{ 348, 292, 50 },
		{ 348, 293, 50 },
		{ 348, 294, 50 },
		{ 348, 295, 50 },
		{ 348, 296, 50 },
		{ 348, 297, 50 },
		{ 348, 298, 50 },
		{ 348, 299, 50 },
		{ 348, 300, 50 },
		{ 348, 301, 50 },
		{ 348, 302, 50 },
		{ 348, 303, 50 },
		{ 348, 304, 50 },
		{ 348, 305, 50 },
		{ 348, 306, 50 },
		{ 348, 307, 50 },
		{ 348, 308, 50 },
		{ 348, 309, 50 },
		{ 348, 310, 50 },
		{ 348, 311, 50 },
		{ 348, 312, 50 },
		{ 348, 313, 50 },
		{ 349, 292, 50 },
		{ 349, 293, 50 },
		{ 349, 294, 50 },
		{ 349, 295, 50 },
		{ 349, 296, 50 },
		{ 349, 297, 50 },
		{ 349, 298, 50 },
		{ 349, 299, 50 },
		{ 349, 300, 50 },
		{ 349, 301, 50 },
		{ 349, 302, 50 },
		{ 349, 303, 50 },
		{ 349, 304, 50 },
		{ 349, 305, 50 },
		{ 349, 306, 50 },
		{ 349, 307, 50 },
		{ 349, 308, 50 },
		{ 349, 309, 50 },
		{ 349, 310, 50 },
		{ 349, 311, 50 },
		{ 349, 312, 50 },
		{ 349, 313, 50 },
		{ 349, 314, 50 },
		{ 350, 292, 50 },
		{ 350, 293, 50 },
		{ 350, 294, 50 },
		{ 350, 295, 50 },
		{ 350, 296, 50 },
		{ 350, 297, 50 },
		{ 350, 298, 50 },
		{ 350, 299, 50 },
		{ 350, 300, 50 },
		{ 350, 301, 50 },
		{ 350, 302, 50 },
		{ 350, 303, 50 },
		{ 350, 304, 50 },
		{ 350, 305, 50 },
		{ 350, 306, 50 },
		{ 350, 307, 50 },
		{ 350, 308, 50 },
		{ 350, 309, 50 },
		{ 350, 310, 50 },
		{ 350, 311, 50 },
		{ 350, 312, 50 },
		{ 350, 313, 50 },
		{ 350, 314, 50 },
		{ 350, 315, 50 },
		{ 351, 293, 50 },
		{ 351, 294, 50 },
		{ 351, 295, 50 },
		{ 351, 296, 50 },
		{ 351, 297, 50 },
		{ 351, 298, 50 },
		{ 351, 299, 50 },
		{ 351, 300, 50 },
		{ 351, 301, 50 },
		{ 351, 302, 50 },
		{ 351, 303, 50 },
		{ 351, 304, 50 },
		{ 351, 305, 50 },
		{ 351, 306, 50 },
		{ 351, 307, 50 },
		{ 351, 308, 50 },
		{ 351, 309, 50 },
		{ 351, 310, 50 },
		{ 351, 311, 50 },
		{ 351, 312, 50 },
		{ 351, 313, 50 },
		{ 351, 314, 50 },
		{ 351, 315, 50 },
		{ 352, 293, 50 },
		{ 352, 294, 50 },
		{ 352, 295, 50 },
		{ 352, 296, 50 },
		{ 352, 297, 50 },
		{ 352, 298, 50 },
		{ 352, 299, 50 },
		{ 352, 300, 50 },
		{ 352, 301, 50 },
		{ 352, 302, 50 },
		{ 352, 303, 50 },
		{ 352, 304, 50 },
		{ 352, 305, 50 },
		{ 352, 306, 50 },
		{ 352, 307, 50 },
		{ 352, 308, 50 },
		{ 352, 309, 50 },
		{ 352, 310, 50 },
		{ 352, 311, 50 },
		{ 352, 312, 50 },
		{ 352, 313, 50 },
		{ 352, 314, 50 },
		{ 352, 315, 50 },
		{ 352, 316, 50 },
		{ 353, 294, 50 },
		{ 353, 295, 50 },
		{ 353, 296, 50 },
		{ 353, 297, 50 },
		{ 353, 298, 50 },
		{ 353, 299, 50 },
		{ 353, 300, 50 },
		{ 353, 301, 50 },
		{ 353, 302, 50 },
		{ 353, 303, 50 },
		{ 353, 304, 50 },
		{ 353, 305, 50 },
		{ 353, 306, 50 },
		{ 353, 307, 50 },
		{ 353, 308, 50 },
		{ 353, 309, 50 },
		{ 353, 310, 50 },
		{ 353, 311, 50 },
		{ 353, 312, 50 },
		{ 353, 313, 50 },
		{ 353, 314, 50 },
		{ 353, 315, 50 },
		{ 353, 316, 50 },
		{ 354, 294, 50 },
		{ 354, 295, 50 },
		{ 354, 296, 50 },
		{ 354, 297, 50 },
		{ 354, 298, 50 },
		{ 354, 299, 50 },
		{ 354, 300, 50 },
		{ 354, 301, 50 },
		{ 354, 302, 50 },
		{ 354, 303, 50 },
		{ 354, 304, 50 },
		{ 354, 305, 50 },
		{ 354, 306, 50 },
		{ 354, 307, 50 },
		{ 354, 308, 50 },
		{ 354, 309, 50 },
		{ 354, 310, 50 },
		{ 354, 311, 50 },
		{ 354, 312, 50 },
		{ 354, 313, 50 },
		{ 354, 314, 50 },
		{ 354, 315, 50 },
		{ 354, 316, 50 },
		{ 354, 317, 50 },
		{ 355, 294, 50 },
		{ 355, 295, 50 },
		{ 355, 296, 50 },
		{ 355, 297, 50 },
		{ 355, 298, 50 },
		{ 355, 299, 50 },
		{ 355, 300, 50 },
		{ 355, 301, 50 },
		{ 355, 302, 50 },
		{ 355, 303, 50 },
		{ 355, 304, 50 },
		{ 355, 305, 50 },
		{ 355, 306, 50 },
		{ 355, 307, 50 },
		{ 355, 308, 50 },
		{ 355, 309, 50 },
		{ 355, 310, 50 },
		{ 355, 311, 50 },
		{ 355, 312, 50 },
		{ 355, 313, 50 },
		{ 355, 314, 50 },
		{ 355, 315, 50 },
		{ 355, 316, 50 },
		{ 355, 317, 50 },
		{ 355, 318, 50 },
		{ 356, 295, 50 },
		{ 356, 296, 50 },
		{ 356, 297, 50 },
		{ 356, 298, 50 },
		{ 356, 299, 50 },
		{ 356, 300, 50 },
		{ 356, 301, 50 },
		{ 356, 302, 50 },
		{ 356, 303, 50 },
		{ 356, 304, 50 },
		{ 356, 305, 50 },
		{ 356, 306, 50 },
		{ 356, 307, 50 },
		{ 356, 308, 50 },
		{ 356, 309, 50 },
		{ 356, 310, 50 },
		{ 356, 311, 50 },
		{ 356, 312, 50 },
		{ 356, 313, 50 },
		{ 356, 314, 50 },
		{ 356, 315, 50 },
		{ 356, 316, 50 },
		{ 356, 317, 50 },
		{ 356, 318, 50 },
		{ 357, 295, 50 },
		{ 357, 296, 50 },
		{ 357, 297, 50 },
		{ 357, 298, 50 },
		{ 357, 299, 50 },
		{ 357, 300, 50 },
		{ 357, 301, 50 },
		{ 357, 302, 50 },
		{ 357, 303, 50 },
		{ 357, 304, 50 },
		{ 357, 305, 50 },
		{ 357, 306, 50 },
		{ 357, 307, 50 },
		{ 357, 308, 50 },
		{ 357, 309, 50 },
		{ 357, 310, 50 },
		{ 357, 311, 50 },
		{ 357, 312, 50 },
		{ 357, 313, 50 },
		{ 357, 314, 50 },
		{ 357, 315, 50 },
		{ 357, 316, 50 },
		{ 357, 317, 50 },
		{ 357, 318, 50 },
		{ 357, 319, 50 },
		{ 358, 295, 50 },
		{ 358, 296, 50 },
		{ 358, 297, 50 },
		{ 358, 298, 50 },
		{ 358, 299, 50 },
		{ 358, 300, 50 },
		{ 358, 301, 50 },
		{ 358, 302, 50 },
		{ 358, 303, 50 },
		{ 358, 304, 50 },
		{ 358, 305, 50 },
		{ 358, 306, 50 },
		{ 358, 307, 50 },
		{ 358, 308, 50 },
		{ 358, 309, 50 },
		{ 358, 310, 50 },
		{ 358, 311, 50 },
		{ 358, 312, 50 },
		{ 358, 313, 50 },
		{ 358, 314, 50 },
		{ 358, 315, 50 },
		{ 358, 316, 50 },
		{ 358, 317, 50 },
		{ 358, 318, 50 },
		{ 358, 319, 50 },
		{ 358, 320, 50 },
		{ 359, 296, 50 },
		{ 359, 297, 50 },
		{ 359, 298, 50 },
		{ 359, 299, 50 },
		{ 359, 300, 50 },
		{ 359, 301, 50 },
		{ 359, 302, 50 },
		{ 359, 303, 50 },
		{ 359, 304, 50 },
		{ 359, 305, 50 },
		{ 359, 306, 50 },
		{ 359, 307, 50 },
		{ 359, 308, 50 },
		{ 359, 309, 50 },
		{ 359, 310, 50 },
		{ 359, 311, 50 },
		{ 359, 312, 50 },
		{ 359, 313, 50 },
		{ 359, 314, 50 },
		{ 359, 315, 50 },
		{ 359, 316, 50 },
		{ 359, 317, 50 },
		{ 359, 318, 50 },
		{ 359, 319, 50 },
		{ 359, 320, 50 },
		{ 360, 296, 50 },
		{ 360, 297, 50 },
		{ 360, 298, 50 },
		{ 360, 299, 50 },
		{ 360, 300, 50 },
		{ 360, 301, 50 },
		{ 360, 302, 50 },
		{ 360, 303, 50 },
		{ 360, 304, 50 },
		{ 360, 305, 50 },
		{ 360, 306, 50 },
		{ 360, 307, 50 },
		{ 360, 308, 50 },
		{ 360, 309, 50 },
		{ 360, 310, 50 },
		{ 360, 311, 50 },
		{ 360, 312, 50 },
		{ 360, 313, 50 },
		{ 360, 314, 50 },
		{ 360, 315, 50 },
		{ 360, 316, 50 },
		{ 360, 317, 50 },
		{ 360, 318, 50 },
		{ 360, 319, 50 },
		{ 360, 320, 50 },
		{ 360, 321, 50 },
		{ 361, 297, 50 },
		{ 361, 298, 50 },
		{ 361, 299, 50 },
		{ 361, 300, 50 },
		{ 361, 301, 50 },
		{ 361, 302, 50 },
		{ 361, 303, 50 },
		{ 361, 304, 50 },
		{ 361, 305, 50 },
		{ 361, 306, 50 },
		{ 361, 307, 50 },
		{ 361, 308, 50 },
		{ 361, 309, 50 },
		{ 361, 310, 50 },
		{ 361, 311, 50 },
		{ 361, 312, 50 },
		{ 361, 313, 50 },
		{ 361, 314, 50 },
		{ 361, 315, 50 },
		{ 361, 316, 50 },
		{ 361, 317, 50 },
		{ 361, 318, 50 },
		{ 361, 319, 50 },
		{ 361, 320, 50 },
		{ 361, 321, 50 },
		{ 362, 297, 50 },
		{ 362, 298, 50 },
		{ 362, 299, 50 },
		{ 362, 300, 50 },
		{ 362, 301, 50 },
		{ 362, 302, 50 },
		{ 362, 303, 50 },
		{ 362, 304, 50 },
		{ 362, 305, 50 },
		{ 362, 306, 50 },
		{ 362, 307, 50 },
		{ 362, 308, 50 },
		{ 362, 309, 50 },
		{ 362, 310, 50 },
		{ 362, 311, 50 },
		{ 362, 312, 50 },
		{ 362, 313, 50 },
		{ 362, 314, 50 },
		{ 362, 315, 50 },
		{ 362, 316, 50 },
		{ 362, 317, 50 },
		{ 362, 318, 50 },
		{ 362, 319, 50 },
		{ 362, 320, 50 },
		{ 362, 321, 50 },
		{ 362, 322, 50 },
		{ 363, 297, 50 },
		{ 363, 298, 50 },
		{ 363, 299, 50 },
		{ 363, 300, 50 },
		{ 363, 301, 50 },
		{ 363, 302, 50 },
		{ 363, 303, 50 },
		{ 363, 304, 50 },
		{ 363, 305, 50 },
		{ 363, 306, 50 },
		{ 363, 307, 50 },
		{ 363, 308, 50 },
		{ 363, 309, 50 },
		{ 363, 310, 50 },
		{ 363, 311, 50 },
		{ 363, 312, 50 },
		{ 363, 313, 50 },
		{ 363, 314, 50 },
		{ 363, 315, 50 },
		{ 363, 316, 50 },
		{ 363, 317, 50 },
		{ 363, 318, 50 },
		{ 363, 319, 50 },
		{ 363, 320, 50 },
		{ 363, 321, 50 },
		{ 363, 322, 50 },
		{ 363, 323, 50 },
		{ 364, 298, 50 },
		{ 364, 299, 50 },
		{ 364, 300, 50 },
		{ 364, 301, 50 },
		{ 364, 302, 50 },
		{ 364, 303, 50 },
		{ 364, 304, 50 },
		{ 364, 305, 50 },
		{ 364, 306, 50 },
		{ 364, 307, 50 },
		{ 364, 308, 50 },
		{ 364, 309, 50 },
		{ 364, 310, 50 },
		{ 364, 311, 50 },
		{ 364, 312, 50 },
		{ 364, 313, 50 },
		{ 364, 314, 50 },
		{ 364, 315, 50 },
		{ 364, 316, 50 },
		{ 364, 317, 50 },
		{ 364, 318, 50 },
		{ 364, 319, 50 },
		{ 364, 320, 50 },
		{ 364, 321, 50 },
		{ 364, 322, 50 },
		{ 364, 323, 50 },
		{ 365, 298, 50 },
		{ 365, 299, 50 },
		{ 365, 300, 50 },
		{ 365, 301, 50 },
		{ 365, 302, 50 },
		{ 365, 303, 50 },
		{ 365, 304, 50 },
		{ 365, 305, 50 },
		{ 365, 306, 50 },
		{ 365, 307, 50 },
		{ 365, 308, 50 },
		{ 365, 309, 50 },
		{ 365, 310, 50 },
		{ 365, 311, 50 },
		{ 365, 312, 50 },
		{ 365, 313, 50 },
		{ 365, 314, 50 },
		{ 365, 315, 50 },
		{ 365, 316, 50 },
		{ 365, 317, 50 },
		{ 365, 318, 50 },
		{ 365, 319, 50 },
		{ 365, 320, 50 },
		{ 365, 321, 50 },
		{ 365, 322, 50 },
		{ 365, 323, 50 },
		{ 365, 324, 50 },
		{ 366, 298, 50 },
		{ 366, 299, 50 },
		{ 366, 300, 50 },
		{ 366, 301, 50 },
		{ 366, 302, 50 },
		{ 366, 303, 50 },
		{ 366, 304, 50 },
		{ 366, 305, 50 },
		{ 366, 306, 50 },
		{ 366, 307, 50 },
		{ 366, 308, 50 },
		{ 366, 309, 50 },
		{ 366, 310, 50 },
		{ 366, 311, 50 },
		{ 366, 312, 50 },
		{ 366, 313, 50 },
		{ 366, 314, 50 },
		{ 366, 315, 50 },
		{ 366, 316, 50 },
		{ 366, 317, 50 },
		{ 366, 318, 50 },
		{ 366, 319, 50 },
		{ 366, 320, 50 },
		{ 366, 321, 50 },
		{ 366, 322, 50 },
		{ 366, 323, 50 },
		{ 366, 324, 50 },
		{ 366, 325, 50 },
		{ 367, 299, 50 },
		{ 367, 300, 50 },
		{ 367, 301, 50 },
		{ 367, 302, 50 },
		{ 367, 303, 50 },
		{ 367, 304, 50 },
		{ 367, 305, 50 },
		{ 367, 306, 50 },
		{ 367, 307, 50 },
		{ 367, 308, 50 },
		{ 367, 309, 50 },
		{ 367, 310, 50 },
		{ 367, 311, 50 },
		{ 367, 312, 50 },
		{ 367, 313, 50 },
		{ 367, 314, 50 },
		{ 367, 315, 50 },
		{ 367, 316, 50 },
		{ 367, 317, 50 },
		{ 367, 318, 50 },
		{ 367, 319, 50 },
		{ 367, 320, 50 },
		{ 367, 321, 50 },
		{ 367, 322, 50 },
		{ 367, 323, 50 },
		{ 367, 324, 50 },
		{ 367, 325, 50 },
		{ 368, 299, 50 },
		{ 368, 300, 50 },
		{ 368, 301, 50 },
		{ 368, 302, 50 },
		{ 368, 303, 50 },
		{ 368, 304, 50 },
		{ 368, 305, 50 },
		{ 368, 306, 50 },
		{ 368, 307, 50 },
		{ 368, 308, 50 },
		{ 368, 309, 50 },
		{ 368, 310, 50 },
		{ 368, 311, 50 },
		{ 368, 312, 50 },
		{ 368, 313, 50 },
		{ 368, 314, 50 },
		{ 368, 315, 50 },
		{ 368, 316, 50 },
		{ 368, 317, 50 },
		{ 368, 318, 50 },
		{ 368, 319, 50 },
		{ 368, 320, 50 },
		{ 368, 321, 50 },
		{ 368, 322, 50 },
		{ 368, 323, 50 },
		{ 368, 324, 50 },
		{ 368, 325, 50 },
		{ 368, 326, 50 },
		{ 369, 300, 50 },
		{ 369, 301, 50 },
		{ 369, 302, 50 },
		{ 369, 303, 50 },
		{ 369, 304, 50 },
		{ 369, 305, 50 },
		{ 369, 306, 50 },
		{ 369, 307, 50 },
		{ 369, 308, 50 },
		{ 369, 309, 50 },
		{ 369, 310, 50 },
		{ 369, 311, 50 },
		{ 369, 312, 50 },
		{ 369, 313, 50 },
		{ 369, 314, 50 },
		{ 369, 315, 50 },
		{ 369, 316, 50 },
		{ 369, 317, 50 },
		{ 369, 318, 50 },
		{ 369, 319, 50 },
		{ 369, 320, 50 },
		{ 369, 321, 50 },
		{ 369, 322, 50 },
		{ 369, 323, 50 },
		{ 369, 324, 50 },
		{ 369, 325, 50 },
		{ 369, 326, 50 },
		{ 370, 300, 50 },
		{ 370, 301, 50 },
		{ 370, 302, 50 },
		{ 370, 303, 50 },
		{ 370, 304, 50 },
		{ 370, 305, 50 },
		{ 370, 306, 50 },
		{ 370, 307, 50 },
		{ 370, 308, 50 },
		{ 370, 309, 50 },
		{ 370, 310, 50 },
		{ 370, 311, 50 },
		{ 370, 312, 50 },
		{ 370, 313, 50 },
		{ 370, 314, 50 },
		{ 370, 315, 50 },
		{ 370, 316, 50 },
		{ 370, 317, 50 },
		{ 370, 318, 50 },
		{ 370, 319, 50 },
		{ 370, 320, 50 },
		{ 370, 321, 50 },
		{ 370, 322, 50 },
		{ 370, 323, 50 },
		{ 370, 324, 50 },
		{ 370, 325, 50 },
		{ 370, 326, 50 },
		{ 370, 327, 50 },
		{ 371, 300, 50 },
		{ 371, 301, 50 },
		{ 371, 302, 50 },
		{ 371, 303, 50 },
		{ 371, 304, 50 },
		{ 371, 305, 50 },
		{ 371, 306, 50 },
		{ 371, 307, 50 },
		{ 371, 308, 50 },
		{ 371, 309, 50 },
		{ 371, 310, 50 },
		{ 371, 311, 50 },
		{ 371, 312, 50 },
		{ 371, 313, 50 },
		{ 371, 314, 50 },
		{ 371, 315, 50 },
		{ 371, 316, 50 },
		{ 371, 317, 50 },
		{ 371, 318, 50 },
		{ 371, 319, 50 },
		{ 371, 320, 50 },
		{ 371, 321, 50 },
		{ 371, 322, 50 },
		{ 371, 323, 50 },
		{ 371, 324, 50 },
		{ 371, 325, 50 },
		{ 371, 326, 50 },
		{ 371, 327, 50 },
		{ 371, 328, 50 },
		{ 372, 301, 50 },
		{ 372, 302, 50 },
		{ 372, 303, 50 },
		{ 372, 304, 50 },
		{ 372, 305, 50 },
		{ 372, 306, 50 },
		{ 372, 307, 50 },
		{ 372, 308, 50 },
		{ 372, 309, 50 },
		{ 372, 310, 50 },
		{ 372, 311, 50 },
		{ 372, 312, 50 },
		{ 372, 313, 50 },
		{ 372, 314, 50 },
		{ 372, 315, 50 },
		{ 372, 316, 50 },
		{ 372, 317, 50 },
		{ 372, 318, 50 },
		{ 372, 319, 50 },
		{ 372, 320, 50 },
		{ 372, 321, 50 },
		{ 372, 322, 50 },
		{ 372, 323, 50 },
		{ 372, 324, 50 },
		{ 372, 325, 50 },
		{ 372, 326, 50 },
		{ 372, 327, 50 },
		{ 372, 328, 50 },
		{ 373, 301, 50 },
		{ 373, 302, 50 },
		{ 373, 303, 50 },
		{ 373, 304, 50 },
		{ 373, 305, 50 },
		{ 373, 306, 50 },
		{ 373, 307, 50 },
		{ 373, 308, 50 },
		{ 373, 309, 50 },
		{ 373, 310, 50 },
		{ 373, 311, 50 },
		{ 373, 312, 50 },
		{ 373, 313, 50 },
		{ 373, 314, 50 },
		{ 373, 315, 50 },
		{ 373, 316, 50 },
		{ 373, 317, 50 },
		{ 373, 318, 50 },
		{ 373, 319, 50 },
		{ 373, 320, 50 },
		{ 373, 321, 50 },
		{ 373, 322, 50 },
		{ 373, 323, 50 },
		{ 373, 324, 50 },
		{ 373, 325, 50 },
		{ 373, 326, 50 },
		{ 373, 327, 50 },
		{ 373, 328, 50 },
		{ 373, 329, 50 },
		{ 374, 301, 50 },
		{ 374, 302, 50 },
		{ 374, 303, 50 },
		{ 374, 304, 50 },
		{ 374, 305, 50 },
		{ 374, 306, 50 },
		{ 374, 307, 50 },
		{ 374, 308, 50 },
		{ 374, 309, 50 },
		{ 374, 310, 50 },
		{ 374, 311, 50 },
		{ 374, 312, 50 },
		{ 374, 313, 50 },
		{ 374, 314, 50 },
		{ 374, 315, 50 },
		{ 374, 316, 50 },
		{ 374, 317, 50 },
		{ 374, 318, 50 },
		{ 374, 319, 50 },
		{ 374, 320, 50 },
		{ 374, 321, 50 },
		{ 374, 322, 50 },
		{ 374, 323, 50 },
		{ 374, 324, 50 },
		{ 374, 325, 50 },
		{ 374, 326, 50 },
		{ 374, 327, 50 },
		{ 374, 328, 50 },
		{ 374, 329, 50 },
		{ 374, 330, 50 },
		{ 375, 302, 50 },
		{ 375, 303, 50 },
		{ 375, 304, 50 },
		{ 375, 305, 50 },
		{ 375, 306, 50 },
		{ 375, 307, 50 },
		{ 375, 308, 50 },
		{ 375, 309, 50 },
		{ 375, 310, 50 },
		{ 375, 311, 50 },
		{ 375, 312, 50 },
		{ 375, 313, 50 },
		{ 375, 314, 50 },
		{ 375, 315, 50 },
		{ 375, 316, 50 },
		{ 375, 317, 50 },
		{ 375, 318, 50 },
		{ 375, 319, 50 },
		{ 375, 320, 50 },
		{ 375, 321, 50 },
		{ 375, 322, 50 },
		{ 375, 323, 50 },
		{ 375, 324, 50 },
		{ 375, 325, 50 },
		{ 375, 326, 50 },
		{ 375, 327, 50 },
		{ 375, 328, 50 },
		{ 375, 329, 50 },
		{ 375, 330, 50 },
		{ 376, 302, 50 },
		{ 376, 303, 50 },
		{ 376, 304, 50 },
		{ 376, 305, 50 },
		{ 376, 306, 50 },
		{ 376, 307, 50 },
		{ 376, 308, 50 },
		{ 376, 309, 50 },
		{ 376, 310, 50 },
		{ 376, 311, 50 },
		{ 376, 312, 50 },
		{ 376, 313, 50 },
		{ 376, 314, 50 },
		{ 376, 315, 50 },
		{ 376, 316, 50 },
		{ 376, 317, 50 },
		{ 376, 318, 50 },
		{ 376, 319, 50 },
		{ 376, 320, 50 },
		{ 376, 321, 50 },
		{ 376, 322, 50 },
		{ 376, 323, 50 },
		{ 376, 324, 50 },
		{ 376, 325, 50 },
		{ 376, 326, 50 },
		{ 376, 327, 50 },
		{ 376, 328, 50 },
		{ 376, 329, 50 },
		{ 376, 330, 50 },
		{ 376, 331, 50 },
		{ 377, 303, 50 },
		{ 377, 304, 50 },
		{ 377, 305, 50 },
		{ 377, 306, 50 },
		{ 377, 307, 50 },
		{ 377, 308, 50 },
		{ 377, 309, 50 },
		{ 377, 310, 50 },
		{ 377, 311, 50 },
		{ 377, 312, 50 },
		{ 377, 313, 50 },
		{ 377, 314, 50 },
		{ 377, 315, 50 },
		{ 377, 316, 50 },
		{ 377, 317, 50 },
		{ 377, 318, 50 },
		{ 377, 319, 50 },
		{ 377, 320, 50 },
		{ 377, 321, 50 },
		{ 377, 322, 50 },
		{ 377, 323, 50 },
		{ 377, 324, 50 },
		{ 377, 325, 50 },
		{ 377, 326, 50 },
		{ 377, 327, 50 },
		{ 377, 328, 50 },
		{ 377, 329, 50 },
		{ 377, 330, 50 },
		{ 377, 331, 50 },
		{ 378, 303, 50 },
		{ 378, 304, 50 },
		{ 378, 305, 50 },
		{ 378, 306, 50 },
		{ 378, 307, 50 },
		{ 378, 308, 50 },
		{ 378, 309, 50 },
		{ 378, 310, 50 },
		{ 378, 311, 50 },
		{ 378, 312, 50 },
		{ 378, 313, 50 },
		{ 378, 314, 50 },
		{ 378, 315, 50 },
		{ 378, 316, 50 },
		{ 378, 317, 50 },
		{ 378, 318, 50 },
		{ 378, 319, 50 },
		{ 378, 320, 50 },
		{ 378, 321, 50 },
		{ 378, 322, 50 },
		{ 378, 323, 50 },
		{ 378, 324, 50 },
		{ 378, 325, 50 },
		{ 378, 326, 50 },
		{ 378, 327, 50 },
		{ 378, 328, 50 },
		{ 378, 329, 50 },
		{ 378, 330, 50 },
		{ 378, 331, 50 },
		{ 378, 332, 50 },
		{ 379, 303, 50 },
		{ 379, 304, 50 },
		{ 379, 305, 50 },
		{ 379, 306, 50 },
		{ 379, 307, 50 },
		{ 379, 308, 50 },
		{ 379, 309, 50 },
		{ 379, 310, 50 },
		{ 379, 311, 50 },
		{ 379, 312, 50 },
		{ 379, 313, 50 },
		{ 379, 314, 50 },
		{ 379, 315, 50 },
		{ 379, 316, 50 },
		{ 379, 317, 50 },
		{ 379, 318, 50 },
		{ 379, 319, 50 },
		{ 379, 320, 50 },
		{ 379, 321, 50 },
		{ 379, 322, 50 },
		{ 379, 323, 50 },
		{ 379, 324, 50 },
		{ 379, 325, 50 },
		{ 379, 326, 50 },
		{ 379, 327, 50 },
		{ 379, 328, 50 },
		{ 379, 329, 50 },
		{ 379, 330, 50 },
		{ 379, 331, 50 },
		{ 379, 332, 50 },
		{ 379, 333, 50 },
		{ 380, 304, 50 },
		{ 380, 305, 50 },
		{ 380, 306, 50 },
		{ 380, 307, 50 },
		{ 380, 308, 50 },
		{ 380, 309, 50 },
		{ 380, 310, 50 },
		{ 380, 311, 50 },
		{ 380, 312, 50 },
		{ 380, 313, 50 },
		{ 380, 314, 50 },
		{ 380, 315, 50 },
		{ 380, 316, 50 },
		{ 380, 317, 50 },
		{ 380, 318, 50 },
		{ 380, 319, 50 },
		{ 380, 320, 50 },
		{ 380, 321, 50 },
		{ 380, 322, 50 },
		{ 380, 323, 50 },
		{ 380, 324, 50 },
		{ 380, 325, 50 },
		{ 380, 326, 50 },
		{ 380, 327, 50 },
		{ 380, 328, 50 },
		{ 380, 329, 50 },
		{ 380, 330, 50 },
		{ 380, 331, 50 },
		{ 380, 332, 50 },
		{ 380, 333, 50 },
		{ 381, 304, 50 },
		{ 381, 305, 50 },
		{ 381, 306, 50 },
		{ 381, 307, 50 },
		{ 381, 308, 50 },
		{ 381, 309, 50 },
		{ 381, 310, 50 },
		{ 381, 311, 50 },
		{ 381, 312, 50 },
		{ 381, 313, 50 },
		{ 381, 314, 50 },
		{ 381, 315, 50 },
		{ 381, 316, 50 },
		{ 381, 317, 50 },
		{ 381, 318, 50 },
		{ 381, 319, 50 },
		{ 381, 320, 50 },
		{ 381, 321, 50 },
		{ 381, 322, 50 },
		{ 381, 323, 50 },
		{ 381, 324, 50 },
		{ 381, 325, 50 },
		{ 381, 326, 50 },
		{ 381, 327, 50 },
		{ 381, 328, 50 },
		{ 381, 329, 50 },
		{ 381, 330, 50 },
		{ 381, 331, 50 },
		{ 381, 332, 50 },
		{ 381, 333, 50 },
		{ 381, 334, 50 },
		{ 382, 304, 50 },
		{ 382, 305, 50 },
		{ 382, 306, 50 },
		{ 382, 307, 50 },
		{ 382, 308, 50 },
		{ 382, 309, 50 },
		{ 382, 310, 50 },
		{ 382, 311, 50 },
		{ 382, 312, 50 },
		{ 382, 313, 50 },
		{ 382, 314, 50 },
		{ 382, 315, 50 },
		{ 382, 316, 50 },
		{ 382, 317, 50 },
		{ 382, 318, 50 },
		{ 382, 319, 50 },
		{ 382, 320, 50 },
		{ 382, 321, 50 },
		{ 382, 322, 50 },
		{ 382, 323, 50 },
		{ 382, 324, 50 },
		{ 382, 325, 50 },
		{ 382, 326, 50 },
		{ 382, 327, 50 },
		{ 382, 328, 50 },
		{ 382, 329, 50 },
		{ 382, 330, 50 },
		{ 382, 331, 50 },
		{ 382, 332, 50 },
		{ 382, 333, 50 },
		{ 382, 334, 50 },
		{ 382, 335, 50 },
		{ 383, 305, 50 },
		{ 383, 306, 50 },
		{ 383, 307, 50 },
		{ 383, 308, 50 },
		{ 383, 309, 50 },
		{ 383, 310, 50 },
		{ 383, 311, 50 },
		{ 383, 312, 50 },
		{ 383, 313, 50 },
		{ 383, 314, 50 },
		{ 383, 315, 50 },
		{ 383, 316, 50 },
		{ 383, 317, 50 },
		{ 383, 318, 50 },
		{ 383, 319, 50 },
		{ 383, 320, 50 },
		{ 383, 321, 50 },
		{ 383, 322, 50 },
		{ 383, 323, 50 },
		{ 383, 324, 50 },
		{ 383, 325, 50 },
		{ 383, 326, 50 },
		{ 383, 327, 50 },
		{ 383, 328, 50 },
		{ 383, 329, 50 },
		{ 383, 330, 50 },
		{ 383, 331, 50 },
		{ 383, 332, 50 },
		{ 383, 333, 50 },
		{ 383, 334, 50 },
		{ 383, 335, 50 },
		{ 384, 305, 50 },
		{ 384, 306, 50 },
		{ 384, 307, 50 },
		{ 384, 308, 50 },
		{ 384, 309, 50 },
		{ 384, 310, 50 },
		{ 384, 311, 50 },
		{ 384, 312, 50 },
		{ 384, 313, 50 },
		{ 384, 314, 50 },
		{ 384, 315, 50 },
		{ 384, 316, 50 },
		{ 384, 317, 50 },
		{ 384, 318, 50 },
		{ 384, 319, 50 },
		{ 384, 320, 50 },
		{ 384, 321, 50 },
		{ 384, 322, 50 },
		{ 384, 323, 50 },
		{ 384, 324, 50 },
		{ 384, 325, 50 },
		{ 384, 326, 50 },
		{ 384, 327, 50 },
		{ 384, 328, 50 },
		{ 384, 329, 50 },
		{ 384, 330, 50 },
		{ 384, 331, 50 },
		{ 384, 332, 50 },
		{ 384, 333, 50 },
		{ 384, 334, 50 },
		{ 384, 335, 50 },
		{ 384, 336, 50 },
		{ 385, 306, 50 },
		{ 385, 307, 50 },
		{ 385, 308, 50 },
		{ 385, 309, 50 },
		{ 385, 310, 50 },
		{ 385, 311, 50 },
		{ 385, 312, 50 },
		{ 385, 313, 50 },
		{ 385, 314, 50 },
		{ 385, 315, 50 },
		{ 385, 316, 50 },
		{ 385, 317, 50 },
		{ 385, 318, 50 },
		{ 385, 319, 50 },
		{ 385, 320, 50 },
		{ 385, 321, 50 },
		{ 385, 322, 50 },
		{ 385, 323, 50 },
		{ 385, 324, 50 },
		{ 385, 325, 50 },
		{ 385, 326, 50 },
		{ 385, 327, 50 },
		{ 385, 328, 50 },
		{ 385, 329, 50 },
		{ 385, 330, 50 },
		{ 385, 331, 50 },
		{ 385, 332, 50 },
		{ 385, 333, 50 },
		{ 385, 334, 50 },
		{ 385, 335, 50 },
		{ 385, 336, 50 },
		{ 386, 306, 50 },
		{ 386, 307, 50 },
		{ 386, 308, 50 },
		{ 386, 309, 50 },
		{ 386, 310, 50 },
		{ 386, 311, 50 },
		{ 386, 312, 50 },
		{ 386, 313, 50 },
		{ 386, 314, 50 },
		{ 386, 315, 50 },
		{ 386, 316, 50 },
		{ 386, 317, 50 },
		{ 386, 318, 50 },
		{ 386, 319, 50 },
		{ 386, 320, 50 },
		{ 386, 321, 50 },
		{ 386, 322, 50 },
		{ 386, 323, 50 },
		{ 386, 324, 50 },
		{ 386, 325, 50 },
		{ 386, 326, 50 },
		{ 386, 327, 50 },
		{ 386, 328, 50 },
		{ 386, 329, 50 },
		{ 386, 330, 50 },
		{ 386, 331, 50 },
		{ 386, 332, 50 },
		{ 386, 333, 50 },
		{ 386, 334, 50 },
		{ 386, 335, 50 },
		{ 386, 336, 50 },
		{ 386, 337, 50 },
		{ 387, 306, 50 },
		{ 387, 307, 50 },
		{ 387, 308, 50 },
		{ 387, 309, 50 },
		{ 387, 310, 50 },
		{ 387, 311, 50 },
		{ 387, 312, 50 },
		{ 387, 313, 50 },
		{ 387, 314, 50 },
		{ 387, 315, 50 },
		{ 387, 316, 50 },
		{ 387, 317, 50 },
		{ 387, 318, 50 },
		{ 387, 319, 50 },
		{ 387, 320, 50 },
		{ 387, 321, 50 },
		{ 387, 322, 50 },
		{ 387, 323, 50 },
		{ 387, 324, 50 },
		{ 387, 325, 50 },
		{ 387, 326, 50 },
		{ 387, 327, 50 },
		{ 387, 328, 50 },
		{ 387, 329, 50 },
		{ 387, 330, 50 },
		{ 387, 331, 50 },
		{ 387, 332, 50 },
		{ 387, 333, 50 },
		{ 387, 334, 50 },
		{ 387, 335, 50 },
		{ 387, 336, 50 },
		{ 387, 337, 50 },
		{ 387, 338, 50 },
		{ 388, 307, 50 },
		{ 388, 308, 50 },
		{ 388, 309, 50 },
		{ 388, 310, 50 },
		{ 388, 311, 50 },
		{ 388, 312, 50 },
		{ 388, 313, 50 },
		{ 388, 314, 50 },
		{ 388, 315, 50 },
		{ 388, 316, 50 },
		{ 388, 317, 50 },
		{ 388, 318, 50 },
		{ 388, 319, 50 },
		{ 388, 320, 50 },
		{ 388, 321, 50 },
		{ 388, 322, 50 },
		{ 388, 323, 50 },
		{ 388, 324, 50 },
		{ 388, 325, 50 },
		{ 388, 326, 50 },
		{ 388, 327, 50 },
		{ 388, 328, 50 },
		{ 388, 329, 50 },
		{ 388, 330, 50 },
		{ 388, 331, 50 },
		{ 388, 332, 50 },
		{ 388, 333, 50 },
		{ 388, 334, 50 },
		{ 388, 335, 50 },
		{ 388, 336, 50 },
		{ 388, 337, 50 },
		{ 388, 338, 50 },
		{ 389, 307, 50 },
		{ 389, 308, 50 },
		{ 389, 309, 50 },
		{ 389, 310, 50 },
		{ 389, 311, 50 },
		{ 389, 312, 50 },
		{ 389, 313, 50 },
		{ 389, 314, 50 },
		{ 389, 315, 50 },
		{ 389, 316, 50 },
		{ 389, 317, 50 },
		{ 389, 318, 50 },
		{ 389, 319, 50 },
		{ 389, 320, 50 },
		{ 389, 321, 50 },
		{ 389, 322, 50 },
		{ 389, 323, 50 },
		{ 389, 324, 50 },
		{ 389, 325, 50 },
		{ 389, 326, 50 },
		{ 389, 327, 50 },
		{ 389, 328, 50 },
		{ 389, 329, 50 },
		{ 389, 330, 50 },
		{ 389, 331, 50 },
		{ 389, 332, 50 },
		{ 389, 333, 50 },
		{ 389, 334, 50 },
		{ 389, 335, 50 },
		{ 389, 336, 50 },
		{ 389, 337, 50 },
		{ 389, 338, 50 },
		{ 389, 339, 50 },
		{ 390, 307, 50 },
		{ 390, 308, 50 },
		{ 390, 309, 50 },
		{ 390, 310, 50 },
		{ 390, 311, 50 },
		{ 390, 312, 50 },
		{ 390, 313, 50 },
		{ 390, 314, 50 },
		{ 390, 315, 50 },
		{ 390, 316, 50 },
		{ 390, 317, 50 },
		{ 390, 318, 50 },
		{ 390, 319, 50 },
		{ 390, 320, 50 },
		{ 390, 321, 50 },
		{ 390, 322, 50 },
		{ 390, 323, 50 },
		{ 390, 324, 50 },
		{ 390, 325, 50 },
		{ 390, 326, 50 },
		{ 390, 327, 50 },
		{ 390, 328, 50 },
		{ 390, 329, 50 },
		{ 390, 330, 50 },
		{ 390, 331, 50 },
		{ 390, 332, 50 },
		{ 390, 333, 50 },
		{ 390, 334, 50 },
		{ 390, 335, 50 },
		{ 390, 336, 50 },
		{ 390, 337, 50 },
		{ 390, 338, 50 },
		{ 390, 339, 50 },
		{ 390, 340, 50 },
		{ 391, 308, 50 },
		{ 391, 309, 50 },
		{ 391, 310, 50 },
		{ 391, 311, 50 },
		{ 391, 312, 50 },
		{ 391, 313, 50 },
		{ 391, 314, 50 },
		{ 391, 315, 50 },
		{ 391, 316, 50 },
		{ 391, 317, 50 },
		{ 391, 318, 50 },
		{ 391, 319, 50 },
		{ 391, 320, 50 },
		{ 391, 321, 50 },
		{ 391, 322, 50 },
		{ 391, 323, 50 },
		{ 391, 324, 50 },
		{ 391, 325, 50 },
		{ 391, 326, 50 },
		{ 391, 327, 50 },
		{ 391, 328, 50 },
		{ 391, 329, 50 },
		{ 391, 330, 50 },
		{ 391, 331, 50 },
		{ 391, 332, 50 },
		{ 391, 333, 50 },
		{ 391, 334, 50 },
		{ 391, 335, 50 },
		{ 391, 336, 50 },
		{ 391, 337, 50 },
		{ 391, 338, 50 },
		{ 391, 339, 50 },
		{ 391, 340, 50 },
		{ 392, 308, 50 },
		{ 392, 309, 50 },
		{ 392, 310, 50 },
		{ 392, 311, 50 },
		{ 392, 312, 50 },
		{ 392, 313, 50 },
		{ 392, 314, 50 },
		{ 392, 315, 50 },
		{ 392, 316, 50 },
		{ 392, 317, 50 },
		{ 392, 318, 50 },
		{ 392, 319, 50 },
		{ 392, 320, 50 },
		{ 392, 321, 50 },
		{ 392, 322, 50 },
		{ 392, 323, 50 },
		{ 392, 324, 50 },
		{ 392, 325, 50 },
		{ 392, 326, 50 },
		{ 392, 327, 50 },
		{ 392, 328, 50 },
		{ 392, 329, 50 },
		{ 392, 330, 50 },
		{ 392, 331, 50 },
		{ 392, 332, 50 },
		{ 392, 333, 50 },
		{ 392, 334, 50 },
		{ 392, 335, 50 },
		{ 392, 336, 50 },
		{ 392, 337, 50 },
		{ 392, 338, 50 },
		{ 392, 339, 50 },
		{ 392, 340, 50 },
		{ 392, 341, 50 },
		{ 393, 309, 50 },
		{ 393, 310, 50 },
		{ 393, 311, 50 },
		{ 393, 312, 50 },
		{ 393, 313, 50 },
		{ 393, 314, 50 },
		{ 393, 315, 50 },
		{ 393, 316, 50 },
		{ 393, 317, 50 },
		{ 393, 318, 50 },
		{ 393, 319, 50 },
		{ 393, 320, 50 },
		{ 393, 321, 50 },
		{ 393, 322, 50 },
		{ 393, 323, 50 },
		{ 393, 324, 50 },
		{ 393, 325, 50 },
		{ 393, 326, 50 },
		{ 393, 327, 50 },
		{ 393, 328, 50 },
		{ 393, 329, 50 },
		{ 393, 330, 50 },
		{ 393, 331, 50 },
		{ 393, 332, 50 },
		{ 393, 333, 50 },
		{ 393, 334, 50 },
		{ 393, 335, 50 },
		{ 393, 336, 50 },
		{ 393, 337, 50 },
		{ 393, 338, 50 },
		{ 393, 339, 50 },
		{ 393, 340, 50 },
		{ 393, 341, 50 },
		{ 394, 309, 50 },
		{ 394, 310, 50 },
		{ 394, 311, 50 },
		{ 394, 312, 50 },
		{ 394, 313, 50 },
		{ 394, 314, 50 },
		{ 394, 315, 50 },
		{ 394, 316, 50 },
		{ 394, 317, 50 },
		{ 394, 318, 50 },
		{ 394, 319, 50 },
		{ 394, 320, 50 },
		{ 394, 321, 50 },
		{ 394, 322, 50 },
		{ 394, 323, 50 },
		{ 394, 324, 50 },
		{ 394, 325, 50 },
		{ 394, 326, 50 },
		{ 394, 327, 50 },
		{ 394, 328, 50 },
		{ 394, 329, 50 },
		{ 394, 330, 50 },
		{ 394, 331, 50 },
		{ 394, 332, 50 },
		{ 394, 333, 50 },
		{ 394, 334, 50 },
		{ 394, 335, 50 },
		{ 394, 336, 50 },
		{ 394, 337, 50 },
		{ 394, 338, 50 },
		{ 394, 339, 50 },
		{ 394, 340, 50 },
		{ 394, 341, 50 },
		{ 394, 342, 50 },
		{ 395, 309, 50 },
		{ 395, 310, 50 },
		{ 395, 311, 50 },
		{ 395, 312, 50 },
		{ 395, 313, 50 },
		{ 395, 314, 50 },
		{ 395, 315, 50 },
		{ 395, 316, 50 },
		{ 395, 317, 50 },
		{ 395, 318, 50 },
		{ 395, 319, 50 },
		{ 395, 320, 50 },
		{ 395, 321, 50 },
		{ 395, 322, 50 },
		{ 395, 323, 50 },
		{ 395, 324, 50 },
		{ 395, 325, 50 },
		{ 395, 326, 50 },
		{ 395, 327, 50 },
		{ 395, 328, 50 },
		{ 395, 329, 50 },
		{ 395, 330, 50 },
		{ 395, 331, 50 },
		{ 395, 332, 50 },
		{ 395, 333, 50 },
		{ 395, 334, 50 },
		{ 395, 335, 50 },
		{ 395, 336, 50 },
		{ 395, 337, 50 },
		{ 395, 338, 50 },
		{ 395, 339, 50 },
		{ 395, 340, 50 },
		{ 395, 341, 50 },
		{ 395, 342, 50 },
		{ 395, 343, 50 },
		{ 396, 310, 50 },
		{ 396, 311, 50 },
		{ 396, 312, 50 },
		{ 396, 313, 50 },
		{ 396, 314, 50 },
		{ 396, 315, 50 },
		{ 396, 316, 50 },
		{ 396, 317, 50 },
		{ 396, 318, 50 },
		{ 396, 319, 50 },
		{ 396, 320, 50 },
		{ 396, 321, 50 },
		{ 396, 322, 50 },
		{ 396, 323, 50 },
		{ 396, 324, 50 },
		{ 396, 325, 50 },
		{ 396, 326, 50 },
		{ 396, 327, 50 },
		{ 396, 328, 50 },
		{ 396, 329, 50 },
		{ 396, 330, 50 },
		{ 396, 331, 50 },
		{ 396, 332, 50 },
		{ 396, 333, 50 },
		{ 396, 334, 50 },
		{ 396, 335, 50 },
		{ 396, 336, 50 },
		{ 396, 337, 50 },
		{ 396, 338, 50 },
		{ 396, 339, 50 },
		{ 396, 340, 50 },
		{ 396, 341, 50 },
		{ 396, 342, 50 },
		{ 396, 343, 50 },
		{ 397, 310, 50 },
		{ 397, 311, 50 },
		{ 397, 312, 50 },
		{ 397, 313, 50 },
		{ 397, 314, 50 },
		{ 397, 315, 50 },
		{ 397, 316, 50 },
		{ 397, 317, 50 },
		{ 397, 318, 50 },
		{ 397, 319, 50 },
		{ 397, 320, 50 },
		{ 397, 321, 50 },
		{ 397, 322, 50 },
		{ 397, 323, 50 },
		{ 397, 324, 50 },
		{ 397, 325, 50 },
		{ 397, 326, 50 },
		{ 397, 327, 50 },
		{ 397, 328, 50 },
		{ 397, 329, 50 },
		{ 397, 330, 50 },
		{ 397, 331, 50 },
		{ 397, 332, 50 },
		{ 397, 333, 50 },
		{ 397, 334, 50 },
		{ 397, 335, 50 },
		{ 397, 336, 50 },
		{ 397, 337, 50 },
		{ 397, 338, 50 },
		{ 397, 339, 50 },
		{ 397, 340, 50 },
		{ 397, 341, 50 },
		{ 397, 342, 50 },
		{ 397, 343, 50 },
		{ 397, 344, 50 },
		{ 398, 310, 50 },
		{ 398, 311, 50 },
		{ 398, 312, 50 },
		{ 398, 313, 50 },
		{ 398, 314, 50 },
		{ 398, 315, 50 },
		{ 398, 316, 50 },
		{ 398, 317, 50 },
		{ 398, 318, 50 },
		{ 398, 319, 50 },
		{ 398, 320, 50 },
		{ 398, 321, 50 },
		{ 398, 322, 50 },
		{ 398, 323, 50 },
		{ 398, 324, 50 },
		{ 398, 325, 50 },
		{ 398, 326, 50 },
		{ 398, 327, 50 },
		{ 398, 328, 50 },
		{ 398, 329, 50 },
		{ 398, 330, 50 },
		{ 398, 331, 50 },
		{ 398, 332, 50 },
		{ 398, 333, 50 },
		{ 398, 334, 50 },
		{ 398, 335, 50 },
		{ 398, 336, 50 },
		{ 398, 337, 50 },
		{ 398, 338, 50 },
		{ 398, 339, 50 },
		{ 398, 340, 50 },
		{ 398, 341, 50 },
		{ 398, 342, 50 },
		{ 398, 343, 50 },
		{ 398, 344, 50 },
		{ 398, 345, 50 },
		{ 399, 311, 50 },
		{ 399, 312, 50 },
		{ 399, 313, 50 },
		{ 399, 314, 50 },
		{ 399, 315, 50 },
		{ 399, 316, 50 },
		{ 399, 317, 50 },
		{ 399, 318, 50 },
		{ 399, 319, 50 },
		{ 399, 320, 50 },
		{ 399, 321, 50 },
		{ 399, 322, 50 },
		{ 399, 323, 50 },
		{ 399, 324, 50 },
		{ 399, 325, 50 },
		{ 399, 326, 50 },
		{ 399, 327, 50 },
		{ 399, 328, 50 },
		{ 399, 329, 50 },
		{ 399, 330, 50 },
		{ 399, 331, 50 },
		{ 399, 332, 50 },
		{ 399, 333, 50 },
		{ 399, 334, 50 },
		{ 399, 335, 50 },
		{ 399, 336, 50 },
		{ 399, 337, 50 },
		{ 399, 338, 50 },
		{ 399, 339, 50 },
		{ 399, 340, 50 },
		{ 399, 341, 50 },
		{ 399, 342, 50 },
		{ 399, 343, 50 },
		{ 399, 344, 50 },
		{ 399, 345, 50 },
		{ 400, 311, 50 },
		{ 400, 312, 50 },
		{ 400, 313, 50 },
		{ 400, 314, 50 },
		{ 400, 315, 50 },
		{ 400, 316, 50 },
		{ 400, 317, 50 },
		{ 400, 318, 50 },
		{ 400, 319, 50 },
		{ 400, 320, 50 },
		{ 400, 321, 50 },
		{ 400, 322, 50 },
		{ 400, 323, 50 },
		{ 400, 324, 50 },
		{ 400, 325, 50 },
		{ 400, 326, 50 },
		{ 400, 327, 50 },
		{ 400, 328, 50 },
		{ 400, 329, 50 },
		{ 400, 330, 50 },
		{ 400, 331, 50 },
		{ 400, 332, 50 },
		{ 400, 333, 50 },
		{ 400, 334, 50 },
		{ 400, 335, 50 },
		{ 400, 336, 50 },
		{ 400, 337, 50 },
		{ 400, 338, 50 },
		{ 400, 339, 50 },
		{ 400, 340, 50 },
		{ 400, 341, 50 },
		{ 400, 342, 50 },
		{ 400, 343, 50 },
		{ 400, 344, 50 },
		{ 400, 345, 50 },
		{ 400, 346, 50 },
		{ 401, 312, 50 },
		{ 401, 313, 50 },
		{ 401, 314, 50 },
		{ 401, 315, 50 },
		{ 401, 316, 50 },
		{ 401, 317, 50 },
		{ 401, 318, 50 },
		{ 401, 319, 50 },
		{ 401, 320, 50 },
		{ 401, 321, 50 },
		{ 401, 322, 50 },
		{ 401, 323, 50 },
		{ 401, 324, 50 },
		{ 401, 325, 50 },
		{ 401, 326, 50 },
		{ 401, 327, 50 },
		{ 401, 328, 50 },
		{ 401, 329, 50 },
		{ 401, 330, 50 },
		{ 401, 331, 50 },
		{ 401, 332, 50 },
		{ 401, 333, 50 },
		{ 401, 334, 50 },
		{ 401, 335, 50 },
		{ 401, 336, 50 },
		{ 401, 337, 50 },
		{ 401, 338, 50 },
		{ 401, 339, 50 },
		{ 401, 340, 50 },
		{ 401, 341, 50 },
		{ 401, 342, 50 },
		{ 401, 343, 50 },
		{ 401, 344, 50 },
		{ 401, 345, 50 },
		{ 401, 346, 50 },
		{ 402, 312, 50 },
		{ 402, 313, 50 },
		{ 402, 314, 50 },
		{ 402, 315, 50 },
		{ 402, 316, 50 },
		{ 402, 317, 50 },
		{ 402, 318, 50 },
		{ 402, 319, 50 },
		{ 402, 320, 50 },
		{ 402, 321, 50 },
		{ 402, 322, 50 },
		{ 402, 323, 50 },
		{ 402, 324, 50 },
		{ 402, 325, 50 },
		{ 402, 326, 50 },
		{ 402, 327, 50 },
		{ 402, 328, 50 },
		{ 402, 329, 50 },
		{ 402, 330, 50 },
		{ 402, 331, 50 },
		{ 402, 332, 50 },
		{ 402, 333, 50 },
		{ 402, 334, 50 },
		{ 402, 335, 50 },
		{ 402, 336, 50 },
		{ 402, 337, 50 },
		{ 402, 338, 50 },
		{ 402, 339, 50 },
		{ 402, 340, 50 },
		{ 402, 341, 50 },
		{ 402, 342, 50 },
		{ 402, 343, 50 },
		{ 402, 344, 50 },
		{ 402, 345, 50 },
		{ 402, 346, 50 },
		{ 402, 347, 50 },
		{ 403, 312, 50 },
		{ 403, 313, 50 },
		{ 403, 314, 50 },
		{ 403, 315, 50 },
		{ 403, 316, 50 },
		{ 403, 317, 50 },
		{ 403, 318, 50 },
		{ 403, 319, 50 },
		{ 403, 320, 50 },
		{ 403, 321, 50 },
		{ 403, 322, 50 },
		{ 403, 323, 50 },
		{ 403, 324, 50 },
		{ 403, 325, 50 },
		{ 403, 326, 50 },
		{ 403, 327, 50 },
		{ 403, 328, 50 },
		{ 403, 329, 50 },
		{ 403, 330, 50 },
		{ 403, 331, 50 },
		{ 403, 332, 50 },
		{ 403, 333, 50 },
		{ 403, 334, 50 },
		{ 403, 335, 50 },
		{ 403, 336, 50 },
		{ 403, 337, 50 },
		{ 403, 338, 50 },
		{ 403, 339, 50 },
		{ 403, 340, 50 },
		{ 403, 341, 50 },
		{ 403, 342, 50 },
		{ 403, 343, 50 },
		{ 403, 344, 50 },
		{ 403, 345, 50 },
		{ 403, 346, 50 },
		{ 403, 347, 50 },
		{ 403, 348, 50 },
		{ 404, 313, 50 },
		{ 404, 314, 50 },
		{ 404, 315, 50 },
		{ 404, 316, 50 },
		{ 404, 317, 50 },
		{ 404, 318, 50 },
		{ 404, 319, 50 },
		{ 404, 320, 50 },
		{ 404, 321, 50 },
		{ 404, 322, 50 },
		{ 404, 323, 50 },
		{ 404, 324, 50 },
		{ 404, 325, 50 },
		{ 404, 326, 50 },
		{ 404, 327, 50 },
		{ 404, 328, 50 },
		{ 404, 329, 50 },
		{ 404, 330, 50 },
		{ 404, 331, 50 },
		{ 404, 332, 50 },
		{ 404, 333, 50 },
		{ 404, 334, 50 },
		{ 404, 335, 50 },
		{ 404, 336, 50 },
		{ 404, 337, 50 },
		{ 404, 338, 50 },
		{ 404, 339, 50 },
		{ 404, 340, 50 },
		{ 404, 341, 50 },
		{ 404, 342, 50 },
		{ 404, 343, 50 },
		{ 404, 344, 50 },
		{ 404, 345, 50 },
		{ 404, 346, 50 },
		{ 404, 347, 50 },
		{ 404, 348, 50 },
		{ 405, 313, 50 },
		{ 405, 314, 50 },
		{ 405, 315, 50 },
		{ 405, 316, 50 },
		{ 405, 317, 50 },
		{ 405, 318, 50 },
		{ 405, 319, 50 },
		{ 405, 320, 50 },
		{ 405, 321, 50 },
		{ 405, 322, 50 },
		{ 405, 323, 50 },
		{ 405, 324, 50 },
		{ 405, 325, 50 },
		{ 405, 326, 50 },
		{ 405, 327, 50 },
		{ 405, 328, 50 },
		{ 405, 329, 50 },
		{ 405, 330, 50 },
		{ 405, 331, 50 },
		{ 405, 332, 50 },
		{ 405, 333, 50 },
		{ 405, 334, 50 },
		{ 405, 335, 50 },
		{ 405, 336, 50 },
		{ 405, 337, 50 },
		{ 405, 338, 50 },
		{ 405, 339, 50 },
		{ 405, 340, 50 },
		{ 405, 341, 50 },
		{ 405, 342, 50 },
		{ 405, 343, 50 },
		{ 405, 344, 50 },
		{ 405, 345, 50 },
		{ 405, 346, 50 },
		{ 405, 347, 50 },
		{ 405, 348, 50 },
		{ 405, 349, 50 },
		{ 406, 313, 50 },
		{ 406, 314, 50 },
		{ 406, 315, 50 },
		{ 406, 316, 50 },
		{ 406, 317, 50 },
		{ 406, 318, 50 },
		{ 406, 319, 50 },
		{ 406, 320, 50 },
		{ 406, 321, 50 },
		{ 406, 322, 50 },
		{ 406, 323, 50 },
		{ 406, 324, 50 },
		{ 406, 325, 50 },
		{ 406, 326, 50 },
		{ 406, 327, 50 },
		{ 406, 328, 50 },
		{ 406, 329, 50 },
		{ 406, 330, 50 },
		{ 406, 331, 50 },
		{ 406, 332, 50 },
		{ 406, 333, 50 },
		{ 406, 334, 50 },
		{ 406, 335, 50 },
		{ 406, 336, 50 },
		{ 406, 337, 50 },
		{ 406, 338, 50 },
		{ 406, 339, 50 },
		{ 406, 340, 50 },
		{ 406, 341, 50 },
		{ 406, 342, 50 },
		{ 406, 343, 50 },
		{ 406, 344, 50 },
		{ 406, 345, 50 },
		{ 406, 346, 50 },
		{ 406, 347, 50 },
		{ 406, 348, 50 },
		{ 406, 349, 50 },
		{ 406, 350, 50 },
		{ 407, 314, 50 },
		{ 407, 315, 50 },
		{ 407, 316, 50 },
		{ 407, 317, 50 },
		{ 407, 318, 50 },
		{ 407, 319, 50 },
		{ 407, 320, 50 },
		{ 407, 321, 50 },
		{ 407, 322, 50 },
		{ 407, 323, 50 },
		{ 407, 324, 50 },
		{ 407, 325, 50 },
		{ 407, 326, 50 },
		{ 407, 327, 50 },
		{ 407, 328, 50 },
		{ 407, 329, 50 },
		{ 407, 330, 50 },
		{ 407, 331, 50 },
		{ 407, 332, 50 },
		{ 407, 333, 50 },
		{ 407, 334, 50 },
		{ 407, 335, 50 },
		{ 407, 336, 50 },
		{ 407, 337, 50 },
		{ 407, 338, 50 },
		{ 407, 339, 50 },
		{ 407, 340, 50 },
		{ 407, 341, 50 },
		{ 407, 342, 50 },
		{ 407, 343, 50 },
		{ 407, 344, 50 },
		{ 407, 345, 50 },
		{ 407, 346, 50 },
		{ 407, 347, 50 },
		{ 407, 348, 50 },
		{ 407, 349, 50 },
		{ 407, 350, 50 },
		{ 408, 314, 50 },
		{ 408, 315, 50 },
		{ 408, 316, 50 },
		{ 408, 317, 50 },
		{ 408, 318, 50 },
		{ 408, 319, 50 },
		{ 408, 320, 50 },
		{ 408, 321, 50 },
		{ 408, 322, 50 },
		{ 408, 323, 50 },
		{ 408, 324, 50 },
		{ 408, 325, 50 },
		{ 408, 326, 50 },
		{ 408, 327, 50 },
		{ 408, 328, 50 },
		{ 408, 329, 50 },
		{ 408, 330, 50 },
		{ 408, 331, 50 },
		{ 408, 332, 50 },
		{ 408, 333, 50 },
		{ 408, 334, 50 },
		{ 408, 335, 50 },
		{ 408, 336, 50 },
		{ 408, 337, 50 },
		{ 408, 338, 50 },
		{ 408, 339, 50 },
		{ 408, 340, 50 },
		{ 408, 341, 50 },
		{ 408, 342, 50 },
		{ 408, 343, 50 },
		{ 408, 344, 50 },
		{ 408, 345, 50 },
		{ 408, 346, 50 },
		{ 408, 347, 50 },
		{ 408, 348, 50 },
		{ 408, 349, 50 },
		{ 408, 350, 50 },
		{ 408, 351, 50 },
		{ 409, 315, 50 },
		{ 409, 316, 50 },
		{ 409, 317, 50 },
		{ 409, 318, 50 },
		{ 409, 319, 50 },
		{ 409, 320, 50 },
		{ 409, 321, 50 },
		{ 409, 322, 50 },
		{ 409, 323, 50 },
		{ 409, 324, 50 },
		{ 409, 325, 50 },
		{ 409, 326, 50 },
		{ 409, 327, 50 },
		{ 409, 328, 50 },
		{ 409, 329, 50 },
		{ 409, 330, 50 },
		{ 409, 331, 50 },
		{ 409, 332, 50 },
		{ 409, 333, 50 },
		{ 409, 334, 50 },
		{ 409, 335, 50 },
		{ 409, 336, 50 },
		{ 409, 337, 50 },
		{ 409, 338, 50 },
		{ 409, 339, 50 },
		{ 409, 340, 50 },
		{ 409, 341, 50 },
		{ 409, 342, 50 },
		{ 409, 343, 50 },
		{ 409, 344, 50 },
		{ 409, 345, 50 },
		{ 409, 346, 50 },
		{ 409, 347, 50 },
		{ 409, 348, 50 },
		{ 409, 349, 50 },
		{ 409, 350, 50 },
		{ 409, 351, 50 },
		{ 410, 315, 50 },
		{ 410, 316, 50 },
		{ 410, 317, 50 },
		{ 410, 318, 50 },
		{ 410, 319, 50 },
		{ 410, 320, 50 },
		{ 410, 321, 50 },
		{ 410, 322, 50 },
		{ 410, 323, 50 },
		{ 410, 324, 50 },
		{ 410, 325, 50 },
		{ 410, 326, 50 },
		{ 410, 327, 50 },
		{ 410, 328, 50 },
		{ 410, 329, 50 },
		{ 410, 330, 50 },
		{ 410, 331, 50 },
		{ 410, 332, 50 },
		{ 410, 333, 50 },
		{ 410, 334, 50 },
		{ 410, 335, 50 },
		{ 410, 336, 50 },
		{ 410, 337, 50 },
		{ 410, 338, 50 },
		{ 410, 339, 50 },
		{ 410, 340, 50 },
		{ 410, 341, 50 },
		{ 410, 342, 50 },
		{ 410, 343, 50 },
		{ 410, 344, 50 },
		{ 410, 345, 50 },
		{ 410, 346, 50 },
		{ 410, 347, 50 },
		{ 410, 348, 50 },
		{ 410, 349, 50 },
		{ 410, 350, 50 },
		{ 410, 351, 50 },
		{ 410, 352, 50 },
		{ 411, 315, 50 },
		{ 411, 316, 50 },
		{ 411, 317, 50 },
		{ 411, 318, 50 },
		{ 411, 319, 50 },
		{ 411, 320, 50 },
		{ 411, 321, 50 },
		{ 411, 322, 50 },
		{ 411, 323, 50 },
		{ 411, 324, 50 },
		{ 411, 325, 50 },
		{ 411, 326, 50 },
		{ 411, 327, 50 },
		{ 411, 328, 50 },
		{ 411, 329, 50 },
		{ 411, 330, 50 },
		{ 411, 331, 50 },
		{ 411, 332, 50 },
		{ 411, 333, 50 },
		{ 411, 334, 50 },
		{ 411, 335, 50 },
		{ 411, 336, 50 },
		{ 411, 337, 50 },
		{ 411, 338, 50 },
		{ 411, 339, 50 },
		{ 411, 340, 50 },
		{ 411, 341, 50 },
		{ 411, 342, 50 },
		{ 411, 343, 50 },
		{ 411, 344, 50 },
		{ 411, 345, 50 },
		{ 411, 346, 50 },
		{ 411, 347, 50 },
		{ 411, 348, 50 },
		{ 411, 349, 50 },
		{ 411, 350, 50 },
		{ 411, 351, 50 },
		{ 411, 352, 50 },
		{ 411, 353, 50 },
		{ 412, 316, 50 },
		{ 412, 317, 50 },
		{ 412, 318, 50 },
		{ 412, 319, 50 },
		{ 412, 320, 50 },
		{ 412, 321, 50 },
		{ 412, 322, 50 },
		{ 412, 323, 50 },
		{ 412, 324, 50 },
		{ 412, 325, 50 },
		{ 412, 326, 50 },
		{ 412, 327, 50 },
		{ 412, 328, 50 },
		{ 412, 329, 50 },
		{ 412, 330, 50 },
		{ 412, 331, 50 },
		{ 412, 332, 50 },
		{ 412, 333, 50 },
		{ 412, 334, 50 },
		{ 412, 335, 50 },
		{ 412, 336, 50 },
		{ 412, 337, 50 },
		{ 412, 338, 50 },
		{ 412, 339, 50 },
		{ 412, 340, 50 },
		{ 412, 341, 50 },
		{ 412, 342, 50 },
		{ 412, 343, 50 },
		{ 412, 344, 50 },
		{ 412, 345, 50 },
		{ 412, 346, 50 },
		{ 412, 347, 50 },
		{ 412, 348, 50 },
		{ 412, 349, 50 },
		{ 412, 350, 50 },
		{ 412, 351, 50 },
		{ 412, 352, 50 },
		{ 412, 353, 50 },
		{ 413, 316, 50 },
		{ 413, 317, 50 },
		{ 413, 318, 50 },
		{ 413, 319, 50 },
		{ 413, 320, 50 },
		{ 413, 321, 50 },
		{ 413, 322, 50 },
		{ 413, 323, 50 },
		{ 413, 324, 50 },
		{ 413, 325, 50 },
		{ 413, 326, 50 },
		{ 413, 327, 50 },
		{ 413, 328, 50 },
		{ 413, 329, 50 },
		{ 413, 330, 50 },
		{ 413, 331, 50 },
		{ 413, 332, 50 },
		{ 413, 333, 50 },
		{ 413, 334, 50 },
		{ 413, 335, 50 },
		{ 413, 336, 50 },
		{ 413, 337, 50 },
		{ 413, 338, 50 },
		{ 413, 339, 50 },
		{ 413, 340, 50 },
		{ 413, 341, 50 },
		{ 413, 342, 50 },
		{ 413, 343, 50 },
		{ 413, 344, 50 },
		{ 413, 345, 50 },
		{ 413, 346, 50 },
		{ 413, 347, 50 },
		{ 413, 348, 50 },
		{ 413, 349, 50 },
		{ 413, 350, 50 },
		{ 413, 351, 50 },
		{ 413, 352, 50 },
		{ 413, 353, 50 },
		{ 413, 354, 50 },
		{ 414, 316, 50 },
		{ 414, 317, 50 },
		{ 414, 318, 50 },
		{ 414, 319, 50 },
		{ 414, 320, 50 },
		{ 414, 321, 50 },
		{ 414, 322, 50 },
		{ 414, 323, 50 },
		{ 414, 324, 50 },
		{ 414, 325, 50 },
		{ 414, 326, 50 },
		{ 414, 327, 50 },
		{ 414, 328, 50 },
		{ 414, 329, 50 },
		{ 414, 330, 50 },
		{ 414, 331, 50 },
		{ 414, 332, 50 },
		{ 414, 333, 50 },
		{ 414, 334, 50 },
		{ 414, 335, 50 },
		{ 414, 336, 50 },
		{ 414, 337, 50 },
		{ 414, 338, 50 },
		{ 414, 339, 50 },
		{ 414, 340, 50 },
		{ 414, 341, 50 },
		{ 414, 342, 50 },
		{ 414, 343, 50 },
		{ 414, 344, 50 },
		{ 414, 345, 50 },
		{ 414, 346, 50 },
		{ 414, 347, 50 },
		{ 414, 348, 50 },
		{ 414, 349, 50 },
		{ 414, 350, 50 },
		{ 414, 351, 50 },
		{ 414, 352, 50 },
		{ 414, 353, 50 },
		{ 414, 354, 50 },
		{ 414, 355, 50 },
		{ 415, 317, 50 },
		{ 415, 318, 50 },
		{ 415, 319, 50 },
		{ 415, 320, 50 },
		{ 415, 321, 50 },
		{ 415, 322, 50 },
		{ 415, 323, 50 },
		{ 415, 324, 50 },
		{ 415, 325, 50 },
		{ 415, 326, 50 },
		{ 415, 327, 50 },
		{ 415, 328, 50 },
		{ 415, 329, 50 },
		{ 415, 330, 50 },
		{ 415, 331, 50 },
		{ 415, 332, 50 },
		{ 415, 333, 50 },
		{ 415, 334, 50 },
		{ 415, 335, 50 },
		{ 415, 336, 50 },
		{ 415, 337, 50 },
		{ 415, 338, 50 },
		{ 415, 339, 50 },
		{ 415, 340, 50 },
		{ 415, 341, 50 },
		{ 415, 342, 50 },
		{ 415, 343, 50 },
		{ 415, 344, 50 },
		{ 415, 345, 50 },
		{ 415, 346, 50 },
		{ 415, 347, 50 },
		{ 415, 348, 50 },
		{ 415, 349, 50 },
		{ 415, 350, 50 },
		{ 415, 351, 50 },
		{ 415, 352, 50 },
		{ 415, 353, 50 },
		{ 415, 354, 50 },
		{ 415, 355, 50 },
		{ 416, 317, 50 },
		{ 416, 318, 50 },
		{ 416, 319, 50 },
		{ 416, 320, 50 },
		{ 416, 321, 50 },
		{ 416, 322, 50 },
		{ 416, 323, 50 },
		{ 416, 324, 50 },
		{ 416, 325, 50 },
		{ 416, 326, 50 },
		{ 416, 327, 50 },
		{ 416, 328, 50 },
		{ 416, 329, 50 },
		{ 416, 330, 50 },
		{ 416, 331, 50 },
		{ 416, 332, 50 },
		{ 416, 333, 50 },
		{ 416, 334, 50 },
		{ 416, 335, 50 },
		{ 416, 336, 50 },
		{ 416, 337, 50 },
		{ 416, 338, 50 },
		{ 416, 339, 50 },
		{ 416, 340, 50 },
		{ 416, 341, 50 },
		{ 416, 342, 50 },
		{ 416, 343, 50 },
		{ 416, 344, 50 },
		{ 416, 345, 50 },
		{ 416, 346, 50 },
		{ 416, 347, 50 },
		{ 416, 348, 50 },
		{ 416, 349, 50 },
		{ 416, 350, 50 },
		{ 416, 351, 50 },
		{ 416, 352, 50 },
		{ 416, 353, 50 },
		{ 416, 354, 50 },
		{ 416, 355, 50 },
		{ 416, 356, 50 },
		{ 417, 318, 50 },
		{ 417, 319, 50 },
		{ 417, 320, 50 },
		{ 417, 321, 50 },
		{ 417, 322, 50 },
		{ 417, 323, 50 },
		{ 417, 324, 50 },
		{ 417, 325, 50 },
		{ 417, 326, 50 },
		{ 417, 327, 50 },
		{ 417, 328, 50 },
		{ 417, 329, 50 },
		{ 417, 330, 50 },
		{ 417, 331, 50 },
		{ 417, 332, 50 },
		{ 417, 333, 50 },
		{ 417, 334, 50 },
		{ 417, 335, 50 },
		{ 417, 336, 50 },
		{ 417, 337, 50 },
		{ 417, 338, 50 },
		{ 417, 339, 50 },
		{ 417, 340, 50 },
		{ 417, 341, 50 },
		{ 417, 342, 50 },
		{ 417, 343, 50 },
		{ 417, 344, 50 },
		{ 417, 345, 50 },
		{ 417, 346, 50 },
		{ 417, 347, 50 },
		{ 417, 348, 50 },
		{ 417, 349, 50 },
		{ 417, 350, 50 },
		{ 417, 351, 50 },
		{ 417, 352, 50 },
		{ 417, 353, 50 },
		{ 417, 354, 50 },
		{ 417, 355, 50 },
		{ 417, 356, 50 },
		{ 418, 318, 50 },
		{ 418, 319, 50 },
		{ 418, 320, 50 },
		{ 418, 321, 50 },
		{ 418, 322, 50 },
		{ 418, 323, 50 },
		{ 418, 324, 50 },
		{ 418, 325, 50 },
		{ 418, 326, 50 },
		{ 418, 327, 50 },
		{ 418, 328, 50 },
		{ 418, 329, 50 },
		{ 418, 330, 50 },
		{ 418, 331, 50 },
		{ 418, 332, 50 },
		{ 418, 333, 50 },
		{ 418, 334, 50 },
		{ 418, 335, 50 },
		{ 418, 336, 50 },
		{ 418, 337, 50 },
		{ 418, 338, 50 },
		{ 418, 339, 50 },
		{ 418, 340, 50 },
		{ 418, 341, 50 },
		{ 418, 342, 50 },
		{ 418, 343, 50 },
		{ 418, 344, 50 },
		{ 418, 345, 50 },
		{ 418, 346, 50 },
		{ 418, 347, 50 },
		{ 418, 348, 50 },
		{ 418, 349, 50 },
		{ 418, 350, 50 },
		{ 418, 351, 50 },
		{ 418, 352, 50 },
		{ 418, 353, 50 },
		{ 418, 354, 50 },
		{ 418, 355, 50 },
		{ 418, 356, 50 },
		{ 418, 357, 50 },
		{ 419, 318, 50 },
		{ 419, 319, 50 },
		{ 419, 320, 50 },
		{ 419, 321, 50 },
		{ 419, 322, 50 },
		{ 419, 323, 50 },
		{ 419, 324, 50 },
		{ 419, 325, 50 },
		{ 419, 326, 50 },
		{ 419, 327, 50 },
		{ 419, 328, 50 },
		{ 419, 329, 50 },
		{ 419, 330, 50 },
		{ 419, 331, 50 },
		{ 419, 332, 50 },
		{ 419, 333, 50 },
		{ 419, 334, 50 },
		{ 419, 335, 50 },
		{ 419, 336, 50 },
		{ 419, 337, 50 },
		{ 419, 338, 50 },
		{ 419, 339, 50 },
		{ 419, 340, 50 },
		{ 419, 341, 50 },
		{ 419, 342, 50 },
		{ 419, 343, 50 },
		{ 419, 344, 50 },
		{ 419, 345, 50 },
		{ 419, 346, 50 },
		{ 419, 347, 50 },
		{ 419, 348, 50 },
		{ 419, 349, 50 },
		{ 419, 350, 50 },
		{ 419, 351, 50 },
		{ 419, 352, 50 },
		{ 419, 353, 50 },
		{ 419, 354, 50 },
		{ 419, 355, 50 },
		{ 419, 356, 50 },
		{ 419, 357, 50 },
		{ 419, 358, 50 },
		{ 420, 319, 50 },
		{ 420, 320, 50 },
		{ 420, 321, 50 },
		{ 420, 322, 50 },
		{ 420, 323, 50 },
		{ 420, 324, 50 },
		{ 420, 325, 50 },
		{ 420, 326, 50 },
		{ 420, 327, 50 },
		{ 420, 328, 50 },
		{ 420, 329, 50 },
		{ 420, 330, 50 },
		{ 420, 331, 50 },
		{ 420, 332, 50 },
		{ 420, 333, 50 },
		{ 420, 334, 50 },
		{ 420, 335, 50 },
		{ 420, 336, 50 },
		{ 420, 337, 50 },
		{ 420, 338, 50 },
		{ 420, 339, 50 },
		{ 420, 340, 50 },
		{ 420, 341, 50 },
		{ 420, 342, 50 },
		{ 420, 343, 50 },
		{ 420, 344, 50 },
		{ 420, 345, 50 },
		{ 420, 346, 50 },
		{ 420, 347, 50 },
		{ 420, 348, 50 },
		{ 420, 349, 50 },
		{ 420, 350, 50 },
		{ 420, 351, 50 },
		{ 420, 352, 50 },
		{ 420, 353, 50 },
		{ 420, 354, 50 },
		{ 420, 355, 50 },
		{ 420, 356, 50 },
		{ 420, 357, 50 },
		{ 420, 358, 50 },
		{ 421, 319, 50 },
		{ 421, 320, 50 },
		{ 421, 321, 50 },
		{ 421, 322, 50 },
		{ 421, 323, 50 },
		{ 421, 324, 50 },
		{ 421, 325, 50 },
		{ 421, 326, 50 },
		{ 421, 327, 50 },
		{ 421, 328, 50 },
		{ 421, 329, 50 },
		{ 421, 330, 50 },
		{ 421, 331, 50 },
		{ 421, 332, 50 },
		{ 421, 333, 50 },
		{ 421, 334, 50 },
		{ 421, 335, 50 },
		{ 421, 336, 50 },
		{ 421, 337, 50 },
		{ 421, 338, 50 },
		{ 421, 339, 50 },
		{ 421, 340, 50 },
		{ 421, 341, 50 },
		{ 421, 342, 50 },
		{ 421, 343, 50 },
		{ 421, 344, 50 },
		{ 421, 345, 50 },
		{ 421, 346, 50 },
		{ 421, 347, 50 },
		{ 421, 348, 50 },
		{ 421, 349, 50 },
		{ 421, 350, 50 },
		{ 421, 351, 50 },
		{ 421, 352, 50 },
		{ 421, 353, 50 },
		{ 421, 354, 50 },
		{ 421, 355, 50 },
		{ 421, 356, 50 },
		{ 421, 357, 50 },
		{ 421, 358, 50 },
		{ 421, 359, 50 },
		{ 422, 319, 50 },
		{ 422, 320, 50 },
		{ 422, 321, 50 },
		{ 422, 322, 50 },
		{ 422, 323, 50 },
		{ 422, 324, 50 },
		{ 422, 325, 50 },
		{ 422, 326, 50 },
		{ 422, 327, 50 },
		{ 422, 328, 50 },
		{ 422, 329, 50 },
		{ 422, 330, 50 },
		{ 422, 331, 50 },
		{ 422, 332, 50 },
		{ 422, 333, 50 },
		{ 422, 334, 50 },
		{ 422, 335, 50 },
		{ 422, 336, 50 },
		{ 422, 337, 50 },
		{ 422, 338, 50 },
		{ 422, 339, 50 },
		{ 422, 340, 50 },
		{ 422, 341, 50 },
		{ 422, 342, 50 },
		{ 422, 343, 50 },
		{ 422, 344, 50 },
		{ 422, 345, 50 },
		{ 422, 346, 50 },
		{ 422, 347, 50 },
		{ 422, 348, 50 },
		{ 422, 349, 50 },
		{ 422, 350, 50 },
		{ 422, 351, 50 },
		{ 422, 352, 50 },
		{ 422, 353, 50 },
		{ 422, 354, 50 },
		{ 422, 355, 50 },
		{ 422, 356, 50 },
		{ 422, 357, 50 },
		{ 422, 358, 50 },
		{ 422, 359, 50 },
		{ 422, 360, 50 },
		{ 423, 320, 50 },
		{ 423, 321, 50 },
		{ 423, 322, 50 },
		{ 423, 323, 50 },
		{ 423, 324, 50 },
		{ 423, 325, 50 },
		{ 423, 326, 50 },
		{ 423, 327, 50 },
		{ 423, 328, 50 },
		{ 423, 329, 50 },
		{ 423, 330, 50 },
		{ 423, 331, 50 },
		{ 423, 332, 50 },
		{ 423, 333, 50 },
		{ 423, 334, 50 },
		{ 423, 335, 50 },
		{ 423, 336, 50 },
		{ 423, 337, 50 },
		{ 423, 338, 50 },
		{ 423, 339, 50 },
		{ 423, 340, 50 },
		{ 423, 341, 50 },
		{ 423, 342, 50 },
		{ 423, 343, 50 },
		{ 423, 344, 50 },
		{ 423, 345, 50 },
		{ 423, 346, 50 },
		{ 423, 347, 50 },
		{ 423, 348, 50 },
		{ 423, 349, 50 },
		{ 423, 350, 50 },
		{ 423, 351, 50 },
		{ 423, 352, 50 },
		{ 423, 353, 50 },
		{ 423, 354, 50 },
		{ 423, 355, 50 },
		{ 423, 356, 50 },
		{ 423, 357, 50 },
		{ 423, 358, 50 },
		{ 423, 359, 50 },
		{ 423, 360, 50 },
		{ 424, 320, 50 },
		{ 424, 321, 50 },
		{ 424, 322, 50 },
		{ 424, 323, 50 },
		{ 424, 324, 50 },
		{ 424, 325, 50 },
		{ 424, 326, 50 },
		{ 424, 327, 50 },
		{ 424, 328, 50 },
		{ 424, 329, 50 },
		{ 424, 330, 50 },
		{ 424, 331, 50 },
		{ 424, 332, 50 },
		{ 424, 333, 50 },
		{ 424, 334, 50 },
		{ 424, 335, 50 },
		{ 424, 336, 50 },
		{ 424, 337, 50 },
		{ 424, 338, 50 },
		{ 424, 339, 50 },
		{ 424, 340, 50 },
		{ 424, 341, 50 },
		{ 424, 342, 50 },
		{ 424, 343, 50 },
		{ 424, 344, 50 },
		{ 424, 345, 50 },
		{ 424, 346, 50 },
		{ 424, 347, 50 },
		{ 424, 348, 50 },
		{ 424, 349, 50 },
		{ 424, 350, 50 },
		{ 424, 351, 50 },
		{ 424, 352, 50 },
		{ 424, 353, 50 },
		{ 424, 354, 50 },
		{ 424, 355, 50 },
		{ 424, 356, 50 },
		{ 424, 357, 50 },
		{ 424, 358, 50 },
		{ 424, 359, 50 },
		{ 424, 360, 50 },
		{ 424, 361, 50 },
		{ 425, 321, 50 },
		{ 425, 322, 50 },
		{ 425, 323, 50 },
		{ 425, 324, 50 },
		{ 425, 325, 50 },
		{ 425, 326, 50 },
		{ 425, 327, 50 },
		{ 425, 328, 50 },
		{ 425, 329, 50 },
		{ 425, 330, 50 },
		{ 425, 331, 50 },
		{ 425, 332, 50 },
		{ 425, 333, 50 },
		{ 425, 334, 50 },
		{ 425, 335, 50 },
		{ 425, 336, 50 },
		{ 425, 337, 50 },
		{ 425, 338, 50 },
		{ 425, 339, 50 },
		{ 425, 340, 50 },
		{ 425, 341, 50 },
		{ 425, 342, 50 },
		{ 425, 343, 50 },
		{ 425, 344, 50 },
		{ 425, 345, 50 },
		{ 425, 346, 50 },
		{ 425, 347, 50 },
		{ 425, 348, 50 },
		{ 425, 349, 50 },
		{ 425, 350, 50 },
		{ 425, 351, 50 },
		{ 425, 352, 50 },
		{ 425, 353, 50 },
		{ 425, 354, 50 },
		{ 425, 355, 50 },
		{ 425, 356, 50 },
		{ 425, 357, 50 },
		{ 425, 358, 50 },
		{ 425, 359, 50 },
		{ 425, 360, 50 },
		{ 425, 361, 50 },
		{ 426, 321, 50 },
		{ 426, 322, 50 },
		{ 426, 323, 50 },
		{ 426, 324, 50 },
		{ 426, 325, 50 },
		{ 426, 326, 50 },
		{ 426, 327, 50 },
		{ 426, 328, 50 },
		{ 426, 329, 50 },
		{ 426, 330, 50 },
		{ 426, 331, 50 },
		{ 426, 332, 50 },
		{ 426, 333, 50 },
		{ 426, 334, 50 },
		{ 426, 335, 50 },
		{ 426, 336, 50 },
		{ 426, 337, 50 },
		{ 426, 338, 50 },
		{ 426, 339, 50 },
		{ 426, 340, 50 },
		{ 426, 341, 50 },
		{ 426, 342, 50 },
		{ 426, 343, 50 },
		{ 426, 344, 50 },
		{ 426, 345, 50 },
		{ 426, 346, 50 },
		{ 426, 347, 50 },
		{ 426, 348, 50 },
		{ 426, 349, 50 },
		{ 426, 350, 50 },
		{ 426, 351, 50 },
		{ 426, 352, 50 },
		{ 426, 353, 50 },
		{ 426, 354, 50 },
		{ 426, 355, 50 },
		{ 426, 356, 50 },
		{ 426, 357, 50 },
		{ 426, 358, 50 },
		{ 426, 359, 50 },
		{ 426, 360, 50 },
		{ 426, 361, 50 },
		{ 426, 362, 50 },
		{ 427, 321, 50 },
		{ 427, 322, 50 },
		{ 427, 323, 50 },
		{ 427, 324, 50 },
		{ 427, 325, 50 },
		{ 427, 326, 50 },
		{ 427, 327, 50 },
		{ 427, 328, 50 },
		{ 427, 329, 50 },
		{ 427, 330, 50 },
		{ 427, 331, 50 },
		{ 427, 332, 50 },
		{ 427, 333, 50 },
		{ 427, 334, 50 },
		{ 427, 335, 50 },
		{ 427, 336, 50 },
		{ 427, 337, 50 },
		{ 427, 338, 50 },
		{ 427, 339, 50 },
		{ 427, 340, 50 },
		{ 427, 341, 50 },
		{ 427, 342, 50 },
		{ 427, 343, 50 },
		{ 427, 344, 50 },
		{ 427, 345, 50 },
		{ 427, 346, 50 },
		{ 427, 347, 50 },
		{ 427, 348, 50 },
		{ 427, 349, 50 },
		{ 427, 350, 50 },
		{ 427, 351, 50 },
		{ 427, 352, 50 },
		{ 427, 353, 50 },
		{ 427, 354, 50 },
		{ 427, 355, 50 },
		{ 427, 356, 50 },
		{ 427, 357, 50 },
		{ 427, 358, 50 },
		{ 427, 359, 50 },
		{ 427, 360, 50 },
		{ 427, 361, 50 },
		{ 427, 362, 50 },
		{ 427, 363, 50 },
		{ 428, 322, 50 },
		{ 428, 323, 50 },
		{ 428, 324, 50 },
		{ 428, 325, 50 },
		{ 428, 326, 50 },
		{ 428, 327, 50 },
		{ 428, 328, 50 },
		{ 428, 329, 50 },
		{ 428, 330, 50 },
		{ 428, 331, 50 },
		{ 428, 332, 50 },
		{ 428, 333, 50 },
		{ 428, 334, 50 },
		{ 428, 335, 50 },
		{ 428, 336, 50 },
		{ 428, 337, 50 },
		{ 428, 338, 50 },
		{ 428, 339, 50 },
		{ 428, 340, 50 },
		{ 428, 341, 50 },
		{ 428, 342, 50 },
		{ 428, 343, 50 },
		{ 428, 344, 50 },
		{ 428, 345, 50 },
		{ 428, 346, 50 },
		{ 428, 347, 50 },
		{ 428, 348, 50 },
		{ 428, 349, 50 },
		{ 428, 350, 50 },
		{ 428, 351, 50 },
		{ 428, 352, 50 },
		{ 428, 353, 50 },
		{ 428, 354, 50 },
		{ 428, 355, 50 },
		{ 428, 356, 50 },
		{ 428, 357, 50 },
		{ 428, 358, 50 },
		{ 428, 359, 50 },
		{ 428, 360, 50 },
		{ 428, 361, 50 },
		{ 428, 362, 50 },
		{ 428, 363, 50 },
		{ 429, 322, 50 },
		{ 429, 323, 50 },
		{ 429, 324, 50 },
		{ 429, 325, 50 },
		{ 429, 326, 50 },
		{ 429, 327, 50 },
		{ 429, 328, 50 },
		{ 429, 329, 50 },
		{ 429, 330, 50 },
		{ 429, 331, 50 },
		{ 429, 332, 50 },
		{ 429, 333, 50 },
		{ 429, 334, 50 },
		{ 429, 335, 50 },
		{ 429, 336, 50 },
		{ 429, 337, 50 },
		{ 429, 338, 50 },
		{ 429, 339, 50 },
		{ 429, 340, 50 },
		{ 429, 341, 50 },
		{ 429, 342, 50 },
		{ 429, 343, 50 },
		{ 429, 344, 50 },
		{ 429, 345, 50 },
		{ 429, 346, 50 },
		{ 429, 347, 50 },
		{ 429, 348, 50 },
		{ 429, 349, 50 },
		{ 429, 350, 50 },
		{ 429, 351, 50 },
		{ 429, 352, 50 },
		{ 429, 353, 50 },
		{ 429, 354, 50 },
		{ 429, 355, 50 },
		{ 429, 356, 50 },
		{ 429, 357, 50 },
		{ 429, 358, 50 },
		{ 429, 359, 50 },
		{ 429, 360, 50 },
		{ 429, 361, 50 },
		{ 429, 362, 50 },
		{ 429, 363, 50 },
		{ 429, 364, 50 },
		{ 430, 322, 50 },
		{ 430, 323, 50 },
		{ 430, 324, 50 },
		{ 430, 325, 50 },
		{ 430, 326, 50 },
		{ 430, 327, 50 },
		{ 430, 328, 50 },
		{ 430, 329, 50 },
		{ 430, 330, 50 },
		{ 430, 331, 50 },
		{ 430, 332, 50 },
		{ 430, 333, 50 },
		{ 430, 334, 50 },
		{ 430, 335, 50 },
		{ 430, 336, 50 },
		{ 430, 337, 50 },
		{ 430, 338, 50 },
		{ 430, 339, 50 },
		{ 430, 340, 50 },
		{ 430, 341, 50 },
		{ 430, 342, 50 },
		{ 430, 343, 50 },
		{ 430, 344, 50 },
		{ 430, 345, 50 },
		{ 430, 346, 50 },
		{ 430, 347, 50 },
		{ 430, 348, 50 },
		{ 430, 349, 50 },
		{ 430, 350, 50 },
		{ 430, 351, 50 },
		{ 430, 352, 50 },
		{ 430, 353, 50 },
		{ 430, 354, 50 },
		{ 430, 355, 50 },
		{ 430, 356, 50 },
		{ 430, 357, 50 },
		{ 430, 358, 50 },
		{ 430, 359, 50 },
		{ 430, 360, 50 },
		{ 430, 361, 50 },
		{ 430, 362, 50 },
		{ 430, 363, 50 },
		{ 430, 364, 50 },
		{ 430, 365, 50 },
		{ 431, 323, 50 },
		{ 431, 324, 50 },
		{ 431, 325, 50 },
		{ 431, 326, 50 },
		{ 431, 327, 50 },
		{ 431, 328, 50 },
		{ 431, 329, 50 },
		{ 431, 330, 50 },
		{ 431, 331, 50 },
		{ 431, 332, 50 },
		{ 431, 333, 50 },
		{ 431, 334, 50 },
		{ 431, 335, 50 },
		{ 431, 336, 50 },
		{ 431, 337, 50 },
		{ 431, 338, 50 },
		{ 431, 339, 50 },
		{ 431, 340, 50 },
		{ 431, 341, 50 },
		{ 431, 342, 50 },
		{ 431, 343, 50 },
		{ 431, 344, 50 },
		{ 431, 345, 50 },
		{ 431, 346, 50 },
		{ 431, 347, 50 },
		{ 431, 348, 50 },
		{ 431, 349, 50 },
		{ 431, 350, 50 },
		{ 431, 351, 50 },
		{ 431, 352, 50 },
		{ 431, 353, 50 },
		{ 431, 354, 50 },
		{ 431, 355, 50 },
		{ 431, 356, 50 },
		{ 431, 357, 50 },
		{ 431, 358, 50 },
		{ 431, 359, 50 },
		{ 431, 360, 50 },
		{ 431, 361, 50 },
		{ 431, 362, 50 },
		{ 431, 363, 50 },
		{ 431, 364, 50 },
		{ 431, 365, 50 },
		{ 432, 323, 50 },
		{ 432, 324, 50 },
		{ 432, 325, 50 },
		{ 432, 326, 50 },
		{ 432, 327, 50 },
		{ 432, 328, 50 },
		{ 432, 329, 50 },
		{ 432, 330, 50 },
		{ 432, 331, 50 },
		{ 432, 332, 50 },
		{ 432, 333, 50 },
		{ 432, 334, 50 },
		{ 432, 335, 50 },
		{ 432, 336, 50 },
		{ 432, 337, 50 },
		{ 432, 338, 50 },
		{ 432, 339, 50 },
		{ 432, 340, 50 },
		{ 432, 341, 50 },
		{ 432, 342, 50 },
		{ 432, 343, 50 },
		{ 432, 344, 50 },
		{ 432, 345, 50 },
		{ 432, 346, 50 },
		{ 432, 347, 50 },
		{ 432, 348, 50 },
		{ 432, 349, 50 },
		{ 432, 350, 50 },
		{ 432, 351, 50 },
		{ 432, 352, 50 },
		{ 432, 353, 50 },
		{ 432, 354, 50 },
		{ 432, 355, 50 },
		{ 432, 356, 50 },
		{ 432, 357, 50 },
		{ 432, 358, 50 },
		{ 432, 359, 50 },
		{ 432, 360, 50 },
		{ 432, 361, 50 },
		{ 432, 362, 50 },
		{ 432, 363, 50 },
		{ 432, 364, 50 },
		{ 432, 365, 50 },
		{ 432, 366, 50 },
		{ 433, 324, 50 },
		{ 433, 325, 50 },
		{ 433, 326, 50 },
		{ 433, 327, 50 },
		{ 433, 328, 50 },
		{ 433, 329, 50 },
		{ 433, 330, 50 },
		{ 433, 331, 50 },
		{ 433, 332, 50 },
		{ 433, 333, 50 },
		{ 433, 334, 50 },
		{ 433, 335, 50 },
		{ 433, 336, 50 },
		{ 433, 337, 50 },
		{ 433, 338, 50 },
		{ 433, 339, 50 },
		{ 433, 340, 50 },
		{ 433, 341, 50 },
		{ 433, 342, 50 },
		{ 433, 343, 50 },
		{ 433, 344, 50 },
		{ 433, 345, 50 },
		{ 433, 346, 50 },
		{ 433, 347, 50 },
		{ 433, 348, 50 },
		{ 433, 349, 50 },
		{ 433, 350, 50 },
		{ 433, 351, 50 },
		{ 433, 352, 50 },
		{ 433, 353, 50 },
		{ 433, 354, 50 },
		{ 433, 355, 50 },
		{ 433, 356, 50 },
		{ 433, 357, 50 },
		{ 433, 358, 50 },
		{ 433, 359, 50 },
		{ 433, 360, 50 },
		{ 433, 361, 50 },
		{ 433, 362, 50 },
		{ 433, 363, 50 },
		{ 433, 364, 50 },
		{ 433, 365, 50 },
		{ 433, 366, 50 },
		{ 434, 324, 50 },
		{ 434, 325, 50 },
		{ 434, 326, 50 },
		{ 434, 327, 50 },
		{ 434, 328, 50 },
		{ 434, 329, 50 },
		{ 434, 330, 50 },
		{ 434, 331, 50 },
		{ 434, 332, 50 },
		{ 434, 333, 50 },
		{ 434, 334, 50 },
		{ 434, 335, 50 },
		{ 434, 336, 50 },
		{ 434, 337, 50 },
		{ 434, 338, 50 },
		{ 434, 339, 50 },
		{ 434, 340, 50 },
		{ 434, 341, 50 },
		{ 434, 342, 50 },
		{ 434, 343, 50 },
		{ 434, 344, 50 },
		{ 434, 345, 50 },
		{ 434, 346, 50 },
		{ 434, 347, 50 },
		{ 434, 348, 50 },
		{ 434, 349, 50 },
		{ 434, 350, 50 },
		{ 434, 351, 50 },
		{ 434, 352, 50 },
		{ 434, 353, 50 },
		{ 434, 354, 50 },
		{ 434, 355, 50 },
		{ 434, 356, 50 },
		{ 434, 357, 50 },
		{ 434, 358, 50 },
		{ 434, 359, 50 },
		{ 434, 360, 50 },
		{ 434, 361, 50 },
		{ 434, 362, 50 },
		{ 434, 363, 50 },
		{ 434, 364, 50 },
		{ 434, 365, 50 },
		{ 434, 366, 50 },
		{ 434, 367, 50 },
		{ 435, 324, 50 },
		{ 435, 325, 50 },
		{ 435, 326, 50 },
		{ 435, 327, 50 },
		{ 435, 328, 50 },
		{ 435, 329, 50 },
		{ 435, 330, 50 },
		{ 435, 331, 50 },
		{ 435, 332, 50 },
		{ 435, 333, 50 },
		{ 435, 334, 50 },
		{ 435, 335, 50 },
		{ 435, 336, 50 },
		{ 435, 337, 50 },
		{ 435, 338, 50 },
		{ 435, 339, 50 },
		{ 435, 340, 50 },
		{ 435, 341, 50 },
		{ 435, 342, 50 },
		{ 435, 343, 50 },
		{ 435, 344, 50 },
		{ 435, 345, 50 },
		{ 435, 346, 50 },
		{ 435, 347, 50 },
		{ 435, 348, 50 },
		{ 435, 349, 50 },
		{ 435, 350, 50 },
		{ 435, 351, 50 },
		{ 435, 352, 50 },
		{ 435, 353, 50 },
		{ 435, 354, 50 },
		{ 435, 355, 50 },
		{ 435, 356, 50 },
		{ 435, 357, 50 },
		{ 435, 358, 50 },
		{ 435, 359, 50 },
		{ 435, 360, 50 },
		{ 435, 361, 50 },
		{ 435, 362, 50 },
		{ 435, 363, 50 },
		{ 435, 364, 50 },
		{ 435, 365, 50 },
		{ 435, 366, 50 },
		{ 435, 367, 50 },
		{ 435, 368, 50 },
		{ 436, 325, 50 },
		{ 436, 326, 50 },
		{ 436, 327, 50 },
		{ 436, 328, 50 },
		{ 436, 329, 50 },
		{ 436, 330, 50 },
		{ 436, 331, 50 },
		{ 436, 332, 50 },
		{ 436, 333, 50 },
		{ 436, 334, 50 },
		{ 436, 335, 50 },
		{ 436, 336, 50 },
		{ 436, 337, 50 },
		{ 436, 338, 50 },
		{ 436, 339, 50 },
		{ 436, 340, 50 },
		{ 436, 341, 50 },
		{ 436, 342, 50 },
		{ 436, 343, 50 },
		{ 436, 344, 50 },
		{ 436, 345, 50 },
		{ 436, 346, 50 },
		{ 436, 347, 50 },
		{ 436, 348, 50 },
		{ 436, 349, 50 },
		{ 436, 350, 50 },
		{ 436, 351, 50 },
		{ 436, 352, 50 },
		{ 436, 353, 50 },
		{ 436, 354, 50 },
		{ 436, 355, 50 },
		{ 436, 356, 50 },
		{ 436, 357, 50 },
		{ 436, 358, 50 },
		{ 436, 359, 50 },
		{ 436, 360, 50 },
		{ 436, 361, 50 },
		{ 436, 362, 50 },
		{ 436, 363, 50 },
		{ 436, 364, 50 },
		{ 436, 365, 50 },
		{ 436, 366, 50 },
		{ 436, 367, 50 },
		{ 436, 368, 50 },
		{ 437, 325, 50 },
		{ 437, 326, 50 },
		{ 437, 327, 50 },
		{ 437, 328, 50 },
		{ 437, 329, 50 },
		{ 437, 330, 50 },
		{ 437, 331, 50 },
		{ 437, 332, 50 },
		{ 437, 333, 50 },
		{ 437, 334, 50 },
		{ 437, 335, 50 },
		{ 437, 336, 50 },
		{ 437, 337, 50 },
		{ 437, 338, 50 },
		{ 437, 339, 50 },
		{ 437, 340, 50 },
		{ 437, 341, 50 },
		{ 437, 342, 50 },
		{ 437, 343, 50 },
		{ 437, 344, 50 },
		{ 437, 345, 50 },
		{ 437, 346, 50 },
		{ 437, 347, 50 },
		{ 437, 348, 50 },
		{ 437, 349, 50 },
		{ 437, 350, 50 },
		{ 437, 351, 50 },
		{ 437, 352, 50 },
		{ 437, 353, 50 },
		{ 437, 354, 50 },
		{ 437, 355, 50 },
		{ 437, 356, 50 },
		{ 437, 357, 50 },
		{ 437, 358, 50 },
		{ 437, 359, 50 },
		{ 437, 360, 50 },
		{ 437, 361, 50 },
		{ 437, 362, 50 },
		{ 437, 363, 50 },
		{ 437, 364, 50 },
		{ 437, 365, 50 },
		{ 437, 366, 50 },
		{ 437, 367, 50 },
		{ 437, 368, 50 },
		{ 437, 369, 50 },
		{ 438, 325, 50 },
		{ 438, 326, 50 },
		{ 438, 327, 50 },
		{ 438, 328, 50 },
		{ 438, 329, 50 },
		{ 438, 330, 50 },
		{ 438, 331, 50 },
		{ 438, 332, 50 },
		{ 438, 333, 50 },
		{ 438, 334, 50 },
		{ 438, 335, 50 },
		{ 438, 336, 50 },
		{ 438, 337, 50 },
		{ 438, 338, 50 },
		{ 438, 339, 50 },
		{ 438, 340, 50 },
		{ 438, 341, 50 },
		{ 438, 342, 50 },
		{ 438, 343, 50 },
		{ 438, 344, 50 },
		{ 438, 345, 50 },
		{ 438, 346, 50 },
		{ 438, 347, 50 },
		{ 438, 348, 50 },
		{ 438, 349, 50 },
		{ 438, 350, 50 },
		{ 438, 351, 50 },
		{ 438, 352, 50 },
		{ 438, 353, 50 },
		{ 438, 354, 50 },
		{ 438, 355, 50 },
		{ 438, 356, 50 },
		{ 438, 357, 50 },
		{ 438, 358, 50 },
		{ 438, 359, 50 },
		{ 438, 360, 50 },
		{ 438, 361, 50 },
		{ 438, 362, 50 },
		{ 438, 363, 50 },
		{ 438, 364, 50 },
		{ 438, 365, 50 },
		{ 438, 366, 50 },
		{ 438, 367, 50 },
		{ 438, 368, 50 },
		{ 438, 369, 50 },
		{ 438, 370, 50 },
		{ 439, 326, 50 },
		{ 439, 327, 50 },
		{ 439, 328, 50 },
		{ 439, 329, 50 },
		{ 439, 330, 50 },
		{ 439, 331, 50 },
		{ 439, 332, 50 },
		{ 439, 333, 50 },
		{ 439, 334, 50 },
		{ 439, 335, 50 },
		{ 439, 336, 50 },
		{ 439, 337, 50 },
		{ 439, 338, 50 },
		{ 439, 339, 50 },
		{ 439, 340, 50 },
		{ 439, 341, 50 },
		{ 439, 342, 50 },
		{ 439, 343, 50 },
		{ 439, 344, 50 },
		{ 439, 345, 50 },
		{ 439, 346, 50 },
		{ 439, 347, 50 },
		{ 439, 348, 50 },
		{ 439, 349, 50 },
		{ 439, 350, 50 },
		{ 439, 351, 50 },
		{ 439, 352, 50 },
		{ 439, 353, 50 },
		{ 439, 354, 50 },
		{ 439, 355, 50 },
		{ 439, 356, 50 },
		{ 439, 357, 50 },
		{ 439, 358, 50 },
		{ 439, 359, 50 },
		{ 439, 360, 50 },
		{ 439, 361, 50 },
		{ 439, 362, 50 },
		{ 439, 363, 50 },
		{ 439, 364, 50 },
		{ 439, 365, 50 },
		{ 439, 366, 50 },
		{ 439, 367, 50 },
		{ 439, 368, 50 },
		{ 439, 369, 50 },
		{ 439, 370, 50 },
		{ 440, 326, 50 },
		{ 440, 327, 50 },
		{ 440, 328, 50 },
		{ 440, 329, 50 },
		{ 440, 330, 50 },
		{ 440, 331, 50 },
		{ 440, 332, 50 },
		{ 440, 333, 50 },
		{ 440, 334, 50 },
		{ 440, 335, 50 },
		{ 440, 336, 50 },
		{ 440, 337, 50 },
		{ 440, 338, 50 },
		{ 440, 339, 50 },
		{ 440, 340, 50 },
		{ 440, 341, 50 },
		{ 440, 342, 50 },
		{ 440, 343, 50 },
		{ 440, 344, 50 },
		{ 440, 345, 50 },
		{ 440, 346, 50 },
		{ 440, 347, 50 },
		{ 440, 348, 50 },
		{ 440, 349, 50 },
		{ 440, 350, 50 },
		{ 440, 351, 50 },
		{ 440, 352, 50 },
		{ 440, 353, 50 },
		{ 440, 354, 50 },
		{ 440, 355, 50 },
		{ 440, 356, 50 },
		{ 440, 357, 50 },
		{ 440, 358, 50 },
		{ 440, 359, 50 },
		{ 440, 360, 50 },
		{ 440, 361, 50 },
		{ 440, 362, 50 },
		{ 440, 363, 50 },
		{ 440, 364, 50 },
		{ 440, 365, 50 },
		{ 440, 366, 50 },
		{ 440, 367, 50 },
		{ 440, 368, 50 },
		{ 440, 369, 50 },
		{ 440, 370, 50 },
		{ 440, 371, 50 },
		{ 441, 327, 50 },
		{ 441, 328, 50 },
		{ 441, 329, 50 },
		{ 441, 330, 50 },
		{ 441, 331, 50 },
		{ 441, 332, 50 },
		{ 441, 333, 50 },
		{ 441, 334, 50 },
		{ 441, 335, 50 },
		{ 441, 336, 50 },
		{ 441, 337, 50 },
		{ 441, 338, 50 },
		{ 441, 339, 50 },
		{ 441, 340, 50 },
		{ 441, 341, 50 },
		{ 441, 342, 50 },
		{ 441, 343, 50 },
		{ 441, 344, 50 },
		{ 441, 345, 50 },
		{ 441, 346, 50 },
		{ 441, 347, 50 },
		{ 441, 348, 50 },
		{ 441, 349, 50 },
		{ 441, 350, 50 },
		{ 441, 351, 50 },
		{ 441, 352, 50 },
		{ 441, 353, 50 },
		{ 441, 354, 50 },
		{ 441, 355, 50 },
		{ 441, 356, 50 },
		{ 441, 357, 50 },
		{ 441, 358, 50 },
		{ 441, 359, 50 },
		{ 441, 360, 50 },
		{ 441, 361, 50 },
		{ 441, 362, 50 },
		{ 441, 363, 50 },
		{ 441, 364, 50 },
		{ 441, 365, 50 },
		{ 441, 366, 50 },
		{ 441, 367, 50 },
		{ 441, 368, 50 },
		{ 441, 369, 50 },
		{ 441, 370, 50 },
		{ 441, 371, 50 },
		{ 442, 327, 50 },
		{ 442, 328, 50 },
		{ 442, 329, 50 },
		{ 442, 330, 50 },
		{ 442, 331, 50 },
		{ 442, 332, 50 },
		{ 442, 333, 50 },
		{ 442, 334, 50 },
		{ 442, 335, 50 },
		{ 442, 336, 50 },
		{ 442, 337, 50 },
		{ 442, 338, 50 },
		{ 442, 339, 50 },
		{ 442, 340, 50 },
		{ 442, 341, 50 },
		{ 442, 342, 50 },
		{ 442, 343, 50 },
		{ 442, 344, 50 },
		{ 442, 345, 50 },
		{ 442, 346, 50 },
		{ 442, 347, 50 },
		{ 442, 348, 50 },
		{ 442, 349, 50 },
		{ 442, 350, 50 },
		{ 442, 351, 50 },
		{ 442, 352, 50 },
		{ 442, 353, 50 },
		{ 442, 354, 50 },
		{ 442, 355, 50 },
		{ 442, 356, 50 },
		{ 442, 357, 50 },
		{ 442, 358, 50 },
		{ 442, 359, 50 },
		{ 442, 360, 50 },
		{ 442, 361, 50 },
		{ 442, 362, 50 },
		{ 442, 363, 50 },
		{ 442, 364, 50 },
		{ 442, 365, 50 },
		{ 442, 366, 50 },
		{ 442, 367, 50 },
		{ 442, 368, 50 },
		{ 442, 369, 50 },
		{ 442, 370, 50 },
		{ 442, 371, 50 },
		{ 442, 372, 50 },
		{ 443, 327, 50 },
		{ 443, 328, 50 },
		{ 443, 329, 50 },
		{ 443, 330, 50 },
		{ 443, 331, 50 },
		{ 443, 332, 50 },
		{ 443, 333, 50 },
		{ 443, 334, 50 },
		{ 443, 335, 50 },
		{ 443, 336, 50 },
		{ 443, 337, 50 },
		{ 443, 338, 50 },
		{ 443, 339, 50 },
		{ 443, 340, 50 },
		{ 443, 341, 50 },
		{ 443, 342, 50 },
		{ 443, 343, 50 },
		{ 443, 344, 50 },
		{ 443, 345, 50 },
		{ 443, 346, 50 },
		{ 443, 347, 50 },
		{ 443, 348, 50 },
		{ 443, 349, 50 },
		{ 443, 350, 50 },
		{ 443, 351, 50 },
		{ 443, 352, 50 },
		{ 443, 353, 50 },
		{ 443, 354, 50 },
		{ 443, 355, 50 },
		{ 443, 356, 50 },
		{ 443, 357, 50 },
		{ 443, 358, 50 },
		{ 443, 359, 50 },
		{ 443, 360, 50 },
		{ 443, 361, 50 },
		{ 443, 362, 50 },
		{ 443, 363, 50 },
		{ 443, 364, 50 },
		{ 443, 365, 50 },
		{ 443, 366, 50 },
		{ 443, 367, 50 },
		{ 443, 368, 50 },
		{ 443, 369, 50 },
		{ 443, 370, 50 },
		{ 443, 371, 50 },
		{ 443, 372, 50 },
		{ 443, 373, 50 },
		{ 444, 328, 50 },
		{ 444, 329, 50 },
		{ 444, 330, 50 },
		{ 444, 331, 50 },
		{ 444, 332, 50 },
		{ 444, 333, 50 },
		{ 444, 334, 50 },
		{ 444, 335, 50 },
		{ 444, 336, 50 },
		{ 444, 337, 50 },
		{ 444, 338, 50 },
		{ 444, 339, 50 },
		{ 444, 340, 50 },
		{ 444, 341, 50 },
		{ 444, 342, 50 },
		{ 444, 343, 50 },
		{ 444, 344, 50 },
		{ 444, 345, 50 },
		{ 444, 346, 50 },
		{ 444, 347, 50 },
		{ 444, 348, 50 },
		{ 444, 349, 50 },
		{ 444, 350, 50 },
		{ 444, 351, 50 },
		{ 444, 352, 50 },
		{ 444, 353, 50 },
		{ 444, 354, 50 },
		{ 444, 355, 50 },
		{ 444, 356, 50 },
		{ 444, 357, 50 },
		{ 444, 358, 50 },
		{ 444, 359, 50 },
		{ 444, 360, 50 },
		{ 444, 361, 50 },
		{ 444, 362, 50 },
		{ 444, 363, 50 },
		{ 444, 364, 50 },
		{ 444, 365, 50 },
		{ 444, 366, 50 },
		{ 444, 367, 50 },
		{ 444, 368, 50 },
		{ 444, 369, 50 },
		{ 444, 370, 50 },
		{ 444, 371, 50 },
		{ 444, 372, 50 },
		{ 444, 373, 50 },
		{ 445, 328, 50 },
		{ 445, 329, 50 },
		{ 445, 330, 50 },
		{ 445, 331, 50 },
		{ 445, 332, 50 },
		{ 445, 333, 50 },
		{ 445, 334, 50 },
		{ 445, 335, 50 },
		{ 445, 336, 50 },
		{ 445, 337, 50 },
		{ 445, 338, 50 },
		{ 445, 339, 50 },
		{ 445, 340, 50 },
		{ 445, 341, 50 },
		{ 445, 342, 50 },
		{ 445, 343, 50 },
		{ 445, 344, 50 },
		{ 445, 345, 50 },
		{ 445, 346, 50 },
		{ 445, 347, 50 },
		{ 445, 348, 50 },
		{ 445, 349, 50 },
		{ 445, 350, 50 },
		{ 445, 351, 50 },
		{ 445, 352, 50 },
		{ 445, 353, 50 },
		{ 445, 354, 50 },
		{ 445, 355, 50 },
		{ 445, 356, 50 },
		{ 445, 357, 50 },
		{ 445, 358, 50 },
		{ 445, 359, 50 },
		{ 445, 360, 50 },
		{ 445, 361, 50 },
		{ 445, 362, 50 },
		{ 445, 363, 50 },
		{ 445, 364, 50 },
		{ 445, 365, 50 },
		{ 445, 366, 50 },
		{ 445, 367, 50 },
		{ 445, 368, 50 },
		{ 445, 369, 50 },
		{ 445, 370, 50 },
		{ 445, 371, 50 },
		{ 445, 372, 50 },
		{ 445, 373, 50 },
		{ 445, 374, 50 },
		{ 446, 328, 50 },
		{ 446, 329, 50 },
		{ 446, 330, 50 },
		{ 446, 331, 50 },
		{ 446, 332, 50 },
		{ 446, 333, 50 },
		{ 446, 334, 50 },
		{ 446, 335, 50 },
		{ 446, 336, 50 },
		{ 446, 337, 50 },
		{ 446, 338, 50 },
		{ 446, 339, 50 },
		{ 446, 340, 50 },
		{ 446, 341, 50 },
		{ 446, 342, 50 },
		{ 446, 343, 50 },
		{ 446, 344, 50 },
		{ 446, 345, 50 },
		{ 446, 346, 50 },
		{ 446, 347, 50 },
		{ 446, 348, 50 },
		{ 446, 349, 50 },
		{ 446, 350, 50 },
		{ 446, 351, 50 },
		{ 446, 352, 50 },
		{ 446, 353, 50 },
		{ 446, 354, 50 },
		{ 446, 355, 50 },
		{ 446, 356, 50 },
		{ 446, 357, 50 },
		{ 446, 358, 50 },
		{ 446, 359, 50 },
		{ 446, 360, 50 },
		{ 446, 361, 50 },
		{ 446, 362, 50 },
		{ 446, 363, 50 },
		{ 446, 364, 50 },
		{ 446, 365, 50 },
		{ 446, 366, 50 },
		{ 446, 367, 50 },
		{ 446, 368, 50 },
		{ 446, 369, 50 },
		{ 446, 370, 50 },
		{ 446, 371, 50 },
		{ 446, 372, 50 },
		{ 446, 373, 50 },
		{ 446, 374, 50 },
		{ 446, 375, 50 },
		{ 447, 329, 50 },
		{ 447, 330, 50 },
		{ 447, 331, 50 },
		{ 447, 332, 50 },
		{ 447, 333, 50 },
		{ 447, 334, 50 },
		{ 447, 335, 50 },
		{ 447, 336, 50 },
		{ 447, 337, 50 },
		{ 447, 338, 50 },
		{ 447, 339, 50 },
		{ 447, 340, 50 },
		{ 447, 341, 50 },
		{ 447, 342, 50 },
		{ 447, 343, 50 },
		{ 447, 344, 50 },
		{ 447, 345, 50 },
		{ 447, 346, 50 },
		{ 447, 347, 50 },
		{ 447, 348, 50 },
		{ 447, 349, 50 },
		{ 447, 350, 50 },
		{ 447, 351, 50 },
		{ 447, 352, 50 },
		{ 447, 353, 50 },
		{ 447, 354, 50 },
		{ 447, 355, 50 },
		{ 447, 356, 50 },
		{ 447, 357, 50 },
		{ 447, 358, 50 },
		{ 447, 359, 50 },
		{ 447, 360, 50 },
		{ 447, 361, 50 },
		{ 447, 362, 50 },
		{ 447, 363, 50 },
		{ 447, 364, 50 },
		{ 447, 365, 50 },
		{ 447, 366, 50 },
		{ 447, 367, 50 },
		{ 447, 368, 50 },
		{ 447, 369, 50 },
		{ 447, 370, 50 },
		{ 447, 371, 50 },
		{ 447, 372, 50 },
		{ 447, 373, 50 },
		{ 447, 374, 50 },
		{ 447, 375, 50 },
		{ 448, 329, 50 },
		{ 448, 330, 50 },
		{ 448, 331, 50 },
		{ 448, 332, 50 },
		{ 448, 333, 50 },
		{ 448, 334, 50 },
		{ 448, 335, 50 },
		{ 448, 336, 50 },
		{ 448, 337, 50 },
		{ 448, 338, 50 },
		{ 448, 339, 50 },
		{ 448, 340, 50 },
		{ 448, 341, 50 },
		{ 448, 342, 50 },
		{ 448, 343, 50 },
		{ 448, 344, 50 },
		{ 448, 345, 50 },
		{ 448, 346, 50 },
		{ 448, 347, 50 },
		{ 448, 348, 50 },
		{ 448, 349, 50 },
		{ 448, 350, 50 },
		{ 448, 351, 50 },
		{ 448, 352, 50 },
		{ 448, 353, 50 },
		{ 448, 354, 50 },
		{ 448, 355, 50 },
		{ 448, 356, 50 },
		{ 448, 357, 50 },
		{ 448, 358, 50 },
		{ 448, 359, 50 },
		{ 448, 360, 50 },
		{ 448, 361, 50 },
		{ 448, 362, 50 },
		{ 448, 363, 50 },
		{ 448, 364, 50 },
		{ 448, 365, 50 },
		{ 448, 366, 50 },
		{ 448, 367, 50 },
		{ 448, 368, 50 },
		{ 448, 369, 50 },
		{ 448, 370, 50 },
		{ 448, 371, 50 },
		{ 448, 372, 50 },
		{ 448, 373, 50 },
		{ 448, 374, 50 },
		{ 448, 375, 50 },
		{ 448, 376, 50 },
		{ 449, 330, 50 },
		{ 449, 331, 50 },
		{ 449, 332, 50 },
		{ 449, 333, 50 },
		{ 449, 334, 50 },
		{ 449, 335, 50 },
		{ 449, 336, 50 },
		{ 449, 337, 50 },
		{ 449, 338, 50 },
		{ 449, 339, 50 },
		{ 449, 340, 50 },
		{ 449, 341, 50 },
		{ 449, 342, 50 },
		{ 449, 343, 50 },
		{ 449, 344, 50 },
		{ 449, 345, 50 },
		{ 449, 346, 50 },
		{ 449, 347, 50 },
		{ 449, 348, 50 },
		{ 449, 349, 50 },
		{ 449, 350, 50 },
		{ 449, 351, 50 },
		{ 449, 352, 50 },
		{ 449, 353, 50 },
		{ 449, 354, 50 },
		{ 449, 355, 50 },
		{ 449, 356, 50 },
		{ 449, 357, 50 },
		{ 449, 358, 50 },
		{ 449, 359, 50 },
		{ 449, 360, 50 },
		{ 449, 361, 50 },
		{ 449, 362, 50 },
		{ 449, 363, 50 },
		{ 449, 364, 50 },
		{ 449, 365, 50 },
		{ 449, 366, 50 },
		{ 449, 367, 50 },
		{ 449, 368, 50 },
		{ 449, 369, 50 },
		{ 449, 370, 50 },
		{ 449, 371, 50 },
		{ 449, 372, 50 },
		{ 449, 373, 50 },
		{ 449, 374, 50 },
		{ 449, 375, 50 },
		{ 449, 376, 50 },
		{ 450, 330, 50 },
		{ 450, 331, 50 },
		{ 450, 332, 50 },
		{ 450, 333, 50 },
		{ 450, 334, 50 },
		{ 450, 335, 50 },
		{ 450, 336, 50 },
		{ 450, 337, 50 },
		{ 450, 338, 50 },
		{ 450, 339, 50 },
		{ 450, 340, 50 },
		{ 450, 341, 50 },
		{ 450, 342, 50 },
		{ 450, 343, 50 },
		{ 450, 344, 50 },
		{ 450, 345, 50 },
		{ 450, 346, 50 },
		{ 450, 347, 50 },
		{ 450, 348, 50 },
		{ 450, 349, 50 },
		{ 450, 350, 50 },
		{ 450, 351, 50 },
		{ 450, 352, 50 },
		{ 450, 353, 50 },
		{ 450, 354, 50 },
		{ 450, 355, 50 },
		{ 450, 356, 50 },
		{ 450, 357, 50 },
		{ 450, 358, 50 },
		{ 450, 359, 50 },
		{ 450, 360, 50 },
		{ 450, 361, 50 },
		{ 450, 362, 50 },
		{ 450, 363, 50 },
		{ 450, 364, 50 },
		{ 450, 365, 50 },
		{ 450, 366, 50 },
		{ 450, 367, 50 },
		{ 450, 368, 50 },
		{ 450, 369, 50 },
		{ 450, 370, 50 },
		{ 450, 371, 50 },
		{ 450, 372, 50 },
		{ 450, 373, 50 },
		{ 450, 374, 50 },
		{ 450, 375, 50 },
		{ 450, 376, 50 },
		{ 450, 377, 50 },
		{ 451, 330, 50 },
		{ 451, 331, 50 },
		{ 451, 332, 50 },
		{ 451, 333, 50 },
		{ 451, 334, 50 },
		{ 451, 335, 50 },
		{ 451, 336, 50 },
		{ 451, 337, 50 },
		{ 451, 338, 50 },
		{ 451, 339, 50 },
		{ 451, 340, 50 },
		{ 451, 341, 50 },
		{ 451, 342, 50 },
		{ 451, 343, 50 },
		{ 451, 344, 50 },
		{ 451, 345, 50 },
		{ 451, 346, 50 },
		{ 451, 347, 50 },
		{ 451, 348, 50 },
		{ 451, 349, 50 },
		{ 451, 350, 50 },
		{ 451, 351, 50 },
		{ 451, 352, 50 },
		{ 451, 353, 50 },
		{ 451, 354, 50 },
		{ 451, 355, 50 },
		{ 451, 356, 50 },
		{ 451, 357, 50 },
		{ 451, 358, 50 },
		{ 451, 359, 50 },
		{ 451, 360, 50 },
		{ 451, 361, 50 },
		{ 451, 362, 50 },
		{ 451, 363, 50 },
		{ 451, 364, 50 },
		{ 451, 365, 50 },
		{ 451, 366, 50 },
		{ 451, 367, 50 },
		{ 451, 368, 50 },
		{ 451, 369, 50 },
		{ 451, 370, 50 },
		{ 451, 371, 50 },
		{ 451, 372, 50 },
		{ 451, 373, 50 },
		{ 451, 374, 50 },
		{ 451, 375, 50 },
		{ 451, 376, 50 },
		{ 451, 377, 50 },
		{ 451, 378, 50 },
		{ 452, 331, 50 },
		{ 452, 332, 50 },
		{ 452, 333, 50 },
		{ 452, 334, 50 },
		{ 452, 335, 50 },
		{ 452, 336, 50 },
		{ 452, 337, 50 },
		{ 452, 338, 50 },
		{ 452, 339, 50 },
		{ 452, 340, 50 },
		{ 452, 341, 50 },
		{ 452, 342, 50 },
		{ 452, 343, 50 },
		{ 452, 344, 50 },
		{ 452, 345, 50 },
		{ 452, 346, 50 },
		{ 452, 347, 50 },
		{ 452, 348, 50 },
		{ 452, 349, 50 },
		{ 452, 350, 50 },
		{ 452, 351, 50 },
		{ 452, 352, 50 },
		{ 452, 353, 50 },
		{ 452, 354, 50 },
		{ 452, 355, 50 },
		{ 452, 356, 50 },
		{ 452, 357, 50 },
		{ 452, 358, 50 },
		{ 452, 359, 50 },
		{ 452, 360, 50 },
		{ 452, 361, 50 },
		{ 452, 362, 50 },
		{ 452, 363, 50 },
		{ 452, 364, 50 },
		{ 452, 365, 50 },
		{ 452, 366, 50 },
		{ 452, 367, 50 },
		{ 452, 368, 50 },
		{ 452, 369, 50 },
		{ 452, 370, 50 },
		{ 452, 371, 50 },
		{ 452, 372, 50 },
		{ 452, 373, 50 },
		{ 452, 374, 50 },
		{ 452, 375, 50 },
		{ 452, 376, 50 },
		{ 452, 377, 50 },
		{ 452, 378, 50 },
		{ 453, 331, 50 },
		{ 453, 332, 50 },
		{ 453, 333, 50 },
		{ 453, 334, 50 },
		{ 453, 335, 50 },
		{ 453, 336, 50 },
		{ 453, 337, 50 },
		{ 453, 338, 50 },
		{ 453, 339, 50 },
		{ 453, 340, 50 },
		{ 453, 341, 50 },
		{ 453, 342, 50 },
		{ 453, 343, 50 },
		{ 453, 344, 50 },
		{ 453, 345, 50 },
		{ 453, 346, 50 },
		{ 453, 347, 50 },
		{ 453, 348, 50 },
		{ 453, 349, 50 },
		{ 453, 350, 50 },
		{ 453, 351, 50 },
		{ 453, 352, 50 },
		{ 453, 353, 50 },
		{ 453, 354, 50 },
		{ 453, 355, 50 },
		{ 453, 356, 50 },
		{ 453, 357, 50 },
		{ 453, 358, 50 },
		{ 453, 359, 50 },
		{ 453, 360, 50 },
		{ 453, 361, 50 },
		{ 453, 362, 50 },
		{ 453, 363, 50 },
		{ 453, 364, 50 },
		{ 453, 365, 50 },
		{ 453, 366, 50 },
		{ 453, 367, 50 },
		{ 453, 368, 50 },
		{ 453, 369, 50 },
		{ 453, 370, 50 },
		{ 453, 371, 50 },
		{ 453, 372, 50 },
		{ 453, 373, 50 },
		{ 453, 374, 50 },
		{ 453, 375, 50 },
		{ 453, 376, 50 },
		{ 453, 377, 50 },
		{ 453, 378, 50 },
		{ 453, 379, 50 },
		{ 454, 331, 50 },
		{ 454, 332, 50 },
		{ 454, 333, 50 },
		{ 454, 334, 50 },
		{ 454, 335, 50 },
		{ 454, 336, 50 },
		{ 454, 337, 50 },
		{ 454, 338, 50 },
		{ 454, 339, 50 },
		{ 454, 340, 50 },
		{ 454, 341, 50 },
		{ 454, 342, 50 },
		{ 454, 343, 50 },
		{ 454, 344, 50 },
		{ 454, 345, 50 },
		{ 454, 346, 50 },
		{ 454, 347, 50 },
		{ 454, 348, 50 },
		{ 454, 349, 50 },
		{ 454, 350, 50 },
		{ 454, 351, 50 },
		{ 454, 352, 50 },
		{ 454, 353, 50 },
		{ 454, 354, 50 },
		{ 454, 355, 50 },
		{ 454, 356, 50 },
		{ 454, 357, 50 },
		{ 454, 358, 50 },
		{ 454, 359, 50 },
		{ 454, 360, 50 },
		{ 454, 361, 50 },
		{ 454, 362, 50 },
		{ 454, 363, 50 },
		{ 454, 364, 50 },
		{ 454, 365, 50 },
		{ 454, 366, 50 },
		{ 454, 367, 50 },
		{ 454, 368, 50 },
		{ 454, 369, 50 },
		{ 454, 370, 50 },
		{ 454, 371, 50 },
		{ 454, 372, 50 },
		{ 454, 373, 50 },
		{ 454, 374, 50 },
		{ 454, 375, 50 },
		{ 454, 376, 50 },
		{ 454, 377, 50 },
		{ 454, 378, 50 },
		{ 454, 379, 50 },
		{ 454, 380, 50 },
		{ 455, 332, 50 },
		{ 455, 333, 50 },
		{ 455, 334, 50 },
		{ 455, 335, 50 },
		{ 455, 336, 50 },
		{ 455, 337, 50 },
		{ 455, 338, 50 },
		{ 455, 339, 50 },
		{ 455, 340, 50 },
		{ 455, 341, 50 },
		{ 455, 342, 50 },
		{ 455, 343, 50 },
		{ 455, 344, 50 },
		{ 455, 345, 50 },
		{ 455, 346, 50 },
		{ 455, 347, 50 },
		{ 455, 348, 50 },
		{ 455, 349, 50 },
		{ 455, 350, 50 },
		{ 455, 351, 50 },
		{ 455, 352, 50 },
		{ 455, 353, 50 },
		{ 455, 354, 50 },
		{ 455, 355, 50 },
		{ 455, 356, 50 },
		{ 455, 357, 50 },
		{ 455, 358, 50 },
		{ 455, 359, 50 },
		{ 455, 360, 50 },
		{ 455, 361, 50 },
		{ 455, 362, 50 },
		{ 455, 363, 50 },
		{ 455, 364, 50 },
		{ 455, 365, 50 },
		{ 455, 366, 50 },
		{ 455, 367, 50 },
		{ 455, 368, 50 },
		{ 455, 369, 50 },
		{ 455, 370, 50 },
		{ 455, 371, 50 },
		{ 455, 372, 50 },
		{ 455, 373, 50 },
		{ 455, 374, 50 },
		{ 455, 375, 50 },
		{ 455, 376, 50 },
		{ 455, 377, 50 },
		{ 455, 378, 50 },
		{ 455, 379, 50 },
		{ 455, 380, 50 },
		{ 456, 332, 50 },
		{ 456, 333, 50 },
		{ 456, 334, 50 },
		{ 456, 335, 50 },
		{ 456, 336, 50 },
		{ 456, 337, 50 },
		{ 456, 338, 50 },
		{ 456, 339, 50 },
		{ 456, 340, 50 },
		{ 456, 341, 50 },
		{ 456, 342, 50 },
		{ 456, 343, 50 },
		{ 456, 344, 50 },
		{ 456, 345, 50 },
		{ 456, 346, 50 },
		{ 456, 347, 50 },
		{ 456, 348, 50 },
		{ 456, 349, 50 },
		{ 456, 350, 50 },
		{ 456, 351, 50 },
		{ 456, 352, 50 },
		{ 456, 353, 50 },
		{ 456, 354, 50 },
		{ 456, 355, 50 },
		{ 456, 356, 50 },
		{ 456, 357, 50 },
		{ 456, 358, 50 },
		{ 456, 359, 50 },
		{ 456, 360, 50 },
		{ 456, 361, 50 },
		{ 456, 362, 50 },
		{ 456, 363, 50 },
		{ 456, 364, 50 },
		{ 456, 365, 50 },
		{ 456, 366, 50 },
		{ 456, 367, 50 },
		{ 456, 368, 50 },
		{ 456, 369, 50 },
		{ 456, 370, 50 },
		{ 456, 371, 50 },
		{ 456, 372, 50 },
		{ 456, 373, 50 },
		{ 456, 374, 50 },
		{ 456, 375, 50 },
		{ 456, 376, 50 },
		{ 456, 377, 50 },
		{ 456, 378, 50 },
		{ 456, 379, 50 },
		{ 456, 380, 50 },
		{ 456, 381, 50 },
		{ 457, 333, 50 },
		{ 457, 334, 50 },
		{ 457, 335, 50 },
		{ 457, 336, 50 },
		{ 457, 337, 50 },
		{ 457, 338, 50 },
		{ 457, 339, 50 },
		{ 457, 340, 50 },
		{ 457, 341, 50 },
		{ 457, 342, 50 },
		{ 457, 343, 50 },
		{ 457, 344, 50 },
		{ 457, 345, 50 },
		{ 457, 346, 50 },
		{ 457, 347, 50 },
		{ 457, 348, 50 },
		{ 457, 349, 50 },
		{ 457, 350, 50 },
		{ 457, 351, 50 },
		{ 457, 352, 50 },
		{ 457, 353, 50 },
		{ 457, 354, 50 },
		{ 457, 355, 50 },
		{ 457, 356, 50 },
		{ 457, 357, 50 },
		{ 457, 358, 50 },
		{ 457, 359, 50 },
		{ 457, 360, 50 },
		{ 457, 361, 50 },
		{ 457, 362, 50 },
		{ 457, 363, 50 },
		{ 457, 364, 50 },
		{ 457, 365, 50 },
		{ 457, 366, 50 },
		{ 457, 367, 50 },
		{ 457, 368, 50 },
		{ 457, 369, 50 },
		{ 457, 370, 50 },
		{ 457, 371, 50 },
		{ 457, 372, 50 },
		{ 457, 373, 50 },
		{ 457, 374, 50 },
		{ 457, 375, 50 },
		{ 457, 376, 50 },
		{ 457, 377, 50 },
		{ 457, 378, 50 },
		{ 457, 379, 50 },
		{ 457, 380, 50 },
		{ 457, 381, 50 },
		{ 458, 333, 50 },
		{ 458, 334, 50 },
		{ 458, 335, 50 },
		{ 458, 336, 50 },
		{ 458, 337, 50 },
		{ 458, 338, 50 },
		{ 458, 339, 50 },
		{ 458, 340, 50 },
		{ 458, 341, 50 },
		{ 458, 342, 50 },
		{ 458, 343, 50 },
		{ 458, 344, 50 },
		{ 458, 345, 50 },
		{ 458, 346, 50 },
		{ 458, 347, 50 },
		{ 458, 348, 50 },
		{ 458, 349, 50 },
		{ 458, 350, 50 },
		{ 458, 351, 50 },
		{ 458, 352, 50 },
		{ 458, 353, 50 },
		{ 458, 354, 50 },
		{ 458, 355, 50 },
		{ 458, 356, 50 },
		{ 458, 357, 50 },
		{ 458, 358, 50 },
		{ 458, 359, 50 },
		{ 458, 360, 50 },
		{ 458, 361, 50 },
		{ 458, 362, 50 },
		{ 458, 363, 50 },
		{ 458, 364, 50 },
		{ 458, 365, 50 },
		{ 458, 366, 50 },
		{ 458, 367, 50 },
		{ 458, 368, 50 },
		{ 458, 369, 50 },
		{ 458, 370, 50 },
		{ 458, 371, 50 },
		{ 458, 372, 50 },
		{ 458, 373, 50 },
		{ 458, 374, 50 },
		{ 458, 375, 50 },
		{ 458, 376, 50 },
		{ 458, 377, 50 },
		{ 458, 378, 50 },
		{ 458, 379, 50 },
		{ 458, 380, 50 },
		{ 458, 381, 50 },
		{ 458, 382, 50 },
		{ 459, 333, 50 },
		{ 459, 334, 50 },
		{ 459, 335, 50 },
		{ 459, 336, 50 },
		{ 459, 337, 50 },
		{ 459, 338, 50 },
		{ 459, 339, 50 },
		{ 459, 340, 50 },
		{ 459, 341, 50 },
		{ 459, 342, 50 },
		{ 459, 343, 50 },
		{ 459, 344, 50 },
		{ 459, 345, 50 },
		{ 459, 346, 50 },
		{ 459, 347, 50 },
		{ 459, 348, 50 },
		{ 459, 349, 50 },
		{ 459, 350, 50 },
		{ 459, 351, 50 },
		{ 459, 352, 50 },
		{ 459, 353, 50 },
		{ 459, 354, 50 },
		{ 459, 355, 50 },
		{ 459, 356, 50 },
		{ 459, 357, 50 },
		{ 459, 358, 50 },
		{ 459, 359, 50 },
		{ 459, 360, 50 },
		{ 459, 361, 50 },
		{ 459, 362, 50 },
		{ 459, 363, 50 },
		{ 459, 364, 50 },
		{ 459, 365, 50 },
		{ 459, 366, 50 },
		{ 459, 367, 50 },
		{ 459, 368, 50 },
		{ 459, 369, 50 },
		{ 459, 370, 50 },
		{ 459, 371, 50 },
		{ 459, 372, 50 },
		{ 459, 373, 50 },
		{ 459, 374, 50 },
		{ 459, 375, 50 },
		{ 459, 376, 50 },
		{ 459, 377, 50 },
		{ 459, 378, 50 },
		{ 459, 379, 50 },
		{ 459, 380, 50 },
		{ 459, 381, 50 },
		{ 459, 382, 50 },
		{ 459, 383, 50 },
		{ 460, 334, 50 },
		{ 460, 335, 50 },
		{ 460, 336, 50 },
		{ 460, 337, 50 },
		{ 460, 338, 50 },
		{ 460, 339, 50 },
		{ 460, 340, 50 },
		{ 460, 341, 50 },
		{ 460, 342, 50 },
		{ 460, 343, 50 },
		{ 460, 344, 50 },
		{ 460, 345, 50 },
		{ 460, 346, 50 },
		{ 460, 347, 50 },
		{ 460, 348, 50 },
		{ 460, 349, 50 },
		{ 460, 350, 50 },
		{ 460, 351, 50 },
		{ 460, 352, 50 },
		{ 460, 353, 50 },
		{ 460, 354, 50 },
		{ 460, 355, 50 },
		{ 460, 356, 50 },
		{ 460, 357, 50 },
		{ 460, 358, 50 },
		{ 460, 359, 50 },
		{ 460, 360, 50 },
		{ 460, 361, 50 },
		{ 460, 362, 50 },
		{ 460, 363, 50 },
		{ 460, 364, 50 },
		{ 460, 365, 50 },
		{ 460, 366, 50 },
		{ 460, 367, 50 },
		{ 460, 368, 50 },
		{ 460, 369, 50 },
		{ 460, 370, 50 },
		{ 460, 371, 50 },
		{ 460, 372, 50 },
		{ 460, 373, 50 },
		{ 460, 374, 50 },
		{ 460, 375, 50 },
		{ 460, 376, 50 },
		{ 460, 377, 50 },
		{ 460, 378, 50 },
		{ 460, 379, 50 },
		{ 460, 380, 50 },
		{ 460, 381, 50 },
		{ 460, 382, 50 },
		{ 460, 383, 50 },
		{ 461, 334, 50 },
		{ 461, 335, 50 },
		{ 461, 336, 50 },
		{ 461, 337, 50 },
		{ 461, 338, 50 },
		{ 461, 339, 50 },
		{ 461, 340, 50 },
		{ 461, 341, 50 },
		{ 461, 342, 50 },
		{ 461, 343, 50 },
		{ 461, 344, 50 },
		{ 461, 345, 50 },
		{ 461, 346, 50 },
		{ 461, 347, 50 },
		{ 461, 348, 50 },
		{ 461, 349, 50 },
		{ 461, 350, 50 },
		{ 461, 351, 50 },
		{ 461, 352, 50 },
		{ 461, 353, 50 },
		{ 461, 354, 50 },
		{ 461, 355, 50 },
		{ 461, 356, 50 },
		{ 461, 357, 50 },
		{ 461, 358, 50 },
		{ 461, 359, 50 },
		{ 461, 360, 50 },
		{ 461, 361, 50 },
		{ 461, 362, 50 },
		{ 461, 363, 50 },
		{ 461, 364, 50 },
		{ 461, 365, 50 },
		{ 461, 366, 50 },
		{ 461, 367, 50 },
		{ 461, 368, 50 },
		{ 461, 369, 50 },
		{ 461, 370, 50 },
		{ 461, 371, 50 },
		{ 461, 372, 50 },
		{ 461, 373, 50 },
		{ 461, 374, 50 },
		{ 461, 375, 50 },
		{ 461, 376, 50 },
		{ 461, 377, 50 },
		{ 461, 378, 50 },
		{ 461, 379, 50 },
		{ 461, 380, 50 },
		{ 461, 381, 50 },
		{ 461, 382, 50 },
		{ 461, 383, 50 },
		{ 461, 384, 50 },
		{ 462, 334, 50 },
		{ 462, 335, 50 },
		{ 462, 336, 50 },
		{ 462, 337, 50 },
		{ 462, 338, 50 },
		{ 462, 339, 50 },
		{ 462, 340, 50 },
		{ 462, 341, 50 },
		{ 462, 342, 50 },
		{ 462, 343, 50 },
		{ 462, 344, 50 },
		{ 462, 345, 50 },
		{ 462, 346, 50 },
		{ 462, 347, 50 },
		{ 462, 348, 50 },
		{ 462, 349, 50 },
		{ 462, 350, 50 },
		{ 462, 351, 50 },
		{ 462, 352, 50 },
		{ 462, 353, 50 },
		{ 462, 354, 50 },
		{ 462, 355, 50 },
		{ 462, 356, 50 },
		{ 462, 357, 50 },
		{ 462, 358, 50 },
		{ 462, 359, 50 },
		{ 462, 360, 50 },
		{ 462, 361, 50 },
		{ 462, 362, 50 },
		{ 462, 363, 50 },
		{ 462, 364, 50 },
		{ 462, 365, 50 },
		{ 462, 366, 50 },
		{ 462, 367, 50 },
		{ 462, 368, 50 },
		{ 462, 369, 50 },
		{ 462, 370, 50 },
		{ 462, 371, 50 },
		{ 462, 372, 50 },
		{ 462, 373, 50 },
		{ 462, 374, 50 },
		{ 462, 375, 50 },
		{ 462, 376, 50 },
		{ 462, 377, 50 },
		{ 462, 378, 50 },
		{ 462, 379, 50 },
		{ 462, 380, 50 },
		{ 462, 381, 50 },
		{ 462, 382, 50 },
		{ 462, 383, 50 },
		{ 462, 384, 50 },
		{ 462, 385, 50 },
		{ 463, 335, 50 },
		{ 463, 336, 50 },
		{ 463, 337, 50 },
		{ 463, 338, 50 },
		{ 463, 339, 50 },
		{ 463, 340, 50 },
		{ 463, 341, 50 },
		{ 463, 342, 50 },
		{ 463, 343, 50 },
		{ 463, 344, 50 },
		{ 463, 345, 50 },
		{ 463, 346, 50 },
		{ 463, 347, 50 },
		{ 463, 348, 50 },
		{ 463, 349, 50 },
		{ 463, 350, 50 },
		{ 463, 351, 50 },
		{ 463, 352, 50 },
		{ 463, 353, 50 },
		{ 463, 354, 50 },
		{ 463, 355, 50 },
		{ 463, 356, 50 },
		{ 463, 357, 50 },
		{ 463, 358, 50 },
		{ 463, 359, 50 },
		{ 463, 360, 50 },
		{ 463, 361, 50 },
		{ 463, 362, 50 },
		{ 463, 363, 50 },
		{ 463, 364, 50 },
		{ 463, 365, 50 },
		{ 463, 366, 50 },
		{ 463, 367, 50 },
		{ 463, 368, 50 },
		{ 463, 369, 50 },
		{ 463, 370, 50 },
		{ 463, 371, 50 },
		{ 463, 372, 50 },
		{ 463, 373, 50 },
		{ 463, 374, 50 },
		{ 463, 375, 50 },
		{ 463, 376, 50 },
		{ 463, 377, 50 },
		{ 463, 378, 50 },
		{ 463, 379, 50 },
		{ 463, 380, 50 },
		{ 463, 381, 50 },
		{ 463, 382, 50 },
		{ 463, 383, 50 },
		{ 463, 384, 50 },
		{ 463, 385, 50 },
		{ 464, 335, 50 },
		{ 464, 336, 50 },
		{ 464, 337, 50 },
		{ 464, 338, 50 },
		{ 464, 339, 50 },
		{ 464, 340, 50 },
		{ 464, 341, 50 },
		{ 464, 342, 50 },
		{ 464, 343, 50 },
		{ 464, 344, 50 },
		{ 464, 345, 50 },
		{ 464, 346, 50 },
		{ 464, 347, 50 },
		{ 464, 348, 50 },
		{ 464, 349, 50 },
		{ 464, 350, 50 },
		{ 464, 351, 50 },
		{ 464, 352, 50 },
		{ 464, 353, 50 },
		{ 464, 354, 50 },
		{ 464, 355, 50 },
		{ 464, 356, 50 },
		{ 464, 357, 50 },
		{ 464, 358, 50 },
		{ 464, 359, 50 },
		{ 464, 360, 50 },
		{ 464, 361, 50 },
		{ 464, 362, 50 },
		{ 464, 363, 50 },
		{ 464, 364, 50 },
		{ 464, 365, 50 },
		{ 464, 366, 50 },
		{ 464, 367, 50 },
		{ 464, 368, 50 },
		{ 464, 369, 50 },
		{ 464, 370, 50 },
		{ 464, 371, 50 },
		{ 464, 372, 50 },
		{ 464, 373, 50 },
		{ 464, 374, 50 },
		{ 464, 375, 50 },
		{ 464, 376, 50 },
		{ 464, 377, 50 },
		{ 464, 378, 50 },
		{ 464, 379, 50 },
		{ 464, 380, 50 },
		{ 464, 381, 50 },
		{ 464, 382, 50 },
		{ 464, 383, 50 },
		{ 464, 384, 50 },
		{ 464, 385, 50 },
		{ 464, 386, 50 },
		{ 465, 336, 50 },
		{ 465, 337, 50 },
		{ 465, 338, 50 },
		{ 465, 339, 50 },
		{ 465, 340, 50 },
		{ 465, 341, 50 },
		{ 465, 342, 50 },
		{ 465, 343, 50 },
		{ 465, 344, 50 },
		{ 465, 345, 50 },
		{ 465, 346, 50 },
		{ 465, 347, 50 },
		{ 465, 348, 50 },
		{ 465, 349, 50 },
		{ 465, 350, 50 },
		{ 465, 351, 50 },
		{ 465, 352, 50 },
		{ 465, 353, 50 },
		{ 465, 354, 50 },
		{ 465, 355, 50 },
		{ 465, 356, 50 },
		{ 465, 357, 50 },
		{ 465, 358, 50 },
		{ 465, 359, 50 },
		{ 465, 360, 50 },
		{ 465, 361, 50 },
		{ 465, 362, 50 },
		{ 465, 363, 50 },
		{ 465, 364, 50 },
		{ 465, 365, 50 },
		{ 465, 366, 50 },
		{ 465, 367, 50 },
		{ 465, 368, 50 },
		{ 465, 369, 50 },
		{ 465, 370, 50 },
		{ 465, 371, 50 },
		{ 465, 372, 50 },
		{ 465, 373, 50 },
		{ 465, 374, 50 },
		{ 465, 375, 50 },
		{ 465, 376, 50 },
		{ 465, 377, 50 },
		{ 465, 378, 50 },
		{ 465, 379, 50 },
		{ 465, 380, 50 },
		{ 465, 381, 50 },
		{ 465, 382, 50 },
		{ 465, 383, 50 },
		{ 465, 384, 50 },
		{ 465, 385, 50 },
		{ 465, 386, 50 },
		{ 466, 336, 50 },
		{ 466, 337, 50 },
		{ 466, 338, 50 },
		{ 466, 339, 50 },
		{ 466, 340, 50 },
		{ 466, 341, 50 },
		{ 466, 342, 50 },
		{ 466, 343, 50 },
		{ 466, 344, 50 },
		{ 466, 345, 50 },
		{ 466, 346, 50 },
		{ 466, 347, 50 },
		{ 466, 348, 50 },
		{ 466, 349, 50 },
		{ 466, 350, 50 },
		{ 466, 351, 50 },
		{ 466, 352, 50 },
		{ 466, 353, 50 },
		{ 466, 354, 50 },
		{ 466, 355, 50 },
		{ 466, 356, 50 },
		{ 466, 357, 50 },
		{ 466, 358, 50 },
		{ 466, 359, 50 },
		{ 466, 360, 50 },
		{ 466, 361, 50 },
		{ 466, 362, 50 },
		{ 466, 363, 50 },
		{ 466, 364, 50 },
		{ 466, 365, 50 },
		{ 466, 366, 50 },
		{ 466, 367, 50 },
		{ 466, 368, 50 },
		{ 466, 369, 50 },
		{ 466, 370, 50 },
		{ 466, 371, 50 },
		{ 466, 372, 50 },
		{ 466, 373, 50 },
		{ 466, 374, 50 },
		{ 466, 375, 50 },
		{ 466, 376, 50 },
		{ 466, 377, 50 },
		{ 466, 378, 50 },
		{ 466, 379, 50 },
		{ 466, 380, 50 },
		{ 466, 381, 50 },
		{ 466, 382, 50 },
		{ 466, 383, 50 },
		{ 466, 384, 50 },
		{ 466, 385, 50 },
		{ 466, 386, 50 },
		{ 466, 387, 50 },
		{ 467, 336, 50 },
		{ 467, 337, 50 },
		{ 467, 338, 50 },
		{ 467, 339, 50 },
		{ 467, 340, 50 },
		{ 467, 341, 50 },
		{ 467, 342, 50 },
		{ 467, 343, 50 },
		{ 467, 344, 50 },
		{ 467, 345, 50 },
		{ 467, 346, 50 },
		{ 467, 347, 50 },
		{ 467, 348, 50 },
		{ 467, 349, 50 },
		{ 467, 350, 50 },
		{ 467, 351, 50 },
		{ 467, 352, 50 },
		{ 467, 353, 50 },
		{ 467, 354, 50 },
		{ 467, 355, 50 },
		{ 467, 356, 50 },
		{ 467, 357, 50 },
		{ 467, 358, 50 },
		{ 467, 359, 50 },
		{ 467, 360, 50 },
		{ 467, 361, 50 },
		{ 467, 362, 50 },
		{ 467, 363, 50 },
		{ 467, 364, 50 },
		{ 467, 365, 50 },
		{ 467, 366, 50 },
		{ 467, 367, 50 },
		{ 467, 368, 50 },
		{ 467, 369, 50 },
		{ 467, 370, 50 },
		{ 467, 371, 50 },
		{ 467, 372, 50 },
		{ 467, 373, 50 },
		{ 467, 374, 50 },
		{ 467, 375, 50 },
		{ 467, 376, 50 },
		{ 467, 377, 50 },
		{ 467, 378, 50 },
		{ 467, 379, 50 },
		{ 467, 380, 50 },
		{ 467, 381, 50 },
		{ 467, 382, 50 },
		{ 467, 383, 50 },
		{ 467, 384, 50 },
		{ 467, 385, 50 },
		{ 467, 386, 50 },
		{ 467, 387, 50 },
		{ 467, 388, 50 },
		{ 468, 337, 50 },
		{ 468, 338, 50 },
		{ 468, 339, 50 },
		{ 468, 340, 50 },
		{ 468, 341, 50 },
		{ 468, 342, 50 },
		{ 468, 343, 50 },
		{ 468, 344, 50 },
		{ 468, 345, 50 },
		{ 468, 346, 50 },
		{ 468, 347, 50 },
		{ 468, 348, 50 },
		{ 468, 349, 50 },
		{ 468, 350, 50 },
		{ 468, 351, 50 },
		{ 468, 352, 50 },
		{ 468, 353, 50 },
		{ 468, 354, 50 },
		{ 468, 355, 50 },
		{ 468, 356, 50 },
		{ 468, 357, 50 },
		{ 468, 358, 50 },
		{ 468, 359, 50 },
		{ 468, 360, 50 },
		{ 468, 361, 50 },
		{ 468, 362, 50 },
		{ 468, 363, 50 },
		{ 468, 364, 50 },
		{ 468, 365, 50 },
		{ 468, 366, 50 },
		{ 468, 367, 50 },
		{ 468, 368, 50 },
		{ 468, 369, 50 },
		{ 468, 370, 50 },
		{ 468, 371, 50 },
		{ 468, 372, 50 },
		{ 468, 373, 50 },
		{ 468, 374, 50 },
		{ 468, 375, 50 },
		{ 468, 376, 50 },
		{ 468, 377, 50 },
		{ 468, 378, 50 },
		{ 468, 379, 50 },
		{ 468, 380, 50 },
		{ 468, 381, 50 },
		{ 468, 382, 50 },
		{ 468, 383, 50 },
		{ 468, 384, 50 },
		{ 468, 385, 50 },
		{ 468, 386, 50 },
		{ 468, 387, 50 },
		{ 468, 388, 50 },
		{ 469, 337, 50 },
		{ 469, 338, 50 },
		{ 469, 339, 50 },
		{ 469, 340, 50 },
		{ 469, 341, 50 },
		{ 469, 342, 50 },
		{ 469, 343, 50 },
		{ 469, 344, 50 },
		{ 469, 345, 50 },
		{ 469, 346, 50 },
		{ 469, 347, 50 },
		{ 469, 348, 50 },
		{ 469, 349, 50 },
		{ 469, 350, 50 },
		{ 469, 351, 50 },
		{ 469, 352, 50 },
		{ 469, 353, 50 },
		{ 469, 354, 50 },
		{ 469, 355, 50 },
		{ 469, 356, 50 },
		{ 469, 357, 50 },
		{ 469, 358, 50 },
		{ 469, 359, 50 },
		{ 469, 360, 50 },
		{ 469, 361, 50 },
		{ 469, 362, 50 },
		{ 469, 363, 50 },
		{ 469, 364, 50 },
		{ 469, 365, 50 },
		{ 469, 366, 50 },
		{ 469, 367, 50 },
		{ 469, 368, 50 },
		{ 469, 369, 50 },
		{ 469, 370, 50 },
		{ 469, 371, 50 },
		{ 469, 372, 50 },
		{ 469, 373, 50 },
		{ 469, 374, 50 },
		{ 469, 375, 50 },
		{ 469, 376, 50 },
		{ 469, 377, 50 },
		{ 469, 378, 50 },
		{ 469, 379, 50 },
		{ 469, 380, 50 },
		{ 469, 381, 50 },
		{ 469, 382, 50 },
		{ 469, 383, 50 },
		{ 469, 384, 50 },
		{ 469, 385, 50 },
		{ 469, 386, 50 },
		{ 469, 387, 50 },
		{ 469, 388, 50 },
		{ 469, 389, 50 },
		{ 470, 337, 50 },
		{ 470, 338, 50 },
		{ 470, 339, 50 },
		{ 470, 340, 50 },
		{ 470, 341, 50 },
		{ 470, 342, 50 },
		{ 470, 343, 50 },
		{ 470, 344, 50 },
		{ 470, 345, 50 },
		{ 470, 346, 50 },
		{ 470, 347, 50 },
		{ 470, 348, 50 },
		{ 470, 349, 50 },
		{ 470, 350, 50 },
		{ 470, 351, 50 },
		{ 470, 352, 50 },
		{ 470, 353, 50 },
		{ 470, 354, 50 },
		{ 470, 355, 50 },
		{ 470, 356, 50 },
		{ 470, 357, 50 },
		{ 470, 358, 50 },
		{ 470, 359, 50 },
		{ 470, 360, 50 },
		{ 470, 361, 50 },
		{ 470, 362, 50 },
		{ 470, 363, 50 },
		{ 470, 364, 50 },
		{ 470, 365, 50 },
		{ 470, 366, 50 },
		{ 470, 367, 50 },
		{ 470, 368, 50 },
		{ 470, 369, 50 },
		{ 470, 370, 50 },
		{ 470, 371, 50 },
		{ 470, 372, 50 },
		{ 470, 373, 50 },
		{ 470, 374, 50 },
		{ 470, 375, 50 },
		{ 470, 376, 50 },
		{ 470, 377, 50 },
		{ 470, 378, 50 },
		{ 470, 379, 50 },
		{ 470, 380, 50 },
		{ 470, 381, 50 },
		{ 470, 382, 50 },
		{ 470, 383, 50 },
		{ 470, 384, 50 },
		{ 470, 385, 50 },
		{ 470, 386, 50 },
		{ 470, 387, 50 },
		{ 470, 388, 50 },
		{ 470, 389, 50 },
		{ 470, 390, 50 },
		{ 471, 338, 50 },
		{ 471, 339, 50 },
		{ 471, 340, 50 },
		{ 471, 341, 50 },
		{ 471, 342, 50 },
		{ 471, 343, 50 },
		{ 471, 344, 50 },
		{ 471, 345, 50 },
		{ 471, 346, 50 },
		{ 471, 347, 50 },
		{ 471, 348, 50 },
		{ 471, 349, 50 },
		{ 471, 350, 50 },
		{ 471, 351, 50 },
		{ 471, 352, 50 },
		{ 471, 353, 50 },
		{ 471, 354, 50 },
		{ 471, 355, 50 },
		{ 471, 356, 50 },
		{ 471, 357, 50 },
		{ 471, 358, 50 },
		{ 471, 359, 50 },
		{ 471, 360, 50 },
		{ 471, 361, 50 },
		{ 471, 362, 50 },
		{ 471, 363, 50 },
		{ 471, 364, 50 },
		{ 471, 365, 50 },
		{ 471, 366, 50 },
		{ 471, 367, 50 },
		{ 471, 368, 50 },
		{ 471, 369, 50 },
		{ 471, 370, 50 },
		{ 471, 371, 50 },
		{ 471, 372, 50 },
		{ 471, 373, 50 },
		{ 471, 374, 50 },
		{ 471, 375, 50 },
		{ 471, 376, 50 },
		{ 471, 377, 50 },
		{ 471, 378, 50 },
		{ 471, 379, 50 },
		{ 471, 380, 50 },
		{ 471, 381, 50 },
		{ 471, 382, 50 },
		{ 471, 383, 50 },
		{ 471, 384, 50 },
		{ 471, 385, 50 },
		{ 471, 386, 50 },
		{ 471, 387, 50 },
		{ 471, 388, 50 },
		{ 471, 389, 50 },
		{ 471, 390, 50 },
		{ 472, 338, 50 },
		{ 472, 339, 50 },
		{ 472, 340, 50 },
		{ 472, 341, 50 },
		{ 472, 342, 50 },
		{ 472, 343, 50 },
		{ 472, 344, 50 },
		{ 472, 345, 50 },
		{ 472, 346, 50 },
		{ 472, 347, 50 },
		{ 472, 348, 50 },
		{ 472, 349, 50 },
		{ 472, 350, 50 },
		{ 472, 351, 50 },
		{ 472, 352, 50 },
		{ 472, 353, 50 },
		{ 472, 354, 50 },
		{ 472, 355, 50 },
		{ 472, 356, 50 },
		{ 472, 357, 50 },
		{ 472, 358, 50 },
		{ 472, 359, 50 },
		{ 472, 360, 50 },
		{ 472, 361, 50 },
		{ 472, 362, 50 },
		{ 472, 363, 50 },
		{ 472, 364, 50 },
		{ 472, 365, 50 },
		{ 472, 366, 50 },
		{ 472, 367, 50 },
		{ 472, 368, 50 },
		{ 472, 369, 50 },
		{ 472, 370, 50 },
		{ 472, 371, 50 },
		{ 472, 372, 50 },
		{ 472, 373, 50 },
		{ 472, 374, 50 },
		{ 472, 375, 50 },
		{ 472, 376, 50 },
		{ 472, 377, 50 },
		{ 472, 378, 50 },
		{ 472, 379, 50 },
		{ 472, 380, 50 },
		{ 472, 381, 50 },
		{ 472, 382, 50 },
		{ 472, 383, 50 },
		{ 472, 384, 50 },
		{ 472, 385, 50 },
		{ 472, 386, 50 },
		{ 472, 387, 50 },
		{ 472, 388, 50 },
		{ 472, 389, 50 },
		{ 472, 390, 50 },
		{ 472, 391, 50 },
		{ 473, 339, 50 },
		{ 473, 340, 50 },
		{ 473, 341, 50 },
		{ 473, 342, 50 },
		{ 473, 343, 50 },
		{ 473, 344, 50 },
		{ 473, 345, 50 },
		{ 473, 346, 50 },
		{ 473, 347, 50 },
		{ 473, 348, 50 },
		{ 473, 349, 50 },
		{ 473, 350, 50 },
		{ 473, 351, 50 },
		{ 473, 352, 50 },
		{ 473, 353, 50 },
		{ 473, 354, 50 },
		{ 473, 355, 50 },
		{ 473, 356, 50 },
		{ 473, 357, 50 },
		{ 473, 358, 50 },
		{ 473, 359, 50 },
		{ 473, 360, 50 },
		{ 473, 361, 50 },
		{ 473, 362, 50 },
		{ 473, 363, 50 },
		{ 473, 364, 50 },
		{ 473, 365, 50 },
		{ 473, 366, 50 },
		{ 473, 367, 50 },
		{ 473, 368, 50 },
		{ 473, 369, 50 },
		{ 473, 370, 50 },
		{ 473, 371, 50 },
		{ 473, 372, 50 },
		{ 473, 373, 50 },
		{ 473, 374, 50 },
		{ 473, 375, 50 },
		{ 473, 376, 50 },
		{ 473, 377, 50 },
		{ 473, 378, 50 },
		{ 473, 379, 50 },
		{ 473, 380, 50 },
		{ 473, 381, 50 },
		{ 473, 382, 50 },
		{ 473, 383, 50 },
		{ 473, 384, 50 },
		{ 473, 385, 50 },
		{ 473, 386, 50 },
		{ 473, 387, 50 },
		{ 473, 388, 50 },
		{ 473, 389, 50 },
		{ 473, 390, 50 },
		{ 473, 391, 50 },
		{ 474, 339, 50 },
		{ 474, 340, 50 },
		{ 474, 341, 50 },
		{ 474, 342, 50 },
		{ 474, 343, 50 },
		{ 474, 344, 50 },
		{ 474, 345, 50 },
		{ 474, 346, 50 },
		{ 474, 347, 50 },
		{ 474, 348, 50 },
		{ 474, 349, 50 },
		{ 474, 350, 50 },
		{ 474, 351, 50 },
		{ 474, 352, 50 },
		{ 474, 353, 50 },
		{ 474, 354, 50 },
		{ 474, 355, 50 },
		{ 474, 356, 50 },
		{ 474, 357, 50 },
		{ 474, 358, 50 },
		{ 474, 359, 50 },
		{ 474, 360, 50 },
		{ 474, 361, 50 },
		{ 474, 362, 50 },
		{ 474, 363, 50 },
		{ 474, 364, 50 },
		{ 474, 365, 50 },
		{ 474, 366, 50 },
		{ 474, 367, 50 },
		{ 474, 368, 50 },
		{ 474, 369, 50 },
		{ 474, 370, 50 },
		{ 474, 371, 50 },
		{ 474, 372, 50 },
		{ 474, 373, 50 },
		{ 474, 374, 50 },
		{ 474, 375, 50 },
		{ 474, 376, 50 },
		{ 474, 377, 50 },
		{ 474, 378, 50 },
		{ 474, 379, 50 },
		{ 474, 380, 50 },
		{ 474, 381, 50 },
		{ 474, 382, 50 },
		{ 474, 383, 50 },
		{ 474, 384, 50 },
		{ 474, 385, 50 },
		{ 474, 386, 50 },
		{ 474, 387, 50 },
		{ 474, 388, 50 },
		{ 474, 389, 50 },
		{ 474, 390, 50 },
		{ 474, 391, 50 },
		{ 474, 392, 50 },
		{ 475, 339, 50 },
		{ 475, 340, 50 },
		{ 475, 341, 50 },
		{ 475, 342, 50 },
		{ 475, 343, 50 },
		{ 475, 344, 50 },
		{ 475, 345, 50 },
		{ 475, 346, 50 },
		{ 475, 347, 50 },
		{ 475, 348, 50 },
		{ 475, 349, 50 },
		{ 475, 350, 50 },
		{ 475, 351, 50 },
		{ 475, 352, 50 },
		{ 475, 353, 50 },
		{ 475, 354, 50 },
		{ 475, 355, 50 },
		{ 475, 356, 50 },
		{ 475, 357, 50 },
		{ 475, 358, 50 },
		{ 475, 359, 50 },
		{ 475, 360, 50 },
		{ 475, 361, 50 },
		{ 475, 362, 50 },
		{ 475, 363, 50 },
		{ 475, 364, 50 },
		{ 475, 365, 50 },
		{ 475, 366, 50 },
		{ 475, 367, 50 },
		{ 475, 368, 50 },
		{ 475, 369, 50 },
		{ 475, 370, 50 },
		{ 475, 371, 50 },
		{ 475, 372, 50 },
		{ 475, 373, 50 },
		{ 475, 374, 50 },
		{ 475, 375, 50 },
		{ 475, 376, 50 },
		{ 475, 377, 50 },
		{ 475, 378, 50 },
		{ 475, 379, 50 },
		{ 475, 380, 50 },
		{ 475, 381, 50 },
		{ 475, 382, 50 },
		{ 475, 383, 50 },
		{ 475, 384, 50 },
		{ 475, 385, 50 },
		{ 475, 386, 50 },
		{ 475, 387, 50 },
		{ 475, 388, 50 },
		{ 475, 389, 50 },
		{ 475, 390, 50 },
		{ 475, 391, 50 },
		{ 475, 392, 50 },
		{ 475, 393, 50 },
		{ 476, 340, 50 },
		{ 476, 341, 50 },
		{ 476, 342, 50 },
		{ 476, 343, 50 },
		{ 476, 344, 50 },
		{ 476, 345, 50 },
		{ 476, 346, 50 },
		{ 476, 347, 50 },
		{ 476, 348, 50 },
		{ 476, 349, 50 },
		{ 476, 350, 50 },
		{ 476, 351, 50 },
		{ 476, 352, 50 },
		{ 476, 353, 50 },
		{ 476, 354, 50 },
		{ 476, 355, 50 },
		{ 476, 356, 50 },
		{ 476, 357, 50 },
		{ 476, 358, 50 },
		{ 476, 359, 50 },
		{ 476, 360, 50 },
		{ 476, 361, 50 },
		{ 476, 362, 50 },
		{ 476, 363, 50 },
		{ 476, 364, 50 },
		{ 476, 365, 50 },
		{ 476, 366, 50 },
		{ 476, 367, 50 },
		{ 476, 368, 50 },
		{ 476, 369, 50 },
		{ 476, 370, 50 },
		{ 476, 371, 50 },
		{ 476, 372, 50 },
		{ 476, 373, 50 },
		{ 476, 374, 50 },
		{ 476, 375, 50 },
		{ 476, 376, 50 },
		{ 476, 377, 50 },
		{ 476, 378, 50 },
		{ 476, 379, 50 },
		{ 476, 380, 50 },
		{ 476, 381, 50 },
		{ 476, 382, 50 },
		{ 476, 383, 50 },
		{ 476, 384, 50 },
		{ 476, 385, 50 },
		{ 476, 386, 50 },
		{ 476, 387, 50 },
		{ 476, 388, 50 },
		{ 476, 389, 50 },
		{ 476, 390, 50 },
		{ 476, 391, 50 },
		{ 476, 392, 50 },
		{ 476, 393, 50 },
		{ 477, 340, 50 },
		{ 477, 341, 50 },
		{ 477, 342, 50 },
		{ 477, 343, 50 },
		{ 477, 344, 50 },
		{ 477, 345, 50 },
		{ 477, 346, 50 },
		{ 477, 347, 50 },
		{ 477, 348, 50 },
		{ 477, 349, 50 },
		{ 477, 350, 50 },
		{ 477, 351, 50 },
		{ 477, 352, 50 },
		{ 477, 353, 50 },
		{ 477, 354, 50 },
		{ 477, 355, 50 },
		{ 477, 356, 50 },
		{ 477, 357, 50 },
		{ 477, 358, 50 },
		{ 477, 359, 50 },
		{ 477, 360, 50 },
		{ 477, 361, 50 },
		{ 477, 362, 50 },
		{ 477, 363, 50 },
		{ 477, 364, 50 },
		{ 477, 365, 50 },
		{ 477, 366, 50 },
		{ 477, 367, 50 },
		{ 477, 368, 50 },
		{ 477, 369, 50 },
		{ 477, 370, 50 },
		{ 477, 371, 50 },
		{ 477, 372, 50 },
		{ 477, 373, 50 },
		{ 477, 374, 50 },
		{ 477, 375, 50 },
		{ 477, 376, 50 },
		{ 477, 377, 50 },
		{ 477, 378, 50 },
		{ 477, 379, 50 },
		{ 477, 380, 50 },
		{ 477, 381, 50 },
		{ 477, 382, 50 },
		{ 477, 383, 50 },
		{ 477, 384, 50 },
		{ 477, 385, 50 },
		{ 477, 386, 50 },
		{ 477, 387, 50 },
		{ 477, 388, 50 },
		{ 477, 389, 50 },
		{ 477, 390, 50 },
		{ 477, 391, 50 },
		{ 477, 392, 50 },
		{ 477, 393, 50 },
		{ 477, 394, 50 },
		{ 478, 340, 50 },
		{ 478, 341, 50 },
		{ 478, 342, 50 },
		{ 478, 343, 50 },
		{ 478, 344, 50 },
		{ 478, 345, 50 },
		{ 478, 346, 50 },
		{ 478, 347, 50 },
		{ 478, 348, 50 },
		{ 478, 349, 50 },
		{ 478, 350, 50 },
		{ 478, 351, 50 },
		{ 478, 352, 50 },
		{ 478, 353, 50 },
		{ 478, 354, 50 },
		{ 478, 355, 50 },
		{ 478, 356, 50 },
		{ 478, 357, 50 },
		{ 478, 358, 50 },
		{ 478, 359, 50 },
		{ 478, 360, 50 },
		{ 478, 361, 50 },
		{ 478, 362, 50 },
		{ 478, 363, 50 },
		{ 478, 364, 50 },
		{ 478, 365, 50 },
		{ 478, 366, 50 },
		{ 478, 367, 50 },
		{ 478, 368, 50 },
		{ 478, 369, 50 },
		{ 478, 370, 50 },
		{ 478, 371, 50 },
		{ 478, 372, 50 },
		{ 478, 373, 50 },
		{ 478, 374, 50 },
		{ 478, 375, 50 },
		{ 478, 376, 50 },
		{ 478, 377, 50 },
		{ 478, 378, 50 },
		{ 478, 379, 50 },
		{ 478, 380, 50 },
		{ 478, 381, 50 },
		{ 478, 382, 50 },
		{ 478, 383, 50 },
		{ 478, 384, 50 },
		{ 478, 385, 50 },
		{ 478, 386, 50 },
		{ 478, 387, 50 },
		{ 478, 388, 50 },
		{ 478, 389, 50 },
		{ 478, 390, 50 },
		{ 478, 391, 50 },
		{ 478, 392, 50 },
		{ 478, 393, 50 },
		{ 478, 394, 50 },
		{ 478, 395, 50 },
		{ 479, 341, 50 },
		{ 479, 342, 50 },
		{ 479, 343, 50 },
		{ 479, 344, 50 },
		{ 479, 345, 50 },
		{ 479, 346, 50 },
		{ 479, 347, 50 },
		{ 479, 348, 50 },
		{ 479, 349, 50 },
		{ 479, 350, 50 },
		{ 479, 351, 50 },
		{ 479, 352, 50 },
		{ 479, 353, 50 },
		{ 479, 354, 50 },
		{ 479, 355, 50 },
		{ 479, 356, 50 },
		{ 479, 357, 50 },
		{ 479, 358, 50 },
		{ 479, 359, 50 },
		{ 479, 360, 50 },
		{ 479, 361, 50 },
		{ 479, 362, 50 },
		{ 479, 363, 50 },
		{ 479, 364, 50 },
		{ 479, 365, 50 },
		{ 479, 366, 50 },
		{ 479, 367, 50 },
		{ 479, 368, 50 },
		{ 479, 369, 50 },
		{ 479, 370, 50 },
		{ 479, 371, 50 },
		{ 479, 372, 50 },
		{ 479, 373, 50 },
		{ 479, 374, 50 },
		{ 479, 375, 50 },
		{ 479, 376, 50 },
		{ 479, 377, 50 },
		{ 479, 378, 50 },
		{ 479, 379, 50 },
		{ 479, 380, 50 },
		{ 479, 381, 50 },
		{ 479, 382, 50 },
		{ 479, 383, 50 },
		{ 479, 384, 50 },
		{ 479, 385, 50 },
		{ 479, 386, 50 },
		{ 479, 387, 50 },
		{ 479, 388, 50 },
		{ 479, 389, 50 },
		{ 479, 390, 50 },
		{ 479, 391, 50 },
		{ 479, 392, 50 },
		{ 479, 393, 50 },
		{ 479, 394, 50 },
		{ 479, 395, 50 },
		{ 480, 341, 50 },
		{ 480, 342, 50 },
		{ 480, 343, 50 },
		{ 480, 344, 50 },
		{ 480, 345, 50 },
		{ 480, 346, 50 },
		{ 480, 347, 50 },
		{ 480, 348, 50 },
		{ 480, 349, 50 },
		{ 480, 350, 50 },
		{ 480, 351, 50 },
		{ 480, 352, 50 },
		{ 480, 353, 50 },
		{ 480, 354, 50 },
		{ 480, 355, 50 },
		{ 480, 356, 50 },
		{ 480, 357, 50 },
		{ 480, 358, 50 },
		{ 480, 359, 50 },
		{ 480, 360, 50 },
		{ 480, 361, 50 },
		{ 480, 362, 50 },
		{ 480, 363, 50 },
		{ 480, 364, 50 },
		{ 480, 365, 50 },
		{ 480, 366, 50 },
		{ 480, 367, 50 },
		{ 480, 368, 50 },
		{ 480, 369, 50 },
		{ 480, 370, 50 },
		{ 480, 371, 50 },
		{ 480, 372, 50 },
		{ 480, 373, 50 },
		{ 480, 374, 50 },
		{ 480, 375, 50 },
		{ 480, 376, 50 },
		{ 480, 377, 50 },
		{ 480, 378, 50 },
		{ 480, 379, 50 },
		{ 480, 380, 50 },
		{ 480, 381, 50 },
		{ 480, 382, 50 },
		{ 480, 383, 50 },
		{ 480, 384, 50 },
		{ 480, 385, 50 },
		{ 480, 386, 50 },
		{ 480, 387, 50 },
		{ 480, 388, 50 },
		{ 480, 389, 50 },
		{ 480, 390, 50 },
		{ 480, 391, 50 },
		{ 480, 392, 50 },
		{ 480, 393, 50 },
		{ 480, 394, 50 },
		{ 480, 395, 50 },
		{ 480, 396, 50 },
		{ 481, 342, 50 },
		{ 481, 343, 50 },
		{ 481, 344, 50 },
		{ 481, 345, 50 },
		{ 481, 346, 50 },
		{ 481, 347, 50 },
		{ 481, 348, 50 },
		{ 481, 349, 50 },
		{ 481, 350, 50 },
		{ 481, 351, 50 },
		{ 481, 352, 50 },
		{ 481, 353, 50 },
		{ 481, 354, 50 },
		{ 481, 355, 50 },
		{ 481, 356, 50 },
		{ 481, 357, 50 },
		{ 481, 358, 50 },
		{ 481, 359, 50 },
		{ 481, 360, 50 },
		{ 481, 361, 50 },
		{ 481, 362, 50 },
		{ 481, 363, 50 },
		{ 481, 364, 50 },
		{ 481, 365, 50 },
		{ 481, 366, 50 },
		{ 481, 367, 50 },
		{ 481, 368, 50 },
		{ 481, 369, 50 },
		{ 481, 370, 50 },
		{ 481, 371, 50 },
		{ 481, 372, 50 },
		{ 481, 373, 50 },
		{ 481, 374, 50 },
		{ 481, 375, 50 },
		{ 481, 376, 50 },
		{ 481, 377, 50 },
		{ 481, 378, 50 },
		{ 481, 379, 50 },
		{ 481, 380, 50 },
		{ 481, 381, 50 },
		{ 481, 382, 50 },
		{ 481, 383, 50 },
		{ 481, 384, 50 },
		{ 481, 385, 50 },
		{ 481, 386, 50 },
		{ 481, 387, 50 },
		{ 481, 388, 50 },
		{ 481, 389, 50 },
		{ 481, 390, 50 },
		{ 481, 391, 50 },
		{ 481, 392, 50 },
		{ 481, 393, 50 },
		{ 481, 394, 50 },
		{ 481, 395, 50 },
		{ 481, 396, 50 },
		{ 482, 342, 50 },
		{ 482, 343, 50 },
		{ 482, 344, 50 },
		{ 482, 345, 50 },
		{ 482, 346, 50 },
		{ 482, 347, 50 },
		{ 482, 348, 50 },
		{ 482, 349, 50 },
		{ 482, 350, 50 },
		{ 482, 351, 50 },
		{ 482, 352, 50 },
		{ 482, 353, 50 },
		{ 482, 354, 50 },
		{ 482, 355, 50 },
		{ 482, 356, 50 },
		{ 482, 357, 50 },
		{ 482, 358, 50 },
		{ 482, 359, 50 },
		{ 482, 360, 50 },
		{ 482, 361, 50 },
		{ 482, 362, 50 },
		{ 482, 363, 50 },
		{ 482, 364, 50 },
		{ 482, 365, 50 },
		{ 482, 366, 50 },
		{ 482, 367, 50 },
		{ 482, 368, 50 },
		{ 482, 369, 50 },
		{ 482, 370, 50 },
		{ 482, 371, 50 },
		{ 482, 372, 50 },
		{ 482, 373, 50 },
		{ 482, 374, 50 },
		{ 482, 375, 50 },
		{ 482, 376, 50 },
		{ 482, 377, 50 },
		{ 482, 378, 50 },
		{ 482, 379, 50 },
		{ 482, 380, 50 },
		{ 482, 381, 50 },
		{ 482, 382, 50 },
		{ 482, 383, 50 },
		{ 482, 384, 50 },
		{ 482, 385, 50 },
		{ 482, 386, 50 },
		{ 482, 387, 50 },
		{ 482, 388, 50 },
		{ 482, 389, 50 },
		{ 482, 390, 50 },
		{ 482, 391, 50 },
		{ 482, 392, 50 },
		{ 482, 393, 50 },
		{ 482, 394, 50 },
		{ 482, 395, 50 },
		{ 482, 396, 50 },
		{ 482, 397, 50 },
		{ 483, 342, 50 },
		{ 483, 343, 50 },
		{ 483, 344, 50 },
		{ 483, 345, 50 },
		{ 483, 346, 50 },
		{ 483, 347, 50 },
		{ 483, 348, 50 },
		{ 483, 349, 50 },
		{ 483, 350, 50 },
		{ 483, 351, 50 },
		{ 483, 352, 50 },
		{ 483, 353, 50 },
		{ 483, 354, 50 },
		{ 483, 355, 50 },
		{ 483, 356, 50 },
		{ 483, 357, 50 },
		{ 483, 358, 50 },
		{ 483, 359, 50 },
		{ 483, 360, 50 },
		{ 483, 361, 50 },
		{ 483, 362, 50 },
		{ 483, 363, 50 },
		{ 483, 364, 50 },
		{ 483, 365, 50 },
		{ 483, 366, 50 },
		{ 483, 367, 50 },
		{ 483, 368, 50 },
		{ 483, 369, 50 },
		{ 483, 370, 50 },
		{ 483, 371, 50 },
		{ 483, 372, 50 },
		{ 483, 373, 50 },
		{ 483, 374, 50 },
		{ 483, 375, 50 },
		{ 483, 376, 50 },
		{ 483, 377, 50 },
		{ 483, 378, 50 },
		{ 483, 379, 50 },
		{ 483, 380, 50 },
		{ 483, 381, 50 },
		{ 483, 382, 50 },
		{ 483, 383, 50 },
		{ 483, 384, 50 },
		{ 483, 385, 50 },
		{ 483, 386, 50 },
		{ 483, 387, 50 },
		{ 483, 388, 50 },
		{ 483, 389, 50 },
		{ 483, 390, 50 },
		{ 483, 391, 50 },
		{ 483, 392, 50 },
		{ 483, 393, 50 },
		{ 483, 394, 50 },
		{ 483, 395, 50 },
		{ 483, 396, 50 },
		{ 483, 397, 50 },
		{ 483, 398, 50 },
		{ 484, 343, 50 },
		{ 484, 344, 50 },
		{ 484, 345, 50 },
		{ 484, 346, 50 },
		{ 484, 347, 50 },
		{ 484, 348, 50 },
		{ 484, 349, 50 },
		{ 484, 350, 50 },
		{ 484, 351, 50 },
		{ 484, 352, 50 },
		{ 484, 353, 50 },
		{ 484, 354, 50 },
		{ 484, 355, 50 },
		{ 484, 356, 50 },
		{ 484, 357, 50 },
		{ 484, 358, 50 },
		{ 484, 359, 50 },
		{ 484, 360, 50 },
		{ 484, 361, 50 },
		{ 484, 362, 50 },
		{ 484, 363, 50 },
		{ 484, 364, 50 },
		{ 484, 365, 50 },
		{ 484, 366, 50 },
		{ 484, 367, 50 },
		{ 484, 368, 50 },
		{ 484, 369, 50 },
		{ 484, 370, 50 },
		{ 484, 371, 50 },
		{ 484, 372, 50 },
		{ 484, 373, 50 },
		{ 484, 374, 50 },
		{ 484, 375, 50 },
		{ 484, 376, 50 },
		{ 484, 377, 50 },
		{ 484, 378, 50 },
		{ 484, 379, 50 },
		{ 484, 380, 50 },
		{ 484, 381, 50 },
		{ 484, 382, 50 },
		{ 484, 383, 50 },
		{ 484, 384, 50 },
		{ 484, 385, 50 },
		{ 484, 386, 50 },
		{ 484, 387, 50 },
		{ 484, 388, 50 },
		{ 484, 389, 50 },
		{ 484, 390, 50 },
		{ 484, 391, 50 },
		{ 484, 392, 50 },
		{ 484, 393, 50 },
		{ 484, 394, 50 },
		{ 484, 395, 50 },
		{ 484, 396, 50 },
		{ 484, 397, 50 },
		{ 484, 398, 50 },
		{ 485, 343, 50 },
		{ 485, 344, 50 },
		{ 485, 345, 50 },
		{ 485, 346, 50 },
		{ 485, 347, 50 },
		{ 485, 348, 50 },
		{ 485, 349, 50 },
		{ 485, 350, 50 },
		{ 485, 351, 50 },
		{ 485, 352, 50 },
		{ 485, 353, 50 },
		{ 485, 354, 50 },
		{ 485, 355, 50 },
		{ 485, 356, 50 },
		{ 485, 357, 50 },
		{ 485, 358, 50 },
		{ 485, 359, 50 },
		{ 485, 360, 50 },
		{ 485, 361, 50 },
		{ 485, 362, 50 },
		{ 485, 363, 50 },
		{ 485, 364, 50 },
		{ 485, 365, 50 },
		{ 485, 366, 50 },
		{ 485, 367, 50 },
		{ 485, 368, 50 },
		{ 485, 369, 50 },
		{ 485, 370, 50 },
		{ 485, 371, 50 },
		{ 485, 372, 50 },
		{ 485, 373, 50 },
		{ 485, 374, 50 },
		{ 485, 375, 50 },
		{ 485, 376, 50 },
		{ 485, 377, 50 },
		{ 485, 378, 50 },
		{ 485, 379, 50 },
		{ 485, 380, 50 },
		{ 485, 381, 50 },
		{ 485, 382, 50 },
		{ 485, 383, 50 },
		{ 485, 384, 50 },
		{ 485, 385, 50 },
		{ 485, 386, 50 },
		{ 485, 387, 50 },
		{ 485, 388, 50 },
		{ 485, 389, 50 },
		{ 485, 390, 50 },
		{ 485, 391, 50 },
		{ 485, 392, 50 },
		{ 485, 393, 50 },
		{ 485, 394, 50 },
		{ 485, 395, 50 },
		{ 485, 396, 50 },
		{ 485, 397, 50 },
		{ 485, 398, 50 },
		{ 485, 399, 50 },
		{ 486, 343, 50 },
		{ 486, 344, 50 },
		{ 486, 345, 50 },
		{ 486, 346, 50 },
		{ 486, 347, 50 },
		{ 486, 348, 50 },
		{ 486, 349, 50 },
		{ 486, 350, 50 },
		{ 486, 351, 50 },
		{ 486, 352, 50 },
		{ 486, 353, 50 },
		{ 486, 354, 50 },
		{ 486, 355, 50 },
		{ 486, 356, 50 },
		{ 486, 357, 50 },
		{ 486, 358, 50 },
		{ 486, 359, 50 },
		{ 486, 360, 50 },
		{ 486, 361, 50 },
		{ 486, 362, 50 },
		{ 486, 363, 50 },
		{ 486, 364, 50 },
		{ 486, 365, 50 },
		{ 486, 366, 50 },
		{ 486, 367, 50 },
		{ 486, 368, 50 },
		{ 486, 369, 50 },
		{ 486, 370, 50 },
		{ 486, 371, 50 },
		{ 486, 372, 50 },
		{ 486, 373, 50 },
		{ 486, 374, 50 },
		{ 486, 375, 50 },
		{ 486, 376, 50 },
		{ 486, 377, 50 },
		{ 486, 378, 50 },
		{ 486, 379, 50 },
		{ 486, 380, 50 },
		{ 486, 381, 50 },
		{ 486, 382, 50 },
		{ 486, 383, 50 },
		{ 486, 384, 50 },
		{ 486, 385, 50 },
		{ 486, 386, 50 },
		{ 486, 387, 50 },
		{ 486, 388, 50 },
		{ 486, 389, 50 },
		{ 486, 390, 50 },
		{ 486, 391, 50 },
		{ 486, 392, 50 },
		{ 486, 393, 50 },
		{ 486, 394, 50 },
		{ 486, 395, 50 },
		{ 486, 396, 50 },
		{ 486, 397, 50 },
		{ 486, 398, 50 },
		{ 486, 399, 50 },
		{ 486, 400, 50 },
		{ 487, 344, 50 },
		{ 487, 345, 50 },
		{ 487, 346, 50 },
		{ 487, 347, 50 },
		{ 487, 348, 50 },
		{ 487, 349, 50 },
		{ 487, 350, 50 },
		{ 487, 351, 50 },
		{ 487, 352, 50 },
		{ 487, 353, 50 },
		{ 487, 354, 50 },
		{ 487, 355, 50 },
		{ 487, 356, 50 },
		{ 487, 357, 50 },
		{ 487, 358, 50 },
		{ 487, 359, 50 },
		{ 487, 360, 50 },
		{ 487, 361, 50 },
		{ 487, 362, 50 },
		{ 487, 363, 50 },
		{ 487, 364, 50 },
		{ 487, 365, 50 },
		{ 487, 366, 50 },
		{ 487, 367, 50 },
		{ 487, 368, 50 },
		{ 487, 369, 50 },
		{ 487, 370, 50 },
		{ 487, 371, 50 },
		{ 487, 372, 50 },
		{ 487, 373, 50 },
		{ 487, 374, 50 },
		{ 487, 375, 50 },
		{ 487, 376, 50 },
		{ 487, 377, 50 },
		{ 487, 378, 50 },
		{ 487, 379, 50 },
		{ 487, 380, 50 },
		{ 487, 381, 50 },
		{ 487, 382, 50 },
		{ 487, 383, 50 },
		{ 487, 384, 50 },
		{ 487, 385, 50 },
		{ 487, 386, 50 },
		{ 487, 387, 50 },
		{ 487, 388, 50 },
		{ 487, 389, 50 },
		{ 487, 390, 50 },
		{ 487, 391, 50 },
		{ 487, 392, 50 },
		{ 487, 393, 50 },
		{ 487, 394, 50 },
		{ 487, 395, 50 },
		{ 487, 396, 50 },
		{ 487, 397, 50 },
		{ 487, 398, 50 },
		{ 487, 399, 50 },
		{ 487, 400, 50 },
		{ 488, 344, 50 },
		{ 488, 345, 50 },
		{ 488, 346, 50 },
		{ 488, 347, 50 },
		{ 488, 348, 50 },
		{ 488, 349, 50 },
		{ 488, 350, 50 },
		{ 488, 351, 50 },
		{ 488, 352, 50 },
		{ 488, 353, 50 },
		{ 488, 354, 50 },
		{ 488, 355, 50 },
		{ 488, 356, 50 },
		{ 488, 357, 50 },
		{ 488, 358, 50 },
		{ 488, 359, 50 },
		{ 488, 360, 50 },
		{ 488, 361, 50 },
		{ 488, 362, 50 },
		{ 488, 363, 50 },
		{ 488, 364, 50 },
		{ 488, 365, 50 },
		{ 488, 366, 50 },
		{ 488, 367, 50 },
		{ 488, 368, 50 },
		{ 488, 369, 50 },
		{ 488, 370, 50 },
		{ 488, 371, 50 },
		{ 488, 372, 50 },
		{ 488, 373, 50 },
		{ 488, 374, 50 },
		{ 488, 375, 50 },
		{ 488, 376, 50 },
		{ 488, 377, 50 },
		{ 488, 378, 50 },
		{ 488, 379, 50 },
		{ 488, 380, 50 },
		{ 488, 381, 50 },
		{ 488, 382, 50 },
		{ 488, 383, 50 },
		{ 488, 384, 50 },
		{ 488, 385, 50 },
		{ 488, 386, 50 },
		{ 488, 387, 50 },
		{ 488, 388, 50 },
		{ 488, 389, 50 },
		{ 488, 390, 50 },
		{ 488, 391, 50 },
		{ 488, 392, 50 },
		{ 488, 393, 50 },
		{ 488, 394, 50 },
		{ 488, 395, 50 },
		{ 488, 396, 50 },
		{ 488, 397, 50 },
		{ 488, 398, 50 },
		{ 488, 399, 50 },
		{ 488, 400, 50 },
		{ 488, 401, 50 },
		{ 489, 345, 50 },
		{ 489, 346, 50 },
		{ 489, 347, 50 },
		{ 489, 348, 50 },
		{ 489, 349, 50 },
		{ 489, 350, 50 },
		{ 489, 351, 50 },
		{ 489, 352, 50 },
		{ 489, 353, 50 },
		{ 489, 354, 50 },
		{ 489, 355, 50 },
		{ 489, 356, 50 },
		{ 489, 357, 50 },
		{ 489, 358, 50 },
		{ 489, 359, 50 },
		{ 489, 360, 50 },
		{ 489, 361, 50 },
		{ 489, 362, 50 },
		{ 489, 363, 50 },
		{ 489, 364, 50 },
		{ 489, 365, 50 },
		{ 489, 366, 50 },
		{ 489, 367, 50 },
		{ 489, 368, 50 },
		{ 489, 369, 50 },
		{ 489, 370, 50 },
		{ 489, 371, 50 },
		{ 489, 372, 50 },
		{ 489, 373, 50 },
		{ 489, 374, 50 },
		{ 489, 375, 50 },
		{ 489, 376, 50 },
		{ 489, 377, 50 },
		{ 489, 378, 50 },
		{ 489, 379, 50 },
		{ 489, 380, 50 },
		{ 489, 381, 50 },
		{ 489, 382, 50 },
		{ 489, 383, 50 },
		{ 489, 384, 50 },
		{ 489, 385, 50 },
		{ 489, 386, 50 },
		{ 489, 387, 50 },
		{ 489, 388, 50 },
		{ 489, 389, 50 },
		{ 489, 390, 50 },
		{ 489, 391, 50 },
		{ 489, 392, 50 },
		{ 489, 393, 50 },
		{ 489, 394, 50 },
		{ 489, 395, 50 },
		{ 489, 396, 50 },
		{ 489, 397, 50 },
		{ 489, 398, 50 },
		{ 489, 399, 50 },
		{ 489, 400, 50 },
		{ 489, 401, 50 },
		{ 490, 345, 50 },
		{ 490, 346, 50 },
		{ 490, 347, 50 },
		{ 490, 348, 50 },
		{ 490, 349, 50 },
		{ 490, 350, 50 },
		{ 490, 351, 50 },
		{ 490, 352, 50 },
		{ 490, 353, 50 },
		{ 490, 354, 50 },
		{ 490, 355, 50 },
		{ 490, 356, 50 },
		{ 490, 357, 50 },
		{ 490, 358, 50 },
		{ 490, 359, 50 },
		{ 490, 360, 50 },
		{ 490, 361, 50 },
		{ 490, 362, 50 },
		{ 490, 363, 50 },
		{ 490, 364, 50 },
		{ 490, 365, 50 },
		{ 490, 366, 50 },
		{ 490, 367, 50 },
		{ 490, 368, 50 },
		{ 490, 369, 50 },
		{ 490, 370, 50 },
		{ 490, 371, 50 },
		{ 490, 372, 50 },
		{ 490, 373, 50 },
		{ 490, 374, 50 },
		{ 490, 375, 50 },
		{ 490, 376, 50 },
		{ 490, 377, 50 },
		{ 490, 378, 50 },
		{ 490, 379, 50 },
		{ 490, 380, 50 },
		{ 490, 381, 50 },
		{ 490, 382, 50 },
		{ 490, 383, 50 },
		{ 490, 384, 50 },
		{ 490, 385, 50 },
		{ 490, 386, 50 },
		{ 490, 387, 50 },
		{ 490, 388, 50 },
		{ 490, 389, 50 },
		{ 490, 390, 50 },
		{ 490, 391, 50 },
		{ 490, 392, 50 },
		{ 490, 393, 50 },
		{ 490, 394, 50 },
		{ 490, 395, 50 },
		{ 490, 396, 50 },
		{ 490, 397, 50 },
		{ 490, 398, 50 },
		{ 490, 399, 50 },
		{ 490, 400, 50 },
		{ 490, 401, 50 },
		{ 490, 402, 50 },
		{ 491, 345, 50 },
		{ 491, 346, 50 },
		{ 491, 347, 50 },
		{ 491, 348, 50 },
		{ 491, 349, 50 },
		{ 491, 350, 50 },
		{ 491, 351, 50 },
		{ 491, 352, 50 },
		{ 491, 353, 50 },
		{ 491, 354, 50 },
		{ 491, 355, 50 },
		{ 491, 356, 50 },
		{ 491, 357, 50 },
		{ 491, 358, 50 },
		{ 491, 359, 50 },
		{ 491, 360, 50 },
		{ 491, 361, 50 },
		{ 491, 362, 50 },
		{ 491, 363, 50 },
		{ 491, 364, 50 },
		{ 491, 365, 50 },
		{ 491, 366, 50 },
		{ 491, 367, 50 },
		{ 491, 368, 50 },
		{ 491, 369, 50 },
		{ 491, 370, 50 },
		{ 491, 371, 50 },
		{ 491, 372, 50 },
		{ 491, 373, 50 },
		{ 491, 374, 50 },
		{ 491, 375, 50 },
		{ 491, 376, 50 },
		{ 491, 377, 50 },
		{ 491, 378, 50 },
		{ 491, 379, 50 },
		{ 491, 380, 50 },
		{ 491, 381, 50 },
		{ 491, 382, 50 },
		{ 491, 383, 50 },
		{ 491, 384, 50 },
		{ 491, 385, 50 },
		{ 491, 386, 50 },
		{ 491, 387, 50 },
		{ 491, 388, 50 },
		{ 491, 389, 50 },
		{ 491, 390, 50 },
		{ 491, 391, 50 },
		{ 491, 392, 50 },
		{ 491, 393, 50 },
		{ 491, 394, 50 },
		{ 491, 395, 50 },
		{ 491, 396, 50 },
		{ 491, 397, 50 },
		{ 491, 398, 50 },
		{ 491, 399, 50 },
		{ 491, 400, 50 },
		{ 491, 401, 50 },
		{ 491, 402, 50 },
		{ 491, 403, 50 },
		{ 492, 346, 50 },
		{ 492, 347, 50 },
		{ 492, 348, 50 },
		{ 492, 349, 50 },
		{ 492, 350, 50 },
		{ 492, 351, 50 },
		{ 492, 352, 50 },
		{ 492, 353, 50 },
		{ 492, 354, 50 },
		{ 492, 355, 50 },
		{ 492, 356, 50 },
		{ 492, 357, 50 },
		{ 492, 358, 50 },
		{ 492, 359, 50 },
		{ 492, 360, 50 },
		{ 492, 361, 50 },
		{ 492, 362, 50 },
		{ 492, 363, 50 },
		{ 492, 364, 50 },
		{ 492, 365, 50 },
		{ 492, 366, 50 },
		{ 492, 367, 50 },
		{ 492, 368, 50 },
		{ 492, 369, 50 },
		{ 492, 370, 50 },
		{ 492, 371, 50 },
		{ 492, 372, 50 },
		{ 492, 373, 50 },
		{ 492, 374, 50 },
		{ 492, 375, 50 },
		{ 492, 376, 50 },
		{ 492, 377, 50 },
		{ 492, 378, 50 },
		{ 492, 379, 50 },
		{ 492, 380, 50 },
		{ 492, 381, 50 },
		{ 492, 382, 50 },
		{ 492, 383, 50 },
		{ 492, 384, 50 },
		{ 492, 385, 50 },
		{ 492, 386, 50 },
		{ 492, 387, 50 },
		{ 492, 388, 50 },
		{ 492, 389, 50 },
		{ 492, 390, 50 },
		{ 492, 391, 50 },
		{ 492, 392, 50 },
		{ 492, 393, 50 },
		{ 492, 394, 50 },
		{ 492, 395, 50 },
		{ 492, 396, 50 },
		{ 492, 397, 50 },
		{ 492, 398, 50 },
		{ 492, 399, 50 },
		{ 492, 400, 50 },
		{ 492, 401, 50 },
		{ 492, 402, 50 },
		{ 492, 403, 50 },
		{ 493, 346, 50 },
		{ 493, 347, 50 },
		{ 493, 348, 50 },
		{ 493, 349, 50 },
		{ 493, 350, 50 },
		{ 493, 351, 50 },
		{ 493, 352, 50 },
		{ 493, 353, 50 },
		{ 493, 354, 50 },
		{ 493, 355, 50 },
		{ 493, 356, 50 },
		{ 493, 357, 50 },
		{ 493, 358, 50 },
		{ 493, 359, 50 },
		{ 493, 360, 50 },
		{ 493, 361, 50 },
		{ 493, 362, 50 },
		{ 493, 363, 50 },
		{ 493, 364, 50 },
		{ 493, 365, 50 },
		{ 493, 366, 50 },
		{ 493, 367, 50 },
		{ 493, 368, 50 },
		{ 493, 369, 50 },
		{ 493, 370, 50 },
		{ 493, 371, 50 },
		{ 493, 372, 50 },
		{ 493, 373, 50 },
		{ 493, 374, 50 },
		{ 493, 375, 50 },
		{ 493, 376, 50 },
		{ 493, 377, 50 },
		{ 493, 378, 50 },
		{ 493, 379, 50 },
		{ 493, 380, 50 },
		{ 493, 381, 50 },
		{ 493, 382, 50 },
		{ 493, 383, 50 },
		{ 493, 384, 50 },
		{ 493, 385, 50 },
		{ 493, 386, 50 },
		{ 493, 387, 50 },
		{ 493, 388, 50 },
		{ 493, 389, 50 },
		{ 493, 390, 50 },
		{ 493, 391, 50 },
		{ 493, 392, 50 },
		{ 493, 393, 50 },
		{ 493, 394, 50 },
		{ 493, 395, 50 },
		{ 493, 396, 50 },
		{ 493, 397, 50 },
		{ 493, 398, 50 },
		{ 493, 399, 50 },
		{ 493, 400, 50 },
		{ 493, 401, 50 },
		{ 493, 402, 50 },
		{ 493, 403, 50 },
		{ 493, 404, 50 },
		{ 494, 346, 50 },
		{ 494, 347, 50 },
		{ 494, 348, 50 },
		{ 494, 349, 50 },
		{ 494, 350, 50 },
		{ 494, 351, 50 },
		{ 494, 352, 50 },
		{ 494, 353, 50 },
		{ 494, 354, 50 },
		{ 494, 355, 50 },
		{ 494, 356, 50 },
		{ 494, 357, 50 },
		{ 494, 358, 50 },
		{ 494, 359, 50 },
		{ 494, 360, 50 },
		{ 494, 361, 50 },
		{ 494, 362, 50 },
		{ 494, 363, 50 },
		{ 494, 364, 50 },
		{ 494, 365, 50 },
		{ 494, 366, 50 },
		{ 494, 367, 50 },
		{ 494, 368, 50 },
		{ 494, 369, 50 },
		{ 494, 370, 50 },
		{ 494, 371, 50 },
		{ 494, 372, 50 },
		{ 494, 373, 50 },
		{ 494, 374, 50 },
		{ 494, 375, 50 },
		{ 494, 376, 50 },
		{ 494, 377, 50 },
		{ 494, 378, 50 },
		{ 494, 379, 50 },
		{ 494, 380, 50 },
		{ 494, 381, 50 },
		{ 494, 382, 50 },
		{ 494, 383, 50 },
		{ 494, 384, 50 },
		{ 494, 385, 50 },
		{ 494, 386, 50 },
		{ 494, 387, 50 },
		{ 494, 388, 50 },
		{ 494, 389, 50 },
		{ 494, 390, 50 },
		{ 494, 391, 50 },
		{ 494, 392, 50 },
		{ 494, 393, 50 },
		{ 494, 394, 50 },
		{ 494, 395, 50 },
		{ 494, 396, 50 },
		{ 494, 397, 50 },
		{ 494, 398, 50 },
		{ 494, 399, 50 },
		{ 494, 400, 50 },
		{ 494, 401, 50 },
		{ 494, 402, 50 },
		{ 494, 403, 50 },
		{ 494, 404, 50 },
		{ 494, 405, 50 },
		{ 495, 347, 50 },
		{ 495, 348, 50 },
		{ 495, 349, 50 },
		{ 495, 350, 50 },
		{ 495, 351, 50 },
		{ 495, 352, 50 },
		{ 495, 353, 50 },
		{ 495, 354, 50 },
		{ 495, 355, 50 },
		{ 495, 356, 50 },
		{ 495, 357, 50 },
		{ 495, 358, 50 },
		{ 495, 359, 50 },
		{ 495, 360, 50 },
		{ 495, 361, 50 },
		{ 495, 362, 50 },
		{ 495, 363, 50 },
		{ 495, 364, 50 },
		{ 495, 365, 50 },
		{ 495, 366, 50 },
		{ 495, 367, 50 },
		{ 495, 368, 50 },
		{ 495, 369, 50 },
		{ 495, 370, 50 },
		{ 495, 371, 50 },
		{ 495, 372, 50 },
		{ 495, 373, 50 },
		{ 495, 374, 50 },
		{ 495, 375, 50 },
		{ 495, 376, 50 },
		{ 495, 377, 50 },
		{ 495, 378, 50 },
		{ 495, 379, 50 },
		{ 495, 380, 50 },
		{ 495, 381, 50 },
		{ 495, 382, 50 },
		{ 495, 383, 50 },
		{ 495, 384, 50 },
		{ 495, 385, 50 },
		{ 495, 386, 50 },
		{ 495, 387, 50 },
		{ 495, 388, 50 },
		{ 495, 389, 50 },
		{ 495, 390, 50 },
		{ 495, 391, 50 },
		{ 495, 392, 50 },
		{ 495, 393, 50 },
		{ 495, 394, 50 },
		{ 495, 395, 50 },
		{ 495, 396, 50 },
		{ 495, 397, 50 },
		{ 495, 398, 50 },
		{ 495, 399, 50 },
		{ 495, 400, 50 },
		{ 495, 401, 50 },
		{ 495, 402, 50 },
		{ 495, 403, 50 },
		{ 495, 404, 50 },
		{ 495, 405, 50 },
		{ 496, 347, 50 },
		{ 496, 348, 50 },
		{ 496, 349, 50 },
		{ 496, 350, 50 },
		{ 496, 351, 50 },
		{ 496, 352, 50 },
		{ 496, 353, 50 },
		{ 496, 354, 50 },
		{ 496, 355, 50 },
		{ 496, 356, 50 },
		{ 496, 357, 50 },
		{ 496, 358, 50 },
		{ 496, 359, 50 },
		{ 496, 360, 50 },
		{ 496, 361, 50 },
		{ 496, 362, 50 },
		{ 496, 363, 50 },
		{ 496, 364, 50 },
		{ 496, 365, 50 },
		{ 496, 366, 50 },
		{ 496, 367, 50 },
		{ 496, 368, 50 },
		{ 496, 369, 50 },
		{ 496, 370, 50 },
		{ 496, 371, 50 },
		{ 496, 372, 50 },
		{ 496, 373, 50 },
		{ 496, 374, 50 },
		{ 496, 375, 50 },
		{ 496, 376, 50 },
		{ 496, 377, 50 },
		{ 496, 378, 50 },
		{ 496, 379, 50 },
		{ 496, 380, 50 },
		{ 496, 381, 50 },
		{ 496, 382, 50 },
		{ 496, 383, 50 },
		{ 496, 384, 50 },
		{ 496, 385, 50 },
		{ 496, 386, 50 },
		{ 496, 387, 50 },
		{ 496, 388, 50 },
		{ 496, 389, 50 },
		{ 496, 390, 50 },
		{ 496, 391, 50 },
		{ 496, 392, 50 },
		{ 496, 393, 50 },
		{ 496, 394, 50 },
		{ 496, 395, 50 },
		{ 496, 396, 50 },
		{ 496, 397, 50 },
		{ 496, 398, 50 },
		{ 496, 399, 50 },
		{ 496, 400, 50 },
		{ 496, 401, 50 },
		{ 496, 402, 50 },
		{ 496, 403, 50 },
		{ 496, 404, 50 },
		{ 496, 405, 50 },
		{ 496, 406, 50 },
		{ 497, 348, 50 },
		{ 497, 349, 50 },
		{ 497, 350, 50 },
		{ 497, 351, 50 },
		{ 497, 352, 50 },
		{ 497, 353, 50 },
		{ 497, 354, 50 },
		{ 497, 355, 50 },
		{ 497, 356, 50 },
		{ 497, 357, 50 },
		{ 497, 358, 50 },
		{ 497, 359, 50 },
		{ 497, 360, 50 },
		{ 497, 361, 50 },
		{ 497, 362, 50 },
		{ 497, 363, 50 },
		{ 497, 364, 50 },
		{ 497, 365, 50 },
		{ 497, 366, 50 },
		{ 497, 367, 50 },
		{ 497, 368, 50 },
		{ 497, 369, 50 },
		{ 497, 370, 50 },
		{ 497, 371, 50 },
		{ 497, 372, 50 },
		{ 497, 373, 50 },
		{ 497, 374, 50 },
		{ 497, 375, 50 },
		{ 497, 376, 50 },
		{ 497, 377, 50 },
		{ 497, 378, 50 },
		{ 497, 379, 50 },
		{ 497, 380, 50 },
		{ 497, 381, 50 },
		{ 497, 382, 50 },
		{ 497, 383, 50 },
		{ 497, 384, 50 },
		{ 497, 385, 50 },
		{ 497, 386, 50 },
		{ 497, 387, 50 },
		{ 497, 388, 50 },
		{ 497, 389, 50 },
		{ 497, 390, 50 },
		{ 497, 391, 50 },
		{ 497, 392, 50 },
		{ 497, 393, 50 },
		{ 497, 394, 50 },
		{ 497, 395, 50 },
		{ 497, 396, 50 },
		{ 497, 397, 50 },
		{ 497, 398, 50 },
		{ 497, 399, 50 },
		{ 497, 400, 50 },
		{ 497, 401, 50 },
		{ 497, 402, 50 },
		{ 497, 403, 50 },
		{ 497, 404, 50 },
		{ 497, 405, 50 },
		{ 497, 406, 50 },
		{ 498, 348, 50 },
		{ 498, 349, 50 },
		{ 498, 350, 50 },
		{ 498, 351, 50 },
		{ 498, 352, 50 },
		{ 498, 353, 50 },
		{ 498, 354, 50 },
		{ 498, 355, 50 },
		{ 498, 356, 50 },
		{ 498, 357, 50 },
		{ 498, 358, 50 },
		{ 498, 359, 50 },
		{ 498, 360, 50 },
		{ 498, 361, 50 },
		{ 498, 362, 50 },
		{ 498, 363, 50 },
		{ 498, 364, 50 },
		{ 498, 365, 50 },
		{ 498, 366, 50 },
		{ 498, 367, 50 },
		{ 498, 368, 50 },
		{ 498, 369, 50 },
		{ 498, 370, 50 },
		{ 498, 371, 50 },
		{ 498, 372, 50 },
		{ 498, 373, 50 },
		{ 498, 374, 50 },
		{ 498, 375, 50 },
		{ 498, 376, 50 },
		{ 498, 377, 50 },
		{ 498, 378, 50 },
		{ 498, 379, 50 },
		{ 498, 380, 50 },
		{ 498, 381, 50 },
		{ 498, 382, 50 },
		{ 498, 383, 50 },
		{ 498, 384, 50 },
		{ 498, 385, 50 },
		{ 498, 386, 50 },
		{ 498, 387, 50 },
		{ 498, 388, 50 },
		{ 498, 389, 50 },
		{ 498, 390, 50 },
		{ 498, 391, 50 },
		{ 498, 392, 50 },
		{ 498, 393, 50 },
		{ 498, 394, 50 },
		{ 498, 395, 50 },
		{ 498, 396, 50 },
		{ 498, 397, 50 },
		{ 498, 398, 50 },
		{ 498, 399, 50 },
		{ 498, 400, 50 },
		{ 498, 401, 50 },
		{ 498, 402, 50 },
		{ 498, 403, 50 },
		{ 498, 404, 50 },
		{ 498, 405, 50 },
		{ 498, 406, 50 },
		{ 498, 407, 50 },
		{ 499, 348, 50 },
		{ 499, 349, 50 },
		{ 499, 350, 50 },
		{ 499, 351, 50 },
		{ 499, 352, 50 },
		{ 499, 353, 50 },
		{ 499, 354, 50 },
		{ 499, 355, 50 },
		{ 499, 356, 50 },
		{ 499, 357, 50 },
		{ 499, 358, 50 },
		{ 499, 359, 50 },
		{ 499, 360, 50 },
		{ 499, 361, 50 },
		{ 499, 362, 50 },
		{ 499, 363, 50 },
		{ 499, 364, 50 },
		{ 499, 365, 50 },
		{ 499, 366, 50 },
		{ 499, 367, 50 },
		{ 499, 368, 50 },
		{ 499, 369, 50 },
		{ 499, 370, 50 },
		{ 499, 371, 50 },
		{ 499, 372, 50 },
		{ 499, 373, 50 },
		{ 499, 374, 50 },
		{ 499, 375, 50 },
		{ 499, 376, 50 },
		{ 499, 377, 50 },
		{ 499, 378, 50 },
		{ 499, 379, 50 },
		{ 499, 380, 50 },
		{ 499, 381, 50 },
		{ 499, 382, 50 },
		{ 499, 383, 50 },
		{ 499, 384, 50 },
		{ 499, 385, 50 },
		{ 499, 386, 50 },
		{ 499, 387, 50 },
		{ 499, 388, 50 },
		{ 499, 389, 50 },
		{ 499, 390, 50 },
		{ 499, 391, 50 },
		{ 499, 392, 50 },
		{ 499, 393, 50 },
		{ 499, 394, 50 },
		{ 499, 395, 50 },
		{ 499, 396, 50 },
		{ 499, 397, 50 },
		{ 499, 398, 50 },
		{ 499, 399, 50 },
		{ 499, 400, 50 },
		{ 499, 401, 50 },
		{ 499, 402, 50 },
		{ 499, 403, 50 },
		{ 499, 404, 50 },
		{ 499, 405, 50 },
		{ 499, 406, 50 },
		{ 499, 407, 50 },
		{ 499, 408, 50 },
		{ 500, 349, 50 },
		{ 500, 350, 50 },
		{ 500, 351, 50 },
		{ 500, 352, 50 },
		{ 500, 353, 50 },
		{ 500, 354, 50 },
		{ 500, 355, 50 },
		{ 500, 356, 50 },
		{ 500, 357, 50 },
		{ 500, 358, 50 },
		{ 500, 359, 50 },
		{ 500, 360, 50 },
		{ 500, 361, 50 },
		{ 500, 362, 50 },
		{ 500, 363, 50 },
		{ 500, 364, 50 },
		{ 500, 365, 50 },
		{ 500, 366, 50 },
		{ 500, 367, 50 },
		{ 500, 368, 50 },
		{ 500, 369, 50 },
		{ 500, 370, 50 },
		{ 500, 371, 50 },
		{ 500, 372, 50 },
		{ 500, 373, 50 },
		{ 500, 374, 50 },
		{ 500, 375, 50 },
		{ 500, 376, 50 },
		{ 500, 377, 50 },
		{ 500, 378, 50 },
		{ 500, 379, 50 },
		{ 500, 380, 50 },
		{ 500, 381, 50 },
		{ 500, 382, 50 },
		{ 500, 383, 50 },
		{ 500, 384, 50 },
		{ 500, 385, 50 },
		{ 500, 386, 50 },
		{ 500, 387, 50 },
		{ 500, 388, 50 },
		{ 500, 389, 50 },
		{ 500, 390, 50 },
		{ 500, 391, 50 },
		{ 500, 392, 50 },
		{ 500, 393, 50 },
		{ 500, 394, 50 },
		{ 500, 395, 50 },
		{ 500, 396, 50 },
		{ 500, 397, 50 },
		{ 500, 398, 50 },
		{ 500, 399, 50 },
		{ 500, 400, 50 },
		{ 500, 401, 50 },
		{ 500, 402, 50 },
		{ 500, 403, 50 },
		{ 500, 404, 50 },
		{ 500, 405, 50 },
		{ 500, 406, 50 },
		{ 500, 407, 50 },
		{ 500, 408, 50 },
		{ 501, 349, 50 },
		{ 501, 350, 50 },
		{ 501, 351, 50 },
		{ 501, 352, 50 },
		{ 501, 353, 50 },
		{ 501, 354, 50 },
		{ 501, 355, 50 },
		{ 501, 356, 50 },
		{ 501, 357, 50 },
		{ 501, 358, 50 },
		{ 501, 359, 50 },
		{ 501, 360, 50 },
		{ 501, 361, 50 },
		{ 501, 362, 50 },
		{ 501, 363, 50 },
		{ 501, 364, 50 },
		{ 501, 365, 50 },
		{ 501, 366, 50 },
		{ 501, 367, 50 },
		{ 501, 368, 50 },
		{ 501, 369, 50 },
		{ 501, 370, 50 },
		{ 501, 371, 50 },
		{ 501, 372, 50 },
		{ 501, 373, 50 },
		{ 501, 374, 50 },
		{ 501, 375, 50 },
		{ 501, 376, 50 },
		{ 501, 377, 50 },
		{ 501, 378, 50 },
		{ 501, 379, 50 },
		{ 501, 380, 50 },
		{ 501, 381, 50 },
		{ 501, 382, 50 },
		{ 501, 383, 50 },
		{ 501, 384, 50 },
		{ 501, 385, 50 },
		{ 501, 386, 50 },
		{ 501, 387, 50 },
		{ 501, 388, 50 },
		{ 501, 389, 50 },
		{ 501, 390, 50 },
		{ 501, 391, 50 },
		{ 501, 392, 50 },
		{ 501, 393, 50 },
		{ 501, 394, 50 },
		{ 501, 395, 50 },
		{ 501, 396, 50 },
		{ 501, 397, 50 },
		{ 501, 398, 50 },
		{ 501, 399, 50 },
		{ 501, 400, 50 },
		{ 501, 401, 50 },
		{ 501, 402, 50 },
		{ 501, 403, 50 },
		{ 501, 404, 50 },
		{ 501, 405, 50 },
		{ 501, 406, 50 },
		{ 501, 407, 50 },
		{ 501, 408, 50 },
		{ 501, 409, 50 },
		{ 502, 349, 50 },
		{ 502, 350, 50 },
		{ 502, 351, 50 },
		{ 502, 352, 50 },
		{ 502, 353, 50 },
		{ 502, 354, 50 },
		{ 502, 355, 50 },
		{ 502, 356, 50 },
		{ 502, 357, 50 },
		{ 502, 358, 50 },
		{ 502, 359, 50 },
		{ 502, 360, 50 },
		{ 502, 361, 50 },
		{ 502, 362, 50 },
		{ 502, 363, 50 },
		{ 502, 364, 50 },
		{ 502, 365, 50 },
		{ 502, 366, 50 },
		{ 502, 367, 50 },
		{ 502, 368, 50 },
		{ 502, 369, 50 },
		{ 502, 370, 50 },
		{ 502, 371, 50 },
		{ 502, 372, 50 },
		{ 502, 373, 50 },
		{ 502, 374, 50 },
		{ 502, 375, 50 },
		{ 502, 376, 50 },
		{ 502, 377, 50 },
		{ 502, 378, 50 },
		{ 502, 379, 50 },
		{ 502, 380, 50 },
		{ 502, 381, 50 },
		{ 502, 382, 50 },
		{ 502, 383, 50 },
		{ 502, 384, 50 },
		{ 502, 385, 50 },
		{ 502, 386, 50 },
		{ 502, 387, 50 },
		{ 502, 388, 50 },
		{ 502, 389, 50 },
		{ 502, 390, 50 },
		{ 502, 391, 50 },
		{ 502, 392, 50 },
		{ 502, 393, 50 },
		{ 502, 394, 50 },
		{ 502, 395, 50 },
		{ 502, 396, 50 },
		{ 502, 397, 50 },
		{ 502, 398, 50 },
		{ 502, 399, 50 },
		{ 502, 400, 50 },
		{ 502, 401, 50 },
		{ 502, 402, 50 },
		{ 502, 403, 50 },
		{ 502, 404, 50 },
		{ 502, 405, 50 },
		{ 502, 406, 50 },
		{ 502, 407, 50 },
		{ 502, 408, 50 },
		{ 502, 409, 50 },
		{ 502, 410, 50 },
		{ 503, 350, 50 },
		{ 503, 351, 50 },
		{ 503, 352, 50 },
		{ 503, 353, 50 },
		{ 503, 354, 50 },
		{ 503, 355, 50 },
		{ 503, 356, 50 },
		{ 503, 357, 50 },
		{ 503, 358, 50 },
		{ 503, 359, 50 },
		{ 503, 360, 50 },
		{ 503, 361, 50 },
		{ 503, 362, 50 },
		{ 503, 363, 50 },
		{ 503, 364, 50 },
		{ 503, 365, 50 },
		{ 503, 366, 50 },
		{ 503, 367, 50 },
		{ 503, 368, 50 },
		{ 503, 369, 50 },
		{ 503, 370, 50 },
		{ 503, 371, 50 },
		{ 503, 372, 50 },
		{ 503, 373, 50 },
		{ 503, 374, 50 },
		{ 503, 375, 50 },
		{ 503, 376, 50 },
		{ 503, 377, 50 },
		{ 503, 378, 50 },
		{ 503, 379, 50 },
		{ 503, 380, 50 },
		{ 503, 381, 50 },
		{ 503, 382, 50 },
		{ 503, 383, 50 },
		{ 503, 384, 50 },
		{ 503, 385, 50 },
		{ 503, 386, 50 },
		{ 503, 387, 50 },
		{ 503, 388, 50 },
		{ 503, 389, 50 },
		{ 503, 390, 50 },
		{ 503, 391, 50 },
		{ 503, 392, 50 },
		{ 503, 393, 50 },
		{ 503, 394, 50 },
		{ 503, 395, 50 },
		{ 503, 396, 50 },
		{ 503, 397, 50 },
		{ 503, 398, 50 },
		{ 503, 399, 50 },
		{ 503, 400, 50 },
		{ 503, 401, 50 },
		{ 503, 402, 50 },
		{ 503, 403, 50 },
		{ 503, 404, 50 },
		{ 503, 405, 50 },
		{ 503, 406, 50 },
		{ 503, 407, 50 },
		{ 503, 408, 50 },
		{ 503, 409, 50 },
		{ 503, 410, 50 },
		{ 504, 350, 50 },
		{ 504, 351, 50 },
		{ 504, 352, 50 },
		{ 504, 353, 50 },
		{ 504, 354, 50 },
		{ 504, 355, 50 },
		{ 504, 356, 50 },
		{ 504, 357, 50 },
		{ 504, 358, 50 },
		{ 504, 359, 50 },
		{ 504, 360, 50 },
		{ 504, 361, 50 },
		{ 504, 362, 50 },
		{ 504, 363, 50 },
		{ 504, 364, 50 },
		{ 504, 365, 50 },
		{ 504, 366, 50 },
		{ 504, 367, 50 },
		{ 504, 368, 50 },
		{ 504, 369, 50 },
		{ 504, 370, 50 },
		{ 504, 371, 50 },
		{ 504, 372, 50 },
		{ 504, 373, 50 },
		{ 504, 374, 50 },
		{ 504, 375, 50 },
		{ 504, 376, 50 },
		{ 504, 377, 50 },
		{ 504, 378, 50 },
		{ 504, 379, 50 },
		{ 504, 380, 50 },
		{ 504, 381, 50 },
		{ 504, 382, 50 },
		{ 504, 383, 50 },
		{ 504, 384, 50 },
		{ 504, 385, 50 },
		{ 504, 386, 50 },
		{ 504, 387, 50 },
		{ 504, 388, 50 },
		{ 504, 389, 50 },
		{ 504, 390, 50 },
		{ 504, 391, 50 },
		{ 504, 392, 50 },
		{ 504, 393, 50 },
		{ 504, 394, 50 },
		{ 504, 395, 50 },
		{ 504, 396, 50 },
		{ 504, 397, 50 },
		{ 504, 398, 50 },
		{ 504, 399, 50 },
		{ 504, 400, 50 },
		{ 504, 401, 50 },
		{ 504, 402, 50 },
		{ 504, 403, 50 },
		{ 504, 404, 50 },
		{ 504, 405, 50 },
		{ 504, 406, 50 },
		{ 504, 407, 50 },
		{ 504, 408, 50 },
		{ 504, 409, 50 },
		{ 504, 410, 50 },
		{ 504, 411, 50 },
		{ 505, 351, 50 },
		{ 505, 352, 50 },
		{ 505, 353, 50 },
		{ 505, 354, 50 },
		{ 505, 355, 50 },
		{ 505, 356, 50 },
		{ 505, 357, 50 },
		{ 505, 358, 50 },
		{ 505, 359, 50 },
		{ 505, 360, 50 },
		{ 505, 361, 50 },
		{ 505, 362, 50 },
		{ 505, 363, 50 },
		{ 505, 364, 50 },
		{ 505, 365, 50 },
		{ 505, 366, 50 },
		{ 505, 367, 50 },
		{ 505, 368, 50 },
		{ 505, 369, 50 },
		{ 505, 370, 50 },
		{ 505, 371, 50 },
		{ 505, 372, 50 },
		{ 505, 373, 50 },
		{ 505, 374, 50 },
		{ 505, 375, 50 },
		{ 505, 376, 50 },
		{ 505, 377, 50 },
		{ 505, 378, 50 },
		{ 505, 379, 50 },
		{ 505, 380, 50 },
		{ 505, 381, 50 },
		{ 505, 382, 50 },
		{ 505, 383, 50 },
		{ 505, 384, 50 },
		{ 505, 385, 50 },
		{ 505, 386, 50 },
		{ 505, 387, 50 },
		{ 505, 388, 50 },
		{ 505, 389, 50 },
		{ 505, 390, 50 },
		{ 505, 391, 50 },
		{ 505, 392, 50 },
		{ 505, 393, 50 },
		{ 505, 394, 50 },
		{ 505, 395, 50 },
		{ 505, 396, 50 },
		{ 505, 397, 50 },
		{ 505, 398, 50 },
		{ 505, 399, 50 },
		{ 505, 400, 50 },
		{ 505, 401, 50 },
		{ 505, 402, 50 },
		{ 505, 403, 50 },
		{ 505, 404, 50 },
		{ 505, 405, 50 },
		{ 505, 406, 50 },
		{ 505, 407, 50 },
		{ 505, 408, 50 },
		{ 505, 409, 50 },
		{ 505, 410, 50 },
		{ 505, 411, 50 },
		{ 506, 351, 50 },
		{ 506, 352, 50 },
		{ 506, 353, 50 },
		{ 506, 354, 50 },
		{ 506, 355, 50 },
		{ 506, 356, 50 },
		{ 506, 357, 50 },
		{ 506, 358, 50 },
		{ 506, 359, 50 },
		{ 506, 360, 50 },
		{ 506, 361, 50 },
		{ 506, 362, 50 },
		{ 506, 363, 50 },
		{ 506, 364, 50 },
		{ 506, 365, 50 },
		{ 506, 366, 50 },
		{ 506, 367, 50 },
		{ 506, 368, 50 },
		{ 506, 369, 50 },
		{ 506, 370, 50 },
		{ 506, 371, 50 },
		{ 506, 372, 50 },
		{ 506, 373, 50 },
		{ 506, 374, 50 },
		{ 506, 375, 50 },
		{ 506, 376, 50 },
		{ 506, 377, 50 },
		{ 506, 378, 50 },
		{ 506, 379, 50 },
		{ 506, 380, 50 },
		{ 506, 381, 50 },
		{ 506, 382, 50 },
		{ 506, 383, 50 },
		{ 506, 384, 50 },
		{ 506, 385, 50 },
		{ 506, 386, 50 },
		{ 506, 387, 50 },
		{ 506, 388, 50 },
		{ 506, 389, 50 },
		{ 506, 390, 50 },
		{ 506, 391, 50 },
		{ 506, 392, 50 },
		{ 506, 393, 50 },
		{ 506, 394, 50 },
		{ 506, 395, 50 },
		{ 506, 396, 50 },
		{ 506, 397, 50 },
		{ 506, 398, 50 },
		{ 506, 399, 50 },
		{ 506, 400, 50 },
		{ 506, 401, 50 },
		{ 506, 402, 50 },
		{ 506, 403, 50 },
		{ 506, 404, 50 },
		{ 506, 405, 50 },
		{ 506, 406, 50 },
		{ 506, 407, 50 },
		{ 506, 408, 50 },
		{ 506, 409, 50 },
		{ 506, 410, 50 },
		{ 506, 411, 50 },
		{ 506, 412, 50 },
		{ 507, 351, 50 },
		{ 507, 352, 50 },
		{ 507, 353, 50 },
		{ 507, 354, 50 },
		{ 507, 355, 50 },
		{ 507, 356, 50 },
		{ 507, 357, 50 },
		{ 507, 358, 50 },
		{ 507, 359, 50 },
		{ 507, 360, 50 },
		{ 507, 361, 50 },
		{ 507, 362, 50 },
		{ 507, 363, 50 },
		{ 507, 364, 50 },
		{ 507, 365, 50 },
		{ 507, 366, 50 },
		{ 507, 367, 50 },
		{ 507, 368, 50 },
		{ 507, 369, 50 },
		{ 507, 370, 50 },
		{ 507, 371, 50 },
		{ 507, 372, 50 },
		{ 507, 373, 50 },
		{ 507, 374, 50 },
		{ 507, 375, 50 },
		{ 507, 376, 50 },
		{ 507, 377, 50 },
		{ 507, 378, 50 },
		{ 507, 379, 50 },
		{ 507, 380, 50 },
		{ 507, 381, 50 },
		{ 507, 382, 50 },
		{ 507, 383, 50 },
		{ 507, 384, 50 },
		{ 507, 385, 50 },
		{ 507, 386, 50 },
		{ 507, 387, 50 },
		{ 507, 388, 50 },
		{ 507, 389, 50 },
		{ 507, 390, 50 },
		{ 507, 391, 50 },
		{ 507, 392, 50 },
		{ 507, 393, 50 },
		{ 507, 394, 50 },
		{ 507, 395, 50 },
		{ 507, 396, 50 },
		{ 507, 397, 50 },
		{ 507, 398, 50 },
		{ 507, 399, 50 },
		{ 507, 400, 50 },
		{ 507, 401, 50 },
		{ 507, 402, 50 },
		{ 507, 403, 50 },
		{ 507, 404, 50 },
		{ 507, 405, 50 },
		{ 507, 406, 50 },
		{ 507, 407, 50 },
		{ 507, 408, 50 },
		{ 507, 409, 50 },
		{ 507, 410, 50 },
		{ 507, 411, 50 },
		{ 507, 412, 50 },
		{ 507, 413, 50 },
		{ 508, 352, 50 },
		{ 508, 353, 50 },
		{ 508, 354, 50 },
		{ 508, 355, 50 },
		{ 508, 356, 50 },
		{ 508, 357, 50 },
		{ 508, 358, 50 },
		{ 508, 359, 50 },
		{ 508, 360, 50 },
		{ 508, 361, 50 },
		{ 508, 362, 50 },
		{ 508, 363, 50 },
		{ 508, 364, 50 },
		{ 508, 365, 50 },
		{ 508, 366, 50 },
		{ 508, 367, 50 },
		{ 508, 368, 50 },
		{ 508, 369, 50 },
		{ 508, 370, 50 },
		{ 508, 371, 50 },
		{ 508, 372, 50 },
		{ 508, 373, 50 },
		{ 508, 374, 50 },
		{ 508, 375, 50 },
		{ 508, 376, 50 },
		{ 508, 377, 50 },
		{ 508, 378, 50 },
		{ 508, 379, 50 },
		{ 508, 380, 50 },
		{ 508, 381, 50 },
		{ 508, 382, 50 },
		{ 508, 383, 50 },
		{ 508, 384, 50 },
		{ 508, 385, 50 },
		{ 508, 386, 50 },
		{ 508, 387, 50 },
		{ 508, 388, 50 },
		{ 508, 389, 50 },
		{ 508, 390, 50 },
		{ 508, 391, 50 },
		{ 508, 392, 50 },
		{ 508, 393, 50 },
		{ 508, 394, 50 },
		{ 508, 395, 50 },
		{ 508, 396, 50 },
		{ 508, 397, 50 },
		{ 508, 398, 50 },
		{ 508, 399, 50 },
		{ 508, 400, 50 },
		{ 508, 401, 50 },
		{ 508, 402, 50 },
		{ 508, 403, 50 },
		{ 508, 404, 50 },
		{ 508, 405, 50 },
		{ 508, 406, 50 },
		{ 508, 407, 50 },
		{ 508, 408, 50 },
		{ 508, 409, 50 },
		{ 508, 410, 50 },
		{ 508, 411, 50 },
		{ 508, 412, 50 },
		{ 508, 413, 50 },
		{ 509, 352, 50 },
		{ 509, 353, 50 },
		{ 509, 354, 50 },
		{ 509, 355, 50 },
		{ 509, 356, 50 },
		{ 509, 357, 50 },
		{ 509, 358, 50 },
		{ 509, 359, 50 },
		{ 509, 360, 50 },
		{ 509, 361, 50 },
		{ 509, 362, 50 },
		{ 509, 363, 50 },
		{ 509, 364, 50 },
		{ 509, 365, 50 },
		{ 509, 366, 50 },
		{ 509, 367, 50 },
		{ 509, 368, 50 },
		{ 509, 369, 50 },
		{ 509, 370, 50 },
		{ 509, 371, 50 },
		{ 509, 372, 50 },
		{ 509, 373, 50 },
		{ 509, 374, 50 },
		{ 509, 375, 50 },
		{ 509, 376, 50 },
		{ 509, 377, 50 },
		{ 509, 378, 50 },
		{ 509, 379, 50 },
		{ 509, 380, 50 },
		{ 509, 381, 50 },
		{ 509, 382, 50 },
		{ 509, 383, 50 },
		{ 509, 384, 50 },
		{ 509, 385, 50 },
		{ 509, 386, 50 },
		{ 509, 387, 50 },
		{ 509, 388, 50 },
		{ 509, 389, 50 },
		{ 509, 390, 50 },
		{ 509, 391, 50 },
		{ 509, 392, 50 },
		{ 509, 393, 50 },
		{ 509, 394, 50 },
		{ 509, 395, 50 },
		{ 509, 396, 50 },
		{ 509, 397, 50 },
		{ 509, 398, 50 },
		{ 509, 399, 50 },
		{ 509, 400, 50 },
		{ 509, 401, 50 },
		{ 509, 402, 50 },
		{ 509, 403, 50 },
		{ 509, 404, 50 },
		{ 509, 405, 50 },
		{ 509, 406, 50 },
		{ 509, 407, 50 },
		{ 509, 408, 50 },
		{ 509, 409, 50 },
		{ 509, 410, 50 },
		{ 509, 411, 50 },
		{ 509, 412, 50 },
		{ 509, 413, 50 },
		{ 509, 414, 50 },
		{ 510, 352, 50 },
		{ 510, 353, 50 },
		{ 510, 354, 50 },
		{ 510, 355, 50 },
		{ 510, 356, 50 },
		{ 510, 357, 50 },
		{ 510, 358, 50 },
		{ 510, 359, 50 },
		{ 510, 360, 50 },
		{ 510, 361, 50 },
		{ 510, 362, 50 },
		{ 510, 363, 50 },
		{ 510, 364, 50 },
		{ 510, 365, 50 },
		{ 510, 366, 50 },
		{ 510, 367, 50 },
		{ 510, 368, 50 },
		{ 510, 369, 50 },
		{ 510, 370, 50 },
		{ 510, 371, 50 },
		{ 510, 372, 50 },
		{ 510, 373, 50 },
		{ 510, 374, 50 },
		{ 510, 375, 50 },
		{ 510, 376, 50 },
		{ 510, 377, 50 },
		{ 510, 378, 50 },
		{ 510, 379, 50 },
		{ 510, 380, 50 },
		{ 510, 381, 50 },
		{ 510, 382, 50 },
		{ 510, 383, 50 },
		{ 510, 384, 50 },
		{ 510, 385, 50 },
		{ 510, 386, 50 },
		{ 510, 387, 50 },
		{ 510, 388, 50 },
		{ 510, 389, 50 },
		{ 510, 390, 50 },
		{ 510, 391, 50 },
		{ 510, 392, 50 },
		{ 510, 393, 50 },
		{ 510, 394, 50 },
		{ 510, 395, 50 },
		{ 510, 396, 50 },
		{ 510, 397, 50 },
		{ 510, 398, 50 },
		{ 510, 399, 50 },
		{ 510, 400, 50 },
		{ 510, 401, 50 },
		{ 510, 402, 50 },
		{ 510, 403, 50 },
		{ 510, 404, 50 },
		{ 510, 405, 50 },
		{ 510, 406, 50 },
		{ 510, 407, 50 },
		{ 510, 408, 50 },
		{ 510, 409, 50 },
		{ 510, 410, 50 },
		{ 510, 411, 50 },
		{ 510, 412, 50 },
		{ 510, 413, 50 },
		{ 510, 414, 50 },
		{ 510, 415, 50 },
		{ 511, 353, 50 },
		{ 511, 354, 50 },
		{ 511, 355, 50 },
		{ 511, 356, 50 },
		{ 511, 357, 50 },
		{ 511, 358, 50 },
		{ 511, 359, 50 },
		{ 511, 360, 50 },
		{ 511, 361, 50 },
		{ 511, 362, 50 },
		{ 511, 363, 50 },
		{ 511, 364, 50 },
		{ 511, 365, 50 },
		{ 511, 366, 50 },
		{ 511, 367, 50 },
		{ 511, 368, 50 },
		{ 511, 369, 50 },
		{ 511, 370, 50 },
		{ 511, 371, 50 },
		{ 511, 372, 50 },
		{ 511, 373, 50 },
		{ 511, 374, 50 },
		{ 511, 375, 50 },
		{ 511, 376, 50 },
		{ 511, 377, 50 },
		{ 511, 378, 50 },
		{ 511, 379, 50 },
		{ 511, 380, 50 },
		{ 511, 381, 50 },
		{ 511, 382, 50 },
		{ 511, 383, 50 },
		{ 511, 384, 50 },
		{ 511, 385, 50 },
		{ 511, 386, 50 },
		{ 511, 387, 50 },
		{ 511, 388, 50 },
		{ 511, 389, 50 },
		{ 511, 390, 50 },
		{ 511, 391, 50 },
		{ 511, 392, 50 },
		{ 511, 393, 50 },
		{ 511, 394, 50 },
		{ 511, 395, 50 },
		{ 511, 396, 50 },
		{ 511, 397, 50 },
		{ 511, 398, 50 },
		{ 511, 399, 50 },
		{ 511, 400, 50 },
		{ 511, 401, 50 },
		{ 511, 402, 50 },
		{ 511, 403, 50 },
		{ 511, 404, 50 },
		{ 511, 405, 50 },
		{ 511, 406, 50 },
		{ 511, 407, 50 },
		{ 511, 408, 50 },
		{ 511, 409, 50 },
		{ 511, 410, 50 },
		{ 511, 411, 50 },
		{ 511, 412, 50 },
		{ 511, 413, 50 },
		{ 511, 414, 50 },
		{ 511, 415, 50 },
		{ 512, 353, 50 },
		{ 512, 354, 50 },
		{ 512, 355, 50 },
		{ 512, 356, 50 },
		{ 512, 357, 50 },
		{ 512, 358, 50 },
		{ 512, 359, 50 },
		{ 512, 360, 50 },
		{ 512, 361, 50 },
		{ 512, 362, 50 },
		{ 512, 363, 50 },
		{ 512, 364, 50 },
		{ 512, 365, 50 },
		{ 512, 366, 50 },
		{ 512, 367, 50 },
		{ 512, 368, 50 },
		{ 512, 369, 50 },
		{ 512, 370, 50 },
		{ 512, 371, 50 },
		{ 512, 372, 50 },
		{ 512, 373, 50 },
		{ 512, 374, 50 },
		{ 512, 375, 50 },
		{ 512, 376, 50 },
		{ 512, 377, 50 },
		{ 512, 378, 50 },
		{ 512, 379, 50 },
		{ 512, 380, 50 },
		{ 512, 381, 50 },
		{ 512, 382, 50 },
		{ 512, 383, 50 },
		{ 512, 384, 50 },
		{ 512, 385, 50 },
		{ 512, 386, 50 },
		{ 512, 387, 50 },
		{ 512, 388, 50 },
		{ 512, 389, 50 },
		{ 512, 390, 50 },
		{ 512, 391, 50 },
		{ 512, 392, 50 },
		{ 512, 393, 50 },
		{ 512, 394, 50 },
		{ 512, 395, 50 },
		{ 512, 396, 50 },
		{ 512, 397, 50 },
		{ 512, 398, 50 },
		{ 512, 399, 50 },
		{ 512, 400, 50 },
		{ 512, 401, 50 },
		{ 512, 402, 50 },
		{ 512, 403, 50 },
		{ 512, 404, 50 },
		{ 512, 405, 50 },
		{ 512, 406, 50 },
		{ 512, 407, 50 },
		{ 512, 408, 50 },
		{ 512, 409, 50 },
		{ 512, 410, 50 },
		{ 512, 411, 50 },
		{ 512, 412, 50 },
		{ 512, 413, 50 },
		{ 512, 414, 50 },
		{ 512, 415, 50 },
		{ 512, 416, 50 },
		{ 1, 66, 51 },
		{ 1, 67, 51 },
		{ 1, 68, 51 },
		{ 1, 69, 51 },
		{ 1, 70, 51 },
		{ 1, 71, 51 },
		{ 1, 72, 51 },
		{ 1, 73, 51 },
		{ 1, 74, 51 },
		{ 1, 75, 51 },
		{ 1, 76, 51 },
		{ 1, 77, 51 },
		{ 1, 78, 51 },
		{ 1, 79, 51 },
		{ 1, 80, 51 },
		{ 1, 81, 51 },
		{ 1, 82, 51 },
		{ 1, 83, 51 },
		{ 1, 84, 51 },
		{ 1, 85, 51 },
		{ 1, 86, 51 },
		{ 1, 87, 51 },
		{ 1, 88, 51 },
		{ 1, 89, 51 },
		{ 1, 90, 51 },
		{ 1, 91, 51 },
		{ 1, 92, 51 },
		{ 1, 93, 51 },
		{ 1, 94, 51 },
		{ 1, 95, 51 },
		{ 1, 96, 51 },
		{ 1, 97, 51 },
		{ 1, 98, 51 },
		{ 1, 99, 51 },
		{ 1, 100, 51 },
		{ 1, 101, 51 },
		{ 1, 102, 51 },
		{ 1, 103, 51 },
		{ 1, 104, 51 },
		{ 1, 105, 51 },
		{ 1, 106, 51 },
		{ 1, 107, 51 },
		{ 1, 108, 51 },
		{ 1, 109, 51 },
		{ 1, 110, 51 },
		{ 1, 111, 51 },
		{ 1, 112, 51 },
		{ 1, 113, 51 },
		{ 1, 114, 51 },
		{ 1, 115, 51 },
		{ 1, 116, 51 },
		{ 1, 117, 51 },
		{ 1, 118, 51 },
		{ 1, 119, 51 },
		{ 1, 120, 51 },
		{ 1, 121, 51 },
		{ 1, 122, 51 },
		{ 1, 123, 51 },
		{ 1, 124, 51 },
		{ 1, 125, 51 },
		{ 1, 126, 51 },
		{ 1, 127, 51 },
		{ 1, 128, 51 },
		{ 2, 66, 51 },
		{ 2, 67, 51 },
		{ 2, 68, 51 },
		{ 2, 69, 51 },
		{ 2, 70, 51 },
		{ 2, 71, 51 },
		{ 2, 72, 51 },
		{ 2, 73, 51 },
		{ 2, 74, 51 },
		{ 2, 75, 51 },
		{ 2, 76, 51 },
		{ 2, 77, 51 },
		{ 2, 78, 51 },
		{ 2, 79, 51 },
		{ 2, 80, 51 },
		{ 2, 81, 51 },
		{ 2, 82, 51 },
		{ 2, 83, 51 },
		{ 2, 84, 51 },
		{ 2, 85, 51 },
		{ 2, 86, 51 },
		{ 2, 87, 51 },
		{ 2, 88, 51 },
		{ 2, 89, 51 },
		{ 2, 90, 51 },
		{ 2, 91, 51 },
		{ 2, 92, 51 },
		{ 2, 93, 51 },
		{ 2, 94, 51 },
		{ 2, 95, 51 },
		{ 2, 96, 51 },
		{ 2, 97, 51 },
		{ 2, 98, 51 },
		{ 2, 99, 51 },
		{ 2, 100, 51 },
		{ 2, 101, 51 },
		{ 2, 102, 51 },
		{ 2, 103, 51 },
		{ 2, 104, 51 },
		{ 2, 105, 51 },
		{ 2, 106, 51 },
		{ 2, 107, 51 },
		{ 2, 108, 51 },
		{ 2, 109, 51 },
		{ 2, 110, 51 },
		{ 2, 111, 51 },
		{ 2, 112, 51 },
		{ 2, 113, 51 },
		{ 2, 114, 51 },
		{ 2, 115, 51 },
		{ 2, 116, 51 },
		{ 2, 117, 51 },
		{ 2, 118, 51 },
		{ 2, 119, 51 },
		{ 2, 120, 51 },
		{ 2, 121, 51 },
		{ 2, 122, 51 },
		{ 2, 123, 51 },
		{ 2, 124, 51 },
		{ 2, 125, 51 },
		{ 2, 126, 51 },
		{ 2, 127, 51 },
		{ 2, 128, 51 },
		{ 2, 129, 51 },
		{ 3, 67, 51 },
		{ 3, 68, 51 },
		{ 3, 69, 51 },
		{ 3, 70, 51 },
		{ 3, 71, 51 },
		{ 3, 72, 51 },
		{ 3, 73, 51 },
		{ 3, 74, 51 },
		{ 3, 75, 51 },
		{ 3, 76, 51 },
		{ 3, 77, 51 },
		{ 3, 78, 51 },
		{ 3, 79, 51 },
		{ 3, 80, 51 },
		{ 3, 81, 51 },
		{ 3, 82, 51 },
		{ 3, 83, 51 },
		{ 3, 84, 51 },
		{ 3, 85, 51 },
		{ 3, 86, 51 },
		{ 3, 87, 51 },
		{ 3, 88, 51 },
		{ 3, 89, 51 },
		{ 3, 90, 51 },
		{ 3, 91, 51 },
		{ 3, 92, 51 },
		{ 3, 93, 51 },
		{ 3, 94, 51 },
		{ 3, 95, 51 },
		{ 3, 96, 51 },
		{ 3, 97, 51 },
		{ 3, 98, 51 },
		{ 3, 99, 51 },
		{ 3, 100, 51 },
		{ 3, 101, 51 },
		{ 3, 102, 51 },
		{ 3, 103, 51 },
		{ 3, 104, 51 },
		{ 3, 105, 51 },
		{ 3, 106, 51 },
		{ 3, 107, 51 },
		{ 3, 108, 51 },
		{ 3, 109, 51 },
		{ 3, 110, 51 },
		{ 3, 111, 51 },
		{ 3, 112, 51 },
		{ 3, 113, 51 },
		{ 3, 114, 51 },
		{ 3, 115, 51 },
		{ 3, 116, 51 },
		{ 3, 117, 51 },
		{ 3, 118, 51 },
		{ 3, 119, 51 },
		{ 3, 120, 51 },
		{ 3, 121, 51 },
		{ 3, 122, 51 },
		{ 3, 123, 51 },
		{ 3, 124, 51 },
		{ 3, 125, 51 },
		{ 3, 126, 51 },
		{ 3, 127, 51 },
		{ 3, 128, 51 },
		{ 3, 129, 51 },
		{ 4, 68, 51 },
		{ 4, 69, 51 },
		{ 4, 70, 51 },
		{ 4, 71, 51 },
		{ 4, 72, 51 },
		{ 4, 73, 51 },
		{ 4, 74, 51 },
		{ 4, 75, 51 },
		{ 4, 76, 51 },
		{ 4, 77, 51 },
		{ 4, 78, 51 },
		{ 4, 79, 51 },
		{ 4, 80, 51 },
		{ 4, 81, 51 },
		{ 4, 82, 51 },
		{ 4, 83, 51 },
		{ 4, 84, 51 },
		{ 4, 85, 51 },
		{ 4, 86, 51 },
		{ 4, 87, 51 },
		{ 4, 88, 51 },
		{ 4, 89, 51 },
		{ 4, 90, 51 },
		{ 4, 91, 51 },
		{ 4, 92, 51 },
		{ 4, 93, 51 },
		{ 4, 94, 51 },
		{ 4, 95, 51 },
		{ 4, 96, 51 },
		{ 4, 97, 51 },
		{ 4, 98, 51 },
		{ 4, 99, 51 },
		{ 4, 100, 51 },
		{ 4, 101, 51 },
		{ 4, 102, 51 },
		{ 4, 103, 51 },
		{ 4, 104, 51 },
		{ 4, 105, 51 },
		{ 4, 106, 51 },
		{ 4, 107, 51 },
		{ 4, 108, 51 },
		{ 4, 109, 51 },
		{ 4, 110, 51 },
		{ 4, 111, 51 },
		{ 4, 112, 51 },
		{ 4, 113, 51 },
		{ 4, 114, 51 },
		{ 4, 115, 51 },
		{ 4, 116, 51 },
		{ 4, 117, 51 },
		{ 4, 118, 51 },
		{ 4, 119, 51 },
		{ 4, 120, 51 },
		{ 4, 121, 51 },
		{ 4, 122, 51 },
		{ 4, 123, 51 },
		{ 4, 124, 51 },
		{ 4, 125, 51 },
		{ 4, 126, 51 },
		{ 4, 127, 51 },
		{ 4, 128, 51 },
		{ 4, 129, 51 },
		{ 4, 130, 51 },
		{ 5, 69, 51 },
		{ 5, 70, 51 },
		{ 5, 71, 51 },
		{ 5, 72, 51 },
		{ 5, 73, 51 },
		{ 5, 74, 51 },
		{ 5, 75, 51 },
		{ 5, 76, 51 },
		{ 5, 77, 51 },
		{ 5, 78, 51 },
		{ 5, 79, 51 },
		{ 5, 80, 51 },
		{ 5, 81, 51 },
		{ 5, 82, 51 },
		{ 5, 83, 51 },
		{ 5, 84, 51 },
		{ 5, 85, 51 },
		{ 5, 86, 51 },
		{ 5, 87, 51 },
		{ 5, 88, 51 },
		{ 5, 89, 51 },
		{ 5, 90, 51 },
		{ 5, 91, 51 },
		{ 5, 92, 51 },
		{ 5, 93, 51 },
		{ 5, 94, 51 },
		{ 5, 95, 51 },
		{ 5, 96, 51 },
		{ 5, 97, 51 },
		{ 5, 98, 51 },
		{ 5, 99, 51 },
		{ 5, 100, 51 },
		{ 5, 101, 51 },
		{ 5, 102, 51 },
		{ 5, 103, 51 },
		{ 5, 104, 51 },
		{ 5, 105, 51 },
		{ 5, 106, 51 },
		{ 5, 107, 51 },
		{ 5, 108, 51 },
		{ 5, 109, 51 },
		{ 5, 110, 51 },
		{ 5, 111, 51 },
		{ 5, 112, 51 },
		{ 5, 113, 51 },
		{ 5, 114, 51 },
		{ 5, 115, 51 },
		{ 5, 116, 51 },
		{ 5, 117, 51 },
		{ 5, 118, 51 },
		{ 5, 119, 51 },
		{ 5, 120, 51 },
		{ 5, 121, 51 },
		{ 5, 122, 51 },
		{ 5, 123, 51 },
		{ 5, 124, 51 },
		{ 5, 125, 51 },
		{ 5, 126, 51 },
		{ 5, 127, 51 },
		{ 5, 128, 51 },
		{ 5, 129, 51 },
		{ 5, 130, 51 },
		{ 6, 69, 51 },
		{ 6, 70, 51 },
		{ 6, 71, 51 },
		{ 6, 72, 51 },
		{ 6, 73, 51 },
		{ 6, 74, 51 },
		{ 6, 75, 51 },
		{ 6, 76, 51 },
		{ 6, 77, 51 },
		{ 6, 78, 51 },
		{ 6, 79, 51 },
		{ 6, 80, 51 },
		{ 6, 81, 51 },
		{ 6, 82, 51 },
		{ 6, 83, 51 },
		{ 6, 84, 51 },
		{ 6, 85, 51 },
		{ 6, 86, 51 },
		{ 6, 87, 51 },
		{ 6, 88, 51 },
		{ 6, 89, 51 },
		{ 6, 90, 51 },
		{ 6, 91, 51 },
		{ 6, 92, 51 },
		{ 6, 93, 51 },
		{ 6, 94, 51 },
		{ 6, 95, 51 },
		{ 6, 96, 51 },
		{ 6, 97, 51 },
		{ 6, 98, 51 },
		{ 6, 99, 51 },
		{ 6, 100, 51 },
		{ 6, 101, 51 },
		{ 6, 102, 51 },
		{ 6, 103, 51 },
		{ 6, 104, 51 },
		{ 6, 105, 51 },
		{ 6, 106, 51 },
		{ 6, 107, 51 },
		{ 6, 108, 51 },
		{ 6, 109, 51 },
		{ 6, 110, 51 },
		{ 6, 111, 51 },
		{ 6, 112, 51 },
		{ 6, 113, 51 },
		{ 6, 114, 51 },
		{ 6, 115, 51 },
		{ 6, 116, 51 },
		{ 6, 117, 51 },
		{ 6, 118, 51 },
		{ 6, 119, 51 },
		{ 6, 120, 51 },
		{ 6, 121, 51 },
		{ 6, 122, 51 },
		{ 6, 123, 51 },
		{ 6, 124, 51 },
		{ 6, 125, 51 },
		{ 6, 126, 51 },
		{ 6, 127, 51 },
		{ 6, 128, 51 },
		{ 6, 129, 51 },
		{ 6, 130, 51 },
		{ 6, 131, 51 },
		{ 7, 70, 51 },
		{ 7, 71, 51 },
		{ 7, 72, 51 },
		{ 7, 73, 51 },
		{ 7, 74, 51 },
		{ 7, 75, 51 },
		{ 7, 76, 51 },
		{ 7, 77, 51 },
		{ 7, 78, 51 },
		{ 7, 79, 51 },
		{ 7, 80, 51 },
		{ 7, 81, 51 },
		{ 7, 82, 51 },
		{ 7, 83, 51 },
		{ 7, 84, 51 },
		{ 7, 85, 51 },
		{ 7, 86, 51 },
		{ 7, 87, 51 },
		{ 7, 88, 51 },
		{ 7, 89, 51 },
		{ 7, 90, 51 },
		{ 7, 91, 51 },
		{ 7, 92, 51 },
		{ 7, 93, 51 },
		{ 7, 94, 51 },
		{ 7, 95, 51 },
		{ 7, 96, 51 },
		{ 7, 97, 51 },
		{ 7, 98, 51 },
		{ 7, 99, 51 },
		{ 7, 100, 51 },
		{ 7, 101, 51 },
		{ 7, 102, 51 },
		{ 7, 103, 51 },
		{ 7, 104, 51 },
		{ 7, 105, 51 },
		{ 7, 106, 51 },
		{ 7, 107, 51 },
		{ 7, 108, 51 },
		{ 7, 109, 51 },
		{ 7, 110, 51 },
		{ 7, 111, 51 },
		{ 7, 112, 51 },
		{ 7, 113, 51 },
		{ 7, 114, 51 },
		{ 7, 115, 51 },
		{ 7, 116, 51 },
		{ 7, 117, 51 },
		{ 7, 118, 51 },
		{ 7, 119, 51 },
		{ 7, 120, 51 },
		{ 7, 121, 51 },
		{ 7, 122, 51 },
		{ 7, 123, 51 },
		{ 7, 124, 51 },
		{ 7, 125, 51 },
		{ 7, 126, 51 },
		{ 7, 127, 51 },
		{ 7, 128, 51 },
		{ 7, 129, 51 },
		{ 7, 130, 51 },
		{ 7, 131, 51 },
		{ 8, 71, 51 },
		{ 8, 72, 51 },
		{ 8, 73, 51 },
		{ 8, 74, 51 },
		{ 8, 75, 51 },
		{ 8, 76, 51 },
		{ 8, 77, 51 },
		{ 8, 78, 51 },
		{ 8, 79, 51 },
		{ 8, 80, 51 },
		{ 8, 81, 51 },
		{ 8, 82, 51 },
		{ 8, 83, 51 },
		{ 8, 84, 51 },
		{ 8, 85, 51 },
		{ 8, 86, 51 },
		{ 8, 87, 51 },
		{ 8, 88, 51 },
		{ 8, 89, 51 },
		{ 8, 90, 51 },
		{ 8, 91, 51 },
		{ 8, 92, 51 },
		{ 8, 93, 51 },
		{ 8, 94, 51 },
		{ 8, 95, 51 },
		{ 8, 96, 51 },
		{ 8, 97, 51 },
		{ 8, 98, 51 },
		{ 8, 99, 51 },
		{ 8, 100, 51 },
		{ 8, 101, 51 },
		{ 8, 102, 51 },
		{ 8, 103, 51 },
		{ 8, 104, 51 },
		{ 8, 105, 51 },
		{ 8, 106, 51 },
		{ 8, 107, 51 },
		{ 8, 108, 51 },
		{ 8, 109, 51 },
		{ 8, 110, 51 },
		{ 8, 111, 51 },
		{ 8, 112, 51 },
		{ 8, 113, 51 },
		{ 8, 114, 51 },
		{ 8, 115, 51 },
		{ 8, 116, 51 },
		{ 8, 117, 51 },
		{ 8, 118, 51 },
		{ 8, 119, 51 },
		{ 8, 120, 51 },
		{ 8, 121, 51 },
		{ 8, 122, 51 },
		{ 8, 123, 51 },
		{ 8, 124, 51 },
		{ 8, 125, 51 },
		{ 8, 126, 51 },
		{ 8, 127, 51 },
		{ 8, 128, 51 },
		{ 8, 129, 51 },
		{ 8, 130, 51 },
		{ 8, 131, 51 },
		{ 8, 132, 51 },
		{ 9, 72, 51 },
		{ 9, 73, 51 },
		{ 9, 74, 51 },
		{ 9, 75, 51 },
		{ 9, 76, 51 },
		{ 9, 77, 51 },
		{ 9, 78, 51 },
		{ 9, 79, 51 },
		{ 9, 80, 51 },
		{ 9, 81, 51 },
		{ 9, 82, 51 },
		{ 9, 83, 51 },
		{ 9, 84, 51 },
		{ 9, 85, 51 },
		{ 9, 86, 51 },
		{ 9, 87, 51 },
		{ 9, 88, 51 },
		{ 9, 89, 51 },
		{ 9, 90, 51 },
		{ 9, 91, 51 },
		{ 9, 92, 51 },
		{ 9, 93, 51 },
		{ 9, 94, 51 },
		{ 9, 95, 51 },
		{ 9, 96, 51 },
		{ 9, 97, 51 },
		{ 9, 98, 51 },
		{ 9, 99, 51 },
		{ 9, 100, 51 },
		{ 9, 101, 51 },
		{ 9, 102, 51 },
		{ 9, 103, 51 },
		{ 9, 104, 51 },
		{ 9, 105, 51 },
		{ 9, 106, 51 },
		{ 9, 107, 51 },
		{ 9, 108, 51 },
		{ 9, 109, 51 },
		{ 9, 110, 51 },
		{ 9, 111, 51 },
		{ 9, 112, 51 },
		{ 9, 113, 51 },
		{ 9, 114, 51 },
		{ 9, 115, 51 },
		{ 9, 116, 51 },
		{ 9, 117, 51 },
		{ 9, 118, 51 },
		{ 9, 119, 51 },
		{ 9, 120, 51 },
		{ 9, 121, 51 },
		{ 9, 122, 51 },
		{ 9, 123, 51 },
		{ 9, 124, 51 },
		{ 9, 125, 51 },
		{ 9, 126, 51 },
		{ 9, 127, 51 },
		{ 9, 128, 51 },
		{ 9, 129, 51 },
		{ 9, 130, 51 },
		{ 9, 131, 51 },
		{ 9, 132, 51 },
		{ 10, 72, 51 },
		{ 10, 73, 51 },
		{ 10, 74, 51 },
		{ 10, 75, 51 },
		{ 10, 76, 51 },
		{ 10, 77, 51 },
		{ 10, 78, 51 },
		{ 10, 79, 51 },
		{ 10, 80, 51 },
		{ 10, 81, 51 },
		{ 10, 82, 51 },
		{ 10, 83, 51 },
		{ 10, 84, 51 },
		{ 10, 85, 51 },
		{ 10, 86, 51 },
		{ 10, 87, 51 },
		{ 10, 88, 51 },
		{ 10, 89, 51 },
		{ 10, 90, 51 },
		{ 10, 91, 51 },
		{ 10, 92, 51 },
		{ 10, 93, 51 },
		{ 10, 94, 51 },
		{ 10, 95, 51 },
		{ 10, 96, 51 },
		{ 10, 97, 51 },
		{ 10, 98, 51 },
		{ 10, 99, 51 },
		{ 10, 100, 51 },
		{ 10, 101, 51 },
		{ 10, 102, 51 },
		{ 10, 103, 51 },
		{ 10, 104, 51 },
		{ 10, 105, 51 },
		{ 10, 106, 51 },
		{ 10, 107, 51 },
		{ 10, 108, 51 },
		{ 10, 109, 51 },
		{ 10, 110, 51 },
		{ 10, 111, 51 },
		{ 10, 112, 51 },
		{ 10, 113, 51 },
		{ 10, 114, 51 },
		{ 10, 115, 51 },
		{ 10, 116, 51 },
		{ 10, 117, 51 },
		{ 10, 118, 51 },
		{ 10, 119, 51 },
		{ 10, 120, 51 },
		{ 10, 121, 51 },
		{ 10, 122, 51 },
		{ 10, 123, 51 },
		{ 10, 124, 51 },
		{ 10, 125, 51 },
		{ 10, 126, 51 },
		{ 10, 127, 51 },
		{ 10, 128, 51 },
		{ 10, 129, 51 },
		{ 10, 130, 51 },
		{ 10, 131, 51 },
		{ 10, 132, 51 },
		{ 10, 133, 51 },
		{ 11, 73, 51 },
		{ 11, 74, 51 },
		{ 11, 75, 51 },
		{ 11, 76, 51 },
		{ 11, 77, 51 },
		{ 11, 78, 51 },
		{ 11, 79, 51 },
		{ 11, 80, 51 },
		{ 11, 81, 51 },
		{ 11, 82, 51 },
		{ 11, 83, 51 },
		{ 11, 84, 51 },
		{ 11, 85, 51 },
		{ 11, 86, 51 },
		{ 11, 87, 51 },
		{ 11, 88, 51 },
		{ 11, 89, 51 },
		{ 11, 90, 51 },
		{ 11, 91, 51 },
		{ 11, 92, 51 },
		{ 11, 93, 51 },
		{ 11, 94, 51 },
		{ 11, 95, 51 },
		{ 11, 96, 51 },
		{ 11, 97, 51 },
		{ 11, 98, 51 },
		{ 11, 99, 51 },
		{ 11, 100, 51 },
		{ 11, 101, 51 },
		{ 11, 102, 51 },
		{ 11, 103, 51 },
		{ 11, 104, 51 },
		{ 11, 105, 51 },
		{ 11, 106, 51 },
		{ 11, 107, 51 },
		{ 11, 108, 51 },
		{ 11, 109, 51 },
		{ 11, 110, 51 },
		{ 11, 111, 51 },
		{ 11, 112, 51 },
		{ 11, 113, 51 },
		{ 11, 114, 51 },
		{ 11, 115, 51 },
		{ 11, 116, 51 },
		{ 11, 117, 51 },
		{ 11, 118, 51 },
		{ 11, 119, 51 },
		{ 11, 120, 51 },
		{ 11, 121, 51 },
		{ 11, 122, 51 },
		{ 11, 123, 51 },
		{ 11, 124, 51 },
		{ 11, 125, 51 },
		{ 11, 126, 51 },
		{ 11, 127, 51 },
		{ 11, 128, 51 },
		{ 11, 129, 51 },
		{ 11, 130, 51 },
		{ 11, 131, 51 },
		{ 11, 132, 51 },
		{ 11, 133, 51 },
		{ 12, 74, 51 },
		{ 12, 75, 51 },
		{ 12, 76, 51 },
		{ 12, 77, 51 },
		{ 12, 78, 51 },
		{ 12, 79, 51 },
		{ 12, 80, 51 },
		{ 12, 81, 51 },
		{ 12, 82, 51 },
		{ 12, 83, 51 },
		{ 12, 84, 51 },
		{ 12, 85, 51 },
		{ 12, 86, 51 },
		{ 12, 87, 51 },
		{ 12, 88, 51 },
		{ 12, 89, 51 },
		{ 12, 90, 51 },
		{ 12, 91, 51 },
		{ 12, 92, 51 },
		{ 12, 93, 51 },
		{ 12, 94, 51 },
		{ 12, 95, 51 },
		{ 12, 96, 51 },
		{ 12, 97, 51 },
		{ 12, 98, 51 },
		{ 12, 99, 51 },
		{ 12, 100, 51 },
		{ 12, 101, 51 },
		{ 12, 102, 51 },
		{ 12, 103, 51 },
		{ 12, 104, 51 },
		{ 12, 105, 51 },
		{ 12, 106, 51 },
		{ 12, 107, 51 },
		{ 12, 108, 51 },
		{ 12, 109, 51 },
		{ 12, 110, 51 },
		{ 12, 111, 51 },
		{ 12, 112, 51 },
		{ 12, 113, 51 },
		{ 12, 114, 51 },
		{ 12, 115, 51 },
		{ 12, 116, 51 },
		{ 12, 117, 51 },
		{ 12, 118, 51 },
		{ 12, 119, 51 },
		{ 12, 120, 51 },
		{ 12, 121, 51 },
		{ 12, 122, 51 },
		{ 12, 123, 51 },
		{ 12, 124, 51 },
		{ 12, 125, 51 },
		{ 12, 126, 51 },
		{ 12, 127, 51 },
		{ 12, 128, 51 },
		{ 12, 129, 51 },
		{ 12, 130, 51 },
		{ 12, 131, 51 },
		{ 12, 132, 51 },
		{ 12, 133, 51 },
		{ 12, 134, 51 },
		{ 13, 75, 51 },
		{ 13, 76, 51 },
		{ 13, 77, 51 },
		{ 13, 78, 51 },
		{ 13, 79, 51 },
		{ 13, 80, 51 },
		{ 13, 81, 51 },
		{ 13, 82, 51 },
		{ 13, 83, 51 },
		{ 13, 84, 51 },
		{ 13, 85, 51 },
		{ 13, 86, 51 },
		{ 13, 87, 51 },
		{ 13, 88, 51 },
		{ 13, 89, 51 },
		{ 13, 90, 51 },
		{ 13, 91, 51 },
		{ 13, 92, 51 },
		{ 13, 93, 51 },
		{ 13, 94, 51 },
		{ 13, 95, 51 },
		{ 13, 96, 51 },
		{ 13, 97, 51 },
		{ 13, 98, 51 },
		{ 13, 99, 51 },
		{ 13, 100, 51 },
		{ 13, 101, 51 },
		{ 13, 102, 51 },
		{ 13, 103, 51 },
		{ 13, 104, 51 },
		{ 13, 105, 51 },
		{ 13, 106, 51 },
		{ 13, 107, 51 },
		{ 13, 108, 51 },
		{ 13, 109, 51 },
		{ 13, 110, 51 },
		{ 13, 111, 51 },
		{ 13, 112, 51 },
		{ 13, 113, 51 },
		{ 13, 114, 51 },
		{ 13, 115, 51 },
		{ 13, 116, 51 },
		{ 13, 117, 51 },
		{ 13, 118, 51 },
		{ 13, 119, 51 },
		{ 13, 120, 51 },
		{ 13, 121, 51 },
		{ 13, 122, 51 },
		{ 13, 123, 51 },
		{ 13, 124, 51 },
		{ 13, 125, 51 },
		{ 13, 126, 51 },
		{ 13, 127, 51 },
		{ 13, 128, 51 },
		{ 13, 129, 51 },
		{ 13, 130, 51 },
		{ 13, 131, 51 },
		{ 13, 132, 51 },
		{ 13, 133, 51 },
		{ 13, 134, 51 },
		{ 14, 75, 51 },
		{ 14, 76, 51 },
		{ 14, 77, 51 },
		{ 14, 78, 51 },
		{ 14, 79, 51 },
		{ 14, 80, 51 },
		{ 14, 81, 51 },
		{ 14, 82, 51 },
		{ 14, 83, 51 },
		{ 14, 84, 51 },
		{ 14, 85, 51 },
		{ 14, 86, 51 },
		{ 14, 87, 51 },
		{ 14, 88, 51 },
		{ 14, 89, 51 },
		{ 14, 90, 51 },
		{ 14, 91, 51 },
		{ 14, 92, 51 },
		{ 14, 93, 51 },
		{ 14, 94, 51 },
		{ 14, 95, 51 },
		{ 14, 96, 51 },
		{ 14, 97, 51 },
		{ 14, 98, 51 },
		{ 14, 99, 51 },
		{ 14, 100, 51 },
		{ 14, 101, 51 },
		{ 14, 102, 51 },
		{ 14, 103, 51 },
		{ 14, 104, 51 },
		{ 14, 105, 51 },
		{ 14, 106, 51 },
		{ 14, 107, 51 },
		{ 14, 108, 51 },
		{ 14, 109, 51 },
		{ 14, 110, 51 },
		{ 14, 111, 51 },
		{ 14, 112, 51 },
		{ 14, 113, 51 },
		{ 14, 114, 51 },
		{ 14, 115, 51 },
		{ 14, 116, 51 },
		{ 14, 117, 51 },
		{ 14, 118, 51 },
		{ 14, 119, 51 },
		{ 14, 120, 51 },
		{ 14, 121, 51 },
		{ 14, 122, 51 },
		{ 14, 123, 51 },
		{ 14, 124, 51 },
		{ 14, 125, 51 },
		{ 14, 126, 51 },
		{ 14, 127, 51 },
		{ 14, 128, 51 },
		{ 14, 129, 51 },
		{ 14, 130, 51 },
		{ 14, 131, 51 },
		{ 14, 132, 51 },
		{ 14, 133, 51 },
		{ 14, 134, 51 },
		{ 14, 135, 51 },
		{ 15, 76, 51 },
		{ 15, 77, 51 },
		{ 15, 78, 51 },
		{ 15, 79, 51 },
		{ 15, 80, 51 },
		{ 15, 81, 51 },
		{ 15, 82, 51 },
		{ 15, 83, 51 },
		{ 15, 84, 51 },
		{ 15, 85, 51 },
		{ 15, 86, 51 },
		{ 15, 87, 51 },
		{ 15, 88, 51 },
		{ 15, 89, 51 },
		{ 15, 90, 51 },
		{ 15, 91, 51 },
		{ 15, 92, 51 },
		{ 15, 93, 51 },
		{ 15, 94, 51 },
		{ 15, 95, 51 },
		{ 15, 96, 51 },
		{ 15, 97, 51 },
		{ 15, 98, 51 },
		{ 15, 99, 51 },
		{ 15, 100, 51 },
		{ 15, 101, 51 },
		{ 15, 102, 51 },
		{ 15, 103, 51 },
		{ 15, 104, 51 },
		{ 15, 105, 51 },
		{ 15, 106, 51 },
		{ 15, 107, 51 },
		{ 15, 108, 51 },
		{ 15, 109, 51 },
		{ 15, 110, 51 },
		{ 15, 111, 51 },
		{ 15, 112, 51 },
		{ 15, 113, 51 },
		{ 15, 114, 51 },
		{ 15, 115, 51 },
		{ 15, 116, 51 },
		{ 15, 117, 51 },
		{ 15, 118, 51 },
		{ 15, 119, 51 },
		{ 15, 120, 51 },
		{ 15, 121, 51 },
		{ 15, 122, 51 },
		{ 15, 123, 51 },
		{ 15, 124, 51 },
		{ 15, 125, 51 },
		{ 15, 126, 51 },
		{ 15, 127, 51 },
		{ 15, 128, 51 },
		{ 15, 129, 51 },
		{ 15, 130, 51 },
		{ 15, 131, 51 },
		{ 15, 132, 51 },
		{ 15, 133, 51 },
		{ 15, 134, 51 },
		{ 15, 135, 51 },
		{ 16, 77, 51 },
		{ 16, 78, 51 },
		{ 16, 79, 51 },
		{ 16, 80, 51 },
		{ 16, 81, 51 },
		{ 16, 82, 51 },
		{ 16, 83, 51 },
		{ 16, 84, 51 },
		{ 16, 85, 51 },
		{ 16, 86, 51 },
		{ 16, 87, 51 },
		{ 16, 88, 51 },
		{ 16, 89, 51 },
		{ 16, 90, 51 },
		{ 16, 91, 51 },
		{ 16, 92, 51 },
		{ 16, 93, 51 },
		{ 16, 94, 51 },
		{ 16, 95, 51 },
		{ 16, 96, 51 },
		{ 16, 97, 51 },
		{ 16, 98, 51 },
		{ 16, 99, 51 },
		{ 16, 100, 51 },
		{ 16, 101, 51 },
		{ 16, 102, 51 },
		{ 16, 103, 51 },
		{ 16, 104, 51 },
		{ 16, 105, 51 },
		{ 16, 106, 51 },
		{ 16, 107, 51 },
		{ 16, 108, 51 },
		{ 16, 109, 51 },
		{ 16, 110, 51 },
		{ 16, 111, 51 },
		{ 16, 112, 51 },
		{ 16, 113, 51 },
		{ 16, 114, 51 },
		{ 16, 115, 51 },
		{ 16, 116, 51 },
		{ 16, 117, 51 },
		{ 16, 118, 51 },
		{ 16, 119, 51 },
		{ 16, 120, 51 },
		{ 16, 121, 51 },
		{ 16, 122, 51 },
		{ 16, 123, 51 },
		{ 16, 124, 51 },
		{ 16, 125, 51 },
		{ 16, 126, 51 },
		{ 16, 127, 51 },
		{ 16, 128, 51 },
		{ 16, 129, 51 },
		{ 16, 130, 51 },
		{ 16, 131, 51 },
		{ 16, 132, 51 },
		{ 16, 133, 51 },
		{ 16, 134, 51 },
		{ 16, 135, 51 },
		{ 16, 136, 51 },
		{ 17, 78, 51 },
		{ 17, 79, 51 },
		{ 17, 80, 51 },
		{ 17, 81, 51 },
		{ 17, 82, 51 },
		{ 17, 83, 51 },
		{ 17, 84, 51 },
		{ 17, 85, 51 },
		{ 17, 86, 51 },
		{ 17, 87, 51 },
		{ 17, 88, 51 },
		{ 17, 89, 51 },
		{ 17, 90, 51 },
		{ 17, 91, 51 },
		{ 17, 92, 51 },
		{ 17, 93, 51 },
		{ 17, 94, 51 },
		{ 17, 95, 51 },
		{ 17, 96, 51 },
		{ 17, 97, 51 },
		{ 17, 98, 51 },
		{ 17, 99, 51 },
		{ 17, 100, 51 },
		{ 17, 101, 51 },
		{ 17, 102, 51 },
		{ 17, 103, 51 },
		{ 17, 104, 51 },
		{ 17, 105, 51 },
		{ 17, 106, 51 },
		{ 17, 107, 51 },
		{ 17, 108, 51 },
		{ 17, 109, 51 },
		{ 17, 110, 51 },
		{ 17, 111, 51 },
		{ 17, 112, 51 },
		{ 17, 113, 51 },
		{ 17, 114, 51 },
		{ 17, 115, 51 },
		{ 17, 116, 51 },
		{ 17, 117, 51 },
		{ 17, 118, 51 },
		{ 17, 119, 51 },
		{ 17, 120, 51 },
		{ 17, 121, 51 },
		{ 17, 122, 51 },
		{ 17, 123, 51 },
		{ 17, 124, 51 },
		{ 17, 125, 51 },
		{ 17, 126, 51 },
		{ 17, 127, 51 },
		{ 17, 128, 51 },
		{ 17, 129, 51 },
		{ 17, 130, 51 },
		{ 17, 131, 51 },
		{ 17, 132, 51 },
		{ 17, 133, 51 },
		{ 17, 134, 51 },
		{ 17, 135, 51 },
		{ 17, 136, 51 },
		{ 18, 78, 51 },
		{ 18, 79, 51 },
		{ 18, 80, 51 },
		{ 18, 81, 51 },
		{ 18, 82, 51 },
		{ 18, 83, 51 },
		{ 18, 84, 51 },
		{ 18, 85, 51 },
		{ 18, 86, 51 },
		{ 18, 87, 51 },
		{ 18, 88, 51 },
		{ 18, 89, 51 },
		{ 18, 90, 51 },
		{ 18, 91, 51 },
		{ 18, 92, 51 },
		{ 18, 93, 51 },
		{ 18, 94, 51 },
		{ 18, 95, 51 },
		{ 18, 96, 51 },
		{ 18, 97, 51 },
		{ 18, 98, 51 },
		{ 18, 99, 51 },
		{ 18, 100, 51 },
		{ 18, 101, 51 },
		{ 18, 102, 51 },
		{ 18, 103, 51 },
		{ 18, 104, 51 },
		{ 18, 105, 51 },
		{ 18, 106, 51 },
		{ 18, 107, 51 },
		{ 18, 108, 51 },
		{ 18, 109, 51 },
		{ 18, 110, 51 },
		{ 18, 111, 51 },
		{ 18, 112, 51 },
		{ 18, 113, 51 },
		{ 18, 114, 51 },
		{ 18, 115, 51 },
		{ 18, 116, 51 },
		{ 18, 117, 51 },
		{ 18, 118, 51 },
		{ 18, 119, 51 },
		{ 18, 120, 51 },
		{ 18, 121, 51 },
		{ 18, 122, 51 },
		{ 18, 123, 51 },
		{ 18, 124, 51 },
		{ 18, 125, 51 },
		{ 18, 126, 51 },
		{ 18, 127, 51 },
		{ 18, 128, 51 },
		{ 18, 129, 51 },
		{ 18, 130, 51 },
		{ 18, 131, 51 },
		{ 18, 132, 51 },
		{ 18, 133, 51 },
		{ 18, 134, 51 },
		{ 18, 135, 51 },
		{ 18, 136, 51 },
		{ 18, 137, 51 },
		{ 19, 79, 51 },
		{ 19, 80, 51 },
		{ 19, 81, 51 },
		{ 19, 82, 51 },
		{ 19, 83, 51 },
		{ 19, 84, 51 },
		{ 19, 85, 51 },
		{ 19, 86, 51 },
		{ 19, 87, 51 },
		{ 19, 88, 51 },
		{ 19, 89, 51 },
		{ 19, 90, 51 },
		{ 19, 91, 51 },
		{ 19, 92, 51 },
		{ 19, 93, 51 },
		{ 19, 94, 51 },
		{ 19, 95, 51 },
		{ 19, 96, 51 },
		{ 19, 97, 51 },
		{ 19, 98, 51 },
		{ 19, 99, 51 },
		{ 19, 100, 51 },
		{ 19, 101, 51 },
		{ 19, 102, 51 },
		{ 19, 103, 51 },
		{ 19, 104, 51 },
		{ 19, 105, 51 },
		{ 19, 106, 51 },
		{ 19, 107, 51 },
		{ 19, 108, 51 },
		{ 19, 109, 51 },
		{ 19, 110, 51 },
		{ 19, 111, 51 },
		{ 19, 112, 51 },
		{ 19, 113, 51 },
		{ 19, 114, 51 },
		{ 19, 115, 51 },
		{ 19, 116, 51 },
		{ 19, 117, 51 },
		{ 19, 118, 51 },
		{ 19, 119, 51 },
		{ 19, 120, 51 },
		{ 19, 121, 51 },
		{ 19, 122, 51 },
		{ 19, 123, 51 },
		{ 19, 124, 51 },
		{ 19, 125, 51 },
		{ 19, 126, 51 },
		{ 19, 127, 51 },
		{ 19, 128, 51 },
		{ 19, 129, 51 },
		{ 19, 130, 51 },
		{ 19, 131, 51 },
		{ 19, 132, 51 },
		{ 19, 133, 51 },
		{ 19, 134, 51 },
		{ 19, 135, 51 },
		{ 19, 136, 51 },
		{ 19, 137, 51 },
		{ 20, 80, 51 },
		{ 20, 81, 51 },
		{ 20, 82, 51 },
		{ 20, 83, 51 },
		{ 20, 84, 51 },
		{ 20, 85, 51 },
		{ 20, 86, 51 },
		{ 20, 87, 51 },
		{ 20, 88, 51 },
		{ 20, 89, 51 },
		{ 20, 90, 51 },
		{ 20, 91, 51 },
		{ 20, 92, 51 },
		{ 20, 93, 51 },
		{ 20, 94, 51 },
		{ 20, 95, 51 },
		{ 20, 96, 51 },
		{ 20, 97, 51 },
		{ 20, 98, 51 },
		{ 20, 99, 51 },
		{ 20, 100, 51 },
		{ 20, 101, 51 },
		{ 20, 102, 51 },
		{ 20, 103, 51 },
		{ 20, 104, 51 },
		{ 20, 105, 51 },
		{ 20, 106, 51 },
		{ 20, 107, 51 },
		{ 20, 108, 51 },
		{ 20, 109, 51 },
		{ 20, 110, 51 },
		{ 20, 111, 51 },
		{ 20, 112, 51 },
		{ 20, 113, 51 },
		{ 20, 114, 51 },
		{ 20, 115, 51 },
		{ 20, 116, 51 },
		{ 20, 117, 51 },
		{ 20, 118, 51 },
		{ 20, 119, 51 },
		{ 20, 120, 51 },
		{ 20, 121, 51 },
		{ 20, 122, 51 },
		{ 20, 123, 51 },
		{ 20, 124, 51 },
		{ 20, 125, 51 },
		{ 20, 126, 51 },
		{ 20, 127, 51 },
		{ 20, 128, 51 },
		{ 20, 129, 51 },
		{ 20, 130, 51 },
		{ 20, 131, 51 },
		{ 20, 132, 51 },
		{ 20, 133, 51 },
		{ 20, 134, 51 },
		{ 20, 135, 51 },
		{ 20, 136, 51 },
		{ 20, 137, 51 },
		{ 20, 138, 51 },
		{ 21, 81, 51 },
		{ 21, 82, 51 },
		{ 21, 83, 51 },
		{ 21, 84, 51 },
		{ 21, 85, 51 },
		{ 21, 86, 51 },
		{ 21, 87, 51 },
		{ 21, 88, 51 },
		{ 21, 89, 51 },
		{ 21, 90, 51 },
		{ 21, 91, 51 },
		{ 21, 92, 51 },
		{ 21, 93, 51 },
		{ 21, 94, 51 },
		{ 21, 95, 51 },
		{ 21, 96, 51 },
		{ 21, 97, 51 },
		{ 21, 98, 51 },
		{ 21, 99, 51 },
		{ 21, 100, 51 },
		{ 21, 101, 51 },
		{ 21, 102, 51 },
		{ 21, 103, 51 },
		{ 21, 104, 51 },
		{ 21, 105, 51 },
		{ 21, 106, 51 },
		{ 21, 107, 51 },
		{ 21, 108, 51 },
		{ 21, 109, 51 },
		{ 21, 110, 51 },
		{ 21, 111, 51 },
		{ 21, 112, 51 },
		{ 21, 113, 51 },
		{ 21, 114, 51 },
		{ 21, 115, 51 },
		{ 21, 116, 51 },
		{ 21, 117, 51 },
		{ 21, 118, 51 },
		{ 21, 119, 51 },
		{ 21, 120, 51 },
		{ 21, 121, 51 },
		{ 21, 122, 51 },
		{ 21, 123, 51 },
		{ 21, 124, 51 },
		{ 21, 125, 51 },
		{ 21, 126, 51 },
		{ 21, 127, 51 },
		{ 21, 128, 51 },
		{ 21, 129, 51 },
		{ 21, 130, 51 },
		{ 21, 131, 51 },
		{ 21, 132, 51 },
		{ 21, 133, 51 },
		{ 21, 134, 51 },
		{ 21, 135, 51 },
		{ 21, 136, 51 },
		{ 21, 137, 51 },
		{ 21, 138, 51 },
		{ 22, 81, 51 },
		{ 22, 82, 51 },
		{ 22, 83, 51 },
		{ 22, 84, 51 },
		{ 22, 85, 51 },
		{ 22, 86, 51 },
		{ 22, 87, 51 },
		{ 22, 88, 51 },
		{ 22, 89, 51 },
		{ 22, 90, 51 },
		{ 22, 91, 51 },
		{ 22, 92, 51 },
		{ 22, 93, 51 },
		{ 22, 94, 51 },
		{ 22, 95, 51 },
		{ 22, 96, 51 },
		{ 22, 97, 51 },
		{ 22, 98, 51 },
		{ 22, 99, 51 },
		{ 22, 100, 51 },
		{ 22, 101, 51 },
		{ 22, 102, 51 },
		{ 22, 103, 51 },
		{ 22, 104, 51 },
		{ 22, 105, 51 },
		{ 22, 106, 51 },
		{ 22, 107, 51 },
		{ 22, 108, 51 },
		{ 22, 109, 51 },
		{ 22, 110, 51 },
		{ 22, 111, 51 },
		{ 22, 112, 51 },
		{ 22, 113, 51 },
		{ 22, 114, 51 },
		{ 22, 115, 51 },
		{ 22, 116, 51 },
		{ 22, 117, 51 },
		{ 22, 118, 51 },
		{ 22, 119, 51 },
		{ 22, 120, 51 },
		{ 22, 121, 51 },
		{ 22, 122, 51 },
		{ 22, 123, 51 },
		{ 22, 124, 51 },
		{ 22, 125, 51 },
		{ 22, 126, 51 },
		{ 22, 127, 51 },
		{ 22, 128, 51 },
		{ 22, 129, 51 },
		{ 22, 130, 51 },
		{ 22, 131, 51 },
		{ 22, 132, 51 },
		{ 22, 133, 51 },
		{ 22, 134, 51 },
		{ 22, 135, 51 },
		{ 22, 136, 51 },
		{ 22, 137, 51 },
		{ 22, 138, 51 },
		{ 22, 139, 51 },
		{ 23, 82, 51 },
		{ 23, 83, 51 },
		{ 23, 84, 51 },
		{ 23, 85, 51 },
		{ 23, 86, 51 },
		{ 23, 87, 51 },
		{ 23, 88, 51 },
		{ 23, 89, 51 },
		{ 23, 90, 51 },
		{ 23, 91, 51 },
		{ 23, 92, 51 },
		{ 23, 93, 51 },
		{ 23, 94, 51 },
		{ 23, 95, 51 },
		{ 23, 96, 51 },
		{ 23, 97, 51 },
		{ 23, 98, 51 },
		{ 23, 99, 51 },
		{ 23, 100, 51 },
		{ 23, 101, 51 },
		{ 23, 102, 51 },
		{ 23, 103, 51 },
		{ 23, 104, 51 },
		{ 23, 105, 51 },
		{ 23, 106, 51 },
		{ 23, 107, 51 },
		{ 23, 108, 51 },
		{ 23, 109, 51 },
		{ 23, 110, 51 },
		{ 23, 111, 51 },
		{ 23, 112, 51 },
		{ 23, 113, 51 },
		{ 23, 114, 51 },
		{ 23, 115, 51 },
		{ 23, 116, 51 },
		{ 23, 117, 51 },
		{ 23, 118, 51 },
		{ 23, 119, 51 },
		{ 23, 120, 51 },
		{ 23, 121, 51 },
		{ 23, 122, 51 },
		{ 23, 123, 51 },
		{ 23, 124, 51 },
		{ 23, 125, 51 },
		{ 23, 126, 51 },
		{ 23, 127, 51 },
		{ 23, 128, 51 },
		{ 23, 129, 51 },
		{ 23, 130, 51 },
		{ 23, 131, 51 },
		{ 23, 132, 51 },
		{ 23, 133, 51 },
		{ 23, 134, 51 },
		{ 23, 135, 51 },
		{ 23, 136, 51 },
		{ 23, 137, 51 },
		{ 23, 138, 51 },
		{ 23, 139, 51 },
		{ 24, 83, 51 },
		{ 24, 84, 51 },
		{ 24, 85, 51 },
		{ 24, 86, 51 },
		{ 24, 87, 51 },
		{ 24, 88, 51 },
		{ 24, 89, 51 },
		{ 24, 90, 51 },
		{ 24, 91, 51 },
		{ 24, 92, 51 },
		{ 24, 93, 51 },
		{ 24, 94, 51 },
		{ 24, 95, 51 },
		{ 24, 96, 51 },
		{ 24, 97, 51 },
		{ 24, 98, 51 },
		{ 24, 99, 51 },
		{ 24, 100, 51 },
		{ 24, 101, 51 },
		{ 24, 102, 51 },
		{ 24, 103, 51 },
		{ 24, 104, 51 },
		{ 24, 105, 51 },
		{ 24, 106, 51 },
		{ 24, 107, 51 },
		{ 24, 108, 51 },
		{ 24, 109, 51 },
		{ 24, 110, 51 },
		{ 24, 111, 51 },
		{ 24, 112, 51 },
		{ 24, 113, 51 },
		{ 24, 114, 51 },
		{ 24, 115, 51 },
		{ 24, 116, 51 },
		{ 24, 117, 51 },
		{ 24, 118, 51 },
		{ 24, 119, 51 },
		{ 24, 120, 51 },
		{ 24, 121, 51 },
		{ 24, 122, 51 },
		{ 24, 123, 51 },
		{ 24, 124, 51 },
		{ 24, 125, 51 },
		{ 24, 126, 51 },
		{ 24, 127, 51 },
		{ 24, 128, 51 },
		{ 24, 129, 51 },
		{ 24, 130, 51 },
		{ 24, 131, 51 },
		{ 24, 132, 51 },
		{ 24, 133, 51 },
		{ 24, 134, 51 },
		{ 24, 135, 51 },
		{ 24, 136, 51 },
		{ 24, 137, 51 },
		{ 24, 138, 51 },
		{ 24, 139, 51 },
		{ 24, 140, 51 },
		{ 25, 84, 51 },
		{ 25, 85, 51 },
		{ 25, 86, 51 },
		{ 25, 87, 51 },
		{ 25, 88, 51 },
		{ 25, 89, 51 },
		{ 25, 90, 51 },
		{ 25, 91, 51 },
		{ 25, 92, 51 },
		{ 25, 93, 51 },
		{ 25, 94, 51 },
		{ 25, 95, 51 },
		{ 25, 96, 51 },
		{ 25, 97, 51 },
		{ 25, 98, 51 },
		{ 25, 99, 51 },
		{ 25, 100, 51 },
		{ 25, 101, 51 },
		{ 25, 102, 51 },
		{ 25, 103, 51 },
		{ 25, 104, 51 },
		{ 25, 105, 51 },
		{ 25, 106, 51 },
		{ 25, 107, 51 },
		{ 25, 108, 51 },
		{ 25, 109, 51 },
		{ 25, 110, 51 },
		{ 25, 111, 51 },
		{ 25, 112, 51 },
		{ 25, 113, 51 },
		{ 25, 114, 51 },
		{ 25, 115, 51 },
		{ 25, 116, 51 },
		{ 25, 117, 51 },
		{ 25, 118, 51 },
		{ 25, 119, 51 },
		{ 25, 120, 51 },
		{ 25, 121, 51 },
		{ 25, 122, 51 },
		{ 25, 123, 51 },
		{ 25, 124, 51 },
		{ 25, 125, 51 },
		{ 25, 126, 51 },
		{ 25, 127, 51 },
		{ 25, 128, 51 },
		{ 25, 129, 51 },
		{ 25, 130, 51 },
		{ 25, 131, 51 },
		{ 25, 132, 51 },
		{ 25, 133, 51 },
		{ 25, 134, 51 },
		{ 25, 135, 51 },
		{ 25, 136, 51 },
		{ 25, 137, 51 },
		{ 25, 138, 51 },
		{ 25, 139, 51 },
		{ 25, 140, 51 },
		{ 26, 84, 51 },
		{ 26, 85, 51 },
		{ 26, 86, 51 },
		{ 26, 87, 51 },
		{ 26, 88, 51 },
		{ 26, 89, 51 },
		{ 26, 90, 51 },
		{ 26, 91, 51 },
		{ 26, 92, 51 },
		{ 26, 93, 51 },
		{ 26, 94, 51 },
		{ 26, 95, 51 },
		{ 26, 96, 51 },
		{ 26, 97, 51 },
		{ 26, 98, 51 },
		{ 26, 99, 51 },
		{ 26, 100, 51 },
		{ 26, 101, 51 },
		{ 26, 102, 51 },
		{ 26, 103, 51 },
		{ 26, 104, 51 },
		{ 26, 105, 51 },
		{ 26, 106, 51 },
		{ 26, 107, 51 },
		{ 26, 108, 51 },
		{ 26, 109, 51 },
		{ 26, 110, 51 },
		{ 26, 111, 51 },
		{ 26, 112, 51 },
		{ 26, 113, 51 },
		{ 26, 114, 51 },
		{ 26, 115, 51 },
		{ 26, 116, 51 },
		{ 26, 117, 51 },
		{ 26, 118, 51 },
		{ 26, 119, 51 },
		{ 26, 120, 51 },
		{ 26, 121, 51 },
		{ 26, 122, 51 },
		{ 26, 123, 51 },
		{ 26, 124, 51 },
		{ 26, 125, 51 },
		{ 26, 126, 51 },
		{ 26, 127, 51 },
		{ 26, 128, 51 },
		{ 26, 129, 51 },
		{ 26, 130, 51 },
		{ 26, 131, 51 },
		{ 26, 132, 51 },
		{ 26, 133, 51 },
		{ 26, 134, 51 },
		{ 26, 135, 51 },
		{ 26, 136, 51 },
		{ 26, 137, 51 },
		{ 26, 138, 51 },
		{ 26, 139, 51 },
		{ 26, 140, 51 },
		{ 26, 141, 51 },
		{ 27, 85, 51 },
		{ 27, 86, 51 },
		{ 27, 87, 51 },
		{ 27, 88, 51 },
		{ 27, 89, 51 },
		{ 27, 90, 51 },
		{ 27, 91, 51 },
		{ 27, 92, 51 },
		{ 27, 93, 51 },
		{ 27, 94, 51 },
		{ 27, 95, 51 },
		{ 27, 96, 51 },
		{ 27, 97, 51 },
		{ 27, 98, 51 },
		{ 27, 99, 51 },
		{ 27, 100, 51 },
		{ 27, 101, 51 },
		{ 27, 102, 51 },
		{ 27, 103, 51 },
		{ 27, 104, 51 },
		{ 27, 105, 51 },
		{ 27, 106, 51 },
		{ 27, 107, 51 },
		{ 27, 108, 51 },
		{ 27, 109, 51 },
		{ 27, 110, 51 },
		{ 27, 111, 51 },
		{ 27, 112, 51 },
		{ 27, 113, 51 },
		{ 27, 114, 51 },
		{ 27, 115, 51 },
		{ 27, 116, 51 },
		{ 27, 117, 51 },
		{ 27, 118, 51 },
		{ 27, 119, 51 },
		{ 27, 120, 51 },
		{ 27, 121, 51 },
		{ 27, 122, 51 },
		{ 27, 123, 51 },
		{ 27, 124, 51 },
		{ 27, 125, 51 },
		{ 27, 126, 51 },
		{ 27, 127, 51 },
		{ 27, 128, 51 },
		{ 27, 129, 51 },
		{ 27, 130, 51 },
		{ 27, 131, 51 },
		{ 27, 132, 51 },
		{ 27, 133, 51 },
		{ 27, 134, 51 },
		{ 27, 135, 51 },
		{ 27, 136, 51 },
		{ 27, 137, 51 },
		{ 27, 138, 51 },
		{ 27, 139, 51 },
		{ 27, 140, 51 },
		{ 27, 141, 51 },
		{ 28, 86, 51 },
		{ 28, 87, 51 },
		{ 28, 88, 51 },
		{ 28, 89, 51 },
		{ 28, 90, 51 },
		{ 28, 91, 51 },
		{ 28, 92, 51 },
		{ 28, 93, 51 },
		{ 28, 94, 51 },
		{ 28, 95, 51 },
		{ 28, 96, 51 },
		{ 28, 97, 51 },
		{ 28, 98, 51 },
		{ 28, 99, 51 },
		{ 28, 100, 51 },
		{ 28, 101, 51 },
		{ 28, 102, 51 },
		{ 28, 103, 51 },
		{ 28, 104, 51 },
		{ 28, 105, 51 },
		{ 28, 106, 51 },
		{ 28, 107, 51 },
		{ 28, 108, 51 },
		{ 28, 109, 51 },
		{ 28, 110, 51 },
		{ 28, 111, 51 },
		{ 28, 112, 51 },
		{ 28, 113, 51 },
		{ 28, 114, 51 },
		{ 28, 115, 51 },
		{ 28, 116, 51 },
		{ 28, 117, 51 },
		{ 28, 118, 51 },
		{ 28, 119, 51 },
		{ 28, 120, 51 },
		{ 28, 121, 51 },
		{ 28, 122, 51 },
		{ 28, 123, 51 },
		{ 28, 124, 51 },
		{ 28, 125, 51 },
		{ 28, 126, 51 },
		{ 28, 127, 51 },
		{ 28, 128, 51 },
		{ 28, 129, 51 },
		{ 28, 130, 51 },
		{ 28, 131, 51 },
		{ 28, 132, 51 },
		{ 28, 133, 51 },
		{ 28, 134, 51 },
		{ 28, 135, 51 },
		{ 28, 136, 51 },
		{ 28, 137, 51 },
		{ 28, 138, 51 },
		{ 28, 139, 51 },
		{ 28, 140, 51 },
		{ 28, 141, 51 },
		{ 28, 142, 51 },
		{ 29, 87, 51 },
		{ 29, 88, 51 },
		{ 29, 89, 51 },
		{ 29, 90, 51 },
		{ 29, 91, 51 },
		{ 29, 92, 51 },
		{ 29, 93, 51 },
		{ 29, 94, 51 },
		{ 29, 95, 51 },
		{ 29, 96, 51 },
		{ 29, 97, 51 },
		{ 29, 98, 51 },
		{ 29, 99, 51 },
		{ 29, 100, 51 },
		{ 29, 101, 51 },
		{ 29, 102, 51 },
		{ 29, 103, 51 },
		{ 29, 104, 51 },
		{ 29, 105, 51 },
		{ 29, 106, 51 },
		{ 29, 107, 51 },
		{ 29, 108, 51 },
		{ 29, 109, 51 },
		{ 29, 110, 51 },
		{ 29, 111, 51 },
		{ 29, 112, 51 },
		{ 29, 113, 51 },
		{ 29, 114, 51 },
		{ 29, 115, 51 },
		{ 29, 116, 51 },
		{ 29, 117, 51 },
		{ 29, 118, 51 },
		{ 29, 119, 51 },
		{ 29, 120, 51 },
		{ 29, 121, 51 },
		{ 29, 122, 51 },
		{ 29, 123, 51 },
		{ 29, 124, 51 },
		{ 29, 125, 51 },
		{ 29, 126, 51 },
		{ 29, 127, 51 },
		{ 29, 128, 51 },
		{ 29, 129, 51 },
		{ 29, 130, 51 },
		{ 29, 131, 51 },
		{ 29, 132, 51 },
		{ 29, 133, 51 },
		{ 29, 134, 51 },
		{ 29, 135, 51 },
		{ 29, 136, 51 },
		{ 29, 137, 51 },
		{ 29, 138, 51 },
		{ 29, 139, 51 },
		{ 29, 140, 51 },
		{ 29, 141, 51 },
		{ 29, 142, 51 },
		{ 30, 87, 51 },
		{ 30, 88, 51 },
		{ 30, 89, 51 },
		{ 30, 90, 51 },
		{ 30, 91, 51 },
		{ 30, 92, 51 },
		{ 30, 93, 51 },
		{ 30, 94, 51 },
		{ 30, 95, 51 },
		{ 30, 96, 51 },
		{ 30, 97, 51 },
		{ 30, 98, 51 },
		{ 30, 99, 51 },
		{ 30, 100, 51 },
		{ 30, 101, 51 },
		{ 30, 102, 51 },
		{ 30, 103, 51 },
		{ 30, 104, 51 },
		{ 30, 105, 51 },
		{ 30, 106, 51 },
		{ 30, 107, 51 },
		{ 30, 108, 51 },
		{ 30, 109, 51 },
		{ 30, 110, 51 },
		{ 30, 111, 51 },
		{ 30, 112, 51 },
		{ 30, 113, 51 },
		{ 30, 114, 51 },
		{ 30, 115, 51 },
		{ 30, 116, 51 },
		{ 30, 117, 51 },
		{ 30, 118, 51 },
		{ 30, 119, 51 },
		{ 30, 120, 51 },
		{ 30, 121, 51 },
		{ 30, 122, 51 },
		{ 30, 123, 51 },
		{ 30, 124, 51 },
		{ 30, 125, 51 },
		{ 30, 126, 51 },
		{ 30, 127, 51 },
		{ 30, 128, 51 },
		{ 30, 129, 51 },
		{ 30, 130, 51 },
		{ 30, 131, 51 },
		{ 30, 132, 51 },
		{ 30, 133, 51 },
		{ 30, 134, 51 },
		{ 30, 135, 51 },
		{ 30, 136, 51 },
		{ 30, 137, 51 },
		{ 30, 138, 51 },
		{ 30, 139, 51 },
		{ 30, 140, 51 },
		{ 30, 141, 51 },
		{ 30, 142, 51 },
		{ 30, 143, 51 },
		{ 31, 88, 51 },
		{ 31, 89, 51 },
		{ 31, 90, 51 },
		{ 31, 91, 51 },
		{ 31, 92, 51 },
		{ 31, 93, 51 },
		{ 31, 94, 51 },
		{ 31, 95, 51 },
		{ 31, 96, 51 },
		{ 31, 97, 51 },
		{ 31, 98, 51 },
		{ 31, 99, 51 },
		{ 31, 100, 51 },
		{ 31, 101, 51 },
		{ 31, 102, 51 },
		{ 31, 103, 51 },
		{ 31, 104, 51 },
		{ 31, 105, 51 },
		{ 31, 106, 51 },
		{ 31, 107, 51 },
		{ 31, 108, 51 },
		{ 31, 109, 51 },
		{ 31, 110, 51 },
		{ 31, 111, 51 },
		{ 31, 112, 51 },
		{ 31, 113, 51 },
		{ 31, 114, 51 },
		{ 31, 115, 51 },
		{ 31, 116, 51 },
		{ 31, 117, 51 },
		{ 31, 118, 51 },
		{ 31, 119, 51 },
		{ 31, 120, 51 },
		{ 31, 121, 51 },
		{ 31, 122, 51 },
		{ 31, 123, 51 },
		{ 31, 124, 51 },
		{ 31, 125, 51 },
		{ 31, 126, 51 },
		{ 31, 127, 51 },
		{ 31, 128, 51 },
		{ 31, 129, 51 },
		{ 31, 130, 51 },
		{ 31, 131, 51 },
		{ 31, 132, 51 },
		{ 31, 133, 51 },
		{ 31, 134, 51 },
		{ 31, 135, 51 },
		{ 31, 136, 51 },
		{ 31, 137, 51 },
		{ 31, 138, 51 },
		{ 31, 139, 51 },
		{ 31, 140, 51 },
		{ 31, 141, 51 },
		{ 31, 142, 51 },
		{ 31, 143, 51 },
		{ 32, 89, 51 },
		{ 32, 90, 51 },
		{ 32, 91, 51 },
		{ 32, 92, 51 },
		{ 32, 93, 51 },
		{ 32, 94, 51 },
		{ 32, 95, 51 },
		{ 32, 96, 51 },
		{ 32, 97, 51 },
		{ 32, 98, 51 },
		{ 32, 99, 51 },
		{ 32, 100, 51 },
		{ 32, 101, 51 },
		{ 32, 102, 51 },
		{ 32, 103, 51 },
		{ 32, 104, 51 },
		{ 32, 105, 51 },
		{ 32, 106, 51 },
		{ 32, 107, 51 },
		{ 32, 108, 51 },
		{ 32, 109, 51 },
		{ 32, 110, 51 },
		{ 32, 111, 51 },
		{ 32, 112, 51 },
		{ 32, 113, 51 },
		{ 32, 114, 51 },
		{ 32, 115, 51 },
		{ 32, 116, 51 },
		{ 32, 117, 51 },
		{ 32, 118, 51 },
		{ 32, 119, 51 },
		{ 32, 120, 51 },
		{ 32, 121, 51 },
		{ 32, 122, 51 },
		{ 32, 123, 51 },
		{ 32, 124, 51 },
		{ 32, 125, 51 },
		{ 32, 126, 51 },
		{ 32, 127, 51 },
		{ 32, 128, 51 },
		{ 32, 129, 51 },
		{ 32, 130, 51 },
		{ 32, 131, 51 },
		{ 32, 132, 51 },
		{ 32, 133, 51 },
		{ 32, 134, 51 },
		{ 32, 135, 51 },
		{ 32, 136, 51 },
		{ 32, 137, 51 },
		{ 32, 138, 51 },
		{ 32, 139, 51 },
		{ 32, 140, 51 },
		{ 32, 141, 51 },
		{ 32, 142, 51 },
		{ 32, 143, 51 },
		{ 32, 144, 51 },
		{ 33, 90, 51 },
		{ 33, 91, 51 },
		{ 33, 92, 51 },
		{ 33, 93, 51 },
		{ 33, 94, 51 },
		{ 33, 95, 51 },
		{ 33, 96, 51 },
		{ 33, 97, 51 },
		{ 33, 98, 51 },
		{ 33, 99, 51 },
		{ 33, 100, 51 },
		{ 33, 101, 51 },
		{ 33, 102, 51 },
		{ 33, 103, 51 },
		{ 33, 104, 51 },
		{ 33, 105, 51 },
		{ 33, 106, 51 },
		{ 33, 107, 51 },
		{ 33, 108, 51 },
		{ 33, 109, 51 },
		{ 33, 110, 51 },
		{ 33, 111, 51 },
		{ 33, 112, 51 },
		{ 33, 113, 51 },
		{ 33, 114, 51 },
		{ 33, 115, 51 },
		{ 33, 116, 51 },
		{ 33, 117, 51 },
		{ 33, 118, 51 },
		{ 33, 119, 51 },
		{ 33, 120, 51 },
		{ 33, 121, 51 },
		{ 33, 122, 51 },
		{ 33, 123, 51 },
		{ 33, 124, 51 },
		{ 33, 125, 51 },
		{ 33, 126, 51 },
		{ 33, 127, 51 },
		{ 33, 128, 51 },
		{ 33, 129, 51 },
		{ 33, 130, 51 },
		{ 33, 131, 51 },
		{ 33, 132, 51 },
		{ 33, 133, 51 },
		{ 33, 134, 51 },
		{ 33, 135, 51 },
		{ 33, 136, 51 },
		{ 33, 137, 51 },
		{ 33, 138, 51 },
		{ 33, 139, 51 },
		{ 33, 140, 51 },
		{ 33, 141, 51 },
		{ 33, 142, 51 },
		{ 33, 143, 51 },
		{ 33, 144, 51 },
		{ 34, 90, 51 },
		{ 34, 91, 51 },
		{ 34, 92, 51 },
		{ 34, 93, 51 },
		{ 34, 94, 51 },
		{ 34, 95, 51 },
		{ 34, 96, 51 },
		{ 34, 97, 51 },
		{ 34, 98, 51 },
		{ 34, 99, 51 },
		{ 34, 100, 51 },
		{ 34, 101, 51 },
		{ 34, 102, 51 },
		{ 34, 103, 51 },
		{ 34, 104, 51 },
		{ 34, 105, 51 },
		{ 34, 106, 51 },
		{ 34, 107, 51 },
		{ 34, 108, 51 },
		{ 34, 109, 51 },
		{ 34, 110, 51 },
		{ 34, 111, 51 },
		{ 34, 112, 51 },
		{ 34, 113, 51 },
		{ 34, 114, 51 },
		{ 34, 115, 51 },
		{ 34, 116, 51 },
		{ 34, 117, 51 },
		{ 34, 118, 51 },
		{ 34, 119, 51 },
		{ 34, 120, 51 },
		{ 34, 121, 51 },
		{ 34, 122, 51 },
		{ 34, 123, 51 },
		{ 34, 124, 51 },
		{ 34, 125, 51 },
		{ 34, 126, 51 },
		{ 34, 127, 51 },
		{ 34, 128, 51 },
		{ 34, 129, 51 },
		{ 34, 130, 51 },
		{ 34, 131, 51 },
		{ 34, 132, 51 },
		{ 34, 133, 51 },
		{ 34, 134, 51 },
		{ 34, 135, 51 },
		{ 34, 136, 51 },
		{ 34, 137, 51 },
		{ 34, 138, 51 },
		{ 34, 139, 51 },
		{ 34, 140, 51 },
		{ 34, 141, 51 },
		{ 34, 142, 51 },
		{ 34, 143, 51 },
		{ 34, 144, 51 },
		{ 34, 145, 51 },
		{ 35, 91, 51 },
		{ 35, 92, 51 },
		{ 35, 93, 51 },
		{ 35, 94, 51 },
		{ 35, 95, 51 },
		{ 35, 96, 51 },
		{ 35, 97, 51 },
		{ 35, 98, 51 },
		{ 35, 99, 51 },
		{ 35, 100, 51 },
		{ 35, 101, 51 },
		{ 35, 102, 51 },
		{ 35, 103, 51 },
		{ 35, 104, 51 },
		{ 35, 105, 51 },
		{ 35, 106, 51 },
		{ 35, 107, 51 },
		{ 35, 108, 51 },
		{ 35, 109, 51 },
		{ 35, 110, 51 },
		{ 35, 111, 51 },
		{ 35, 112, 51 },
		{ 35, 113, 51 },
		{ 35, 114, 51 },
		{ 35, 115, 51 },
		{ 35, 116, 51 },
		{ 35, 117, 51 },
		{ 35, 118, 51 },
		{ 35, 119, 51 },
		{ 35, 120, 51 },
		{ 35, 121, 51 },
		{ 35, 122, 51 },
		{ 35, 123, 51 },
		{ 35, 124, 51 },
		{ 35, 125, 51 },
		{ 35, 126, 51 },
		{ 35, 127, 51 },
		{ 35, 128, 51 },
		{ 35, 129, 51 },
		{ 35, 130, 51 },
		{ 35, 131, 51 },
		{ 35, 132, 51 },
		{ 35, 133, 51 },
		{ 35, 134, 51 },
		{ 35, 135, 51 },
		{ 35, 136, 51 },
		{ 35, 137, 51 },
		{ 35, 138, 51 },
		{ 35, 139, 51 },
		{ 35, 140, 51 },
		{ 35, 141, 51 },
		{ 35, 142, 51 },
		{ 35, 143, 51 },
		{ 35, 144, 51 },
		{ 35, 145, 51 },
		{ 36, 92, 51 },
		{ 36, 93, 51 },
		{ 36, 94, 51 },
		{ 36, 95, 51 },
		{ 36, 96, 51 },
		{ 36, 97, 51 },
		{ 36, 98, 51 },
		{ 36, 99, 51 },
		{ 36, 100, 51 },
		{ 36, 101, 51 },
		{ 36, 102, 51 },
		{ 36, 103, 51 },
		{ 36, 104, 51 },
		{ 36, 105, 51 },
		{ 36, 106, 51 },
		{ 36, 107, 51 },
		{ 36, 108, 51 },
		{ 36, 109, 51 },
		{ 36, 110, 51 },
		{ 36, 111, 51 },
		{ 36, 112, 51 },
		{ 36, 113, 51 },
		{ 36, 114, 51 },
		{ 36, 115, 51 },
		{ 36, 116, 51 },
		{ 36, 117, 51 },
		{ 36, 118, 51 },
		{ 36, 119, 51 },
		{ 36, 120, 51 },
		{ 36, 121, 51 },
		{ 36, 122, 51 },
		{ 36, 123, 51 },
		{ 36, 124, 51 },
		{ 36, 125, 51 },
		{ 36, 126, 51 },
		{ 36, 127, 51 },
		{ 36, 128, 51 },
		{ 36, 129, 51 },
		{ 36, 130, 51 },
		{ 36, 131, 51 },
		{ 36, 132, 51 },
		{ 36, 133, 51 },
		{ 36, 134, 51 },
		{ 36, 135, 51 },
		{ 36, 136, 51 },
		{ 36, 137, 51 },
		{ 36, 138, 51 },
		{ 36, 139, 51 },
		{ 36, 140, 51 },
		{ 36, 141, 51 },
		{ 36, 142, 51 },
		{ 36, 143, 51 },
		{ 36, 144, 51 },
		{ 36, 145, 51 },
		{ 36, 146, 51 },
		{ 37, 93, 51 },
		{ 37, 94, 51 },
		{ 37, 95, 51 },
		{ 37, 96, 51 },
		{ 37, 97, 51 },
		{ 37, 98, 51 },
		{ 37, 99, 51 },
		{ 37, 100, 51 },
		{ 37, 101, 51 },
		{ 37, 102, 51 },
		{ 37, 103, 51 },
		{ 37, 104, 51 },
		{ 37, 105, 51 },
		{ 37, 106, 51 },
		{ 37, 107, 51 },
		{ 37, 108, 51 },
		{ 37, 109, 51 },
		{ 37, 110, 51 },
		{ 37, 111, 51 },
		{ 37, 112, 51 },
		{ 37, 113, 51 },
		{ 37, 114, 51 },
		{ 37, 115, 51 },
		{ 37, 116, 51 },
		{ 37, 117, 51 },
		{ 37, 118, 51 },
		{ 37, 119, 51 },
		{ 37, 120, 51 },
		{ 37, 121, 51 },
		{ 37, 122, 51 },
		{ 37, 123, 51 },
		{ 37, 124, 51 },
		{ 37, 125, 51 },
		{ 37, 126, 51 },
		{ 37, 127, 51 },
		{ 37, 128, 51 },
		{ 37, 129, 51 },
		{ 37, 130, 51 },
		{ 37, 131, 51 },
		{ 37, 132, 51 },
		{ 37, 133, 51 },
		{ 37, 134, 51 },
		{ 37, 135, 51 },
		{ 37, 136, 51 },
		{ 37, 137, 51 },
		{ 37, 138, 51 },
		{ 37, 139, 51 },
		{ 37, 140, 51 },
		{ 37, 141, 51 },
		{ 37, 142, 51 },
		{ 37, 143, 51 },
		{ 37, 144, 51 },
		{ 37, 145, 51 },
		{ 37, 146, 51 },
		{ 38, 93, 51 },
		{ 38, 94, 51 },
		{ 38, 95, 51 },
		{ 38, 96, 51 },
		{ 38, 97, 51 },
		{ 38, 98, 51 },
		{ 38, 99, 51 },
		{ 38, 100, 51 },
		{ 38, 101, 51 },
		{ 38, 102, 51 },
		{ 38, 103, 51 },
		{ 38, 104, 51 },
		{ 38, 105, 51 },
		{ 38, 106, 51 },
		{ 38, 107, 51 },
		{ 38, 108, 51 },
		{ 38, 109, 51 },
		{ 38, 110, 51 },
		{ 38, 111, 51 },
		{ 38, 112, 51 },
		{ 38, 113, 51 },
		{ 38, 114, 51 },
		{ 38, 115, 51 },
		{ 38, 116, 51 },
		{ 38, 117, 51 },
		{ 38, 118, 51 },
		{ 38, 119, 51 },
		{ 38, 120, 51 },
		{ 38, 121, 51 },
		{ 38, 122, 51 },
		{ 38, 123, 51 },
		{ 38, 124, 51 },
		{ 38, 125, 51 },
		{ 38, 126, 51 },
		{ 38, 127, 51 },
		{ 38, 128, 51 },
		{ 38, 129, 51 },
		{ 38, 130, 51 },
		{ 38, 131, 51 },
		{ 38, 132, 51 },
		{ 38, 133, 51 },
		{ 38, 134, 51 },
		{ 38, 135, 51 },
		{ 38, 136, 51 },
		{ 38, 137, 51 },
		{ 38, 138, 51 },
		{ 38, 139, 51 },
		{ 38, 140, 51 },
		{ 38, 141, 51 },
		{ 38, 142, 51 },
		{ 38, 143, 51 },
		{ 38, 144, 51 },
		{ 38, 145, 51 },
		{ 38, 146, 51 },
		{ 38, 147, 51 },
		{ 39, 94, 51 },
		{ 39, 95, 51 },
		{ 39, 96, 51 },
		{ 39, 97, 51 },
		{ 39, 98, 51 },
		{ 39, 99, 51 },
		{ 39, 100, 51 },
		{ 39, 101, 51 },
		{ 39, 102, 51 },
		{ 39, 103, 51 },
		{ 39, 104, 51 },
		{ 39, 105, 51 },
		{ 39, 106, 51 },
		{ 39, 107, 51 },
		{ 39, 108, 51 },
		{ 39, 109, 51 },
		{ 39, 110, 51 },
		{ 39, 111, 51 },
		{ 39, 112, 51 },
		{ 39, 113, 51 },
		{ 39, 114, 51 },
		{ 39, 115, 51 },
		{ 39, 116, 51 },
		{ 39, 117, 51 },
		{ 39, 118, 51 },
		{ 39, 119, 51 },
		{ 39, 120, 51 },
		{ 39, 121, 51 },
		{ 39, 122, 51 },
		{ 39, 123, 51 },
		{ 39, 124, 51 },
		{ 39, 125, 51 },
		{ 39, 126, 51 },
		{ 39, 127, 51 },
		{ 39, 128, 51 },
		{ 39, 129, 51 },
		{ 39, 130, 51 },
		{ 39, 131, 51 },
		{ 39, 132, 51 },
		{ 39, 133, 51 },
		{ 39, 134, 51 },
		{ 39, 135, 51 },
		{ 39, 136, 51 },
		{ 39, 137, 51 },
		{ 39, 138, 51 },
		{ 39, 139, 51 },
		{ 39, 140, 51 },
		{ 39, 141, 51 },
		{ 39, 142, 51 },
		{ 39, 143, 51 },
		{ 39, 144, 51 },
		{ 39, 145, 51 },
		{ 39, 146, 51 },
		{ 39, 147, 51 },
		{ 40, 95, 51 },
		{ 40, 96, 51 },
		{ 40, 97, 51 },
		{ 40, 98, 51 },
		{ 40, 99, 51 },
		{ 40, 100, 51 },
		{ 40, 101, 51 },
		{ 40, 102, 51 },
		{ 40, 103, 51 },
		{ 40, 104, 51 },
		{ 40, 105, 51 },
		{ 40, 106, 51 },
		{ 40, 107, 51 },
		{ 40, 108, 51 },
		{ 40, 109, 51 },
		{ 40, 110, 51 },
		{ 40, 111, 51 },
		{ 40, 112, 51 },
		{ 40, 113, 51 },
		{ 40, 114, 51 },
		{ 40, 115, 51 },
		{ 40, 116, 51 },
		{ 40, 117, 51 },
		{ 40, 118, 51 },
		{ 40, 119, 51 },
		{ 40, 120, 51 },
		{ 40, 121, 51 },
		{ 40, 122, 51 },
		{ 40, 123, 51 },
		{ 40, 124, 51 },
		{ 40, 125, 51 },
		{ 40, 126, 51 },
		{ 40, 127, 51 },
		{ 40, 128, 51 },
		{ 40, 129, 51 },
		{ 40, 130, 51 },
		{ 40, 131, 51 },
		{ 40, 132, 51 },
		{ 40, 133, 51 },
		{ 40, 134, 51 },
		{ 40, 135, 51 },
		{ 40, 136, 51 },
		{ 40, 137, 51 },
		{ 40, 138, 51 },
		{ 40, 139, 51 },
		{ 40, 140, 51 },
		{ 40, 141, 51 },
		{ 40, 142, 51 },
		{ 40, 143, 51 },
		{ 40, 144, 51 },
		{ 40, 145, 51 },
		{ 40, 146, 51 },
		{ 40, 147, 51 },
		{ 40, 148, 51 },
		{ 41, 96, 51 },
		{ 41, 97, 51 },
		{ 41, 98, 51 },
		{ 41, 99, 51 },
		{ 41, 100, 51 },
		{ 41, 101, 51 },
		{ 41, 102, 51 },
		{ 41, 103, 51 },
		{ 41, 104, 51 },
		{ 41, 105, 51 },
		{ 41, 106, 51 },
		{ 41, 107, 51 },
		{ 41, 108, 51 },
		{ 41, 109, 51 },
		{ 41, 110, 51 },
		{ 41, 111, 51 },
		{ 41, 112, 51 },
		{ 41, 113, 51 },
		{ 41, 114, 51 },
		{ 41, 115, 51 },
		{ 41, 116, 51 },
		{ 41, 117, 51 },
		{ 41, 118, 51 },
		{ 41, 119, 51 },
		{ 41, 120, 51 },
		{ 41, 121, 51 },
		{ 41, 122, 51 },
		{ 41, 123, 51 },
		{ 41, 124, 51 },
		{ 41, 125, 51 },
		{ 41, 126, 51 },
		{ 41, 127, 51 },
		{ 41, 128, 51 },
		{ 41, 129, 51 },
		{ 41, 130, 51 },
		{ 41, 131, 51 },
		{ 41, 132, 51 },
		{ 41, 133, 51 },
		{ 41, 134, 51 },
		{ 41, 135, 51 },
		{ 41, 136, 51 },
		{ 41, 137, 51 },
		{ 41, 138, 51 },
		{ 41, 139, 51 },
		{ 41, 140, 51 },
		{ 41, 141, 51 },
		{ 41, 142, 51 },
		{ 41, 143, 51 },
		{ 41, 144, 51 },
		{ 41, 145, 51 },
		{ 41, 146, 51 },
		{ 41, 147, 51 },
		{ 41, 148, 51 },
		{ 42, 96, 51 },
		{ 42, 97, 51 },
		{ 42, 98, 51 },
		{ 42, 99, 51 },
		{ 42, 100, 51 },
		{ 42, 101, 51 },
		{ 42, 102, 51 },
		{ 42, 103, 51 },
		{ 42, 104, 51 },
		{ 42, 105, 51 },
		{ 42, 106, 51 },
		{ 42, 107, 51 },
		{ 42, 108, 51 },
		{ 42, 109, 51 },
		{ 42, 110, 51 },
		{ 42, 111, 51 },
		{ 42, 112, 51 },
		{ 42, 113, 51 },
		{ 42, 114, 51 },
		{ 42, 115, 51 },
		{ 42, 116, 51 },
		{ 42, 117, 51 },
		{ 42, 118, 51 },
		{ 42, 119, 51 },
		{ 42, 120, 51 },
		{ 42, 121, 51 },
		{ 42, 122, 51 },
		{ 42, 123, 51 },
		{ 42, 124, 51 },
		{ 42, 125, 51 },
		{ 42, 126, 51 },
		{ 42, 127, 51 },
		{ 42, 128, 51 },
		{ 42, 129, 51 },
		{ 42, 130, 51 },
		{ 42, 131, 51 },
		{ 42, 132, 51 },
		{ 42, 133, 51 },
		{ 42, 134, 51 },
		{ 42, 135, 51 },
		{ 42, 136, 51 },
		{ 42, 137, 51 },
		{ 42, 138, 51 },
		{ 42, 139, 51 },
		{ 42, 140, 51 },
		{ 42, 141, 51 },
		{ 42, 142, 51 },
		{ 42, 143, 51 },
		{ 42, 144, 51 },
		{ 42, 145, 51 },
		{ 42, 146, 51 },
		{ 42, 147, 51 },
		{ 42, 148, 51 },
		{ 42, 149, 51 },
		{ 43, 97, 51 },
		{ 43, 98, 51 },
		{ 43, 99, 51 },
		{ 43, 100, 51 },
		{ 43, 101, 51 },
		{ 43, 102, 51 },
		{ 43, 103, 51 },
		{ 43, 104, 51 },
		{ 43, 105, 51 },
		{ 43, 106, 51 },
		{ 43, 107, 51 },
		{ 43, 108, 51 },
		{ 43, 109, 51 },
		{ 43, 110, 51 },
		{ 43, 111, 51 },
		{ 43, 112, 51 },
		{ 43, 113, 51 },
		{ 43, 114, 51 },
		{ 43, 115, 51 },
		{ 43, 116, 51 },
		{ 43, 117, 51 },
		{ 43, 118, 51 },
		{ 43, 119, 51 },
		{ 43, 120, 51 },
		{ 43, 121, 51 },
		{ 43, 122, 51 },
		{ 43, 123, 51 },
		{ 43, 124, 51 },
		{ 43, 125, 51 },
		{ 43, 126, 51 },
		{ 43, 127, 51 },
		{ 43, 128, 51 },
		{ 43, 129, 51 },
		{ 43, 130, 51 },
		{ 43, 131, 51 },
		{ 43, 132, 51 },
		{ 43, 133, 51 },
		{ 43, 134, 51 },
		{ 43, 135, 51 },
		{ 43, 136, 51 },
		{ 43, 137, 51 },
		{ 43, 138, 51 },
		{ 43, 139, 51 },
		{ 43, 140, 51 },
		{ 43, 141, 51 },
		{ 43, 142, 51 },
		{ 43, 143, 51 },
		{ 43, 144, 51 },
		{ 43, 145, 51 },
		{ 43, 146, 51 },
		{ 43, 147, 51 },
		{ 43, 148, 51 },
		{ 43, 149, 51 },
		{ 44, 98, 51 },
		{ 44, 99, 51 },
		{ 44, 100, 51 },
		{ 44, 101, 51 },
		{ 44, 102, 51 },
		{ 44, 103, 51 },
		{ 44, 104, 51 },
		{ 44, 105, 51 },
		{ 44, 106, 51 },
		{ 44, 107, 51 },
		{ 44, 108, 51 },
		{ 44, 109, 51 },
		{ 44, 110, 51 },
		{ 44, 111, 51 },
		{ 44, 112, 51 },
		{ 44, 113, 51 },
		{ 44, 114, 51 },
		{ 44, 115, 51 },
		{ 44, 116, 51 },
		{ 44, 117, 51 },
		{ 44, 118, 51 },
		{ 44, 119, 51 },
		{ 44, 120, 51 },
		{ 44, 121, 51 },
		{ 44, 122, 51 },
		{ 44, 123, 51 },
		{ 44, 124, 51 },
		{ 44, 125, 51 },
		{ 44, 126, 51 },
		{ 44, 127, 51 },
		{ 44, 128, 51 },
		{ 44, 129, 51 },
		{ 44, 130, 51 },
		{ 44, 131, 51 },
		{ 44, 132, 51 },
		{ 44, 133, 51 },
		{ 44, 134, 51 },
		{ 44, 135, 51 },
		{ 44, 136, 51 },
		{ 44, 137, 51 },
		{ 44, 138, 51 },
		{ 44, 139, 51 },
		{ 44, 140, 51 },
		{ 44, 141, 51 },
		{ 44, 142, 51 },
		{ 44, 143, 51 },
		{ 44, 144, 51 },
		{ 44, 145, 51 },
		{ 44, 146, 51 },
		{ 44, 147, 51 },
		{ 44, 148, 51 },
		{ 44, 149, 51 },
		{ 44, 150, 51 },
		{ 45, 99, 51 },
		{ 45, 100, 51 },
		{ 45, 101, 51 },
		{ 45, 102, 51 },
		{ 45, 103, 51 },
		{ 45, 104, 51 },
		{ 45, 105, 51 },
		{ 45, 106, 51 },
		{ 45, 107, 51 },
		{ 45, 108, 51 },
		{ 45, 109, 51 },
		{ 45, 110, 51 },
		{ 45, 111, 51 },
		{ 45, 112, 51 },
		{ 45, 113, 51 },
		{ 45, 114, 51 },
		{ 45, 115, 51 },
		{ 45, 116, 51 },
		{ 45, 117, 51 },
		{ 45, 118, 51 },
		{ 45, 119, 51 },
		{ 45, 120, 51 },
		{ 45, 121, 51 },
		{ 45, 122, 51 },
		{ 45, 123, 51 },
		{ 45, 124, 51 },
		{ 45, 125, 51 },
		{ 45, 126, 51 },
		{ 45, 127, 51 },
		{ 45, 128, 51 },
		{ 45, 129, 51 },
		{ 45, 130, 51 },
		{ 45, 131, 51 },
		{ 45, 132, 51 },
		{ 45, 133, 51 },
		{ 45, 134, 51 },
		{ 45, 135, 51 },
		{ 45, 136, 51 },
		{ 45, 137, 51 },
		{ 45, 138, 51 },
		{ 45, 139, 51 },
		{ 45, 140, 51 },
		{ 45, 141, 51 },
		{ 45, 142, 51 },
		{ 45, 143, 51 },
		{ 45, 144, 51 },
		{ 45, 145, 51 },
		{ 45, 146, 51 },
		{ 45, 147, 51 },
		{ 45, 148, 51 },
		{ 45, 149, 51 },
		{ 45, 150, 51 },
		{ 46, 99, 51 },
		{ 46, 100, 51 },
		{ 46, 101, 51 },
		{ 46, 102, 51 },
		{ 46, 103, 51 },
		{ 46, 104, 51 },
		{ 46, 105, 51 },
		{ 46, 106, 51 },
		{ 46, 107, 51 },
		{ 46, 108, 51 },
		{ 46, 109, 51 },
		{ 46, 110, 51 },
		{ 46, 111, 51 },
		{ 46, 112, 51 },
		{ 46, 113, 51 },
		{ 46, 114, 51 },
		{ 46, 115, 51 },
		{ 46, 116, 51 },
		{ 46, 117, 51 },
		{ 46, 118, 51 },
		{ 46, 119, 51 },
		{ 46, 120, 51 },
		{ 46, 121, 51 },
		{ 46, 122, 51 },
		{ 46, 123, 51 },
		{ 46, 124, 51 },
		{ 46, 125, 51 },
		{ 46, 126, 51 },
		{ 46, 127, 51 },
		{ 46, 128, 51 },
		{ 46, 129, 51 },
		{ 46, 130, 51 },
		{ 46, 131, 51 },
		{ 46, 132, 51 },
		{ 46, 133, 51 },
		{ 46, 134, 51 },
		{ 46, 135, 51 },
		{ 46, 136, 51 },
		{ 46, 137, 51 },
		{ 46, 138, 51 },
		{ 46, 139, 51 },
		{ 46, 140, 51 },
		{ 46, 141, 51 },
		{ 46, 142, 51 },
		{ 46, 143, 51 },
		{ 46, 144, 51 },
		{ 46, 145, 51 },
		{ 46, 146, 51 },
		{ 46, 147, 51 },
		{ 46, 148, 51 },
		{ 46, 149, 51 },
		{ 46, 150, 51 },
		{ 46, 151, 51 },
		{ 47, 100, 51 },
		{ 47, 101, 51 },
		{ 47, 102, 51 },
		{ 47, 103, 51 },
		{ 47, 104, 51 },
		{ 47, 105, 51 },
		{ 47, 106, 51 },
		{ 47, 107, 51 },
		{ 47, 108, 51 },
		{ 47, 109, 51 },
		{ 47, 110, 51 },
		{ 47, 111, 51 },
		{ 47, 112, 51 },
		{ 47, 113, 51 },
		{ 47, 114, 51 },
		{ 47, 115, 51 },
		{ 47, 116, 51 },
		{ 47, 117, 51 },
		{ 47, 118, 51 },
		{ 47, 119, 51 },
		{ 47, 120, 51 },
		{ 47, 121, 51 },
		{ 47, 122, 51 },
		{ 47, 123, 51 },
		{ 47, 124, 51 },
		{ 47, 125, 51 },
		{ 47, 126, 51 },
		{ 47, 127, 51 },
		{ 47, 128, 51 },
		{ 47, 129, 51 },
		{ 47, 130, 51 },
		{ 47, 131, 51 },
		{ 47, 132, 51 },
		{ 47, 133, 51 },
		{ 47, 134, 51 },
		{ 47, 135, 51 },
		{ 47, 136, 51 },
		{ 47, 137, 51 },
		{ 47, 138, 51 },
		{ 47, 139, 51 },
		{ 47, 140, 51 },
		{ 47, 141, 51 },
		{ 47, 142, 51 },
		{ 47, 143, 51 },
		{ 47, 144, 51 },
		{ 47, 145, 51 },
		{ 47, 146, 51 },
		{ 47, 147, 51 },
		{ 47, 148, 51 },
		{ 47, 149, 51 },
		{ 47, 150, 51 },
		{ 47, 151, 51 },
		{ 48, 101, 51 },
		{ 48, 102, 51 },
		{ 48, 103, 51 },
		{ 48, 104, 51 },
		{ 48, 105, 51 },
		{ 48, 106, 51 },
		{ 48, 107, 51 },
		{ 48, 108, 51 },
		{ 48, 109, 51 },
		{ 48, 110, 51 },
		{ 48, 111, 51 },
		{ 48, 112, 51 },
		{ 48, 113, 51 },
		{ 48, 114, 51 },
		{ 48, 115, 51 },
		{ 48, 116, 51 },
		{ 48, 117, 51 },
		{ 48, 118, 51 },
		{ 48, 119, 51 },
		{ 48, 120, 51 },
		{ 48, 121, 51 },
		{ 48, 122, 51 },
		{ 48, 123, 51 },
		{ 48, 124, 51 },
		{ 48, 125, 51 },
		{ 48, 126, 51 },
		{ 48, 127, 51 },
		{ 48, 128, 51 },
		{ 48, 129, 51 },
		{ 48, 130, 51 },
		{ 48, 131, 51 },
		{ 48, 132, 51 },
		{ 48, 133, 51 },
		{ 48, 134, 51 },
		{ 48, 135, 51 },
		{ 48, 136, 51 },
		{ 48, 137, 51 },
		{ 48, 138, 51 },
		{ 48, 139, 51 },
		{ 48, 140, 51 },
		{ 48, 141, 51 },
		{ 48, 142, 51 },
		{ 48, 143, 51 },
		{ 48, 144, 51 },
		{ 48, 145, 51 },
		{ 48, 146, 51 },
		{ 48, 147, 51 },
		{ 48, 148, 51 },
		{ 48, 149, 51 },
		{ 48, 150, 51 },
		{ 48, 151, 51 },
		{ 48, 152, 51 },
		{ 49, 102, 51 },
		{ 49, 103, 51 },
		{ 49, 104, 51 },
		{ 49, 105, 51 },
		{ 49, 106, 51 },
		{ 49, 107, 51 },
		{ 49, 108, 51 },
		{ 49, 109, 51 },
		{ 49, 110, 51 },
		{ 49, 111, 51 },
		{ 49, 112, 51 },
		{ 49, 113, 51 },
		{ 49, 114, 51 },
		{ 49, 115, 51 },
		{ 49, 116, 51 },
		{ 49, 117, 51 },
		{ 49, 118, 51 },
		{ 49, 119, 51 },
		{ 49, 120, 51 },
		{ 49, 121, 51 },
		{ 49, 122, 51 },
		{ 49, 123, 51 },
		{ 49, 124, 51 },
		{ 49, 125, 51 },
		{ 49, 126, 51 },
		{ 49, 127, 51 },
		{ 49, 128, 51 },
		{ 49, 129, 51 },
		{ 49, 130, 51 },
		{ 49, 131, 51 },
		{ 49, 132, 51 },
		{ 49, 133, 51 },
		{ 49, 134, 51 },
		{ 49, 135, 51 },
		{ 49, 136, 51 },
		{ 49, 137, 51 },
		{ 49, 138, 51 },
		{ 49, 139, 51 },
		{ 49, 140, 51 },
		{ 49, 141, 51 },
		{ 49, 142, 51 },
		{ 49, 143, 51 },
		{ 49, 144, 51 },
		{ 49, 145, 51 },
		{ 49, 146, 51 },
		{ 49, 147, 51 },
		{ 49, 148, 51 },
		{ 49, 149, 51 },
		{ 49, 150, 51 },
		{ 49, 151, 51 },
		{ 49, 152, 51 },
		{ 50, 102, 51 },
		{ 50, 103, 51 },
		{ 50, 104, 51 },
		{ 50, 105, 51 },
		{ 50, 106, 51 },
		{ 50, 107, 51 },
		{ 50, 108, 51 },
		{ 50, 109, 51 },
		{ 50, 110, 51 },
		{ 50, 111, 51 },
		{ 50, 112, 51 },
		{ 50, 113, 51 },
		{ 50, 114, 51 },
		{ 50, 115, 51 },
		{ 50, 116, 51 },
		{ 50, 117, 51 },
		{ 50, 118, 51 },
		{ 50, 119, 51 },
		{ 50, 120, 51 },
		{ 50, 121, 51 },
		{ 50, 122, 51 },
		{ 50, 123, 51 },
		{ 50, 124, 51 },
		{ 50, 125, 51 },
		{ 50, 126, 51 },
		{ 50, 127, 51 },
		{ 50, 128, 51 },
		{ 50, 129, 51 },
		{ 50, 130, 51 },
		{ 50, 131, 51 },
		{ 50, 132, 51 },
		{ 50, 133, 51 },
		{ 50, 134, 51 },
		{ 50, 135, 51 },
		{ 50, 136, 51 },
		{ 50, 137, 51 },
		{ 50, 138, 51 },
		{ 50, 139, 51 },
		{ 50, 140, 51 },
		{ 50, 141, 51 },
		{ 50, 142, 51 },
		{ 50, 143, 51 },
		{ 50, 144, 51 },
		{ 50, 145, 51 },
		{ 50, 146, 51 },
		{ 50, 147, 51 },
		{ 50, 148, 51 },
		{ 50, 149, 51 },
		{ 50, 150, 51 },
		{ 50, 151, 51 },
		{ 50, 152, 51 },
		{ 50, 153, 51 },
		{ 51, 103, 51 },
		{ 51, 104, 51 },
		{ 51, 105, 51 },
		{ 51, 106, 51 },
		{ 51, 107, 51 },
		{ 51, 108, 51 },
		{ 51, 109, 51 },
		{ 51, 110, 51 },
		{ 51, 111, 51 },
		{ 51, 112, 51 },
		{ 51, 113, 51 },
		{ 51, 114, 51 },
		{ 51, 115, 51 },
		{ 51, 116, 51 },
		{ 51, 117, 51 },
		{ 51, 118, 51 },
		{ 51, 119, 51 },
		{ 51, 120, 51 },
		{ 51, 121, 51 },
		{ 51, 122, 51 },
		{ 51, 123, 51 },
		{ 51, 124, 51 },
		{ 51, 125, 51 },
		{ 51, 126, 51 },
		{ 51, 127, 51 },
		{ 51, 128, 51 },
		{ 51, 129, 51 },
		{ 51, 130, 51 },
		{ 51, 131, 51 },
		{ 51, 132, 51 },
		{ 51, 133, 51 },
		{ 51, 134, 51 },
		{ 51, 135, 51 },
		{ 51, 136, 51 },
		{ 51, 137, 51 },
		{ 51, 138, 51 },
		{ 51, 139, 51 },
		{ 51, 140, 51 },
		{ 51, 141, 51 },
		{ 51, 142, 51 },
		{ 51, 143, 51 },
		{ 51, 144, 51 },
		{ 51, 145, 51 },
		{ 51, 146, 51 },
		{ 51, 147, 51 },
		{ 51, 148, 51 },
		{ 51, 149, 51 },
		{ 51, 150, 51 },
		{ 51, 151, 51 },
		{ 51, 152, 51 },
		{ 51, 153, 51 },
		{ 52, 104, 51 },
		{ 52, 105, 51 },
		{ 52, 106, 51 },
		{ 52, 107, 51 },
		{ 52, 108, 51 },
		{ 52, 109, 51 },
		{ 52, 110, 51 },
		{ 52, 111, 51 },
		{ 52, 112, 51 },
		{ 52, 113, 51 },
		{ 52, 114, 51 },
		{ 52, 115, 51 },
		{ 52, 116, 51 },
		{ 52, 117, 51 },
		{ 52, 118, 51 },
		{ 52, 119, 51 },
		{ 52, 120, 51 },
		{ 52, 121, 51 },
		{ 52, 122, 51 },
		{ 52, 123, 51 },
		{ 52, 124, 51 },
		{ 52, 125, 51 },
		{ 52, 126, 51 },
		{ 52, 127, 51 },
		{ 52, 128, 51 },
		{ 52, 129, 51 },
		{ 52, 130, 51 },
		{ 52, 131, 51 },
		{ 52, 132, 51 },
		{ 52, 133, 51 },
		{ 52, 134, 51 },
		{ 52, 135, 51 },
		{ 52, 136, 51 },
		{ 52, 137, 51 },
		{ 52, 138, 51 },
		{ 52, 139, 51 },
		{ 52, 140, 51 },
		{ 52, 141, 51 },
		{ 52, 142, 51 },
		{ 52, 143, 51 },
		{ 52, 144, 51 },
		{ 52, 145, 51 },
		{ 52, 146, 51 },
		{ 52, 147, 51 },
		{ 52, 148, 51 },
		{ 52, 149, 51 },
		{ 52, 150, 51 },
		{ 52, 151, 51 },
		{ 52, 152, 51 },
		{ 52, 153, 51 },
		{ 52, 154, 51 },
		{ 53, 105, 51 },
		{ 53, 106, 51 },
		{ 53, 107, 51 },
		{ 53, 108, 51 },
		{ 53, 109, 51 },
		{ 53, 110, 51 },
		{ 53, 111, 51 },
		{ 53, 112, 51 },
		{ 53, 113, 51 },
		{ 53, 114, 51 },
		{ 53, 115, 51 },
		{ 53, 116, 51 },
		{ 53, 117, 51 },
		{ 53, 118, 51 },
		{ 53, 119, 51 },
		{ 53, 120, 51 },
		{ 53, 121, 51 },
		{ 53, 122, 51 },
		{ 53, 123, 51 },
		{ 53, 124, 51 },
		{ 53, 125, 51 },
		{ 53, 126, 51 },
		{ 53, 127, 51 },
		{ 53, 128, 51 },
		{ 53, 129, 51 },
		{ 53, 130, 51 },
		{ 53, 131, 51 },
		{ 53, 132, 51 },
		{ 53, 133, 51 },
		{ 53, 134, 51 },
		{ 53, 135, 51 },
		{ 53, 136, 51 },
		{ 53, 137, 51 },
		{ 53, 138, 51 },
		{ 53, 139, 51 },
		{ 53, 140, 51 },
		{ 53, 141, 51 },
		{ 53, 142, 51 },
		{ 53, 143, 51 },
		{ 53, 144, 51 },
		{ 53, 145, 51 },
		{ 53, 146, 51 },
		{ 53, 147, 51 },
		{ 53, 148, 51 },
		{ 53, 149, 51 },
		{ 53, 150, 51 },
		{ 53, 151, 51 },
		{ 53, 152, 51 },
		{ 53, 153, 51 },
		{ 53, 154, 51 },
		{ 54, 105, 51 },
		{ 54, 106, 51 },
		{ 54, 107, 51 },
		{ 54, 108, 51 },
		{ 54, 109, 51 },
		{ 54, 110, 51 },
		{ 54, 111, 51 },
		{ 54, 112, 51 },
		{ 54, 113, 51 },
		{ 54, 114, 51 },
		{ 54, 115, 51 },
		{ 54, 116, 51 },
		{ 54, 117, 51 },
		{ 54, 118, 51 },
		{ 54, 119, 51 },
		{ 54, 120, 51 },
		{ 54, 121, 51 },
		{ 54, 122, 51 },
		{ 54, 123, 51 },
		{ 54, 124, 51 },
		{ 54, 125, 51 },
		{ 54, 126, 51 },
		{ 54, 127, 51 },
		{ 54, 128, 51 },
		{ 54, 129, 51 },
		{ 54, 130, 51 },
		{ 54, 131, 51 },
		{ 54, 132, 51 },
		{ 54, 133, 51 },
		{ 54, 134, 51 },
		{ 54, 135, 51 },
		{ 54, 136, 51 },
		{ 54, 137, 51 },
		{ 54, 138, 51 },
		{ 54, 139, 51 },
		{ 54, 140, 51 },
		{ 54, 141, 51 },
		{ 54, 142, 51 },
		{ 54, 143, 51 },
		{ 54, 144, 51 },
		{ 54, 145, 51 },
		{ 54, 146, 51 },
		{ 54, 147, 51 },
		{ 54, 148, 51 },
		{ 54, 149, 51 },
		{ 54, 150, 51 },
		{ 54, 151, 51 },
		{ 54, 152, 51 },
		{ 54, 153, 51 },
		{ 54, 154, 51 },
		{ 54, 155, 51 },
		{ 55, 106, 51 },
		{ 55, 107, 51 },
		{ 55, 108, 51 },
		{ 55, 109, 51 },
		{ 55, 110, 51 },
		{ 55, 111, 51 },
		{ 55, 112, 51 },
		{ 55, 113, 51 },
		{ 55, 114, 51 },
		{ 55, 115, 51 },
		{ 55, 116, 51 },
		{ 55, 117, 51 },
		{ 55, 118, 51 },
		{ 55, 119, 51 },
		{ 55, 120, 51 },
		{ 55, 121, 51 },
		{ 55, 122, 51 },
		{ 55, 123, 51 },
		{ 55, 124, 51 },
		{ 55, 125, 51 },
		{ 55, 126, 51 },
		{ 55, 127, 51 },
		{ 55, 128, 51 },
		{ 55, 129, 51 },
		{ 55, 130, 51 },
		{ 55, 131, 51 },
		{ 55, 132, 51 },
		{ 55, 133, 51 },
		{ 55, 134, 51 },
		{ 55, 135, 51 },
		{ 55, 136, 51 },
		{ 55, 137, 51 },
		{ 55, 138, 51 },
		{ 55, 139, 51 },
		{ 55, 140, 51 },
		{ 55, 141, 51 },
		{ 55, 142, 51 },
		{ 55, 143, 51 },
		{ 55, 144, 51 },
		{ 55, 145, 51 },
		{ 55, 146, 51 },
		{ 55, 147, 51 },
		{ 55, 148, 51 },
		{ 55, 149, 51 },
		{ 55, 150, 51 },
		{ 55, 151, 51 },
		{ 55, 152, 51 },
		{ 55, 153, 51 },
		{ 55, 154, 51 },
		{ 55, 155, 51 },
		{ 56, 107, 51 },
		{ 56, 108, 51 },
		{ 56, 109, 51 },
		{ 56, 110, 51 },
		{ 56, 111, 51 },
		{ 56, 112, 51 },
		{ 56, 113, 51 },
		{ 56, 114, 51 },
		{ 56, 115, 51 },
		{ 56, 116, 51 },
		{ 56, 117, 51 },
		{ 56, 118, 51 },
		{ 56, 119, 51 },
		{ 56, 120, 51 },
		{ 56, 121, 51 },
		{ 56, 122, 51 },
		{ 56, 123, 51 },
		{ 56, 124, 51 },
		{ 56, 125, 51 },
		{ 56, 126, 51 },
		{ 56, 127, 51 },
		{ 56, 128, 51 },
		{ 56, 129, 51 },
		{ 56, 130, 51 },
		{ 56, 131, 51 },
		{ 56, 132, 51 },
		{ 56, 133, 51 },
		{ 56, 134, 51 },
		{ 56, 135, 51 },
		{ 56, 136, 51 },
		{ 56, 137, 51 },
		{ 56, 138, 51 },
		{ 56, 139, 51 },
		{ 56, 140, 51 },
		{ 56, 141, 51 },
		{ 56, 142, 51 },
		{ 56, 143, 51 },
		{ 56, 144, 51 },
		{ 56, 145, 51 },
		{ 56, 146, 51 },
		{ 56, 147, 51 },
		{ 56, 148, 51 },
		{ 56, 149, 51 },
		{ 56, 150, 51 },
		{ 56, 151, 51 },
		{ 56, 152, 51 },
		{ 56, 153, 51 },
		{ 56, 154, 51 },
		{ 56, 155, 51 },
		{ 56, 156, 51 },
		{ 57, 108, 51 },
		{ 57, 109, 51 },
		{ 57, 110, 51 },
		{ 57, 111, 51 },
		{ 57, 112, 51 },
		{ 57, 113, 51 },
		{ 57, 114, 51 },
		{ 57, 115, 51 },
		{ 57, 116, 51 },
		{ 57, 117, 51 },
		{ 57, 118, 51 },
		{ 57, 119, 51 },
		{ 57, 120, 51 },
		{ 57, 121, 51 },
		{ 57, 122, 51 },
		{ 57, 123, 51 },
		{ 57, 124, 51 },
		{ 57, 125, 51 },
		{ 57, 126, 51 },
		{ 57, 127, 51 },
		{ 57, 128, 51 },
		{ 57, 129, 51 },
		{ 57, 130, 51 },
		{ 57, 131, 51 },
		{ 57, 132, 51 },
		{ 57, 133, 51 },
		{ 57, 134, 51 },
		{ 57, 135, 51 },
		{ 57, 136, 51 },
		{ 57, 137, 51 },
		{ 57, 138, 51 },
		{ 57, 139, 51 },
		{ 57, 140, 51 },
		{ 57, 141, 51 },
		{ 57, 142, 51 },
		{ 57, 143, 51 },
		{ 57, 144, 51 },
		{ 57, 145, 51 },
		{ 57, 146, 51 },
		{ 57, 147, 51 },
		{ 57, 148, 51 },
		{ 57, 149, 51 },
		{ 57, 150, 51 },
		{ 57, 151, 51 },
		{ 57, 152, 51 },
		{ 57, 153, 51 },
		{ 57, 154, 51 },
		{ 57, 155, 51 },
		{ 57, 156, 51 },
		{ 58, 108, 51 },
		{ 58, 109, 51 },
		{ 58, 110, 51 },
		{ 58, 111, 51 },
		{ 58, 112, 51 },
		{ 58, 113, 51 },
		{ 58, 114, 51 },
		{ 58, 115, 51 },
		{ 58, 116, 51 },
		{ 58, 117, 51 },
		{ 58, 118, 51 },
		{ 58, 119, 51 },
		{ 58, 120, 51 },
		{ 58, 121, 51 },
		{ 58, 122, 51 },
		{ 58, 123, 51 },
		{ 58, 124, 51 },
		{ 58, 125, 51 },
		{ 58, 126, 51 },
		{ 58, 127, 51 },
		{ 58, 128, 51 },
		{ 58, 129, 51 },
		{ 58, 130, 51 },
		{ 58, 131, 51 },
		{ 58, 132, 51 },
		{ 58, 133, 51 },
		{ 58, 134, 51 },
		{ 58, 135, 51 },
		{ 58, 136, 51 },
		{ 58, 137, 51 },
		{ 58, 138, 51 },
		{ 58, 139, 51 },
		{ 58, 140, 51 },
		{ 58, 141, 51 },
		{ 58, 142, 51 },
		{ 58, 143, 51 },
		{ 58, 144, 51 },
		{ 58, 145, 51 },
		{ 58, 146, 51 },
		{ 58, 147, 51 },
		{ 58, 148, 51 },
		{ 58, 149, 51 },
		{ 58, 150, 51 },
		{ 58, 151, 51 },
		{ 58, 152, 51 },
		{ 58, 153, 51 },
		{ 58, 154, 51 },
		{ 58, 155, 51 },
		{ 58, 156, 51 },
		{ 58, 157, 51 },
		{ 59, 109, 51 },
		{ 59, 110, 51 },
		{ 59, 111, 51 },
		{ 59, 112, 51 },
		{ 59, 113, 51 },
		{ 59, 114, 51 },
		{ 59, 115, 51 },
		{ 59, 116, 51 },
		{ 59, 117, 51 },
		{ 59, 118, 51 },
		{ 59, 119, 51 },
		{ 59, 120, 51 },
		{ 59, 121, 51 },
		{ 59, 122, 51 },
		{ 59, 123, 51 },
		{ 59, 124, 51 },
		{ 59, 125, 51 },
		{ 59, 126, 51 },
		{ 59, 127, 51 },
		{ 59, 128, 51 },
		{ 59, 129, 51 },
		{ 59, 130, 51 },
		{ 59, 131, 51 },
		{ 59, 132, 51 },
		{ 59, 133, 51 },
		{ 59, 134, 51 },
		{ 59, 135, 51 },
		{ 59, 136, 51 },
		{ 59, 137, 51 },
		{ 59, 138, 51 },
		{ 59, 139, 51 },
		{ 59, 140, 51 },
		{ 59, 141, 51 },
		{ 59, 142, 51 },
		{ 59, 143, 51 },
		{ 59, 144, 51 },
		{ 59, 145, 51 },
		{ 59, 146, 51 },
		{ 59, 147, 51 },
		{ 59, 148, 51 },
		{ 59, 149, 51 },
		{ 59, 150, 51 },
		{ 59, 151, 51 },
		{ 59, 152, 51 },
		{ 59, 153, 51 },
		{ 59, 154, 51 },
		{ 59, 155, 51 },
		{ 59, 156, 51 },
		{ 59, 157, 51 },
		{ 60, 110, 51 },
		{ 60, 111, 51 },
		{ 60, 112, 51 },
		{ 60, 113, 51 },
		{ 60, 114, 51 },
		{ 60, 115, 51 },
		{ 60, 116, 51 },
		{ 60, 117, 51 },
		{ 60, 118, 51 },
		{ 60, 119, 51 },
		{ 60, 120, 51 },
		{ 60, 121, 51 },
		{ 60, 122, 51 },
		{ 60, 123, 51 },
		{ 60, 124, 51 },
		{ 60, 125, 51 },
		{ 60, 126, 51 },
		{ 60, 127, 51 },
		{ 60, 128, 51 },
		{ 60, 129, 51 },
		{ 60, 130, 51 },
		{ 60, 131, 51 },
		{ 60, 132, 51 },
		{ 60, 133, 51 },
		{ 60, 134, 51 },
		{ 60, 135, 51 },
		{ 60, 136, 51 },
		{ 60, 137, 51 },
		{ 60, 138, 51 },
		{ 60, 139, 51 },
		{ 60, 140, 51 },
		{ 60, 141, 51 },
		{ 60, 142, 51 },
		{ 60, 143, 51 },
		{ 60, 144, 51 },
		{ 60, 145, 51 },
		{ 60, 146, 51 },
		{ 60, 147, 51 },
		{ 60, 148, 51 },
		{ 60, 149, 51 },
		{ 60, 150, 51 },
		{ 60, 151, 51 },
		{ 60, 152, 51 },
		{ 60, 153, 51 },
		{ 60, 154, 51 },
		{ 60, 155, 51 },
		{ 60, 156, 51 },
		{ 60, 157, 51 },
		{ 60, 158, 51 },
		{ 61, 111, 51 },
		{ 61, 112, 51 },
		{ 61, 113, 51 },
		{ 61, 114, 51 },
		{ 61, 115, 51 },
		{ 61, 116, 51 },
		{ 61, 117, 51 },
		{ 61, 118, 51 },
		{ 61, 119, 51 },
		{ 61, 120, 51 },
		{ 61, 121, 51 },
		{ 61, 122, 51 },
		{ 61, 123, 51 },
		{ 61, 124, 51 },
		{ 61, 125, 51 },
		{ 61, 126, 51 },
		{ 61, 127, 51 },
		{ 61, 128, 51 },
		{ 61, 129, 51 },
		{ 61, 130, 51 },
		{ 61, 131, 51 },
		{ 61, 132, 51 },
		{ 61, 133, 51 },
		{ 61, 134, 51 },
		{ 61, 135, 51 },
		{ 61, 136, 51 },
		{ 61, 137, 51 },
		{ 61, 138, 51 },
		{ 61, 139, 51 },
		{ 61, 140, 51 },
		{ 61, 141, 51 },
		{ 61, 142, 51 },
		{ 61, 143, 51 },
		{ 61, 144, 51 },
		{ 61, 145, 51 },
		{ 61, 146, 51 },
		{ 61, 147, 51 },
		{ 61, 148, 51 },
		{ 61, 149, 51 },
		{ 61, 150, 51 },
		{ 61, 151, 51 },
		{ 61, 152, 51 },
		{ 61, 153, 51 },
		{ 61, 154, 51 },
		{ 61, 155, 51 },
		{ 61, 156, 51 },
		{ 61, 157, 51 },
		{ 61, 158, 51 },
		{ 62, 111, 51 },
		{ 62, 112, 51 },
		{ 62, 113, 51 },
		{ 62, 114, 51 },
		{ 62, 115, 51 },
		{ 62, 116, 51 },
		{ 62, 117, 51 },
		{ 62, 118, 51 },
		{ 62, 119, 51 },
		{ 62, 120, 51 },
		{ 62, 121, 51 },
		{ 62, 122, 51 },
		{ 62, 123, 51 },
		{ 62, 124, 51 },
		{ 62, 125, 51 },
		{ 62, 126, 51 },
		{ 62, 127, 51 },
		{ 62, 128, 51 },
		{ 62, 129, 51 },
		{ 62, 130, 51 },
		{ 62, 131, 51 },
		{ 62, 132, 51 },
		{ 62, 133, 51 },
		{ 62, 134, 51 },
		{ 62, 135, 51 },
		{ 62, 136, 51 },
		{ 62, 137, 51 },
		{ 62, 138, 51 },
		{ 62, 139, 51 },
		{ 62, 140, 51 },
		{ 62, 141, 51 },
		{ 62, 142, 51 },
		{ 62, 143, 51 },
		{ 62, 144, 51 },
		{ 62, 145, 51 },
		{ 62, 146, 51 },
		{ 62, 147, 51 },
		{ 62, 148, 51 },
		{ 62, 149, 51 },
		{ 62, 150, 51 },
		{ 62, 151, 51 },
		{ 62, 152, 51 },
		{ 62, 153, 51 },
		{ 62, 154, 51 },
		{ 62, 155, 51 },
		{ 62, 156, 51 },
		{ 62, 157, 51 },
		{ 62, 158, 51 },
		{ 62, 159, 51 },
		{ 63, 112, 51 },
		{ 63, 113, 51 },
		{ 63, 114, 51 },
		{ 63, 115, 51 },
		{ 63, 116, 51 },
		{ 63, 117, 51 },
		{ 63, 118, 51 },
		{ 63, 119, 51 },
		{ 63, 120, 51 },
		{ 63, 121, 51 },
		{ 63, 122, 51 },
		{ 63, 123, 51 },
		{ 63, 124, 51 },
		{ 63, 125, 51 },
		{ 63, 126, 51 },
		{ 63, 127, 51 },
		{ 63, 128, 51 },
		{ 63, 129, 51 },
		{ 63, 130, 51 },
		{ 63, 131, 51 },
		{ 63, 132, 51 },
		{ 63, 133, 51 },
		{ 63, 134, 51 },
		{ 63, 135, 51 },
		{ 63, 136, 51 },
		{ 63, 137, 51 },
		{ 63, 138, 51 },
		{ 63, 139, 51 },
		{ 63, 140, 51 },
		{ 63, 141, 51 },
		{ 63, 142, 51 },
		{ 63, 143, 51 },
		{ 63, 144, 51 },
		{ 63, 145, 51 },
		{ 63, 146, 51 },
		{ 63, 147, 51 },
		{ 63, 148, 51 },
		{ 63, 149, 51 },
		{ 63, 150, 51 },
		{ 63, 151, 51 },
		{ 63, 152, 51 },
		{ 63, 153, 51 },
		{ 63, 154, 51 },
		{ 63, 155, 51 },
		{ 63, 156, 51 },
		{ 63, 157, 51 },
		{ 63, 158, 51 },
		{ 63, 159, 51 },
		{ 64, 113, 51 },
		{ 64, 114, 51 },
		{ 64, 115, 51 },
		{ 64, 116, 51 },
		{ 64, 117, 51 },
		{ 64, 118, 51 },
		{ 64, 119, 51 },
		{ 64, 120, 51 },
		{ 64, 121, 51 },
		{ 64, 122, 51 },
		{ 64, 123, 51 },
		{ 64, 124, 51 },
		{ 64, 125, 51 },
		{ 64, 126, 51 },
		{ 64, 127, 51 },
		{ 64, 128, 51 },
		{ 64, 129, 51 },
		{ 64, 130, 51 },
		{ 64, 131, 51 },
		{ 64, 132, 51 },
		{ 64, 133, 51 },
		{ 64, 134, 51 },
		{ 64, 135, 51 },
		{ 64, 136, 51 },
		{ 64, 137, 51 },
		{ 64, 138, 51 },
		{ 64, 139, 51 },
		{ 64, 140, 51 },
		{ 64, 141, 51 },
		{ 64, 142, 51 },
		{ 64, 143, 51 },
		{ 64, 144, 51 },
		{ 64, 145, 51 },
		{ 64, 146, 51 },
		{ 64, 147, 51 },
		{ 64, 148, 51 },
		{ 64, 149, 51 },
		{ 64, 150, 51 },
		{ 64, 151, 51 },
		{ 64, 152, 51 },
		{ 64, 153, 51 },
		{ 64, 154, 51 },
		{ 64, 155, 51 },
		{ 64, 156, 51 },
		{ 64, 157, 51 },
		{ 64, 158, 51 },
		{ 64, 159, 51 },
		{ 64, 160, 51 },
		{ 65, 114, 51 },
		{ 65, 115, 51 },
		{ 65, 116, 51 },
		{ 65, 117, 51 },
		{ 65, 118, 51 },
		{ 65, 119, 51 },
		{ 65, 120, 51 },
		{ 65, 121, 51 },
		{ 65, 122, 51 },
		{ 65, 123, 51 },
		{ 65, 124, 51 },
		{ 65, 125, 51 },
		{ 65, 126, 51 },
		{ 65, 127, 51 },
		{ 65, 128, 51 },
		{ 65, 129, 51 },
		{ 65, 130, 51 },
		{ 65, 131, 51 },
		{ 65, 132, 51 },
		{ 65, 133, 51 },
		{ 65, 134, 51 },
		{ 65, 135, 51 },
		{ 65, 136, 51 },
		{ 65, 137, 51 },
		{ 65, 138, 51 },
		{ 65, 139, 51 },
		{ 65, 140, 51 },
		{ 65, 141, 51 },
		{ 65, 142, 51 },
		{ 65, 143, 51 },
		{ 65, 144, 51 },
		{ 65, 145, 51 },
		{ 65, 146, 51 },
		{ 65, 147, 51 },
		{ 65, 148, 51 },
		{ 65, 149, 51 },
		{ 65, 150, 51 },
		{ 65, 151, 51 },
		{ 65, 152, 51 },
		{ 65, 153, 51 },
		{ 65, 154, 51 },
		{ 65, 155, 51 },
		{ 65, 156, 51 },
		{ 65, 157, 51 },
		{ 65, 158, 51 },
		{ 65, 159, 51 },
		{ 65, 160, 51 },
		{ 66, 114, 51 },
		{ 66, 115, 51 },
		{ 66, 116, 51 },
		{ 66, 117, 51 },
		{ 66, 118, 51 },
		{ 66, 119, 51 },
		{ 66, 120, 51 },
		{ 66, 121, 51 },
		{ 66, 122, 51 },
		{ 66, 123, 51 },
		{ 66, 124, 51 },
		{ 66, 125, 51 },
		{ 66, 126, 51 },
		{ 66, 127, 51 },
		{ 66, 128, 51 },
		{ 66, 129, 51 },
		{ 66, 130, 51 },
		{ 66, 131, 51 },
		{ 66, 132, 51 },
		{ 66, 133, 51 },
		{ 66, 134, 51 },
		{ 66, 135, 51 },
		{ 66, 136, 51 },
		{ 66, 137, 51 },
		{ 66, 138, 51 },
		{ 66, 139, 51 },
		{ 66, 140, 51 },
		{ 66, 141, 51 },
		{ 66, 142, 51 },
		{ 66, 143, 51 },
		{ 66, 144, 51 },
		{ 66, 145, 51 },
		{ 66, 146, 51 },
		{ 66, 147, 51 },
		{ 66, 148, 51 },
		{ 66, 149, 51 },
		{ 66, 150, 51 },
		{ 66, 151, 51 },
		{ 66, 152, 51 },
		{ 66, 153, 51 },
		{ 66, 154, 51 },
		{ 66, 155, 51 },
		{ 66, 156, 51 },
		{ 66, 157, 51 },
		{ 66, 158, 51 },
		{ 66, 159, 51 },
		{ 66, 160, 51 },
		{ 66, 161, 51 },
		{ 67, 115, 51 },
		{ 67, 116, 51 },
		{ 67, 117, 51 },
		{ 67, 118, 51 },
		{ 67, 119, 51 },
		{ 67, 120, 51 },
		{ 67, 121, 51 },
		{ 67, 122, 51 },
		{ 67, 123, 51 },
		{ 67, 124, 51 },
		{ 67, 125, 51 },
		{ 67, 126, 51 },
		{ 67, 127, 51 },
		{ 67, 128, 51 },
		{ 67, 129, 51 },
		{ 67, 130, 51 },
		{ 67, 131, 51 },
		{ 67, 132, 51 },
		{ 67, 133, 51 },
		{ 67, 134, 51 },
		{ 67, 135, 51 },
		{ 67, 136, 51 },
		{ 67, 137, 51 },
		{ 67, 138, 51 },
		{ 67, 139, 51 },
		{ 67, 140, 51 },
		{ 67, 141, 51 },
		{ 67, 142, 51 },
		{ 67, 143, 51 },
		{ 67, 144, 51 },
		{ 67, 145, 51 },
		{ 67, 146, 51 },
		{ 67, 147, 51 },
		{ 67, 148, 51 },
		{ 67, 149, 51 },
		{ 67, 150, 51 },
		{ 67, 151, 51 },
		{ 67, 152, 51 },
		{ 67, 153, 51 },
		{ 67, 154, 51 },
		{ 67, 155, 51 },
		{ 67, 156, 51 },
		{ 67, 157, 51 },
		{ 67, 158, 51 },
		{ 67, 159, 51 },
		{ 67, 160, 51 },
		{ 67, 161, 51 },
		{ 68, 116, 51 },
		{ 68, 117, 51 },
		{ 68, 118, 51 },
		{ 68, 119, 51 },
		{ 68, 120, 51 },
		{ 68, 121, 51 },
		{ 68, 122, 51 },
		{ 68, 123, 51 },
		{ 68, 124, 51 },
		{ 68, 125, 51 },
		{ 68, 126, 51 },
		{ 68, 127, 51 },
		{ 68, 128, 51 },
		{ 68, 129, 51 },
		{ 68, 130, 51 },
		{ 68, 131, 51 },
		{ 68, 132, 51 },
		{ 68, 133, 51 },
		{ 68, 134, 51 },
		{ 68, 135, 51 },
		{ 68, 136, 51 },
		{ 68, 137, 51 },
		{ 68, 138, 51 },
		{ 68, 139, 51 },
		{ 68, 140, 51 },
		{ 68, 141, 51 },
		{ 68, 142, 51 },
		{ 68, 143, 51 },
		{ 68, 144, 51 },
		{ 68, 145, 51 },
		{ 68, 146, 51 },
		{ 68, 147, 51 },
		{ 68, 148, 51 },
		{ 68, 149, 51 },
		{ 68, 150, 51 },
		{ 68, 151, 51 },
		{ 68, 152, 51 },
		{ 68, 153, 51 },
		{ 68, 154, 51 },
		{ 68, 155, 51 },
		{ 68, 156, 51 },
		{ 68, 157, 51 },
		{ 68, 158, 51 },
		{ 68, 159, 51 },
		{ 68, 160, 51 },
		{ 68, 161, 51 },
		{ 68, 162, 51 },
		{ 69, 117, 51 },
		{ 69, 118, 51 },
		{ 69, 119, 51 },
		{ 69, 120, 51 },
		{ 69, 121, 51 },
		{ 69, 122, 51 },
		{ 69, 123, 51 },
		{ 69, 124, 51 },
		{ 69, 125, 51 },
		{ 69, 126, 51 },
		{ 69, 127, 51 },
		{ 69, 128, 51 },
		{ 69, 129, 51 },
		{ 69, 130, 51 },
		{ 69, 131, 51 },
		{ 69, 132, 51 },
		{ 69, 133, 51 },
		{ 69, 134, 51 },
		{ 69, 135, 51 },
		{ 69, 136, 51 },
		{ 69, 137, 51 },
		{ 69, 138, 51 },
		{ 69, 139, 51 },
		{ 69, 140, 51 },
		{ 69, 141, 51 },
		{ 69, 142, 51 },
		{ 69, 143, 51 },
		{ 69, 144, 51 },
		{ 69, 145, 51 },
		{ 69, 146, 51 },
		{ 69, 147, 51 },
		{ 69, 148, 51 },
		{ 69, 149, 51 },
		{ 69, 150, 51 },
		{ 69, 151, 51 },
		{ 69, 152, 51 },
		{ 69, 153, 51 },
		{ 69, 154, 51 },
		{ 69, 155, 51 },
		{ 69, 156, 51 },
		{ 69, 157, 51 },
		{ 69, 158, 51 },
		{ 69, 159, 51 },
		{ 69, 160, 51 },
		{ 69, 161, 51 },
		{ 69, 162, 51 },
		{ 70, 117, 51 },
		{ 70, 118, 51 },
		{ 70, 119, 51 },
		{ 70, 120, 51 },
		{ 70, 121, 51 },
		{ 70, 122, 51 },
		{ 70, 123, 51 },
		{ 70, 124, 51 },
		{ 70, 125, 51 },
		{ 70, 126, 51 },
		{ 70, 127, 51 },
		{ 70, 128, 51 },
		{ 70, 129, 51 },
		{ 70, 130, 51 },
		{ 70, 131, 51 },
		{ 70, 132, 51 },
		{ 70, 133, 51 },
		{ 70, 134, 51 },
		{ 70, 135, 51 },
		{ 70, 136, 51 },
		{ 70, 137, 51 },
		{ 70, 138, 51 },
		{ 70, 139, 51 },
		{ 70, 140, 51 },
		{ 70, 141, 51 },
		{ 70, 142, 51 },
		{ 70, 143, 51 },
		{ 70, 144, 51 },
		{ 70, 145, 51 },
		{ 70, 146, 51 },
		{ 70, 147, 51 },
		{ 70, 148, 51 },
		{ 70, 149, 51 },
		{ 70, 150, 51 },
		{ 70, 151, 51 },
		{ 70, 152, 51 },
		{ 70, 153, 51 },
		{ 70, 154, 51 },
		{ 70, 155, 51 },
		{ 70, 156, 51 },
		{ 70, 157, 51 },
		{ 70, 158, 51 },
		{ 70, 159, 51 },
		{ 70, 160, 51 },
		{ 70, 161, 51 },
		{ 70, 162, 51 },
		{ 70, 163, 51 },
		{ 71, 118, 51 },
		{ 71, 119, 51 },
		{ 71, 120, 51 },
		{ 71, 121, 51 },
		{ 71, 122, 51 },
		{ 71, 123, 51 },
		{ 71, 124, 51 },
		{ 71, 125, 51 },
		{ 71, 126, 51 },
		{ 71, 127, 51 },
		{ 71, 128, 51 },
		{ 71, 129, 51 },
		{ 71, 130, 51 },
		{ 71, 131, 51 },
		{ 71, 132, 51 },
		{ 71, 133, 51 },
		{ 71, 134, 51 },
		{ 71, 135, 51 },
		{ 71, 136, 51 },
		{ 71, 137, 51 },
		{ 71, 138, 51 },
		{ 71, 139, 51 },
		{ 71, 140, 51 },
		{ 71, 141, 51 },
		{ 71, 142, 51 },
		{ 71, 143, 51 },
		{ 71, 144, 51 },
		{ 71, 145, 51 },
		{ 71, 146, 51 },
		{ 71, 147, 51 },
		{ 71, 148, 51 },
		{ 71, 149, 51 },
		{ 71, 150, 51 },
		{ 71, 151, 51 },
		{ 71, 152, 51 },
		{ 71, 153, 51 },
		{ 71, 154, 51 },
		{ 71, 155, 51 },
		{ 71, 156, 51 },
		{ 71, 157, 51 },
		{ 71, 158, 51 },
		{ 71, 159, 51 },
		{ 71, 160, 51 },
		{ 71, 161, 51 },
		{ 71, 162, 51 },
		{ 71, 163, 51 },
		{ 72, 119, 51 },
		{ 72, 120, 51 },
		{ 72, 121, 51 },
		{ 72, 122, 51 },
		{ 72, 123, 51 },
		{ 72, 124, 51 },
		{ 72, 125, 51 },
		{ 72, 126, 51 },
		{ 72, 127, 51 },
		{ 72, 128, 51 },
		{ 72, 129, 51 },
		{ 72, 130, 51 },
		{ 72, 131, 51 },
		{ 72, 132, 51 },
		{ 72, 133, 51 },
		{ 72, 134, 51 },
		{ 72, 135, 51 },
		{ 72, 136, 51 },
		{ 72, 137, 51 },
		{ 72, 138, 51 },
		{ 72, 139, 51 },
		{ 72, 140, 51 },
		{ 72, 141, 51 },
		{ 72, 142, 51 },
		{ 72, 143, 51 },
		{ 72, 144, 51 },
		{ 72, 145, 51 },
		{ 72, 146, 51 },
		{ 72, 147, 51 },
		{ 72, 148, 51 },
		{ 72, 149, 51 },
		{ 72, 150, 51 },
		{ 72, 151, 51 },
		{ 72, 152, 51 },
		{ 72, 153, 51 },
		{ 72, 154, 51 },
		{ 72, 155, 51 },
		{ 72, 156, 51 },
		{ 72, 157, 51 },
		{ 72, 158, 51 },
		{ 72, 159, 51 },
		{ 72, 160, 51 },
		{ 72, 161, 51 },
		{ 72, 162, 51 },
		{ 72, 163, 51 },
		{ 72, 164, 51 },
		{ 73, 120, 51 },
		{ 73, 121, 51 },
		{ 73, 122, 51 },
		{ 73, 123, 51 },
		{ 73, 124, 51 },
		{ 73, 125, 51 },
		{ 73, 126, 51 },
		{ 73, 127, 51 },
		{ 73, 128, 51 },
		{ 73, 129, 51 },
		{ 73, 130, 51 },
		{ 73, 131, 51 },
		{ 73, 132, 51 },
		{ 73, 133, 51 },
		{ 73, 134, 51 },
		{ 73, 135, 51 },
		{ 73, 136, 51 },
		{ 73, 137, 51 },
		{ 73, 138, 51 },
		{ 73, 139, 51 },
		{ 73, 140, 51 },
		{ 73, 141, 51 },
		{ 73, 142, 51 },
		{ 73, 143, 51 },
		{ 73, 144, 51 },
		{ 73, 145, 51 },
		{ 73, 146, 51 },
		{ 73, 147, 51 },
		{ 73, 148, 51 },
		{ 73, 149, 51 },
		{ 73, 150, 51 },
		{ 73, 151, 51 },
		{ 73, 152, 51 },
		{ 73, 153, 51 },
		{ 73, 154, 51 },
		{ 73, 155, 51 },
		{ 73, 156, 51 },
		{ 73, 157, 51 },
		{ 73, 158, 51 },
		{ 73, 159, 51 },
		{ 73, 160, 51 },
		{ 73, 161, 51 },
		{ 73, 162, 51 },
		{ 73, 163, 51 },
		{ 73, 164, 51 },
		{ 74, 120, 51 },
		{ 74, 121, 51 },
		{ 74, 122, 51 },
		{ 74, 123, 51 },
		{ 74, 124, 51 },
		{ 74, 125, 51 },
		{ 74, 126, 51 },
		{ 74, 127, 51 },
		{ 74, 128, 51 },
		{ 74, 129, 51 },
		{ 74, 130, 51 },
		{ 74, 131, 51 },
		{ 74, 132, 51 },
		{ 74, 133, 51 },
		{ 74, 134, 51 },
		{ 74, 135, 51 },
		{ 74, 136, 51 },
		{ 74, 137, 51 },
		{ 74, 138, 51 },
		{ 74, 139, 51 },
		{ 74, 140, 51 },
		{ 74, 141, 51 },
		{ 74, 142, 51 },
		{ 74, 143, 51 },
		{ 74, 144, 51 },
		{ 74, 145, 51 },
		{ 74, 146, 51 },
		{ 74, 147, 51 },
		{ 74, 148, 51 },
		{ 74, 149, 51 },
		{ 74, 150, 51 },
		{ 74, 151, 51 },
		{ 74, 152, 51 },
		{ 74, 153, 51 },
		{ 74, 154, 51 },
		{ 74, 155, 51 },
		{ 74, 156, 51 },
		{ 74, 157, 51 },
		{ 74, 158, 51 },
		{ 74, 159, 51 },
		{ 74, 160, 51 },
		{ 74, 161, 51 },
		{ 74, 162, 51 },
		{ 74, 163, 51 },
		{ 74, 164, 51 },
		{ 74, 165, 51 },
		{ 75, 121, 51 },
		{ 75, 122, 51 },
		{ 75, 123, 51 },
		{ 75, 124, 51 },
		{ 75, 125, 51 },
		{ 75, 126, 51 },
		{ 75, 127, 51 },
		{ 75, 128, 51 },
		{ 75, 129, 51 },
		{ 75, 130, 51 },
		{ 75, 131, 51 },
		{ 75, 132, 51 },
		{ 75, 133, 51 },
		{ 75, 134, 51 },
		{ 75, 135, 51 },
		{ 75, 136, 51 },
		{ 75, 137, 51 },
		{ 75, 138, 51 },
		{ 75, 139, 51 },
		{ 75, 140, 51 },
		{ 75, 141, 51 },
		{ 75, 142, 51 },
		{ 75, 143, 51 },
		{ 75, 144, 51 },
		{ 75, 145, 51 },
		{ 75, 146, 51 },
		{ 75, 147, 51 },
		{ 75, 148, 51 },
		{ 75, 149, 51 },
		{ 75, 150, 51 },
		{ 75, 151, 51 },
		{ 75, 152, 51 },
		{ 75, 153, 51 },
		{ 75, 154, 51 },
		{ 75, 155, 51 },
		{ 75, 156, 51 },
		{ 75, 157, 51 },
		{ 75, 158, 51 },
		{ 75, 159, 51 },
		{ 75, 160, 51 },
		{ 75, 161, 51 },
		{ 75, 162, 51 },
		{ 75, 163, 51 },
		{ 75, 164, 51 },
		{ 75, 165, 51 },
		{ 76, 122, 51 },
		{ 76, 123, 51 },
		{ 76, 124, 51 },
		{ 76, 125, 51 },
		{ 76, 126, 51 },
		{ 76, 127, 51 },
		{ 76, 128, 51 },
		{ 76, 129, 51 },
		{ 76, 130, 51 },
		{ 76, 131, 51 },
		{ 76, 132, 51 },
		{ 76, 133, 51 },
		{ 76, 134, 51 },
		{ 76, 135, 51 },
		{ 76, 136, 51 },
		{ 76, 137, 51 },
		{ 76, 138, 51 },
		{ 76, 139, 51 },
		{ 76, 140, 51 },
		{ 76, 141, 51 },
		{ 76, 142, 51 },
		{ 76, 143, 51 },
		{ 76, 144, 51 },
		{ 76, 145, 51 },
		{ 76, 146, 51 },
		{ 76, 147, 51 },
		{ 76, 148, 51 },
		{ 76, 149, 51 },
		{ 76, 150, 51 },
		{ 76, 151, 51 },
		{ 76, 152, 51 },
		{ 76, 153, 51 },
		{ 76, 154, 51 },
		{ 76, 155, 51 },
		{ 76, 156, 51 },
		{ 76, 157, 51 },
		{ 76, 158, 51 },
		{ 76, 159, 51 },
		{ 76, 160, 51 },
		{ 76, 161, 51 },
		{ 76, 162, 51 },
		{ 76, 163, 51 },
		{ 76, 164, 51 },
		{ 76, 165, 51 },
		{ 76, 166, 51 },
		{ 77, 123, 51 },
		{ 77, 124, 51 },
		{ 77, 125, 51 },
		{ 77, 126, 51 },
		{ 77, 127, 51 },
		{ 77, 128, 51 },
		{ 77, 129, 51 },
		{ 77, 130, 51 },
		{ 77, 131, 51 },
		{ 77, 132, 51 },
		{ 77, 133, 51 },
		{ 77, 134, 51 },
		{ 77, 135, 51 },
		{ 77, 136, 51 },
		{ 77, 137, 51 },
		{ 77, 138, 51 },
		{ 77, 139, 51 },
		{ 77, 140, 51 },
		{ 77, 141, 51 },
		{ 77, 142, 51 },
		{ 77, 143, 51 },
		{ 77, 144, 51 },
		{ 77, 145, 51 },
		{ 77, 146, 51 },
		{ 77, 147, 51 },
		{ 77, 148, 51 },
		{ 77, 149, 51 },
		{ 77, 150, 51 },
		{ 77, 151, 51 },
		{ 77, 152, 51 },
		{ 77, 153, 51 },
		{ 77, 154, 51 },
		{ 77, 155, 51 },
		{ 77, 156, 51 },
		{ 77, 157, 51 },
		{ 77, 158, 51 },
		{ 77, 159, 51 },
		{ 77, 160, 51 },
		{ 77, 161, 51 },
		{ 77, 162, 51 },
		{ 77, 163, 51 },
		{ 77, 164, 51 },
		{ 77, 165, 51 },
		{ 77, 166, 51 },
		{ 78, 123, 51 },
		{ 78, 124, 51 },
		{ 78, 125, 51 },
		{ 78, 126, 51 },
		{ 78, 127, 51 },
		{ 78, 128, 51 },
		{ 78, 129, 51 },
		{ 78, 130, 51 },
		{ 78, 131, 51 },
		{ 78, 132, 51 },
		{ 78, 133, 51 },
		{ 78, 134, 51 },
		{ 78, 135, 51 },
		{ 78, 136, 51 },
		{ 78, 137, 51 },
		{ 78, 138, 51 },
		{ 78, 139, 51 },
		{ 78, 140, 51 },
		{ 78, 141, 51 },
		{ 78, 142, 51 },
		{ 78, 143, 51 },
		{ 78, 144, 51 },
		{ 78, 145, 51 },
		{ 78, 146, 51 },
		{ 78, 147, 51 },
		{ 78, 148, 51 },
		{ 78, 149, 51 },
		{ 78, 150, 51 },
		{ 78, 151, 51 },
		{ 78, 152, 51 },
		{ 78, 153, 51 },
		{ 78, 154, 51 },
		{ 78, 155, 51 },
		{ 78, 156, 51 },
		{ 78, 157, 51 },
		{ 78, 158, 51 },
		{ 78, 159, 51 },
		{ 78, 160, 51 },
		{ 78, 161, 51 },
		{ 78, 162, 51 },
		{ 78, 163, 51 },
		{ 78, 164, 51 },
		{ 78, 165, 51 },
		{ 78, 166, 51 },
		{ 78, 167, 51 },
		{ 79, 124, 51 },
		{ 79, 125, 51 },
		{ 79, 126, 51 },
		{ 79, 127, 51 },
		{ 79, 128, 51 },
		{ 79, 129, 51 },
		{ 79, 130, 51 },
		{ 79, 131, 51 },
		{ 79, 132, 51 },
		{ 79, 133, 51 },
		{ 79, 134, 51 },
		{ 79, 135, 51 },
		{ 79, 136, 51 },
		{ 79, 137, 51 },
		{ 79, 138, 51 },
		{ 79, 139, 51 },
		{ 79, 140, 51 },
		{ 79, 141, 51 },
		{ 79, 142, 51 },
		{ 79, 143, 51 },
		{ 79, 144, 51 },
		{ 79, 145, 51 },
		{ 79, 146, 51 },
		{ 79, 147, 51 },
		{ 79, 148, 51 },
		{ 79, 149, 51 },
		{ 79, 150, 51 },
		{ 79, 151, 51 },
		{ 79, 152, 51 },
		{ 79, 153, 51 },
		{ 79, 154, 51 },
		{ 79, 155, 51 },
		{ 79, 156, 51 },
		{ 79, 157, 51 },
		{ 79, 158, 51 },
		{ 79, 159, 51 },
		{ 79, 160, 51 },
		{ 79, 161, 51 },
		{ 79, 162, 51 },
		{ 79, 163, 51 },
		{ 79, 164, 51 },
		{ 79, 165, 51 },
		{ 79, 166, 51 },
		{ 79, 167, 51 },
		{ 80, 125, 51 },
		{ 80, 126, 51 },
		{ 80, 127, 51 },
		{ 80, 128, 51 },
		{ 80, 129, 51 },
		{ 80, 130, 51 },
		{ 80, 131, 51 },
		{ 80, 132, 51 },
		{ 80, 133, 51 },
		{ 80, 134, 51 },
		{ 80, 135, 51 },
		{ 80, 136, 51 },
		{ 80, 137, 51 },
		{ 80, 138, 51 },
		{ 80, 139, 51 },
		{ 80, 140, 51 },
		{ 80, 141, 51 },
		{ 80, 142, 51 },
		{ 80, 143, 51 },
		{ 80, 144, 51 },
		{ 80, 145, 51 },
		{ 80, 146, 51 },
		{ 80, 147, 51 },
		{ 80, 148, 51 },
		{ 80, 149, 51 },
		{ 80, 150, 51 },
		{ 80, 151, 51 },
		{ 80, 152, 51 },
		{ 80, 153, 51 },
		{ 80, 154, 51 },
		{ 80, 155, 51 },
		{ 80, 156, 51 },
		{ 80, 157, 51 },
		{ 80, 158, 51 },
		{ 80, 159, 51 },
		{ 80, 160, 51 },
		{ 80, 161, 51 },
		{ 80, 162, 51 },
		{ 80, 163, 51 },
		{ 80, 164, 51 },
		{ 80, 165, 51 },
		{ 80, 166, 51 },
		{ 80, 167, 51 },
		{ 80, 168, 51 },
		{ 81, 126, 51 },
		{ 81, 127, 51 },
		{ 81, 128, 51 },
		{ 81, 129, 51 },
		{ 81, 130, 51 },
		{ 81, 131, 51 },
		{ 81, 132, 51 },
		{ 81, 133, 51 },
		{ 81, 134, 51 },
		{ 81, 135, 51 },
		{ 81, 136, 51 },
		{ 81, 137, 51 },
		{ 81, 138, 51 },
		{ 81, 139, 51 },
		{ 81, 140, 51 },
		{ 81, 141, 51 },
		{ 81, 142, 51 },
		{ 81, 143, 51 },
		{ 81, 144, 51 },
		{ 81, 145, 51 },
		{ 81, 146, 51 },
		{ 81, 147, 51 },
		{ 81, 148, 51 },
		{ 81, 149, 51 },
		{ 81, 150, 51 },
		{ 81, 151, 51 },
		{ 81, 152, 51 },
		{ 81, 153, 51 },
		{ 81, 154, 51 },
		{ 81, 155, 51 },
		{ 81, 156, 51 },
		{ 81, 157, 51 },
		{ 81, 158, 51 },
		{ 81, 159, 51 },
		{ 81, 160, 51 },
		{ 81, 161, 51 },
		{ 81, 162, 51 },
		{ 81, 163, 51 },
		{ 81, 164, 51 },
		{ 81, 165, 51 },
		{ 81, 166, 51 },
		{ 81, 167, 51 },
		{ 81, 168, 51 },
		{ 82, 126, 51 },
		{ 82, 127, 51 },
		{ 82, 128, 51 },
		{ 82, 129, 51 },
		{ 82, 130, 51 },
		{ 82, 131, 51 },
		{ 82, 132, 51 },
		{ 82, 133, 51 },
		{ 82, 134, 51 },
		{ 82, 135, 51 },
		{ 82, 136, 51 },
		{ 82, 137, 51 },
		{ 82, 138, 51 },
		{ 82, 139, 51 },
		{ 82, 140, 51 },
		{ 82, 141, 51 },
		{ 82, 142, 51 },
		{ 82, 143, 51 },
		{ 82, 144, 51 },
		{ 82, 145, 51 },
		{ 82, 146, 51 },
		{ 82, 147, 51 },
		{ 82, 148, 51 },
		{ 82, 149, 51 },
		{ 82, 150, 51 },
		{ 82, 151, 51 },
		{ 82, 152, 51 },
		{ 82, 153, 51 },
		{ 82, 154, 51 },
		{ 82, 155, 51 },
		{ 82, 156, 51 },
		{ 82, 157, 51 },
		{ 82, 158, 51 },
		{ 82, 159, 51 },
		{ 82, 160, 51 },
		{ 82, 161, 51 },
		{ 82, 162, 51 },
		{ 82, 163, 51 },
		{ 82, 164, 51 },
		{ 82, 165, 51 },
		{ 82, 166, 51 },
		{ 82, 167, 51 },
		{ 82, 168, 51 },
		{ 82, 169, 51 },
		{ 83, 127, 51 },
		{ 83, 128, 51 },
		{ 83, 129, 51 },
		{ 83, 130, 51 },
		{ 83, 131, 51 },
		{ 83, 132, 51 },
		{ 83, 133, 51 },
		{ 83, 134, 51 },
		{ 83, 135, 51 },
		{ 83, 136, 51 },
		{ 83, 137, 51 },
		{ 83, 138, 51 },
		{ 83, 139, 51 },
		{ 83, 140, 51 },
		{ 83, 141, 51 },
		{ 83, 142, 51 },
		{ 83, 143, 51 },
		{ 83, 144, 51 },
		{ 83, 145, 51 },
		{ 83, 146, 51 },
		{ 83, 147, 51 },
		{ 83, 148, 51 },
		{ 83, 149, 51 },
		{ 83, 150, 51 },
		{ 83, 151, 51 },
		{ 83, 152, 51 },
		{ 83, 153, 51 },
		{ 83, 154, 51 },
		{ 83, 155, 51 },
		{ 83, 156, 51 },
		{ 83, 157, 51 },
		{ 83, 158, 51 },
		{ 83, 159, 51 },
		{ 83, 160, 51 },
		{ 83, 161, 51 },
		{ 83, 162, 51 },
		{ 83, 163, 51 },
		{ 83, 164, 51 },
		{ 83, 165, 51 },
		{ 83, 166, 51 },
		{ 83, 167, 51 },
		{ 83, 168, 51 },
		{ 83, 169, 51 },
		{ 84, 128, 51 },
		{ 84, 129, 51 },
		{ 84, 130, 51 },
		{ 84, 131, 51 },
		{ 84, 132, 51 },
		{ 84, 133, 51 },
		{ 84, 134, 51 },
		{ 84, 135, 51 },
		{ 84, 136, 51 },
		{ 84, 137, 51 },
		{ 84, 138, 51 },
		{ 84, 139, 51 },
		{ 84, 140, 51 },
		{ 84, 141, 51 },
		{ 84, 142, 51 },
		{ 84, 143, 51 },
		{ 84, 144, 51 },
		{ 84, 145, 51 },
		{ 84, 146, 51 },
		{ 84, 147, 51 },
		{ 84, 148, 51 },
		{ 84, 149, 51 },
		{ 84, 150, 51 },
		{ 84, 151, 51 },
		{ 84, 152, 51 },
		{ 84, 153, 51 },
		{ 84, 154, 51 },
		{ 84, 155, 51 },
		{ 84, 156, 51 },
		{ 84, 157, 51 },
		{ 84, 158, 51 },
		{ 84, 159, 51 },
		{ 84, 160, 51 },
		{ 84, 161, 51 },
		{ 84, 162, 51 },
		{ 84, 163, 51 },
		{ 84, 164, 51 },
		{ 84, 165, 51 },
		{ 84, 166, 51 },
		{ 84, 167, 51 },
		{ 84, 168, 51 },
		{ 84, 169, 51 },
		{ 84, 170, 51 },
		{ 85, 129, 51 },
		{ 85, 130, 51 },
		{ 85, 131, 51 },
		{ 85, 132, 51 },
		{ 85, 133, 51 },
		{ 85, 134, 51 },
		{ 85, 135, 51 },
		{ 85, 136, 51 },
		{ 85, 137, 51 },
		{ 85, 138, 51 },
		{ 85, 139, 51 },
		{ 85, 140, 51 },
		{ 85, 141, 51 },
		{ 85, 142, 51 },
		{ 85, 143, 51 },
		{ 85, 144, 51 },
		{ 85, 145, 51 },
		{ 85, 146, 51 },
		{ 85, 147, 51 },
		{ 85, 148, 51 },
		{ 85, 149, 51 },
		{ 85, 150, 51 },
		{ 85, 151, 51 },
		{ 85, 152, 51 },
		{ 85, 153, 51 },
		{ 85, 154, 51 },
		{ 85, 155, 51 },
		{ 85, 156, 51 },
		{ 85, 157, 51 },
		{ 85, 158, 51 },
		{ 85, 159, 51 },
		{ 85, 160, 51 },
		{ 85, 161, 51 },
		{ 85, 162, 51 },
		{ 85, 163, 51 },
		{ 85, 164, 51 },
		{ 85, 165, 51 },
		{ 85, 166, 51 },
		{ 85, 167, 51 },
		{ 85, 168, 51 },
		{ 85, 169, 51 },
		{ 85, 170, 51 },
		{ 86, 129, 51 },
		{ 86, 130, 51 },
		{ 86, 131, 51 },
		{ 86, 132, 51 },
		{ 86, 133, 51 },
		{ 86, 134, 51 },
		{ 86, 135, 51 },
		{ 86, 136, 51 },
		{ 86, 137, 51 },
		{ 86, 138, 51 },
		{ 86, 139, 51 },
		{ 86, 140, 51 },
		{ 86, 141, 51 },
		{ 86, 142, 51 },
		{ 86, 143, 51 },
		{ 86, 144, 51 },
		{ 86, 145, 51 },
		{ 86, 146, 51 },
		{ 86, 147, 51 },
		{ 86, 148, 51 },
		{ 86, 149, 51 },
		{ 86, 150, 51 },
		{ 86, 151, 51 },
		{ 86, 152, 51 },
		{ 86, 153, 51 },
		{ 86, 154, 51 },
		{ 86, 155, 51 },
		{ 86, 156, 51 },
		{ 86, 157, 51 },
		{ 86, 158, 51 },
		{ 86, 159, 51 },
		{ 86, 160, 51 },
		{ 86, 161, 51 },
		{ 86, 162, 51 },
		{ 86, 163, 51 },
		{ 86, 164, 51 },
		{ 86, 165, 51 },
		{ 86, 166, 51 },
		{ 86, 167, 51 },
		{ 86, 168, 51 },
		{ 86, 169, 51 },
		{ 86, 170, 51 },
		{ 86, 171, 51 },
		{ 87, 130, 51 },
		{ 87, 131, 51 },
		{ 87, 132, 51 },
		{ 87, 133, 51 },
		{ 87, 134, 51 },
		{ 87, 135, 51 },
		{ 87, 136, 51 },
		{ 87, 137, 51 },
		{ 87, 138, 51 },
		{ 87, 139, 51 },
		{ 87, 140, 51 },
		{ 87, 141, 51 },
		{ 87, 142, 51 },
		{ 87, 143, 51 },
		{ 87, 144, 51 },
		{ 87, 145, 51 },
		{ 87, 146, 51 },
		{ 87, 147, 51 },
		{ 87, 148, 51 },
		{ 87, 149, 51 },
		{ 87, 150, 51 },
		{ 87, 151, 51 },
		{ 87, 152, 51 },
		{ 87, 153, 51 },
		{ 87, 154, 51 },
		{ 87, 155, 51 },
		{ 87, 156, 51 },
		{ 87, 157, 51 },
		{ 87, 158, 51 },
		{ 87, 159, 51 },
		{ 87, 160, 51 },
		{ 87, 161, 51 },
		{ 87, 162, 51 },
		{ 87, 163, 51 },
		{ 87, 164, 51 },
		{ 87, 165, 51 },
		{ 87, 166, 51 },
		{ 87, 167, 51 },
		{ 87, 168, 51 },
		{ 87, 169, 51 },
		{ 87, 170, 51 },
		{ 87, 171, 51 },
		{ 88, 131, 51 },
		{ 88, 132, 51 },
		{ 88, 133, 51 },
		{ 88, 134, 51 },
		{ 88, 135, 51 },
		{ 88, 136, 51 },
		{ 88, 137, 51 },
		{ 88, 138, 51 },
		{ 88, 139, 51 },
		{ 88, 140, 51 },
		{ 88, 141, 51 },
		{ 88, 142, 51 },
		{ 88, 143, 51 },
		{ 88, 144, 51 },
		{ 88, 145, 51 },
		{ 88, 146, 51 },
		{ 88, 147, 51 },
		{ 88, 148, 51 },
		{ 88, 149, 51 },
		{ 88, 150, 51 },
		{ 88, 151, 51 },
		{ 88, 152, 51 },
		{ 88, 153, 51 },
		{ 88, 154, 51 },
		{ 88, 155, 51 },
		{ 88, 156, 51 },
		{ 88, 157, 51 },
		{ 88, 158, 51 },
		{ 88, 159, 51 },
		{ 88, 160, 51 },
		{ 88, 161, 51 },
		{ 88, 162, 51 },
		{ 88, 163, 51 },
		{ 88, 164, 51 },
		{ 88, 165, 51 },
		{ 88, 166, 51 },
		{ 88, 167, 51 },
		{ 88, 168, 51 },
		{ 88, 169, 51 },
		{ 88, 170, 51 },
		{ 88, 171, 51 },
		{ 88, 172, 51 },
		{ 89, 132, 51 },
		{ 89, 133, 51 },
		{ 89, 134, 51 },
		{ 89, 135, 51 },
		{ 89, 136, 51 },
		{ 89, 137, 51 },
		{ 89, 138, 51 },
		{ 89, 139, 51 },
		{ 89, 140, 51 },
		{ 89, 141, 51 },
		{ 89, 142, 51 },
		{ 89, 143, 51 },
		{ 89, 144, 51 },
		{ 89, 145, 51 },
		{ 89, 146, 51 },
		{ 89, 147, 51 },
		{ 89, 148, 51 },
		{ 89, 149, 51 },
		{ 89, 150, 51 },
		{ 89, 151, 51 },
		{ 89, 152, 51 },
		{ 89, 153, 51 },
		{ 89, 154, 51 },
		{ 89, 155, 51 },
		{ 89, 156, 51 },
		{ 89, 157, 51 },
		{ 89, 158, 51 },
		{ 89, 159, 51 },
		{ 89, 160, 51 },
		{ 89, 161, 51 },
		{ 89, 162, 51 },
		{ 89, 163, 51 },
		{ 89, 164, 51 },
		{ 89, 165, 51 },
		{ 89, 166, 51 },
		{ 89, 167, 51 },
		{ 89, 168, 51 },
		{ 89, 169, 51 },
		{ 89, 170, 51 },
		{ 89, 171, 51 },
		{ 89, 172, 51 },
		{ 90, 132, 51 },
		{ 90, 133, 51 },
		{ 90, 134, 51 },
		{ 90, 135, 51 },
		{ 90, 136, 51 },
		{ 90, 137, 51 },
		{ 90, 138, 51 },
		{ 90, 139, 51 },
		{ 90, 140, 51 },
		{ 90, 141, 51 },
		{ 90, 142, 51 },
		{ 90, 143, 51 },
		{ 90, 144, 51 },
		{ 90, 145, 51 },
		{ 90, 146, 51 },
		{ 90, 147, 51 },
		{ 90, 148, 51 },
		{ 90, 149, 51 },
		{ 90, 150, 51 },
		{ 90, 151, 51 },
		{ 90, 152, 51 },
		{ 90, 153, 51 },
		{ 90, 154, 51 },
		{ 90, 155, 51 },
		{ 90, 156, 51 },
		{ 90, 157, 51 },
		{ 90, 158, 51 },
		{ 90, 159, 51 },
		{ 90, 160, 51 },
		{ 90, 161, 51 },
		{ 90, 162, 51 },
		{ 90, 163, 51 },
		{ 90, 164, 51 },
		{ 90, 165, 51 },
		{ 90, 166, 51 },
		{ 90, 167, 51 },
		{ 90, 168, 51 },
		{ 90, 169, 51 },
		{ 90, 170, 51 },
		{ 90, 171, 51 },
		{ 90, 172, 51 },
		{ 90, 173, 51 },
		{ 91, 133, 51 },
		{ 91, 134, 51 },
		{ 91, 135, 51 },
		{ 91, 136, 51 },
		{ 91, 137, 51 },
		{ 91, 138, 51 },
		{ 91, 139, 51 },
		{ 91, 140, 51 },
		{ 91, 141, 51 },
		{ 91, 142, 51 },
		{ 91, 143, 51 },
		{ 91, 144, 51 },
		{ 91, 145, 51 },
		{ 91, 146, 51 },
		{ 91, 147, 51 },
		{ 91, 148, 51 },
		{ 91, 149, 51 },
		{ 91, 150, 51 },
		{ 91, 151, 51 },
		{ 91, 152, 51 },
		{ 91, 153, 51 },
		{ 91, 154, 51 },
		{ 91, 155, 51 },
		{ 91, 156, 51 },
		{ 91, 157, 51 },
		{ 91, 158, 51 },
		{ 91, 159, 51 },
		{ 91, 160, 51 },
		{ 91, 161, 51 },
		{ 91, 162, 51 },
		{ 91, 163, 51 },
		{ 91, 164, 51 },
		{ 91, 165, 51 },
		{ 91, 166, 51 },
		{ 91, 167, 51 },
		{ 91, 168, 51 },
		{ 91, 169, 51 },
		{ 91, 170, 51 },
		{ 91, 171, 51 },
		{ 91, 172, 51 },
		{ 91, 173, 51 },
		{ 92, 134, 51 },
		{ 92, 135, 51 },
		{ 92, 136, 51 },
		{ 92, 137, 51 },
		{ 92, 138, 51 },
		{ 92, 139, 51 },
		{ 92, 140, 51 },
		{ 92, 141, 51 },
		{ 92, 142, 51 },
		{ 92, 143, 51 },
		{ 92, 144, 51 },
		{ 92, 145, 51 },
		{ 92, 146, 51 },
		{ 92, 147, 51 },
		{ 92, 148, 51 },
		{ 92, 149, 51 },
		{ 92, 150, 51 },
		{ 92, 151, 51 },
		{ 92, 152, 51 },
		{ 92, 153, 51 },
		{ 92, 154, 51 },
		{ 92, 155, 51 },
		{ 92, 156, 51 },
		{ 92, 157, 51 },
		{ 92, 158, 51 },
		{ 92, 159, 51 },
		{ 92, 160, 51 },
		{ 92, 161, 51 },
		{ 92, 162, 51 },
		{ 92, 163, 51 },
		{ 92, 164, 51 },
		{ 92, 165, 51 },
		{ 92, 166, 51 },
		{ 92, 167, 51 },
		{ 92, 168, 51 },
		{ 92, 169, 51 },
		{ 92, 170, 51 },
		{ 92, 171, 51 },
		{ 92, 172, 51 },
		{ 92, 173, 51 },
		{ 92, 174, 51 },
		{ 93, 135, 51 },
		{ 93, 136, 51 },
		{ 93, 137, 51 },
		{ 93, 138, 51 },
		{ 93, 139, 51 },
		{ 93, 140, 51 },
		{ 93, 141, 51 },
		{ 93, 142, 51 },
		{ 93, 143, 51 },
		{ 93, 144, 51 },
		{ 93, 145, 51 },
		{ 93, 146, 51 },
		{ 93, 147, 51 },
		{ 93, 148, 51 },
		{ 93, 149, 51 },
		{ 93, 150, 51 },
		{ 93, 151, 51 },
		{ 93, 152, 51 },
		{ 93, 153, 51 },
		{ 93, 154, 51 },
		{ 93, 155, 51 },
		{ 93, 156, 51 },
		{ 93, 157, 51 },
		{ 93, 158, 51 },
		{ 93, 159, 51 },
		{ 93, 160, 51 },
		{ 93, 161, 51 },
		{ 93, 162, 51 },
		{ 93, 163, 51 },
		{ 93, 164, 51 },
		{ 93, 165, 51 },
		{ 93, 166, 51 },
		{ 93, 167, 51 },
		{ 93, 168, 51 },
		{ 93, 169, 51 },
		{ 93, 170, 51 },
		{ 93, 171, 51 },
		{ 93, 172, 51 },
		{ 93, 173, 51 },
		{ 93, 174, 51 },
		{ 94, 135, 51 },
		{ 94, 136, 51 },
		{ 94, 137, 51 },
		{ 94, 138, 51 },
		{ 94, 139, 51 },
		{ 94, 140, 51 },
		{ 94, 141, 51 },
		{ 94, 142, 51 },
		{ 94, 143, 51 },
		{ 94, 144, 51 },
		{ 94, 145, 51 },
		{ 94, 146, 51 },
		{ 94, 147, 51 },
		{ 94, 148, 51 },
		{ 94, 149, 51 },
		{ 94, 150, 51 },
		{ 94, 151, 51 },
		{ 94, 152, 51 },
		{ 94, 153, 51 },
		{ 94, 154, 51 },
		{ 94, 155, 51 },
		{ 94, 156, 51 },
		{ 94, 157, 51 },
		{ 94, 158, 51 },
		{ 94, 159, 51 },
		{ 94, 160, 51 },
		{ 94, 161, 51 },
		{ 94, 162, 51 },
		{ 94, 163, 51 },
		{ 94, 164, 51 },
		{ 94, 165, 51 },
		{ 94, 166, 51 },
		{ 94, 167, 51 },
		{ 94, 168, 51 },
		{ 94, 169, 51 },
		{ 94, 170, 51 },
		{ 94, 171, 51 },
		{ 94, 172, 51 },
		{ 94, 173, 51 },
		{ 94, 174, 51 },
		{ 94, 175, 51 },
		{ 95, 136, 51 },
		{ 95, 137, 51 },
		{ 95, 138, 51 },
		{ 95, 139, 51 },
		{ 95, 140, 51 },
		{ 95, 141, 51 },
		{ 95, 142, 51 },
		{ 95, 143, 51 },
		{ 95, 144, 51 },
		{ 95, 145, 51 },
		{ 95, 146, 51 },
		{ 95, 147, 51 },
		{ 95, 148, 51 },
		{ 95, 149, 51 },
		{ 95, 150, 51 },
		{ 95, 151, 51 },
		{ 95, 152, 51 },
		{ 95, 153, 51 },
		{ 95, 154, 51 },
		{ 95, 155, 51 },
		{ 95, 156, 51 },
		{ 95, 157, 51 },
		{ 95, 158, 51 },
		{ 95, 159, 51 },
		{ 95, 160, 51 },
		{ 95, 161, 51 },
		{ 95, 162, 51 },
		{ 95, 163, 51 },
		{ 95, 164, 51 },
		{ 95, 165, 51 },
		{ 95, 166, 51 },
		{ 95, 167, 51 },
		{ 95, 168, 51 },
		{ 95, 169, 51 },
		{ 95, 170, 51 },
		{ 95, 171, 51 },
		{ 95, 172, 51 },
		{ 95, 173, 51 },
		{ 95, 174, 51 },
		{ 95, 175, 51 },
		{ 96, 137, 51 },
		{ 96, 138, 51 },
		{ 96, 139, 51 },
		{ 96, 140, 51 },
		{ 96, 141, 51 },
		{ 96, 142, 51 },
		{ 96, 143, 51 },
		{ 96, 144, 51 },
		{ 96, 145, 51 },
		{ 96, 146, 51 },
		{ 96, 147, 51 },
		{ 96, 148, 51 },
		{ 96, 149, 51 },
		{ 96, 150, 51 },
		{ 96, 151, 51 },
		{ 96, 152, 51 },
		{ 96, 153, 51 },
		{ 96, 154, 51 },
		{ 96, 155, 51 },
		{ 96, 156, 51 },
		{ 96, 157, 51 },
		{ 96, 158, 51 },
		{ 96, 159, 51 },
		{ 96, 160, 51 },
		{ 96, 161, 51 },
		{ 96, 162, 51 },
		{ 96, 163, 51 },
		{ 96, 164, 51 },
		{ 96, 165, 51 },
		{ 96, 166, 51 },
		{ 96, 167, 51 },
		{ 96, 168, 51 },
		{ 96, 169, 51 },
		{ 96, 170, 51 },
		{ 96, 171, 51 },
		{ 96, 172, 51 },
		{ 96, 173, 51 },
		{ 96, 174, 51 },
		{ 96, 175, 51 },
		{ 96, 176, 51 },
		{ 97, 138, 51 },
		{ 97, 139, 51 },
		{ 97, 140, 51 },
		{ 97, 141, 51 },
		{ 97, 142, 51 },
		{ 97, 143, 51 },
		{ 97, 144, 51 },
		{ 97, 145, 51 },
		{ 97, 146, 51 },
		{ 97, 147, 51 },
		{ 97, 148, 51 },
		{ 97, 149, 51 },
		{ 97, 150, 51 },
		{ 97, 151, 51 },
		{ 97, 152, 51 },
		{ 97, 153, 51 },
		{ 97, 154, 51 },
		{ 97, 155, 51 },
		{ 97, 156, 51 },
		{ 97, 157, 51 },
		{ 97, 158, 51 },
		{ 97, 159, 51 },
		{ 97, 160, 51 },
		{ 97, 161, 51 },
		{ 97, 162, 51 },
		{ 97, 163, 51 },
		{ 97, 164, 51 },
		{ 97, 165, 51 },
		{ 97, 166, 51 },
		{ 97, 167, 51 },
		{ 97, 168, 51 },
		{ 97, 169, 51 },
		{ 97, 170, 51 },
		{ 97, 171, 51 },
		{ 97, 172, 51 },
		{ 97, 173, 51 },
		{ 97, 174, 51 },
		{ 97, 175, 51 },
		{ 97, 176, 51 },
		{ 98, 138, 51 },
		{ 98, 139, 51 },
		{ 98, 140, 51 },
		{ 98, 141, 51 },
		{ 98, 142, 51 },
		{ 98, 143, 51 },
		{ 98, 144, 51 },
		{ 98, 145, 51 },
		{ 98, 146, 51 },
		{ 98, 147, 51 },
		{ 98, 148, 51 },
		{ 98, 149, 51 },
		{ 98, 150, 51 },
		{ 98, 151, 51 },
		{ 98, 152, 51 },
		{ 98, 153, 51 },
		{ 98, 154, 51 },
		{ 98, 155, 51 },
		{ 98, 156, 51 },
		{ 98, 157, 51 },
		{ 98, 158, 51 },
		{ 98, 159, 51 },
		{ 98, 160, 51 },
		{ 98, 161, 51 },
		{ 98, 162, 51 },
		{ 98, 163, 51 },
		{ 98, 164, 51 },
		{ 98, 165, 51 },
		{ 98, 166, 51 },
		{ 98, 167, 51 },
		{ 98, 168, 51 },
		{ 98, 169, 51 },
		{ 98, 170, 51 },
		{ 98, 171, 51 },
		{ 98, 172, 51 },
		{ 98, 173, 51 },
		{ 98, 174, 51 },
		{ 98, 175, 51 },
		{ 98, 176, 51 },
		{ 98, 177, 51 },
		{ 99, 139, 51 },
		{ 99, 140, 51 },
		{ 99, 141, 51 },
		{ 99, 142, 51 },
		{ 99, 143, 51 },
		{ 99, 144, 51 },
		{ 99, 145, 51 },
		{ 99, 146, 51 },
		{ 99, 147, 51 },
		{ 99, 148, 51 },
		{ 99, 149, 51 },
		{ 99, 150, 51 },
		{ 99, 151, 51 },
		{ 99, 152, 51 },
		{ 99, 153, 51 },
		{ 99, 154, 51 },
		{ 99, 155, 51 },
		{ 99, 156, 51 },
		{ 99, 157, 51 },
		{ 99, 158, 51 },
		{ 99, 159, 51 },
		{ 99, 160, 51 },
		{ 99, 161, 51 },
		{ 99, 162, 51 },
		{ 99, 163, 51 },
		{ 99, 164, 51 },
		{ 99, 165, 51 },
		{ 99, 166, 51 },
		{ 99, 167, 51 },
		{ 99, 168, 51 },
		{ 99, 169, 51 },
		{ 99, 170, 51 },
		{ 99, 171, 51 },
		{ 99, 172, 51 },
		{ 99, 173, 51 },
		{ 99, 174, 51 },
		{ 99, 175, 51 },
		{ 99, 176, 51 },
		{ 99, 177, 51 },
		{ 100, 140, 51 },
		{ 100, 141, 51 },
		{ 100, 142, 51 },
		{ 100, 143, 51 },
		{ 100, 144, 51 },
		{ 100, 145, 51 },
		{ 100, 146, 51 },
		{ 100, 147, 51 },
		{ 100, 148, 51 },
		{ 100, 149, 51 },
		{ 100, 150, 51 },
		{ 100, 151, 51 },
		{ 100, 152, 51 },
		{ 100, 153, 51 },
		{ 100, 154, 51 },
		{ 100, 155, 51 },
		{ 100, 156, 51 },
		{ 100, 157, 51 },
		{ 100, 158, 51 },
		{ 100, 159, 51 },
		{ 100, 160, 51 },
		{ 100, 161, 51 },
		{ 100, 162, 51 },
		{ 100, 163, 51 },
		{ 100, 164, 51 },
		{ 100, 165, 51 },
		{ 100, 166, 51 },
		{ 100, 167, 51 },
		{ 100, 168, 51 },
		{ 100, 169, 51 },
		{ 100, 170, 51 },
		{ 100, 171, 51 },
		{ 100, 172, 51 },
		{ 100, 173, 51 },
		{ 100, 174, 51 },
		{ 100, 175, 51 },
		{ 100, 176, 51 },
		{ 100, 177, 51 },
		{ 100, 178, 51 },
		{ 101, 141, 51 },
		{ 101, 142, 51 },
		{ 101, 143, 51 },
		{ 101, 144, 51 },
		{ 101, 145, 51 },
		{ 101, 146, 51 },
		{ 101, 147, 51 },
		{ 101, 148, 51 },
		{ 101, 149, 51 },
		{ 101, 150, 51 },
		{ 101, 151, 51 },
		{ 101, 152, 51 },
		{ 101, 153, 51 },
		{ 101, 154, 51 },
		{ 101, 155, 51 },
		{ 101, 156, 51 },
		{ 101, 157, 51 },
		{ 101, 158, 51 },
		{ 101, 159, 51 },
		{ 101, 160, 51 },
		{ 101, 161, 51 },
		{ 101, 162, 51 },
		{ 101, 163, 51 },
		{ 101, 164, 51 },
		{ 101, 165, 51 },
		{ 101, 166, 51 },
		{ 101, 167, 51 },
		{ 101, 168, 51 },
		{ 101, 169, 51 },
		{ 101, 170, 51 },
		{ 101, 171, 51 },
		{ 101, 172, 51 },
		{ 101, 173, 51 },
		{ 101, 174, 51 },
		{ 101, 175, 51 },
		{ 101, 176, 51 },
		{ 101, 177, 51 },
		{ 101, 178, 51 },
		{ 102, 141, 51 },
		{ 102, 142, 51 },
		{ 102, 143, 51 },
		{ 102, 144, 51 },
		{ 102, 145, 51 },
		{ 102, 146, 51 },
		{ 102, 147, 51 },
		{ 102, 148, 51 },
		{ 102, 149, 51 },
		{ 102, 150, 51 },
		{ 102, 151, 51 },
		{ 102, 152, 51 },
		{ 102, 153, 51 },
		{ 102, 154, 51 },
		{ 102, 155, 51 },
		{ 102, 156, 51 },
		{ 102, 157, 51 },
		{ 102, 158, 51 },
		{ 102, 159, 51 },
		{ 102, 160, 51 },
		{ 102, 161, 51 },
		{ 102, 162, 51 },
		{ 102, 163, 51 },
		{ 102, 164, 51 },
		{ 102, 165, 51 },
		{ 102, 166, 51 },
		{ 102, 167, 51 },
		{ 102, 168, 51 },
		{ 102, 169, 51 },
		{ 102, 170, 51 },
		{ 102, 171, 51 },
		{ 102, 172, 51 },
		{ 102, 173, 51 },
		{ 102, 174, 51 },
		{ 102, 175, 51 },
		{ 102, 176, 51 },
		{ 102, 177, 51 },
		{ 102, 178, 51 },
		{ 102, 179, 51 },
		{ 103, 142, 51 },
		{ 103, 143, 51 },
		{ 103, 144, 51 },
		{ 103, 145, 51 },
		{ 103, 146, 51 },
		{ 103, 147, 51 },
		{ 103, 148, 51 },
		{ 103, 149, 51 },
		{ 103, 150, 51 },
		{ 103, 151, 51 },
		{ 103, 152, 51 },
		{ 103, 153, 51 },
		{ 103, 154, 51 },
		{ 103, 155, 51 },
		{ 103, 156, 51 },
		{ 103, 157, 51 },
		{ 103, 158, 51 },
		{ 103, 159, 51 },
		{ 103, 160, 51 },
		{ 103, 161, 51 },
		{ 103, 162, 51 },
		{ 103, 163, 51 },
		{ 103, 164, 51 },
		{ 103, 165, 51 },
		{ 103, 166, 51 },
		{ 103, 167, 51 },
		{ 103, 168, 51 },
		{ 103, 169, 51 },
		{ 103, 170, 51 },
		{ 103, 171, 51 },
		{ 103, 172, 51 },
		{ 103, 173, 51 },
		{ 103, 174, 51 },
		{ 103, 175, 51 },
		{ 103, 176, 51 },
		{ 103, 177, 51 },
		{ 103, 178, 51 },
		{ 103, 179, 51 },
		{ 104, 143, 51 },
		{ 104, 144, 51 },
		{ 104, 145, 51 },
		{ 104, 146, 51 },
		{ 104, 147, 51 },
		{ 104, 148, 51 },
		{ 104, 149, 51 },
		{ 104, 150, 51 },
		{ 104, 151, 51 },
		{ 104, 152, 51 },
		{ 104, 153, 51 },
		{ 104, 154, 51 },
		{ 104, 155, 51 },
		{ 104, 156, 51 },
		{ 104, 157, 51 },
		{ 104, 158, 51 },
		{ 104, 159, 51 },
		{ 104, 160, 51 },
		{ 104, 161, 51 },
		{ 104, 162, 51 },
		{ 104, 163, 51 },
		{ 104, 164, 51 },
		{ 104, 165, 51 },
		{ 104, 166, 51 },
		{ 104, 167, 51 },
		{ 104, 168, 51 },
		{ 104, 169, 51 },
		{ 104, 170, 51 },
		{ 104, 171, 51 },
		{ 104, 172, 51 },
		{ 104, 173, 51 },
		{ 104, 174, 51 },
		{ 104, 175, 51 },
		{ 104, 176, 51 },
		{ 104, 177, 51 },
		{ 104, 178, 51 },
		{ 104, 179, 51 },
		{ 104, 180, 51 },
		{ 105, 144, 51 },
		{ 105, 145, 51 },
		{ 105, 146, 51 },
		{ 105, 147, 51 },
		{ 105, 148, 51 },
		{ 105, 149, 51 },
		{ 105, 150, 51 },
		{ 105, 151, 51 },
		{ 105, 152, 51 },
		{ 105, 153, 51 },
		{ 105, 154, 51 },
		{ 105, 155, 51 },
		{ 105, 156, 51 },
		{ 105, 157, 51 },
		{ 105, 158, 51 },
		{ 105, 159, 51 },
		{ 105, 160, 51 },
		{ 105, 161, 51 },
		{ 105, 162, 51 },
		{ 105, 163, 51 },
		{ 105, 164, 51 },
		{ 105, 165, 51 },
		{ 105, 166, 51 },
		{ 105, 167, 51 },
		{ 105, 168, 51 },
		{ 105, 169, 51 },
		{ 105, 170, 51 },
		{ 105, 171, 51 },
		{ 105, 172, 51 },
		{ 105, 173, 51 },
		{ 105, 174, 51 },
		{ 105, 175, 51 },
		{ 105, 176, 51 },
		{ 105, 177, 51 },
		{ 105, 178, 51 },
		{ 105, 179, 51 },
		{ 105, 180, 51 },
		{ 106, 144, 51 },
		{ 106, 145, 51 },
		{ 106, 146, 51 },
		{ 106, 147, 51 },
		{ 106, 148, 51 },
		{ 106, 149, 51 },
		{ 106, 150, 51 },
		{ 106, 151, 51 },
		{ 106, 152, 51 },
		{ 106, 153, 51 },
		{ 106, 154, 51 },
		{ 106, 155, 51 },
		{ 106, 156, 51 },
		{ 106, 157, 51 },
		{ 106, 158, 51 },
		{ 106, 159, 51 },
		{ 106, 160, 51 },
		{ 106, 161, 51 },
		{ 106, 162, 51 },
		{ 106, 163, 51 },
		{ 106, 164, 51 },
		{ 106, 165, 51 },
		{ 106, 166, 51 },
		{ 106, 167, 51 },
		{ 106, 168, 51 },
		{ 106, 169, 51 },
		{ 106, 170, 51 },
		{ 106, 171, 51 },
		{ 106, 172, 51 },
		{ 106, 173, 51 },
		{ 106, 174, 51 },
		{ 106, 175, 51 },
		{ 106, 176, 51 },
		{ 106, 177, 51 },
		{ 106, 178, 51 },
		{ 106, 179, 51 },
		{ 106, 180, 51 },
		{ 106, 181, 51 },
		{ 107, 145, 51 },
		{ 107, 146, 51 },
		{ 107, 147, 51 },
		{ 107, 148, 51 },
		{ 107, 149, 51 },
		{ 107, 150, 51 },
		{ 107, 151, 51 },
		{ 107, 152, 51 },
		{ 107, 153, 51 },
		{ 107, 154, 51 },
		{ 107, 155, 51 },
		{ 107, 156, 51 },
		{ 107, 157, 51 },
		{ 107, 158, 51 },
		{ 107, 159, 51 },
		{ 107, 160, 51 },
		{ 107, 161, 51 },
		{ 107, 162, 51 },
		{ 107, 163, 51 },
		{ 107, 164, 51 },
		{ 107, 165, 51 },
		{ 107, 166, 51 },
		{ 107, 167, 51 },
		{ 107, 168, 51 },
		{ 107, 169, 51 },
		{ 107, 170, 51 },
		{ 107, 171, 51 },
		{ 107, 172, 51 },
		{ 107, 173, 51 },
		{ 107, 174, 51 },
		{ 107, 175, 51 },
		{ 107, 176, 51 },
		{ 107, 177, 51 },
		{ 107, 178, 51 },
		{ 107, 179, 51 },
		{ 107, 180, 51 },
		{ 107, 181, 51 },
		{ 108, 146, 51 },
		{ 108, 147, 51 },
		{ 108, 148, 51 },
		{ 108, 149, 51 },
		{ 108, 150, 51 },
		{ 108, 151, 51 },
		{ 108, 152, 51 },
		{ 108, 153, 51 },
		{ 108, 154, 51 },
		{ 108, 155, 51 },
		{ 108, 156, 51 },
		{ 108, 157, 51 },
		{ 108, 158, 51 },
		{ 108, 159, 51 },
		{ 108, 160, 51 },
		{ 108, 161, 51 },
		{ 108, 162, 51 },
		{ 108, 163, 51 },
		{ 108, 164, 51 },
		{ 108, 165, 51 },
		{ 108, 166, 51 },
		{ 108, 167, 51 },
		{ 108, 168, 51 },
		{ 108, 169, 51 },
		{ 108, 170, 51 },
		{ 108, 171, 51 },
		{ 108, 172, 51 },
		{ 108, 173, 51 },
		{ 108, 174, 51 },
		{ 108, 175, 51 },
		{ 108, 176, 51 },
		{ 108, 177, 51 },
		{ 108, 178, 51 },
		{ 108, 179, 51 },
		{ 108, 180, 51 },
		{ 108, 181, 51 },
		{ 108, 182, 51 },
		{ 109, 147, 51 },
		{ 109, 148, 51 },
		{ 109, 149, 51 },
		{ 109, 150, 51 },
		{ 109, 151, 51 },
		{ 109, 152, 51 },
		{ 109, 153, 51 },
		{ 109, 154, 51 },
		{ 109, 155, 51 },
		{ 109, 156, 51 },
		{ 109, 157, 51 },
		{ 109, 158, 51 },
		{ 109, 159, 51 },
		{ 109, 160, 51 },
		{ 109, 161, 51 },
		{ 109, 162, 51 },
		{ 109, 163, 51 },
		{ 109, 164, 51 },
		{ 109, 165, 51 },
		{ 109, 166, 51 },
		{ 109, 167, 51 },
		{ 109, 168, 51 },
		{ 109, 169, 51 },
		{ 109, 170, 51 },
		{ 109, 171, 51 },
		{ 109, 172, 51 },
		{ 109, 173, 51 },
		{ 109, 174, 51 },
		{ 109, 175, 51 },
		{ 109, 176, 51 },
		{ 109, 177, 51 },
		{ 109, 178, 51 },
		{ 109, 179, 51 },
		{ 109, 180, 51 },
		{ 109, 181, 51 },
		{ 109, 182, 51 },
		{ 110, 147, 51 },
		{ 110, 148, 51 },
		{ 110, 149, 51 },
		{ 110, 150, 51 },
		{ 110, 151, 51 },
		{ 110, 152, 51 },
		{ 110, 153, 51 },
		{ 110, 154, 51 },
		{ 110, 155, 51 },
		{ 110, 156, 51 },
		{ 110, 157, 51 },
		{ 110, 158, 51 },
		{ 110, 159, 51 },
		{ 110, 160, 51 },
		{ 110, 161, 51 },
		{ 110, 162, 51 },
		{ 110, 163, 51 },
		{ 110, 164, 51 },
		{ 110, 165, 51 },
		{ 110, 166, 51 },
		{ 110, 167, 51 },
		{ 110, 168, 51 },
		{ 110, 169, 51 },
		{ 110, 170, 51 },
		{ 110, 171, 51 },
		{ 110, 172, 51 },
		{ 110, 173, 51 },
		{ 110, 174, 51 },
		{ 110, 175, 51 },
		{ 110, 176, 51 },
		{ 110, 177, 51 },
		{ 110, 178, 51 },
		{ 110, 179, 51 },
		{ 110, 180, 51 },
		{ 110, 181, 51 },
		{ 110, 182, 51 },
		{ 110, 183, 51 },
		{ 111, 148, 51 },
		{ 111, 149, 51 },
		{ 111, 150, 51 },
		{ 111, 151, 51 },
		{ 111, 152, 51 },
		{ 111, 153, 51 },
		{ 111, 154, 51 },
		{ 111, 155, 51 },
		{ 111, 156, 51 },
		{ 111, 157, 51 },
		{ 111, 158, 51 },
		{ 111, 159, 51 },
		{ 111, 160, 51 },
		{ 111, 161, 51 },
		{ 111, 162, 51 },
		{ 111, 163, 51 },
		{ 111, 164, 51 },
		{ 111, 165, 51 },
		{ 111, 166, 51 },
		{ 111, 167, 51 },
		{ 111, 168, 51 },
		{ 111, 169, 51 },
		{ 111, 170, 51 },
		{ 111, 171, 51 },
		{ 111, 172, 51 },
		{ 111, 173, 51 },
		{ 111, 174, 51 },
		{ 111, 175, 51 },
		{ 111, 176, 51 },
		{ 111, 177, 51 },
		{ 111, 178, 51 },
		{ 111, 179, 51 },
		{ 111, 180, 51 },
		{ 111, 181, 51 },
		{ 111, 182, 51 },
		{ 111, 183, 51 },
		{ 112, 149, 51 },
		{ 112, 150, 51 },
		{ 112, 151, 51 },
		{ 112, 152, 51 },
		{ 112, 153, 51 },
		{ 112, 154, 51 },
		{ 112, 155, 51 },
		{ 112, 156, 51 },
		{ 112, 157, 51 },
		{ 112, 158, 51 },
		{ 112, 159, 51 },
		{ 112, 160, 51 },
		{ 112, 161, 51 },
		{ 112, 162, 51 },
		{ 112, 163, 51 },
		{ 112, 164, 51 },
		{ 112, 165, 51 },
		{ 112, 166, 51 },
		{ 112, 167, 51 },
		{ 112, 168, 51 },
		{ 112, 169, 51 },
		{ 112, 170, 51 },
		{ 112, 171, 51 },
		{ 112, 172, 51 },
		{ 112, 173, 51 },
		{ 112, 174, 51 },
		{ 112, 175, 51 },
		{ 112, 176, 51 },
		{ 112, 177, 51 },
		{ 112, 178, 51 },
		{ 112, 179, 51 },
		{ 112, 180, 51 },
		{ 112, 181, 51 },
		{ 112, 182, 51 },
		{ 112, 183, 51 },
		{ 112, 184, 51 },
		{ 113, 150, 51 },
		{ 113, 151, 51 },
		{ 113, 152, 51 },
		{ 113, 153, 51 },
		{ 113, 154, 51 },
		{ 113, 155, 51 },
		{ 113, 156, 51 },
		{ 113, 157, 51 },
		{ 113, 158, 51 },
		{ 113, 159, 51 },
		{ 113, 160, 51 },
		{ 113, 161, 51 },
		{ 113, 162, 51 },
		{ 113, 163, 51 },
		{ 113, 164, 51 },
		{ 113, 165, 51 },
		{ 113, 166, 51 },
		{ 113, 167, 51 },
		{ 113, 168, 51 },
		{ 113, 169, 51 },
		{ 113, 170, 51 },
		{ 113, 171, 51 },
		{ 113, 172, 51 },
		{ 113, 173, 51 },
		{ 113, 174, 51 },
		{ 113, 175, 51 },
		{ 113, 176, 51 },
		{ 113, 177, 51 },
		{ 113, 178, 51 },
		{ 113, 179, 51 },
		{ 113, 180, 51 },
		{ 113, 181, 51 },
		{ 113, 182, 51 },
		{ 113, 183, 51 },
		{ 113, 184, 51 },
		{ 114, 150, 51 },
		{ 114, 151, 51 },
		{ 114, 152, 51 },
		{ 114, 153, 51 },
		{ 114, 154, 51 },
		{ 114, 155, 51 },
		{ 114, 156, 51 },
		{ 114, 157, 51 },
		{ 114, 158, 51 },
		{ 114, 159, 51 },
		{ 114, 160, 51 },
		{ 114, 161, 51 },
		{ 114, 162, 51 },
		{ 114, 163, 51 },
		{ 114, 164, 51 },
		{ 114, 165, 51 },
		{ 114, 166, 51 },
		{ 114, 167, 51 },
		{ 114, 168, 51 },
		{ 114, 169, 51 },
		{ 114, 170, 51 },
		{ 114, 171, 51 },
		{ 114, 172, 51 },
		{ 114, 173, 51 },
		{ 114, 174, 51 },
		{ 114, 175, 51 },
		{ 114, 176, 51 },
		{ 114, 177, 51 },
		{ 114, 178, 51 },
		{ 114, 179, 51 },
		{ 114, 180, 51 },
		{ 114, 181, 51 },
		{ 114, 182, 51 },
		{ 114, 183, 51 },
		{ 114, 184, 51 },
		{ 114, 185, 51 },
		{ 115, 151, 51 },
		{ 115, 152, 51 },
		{ 115, 153, 51 },
		{ 115, 154, 51 },
		{ 115, 155, 51 },
		{ 115, 156, 51 },
		{ 115, 157, 51 },
		{ 115, 158, 51 },
		{ 115, 159, 51 },
		{ 115, 160, 51 },
		{ 115, 161, 51 },
		{ 115, 162, 51 },
		{ 115, 163, 51 },
		{ 115, 164, 51 },
		{ 115, 165, 51 },
		{ 115, 166, 51 },
		{ 115, 167, 51 },
		{ 115, 168, 51 },
		{ 115, 169, 51 },
		{ 115, 170, 51 },
		{ 115, 171, 51 },
		{ 115, 172, 51 },
		{ 115, 173, 51 },
		{ 115, 174, 51 },
		{ 115, 175, 51 },
		{ 115, 176, 51 },
		{ 115, 177, 51 },
		{ 115, 178, 51 },
		{ 115, 179, 51 },
		{ 115, 180, 51 },
		{ 115, 181, 51 },
		{ 115, 182, 51 },
		{ 115, 183, 51 },
		{ 115, 184, 51 },
		{ 115, 185, 51 },
		{ 116, 152, 51 },
		{ 116, 153, 51 },
		{ 116, 154, 51 },
		{ 116, 155, 51 },
		{ 116, 156, 51 },
		{ 116, 157, 51 },
		{ 116, 158, 51 },
		{ 116, 159, 51 },
		{ 116, 160, 51 },
		{ 116, 161, 51 },
		{ 116, 162, 51 },
		{ 116, 163, 51 },
		{ 116, 164, 51 },
		{ 116, 165, 51 },
		{ 116, 166, 51 },
		{ 116, 167, 51 },
		{ 116, 168, 51 },
		{ 116, 169, 51 },
		{ 116, 170, 51 },
		{ 116, 171, 51 },
		{ 116, 172, 51 },
		{ 116, 173, 51 },
		{ 116, 174, 51 },
		{ 116, 175, 51 },
		{ 116, 176, 51 },
		{ 116, 177, 51 },
		{ 116, 178, 51 },
		{ 116, 179, 51 },
		{ 116, 180, 51 },
		{ 116, 181, 51 },
		{ 116, 182, 51 },
		{ 116, 183, 51 },
		{ 116, 184, 51 },
		{ 116, 185, 51 },
		{ 116, 186, 51 },
		{ 117, 153, 51 },
		{ 117, 154, 51 },
		{ 117, 155, 51 },
		{ 117, 156, 51 },
		{ 117, 157, 51 },
		{ 117, 158, 51 },
		{ 117, 159, 51 },
		{ 117, 160, 51 },
		{ 117, 161, 51 },
		{ 117, 162, 51 },
		{ 117, 163, 51 },
		{ 117, 164, 51 },
		{ 117, 165, 51 },
		{ 117, 166, 51 },
		{ 117, 167, 51 },
		{ 117, 168, 51 },
		{ 117, 169, 51 },
		{ 117, 170, 51 },
		{ 117, 171, 51 },
		{ 117, 172, 51 },
		{ 117, 173, 51 },
		{ 117, 174, 51 },
		{ 117, 175, 51 },
		{ 117, 176, 51 },
		{ 117, 177, 51 },
		{ 117, 178, 51 },
		{ 117, 179, 51 },
		{ 117, 180, 51 },
		{ 117, 181, 51 },
		{ 117, 182, 51 },
		{ 117, 183, 51 },
		{ 117, 184, 51 },
		{ 117, 185, 51 },
		{ 117, 186, 51 },
		{ 118, 153, 51 },
		{ 118, 154, 51 },
		{ 118, 155, 51 },
		{ 118, 156, 51 },
		{ 118, 157, 51 },
		{ 118, 158, 51 },
		{ 118, 159, 51 },
		{ 118, 160, 51 },
		{ 118, 161, 51 },
		{ 118, 162, 51 },
		{ 118, 163, 51 },
		{ 118, 164, 51 },
		{ 118, 165, 51 },
		{ 118, 166, 51 },
		{ 118, 167, 51 },
		{ 118, 168, 51 },
		{ 118, 169, 51 },
		{ 118, 170, 51 },
		{ 118, 171, 51 },
		{ 118, 172, 51 },
		{ 118, 173, 51 },
		{ 118, 174, 51 },
		{ 118, 175, 51 },
		{ 118, 176, 51 },
		{ 118, 177, 51 },
		{ 118, 178, 51 },
		{ 118, 179, 51 },
		{ 118, 180, 51 },
		{ 118, 181, 51 },
		{ 118, 182, 51 },
		{ 118, 183, 51 },
		{ 118, 184, 51 },
		{ 118, 185, 51 },
		{ 118, 186, 51 },
		{ 118, 187, 51 },
		{ 119, 154, 51 },
		{ 119, 155, 51 },
		{ 119, 156, 51 },
		{ 119, 157, 51 },
		{ 119, 158, 51 },
		{ 119, 159, 51 },
		{ 119, 160, 51 },
		{ 119, 161, 51 },
		{ 119, 162, 51 },
		{ 119, 163, 51 },
		{ 119, 164, 51 },
		{ 119, 165, 51 },
		{ 119, 166, 51 },
		{ 119, 167, 51 },
		{ 119, 168, 51 },
		{ 119, 169, 51 },
		{ 119, 170, 51 },
		{ 119, 171, 51 },
		{ 119, 172, 51 },
		{ 119, 173, 51 },
		{ 119, 174, 51 },
		{ 119, 175, 51 },
		{ 119, 176, 51 },
		{ 119, 177, 51 },
		{ 119, 178, 51 },
		{ 119, 179, 51 },
		{ 119, 180, 51 },
		{ 119, 181, 51 },
		{ 119, 182, 51 },
		{ 119, 183, 51 },
		{ 119, 184, 51 },
		{ 119, 185, 51 },
		{ 119, 186, 51 },
		{ 119, 187, 51 },
		{ 120, 155, 51 },
		{ 120, 156, 51 },
		{ 120, 157, 51 },
		{ 120, 158, 51 },
		{ 120, 159, 51 },
		{ 120, 160, 51 },
		{ 120, 161, 51 },
		{ 120, 162, 51 },
		{ 120, 163, 51 },
		{ 120, 164, 51 },
		{ 120, 165, 51 },
		{ 120, 166, 51 },
		{ 120, 167, 51 },
		{ 120, 168, 51 },
		{ 120, 169, 51 },
		{ 120, 170, 51 },
		{ 120, 171, 51 },
		{ 120, 172, 51 },
		{ 120, 173, 51 },
		{ 120, 174, 51 },
		{ 120, 175, 51 },
		{ 120, 176, 51 },
		{ 120, 177, 51 },
		{ 120, 178, 51 },
		{ 120, 179, 51 },
		{ 120, 180, 51 },
		{ 120, 181, 51 },
		{ 120, 182, 51 },
		{ 120, 183, 51 },
		{ 120, 184, 51 },
		{ 120, 185, 51 },
		{ 120, 186, 51 },
		{ 120, 187, 51 },
		{ 120, 188, 51 },
		{ 121, 156, 51 },
		{ 121, 157, 51 },
		{ 121, 158, 51 },
		{ 121, 159, 51 },
		{ 121, 160, 51 },
		{ 121, 161, 51 },
		{ 121, 162, 51 },
		{ 121, 163, 51 },
		{ 121, 164, 51 },
		{ 121, 165, 51 },
		{ 121, 166, 51 },
		{ 121, 167, 51 },
		{ 121, 168, 51 },
		{ 121, 169, 51 },
		{ 121, 170, 51 },
		{ 121, 171, 51 },
		{ 121, 172, 51 },
		{ 121, 173, 51 },
		{ 121, 174, 51 },
		{ 121, 175, 51 },
		{ 121, 176, 51 },
		{ 121, 177, 51 },
		{ 121, 178, 51 },
		{ 121, 179, 51 },
		{ 121, 180, 51 },
		{ 121, 181, 51 },
		{ 121, 182, 51 },
		{ 121, 183, 51 },
		{ 121, 184, 51 },
		{ 121, 185, 51 },
		{ 121, 186, 51 },
		{ 121, 187, 51 },
		{ 121, 188, 51 },
		{ 122, 156, 51 },
		{ 122, 157, 51 },
		{ 122, 158, 51 },
		{ 122, 159, 51 },
		{ 122, 160, 51 },
		{ 122, 161, 51 },
		{ 122, 162, 51 },
		{ 122, 163, 51 },
		{ 122, 164, 51 },
		{ 122, 165, 51 },
		{ 122, 166, 51 },
		{ 122, 167, 51 },
		{ 122, 168, 51 },
		{ 122, 169, 51 },
		{ 122, 170, 51 },
		{ 122, 171, 51 },
		{ 122, 172, 51 },
		{ 122, 173, 51 },
		{ 122, 174, 51 },
		{ 122, 175, 51 },
		{ 122, 176, 51 },
		{ 122, 177, 51 },
		{ 122, 178, 51 },
		{ 122, 179, 51 },
		{ 122, 180, 51 },
		{ 122, 181, 51 },
		{ 122, 182, 51 },
		{ 122, 183, 51 },
		{ 122, 184, 51 },
		{ 122, 185, 51 },
		{ 122, 186, 51 },
		{ 122, 187, 51 },
		{ 122, 188, 51 },
		{ 122, 189, 51 },
		{ 123, 157, 51 },
		{ 123, 158, 51 },
		{ 123, 159, 51 },
		{ 123, 160, 51 },
		{ 123, 161, 51 },
		{ 123, 162, 51 },
		{ 123, 163, 51 },
		{ 123, 164, 51 },
		{ 123, 165, 51 },
		{ 123, 166, 51 },
		{ 123, 167, 51 },
		{ 123, 168, 51 },
		{ 123, 169, 51 },
		{ 123, 170, 51 },
		{ 123, 171, 51 },
		{ 123, 172, 51 },
		{ 123, 173, 51 },
		{ 123, 174, 51 },
		{ 123, 175, 51 },
		{ 123, 176, 51 },
		{ 123, 177, 51 },
		{ 123, 178, 51 },
		{ 123, 179, 51 },
		{ 123, 180, 51 },
		{ 123, 181, 51 },
		{ 123, 182, 51 },
		{ 123, 183, 51 },
		{ 123, 184, 51 },
		{ 123, 185, 51 },
		{ 123, 186, 51 },
		{ 123, 187, 51 },
		{ 123, 188, 51 },
		{ 123, 189, 51 },
		{ 124, 158, 51 },
		{ 124, 159, 51 },
		{ 124, 160, 51 },
		{ 124, 161, 51 },
		{ 124, 162, 51 },
		{ 124, 163, 51 },
		{ 124, 164, 51 },
		{ 124, 165, 51 },
		{ 124, 166, 51 },
		{ 124, 167, 51 },
		{ 124, 168, 51 },
		{ 124, 169, 51 },
		{ 124, 170, 51 },
		{ 124, 171, 51 },
		{ 124, 172, 51 },
		{ 124, 173, 51 },
		{ 124, 174, 51 },
		{ 124, 175, 51 },
		{ 124, 176, 51 },
		{ 124, 177, 51 },
		{ 124, 178, 51 },
		{ 124, 179, 51 },
		{ 124, 180, 51 },
		{ 124, 181, 51 },
		{ 124, 182, 51 },
		{ 124, 183, 51 },
		{ 124, 184, 51 },
		{ 124, 185, 51 },
		{ 124, 186, 51 },
		{ 124, 187, 51 },
		{ 124, 188, 51 },
		{ 124, 189, 51 },
		{ 124, 190, 51 },
		{ 125, 159, 51 },
		{ 125, 160, 51 },
		{ 125, 161, 51 },
		{ 125, 162, 51 },
		{ 125, 163, 51 },
		{ 125, 164, 51 },
		{ 125, 165, 51 },
		{ 125, 166, 51 },
		{ 125, 167, 51 },
		{ 125, 168, 51 },
		{ 125, 169, 51 },
		{ 125, 170, 51 },
		{ 125, 171, 51 },
		{ 125, 172, 51 },
		{ 125, 173, 51 },
		{ 125, 174, 51 },
		{ 125, 175, 51 },
		{ 125, 176, 51 },
		{ 125, 177, 51 },
		{ 125, 178, 51 },
		{ 125, 179, 51 },
		{ 125, 180, 51 },
		{ 125, 181, 51 },
		{ 125, 182, 51 },
		{ 125, 183, 51 },
		{ 125, 184, 51 },
		{ 125, 185, 51 },
		{ 125, 186, 51 },
		{ 125, 187, 51 },
		{ 125, 188, 51 },
		{ 125, 189, 51 },
		{ 125, 190, 51 },
		{ 126, 159, 51 },
		{ 126, 160, 51 },
		{ 126, 161, 51 },
		{ 126, 162, 51 },
		{ 126, 163, 51 },
		{ 126, 164, 51 },
		{ 126, 165, 51 },
		{ 126, 166, 51 },
		{ 126, 167, 51 },
		{ 126, 168, 51 },
		{ 126, 169, 51 },
		{ 126, 170, 51 },
		{ 126, 171, 51 },
		{ 126, 172, 51 },
		{ 126, 173, 51 },
		{ 126, 174, 51 },
		{ 126, 175, 51 },
		{ 126, 176, 51 },
		{ 126, 177, 51 },
		{ 126, 178, 51 },
		{ 126, 179, 51 },
		{ 126, 180, 51 },
		{ 126, 181, 51 },
		{ 126, 182, 51 },
		{ 126, 183, 51 },
		{ 126, 184, 51 },
		{ 126, 185, 51 },
		{ 126, 186, 51 },
		{ 126, 187, 51 },
		{ 126, 188, 51 },
		{ 126, 189, 51 },
		{ 126, 190, 51 },
		{ 126, 191, 51 },
		{ 127, 160, 51 },
		{ 127, 161, 51 },
		{ 127, 162, 51 },
		{ 127, 163, 51 },
		{ 127, 164, 51 },
		{ 127, 165, 51 },
		{ 127, 166, 51 },
		{ 127, 167, 51 },
		{ 127, 168, 51 },
		{ 127, 169, 51 },
		{ 127, 170, 51 },
		{ 127, 171, 51 },
		{ 127, 172, 51 },
		{ 127, 173, 51 },
		{ 127, 174, 51 },
		{ 127, 175, 51 },
		{ 127, 176, 51 },
		{ 127, 177, 51 },
		{ 127, 178, 51 },
		{ 127, 179, 51 },
		{ 127, 180, 51 },
		{ 127, 181, 51 },
		{ 127, 182, 51 },
		{ 127, 183, 51 },
		{ 127, 184, 51 },
		{ 127, 185, 51 },
		{ 127, 186, 51 },
		{ 127, 187, 51 },
		{ 127, 188, 51 },
		{ 127, 189, 51 },
		{ 127, 190, 51 },
		{ 127, 191, 51 },
		{ 128, 161, 51 },
		{ 128, 162, 51 },
		{ 128, 163, 51 },
		{ 128, 164, 51 },
		{ 128, 165, 51 },
		{ 128, 166, 51 },
		{ 128, 167, 51 },
		{ 128, 168, 51 },
		{ 128, 169, 51 },
		{ 128, 170, 51 },
		{ 128, 171, 51 },
		{ 128, 172, 51 },
		{ 128, 173, 51 },
		{ 128, 174, 51 },
		{ 128, 175, 51 },
		{ 128, 176, 51 },
		{ 128, 177, 51 },
		{ 128, 178, 51 },
		{ 128, 179, 51 },
		{ 128, 180, 51 },
		{ 128, 181, 51 },
		{ 128, 182, 51 },
		{ 128, 183, 51 },
		{ 128, 184, 51 },
		{ 128, 185, 51 },
		{ 128, 186, 51 },
		{ 128, 187, 51 },
		{ 128, 188, 51 },
		{ 128, 189, 51 },
		{ 128, 190, 51 },
		{ 128, 191, 51 },
		{ 128, 192, 51 },
		{ 129, 162, 51 },
		{ 129, 163, 51 },
		{ 129, 164, 51 },
		{ 129, 165, 51 },
		{ 129, 166, 51 },
		{ 129, 167, 51 },
		{ 129, 168, 51 },
		{ 129, 169, 51 },
		{ 129, 170, 51 },
		{ 129, 171, 51 },
		{ 129, 172, 51 },
		{ 129, 173, 51 },
		{ 129, 174, 51 },
		{ 129, 175, 51 },
		{ 129, 176, 51 },
		{ 129, 177, 51 },
		{ 129, 178, 51 },
		{ 129, 179, 51 },
		{ 129, 180, 51 },
		{ 129, 181, 51 },
		{ 129, 182, 51 },
		{ 129, 183, 51 },
		{ 129, 184, 51 },
		{ 129, 185, 51 },
		{ 129, 186, 51 },
		{ 129, 187, 51 },
		{ 129, 188, 51 },
		{ 129, 189, 51 },
		{ 129, 190, 51 },
		{ 129, 191, 51 },
		{ 129, 192, 51 },
		{ 130, 162, 51 },
		{ 130, 163, 51 },
		{ 130, 164, 51 },
		{ 130, 165, 51 },
		{ 130, 166, 51 },
		{ 130, 167, 51 },
		{ 130, 168, 51 },
		{ 130, 169, 51 },
		{ 130, 170, 51 },
		{ 130, 171, 51 },
		{ 130, 172, 51 },
		{ 130, 173, 51 },
		{ 130, 174, 51 },
		{ 130, 175, 51 },
		{ 130, 176, 51 },
		{ 130, 177, 51 },
		{ 130, 178, 51 },
		{ 130, 179, 51 },
		{ 130, 180, 51 },
		{ 130, 181, 51 },
		{ 130, 182, 51 },
		{ 130, 183, 51 },
		{ 130, 184, 51 },
		{ 130, 185, 51 },
		{ 130, 186, 51 },
		{ 130, 187, 51 },
		{ 130, 188, 51 },
		{ 130, 189, 51 },
		{ 130, 190, 51 },
		{ 130, 191, 51 },
		{ 130, 192, 51 },
		{ 130, 193, 51 },
		{ 131, 163, 51 },
		{ 131, 164, 51 },
		{ 131, 165, 51 },
		{ 131, 166, 51 },
		{ 131, 167, 51 },
		{ 131, 168, 51 },
		{ 131, 169, 51 },
		{ 131, 170, 51 },
		{ 131, 171, 51 },
		{ 131, 172, 51 },
		{ 131, 173, 51 },
		{ 131, 174, 51 },
		{ 131, 175, 51 },
		{ 131, 176, 51 },
		{ 131, 177, 51 },
		{ 131, 178, 51 },
		{ 131, 179, 51 },
		{ 131, 180, 51 },
		{ 131, 181, 51 },
		{ 131, 182, 51 },
		{ 131, 183, 51 },
		{ 131, 184, 51 },
		{ 131, 185, 51 },
		{ 131, 186, 51 },
		{ 131, 187, 51 },
		{ 131, 188, 51 },
		{ 131, 189, 51 },
		{ 131, 190, 51 },
		{ 131, 191, 51 },
		{ 131, 192, 51 },
		{ 131, 193, 51 },
		{ 132, 164, 51 },
		{ 132, 165, 51 },
		{ 132, 166, 51 },
		{ 132, 167, 51 },
		{ 132, 168, 51 },
		{ 132, 169, 51 },
		{ 132, 170, 51 },
		{ 132, 171, 51 },
		{ 132, 172, 51 },
		{ 132, 173, 51 },
		{ 132, 174, 51 },
		{ 132, 175, 51 },
		{ 132, 176, 51 },
		{ 132, 177, 51 },
		{ 132, 178, 51 },
		{ 132, 179, 51 },
		{ 132, 180, 51 },
		{ 132, 181, 51 },
		{ 132, 182, 51 },
		{ 132, 183, 51 },
		{ 132, 184, 51 },
		{ 132, 185, 51 },
		{ 132, 186, 51 },
		{ 132, 187, 51 },
		{ 132, 188, 51 },
		{ 132, 189, 51 },
		{ 132, 190, 51 },
		{ 132, 191, 51 },
		{ 132, 192, 51 },
		{ 132, 193, 51 },
		{ 132, 194, 51 },
		{ 133, 165, 51 },
		{ 133, 166, 51 },
		{ 133, 167, 51 },
		{ 133, 168, 51 },
		{ 133, 169, 51 },
		{ 133, 170, 51 },
		{ 133, 171, 51 },
		{ 133, 172, 51 },
		{ 133, 173, 51 },
		{ 133, 174, 51 },
		{ 133, 175, 51 },
		{ 133, 176, 51 },
		{ 133, 177, 51 },
		{ 133, 178, 51 },
		{ 133, 179, 51 },
		{ 133, 180, 51 },
		{ 133, 181, 51 },
		{ 133, 182, 51 },
		{ 133, 183, 51 },
		{ 133, 184, 51 },
		{ 133, 185, 51 },
		{ 133, 186, 51 },
		{ 133, 187, 51 },
		{ 133, 188, 51 },
		{ 133, 189, 51 },
		{ 133, 190, 51 },
		{ 133, 191, 51 },
		{ 133, 192, 51 },
		{ 133, 193, 51 },
		{ 133, 194, 51 },
		{ 134, 165, 51 },
		{ 134, 166, 51 },
		{ 134, 167, 51 },
		{ 134, 168, 51 },
		{ 134, 169, 51 },
		{ 134, 170, 51 },
		{ 134, 171, 51 },
		{ 134, 172, 51 },
		{ 134, 173, 51 },
		{ 134, 174, 51 },
		{ 134, 175, 51 },
		{ 134, 176, 51 },
		{ 134, 177, 51 },
		{ 134, 178, 51 },
		{ 134, 179, 51 },
		{ 134, 180, 51 },
		{ 134, 181, 51 },
		{ 134, 182, 51 },
		{ 134, 183, 51 },
		{ 134, 184, 51 },
		{ 134, 185, 51 },
		{ 134, 186, 51 },
		{ 134, 187, 51 },
		{ 134, 188, 51 },
		{ 134, 189, 51 },
		{ 134, 190, 51 },
		{ 134, 191, 51 },
		{ 134, 192, 51 },
		{ 134, 193, 51 },
		{ 134, 194, 51 },
		{ 134, 195, 51 },
		{ 135, 166, 51 },
		{ 135, 167, 51 },
		{ 135, 168, 51 },
		{ 135, 169, 51 },
		{ 135, 170, 51 },
		{ 135, 171, 51 },
		{ 135, 172, 51 },
		{ 135, 173, 51 },
		{ 135, 174, 51 },
		{ 135, 175, 51 },
		{ 135, 176, 51 },
		{ 135, 177, 51 },
		{ 135, 178, 51 },
		{ 135, 179, 51 },
		{ 135, 180, 51 },
		{ 135, 181, 51 },
		{ 135, 182, 51 },
		{ 135, 183, 51 },
		{ 135, 184, 51 },
		{ 135, 185, 51 },
		{ 135, 186, 51 },
		{ 135, 187, 51 },
		{ 135, 188, 51 },
		{ 135, 189, 51 },
		{ 135, 190, 51 },
		{ 135, 191, 51 },
		{ 135, 192, 51 },
		{ 135, 193, 51 },
		{ 135, 194, 51 },
		{ 135, 195, 51 },
		{ 136, 167, 51 },
		{ 136, 168, 51 },
		{ 136, 169, 51 },
		{ 136, 170, 51 },
		{ 136, 171, 51 },
		{ 136, 172, 51 },
		{ 136, 173, 51 },
		{ 136, 174, 51 },
		{ 136, 175, 51 },
		{ 136, 176, 51 },
		{ 136, 177, 51 },
		{ 136, 178, 51 },
		{ 136, 179, 51 },
		{ 136, 180, 51 },
		{ 136, 181, 51 },
		{ 136, 182, 51 },
		{ 136, 183, 51 },
		{ 136, 184, 51 },
		{ 136, 185, 51 },
		{ 136, 186, 51 },
		{ 136, 187, 51 },
		{ 136, 188, 51 },
		{ 136, 189, 51 },
		{ 136, 190, 51 },
		{ 136, 191, 51 },
		{ 136, 192, 51 },
		{ 136, 193, 51 },
		{ 136, 194, 51 },
		{ 136, 195, 51 },
		{ 136, 196, 51 },
		{ 137, 168, 51 },
		{ 137, 169, 51 },
		{ 137, 170, 51 },
		{ 137, 171, 51 },
		{ 137, 172, 51 },
		{ 137, 173, 51 },
		{ 137, 174, 51 },
		{ 137, 175, 51 },
		{ 137, 176, 51 },
		{ 137, 177, 51 },
		{ 137, 178, 51 },
		{ 137, 179, 51 },
		{ 137, 180, 51 },
		{ 137, 181, 51 },
		{ 137, 182, 51 },
		{ 137, 183, 51 },
		{ 137, 184, 51 },
		{ 137, 185, 51 },
		{ 137, 186, 51 },
		{ 137, 187, 51 },
		{ 137, 188, 51 },
		{ 137, 189, 51 },
		{ 137, 190, 51 },
		{ 137, 191, 51 },
		{ 137, 192, 51 },
		{ 137, 193, 51 },
		{ 137, 194, 51 },
		{ 137, 195, 51 },
		{ 137, 196, 51 },
		{ 138, 168, 51 },
		{ 138, 169, 51 },
		{ 138, 170, 51 },
		{ 138, 171, 51 },
		{ 138, 172, 51 },
		{ 138, 173, 51 },
		{ 138, 174, 51 },
		{ 138, 175, 51 },
		{ 138, 176, 51 },
		{ 138, 177, 51 },
		{ 138, 178, 51 },
		{ 138, 179, 51 },
		{ 138, 180, 51 },
		{ 138, 181, 51 },
		{ 138, 182, 51 },
		{ 138, 183, 51 },
		{ 138, 184, 51 },
		{ 138, 185, 51 },
		{ 138, 186, 51 },
		{ 138, 187, 51 },
		{ 138, 188, 51 },
		{ 138, 189, 51 },
		{ 138, 190, 51 },
		{ 138, 191, 51 },
		{ 138, 192, 51 },
		{ 138, 193, 51 },
		{ 138, 194, 51 },
		{ 138, 195, 51 },
		{ 138, 196, 51 },
		{ 138, 197, 51 },
		{ 139, 169, 51 },
		{ 139, 170, 51 },
		{ 139, 171, 51 },
		{ 139, 172, 51 },
		{ 139, 173, 51 },
		{ 139, 174, 51 },
		{ 139, 175, 51 },
		{ 139, 176, 51 },
		{ 139, 177, 51 },
		{ 139, 178, 51 },
		{ 139, 179, 51 },
		{ 139, 180, 51 },
		{ 139, 181, 51 },
		{ 139, 182, 51 },
		{ 139, 183, 51 },
		{ 139, 184, 51 },
		{ 139, 185, 51 },
		{ 139, 186, 51 },
		{ 139, 187, 51 },
		{ 139, 188, 51 },
		{ 139, 189, 51 },
		{ 139, 190, 51 },
		{ 139, 191, 51 },
		{ 139, 192, 51 },
		{ 139, 193, 51 },
		{ 139, 194, 51 },
		{ 139, 195, 51 },
		{ 139, 196, 51 },
		{ 139, 197, 51 },
		{ 140, 170, 51 },
		{ 140, 171, 51 },
		{ 140, 172, 51 },
		{ 140, 173, 51 },
		{ 140, 174, 51 },
		{ 140, 175, 51 },
		{ 140, 176, 51 },
		{ 140, 177, 51 },
		{ 140, 178, 51 },
		{ 140, 179, 51 },
		{ 140, 180, 51 },
		{ 140, 181, 51 },
		{ 140, 182, 51 },
		{ 140, 183, 51 },
		{ 140, 184, 51 },
		{ 140, 185, 51 },
		{ 140, 186, 51 },
		{ 140, 187, 51 },
		{ 140, 188, 51 },
		{ 140, 189, 51 },
		{ 140, 190, 51 },
		{ 140, 191, 51 },
		{ 140, 192, 51 },
		{ 140, 193, 51 },
		{ 140, 194, 51 },
		{ 140, 195, 51 },
		{ 140, 196, 51 },
		{ 140, 197, 51 },
		{ 140, 198, 51 },
		{ 141, 171, 51 },
		{ 141, 172, 51 },
		{ 141, 173, 51 },
		{ 141, 174, 51 },
		{ 141, 175, 51 },
		{ 141, 176, 51 },
		{ 141, 177, 51 },
		{ 141, 178, 51 },
		{ 141, 179, 51 },
		{ 141, 180, 51 },
		{ 141, 181, 51 },
		{ 141, 182, 51 },
		{ 141, 183, 51 },
		{ 141, 184, 51 },
		{ 141, 185, 51 },
		{ 141, 186, 51 },
		{ 141, 187, 51 },
		{ 141, 188, 51 },
		{ 141, 189, 51 },
		{ 141, 190, 51 },
		{ 141, 191, 51 },
		{ 141, 192, 51 },
		{ 141, 193, 51 },
		{ 141, 194, 51 },
		{ 141, 195, 51 },
		{ 141, 196, 51 },
		{ 141, 197, 51 },
		{ 141, 198, 51 },
		{ 142, 171, 51 },
		{ 142, 172, 51 },
		{ 142, 173, 51 },
		{ 142, 174, 51 },
		{ 142, 175, 51 },
		{ 142, 176, 51 },
		{ 142, 177, 51 },
		{ 142, 178, 51 },
		{ 142, 179, 51 },
		{ 142, 180, 51 },
		{ 142, 181, 51 },
		{ 142, 182, 51 },
		{ 142, 183, 51 },
		{ 142, 184, 51 },
		{ 142, 185, 51 },
		{ 142, 186, 51 },
		{ 142, 187, 51 },
		{ 142, 188, 51 },
		{ 142, 189, 51 },
		{ 142, 190, 51 },
		{ 142, 191, 51 },
		{ 142, 192, 51 },
		{ 142, 193, 51 },
		{ 142, 194, 51 },
		{ 142, 195, 51 },
		{ 142, 196, 51 },
		{ 142, 197, 51 },
		{ 142, 198, 51 },
		{ 142, 199, 51 },
		{ 143, 172, 51 },
		{ 143, 173, 51 },
		{ 143, 174, 51 },
		{ 143, 175, 51 },
		{ 143, 176, 51 },
		{ 143, 177, 51 },
		{ 143, 178, 51 },
		{ 143, 179, 51 },
		{ 143, 180, 51 },
		{ 143, 181, 51 },
		{ 143, 182, 51 },
		{ 143, 183, 51 },
		{ 143, 184, 51 },
		{ 143, 185, 51 },
		{ 143, 186, 51 },
		{ 143, 187, 51 },
		{ 143, 188, 51 },
		{ 143, 189, 51 },
		{ 143, 190, 51 },
		{ 143, 191, 51 },
		{ 143, 192, 51 },
		{ 143, 193, 51 },
		{ 143, 194, 51 },
		{ 143, 195, 51 },
		{ 143, 196, 51 },
		{ 143, 197, 51 },
		{ 143, 198, 51 },
		{ 143, 199, 51 },
		{ 144, 173, 51 },
		{ 144, 174, 51 },
		{ 144, 175, 51 },
		{ 144, 176, 51 },
		{ 144, 177, 51 },
		{ 144, 178, 51 },
		{ 144, 179, 51 },
		{ 144, 180, 51 },
		{ 144, 181, 51 },
		{ 144, 182, 51 },
		{ 144, 183, 51 },
		{ 144, 184, 51 },
		{ 144, 185, 51 },
		{ 144, 186, 51 },
		{ 144, 187, 51 },
		{ 144, 188, 51 },
		{ 144, 189, 51 },
		{ 144, 190, 51 },
		{ 144, 191, 51 },
		{ 144, 192, 51 },
		{ 144, 193, 51 },
		{ 144, 194, 51 },
		{ 144, 195, 51 },
		{ 144, 196, 51 },
		{ 144, 197, 51 },
		{ 144, 198, 51 },
		{ 144, 199, 51 },
		{ 144, 200, 51 },
		{ 145, 174, 51 },
		{ 145, 175, 51 },
		{ 145, 176, 51 },
		{ 145, 177, 51 },
		{ 145, 178, 51 },
		{ 145, 179, 51 },
		{ 145, 180, 51 },
		{ 145, 181, 51 },
		{ 145, 182, 51 },
		{ 145, 183, 51 },
		{ 145, 184, 51 },
		{ 145, 185, 51 },
		{ 145, 186, 51 },
		{ 145, 187, 51 },
		{ 145, 188, 51 },
		{ 145, 189, 51 },
		{ 145, 190, 51 },
		{ 145, 191, 51 },
		{ 145, 192, 51 },
		{ 145, 193, 51 },
		{ 145, 194, 51 },
		{ 145, 195, 51 },
		{ 145, 196, 51 },
		{ 145, 197, 51 },
		{ 145, 198, 51 },
		{ 145, 199, 51 },
		{ 145, 200, 51 },
		{ 146, 174, 51 },
		{ 146, 175, 51 },
		{ 146, 176, 51 },
		{ 146, 177, 51 },
		{ 146, 178, 51 },
		{ 146, 179, 51 },
		{ 146, 180, 51 },
		{ 146, 181, 51 },
		{ 146, 182, 51 },
		{ 146, 183, 51 },
		{ 146, 184, 51 },
		{ 146, 185, 51 },
		{ 146, 186, 51 },
		{ 146, 187, 51 },
		{ 146, 188, 51 },
		{ 146, 189, 51 },
		{ 146, 190, 51 },
		{ 146, 191, 51 },
		{ 146, 192, 51 },
		{ 146, 193, 51 },
		{ 146, 194, 51 },
		{ 146, 195, 51 },
		{ 146, 196, 51 },
		{ 146, 197, 51 },
		{ 146, 198, 51 },
		{ 146, 199, 51 },
		{ 146, 200, 51 },
		{ 146, 201, 51 },
		{ 147, 175, 51 },
		{ 147, 176, 51 },
		{ 147, 177, 51 },
		{ 147, 178, 51 },
		{ 147, 179, 51 },
		{ 147, 180, 51 },
		{ 147, 181, 51 },
		{ 147, 182, 51 },
		{ 147, 183, 51 },
		{ 147, 184, 51 },
		{ 147, 185, 51 },
		{ 147, 186, 51 },
		{ 147, 187, 51 },
		{ 147, 188, 51 },
		{ 147, 189, 51 },
		{ 147, 190, 51 },
		{ 147, 191, 51 },
		{ 147, 192, 51 },
		{ 147, 193, 51 },
		{ 147, 194, 51 },
		{ 147, 195, 51 },
		{ 147, 196, 51 },
		{ 147, 197, 51 },
		{ 147, 198, 51 },
		{ 147, 199, 51 },
		{ 147, 200, 51 },
		{ 147, 201, 51 },
		{ 148, 176, 51 },
		{ 148, 177, 51 },
		{ 148, 178, 51 },
		{ 148, 179, 51 },
		{ 148, 180, 51 },
		{ 148, 181, 51 },
		{ 148, 182, 51 },
		{ 148, 183, 51 },
		{ 148, 184, 51 },
		{ 148, 185, 51 },
		{ 148, 186, 51 },
		{ 148, 187, 51 },
		{ 148, 188, 51 },
		{ 148, 189, 51 },
		{ 148, 190, 51 },
		{ 148, 191, 51 },
		{ 148, 192, 51 },
		{ 148, 193, 51 },
		{ 148, 194, 51 },
		{ 148, 195, 51 },
		{ 148, 196, 51 },
		{ 148, 197, 51 },
		{ 148, 198, 51 },
		{ 148, 199, 51 },
		{ 148, 200, 51 },
		{ 148, 201, 51 },
		{ 148, 202, 51 },
		{ 149, 177, 51 },
		{ 149, 178, 51 },
		{ 149, 179, 51 },
		{ 149, 180, 51 },
		{ 149, 181, 51 },
		{ 149, 182, 51 },
		{ 149, 183, 51 },
		{ 149, 184, 51 },
		{ 149, 185, 51 },
		{ 149, 186, 51 },
		{ 149, 187, 51 },
		{ 149, 188, 51 },
		{ 149, 189, 51 },
		{ 149, 190, 51 },
		{ 149, 191, 51 },
		{ 149, 192, 51 },
		{ 149, 193, 51 },
		{ 149, 194, 51 },
		{ 149, 195, 51 },
		{ 149, 196, 51 },
		{ 149, 197, 51 },
		{ 149, 198, 51 },
		{ 149, 199, 51 },
		{ 149, 200, 51 },
		{ 149, 201, 51 },
		{ 149, 202, 51 },
		{ 150, 177, 51 },
		{ 150, 178, 51 },
		{ 150, 179, 51 },
		{ 150, 180, 51 },
		{ 150, 181, 51 },
		{ 150, 182, 51 },
		{ 150, 183, 51 },
		{ 150, 184, 51 },
		{ 150, 185, 51 },
		{ 150, 186, 51 },
		{ 150, 187, 51 },
		{ 150, 188, 51 },
		{ 150, 189, 51 },
		{ 150, 190, 51 },
		{ 150, 191, 51 },
		{ 150, 192, 51 },
		{ 150, 193, 51 },
		{ 150, 194, 51 },
		{ 150, 195, 51 },
		{ 150, 196, 51 },
		{ 150, 197, 51 },
		{ 150, 198, 51 },
		{ 150, 199, 51 },
		{ 150, 200, 51 },
		{ 150, 201, 51 },
		{ 150, 202, 51 },
		{ 150, 203, 51 },
		{ 151, 178, 51 },
		{ 151, 179, 51 },
		{ 151, 180, 51 },
		{ 151, 181, 51 },
		{ 151, 182, 51 },
		{ 151, 183, 51 },
		{ 151, 184, 51 },
		{ 151, 185, 51 },
		{ 151, 186, 51 },
		{ 151, 187, 51 },
		{ 151, 188, 51 },
		{ 151, 189, 51 },
		{ 151, 190, 51 },
		{ 151, 191, 51 },
		{ 151, 192, 51 },
		{ 151, 193, 51 },
		{ 151, 194, 51 },
		{ 151, 195, 51 },
		{ 151, 196, 51 },
		{ 151, 197, 51 },
		{ 151, 198, 51 },
		{ 151, 199, 51 },
		{ 151, 200, 51 },
		{ 151, 201, 51 },
		{ 151, 202, 51 },
		{ 151, 203, 51 },
		{ 152, 179, 51 },
		{ 152, 180, 51 },
		{ 152, 181, 51 },
		{ 152, 182, 51 },
		{ 152, 183, 51 },
		{ 152, 184, 51 },
		{ 152, 185, 51 },
		{ 152, 186, 51 },
		{ 152, 187, 51 },
		{ 152, 188, 51 },
		{ 152, 189, 51 },
		{ 152, 190, 51 },
		{ 152, 191, 51 },
		{ 152, 192, 51 },
		{ 152, 193, 51 },
		{ 152, 194, 51 },
		{ 152, 195, 51 },
		{ 152, 196, 51 },
		{ 152, 197, 51 },
		{ 152, 198, 51 },
		{ 152, 199, 51 },
		{ 152, 200, 51 },
		{ 152, 201, 51 },
		{ 152, 202, 51 },
		{ 152, 203, 51 },
		{ 152, 204, 51 },
		{ 153, 180, 51 },
		{ 153, 181, 51 },
		{ 153, 182, 51 },
		{ 153, 183, 51 },
		{ 153, 184, 51 },
		{ 153, 185, 51 },
		{ 153, 186, 51 },
		{ 153, 187, 51 },
		{ 153, 188, 51 },
		{ 153, 189, 51 },
		{ 153, 190, 51 },
		{ 153, 191, 51 },
		{ 153, 192, 51 },
		{ 153, 193, 51 },
		{ 153, 194, 51 },
		{ 153, 195, 51 },
		{ 153, 196, 51 },
		{ 153, 197, 51 },
		{ 153, 198, 51 },
		{ 153, 199, 51 },
		{ 153, 200, 51 },
		{ 153, 201, 51 },
		{ 153, 202, 51 },
		{ 153, 203, 51 },
		{ 153, 204, 51 },
		{ 154, 180, 51 },
		{ 154, 181, 51 },
		{ 154, 182, 51 },
		{ 154, 183, 51 },
		{ 154, 184, 51 },
		{ 154, 185, 51 },
		{ 154, 186, 51 },
		{ 154, 187, 51 },
		{ 154, 188, 51 },
		{ 154, 189, 51 },
		{ 154, 190, 51 },
		{ 154, 191, 51 },
		{ 154, 192, 51 },
		{ 154, 193, 51 },
		{ 154, 194, 51 },
		{ 154, 195, 51 },
		{ 154, 196, 51 },
		{ 154, 197, 51 },
		{ 154, 198, 51 },
		{ 154, 199, 51 },
		{ 154, 200, 51 },
		{ 154, 201, 51 },
		{ 154, 202, 51 },
		{ 154, 203, 51 },
		{ 154, 204, 51 },
		{ 154, 205, 51 },
		{ 155, 181, 51 },
		{ 155, 182, 51 },
		{ 155, 183, 51 },
		{ 155, 184, 51 },
		{ 155, 185, 51 },
		{ 155, 186, 51 },
		{ 155, 187, 51 },
		{ 155, 188, 51 },
		{ 155, 189, 51 },
		{ 155, 190, 51 },
		{ 155, 191, 51 },
		{ 155, 192, 51 },
		{ 155, 193, 51 },
		{ 155, 194, 51 },
		{ 155, 195, 51 },
		{ 155, 196, 51 },
		{ 155, 197, 51 },
		{ 155, 198, 51 },
		{ 155, 199, 51 },
		{ 155, 200, 51 },
		{ 155, 201, 51 },
		{ 155, 202, 51 },
		{ 155, 203, 51 },
		{ 155, 204, 51 },
		{ 155, 205, 51 },
		{ 156, 182, 51 },
		{ 156, 183, 51 },
		{ 156, 184, 51 },
		{ 156, 185, 51 },
		{ 156, 186, 51 },
		{ 156, 187, 51 },
		{ 156, 188, 51 },
		{ 156, 189, 51 },
		{ 156, 190, 51 },
		{ 156, 191, 51 },
		{ 156, 192, 51 },
		{ 156, 193, 51 },
		{ 156, 194, 51 },
		{ 156, 195, 51 },
		{ 156, 196, 51 },
		{ 156, 197, 51 },
		{ 156, 198, 51 },
		{ 156, 199, 51 },
		{ 156, 200, 51 },
		{ 156, 201, 51 },
		{ 156, 202, 51 },
		{ 156, 203, 51 },
		{ 156, 204, 51 },
		{ 156, 205, 51 },
		{ 156, 206, 51 },
		{ 157, 183, 51 },
		{ 157, 184, 51 },
		{ 157, 185, 51 },
		{ 157, 186, 51 },
		{ 157, 187, 51 },
		{ 157, 188, 51 },
		{ 157, 189, 51 },
		{ 157, 190, 51 },
		{ 157, 191, 51 },
		{ 157, 192, 51 },
		{ 157, 193, 51 },
		{ 157, 194, 51 },
		{ 157, 195, 51 },
		{ 157, 196, 51 },
		{ 157, 197, 51 },
		{ 157, 198, 51 },
		{ 157, 199, 51 },
		{ 157, 200, 51 },
		{ 157, 201, 51 },
		{ 157, 202, 51 },
		{ 157, 203, 51 },
		{ 157, 204, 51 },
		{ 157, 205, 51 },
		{ 157, 206, 51 },
		{ 158, 183, 51 },
		{ 158, 184, 51 },
		{ 158, 185, 51 },
		{ 158, 186, 51 },
		{ 158, 187, 51 },
		{ 158, 188, 51 },
		{ 158, 189, 51 },
		{ 158, 190, 51 },
		{ 158, 191, 51 },
		{ 158, 192, 51 },
		{ 158, 193, 51 },
		{ 158, 194, 51 },
		{ 158, 195, 51 },
		{ 158, 196, 51 },
		{ 158, 197, 51 },
		{ 158, 198, 51 },
		{ 158, 199, 51 },
		{ 158, 200, 51 },
		{ 158, 201, 51 },
		{ 158, 202, 51 },
		{ 158, 203, 51 },
		{ 158, 204, 51 },
		{ 158, 205, 51 },
		{ 158, 206, 51 },
		{ 158, 207, 51 },
		{ 159, 184, 51 },
		{ 159, 185, 51 },
		{ 159, 186, 51 },
		{ 159, 187, 51 },
		{ 159, 188, 51 },
		{ 159, 189, 51 },
		{ 159, 190, 51 },
		{ 159, 191, 51 },
		{ 159, 192, 51 },
		{ 159, 193, 51 },
		{ 159, 194, 51 },
		{ 159, 195, 51 },
		{ 159, 196, 51 },
		{ 159, 197, 51 },
		{ 159, 198, 51 },
		{ 159, 199, 51 },
		{ 159, 200, 51 },
		{ 159, 201, 51 },
		{ 159, 202, 51 },
		{ 159, 203, 51 },
		{ 159, 204, 51 },
		{ 159, 205, 51 },
		{ 159, 206, 51 },
		{ 159, 207, 51 },
		{ 160, 185, 51 },
		{ 160, 186, 51 },
		{ 160, 187, 51 },
		{ 160, 188, 51 },
		{ 160, 189, 51 },
		{ 160, 190, 51 },
		{ 160, 191, 51 },
		{ 160, 192, 51 },
		{ 160, 193, 51 },
		{ 160, 194, 51 },
		{ 160, 195, 51 },
		{ 160, 196, 51 },
		{ 160, 197, 51 },
		{ 160, 198, 51 },
		{ 160, 199, 51 },
		{ 160, 200, 51 },
		{ 160, 201, 51 },
		{ 160, 202, 51 },
		{ 160, 203, 51 },
		{ 160, 204, 51 },
		{ 160, 205, 51 },
		{ 160, 206, 51 },
		{ 160, 207, 51 },
		{ 160, 208, 51 },
		{ 161, 186, 51 },
		{ 161, 187, 51 },
		{ 161, 188, 51 },
		{ 161, 189, 51 },
		{ 161, 190, 51 },
		{ 161, 191, 51 },
		{ 161, 192, 51 },
		{ 161, 193, 51 },
		{ 161, 194, 51 },
		{ 161, 195, 51 },
		{ 161, 196, 51 },
		{ 161, 197, 51 },
		{ 161, 198, 51 },
		{ 161, 199, 51 },
		{ 161, 200, 51 },
		{ 161, 201, 51 },
		{ 161, 202, 51 },
		{ 161, 203, 51 },
		{ 161, 204, 51 },
		{ 161, 205, 51 },
		{ 161, 206, 51 },
		{ 161, 207, 51 },
		{ 161, 208, 51 },
		{ 162, 186, 51 },
		{ 162, 187, 51 },
		{ 162, 188, 51 },
		{ 162, 189, 51 },
		{ 162, 190, 51 },
		{ 162, 191, 51 },
		{ 162, 192, 51 },
		{ 162, 193, 51 },
		{ 162, 194, 51 },
		{ 162, 195, 51 },
		{ 162, 196, 51 },
		{ 162, 197, 51 },
		{ 162, 198, 51 },
		{ 162, 199, 51 },
		{ 162, 200, 51 },
		{ 162, 201, 51 },
		{ 162, 202, 51 },
		{ 162, 203, 51 },
		{ 162, 204, 51 },
		{ 162, 205, 51 },
		{ 162, 206, 51 },
		{ 162, 207, 51 },
		{ 162, 208, 51 },
		{ 162, 209, 51 },
		{ 163, 187, 51 },
		{ 163, 188, 51 },
		{ 163, 189, 51 },
		{ 163, 190, 51 },
		{ 163, 191, 51 },
		{ 163, 192, 51 },
		{ 163, 193, 51 },
		{ 163, 194, 51 },
		{ 163, 195, 51 },
		{ 163, 196, 51 },
		{ 163, 197, 51 },
		{ 163, 198, 51 },
		{ 163, 199, 51 },
		{ 163, 200, 51 },
		{ 163, 201, 51 },
		{ 163, 202, 51 },
		{ 163, 203, 51 },
		{ 163, 204, 51 },
		{ 163, 205, 51 },
		{ 163, 206, 51 },
		{ 163, 207, 51 },
		{ 163, 208, 51 },
		{ 163, 209, 51 },
		{ 164, 188, 51 },
		{ 164, 189, 51 },
		{ 164, 190, 51 },
		{ 164, 191, 51 },
		{ 164, 192, 51 },
		{ 164, 193, 51 },
		{ 164, 194, 51 },
		{ 164, 195, 51 },
		{ 164, 196, 51 },
		{ 164, 197, 51 },
		{ 164, 198, 51 },
		{ 164, 199, 51 },
		{ 164, 200, 51 },
		{ 164, 201, 51 },
		{ 164, 202, 51 },
		{ 164, 203, 51 },
		{ 164, 204, 51 },
		{ 164, 205, 51 },
		{ 164, 206, 51 },
		{ 164, 207, 51 },
		{ 164, 208, 51 },
		{ 164, 209, 51 },
		{ 164, 210, 51 },
		{ 165, 189, 51 },
		{ 165, 190, 51 },
		{ 165, 191, 51 },
		{ 165, 192, 51 },
		{ 165, 193, 51 },
		{ 165, 194, 51 },
		{ 165, 195, 51 },
		{ 165, 196, 51 },
		{ 165, 197, 51 },
		{ 165, 198, 51 },
		{ 165, 199, 51 },
		{ 165, 200, 51 },
		{ 165, 201, 51 },
		{ 165, 202, 51 },
		{ 165, 203, 51 },
		{ 165, 204, 51 },
		{ 165, 205, 51 },
		{ 165, 206, 51 },
		{ 165, 207, 51 },
		{ 165, 208, 51 },
		{ 165, 209, 51 },
		{ 165, 210, 51 },
		{ 166, 189, 51 },
		{ 166, 190, 51 },
		{ 166, 191, 51 },
		{ 166, 192, 51 },
		{ 166, 193, 51 },
		{ 166, 194, 51 },
		{ 166, 195, 51 },
		{ 166, 196, 51 },
		{ 166, 197, 51 },
		{ 166, 198, 51 },
		{ 166, 199, 51 },
		{ 166, 200, 51 },
		{ 166, 201, 51 },
		{ 166, 202, 51 },
		{ 166, 203, 51 },
		{ 166, 204, 51 },
		{ 166, 205, 51 },
		{ 166, 206, 51 },
		{ 166, 207, 51 },
		{ 166, 208, 51 },
		{ 166, 209, 51 },
		{ 166, 210, 51 },
		{ 166, 211, 51 },
		{ 167, 190, 51 },
		{ 167, 191, 51 },
		{ 167, 192, 51 },
		{ 167, 193, 51 },
		{ 167, 194, 51 },
		{ 167, 195, 51 },
		{ 167, 196, 51 },
		{ 167, 197, 51 },
		{ 167, 198, 51 },
		{ 167, 199, 51 },
		{ 167, 200, 51 },
		{ 167, 201, 51 },
		{ 167, 202, 51 },
		{ 167, 203, 51 },
		{ 167, 204, 51 },
		{ 167, 205, 51 },
		{ 167, 206, 51 },
		{ 167, 207, 51 },
		{ 167, 208, 51 },
		{ 167, 209, 51 },
		{ 167, 210, 51 },
		{ 167, 211, 51 },
		{ 168, 191, 51 },
		{ 168, 192, 51 },
		{ 168, 193, 51 },
		{ 168, 194, 51 },
		{ 168, 195, 51 },
		{ 168, 196, 51 },
		{ 168, 197, 51 },
		{ 168, 198, 51 },
		{ 168, 199, 51 },
		{ 168, 200, 51 },
		{ 168, 201, 51 },
		{ 168, 202, 51 },
		{ 168, 203, 51 },
		{ 168, 204, 51 },
		{ 168, 205, 51 },
		{ 168, 206, 51 },
		{ 168, 207, 51 },
		{ 168, 208, 51 },
		{ 168, 209, 51 },
		{ 168, 210, 51 },
		{ 168, 211, 51 },
		{ 168, 212, 51 },
		{ 169, 192, 51 },
		{ 169, 193, 51 },
		{ 169, 194, 51 },
		{ 169, 195, 51 },
		{ 169, 196, 51 },
		{ 169, 197, 51 },
		{ 169, 198, 51 },
		{ 169, 199, 51 },
		{ 169, 200, 51 },
		{ 169, 201, 51 },
		{ 169, 202, 51 },
		{ 169, 203, 51 },
		{ 169, 204, 51 },
		{ 169, 205, 51 },
		{ 169, 206, 51 },
		{ 169, 207, 51 },
		{ 169, 208, 51 },
		{ 169, 209, 51 },
		{ 169, 210, 51 },
		{ 169, 211, 51 },
		{ 169, 212, 51 },
		{ 170, 192, 51 },
		{ 170, 193, 51 },
		{ 170, 194, 51 },
		{ 170, 195, 51 },
		{ 170, 196, 51 },
		{ 170, 197, 51 },
		{ 170, 198, 51 },
		{ 170, 199, 51 },
		{ 170, 200, 51 },
		{ 170, 201, 51 },
		{ 170, 202, 51 },
		{ 170, 203, 51 },
		{ 170, 204, 51 },
		{ 170, 205, 51 },
		{ 170, 206, 51 },
		{ 170, 207, 51 },
		{ 170, 208, 51 },
		{ 170, 209, 51 },
		{ 170, 210, 51 },
		{ 170, 211, 51 },
		{ 170, 212, 51 },
		{ 170, 213, 51 },
		{ 171, 193, 51 },
		{ 171, 194, 51 },
		{ 171, 195, 51 },
		{ 171, 196, 51 },
		{ 171, 197, 51 },
		{ 171, 198, 51 },
		{ 171, 199, 51 },
		{ 171, 200, 51 },
		{ 171, 201, 51 },
		{ 171, 202, 51 },
		{ 171, 203, 51 },
		{ 171, 204, 51 },
		{ 171, 205, 51 },
		{ 171, 206, 51 },
		{ 171, 207, 51 },
		{ 171, 208, 51 },
		{ 171, 209, 51 },
		{ 171, 210, 51 },
		{ 171, 211, 51 },
		{ 171, 212, 51 },
		{ 171, 213, 51 },
		{ 172, 194, 51 },
		{ 172, 195, 51 },
		{ 172, 196, 51 },
		{ 172, 197, 51 },
		{ 172, 198, 51 },
		{ 172, 199, 51 },
		{ 172, 200, 51 },
		{ 172, 201, 51 },
		{ 172, 202, 51 },
		{ 172, 203, 51 },
		{ 172, 204, 51 },
		{ 172, 205, 51 },
		{ 172, 206, 51 },
		{ 172, 207, 51 },
		{ 172, 208, 51 },
		{ 172, 209, 51 },
		{ 172, 210, 51 },
		{ 172, 211, 51 },
		{ 172, 212, 51 },
		{ 172, 213, 51 },
		{ 172, 214, 51 },
		{ 173, 195, 51 },
		{ 173, 196, 51 },
		{ 173, 197, 51 },
		{ 173, 198, 51 },
		{ 173, 199, 51 },
		{ 173, 200, 51 },
		{ 173, 201, 51 },
		{ 173, 202, 51 },
		{ 173, 203, 51 },
		{ 173, 204, 51 },
		{ 173, 205, 51 },
		{ 173, 206, 51 },
		{ 173, 207, 51 },
		{ 173, 208, 51 },
		{ 173, 209, 51 },
		{ 173, 210, 51 },
		{ 173, 211, 51 },
		{ 173, 212, 51 },
		{ 173, 213, 51 },
		{ 173, 214, 51 },
		{ 174, 195, 51 },
		{ 174, 196, 51 },
		{ 174, 197, 51 },
		{ 174, 198, 51 },
		{ 174, 199, 51 },
		{ 174, 200, 51 },
		{ 174, 201, 51 },
		{ 174, 202, 51 },
		{ 174, 203, 51 },
		{ 174, 204, 51 },
		{ 174, 205, 51 },
		{ 174, 206, 51 },
		{ 174, 207, 51 },
		{ 174, 208, 51 },
		{ 174, 209, 51 },
		{ 174, 210, 51 },
		{ 174, 211, 51 },
		{ 174, 212, 51 },
		{ 174, 213, 51 },
		{ 174, 214, 51 },
		{ 174, 215, 51 },
		{ 175, 196, 51 },
		{ 175, 197, 51 },
		{ 175, 198, 51 },
		{ 175, 199, 51 },
		{ 175, 200, 51 },
		{ 175, 201, 51 },
		{ 175, 202, 51 },
		{ 175, 203, 51 },
		{ 175, 204, 51 },
		{ 175, 205, 51 },
		{ 175, 206, 51 },
		{ 175, 207, 51 },
		{ 175, 208, 51 },
		{ 175, 209, 51 },
		{ 175, 210, 51 },
		{ 175, 211, 51 },
		{ 175, 212, 51 },
		{ 175, 213, 51 },
		{ 175, 214, 51 },
		{ 175, 215, 51 },
		{ 176, 197, 51 },
		{ 176, 198, 51 },
		{ 176, 199, 51 },
		{ 176, 200, 51 },
		{ 176, 201, 51 },
		{ 176, 202, 51 },
		{ 176, 203, 51 },
		{ 176, 204, 51 },
		{ 176, 205, 51 },
		{ 176, 206, 51 },
		{ 176, 207, 51 },
		{ 176, 208, 51 },
		{ 176, 209, 51 },
		{ 176, 210, 51 },
		{ 176, 211, 51 },
		{ 176, 212, 51 },
		{ 176, 213, 51 },
		{ 176, 214, 51 },
		{ 176, 215, 51 },
		{ 176, 216, 51 },
		{ 177, 198, 51 },
		{ 177, 199, 51 },
		{ 177, 200, 51 },
		{ 177, 201, 51 },
		{ 177, 202, 51 },
		{ 177, 203, 51 },
		{ 177, 204, 51 },
		{ 177, 205, 51 },
		{ 177, 206, 51 },
		{ 177, 207, 51 },
		{ 177, 208, 51 },
		{ 177, 209, 51 },
		{ 177, 210, 51 },
		{ 177, 211, 51 },
		{ 177, 212, 51 },
		{ 177, 213, 51 },
		{ 177, 214, 51 },
		{ 177, 215, 51 },
		{ 177, 216, 51 },
		{ 178, 198, 51 },
		{ 178, 199, 51 },
		{ 178, 200, 51 },
		{ 178, 201, 51 },
		{ 178, 202, 51 },
		{ 178, 203, 51 },
		{ 178, 204, 51 },
		{ 178, 205, 51 },
		{ 178, 206, 51 },
		{ 178, 207, 51 },
		{ 178, 208, 51 },
		{ 178, 209, 51 },
		{ 178, 210, 51 },
		{ 178, 211, 51 },
		{ 178, 212, 51 },
		{ 178, 213, 51 },
		{ 178, 214, 51 },
		{ 178, 215, 51 },
		{ 178, 216, 51 },
		{ 178, 217, 51 },
		{ 179, 199, 51 },
		{ 179, 200, 51 },
		{ 179, 201, 51 },
		{ 179, 202, 51 },
		{ 179, 203, 51 },
		{ 179, 204, 51 },
		{ 179, 205, 51 },
		{ 179, 206, 51 },
		{ 179, 207, 51 },
		{ 179, 208, 51 },
		{ 179, 209, 51 },
		{ 179, 210, 51 },
		{ 179, 211, 51 },
		{ 179, 212, 51 },
		{ 179, 213, 51 },
		{ 179, 214, 51 },
		{ 179, 215, 51 },
		{ 179, 216, 51 },
		{ 179, 217, 51 },
		{ 180, 200, 51 },
		{ 180, 201, 51 },
		{ 180, 202, 51 },
		{ 180, 203, 51 },
		{ 180, 204, 51 },
		{ 180, 205, 51 },
		{ 180, 206, 51 },
		{ 180, 207, 51 },
		{ 180, 208, 51 },
		{ 180, 209, 51 },
		{ 180, 210, 51 },
		{ 180, 211, 51 },
		{ 180, 212, 51 },
		{ 180, 213, 51 },
		{ 180, 214, 51 },
		{ 180, 215, 51 },
		{ 180, 216, 51 },
		{ 180, 217, 51 },
		{ 180, 218, 51 },
		{ 181, 201, 51 },
		{ 181, 202, 51 },
		{ 181, 203, 51 },
		{ 181, 204, 51 },
		{ 181, 205, 51 },
		{ 181, 206, 51 },
		{ 181, 207, 51 },
		{ 181, 208, 51 },
		{ 181, 209, 51 },
		{ 181, 210, 51 },
		{ 181, 211, 51 },
		{ 181, 212, 51 },
		{ 181, 213, 51 },
		{ 181, 214, 51 },
		{ 181, 215, 51 },
		{ 181, 216, 51 },
		{ 181, 217, 51 },
		{ 181, 218, 51 },
		{ 182, 201, 51 },
		{ 182, 202, 51 },
		{ 182, 203, 51 },
		{ 182, 204, 51 },
		{ 182, 205, 51 },
		{ 182, 206, 51 },
		{ 182, 207, 51 },
		{ 182, 208, 51 },
		{ 182, 209, 51 },
		{ 182, 210, 51 },
		{ 182, 211, 51 },
		{ 182, 212, 51 },
		{ 182, 213, 51 },
		{ 182, 214, 51 },
		{ 182, 215, 51 },
		{ 182, 216, 51 },
		{ 182, 217, 51 },
		{ 182, 218, 51 },
		{ 182, 219, 51 },
		{ 183, 202, 51 },
		{ 183, 203, 51 },
		{ 183, 204, 51 },
		{ 183, 205, 51 },
		{ 183, 206, 51 },
		{ 183, 207, 51 },
		{ 183, 208, 51 },
		{ 183, 209, 51 },
		{ 183, 210, 51 },
		{ 183, 211, 51 },
		{ 183, 212, 51 },
		{ 183, 213, 51 },
		{ 183, 214, 51 },
		{ 183, 215, 51 },
		{ 183, 216, 51 },
		{ 183, 217, 51 },
		{ 183, 218, 51 },
		{ 183, 219, 51 },
		{ 184, 203, 51 },
		{ 184, 204, 51 },
		{ 184, 205, 51 },
		{ 184, 206, 51 },
		{ 184, 207, 51 },
		{ 184, 208, 51 },
		{ 184, 209, 51 },
		{ 184, 210, 51 },
		{ 184, 211, 51 },
		{ 184, 212, 51 },
		{ 184, 213, 51 },
		{ 184, 214, 51 },
		{ 184, 215, 51 },
		{ 184, 216, 51 },
		{ 184, 217, 51 },
		{ 184, 218, 51 },
		{ 184, 219, 51 },
		{ 184, 220, 51 },
		{ 185, 204, 51 },
		{ 185, 205, 51 },
		{ 185, 206, 51 },
		{ 185, 207, 51 },
		{ 185, 208, 51 },
		{ 185, 209, 51 },
		{ 185, 210, 51 },
		{ 185, 211, 51 },
		{ 185, 212, 51 },
		{ 185, 213, 51 },
		{ 185, 214, 51 },
		{ 185, 215, 51 },
		{ 185, 216, 51 },
		{ 185, 217, 51 },
		{ 185, 218, 51 },
		{ 185, 219, 51 },
		{ 185, 220, 51 },
		{ 186, 204, 51 },
		{ 186, 205, 51 },
		{ 186, 206, 51 },
		{ 186, 207, 51 },
		{ 186, 208, 51 },
		{ 186, 209, 51 },
		{ 186, 210, 51 },
		{ 186, 211, 51 },
		{ 186, 212, 51 },
		{ 186, 213, 51 },
		{ 186, 214, 51 },
		{ 186, 215, 51 },
		{ 186, 216, 51 },
		{ 186, 217, 51 },
		{ 186, 218, 51 },
		{ 186, 219, 51 },
		{ 186, 220, 51 },
		{ 186, 221, 51 },
		{ 187, 205, 51 },
		{ 187, 206, 51 },
		{ 187, 207, 51 },
		{ 187, 208, 51 },
		{ 187, 209, 51 },
		{ 187, 210, 51 },
		{ 187, 211, 51 },
		{ 187, 212, 51 },
		{ 187, 213, 51 },
		{ 187, 214, 51 },
		{ 187, 215, 51 },
		{ 187, 216, 51 },
		{ 187, 217, 51 },
		{ 187, 218, 51 },
		{ 187, 219, 51 },
		{ 187, 220, 51 },
		{ 187, 221, 51 },
		{ 188, 206, 51 },
		{ 188, 207, 51 },
		{ 188, 208, 51 },
		{ 188, 209, 51 },
		{ 188, 210, 51 },
		{ 188, 211, 51 },
		{ 188, 212, 51 },
		{ 188, 213, 51 },
		{ 188, 214, 51 },
		{ 188, 215, 51 },
		{ 188, 216, 51 },
		{ 188, 217, 51 },
		{ 188, 218, 51 },
		{ 188, 219, 51 },
		{ 188, 220, 51 },
		{ 188, 221, 51 },
		{ 188, 222, 51 },
		{ 189, 207, 51 },
		{ 189, 208, 51 },
		{ 189, 209, 51 },
		{ 189, 210, 51 },
		{ 189, 211, 51 },
		{ 189, 212, 51 },
		{ 189, 213, 51 },
		{ 189, 214, 51 },
		{ 189, 215, 51 },
		{ 189, 216, 51 },
		{ 189, 217, 51 },
		{ 189, 218, 51 },
		{ 189, 219, 51 },
		{ 189, 220, 51 },
		{ 189, 221, 51 },
		{ 189, 222, 51 },
		{ 190, 208, 51 },
		{ 190, 209, 51 },
		{ 190, 210, 51 },
		{ 190, 211, 51 },
		{ 190, 212, 51 },
		{ 190, 213, 51 },
		{ 190, 214, 51 },
		{ 190, 215, 51 },
		{ 190, 216, 51 },
		{ 190, 217, 51 },
		{ 190, 218, 51 },
		{ 190, 219, 51 },
		{ 190, 220, 51 },
		{ 190, 221, 51 },
		{ 190, 222, 51 },
		{ 191, 210, 51 },
		{ 191, 211, 51 },
		{ 191, 212, 51 },
		{ 191, 213, 51 },
		{ 191, 214, 51 },
		{ 191, 215, 51 },
		{ 191, 216, 51 },
		{ 191, 217, 51 },
		{ 191, 218, 51 },
		{ 191, 219, 51 },
		{ 191, 220, 51 },
		{ 191, 221, 51 },
		{ 191, 222, 51 },
		{ 323, 292, 51 },
		{ 323, 293, 51 },
		{ 323, 294, 51 },
		{ 323, 295, 51 },
		{ 323, 296, 51 },
		{ 323, 297, 51 },
		{ 323, 298, 51 },
		{ 323, 299, 51 },
		{ 323, 300, 51 },
		{ 323, 301, 51 },
		{ 323, 302, 51 },
		{ 323, 303, 51 },
		{ 323, 304, 51 },
		{ 324, 292, 51 },
		{ 324, 293, 51 },
		{ 324, 294, 51 },
		{ 324, 295, 51 },
		{ 324, 296, 51 },
		{ 324, 297, 51 },
		{ 324, 298, 51 },
		{ 324, 299, 51 },
		{ 324, 300, 51 },
		{ 324, 301, 51 },
		{ 324, 302, 51 },
		{ 324, 303, 51 },
		{ 324, 304, 51 },
		{ 324, 305, 51 },
		{ 324, 306, 51 },
		{ 325, 292, 51 },
		{ 325, 293, 51 },
		{ 325, 294, 51 },
		{ 325, 295, 51 },
		{ 325, 296, 51 },
		{ 325, 297, 51 },
		{ 325, 298, 51 },
		{ 325, 299, 51 },
		{ 325, 300, 51 },
		{ 325, 301, 51 },
		{ 325, 302, 51 },
		{ 325, 303, 51 },
		{ 325, 304, 51 },
		{ 325, 305, 51 },
		{ 325, 306, 51 },
		{ 325, 307, 51 },
		{ 326, 292, 51 },
		{ 326, 293, 51 },
		{ 326, 294, 51 },
		{ 326, 295, 51 },
		{ 326, 296, 51 },
		{ 326, 297, 51 },
		{ 326, 298, 51 },
		{ 326, 299, 51 },
		{ 326, 300, 51 },
		{ 326, 301, 51 },
		{ 326, 302, 51 },
		{ 326, 303, 51 },
		{ 326, 304, 51 },
		{ 326, 305, 51 },
		{ 326, 306, 51 },
		{ 326, 307, 51 },
		{ 326, 308, 51 },
		{ 327, 293, 51 },
		{ 327, 294, 51 },
		{ 327, 295, 51 },
		{ 327, 296, 51 },
		{ 327, 297, 51 },
		{ 327, 298, 51 },
		{ 327, 299, 51 },
		{ 327, 300, 51 },
		{ 327, 301, 51 },
		{ 327, 302, 51 },
		{ 327, 303, 51 },
		{ 327, 304, 51 },
		{ 327, 305, 51 },
		{ 327, 306, 51 },
		{ 327, 307, 51 },
		{ 327, 308, 51 },
		{ 327, 309, 51 },
		{ 328, 293, 51 },
		{ 328, 294, 51 },
		{ 328, 295, 51 },
		{ 328, 296, 51 },
		{ 328, 297, 51 },
		{ 328, 298, 51 },
		{ 328, 299, 51 },
		{ 328, 300, 51 },
		{ 328, 301, 51 },
		{ 328, 302, 51 },
		{ 328, 303, 51 },
		{ 328, 304, 51 },
		{ 328, 305, 51 },
		{ 328, 306, 51 },
		{ 328, 307, 51 },
		{ 328, 308, 51 },
		{ 328, 309, 51 },
		{ 328, 310, 51 },
		{ 329, 294, 51 },
		{ 329, 295, 51 },
		{ 329, 296, 51 },
		{ 329, 297, 51 },
		{ 329, 298, 51 },
		{ 329, 299, 51 },
		{ 329, 300, 51 },
		{ 329, 301, 51 },
		{ 329, 302, 51 },
		{ 329, 303, 51 },
		{ 329, 304, 51 },
		{ 329, 305, 51 },
		{ 329, 306, 51 },
		{ 329, 307, 51 },
		{ 329, 308, 51 },
		{ 329, 309, 51 },
		{ 329, 310, 51 },
		{ 330, 294, 51 },
		{ 330, 295, 51 },
		{ 330, 296, 51 },
		{ 330, 297, 51 },
		{ 330, 298, 51 },
		{ 330, 299, 51 },
		{ 330, 300, 51 },
		{ 330, 301, 51 },
		{ 330, 302, 51 },
		{ 330, 303, 51 },
		{ 330, 304, 51 },
		{ 330, 305, 51 },
		{ 330, 306, 51 },
		{ 330, 307, 51 },
		{ 330, 308, 51 },
		{ 330, 309, 51 },
		{ 330, 310, 51 },
		{ 330, 311, 51 },
		{ 331, 295, 51 },
		{ 331, 296, 51 },
		{ 331, 297, 51 },
		{ 331, 298, 51 },
		{ 331, 299, 51 },
		{ 331, 300, 51 },
		{ 331, 301, 51 },
		{ 331, 302, 51 },
		{ 331, 303, 51 },
		{ 331, 304, 51 },
		{ 331, 305, 51 },
		{ 331, 306, 51 },
		{ 331, 307, 51 },
		{ 331, 308, 51 },
		{ 331, 309, 51 },
		{ 331, 310, 51 },
		{ 331, 311, 51 },
		{ 331, 312, 51 },
		{ 332, 295, 51 },
		{ 332, 296, 51 },
		{ 332, 297, 51 },
		{ 332, 298, 51 },
		{ 332, 299, 51 },
		{ 332, 300, 51 },
		{ 332, 301, 51 },
		{ 332, 302, 51 },
		{ 332, 303, 51 },
		{ 332, 304, 51 },
		{ 332, 305, 51 },
		{ 332, 306, 51 },
		{ 332, 307, 51 },
		{ 332, 308, 51 },
		{ 332, 309, 51 },
		{ 332, 310, 51 },
		{ 332, 311, 51 },
		{ 332, 312, 51 },
		{ 332, 313, 51 },
		{ 333, 296, 51 },
		{ 333, 297, 51 },
		{ 333, 298, 51 },
		{ 333, 299, 51 },
		{ 333, 300, 51 },
		{ 333, 301, 51 },
		{ 333, 302, 51 },
		{ 333, 303, 51 },
		{ 333, 304, 51 },
		{ 333, 305, 51 },
		{ 333, 306, 51 },
		{ 333, 307, 51 },
		{ 333, 308, 51 },
		{ 333, 309, 51 },
		{ 333, 310, 51 },
		{ 333, 311, 51 },
		{ 333, 312, 51 },
		{ 333, 313, 51 },
		{ 334, 296, 51 },
		{ 334, 297, 51 },
		{ 334, 298, 51 },
		{ 334, 299, 51 },
		{ 334, 300, 51 },
		{ 334, 301, 51 },
		{ 334, 302, 51 },
		{ 334, 303, 51 },
		{ 334, 304, 51 },
		{ 334, 305, 51 },
		{ 334, 306, 51 },
		{ 334, 307, 51 },
		{ 334, 308, 51 },
		{ 334, 309, 51 },
		{ 334, 310, 51 },
		{ 334, 311, 51 },
		{ 334, 312, 51 },
		{ 334, 313, 51 },
		{ 334, 314, 51 },
		{ 335, 297, 51 },
		{ 335, 298, 51 },
		{ 335, 299, 51 },
		{ 335, 300, 51 },
		{ 335, 301, 51 },
		{ 335, 302, 51 },
		{ 335, 303, 51 },
		{ 335, 304, 51 },
		{ 335, 305, 51 },
		{ 335, 306, 51 },
		{ 335, 307, 51 },
		{ 335, 308, 51 },
		{ 335, 309, 51 },
		{ 335, 310, 51 },
		{ 335, 311, 51 },
		{ 335, 312, 51 },
		{ 335, 313, 51 },
		{ 335, 314, 51 },
		{ 335, 315, 51 },
		{ 336, 297, 51 },
		{ 336, 298, 51 },
		{ 336, 299, 51 },
		{ 336, 300, 51 },
		{ 336, 301, 51 },
		{ 336, 302, 51 },
		{ 336, 303, 51 },
		{ 336, 304, 51 },
		{ 336, 305, 51 },
		{ 336, 306, 51 },
		{ 336, 307, 51 },
		{ 336, 308, 51 },
		{ 336, 309, 51 },
		{ 336, 310, 51 },
		{ 336, 311, 51 },
		{ 336, 312, 51 },
		{ 336, 313, 51 },
		{ 336, 314, 51 },
		{ 336, 315, 51 },
		{ 336, 316, 51 },
		{ 337, 298, 51 },
		{ 337, 299, 51 },
		{ 337, 300, 51 },
		{ 337, 301, 51 },
		{ 337, 302, 51 },
		{ 337, 303, 51 },
		{ 337, 304, 51 },
		{ 337, 305, 51 },
		{ 337, 306, 51 },
		{ 337, 307, 51 },
		{ 337, 308, 51 },
		{ 337, 309, 51 },
		{ 337, 310, 51 },
		{ 337, 311, 51 },
		{ 337, 312, 51 },
		{ 337, 313, 51 },
		{ 337, 314, 51 },
		{ 337, 315, 51 },
		{ 337, 316, 51 },
		{ 338, 298, 51 },
		{ 338, 299, 51 },
		{ 338, 300, 51 },
		{ 338, 301, 51 },
		{ 338, 302, 51 },
		{ 338, 303, 51 },
		{ 338, 304, 51 },
		{ 338, 305, 51 },
		{ 338, 306, 51 },
		{ 338, 307, 51 },
		{ 338, 308, 51 },
		{ 338, 309, 51 },
		{ 338, 310, 51 },
		{ 338, 311, 51 },
		{ 338, 312, 51 },
		{ 338, 313, 51 },
		{ 338, 314, 51 },
		{ 338, 315, 51 },
		{ 338, 316, 51 },
		{ 338, 317, 51 },
		{ 339, 299, 51 },
		{ 339, 300, 51 },
		{ 339, 301, 51 },
		{ 339, 302, 51 },
		{ 339, 303, 51 },
		{ 339, 304, 51 },
		{ 339, 305, 51 },
		{ 339, 306, 51 },
		{ 339, 307, 51 },
		{ 339, 308, 51 },
		{ 339, 309, 51 },
		{ 339, 310, 51 },
		{ 339, 311, 51 },
		{ 339, 312, 51 },
		{ 339, 313, 51 },
		{ 339, 314, 51 },
		{ 339, 315, 51 },
		{ 339, 316, 51 },
		{ 339, 317, 51 },
		{ 339, 318, 51 },
		{ 340, 299, 51 },
		{ 340, 300, 51 },
		{ 340, 301, 51 },
		{ 340, 302, 51 },
		{ 340, 303, 51 },
		{ 340, 304, 51 },
		{ 340, 305, 51 },
		{ 340, 306, 51 },
		{ 340, 307, 51 },
		{ 340, 308, 51 },
		{ 340, 309, 51 },
		{ 340, 310, 51 },
		{ 340, 311, 51 },
		{ 340, 312, 51 },
		{ 340, 313, 51 },
		{ 340, 314, 51 },
		{ 340, 315, 51 },
		{ 340, 316, 51 },
		{ 340, 317, 51 },
		{ 340, 318, 51 },
		{ 340, 319, 51 },
		{ 341, 300, 51 },
		{ 341, 301, 51 },
		{ 341, 302, 51 },
		{ 341, 303, 51 },
		{ 341, 304, 51 },
		{ 341, 305, 51 },
		{ 341, 306, 51 },
		{ 341, 307, 51 },
		{ 341, 308, 51 },
		{ 341, 309, 51 },
		{ 341, 310, 51 },
		{ 341, 311, 51 },
		{ 341, 312, 51 },
		{ 341, 313, 51 },
		{ 341, 314, 51 },
		{ 341, 315, 51 },
		{ 341, 316, 51 },
		{ 341, 317, 51 },
		{ 341, 318, 51 },
		{ 341, 319, 51 },
		{ 342, 300, 51 },
		{ 342, 301, 51 },
		{ 342, 302, 51 },
		{ 342, 303, 51 },
		{ 342, 304, 51 },
		{ 342, 305, 51 },
		{ 342, 306, 51 },
		{ 342, 307, 51 },
		{ 342, 308, 51 },
		{ 342, 309, 51 },
		{ 342, 310, 51 },
		{ 342, 311, 51 },
		{ 342, 312, 51 },
		{ 342, 313, 51 },
		{ 342, 314, 51 },
		{ 342, 315, 51 },
		{ 342, 316, 51 },
		{ 342, 317, 51 },
		{ 342, 318, 51 },
		{ 342, 319, 51 },
		{ 342, 320, 51 },
		{ 343, 301, 51 },
		{ 343, 302, 51 },
		{ 343, 303, 51 },
		{ 343, 304, 51 },
		{ 343, 305, 51 },
		{ 343, 306, 51 },
		{ 343, 307, 51 },
		{ 343, 308, 51 },
		{ 343, 309, 51 },
		{ 343, 310, 51 },
		{ 343, 311, 51 },
		{ 343, 312, 51 },
		{ 343, 313, 51 },
		{ 343, 314, 51 },
		{ 343, 315, 51 },
		{ 343, 316, 51 },
		{ 343, 317, 51 },
		{ 343, 318, 51 },
		{ 343, 319, 51 },
		{ 343, 320, 51 },
		{ 343, 321, 51 },
		{ 344, 301, 51 },
		{ 344, 302, 51 },
		{ 344, 303, 51 },
		{ 344, 304, 51 },
		{ 344, 305, 51 },
		{ 344, 306, 51 },
		{ 344, 307, 51 },
		{ 344, 308, 51 },
		{ 344, 309, 51 },
		{ 344, 310, 51 },
		{ 344, 311, 51 },
		{ 344, 312, 51 },
		{ 344, 313, 51 },
		{ 344, 314, 51 },
		{ 344, 315, 51 },
		{ 344, 316, 51 },
		{ 344, 317, 51 },
		{ 344, 318, 51 },
		{ 344, 319, 51 },
		{ 344, 320, 51 },
		{ 344, 321, 51 },
		{ 344, 322, 51 },
		{ 345, 302, 51 },
		{ 345, 303, 51 },
		{ 345, 304, 51 },
		{ 345, 305, 51 },
		{ 345, 306, 51 },
		{ 345, 307, 51 },
		{ 345, 308, 51 },
		{ 345, 309, 51 },
		{ 345, 310, 51 },
		{ 345, 311, 51 },
		{ 345, 312, 51 },
		{ 345, 313, 51 },
		{ 345, 314, 51 },
		{ 345, 315, 51 },
		{ 345, 316, 51 },
		{ 345, 317, 51 },
		{ 345, 318, 51 },
		{ 345, 319, 51 },
		{ 345, 320, 51 },
		{ 345, 321, 51 },
		{ 345, 322, 51 },
		{ 346, 302, 51 },
		{ 346, 303, 51 },
		{ 346, 304, 51 },
		{ 346, 305, 51 },
		{ 346, 306, 51 },
		{ 346, 307, 51 },
		{ 346, 308, 51 },
		{ 346, 309, 51 },
		{ 346, 310, 51 },
		{ 346, 311, 51 },
		{ 346, 312, 51 },
		{ 346, 313, 51 },
		{ 346, 314, 51 },
		{ 346, 315, 51 },
		{ 346, 316, 51 },
		{ 346, 317, 51 },
		{ 346, 318, 51 },
		{ 346, 319, 51 },
		{ 346, 320, 51 },
		{ 346, 321, 51 },
		{ 346, 322, 51 },
		{ 346, 323, 51 },
		{ 347, 303, 51 },
		{ 347, 304, 51 },
		{ 347, 305, 51 },
		{ 347, 306, 51 },
		{ 347, 307, 51 },
		{ 347, 308, 51 },
		{ 347, 309, 51 },
		{ 347, 310, 51 },
		{ 347, 311, 51 },
		{ 347, 312, 51 },
		{ 347, 313, 51 },
		{ 347, 314, 51 },
		{ 347, 315, 51 },
		{ 347, 316, 51 },
		{ 347, 317, 51 },
		{ 347, 318, 51 },
		{ 347, 319, 51 },
		{ 347, 320, 51 },
		{ 347, 321, 51 },
		{ 347, 322, 51 },
		{ 347, 323, 51 },
		{ 347, 324, 51 },
		{ 348, 303, 51 },
		{ 348, 304, 51 },
		{ 348, 305, 51 },
		{ 348, 306, 51 },
		{ 348, 307, 51 },
		{ 348, 308, 51 },
		{ 348, 309, 51 },
		{ 348, 310, 51 },
		{ 348, 311, 51 },
		{ 348, 312, 51 },
		{ 348, 313, 51 },
		{ 348, 314, 51 },
		{ 348, 315, 51 },
		{ 348, 316, 51 },
		{ 348, 317, 51 },
		{ 348, 318, 51 },
		{ 348, 319, 51 },
		{ 348, 320, 51 },
		{ 348, 321, 51 },
		{ 348, 322, 51 },
		{ 348, 323, 51 },
		{ 348, 324, 51 },
		{ 348, 325, 51 },
		{ 349, 304, 51 },
		{ 349, 305, 51 },
		{ 349, 306, 51 },
		{ 349, 307, 51 },
		{ 349, 308, 51 },
		{ 349, 309, 51 },
		{ 349, 310, 51 },
		{ 349, 311, 51 },
		{ 349, 312, 51 },
		{ 349, 313, 51 },
		{ 349, 314, 51 },
		{ 349, 315, 51 },
		{ 349, 316, 51 },
		{ 349, 317, 51 },
		{ 349, 318, 51 },
		{ 349, 319, 51 },
		{ 349, 320, 51 },
		{ 349, 321, 51 },
		{ 349, 322, 51 },
		{ 349, 323, 51 },
		{ 349, 324, 51 },
		{ 349, 325, 51 },
		{ 350, 304, 51 },
		{ 350, 305, 51 },
		{ 350, 306, 51 },
		{ 350, 307, 51 },
		{ 350, 308, 51 },
		{ 350, 309, 51 },
		{ 350, 310, 51 },
		{ 350, 311, 51 },
		{ 350, 312, 51 },
		{ 350, 313, 51 },
		{ 350, 314, 51 },
		{ 350, 315, 51 },
		{ 350, 316, 51 },
		{ 350, 317, 51 },
		{ 350, 318, 51 },
		{ 350, 319, 51 },
		{ 350, 320, 51 },
		{ 350, 321, 51 },
		{ 350, 322, 51 },
		{ 350, 323, 51 },
		{ 350, 324, 51 },
		{ 350, 325, 51 },
		{ 350, 326, 51 },
		{ 351, 305, 51 },
		{ 351, 306, 51 },
		{ 351, 307, 51 },
		{ 351, 308, 51 },
		{ 351, 309, 51 },
		{ 351, 310, 51 },
		{ 351, 311, 51 },
		{ 351, 312, 51 },
		{ 351, 313, 51 },
		{ 351, 314, 51 },
		{ 351, 315, 51 },
		{ 351, 316, 51 },
		{ 351, 317, 51 },
		{ 351, 318, 51 },
		{ 351, 319, 51 },
		{ 351, 320, 51 },
		{ 351, 321, 51 },
		{ 351, 322, 51 },
		{ 351, 323, 51 },
		{ 351, 324, 51 },
		{ 351, 325, 51 },
		{ 351, 326, 51 },
		{ 351, 327, 51 },
		{ 352, 305, 51 },
		{ 352, 306, 51 },
		{ 352, 307, 51 },
		{ 352, 308, 51 },
		{ 352, 309, 51 },
		{ 352, 310, 51 },
		{ 352, 311, 51 },
		{ 352, 312, 51 },
		{ 352, 313, 51 },
		{ 352, 314, 51 },
		{ 352, 315, 51 },
		{ 352, 316, 51 },
		{ 352, 317, 51 },
		{ 352, 318, 51 },
		{ 352, 319, 51 },
		{ 352, 320, 51 },
		{ 352, 321, 51 },
		{ 352, 322, 51 },
		{ 352, 323, 51 },
		{ 352, 324, 51 },
		{ 352, 325, 51 },
		{ 352, 326, 51 },
		{ 352, 327, 51 },
		{ 352, 328, 51 },
		{ 353, 306, 51 },
		{ 353, 307, 51 },
		{ 353, 308, 51 },
		{ 353, 309, 51 },
		{ 353, 310, 51 },
		{ 353, 311, 51 },
		{ 353, 312, 51 },
		{ 353, 313, 51 },
		{ 353, 314, 51 },
		{ 353, 315, 51 },
		{ 353, 316, 51 },
		{ 353, 317, 51 },
		{ 353, 318, 51 },
		{ 353, 319, 51 },
		{ 353, 320, 51 },
		{ 353, 321, 51 },
		{ 353, 322, 51 },
		{ 353, 323, 51 },
		{ 353, 324, 51 },
		{ 353, 325, 51 },
		{ 353, 326, 51 },
		{ 353, 327, 51 },
		{ 353, 328, 51 },
		{ 354, 306, 51 },
		{ 354, 307, 51 },
		{ 354, 308, 51 },
		{ 354, 309, 51 },
		{ 354, 310, 51 },
		{ 354, 311, 51 },
		{ 354, 312, 51 },
		{ 354, 313, 51 },
		{ 354, 314, 51 },
		{ 354, 315, 51 },
		{ 354, 316, 51 },
		{ 354, 317, 51 },
		{ 354, 318, 51 },
		{ 354, 319, 51 },
		{ 354, 320, 51 },
		{ 354, 321, 51 },
		{ 354, 322, 51 },
		{ 354, 323, 51 },
		{ 354, 324, 51 },
		{ 354, 325, 51 },
		{ 354, 326, 51 },
		{ 354, 327, 51 },
		{ 354, 328, 51 },
		{ 354, 329, 51 },
		{ 355, 307, 51 },
		{ 355, 308, 51 },
		{ 355, 309, 51 },
		{ 355, 310, 51 },
		{ 355, 311, 51 },
		{ 355, 312, 51 },
		{ 355, 313, 51 },
		{ 355, 314, 51 },
		{ 355, 315, 51 },
		{ 355, 316, 51 },
		{ 355, 317, 51 },
		{ 355, 318, 51 },
		{ 355, 319, 51 },
		{ 355, 320, 51 },
		{ 355, 321, 51 },
		{ 355, 322, 51 },
		{ 355, 323, 51 },
		{ 355, 324, 51 },
		{ 355, 325, 51 },
		{ 355, 326, 51 },
		{ 355, 327, 51 },
		{ 355, 328, 51 },
		{ 355, 329, 51 },
		{ 355, 330, 51 },
		{ 356, 307, 51 },
		{ 356, 308, 51 },
		{ 356, 309, 51 },
		{ 356, 310, 51 },
		{ 356, 311, 51 },
		{ 356, 312, 51 },
		{ 356, 313, 51 },
		{ 356, 314, 51 },
		{ 356, 315, 51 },
		{ 356, 316, 51 },
		{ 356, 317, 51 },
		{ 356, 318, 51 },
		{ 356, 319, 51 },
		{ 356, 320, 51 },
		{ 356, 321, 51 },
		{ 356, 322, 51 },
		{ 356, 323, 51 },
		{ 356, 324, 51 },
		{ 356, 325, 51 },
		{ 356, 326, 51 },
		{ 356, 327, 51 },
		{ 356, 328, 51 },
		{ 356, 329, 51 },
		{ 356, 330, 51 },
		{ 356, 331, 51 },
		{ 357, 308, 51 },
		{ 357, 309, 51 },
		{ 357, 310, 51 },
		{ 357, 311, 51 },
		{ 357, 312, 51 },
		{ 357, 313, 51 },
		{ 357, 314, 51 },
		{ 357, 315, 51 },
		{ 357, 316, 51 },
		{ 357, 317, 51 },
		{ 357, 318, 51 },
		{ 357, 319, 51 },
		{ 357, 320, 51 },
		{ 357, 321, 51 },
		{ 357, 322, 51 },
		{ 357, 323, 51 },
		{ 357, 324, 51 },
		{ 357, 325, 51 },
		{ 357, 326, 51 },
		{ 357, 327, 51 },
		{ 357, 328, 51 },
		{ 357, 329, 51 },
		{ 357, 330, 51 },
		{ 357, 331, 51 },
		{ 358, 308, 51 },
		{ 358, 309, 51 },
		{ 358, 310, 51 },
		{ 358, 311, 51 },
		{ 358, 312, 51 },
		{ 358, 313, 51 },
		{ 358, 314, 51 },
		{ 358, 315, 51 },
		{ 358, 316, 51 },
		{ 358, 317, 51 },
		{ 358, 318, 51 },
		{ 358, 319, 51 },
		{ 358, 320, 51 },
		{ 358, 321, 51 },
		{ 358, 322, 51 },
		{ 358, 323, 51 },
		{ 358, 324, 51 },
		{ 358, 325, 51 },
		{ 358, 326, 51 },
		{ 358, 327, 51 },
		{ 358, 328, 51 },
		{ 358, 329, 51 },
		{ 358, 330, 51 },
		{ 358, 331, 51 },
		{ 358, 332, 51 },
		{ 359, 309, 51 },
		{ 359, 310, 51 },
		{ 359, 311, 51 },
		{ 359, 312, 51 },
		{ 359, 313, 51 },
		{ 359, 314, 51 },
		{ 359, 315, 51 },
		{ 359, 316, 51 },
		{ 359, 317, 51 },
		{ 359, 318, 51 },
		{ 359, 319, 51 },
		{ 359, 320, 51 },
		{ 359, 321, 51 },
		{ 359, 322, 51 },
		{ 359, 323, 51 },
		{ 359, 324, 51 },
		{ 359, 325, 51 },
		{ 359, 326, 51 },
		{ 359, 327, 51 },
		{ 359, 328, 51 },
		{ 359, 329, 51 },
		{ 359, 330, 51 },
		{ 359, 331, 51 },
		{ 359, 332, 51 },
		{ 359, 333, 51 },
		{ 360, 309, 51 },
		{ 360, 310, 51 },
		{ 360, 311, 51 },
		{ 360, 312, 51 },
		{ 360, 313, 51 },
		{ 360, 314, 51 },
		{ 360, 315, 51 },
		{ 360, 316, 51 },
		{ 360, 317, 51 },
		{ 360, 318, 51 },
		{ 360, 319, 51 },
		{ 360, 320, 51 },
		{ 360, 321, 51 },
		{ 360, 322, 51 },
		{ 360, 323, 51 },
		{ 360, 324, 51 },
		{ 360, 325, 51 },
		{ 360, 326, 51 },
		{ 360, 327, 51 },
		{ 360, 328, 51 },
		{ 360, 329, 51 },
		{ 360, 330, 51 },
		{ 360, 331, 51 },
		{ 360, 332, 51 },
		{ 360, 333, 51 },
		{ 360, 334, 51 },
		{ 361, 310, 51 },
		{ 361, 311, 51 },
		{ 361, 312, 51 },
		{ 361, 313, 51 },
		{ 361, 314, 51 },
		{ 361, 315, 51 },
		{ 361, 316, 51 },
		{ 361, 317, 51 },
		{ 361, 318, 51 },
		{ 361, 319, 51 },
		{ 361, 320, 51 },
		{ 361, 321, 51 },
		{ 361, 322, 51 },
		{ 361, 323, 51 },
		{ 361, 324, 51 },
		{ 361, 325, 51 },
		{ 361, 326, 51 },
		{ 361, 327, 51 },
		{ 361, 328, 51 },
		{ 361, 329, 51 },
		{ 361, 330, 51 },
		{ 361, 331, 51 },
		{ 361, 332, 51 },
		{ 361, 333, 51 },
		{ 361, 334, 51 },
		{ 362, 310, 51 },
		{ 362, 311, 51 },
		{ 362, 312, 51 },
		{ 362, 313, 51 },
		{ 362, 314, 51 },
		{ 362, 315, 51 },
		{ 362, 316, 51 },
		{ 362, 317, 51 },
		{ 362, 318, 51 },
		{ 362, 319, 51 },
		{ 362, 320, 51 },
		{ 362, 321, 51 },
		{ 362, 322, 51 },
		{ 362, 323, 51 },
		{ 362, 324, 51 },
		{ 362, 325, 51 },
		{ 362, 326, 51 },
		{ 362, 327, 51 },
		{ 362, 328, 51 },
		{ 362, 329, 51 },
		{ 362, 330, 51 },
		{ 362, 331, 51 },
		{ 362, 332, 51 },
		{ 362, 333, 51 },
		{ 362, 334, 51 },
		{ 362, 335, 51 },
		{ 363, 311, 51 },
		{ 363, 312, 51 },
		{ 363, 313, 51 },
		{ 363, 314, 51 },
		{ 363, 315, 51 },
		{ 363, 316, 51 },
		{ 363, 317, 51 },
		{ 363, 318, 51 },
		{ 363, 319, 51 },
		{ 363, 320, 51 },
		{ 363, 321, 51 },
		{ 363, 322, 51 },
		{ 363, 323, 51 },
		{ 363, 324, 51 },
		{ 363, 325, 51 },
		{ 363, 326, 51 },
		{ 363, 327, 51 },
		{ 363, 328, 51 },
		{ 363, 329, 51 },
		{ 363, 330, 51 },
		{ 363, 331, 51 },
		{ 363, 332, 51 },
		{ 363, 333, 51 },
		{ 363, 334, 51 },
		{ 363, 335, 51 },
		{ 363, 336, 51 },
		{ 364, 311, 51 },
		{ 364, 312, 51 },
		{ 364, 313, 51 },
		{ 364, 314, 51 },
		{ 364, 315, 51 },
		{ 364, 316, 51 },
		{ 364, 317, 51 },
		{ 364, 318, 51 },
		{ 364, 319, 51 },
		{ 364, 320, 51 },
		{ 364, 321, 51 },
		{ 364, 322, 51 },
		{ 364, 323, 51 },
		{ 364, 324, 51 },
		{ 364, 325, 51 },
		{ 364, 326, 51 },
		{ 364, 327, 51 },
		{ 364, 328, 51 },
		{ 364, 329, 51 },
		{ 364, 330, 51 },
		{ 364, 331, 51 },
		{ 364, 332, 51 },
		{ 364, 333, 51 },
		{ 364, 334, 51 },
		{ 364, 335, 51 },
		{ 364, 336, 51 },
		{ 364, 337, 51 },
		{ 365, 312, 51 },
		{ 365, 313, 51 },
		{ 365, 314, 51 },
		{ 365, 315, 51 },
		{ 365, 316, 51 },
		{ 365, 317, 51 },
		{ 365, 318, 51 },
		{ 365, 319, 51 },
		{ 365, 320, 51 },
		{ 365, 321, 51 },
		{ 365, 322, 51 },
		{ 365, 323, 51 },
		{ 365, 324, 51 },
		{ 365, 325, 51 },
		{ 365, 326, 51 },
		{ 365, 327, 51 },
		{ 365, 328, 51 },
		{ 365, 329, 51 },
		{ 365, 330, 51 },
		{ 365, 331, 51 },
		{ 365, 332, 51 },
		{ 365, 333, 51 },
		{ 365, 334, 51 },
		{ 365, 335, 51 },
		{ 365, 336, 51 },
		{ 365, 337, 51 },
		{ 366, 312, 51 },
		{ 366, 313, 51 },
		{ 366, 314, 51 },
		{ 366, 315, 51 },
		{ 366, 316, 51 },
		{ 366, 317, 51 },
		{ 366, 318, 51 },
		{ 366, 319, 51 },
		{ 366, 320, 51 },
		{ 366, 321, 51 },
		{ 366, 322, 51 },
		{ 366, 323, 51 },
		{ 366, 324, 51 },
		{ 366, 325, 51 },
		{ 366, 326, 51 },
		{ 366, 327, 51 },
		{ 366, 328, 51 },
		{ 366, 329, 51 },
		{ 366, 330, 51 },
		{ 366, 331, 51 },
		{ 366, 332, 51 },
		{ 366, 333, 51 },
		{ 366, 334, 51 },
		{ 366, 335, 51 },
		{ 366, 336, 51 },
		{ 366, 337, 51 },
		{ 366, 338, 51 },
		{ 367, 313, 51 },
		{ 367, 314, 51 },
		{ 367, 315, 51 },
		{ 367, 316, 51 },
		{ 367, 317, 51 },
		{ 367, 318, 51 },
		{ 367, 319, 51 },
		{ 367, 320, 51 },
		{ 367, 321, 51 },
		{ 367, 322, 51 },
		{ 367, 323, 51 },
		{ 367, 324, 51 },
		{ 367, 325, 51 },
		{ 367, 326, 51 },
		{ 367, 327, 51 },
		{ 367, 328, 51 },
		{ 367, 329, 51 },
		{ 367, 330, 51 },
		{ 367, 331, 51 },
		{ 367, 332, 51 },
		{ 367, 333, 51 },
		{ 367, 334, 51 },
		{ 367, 335, 51 },
		{ 367, 336, 51 },
		{ 367, 337, 51 },
		{ 367, 338, 51 },
		{ 367, 339, 51 },
		{ 368, 313, 51 },
		{ 368, 314, 51 },
		{ 368, 315, 51 },
		{ 368, 316, 51 },
		{ 368, 317, 51 },
		{ 368, 318, 51 },
		{ 368, 319, 51 },
		{ 368, 320, 51 },
		{ 368, 321, 51 },
		{ 368, 322, 51 },
		{ 368, 323, 51 },
		{ 368, 324, 51 },
		{ 368, 325, 51 },
		{ 368, 326, 51 },
		{ 368, 327, 51 },
		{ 368, 328, 51 },
		{ 368, 329, 51 },
		{ 368, 330, 51 },
		{ 368, 331, 51 },
		{ 368, 332, 51 },
		{ 368, 333, 51 },
		{ 368, 334, 51 },
		{ 368, 335, 51 },
		{ 368, 336, 51 },
		{ 368, 337, 51 },
		{ 368, 338, 51 },
		{ 368, 339, 51 },
		{ 368, 340, 51 },
		{ 369, 314, 51 },
		{ 369, 315, 51 },
		{ 369, 316, 51 },
		{ 369, 317, 51 },
		{ 369, 318, 51 },
		{ 369, 319, 51 },
		{ 369, 320, 51 },
		{ 369, 321, 51 },
		{ 369, 322, 51 },
		{ 369, 323, 51 },
		{ 369, 324, 51 },
		{ 369, 325, 51 },
		{ 369, 326, 51 },
		{ 369, 327, 51 },
		{ 369, 328, 51 },
		{ 369, 329, 51 },
		{ 369, 330, 51 },
		{ 369, 331, 51 },
		{ 369, 332, 51 },
		{ 369, 333, 51 },
		{ 369, 334, 51 },
		{ 369, 335, 51 },
		{ 369, 336, 51 },
		{ 369, 337, 51 },
		{ 369, 338, 51 },
		{ 369, 339, 51 },
		{ 369, 340, 51 },
		{ 370, 314, 51 },
		{ 370, 315, 51 },
		{ 370, 316, 51 },
		{ 370, 317, 51 },
		{ 370, 318, 51 },
		{ 370, 319, 51 },
		{ 370, 320, 51 },
		{ 370, 321, 51 },
		{ 370, 322, 51 },
		{ 370, 323, 51 },
		{ 370, 324, 51 },
		{ 370, 325, 51 },
		{ 370, 326, 51 },
		{ 370, 327, 51 },
		{ 370, 328, 51 },
		{ 370, 329, 51 },
		{ 370, 330, 51 },
		{ 370, 331, 51 },
		{ 370, 332, 51 },
		{ 370, 333, 51 },
		{ 370, 334, 51 },
		{ 370, 335, 51 },
		{ 370, 336, 51 },
		{ 370, 337, 51 },
		{ 370, 338, 51 },
		{ 370, 339, 51 },
		{ 370, 340, 51 },
		{ 370, 341, 51 },
		{ 371, 315, 51 },
		{ 371, 316, 51 },
		{ 371, 317, 51 },
		{ 371, 318, 51 },
		{ 371, 319, 51 },
		{ 371, 320, 51 },
		{ 371, 321, 51 },
		{ 371, 322, 51 },
		{ 371, 323, 51 },
		{ 371, 324, 51 },
		{ 371, 325, 51 },
		{ 371, 326, 51 },
		{ 371, 327, 51 },
		{ 371, 328, 51 },
		{ 371, 329, 51 },
		{ 371, 330, 51 },
		{ 371, 331, 51 },
		{ 371, 332, 51 },
		{ 371, 333, 51 },
		{ 371, 334, 51 },
		{ 371, 335, 51 },
		{ 371, 336, 51 },
		{ 371, 337, 51 },
		{ 371, 338, 51 },
		{ 371, 339, 51 },
		{ 371, 340, 51 },
		{ 371, 341, 51 },
		{ 371, 342, 51 },
		{ 372, 315, 51 },
		{ 372, 316, 51 },
		{ 372, 317, 51 },
		{ 372, 318, 51 },
		{ 372, 319, 51 },
		{ 372, 320, 51 },
		{ 372, 321, 51 },
		{ 372, 322, 51 },
		{ 372, 323, 51 },
		{ 372, 324, 51 },
		{ 372, 325, 51 },
		{ 372, 326, 51 },
		{ 372, 327, 51 },
		{ 372, 328, 51 },
		{ 372, 329, 51 },
		{ 372, 330, 51 },
		{ 372, 331, 51 },
		{ 372, 332, 51 },
		{ 372, 333, 51 },
		{ 372, 334, 51 },
		{ 372, 335, 51 },
		{ 372, 336, 51 },
		{ 372, 337, 51 },
		{ 372, 338, 51 },
		{ 372, 339, 51 },
		{ 372, 340, 51 },
		{ 372, 341, 51 },
		{ 372, 342, 51 },
		{ 372, 343, 51 },
		{ 373, 316, 51 },
		{ 373, 317, 51 },
		{ 373, 318, 51 },
		{ 373, 319, 51 },
		{ 373, 320, 51 },
		{ 373, 321, 51 },
		{ 373, 322, 51 },
		{ 373, 323, 51 },
		{ 373, 324, 51 },
		{ 373, 325, 51 },
		{ 373, 326, 51 },
		{ 373, 327, 51 },
		{ 373, 328, 51 },
		{ 373, 329, 51 },
		{ 373, 330, 51 },
		{ 373, 331, 51 },
		{ 373, 332, 51 },
		{ 373, 333, 51 },
		{ 373, 334, 51 },
		{ 373, 335, 51 },
		{ 373, 336, 51 },
		{ 373, 337, 51 },
		{ 373, 338, 51 },
		{ 373, 339, 51 },
		{ 373, 340, 51 },
		{ 373, 341, 51 },
		{ 373, 342, 51 },
		{ 373, 343, 51 },
		{ 374, 316, 51 },
		{ 374, 317, 51 },
		{ 374, 318, 51 },
		{ 374, 319, 51 },
		{ 374, 320, 51 },
		{ 374, 321, 51 },
		{ 374, 322, 51 },
		{ 374, 323, 51 },
		{ 374, 324, 51 },
		{ 374, 325, 51 },
		{ 374, 326, 51 },
		{ 374, 327, 51 },
		{ 374, 328, 51 },
		{ 374, 329, 51 },
		{ 374, 330, 51 },
		{ 374, 331, 51 },
		{ 374, 332, 51 },
		{ 374, 333, 51 },
		{ 374, 334, 51 },
		{ 374, 335, 51 },
		{ 374, 336, 51 },
		{ 374, 337, 51 },
		{ 374, 338, 51 },
		{ 374, 339, 51 },
		{ 374, 340, 51 },
		{ 374, 341, 51 },
		{ 374, 342, 51 },
		{ 374, 343, 51 },
		{ 374, 344, 51 },
		{ 375, 317, 51 },
		{ 375, 318, 51 },
		{ 375, 319, 51 },
		{ 375, 320, 51 },
		{ 375, 321, 51 },
		{ 375, 322, 51 },
		{ 375, 323, 51 },
		{ 375, 324, 51 },
		{ 375, 325, 51 },
		{ 375, 326, 51 },
		{ 375, 327, 51 },
		{ 375, 328, 51 },
		{ 375, 329, 51 },
		{ 375, 330, 51 },
		{ 375, 331, 51 },
		{ 375, 332, 51 },
		{ 375, 333, 51 },
		{ 375, 334, 51 },
		{ 375, 335, 51 },
		{ 375, 336, 51 },
		{ 375, 337, 51 },
		{ 375, 338, 51 },
		{ 375, 339, 51 },
		{ 375, 340, 51 },
		{ 375, 341, 51 },
		{ 375, 342, 51 },
		{ 375, 343, 51 },
		{ 375, 344, 51 },
		{ 375, 345, 51 },
		{ 376, 317, 51 },
		{ 376, 318, 51 },
		{ 376, 319, 51 },
		{ 376, 320, 51 },
		{ 376, 321, 51 },
		{ 376, 322, 51 },
		{ 376, 323, 51 },
		{ 376, 324, 51 },
		{ 376, 325, 51 },
		{ 376, 326, 51 },
		{ 376, 327, 51 },
		{ 376, 328, 51 },
		{ 376, 329, 51 },
		{ 376, 330, 51 },
		{ 376, 331, 51 },
		{ 376, 332, 51 },
		{ 376, 333, 51 },
		{ 376, 334, 51 },
		{ 376, 335, 51 },
		{ 376, 336, 51 },
		{ 376, 337, 51 },
		{ 376, 338, 51 },
		{ 376, 339, 51 },
		{ 376, 340, 51 },
		{ 376, 341, 51 },
		{ 376, 342, 51 },
		{ 376, 343, 51 },
		{ 376, 344, 51 },
		{ 376, 345, 51 },
		{ 376, 346, 51 },
		{ 377, 318, 51 },
		{ 377, 319, 51 },
		{ 377, 320, 51 },
		{ 377, 321, 51 },
		{ 377, 322, 51 },
		{ 377, 323, 51 },
		{ 377, 324, 51 },
		{ 377, 325, 51 },
		{ 377, 326, 51 },
		{ 377, 327, 51 },
		{ 377, 328, 51 },
		{ 377, 329, 51 },
		{ 377, 330, 51 },
		{ 377, 331, 51 },
		{ 377, 332, 51 },
		{ 377, 333, 51 },
		{ 377, 334, 51 },
		{ 377, 335, 51 },
		{ 377, 336, 51 },
		{ 377, 337, 51 },
		{ 377, 338, 51 },
		{ 377, 339, 51 },
		{ 377, 340, 51 },
		{ 377, 341, 51 },
		{ 377, 342, 51 },
		{ 377, 343, 51 },
		{ 377, 344, 51 },
		{ 377, 345, 51 },
		{ 377, 346, 51 },
		{ 378, 318, 51 },
		{ 378, 319, 51 },
		{ 378, 320, 51 },
		{ 378, 321, 51 },
		{ 378, 322, 51 },
		{ 378, 323, 51 },
		{ 378, 324, 51 },
		{ 378, 325, 51 },
		{ 378, 326, 51 },
		{ 378, 327, 51 },
		{ 378, 328, 51 },
		{ 378, 329, 51 },
		{ 378, 330, 51 },
		{ 378, 331, 51 },
		{ 378, 332, 51 },
		{ 378, 333, 51 },
		{ 378, 334, 51 },
		{ 378, 335, 51 },
		{ 378, 336, 51 },
		{ 378, 337, 51 },
		{ 378, 338, 51 },
		{ 378, 339, 51 },
		{ 378, 340, 51 },
		{ 378, 341, 51 },
		{ 378, 342, 51 },
		{ 378, 343, 51 },
		{ 378, 344, 51 },
		{ 378, 345, 51 },
		{ 378, 346, 51 },
		{ 378, 347, 51 },
		{ 379, 319, 51 },
		{ 379, 320, 51 },
		{ 379, 321, 51 },
		{ 379, 322, 51 },
		{ 379, 323, 51 },
		{ 379, 324, 51 },
		{ 379, 325, 51 },
		{ 379, 326, 51 },
		{ 379, 327, 51 },
		{ 379, 328, 51 },
		{ 379, 329, 51 },
		{ 379, 330, 51 },
		{ 379, 331, 51 },
		{ 379, 332, 51 },
		{ 379, 333, 51 },
		{ 379, 334, 51 },
		{ 379, 335, 51 },
		{ 379, 336, 51 },
		{ 379, 337, 51 },
		{ 379, 338, 51 },
		{ 379, 339, 51 },
		{ 379, 340, 51 },
		{ 379, 341, 51 },
		{ 379, 342, 51 },
		{ 379, 343, 51 },
		{ 379, 344, 51 },
		{ 379, 345, 51 },
		{ 379, 346, 51 },
		{ 379, 347, 51 },
		{ 379, 348, 51 },
		{ 380, 319, 51 },
		{ 380, 320, 51 },
		{ 380, 321, 51 },
		{ 380, 322, 51 },
		{ 380, 323, 51 },
		{ 380, 324, 51 },
		{ 380, 325, 51 },
		{ 380, 326, 51 },
		{ 380, 327, 51 },
		{ 380, 328, 51 },
		{ 380, 329, 51 },
		{ 380, 330, 51 },
		{ 380, 331, 51 },
		{ 380, 332, 51 },
		{ 380, 333, 51 },
		{ 380, 334, 51 },
		{ 380, 335, 51 },
		{ 380, 336, 51 },
		{ 380, 337, 51 },
		{ 380, 338, 51 },
		{ 380, 339, 51 },
		{ 380, 340, 51 },
		{ 380, 341, 51 },
		{ 380, 342, 51 },
		{ 380, 343, 51 },
		{ 380, 344, 51 },
		{ 380, 345, 51 },
		{ 380, 346, 51 },
		{ 380, 347, 51 },
		{ 380, 348, 51 },
		{ 380, 349, 51 },
		{ 381, 320, 51 },
		{ 381, 321, 51 },
		{ 381, 322, 51 },
		{ 381, 323, 51 },
		{ 381, 324, 51 },
		{ 381, 325, 51 },
		{ 381, 326, 51 },
		{ 381, 327, 51 },
		{ 381, 328, 51 },
		{ 381, 329, 51 },
		{ 381, 330, 51 },
		{ 381, 331, 51 },
		{ 381, 332, 51 },
		{ 381, 333, 51 },
		{ 381, 334, 51 },
		{ 381, 335, 51 },
		{ 381, 336, 51 },
		{ 381, 337, 51 },
		{ 381, 338, 51 },
		{ 381, 339, 51 },
		{ 381, 340, 51 },
		{ 381, 341, 51 },
		{ 381, 342, 51 },
		{ 381, 343, 51 },
		{ 381, 344, 51 },
		{ 381, 345, 51 },
		{ 381, 346, 51 },
		{ 381, 347, 51 },
		{ 381, 348, 51 },
		{ 381, 349, 51 },
		{ 382, 320, 51 },
		{ 382, 321, 51 },
		{ 382, 322, 51 },
		{ 382, 323, 51 },
		{ 382, 324, 51 },
		{ 382, 325, 51 },
		{ 382, 326, 51 },
		{ 382, 327, 51 },
		{ 382, 328, 51 },
		{ 382, 329, 51 },
		{ 382, 330, 51 },
		{ 382, 331, 51 },
		{ 382, 332, 51 },
		{ 382, 333, 51 },
		{ 382, 334, 51 },
		{ 382, 335, 51 },
		{ 382, 336, 51 },
		{ 382, 337, 51 },
		{ 382, 338, 51 },
		{ 382, 339, 51 },
		{ 382, 340, 51 },
		{ 382, 341, 51 },
		{ 382, 342, 51 },
		{ 382, 343, 51 },
		{ 382, 344, 51 },
		{ 382, 345, 51 },
		{ 382, 346, 51 },
		{ 382, 347, 51 },
		{ 382, 348, 51 },
		{ 382, 349, 51 },
		{ 382, 350, 51 },
		{ 383, 321, 51 },
		{ 383, 322, 51 },
		{ 383, 323, 51 },
		{ 383, 324, 51 },
		{ 383, 325, 51 },
		{ 383, 326, 51 },
		{ 383, 327, 51 },
		{ 383, 328, 51 },
		{ 383, 329, 51 },
		{ 383, 330, 51 },
		{ 383, 331, 51 },
		{ 383, 332, 51 },
		{ 383, 333, 51 },
		{ 383, 334, 51 },
		{ 383, 335, 51 },
		{ 383, 336, 51 },
		{ 383, 337, 51 },
		{ 383, 338, 51 },
		{ 383, 339, 51 },
		{ 383, 340, 51 },
		{ 383, 341, 51 },
		{ 383, 342, 51 },
		{ 383, 343, 51 },
		{ 383, 344, 51 },
		{ 383, 345, 51 },
		{ 383, 346, 51 },
		{ 383, 347, 51 },
		{ 383, 348, 51 },
		{ 383, 349, 51 },
		{ 383, 350, 51 },
		{ 383, 351, 51 },
		{ 384, 321, 51 },
		{ 384, 322, 51 },
		{ 384, 323, 51 },
		{ 384, 324, 51 },
		{ 384, 325, 51 },
		{ 384, 326, 51 },
		{ 384, 327, 51 },
		{ 384, 328, 51 },
		{ 384, 329, 51 },
		{ 384, 330, 51 },
		{ 384, 331, 51 },
		{ 384, 332, 51 },
		{ 384, 333, 51 },
		{ 384, 334, 51 },
		{ 384, 335, 51 },
		{ 384, 336, 51 },
		{ 384, 337, 51 },
		{ 384, 338, 51 },
		{ 384, 339, 51 },
		{ 384, 340, 51 },
		{ 384, 341, 51 },
		{ 384, 342, 51 },
		{ 384, 343, 51 },
		{ 384, 344, 51 },
		{ 384, 345, 51 },
		{ 384, 346, 51 },
		{ 384, 347, 51 },
		{ 384, 348, 51 },
		{ 384, 349, 51 },
		{ 384, 350, 51 },
		{ 384, 351, 51 },
		{ 384, 352, 51 },
		{ 385, 322, 51 },
		{ 385, 323, 51 },
		{ 385, 324, 51 },
		{ 385, 325, 51 },
		{ 385, 326, 51 },
		{ 385, 327, 51 },
		{ 385, 328, 51 },
		{ 385, 329, 51 },
		{ 385, 330, 51 },
		{ 385, 331, 51 },
		{ 385, 332, 51 },
		{ 385, 333, 51 },
		{ 385, 334, 51 },
		{ 385, 335, 51 },
		{ 385, 336, 51 },
		{ 385, 337, 51 },
		{ 385, 338, 51 },
		{ 385, 339, 51 },
		{ 385, 340, 51 },
		{ 385, 341, 51 },
		{ 385, 342, 51 },
		{ 385, 343, 51 },
		{ 385, 344, 51 },
		{ 385, 345, 51 },
		{ 385, 346, 51 },
		{ 385, 347, 51 },
		{ 385, 348, 51 },
		{ 385, 349, 51 },
		{ 385, 350, 51 },
		{ 385, 351, 51 },
		{ 385, 352, 51 },
		{ 386, 322, 51 },
		{ 386, 323, 51 },
		{ 386, 324, 51 },
		{ 386, 325, 51 },
		{ 386, 326, 51 },
		{ 386, 327, 51 },
		{ 386, 328, 51 },
		{ 386, 329, 51 },
		{ 386, 330, 51 },
		{ 386, 331, 51 },
		{ 386, 332, 51 },
		{ 386, 333, 51 },
		{ 386, 334, 51 },
		{ 386, 335, 51 },
		{ 386, 336, 51 },
		{ 386, 337, 51 },
		{ 386, 338, 51 },
		{ 386, 339, 51 },
		{ 386, 340, 51 },
		{ 386, 341, 51 },
		{ 386, 342, 51 },
		{ 386, 343, 51 },
		{ 386, 344, 51 },
		{ 386, 345, 51 },
		{ 386, 346, 51 },
		{ 386, 347, 51 },
		{ 386, 348, 51 },
		{ 386, 349, 51 },
		{ 386, 350, 51 },
		{ 386, 351, 51 },
		{ 386, 352, 51 },
		{ 386, 353, 51 },
		{ 387, 323, 51 },
		{ 387, 324, 51 },
		{ 387, 325, 51 },
		{ 387, 326, 51 },
		{ 387, 327, 51 },
		{ 387, 328, 51 },
		{ 387, 329, 51 },
		{ 387, 330, 51 },
		{ 387, 331, 51 },
		{ 387, 332, 51 },
		{ 387, 333, 51 },
		{ 387, 334, 51 },
		{ 387, 335, 51 },
		{ 387, 336, 51 },
		{ 387, 337, 51 },
		{ 387, 338, 51 },
		{ 387, 339, 51 },
		{ 387, 340, 51 },
		{ 387, 341, 51 },
		{ 387, 342, 51 },
		{ 387, 343, 51 },
		{ 387, 344, 51 },
		{ 387, 345, 51 },
		{ 387, 346, 51 },
		{ 387, 347, 51 },
		{ 387, 348, 51 },
		{ 387, 349, 51 },
		{ 387, 350, 51 },
		{ 387, 351, 51 },
		{ 387, 352, 51 },
		{ 387, 353, 51 },
		{ 387, 354, 51 },
		{ 388, 323, 51 },
		{ 388, 324, 51 },
		{ 388, 325, 51 },
		{ 388, 326, 51 },
		{ 388, 327, 51 },
		{ 388, 328, 51 },
		{ 388, 329, 51 },
		{ 388, 330, 51 },
		{ 388, 331, 51 },
		{ 388, 332, 51 },
		{ 388, 333, 51 },
		{ 388, 334, 51 },
		{ 388, 335, 51 },
		{ 388, 336, 51 },
		{ 388, 337, 51 },
		{ 388, 338, 51 },
		{ 388, 339, 51 },
		{ 388, 340, 51 },
		{ 388, 341, 51 },
		{ 388, 342, 51 },
		{ 388, 343, 51 },
		{ 388, 344, 51 },
		{ 388, 345, 51 },
		{ 388, 346, 51 },
		{ 388, 347, 51 },
		{ 388, 348, 51 },
		{ 388, 349, 51 },
		{ 388, 350, 51 },
		{ 388, 351, 51 },
		{ 388, 352, 51 },
		{ 388, 353, 51 },
		{ 388, 354, 51 },
		{ 388, 355, 51 },
		{ 389, 324, 51 },
		{ 389, 325, 51 },
		{ 389, 326, 51 },
		{ 389, 327, 51 },
		{ 389, 328, 51 },
		{ 389, 329, 51 },
		{ 389, 330, 51 },
		{ 389, 331, 51 },
		{ 389, 332, 51 },
		{ 389, 333, 51 },
		{ 389, 334, 51 },
		{ 389, 335, 51 },
		{ 389, 336, 51 },
		{ 389, 337, 51 },
		{ 389, 338, 51 },
		{ 389, 339, 51 },
		{ 389, 340, 51 },
		{ 389, 341, 51 },
		{ 389, 342, 51 },
		{ 389, 343, 51 },
		{ 389, 344, 51 },
		{ 389, 345, 51 },
		{ 389, 346, 51 },
		{ 389, 347, 51 },
		{ 389, 348, 51 },
		{ 389, 349, 51 },
		{ 389, 350, 51 },
		{ 389, 351, 51 },
		{ 389, 352, 51 },
		{ 389, 353, 51 },
		{ 389, 354, 51 },
		{ 389, 355, 51 },
		{ 390, 324, 51 },
		{ 390, 325, 51 },
		{ 390, 326, 51 },
		{ 390, 327, 51 },
		{ 390, 328, 51 },
		{ 390, 329, 51 },
		{ 390, 330, 51 },
		{ 390, 331, 51 },
		{ 390, 332, 51 },
		{ 390, 333, 51 },
		{ 390, 334, 51 },
		{ 390, 335, 51 },
		{ 390, 336, 51 },
		{ 390, 337, 51 },
		{ 390, 338, 51 },
		{ 390, 339, 51 },
		{ 390, 340, 51 },
		{ 390, 341, 51 },
		{ 390, 342, 51 },
		{ 390, 343, 51 },
		{ 390, 344, 51 },
		{ 390, 345, 51 },
		{ 390, 346, 51 },
		{ 390, 347, 51 },
		{ 390, 348, 51 },
		{ 390, 349, 51 },
		{ 390, 350, 51 },
		{ 390, 351, 51 },
		{ 390, 352, 51 },
		{ 390, 353, 51 },
		{ 390, 354, 51 },
		{ 390, 355, 51 },
		{ 390, 356, 51 },
		{ 391, 325, 51 },
		{ 391, 326, 51 },
		{ 391, 327, 51 },
		{ 391, 328, 51 },
		{ 391, 329, 51 },
		{ 391, 330, 51 },
		{ 391, 331, 51 },
		{ 391, 332, 51 },
		{ 391, 333, 51 },
		{ 391, 334, 51 },
		{ 391, 335, 51 },
		{ 391, 336, 51 },
		{ 391, 337, 51 },
		{ 391, 338, 51 },
		{ 391, 339, 51 },
		{ 391, 340, 51 },
		{ 391, 341, 51 },
		{ 391, 342, 51 },
		{ 391, 343, 51 },
		{ 391, 344, 51 },
		{ 391, 345, 51 },
		{ 391, 346, 51 },
		{ 391, 347, 51 },
		{ 391, 348, 51 },
		{ 391, 349, 51 },
		{ 391, 350, 51 },
		{ 391, 351, 51 },
		{ 391, 352, 51 },
		{ 391, 353, 51 },
		{ 391, 354, 51 },
		{ 391, 355, 51 },
		{ 391, 356, 51 },
		{ 391, 357, 51 },
		{ 392, 325, 51 },
		{ 392, 326, 51 },
		{ 392, 327, 51 },
		{ 392, 328, 51 },
		{ 392, 329, 51 },
		{ 392, 330, 51 },
		{ 392, 331, 51 },
		{ 392, 332, 51 },
		{ 392, 333, 51 },
		{ 392, 334, 51 },
		{ 392, 335, 51 },
		{ 392, 336, 51 },
		{ 392, 337, 51 },
		{ 392, 338, 51 },
		{ 392, 339, 51 },
		{ 392, 340, 51 },
		{ 392, 341, 51 },
		{ 392, 342, 51 },
		{ 392, 343, 51 },
		{ 392, 344, 51 },
		{ 392, 345, 51 },
		{ 392, 346, 51 },
		{ 392, 347, 51 },
		{ 392, 348, 51 },
		{ 392, 349, 51 },
		{ 392, 350, 51 },
		{ 392, 351, 51 },
		{ 392, 352, 51 },
		{ 392, 353, 51 },
		{ 392, 354, 51 },
		{ 392, 355, 51 },
		{ 392, 356, 51 },
		{ 392, 357, 51 },
		{ 392, 358, 51 },
		{ 393, 326, 51 },
		{ 393, 327, 51 },
		{ 393, 328, 51 },
		{ 393, 329, 51 },
		{ 393, 330, 51 },
		{ 393, 331, 51 },
		{ 393, 332, 51 },
		{ 393, 333, 51 },
		{ 393, 334, 51 },
		{ 393, 335, 51 },
		{ 393, 336, 51 },
		{ 393, 337, 51 },
		{ 393, 338, 51 },
		{ 393, 339, 51 },
		{ 393, 340, 51 },
		{ 393, 341, 51 },
		{ 393, 342, 51 },
		{ 393, 343, 51 },
		{ 393, 344, 51 },
		{ 393, 345, 51 },
		{ 393, 346, 51 },
		{ 393, 347, 51 },
		{ 393, 348, 51 },
		{ 393, 349, 51 },
		{ 393, 350, 51 },
		{ 393, 351, 51 },
		{ 393, 352, 51 },
		{ 393, 353, 51 },
		{ 393, 354, 51 },
		{ 393, 355, 51 },
		{ 393, 356, 51 },
		{ 393, 357, 51 },
		{ 393, 358, 51 },
		{ 394, 326, 51 },
		{ 394, 327, 51 },
		{ 394, 328, 51 },
		{ 394, 329, 51 },
		{ 394, 330, 51 },
		{ 394, 331, 51 },
		{ 394, 332, 51 },
		{ 394, 333, 51 },
		{ 394, 334, 51 },
		{ 394, 335, 51 },
		{ 394, 336, 51 },
		{ 394, 337, 51 },
		{ 394, 338, 51 },
		{ 394, 339, 51 },
		{ 394, 340, 51 },
		{ 394, 341, 51 },
		{ 394, 342, 51 },
		{ 394, 343, 51 },
		{ 394, 344, 51 },
		{ 394, 345, 51 },
		{ 394, 346, 51 },
		{ 394, 347, 51 },
		{ 394, 348, 51 },
		{ 394, 349, 51 },
		{ 394, 350, 51 },
		{ 394, 351, 51 },
		{ 394, 352, 51 },
		{ 394, 353, 51 },
		{ 394, 354, 51 },
		{ 394, 355, 51 },
		{ 394, 356, 51 },
		{ 394, 357, 51 },
		{ 394, 358, 51 },
		{ 394, 359, 51 },
		{ 395, 327, 51 },
		{ 395, 328, 51 },
		{ 395, 329, 51 },
		{ 395, 330, 51 },
		{ 395, 331, 51 },
		{ 395, 332, 51 },
		{ 395, 333, 51 },
		{ 395, 334, 51 },
		{ 395, 335, 51 },
		{ 395, 336, 51 },
		{ 395, 337, 51 },
		{ 395, 338, 51 },
		{ 395, 339, 51 },
		{ 395, 340, 51 },
		{ 395, 341, 51 },
		{ 395, 342, 51 },
		{ 395, 343, 51 },
		{ 395, 344, 51 },
		{ 395, 345, 51 },
		{ 395, 346, 51 },
		{ 395, 347, 51 },
		{ 395, 348, 51 },
		{ 395, 349, 51 },
		{ 395, 350, 51 },
		{ 395, 351, 51 },
		{ 395, 352, 51 },
		{ 395, 353, 51 },
		{ 395, 354, 51 },
		{ 395, 355, 51 },
		{ 395, 356, 51 },
		{ 395, 357, 51 },
		{ 395, 358, 51 },
		{ 395, 359, 51 },
		{ 395, 360, 51 },
		{ 396, 327, 51 },
		{ 396, 328, 51 },
		{ 396, 329, 51 },
		{ 396, 330, 51 },
		{ 396, 331, 51 },
		{ 396, 332, 51 },
		{ 396, 333, 51 },
		{ 396, 334, 51 },
		{ 396, 335, 51 },
		{ 396, 336, 51 },
		{ 396, 337, 51 },
		{ 396, 338, 51 },
		{ 396, 339, 51 },
		{ 396, 340, 51 },
		{ 396, 341, 51 },
		{ 396, 342, 51 },
		{ 396, 343, 51 },
		{ 396, 344, 51 },
		{ 396, 345, 51 },
		{ 396, 346, 51 },
		{ 396, 347, 51 },
		{ 396, 348, 51 },
		{ 396, 349, 51 },
		{ 396, 350, 51 },
		{ 396, 351, 51 },
		{ 396, 352, 51 },
		{ 396, 353, 51 },
		{ 396, 354, 51 },
		{ 396, 355, 51 },
		{ 396, 356, 51 },
		{ 396, 357, 51 },
		{ 396, 358, 51 },
		{ 396, 359, 51 },
		{ 396, 360, 51 },
		{ 396, 361, 51 },
		{ 397, 328, 51 },
		{ 397, 329, 51 },
		{ 397, 330, 51 },
		{ 397, 331, 51 },
		{ 397, 332, 51 },
		{ 397, 333, 51 },
		{ 397, 334, 51 },
		{ 397, 335, 51 },
		{ 397, 336, 51 },
		{ 397, 337, 51 },
		{ 397, 338, 51 },
		{ 397, 339, 51 },
		{ 397, 340, 51 },
		{ 397, 341, 51 },
		{ 397, 342, 51 },
		{ 397, 343, 51 },
		{ 397, 344, 51 },
		{ 397, 345, 51 },
		{ 397, 346, 51 },
		{ 397, 347, 51 },
		{ 397, 348, 51 },
		{ 397, 349, 51 },
		{ 397, 350, 51 },
		{ 397, 351, 51 },
		{ 397, 352, 51 },
		{ 397, 353, 51 },
		{ 397, 354, 51 },
		{ 397, 355, 51 },
		{ 397, 356, 51 },
		{ 397, 357, 51 },
		{ 397, 358, 51 },
		{ 397, 359, 51 },
		{ 397, 360, 51 },
		{ 397, 361, 51 },
		{ 398, 328, 51 },
		{ 398, 329, 51 },
		{ 398, 330, 51 },
		{ 398, 331, 51 },
		{ 398, 332, 51 },
		{ 398, 333, 51 },
		{ 398, 334, 51 },
		{ 398, 335, 51 },
		{ 398, 336, 51 },
		{ 398, 337, 51 },
		{ 398, 338, 51 },
		{ 398, 339, 51 },
		{ 398, 340, 51 },
		{ 398, 341, 51 },
		{ 398, 342, 51 },
		{ 398, 343, 51 },
		{ 398, 344, 51 },
		{ 398, 345, 51 },
		{ 398, 346, 51 },
		{ 398, 347, 51 },
		{ 398, 348, 51 },
		{ 398, 349, 51 },
		{ 398, 350, 51 },
		{ 398, 351, 51 },
		{ 398, 352, 51 },
		{ 398, 353, 51 },
		{ 398, 354, 51 },
		{ 398, 355, 51 },
		{ 398, 356, 51 },
		{ 398, 357, 51 },
		{ 398, 358, 51 },
		{ 398, 359, 51 },
		{ 398, 360, 51 },
		{ 398, 361, 51 },
		{ 398, 362, 51 },
		{ 399, 329, 51 },
		{ 399, 330, 51 },
		{ 399, 331, 51 },
		{ 399, 332, 51 },
		{ 399, 333, 51 },
		{ 399, 334, 51 },
		{ 399, 335, 51 },
		{ 399, 336, 51 },
		{ 399, 337, 51 },
		{ 399, 338, 51 },
		{ 399, 339, 51 },
		{ 399, 340, 51 },
		{ 399, 341, 51 },
		{ 399, 342, 51 },
		{ 399, 343, 51 },
		{ 399, 344, 51 },
		{ 399, 345, 51 },
		{ 399, 346, 51 },
		{ 399, 347, 51 },
		{ 399, 348, 51 },
		{ 399, 349, 51 },
		{ 399, 350, 51 },
		{ 399, 351, 51 },
		{ 399, 352, 51 },
		{ 399, 353, 51 },
		{ 399, 354, 51 },
		{ 399, 355, 51 },
		{ 399, 356, 51 },
		{ 399, 357, 51 },
		{ 399, 358, 51 },
		{ 399, 359, 51 },
		{ 399, 360, 51 },
		{ 399, 361, 51 },
		{ 399, 362, 51 },
		{ 399, 363, 51 },
		{ 400, 329, 51 },
		{ 400, 330, 51 },
		{ 400, 331, 51 },
		{ 400, 332, 51 },
		{ 400, 333, 51 },
		{ 400, 334, 51 },
		{ 400, 335, 51 },
		{ 400, 336, 51 },
		{ 400, 337, 51 },
		{ 400, 338, 51 },
		{ 400, 339, 51 },
		{ 400, 340, 51 },
		{ 400, 341, 51 },
		{ 400, 342, 51 },
		{ 400, 343, 51 },
		{ 400, 344, 51 },
		{ 400, 345, 51 },
		{ 400, 346, 51 },
		{ 400, 347, 51 },
		{ 400, 348, 51 },
		{ 400, 349, 51 },
		{ 400, 350, 51 },
		{ 400, 351, 51 },
		{ 400, 352, 51 },
		{ 400, 353, 51 },
		{ 400, 354, 51 },
		{ 400, 355, 51 },
		{ 400, 356, 51 },
		{ 400, 357, 51 },
		{ 400, 358, 51 },
		{ 400, 359, 51 },
		{ 400, 360, 51 },
		{ 400, 361, 51 },
		{ 400, 362, 51 },
		{ 400, 363, 51 },
		{ 400, 364, 51 },
		{ 401, 330, 51 },
		{ 401, 331, 51 },
		{ 401, 332, 51 },
		{ 401, 333, 51 },
		{ 401, 334, 51 },
		{ 401, 335, 51 },
		{ 401, 336, 51 },
		{ 401, 337, 51 },
		{ 401, 338, 51 },
		{ 401, 339, 51 },
		{ 401, 340, 51 },
		{ 401, 341, 51 },
		{ 401, 342, 51 },
		{ 401, 343, 51 },
		{ 401, 344, 51 },
		{ 401, 345, 51 },
		{ 401, 346, 51 },
		{ 401, 347, 51 },
		{ 401, 348, 51 },
		{ 401, 349, 51 },
		{ 401, 350, 51 },
		{ 401, 351, 51 },
		{ 401, 352, 51 },
		{ 401, 353, 51 },
		{ 401, 354, 51 },
		{ 401, 355, 51 },
		{ 401, 356, 51 },
		{ 401, 357, 51 },
		{ 401, 358, 51 },
		{ 401, 359, 51 },
		{ 401, 360, 51 },
		{ 401, 361, 51 },
		{ 401, 362, 51 },
		{ 401, 363, 51 },
		{ 401, 364, 51 },
		{ 402, 330, 51 },
		{ 402, 331, 51 },
		{ 402, 332, 51 },
		{ 402, 333, 51 },
		{ 402, 334, 51 },
		{ 402, 335, 51 },
		{ 402, 336, 51 },
		{ 402, 337, 51 },
		{ 402, 338, 51 },
		{ 402, 339, 51 },
		{ 402, 340, 51 },
		{ 402, 341, 51 },
		{ 402, 342, 51 },
		{ 402, 343, 51 },
		{ 402, 344, 51 },
		{ 402, 345, 51 },
		{ 402, 346, 51 },
		{ 402, 347, 51 },
		{ 402, 348, 51 },
		{ 402, 349, 51 },
		{ 402, 350, 51 },
		{ 402, 351, 51 },
		{ 402, 352, 51 },
		{ 402, 353, 51 },
		{ 402, 354, 51 },
		{ 402, 355, 51 },
		{ 402, 356, 51 },
		{ 402, 357, 51 },
		{ 402, 358, 51 },
		{ 402, 359, 51 },
		{ 402, 360, 51 },
		{ 402, 361, 51 },
		{ 402, 362, 51 },
		{ 402, 363, 51 },
		{ 402, 364, 51 },
		{ 402, 365, 51 },
		{ 403, 331, 51 },
		{ 403, 332, 51 },
		{ 403, 333, 51 },
		{ 403, 334, 51 },
		{ 403, 335, 51 },
		{ 403, 336, 51 },
		{ 403, 337, 51 },
		{ 403, 338, 51 },
		{ 403, 339, 51 },
		{ 403, 340, 51 },
		{ 403, 341, 51 },
		{ 403, 342, 51 },
		{ 403, 343, 51 },
		{ 403, 344, 51 },
		{ 403, 345, 51 },
		{ 403, 346, 51 },
		{ 403, 347, 51 },
		{ 403, 348, 51 },
		{ 403, 349, 51 },
		{ 403, 350, 51 },
		{ 403, 351, 51 },
		{ 403, 352, 51 },
		{ 403, 353, 51 },
		{ 403, 354, 51 },
		{ 403, 355, 51 },
		{ 403, 356, 51 },
		{ 403, 357, 51 },
		{ 403, 358, 51 },
		{ 403, 359, 51 },
		{ 403, 360, 51 },
		{ 403, 361, 51 },
		{ 403, 362, 51 },
		{ 403, 363, 51 },
		{ 403, 364, 51 },
		{ 403, 365, 51 },
		{ 403, 366, 51 },
		{ 404, 331, 51 },
		{ 404, 332, 51 },
		{ 404, 333, 51 },
		{ 404, 334, 51 },
		{ 404, 335, 51 },
		{ 404, 336, 51 },
		{ 404, 337, 51 },
		{ 404, 338, 51 },
		{ 404, 339, 51 },
		{ 404, 340, 51 },
		{ 404, 341, 51 },
		{ 404, 342, 51 },
		{ 404, 343, 51 },
		{ 404, 344, 51 },
		{ 404, 345, 51 },
		{ 404, 346, 51 },
		{ 404, 347, 51 },
		{ 404, 348, 51 },
		{ 404, 349, 51 },
		{ 404, 350, 51 },
		{ 404, 351, 51 },
		{ 404, 352, 51 },
		{ 404, 353, 51 },
		{ 404, 354, 51 },
		{ 404, 355, 51 },
		{ 404, 356, 51 },
		{ 404, 357, 51 },
		{ 404, 358, 51 },
		{ 404, 359, 51 },
		{ 404, 360, 51 },
		{ 404, 361, 51 },
		{ 404, 362, 51 },
		{ 404, 363, 51 },
		{ 404, 364, 51 },
		{ 404, 365, 51 },
		{ 404, 366, 51 },
		{ 404, 367, 51 },
		{ 405, 332, 51 },
		{ 405, 333, 51 },
		{ 405, 334, 51 },
		{ 405, 335, 51 },
		{ 405, 336, 51 },
		{ 405, 337, 51 },
		{ 405, 338, 51 },
		{ 405, 339, 51 },
		{ 405, 340, 51 },
		{ 405, 341, 51 },
		{ 405, 342, 51 },
		{ 405, 343, 51 },
		{ 405, 344, 51 },
		{ 405, 345, 51 },
		{ 405, 346, 51 },
		{ 405, 347, 51 },
		{ 405, 348, 51 },
		{ 405, 349, 51 },
		{ 405, 350, 51 },
		{ 405, 351, 51 },
		{ 405, 352, 51 },
		{ 405, 353, 51 },
		{ 405, 354, 51 },
		{ 405, 355, 51 },
		{ 405, 356, 51 },
		{ 405, 357, 51 },
		{ 405, 358, 51 },
		{ 405, 359, 51 },
		{ 405, 360, 51 },
		{ 405, 361, 51 },
		{ 405, 362, 51 },
		{ 405, 363, 51 },
		{ 405, 364, 51 },
		{ 405, 365, 51 },
		{ 405, 366, 51 },
		{ 405, 367, 51 },
		{ 406, 332, 51 },
		{ 406, 333, 51 },
		{ 406, 334, 51 },
		{ 406, 335, 51 },
		{ 406, 336, 51 },
		{ 406, 337, 51 },
		{ 406, 338, 51 },
		{ 406, 339, 51 },
		{ 406, 340, 51 },
		{ 406, 341, 51 },
		{ 406, 342, 51 },
		{ 406, 343, 51 },
		{ 406, 344, 51 },
		{ 406, 345, 51 },
		{ 406, 346, 51 },
		{ 406, 347, 51 },
		{ 406, 348, 51 },
		{ 406, 349, 51 },
		{ 406, 350, 51 },
		{ 406, 351, 51 },
		{ 406, 352, 51 },
		{ 406, 353, 51 },
		{ 406, 354, 51 },
		{ 406, 355, 51 },
		{ 406, 356, 51 },
		{ 406, 357, 51 },
		{ 406, 358, 51 },
		{ 406, 359, 51 },
		{ 406, 360, 51 },
		{ 406, 361, 51 },
		{ 406, 362, 51 },
		{ 406, 363, 51 },
		{ 406, 364, 51 },
		{ 406, 365, 51 },
		{ 406, 366, 51 },
		{ 406, 367, 51 },
		{ 406, 368, 51 },
		{ 407, 333, 51 },
		{ 407, 334, 51 },
		{ 407, 335, 51 },
		{ 407, 336, 51 },
		{ 407, 337, 51 },
		{ 407, 338, 51 },
		{ 407, 339, 51 },
		{ 407, 340, 51 },
		{ 407, 341, 51 },
		{ 407, 342, 51 },
		{ 407, 343, 51 },
		{ 407, 344, 51 },
		{ 407, 345, 51 },
		{ 407, 346, 51 },
		{ 407, 347, 51 },
		{ 407, 348, 51 },
		{ 407, 349, 51 },
		{ 407, 350, 51 },
		{ 407, 351, 51 },
		{ 407, 352, 51 },
		{ 407, 353, 51 },
		{ 407, 354, 51 },
		{ 407, 355, 51 },
		{ 407, 356, 51 },
		{ 407, 357, 51 },
		{ 407, 358, 51 },
		{ 407, 359, 51 },
		{ 407, 360, 51 },
		{ 407, 361, 51 },
		{ 407, 362, 51 },
		{ 407, 363, 51 },
		{ 407, 364, 51 },
		{ 407, 365, 51 },
		{ 407, 366, 51 },
		{ 407, 367, 51 },
		{ 407, 368, 51 },
		{ 407, 369, 51 },
		{ 408, 333, 51 },
		{ 408, 334, 51 },
		{ 408, 335, 51 },
		{ 408, 336, 51 },
		{ 408, 337, 51 },
		{ 408, 338, 51 },
		{ 408, 339, 51 },
		{ 408, 340, 51 },
		{ 408, 341, 51 },
		{ 408, 342, 51 },
		{ 408, 343, 51 },
		{ 408, 344, 51 },
		{ 408, 345, 51 },
		{ 408, 346, 51 },
		{ 408, 347, 51 },
		{ 408, 348, 51 },
		{ 408, 349, 51 },
		{ 408, 350, 51 },
		{ 408, 351, 51 },
		{ 408, 352, 51 },
		{ 408, 353, 51 },
		{ 408, 354, 51 },
		{ 408, 355, 51 },
		{ 408, 356, 51 },
		{ 408, 357, 51 },
		{ 408, 358, 51 },
		{ 408, 359, 51 },
		{ 408, 360, 51 },
		{ 408, 361, 51 },
		{ 408, 362, 51 },
		{ 408, 363, 51 },
		{ 408, 364, 51 },
		{ 408, 365, 51 },
		{ 408, 366, 51 },
		{ 408, 367, 51 },
		{ 408, 368, 51 },
		{ 408, 369, 51 },
		{ 408, 370, 51 },
		{ 409, 334, 51 },
		{ 409, 335, 51 },
		{ 409, 336, 51 },
		{ 409, 337, 51 },
		{ 409, 338, 51 },
		{ 409, 339, 51 },
		{ 409, 340, 51 },
		{ 409, 341, 51 },
		{ 409, 342, 51 },
		{ 409, 343, 51 },
		{ 409, 344, 51 },
		{ 409, 345, 51 },
		{ 409, 346, 51 },
		{ 409, 347, 51 },
		{ 409, 348, 51 },
		{ 409, 349, 51 },
		{ 409, 350, 51 },
		{ 409, 351, 51 },
		{ 409, 352, 51 },
		{ 409, 353, 51 },
		{ 409, 354, 51 },
		{ 409, 355, 51 },
		{ 409, 356, 51 },
		{ 409, 357, 51 },
		{ 409, 358, 51 },
		{ 409, 359, 51 },
		{ 409, 360, 51 },
		{ 409, 361, 51 },
		{ 409, 362, 51 },
		{ 409, 363, 51 },
		{ 409, 364, 51 },
		{ 409, 365, 51 },
		{ 409, 366, 51 },
		{ 409, 367, 51 },
		{ 409, 368, 51 },
		{ 409, 369, 51 },
		{ 409, 370, 51 },
		{ 410, 334, 51 },
		{ 410, 335, 51 },
		{ 410, 336, 51 },
		{ 410, 337, 51 },
		{ 410, 338, 51 },
		{ 410, 339, 51 },
		{ 410, 340, 51 },
		{ 410, 341, 51 },
		{ 410, 342, 51 },
		{ 410, 343, 51 },
		{ 410, 344, 51 },
		{ 410, 345, 51 },
		{ 410, 346, 51 },
		{ 410, 347, 51 },
		{ 410, 348, 51 },
		{ 410, 349, 51 },
		{ 410, 350, 51 },
		{ 410, 351, 51 },
		{ 410, 352, 51 },
		{ 410, 353, 51 },
		{ 410, 354, 51 },
		{ 410, 355, 51 },
		{ 410, 356, 51 },
		{ 410, 357, 51 },
		{ 410, 358, 51 },
		{ 410, 359, 51 },
		{ 410, 360, 51 },
		{ 410, 361, 51 },
		{ 410, 362, 51 },
		{ 410, 363, 51 },
		{ 410, 364, 51 },
		{ 410, 365, 51 },
		{ 410, 366, 51 },
		{ 410, 367, 51 },
		{ 410, 368, 51 },
		{ 410, 369, 51 },
		{ 410, 370, 51 },
		{ 410, 371, 51 },
		{ 411, 335, 51 },
		{ 411, 336, 51 },
		{ 411, 337, 51 },
		{ 411, 338, 51 },
		{ 411, 339, 51 },
		{ 411, 340, 51 },
		{ 411, 341, 51 },
		{ 411, 342, 51 },
		{ 411, 343, 51 },
		{ 411, 344, 51 },
		{ 411, 345, 51 },
		{ 411, 346, 51 },
		{ 411, 347, 51 },
		{ 411, 348, 51 },
		{ 411, 349, 51 },
		{ 411, 350, 51 },
		{ 411, 351, 51 },
		{ 411, 352, 51 },
		{ 411, 353, 51 },
		{ 411, 354, 51 },
		{ 411, 355, 51 },
		{ 411, 356, 51 },
		{ 411, 357, 51 },
		{ 411, 358, 51 },
		{ 411, 359, 51 },
		{ 411, 360, 51 },
		{ 411, 361, 51 },
		{ 411, 362, 51 },
		{ 411, 363, 51 },
		{ 411, 364, 51 },
		{ 411, 365, 51 },
		{ 411, 366, 51 },
		{ 411, 367, 51 },
		{ 411, 368, 51 },
		{ 411, 369, 51 },
		{ 411, 370, 51 },
		{ 411, 371, 51 },
		{ 411, 372, 51 },
		{ 412, 335, 51 },
		{ 412, 336, 51 },
		{ 412, 337, 51 },
		{ 412, 338, 51 },
		{ 412, 339, 51 },
		{ 412, 340, 51 },
		{ 412, 341, 51 },
		{ 412, 342, 51 },
		{ 412, 343, 51 },
		{ 412, 344, 51 },
		{ 412, 345, 51 },
		{ 412, 346, 51 },
		{ 412, 347, 51 },
		{ 412, 348, 51 },
		{ 412, 349, 51 },
		{ 412, 350, 51 },
		{ 412, 351, 51 },
		{ 412, 352, 51 },
		{ 412, 353, 51 },
		{ 412, 354, 51 },
		{ 412, 355, 51 },
		{ 412, 356, 51 },
		{ 412, 357, 51 },
		{ 412, 358, 51 },
		{ 412, 359, 51 },
		{ 412, 360, 51 },
		{ 412, 361, 51 },
		{ 412, 362, 51 },
		{ 412, 363, 51 },
		{ 412, 364, 51 },
		{ 412, 365, 51 },
		{ 412, 366, 51 },
		{ 412, 367, 51 },
		{ 412, 368, 51 },
		{ 412, 369, 51 },
		{ 412, 370, 51 },
		{ 412, 371, 51 },
		{ 412, 372, 51 },
		{ 412, 373, 51 },
		{ 413, 336, 51 },
		{ 413, 337, 51 },
		{ 413, 338, 51 },
		{ 413, 339, 51 },
		{ 413, 340, 51 },
		{ 413, 341, 51 },
		{ 413, 342, 51 },
		{ 413, 343, 51 },
		{ 413, 344, 51 },
		{ 413, 345, 51 },
		{ 413, 346, 51 },
		{ 413, 347, 51 },
		{ 413, 348, 51 },
		{ 413, 349, 51 },
		{ 413, 350, 51 },
		{ 413, 351, 51 },
		{ 413, 352, 51 },
		{ 413, 353, 51 },
		{ 413, 354, 51 },
		{ 413, 355, 51 },
		{ 413, 356, 51 },
		{ 413, 357, 51 },
		{ 413, 358, 51 },
		{ 413, 359, 51 },
		{ 413, 360, 51 },
		{ 413, 361, 51 },
		{ 413, 362, 51 },
		{ 413, 363, 51 },
		{ 413, 364, 51 },
		{ 413, 365, 51 },
		{ 413, 366, 51 },
		{ 413, 367, 51 },
		{ 413, 368, 51 },
		{ 413, 369, 51 },
		{ 413, 370, 51 },
		{ 413, 371, 51 },
		{ 413, 372, 51 },
		{ 413, 373, 51 },
		{ 414, 336, 51 },
		{ 414, 337, 51 },
		{ 414, 338, 51 },
		{ 414, 339, 51 },
		{ 414, 340, 51 },
		{ 414, 341, 51 },
		{ 414, 342, 51 },
		{ 414, 343, 51 },
		{ 414, 344, 51 },
		{ 414, 345, 51 },
		{ 414, 346, 51 },
		{ 414, 347, 51 },
		{ 414, 348, 51 },
		{ 414, 349, 51 },
		{ 414, 350, 51 },
		{ 414, 351, 51 },
		{ 414, 352, 51 },
		{ 414, 353, 51 },
		{ 414, 354, 51 },
		{ 414, 355, 51 },
		{ 414, 356, 51 },
		{ 414, 357, 51 },
		{ 414, 358, 51 },
		{ 414, 359, 51 },
		{ 414, 360, 51 },
		{ 414, 361, 51 },
		{ 414, 362, 51 },
		{ 414, 363, 51 },
		{ 414, 364, 51 },
		{ 414, 365, 51 },
		{ 414, 366, 51 },
		{ 414, 367, 51 },
		{ 414, 368, 51 },
		{ 414, 369, 51 },
		{ 414, 370, 51 },
		{ 414, 371, 51 },
		{ 414, 372, 51 },
		{ 414, 373, 51 },
		{ 414, 374, 51 },
		{ 415, 337, 51 },
		{ 415, 338, 51 },
		{ 415, 339, 51 },
		{ 415, 340, 51 },
		{ 415, 341, 51 },
		{ 415, 342, 51 },
		{ 415, 343, 51 },
		{ 415, 344, 51 },
		{ 415, 345, 51 },
		{ 415, 346, 51 },
		{ 415, 347, 51 },
		{ 415, 348, 51 },
		{ 415, 349, 51 },
		{ 415, 350, 51 },
		{ 415, 351, 51 },
		{ 415, 352, 51 },
		{ 415, 353, 51 },
		{ 415, 354, 51 },
		{ 415, 355, 51 },
		{ 415, 356, 51 },
		{ 415, 357, 51 },
		{ 415, 358, 51 },
		{ 415, 359, 51 },
		{ 415, 360, 51 },
		{ 415, 361, 51 },
		{ 415, 362, 51 },
		{ 415, 363, 51 },
		{ 415, 364, 51 },
		{ 415, 365, 51 },
		{ 415, 366, 51 },
		{ 415, 367, 51 },
		{ 415, 368, 51 },
		{ 415, 369, 51 },
		{ 415, 370, 51 },
		{ 415, 371, 51 },
		{ 415, 372, 51 },
		{ 415, 373, 51 },
		{ 415, 374, 51 },
		{ 415, 375, 51 },
		{ 416, 337, 51 },
		{ 416, 338, 51 },
		{ 416, 339, 51 },
		{ 416, 340, 51 },
		{ 416, 341, 51 },
		{ 416, 342, 51 },
		{ 416, 343, 51 },
		{ 416, 344, 51 },
		{ 416, 345, 51 },
		{ 416, 346, 51 },
		{ 416, 347, 51 },
		{ 416, 348, 51 },
		{ 416, 349, 51 },
		{ 416, 350, 51 },
		{ 416, 351, 51 },
		{ 416, 352, 51 },
		{ 416, 353, 51 },
		{ 416, 354, 51 },
		{ 416, 355, 51 },
		{ 416, 356, 51 },
		{ 416, 357, 51 },
		{ 416, 358, 51 },
		{ 416, 359, 51 },
		{ 416, 360, 51 },
		{ 416, 361, 51 },
		{ 416, 362, 51 },
		{ 416, 363, 51 },
		{ 416, 364, 51 },
		{ 416, 365, 51 },
		{ 416, 366, 51 },
		{ 416, 367, 51 },
		{ 416, 368, 51 },
		{ 416, 369, 51 },
		{ 416, 370, 51 },
		{ 416, 371, 51 },
		{ 416, 372, 51 },
		{ 416, 373, 51 },
		{ 416, 374, 51 },
		{ 416, 375, 51 },
		{ 416, 376, 51 },
		{ 417, 338, 51 },
		{ 417, 339, 51 },
		{ 417, 340, 51 },
		{ 417, 341, 51 },
		{ 417, 342, 51 },
		{ 417, 343, 51 },
		{ 417, 344, 51 },
		{ 417, 345, 51 },
		{ 417, 346, 51 },
		{ 417, 347, 51 },
		{ 417, 348, 51 },
		{ 417, 349, 51 },
		{ 417, 350, 51 },
		{ 417, 351, 51 },
		{ 417, 352, 51 },
		{ 417, 353, 51 },
		{ 417, 354, 51 },
		{ 417, 355, 51 },
		{ 417, 356, 51 },
		{ 417, 357, 51 },
		{ 417, 358, 51 },
		{ 417, 359, 51 },
		{ 417, 360, 51 },
		{ 417, 361, 51 },
		{ 417, 362, 51 },
		{ 417, 363, 51 },
		{ 417, 364, 51 },
		{ 417, 365, 51 },
		{ 417, 366, 51 },
		{ 417, 367, 51 },
		{ 417, 368, 51 },
		{ 417, 369, 51 },
		{ 417, 370, 51 },
		{ 417, 371, 51 },
		{ 417, 372, 51 },
		{ 417, 373, 51 },
		{ 417, 374, 51 },
		{ 417, 375, 51 },
		{ 417, 376, 51 },
		{ 418, 338, 51 },
		{ 418, 339, 51 },
		{ 418, 340, 51 },
		{ 418, 341, 51 },
		{ 418, 342, 51 },
		{ 418, 343, 51 },
		{ 418, 344, 51 },
		{ 418, 345, 51 },
		{ 418, 346, 51 },
		{ 418, 347, 51 },
		{ 418, 348, 51 },
		{ 418, 349, 51 },
		{ 418, 350, 51 },
		{ 418, 351, 51 },
		{ 418, 352, 51 },
		{ 418, 353, 51 },
		{ 418, 354, 51 },
		{ 418, 355, 51 },
		{ 418, 356, 51 },
		{ 418, 357, 51 },
		{ 418, 358, 51 },
		{ 418, 359, 51 },
		{ 418, 360, 51 },
		{ 418, 361, 51 },
		{ 418, 362, 51 },
		{ 418, 363, 51 },
		{ 418, 364, 51 },
		{ 418, 365, 51 },
		{ 418, 366, 51 },
		{ 418, 367, 51 },
		{ 418, 368, 51 },
		{ 418, 369, 51 },
		{ 418, 370, 51 },
		{ 418, 371, 51 },
		{ 418, 372, 51 },
		{ 418, 373, 51 },
		{ 418, 374, 51 },
		{ 418, 375, 51 },
		{ 418, 376, 51 },
		{ 418, 377, 51 },
		{ 419, 339, 51 },
		{ 419, 340, 51 },
		{ 419, 341, 51 },
		{ 419, 342, 51 },
		{ 419, 343, 51 },
		{ 419, 344, 51 },
		{ 419, 345, 51 },
		{ 419, 346, 51 },
		{ 419, 347, 51 },
		{ 419, 348, 51 },
		{ 419, 349, 51 },
		{ 419, 350, 51 },
		{ 419, 351, 51 },
		{ 419, 352, 51 },
		{ 419, 353, 51 },
		{ 419, 354, 51 },
		{ 419, 355, 51 },
		{ 419, 356, 51 },
		{ 419, 357, 51 },
		{ 419, 358, 51 },
		{ 419, 359, 51 },
		{ 419, 360, 51 },
		{ 419, 361, 51 },
		{ 419, 362, 51 },
		{ 419, 363, 51 },
		{ 419, 364, 51 },
		{ 419, 365, 51 },
		{ 419, 366, 51 },
		{ 419, 367, 51 },
		{ 419, 368, 51 },
		{ 419, 369, 51 },
		{ 419, 370, 51 },
		{ 419, 371, 51 },
		{ 419, 372, 51 },
		{ 419, 373, 51 },
		{ 419, 374, 51 },
		{ 419, 375, 51 },
		{ 419, 376, 51 },
		{ 419, 377, 51 },
		{ 419, 378, 51 },
		{ 420, 339, 51 },
		{ 420, 340, 51 },
		{ 420, 341, 51 },
		{ 420, 342, 51 },
		{ 420, 343, 51 },
		{ 420, 344, 51 },
		{ 420, 345, 51 },
		{ 420, 346, 51 },
		{ 420, 347, 51 },
		{ 420, 348, 51 },
		{ 420, 349, 51 },
		{ 420, 350, 51 },
		{ 420, 351, 51 },
		{ 420, 352, 51 },
		{ 420, 353, 51 },
		{ 420, 354, 51 },
		{ 420, 355, 51 },
		{ 420, 356, 51 },
		{ 420, 357, 51 },
		{ 420, 358, 51 },
		{ 420, 359, 51 },
		{ 420, 360, 51 },
		{ 420, 361, 51 },
		{ 420, 362, 51 },
		{ 420, 363, 51 },
		{ 420, 364, 51 },
		{ 420, 365, 51 },
		{ 420, 366, 51 },
		{ 420, 367, 51 },
		{ 420, 368, 51 },
		{ 420, 369, 51 },
		{ 420, 370, 51 },
		{ 420, 371, 51 },
		{ 420, 372, 51 },
		{ 420, 373, 51 },
		{ 420, 374, 51 },
		{ 420, 375, 51 },
		{ 420, 376, 51 },
		{ 420, 377, 51 },
		{ 420, 378, 51 },
		{ 420, 379, 51 },
		{ 421, 340, 51 },
		{ 421, 341, 51 },
		{ 421, 342, 51 },
		{ 421, 343, 51 },
		{ 421, 344, 51 },
		{ 421, 345, 51 },
		{ 421, 346, 51 },
		{ 421, 347, 51 },
		{ 421, 348, 51 },
		{ 421, 349, 51 },
		{ 421, 350, 51 },
		{ 421, 351, 51 },
		{ 421, 352, 51 },
		{ 421, 353, 51 },
		{ 421, 354, 51 },
		{ 421, 355, 51 },
		{ 421, 356, 51 },
		{ 421, 357, 51 },
		{ 421, 358, 51 },
		{ 421, 359, 51 },
		{ 421, 360, 51 },
		{ 421, 361, 51 },
		{ 421, 362, 51 },
		{ 421, 363, 51 },
		{ 421, 364, 51 },
		{ 421, 365, 51 },
		{ 421, 366, 51 },
		{ 421, 367, 51 },
		{ 421, 368, 51 },
		{ 421, 369, 51 },
		{ 421, 370, 51 },
		{ 421, 371, 51 },
		{ 421, 372, 51 },
		{ 421, 373, 51 },
		{ 421, 374, 51 },
		{ 421, 375, 51 },
		{ 421, 376, 51 },
		{ 421, 377, 51 },
		{ 421, 378, 51 },
		{ 421, 379, 51 },
		{ 422, 340, 51 },
		{ 422, 341, 51 },
		{ 422, 342, 51 },
		{ 422, 343, 51 },
		{ 422, 344, 51 },
		{ 422, 345, 51 },
		{ 422, 346, 51 },
		{ 422, 347, 51 },
		{ 422, 348, 51 },
		{ 422, 349, 51 },
		{ 422, 350, 51 },
		{ 422, 351, 51 },
		{ 422, 352, 51 },
		{ 422, 353, 51 },
		{ 422, 354, 51 },
		{ 422, 355, 51 },
		{ 422, 356, 51 },
		{ 422, 357, 51 },
		{ 422, 358, 51 },
		{ 422, 359, 51 },
		{ 422, 360, 51 },
		{ 422, 361, 51 },
		{ 422, 362, 51 },
		{ 422, 363, 51 },
		{ 422, 364, 51 },
		{ 422, 365, 51 },
		{ 422, 366, 51 },
		{ 422, 367, 51 },
		{ 422, 368, 51 },
		{ 422, 369, 51 },
		{ 422, 370, 51 },
		{ 422, 371, 51 },
		{ 422, 372, 51 },
		{ 422, 373, 51 },
		{ 422, 374, 51 },
		{ 422, 375, 51 },
		{ 422, 376, 51 },
		{ 422, 377, 51 },
		{ 422, 378, 51 },
		{ 422, 379, 51 },
		{ 422, 380, 51 },
		{ 423, 341, 51 },
		{ 423, 342, 51 },
		{ 423, 343, 51 },
		{ 423, 344, 51 },
		{ 423, 345, 51 },
		{ 423, 346, 51 },
		{ 423, 347, 51 },
		{ 423, 348, 51 },
		{ 423, 349, 51 },
		{ 423, 350, 51 },
		{ 423, 351, 51 },
		{ 423, 352, 51 },
		{ 423, 353, 51 },
		{ 423, 354, 51 },
		{ 423, 355, 51 },
		{ 423, 356, 51 },
		{ 423, 357, 51 },
		{ 423, 358, 51 },
		{ 423, 359, 51 },
		{ 423, 360, 51 },
		{ 423, 361, 51 },
		{ 423, 362, 51 },
		{ 423, 363, 51 },
		{ 423, 364, 51 },
		{ 423, 365, 51 },
		{ 423, 366, 51 },
		{ 423, 367, 51 },
		{ 423, 368, 51 },
		{ 423, 369, 51 },
		{ 423, 370, 51 },
		{ 423, 371, 51 },
		{ 423, 372, 51 },
		{ 423, 373, 51 },
		{ 423, 374, 51 },
		{ 423, 375, 51 },
		{ 423, 376, 51 },
		{ 423, 377, 51 },
		{ 423, 378, 51 },
		{ 423, 379, 51 },
		{ 423, 380, 51 },
		{ 423, 381, 51 },
		{ 424, 341, 51 },
		{ 424, 342, 51 },
		{ 424, 343, 51 },
		{ 424, 344, 51 },
		{ 424, 345, 51 },
		{ 424, 346, 51 },
		{ 424, 347, 51 },
		{ 424, 348, 51 },
		{ 424, 349, 51 },
		{ 424, 350, 51 },
		{ 424, 351, 51 },
		{ 424, 352, 51 },
		{ 424, 353, 51 },
		{ 424, 354, 51 },
		{ 424, 355, 51 },
		{ 424, 356, 51 },
		{ 424, 357, 51 },
		{ 424, 358, 51 },
		{ 424, 359, 51 },
		{ 424, 360, 51 },
		{ 424, 361, 51 },
		{ 424, 362, 51 },
		{ 424, 363, 51 },
		{ 424, 364, 51 },
		{ 424, 365, 51 },
		{ 424, 366, 51 },
		{ 424, 367, 51 },
		{ 424, 368, 51 },
		{ 424, 369, 51 },
		{ 424, 370, 51 },
		{ 424, 371, 51 },
		{ 424, 372, 51 },
		{ 424, 373, 51 },
		{ 424, 374, 51 },
		{ 424, 375, 51 },
		{ 424, 376, 51 },
		{ 424, 377, 51 },
		{ 424, 378, 51 },
		{ 424, 379, 51 },
		{ 424, 380, 51 },
		{ 424, 381, 51 },
		{ 424, 382, 51 },
		{ 425, 342, 51 },
		{ 425, 343, 51 },
		{ 425, 344, 51 },
		{ 425, 345, 51 },
		{ 425, 346, 51 },
		{ 425, 347, 51 },
		{ 425, 348, 51 },
		{ 425, 349, 51 },
		{ 425, 350, 51 },
		{ 425, 351, 51 },
		{ 425, 352, 51 },
		{ 425, 353, 51 },
		{ 425, 354, 51 },
		{ 425, 355, 51 },
		{ 425, 356, 51 },
		{ 425, 357, 51 },
		{ 425, 358, 51 },
		{ 425, 359, 51 },
		{ 425, 360, 51 },
		{ 425, 361, 51 },
		{ 425, 362, 51 },
		{ 425, 363, 51 },
		{ 425, 364, 51 },
		{ 425, 365, 51 },
		{ 425, 366, 51 },
		{ 425, 367, 51 },
		{ 425, 368, 51 },
		{ 425, 369, 51 },
		{ 425, 370, 51 },
		{ 425, 371, 51 },
		{ 425, 372, 51 },
		{ 425, 373, 51 },
		{ 425, 374, 51 },
		{ 425, 375, 51 },
		{ 425, 376, 51 },
		{ 425, 377, 51 },
		{ 425, 378, 51 },
		{ 425, 379, 51 },
		{ 425, 380, 51 },
		{ 425, 381, 51 },
		{ 425, 382, 51 },
		{ 426, 342, 51 },
		{ 426, 343, 51 },
		{ 426, 344, 51 },
		{ 426, 345, 51 },
		{ 426, 346, 51 },
		{ 426, 347, 51 },
		{ 426, 348, 51 },
		{ 426, 349, 51 },
		{ 426, 350, 51 },
		{ 426, 351, 51 },
		{ 426, 352, 51 },
		{ 426, 353, 51 },
		{ 426, 354, 51 },
		{ 426, 355, 51 },
		{ 426, 356, 51 },
		{ 426, 357, 51 },
		{ 426, 358, 51 },
		{ 426, 359, 51 },
		{ 426, 360, 51 },
		{ 426, 361, 51 },
		{ 426, 362, 51 },
		{ 426, 363, 51 },
		{ 426, 364, 51 },
		{ 426, 365, 51 },
		{ 426, 366, 51 },
		{ 426, 367, 51 },
		{ 426, 368, 51 },
		{ 426, 369, 51 },
		{ 426, 370, 51 },
		{ 426, 371, 51 },
		{ 426, 372, 51 },
		{ 426, 373, 51 },
		{ 426, 374, 51 },
		{ 426, 375, 51 },
		{ 426, 376, 51 },
		{ 426, 377, 51 },
		{ 426, 378, 51 },
		{ 426, 379, 51 },
		{ 426, 380, 51 },
		{ 426, 381, 51 },
		{ 426, 382, 51 },
		{ 426, 383, 51 },
		{ 427, 343, 51 },
		{ 427, 344, 51 },
		{ 427, 345, 51 },
		{ 427, 346, 51 },
		{ 427, 347, 51 },
		{ 427, 348, 51 },
		{ 427, 349, 51 },
		{ 427, 350, 51 },
		{ 427, 351, 51 },
		{ 427, 352, 51 },
		{ 427, 353, 51 },
		{ 427, 354, 51 },
		{ 427, 355, 51 },
		{ 427, 356, 51 },
		{ 427, 357, 51 },
		{ 427, 358, 51 },
		{ 427, 359, 51 },
		{ 427, 360, 51 },
		{ 427, 361, 51 },
		{ 427, 362, 51 },
		{ 427, 363, 51 },
		{ 427, 364, 51 },
		{ 427, 365, 51 },
		{ 427, 366, 51 },
		{ 427, 367, 51 },
		{ 427, 368, 51 },
		{ 427, 369, 51 },
		{ 427, 370, 51 },
		{ 427, 371, 51 },
		{ 427, 372, 51 },
		{ 427, 373, 51 },
		{ 427, 374, 51 },
		{ 427, 375, 51 },
		{ 427, 376, 51 },
		{ 427, 377, 51 },
		{ 427, 378, 51 },
		{ 427, 379, 51 },
		{ 427, 380, 51 },
		{ 427, 381, 51 },
		{ 427, 382, 51 },
		{ 427, 383, 51 },
		{ 427, 384, 51 },
		{ 428, 343, 51 },
		{ 428, 344, 51 },
		{ 428, 345, 51 },
		{ 428, 346, 51 },
		{ 428, 347, 51 },
		{ 428, 348, 51 },
		{ 428, 349, 51 },
		{ 428, 350, 51 },
		{ 428, 351, 51 },
		{ 428, 352, 51 },
		{ 428, 353, 51 },
		{ 428, 354, 51 },
		{ 428, 355, 51 },
		{ 428, 356, 51 },
		{ 428, 357, 51 },
		{ 428, 358, 51 },
		{ 428, 359, 51 },
		{ 428, 360, 51 },
		{ 428, 361, 51 },
		{ 428, 362, 51 },
		{ 428, 363, 51 },
		{ 428, 364, 51 },
		{ 428, 365, 51 },
		{ 428, 366, 51 },
		{ 428, 367, 51 },
		{ 428, 368, 51 },
		{ 428, 369, 51 },
		{ 428, 370, 51 },
		{ 428, 371, 51 },
		{ 428, 372, 51 },
		{ 428, 373, 51 },
		{ 428, 374, 51 },
		{ 428, 375, 51 },
		{ 428, 376, 51 },
		{ 428, 377, 51 },
		{ 428, 378, 51 },
		{ 428, 379, 51 },
		{ 428, 380, 51 },
		{ 428, 381, 51 },
		{ 428, 382, 51 },
		{ 428, 383, 51 },
		{ 428, 384, 51 },
		{ 428, 385, 51 },
		{ 429, 344, 51 },
		{ 429, 345, 51 },
		{ 429, 346, 51 },
		{ 429, 347, 51 },
		{ 429, 348, 51 },
		{ 429, 349, 51 },
		{ 429, 350, 51 },
		{ 429, 351, 51 },
		{ 429, 352, 51 },
		{ 429, 353, 51 },
		{ 429, 354, 51 },
		{ 429, 355, 51 },
		{ 429, 356, 51 },
		{ 429, 357, 51 },
		{ 429, 358, 51 },
		{ 429, 359, 51 },
		{ 429, 360, 51 },
		{ 429, 361, 51 },
		{ 429, 362, 51 },
		{ 429, 363, 51 },
		{ 429, 364, 51 },
		{ 429, 365, 51 },
		{ 429, 366, 51 },
		{ 429, 367, 51 },
		{ 429, 368, 51 },
		{ 429, 369, 51 },
		{ 429, 370, 51 },
		{ 429, 371, 51 },
		{ 429, 372, 51 },
		{ 429, 373, 51 },
		{ 429, 374, 51 },
		{ 429, 375, 51 },
		{ 429, 376, 51 },
		{ 429, 377, 51 },
		{ 429, 378, 51 },
		{ 429, 379, 51 },
		{ 429, 380, 51 },
		{ 429, 381, 51 },
		{ 429, 382, 51 },
		{ 429, 383, 51 },
		{ 429, 384, 51 },
		{ 429, 385, 51 },
		{ 430, 344, 51 },
		{ 430, 345, 51 },
		{ 430, 346, 51 },
		{ 430, 347, 51 },
		{ 430, 348, 51 },
		{ 430, 349, 51 },
		{ 430, 350, 51 },
		{ 430, 351, 51 },
		{ 430, 352, 51 },
		{ 430, 353, 51 },
		{ 430, 354, 51 },
		{ 430, 355, 51 },
		{ 430, 356, 51 },
		{ 430, 357, 51 },
		{ 430, 358, 51 },
		{ 430, 359, 51 },
		{ 430, 360, 51 },
		{ 430, 361, 51 },
		{ 430, 362, 51 },
		{ 430, 363, 51 },
		{ 430, 364, 51 },
		{ 430, 365, 51 },
		{ 430, 366, 51 },
		{ 430, 367, 51 },
		{ 430, 368, 51 },
		{ 430, 369, 51 },
		{ 430, 370, 51 },
		{ 430, 371, 51 },
		{ 430, 372, 51 },
		{ 430, 373, 51 },
		{ 430, 374, 51 },
		{ 430, 375, 51 },
		{ 430, 376, 51 },
		{ 430, 377, 51 },
		{ 430, 378, 51 },
		{ 430, 379, 51 },
		{ 430, 380, 51 },
		{ 430, 381, 51 },
		{ 430, 382, 51 },
		{ 430, 383, 51 },
		{ 430, 384, 51 },
		{ 430, 385, 51 },
		{ 430, 386, 51 },
		{ 431, 345, 51 },
		{ 431, 346, 51 },
		{ 431, 347, 51 },
		{ 431, 348, 51 },
		{ 431, 349, 51 },
		{ 431, 350, 51 },
		{ 431, 351, 51 },
		{ 431, 352, 51 },
		{ 431, 353, 51 },
		{ 431, 354, 51 },
		{ 431, 355, 51 },
		{ 431, 356, 51 },
		{ 431, 357, 51 },
		{ 431, 358, 51 },
		{ 431, 359, 51 },
		{ 431, 360, 51 },
		{ 431, 361, 51 },
		{ 431, 362, 51 },
		{ 431, 363, 51 },
		{ 431, 364, 51 },
		{ 431, 365, 51 },
		{ 431, 366, 51 },
		{ 431, 367, 51 },
		{ 431, 368, 51 },
		{ 431, 369, 51 },
		{ 431, 370, 51 },
		{ 431, 371, 51 },
		{ 431, 372, 51 },
		{ 431, 373, 51 },
		{ 431, 374, 51 },
		{ 431, 375, 51 },
		{ 431, 376, 51 },
		{ 431, 377, 51 },
		{ 431, 378, 51 },
		{ 431, 379, 51 },
		{ 431, 380, 51 },
		{ 431, 381, 51 },
		{ 431, 382, 51 },
		{ 431, 383, 51 },
		{ 431, 384, 51 },
		{ 431, 385, 51 },
		{ 431, 386, 51 },
		{ 431, 387, 51 },
		{ 432, 345, 51 },
		{ 432, 346, 51 },
		{ 432, 347, 51 },
		{ 432, 348, 51 },
		{ 432, 349, 51 },
		{ 432, 350, 51 },
		{ 432, 351, 51 },
		{ 432, 352, 51 },
		{ 432, 353, 51 },
		{ 432, 354, 51 },
		{ 432, 355, 51 },
		{ 432, 356, 51 },
		{ 432, 357, 51 },
		{ 432, 358, 51 },
		{ 432, 359, 51 },
		{ 432, 360, 51 },
		{ 432, 361, 51 },
		{ 432, 362, 51 },
		{ 432, 363, 51 },
		{ 432, 364, 51 },
		{ 432, 365, 51 },
		{ 432, 366, 51 },
		{ 432, 367, 51 },
		{ 432, 368, 51 },
		{ 432, 369, 51 },
		{ 432, 370, 51 },
		{ 432, 371, 51 },
		{ 432, 372, 51 },
		{ 432, 373, 51 },
		{ 432, 374, 51 },
		{ 432, 375, 51 },
		{ 432, 376, 51 },
		{ 432, 377, 51 },
		{ 432, 378, 51 },
		{ 432, 379, 51 },
		{ 432, 380, 51 },
		{ 432, 381, 51 },
		{ 432, 382, 51 },
		{ 432, 383, 51 },
		{ 432, 384, 51 },
		{ 432, 385, 51 },
		{ 432, 386, 51 },
		{ 432, 387, 51 },
		{ 432, 388, 51 },
		{ 433, 346, 51 },
		{ 433, 347, 51 },
		{ 433, 348, 51 },
		{ 433, 349, 51 },
		{ 433, 350, 51 },
		{ 433, 351, 51 },
		{ 433, 352, 51 },
		{ 433, 353, 51 },
		{ 433, 354, 51 },
		{ 433, 355, 51 },
		{ 433, 356, 51 },
		{ 433, 357, 51 },
		{ 433, 358, 51 },
		{ 433, 359, 51 },
		{ 433, 360, 51 },
		{ 433, 361, 51 },
		{ 433, 362, 51 },
		{ 433, 363, 51 },
		{ 433, 364, 51 },
		{ 433, 365, 51 },
		{ 433, 366, 51 },
		{ 433, 367, 51 },
		{ 433, 368, 51 },
		{ 433, 369, 51 },
		{ 433, 370, 51 },
		{ 433, 371, 51 },
		{ 433, 372, 51 },
		{ 433, 373, 51 },
		{ 433, 374, 51 },
		{ 433, 375, 51 },
		{ 433, 376, 51 },
		{ 433, 377, 51 },
		{ 433, 378, 51 },
		{ 433, 379, 51 },
		{ 433, 380, 51 },
		{ 433, 381, 51 },
		{ 433, 382, 51 },
		{ 433, 383, 51 },
		{ 433, 384, 51 },
		{ 433, 385, 51 },
		{ 433, 386, 51 },
		{ 433, 387, 51 },
		{ 433, 388, 51 },
		{ 434, 346, 51 },
		{ 434, 347, 51 },
		{ 434, 348, 51 },
		{ 434, 349, 51 },
		{ 434, 350, 51 },
		{ 434, 351, 51 },
		{ 434, 352, 51 },
		{ 434, 353, 51 },
		{ 434, 354, 51 },
		{ 434, 355, 51 },
		{ 434, 356, 51 },
		{ 434, 357, 51 },
		{ 434, 358, 51 },
		{ 434, 359, 51 },
		{ 434, 360, 51 },
		{ 434, 361, 51 },
		{ 434, 362, 51 },
		{ 434, 363, 51 },
		{ 434, 364, 51 },
		{ 434, 365, 51 },
		{ 434, 366, 51 },
		{ 434, 367, 51 },
		{ 434, 368, 51 },
		{ 434, 369, 51 },
		{ 434, 370, 51 },
		{ 434, 371, 51 },
		{ 434, 372, 51 },
		{ 434, 373, 51 },
		{ 434, 374, 51 },
		{ 434, 375, 51 },
		{ 434, 376, 51 },
		{ 434, 377, 51 },
		{ 434, 378, 51 },
		{ 434, 379, 51 },
		{ 434, 380, 51 },
		{ 434, 381, 51 },
		{ 434, 382, 51 },
		{ 434, 383, 51 },
		{ 434, 384, 51 },
		{ 434, 385, 51 },
		{ 434, 386, 51 },
		{ 434, 387, 51 },
		{ 434, 388, 51 },
		{ 434, 389, 51 },
		{ 435, 347, 51 },
		{ 435, 348, 51 },
		{ 435, 349, 51 },
		{ 435, 350, 51 },
		{ 435, 351, 51 },
		{ 435, 352, 51 },
		{ 435, 353, 51 },
		{ 435, 354, 51 },
		{ 435, 355, 51 },
		{ 435, 356, 51 },
		{ 435, 357, 51 },
		{ 435, 358, 51 },
		{ 435, 359, 51 },
		{ 435, 360, 51 },
		{ 435, 361, 51 },
		{ 435, 362, 51 },
		{ 435, 363, 51 },
		{ 435, 364, 51 },
		{ 435, 365, 51 },
		{ 435, 366, 51 },
		{ 435, 367, 51 },
		{ 435, 368, 51 },
		{ 435, 369, 51 },
		{ 435, 370, 51 },
		{ 435, 371, 51 },
		{ 435, 372, 51 },
		{ 435, 373, 51 },
		{ 435, 374, 51 },
		{ 435, 375, 51 },
		{ 435, 376, 51 },
		{ 435, 377, 51 },
		{ 435, 378, 51 },
		{ 435, 379, 51 },
		{ 435, 380, 51 },
		{ 435, 381, 51 },
		{ 435, 382, 51 },
		{ 435, 383, 51 },
		{ 435, 384, 51 },
		{ 435, 385, 51 },
		{ 435, 386, 51 },
		{ 435, 387, 51 },
		{ 435, 388, 51 },
		{ 435, 389, 51 },
		{ 435, 390, 51 },
		{ 436, 347, 51 },
		{ 436, 348, 51 },
		{ 436, 349, 51 },
		{ 436, 350, 51 },
		{ 436, 351, 51 },
		{ 436, 352, 51 },
		{ 436, 353, 51 },
		{ 436, 354, 51 },
		{ 436, 355, 51 },
		{ 436, 356, 51 },
		{ 436, 357, 51 },
		{ 436, 358, 51 },
		{ 436, 359, 51 },
		{ 436, 360, 51 },
		{ 436, 361, 51 },
		{ 436, 362, 51 },
		{ 436, 363, 51 },
		{ 436, 364, 51 },
		{ 436, 365, 51 },
		{ 436, 366, 51 },
		{ 436, 367, 51 },
		{ 436, 368, 51 },
		{ 436, 369, 51 },
		{ 436, 370, 51 },
		{ 436, 371, 51 },
		{ 436, 372, 51 },
		{ 436, 373, 51 },
		{ 436, 374, 51 },
		{ 436, 375, 51 },
		{ 436, 376, 51 },
		{ 436, 377, 51 },
		{ 436, 378, 51 },
		{ 436, 379, 51 },
		{ 436, 380, 51 },
		{ 436, 381, 51 },
		{ 436, 382, 51 },
		{ 436, 383, 51 },
		{ 436, 384, 51 },
		{ 436, 385, 51 },
		{ 436, 386, 51 },
		{ 436, 387, 51 },
		{ 436, 388, 51 },
		{ 436, 389, 51 },
		{ 436, 390, 51 },
		{ 436, 391, 51 },
		{ 437, 348, 51 },
		{ 437, 349, 51 },
		{ 437, 350, 51 },
		{ 437, 351, 51 },
		{ 437, 352, 51 },
		{ 437, 353, 51 },
		{ 437, 354, 51 },
		{ 437, 355, 51 },
		{ 437, 356, 51 },
		{ 437, 357, 51 },
		{ 437, 358, 51 },
		{ 437, 359, 51 },
		{ 437, 360, 51 },
		{ 437, 361, 51 },
		{ 437, 362, 51 },
		{ 437, 363, 51 },
		{ 437, 364, 51 },
		{ 437, 365, 51 },
		{ 437, 366, 51 },
		{ 437, 367, 51 },
		{ 437, 368, 51 },
		{ 437, 369, 51 },
		{ 437, 370, 51 },
		{ 437, 371, 51 },
		{ 437, 372, 51 },
		{ 437, 373, 51 },
		{ 437, 374, 51 },
		{ 437, 375, 51 },
		{ 437, 376, 51 },
		{ 437, 377, 51 },
		{ 437, 378, 51 },
		{ 437, 379, 51 },
		{ 437, 380, 51 },
		{ 437, 381, 51 },
		{ 437, 382, 51 },
		{ 437, 383, 51 },
		{ 437, 384, 51 },
		{ 437, 385, 51 },
		{ 437, 386, 51 },
		{ 437, 387, 51 },
		{ 437, 388, 51 },
		{ 437, 389, 51 },
		{ 437, 390, 51 },
		{ 437, 391, 51 },
		{ 438, 348, 51 },
		{ 438, 349, 51 },
		{ 438, 350, 51 },
		{ 438, 351, 51 },
		{ 438, 352, 51 },
		{ 438, 353, 51 },
		{ 438, 354, 51 },
		{ 438, 355, 51 },
		{ 438, 356, 51 },
		{ 438, 357, 51 },
		{ 438, 358, 51 },
		{ 438, 359, 51 },
		{ 438, 360, 51 },
		{ 438, 361, 51 },
		{ 438, 362, 51 },
		{ 438, 363, 51 },
		{ 438, 364, 51 },
		{ 438, 365, 51 },
		{ 438, 366, 51 },
		{ 438, 367, 51 },
		{ 438, 368, 51 },
		{ 438, 369, 51 },
		{ 438, 370, 51 },
		{ 438, 371, 51 },
		{ 438, 372, 51 },
		{ 438, 373, 51 },
		{ 438, 374, 51 },
		{ 438, 375, 51 },
		{ 438, 376, 51 },
		{ 438, 377, 51 },
		{ 438, 378, 51 },
		{ 438, 379, 51 },
		{ 438, 380, 51 },
		{ 438, 381, 51 },
		{ 438, 382, 51 },
		{ 438, 383, 51 },
		{ 438, 384, 51 },
		{ 438, 385, 51 },
		{ 438, 386, 51 },
		{ 438, 387, 51 },
		{ 438, 388, 51 },
		{ 438, 389, 51 },
		{ 438, 390, 51 },
		{ 438, 391, 51 },
		{ 438, 392, 51 },
		{ 439, 349, 51 },
		{ 439, 350, 51 },
		{ 439, 351, 51 },
		{ 439, 352, 51 },
		{ 439, 353, 51 },
		{ 439, 354, 51 },
		{ 439, 355, 51 },
		{ 439, 356, 51 },
		{ 439, 357, 51 },
		{ 439, 358, 51 },
		{ 439, 359, 51 },
		{ 439, 360, 51 },
		{ 439, 361, 51 },
		{ 439, 362, 51 },
		{ 439, 363, 51 },
		{ 439, 364, 51 },
		{ 439, 365, 51 },
		{ 439, 366, 51 },
		{ 439, 367, 51 },
		{ 439, 368, 51 },
		{ 439, 369, 51 },
		{ 439, 370, 51 },
		{ 439, 371, 51 },
		{ 439, 372, 51 },
		{ 439, 373, 51 },
		{ 439, 374, 51 },
		{ 439, 375, 51 },
		{ 439, 376, 51 },
		{ 439, 377, 51 },
		{ 439, 378, 51 },
		{ 439, 379, 51 },
		{ 439, 380, 51 },
		{ 439, 381, 51 },
		{ 439, 382, 51 },
		{ 439, 383, 51 },
		{ 439, 384, 51 },
		{ 439, 385, 51 },
		{ 439, 386, 51 },
		{ 439, 387, 51 },
		{ 439, 388, 51 },
		{ 439, 389, 51 },
		{ 439, 390, 51 },
		{ 439, 391, 51 },
		{ 439, 392, 51 },
		{ 439, 393, 51 },
		{ 440, 349, 51 },
		{ 440, 350, 51 },
		{ 440, 351, 51 },
		{ 440, 352, 51 },
		{ 440, 353, 51 },
		{ 440, 354, 51 },
		{ 440, 355, 51 },
		{ 440, 356, 51 },
		{ 440, 357, 51 },
		{ 440, 358, 51 },
		{ 440, 359, 51 },
		{ 440, 360, 51 },
		{ 440, 361, 51 },
		{ 440, 362, 51 },
		{ 440, 363, 51 },
		{ 440, 364, 51 },
		{ 440, 365, 51 },
		{ 440, 366, 51 },
		{ 440, 367, 51 },
		{ 440, 368, 51 },
		{ 440, 369, 51 },
		{ 440, 370, 51 },
		{ 440, 371, 51 },
		{ 440, 372, 51 },
		{ 440, 373, 51 },
		{ 440, 374, 51 },
		{ 440, 375, 51 },
		{ 440, 376, 51 },
		{ 440, 377, 51 },
		{ 440, 378, 51 },
		{ 440, 379, 51 },
		{ 440, 380, 51 },
		{ 440, 381, 51 },
		{ 440, 382, 51 },
		{ 440, 383, 51 },
		{ 440, 384, 51 },
		{ 440, 385, 51 },
		{ 440, 386, 51 },
		{ 440, 387, 51 },
		{ 440, 388, 51 },
		{ 440, 389, 51 },
		{ 440, 390, 51 },
		{ 440, 391, 51 },
		{ 440, 392, 51 },
		{ 440, 393, 51 },
		{ 440, 394, 51 },
		{ 441, 350, 51 },
		{ 441, 351, 51 },
		{ 441, 352, 51 },
		{ 441, 353, 51 },
		{ 441, 354, 51 },
		{ 441, 355, 51 },
		{ 441, 356, 51 },
		{ 441, 357, 51 },
		{ 441, 358, 51 },
		{ 441, 359, 51 },
		{ 441, 360, 51 },
		{ 441, 361, 51 },
		{ 441, 362, 51 },
		{ 441, 363, 51 },
		{ 441, 364, 51 },
		{ 441, 365, 51 },
		{ 441, 366, 51 },
		{ 441, 367, 51 },
		{ 441, 368, 51 },
		{ 441, 369, 51 },
		{ 441, 370, 51 },
		{ 441, 371, 51 },
		{ 441, 372, 51 },
		{ 441, 373, 51 },
		{ 441, 374, 51 },
		{ 441, 375, 51 },
		{ 441, 376, 51 },
		{ 441, 377, 51 },
		{ 441, 378, 51 },
		{ 441, 379, 51 },
		{ 441, 380, 51 },
		{ 441, 381, 51 },
		{ 441, 382, 51 },
		{ 441, 383, 51 },
		{ 441, 384, 51 },
		{ 441, 385, 51 },
		{ 441, 386, 51 },
		{ 441, 387, 51 },
		{ 441, 388, 51 },
		{ 441, 389, 51 },
		{ 441, 390, 51 },
		{ 441, 391, 51 },
		{ 441, 392, 51 },
		{ 441, 393, 51 },
		{ 441, 394, 51 },
		{ 442, 350, 51 },
		{ 442, 351, 51 },
		{ 442, 352, 51 },
		{ 442, 353, 51 },
		{ 442, 354, 51 },
		{ 442, 355, 51 },
		{ 442, 356, 51 },
		{ 442, 357, 51 },
		{ 442, 358, 51 },
		{ 442, 359, 51 },
		{ 442, 360, 51 },
		{ 442, 361, 51 },
		{ 442, 362, 51 },
		{ 442, 363, 51 },
		{ 442, 364, 51 },
		{ 442, 365, 51 },
		{ 442, 366, 51 },
		{ 442, 367, 51 },
		{ 442, 368, 51 },
		{ 442, 369, 51 },
		{ 442, 370, 51 },
		{ 442, 371, 51 },
		{ 442, 372, 51 },
		{ 442, 373, 51 },
		{ 442, 374, 51 },
		{ 442, 375, 51 },
		{ 442, 376, 51 },
		{ 442, 377, 51 },
		{ 442, 378, 51 },
		{ 442, 379, 51 },
		{ 442, 380, 51 },
		{ 442, 381, 51 },
		{ 442, 382, 51 },
		{ 442, 383, 51 },
		{ 442, 384, 51 },
		{ 442, 385, 51 },
		{ 442, 386, 51 },
		{ 442, 387, 51 },
		{ 442, 388, 51 },
		{ 442, 389, 51 },
		{ 442, 390, 51 },
		{ 442, 391, 51 },
		{ 442, 392, 51 },
		{ 442, 393, 51 },
		{ 442, 394, 51 },
		{ 442, 395, 51 },
		{ 443, 351, 51 },
		{ 443, 352, 51 },
		{ 443, 353, 51 },
		{ 443, 354, 51 },
		{ 443, 355, 51 },
		{ 443, 356, 51 },
		{ 443, 357, 51 },
		{ 443, 358, 51 },
		{ 443, 359, 51 },
		{ 443, 360, 51 },
		{ 443, 361, 51 },
		{ 443, 362, 51 },
		{ 443, 363, 51 },
		{ 443, 364, 51 },
		{ 443, 365, 51 },
		{ 443, 366, 51 },
		{ 443, 367, 51 },
		{ 443, 368, 51 },
		{ 443, 369, 51 },
		{ 443, 370, 51 },
		{ 443, 371, 51 },
		{ 443, 372, 51 },
		{ 443, 373, 51 },
		{ 443, 374, 51 },
		{ 443, 375, 51 },
		{ 443, 376, 51 },
		{ 443, 377, 51 },
		{ 443, 378, 51 },
		{ 443, 379, 51 },
		{ 443, 380, 51 },
		{ 443, 381, 51 },
		{ 443, 382, 51 },
		{ 443, 383, 51 },
		{ 443, 384, 51 },
		{ 443, 385, 51 },
		{ 443, 386, 51 },
		{ 443, 387, 51 },
		{ 443, 388, 51 },
		{ 443, 389, 51 },
		{ 443, 390, 51 },
		{ 443, 391, 51 },
		{ 443, 392, 51 },
		{ 443, 393, 51 },
		{ 443, 394, 51 },
		{ 443, 395, 51 },
		{ 443, 396, 51 },
		{ 444, 351, 51 },
		{ 444, 352, 51 },
		{ 444, 353, 51 },
		{ 444, 354, 51 },
		{ 444, 355, 51 },
		{ 444, 356, 51 },
		{ 444, 357, 51 },
		{ 444, 358, 51 },
		{ 444, 359, 51 },
		{ 444, 360, 51 },
		{ 444, 361, 51 },
		{ 444, 362, 51 },
		{ 444, 363, 51 },
		{ 444, 364, 51 },
		{ 444, 365, 51 },
		{ 444, 366, 51 },
		{ 444, 367, 51 },
		{ 444, 368, 51 },
		{ 444, 369, 51 },
		{ 444, 370, 51 },
		{ 444, 371, 51 },
		{ 444, 372, 51 },
		{ 444, 373, 51 },
		{ 444, 374, 51 },
		{ 444, 375, 51 },
		{ 444, 376, 51 },
		{ 444, 377, 51 },
		{ 444, 378, 51 },
		{ 444, 379, 51 },
		{ 444, 380, 51 },
		{ 444, 381, 51 },
		{ 444, 382, 51 },
		{ 444, 383, 51 },
		{ 444, 384, 51 },
		{ 444, 385, 51 },
		{ 444, 386, 51 },
		{ 444, 387, 51 },
		{ 444, 388, 51 },
		{ 444, 389, 51 },
		{ 444, 390, 51 },
		{ 444, 391, 51 },
		{ 444, 392, 51 },
		{ 444, 393, 51 },
		{ 444, 394, 51 },
		{ 444, 395, 51 },
		{ 444, 396, 51 },
		{ 444, 397, 51 },
		{ 445, 352, 51 },
		{ 445, 353, 51 },
		{ 445, 354, 51 },
		{ 445, 355, 51 },
		{ 445, 356, 51 },
		{ 445, 357, 51 },
		{ 445, 358, 51 },
		{ 445, 359, 51 },
		{ 445, 360, 51 },
		{ 445, 361, 51 },
		{ 445, 362, 51 },
		{ 445, 363, 51 },
		{ 445, 364, 51 },
		{ 445, 365, 51 },
		{ 445, 366, 51 },
		{ 445, 367, 51 },
		{ 445, 368, 51 },
		{ 445, 369, 51 },
		{ 445, 370, 51 },
		{ 445, 371, 51 },
		{ 445, 372, 51 },
		{ 445, 373, 51 },
		{ 445, 374, 51 },
		{ 445, 375, 51 },
		{ 445, 376, 51 },
		{ 445, 377, 51 },
		{ 445, 378, 51 },
		{ 445, 379, 51 },
		{ 445, 380, 51 },
		{ 445, 381, 51 },
		{ 445, 382, 51 },
		{ 445, 383, 51 },
		{ 445, 384, 51 },
		{ 445, 385, 51 },
		{ 445, 386, 51 },
		{ 445, 387, 51 },
		{ 445, 388, 51 },
		{ 445, 389, 51 },
		{ 445, 390, 51 },
		{ 445, 391, 51 },
		{ 445, 392, 51 },
		{ 445, 393, 51 },
		{ 445, 394, 51 },
		{ 445, 395, 51 },
		{ 445, 396, 51 },
		{ 445, 397, 51 },
		{ 446, 352, 51 },
		{ 446, 353, 51 },
		{ 446, 354, 51 },
		{ 446, 355, 51 },
		{ 446, 356, 51 },
		{ 446, 357, 51 },
		{ 446, 358, 51 },
		{ 446, 359, 51 },
		{ 446, 360, 51 },
		{ 446, 361, 51 },
		{ 446, 362, 51 },
		{ 446, 363, 51 },
		{ 446, 364, 51 },
		{ 446, 365, 51 },
		{ 446, 366, 51 },
		{ 446, 367, 51 },
		{ 446, 368, 51 },
		{ 446, 369, 51 },
		{ 446, 370, 51 },
		{ 446, 371, 51 },
		{ 446, 372, 51 },
		{ 446, 373, 51 },
		{ 446, 374, 51 },
		{ 446, 375, 51 },
		{ 446, 376, 51 },
		{ 446, 377, 51 },
		{ 446, 378, 51 },
		{ 446, 379, 51 },
		{ 446, 380, 51 },
		{ 446, 381, 51 },
		{ 446, 382, 51 },
		{ 446, 383, 51 },
		{ 446, 384, 51 },
		{ 446, 385, 51 },
		{ 446, 386, 51 },
		{ 446, 387, 51 },
		{ 446, 388, 51 },
		{ 446, 389, 51 },
		{ 446, 390, 51 },
		{ 446, 391, 51 },
		{ 446, 392, 51 },
		{ 446, 393, 51 },
		{ 446, 394, 51 },
		{ 446, 395, 51 },
		{ 446, 396, 51 },
		{ 446, 397, 51 },
		{ 446, 398, 51 },
		{ 447, 353, 51 },
		{ 447, 354, 51 },
		{ 447, 355, 51 },
		{ 447, 356, 51 },
		{ 447, 357, 51 },
		{ 447, 358, 51 },
		{ 447, 359, 51 },
		{ 447, 360, 51 },
		{ 447, 361, 51 },
		{ 447, 362, 51 },
		{ 447, 363, 51 },
		{ 447, 364, 51 },
		{ 447, 365, 51 },
		{ 447, 366, 51 },
		{ 447, 367, 51 },
		{ 447, 368, 51 },
		{ 447, 369, 51 },
		{ 447, 370, 51 },
		{ 447, 371, 51 },
		{ 447, 372, 51 },
		{ 447, 373, 51 },
		{ 447, 374, 51 },
		{ 447, 375, 51 },
		{ 447, 376, 51 },
		{ 447, 377, 51 },
		{ 447, 378, 51 },
		{ 447, 379, 51 },
		{ 447, 380, 51 },
		{ 447, 381, 51 },
		{ 447, 382, 51 },
		{ 447, 383, 51 },
		{ 447, 384, 51 },
		{ 447, 385, 51 },
		{ 447, 386, 51 },
		{ 447, 387, 51 },
		{ 447, 388, 51 },
		{ 447, 389, 51 },
		{ 447, 390, 51 },
		{ 447, 391, 51 },
		{ 447, 392, 51 },
		{ 447, 393, 51 },
		{ 447, 394, 51 },
		{ 447, 395, 51 },
		{ 447, 396, 51 },
		{ 447, 397, 51 },
		{ 447, 398, 51 },
		{ 447, 399, 51 },
		{ 448, 353, 51 },
		{ 448, 354, 51 },
		{ 448, 355, 51 },
		{ 448, 356, 51 },
		{ 448, 357, 51 },
		{ 448, 358, 51 },
		{ 448, 359, 51 },
		{ 448, 360, 51 },
		{ 448, 361, 51 },
		{ 448, 362, 51 },
		{ 448, 363, 51 },
		{ 448, 364, 51 },
		{ 448, 365, 51 },
		{ 448, 366, 51 },
		{ 448, 367, 51 },
		{ 448, 368, 51 },
		{ 448, 369, 51 },
		{ 448, 370, 51 },
		{ 448, 371, 51 },
		{ 448, 372, 51 },
		{ 448, 373, 51 },
		{ 448, 374, 51 },
		{ 448, 375, 51 },
		{ 448, 376, 51 },
		{ 448, 377, 51 },
		{ 448, 378, 51 },
		{ 448, 379, 51 },
		{ 448, 380, 51 },
		{ 448, 381, 51 },
		{ 448, 382, 51 },
		{ 448, 383, 51 },
		{ 448, 384, 51 },
		{ 448, 385, 51 },
		{ 448, 386, 51 },
		{ 448, 387, 51 },
		{ 448, 388, 51 },
		{ 448, 389, 51 },
		{ 448, 390, 51 },
		{ 448, 391, 51 },
		{ 448, 392, 51 },
		{ 448, 393, 51 },
		{ 448, 394, 51 },
		{ 448, 395, 51 },
		{ 448, 396, 51 },
		{ 448, 397, 51 },
		{ 448, 398, 51 },
		{ 448, 399, 51 },
		{ 448, 400, 51 },
		{ 449, 354, 51 },
		{ 449, 355, 51 },
		{ 449, 356, 51 },
		{ 449, 357, 51 },
		{ 449, 358, 51 },
		{ 449, 359, 51 },
		{ 449, 360, 51 },
		{ 449, 361, 51 },
		{ 449, 362, 51 },
		{ 449, 363, 51 },
		{ 449, 364, 51 },
		{ 449, 365, 51 },
		{ 449, 366, 51 },
		{ 449, 367, 51 },
		{ 449, 368, 51 },
		{ 449, 369, 51 },
		{ 449, 370, 51 },
		{ 449, 371, 51 },
		{ 449, 372, 51 },
		{ 449, 373, 51 },
		{ 449, 374, 51 },
		{ 449, 375, 51 },
		{ 449, 376, 51 },
		{ 449, 377, 51 },
		{ 449, 378, 51 },
		{ 449, 379, 51 },
		{ 449, 380, 51 },
		{ 449, 381, 51 },
		{ 449, 382, 51 },
		{ 449, 383, 51 },
		{ 449, 384, 51 },
		{ 449, 385, 51 },
		{ 449, 386, 51 },
		{ 449, 387, 51 },
		{ 449, 388, 51 },
		{ 449, 389, 51 },
		{ 449, 390, 51 },
		{ 449, 391, 51 },
		{ 449, 392, 51 },
		{ 449, 393, 51 },
		{ 449, 394, 51 },
		{ 449, 395, 51 },
		{ 449, 396, 51 },
		{ 449, 397, 51 },
		{ 449, 398, 51 },
		{ 449, 399, 51 },
		{ 449, 400, 51 },
		{ 450, 354, 51 },
		{ 450, 355, 51 },
		{ 450, 356, 51 },
		{ 450, 357, 51 },
		{ 450, 358, 51 },
		{ 450, 359, 51 },
		{ 450, 360, 51 },
		{ 450, 361, 51 },
		{ 450, 362, 51 },
		{ 450, 363, 51 },
		{ 450, 364, 51 },
		{ 450, 365, 51 },
		{ 450, 366, 51 },
		{ 450, 367, 51 },
		{ 450, 368, 51 },
		{ 450, 369, 51 },
		{ 450, 370, 51 },
		{ 450, 371, 51 },
		{ 450, 372, 51 },
		{ 450, 373, 51 },
		{ 450, 374, 51 },
		{ 450, 375, 51 },
		{ 450, 376, 51 },
		{ 450, 377, 51 },
		{ 450, 378, 51 },
		{ 450, 379, 51 },
		{ 450, 380, 51 },
		{ 450, 381, 51 },
		{ 450, 382, 51 },
		{ 450, 383, 51 },
		{ 450, 384, 51 },
		{ 450, 385, 51 },
		{ 450, 386, 51 },
		{ 450, 387, 51 },
		{ 450, 388, 51 },
		{ 450, 389, 51 },
		{ 450, 390, 51 },
		{ 450, 391, 51 },
		{ 450, 392, 51 },
		{ 450, 393, 51 },
		{ 450, 394, 51 },
		{ 450, 395, 51 },
		{ 450, 396, 51 },
		{ 450, 397, 51 },
		{ 450, 398, 51 },
		{ 450, 399, 51 },
		{ 450, 400, 51 },
		{ 450, 401, 51 },
		{ 451, 355, 51 },
		{ 451, 356, 51 },
		{ 451, 357, 51 },
		{ 451, 358, 51 },
		{ 451, 359, 51 },
		{ 451, 360, 51 },
		{ 451, 361, 51 },
		{ 451, 362, 51 },
		{ 451, 363, 51 },
		{ 451, 364, 51 },
		{ 451, 365, 51 },
		{ 451, 366, 51 },
		{ 451, 367, 51 },
		{ 451, 368, 51 },
		{ 451, 369, 51 },
		{ 451, 370, 51 },
		{ 451, 371, 51 },
		{ 451, 372, 51 },
		{ 451, 373, 51 },
		{ 451, 374, 51 },
		{ 451, 375, 51 },
		{ 451, 376, 51 },
		{ 451, 377, 51 },
		{ 451, 378, 51 },
		{ 451, 379, 51 },
		{ 451, 380, 51 },
		{ 451, 381, 51 },
		{ 451, 382, 51 },
		{ 451, 383, 51 },
		{ 451, 384, 51 },
		{ 451, 385, 51 },
		{ 451, 386, 51 },
		{ 451, 387, 51 },
		{ 451, 388, 51 },
		{ 451, 389, 51 },
		{ 451, 390, 51 },
		{ 451, 391, 51 },
		{ 451, 392, 51 },
		{ 451, 393, 51 },
		{ 451, 394, 51 },
		{ 451, 395, 51 },
		{ 451, 396, 51 },
		{ 451, 397, 51 },
		{ 451, 398, 51 },
		{ 451, 399, 51 },
		{ 451, 400, 51 },
		{ 451, 401, 51 },
		{ 451, 402, 51 },
		{ 452, 355, 51 },
		{ 452, 356, 51 },
		{ 452, 357, 51 },
		{ 452, 358, 51 },
		{ 452, 359, 51 },
		{ 452, 360, 51 },
		{ 452, 361, 51 },
		{ 452, 362, 51 },
		{ 452, 363, 51 },
		{ 452, 364, 51 },
		{ 452, 365, 51 },
		{ 452, 366, 51 },
		{ 452, 367, 51 },
		{ 452, 368, 51 },
		{ 452, 369, 51 },
		{ 452, 370, 51 },
		{ 452, 371, 51 },
		{ 452, 372, 51 },
		{ 452, 373, 51 },
		{ 452, 374, 51 },
		{ 452, 375, 51 },
		{ 452, 376, 51 },
		{ 452, 377, 51 },
		{ 452, 378, 51 },
		{ 452, 379, 51 },
		{ 452, 380, 51 },
		{ 452, 381, 51 },
		{ 452, 382, 51 },
		{ 452, 383, 51 },
		{ 452, 384, 51 },
		{ 452, 385, 51 },
		{ 452, 386, 51 },
		{ 452, 387, 51 },
		{ 452, 388, 51 },
		{ 452, 389, 51 },
		{ 452, 390, 51 },
		{ 452, 391, 51 },
		{ 452, 392, 51 },
		{ 452, 393, 51 },
		{ 452, 394, 51 },
		{ 452, 395, 51 },
		{ 452, 396, 51 },
		{ 452, 397, 51 },
		{ 452, 398, 51 },
		{ 452, 399, 51 },
		{ 452, 400, 51 },
		{ 452, 401, 51 },
		{ 452, 402, 51 },
		{ 452, 403, 51 },
		{ 453, 356, 51 },
		{ 453, 357, 51 },
		{ 453, 358, 51 },
		{ 453, 359, 51 },
		{ 453, 360, 51 },
		{ 453, 361, 51 },
		{ 453, 362, 51 },
		{ 453, 363, 51 },
		{ 453, 364, 51 },
		{ 453, 365, 51 },
		{ 453, 366, 51 },
		{ 453, 367, 51 },
		{ 453, 368, 51 },
		{ 453, 369, 51 },
		{ 453, 370, 51 },
		{ 453, 371, 51 },
		{ 453, 372, 51 },
		{ 453, 373, 51 },
		{ 453, 374, 51 },
		{ 453, 375, 51 },
		{ 453, 376, 51 },
		{ 453, 377, 51 },
		{ 453, 378, 51 },
		{ 453, 379, 51 },
		{ 453, 380, 51 },
		{ 453, 381, 51 },
		{ 453, 382, 51 },
		{ 453, 383, 51 },
		{ 453, 384, 51 },
		{ 453, 385, 51 },
		{ 453, 386, 51 },
		{ 453, 387, 51 },
		{ 453, 388, 51 },
		{ 453, 389, 51 },
		{ 453, 390, 51 },
		{ 453, 391, 51 },
		{ 453, 392, 51 },
		{ 453, 393, 51 },
		{ 453, 394, 51 },
		{ 453, 395, 51 },
		{ 453, 396, 51 },
		{ 453, 397, 51 },
		{ 453, 398, 51 },
		{ 453, 399, 51 },
		{ 453, 400, 51 },
		{ 453, 401, 51 },
		{ 453, 402, 51 },
		{ 453, 403, 51 },
		{ 454, 356, 51 },
		{ 454, 357, 51 },
		{ 454, 358, 51 },
		{ 454, 359, 51 },
		{ 454, 360, 51 },
		{ 454, 361, 51 },
		{ 454, 362, 51 },
		{ 454, 363, 51 },
		{ 454, 364, 51 },
		{ 454, 365, 51 },
		{ 454, 366, 51 },
		{ 454, 367, 51 },
		{ 454, 368, 51 },
		{ 454, 369, 51 },
		{ 454, 370, 51 },
		{ 454, 371, 51 },
		{ 454, 372, 51 },
		{ 454, 373, 51 },
		{ 454, 374, 51 },
		{ 454, 375, 51 },
		{ 454, 376, 51 },
		{ 454, 377, 51 },
		{ 454, 378, 51 },
		{ 454, 379, 51 },
		{ 454, 380, 51 },
		{ 454, 381, 51 },
		{ 454, 382, 51 },
		{ 454, 383, 51 },
		{ 454, 384, 51 },
		{ 454, 385, 51 },
		{ 454, 386, 51 },
		{ 454, 387, 51 },
		{ 454, 388, 51 },
		{ 454, 389, 51 },
		{ 454, 390, 51 },
		{ 454, 391, 51 },
		{ 454, 392, 51 },
		{ 454, 393, 51 },
		{ 454, 394, 51 },
		{ 454, 395, 51 },
		{ 454, 396, 51 },
		{ 454, 397, 51 },
		{ 454, 398, 51 },
		{ 454, 399, 51 },
		{ 454, 400, 51 },
		{ 454, 401, 51 },
		{ 454, 402, 51 },
		{ 454, 403, 51 },
		{ 454, 404, 51 },
		{ 455, 357, 51 },
		{ 455, 358, 51 },
		{ 455, 359, 51 },
		{ 455, 360, 51 },
		{ 455, 361, 51 },
		{ 455, 362, 51 },
		{ 455, 363, 51 },
		{ 455, 364, 51 },
		{ 455, 365, 51 },
		{ 455, 366, 51 },
		{ 455, 367, 51 },
		{ 455, 368, 51 },
		{ 455, 369, 51 },
		{ 455, 370, 51 },
		{ 455, 371, 51 },
		{ 455, 372, 51 },
		{ 455, 373, 51 },
		{ 455, 374, 51 },
		{ 455, 375, 51 },
		{ 455, 376, 51 },
		{ 455, 377, 51 },
		{ 455, 378, 51 },
		{ 455, 379, 51 },
		{ 455, 380, 51 },
		{ 455, 381, 51 },
		{ 455, 382, 51 },
		{ 455, 383, 51 },
		{ 455, 384, 51 },
		{ 455, 385, 51 },
		{ 455, 386, 51 },
		{ 455, 387, 51 },
		{ 455, 388, 51 },
		{ 455, 389, 51 },
		{ 455, 390, 51 },
		{ 455, 391, 51 },
		{ 455, 392, 51 },
		{ 455, 393, 51 },
		{ 455, 394, 51 },
		{ 455, 395, 51 },
		{ 455, 396, 51 },
		{ 455, 397, 51 },
		{ 455, 398, 51 },
		{ 455, 399, 51 },
		{ 455, 400, 51 },
		{ 455, 401, 51 },
		{ 455, 402, 51 },
		{ 455, 403, 51 },
		{ 455, 404, 51 },
		{ 455, 405, 51 },
		{ 456, 357, 51 },
		{ 456, 358, 51 },
		{ 456, 359, 51 },
		{ 456, 360, 51 },
		{ 456, 361, 51 },
		{ 456, 362, 51 },
		{ 456, 363, 51 },
		{ 456, 364, 51 },
		{ 456, 365, 51 },
		{ 456, 366, 51 },
		{ 456, 367, 51 },
		{ 456, 368, 51 },
		{ 456, 369, 51 },
		{ 456, 370, 51 },
		{ 456, 371, 51 },
		{ 456, 372, 51 },
		{ 456, 373, 51 },
		{ 456, 374, 51 },
		{ 456, 375, 51 },
		{ 456, 376, 51 },
		{ 456, 377, 51 },
		{ 456, 378, 51 },
		{ 456, 379, 51 },
		{ 456, 380, 51 },
		{ 456, 381, 51 },
		{ 456, 382, 51 },
		{ 456, 383, 51 },
		{ 456, 384, 51 },
		{ 456, 385, 51 },
		{ 456, 386, 51 },
		{ 456, 387, 51 },
		{ 456, 388, 51 },
		{ 456, 389, 51 },
		{ 456, 390, 51 },
		{ 456, 391, 51 },
		{ 456, 392, 51 },
		{ 456, 393, 51 },
		{ 456, 394, 51 },
		{ 456, 395, 51 },
		{ 456, 396, 51 },
		{ 456, 397, 51 },
		{ 456, 398, 51 },
		{ 456, 399, 51 },
		{ 456, 400, 51 },
		{ 456, 401, 51 },
		{ 456, 402, 51 },
		{ 456, 403, 51 },
		{ 456, 404, 51 },
		{ 456, 405, 51 },
		{ 456, 406, 51 },
		{ 457, 358, 51 },
		{ 457, 359, 51 },
		{ 457, 360, 51 },
		{ 457, 361, 51 },
		{ 457, 362, 51 },
		{ 457, 363, 51 },
		{ 457, 364, 51 },
		{ 457, 365, 51 },
		{ 457, 366, 51 },
		{ 457, 367, 51 },
		{ 457, 368, 51 },
		{ 457, 369, 51 },
		{ 457, 370, 51 },
		{ 457, 371, 51 },
		{ 457, 372, 51 },
		{ 457, 373, 51 },
		{ 457, 374, 51 },
		{ 457, 375, 51 },
		{ 457, 376, 51 },
		{ 457, 377, 51 },
		{ 457, 378, 51 },
		{ 457, 379, 51 },
		{ 457, 380, 51 },
		{ 457, 381, 51 },
		{ 457, 382, 51 },
		{ 457, 383, 51 },
		{ 457, 384, 51 },
		{ 457, 385, 51 },
		{ 457, 386, 51 },
		{ 457, 387, 51 },
		{ 457, 388, 51 },
		{ 457, 389, 51 },
		{ 457, 390, 51 },
		{ 457, 391, 51 },
		{ 457, 392, 51 },
		{ 457, 393, 51 },
		{ 457, 394, 51 },
		{ 457, 395, 51 },
		{ 457, 396, 51 },
		{ 457, 397, 51 },
		{ 457, 398, 51 },
		{ 457, 399, 51 },
		{ 457, 400, 51 },
		{ 457, 401, 51 },
		{ 457, 402, 51 },
		{ 457, 403, 51 },
		{ 457, 404, 51 },
		{ 457, 405, 51 },
		{ 457, 406, 51 },
		{ 458, 358, 51 },
		{ 458, 359, 51 },
		{ 458, 360, 51 },
		{ 458, 361, 51 },
		{ 458, 362, 51 },
		{ 458, 363, 51 },
		{ 458, 364, 51 },
		{ 458, 365, 51 },
		{ 458, 366, 51 },
		{ 458, 367, 51 },
		{ 458, 368, 51 },
		{ 458, 369, 51 },
		{ 458, 370, 51 },
		{ 458, 371, 51 },
		{ 458, 372, 51 },
		{ 458, 373, 51 },
		{ 458, 374, 51 },
		{ 458, 375, 51 },
		{ 458, 376, 51 },
		{ 458, 377, 51 },
		{ 458, 378, 51 },
		{ 458, 379, 51 },
		{ 458, 380, 51 },
		{ 458, 381, 51 },
		{ 458, 382, 51 },
		{ 458, 383, 51 },
		{ 458, 384, 51 },
		{ 458, 385, 51 },
		{ 458, 386, 51 },
		{ 458, 387, 51 },
		{ 458, 388, 51 },
		{ 458, 389, 51 },
		{ 458, 390, 51 },
		{ 458, 391, 51 },
		{ 458, 392, 51 },
		{ 458, 393, 51 },
		{ 458, 394, 51 },
		{ 458, 395, 51 },
		{ 458, 396, 51 },
		{ 458, 397, 51 },
		{ 458, 398, 51 },
		{ 458, 399, 51 },
		{ 458, 400, 51 },
		{ 458, 401, 51 },
		{ 458, 402, 51 },
		{ 458, 403, 51 },
		{ 458, 404, 51 },
		{ 458, 405, 51 },
		{ 458, 406, 51 },
		{ 458, 407, 51 },
		{ 459, 359, 51 },
		{ 459, 360, 51 },
		{ 459, 361, 51 },
		{ 459, 362, 51 },
		{ 459, 363, 51 },
		{ 459, 364, 51 },
		{ 459, 365, 51 },
		{ 459, 366, 51 },
		{ 459, 367, 51 },
		{ 459, 368, 51 },
		{ 459, 369, 51 },
		{ 459, 370, 51 },
		{ 459, 371, 51 },
		{ 459, 372, 51 },
		{ 459, 373, 51 },
		{ 459, 374, 51 },
		{ 459, 375, 51 },
		{ 459, 376, 51 },
		{ 459, 377, 51 },
		{ 459, 378, 51 },
		{ 459, 379, 51 },
		{ 459, 380, 51 },
		{ 459, 381, 51 },
		{ 459, 382, 51 },
		{ 459, 383, 51 },
		{ 459, 384, 51 },
		{ 459, 385, 51 },
		{ 459, 386, 51 },
		{ 459, 387, 51 },
		{ 459, 388, 51 },
		{ 459, 389, 51 },
		{ 459, 390, 51 },
		{ 459, 391, 51 },
		{ 459, 392, 51 },
		{ 459, 393, 51 },
		{ 459, 394, 51 },
		{ 459, 395, 51 },
		{ 459, 396, 51 },
		{ 459, 397, 51 },
		{ 459, 398, 51 },
		{ 459, 399, 51 },
		{ 459, 400, 51 },
		{ 459, 401, 51 },
		{ 459, 402, 51 },
		{ 459, 403, 51 },
		{ 459, 404, 51 },
		{ 459, 405, 51 },
		{ 459, 406, 51 },
		{ 459, 407, 51 },
		{ 459, 408, 51 },
		{ 460, 359, 51 },
		{ 460, 360, 51 },
		{ 460, 361, 51 },
		{ 460, 362, 51 },
		{ 460, 363, 51 },
		{ 460, 364, 51 },
		{ 460, 365, 51 },
		{ 460, 366, 51 },
		{ 460, 367, 51 },
		{ 460, 368, 51 },
		{ 460, 369, 51 },
		{ 460, 370, 51 },
		{ 460, 371, 51 },
		{ 460, 372, 51 },
		{ 460, 373, 51 },
		{ 460, 374, 51 },
		{ 460, 375, 51 },
		{ 460, 376, 51 },
		{ 460, 377, 51 },
		{ 460, 378, 51 },
		{ 460, 379, 51 },
		{ 460, 380, 51 },
		{ 460, 381, 51 },
		{ 460, 382, 51 },
		{ 460, 383, 51 },
		{ 460, 384, 51 },
		{ 460, 385, 51 },
		{ 460, 386, 51 },
		{ 460, 387, 51 },
		{ 460, 388, 51 },
		{ 460, 389, 51 },
		{ 460, 390, 51 },
		{ 460, 391, 51 },
		{ 460, 392, 51 },
		{ 460, 393, 51 },
		{ 460, 394, 51 },
		{ 460, 395, 51 },
		{ 460, 396, 51 },
		{ 460, 397, 51 },
		{ 460, 398, 51 },
		{ 460, 399, 51 },
		{ 460, 400, 51 },
		{ 460, 401, 51 },
		{ 460, 402, 51 },
		{ 460, 403, 51 },
		{ 460, 404, 51 },
		{ 460, 405, 51 },
		{ 460, 406, 51 },
		{ 460, 407, 51 },
		{ 460, 408, 51 },
		{ 460, 409, 51 },
		{ 461, 360, 51 },
		{ 461, 361, 51 },
		{ 461, 362, 51 },
		{ 461, 363, 51 },
		{ 461, 364, 51 },
		{ 461, 365, 51 },
		{ 461, 366, 51 },
		{ 461, 367, 51 },
		{ 461, 368, 51 },
		{ 461, 369, 51 },
		{ 461, 370, 51 },
		{ 461, 371, 51 },
		{ 461, 372, 51 },
		{ 461, 373, 51 },
		{ 461, 374, 51 },
		{ 461, 375, 51 },
		{ 461, 376, 51 },
		{ 461, 377, 51 },
		{ 461, 378, 51 },
		{ 461, 379, 51 },
		{ 461, 380, 51 },
		{ 461, 381, 51 },
		{ 461, 382, 51 },
		{ 461, 383, 51 },
		{ 461, 384, 51 },
		{ 461, 385, 51 },
		{ 461, 386, 51 },
		{ 461, 387, 51 },
		{ 461, 388, 51 },
		{ 461, 389, 51 },
		{ 461, 390, 51 },
		{ 461, 391, 51 },
		{ 461, 392, 51 },
		{ 461, 393, 51 },
		{ 461, 394, 51 },
		{ 461, 395, 51 },
		{ 461, 396, 51 },
		{ 461, 397, 51 },
		{ 461, 398, 51 },
		{ 461, 399, 51 },
		{ 461, 400, 51 },
		{ 461, 401, 51 },
		{ 461, 402, 51 },
		{ 461, 403, 51 },
		{ 461, 404, 51 },
		{ 461, 405, 51 },
		{ 461, 406, 51 },
		{ 461, 407, 51 },
		{ 461, 408, 51 },
		{ 461, 409, 51 },
		{ 462, 360, 51 },
		{ 462, 361, 51 },
		{ 462, 362, 51 },
		{ 462, 363, 51 },
		{ 462, 364, 51 },
		{ 462, 365, 51 },
		{ 462, 366, 51 },
		{ 462, 367, 51 },
		{ 462, 368, 51 },
		{ 462, 369, 51 },
		{ 462, 370, 51 },
		{ 462, 371, 51 },
		{ 462, 372, 51 },
		{ 462, 373, 51 },
		{ 462, 374, 51 },
		{ 462, 375, 51 },
		{ 462, 376, 51 },
		{ 462, 377, 51 },
		{ 462, 378, 51 },
		{ 462, 379, 51 },
		{ 462, 380, 51 },
		{ 462, 381, 51 },
		{ 462, 382, 51 },
		{ 462, 383, 51 },
		{ 462, 384, 51 },
		{ 462, 385, 51 },
		{ 462, 386, 51 },
		{ 462, 387, 51 },
		{ 462, 388, 51 },
		{ 462, 389, 51 },
		{ 462, 390, 51 },
		{ 462, 391, 51 },
		{ 462, 392, 51 },
		{ 462, 393, 51 },
		{ 462, 394, 51 },
		{ 462, 395, 51 },
		{ 462, 396, 51 },
		{ 462, 397, 51 },
		{ 462, 398, 51 },
		{ 462, 399, 51 },
		{ 462, 400, 51 },
		{ 462, 401, 51 },
		{ 462, 402, 51 },
		{ 462, 403, 51 },
		{ 462, 404, 51 },
		{ 462, 405, 51 },
		{ 462, 406, 51 },
		{ 462, 407, 51 },
		{ 462, 408, 51 },
		{ 462, 409, 51 },
		{ 462, 410, 51 },
		{ 463, 361, 51 },
		{ 463, 362, 51 },
		{ 463, 363, 51 },
		{ 463, 364, 51 },
		{ 463, 365, 51 },
		{ 463, 366, 51 },
		{ 463, 367, 51 },
		{ 463, 368, 51 },
		{ 463, 369, 51 },
		{ 463, 370, 51 },
		{ 463, 371, 51 },
		{ 463, 372, 51 },
		{ 463, 373, 51 },
		{ 463, 374, 51 },
		{ 463, 375, 51 },
		{ 463, 376, 51 },
		{ 463, 377, 51 },
		{ 463, 378, 51 },
		{ 463, 379, 51 },
		{ 463, 380, 51 },
		{ 463, 381, 51 },
		{ 463, 382, 51 },
		{ 463, 383, 51 },
		{ 463, 384, 51 },
		{ 463, 385, 51 },
		{ 463, 386, 51 },
		{ 463, 387, 51 },
		{ 463, 388, 51 },
		{ 463, 389, 51 },
		{ 463, 390, 51 },
		{ 463, 391, 51 },
		{ 463, 392, 51 },
		{ 463, 393, 51 },
		{ 463, 394, 51 },
		{ 463, 395, 51 },
		{ 463, 396, 51 },
		{ 463, 397, 51 },
		{ 463, 398, 51 },
		{ 463, 399, 51 },
		{ 463, 400, 51 },
		{ 463, 401, 51 },
		{ 463, 402, 51 },
		{ 463, 403, 51 },
		{ 463, 404, 51 },
		{ 463, 405, 51 },
		{ 463, 406, 51 },
		{ 463, 407, 51 },
		{ 463, 408, 51 },
		{ 463, 409, 51 },
		{ 463, 410, 51 },
		{ 463, 411, 51 },
		{ 464, 361, 51 },
		{ 464, 362, 51 },
		{ 464, 363, 51 },
		{ 464, 364, 51 },
		{ 464, 365, 51 },
		{ 464, 366, 51 },
		{ 464, 367, 51 },
		{ 464, 368, 51 },
		{ 464, 369, 51 },
		{ 464, 370, 51 },
		{ 464, 371, 51 },
		{ 464, 372, 51 },
		{ 464, 373, 51 },
		{ 464, 374, 51 },
		{ 464, 375, 51 },
		{ 464, 376, 51 },
		{ 464, 377, 51 },
		{ 464, 378, 51 },
		{ 464, 379, 51 },
		{ 464, 380, 51 },
		{ 464, 381, 51 },
		{ 464, 382, 51 },
		{ 464, 383, 51 },
		{ 464, 384, 51 },
		{ 464, 385, 51 },
		{ 464, 386, 51 },
		{ 464, 387, 51 },
		{ 464, 388, 51 },
		{ 464, 389, 51 },
		{ 464, 390, 51 },
		{ 464, 391, 51 },
		{ 464, 392, 51 },
		{ 464, 393, 51 },
		{ 464, 394, 51 },
		{ 464, 395, 51 },
		{ 464, 396, 51 },
		{ 464, 397, 51 },
		{ 464, 398, 51 },
		{ 464, 399, 51 },
		{ 464, 400, 51 },
		{ 464, 401, 51 },
		{ 464, 402, 51 },
		{ 464, 403, 51 },
		{ 464, 404, 51 },
		{ 464, 405, 51 },
		{ 464, 406, 51 },
		{ 464, 407, 51 },
		{ 464, 408, 51 },
		{ 464, 409, 51 },
		{ 464, 410, 51 },
		{ 464, 411, 51 },
		{ 464, 412, 51 },
		{ 465, 362, 51 },
		{ 465, 363, 51 },
		{ 465, 364, 51 },
		{ 465, 365, 51 },
		{ 465, 366, 51 },
		{ 465, 367, 51 },
		{ 465, 368, 51 },
		{ 465, 369, 51 },
		{ 465, 370, 51 },
		{ 465, 371, 51 },
		{ 465, 372, 51 },
		{ 465, 373, 51 },
		{ 465, 374, 51 },
		{ 465, 375, 51 },
		{ 465, 376, 51 },
		{ 465, 377, 51 },
		{ 465, 378, 51 },
		{ 465, 379, 51 },
		{ 465, 380, 51 },
		{ 465, 381, 51 },
		{ 465, 382, 51 },
		{ 465, 383, 51 },
		{ 465, 384, 51 },
		{ 465, 385, 51 },
		{ 465, 386, 51 },
		{ 465, 387, 51 },
		{ 465, 388, 51 },
		{ 465, 389, 51 },
		{ 465, 390, 51 },
		{ 465, 391, 51 },
		{ 465, 392, 51 },
		{ 465, 393, 51 },
		{ 465, 394, 51 },
		{ 465, 395, 51 },
		{ 465, 396, 51 },
		{ 465, 397, 51 },
		{ 465, 398, 51 },
		{ 465, 399, 51 },
		{ 465, 400, 51 },
		{ 465, 401, 51 },
		{ 465, 402, 51 },
		{ 465, 403, 51 },
		{ 465, 404, 51 },
		{ 465, 405, 51 },
		{ 465, 406, 51 },
		{ 465, 407, 51 },
		{ 465, 408, 51 },
		{ 465, 409, 51 },
		{ 465, 410, 51 },
		{ 465, 411, 51 },
		{ 465, 412, 51 },
		{ 466, 362, 51 },
		{ 466, 363, 51 },
		{ 466, 364, 51 },
		{ 466, 365, 51 },
		{ 466, 366, 51 },
		{ 466, 367, 51 },
		{ 466, 368, 51 },
		{ 466, 369, 51 },
		{ 466, 370, 51 },
		{ 466, 371, 51 },
		{ 466, 372, 51 },
		{ 466, 373, 51 },
		{ 466, 374, 51 },
		{ 466, 375, 51 },
		{ 466, 376, 51 },
		{ 466, 377, 51 },
		{ 466, 378, 51 },
		{ 466, 379, 51 },
		{ 466, 380, 51 },
		{ 466, 381, 51 },
		{ 466, 382, 51 },
		{ 466, 383, 51 },
		{ 466, 384, 51 },
		{ 466, 385, 51 },
		{ 466, 386, 51 },
		{ 466, 387, 51 },
		{ 466, 388, 51 },
		{ 466, 389, 51 },
		{ 466, 390, 51 },
		{ 466, 391, 51 },
		{ 466, 392, 51 },
		{ 466, 393, 51 },
		{ 466, 394, 51 },
		{ 466, 395, 51 },
		{ 466, 396, 51 },
		{ 466, 397, 51 },
		{ 466, 398, 51 },
		{ 466, 399, 51 },
		{ 466, 400, 51 },
		{ 466, 401, 51 },
		{ 466, 402, 51 },
		{ 466, 403, 51 },
		{ 466, 404, 51 },
		{ 466, 405, 51 },
		{ 466, 406, 51 },
		{ 466, 407, 51 },
		{ 466, 408, 51 },
		{ 466, 409, 51 },
		{ 466, 410, 51 },
		{ 466, 411, 51 },
		{ 466, 412, 51 },
		{ 466, 413, 51 },
		{ 467, 363, 51 },
		{ 467, 364, 51 },
		{ 467, 365, 51 },
		{ 467, 366, 51 },
		{ 467, 367, 51 },
		{ 467, 368, 51 },
		{ 467, 369, 51 },
		{ 467, 370, 51 },
		{ 467, 371, 51 },
		{ 467, 372, 51 },
		{ 467, 373, 51 },
		{ 467, 374, 51 },
		{ 467, 375, 51 },
		{ 467, 376, 51 },
		{ 467, 377, 51 },
		{ 467, 378, 51 },
		{ 467, 379, 51 },
		{ 467, 380, 51 },
		{ 467, 381, 51 },
		{ 467, 382, 51 },
		{ 467, 383, 51 },
		{ 467, 384, 51 },
		{ 467, 385, 51 },
		{ 467, 386, 51 },
		{ 467, 387, 51 },
		{ 467, 388, 51 },
		{ 467, 389, 51 },
		{ 467, 390, 51 },
		{ 467, 391, 51 },
		{ 467, 392, 51 },
		{ 467, 393, 51 },
		{ 467, 394, 51 },
		{ 467, 395, 51 },
		{ 467, 396, 51 },
		{ 467, 397, 51 },
		{ 467, 398, 51 },
		{ 467, 399, 51 },
		{ 467, 400, 51 },
		{ 467, 401, 51 },
		{ 467, 402, 51 },
		{ 467, 403, 51 },
		{ 467, 404, 51 },
		{ 467, 405, 51 },
		{ 467, 406, 51 },
		{ 467, 407, 51 },
		{ 467, 408, 51 },
		{ 467, 409, 51 },
		{ 467, 410, 51 },
		{ 467, 411, 51 },
		{ 467, 412, 51 },
		{ 467, 413, 51 },
		{ 467, 414, 51 },
		{ 468, 363, 51 },
		{ 468, 364, 51 },
		{ 468, 365, 51 },
		{ 468, 366, 51 },
		{ 468, 367, 51 },
		{ 468, 368, 51 },
		{ 468, 369, 51 },
		{ 468, 370, 51 },
		{ 468, 371, 51 },
		{ 468, 372, 51 },
		{ 468, 373, 51 },
		{ 468, 374, 51 },
		{ 468, 375, 51 },
		{ 468, 376, 51 },
		{ 468, 377, 51 },
		{ 468, 378, 51 },
		{ 468, 379, 51 },
		{ 468, 380, 51 },
		{ 468, 381, 51 },
		{ 468, 382, 51 },
		{ 468, 383, 51 },
		{ 468, 384, 51 },
		{ 468, 385, 51 },
		{ 468, 386, 51 },
		{ 468, 387, 51 },
		{ 468, 388, 51 },
		{ 468, 389, 51 },
		{ 468, 390, 51 },
		{ 468, 391, 51 },
		{ 468, 392, 51 },
		{ 468, 393, 51 },
		{ 468, 394, 51 },
		{ 468, 395, 51 },
		{ 468, 396, 51 },
		{ 468, 397, 51 },
		{ 468, 398, 51 },
		{ 468, 399, 51 },
		{ 468, 400, 51 },
		{ 468, 401, 51 },
		{ 468, 402, 51 },
		{ 468, 403, 51 },
		{ 468, 404, 51 },
		{ 468, 405, 51 },
		{ 468, 406, 51 },
		{ 468, 407, 51 },
		{ 468, 408, 51 },
		{ 468, 409, 51 },
		{ 468, 410, 51 },
		{ 468, 411, 51 },
		{ 468, 412, 51 },
		{ 468, 413, 51 },
		{ 468, 414, 51 },
		{ 468, 415, 51 },
		{ 469, 364, 51 },
		{ 469, 365, 51 },
		{ 469, 366, 51 },
		{ 469, 367, 51 },
		{ 469, 368, 51 },
		{ 469, 369, 51 },
		{ 469, 370, 51 },
		{ 469, 371, 51 },
		{ 469, 372, 51 },
		{ 469, 373, 51 },
		{ 469, 374, 51 },
		{ 469, 375, 51 },
		{ 469, 376, 51 },
		{ 469, 377, 51 },
		{ 469, 378, 51 },
		{ 469, 379, 51 },
		{ 469, 380, 51 },
		{ 469, 381, 51 },
		{ 469, 382, 51 },
		{ 469, 383, 51 },
		{ 469, 384, 51 },
		{ 469, 385, 51 },
		{ 469, 386, 51 },
		{ 469, 387, 51 },
		{ 469, 388, 51 },
		{ 469, 389, 51 },
		{ 469, 390, 51 },
		{ 469, 391, 51 },
		{ 469, 392, 51 },
		{ 469, 393, 51 },
		{ 469, 394, 51 },
		{ 469, 395, 51 },
		{ 469, 396, 51 },
		{ 469, 397, 51 },
		{ 469, 398, 51 },
		{ 469, 399, 51 },
		{ 469, 400, 51 },
		{ 469, 401, 51 },
		{ 469, 402, 51 },
		{ 469, 403, 51 },
		{ 469, 404, 51 },
		{ 469, 405, 51 },
		{ 469, 406, 51 },
		{ 469, 407, 51 },
		{ 469, 408, 51 },
		{ 469, 409, 51 },
		{ 469, 410, 51 },
		{ 469, 411, 51 },
		{ 469, 412, 51 },
		{ 469, 413, 51 },
		{ 469, 414, 51 },
		{ 469, 415, 51 },
		{ 470, 364, 51 },
		{ 470, 365, 51 },
		{ 470, 366, 51 },
		{ 470, 367, 51 },
		{ 470, 368, 51 },
		{ 470, 369, 51 },
		{ 470, 370, 51 },
		{ 470, 371, 51 },
		{ 470, 372, 51 },
		{ 470, 373, 51 },
		{ 470, 374, 51 },
		{ 470, 375, 51 },
		{ 470, 376, 51 },
		{ 470, 377, 51 },
		{ 470, 378, 51 },
		{ 470, 379, 51 },
		{ 470, 380, 51 },
		{ 470, 381, 51 },
		{ 470, 382, 51 },
		{ 470, 383, 51 },
		{ 470, 384, 51 },
		{ 470, 385, 51 },
		{ 470, 386, 51 },
		{ 470, 387, 51 },
		{ 470, 388, 51 },
		{ 470, 389, 51 },
		{ 470, 390, 51 },
		{ 470, 391, 51 },
		{ 470, 392, 51 },
		{ 470, 393, 51 },
		{ 470, 394, 51 },
		{ 470, 395, 51 },
		{ 470, 396, 51 },
		{ 470, 397, 51 },
		{ 470, 398, 51 },
		{ 470, 399, 51 },
		{ 470, 400, 51 },
		{ 470, 401, 51 },
		{ 470, 402, 51 },
		{ 470, 403, 51 },
		{ 470, 404, 51 },
		{ 470, 405, 51 },
		{ 470, 406, 51 },
		{ 470, 407, 51 },
		{ 470, 408, 51 },
		{ 470, 409, 51 },
		{ 470, 410, 51 },
		{ 470, 411, 51 },
		{ 470, 412, 51 },
		{ 470, 413, 51 },
		{ 470, 414, 51 },
		{ 470, 415, 51 },
		{ 470, 416, 51 },
		{ 471, 365, 51 },
		{ 471, 366, 51 },
		{ 471, 367, 51 },
		{ 471, 368, 51 },
		{ 471, 369, 51 },
		{ 471, 370, 51 },
		{ 471, 371, 51 },
		{ 471, 372, 51 },
		{ 471, 373, 51 },
		{ 471, 374, 51 },
		{ 471, 375, 51 },
		{ 471, 376, 51 },
		{ 471, 377, 51 },
		{ 471, 378, 51 },
		{ 471, 379, 51 },
		{ 471, 380, 51 },
		{ 471, 381, 51 },
		{ 471, 382, 51 },
		{ 471, 383, 51 },
		{ 471, 384, 51 },
		{ 471, 385, 51 },
		{ 471, 386, 51 },
		{ 471, 387, 51 },
		{ 471, 388, 51 },
		{ 471, 389, 51 },
		{ 471, 390, 51 },
		{ 471, 391, 51 },
		{ 471, 392, 51 },
		{ 471, 393, 51 },
		{ 471, 394, 51 },
		{ 471, 395, 51 },
		{ 471, 396, 51 },
		{ 471, 397, 51 },
		{ 471, 398, 51 },
		{ 471, 399, 51 },
		{ 471, 400, 51 },
		{ 471, 401, 51 },
		{ 471, 402, 51 },
		{ 471, 403, 51 },
		{ 471, 404, 51 },
		{ 471, 405, 51 },
		{ 471, 406, 51 },
		{ 471, 407, 51 },
		{ 471, 408, 51 },
		{ 471, 409, 51 },
		{ 471, 410, 51 },
		{ 471, 411, 51 },
		{ 471, 412, 51 },
		{ 471, 413, 51 },
		{ 471, 414, 51 },
		{ 471, 415, 51 },
		{ 471, 416, 51 },
		{ 471, 417, 51 },
		{ 472, 365, 51 },
		{ 472, 366, 51 },
		{ 472, 367, 51 },
		{ 472, 368, 51 },
		{ 472, 369, 51 },
		{ 472, 370, 51 },
		{ 472, 371, 51 },
		{ 472, 372, 51 },
		{ 472, 373, 51 },
		{ 472, 374, 51 },
		{ 472, 375, 51 },
		{ 472, 376, 51 },
		{ 472, 377, 51 },
		{ 472, 378, 51 },
		{ 472, 379, 51 },
		{ 472, 380, 51 },
		{ 472, 381, 51 },
		{ 472, 382, 51 },
		{ 472, 383, 51 },
		{ 472, 384, 51 },
		{ 472, 385, 51 },
		{ 472, 386, 51 },
		{ 472, 387, 51 },
		{ 472, 388, 51 },
		{ 472, 389, 51 },
		{ 472, 390, 51 },
		{ 472, 391, 51 },
		{ 472, 392, 51 },
		{ 472, 393, 51 },
		{ 472, 394, 51 },
		{ 472, 395, 51 },
		{ 472, 396, 51 },
		{ 472, 397, 51 },
		{ 472, 398, 51 },
		{ 472, 399, 51 },
		{ 472, 400, 51 },
		{ 472, 401, 51 },
		{ 472, 402, 51 },
		{ 472, 403, 51 },
		{ 472, 404, 51 },
		{ 472, 405, 51 },
		{ 472, 406, 51 },
		{ 472, 407, 51 },
		{ 472, 408, 51 },
		{ 472, 409, 51 },
		{ 472, 410, 51 },
		{ 472, 411, 51 },
		{ 472, 412, 51 },
		{ 472, 413, 51 },
		{ 472, 414, 51 },
		{ 472, 415, 51 },
		{ 472, 416, 51 },
		{ 472, 417, 51 },
		{ 472, 418, 51 },
		{ 473, 366, 51 },
		{ 473, 367, 51 },
		{ 473, 368, 51 },
		{ 473, 369, 51 },
		{ 473, 370, 51 },
		{ 473, 371, 51 },
		{ 473, 372, 51 },
		{ 473, 373, 51 },
		{ 473, 374, 51 },
		{ 473, 375, 51 },
		{ 473, 376, 51 },
		{ 473, 377, 51 },
		{ 473, 378, 51 },
		{ 473, 379, 51 },
		{ 473, 380, 51 },
		{ 473, 381, 51 },
		{ 473, 382, 51 },
		{ 473, 383, 51 },
		{ 473, 384, 51 },
		{ 473, 385, 51 },
		{ 473, 386, 51 },
		{ 473, 387, 51 },
		{ 473, 388, 51 },
		{ 473, 389, 51 },
		{ 473, 390, 51 },
		{ 473, 391, 51 },
		{ 473, 392, 51 },
		{ 473, 393, 51 },
		{ 473, 394, 51 },
		{ 473, 395, 51 },
		{ 473, 396, 51 },
		{ 473, 397, 51 },
		{ 473, 398, 51 },
		{ 473, 399, 51 },
		{ 473, 400, 51 },
		{ 473, 401, 51 },
		{ 473, 402, 51 },
		{ 473, 403, 51 },
		{ 473, 404, 51 },
		{ 473, 405, 51 },
		{ 473, 406, 51 },
		{ 473, 407, 51 },
		{ 473, 408, 51 },
		{ 473, 409, 51 },
		{ 473, 410, 51 },
		{ 473, 411, 51 },
		{ 473, 412, 51 },
		{ 473, 413, 51 },
		{ 473, 414, 51 },
		{ 473, 415, 51 },
		{ 473, 416, 51 },
		{ 473, 417, 51 },
		{ 473, 418, 51 },
		{ 474, 366, 51 },
		{ 474, 367, 51 },
		{ 474, 368, 51 },
		{ 474, 369, 51 },
		{ 474, 370, 51 },
		{ 474, 371, 51 },
		{ 474, 372, 51 },
		{ 474, 373, 51 },
		{ 474, 374, 51 },
		{ 474, 375, 51 },
		{ 474, 376, 51 },
		{ 474, 377, 51 },
		{ 474, 378, 51 },
		{ 474, 379, 51 },
		{ 474, 380, 51 },
		{ 474, 381, 51 },
		{ 474, 382, 51 },
		{ 474, 383, 51 },
		{ 474, 384, 51 },
		{ 474, 385, 51 },
		{ 474, 386, 51 },
		{ 474, 387, 51 },
		{ 474, 388, 51 },
		{ 474, 389, 51 },
		{ 474, 390, 51 },
		{ 474, 391, 51 },
		{ 474, 392, 51 },
		{ 474, 393, 51 },
		{ 474, 394, 51 },
		{ 474, 395, 51 },
		{ 474, 396, 51 },
		{ 474, 397, 51 },
		{ 474, 398, 51 },
		{ 474, 399, 51 },
		{ 474, 400, 51 },
		{ 474, 401, 51 },
		{ 474, 402, 51 },
		{ 474, 403, 51 },
		{ 474, 404, 51 },
		{ 474, 405, 51 },
		{ 474, 406, 51 },
		{ 474, 407, 51 },
		{ 474, 408, 51 },
		{ 474, 409, 51 },
		{ 474, 410, 51 },
		{ 474, 411, 51 },
		{ 474, 412, 51 },
		{ 474, 413, 51 },
		{ 474, 414, 51 },
		{ 474, 415, 51 },
		{ 474, 416, 51 },
		{ 474, 417, 51 },
		{ 474, 418, 51 },
		{ 474, 419, 51 },
		{ 475, 367, 51 },
		{ 475, 368, 51 },
		{ 475, 369, 51 },
		{ 475, 370, 51 },
		{ 475, 371, 51 },
		{ 475, 372, 51 },
		{ 475, 373, 51 },
		{ 475, 374, 51 },
		{ 475, 375, 51 },
		{ 475, 376, 51 },
		{ 475, 377, 51 },
		{ 475, 378, 51 },
		{ 475, 379, 51 },
		{ 475, 380, 51 },
		{ 475, 381, 51 },
		{ 475, 382, 51 },
		{ 475, 383, 51 },
		{ 475, 384, 51 },
		{ 475, 385, 51 },
		{ 475, 386, 51 },
		{ 475, 387, 51 },
		{ 475, 388, 51 },
		{ 475, 389, 51 },
		{ 475, 390, 51 },
		{ 475, 391, 51 },
		{ 475, 392, 51 },
		{ 475, 393, 51 },
		{ 475, 394, 51 },
		{ 475, 395, 51 },
		{ 475, 396, 51 },
		{ 475, 397, 51 },
		{ 475, 398, 51 },
		{ 475, 399, 51 },
		{ 475, 400, 51 },
		{ 475, 401, 51 },
		{ 475, 402, 51 },
		{ 475, 403, 51 },
		{ 475, 404, 51 },
		{ 475, 405, 51 },
		{ 475, 406, 51 },
		{ 475, 407, 51 },
		{ 475, 408, 51 },
		{ 475, 409, 51 },
		{ 475, 410, 51 },
		{ 475, 411, 51 },
		{ 475, 412, 51 },
		{ 475, 413, 51 },
		{ 475, 414, 51 },
		{ 475, 415, 51 },
		{ 475, 416, 51 },
		{ 475, 417, 51 },
		{ 475, 418, 51 },
		{ 475, 419, 51 },
		{ 475, 420, 51 },
		{ 476, 367, 51 },
		{ 476, 368, 51 },
		{ 476, 369, 51 },
		{ 476, 370, 51 },
		{ 476, 371, 51 },
		{ 476, 372, 51 },
		{ 476, 373, 51 },
		{ 476, 374, 51 },
		{ 476, 375, 51 },
		{ 476, 376, 51 },
		{ 476, 377, 51 },
		{ 476, 378, 51 },
		{ 476, 379, 51 },
		{ 476, 380, 51 },
		{ 476, 381, 51 },
		{ 476, 382, 51 },
		{ 476, 383, 51 },
		{ 476, 384, 51 },
		{ 476, 385, 51 },
		{ 476, 386, 51 },
		{ 476, 387, 51 },
		{ 476, 388, 51 },
		{ 476, 389, 51 },
		{ 476, 390, 51 },
		{ 476, 391, 51 },
		{ 476, 392, 51 },
		{ 476, 393, 51 },
		{ 476, 394, 51 },
		{ 476, 395, 51 },
		{ 476, 396, 51 },
		{ 476, 397, 51 },
		{ 476, 398, 51 },
		{ 476, 399, 51 },
		{ 476, 400, 51 },
		{ 476, 401, 51 },
		{ 476, 402, 51 },
		{ 476, 403, 51 },
		{ 476, 404, 51 },
		{ 476, 405, 51 },
		{ 476, 406, 51 },
		{ 476, 407, 51 },
		{ 476, 408, 51 },
		{ 476, 409, 51 },
		{ 476, 410, 51 },
		{ 476, 411, 51 },
		{ 476, 412, 51 },
		{ 476, 413, 51 },
		{ 476, 414, 51 },
		{ 476, 415, 51 },
		{ 476, 416, 51 },
		{ 476, 417, 51 },
		{ 476, 418, 51 },
		{ 476, 419, 51 },
		{ 476, 420, 51 },
		{ 476, 421, 51 },
		{ 477, 368, 51 },
		{ 477, 369, 51 },
		{ 477, 370, 51 },
		{ 477, 371, 51 },
		{ 477, 372, 51 },
		{ 477, 373, 51 },
		{ 477, 374, 51 },
		{ 477, 375, 51 },
		{ 477, 376, 51 },
		{ 477, 377, 51 },
		{ 477, 378, 51 },
		{ 477, 379, 51 },
		{ 477, 380, 51 },
		{ 477, 381, 51 },
		{ 477, 382, 51 },
		{ 477, 383, 51 },
		{ 477, 384, 51 },
		{ 477, 385, 51 },
		{ 477, 386, 51 },
		{ 477, 387, 51 },
		{ 477, 388, 51 },
		{ 477, 389, 51 },
		{ 477, 390, 51 },
		{ 477, 391, 51 },
		{ 477, 392, 51 },
		{ 477, 393, 51 },
		{ 477, 394, 51 },
		{ 477, 395, 51 },
		{ 477, 396, 51 },
		{ 477, 397, 51 },
		{ 477, 398, 51 },
		{ 477, 399, 51 },
		{ 477, 400, 51 },
		{ 477, 401, 51 },
		{ 477, 402, 51 },
		{ 477, 403, 51 },
		{ 477, 404, 51 },
		{ 477, 405, 51 },
		{ 477, 406, 51 },
		{ 477, 407, 51 },
		{ 477, 408, 51 },
		{ 477, 409, 51 },
		{ 477, 410, 51 },
		{ 477, 411, 51 },
		{ 477, 412, 51 },
		{ 477, 413, 51 },
		{ 477, 414, 51 },
		{ 477, 415, 51 },
		{ 477, 416, 51 },
		{ 477, 417, 51 },
		{ 477, 418, 51 },
		{ 477, 419, 51 },
		{ 477, 420, 51 },
		{ 477, 421, 51 },
		{ 478, 368, 51 },
		{ 478, 369, 51 },
		{ 478, 370, 51 },
		{ 478, 371, 51 },
		{ 478, 372, 51 },
		{ 478, 373, 51 },
		{ 478, 374, 51 },
		{ 478, 375, 51 },
		{ 478, 376, 51 },
		{ 478, 377, 51 },
		{ 478, 378, 51 },
		{ 478, 379, 51 },
		{ 478, 380, 51 },
		{ 478, 381, 51 },
		{ 478, 382, 51 },
		{ 478, 383, 51 },
		{ 478, 384, 51 },
		{ 478, 385, 51 },
		{ 478, 386, 51 },
		{ 478, 387, 51 },
		{ 478, 388, 51 },
		{ 478, 389, 51 },
		{ 478, 390, 51 },
		{ 478, 391, 51 },
		{ 478, 392, 51 },
		{ 478, 393, 51 },
		{ 478, 394, 51 },
		{ 478, 395, 51 },
		{ 478, 396, 51 },
		{ 478, 397, 51 },
		{ 478, 398, 51 },
		{ 478, 399, 51 },
		{ 478, 400, 51 },
		{ 478, 401, 51 },
		{ 478, 402, 51 },
		{ 478, 403, 51 },
		{ 478, 404, 51 },
		{ 478, 405, 51 },
		{ 478, 406, 51 },
		{ 478, 407, 51 },
		{ 478, 408, 51 },
		{ 478, 409, 51 },
		{ 478, 410, 51 },
		{ 478, 411, 51 },
		{ 478, 412, 51 },
		{ 478, 413, 51 },
		{ 478, 414, 51 },
		{ 478, 415, 51 },
		{ 478, 416, 51 },
		{ 478, 417, 51 },
		{ 478, 418, 51 },
		{ 478, 419, 51 },
		{ 478, 420, 51 },
		{ 478, 421, 51 },
		{ 478, 422, 51 },
		{ 479, 369, 51 },
		{ 479, 370, 51 },
		{ 479, 371, 51 },
		{ 479, 372, 51 },
		{ 479, 373, 51 },
		{ 479, 374, 51 },
		{ 479, 375, 51 },
		{ 479, 376, 51 },
		{ 479, 377, 51 },
		{ 479, 378, 51 },
		{ 479, 379, 51 },
		{ 479, 380, 51 },
		{ 479, 381, 51 },
		{ 479, 382, 51 },
		{ 479, 383, 51 },
		{ 479, 384, 51 },
		{ 479, 385, 51 },
		{ 479, 386, 51 },
		{ 479, 387, 51 },
		{ 479, 388, 51 },
		{ 479, 389, 51 },
		{ 479, 390, 51 },
		{ 479, 391, 51 },
		{ 479, 392, 51 },
		{ 479, 393, 51 },
		{ 479, 394, 51 },
		{ 479, 395, 51 },
		{ 479, 396, 51 },
		{ 479, 397, 51 },
		{ 479, 398, 51 },
		{ 479, 399, 51 },
		{ 479, 400, 51 },
		{ 479, 401, 51 },
		{ 479, 402, 51 },
		{ 479, 403, 51 },
		{ 479, 404, 51 },
		{ 479, 405, 51 },
		{ 479, 406, 51 },
		{ 479, 407, 51 },
		{ 479, 408, 51 },
		{ 479, 409, 51 },
		{ 479, 410, 51 },
		{ 479, 411, 51 },
		{ 479, 412, 51 },
		{ 479, 413, 51 },
		{ 479, 414, 51 },
		{ 479, 415, 51 },
		{ 479, 416, 51 },
		{ 479, 417, 51 },
		{ 479, 418, 51 },
		{ 479, 419, 51 },
		{ 479, 420, 51 },
		{ 479, 421, 51 },
		{ 479, 422, 51 },
		{ 479, 423, 51 },
		{ 480, 369, 51 },
		{ 480, 370, 51 },
		{ 480, 371, 51 },
		{ 480, 372, 51 },
		{ 480, 373, 51 },
		{ 480, 374, 51 },
		{ 480, 375, 51 },
		{ 480, 376, 51 },
		{ 480, 377, 51 },
		{ 480, 378, 51 },
		{ 480, 379, 51 },
		{ 480, 380, 51 },
		{ 480, 381, 51 },
		{ 480, 382, 51 },
		{ 480, 383, 51 },
		{ 480, 384, 51 },
		{ 480, 385, 51 },
		{ 480, 386, 51 },
		{ 480, 387, 51 },
		{ 480, 388, 51 },
		{ 480, 389, 51 },
		{ 480, 390, 51 },
		{ 480, 391, 51 },
		{ 480, 392, 51 },
		{ 480, 393, 51 },
		{ 480, 394, 51 },
		{ 480, 395, 51 },
		{ 480, 396, 51 },
		{ 480, 397, 51 },
		{ 480, 398, 51 },
		{ 480, 399, 51 },
		{ 480, 400, 51 },
		{ 480, 401, 51 },
		{ 480, 402, 51 },
		{ 480, 403, 51 },
		{ 480, 404, 51 },
		{ 480, 405, 51 },
		{ 480, 406, 51 },
		{ 480, 407, 51 },
		{ 480, 408, 51 },
		{ 480, 409, 51 },
		{ 480, 410, 51 },
		{ 480, 411, 51 },
		{ 480, 412, 51 },
		{ 480, 413, 51 },
		{ 480, 414, 51 },
		{ 480, 415, 51 },
		{ 480, 416, 51 },
		{ 480, 417, 51 },
		{ 480, 418, 51 },
		{ 480, 419, 51 },
		{ 480, 420, 51 },
		{ 480, 421, 51 },
		{ 480, 422, 51 },
		{ 480, 423, 51 },
		{ 480, 424, 51 },
		{ 481, 370, 51 },
		{ 481, 371, 51 },
		{ 481, 372, 51 },
		{ 481, 373, 51 },
		{ 481, 374, 51 },
		{ 481, 375, 51 },
		{ 481, 376, 51 },
		{ 481, 377, 51 },
		{ 481, 378, 51 },
		{ 481, 379, 51 },
		{ 481, 380, 51 },
		{ 481, 381, 51 },
		{ 481, 382, 51 },
		{ 481, 383, 51 },
		{ 481, 384, 51 },
		{ 481, 385, 51 },
		{ 481, 386, 51 },
		{ 481, 387, 51 },
		{ 481, 388, 51 },
		{ 481, 389, 51 },
		{ 481, 390, 51 },
		{ 481, 391, 51 },
		{ 481, 392, 51 },
		{ 481, 393, 51 },
		{ 481, 394, 51 },
		{ 481, 395, 51 },
		{ 481, 396, 51 },
		{ 481, 397, 51 },
		{ 481, 398, 51 },
		{ 481, 399, 51 },
		{ 481, 400, 51 },
		{ 481, 401, 51 },
		{ 481, 402, 51 },
		{ 481, 403, 51 },
		{ 481, 404, 51 },
		{ 481, 405, 51 },
		{ 481, 406, 51 },
		{ 481, 407, 51 },
		{ 481, 408, 51 },
		{ 481, 409, 51 },
		{ 481, 410, 51 },
		{ 481, 411, 51 },
		{ 481, 412, 51 },
		{ 481, 413, 51 },
		{ 481, 414, 51 },
		{ 481, 415, 51 },
		{ 481, 416, 51 },
		{ 481, 417, 51 },
		{ 481, 418, 51 },
		{ 481, 419, 51 },
		{ 481, 420, 51 },
		{ 481, 421, 51 },
		{ 481, 422, 51 },
		{ 481, 423, 51 },
		{ 481, 424, 51 },
		{ 482, 370, 51 },
		{ 482, 371, 51 },
		{ 482, 372, 51 },
		{ 482, 373, 51 },
		{ 482, 374, 51 },
		{ 482, 375, 51 },
		{ 482, 376, 51 },
		{ 482, 377, 51 },
		{ 482, 378, 51 },
		{ 482, 379, 51 },
		{ 482, 380, 51 },
		{ 482, 381, 51 },
		{ 482, 382, 51 },
		{ 482, 383, 51 },
		{ 482, 384, 51 },
		{ 482, 385, 51 },
		{ 482, 386, 51 },
		{ 482, 387, 51 },
		{ 482, 388, 51 },
		{ 482, 389, 51 },
		{ 482, 390, 51 },
		{ 482, 391, 51 },
		{ 482, 392, 51 },
		{ 482, 393, 51 },
		{ 482, 394, 51 },
		{ 482, 395, 51 },
		{ 482, 396, 51 },
		{ 482, 397, 51 },
		{ 482, 398, 51 },
		{ 482, 399, 51 },
		{ 482, 400, 51 },
		{ 482, 401, 51 },
		{ 482, 402, 51 },
		{ 482, 403, 51 },
		{ 482, 404, 51 },
		{ 482, 405, 51 },
		{ 482, 406, 51 },
		{ 482, 407, 51 },
		{ 482, 408, 51 },
		{ 482, 409, 51 },
		{ 482, 410, 51 },
		{ 482, 411, 51 },
		{ 482, 412, 51 },
		{ 482, 413, 51 },
		{ 482, 414, 51 },
		{ 482, 415, 51 },
		{ 482, 416, 51 },
		{ 482, 417, 51 },
		{ 482, 418, 51 },
		{ 482, 419, 51 },
		{ 482, 420, 51 },
		{ 482, 421, 51 },
		{ 482, 422, 51 },
		{ 482, 423, 51 },
		{ 482, 424, 51 },
		{ 482, 425, 51 },
		{ 483, 371, 51 },
		{ 483, 372, 51 },
		{ 483, 373, 51 },
		{ 483, 374, 51 },
		{ 483, 375, 51 },
		{ 483, 376, 51 },
		{ 483, 377, 51 },
		{ 483, 378, 51 },
		{ 483, 379, 51 },
		{ 483, 380, 51 },
		{ 483, 381, 51 },
		{ 483, 382, 51 },
		{ 483, 383, 51 },
		{ 483, 384, 51 },
		{ 483, 385, 51 },
		{ 483, 386, 51 },
		{ 483, 387, 51 },
		{ 483, 388, 51 },
		{ 483, 389, 51 },
		{ 483, 390, 51 },
		{ 483, 391, 51 },
		{ 483, 392, 51 },
		{ 483, 393, 51 },
		{ 483, 394, 51 },
		{ 483, 395, 51 },
		{ 483, 396, 51 },
		{ 483, 397, 51 },
		{ 483, 398, 51 },
		{ 483, 399, 51 },
		{ 483, 400, 51 },
		{ 483, 401, 51 },
		{ 483, 402, 51 },
		{ 483, 403, 51 },
		{ 483, 404, 51 },
		{ 483, 405, 51 },
		{ 483, 406, 51 },
		{ 483, 407, 51 },
		{ 483, 408, 51 },
		{ 483, 409, 51 },
		{ 483, 410, 51 },
		{ 483, 411, 51 },
		{ 483, 412, 51 },
		{ 483, 413, 51 },
		{ 483, 414, 51 },
		{ 483, 415, 51 },
		{ 483, 416, 51 },
		{ 483, 417, 51 },
		{ 483, 418, 51 },
		{ 483, 419, 51 },
		{ 483, 420, 51 },
		{ 483, 421, 51 },
		{ 483, 422, 51 },
		{ 483, 423, 51 },
		{ 483, 424, 51 },
		{ 483, 425, 51 },
		{ 483, 426, 51 },
		{ 484, 371, 51 },
		{ 484, 372, 51 },
		{ 484, 373, 51 },
		{ 484, 374, 51 },
		{ 484, 375, 51 },
		{ 484, 376, 51 },
		{ 484, 377, 51 },
		{ 484, 378, 51 },
		{ 484, 379, 51 },
		{ 484, 380, 51 },
		{ 484, 381, 51 },
		{ 484, 382, 51 },
		{ 484, 383, 51 },
		{ 484, 384, 51 },
		{ 484, 385, 51 },
		{ 484, 386, 51 },
		{ 484, 387, 51 },
		{ 484, 388, 51 },
		{ 484, 389, 51 },
		{ 484, 390, 51 },
		{ 484, 391, 51 },
		{ 484, 392, 51 },
		{ 484, 393, 51 },
		{ 484, 394, 51 },
		{ 484, 395, 51 },
		{ 484, 396, 51 },
		{ 484, 397, 51 },
		{ 484, 398, 51 },
		{ 484, 399, 51 },
		{ 484, 400, 51 },
		{ 484, 401, 51 },
		{ 484, 402, 51 },
		{ 484, 403, 51 },
		{ 484, 404, 51 },
		{ 484, 405, 51 },
		{ 484, 406, 51 },
		{ 484, 407, 51 },
		{ 484, 408, 51 },
		{ 484, 409, 51 },
		{ 484, 410, 51 },
		{ 484, 411, 51 },
		{ 484, 412, 51 },
		{ 484, 413, 51 },
		{ 484, 414, 51 },
		{ 484, 415, 51 },
		{ 484, 416, 51 },
		{ 484, 417, 51 },
		{ 484, 418, 51 },
		{ 484, 419, 51 },
		{ 484, 420, 51 },
		{ 484, 421, 51 },
		{ 484, 422, 51 },
		{ 484, 423, 51 },
		{ 484, 424, 51 },
		{ 484, 425, 51 },
		{ 484, 426, 51 },
		{ 484, 427, 51 },
		{ 485, 372, 51 },
		{ 485, 373, 51 },
		{ 485, 374, 51 },
		{ 485, 375, 51 },
		{ 485, 376, 51 },
		{ 485, 377, 51 },
		{ 485, 378, 51 },
		{ 485, 379, 51 },
		{ 485, 380, 51 },
		{ 485, 381, 51 },
		{ 485, 382, 51 },
		{ 485, 383, 51 },
		{ 485, 384, 51 },
		{ 485, 385, 51 },
		{ 485, 386, 51 },
		{ 485, 387, 51 },
		{ 485, 388, 51 },
		{ 485, 389, 51 },
		{ 485, 390, 51 },
		{ 485, 391, 51 },
		{ 485, 392, 51 },
		{ 485, 393, 51 },
		{ 485, 394, 51 },
		{ 485, 395, 51 },
		{ 485, 396, 51 },
		{ 485, 397, 51 },
		{ 485, 398, 51 },
		{ 485, 399, 51 },
		{ 485, 400, 51 },
		{ 485, 401, 51 },
		{ 485, 402, 51 },
		{ 485, 403, 51 },
		{ 485, 404, 51 },
		{ 485, 405, 51 },
		{ 485, 406, 51 },
		{ 485, 407, 51 },
		{ 485, 408, 51 },
		{ 485, 409, 51 },
		{ 485, 410, 51 },
		{ 485, 411, 51 },
		{ 485, 412, 51 },
		{ 485, 413, 51 },
		{ 485, 414, 51 },
		{ 485, 415, 51 },
		{ 485, 416, 51 },
		{ 485, 417, 51 },
		{ 485, 418, 51 },
		{ 485, 419, 51 },
		{ 485, 420, 51 },
		{ 485, 421, 51 },
		{ 485, 422, 51 },
		{ 485, 423, 51 },
		{ 485, 424, 51 },
		{ 485, 425, 51 },
		{ 485, 426, 51 },
		{ 485, 427, 51 },
		{ 486, 372, 51 },
		{ 486, 373, 51 },
		{ 486, 374, 51 },
		{ 486, 375, 51 },
		{ 486, 376, 51 },
		{ 486, 377, 51 },
		{ 486, 378, 51 },
		{ 486, 379, 51 },
		{ 486, 380, 51 },
		{ 486, 381, 51 },
		{ 486, 382, 51 },
		{ 486, 383, 51 },
		{ 486, 384, 51 },
		{ 486, 385, 51 },
		{ 486, 386, 51 },
		{ 486, 387, 51 },
		{ 486, 388, 51 },
		{ 486, 389, 51 },
		{ 486, 390, 51 },
		{ 486, 391, 51 },
		{ 486, 392, 51 },
		{ 486, 393, 51 },
		{ 486, 394, 51 },
		{ 486, 395, 51 },
		{ 486, 396, 51 },
		{ 486, 397, 51 },
		{ 486, 398, 51 },
		{ 486, 399, 51 },
		{ 486, 400, 51 },
		{ 486, 401, 51 },
		{ 486, 402, 51 },
		{ 486, 403, 51 },
		{ 486, 404, 51 },
		{ 486, 405, 51 },
		{ 486, 406, 51 },
		{ 486, 407, 51 },
		{ 486, 408, 51 },
		{ 486, 409, 51 },
		{ 486, 410, 51 },
		{ 486, 411, 51 },
		{ 486, 412, 51 },
		{ 486, 413, 51 },
		{ 486, 414, 51 },
		{ 486, 415, 51 },
		{ 486, 416, 51 },
		{ 486, 417, 51 },
		{ 486, 418, 51 },
		{ 486, 419, 51 },
		{ 486, 420, 51 },
		{ 486, 421, 51 },
		{ 486, 422, 51 },
		{ 486, 423, 51 },
		{ 486, 424, 51 },
		{ 486, 425, 51 },
		{ 486, 426, 51 },
		{ 486, 427, 51 },
		{ 486, 428, 51 },
		{ 487, 373, 51 },
		{ 487, 374, 51 },
		{ 487, 375, 51 },
		{ 487, 376, 51 },
		{ 487, 377, 51 },
		{ 487, 378, 51 },
		{ 487, 379, 51 },
		{ 487, 380, 51 },
		{ 487, 381, 51 },
		{ 487, 382, 51 },
		{ 487, 383, 51 },
		{ 487, 384, 51 },
		{ 487, 385, 51 },
		{ 487, 386, 51 },
		{ 487, 387, 51 },
		{ 487, 388, 51 },
		{ 487, 389, 51 },
		{ 487, 390, 51 },
		{ 487, 391, 51 },
		{ 487, 392, 51 },
		{ 487, 393, 51 },
		{ 487, 394, 51 },
		{ 487, 395, 51 },
		{ 487, 396, 51 },
		{ 487, 397, 51 },
		{ 487, 398, 51 },
		{ 487, 399, 51 },
		{ 487, 400, 51 },
		{ 487, 401, 51 },
		{ 487, 402, 51 },
		{ 487, 403, 51 },
		{ 487, 404, 51 },
		{ 487, 405, 51 },
		{ 487, 406, 51 },
		{ 487, 407, 51 },
		{ 487, 408, 51 },
		{ 487, 409, 51 },
		{ 487, 410, 51 },
		{ 487, 411, 51 },
		{ 487, 412, 51 },
		{ 487, 413, 51 },
		{ 487, 414, 51 },
		{ 487, 415, 51 },
		{ 487, 416, 51 },
		{ 487, 417, 51 },
		{ 487, 418, 51 },
		{ 487, 419, 51 },
		{ 487, 420, 51 },
		{ 487, 421, 51 },
		{ 487, 422, 51 },
		{ 487, 423, 51 },
		{ 487, 424, 51 },
		{ 487, 425, 51 },
		{ 487, 426, 51 },
		{ 487, 427, 51 },
		{ 487, 428, 51 },
		{ 487, 429, 51 },
		{ 488, 373, 51 },
		{ 488, 374, 51 },
		{ 488, 375, 51 },
		{ 488, 376, 51 },
		{ 488, 377, 51 },
		{ 488, 378, 51 },
		{ 488, 379, 51 },
		{ 488, 380, 51 },
		{ 488, 381, 51 },
		{ 488, 382, 51 },
		{ 488, 383, 51 },
		{ 488, 384, 51 },
		{ 488, 385, 51 },
		{ 488, 386, 51 },
		{ 488, 387, 51 },
		{ 488, 388, 51 },
		{ 488, 389, 51 },
		{ 488, 390, 51 },
		{ 488, 391, 51 },
		{ 488, 392, 51 },
		{ 488, 393, 51 },
		{ 488, 394, 51 },
		{ 488, 395, 51 },
		{ 488, 396, 51 },
		{ 488, 397, 51 },
		{ 488, 398, 51 },
		{ 488, 399, 51 },
		{ 488, 400, 51 },
		{ 488, 401, 51 },
		{ 488, 402, 51 },
		{ 488, 403, 51 },
		{ 488, 404, 51 },
		{ 488, 405, 51 },
		{ 488, 406, 51 },
		{ 488, 407, 51 },
		{ 488, 408, 51 },
		{ 488, 409, 51 },
		{ 488, 410, 51 },
		{ 488, 411, 51 },
		{ 488, 412, 51 },
		{ 488, 413, 51 },
		{ 488, 414, 51 },
		{ 488, 415, 51 },
		{ 488, 416, 51 },
		{ 488, 417, 51 },
		{ 488, 418, 51 },
		{ 488, 419, 51 },
		{ 488, 420, 51 },
		{ 488, 421, 51 },
		{ 488, 422, 51 },
		{ 488, 423, 51 },
		{ 488, 424, 51 },
		{ 488, 425, 51 },
		{ 488, 426, 51 },
		{ 488, 427, 51 },
		{ 488, 428, 51 },
		{ 488, 429, 51 },
		{ 488, 430, 51 },
		{ 489, 374, 51 },
		{ 489, 375, 51 },
		{ 489, 376, 51 },
		{ 489, 377, 51 },
		{ 489, 378, 51 },
		{ 489, 379, 51 },
		{ 489, 380, 51 },
		{ 489, 381, 51 },
		{ 489, 382, 51 },
		{ 489, 383, 51 },
		{ 489, 384, 51 },
		{ 489, 385, 51 },
		{ 489, 386, 51 },
		{ 489, 387, 51 },
		{ 489, 388, 51 },
		{ 489, 389, 51 },
		{ 489, 390, 51 },
		{ 489, 391, 51 },
		{ 489, 392, 51 },
		{ 489, 393, 51 },
		{ 489, 394, 51 },
		{ 489, 395, 51 },
		{ 489, 396, 51 },
		{ 489, 397, 51 },
		{ 489, 398, 51 },
		{ 489, 399, 51 },
		{ 489, 400, 51 },
		{ 489, 401, 51 },
		{ 489, 402, 51 },
		{ 489, 403, 51 },
		{ 489, 404, 51 },
		{ 489, 405, 51 },
		{ 489, 406, 51 },
		{ 489, 407, 51 },
		{ 489, 408, 51 },
		{ 489, 409, 51 },
		{ 489, 410, 51 },
		{ 489, 411, 51 },
		{ 489, 412, 51 },
		{ 489, 413, 51 },
		{ 489, 414, 51 },
		{ 489, 415, 51 },
		{ 489, 416, 51 },
		{ 489, 417, 51 },
		{ 489, 418, 51 },
		{ 489, 419, 51 },
		{ 489, 420, 51 },
		{ 489, 421, 51 },
		{ 489, 422, 51 },
		{ 489, 423, 51 },
		{ 489, 424, 51 },
		{ 489, 425, 51 },
		{ 489, 426, 51 },
		{ 489, 427, 51 },
		{ 489, 428, 51 },
		{ 489, 429, 51 },
		{ 489, 430, 51 },
		{ 490, 374, 51 },
		{ 490, 375, 51 },
		{ 490, 376, 51 },
		{ 490, 377, 51 },
		{ 490, 378, 51 },
		{ 490, 379, 51 },
		{ 490, 380, 51 },
		{ 490, 381, 51 },
		{ 490, 382, 51 },
		{ 490, 383, 51 },
		{ 490, 384, 51 },
		{ 490, 385, 51 },
		{ 490, 386, 51 },
		{ 490, 387, 51 },
		{ 490, 388, 51 },
		{ 490, 389, 51 },
		{ 490, 390, 51 },
		{ 490, 391, 51 },
		{ 490, 392, 51 },
		{ 490, 393, 51 },
		{ 490, 394, 51 },
		{ 490, 395, 51 },
		{ 490, 396, 51 },
		{ 490, 397, 51 },
		{ 490, 398, 51 },
		{ 490, 399, 51 },
		{ 490, 400, 51 },
		{ 490, 401, 51 },
		{ 490, 402, 51 },
		{ 490, 403, 51 },
		{ 490, 404, 51 },
		{ 490, 405, 51 },
		{ 490, 406, 51 },
		{ 490, 407, 51 },
		{ 490, 408, 51 },
		{ 490, 409, 51 },
		{ 490, 410, 51 },
		{ 490, 411, 51 },
		{ 490, 412, 51 },
		{ 490, 413, 51 },
		{ 490, 414, 51 },
		{ 490, 415, 51 },
		{ 490, 416, 51 },
		{ 490, 417, 51 },
		{ 490, 418, 51 },
		{ 490, 419, 51 },
		{ 490, 420, 51 },
		{ 490, 421, 51 },
		{ 490, 422, 51 },
		{ 490, 423, 51 },
		{ 490, 424, 51 },
		{ 490, 425, 51 },
		{ 490, 426, 51 },
		{ 490, 427, 51 },
		{ 490, 428, 51 },
		{ 490, 429, 51 },
		{ 490, 430, 51 },
		{ 490, 431, 51 },
		{ 491, 375, 51 },
		{ 491, 376, 51 },
		{ 491, 377, 51 },
		{ 491, 378, 51 },
		{ 491, 379, 51 },
		{ 491, 380, 51 },
		{ 491, 381, 51 },
		{ 491, 382, 51 },
		{ 491, 383, 51 },
		{ 491, 384, 51 },
		{ 491, 385, 51 },
		{ 491, 386, 51 },
		{ 491, 387, 51 },
		{ 491, 388, 51 },
		{ 491, 389, 51 },
		{ 491, 390, 51 },
		{ 491, 391, 51 },
		{ 491, 392, 51 },
		{ 491, 393, 51 },
		{ 491, 394, 51 },
		{ 491, 395, 51 },
		{ 491, 396, 51 },
		{ 491, 397, 51 },
		{ 491, 398, 51 },
		{ 491, 399, 51 },
		{ 491, 400, 51 },
		{ 491, 401, 51 },
		{ 491, 402, 51 },
		{ 491, 403, 51 },
		{ 491, 404, 51 },
		{ 491, 405, 51 },
		{ 491, 406, 51 },
		{ 491, 407, 51 },
		{ 491, 408, 51 },
		{ 491, 409, 51 },
		{ 491, 410, 51 },
		{ 491, 411, 51 },
		{ 491, 412, 51 },
		{ 491, 413, 51 },
		{ 491, 414, 51 },
		{ 491, 415, 51 },
		{ 491, 416, 51 },
		{ 491, 417, 51 },
		{ 491, 418, 51 },
		{ 491, 419, 51 },
		{ 491, 420, 51 },
		{ 491, 421, 51 },
		{ 491, 422, 51 },
		{ 491, 423, 51 },
		{ 491, 424, 51 },
		{ 491, 425, 51 },
		{ 491, 426, 51 },
		{ 491, 427, 51 },
		{ 491, 428, 51 },
		{ 491, 429, 51 },
		{ 491, 430, 51 },
		{ 491, 431, 51 },
		{ 491, 432, 51 },
		{ 492, 375, 51 },
		{ 492, 376, 51 },
		{ 492, 377, 51 },
		{ 492, 378, 51 },
		{ 492, 379, 51 },
		{ 492, 380, 51 },
		{ 492, 381, 51 },
		{ 492, 382, 51 },
		{ 492, 383, 51 },
		{ 492, 384, 51 },
		{ 492, 385, 51 },
		{ 492, 386, 51 },
		{ 492, 387, 51 },
		{ 492, 388, 51 },
		{ 492, 389, 51 },
		{ 492, 390, 51 },
		{ 492, 391, 51 },
		{ 492, 392, 51 },
		{ 492, 393, 51 },
		{ 492, 394, 51 },
		{ 492, 395, 51 },
		{ 492, 396, 51 },
		{ 492, 397, 51 },
		{ 492, 398, 51 },
		{ 492, 399, 51 },
		{ 492, 400, 51 },
		{ 492, 401, 51 },
		{ 492, 402, 51 },
		{ 492, 403, 51 },
		{ 492, 404, 51 },
		{ 492, 405, 51 },
		{ 492, 406, 51 },
		{ 492, 407, 51 },
		{ 492, 408, 51 },
		{ 492, 409, 51 },
		{ 492, 410, 51 },
		{ 492, 411, 51 },
		{ 492, 412, 51 },
		{ 492, 413, 51 },
		{ 492, 414, 51 },
		{ 492, 415, 51 },
		{ 492, 416, 51 },
		{ 492, 417, 51 },
		{ 492, 418, 51 },
		{ 492, 419, 51 },
		{ 492, 420, 51 },
		{ 492, 421, 51 },
		{ 492, 422, 51 },
		{ 492, 423, 51 },
		{ 492, 424, 51 },
		{ 492, 425, 51 },
		{ 492, 426, 51 },
		{ 492, 427, 51 },
		{ 492, 428, 51 },
		{ 492, 429, 51 },
		{ 492, 430, 51 },
		{ 492, 431, 51 },
		{ 492, 432, 51 },
		{ 492, 433, 51 },
		{ 493, 376, 51 },
		{ 493, 377, 51 },
		{ 493, 378, 51 },
		{ 493, 379, 51 },
		{ 493, 380, 51 },
		{ 493, 381, 51 },
		{ 493, 382, 51 },
		{ 493, 383, 51 },
		{ 493, 384, 51 },
		{ 493, 385, 51 },
		{ 493, 386, 51 },
		{ 493, 387, 51 },
		{ 493, 388, 51 },
		{ 493, 389, 51 },
		{ 493, 390, 51 },
		{ 493, 391, 51 },
		{ 493, 392, 51 },
		{ 493, 393, 51 },
		{ 493, 394, 51 },
		{ 493, 395, 51 },
		{ 493, 396, 51 },
		{ 493, 397, 51 },
		{ 493, 398, 51 },
		{ 493, 399, 51 },
		{ 493, 400, 51 },
		{ 493, 401, 51 },
		{ 493, 402, 51 },
		{ 493, 403, 51 },
		{ 493, 404, 51 },
		{ 493, 405, 51 },
		{ 493, 406, 51 },
		{ 493, 407, 51 },
		{ 493, 408, 51 },
		{ 493, 409, 51 },
		{ 493, 410, 51 },
		{ 493, 411, 51 },
		{ 493, 412, 51 },
		{ 493, 413, 51 },
		{ 493, 414, 51 },
		{ 493, 415, 51 },
		{ 493, 416, 51 },
		{ 493, 417, 51 },
		{ 493, 418, 51 },
		{ 493, 419, 51 },
		{ 493, 420, 51 },
		{ 493, 421, 51 },
		{ 493, 422, 51 },
		{ 493, 423, 51 },
		{ 493, 424, 51 },
		{ 493, 425, 51 },
		{ 493, 426, 51 },
		{ 493, 427, 51 },
		{ 493, 428, 51 },
		{ 493, 429, 51 },
		{ 493, 430, 51 },
		{ 493, 431, 51 },
		{ 493, 432, 51 },
		{ 493, 433, 51 },
		{ 494, 376, 51 },
		{ 494, 377, 51 },
		{ 494, 378, 51 },
		{ 494, 379, 51 },
		{ 494, 380, 51 },
		{ 494, 381, 51 },
		{ 494, 382, 51 },
		{ 494, 383, 51 },
		{ 494, 384, 51 },
		{ 494, 385, 51 },
		{ 494, 386, 51 },
		{ 494, 387, 51 },
		{ 494, 388, 51 },
		{ 494, 389, 51 },
		{ 494, 390, 51 },
		{ 494, 391, 51 },
		{ 494, 392, 51 },
		{ 494, 393, 51 },
		{ 494, 394, 51 },
		{ 494, 395, 51 },
		{ 494, 396, 51 },
		{ 494, 397, 51 },
		{ 494, 398, 51 },
		{ 494, 399, 51 },
		{ 494, 400, 51 },
		{ 494, 401, 51 },
		{ 494, 402, 51 },
		{ 494, 403, 51 },
		{ 494, 404, 51 },
		{ 494, 405, 51 },
		{ 494, 406, 51 },
		{ 494, 407, 51 },
		{ 494, 408, 51 },
		{ 494, 409, 51 },
		{ 494, 410, 51 },
		{ 494, 411, 51 },
		{ 494, 412, 51 },
		{ 494, 413, 51 },
		{ 494, 414, 51 },
		{ 494, 415, 51 },
		{ 494, 416, 51 },
		{ 494, 417, 51 },
		{ 494, 418, 51 },
		{ 494, 419, 51 },
		{ 494, 420, 51 },
		{ 494, 421, 51 },
		{ 494, 422, 51 },
		{ 494, 423, 51 },
		{ 494, 424, 51 },
		{ 494, 425, 51 },
		{ 494, 426, 51 },
		{ 494, 427, 51 },
		{ 494, 428, 51 },
		{ 494, 429, 51 },
		{ 494, 430, 51 },
		{ 494, 431, 51 },
		{ 494, 432, 51 },
		{ 494, 433, 51 },
		{ 494, 434, 51 },
		{ 495, 377, 51 },
		{ 495, 378, 51 },
		{ 495, 379, 51 },
		{ 495, 380, 51 },
		{ 495, 381, 51 },
		{ 495, 382, 51 },
		{ 495, 383, 51 },
		{ 495, 384, 51 },
		{ 495, 385, 51 },
		{ 495, 386, 51 },
		{ 495, 387, 51 },
		{ 495, 388, 51 },
		{ 495, 389, 51 },
		{ 495, 390, 51 },
		{ 495, 391, 51 },
		{ 495, 392, 51 },
		{ 495, 393, 51 },
		{ 495, 394, 51 },
		{ 495, 395, 51 },
		{ 495, 396, 51 },
		{ 495, 397, 51 },
		{ 495, 398, 51 },
		{ 495, 399, 51 },
		{ 495, 400, 51 },
		{ 495, 401, 51 },
		{ 495, 402, 51 },
		{ 495, 403, 51 },
		{ 495, 404, 51 },
		{ 495, 405, 51 },
		{ 495, 406, 51 },
		{ 495, 407, 51 },
		{ 495, 408, 51 },
		{ 495, 409, 51 },
		{ 495, 410, 51 },
		{ 495, 411, 51 },
		{ 495, 412, 51 },
		{ 495, 413, 51 },
		{ 495, 414, 51 },
		{ 495, 415, 51 },
		{ 495, 416, 51 },
		{ 495, 417, 51 },
		{ 495, 418, 51 },
		{ 495, 419, 51 },
		{ 495, 420, 51 },
		{ 495, 421, 51 },
		{ 495, 422, 51 },
		{ 495, 423, 51 },
		{ 495, 424, 51 },
		{ 495, 425, 51 },
		{ 495, 426, 51 },
		{ 495, 427, 51 },
		{ 495, 428, 51 },
		{ 495, 429, 51 },
		{ 495, 430, 51 },
		{ 495, 431, 51 },
		{ 495, 432, 51 },
		{ 495, 433, 51 },
		{ 495, 434, 51 },
		{ 495, 435, 51 },
		{ 496, 377, 51 },
		{ 496, 378, 51 },
		{ 496, 379, 51 },
		{ 496, 380, 51 },
		{ 496, 381, 51 },
		{ 496, 382, 51 },
		{ 496, 383, 51 },
		{ 496, 384, 51 },
		{ 496, 385, 51 },
		{ 496, 386, 51 },
		{ 496, 387, 51 },
		{ 496, 388, 51 },
		{ 496, 389, 51 },
		{ 496, 390, 51 },
		{ 496, 391, 51 },
		{ 496, 392, 51 },
		{ 496, 393, 51 },
		{ 496, 394, 51 },
		{ 496, 395, 51 },
		{ 496, 396, 51 },
		{ 496, 397, 51 },
		{ 496, 398, 51 },
		{ 496, 399, 51 },
		{ 496, 400, 51 },
		{ 496, 401, 51 },
		{ 496, 402, 51 },
		{ 496, 403, 51 },
		{ 496, 404, 51 },
		{ 496, 405, 51 },
		{ 496, 406, 51 },
		{ 496, 407, 51 },
		{ 496, 408, 51 },
		{ 496, 409, 51 },
		{ 496, 410, 51 },
		{ 496, 411, 51 },
		{ 496, 412, 51 },
		{ 496, 413, 51 },
		{ 496, 414, 51 },
		{ 496, 415, 51 },
		{ 496, 416, 51 },
		{ 496, 417, 51 },
		{ 496, 418, 51 },
		{ 496, 419, 51 },
		{ 496, 420, 51 },
		{ 496, 421, 51 },
		{ 496, 422, 51 },
		{ 496, 423, 51 },
		{ 496, 424, 51 },
		{ 496, 425, 51 },
		{ 496, 426, 51 },
		{ 496, 427, 51 },
		{ 496, 428, 51 },
		{ 496, 429, 51 },
		{ 496, 430, 51 },
		{ 496, 431, 51 },
		{ 496, 432, 51 },
		{ 496, 433, 51 },
		{ 496, 434, 51 },
		{ 496, 435, 51 },
		{ 496, 436, 51 },
		{ 497, 378, 51 },
		{ 497, 379, 51 },
		{ 497, 380, 51 },
		{ 497, 381, 51 },
		{ 497, 382, 51 },
		{ 497, 383, 51 },
		{ 497, 384, 51 },
		{ 497, 385, 51 },
		{ 497, 386, 51 },
		{ 497, 387, 51 },
		{ 497, 388, 51 },
		{ 497, 389, 51 },
		{ 497, 390, 51 },
		{ 497, 391, 51 },
		{ 497, 392, 51 },
		{ 497, 393, 51 },
		{ 497, 394, 51 },
		{ 497, 395, 51 },
		{ 497, 396, 51 },
		{ 497, 397, 51 },
		{ 497, 398, 51 },
		{ 497, 399, 51 },
		{ 497, 400, 51 },
		{ 497, 401, 51 },
		{ 497, 402, 51 },
		{ 497, 403, 51 },
		{ 497, 404, 51 },
		{ 497, 405, 51 },
		{ 497, 406, 51 },
		{ 497, 407, 51 },
		{ 497, 408, 51 },
		{ 497, 409, 51 },
		{ 497, 410, 51 },
		{ 497, 411, 51 },
		{ 497, 412, 51 },
		{ 497, 413, 51 },
		{ 497, 414, 51 },
		{ 497, 415, 51 },
		{ 497, 416, 51 },
		{ 497, 417, 51 },
		{ 497, 418, 51 },
		{ 497, 419, 51 },
		{ 497, 420, 51 },
		{ 497, 421, 51 },
		{ 497, 422, 51 },
		{ 497, 423, 51 },
		{ 497, 424, 51 },
		{ 497, 425, 51 },
		{ 497, 426, 51 },
		{ 497, 427, 51 },
		{ 497, 428, 51 },
		{ 497, 429, 51 },
		{ 497, 430, 51 },
		{ 497, 431, 51 },
		{ 497, 432, 51 },
		{ 497, 433, 51 },
		{ 497, 434, 51 },
		{ 497, 435, 51 },
		{ 497, 436, 51 },
		{ 498, 378, 51 },
		{ 498, 379, 51 },
		{ 498, 380, 51 },
		{ 498, 381, 51 },
		{ 498, 382, 51 },
		{ 498, 383, 51 },
		{ 498, 384, 51 },
		{ 498, 385, 51 },
		{ 498, 386, 51 },
		{ 498, 387, 51 },
		{ 498, 388, 51 },
		{ 498, 389, 51 },
		{ 498, 390, 51 },
		{ 498, 391, 51 },
		{ 498, 392, 51 },
		{ 498, 393, 51 },
		{ 498, 394, 51 },
		{ 498, 395, 51 },
		{ 498, 396, 51 },
		{ 498, 397, 51 },
		{ 498, 398, 51 },
		{ 498, 399, 51 },
		{ 498, 400, 51 },
		{ 498, 401, 51 },
		{ 498, 402, 51 },
		{ 498, 403, 51 },
		{ 498, 404, 51 },
		{ 498, 405, 51 },
		{ 498, 406, 51 },
		{ 498, 407, 51 },
		{ 498, 408, 51 },
		{ 498, 409, 51 },
		{ 498, 410, 51 },
		{ 498, 411, 51 },
		{ 498, 412, 51 },
		{ 498, 413, 51 },
		{ 498, 414, 51 },
		{ 498, 415, 51 },
		{ 498, 416, 51 },
		{ 498, 417, 51 },
		{ 498, 418, 51 },
		{ 498, 419, 51 },
		{ 498, 420, 51 },
		{ 498, 421, 51 },
		{ 498, 422, 51 },
		{ 498, 423, 51 },
		{ 498, 424, 51 },
		{ 498, 425, 51 },
		{ 498, 426, 51 },
		{ 498, 427, 51 },
		{ 498, 428, 51 },
		{ 498, 429, 51 },
		{ 498, 430, 51 },
		{ 498, 431, 51 },
		{ 498, 432, 51 },
		{ 498, 433, 51 },
		{ 498, 434, 51 },
		{ 498, 435, 51 },
		{ 498, 436, 51 },
		{ 498, 437, 51 },
		{ 499, 379, 51 },
		{ 499, 380, 51 },
		{ 499, 381, 51 },
		{ 499, 382, 51 },
		{ 499, 383, 51 },
		{ 499, 384, 51 },
		{ 499, 385, 51 },
		{ 499, 386, 51 },
		{ 499, 387, 51 },
		{ 499, 388, 51 },
		{ 499, 389, 51 },
		{ 499, 390, 51 },
		{ 499, 391, 51 },
		{ 499, 392, 51 },
		{ 499, 393, 51 },
		{ 499, 394, 51 },
		{ 499, 395, 51 },
		{ 499, 396, 51 },
		{ 499, 397, 51 },
		{ 499, 398, 51 },
		{ 499, 399, 51 },
		{ 499, 400, 51 },
		{ 499, 401, 51 },
		{ 499, 402, 51 },
		{ 499, 403, 51 },
		{ 499, 404, 51 },
		{ 499, 405, 51 },
		{ 499, 406, 51 },
		{ 499, 407, 51 },
		{ 499, 408, 51 },
		{ 499, 409, 51 },
		{ 499, 410, 51 },
		{ 499, 411, 51 },
		{ 499, 412, 51 },
		{ 499, 413, 51 },
		{ 499, 414, 51 },
		{ 499, 415, 51 },
		{ 499, 416, 51 },
		{ 499, 417, 51 },
		{ 499, 418, 51 },
		{ 499, 419, 51 },
		{ 499, 420, 51 },
		{ 499, 421, 51 },
		{ 499, 422, 51 },
		{ 499, 423, 51 },
		{ 499, 424, 51 },
		{ 499, 425, 51 },
		{ 499, 426, 51 },
		{ 499, 427, 51 },
		{ 499, 428, 51 },
		{ 499, 429, 51 },
		{ 499, 430, 51 },
		{ 499, 431, 51 },
		{ 499, 432, 51 },
		{ 499, 433, 51 },
		{ 499, 434, 51 },
		{ 499, 435, 51 },
		{ 499, 436, 51 },
		{ 499, 437, 51 },
		{ 499, 438, 51 },
		{ 500, 379, 51 },
		{ 500, 380, 51 },
		{ 500, 381, 51 },
		{ 500, 382, 51 },
		{ 500, 383, 51 },
		{ 500, 384, 51 },
		{ 500, 385, 51 },
		{ 500, 386, 51 },
		{ 500, 387, 51 },
		{ 500, 388, 51 },
		{ 500, 389, 51 },
		{ 500, 390, 51 },
		{ 500, 391, 51 },
		{ 500, 392, 51 },
		{ 500, 393, 51 },
		{ 500, 394, 51 },
		{ 500, 395, 51 },
		{ 500, 396, 51 },
		{ 500, 397, 51 },
		{ 500, 398, 51 },
		{ 500, 399, 51 },
		{ 500, 400, 51 },
		{ 500, 401, 51 },
		{ 500, 402, 51 },
		{ 500, 403, 51 },
		{ 500, 404, 51 },
		{ 500, 405, 51 },
		{ 500, 406, 51 },
		{ 500, 407, 51 },
		{ 500, 408, 51 },
		{ 500, 409, 51 },
		{ 500, 410, 51 },
		{ 500, 411, 51 },
		{ 500, 412, 51 },
		{ 500, 413, 51 },
		{ 500, 414, 51 },
		{ 500, 415, 51 },
		{ 500, 416, 51 },
		{ 500, 417, 51 },
		{ 500, 418, 51 },
		{ 500, 419, 51 },
		{ 500, 420, 51 },
		{ 500, 421, 51 },
		{ 500, 422, 51 },
		{ 500, 423, 51 },
		{ 500, 424, 51 },
		{ 500, 425, 51 },
		{ 500, 426, 51 },
		{ 500, 427, 51 },
		{ 500, 428, 51 },
		{ 500, 429, 51 },
		{ 500, 430, 51 },
		{ 500, 431, 51 },
		{ 500, 432, 51 },
		{ 500, 433, 51 },
		{ 500, 434, 51 },
		{ 500, 435, 51 },
		{ 500, 436, 51 },
		{ 500, 437, 51 },
		{ 500, 438, 51 },
		{ 500, 439, 51 },
		{ 501, 380, 51 },
		{ 501, 381, 51 },
		{ 501, 382, 51 },
		{ 501, 383, 51 },
		{ 501, 384, 51 },
		{ 501, 385, 51 },
		{ 501, 386, 51 },
		{ 501, 387, 51 },
		{ 501, 388, 51 },
		{ 501, 389, 51 },
		{ 501, 390, 51 },
		{ 501, 391, 51 },
		{ 501, 392, 51 },
		{ 501, 393, 51 },
		{ 501, 394, 51 },
		{ 501, 395, 51 },
		{ 501, 396, 51 },
		{ 501, 397, 51 },
		{ 501, 398, 51 },
		{ 501, 399, 51 },
		{ 501, 400, 51 },
		{ 501, 401, 51 },
		{ 501, 402, 51 },
		{ 501, 403, 51 },
		{ 501, 404, 51 },
		{ 501, 405, 51 },
		{ 501, 406, 51 },
		{ 501, 407, 51 },
		{ 501, 408, 51 },
		{ 501, 409, 51 },
		{ 501, 410, 51 },
		{ 501, 411, 51 },
		{ 501, 412, 51 },
		{ 501, 413, 51 },
		{ 501, 414, 51 },
		{ 501, 415, 51 },
		{ 501, 416, 51 },
		{ 501, 417, 51 },
		{ 501, 418, 51 },
		{ 501, 419, 51 },
		{ 501, 420, 51 },
		{ 501, 421, 51 },
		{ 501, 422, 51 },
		{ 501, 423, 51 },
		{ 501, 424, 51 },
		{ 501, 425, 51 },
		{ 501, 426, 51 },
		{ 501, 427, 51 },
		{ 501, 428, 51 },
		{ 501, 429, 51 },
		{ 501, 430, 51 },
		{ 501, 431, 51 },
		{ 501, 432, 51 },
		{ 501, 433, 51 },
		{ 501, 434, 51 },
		{ 501, 435, 51 },
		{ 501, 436, 51 },
		{ 501, 437, 51 },
		{ 501, 438, 51 },
		{ 501, 439, 51 },
		{ 502, 380, 51 },
		{ 502, 381, 51 },
		{ 502, 382, 51 },
		{ 502, 383, 51 },
		{ 502, 384, 51 },
		{ 502, 385, 51 },
		{ 502, 386, 51 },
		{ 502, 387, 51 },
		{ 502, 388, 51 },
		{ 502, 389, 51 },
		{ 502, 390, 51 },
		{ 502, 391, 51 },
		{ 502, 392, 51 },
		{ 502, 393, 51 },
		{ 502, 394, 51 },
		{ 502, 395, 51 },
		{ 502, 396, 51 },
		{ 502, 397, 51 },
		{ 502, 398, 51 },
		{ 502, 399, 51 },
		{ 502, 400, 51 },
		{ 502, 401, 51 },
		{ 502, 402, 51 },
		{ 502, 403, 51 },
		{ 502, 404, 51 },
		{ 502, 405, 51 },
		{ 502, 406, 51 },
		{ 502, 407, 51 },
		{ 502, 408, 51 },
		{ 502, 409, 51 },
		{ 502, 410, 51 },
		{ 502, 411, 51 },
		{ 502, 412, 51 },
		{ 502, 413, 51 },
		{ 502, 414, 51 },
		{ 502, 415, 51 },
		{ 502, 416, 51 },
		{ 502, 417, 51 },
		{ 502, 418, 51 },
		{ 502, 419, 51 },
		{ 502, 420, 51 },
		{ 502, 421, 51 },
		{ 502, 422, 51 },
		{ 502, 423, 51 },
		{ 502, 424, 51 },
		{ 502, 425, 51 },
		{ 502, 426, 51 },
		{ 502, 427, 51 },
		{ 502, 428, 51 },
		{ 502, 429, 51 },
		{ 502, 430, 51 },
		{ 502, 431, 51 },
		{ 502, 432, 51 },
		{ 502, 433, 51 },
		{ 502, 434, 51 },
		{ 502, 435, 51 },
		{ 502, 436, 51 },
		{ 502, 437, 51 },
		{ 502, 438, 51 },
		{ 502, 439, 51 },
		{ 502, 440, 51 },
		{ 503, 381, 51 },
		{ 503, 382, 51 },
		{ 503, 383, 51 },
		{ 503, 384, 51 },
		{ 503, 385, 51 },
		{ 503, 386, 51 },
		{ 503, 387, 51 },
		{ 503, 388, 51 },
		{ 503, 389, 51 },
		{ 503, 390, 51 },
		{ 503, 391, 51 },
		{ 503, 392, 51 },
		{ 503, 393, 51 },
		{ 503, 394, 51 },
		{ 503, 395, 51 },
		{ 503, 396, 51 },
		{ 503, 397, 51 },
		{ 503, 398, 51 },
		{ 503, 399, 51 },
		{ 503, 400, 51 },
		{ 503, 401, 51 },
		{ 503, 402, 51 },
		{ 503, 403, 51 },
		{ 503, 404, 51 },
		{ 503, 405, 51 },
		{ 503, 406, 51 },
		{ 503, 407, 51 },
		{ 503, 408, 51 },
		{ 503, 409, 51 },
		{ 503, 410, 51 },
		{ 503, 411, 51 },
		{ 503, 412, 51 },
		{ 503, 413, 51 },
		{ 503, 414, 51 },
		{ 503, 415, 51 },
		{ 503, 416, 51 },
		{ 503, 417, 51 },
		{ 503, 418, 51 },
		{ 503, 419, 51 },
		{ 503, 420, 51 },
		{ 503, 421, 51 },
		{ 503, 422, 51 },
		{ 503, 423, 51 },
		{ 503, 424, 51 },
		{ 503, 425, 51 },
		{ 503, 426, 51 },
		{ 503, 427, 51 },
		{ 503, 428, 51 },
		{ 503, 429, 51 },
		{ 503, 430, 51 },
		{ 503, 431, 51 },
		{ 503, 432, 51 },
		{ 503, 433, 51 },
		{ 503, 434, 51 },
		{ 503, 435, 51 },
		{ 503, 436, 51 },
		{ 503, 437, 51 },
		{ 503, 438, 51 },
		{ 503, 439, 51 },
		{ 503, 440, 51 },
		{ 503, 441, 51 },
		{ 504, 381, 51 },
		{ 504, 382, 51 },
		{ 504, 383, 51 },
		{ 504, 384, 51 },
		{ 504, 385, 51 },
		{ 504, 386, 51 },
		{ 504, 387, 51 },
		{ 504, 388, 51 },
		{ 504, 389, 51 },
		{ 504, 390, 51 },
		{ 504, 391, 51 },
		{ 504, 392, 51 },
		{ 504, 393, 51 },
		{ 504, 394, 51 },
		{ 504, 395, 51 },
		{ 504, 396, 51 },
		{ 504, 397, 51 },
		{ 504, 398, 51 },
		{ 504, 399, 51 },
		{ 504, 400, 51 },
		{ 504, 401, 51 },
		{ 504, 402, 51 },
		{ 504, 403, 51 },
		{ 504, 404, 51 },
		{ 504, 405, 51 },
		{ 504, 406, 51 },
		{ 504, 407, 51 },
		{ 504, 408, 51 },
		{ 504, 409, 51 },
		{ 504, 410, 51 },
		{ 504, 411, 51 },
		{ 504, 412, 51 },
		{ 504, 413, 51 },
		{ 504, 414, 51 },
		{ 504, 415, 51 },
		{ 504, 416, 51 },
		{ 504, 417, 51 },
		{ 504, 418, 51 },
		{ 504, 419, 51 },
		{ 504, 420, 51 },
		{ 504, 421, 51 },
		{ 504, 422, 51 },
		{ 504, 423, 51 },
		{ 504, 424, 51 },
		{ 504, 425, 51 },
		{ 504, 426, 51 },
		{ 504, 427, 51 },
		{ 504, 428, 51 },
		{ 504, 429, 51 },
		{ 504, 430, 51 },
		{ 504, 431, 51 },
		{ 504, 432, 51 },
		{ 504, 433, 51 },
		{ 504, 434, 51 },
		{ 504, 435, 51 },
		{ 504, 436, 51 },
		{ 504, 437, 51 },
		{ 504, 438, 51 },
		{ 504, 439, 51 },
		{ 504, 440, 51 },
		{ 504, 441, 51 },
		{ 504, 442, 51 },
		{ 505, 382, 51 },
		{ 505, 383, 51 },
		{ 505, 384, 51 },
		{ 505, 385, 51 },
		{ 505, 386, 51 },
		{ 505, 387, 51 },
		{ 505, 388, 51 },
		{ 505, 389, 51 },
		{ 505, 390, 51 },
		{ 505, 391, 51 },
		{ 505, 392, 51 },
		{ 505, 393, 51 },
		{ 505, 394, 51 },
		{ 505, 395, 51 },
		{ 505, 396, 51 },
		{ 505, 397, 51 },
		{ 505, 398, 51 },
		{ 505, 399, 51 },
		{ 505, 400, 51 },
		{ 505, 401, 51 },
		{ 505, 402, 51 },
		{ 505, 403, 51 },
		{ 505, 404, 51 },
		{ 505, 405, 51 },
		{ 505, 406, 51 },
		{ 505, 407, 51 },
		{ 505, 408, 51 },
		{ 505, 409, 51 },
		{ 505, 410, 51 },
		{ 505, 411, 51 },
		{ 505, 412, 51 },
		{ 505, 413, 51 },
		{ 505, 414, 51 },
		{ 505, 415, 51 },
		{ 505, 416, 51 },
		{ 505, 417, 51 },
		{ 505, 418, 51 },
		{ 505, 419, 51 },
		{ 505, 420, 51 },
		{ 505, 421, 51 },
		{ 505, 422, 51 },
		{ 505, 423, 51 },
		{ 505, 424, 51 },
		{ 505, 425, 51 },
		{ 505, 426, 51 },
		{ 505, 427, 51 },
		{ 505, 428, 51 },
		{ 505, 429, 51 },
		{ 505, 430, 51 },
		{ 505, 431, 51 },
		{ 505, 432, 51 },
		{ 505, 433, 51 },
		{ 505, 434, 51 },
		{ 505, 435, 51 },
		{ 505, 436, 51 },
		{ 505, 437, 51 },
		{ 505, 438, 51 },
		{ 505, 439, 51 },
		{ 505, 440, 51 },
		{ 505, 441, 51 },
		{ 505, 442, 51 },
		{ 506, 382, 51 },
		{ 506, 383, 51 },
		{ 506, 384, 51 },
		{ 506, 385, 51 },
		{ 506, 386, 51 },
		{ 506, 387, 51 },
		{ 506, 388, 51 },
		{ 506, 389, 51 },
		{ 506, 390, 51 },
		{ 506, 391, 51 },
		{ 506, 392, 51 },
		{ 506, 393, 51 },
		{ 506, 394, 51 },
		{ 506, 395, 51 },
		{ 506, 396, 51 },
		{ 506, 397, 51 },
		{ 506, 398, 51 },
		{ 506, 399, 51 },
		{ 506, 400, 51 },
		{ 506, 401, 51 },
		{ 506, 402, 51 },
		{ 506, 403, 51 },
		{ 506, 404, 51 },
		{ 506, 405, 51 },
		{ 506, 406, 51 },
		{ 506, 407, 51 },
		{ 506, 408, 51 },
		{ 506, 409, 51 },
		{ 506, 410, 51 },
		{ 506, 411, 51 },
		{ 506, 412, 51 },
		{ 506, 413, 51 },
		{ 506, 414, 51 },
		{ 506, 415, 51 },
		{ 506, 416, 51 },
		{ 506, 417, 51 },
		{ 506, 418, 51 },
		{ 506, 419, 51 },
		{ 506, 420, 51 },
		{ 506, 421, 51 },
		{ 506, 422, 51 },
		{ 506, 423, 51 },
		{ 506, 424, 51 },
		{ 506, 425, 51 },
		{ 506, 426, 51 },
		{ 506, 427, 51 },
		{ 506, 428, 51 },
		{ 506, 429, 51 },
		{ 506, 430, 51 },
		{ 506, 431, 51 },
		{ 506, 432, 51 },
		{ 506, 433, 51 },
		{ 506, 434, 51 },
		{ 506, 435, 51 },
		{ 506, 436, 51 },
		{ 506, 437, 51 },
		{ 506, 438, 51 },
		{ 506, 439, 51 },
		{ 506, 440, 51 },
		{ 506, 441, 51 },
		{ 506, 442, 51 },
		{ 506, 443, 51 },
		{ 507, 383, 51 },
		{ 507, 384, 51 },
		{ 507, 385, 51 },
		{ 507, 386, 51 },
		{ 507, 387, 51 },
		{ 507, 388, 51 },
		{ 507, 389, 51 },
		{ 507, 390, 51 },
		{ 507, 391, 51 },
		{ 507, 392, 51 },
		{ 507, 393, 51 },
		{ 507, 394, 51 },
		{ 507, 395, 51 },
		{ 507, 396, 51 },
		{ 507, 397, 51 },
		{ 507, 398, 51 },
		{ 507, 399, 51 },
		{ 507, 400, 51 },
		{ 507, 401, 51 },
		{ 507, 402, 51 },
		{ 507, 403, 51 },
		{ 507, 404, 51 },
		{ 507, 405, 51 },
		{ 507, 406, 51 },
		{ 507, 407, 51 },
		{ 507, 408, 51 },
		{ 507, 409, 51 },
		{ 507, 410, 51 },
		{ 507, 411, 51 },
		{ 507, 412, 51 },
		{ 507, 413, 51 },
		{ 507, 414, 51 },
		{ 507, 415, 51 },
		{ 507, 416, 51 },
		{ 507, 417, 51 },
		{ 507, 418, 51 },
		{ 507, 419, 51 },
		{ 507, 420, 51 },
		{ 507, 421, 51 },
		{ 507, 422, 51 },
		{ 507, 423, 51 },
		{ 507, 424, 51 },
		{ 507, 425, 51 },
		{ 507, 426, 51 },
		{ 507, 427, 51 },
		{ 507, 428, 51 },
		{ 507, 429, 51 },
		{ 507, 430, 51 },
		{ 507, 431, 51 },
		{ 507, 432, 51 },
		{ 507, 433, 51 },
		{ 507, 434, 51 },
		{ 507, 435, 51 },
		{ 507, 436, 51 },
		{ 507, 437, 51 },
		{ 507, 438, 51 },
		{ 507, 439, 51 },
		{ 507, 440, 51 },
		{ 507, 441, 51 },
		{ 507, 442, 51 },
		{ 507, 443, 51 },
		{ 507, 444, 51 },
		{ 508, 383, 51 },
		{ 508, 384, 51 },
		{ 508, 385, 51 },
		{ 508, 386, 51 },
		{ 508, 387, 51 },
		{ 508, 388, 51 },
		{ 508, 389, 51 },
		{ 508, 390, 51 },
		{ 508, 391, 51 },
		{ 508, 392, 51 },
		{ 508, 393, 51 },
		{ 508, 394, 51 },
		{ 508, 395, 51 },
		{ 508, 396, 51 },
		{ 508, 397, 51 },
		{ 508, 398, 51 },
		{ 508, 399, 51 },
		{ 508, 400, 51 },
		{ 508, 401, 51 },
		{ 508, 402, 51 },
		{ 508, 403, 51 },
		{ 508, 404, 51 },
		{ 508, 405, 51 },
		{ 508, 406, 51 },
		{ 508, 407, 51 },
		{ 508, 408, 51 },
		{ 508, 409, 51 },
		{ 508, 410, 51 },
		{ 508, 411, 51 },
		{ 508, 412, 51 },
		{ 508, 413, 51 },
		{ 508, 414, 51 },
		{ 508, 415, 51 },
		{ 508, 416, 51 },
		{ 508, 417, 51 },
		{ 508, 418, 51 },
		{ 508, 419, 51 },
		{ 508, 420, 51 },
		{ 508, 421, 51 },
		{ 508, 422, 51 },
		{ 508, 423, 51 },
		{ 508, 424, 51 },
		{ 508, 425, 51 },
		{ 508, 426, 51 },
		{ 508, 427, 51 },
		{ 508, 428, 51 },
		{ 508, 429, 51 },
		{ 508, 430, 51 },
		{ 508, 431, 51 },
		{ 508, 432, 51 },
		{ 508, 433, 51 },
		{ 508, 434, 51 },
		{ 508, 435, 51 },
		{ 508, 436, 51 },
		{ 508, 437, 51 },
		{ 508, 438, 51 },
		{ 508, 439, 51 },
		{ 508, 440, 51 },
		{ 508, 441, 51 },
		{ 508, 442, 51 },
		{ 508, 443, 51 },
		{ 508, 444, 51 },
		{ 508, 445, 51 },
		{ 509, 384, 51 },
		{ 509, 385, 51 },
		{ 509, 386, 51 },
		{ 509, 387, 51 },
		{ 509, 388, 51 },
		{ 509, 389, 51 },
		{ 509, 390, 51 },
		{ 509, 391, 51 },
		{ 509, 392, 51 },
		{ 509, 393, 51 },
		{ 509, 394, 51 },
		{ 509, 395, 51 },
		{ 509, 396, 51 },
		{ 509, 397, 51 },
		{ 509, 398, 51 },
		{ 509, 399, 51 },
		{ 509, 400, 51 },
		{ 509, 401, 51 },
		{ 509, 402, 51 },
		{ 509, 403, 51 },
		{ 509, 404, 51 },
		{ 509, 405, 51 },
		{ 509, 406, 51 },
		{ 509, 407, 51 },
		{ 509, 408, 51 },
		{ 509, 409, 51 },
		{ 509, 410, 51 },
		{ 509, 411, 51 },
		{ 509, 412, 51 },
		{ 509, 413, 51 },
		{ 509, 414, 51 },
		{ 509, 415, 51 },
		{ 509, 416, 51 },
		{ 509, 417, 51 },
		{ 509, 418, 51 },
		{ 509, 419, 51 },
		{ 509, 420, 51 },
		{ 509, 421, 51 },
		{ 509, 422, 51 },
		{ 509, 423, 51 },
		{ 509, 424, 51 },
		{ 509, 425, 51 },
		{ 509, 426, 51 },
		{ 509, 427, 51 },
		{ 509, 428, 51 },
		{ 509, 429, 51 },
		{ 509, 430, 51 },
		{ 509, 431, 51 },
		{ 509, 432, 51 },
		{ 509, 433, 51 },
		{ 509, 434, 51 },
		{ 509, 435, 51 },
		{ 509, 436, 51 },
		{ 509, 437, 51 },
		{ 509, 438, 51 },
		{ 509, 439, 51 },
		{ 509, 440, 51 },
		{ 509, 441, 51 },
		{ 509, 442, 51 },
		{ 509, 443, 51 },
		{ 509, 444, 51 },
		{ 509, 445, 51 },
		{ 510, 384, 51 },
		{ 510, 385, 51 },
		{ 510, 386, 51 },
		{ 510, 387, 51 },
		{ 510, 388, 51 },
		{ 510, 389, 51 },
		{ 510, 390, 51 },
		{ 510, 391, 51 },
		{ 510, 392, 51 },
		{ 510, 393, 51 },
		{ 510, 394, 51 },
		{ 510, 395, 51 },
		{ 510, 396, 51 },
		{ 510, 397, 51 },
		{ 510, 398, 51 },
		{ 510, 399, 51 },
		{ 510, 400, 51 },
		{ 510, 401, 51 },
		{ 510, 402, 51 },
		{ 510, 403, 51 },
		{ 510, 404, 51 },
		{ 510, 405, 51 },
		{ 510, 406, 51 },
		{ 510, 407, 51 },
		{ 510, 408, 51 },
		{ 510, 409, 51 },
		{ 510, 410, 51 },
		{ 510, 411, 51 },
		{ 510, 412, 51 },
		{ 510, 413, 51 },
		{ 510, 414, 51 },
		{ 510, 415, 51 },
		{ 510, 416, 51 },
		{ 510, 417, 51 },
		{ 510, 418, 51 },
		{ 510, 419, 51 },
		{ 510, 420, 51 },
		{ 510, 421, 51 },
		{ 510, 422, 51 },
		{ 510, 423, 51 },
		{ 510, 424, 51 },
		{ 510, 425, 51 },
		{ 510, 426, 51 },
		{ 510, 427, 51 },
		{ 510, 428, 51 },
		{ 510, 429, 51 },
		{ 510, 430, 51 },
		{ 510, 431, 51 },
		{ 510, 432, 51 },
		{ 510, 433, 51 },
		{ 510, 434, 51 },
		{ 510, 435, 51 },
		{ 510, 436, 51 },
		{ 510, 437, 51 },
		{ 510, 438, 51 },
		{ 510, 439, 51 },
		{ 510, 440, 51 },
		{ 510, 441, 51 },
		{ 510, 442, 51 },
		{ 510, 443, 51 },
		{ 510, 444, 51 },
		{ 510, 445, 51 },
		{ 510, 446, 51 },
		{ 511, 385, 51 },
		{ 511, 386, 51 },
		{ 511, 387, 51 },
		{ 511, 388, 51 },
		{ 511, 389, 51 },
		{ 511, 390, 51 },
		{ 511, 391, 51 },
		{ 511, 392, 51 },
		{ 511, 393, 51 },
		{ 511, 394, 51 },
		{ 511, 395, 51 },
		{ 511, 396, 51 },
		{ 511, 397, 51 },
		{ 511, 398, 51 },
		{ 511, 399, 51 },
		{ 511, 400, 51 },
		{ 511, 401, 51 },
		{ 511, 402, 51 },
		{ 511, 403, 51 },
		{ 511, 404, 51 },
		{ 511, 405, 51 },
		{ 511, 406, 51 },
		{ 511, 407, 51 },
		{ 511, 408, 51 },
		{ 511, 409, 51 },
		{ 511, 410, 51 },
		{ 511, 411, 51 },
		{ 511, 412, 51 },
		{ 511, 413, 51 },
		{ 511, 414, 51 },
		{ 511, 415, 51 },
		{ 511, 416, 51 },
		{ 511, 417, 51 },
		{ 511, 418, 51 },
		{ 511, 419, 51 },
		{ 511, 420, 51 },
		{ 511, 421, 51 },
		{ 511, 422, 51 },
		{ 511, 423, 51 },
		{ 511, 424, 51 },
		{ 511, 425, 51 },
		{ 511, 426, 51 },
		{ 511, 427, 51 },
		{ 511, 428, 51 },
		{ 511, 429, 51 },
		{ 511, 430, 51 },
		{ 511, 431, 51 },
		{ 511, 432, 51 },
		{ 511, 433, 51 },
		{ 511, 434, 51 },
		{ 511, 435, 51 },
		{ 511, 436, 51 },
		{ 511, 437, 51 },
		{ 511, 438, 51 },
		{ 511, 439, 51 },
		{ 511, 440, 51 },
		{ 511, 441, 51 },
		{ 511, 442, 51 },
		{ 511, 443, 51 },
		{ 511, 444, 51 },
		{ 511, 445, 51 },
		{ 511, 446, 51 },
		{ 511, 447, 51 },
		{ 512, 385, 51 },
		{ 512, 386, 51 },
		{ 512, 387, 51 },
		{ 512, 388, 51 },
		{ 512, 389, 51 },
		{ 512, 390, 51 },
		{ 512, 391, 51 },
		{ 512, 392, 51 },
		{ 512, 393, 51 },
		{ 512, 394, 51 },
		{ 512, 395, 51 },
		{ 512, 396, 51 },
		{ 512, 397, 51 },
		{ 512, 398, 51 },
		{ 512, 399, 51 },
		{ 512, 400, 51 },
		{ 512, 401, 51 },
		{ 512, 402, 51 },
		{ 512, 403, 51 },
		{ 512, 404, 51 },
		{ 512, 405, 51 },
		{ 512, 406, 51 },
		{ 512, 407, 51 },
		{ 512, 408, 51 },
		{ 512, 409, 51 },
		{ 512, 410, 51 },
		{ 512, 411, 51 },
		{ 512, 412, 51 },
		{ 512, 413, 51 },
		{ 512, 414, 51 },
		{ 512, 415, 51 },
		{ 512, 416, 51 },
		{ 512, 417, 51 },
		{ 512, 418, 51 },
		{ 512, 419, 51 },
		{ 512, 420, 51 },
		{ 512, 421, 51 },
		{ 512, 422, 51 },
		{ 512, 423, 51 },
		{ 512, 424, 51 },
		{ 512, 425, 51 },
		{ 512, 426, 51 },
		{ 512, 427, 51 },
		{ 512, 428, 51 },
		{ 512, 429, 51 },
		{ 512, 430, 51 },
		{ 512, 431, 51 },
		{ 512, 432, 51 },
		{ 512, 433, 51 },
		{ 512, 434, 51 },
		{ 512, 435, 51 },
		{ 512, 436, 51 },
		{ 512, 437, 51 },
		{ 512, 438, 51 },
		{ 512, 439, 51 },
		{ 512, 440, 51 },
		{ 512, 441, 51 },
		{ 512, 442, 51 },
		{ 512, 443, 51 },
		{ 512, 444, 51 },
		{ 512, 445, 51 },
		{ 512, 446, 51 },
		{ 512, 447, 51 },
		{ 512, 448, 51 },
		{ 1, 34, 52 },
		{ 1, 35, 52 },
		{ 1, 36, 52 },
		{ 1, 37, 52 },
		{ 1, 38, 52 },
		{ 1, 39, 52 },
		{ 1, 40, 52 },
		{ 1, 41, 52 },
		{ 1, 42, 52 },
		{ 1, 43, 52 },
		{ 1, 44, 52 },
		{ 1, 45, 52 },
		{ 1, 46, 52 },
		{ 1, 47, 52 },
		{ 1, 48, 52 },
		{ 1, 49, 52 },
		{ 1, 50, 52 },
		{ 1, 51, 52 },
		{ 1, 52, 52 },
		{ 1, 53, 52 },
		{ 1, 54, 52 },
		{ 1, 55, 52 },
		{ 1, 56, 52 },
		{ 1, 57, 52 },
		{ 1, 58, 52 },
		{ 1, 59, 52 },
		{ 1, 60, 52 },
		{ 1, 61, 52 },
		{ 1, 62, 52 },
		{ 1, 63, 52 },
		{ 1, 64, 52 },
		{ 1, 65, 52 },
		{ 1, 66, 52 },
		{ 1, 67, 52 },
		{ 1, 68, 52 },
		{ 1, 69, 52 },
		{ 1, 70, 52 },
		{ 1, 71, 52 },
		{ 1, 72, 52 },
		{ 1, 73, 52 },
		{ 1, 74, 52 },
		{ 1, 75, 52 },
		{ 1, 76, 52 },
		{ 1, 77, 52 },
		{ 1, 78, 52 },
		{ 1, 79, 52 },
		{ 1, 80, 52 },
		{ 1, 81, 52 },
		{ 1, 82, 52 },
		{ 1, 83, 52 },
		{ 1, 84, 52 },
		{ 1, 85, 52 },
		{ 1, 86, 52 },
		{ 1, 87, 52 },
		{ 1, 88, 52 },
		{ 1, 89, 52 },
		{ 1, 90, 52 },
		{ 1, 91, 52 },
		{ 1, 92, 52 },
		{ 1, 93, 52 },
		{ 1, 94, 52 },
		{ 1, 95, 52 },
		{ 1, 96, 52 },
		{ 2, 34, 52 },
		{ 2, 35, 52 },
		{ 2, 36, 52 },
		{ 2, 37, 52 },
		{ 2, 38, 52 },
		{ 2, 39, 52 },
		{ 2, 40, 52 },
		{ 2, 41, 52 },
		{ 2, 42, 52 },
		{ 2, 43, 52 },
		{ 2, 44, 52 },
		{ 2, 45, 52 },
		{ 2, 46, 52 },
		{ 2, 47, 52 },
		{ 2, 48, 52 },
		{ 2, 49, 52 },
		{ 2, 50, 52 },
		{ 2, 51, 52 },
		{ 2, 52, 52 },
		{ 2, 53, 52 },
		{ 2, 54, 52 },
		{ 2, 55, 52 },
		{ 2, 56, 52 },
		{ 2, 57, 52 },
		{ 2, 58, 52 },
		{ 2, 59, 52 },
		{ 2, 60, 52 },
		{ 2, 61, 52 },
		{ 2, 62, 52 },
		{ 2, 63, 52 },
		{ 2, 64, 52 },
		{ 2, 65, 52 },
		{ 2, 66, 52 },
		{ 2, 67, 52 },
		{ 2, 68, 52 },
		{ 2, 69, 52 },
		{ 2, 70, 52 },
		{ 2, 71, 52 },
		{ 2, 72, 52 },
		{ 2, 73, 52 },
		{ 2, 74, 52 },
		{ 2, 75, 52 },
		{ 2, 76, 52 },
		{ 2, 77, 52 },
		{ 2, 78, 52 },
		{ 2, 79, 52 },
		{ 2, 80, 52 },
		{ 2, 81, 52 },
		{ 2, 82, 52 },
		{ 2, 83, 52 },
		{ 2, 84, 52 },
		{ 2, 85, 52 },
		{ 2, 86, 52 },
		{ 2, 87, 52 },
		{ 2, 88, 52 },
		{ 2, 89, 52 },
		{ 2, 90, 52 },
		{ 2, 91, 52 },
		{ 2, 92, 52 },
		{ 2, 93, 52 },
		{ 2, 94, 52 },
		{ 2, 95, 52 },
		{ 2, 96, 52 },
		{ 2, 97, 52 },
		{ 3, 35, 52 },
		{ 3, 36, 52 },
		{ 3, 37, 52 },
		{ 3, 38, 52 },
		{ 3, 39, 52 },
		{ 3, 40, 52 },
		{ 3, 41, 52 },
		{ 3, 42, 52 },
		{ 3, 43, 52 },
		{ 3, 44, 52 },
		{ 3, 45, 52 },
		{ 3, 46, 52 },
		{ 3, 47, 52 },
		{ 3, 48, 52 },
		{ 3, 49, 52 },
		{ 3, 50, 52 },
		{ 3, 51, 52 },
		{ 3, 52, 52 },
		{ 3, 53, 52 },
		{ 3, 54, 52 },
		{ 3, 55, 52 },
		{ 3, 56, 52 },
		{ 3, 57, 52 },
		{ 3, 58, 52 },
		{ 3, 59, 52 },
		{ 3, 60, 52 },
		{ 3, 61, 52 },
		{ 3, 62, 52 },
		{ 3, 63, 52 },
		{ 3, 64, 52 },
		{ 3, 65, 52 },
		{ 3, 66, 52 },
		{ 3, 67, 52 },
		{ 3, 68, 52 },
		{ 3, 69, 52 },
		{ 3, 70, 52 },
		{ 3, 71, 52 },
		{ 3, 72, 52 },
		{ 3, 73, 52 },
		{ 3, 74, 52 },
		{ 3, 75, 52 },
		{ 3, 76, 52 },
		{ 3, 77, 52 },
		{ 3, 78, 52 },
		{ 3, 79, 52 },
		{ 3, 80, 52 },
		{ 3, 81, 52 },
		{ 3, 82, 52 },
		{ 3, 83, 52 },
		{ 3, 84, 52 },
		{ 3, 85, 52 },
		{ 3, 86, 52 },
		{ 3, 87, 52 },
		{ 3, 88, 52 },
		{ 3, 89, 52 },
		{ 3, 90, 52 },
		{ 3, 91, 52 },
		{ 3, 92, 52 },
		{ 3, 93, 52 },
		{ 3, 94, 52 },
		{ 3, 95, 52 },
		{ 3, 96, 52 },
		{ 3, 97, 52 },
		{ 3, 98, 52 },
		{ 4, 36, 52 },
		{ 4, 37, 52 },
		{ 4, 38, 52 },
		{ 4, 39, 52 },
		{ 4, 40, 52 },
		{ 4, 41, 52 },
		{ 4, 42, 52 },
		{ 4, 43, 52 },
		{ 4, 44, 52 },
		{ 4, 45, 52 },
		{ 4, 46, 52 },
		{ 4, 47, 52 },
		{ 4, 48, 52 },
		{ 4, 49, 52 },
		{ 4, 50, 52 },
		{ 4, 51, 52 },
		{ 4, 52, 52 },
		{ 4, 53, 52 },
		{ 4, 54, 52 },
		{ 4, 55, 52 },
		{ 4, 56, 52 },
		{ 4, 57, 52 },
		{ 4, 58, 52 },
		{ 4, 59, 52 },
		{ 4, 60, 52 },
		{ 4, 61, 52 },
		{ 4, 62, 52 },
		{ 4, 63, 52 },
		{ 4, 64, 52 },
		{ 4, 65, 52 },
		{ 4, 66, 52 },
		{ 4, 67, 52 },
		{ 4, 68, 52 },
		{ 4, 69, 52 },
		{ 4, 70, 52 },
		{ 4, 71, 52 },
		{ 4, 72, 52 },
		{ 4, 73, 52 },
		{ 4, 74, 52 },
		{ 4, 75, 52 },
		{ 4, 76, 52 },
		{ 4, 77, 52 },
		{ 4, 78, 52 },
		{ 4, 79, 52 },
		{ 4, 80, 52 },
		{ 4, 81, 52 },
		{ 4, 82, 52 },
		{ 4, 83, 52 },
		{ 4, 84, 52 },
		{ 4, 85, 52 },
		{ 4, 86, 52 },
		{ 4, 87, 52 },
		{ 4, 88, 52 },
		{ 4, 89, 52 },
		{ 4, 90, 52 },
		{ 4, 91, 52 },
		{ 4, 92, 52 },
		{ 4, 93, 52 },
		{ 4, 94, 52 },
		{ 4, 95, 52 },
		{ 4, 96, 52 },
		{ 4, 97, 52 },
		{ 4, 98, 52 },
		{ 5, 37, 52 },
		{ 5, 38, 52 },
		{ 5, 39, 52 },
		{ 5, 40, 52 },
		{ 5, 41, 52 },
		{ 5, 42, 52 },
		{ 5, 43, 52 },
		{ 5, 44, 52 },
		{ 5, 45, 52 },
		{ 5, 46, 52 },
		{ 5, 47, 52 },
		{ 5, 48, 52 },
		{ 5, 49, 52 },
		{ 5, 50, 52 },
		{ 5, 51, 52 },
		{ 5, 52, 52 },
		{ 5, 53, 52 },
		{ 5, 54, 52 },
		{ 5, 55, 52 },
		{ 5, 56, 52 },
		{ 5, 57, 52 },
		{ 5, 58, 52 },
		{ 5, 59, 52 },
		{ 5, 60, 52 },
		{ 5, 61, 52 },
		{ 5, 62, 52 },
		{ 5, 63, 52 },
		{ 5, 64, 52 },
		{ 5, 65, 52 },
		{ 5, 66, 52 },
		{ 5, 67, 52 },
		{ 5, 68, 52 },
		{ 5, 69, 52 },
		{ 5, 70, 52 },
		{ 5, 71, 52 },
		{ 5, 72, 52 },
		{ 5, 73, 52 },
		{ 5, 74, 52 },
		{ 5, 75, 52 },
		{ 5, 76, 52 },
		{ 5, 77, 52 },
		{ 5, 78, 52 },
		{ 5, 79, 52 },
		{ 5, 80, 52 },
		{ 5, 81, 52 },
		{ 5, 82, 52 },
		{ 5, 83, 52 },
		{ 5, 84, 52 },
		{ 5, 85, 52 },
		{ 5, 86, 52 },
		{ 5, 87, 52 },
		{ 5, 88, 52 },
		{ 5, 89, 52 },
		{ 5, 90, 52 },
		{ 5, 91, 52 },
		{ 5, 92, 52 },
		{ 5, 93, 52 },
		{ 5, 94, 52 },
		{ 5, 95, 52 },
		{ 5, 96, 52 },
		{ 5, 97, 52 },
		{ 5, 98, 52 },
		{ 5, 99, 52 },
		{ 6, 38, 52 },
		{ 6, 39, 52 },
		{ 6, 40, 52 },
		{ 6, 41, 52 },
		{ 6, 42, 52 },
		{ 6, 43, 52 },
		{ 6, 44, 52 },
		{ 6, 45, 52 },
		{ 6, 46, 52 },
		{ 6, 47, 52 },
		{ 6, 48, 52 },
		{ 6, 49, 52 },
		{ 6, 50, 52 },
		{ 6, 51, 52 },
		{ 6, 52, 52 },
		{ 6, 53, 52 },
		{ 6, 54, 52 },
		{ 6, 55, 52 },
		{ 6, 56, 52 },
		{ 6, 57, 52 },
		{ 6, 58, 52 },
		{ 6, 59, 52 },
		{ 6, 60, 52 },
		{ 6, 61, 52 },
		{ 6, 62, 52 },
		{ 6, 63, 52 },
		{ 6, 64, 52 },
		{ 6, 65, 52 },
		{ 6, 66, 52 },
		{ 6, 67, 52 },
		{ 6, 68, 52 },
		{ 6, 69, 52 },
		{ 6, 70, 52 },
		{ 6, 71, 52 },
		{ 6, 72, 52 },
		{ 6, 73, 52 },
		{ 6, 74, 52 },
		{ 6, 75, 52 },
		{ 6, 76, 52 },
		{ 6, 77, 52 },
		{ 6, 78, 52 },
		{ 6, 79, 52 },
		{ 6, 80, 52 },
		{ 6, 81, 52 },
		{ 6, 82, 52 },
		{ 6, 83, 52 },
		{ 6, 84, 52 },
		{ 6, 85, 52 },
		{ 6, 86, 52 },
		{ 6, 87, 52 },
		{ 6, 88, 52 },
		{ 6, 89, 52 },
		{ 6, 90, 52 },
		{ 6, 91, 52 },
		{ 6, 92, 52 },
		{ 6, 93, 52 },
		{ 6, 94, 52 },
		{ 6, 95, 52 },
		{ 6, 96, 52 },
		{ 6, 97, 52 },
		{ 6, 98, 52 },
		{ 6, 99, 52 },
		{ 6, 100, 52 },
		{ 7, 39, 52 },
		{ 7, 40, 52 },
		{ 7, 41, 52 },
		{ 7, 42, 52 },
		{ 7, 43, 52 },
		{ 7, 44, 52 },
		{ 7, 45, 52 },
		{ 7, 46, 52 },
		{ 7, 47, 52 },
		{ 7, 48, 52 },
		{ 7, 49, 52 },
		{ 7, 50, 52 },
		{ 7, 51, 52 },
		{ 7, 52, 52 },
		{ 7, 53, 52 },
		{ 7, 54, 52 },
		{ 7, 55, 52 },
		{ 7, 56, 52 },
		{ 7, 57, 52 },
		{ 7, 58, 52 },
		{ 7, 59, 52 },
		{ 7, 60, 52 },
		{ 7, 61, 52 },
		{ 7, 62, 52 },
		{ 7, 63, 52 },
		{ 7, 64, 52 },
		{ 7, 65, 52 },
		{ 7, 66, 52 },
		{ 7, 67, 52 },
		{ 7, 68, 52 },
		{ 7, 69, 52 },
		{ 7, 70, 52 },
		{ 7, 71, 52 },
		{ 7, 72, 52 },
		{ 7, 73, 52 },
		{ 7, 74, 52 },
		{ 7, 75, 52 },
		{ 7, 76, 52 },
		{ 7, 77, 52 },
		{ 7, 78, 52 },
		{ 7, 79, 52 },
		{ 7, 80, 52 },
		{ 7, 81, 52 },
		{ 7, 82, 52 },
		{ 7, 83, 52 },
		{ 7, 84, 52 },
		{ 7, 85, 52 },
		{ 7, 86, 52 },
		{ 7, 87, 52 },
		{ 7, 88, 52 },
		{ 7, 89, 52 },
		{ 7, 90, 52 },
		{ 7, 91, 52 },
		{ 7, 92, 52 },
		{ 7, 93, 52 },
		{ 7, 94, 52 },
		{ 7, 95, 52 },
		{ 7, 96, 52 },
		{ 7, 97, 52 },
		{ 7, 98, 52 },
		{ 7, 99, 52 },
		{ 7, 100, 52 },
		{ 8, 40, 52 },
		{ 8, 41, 52 },
		{ 8, 42, 52 },
		{ 8, 43, 52 },
		{ 8, 44, 52 },
		{ 8, 45, 52 },
		{ 8, 46, 52 },
		{ 8, 47, 52 },
		{ 8, 48, 52 },
		{ 8, 49, 52 },
		{ 8, 50, 52 },
		{ 8, 51, 52 },
		{ 8, 52, 52 },
		{ 8, 53, 52 },
		{ 8, 54, 52 },
		{ 8, 55, 52 },
		{ 8, 56, 52 },
		{ 8, 57, 52 },
		{ 8, 58, 52 },
		{ 8, 59, 52 },
		{ 8, 60, 52 },
		{ 8, 61, 52 },
		{ 8, 62, 52 },
		{ 8, 63, 52 },
		{ 8, 64, 52 },
		{ 8, 65, 52 },
		{ 8, 66, 52 },
		{ 8, 67, 52 },
		{ 8, 68, 52 },
		{ 8, 69, 52 },
		{ 8, 70, 52 },
		{ 8, 71, 52 },
		{ 8, 72, 52 },
		{ 8, 73, 52 },
		{ 8, 74, 52 },
		{ 8, 75, 52 },
		{ 8, 76, 52 },
		{ 8, 77, 52 },
		{ 8, 78, 52 },
		{ 8, 79, 52 },
		{ 8, 80, 52 },
		{ 8, 81, 52 },
		{ 8, 82, 52 },
		{ 8, 83, 52 },
		{ 8, 84, 52 },
		{ 8, 85, 52 },
		{ 8, 86, 52 },
		{ 8, 87, 52 },
		{ 8, 88, 52 },
		{ 8, 89, 52 },
		{ 8, 90, 52 },
		{ 8, 91, 52 },
		{ 8, 92, 52 },
		{ 8, 93, 52 },
		{ 8, 94, 52 },
		{ 8, 95, 52 },
		{ 8, 96, 52 },
		{ 8, 97, 52 },
		{ 8, 98, 52 },
		{ 8, 99, 52 },
		{ 8, 100, 52 },
		{ 8, 101, 52 },
		{ 9, 41, 52 },
		{ 9, 42, 52 },
		{ 9, 43, 52 },
		{ 9, 44, 52 },
		{ 9, 45, 52 },
		{ 9, 46, 52 },
		{ 9, 47, 52 },
		{ 9, 48, 52 },
		{ 9, 49, 52 },
		{ 9, 50, 52 },
		{ 9, 51, 52 },
		{ 9, 52, 52 },
		{ 9, 53, 52 },
		{ 9, 54, 52 },
		{ 9, 55, 52 },
		{ 9, 56, 52 },
		{ 9, 57, 52 },
		{ 9, 58, 52 },
		{ 9, 59, 52 },
		{ 9, 60, 52 },
		{ 9, 61, 52 },
		{ 9, 62, 52 },
		{ 9, 63, 52 },
		{ 9, 64, 52 },
		{ 9, 65, 52 },
		{ 9, 66, 52 },
		{ 9, 67, 52 },
		{ 9, 68, 52 },
		{ 9, 69, 52 },
		{ 9, 70, 52 },
		{ 9, 71, 52 },
		{ 9, 72, 52 },
		{ 9, 73, 52 },
		{ 9, 74, 52 },
		{ 9, 75, 52 },
		{ 9, 76, 52 },
		{ 9, 77, 52 },
		{ 9, 78, 52 },
		{ 9, 79, 52 },
		{ 9, 80, 52 },
		{ 9, 81, 52 },
		{ 9, 82, 52 },
		{ 9, 83, 52 },
		{ 9, 84, 52 },
		{ 9, 85, 52 },
		{ 9, 86, 52 },
		{ 9, 87, 52 },
		{ 9, 88, 52 },
		{ 9, 89, 52 },
		{ 9, 90, 52 },
		{ 9, 91, 52 },
		{ 9, 92, 52 },
		{ 9, 93, 52 },
		{ 9, 94, 52 },
		{ 9, 95, 52 },
		{ 9, 96, 52 },
		{ 9, 97, 52 },
		{ 9, 98, 52 },
		{ 9, 99, 52 },
		{ 9, 100, 52 },
		{ 9, 101, 52 },
		{ 10, 41, 52 },
		{ 10, 42, 52 },
		{ 10, 43, 52 },
		{ 10, 44, 52 },
		{ 10, 45, 52 },
		{ 10, 46, 52 },
		{ 10, 47, 52 },
		{ 10, 48, 52 },
		{ 10, 49, 52 },
		{ 10, 50, 52 },
		{ 10, 51, 52 },
		{ 10, 52, 52 },
		{ 10, 53, 52 },
		{ 10, 54, 52 },
		{ 10, 55, 52 },
		{ 10, 56, 52 },
		{ 10, 57, 52 },
		{ 10, 58, 52 },
		{ 10, 59, 52 },
		{ 10, 60, 52 },
		{ 10, 61, 52 },
		{ 10, 62, 52 },
		{ 10, 63, 52 },
		{ 10, 64, 52 },
		{ 10, 65, 52 },
		{ 10, 66, 52 },
		{ 10, 67, 52 },
		{ 10, 68, 52 },
		{ 10, 69, 52 },
		{ 10, 70, 52 },
		{ 10, 71, 52 },
		{ 10, 72, 52 },
		{ 10, 73, 52 },
		{ 10, 74, 52 },
		{ 10, 75, 52 },
		{ 10, 76, 52 },
		{ 10, 77, 52 },
		{ 10, 78, 52 },
		{ 10, 79, 52 },
		{ 10, 80, 52 },
		{ 10, 81, 52 },
		{ 10, 82, 52 },
		{ 10, 83, 52 },
		{ 10, 84, 52 },
		{ 10, 85, 52 },
		{ 10, 86, 52 },
		{ 10, 87, 52 },
		{ 10, 88, 52 },
		{ 10, 89, 52 },
		{ 10, 90, 52 },
		{ 10, 91, 52 },
		{ 10, 92, 52 },
		{ 10, 93, 52 },
		{ 10, 94, 52 },
		{ 10, 95, 52 },
		{ 10, 96, 52 },
		{ 10, 97, 52 },
		{ 10, 98, 52 },
		{ 10, 99, 52 },
		{ 10, 100, 52 },
		{ 10, 101, 52 },
		{ 10, 102, 52 },
		{ 11, 42, 52 },
		{ 11, 43, 52 },
		{ 11, 44, 52 },
		{ 11, 45, 52 },
		{ 11, 46, 52 },
		{ 11, 47, 52 },
		{ 11, 48, 52 },
		{ 11, 49, 52 },
		{ 11, 50, 52 },
		{ 11, 51, 52 },
		{ 11, 52, 52 },
		{ 11, 53, 52 },
		{ 11, 54, 52 },
		{ 11, 55, 52 },
		{ 11, 56, 52 },
		{ 11, 57, 52 },
		{ 11, 58, 52 },
		{ 11, 59, 52 },
		{ 11, 60, 52 },
		{ 11, 61, 52 },
		{ 11, 62, 52 },
		{ 11, 63, 52 },
		{ 11, 64, 52 },
		{ 11, 65, 52 },
		{ 11, 66, 52 },
		{ 11, 67, 52 },
		{ 11, 68, 52 },
		{ 11, 69, 52 },
		{ 11, 70, 52 },
		{ 11, 71, 52 },
		{ 11, 72, 52 },
		{ 11, 73, 52 },
		{ 11, 74, 52 },
		{ 11, 75, 52 },
		{ 11, 76, 52 },
		{ 11, 77, 52 },
		{ 11, 78, 52 },
		{ 11, 79, 52 },
		{ 11, 80, 52 },
		{ 11, 81, 52 },
		{ 11, 82, 52 },
		{ 11, 83, 52 },
		{ 11, 84, 52 },
		{ 11, 85, 52 },
		{ 11, 86, 52 },
		{ 11, 87, 52 },
		{ 11, 88, 52 },
		{ 11, 89, 52 },
		{ 11, 90, 52 },
		{ 11, 91, 52 },
		{ 11, 92, 52 },
		{ 11, 93, 52 },
		{ 11, 94, 52 },
		{ 11, 95, 52 },
		{ 11, 96, 52 },
		{ 11, 97, 52 },
		{ 11, 98, 52 },
		{ 11, 99, 52 },
		{ 11, 100, 52 },
		{ 11, 101, 52 },
		{ 11, 102, 52 },
		{ 11, 103, 52 },
		{ 12, 43, 52 },
		{ 12, 44, 52 },
		{ 12, 45, 52 },
		{ 12, 46, 52 },
		{ 12, 47, 52 },
		{ 12, 48, 52 },
		{ 12, 49, 52 },
		{ 12, 50, 52 },
		{ 12, 51, 52 },
		{ 12, 52, 52 },
		{ 12, 53, 52 },
		{ 12, 54, 52 },
		{ 12, 55, 52 },
		{ 12, 56, 52 },
		{ 12, 57, 52 },
		{ 12, 58, 52 },
		{ 12, 59, 52 },
		{ 12, 60, 52 },
		{ 12, 61, 52 },
		{ 12, 62, 52 },
		{ 12, 63, 52 },
		{ 12, 64, 52 },
		{ 12, 65, 52 },
		{ 12, 66, 52 },
		{ 12, 67, 52 },
		{ 12, 68, 52 },
		{ 12, 69, 52 },
		{ 12, 70, 52 },
		{ 12, 71, 52 },
		{ 12, 72, 52 },
		{ 12, 73, 52 },
		{ 12, 74, 52 },
		{ 12, 75, 52 },
		{ 12, 76, 52 },
		{ 12, 77, 52 },
		{ 12, 78, 52 },
		{ 12, 79, 52 },
		{ 12, 80, 52 },
		{ 12, 81, 52 },
		{ 12, 82, 52 },
		{ 12, 83, 52 },
		{ 12, 84, 52 },
		{ 12, 85, 52 },
		{ 12, 86, 52 },
		{ 12, 87, 52 },
		{ 12, 88, 52 },
		{ 12, 89, 52 },
		{ 12, 90, 52 },
		{ 12, 91, 52 },
		{ 12, 92, 52 },
		{ 12, 93, 52 },
		{ 12, 94, 52 },
		{ 12, 95, 52 },
		{ 12, 96, 52 },
		{ 12, 97, 52 },
		{ 12, 98, 52 },
		{ 12, 99, 52 },
		{ 12, 100, 52 },
		{ 12, 101, 52 },
		{ 12, 102, 52 },
		{ 12, 103, 52 },
		{ 13, 44, 52 },
		{ 13, 45, 52 },
		{ 13, 46, 52 },
		{ 13, 47, 52 },
		{ 13, 48, 52 },
		{ 13, 49, 52 },
		{ 13, 50, 52 },
		{ 13, 51, 52 },
		{ 13, 52, 52 },
		{ 13, 53, 52 },
		{ 13, 54, 52 },
		{ 13, 55, 52 },
		{ 13, 56, 52 },
		{ 13, 57, 52 },
		{ 13, 58, 52 },
		{ 13, 59, 52 },
		{ 13, 60, 52 },
		{ 13, 61, 52 },
		{ 13, 62, 52 },
		{ 13, 63, 52 },
		{ 13, 64, 52 },
		{ 13, 65, 52 },
		{ 13, 66, 52 },
		{ 13, 67, 52 },
		{ 13, 68, 52 },
		{ 13, 69, 52 },
		{ 13, 70, 52 },
		{ 13, 71, 52 },
		{ 13, 72, 52 },
		{ 13, 73, 52 },
		{ 13, 74, 52 },
		{ 13, 75, 52 },
		{ 13, 76, 52 },
		{ 13, 77, 52 },
		{ 13, 78, 52 },
		{ 13, 79, 52 },
		{ 13, 80, 52 },
		{ 13, 81, 52 },
		{ 13, 82, 52 },
		{ 13, 83, 52 },
		{ 13, 84, 52 },
		{ 13, 85, 52 },
		{ 13, 86, 52 },
		{ 13, 87, 52 },
		{ 13, 88, 52 },
		{ 13, 89, 52 },
		{ 13, 90, 52 },
		{ 13, 91, 52 },
		{ 13, 92, 52 },
		{ 13, 93, 52 },
		{ 13, 94, 52 },
		{ 13, 95, 52 },
		{ 13, 96, 52 },
		{ 13, 97, 52 },
		{ 13, 98, 52 },
		{ 13, 99, 52 },
		{ 13, 100, 52 },
		{ 13, 101, 52 },
		{ 13, 102, 52 },
		{ 13, 103, 52 },
		{ 13, 104, 52 },
		{ 14, 45, 52 },
		{ 14, 46, 52 },
		{ 14, 47, 52 },
		{ 14, 48, 52 },
		{ 14, 49, 52 },
		{ 14, 50, 52 },
		{ 14, 51, 52 },
		{ 14, 52, 52 },
		{ 14, 53, 52 },
		{ 14, 54, 52 },
		{ 14, 55, 52 },
		{ 14, 56, 52 },
		{ 14, 57, 52 },
		{ 14, 58, 52 },
		{ 14, 59, 52 },
		{ 14, 60, 52 },
		{ 14, 61, 52 },
		{ 14, 62, 52 },
		{ 14, 63, 52 },
		{ 14, 64, 52 },
		{ 14, 65, 52 },
		{ 14, 66, 52 },
		{ 14, 67, 52 },
		{ 14, 68, 52 },
		{ 14, 69, 52 },
		{ 14, 70, 52 },
		{ 14, 71, 52 },
		{ 14, 72, 52 },
		{ 14, 73, 52 },
		{ 14, 74, 52 },
		{ 14, 75, 52 },
		{ 14, 76, 52 },
		{ 14, 77, 52 },
		{ 14, 78, 52 },
		{ 14, 79, 52 },
		{ 14, 80, 52 },
		{ 14, 81, 52 },
		{ 14, 82, 52 },
		{ 14, 83, 52 },
		{ 14, 84, 52 },
		{ 14, 85, 52 },
		{ 14, 86, 52 },
		{ 14, 87, 52 },
		{ 14, 88, 52 },
		{ 14, 89, 52 },
		{ 14, 90, 52 },
		{ 14, 91, 52 },
		{ 14, 92, 52 },
		{ 14, 93, 52 },
		{ 14, 94, 52 },
		{ 14, 95, 52 },
		{ 14, 96, 52 },
		{ 14, 97, 52 },
		{ 14, 98, 52 },
		{ 14, 99, 52 },
		{ 14, 100, 52 },
		{ 14, 101, 52 },
		{ 14, 102, 52 },
		{ 14, 103, 52 },
		{ 14, 104, 52 },
		{ 14, 105, 52 },
		{ 15, 46, 52 },
		{ 15, 47, 52 },
		{ 15, 48, 52 },
		{ 15, 49, 52 },
		{ 15, 50, 52 },
		{ 15, 51, 52 },
		{ 15, 52, 52 },
		{ 15, 53, 52 },
		{ 15, 54, 52 },
		{ 15, 55, 52 },
		{ 15, 56, 52 },
		{ 15, 57, 52 },
		{ 15, 58, 52 },
		{ 15, 59, 52 },
		{ 15, 60, 52 },
		{ 15, 61, 52 },
		{ 15, 62, 52 },
		{ 15, 63, 52 },
		{ 15, 64, 52 },
		{ 15, 65, 52 },
		{ 15, 66, 52 },
		{ 15, 67, 52 },
		{ 15, 68, 52 },
		{ 15, 69, 52 },
		{ 15, 70, 52 },
		{ 15, 71, 52 },
		{ 15, 72, 52 },
		{ 15, 73, 52 },
		{ 15, 74, 52 },
		{ 15, 75, 52 },
		{ 15, 76, 52 },
		{ 15, 77, 52 },
		{ 15, 78, 52 },
		{ 15, 79, 52 },
		{ 15, 80, 52 },
		{ 15, 81, 52 },
		{ 15, 82, 52 },
		{ 15, 83, 52 },
		{ 15, 84, 52 },
		{ 15, 85, 52 },
		{ 15, 86, 52 },
		{ 15, 87, 52 },
		{ 15, 88, 52 },
		{ 15, 89, 52 },
		{ 15, 90, 52 },
		{ 15, 91, 52 },
		{ 15, 92, 52 },
		{ 15, 93, 52 },
		{ 15, 94, 52 },
		{ 15, 95, 52 },
		{ 15, 96, 52 },
		{ 15, 97, 52 },
		{ 15, 98, 52 },
		{ 15, 99, 52 },
		{ 15, 100, 52 },
		{ 15, 101, 52 },
		{ 15, 102, 52 },
		{ 15, 103, 52 },
		{ 15, 104, 52 },
		{ 15, 105, 52 },
		{ 16, 47, 52 },
		{ 16, 48, 52 },
		{ 16, 49, 52 },
		{ 16, 50, 52 },
		{ 16, 51, 52 },
		{ 16, 52, 52 },
		{ 16, 53, 52 },
		{ 16, 54, 52 },
		{ 16, 55, 52 },
		{ 16, 56, 52 },
		{ 16, 57, 52 },
		{ 16, 58, 52 },
		{ 16, 59, 52 },
		{ 16, 60, 52 },
		{ 16, 61, 52 },
		{ 16, 62, 52 },
		{ 16, 63, 52 },
		{ 16, 64, 52 },
		{ 16, 65, 52 },
		{ 16, 66, 52 },
		{ 16, 67, 52 },
		{ 16, 68, 52 },
		{ 16, 69, 52 },
		{ 16, 70, 52 },
		{ 16, 71, 52 },
		{ 16, 72, 52 },
		{ 16, 73, 52 },
		{ 16, 74, 52 },
		{ 16, 75, 52 },
		{ 16, 76, 52 },
		{ 16, 77, 52 },
		{ 16, 78, 52 },
		{ 16, 79, 52 },
		{ 16, 80, 52 },
		{ 16, 81, 52 },
		{ 16, 82, 52 },
		{ 16, 83, 52 },
		{ 16, 84, 52 },
		{ 16, 85, 52 },
		{ 16, 86, 52 },
		{ 16, 87, 52 },
		{ 16, 88, 52 },
		{ 16, 89, 52 },
		{ 16, 90, 52 },
		{ 16, 91, 52 },
		{ 16, 92, 52 },
		{ 16, 93, 52 },
		{ 16, 94, 52 },
		{ 16, 95, 52 },
		{ 16, 96, 52 },
		{ 16, 97, 52 },
		{ 16, 98, 52 },
		{ 16, 99, 52 },
		{ 16, 100, 52 },
		{ 16, 101, 52 },
		{ 16, 102, 52 },
		{ 16, 103, 52 },
		{ 16, 104, 52 },
		{ 16, 105, 52 },
		{ 16, 106, 52 },
		{ 17, 48, 52 },
		{ 17, 49, 52 },
		{ 17, 50, 52 },
		{ 17, 51, 52 },
		{ 17, 52, 52 },
		{ 17, 53, 52 },
		{ 17, 54, 52 },
		{ 17, 55, 52 },
		{ 17, 56, 52 },
		{ 17, 57, 52 },
		{ 17, 58, 52 },
		{ 17, 59, 52 },
		{ 17, 60, 52 },
		{ 17, 61, 52 },
		{ 17, 62, 52 },
		{ 17, 63, 52 },
		{ 17, 64, 52 },
		{ 17, 65, 52 },
		{ 17, 66, 52 },
		{ 17, 67, 52 },
		{ 17, 68, 52 },
		{ 17, 69, 52 },
		{ 17, 70, 52 },
		{ 17, 71, 52 },
		{ 17, 72, 52 },
		{ 17, 73, 52 },
		{ 17, 74, 52 },
		{ 17, 75, 52 },
		{ 17, 76, 52 },
		{ 17, 77, 52 },
		{ 17, 78, 52 },
		{ 17, 79, 52 },
		{ 17, 80, 52 },
		{ 17, 81, 52 },
		{ 17, 82, 52 },
		{ 17, 83, 52 },
		{ 17, 84, 52 },
		{ 17, 85, 52 },
		{ 17, 86, 52 },
		{ 17, 87, 52 },
		{ 17, 88, 52 },
		{ 17, 89, 52 },
		{ 17, 90, 52 },
		{ 17, 91, 52 },
		{ 17, 92, 52 },
		{ 17, 93, 52 },
		{ 17, 94, 52 },
		{ 17, 95, 52 },
		{ 17, 96, 52 },
		{ 17, 97, 52 },
		{ 17, 98, 52 },
		{ 17, 99, 52 },
		{ 17, 100, 52 },
		{ 17, 101, 52 },
		{ 17, 102, 52 },
		{ 17, 103, 52 },
		{ 17, 104, 52 },
		{ 17, 105, 52 },
		{ 17, 106, 52 },
		{ 18, 48, 52 },
		{ 18, 49, 52 },
		{ 18, 50, 52 },
		{ 18, 51, 52 },
		{ 18, 52, 52 },
		{ 18, 53, 52 },
		{ 18, 54, 52 },
		{ 18, 55, 52 },
		{ 18, 56, 52 },
		{ 18, 57, 52 },
		{ 18, 58, 52 },
		{ 18, 59, 52 },
		{ 18, 60, 52 },
		{ 18, 61, 52 },
		{ 18, 62, 52 },
		{ 18, 63, 52 },
		{ 18, 64, 52 },
		{ 18, 65, 52 },
		{ 18, 66, 52 },
		{ 18, 67, 52 },
		{ 18, 68, 52 },
		{ 18, 69, 52 },
		{ 18, 70, 52 },
		{ 18, 71, 52 },
		{ 18, 72, 52 },
		{ 18, 73, 52 },
		{ 18, 74, 52 },
		{ 18, 75, 52 },
		{ 18, 76, 52 },
		{ 18, 77, 52 },
		{ 18, 78, 52 },
		{ 18, 79, 52 },
		{ 18, 80, 52 },
		{ 18, 81, 52 },
		{ 18, 82, 52 },
		{ 18, 83, 52 },
		{ 18, 84, 52 },
		{ 18, 85, 52 },
		{ 18, 86, 52 },
		{ 18, 87, 52 },
		{ 18, 88, 52 },
		{ 18, 89, 52 },
		{ 18, 90, 52 },
		{ 18, 91, 52 },
		{ 18, 92, 52 },
		{ 18, 93, 52 },
		{ 18, 94, 52 },
		{ 18, 95, 52 },
		{ 18, 96, 52 },
		{ 18, 97, 52 },
		{ 18, 98, 52 },
		{ 18, 99, 52 },
		{ 18, 100, 52 },
		{ 18, 101, 52 },
		{ 18, 102, 52 },
		{ 18, 103, 52 },
		{ 18, 104, 52 },
		{ 18, 105, 52 },
		{ 18, 106, 52 },
		{ 18, 107, 52 },
		{ 19, 49, 52 },
		{ 19, 50, 52 },
		{ 19, 51, 52 },
		{ 19, 52, 52 },
		{ 19, 53, 52 },
		{ 19, 54, 52 },
		{ 19, 55, 52 },
		{ 19, 56, 52 },
		{ 19, 57, 52 },
		{ 19, 58, 52 },
		{ 19, 59, 52 },
		{ 19, 60, 52 },
		{ 19, 61, 52 },
		{ 19, 62, 52 },
		{ 19, 63, 52 },
		{ 19, 64, 52 },
		{ 19, 65, 52 },
		{ 19, 66, 52 },
		{ 19, 67, 52 },
		{ 19, 68, 52 },
		{ 19, 69, 52 },
		{ 19, 70, 52 },
		{ 19, 71, 52 },
		{ 19, 72, 52 },
		{ 19, 73, 52 },
		{ 19, 74, 52 },
		{ 19, 75, 52 },
		{ 19, 76, 52 },
		{ 19, 77, 52 },
		{ 19, 78, 52 },
		{ 19, 79, 52 },
		{ 19, 80, 52 },
		{ 19, 81, 52 },
		{ 19, 82, 52 },
		{ 19, 83, 52 },
		{ 19, 84, 52 },
		{ 19, 85, 52 },
		{ 19, 86, 52 },
		{ 19, 87, 52 },
		{ 19, 88, 52 },
		{ 19, 89, 52 },
		{ 19, 90, 52 },
		{ 19, 91, 52 },
		{ 19, 92, 52 },
		{ 19, 93, 52 },
		{ 19, 94, 52 },
		{ 19, 95, 52 },
		{ 19, 96, 52 },
		{ 19, 97, 52 },
		{ 19, 98, 52 },
		{ 19, 99, 52 },
		{ 19, 100, 52 },
		{ 19, 101, 52 },
		{ 19, 102, 52 },
		{ 19, 103, 52 },
		{ 19, 104, 52 },
		{ 19, 105, 52 },
		{ 19, 106, 52 },
		{ 19, 107, 52 },
		{ 19, 108, 52 },
		{ 20, 50, 52 },
		{ 20, 51, 52 },
		{ 20, 52, 52 },
		{ 20, 53, 52 },
		{ 20, 54, 52 },
		{ 20, 55, 52 },
		{ 20, 56, 52 },
		{ 20, 57, 52 },
		{ 20, 58, 52 },
		{ 20, 59, 52 },
		{ 20, 60, 52 },
		{ 20, 61, 52 },
		{ 20, 62, 52 },
		{ 20, 63, 52 },
		{ 20, 64, 52 },
		{ 20, 65, 52 },
		{ 20, 66, 52 },
		{ 20, 67, 52 },
		{ 20, 68, 52 },
		{ 20, 69, 52 },
		{ 20, 70, 52 },
		{ 20, 71, 52 },
		{ 20, 72, 52 },
		{ 20, 73, 52 },
		{ 20, 74, 52 },
		{ 20, 75, 52 },
		{ 20, 76, 52 },
		{ 20, 77, 52 },
		{ 20, 78, 52 },
		{ 20, 79, 52 },
		{ 20, 80, 52 },
		{ 20, 81, 52 },
		{ 20, 82, 52 },
		{ 20, 83, 52 },
		{ 20, 84, 52 },
		{ 20, 85, 52 },
		{ 20, 86, 52 },
		{ 20, 87, 52 },
		{ 20, 88, 52 },
		{ 20, 89, 52 },
		{ 20, 90, 52 },
		{ 20, 91, 52 },
		{ 20, 92, 52 },
		{ 20, 93, 52 },
		{ 20, 94, 52 },
		{ 20, 95, 52 },
		{ 20, 96, 52 },
		{ 20, 97, 52 },
		{ 20, 98, 52 },
		{ 20, 99, 52 },
		{ 20, 100, 52 },
		{ 20, 101, 52 },
		{ 20, 102, 52 },
		{ 20, 103, 52 },
		{ 20, 104, 52 },
		{ 20, 105, 52 },
		{ 20, 106, 52 },
		{ 20, 107, 52 },
		{ 20, 108, 52 },
		{ 21, 51, 52 },
		{ 21, 52, 52 },
		{ 21, 53, 52 },
		{ 21, 54, 52 },
		{ 21, 55, 52 },
		{ 21, 56, 52 },
		{ 21, 57, 52 },
		{ 21, 58, 52 },
		{ 21, 59, 52 },
		{ 21, 60, 52 },
		{ 21, 61, 52 },
		{ 21, 62, 52 },
		{ 21, 63, 52 },
		{ 21, 64, 52 },
		{ 21, 65, 52 },
		{ 21, 66, 52 },
		{ 21, 67, 52 },
		{ 21, 68, 52 },
		{ 21, 69, 52 },
		{ 21, 70, 52 },
		{ 21, 71, 52 },
		{ 21, 72, 52 },
		{ 21, 73, 52 },
		{ 21, 74, 52 },
		{ 21, 75, 52 },
		{ 21, 76, 52 },
		{ 21, 77, 52 },
		{ 21, 78, 52 },
		{ 21, 79, 52 },
		{ 21, 80, 52 },
		{ 21, 81, 52 },
		{ 21, 82, 52 },
		{ 21, 83, 52 },
		{ 21, 84, 52 },
		{ 21, 85, 52 },
		{ 21, 86, 52 },
		{ 21, 87, 52 },
		{ 21, 88, 52 },
		{ 21, 89, 52 },
		{ 21, 90, 52 },
		{ 21, 91, 52 },
		{ 21, 92, 52 },
		{ 21, 93, 52 },
		{ 21, 94, 52 },
		{ 21, 95, 52 },
		{ 21, 96, 52 },
		{ 21, 97, 52 },
		{ 21, 98, 52 },
		{ 21, 99, 52 },
		{ 21, 100, 52 },
		{ 21, 101, 52 },
		{ 21, 102, 52 },
		{ 21, 103, 52 },
		{ 21, 104, 52 },
		{ 21, 105, 52 },
		{ 21, 106, 52 },
		{ 21, 107, 52 },
		{ 21, 108, 52 },
		{ 21, 109, 52 },
		{ 22, 52, 52 },
		{ 22, 53, 52 },
		{ 22, 54, 52 },
		{ 22, 55, 52 },
		{ 22, 56, 52 },
		{ 22, 57, 52 },
		{ 22, 58, 52 },
		{ 22, 59, 52 },
		{ 22, 60, 52 },
		{ 22, 61, 52 },
		{ 22, 62, 52 },
		{ 22, 63, 52 },
		{ 22, 64, 52 },
		{ 22, 65, 52 },
		{ 22, 66, 52 },
		{ 22, 67, 52 },
		{ 22, 68, 52 },
		{ 22, 69, 52 },
		{ 22, 70, 52 },
		{ 22, 71, 52 },
		{ 22, 72, 52 },
		{ 22, 73, 52 },
		{ 22, 74, 52 },
		{ 22, 75, 52 },
		{ 22, 76, 52 },
		{ 22, 77, 52 },
		{ 22, 78, 52 },
		{ 22, 79, 52 },
		{ 22, 80, 52 },
		{ 22, 81, 52 },
		{ 22, 82, 52 },
		{ 22, 83, 52 },
		{ 22, 84, 52 },
		{ 22, 85, 52 },
		{ 22, 86, 52 },
		{ 22, 87, 52 },
		{ 22, 88, 52 },
		{ 22, 89, 52 },
		{ 22, 90, 52 },
		{ 22, 91, 52 },
		{ 22, 92, 52 },
		{ 22, 93, 52 },
		{ 22, 94, 52 },
		{ 22, 95, 52 },
		{ 22, 96, 52 },
		{ 22, 97, 52 },
		{ 22, 98, 52 },
		{ 22, 99, 52 },
		{ 22, 100, 52 },
		{ 22, 101, 52 },
		{ 22, 102, 52 },
		{ 22, 103, 52 },
		{ 22, 104, 52 },
		{ 22, 105, 52 },
		{ 22, 106, 52 },
		{ 22, 107, 52 },
		{ 22, 108, 52 },
		{ 22, 109, 52 },
		{ 22, 110, 52 },
		{ 23, 53, 52 },
		{ 23, 54, 52 },
		{ 23, 55, 52 },
		{ 23, 56, 52 },
		{ 23, 57, 52 },
		{ 23, 58, 52 },
		{ 23, 59, 52 },
		{ 23, 60, 52 },
		{ 23, 61, 52 },
		{ 23, 62, 52 },
		{ 23, 63, 52 },
		{ 23, 64, 52 },
		{ 23, 65, 52 },
		{ 23, 66, 52 },
		{ 23, 67, 52 },
		{ 23, 68, 52 },
		{ 23, 69, 52 },
		{ 23, 70, 52 },
		{ 23, 71, 52 },
		{ 23, 72, 52 },
		{ 23, 73, 52 },
		{ 23, 74, 52 },
		{ 23, 75, 52 },
		{ 23, 76, 52 },
		{ 23, 77, 52 },
		{ 23, 78, 52 },
		{ 23, 79, 52 },
		{ 23, 80, 52 },
		{ 23, 81, 52 },
		{ 23, 82, 52 },
		{ 23, 83, 52 },
		{ 23, 84, 52 },
		{ 23, 85, 52 },
		{ 23, 86, 52 },
		{ 23, 87, 52 },
		{ 23, 88, 52 },
		{ 23, 89, 52 },
		{ 23, 90, 52 },
		{ 23, 91, 52 },
		{ 23, 92, 52 },
		{ 23, 93, 52 },
		{ 23, 94, 52 },
		{ 23, 95, 52 },
		{ 23, 96, 52 },
		{ 23, 97, 52 },
		{ 23, 98, 52 },
		{ 23, 99, 52 },
		{ 23, 100, 52 },
		{ 23, 101, 52 },
		{ 23, 102, 52 },
		{ 23, 103, 52 },
		{ 23, 104, 52 },
		{ 23, 105, 52 },
		{ 23, 106, 52 },
		{ 23, 107, 52 },
		{ 23, 108, 52 },
		{ 23, 109, 52 },
		{ 23, 110, 52 },
		{ 24, 54, 52 },
		{ 24, 55, 52 },
		{ 24, 56, 52 },
		{ 24, 57, 52 },
		{ 24, 58, 52 },
		{ 24, 59, 52 },
		{ 24, 60, 52 },
		{ 24, 61, 52 },
		{ 24, 62, 52 },
		{ 24, 63, 52 },
		{ 24, 64, 52 },
		{ 24, 65, 52 },
		{ 24, 66, 52 },
		{ 24, 67, 52 },
		{ 24, 68, 52 },
		{ 24, 69, 52 },
		{ 24, 70, 52 },
		{ 24, 71, 52 },
		{ 24, 72, 52 },
		{ 24, 73, 52 },
		{ 24, 74, 52 },
		{ 24, 75, 52 },
		{ 24, 76, 52 },
		{ 24, 77, 52 },
		{ 24, 78, 52 },
		{ 24, 79, 52 },
		{ 24, 80, 52 },
		{ 24, 81, 52 },
		{ 24, 82, 52 },
		{ 24, 83, 52 },
		{ 24, 84, 52 },
		{ 24, 85, 52 },
		{ 24, 86, 52 },
		{ 24, 87, 52 },
		{ 24, 88, 52 },
		{ 24, 89, 52 },
		{ 24, 90, 52 },
		{ 24, 91, 52 },
		{ 24, 92, 52 },
		{ 24, 93, 52 },
		{ 24, 94, 52 },
		{ 24, 95, 52 },
		{ 24, 96, 52 },
		{ 24, 97, 52 },
		{ 24, 98, 52 },
		{ 24, 99, 52 },
		{ 24, 100, 52 },
		{ 24, 101, 52 },
		{ 24, 102, 52 },
		{ 24, 103, 52 },
		{ 24, 104, 52 },
		{ 24, 105, 52 },
		{ 24, 106, 52 },
		{ 24, 107, 52 },
		{ 24, 108, 52 },
		{ 24, 109, 52 },
		{ 24, 110, 52 },
		{ 24, 111, 52 },
		{ 25, 55, 52 },
		{ 25, 56, 52 },
		{ 25, 57, 52 },
		{ 25, 58, 52 },
		{ 25, 59, 52 },
		{ 25, 60, 52 },
		{ 25, 61, 52 },
		{ 25, 62, 52 },
		{ 25, 63, 52 },
		{ 25, 64, 52 },
		{ 25, 65, 52 },
		{ 25, 66, 52 },
		{ 25, 67, 52 },
		{ 25, 68, 52 },
		{ 25, 69, 52 },
		{ 25, 70, 52 },
		{ 25, 71, 52 },
		{ 25, 72, 52 },
		{ 25, 73, 52 },
		{ 25, 74, 52 },
		{ 25, 75, 52 },
		{ 25, 76, 52 },
		{ 25, 77, 52 },
		{ 25, 78, 52 },
		{ 25, 79, 52 },
		{ 25, 80, 52 },
		{ 25, 81, 52 },
		{ 25, 82, 52 },
		{ 25, 83, 52 },
		{ 25, 84, 52 },
		{ 25, 85, 52 },
		{ 25, 86, 52 },
		{ 25, 87, 52 },
		{ 25, 88, 52 },
		{ 25, 89, 52 },
		{ 25, 90, 52 },
		{ 25, 91, 52 },
		{ 25, 92, 52 },
		{ 25, 93, 52 },
		{ 25, 94, 52 },
		{ 25, 95, 52 },
		{ 25, 96, 52 },
		{ 25, 97, 52 },
		{ 25, 98, 52 },
		{ 25, 99, 52 },
		{ 25, 100, 52 },
		{ 25, 101, 52 },
		{ 25, 102, 52 },
		{ 25, 103, 52 },
		{ 25, 104, 52 },
		{ 25, 105, 52 },
		{ 25, 106, 52 },
		{ 25, 107, 52 },
		{ 25, 108, 52 },
		{ 25, 109, 52 },
		{ 25, 110, 52 },
		{ 25, 111, 52 },
		{ 26, 55, 52 },
		{ 26, 56, 52 },
		{ 26, 57, 52 },
		{ 26, 58, 52 },
		{ 26, 59, 52 },
		{ 26, 60, 52 },
		{ 26, 61, 52 },
		{ 26, 62, 52 },
		{ 26, 63, 52 },
		{ 26, 64, 52 },
		{ 26, 65, 52 },
		{ 26, 66, 52 },
		{ 26, 67, 52 },
		{ 26, 68, 52 },
		{ 26, 69, 52 },
		{ 26, 70, 52 },
		{ 26, 71, 52 },
		{ 26, 72, 52 },
		{ 26, 73, 52 },
		{ 26, 74, 52 },
		{ 26, 75, 52 },
		{ 26, 76, 52 },
		{ 26, 77, 52 },
		{ 26, 78, 52 },
		{ 26, 79, 52 },
		{ 26, 80, 52 },
		{ 26, 81, 52 },
		{ 26, 82, 52 },
		{ 26, 83, 52 },
		{ 26, 84, 52 },
		{ 26, 85, 52 },
		{ 26, 86, 52 },
		{ 26, 87, 52 },
		{ 26, 88, 52 },
		{ 26, 89, 52 },
		{ 26, 90, 52 },
		{ 26, 91, 52 },
		{ 26, 92, 52 },
		{ 26, 93, 52 },
		{ 26, 94, 52 },
		{ 26, 95, 52 },
		{ 26, 96, 52 },
		{ 26, 97, 52 },
		{ 26, 98, 52 },
		{ 26, 99, 52 },
		{ 26, 100, 52 },
		{ 26, 101, 52 },
		{ 26, 102, 52 },
		{ 26, 103, 52 },
		{ 26, 104, 52 },
		{ 26, 105, 52 },
		{ 26, 106, 52 },
		{ 26, 107, 52 },
		{ 26, 108, 52 },
		{ 26, 109, 52 },
		{ 26, 110, 52 },
		{ 26, 111, 52 },
		{ 26, 112, 52 },
		{ 27, 56, 52 },
		{ 27, 57, 52 },
		{ 27, 58, 52 },
		{ 27, 59, 52 },
		{ 27, 60, 52 },
		{ 27, 61, 52 },
		{ 27, 62, 52 },
		{ 27, 63, 52 },
		{ 27, 64, 52 },
		{ 27, 65, 52 },
		{ 27, 66, 52 },
		{ 27, 67, 52 },
		{ 27, 68, 52 },
		{ 27, 69, 52 },
		{ 27, 70, 52 },
		{ 27, 71, 52 },
		{ 27, 72, 52 },
		{ 27, 73, 52 },
		{ 27, 74, 52 },
		{ 27, 75, 52 },
		{ 27, 76, 52 },
		{ 27, 77, 52 },
		{ 27, 78, 52 },
		{ 27, 79, 52 },
		{ 27, 80, 52 },
		{ 27, 81, 52 },
		{ 27, 82, 52 },
		{ 27, 83, 52 },
		{ 27, 84, 52 },
		{ 27, 85, 52 },
		{ 27, 86, 52 },
		{ 27, 87, 52 },
		{ 27, 88, 52 },
		{ 27, 89, 52 },
		{ 27, 90, 52 },
		{ 27, 91, 52 },
		{ 27, 92, 52 },
		{ 27, 93, 52 },
		{ 27, 94, 52 },
		{ 27, 95, 52 },
		{ 27, 96, 52 },
		{ 27, 97, 52 },
		{ 27, 98, 52 },
		{ 27, 99, 52 },
		{ 27, 100, 52 },
		{ 27, 101, 52 },
		{ 27, 102, 52 },
		{ 27, 103, 52 },
		{ 27, 104, 52 },
		{ 27, 105, 52 },
		{ 27, 106, 52 },
		{ 27, 107, 52 },
		{ 27, 108, 52 },
		{ 27, 109, 52 },
		{ 27, 110, 52 },
		{ 27, 111, 52 },
		{ 27, 112, 52 },
		{ 27, 113, 52 },
		{ 28, 57, 52 },
		{ 28, 58, 52 },
		{ 28, 59, 52 },
		{ 28, 60, 52 },
		{ 28, 61, 52 },
		{ 28, 62, 52 },
		{ 28, 63, 52 },
		{ 28, 64, 52 },
		{ 28, 65, 52 },
		{ 28, 66, 52 },
		{ 28, 67, 52 },
		{ 28, 68, 52 },
		{ 28, 69, 52 },
		{ 28, 70, 52 },
		{ 28, 71, 52 },
		{ 28, 72, 52 },
		{ 28, 73, 52 },
		{ 28, 74, 52 },
		{ 28, 75, 52 },
		{ 28, 76, 52 },
		{ 28, 77, 52 },
		{ 28, 78, 52 },
		{ 28, 79, 52 },
		{ 28, 80, 52 },
		{ 28, 81, 52 },
		{ 28, 82, 52 },
		{ 28, 83, 52 },
		{ 28, 84, 52 },
		{ 28, 85, 52 },
		{ 28, 86, 52 },
		{ 28, 87, 52 },
		{ 28, 88, 52 },
		{ 28, 89, 52 },
		{ 28, 90, 52 },
		{ 28, 91, 52 },
		{ 28, 92, 52 },
		{ 28, 93, 52 },
		{ 28, 94, 52 },
		{ 28, 95, 52 },
		{ 28, 96, 52 },
		{ 28, 97, 52 },
		{ 28, 98, 52 },
		{ 28, 99, 52 },
		{ 28, 100, 52 },
		{ 28, 101, 52 },
		{ 28, 102, 52 },
		{ 28, 103, 52 },
		{ 28, 104, 52 },
		{ 28, 105, 52 },
		{ 28, 106, 52 },
		{ 28, 107, 52 },
		{ 28, 108, 52 },
		{ 28, 109, 52 },
		{ 28, 110, 52 },
		{ 28, 111, 52 },
		{ 28, 112, 52 },
		{ 28, 113, 52 },
		{ 29, 58, 52 },
		{ 29, 59, 52 },
		{ 29, 60, 52 },
		{ 29, 61, 52 },
		{ 29, 62, 52 },
		{ 29, 63, 52 },
		{ 29, 64, 52 },
		{ 29, 65, 52 },
		{ 29, 66, 52 },
		{ 29, 67, 52 },
		{ 29, 68, 52 },
		{ 29, 69, 52 },
		{ 29, 70, 52 },
		{ 29, 71, 52 },
		{ 29, 72, 52 },
		{ 29, 73, 52 },
		{ 29, 74, 52 },
		{ 29, 75, 52 },
		{ 29, 76, 52 },
		{ 29, 77, 52 },
		{ 29, 78, 52 },
		{ 29, 79, 52 },
		{ 29, 80, 52 },
		{ 29, 81, 52 },
		{ 29, 82, 52 },
		{ 29, 83, 52 },
		{ 29, 84, 52 },
		{ 29, 85, 52 },
		{ 29, 86, 52 },
		{ 29, 87, 52 },
		{ 29, 88, 52 },
		{ 29, 89, 52 },
		{ 29, 90, 52 },
		{ 29, 91, 52 },
		{ 29, 92, 52 },
		{ 29, 93, 52 },
		{ 29, 94, 52 },
		{ 29, 95, 52 },
		{ 29, 96, 52 },
		{ 29, 97, 52 },
		{ 29, 98, 52 },
		{ 29, 99, 52 },
		{ 29, 100, 52 },
		{ 29, 101, 52 },
		{ 29, 102, 52 },
		{ 29, 103, 52 },
		{ 29, 104, 52 },
		{ 29, 105, 52 },
		{ 29, 106, 52 },
		{ 29, 107, 52 },
		{ 29, 108, 52 },
		{ 29, 109, 52 },
		{ 29, 110, 52 },
		{ 29, 111, 52 },
		{ 29, 112, 52 },
		{ 29, 113, 52 },
		{ 29, 114, 52 },
		{ 30, 59, 52 },
		{ 30, 60, 52 },
		{ 30, 61, 52 },
		{ 30, 62, 52 },
		{ 30, 63, 52 },
		{ 30, 64, 52 },
		{ 30, 65, 52 },
		{ 30, 66, 52 },
		{ 30, 67, 52 },
		{ 30, 68, 52 },
		{ 30, 69, 52 },
		{ 30, 70, 52 },
		{ 30, 71, 52 },
		{ 30, 72, 52 },
		{ 30, 73, 52 },
		{ 30, 74, 52 },
		{ 30, 75, 52 },
		{ 30, 76, 52 },
		{ 30, 77, 52 },
		{ 30, 78, 52 },
		{ 30, 79, 52 },
		{ 30, 80, 52 },
		{ 30, 81, 52 },
		{ 30, 82, 52 },
		{ 30, 83, 52 },
		{ 30, 84, 52 },
		{ 30, 85, 52 },
		{ 30, 86, 52 },
		{ 30, 87, 52 },
		{ 30, 88, 52 },
		{ 30, 89, 52 },
		{ 30, 90, 52 },
		{ 30, 91, 52 },
		{ 30, 92, 52 },
		{ 30, 93, 52 },
		{ 30, 94, 52 },
		{ 30, 95, 52 },
		{ 30, 96, 52 },
		{ 30, 97, 52 },
		{ 30, 98, 52 },
		{ 30, 99, 52 },
		{ 30, 100, 52 },
		{ 30, 101, 52 },
		{ 30, 102, 52 },
		{ 30, 103, 52 },
		{ 30, 104, 52 },
		{ 30, 105, 52 },
		{ 30, 106, 52 },
		{ 30, 107, 52 },
		{ 30, 108, 52 },
		{ 30, 109, 52 },
		{ 30, 110, 52 },
		{ 30, 111, 52 },
		{ 30, 112, 52 },
		{ 30, 113, 52 },
		{ 30, 114, 52 },
		{ 30, 115, 52 },
		{ 31, 60, 52 },
		{ 31, 61, 52 },
		{ 31, 62, 52 },
		{ 31, 63, 52 },
		{ 31, 64, 52 },
		{ 31, 65, 52 },
		{ 31, 66, 52 },
		{ 31, 67, 52 },
		{ 31, 68, 52 },
		{ 31, 69, 52 },
		{ 31, 70, 52 },
		{ 31, 71, 52 },
		{ 31, 72, 52 },
		{ 31, 73, 52 },
		{ 31, 74, 52 },
		{ 31, 75, 52 },
		{ 31, 76, 52 },
		{ 31, 77, 52 },
		{ 31, 78, 52 },
		{ 31, 79, 52 },
		{ 31, 80, 52 },
		{ 31, 81, 52 },
		{ 31, 82, 52 },
		{ 31, 83, 52 },
		{ 31, 84, 52 },
		{ 31, 85, 52 },
		{ 31, 86, 52 },
		{ 31, 87, 52 },
		{ 31, 88, 52 },
		{ 31, 89, 52 },
		{ 31, 90, 52 },
		{ 31, 91, 52 },
		{ 31, 92, 52 },
		{ 31, 93, 52 },
		{ 31, 94, 52 },
		{ 31, 95, 52 },
		{ 31, 96, 52 },
		{ 31, 97, 52 },
		{ 31, 98, 52 },
		{ 31, 99, 52 },
		{ 31, 100, 52 },
		{ 31, 101, 52 },
		{ 31, 102, 52 },
		{ 31, 103, 52 },
		{ 31, 104, 52 },
		{ 31, 105, 52 },
		{ 31, 106, 52 },
		{ 31, 107, 52 },
		{ 31, 108, 52 },
		{ 31, 109, 52 },
		{ 31, 110, 52 },
		{ 31, 111, 52 },
		{ 31, 112, 52 },
		{ 31, 113, 52 },
		{ 31, 114, 52 },
		{ 31, 115, 52 },
		{ 32, 61, 52 },
		{ 32, 62, 52 },
		{ 32, 63, 52 },
		{ 32, 64, 52 },
		{ 32, 65, 52 },
		{ 32, 66, 52 },
		{ 32, 67, 52 },
		{ 32, 68, 52 },
		{ 32, 69, 52 },
		{ 32, 70, 52 },
		{ 32, 71, 52 },
		{ 32, 72, 52 },
		{ 32, 73, 52 },
		{ 32, 74, 52 },
		{ 32, 75, 52 },
		{ 32, 76, 52 },
		{ 32, 77, 52 },
		{ 32, 78, 52 },
		{ 32, 79, 52 },
		{ 32, 80, 52 },
		{ 32, 81, 52 },
		{ 32, 82, 52 },
		{ 32, 83, 52 },
		{ 32, 84, 52 },
		{ 32, 85, 52 },
		{ 32, 86, 52 },
		{ 32, 87, 52 },
		{ 32, 88, 52 },
		{ 32, 89, 52 },
		{ 32, 90, 52 },
		{ 32, 91, 52 },
		{ 32, 92, 52 },
		{ 32, 93, 52 },
		{ 32, 94, 52 },
		{ 32, 95, 52 },
		{ 32, 96, 52 },
		{ 32, 97, 52 },
		{ 32, 98, 52 },
		{ 32, 99, 52 },
		{ 32, 100, 52 },
		{ 32, 101, 52 },
		{ 32, 102, 52 },
		{ 32, 103, 52 },
		{ 32, 104, 52 },
		{ 32, 105, 52 },
		{ 32, 106, 52 },
		{ 32, 107, 52 },
		{ 32, 108, 52 },
		{ 32, 109, 52 },
		{ 32, 110, 52 },
		{ 32, 111, 52 },
		{ 32, 112, 52 },
		{ 32, 113, 52 },
		{ 32, 114, 52 },
		{ 32, 115, 52 },
		{ 32, 116, 52 },
		{ 33, 62, 52 },
		{ 33, 63, 52 },
		{ 33, 64, 52 },
		{ 33, 65, 52 },
		{ 33, 66, 52 },
		{ 33, 67, 52 },
		{ 33, 68, 52 },
		{ 33, 69, 52 },
		{ 33, 70, 52 },
		{ 33, 71, 52 },
		{ 33, 72, 52 },
		{ 33, 73, 52 },
		{ 33, 74, 52 },
		{ 33, 75, 52 },
		{ 33, 76, 52 },
		{ 33, 77, 52 },
		{ 33, 78, 52 },
		{ 33, 79, 52 },
		{ 33, 80, 52 },
		{ 33, 81, 52 },
		{ 33, 82, 52 },
		{ 33, 83, 52 },
		{ 33, 84, 52 },
		{ 33, 85, 52 },
		{ 33, 86, 52 },
		{ 33, 87, 52 },
		{ 33, 88, 52 },
		{ 33, 89, 52 },
		{ 33, 90, 52 },
		{ 33, 91, 52 },
		{ 33, 92, 52 },
		{ 33, 93, 52 },
		{ 33, 94, 52 },
		{ 33, 95, 52 },
		{ 33, 96, 52 },
		{ 33, 97, 52 },
		{ 33, 98, 52 },
		{ 33, 99, 52 },
		{ 33, 100, 52 },
		{ 33, 101, 52 },
		{ 33, 102, 52 },
		{ 33, 103, 52 },
		{ 33, 104, 52 },
		{ 33, 105, 52 },
		{ 33, 106, 52 },
		{ 33, 107, 52 },
		{ 33, 108, 52 },
		{ 33, 109, 52 },
		{ 33, 110, 52 },
		{ 33, 111, 52 },
		{ 33, 112, 52 },
		{ 33, 113, 52 },
		{ 33, 114, 52 },
		{ 33, 115, 52 },
		{ 33, 116, 52 },
		{ 34, 62, 52 },
		{ 34, 63, 52 },
		{ 34, 64, 52 },
		{ 34, 65, 52 },
		{ 34, 66, 52 },
		{ 34, 67, 52 },
		{ 34, 68, 52 },
		{ 34, 69, 52 },
		{ 34, 70, 52 },
		{ 34, 71, 52 },
		{ 34, 72, 52 },
		{ 34, 73, 52 },
		{ 34, 74, 52 },
		{ 34, 75, 52 },
		{ 34, 76, 52 },
		{ 34, 77, 52 },
		{ 34, 78, 52 },
		{ 34, 79, 52 },
		{ 34, 80, 52 },
		{ 34, 81, 52 },
		{ 34, 82, 52 },
		{ 34, 83, 52 },
		{ 34, 84, 52 },
		{ 34, 85, 52 },
		{ 34, 86, 52 },
		{ 34, 87, 52 },
		{ 34, 88, 52 },
		{ 34, 89, 52 },
		{ 34, 90, 52 },
		{ 34, 91, 52 },
		{ 34, 92, 52 },
		{ 34, 93, 52 },
		{ 34, 94, 52 },
		{ 34, 95, 52 },
		{ 34, 96, 52 },
		{ 34, 97, 52 },
		{ 34, 98, 52 },
		{ 34, 99, 52 },
		{ 34, 100, 52 },
		{ 34, 101, 52 },
		{ 34, 102, 52 },
		{ 34, 103, 52 },
		{ 34, 104, 52 },
		{ 34, 105, 52 },
		{ 34, 106, 52 },
		{ 34, 107, 52 },
		{ 34, 108, 52 },
		{ 34, 109, 52 },
		{ 34, 110, 52 },
		{ 34, 111, 52 },
		{ 34, 112, 52 },
		{ 34, 113, 52 },
		{ 34, 114, 52 },
		{ 34, 115, 52 },
		{ 34, 116, 52 },
		{ 34, 117, 52 },
		{ 35, 63, 52 },
		{ 35, 64, 52 },
		{ 35, 65, 52 },
		{ 35, 66, 52 },
		{ 35, 67, 52 },
		{ 35, 68, 52 },
		{ 35, 69, 52 },
		{ 35, 70, 52 },
		{ 35, 71, 52 },
		{ 35, 72, 52 },
		{ 35, 73, 52 },
		{ 35, 74, 52 },
		{ 35, 75, 52 },
		{ 35, 76, 52 },
		{ 35, 77, 52 },
		{ 35, 78, 52 },
		{ 35, 79, 52 },
		{ 35, 80, 52 },
		{ 35, 81, 52 },
		{ 35, 82, 52 },
		{ 35, 83, 52 },
		{ 35, 84, 52 },
		{ 35, 85, 52 },
		{ 35, 86, 52 },
		{ 35, 87, 52 },
		{ 35, 88, 52 },
		{ 35, 89, 52 },
		{ 35, 90, 52 },
		{ 35, 91, 52 },
		{ 35, 92, 52 },
		{ 35, 93, 52 },
		{ 35, 94, 52 },
		{ 35, 95, 52 },
		{ 35, 96, 52 },
		{ 35, 97, 52 },
		{ 35, 98, 52 },
		{ 35, 99, 52 },
		{ 35, 100, 52 },
		{ 35, 101, 52 },
		{ 35, 102, 52 },
		{ 35, 103, 52 },
		{ 35, 104, 52 },
		{ 35, 105, 52 },
		{ 35, 106, 52 },
		{ 35, 107, 52 },
		{ 35, 108, 52 },
		{ 35, 109, 52 },
		{ 35, 110, 52 },
		{ 35, 111, 52 },
		{ 35, 112, 52 },
		{ 35, 113, 52 },
		{ 35, 114, 52 },
		{ 35, 115, 52 },
		{ 35, 116, 52 },
		{ 35, 117, 52 },
		{ 35, 118, 52 },
		{ 36, 64, 52 },
		{ 36, 65, 52 },
		{ 36, 66, 52 },
		{ 36, 67, 52 },
		{ 36, 68, 52 },
		{ 36, 69, 52 },
		{ 36, 70, 52 },
		{ 36, 71, 52 },
		{ 36, 72, 52 },
		{ 36, 73, 52 },
		{ 36, 74, 52 },
		{ 36, 75, 52 },
		{ 36, 76, 52 },
		{ 36, 77, 52 },
		{ 36, 78, 52 },
		{ 36, 79, 52 },
		{ 36, 80, 52 },
		{ 36, 81, 52 },
		{ 36, 82, 52 },
		{ 36, 83, 52 },
		{ 36, 84, 52 },
		{ 36, 85, 52 },
		{ 36, 86, 52 },
		{ 36, 87, 52 },
		{ 36, 88, 52 },
		{ 36, 89, 52 },
		{ 36, 90, 52 },
		{ 36, 91, 52 },
		{ 36, 92, 52 },
		{ 36, 93, 52 },
		{ 36, 94, 52 },
		{ 36, 95, 52 },
		{ 36, 96, 52 },
		{ 36, 97, 52 },
		{ 36, 98, 52 },
		{ 36, 99, 52 },
		{ 36, 100, 52 },
		{ 36, 101, 52 },
		{ 36, 102, 52 },
		{ 36, 103, 52 },
		{ 36, 104, 52 },
		{ 36, 105, 52 },
		{ 36, 106, 52 },
		{ 36, 107, 52 },
		{ 36, 108, 52 },
		{ 36, 109, 52 },
		{ 36, 110, 52 },
		{ 36, 111, 52 },
		{ 36, 112, 52 },
		{ 36, 113, 52 },
		{ 36, 114, 52 },
		{ 36, 115, 52 },
		{ 36, 116, 52 },
		{ 36, 117, 52 },
		{ 36, 118, 52 },
		{ 37, 65, 52 },
		{ 37, 66, 52 },
		{ 37, 67, 52 },
		{ 37, 68, 52 },
		{ 37, 69, 52 },
		{ 37, 70, 52 },
		{ 37, 71, 52 },
		{ 37, 72, 52 },
		{ 37, 73, 52 },
		{ 37, 74, 52 },
		{ 37, 75, 52 },
		{ 37, 76, 52 },
		{ 37, 77, 52 },
		{ 37, 78, 52 },
		{ 37, 79, 52 },
		{ 37, 80, 52 },
		{ 37, 81, 52 },
		{ 37, 82, 52 },
		{ 37, 83, 52 },
		{ 37, 84, 52 },
		{ 37, 85, 52 },
		{ 37, 86, 52 },
		{ 37, 87, 52 },
		{ 37, 88, 52 },
		{ 37, 89, 52 },
		{ 37, 90, 52 },
		{ 37, 91, 52 },
		{ 37, 92, 52 },
		{ 37, 93, 52 },
		{ 37, 94, 52 },
		{ 37, 95, 52 },
		{ 37, 96, 52 },
		{ 37, 97, 52 },
		{ 37, 98, 52 },
		{ 37, 99, 52 },
		{ 37, 100, 52 },
		{ 37, 101, 52 },
		{ 37, 102, 52 },
		{ 37, 103, 52 },
		{ 37, 104, 52 },
		{ 37, 105, 52 },
		{ 37, 106, 52 },
		{ 37, 107, 52 },
		{ 37, 108, 52 },
		{ 37, 109, 52 },
		{ 37, 110, 52 },
		{ 37, 111, 52 },
		{ 37, 112, 52 },
		{ 37, 113, 52 },
		{ 37, 114, 52 },
		{ 37, 115, 52 },
		{ 37, 116, 52 },
		{ 37, 117, 52 },
		{ 37, 118, 52 },
		{ 37, 119, 52 },
		{ 38, 66, 52 },
		{ 38, 67, 52 },
		{ 38, 68, 52 },
		{ 38, 69, 52 },
		{ 38, 70, 52 },
		{ 38, 71, 52 },
		{ 38, 72, 52 },
		{ 38, 73, 52 },
		{ 38, 74, 52 },
		{ 38, 75, 52 },
		{ 38, 76, 52 },
		{ 38, 77, 52 },
		{ 38, 78, 52 },
		{ 38, 79, 52 },
		{ 38, 80, 52 },
		{ 38, 81, 52 },
		{ 38, 82, 52 },
		{ 38, 83, 52 },
		{ 38, 84, 52 },
		{ 38, 85, 52 },
		{ 38, 86, 52 },
		{ 38, 87, 52 },
		{ 38, 88, 52 },
		{ 38, 89, 52 },
		{ 38, 90, 52 },
		{ 38, 91, 52 },
		{ 38, 92, 52 },
		{ 38, 93, 52 },
		{ 38, 94, 52 },
		{ 38, 95, 52 },
		{ 38, 96, 52 },
		{ 38, 97, 52 },
		{ 38, 98, 52 },
		{ 38, 99, 52 },
		{ 38, 100, 52 },
		{ 38, 101, 52 },
		{ 38, 102, 52 },
		{ 38, 103, 52 },
		{ 38, 104, 52 },
		{ 38, 105, 52 },
		{ 38, 106, 52 },
		{ 38, 107, 52 },
		{ 38, 108, 52 },
		{ 38, 109, 52 },
		{ 38, 110, 52 },
		{ 38, 111, 52 },
		{ 38, 112, 52 },
		{ 38, 113, 52 },
		{ 38, 114, 52 },
		{ 38, 115, 52 },
		{ 38, 116, 52 },
		{ 38, 117, 52 },
		{ 38, 118, 52 },
		{ 38, 119, 52 },
		{ 38, 120, 52 },
		{ 39, 67, 52 },
		{ 39, 68, 52 },
		{ 39, 69, 52 },
		{ 39, 70, 52 },
		{ 39, 71, 52 },
		{ 39, 72, 52 },
		{ 39, 73, 52 },
		{ 39, 74, 52 },
		{ 39, 75, 52 },
		{ 39, 76, 52 },
		{ 39, 77, 52 },
		{ 39, 78, 52 },
		{ 39, 79, 52 },
		{ 39, 80, 52 },
		{ 39, 81, 52 },
		{ 39, 82, 52 },
		{ 39, 83, 52 },
		{ 39, 84, 52 },
		{ 39, 85, 52 },
		{ 39, 86, 52 },
		{ 39, 87, 52 },
		{ 39, 88, 52 },
		{ 39, 89, 52 },
		{ 39, 90, 52 },
		{ 39, 91, 52 },
		{ 39, 92, 52 },
		{ 39, 93, 52 },
		{ 39, 94, 52 },
		{ 39, 95, 52 },
		{ 39, 96, 52 },
		{ 39, 97, 52 },
		{ 39, 98, 52 },
		{ 39, 99, 52 },
		{ 39, 100, 52 },
		{ 39, 101, 52 },
		{ 39, 102, 52 },
		{ 39, 103, 52 },
		{ 39, 104, 52 },
		{ 39, 105, 52 },
		{ 39, 106, 52 },
		{ 39, 107, 52 },
		{ 39, 108, 52 },
		{ 39, 109, 52 },
		{ 39, 110, 52 },
		{ 39, 111, 52 },
		{ 39, 112, 52 },
		{ 39, 113, 52 },
		{ 39, 114, 52 },
		{ 39, 115, 52 },
		{ 39, 116, 52 },
		{ 39, 117, 52 },
		{ 39, 118, 52 },
		{ 39, 119, 52 },
		{ 39, 120, 52 },
		{ 40, 68, 52 },
		{ 40, 69, 52 },
		{ 40, 70, 52 },
		{ 40, 71, 52 },
		{ 40, 72, 52 },
		{ 40, 73, 52 },
		{ 40, 74, 52 },
		{ 40, 75, 52 },
		{ 40, 76, 52 },
		{ 40, 77, 52 },
		{ 40, 78, 52 },
		{ 40, 79, 52 },
		{ 40, 80, 52 },
		{ 40, 81, 52 },
		{ 40, 82, 52 },
		{ 40, 83, 52 },
		{ 40, 84, 52 },
		{ 40, 85, 52 },
		{ 40, 86, 52 },
		{ 40, 87, 52 },
		{ 40, 88, 52 },
		{ 40, 89, 52 },
		{ 40, 90, 52 },
		{ 40, 91, 52 },
		{ 40, 92, 52 },
		{ 40, 93, 52 },
		{ 40, 94, 52 },
		{ 40, 95, 52 },
		{ 40, 96, 52 },
		{ 40, 97, 52 },
		{ 40, 98, 52 },
		{ 40, 99, 52 },
		{ 40, 100, 52 },
		{ 40, 101, 52 },
		{ 40, 102, 52 },
		{ 40, 103, 52 },
		{ 40, 104, 52 },
		{ 40, 105, 52 },
		{ 40, 106, 52 },
		{ 40, 107, 52 },
		{ 40, 108, 52 },
		{ 40, 109, 52 },
		{ 40, 110, 52 },
		{ 40, 111, 52 },
		{ 40, 112, 52 },
		{ 40, 113, 52 },
		{ 40, 114, 52 },
		{ 40, 115, 52 },
		{ 40, 116, 52 },
		{ 40, 117, 52 },
		{ 40, 118, 52 },
		{ 40, 119, 52 },
		{ 40, 120, 52 },
		{ 40, 121, 52 },
		{ 41, 69, 52 },
		{ 41, 70, 52 },
		{ 41, 71, 52 },
		{ 41, 72, 52 },
		{ 41, 73, 52 },
		{ 41, 74, 52 },
		{ 41, 75, 52 },
		{ 41, 76, 52 },
		{ 41, 77, 52 },
		{ 41, 78, 52 },
		{ 41, 79, 52 },
		{ 41, 80, 52 },
		{ 41, 81, 52 },
		{ 41, 82, 52 },
		{ 41, 83, 52 },
		{ 41, 84, 52 },
		{ 41, 85, 52 },
		{ 41, 86, 52 },
		{ 41, 87, 52 },
		{ 41, 88, 52 },
		{ 41, 89, 52 },
		{ 41, 90, 52 },
		{ 41, 91, 52 },
		{ 41, 92, 52 },
		{ 41, 93, 52 },
		{ 41, 94, 52 },
		{ 41, 95, 52 },
		{ 41, 96, 52 },
		{ 41, 97, 52 },
		{ 41, 98, 52 },
		{ 41, 99, 52 },
		{ 41, 100, 52 },
		{ 41, 101, 52 },
		{ 41, 102, 52 },
		{ 41, 103, 52 },
		{ 41, 104, 52 },
		{ 41, 105, 52 },
		{ 41, 106, 52 },
		{ 41, 107, 52 },
		{ 41, 108, 52 },
		{ 41, 109, 52 },
		{ 41, 110, 52 },
		{ 41, 111, 52 },
		{ 41, 112, 52 },
		{ 41, 113, 52 },
		{ 41, 114, 52 },
		{ 41, 115, 52 },
		{ 41, 116, 52 },
		{ 41, 117, 52 },
		{ 41, 118, 52 },
		{ 41, 119, 52 },
		{ 41, 120, 52 },
		{ 41, 121, 52 },
		{ 42, 69, 52 },
		{ 42, 70, 52 },
		{ 42, 71, 52 },
		{ 42, 72, 52 },
		{ 42, 73, 52 },
		{ 42, 74, 52 },
		{ 42, 75, 52 },
		{ 42, 76, 52 },
		{ 42, 77, 52 },
		{ 42, 78, 52 },
		{ 42, 79, 52 },
		{ 42, 80, 52 },
		{ 42, 81, 52 },
		{ 42, 82, 52 },
		{ 42, 83, 52 },
		{ 42, 84, 52 },
		{ 42, 85, 52 },
		{ 42, 86, 52 },
		{ 42, 87, 52 },
		{ 42, 88, 52 },
		{ 42, 89, 52 },
		{ 42, 90, 52 },
		{ 42, 91, 52 },
		{ 42, 92, 52 },
		{ 42, 93, 52 },
		{ 42, 94, 52 },
		{ 42, 95, 52 },
		{ 42, 96, 52 },
		{ 42, 97, 52 },
		{ 42, 98, 52 },
		{ 42, 99, 52 },
		{ 42, 100, 52 },
		{ 42, 101, 52 },
		{ 42, 102, 52 },
		{ 42, 103, 52 },
		{ 42, 104, 52 },
		{ 42, 105, 52 },
		{ 42, 106, 52 },
		{ 42, 107, 52 },
		{ 42, 108, 52 },
		{ 42, 109, 52 },
		{ 42, 110, 52 },
		{ 42, 111, 52 },
		{ 42, 112, 52 },
		{ 42, 113, 52 },
		{ 42, 114, 52 },
		{ 42, 115, 52 },
		{ 42, 116, 52 },
		{ 42, 117, 52 },
		{ 42, 118, 52 },
		{ 42, 119, 52 },
		{ 42, 120, 52 },
		{ 42, 121, 52 },
		{ 42, 122, 52 },
		{ 43, 70, 52 },
		{ 43, 71, 52 },
		{ 43, 72, 52 },
		{ 43, 73, 52 },
		{ 43, 74, 52 },
		{ 43, 75, 52 },
		{ 43, 76, 52 },
		{ 43, 77, 52 },
		{ 43, 78, 52 },
		{ 43, 79, 52 },
		{ 43, 80, 52 },
		{ 43, 81, 52 },
		{ 43, 82, 52 },
		{ 43, 83, 52 },
		{ 43, 84, 52 },
		{ 43, 85, 52 },
		{ 43, 86, 52 },
		{ 43, 87, 52 },
		{ 43, 88, 52 },
		{ 43, 89, 52 },
		{ 43, 90, 52 },
		{ 43, 91, 52 },
		{ 43, 92, 52 },
		{ 43, 93, 52 },
		{ 43, 94, 52 },
		{ 43, 95, 52 },
		{ 43, 96, 52 },
		{ 43, 97, 52 },
		{ 43, 98, 52 },
		{ 43, 99, 52 },
		{ 43, 100, 52 },
		{ 43, 101, 52 },
		{ 43, 102, 52 },
		{ 43, 103, 52 },
		{ 43, 104, 52 },
		{ 43, 105, 52 },
		{ 43, 106, 52 },
		{ 43, 107, 52 },
		{ 43, 108, 52 },
		{ 43, 109, 52 },
		{ 43, 110, 52 },
		{ 43, 111, 52 },
		{ 43, 112, 52 },
		{ 43, 113, 52 },
		{ 43, 114, 52 },
		{ 43, 115, 52 },
		{ 43, 116, 52 },
		{ 43, 117, 52 },
		{ 43, 118, 52 },
		{ 43, 119, 52 },
		{ 43, 120, 52 },
		{ 43, 121, 52 },
		{ 43, 122, 52 },
		{ 43, 123, 52 },
		{ 44, 71, 52 },
		{ 44, 72, 52 },
		{ 44, 73, 52 },
		{ 44, 74, 52 },
		{ 44, 75, 52 },
		{ 44, 76, 52 },
		{ 44, 77, 52 },
		{ 44, 78, 52 },
		{ 44, 79, 52 },
		{ 44, 80, 52 },
		{ 44, 81, 52 },
		{ 44, 82, 52 },
		{ 44, 83, 52 },
		{ 44, 84, 52 },
		{ 44, 85, 52 },
		{ 44, 86, 52 },
		{ 44, 87, 52 },
		{ 44, 88, 52 },
		{ 44, 89, 52 },
		{ 44, 90, 52 },
		{ 44, 91, 52 },
		{ 44, 92, 52 },
		{ 44, 93, 52 },
		{ 44, 94, 52 },
		{ 44, 95, 52 },
		{ 44, 96, 52 },
		{ 44, 97, 52 },
		{ 44, 98, 52 },
		{ 44, 99, 52 },
		{ 44, 100, 52 },
		{ 44, 101, 52 },
		{ 44, 102, 52 },
		{ 44, 103, 52 },
		{ 44, 104, 52 },
		{ 44, 105, 52 },
		{ 44, 106, 52 },
		{ 44, 107, 52 },
		{ 44, 108, 52 },
		{ 44, 109, 52 },
		{ 44, 110, 52 },
		{ 44, 111, 52 },
		{ 44, 112, 52 },
		{ 44, 113, 52 },
		{ 44, 114, 52 },
		{ 44, 115, 52 },
		{ 44, 116, 52 },
		{ 44, 117, 52 },
		{ 44, 118, 52 },
		{ 44, 119, 52 },
		{ 44, 120, 52 },
		{ 44, 121, 52 },
		{ 44, 122, 52 },
		{ 44, 123, 52 },
		{ 45, 72, 52 },
		{ 45, 73, 52 },
		{ 45, 74, 52 },
		{ 45, 75, 52 },
		{ 45, 76, 52 },
		{ 45, 77, 52 },
		{ 45, 78, 52 },
		{ 45, 79, 52 },
		{ 45, 80, 52 },
		{ 45, 81, 52 },
		{ 45, 82, 52 },
		{ 45, 83, 52 },
		{ 45, 84, 52 },
		{ 45, 85, 52 },
		{ 45, 86, 52 },
		{ 45, 87, 52 },
		{ 45, 88, 52 },
		{ 45, 89, 52 },
		{ 45, 90, 52 },
		{ 45, 91, 52 },
		{ 45, 92, 52 },
		{ 45, 93, 52 },
		{ 45, 94, 52 },
		{ 45, 95, 52 },
		{ 45, 96, 52 },
		{ 45, 97, 52 },
		{ 45, 98, 52 },
		{ 45, 99, 52 },
		{ 45, 100, 52 },
		{ 45, 101, 52 },
		{ 45, 102, 52 },
		{ 45, 103, 52 },
		{ 45, 104, 52 },
		{ 45, 105, 52 },
		{ 45, 106, 52 },
		{ 45, 107, 52 },
		{ 45, 108, 52 },
		{ 45, 109, 52 },
		{ 45, 110, 52 },
		{ 45, 111, 52 },
		{ 45, 112, 52 },
		{ 45, 113, 52 },
		{ 45, 114, 52 },
		{ 45, 115, 52 },
		{ 45, 116, 52 },
		{ 45, 117, 52 },
		{ 45, 118, 52 },
		{ 45, 119, 52 },
		{ 45, 120, 52 },
		{ 45, 121, 52 },
		{ 45, 122, 52 },
		{ 45, 123, 52 },
		{ 45, 124, 52 },
		{ 46, 73, 52 },
		{ 46, 74, 52 },
		{ 46, 75, 52 },
		{ 46, 76, 52 },
		{ 46, 77, 52 },
		{ 46, 78, 52 },
		{ 46, 79, 52 },
		{ 46, 80, 52 },
		{ 46, 81, 52 },
		{ 46, 82, 52 },
		{ 46, 83, 52 },
		{ 46, 84, 52 },
		{ 46, 85, 52 },
		{ 46, 86, 52 },
		{ 46, 87, 52 },
		{ 46, 88, 52 },
		{ 46, 89, 52 },
		{ 46, 90, 52 },
		{ 46, 91, 52 },
		{ 46, 92, 52 },
		{ 46, 93, 52 },
		{ 46, 94, 52 },
		{ 46, 95, 52 },
		{ 46, 96, 52 },
		{ 46, 97, 52 },
		{ 46, 98, 52 },
		{ 46, 99, 52 },
		{ 46, 100, 52 },
		{ 46, 101, 52 },
		{ 46, 102, 52 },
		{ 46, 103, 52 },
		{ 46, 104, 52 },
		{ 46, 105, 52 },
		{ 46, 106, 52 },
		{ 46, 107, 52 },
		{ 46, 108, 52 },
		{ 46, 109, 52 },
		{ 46, 110, 52 },
		{ 46, 111, 52 },
		{ 46, 112, 52 },
		{ 46, 113, 52 },
		{ 46, 114, 52 },
		{ 46, 115, 52 },
		{ 46, 116, 52 },
		{ 46, 117, 52 },
		{ 46, 118, 52 },
		{ 46, 119, 52 },
		{ 46, 120, 52 },
		{ 46, 121, 52 },
		{ 46, 122, 52 },
		{ 46, 123, 52 },
		{ 46, 124, 52 },
		{ 46, 125, 52 },
		{ 47, 74, 52 },
		{ 47, 75, 52 },
		{ 47, 76, 52 },
		{ 47, 77, 52 },
		{ 47, 78, 52 },
		{ 47, 79, 52 },
		{ 47, 80, 52 },
		{ 47, 81, 52 },
		{ 47, 82, 52 },
		{ 47, 83, 52 },
		{ 47, 84, 52 },
		{ 47, 85, 52 },
		{ 47, 86, 52 },
		{ 47, 87, 52 },
		{ 47, 88, 52 },
		{ 47, 89, 52 },
		{ 47, 90, 52 },
		{ 47, 91, 52 },
		{ 47, 92, 52 },
		{ 47, 93, 52 },
		{ 47, 94, 52 },
		{ 47, 95, 52 },
		{ 47, 96, 52 },
		{ 47, 97, 52 },
		{ 47, 98, 52 },
		{ 47, 99, 52 },
		{ 47, 100, 52 },
		{ 47, 101, 52 },
		{ 47, 102, 52 },
		{ 47, 103, 52 },
		{ 47, 104, 52 },
		{ 47, 105, 52 },
		{ 47, 106, 52 },
		{ 47, 107, 52 },
		{ 47, 108, 52 },
		{ 47, 109, 52 },
		{ 47, 110, 52 },
		{ 47, 111, 52 },
		{ 47, 112, 52 },
		{ 47, 113, 52 },
		{ 47, 114, 52 },
		{ 47, 115, 52 },
		{ 47, 116, 52 },
		{ 47, 117, 52 },
		{ 47, 118, 52 },
		{ 47, 119, 52 },
		{ 47, 120, 52 },
		{ 47, 121, 52 },
		{ 47, 122, 52 },
		{ 47, 123, 52 },
		{ 47, 124, 52 },
		{ 47, 125, 52 },
		{ 48, 75, 52 },
		{ 48, 76, 52 },
		{ 48, 77, 52 },
		{ 48, 78, 52 },
		{ 48, 79, 52 },
		{ 48, 80, 52 },
		{ 48, 81, 52 },
		{ 48, 82, 52 },
		{ 48, 83, 52 },
		{ 48, 84, 52 },
		{ 48, 85, 52 },
		{ 48, 86, 52 },
		{ 48, 87, 52 },
		{ 48, 88, 52 },
		{ 48, 89, 52 },
		{ 48, 90, 52 },
		{ 48, 91, 52 },
		{ 48, 92, 52 },
		{ 48, 93, 52 },
		{ 48, 94, 52 },
		{ 48, 95, 52 },
		{ 48, 96, 52 },
		{ 48, 97, 52 },
		{ 48, 98, 52 },
		{ 48, 99, 52 },
		{ 48, 100, 52 },
		{ 48, 101, 52 },
		{ 48, 102, 52 },
		{ 48, 103, 52 },
		{ 48, 104, 52 },
		{ 48, 105, 52 },
		{ 48, 106, 52 },
		{ 48, 107, 52 },
		{ 48, 108, 52 },
		{ 48, 109, 52 },
		{ 48, 110, 52 },
		{ 48, 111, 52 },
		{ 48, 112, 52 },
		{ 48, 113, 52 },
		{ 48, 114, 52 },
		{ 48, 115, 52 },
		{ 48, 116, 52 },
		{ 48, 117, 52 },
		{ 48, 118, 52 },
		{ 48, 119, 52 },
		{ 48, 120, 52 },
		{ 48, 121, 52 },
		{ 48, 122, 52 },
		{ 48, 123, 52 },
		{ 48, 124, 52 },
		{ 48, 125, 52 },
		{ 48, 126, 52 },
		{ 49, 76, 52 },
		{ 49, 77, 52 },
		{ 49, 78, 52 },
		{ 49, 79, 52 },
		{ 49, 80, 52 },
		{ 49, 81, 52 },
		{ 49, 82, 52 },
		{ 49, 83, 52 },
		{ 49, 84, 52 },
		{ 49, 85, 52 },
		{ 49, 86, 52 },
		{ 49, 87, 52 },
		{ 49, 88, 52 },
		{ 49, 89, 52 },
		{ 49, 90, 52 },
		{ 49, 91, 52 },
		{ 49, 92, 52 },
		{ 49, 93, 52 },
		{ 49, 94, 52 },
		{ 49, 95, 52 },
		{ 49, 96, 52 },
		{ 49, 97, 52 },
		{ 49, 98, 52 },
		{ 49, 99, 52 },
		{ 49, 100, 52 },
		{ 49, 101, 52 },
		{ 49, 102, 52 },
		{ 49, 103, 52 },
		{ 49, 104, 52 },
		{ 49, 105, 52 },
		{ 49, 106, 52 },
		{ 49, 107, 52 },
		{ 49, 108, 52 },
		{ 49, 109, 52 },
		{ 49, 110, 52 },
		{ 49, 111, 52 },
		{ 49, 112, 52 },
		{ 49, 113, 52 },
		{ 49, 114, 52 },
		{ 49, 115, 52 },
		{ 49, 116, 52 },
		{ 49, 117, 52 },
		{ 49, 118, 52 },
		{ 49, 119, 52 },
		{ 49, 120, 52 },
		{ 49, 121, 52 },
		{ 49, 122, 52 },
		{ 49, 123, 52 },
		{ 49, 124, 52 },
		{ 49, 125, 52 },
		{ 49, 126, 52 },
		{ 50, 76, 52 },
		{ 50, 77, 52 },
		{ 50, 78, 52 },
		{ 50, 79, 52 },
		{ 50, 80, 52 },
		{ 50, 81, 52 },
		{ 50, 82, 52 },
		{ 50, 83, 52 },
		{ 50, 84, 52 },
		{ 50, 85, 52 },
		{ 50, 86, 52 },
		{ 50, 87, 52 },
		{ 50, 88, 52 },
		{ 50, 89, 52 },
		{ 50, 90, 52 },
		{ 50, 91, 52 },
		{ 50, 92, 52 },
		{ 50, 93, 52 },
		{ 50, 94, 52 },
		{ 50, 95, 52 },
		{ 50, 96, 52 },
		{ 50, 97, 52 },
		{ 50, 98, 52 },
		{ 50, 99, 52 },
		{ 50, 100, 52 },
		{ 50, 101, 52 },
		{ 50, 102, 52 },
		{ 50, 103, 52 },
		{ 50, 104, 52 },
		{ 50, 105, 52 },
		{ 50, 106, 52 },
		{ 50, 107, 52 },
		{ 50, 108, 52 },
		{ 50, 109, 52 },
		{ 50, 110, 52 },
		{ 50, 111, 52 },
		{ 50, 112, 52 },
		{ 50, 113, 52 },
		{ 50, 114, 52 },
		{ 50, 115, 52 },
		{ 50, 116, 52 },
		{ 50, 117, 52 },
		{ 50, 118, 52 },
		{ 50, 119, 52 },
		{ 50, 120, 52 },
		{ 50, 121, 52 },
		{ 50, 122, 52 },
		{ 50, 123, 52 },
		{ 50, 124, 52 },
		{ 50, 125, 52 },
		{ 50, 126, 52 },
		{ 50, 127, 52 },
		{ 51, 77, 52 },
		{ 51, 78, 52 },
		{ 51, 79, 52 },
		{ 51, 80, 52 },
		{ 51, 81, 52 },
		{ 51, 82, 52 },
		{ 51, 83, 52 },
		{ 51, 84, 52 },
		{ 51, 85, 52 },
		{ 51, 86, 52 },
		{ 51, 87, 52 },
		{ 51, 88, 52 },
		{ 51, 89, 52 },
		{ 51, 90, 52 },
		{ 51, 91, 52 },
		{ 51, 92, 52 },
		{ 51, 93, 52 },
		{ 51, 94, 52 },
		{ 51, 95, 52 },
		{ 51, 96, 52 },
		{ 51, 97, 52 },
		{ 51, 98, 52 },
		{ 51, 99, 52 },
		{ 51, 100, 52 },
		{ 51, 101, 52 },
		{ 51, 102, 52 },
		{ 51, 103, 52 },
		{ 51, 104, 52 },
		{ 51, 105, 52 },
		{ 51, 106, 52 },
		{ 51, 107, 52 },
		{ 51, 108, 52 },
		{ 51, 109, 52 },
		{ 51, 110, 52 },
		{ 51, 111, 52 },
		{ 51, 112, 52 },
		{ 51, 113, 52 },
		{ 51, 114, 52 },
		{ 51, 115, 52 },
		{ 51, 116, 52 },
		{ 51, 117, 52 },
		{ 51, 118, 52 },
		{ 51, 119, 52 },
		{ 51, 120, 52 },
		{ 51, 121, 52 },
		{ 51, 122, 52 },
		{ 51, 123, 52 },
		{ 51, 124, 52 },
		{ 51, 125, 52 },
		{ 51, 126, 52 },
		{ 51, 127, 52 },
		{ 51, 128, 52 },
		{ 52, 78, 52 },
		{ 52, 79, 52 },
		{ 52, 80, 52 },
		{ 52, 81, 52 },
		{ 52, 82, 52 },
		{ 52, 83, 52 },
		{ 52, 84, 52 },
		{ 52, 85, 52 },
		{ 52, 86, 52 },
		{ 52, 87, 52 },
		{ 52, 88, 52 },
		{ 52, 89, 52 },
		{ 52, 90, 52 },
		{ 52, 91, 52 },
		{ 52, 92, 52 },
		{ 52, 93, 52 },
		{ 52, 94, 52 },
		{ 52, 95, 52 },
		{ 52, 96, 52 },
		{ 52, 97, 52 },
		{ 52, 98, 52 },
		{ 52, 99, 52 },
		{ 52, 100, 52 },
		{ 52, 101, 52 },
		{ 52, 102, 52 },
		{ 52, 103, 52 },
		{ 52, 104, 52 },
		{ 52, 105, 52 },
		{ 52, 106, 52 },
		{ 52, 107, 52 },
		{ 52, 108, 52 },
		{ 52, 109, 52 },
		{ 52, 110, 52 },
		{ 52, 111, 52 },
		{ 52, 112, 52 },
		{ 52, 113, 52 },
		{ 52, 114, 52 },
		{ 52, 115, 52 },
		{ 52, 116, 52 },
		{ 52, 117, 52 },
		{ 52, 118, 52 },
		{ 52, 119, 52 },
		{ 52, 120, 52 },
		{ 52, 121, 52 },
		{ 52, 122, 52 },
		{ 52, 123, 52 },
		{ 52, 124, 52 },
		{ 52, 125, 52 },
		{ 52, 126, 52 },
		{ 52, 127, 52 },
		{ 52, 128, 52 },
		{ 53, 79, 52 },
		{ 53, 80, 52 },
		{ 53, 81, 52 },
		{ 53, 82, 52 },
		{ 53, 83, 52 },
		{ 53, 84, 52 },
		{ 53, 85, 52 },
		{ 53, 86, 52 },
		{ 53, 87, 52 },
		{ 53, 88, 52 },
		{ 53, 89, 52 },
		{ 53, 90, 52 },
		{ 53, 91, 52 },
		{ 53, 92, 52 },
		{ 53, 93, 52 },
		{ 53, 94, 52 },
		{ 53, 95, 52 },
		{ 53, 96, 52 },
		{ 53, 97, 52 },
		{ 53, 98, 52 },
		{ 53, 99, 52 },
		{ 53, 100, 52 },
		{ 53, 101, 52 },
		{ 53, 102, 52 },
		{ 53, 103, 52 },
		{ 53, 104, 52 },
		{ 53, 105, 52 },
		{ 53, 106, 52 },
		{ 53, 107, 52 },
		{ 53, 108, 52 },
		{ 53, 109, 52 },
		{ 53, 110, 52 },
		{ 53, 111, 52 },
		{ 53, 112, 52 },
		{ 53, 113, 52 },
		{ 53, 114, 52 },
		{ 53, 115, 52 },
		{ 53, 116, 52 },
		{ 53, 117, 52 },
		{ 53, 118, 52 },
		{ 53, 119, 52 },
		{ 53, 120, 52 },
		{ 53, 121, 52 },
		{ 53, 122, 52 },
		{ 53, 123, 52 },
		{ 53, 124, 52 },
		{ 53, 125, 52 },
		{ 53, 126, 52 },
		{ 53, 127, 52 },
		{ 53, 128, 52 },
		{ 53, 129, 52 },
		{ 54, 80, 52 },
		{ 54, 81, 52 },
		{ 54, 82, 52 },
		{ 54, 83, 52 },
		{ 54, 84, 52 },
		{ 54, 85, 52 },
		{ 54, 86, 52 },
		{ 54, 87, 52 },
		{ 54, 88, 52 },
		{ 54, 89, 52 },
		{ 54, 90, 52 },
		{ 54, 91, 52 },
		{ 54, 92, 52 },
		{ 54, 93, 52 },
		{ 54, 94, 52 },
		{ 54, 95, 52 },
		{ 54, 96, 52 },
		{ 54, 97, 52 },
		{ 54, 98, 52 },
		{ 54, 99, 52 },
		{ 54, 100, 52 },
		{ 54, 101, 52 },
		{ 54, 102, 52 },
		{ 54, 103, 52 },
		{ 54, 104, 52 },
		{ 54, 105, 52 },
		{ 54, 106, 52 },
		{ 54, 107, 52 },
		{ 54, 108, 52 },
		{ 54, 109, 52 },
		{ 54, 110, 52 },
		{ 54, 111, 52 },
		{ 54, 112, 52 },
		{ 54, 113, 52 },
		{ 54, 114, 52 },
		{ 54, 115, 52 },
		{ 54, 116, 52 },
		{ 54, 117, 52 },
		{ 54, 118, 52 },
		{ 54, 119, 52 },
		{ 54, 120, 52 },
		{ 54, 121, 52 },
		{ 54, 122, 52 },
		{ 54, 123, 52 },
		{ 54, 124, 52 },
		{ 54, 125, 52 },
		{ 54, 126, 52 },
		{ 54, 127, 52 },
		{ 54, 128, 52 },
		{ 54, 129, 52 },
		{ 54, 130, 52 },
		{ 55, 81, 52 },
		{ 55, 82, 52 },
		{ 55, 83, 52 },
		{ 55, 84, 52 },
		{ 55, 85, 52 },
		{ 55, 86, 52 },
		{ 55, 87, 52 },
		{ 55, 88, 52 },
		{ 55, 89, 52 },
		{ 55, 90, 52 },
		{ 55, 91, 52 },
		{ 55, 92, 52 },
		{ 55, 93, 52 },
		{ 55, 94, 52 },
		{ 55, 95, 52 },
		{ 55, 96, 52 },
		{ 55, 97, 52 },
		{ 55, 98, 52 },
		{ 55, 99, 52 },
		{ 55, 100, 52 },
		{ 55, 101, 52 },
		{ 55, 102, 52 },
		{ 55, 103, 52 },
		{ 55, 104, 52 },
		{ 55, 105, 52 },
		{ 55, 106, 52 },
		{ 55, 107, 52 },
		{ 55, 108, 52 },
		{ 55, 109, 52 },
		{ 55, 110, 52 },
		{ 55, 111, 52 },
		{ 55, 112, 52 },
		{ 55, 113, 52 },
		{ 55, 114, 52 },
		{ 55, 115, 52 },
		{ 55, 116, 52 },
		{ 55, 117, 52 },
		{ 55, 118, 52 },
		{ 55, 119, 52 },
		{ 55, 120, 52 },
		{ 55, 121, 52 },
		{ 55, 122, 52 },
		{ 55, 123, 52 },
		{ 55, 124, 52 },
		{ 55, 125, 52 },
		{ 55, 126, 52 },
		{ 55, 127, 52 },
		{ 55, 128, 52 },
		{ 55, 129, 52 },
		{ 55, 130, 52 },
		{ 56, 82, 52 },
		{ 56, 83, 52 },
		{ 56, 84, 52 },
		{ 56, 85, 52 },
		{ 56, 86, 52 },
		{ 56, 87, 52 },
		{ 56, 88, 52 },
		{ 56, 89, 52 },
		{ 56, 90, 52 },
		{ 56, 91, 52 },
		{ 56, 92, 52 },
		{ 56, 93, 52 },
		{ 56, 94, 52 },
		{ 56, 95, 52 },
		{ 56, 96, 52 },
		{ 56, 97, 52 },
		{ 56, 98, 52 },
		{ 56, 99, 52 },
		{ 56, 100, 52 },
		{ 56, 101, 52 },
		{ 56, 102, 52 },
		{ 56, 103, 52 },
		{ 56, 104, 52 },
		{ 56, 105, 52 },
		{ 56, 106, 52 },
		{ 56, 107, 52 },
		{ 56, 108, 52 },
		{ 56, 109, 52 },
		{ 56, 110, 52 },
		{ 56, 111, 52 },
		{ 56, 112, 52 },
		{ 56, 113, 52 },
		{ 56, 114, 52 },
		{ 56, 115, 52 },
		{ 56, 116, 52 },
		{ 56, 117, 52 },
		{ 56, 118, 52 },
		{ 56, 119, 52 },
		{ 56, 120, 52 },
		{ 56, 121, 52 },
		{ 56, 122, 52 },
		{ 56, 123, 52 },
		{ 56, 124, 52 },
		{ 56, 125, 52 },
		{ 56, 126, 52 },
		{ 56, 127, 52 },
		{ 56, 128, 52 },
		{ 56, 129, 52 },
		{ 56, 130, 52 },
		{ 56, 131, 52 },
		{ 57, 83, 52 },
		{ 57, 84, 52 },
		{ 57, 85, 52 },
		{ 57, 86, 52 },
		{ 57, 87, 52 },
		{ 57, 88, 52 },
		{ 57, 89, 52 },
		{ 57, 90, 52 },
		{ 57, 91, 52 },
		{ 57, 92, 52 },
		{ 57, 93, 52 },
		{ 57, 94, 52 },
		{ 57, 95, 52 },
		{ 57, 96, 52 },
		{ 57, 97, 52 },
		{ 57, 98, 52 },
		{ 57, 99, 52 },
		{ 57, 100, 52 },
		{ 57, 101, 52 },
		{ 57, 102, 52 },
		{ 57, 103, 52 },
		{ 57, 104, 52 },
		{ 57, 105, 52 },
		{ 57, 106, 52 },
		{ 57, 107, 52 },
		{ 57, 108, 52 },
		{ 57, 109, 52 },
		{ 57, 110, 52 },
		{ 57, 111, 52 },
		{ 57, 112, 52 },
		{ 57, 113, 52 },
		{ 57, 114, 52 },
		{ 57, 115, 52 },
		{ 57, 116, 52 },
		{ 57, 117, 52 },
		{ 57, 118, 52 },
		{ 57, 119, 52 },
		{ 57, 120, 52 },
		{ 57, 121, 52 },
		{ 57, 122, 52 },
		{ 57, 123, 52 },
		{ 57, 124, 52 },
		{ 57, 125, 52 },
		{ 57, 126, 52 },
		{ 57, 127, 52 },
		{ 57, 128, 52 },
		{ 57, 129, 52 },
		{ 57, 130, 52 },
		{ 57, 131, 52 },
		{ 58, 83, 52 },
		{ 58, 84, 52 },
		{ 58, 85, 52 },
		{ 58, 86, 52 },
		{ 58, 87, 52 },
		{ 58, 88, 52 },
		{ 58, 89, 52 },
		{ 58, 90, 52 },
		{ 58, 91, 52 },
		{ 58, 92, 52 },
		{ 58, 93, 52 },
		{ 58, 94, 52 },
		{ 58, 95, 52 },
		{ 58, 96, 52 },
		{ 58, 97, 52 },
		{ 58, 98, 52 },
		{ 58, 99, 52 },
		{ 58, 100, 52 },
		{ 58, 101, 52 },
		{ 58, 102, 52 },
		{ 58, 103, 52 },
		{ 58, 104, 52 },
		{ 58, 105, 52 },
		{ 58, 106, 52 },
		{ 58, 107, 52 },
		{ 58, 108, 52 },
		{ 58, 109, 52 },
		{ 58, 110, 52 },
		{ 58, 111, 52 },
		{ 58, 112, 52 },
		{ 58, 113, 52 },
		{ 58, 114, 52 },
		{ 58, 115, 52 },
		{ 58, 116, 52 },
		{ 58, 117, 52 },
		{ 58, 118, 52 },
		{ 58, 119, 52 },
		{ 58, 120, 52 },
		{ 58, 121, 52 },
		{ 58, 122, 52 },
		{ 58, 123, 52 },
		{ 58, 124, 52 },
		{ 58, 125, 52 },
		{ 58, 126, 52 },
		{ 58, 127, 52 },
		{ 58, 128, 52 },
		{ 58, 129, 52 },
		{ 58, 130, 52 },
		{ 58, 131, 52 },
		{ 58, 132, 52 },
		{ 59, 84, 52 },
		{ 59, 85, 52 },
		{ 59, 86, 52 },
		{ 59, 87, 52 },
		{ 59, 88, 52 },
		{ 59, 89, 52 },
		{ 59, 90, 52 },
		{ 59, 91, 52 },
		{ 59, 92, 52 },
		{ 59, 93, 52 },
		{ 59, 94, 52 },
		{ 59, 95, 52 },
		{ 59, 96, 52 },
		{ 59, 97, 52 },
		{ 59, 98, 52 },
		{ 59, 99, 52 },
		{ 59, 100, 52 },
		{ 59, 101, 52 },
		{ 59, 102, 52 },
		{ 59, 103, 52 },
		{ 59, 104, 52 },
		{ 59, 105, 52 },
		{ 59, 106, 52 },
		{ 59, 107, 52 },
		{ 59, 108, 52 },
		{ 59, 109, 52 },
		{ 59, 110, 52 },
		{ 59, 111, 52 },
		{ 59, 112, 52 },
		{ 59, 113, 52 },
		{ 59, 114, 52 },
		{ 59, 115, 52 },
		{ 59, 116, 52 },
		{ 59, 117, 52 },
		{ 59, 118, 52 },
		{ 59, 119, 52 },
		{ 59, 120, 52 },
		{ 59, 121, 52 },
		{ 59, 122, 52 },
		{ 59, 123, 52 },
		{ 59, 124, 52 },
		{ 59, 125, 52 },
		{ 59, 126, 52 },
		{ 59, 127, 52 },
		{ 59, 128, 52 },
		{ 59, 129, 52 },
		{ 59, 130, 52 },
		{ 59, 131, 52 },
		{ 59, 132, 52 },
		{ 59, 133, 52 },
		{ 60, 85, 52 },
		{ 60, 86, 52 },
		{ 60, 87, 52 },
		{ 60, 88, 52 },
		{ 60, 89, 52 },
		{ 60, 90, 52 },
		{ 60, 91, 52 },
		{ 60, 92, 52 },
		{ 60, 93, 52 },
		{ 60, 94, 52 },
		{ 60, 95, 52 },
		{ 60, 96, 52 },
		{ 60, 97, 52 },
		{ 60, 98, 52 },
		{ 60, 99, 52 },
		{ 60, 100, 52 },
		{ 60, 101, 52 },
		{ 60, 102, 52 },
		{ 60, 103, 52 },
		{ 60, 104, 52 },
		{ 60, 105, 52 },
		{ 60, 106, 52 },
		{ 60, 107, 52 },
		{ 60, 108, 52 },
		{ 60, 109, 52 },
		{ 60, 110, 52 },
		{ 60, 111, 52 },
		{ 60, 112, 52 },
		{ 60, 113, 52 },
		{ 60, 114, 52 },
		{ 60, 115, 52 },
		{ 60, 116, 52 },
		{ 60, 117, 52 },
		{ 60, 118, 52 },
		{ 60, 119, 52 },
		{ 60, 120, 52 },
		{ 60, 121, 52 },
		{ 60, 122, 52 },
		{ 60, 123, 52 },
		{ 60, 124, 52 },
		{ 60, 125, 52 },
		{ 60, 126, 52 },
		{ 60, 127, 52 },
		{ 60, 128, 52 },
		{ 60, 129, 52 },
		{ 60, 130, 52 },
		{ 60, 131, 52 },
		{ 60, 132, 52 },
		{ 60, 133, 52 },
		{ 61, 86, 52 },
		{ 61, 87, 52 },
		{ 61, 88, 52 },
		{ 61, 89, 52 },
		{ 61, 90, 52 },
		{ 61, 91, 52 },
		{ 61, 92, 52 },
		{ 61, 93, 52 },
		{ 61, 94, 52 },
		{ 61, 95, 52 },
		{ 61, 96, 52 },
		{ 61, 97, 52 },
		{ 61, 98, 52 },
		{ 61, 99, 52 },
		{ 61, 100, 52 },
		{ 61, 101, 52 },
		{ 61, 102, 52 },
		{ 61, 103, 52 },
		{ 61, 104, 52 },
		{ 61, 105, 52 },
		{ 61, 106, 52 },
		{ 61, 107, 52 },
		{ 61, 108, 52 },
		{ 61, 109, 52 },
		{ 61, 110, 52 },
		{ 61, 111, 52 },
		{ 61, 112, 52 },
		{ 61, 113, 52 },
		{ 61, 114, 52 },
		{ 61, 115, 52 },
		{ 61, 116, 52 },
		{ 61, 117, 52 },
		{ 61, 118, 52 },
		{ 61, 119, 52 },
		{ 61, 120, 52 },
		{ 61, 121, 52 },
		{ 61, 122, 52 },
		{ 61, 123, 52 },
		{ 61, 124, 52 },
		{ 61, 125, 52 },
		{ 61, 126, 52 },
		{ 61, 127, 52 },
		{ 61, 128, 52 },
		{ 61, 129, 52 },
		{ 61, 130, 52 },
		{ 61, 131, 52 },
		{ 61, 132, 52 },
		{ 61, 133, 52 },
		{ 61, 134, 52 },
		{ 62, 87, 52 },
		{ 62, 88, 52 },
		{ 62, 89, 52 },
		{ 62, 90, 52 },
		{ 62, 91, 52 },
		{ 62, 92, 52 },
		{ 62, 93, 52 },
		{ 62, 94, 52 },
		{ 62, 95, 52 },
		{ 62, 96, 52 },
		{ 62, 97, 52 },
		{ 62, 98, 52 },
		{ 62, 99, 52 },
		{ 62, 100, 52 },
		{ 62, 101, 52 },
		{ 62, 102, 52 },
		{ 62, 103, 52 },
		{ 62, 104, 52 },
		{ 62, 105, 52 },
		{ 62, 106, 52 },
		{ 62, 107, 52 },
		{ 62, 108, 52 },
		{ 62, 109, 52 },
		{ 62, 110, 52 },
		{ 62, 111, 52 },
		{ 62, 112, 52 },
		{ 62, 113, 52 },
		{ 62, 114, 52 },
		{ 62, 115, 52 },
		{ 62, 116, 52 },
		{ 62, 117, 52 },
		{ 62, 118, 52 },
		{ 62, 119, 52 },
		{ 62, 120, 52 },
		{ 62, 121, 52 },
		{ 62, 122, 52 },
		{ 62, 123, 52 },
		{ 62, 124, 52 },
		{ 62, 125, 52 },
		{ 62, 126, 52 },
		{ 62, 127, 52 },
		{ 62, 128, 52 },
		{ 62, 129, 52 },
		{ 62, 130, 52 },
		{ 62, 131, 52 },
		{ 62, 132, 52 },
		{ 62, 133, 52 },
		{ 62, 134, 52 },
		{ 62, 135, 52 },
		{ 63, 88, 52 },
		{ 63, 89, 52 },
		{ 63, 90, 52 },
		{ 63, 91, 52 },
		{ 63, 92, 52 },
		{ 63, 93, 52 },
		{ 63, 94, 52 },
		{ 63, 95, 52 },
		{ 63, 96, 52 },
		{ 63, 97, 52 },
		{ 63, 98, 52 },
		{ 63, 99, 52 },
		{ 63, 100, 52 },
		{ 63, 101, 52 },
		{ 63, 102, 52 },
		{ 63, 103, 52 },
		{ 63, 104, 52 },
		{ 63, 105, 52 },
		{ 63, 106, 52 },
		{ 63, 107, 52 },
		{ 63, 108, 52 },
		{ 63, 109, 52 },
		{ 63, 110, 52 },
		{ 63, 111, 52 },
		{ 63, 112, 52 },
		{ 63, 113, 52 },
		{ 63, 114, 52 },
		{ 63, 115, 52 },
		{ 63, 116, 52 },
		{ 63, 117, 52 },
		{ 63, 118, 52 },
		{ 63, 119, 52 },
		{ 63, 120, 52 },
		{ 63, 121, 52 },
		{ 63, 122, 52 },
		{ 63, 123, 52 },
		{ 63, 124, 52 },
		{ 63, 125, 52 },
		{ 63, 126, 52 },
		{ 63, 127, 52 },
		{ 63, 128, 52 },
		{ 63, 129, 52 },
		{ 63, 130, 52 },
		{ 63, 131, 52 },
		{ 63, 132, 52 },
		{ 63, 133, 52 },
		{ 63, 134, 52 },
		{ 63, 135, 52 },
		{ 64, 89, 52 },
		{ 64, 90, 52 },
		{ 64, 91, 52 },
		{ 64, 92, 52 },
		{ 64, 93, 52 },
		{ 64, 94, 52 },
		{ 64, 95, 52 },
		{ 64, 96, 52 },
		{ 64, 97, 52 },
		{ 64, 98, 52 },
		{ 64, 99, 52 },
		{ 64, 100, 52 },
		{ 64, 101, 52 },
		{ 64, 102, 52 },
		{ 64, 103, 52 },
		{ 64, 104, 52 },
		{ 64, 105, 52 },
		{ 64, 106, 52 },
		{ 64, 107, 52 },
		{ 64, 108, 52 },
		{ 64, 109, 52 },
		{ 64, 110, 52 },
		{ 64, 111, 52 },
		{ 64, 112, 52 },
		{ 64, 113, 52 },
		{ 64, 114, 52 },
		{ 64, 115, 52 },
		{ 64, 116, 52 },
		{ 64, 117, 52 },
		{ 64, 118, 52 },
		{ 64, 119, 52 },
		{ 64, 120, 52 },
		{ 64, 121, 52 },
		{ 64, 122, 52 },
		{ 64, 123, 52 },
		{ 64, 124, 52 },
		{ 64, 125, 52 },
		{ 64, 126, 52 },
		{ 64, 127, 52 },
		{ 64, 128, 52 },
		{ 64, 129, 52 },
		{ 64, 130, 52 },
		{ 64, 131, 52 },
		{ 64, 132, 52 },
		{ 64, 133, 52 },
		{ 64, 134, 52 },
		{ 64, 135, 52 },
		{ 64, 136, 52 },
		{ 65, 90, 52 },
		{ 65, 91, 52 },
		{ 65, 92, 52 },
		{ 65, 93, 52 },
		{ 65, 94, 52 },
		{ 65, 95, 52 },
		{ 65, 96, 52 },
		{ 65, 97, 52 },
		{ 65, 98, 52 },
		{ 65, 99, 52 },
		{ 65, 100, 52 },
		{ 65, 101, 52 },
		{ 65, 102, 52 },
		{ 65, 103, 52 },
		{ 65, 104, 52 },
		{ 65, 105, 52 },
		{ 65, 106, 52 },
		{ 65, 107, 52 },
		{ 65, 108, 52 },
		{ 65, 109, 52 },
		{ 65, 110, 52 },
		{ 65, 111, 52 },
		{ 65, 112, 52 },
		{ 65, 113, 52 },
		{ 65, 114, 52 },
		{ 65, 115, 52 },
		{ 65, 116, 52 },
		{ 65, 117, 52 },
		{ 65, 118, 52 },
		{ 65, 119, 52 },
		{ 65, 120, 52 },
		{ 65, 121, 52 },
		{ 65, 122, 52 },
		{ 65, 123, 52 },
		{ 65, 124, 52 },
		{ 65, 125, 52 },
		{ 65, 126, 52 },
		{ 65, 127, 52 },
		{ 65, 128, 52 },
		{ 65, 129, 52 },
		{ 65, 130, 52 },
		{ 65, 131, 52 },
		{ 65, 132, 52 },
		{ 65, 133, 52 },
		{ 65, 134, 52 },
		{ 65, 135, 52 },
		{ 65, 136, 52 },
		{ 66, 90, 52 },
		{ 66, 91, 52 },
		{ 66, 92, 52 },
		{ 66, 93, 52 },
		{ 66, 94, 52 },
		{ 66, 95, 52 },
		{ 66, 96, 52 },
		{ 66, 97, 52 },
		{ 66, 98, 52 },
		{ 66, 99, 52 },
		{ 66, 100, 52 },
		{ 66, 101, 52 },
		{ 66, 102, 52 },
		{ 66, 103, 52 },
		{ 66, 104, 52 },
		{ 66, 105, 52 },
		{ 66, 106, 52 },
		{ 66, 107, 52 },
		{ 66, 108, 52 },
		{ 66, 109, 52 },
		{ 66, 110, 52 },
		{ 66, 111, 52 },
		{ 66, 112, 52 },
		{ 66, 113, 52 },
		{ 66, 114, 52 },
		{ 66, 115, 52 },
		{ 66, 116, 52 },
		{ 66, 117, 52 },
		{ 66, 118, 52 },
		{ 66, 119, 52 },
		{ 66, 120, 52 },
		{ 66, 121, 52 },
		{ 66, 122, 52 },
		{ 66, 123, 52 },
		{ 66, 124, 52 },
		{ 66, 125, 52 },
		{ 66, 126, 52 },
		{ 66, 127, 52 },
		{ 66, 128, 52 },
		{ 66, 129, 52 },
		{ 66, 130, 52 },
		{ 66, 131, 52 },
		{ 66, 132, 52 },
		{ 66, 133, 52 },
		{ 66, 134, 52 },
		{ 66, 135, 52 },
		{ 66, 136, 52 },
		{ 66, 137, 52 },
		{ 67, 91, 52 },
		{ 67, 92, 52 },
		{ 67, 93, 52 },
		{ 67, 94, 52 },
		{ 67, 95, 52 },
		{ 67, 96, 52 },
		{ 67, 97, 52 },
		{ 67, 98, 52 },
		{ 67, 99, 52 },
		{ 67, 100, 52 },
		{ 67, 101, 52 },
		{ 67, 102, 52 },
		{ 67, 103, 52 },
		{ 67, 104, 52 },
		{ 67, 105, 52 },
		{ 67, 106, 52 },
		{ 67, 107, 52 },
		{ 67, 108, 52 },
		{ 67, 109, 52 },
		{ 67, 110, 52 },
		{ 67, 111, 52 },
		{ 67, 112, 52 },
		{ 67, 113, 52 },
		{ 67, 114, 52 },
		{ 67, 115, 52 },
		{ 67, 116, 52 },
		{ 67, 117, 52 },
		{ 67, 118, 52 },
		{ 67, 119, 52 },
		{ 67, 120, 52 },
		{ 67, 121, 52 },
		{ 67, 122, 52 },
		{ 67, 123, 52 },
		{ 67, 124, 52 },
		{ 67, 125, 52 },
		{ 67, 126, 52 },
		{ 67, 127, 52 },
		{ 67, 128, 52 },
		{ 67, 129, 52 },
		{ 67, 130, 52 },
		{ 67, 131, 52 },
		{ 67, 132, 52 },
		{ 67, 133, 52 },
		{ 67, 134, 52 },
		{ 67, 135, 52 },
		{ 67, 136, 52 },
		{ 67, 137, 52 },
		{ 67, 138, 52 },
		{ 68, 92, 52 },
		{ 68, 93, 52 },
		{ 68, 94, 52 },
		{ 68, 95, 52 },
		{ 68, 96, 52 },
		{ 68, 97, 52 },
		{ 68, 98, 52 },
		{ 68, 99, 52 },
		{ 68, 100, 52 },
		{ 68, 101, 52 },
		{ 68, 102, 52 },
		{ 68, 103, 52 },
		{ 68, 104, 52 },
		{ 68, 105, 52 },
		{ 68, 106, 52 },
		{ 68, 107, 52 },
		{ 68, 108, 52 },
		{ 68, 109, 52 },
		{ 68, 110, 52 },
		{ 68, 111, 52 },
		{ 68, 112, 52 },
		{ 68, 113, 52 },
		{ 68, 114, 52 },
		{ 68, 115, 52 },
		{ 68, 116, 52 },
		{ 68, 117, 52 },
		{ 68, 118, 52 },
		{ 68, 119, 52 },
		{ 68, 120, 52 },
		{ 68, 121, 52 },
		{ 68, 122, 52 },
		{ 68, 123, 52 },
		{ 68, 124, 52 },
		{ 68, 125, 52 },
		{ 68, 126, 52 },
		{ 68, 127, 52 },
		{ 68, 128, 52 },
		{ 68, 129, 52 },
		{ 68, 130, 52 },
		{ 68, 131, 52 },
		{ 68, 132, 52 },
		{ 68, 133, 52 },
		{ 68, 134, 52 },
		{ 68, 135, 52 },
		{ 68, 136, 52 },
		{ 68, 137, 52 },
		{ 68, 138, 52 },
		{ 69, 93, 52 },
		{ 69, 94, 52 },
		{ 69, 95, 52 },
		{ 69, 96, 52 },
		{ 69, 97, 52 },
		{ 69, 98, 52 },
		{ 69, 99, 52 },
		{ 69, 100, 52 },
		{ 69, 101, 52 },
		{ 69, 102, 52 },
		{ 69, 103, 52 },
		{ 69, 104, 52 },
		{ 69, 105, 52 },
		{ 69, 106, 52 },
		{ 69, 107, 52 },
		{ 69, 108, 52 },
		{ 69, 109, 52 },
		{ 69, 110, 52 },
		{ 69, 111, 52 },
		{ 69, 112, 52 },
		{ 69, 113, 52 },
		{ 69, 114, 52 },
		{ 69, 115, 52 },
		{ 69, 116, 52 },
		{ 69, 117, 52 },
		{ 69, 118, 52 },
		{ 69, 119, 52 },
		{ 69, 120, 52 },
		{ 69, 121, 52 },
		{ 69, 122, 52 },
		{ 69, 123, 52 },
		{ 69, 124, 52 },
		{ 69, 125, 52 },
		{ 69, 126, 52 },
		{ 69, 127, 52 },
		{ 69, 128, 52 },
		{ 69, 129, 52 },
		{ 69, 130, 52 },
		{ 69, 131, 52 },
		{ 69, 132, 52 },
		{ 69, 133, 52 },
		{ 69, 134, 52 },
		{ 69, 135, 52 },
		{ 69, 136, 52 },
		{ 69, 137, 52 },
		{ 69, 138, 52 },
		{ 69, 139, 52 },
		{ 70, 94, 52 },
		{ 70, 95, 52 },
		{ 70, 96, 52 },
		{ 70, 97, 52 },
		{ 70, 98, 52 },
		{ 70, 99, 52 },
		{ 70, 100, 52 },
		{ 70, 101, 52 },
		{ 70, 102, 52 },
		{ 70, 103, 52 },
		{ 70, 104, 52 },
		{ 70, 105, 52 },
		{ 70, 106, 52 },
		{ 70, 107, 52 },
		{ 70, 108, 52 },
		{ 70, 109, 52 },
		{ 70, 110, 52 },
		{ 70, 111, 52 },
		{ 70, 112, 52 },
		{ 70, 113, 52 },
		{ 70, 114, 52 },
		{ 70, 115, 52 },
		{ 70, 116, 52 },
		{ 70, 117, 52 },
		{ 70, 118, 52 },
		{ 70, 119, 52 },
		{ 70, 120, 52 },
		{ 70, 121, 52 },
		{ 70, 122, 52 },
		{ 70, 123, 52 },
		{ 70, 124, 52 },
		{ 70, 125, 52 },
		{ 70, 126, 52 },
		{ 70, 127, 52 },
		{ 70, 128, 52 },
		{ 70, 129, 52 },
		{ 70, 130, 52 },
		{ 70, 131, 52 },
		{ 70, 132, 52 },
		{ 70, 133, 52 },
		{ 70, 134, 52 },
		{ 70, 135, 52 },
		{ 70, 136, 52 },
		{ 70, 137, 52 },
		{ 70, 138, 52 },
		{ 70, 139, 52 },
		{ 70, 140, 52 },
		{ 71, 95, 52 },
		{ 71, 96, 52 },
		{ 71, 97, 52 },
		{ 71, 98, 52 },
		{ 71, 99, 52 },
		{ 71, 100, 52 },
		{ 71, 101, 52 },
		{ 71, 102, 52 },
		{ 71, 103, 52 },
		{ 71, 104, 52 },
		{ 71, 105, 52 },
		{ 71, 106, 52 },
		{ 71, 107, 52 },
		{ 71, 108, 52 },
		{ 71, 109, 52 },
		{ 71, 110, 52 },
		{ 71, 111, 52 },
		{ 71, 112, 52 },
		{ 71, 113, 52 },
		{ 71, 114, 52 },
		{ 71, 115, 52 },
		{ 71, 116, 52 },
		{ 71, 117, 52 },
		{ 71, 118, 52 },
		{ 71, 119, 52 },
		{ 71, 120, 52 },
		{ 71, 121, 52 },
		{ 71, 122, 52 },
		{ 71, 123, 52 },
		{ 71, 124, 52 },
		{ 71, 125, 52 },
		{ 71, 126, 52 },
		{ 71, 127, 52 },
		{ 71, 128, 52 },
		{ 71, 129, 52 },
		{ 71, 130, 52 },
		{ 71, 131, 52 },
		{ 71, 132, 52 },
		{ 71, 133, 52 },
		{ 71, 134, 52 },
		{ 71, 135, 52 },
		{ 71, 136, 52 },
		{ 71, 137, 52 },
		{ 71, 138, 52 },
		{ 71, 139, 52 },
		{ 71, 140, 52 },
		{ 72, 96, 52 },
		{ 72, 97, 52 },
		{ 72, 98, 52 },
		{ 72, 99, 52 },
		{ 72, 100, 52 },
		{ 72, 101, 52 },
		{ 72, 102, 52 },
		{ 72, 103, 52 },
		{ 72, 104, 52 },
		{ 72, 105, 52 },
		{ 72, 106, 52 },
		{ 72, 107, 52 },
		{ 72, 108, 52 },
		{ 72, 109, 52 },
		{ 72, 110, 52 },
		{ 72, 111, 52 },
		{ 72, 112, 52 },
		{ 72, 113, 52 },
		{ 72, 114, 52 },
		{ 72, 115, 52 },
		{ 72, 116, 52 },
		{ 72, 117, 52 },
		{ 72, 118, 52 },
		{ 72, 119, 52 },
		{ 72, 120, 52 },
		{ 72, 121, 52 },
		{ 72, 122, 52 },
		{ 72, 123, 52 },
		{ 72, 124, 52 },
		{ 72, 125, 52 },
		{ 72, 126, 52 },
		{ 72, 127, 52 },
		{ 72, 128, 52 },
		{ 72, 129, 52 },
		{ 72, 130, 52 },
		{ 72, 131, 52 },
		{ 72, 132, 52 },
		{ 72, 133, 52 },
		{ 72, 134, 52 },
		{ 72, 135, 52 },
		{ 72, 136, 52 },
		{ 72, 137, 52 },
		{ 72, 138, 52 },
		{ 72, 139, 52 },
		{ 72, 140, 52 },
		{ 72, 141, 52 },
		{ 73, 97, 52 },
		{ 73, 98, 52 },
		{ 73, 99, 52 },
		{ 73, 100, 52 },
		{ 73, 101, 52 },
		{ 73, 102, 52 },
		{ 73, 103, 52 },
		{ 73, 104, 52 },
		{ 73, 105, 52 },
		{ 73, 106, 52 },
		{ 73, 107, 52 },
		{ 73, 108, 52 },
		{ 73, 109, 52 },
		{ 73, 110, 52 },
		{ 73, 111, 52 },
		{ 73, 112, 52 },
		{ 73, 113, 52 },
		{ 73, 114, 52 },
		{ 73, 115, 52 },
		{ 73, 116, 52 },
		{ 73, 117, 52 },
		{ 73, 118, 52 },
		{ 73, 119, 52 },
		{ 73, 120, 52 },
		{ 73, 121, 52 },
		{ 73, 122, 52 },
		{ 73, 123, 52 },
		{ 73, 124, 52 },
		{ 73, 125, 52 },
		{ 73, 126, 52 },
		{ 73, 127, 52 },
		{ 73, 128, 52 },
		{ 73, 129, 52 },
		{ 73, 130, 52 },
		{ 73, 131, 52 },
		{ 73, 132, 52 },
		{ 73, 133, 52 },
		{ 73, 134, 52 },
		{ 73, 135, 52 },
		{ 73, 136, 52 },
		{ 73, 137, 52 },
		{ 73, 138, 52 },
		{ 73, 139, 52 },
		{ 73, 140, 52 },
		{ 73, 141, 52 },
		{ 74, 97, 52 },
		{ 74, 98, 52 },
		{ 74, 99, 52 },
		{ 74, 100, 52 },
		{ 74, 101, 52 },
		{ 74, 102, 52 },
		{ 74, 103, 52 },
		{ 74, 104, 52 },
		{ 74, 105, 52 },
		{ 74, 106, 52 },
		{ 74, 107, 52 },
		{ 74, 108, 52 },
		{ 74, 109, 52 },
		{ 74, 110, 52 },
		{ 74, 111, 52 },
		{ 74, 112, 52 },
		{ 74, 113, 52 },
		{ 74, 114, 52 },
		{ 74, 115, 52 },
		{ 74, 116, 52 },
		{ 74, 117, 52 },
		{ 74, 118, 52 },
		{ 74, 119, 52 },
		{ 74, 120, 52 },
		{ 74, 121, 52 },
		{ 74, 122, 52 },
		{ 74, 123, 52 },
		{ 74, 124, 52 },
		{ 74, 125, 52 },
		{ 74, 126, 52 },
		{ 74, 127, 52 },
		{ 74, 128, 52 },
		{ 74, 129, 52 },
		{ 74, 130, 52 },
		{ 74, 131, 52 },
		{ 74, 132, 52 },
		{ 74, 133, 52 },
		{ 74, 134, 52 },
		{ 74, 135, 52 },
		{ 74, 136, 52 },
		{ 74, 137, 52 },
		{ 74, 138, 52 },
		{ 74, 139, 52 },
		{ 74, 140, 52 },
		{ 74, 141, 52 },
		{ 74, 142, 52 },
		{ 75, 98, 52 },
		{ 75, 99, 52 },
		{ 75, 100, 52 },
		{ 75, 101, 52 },
		{ 75, 102, 52 },
		{ 75, 103, 52 },
		{ 75, 104, 52 },
		{ 75, 105, 52 },
		{ 75, 106, 52 },
		{ 75, 107, 52 },
		{ 75, 108, 52 },
		{ 75, 109, 52 },
		{ 75, 110, 52 },
		{ 75, 111, 52 },
		{ 75, 112, 52 },
		{ 75, 113, 52 },
		{ 75, 114, 52 },
		{ 75, 115, 52 },
		{ 75, 116, 52 },
		{ 75, 117, 52 },
		{ 75, 118, 52 },
		{ 75, 119, 52 },
		{ 75, 120, 52 },
		{ 75, 121, 52 },
		{ 75, 122, 52 },
		{ 75, 123, 52 },
		{ 75, 124, 52 },
		{ 75, 125, 52 },
		{ 75, 126, 52 },
		{ 75, 127, 52 },
		{ 75, 128, 52 },
		{ 75, 129, 52 },
		{ 75, 130, 52 },
		{ 75, 131, 52 },
		{ 75, 132, 52 },
		{ 75, 133, 52 },
		{ 75, 134, 52 },
		{ 75, 135, 52 },
		{ 75, 136, 52 },
		{ 75, 137, 52 },
		{ 75, 138, 52 },
		{ 75, 139, 52 },
		{ 75, 140, 52 },
		{ 75, 141, 52 },
		{ 75, 142, 52 },
		{ 75, 143, 52 },
		{ 76, 99, 52 },
		{ 76, 100, 52 },
		{ 76, 101, 52 },
		{ 76, 102, 52 },
		{ 76, 103, 52 },
		{ 76, 104, 52 },
		{ 76, 105, 52 },
		{ 76, 106, 52 },
		{ 76, 107, 52 },
		{ 76, 108, 52 },
		{ 76, 109, 52 },
		{ 76, 110, 52 },
		{ 76, 111, 52 },
		{ 76, 112, 52 },
		{ 76, 113, 52 },
		{ 76, 114, 52 },
		{ 76, 115, 52 },
		{ 76, 116, 52 },
		{ 76, 117, 52 },
		{ 76, 118, 52 },
		{ 76, 119, 52 },
		{ 76, 120, 52 },
		{ 76, 121, 52 },
		{ 76, 122, 52 },
		{ 76, 123, 52 },
		{ 76, 124, 52 },
		{ 76, 125, 52 },
		{ 76, 126, 52 },
		{ 76, 127, 52 },
		{ 76, 128, 52 },
		{ 76, 129, 52 },
		{ 76, 130, 52 },
		{ 76, 131, 52 },
		{ 76, 132, 52 },
		{ 76, 133, 52 },
		{ 76, 134, 52 },
		{ 76, 135, 52 },
		{ 76, 136, 52 },
		{ 76, 137, 52 },
		{ 76, 138, 52 },
		{ 76, 139, 52 },
		{ 76, 140, 52 },
		{ 76, 141, 52 },
		{ 76, 142, 52 },
		{ 76, 143, 52 },
		{ 77, 100, 52 },
		{ 77, 101, 52 },
		{ 77, 102, 52 },
		{ 77, 103, 52 },
		{ 77, 104, 52 },
		{ 77, 105, 52 },
		{ 77, 106, 52 },
		{ 77, 107, 52 },
		{ 77, 108, 52 },
		{ 77, 109, 52 },
		{ 77, 110, 52 },
		{ 77, 111, 52 },
		{ 77, 112, 52 },
		{ 77, 113, 52 },
		{ 77, 114, 52 },
		{ 77, 115, 52 },
		{ 77, 116, 52 },
		{ 77, 117, 52 },
		{ 77, 118, 52 },
		{ 77, 119, 52 },
		{ 77, 120, 52 },
		{ 77, 121, 52 },
		{ 77, 122, 52 },
		{ 77, 123, 52 },
		{ 77, 124, 52 },
		{ 77, 125, 52 },
		{ 77, 126, 52 },
		{ 77, 127, 52 },
		{ 77, 128, 52 },
		{ 77, 129, 52 },
		{ 77, 130, 52 },
		{ 77, 131, 52 },
		{ 77, 132, 52 },
		{ 77, 133, 52 },
		{ 77, 134, 52 },
		{ 77, 135, 52 },
		{ 77, 136, 52 },
		{ 77, 137, 52 },
		{ 77, 138, 52 },
		{ 77, 139, 52 },
		{ 77, 140, 52 },
		{ 77, 141, 52 },
		{ 77, 142, 52 },
		{ 77, 143, 52 },
		{ 77, 144, 52 },
		{ 78, 101, 52 },
		{ 78, 102, 52 },
		{ 78, 103, 52 },
		{ 78, 104, 52 },
		{ 78, 105, 52 },
		{ 78, 106, 52 },
		{ 78, 107, 52 },
		{ 78, 108, 52 },
		{ 78, 109, 52 },
		{ 78, 110, 52 },
		{ 78, 111, 52 },
		{ 78, 112, 52 },
		{ 78, 113, 52 },
		{ 78, 114, 52 },
		{ 78, 115, 52 },
		{ 78, 116, 52 },
		{ 78, 117, 52 },
		{ 78, 118, 52 },
		{ 78, 119, 52 },
		{ 78, 120, 52 },
		{ 78, 121, 52 },
		{ 78, 122, 52 },
		{ 78, 123, 52 },
		{ 78, 124, 52 },
		{ 78, 125, 52 },
		{ 78, 126, 52 },
		{ 78, 127, 52 },
		{ 78, 128, 52 },
		{ 78, 129, 52 },
		{ 78, 130, 52 },
		{ 78, 131, 52 },
		{ 78, 132, 52 },
		{ 78, 133, 52 },
		{ 78, 134, 52 },
		{ 78, 135, 52 },
		{ 78, 136, 52 },
		{ 78, 137, 52 },
		{ 78, 138, 52 },
		{ 78, 139, 52 },
		{ 78, 140, 52 },
		{ 78, 141, 52 },
		{ 78, 142, 52 },
		{ 78, 143, 52 },
		{ 78, 144, 52 },
		{ 78, 145, 52 },
		{ 79, 102, 52 },
		{ 79, 103, 52 },
		{ 79, 104, 52 },
		{ 79, 105, 52 },
		{ 79, 106, 52 },
		{ 79, 107, 52 },
		{ 79, 108, 52 },
		{ 79, 109, 52 },
		{ 79, 110, 52 },
		{ 79, 111, 52 },
		{ 79, 112, 52 },
		{ 79, 113, 52 },
		{ 79, 114, 52 },
		{ 79, 115, 52 },
		{ 79, 116, 52 },
		{ 79, 117, 52 },
		{ 79, 118, 52 },
		{ 79, 119, 52 },
		{ 79, 120, 52 },
		{ 79, 121, 52 },
		{ 79, 122, 52 },
		{ 79, 123, 52 },
		{ 79, 124, 52 },
		{ 79, 125, 52 },
		{ 79, 126, 52 },
		{ 79, 127, 52 },
		{ 79, 128, 52 },
		{ 79, 129, 52 },
		{ 79, 130, 52 },
		{ 79, 131, 52 },
		{ 79, 132, 52 },
		{ 79, 133, 52 },
		{ 79, 134, 52 },
		{ 79, 135, 52 },
		{ 79, 136, 52 },
		{ 79, 137, 52 },
		{ 79, 138, 52 },
		{ 79, 139, 52 },
		{ 79, 140, 52 },
		{ 79, 141, 52 },
		{ 79, 142, 52 },
		{ 79, 143, 52 },
		{ 79, 144, 52 },
		{ 79, 145, 52 },
		{ 80, 103, 52 },
		{ 80, 104, 52 },
		{ 80, 105, 52 },
		{ 80, 106, 52 },
		{ 80, 107, 52 },
		{ 80, 108, 52 },
		{ 80, 109, 52 },
		{ 80, 110, 52 },
		{ 80, 111, 52 },
		{ 80, 112, 52 },
		{ 80, 113, 52 },
		{ 80, 114, 52 },
		{ 80, 115, 52 },
		{ 80, 116, 52 },
		{ 80, 117, 52 },
		{ 80, 118, 52 },
		{ 80, 119, 52 },
		{ 80, 120, 52 },
		{ 80, 121, 52 },
		{ 80, 122, 52 },
		{ 80, 123, 52 },
		{ 80, 124, 52 },
		{ 80, 125, 52 },
		{ 80, 126, 52 },
		{ 80, 127, 52 },
		{ 80, 128, 52 },
		{ 80, 129, 52 },
		{ 80, 130, 52 },
		{ 80, 131, 52 },
		{ 80, 132, 52 },
		{ 80, 133, 52 },
		{ 80, 134, 52 },
		{ 80, 135, 52 },
		{ 80, 136, 52 },
		{ 80, 137, 52 },
		{ 80, 138, 52 },
		{ 80, 139, 52 },
		{ 80, 140, 52 },
		{ 80, 141, 52 },
		{ 80, 142, 52 },
		{ 80, 143, 52 },
		{ 80, 144, 52 },
		{ 80, 145, 52 },
		{ 80, 146, 52 },
		{ 81, 104, 52 },
		{ 81, 105, 52 },
		{ 81, 106, 52 },
		{ 81, 107, 52 },
		{ 81, 108, 52 },
		{ 81, 109, 52 },
		{ 81, 110, 52 },
		{ 81, 111, 52 },
		{ 81, 112, 52 },
		{ 81, 113, 52 },
		{ 81, 114, 52 },
		{ 81, 115, 52 },
		{ 81, 116, 52 },
		{ 81, 117, 52 },
		{ 81, 118, 52 },
		{ 81, 119, 52 },
		{ 81, 120, 52 },
		{ 81, 121, 52 },
		{ 81, 122, 52 },
		{ 81, 123, 52 },
		{ 81, 124, 52 },
		{ 81, 125, 52 },
		{ 81, 126, 52 },
		{ 81, 127, 52 },
		{ 81, 128, 52 },
		{ 81, 129, 52 },
		{ 81, 130, 52 },
		{ 81, 131, 52 },
		{ 81, 132, 52 },
		{ 81, 133, 52 },
		{ 81, 134, 52 },
		{ 81, 135, 52 },
		{ 81, 136, 52 },
		{ 81, 137, 52 },
		{ 81, 138, 52 },
		{ 81, 139, 52 },
		{ 81, 140, 52 },
		{ 81, 141, 52 },
		{ 81, 142, 52 },
		{ 81, 143, 52 },
		{ 81, 144, 52 },
		{ 81, 145, 52 },
		{ 81, 146, 52 },
		{ 82, 104, 52 },
		{ 82, 105, 52 },
		{ 82, 106, 52 },
		{ 82, 107, 52 },
		{ 82, 108, 52 },
		{ 82, 109, 52 },
		{ 82, 110, 52 },
		{ 82, 111, 52 },
		{ 82, 112, 52 },
		{ 82, 113, 52 },
		{ 82, 114, 52 },
		{ 82, 115, 52 },
		{ 82, 116, 52 },
		{ 82, 117, 52 },
		{ 82, 118, 52 },
		{ 82, 119, 52 },
		{ 82, 120, 52 },
		{ 82, 121, 52 },
		{ 82, 122, 52 },
		{ 82, 123, 52 },
		{ 82, 124, 52 },
		{ 82, 125, 52 },
		{ 82, 126, 52 },
		{ 82, 127, 52 },
		{ 82, 128, 52 },
		{ 82, 129, 52 },
		{ 82, 130, 52 },
		{ 82, 131, 52 },
		{ 82, 132, 52 },
		{ 82, 133, 52 },
		{ 82, 134, 52 },
		{ 82, 135, 52 },
		{ 82, 136, 52 },
		{ 82, 137, 52 },
		{ 82, 138, 52 },
		{ 82, 139, 52 },
		{ 82, 140, 52 },
		{ 82, 141, 52 },
		{ 82, 142, 52 },
		{ 82, 143, 52 },
		{ 82, 144, 52 },
		{ 82, 145, 52 },
		{ 82, 146, 52 },
		{ 82, 147, 52 },
		{ 83, 105, 52 },
		{ 83, 106, 52 },
		{ 83, 107, 52 },
		{ 83, 108, 52 },
		{ 83, 109, 52 },
		{ 83, 110, 52 },
		{ 83, 111, 52 },
		{ 83, 112, 52 },
		{ 83, 113, 52 },
		{ 83, 114, 52 },
		{ 83, 115, 52 },
		{ 83, 116, 52 },
		{ 83, 117, 52 },
		{ 83, 118, 52 },
		{ 83, 119, 52 },
		{ 83, 120, 52 },
		{ 83, 121, 52 },
		{ 83, 122, 52 },
		{ 83, 123, 52 },
		{ 83, 124, 52 },
		{ 83, 125, 52 },
		{ 83, 126, 52 },
		{ 83, 127, 52 },
		{ 83, 128, 52 },
		{ 83, 129, 52 },
		{ 83, 130, 52 },
		{ 83, 131, 52 },
		{ 83, 132, 52 },
		{ 83, 133, 52 },
		{ 83, 134, 52 },
		{ 83, 135, 52 },
		{ 83, 136, 52 },
		{ 83, 137, 52 },
		{ 83, 138, 52 },
		{ 83, 139, 52 },
		{ 83, 140, 52 },
		{ 83, 141, 52 },
		{ 83, 142, 52 },
		{ 83, 143, 52 },
		{ 83, 144, 52 },
		{ 83, 145, 52 },
		{ 83, 146, 52 },
		{ 83, 147, 52 },
		{ 83, 148, 52 },
		{ 84, 106, 52 },
		{ 84, 107, 52 },
		{ 84, 108, 52 },
		{ 84, 109, 52 },
		{ 84, 110, 52 },
		{ 84, 111, 52 },
		{ 84, 112, 52 },
		{ 84, 113, 52 },
		{ 84, 114, 52 },
		{ 84, 115, 52 },
		{ 84, 116, 52 },
		{ 84, 117, 52 },
		{ 84, 118, 52 },
		{ 84, 119, 52 },
		{ 84, 120, 52 },
		{ 84, 121, 52 },
		{ 84, 122, 52 },
		{ 84, 123, 52 },
		{ 84, 124, 52 },
		{ 84, 125, 52 },
		{ 84, 126, 52 },
		{ 84, 127, 52 },
		{ 84, 128, 52 },
		{ 84, 129, 52 },
		{ 84, 130, 52 },
		{ 84, 131, 52 },
		{ 84, 132, 52 },
		{ 84, 133, 52 },
		{ 84, 134, 52 },
		{ 84, 135, 52 },
		{ 84, 136, 52 },
		{ 84, 137, 52 },
		{ 84, 138, 52 },
		{ 84, 139, 52 },
		{ 84, 140, 52 },
		{ 84, 141, 52 },
		{ 84, 142, 52 },
		{ 84, 143, 52 },
		{ 84, 144, 52 },
		{ 84, 145, 52 },
		{ 84, 146, 52 },
		{ 84, 147, 52 },
		{ 84, 148, 52 },
		{ 85, 107, 52 },
		{ 85, 108, 52 },
		{ 85, 109, 52 },
		{ 85, 110, 52 },
		{ 85, 111, 52 },
		{ 85, 112, 52 },
		{ 85, 113, 52 },
		{ 85, 114, 52 },
		{ 85, 115, 52 },
		{ 85, 116, 52 },
		{ 85, 117, 52 },
		{ 85, 118, 52 },
		{ 85, 119, 52 },
		{ 85, 120, 52 },
		{ 85, 121, 52 },
		{ 85, 122, 52 },
		{ 85, 123, 52 },
		{ 85, 124, 52 },
		{ 85, 125, 52 },
		{ 85, 126, 52 },
		{ 85, 127, 52 },
		{ 85, 128, 52 },
		{ 85, 129, 52 },
		{ 85, 130, 52 },
		{ 85, 131, 52 },
		{ 85, 132, 52 },
		{ 85, 133, 52 },
		{ 85, 134, 52 },
		{ 85, 135, 52 },
		{ 85, 136, 52 },
		{ 85, 137, 52 },
		{ 85, 138, 52 },
		{ 85, 139, 52 },
		{ 85, 140, 52 },
		{ 85, 141, 52 },
		{ 85, 142, 52 },
		{ 85, 143, 52 },
		{ 85, 144, 52 },
		{ 85, 145, 52 },
		{ 85, 146, 52 },
		{ 85, 147, 52 },
		{ 85, 148, 52 },
		{ 85, 149, 52 },
		{ 86, 108, 52 },
		{ 86, 109, 52 },
		{ 86, 110, 52 },
		{ 86, 111, 52 },
		{ 86, 112, 52 },
		{ 86, 113, 52 },
		{ 86, 114, 52 },
		{ 86, 115, 52 },
		{ 86, 116, 52 },
		{ 86, 117, 52 },
		{ 86, 118, 52 },
		{ 86, 119, 52 },
		{ 86, 120, 52 },
		{ 86, 121, 52 },
		{ 86, 122, 52 },
		{ 86, 123, 52 },
		{ 86, 124, 52 },
		{ 86, 125, 52 },
		{ 86, 126, 52 },
		{ 86, 127, 52 },
		{ 86, 128, 52 },
		{ 86, 129, 52 },
		{ 86, 130, 52 },
		{ 86, 131, 52 },
		{ 86, 132, 52 },
		{ 86, 133, 52 },
		{ 86, 134, 52 },
		{ 86, 135, 52 },
		{ 86, 136, 52 },
		{ 86, 137, 52 },
		{ 86, 138, 52 },
		{ 86, 139, 52 },
		{ 86, 140, 52 },
		{ 86, 141, 52 },
		{ 86, 142, 52 },
		{ 86, 143, 52 },
		{ 86, 144, 52 },
		{ 86, 145, 52 },
		{ 86, 146, 52 },
		{ 86, 147, 52 },
		{ 86, 148, 52 },
		{ 86, 149, 52 },
		{ 86, 150, 52 },
		{ 87, 109, 52 },
		{ 87, 110, 52 },
		{ 87, 111, 52 },
		{ 87, 112, 52 },
		{ 87, 113, 52 },
		{ 87, 114, 52 },
		{ 87, 115, 52 },
		{ 87, 116, 52 },
		{ 87, 117, 52 },
		{ 87, 118, 52 },
		{ 87, 119, 52 },
		{ 87, 120, 52 },
		{ 87, 121, 52 },
		{ 87, 122, 52 },
		{ 87, 123, 52 },
		{ 87, 124, 52 },
		{ 87, 125, 52 },
		{ 87, 126, 52 },
		{ 87, 127, 52 },
		{ 87, 128, 52 },
		{ 87, 129, 52 },
		{ 87, 130, 52 },
		{ 87, 131, 52 },
		{ 87, 132, 52 },
		{ 87, 133, 52 },
		{ 87, 134, 52 },
		{ 87, 135, 52 },
		{ 87, 136, 52 },
		{ 87, 137, 52 },
		{ 87, 138, 52 },
		{ 87, 139, 52 },
		{ 87, 140, 52 },
		{ 87, 141, 52 },
		{ 87, 142, 52 },
		{ 87, 143, 52 },
		{ 87, 144, 52 },
		{ 87, 145, 52 },
		{ 87, 146, 52 },
		{ 87, 147, 52 },
		{ 87, 148, 52 },
		{ 87, 149, 52 },
		{ 87, 150, 52 },
		{ 88, 110, 52 },
		{ 88, 111, 52 },
		{ 88, 112, 52 },
		{ 88, 113, 52 },
		{ 88, 114, 52 },
		{ 88, 115, 52 },
		{ 88, 116, 52 },
		{ 88, 117, 52 },
		{ 88, 118, 52 },
		{ 88, 119, 52 },
		{ 88, 120, 52 },
		{ 88, 121, 52 },
		{ 88, 122, 52 },
		{ 88, 123, 52 },
		{ 88, 124, 52 },
		{ 88, 125, 52 },
		{ 88, 126, 52 },
		{ 88, 127, 52 },
		{ 88, 128, 52 },
		{ 88, 129, 52 },
		{ 88, 130, 52 },
		{ 88, 131, 52 },
		{ 88, 132, 52 },
		{ 88, 133, 52 },
		{ 88, 134, 52 },
		{ 88, 135, 52 },
		{ 88, 136, 52 },
		{ 88, 137, 52 },
		{ 88, 138, 52 },
		{ 88, 139, 52 },
		{ 88, 140, 52 },
		{ 88, 141, 52 },
		{ 88, 142, 52 },
		{ 88, 143, 52 },
		{ 88, 144, 52 },
		{ 88, 145, 52 },
		{ 88, 146, 52 },
		{ 88, 147, 52 },
		{ 88, 148, 52 },
		{ 88, 149, 52 },
		{ 88, 150, 52 },
		{ 88, 151, 52 },
		{ 89, 111, 52 },
		{ 89, 112, 52 },
		{ 89, 113, 52 },
		{ 89, 114, 52 },
		{ 89, 115, 52 },
		{ 89, 116, 52 },
		{ 89, 117, 52 },
		{ 89, 118, 52 },
		{ 89, 119, 52 },
		{ 89, 120, 52 },
		{ 89, 121, 52 },
		{ 89, 122, 52 },
		{ 89, 123, 52 },
		{ 89, 124, 52 },
		{ 89, 125, 52 },
		{ 89, 126, 52 },
		{ 89, 127, 52 },
		{ 89, 128, 52 },
		{ 89, 129, 52 },
		{ 89, 130, 52 },
		{ 89, 131, 52 },
		{ 89, 132, 52 },
		{ 89, 133, 52 },
		{ 89, 134, 52 },
		{ 89, 135, 52 },
		{ 89, 136, 52 },
		{ 89, 137, 52 },
		{ 89, 138, 52 },
		{ 89, 139, 52 },
		{ 89, 140, 52 },
		{ 89, 141, 52 },
		{ 89, 142, 52 },
		{ 89, 143, 52 },
		{ 89, 144, 52 },
		{ 89, 145, 52 },
		{ 89, 146, 52 },
		{ 89, 147, 52 },
		{ 89, 148, 52 },
		{ 89, 149, 52 },
		{ 89, 150, 52 },
		{ 89, 151, 52 },
		{ 90, 111, 52 },
		{ 90, 112, 52 },
		{ 90, 113, 52 },
		{ 90, 114, 52 },
		{ 90, 115, 52 },
		{ 90, 116, 52 },
		{ 90, 117, 52 },
		{ 90, 118, 52 },
		{ 90, 119, 52 },
		{ 90, 120, 52 },
		{ 90, 121, 52 },
		{ 90, 122, 52 },
		{ 90, 123, 52 },
		{ 90, 124, 52 },
		{ 90, 125, 52 },
		{ 90, 126, 52 },
		{ 90, 127, 52 },
		{ 90, 128, 52 },
		{ 90, 129, 52 },
		{ 90, 130, 52 },
		{ 90, 131, 52 },
		{ 90, 132, 52 },
		{ 90, 133, 52 },
		{ 90, 134, 52 },
		{ 90, 135, 52 },
		{ 90, 136, 52 },
		{ 90, 137, 52 },
		{ 90, 138, 52 },
		{ 90, 139, 52 },
		{ 90, 140, 52 },
		{ 90, 141, 52 },
		{ 90, 142, 52 },
		{ 90, 143, 52 },
		{ 90, 144, 52 },
		{ 90, 145, 52 },
		{ 90, 146, 52 },
		{ 90, 147, 52 },
		{ 90, 148, 52 },
		{ 90, 149, 52 },
		{ 90, 150, 52 },
		{ 90, 151, 52 },
		{ 90, 152, 52 },
		{ 91, 112, 52 },
		{ 91, 113, 52 },
		{ 91, 114, 52 },
		{ 91, 115, 52 },
		{ 91, 116, 52 },
		{ 91, 117, 52 },
		{ 91, 118, 52 },
		{ 91, 119, 52 },
		{ 91, 120, 52 },
		{ 91, 121, 52 },
		{ 91, 122, 52 },
		{ 91, 123, 52 },
		{ 91, 124, 52 },
		{ 91, 125, 52 },
		{ 91, 126, 52 },
		{ 91, 127, 52 },
		{ 91, 128, 52 },
		{ 91, 129, 52 },
		{ 91, 130, 52 },
		{ 91, 131, 52 },
		{ 91, 132, 52 },
		{ 91, 133, 52 },
		{ 91, 134, 52 },
		{ 91, 135, 52 },
		{ 91, 136, 52 },
		{ 91, 137, 52 },
		{ 91, 138, 52 },
		{ 91, 139, 52 },
		{ 91, 140, 52 },
		{ 91, 141, 52 },
		{ 91, 142, 52 },
		{ 91, 143, 52 },
		{ 91, 144, 52 },
		{ 91, 145, 52 },
		{ 91, 146, 52 },
		{ 91, 147, 52 },
		{ 91, 148, 52 },
		{ 91, 149, 52 },
		{ 91, 150, 52 },
		{ 91, 151, 52 },
		{ 91, 152, 52 },
		{ 91, 153, 52 },
		{ 92, 113, 52 },
		{ 92, 114, 52 },
		{ 92, 115, 52 },
		{ 92, 116, 52 },
		{ 92, 117, 52 },
		{ 92, 118, 52 },
		{ 92, 119, 52 },
		{ 92, 120, 52 },
		{ 92, 121, 52 },
		{ 92, 122, 52 },
		{ 92, 123, 52 },
		{ 92, 124, 52 },
		{ 92, 125, 52 },
		{ 92, 126, 52 },
		{ 92, 127, 52 },
		{ 92, 128, 52 },
		{ 92, 129, 52 },
		{ 92, 130, 52 },
		{ 92, 131, 52 },
		{ 92, 132, 52 },
		{ 92, 133, 52 },
		{ 92, 134, 52 },
		{ 92, 135, 52 },
		{ 92, 136, 52 },
		{ 92, 137, 52 },
		{ 92, 138, 52 },
		{ 92, 139, 52 },
		{ 92, 140, 52 },
		{ 92, 141, 52 },
		{ 92, 142, 52 },
		{ 92, 143, 52 },
		{ 92, 144, 52 },
		{ 92, 145, 52 },
		{ 92, 146, 52 },
		{ 92, 147, 52 },
		{ 92, 148, 52 },
		{ 92, 149, 52 },
		{ 92, 150, 52 },
		{ 92, 151, 52 },
		{ 92, 152, 52 },
		{ 92, 153, 52 },
		{ 93, 114, 52 },
		{ 93, 115, 52 },
		{ 93, 116, 52 },
		{ 93, 117, 52 },
		{ 93, 118, 52 },
		{ 93, 119, 52 },
		{ 93, 120, 52 },
		{ 93, 121, 52 },
		{ 93, 122, 52 },
		{ 93, 123, 52 },
		{ 93, 124, 52 },
		{ 93, 125, 52 },
		{ 93, 126, 52 },
		{ 93, 127, 52 },
		{ 93, 128, 52 },
		{ 93, 129, 52 },
		{ 93, 130, 52 },
		{ 93, 131, 52 },
		{ 93, 132, 52 },
		{ 93, 133, 52 },
		{ 93, 134, 52 },
		{ 93, 135, 52 },
		{ 93, 136, 52 },
		{ 93, 137, 52 },
		{ 93, 138, 52 },
		{ 93, 139, 52 },
		{ 93, 140, 52 },
		{ 93, 141, 52 },
		{ 93, 142, 52 },
		{ 93, 143, 52 },
		{ 93, 144, 52 },
		{ 93, 145, 52 },
		{ 93, 146, 52 },
		{ 93, 147, 52 },
		{ 93, 148, 52 },
		{ 93, 149, 52 },
		{ 93, 150, 52 },
		{ 93, 151, 52 },
		{ 93, 152, 52 },
		{ 93, 153, 52 },
		{ 93, 154, 52 },
		{ 94, 115, 52 },
		{ 94, 116, 52 },
		{ 94, 117, 52 },
		{ 94, 118, 52 },
		{ 94, 119, 52 },
		{ 94, 120, 52 },
		{ 94, 121, 52 },
		{ 94, 122, 52 },
		{ 94, 123, 52 },
		{ 94, 124, 52 },
		{ 94, 125, 52 },
		{ 94, 126, 52 },
		{ 94, 127, 52 },
		{ 94, 128, 52 },
		{ 94, 129, 52 },
		{ 94, 130, 52 },
		{ 94, 131, 52 },
		{ 94, 132, 52 },
		{ 94, 133, 52 },
		{ 94, 134, 52 },
		{ 94, 135, 52 },
		{ 94, 136, 52 },
		{ 94, 137, 52 },
		{ 94, 138, 52 },
		{ 94, 139, 52 },
		{ 94, 140, 52 },
		{ 94, 141, 52 },
		{ 94, 142, 52 },
		{ 94, 143, 52 },
		{ 94, 144, 52 },
		{ 94, 145, 52 },
		{ 94, 146, 52 },
		{ 94, 147, 52 },
		{ 94, 148, 52 },
		{ 94, 149, 52 },
		{ 94, 150, 52 },
		{ 94, 151, 52 },
		{ 94, 152, 52 },
		{ 94, 153, 52 },
		{ 94, 154, 52 },
		{ 94, 155, 52 },
		{ 95, 116, 52 },
		{ 95, 117, 52 },
		{ 95, 118, 52 },
		{ 95, 119, 52 },
		{ 95, 120, 52 },
		{ 95, 121, 52 },
		{ 95, 122, 52 },
		{ 95, 123, 52 },
		{ 95, 124, 52 },
		{ 95, 125, 52 },
		{ 95, 126, 52 },
		{ 95, 127, 52 },
		{ 95, 128, 52 },
		{ 95, 129, 52 },
		{ 95, 130, 52 },
		{ 95, 131, 52 },
		{ 95, 132, 52 },
		{ 95, 133, 52 },
		{ 95, 134, 52 },
		{ 95, 135, 52 },
		{ 95, 136, 52 },
		{ 95, 137, 52 },
		{ 95, 138, 52 },
		{ 95, 139, 52 },
		{ 95, 140, 52 },
		{ 95, 141, 52 },
		{ 95, 142, 52 },
		{ 95, 143, 52 },
		{ 95, 144, 52 },
		{ 95, 145, 52 },
		{ 95, 146, 52 },
		{ 95, 147, 52 },
		{ 95, 148, 52 },
		{ 95, 149, 52 },
		{ 95, 150, 52 },
		{ 95, 151, 52 },
		{ 95, 152, 52 },
		{ 95, 153, 52 },
		{ 95, 154, 52 },
		{ 95, 155, 52 },
		{ 96, 117, 52 },
		{ 96, 118, 52 },
		{ 96, 119, 52 },
		{ 96, 120, 52 },
		{ 96, 121, 52 },
		{ 96, 122, 52 },
		{ 96, 123, 52 },
		{ 96, 124, 52 },
		{ 96, 125, 52 },
		{ 96, 126, 52 },
		{ 96, 127, 52 },
		{ 96, 128, 52 },
		{ 96, 129, 52 },
		{ 96, 130, 52 },
		{ 96, 131, 52 },
		{ 96, 132, 52 },
		{ 96, 133, 52 },
		{ 96, 134, 52 },
		{ 96, 135, 52 },
		{ 96, 136, 52 },
		{ 96, 137, 52 },
		{ 96, 138, 52 },
		{ 96, 139, 52 },
		{ 96, 140, 52 },
		{ 96, 141, 52 },
		{ 96, 142, 52 },
		{ 96, 143, 52 },
		{ 96, 144, 52 },
		{ 96, 145, 52 },
		{ 96, 146, 52 },
		{ 96, 147, 52 },
		{ 96, 148, 52 },
		{ 96, 149, 52 },
		{ 96, 150, 52 },
		{ 96, 151, 52 },
		{ 96, 152, 52 },
		{ 96, 153, 52 },
		{ 96, 154, 52 },
		{ 96, 155, 52 },
		{ 96, 156, 52 },
		{ 97, 118, 52 },
		{ 97, 119, 52 },
		{ 97, 120, 52 },
		{ 97, 121, 52 },
		{ 97, 122, 52 },
		{ 97, 123, 52 },
		{ 97, 124, 52 },
		{ 97, 125, 52 },
		{ 97, 126, 52 },
		{ 97, 127, 52 },
		{ 97, 128, 52 },
		{ 97, 129, 52 },
		{ 97, 130, 52 },
		{ 97, 131, 52 },
		{ 97, 132, 52 },
		{ 97, 133, 52 },
		{ 97, 134, 52 },
		{ 97, 135, 52 },
		{ 97, 136, 52 },
		{ 97, 137, 52 },
		{ 97, 138, 52 },
		{ 97, 139, 52 },
		{ 97, 140, 52 },
		{ 97, 141, 52 },
		{ 97, 142, 52 },
		{ 97, 143, 52 },
		{ 97, 144, 52 },
		{ 97, 145, 52 },
		{ 97, 146, 52 },
		{ 97, 147, 52 },
		{ 97, 148, 52 },
		{ 97, 149, 52 },
		{ 97, 150, 52 },
		{ 97, 151, 52 },
		{ 97, 152, 52 },
		{ 97, 153, 52 },
		{ 97, 154, 52 },
		{ 97, 155, 52 },
		{ 97, 156, 52 },
		{ 98, 118, 52 },
		{ 98, 119, 52 },
		{ 98, 120, 52 },
		{ 98, 121, 52 },
		{ 98, 122, 52 },
		{ 98, 123, 52 },
		{ 98, 124, 52 },
		{ 98, 125, 52 },
		{ 98, 126, 52 },
		{ 98, 127, 52 },
		{ 98, 128, 52 },
		{ 98, 129, 52 },
		{ 98, 130, 52 },
		{ 98, 131, 52 },
		{ 98, 132, 52 },
		{ 98, 133, 52 },
		{ 98, 134, 52 },
		{ 98, 135, 52 },
		{ 98, 136, 52 },
		{ 98, 137, 52 },
		{ 98, 138, 52 },
		{ 98, 139, 52 },
		{ 98, 140, 52 },
		{ 98, 141, 52 },
		{ 98, 142, 52 },
		{ 98, 143, 52 },
		{ 98, 144, 52 },
		{ 98, 145, 52 },
		{ 98, 146, 52 },
		{ 98, 147, 52 },
		{ 98, 148, 52 },
		{ 98, 149, 52 },
		{ 98, 150, 52 },
		{ 98, 151, 52 },
		{ 98, 152, 52 },
		{ 98, 153, 52 },
		{ 98, 154, 52 },
		{ 98, 155, 52 },
		{ 98, 156, 52 },
		{ 98, 157, 52 },
		{ 99, 119, 52 },
		{ 99, 120, 52 },
		{ 99, 121, 52 },
		{ 99, 122, 52 },
		{ 99, 123, 52 },
		{ 99, 124, 52 },
		{ 99, 125, 52 },
		{ 99, 126, 52 },
		{ 99, 127, 52 },
		{ 99, 128, 52 },
		{ 99, 129, 52 },
		{ 99, 130, 52 },
		{ 99, 131, 52 },
		{ 99, 132, 52 },
		{ 99, 133, 52 },
		{ 99, 134, 52 },
		{ 99, 135, 52 },
		{ 99, 136, 52 },
		{ 99, 137, 52 },
		{ 99, 138, 52 },
		{ 99, 139, 52 },
		{ 99, 140, 52 },
		{ 99, 141, 52 },
		{ 99, 142, 52 },
		{ 99, 143, 52 },
		{ 99, 144, 52 },
		{ 99, 145, 52 },
		{ 99, 146, 52 },
		{ 99, 147, 52 },
		{ 99, 148, 52 },
		{ 99, 149, 52 },
		{ 99, 150, 52 },
		{ 99, 151, 52 },
		{ 99, 152, 52 },
		{ 99, 153, 52 },
		{ 99, 154, 52 },
		{ 99, 155, 52 },
		{ 99, 156, 52 },
		{ 99, 157, 52 },
		{ 99, 158, 52 },
		{ 100, 120, 52 },
		{ 100, 121, 52 },
		{ 100, 122, 52 },
		{ 100, 123, 52 },
		{ 100, 124, 52 },
		{ 100, 125, 52 },
		{ 100, 126, 52 },
		{ 100, 127, 52 },
		{ 100, 128, 52 },
		{ 100, 129, 52 },
		{ 100, 130, 52 },
		{ 100, 131, 52 },
		{ 100, 132, 52 },
		{ 100, 133, 52 },
		{ 100, 134, 52 },
		{ 100, 135, 52 },
		{ 100, 136, 52 },
		{ 100, 137, 52 },
		{ 100, 138, 52 },
		{ 100, 139, 52 },
		{ 100, 140, 52 },
		{ 100, 141, 52 },
		{ 100, 142, 52 },
		{ 100, 143, 52 },
		{ 100, 144, 52 },
		{ 100, 145, 52 },
		{ 100, 146, 52 },
		{ 100, 147, 52 },
		{ 100, 148, 52 },
		{ 100, 149, 52 },
		{ 100, 150, 52 },
		{ 100, 151, 52 },
		{ 100, 152, 52 },
		{ 100, 153, 52 },
		{ 100, 154, 52 },
		{ 100, 155, 52 },
		{ 100, 156, 52 },
		{ 100, 157, 52 },
		{ 100, 158, 52 },
		{ 101, 121, 52 },
		{ 101, 122, 52 },
		{ 101, 123, 52 },
		{ 101, 124, 52 },
		{ 101, 125, 52 },
		{ 101, 126, 52 },
		{ 101, 127, 52 },
		{ 101, 128, 52 },
		{ 101, 129, 52 },
		{ 101, 130, 52 },
		{ 101, 131, 52 },
		{ 101, 132, 52 },
		{ 101, 133, 52 },
		{ 101, 134, 52 },
		{ 101, 135, 52 },
		{ 101, 136, 52 },
		{ 101, 137, 52 },
		{ 101, 138, 52 },
		{ 101, 139, 52 },
		{ 101, 140, 52 },
		{ 101, 141, 52 },
		{ 101, 142, 52 },
		{ 101, 143, 52 },
		{ 101, 144, 52 },
		{ 101, 145, 52 },
		{ 101, 146, 52 },
		{ 101, 147, 52 },
		{ 101, 148, 52 },
		{ 101, 149, 52 },
		{ 101, 150, 52 },
		{ 101, 151, 52 },
		{ 101, 152, 52 },
		{ 101, 153, 52 },
		{ 101, 154, 52 },
		{ 101, 155, 52 },
		{ 101, 156, 52 },
		{ 101, 157, 52 },
		{ 101, 158, 52 },
		{ 101, 159, 52 },
		{ 102, 122, 52 },
		{ 102, 123, 52 },
		{ 102, 124, 52 },
		{ 102, 125, 52 },
		{ 102, 126, 52 },
		{ 102, 127, 52 },
		{ 102, 128, 52 },
		{ 102, 129, 52 },
		{ 102, 130, 52 },
		{ 102, 131, 52 },
		{ 102, 132, 52 },
		{ 102, 133, 52 },
		{ 102, 134, 52 },
		{ 102, 135, 52 },
		{ 102, 136, 52 },
		{ 102, 137, 52 },
		{ 102, 138, 52 },
		{ 102, 139, 52 },
		{ 102, 140, 52 },
		{ 102, 141, 52 },
		{ 102, 142, 52 },
		{ 102, 143, 52 },
		{ 102, 144, 52 },
		{ 102, 145, 52 },
		{ 102, 146, 52 },
		{ 102, 147, 52 },
		{ 102, 148, 52 },
		{ 102, 149, 52 },
		{ 102, 150, 52 },
		{ 102, 151, 52 },
		{ 102, 152, 52 },
		{ 102, 153, 52 },
		{ 102, 154, 52 },
		{ 102, 155, 52 },
		{ 102, 156, 52 },
		{ 102, 157, 52 },
		{ 102, 158, 52 },
		{ 102, 159, 52 },
		{ 102, 160, 52 },
		{ 103, 123, 52 },
		{ 103, 124, 52 },
		{ 103, 125, 52 },
		{ 103, 126, 52 },
		{ 103, 127, 52 },
		{ 103, 128, 52 },
		{ 103, 129, 52 },
		{ 103, 130, 52 },
		{ 103, 131, 52 },
		{ 103, 132, 52 },
		{ 103, 133, 52 },
		{ 103, 134, 52 },
		{ 103, 135, 52 },
		{ 103, 136, 52 },
		{ 103, 137, 52 },
		{ 103, 138, 52 },
		{ 103, 139, 52 },
		{ 103, 140, 52 },
		{ 103, 141, 52 },
		{ 103, 142, 52 },
		{ 103, 143, 52 },
		{ 103, 144, 52 },
		{ 103, 145, 52 },
		{ 103, 146, 52 },
		{ 103, 147, 52 },
		{ 103, 148, 52 },
		{ 103, 149, 52 },
		{ 103, 150, 52 },
		{ 103, 151, 52 },
		{ 103, 152, 52 },
		{ 103, 153, 52 },
		{ 103, 154, 52 },
		{ 103, 155, 52 },
		{ 103, 156, 52 },
		{ 103, 157, 52 },
		{ 103, 158, 52 },
		{ 103, 159, 52 },
		{ 103, 160, 52 },
		{ 104, 124, 52 },
		{ 104, 125, 52 },
		{ 104, 126, 52 },
		{ 104, 127, 52 },
		{ 104, 128, 52 },
		{ 104, 129, 52 },
		{ 104, 130, 52 },
		{ 104, 131, 52 },
		{ 104, 132, 52 },
		{ 104, 133, 52 },
		{ 104, 134, 52 },
		{ 104, 135, 52 },
		{ 104, 136, 52 },
		{ 104, 137, 52 },
		{ 104, 138, 52 },
		{ 104, 139, 52 },
		{ 104, 140, 52 },
		{ 104, 141, 52 },
		{ 104, 142, 52 },
		{ 104, 143, 52 },
		{ 104, 144, 52 },
		{ 104, 145, 52 },
		{ 104, 146, 52 },
		{ 104, 147, 52 },
		{ 104, 148, 52 },
		{ 104, 149, 52 },
		{ 104, 150, 52 },
		{ 104, 151, 52 },
		{ 104, 152, 52 },
		{ 104, 153, 52 },
		{ 104, 154, 52 },
		{ 104, 155, 52 },
		{ 104, 156, 52 },
		{ 104, 157, 52 },
		{ 104, 158, 52 },
		{ 104, 159, 52 },
		{ 104, 160, 52 },
		{ 104, 161, 52 },
		{ 105, 125, 52 },
		{ 105, 126, 52 },
		{ 105, 127, 52 },
		{ 105, 128, 52 },
		{ 105, 129, 52 },
		{ 105, 130, 52 },
		{ 105, 131, 52 },
		{ 105, 132, 52 },
		{ 105, 133, 52 },
		{ 105, 134, 52 },
		{ 105, 135, 52 },
		{ 105, 136, 52 },
		{ 105, 137, 52 },
		{ 105, 138, 52 },
		{ 105, 139, 52 },
		{ 105, 140, 52 },
		{ 105, 141, 52 },
		{ 105, 142, 52 },
		{ 105, 143, 52 },
		{ 105, 144, 52 },
		{ 105, 145, 52 },
		{ 105, 146, 52 },
		{ 105, 147, 52 },
		{ 105, 148, 52 },
		{ 105, 149, 52 },
		{ 105, 150, 52 },
		{ 105, 151, 52 },
		{ 105, 152, 52 },
		{ 105, 153, 52 },
		{ 105, 154, 52 },
		{ 105, 155, 52 },
		{ 105, 156, 52 },
		{ 105, 157, 52 },
		{ 105, 158, 52 },
		{ 105, 159, 52 },
		{ 105, 160, 52 },
		{ 105, 161, 52 },
		{ 106, 125, 52 },
		{ 106, 126, 52 },
		{ 106, 127, 52 },
		{ 106, 128, 52 },
		{ 106, 129, 52 },
		{ 106, 130, 52 },
		{ 106, 131, 52 },
		{ 106, 132, 52 },
		{ 106, 133, 52 },
		{ 106, 134, 52 },
		{ 106, 135, 52 },
		{ 106, 136, 52 },
		{ 106, 137, 52 },
		{ 106, 138, 52 },
		{ 106, 139, 52 },
		{ 106, 140, 52 },
		{ 106, 141, 52 },
		{ 106, 142, 52 },
		{ 106, 143, 52 },
		{ 106, 144, 52 },
		{ 106, 145, 52 },
		{ 106, 146, 52 },
		{ 106, 147, 52 },
		{ 106, 148, 52 },
		{ 106, 149, 52 },
		{ 106, 150, 52 },
		{ 106, 151, 52 },
		{ 106, 152, 52 },
		{ 106, 153, 52 },
		{ 106, 154, 52 },
		{ 106, 155, 52 },
		{ 106, 156, 52 },
		{ 106, 157, 52 },
		{ 106, 158, 52 },
		{ 106, 159, 52 },
		{ 106, 160, 52 },
		{ 106, 161, 52 },
		{ 106, 162, 52 },
		{ 107, 126, 52 },
		{ 107, 127, 52 },
		{ 107, 128, 52 },
		{ 107, 129, 52 },
		{ 107, 130, 52 },
		{ 107, 131, 52 },
		{ 107, 132, 52 },
		{ 107, 133, 52 },
		{ 107, 134, 52 },
		{ 107, 135, 52 },
		{ 107, 136, 52 },
		{ 107, 137, 52 },
		{ 107, 138, 52 },
		{ 107, 139, 52 },
		{ 107, 140, 52 },
		{ 107, 141, 52 },
		{ 107, 142, 52 },
		{ 107, 143, 52 },
		{ 107, 144, 52 },
		{ 107, 145, 52 },
		{ 107, 146, 52 },
		{ 107, 147, 52 },
		{ 107, 148, 52 },
		{ 107, 149, 52 },
		{ 107, 150, 52 },
		{ 107, 151, 52 },
		{ 107, 152, 52 },
		{ 107, 153, 52 },
		{ 107, 154, 52 },
		{ 107, 155, 52 },
		{ 107, 156, 52 },
		{ 107, 157, 52 },
		{ 107, 158, 52 },
		{ 107, 159, 52 },
		{ 107, 160, 52 },
		{ 107, 161, 52 },
		{ 107, 162, 52 },
		{ 107, 163, 52 },
		{ 108, 127, 52 },
		{ 108, 128, 52 },
		{ 108, 129, 52 },
		{ 108, 130, 52 },
		{ 108, 131, 52 },
		{ 108, 132, 52 },
		{ 108, 133, 52 },
		{ 108, 134, 52 },
		{ 108, 135, 52 },
		{ 108, 136, 52 },
		{ 108, 137, 52 },
		{ 108, 138, 52 },
		{ 108, 139, 52 },
		{ 108, 140, 52 },
		{ 108, 141, 52 },
		{ 108, 142, 52 },
		{ 108, 143, 52 },
		{ 108, 144, 52 },
		{ 108, 145, 52 },
		{ 108, 146, 52 },
		{ 108, 147, 52 },
		{ 108, 148, 52 },
		{ 108, 149, 52 },
		{ 108, 150, 52 },
		{ 108, 151, 52 },
		{ 108, 152, 52 },
		{ 108, 153, 52 },
		{ 108, 154, 52 },
		{ 108, 155, 52 },
		{ 108, 156, 52 },
		{ 108, 157, 52 },
		{ 108, 158, 52 },
		{ 108, 159, 52 },
		{ 108, 160, 52 },
		{ 108, 161, 52 },
		{ 108, 162, 52 },
		{ 108, 163, 52 },
		{ 109, 128, 52 },
		{ 109, 129, 52 },
		{ 109, 130, 52 },
		{ 109, 131, 52 },
		{ 109, 132, 52 },
		{ 109, 133, 52 },
		{ 109, 134, 52 },
		{ 109, 135, 52 },
		{ 109, 136, 52 },
		{ 109, 137, 52 },
		{ 109, 138, 52 },
		{ 109, 139, 52 },
		{ 109, 140, 52 },
		{ 109, 141, 52 },
		{ 109, 142, 52 },
		{ 109, 143, 52 },
		{ 109, 144, 52 },
		{ 109, 145, 52 },
		{ 109, 146, 52 },
		{ 109, 147, 52 },
		{ 109, 148, 52 },
		{ 109, 149, 52 },
		{ 109, 150, 52 },
		{ 109, 151, 52 },
		{ 109, 152, 52 },
		{ 109, 153, 52 },
		{ 109, 154, 52 },
		{ 109, 155, 52 },
		{ 109, 156, 52 },
		{ 109, 157, 52 },
		{ 109, 158, 52 },
		{ 109, 159, 52 },
		{ 109, 160, 52 },
		{ 109, 161, 52 },
		{ 109, 162, 52 },
		{ 109, 163, 52 },
		{ 109, 164, 52 },
		{ 110, 129, 52 },
		{ 110, 130, 52 },
		{ 110, 131, 52 },
		{ 110, 132, 52 },
		{ 110, 133, 52 },
		{ 110, 134, 52 },
		{ 110, 135, 52 },
		{ 110, 136, 52 },
		{ 110, 137, 52 },
		{ 110, 138, 52 },
		{ 110, 139, 52 },
		{ 110, 140, 52 },
		{ 110, 141, 52 },
		{ 110, 142, 52 },
		{ 110, 143, 52 },
		{ 110, 144, 52 },
		{ 110, 145, 52 },
		{ 110, 146, 52 },
		{ 110, 147, 52 },
		{ 110, 148, 52 },
		{ 110, 149, 52 },
		{ 110, 150, 52 },
		{ 110, 151, 52 },
		{ 110, 152, 52 },
		{ 110, 153, 52 },
		{ 110, 154, 52 },
		{ 110, 155, 52 },
		{ 110, 156, 52 },
		{ 110, 157, 52 },
		{ 110, 158, 52 },
		{ 110, 159, 52 },
		{ 110, 160, 52 },
		{ 110, 161, 52 },
		{ 110, 162, 52 },
		{ 110, 163, 52 },
		{ 110, 164, 52 },
		{ 110, 165, 52 },
		{ 111, 130, 52 },
		{ 111, 131, 52 },
		{ 111, 132, 52 },
		{ 111, 133, 52 },
		{ 111, 134, 52 },
		{ 111, 135, 52 },
		{ 111, 136, 52 },
		{ 111, 137, 52 },
		{ 111, 138, 52 },
		{ 111, 139, 52 },
		{ 111, 140, 52 },
		{ 111, 141, 52 },
		{ 111, 142, 52 },
		{ 111, 143, 52 },
		{ 111, 144, 52 },
		{ 111, 145, 52 },
		{ 111, 146, 52 },
		{ 111, 147, 52 },
		{ 111, 148, 52 },
		{ 111, 149, 52 },
		{ 111, 150, 52 },
		{ 111, 151, 52 },
		{ 111, 152, 52 },
		{ 111, 153, 52 },
		{ 111, 154, 52 },
		{ 111, 155, 52 },
		{ 111, 156, 52 },
		{ 111, 157, 52 },
		{ 111, 158, 52 },
		{ 111, 159, 52 },
		{ 111, 160, 52 },
		{ 111, 161, 52 },
		{ 111, 162, 52 },
		{ 111, 163, 52 },
		{ 111, 164, 52 },
		{ 111, 165, 52 },
		{ 112, 131, 52 },
		{ 112, 132, 52 },
		{ 112, 133, 52 },
		{ 112, 134, 52 },
		{ 112, 135, 52 },
		{ 112, 136, 52 },
		{ 112, 137, 52 },
		{ 112, 138, 52 },
		{ 112, 139, 52 },
		{ 112, 140, 52 },
		{ 112, 141, 52 },
		{ 112, 142, 52 },
		{ 112, 143, 52 },
		{ 112, 144, 52 },
		{ 112, 145, 52 },
		{ 112, 146, 52 },
		{ 112, 147, 52 },
		{ 112, 148, 52 },
		{ 112, 149, 52 },
		{ 112, 150, 52 },
		{ 112, 151, 52 },
		{ 112, 152, 52 },
		{ 112, 153, 52 },
		{ 112, 154, 52 },
		{ 112, 155, 52 },
		{ 112, 156, 52 },
		{ 112, 157, 52 },
		{ 112, 158, 52 },
		{ 112, 159, 52 },
		{ 112, 160, 52 },
		{ 112, 161, 52 },
		{ 112, 162, 52 },
		{ 112, 163, 52 },
		{ 112, 164, 52 },
		{ 112, 165, 52 },
		{ 112, 166, 52 },
		{ 113, 132, 52 },
		{ 113, 133, 52 },
		{ 113, 134, 52 },
		{ 113, 135, 52 },
		{ 113, 136, 52 },
		{ 113, 137, 52 },
		{ 113, 138, 52 },
		{ 113, 139, 52 },
		{ 113, 140, 52 },
		{ 113, 141, 52 },
		{ 113, 142, 52 },
		{ 113, 143, 52 },
		{ 113, 144, 52 },
		{ 113, 145, 52 },
		{ 113, 146, 52 },
		{ 113, 147, 52 },
		{ 113, 148, 52 },
		{ 113, 149, 52 },
		{ 113, 150, 52 },
		{ 113, 151, 52 },
		{ 113, 152, 52 },
		{ 113, 153, 52 },
		{ 113, 154, 52 },
		{ 113, 155, 52 },
		{ 113, 156, 52 },
		{ 113, 157, 52 },
		{ 113, 158, 52 },
		{ 113, 159, 52 },
		{ 113, 160, 52 },
		{ 113, 161, 52 },
		{ 113, 162, 52 },
		{ 113, 163, 52 },
		{ 113, 164, 52 },
		{ 113, 165, 52 },
		{ 113, 166, 52 },
		{ 114, 132, 52 },
		{ 114, 133, 52 },
		{ 114, 134, 52 },
		{ 114, 135, 52 },
		{ 114, 136, 52 },
		{ 114, 137, 52 },
		{ 114, 138, 52 },
		{ 114, 139, 52 },
		{ 114, 140, 52 },
		{ 114, 141, 52 },
		{ 114, 142, 52 },
		{ 114, 143, 52 },
		{ 114, 144, 52 },
		{ 114, 145, 52 },
		{ 114, 146, 52 },
		{ 114, 147, 52 },
		{ 114, 148, 52 },
		{ 114, 149, 52 },
		{ 114, 150, 52 },
		{ 114, 151, 52 },
		{ 114, 152, 52 },
		{ 114, 153, 52 },
		{ 114, 154, 52 },
		{ 114, 155, 52 },
		{ 114, 156, 52 },
		{ 114, 157, 52 },
		{ 114, 158, 52 },
		{ 114, 159, 52 },
		{ 114, 160, 52 },
		{ 114, 161, 52 },
		{ 114, 162, 52 },
		{ 114, 163, 52 },
		{ 114, 164, 52 },
		{ 114, 165, 52 },
		{ 114, 166, 52 },
		{ 114, 167, 52 },
		{ 115, 133, 52 },
		{ 115, 134, 52 },
		{ 115, 135, 52 },
		{ 115, 136, 52 },
		{ 115, 137, 52 },
		{ 115, 138, 52 },
		{ 115, 139, 52 },
		{ 115, 140, 52 },
		{ 115, 141, 52 },
		{ 115, 142, 52 },
		{ 115, 143, 52 },
		{ 115, 144, 52 },
		{ 115, 145, 52 },
		{ 115, 146, 52 },
		{ 115, 147, 52 },
		{ 115, 148, 52 },
		{ 115, 149, 52 },
		{ 115, 150, 52 },
		{ 115, 151, 52 },
		{ 115, 152, 52 },
		{ 115, 153, 52 },
		{ 115, 154, 52 },
		{ 115, 155, 52 },
		{ 115, 156, 52 },
		{ 115, 157, 52 },
		{ 115, 158, 52 },
		{ 115, 159, 52 },
		{ 115, 160, 52 },
		{ 115, 161, 52 },
		{ 115, 162, 52 },
		{ 115, 163, 52 },
		{ 115, 164, 52 },
		{ 115, 165, 52 },
		{ 115, 166, 52 },
		{ 115, 167, 52 },
		{ 115, 168, 52 },
		{ 116, 134, 52 },
		{ 116, 135, 52 },
		{ 116, 136, 52 },
		{ 116, 137, 52 },
		{ 116, 138, 52 },
		{ 116, 139, 52 },
		{ 116, 140, 52 },
		{ 116, 141, 52 },
		{ 116, 142, 52 },
		{ 116, 143, 52 },
		{ 116, 144, 52 },
		{ 116, 145, 52 },
		{ 116, 146, 52 },
		{ 116, 147, 52 },
		{ 116, 148, 52 },
		{ 116, 149, 52 },
		{ 116, 150, 52 },
		{ 116, 151, 52 },
		{ 116, 152, 52 },
		{ 116, 153, 52 },
		{ 116, 154, 52 },
		{ 116, 155, 52 },
		{ 116, 156, 52 },
		{ 116, 157, 52 },
		{ 116, 158, 52 },
		{ 116, 159, 52 },
		{ 116, 160, 52 },
		{ 116, 161, 52 },
		{ 116, 162, 52 },
		{ 116, 163, 52 },
		{ 116, 164, 52 },
		{ 116, 165, 52 },
		{ 116, 166, 52 },
		{ 116, 167, 52 },
		{ 116, 168, 52 },
		{ 117, 135, 52 },
		{ 117, 136, 52 },
		{ 117, 137, 52 },
		{ 117, 138, 52 },
		{ 117, 139, 52 },
		{ 117, 140, 52 },
		{ 117, 141, 52 },
		{ 117, 142, 52 },
		{ 117, 143, 52 },
		{ 117, 144, 52 },
		{ 117, 145, 52 },
		{ 117, 146, 52 },
		{ 117, 147, 52 },
		{ 117, 148, 52 },
		{ 117, 149, 52 },
		{ 117, 150, 52 },
		{ 117, 151, 52 },
		{ 117, 152, 52 },
		{ 117, 153, 52 },
		{ 117, 154, 52 },
		{ 117, 155, 52 },
		{ 117, 156, 52 },
		{ 117, 157, 52 },
		{ 117, 158, 52 },
		{ 117, 159, 52 },
		{ 117, 160, 52 },
		{ 117, 161, 52 },
		{ 117, 162, 52 },
		{ 117, 163, 52 },
		{ 117, 164, 52 },
		{ 117, 165, 52 },
		{ 117, 166, 52 },
		{ 117, 167, 52 },
		{ 117, 168, 52 },
		{ 117, 169, 52 },
		{ 118, 136, 52 },
		{ 118, 137, 52 },
		{ 118, 138, 52 },
		{ 118, 139, 52 },
		{ 118, 140, 52 },
		{ 118, 141, 52 },
		{ 118, 142, 52 },
		{ 118, 143, 52 },
		{ 118, 144, 52 },
		{ 118, 145, 52 },
		{ 118, 146, 52 },
		{ 118, 147, 52 },
		{ 118, 148, 52 },
		{ 118, 149, 52 },
		{ 118, 150, 52 },
		{ 118, 151, 52 },
		{ 118, 152, 52 },
		{ 118, 153, 52 },
		{ 118, 154, 52 },
		{ 118, 155, 52 },
		{ 118, 156, 52 },
		{ 118, 157, 52 },
		{ 118, 158, 52 },
		{ 118, 159, 52 },
		{ 118, 160, 52 },
		{ 118, 161, 52 },
		{ 118, 162, 52 },
		{ 118, 163, 52 },
		{ 118, 164, 52 },
		{ 118, 165, 52 },
		{ 118, 166, 52 },
		{ 118, 167, 52 },
		{ 118, 168, 52 },
		{ 118, 169, 52 },
		{ 118, 170, 52 },
		{ 119, 137, 52 },
		{ 119, 138, 52 },
		{ 119, 139, 52 },
		{ 119, 140, 52 },
		{ 119, 141, 52 },
		{ 119, 142, 52 },
		{ 119, 143, 52 },
		{ 119, 144, 52 },
		{ 119, 145, 52 },
		{ 119, 146, 52 },
		{ 119, 147, 52 },
		{ 119, 148, 52 },
		{ 119, 149, 52 },
		{ 119, 150, 52 },
		{ 119, 151, 52 },
		{ 119, 152, 52 },
		{ 119, 153, 52 },
		{ 119, 154, 52 },
		{ 119, 155, 52 },
		{ 119, 156, 52 },
		{ 119, 157, 52 },
		{ 119, 158, 52 },
		{ 119, 159, 52 },
		{ 119, 160, 52 },
		{ 119, 161, 52 },
		{ 119, 162, 52 },
		{ 119, 163, 52 },
		{ 119, 164, 52 },
		{ 119, 165, 52 },
		{ 119, 166, 52 },
		{ 119, 167, 52 },
		{ 119, 168, 52 },
		{ 119, 169, 52 },
		{ 119, 170, 52 },
		{ 120, 138, 52 },
		{ 120, 139, 52 },
		{ 120, 140, 52 },
		{ 120, 141, 52 },
		{ 120, 142, 52 },
		{ 120, 143, 52 },
		{ 120, 144, 52 },
		{ 120, 145, 52 },
		{ 120, 146, 52 },
		{ 120, 147, 52 },
		{ 120, 148, 52 },
		{ 120, 149, 52 },
		{ 120, 150, 52 },
		{ 120, 151, 52 },
		{ 120, 152, 52 },
		{ 120, 153, 52 },
		{ 120, 154, 52 },
		{ 120, 155, 52 },
		{ 120, 156, 52 },
		{ 120, 157, 52 },
		{ 120, 158, 52 },
		{ 120, 159, 52 },
		{ 120, 160, 52 },
		{ 120, 161, 52 },
		{ 120, 162, 52 },
		{ 120, 163, 52 },
		{ 120, 164, 52 },
		{ 120, 165, 52 },
		{ 120, 166, 52 },
		{ 120, 167, 52 },
		{ 120, 168, 52 },
		{ 120, 169, 52 },
		{ 120, 170, 52 },
		{ 120, 171, 52 },
		{ 121, 139, 52 },
		{ 121, 140, 52 },
		{ 121, 141, 52 },
		{ 121, 142, 52 },
		{ 121, 143, 52 },
		{ 121, 144, 52 },
		{ 121, 145, 52 },
		{ 121, 146, 52 },
		{ 121, 147, 52 },
		{ 121, 148, 52 },
		{ 121, 149, 52 },
		{ 121, 150, 52 },
		{ 121, 151, 52 },
		{ 121, 152, 52 },
		{ 121, 153, 52 },
		{ 121, 154, 52 },
		{ 121, 155, 52 },
		{ 121, 156, 52 },
		{ 121, 157, 52 },
		{ 121, 158, 52 },
		{ 121, 159, 52 },
		{ 121, 160, 52 },
		{ 121, 161, 52 },
		{ 121, 162, 52 },
		{ 121, 163, 52 },
		{ 121, 164, 52 },
		{ 121, 165, 52 },
		{ 121, 166, 52 },
		{ 121, 167, 52 },
		{ 121, 168, 52 },
		{ 121, 169, 52 },
		{ 121, 170, 52 },
		{ 121, 171, 52 },
		{ 122, 139, 52 },
		{ 122, 140, 52 },
		{ 122, 141, 52 },
		{ 122, 142, 52 },
		{ 122, 143, 52 },
		{ 122, 144, 52 },
		{ 122, 145, 52 },
		{ 122, 146, 52 },
		{ 122, 147, 52 },
		{ 122, 148, 52 },
		{ 122, 149, 52 },
		{ 122, 150, 52 },
		{ 122, 151, 52 },
		{ 122, 152, 52 },
		{ 122, 153, 52 },
		{ 122, 154, 52 },
		{ 122, 155, 52 },
		{ 122, 156, 52 },
		{ 122, 157, 52 },
		{ 122, 158, 52 },
		{ 122, 159, 52 },
		{ 122, 160, 52 },
		{ 122, 161, 52 },
		{ 122, 162, 52 },
		{ 122, 163, 52 },
		{ 122, 164, 52 },
		{ 122, 165, 52 },
		{ 122, 166, 52 },
		{ 122, 167, 52 },
		{ 122, 168, 52 },
		{ 122, 169, 52 },
		{ 122, 170, 52 },
		{ 122, 171, 52 },
		{ 122, 172, 52 },
		{ 123, 140, 52 },
		{ 123, 141, 52 },
		{ 123, 142, 52 },
		{ 123, 143, 52 },
		{ 123, 144, 52 },
		{ 123, 145, 52 },
		{ 123, 146, 52 },
		{ 123, 147, 52 },
		{ 123, 148, 52 },
		{ 123, 149, 52 },
		{ 123, 150, 52 },
		{ 123, 151, 52 },
		{ 123, 152, 52 },
		{ 123, 153, 52 },
		{ 123, 154, 52 },
		{ 123, 155, 52 },
		{ 123, 156, 52 },
		{ 123, 157, 52 },
		{ 123, 158, 52 },
		{ 123, 159, 52 },
		{ 123, 160, 52 },
		{ 123, 161, 52 },
		{ 123, 162, 52 },
		{ 123, 163, 52 },
		{ 123, 164, 52 },
		{ 123, 165, 52 },
		{ 123, 166, 52 },
		{ 123, 167, 52 },
		{ 123, 168, 52 },
		{ 123, 169, 52 },
		{ 123, 170, 52 },
		{ 123, 171, 52 },
		{ 123, 172, 52 },
		{ 123, 173, 52 },
		{ 124, 141, 52 },
		{ 124, 142, 52 },
		{ 124, 143, 52 },
		{ 124, 144, 52 },
		{ 124, 145, 52 },
		{ 124, 146, 52 },
		{ 124, 147, 52 },
		{ 124, 148, 52 },
		{ 124, 149, 52 },
		{ 124, 150, 52 },
		{ 124, 151, 52 },
		{ 124, 152, 52 },
		{ 124, 153, 52 },
		{ 124, 154, 52 },
		{ 124, 155, 52 },
		{ 124, 156, 52 },
		{ 124, 157, 52 },
		{ 124, 158, 52 },
		{ 124, 159, 52 },
		{ 124, 160, 52 },
		{ 124, 161, 52 },
		{ 124, 162, 52 },
		{ 124, 163, 52 },
		{ 124, 164, 52 },
		{ 124, 165, 52 },
		{ 124, 166, 52 },
		{ 124, 167, 52 },
		{ 124, 168, 52 },
		{ 124, 169, 52 },
		{ 124, 170, 52 },
		{ 124, 171, 52 },
		{ 124, 172, 52 },
		{ 124, 173, 52 },
		{ 125, 142, 52 },
		{ 125, 143, 52 },
		{ 125, 144, 52 },
		{ 125, 145, 52 },
		{ 125, 146, 52 },
		{ 125, 147, 52 },
		{ 125, 148, 52 },
		{ 125, 149, 52 },
		{ 125, 150, 52 },
		{ 125, 151, 52 },
		{ 125, 152, 52 },
		{ 125, 153, 52 },
		{ 125, 154, 52 },
		{ 125, 155, 52 },
		{ 125, 156, 52 },
		{ 125, 157, 52 },
		{ 125, 158, 52 },
		{ 125, 159, 52 },
		{ 125, 160, 52 },
		{ 125, 161, 52 },
		{ 125, 162, 52 },
		{ 125, 163, 52 },
		{ 125, 164, 52 },
		{ 125, 165, 52 },
		{ 125, 166, 52 },
		{ 125, 167, 52 },
		{ 125, 168, 52 },
		{ 125, 169, 52 },
		{ 125, 170, 52 },
		{ 125, 171, 52 },
		{ 125, 172, 52 },
		{ 125, 173, 52 },
		{ 125, 174, 52 },
		{ 126, 143, 52 },
		{ 126, 144, 52 },
		{ 126, 145, 52 },
		{ 126, 146, 52 },
		{ 126, 147, 52 },
		{ 126, 148, 52 },
		{ 126, 149, 52 },
		{ 126, 150, 52 },
		{ 126, 151, 52 },
		{ 126, 152, 52 },
		{ 126, 153, 52 },
		{ 126, 154, 52 },
		{ 126, 155, 52 },
		{ 126, 156, 52 },
		{ 126, 157, 52 },
		{ 126, 158, 52 },
		{ 126, 159, 52 },
		{ 126, 160, 52 },
		{ 126, 161, 52 },
		{ 126, 162, 52 },
		{ 126, 163, 52 },
		{ 126, 164, 52 },
		{ 126, 165, 52 },
		{ 126, 166, 52 },
		{ 126, 167, 52 },
		{ 126, 168, 52 },
		{ 126, 169, 52 },
		{ 126, 170, 52 },
		{ 126, 171, 52 },
		{ 126, 172, 52 },
		{ 126, 173, 52 },
		{ 126, 174, 52 },
		{ 126, 175, 52 },
		{ 127, 144, 52 },
		{ 127, 145, 52 },
		{ 127, 146, 52 },
		{ 127, 147, 52 },
		{ 127, 148, 52 },
		{ 127, 149, 52 },
		{ 127, 150, 52 },
		{ 127, 151, 52 },
		{ 127, 152, 52 },
		{ 127, 153, 52 },
		{ 127, 154, 52 },
		{ 127, 155, 52 },
		{ 127, 156, 52 },
		{ 127, 157, 52 },
		{ 127, 158, 52 },
		{ 127, 159, 52 },
		{ 127, 160, 52 },
		{ 127, 161, 52 },
		{ 127, 162, 52 },
		{ 127, 163, 52 },
		{ 127, 164, 52 },
		{ 127, 165, 52 },
		{ 127, 166, 52 },
		{ 127, 167, 52 },
		{ 127, 168, 52 },
		{ 127, 169, 52 },
		{ 127, 170, 52 },
		{ 127, 171, 52 },
		{ 127, 172, 52 },
		{ 127, 173, 52 },
		{ 127, 174, 52 },
		{ 127, 175, 52 },
		{ 128, 145, 52 },
		{ 128, 146, 52 },
		{ 128, 147, 52 },
		{ 128, 148, 52 },
		{ 128, 149, 52 },
		{ 128, 150, 52 },
		{ 128, 151, 52 },
		{ 128, 152, 52 },
		{ 128, 153, 52 },
		{ 128, 154, 52 },
		{ 128, 155, 52 },
		{ 128, 156, 52 },
		{ 128, 157, 52 },
		{ 128, 158, 52 },
		{ 128, 159, 52 },
		{ 128, 160, 52 },
		{ 128, 161, 52 },
		{ 128, 162, 52 },
		{ 128, 163, 52 },
		{ 128, 164, 52 },
		{ 128, 165, 52 },
		{ 128, 166, 52 },
		{ 128, 167, 52 },
		{ 128, 168, 52 },
		{ 128, 169, 52 },
		{ 128, 170, 52 },
		{ 128, 171, 52 },
		{ 128, 172, 52 },
		{ 128, 173, 52 },
		{ 128, 174, 52 },
		{ 128, 175, 52 },
		{ 128, 176, 52 },
		{ 129, 146, 52 },
		{ 129, 147, 52 },
		{ 129, 148, 52 },
		{ 129, 149, 52 },
		{ 129, 150, 52 },
		{ 129, 151, 52 },
		{ 129, 152, 52 },
		{ 129, 153, 52 },
		{ 129, 154, 52 },
		{ 129, 155, 52 },
		{ 129, 156, 52 },
		{ 129, 157, 52 },
		{ 129, 158, 52 },
		{ 129, 159, 52 },
		{ 129, 160, 52 },
		{ 129, 161, 52 },
		{ 129, 162, 52 },
		{ 129, 163, 52 },
		{ 129, 164, 52 },
		{ 129, 165, 52 },
		{ 129, 166, 52 },
		{ 129, 167, 52 },
		{ 129, 168, 52 },
		{ 129, 169, 52 },
		{ 129, 170, 52 },
		{ 129, 171, 52 },
		{ 129, 172, 52 },
		{ 129, 173, 52 },
		{ 129, 174, 52 },
		{ 129, 175, 52 },
		{ 129, 176, 52 },
		{ 130, 146, 52 },
		{ 130, 147, 52 },
		{ 130, 148, 52 },
		{ 130, 149, 52 },
		{ 130, 150, 52 },
		{ 130, 151, 52 },
		{ 130, 152, 52 },
		{ 130, 153, 52 },
		{ 130, 154, 52 },
		{ 130, 155, 52 },
		{ 130, 156, 52 },
		{ 130, 157, 52 },
		{ 130, 158, 52 },
		{ 130, 159, 52 },
		{ 130, 160, 52 },
		{ 130, 161, 52 },
		{ 130, 162, 52 },
		{ 130, 163, 52 },
		{ 130, 164, 52 },
		{ 130, 165, 52 },
		{ 130, 166, 52 },
		{ 130, 167, 52 },
		{ 130, 168, 52 },
		{ 130, 169, 52 },
		{ 130, 170, 52 },
		{ 130, 171, 52 },
		{ 130, 172, 52 },
		{ 130, 173, 52 },
		{ 130, 174, 52 },
		{ 130, 175, 52 },
		{ 130, 176, 52 },
		{ 130, 177, 52 },
		{ 131, 147, 52 },
		{ 131, 148, 52 },
		{ 131, 149, 52 },
		{ 131, 150, 52 },
		{ 131, 151, 52 },
		{ 131, 152, 52 },
		{ 131, 153, 52 },
		{ 131, 154, 52 },
		{ 131, 155, 52 },
		{ 131, 156, 52 },
		{ 131, 157, 52 },
		{ 131, 158, 52 },
		{ 131, 159, 52 },
		{ 131, 160, 52 },
		{ 131, 161, 52 },
		{ 131, 162, 52 },
		{ 131, 163, 52 },
		{ 131, 164, 52 },
		{ 131, 165, 52 },
		{ 131, 166, 52 },
		{ 131, 167, 52 },
		{ 131, 168, 52 },
		{ 131, 169, 52 },
		{ 131, 170, 52 },
		{ 131, 171, 52 },
		{ 131, 172, 52 },
		{ 131, 173, 52 },
		{ 131, 174, 52 },
		{ 131, 175, 52 },
		{ 131, 176, 52 },
		{ 131, 177, 52 },
		{ 131, 178, 52 },
		{ 132, 148, 52 },
		{ 132, 149, 52 },
		{ 132, 150, 52 },
		{ 132, 151, 52 },
		{ 132, 152, 52 },
		{ 132, 153, 52 },
		{ 132, 154, 52 },
		{ 132, 155, 52 },
		{ 132, 156, 52 },
		{ 132, 157, 52 },
		{ 132, 158, 52 },
		{ 132, 159, 52 },
		{ 132, 160, 52 },
		{ 132, 161, 52 },
		{ 132, 162, 52 },
		{ 132, 163, 52 },
		{ 132, 164, 52 },
		{ 132, 165, 52 },
		{ 132, 166, 52 },
		{ 132, 167, 52 },
		{ 132, 168, 52 },
		{ 132, 169, 52 },
		{ 132, 170, 52 },
		{ 132, 171, 52 },
		{ 132, 172, 52 },
		{ 132, 173, 52 },
		{ 132, 174, 52 },
		{ 132, 175, 52 },
		{ 132, 176, 52 },
		{ 132, 177, 52 },
		{ 132, 178, 52 },
		{ 133, 149, 52 },
		{ 133, 150, 52 },
		{ 133, 151, 52 },
		{ 133, 152, 52 },
		{ 133, 153, 52 },
		{ 133, 154, 52 },
		{ 133, 155, 52 },
		{ 133, 156, 52 },
		{ 133, 157, 52 },
		{ 133, 158, 52 },
		{ 133, 159, 52 },
		{ 133, 160, 52 },
		{ 133, 161, 52 },
		{ 133, 162, 52 },
		{ 133, 163, 52 },
		{ 133, 164, 52 },
		{ 133, 165, 52 },
		{ 133, 166, 52 },
		{ 133, 167, 52 },
		{ 133, 168, 52 },
		{ 133, 169, 52 },
		{ 133, 170, 52 },
		{ 133, 171, 52 },
		{ 133, 172, 52 },
		{ 133, 173, 52 },
		{ 133, 174, 52 },
		{ 133, 175, 52 },
		{ 133, 176, 52 },
		{ 133, 177, 52 },
		{ 133, 178, 52 },
		{ 133, 179, 52 },
		{ 134, 150, 52 },
		{ 134, 151, 52 },
		{ 134, 152, 52 },
		{ 134, 153, 52 },
		{ 134, 154, 52 },
		{ 134, 155, 52 },
		{ 134, 156, 52 },
		{ 134, 157, 52 },
		{ 134, 158, 52 },
		{ 134, 159, 52 },
		{ 134, 160, 52 },
		{ 134, 161, 52 },
		{ 134, 162, 52 },
		{ 134, 163, 52 },
		{ 134, 164, 52 },
		{ 134, 165, 52 },
		{ 134, 166, 52 },
		{ 134, 167, 52 },
		{ 134, 168, 52 },
		{ 134, 169, 52 },
		{ 134, 170, 52 },
		{ 134, 171, 52 },
		{ 134, 172, 52 },
		{ 134, 173, 52 },
		{ 134, 174, 52 },
		{ 134, 175, 52 },
		{ 134, 176, 52 },
		{ 134, 177, 52 },
		{ 134, 178, 52 },
		{ 134, 179, 52 },
		{ 134, 180, 52 },
		{ 135, 151, 52 },
		{ 135, 152, 52 },
		{ 135, 153, 52 },
		{ 135, 154, 52 },
		{ 135, 155, 52 },
		{ 135, 156, 52 },
		{ 135, 157, 52 },
		{ 135, 158, 52 },
		{ 135, 159, 52 },
		{ 135, 160, 52 },
		{ 135, 161, 52 },
		{ 135, 162, 52 },
		{ 135, 163, 52 },
		{ 135, 164, 52 },
		{ 135, 165, 52 },
		{ 135, 166, 52 },
		{ 135, 167, 52 },
		{ 135, 168, 52 },
		{ 135, 169, 52 },
		{ 135, 170, 52 },
		{ 135, 171, 52 },
		{ 135, 172, 52 },
		{ 135, 173, 52 },
		{ 135, 174, 52 },
		{ 135, 175, 52 },
		{ 135, 176, 52 },
		{ 135, 177, 52 },
		{ 135, 178, 52 },
		{ 135, 179, 52 },
		{ 135, 180, 52 },
		{ 136, 152, 52 },
		{ 136, 153, 52 },
		{ 136, 154, 52 },
		{ 136, 155, 52 },
		{ 136, 156, 52 },
		{ 136, 157, 52 },
		{ 136, 158, 52 },
		{ 136, 159, 52 },
		{ 136, 160, 52 },
		{ 136, 161, 52 },
		{ 136, 162, 52 },
		{ 136, 163, 52 },
		{ 136, 164, 52 },
		{ 136, 165, 52 },
		{ 136, 166, 52 },
		{ 136, 167, 52 },
		{ 136, 168, 52 },
		{ 136, 169, 52 },
		{ 136, 170, 52 },
		{ 136, 171, 52 },
		{ 136, 172, 52 },
		{ 136, 173, 52 },
		{ 136, 174, 52 },
		{ 136, 175, 52 },
		{ 136, 176, 52 },
		{ 136, 177, 52 },
		{ 136, 178, 52 },
		{ 136, 179, 52 },
		{ 136, 180, 52 },
		{ 136, 181, 52 },
		{ 137, 153, 52 },
		{ 137, 154, 52 },
		{ 137, 155, 52 },
		{ 137, 156, 52 },
		{ 137, 157, 52 },
		{ 137, 158, 52 },
		{ 137, 159, 52 },
		{ 137, 160, 52 },
		{ 137, 161, 52 },
		{ 137, 162, 52 },
		{ 137, 163, 52 },
		{ 137, 164, 52 },
		{ 137, 165, 52 },
		{ 137, 166, 52 },
		{ 137, 167, 52 },
		{ 137, 168, 52 },
		{ 137, 169, 52 },
		{ 137, 170, 52 },
		{ 137, 171, 52 },
		{ 137, 172, 52 },
		{ 137, 173, 52 },
		{ 137, 174, 52 },
		{ 137, 175, 52 },
		{ 137, 176, 52 },
		{ 137, 177, 52 },
		{ 137, 178, 52 },
		{ 137, 179, 52 },
		{ 137, 180, 52 },
		{ 137, 181, 52 },
		{ 138, 153, 52 },
		{ 138, 154, 52 },
		{ 138, 155, 52 },
		{ 138, 156, 52 },
		{ 138, 157, 52 },
		{ 138, 158, 52 },
		{ 138, 159, 52 },
		{ 138, 160, 52 },
		{ 138, 161, 52 },
		{ 138, 162, 52 },
		{ 138, 163, 52 },
		{ 138, 164, 52 },
		{ 138, 165, 52 },
		{ 138, 166, 52 },
		{ 138, 167, 52 },
		{ 138, 168, 52 },
		{ 138, 169, 52 },
		{ 138, 170, 52 },
		{ 138, 171, 52 },
		{ 138, 172, 52 },
		{ 138, 173, 52 },
		{ 138, 174, 52 },
		{ 138, 175, 52 },
		{ 138, 176, 52 },
		{ 138, 177, 52 },
		{ 138, 178, 52 },
		{ 138, 179, 52 },
		{ 138, 180, 52 },
		{ 138, 181, 52 },
		{ 138, 182, 52 },
		{ 139, 154, 52 },
		{ 139, 155, 52 },
		{ 139, 156, 52 },
		{ 139, 157, 52 },
		{ 139, 158, 52 },
		{ 139, 159, 52 },
		{ 139, 160, 52 },
		{ 139, 161, 52 },
		{ 139, 162, 52 },
		{ 139, 163, 52 },
		{ 139, 164, 52 },
		{ 139, 165, 52 },
		{ 139, 166, 52 },
		{ 139, 167, 52 },
		{ 139, 168, 52 },
		{ 139, 169, 52 },
		{ 139, 170, 52 },
		{ 139, 171, 52 },
		{ 139, 172, 52 },
		{ 139, 173, 52 },
		{ 139, 174, 52 },
		{ 139, 175, 52 },
		{ 139, 176, 52 },
		{ 139, 177, 52 },
		{ 139, 178, 52 },
		{ 139, 179, 52 },
		{ 139, 180, 52 },
		{ 139, 181, 52 },
		{ 139, 182, 52 },
		{ 139, 183, 52 },
		{ 140, 155, 52 },
		{ 140, 156, 52 },
		{ 140, 157, 52 },
		{ 140, 158, 52 },
		{ 140, 159, 52 },
		{ 140, 160, 52 },
		{ 140, 161, 52 },
		{ 140, 162, 52 },
		{ 140, 163, 52 },
		{ 140, 164, 52 },
		{ 140, 165, 52 },
		{ 140, 166, 52 },
		{ 140, 167, 52 },
		{ 140, 168, 52 },
		{ 140, 169, 52 },
		{ 140, 170, 52 },
		{ 140, 171, 52 },
		{ 140, 172, 52 },
		{ 140, 173, 52 },
		{ 140, 174, 52 },
		{ 140, 175, 52 },
		{ 140, 176, 52 },
		{ 140, 177, 52 },
		{ 140, 178, 52 },
		{ 140, 179, 52 },
		{ 140, 180, 52 },
		{ 140, 181, 52 },
		{ 140, 182, 52 },
		{ 140, 183, 52 },
		{ 141, 156, 52 },
		{ 141, 157, 52 },
		{ 141, 158, 52 },
		{ 141, 159, 52 },
		{ 141, 160, 52 },
		{ 141, 161, 52 },
		{ 141, 162, 52 },
		{ 141, 163, 52 },
		{ 141, 164, 52 },
		{ 141, 165, 52 },
		{ 141, 166, 52 },
		{ 141, 167, 52 },
		{ 141, 168, 52 },
		{ 141, 169, 52 },
		{ 141, 170, 52 },
		{ 141, 171, 52 },
		{ 141, 172, 52 },
		{ 141, 173, 52 },
		{ 141, 174, 52 },
		{ 141, 175, 52 },
		{ 141, 176, 52 },
		{ 141, 177, 52 },
		{ 141, 178, 52 },
		{ 141, 179, 52 },
		{ 141, 180, 52 },
		{ 141, 181, 52 },
		{ 141, 182, 52 },
		{ 141, 183, 52 },
		{ 141, 184, 52 },
		{ 142, 157, 52 },
		{ 142, 158, 52 },
		{ 142, 159, 52 },
		{ 142, 160, 52 },
		{ 142, 161, 52 },
		{ 142, 162, 52 },
		{ 142, 163, 52 },
		{ 142, 164, 52 },
		{ 142, 165, 52 },
		{ 142, 166, 52 },
		{ 142, 167, 52 },
		{ 142, 168, 52 },
		{ 142, 169, 52 },
		{ 142, 170, 52 },
		{ 142, 171, 52 },
		{ 142, 172, 52 },
		{ 142, 173, 52 },
		{ 142, 174, 52 },
		{ 142, 175, 52 },
		{ 142, 176, 52 },
		{ 142, 177, 52 },
		{ 142, 178, 52 },
		{ 142, 179, 52 },
		{ 142, 180, 52 },
		{ 142, 181, 52 },
		{ 142, 182, 52 },
		{ 142, 183, 52 },
		{ 142, 184, 52 },
		{ 142, 185, 52 },
		{ 143, 158, 52 },
		{ 143, 159, 52 },
		{ 143, 160, 52 },
		{ 143, 161, 52 },
		{ 143, 162, 52 },
		{ 143, 163, 52 },
		{ 143, 164, 52 },
		{ 143, 165, 52 },
		{ 143, 166, 52 },
		{ 143, 167, 52 },
		{ 143, 168, 52 },
		{ 143, 169, 52 },
		{ 143, 170, 52 },
		{ 143, 171, 52 },
		{ 143, 172, 52 },
		{ 143, 173, 52 },
		{ 143, 174, 52 },
		{ 143, 175, 52 },
		{ 143, 176, 52 },
		{ 143, 177, 52 },
		{ 143, 178, 52 },
		{ 143, 179, 52 },
		{ 143, 180, 52 },
		{ 143, 181, 52 },
		{ 143, 182, 52 },
		{ 143, 183, 52 },
		{ 143, 184, 52 },
		{ 143, 185, 52 },
		{ 144, 159, 52 },
		{ 144, 160, 52 },
		{ 144, 161, 52 },
		{ 144, 162, 52 },
		{ 144, 163, 52 },
		{ 144, 164, 52 },
		{ 144, 165, 52 },
		{ 144, 166, 52 },
		{ 144, 167, 52 },
		{ 144, 168, 52 },
		{ 144, 169, 52 },
		{ 144, 170, 52 },
		{ 144, 171, 52 },
		{ 144, 172, 52 },
		{ 144, 173, 52 },
		{ 144, 174, 52 },
		{ 144, 175, 52 },
		{ 144, 176, 52 },
		{ 144, 177, 52 },
		{ 144, 178, 52 },
		{ 144, 179, 52 },
		{ 144, 180, 52 },
		{ 144, 181, 52 },
		{ 144, 182, 52 },
		{ 144, 183, 52 },
		{ 144, 184, 52 },
		{ 144, 185, 52 },
		{ 144, 186, 52 },
		{ 145, 160, 52 },
		{ 145, 161, 52 },
		{ 145, 162, 52 },
		{ 145, 163, 52 },
		{ 145, 164, 52 },
		{ 145, 165, 52 },
		{ 145, 166, 52 },
		{ 145, 167, 52 },
		{ 145, 168, 52 },
		{ 145, 169, 52 },
		{ 145, 170, 52 },
		{ 145, 171, 52 },
		{ 145, 172, 52 },
		{ 145, 173, 52 },
		{ 145, 174, 52 },
		{ 145, 175, 52 },
		{ 145, 176, 52 },
		{ 145, 177, 52 },
		{ 145, 178, 52 },
		{ 145, 179, 52 },
		{ 145, 180, 52 },
		{ 145, 181, 52 },
		{ 145, 182, 52 },
		{ 145, 183, 52 },
		{ 145, 184, 52 },
		{ 145, 185, 52 },
		{ 145, 186, 52 },
		{ 146, 160, 52 },
		{ 146, 161, 52 },
		{ 146, 162, 52 },
		{ 146, 163, 52 },
		{ 146, 164, 52 },
		{ 146, 165, 52 },
		{ 146, 166, 52 },
		{ 146, 167, 52 },
		{ 146, 168, 52 },
		{ 146, 169, 52 },
		{ 146, 170, 52 },
		{ 146, 171, 52 },
		{ 146, 172, 52 },
		{ 146, 173, 52 },
		{ 146, 174, 52 },
		{ 146, 175, 52 },
		{ 146, 176, 52 },
		{ 146, 177, 52 },
		{ 146, 178, 52 },
		{ 146, 179, 52 },
		{ 146, 180, 52 },
		{ 146, 181, 52 },
		{ 146, 182, 52 },
		{ 146, 183, 52 },
		{ 146, 184, 52 },
		{ 146, 185, 52 },
		{ 146, 186, 52 },
		{ 146, 187, 52 },
		{ 147, 161, 52 },
		{ 147, 162, 52 },
		{ 147, 163, 52 },
		{ 147, 164, 52 },
		{ 147, 165, 52 },
		{ 147, 166, 52 },
		{ 147, 167, 52 },
		{ 147, 168, 52 },
		{ 147, 169, 52 },
		{ 147, 170, 52 },
		{ 147, 171, 52 },
		{ 147, 172, 52 },
		{ 147, 173, 52 },
		{ 147, 174, 52 },
		{ 147, 175, 52 },
		{ 147, 176, 52 },
		{ 147, 177, 52 },
		{ 147, 178, 52 },
		{ 147, 179, 52 },
		{ 147, 180, 52 },
		{ 147, 181, 52 },
		{ 147, 182, 52 },
		{ 147, 183, 52 },
		{ 147, 184, 52 },
		{ 147, 185, 52 },
		{ 147, 186, 52 },
		{ 147, 187, 52 },
		{ 147, 188, 52 },
		{ 148, 162, 52 },
		{ 148, 163, 52 },
		{ 148, 164, 52 },
		{ 148, 165, 52 },
		{ 148, 166, 52 },
		{ 148, 167, 52 },
		{ 148, 168, 52 },
		{ 148, 169, 52 },
		{ 148, 170, 52 },
		{ 148, 171, 52 },
		{ 148, 172, 52 },
		{ 148, 173, 52 },
		{ 148, 174, 52 },
		{ 148, 175, 52 },
		{ 148, 176, 52 },
		{ 148, 177, 52 },
		{ 148, 178, 52 },
		{ 148, 179, 52 },
		{ 148, 180, 52 },
		{ 148, 181, 52 },
		{ 148, 182, 52 },
		{ 148, 183, 52 },
		{ 148, 184, 52 },
		{ 148, 185, 52 },
		{ 148, 186, 52 },
		{ 148, 187, 52 },
		{ 148, 188, 52 },
		{ 149, 163, 52 },
		{ 149, 164, 52 },
		{ 149, 165, 52 },
		{ 149, 166, 52 },
		{ 149, 167, 52 },
		{ 149, 168, 52 },
		{ 149, 169, 52 },
		{ 149, 170, 52 },
		{ 149, 171, 52 },
		{ 149, 172, 52 },
		{ 149, 173, 52 },
		{ 149, 174, 52 },
		{ 149, 175, 52 },
		{ 149, 176, 52 },
		{ 149, 177, 52 },
		{ 149, 178, 52 },
		{ 149, 179, 52 },
		{ 149, 180, 52 },
		{ 149, 181, 52 },
		{ 149, 182, 52 },
		{ 149, 183, 52 },
		{ 149, 184, 52 },
		{ 149, 185, 52 },
		{ 149, 186, 52 },
		{ 149, 187, 52 },
		{ 149, 188, 52 },
		{ 149, 189, 52 },
		{ 150, 164, 52 },
		{ 150, 165, 52 },
		{ 150, 166, 52 },
		{ 150, 167, 52 },
		{ 150, 168, 52 },
		{ 150, 169, 52 },
		{ 150, 170, 52 },
		{ 150, 171, 52 },
		{ 150, 172, 52 },
		{ 150, 173, 52 },
		{ 150, 174, 52 },
		{ 150, 175, 52 },
		{ 150, 176, 52 },
		{ 150, 177, 52 },
		{ 150, 178, 52 },
		{ 150, 179, 52 },
		{ 150, 180, 52 },
		{ 150, 181, 52 },
		{ 150, 182, 52 },
		{ 150, 183, 52 },
		{ 150, 184, 52 },
		{ 150, 185, 52 },
		{ 150, 186, 52 },
		{ 150, 187, 52 },
		{ 150, 188, 52 },
		{ 150, 189, 52 },
		{ 150, 190, 52 },
		{ 151, 165, 52 },
		{ 151, 166, 52 },
		{ 151, 167, 52 },
		{ 151, 168, 52 },
		{ 151, 169, 52 },
		{ 151, 170, 52 },
		{ 151, 171, 52 },
		{ 151, 172, 52 },
		{ 151, 173, 52 },
		{ 151, 174, 52 },
		{ 151, 175, 52 },
		{ 151, 176, 52 },
		{ 151, 177, 52 },
		{ 151, 178, 52 },
		{ 151, 179, 52 },
		{ 151, 180, 52 },
		{ 151, 181, 52 },
		{ 151, 182, 52 },
		{ 151, 183, 52 },
		{ 151, 184, 52 },
		{ 151, 185, 52 },
		{ 151, 186, 52 },
		{ 151, 187, 52 },
		{ 151, 188, 52 },
		{ 151, 189, 52 },
		{ 151, 190, 52 },
		{ 152, 166, 52 },
		{ 152, 167, 52 },
		{ 152, 168, 52 },
		{ 152, 169, 52 },
		{ 152, 170, 52 },
		{ 152, 171, 52 },
		{ 152, 172, 52 },
		{ 152, 173, 52 },
		{ 152, 174, 52 },
		{ 152, 175, 52 },
		{ 152, 176, 52 },
		{ 152, 177, 52 },
		{ 152, 178, 52 },
		{ 152, 179, 52 },
		{ 152, 180, 52 },
		{ 152, 181, 52 },
		{ 152, 182, 52 },
		{ 152, 183, 52 },
		{ 152, 184, 52 },
		{ 152, 185, 52 },
		{ 152, 186, 52 },
		{ 152, 187, 52 },
		{ 152, 188, 52 },
		{ 152, 189, 52 },
		{ 152, 190, 52 },
		{ 152, 191, 52 },
		{ 153, 167, 52 },
		{ 153, 168, 52 },
		{ 153, 169, 52 },
		{ 153, 170, 52 },
		{ 153, 171, 52 },
		{ 153, 172, 52 },
		{ 153, 173, 52 },
		{ 153, 174, 52 },
		{ 153, 175, 52 },
		{ 153, 176, 52 },
		{ 153, 177, 52 },
		{ 153, 178, 52 },
		{ 153, 179, 52 },
		{ 153, 180, 52 },
		{ 153, 181, 52 },
		{ 153, 182, 52 },
		{ 153, 183, 52 },
		{ 153, 184, 52 },
		{ 153, 185, 52 },
		{ 153, 186, 52 },
		{ 153, 187, 52 },
		{ 153, 188, 52 },
		{ 153, 189, 52 },
		{ 153, 190, 52 },
		{ 153, 191, 52 },
		{ 154, 167, 52 },
		{ 154, 168, 52 },
		{ 154, 169, 52 },
		{ 154, 170, 52 },
		{ 154, 171, 52 },
		{ 154, 172, 52 },
		{ 154, 173, 52 },
		{ 154, 174, 52 },
		{ 154, 175, 52 },
		{ 154, 176, 52 },
		{ 154, 177, 52 },
		{ 154, 178, 52 },
		{ 154, 179, 52 },
		{ 154, 180, 52 },
		{ 154, 181, 52 },
		{ 154, 182, 52 },
		{ 154, 183, 52 },
		{ 154, 184, 52 },
		{ 154, 185, 52 },
		{ 154, 186, 52 },
		{ 154, 187, 52 },
		{ 154, 188, 52 },
		{ 154, 189, 52 },
		{ 154, 190, 52 },
		{ 154, 191, 52 },
		{ 154, 192, 52 },
		{ 155, 168, 52 },
		{ 155, 169, 52 },
		{ 155, 170, 52 },
		{ 155, 171, 52 },
		{ 155, 172, 52 },
		{ 155, 173, 52 },
		{ 155, 174, 52 },
		{ 155, 175, 52 },
		{ 155, 176, 52 },
		{ 155, 177, 52 },
		{ 155, 178, 52 },
		{ 155, 179, 52 },
		{ 155, 180, 52 },
		{ 155, 181, 52 },
		{ 155, 182, 52 },
		{ 155, 183, 52 },
		{ 155, 184, 52 },
		{ 155, 185, 52 },
		{ 155, 186, 52 },
		{ 155, 187, 52 },
		{ 155, 188, 52 },
		{ 155, 189, 52 },
		{ 155, 190, 52 },
		{ 155, 191, 52 },
		{ 155, 192, 52 },
		{ 155, 193, 52 },
		{ 156, 169, 52 },
		{ 156, 170, 52 },
		{ 156, 171, 52 },
		{ 156, 172, 52 },
		{ 156, 173, 52 },
		{ 156, 174, 52 },
		{ 156, 175, 52 },
		{ 156, 176, 52 },
		{ 156, 177, 52 },
		{ 156, 178, 52 },
		{ 156, 179, 52 },
		{ 156, 180, 52 },
		{ 156, 181, 52 },
		{ 156, 182, 52 },
		{ 156, 183, 52 },
		{ 156, 184, 52 },
		{ 156, 185, 52 },
		{ 156, 186, 52 },
		{ 156, 187, 52 },
		{ 156, 188, 52 },
		{ 156, 189, 52 },
		{ 156, 190, 52 },
		{ 156, 191, 52 },
		{ 156, 192, 52 },
		{ 156, 193, 52 },
		{ 157, 170, 52 },
		{ 157, 171, 52 },
		{ 157, 172, 52 },
		{ 157, 173, 52 },
		{ 157, 174, 52 },
		{ 157, 175, 52 },
		{ 157, 176, 52 },
		{ 157, 177, 52 },
		{ 157, 178, 52 },
		{ 157, 179, 52 },
		{ 157, 180, 52 },
		{ 157, 181, 52 },
		{ 157, 182, 52 },
		{ 157, 183, 52 },
		{ 157, 184, 52 },
		{ 157, 185, 52 },
		{ 157, 186, 52 },
		{ 157, 187, 52 },
		{ 157, 188, 52 },
		{ 157, 189, 52 },
		{ 157, 190, 52 },
		{ 157, 191, 52 },
		{ 157, 192, 52 },
		{ 157, 193, 52 },
		{ 157, 194, 52 },
		{ 158, 171, 52 },
		{ 158, 172, 52 },
		{ 158, 173, 52 },
		{ 158, 174, 52 },
		{ 158, 175, 52 },
		{ 158, 176, 52 },
		{ 158, 177, 52 },
		{ 158, 178, 52 },
		{ 158, 179, 52 },
		{ 158, 180, 52 },
		{ 158, 181, 52 },
		{ 158, 182, 52 },
		{ 158, 183, 52 },
		{ 158, 184, 52 },
		{ 158, 185, 52 },
		{ 158, 186, 52 },
		{ 158, 187, 52 },
		{ 158, 188, 52 },
		{ 158, 189, 52 },
		{ 158, 190, 52 },
		{ 158, 191, 52 },
		{ 158, 192, 52 },
		{ 158, 193, 52 },
		{ 158, 194, 52 },
		{ 158, 195, 52 },
		{ 159, 172, 52 },
		{ 159, 173, 52 },
		{ 159, 174, 52 },
		{ 159, 175, 52 },
		{ 159, 176, 52 },
		{ 159, 177, 52 },
		{ 159, 178, 52 },
		{ 159, 179, 52 },
		{ 159, 180, 52 },
		{ 159, 181, 52 },
		{ 159, 182, 52 },
		{ 159, 183, 52 },
		{ 159, 184, 52 },
		{ 159, 185, 52 },
		{ 159, 186, 52 },
		{ 159, 187, 52 },
		{ 159, 188, 52 },
		{ 159, 189, 52 },
		{ 159, 190, 52 },
		{ 159, 191, 52 },
		{ 159, 192, 52 },
		{ 159, 193, 52 },
		{ 159, 194, 52 },
		{ 159, 195, 52 },
		{ 160, 173, 52 },
		{ 160, 174, 52 },
		{ 160, 175, 52 },
		{ 160, 176, 52 },
		{ 160, 177, 52 },
		{ 160, 178, 52 },
		{ 160, 179, 52 },
		{ 160, 180, 52 },
		{ 160, 181, 52 },
		{ 160, 182, 52 },
		{ 160, 183, 52 },
		{ 160, 184, 52 },
		{ 160, 185, 52 },
		{ 160, 186, 52 },
		{ 160, 187, 52 },
		{ 160, 188, 52 },
		{ 160, 189, 52 },
		{ 160, 190, 52 },
		{ 160, 191, 52 },
		{ 160, 192, 52 },
		{ 160, 193, 52 },
		{ 160, 194, 52 },
		{ 160, 195, 52 },
		{ 160, 196, 52 },
		{ 161, 174, 52 },
		{ 161, 175, 52 },
		{ 161, 176, 52 },
		{ 161, 177, 52 },
		{ 161, 178, 52 },
		{ 161, 179, 52 },
		{ 161, 180, 52 },
		{ 161, 181, 52 },
		{ 161, 182, 52 },
		{ 161, 183, 52 },
		{ 161, 184, 52 },
		{ 161, 185, 52 },
		{ 161, 186, 52 },
		{ 161, 187, 52 },
		{ 161, 188, 52 },
		{ 161, 189, 52 },
		{ 161, 190, 52 },
		{ 161, 191, 52 },
		{ 161, 192, 52 },
		{ 161, 193, 52 },
		{ 161, 194, 52 },
		{ 161, 195, 52 },
		{ 161, 196, 52 },
		{ 162, 174, 52 },
		{ 162, 175, 52 },
		{ 162, 176, 52 },
		{ 162, 177, 52 },
		{ 162, 178, 52 },
		{ 162, 179, 52 },
		{ 162, 180, 52 },
		{ 162, 181, 52 },
		{ 162, 182, 52 },
		{ 162, 183, 52 },
		{ 162, 184, 52 },
		{ 162, 185, 52 },
		{ 162, 186, 52 },
		{ 162, 187, 52 },
		{ 162, 188, 52 },
		{ 162, 189, 52 },
		{ 162, 190, 52 },
		{ 162, 191, 52 },
		{ 162, 192, 52 },
		{ 162, 193, 52 },
		{ 162, 194, 52 },
		{ 162, 195, 52 },
		{ 162, 196, 52 },
		{ 162, 197, 52 },
		{ 163, 175, 52 },
		{ 163, 176, 52 },
		{ 163, 177, 52 },
		{ 163, 178, 52 },
		{ 163, 179, 52 },
		{ 163, 180, 52 },
		{ 163, 181, 52 },
		{ 163, 182, 52 },
		{ 163, 183, 52 },
		{ 163, 184, 52 },
		{ 163, 185, 52 },
		{ 163, 186, 52 },
		{ 163, 187, 52 },
		{ 163, 188, 52 },
		{ 163, 189, 52 },
		{ 163, 190, 52 },
		{ 163, 191, 52 },
		{ 163, 192, 52 },
		{ 163, 193, 52 },
		{ 163, 194, 52 },
		{ 163, 195, 52 },
		{ 163, 196, 52 },
		{ 163, 197, 52 },
		{ 163, 198, 52 },
		{ 164, 176, 52 },
		{ 164, 177, 52 },
		{ 164, 178, 52 },
		{ 164, 179, 52 },
		{ 164, 180, 52 },
		{ 164, 181, 52 },
		{ 164, 182, 52 },
		{ 164, 183, 52 },
		{ 164, 184, 52 },
		{ 164, 185, 52 },
		{ 164, 186, 52 },
		{ 164, 187, 52 },
		{ 164, 188, 52 },
		{ 164, 189, 52 },
		{ 164, 190, 52 },
		{ 164, 191, 52 },
		{ 164, 192, 52 },
		{ 164, 193, 52 },
		{ 164, 194, 52 },
		{ 164, 195, 52 },
		{ 164, 196, 52 },
		{ 164, 197, 52 },
		{ 164, 198, 52 },
		{ 165, 177, 52 },
		{ 165, 178, 52 },
		{ 165, 179, 52 },
		{ 165, 180, 52 },
		{ 165, 181, 52 },
		{ 165, 182, 52 },
		{ 165, 183, 52 },
		{ 165, 184, 52 },
		{ 165, 185, 52 },
		{ 165, 186, 52 },
		{ 165, 187, 52 },
		{ 165, 188, 52 },
		{ 165, 189, 52 },
		{ 165, 190, 52 },
		{ 165, 191, 52 },
		{ 165, 192, 52 },
		{ 165, 193, 52 },
		{ 165, 194, 52 },
		{ 165, 195, 52 },
		{ 165, 196, 52 },
		{ 165, 197, 52 },
		{ 165, 198, 52 },
		{ 165, 199, 52 },
		{ 166, 178, 52 },
		{ 166, 179, 52 },
		{ 166, 180, 52 },
		{ 166, 181, 52 },
		{ 166, 182, 52 },
		{ 166, 183, 52 },
		{ 166, 184, 52 },
		{ 166, 185, 52 },
		{ 166, 186, 52 },
		{ 166, 187, 52 },
		{ 166, 188, 52 },
		{ 166, 189, 52 },
		{ 166, 190, 52 },
		{ 166, 191, 52 },
		{ 166, 192, 52 },
		{ 166, 193, 52 },
		{ 166, 194, 52 },
		{ 166, 195, 52 },
		{ 166, 196, 52 },
		{ 166, 197, 52 },
		{ 166, 198, 52 },
		{ 166, 199, 52 },
		{ 166, 200, 52 },
		{ 167, 179, 52 },
		{ 167, 180, 52 },
		{ 167, 181, 52 },
		{ 167, 182, 52 },
		{ 167, 183, 52 },
		{ 167, 184, 52 },
		{ 167, 185, 52 },
		{ 167, 186, 52 },
		{ 167, 187, 52 },
		{ 167, 188, 52 },
		{ 167, 189, 52 },
		{ 167, 190, 52 },
		{ 167, 191, 52 },
		{ 167, 192, 52 },
		{ 167, 193, 52 },
		{ 167, 194, 52 },
		{ 167, 195, 52 },
		{ 167, 196, 52 },
		{ 167, 197, 52 },
		{ 167, 198, 52 },
		{ 167, 199, 52 },
		{ 167, 200, 52 },
		{ 168, 180, 52 },
		{ 168, 181, 52 },
		{ 168, 182, 52 },
		{ 168, 183, 52 },
		{ 168, 184, 52 },
		{ 168, 185, 52 },
		{ 168, 186, 52 },
		{ 168, 187, 52 },
		{ 168, 188, 52 },
		{ 168, 189, 52 },
		{ 168, 190, 52 },
		{ 168, 191, 52 },
		{ 168, 192, 52 },
		{ 168, 193, 52 },
		{ 168, 194, 52 },
		{ 168, 195, 52 },
		{ 168, 196, 52 },
		{ 168, 197, 52 },
		{ 168, 198, 52 },
		{ 168, 199, 52 },
		{ 168, 200, 52 },
		{ 168, 201, 52 },
		{ 169, 181, 52 },
		{ 169, 182, 52 },
		{ 169, 183, 52 },
		{ 169, 184, 52 },
		{ 169, 185, 52 },
		{ 169, 186, 52 },
		{ 169, 187, 52 },
		{ 169, 188, 52 },
		{ 169, 189, 52 },
		{ 169, 190, 52 },
		{ 169, 191, 52 },
		{ 169, 192, 52 },
		{ 169, 193, 52 },
		{ 169, 194, 52 },
		{ 169, 195, 52 },
		{ 169, 196, 52 },
		{ 169, 197, 52 },
		{ 169, 198, 52 },
		{ 169, 199, 52 },
		{ 169, 200, 52 },
		{ 169, 201, 52 },
		{ 170, 181, 52 },
		{ 170, 182, 52 },
		{ 170, 183, 52 },
		{ 170, 184, 52 },
		{ 170, 185, 52 },
		{ 170, 186, 52 },
		{ 170, 187, 52 },
		{ 170, 188, 52 },
		{ 170, 189, 52 },
		{ 170, 190, 52 },
		{ 170, 191, 52 },
		{ 170, 192, 52 },
		{ 170, 193, 52 },
		{ 170, 194, 52 },
		{ 170, 195, 52 },
		{ 170, 196, 52 },
		{ 170, 197, 52 },
		{ 170, 198, 52 },
		{ 170, 199, 52 },
		{ 170, 200, 52 },
		{ 170, 201, 52 },
		{ 170, 202, 52 },
		{ 171, 182, 52 },
		{ 171, 183, 52 },
		{ 171, 184, 52 },
		{ 171, 185, 52 },
		{ 171, 186, 52 },
		{ 171, 187, 52 },
		{ 171, 188, 52 },
		{ 171, 189, 52 },
		{ 171, 190, 52 },
		{ 171, 191, 52 },
		{ 171, 192, 52 },
		{ 171, 193, 52 },
		{ 171, 194, 52 },
		{ 171, 195, 52 },
		{ 171, 196, 52 },
		{ 171, 197, 52 },
		{ 171, 198, 52 },
		{ 171, 199, 52 },
		{ 171, 200, 52 },
		{ 171, 201, 52 },
		{ 171, 202, 52 },
		{ 171, 203, 52 },
		{ 172, 183, 52 },
		{ 172, 184, 52 },
		{ 172, 185, 52 },
		{ 172, 186, 52 },
		{ 172, 187, 52 },
		{ 172, 188, 52 },
		{ 172, 189, 52 },
		{ 172, 190, 52 },
		{ 172, 191, 52 },
		{ 172, 192, 52 },
		{ 172, 193, 52 },
		{ 172, 194, 52 },
		{ 172, 195, 52 },
		{ 172, 196, 52 },
		{ 172, 197, 52 },
		{ 172, 198, 52 },
		{ 172, 199, 52 },
		{ 172, 200, 52 },
		{ 172, 201, 52 },
		{ 172, 202, 52 },
		{ 172, 203, 52 },
		{ 173, 184, 52 },
		{ 173, 185, 52 },
		{ 173, 186, 52 },
		{ 173, 187, 52 },
		{ 173, 188, 52 },
		{ 173, 189, 52 },
		{ 173, 190, 52 },
		{ 173, 191, 52 },
		{ 173, 192, 52 },
		{ 173, 193, 52 },
		{ 173, 194, 52 },
		{ 173, 195, 52 },
		{ 173, 196, 52 },
		{ 173, 197, 52 },
		{ 173, 198, 52 },
		{ 173, 199, 52 },
		{ 173, 200, 52 },
		{ 173, 201, 52 },
		{ 173, 202, 52 },
		{ 173, 203, 52 },
		{ 173, 204, 52 },
		{ 174, 185, 52 },
		{ 174, 186, 52 },
		{ 174, 187, 52 },
		{ 174, 188, 52 },
		{ 174, 189, 52 },
		{ 174, 190, 52 },
		{ 174, 191, 52 },
		{ 174, 192, 52 },
		{ 174, 193, 52 },
		{ 174, 194, 52 },
		{ 174, 195, 52 },
		{ 174, 196, 52 },
		{ 174, 197, 52 },
		{ 174, 198, 52 },
		{ 174, 199, 52 },
		{ 174, 200, 52 },
		{ 174, 201, 52 },
		{ 174, 202, 52 },
		{ 174, 203, 52 },
		{ 174, 204, 52 },
		{ 174, 205, 52 },
		{ 175, 186, 52 },
		{ 175, 187, 52 },
		{ 175, 188, 52 },
		{ 175, 189, 52 },
		{ 175, 190, 52 },
		{ 175, 191, 52 },
		{ 175, 192, 52 },
		{ 175, 193, 52 },
		{ 175, 194, 52 },
		{ 175, 195, 52 },
		{ 175, 196, 52 },
		{ 175, 197, 52 },
		{ 175, 198, 52 },
		{ 175, 199, 52 },
		{ 175, 200, 52 },
		{ 175, 201, 52 },
		{ 175, 202, 52 },
		{ 175, 203, 52 },
		{ 175, 204, 52 },
		{ 175, 205, 52 },
		{ 176, 187, 52 },
		{ 176, 188, 52 },
		{ 176, 189, 52 },
		{ 176, 190, 52 },
		{ 176, 191, 52 },
		{ 176, 192, 52 },
		{ 176, 193, 52 },
		{ 176, 194, 52 },
		{ 176, 195, 52 },
		{ 176, 196, 52 },
		{ 176, 197, 52 },
		{ 176, 198, 52 },
		{ 176, 199, 52 },
		{ 176, 200, 52 },
		{ 176, 201, 52 },
		{ 176, 202, 52 },
		{ 176, 203, 52 },
		{ 176, 204, 52 },
		{ 176, 205, 52 },
		{ 176, 206, 52 },
		{ 177, 188, 52 },
		{ 177, 189, 52 },
		{ 177, 190, 52 },
		{ 177, 191, 52 },
		{ 177, 192, 52 },
		{ 177, 193, 52 },
		{ 177, 194, 52 },
		{ 177, 195, 52 },
		{ 177, 196, 52 },
		{ 177, 197, 52 },
		{ 177, 198, 52 },
		{ 177, 199, 52 },
		{ 177, 200, 52 },
		{ 177, 201, 52 },
		{ 177, 202, 52 },
		{ 177, 203, 52 },
		{ 177, 204, 52 },
		{ 177, 205, 52 },
		{ 177, 206, 52 },
		{ 178, 188, 52 },
		{ 178, 189, 52 },
		{ 178, 190, 52 },
		{ 178, 191, 52 },
		{ 178, 192, 52 },
		{ 178, 193, 52 },
		{ 178, 194, 52 },
		{ 178, 195, 52 },
		{ 178, 196, 52 },
		{ 178, 197, 52 },
		{ 178, 198, 52 },
		{ 178, 199, 52 },
		{ 178, 200, 52 },
		{ 178, 201, 52 },
		{ 178, 202, 52 },
		{ 178, 203, 52 },
		{ 178, 204, 52 },
		{ 178, 205, 52 },
		{ 178, 206, 52 },
		{ 178, 207, 52 },
		{ 179, 189, 52 },
		{ 179, 190, 52 },
		{ 179, 191, 52 },
		{ 179, 192, 52 },
		{ 179, 193, 52 },
		{ 179, 194, 52 },
		{ 179, 195, 52 },
		{ 179, 196, 52 },
		{ 179, 197, 52 },
		{ 179, 198, 52 },
		{ 179, 199, 52 },
		{ 179, 200, 52 },
		{ 179, 201, 52 },
		{ 179, 202, 52 },
		{ 179, 203, 52 },
		{ 179, 204, 52 },
		{ 179, 205, 52 },
		{ 179, 206, 52 },
		{ 179, 207, 52 },
		{ 179, 208, 52 },
		{ 180, 190, 52 },
		{ 180, 191, 52 },
		{ 180, 192, 52 },
		{ 180, 193, 52 },
		{ 180, 194, 52 },
		{ 180, 195, 52 },
		{ 180, 196, 52 },
		{ 180, 197, 52 },
		{ 180, 198, 52 },
		{ 180, 199, 52 },
		{ 180, 200, 52 },
		{ 180, 201, 52 },
		{ 180, 202, 52 },
		{ 180, 203, 52 },
		{ 180, 204, 52 },
		{ 180, 205, 52 },
		{ 180, 206, 52 },
		{ 180, 207, 52 },
		{ 180, 208, 52 },
		{ 181, 191, 52 },
		{ 181, 192, 52 },
		{ 181, 193, 52 },
		{ 181, 194, 52 },
		{ 181, 195, 52 },
		{ 181, 196, 52 },
		{ 181, 197, 52 },
		{ 181, 198, 52 },
		{ 181, 199, 52 },
		{ 181, 200, 52 },
		{ 181, 201, 52 },
		{ 181, 202, 52 },
		{ 181, 203, 52 },
		{ 181, 204, 52 },
		{ 181, 205, 52 },
		{ 181, 206, 52 },
		{ 181, 207, 52 },
		{ 181, 208, 52 },
		{ 181, 209, 52 },
		{ 182, 192, 52 },
		{ 182, 193, 52 },
		{ 182, 194, 52 },
		{ 182, 195, 52 },
		{ 182, 196, 52 },
		{ 182, 197, 52 },
		{ 182, 198, 52 },
		{ 182, 199, 52 },
		{ 182, 200, 52 },
		{ 182, 201, 52 },
		{ 182, 202, 52 },
		{ 182, 203, 52 },
		{ 182, 204, 52 },
		{ 182, 205, 52 },
		{ 182, 206, 52 },
		{ 182, 207, 52 },
		{ 182, 208, 52 },
		{ 182, 209, 52 },
		{ 182, 210, 52 },
		{ 183, 193, 52 },
		{ 183, 194, 52 },
		{ 183, 195, 52 },
		{ 183, 196, 52 },
		{ 183, 197, 52 },
		{ 183, 198, 52 },
		{ 183, 199, 52 },
		{ 183, 200, 52 },
		{ 183, 201, 52 },
		{ 183, 202, 52 },
		{ 183, 203, 52 },
		{ 183, 204, 52 },
		{ 183, 205, 52 },
		{ 183, 206, 52 },
		{ 183, 207, 52 },
		{ 183, 208, 52 },
		{ 183, 209, 52 },
		{ 183, 210, 52 },
		{ 184, 194, 52 },
		{ 184, 195, 52 },
		{ 184, 196, 52 },
		{ 184, 197, 52 },
		{ 184, 198, 52 },
		{ 184, 199, 52 },
		{ 184, 200, 52 },
		{ 184, 201, 52 },
		{ 184, 202, 52 },
		{ 184, 203, 52 },
		{ 184, 204, 52 },
		{ 184, 205, 52 },
		{ 184, 206, 52 },
		{ 184, 207, 52 },
		{ 184, 208, 52 },
		{ 184, 209, 52 },
		{ 184, 210, 52 },
		{ 184, 211, 52 },
		{ 185, 195, 52 },
		{ 185, 196, 52 },
		{ 185, 197, 52 },
		{ 185, 198, 52 },
		{ 185, 199, 52 },
		{ 185, 200, 52 },
		{ 185, 201, 52 },
		{ 185, 202, 52 },
		{ 185, 203, 52 },
		{ 185, 204, 52 },
		{ 185, 205, 52 },
		{ 185, 206, 52 },
		{ 185, 207, 52 },
		{ 185, 208, 52 },
		{ 185, 209, 52 },
		{ 185, 210, 52 },
		{ 185, 211, 52 },
		{ 186, 196, 52 },
		{ 186, 197, 52 },
		{ 186, 198, 52 },
		{ 186, 199, 52 },
		{ 186, 200, 52 },
		{ 186, 201, 52 },
		{ 186, 202, 52 },
		{ 186, 203, 52 },
		{ 186, 204, 52 },
		{ 186, 205, 52 },
		{ 186, 206, 52 },
		{ 186, 207, 52 },
		{ 186, 208, 52 },
		{ 186, 209, 52 },
		{ 186, 210, 52 },
		{ 186, 211, 52 },
		{ 186, 212, 52 },
		{ 187, 196, 52 },
		{ 187, 197, 52 },
		{ 187, 198, 52 },
		{ 187, 199, 52 },
		{ 187, 200, 52 },
		{ 187, 201, 52 },
		{ 187, 202, 52 },
		{ 187, 203, 52 },
		{ 187, 204, 52 },
		{ 187, 205, 52 },
		{ 187, 206, 52 },
		{ 187, 207, 52 },
		{ 187, 208, 52 },
		{ 187, 209, 52 },
		{ 187, 210, 52 },
		{ 187, 211, 52 },
		{ 187, 212, 52 },
		{ 187, 213, 52 },
		{ 188, 197, 52 },
		{ 188, 198, 52 },
		{ 188, 199, 52 },
		{ 188, 200, 52 },
		{ 188, 201, 52 },
		{ 188, 202, 52 },
		{ 188, 203, 52 },
		{ 188, 204, 52 },
		{ 188, 205, 52 },
		{ 188, 206, 52 },
		{ 188, 207, 52 },
		{ 188, 208, 52 },
		{ 188, 209, 52 },
		{ 188, 210, 52 },
		{ 188, 211, 52 },
		{ 188, 212, 52 },
		{ 188, 213, 52 },
		{ 189, 198, 52 },
		{ 189, 199, 52 },
		{ 189, 200, 52 },
		{ 189, 201, 52 },
		{ 189, 202, 52 },
		{ 189, 203, 52 },
		{ 189, 204, 52 },
		{ 189, 205, 52 },
		{ 189, 206, 52 },
		{ 189, 207, 52 },
		{ 189, 208, 52 },
		{ 189, 209, 52 },
		{ 189, 210, 52 },
		{ 189, 211, 52 },
		{ 189, 212, 52 },
		{ 189, 213, 52 },
		{ 189, 214, 52 },
		{ 190, 200, 52 },
		{ 190, 201, 52 },
		{ 190, 202, 52 },
		{ 190, 203, 52 },
		{ 190, 204, 52 },
		{ 190, 205, 52 },
		{ 190, 206, 52 },
		{ 190, 207, 52 },
		{ 190, 208, 52 },
		{ 190, 209, 52 },
		{ 190, 210, 52 },
		{ 190, 211, 52 },
		{ 190, 212, 52 },
		{ 190, 213, 52 },
		{ 190, 214, 52 },
		{ 191, 202, 52 },
		{ 191, 203, 52 },
		{ 191, 204, 52 },
		{ 191, 205, 52 },
		{ 191, 206, 52 },
		{ 191, 207, 52 },
		{ 191, 208, 52 },
		{ 191, 209, 52 },
		{ 191, 210, 52 },
		{ 191, 211, 52 },
		{ 191, 212, 52 },
		{ 191, 213, 52 },
		{ 323, 301, 52 },
		{ 323, 302, 52 },
		{ 323, 303, 52 },
		{ 323, 304, 52 },
		{ 323, 305, 52 },
		{ 323, 306, 52 },
		{ 323, 307, 52 },
		{ 323, 308, 52 },
		{ 323, 309, 52 },
		{ 323, 310, 52 },
		{ 323, 311, 52 },
		{ 323, 312, 52 },
		{ 324, 300, 52 },
		{ 324, 301, 52 },
		{ 324, 302, 52 },
		{ 324, 303, 52 },
		{ 324, 304, 52 },
		{ 324, 305, 52 },
		{ 324, 306, 52 },
		{ 324, 307, 52 },
		{ 324, 308, 52 },
		{ 324, 309, 52 },
		{ 324, 310, 52 },
		{ 324, 311, 52 },
		{ 324, 312, 52 },
		{ 324, 313, 52 },
		{ 324, 314, 52 },
		{ 325, 300, 52 },
		{ 325, 301, 52 },
		{ 325, 302, 52 },
		{ 325, 303, 52 },
		{ 325, 304, 52 },
		{ 325, 305, 52 },
		{ 325, 306, 52 },
		{ 325, 307, 52 },
		{ 325, 308, 52 },
		{ 325, 309, 52 },
		{ 325, 310, 52 },
		{ 325, 311, 52 },
		{ 325, 312, 52 },
		{ 325, 313, 52 },
		{ 325, 314, 52 },
		{ 325, 315, 52 },
		{ 325, 316, 52 },
		{ 326, 301, 52 },
		{ 326, 302, 52 },
		{ 326, 303, 52 },
		{ 326, 304, 52 },
		{ 326, 305, 52 },
		{ 326, 306, 52 },
		{ 326, 307, 52 },
		{ 326, 308, 52 },
		{ 326, 309, 52 },
		{ 326, 310, 52 },
		{ 326, 311, 52 },
		{ 326, 312, 52 },
		{ 326, 313, 52 },
		{ 326, 314, 52 },
		{ 326, 315, 52 },
		{ 326, 316, 52 },
		{ 326, 317, 52 },
		{ 327, 301, 52 },
		{ 327, 302, 52 },
		{ 327, 303, 52 },
		{ 327, 304, 52 },
		{ 327, 305, 52 },
		{ 327, 306, 52 },
		{ 327, 307, 52 },
		{ 327, 308, 52 },
		{ 327, 309, 52 },
		{ 327, 310, 52 },
		{ 327, 311, 52 },
		{ 327, 312, 52 },
		{ 327, 313, 52 },
		{ 327, 314, 52 },
		{ 327, 315, 52 },
		{ 327, 316, 52 },
		{ 327, 317, 52 },
		{ 327, 318, 52 },
		{ 328, 302, 52 },
		{ 328, 303, 52 },
		{ 328, 304, 52 },
		{ 328, 305, 52 },
		{ 328, 306, 52 },
		{ 328, 307, 52 },
		{ 328, 308, 52 },
		{ 328, 309, 52 },
		{ 328, 310, 52 },
		{ 328, 311, 52 },
		{ 328, 312, 52 },
		{ 328, 313, 52 },
		{ 328, 314, 52 },
		{ 328, 315, 52 },
		{ 328, 316, 52 },
		{ 328, 317, 52 },
		{ 328, 318, 52 },
		{ 329, 303, 52 },
		{ 329, 304, 52 },
		{ 329, 305, 52 },
		{ 329, 306, 52 },
		{ 329, 307, 52 },
		{ 329, 308, 52 },
		{ 329, 309, 52 },
		{ 329, 310, 52 },
		{ 329, 311, 52 },
		{ 329, 312, 52 },
		{ 329, 313, 52 },
		{ 329, 314, 52 },
		{ 329, 315, 52 },
		{ 329, 316, 52 },
		{ 329, 317, 52 },
		{ 329, 318, 52 },
		{ 329, 319, 52 },
		{ 330, 303, 52 },
		{ 330, 304, 52 },
		{ 330, 305, 52 },
		{ 330, 306, 52 },
		{ 330, 307, 52 },
		{ 330, 308, 52 },
		{ 330, 309, 52 },
		{ 330, 310, 52 },
		{ 330, 311, 52 },
		{ 330, 312, 52 },
		{ 330, 313, 52 },
		{ 330, 314, 52 },
		{ 330, 315, 52 },
		{ 330, 316, 52 },
		{ 330, 317, 52 },
		{ 330, 318, 52 },
		{ 330, 319, 52 },
		{ 330, 320, 52 },
		{ 331, 304, 52 },
		{ 331, 305, 52 },
		{ 331, 306, 52 },
		{ 331, 307, 52 },
		{ 331, 308, 52 },
		{ 331, 309, 52 },
		{ 331, 310, 52 },
		{ 331, 311, 52 },
		{ 331, 312, 52 },
		{ 331, 313, 52 },
		{ 331, 314, 52 },
		{ 331, 315, 52 },
		{ 331, 316, 52 },
		{ 331, 317, 52 },
		{ 331, 318, 52 },
		{ 331, 319, 52 },
		{ 331, 320, 52 },
		{ 331, 321, 52 },
		{ 332, 304, 52 },
		{ 332, 305, 52 },
		{ 332, 306, 52 },
		{ 332, 307, 52 },
		{ 332, 308, 52 },
		{ 332, 309, 52 },
		{ 332, 310, 52 },
		{ 332, 311, 52 },
		{ 332, 312, 52 },
		{ 332, 313, 52 },
		{ 332, 314, 52 },
		{ 332, 315, 52 },
		{ 332, 316, 52 },
		{ 332, 317, 52 },
		{ 332, 318, 52 },
		{ 332, 319, 52 },
		{ 332, 320, 52 },
		{ 332, 321, 52 },
		{ 332, 322, 52 },
		{ 333, 305, 52 },
		{ 333, 306, 52 },
		{ 333, 307, 52 },
		{ 333, 308, 52 },
		{ 333, 309, 52 },
		{ 333, 310, 52 },
		{ 333, 311, 52 },
		{ 333, 312, 52 },
		{ 333, 313, 52 },
		{ 333, 314, 52 },
		{ 333, 315, 52 },
		{ 333, 316, 52 },
		{ 333, 317, 52 },
		{ 333, 318, 52 },
		{ 333, 319, 52 },
		{ 333, 320, 52 },
		{ 333, 321, 52 },
		{ 333, 322, 52 },
		{ 333, 323, 52 },
		{ 334, 306, 52 },
		{ 334, 307, 52 },
		{ 334, 308, 52 },
		{ 334, 309, 52 },
		{ 334, 310, 52 },
		{ 334, 311, 52 },
		{ 334, 312, 52 },
		{ 334, 313, 52 },
		{ 334, 314, 52 },
		{ 334, 315, 52 },
		{ 334, 316, 52 },
		{ 334, 317, 52 },
		{ 334, 318, 52 },
		{ 334, 319, 52 },
		{ 334, 320, 52 },
		{ 334, 321, 52 },
		{ 334, 322, 52 },
		{ 334, 323, 52 },
		{ 334, 324, 52 },
		{ 335, 306, 52 },
		{ 335, 307, 52 },
		{ 335, 308, 52 },
		{ 335, 309, 52 },
		{ 335, 310, 52 },
		{ 335, 311, 52 },
		{ 335, 312, 52 },
		{ 335, 313, 52 },
		{ 335, 314, 52 },
		{ 335, 315, 52 },
		{ 335, 316, 52 },
		{ 335, 317, 52 },
		{ 335, 318, 52 },
		{ 335, 319, 52 },
		{ 335, 320, 52 },
		{ 335, 321, 52 },
		{ 335, 322, 52 },
		{ 335, 323, 52 },
		{ 335, 324, 52 },
		{ 335, 325, 52 },
		{ 336, 307, 52 },
		{ 336, 308, 52 },
		{ 336, 309, 52 },
		{ 336, 310, 52 },
		{ 336, 311, 52 },
		{ 336, 312, 52 },
		{ 336, 313, 52 },
		{ 336, 314, 52 },
		{ 336, 315, 52 },
		{ 336, 316, 52 },
		{ 336, 317, 52 },
		{ 336, 318, 52 },
		{ 336, 319, 52 },
		{ 336, 320, 52 },
		{ 336, 321, 52 },
		{ 336, 322, 52 },
		{ 336, 323, 52 },
		{ 336, 324, 52 },
		{ 336, 325, 52 },
		{ 336, 326, 52 },
		{ 337, 308, 52 },
		{ 337, 309, 52 },
		{ 337, 310, 52 },
		{ 337, 311, 52 },
		{ 337, 312, 52 },
		{ 337, 313, 52 },
		{ 337, 314, 52 },
		{ 337, 315, 52 },
		{ 337, 316, 52 },
		{ 337, 317, 52 },
		{ 337, 318, 52 },
		{ 337, 319, 52 },
		{ 337, 320, 52 },
		{ 337, 321, 52 },
		{ 337, 322, 52 },
		{ 337, 323, 52 },
		{ 337, 324, 52 },
		{ 337, 325, 52 },
		{ 337, 326, 52 },
		{ 338, 308, 52 },
		{ 338, 309, 52 },
		{ 338, 310, 52 },
		{ 338, 311, 52 },
		{ 338, 312, 52 },
		{ 338, 313, 52 },
		{ 338, 314, 52 },
		{ 338, 315, 52 },
		{ 338, 316, 52 },
		{ 338, 317, 52 },
		{ 338, 318, 52 },
		{ 338, 319, 52 },
		{ 338, 320, 52 },
		{ 338, 321, 52 },
		{ 338, 322, 52 },
		{ 338, 323, 52 },
		{ 338, 324, 52 },
		{ 338, 325, 52 },
		{ 338, 326, 52 },
		{ 338, 327, 52 },
		{ 339, 309, 52 },
		{ 339, 310, 52 },
		{ 339, 311, 52 },
		{ 339, 312, 52 },
		{ 339, 313, 52 },
		{ 339, 314, 52 },
		{ 339, 315, 52 },
		{ 339, 316, 52 },
		{ 339, 317, 52 },
		{ 339, 318, 52 },
		{ 339, 319, 52 },
		{ 339, 320, 52 },
		{ 339, 321, 52 },
		{ 339, 322, 52 },
		{ 339, 323, 52 },
		{ 339, 324, 52 },
		{ 339, 325, 52 },
		{ 339, 326, 52 },
		{ 339, 327, 52 },
		{ 339, 328, 52 },
		{ 340, 309, 52 },
		{ 340, 310, 52 },
		{ 340, 311, 52 },
		{ 340, 312, 52 },
		{ 340, 313, 52 },
		{ 340, 314, 52 },
		{ 340, 315, 52 },
		{ 340, 316, 52 },
		{ 340, 317, 52 },
		{ 340, 318, 52 },
		{ 340, 319, 52 },
		{ 340, 320, 52 },
		{ 340, 321, 52 },
		{ 340, 322, 52 },
		{ 340, 323, 52 },
		{ 340, 324, 52 },
		{ 340, 325, 52 },
		{ 340, 326, 52 },
		{ 340, 327, 52 },
		{ 340, 328, 52 },
		{ 340, 329, 52 },
		{ 341, 310, 52 },
		{ 341, 311, 52 },
		{ 341, 312, 52 },
		{ 341, 313, 52 },
		{ 341, 314, 52 },
		{ 341, 315, 52 },
		{ 341, 316, 52 },
		{ 341, 317, 52 },
		{ 341, 318, 52 },
		{ 341, 319, 52 },
		{ 341, 320, 52 },
		{ 341, 321, 52 },
		{ 341, 322, 52 },
		{ 341, 323, 52 },
		{ 341, 324, 52 },
		{ 341, 325, 52 },
		{ 341, 326, 52 },
		{ 341, 327, 52 },
		{ 341, 328, 52 },
		{ 341, 329, 52 },
		{ 341, 330, 52 },
		{ 342, 311, 52 },
		{ 342, 312, 52 },
		{ 342, 313, 52 },
		{ 342, 314, 52 },
		{ 342, 315, 52 },
		{ 342, 316, 52 },
		{ 342, 317, 52 },
		{ 342, 318, 52 },
		{ 342, 319, 52 },
		{ 342, 320, 52 },
		{ 342, 321, 52 },
		{ 342, 322, 52 },
		{ 342, 323, 52 },
		{ 342, 324, 52 },
		{ 342, 325, 52 },
		{ 342, 326, 52 },
		{ 342, 327, 52 },
		{ 342, 328, 52 },
		{ 342, 329, 52 },
		{ 342, 330, 52 },
		{ 342, 331, 52 },
		{ 343, 311, 52 },
		{ 343, 312, 52 },
		{ 343, 313, 52 },
		{ 343, 314, 52 },
		{ 343, 315, 52 },
		{ 343, 316, 52 },
		{ 343, 317, 52 },
		{ 343, 318, 52 },
		{ 343, 319, 52 },
		{ 343, 320, 52 },
		{ 343, 321, 52 },
		{ 343, 322, 52 },
		{ 343, 323, 52 },
		{ 343, 324, 52 },
		{ 343, 325, 52 },
		{ 343, 326, 52 },
		{ 343, 327, 52 },
		{ 343, 328, 52 },
		{ 343, 329, 52 },
		{ 343, 330, 52 },
		{ 343, 331, 52 },
		{ 343, 332, 52 },
		{ 344, 312, 52 },
		{ 344, 313, 52 },
		{ 344, 314, 52 },
		{ 344, 315, 52 },
		{ 344, 316, 52 },
		{ 344, 317, 52 },
		{ 344, 318, 52 },
		{ 344, 319, 52 },
		{ 344, 320, 52 },
		{ 344, 321, 52 },
		{ 344, 322, 52 },
		{ 344, 323, 52 },
		{ 344, 324, 52 },
		{ 344, 325, 52 },
		{ 344, 326, 52 },
		{ 344, 327, 52 },
		{ 344, 328, 52 },
		{ 344, 329, 52 },
		{ 344, 330, 52 },
		{ 344, 331, 52 },
		{ 344, 332, 52 },
		{ 344, 333, 52 },
		{ 345, 313, 52 },
		{ 345, 314, 52 },
		{ 345, 315, 52 },
		{ 345, 316, 52 },
		{ 345, 317, 52 },
		{ 345, 318, 52 },
		{ 345, 319, 52 },
		{ 345, 320, 52 },
		{ 345, 321, 52 },
		{ 345, 322, 52 },
		{ 345, 323, 52 },
		{ 345, 324, 52 },
		{ 345, 325, 52 },
		{ 345, 326, 52 },
		{ 345, 327, 52 },
		{ 345, 328, 52 },
		{ 345, 329, 52 },
		{ 345, 330, 52 },
		{ 345, 331, 52 },
		{ 345, 332, 52 },
		{ 345, 333, 52 },
		{ 346, 313, 52 },
		{ 346, 314, 52 },
		{ 346, 315, 52 },
		{ 346, 316, 52 },
		{ 346, 317, 52 },
		{ 346, 318, 52 },
		{ 346, 319, 52 },
		{ 346, 320, 52 },
		{ 346, 321, 52 },
		{ 346, 322, 52 },
		{ 346, 323, 52 },
		{ 346, 324, 52 },
		{ 346, 325, 52 },
		{ 346, 326, 52 },
		{ 346, 327, 52 },
		{ 346, 328, 52 },
		{ 346, 329, 52 },
		{ 346, 330, 52 },
		{ 346, 331, 52 },
		{ 346, 332, 52 },
		{ 346, 333, 52 },
		{ 346, 334, 52 },
		{ 347, 314, 52 },
		{ 347, 315, 52 },
		{ 347, 316, 52 },
		{ 347, 317, 52 },
		{ 347, 318, 52 },
		{ 347, 319, 52 },
		{ 347, 320, 52 },
		{ 347, 321, 52 },
		{ 347, 322, 52 },
		{ 347, 323, 52 },
		{ 347, 324, 52 },
		{ 347, 325, 52 },
		{ 347, 326, 52 },
		{ 347, 327, 52 },
		{ 347, 328, 52 },
		{ 347, 329, 52 },
		{ 347, 330, 52 },
		{ 347, 331, 52 },
		{ 347, 332, 52 },
		{ 347, 333, 52 },
		{ 347, 334, 52 },
		{ 347, 335, 52 },
		{ 348, 314, 52 },
		{ 348, 315, 52 },
		{ 348, 316, 52 },
		{ 348, 317, 52 },
		{ 348, 318, 52 },
		{ 348, 319, 52 },
		{ 348, 320, 52 },
		{ 348, 321, 52 },
		{ 348, 322, 52 },
		{ 348, 323, 52 },
		{ 348, 324, 52 },
		{ 348, 325, 52 },
		{ 348, 326, 52 },
		{ 348, 327, 52 },
		{ 348, 328, 52 },
		{ 348, 329, 52 },
		{ 348, 330, 52 },
		{ 348, 331, 52 },
		{ 348, 332, 52 },
		{ 348, 333, 52 },
		{ 348, 334, 52 },
		{ 348, 335, 52 },
		{ 348, 336, 52 },
		{ 349, 315, 52 },
		{ 349, 316, 52 },
		{ 349, 317, 52 },
		{ 349, 318, 52 },
		{ 349, 319, 52 },
		{ 349, 320, 52 },
		{ 349, 321, 52 },
		{ 349, 322, 52 },
		{ 349, 323, 52 },
		{ 349, 324, 52 },
		{ 349, 325, 52 },
		{ 349, 326, 52 },
		{ 349, 327, 52 },
		{ 349, 328, 52 },
		{ 349, 329, 52 },
		{ 349, 330, 52 },
		{ 349, 331, 52 },
		{ 349, 332, 52 },
		{ 349, 333, 52 },
		{ 349, 334, 52 },
		{ 349, 335, 52 },
		{ 349, 336, 52 },
		{ 349, 337, 52 },
		{ 350, 316, 52 },
		{ 350, 317, 52 },
		{ 350, 318, 52 },
		{ 350, 319, 52 },
		{ 350, 320, 52 },
		{ 350, 321, 52 },
		{ 350, 322, 52 },
		{ 350, 323, 52 },
		{ 350, 324, 52 },
		{ 350, 325, 52 },
		{ 350, 326, 52 },
		{ 350, 327, 52 },
		{ 350, 328, 52 },
		{ 350, 329, 52 },
		{ 350, 330, 52 },
		{ 350, 331, 52 },
		{ 350, 332, 52 },
		{ 350, 333, 52 },
		{ 350, 334, 52 },
		{ 350, 335, 52 },
		{ 350, 336, 52 },
		{ 350, 337, 52 },
		{ 350, 338, 52 },
		{ 351, 316, 52 },
		{ 351, 317, 52 },
		{ 351, 318, 52 },
		{ 351, 319, 52 },
		{ 351, 320, 52 },
		{ 351, 321, 52 },
		{ 351, 322, 52 },
		{ 351, 323, 52 },
		{ 351, 324, 52 },
		{ 351, 325, 52 },
		{ 351, 326, 52 },
		{ 351, 327, 52 },
		{ 351, 328, 52 },
		{ 351, 329, 52 },
		{ 351, 330, 52 },
		{ 351, 331, 52 },
		{ 351, 332, 52 },
		{ 351, 333, 52 },
		{ 351, 334, 52 },
		{ 351, 335, 52 },
		{ 351, 336, 52 },
		{ 351, 337, 52 },
		{ 351, 338, 52 },
		{ 351, 339, 52 },
		{ 352, 317, 52 },
		{ 352, 318, 52 },
		{ 352, 319, 52 },
		{ 352, 320, 52 },
		{ 352, 321, 52 },
		{ 352, 322, 52 },
		{ 352, 323, 52 },
		{ 352, 324, 52 },
		{ 352, 325, 52 },
		{ 352, 326, 52 },
		{ 352, 327, 52 },
		{ 352, 328, 52 },
		{ 352, 329, 52 },
		{ 352, 330, 52 },
		{ 352, 331, 52 },
		{ 352, 332, 52 },
		{ 352, 333, 52 },
		{ 352, 334, 52 },
		{ 352, 335, 52 },
		{ 352, 336, 52 },
		{ 352, 337, 52 },
		{ 352, 338, 52 },
		{ 352, 339, 52 },
		{ 352, 340, 52 },
		{ 353, 318, 52 },
		{ 353, 319, 52 },
		{ 353, 320, 52 },
		{ 353, 321, 52 },
		{ 353, 322, 52 },
		{ 353, 323, 52 },
		{ 353, 324, 52 },
		{ 353, 325, 52 },
		{ 353, 326, 52 },
		{ 353, 327, 52 },
		{ 353, 328, 52 },
		{ 353, 329, 52 },
		{ 353, 330, 52 },
		{ 353, 331, 52 },
		{ 353, 332, 52 },
		{ 353, 333, 52 },
		{ 353, 334, 52 },
		{ 353, 335, 52 },
		{ 353, 336, 52 },
		{ 353, 337, 52 },
		{ 353, 338, 52 },
		{ 353, 339, 52 },
		{ 353, 340, 52 },
		{ 354, 318, 52 },
		{ 354, 319, 52 },
		{ 354, 320, 52 },
		{ 354, 321, 52 },
		{ 354, 322, 52 },
		{ 354, 323, 52 },
		{ 354, 324, 52 },
		{ 354, 325, 52 },
		{ 354, 326, 52 },
		{ 354, 327, 52 },
		{ 354, 328, 52 },
		{ 354, 329, 52 },
		{ 354, 330, 52 },
		{ 354, 331, 52 },
		{ 354, 332, 52 },
		{ 354, 333, 52 },
		{ 354, 334, 52 },
		{ 354, 335, 52 },
		{ 354, 336, 52 },
		{ 354, 337, 52 },
		{ 354, 338, 52 },
		{ 354, 339, 52 },
		{ 354, 340, 52 },
		{ 354, 341, 52 },
		{ 355, 319, 52 },
		{ 355, 320, 52 },
		{ 355, 321, 52 },
		{ 355, 322, 52 },
		{ 355, 323, 52 },
		{ 355, 324, 52 },
		{ 355, 325, 52 },
		{ 355, 326, 52 },
		{ 355, 327, 52 },
		{ 355, 328, 52 },
		{ 355, 329, 52 },
		{ 355, 330, 52 },
		{ 355, 331, 52 },
		{ 355, 332, 52 },
		{ 355, 333, 52 },
		{ 355, 334, 52 },
		{ 355, 335, 52 },
		{ 355, 336, 52 },
		{ 355, 337, 52 },
		{ 355, 338, 52 },
		{ 355, 339, 52 },
		{ 355, 340, 52 },
		{ 355, 341, 52 },
		{ 355, 342, 52 },
		{ 356, 319, 52 },
		{ 356, 320, 52 },
		{ 356, 321, 52 },
		{ 356, 322, 52 },
		{ 356, 323, 52 },
		{ 356, 324, 52 },
		{ 356, 325, 52 },
		{ 356, 326, 52 },
		{ 356, 327, 52 },
		{ 356, 328, 52 },
		{ 356, 329, 52 },
		{ 356, 330, 52 },
		{ 356, 331, 52 },
		{ 356, 332, 52 },
		{ 356, 333, 52 },
		{ 356, 334, 52 },
		{ 356, 335, 52 },
		{ 356, 336, 52 },
		{ 356, 337, 52 },
		{ 356, 338, 52 },
		{ 356, 339, 52 },
		{ 356, 340, 52 },
		{ 356, 341, 52 },
		{ 356, 342, 52 },
		{ 356, 343, 52 },
		{ 357, 320, 52 },
		{ 357, 321, 52 },
		{ 357, 322, 52 },
		{ 357, 323, 52 },
		{ 357, 324, 52 },
		{ 357, 325, 52 },
		{ 357, 326, 52 },
		{ 357, 327, 52 },
		{ 357, 328, 52 },
		{ 357, 329, 52 },
		{ 357, 330, 52 },
		{ 357, 331, 52 },
		{ 357, 332, 52 },
		{ 357, 333, 52 },
		{ 357, 334, 52 },
		{ 357, 335, 52 },
		{ 357, 336, 52 },
		{ 357, 337, 52 },
		{ 357, 338, 52 },
		{ 357, 339, 52 },
		{ 357, 340, 52 },
		{ 357, 341, 52 },
		{ 357, 342, 52 },
		{ 357, 343, 52 },
		{ 357, 344, 52 },
		{ 358, 321, 52 },
		{ 358, 322, 52 },
		{ 358, 323, 52 },
		{ 358, 324, 52 },
		{ 358, 325, 52 },
		{ 358, 326, 52 },
		{ 358, 327, 52 },
		{ 358, 328, 52 },
		{ 358, 329, 52 },
		{ 358, 330, 52 },
		{ 358, 331, 52 },
		{ 358, 332, 52 },
		{ 358, 333, 52 },
		{ 358, 334, 52 },
		{ 358, 335, 52 },
		{ 358, 336, 52 },
		{ 358, 337, 52 },
		{ 358, 338, 52 },
		{ 358, 339, 52 },
		{ 358, 340, 52 },
		{ 358, 341, 52 },
		{ 358, 342, 52 },
		{ 358, 343, 52 },
		{ 358, 344, 52 },
		{ 358, 345, 52 },
		{ 359, 321, 52 },
		{ 359, 322, 52 },
		{ 359, 323, 52 },
		{ 359, 324, 52 },
		{ 359, 325, 52 },
		{ 359, 326, 52 },
		{ 359, 327, 52 },
		{ 359, 328, 52 },
		{ 359, 329, 52 },
		{ 359, 330, 52 },
		{ 359, 331, 52 },
		{ 359, 332, 52 },
		{ 359, 333, 52 },
		{ 359, 334, 52 },
		{ 359, 335, 52 },
		{ 359, 336, 52 },
		{ 359, 337, 52 },
		{ 359, 338, 52 },
		{ 359, 339, 52 },
		{ 359, 340, 52 },
		{ 359, 341, 52 },
		{ 359, 342, 52 },
		{ 359, 343, 52 },
		{ 359, 344, 52 },
		{ 359, 345, 52 },
		{ 359, 346, 52 },
		{ 360, 322, 52 },
		{ 360, 323, 52 },
		{ 360, 324, 52 },
		{ 360, 325, 52 },
		{ 360, 326, 52 },
		{ 360, 327, 52 },
		{ 360, 328, 52 },
		{ 360, 329, 52 },
		{ 360, 330, 52 },
		{ 360, 331, 52 },
		{ 360, 332, 52 },
		{ 360, 333, 52 },
		{ 360, 334, 52 },
		{ 360, 335, 52 },
		{ 360, 336, 52 },
		{ 360, 337, 52 },
		{ 360, 338, 52 },
		{ 360, 339, 52 },
		{ 360, 340, 52 },
		{ 360, 341, 52 },
		{ 360, 342, 52 },
		{ 360, 343, 52 },
		{ 360, 344, 52 },
		{ 360, 345, 52 },
		{ 360, 346, 52 },
		{ 360, 347, 52 },
		{ 361, 323, 52 },
		{ 361, 324, 52 },
		{ 361, 325, 52 },
		{ 361, 326, 52 },
		{ 361, 327, 52 },
		{ 361, 328, 52 },
		{ 361, 329, 52 },
		{ 361, 330, 52 },
		{ 361, 331, 52 },
		{ 361, 332, 52 },
		{ 361, 333, 52 },
		{ 361, 334, 52 },
		{ 361, 335, 52 },
		{ 361, 336, 52 },
		{ 361, 337, 52 },
		{ 361, 338, 52 },
		{ 361, 339, 52 },
		{ 361, 340, 52 },
		{ 361, 341, 52 },
		{ 361, 342, 52 },
		{ 361, 343, 52 },
		{ 361, 344, 52 },
		{ 361, 345, 52 },
		{ 361, 346, 52 },
		{ 361, 347, 52 },
		{ 362, 323, 52 },
		{ 362, 324, 52 },
		{ 362, 325, 52 },
		{ 362, 326, 52 },
		{ 362, 327, 52 },
		{ 362, 328, 52 },
		{ 362, 329, 52 },
		{ 362, 330, 52 },
		{ 362, 331, 52 },
		{ 362, 332, 52 },
		{ 362, 333, 52 },
		{ 362, 334, 52 },
		{ 362, 335, 52 },
		{ 362, 336, 52 },
		{ 362, 337, 52 },
		{ 362, 338, 52 },
		{ 362, 339, 52 },
		{ 362, 340, 52 },
		{ 362, 341, 52 },
		{ 362, 342, 52 },
		{ 362, 343, 52 },
		{ 362, 344, 52 },
		{ 362, 345, 52 },
		{ 362, 346, 52 },
		{ 362, 347, 52 },
		{ 362, 348, 52 },
		{ 363, 324, 52 },
		{ 363, 325, 52 },
		{ 363, 326, 52 },
		{ 363, 327, 52 },
		{ 363, 328, 52 },
		{ 363, 329, 52 },
		{ 363, 330, 52 },
		{ 363, 331, 52 },
		{ 363, 332, 52 },
		{ 363, 333, 52 },
		{ 363, 334, 52 },
		{ 363, 335, 52 },
		{ 363, 336, 52 },
		{ 363, 337, 52 },
		{ 363, 338, 52 },
		{ 363, 339, 52 },
		{ 363, 340, 52 },
		{ 363, 341, 52 },
		{ 363, 342, 52 },
		{ 363, 343, 52 },
		{ 363, 344, 52 },
		{ 363, 345, 52 },
		{ 363, 346, 52 },
		{ 363, 347, 52 },
		{ 363, 348, 52 },
		{ 363, 349, 52 },
		{ 364, 324, 52 },
		{ 364, 325, 52 },
		{ 364, 326, 52 },
		{ 364, 327, 52 },
		{ 364, 328, 52 },
		{ 364, 329, 52 },
		{ 364, 330, 52 },
		{ 364, 331, 52 },
		{ 364, 332, 52 },
		{ 364, 333, 52 },
		{ 364, 334, 52 },
		{ 364, 335, 52 },
		{ 364, 336, 52 },
		{ 364, 337, 52 },
		{ 364, 338, 52 },
		{ 364, 339, 52 },
		{ 364, 340, 52 },
		{ 364, 341, 52 },
		{ 364, 342, 52 },
		{ 364, 343, 52 },
		{ 364, 344, 52 },
		{ 364, 345, 52 },
		{ 364, 346, 52 },
		{ 364, 347, 52 },
		{ 364, 348, 52 },
		{ 364, 349, 52 },
		{ 364, 350, 52 },
		{ 365, 325, 52 },
		{ 365, 326, 52 },
		{ 365, 327, 52 },
		{ 365, 328, 52 },
		{ 365, 329, 52 },
		{ 365, 330, 52 },
		{ 365, 331, 52 },
		{ 365, 332, 52 },
		{ 365, 333, 52 },
		{ 365, 334, 52 },
		{ 365, 335, 52 },
		{ 365, 336, 52 },
		{ 365, 337, 52 },
		{ 365, 338, 52 },
		{ 365, 339, 52 },
		{ 365, 340, 52 },
		{ 365, 341, 52 },
		{ 365, 342, 52 },
		{ 365, 343, 52 },
		{ 365, 344, 52 },
		{ 365, 345, 52 },
		{ 365, 346, 52 },
		{ 365, 347, 52 },
		{ 365, 348, 52 },
		{ 365, 349, 52 },
		{ 365, 350, 52 },
		{ 365, 351, 52 },
		{ 366, 326, 52 },
		{ 366, 327, 52 },
		{ 366, 328, 52 },
		{ 366, 329, 52 },
		{ 366, 330, 52 },
		{ 366, 331, 52 },
		{ 366, 332, 52 },
		{ 366, 333, 52 },
		{ 366, 334, 52 },
		{ 366, 335, 52 },
		{ 366, 336, 52 },
		{ 366, 337, 52 },
		{ 366, 338, 52 },
		{ 366, 339, 52 },
		{ 366, 340, 52 },
		{ 366, 341, 52 },
		{ 366, 342, 52 },
		{ 366, 343, 52 },
		{ 366, 344, 52 },
		{ 366, 345, 52 },
		{ 366, 346, 52 },
		{ 366, 347, 52 },
		{ 366, 348, 52 },
		{ 366, 349, 52 },
		{ 366, 350, 52 },
		{ 366, 351, 52 },
		{ 366, 352, 52 },
		{ 367, 326, 52 },
		{ 367, 327, 52 },
		{ 367, 328, 52 },
		{ 367, 329, 52 },
		{ 367, 330, 52 },
		{ 367, 331, 52 },
		{ 367, 332, 52 },
		{ 367, 333, 52 },
		{ 367, 334, 52 },
		{ 367, 335, 52 },
		{ 367, 336, 52 },
		{ 367, 337, 52 },
		{ 367, 338, 52 },
		{ 367, 339, 52 },
		{ 367, 340, 52 },
		{ 367, 341, 52 },
		{ 367, 342, 52 },
		{ 367, 343, 52 },
		{ 367, 344, 52 },
		{ 367, 345, 52 },
		{ 367, 346, 52 },
		{ 367, 347, 52 },
		{ 367, 348, 52 },
		{ 367, 349, 52 },
		{ 367, 350, 52 },
		{ 367, 351, 52 },
		{ 367, 352, 52 },
		{ 367, 353, 52 },
		{ 368, 327, 52 },
		{ 368, 328, 52 },
		{ 368, 329, 52 },
		{ 368, 330, 52 },
		{ 368, 331, 52 },
		{ 368, 332, 52 },
		{ 368, 333, 52 },
		{ 368, 334, 52 },
		{ 368, 335, 52 },
		{ 368, 336, 52 },
		{ 368, 337, 52 },
		{ 368, 338, 52 },
		{ 368, 339, 52 },
		{ 368, 340, 52 },
		{ 368, 341, 52 },
		{ 368, 342, 52 },
		{ 368, 343, 52 },
		{ 368, 344, 52 },
		{ 368, 345, 52 },
		{ 368, 346, 52 },
		{ 368, 347, 52 },
		{ 368, 348, 52 },
		{ 368, 349, 52 },
		{ 368, 350, 52 },
		{ 368, 351, 52 },
		{ 368, 352, 52 },
		{ 368, 353, 52 },
		{ 368, 354, 52 },
		{ 369, 328, 52 },
		{ 369, 329, 52 },
		{ 369, 330, 52 },
		{ 369, 331, 52 },
		{ 369, 332, 52 },
		{ 369, 333, 52 },
		{ 369, 334, 52 },
		{ 369, 335, 52 },
		{ 369, 336, 52 },
		{ 369, 337, 52 },
		{ 369, 338, 52 },
		{ 369, 339, 52 },
		{ 369, 340, 52 },
		{ 369, 341, 52 },
		{ 369, 342, 52 },
		{ 369, 343, 52 },
		{ 369, 344, 52 },
		{ 369, 345, 52 },
		{ 369, 346, 52 },
		{ 369, 347, 52 },
		{ 369, 348, 52 },
		{ 369, 349, 52 },
		{ 369, 350, 52 },
		{ 369, 351, 52 },
		{ 369, 352, 52 },
		{ 369, 353, 52 },
		{ 369, 354, 52 },
		{ 370, 328, 52 },
		{ 370, 329, 52 },
		{ 370, 330, 52 },
		{ 370, 331, 52 },
		{ 370, 332, 52 },
		{ 370, 333, 52 },
		{ 370, 334, 52 },
		{ 370, 335, 52 },
		{ 370, 336, 52 },
		{ 370, 337, 52 },
		{ 370, 338, 52 },
		{ 370, 339, 52 },
		{ 370, 340, 52 },
		{ 370, 341, 52 },
		{ 370, 342, 52 },
		{ 370, 343, 52 },
		{ 370, 344, 52 },
		{ 370, 345, 52 },
		{ 370, 346, 52 },
		{ 370, 347, 52 },
		{ 370, 348, 52 },
		{ 370, 349, 52 },
		{ 370, 350, 52 },
		{ 370, 351, 52 },
		{ 370, 352, 52 },
		{ 370, 353, 52 },
		{ 370, 354, 52 },
		{ 370, 355, 52 },
		{ 371, 329, 52 },
		{ 371, 330, 52 },
		{ 371, 331, 52 },
		{ 371, 332, 52 },
		{ 371, 333, 52 },
		{ 371, 334, 52 },
		{ 371, 335, 52 },
		{ 371, 336, 52 },
		{ 371, 337, 52 },
		{ 371, 338, 52 },
		{ 371, 339, 52 },
		{ 371, 340, 52 },
		{ 371, 341, 52 },
		{ 371, 342, 52 },
		{ 371, 343, 52 },
		{ 371, 344, 52 },
		{ 371, 345, 52 },
		{ 371, 346, 52 },
		{ 371, 347, 52 },
		{ 371, 348, 52 },
		{ 371, 349, 52 },
		{ 371, 350, 52 },
		{ 371, 351, 52 },
		{ 371, 352, 52 },
		{ 371, 353, 52 },
		{ 371, 354, 52 },
		{ 371, 355, 52 },
		{ 371, 356, 52 },
		{ 372, 329, 52 },
		{ 372, 330, 52 },
		{ 372, 331, 52 },
		{ 372, 332, 52 },
		{ 372, 333, 52 },
		{ 372, 334, 52 },
		{ 372, 335, 52 },
		{ 372, 336, 52 },
		{ 372, 337, 52 },
		{ 372, 338, 52 },
		{ 372, 339, 52 },
		{ 372, 340, 52 },
		{ 372, 341, 52 },
		{ 372, 342, 52 },
		{ 372, 343, 52 },
		{ 372, 344, 52 },
		{ 372, 345, 52 },
		{ 372, 346, 52 },
		{ 372, 347, 52 },
		{ 372, 348, 52 },
		{ 372, 349, 52 },
		{ 372, 350, 52 },
		{ 372, 351, 52 },
		{ 372, 352, 52 },
		{ 372, 353, 52 },
		{ 372, 354, 52 },
		{ 372, 355, 52 },
		{ 372, 356, 52 },
		{ 372, 357, 52 },
		{ 373, 330, 52 },
		{ 373, 331, 52 },
		{ 373, 332, 52 },
		{ 373, 333, 52 },
		{ 373, 334, 52 },
		{ 373, 335, 52 },
		{ 373, 336, 52 },
		{ 373, 337, 52 },
		{ 373, 338, 52 },
		{ 373, 339, 52 },
		{ 373, 340, 52 },
		{ 373, 341, 52 },
		{ 373, 342, 52 },
		{ 373, 343, 52 },
		{ 373, 344, 52 },
		{ 373, 345, 52 },
		{ 373, 346, 52 },
		{ 373, 347, 52 },
		{ 373, 348, 52 },
		{ 373, 349, 52 },
		{ 373, 350, 52 },
		{ 373, 351, 52 },
		{ 373, 352, 52 },
		{ 373, 353, 52 },
		{ 373, 354, 52 },
		{ 373, 355, 52 },
		{ 373, 356, 52 },
		{ 373, 357, 52 },
		{ 373, 358, 52 },
		{ 374, 331, 52 },
		{ 374, 332, 52 },
		{ 374, 333, 52 },
		{ 374, 334, 52 },
		{ 374, 335, 52 },
		{ 374, 336, 52 },
		{ 374, 337, 52 },
		{ 374, 338, 52 },
		{ 374, 339, 52 },
		{ 374, 340, 52 },
		{ 374, 341, 52 },
		{ 374, 342, 52 },
		{ 374, 343, 52 },
		{ 374, 344, 52 },
		{ 374, 345, 52 },
		{ 374, 346, 52 },
		{ 374, 347, 52 },
		{ 374, 348, 52 },
		{ 374, 349, 52 },
		{ 374, 350, 52 },
		{ 374, 351, 52 },
		{ 374, 352, 52 },
		{ 374, 353, 52 },
		{ 374, 354, 52 },
		{ 374, 355, 52 },
		{ 374, 356, 52 },
		{ 374, 357, 52 },
		{ 374, 358, 52 },
		{ 374, 359, 52 },
		{ 375, 331, 52 },
		{ 375, 332, 52 },
		{ 375, 333, 52 },
		{ 375, 334, 52 },
		{ 375, 335, 52 },
		{ 375, 336, 52 },
		{ 375, 337, 52 },
		{ 375, 338, 52 },
		{ 375, 339, 52 },
		{ 375, 340, 52 },
		{ 375, 341, 52 },
		{ 375, 342, 52 },
		{ 375, 343, 52 },
		{ 375, 344, 52 },
		{ 375, 345, 52 },
		{ 375, 346, 52 },
		{ 375, 347, 52 },
		{ 375, 348, 52 },
		{ 375, 349, 52 },
		{ 375, 350, 52 },
		{ 375, 351, 52 },
		{ 375, 352, 52 },
		{ 375, 353, 52 },
		{ 375, 354, 52 },
		{ 375, 355, 52 },
		{ 375, 356, 52 },
		{ 375, 357, 52 },
		{ 375, 358, 52 },
		{ 375, 359, 52 },
		{ 375, 360, 52 },
		{ 376, 332, 52 },
		{ 376, 333, 52 },
		{ 376, 334, 52 },
		{ 376, 335, 52 },
		{ 376, 336, 52 },
		{ 376, 337, 52 },
		{ 376, 338, 52 },
		{ 376, 339, 52 },
		{ 376, 340, 52 },
		{ 376, 341, 52 },
		{ 376, 342, 52 },
		{ 376, 343, 52 },
		{ 376, 344, 52 },
		{ 376, 345, 52 },
		{ 376, 346, 52 },
		{ 376, 347, 52 },
		{ 376, 348, 52 },
		{ 376, 349, 52 },
		{ 376, 350, 52 },
		{ 376, 351, 52 },
		{ 376, 352, 52 },
		{ 376, 353, 52 },
		{ 376, 354, 52 },
		{ 376, 355, 52 },
		{ 376, 356, 52 },
		{ 376, 357, 52 },
		{ 376, 358, 52 },
		{ 376, 359, 52 },
		{ 376, 360, 52 },
		{ 376, 361, 52 },
		{ 377, 333, 52 },
		{ 377, 334, 52 },
		{ 377, 335, 52 },
		{ 377, 336, 52 },
		{ 377, 337, 52 },
		{ 377, 338, 52 },
		{ 377, 339, 52 },
		{ 377, 340, 52 },
		{ 377, 341, 52 },
		{ 377, 342, 52 },
		{ 377, 343, 52 },
		{ 377, 344, 52 },
		{ 377, 345, 52 },
		{ 377, 346, 52 },
		{ 377, 347, 52 },
		{ 377, 348, 52 },
		{ 377, 349, 52 },
		{ 377, 350, 52 },
		{ 377, 351, 52 },
		{ 377, 352, 52 },
		{ 377, 353, 52 },
		{ 377, 354, 52 },
		{ 377, 355, 52 },
		{ 377, 356, 52 },
		{ 377, 357, 52 },
		{ 377, 358, 52 },
		{ 377, 359, 52 },
		{ 377, 360, 52 },
		{ 377, 361, 52 },
		{ 378, 333, 52 },
		{ 378, 334, 52 },
		{ 378, 335, 52 },
		{ 378, 336, 52 },
		{ 378, 337, 52 },
		{ 378, 338, 52 },
		{ 378, 339, 52 },
		{ 378, 340, 52 },
		{ 378, 341, 52 },
		{ 378, 342, 52 },
		{ 378, 343, 52 },
		{ 378, 344, 52 },
		{ 378, 345, 52 },
		{ 378, 346, 52 },
		{ 378, 347, 52 },
		{ 378, 348, 52 },
		{ 378, 349, 52 },
		{ 378, 350, 52 },
		{ 378, 351, 52 },
		{ 378, 352, 52 },
		{ 378, 353, 52 },
		{ 378, 354, 52 },
		{ 378, 355, 52 },
		{ 378, 356, 52 },
		{ 378, 357, 52 },
		{ 378, 358, 52 },
		{ 378, 359, 52 },
		{ 378, 360, 52 },
		{ 378, 361, 52 },
		{ 378, 362, 52 },
		{ 379, 334, 52 },
		{ 379, 335, 52 },
		{ 379, 336, 52 },
		{ 379, 337, 52 },
		{ 379, 338, 52 },
		{ 379, 339, 52 },
		{ 379, 340, 52 },
		{ 379, 341, 52 },
		{ 379, 342, 52 },
		{ 379, 343, 52 },
		{ 379, 344, 52 },
		{ 379, 345, 52 },
		{ 379, 346, 52 },
		{ 379, 347, 52 },
		{ 379, 348, 52 },
		{ 379, 349, 52 },
		{ 379, 350, 52 },
		{ 379, 351, 52 },
		{ 379, 352, 52 },
		{ 379, 353, 52 },
		{ 379, 354, 52 },
		{ 379, 355, 52 },
		{ 379, 356, 52 },
		{ 379, 357, 52 },
		{ 379, 358, 52 },
		{ 379, 359, 52 },
		{ 379, 360, 52 },
		{ 379, 361, 52 },
		{ 379, 362, 52 },
		{ 379, 363, 52 },
		{ 380, 334, 52 },
		{ 380, 335, 52 },
		{ 380, 336, 52 },
		{ 380, 337, 52 },
		{ 380, 338, 52 },
		{ 380, 339, 52 },
		{ 380, 340, 52 },
		{ 380, 341, 52 },
		{ 380, 342, 52 },
		{ 380, 343, 52 },
		{ 380, 344, 52 },
		{ 380, 345, 52 },
		{ 380, 346, 52 },
		{ 380, 347, 52 },
		{ 380, 348, 52 },
		{ 380, 349, 52 },
		{ 380, 350, 52 },
		{ 380, 351, 52 },
		{ 380, 352, 52 },
		{ 380, 353, 52 },
		{ 380, 354, 52 },
		{ 380, 355, 52 },
		{ 380, 356, 52 },
		{ 380, 357, 52 },
		{ 380, 358, 52 },
		{ 380, 359, 52 },
		{ 380, 360, 52 },
		{ 380, 361, 52 },
		{ 380, 362, 52 },
		{ 380, 363, 52 },
		{ 380, 364, 52 },
		{ 381, 335, 52 },
		{ 381, 336, 52 },
		{ 381, 337, 52 },
		{ 381, 338, 52 },
		{ 381, 339, 52 },
		{ 381, 340, 52 },
		{ 381, 341, 52 },
		{ 381, 342, 52 },
		{ 381, 343, 52 },
		{ 381, 344, 52 },
		{ 381, 345, 52 },
		{ 381, 346, 52 },
		{ 381, 347, 52 },
		{ 381, 348, 52 },
		{ 381, 349, 52 },
		{ 381, 350, 52 },
		{ 381, 351, 52 },
		{ 381, 352, 52 },
		{ 381, 353, 52 },
		{ 381, 354, 52 },
		{ 381, 355, 52 },
		{ 381, 356, 52 },
		{ 381, 357, 52 },
		{ 381, 358, 52 },
		{ 381, 359, 52 },
		{ 381, 360, 52 },
		{ 381, 361, 52 },
		{ 381, 362, 52 },
		{ 381, 363, 52 },
		{ 381, 364, 52 },
		{ 381, 365, 52 },
		{ 382, 336, 52 },
		{ 382, 337, 52 },
		{ 382, 338, 52 },
		{ 382, 339, 52 },
		{ 382, 340, 52 },
		{ 382, 341, 52 },
		{ 382, 342, 52 },
		{ 382, 343, 52 },
		{ 382, 344, 52 },
		{ 382, 345, 52 },
		{ 382, 346, 52 },
		{ 382, 347, 52 },
		{ 382, 348, 52 },
		{ 382, 349, 52 },
		{ 382, 350, 52 },
		{ 382, 351, 52 },
		{ 382, 352, 52 },
		{ 382, 353, 52 },
		{ 382, 354, 52 },
		{ 382, 355, 52 },
		{ 382, 356, 52 },
		{ 382, 357, 52 },
		{ 382, 358, 52 },
		{ 382, 359, 52 },
		{ 382, 360, 52 },
		{ 382, 361, 52 },
		{ 382, 362, 52 },
		{ 382, 363, 52 },
		{ 382, 364, 52 },
		{ 382, 365, 52 },
		{ 382, 366, 52 },
		{ 383, 336, 52 },
		{ 383, 337, 52 },
		{ 383, 338, 52 },
		{ 383, 339, 52 },
		{ 383, 340, 52 },
		{ 383, 341, 52 },
		{ 383, 342, 52 },
		{ 383, 343, 52 },
		{ 383, 344, 52 },
		{ 383, 345, 52 },
		{ 383, 346, 52 },
		{ 383, 347, 52 },
		{ 383, 348, 52 },
		{ 383, 349, 52 },
		{ 383, 350, 52 },
		{ 383, 351, 52 },
		{ 383, 352, 52 },
		{ 383, 353, 52 },
		{ 383, 354, 52 },
		{ 383, 355, 52 },
		{ 383, 356, 52 },
		{ 383, 357, 52 },
		{ 383, 358, 52 },
		{ 383, 359, 52 },
		{ 383, 360, 52 },
		{ 383, 361, 52 },
		{ 383, 362, 52 },
		{ 383, 363, 52 },
		{ 383, 364, 52 },
		{ 383, 365, 52 },
		{ 383, 366, 52 },
		{ 383, 367, 52 },
		{ 384, 337, 52 },
		{ 384, 338, 52 },
		{ 384, 339, 52 },
		{ 384, 340, 52 },
		{ 384, 341, 52 },
		{ 384, 342, 52 },
		{ 384, 343, 52 },
		{ 384, 344, 52 },
		{ 384, 345, 52 },
		{ 384, 346, 52 },
		{ 384, 347, 52 },
		{ 384, 348, 52 },
		{ 384, 349, 52 },
		{ 384, 350, 52 },
		{ 384, 351, 52 },
		{ 384, 352, 52 },
		{ 384, 353, 52 },
		{ 384, 354, 52 },
		{ 384, 355, 52 },
		{ 384, 356, 52 },
		{ 384, 357, 52 },
		{ 384, 358, 52 },
		{ 384, 359, 52 },
		{ 384, 360, 52 },
		{ 384, 361, 52 },
		{ 384, 362, 52 },
		{ 384, 363, 52 },
		{ 384, 364, 52 },
		{ 384, 365, 52 },
		{ 384, 366, 52 },
		{ 384, 367, 52 },
		{ 384, 368, 52 },
		{ 385, 338, 52 },
		{ 385, 339, 52 },
		{ 385, 340, 52 },
		{ 385, 341, 52 },
		{ 385, 342, 52 },
		{ 385, 343, 52 },
		{ 385, 344, 52 },
		{ 385, 345, 52 },
		{ 385, 346, 52 },
		{ 385, 347, 52 },
		{ 385, 348, 52 },
		{ 385, 349, 52 },
		{ 385, 350, 52 },
		{ 385, 351, 52 },
		{ 385, 352, 52 },
		{ 385, 353, 52 },
		{ 385, 354, 52 },
		{ 385, 355, 52 },
		{ 385, 356, 52 },
		{ 385, 357, 52 },
		{ 385, 358, 52 },
		{ 385, 359, 52 },
		{ 385, 360, 52 },
		{ 385, 361, 52 },
		{ 385, 362, 52 },
		{ 385, 363, 52 },
		{ 385, 364, 52 },
		{ 385, 365, 52 },
		{ 385, 366, 52 },
		{ 385, 367, 52 },
		{ 385, 368, 52 },
		{ 386, 338, 52 },
		{ 386, 339, 52 },
		{ 386, 340, 52 },
		{ 386, 341, 52 },
		{ 386, 342, 52 },
		{ 386, 343, 52 },
		{ 386, 344, 52 },
		{ 386, 345, 52 },
		{ 386, 346, 52 },
		{ 386, 347, 52 },
		{ 386, 348, 52 },
		{ 386, 349, 52 },
		{ 386, 350, 52 },
		{ 386, 351, 52 },
		{ 386, 352, 52 },
		{ 386, 353, 52 },
		{ 386, 354, 52 },
		{ 386, 355, 52 },
		{ 386, 356, 52 },
		{ 386, 357, 52 },
		{ 386, 358, 52 },
		{ 386, 359, 52 },
		{ 386, 360, 52 },
		{ 386, 361, 52 },
		{ 386, 362, 52 },
		{ 386, 363, 52 },
		{ 386, 364, 52 },
		{ 386, 365, 52 },
		{ 386, 366, 52 },
		{ 386, 367, 52 },
		{ 386, 368, 52 },
		{ 386, 369, 52 },
		{ 387, 339, 52 },
		{ 387, 340, 52 },
		{ 387, 341, 52 },
		{ 387, 342, 52 },
		{ 387, 343, 52 },
		{ 387, 344, 52 },
		{ 387, 345, 52 },
		{ 387, 346, 52 },
		{ 387, 347, 52 },
		{ 387, 348, 52 },
		{ 387, 349, 52 },
		{ 387, 350, 52 },
		{ 387, 351, 52 },
		{ 387, 352, 52 },
		{ 387, 353, 52 },
		{ 387, 354, 52 },
		{ 387, 355, 52 },
		{ 387, 356, 52 },
		{ 387, 357, 52 },
		{ 387, 358, 52 },
		{ 387, 359, 52 },
		{ 387, 360, 52 },
		{ 387, 361, 52 },
		{ 387, 362, 52 },
		{ 387, 363, 52 },
		{ 387, 364, 52 },
		{ 387, 365, 52 },
		{ 387, 366, 52 },
		{ 387, 367, 52 },
		{ 387, 368, 52 },
		{ 387, 369, 52 },
		{ 387, 370, 52 },
		{ 388, 339, 52 },
		{ 388, 340, 52 },
		{ 388, 341, 52 },
		{ 388, 342, 52 },
		{ 388, 343, 52 },
		{ 388, 344, 52 },
		{ 388, 345, 52 },
		{ 388, 346, 52 },
		{ 388, 347, 52 },
		{ 388, 348, 52 },
		{ 388, 349, 52 },
		{ 388, 350, 52 },
		{ 388, 351, 52 },
		{ 388, 352, 52 },
		{ 388, 353, 52 },
		{ 388, 354, 52 },
		{ 388, 355, 52 },
		{ 388, 356, 52 },
		{ 388, 357, 52 },
		{ 388, 358, 52 },
		{ 388, 359, 52 },
		{ 388, 360, 52 },
		{ 388, 361, 52 },
		{ 388, 362, 52 },
		{ 388, 363, 52 },
		{ 388, 364, 52 },
		{ 388, 365, 52 },
		{ 388, 366, 52 },
		{ 388, 367, 52 },
		{ 388, 368, 52 },
		{ 388, 369, 52 },
		{ 388, 370, 52 },
		{ 388, 371, 52 },
		{ 389, 340, 52 },
		{ 389, 341, 52 },
		{ 389, 342, 52 },
		{ 389, 343, 52 },
		{ 389, 344, 52 },
		{ 389, 345, 52 },
		{ 389, 346, 52 },
		{ 389, 347, 52 },
		{ 389, 348, 52 },
		{ 389, 349, 52 },
		{ 389, 350, 52 },
		{ 389, 351, 52 },
		{ 389, 352, 52 },
		{ 389, 353, 52 },
		{ 389, 354, 52 },
		{ 389, 355, 52 },
		{ 389, 356, 52 },
		{ 389, 357, 52 },
		{ 389, 358, 52 },
		{ 389, 359, 52 },
		{ 389, 360, 52 },
		{ 389, 361, 52 },
		{ 389, 362, 52 },
		{ 389, 363, 52 },
		{ 389, 364, 52 },
		{ 389, 365, 52 },
		{ 389, 366, 52 },
		{ 389, 367, 52 },
		{ 389, 368, 52 },
		{ 389, 369, 52 },
		{ 389, 370, 52 },
		{ 389, 371, 52 },
		{ 389, 372, 52 },
		{ 390, 341, 52 },
		{ 390, 342, 52 },
		{ 390, 343, 52 },
		{ 390, 344, 52 },
		{ 390, 345, 52 },
		{ 390, 346, 52 },
		{ 390, 347, 52 },
		{ 390, 348, 52 },
		{ 390, 349, 52 },
		{ 390, 350, 52 },
		{ 390, 351, 52 },
		{ 390, 352, 52 },
		{ 390, 353, 52 },
		{ 390, 354, 52 },
		{ 390, 355, 52 },
		{ 390, 356, 52 },
		{ 390, 357, 52 },
		{ 390, 358, 52 },
		{ 390, 359, 52 },
		{ 390, 360, 52 },
		{ 390, 361, 52 },
		{ 390, 362, 52 },
		{ 390, 363, 52 },
		{ 390, 364, 52 },
		{ 390, 365, 52 },
		{ 390, 366, 52 },
		{ 390, 367, 52 },
		{ 390, 368, 52 },
		{ 390, 369, 52 },
		{ 390, 370, 52 },
		{ 390, 371, 52 },
		{ 390, 372, 52 },
		{ 390, 373, 52 },
		{ 391, 341, 52 },
		{ 391, 342, 52 },
		{ 391, 343, 52 },
		{ 391, 344, 52 },
		{ 391, 345, 52 },
		{ 391, 346, 52 },
		{ 391, 347, 52 },
		{ 391, 348, 52 },
		{ 391, 349, 52 },
		{ 391, 350, 52 },
		{ 391, 351, 52 },
		{ 391, 352, 52 },
		{ 391, 353, 52 },
		{ 391, 354, 52 },
		{ 391, 355, 52 },
		{ 391, 356, 52 },
		{ 391, 357, 52 },
		{ 391, 358, 52 },
		{ 391, 359, 52 },
		{ 391, 360, 52 },
		{ 391, 361, 52 },
		{ 391, 362, 52 },
		{ 391, 363, 52 },
		{ 391, 364, 52 },
		{ 391, 365, 52 },
		{ 391, 366, 52 },
		{ 391, 367, 52 },
		{ 391, 368, 52 },
		{ 391, 369, 52 },
		{ 391, 370, 52 },
		{ 391, 371, 52 },
		{ 391, 372, 52 },
		{ 391, 373, 52 },
		{ 391, 374, 52 },
		{ 392, 342, 52 },
		{ 392, 343, 52 },
		{ 392, 344, 52 },
		{ 392, 345, 52 },
		{ 392, 346, 52 },
		{ 392, 347, 52 },
		{ 392, 348, 52 },
		{ 392, 349, 52 },
		{ 392, 350, 52 },
		{ 392, 351, 52 },
		{ 392, 352, 52 },
		{ 392, 353, 52 },
		{ 392, 354, 52 },
		{ 392, 355, 52 },
		{ 392, 356, 52 },
		{ 392, 357, 52 },
		{ 392, 358, 52 },
		{ 392, 359, 52 },
		{ 392, 360, 52 },
		{ 392, 361, 52 },
		{ 392, 362, 52 },
		{ 392, 363, 52 },
		{ 392, 364, 52 },
		{ 392, 365, 52 },
		{ 392, 366, 52 },
		{ 392, 367, 52 },
		{ 392, 368, 52 },
		{ 392, 369, 52 },
		{ 392, 370, 52 },
		{ 392, 371, 52 },
		{ 392, 372, 52 },
		{ 392, 373, 52 },
		{ 392, 374, 52 },
		{ 392, 375, 52 },
		{ 393, 343, 52 },
		{ 393, 344, 52 },
		{ 393, 345, 52 },
		{ 393, 346, 52 },
		{ 393, 347, 52 },
		{ 393, 348, 52 },
		{ 393, 349, 52 },
		{ 393, 350, 52 },
		{ 393, 351, 52 },
		{ 393, 352, 52 },
		{ 393, 353, 52 },
		{ 393, 354, 52 },
		{ 393, 355, 52 },
		{ 393, 356, 52 },
		{ 393, 357, 52 },
		{ 393, 358, 52 },
		{ 393, 359, 52 },
		{ 393, 360, 52 },
		{ 393, 361, 52 },
		{ 393, 362, 52 },
		{ 393, 363, 52 },
		{ 393, 364, 52 },
		{ 393, 365, 52 },
		{ 393, 366, 52 },
		{ 393, 367, 52 },
		{ 393, 368, 52 },
		{ 393, 369, 52 },
		{ 393, 370, 52 },
		{ 393, 371, 52 },
		{ 393, 372, 52 },
		{ 393, 373, 52 },
		{ 393, 374, 52 },
		{ 393, 375, 52 },
		{ 394, 343, 52 },
		{ 394, 344, 52 },
		{ 394, 345, 52 },
		{ 394, 346, 52 },
		{ 394, 347, 52 },
		{ 394, 348, 52 },
		{ 394, 349, 52 },
		{ 394, 350, 52 },
		{ 394, 351, 52 },
		{ 394, 352, 52 },
		{ 394, 353, 52 },
		{ 394, 354, 52 },
		{ 394, 355, 52 },
		{ 394, 356, 52 },
		{ 394, 357, 52 },
		{ 394, 358, 52 },
		{ 394, 359, 52 },
		{ 394, 360, 52 },
		{ 394, 361, 52 },
		{ 394, 362, 52 },
		{ 394, 363, 52 },
		{ 394, 364, 52 },
		{ 394, 365, 52 },
		{ 394, 366, 52 },
		{ 394, 367, 52 },
		{ 394, 368, 52 },
		{ 394, 369, 52 },
		{ 394, 370, 52 },
		{ 394, 371, 52 },
		{ 394, 372, 52 },
		{ 394, 373, 52 },
		{ 394, 374, 52 },
		{ 394, 375, 52 },
		{ 394, 376, 52 },
		{ 395, 344, 52 },
		{ 395, 345, 52 },
		{ 395, 346, 52 },
		{ 395, 347, 52 },
		{ 395, 348, 52 },
		{ 395, 349, 52 },
		{ 395, 350, 52 },
		{ 395, 351, 52 },
		{ 395, 352, 52 },
		{ 395, 353, 52 },
		{ 395, 354, 52 },
		{ 395, 355, 52 },
		{ 395, 356, 52 },
		{ 395, 357, 52 },
		{ 395, 358, 52 },
		{ 395, 359, 52 },
		{ 395, 360, 52 },
		{ 395, 361, 52 },
		{ 395, 362, 52 },
		{ 395, 363, 52 },
		{ 395, 364, 52 },
		{ 395, 365, 52 },
		{ 395, 366, 52 },
		{ 395, 367, 52 },
		{ 395, 368, 52 },
		{ 395, 369, 52 },
		{ 395, 370, 52 },
		{ 395, 371, 52 },
		{ 395, 372, 52 },
		{ 395, 373, 52 },
		{ 395, 374, 52 },
		{ 395, 375, 52 },
		{ 395, 376, 52 },
		{ 395, 377, 52 },
		{ 396, 344, 52 },
		{ 396, 345, 52 },
		{ 396, 346, 52 },
		{ 396, 347, 52 },
		{ 396, 348, 52 },
		{ 396, 349, 52 },
		{ 396, 350, 52 },
		{ 396, 351, 52 },
		{ 396, 352, 52 },
		{ 396, 353, 52 },
		{ 396, 354, 52 },
		{ 396, 355, 52 },
		{ 396, 356, 52 },
		{ 396, 357, 52 },
		{ 396, 358, 52 },
		{ 396, 359, 52 },
		{ 396, 360, 52 },
		{ 396, 361, 52 },
		{ 396, 362, 52 },
		{ 396, 363, 52 },
		{ 396, 364, 52 },
		{ 396, 365, 52 },
		{ 396, 366, 52 },
		{ 396, 367, 52 },
		{ 396, 368, 52 },
		{ 396, 369, 52 },
		{ 396, 370, 52 },
		{ 396, 371, 52 },
		{ 396, 372, 52 },
		{ 396, 373, 52 },
		{ 396, 374, 52 },
		{ 396, 375, 52 },
		{ 396, 376, 52 },
		{ 396, 377, 52 },
		{ 396, 378, 52 },
		{ 397, 345, 52 },
		{ 397, 346, 52 },
		{ 397, 347, 52 },
		{ 397, 348, 52 },
		{ 397, 349, 52 },
		{ 397, 350, 52 },
		{ 397, 351, 52 },
		{ 397, 352, 52 },
		{ 397, 353, 52 },
		{ 397, 354, 52 },
		{ 397, 355, 52 },
		{ 397, 356, 52 },
		{ 397, 357, 52 },
		{ 397, 358, 52 },
		{ 397, 359, 52 },
		{ 397, 360, 52 },
		{ 397, 361, 52 },
		{ 397, 362, 52 },
		{ 397, 363, 52 },
		{ 397, 364, 52 },
		{ 397, 365, 52 },
		{ 397, 366, 52 },
		{ 397, 367, 52 },
		{ 397, 368, 52 },
		{ 397, 369, 52 },
		{ 397, 370, 52 },
		{ 397, 371, 52 },
		{ 397, 372, 52 },
		{ 397, 373, 52 },
		{ 397, 374, 52 },
		{ 397, 375, 52 },
		{ 397, 376, 52 },
		{ 397, 377, 52 },
		{ 397, 378, 52 },
		{ 397, 379, 52 },
		{ 398, 346, 52 },
		{ 398, 347, 52 },
		{ 398, 348, 52 },
		{ 398, 349, 52 },
		{ 398, 350, 52 },
		{ 398, 351, 52 },
		{ 398, 352, 52 },
		{ 398, 353, 52 },
		{ 398, 354, 52 },
		{ 398, 355, 52 },
		{ 398, 356, 52 },
		{ 398, 357, 52 },
		{ 398, 358, 52 },
		{ 398, 359, 52 },
		{ 398, 360, 52 },
		{ 398, 361, 52 },
		{ 398, 362, 52 },
		{ 398, 363, 52 },
		{ 398, 364, 52 },
		{ 398, 365, 52 },
		{ 398, 366, 52 },
		{ 398, 367, 52 },
		{ 398, 368, 52 },
		{ 398, 369, 52 },
		{ 398, 370, 52 },
		{ 398, 371, 52 },
		{ 398, 372, 52 },
		{ 398, 373, 52 },
		{ 398, 374, 52 },
		{ 398, 375, 52 },
		{ 398, 376, 52 },
		{ 398, 377, 52 },
		{ 398, 378, 52 },
		{ 398, 379, 52 },
		{ 398, 380, 52 },
		{ 399, 346, 52 },
		{ 399, 347, 52 },
		{ 399, 348, 52 },
		{ 399, 349, 52 },
		{ 399, 350, 52 },
		{ 399, 351, 52 },
		{ 399, 352, 52 },
		{ 399, 353, 52 },
		{ 399, 354, 52 },
		{ 399, 355, 52 },
		{ 399, 356, 52 },
		{ 399, 357, 52 },
		{ 399, 358, 52 },
		{ 399, 359, 52 },
		{ 399, 360, 52 },
		{ 399, 361, 52 },
		{ 399, 362, 52 },
		{ 399, 363, 52 },
		{ 399, 364, 52 },
		{ 399, 365, 52 },
		{ 399, 366, 52 },
		{ 399, 367, 52 },
		{ 399, 368, 52 },
		{ 399, 369, 52 },
		{ 399, 370, 52 },
		{ 399, 371, 52 },
		{ 399, 372, 52 },
		{ 399, 373, 52 },
		{ 399, 374, 52 },
		{ 399, 375, 52 },
		{ 399, 376, 52 },
		{ 399, 377, 52 },
		{ 399, 378, 52 },
		{ 399, 379, 52 },
		{ 399, 380, 52 },
		{ 399, 381, 52 },
		{ 400, 347, 52 },
		{ 400, 348, 52 },
		{ 400, 349, 52 },
		{ 400, 350, 52 },
		{ 400, 351, 52 },
		{ 400, 352, 52 },
		{ 400, 353, 52 },
		{ 400, 354, 52 },
		{ 400, 355, 52 },
		{ 400, 356, 52 },
		{ 400, 357, 52 },
		{ 400, 358, 52 },
		{ 400, 359, 52 },
		{ 400, 360, 52 },
		{ 400, 361, 52 },
		{ 400, 362, 52 },
		{ 400, 363, 52 },
		{ 400, 364, 52 },
		{ 400, 365, 52 },
		{ 400, 366, 52 },
		{ 400, 367, 52 },
		{ 400, 368, 52 },
		{ 400, 369, 52 },
		{ 400, 370, 52 },
		{ 400, 371, 52 },
		{ 400, 372, 52 },
		{ 400, 373, 52 },
		{ 400, 374, 52 },
		{ 400, 375, 52 },
		{ 400, 376, 52 },
		{ 400, 377, 52 },
		{ 400, 378, 52 },
		{ 400, 379, 52 },
		{ 400, 380, 52 },
		{ 400, 381, 52 },
		{ 400, 382, 52 },
		{ 401, 348, 52 },
		{ 401, 349, 52 },
		{ 401, 350, 52 },
		{ 401, 351, 52 },
		{ 401, 352, 52 },
		{ 401, 353, 52 },
		{ 401, 354, 52 },
		{ 401, 355, 52 },
		{ 401, 356, 52 },
		{ 401, 357, 52 },
		{ 401, 358, 52 },
		{ 401, 359, 52 },
		{ 401, 360, 52 },
		{ 401, 361, 52 },
		{ 401, 362, 52 },
		{ 401, 363, 52 },
		{ 401, 364, 52 },
		{ 401, 365, 52 },
		{ 401, 366, 52 },
		{ 401, 367, 52 },
		{ 401, 368, 52 },
		{ 401, 369, 52 },
		{ 401, 370, 52 },
		{ 401, 371, 52 },
		{ 401, 372, 52 },
		{ 401, 373, 52 },
		{ 401, 374, 52 },
		{ 401, 375, 52 },
		{ 401, 376, 52 },
		{ 401, 377, 52 },
		{ 401, 378, 52 },
		{ 401, 379, 52 },
		{ 401, 380, 52 },
		{ 401, 381, 52 },
		{ 401, 382, 52 },
		{ 402, 348, 52 },
		{ 402, 349, 52 },
		{ 402, 350, 52 },
		{ 402, 351, 52 },
		{ 402, 352, 52 },
		{ 402, 353, 52 },
		{ 402, 354, 52 },
		{ 402, 355, 52 },
		{ 402, 356, 52 },
		{ 402, 357, 52 },
		{ 402, 358, 52 },
		{ 402, 359, 52 },
		{ 402, 360, 52 },
		{ 402, 361, 52 },
		{ 402, 362, 52 },
		{ 402, 363, 52 },
		{ 402, 364, 52 },
		{ 402, 365, 52 },
		{ 402, 366, 52 },
		{ 402, 367, 52 },
		{ 402, 368, 52 },
		{ 402, 369, 52 },
		{ 402, 370, 52 },
		{ 402, 371, 52 },
		{ 402, 372, 52 },
		{ 402, 373, 52 },
		{ 402, 374, 52 },
		{ 402, 375, 52 },
		{ 402, 376, 52 },
		{ 402, 377, 52 },
		{ 402, 378, 52 },
		{ 402, 379, 52 },
		{ 402, 380, 52 },
		{ 402, 381, 52 },
		{ 402, 382, 52 },
		{ 402, 383, 52 },
		{ 403, 349, 52 },
		{ 403, 350, 52 },
		{ 403, 351, 52 },
		{ 403, 352, 52 },
		{ 403, 353, 52 },
		{ 403, 354, 52 },
		{ 403, 355, 52 },
		{ 403, 356, 52 },
		{ 403, 357, 52 },
		{ 403, 358, 52 },
		{ 403, 359, 52 },
		{ 403, 360, 52 },
		{ 403, 361, 52 },
		{ 403, 362, 52 },
		{ 403, 363, 52 },
		{ 403, 364, 52 },
		{ 403, 365, 52 },
		{ 403, 366, 52 },
		{ 403, 367, 52 },
		{ 403, 368, 52 },
		{ 403, 369, 52 },
		{ 403, 370, 52 },
		{ 403, 371, 52 },
		{ 403, 372, 52 },
		{ 403, 373, 52 },
		{ 403, 374, 52 },
		{ 403, 375, 52 },
		{ 403, 376, 52 },
		{ 403, 377, 52 },
		{ 403, 378, 52 },
		{ 403, 379, 52 },
		{ 403, 380, 52 },
		{ 403, 381, 52 },
		{ 403, 382, 52 },
		{ 403, 383, 52 },
		{ 403, 384, 52 },
		{ 404, 349, 52 },
		{ 404, 350, 52 },
		{ 404, 351, 52 },
		{ 404, 352, 52 },
		{ 404, 353, 52 },
		{ 404, 354, 52 },
		{ 404, 355, 52 },
		{ 404, 356, 52 },
		{ 404, 357, 52 },
		{ 404, 358, 52 },
		{ 404, 359, 52 },
		{ 404, 360, 52 },
		{ 404, 361, 52 },
		{ 404, 362, 52 },
		{ 404, 363, 52 },
		{ 404, 364, 52 },
		{ 404, 365, 52 },
		{ 404, 366, 52 },
		{ 404, 367, 52 },
		{ 404, 368, 52 },
		{ 404, 369, 52 },
		{ 404, 370, 52 },
		{ 404, 371, 52 },
		{ 404, 372, 52 },
		{ 404, 373, 52 },
		{ 404, 374, 52 },
		{ 404, 375, 52 },
		{ 404, 376, 52 },
		{ 404, 377, 52 },
		{ 404, 378, 52 },
		{ 404, 379, 52 },
		{ 404, 380, 52 },
		{ 404, 381, 52 },
		{ 404, 382, 52 },
		{ 404, 383, 52 },
		{ 404, 384, 52 },
		{ 404, 385, 52 },
		{ 405, 350, 52 },
		{ 405, 351, 52 },
		{ 405, 352, 52 },
		{ 405, 353, 52 },
		{ 405, 354, 52 },
		{ 405, 355, 52 },
		{ 405, 356, 52 },
		{ 405, 357, 52 },
		{ 405, 358, 52 },
		{ 405, 359, 52 },
		{ 405, 360, 52 },
		{ 405, 361, 52 },
		{ 405, 362, 52 },
		{ 405, 363, 52 },
		{ 405, 364, 52 },
		{ 405, 365, 52 },
		{ 405, 366, 52 },
		{ 405, 367, 52 },
		{ 405, 368, 52 },
		{ 405, 369, 52 },
		{ 405, 370, 52 },
		{ 405, 371, 52 },
		{ 405, 372, 52 },
		{ 405, 373, 52 },
		{ 405, 374, 52 },
		{ 405, 375, 52 },
		{ 405, 376, 52 },
		{ 405, 377, 52 },
		{ 405, 378, 52 },
		{ 405, 379, 52 },
		{ 405, 380, 52 },
		{ 405, 381, 52 },
		{ 405, 382, 52 },
		{ 405, 383, 52 },
		{ 405, 384, 52 },
		{ 405, 385, 52 },
		{ 405, 386, 52 },
		{ 406, 351, 52 },
		{ 406, 352, 52 },
		{ 406, 353, 52 },
		{ 406, 354, 52 },
		{ 406, 355, 52 },
		{ 406, 356, 52 },
		{ 406, 357, 52 },
		{ 406, 358, 52 },
		{ 406, 359, 52 },
		{ 406, 360, 52 },
		{ 406, 361, 52 },
		{ 406, 362, 52 },
		{ 406, 363, 52 },
		{ 406, 364, 52 },
		{ 406, 365, 52 },
		{ 406, 366, 52 },
		{ 406, 367, 52 },
		{ 406, 368, 52 },
		{ 406, 369, 52 },
		{ 406, 370, 52 },
		{ 406, 371, 52 },
		{ 406, 372, 52 },
		{ 406, 373, 52 },
		{ 406, 374, 52 },
		{ 406, 375, 52 },
		{ 406, 376, 52 },
		{ 406, 377, 52 },
		{ 406, 378, 52 },
		{ 406, 379, 52 },
		{ 406, 380, 52 },
		{ 406, 381, 52 },
		{ 406, 382, 52 },
		{ 406, 383, 52 },
		{ 406, 384, 52 },
		{ 406, 385, 52 },
		{ 406, 386, 52 },
		{ 406, 387, 52 },
		{ 407, 351, 52 },
		{ 407, 352, 52 },
		{ 407, 353, 52 },
		{ 407, 354, 52 },
		{ 407, 355, 52 },
		{ 407, 356, 52 },
		{ 407, 357, 52 },
		{ 407, 358, 52 },
		{ 407, 359, 52 },
		{ 407, 360, 52 },
		{ 407, 361, 52 },
		{ 407, 362, 52 },
		{ 407, 363, 52 },
		{ 407, 364, 52 },
		{ 407, 365, 52 },
		{ 407, 366, 52 },
		{ 407, 367, 52 },
		{ 407, 368, 52 },
		{ 407, 369, 52 },
		{ 407, 370, 52 },
		{ 407, 371, 52 },
		{ 407, 372, 52 },
		{ 407, 373, 52 },
		{ 407, 374, 52 },
		{ 407, 375, 52 },
		{ 407, 376, 52 },
		{ 407, 377, 52 },
		{ 407, 378, 52 },
		{ 407, 379, 52 },
		{ 407, 380, 52 },
		{ 407, 381, 52 },
		{ 407, 382, 52 },
		{ 407, 383, 52 },
		{ 407, 384, 52 },
		{ 407, 385, 52 },
		{ 407, 386, 52 },
		{ 407, 387, 52 },
		{ 407, 388, 52 },
		{ 408, 352, 52 },
		{ 408, 353, 52 },
		{ 408, 354, 52 },
		{ 408, 355, 52 },
		{ 408, 356, 52 },
		{ 408, 357, 52 },
		{ 408, 358, 52 },
		{ 408, 359, 52 },
		{ 408, 360, 52 },
		{ 408, 361, 52 },
		{ 408, 362, 52 },
		{ 408, 363, 52 },
		{ 408, 364, 52 },
		{ 408, 365, 52 },
		{ 408, 366, 52 },
		{ 408, 367, 52 },
		{ 408, 368, 52 },
		{ 408, 369, 52 },
		{ 408, 370, 52 },
		{ 408, 371, 52 },
		{ 408, 372, 52 },
		{ 408, 373, 52 },
		{ 408, 374, 52 },
		{ 408, 375, 52 },
		{ 408, 376, 52 },
		{ 408, 377, 52 },
		{ 408, 378, 52 },
		{ 408, 379, 52 },
		{ 408, 380, 52 },
		{ 408, 381, 52 },
		{ 408, 382, 52 },
		{ 408, 383, 52 },
		{ 408, 384, 52 },
		{ 408, 385, 52 },
		{ 408, 386, 52 },
		{ 408, 387, 52 },
		{ 408, 388, 52 },
		{ 408, 389, 52 },
		{ 409, 353, 52 },
		{ 409, 354, 52 },
		{ 409, 355, 52 },
		{ 409, 356, 52 },
		{ 409, 357, 52 },
		{ 409, 358, 52 },
		{ 409, 359, 52 },
		{ 409, 360, 52 },
		{ 409, 361, 52 },
		{ 409, 362, 52 },
		{ 409, 363, 52 },
		{ 409, 364, 52 },
		{ 409, 365, 52 },
		{ 409, 366, 52 },
		{ 409, 367, 52 },
		{ 409, 368, 52 },
		{ 409, 369, 52 },
		{ 409, 370, 52 },
		{ 409, 371, 52 },
		{ 409, 372, 52 },
		{ 409, 373, 52 },
		{ 409, 374, 52 },
		{ 409, 375, 52 },
		{ 409, 376, 52 },
		{ 409, 377, 52 },
		{ 409, 378, 52 },
		{ 409, 379, 52 },
		{ 409, 380, 52 },
		{ 409, 381, 52 },
		{ 409, 382, 52 },
		{ 409, 383, 52 },
		{ 409, 384, 52 },
		{ 409, 385, 52 },
		{ 409, 386, 52 },
		{ 409, 387, 52 },
		{ 409, 388, 52 },
		{ 409, 389, 52 },
		{ 410, 353, 52 },
		{ 410, 354, 52 },
		{ 410, 355, 52 },
		{ 410, 356, 52 },
		{ 410, 357, 52 },
		{ 410, 358, 52 },
		{ 410, 359, 52 },
		{ 410, 360, 52 },
		{ 410, 361, 52 },
		{ 410, 362, 52 },
		{ 410, 363, 52 },
		{ 410, 364, 52 },
		{ 410, 365, 52 },
		{ 410, 366, 52 },
		{ 410, 367, 52 },
		{ 410, 368, 52 },
		{ 410, 369, 52 },
		{ 410, 370, 52 },
		{ 410, 371, 52 },
		{ 410, 372, 52 },
		{ 410, 373, 52 },
		{ 410, 374, 52 },
		{ 410, 375, 52 },
		{ 410, 376, 52 },
		{ 410, 377, 52 },
		{ 410, 378, 52 },
		{ 410, 379, 52 },
		{ 410, 380, 52 },
		{ 410, 381, 52 },
		{ 410, 382, 52 },
		{ 410, 383, 52 },
		{ 410, 384, 52 },
		{ 410, 385, 52 },
		{ 410, 386, 52 },
		{ 410, 387, 52 },
		{ 410, 388, 52 },
		{ 410, 389, 52 },
		{ 410, 390, 52 },
		{ 411, 354, 52 },
		{ 411, 355, 52 },
		{ 411, 356, 52 },
		{ 411, 357, 52 },
		{ 411, 358, 52 },
		{ 411, 359, 52 },
		{ 411, 360, 52 },
		{ 411, 361, 52 },
		{ 411, 362, 52 },
		{ 411, 363, 52 },
		{ 411, 364, 52 },
		{ 411, 365, 52 },
		{ 411, 366, 52 },
		{ 411, 367, 52 },
		{ 411, 368, 52 },
		{ 411, 369, 52 },
		{ 411, 370, 52 },
		{ 411, 371, 52 },
		{ 411, 372, 52 },
		{ 411, 373, 52 },
		{ 411, 374, 52 },
		{ 411, 375, 52 },
		{ 411, 376, 52 },
		{ 411, 377, 52 },
		{ 411, 378, 52 },
		{ 411, 379, 52 },
		{ 411, 380, 52 },
		{ 411, 381, 52 },
		{ 411, 382, 52 },
		{ 411, 383, 52 },
		{ 411, 384, 52 },
		{ 411, 385, 52 },
		{ 411, 386, 52 },
		{ 411, 387, 52 },
		{ 411, 388, 52 },
		{ 411, 389, 52 },
		{ 411, 390, 52 },
		{ 411, 391, 52 },
		{ 412, 354, 52 },
		{ 412, 355, 52 },
		{ 412, 356, 52 },
		{ 412, 357, 52 },
		{ 412, 358, 52 },
		{ 412, 359, 52 },
		{ 412, 360, 52 },
		{ 412, 361, 52 },
		{ 412, 362, 52 },
		{ 412, 363, 52 },
		{ 412, 364, 52 },
		{ 412, 365, 52 },
		{ 412, 366, 52 },
		{ 412, 367, 52 },
		{ 412, 368, 52 },
		{ 412, 369, 52 },
		{ 412, 370, 52 },
		{ 412, 371, 52 },
		{ 412, 372, 52 },
		{ 412, 373, 52 },
		{ 412, 374, 52 },
		{ 412, 375, 52 },
		{ 412, 376, 52 },
		{ 412, 377, 52 },
		{ 412, 378, 52 },
		{ 412, 379, 52 },
		{ 412, 380, 52 },
		{ 412, 381, 52 },
		{ 412, 382, 52 },
		{ 412, 383, 52 },
		{ 412, 384, 52 },
		{ 412, 385, 52 },
		{ 412, 386, 52 },
		{ 412, 387, 52 },
		{ 412, 388, 52 },
		{ 412, 389, 52 },
		{ 412, 390, 52 },
		{ 412, 391, 52 },
		{ 412, 392, 52 },
		{ 413, 355, 52 },
		{ 413, 356, 52 },
		{ 413, 357, 52 },
		{ 413, 358, 52 },
		{ 413, 359, 52 },
		{ 413, 360, 52 },
		{ 413, 361, 52 },
		{ 413, 362, 52 },
		{ 413, 363, 52 },
		{ 413, 364, 52 },
		{ 413, 365, 52 },
		{ 413, 366, 52 },
		{ 413, 367, 52 },
		{ 413, 368, 52 },
		{ 413, 369, 52 },
		{ 413, 370, 52 },
		{ 413, 371, 52 },
		{ 413, 372, 52 },
		{ 413, 373, 52 },
		{ 413, 374, 52 },
		{ 413, 375, 52 },
		{ 413, 376, 52 },
		{ 413, 377, 52 },
		{ 413, 378, 52 },
		{ 413, 379, 52 },
		{ 413, 380, 52 },
		{ 413, 381, 52 },
		{ 413, 382, 52 },
		{ 413, 383, 52 },
		{ 413, 384, 52 },
		{ 413, 385, 52 },
		{ 413, 386, 52 },
		{ 413, 387, 52 },
		{ 413, 388, 52 },
		{ 413, 389, 52 },
		{ 413, 390, 52 },
		{ 413, 391, 52 },
		{ 413, 392, 52 },
		{ 413, 393, 52 },
		{ 414, 356, 52 },
		{ 414, 357, 52 },
		{ 414, 358, 52 },
		{ 414, 359, 52 },
		{ 414, 360, 52 },
		{ 414, 361, 52 },
		{ 414, 362, 52 },
		{ 414, 363, 52 },
		{ 414, 364, 52 },
		{ 414, 365, 52 },
		{ 414, 366, 52 },
		{ 414, 367, 52 },
		{ 414, 368, 52 },
		{ 414, 369, 52 },
		{ 414, 370, 52 },
		{ 414, 371, 52 },
		{ 414, 372, 52 },
		{ 414, 373, 52 },
		{ 414, 374, 52 },
		{ 414, 375, 52 },
		{ 414, 376, 52 },
		{ 414, 377, 52 },
		{ 414, 378, 52 },
		{ 414, 379, 52 },
		{ 414, 380, 52 },
		{ 414, 381, 52 },
		{ 414, 382, 52 },
		{ 414, 383, 52 },
		{ 414, 384, 52 },
		{ 414, 385, 52 },
		{ 414, 386, 52 },
		{ 414, 387, 52 },
		{ 414, 388, 52 },
		{ 414, 389, 52 },
		{ 414, 390, 52 },
		{ 414, 391, 52 },
		{ 414, 392, 52 },
		{ 414, 393, 52 },
		{ 414, 394, 52 },
		{ 415, 356, 52 },
		{ 415, 357, 52 },
		{ 415, 358, 52 },
		{ 415, 359, 52 },
		{ 415, 360, 52 },
		{ 415, 361, 52 },
		{ 415, 362, 52 },
		{ 415, 363, 52 },
		{ 415, 364, 52 },
		{ 415, 365, 52 },
		{ 415, 366, 52 },
		{ 415, 367, 52 },
		{ 415, 368, 52 },
		{ 415, 369, 52 },
		{ 415, 370, 52 },
		{ 415, 371, 52 },
		{ 415, 372, 52 },
		{ 415, 373, 52 },
		{ 415, 374, 52 },
		{ 415, 375, 52 },
		{ 415, 376, 52 },
		{ 415, 377, 52 },
		{ 415, 378, 52 },
		{ 415, 379, 52 },
		{ 415, 380, 52 },
		{ 415, 381, 52 },
		{ 415, 382, 52 },
		{ 415, 383, 52 },
		{ 415, 384, 52 },
		{ 415, 385, 52 },
		{ 415, 386, 52 },
		{ 415, 387, 52 },
		{ 415, 388, 52 },
		{ 415, 389, 52 },
		{ 415, 390, 52 },
		{ 415, 391, 52 },
		{ 415, 392, 52 },
		{ 415, 393, 52 },
		{ 415, 394, 52 },
		{ 415, 395, 52 },
		{ 416, 357, 52 },
		{ 416, 358, 52 },
		{ 416, 359, 52 },
		{ 416, 360, 52 },
		{ 416, 361, 52 },
		{ 416, 362, 52 },
		{ 416, 363, 52 },
		{ 416, 364, 52 },
		{ 416, 365, 52 },
		{ 416, 366, 52 },
		{ 416, 367, 52 },
		{ 416, 368, 52 },
		{ 416, 369, 52 },
		{ 416, 370, 52 },
		{ 416, 371, 52 },
		{ 416, 372, 52 },
		{ 416, 373, 52 },
		{ 416, 374, 52 },
		{ 416, 375, 52 },
		{ 416, 376, 52 },
		{ 416, 377, 52 },
		{ 416, 378, 52 },
		{ 416, 379, 52 },
		{ 416, 380, 52 },
		{ 416, 381, 52 },
		{ 416, 382, 52 },
		{ 416, 383, 52 },
		{ 416, 384, 52 },
		{ 416, 385, 52 },
		{ 416, 386, 52 },
		{ 416, 387, 52 },
		{ 416, 388, 52 },
		{ 416, 389, 52 },
		{ 416, 390, 52 },
		{ 416, 391, 52 },
		{ 416, 392, 52 },
		{ 416, 393, 52 },
		{ 416, 394, 52 },
		{ 416, 395, 52 },
		{ 416, 396, 52 },
		{ 417, 358, 52 },
		{ 417, 359, 52 },
		{ 417, 360, 52 },
		{ 417, 361, 52 },
		{ 417, 362, 52 },
		{ 417, 363, 52 },
		{ 417, 364, 52 },
		{ 417, 365, 52 },
		{ 417, 366, 52 },
		{ 417, 367, 52 },
		{ 417, 368, 52 },
		{ 417, 369, 52 },
		{ 417, 370, 52 },
		{ 417, 371, 52 },
		{ 417, 372, 52 },
		{ 417, 373, 52 },
		{ 417, 374, 52 },
		{ 417, 375, 52 },
		{ 417, 376, 52 },
		{ 417, 377, 52 },
		{ 417, 378, 52 },
		{ 417, 379, 52 },
		{ 417, 380, 52 },
		{ 417, 381, 52 },
		{ 417, 382, 52 },
		{ 417, 383, 52 },
		{ 417, 384, 52 },
		{ 417, 385, 52 },
		{ 417, 386, 52 },
		{ 417, 387, 52 },
		{ 417, 388, 52 },
		{ 417, 389, 52 },
		{ 417, 390, 52 },
		{ 417, 391, 52 },
		{ 417, 392, 52 },
		{ 417, 393, 52 },
		{ 417, 394, 52 },
		{ 417, 395, 52 },
		{ 417, 396, 52 },
		{ 418, 358, 52 },
		{ 418, 359, 52 },
		{ 418, 360, 52 },
		{ 418, 361, 52 },
		{ 418, 362, 52 },
		{ 418, 363, 52 },
		{ 418, 364, 52 },
		{ 418, 365, 52 },
		{ 418, 366, 52 },
		{ 418, 367, 52 },
		{ 418, 368, 52 },
		{ 418, 369, 52 },
		{ 418, 370, 52 },
		{ 418, 371, 52 },
		{ 418, 372, 52 },
		{ 418, 373, 52 },
		{ 418, 374, 52 },
		{ 418, 375, 52 },
		{ 418, 376, 52 },
		{ 418, 377, 52 },
		{ 418, 378, 52 },
		{ 418, 379, 52 },
		{ 418, 380, 52 },
		{ 418, 381, 52 },
		{ 418, 382, 52 },
		{ 418, 383, 52 },
		{ 418, 384, 52 },
		{ 418, 385, 52 },
		{ 418, 386, 52 },
		{ 418, 387, 52 },
		{ 418, 388, 52 },
		{ 418, 389, 52 },
		{ 418, 390, 52 },
		{ 418, 391, 52 },
		{ 418, 392, 52 },
		{ 418, 393, 52 },
		{ 418, 394, 52 },
		{ 418, 395, 52 },
		{ 418, 396, 52 },
		{ 418, 397, 52 },
		{ 419, 359, 52 },
		{ 419, 360, 52 },
		{ 419, 361, 52 },
		{ 419, 362, 52 },
		{ 419, 363, 52 },
		{ 419, 364, 52 },
		{ 419, 365, 52 },
		{ 419, 366, 52 },
		{ 419, 367, 52 },
		{ 419, 368, 52 },
		{ 419, 369, 52 },
		{ 419, 370, 52 },
		{ 419, 371, 52 },
		{ 419, 372, 52 },
		{ 419, 373, 52 },
		{ 419, 374, 52 },
		{ 419, 375, 52 },
		{ 419, 376, 52 },
		{ 419, 377, 52 },
		{ 419, 378, 52 },
		{ 419, 379, 52 },
		{ 419, 380, 52 },
		{ 419, 381, 52 },
		{ 419, 382, 52 },
		{ 419, 383, 52 },
		{ 419, 384, 52 },
		{ 419, 385, 52 },
		{ 419, 386, 52 },
		{ 419, 387, 52 },
		{ 419, 388, 52 },
		{ 419, 389, 52 },
		{ 419, 390, 52 },
		{ 419, 391, 52 },
		{ 419, 392, 52 },
		{ 419, 393, 52 },
		{ 419, 394, 52 },
		{ 419, 395, 52 },
		{ 419, 396, 52 },
		{ 419, 397, 52 },
		{ 419, 398, 52 },
		{ 420, 359, 52 },
		{ 420, 360, 52 },
		{ 420, 361, 52 },
		{ 420, 362, 52 },
		{ 420, 363, 52 },
		{ 420, 364, 52 },
		{ 420, 365, 52 },
		{ 420, 366, 52 },
		{ 420, 367, 52 },
		{ 420, 368, 52 },
		{ 420, 369, 52 },
		{ 420, 370, 52 },
		{ 420, 371, 52 },
		{ 420, 372, 52 },
		{ 420, 373, 52 },
		{ 420, 374, 52 },
		{ 420, 375, 52 },
		{ 420, 376, 52 },
		{ 420, 377, 52 },
		{ 420, 378, 52 },
		{ 420, 379, 52 },
		{ 420, 380, 52 },
		{ 420, 381, 52 },
		{ 420, 382, 52 },
		{ 420, 383, 52 },
		{ 420, 384, 52 },
		{ 420, 385, 52 },
		{ 420, 386, 52 },
		{ 420, 387, 52 },
		{ 420, 388, 52 },
		{ 420, 389, 52 },
		{ 420, 390, 52 },
		{ 420, 391, 52 },
		{ 420, 392, 52 },
		{ 420, 393, 52 },
		{ 420, 394, 52 },
		{ 420, 395, 52 },
		{ 420, 396, 52 },
		{ 420, 397, 52 },
		{ 420, 398, 52 },
		{ 420, 399, 52 },
		{ 421, 360, 52 },
		{ 421, 361, 52 },
		{ 421, 362, 52 },
		{ 421, 363, 52 },
		{ 421, 364, 52 },
		{ 421, 365, 52 },
		{ 421, 366, 52 },
		{ 421, 367, 52 },
		{ 421, 368, 52 },
		{ 421, 369, 52 },
		{ 421, 370, 52 },
		{ 421, 371, 52 },
		{ 421, 372, 52 },
		{ 421, 373, 52 },
		{ 421, 374, 52 },
		{ 421, 375, 52 },
		{ 421, 376, 52 },
		{ 421, 377, 52 },
		{ 421, 378, 52 },
		{ 421, 379, 52 },
		{ 421, 380, 52 },
		{ 421, 381, 52 },
		{ 421, 382, 52 },
		{ 421, 383, 52 },
		{ 421, 384, 52 },
		{ 421, 385, 52 },
		{ 421, 386, 52 },
		{ 421, 387, 52 },
		{ 421, 388, 52 },
		{ 421, 389, 52 },
		{ 421, 390, 52 },
		{ 421, 391, 52 },
		{ 421, 392, 52 },
		{ 421, 393, 52 },
		{ 421, 394, 52 },
		{ 421, 395, 52 },
		{ 421, 396, 52 },
		{ 421, 397, 52 },
		{ 421, 398, 52 },
		{ 421, 399, 52 },
		{ 421, 400, 52 },
		{ 422, 361, 52 },
		{ 422, 362, 52 },
		{ 422, 363, 52 },
		{ 422, 364, 52 },
		{ 422, 365, 52 },
		{ 422, 366, 52 },
		{ 422, 367, 52 },
		{ 422, 368, 52 },
		{ 422, 369, 52 },
		{ 422, 370, 52 },
		{ 422, 371, 52 },
		{ 422, 372, 52 },
		{ 422, 373, 52 },
		{ 422, 374, 52 },
		{ 422, 375, 52 },
		{ 422, 376, 52 },
		{ 422, 377, 52 },
		{ 422, 378, 52 },
		{ 422, 379, 52 },
		{ 422, 380, 52 },
		{ 422, 381, 52 },
		{ 422, 382, 52 },
		{ 422, 383, 52 },
		{ 422, 384, 52 },
		{ 422, 385, 52 },
		{ 422, 386, 52 },
		{ 422, 387, 52 },
		{ 422, 388, 52 },
		{ 422, 389, 52 },
		{ 422, 390, 52 },
		{ 422, 391, 52 },
		{ 422, 392, 52 },
		{ 422, 393, 52 },
		{ 422, 394, 52 },
		{ 422, 395, 52 },
		{ 422, 396, 52 },
		{ 422, 397, 52 },
		{ 422, 398, 52 },
		{ 422, 399, 52 },
		{ 422, 400, 52 },
		{ 422, 401, 52 },
		{ 423, 361, 52 },
		{ 423, 362, 52 },
		{ 423, 363, 52 },
		{ 423, 364, 52 },
		{ 423, 365, 52 },
		{ 423, 366, 52 },
		{ 423, 367, 52 },
		{ 423, 368, 52 },
		{ 423, 369, 52 },
		{ 423, 370, 52 },
		{ 423, 371, 52 },
		{ 423, 372, 52 },
		{ 423, 373, 52 },
		{ 423, 374, 52 },
		{ 423, 375, 52 },
		{ 423, 376, 52 },
		{ 423, 377, 52 },
		{ 423, 378, 52 },
		{ 423, 379, 52 },
		{ 423, 380, 52 },
		{ 423, 381, 52 },
		{ 423, 382, 52 },
		{ 423, 383, 52 },
		{ 423, 384, 52 },
		{ 423, 385, 52 },
		{ 423, 386, 52 },
		{ 423, 387, 52 },
		{ 423, 388, 52 },
		{ 423, 389, 52 },
		{ 423, 390, 52 },
		{ 423, 391, 52 },
		{ 423, 392, 52 },
		{ 423, 393, 52 },
		{ 423, 394, 52 },
		{ 423, 395, 52 },
		{ 423, 396, 52 },
		{ 423, 397, 52 },
		{ 423, 398, 52 },
		{ 423, 399, 52 },
		{ 423, 400, 52 },
		{ 423, 401, 52 },
		{ 423, 402, 52 },
		{ 424, 362, 52 },
		{ 424, 363, 52 },
		{ 424, 364, 52 },
		{ 424, 365, 52 },
		{ 424, 366, 52 },
		{ 424, 367, 52 },
		{ 424, 368, 52 },
		{ 424, 369, 52 },
		{ 424, 370, 52 },
		{ 424, 371, 52 },
		{ 424, 372, 52 },
		{ 424, 373, 52 },
		{ 424, 374, 52 },
		{ 424, 375, 52 },
		{ 424, 376, 52 },
		{ 424, 377, 52 },
		{ 424, 378, 52 },
		{ 424, 379, 52 },
		{ 424, 380, 52 },
		{ 424, 381, 52 },
		{ 424, 382, 52 },
		{ 424, 383, 52 },
		{ 424, 384, 52 },
		{ 424, 385, 52 },
		{ 424, 386, 52 },
		{ 424, 387, 52 },
		{ 424, 388, 52 },
		{ 424, 389, 52 },
		{ 424, 390, 52 },
		{ 424, 391, 52 },
		{ 424, 392, 52 },
		{ 424, 393, 52 },
		{ 424, 394, 52 },
		{ 424, 395, 52 },
		{ 424, 396, 52 },
		{ 424, 397, 52 },
		{ 424, 398, 52 },
		{ 424, 399, 52 },
		{ 424, 400, 52 },
		{ 424, 401, 52 },
		{ 424, 402, 52 },
		{ 424, 403, 52 },
		{ 425, 363, 52 },
		{ 425, 364, 52 },
		{ 425, 365, 52 },
		{ 425, 366, 52 },
		{ 425, 367, 52 },
		{ 425, 368, 52 },
		{ 425, 369, 52 },
		{ 425, 370, 52 },
		{ 425, 371, 52 },
		{ 425, 372, 52 },
		{ 425, 373, 52 },
		{ 425, 374, 52 },
		{ 425, 375, 52 },
		{ 425, 376, 52 },
		{ 425, 377, 52 },
		{ 425, 378, 52 },
		{ 425, 379, 52 },
		{ 425, 380, 52 },
		{ 425, 381, 52 },
		{ 425, 382, 52 },
		{ 425, 383, 52 },
		{ 425, 384, 52 },
		{ 425, 385, 52 },
		{ 425, 386, 52 },
		{ 425, 387, 52 },
		{ 425, 388, 52 },
		{ 425, 389, 52 },
		{ 425, 390, 52 },
		{ 425, 391, 52 },
		{ 425, 392, 52 },
		{ 425, 393, 52 },
		{ 425, 394, 52 },
		{ 425, 395, 52 },
		{ 425, 396, 52 },
		{ 425, 397, 52 },
		{ 425, 398, 52 },
		{ 425, 399, 52 },
		{ 425, 400, 52 },
		{ 425, 401, 52 },
		{ 425, 402, 52 },
		{ 425, 403, 52 },
		{ 426, 363, 52 },
		{ 426, 364, 52 },
		{ 426, 365, 52 },
		{ 426, 366, 52 },
		{ 426, 367, 52 },
		{ 426, 368, 52 },
		{ 426, 369, 52 },
		{ 426, 370, 52 },
		{ 426, 371, 52 },
		{ 426, 372, 52 },
		{ 426, 373, 52 },
		{ 426, 374, 52 },
		{ 426, 375, 52 },
		{ 426, 376, 52 },
		{ 426, 377, 52 },
		{ 426, 378, 52 },
		{ 426, 379, 52 },
		{ 426, 380, 52 },
		{ 426, 381, 52 },
		{ 426, 382, 52 },
		{ 426, 383, 52 },
		{ 426, 384, 52 },
		{ 426, 385, 52 },
		{ 426, 386, 52 },
		{ 426, 387, 52 },
		{ 426, 388, 52 },
		{ 426, 389, 52 },
		{ 426, 390, 52 },
		{ 426, 391, 52 },
		{ 426, 392, 52 },
		{ 426, 393, 52 },
		{ 426, 394, 52 },
		{ 426, 395, 52 },
		{ 426, 396, 52 },
		{ 426, 397, 52 },
		{ 426, 398, 52 },
		{ 426, 399, 52 },
		{ 426, 400, 52 },
		{ 426, 401, 52 },
		{ 426, 402, 52 },
		{ 426, 403, 52 },
		{ 426, 404, 52 },
		{ 427, 364, 52 },
		{ 427, 365, 52 },
		{ 427, 366, 52 },
		{ 427, 367, 52 },
		{ 427, 368, 52 },
		{ 427, 369, 52 },
		{ 427, 370, 52 },
		{ 427, 371, 52 },
		{ 427, 372, 52 },
		{ 427, 373, 52 },
		{ 427, 374, 52 },
		{ 427, 375, 52 },
		{ 427, 376, 52 },
		{ 427, 377, 52 },
		{ 427, 378, 52 },
		{ 427, 379, 52 },
		{ 427, 380, 52 },
		{ 427, 381, 52 },
		{ 427, 382, 52 },
		{ 427, 383, 52 },
		{ 427, 384, 52 },
		{ 427, 385, 52 },
		{ 427, 386, 52 },
		{ 427, 387, 52 },
		{ 427, 388, 52 },
		{ 427, 389, 52 },
		{ 427, 390, 52 },
		{ 427, 391, 52 },
		{ 427, 392, 52 },
		{ 427, 393, 52 },
		{ 427, 394, 52 },
		{ 427, 395, 52 },
		{ 427, 396, 52 },
		{ 427, 397, 52 },
		{ 427, 398, 52 },
		{ 427, 399, 52 },
		{ 427, 400, 52 },
		{ 427, 401, 52 },
		{ 427, 402, 52 },
		{ 427, 403, 52 },
		{ 427, 404, 52 },
		{ 427, 405, 52 },
		{ 428, 364, 52 },
		{ 428, 365, 52 },
		{ 428, 366, 52 },
		{ 428, 367, 52 },
		{ 428, 368, 52 },
		{ 428, 369, 52 },
		{ 428, 370, 52 },
		{ 428, 371, 52 },
		{ 428, 372, 52 },
		{ 428, 373, 52 },
		{ 428, 374, 52 },
		{ 428, 375, 52 },
		{ 428, 376, 52 },
		{ 428, 377, 52 },
		{ 428, 378, 52 },
		{ 428, 379, 52 },
		{ 428, 380, 52 },
		{ 428, 381, 52 },
		{ 428, 382, 52 },
		{ 428, 383, 52 },
		{ 428, 384, 52 },
		{ 428, 385, 52 },
		{ 428, 386, 52 },
		{ 428, 387, 52 },
		{ 428, 388, 52 },
		{ 428, 389, 52 },
		{ 428, 390, 52 },
		{ 428, 391, 52 },
		{ 428, 392, 52 },
		{ 428, 393, 52 },
		{ 428, 394, 52 },
		{ 428, 395, 52 },
		{ 428, 396, 52 },
		{ 428, 397, 52 },
		{ 428, 398, 52 },
		{ 428, 399, 52 },
		{ 428, 400, 52 },
		{ 428, 401, 52 },
		{ 428, 402, 52 },
		{ 428, 403, 52 },
		{ 428, 404, 52 },
		{ 428, 405, 52 },
		{ 428, 406, 52 },
		{ 429, 365, 52 },
		{ 429, 366, 52 },
		{ 429, 367, 52 },
		{ 429, 368, 52 },
		{ 429, 369, 52 },
		{ 429, 370, 52 },
		{ 429, 371, 52 },
		{ 429, 372, 52 },
		{ 429, 373, 52 },
		{ 429, 374, 52 },
		{ 429, 375, 52 },
		{ 429, 376, 52 },
		{ 429, 377, 52 },
		{ 429, 378, 52 },
		{ 429, 379, 52 },
		{ 429, 380, 52 },
		{ 429, 381, 52 },
		{ 429, 382, 52 },
		{ 429, 383, 52 },
		{ 429, 384, 52 },
		{ 429, 385, 52 },
		{ 429, 386, 52 },
		{ 429, 387, 52 },
		{ 429, 388, 52 },
		{ 429, 389, 52 },
		{ 429, 390, 52 },
		{ 429, 391, 52 },
		{ 429, 392, 52 },
		{ 429, 393, 52 },
		{ 429, 394, 52 },
		{ 429, 395, 52 },
		{ 429, 396, 52 },
		{ 429, 397, 52 },
		{ 429, 398, 52 },
		{ 429, 399, 52 },
		{ 429, 400, 52 },
		{ 429, 401, 52 },
		{ 429, 402, 52 },
		{ 429, 403, 52 },
		{ 429, 404, 52 },
		{ 429, 405, 52 },
		{ 429, 406, 52 },
		{ 429, 407, 52 },
		{ 430, 366, 52 },
		{ 430, 367, 52 },
		{ 430, 368, 52 },
		{ 430, 369, 52 },
		{ 430, 370, 52 },
		{ 430, 371, 52 },
		{ 430, 372, 52 },
		{ 430, 373, 52 },
		{ 430, 374, 52 },
		{ 430, 375, 52 },
		{ 430, 376, 52 },
		{ 430, 377, 52 },
		{ 430, 378, 52 },
		{ 430, 379, 52 },
		{ 430, 380, 52 },
		{ 430, 381, 52 },
		{ 430, 382, 52 },
		{ 430, 383, 52 },
		{ 430, 384, 52 },
		{ 430, 385, 52 },
		{ 430, 386, 52 },
		{ 430, 387, 52 },
		{ 430, 388, 52 },
		{ 430, 389, 52 },
		{ 430, 390, 52 },
		{ 430, 391, 52 },
		{ 430, 392, 52 },
		{ 430, 393, 52 },
		{ 430, 394, 52 },
		{ 430, 395, 52 },
		{ 430, 396, 52 },
		{ 430, 397, 52 },
		{ 430, 398, 52 },
		{ 430, 399, 52 },
		{ 430, 400, 52 },
		{ 430, 401, 52 },
		{ 430, 402, 52 },
		{ 430, 403, 52 },
		{ 430, 404, 52 },
		{ 430, 405, 52 },
		{ 430, 406, 52 },
		{ 430, 407, 52 },
		{ 430, 408, 52 },
		{ 431, 366, 52 },
		{ 431, 367, 52 },
		{ 431, 368, 52 },
		{ 431, 369, 52 },
		{ 431, 370, 52 },
		{ 431, 371, 52 },
		{ 431, 372, 52 },
		{ 431, 373, 52 },
		{ 431, 374, 52 },
		{ 431, 375, 52 },
		{ 431, 376, 52 },
		{ 431, 377, 52 },
		{ 431, 378, 52 },
		{ 431, 379, 52 },
		{ 431, 380, 52 },
		{ 431, 381, 52 },
		{ 431, 382, 52 },
		{ 431, 383, 52 },
		{ 431, 384, 52 },
		{ 431, 385, 52 },
		{ 431, 386, 52 },
		{ 431, 387, 52 },
		{ 431, 388, 52 },
		{ 431, 389, 52 },
		{ 431, 390, 52 },
		{ 431, 391, 52 },
		{ 431, 392, 52 },
		{ 431, 393, 52 },
		{ 431, 394, 52 },
		{ 431, 395, 52 },
		{ 431, 396, 52 },
		{ 431, 397, 52 },
		{ 431, 398, 52 },
		{ 431, 399, 52 },
		{ 431, 400, 52 },
		{ 431, 401, 52 },
		{ 431, 402, 52 },
		{ 431, 403, 52 },
		{ 431, 404, 52 },
		{ 431, 405, 52 },
		{ 431, 406, 52 },
		{ 431, 407, 52 },
		{ 431, 408, 52 },
		{ 431, 409, 52 },
		{ 432, 367, 52 },
		{ 432, 368, 52 },
		{ 432, 369, 52 },
		{ 432, 370, 52 },
		{ 432, 371, 52 },
		{ 432, 372, 52 },
		{ 432, 373, 52 },
		{ 432, 374, 52 },
		{ 432, 375, 52 },
		{ 432, 376, 52 },
		{ 432, 377, 52 },
		{ 432, 378, 52 },
		{ 432, 379, 52 },
		{ 432, 380, 52 },
		{ 432, 381, 52 },
		{ 432, 382, 52 },
		{ 432, 383, 52 },
		{ 432, 384, 52 },
		{ 432, 385, 52 },
		{ 432, 386, 52 },
		{ 432, 387, 52 },
		{ 432, 388, 52 },
		{ 432, 389, 52 },
		{ 432, 390, 52 },
		{ 432, 391, 52 },
		{ 432, 392, 52 },
		{ 432, 393, 52 },
		{ 432, 394, 52 },
		{ 432, 395, 52 },
		{ 432, 396, 52 },
		{ 432, 397, 52 },
		{ 432, 398, 52 },
		{ 432, 399, 52 },
		{ 432, 400, 52 },
		{ 432, 401, 52 },
		{ 432, 402, 52 },
		{ 432, 403, 52 },
		{ 432, 404, 52 },
		{ 432, 405, 52 },
		{ 432, 406, 52 },
		{ 432, 407, 52 },
		{ 432, 408, 52 },
		{ 432, 409, 52 },
		{ 432, 410, 52 },
		{ 433, 368, 52 },
		{ 433, 369, 52 },
		{ 433, 370, 52 },
		{ 433, 371, 52 },
		{ 433, 372, 52 },
		{ 433, 373, 52 },
		{ 433, 374, 52 },
		{ 433, 375, 52 },
		{ 433, 376, 52 },
		{ 433, 377, 52 },
		{ 433, 378, 52 },
		{ 433, 379, 52 },
		{ 433, 380, 52 },
		{ 433, 381, 52 },
		{ 433, 382, 52 },
		{ 433, 383, 52 },
		{ 433, 384, 52 },
		{ 433, 385, 52 },
		{ 433, 386, 52 },
		{ 433, 387, 52 },
		{ 433, 388, 52 },
		{ 433, 389, 52 },
		{ 433, 390, 52 },
		{ 433, 391, 52 },
		{ 433, 392, 52 },
		{ 433, 393, 52 },
		{ 433, 394, 52 },
		{ 433, 395, 52 },
		{ 433, 396, 52 },
		{ 433, 397, 52 },
		{ 433, 398, 52 },
		{ 433, 399, 52 },
		{ 433, 400, 52 },
		{ 433, 401, 52 },
		{ 433, 402, 52 },
		{ 433, 403, 52 },
		{ 433, 404, 52 },
		{ 433, 405, 52 },
		{ 433, 406, 52 },
		{ 433, 407, 52 },
		{ 433, 408, 52 },
		{ 433, 409, 52 },
		{ 433, 410, 52 },
		{ 434, 368, 52 },
		{ 434, 369, 52 },
		{ 434, 370, 52 },
		{ 434, 371, 52 },
		{ 434, 372, 52 },
		{ 434, 373, 52 },
		{ 434, 374, 52 },
		{ 434, 375, 52 },
		{ 434, 376, 52 },
		{ 434, 377, 52 },
		{ 434, 378, 52 },
		{ 434, 379, 52 },
		{ 434, 380, 52 },
		{ 434, 381, 52 },
		{ 434, 382, 52 },
		{ 434, 383, 52 },
		{ 434, 384, 52 },
		{ 434, 385, 52 },
		{ 434, 386, 52 },
		{ 434, 387, 52 },
		{ 434, 388, 52 },
		{ 434, 389, 52 },
		{ 434, 390, 52 },
		{ 434, 391, 52 },
		{ 434, 392, 52 },
		{ 434, 393, 52 },
		{ 434, 394, 52 },
		{ 434, 395, 52 },
		{ 434, 396, 52 },
		{ 434, 397, 52 },
		{ 434, 398, 52 },
		{ 434, 399, 52 },
		{ 434, 400, 52 },
		{ 434, 401, 52 },
		{ 434, 402, 52 },
		{ 434, 403, 52 },
		{ 434, 404, 52 },
		{ 434, 405, 52 },
		{ 434, 406, 52 },
		{ 434, 407, 52 },
		{ 434, 408, 52 },
		{ 434, 409, 52 },
		{ 434, 410, 52 },
		{ 434, 411, 52 },
		{ 435, 369, 52 },
		{ 435, 370, 52 },
		{ 435, 371, 52 },
		{ 435, 372, 52 },
		{ 435, 373, 52 },
		{ 435, 374, 52 },
		{ 435, 375, 52 },
		{ 435, 376, 52 },
		{ 435, 377, 52 },
		{ 435, 378, 52 },
		{ 435, 379, 52 },
		{ 435, 380, 52 },
		{ 435, 381, 52 },
		{ 435, 382, 52 },
		{ 435, 383, 52 },
		{ 435, 384, 52 },
		{ 435, 385, 52 },
		{ 435, 386, 52 },
		{ 435, 387, 52 },
		{ 435, 388, 52 },
		{ 435, 389, 52 },
		{ 435, 390, 52 },
		{ 435, 391, 52 },
		{ 435, 392, 52 },
		{ 435, 393, 52 },
		{ 435, 394, 52 },
		{ 435, 395, 52 },
		{ 435, 396, 52 },
		{ 435, 397, 52 },
		{ 435, 398, 52 },
		{ 435, 399, 52 },
		{ 435, 400, 52 },
		{ 435, 401, 52 },
		{ 435, 402, 52 },
		{ 435, 403, 52 },
		{ 435, 404, 52 },
		{ 435, 405, 52 },
		{ 435, 406, 52 },
		{ 435, 407, 52 },
		{ 435, 408, 52 },
		{ 435, 409, 52 },
		{ 435, 410, 52 },
		{ 435, 411, 52 },
		{ 435, 412, 52 },
		{ 436, 369, 52 },
		{ 436, 370, 52 },
		{ 436, 371, 52 },
		{ 436, 372, 52 },
		{ 436, 373, 52 },
		{ 436, 374, 52 },
		{ 436, 375, 52 },
		{ 436, 376, 52 },
		{ 436, 377, 52 },
		{ 436, 378, 52 },
		{ 436, 379, 52 },
		{ 436, 380, 52 },
		{ 436, 381, 52 },
		{ 436, 382, 52 },
		{ 436, 383, 52 },
		{ 436, 384, 52 },
		{ 436, 385, 52 },
		{ 436, 386, 52 },
		{ 436, 387, 52 },
		{ 436, 388, 52 },
		{ 436, 389, 52 },
		{ 436, 390, 52 },
		{ 436, 391, 52 },
		{ 436, 392, 52 },
		{ 436, 393, 52 },
		{ 436, 394, 52 },
		{ 436, 395, 52 },
		{ 436, 396, 52 },
		{ 436, 397, 52 },
		{ 436, 398, 52 },
		{ 436, 399, 52 },
		{ 436, 400, 52 },
		{ 436, 401, 52 },
		{ 436, 402, 52 },
		{ 436, 403, 52 },
		{ 436, 404, 52 },
		{ 436, 405, 52 },
		{ 436, 406, 52 },
		{ 436, 407, 52 },
		{ 436, 408, 52 },
		{ 436, 409, 52 },
		{ 436, 410, 52 },
		{ 436, 411, 52 },
		{ 436, 412, 52 },
		{ 436, 413, 52 },
		{ 437, 370, 52 },
		{ 437, 371, 52 },
		{ 437, 372, 52 },
		{ 437, 373, 52 },
		{ 437, 374, 52 },
		{ 437, 375, 52 },
		{ 437, 376, 52 },
		{ 437, 377, 52 },
		{ 437, 378, 52 },
		{ 437, 379, 52 },
		{ 437, 380, 52 },
		{ 437, 381, 52 },
		{ 437, 382, 52 },
		{ 437, 383, 52 },
		{ 437, 384, 52 },
		{ 437, 385, 52 },
		{ 437, 386, 52 },
		{ 437, 387, 52 },
		{ 437, 388, 52 },
		{ 437, 389, 52 },
		{ 437, 390, 52 },
		{ 437, 391, 52 },
		{ 437, 392, 52 },
		{ 437, 393, 52 },
		{ 437, 394, 52 },
		{ 437, 395, 52 },
		{ 437, 396, 52 },
		{ 437, 397, 52 },
		{ 437, 398, 52 },
		{ 437, 399, 52 },
		{ 437, 400, 52 },
		{ 437, 401, 52 },
		{ 437, 402, 52 },
		{ 437, 403, 52 },
		{ 437, 404, 52 },
		{ 437, 405, 52 },
		{ 437, 406, 52 },
		{ 437, 407, 52 },
		{ 437, 408, 52 },
		{ 437, 409, 52 },
		{ 437, 410, 52 },
		{ 437, 411, 52 },
		{ 437, 412, 52 },
		{ 437, 413, 52 },
		{ 437, 414, 52 },
		{ 438, 371, 52 },
		{ 438, 372, 52 },
		{ 438, 373, 52 },
		{ 438, 374, 52 },
		{ 438, 375, 52 },
		{ 438, 376, 52 },
		{ 438, 377, 52 },
		{ 438, 378, 52 },
		{ 438, 379, 52 },
		{ 438, 380, 52 },
		{ 438, 381, 52 },
		{ 438, 382, 52 },
		{ 438, 383, 52 },
		{ 438, 384, 52 },
		{ 438, 385, 52 },
		{ 438, 386, 52 },
		{ 438, 387, 52 },
		{ 438, 388, 52 },
		{ 438, 389, 52 },
		{ 438, 390, 52 },
		{ 438, 391, 52 },
		{ 438, 392, 52 },
		{ 438, 393, 52 },
		{ 438, 394, 52 },
		{ 438, 395, 52 },
		{ 438, 396, 52 },
		{ 438, 397, 52 },
		{ 438, 398, 52 },
		{ 438, 399, 52 },
		{ 438, 400, 52 },
		{ 438, 401, 52 },
		{ 438, 402, 52 },
		{ 438, 403, 52 },
		{ 438, 404, 52 },
		{ 438, 405, 52 },
		{ 438, 406, 52 },
		{ 438, 407, 52 },
		{ 438, 408, 52 },
		{ 438, 409, 52 },
		{ 438, 410, 52 },
		{ 438, 411, 52 },
		{ 438, 412, 52 },
		{ 438, 413, 52 },
		{ 438, 414, 52 },
		{ 438, 415, 52 },
		{ 439, 371, 52 },
		{ 439, 372, 52 },
		{ 439, 373, 52 },
		{ 439, 374, 52 },
		{ 439, 375, 52 },
		{ 439, 376, 52 },
		{ 439, 377, 52 },
		{ 439, 378, 52 },
		{ 439, 379, 52 },
		{ 439, 380, 52 },
		{ 439, 381, 52 },
		{ 439, 382, 52 },
		{ 439, 383, 52 },
		{ 439, 384, 52 },
		{ 439, 385, 52 },
		{ 439, 386, 52 },
		{ 439, 387, 52 },
		{ 439, 388, 52 },
		{ 439, 389, 52 },
		{ 439, 390, 52 },
		{ 439, 391, 52 },
		{ 439, 392, 52 },
		{ 439, 393, 52 },
		{ 439, 394, 52 },
		{ 439, 395, 52 },
		{ 439, 396, 52 },
		{ 439, 397, 52 },
		{ 439, 398, 52 },
		{ 439, 399, 52 },
		{ 439, 400, 52 },
		{ 439, 401, 52 },
		{ 439, 402, 52 },
		{ 439, 403, 52 },
		{ 439, 404, 52 },
		{ 439, 405, 52 },
		{ 439, 406, 52 },
		{ 439, 407, 52 },
		{ 439, 408, 52 },
		{ 439, 409, 52 },
		{ 439, 410, 52 },
		{ 439, 411, 52 },
		{ 439, 412, 52 },
		{ 439, 413, 52 },
		{ 439, 414, 52 },
		{ 439, 415, 52 },
		{ 439, 416, 52 },
		{ 440, 372, 52 },
		{ 440, 373, 52 },
		{ 440, 374, 52 },
		{ 440, 375, 52 },
		{ 440, 376, 52 },
		{ 440, 377, 52 },
		{ 440, 378, 52 },
		{ 440, 379, 52 },
		{ 440, 380, 52 },
		{ 440, 381, 52 },
		{ 440, 382, 52 },
		{ 440, 383, 52 },
		{ 440, 384, 52 },
		{ 440, 385, 52 },
		{ 440, 386, 52 },
		{ 440, 387, 52 },
		{ 440, 388, 52 },
		{ 440, 389, 52 },
		{ 440, 390, 52 },
		{ 440, 391, 52 },
		{ 440, 392, 52 },
		{ 440, 393, 52 },
		{ 440, 394, 52 },
		{ 440, 395, 52 },
		{ 440, 396, 52 },
		{ 440, 397, 52 },
		{ 440, 398, 52 },
		{ 440, 399, 52 },
		{ 440, 400, 52 },
		{ 440, 401, 52 },
		{ 440, 402, 52 },
		{ 440, 403, 52 },
		{ 440, 404, 52 },
		{ 440, 405, 52 },
		{ 440, 406, 52 },
		{ 440, 407, 52 },
		{ 440, 408, 52 },
		{ 440, 409, 52 },
		{ 440, 410, 52 },
		{ 440, 411, 52 },
		{ 440, 412, 52 },
		{ 440, 413, 52 },
		{ 440, 414, 52 },
		{ 440, 415, 52 },
		{ 440, 416, 52 },
		{ 440, 417, 52 },
		{ 441, 373, 52 },
		{ 441, 374, 52 },
		{ 441, 375, 52 },
		{ 441, 376, 52 },
		{ 441, 377, 52 },
		{ 441, 378, 52 },
		{ 441, 379, 52 },
		{ 441, 380, 52 },
		{ 441, 381, 52 },
		{ 441, 382, 52 },
		{ 441, 383, 52 },
		{ 441, 384, 52 },
		{ 441, 385, 52 },
		{ 441, 386, 52 },
		{ 441, 387, 52 },
		{ 441, 388, 52 },
		{ 441, 389, 52 },
		{ 441, 390, 52 },
		{ 441, 391, 52 },
		{ 441, 392, 52 },
		{ 441, 393, 52 },
		{ 441, 394, 52 },
		{ 441, 395, 52 },
		{ 441, 396, 52 },
		{ 441, 397, 52 },
		{ 441, 398, 52 },
		{ 441, 399, 52 },
		{ 441, 400, 52 },
		{ 441, 401, 52 },
		{ 441, 402, 52 },
		{ 441, 403, 52 },
		{ 441, 404, 52 },
		{ 441, 405, 52 },
		{ 441, 406, 52 },
		{ 441, 407, 52 },
		{ 441, 408, 52 },
		{ 441, 409, 52 },
		{ 441, 410, 52 },
		{ 441, 411, 52 },
		{ 441, 412, 52 },
		{ 441, 413, 52 },
		{ 441, 414, 52 },
		{ 441, 415, 52 },
		{ 441, 416, 52 },
		{ 441, 417, 52 },
		{ 442, 373, 52 },
		{ 442, 374, 52 },
		{ 442, 375, 52 },
		{ 442, 376, 52 },
		{ 442, 377, 52 },
		{ 442, 378, 52 },
		{ 442, 379, 52 },
		{ 442, 380, 52 },
		{ 442, 381, 52 },
		{ 442, 382, 52 },
		{ 442, 383, 52 },
		{ 442, 384, 52 },
		{ 442, 385, 52 },
		{ 442, 386, 52 },
		{ 442, 387, 52 },
		{ 442, 388, 52 },
		{ 442, 389, 52 },
		{ 442, 390, 52 },
		{ 442, 391, 52 },
		{ 442, 392, 52 },
		{ 442, 393, 52 },
		{ 442, 394, 52 },
		{ 442, 395, 52 },
		{ 442, 396, 52 },
		{ 442, 397, 52 },
		{ 442, 398, 52 },
		{ 442, 399, 52 },
		{ 442, 400, 52 },
		{ 442, 401, 52 },
		{ 442, 402, 52 },
		{ 442, 403, 52 },
		{ 442, 404, 52 },
		{ 442, 405, 52 },
		{ 442, 406, 52 },
		{ 442, 407, 52 },
		{ 442, 408, 52 },
		{ 442, 409, 52 },
		{ 442, 410, 52 },
		{ 442, 411, 52 },
		{ 442, 412, 52 },
		{ 442, 413, 52 },
		{ 442, 414, 52 },
		{ 442, 415, 52 },
		{ 442, 416, 52 },
		{ 442, 417, 52 },
		{ 442, 418, 52 },
		{ 443, 374, 52 },
		{ 443, 375, 52 },
		{ 443, 376, 52 },
		{ 443, 377, 52 },
		{ 443, 378, 52 },
		{ 443, 379, 52 },
		{ 443, 380, 52 },
		{ 443, 381, 52 },
		{ 443, 382, 52 },
		{ 443, 383, 52 },
		{ 443, 384, 52 },
		{ 443, 385, 52 },
		{ 443, 386, 52 },
		{ 443, 387, 52 },
		{ 443, 388, 52 },
		{ 443, 389, 52 },
		{ 443, 390, 52 },
		{ 443, 391, 52 },
		{ 443, 392, 52 },
		{ 443, 393, 52 },
		{ 443, 394, 52 },
		{ 443, 395, 52 },
		{ 443, 396, 52 },
		{ 443, 397, 52 },
		{ 443, 398, 52 },
		{ 443, 399, 52 },
		{ 443, 400, 52 },
		{ 443, 401, 52 },
		{ 443, 402, 52 },
		{ 443, 403, 52 },
		{ 443, 404, 52 },
		{ 443, 405, 52 },
		{ 443, 406, 52 },
		{ 443, 407, 52 },
		{ 443, 408, 52 },
		{ 443, 409, 52 },
		{ 443, 410, 52 },
		{ 443, 411, 52 },
		{ 443, 412, 52 },
		{ 443, 413, 52 },
		{ 443, 414, 52 },
		{ 443, 415, 52 },
		{ 443, 416, 52 },
		{ 443, 417, 52 },
		{ 443, 418, 52 },
		{ 443, 419, 52 },
		{ 444, 374, 52 },
		{ 444, 375, 52 },
		{ 444, 376, 52 },
		{ 444, 377, 52 },
		{ 444, 378, 52 },
		{ 444, 379, 52 },
		{ 444, 380, 52 },
		{ 444, 381, 52 },
		{ 444, 382, 52 },
		{ 444, 383, 52 },
		{ 444, 384, 52 },
		{ 444, 385, 52 },
		{ 444, 386, 52 },
		{ 444, 387, 52 },
		{ 444, 388, 52 },
		{ 444, 389, 52 },
		{ 444, 390, 52 },
		{ 444, 391, 52 },
		{ 444, 392, 52 },
		{ 444, 393, 52 },
		{ 444, 394, 52 },
		{ 444, 395, 52 },
		{ 444, 396, 52 },
		{ 444, 397, 52 },
		{ 444, 398, 52 },
		{ 444, 399, 52 },
		{ 444, 400, 52 },
		{ 444, 401, 52 },
		{ 444, 402, 52 },
		{ 444, 403, 52 },
		{ 444, 404, 52 },
		{ 444, 405, 52 },
		{ 444, 406, 52 },
		{ 444, 407, 52 },
		{ 444, 408, 52 },
		{ 444, 409, 52 },
		{ 444, 410, 52 },
		{ 444, 411, 52 },
		{ 444, 412, 52 },
		{ 444, 413, 52 },
		{ 444, 414, 52 },
		{ 444, 415, 52 },
		{ 444, 416, 52 },
		{ 444, 417, 52 },
		{ 444, 418, 52 },
		{ 444, 419, 52 },
		{ 444, 420, 52 },
		{ 445, 375, 52 },
		{ 445, 376, 52 },
		{ 445, 377, 52 },
		{ 445, 378, 52 },
		{ 445, 379, 52 },
		{ 445, 380, 52 },
		{ 445, 381, 52 },
		{ 445, 382, 52 },
		{ 445, 383, 52 },
		{ 445, 384, 52 },
		{ 445, 385, 52 },
		{ 445, 386, 52 },
		{ 445, 387, 52 },
		{ 445, 388, 52 },
		{ 445, 389, 52 },
		{ 445, 390, 52 },
		{ 445, 391, 52 },
		{ 445, 392, 52 },
		{ 445, 393, 52 },
		{ 445, 394, 52 },
		{ 445, 395, 52 },
		{ 445, 396, 52 },
		{ 445, 397, 52 },
		{ 445, 398, 52 },
		{ 445, 399, 52 },
		{ 445, 400, 52 },
		{ 445, 401, 52 },
		{ 445, 402, 52 },
		{ 445, 403, 52 },
		{ 445, 404, 52 },
		{ 445, 405, 52 },
		{ 445, 406, 52 },
		{ 445, 407, 52 },
		{ 445, 408, 52 },
		{ 445, 409, 52 },
		{ 445, 410, 52 },
		{ 445, 411, 52 },
		{ 445, 412, 52 },
		{ 445, 413, 52 },
		{ 445, 414, 52 },
		{ 445, 415, 52 },
		{ 445, 416, 52 },
		{ 445, 417, 52 },
		{ 445, 418, 52 },
		{ 445, 419, 52 },
		{ 445, 420, 52 },
		{ 445, 421, 52 },
		{ 446, 376, 52 },
		{ 446, 377, 52 },
		{ 446, 378, 52 },
		{ 446, 379, 52 },
		{ 446, 380, 52 },
		{ 446, 381, 52 },
		{ 446, 382, 52 },
		{ 446, 383, 52 },
		{ 446, 384, 52 },
		{ 446, 385, 52 },
		{ 446, 386, 52 },
		{ 446, 387, 52 },
		{ 446, 388, 52 },
		{ 446, 389, 52 },
		{ 446, 390, 52 },
		{ 446, 391, 52 },
		{ 446, 392, 52 },
		{ 446, 393, 52 },
		{ 446, 394, 52 },
		{ 446, 395, 52 },
		{ 446, 396, 52 },
		{ 446, 397, 52 },
		{ 446, 398, 52 },
		{ 446, 399, 52 },
		{ 446, 400, 52 },
		{ 446, 401, 52 },
		{ 446, 402, 52 },
		{ 446, 403, 52 },
		{ 446, 404, 52 },
		{ 446, 405, 52 },
		{ 446, 406, 52 },
		{ 446, 407, 52 },
		{ 446, 408, 52 },
		{ 446, 409, 52 },
		{ 446, 410, 52 },
		{ 446, 411, 52 },
		{ 446, 412, 52 },
		{ 446, 413, 52 },
		{ 446, 414, 52 },
		{ 446, 415, 52 },
		{ 446, 416, 52 },
		{ 446, 417, 52 },
		{ 446, 418, 52 },
		{ 446, 419, 52 },
		{ 446, 420, 52 },
		{ 446, 421, 52 },
		{ 446, 422, 52 },
		{ 447, 376, 52 },
		{ 447, 377, 52 },
		{ 447, 378, 52 },
		{ 447, 379, 52 },
		{ 447, 380, 52 },
		{ 447, 381, 52 },
		{ 447, 382, 52 },
		{ 447, 383, 52 },
		{ 447, 384, 52 },
		{ 447, 385, 52 },
		{ 447, 386, 52 },
		{ 447, 387, 52 },
		{ 447, 388, 52 },
		{ 447, 389, 52 },
		{ 447, 390, 52 },
		{ 447, 391, 52 },
		{ 447, 392, 52 },
		{ 447, 393, 52 },
		{ 447, 394, 52 },
		{ 447, 395, 52 },
		{ 447, 396, 52 },
		{ 447, 397, 52 },
		{ 447, 398, 52 },
		{ 447, 399, 52 },
		{ 447, 400, 52 },
		{ 447, 401, 52 },
		{ 447, 402, 52 },
		{ 447, 403, 52 },
		{ 447, 404, 52 },
		{ 447, 405, 52 },
		{ 447, 406, 52 },
		{ 447, 407, 52 },
		{ 447, 408, 52 },
		{ 447, 409, 52 },
		{ 447, 410, 52 },
		{ 447, 411, 52 },
		{ 447, 412, 52 },
		{ 447, 413, 52 },
		{ 447, 414, 52 },
		{ 447, 415, 52 },
		{ 447, 416, 52 },
		{ 447, 417, 52 },
		{ 447, 418, 52 },
		{ 447, 419, 52 },
		{ 447, 420, 52 },
		{ 447, 421, 52 },
		{ 447, 422, 52 },
		{ 447, 423, 52 },
		{ 448, 377, 52 },
		{ 448, 378, 52 },
		{ 448, 379, 52 },
		{ 448, 380, 52 },
		{ 448, 381, 52 },
		{ 448, 382, 52 },
		{ 448, 383, 52 },
		{ 448, 384, 52 },
		{ 448, 385, 52 },
		{ 448, 386, 52 },
		{ 448, 387, 52 },
		{ 448, 388, 52 },
		{ 448, 389, 52 },
		{ 448, 390, 52 },
		{ 448, 391, 52 },
		{ 448, 392, 52 },
		{ 448, 393, 52 },
		{ 448, 394, 52 },
		{ 448, 395, 52 },
		{ 448, 396, 52 },
		{ 448, 397, 52 },
		{ 448, 398, 52 },
		{ 448, 399, 52 },
		{ 448, 400, 52 },
		{ 448, 401, 52 },
		{ 448, 402, 52 },
		{ 448, 403, 52 },
		{ 448, 404, 52 },
		{ 448, 405, 52 },
		{ 448, 406, 52 },
		{ 448, 407, 52 },
		{ 448, 408, 52 },
		{ 448, 409, 52 },
		{ 448, 410, 52 },
		{ 448, 411, 52 },
		{ 448, 412, 52 },
		{ 448, 413, 52 },
		{ 448, 414, 52 },
		{ 448, 415, 52 },
		{ 448, 416, 52 },
		{ 448, 417, 52 },
		{ 448, 418, 52 },
		{ 448, 419, 52 },
		{ 448, 420, 52 },
		{ 448, 421, 52 },
		{ 448, 422, 52 },
		{ 448, 423, 52 },
		{ 448, 424, 52 },
		{ 449, 378, 52 },
		{ 449, 379, 52 },
		{ 449, 380, 52 },
		{ 449, 381, 52 },
		{ 449, 382, 52 },
		{ 449, 383, 52 },
		{ 449, 384, 52 },
		{ 449, 385, 52 },
		{ 449, 386, 52 },
		{ 449, 387, 52 },
		{ 449, 388, 52 },
		{ 449, 389, 52 },
		{ 449, 390, 52 },
		{ 449, 391, 52 },
		{ 449, 392, 52 },
		{ 449, 393, 52 },
		{ 449, 394, 52 },
		{ 449, 395, 52 },
		{ 449, 396, 52 },
		{ 449, 397, 52 },
		{ 449, 398, 52 },
		{ 449, 399, 52 },
		{ 449, 400, 52 },
		{ 449, 401, 52 },
		{ 449, 402, 52 },
		{ 449, 403, 52 },
		{ 449, 404, 52 },
		{ 449, 405, 52 },
		{ 449, 406, 52 },
		{ 449, 407, 52 },
		{ 449, 408, 52 },
		{ 449, 409, 52 },
		{ 449, 410, 52 },
		{ 449, 411, 52 },
		{ 449, 412, 52 },
		{ 449, 413, 52 },
		{ 449, 414, 52 },
		{ 449, 415, 52 },
		{ 449, 416, 52 },
		{ 449, 417, 52 },
		{ 449, 418, 52 },
		{ 449, 419, 52 },
		{ 449, 420, 52 },
		{ 449, 421, 52 },
		{ 449, 422, 52 },
		{ 449, 423, 52 },
		{ 449, 424, 52 },
		{ 450, 378, 52 },
		{ 450, 379, 52 },
		{ 450, 380, 52 },
		{ 450, 381, 52 },
		{ 450, 382, 52 },
		{ 450, 383, 52 },
		{ 450, 384, 52 },
		{ 450, 385, 52 },
		{ 450, 386, 52 },
		{ 450, 387, 52 },
		{ 450, 388, 52 },
		{ 450, 389, 52 },
		{ 450, 390, 52 },
		{ 450, 391, 52 },
		{ 450, 392, 52 },
		{ 450, 393, 52 },
		{ 450, 394, 52 },
		{ 450, 395, 52 },
		{ 450, 396, 52 },
		{ 450, 397, 52 },
		{ 450, 398, 52 },
		{ 450, 399, 52 },
		{ 450, 400, 52 },
		{ 450, 401, 52 },
		{ 450, 402, 52 },
		{ 450, 403, 52 },
		{ 450, 404, 52 },
		{ 450, 405, 52 },
		{ 450, 406, 52 },
		{ 450, 407, 52 },
		{ 450, 408, 52 },
		{ 450, 409, 52 },
		{ 450, 410, 52 },
		{ 450, 411, 52 },
		{ 450, 412, 52 },
		{ 450, 413, 52 },
		{ 450, 414, 52 },
		{ 450, 415, 52 },
		{ 450, 416, 52 },
		{ 450, 417, 52 },
		{ 450, 418, 52 },
		{ 450, 419, 52 },
		{ 450, 420, 52 },
		{ 450, 421, 52 },
		{ 450, 422, 52 },
		{ 450, 423, 52 },
		{ 450, 424, 52 },
		{ 450, 425, 52 },
		{ 451, 379, 52 },
		{ 451, 380, 52 },
		{ 451, 381, 52 },
		{ 451, 382, 52 },
		{ 451, 383, 52 },
		{ 451, 384, 52 },
		{ 451, 385, 52 },
		{ 451, 386, 52 },
		{ 451, 387, 52 },
		{ 451, 388, 52 },
		{ 451, 389, 52 },
		{ 451, 390, 52 },
		{ 451, 391, 52 },
		{ 451, 392, 52 },
		{ 451, 393, 52 },
		{ 451, 394, 52 },
		{ 451, 395, 52 },
		{ 451, 396, 52 },
		{ 451, 397, 52 },
		{ 451, 398, 52 },
		{ 451, 399, 52 },
		{ 451, 400, 52 },
		{ 451, 401, 52 },
		{ 451, 402, 52 },
		{ 451, 403, 52 },
		{ 451, 404, 52 },
		{ 451, 405, 52 },
		{ 451, 406, 52 },
		{ 451, 407, 52 },
		{ 451, 408, 52 },
		{ 451, 409, 52 },
		{ 451, 410, 52 },
		{ 451, 411, 52 },
		{ 451, 412, 52 },
		{ 451, 413, 52 },
		{ 451, 414, 52 },
		{ 451, 415, 52 },
		{ 451, 416, 52 },
		{ 451, 417, 52 },
		{ 451, 418, 52 },
		{ 451, 419, 52 },
		{ 451, 420, 52 },
		{ 451, 421, 52 },
		{ 451, 422, 52 },
		{ 451, 423, 52 },
		{ 451, 424, 52 },
		{ 451, 425, 52 },
		{ 451, 426, 52 },
		{ 452, 379, 52 },
		{ 452, 380, 52 },
		{ 452, 381, 52 },
		{ 452, 382, 52 },
		{ 452, 383, 52 },
		{ 452, 384, 52 },
		{ 452, 385, 52 },
		{ 452, 386, 52 },
		{ 452, 387, 52 },
		{ 452, 388, 52 },
		{ 452, 389, 52 },
		{ 452, 390, 52 },
		{ 452, 391, 52 },
		{ 452, 392, 52 },
		{ 452, 393, 52 },
		{ 452, 394, 52 },
		{ 452, 395, 52 },
		{ 452, 396, 52 },
		{ 452, 397, 52 },
		{ 452, 398, 52 },
		{ 452, 399, 52 },
		{ 452, 400, 52 },
		{ 452, 401, 52 },
		{ 452, 402, 52 },
		{ 452, 403, 52 },
		{ 452, 404, 52 },
		{ 452, 405, 52 },
		{ 452, 406, 52 },
		{ 452, 407, 52 },
		{ 452, 408, 52 },
		{ 452, 409, 52 },
		{ 452, 410, 52 },
		{ 452, 411, 52 },
		{ 452, 412, 52 },
		{ 452, 413, 52 },
		{ 452, 414, 52 },
		{ 452, 415, 52 },
		{ 452, 416, 52 },
		{ 452, 417, 52 },
		{ 452, 418, 52 },
		{ 452, 419, 52 },
		{ 452, 420, 52 },
		{ 452, 421, 52 },
		{ 452, 422, 52 },
		{ 452, 423, 52 },
		{ 452, 424, 52 },
		{ 452, 425, 52 },
		{ 452, 426, 52 },
		{ 452, 427, 52 },
		{ 453, 380, 52 },
		{ 453, 381, 52 },
		{ 453, 382, 52 },
		{ 453, 383, 52 },
		{ 453, 384, 52 },
		{ 453, 385, 52 },
		{ 453, 386, 52 },
		{ 453, 387, 52 },
		{ 453, 388, 52 },
		{ 453, 389, 52 },
		{ 453, 390, 52 },
		{ 453, 391, 52 },
		{ 453, 392, 52 },
		{ 453, 393, 52 },
		{ 453, 394, 52 },
		{ 453, 395, 52 },
		{ 453, 396, 52 },
		{ 453, 397, 52 },
		{ 453, 398, 52 },
		{ 453, 399, 52 },
		{ 453, 400, 52 },
		{ 453, 401, 52 },
		{ 453, 402, 52 },
		{ 453, 403, 52 },
		{ 453, 404, 52 },
		{ 453, 405, 52 },
		{ 453, 406, 52 },
		{ 453, 407, 52 },
		{ 453, 408, 52 },
		{ 453, 409, 52 },
		{ 453, 410, 52 },
		{ 453, 411, 52 },
		{ 453, 412, 52 },
		{ 453, 413, 52 },
		{ 453, 414, 52 },
		{ 453, 415, 52 },
		{ 453, 416, 52 },
		{ 453, 417, 52 },
		{ 453, 418, 52 },
		{ 453, 419, 52 },
		{ 453, 420, 52 },
		{ 453, 421, 52 },
		{ 453, 422, 52 },
		{ 453, 423, 52 },
		{ 453, 424, 52 },
		{ 453, 425, 52 },
		{ 453, 426, 52 },
		{ 453, 427, 52 },
		{ 453, 428, 52 },
		{ 454, 381, 52 },
		{ 454, 382, 52 },
		{ 454, 383, 52 },
		{ 454, 384, 52 },
		{ 454, 385, 52 },
		{ 454, 386, 52 },
		{ 454, 387, 52 },
		{ 454, 388, 52 },
		{ 454, 389, 52 },
		{ 454, 390, 52 },
		{ 454, 391, 52 },
		{ 454, 392, 52 },
		{ 454, 393, 52 },
		{ 454, 394, 52 },
		{ 454, 395, 52 },
		{ 454, 396, 52 },
		{ 454, 397, 52 },
		{ 454, 398, 52 },
		{ 454, 399, 52 },
		{ 454, 400, 52 },
		{ 454, 401, 52 },
		{ 454, 402, 52 },
		{ 454, 403, 52 },
		{ 454, 404, 52 },
		{ 454, 405, 52 },
		{ 454, 406, 52 },
		{ 454, 407, 52 },
		{ 454, 408, 52 },
		{ 454, 409, 52 },
		{ 454, 410, 52 },
		{ 454, 411, 52 },
		{ 454, 412, 52 },
		{ 454, 413, 52 },
		{ 454, 414, 52 },
		{ 454, 415, 52 },
		{ 454, 416, 52 },
		{ 454, 417, 52 },
		{ 454, 418, 52 },
		{ 454, 419, 52 },
		{ 454, 420, 52 },
		{ 454, 421, 52 },
		{ 454, 422, 52 },
		{ 454, 423, 52 },
		{ 454, 424, 52 },
		{ 454, 425, 52 },
		{ 454, 426, 52 },
		{ 454, 427, 52 },
		{ 454, 428, 52 },
		{ 454, 429, 52 },
		{ 455, 381, 52 },
		{ 455, 382, 52 },
		{ 455, 383, 52 },
		{ 455, 384, 52 },
		{ 455, 385, 52 },
		{ 455, 386, 52 },
		{ 455, 387, 52 },
		{ 455, 388, 52 },
		{ 455, 389, 52 },
		{ 455, 390, 52 },
		{ 455, 391, 52 },
		{ 455, 392, 52 },
		{ 455, 393, 52 },
		{ 455, 394, 52 },
		{ 455, 395, 52 },
		{ 455, 396, 52 },
		{ 455, 397, 52 },
		{ 455, 398, 52 },
		{ 455, 399, 52 },
		{ 455, 400, 52 },
		{ 455, 401, 52 },
		{ 455, 402, 52 },
		{ 455, 403, 52 },
		{ 455, 404, 52 },
		{ 455, 405, 52 },
		{ 455, 406, 52 },
		{ 455, 407, 52 },
		{ 455, 408, 52 },
		{ 455, 409, 52 },
		{ 455, 410, 52 },
		{ 455, 411, 52 },
		{ 455, 412, 52 },
		{ 455, 413, 52 },
		{ 455, 414, 52 },
		{ 455, 415, 52 },
		{ 455, 416, 52 },
		{ 455, 417, 52 },
		{ 455, 418, 52 },
		{ 455, 419, 52 },
		{ 455, 420, 52 },
		{ 455, 421, 52 },
		{ 455, 422, 52 },
		{ 455, 423, 52 },
		{ 455, 424, 52 },
		{ 455, 425, 52 },
		{ 455, 426, 52 },
		{ 455, 427, 52 },
		{ 455, 428, 52 },
		{ 455, 429, 52 },
		{ 455, 430, 52 },
		{ 456, 382, 52 },
		{ 456, 383, 52 },
		{ 456, 384, 52 },
		{ 456, 385, 52 },
		{ 456, 386, 52 },
		{ 456, 387, 52 },
		{ 456, 388, 52 },
		{ 456, 389, 52 },
		{ 456, 390, 52 },
		{ 456, 391, 52 },
		{ 456, 392, 52 },
		{ 456, 393, 52 },
		{ 456, 394, 52 },
		{ 456, 395, 52 },
		{ 456, 396, 52 },
		{ 456, 397, 52 },
		{ 456, 398, 52 },
		{ 456, 399, 52 },
		{ 456, 400, 52 },
		{ 456, 401, 52 },
		{ 456, 402, 52 },
		{ 456, 403, 52 },
		{ 456, 404, 52 },
		{ 456, 405, 52 },
		{ 456, 406, 52 },
		{ 456, 407, 52 },
		{ 456, 408, 52 },
		{ 456, 409, 52 },
		{ 456, 410, 52 },
		{ 456, 411, 52 },
		{ 456, 412, 52 },
		{ 456, 413, 52 },
		{ 456, 414, 52 },
		{ 456, 415, 52 },
		{ 456, 416, 52 },
		{ 456, 417, 52 },
		{ 456, 418, 52 },
		{ 456, 419, 52 },
		{ 456, 420, 52 },
		{ 456, 421, 52 },
		{ 456, 422, 52 },
		{ 456, 423, 52 },
		{ 456, 424, 52 },
		{ 456, 425, 52 },
		{ 456, 426, 52 },
		{ 456, 427, 52 },
		{ 456, 428, 52 },
		{ 456, 429, 52 },
		{ 456, 430, 52 },
		{ 456, 431, 52 },
		{ 457, 383, 52 },
		{ 457, 384, 52 },
		{ 457, 385, 52 },
		{ 457, 386, 52 },
		{ 457, 387, 52 },
		{ 457, 388, 52 },
		{ 457, 389, 52 },
		{ 457, 390, 52 },
		{ 457, 391, 52 },
		{ 457, 392, 52 },
		{ 457, 393, 52 },
		{ 457, 394, 52 },
		{ 457, 395, 52 },
		{ 457, 396, 52 },
		{ 457, 397, 52 },
		{ 457, 398, 52 },
		{ 457, 399, 52 },
		{ 457, 400, 52 },
		{ 457, 401, 52 },
		{ 457, 402, 52 },
		{ 457, 403, 52 },
		{ 457, 404, 52 },
		{ 457, 405, 52 },
		{ 457, 406, 52 },
		{ 457, 407, 52 },
		{ 457, 408, 52 },
		{ 457, 409, 52 },
		{ 457, 410, 52 },
		{ 457, 411, 52 },
		{ 457, 412, 52 },
		{ 457, 413, 52 },
		{ 457, 414, 52 },
		{ 457, 415, 52 },
		{ 457, 416, 52 },
		{ 457, 417, 52 },
		{ 457, 418, 52 },
		{ 457, 419, 52 },
		{ 457, 420, 52 },
		{ 457, 421, 52 },
		{ 457, 422, 52 },
		{ 457, 423, 52 },
		{ 457, 424, 52 },
		{ 457, 425, 52 },
		{ 457, 426, 52 },
		{ 457, 427, 52 },
		{ 457, 428, 52 },
		{ 457, 429, 52 },
		{ 457, 430, 52 },
		{ 457, 431, 52 },
		{ 458, 383, 52 },
		{ 458, 384, 52 },
		{ 458, 385, 52 },
		{ 458, 386, 52 },
		{ 458, 387, 52 },
		{ 458, 388, 52 },
		{ 458, 389, 52 },
		{ 458, 390, 52 },
		{ 458, 391, 52 },
		{ 458, 392, 52 },
		{ 458, 393, 52 },
		{ 458, 394, 52 },
		{ 458, 395, 52 },
		{ 458, 396, 52 },
		{ 458, 397, 52 },
		{ 458, 398, 52 },
		{ 458, 399, 52 },
		{ 458, 400, 52 },
		{ 458, 401, 52 },
		{ 458, 402, 52 },
		{ 458, 403, 52 },
		{ 458, 404, 52 },
		{ 458, 405, 52 },
		{ 458, 406, 52 },
		{ 458, 407, 52 },
		{ 458, 408, 52 },
		{ 458, 409, 52 },
		{ 458, 410, 52 },
		{ 458, 411, 52 },
		{ 458, 412, 52 },
		{ 458, 413, 52 },
		{ 458, 414, 52 },
		{ 458, 415, 52 },
		{ 458, 416, 52 },
		{ 458, 417, 52 },
		{ 458, 418, 52 },
		{ 458, 419, 52 },
		{ 458, 420, 52 },
		{ 458, 421, 52 },
		{ 458, 422, 52 },
		{ 458, 423, 52 },
		{ 458, 424, 52 },
		{ 458, 425, 52 },
		{ 458, 426, 52 },
		{ 458, 427, 52 },
		{ 458, 428, 52 },
		{ 458, 429, 52 },
		{ 458, 430, 52 },
		{ 458, 431, 52 },
		{ 458, 432, 52 },
		{ 459, 384, 52 },
		{ 459, 385, 52 },
		{ 459, 386, 52 },
		{ 459, 387, 52 },
		{ 459, 388, 52 },
		{ 459, 389, 52 },
		{ 459, 390, 52 },
		{ 459, 391, 52 },
		{ 459, 392, 52 },
		{ 459, 393, 52 },
		{ 459, 394, 52 },
		{ 459, 395, 52 },
		{ 459, 396, 52 },
		{ 459, 397, 52 },
		{ 459, 398, 52 },
		{ 459, 399, 52 },
		{ 459, 400, 52 },
		{ 459, 401, 52 },
		{ 459, 402, 52 },
		{ 459, 403, 52 },
		{ 459, 404, 52 },
		{ 459, 405, 52 },
		{ 459, 406, 52 },
		{ 459, 407, 52 },
		{ 459, 408, 52 },
		{ 459, 409, 52 },
		{ 459, 410, 52 },
		{ 459, 411, 52 },
		{ 459, 412, 52 },
		{ 459, 413, 52 },
		{ 459, 414, 52 },
		{ 459, 415, 52 },
		{ 459, 416, 52 },
		{ 459, 417, 52 },
		{ 459, 418, 52 },
		{ 459, 419, 52 },
		{ 459, 420, 52 },
		{ 459, 421, 52 },
		{ 459, 422, 52 },
		{ 459, 423, 52 },
		{ 459, 424, 52 },
		{ 459, 425, 52 },
		{ 459, 426, 52 },
		{ 459, 427, 52 },
		{ 459, 428, 52 },
		{ 459, 429, 52 },
		{ 459, 430, 52 },
		{ 459, 431, 52 },
		{ 459, 432, 52 },
		{ 459, 433, 52 },
		{ 460, 384, 52 },
		{ 460, 385, 52 },
		{ 460, 386, 52 },
		{ 460, 387, 52 },
		{ 460, 388, 52 },
		{ 460, 389, 52 },
		{ 460, 390, 52 },
		{ 460, 391, 52 },
		{ 460, 392, 52 },
		{ 460, 393, 52 },
		{ 460, 394, 52 },
		{ 460, 395, 52 },
		{ 460, 396, 52 },
		{ 460, 397, 52 },
		{ 460, 398, 52 },
		{ 460, 399, 52 },
		{ 460, 400, 52 },
		{ 460, 401, 52 },
		{ 460, 402, 52 },
		{ 460, 403, 52 },
		{ 460, 404, 52 },
		{ 460, 405, 52 },
		{ 460, 406, 52 },
		{ 460, 407, 52 },
		{ 460, 408, 52 },
		{ 460, 409, 52 },
		{ 460, 410, 52 },
		{ 460, 411, 52 },
		{ 460, 412, 52 },
		{ 460, 413, 52 },
		{ 460, 414, 52 },
		{ 460, 415, 52 },
		{ 460, 416, 52 },
		{ 460, 417, 52 },
		{ 460, 418, 52 },
		{ 460, 419, 52 },
		{ 460, 420, 52 },
		{ 460, 421, 52 },
		{ 460, 422, 52 },
		{ 460, 423, 52 },
		{ 460, 424, 52 },
		{ 460, 425, 52 },
		{ 460, 426, 52 },
		{ 460, 427, 52 },
		{ 460, 428, 52 },
		{ 460, 429, 52 },
		{ 460, 430, 52 },
		{ 460, 431, 52 },
		{ 460, 432, 52 },
		{ 460, 433, 52 },
		{ 460, 434, 52 },
		{ 461, 385, 52 },
		{ 461, 386, 52 },
		{ 461, 387, 52 },
		{ 461, 388, 52 },
		{ 461, 389, 52 },
		{ 461, 390, 52 },
		{ 461, 391, 52 },
		{ 461, 392, 52 },
		{ 461, 393, 52 },
		{ 461, 394, 52 },
		{ 461, 395, 52 },
		{ 461, 396, 52 },
		{ 461, 397, 52 },
		{ 461, 398, 52 },
		{ 461, 399, 52 },
		{ 461, 400, 52 },
		{ 461, 401, 52 },
		{ 461, 402, 52 },
		{ 461, 403, 52 },
		{ 461, 404, 52 },
		{ 461, 405, 52 },
		{ 461, 406, 52 },
		{ 461, 407, 52 },
		{ 461, 408, 52 },
		{ 461, 409, 52 },
		{ 461, 410, 52 },
		{ 461, 411, 52 },
		{ 461, 412, 52 },
		{ 461, 413, 52 },
		{ 461, 414, 52 },
		{ 461, 415, 52 },
		{ 461, 416, 52 },
		{ 461, 417, 52 },
		{ 461, 418, 52 },
		{ 461, 419, 52 },
		{ 461, 420, 52 },
		{ 461, 421, 52 },
		{ 461, 422, 52 },
		{ 461, 423, 52 },
		{ 461, 424, 52 },
		{ 461, 425, 52 },
		{ 461, 426, 52 },
		{ 461, 427, 52 },
		{ 461, 428, 52 },
		{ 461, 429, 52 },
		{ 461, 430, 52 },
		{ 461, 431, 52 },
		{ 461, 432, 52 },
		{ 461, 433, 52 },
		{ 461, 434, 52 },
		{ 461, 435, 52 },
		{ 462, 386, 52 },
		{ 462, 387, 52 },
		{ 462, 388, 52 },
		{ 462, 389, 52 },
		{ 462, 390, 52 },
		{ 462, 391, 52 },
		{ 462, 392, 52 },
		{ 462, 393, 52 },
		{ 462, 394, 52 },
		{ 462, 395, 52 },
		{ 462, 396, 52 },
		{ 462, 397, 52 },
		{ 462, 398, 52 },
		{ 462, 399, 52 },
		{ 462, 400, 52 },
		{ 462, 401, 52 },
		{ 462, 402, 52 },
		{ 462, 403, 52 },
		{ 462, 404, 52 },
		{ 462, 405, 52 },
		{ 462, 406, 52 },
		{ 462, 407, 52 },
		{ 462, 408, 52 },
		{ 462, 409, 52 },
		{ 462, 410, 52 },
		{ 462, 411, 52 },
		{ 462, 412, 52 },
		{ 462, 413, 52 },
		{ 462, 414, 52 },
		{ 462, 415, 52 },
		{ 462, 416, 52 },
		{ 462, 417, 52 },
		{ 462, 418, 52 },
		{ 462, 419, 52 },
		{ 462, 420, 52 },
		{ 462, 421, 52 },
		{ 462, 422, 52 },
		{ 462, 423, 52 },
		{ 462, 424, 52 },
		{ 462, 425, 52 },
		{ 462, 426, 52 },
		{ 462, 427, 52 },
		{ 462, 428, 52 },
		{ 462, 429, 52 },
		{ 462, 430, 52 },
		{ 462, 431, 52 },
		{ 462, 432, 52 },
		{ 462, 433, 52 },
		{ 462, 434, 52 },
		{ 462, 435, 52 },
		{ 462, 436, 52 },
		{ 463, 386, 52 },
		{ 463, 387, 52 },
		{ 463, 388, 52 },
		{ 463, 389, 52 },
		{ 463, 390, 52 },
		{ 463, 391, 52 },
		{ 463, 392, 52 },
		{ 463, 393, 52 },
		{ 463, 394, 52 },
		{ 463, 395, 52 },
		{ 463, 396, 52 },
		{ 463, 397, 52 },
		{ 463, 398, 52 },
		{ 463, 399, 52 },
		{ 463, 400, 52 },
		{ 463, 401, 52 },
		{ 463, 402, 52 },
		{ 463, 403, 52 },
		{ 463, 404, 52 },
		{ 463, 405, 52 },
		{ 463, 406, 52 },
		{ 463, 407, 52 },
		{ 463, 408, 52 },
		{ 463, 409, 52 },
		{ 463, 410, 52 },
		{ 463, 411, 52 },
		{ 463, 412, 52 },
		{ 463, 413, 52 },
		{ 463, 414, 52 },
		{ 463, 415, 52 },
		{ 463, 416, 52 },
		{ 463, 417, 52 },
		{ 463, 418, 52 },
		{ 463, 419, 52 },
		{ 463, 420, 52 },
		{ 463, 421, 52 },
		{ 463, 422, 52 },
		{ 463, 423, 52 },
		{ 463, 424, 52 },
		{ 463, 425, 52 },
		{ 463, 426, 52 },
		{ 463, 427, 52 },
		{ 463, 428, 52 },
		{ 463, 429, 52 },
		{ 463, 430, 52 },
		{ 463, 431, 52 },
		{ 463, 432, 52 },
		{ 463, 433, 52 },
		{ 463, 434, 52 },
		{ 463, 435, 52 },
		{ 463, 436, 52 },
		{ 463, 437, 52 },
		{ 464, 387, 52 },
		{ 464, 388, 52 },
		{ 464, 389, 52 },
		{ 464, 390, 52 },
		{ 464, 391, 52 },
		{ 464, 392, 52 },
		{ 464, 393, 52 },
		{ 464, 394, 52 },
		{ 464, 395, 52 },
		{ 464, 396, 52 },
		{ 464, 397, 52 },
		{ 464, 398, 52 },
		{ 464, 399, 52 },
		{ 464, 400, 52 },
		{ 464, 401, 52 },
		{ 464, 402, 52 },
		{ 464, 403, 52 },
		{ 464, 404, 52 },
		{ 464, 405, 52 },
		{ 464, 406, 52 },
		{ 464, 407, 52 },
		{ 464, 408, 52 },
		{ 464, 409, 52 },
		{ 464, 410, 52 },
		{ 464, 411, 52 },
		{ 464, 412, 52 },
		{ 464, 413, 52 },
		{ 464, 414, 52 },
		{ 464, 415, 52 },
		{ 464, 416, 52 },
		{ 464, 417, 52 },
		{ 464, 418, 52 },
		{ 464, 419, 52 },
		{ 464, 420, 52 },
		{ 464, 421, 52 },
		{ 464, 422, 52 },
		{ 464, 423, 52 },
		{ 464, 424, 52 },
		{ 464, 425, 52 },
		{ 464, 426, 52 },
		{ 464, 427, 52 },
		{ 464, 428, 52 },
		{ 464, 429, 52 },
		{ 464, 430, 52 },
		{ 464, 431, 52 },
		{ 464, 432, 52 },
		{ 464, 433, 52 },
		{ 464, 434, 52 },
		{ 464, 435, 52 },
		{ 464, 436, 52 },
		{ 464, 437, 52 },
		{ 464, 438, 52 },
		{ 465, 388, 52 },
		{ 465, 389, 52 },
		{ 465, 390, 52 },
		{ 465, 391, 52 },
		{ 465, 392, 52 },
		{ 465, 393, 52 },
		{ 465, 394, 52 },
		{ 465, 395, 52 },
		{ 465, 396, 52 },
		{ 465, 397, 52 },
		{ 465, 398, 52 },
		{ 465, 399, 52 },
		{ 465, 400, 52 },
		{ 465, 401, 52 },
		{ 465, 402, 52 },
		{ 465, 403, 52 },
		{ 465, 404, 52 },
		{ 465, 405, 52 },
		{ 465, 406, 52 },
		{ 465, 407, 52 },
		{ 465, 408, 52 },
		{ 465, 409, 52 },
		{ 465, 410, 52 },
		{ 465, 411, 52 },
		{ 465, 412, 52 },
		{ 465, 413, 52 },
		{ 465, 414, 52 },
		{ 465, 415, 52 },
		{ 465, 416, 52 },
		{ 465, 417, 52 },
		{ 465, 418, 52 },
		{ 465, 419, 52 },
		{ 465, 420, 52 },
		{ 465, 421, 52 },
		{ 465, 422, 52 },
		{ 465, 423, 52 },
		{ 465, 424, 52 },
		{ 465, 425, 52 },
		{ 465, 426, 52 },
		{ 465, 427, 52 },
		{ 465, 428, 52 },
		{ 465, 429, 52 },
		{ 465, 430, 52 },
		{ 465, 431, 52 },
		{ 465, 432, 52 },
		{ 465, 433, 52 },
		{ 465, 434, 52 },
		{ 465, 435, 52 },
		{ 465, 436, 52 },
		{ 465, 437, 52 },
		{ 465, 438, 52 },
		{ 466, 388, 52 },
		{ 466, 389, 52 },
		{ 466, 390, 52 },
		{ 466, 391, 52 },
		{ 466, 392, 52 },
		{ 466, 393, 52 },
		{ 466, 394, 52 },
		{ 466, 395, 52 },
		{ 466, 396, 52 },
		{ 466, 397, 52 },
		{ 466, 398, 52 },
		{ 466, 399, 52 },
		{ 466, 400, 52 },
		{ 466, 401, 52 },
		{ 466, 402, 52 },
		{ 466, 403, 52 },
		{ 466, 404, 52 },
		{ 466, 405, 52 },
		{ 466, 406, 52 },
		{ 466, 407, 52 },
		{ 466, 408, 52 },
		{ 466, 409, 52 },
		{ 466, 410, 52 },
		{ 466, 411, 52 },
		{ 466, 412, 52 },
		{ 466, 413, 52 },
		{ 466, 414, 52 },
		{ 466, 415, 52 },
		{ 466, 416, 52 },
		{ 466, 417, 52 },
		{ 466, 418, 52 },
		{ 466, 419, 52 },
		{ 466, 420, 52 },
		{ 466, 421, 52 },
		{ 466, 422, 52 },
		{ 466, 423, 52 },
		{ 466, 424, 52 },
		{ 466, 425, 52 },
		{ 466, 426, 52 },
		{ 466, 427, 52 },
		{ 466, 428, 52 },
		{ 466, 429, 52 },
		{ 466, 430, 52 },
		{ 466, 431, 52 },
		{ 466, 432, 52 },
		{ 466, 433, 52 },
		{ 466, 434, 52 },
		{ 466, 435, 52 },
		{ 466, 436, 52 },
		{ 466, 437, 52 },
		{ 466, 438, 52 },
		{ 466, 439, 52 },
		{ 467, 389, 52 },
		{ 467, 390, 52 },
		{ 467, 391, 52 },
		{ 467, 392, 52 },
		{ 467, 393, 52 },
		{ 467, 394, 52 },
		{ 467, 395, 52 },
		{ 467, 396, 52 },
		{ 467, 397, 52 },
		{ 467, 398, 52 },
		{ 467, 399, 52 },
		{ 467, 400, 52 },
		{ 467, 401, 52 },
		{ 467, 402, 52 },
		{ 467, 403, 52 },
		{ 467, 404, 52 },
		{ 467, 405, 52 },
		{ 467, 406, 52 },
		{ 467, 407, 52 },
		{ 467, 408, 52 },
		{ 467, 409, 52 },
		{ 467, 410, 52 },
		{ 467, 411, 52 },
		{ 467, 412, 52 },
		{ 467, 413, 52 },
		{ 467, 414, 52 },
		{ 467, 415, 52 },
		{ 467, 416, 52 },
		{ 467, 417, 52 },
		{ 467, 418, 52 },
		{ 467, 419, 52 },
		{ 467, 420, 52 },
		{ 467, 421, 52 },
		{ 467, 422, 52 },
		{ 467, 423, 52 },
		{ 467, 424, 52 },
		{ 467, 425, 52 },
		{ 467, 426, 52 },
		{ 467, 427, 52 },
		{ 467, 428, 52 },
		{ 467, 429, 52 },
		{ 467, 430, 52 },
		{ 467, 431, 52 },
		{ 467, 432, 52 },
		{ 467, 433, 52 },
		{ 467, 434, 52 },
		{ 467, 435, 52 },
		{ 467, 436, 52 },
		{ 467, 437, 52 },
		{ 467, 438, 52 },
		{ 467, 439, 52 },
		{ 467, 440, 52 },
		{ 468, 389, 52 },
		{ 468, 390, 52 },
		{ 468, 391, 52 },
		{ 468, 392, 52 },
		{ 468, 393, 52 },
		{ 468, 394, 52 },
		{ 468, 395, 52 },
		{ 468, 396, 52 },
		{ 468, 397, 52 },
		{ 468, 398, 52 },
		{ 468, 399, 52 },
		{ 468, 400, 52 },
		{ 468, 401, 52 },
		{ 468, 402, 52 },
		{ 468, 403, 52 },
		{ 468, 404, 52 },
		{ 468, 405, 52 },
		{ 468, 406, 52 },
		{ 468, 407, 52 },
		{ 468, 408, 52 },
		{ 468, 409, 52 },
		{ 468, 410, 52 },
		{ 468, 411, 52 },
		{ 468, 412, 52 },
		{ 468, 413, 52 },
		{ 468, 414, 52 },
		{ 468, 415, 52 },
		{ 468, 416, 52 },
		{ 468, 417, 52 },
		{ 468, 418, 52 },
		{ 468, 419, 52 },
		{ 468, 420, 52 },
		{ 468, 421, 52 },
		{ 468, 422, 52 },
		{ 468, 423, 52 },
		{ 468, 424, 52 },
		{ 468, 425, 52 },
		{ 468, 426, 52 },
		{ 468, 427, 52 },
		{ 468, 428, 52 },
		{ 468, 429, 52 },
		{ 468, 430, 52 },
		{ 468, 431, 52 },
		{ 468, 432, 52 },
		{ 468, 433, 52 },
		{ 468, 434, 52 },
		{ 468, 435, 52 },
		{ 468, 436, 52 },
		{ 468, 437, 52 },
		{ 468, 438, 52 },
		{ 468, 439, 52 },
		{ 468, 440, 52 },
		{ 468, 441, 52 },
		{ 469, 390, 52 },
		{ 469, 391, 52 },
		{ 469, 392, 52 },
		{ 469, 393, 52 },
		{ 469, 394, 52 },
		{ 469, 395, 52 },
		{ 469, 396, 52 },
		{ 469, 397, 52 },
		{ 469, 398, 52 },
		{ 469, 399, 52 },
		{ 469, 400, 52 },
		{ 469, 401, 52 },
		{ 469, 402, 52 },
		{ 469, 403, 52 },
		{ 469, 404, 52 },
		{ 469, 405, 52 },
		{ 469, 406, 52 },
		{ 469, 407, 52 },
		{ 469, 408, 52 },
		{ 469, 409, 52 },
		{ 469, 410, 52 },
		{ 469, 411, 52 },
		{ 469, 412, 52 },
		{ 469, 413, 52 },
		{ 469, 414, 52 },
		{ 469, 415, 52 },
		{ 469, 416, 52 },
		{ 469, 417, 52 },
		{ 469, 418, 52 },
		{ 469, 419, 52 },
		{ 469, 420, 52 },
		{ 469, 421, 52 },
		{ 469, 422, 52 },
		{ 469, 423, 52 },
		{ 469, 424, 52 },
		{ 469, 425, 52 },
		{ 469, 426, 52 },
		{ 469, 427, 52 },
		{ 469, 428, 52 },
		{ 469, 429, 52 },
		{ 469, 430, 52 },
		{ 469, 431, 52 },
		{ 469, 432, 52 },
		{ 469, 433, 52 },
		{ 469, 434, 52 },
		{ 469, 435, 52 },
		{ 469, 436, 52 },
		{ 469, 437, 52 },
		{ 469, 438, 52 },
		{ 469, 439, 52 },
		{ 469, 440, 52 },
		{ 469, 441, 52 },
		{ 469, 442, 52 },
		{ 470, 391, 52 },
		{ 470, 392, 52 },
		{ 470, 393, 52 },
		{ 470, 394, 52 },
		{ 470, 395, 52 },
		{ 470, 396, 52 },
		{ 470, 397, 52 },
		{ 470, 398, 52 },
		{ 470, 399, 52 },
		{ 470, 400, 52 },
		{ 470, 401, 52 },
		{ 470, 402, 52 },
		{ 470, 403, 52 },
		{ 470, 404, 52 },
		{ 470, 405, 52 },
		{ 470, 406, 52 },
		{ 470, 407, 52 },
		{ 470, 408, 52 },
		{ 470, 409, 52 },
		{ 470, 410, 52 },
		{ 470, 411, 52 },
		{ 470, 412, 52 },
		{ 470, 413, 52 },
		{ 470, 414, 52 },
		{ 470, 415, 52 },
		{ 470, 416, 52 },
		{ 470, 417, 52 },
		{ 470, 418, 52 },
		{ 470, 419, 52 },
		{ 470, 420, 52 },
		{ 470, 421, 52 },
		{ 470, 422, 52 },
		{ 470, 423, 52 },
		{ 470, 424, 52 },
		{ 470, 425, 52 },
		{ 470, 426, 52 },
		{ 470, 427, 52 },
		{ 470, 428, 52 },
		{ 470, 429, 52 },
		{ 470, 430, 52 },
		{ 470, 431, 52 },
		{ 470, 432, 52 },
		{ 470, 433, 52 },
		{ 470, 434, 52 },
		{ 470, 435, 52 },
		{ 470, 436, 52 },
		{ 470, 437, 52 },
		{ 470, 438, 52 },
		{ 470, 439, 52 },
		{ 470, 440, 52 },
		{ 470, 441, 52 },
		{ 470, 442, 52 },
		{ 470, 443, 52 },
		{ 471, 391, 52 },
		{ 471, 392, 52 },
		{ 471, 393, 52 },
		{ 471, 394, 52 },
		{ 471, 395, 52 },
		{ 471, 396, 52 },
		{ 471, 397, 52 },
		{ 471, 398, 52 },
		{ 471, 399, 52 },
		{ 471, 400, 52 },
		{ 471, 401, 52 },
		{ 471, 402, 52 },
		{ 471, 403, 52 },
		{ 471, 404, 52 },
		{ 471, 405, 52 },
		{ 471, 406, 52 },
		{ 471, 407, 52 },
		{ 471, 408, 52 },
		{ 471, 409, 52 },
		{ 471, 410, 52 },
		{ 471, 411, 52 },
		{ 471, 412, 52 },
		{ 471, 413, 52 },
		{ 471, 414, 52 },
		{ 471, 415, 52 },
		{ 471, 416, 52 },
		{ 471, 417, 52 },
		{ 471, 418, 52 },
		{ 471, 419, 52 },
		{ 471, 420, 52 },
		{ 471, 421, 52 },
		{ 471, 422, 52 },
		{ 471, 423, 52 },
		{ 471, 424, 52 },
		{ 471, 425, 52 },
		{ 471, 426, 52 },
		{ 471, 427, 52 },
		{ 471, 428, 52 },
		{ 471, 429, 52 },
		{ 471, 430, 52 },
		{ 471, 431, 52 },
		{ 471, 432, 52 },
		{ 471, 433, 52 },
		{ 471, 434, 52 },
		{ 471, 435, 52 },
		{ 471, 436, 52 },
		{ 471, 437, 52 },
		{ 471, 438, 52 },
		{ 471, 439, 52 },
		{ 471, 440, 52 },
		{ 471, 441, 52 },
		{ 471, 442, 52 },
		{ 471, 443, 52 },
		{ 471, 444, 52 },
		{ 472, 392, 52 },
		{ 472, 393, 52 },
		{ 472, 394, 52 },
		{ 472, 395, 52 },
		{ 472, 396, 52 },
		{ 472, 397, 52 },
		{ 472, 398, 52 },
		{ 472, 399, 52 },
		{ 472, 400, 52 },
		{ 472, 401, 52 },
		{ 472, 402, 52 },
		{ 472, 403, 52 },
		{ 472, 404, 52 },
		{ 472, 405, 52 },
		{ 472, 406, 52 },
		{ 472, 407, 52 },
		{ 472, 408, 52 },
		{ 472, 409, 52 },
		{ 472, 410, 52 },
		{ 472, 411, 52 },
		{ 472, 412, 52 },
		{ 472, 413, 52 },
		{ 472, 414, 52 },
		{ 472, 415, 52 },
		{ 472, 416, 52 },
		{ 472, 417, 52 },
		{ 472, 418, 52 },
		{ 472, 419, 52 },
		{ 472, 420, 52 },
		{ 472, 421, 52 },
		{ 472, 422, 52 },
		{ 472, 423, 52 },
		{ 472, 424, 52 },
		{ 472, 425, 52 },
		{ 472, 426, 52 },
		{ 472, 427, 52 },
		{ 472, 428, 52 },
		{ 472, 429, 52 },
		{ 472, 430, 52 },
		{ 472, 431, 52 },
		{ 472, 432, 52 },
		{ 472, 433, 52 },
		{ 472, 434, 52 },
		{ 472, 435, 52 },
		{ 472, 436, 52 },
		{ 472, 437, 52 },
		{ 472, 438, 52 },
		{ 472, 439, 52 },
		{ 472, 440, 52 },
		{ 472, 441, 52 },
		{ 472, 442, 52 },
		{ 472, 443, 52 },
		{ 472, 444, 52 },
		{ 472, 445, 52 },
		{ 473, 393, 52 },
		{ 473, 394, 52 },
		{ 473, 395, 52 },
		{ 473, 396, 52 },
		{ 473, 397, 52 },
		{ 473, 398, 52 },
		{ 473, 399, 52 },
		{ 473, 400, 52 },
		{ 473, 401, 52 },
		{ 473, 402, 52 },
		{ 473, 403, 52 },
		{ 473, 404, 52 },
		{ 473, 405, 52 },
		{ 473, 406, 52 },
		{ 473, 407, 52 },
		{ 473, 408, 52 },
		{ 473, 409, 52 },
		{ 473, 410, 52 },
		{ 473, 411, 52 },
		{ 473, 412, 52 },
		{ 473, 413, 52 },
		{ 473, 414, 52 },
		{ 473, 415, 52 },
		{ 473, 416, 52 },
		{ 473, 417, 52 },
		{ 473, 418, 52 },
		{ 473, 419, 52 },
		{ 473, 420, 52 },
		{ 473, 421, 52 },
		{ 473, 422, 52 },
		{ 473, 423, 52 },
		{ 473, 424, 52 },
		{ 473, 425, 52 },
		{ 473, 426, 52 },
		{ 473, 427, 52 },
		{ 473, 428, 52 },
		{ 473, 429, 52 },
		{ 473, 430, 52 },
		{ 473, 431, 52 },
		{ 473, 432, 52 },
		{ 473, 433, 52 },
		{ 473, 434, 52 },
		{ 473, 435, 52 },
		{ 473, 436, 52 },
		{ 473, 437, 52 },
		{ 473, 438, 52 },
		{ 473, 439, 52 },
		{ 473, 440, 52 },
		{ 473, 441, 52 },
		{ 473, 442, 52 },
		{ 473, 443, 52 },
		{ 473, 444, 52 },
		{ 473, 445, 52 },
		{ 474, 393, 52 },
		{ 474, 394, 52 },
		{ 474, 395, 52 },
		{ 474, 396, 52 },
		{ 474, 397, 52 },
		{ 474, 398, 52 },
		{ 474, 399, 52 },
		{ 474, 400, 52 },
		{ 474, 401, 52 },
		{ 474, 402, 52 },
		{ 474, 403, 52 },
		{ 474, 404, 52 },
		{ 474, 405, 52 },
		{ 474, 406, 52 },
		{ 474, 407, 52 },
		{ 474, 408, 52 },
		{ 474, 409, 52 },
		{ 474, 410, 52 },
		{ 474, 411, 52 },
		{ 474, 412, 52 },
		{ 474, 413, 52 },
		{ 474, 414, 52 },
		{ 474, 415, 52 },
		{ 474, 416, 52 },
		{ 474, 417, 52 },
		{ 474, 418, 52 },
		{ 474, 419, 52 },
		{ 474, 420, 52 },
		{ 474, 421, 52 },
		{ 474, 422, 52 },
		{ 474, 423, 52 },
		{ 474, 424, 52 },
		{ 474, 425, 52 },
		{ 474, 426, 52 },
		{ 474, 427, 52 },
		{ 474, 428, 52 },
		{ 474, 429, 52 },
		{ 474, 430, 52 },
		{ 474, 431, 52 },
		{ 474, 432, 52 },
		{ 474, 433, 52 },
		{ 474, 434, 52 },
		{ 474, 435, 52 },
		{ 474, 436, 52 },
		{ 474, 437, 52 },
		{ 474, 438, 52 },
		{ 474, 439, 52 },
		{ 474, 440, 52 },
		{ 474, 441, 52 },
		{ 474, 442, 52 },
		{ 474, 443, 52 },
		{ 474, 444, 52 },
		{ 474, 445, 52 },
		{ 474, 446, 52 },
		{ 475, 394, 52 },
		{ 475, 395, 52 },
		{ 475, 396, 52 },
		{ 475, 397, 52 },
		{ 475, 398, 52 },
		{ 475, 399, 52 },
		{ 475, 400, 52 },
		{ 475, 401, 52 },
		{ 475, 402, 52 },
		{ 475, 403, 52 },
		{ 475, 404, 52 },
		{ 475, 405, 52 },
		{ 475, 406, 52 },
		{ 475, 407, 52 },
		{ 475, 408, 52 },
		{ 475, 409, 52 },
		{ 475, 410, 52 },
		{ 475, 411, 52 },
		{ 475, 412, 52 },
		{ 475, 413, 52 },
		{ 475, 414, 52 },
		{ 475, 415, 52 },
		{ 475, 416, 52 },
		{ 475, 417, 52 },
		{ 475, 418, 52 },
		{ 475, 419, 52 },
		{ 475, 420, 52 },
		{ 475, 421, 52 },
		{ 475, 422, 52 },
		{ 475, 423, 52 },
		{ 475, 424, 52 },
		{ 475, 425, 52 },
		{ 475, 426, 52 },
		{ 475, 427, 52 },
		{ 475, 428, 52 },
		{ 475, 429, 52 },
		{ 475, 430, 52 },
		{ 475, 431, 52 },
		{ 475, 432, 52 },
		{ 475, 433, 52 },
		{ 475, 434, 52 },
		{ 475, 435, 52 },
		{ 475, 436, 52 },
		{ 475, 437, 52 },
		{ 475, 438, 52 },
		{ 475, 439, 52 },
		{ 475, 440, 52 },
		{ 475, 441, 52 },
		{ 475, 442, 52 },
		{ 475, 443, 52 },
		{ 475, 444, 52 },
		{ 475, 445, 52 },
		{ 475, 446, 52 },
		{ 475, 447, 52 },
		{ 476, 394, 52 },
		{ 476, 395, 52 },
		{ 476, 396, 52 },
		{ 476, 397, 52 },
		{ 476, 398, 52 },
		{ 476, 399, 52 },
		{ 476, 400, 52 },
		{ 476, 401, 52 },
		{ 476, 402, 52 },
		{ 476, 403, 52 },
		{ 476, 404, 52 },
		{ 476, 405, 52 },
		{ 476, 406, 52 },
		{ 476, 407, 52 },
		{ 476, 408, 52 },
		{ 476, 409, 52 },
		{ 476, 410, 52 },
		{ 476, 411, 52 },
		{ 476, 412, 52 },
		{ 476, 413, 52 },
		{ 476, 414, 52 },
		{ 476, 415, 52 },
		{ 476, 416, 52 },
		{ 476, 417, 52 },
		{ 476, 418, 52 },
		{ 476, 419, 52 },
		{ 476, 420, 52 },
		{ 476, 421, 52 },
		{ 476, 422, 52 },
		{ 476, 423, 52 },
		{ 476, 424, 52 },
		{ 476, 425, 52 },
		{ 476, 426, 52 },
		{ 476, 427, 52 },
		{ 476, 428, 52 },
		{ 476, 429, 52 },
		{ 476, 430, 52 },
		{ 476, 431, 52 },
		{ 476, 432, 52 },
		{ 476, 433, 52 },
		{ 476, 434, 52 },
		{ 476, 435, 52 },
		{ 476, 436, 52 },
		{ 476, 437, 52 },
		{ 476, 438, 52 },
		{ 476, 439, 52 },
		{ 476, 440, 52 },
		{ 476, 441, 52 },
		{ 476, 442, 52 },
		{ 476, 443, 52 },
		{ 476, 444, 52 },
		{ 476, 445, 52 },
		{ 476, 446, 52 },
		{ 476, 447, 52 },
		{ 476, 448, 52 },
		{ 477, 395, 52 },
		{ 477, 396, 52 },
		{ 477, 397, 52 },
		{ 477, 398, 52 },
		{ 477, 399, 52 },
		{ 477, 400, 52 },
		{ 477, 401, 52 },
		{ 477, 402, 52 },
		{ 477, 403, 52 },
		{ 477, 404, 52 },
		{ 477, 405, 52 },
		{ 477, 406, 52 },
		{ 477, 407, 52 },
		{ 477, 408, 52 },
		{ 477, 409, 52 },
		{ 477, 410, 52 },
		{ 477, 411, 52 },
		{ 477, 412, 52 },
		{ 477, 413, 52 },
		{ 477, 414, 52 },
		{ 477, 415, 52 },
		{ 477, 416, 52 },
		{ 477, 417, 52 },
		{ 477, 418, 52 },
		{ 477, 419, 52 },
		{ 477, 420, 52 },
		{ 477, 421, 52 },
		{ 477, 422, 52 },
		{ 477, 423, 52 },
		{ 477, 424, 52 },
		{ 477, 425, 52 },
		{ 477, 426, 52 },
		{ 477, 427, 52 },
		{ 477, 428, 52 },
		{ 477, 429, 52 },
		{ 477, 430, 52 },
		{ 477, 431, 52 },
		{ 477, 432, 52 },
		{ 477, 433, 52 },
		{ 477, 434, 52 },
		{ 477, 435, 52 },
		{ 477, 436, 52 },
		{ 477, 437, 52 },
		{ 477, 438, 52 },
		{ 477, 439, 52 },
		{ 477, 440, 52 },
		{ 477, 441, 52 },
		{ 477, 442, 52 },
		{ 477, 443, 52 },
		{ 477, 444, 52 },
		{ 477, 445, 52 },
		{ 477, 446, 52 },
		{ 477, 447, 52 },
		{ 477, 448, 52 },
		{ 477, 449, 52 },
		{ 478, 396, 52 },
		{ 478, 397, 52 },
		{ 478, 398, 52 },
		{ 478, 399, 52 },
		{ 478, 400, 52 },
		{ 478, 401, 52 },
		{ 478, 402, 52 },
		{ 478, 403, 52 },
		{ 478, 404, 52 },
		{ 478, 405, 52 },
		{ 478, 406, 52 },
		{ 478, 407, 52 },
		{ 478, 408, 52 },
		{ 478, 409, 52 },
		{ 478, 410, 52 },
		{ 478, 411, 52 },
		{ 478, 412, 52 },
		{ 478, 413, 52 },
		{ 478, 414, 52 },
		{ 478, 415, 52 },
		{ 478, 416, 52 },
		{ 478, 417, 52 },
		{ 478, 418, 52 },
		{ 478, 419, 52 },
		{ 478, 420, 52 },
		{ 478, 421, 52 },
		{ 478, 422, 52 },
		{ 478, 423, 52 },
		{ 478, 424, 52 },
		{ 478, 425, 52 },
		{ 478, 426, 52 },
		{ 478, 427, 52 },
		{ 478, 428, 52 },
		{ 478, 429, 52 },
		{ 478, 430, 52 },
		{ 478, 431, 52 },
		{ 478, 432, 52 },
		{ 478, 433, 52 },
		{ 478, 434, 52 },
		{ 478, 435, 52 },
		{ 478, 436, 52 },
		{ 478, 437, 52 },
		{ 478, 438, 52 },
		{ 478, 439, 52 },
		{ 478, 440, 52 },
		{ 478, 441, 52 },
		{ 478, 442, 52 },
		{ 478, 443, 52 },
		{ 478, 444, 52 },
		{ 478, 445, 52 },
		{ 478, 446, 52 },
		{ 478, 447, 52 },
		{ 478, 448, 52 },
		{ 478, 449, 52 },
		{ 478, 450, 52 },
		{ 479, 396, 52 },
		{ 479, 397, 52 },
		{ 479, 398, 52 },
		{ 479, 399, 52 },
		{ 479, 400, 52 },
		{ 479, 401, 52 },
		{ 479, 402, 52 },
		{ 479, 403, 52 },
		{ 479, 404, 52 },
		{ 479, 405, 52 },
		{ 479, 406, 52 },
		{ 479, 407, 52 },
		{ 479, 408, 52 },
		{ 479, 409, 52 },
		{ 479, 410, 52 },
		{ 479, 411, 52 },
		{ 479, 412, 52 },
		{ 479, 413, 52 },
		{ 479, 414, 52 },
		{ 479, 415, 52 },
		{ 479, 416, 52 },
		{ 479, 417, 52 },
		{ 479, 418, 52 },
		{ 479, 419, 52 },
		{ 479, 420, 52 },
		{ 479, 421, 52 },
		{ 479, 422, 52 },
		{ 479, 423, 52 },
		{ 479, 424, 52 },
		{ 479, 425, 52 },
		{ 479, 426, 52 },
		{ 479, 427, 52 },
		{ 479, 428, 52 },
		{ 479, 429, 52 },
		{ 479, 430, 52 },
		{ 479, 431, 52 },
		{ 479, 432, 52 },
		{ 479, 433, 52 },
		{ 479, 434, 52 },
		{ 479, 435, 52 },
		{ 479, 436, 52 },
		{ 479, 437, 52 },
		{ 479, 438, 52 },
		{ 479, 439, 52 },
		{ 479, 440, 52 },
		{ 479, 441, 52 },
		{ 479, 442, 52 },
		{ 479, 443, 52 },
		{ 479, 444, 52 },
		{ 479, 445, 52 },
		{ 479, 446, 52 },
		{ 479, 447, 52 },
		{ 479, 448, 52 },
		{ 479, 449, 52 },
		{ 479, 450, 52 },
		{ 479, 451, 52 },
		{ 480, 397, 52 },
		{ 480, 398, 52 },
		{ 480, 399, 52 },
		{ 480, 400, 52 },
		{ 480, 401, 52 },
		{ 480, 402, 52 },
		{ 480, 403, 52 },
		{ 480, 404, 52 },
		{ 480, 405, 52 },
		{ 480, 406, 52 },
		{ 480, 407, 52 },
		{ 480, 408, 52 },
		{ 480, 409, 52 },
		{ 480, 410, 52 },
		{ 480, 411, 52 },
		{ 480, 412, 52 },
		{ 480, 413, 52 },
		{ 480, 414, 52 },
		{ 480, 415, 52 },
		{ 480, 416, 52 },
		{ 480, 417, 52 },
		{ 480, 418, 52 },
		{ 480, 419, 52 },
		{ 480, 420, 52 },
		{ 480, 421, 52 },
		{ 480, 422, 52 },
		{ 480, 423, 52 },
		{ 480, 424, 52 },
		{ 480, 425, 52 },
		{ 480, 426, 52 },
		{ 480, 427, 52 },
		{ 480, 428, 52 },
		{ 480, 429, 52 },
		{ 480, 430, 52 },
		{ 480, 431, 52 },
		{ 480, 432, 52 },
		{ 480, 433, 52 },
		{ 480, 434, 52 },
		{ 480, 435, 52 },
		{ 480, 436, 52 },
		{ 480, 437, 52 },
		{ 480, 438, 52 },
		{ 480, 439, 52 },
		{ 480, 440, 52 },
		{ 480, 441, 52 },
		{ 480, 442, 52 },
		{ 480, 443, 52 },
		{ 480, 444, 52 },
		{ 480, 445, 52 },
		{ 480, 446, 52 },
		{ 480, 447, 52 },
		{ 480, 448, 52 },
		{ 480, 449, 52 },
		{ 480, 450, 52 },
		{ 480, 451, 52 },
		{ 480, 452, 52 },
		{ 481, 398, 52 },
		{ 481, 399, 52 },
		{ 481, 400, 52 },
		{ 481, 401, 52 },
		{ 481, 402, 52 },
		{ 481, 403, 52 },
		{ 481, 404, 52 },
		{ 481, 405, 52 },
		{ 481, 406, 52 },
		{ 481, 407, 52 },
		{ 481, 408, 52 },
		{ 481, 409, 52 },
		{ 481, 410, 52 },
		{ 481, 411, 52 },
		{ 481, 412, 52 },
		{ 481, 413, 52 },
		{ 481, 414, 52 },
		{ 481, 415, 52 },
		{ 481, 416, 52 },
		{ 481, 417, 52 },
		{ 481, 418, 52 },
		{ 481, 419, 52 },
		{ 481, 420, 52 },
		{ 481, 421, 52 },
		{ 481, 422, 52 },
		{ 481, 423, 52 },
		{ 481, 424, 52 },
		{ 481, 425, 52 },
		{ 481, 426, 52 },
		{ 481, 427, 52 },
		{ 481, 428, 52 },
		{ 481, 429, 52 },
		{ 481, 430, 52 },
		{ 481, 431, 52 },
		{ 481, 432, 52 },
		{ 481, 433, 52 },
		{ 481, 434, 52 },
		{ 481, 435, 52 },
		{ 481, 436, 52 },
		{ 481, 437, 52 },
		{ 481, 438, 52 },
		{ 481, 439, 52 },
		{ 481, 440, 52 },
		{ 481, 441, 52 },
		{ 481, 442, 52 },
		{ 481, 443, 52 },
		{ 481, 444, 52 },
		{ 481, 445, 52 },
		{ 481, 446, 52 },
		{ 481, 447, 52 },
		{ 481, 448, 52 },
		{ 481, 449, 52 },
		{ 481, 450, 52 },
		{ 481, 451, 52 },
		{ 481, 452, 52 },
		{ 482, 398, 52 },
		{ 482, 399, 52 },
		{ 482, 400, 52 },
		{ 482, 401, 52 },
		{ 482, 402, 52 },
		{ 482, 403, 52 },
		{ 482, 404, 52 },
		{ 482, 405, 52 },
		{ 482, 406, 52 },
		{ 482, 407, 52 },
		{ 482, 408, 52 },
		{ 482, 409, 52 },
		{ 482, 410, 52 },
		{ 482, 411, 52 },
		{ 482, 412, 52 },
		{ 482, 413, 52 },
		{ 482, 414, 52 },
		{ 482, 415, 52 },
		{ 482, 416, 52 },
		{ 482, 417, 52 },
		{ 482, 418, 52 },
		{ 482, 419, 52 },
		{ 482, 420, 52 },
		{ 482, 421, 52 },
		{ 482, 422, 52 },
		{ 482, 423, 52 },
		{ 482, 424, 52 },
		{ 482, 425, 52 },
		{ 482, 426, 52 },
		{ 482, 427, 52 },
		{ 482, 428, 52 },
		{ 482, 429, 52 },
		{ 482, 430, 52 },
		{ 482, 431, 52 },
		{ 482, 432, 52 },
		{ 482, 433, 52 },
		{ 482, 434, 52 },
		{ 482, 435, 52 },
		{ 482, 436, 52 },
		{ 482, 437, 52 },
		{ 482, 438, 52 },
		{ 482, 439, 52 },
		{ 482, 440, 52 },
		{ 482, 441, 52 },
		{ 482, 442, 52 },
		{ 482, 443, 52 },
		{ 482, 444, 52 },
		{ 482, 445, 52 },
		{ 482, 446, 52 },
		{ 482, 447, 52 },
		{ 482, 448, 52 },
		{ 482, 449, 52 },
		{ 482, 450, 52 },
		{ 482, 451, 52 },
		{ 482, 452, 52 },
		{ 482, 453, 52 },
		{ 483, 399, 52 },
		{ 483, 400, 52 },
		{ 483, 401, 52 },
		{ 483, 402, 52 },
		{ 483, 403, 52 },
		{ 483, 404, 52 },
		{ 483, 405, 52 },
		{ 483, 406, 52 },
		{ 483, 407, 52 },
		{ 483, 408, 52 },
		{ 483, 409, 52 },
		{ 483, 410, 52 },
		{ 483, 411, 52 },
		{ 483, 412, 52 },
		{ 483, 413, 52 },
		{ 483, 414, 52 },
		{ 483, 415, 52 },
		{ 483, 416, 52 },
		{ 483, 417, 52 },
		{ 483, 418, 52 },
		{ 483, 419, 52 },
		{ 483, 420, 52 },
		{ 483, 421, 52 },
		{ 483, 422, 52 },
		{ 483, 423, 52 },
		{ 483, 424, 52 },
		{ 483, 425, 52 },
		{ 483, 426, 52 },
		{ 483, 427, 52 },
		{ 483, 428, 52 },
		{ 483, 429, 52 },
		{ 483, 430, 52 },
		{ 483, 431, 52 },
		{ 483, 432, 52 },
		{ 483, 433, 52 },
		{ 483, 434, 52 },
		{ 483, 435, 52 },
		{ 483, 436, 52 },
		{ 483, 437, 52 },
		{ 483, 438, 52 },
		{ 483, 439, 52 },
		{ 483, 440, 52 },
		{ 483, 441, 52 },
		{ 483, 442, 52 },
		{ 483, 443, 52 },
		{ 483, 444, 52 },
		{ 483, 445, 52 },
		{ 483, 446, 52 },
		{ 483, 447, 52 },
		{ 483, 448, 52 },
		{ 483, 449, 52 },
		{ 483, 450, 52 },
		{ 483, 451, 52 },
		{ 483, 452, 52 },
		{ 483, 453, 52 },
		{ 483, 454, 52 },
		{ 484, 399, 52 },
		{ 484, 400, 52 },
		{ 484, 401, 52 },
		{ 484, 402, 52 },
		{ 484, 403, 52 },
		{ 484, 404, 52 },
		{ 484, 405, 52 },
		{ 484, 406, 52 },
		{ 484, 407, 52 },
		{ 484, 408, 52 },
		{ 484, 409, 52 },
		{ 484, 410, 52 },
		{ 484, 411, 52 },
		{ 484, 412, 52 },
		{ 484, 413, 52 },
		{ 484, 414, 52 },
		{ 484, 415, 52 },
		{ 484, 416, 52 },
		{ 484, 417, 52 },
		{ 484, 418, 52 },
		{ 484, 419, 52 },
		{ 484, 420, 52 },
		{ 484, 421, 52 },
		{ 484, 422, 52 },
		{ 484, 423, 52 },
		{ 484, 424, 52 },
		{ 484, 425, 52 },
		{ 484, 426, 52 },
		{ 484, 427, 52 },
		{ 484, 428, 52 },
		{ 484, 429, 52 },
		{ 484, 430, 52 },
		{ 484, 431, 52 },
		{ 484, 432, 52 },
		{ 484, 433, 52 },
		{ 484, 434, 52 },
		{ 484, 435, 52 },
		{ 484, 436, 52 },
		{ 484, 437, 52 },
		{ 484, 438, 52 },
		{ 484, 439, 52 },
		{ 484, 440, 52 },
		{ 484, 441, 52 },
		{ 484, 442, 52 },
		{ 484, 443, 52 },
		{ 484, 444, 52 },
		{ 484, 445, 52 },
		{ 484, 446, 52 },
		{ 484, 447, 52 },
		{ 484, 448, 52 },
		{ 484, 449, 52 },
		{ 484, 450, 52 },
		{ 484, 451, 52 },
		{ 484, 452, 52 },
		{ 484, 453, 52 },
		{ 484, 454, 52 },
		{ 484, 455, 52 },
		{ 485, 400, 52 },
		{ 485, 401, 52 },
		{ 485, 402, 52 },
		{ 485, 403, 52 },
		{ 485, 404, 52 },
		{ 485, 405, 52 },
		{ 485, 406, 52 },
		{ 485, 407, 52 },
		{ 485, 408, 52 },
		{ 485, 409, 52 },
		{ 485, 410, 52 },
		{ 485, 411, 52 },
		{ 485, 412, 52 },
		{ 485, 413, 52 },
		{ 485, 414, 52 },
		{ 485, 415, 52 },
		{ 485, 416, 52 },
		{ 485, 417, 52 },
		{ 485, 418, 52 },
		{ 485, 419, 52 },
		{ 485, 420, 52 },
		{ 485, 421, 52 },
		{ 485, 422, 52 },
		{ 485, 423, 52 },
		{ 485, 424, 52 },
		{ 485, 425, 52 },
		{ 485, 426, 52 },
		{ 485, 427, 52 },
		{ 485, 428, 52 },
		{ 485, 429, 52 },
		{ 485, 430, 52 },
		{ 485, 431, 52 },
		{ 485, 432, 52 },
		{ 485, 433, 52 },
		{ 485, 434, 52 },
		{ 485, 435, 52 },
		{ 485, 436, 52 },
		{ 485, 437, 52 },
		{ 485, 438, 52 },
		{ 485, 439, 52 },
		{ 485, 440, 52 },
		{ 485, 441, 52 },
		{ 485, 442, 52 },
		{ 485, 443, 52 },
		{ 485, 444, 52 },
		{ 485, 445, 52 },
		{ 485, 446, 52 },
		{ 485, 447, 52 },
		{ 485, 448, 52 },
		{ 485, 449, 52 },
		{ 485, 450, 52 },
		{ 485, 451, 52 },
		{ 485, 452, 52 },
		{ 485, 453, 52 },
		{ 485, 454, 52 },
		{ 485, 455, 52 },
		{ 485, 456, 52 },
		{ 486, 401, 52 },
		{ 486, 402, 52 },
		{ 486, 403, 52 },
		{ 486, 404, 52 },
		{ 486, 405, 52 },
		{ 486, 406, 52 },
		{ 486, 407, 52 },
		{ 486, 408, 52 },
		{ 486, 409, 52 },
		{ 486, 410, 52 },
		{ 486, 411, 52 },
		{ 486, 412, 52 },
		{ 486, 413, 52 },
		{ 486, 414, 52 },
		{ 486, 415, 52 },
		{ 486, 416, 52 },
		{ 486, 417, 52 },
		{ 486, 418, 52 },
		{ 486, 419, 52 },
		{ 486, 420, 52 },
		{ 486, 421, 52 },
		{ 486, 422, 52 },
		{ 486, 423, 52 },
		{ 486, 424, 52 },
		{ 486, 425, 52 },
		{ 486, 426, 52 },
		{ 486, 427, 52 },
		{ 486, 428, 52 },
		{ 486, 429, 52 },
		{ 486, 430, 52 },
		{ 486, 431, 52 },
		{ 486, 432, 52 },
		{ 486, 433, 52 },
		{ 486, 434, 52 },
		{ 486, 435, 52 },
		{ 486, 436, 52 },
		{ 486, 437, 52 },
		{ 486, 438, 52 },
		{ 486, 439, 52 },
		{ 486, 440, 52 },
		{ 486, 441, 52 },
		{ 486, 442, 52 },
		{ 486, 443, 52 },
		{ 486, 444, 52 },
		{ 486, 445, 52 },
		{ 486, 446, 52 },
		{ 486, 447, 52 },
		{ 486, 448, 52 },
		{ 486, 449, 52 },
		{ 486, 450, 52 },
		{ 486, 451, 52 },
		{ 486, 452, 52 },
		{ 486, 453, 52 },
		{ 486, 454, 52 },
		{ 486, 455, 52 },
		{ 486, 456, 52 },
		{ 486, 457, 52 },
		{ 487, 401, 52 },
		{ 487, 402, 52 },
		{ 487, 403, 52 },
		{ 487, 404, 52 },
		{ 487, 405, 52 },
		{ 487, 406, 52 },
		{ 487, 407, 52 },
		{ 487, 408, 52 },
		{ 487, 409, 52 },
		{ 487, 410, 52 },
		{ 487, 411, 52 },
		{ 487, 412, 52 },
		{ 487, 413, 52 },
		{ 487, 414, 52 },
		{ 487, 415, 52 },
		{ 487, 416, 52 },
		{ 487, 417, 52 },
		{ 487, 418, 52 },
		{ 487, 419, 52 },
		{ 487, 420, 52 },
		{ 487, 421, 52 },
		{ 487, 422, 52 },
		{ 487, 423, 52 },
		{ 487, 424, 52 },
		{ 487, 425, 52 },
		{ 487, 426, 52 },
		{ 487, 427, 52 },
		{ 487, 428, 52 },
		{ 487, 429, 52 },
		{ 487, 430, 52 },
		{ 487, 431, 52 },
		{ 487, 432, 52 },
		{ 487, 433, 52 },
		{ 487, 434, 52 },
		{ 487, 435, 52 },
		{ 487, 436, 52 },
		{ 487, 437, 52 },
		{ 487, 438, 52 },
		{ 487, 439, 52 },
		{ 487, 440, 52 },
		{ 487, 441, 52 },
		{ 487, 442, 52 },
		{ 487, 443, 52 },
		{ 487, 444, 52 },
		{ 487, 445, 52 },
		{ 487, 446, 52 },
		{ 487, 447, 52 },
		{ 487, 448, 52 },
		{ 487, 449, 52 },
		{ 487, 450, 52 },
		{ 487, 451, 52 },
		{ 487, 452, 52 },
		{ 487, 453, 52 },
		{ 487, 454, 52 },
		{ 487, 455, 52 },
		{ 487, 456, 52 },
		{ 487, 457, 52 },
		{ 487, 458, 52 },
		{ 488, 402, 52 },
		{ 488, 403, 52 },
		{ 488, 404, 52 },
		{ 488, 405, 52 },
		{ 488, 406, 52 },
		{ 488, 407, 52 },
		{ 488, 408, 52 },
		{ 488, 409, 52 },
		{ 488, 410, 52 },
		{ 488, 411, 52 },
		{ 488, 412, 52 },
		{ 488, 413, 52 },
		{ 488, 414, 52 },
		{ 488, 415, 52 },
		{ 488, 416, 52 },
		{ 488, 417, 52 },
		{ 488, 418, 52 },
		{ 488, 419, 52 },
		{ 488, 420, 52 },
		{ 488, 421, 52 },
		{ 488, 422, 52 },
		{ 488, 423, 52 },
		{ 488, 424, 52 },
		{ 488, 425, 52 },
		{ 488, 426, 52 },
		{ 488, 427, 52 },
		{ 488, 428, 52 },
		{ 488, 429, 52 },
		{ 488, 430, 52 },
		{ 488, 431, 52 },
		{ 488, 432, 52 },
		{ 488, 433, 52 },
		{ 488, 434, 52 },
		{ 488, 435, 52 },
		{ 488, 436, 52 },
		{ 488, 437, 52 },
		{ 488, 438, 52 },
		{ 488, 439, 52 },
		{ 488, 440, 52 },
		{ 488, 441, 52 },
		{ 488, 442, 52 },
		{ 488, 443, 52 },
		{ 488, 444, 52 },
		{ 488, 445, 52 },
		{ 488, 446, 52 },
		{ 488, 447, 52 },
		{ 488, 448, 52 },
		{ 488, 449, 52 },
		{ 488, 450, 52 },
		{ 488, 451, 52 },
		{ 488, 452, 52 },
		{ 488, 453, 52 },
		{ 488, 454, 52 },
		{ 488, 455, 52 },
		{ 488, 456, 52 },
		{ 488, 457, 52 },
		{ 488, 458, 52 },
		{ 488, 459, 52 },
		{ 489, 403, 52 },
		{ 489, 404, 52 },
		{ 489, 405, 52 },
		{ 489, 406, 52 },
		{ 489, 407, 52 },
		{ 489, 408, 52 },
		{ 489, 409, 52 },
		{ 489, 410, 52 },
		{ 489, 411, 52 },
		{ 489, 412, 52 },
		{ 489, 413, 52 },
		{ 489, 414, 52 },
		{ 489, 415, 52 },
		{ 489, 416, 52 },
		{ 489, 417, 52 },
		{ 489, 418, 52 },
		{ 489, 419, 52 },
		{ 489, 420, 52 },
		{ 489, 421, 52 },
		{ 489, 422, 52 },
		{ 489, 423, 52 },
		{ 489, 424, 52 },
		{ 489, 425, 52 },
		{ 489, 426, 52 },
		{ 489, 427, 52 },
		{ 489, 428, 52 },
		{ 489, 429, 52 },
		{ 489, 430, 52 },
		{ 489, 431, 52 },
		{ 489, 432, 52 },
		{ 489, 433, 52 },
		{ 489, 434, 52 },
		{ 489, 435, 52 },
		{ 489, 436, 52 },
		{ 489, 437, 52 },
		{ 489, 438, 52 },
		{ 489, 439, 52 },
		{ 489, 440, 52 },
		{ 489, 441, 52 },
		{ 489, 442, 52 },
		{ 489, 443, 52 },
		{ 489, 444, 52 },
		{ 489, 445, 52 },
		{ 489, 446, 52 },
		{ 489, 447, 52 },
		{ 489, 448, 52 },
		{ 489, 449, 52 },
		{ 489, 450, 52 },
		{ 489, 451, 52 },
		{ 489, 452, 52 },
		{ 489, 453, 52 },
		{ 489, 454, 52 },
		{ 489, 455, 52 },
		{ 489, 456, 52 },
		{ 489, 457, 52 },
		{ 489, 458, 52 },
		{ 489, 459, 52 },
		{ 490, 403, 52 },
		{ 490, 404, 52 },
		{ 490, 405, 52 },
		{ 490, 406, 52 },
		{ 490, 407, 52 },
		{ 490, 408, 52 },
		{ 490, 409, 52 },
		{ 490, 410, 52 },
		{ 490, 411, 52 },
		{ 490, 412, 52 },
		{ 490, 413, 52 },
		{ 490, 414, 52 },
		{ 490, 415, 52 },
		{ 490, 416, 52 },
		{ 490, 417, 52 },
		{ 490, 418, 52 },
		{ 490, 419, 52 },
		{ 490, 420, 52 },
		{ 490, 421, 52 },
		{ 490, 422, 52 },
		{ 490, 423, 52 },
		{ 490, 424, 52 },
		{ 490, 425, 52 },
		{ 490, 426, 52 },
		{ 490, 427, 52 },
		{ 490, 428, 52 },
		{ 490, 429, 52 },
		{ 490, 430, 52 },
		{ 490, 431, 52 },
		{ 490, 432, 52 },
		{ 490, 433, 52 },
		{ 490, 434, 52 },
		{ 490, 435, 52 },
		{ 490, 436, 52 },
		{ 490, 437, 52 },
		{ 490, 438, 52 },
		{ 490, 439, 52 },
		{ 490, 440, 52 },
		{ 490, 441, 52 },
		{ 490, 442, 52 },
		{ 490, 443, 52 },
		{ 490, 444, 52 },
		{ 490, 445, 52 },
		{ 490, 446, 52 },
		{ 490, 447, 52 },
		{ 490, 448, 52 },
		{ 490, 449, 52 },
		{ 490, 450, 52 },
		{ 490, 451, 52 },
		{ 490, 452, 52 },
		{ 490, 453, 52 },
		{ 490, 454, 52 },
		{ 490, 455, 52 },
		{ 490, 456, 52 },
		{ 490, 457, 52 },
		{ 490, 458, 52 },
		{ 490, 459, 52 },
		{ 490, 460, 52 },
		{ 491, 404, 52 },
		{ 491, 405, 52 },
		{ 491, 406, 52 },
		{ 491, 407, 52 },
		{ 491, 408, 52 },
		{ 491, 409, 52 },
		{ 491, 410, 52 },
		{ 491, 411, 52 },
		{ 491, 412, 52 },
		{ 491, 413, 52 },
		{ 491, 414, 52 },
		{ 491, 415, 52 },
		{ 491, 416, 52 },
		{ 491, 417, 52 },
		{ 491, 418, 52 },
		{ 491, 419, 52 },
		{ 491, 420, 52 },
		{ 491, 421, 52 },
		{ 491, 422, 52 },
		{ 491, 423, 52 },
		{ 491, 424, 52 },
		{ 491, 425, 52 },
		{ 491, 426, 52 },
		{ 491, 427, 52 },
		{ 491, 428, 52 },
		{ 491, 429, 52 },
		{ 491, 430, 52 },
		{ 491, 431, 52 },
		{ 491, 432, 52 },
		{ 491, 433, 52 },
		{ 491, 434, 52 },
		{ 491, 435, 52 },
		{ 491, 436, 52 },
		{ 491, 437, 52 },
		{ 491, 438, 52 },
		{ 491, 439, 52 },
		{ 491, 440, 52 },
		{ 491, 441, 52 },
		{ 491, 442, 52 },
		{ 491, 443, 52 },
		{ 491, 444, 52 },
		{ 491, 445, 52 },
		{ 491, 446, 52 },
		{ 491, 447, 52 },
		{ 491, 448, 52 },
		{ 491, 449, 52 },
		{ 491, 450, 52 },
		{ 491, 451, 52 },
		{ 491, 452, 52 },
		{ 491, 453, 52 },
		{ 491, 454, 52 },
		{ 491, 455, 52 },
		{ 491, 456, 52 },
		{ 491, 457, 52 },
		{ 491, 458, 52 },
		{ 491, 459, 52 },
		{ 491, 460, 52 },
		{ 491, 461, 52 },
		{ 492, 404, 52 },
		{ 492, 405, 52 },
		{ 492, 406, 52 },
		{ 492, 407, 52 },
		{ 492, 408, 52 },
		{ 492, 409, 52 },
		{ 492, 410, 52 },
		{ 492, 411, 52 },
		{ 492, 412, 52 },
		{ 492, 413, 52 },
		{ 492, 414, 52 },
		{ 492, 415, 52 },
		{ 492, 416, 52 },
		{ 492, 417, 52 },
		{ 492, 418, 52 },
		{ 492, 419, 52 },
		{ 492, 420, 52 },
		{ 492, 421, 52 },
		{ 492, 422, 52 },
		{ 492, 423, 52 },
		{ 492, 424, 52 },
		{ 492, 425, 52 },
		{ 492, 426, 52 },
		{ 492, 427, 52 },
		{ 492, 428, 52 },
		{ 492, 429, 52 },
		{ 492, 430, 52 },
		{ 492, 431, 52 },
		{ 492, 432, 52 },
		{ 492, 433, 52 },
		{ 492, 434, 52 },
		{ 492, 435, 52 },
		{ 492, 436, 52 },
		{ 492, 437, 52 },
		{ 492, 438, 52 },
		{ 492, 439, 52 },
		{ 492, 440, 52 },
		{ 492, 441, 52 },
		{ 492, 442, 52 },
		{ 492, 443, 52 },
		{ 492, 444, 52 },
		{ 492, 445, 52 },
		{ 492, 446, 52 },
		{ 492, 447, 52 },
		{ 492, 448, 52 },
		{ 492, 449, 52 },
		{ 492, 450, 52 },
		{ 492, 451, 52 },
		{ 492, 452, 52 },
		{ 492, 453, 52 },
		{ 492, 454, 52 },
		{ 492, 455, 52 },
		{ 492, 456, 52 },
		{ 492, 457, 52 },
		{ 492, 458, 52 },
		{ 492, 459, 52 },
		{ 492, 460, 52 },
		{ 492, 461, 52 },
		{ 492, 462, 52 },
		{ 493, 405, 52 },
		{ 493, 406, 52 },
		{ 493, 407, 52 },
		{ 493, 408, 52 },
		{ 493, 409, 52 },
		{ 493, 410, 52 },
		{ 493, 411, 52 },
		{ 493, 412, 52 },
		{ 493, 413, 52 },
		{ 493, 414, 52 },
		{ 493, 415, 52 },
		{ 493, 416, 52 },
		{ 493, 417, 52 },
		{ 493, 418, 52 },
		{ 493, 419, 52 },
		{ 493, 420, 52 },
		{ 493, 421, 52 },
		{ 493, 422, 52 },
		{ 493, 423, 52 },
		{ 493, 424, 52 },
		{ 493, 425, 52 },
		{ 493, 426, 52 },
		{ 493, 427, 52 },
		{ 493, 428, 52 },
		{ 493, 429, 52 },
		{ 493, 430, 52 },
		{ 493, 431, 52 },
		{ 493, 432, 52 },
		{ 493, 433, 52 },
		{ 493, 434, 52 },
		{ 493, 435, 52 },
		{ 493, 436, 52 },
		{ 493, 437, 52 },
		{ 493, 438, 52 },
		{ 493, 439, 52 },
		{ 493, 440, 52 },
		{ 493, 441, 52 },
		{ 493, 442, 52 },
		{ 493, 443, 52 },
		{ 493, 444, 52 },
		{ 493, 445, 52 },
		{ 493, 446, 52 },
		{ 493, 447, 52 },
		{ 493, 448, 52 },
		{ 493, 449, 52 },
		{ 493, 450, 52 },
		{ 493, 451, 52 },
		{ 493, 452, 52 },
		{ 493, 453, 52 },
		{ 493, 454, 52 },
		{ 493, 455, 52 },
		{ 493, 456, 52 },
		{ 493, 457, 52 },
		{ 493, 458, 52 },
		{ 493, 459, 52 },
		{ 493, 460, 52 },
		{ 493, 461, 52 },
		{ 493, 462, 52 },
		{ 493, 463, 52 },
		{ 494, 406, 52 },
		{ 494, 407, 52 },
		{ 494, 408, 52 },
		{ 494, 409, 52 },
		{ 494, 410, 52 },
		{ 494, 411, 52 },
		{ 494, 412, 52 },
		{ 494, 413, 52 },
		{ 494, 414, 52 },
		{ 494, 415, 52 },
		{ 494, 416, 52 },
		{ 494, 417, 52 },
		{ 494, 418, 52 },
		{ 494, 419, 52 },
		{ 494, 420, 52 },
		{ 494, 421, 52 },
		{ 494, 422, 52 },
		{ 494, 423, 52 },
		{ 494, 424, 52 },
		{ 494, 425, 52 },
		{ 494, 426, 52 },
		{ 494, 427, 52 },
		{ 494, 428, 52 },
		{ 494, 429, 52 },
		{ 494, 430, 52 },
		{ 494, 431, 52 },
		{ 494, 432, 52 },
		{ 494, 433, 52 },
		{ 494, 434, 52 },
		{ 494, 435, 52 },
		{ 494, 436, 52 },
		{ 494, 437, 52 },
		{ 494, 438, 52 },
		{ 494, 439, 52 },
		{ 494, 440, 52 },
		{ 494, 441, 52 },
		{ 494, 442, 52 },
		{ 494, 443, 52 },
		{ 494, 444, 52 },
		{ 494, 445, 52 },
		{ 494, 446, 52 },
		{ 494, 447, 52 },
		{ 494, 448, 52 },
		{ 494, 449, 52 },
		{ 494, 450, 52 },
		{ 494, 451, 52 },
		{ 494, 452, 52 },
		{ 494, 453, 52 },
		{ 494, 454, 52 },
		{ 494, 455, 52 },
		{ 494, 456, 52 },
		{ 494, 457, 52 },
		{ 494, 458, 52 },
		{ 494, 459, 52 },
		{ 494, 460, 52 },
		{ 494, 461, 52 },
		{ 494, 462, 52 },
		{ 494, 463, 52 },
		{ 494, 464, 52 },
		{ 495, 406, 52 },
		{ 495, 407, 52 },
		{ 495, 408, 52 },
		{ 495, 409, 52 },
		{ 495, 410, 52 },
		{ 495, 411, 52 },
		{ 495, 412, 52 },
		{ 495, 413, 52 },
		{ 495, 414, 52 },
		{ 495, 415, 52 },
		{ 495, 416, 52 },
		{ 495, 417, 52 },
		{ 495, 418, 52 },
		{ 495, 419, 52 },
		{ 495, 420, 52 },
		{ 495, 421, 52 },
		{ 495, 422, 52 },
		{ 495, 423, 52 },
		{ 495, 424, 52 },
		{ 495, 425, 52 },
		{ 495, 426, 52 },
		{ 495, 427, 52 },
		{ 495, 428, 52 },
		{ 495, 429, 52 },
		{ 495, 430, 52 },
		{ 495, 431, 52 },
		{ 495, 432, 52 },
		{ 495, 433, 52 },
		{ 495, 434, 52 },
		{ 495, 435, 52 },
		{ 495, 436, 52 },
		{ 495, 437, 52 },
		{ 495, 438, 52 },
		{ 495, 439, 52 },
		{ 495, 440, 52 },
		{ 495, 441, 52 },
		{ 495, 442, 52 },
		{ 495, 443, 52 },
		{ 495, 444, 52 },
		{ 495, 445, 52 },
		{ 495, 446, 52 },
		{ 495, 447, 52 },
		{ 495, 448, 52 },
		{ 495, 449, 52 },
		{ 495, 450, 52 },
		{ 495, 451, 52 },
		{ 495, 452, 52 },
		{ 495, 453, 52 },
		{ 495, 454, 52 },
		{ 495, 455, 52 },
		{ 495, 456, 52 },
		{ 495, 457, 52 },
		{ 495, 458, 52 },
		{ 495, 459, 52 },
		{ 495, 460, 52 },
		{ 495, 461, 52 },
		{ 495, 462, 52 },
		{ 495, 463, 52 },
		{ 495, 464, 52 },
		{ 495, 465, 52 },
		{ 496, 407, 52 },
		{ 496, 408, 52 },
		{ 496, 409, 52 },
		{ 496, 410, 52 },
		{ 496, 411, 52 },
		{ 496, 412, 52 },
		{ 496, 413, 52 },
		{ 496, 414, 52 },
		{ 496, 415, 52 },
		{ 496, 416, 52 },
		{ 496, 417, 52 },
		{ 496, 418, 52 },
		{ 496, 419, 52 },
		{ 496, 420, 52 },
		{ 496, 421, 52 },
		{ 496, 422, 52 },
		{ 496, 423, 52 },
		{ 496, 424, 52 },
		{ 496, 425, 52 },
		{ 496, 426, 52 },
		{ 496, 427, 52 },
		{ 496, 428, 52 },
		{ 496, 429, 52 },
		{ 496, 430, 52 },
		{ 496, 431, 52 },
		{ 496, 432, 52 },
		{ 496, 433, 52 },
		{ 496, 434, 52 },
		{ 496, 435, 52 },
		{ 496, 436, 52 },
		{ 496, 437, 52 },
		{ 496, 438, 52 },
		{ 496, 439, 52 },
		{ 496, 440, 52 },
		{ 496, 441, 52 },
		{ 496, 442, 52 },
		{ 496, 443, 52 },
		{ 496, 444, 52 },
		{ 496, 445, 52 },
		{ 496, 446, 52 },
		{ 496, 447, 52 },
		{ 496, 448, 52 },
		{ 496, 449, 52 },
		{ 496, 450, 52 },
		{ 496, 451, 52 },
		{ 496, 452, 52 },
		{ 496, 453, 52 },
		{ 496, 454, 52 },
		{ 496, 455, 52 },
		{ 496, 456, 52 },
		{ 496, 457, 52 },
		{ 496, 458, 52 },
		{ 496, 459, 52 },
		{ 496, 460, 52 },
		{ 496, 461, 52 },
		{ 496, 462, 52 },
		{ 496, 463, 52 },
		{ 496, 464, 52 },
		{ 496, 465, 52 },
		{ 496, 466, 52 },
		{ 497, 408, 52 },
		{ 497, 409, 52 },
		{ 497, 410, 52 },
		{ 497, 411, 52 },
		{ 497, 412, 52 },
		{ 497, 413, 52 },
		{ 497, 414, 52 },
		{ 497, 415, 52 },
		{ 497, 416, 52 },
		{ 497, 417, 52 },
		{ 497, 418, 52 },
		{ 497, 419, 52 },
		{ 497, 420, 52 },
		{ 497, 421, 52 },
		{ 497, 422, 52 },
		{ 497, 423, 52 },
		{ 497, 424, 52 },
		{ 497, 425, 52 },
		{ 497, 426, 52 },
		{ 497, 427, 52 },
		{ 497, 428, 52 },
		{ 497, 429, 52 },
		{ 497, 430, 52 },
		{ 497, 431, 52 },
		{ 497, 432, 52 },
		{ 497, 433, 52 },
		{ 497, 434, 52 },
		{ 497, 435, 52 },
		{ 497, 436, 52 },
		{ 497, 437, 52 },
		{ 497, 438, 52 },
		{ 497, 439, 52 },
		{ 497, 440, 52 },
		{ 497, 441, 52 },
		{ 497, 442, 52 },
		{ 497, 443, 52 },
		{ 497, 444, 52 },
		{ 497, 445, 52 },
		{ 497, 446, 52 },
		{ 497, 447, 52 },
		{ 497, 448, 52 },
		{ 497, 449, 52 },
		{ 497, 450, 52 },
		{ 497, 451, 52 },
		{ 497, 452, 52 },
		{ 497, 453, 52 },
		{ 497, 454, 52 },
		{ 497, 455, 52 },
		{ 497, 456, 52 },
		{ 497, 457, 52 },
		{ 497, 458, 52 },
		{ 497, 459, 52 },
		{ 497, 460, 52 },
		{ 497, 461, 52 },
		{ 497, 462, 52 },
		{ 497, 463, 52 },
		{ 497, 464, 52 },
		{ 497, 465, 52 },
		{ 497, 466, 52 },
		{ 498, 408, 52 },
		{ 498, 409, 52 },
		{ 498, 410, 52 },
		{ 498, 411, 52 },
		{ 498, 412, 52 },
		{ 498, 413, 52 },
		{ 498, 414, 52 },
		{ 498, 415, 52 },
		{ 498, 416, 52 },
		{ 498, 417, 52 },
		{ 498, 418, 52 },
		{ 498, 419, 52 },
		{ 498, 420, 52 },
		{ 498, 421, 52 },
		{ 498, 422, 52 },
		{ 498, 423, 52 },
		{ 498, 424, 52 },
		{ 498, 425, 52 },
		{ 498, 426, 52 },
		{ 498, 427, 52 },
		{ 498, 428, 52 },
		{ 498, 429, 52 },
		{ 498, 430, 52 },
		{ 498, 431, 52 },
		{ 498, 432, 52 },
		{ 498, 433, 52 },
		{ 498, 434, 52 },
		{ 498, 435, 52 },
		{ 498, 436, 52 },
		{ 498, 437, 52 },
		{ 498, 438, 52 },
		{ 498, 439, 52 },
		{ 498, 440, 52 },
		{ 498, 441, 52 },
		{ 498, 442, 52 },
		{ 498, 443, 52 },
		{ 498, 444, 52 },
		{ 498, 445, 52 },
		{ 498, 446, 52 },
		{ 498, 447, 52 },
		{ 498, 448, 52 },
		{ 498, 449, 52 },
		{ 498, 450, 52 },
		{ 498, 451, 52 },
		{ 498, 452, 52 },
		{ 498, 453, 52 },
		{ 498, 454, 52 },
		{ 498, 455, 52 },
		{ 498, 456, 52 },
		{ 498, 457, 52 },
		{ 498, 458, 52 },
		{ 498, 459, 52 },
		{ 498, 460, 52 },
		{ 498, 461, 52 },
		{ 498, 462, 52 },
		{ 498, 463, 52 },
		{ 498, 464, 52 },
		{ 498, 465, 52 },
		{ 498, 466, 52 },
		{ 498, 467, 52 },
		{ 499, 409, 52 },
		{ 499, 410, 52 },
		{ 499, 411, 52 },
		{ 499, 412, 52 },
		{ 499, 413, 52 },
		{ 499, 414, 52 },
		{ 499, 415, 52 },
		{ 499, 416, 52 },
		{ 499, 417, 52 },
		{ 499, 418, 52 },
		{ 499, 419, 52 },
		{ 499, 420, 52 },
		{ 499, 421, 52 },
		{ 499, 422, 52 },
		{ 499, 423, 52 },
		{ 499, 424, 52 },
		{ 499, 425, 52 },
		{ 499, 426, 52 },
		{ 499, 427, 52 },
		{ 499, 428, 52 },
		{ 499, 429, 52 },
		{ 499, 430, 52 },
		{ 499, 431, 52 },
		{ 499, 432, 52 },
		{ 499, 433, 52 },
		{ 499, 434, 52 },
		{ 499, 435, 52 },
		{ 499, 436, 52 },
		{ 499, 437, 52 },
		{ 499, 438, 52 },
		{ 499, 439, 52 },
		{ 499, 440, 52 },
		{ 499, 441, 52 },
		{ 499, 442, 52 },
		{ 499, 443, 52 },
		{ 499, 444, 52 },
		{ 499, 445, 52 },
		{ 499, 446, 52 },
		{ 499, 447, 52 },
		{ 499, 448, 52 },
		{ 499, 449, 52 },
		{ 499, 450, 52 },
		{ 499, 451, 52 },
		{ 499, 452, 52 },
		{ 499, 453, 52 },
		{ 499, 454, 52 },
		{ 499, 455, 52 },
		{ 499, 456, 52 },
		{ 499, 457, 52 },
		{ 499, 458, 52 },
		{ 499, 459, 52 },
		{ 499, 460, 52 },
		{ 499, 461, 52 },
		{ 499, 462, 52 },
		{ 499, 463, 52 },
		{ 499, 464, 52 },
		{ 499, 465, 52 },
		{ 499, 466, 52 },
		{ 499, 467, 52 },
		{ 499, 468, 52 },
		{ 500, 409, 52 },
		{ 500, 410, 52 },
		{ 500, 411, 52 },
		{ 500, 412, 52 },
		{ 500, 413, 52 },
		{ 500, 414, 52 },
		{ 500, 415, 52 },
		{ 500, 416, 52 },
		{ 500, 417, 52 },
		{ 500, 418, 52 },
		{ 500, 419, 52 },
		{ 500, 420, 52 },
		{ 500, 421, 52 },
		{ 500, 422, 52 },
		{ 500, 423, 52 },
		{ 500, 424, 52 },
		{ 500, 425, 52 },
		{ 500, 426, 52 },
		{ 500, 427, 52 },
		{ 500, 428, 52 },
		{ 500, 429, 52 },
		{ 500, 430, 52 },
		{ 500, 431, 52 },
		{ 500, 432, 52 },
		{ 500, 433, 52 },
		{ 500, 434, 52 },
		{ 500, 435, 52 },
		{ 500, 436, 52 },
		{ 500, 437, 52 },
		{ 500, 438, 52 },
		{ 500, 439, 52 },
		{ 500, 440, 52 },
		{ 500, 441, 52 },
		{ 500, 442, 52 },
		{ 500, 443, 52 },
		{ 500, 444, 52 },
		{ 500, 445, 52 },
		{ 500, 446, 52 },
		{ 500, 447, 52 },
		{ 500, 448, 52 },
		{ 500, 449, 52 },
		{ 500, 450, 52 },
		{ 500, 451, 52 },
		{ 500, 452, 52 },
		{ 500, 453, 52 },
		{ 500, 454, 52 },
		{ 500, 455, 52 },
		{ 500, 456, 52 },
		{ 500, 457, 52 },
		{ 500, 458, 52 },
		{ 500, 459, 52 },
		{ 500, 460, 52 },
		{ 500, 461, 52 },
		{ 500, 462, 52 },
		{ 500, 463, 52 },
		{ 500, 464, 52 },
		{ 500, 465, 52 },
		{ 500, 466, 52 },
		{ 500, 467, 52 },
		{ 500, 468, 52 },
		{ 500, 469, 52 },
		{ 501, 410, 52 },
		{ 501, 411, 52 },
		{ 501, 412, 52 },
		{ 501, 413, 52 },
		{ 501, 414, 52 },
		{ 501, 415, 52 },
		{ 501, 416, 52 },
		{ 501, 417, 52 },
		{ 501, 418, 52 },
		{ 501, 419, 52 },
		{ 501, 420, 52 },
		{ 501, 421, 52 },
		{ 501, 422, 52 },
		{ 501, 423, 52 },
		{ 501, 424, 52 },
		{ 501, 425, 52 },
		{ 501, 426, 52 },
		{ 501, 427, 52 },
		{ 501, 428, 52 },
		{ 501, 429, 52 },
		{ 501, 430, 52 },
		{ 501, 431, 52 },
		{ 501, 432, 52 },
		{ 501, 433, 52 },
		{ 501, 434, 52 },
		{ 501, 435, 52 },
		{ 501, 436, 52 },
		{ 501, 437, 52 },
		{ 501, 438, 52 },
		{ 501, 439, 52 },
		{ 501, 440, 52 },
		{ 501, 441, 52 },
		{ 501, 442, 52 },
		{ 501, 443, 52 },
		{ 501, 444, 52 },
		{ 501, 445, 52 },
		{ 501, 446, 52 },
		{ 501, 447, 52 },
		{ 501, 448, 52 },
		{ 501, 449, 52 },
		{ 501, 450, 52 },
		{ 501, 451, 52 },
		{ 501, 452, 52 },
		{ 501, 453, 52 },
		{ 501, 454, 52 },
		{ 501, 455, 52 },
		{ 501, 456, 52 },
		{ 501, 457, 52 },
		{ 501, 458, 52 },
		{ 501, 459, 52 },
		{ 501, 460, 52 },
		{ 501, 461, 52 },
		{ 501, 462, 52 },
		{ 501, 463, 52 },
		{ 501, 464, 52 },
		{ 501, 465, 52 },
		{ 501, 466, 52 },
		{ 501, 467, 52 },
		{ 501, 468, 52 },
		{ 501, 469, 52 },
		{ 501, 470, 52 },
		{ 502, 411, 52 },
		{ 502, 412, 52 },
		{ 502, 413, 52 },
		{ 502, 414, 52 },
		{ 502, 415, 52 },
		{ 502, 416, 52 },
		{ 502, 417, 52 },
		{ 502, 418, 52 },
		{ 502, 419, 52 },
		{ 502, 420, 52 },
		{ 502, 421, 52 },
		{ 502, 422, 52 },
		{ 502, 423, 52 },
		{ 502, 424, 52 },
		{ 502, 425, 52 },
		{ 502, 426, 52 },
		{ 502, 427, 52 },
		{ 502, 428, 52 },
		{ 502, 429, 52 },
		{ 502, 430, 52 },
		{ 502, 431, 52 },
		{ 502, 432, 52 },
		{ 502, 433, 52 },
		{ 502, 434, 52 },
		{ 502, 435, 52 },
		{ 502, 436, 52 },
		{ 502, 437, 52 },
		{ 502, 438, 52 },
		{ 502, 439, 52 },
		{ 502, 440, 52 },
		{ 502, 441, 52 },
		{ 502, 442, 52 },
		{ 502, 443, 52 },
		{ 502, 444, 52 },
		{ 502, 445, 52 },
		{ 502, 446, 52 },
		{ 502, 447, 52 },
		{ 502, 448, 52 },
		{ 502, 449, 52 },
		{ 502, 450, 52 },
		{ 502, 451, 52 },
		{ 502, 452, 52 },
		{ 502, 453, 52 },
		{ 502, 454, 52 },
		{ 502, 455, 52 },
		{ 502, 456, 52 },
		{ 502, 457, 52 },
		{ 502, 458, 52 },
		{ 502, 459, 52 },
		{ 502, 460, 52 },
		{ 502, 461, 52 },
		{ 502, 462, 52 },
		{ 502, 463, 52 },
		{ 502, 464, 52 },
		{ 502, 465, 52 },
		{ 502, 466, 52 },
		{ 502, 467, 52 },
		{ 502, 468, 52 },
		{ 502, 469, 52 },
		{ 502, 470, 52 },
		{ 502, 471, 52 },
		{ 503, 411, 52 },
		{ 503, 412, 52 },
		{ 503, 413, 52 },
		{ 503, 414, 52 },
		{ 503, 415, 52 },
		{ 503, 416, 52 },
		{ 503, 417, 52 },
		{ 503, 418, 52 },
		{ 503, 419, 52 },
		{ 503, 420, 52 },
		{ 503, 421, 52 },
		{ 503, 422, 52 },
		{ 503, 423, 52 },
		{ 503, 424, 52 },
		{ 503, 425, 52 },
		{ 503, 426, 52 },
		{ 503, 427, 52 },
		{ 503, 428, 52 },
		{ 503, 429, 52 },
		{ 503, 430, 52 },
		{ 503, 431, 52 },
		{ 503, 432, 52 },
		{ 503, 433, 52 },
		{ 503, 434, 52 },
		{ 503, 435, 52 },
		{ 503, 436, 52 },
		{ 503, 437, 52 },
		{ 503, 438, 52 },
		{ 503, 439, 52 },
		{ 503, 440, 52 },
		{ 503, 441, 52 },
		{ 503, 442, 52 },
		{ 503, 443, 52 },
		{ 503, 444, 52 },
		{ 503, 445, 52 },
		{ 503, 446, 52 },
		{ 503, 447, 52 },
		{ 503, 448, 52 },
		{ 503, 449, 52 },
		{ 503, 450, 52 },
		{ 503, 451, 52 },
		{ 503, 452, 52 },
		{ 503, 453, 52 },
		{ 503, 454, 52 },
		{ 503, 455, 52 },
		{ 503, 456, 52 },
		{ 503, 457, 52 },
		{ 503, 458, 52 },
		{ 503, 459, 52 },
		{ 503, 460, 52 },
		{ 503, 461, 52 },
		{ 503, 462, 52 },
		{ 503, 463, 52 },
		{ 503, 464, 52 },
		{ 503, 465, 52 },
		{ 503, 466, 52 },
		{ 503, 467, 52 },
		{ 503, 468, 52 },
		{ 503, 469, 52 },
		{ 503, 470, 52 },
		{ 503, 471, 52 },
		{ 503, 472, 52 },
		{ 504, 412, 52 },
		{ 504, 413, 52 },
		{ 504, 414, 52 },
		{ 504, 415, 52 },
		{ 504, 416, 52 },
		{ 504, 417, 52 },
		{ 504, 418, 52 },
		{ 504, 419, 52 },
		{ 504, 420, 52 },
		{ 504, 421, 52 },
		{ 504, 422, 52 },
		{ 504, 423, 52 },
		{ 504, 424, 52 },
		{ 504, 425, 52 },
		{ 504, 426, 52 },
		{ 504, 427, 52 },
		{ 504, 428, 52 },
		{ 504, 429, 52 },
		{ 504, 430, 52 },
		{ 504, 431, 52 },
		{ 504, 432, 52 },
		{ 504, 433, 52 },
		{ 504, 434, 52 },
		{ 504, 435, 52 },
		{ 504, 436, 52 },
		{ 504, 437, 52 },
		{ 504, 438, 52 },
		{ 504, 439, 52 },
		{ 504, 440, 52 },
		{ 504, 441, 52 },
		{ 504, 442, 52 },
		{ 504, 443, 52 },
		{ 504, 444, 52 },
		{ 504, 445, 52 },
		{ 504, 446, 52 },
		{ 504, 447, 52 },
		{ 504, 448, 52 },
		{ 504, 449, 52 },
		{ 504, 450, 52 },
		{ 504, 451, 52 },
		{ 504, 452, 52 },
		{ 504, 453, 52 },
		{ 504, 454, 52 },
		{ 504, 455, 52 },
		{ 504, 456, 52 },
		{ 504, 457, 52 },
		{ 504, 458, 52 },
		{ 504, 459, 52 },
		{ 504, 460, 52 },
		{ 504, 461, 52 },
		{ 504, 462, 52 },
		{ 504, 463, 52 },
		{ 504, 464, 52 },
		{ 504, 465, 52 },
		{ 504, 466, 52 },
		{ 504, 467, 52 },
		{ 504, 468, 52 },
		{ 504, 469, 52 },
		{ 504, 470, 52 },
		{ 504, 471, 52 },
		{ 504, 472, 52 },
		{ 504, 473, 52 },
		{ 505, 413, 52 },
		{ 505, 414, 52 },
		{ 505, 415, 52 },
		{ 505, 416, 52 },
		{ 505, 417, 52 },
		{ 505, 418, 52 },
		{ 505, 419, 52 },
		{ 505, 420, 52 },
		{ 505, 421, 52 },
		{ 505, 422, 52 },
		{ 505, 423, 52 },
		{ 505, 424, 52 },
		{ 505, 425, 52 },
		{ 505, 426, 52 },
		{ 505, 427, 52 },
		{ 505, 428, 52 },
		{ 505, 429, 52 },
		{ 505, 430, 52 },
		{ 505, 431, 52 },
		{ 505, 432, 52 },
		{ 505, 433, 52 },
		{ 505, 434, 52 },
		{ 505, 435, 52 },
		{ 505, 436, 52 },
		{ 505, 437, 52 },
		{ 505, 438, 52 },
		{ 505, 439, 52 },
		{ 505, 440, 52 },
		{ 505, 441, 52 },
		{ 505, 442, 52 },
		{ 505, 443, 52 },
		{ 505, 444, 52 },
		{ 505, 445, 52 },
		{ 505, 446, 52 },
		{ 505, 447, 52 },
		{ 505, 448, 52 },
		{ 505, 449, 52 },
		{ 505, 450, 52 },
		{ 505, 451, 52 },
		{ 505, 452, 52 },
		{ 505, 453, 52 },
		{ 505, 454, 52 },
		{ 505, 455, 52 },
		{ 505, 456, 52 },
		{ 505, 457, 52 },
		{ 505, 458, 52 },
		{ 505, 459, 52 },
		{ 505, 460, 52 },
		{ 505, 461, 52 },
		{ 505, 462, 52 },
		{ 505, 463, 52 },
		{ 505, 464, 52 },
		{ 505, 465, 52 },
		{ 505, 466, 52 },
		{ 505, 467, 52 },
		{ 505, 468, 52 },
		{ 505, 469, 52 },
		{ 505, 470, 52 },
		{ 505, 471, 52 },
		{ 505, 472, 52 },
		{ 505, 473, 52 },
		{ 506, 413, 52 },
		{ 506, 414, 52 },
		{ 506, 415, 52 },
		{ 506, 416, 52 },
		{ 506, 417, 52 },
		{ 506, 418, 52 },
		{ 506, 419, 52 },
		{ 506, 420, 52 },
		{ 506, 421, 52 },
		{ 506, 422, 52 },
		{ 506, 423, 52 },
		{ 506, 424, 52 },
		{ 506, 425, 52 },
		{ 506, 426, 52 },
		{ 506, 427, 52 },
		{ 506, 428, 52 },
		{ 506, 429, 52 },
		{ 506, 430, 52 },
		{ 506, 431, 52 },
		{ 506, 432, 52 },
		{ 506, 433, 52 },
		{ 506, 434, 52 },
		{ 506, 435, 52 },
		{ 506, 436, 52 },
		{ 506, 437, 52 },
		{ 506, 438, 52 },
		{ 506, 439, 52 },
		{ 506, 440, 52 },
		{ 506, 441, 52 },
		{ 506, 442, 52 },
		{ 506, 443, 52 },
		{ 506, 444, 52 },
		{ 506, 445, 52 },
		{ 506, 446, 52 },
		{ 506, 447, 52 },
		{ 506, 448, 52 },
		{ 506, 449, 52 },
		{ 506, 450, 52 },
		{ 506, 451, 52 },
		{ 506, 452, 52 },
		{ 506, 453, 52 },
		{ 506, 454, 52 },
		{ 506, 455, 52 },
		{ 506, 456, 52 },
		{ 506, 457, 52 },
		{ 506, 458, 52 },
		{ 506, 459, 52 },
		{ 506, 460, 52 },
		{ 506, 461, 52 },
		{ 506, 462, 52 },
		{ 506, 463, 52 },
		{ 506, 464, 52 },
		{ 506, 465, 52 },
		{ 506, 466, 52 },
		{ 506, 467, 52 },
		{ 506, 468, 52 },
		{ 506, 469, 52 },
		{ 506, 470, 52 },
		{ 506, 471, 52 },
		{ 506, 472, 52 },
		{ 506, 473, 52 },
		{ 506, 474, 52 },
		{ 507, 414, 52 },
		{ 507, 415, 52 },
		{ 507, 416, 52 },
		{ 507, 417, 52 },
		{ 507, 418, 52 },
		{ 507, 419, 52 },
		{ 507, 420, 52 },
		{ 507, 421, 52 },
		{ 507, 422, 52 },
		{ 507, 423, 52 },
		{ 507, 424, 52 },
		{ 507, 425, 52 },
		{ 507, 426, 52 },
		{ 507, 427, 52 },
		{ 507, 428, 52 },
		{ 507, 429, 52 },
		{ 507, 430, 52 },
		{ 507, 431, 52 },
		{ 507, 432, 52 },
		{ 507, 433, 52 },
		{ 507, 434, 52 },
		{ 507, 435, 52 },
		{ 507, 436, 52 },
		{ 507, 437, 52 },
		{ 507, 438, 52 },
		{ 507, 439, 52 },
		{ 507, 440, 52 },
		{ 507, 441, 52 },
		{ 507, 442, 52 },
		{ 507, 443, 52 },
		{ 507, 444, 52 },
		{ 507, 445, 52 },
		{ 507, 446, 52 },
		{ 507, 447, 52 },
		{ 507, 448, 52 },
		{ 507, 449, 52 },
		{ 507, 450, 52 },
		{ 507, 451, 52 },
		{ 507, 452, 52 },
		{ 507, 453, 52 },
		{ 507, 454, 52 },
		{ 507, 455, 52 },
		{ 507, 456, 52 },
		{ 507, 457, 52 },
		{ 507, 458, 52 },
		{ 507, 459, 52 },
		{ 507, 460, 52 },
		{ 507, 461, 52 },
		{ 507, 462, 52 },
		{ 507, 463, 52 },
		{ 507, 464, 52 },
		{ 507, 465, 52 },
		{ 507, 466, 52 },
		{ 507, 467, 52 },
		{ 507, 468, 52 },
		{ 507, 469, 52 },
		{ 507, 470, 52 },
		{ 507, 471, 52 },
		{ 507, 472, 52 },
		{ 507, 473, 52 },
		{ 507, 474, 52 },
		{ 507, 475, 52 },
		{ 508, 414, 52 },
		{ 508, 415, 52 },
		{ 508, 416, 52 },
		{ 508, 417, 52 },
		{ 508, 418, 52 },
		{ 508, 419, 52 },
		{ 508, 420, 52 },
		{ 508, 421, 52 },
		{ 508, 422, 52 },
		{ 508, 423, 52 },
		{ 508, 424, 52 },
		{ 508, 425, 52 },
		{ 508, 426, 52 },
		{ 508, 427, 52 },
		{ 508, 428, 52 },
		{ 508, 429, 52 },
		{ 508, 430, 52 },
		{ 508, 431, 52 },
		{ 508, 432, 52 },
		{ 508, 433, 52 },
		{ 508, 434, 52 },
		{ 508, 435, 52 },
		{ 508, 436, 52 },
		{ 508, 437, 52 },
		{ 508, 438, 52 },
		{ 508, 439, 52 },
		{ 508, 440, 52 },
		{ 508, 441, 52 },
		{ 508, 442, 52 },
		{ 508, 443, 52 },
		{ 508, 444, 52 },
		{ 508, 445, 52 },
		{ 508, 446, 52 },
		{ 508, 447, 52 },
		{ 508, 448, 52 },
		{ 508, 449, 52 },
		{ 508, 450, 52 },
		{ 508, 451, 52 },
		{ 508, 452, 52 },
		{ 508, 453, 52 },
		{ 508, 454, 52 },
		{ 508, 455, 52 },
		{ 508, 456, 52 },
		{ 508, 457, 52 },
		{ 508, 458, 52 },
		{ 508, 459, 52 },
		{ 508, 460, 52 },
		{ 508, 461, 52 },
		{ 508, 462, 52 },
		{ 508, 463, 52 },
		{ 508, 464, 52 },
		{ 508, 465, 52 },
		{ 508, 466, 52 },
		{ 508, 467, 52 },
		{ 508, 468, 52 },
		{ 508, 469, 52 },
		{ 508, 470, 52 },
		{ 508, 471, 52 },
		{ 508, 472, 52 },
		{ 508, 473, 52 },
		{ 508, 474, 52 },
		{ 508, 475, 52 },
		{ 508, 476, 52 },
		{ 509, 415, 52 },
		{ 509, 416, 52 },
		{ 509, 417, 52 },
		{ 509, 418, 52 },
		{ 509, 419, 52 },
		{ 509, 420, 52 },
		{ 509, 421, 52 },
		{ 509, 422, 52 },
		{ 509, 423, 52 },
		{ 509, 424, 52 },
		{ 509, 425, 52 },
		{ 509, 426, 52 },
		{ 509, 427, 52 },
		{ 509, 428, 52 },
		{ 509, 429, 52 },
		{ 509, 430, 52 },
		{ 509, 431, 52 },
		{ 509, 432, 52 },
		{ 509, 433, 52 },
		{ 509, 434, 52 },
		{ 509, 435, 52 },
		{ 509, 436, 52 },
		{ 509, 437, 52 },
		{ 509, 438, 52 },
		{ 509, 439, 52 },
		{ 509, 440, 52 },
		{ 509, 441, 52 },
		{ 509, 442, 52 },
		{ 509, 443, 52 },
		{ 509, 444, 52 },
		{ 509, 445, 52 },
		{ 509, 446, 52 },
		{ 509, 447, 52 },
		{ 509, 448, 52 },
		{ 509, 449, 52 },
		{ 509, 450, 52 },
		{ 509, 451, 52 },
		{ 509, 452, 52 },
		{ 509, 453, 52 },
		{ 509, 454, 52 },
		{ 509, 455, 52 },
		{ 509, 456, 52 },
		{ 509, 457, 52 },
		{ 509, 458, 52 },
		{ 509, 459, 52 },
		{ 509, 460, 52 },
		{ 509, 461, 52 },
		{ 509, 462, 52 },
		{ 509, 463, 52 },
		{ 509, 464, 52 },
		{ 509, 465, 52 },
		{ 509, 466, 52 },
		{ 509, 467, 52 },
		{ 509, 468, 52 },
		{ 509, 469, 52 },
		{ 509, 470, 52 },
		{ 509, 471, 52 },
		{ 509, 472, 52 },
		{ 509, 473, 52 },
		{ 509, 474, 52 },
		{ 509, 475, 52 },
		{ 509, 476, 52 },
		{ 509, 477, 52 },
		{ 510, 416, 52 },
		{ 510, 417, 52 },
		{ 510, 418, 52 },
		{ 510, 419, 52 },
		{ 510, 420, 52 },
		{ 510, 421, 52 },
		{ 510, 422, 52 },
		{ 510, 423, 52 },
		{ 510, 424, 52 },
		{ 510, 425, 52 },
		{ 510, 426, 52 },
		{ 510, 427, 52 },
		{ 510, 428, 52 },
		{ 510, 429, 52 },
		{ 510, 430, 52 },
		{ 510, 431, 52 },
		{ 510, 432, 52 },
		{ 510, 433, 52 },
		{ 510, 434, 52 },
		{ 510, 435, 52 },
		{ 510, 436, 52 },
		{ 510, 437, 52 },
		{ 510, 438, 52 },
		{ 510, 439, 52 },
		{ 510, 440, 52 },
		{ 510, 441, 52 },
		{ 510, 442, 52 },
		{ 510, 443, 52 },
		{ 510, 444, 52 },
		{ 510, 445, 52 },
		{ 510, 446, 52 },
		{ 510, 447, 52 },
		{ 510, 448, 52 },
		{ 510, 449, 52 },
		{ 510, 450, 52 },
		{ 510, 451, 52 },
		{ 510, 452, 52 },
		{ 510, 453, 52 },
		{ 510, 454, 52 },
		{ 510, 455, 52 },
		{ 510, 456, 52 },
		{ 510, 457, 52 },
		{ 510, 458, 52 },
		{ 510, 459, 52 },
		{ 510, 460, 52 },
		{ 510, 461, 52 },
		{ 510, 462, 52 },
		{ 510, 463, 52 },
		{ 510, 464, 52 },
		{ 510, 465, 52 },
		{ 510, 466, 52 },
		{ 510, 467, 52 },
		{ 510, 468, 52 },
		{ 510, 469, 52 },
		{ 510, 470, 52 },
		{ 510, 471, 52 },
		{ 510, 472, 52 },
		{ 510, 473, 52 },
		{ 510, 474, 52 },
		{ 510, 475, 52 },
		{ 510, 476, 52 },
		{ 510, 477, 52 },
		{ 510, 478, 52 },
		{ 511, 416, 52 },
		{ 511, 417, 52 },
		{ 511, 418, 52 },
		{ 511, 419, 52 },
		{ 511, 420, 52 },
		{ 511, 421, 52 },
		{ 511, 422, 52 },
		{ 511, 423, 52 },
		{ 511, 424, 52 },
		{ 511, 425, 52 },
		{ 511, 426, 52 },
		{ 511, 427, 52 },
		{ 511, 428, 52 },
		{ 511, 429, 52 },
		{ 511, 430, 52 },
		{ 511, 431, 52 },
		{ 511, 432, 52 },
		{ 511, 433, 52 },
		{ 511, 434, 52 },
		{ 511, 435, 52 },
		{ 511, 436, 52 },
		{ 511, 437, 52 },
		{ 511, 438, 52 },
		{ 511, 439, 52 },
		{ 511, 440, 52 },
		{ 511, 441, 52 },
		{ 511, 442, 52 },
		{ 511, 443, 52 },
		{ 511, 444, 52 },
		{ 511, 445, 52 },
		{ 511, 446, 52 },
		{ 511, 447, 52 },
		{ 511, 448, 52 },
		{ 511, 449, 52 },
		{ 511, 450, 52 },
		{ 511, 451, 52 },
		{ 511, 452, 52 },
		{ 511, 453, 52 },
		{ 511, 454, 52 },
		{ 511, 455, 52 },
		{ 511, 456, 52 },
		{ 511, 457, 52 },
		{ 511, 458, 52 },
		{ 511, 459, 52 },
		{ 511, 460, 52 },
		{ 511, 461, 52 },
		{ 511, 462, 52 },
		{ 511, 463, 52 },
		{ 511, 464, 52 },
		{ 511, 465, 52 },
		{ 511, 466, 52 },
		{ 511, 467, 52 },
		{ 511, 468, 52 },
		{ 511, 469, 52 },
		{ 511, 470, 52 },
		{ 511, 471, 52 },
		{ 511, 472, 52 },
		{ 511, 473, 52 },
		{ 511, 474, 52 },
		{ 511, 475, 52 },
		{ 511, 476, 52 },
		{ 511, 477, 52 },
		{ 511, 478, 52 },
		{ 511, 479, 52 },
		{ 512, 417, 52 },
		{ 512, 418, 52 },
		{ 512, 419, 52 },
		{ 512, 420, 52 },
		{ 512, 421, 52 },
		{ 512, 422, 52 },
		{ 512, 423, 52 },
		{ 512, 424, 52 },
		{ 512, 425, 52 },
		{ 512, 426, 52 },
		{ 512, 427, 52 },
		{ 512, 428, 52 },
		{ 512, 429, 52 },
		{ 512, 430, 52 },
		{ 512, 431, 52 },
		{ 512, 432, 52 },
		{ 512, 433, 52 },
		{ 512, 434, 52 },
		{ 512, 435, 52 },
		{ 512, 436, 52 },
		{ 512, 437, 52 },
		{ 512, 438, 52 },
		{ 512, 439, 52 },
		{ 512, 440, 52 },
		{ 512, 441, 52 },
		{ 512, 442, 52 },
		{ 512, 443, 52 },
		{ 512, 444, 52 },
		{ 512, 445, 52 },
		{ 512, 446, 52 },
		{ 512, 447, 52 },
		{ 512, 448, 52 },
		{ 512, 449, 52 },
		{ 512, 450, 52 },
		{ 512, 451, 52 },
		{ 512, 452, 52 },
		{ 512, 453, 52 },
		{ 512, 454, 52 },
		{ 512, 455, 52 },
		{ 512, 456, 52 },
		{ 512, 457, 52 },
		{ 512, 458, 52 },
		{ 512, 459, 52 },
		{ 512, 460, 52 },
		{ 512, 461, 52 },
		{ 512, 462, 52 },
		{ 512, 463, 52 },
		{ 512, 464, 52 },
		{ 512, 465, 52 },
		{ 512, 466, 52 },
		{ 512, 467, 52 },
		{ 512, 468, 52 },
		{ 512, 469, 52 },
		{ 512, 470, 52 },
		{ 512, 471, 52 },
		{ 512, 472, 52 },
		{ 512, 473, 52 },
		{ 512, 474, 52 },
		{ 512, 475, 52 },
		{ 512, 476, 52 },
		{ 512, 477, 52 },
		{ 512, 478, 52 },
		{ 512, 479, 52 },
		{ 512, 480, 52 },
		{ 1, 2, 53 },
		{ 1, 3, 53 },
		{ 1, 4, 53 },
		{ 1, 5, 53 },
		{ 1, 6, 53 },
		{ 1, 7, 53 },
		{ 1, 8, 53 },
		{ 1, 9, 53 },
		{ 1, 10, 53 },
		{ 1, 11, 53 },
		{ 1, 12, 53 },
		{ 1, 13, 53 },
		{ 1, 14, 53 },
		{ 1, 15, 53 },
		{ 1, 16, 53 },
		{ 1, 17, 53 },
		{ 1, 18, 53 },
		{ 1, 19, 53 },
		{ 1, 20, 53 },
		{ 1, 21, 53 },
		{ 1, 22, 53 },
		{ 1, 23, 53 },
		{ 1, 24, 53 },
		{ 1, 25, 53 },
		{ 1, 26, 53 },
		{ 1, 27, 53 },
		{ 1, 28, 53 },
		{ 1, 29, 53 },
		{ 1, 30, 53 },
		{ 1, 31, 53 },
		{ 1, 32, 53 },
		{ 1, 33, 53 },
		{ 1, 34, 53 },
		{ 1, 35, 53 },
		{ 1, 36, 53 },
		{ 1, 37, 53 },
		{ 1, 38, 53 },
		{ 1, 39, 53 },
		{ 1, 40, 53 },
		{ 1, 41, 53 },
		{ 1, 42, 53 },
		{ 1, 43, 53 },
		{ 1, 44, 53 },
		{ 1, 45, 53 },
		{ 1, 46, 53 },
		{ 1, 47, 53 },
		{ 1, 48, 53 },
		{ 1, 49, 53 },
		{ 1, 50, 53 },
		{ 1, 51, 53 },
		{ 1, 52, 53 },
		{ 1, 53, 53 },
		{ 1, 54, 53 },
		{ 1, 55, 53 },
		{ 1, 56, 53 },
		{ 1, 57, 53 },
		{ 1, 58, 53 },
		{ 1, 59, 53 },
		{ 1, 60, 53 },
		{ 1, 61, 53 },
		{ 1, 62, 53 },
		{ 1, 63, 53 },
		{ 1, 64, 53 },
		{ 2, 3, 53 },
		{ 2, 4, 53 },
		{ 2, 5, 53 },
		{ 2, 6, 53 },
		{ 2, 7, 53 },
		{ 2, 8, 53 },
		{ 2, 9, 53 },
		{ 2, 10, 53 },
		{ 2, 11, 53 },
		{ 2, 12, 53 },
		{ 2, 13, 53 },
		{ 2, 14, 53 },
		{ 2, 15, 53 },
		{ 2, 16, 53 },
		{ 2, 17, 53 },
		{ 2, 18, 53 },
		{ 2, 19, 53 },
		{ 2, 20, 53 },
		{ 2, 21, 53 },
		{ 2, 22, 53 },
		{ 2, 23, 53 },
		{ 2, 24, 53 },
		{ 2, 25, 53 },
		{ 2, 26, 53 },
		{ 2, 27, 53 },
		{ 2, 28, 53 },
		{ 2, 29, 53 },
		{ 2, 30, 53 },
		{ 2, 31, 53 },
		{ 2, 32, 53 },
		{ 2, 33, 53 },
		{ 2, 34, 53 },
		{ 2, 35, 53 },
		{ 2, 36, 53 },
		{ 2, 37, 53 },
		{ 2, 38, 53 },
		{ 2, 39, 53 },
		{ 2, 40, 53 },
		{ 2, 41, 53 },
		{ 2, 42, 53 },
		{ 2, 43, 53 },
		{ 2, 44, 53 },
		{ 2, 45, 53 },
		{ 2, 46, 53 },
		{ 2, 47, 53 },
		{ 2, 48, 53 },
		{ 2, 49, 53 },
		{ 2, 50, 53 },
		{ 2, 51, 53 },
		{ 2, 52, 53 },
		{ 2, 53, 53 },
		{ 2, 54, 53 },
		{ 2, 55, 53 },
		{ 2, 56, 53 },
		{ 2, 57, 53 },
		{ 2, 58, 53 },
		{ 2, 59, 53 },
		{ 2, 60, 53 },
		{ 2, 61, 53 },
		{ 2, 62, 53 },
		{ 2, 63, 53 },
		{ 2, 64, 53 },
		{ 2, 65, 53 },
		{ 3, 4, 53 },
		{ 3, 5, 53 },
		{ 3, 6, 53 },
		{ 3, 7, 53 },
		{ 3, 8, 53 },
		{ 3, 9, 53 },
		{ 3, 10, 53 },
		{ 3, 11, 53 },
		{ 3, 12, 53 },
		{ 3, 13, 53 },
		{ 3, 14, 53 },
		{ 3, 15, 53 },
		{ 3, 16, 53 },
		{ 3, 17, 53 },
		{ 3, 18, 53 },
		{ 3, 19, 53 },
		{ 3, 20, 53 },
		{ 3, 21, 53 },
		{ 3, 22, 53 },
		{ 3, 23, 53 },
		{ 3, 24, 53 },
		{ 3, 25, 53 },
		{ 3, 26, 53 },
		{ 3, 27, 53 },
		{ 3, 28, 53 },
		{ 3, 29, 53 },
		{ 3, 30, 53 },
		{ 3, 31, 53 },
		{ 3, 32, 53 },
		{ 3, 33, 53 },
		{ 3, 34, 53 },
		{ 3, 35, 53 },
		{ 3, 36, 53 },
		{ 3, 37, 53 },
		{ 3, 38, 53 },
		{ 3, 39, 53 },
		{ 3, 40, 53 },
		{ 3, 41, 53 },
		{ 3, 42, 53 },
		{ 3, 43, 53 },
		{ 3, 44, 53 },
		{ 3, 45, 53 },
		{ 3, 46, 53 },
		{ 3, 47, 53 },
		{ 3, 48, 53 },
		{ 3, 49, 53 },
		{ 3, 50, 53 },
		{ 3, 51, 53 },
		{ 3, 52, 53 },
		{ 3, 53, 53 },
		{ 3, 54, 53 },
		{ 3, 55, 53 },
		{ 3, 56, 53 },
		{ 3, 57, 53 },
		{ 3, 58, 53 },
		{ 3, 59, 53 },
		{ 3, 60, 53 },
		{ 3, 61, 53 },
		{ 3, 62, 53 },
		{ 3, 63, 53 },
		{ 3, 64, 53 },
		{ 3, 65, 53 },
		{ 3, 66, 53 },
		{ 4, 5, 53 },
		{ 4, 6, 53 },
		{ 4, 7, 53 },
		{ 4, 8, 53 },
		{ 4, 9, 53 },
		{ 4, 10, 53 },
		{ 4, 11, 53 },
		{ 4, 12, 53 },
		{ 4, 13, 53 },
		{ 4, 14, 53 },
		{ 4, 15, 53 },
		{ 4, 16, 53 },
		{ 4, 17, 53 },
		{ 4, 18, 53 },
		{ 4, 19, 53 },
		{ 4, 20, 53 },
		{ 4, 21, 53 },
		{ 4, 22, 53 },
		{ 4, 23, 53 },
		{ 4, 24, 53 },
		{ 4, 25, 53 },
		{ 4, 26, 53 },
		{ 4, 27, 53 },
		{ 4, 28, 53 },
		{ 4, 29, 53 },
		{ 4, 30, 53 },
		{ 4, 31, 53 },
		{ 4, 32, 53 },
		{ 4, 33, 53 },
		{ 4, 34, 53 },
		{ 4, 35, 53 },
		{ 4, 36, 53 },
		{ 4, 37, 53 },
		{ 4, 38, 53 },
		{ 4, 39, 53 },
		{ 4, 40, 53 },
		{ 4, 41, 53 },
		{ 4, 42, 53 },
		{ 4, 43, 53 },
		{ 4, 44, 53 },
		{ 4, 45, 53 },
		{ 4, 46, 53 },
		{ 4, 47, 53 },
		{ 4, 48, 53 },
		{ 4, 49, 53 },
		{ 4, 50, 53 },
		{ 4, 51, 53 },
		{ 4, 52, 53 },
		{ 4, 53, 53 },
		{ 4, 54, 53 },
		{ 4, 55, 53 },
		{ 4, 56, 53 },
		{ 4, 57, 53 },
		{ 4, 58, 53 },
		{ 4, 59, 53 },
		{ 4, 60, 53 },
		{ 4, 61, 53 },
		{ 4, 62, 53 },
		{ 4, 63, 53 },
		{ 4, 64, 53 },
		{ 4, 65, 53 },
		{ 4, 66, 53 },
		{ 4, 67, 53 },
		{ 5, 6, 53 },
		{ 5, 7, 53 },
		{ 5, 8, 53 },
		{ 5, 9, 53 },
		{ 5, 10, 53 },
		{ 5, 11, 53 },
		{ 5, 12, 53 },
		{ 5, 13, 53 },
		{ 5, 14, 53 },
		{ 5, 15, 53 },
		{ 5, 16, 53 },
		{ 5, 17, 53 },
		{ 5, 18, 53 },
		{ 5, 19, 53 },
		{ 5, 20, 53 },
		{ 5, 21, 53 },
		{ 5, 22, 53 },
		{ 5, 23, 53 },
		{ 5, 24, 53 },
		{ 5, 25, 53 },
		{ 5, 26, 53 },
		{ 5, 27, 53 },
		{ 5, 28, 53 },
		{ 5, 29, 53 },
		{ 5, 30, 53 },
		{ 5, 31, 53 },
		{ 5, 32, 53 },
		{ 5, 33, 53 },
		{ 5, 34, 53 },
		{ 5, 35, 53 },
		{ 5, 36, 53 },
		{ 5, 37, 53 },
		{ 5, 38, 53 },
		{ 5, 39, 53 },
		{ 5, 40, 53 },
		{ 5, 41, 53 },
		{ 5, 42, 53 },
		{ 5, 43, 53 },
		{ 5, 44, 53 },
		{ 5, 45, 53 },
		{ 5, 46, 53 },
		{ 5, 47, 53 },
		{ 5, 48, 53 },
		{ 5, 49, 53 },
		{ 5, 50, 53 },
		{ 5, 51, 53 },
		{ 5, 52, 53 },
		{ 5, 53, 53 },
		{ 5, 54, 53 },
		{ 5, 55, 53 },
		{ 5, 56, 53 },
		{ 5, 57, 53 },
		{ 5, 58, 53 },
		{ 5, 59, 53 },
		{ 5, 60, 53 },
		{ 5, 61, 53 },
		{ 5, 62, 53 },
		{ 5, 63, 53 },
		{ 5, 64, 53 },
		{ 5, 65, 53 },
		{ 5, 66, 53 },
		{ 5, 67, 53 },
		{ 6, 7, 53 },
		{ 6, 8, 53 },
		{ 6, 9, 53 },
		{ 6, 10, 53 },
		{ 6, 11, 53 },
		{ 6, 12, 53 },
		{ 6, 13, 53 },
		{ 6, 14, 53 },
		{ 6, 15, 53 },
		{ 6, 16, 53 },
		{ 6, 17, 53 },
		{ 6, 18, 53 },
		{ 6, 19, 53 },
		{ 6, 20, 53 },
		{ 6, 21, 53 },
		{ 6, 22, 53 },
		{ 6, 23, 53 },
		{ 6, 24, 53 },
		{ 6, 25, 53 },
		{ 6, 26, 53 },
		{ 6, 27, 53 },
		{ 6, 28, 53 },
		{ 6, 29, 53 },
		{ 6, 30, 53 },
		{ 6, 31, 53 },
		{ 6, 32, 53 },
		{ 6, 33, 53 },
		{ 6, 34, 53 },
		{ 6, 35, 53 },
		{ 6, 36, 53 },
		{ 6, 37, 53 },
		{ 6, 38, 53 },
		{ 6, 39, 53 },
		{ 6, 40, 53 },
		{ 6, 41, 53 },
		{ 6, 42, 53 },
		{ 6, 43, 53 },
		{ 6, 44, 53 },
		{ 6, 45, 53 },
		{ 6, 46, 53 },
		{ 6, 47, 53 },
		{ 6, 48, 53 },
		{ 6, 49, 53 },
		{ 6, 50, 53 },
		{ 6, 51, 53 },
		{ 6, 52, 53 },
		{ 6, 53, 53 },
		{ 6, 54, 53 },
		{ 6, 55, 53 },
		{ 6, 56, 53 },
		{ 6, 57, 53 },
		{ 6, 58, 53 },
		{ 6, 59, 53 },
		{ 6, 60, 53 },
		{ 6, 61, 53 },
		{ 6, 62, 53 },
		{ 6, 63, 53 },
		{ 6, 64, 53 },
		{ 6, 65, 53 },
		{ 6, 66, 53 },
		{ 6, 67, 53 },
		{ 6, 68, 53 },
		{ 7, 8, 53 },
		{ 7, 9, 53 },
		{ 7, 10, 53 },
		{ 7, 11, 53 },
		{ 7, 12, 53 },
		{ 7, 13, 53 },
		{ 7, 14, 53 },
		{ 7, 15, 53 },
		{ 7, 16, 53 },
		{ 7, 17, 53 },
		{ 7, 18, 53 },
		{ 7, 19, 53 },
		{ 7, 20, 53 },
		{ 7, 21, 53 },
		{ 7, 22, 53 },
		{ 7, 23, 53 },
		{ 7, 24, 53 },
		{ 7, 25, 53 },
		{ 7, 26, 53 },
		{ 7, 27, 53 },
		{ 7, 28, 53 },
		{ 7, 29, 53 },
		{ 7, 30, 53 },
		{ 7, 31, 53 },
		{ 7, 32, 53 },
		{ 7, 33, 53 },
		{ 7, 34, 53 },
		{ 7, 35, 53 },
		{ 7, 36, 53 },
		{ 7, 37, 53 },
		{ 7, 38, 53 },
		{ 7, 39, 53 },
		{ 7, 40, 53 },
		{ 7, 41, 53 },
		{ 7, 42, 53 },
		{ 7, 43, 53 },
		{ 7, 44, 53 },
		{ 7, 45, 53 },
		{ 7, 46, 53 },
		{ 7, 47, 53 },
		{ 7, 48, 53 },
		{ 7, 49, 53 },
		{ 7, 50, 53 },
		{ 7, 51, 53 },
		{ 7, 52, 53 },
		{ 7, 53, 53 },
		{ 7, 54, 53 },
		{ 7, 55, 53 },
		{ 7, 56, 53 },
		{ 7, 57, 53 },
		{ 7, 58, 53 },
		{ 7, 59, 53 },
		{ 7, 60, 53 },
		{ 7, 61, 53 },
		{ 7, 62, 53 },
		{ 7, 63, 53 },
		{ 7, 64, 53 },
		{ 7, 65, 53 },
		{ 7, 66, 53 },
		{ 7, 67, 53 },
		{ 7, 68, 53 },
		{ 7, 69, 53 },
		{ 8, 9, 53 },
		{ 8, 10, 53 },
		{ 8, 11, 53 },
		{ 8, 12, 53 },
		{ 8, 13, 53 },
		{ 8, 14, 53 },
		{ 8, 15, 53 },
		{ 8, 16, 53 },
		{ 8, 17, 53 },
		{ 8, 18, 53 },
		{ 8, 19, 53 },
		{ 8, 20, 53 },
		{ 8, 21, 53 },
		{ 8, 22, 53 },
		{ 8, 23, 53 },
		{ 8, 24, 53 },
		{ 8, 25, 53 },
		{ 8, 26, 53 },
		{ 8, 27, 53 },
		{ 8, 28, 53 },
		{ 8, 29, 53 },
		{ 8, 30, 53 },
		{ 8, 31, 53 },
		{ 8, 32, 53 },
		{ 8, 33, 53 },
		{ 8, 34, 53 },
		{ 8, 35, 53 },
		{ 8, 36, 53 },
		{ 8, 37, 53 },
		{ 8, 38, 53 },
		{ 8, 39, 53 },
		{ 8, 40, 53 },
		{ 8, 41, 53 },
		{ 8, 42, 53 },
		{ 8, 43, 53 },
		{ 8, 44, 53 },
		{ 8, 45, 53 },
		{ 8, 46, 53 },
		{ 8, 47, 53 },
		{ 8, 48, 53 },
		{ 8, 49, 53 },
		{ 8, 50, 53 },
		{ 8, 51, 53 },
		{ 8, 52, 53 },
		{ 8, 53, 53 },
		{ 8, 54, 53 },
		{ 8, 55, 53 },
		{ 8, 56, 53 },
		{ 8, 57, 53 },
		{ 8, 58, 53 },
		{ 8, 59, 53 },
		{ 8, 60, 53 },
		{ 8, 61, 53 },
		{ 8, 62, 53 },
		{ 8, 63, 53 },
		{ 8, 64, 53 },
		{ 8, 65, 53 },
		{ 8, 66, 53 },
		{ 8, 67, 53 },
		{ 8, 68, 53 },
		{ 8, 69, 53 },
		{ 8, 70, 53 },
		{ 9, 10, 53 },
		{ 9, 11, 53 },
		{ 9, 12, 53 },
		{ 9, 13, 53 },
		{ 9, 14, 53 },
		{ 9, 15, 53 },
		{ 9, 16, 53 },
		{ 9, 17, 53 },
		{ 9, 18, 53 },
		{ 9, 19, 53 },
		{ 9, 20, 53 },
		{ 9, 21, 53 },
		{ 9, 22, 53 },
		{ 9, 23, 53 },
		{ 9, 24, 53 },
		{ 9, 25, 53 },
		{ 9, 26, 53 },
		{ 9, 27, 53 },
		{ 9, 28, 53 },
		{ 9, 29, 53 },
		{ 9, 30, 53 },
		{ 9, 31, 53 },
		{ 9, 32, 53 },
		{ 9, 33, 53 },
		{ 9, 34, 53 },
		{ 9, 35, 53 },
		{ 9, 36, 53 },
		{ 9, 37, 53 },
		{ 9, 38, 53 },
		{ 9, 39, 53 },
		{ 9, 40, 53 },
		{ 9, 41, 53 },
		{ 9, 42, 53 },
		{ 9, 43, 53 },
		{ 9, 44, 53 },
		{ 9, 45, 53 },
		{ 9, 46, 53 },
		{ 9, 47, 53 },
		{ 9, 48, 53 },
		{ 9, 49, 53 },
		{ 9, 50, 53 },
		{ 9, 51, 53 },
		{ 9, 52, 53 },
		{ 9, 53, 53 },
		{ 9, 54, 53 },
		{ 9, 55, 53 },
		{ 9, 56, 53 },
		{ 9, 57, 53 },
		{ 9, 58, 53 },
		{ 9, 59, 53 },
		{ 9, 60, 53 },
		{ 9, 61, 53 },
		{ 9, 62, 53 },
		{ 9, 63, 53 },
		{ 9, 64, 53 },
		{ 9, 65, 53 },
		{ 9, 66, 53 },
		{ 9, 67, 53 },
		{ 9, 68, 53 },
		{ 9, 69, 53 },
		{ 9, 70, 53 },
		{ 10, 11, 53 },
		{ 10, 12, 53 },
		{ 10, 13, 53 },
		{ 10, 14, 53 },
		{ 10, 15, 53 },
		{ 10, 16, 53 },
		{ 10, 17, 53 },
		{ 10, 18, 53 },
		{ 10, 19, 53 },
		{ 10, 20, 53 },
		{ 10, 21, 53 },
		{ 10, 22, 53 },
		{ 10, 23, 53 },
		{ 10, 24, 53 },
		{ 10, 25, 53 },
		{ 10, 26, 53 },
		{ 10, 27, 53 },
		{ 10, 28, 53 },
		{ 10, 29, 53 },
		{ 10, 30, 53 },
		{ 10, 31, 53 },
		{ 10, 32, 53 },
		{ 10, 33, 53 },
		{ 10, 34, 53 },
		{ 10, 35, 53 },
		{ 10, 36, 53 },
		{ 10, 37, 53 },
		{ 10, 38, 53 },
		{ 10, 39, 53 },
		{ 10, 40, 53 },
		{ 10, 41, 53 },
		{ 10, 42, 53 },
		{ 10, 43, 53 },
		{ 10, 44, 53 },
		{ 10, 45, 53 },
		{ 10, 46, 53 },
		{ 10, 47, 53 },
		{ 10, 48, 53 },
		{ 10, 49, 53 },
		{ 10, 50, 53 },
		{ 10, 51, 53 },
		{ 10, 52, 53 },
		{ 10, 53, 53 },
		{ 10, 54, 53 },
		{ 10, 55, 53 },
		{ 10, 56, 53 },
		{ 10, 57, 53 },
		{ 10, 58, 53 },
		{ 10, 59, 53 },
		{ 10, 60, 53 },
		{ 10, 61, 53 },
		{ 10, 62, 53 },
		{ 10, 63, 53 },
		{ 10, 64, 53 },
		{ 10, 65, 53 },
		{ 10, 66, 53 },
		{ 10, 67, 53 },
		{ 10, 68, 53 },
		{ 10, 69, 53 },
		{ 10, 70, 53 },
		{ 10, 71, 53 },
		{ 11, 12, 53 },
		{ 11, 13, 53 },
		{ 11, 14, 53 },
		{ 11, 15, 53 },
		{ 11, 16, 53 },
		{ 11, 17, 53 },
		{ 11, 18, 53 },
		{ 11, 19, 53 },
		{ 11, 20, 53 },
		{ 11, 21, 53 },
		{ 11, 22, 53 },
		{ 11, 23, 53 },
		{ 11, 24, 53 },
		{ 11, 25, 53 },
		{ 11, 26, 53 },
		{ 11, 27, 53 },
		{ 11, 28, 53 },
		{ 11, 29, 53 },
		{ 11, 30, 53 },
		{ 11, 31, 53 },
		{ 11, 32, 53 },
		{ 11, 33, 53 },
		{ 11, 34, 53 },
		{ 11, 35, 53 },
		{ 11, 36, 53 },
		{ 11, 37, 53 },
		{ 11, 38, 53 },
		{ 11, 39, 53 },
		{ 11, 40, 53 },
		{ 11, 41, 53 },
		{ 11, 42, 53 },
		{ 11, 43, 53 },
		{ 11, 44, 53 },
		{ 11, 45, 53 },
		{ 11, 46, 53 },
		{ 11, 47, 53 },
		{ 11, 48, 53 },
		{ 11, 49, 53 },
		{ 11, 50, 53 },
		{ 11, 51, 53 },
		{ 11, 52, 53 },
		{ 11, 53, 53 },
		{ 11, 54, 53 },
		{ 11, 55, 53 },
		{ 11, 56, 53 },
		{ 11, 57, 53 },
		{ 11, 58, 53 },
		{ 11, 59, 53 },
		{ 11, 60, 53 },
		{ 11, 61, 53 },
		{ 11, 62, 53 },
		{ 11, 63, 53 },
		{ 11, 64, 53 },
		{ 11, 65, 53 },
		{ 11, 66, 53 },
		{ 11, 67, 53 },
		{ 11, 68, 53 },
		{ 11, 69, 53 },
		{ 11, 70, 53 },
		{ 11, 71, 53 },
		{ 11, 72, 53 },
		{ 12, 13, 53 },
		{ 12, 14, 53 },
		{ 12, 15, 53 },
		{ 12, 16, 53 },
		{ 12, 17, 53 },
		{ 12, 18, 53 },
		{ 12, 19, 53 },
		{ 12, 20, 53 },
		{ 12, 21, 53 },
		{ 12, 22, 53 },
		{ 12, 23, 53 },
		{ 12, 24, 53 },
		{ 12, 25, 53 },
		{ 12, 26, 53 },
		{ 12, 27, 53 },
		{ 12, 28, 53 },
		{ 12, 29, 53 },
		{ 12, 30, 53 },
		{ 12, 31, 53 },
		{ 12, 32, 53 },
		{ 12, 33, 53 },
		{ 12, 34, 53 },
		{ 12, 35, 53 },
		{ 12, 36, 53 },
		{ 12, 37, 53 },
		{ 12, 38, 53 },
		{ 12, 39, 53 },
		{ 12, 40, 53 },
		{ 12, 41, 53 },
		{ 12, 42, 53 },
		{ 12, 43, 53 },
		{ 12, 44, 53 },
		{ 12, 45, 53 },
		{ 12, 46, 53 },
		{ 12, 47, 53 },
		{ 12, 48, 53 },
		{ 12, 49, 53 },
		{ 12, 50, 53 },
		{ 12, 51, 53 },
		{ 12, 52, 53 },
		{ 12, 53, 53 },
		{ 12, 54, 53 },
		{ 12, 55, 53 },
		{ 12, 56, 53 },
		{ 12, 57, 53 },
		{ 12, 58, 53 },
		{ 12, 59, 53 },
		{ 12, 60, 53 },
		{ 12, 61, 53 },
		{ 12, 62, 53 },
		{ 12, 63, 53 },
		{ 12, 64, 53 },
		{ 12, 65, 53 },
		{ 12, 66, 53 },
		{ 12, 67, 53 },
		{ 12, 68, 53 },
		{ 12, 69, 53 },
		{ 12, 70, 53 },
		{ 12, 71, 53 },
		{ 12, 72, 53 },
		{ 12, 73, 53 },
		{ 13, 14, 53 },
		{ 13, 15, 53 },
		{ 13, 16, 53 },
		{ 13, 17, 53 },
		{ 13, 18, 53 },
		{ 13, 19, 53 },
		{ 13, 20, 53 },
		{ 13, 21, 53 },
		{ 13, 22, 53 },
		{ 13, 23, 53 },
		{ 13, 24, 53 },
		{ 13, 25, 53 },
		{ 13, 26, 53 },
		{ 13, 27, 53 },
		{ 13, 28, 53 },
		{ 13, 29, 53 },
		{ 13, 30, 53 },
		{ 13, 31, 53 },
		{ 13, 32, 53 },
		{ 13, 33, 53 },
		{ 13, 34, 53 },
		{ 13, 35, 53 },
		{ 13, 36, 53 },
		{ 13, 37, 53 },
		{ 13, 38, 53 },
		{ 13, 39, 53 },
		{ 13, 40, 53 },
		{ 13, 41, 53 },
		{ 13, 42, 53 },
		{ 13, 43, 53 },
		{ 13, 44, 53 },
		{ 13, 45, 53 },
		{ 13, 46, 53 },
		{ 13, 47, 53 },
		{ 13, 48, 53 },
		{ 13, 49, 53 },
		{ 13, 50, 53 },
		{ 13, 51, 53 },
		{ 13, 52, 53 },
		{ 13, 53, 53 },
		{ 13, 54, 53 },
		{ 13, 55, 53 },
		{ 13, 56, 53 },
		{ 13, 57, 53 },
		{ 13, 58, 53 },
		{ 13, 59, 53 },
		{ 13, 60, 53 },
		{ 13, 61, 53 },
		{ 13, 62, 53 },
		{ 13, 63, 53 },
		{ 13, 64, 53 },
		{ 13, 65, 53 },
		{ 13, 66, 53 },
		{ 13, 67, 53 },
		{ 13, 68, 53 },
		{ 13, 69, 53 },
		{ 13, 70, 53 },
		{ 13, 71, 53 },
		{ 13, 72, 53 },
		{ 13, 73, 53 },
		{ 14, 15, 53 },
		{ 14, 16, 53 },
		{ 14, 17, 53 },
		{ 14, 18, 53 },
		{ 14, 19, 53 },
		{ 14, 20, 53 },
		{ 14, 21, 53 },
		{ 14, 22, 53 },
		{ 14, 23, 53 },
		{ 14, 24, 53 },
		{ 14, 25, 53 },
		{ 14, 26, 53 },
		{ 14, 27, 53 },
		{ 14, 28, 53 },
		{ 14, 29, 53 },
		{ 14, 30, 53 },
		{ 14, 31, 53 },
		{ 14, 32, 53 },
		{ 14, 33, 53 },
		{ 14, 34, 53 },
		{ 14, 35, 53 },
		{ 14, 36, 53 },
		{ 14, 37, 53 },
		{ 14, 38, 53 },
		{ 14, 39, 53 },
		{ 14, 40, 53 },
		{ 14, 41, 53 },
		{ 14, 42, 53 },
		{ 14, 43, 53 },
		{ 14, 44, 53 },
		{ 14, 45, 53 },
		{ 14, 46, 53 },
		{ 14, 47, 53 },
		{ 14, 48, 53 },
		{ 14, 49, 53 },
		{ 14, 50, 53 },
		{ 14, 51, 53 },
		{ 14, 52, 53 },
		{ 14, 53, 53 },
		{ 14, 54, 53 },
		{ 14, 55, 53 },
		{ 14, 56, 53 },
		{ 14, 57, 53 },
		{ 14, 58, 53 },
		{ 14, 59, 53 },
		{ 14, 60, 53 },
		{ 14, 61, 53 },
		{ 14, 62, 53 },
		{ 14, 63, 53 },
		{ 14, 64, 53 },
		{ 14, 65, 53 },
		{ 14, 66, 53 },
		{ 14, 67, 53 },
		{ 14, 68, 53 },
		{ 14, 69, 53 },
		{ 14, 70, 53 },
		{ 14, 71, 53 },
		{ 14, 72, 53 },
		{ 14, 73, 53 },
		{ 14, 74, 53 },
		{ 15, 16, 53 },
		{ 15, 17, 53 },
		{ 15, 18, 53 },
		{ 15, 19, 53 },
		{ 15, 20, 53 },
		{ 15, 21, 53 },
		{ 15, 22, 53 },
		{ 15, 23, 53 },
		{ 15, 24, 53 },
		{ 15, 25, 53 },
		{ 15, 26, 53 },
		{ 15, 27, 53 },
		{ 15, 28, 53 },
		{ 15, 29, 53 },
		{ 15, 30, 53 },
		{ 15, 31, 53 },
		{ 15, 32, 53 },
		{ 15, 33, 53 },
		{ 15, 34, 53 },
		{ 15, 35, 53 },
		{ 15, 36, 53 },
		{ 15, 37, 53 },
		{ 15, 38, 53 },
		{ 15, 39, 53 },
		{ 15, 40, 53 },
		{ 15, 41, 53 },
		{ 15, 42, 53 },
		{ 15, 43, 53 },
		{ 15, 44, 53 },
		{ 15, 45, 53 },
		{ 15, 46, 53 },
		{ 15, 47, 53 },
		{ 15, 48, 53 },
		{ 15, 49, 53 },
		{ 15, 50, 53 },
		{ 15, 51, 53 },
		{ 15, 52, 53 },
		{ 15, 53, 53 },
		{ 15, 54, 53 },
		{ 15, 55, 53 },
		{ 15, 56, 53 },
		{ 15, 57, 53 },
		{ 15, 58, 53 },
		{ 15, 59, 53 },
		{ 15, 60, 53 },
		{ 15, 61, 53 },
		{ 15, 62, 53 },
		{ 15, 63, 53 },
		{ 15, 64, 53 },
		{ 15, 65, 53 },
		{ 15, 66, 53 },
		{ 15, 67, 53 },
		{ 15, 68, 53 },
		{ 15, 69, 53 },
		{ 15, 70, 53 },
		{ 15, 71, 53 },
		{ 15, 72, 53 },
		{ 15, 73, 53 },
		{ 15, 74, 53 },
		{ 15, 75, 53 },
		{ 16, 17, 53 },
		{ 16, 18, 53 },
		{ 16, 19, 53 },
		{ 16, 20, 53 },
		{ 16, 21, 53 },
		{ 16, 22, 53 },
		{ 16, 23, 53 },
		{ 16, 24, 53 },
		{ 16, 25, 53 },
		{ 16, 26, 53 },
		{ 16, 27, 53 },
		{ 16, 28, 53 },
		{ 16, 29, 53 },
		{ 16, 30, 53 },
		{ 16, 31, 53 },
		{ 16, 32, 53 },
		{ 16, 33, 53 },
		{ 16, 34, 53 },
		{ 16, 35, 53 },
		{ 16, 36, 53 },
		{ 16, 37, 53 },
		{ 16, 38, 53 },
		{ 16, 39, 53 },
		{ 16, 40, 53 },
		{ 16, 41, 53 },
		{ 16, 42, 53 },
		{ 16, 43, 53 },
		{ 16, 44, 53 },
		{ 16, 45, 53 },
		{ 16, 46, 53 },
		{ 16, 47, 53 },
		{ 16, 48, 53 },
		{ 16, 49, 53 },
		{ 16, 50, 53 },
		{ 16, 51, 53 },
		{ 16, 52, 53 },
		{ 16, 53, 53 },
		{ 16, 54, 53 },
		{ 16, 55, 53 },
		{ 16, 56, 53 },
		{ 16, 57, 53 },
		{ 16, 58, 53 },
		{ 16, 59, 53 },
		{ 16, 60, 53 },
		{ 16, 61, 53 },
		{ 16, 62, 53 },
		{ 16, 63, 53 },
		{ 16, 64, 53 },
		{ 16, 65, 53 },
		{ 16, 66, 53 },
		{ 16, 67, 53 },
		{ 16, 68, 53 },
		{ 16, 69, 53 },
		{ 16, 70, 53 },
		{ 16, 71, 53 },
		{ 16, 72, 53 },
		{ 16, 73, 53 },
		{ 16, 74, 53 },
		{ 16, 75, 53 },
		{ 16, 76, 53 },
		{ 17, 18, 53 },
		{ 17, 19, 53 },
		{ 17, 20, 53 },
		{ 17, 21, 53 },
		{ 17, 22, 53 },
		{ 17, 23, 53 },
		{ 17, 24, 53 },
		{ 17, 25, 53 },
		{ 17, 26, 53 },
		{ 17, 27, 53 },
		{ 17, 28, 53 },
		{ 17, 29, 53 },
		{ 17, 30, 53 },
		{ 17, 31, 53 },
		{ 17, 32, 53 },
		{ 17, 33, 53 },
		{ 17, 34, 53 },
		{ 17, 35, 53 },
		{ 17, 36, 53 },
		{ 17, 37, 53 },
		{ 17, 38, 53 },
		{ 17, 39, 53 },
		{ 17, 40, 53 },
		{ 17, 41, 53 },
		{ 17, 42, 53 },
		{ 17, 43, 53 },
		{ 17, 44, 53 },
		{ 17, 45, 53 },
		{ 17, 46, 53 },
		{ 17, 47, 53 },
		{ 17, 48, 53 },
		{ 17, 49, 53 },
		{ 17, 50, 53 },
		{ 17, 51, 53 },
		{ 17, 52, 53 },
		{ 17, 53, 53 },
		{ 17, 54, 53 },
		{ 17, 55, 53 },
		{ 17, 56, 53 },
		{ 17, 57, 53 },
		{ 17, 58, 53 },
		{ 17, 59, 53 },
		{ 17, 60, 53 },
		{ 17, 61, 53 },
		{ 17, 62, 53 },
		{ 17, 63, 53 },
		{ 17, 64, 53 },
		{ 17, 65, 53 },
		{ 17, 66, 53 },
		{ 17, 67, 53 },
		{ 17, 68, 53 },
		{ 17, 69, 53 },
		{ 17, 70, 53 },
		{ 17, 71, 53 },
		{ 17, 72, 53 },
		{ 17, 73, 53 },
		{ 17, 74, 53 },
		{ 17, 75, 53 },
		{ 17, 76, 53 },
		{ 18, 19, 53 },
		{ 18, 20, 53 },
		{ 18, 21, 53 },
		{ 18, 22, 53 },
		{ 18, 23, 53 },
		{ 18, 24, 53 },
		{ 18, 25, 53 },
		{ 18, 26, 53 },
		{ 18, 27, 53 },
		{ 18, 28, 53 },
		{ 18, 29, 53 },
		{ 18, 30, 53 },
		{ 18, 31, 53 },
		{ 18, 32, 53 },
		{ 18, 33, 53 },
		{ 18, 34, 53 },
		{ 18, 35, 53 },
		{ 18, 36, 53 },
		{ 18, 37, 53 },
		{ 18, 38, 53 },
		{ 18, 39, 53 },
		{ 18, 40, 53 },
		{ 18, 41, 53 },
		{ 18, 42, 53 },
		{ 18, 43, 53 },
		{ 18, 44, 53 },
		{ 18, 45, 53 },
		{ 18, 46, 53 },
		{ 18, 47, 53 },
		{ 18, 48, 53 },
		{ 18, 49, 53 },
		{ 18, 50, 53 },
		{ 18, 51, 53 },
		{ 18, 52, 53 },
		{ 18, 53, 53 },
		{ 18, 54, 53 },
		{ 18, 55, 53 },
		{ 18, 56, 53 },
		{ 18, 57, 53 },
		{ 18, 58, 53 },
		{ 18, 59, 53 },
		{ 18, 60, 53 },
		{ 18, 61, 53 },
		{ 18, 62, 53 },
		{ 18, 63, 53 },
		{ 18, 64, 53 },
		{ 18, 65, 53 },
		{ 18, 66, 53 },
		{ 18, 67, 53 },
		{ 18, 68, 53 },
		{ 18, 69, 53 },
		{ 18, 70, 53 },
		{ 18, 71, 53 },
		{ 18, 72, 53 },
		{ 18, 73, 53 },
		{ 18, 74, 53 },
		{ 18, 75, 53 },
		{ 18, 76, 53 },
		{ 18, 77, 53 },
		{ 19, 20, 53 },
		{ 19, 21, 53 },
		{ 19, 22, 53 },
		{ 19, 23, 53 },
		{ 19, 24, 53 },
		{ 19, 25, 53 },
		{ 19, 26, 53 },
		{ 19, 27, 53 },
		{ 19, 28, 53 },
		{ 19, 29, 53 },
		{ 19, 30, 53 },
		{ 19, 31, 53 },
		{ 19, 32, 53 },
		{ 19, 33, 53 },
		{ 19, 34, 53 },
		{ 19, 35, 53 },
		{ 19, 36, 53 },
		{ 19, 37, 53 },
		{ 19, 38, 53 },
		{ 19, 39, 53 },
		{ 19, 40, 53 },
		{ 19, 41, 53 },
		{ 19, 42, 53 },
		{ 19, 43, 53 },
		{ 19, 44, 53 },
		{ 19, 45, 53 },
		{ 19, 46, 53 },
		{ 19, 47, 53 },
		{ 19, 48, 53 },
		{ 19, 49, 53 },
		{ 19, 50, 53 },
		{ 19, 51, 53 },
		{ 19, 52, 53 },
		{ 19, 53, 53 },
		{ 19, 54, 53 },
		{ 19, 55, 53 },
		{ 19, 56, 53 },
		{ 19, 57, 53 },
		{ 19, 58, 53 },
		{ 19, 59, 53 },
		{ 19, 60, 53 },
		{ 19, 61, 53 },
		{ 19, 62, 53 },
		{ 19, 63, 53 },
		{ 19, 64, 53 },
		{ 19, 65, 53 },
		{ 19, 66, 53 },
		{ 19, 67, 53 },
		{ 19, 68, 53 },
		{ 19, 69, 53 },
		{ 19, 70, 53 },
		{ 19, 71, 53 },
		{ 19, 72, 53 },
		{ 19, 73, 53 },
		{ 19, 74, 53 },
		{ 19, 75, 53 },
		{ 19, 76, 53 },
		{ 19, 77, 53 },
		{ 19, 78, 53 },
		{ 20, 21, 53 },
		{ 20, 22, 53 },
		{ 20, 23, 53 },
		{ 20, 24, 53 },
		{ 20, 25, 53 },
		{ 20, 26, 53 },
		{ 20, 27, 53 },
		{ 20, 28, 53 },
		{ 20, 29, 53 },
		{ 20, 30, 53 },
		{ 20, 31, 53 },
		{ 20, 32, 53 },
		{ 20, 33, 53 },
		{ 20, 34, 53 },
		{ 20, 35, 53 },
		{ 20, 36, 53 },
		{ 20, 37, 53 },
		{ 20, 38, 53 },
		{ 20, 39, 53 },
		{ 20, 40, 53 },
		{ 20, 41, 53 },
		{ 20, 42, 53 },
		{ 20, 43, 53 },
		{ 20, 44, 53 },
		{ 20, 45, 53 },
		{ 20, 46, 53 },
		{ 20, 47, 53 },
		{ 20, 48, 53 },
		{ 20, 49, 53 },
		{ 20, 50, 53 },
		{ 20, 51, 53 },
		{ 20, 52, 53 },
		{ 20, 53, 53 },
		{ 20, 54, 53 },
		{ 20, 55, 53 },
		{ 20, 56, 53 },
		{ 20, 57, 53 },
		{ 20, 58, 53 },
		{ 20, 59, 53 },
		{ 20, 60, 53 },
		{ 20, 61, 53 },
		{ 20, 62, 53 },
		{ 20, 63, 53 },
		{ 20, 64, 53 },
		{ 20, 65, 53 },
		{ 20, 66, 53 },
		{ 20, 67, 53 },
		{ 20, 68, 53 },
		{ 20, 69, 53 },
		{ 20, 70, 53 },
		{ 20, 71, 53 },
		{ 20, 72, 53 },
		{ 20, 73, 53 },
		{ 20, 74, 53 },
		{ 20, 75, 53 },
		{ 20, 76, 53 },
		{ 20, 77, 53 },
		{ 20, 78, 53 },
		{ 20, 79, 53 },
		{ 21, 22, 53 },
		{ 21, 23, 53 },
		{ 21, 24, 53 },
		{ 21, 25, 53 },
		{ 21, 26, 53 },
		{ 21, 27, 53 },
		{ 21, 28, 53 },
		{ 21, 29, 53 },
		{ 21, 30, 53 },
		{ 21, 31, 53 },
		{ 21, 32, 53 },
		{ 21, 33, 53 },
		{ 21, 34, 53 },
		{ 21, 35, 53 },
		{ 21, 36, 53 },
		{ 21, 37, 53 },
		{ 21, 38, 53 },
		{ 21, 39, 53 },
		{ 21, 40, 53 },
		{ 21, 41, 53 },
		{ 21, 42, 53 },
		{ 21, 43, 53 },
		{ 21, 44, 53 },
		{ 21, 45, 53 },
		{ 21, 46, 53 },
		{ 21, 47, 53 },
		{ 21, 48, 53 },
		{ 21, 49, 53 },
		{ 21, 50, 53 },
		{ 21, 51, 53 },
		{ 21, 52, 53 },
		{ 21, 53, 53 },
		{ 21, 54, 53 },
		{ 21, 55, 53 },
		{ 21, 56, 53 },
		{ 21, 57, 53 },
		{ 21, 58, 53 },
		{ 21, 59, 53 },
		{ 21, 60, 53 },
		{ 21, 61, 53 },
		{ 21, 62, 53 },
		{ 21, 63, 53 },
		{ 21, 64, 53 },
		{ 21, 65, 53 },
		{ 21, 66, 53 },
		{ 21, 67, 53 },
		{ 21, 68, 53 },
		{ 21, 69, 53 },
		{ 21, 70, 53 },
		{ 21, 71, 53 },
		{ 21, 72, 53 },
		{ 21, 73, 53 },
		{ 21, 74, 53 },
		{ 21, 75, 53 },
		{ 21, 76, 53 },
		{ 21, 77, 53 },
		{ 21, 78, 53 },
		{ 21, 79, 53 },
		{ 22, 23, 53 },
		{ 22, 24, 53 },
		{ 22, 25, 53 },
		{ 22, 26, 53 },
		{ 22, 27, 53 },
		{ 22, 28, 53 },
		{ 22, 29, 53 },
		{ 22, 30, 53 },
		{ 22, 31, 53 },
		{ 22, 32, 53 },
		{ 22, 33, 53 },
		{ 22, 34, 53 },
		{ 22, 35, 53 },
		{ 22, 36, 53 },
		{ 22, 37, 53 },
		{ 22, 38, 53 },
		{ 22, 39, 53 },
		{ 22, 40, 53 },
		{ 22, 41, 53 },
		{ 22, 42, 53 },
		{ 22, 43, 53 },
		{ 22, 44, 53 },
		{ 22, 45, 53 },
		{ 22, 46, 53 },
		{ 22, 47, 53 },
		{ 22, 48, 53 },
		{ 22, 49, 53 },
		{ 22, 50, 53 },
		{ 22, 51, 53 },
		{ 22, 52, 53 },
		{ 22, 53, 53 },
		{ 22, 54, 53 },
		{ 22, 55, 53 },
		{ 22, 56, 53 },
		{ 22, 57, 53 },
		{ 22, 58, 53 },
		{ 22, 59, 53 },
		{ 22, 60, 53 },
		{ 22, 61, 53 },
		{ 22, 62, 53 },
		{ 22, 63, 53 },
		{ 22, 64, 53 },
		{ 22, 65, 53 },
		{ 22, 66, 53 },
		{ 22, 67, 53 },
		{ 22, 68, 53 },
		{ 22, 69, 53 },
		{ 22, 70, 53 },
		{ 22, 71, 53 },
		{ 22, 72, 53 },
		{ 22, 73, 53 },
		{ 22, 74, 53 },
		{ 22, 75, 53 },
		{ 22, 76, 53 },
		{ 22, 77, 53 },
		{ 22, 78, 53 },
		{ 22, 79, 53 },
		{ 22, 80, 53 },
		{ 23, 24, 53 },
		{ 23, 25, 53 },
		{ 23, 26, 53 },
		{ 23, 27, 53 },
		{ 23, 28, 53 },
		{ 23, 29, 53 },
		{ 23, 30, 53 },
		{ 23, 31, 53 },
		{ 23, 32, 53 },
		{ 23, 33, 53 },
		{ 23, 34, 53 },
		{ 23, 35, 53 },
		{ 23, 36, 53 },
		{ 23, 37, 53 },
		{ 23, 38, 53 },
		{ 23, 39, 53 },
		{ 23, 40, 53 },
		{ 23, 41, 53 },
		{ 23, 42, 53 },
		{ 23, 43, 53 },
		{ 23, 44, 53 },
		{ 23, 45, 53 },
		{ 23, 46, 53 },
		{ 23, 47, 53 },
		{ 23, 48, 53 },
		{ 23, 49, 53 },
		{ 23, 50, 53 },
		{ 23, 51, 53 },
		{ 23, 52, 53 },
		{ 23, 53, 53 },
		{ 23, 54, 53 },
		{ 23, 55, 53 },
		{ 23, 56, 53 },
		{ 23, 57, 53 },
		{ 23, 58, 53 },
		{ 23, 59, 53 },
		{ 23, 60, 53 },
		{ 23, 61, 53 },
		{ 23, 62, 53 },
		{ 23, 63, 53 },
		{ 23, 64, 53 },
		{ 23, 65, 53 },
		{ 23, 66, 53 },
		{ 23, 67, 53 },
		{ 23, 68, 53 },
		{ 23, 69, 53 },
		{ 23, 70, 53 },
		{ 23, 71, 53 },
		{ 23, 72, 53 },
		{ 23, 73, 53 },
		{ 23, 74, 53 },
		{ 23, 75, 53 },
		{ 23, 76, 53 },
		{ 23, 77, 53 },
		{ 23, 78, 53 },
		{ 23, 79, 53 },
		{ 23, 80, 53 },
		{ 23, 81, 53 },
		{ 24, 25, 53 },
		{ 24, 26, 53 },
		{ 24, 27, 53 },
		{ 24, 28, 53 },
		{ 24, 29, 53 },
		{ 24, 30, 53 },
		{ 24, 31, 53 },
		{ 24, 32, 53 },
		{ 24, 33, 53 },
		{ 24, 34, 53 },
		{ 24, 35, 53 },
		{ 24, 36, 53 },
		{ 24, 37, 53 },
		{ 24, 38, 53 },
		{ 24, 39, 53 },
		{ 24, 40, 53 },
		{ 24, 41, 53 },
		{ 24, 42, 53 },
		{ 24, 43, 53 },
		{ 24, 44, 53 },
		{ 24, 45, 53 },
		{ 24, 46, 53 },
		{ 24, 47, 53 },
		{ 24, 48, 53 },
		{ 24, 49, 53 },
		{ 24, 50, 53 },
		{ 24, 51, 53 },
		{ 24, 52, 53 },
		{ 24, 53, 53 },
		{ 24, 54, 53 },
		{ 24, 55, 53 },
		{ 24, 56, 53 },
		{ 24, 57, 53 },
		{ 24, 58, 53 },
		{ 24, 59, 53 },
		{ 24, 60, 53 },
		{ 24, 61, 53 },
		{ 24, 62, 53 },
		{ 24, 63, 53 },
		{ 24, 64, 53 },
		{ 24, 65, 53 },
		{ 24, 66, 53 },
		{ 24, 67, 53 },
		{ 24, 68, 53 },
		{ 24, 69, 53 },
		{ 24, 70, 53 },
		{ 24, 71, 53 },
		{ 24, 72, 53 },
		{ 24, 73, 53 },
		{ 24, 74, 53 },
		{ 24, 75, 53 },
		{ 24, 76, 53 },
		{ 24, 77, 53 },
		{ 24, 78, 53 },
		{ 24, 79, 53 },
		{ 24, 80, 53 },
		{ 24, 81, 53 },
		{ 24, 82, 53 },
		{ 25, 26, 53 },
		{ 25, 27, 53 },
		{ 25, 28, 53 },
		{ 25, 29, 53 },
		{ 25, 30, 53 },
		{ 25, 31, 53 },
		{ 25, 32, 53 },
		{ 25, 33, 53 },
		{ 25, 34, 53 },
		{ 25, 35, 53 },
		{ 25, 36, 53 },
		{ 25, 37, 53 },
		{ 25, 38, 53 },
		{ 25, 39, 53 },
		{ 25, 40, 53 },
		{ 25, 41, 53 },
		{ 25, 42, 53 },
		{ 25, 43, 53 },
		{ 25, 44, 53 },
		{ 25, 45, 53 },
		{ 25, 46, 53 },
		{ 25, 47, 53 },
		{ 25, 48, 53 },
		{ 25, 49, 53 },
		{ 25, 50, 53 },
		{ 25, 51, 53 },
		{ 25, 52, 53 },
		{ 25, 53, 53 },
		{ 25, 54, 53 },
		{ 25, 55, 53 },
		{ 25, 56, 53 },
		{ 25, 57, 53 },
		{ 25, 58, 53 },
		{ 25, 59, 53 },
		{ 25, 60, 53 },
		{ 25, 61, 53 },
		{ 25, 62, 53 },
		{ 25, 63, 53 },
		{ 25, 64, 53 },
		{ 25, 65, 53 },
		{ 25, 66, 53 },
		{ 25, 67, 53 },
		{ 25, 68, 53 },
		{ 25, 69, 53 },
		{ 25, 70, 53 },
		{ 25, 71, 53 },
		{ 25, 72, 53 },
		{ 25, 73, 53 },
		{ 25, 74, 53 },
		{ 25, 75, 53 },
		{ 25, 76, 53 },
		{ 25, 77, 53 },
		{ 25, 78, 53 },
		{ 25, 79, 53 },
		{ 25, 80, 53 },
		{ 25, 81, 53 },
		{ 25, 82, 53 },
		{ 26, 27, 53 },
		{ 26, 28, 53 },
		{ 26, 29, 53 },
		{ 26, 30, 53 },
		{ 26, 31, 53 },
		{ 26, 32, 53 },
		{ 26, 33, 53 },
		{ 26, 34, 53 },
		{ 26, 35, 53 },
		{ 26, 36, 53 },
		{ 26, 37, 53 },
		{ 26, 38, 53 },
		{ 26, 39, 53 },
		{ 26, 40, 53 },
		{ 26, 41, 53 },
		{ 26, 42, 53 },
		{ 26, 43, 53 },
		{ 26, 44, 53 },
		{ 26, 45, 53 },
		{ 26, 46, 53 },
		{ 26, 47, 53 },
		{ 26, 48, 53 },
		{ 26, 49, 53 },
		{ 26, 50, 53 },
		{ 26, 51, 53 },
		{ 26, 52, 53 },
		{ 26, 53, 53 },
		{ 26, 54, 53 },
		{ 26, 55, 53 },
		{ 26, 56, 53 },
		{ 26, 57, 53 },
		{ 26, 58, 53 },
		{ 26, 59, 53 },
		{ 26, 60, 53 },
		{ 26, 61, 53 },
		{ 26, 62, 53 },
		{ 26, 63, 53 },
		{ 26, 64, 53 },
		{ 26, 65, 53 },
		{ 26, 66, 53 },
		{ 26, 67, 53 },
		{ 26, 68, 53 },
		{ 26, 69, 53 },
		{ 26, 70, 53 },
		{ 26, 71, 53 },
		{ 26, 72, 53 },
		{ 26, 73, 53 },
		{ 26, 74, 53 },
		{ 26, 75, 53 },
		{ 26, 76, 53 },
		{ 26, 77, 53 },
		{ 26, 78, 53 },
		{ 26, 79, 53 },
		{ 26, 80, 53 },
		{ 26, 81, 53 },
		{ 26, 82, 53 },
		{ 26, 83, 53 },
		{ 27, 28, 53 },
		{ 27, 29, 53 },
		{ 27, 30, 53 },
		{ 27, 31, 53 },
		{ 27, 32, 53 },
		{ 27, 33, 53 },
		{ 27, 34, 53 },
		{ 27, 35, 53 },
		{ 27, 36, 53 },
		{ 27, 37, 53 },
		{ 27, 38, 53 },
		{ 27, 39, 53 },
		{ 27, 40, 53 },
		{ 27, 41, 53 },
		{ 27, 42, 53 },
		{ 27, 43, 53 },
		{ 27, 44, 53 },
		{ 27, 45, 53 },
		{ 27, 46, 53 },
		{ 27, 47, 53 },
		{ 27, 48, 53 },
		{ 27, 49, 53 },
		{ 27, 50, 53 },
		{ 27, 51, 53 },
		{ 27, 52, 53 },
		{ 27, 53, 53 },
		{ 27, 54, 53 },
		{ 27, 55, 53 },
		{ 27, 56, 53 },
		{ 27, 57, 53 },
		{ 27, 58, 53 },
		{ 27, 59, 53 },
		{ 27, 60, 53 },
		{ 27, 61, 53 },
		{ 27, 62, 53 },
		{ 27, 63, 53 },
		{ 27, 64, 53 },
		{ 27, 65, 53 },
		{ 27, 66, 53 },
		{ 27, 67, 53 },
		{ 27, 68, 53 },
		{ 27, 69, 53 },
		{ 27, 70, 53 },
		{ 27, 71, 53 },
		{ 27, 72, 53 },
		{ 27, 73, 53 },
		{ 27, 74, 53 },
		{ 27, 75, 53 },
		{ 27, 76, 53 },
		{ 27, 77, 53 },
		{ 27, 78, 53 },
		{ 27, 79, 53 },
		{ 27, 80, 53 },
		{ 27, 81, 53 },
		{ 27, 82, 53 },
		{ 27, 83, 53 },
		{ 27, 84, 53 },
		{ 28, 29, 53 },
		{ 28, 30, 53 },
		{ 28, 31, 53 },
		{ 28, 32, 53 },
		{ 28, 33, 53 },
		{ 28, 34, 53 },
		{ 28, 35, 53 },
		{ 28, 36, 53 },
		{ 28, 37, 53 },
		{ 28, 38, 53 },
		{ 28, 39, 53 },
		{ 28, 40, 53 },
		{ 28, 41, 53 },
		{ 28, 42, 53 },
		{ 28, 43, 53 },
		{ 28, 44, 53 },
		{ 28, 45, 53 },
		{ 28, 46, 53 },
		{ 28, 47, 53 },
		{ 28, 48, 53 },
		{ 28, 49, 53 },
		{ 28, 50, 53 },
		{ 28, 51, 53 },
		{ 28, 52, 53 },
		{ 28, 53, 53 },
		{ 28, 54, 53 },
		{ 28, 55, 53 },
		{ 28, 56, 53 },
		{ 28, 57, 53 },
		{ 28, 58, 53 },
		{ 28, 59, 53 },
		{ 28, 60, 53 },
		{ 28, 61, 53 },
		{ 28, 62, 53 },
		{ 28, 63, 53 },
		{ 28, 64, 53 },
		{ 28, 65, 53 },
		{ 28, 66, 53 },
		{ 28, 67, 53 },
		{ 28, 68, 53 },
		{ 28, 69, 53 },
		{ 28, 70, 53 },
		{ 28, 71, 53 },
		{ 28, 72, 53 },
		{ 28, 73, 53 },
		{ 28, 74, 53 },
		{ 28, 75, 53 },
		{ 28, 76, 53 },
		{ 28, 77, 53 },
		{ 28, 78, 53 },
		{ 28, 79, 53 },
		{ 28, 80, 53 },
		{ 28, 81, 53 },
		{ 28, 82, 53 },
		{ 28, 83, 53 },
		{ 28, 84, 53 },
		{ 28, 85, 53 },
		{ 29, 30, 53 },
		{ 29, 31, 53 },
		{ 29, 32, 53 },
		{ 29, 33, 53 },
		{ 29, 34, 53 },
		{ 29, 35, 53 },
		{ 29, 36, 53 },
		{ 29, 37, 53 },
		{ 29, 38, 53 },
		{ 29, 39, 53 },
		{ 29, 40, 53 },
		{ 29, 41, 53 },
		{ 29, 42, 53 },
		{ 29, 43, 53 },
		{ 29, 44, 53 },
		{ 29, 45, 53 },
		{ 29, 46, 53 },
		{ 29, 47, 53 },
		{ 29, 48, 53 },
		{ 29, 49, 53 },
		{ 29, 50, 53 },
		{ 29, 51, 53 },
		{ 29, 52, 53 },
		{ 29, 53, 53 },
		{ 29, 54, 53 },
		{ 29, 55, 53 },
		{ 29, 56, 53 },
		{ 29, 57, 53 },
		{ 29, 58, 53 },
		{ 29, 59, 53 },
		{ 29, 60, 53 },
		{ 29, 61, 53 },
		{ 29, 62, 53 },
		{ 29, 63, 53 },
		{ 29, 64, 53 },
		{ 29, 65, 53 },
		{ 29, 66, 53 },
		{ 29, 67, 53 },
		{ 29, 68, 53 },
		{ 29, 69, 53 },
		{ 29, 70, 53 },
		{ 29, 71, 53 },
		{ 29, 72, 53 },
		{ 29, 73, 53 },
		{ 29, 74, 53 },
		{ 29, 75, 53 },
		{ 29, 76, 53 },
		{ 29, 77, 53 },
		{ 29, 78, 53 },
		{ 29, 79, 53 },
		{ 29, 80, 53 },
		{ 29, 81, 53 },
		{ 29, 82, 53 },
		{ 29, 83, 53 },
		{ 29, 84, 53 },
		{ 29, 85, 53 },
		{ 30, 31, 53 },
		{ 30, 32, 53 },
		{ 30, 33, 53 },
		{ 30, 34, 53 },
		{ 30, 35, 53 },
		{ 30, 36, 53 },
		{ 30, 37, 53 },
		{ 30, 38, 53 },
		{ 30, 39, 53 },
		{ 30, 40, 53 },
		{ 30, 41, 53 },
		{ 30, 42, 53 },
		{ 30, 43, 53 },
		{ 30, 44, 53 },
		{ 30, 45, 53 },
		{ 30, 46, 53 },
		{ 30, 47, 53 },
		{ 30, 48, 53 },
		{ 30, 49, 53 },
		{ 30, 50, 53 },
		{ 30, 51, 53 },
		{ 30, 52, 53 },
		{ 30, 53, 53 },
		{ 30, 54, 53 },
		{ 30, 55, 53 },
		{ 30, 56, 53 },
		{ 30, 57, 53 },
		{ 30, 58, 53 },
		{ 30, 59, 53 },
		{ 30, 60, 53 },
		{ 30, 61, 53 },
		{ 30, 62, 53 },
		{ 30, 63, 53 },
		{ 30, 64, 53 },
		{ 30, 65, 53 },
		{ 30, 66, 53 },
		{ 30, 67, 53 },
		{ 30, 68, 53 },
		{ 30, 69, 53 },
		{ 30, 70, 53 },
		{ 30, 71, 53 },
		{ 30, 72, 53 },
		{ 30, 73, 53 },
		{ 30, 74, 53 },
		{ 30, 75, 53 },
		{ 30, 76, 53 },
		{ 30, 77, 53 },
		{ 30, 78, 53 },
		{ 30, 79, 53 },
		{ 30, 80, 53 },
		{ 30, 81, 53 },
		{ 30, 82, 53 },
		{ 30, 83, 53 },
		{ 30, 84, 53 },
		{ 30, 85, 53 },
		{ 30, 86, 53 },
		{ 31, 32, 53 },
		{ 31, 33, 53 },
		{ 31, 34, 53 },
		{ 31, 35, 53 },
		{ 31, 36, 53 },
		{ 31, 37, 53 },
		{ 31, 38, 53 },
		{ 31, 39, 53 },
		{ 31, 40, 53 },
		{ 31, 41, 53 },
		{ 31, 42, 53 },
		{ 31, 43, 53 },
		{ 31, 44, 53 },
		{ 31, 45, 53 },
		{ 31, 46, 53 },
		{ 31, 47, 53 },
		{ 31, 48, 53 },
		{ 31, 49, 53 },
		{ 31, 50, 53 },
		{ 31, 51, 53 },
		{ 31, 52, 53 },
		{ 31, 53, 53 },
		{ 31, 54, 53 },
		{ 31, 55, 53 },
		{ 31, 56, 53 },
		{ 31, 57, 53 },
		{ 31, 58, 53 },
		{ 31, 59, 53 },
		{ 31, 60, 53 },
		{ 31, 61, 53 },
		{ 31, 62, 53 },
		{ 31, 63, 53 },
		{ 31, 64, 53 },
		{ 31, 65, 53 },
		{ 31, 66, 53 },
		{ 31, 67, 53 },
		{ 31, 68, 53 },
		{ 31, 69, 53 },
		{ 31, 70, 53 },
		{ 31, 71, 53 },
		{ 31, 72, 53 },
		{ 31, 73, 53 },
		{ 31, 74, 53 },
		{ 31, 75, 53 },
		{ 31, 76, 53 },
		{ 31, 77, 53 },
		{ 31, 78, 53 },
		{ 31, 79, 53 },
		{ 31, 80, 53 },
		{ 31, 81, 53 },
		{ 31, 82, 53 },
		{ 31, 83, 53 },
		{ 31, 84, 53 },
		{ 31, 85, 53 },
		{ 31, 86, 53 },
		{ 31, 87, 53 },
		{ 32, 33, 53 },
		{ 32, 34, 53 },
		{ 32, 35, 53 },
		{ 32, 36, 53 },
		{ 32, 37, 53 },
		{ 32, 38, 53 },
		{ 32, 39, 53 },
		{ 32, 40, 53 },
		{ 32, 41, 53 },
		{ 32, 42, 53 },
		{ 32, 43, 53 },
		{ 32, 44, 53 },
		{ 32, 45, 53 },
		{ 32, 46, 53 },
		{ 32, 47, 53 },
		{ 32, 48, 53 },
		{ 32, 49, 53 },
		{ 32, 50, 53 },
		{ 32, 51, 53 },
		{ 32, 52, 53 },
		{ 32, 53, 53 },
		{ 32, 54, 53 },
		{ 32, 55, 53 },
		{ 32, 56, 53 },
		{ 32, 57, 53 },
		{ 32, 58, 53 },
		{ 32, 59, 53 },
		{ 32, 60, 53 },
		{ 32, 61, 53 },
		{ 32, 62, 53 },
		{ 32, 63, 53 },
		{ 32, 64, 53 },
		{ 32, 65, 53 },
		{ 32, 66, 53 },
		{ 32, 67, 53 },
		{ 32, 68, 53 },
		{ 32, 69, 53 },
		{ 32, 70, 53 },
		{ 32, 71, 53 },
		{ 32, 72, 53 },
		{ 32, 73, 53 },
		{ 32, 74, 53 },
		{ 32, 75, 53 },
		{ 32, 76, 53 },
		{ 32, 77, 53 },
		{ 32, 78, 53 },
		{ 32, 79, 53 },
		{ 32, 80, 53 },
		{ 32, 81, 53 },
		{ 32, 82, 53 },
		{ 32, 83, 53 },
		{ 32, 84, 53 },
		{ 32, 85, 53 },
		{ 32, 86, 53 },
		{ 32, 87, 53 },
		{ 32, 88, 53 },
		{ 33, 34, 53 },
		{ 33, 35, 53 },
		{ 33, 36, 53 },
		{ 33, 37, 53 },
		{ 33, 38, 53 },
		{ 33, 39, 53 },
		{ 33, 40, 53 },
		{ 33, 41, 53 },
		{ 33, 42, 53 },
		{ 33, 43, 53 },
		{ 33, 44, 53 },
		{ 33, 45, 53 },
		{ 33, 46, 53 },
		{ 33, 47, 53 },
		{ 33, 48, 53 },
		{ 33, 49, 53 },
		{ 33, 50, 53 },
		{ 33, 51, 53 },
		{ 33, 52, 53 },
		{ 33, 53, 53 },
		{ 33, 54, 53 },
		{ 33, 55, 53 },
		{ 33, 56, 53 },
		{ 33, 57, 53 },
		{ 33, 58, 53 },
		{ 33, 59, 53 },
		{ 33, 60, 53 },
		{ 33, 61, 53 },
		{ 33, 62, 53 },
		{ 33, 63, 53 },
		{ 33, 64, 53 },
		{ 33, 65, 53 },
		{ 33, 66, 53 },
		{ 33, 67, 53 },
		{ 33, 68, 53 },
		{ 33, 69, 53 },
		{ 33, 70, 53 },
		{ 33, 71, 53 },
		{ 33, 72, 53 },
		{ 33, 73, 53 },
		{ 33, 74, 53 },
		{ 33, 75, 53 },
		{ 33, 76, 53 },
		{ 33, 77, 53 },
		{ 33, 78, 53 },
		{ 33, 79, 53 },
		{ 33, 80, 53 },
		{ 33, 81, 53 },
		{ 33, 82, 53 },
		{ 33, 83, 53 },
		{ 33, 84, 53 },
		{ 33, 85, 53 },
		{ 33, 86, 53 },
		{ 33, 87, 53 },
		{ 33, 88, 53 },
		{ 34, 35, 53 },
		{ 34, 36, 53 },
		{ 34, 37, 53 },
		{ 34, 38, 53 },
		{ 34, 39, 53 },
		{ 34, 40, 53 },
		{ 34, 41, 53 },
		{ 34, 42, 53 },
		{ 34, 43, 53 },
		{ 34, 44, 53 },
		{ 34, 45, 53 },
		{ 34, 46, 53 },
		{ 34, 47, 53 },
		{ 34, 48, 53 },
		{ 34, 49, 53 },
		{ 34, 50, 53 },
		{ 34, 51, 53 },
		{ 34, 52, 53 },
		{ 34, 53, 53 },
		{ 34, 54, 53 },
		{ 34, 55, 53 },
		{ 34, 56, 53 },
		{ 34, 57, 53 },
		{ 34, 58, 53 },
		{ 34, 59, 53 },
		{ 34, 60, 53 },
		{ 34, 61, 53 },
		{ 34, 62, 53 },
		{ 34, 63, 53 },
		{ 34, 64, 53 },
		{ 34, 65, 53 },
		{ 34, 66, 53 },
		{ 34, 67, 53 },
		{ 34, 68, 53 },
		{ 34, 69, 53 },
		{ 34, 70, 53 },
		{ 34, 71, 53 },
		{ 34, 72, 53 },
		{ 34, 73, 53 },
		{ 34, 74, 53 },
		{ 34, 75, 53 },
		{ 34, 76, 53 },
		{ 34, 77, 53 },
		{ 34, 78, 53 },
		{ 34, 79, 53 },
		{ 34, 80, 53 },
		{ 34, 81, 53 },
		{ 34, 82, 53 },
		{ 34, 83, 53 },
		{ 34, 84, 53 },
		{ 34, 85, 53 },
		{ 34, 86, 53 },
		{ 34, 87, 53 },
		{ 34, 88, 53 },
		{ 34, 89, 53 },
		{ 35, 36, 53 },
		{ 35, 37, 53 },
		{ 35, 38, 53 },
		{ 35, 39, 53 },
		{ 35, 40, 53 },
		{ 35, 41, 53 },
		{ 35, 42, 53 },
		{ 35, 43, 53 },
		{ 35, 44, 53 },
		{ 35, 45, 53 },
		{ 35, 46, 53 },
		{ 35, 47, 53 },
		{ 35, 48, 53 },
		{ 35, 49, 53 },
		{ 35, 50, 53 },
		{ 35, 51, 53 },
		{ 35, 52, 53 },
		{ 35, 53, 53 },
		{ 35, 54, 53 },
		{ 35, 55, 53 },
		{ 35, 56, 53 },
		{ 35, 57, 53 },
		{ 35, 58, 53 },
		{ 35, 59, 53 },
		{ 35, 60, 53 },
		{ 35, 61, 53 },
		{ 35, 62, 53 },
		{ 35, 63, 53 },
		{ 35, 64, 53 },
		{ 35, 65, 53 },
		{ 35, 66, 53 },
		{ 35, 67, 53 },
		{ 35, 68, 53 },
		{ 35, 69, 53 },
		{ 35, 70, 53 },
		{ 35, 71, 53 },
		{ 35, 72, 53 },
		{ 35, 73, 53 },
		{ 35, 74, 53 },
		{ 35, 75, 53 },
		{ 35, 76, 53 },
		{ 35, 77, 53 },
		{ 35, 78, 53 },
		{ 35, 79, 53 },
		{ 35, 80, 53 },
		{ 35, 81, 53 },
		{ 35, 82, 53 },
		{ 35, 83, 53 },
		{ 35, 84, 53 },
		{ 35, 85, 53 },
		{ 35, 86, 53 },
		{ 35, 87, 53 },
		{ 35, 88, 53 },
		{ 35, 89, 53 },
		{ 35, 90, 53 },
		{ 36, 37, 53 },
		{ 36, 38, 53 },
		{ 36, 39, 53 },
		{ 36, 40, 53 },
		{ 36, 41, 53 },
		{ 36, 42, 53 },
		{ 36, 43, 53 },
		{ 36, 44, 53 },
		{ 36, 45, 53 },
		{ 36, 46, 53 },
		{ 36, 47, 53 },
		{ 36, 48, 53 },
		{ 36, 49, 53 },
		{ 36, 50, 53 },
		{ 36, 51, 53 },
		{ 36, 52, 53 },
		{ 36, 53, 53 },
		{ 36, 54, 53 },
		{ 36, 55, 53 },
		{ 36, 56, 53 },
		{ 36, 57, 53 },
		{ 36, 58, 53 },
		{ 36, 59, 53 },
		{ 36, 60, 53 },
		{ 36, 61, 53 },
		{ 36, 62, 53 },
		{ 36, 63, 53 },
		{ 36, 64, 53 },
		{ 36, 65, 53 },
		{ 36, 66, 53 },
		{ 36, 67, 53 },
		{ 36, 68, 53 },
		{ 36, 69, 53 },
		{ 36, 70, 53 },
		{ 36, 71, 53 },
		{ 36, 72, 53 },
		{ 36, 73, 53 },
		{ 36, 74, 53 },
		{ 36, 75, 53 },
		{ 36, 76, 53 },
		{ 36, 77, 53 },
		{ 36, 78, 53 },
		{ 36, 79, 53 },
		{ 36, 80, 53 },
		{ 36, 81, 53 },
		{ 36, 82, 53 },
		{ 36, 83, 53 },
		{ 36, 84, 53 },
		{ 36, 85, 53 },
		{ 36, 86, 53 },
		{ 36, 87, 53 },
		{ 36, 88, 53 },
		{ 36, 89, 53 },
		{ 36, 90, 53 },
		{ 36, 91, 53 },
		{ 37, 38, 53 },
		{ 37, 39, 53 },
		{ 37, 40, 53 },
		{ 37, 41, 53 },
		{ 37, 42, 53 },
		{ 37, 43, 53 },
		{ 37, 44, 53 },
		{ 37, 45, 53 },
		{ 37, 46, 53 },
		{ 37, 47, 53 },
		{ 37, 48, 53 },
		{ 37, 49, 53 },
		{ 37, 50, 53 },
		{ 37, 51, 53 },
		{ 37, 52, 53 },
		{ 37, 53, 53 },
		{ 37, 54, 53 },
		{ 37, 55, 53 },
		{ 37, 56, 53 },
		{ 37, 57, 53 },
		{ 37, 58, 53 },
		{ 37, 59, 53 },
		{ 37, 60, 53 },
		{ 37, 61, 53 },
		{ 37, 62, 53 },
		{ 37, 63, 53 },
		{ 37, 64, 53 },
		{ 37, 65, 53 },
		{ 37, 66, 53 },
		{ 37, 67, 53 },
		{ 37, 68, 53 },
		{ 37, 69, 53 },
		{ 37, 70, 53 },
		{ 37, 71, 53 },
		{ 37, 72, 53 },
		{ 37, 73, 53 },
		{ 37, 74, 53 },
		{ 37, 75, 53 },
		{ 37, 76, 53 },
		{ 37, 77, 53 },
		{ 37, 78, 53 },
		{ 37, 79, 53 },
		{ 37, 80, 53 },
		{ 37, 81, 53 },
		{ 37, 82, 53 },
		{ 37, 83, 53 },
		{ 37, 84, 53 },
		{ 37, 85, 53 },
		{ 37, 86, 53 },
		{ 37, 87, 53 },
		{ 37, 88, 53 },
		{ 37, 89, 53 },
		{ 37, 90, 53 },
		{ 37, 91, 53 },
		{ 38, 39, 53 },
		{ 38, 40, 53 },
		{ 38, 41, 53 },
		{ 38, 42, 53 },
		{ 38, 43, 53 },
		{ 38, 44, 53 },
		{ 38, 45, 53 },
		{ 38, 46, 53 },
		{ 38, 47, 53 },
		{ 38, 48, 53 },
		{ 38, 49, 53 },
		{ 38, 50, 53 },
		{ 38, 51, 53 },
		{ 38, 52, 53 },
		{ 38, 53, 53 },
		{ 38, 54, 53 },
		{ 38, 55, 53 },
		{ 38, 56, 53 },
		{ 38, 57, 53 },
		{ 38, 58, 53 },
		{ 38, 59, 53 },
		{ 38, 60, 53 },
		{ 38, 61, 53 },
		{ 38, 62, 53 },
		{ 38, 63, 53 },
		{ 38, 64, 53 },
		{ 38, 65, 53 },
		{ 38, 66, 53 },
		{ 38, 67, 53 },
		{ 38, 68, 53 },
		{ 38, 69, 53 },
		{ 38, 70, 53 },
		{ 38, 71, 53 },
		{ 38, 72, 53 },
		{ 38, 73, 53 },
		{ 38, 74, 53 },
		{ 38, 75, 53 },
		{ 38, 76, 53 },
		{ 38, 77, 53 },
		{ 38, 78, 53 },
		{ 38, 79, 53 },
		{ 38, 80, 53 },
		{ 38, 81, 53 },
		{ 38, 82, 53 },
		{ 38, 83, 53 },
		{ 38, 84, 53 },
		{ 38, 85, 53 },
		{ 38, 86, 53 },
		{ 38, 87, 53 },
		{ 38, 88, 53 },
		{ 38, 89, 53 },
		{ 38, 90, 53 },
		{ 38, 91, 53 },
		{ 38, 92, 53 },
		{ 39, 40, 53 },
		{ 39, 41, 53 },
		{ 39, 42, 53 },
		{ 39, 43, 53 },
		{ 39, 44, 53 },
		{ 39, 45, 53 },
		{ 39, 46, 53 },
		{ 39, 47, 53 },
		{ 39, 48, 53 },
		{ 39, 49, 53 },
		{ 39, 50, 53 },
		{ 39, 51, 53 },
		{ 39, 52, 53 },
		{ 39, 53, 53 },
		{ 39, 54, 53 },
		{ 39, 55, 53 },
		{ 39, 56, 53 },
		{ 39, 57, 53 },
		{ 39, 58, 53 },
		{ 39, 59, 53 },
		{ 39, 60, 53 },
		{ 39, 61, 53 },
		{ 39, 62, 53 },
		{ 39, 63, 53 },
		{ 39, 64, 53 },
		{ 39, 65, 53 },
		{ 39, 66, 53 },
		{ 39, 67, 53 },
		{ 39, 68, 53 },
		{ 39, 69, 53 },
		{ 39, 70, 53 },
		{ 39, 71, 53 },
		{ 39, 72, 53 },
		{ 39, 73, 53 },
		{ 39, 74, 53 },
		{ 39, 75, 53 },
		{ 39, 76, 53 },
		{ 39, 77, 53 },
		{ 39, 78, 53 },
		{ 39, 79, 53 },
		{ 39, 80, 53 },
		{ 39, 81, 53 },
		{ 39, 82, 53 },
		{ 39, 83, 53 },
		{ 39, 84, 53 },
		{ 39, 85, 53 },
		{ 39, 86, 53 },
		{ 39, 87, 53 },
		{ 39, 88, 53 },
		{ 39, 89, 53 },
		{ 39, 90, 53 },
		{ 39, 91, 53 },
		{ 39, 92, 53 },
		{ 39, 93, 53 },
		{ 40, 41, 53 },
		{ 40, 42, 53 },
		{ 40, 43, 53 },
		{ 40, 44, 53 },
		{ 40, 45, 53 },
		{ 40, 46, 53 },
		{ 40, 47, 53 },
		{ 40, 48, 53 },
		{ 40, 49, 53 },
		{ 40, 50, 53 },
		{ 40, 51, 53 },
		{ 40, 52, 53 },
		{ 40, 53, 53 },
		{ 40, 54, 53 },
		{ 40, 55, 53 },
		{ 40, 56, 53 },
		{ 40, 57, 53 },
		{ 40, 58, 53 },
		{ 40, 59, 53 },
		{ 40, 60, 53 },
		{ 40, 61, 53 },
		{ 40, 62, 53 },
		{ 40, 63, 53 },
		{ 40, 64, 53 },
		{ 40, 65, 53 },
		{ 40, 66, 53 },
		{ 40, 67, 53 },
		{ 40, 68, 53 },
		{ 40, 69, 53 },
		{ 40, 70, 53 },
		{ 40, 71, 53 },
		{ 40, 72, 53 },
		{ 40, 73, 53 },
		{ 40, 74, 53 },
		{ 40, 75, 53 },
		{ 40, 76, 53 },
		{ 40, 77, 53 },
		{ 40, 78, 53 },
		{ 40, 79, 53 },
		{ 40, 80, 53 },
		{ 40, 81, 53 },
		{ 40, 82, 53 },
		{ 40, 83, 53 },
		{ 40, 84, 53 },
		{ 40, 85, 53 },
		{ 40, 86, 53 },
		{ 40, 87, 53 },
		{ 40, 88, 53 },
		{ 40, 89, 53 },
		{ 40, 90, 53 },
		{ 40, 91, 53 },
		{ 40, 92, 53 },
		{ 40, 93, 53 },
		{ 40, 94, 53 },
		{ 41, 42, 53 },
		{ 41, 43, 53 },
		{ 41, 44, 53 },
		{ 41, 45, 53 },
		{ 41, 46, 53 },
		{ 41, 47, 53 },
		{ 41, 48, 53 },
		{ 41, 49, 53 },
		{ 41, 50, 53 },
		{ 41, 51, 53 },
		{ 41, 52, 53 },
		{ 41, 53, 53 },
		{ 41, 54, 53 },
		{ 41, 55, 53 },
		{ 41, 56, 53 },
		{ 41, 57, 53 },
		{ 41, 58, 53 },
		{ 41, 59, 53 },
		{ 41, 60, 53 },
		{ 41, 61, 53 },
		{ 41, 62, 53 },
		{ 41, 63, 53 },
		{ 41, 64, 53 },
		{ 41, 65, 53 },
		{ 41, 66, 53 },
		{ 41, 67, 53 },
		{ 41, 68, 53 },
		{ 41, 69, 53 },
		{ 41, 70, 53 },
		{ 41, 71, 53 },
		{ 41, 72, 53 },
		{ 41, 73, 53 },
		{ 41, 74, 53 },
		{ 41, 75, 53 },
		{ 41, 76, 53 },
		{ 41, 77, 53 },
		{ 41, 78, 53 },
		{ 41, 79, 53 },
		{ 41, 80, 53 },
		{ 41, 81, 53 },
		{ 41, 82, 53 },
		{ 41, 83, 53 },
		{ 41, 84, 53 },
		{ 41, 85, 53 },
		{ 41, 86, 53 },
		{ 41, 87, 53 },
		{ 41, 88, 53 },
		{ 41, 89, 53 },
		{ 41, 90, 53 },
		{ 41, 91, 53 },
		{ 41, 92, 53 },
		{ 41, 93, 53 },
		{ 41, 94, 53 },
		{ 42, 43, 53 },
		{ 42, 44, 53 },
		{ 42, 45, 53 },
		{ 42, 46, 53 },
		{ 42, 47, 53 },
		{ 42, 48, 53 },
		{ 42, 49, 53 },
		{ 42, 50, 53 },
		{ 42, 51, 53 },
		{ 42, 52, 53 },
		{ 42, 53, 53 },
		{ 42, 54, 53 },
		{ 42, 55, 53 },
		{ 42, 56, 53 },
		{ 42, 57, 53 },
		{ 42, 58, 53 },
		{ 42, 59, 53 },
		{ 42, 60, 53 },
		{ 42, 61, 53 },
		{ 42, 62, 53 },
		{ 42, 63, 53 },
		{ 42, 64, 53 },
		{ 42, 65, 53 },
		{ 42, 66, 53 },
		{ 42, 67, 53 },
		{ 42, 68, 53 },
		{ 42, 69, 53 },
		{ 42, 70, 53 },
		{ 42, 71, 53 },
		{ 42, 72, 53 },
		{ 42, 73, 53 },
		{ 42, 74, 53 },
		{ 42, 75, 53 },
		{ 42, 76, 53 },
		{ 42, 77, 53 },
		{ 42, 78, 53 },
		{ 42, 79, 53 },
		{ 42, 80, 53 },
		{ 42, 81, 53 },
		{ 42, 82, 53 },
		{ 42, 83, 53 },
		{ 42, 84, 53 },
		{ 42, 85, 53 },
		{ 42, 86, 53 },
		{ 42, 87, 53 },
		{ 42, 88, 53 },
		{ 42, 89, 53 },
		{ 42, 90, 53 },
		{ 42, 91, 53 },
		{ 42, 92, 53 },
		{ 42, 93, 53 },
		{ 42, 94, 53 },
		{ 42, 95, 53 },
		{ 43, 44, 53 },
		{ 43, 45, 53 },
		{ 43, 46, 53 },
		{ 43, 47, 53 },
		{ 43, 48, 53 },
		{ 43, 49, 53 },
		{ 43, 50, 53 },
		{ 43, 51, 53 },
		{ 43, 52, 53 },
		{ 43, 53, 53 },
		{ 43, 54, 53 },
		{ 43, 55, 53 },
		{ 43, 56, 53 },
		{ 43, 57, 53 },
		{ 43, 58, 53 },
		{ 43, 59, 53 },
		{ 43, 60, 53 },
		{ 43, 61, 53 },
		{ 43, 62, 53 },
		{ 43, 63, 53 },
		{ 43, 64, 53 },
		{ 43, 65, 53 },
		{ 43, 66, 53 },
		{ 43, 67, 53 },
		{ 43, 68, 53 },
		{ 43, 69, 53 },
		{ 43, 70, 53 },
		{ 43, 71, 53 },
		{ 43, 72, 53 },
		{ 43, 73, 53 },
		{ 43, 74, 53 },
		{ 43, 75, 53 },
		{ 43, 76, 53 },
		{ 43, 77, 53 },
		{ 43, 78, 53 },
		{ 43, 79, 53 },
		{ 43, 80, 53 },
		{ 43, 81, 53 },
		{ 43, 82, 53 },
		{ 43, 83, 53 },
		{ 43, 84, 53 },
		{ 43, 85, 53 },
		{ 43, 86, 53 },
		{ 43, 87, 53 },
		{ 43, 88, 53 },
		{ 43, 89, 53 },
		{ 43, 90, 53 },
		{ 43, 91, 53 },
		{ 43, 92, 53 },
		{ 43, 93, 53 },
		{ 43, 94, 53 },
		{ 43, 95, 53 },
		{ 43, 96, 53 },
		{ 44, 45, 53 },
		{ 44, 46, 53 },
		{ 44, 47, 53 },
		{ 44, 48, 53 },
		{ 44, 49, 53 },
		{ 44, 50, 53 },
		{ 44, 51, 53 },
		{ 44, 52, 53 },
		{ 44, 53, 53 },
		{ 44, 54, 53 },
		{ 44, 55, 53 },
		{ 44, 56, 53 },
		{ 44, 57, 53 },
		{ 44, 58, 53 },
		{ 44, 59, 53 },
		{ 44, 60, 53 },
		{ 44, 61, 53 },
		{ 44, 62, 53 },
		{ 44, 63, 53 },
		{ 44, 64, 53 },
		{ 44, 65, 53 },
		{ 44, 66, 53 },
		{ 44, 67, 53 },
		{ 44, 68, 53 },
		{ 44, 69, 53 },
		{ 44, 70, 53 },
		{ 44, 71, 53 },
		{ 44, 72, 53 },
		{ 44, 73, 53 },
		{ 44, 74, 53 },
		{ 44, 75, 53 },
		{ 44, 76, 53 },
		{ 44, 77, 53 },
		{ 44, 78, 53 },
		{ 44, 79, 53 },
		{ 44, 80, 53 },
		{ 44, 81, 53 },
		{ 44, 82, 53 },
		{ 44, 83, 53 },
		{ 44, 84, 53 },
		{ 44, 85, 53 },
		{ 44, 86, 53 },
		{ 44, 87, 53 },
		{ 44, 88, 53 },
		{ 44, 89, 53 },
		{ 44, 90, 53 },
		{ 44, 91, 53 },
		{ 44, 92, 53 },
		{ 44, 93, 53 },
		{ 44, 94, 53 },
		{ 44, 95, 53 },
		{ 44, 96, 53 },
		{ 44, 97, 53 },
		{ 45, 46, 53 },
		{ 45, 47, 53 },
		{ 45, 48, 53 },
		{ 45, 49, 53 },
		{ 45, 50, 53 },
		{ 45, 51, 53 },
		{ 45, 52, 53 },
		{ 45, 53, 53 },
		{ 45, 54, 53 },
		{ 45, 55, 53 },
		{ 45, 56, 53 },
		{ 45, 57, 53 },
		{ 45, 58, 53 },
		{ 45, 59, 53 },
		{ 45, 60, 53 },
		{ 45, 61, 53 },
		{ 45, 62, 53 },
		{ 45, 63, 53 },
		{ 45, 64, 53 },
		{ 45, 65, 53 },
		{ 45, 66, 53 },
		{ 45, 67, 53 },
		{ 45, 68, 53 },
		{ 45, 69, 53 },
		{ 45, 70, 53 },
		{ 45, 71, 53 },
		{ 45, 72, 53 },
		{ 45, 73, 53 },
		{ 45, 74, 53 },
		{ 45, 75, 53 },
		{ 45, 76, 53 },
		{ 45, 77, 53 },
		{ 45, 78, 53 },
		{ 45, 79, 53 },
		{ 45, 80, 53 },
		{ 45, 81, 53 },
		{ 45, 82, 53 },
		{ 45, 83, 53 },
		{ 45, 84, 53 },
		{ 45, 85, 53 },
		{ 45, 86, 53 },
		{ 45, 87, 53 },
		{ 45, 88, 53 },
		{ 45, 89, 53 },
		{ 45, 90, 53 },
		{ 45, 91, 53 },
		{ 45, 92, 53 },
		{ 45, 93, 53 },
		{ 45, 94, 53 },
		{ 45, 95, 53 },
		{ 45, 96, 53 },
		{ 45, 97, 53 },
		{ 46, 47, 53 },
		{ 46, 48, 53 },
		{ 46, 49, 53 },
		{ 46, 50, 53 },
		{ 46, 51, 53 },
		{ 46, 52, 53 },
		{ 46, 53, 53 },
		{ 46, 54, 53 },
		{ 46, 55, 53 },
		{ 46, 56, 53 },
		{ 46, 57, 53 },
		{ 46, 58, 53 },
		{ 46, 59, 53 },
		{ 46, 60, 53 },
		{ 46, 61, 53 },
		{ 46, 62, 53 },
		{ 46, 63, 53 },
		{ 46, 64, 53 },
		{ 46, 65, 53 },
		{ 46, 66, 53 },
		{ 46, 67, 53 },
		{ 46, 68, 53 },
		{ 46, 69, 53 },
		{ 46, 70, 53 },
		{ 46, 71, 53 },
		{ 46, 72, 53 },
		{ 46, 73, 53 },
		{ 46, 74, 53 },
		{ 46, 75, 53 },
		{ 46, 76, 53 },
		{ 46, 77, 53 },
		{ 46, 78, 53 },
		{ 46, 79, 53 },
		{ 46, 80, 53 },
		{ 46, 81, 53 },
		{ 46, 82, 53 },
		{ 46, 83, 53 },
		{ 46, 84, 53 },
		{ 46, 85, 53 },
		{ 46, 86, 53 },
		{ 46, 87, 53 },
		{ 46, 88, 53 },
		{ 46, 89, 53 },
		{ 46, 90, 53 },
		{ 46, 91, 53 },
		{ 46, 92, 53 },
		{ 46, 93, 53 },
		{ 46, 94, 53 },
		{ 46, 95, 53 },
		{ 46, 96, 53 },
		{ 46, 97, 53 },
		{ 46, 98, 53 },
		{ 47, 48, 53 },
		{ 47, 49, 53 },
		{ 47, 50, 53 },
		{ 47, 51, 53 },
		{ 47, 52, 53 },
		{ 47, 53, 53 },
		{ 47, 54, 53 },
		{ 47, 55, 53 },
		{ 47, 56, 53 },
		{ 47, 57, 53 },
		{ 47, 58, 53 },
		{ 47, 59, 53 },
		{ 47, 60, 53 },
		{ 47, 61, 53 },
		{ 47, 62, 53 },
		{ 47, 63, 53 },
		{ 47, 64, 53 },
		{ 47, 65, 53 },
		{ 47, 66, 53 },
		{ 47, 67, 53 },
		{ 47, 68, 53 },
		{ 47, 69, 53 },
		{ 47, 70, 53 },
		{ 47, 71, 53 },
		{ 47, 72, 53 },
		{ 47, 73, 53 },
		{ 47, 74, 53 },
		{ 47, 75, 53 },
		{ 47, 76, 53 },
		{ 47, 77, 53 },
		{ 47, 78, 53 },
		{ 47, 79, 53 },
		{ 47, 80, 53 },
		{ 47, 81, 53 },
		{ 47, 82, 53 },
		{ 47, 83, 53 },
		{ 47, 84, 53 },
		{ 47, 85, 53 },
		{ 47, 86, 53 },
		{ 47, 87, 53 },
		{ 47, 88, 53 },
		{ 47, 89, 53 },
		{ 47, 90, 53 },
		{ 47, 91, 53 },
		{ 47, 92, 53 },
		{ 47, 93, 53 },
		{ 47, 94, 53 },
		{ 47, 95, 53 },
		{ 47, 96, 53 },
		{ 47, 97, 53 },
		{ 47, 98, 53 },
		{ 47, 99, 53 },
		{ 48, 49, 53 },
		{ 48, 50, 53 },
		{ 48, 51, 53 },
		{ 48, 52, 53 },
		{ 48, 53, 53 },
		{ 48, 54, 53 },
		{ 48, 55, 53 },
		{ 48, 56, 53 },
		{ 48, 57, 53 },
		{ 48, 58, 53 },
		{ 48, 59, 53 },
		{ 48, 60, 53 },
		{ 48, 61, 53 },
		{ 48, 62, 53 },
		{ 48, 63, 53 },
		{ 48, 64, 53 },
		{ 48, 65, 53 },
		{ 48, 66, 53 },
		{ 48, 67, 53 },
		{ 48, 68, 53 },
		{ 48, 69, 53 },
		{ 48, 70, 53 },
		{ 48, 71, 53 },
		{ 48, 72, 53 },
		{ 48, 73, 53 },
		{ 48, 74, 53 },
		{ 48, 75, 53 },
		{ 48, 76, 53 },
		{ 48, 77, 53 },
		{ 48, 78, 53 },
		{ 48, 79, 53 },
		{ 48, 80, 53 },
		{ 48, 81, 53 },
		{ 48, 82, 53 },
		{ 48, 83, 53 },
		{ 48, 84, 53 },
		{ 48, 85, 53 },
		{ 48, 86, 53 },
		{ 48, 87, 53 },
		{ 48, 88, 53 },
		{ 48, 89, 53 },
		{ 48, 90, 53 },
		{ 48, 91, 53 },
		{ 48, 92, 53 },
		{ 48, 93, 53 },
		{ 48, 94, 53 },
		{ 48, 95, 53 },
		{ 48, 96, 53 },
		{ 48, 97, 53 },
		{ 48, 98, 53 },
		{ 48, 99, 53 },
		{ 48, 100, 53 },
		{ 49, 50, 53 },
		{ 49, 51, 53 },
		{ 49, 52, 53 },
		{ 49, 53, 53 },
		{ 49, 54, 53 },
		{ 49, 55, 53 },
		{ 49, 56, 53 },
		{ 49, 57, 53 },
		{ 49, 58, 53 },
		{ 49, 59, 53 },
		{ 49, 60, 53 },
		{ 49, 61, 53 },
		{ 49, 62, 53 },
		{ 49, 63, 53 },
		{ 49, 64, 53 },
		{ 49, 65, 53 },
		{ 49, 66, 53 },
		{ 49, 67, 53 },
		{ 49, 68, 53 },
		{ 49, 69, 53 },
		{ 49, 70, 53 },
		{ 49, 71, 53 },
		{ 49, 72, 53 },
		{ 49, 73, 53 },
		{ 49, 74, 53 },
		{ 49, 75, 53 },
		{ 49, 76, 53 },
		{ 49, 77, 53 },
		{ 49, 78, 53 },
		{ 49, 79, 53 },
		{ 49, 80, 53 },
		{ 49, 81, 53 },
		{ 49, 82, 53 },
		{ 49, 83, 53 },
		{ 49, 84, 53 },
		{ 49, 85, 53 },
		{ 49, 86, 53 },
		{ 49, 87, 53 },
		{ 49, 88, 53 },
		{ 49, 89, 53 },
		{ 49, 90, 53 },
		{ 49, 91, 53 },
		{ 49, 92, 53 },
		{ 49, 93, 53 },
		{ 49, 94, 53 },
		{ 49, 95, 53 },
		{ 49, 96, 53 },
		{ 49, 97, 53 },
		{ 49, 98, 53 },
		{ 49, 99, 53 },
		{ 49, 100, 53 },
		{ 50, 51, 53 },
		{ 50, 52, 53 },
		{ 50, 53, 53 },
		{ 50, 54, 53 },
		{ 50, 55, 53 },
		{ 50, 56, 53 },
		{ 50, 57, 53 },
		{ 50, 58, 53 },
		{ 50, 59, 53 },
		{ 50, 60, 53 },
		{ 50, 61, 53 },
		{ 50, 62, 53 },
		{ 50, 63, 53 },
		{ 50, 64, 53 },
		{ 50, 65, 53 },
		{ 50, 66, 53 },
		{ 50, 67, 53 },
		{ 50, 68, 53 },
		{ 50, 69, 53 },
		{ 50, 70, 53 },
		{ 50, 71, 53 },
		{ 50, 72, 53 },
		{ 50, 73, 53 },
		{ 50, 74, 53 },
		{ 50, 75, 53 },
		{ 50, 76, 53 },
		{ 50, 77, 53 },
		{ 50, 78, 53 },
		{ 50, 79, 53 },
		{ 50, 80, 53 },
		{ 50, 81, 53 },
		{ 50, 82, 53 },
		{ 50, 83, 53 },
		{ 50, 84, 53 },
		{ 50, 85, 53 },
		{ 50, 86, 53 },
		{ 50, 87, 53 },
		{ 50, 88, 53 },
		{ 50, 89, 53 },
		{ 50, 90, 53 },
		{ 50, 91, 53 },
		{ 50, 92, 53 },
		{ 50, 93, 53 },
		{ 50, 94, 53 },
		{ 50, 95, 53 },
		{ 50, 96, 53 },
		{ 50, 97, 53 },
		{ 50, 98, 53 },
		{ 50, 99, 53 },
		{ 50, 100, 53 },
		{ 50, 101, 53 },
		{ 51, 52, 53 },
		{ 51, 53, 53 },
		{ 51, 54, 53 },
		{ 51, 55, 53 },
		{ 51, 56, 53 },
		{ 51, 57, 53 },
		{ 51, 58, 53 },
		{ 51, 59, 53 },
		{ 51, 60, 53 },
		{ 51, 61, 53 },
		{ 51, 62, 53 },
		{ 51, 63, 53 },
		{ 51, 64, 53 },
		{ 51, 65, 53 },
		{ 51, 66, 53 },
		{ 51, 67, 53 },
		{ 51, 68, 53 },
		{ 51, 69, 53 },
		{ 51, 70, 53 },
		{ 51, 71, 53 },
		{ 51, 72, 53 },
		{ 51, 73, 53 },
		{ 51, 74, 53 },
		{ 51, 75, 53 },
		{ 51, 76, 53 },
		{ 51, 77, 53 },
		{ 51, 78, 53 },
		{ 51, 79, 53 },
		{ 51, 80, 53 },
		{ 51, 81, 53 },
		{ 51, 82, 53 },
		{ 51, 83, 53 },
		{ 51, 84, 53 },
		{ 51, 85, 53 },
		{ 51, 86, 53 },
		{ 51, 87, 53 },
		{ 51, 88, 53 },
		{ 51, 89, 53 },
		{ 51, 90, 53 },
		{ 51, 91, 53 },
		{ 51, 92, 53 },
		{ 51, 93, 53 },
		{ 51, 94, 53 },
		{ 51, 95, 53 },
		{ 51, 96, 53 },
		{ 51, 97, 53 },
		{ 51, 98, 53 },
		{ 51, 99, 53 },
		{ 51, 100, 53 },
		{ 51, 101, 53 },
		{ 51, 102, 53 },
		{ 52, 53, 53 },
		{ 52, 54, 53 },
		{ 52, 55, 53 },
		{ 52, 56, 53 },
		{ 52, 57, 53 },
		{ 52, 58, 53 },
		{ 52, 59, 53 },
		{ 52, 60, 53 },
		{ 52, 61, 53 },
		{ 52, 62, 53 },
		{ 52, 63, 53 },
		{ 52, 64, 53 },
		{ 52, 65, 53 },
		{ 52, 66, 53 },
		{ 52, 67, 53 },
		{ 52, 68, 53 },
		{ 52, 69, 53 },
		{ 52, 70, 53 },
		{ 52, 71, 53 },
		{ 52, 72, 53 },
		{ 52, 73, 53 },
		{ 52, 74, 53 },
		{ 52, 75, 53 },
		{ 52, 76, 53 },
		{ 52, 77, 53 },
		{ 52, 78, 53 },
		{ 52, 79, 53 },
		{ 52, 80, 53 },
		{ 52, 81, 53 },
		{ 52, 82, 53 },
		{ 52, 83, 53 },
		{ 52, 84, 53 },
		{ 52, 85, 53 },
		{ 52, 86, 53 },
		{ 52, 87, 53 },
		{ 52, 88, 53 },
		{ 52, 89, 53 },
		{ 52, 90, 53 },
		{ 52, 91, 53 },
		{ 52, 92, 53 },
		{ 52, 93, 53 },
		{ 52, 94, 53 },
		{ 52, 95, 53 },
		{ 52, 96, 53 },
		{ 52, 97, 53 },
		{ 52, 98, 53 },
		{ 52, 99, 53 },
		{ 52, 100, 53 },
		{ 52, 101, 53 },
		{ 52, 102, 53 },
		{ 52, 103, 53 },
		{ 53, 54, 53 },
		{ 53, 55, 53 },
		{ 53, 56, 53 },
		{ 53, 57, 53 },
		{ 53, 58, 53 },
		{ 53, 59, 53 },
		{ 53, 60, 53 },
		{ 53, 61, 53 },
		{ 53, 62, 53 },
		{ 53, 63, 53 },
		{ 53, 64, 53 },
		{ 53, 65, 53 },
		{ 53, 66, 53 },
		{ 53, 67, 53 },
		{ 53, 68, 53 },
		{ 53, 69, 53 },
		{ 53, 70, 53 },
		{ 53, 71, 53 },
		{ 53, 72, 53 },
		{ 53, 73, 53 },
		{ 53, 74, 53 },
		{ 53, 75, 53 },
		{ 53, 76, 53 },
		{ 53, 77, 53 },
		{ 53, 78, 53 },
		{ 53, 79, 53 },
		{ 53, 80, 53 },
		{ 53, 81, 53 },
		{ 53, 82, 53 },
		{ 53, 83, 53 },
		{ 53, 84, 53 },
		{ 53, 85, 53 },
		{ 53, 86, 53 },
		{ 53, 87, 53 },
		{ 53, 88, 53 },
		{ 53, 89, 53 },
		{ 53, 90, 53 },
		{ 53, 91, 53 },
		{ 53, 92, 53 },
		{ 53, 93, 53 },
		{ 53, 94, 53 },
		{ 53, 95, 53 },
		{ 53, 96, 53 },
		{ 53, 97, 53 },
		{ 53, 98, 53 },
		{ 53, 99, 53 },
		{ 53, 100, 53 },
		{ 53, 101, 53 },
		{ 53, 102, 53 },
		{ 53, 103, 53 },
		{ 54, 55, 53 },
		{ 54, 56, 53 },
		{ 54, 57, 53 },
		{ 54, 58, 53 },
		{ 54, 59, 53 },
		{ 54, 60, 53 },
		{ 54, 61, 53 },
		{ 54, 62, 53 },
		{ 54, 63, 53 },
		{ 54, 64, 53 },
		{ 54, 65, 53 },
		{ 54, 66, 53 },
		{ 54, 67, 53 },
		{ 54, 68, 53 },
		{ 54, 69, 53 },
		{ 54, 70, 53 },
		{ 54, 71, 53 },
		{ 54, 72, 53 },
		{ 54, 73, 53 },
		{ 54, 74, 53 },
		{ 54, 75, 53 },
		{ 54, 76, 53 },
		{ 54, 77, 53 },
		{ 54, 78, 53 },
		{ 54, 79, 53 },
		{ 54, 80, 53 },
		{ 54, 81, 53 },
		{ 54, 82, 53 },
		{ 54, 83, 53 },
		{ 54, 84, 53 },
		{ 54, 85, 53 },
		{ 54, 86, 53 },
		{ 54, 87, 53 },
		{ 54, 88, 53 },
		{ 54, 89, 53 },
		{ 54, 90, 53 },
		{ 54, 91, 53 },
		{ 54, 92, 53 },
		{ 54, 93, 53 },
		{ 54, 94, 53 },
		{ 54, 95, 53 },
		{ 54, 96, 53 },
		{ 54, 97, 53 },
		{ 54, 98, 53 },
		{ 54, 99, 53 },
		{ 54, 100, 53 },
		{ 54, 101, 53 },
		{ 54, 102, 53 },
		{ 54, 103, 53 },
		{ 54, 104, 53 },
		{ 55, 56, 53 },
		{ 55, 57, 53 },
		{ 55, 58, 53 },
		{ 55, 59, 53 },
		{ 55, 60, 53 },
		{ 55, 61, 53 },
		{ 55, 62, 53 },
		{ 55, 63, 53 },
		{ 55, 64, 53 },
		{ 55, 65, 53 },
		{ 55, 66, 53 },
		{ 55, 67, 53 },
		{ 55, 68, 53 },
		{ 55, 69, 53 },
		{ 55, 70, 53 },
		{ 55, 71, 53 },
		{ 55, 72, 53 },
		{ 55, 73, 53 },
		{ 55, 74, 53 },
		{ 55, 75, 53 },
		{ 55, 76, 53 },
		{ 55, 77, 53 },
		{ 55, 78, 53 },
		{ 55, 79, 53 },
		{ 55, 80, 53 },
		{ 55, 81, 53 },
		{ 55, 82, 53 },
		{ 55, 83, 53 },
		{ 55, 84, 53 },
		{ 55, 85, 53 },
		{ 55, 86, 53 },
		{ 55, 87, 53 },
		{ 55, 88, 53 },
		{ 55, 89, 53 },
		{ 55, 90, 53 },
		{ 55, 91, 53 },
		{ 55, 92, 53 },
		{ 55, 93, 53 },
		{ 55, 94, 53 },
		{ 55, 95, 53 },
		{ 55, 96, 53 },
		{ 55, 97, 53 },
		{ 55, 98, 53 },
		{ 55, 99, 53 },
		{ 55, 100, 53 },
		{ 55, 101, 53 },
		{ 55, 102, 53 },
		{ 55, 103, 53 },
		{ 55, 104, 53 },
		{ 55, 105, 53 },
		{ 56, 57, 53 },
		{ 56, 58, 53 },
		{ 56, 59, 53 },
		{ 56, 60, 53 },
		{ 56, 61, 53 },
		{ 56, 62, 53 },
		{ 56, 63, 53 },
		{ 56, 64, 53 },
		{ 56, 65, 53 },
		{ 56, 66, 53 },
		{ 56, 67, 53 },
		{ 56, 68, 53 },
		{ 56, 69, 53 },
		{ 56, 70, 53 },
		{ 56, 71, 53 },
		{ 56, 72, 53 },
		{ 56, 73, 53 },
		{ 56, 74, 53 },
		{ 56, 75, 53 },
		{ 56, 76, 53 },
		{ 56, 77, 53 },
		{ 56, 78, 53 },
		{ 56, 79, 53 },
		{ 56, 80, 53 },
		{ 56, 81, 53 },
		{ 56, 82, 53 },
		{ 56, 83, 53 },
		{ 56, 84, 53 },
		{ 56, 85, 53 },
		{ 56, 86, 53 },
		{ 56, 87, 53 },
		{ 56, 88, 53 },
		{ 56, 89, 53 },
		{ 56, 90, 53 },
		{ 56, 91, 53 },
		{ 56, 92, 53 },
		{ 56, 93, 53 },
		{ 56, 94, 53 },
		{ 56, 95, 53 },
		{ 56, 96, 53 },
		{ 56, 97, 53 },
		{ 56, 98, 53 },
		{ 56, 99, 53 },
		{ 56, 100, 53 },
		{ 56, 101, 53 },
		{ 56, 102, 53 },
		{ 56, 103, 53 },
		{ 56, 104, 53 },
		{ 56, 105, 53 },
		{ 56, 106, 53 },
		{ 57, 58, 53 },
		{ 57, 59, 53 },
		{ 57, 60, 53 },
		{ 57, 61, 53 },
		{ 57, 62, 53 },
		{ 57, 63, 53 },
		{ 57, 64, 53 },
		{ 57, 65, 53 },
		{ 57, 66, 53 },
		{ 57, 67, 53 },
		{ 57, 68, 53 },
		{ 57, 69, 53 },
		{ 57, 70, 53 },
		{ 57, 71, 53 },
		{ 57, 72, 53 },
		{ 57, 73, 53 },
		{ 57, 74, 53 },
		{ 57, 75, 53 },
		{ 57, 76, 53 },
		{ 57, 77, 53 },
		{ 57, 78, 53 },
		{ 57, 79, 53 },
		{ 57, 80, 53 },
		{ 57, 81, 53 },
		{ 57, 82, 53 },
		{ 57, 83, 53 },
		{ 57, 84, 53 },
		{ 57, 85, 53 },
		{ 57, 86, 53 },
		{ 57, 87, 53 },
		{ 57, 88, 53 },
		{ 57, 89, 53 },
		{ 57, 90, 53 },
		{ 57, 91, 53 },
		{ 57, 92, 53 },
		{ 57, 93, 53 },
		{ 57, 94, 53 },
		{ 57, 95, 53 },
		{ 57, 96, 53 },
		{ 57, 97, 53 },
		{ 57, 98, 53 },
		{ 57, 99, 53 },
		{ 57, 100, 53 },
		{ 57, 101, 53 },
		{ 57, 102, 53 },
		{ 57, 103, 53 },
		{ 57, 104, 53 },
		{ 57, 105, 53 },
		{ 57, 106, 53 },
		{ 58, 59, 53 },
		{ 58, 60, 53 },
		{ 58, 61, 53 },
		{ 58, 62, 53 },
		{ 58, 63, 53 },
		{ 58, 64, 53 },
		{ 58, 65, 53 },
		{ 58, 66, 53 },
		{ 58, 67, 53 },
		{ 58, 68, 53 },
		{ 58, 69, 53 },
		{ 58, 70, 53 },
		{ 58, 71, 53 },
		{ 58, 72, 53 },
		{ 58, 73, 53 },
		{ 58, 74, 53 },
		{ 58, 75, 53 },
		{ 58, 76, 53 },
		{ 58, 77, 53 },
		{ 58, 78, 53 },
		{ 58, 79, 53 },
		{ 58, 80, 53 },
		{ 58, 81, 53 },
		{ 58, 82, 53 },
		{ 58, 83, 53 },
		{ 58, 84, 53 },
		{ 58, 85, 53 },
		{ 58, 86, 53 },
		{ 58, 87, 53 },
		{ 58, 88, 53 },
		{ 58, 89, 53 },
		{ 58, 90, 53 },
		{ 58, 91, 53 },
		{ 58, 92, 53 },
		{ 58, 93, 53 },
		{ 58, 94, 53 },
		{ 58, 95, 53 },
		{ 58, 96, 53 },
		{ 58, 97, 53 },
		{ 58, 98, 53 },
		{ 58, 99, 53 },
		{ 58, 100, 53 },
		{ 58, 101, 53 },
		{ 58, 102, 53 },
		{ 58, 103, 53 },
		{ 58, 104, 53 },
		{ 58, 105, 53 },
		{ 58, 106, 53 },
		{ 58, 107, 53 },
		{ 59, 60, 53 },
		{ 59, 61, 53 },
		{ 59, 62, 53 },
		{ 59, 63, 53 },
		{ 59, 64, 53 },
		{ 59, 65, 53 },
		{ 59, 66, 53 },
		{ 59, 67, 53 },
		{ 59, 68, 53 },
		{ 59, 69, 53 },
		{ 59, 70, 53 },
		{ 59, 71, 53 },
		{ 59, 72, 53 },
		{ 59, 73, 53 },
		{ 59, 74, 53 },
		{ 59, 75, 53 },
		{ 59, 76, 53 },
		{ 59, 77, 53 },
		{ 59, 78, 53 },
		{ 59, 79, 53 },
		{ 59, 80, 53 },
		{ 59, 81, 53 },
		{ 59, 82, 53 },
		{ 59, 83, 53 },
		{ 59, 84, 53 },
		{ 59, 85, 53 },
		{ 59, 86, 53 },
		{ 59, 87, 53 },
		{ 59, 88, 53 },
		{ 59, 89, 53 },
		{ 59, 90, 53 },
		{ 59, 91, 53 },
		{ 59, 92, 53 },
		{ 59, 93, 53 },
		{ 59, 94, 53 },
		{ 59, 95, 53 },
		{ 59, 96, 53 },
		{ 59, 97, 53 },
		{ 59, 98, 53 },
		{ 59, 99, 53 },
		{ 59, 100, 53 },
		{ 59, 101, 53 },
		{ 59, 102, 53 },
		{ 59, 103, 53 },
		{ 59, 104, 53 },
		{ 59, 105, 53 },
		{ 59, 106, 53 },
		{ 59, 107, 53 },
		{ 59, 108, 53 },
		{ 60, 61, 53 },
		{ 60, 62, 53 },
		{ 60, 63, 53 },
		{ 60, 64, 53 },
		{ 60, 65, 53 },
		{ 60, 66, 53 },
		{ 60, 67, 53 },
		{ 60, 68, 53 },
		{ 60, 69, 53 },
		{ 60, 70, 53 },
		{ 60, 71, 53 },
		{ 60, 72, 53 },
		{ 60, 73, 53 },
		{ 60, 74, 53 },
		{ 60, 75, 53 },
		{ 60, 76, 53 },
		{ 60, 77, 53 },
		{ 60, 78, 53 },
		{ 60, 79, 53 },
		{ 60, 80, 53 },
		{ 60, 81, 53 },
		{ 60, 82, 53 },
		{ 60, 83, 53 },
		{ 60, 84, 53 },
		{ 60, 85, 53 },
		{ 60, 86, 53 },
		{ 60, 87, 53 },
		{ 60, 88, 53 },
		{ 60, 89, 53 },
		{ 60, 90, 53 },
		{ 60, 91, 53 },
		{ 60, 92, 53 },
		{ 60, 93, 53 },
		{ 60, 94, 53 },
		{ 60, 95, 53 },
		{ 60, 96, 53 },
		{ 60, 97, 53 },
		{ 60, 98, 53 },
		{ 60, 99, 53 },
		{ 60, 100, 53 },
		{ 60, 101, 53 },
		{ 60, 102, 53 },
		{ 60, 103, 53 },
		{ 60, 104, 53 },
		{ 60, 105, 53 },
		{ 60, 106, 53 },
		{ 60, 107, 53 },
		{ 60, 108, 53 },
		{ 60, 109, 53 },
		{ 61, 62, 53 },
		{ 61, 63, 53 },
		{ 61, 64, 53 },
		{ 61, 65, 53 },
		{ 61, 66, 53 },
		{ 61, 67, 53 },
		{ 61, 68, 53 },
		{ 61, 69, 53 },
		{ 61, 70, 53 },
		{ 61, 71, 53 },
		{ 61, 72, 53 },
		{ 61, 73, 53 },
		{ 61, 74, 53 },
		{ 61, 75, 53 },
		{ 61, 76, 53 },
		{ 61, 77, 53 },
		{ 61, 78, 53 },
		{ 61, 79, 53 },
		{ 61, 80, 53 },
		{ 61, 81, 53 },
		{ 61, 82, 53 },
		{ 61, 83, 53 },
		{ 61, 84, 53 },
		{ 61, 85, 53 },
		{ 61, 86, 53 },
		{ 61, 87, 53 },
		{ 61, 88, 53 },
		{ 61, 89, 53 },
		{ 61, 90, 53 },
		{ 61, 91, 53 },
		{ 61, 92, 53 },
		{ 61, 93, 53 },
		{ 61, 94, 53 },
		{ 61, 95, 53 },
		{ 61, 96, 53 },
		{ 61, 97, 53 },
		{ 61, 98, 53 },
		{ 61, 99, 53 },
		{ 61, 100, 53 },
		{ 61, 101, 53 },
		{ 61, 102, 53 },
		{ 61, 103, 53 },
		{ 61, 104, 53 },
		{ 61, 105, 53 },
		{ 61, 106, 53 },
		{ 61, 107, 53 },
		{ 61, 108, 53 },
		{ 61, 109, 53 },
		{ 62, 63, 53 },
		{ 62, 64, 53 },
		{ 62, 65, 53 },
		{ 62, 66, 53 },
		{ 62, 67, 53 },
		{ 62, 68, 53 },
		{ 62, 69, 53 },
		{ 62, 70, 53 },
		{ 62, 71, 53 },
		{ 62, 72, 53 },
		{ 62, 73, 53 },
		{ 62, 74, 53 },
		{ 62, 75, 53 },
		{ 62, 76, 53 },
		{ 62, 77, 53 },
		{ 62, 78, 53 },
		{ 62, 79, 53 },
		{ 62, 80, 53 },
		{ 62, 81, 53 },
		{ 62, 82, 53 },
		{ 62, 83, 53 },
		{ 62, 84, 53 },
		{ 62, 85, 53 },
		{ 62, 86, 53 },
		{ 62, 87, 53 },
		{ 62, 88, 53 },
		{ 62, 89, 53 },
		{ 62, 90, 53 },
		{ 62, 91, 53 },
		{ 62, 92, 53 },
		{ 62, 93, 53 },
		{ 62, 94, 53 },
		{ 62, 95, 53 },
		{ 62, 96, 53 },
		{ 62, 97, 53 },
		{ 62, 98, 53 },
		{ 62, 99, 53 },
		{ 62, 100, 53 },
		{ 62, 101, 53 },
		{ 62, 102, 53 },
		{ 62, 103, 53 },
		{ 62, 104, 53 },
		{ 62, 105, 53 },
		{ 62, 106, 53 },
		{ 62, 107, 53 },
		{ 62, 108, 53 },
		{ 62, 109, 53 },
		{ 62, 110, 53 },
		{ 63, 64, 53 },
		{ 63, 65, 53 },
		{ 63, 66, 53 },
		{ 63, 67, 53 },
		{ 63, 68, 53 },
		{ 63, 69, 53 },
		{ 63, 70, 53 },
		{ 63, 71, 53 },
		{ 63, 72, 53 },
		{ 63, 73, 53 },
		{ 63, 74, 53 },
		{ 63, 75, 53 },
		{ 63, 76, 53 },
		{ 63, 77, 53 },
		{ 63, 78, 53 },
		{ 63, 79, 53 },
		{ 63, 80, 53 },
		{ 63, 81, 53 },
		{ 63, 82, 53 },
		{ 63, 83, 53 },
		{ 63, 84, 53 },
		{ 63, 85, 53 },
		{ 63, 86, 53 },
		{ 63, 87, 53 },
		{ 63, 88, 53 },
		{ 63, 89, 53 },
		{ 63, 90, 53 },
		{ 63, 91, 53 },
		{ 63, 92, 53 },
		{ 63, 93, 53 },
		{ 63, 94, 53 },
		{ 63, 95, 53 },
		{ 63, 96, 53 },
		{ 63, 97, 53 },
		{ 63, 98, 53 },
		{ 63, 99, 53 },
		{ 63, 100, 53 },
		{ 63, 101, 53 },
		{ 63, 102, 53 },
		{ 63, 103, 53 },
		{ 63, 104, 53 },
		{ 63, 105, 53 },
		{ 63, 106, 53 },
		{ 63, 107, 53 },
		{ 63, 108, 53 },
		{ 63, 109, 53 },
		{ 63, 110, 53 },
		{ 63, 111, 53 },
		{ 64, 65, 53 },
		{ 64, 66, 53 },
		{ 64, 67, 53 },
		{ 64, 68, 53 },
		{ 64, 69, 53 },
		{ 64, 70, 53 },
		{ 64, 71, 53 },
		{ 64, 72, 53 },
		{ 64, 73, 53 },
		{ 64, 74, 53 },
		{ 64, 75, 53 },
		{ 64, 76, 53 },
		{ 64, 77, 53 },
		{ 64, 78, 53 },
		{ 64, 79, 53 },
		{ 64, 80, 53 },
		{ 64, 81, 53 },
		{ 64, 82, 53 },
		{ 64, 83, 53 },
		{ 64, 84, 53 },
		{ 64, 85, 53 },
		{ 64, 86, 53 },
		{ 64, 87, 53 },
		{ 64, 88, 53 },
		{ 64, 89, 53 },
		{ 64, 90, 53 },
		{ 64, 91, 53 },
		{ 64, 92, 53 },
		{ 64, 93, 53 },
		{ 64, 94, 53 },
		{ 64, 95, 53 },
		{ 64, 96, 53 },
		{ 64, 97, 53 },
		{ 64, 98, 53 },
		{ 64, 99, 53 },
		{ 64, 100, 53 },
		{ 64, 101, 53 },
		{ 64, 102, 53 },
		{ 64, 103, 53 },
		{ 64, 104, 53 },
		{ 64, 105, 53 },
		{ 64, 106, 53 },
		{ 64, 107, 53 },
		{ 64, 108, 53 },
		{ 64, 109, 53 },
		{ 64, 110, 53 },
		{ 64, 111, 53 },
		{ 64, 112, 53 },
		{ 65, 66, 53 },
		{ 65, 67, 53 },
		{ 65, 68, 53 },
		{ 65, 69, 53 },
		{ 65, 70, 53 },
		{ 65, 71, 53 },
		{ 65, 72, 53 },
		{ 65, 73, 53 },
		{ 65, 74, 53 },
		{ 65, 75, 53 },
		{ 65, 76, 53 },
		{ 65, 77, 53 },
		{ 65, 78, 53 },
		{ 65, 79, 53 },
		{ 65, 80, 53 },
		{ 65, 81, 53 },
		{ 65, 82, 53 },
		{ 65, 83, 53 },
		{ 65, 84, 53 },
		{ 65, 85, 53 },
		{ 65, 86, 53 },
		{ 65, 87, 53 },
		{ 65, 88, 53 },
		{ 65, 89, 53 },
		{ 65, 90, 53 },
		{ 65, 91, 53 },
		{ 65, 92, 53 },
		{ 65, 93, 53 },
		{ 65, 94, 53 },
		{ 65, 95, 53 },
		{ 65, 96, 53 },
		{ 65, 97, 53 },
		{ 65, 98, 53 },
		{ 65, 99, 53 },
		{ 65, 100, 53 },
		{ 65, 101, 53 },
		{ 65, 102, 53 },
		{ 65, 103, 53 },
		{ 65, 104, 53 },
		{ 65, 105, 53 },
		{ 65, 106, 53 },
		{ 65, 107, 53 },
		{ 65, 108, 53 },
		{ 65, 109, 53 },
		{ 65, 110, 53 },
		{ 65, 111, 53 },
		{ 65, 112, 53 },
		{ 66, 67, 53 },
		{ 66, 68, 53 },
		{ 66, 69, 53 },
		{ 66, 70, 53 },
		{ 66, 71, 53 },
		{ 66, 72, 53 },
		{ 66, 73, 53 },
		{ 66, 74, 53 },
		{ 66, 75, 53 },
		{ 66, 76, 53 },
		{ 66, 77, 53 },
		{ 66, 78, 53 },
		{ 66, 79, 53 },
		{ 66, 80, 53 },
		{ 66, 81, 53 },
		{ 66, 82, 53 },
		{ 66, 83, 53 },
		{ 66, 84, 53 },
		{ 66, 85, 53 },
		{ 66, 86, 53 },
		{ 66, 87, 53 },
		{ 66, 88, 53 },
		{ 66, 89, 53 },
		{ 66, 90, 53 },
		{ 66, 91, 53 },
		{ 66, 92, 53 },
		{ 66, 93, 53 },
		{ 66, 94, 53 },
		{ 66, 95, 53 },
		{ 66, 96, 53 },
		{ 66, 97, 53 },
		{ 66, 98, 53 },
		{ 66, 99, 53 },
		{ 66, 100, 53 },
		{ 66, 101, 53 },
		{ 66, 102, 53 },
		{ 66, 103, 53 },
		{ 66, 104, 53 },
		{ 66, 105, 53 },
		{ 66, 106, 53 },
		{ 66, 107, 53 },
		{ 66, 108, 53 },
		{ 66, 109, 53 },
		{ 66, 110, 53 },
		{ 66, 111, 53 },
		{ 66, 112, 53 },
		{ 66, 113, 53 },
		{ 67, 68, 53 },
		{ 67, 69, 53 },
		{ 67, 70, 53 },
		{ 67, 71, 53 },
		{ 67, 72, 53 },
		{ 67, 73, 53 },
		{ 67, 74, 53 },
		{ 67, 75, 53 },
		{ 67, 76, 53 },
		{ 67, 77, 53 },
		{ 67, 78, 53 },
		{ 67, 79, 53 },
		{ 67, 80, 53 },
		{ 67, 81, 53 },
		{ 67, 82, 53 },
		{ 67, 83, 53 },
		{ 67, 84, 53 },
		{ 67, 85, 53 },
		{ 67, 86, 53 },
		{ 67, 87, 53 },
		{ 67, 88, 53 },
		{ 67, 89, 53 },
		{ 67, 90, 53 },
		{ 67, 91, 53 },
		{ 67, 92, 53 },
		{ 67, 93, 53 },
		{ 67, 94, 53 },
		{ 67, 95, 53 },
		{ 67, 96, 53 },
		{ 67, 97, 53 },
		{ 67, 98, 53 },
		{ 67, 99, 53 },
		{ 67, 100, 53 },
		{ 67, 101, 53 },
		{ 67, 102, 53 },
		{ 67, 103, 53 },
		{ 67, 104, 53 },
		{ 67, 105, 53 },
		{ 67, 106, 53 },
		{ 67, 107, 53 },
		{ 67, 108, 53 },
		{ 67, 109, 53 },
		{ 67, 110, 53 },
		{ 67, 111, 53 },
		{ 67, 112, 53 },
		{ 67, 113, 53 },
		{ 67, 114, 53 },
		{ 68, 69, 53 },
		{ 68, 70, 53 },
		{ 68, 71, 53 },
		{ 68, 72, 53 },
		{ 68, 73, 53 },
		{ 68, 74, 53 },
		{ 68, 75, 53 },
		{ 68, 76, 53 },
		{ 68, 77, 53 },
		{ 68, 78, 53 },
		{ 68, 79, 53 },
		{ 68, 80, 53 },
		{ 68, 81, 53 },
		{ 68, 82, 53 },
		{ 68, 83, 53 },
		{ 68, 84, 53 },
		{ 68, 85, 53 },
		{ 68, 86, 53 },
		{ 68, 87, 53 },
		{ 68, 88, 53 },
		{ 68, 89, 53 },
		{ 68, 90, 53 },
		{ 68, 91, 53 },
		{ 68, 92, 53 },
		{ 68, 93, 53 },
		{ 68, 94, 53 },
		{ 68, 95, 53 },
		{ 68, 96, 53 },
		{ 68, 97, 53 },
		{ 68, 98, 53 },
		{ 68, 99, 53 },
		{ 68, 100, 53 },
		{ 68, 101, 53 },
		{ 68, 102, 53 },
		{ 68, 103, 53 },
		{ 68, 104, 53 },
		{ 68, 105, 53 },
		{ 68, 106, 53 },
		{ 68, 107, 53 },
		{ 68, 108, 53 },
		{ 68, 109, 53 },
		{ 68, 110, 53 },
		{ 68, 111, 53 },
		{ 68, 112, 53 },
		{ 68, 113, 53 },
		{ 68, 114, 53 },
		{ 68, 115, 53 },
		{ 69, 70, 53 },
		{ 69, 71, 53 },
		{ 69, 72, 53 },
		{ 69, 73, 53 },
		{ 69, 74, 53 },
		{ 69, 75, 53 },
		{ 69, 76, 53 },
		{ 69, 77, 53 },
		{ 69, 78, 53 },
		{ 69, 79, 53 },
		{ 69, 80, 53 },
		{ 69, 81, 53 },
		{ 69, 82, 53 },
		{ 69, 83, 53 },
		{ 69, 84, 53 },
		{ 69, 85, 53 },
		{ 69, 86, 53 },
		{ 69, 87, 53 },
		{ 69, 88, 53 },
		{ 69, 89, 53 },
		{ 69, 90, 53 },
		{ 69, 91, 53 },
		{ 69, 92, 53 },
		{ 69, 93, 53 },
		{ 69, 94, 53 },
		{ 69, 95, 53 },
		{ 69, 96, 53 },
		{ 69, 97, 53 },
		{ 69, 98, 53 },
		{ 69, 99, 53 },
		{ 69, 100, 53 },
		{ 69, 101, 53 },
		{ 69, 102, 53 },
		{ 69, 103, 53 },
		{ 69, 104, 53 },
		{ 69, 105, 53 },
		{ 69, 106, 53 },
		{ 69, 107, 53 },
		{ 69, 108, 53 },
		{ 69, 109, 53 },
		{ 69, 110, 53 },
		{ 69, 111, 53 },
		{ 69, 112, 53 },
		{ 69, 113, 53 },
		{ 69, 114, 53 },
		{ 69, 115, 53 },
		{ 70, 71, 53 },
		{ 70, 72, 53 },
		{ 70, 73, 53 },
		{ 70, 74, 53 },
		{ 70, 75, 53 },
		{ 70, 76, 53 },
		{ 70, 77, 53 },
		{ 70, 78, 53 },
		{ 70, 79, 53 },
		{ 70, 80, 53 },
		{ 70, 81, 53 },
		{ 70, 82, 53 },
		{ 70, 83, 53 },
		{ 70, 84, 53 },
		{ 70, 85, 53 },
		{ 70, 86, 53 },
		{ 70, 87, 53 },
		{ 70, 88, 53 },
		{ 70, 89, 53 },
		{ 70, 90, 53 },
		{ 70, 91, 53 },
		{ 70, 92, 53 },
		{ 70, 93, 53 },
		{ 70, 94, 53 },
		{ 70, 95, 53 },
		{ 70, 96, 53 },
		{ 70, 97, 53 },
		{ 70, 98, 53 },
		{ 70, 99, 53 },
		{ 70, 100, 53 },
		{ 70, 101, 53 },
		{ 70, 102, 53 },
		{ 70, 103, 53 },
		{ 70, 104, 53 },
		{ 70, 105, 53 },
		{ 70, 106, 53 },
		{ 70, 107, 53 },
		{ 70, 108, 53 },
		{ 70, 109, 53 },
		{ 70, 110, 53 },
		{ 70, 111, 53 },
		{ 70, 112, 53 },
		{ 70, 113, 53 },
		{ 70, 114, 53 },
		{ 70, 115, 53 },
		{ 70, 116, 53 },
		{ 71, 72, 53 },
		{ 71, 73, 53 },
		{ 71, 74, 53 },
		{ 71, 75, 53 },
		{ 71, 76, 53 },
		{ 71, 77, 53 },
		{ 71, 78, 53 },
		{ 71, 79, 53 },
		{ 71, 80, 53 },
		{ 71, 81, 53 },
		{ 71, 82, 53 },
		{ 71, 83, 53 },
		{ 71, 84, 53 },
		{ 71, 85, 53 },
		{ 71, 86, 53 },
		{ 71, 87, 53 },
		{ 71, 88, 53 },
		{ 71, 89, 53 },
		{ 71, 90, 53 },
		{ 71, 91, 53 },
		{ 71, 92, 53 },
		{ 71, 93, 53 },
		{ 71, 94, 53 },
		{ 71, 95, 53 },
		{ 71, 96, 53 },
		{ 71, 97, 53 },
		{ 71, 98, 53 },
		{ 71, 99, 53 },
		{ 71, 100, 53 },
		{ 71, 101, 53 },
		{ 71, 102, 53 },
		{ 71, 103, 53 },
		{ 71, 104, 53 },
		{ 71, 105, 53 },
		{ 71, 106, 53 },
		{ 71, 107, 53 },
		{ 71, 108, 53 },
		{ 71, 109, 53 },
		{ 71, 110, 53 },
		{ 71, 111, 53 },
		{ 71, 112, 53 },
		{ 71, 113, 53 },
		{ 71, 114, 53 },
		{ 71, 115, 53 },
		{ 71, 116, 53 },
		{ 71, 117, 53 },
		{ 72, 73, 53 },
		{ 72, 74, 53 },
		{ 72, 75, 53 },
		{ 72, 76, 53 },
		{ 72, 77, 53 },
		{ 72, 78, 53 },
		{ 72, 79, 53 },
		{ 72, 80, 53 },
		{ 72, 81, 53 },
		{ 72, 82, 53 },
		{ 72, 83, 53 },
		{ 72, 84, 53 },
		{ 72, 85, 53 },
		{ 72, 86, 53 },
		{ 72, 87, 53 },
		{ 72, 88, 53 },
		{ 72, 89, 53 },
		{ 72, 90, 53 },
		{ 72, 91, 53 },
		{ 72, 92, 53 },
		{ 72, 93, 53 },
		{ 72, 94, 53 },
		{ 72, 95, 53 },
		{ 72, 96, 53 },
		{ 72, 97, 53 },
		{ 72, 98, 53 },
		{ 72, 99, 53 },
		{ 72, 100, 53 },
		{ 72, 101, 53 },
		{ 72, 102, 53 },
		{ 72, 103, 53 },
		{ 72, 104, 53 },
		{ 72, 105, 53 },
		{ 72, 106, 53 },
		{ 72, 107, 53 },
		{ 72, 108, 53 },
		{ 72, 109, 53 },
		{ 72, 110, 53 },
		{ 72, 111, 53 },
		{ 72, 112, 53 },
		{ 72, 113, 53 },
		{ 72, 114, 53 },
		{ 72, 115, 53 },
		{ 72, 116, 53 },
		{ 72, 117, 53 },
		{ 72, 118, 53 },
		{ 73, 74, 53 },
		{ 73, 75, 53 },
		{ 73, 76, 53 },
		{ 73, 77, 53 },
		{ 73, 78, 53 },
		{ 73, 79, 53 },
		{ 73, 80, 53 },
		{ 73, 81, 53 },
		{ 73, 82, 53 },
		{ 73, 83, 53 },
		{ 73, 84, 53 },
		{ 73, 85, 53 },
		{ 73, 86, 53 },
		{ 73, 87, 53 },
		{ 73, 88, 53 },
		{ 73, 89, 53 },
		{ 73, 90, 53 },
		{ 73, 91, 53 },
		{ 73, 92, 53 },
		{ 73, 93, 53 },
		{ 73, 94, 53 },
		{ 73, 95, 53 },
		{ 73, 96, 53 },
		{ 73, 97, 53 },
		{ 73, 98, 53 },
		{ 73, 99, 53 },
		{ 73, 100, 53 },
		{ 73, 101, 53 },
		{ 73, 102, 53 },
		{ 73, 103, 53 },
		{ 73, 104, 53 },
		{ 73, 105, 53 },
		{ 73, 106, 53 },
		{ 73, 107, 53 },
		{ 73, 108, 53 },
		{ 73, 109, 53 },
		{ 73, 110, 53 },
		{ 73, 111, 53 },
		{ 73, 112, 53 },
		{ 73, 113, 53 },
		{ 73, 114, 53 },
		{ 73, 115, 53 },
		{ 73, 116, 53 },
		{ 73, 117, 53 },
		{ 73, 118, 53 },
		{ 74, 75, 53 },
		{ 74, 76, 53 },
		{ 74, 77, 53 },
		{ 74, 78, 53 },
		{ 74, 79, 53 },
		{ 74, 80, 53 },
		{ 74, 81, 53 },
		{ 74, 82, 53 },
		{ 74, 83, 53 },
		{ 74, 84, 53 },
		{ 74, 85, 53 },
		{ 74, 86, 53 },
		{ 74, 87, 53 },
		{ 74, 88, 53 },
		{ 74, 89, 53 },
		{ 74, 90, 53 },
		{ 74, 91, 53 },
		{ 74, 92, 53 },
		{ 74, 93, 53 },
		{ 74, 94, 53 },
		{ 74, 95, 53 },
		{ 74, 96, 53 },
		{ 74, 97, 53 },
		{ 74, 98, 53 },
		{ 74, 99, 53 },
		{ 74, 100, 53 },
		{ 74, 101, 53 },
		{ 74, 102, 53 },
		{ 74, 103, 53 },
		{ 74, 104, 53 },
		{ 74, 105, 53 },
		{ 74, 106, 53 },
		{ 74, 107, 53 },
		{ 74, 108, 53 },
		{ 74, 109, 53 },
		{ 74, 110, 53 },
		{ 74, 111, 53 },
		{ 74, 112, 53 },
		{ 74, 113, 53 },
		{ 74, 114, 53 },
		{ 74, 115, 53 },
		{ 74, 116, 53 },
		{ 74, 117, 53 },
		{ 74, 118, 53 },
		{ 74, 119, 53 },
		{ 75, 76, 53 },
		{ 75, 77, 53 },
		{ 75, 78, 53 },
		{ 75, 79, 53 },
		{ 75, 80, 53 },
		{ 75, 81, 53 },
		{ 75, 82, 53 },
		{ 75, 83, 53 },
		{ 75, 84, 53 },
		{ 75, 85, 53 },
		{ 75, 86, 53 },
		{ 75, 87, 53 },
		{ 75, 88, 53 },
		{ 75, 89, 53 },
		{ 75, 90, 53 },
		{ 75, 91, 53 },
		{ 75, 92, 53 },
		{ 75, 93, 53 },
		{ 75, 94, 53 },
		{ 75, 95, 53 },
		{ 75, 96, 53 },
		{ 75, 97, 53 },
		{ 75, 98, 53 },
		{ 75, 99, 53 },
		{ 75, 100, 53 },
		{ 75, 101, 53 },
		{ 75, 102, 53 },
		{ 75, 103, 53 },
		{ 75, 104, 53 },
		{ 75, 105, 53 },
		{ 75, 106, 53 },
		{ 75, 107, 53 },
		{ 75, 108, 53 },
		{ 75, 109, 53 },
		{ 75, 110, 53 },
		{ 75, 111, 53 },
		{ 75, 112, 53 },
		{ 75, 113, 53 },
		{ 75, 114, 53 },
		{ 75, 115, 53 },
		{ 75, 116, 53 },
		{ 75, 117, 53 },
		{ 75, 118, 53 },
		{ 75, 119, 53 },
		{ 75, 120, 53 },
		{ 76, 77, 53 },
		{ 76, 78, 53 },
		{ 76, 79, 53 },
		{ 76, 80, 53 },
		{ 76, 81, 53 },
		{ 76, 82, 53 },
		{ 76, 83, 53 },
		{ 76, 84, 53 },
		{ 76, 85, 53 },
		{ 76, 86, 53 },
		{ 76, 87, 53 },
		{ 76, 88, 53 },
		{ 76, 89, 53 },
		{ 76, 90, 53 },
		{ 76, 91, 53 },
		{ 76, 92, 53 },
		{ 76, 93, 53 },
		{ 76, 94, 53 },
		{ 76, 95, 53 },
		{ 76, 96, 53 },
		{ 76, 97, 53 },
		{ 76, 98, 53 },
		{ 76, 99, 53 },
		{ 76, 100, 53 },
		{ 76, 101, 53 },
		{ 76, 102, 53 },
		{ 76, 103, 53 },
		{ 76, 104, 53 },
		{ 76, 105, 53 },
		{ 76, 106, 53 },
		{ 76, 107, 53 },
		{ 76, 108, 53 },
		{ 76, 109, 53 },
		{ 76, 110, 53 },
		{ 76, 111, 53 },
		{ 76, 112, 53 },
		{ 76, 113, 53 },
		{ 76, 114, 53 },
		{ 76, 115, 53 },
		{ 76, 116, 53 },
		{ 76, 117, 53 },
		{ 76, 118, 53 },
		{ 76, 119, 53 },
		{ 76, 120, 53 },
		{ 76, 121, 53 },
		{ 77, 78, 53 },
		{ 77, 79, 53 },
		{ 77, 80, 53 },
		{ 77, 81, 53 },
		{ 77, 82, 53 },
		{ 77, 83, 53 },
		{ 77, 84, 53 },
		{ 77, 85, 53 },
		{ 77, 86, 53 },
		{ 77, 87, 53 },
		{ 77, 88, 53 },
		{ 77, 89, 53 },
		{ 77, 90, 53 },
		{ 77, 91, 53 },
		{ 77, 92, 53 },
		{ 77, 93, 53 },
		{ 77, 94, 53 },
		{ 77, 95, 53 },
		{ 77, 96, 53 },
		{ 77, 97, 53 },
		{ 77, 98, 53 },
		{ 77, 99, 53 },
		{ 77, 100, 53 },
		{ 77, 101, 53 },
		{ 77, 102, 53 },
		{ 77, 103, 53 },
		{ 77, 104, 53 },
		{ 77, 105, 53 },
		{ 77, 106, 53 },
		{ 77, 107, 53 },
		{ 77, 108, 53 },
		{ 77, 109, 53 },
		{ 77, 110, 53 },
		{ 77, 111, 53 },
		{ 77, 112, 53 },
		{ 77, 113, 53 },
		{ 77, 114, 53 },
		{ 77, 115, 53 },
		{ 77, 116, 53 },
		{ 77, 117, 53 },
		{ 77, 118, 53 },
		{ 77, 119, 53 },
		{ 77, 120, 53 },
		{ 77, 121, 53 },
		{ 78, 79, 53 },
		{ 78, 80, 53 },
		{ 78, 81, 53 },
		{ 78, 82, 53 },
		{ 78, 83, 53 },
		{ 78, 84, 53 },
		{ 78, 85, 53 },
		{ 78, 86, 53 },
		{ 78, 87, 53 },
		{ 78, 88, 53 },
		{ 78, 89, 53 },
		{ 78, 90, 53 },
		{ 78, 91, 53 },
		{ 78, 92, 53 },
		{ 78, 93, 53 },
		{ 78, 94, 53 },
		{ 78, 95, 53 },
		{ 78, 96, 53 },
		{ 78, 97, 53 },
		{ 78, 98, 53 },
		{ 78, 99, 53 },
		{ 78, 100, 53 },
		{ 78, 101, 53 },
		{ 78, 102, 53 },
		{ 78, 103, 53 },
		{ 78, 104, 53 },
		{ 78, 105, 53 },
		{ 78, 106, 53 },
		{ 78, 107, 53 },
		{ 78, 108, 53 },
		{ 78, 109, 53 },
		{ 78, 110, 53 },
		{ 78, 111, 53 },
		{ 78, 112, 53 },
		{ 78, 113, 53 },
		{ 78, 114, 53 },
		{ 78, 115, 53 },
		{ 78, 116, 53 },
		{ 78, 117, 53 },
		{ 78, 118, 53 },
		{ 78, 119, 53 },
		{ 78, 120, 53 },
		{ 78, 121, 53 },
		{ 78, 122, 53 },
		{ 79, 80, 53 },
		{ 79, 81, 53 },
		{ 79, 82, 53 },
		{ 79, 83, 53 },
		{ 79, 84, 53 },
		{ 79, 85, 53 },
		{ 79, 86, 53 },
		{ 79, 87, 53 },
		{ 79, 88, 53 },
		{ 79, 89, 53 },
		{ 79, 90, 53 },
		{ 79, 91, 53 },
		{ 79, 92, 53 },
		{ 79, 93, 53 },
		{ 79, 94, 53 },
		{ 79, 95, 53 },
		{ 79, 96, 53 },
		{ 79, 97, 53 },
		{ 79, 98, 53 },
		{ 79, 99, 53 },
		{ 79, 100, 53 },
		{ 79, 101, 53 },
		{ 79, 102, 53 },
		{ 79, 103, 53 },
		{ 79, 104, 53 },
		{ 79, 105, 53 },
		{ 79, 106, 53 },
		{ 79, 107, 53 },
		{ 79, 108, 53 },
		{ 79, 109, 53 },
		{ 79, 110, 53 },
		{ 79, 111, 53 },
		{ 79, 112, 53 },
		{ 79, 113, 53 },
		{ 79, 114, 53 },
		{ 79, 115, 53 },
		{ 79, 116, 53 },
		{ 79, 117, 53 },
		{ 79, 118, 53 },
		{ 79, 119, 53 },
		{ 79, 120, 53 },
		{ 79, 121, 53 },
		{ 79, 122, 53 },
		{ 79, 123, 53 },
		{ 80, 81, 53 },
		{ 80, 82, 53 },
		{ 80, 83, 53 },
		{ 80, 84, 53 },
		{ 80, 85, 53 },
		{ 80, 86, 53 },
		{ 80, 87, 53 },
		{ 80, 88, 53 },
		{ 80, 89, 53 },
		{ 80, 90, 53 },
		{ 80, 91, 53 },
		{ 80, 92, 53 },
		{ 80, 93, 53 },
		{ 80, 94, 53 },
		{ 80, 95, 53 },
		{ 80, 96, 53 },
		{ 80, 97, 53 },
		{ 80, 98, 53 },
		{ 80, 99, 53 },
		{ 80, 100, 53 },
		{ 80, 101, 53 },
		{ 80, 102, 53 },
		{ 80, 103, 53 },
		{ 80, 104, 53 },
		{ 80, 105, 53 },
		{ 80, 106, 53 },
		{ 80, 107, 53 },
		{ 80, 108, 53 },
		{ 80, 109, 53 },
		{ 80, 110, 53 },
		{ 80, 111, 53 },
		{ 80, 112, 53 },
		{ 80, 113, 53 },
		{ 80, 114, 53 },
		{ 80, 115, 53 },
		{ 80, 116, 53 },
		{ 80, 117, 53 },
		{ 80, 118, 53 },
		{ 80, 119, 53 },
		{ 80, 120, 53 },
		{ 80, 121, 53 },
		{ 80, 122, 53 },
		{ 80, 123, 53 },
		{ 80, 124, 53 },
		{ 81, 82, 53 },
		{ 81, 83, 53 },
		{ 81, 84, 53 },
		{ 81, 85, 53 },
		{ 81, 86, 53 },
		{ 81, 87, 53 },
		{ 81, 88, 53 },
		{ 81, 89, 53 },
		{ 81, 90, 53 },
		{ 81, 91, 53 },
		{ 81, 92, 53 },
		{ 81, 93, 53 },
		{ 81, 94, 53 },
		{ 81, 95, 53 },
		{ 81, 96, 53 },
		{ 81, 97, 53 },
		{ 81, 98, 53 },
		{ 81, 99, 53 },
		{ 81, 100, 53 },
		{ 81, 101, 53 },
		{ 81, 102, 53 },
		{ 81, 103, 53 },
		{ 81, 104, 53 },
		{ 81, 105, 53 },
		{ 81, 106, 53 },
		{ 81, 107, 53 },
		{ 81, 108, 53 },
		{ 81, 109, 53 },
		{ 81, 110, 53 },
		{ 81, 111, 53 },
		{ 81, 112, 53 },
		{ 81, 113, 53 },
		{ 81, 114, 53 },
		{ 81, 115, 53 },
		{ 81, 116, 53 },
		{ 81, 117, 53 },
		{ 81, 118, 53 },
		{ 81, 119, 53 },
		{ 81, 120, 53 },
		{ 81, 121, 53 },
		{ 81, 122, 53 },
		{ 81, 123, 53 },
		{ 81, 124, 53 },
		{ 82, 83, 53 },
		{ 82, 84, 53 },
		{ 82, 85, 53 },
		{ 82, 86, 53 },
		{ 82, 87, 53 },
		{ 82, 88, 53 },
		{ 82, 89, 53 },
		{ 82, 90, 53 },
		{ 82, 91, 53 },
		{ 82, 92, 53 },
		{ 82, 93, 53 },
		{ 82, 94, 53 },
		{ 82, 95, 53 },
		{ 82, 96, 53 },
		{ 82, 97, 53 },
		{ 82, 98, 53 },
		{ 82, 99, 53 },
		{ 82, 100, 53 },
		{ 82, 101, 53 },
		{ 82, 102, 53 },
		{ 82, 103, 53 },
		{ 82, 104, 53 },
		{ 82, 105, 53 },
		{ 82, 106, 53 },
		{ 82, 107, 53 },
		{ 82, 108, 53 },
		{ 82, 109, 53 },
		{ 82, 110, 53 },
		{ 82, 111, 53 },
		{ 82, 112, 53 },
		{ 82, 113, 53 },
		{ 82, 114, 53 },
		{ 82, 115, 53 },
		{ 82, 116, 53 },
		{ 82, 117, 53 },
		{ 82, 118, 53 },
		{ 82, 119, 53 },
		{ 82, 120, 53 },
		{ 82, 121, 53 },
		{ 82, 122, 53 },
		{ 82, 123, 53 },
		{ 82, 124, 53 },
		{ 82, 125, 53 },
		{ 83, 84, 53 },
		{ 83, 85, 53 },
		{ 83, 86, 53 },
		{ 83, 87, 53 },
		{ 83, 88, 53 },
		{ 83, 89, 53 },
		{ 83, 90, 53 },
		{ 83, 91, 53 },
		{ 83, 92, 53 },
		{ 83, 93, 53 },
		{ 83, 94, 53 },
		{ 83, 95, 53 },
		{ 83, 96, 53 },
		{ 83, 97, 53 },
		{ 83, 98, 53 },
		{ 83, 99, 53 },
		{ 83, 100, 53 },
		{ 83, 101, 53 },
		{ 83, 102, 53 },
		{ 83, 103, 53 },
		{ 83, 104, 53 },
		{ 83, 105, 53 },
		{ 83, 106, 53 },
		{ 83, 107, 53 },
		{ 83, 108, 53 },
		{ 83, 109, 53 },
		{ 83, 110, 53 },
		{ 83, 111, 53 },
		{ 83, 112, 53 },
		{ 83, 113, 53 },
		{ 83, 114, 53 },
		{ 83, 115, 53 },
		{ 83, 116, 53 },
		{ 83, 117, 53 },
		{ 83, 118, 53 },
		{ 83, 119, 53 },
		{ 83, 120, 53 },
		{ 83, 121, 53 },
		{ 83, 122, 53 },
		{ 83, 123, 53 },
		{ 83, 124, 53 },
		{ 83, 125, 53 },
		{ 83, 126, 53 },
		{ 84, 85, 53 },
		{ 84, 86, 53 },
		{ 84, 87, 53 },
		{ 84, 88, 53 },
		{ 84, 89, 53 },
		{ 84, 90, 53 },
		{ 84, 91, 53 },
		{ 84, 92, 53 },
		{ 84, 93, 53 },
		{ 84, 94, 53 },
		{ 84, 95, 53 },
		{ 84, 96, 53 },
		{ 84, 97, 53 },
		{ 84, 98, 53 },
		{ 84, 99, 53 },
		{ 84, 100, 53 },
		{ 84, 101, 53 },
		{ 84, 102, 53 },
		{ 84, 103, 53 },
		{ 84, 104, 53 },
		{ 84, 105, 53 },
		{ 84, 106, 53 },
		{ 84, 107, 53 },
		{ 84, 108, 53 },
		{ 84, 109, 53 },
		{ 84, 110, 53 },
		{ 84, 111, 53 },
		{ 84, 112, 53 },
		{ 84, 113, 53 },
		{ 84, 114, 53 },
		{ 84, 115, 53 },
		{ 84, 116, 53 },
		{ 84, 117, 53 },
		{ 84, 118, 53 },
		{ 84, 119, 53 },
		{ 84, 120, 53 },
		{ 84, 121, 53 },
		{ 84, 122, 53 },
		{ 84, 123, 53 },
		{ 84, 124, 53 },
		{ 84, 125, 53 },
		{ 84, 126, 53 },
		{ 84, 127, 53 },
		{ 85, 86, 53 },
		{ 85, 87, 53 },
		{ 85, 88, 53 },
		{ 85, 89, 53 },
		{ 85, 90, 53 },
		{ 85, 91, 53 },
		{ 85, 92, 53 },
		{ 85, 93, 53 },
		{ 85, 94, 53 },
		{ 85, 95, 53 },
		{ 85, 96, 53 },
		{ 85, 97, 53 },
		{ 85, 98, 53 },
		{ 85, 99, 53 },
		{ 85, 100, 53 },
		{ 85, 101, 53 },
		{ 85, 102, 53 },
		{ 85, 103, 53 },
		{ 85, 104, 53 },
		{ 85, 105, 53 },
		{ 85, 106, 53 },
		{ 85, 107, 53 },
		{ 85, 108, 53 },
		{ 85, 109, 53 },
		{ 85, 110, 53 },
		{ 85, 111, 53 },
		{ 85, 112, 53 },
		{ 85, 113, 53 },
		{ 85, 114, 53 },
		{ 85, 115, 53 },
		{ 85, 116, 53 },
		{ 85, 117, 53 },
		{ 85, 118, 53 },
		{ 85, 119, 53 },
		{ 85, 120, 53 },
		{ 85, 121, 53 },
		{ 85, 122, 53 },
		{ 85, 123, 53 },
		{ 85, 124, 53 },
		{ 85, 125, 53 },
		{ 85, 126, 53 },
		{ 85, 127, 53 },
		{ 86, 87, 53 },
		{ 86, 88, 53 },
		{ 86, 89, 53 },
		{ 86, 90, 53 },
		{ 86, 91, 53 },
		{ 86, 92, 53 },
		{ 86, 93, 53 },
		{ 86, 94, 53 },
		{ 86, 95, 53 },
		{ 86, 96, 53 },
		{ 86, 97, 53 },
		{ 86, 98, 53 },
		{ 86, 99, 53 },
		{ 86, 100, 53 },
		{ 86, 101, 53 },
		{ 86, 102, 53 },
		{ 86, 103, 53 },
		{ 86, 104, 53 },
		{ 86, 105, 53 },
		{ 86, 106, 53 },
		{ 86, 107, 53 },
		{ 86, 108, 53 },
		{ 86, 109, 53 },
		{ 86, 110, 53 },
		{ 86, 111, 53 },
		{ 86, 112, 53 },
		{ 86, 113, 53 },
		{ 86, 114, 53 },
		{ 86, 115, 53 },
		{ 86, 116, 53 },
		{ 86, 117, 53 },
		{ 86, 118, 53 },
		{ 86, 119, 53 },
		{ 86, 120, 53 },
		{ 86, 121, 53 },
		{ 86, 122, 53 },
		{ 86, 123, 53 },
		{ 86, 124, 53 },
		{ 86, 125, 53 },
		{ 86, 126, 53 },
		{ 86, 127, 53 },
		{ 86, 128, 53 },
		{ 87, 88, 53 },
		{ 87, 89, 53 },
		{ 87, 90, 53 },
		{ 87, 91, 53 },
		{ 87, 92, 53 },
		{ 87, 93, 53 },
		{ 87, 94, 53 },
		{ 87, 95, 53 },
		{ 87, 96, 53 },
		{ 87, 97, 53 },
		{ 87, 98, 53 },
		{ 87, 99, 53 },
		{ 87, 100, 53 },
		{ 87, 101, 53 },
		{ 87, 102, 53 },
		{ 87, 103, 53 },
		{ 87, 104, 53 },
		{ 87, 105, 53 },
		{ 87, 106, 53 },
		{ 87, 107, 53 },
		{ 87, 108, 53 },
		{ 87, 109, 53 },
		{ 87, 110, 53 },
		{ 87, 111, 53 },
		{ 87, 112, 53 },
		{ 87, 113, 53 },
		{ 87, 114, 53 },
		{ 87, 115, 53 },
		{ 87, 116, 53 },
		{ 87, 117, 53 },
		{ 87, 118, 53 },
		{ 87, 119, 53 },
		{ 87, 120, 53 },
		{ 87, 121, 53 },
		{ 87, 122, 53 },
		{ 87, 123, 53 },
		{ 87, 124, 53 },
		{ 87, 125, 53 },
		{ 87, 126, 53 },
		{ 87, 127, 53 },
		{ 87, 128, 53 },
		{ 87, 129, 53 },
		{ 88, 89, 53 },
		{ 88, 90, 53 },
		{ 88, 91, 53 },
		{ 88, 92, 53 },
		{ 88, 93, 53 },
		{ 88, 94, 53 },
		{ 88, 95, 53 },
		{ 88, 96, 53 },
		{ 88, 97, 53 },
		{ 88, 98, 53 },
		{ 88, 99, 53 },
		{ 88, 100, 53 },
		{ 88, 101, 53 },
		{ 88, 102, 53 },
		{ 88, 103, 53 },
		{ 88, 104, 53 },
		{ 88, 105, 53 },
		{ 88, 106, 53 },
		{ 88, 107, 53 },
		{ 88, 108, 53 },
		{ 88, 109, 53 },
		{ 88, 110, 53 },
		{ 88, 111, 53 },
		{ 88, 112, 53 },
		{ 88, 113, 53 },
		{ 88, 114, 53 },
		{ 88, 115, 53 },
		{ 88, 116, 53 },
		{ 88, 117, 53 },
		{ 88, 118, 53 },
		{ 88, 119, 53 },
		{ 88, 120, 53 },
		{ 88, 121, 53 },
		{ 88, 122, 53 },
		{ 88, 123, 53 },
		{ 88, 124, 53 },
		{ 88, 125, 53 },
		{ 88, 126, 53 },
		{ 88, 127, 53 },
		{ 88, 128, 53 },
		{ 88, 129, 53 },
		{ 88, 130, 53 },
		{ 89, 90, 53 },
		{ 89, 91, 53 },
		{ 89, 92, 53 },
		{ 89, 93, 53 },
		{ 89, 94, 53 },
		{ 89, 95, 53 },
		{ 89, 96, 53 },
		{ 89, 97, 53 },
		{ 89, 98, 53 },
		{ 89, 99, 53 },
		{ 89, 100, 53 },
		{ 89, 101, 53 },
		{ 89, 102, 53 },
		{ 89, 103, 53 },
		{ 89, 104, 53 },
		{ 89, 105, 53 },
		{ 89, 106, 53 },
		{ 89, 107, 53 },
		{ 89, 108, 53 },
		{ 89, 109, 53 },
		{ 89, 110, 53 },
		{ 89, 111, 53 },
		{ 89, 112, 53 },
		{ 89, 113, 53 },
		{ 89, 114, 53 },
		{ 89, 115, 53 },
		{ 89, 116, 53 },
		{ 89, 117, 53 },
		{ 89, 118, 53 },
		{ 89, 119, 53 },
		{ 89, 120, 53 },
		{ 89, 121, 53 },
		{ 89, 122, 53 },
		{ 89, 123, 53 },
		{ 89, 124, 53 },
		{ 89, 125, 53 },
		{ 89, 126, 53 },
		{ 89, 127, 53 },
		{ 89, 128, 53 },
		{ 89, 129, 53 },
		{ 89, 130, 53 },
		{ 90, 91, 53 },
		{ 90, 92, 53 },
		{ 90, 93, 53 },
		{ 90, 94, 53 },
		{ 90, 95, 53 },
		{ 90, 96, 53 },
		{ 90, 97, 53 },
		{ 90, 98, 53 },
		{ 90, 99, 53 },
		{ 90, 100, 53 },
		{ 90, 101, 53 },
		{ 90, 102, 53 },
		{ 90, 103, 53 },
		{ 90, 104, 53 },
		{ 90, 105, 53 },
		{ 90, 106, 53 },
		{ 90, 107, 53 },
		{ 90, 108, 53 },
		{ 90, 109, 53 },
		{ 90, 110, 53 },
		{ 90, 111, 53 },
		{ 90, 112, 53 },
		{ 90, 113, 53 },
		{ 90, 114, 53 },
		{ 90, 115, 53 },
		{ 90, 116, 53 },
		{ 90, 117, 53 },
		{ 90, 118, 53 },
		{ 90, 119, 53 },
		{ 90, 120, 53 },
		{ 90, 121, 53 },
		{ 90, 122, 53 },
		{ 90, 123, 53 },
		{ 90, 124, 53 },
		{ 90, 125, 53 },
		{ 90, 126, 53 },
		{ 90, 127, 53 },
		{ 90, 128, 53 },
		{ 90, 129, 53 },
		{ 90, 130, 53 },
		{ 90, 131, 53 },
		{ 91, 92, 53 },
		{ 91, 93, 53 },
		{ 91, 94, 53 },
		{ 91, 95, 53 },
		{ 91, 96, 53 },
		{ 91, 97, 53 },
		{ 91, 98, 53 },
		{ 91, 99, 53 },
		{ 91, 100, 53 },
		{ 91, 101, 53 },
		{ 91, 102, 53 },
		{ 91, 103, 53 },
		{ 91, 104, 53 },
		{ 91, 105, 53 },
		{ 91, 106, 53 },
		{ 91, 107, 53 },
		{ 91, 108, 53 },
		{ 91, 109, 53 },
		{ 91, 110, 53 },
		{ 91, 111, 53 },
		{ 91, 112, 53 },
		{ 91, 113, 53 },
		{ 91, 114, 53 },
		{ 91, 115, 53 },
		{ 91, 116, 53 },
		{ 91, 117, 53 },
		{ 91, 118, 53 },
		{ 91, 119, 53 },
		{ 91, 120, 53 },
		{ 91, 121, 53 },
		{ 91, 122, 53 },
		{ 91, 123, 53 },
		{ 91, 124, 53 },
		{ 91, 125, 53 },
		{ 91, 126, 53 },
		{ 91, 127, 53 },
		{ 91, 128, 53 },
		{ 91, 129, 53 },
		{ 91, 130, 53 },
		{ 91, 131, 53 },
		{ 91, 132, 53 },
		{ 92, 93, 53 },
		{ 92, 94, 53 },
		{ 92, 95, 53 },
		{ 92, 96, 53 },
		{ 92, 97, 53 },
		{ 92, 98, 53 },
		{ 92, 99, 53 },
		{ 92, 100, 53 },
		{ 92, 101, 53 },
		{ 92, 102, 53 },
		{ 92, 103, 53 },
		{ 92, 104, 53 },
		{ 92, 105, 53 },
		{ 92, 106, 53 },
		{ 92, 107, 53 },
		{ 92, 108, 53 },
		{ 92, 109, 53 },
		{ 92, 110, 53 },
		{ 92, 111, 53 },
		{ 92, 112, 53 },
		{ 92, 113, 53 },
		{ 92, 114, 53 },
		{ 92, 115, 53 },
		{ 92, 116, 53 },
		{ 92, 117, 53 },
		{ 92, 118, 53 },
		{ 92, 119, 53 },
		{ 92, 120, 53 },
		{ 92, 121, 53 },
		{ 92, 122, 53 },
		{ 92, 123, 53 },
		{ 92, 124, 53 },
		{ 92, 125, 53 },
		{ 92, 126, 53 },
		{ 92, 127, 53 },
		{ 92, 128, 53 },
		{ 92, 129, 53 },
		{ 92, 130, 53 },
		{ 92, 131, 53 },
		{ 92, 132, 53 },
		{ 92, 133, 53 },
		{ 93, 94, 53 },
		{ 93, 95, 53 },
		{ 93, 96, 53 },
		{ 93, 97, 53 },
		{ 93, 98, 53 },
		{ 93, 99, 53 },
		{ 93, 100, 53 },
		{ 93, 101, 53 },
		{ 93, 102, 53 },
		{ 93, 103, 53 },
		{ 93, 104, 53 },
		{ 93, 105, 53 },
		{ 93, 106, 53 },
		{ 93, 107, 53 },
		{ 93, 108, 53 },
		{ 93, 109, 53 },
		{ 93, 110, 53 },
		{ 93, 111, 53 },
		{ 93, 112, 53 },
		{ 93, 113, 53 },
		{ 93, 114, 53 },
		{ 93, 115, 53 },
		{ 93, 116, 53 },
		{ 93, 117, 53 },
		{ 93, 118, 53 },
		{ 93, 119, 53 },
		{ 93, 120, 53 },
		{ 93, 121, 53 },
		{ 93, 122, 53 },
		{ 93, 123, 53 },
		{ 93, 124, 53 },
		{ 93, 125, 53 },
		{ 93, 126, 53 },
		{ 93, 127, 53 },
		{ 93, 128, 53 },
		{ 93, 129, 53 },
		{ 93, 130, 53 },
		{ 93, 131, 53 },
		{ 93, 132, 53 },
		{ 93, 133, 53 },
		{ 94, 95, 53 },
		{ 94, 96, 53 },
		{ 94, 97, 53 },
		{ 94, 98, 53 },
		{ 94, 99, 53 },
		{ 94, 100, 53 },
		{ 94, 101, 53 },
		{ 94, 102, 53 },
		{ 94, 103, 53 },
		{ 94, 104, 53 },
		{ 94, 105, 53 },
		{ 94, 106, 53 },
		{ 94, 107, 53 },
		{ 94, 108, 53 },
		{ 94, 109, 53 },
		{ 94, 110, 53 },
		{ 94, 111, 53 },
		{ 94, 112, 53 },
		{ 94, 113, 53 },
		{ 94, 114, 53 },
		{ 94, 115, 53 },
		{ 94, 116, 53 },
		{ 94, 117, 53 },
		{ 94, 118, 53 },
		{ 94, 119, 53 },
		{ 94, 120, 53 },
		{ 94, 121, 53 },
		{ 94, 122, 53 },
		{ 94, 123, 53 },
		{ 94, 124, 53 },
		{ 94, 125, 53 },
		{ 94, 126, 53 },
		{ 94, 127, 53 },
		{ 94, 128, 53 },
		{ 94, 129, 53 },
		{ 94, 130, 53 },
		{ 94, 131, 53 },
		{ 94, 132, 53 },
		{ 94, 133, 53 },
		{ 94, 134, 53 },
		{ 95, 96, 53 },
		{ 95, 97, 53 },
		{ 95, 98, 53 },
		{ 95, 99, 53 },
		{ 95, 100, 53 },
		{ 95, 101, 53 },
		{ 95, 102, 53 },
		{ 95, 103, 53 },
		{ 95, 104, 53 },
		{ 95, 105, 53 },
		{ 95, 106, 53 },
		{ 95, 107, 53 },
		{ 95, 108, 53 },
		{ 95, 109, 53 },
		{ 95, 110, 53 },
		{ 95, 111, 53 },
		{ 95, 112, 53 },
		{ 95, 113, 53 },
		{ 95, 114, 53 },
		{ 95, 115, 53 },
		{ 95, 116, 53 },
		{ 95, 117, 53 },
		{ 95, 118, 53 },
		{ 95, 119, 53 },
		{ 95, 120, 53 },
		{ 95, 121, 53 },
		{ 95, 122, 53 },
		{ 95, 123, 53 },
		{ 95, 124, 53 },
		{ 95, 125, 53 },
		{ 95, 126, 53 },
		{ 95, 127, 53 },
		{ 95, 128, 53 },
		{ 95, 129, 53 },
		{ 95, 130, 53 },
		{ 95, 131, 53 },
		{ 95, 132, 53 },
		{ 95, 133, 53 },
		{ 95, 134, 53 },
		{ 95, 135, 53 },
		{ 96, 97, 53 },
		{ 96, 98, 53 },
		{ 96, 99, 53 },
		{ 96, 100, 53 },
		{ 96, 101, 53 },
		{ 96, 102, 53 },
		{ 96, 103, 53 },
		{ 96, 104, 53 },
		{ 96, 105, 53 },
		{ 96, 106, 53 },
		{ 96, 107, 53 },
		{ 96, 108, 53 },
		{ 96, 109, 53 },
		{ 96, 110, 53 },
		{ 96, 111, 53 },
		{ 96, 112, 53 },
		{ 96, 113, 53 },
		{ 96, 114, 53 },
		{ 96, 115, 53 },
		{ 96, 116, 53 },
		{ 96, 117, 53 },
		{ 96, 118, 53 },
		{ 96, 119, 53 },
		{ 96, 120, 53 },
		{ 96, 121, 53 },
		{ 96, 122, 53 },
		{ 96, 123, 53 },
		{ 96, 124, 53 },
		{ 96, 125, 53 },
		{ 96, 126, 53 },
		{ 96, 127, 53 },
		{ 96, 128, 53 },
		{ 96, 129, 53 },
		{ 96, 130, 53 },
		{ 96, 131, 53 },
		{ 96, 132, 53 },
		{ 96, 133, 53 },
		{ 96, 134, 53 },
		{ 96, 135, 53 },
		{ 96, 136, 53 },
		{ 97, 98, 53 },
		{ 97, 99, 53 },
		{ 97, 100, 53 },
		{ 97, 101, 53 },
		{ 97, 102, 53 },
		{ 97, 103, 53 },
		{ 97, 104, 53 },
		{ 97, 105, 53 },
		{ 97, 106, 53 },
		{ 97, 107, 53 },
		{ 97, 108, 53 },
		{ 97, 109, 53 },
		{ 97, 110, 53 },
		{ 97, 111, 53 },
		{ 97, 112, 53 },
		{ 97, 113, 53 },
		{ 97, 114, 53 },
		{ 97, 115, 53 },
		{ 97, 116, 53 },
		{ 97, 117, 53 },
		{ 97, 118, 53 },
		{ 97, 119, 53 },
		{ 97, 120, 53 },
		{ 97, 121, 53 },
		{ 97, 122, 53 },
		{ 97, 123, 53 },
		{ 97, 124, 53 },
		{ 97, 125, 53 },
		{ 97, 126, 53 },
		{ 97, 127, 53 },
		{ 97, 128, 53 },
		{ 97, 129, 53 },
		{ 97, 130, 53 },
		{ 97, 131, 53 },
		{ 97, 132, 53 },
		{ 97, 133, 53 },
		{ 97, 134, 53 },
		{ 97, 135, 53 },
		{ 97, 136, 53 },
		{ 98, 99, 53 },
		{ 98, 100, 53 },
		{ 98, 101, 53 },
		{ 98, 102, 53 },
		{ 98, 103, 53 },
		{ 98, 104, 53 },
		{ 98, 105, 53 },
		{ 98, 106, 53 },
		{ 98, 107, 53 },
		{ 98, 108, 53 },
		{ 98, 109, 53 },
		{ 98, 110, 53 },
		{ 98, 111, 53 },
		{ 98, 112, 53 },
		{ 98, 113, 53 },
		{ 98, 114, 53 },
		{ 98, 115, 53 },
		{ 98, 116, 53 },
		{ 98, 117, 53 },
		{ 98, 118, 53 },
		{ 98, 119, 53 },
		{ 98, 120, 53 },
		{ 98, 121, 53 },
		{ 98, 122, 53 },
		{ 98, 123, 53 },
		{ 98, 124, 53 },
		{ 98, 125, 53 },
		{ 98, 126, 53 },
		{ 98, 127, 53 },
		{ 98, 128, 53 },
		{ 98, 129, 53 },
		{ 98, 130, 53 },
		{ 98, 131, 53 },
		{ 98, 132, 53 },
		{ 98, 133, 53 },
		{ 98, 134, 53 },
		{ 98, 135, 53 },
		{ 98, 136, 53 },
		{ 98, 137, 53 },
		{ 99, 100, 53 },
		{ 99, 101, 53 },
		{ 99, 102, 53 },
		{ 99, 103, 53 },
		{ 99, 104, 53 },
		{ 99, 105, 53 },
		{ 99, 106, 53 },
		{ 99, 107, 53 },
		{ 99, 108, 53 },
		{ 99, 109, 53 },
		{ 99, 110, 53 },
		{ 99, 111, 53 },
		{ 99, 112, 53 },
		{ 99, 113, 53 },
		{ 99, 114, 53 },
		{ 99, 115, 53 },
		{ 99, 116, 53 },
		{ 99, 117, 53 },
		{ 99, 118, 53 },
		{ 99, 119, 53 },
		{ 99, 120, 53 },
		{ 99, 121, 53 },
		{ 99, 122, 53 },
		{ 99, 123, 53 },
		{ 99, 124, 53 },
		{ 99, 125, 53 },
		{ 99, 126, 53 },
		{ 99, 127, 53 },
		{ 99, 128, 53 },
		{ 99, 129, 53 },
		{ 99, 130, 53 },
		{ 99, 131, 53 },
		{ 99, 132, 53 },
		{ 99, 133, 53 },
		{ 99, 134, 53 },
		{ 99, 135, 53 },
		{ 99, 136, 53 },
		{ 99, 137, 53 },
		{ 99, 138, 53 },
		{ 100, 101, 53 },
		{ 100, 102, 53 },
		{ 100, 103, 53 },
		{ 100, 104, 53 },
		{ 100, 105, 53 },
		{ 100, 106, 53 },
		{ 100, 107, 53 },
		{ 100, 108, 53 },
		{ 100, 109, 53 },
		{ 100, 110, 53 },
		{ 100, 111, 53 },
		{ 100, 112, 53 },
		{ 100, 113, 53 },
		{ 100, 114, 53 },
		{ 100, 115, 53 },
		{ 100, 116, 53 },
		{ 100, 117, 53 },
		{ 100, 118, 53 },
		{ 100, 119, 53 },
		{ 100, 120, 53 },
		{ 100, 121, 53 },
		{ 100, 122, 53 },
		{ 100, 123, 53 },
		{ 100, 124, 53 },
		{ 100, 125, 53 },
		{ 100, 126, 53 },
		{ 100, 127, 53 },
		{ 100, 128, 53 },
		{ 100, 129, 53 },
		{ 100, 130, 53 },
		{ 100, 131, 53 },
		{ 100, 132, 53 },
		{ 100, 133, 53 },
		{ 100, 134, 53 },
		{ 100, 135, 53 },
		{ 100, 136, 53 },
		{ 100, 137, 53 },
		{ 100, 138, 53 },
		{ 100, 139, 53 },
		{ 101, 102, 53 },
		{ 101, 103, 53 },
		{ 101, 104, 53 },
		{ 101, 105, 53 },
		{ 101, 106, 53 },
		{ 101, 107, 53 },
		{ 101, 108, 53 },
		{ 101, 109, 53 },
		{ 101, 110, 53 },
		{ 101, 111, 53 },
		{ 101, 112, 53 },
		{ 101, 113, 53 },
		{ 101, 114, 53 },
		{ 101, 115, 53 },
		{ 101, 116, 53 },
		{ 101, 117, 53 },
		{ 101, 118, 53 },
		{ 101, 119, 53 },
		{ 101, 120, 53 },
		{ 101, 121, 53 },
		{ 101, 122, 53 },
		{ 101, 123, 53 },
		{ 101, 124, 53 },
		{ 101, 125, 53 },
		{ 101, 126, 53 },
		{ 101, 127, 53 },
		{ 101, 128, 53 },
		{ 101, 129, 53 },
		{ 101, 130, 53 },
		{ 101, 131, 53 },
		{ 101, 132, 53 },
		{ 101, 133, 53 },
		{ 101, 134, 53 },
		{ 101, 135, 53 },
		{ 101, 136, 53 },
		{ 101, 137, 53 },
		{ 101, 138, 53 },
		{ 101, 139, 53 },
		{ 102, 103, 53 },
		{ 102, 104, 53 },
		{ 102, 105, 53 },
		{ 102, 106, 53 },
		{ 102, 107, 53 },
		{ 102, 108, 53 },
		{ 102, 109, 53 },
		{ 102, 110, 53 },
		{ 102, 111, 53 },
		{ 102, 112, 53 },
		{ 102, 113, 53 },
		{ 102, 114, 53 },
		{ 102, 115, 53 },
		{ 102, 116, 53 },
		{ 102, 117, 53 },
		{ 102, 118, 53 },
		{ 102, 119, 53 },
		{ 102, 120, 53 },
		{ 102, 121, 53 },
		{ 102, 122, 53 },
		{ 102, 123, 53 },
		{ 102, 124, 53 },
		{ 102, 125, 53 },
		{ 102, 126, 53 },
		{ 102, 127, 53 },
		{ 102, 128, 53 },
		{ 102, 129, 53 },
		{ 102, 130, 53 },
		{ 102, 131, 53 },
		{ 102, 132, 53 },
		{ 102, 133, 53 },
		{ 102, 134, 53 },
		{ 102, 135, 53 },
		{ 102, 136, 53 },
		{ 102, 137, 53 },
		{ 102, 138, 53 },
		{ 102, 139, 53 },
		{ 102, 140, 53 },
		{ 103, 104, 53 },
		{ 103, 105, 53 },
		{ 103, 106, 53 },
		{ 103, 107, 53 },
		{ 103, 108, 53 },
		{ 103, 109, 53 },
		{ 103, 110, 53 },
		{ 103, 111, 53 },
		{ 103, 112, 53 },
		{ 103, 113, 53 },
		{ 103, 114, 53 },
		{ 103, 115, 53 },
		{ 103, 116, 53 },
		{ 103, 117, 53 },
		{ 103, 118, 53 },
		{ 103, 119, 53 },
		{ 103, 120, 53 },
		{ 103, 121, 53 },
		{ 103, 122, 53 },
		{ 103, 123, 53 },
		{ 103, 124, 53 },
		{ 103, 125, 53 },
		{ 103, 126, 53 },
		{ 103, 127, 53 },
		{ 103, 128, 53 },
		{ 103, 129, 53 },
		{ 103, 130, 53 },
		{ 103, 131, 53 },
		{ 103, 132, 53 },
		{ 103, 133, 53 },
		{ 103, 134, 53 },
		{ 103, 135, 53 },
		{ 103, 136, 53 },
		{ 103, 137, 53 },
		{ 103, 138, 53 },
		{ 103, 139, 53 },
		{ 103, 140, 53 },
		{ 103, 141, 53 },
		{ 104, 105, 53 },
		{ 104, 106, 53 },
		{ 104, 107, 53 },
		{ 104, 108, 53 },
		{ 104, 109, 53 },
		{ 104, 110, 53 },
		{ 104, 111, 53 },
		{ 104, 112, 53 },
		{ 104, 113, 53 },
		{ 104, 114, 53 },
		{ 104, 115, 53 },
		{ 104, 116, 53 },
		{ 104, 117, 53 },
		{ 104, 118, 53 },
		{ 104, 119, 53 },
		{ 104, 120, 53 },
		{ 104, 121, 53 },
		{ 104, 122, 53 },
		{ 104, 123, 53 },
		{ 104, 124, 53 },
		{ 104, 125, 53 },
		{ 104, 126, 53 },
		{ 104, 127, 53 },
		{ 104, 128, 53 },
		{ 104, 129, 53 },
		{ 104, 130, 53 },
		{ 104, 131, 53 },
		{ 104, 132, 53 },
		{ 104, 133, 53 },
		{ 104, 134, 53 },
		{ 104, 135, 53 },
		{ 104, 136, 53 },
		{ 104, 137, 53 },
		{ 104, 138, 53 },
		{ 104, 139, 53 },
		{ 104, 140, 53 },
		{ 104, 141, 53 },
		{ 104, 142, 53 },
		{ 105, 106, 53 },
		{ 105, 107, 53 },
		{ 105, 108, 53 },
		{ 105, 109, 53 },
		{ 105, 110, 53 },
		{ 105, 111, 53 },
		{ 105, 112, 53 },
		{ 105, 113, 53 },
		{ 105, 114, 53 },
		{ 105, 115, 53 },
		{ 105, 116, 53 },
		{ 105, 117, 53 },
		{ 105, 118, 53 },
		{ 105, 119, 53 },
		{ 105, 120, 53 },
		{ 105, 121, 53 },
		{ 105, 122, 53 },
		{ 105, 123, 53 },
		{ 105, 124, 53 },
		{ 105, 125, 53 },
		{ 105, 126, 53 },
		{ 105, 127, 53 },
		{ 105, 128, 53 },
		{ 105, 129, 53 },
		{ 105, 130, 53 },
		{ 105, 131, 53 },
		{ 105, 132, 53 },
		{ 105, 133, 53 },
		{ 105, 134, 53 },
		{ 105, 135, 53 },
		{ 105, 136, 53 },
		{ 105, 137, 53 },
		{ 105, 138, 53 },
		{ 105, 139, 53 },
		{ 105, 140, 53 },
		{ 105, 141, 53 },
		{ 105, 142, 53 },
		{ 106, 107, 53 },
		{ 106, 108, 53 },
		{ 106, 109, 53 },
		{ 106, 110, 53 },
		{ 106, 111, 53 },
		{ 106, 112, 53 },
		{ 106, 113, 53 },
		{ 106, 114, 53 },
		{ 106, 115, 53 },
		{ 106, 116, 53 },
		{ 106, 117, 53 },
		{ 106, 118, 53 },
		{ 106, 119, 53 },
		{ 106, 120, 53 },
		{ 106, 121, 53 },
		{ 106, 122, 53 },
		{ 106, 123, 53 },
		{ 106, 124, 53 },
		{ 106, 125, 53 },
		{ 106, 126, 53 },
		{ 106, 127, 53 },
		{ 106, 128, 53 },
		{ 106, 129, 53 },
		{ 106, 130, 53 },
		{ 106, 131, 53 },
		{ 106, 132, 53 },
		{ 106, 133, 53 },
		{ 106, 134, 53 },
		{ 106, 135, 53 },
		{ 106, 136, 53 },
		{ 106, 137, 53 },
		{ 106, 138, 53 },
		{ 106, 139, 53 },
		{ 106, 140, 53 },
		{ 106, 141, 53 },
		{ 106, 142, 53 },
		{ 106, 143, 53 },
		{ 107, 108, 53 },
		{ 107, 109, 53 },
		{ 107, 110, 53 },
		{ 107, 111, 53 },
		{ 107, 112, 53 },
		{ 107, 113, 53 },
		{ 107, 114, 53 },
		{ 107, 115, 53 },
		{ 107, 116, 53 },
		{ 107, 117, 53 },
		{ 107, 118, 53 },
		{ 107, 119, 53 },
		{ 107, 120, 53 },
		{ 107, 121, 53 },
		{ 107, 122, 53 },
		{ 107, 123, 53 },
		{ 107, 124, 53 },
		{ 107, 125, 53 },
		{ 107, 126, 53 },
		{ 107, 127, 53 },
		{ 107, 128, 53 },
		{ 107, 129, 53 },
		{ 107, 130, 53 },
		{ 107, 131, 53 },
		{ 107, 132, 53 },
		{ 107, 133, 53 },
		{ 107, 134, 53 },
		{ 107, 135, 53 },
		{ 107, 136, 53 },
		{ 107, 137, 53 },
		{ 107, 138, 53 },
		{ 107, 139, 53 },
		{ 107, 140, 53 },
		{ 107, 141, 53 },
		{ 107, 142, 53 },
		{ 107, 143, 53 },
		{ 107, 144, 53 },
		{ 108, 109, 53 },
		{ 108, 110, 53 },
		{ 108, 111, 53 },
		{ 108, 112, 53 },
		{ 108, 113, 53 },
		{ 108, 114, 53 },
		{ 108, 115, 53 },
		{ 108, 116, 53 },
		{ 108, 117, 53 },
		{ 108, 118, 53 },
		{ 108, 119, 53 },
		{ 108, 120, 53 },
		{ 108, 121, 53 },
		{ 108, 122, 53 },
		{ 108, 123, 53 },
		{ 108, 124, 53 },
		{ 108, 125, 53 },
		{ 108, 126, 53 },
		{ 108, 127, 53 },
		{ 108, 128, 53 },
		{ 108, 129, 53 },
		{ 108, 130, 53 },
		{ 108, 131, 53 },
		{ 108, 132, 53 },
		{ 108, 133, 53 },
		{ 108, 134, 53 },
		{ 108, 135, 53 },
		{ 108, 136, 53 },
		{ 108, 137, 53 },
		{ 108, 138, 53 },
		{ 108, 139, 53 },
		{ 108, 140, 53 },
		{ 108, 141, 53 },
		{ 108, 142, 53 },
		{ 108, 143, 53 },
		{ 108, 144, 53 },
		{ 108, 145, 53 },
		{ 109, 110, 53 },
		{ 109, 111, 53 },
		{ 109, 112, 53 },
		{ 109, 113, 53 },
		{ 109, 114, 53 },
		{ 109, 115, 53 },
		{ 109, 116, 53 },
		{ 109, 117, 53 },
		{ 109, 118, 53 },
		{ 109, 119, 53 },
		{ 109, 120, 53 },
		{ 109, 121, 53 },
		{ 109, 122, 53 },
		{ 109, 123, 53 },
		{ 109, 124, 53 },
		{ 109, 125, 53 },
		{ 109, 126, 53 },
		{ 109, 127, 53 },
		{ 109, 128, 53 },
		{ 109, 129, 53 },
		{ 109, 130, 53 },
		{ 109, 131, 53 },
		{ 109, 132, 53 },
		{ 109, 133, 53 },
		{ 109, 134, 53 },
		{ 109, 135, 53 },
		{ 109, 136, 53 },
		{ 109, 137, 53 },
		{ 109, 138, 53 },
		{ 109, 139, 53 },
		{ 109, 140, 53 },
		{ 109, 141, 53 },
		{ 109, 142, 53 },
		{ 109, 143, 53 },
		{ 109, 144, 53 },
		{ 109, 145, 53 },
		{ 110, 111, 53 },
		{ 110, 112, 53 },
		{ 110, 113, 53 },
		{ 110, 114, 53 },
		{ 110, 115, 53 },
		{ 110, 116, 53 },
		{ 110, 117, 53 },
		{ 110, 118, 53 },
		{ 110, 119, 53 },
		{ 110, 120, 53 },
		{ 110, 121, 53 },
		{ 110, 122, 53 },
		{ 110, 123, 53 },
		{ 110, 124, 53 },
		{ 110, 125, 53 },
		{ 110, 126, 53 },
		{ 110, 127, 53 },
		{ 110, 128, 53 },
		{ 110, 129, 53 },
		{ 110, 130, 53 },
		{ 110, 131, 53 },
		{ 110, 132, 53 },
		{ 110, 133, 53 },
		{ 110, 134, 53 },
		{ 110, 135, 53 },
		{ 110, 136, 53 },
		{ 110, 137, 53 },
		{ 110, 138, 53 },
		{ 110, 139, 53 },
		{ 110, 140, 53 },
		{ 110, 141, 53 },
		{ 110, 142, 53 },
		{ 110, 143, 53 },
		{ 110, 144, 53 },
		{ 110, 145, 53 },
		{ 110, 146, 53 },
		{ 111, 112, 53 },
		{ 111, 113, 53 },
		{ 111, 114, 53 },
		{ 111, 115, 53 },
		{ 111, 116, 53 },
		{ 111, 117, 53 },
		{ 111, 118, 53 },
		{ 111, 119, 53 },
		{ 111, 120, 53 },
		{ 111, 121, 53 },
		{ 111, 122, 53 },
		{ 111, 123, 53 },
		{ 111, 124, 53 },
		{ 111, 125, 53 },
		{ 111, 126, 53 },
		{ 111, 127, 53 },
		{ 111, 128, 53 },
		{ 111, 129, 53 },
		{ 111, 130, 53 },
		{ 111, 131, 53 },
		{ 111, 132, 53 },
		{ 111, 133, 53 },
		{ 111, 134, 53 },
		{ 111, 135, 53 },
		{ 111, 136, 53 },
		{ 111, 137, 53 },
		{ 111, 138, 53 },
		{ 111, 139, 53 },
		{ 111, 140, 53 },
		{ 111, 141, 53 },
		{ 111, 142, 53 },
		{ 111, 143, 53 },
		{ 111, 144, 53 },
		{ 111, 145, 53 },
		{ 111, 146, 53 },
		{ 111, 147, 53 },
		{ 112, 113, 53 },
		{ 112, 114, 53 },
		{ 112, 115, 53 },
		{ 112, 116, 53 },
		{ 112, 117, 53 },
		{ 112, 118, 53 },
		{ 112, 119, 53 },
		{ 112, 120, 53 },
		{ 112, 121, 53 },
		{ 112, 122, 53 },
		{ 112, 123, 53 },
		{ 112, 124, 53 },
		{ 112, 125, 53 },
		{ 112, 126, 53 },
		{ 112, 127, 53 },
		{ 112, 128, 53 },
		{ 112, 129, 53 },
		{ 112, 130, 53 },
		{ 112, 131, 53 },
		{ 112, 132, 53 },
		{ 112, 133, 53 },
		{ 112, 134, 53 },
		{ 112, 135, 53 },
		{ 112, 136, 53 },
		{ 112, 137, 53 },
		{ 112, 138, 53 },
		{ 112, 139, 53 },
		{ 112, 140, 53 },
		{ 112, 141, 53 },
		{ 112, 142, 53 },
		{ 112, 143, 53 },
		{ 112, 144, 53 },
		{ 112, 145, 53 },
		{ 112, 146, 53 },
		{ 112, 147, 53 },
		{ 112, 148, 53 },
		{ 113, 114, 53 },
		{ 113, 115, 53 },
		{ 113, 116, 53 },
		{ 113, 117, 53 },
		{ 113, 118, 53 },
		{ 113, 119, 53 },
		{ 113, 120, 53 },
		{ 113, 121, 53 },
		{ 113, 122, 53 },
		{ 113, 123, 53 },
		{ 113, 124, 53 },
		{ 113, 125, 53 },
		{ 113, 126, 53 },
		{ 113, 127, 53 },
		{ 113, 128, 53 },
		{ 113, 129, 53 },
		{ 113, 130, 53 },
		{ 113, 131, 53 },
		{ 113, 132, 53 },
		{ 113, 133, 53 },
		{ 113, 134, 53 },
		{ 113, 135, 53 },
		{ 113, 136, 53 },
		{ 113, 137, 53 },
		{ 113, 138, 53 },
		{ 113, 139, 53 },
		{ 113, 140, 53 },
		{ 113, 141, 53 },
		{ 113, 142, 53 },
		{ 113, 143, 53 },
		{ 113, 144, 53 },
		{ 113, 145, 53 },
		{ 113, 146, 53 },
		{ 113, 147, 53 },
		{ 113, 148, 53 },
		{ 114, 115, 53 },
		{ 114, 116, 53 },
		{ 114, 117, 53 },
		{ 114, 118, 53 },
		{ 114, 119, 53 },
		{ 114, 120, 53 },
		{ 114, 121, 53 },
		{ 114, 122, 53 },
		{ 114, 123, 53 },
		{ 114, 124, 53 },
		{ 114, 125, 53 },
		{ 114, 126, 53 },
		{ 114, 127, 53 },
		{ 114, 128, 53 },
		{ 114, 129, 53 },
		{ 114, 130, 53 },
		{ 114, 131, 53 },
		{ 114, 132, 53 },
		{ 114, 133, 53 },
		{ 114, 134, 53 },
		{ 114, 135, 53 },
		{ 114, 136, 53 },
		{ 114, 137, 53 },
		{ 114, 138, 53 },
		{ 114, 139, 53 },
		{ 114, 140, 53 },
		{ 114, 141, 53 },
		{ 114, 142, 53 },
		{ 114, 143, 53 },
		{ 114, 144, 53 },
		{ 114, 145, 53 },
		{ 114, 146, 53 },
		{ 114, 147, 53 },
		{ 114, 148, 53 },
		{ 114, 149, 53 },
		{ 115, 116, 53 },
		{ 115, 117, 53 },
		{ 115, 118, 53 },
		{ 115, 119, 53 },
		{ 115, 120, 53 },
		{ 115, 121, 53 },
		{ 115, 122, 53 },
		{ 115, 123, 53 },
		{ 115, 124, 53 },
		{ 115, 125, 53 },
		{ 115, 126, 53 },
		{ 115, 127, 53 },
		{ 115, 128, 53 },
		{ 115, 129, 53 },
		{ 115, 130, 53 },
		{ 115, 131, 53 },
		{ 115, 132, 53 },
		{ 115, 133, 53 },
		{ 115, 134, 53 },
		{ 115, 135, 53 },
		{ 115, 136, 53 },
		{ 115, 137, 53 },
		{ 115, 138, 53 },
		{ 115, 139, 53 },
		{ 115, 140, 53 },
		{ 115, 141, 53 },
		{ 115, 142, 53 },
		{ 115, 143, 53 },
		{ 115, 144, 53 },
		{ 115, 145, 53 },
		{ 115, 146, 53 },
		{ 115, 147, 53 },
		{ 115, 148, 53 },
		{ 115, 149, 53 },
		{ 115, 150, 53 },
		{ 116, 117, 53 },
		{ 116, 118, 53 },
		{ 116, 119, 53 },
		{ 116, 120, 53 },
		{ 116, 121, 53 },
		{ 116, 122, 53 },
		{ 116, 123, 53 },
		{ 116, 124, 53 },
		{ 116, 125, 53 },
		{ 116, 126, 53 },
		{ 116, 127, 53 },
		{ 116, 128, 53 },
		{ 116, 129, 53 },
		{ 116, 130, 53 },
		{ 116, 131, 53 },
		{ 116, 132, 53 },
		{ 116, 133, 53 },
		{ 116, 134, 53 },
		{ 116, 135, 53 },
		{ 116, 136, 53 },
		{ 116, 137, 53 },
		{ 116, 138, 53 },
		{ 116, 139, 53 },
		{ 116, 140, 53 },
		{ 116, 141, 53 },
		{ 116, 142, 53 },
		{ 116, 143, 53 },
		{ 116, 144, 53 },
		{ 116, 145, 53 },
		{ 116, 146, 53 },
		{ 116, 147, 53 },
		{ 116, 148, 53 },
		{ 116, 149, 53 },
		{ 116, 150, 53 },
		{ 116, 151, 53 },
		{ 117, 118, 53 },
		{ 117, 119, 53 },
		{ 117, 120, 53 },
		{ 117, 121, 53 },
		{ 117, 122, 53 },
		{ 117, 123, 53 },
		{ 117, 124, 53 },
		{ 117, 125, 53 },
		{ 117, 126, 53 },
		{ 117, 127, 53 },
		{ 117, 128, 53 },
		{ 117, 129, 53 },
		{ 117, 130, 53 },
		{ 117, 131, 53 },
		{ 117, 132, 53 },
		{ 117, 133, 53 },
		{ 117, 134, 53 },
		{ 117, 135, 53 },
		{ 117, 136, 53 },
		{ 117, 137, 53 },
		{ 117, 138, 53 },
		{ 117, 139, 53 },
		{ 117, 140, 53 },
		{ 117, 141, 53 },
		{ 117, 142, 53 },
		{ 117, 143, 53 },
		{ 117, 144, 53 },
		{ 117, 145, 53 },
		{ 117, 146, 53 },
		{ 117, 147, 53 },
		{ 117, 148, 53 },
		{ 117, 149, 53 },
		{ 117, 150, 53 },
		{ 117, 151, 53 },
		{ 118, 119, 53 },
		{ 118, 120, 53 },
		{ 118, 121, 53 },
		{ 118, 122, 53 },
		{ 118, 123, 53 },
		{ 118, 124, 53 },
		{ 118, 125, 53 },
		{ 118, 126, 53 },
		{ 118, 127, 53 },
		{ 118, 128, 53 },
		{ 118, 129, 53 },
		{ 118, 130, 53 },
		{ 118, 131, 53 },
		{ 118, 132, 53 },
		{ 118, 133, 53 },
		{ 118, 134, 53 },
		{ 118, 135, 53 },
		{ 118, 136, 53 },
		{ 118, 137, 53 },
		{ 118, 138, 53 },
		{ 118, 139, 53 },
		{ 118, 140, 53 },
		{ 118, 141, 53 },
		{ 118, 142, 53 },
		{ 118, 143, 53 },
		{ 118, 144, 53 },
		{ 118, 145, 53 },
		{ 118, 146, 53 },
		{ 118, 147, 53 },
		{ 118, 148, 53 },
		{ 118, 149, 53 },
		{ 118, 150, 53 },
		{ 118, 151, 53 },
		{ 118, 152, 53 },
		{ 119, 120, 53 },
		{ 119, 121, 53 },
		{ 119, 122, 53 },
		{ 119, 123, 53 },
		{ 119, 124, 53 },
		{ 119, 125, 53 },
		{ 119, 126, 53 },
		{ 119, 127, 53 },
		{ 119, 128, 53 },
		{ 119, 129, 53 },
		{ 119, 130, 53 },
		{ 119, 131, 53 },
		{ 119, 132, 53 },
		{ 119, 133, 53 },
		{ 119, 134, 53 },
		{ 119, 135, 53 },
		{ 119, 136, 53 },
		{ 119, 137, 53 },
		{ 119, 138, 53 },
		{ 119, 139, 53 },
		{ 119, 140, 53 },
		{ 119, 141, 53 },
		{ 119, 142, 53 },
		{ 119, 143, 53 },
		{ 119, 144, 53 },
		{ 119, 145, 53 },
		{ 119, 146, 53 },
		{ 119, 147, 53 },
		{ 119, 148, 53 },
		{ 119, 149, 53 },
		{ 119, 150, 53 },
		{ 119, 151, 53 },
		{ 119, 152, 53 },
		{ 119, 153, 53 },
		{ 120, 121, 53 },
		{ 120, 122, 53 },
		{ 120, 123, 53 },
		{ 120, 124, 53 },
		{ 120, 125, 53 },
		{ 120, 126, 53 },
		{ 120, 127, 53 },
		{ 120, 128, 53 },
		{ 120, 129, 53 },
		{ 120, 130, 53 },
		{ 120, 131, 53 },
		{ 120, 132, 53 },
		{ 120, 133, 53 },
		{ 120, 134, 53 },
		{ 120, 135, 53 },
		{ 120, 136, 53 },
		{ 120, 137, 53 },
		{ 120, 138, 53 },
		{ 120, 139, 53 },
		{ 120, 140, 53 },
		{ 120, 141, 53 },
		{ 120, 142, 53 },
		{ 120, 143, 53 },
		{ 120, 144, 53 },
		{ 120, 145, 53 },
		{ 120, 146, 53 },
		{ 120, 147, 53 },
		{ 120, 148, 53 },
		{ 120, 149, 53 },
		{ 120, 150, 53 },
		{ 120, 151, 53 },
		{ 120, 152, 53 },
		{ 120, 153, 53 },
		{ 120, 154, 53 },
		{ 121, 122, 53 },
		{ 121, 123, 53 },
		{ 121, 124, 53 },
		{ 121, 125, 53 },
		{ 121, 126, 53 },
		{ 121, 127, 53 },
		{ 121, 128, 53 },
		{ 121, 129, 53 },
		{ 121, 130, 53 },
		{ 121, 131, 53 },
		{ 121, 132, 53 },
		{ 121, 133, 53 },
		{ 121, 134, 53 },
		{ 121, 135, 53 },
		{ 121, 136, 53 },
		{ 121, 137, 53 },
		{ 121, 138, 53 },
		{ 121, 139, 53 },
		{ 121, 140, 53 },
		{ 121, 141, 53 },
		{ 121, 142, 53 },
		{ 121, 143, 53 },
		{ 121, 144, 53 },
		{ 121, 145, 53 },
		{ 121, 146, 53 },
		{ 121, 147, 53 },
		{ 121, 148, 53 },
		{ 121, 149, 53 },
		{ 121, 150, 53 },
		{ 121, 151, 53 },
		{ 121, 152, 53 },
		{ 121, 153, 53 },
		{ 121, 154, 53 },
		{ 122, 123, 53 },
		{ 122, 124, 53 },
		{ 122, 125, 53 },
		{ 122, 126, 53 },
		{ 122, 127, 53 },
		{ 122, 128, 53 },
		{ 122, 129, 53 },
		{ 122, 130, 53 },
		{ 122, 131, 53 },
		{ 122, 132, 53 },
		{ 122, 133, 53 },
		{ 122, 134, 53 },
		{ 122, 135, 53 },
		{ 122, 136, 53 },
		{ 122, 137, 53 },
		{ 122, 138, 53 },
		{ 122, 139, 53 },
		{ 122, 140, 53 },
		{ 122, 141, 53 },
		{ 122, 142, 53 },
		{ 122, 143, 53 },
		{ 122, 144, 53 },
		{ 122, 145, 53 },
		{ 122, 146, 53 },
		{ 122, 147, 53 },
		{ 122, 148, 53 },
		{ 122, 149, 53 },
		{ 122, 150, 53 },
		{ 122, 151, 53 },
		{ 122, 152, 53 },
		{ 122, 153, 53 },
		{ 122, 154, 53 },
		{ 122, 155, 53 },
		{ 123, 124, 53 },
		{ 123, 125, 53 },
		{ 123, 126, 53 },
		{ 123, 127, 53 },
		{ 123, 128, 53 },
		{ 123, 129, 53 },
		{ 123, 130, 53 },
		{ 123, 131, 53 },
		{ 123, 132, 53 },
		{ 123, 133, 53 },
		{ 123, 134, 53 },
		{ 123, 135, 53 },
		{ 123, 136, 53 },
		{ 123, 137, 53 },
		{ 123, 138, 53 },
		{ 123, 139, 53 },
		{ 123, 140, 53 },
		{ 123, 141, 53 },
		{ 123, 142, 53 },
		{ 123, 143, 53 },
		{ 123, 144, 53 },
		{ 123, 145, 53 },
		{ 123, 146, 53 },
		{ 123, 147, 53 },
		{ 123, 148, 53 },
		{ 123, 149, 53 },
		{ 123, 150, 53 },
		{ 123, 151, 53 },
		{ 123, 152, 53 },
		{ 123, 153, 53 },
		{ 123, 154, 53 },
		{ 123, 155, 53 },
		{ 123, 156, 53 },
		{ 124, 125, 53 },
		{ 124, 126, 53 },
		{ 124, 127, 53 },
		{ 124, 128, 53 },
		{ 124, 129, 53 },
		{ 124, 130, 53 },
		{ 124, 131, 53 },
		{ 124, 132, 53 },
		{ 124, 133, 53 },
		{ 124, 134, 53 },
		{ 124, 135, 53 },
		{ 124, 136, 53 },
		{ 124, 137, 53 },
		{ 124, 138, 53 },
		{ 124, 139, 53 },
		{ 124, 140, 53 },
		{ 124, 141, 53 },
		{ 124, 142, 53 },
		{ 124, 143, 53 },
		{ 124, 144, 53 },
		{ 124, 145, 53 },
		{ 124, 146, 53 },
		{ 124, 147, 53 },
		{ 124, 148, 53 },
		{ 124, 149, 53 },
		{ 124, 150, 53 },
		{ 124, 151, 53 },
		{ 124, 152, 53 },
		{ 124, 153, 53 },
		{ 124, 154, 53 },
		{ 124, 155, 53 },
		{ 124, 156, 53 },
		{ 124, 157, 53 },
		{ 125, 126, 53 },
		{ 125, 127, 53 },
		{ 125, 128, 53 },
		{ 125, 129, 53 },
		{ 125, 130, 53 },
		{ 125, 131, 53 },
		{ 125, 132, 53 },
		{ 125, 133, 53 },
		{ 125, 134, 53 },
		{ 125, 135, 53 },
		{ 125, 136, 53 },
		{ 125, 137, 53 },
		{ 125, 138, 53 },
		{ 125, 139, 53 },
		{ 125, 140, 53 },
		{ 125, 141, 53 },
		{ 125, 142, 53 },
		{ 125, 143, 53 },
		{ 125, 144, 53 },
		{ 125, 145, 53 },
		{ 125, 146, 53 },
		{ 125, 147, 53 },
		{ 125, 148, 53 },
		{ 125, 149, 53 },
		{ 125, 150, 53 },
		{ 125, 151, 53 },
		{ 125, 152, 53 },
		{ 125, 153, 53 },
		{ 125, 154, 53 },
		{ 125, 155, 53 },
		{ 125, 156, 53 },
		{ 125, 157, 53 },
		{ 126, 127, 53 },
		{ 126, 128, 53 },
		{ 126, 129, 53 },
		{ 126, 130, 53 },
		{ 126, 131, 53 },
		{ 126, 132, 53 },
		{ 126, 133, 53 },
		{ 126, 134, 53 },
		{ 126, 135, 53 },
		{ 126, 136, 53 },
		{ 126, 137, 53 },
		{ 126, 138, 53 },
		{ 126, 139, 53 },
		{ 126, 140, 53 },
		{ 126, 141, 53 },
		{ 126, 142, 53 },
		{ 126, 143, 53 },
		{ 126, 144, 53 },
		{ 126, 145, 53 },
		{ 126, 146, 53 },
		{ 126, 147, 53 },
		{ 126, 148, 53 },
		{ 126, 149, 53 },
		{ 126, 150, 53 },
		{ 126, 151, 53 },
		{ 126, 152, 53 },
		{ 126, 153, 53 },
		{ 126, 154, 53 },
		{ 126, 155, 53 },
		{ 126, 156, 53 },
		{ 126, 157, 53 },
		{ 126, 158, 53 },
		{ 127, 128, 53 },
		{ 127, 129, 53 },
		{ 127, 130, 53 },
		{ 127, 131, 53 },
		{ 127, 132, 53 },
		{ 127, 133, 53 },
		{ 127, 134, 53 },
		{ 127, 135, 53 },
		{ 127, 136, 53 },
		{ 127, 137, 53 },
		{ 127, 138, 53 },
		{ 127, 139, 53 },
		{ 127, 140, 53 },
		{ 127, 141, 53 },
		{ 127, 142, 53 },
		{ 127, 143, 53 },
		{ 127, 144, 53 },
		{ 127, 145, 53 },
		{ 127, 146, 53 },
		{ 127, 147, 53 },
		{ 127, 148, 53 },
		{ 127, 149, 53 },
		{ 127, 150, 53 },
		{ 127, 151, 53 },
		{ 127, 152, 53 },
		{ 127, 153, 53 },
		{ 127, 154, 53 },
		{ 127, 155, 53 },
		{ 127, 156, 53 },
		{ 127, 157, 53 },
		{ 127, 158, 53 },
		{ 127, 159, 53 },
		{ 128, 129, 53 },
		{ 128, 130, 53 },
		{ 128, 131, 53 },
		{ 128, 132, 53 },
		{ 128, 133, 53 },
		{ 128, 134, 53 },
		{ 128, 135, 53 },
		{ 128, 136, 53 },
		{ 128, 137, 53 },
		{ 128, 138, 53 },
		{ 128, 139, 53 },
		{ 128, 140, 53 },
		{ 128, 141, 53 },
		{ 128, 142, 53 },
		{ 128, 143, 53 },
		{ 128, 144, 53 },
		{ 128, 145, 53 },
		{ 128, 146, 53 },
		{ 128, 147, 53 },
		{ 128, 148, 53 },
		{ 128, 149, 53 },
		{ 128, 150, 53 },
		{ 128, 151, 53 },
		{ 128, 152, 53 },
		{ 128, 153, 53 },
		{ 128, 154, 53 },
		{ 128, 155, 53 },
		{ 128, 156, 53 },
		{ 128, 157, 53 },
		{ 128, 158, 53 },
		{ 128, 159, 53 },
		{ 128, 160, 53 },
		{ 129, 130, 53 },
		{ 129, 131, 53 },
		{ 129, 132, 53 },
		{ 129, 133, 53 },
		{ 129, 134, 53 },
		{ 129, 135, 53 },
		{ 129, 136, 53 },
		{ 129, 137, 53 },
		{ 129, 138, 53 },
		{ 129, 139, 53 },
		{ 129, 140, 53 },
		{ 129, 141, 53 },
		{ 129, 142, 53 },
		{ 129, 143, 53 },
		{ 129, 144, 53 },
		{ 129, 145, 53 },
		{ 129, 146, 53 },
		{ 129, 147, 53 },
		{ 129, 148, 53 },
		{ 129, 149, 53 },
		{ 129, 150, 53 },
		{ 129, 151, 53 },
		{ 129, 152, 53 },
		{ 129, 153, 53 },
		{ 129, 154, 53 },
		{ 129, 155, 53 },
		{ 129, 156, 53 },
		{ 129, 157, 53 },
		{ 129, 158, 53 },
		{ 129, 159, 53 },
		{ 129, 160, 53 },
		{ 130, 131, 53 },
		{ 130, 132, 53 },
		{ 130, 133, 53 },
		{ 130, 134, 53 },
		{ 130, 135, 53 },
		{ 130, 136, 53 },
		{ 130, 137, 53 },
		{ 130, 138, 53 },
		{ 130, 139, 53 },
		{ 130, 140, 53 },
		{ 130, 141, 53 },
		{ 130, 142, 53 },
		{ 130, 143, 53 },
		{ 130, 144, 53 },
		{ 130, 145, 53 },
		{ 130, 146, 53 },
		{ 130, 147, 53 },
		{ 130, 148, 53 },
		{ 130, 149, 53 },
		{ 130, 150, 53 },
		{ 130, 151, 53 },
		{ 130, 152, 53 },
		{ 130, 153, 53 },
		{ 130, 154, 53 },
		{ 130, 155, 53 },
		{ 130, 156, 53 },
		{ 130, 157, 53 },
		{ 130, 158, 53 },
		{ 130, 159, 53 },
		{ 130, 160, 53 },
		{ 130, 161, 53 },
		{ 131, 132, 53 },
		{ 131, 133, 53 },
		{ 131, 134, 53 },
		{ 131, 135, 53 },
		{ 131, 136, 53 },
		{ 131, 137, 53 },
		{ 131, 138, 53 },
		{ 131, 139, 53 },
		{ 131, 140, 53 },
		{ 131, 141, 53 },
		{ 131, 142, 53 },
		{ 131, 143, 53 },
		{ 131, 144, 53 },
		{ 131, 145, 53 },
		{ 131, 146, 53 },
		{ 131, 147, 53 },
		{ 131, 148, 53 },
		{ 131, 149, 53 },
		{ 131, 150, 53 },
		{ 131, 151, 53 },
		{ 131, 152, 53 },
		{ 131, 153, 53 },
		{ 131, 154, 53 },
		{ 131, 155, 53 },
		{ 131, 156, 53 },
		{ 131, 157, 53 },
		{ 131, 158, 53 },
		{ 131, 159, 53 },
		{ 131, 160, 53 },
		{ 131, 161, 53 },
		{ 131, 162, 53 },
		{ 132, 133, 53 },
		{ 132, 134, 53 },
		{ 132, 135, 53 },
		{ 132, 136, 53 },
		{ 132, 137, 53 },
		{ 132, 138, 53 },
		{ 132, 139, 53 },
		{ 132, 140, 53 },
		{ 132, 141, 53 },
		{ 132, 142, 53 },
		{ 132, 143, 53 },
		{ 132, 144, 53 },
		{ 132, 145, 53 },
		{ 132, 146, 53 },
		{ 132, 147, 53 },
		{ 132, 148, 53 },
		{ 132, 149, 53 },
		{ 132, 150, 53 },
		{ 132, 151, 53 },
		{ 132, 152, 53 },
		{ 132, 153, 53 },
		{ 132, 154, 53 },
		{ 132, 155, 53 },
		{ 132, 156, 53 },
		{ 132, 157, 53 },
		{ 132, 158, 53 },
		{ 132, 159, 53 },
		{ 132, 160, 53 },
		{ 132, 161, 53 },
		{ 132, 162, 53 },
		{ 132, 163, 53 },
		{ 133, 134, 53 },
		{ 133, 135, 53 },
		{ 133, 136, 53 },
		{ 133, 137, 53 },
		{ 133, 138, 53 },
		{ 133, 139, 53 },
		{ 133, 140, 53 },
		{ 133, 141, 53 },
		{ 133, 142, 53 },
		{ 133, 143, 53 },
		{ 133, 144, 53 },
		{ 133, 145, 53 },
		{ 133, 146, 53 },
		{ 133, 147, 53 },
		{ 133, 148, 53 },
		{ 133, 149, 53 },
		{ 133, 150, 53 },
		{ 133, 151, 53 },
		{ 133, 152, 53 },
		{ 133, 153, 53 },
		{ 133, 154, 53 },
		{ 133, 155, 53 },
		{ 133, 156, 53 },
		{ 133, 157, 53 },
		{ 133, 158, 53 },
		{ 133, 159, 53 },
		{ 133, 160, 53 },
		{ 133, 161, 53 },
		{ 133, 162, 53 },
		{ 133, 163, 53 },
		{ 134, 135, 53 },
		{ 134, 136, 53 },
		{ 134, 137, 53 },
		{ 134, 138, 53 },
		{ 134, 139, 53 },
		{ 134, 140, 53 },
		{ 134, 141, 53 },
		{ 134, 142, 53 },
		{ 134, 143, 53 },
		{ 134, 144, 53 },
		{ 134, 145, 53 },
		{ 134, 146, 53 },
		{ 134, 147, 53 },
		{ 134, 148, 53 },
		{ 134, 149, 53 },
		{ 134, 150, 53 },
		{ 134, 151, 53 },
		{ 134, 152, 53 },
		{ 134, 153, 53 },
		{ 134, 154, 53 },
		{ 134, 155, 53 },
		{ 134, 156, 53 },
		{ 134, 157, 53 },
		{ 134, 158, 53 },
		{ 134, 159, 53 },
		{ 134, 160, 53 },
		{ 134, 161, 53 },
		{ 134, 162, 53 },
		{ 134, 163, 53 },
		{ 134, 164, 53 },
		{ 135, 136, 53 },
		{ 135, 137, 53 },
		{ 135, 138, 53 },
		{ 135, 139, 53 },
		{ 135, 140, 53 },
		{ 135, 141, 53 },
		{ 135, 142, 53 },
		{ 135, 143, 53 },
		{ 135, 144, 53 },
		{ 135, 145, 53 },
		{ 135, 146, 53 },
		{ 135, 147, 53 },
		{ 135, 148, 53 },
		{ 135, 149, 53 },
		{ 135, 150, 53 },
		{ 135, 151, 53 },
		{ 135, 152, 53 },
		{ 135, 153, 53 },
		{ 135, 154, 53 },
		{ 135, 155, 53 },
		{ 135, 156, 53 },
		{ 135, 157, 53 },
		{ 135, 158, 53 },
		{ 135, 159, 53 },
		{ 135, 160, 53 },
		{ 135, 161, 53 },
		{ 135, 162, 53 },
		{ 135, 163, 53 },
		{ 135, 164, 53 },
		{ 135, 165, 53 },
		{ 136, 137, 53 },
		{ 136, 138, 53 },
		{ 136, 139, 53 },
		{ 136, 140, 53 },
		{ 136, 141, 53 },
		{ 136, 142, 53 },
		{ 136, 143, 53 },
		{ 136, 144, 53 },
		{ 136, 145, 53 },
		{ 136, 146, 53 },
		{ 136, 147, 53 },
		{ 136, 148, 53 },
		{ 136, 149, 53 },
		{ 136, 150, 53 },
		{ 136, 151, 53 },
		{ 136, 152, 53 },
		{ 136, 153, 53 },
		{ 136, 154, 53 },
		{ 136, 155, 53 },
		{ 136, 156, 53 },
		{ 136, 157, 53 },
		{ 136, 158, 53 },
		{ 136, 159, 53 },
		{ 136, 160, 53 },
		{ 136, 161, 53 },
		{ 136, 162, 53 },
		{ 136, 163, 53 },
		{ 136, 164, 53 },
		{ 136, 165, 53 },
		{ 136, 166, 53 },
		{ 137, 138, 53 },
		{ 137, 139, 53 },
		{ 137, 140, 53 },
		{ 137, 141, 53 },
		{ 137, 142, 53 },
		{ 137, 143, 53 },
		{ 137, 144, 53 },
		{ 137, 145, 53 },
		{ 137, 146, 53 },
		{ 137, 147, 53 },
		{ 137, 148, 53 },
		{ 137, 149, 53 },
		{ 137, 150, 53 },
		{ 137, 151, 53 },
		{ 137, 152, 53 },
		{ 137, 153, 53 },
		{ 137, 154, 53 },
		{ 137, 155, 53 },
		{ 137, 156, 53 },
		{ 137, 157, 53 },
		{ 137, 158, 53 },
		{ 137, 159, 53 },
		{ 137, 160, 53 },
		{ 137, 161, 53 },
		{ 137, 162, 53 },
		{ 137, 163, 53 },
		{ 137, 164, 53 },
		{ 137, 165, 53 },
		{ 137, 166, 53 },
		{ 138, 139, 53 },
		{ 138, 140, 53 },
		{ 138, 141, 53 },
		{ 138, 142, 53 },
		{ 138, 143, 53 },
		{ 138, 144, 53 },
		{ 138, 145, 53 },
		{ 138, 146, 53 },
		{ 138, 147, 53 },
		{ 138, 148, 53 },
		{ 138, 149, 53 },
		{ 138, 150, 53 },
		{ 138, 151, 53 },
		{ 138, 152, 53 },
		{ 138, 153, 53 },
		{ 138, 154, 53 },
		{ 138, 155, 53 },
		{ 138, 156, 53 },
		{ 138, 157, 53 },
		{ 138, 158, 53 },
		{ 138, 159, 53 },
		{ 138, 160, 53 },
		{ 138, 161, 53 },
		{ 138, 162, 53 },
		{ 138, 163, 53 },
		{ 138, 164, 53 },
		{ 138, 165, 53 },
		{ 138, 166, 53 },
		{ 138, 167, 53 },
		{ 139, 140, 53 },
		{ 139, 141, 53 },
		{ 139, 142, 53 },
		{ 139, 143, 53 },
		{ 139, 144, 53 },
		{ 139, 145, 53 },
		{ 139, 146, 53 },
		{ 139, 147, 53 },
		{ 139, 148, 53 },
		{ 139, 149, 53 },
		{ 139, 150, 53 },
		{ 139, 151, 53 },
		{ 139, 152, 53 },
		{ 139, 153, 53 },
		{ 139, 154, 53 },
		{ 139, 155, 53 },
		{ 139, 156, 53 },
		{ 139, 157, 53 },
		{ 139, 158, 53 },
		{ 139, 159, 53 },
		{ 139, 160, 53 },
		{ 139, 161, 53 },
		{ 139, 162, 53 },
		{ 139, 163, 53 },
		{ 139, 164, 53 },
		{ 139, 165, 53 },
		{ 139, 166, 53 },
		{ 139, 167, 53 },
		{ 139, 168, 53 },
		{ 140, 141, 53 },
		{ 140, 142, 53 },
		{ 140, 143, 53 },
		{ 140, 144, 53 },
		{ 140, 145, 53 },
		{ 140, 146, 53 },
		{ 140, 147, 53 },
		{ 140, 148, 53 },
		{ 140, 149, 53 },
		{ 140, 150, 53 },
		{ 140, 151, 53 },
		{ 140, 152, 53 },
		{ 140, 153, 53 },
		{ 140, 154, 53 },
		{ 140, 155, 53 },
		{ 140, 156, 53 },
		{ 140, 157, 53 },
		{ 140, 158, 53 },
		{ 140, 159, 53 },
		{ 140, 160, 53 },
		{ 140, 161, 53 },
		{ 140, 162, 53 },
		{ 140, 163, 53 },
		{ 140, 164, 53 },
		{ 140, 165, 53 },
		{ 140, 166, 53 },
		{ 140, 167, 53 },
		{ 140, 168, 53 },
		{ 140, 169, 53 },
		{ 141, 142, 53 },
		{ 141, 143, 53 },
		{ 141, 144, 53 },
		{ 141, 145, 53 },
		{ 141, 146, 53 },
		{ 141, 147, 53 },
		{ 141, 148, 53 },
		{ 141, 149, 53 },
		{ 141, 150, 53 },
		{ 141, 151, 53 },
		{ 141, 152, 53 },
		{ 141, 153, 53 },
		{ 141, 154, 53 },
		{ 141, 155, 53 },
		{ 141, 156, 53 },
		{ 141, 157, 53 },
		{ 141, 158, 53 },
		{ 141, 159, 53 },
		{ 141, 160, 53 },
		{ 141, 161, 53 },
		{ 141, 162, 53 },
		{ 141, 163, 53 },
		{ 141, 164, 53 },
		{ 141, 165, 53 },
		{ 141, 166, 53 },
		{ 141, 167, 53 },
		{ 141, 168, 53 },
		{ 141, 169, 53 },
		{ 142, 143, 53 },
		{ 142, 144, 53 },
		{ 142, 145, 53 },
		{ 142, 146, 53 },
		{ 142, 147, 53 },
		{ 142, 148, 53 },
		{ 142, 149, 53 },
		{ 142, 150, 53 },
		{ 142, 151, 53 },
		{ 142, 152, 53 },
		{ 142, 153, 53 },
		{ 142, 154, 53 },
		{ 142, 155, 53 },
		{ 142, 156, 53 },
		{ 142, 157, 53 },
		{ 142, 158, 53 },
		{ 142, 159, 53 },
		{ 142, 160, 53 },
		{ 142, 161, 53 },
		{ 142, 162, 53 },
		{ 142, 163, 53 },
		{ 142, 164, 53 },
		{ 142, 165, 53 },
		{ 142, 166, 53 },
		{ 142, 167, 53 },
		{ 142, 168, 53 },
		{ 142, 169, 53 },
		{ 142, 170, 53 },
		{ 143, 144, 53 },
		{ 143, 145, 53 },
		{ 143, 146, 53 },
		{ 143, 147, 53 },
		{ 143, 148, 53 },
		{ 143, 149, 53 },
		{ 143, 150, 53 },
		{ 143, 151, 53 },
		{ 143, 152, 53 },
		{ 143, 153, 53 },
		{ 143, 154, 53 },
		{ 143, 155, 53 },
		{ 143, 156, 53 },
		{ 143, 157, 53 },
		{ 143, 158, 53 },
		{ 143, 159, 53 },
		{ 143, 160, 53 },
		{ 143, 161, 53 },
		{ 143, 162, 53 },
		{ 143, 163, 53 },
		{ 143, 164, 53 },
		{ 143, 165, 53 },
		{ 143, 166, 53 },
		{ 143, 167, 53 },
		{ 143, 168, 53 },
		{ 143, 169, 53 },
		{ 143, 170, 53 },
		{ 143, 171, 53 },
		{ 144, 145, 53 },
		{ 144, 146, 53 },
		{ 144, 147, 53 },
		{ 144, 148, 53 },
		{ 144, 149, 53 },
		{ 144, 150, 53 },
		{ 144, 151, 53 },
		{ 144, 152, 53 },
		{ 144, 153, 53 },
		{ 144, 154, 53 },
		{ 144, 155, 53 },
		{ 144, 156, 53 },
		{ 144, 157, 53 },
		{ 144, 158, 53 },
		{ 144, 159, 53 },
		{ 144, 160, 53 },
		{ 144, 161, 53 },
		{ 144, 162, 53 },
		{ 144, 163, 53 },
		{ 144, 164, 53 },
		{ 144, 165, 53 },
		{ 144, 166, 53 },
		{ 144, 167, 53 },
		{ 144, 168, 53 },
		{ 144, 169, 53 },
		{ 144, 170, 53 },
		{ 144, 171, 53 },
		{ 144, 172, 53 },
		{ 145, 146, 53 },
		{ 145, 147, 53 },
		{ 145, 148, 53 },
		{ 145, 149, 53 },
		{ 145, 150, 53 },
		{ 145, 151, 53 },
		{ 145, 152, 53 },
		{ 145, 153, 53 },
		{ 145, 154, 53 },
		{ 145, 155, 53 },
		{ 145, 156, 53 },
		{ 145, 157, 53 },
		{ 145, 158, 53 },
		{ 145, 159, 53 },
		{ 145, 160, 53 },
		{ 145, 161, 53 },
		{ 145, 162, 53 },
		{ 145, 163, 53 },
		{ 145, 164, 53 },
		{ 145, 165, 53 },
		{ 145, 166, 53 },
		{ 145, 167, 53 },
		{ 145, 168, 53 },
		{ 145, 169, 53 },
		{ 145, 170, 53 },
		{ 145, 171, 53 },
		{ 145, 172, 53 },
		{ 146, 147, 53 },
		{ 146, 148, 53 },
		{ 146, 149, 53 },
		{ 146, 150, 53 },
		{ 146, 151, 53 },
		{ 146, 152, 53 },
		{ 146, 153, 53 },
		{ 146, 154, 53 },
		{ 146, 155, 53 },
		{ 146, 156, 53 },
		{ 146, 157, 53 },
		{ 146, 158, 53 },
		{ 146, 159, 53 },
		{ 146, 160, 53 },
		{ 146, 161, 53 },
		{ 146, 162, 53 },
		{ 146, 163, 53 },
		{ 146, 164, 53 },
		{ 146, 165, 53 },
		{ 146, 166, 53 },
		{ 146, 167, 53 },
		{ 146, 168, 53 },
		{ 146, 169, 53 },
		{ 146, 170, 53 },
		{ 146, 171, 53 },
		{ 146, 172, 53 },
		{ 146, 173, 53 },
		{ 147, 148, 53 },
		{ 147, 149, 53 },
		{ 147, 150, 53 },
		{ 147, 151, 53 },
		{ 147, 152, 53 },
		{ 147, 153, 53 },
		{ 147, 154, 53 },
		{ 147, 155, 53 },
		{ 147, 156, 53 },
		{ 147, 157, 53 },
		{ 147, 158, 53 },
		{ 147, 159, 53 },
		{ 147, 160, 53 },
		{ 147, 161, 53 },
		{ 147, 162, 53 },
		{ 147, 163, 53 },
		{ 147, 164, 53 },
		{ 147, 165, 53 },
		{ 147, 166, 53 },
		{ 147, 167, 53 },
		{ 147, 168, 53 },
		{ 147, 169, 53 },
		{ 147, 170, 53 },
		{ 147, 171, 53 },
		{ 147, 172, 53 },
		{ 147, 173, 53 },
		{ 147, 174, 53 },
		{ 148, 149, 53 },
		{ 148, 150, 53 },
		{ 148, 151, 53 },
		{ 148, 152, 53 },
		{ 148, 153, 53 },
		{ 148, 154, 53 },
		{ 148, 155, 53 },
		{ 148, 156, 53 },
		{ 148, 157, 53 },
		{ 148, 158, 53 },
		{ 148, 159, 53 },
		{ 148, 160, 53 },
		{ 148, 161, 53 },
		{ 148, 162, 53 },
		{ 148, 163, 53 },
		{ 148, 164, 53 },
		{ 148, 165, 53 },
		{ 148, 166, 53 },
		{ 148, 167, 53 },
		{ 148, 168, 53 },
		{ 148, 169, 53 },
		{ 148, 170, 53 },
		{ 148, 171, 53 },
		{ 148, 172, 53 },
		{ 148, 173, 53 },
		{ 148, 174, 53 },
		{ 148, 175, 53 },
		{ 149, 150, 53 },
		{ 149, 151, 53 },
		{ 149, 152, 53 },
		{ 149, 153, 53 },
		{ 149, 154, 53 },
		{ 149, 155, 53 },
		{ 149, 156, 53 },
		{ 149, 157, 53 },
		{ 149, 158, 53 },
		{ 149, 159, 53 },
		{ 149, 160, 53 },
		{ 149, 161, 53 },
		{ 149, 162, 53 },
		{ 149, 163, 53 },
		{ 149, 164, 53 },
		{ 149, 165, 53 },
		{ 149, 166, 53 },
		{ 149, 167, 53 },
		{ 149, 168, 53 },
		{ 149, 169, 53 },
		{ 149, 170, 53 },
		{ 149, 171, 53 },
		{ 149, 172, 53 },
		{ 149, 173, 53 },
		{ 149, 174, 53 },
		{ 149, 175, 53 },
		{ 150, 151, 53 },
		{ 150, 152, 53 },
		{ 150, 153, 53 },
		{ 150, 154, 53 },
		{ 150, 155, 53 },
		{ 150, 156, 53 },
		{ 150, 157, 53 },
		{ 150, 158, 53 },
		{ 150, 159, 53 },
		{ 150, 160, 53 },
		{ 150, 161, 53 },
		{ 150, 162, 53 },
		{ 150, 163, 53 },
		{ 150, 164, 53 },
		{ 150, 165, 53 },
		{ 150, 166, 53 },
		{ 150, 167, 53 },
		{ 150, 168, 53 },
		{ 150, 169, 53 },
		{ 150, 170, 53 },
		{ 150, 171, 53 },
		{ 150, 172, 53 },
		{ 150, 173, 53 },
		{ 150, 174, 53 },
		{ 150, 175, 53 },
		{ 150, 176, 53 },
		{ 151, 152, 53 },
		{ 151, 153, 53 },
		{ 151, 154, 53 },
		{ 151, 155, 53 },
		{ 151, 156, 53 },
		{ 151, 157, 53 },
		{ 151, 158, 53 },
		{ 151, 159, 53 },
		{ 151, 160, 53 },
		{ 151, 161, 53 },
		{ 151, 162, 53 },
		{ 151, 163, 53 },
		{ 151, 164, 53 },
		{ 151, 165, 53 },
		{ 151, 166, 53 },
		{ 151, 167, 53 },
		{ 151, 168, 53 },
		{ 151, 169, 53 },
		{ 151, 170, 53 },
		{ 151, 171, 53 },
		{ 151, 172, 53 },
		{ 151, 173, 53 },
		{ 151, 174, 53 },
		{ 151, 175, 53 },
		{ 151, 176, 53 },
		{ 151, 177, 53 },
		{ 152, 153, 53 },
		{ 152, 154, 53 },
		{ 152, 155, 53 },
		{ 152, 156, 53 },
		{ 152, 157, 53 },
		{ 152, 158, 53 },
		{ 152, 159, 53 },
		{ 152, 160, 53 },
		{ 152, 161, 53 },
		{ 152, 162, 53 },
		{ 152, 163, 53 },
		{ 152, 164, 53 },
		{ 152, 165, 53 },
		{ 152, 166, 53 },
		{ 152, 167, 53 },
		{ 152, 168, 53 },
		{ 152, 169, 53 },
		{ 152, 170, 53 },
		{ 152, 171, 53 },
		{ 152, 172, 53 },
		{ 152, 173, 53 },
		{ 152, 174, 53 },
		{ 152, 175, 53 },
		{ 152, 176, 53 },
		{ 152, 177, 53 },
		{ 152, 178, 53 },
		{ 153, 154, 53 },
		{ 153, 155, 53 },
		{ 153, 156, 53 },
		{ 153, 157, 53 },
		{ 153, 158, 53 },
		{ 153, 159, 53 },
		{ 153, 160, 53 },
		{ 153, 161, 53 },
		{ 153, 162, 53 },
		{ 153, 163, 53 },
		{ 153, 164, 53 },
		{ 153, 165, 53 },
		{ 153, 166, 53 },
		{ 153, 167, 53 },
		{ 153, 168, 53 },
		{ 153, 169, 53 },
		{ 153, 170, 53 },
		{ 153, 171, 53 },
		{ 153, 172, 53 },
		{ 153, 173, 53 },
		{ 153, 174, 53 },
		{ 153, 175, 53 },
		{ 153, 176, 53 },
		{ 153, 177, 53 },
		{ 153, 178, 53 },
		{ 154, 155, 53 },
		{ 154, 156, 53 },
		{ 154, 157, 53 },
		{ 154, 158, 53 },
		{ 154, 159, 53 },
		{ 154, 160, 53 },
		{ 154, 161, 53 },
		{ 154, 162, 53 },
		{ 154, 163, 53 },
		{ 154, 164, 53 },
		{ 154, 165, 53 },
		{ 154, 166, 53 },
		{ 154, 167, 53 },
		{ 154, 168, 53 },
		{ 154, 169, 53 },
		{ 154, 170, 53 },
		{ 154, 171, 53 },
		{ 154, 172, 53 },
		{ 154, 173, 53 },
		{ 154, 174, 53 },
		{ 154, 175, 53 },
		{ 154, 176, 53 },
		{ 154, 177, 53 },
		{ 154, 178, 53 },
		{ 154, 179, 53 },
		{ 155, 156, 53 },
		{ 155, 157, 53 },
		{ 155, 158, 53 },
		{ 155, 159, 53 },
		{ 155, 160, 53 },
		{ 155, 161, 53 },
		{ 155, 162, 53 },
		{ 155, 163, 53 },
		{ 155, 164, 53 },
		{ 155, 165, 53 },
		{ 155, 166, 53 },
		{ 155, 167, 53 },
		{ 155, 168, 53 },
		{ 155, 169, 53 },
		{ 155, 170, 53 },
		{ 155, 171, 53 },
		{ 155, 172, 53 },
		{ 155, 173, 53 },
		{ 155, 174, 53 },
		{ 155, 175, 53 },
		{ 155, 176, 53 },
		{ 155, 177, 53 },
		{ 155, 178, 53 },
		{ 155, 179, 53 },
		{ 155, 180, 53 },
		{ 156, 157, 53 },
		{ 156, 158, 53 },
		{ 156, 159, 53 },
		{ 156, 160, 53 },
		{ 156, 161, 53 },
		{ 156, 162, 53 },
		{ 156, 163, 53 },
		{ 156, 164, 53 },
		{ 156, 165, 53 },
		{ 156, 166, 53 },
		{ 156, 167, 53 },
		{ 156, 168, 53 },
		{ 156, 169, 53 },
		{ 156, 170, 53 },
		{ 156, 171, 53 },
		{ 156, 172, 53 },
		{ 156, 173, 53 },
		{ 156, 174, 53 },
		{ 156, 175, 53 },
		{ 156, 176, 53 },
		{ 156, 177, 53 },
		{ 156, 178, 53 },
		{ 156, 179, 53 },
		{ 156, 180, 53 },
		{ 156, 181, 53 },
		{ 157, 158, 53 },
		{ 157, 159, 53 },
		{ 157, 160, 53 },
		{ 157, 161, 53 },
		{ 157, 162, 53 },
		{ 157, 163, 53 },
		{ 157, 164, 53 },
		{ 157, 165, 53 },
		{ 157, 166, 53 },
		{ 157, 167, 53 },
		{ 157, 168, 53 },
		{ 157, 169, 53 },
		{ 157, 170, 53 },
		{ 157, 171, 53 },
		{ 157, 172, 53 },
		{ 157, 173, 53 },
		{ 157, 174, 53 },
		{ 157, 175, 53 },
		{ 157, 176, 53 },
		{ 157, 177, 53 },
		{ 157, 178, 53 },
		{ 157, 179, 53 },
		{ 157, 180, 53 },
		{ 157, 181, 53 },
		{ 158, 159, 53 },
		{ 158, 160, 53 },
		{ 158, 161, 53 },
		{ 158, 162, 53 },
		{ 158, 163, 53 },
		{ 158, 164, 53 },
		{ 158, 165, 53 },
		{ 158, 166, 53 },
		{ 158, 167, 53 },
		{ 158, 168, 53 },
		{ 158, 169, 53 },
		{ 158, 170, 53 },
		{ 158, 171, 53 },
		{ 158, 172, 53 },
		{ 158, 173, 53 },
		{ 158, 174, 53 },
		{ 158, 175, 53 },
		{ 158, 176, 53 },
		{ 158, 177, 53 },
		{ 158, 178, 53 },
		{ 158, 179, 53 },
		{ 158, 180, 53 },
		{ 158, 181, 53 },
		{ 158, 182, 53 },
		{ 159, 160, 53 },
		{ 159, 161, 53 },
		{ 159, 162, 53 },
		{ 159, 163, 53 },
		{ 159, 164, 53 },
		{ 159, 165, 53 },
		{ 159, 166, 53 },
		{ 159, 167, 53 },
		{ 159, 168, 53 },
		{ 159, 169, 53 },
		{ 159, 170, 53 },
		{ 159, 171, 53 },
		{ 159, 172, 53 },
		{ 159, 173, 53 },
		{ 159, 174, 53 },
		{ 159, 175, 53 },
		{ 159, 176, 53 },
		{ 159, 177, 53 },
		{ 159, 178, 53 },
		{ 159, 179, 53 },
		{ 159, 180, 53 },
		{ 159, 181, 53 },
		{ 159, 182, 53 },
		{ 159, 183, 53 },
		{ 160, 161, 53 },
		{ 160, 162, 53 },
		{ 160, 163, 53 },
		{ 160, 164, 53 },
		{ 160, 165, 53 },
		{ 160, 166, 53 },
		{ 160, 167, 53 },
		{ 160, 168, 53 },
		{ 160, 169, 53 },
		{ 160, 170, 53 },
		{ 160, 171, 53 },
		{ 160, 172, 53 },
		{ 160, 173, 53 },
		{ 160, 174, 53 },
		{ 160, 175, 53 },
		{ 160, 176, 53 },
		{ 160, 177, 53 },
		{ 160, 178, 53 },
		{ 160, 179, 53 },
		{ 160, 180, 53 },
		{ 160, 181, 53 },
		{ 160, 182, 53 },
		{ 160, 183, 53 },
		{ 160, 184, 53 },
		{ 161, 162, 53 },
		{ 161, 163, 53 },
		{ 161, 164, 53 },
		{ 161, 165, 53 },
		{ 161, 166, 53 },
		{ 161, 167, 53 },
		{ 161, 168, 53 },
		{ 161, 169, 53 },
		{ 161, 170, 53 },
		{ 161, 171, 53 },
		{ 161, 172, 53 },
		{ 161, 173, 53 },
		{ 161, 174, 53 },
		{ 161, 175, 53 },
		{ 161, 176, 53 },
		{ 161, 177, 53 },
		{ 161, 178, 53 },
		{ 161, 179, 53 },
		{ 161, 180, 53 },
		{ 161, 181, 53 },
		{ 161, 182, 53 },
		{ 161, 183, 53 },
		{ 161, 184, 53 },
		{ 162, 163, 53 },
		{ 162, 164, 53 },
		{ 162, 165, 53 },
		{ 162, 166, 53 },
		{ 162, 167, 53 },
		{ 162, 168, 53 },
		{ 162, 169, 53 },
		{ 162, 170, 53 },
		{ 162, 171, 53 },
		{ 162, 172, 53 },
		{ 162, 173, 53 },
		{ 162, 174, 53 },
		{ 162, 175, 53 },
		{ 162, 176, 53 },
		{ 162, 177, 53 },
		{ 162, 178, 53 },
		{ 162, 179, 53 },
		{ 162, 180, 53 },
		{ 162, 181, 53 },
		{ 162, 182, 53 },
		{ 162, 183, 53 },
		{ 162, 184, 53 },
		{ 162, 185, 53 },
		{ 163, 164, 53 },
		{ 163, 165, 53 },
		{ 163, 166, 53 },
		{ 163, 167, 53 },
		{ 163, 168, 53 },
		{ 163, 169, 53 },
		{ 163, 170, 53 },
		{ 163, 171, 53 },
		{ 163, 172, 53 },
		{ 163, 173, 53 },
		{ 163, 174, 53 },
		{ 163, 175, 53 },
		{ 163, 176, 53 },
		{ 163, 177, 53 },
		{ 163, 178, 53 },
		{ 163, 179, 53 },
		{ 163, 180, 53 },
		{ 163, 181, 53 },
		{ 163, 182, 53 },
		{ 163, 183, 53 },
		{ 163, 184, 53 },
		{ 163, 185, 53 },
		{ 163, 186, 53 },
		{ 164, 165, 53 },
		{ 164, 166, 53 },
		{ 164, 167, 53 },
		{ 164, 168, 53 },
		{ 164, 169, 53 },
		{ 164, 170, 53 },
		{ 164, 171, 53 },
		{ 164, 172, 53 },
		{ 164, 173, 53 },
		{ 164, 174, 53 },
		{ 164, 175, 53 },
		{ 164, 176, 53 },
		{ 164, 177, 53 },
		{ 164, 178, 53 },
		{ 164, 179, 53 },
		{ 164, 180, 53 },
		{ 164, 181, 53 },
		{ 164, 182, 53 },
		{ 164, 183, 53 },
		{ 164, 184, 53 },
		{ 164, 185, 53 },
		{ 164, 186, 53 },
		{ 164, 187, 53 },
		{ 165, 166, 53 },
		{ 165, 167, 53 },
		{ 165, 168, 53 },
		{ 165, 169, 53 },
		{ 165, 170, 53 },
		{ 165, 171, 53 },
		{ 165, 172, 53 },
		{ 165, 173, 53 },
		{ 165, 174, 53 },
		{ 165, 175, 53 },
		{ 165, 176, 53 },
		{ 165, 177, 53 },
		{ 165, 178, 53 },
		{ 165, 179, 53 },
		{ 165, 180, 53 },
		{ 165, 181, 53 },
		{ 165, 182, 53 },
		{ 165, 183, 53 },
		{ 165, 184, 53 },
		{ 165, 185, 53 },
		{ 165, 186, 53 },
		{ 165, 187, 53 },
		{ 166, 167, 53 },
		{ 166, 168, 53 },
		{ 166, 169, 53 },
		{ 166, 170, 53 },
		{ 166, 171, 53 },
		{ 166, 172, 53 },
		{ 166, 173, 53 },
		{ 166, 174, 53 },
		{ 166, 175, 53 },
		{ 166, 176, 53 },
		{ 166, 177, 53 },
		{ 166, 178, 53 },
		{ 166, 179, 53 },
		{ 166, 180, 53 },
		{ 166, 181, 53 },
		{ 166, 182, 53 },
		{ 166, 183, 53 },
		{ 166, 184, 53 },
		{ 166, 185, 53 },
		{ 166, 186, 53 },
		{ 166, 187, 53 },
		{ 166, 188, 53 },
		{ 167, 168, 53 },
		{ 167, 169, 53 },
		{ 167, 170, 53 },
		{ 167, 171, 53 },
		{ 167, 172, 53 },
		{ 167, 173, 53 },
		{ 167, 174, 53 },
		{ 167, 175, 53 },
		{ 167, 176, 53 },
		{ 167, 177, 53 },
		{ 167, 178, 53 },
		{ 167, 179, 53 },
		{ 167, 180, 53 },
		{ 167, 181, 53 },
		{ 167, 182, 53 },
		{ 167, 183, 53 },
		{ 167, 184, 53 },
		{ 167, 185, 53 },
		{ 167, 186, 53 },
		{ 167, 187, 53 },
		{ 167, 188, 53 },
		{ 167, 189, 53 },
		{ 168, 169, 53 },
		{ 168, 170, 53 },
		{ 168, 171, 53 },
		{ 168, 172, 53 },
		{ 168, 173, 53 },
		{ 168, 174, 53 },
		{ 168, 175, 53 },
		{ 168, 176, 53 },
		{ 168, 177, 53 },
		{ 168, 178, 53 },
		{ 168, 179, 53 },
		{ 168, 180, 53 },
		{ 168, 181, 53 },
		{ 168, 182, 53 },
		{ 168, 183, 53 },
		{ 168, 184, 53 },
		{ 168, 185, 53 },
		{ 168, 186, 53 },
		{ 168, 187, 53 },
		{ 168, 188, 53 },
		{ 168, 189, 53 },
		{ 168, 190, 53 },
		{ 169, 170, 53 },
		{ 169, 171, 53 },
		{ 169, 172, 53 },
		{ 169, 173, 53 },
		{ 169, 174, 53 },
		{ 169, 175, 53 },
		{ 169, 176, 53 },
		{ 169, 177, 53 },
		{ 169, 178, 53 },
		{ 169, 179, 53 },
		{ 169, 180, 53 },
		{ 169, 181, 53 },
		{ 169, 182, 53 },
		{ 169, 183, 53 },
		{ 169, 184, 53 },
		{ 169, 185, 53 },
		{ 169, 186, 53 },
		{ 169, 187, 53 },
		{ 169, 188, 53 },
		{ 169, 189, 53 },
		{ 169, 190, 53 },
		{ 170, 171, 53 },
		{ 170, 172, 53 },
		{ 170, 173, 53 },
		{ 170, 174, 53 },
		{ 170, 175, 53 },
		{ 170, 176, 53 },
		{ 170, 177, 53 },
		{ 170, 178, 53 },
		{ 170, 179, 53 },
		{ 170, 180, 53 },
		{ 170, 181, 53 },
		{ 170, 182, 53 },
		{ 170, 183, 53 },
		{ 170, 184, 53 },
		{ 170, 185, 53 },
		{ 170, 186, 53 },
		{ 170, 187, 53 },
		{ 170, 188, 53 },
		{ 170, 189, 53 },
		{ 170, 190, 53 },
		{ 170, 191, 53 },
		{ 171, 172, 53 },
		{ 171, 173, 53 },
		{ 171, 174, 53 },
		{ 171, 175, 53 },
		{ 171, 176, 53 },
		{ 171, 177, 53 },
		{ 171, 178, 53 },
		{ 171, 179, 53 },
		{ 171, 180, 53 },
		{ 171, 181, 53 },
		{ 171, 182, 53 },
		{ 171, 183, 53 },
		{ 171, 184, 53 },
		{ 171, 185, 53 },
		{ 171, 186, 53 },
		{ 171, 187, 53 },
		{ 171, 188, 53 },
		{ 171, 189, 53 },
		{ 171, 190, 53 },
		{ 171, 191, 53 },
		{ 171, 192, 53 },
		{ 172, 173, 53 },
		{ 172, 174, 53 },
		{ 172, 175, 53 },
		{ 172, 176, 53 },
		{ 172, 177, 53 },
		{ 172, 178, 53 },
		{ 172, 179, 53 },
		{ 172, 180, 53 },
		{ 172, 181, 53 },
		{ 172, 182, 53 },
		{ 172, 183, 53 },
		{ 172, 184, 53 },
		{ 172, 185, 53 },
		{ 172, 186, 53 },
		{ 172, 187, 53 },
		{ 172, 188, 53 },
		{ 172, 189, 53 },
		{ 172, 190, 53 },
		{ 172, 191, 53 },
		{ 172, 192, 53 },
		{ 172, 193, 53 },
		{ 173, 174, 53 },
		{ 173, 175, 53 },
		{ 173, 176, 53 },
		{ 173, 177, 53 },
		{ 173, 178, 53 },
		{ 173, 179, 53 },
		{ 173, 180, 53 },
		{ 173, 181, 53 },
		{ 173, 182, 53 },
		{ 173, 183, 53 },
		{ 173, 184, 53 },
		{ 173, 185, 53 },
		{ 173, 186, 53 },
		{ 173, 187, 53 },
		{ 173, 188, 53 },
		{ 173, 189, 53 },
		{ 173, 190, 53 },
		{ 173, 191, 53 },
		{ 173, 192, 53 },
		{ 173, 193, 53 },
		{ 174, 175, 53 },
		{ 174, 176, 53 },
		{ 174, 177, 53 },
		{ 174, 178, 53 },
		{ 174, 179, 53 },
		{ 174, 180, 53 },
		{ 174, 181, 53 },
		{ 174, 182, 53 },
		{ 174, 183, 53 },
		{ 174, 184, 53 },
		{ 174, 185, 53 },
		{ 174, 186, 53 },
		{ 174, 187, 53 },
		{ 174, 188, 53 },
		{ 174, 189, 53 },
		{ 174, 190, 53 },
		{ 174, 191, 53 },
		{ 174, 192, 53 },
		{ 174, 193, 53 },
		{ 174, 194, 53 },
		{ 175, 176, 53 },
		{ 175, 177, 53 },
		{ 175, 178, 53 },
		{ 175, 179, 53 },
		{ 175, 180, 53 },
		{ 175, 181, 53 },
		{ 175, 182, 53 },
		{ 175, 183, 53 },
		{ 175, 184, 53 },
		{ 175, 185, 53 },
		{ 175, 186, 53 },
		{ 175, 187, 53 },
		{ 175, 188, 53 },
		{ 175, 189, 53 },
		{ 175, 190, 53 },
		{ 175, 191, 53 },
		{ 175, 192, 53 },
		{ 175, 193, 53 },
		{ 175, 194, 53 },
		{ 175, 195, 53 },
		{ 176, 177, 53 },
		{ 176, 178, 53 },
		{ 176, 179, 53 },
		{ 176, 180, 53 },
		{ 176, 181, 53 },
		{ 176, 182, 53 },
		{ 176, 183, 53 },
		{ 176, 184, 53 },
		{ 176, 185, 53 },
		{ 176, 186, 53 },
		{ 176, 187, 53 },
		{ 176, 188, 53 },
		{ 176, 189, 53 },
		{ 176, 190, 53 },
		{ 176, 191, 53 },
		{ 176, 192, 53 },
		{ 176, 193, 53 },
		{ 176, 194, 53 },
		{ 176, 195, 53 },
		{ 176, 196, 53 },
		{ 177, 178, 53 },
		{ 177, 179, 53 },
		{ 177, 180, 53 },
		{ 177, 181, 53 },
		{ 177, 182, 53 },
		{ 177, 183, 53 },
		{ 177, 184, 53 },
		{ 177, 185, 53 },
		{ 177, 186, 53 },
		{ 177, 187, 53 },
		{ 177, 188, 53 },
		{ 177, 189, 53 },
		{ 177, 190, 53 },
		{ 177, 191, 53 },
		{ 177, 192, 53 },
		{ 177, 193, 53 },
		{ 177, 194, 53 },
		{ 177, 195, 53 },
		{ 177, 196, 53 },
		{ 178, 179, 53 },
		{ 178, 180, 53 },
		{ 178, 181, 53 },
		{ 178, 182, 53 },
		{ 178, 183, 53 },
		{ 178, 184, 53 },
		{ 178, 185, 53 },
		{ 178, 186, 53 },
		{ 178, 187, 53 },
		{ 178, 188, 53 },
		{ 178, 189, 53 },
		{ 178, 190, 53 },
		{ 178, 191, 53 },
		{ 178, 192, 53 },
		{ 178, 193, 53 },
		{ 178, 194, 53 },
		{ 178, 195, 53 },
		{ 178, 196, 53 },
		{ 178, 197, 53 },
		{ 179, 180, 53 },
		{ 179, 181, 53 },
		{ 179, 182, 53 },
		{ 179, 183, 53 },
		{ 179, 184, 53 },
		{ 179, 185, 53 },
		{ 179, 186, 53 },
		{ 179, 187, 53 },
		{ 179, 188, 53 },
		{ 179, 189, 53 },
		{ 179, 190, 53 },
		{ 179, 191, 53 },
		{ 179, 192, 53 },
		{ 179, 193, 53 },
		{ 179, 194, 53 },
		{ 179, 195, 53 },
		{ 179, 196, 53 },
		{ 179, 197, 53 },
		{ 179, 198, 53 },
		{ 180, 181, 53 },
		{ 180, 182, 53 },
		{ 180, 183, 53 },
		{ 180, 184, 53 },
		{ 180, 185, 53 },
		{ 180, 186, 53 },
		{ 180, 187, 53 },
		{ 180, 188, 53 },
		{ 180, 189, 53 },
		{ 180, 190, 53 },
		{ 180, 191, 53 },
		{ 180, 192, 53 },
		{ 180, 193, 53 },
		{ 180, 194, 53 },
		{ 180, 195, 53 },
		{ 180, 196, 53 },
		{ 180, 197, 53 },
		{ 180, 198, 53 },
		{ 180, 199, 53 },
		{ 181, 182, 53 },
		{ 181, 183, 53 },
		{ 181, 184, 53 },
		{ 181, 185, 53 },
		{ 181, 186, 53 },
		{ 181, 187, 53 },
		{ 181, 188, 53 },
		{ 181, 189, 53 },
		{ 181, 190, 53 },
		{ 181, 191, 53 },
		{ 181, 192, 53 },
		{ 181, 193, 53 },
		{ 181, 194, 53 },
		{ 181, 195, 53 },
		{ 181, 196, 53 },
		{ 181, 197, 53 },
		{ 181, 198, 53 },
		{ 181, 199, 53 },
		{ 182, 183, 53 },
		{ 182, 184, 53 },
		{ 182, 185, 53 },
		{ 182, 186, 53 },
		{ 182, 187, 53 },
		{ 182, 188, 53 },
		{ 182, 189, 53 },
		{ 182, 190, 53 },
		{ 182, 191, 53 },
		{ 182, 192, 53 },
		{ 182, 193, 53 },
		{ 182, 194, 53 },
		{ 182, 195, 53 },
		{ 182, 196, 53 },
		{ 182, 197, 53 },
		{ 182, 198, 53 },
		{ 182, 199, 53 },
		{ 182, 200, 53 },
		{ 183, 184, 53 },
		{ 183, 185, 53 },
		{ 183, 186, 53 },
		{ 183, 187, 53 },
		{ 183, 188, 53 },
		{ 183, 189, 53 },
		{ 183, 190, 53 },
		{ 183, 191, 53 },
		{ 183, 192, 53 },
		{ 183, 193, 53 },
		{ 183, 194, 53 },
		{ 183, 195, 53 },
		{ 183, 196, 53 },
		{ 183, 197, 53 },
		{ 183, 198, 53 },
		{ 183, 199, 53 },
		{ 183, 200, 53 },
		{ 183, 201, 53 },
		{ 184, 185, 53 },
		{ 184, 186, 53 },
		{ 184, 187, 53 },
		{ 184, 188, 53 },
		{ 184, 189, 53 },
		{ 184, 190, 53 },
		{ 184, 191, 53 },
		{ 184, 192, 53 },
		{ 184, 193, 53 },
		{ 184, 194, 53 },
		{ 184, 195, 53 },
		{ 184, 196, 53 },
		{ 184, 197, 53 },
		{ 184, 198, 53 },
		{ 184, 199, 53 },
		{ 184, 200, 53 },
		{ 184, 201, 53 },
		{ 184, 202, 53 },
		{ 185, 186, 53 },
		{ 185, 187, 53 },
		{ 185, 188, 53 },
		{ 185, 189, 53 },
		{ 185, 190, 53 },
		{ 185, 191, 53 },
		{ 185, 192, 53 },
		{ 185, 193, 53 },
		{ 185, 194, 53 },
		{ 185, 195, 53 },
		{ 185, 196, 53 },
		{ 185, 197, 53 },
		{ 185, 198, 53 },
		{ 185, 199, 53 },
		{ 185, 200, 53 },
		{ 185, 201, 53 },
		{ 185, 202, 53 },
		{ 186, 187, 53 },
		{ 186, 188, 53 },
		{ 186, 189, 53 },
		{ 186, 190, 53 },
		{ 186, 191, 53 },
		{ 186, 192, 53 },
		{ 186, 193, 53 },
		{ 186, 194, 53 },
		{ 186, 195, 53 },
		{ 186, 196, 53 },
		{ 186, 197, 53 },
		{ 186, 198, 53 },
		{ 186, 199, 53 },
		{ 186, 200, 53 },
		{ 186, 201, 53 },
		{ 186, 202, 53 },
		{ 186, 203, 53 },
		{ 187, 188, 53 },
		{ 187, 189, 53 },
		{ 187, 190, 53 },
		{ 187, 191, 53 },
		{ 187, 192, 53 },
		{ 187, 193, 53 },
		{ 187, 194, 53 },
		{ 187, 195, 53 },
		{ 187, 196, 53 },
		{ 187, 197, 53 },
		{ 187, 198, 53 },
		{ 187, 199, 53 },
		{ 187, 200, 53 },
		{ 187, 201, 53 },
		{ 187, 202, 53 },
		{ 187, 203, 53 },
		{ 187, 204, 53 },
		{ 188, 189, 53 },
		{ 188, 190, 53 },
		{ 188, 191, 53 },
		{ 188, 192, 53 },
		{ 188, 193, 53 },
		{ 188, 194, 53 },
		{ 188, 195, 53 },
		{ 188, 196, 53 },
		{ 188, 197, 53 },
		{ 188, 198, 53 },
		{ 188, 199, 53 },
		{ 188, 200, 53 },
		{ 188, 201, 53 },
		{ 188, 202, 53 },
		{ 188, 203, 53 },
		{ 188, 204, 53 },
		{ 189, 190, 53 },
		{ 189, 191, 53 },
		{ 189, 192, 53 },
		{ 189, 193, 53 },
		{ 189, 194, 53 },
		{ 189, 195, 53 },
		{ 189, 196, 53 },
		{ 189, 197, 53 },
		{ 189, 198, 53 },
		{ 189, 199, 53 },
		{ 189, 200, 53 },
		{ 189, 201, 53 },
		{ 189, 202, 53 },
		{ 189, 203, 53 },
		{ 189, 204, 53 },
		{ 189, 205, 53 },
		{ 190, 191, 53 },
		{ 190, 192, 53 },
		{ 190, 193, 53 },
		{ 190, 194, 53 },
		{ 190, 195, 53 },
		{ 190, 196, 53 },
		{ 190, 197, 53 },
		{ 190, 198, 53 },
		{ 190, 199, 53 },
		{ 190, 200, 53 },
		{ 190, 201, 53 },
		{ 190, 202, 53 },
		{ 190, 203, 53 },
		{ 190, 204, 53 },
		{ 190, 205, 53 },
		{ 190, 206, 53 },
		{ 191, 193, 53 },
		{ 191, 194, 53 },
		{ 191, 195, 53 },
		{ 191, 196, 53 },
		{ 191, 197, 53 },
		{ 191, 198, 53 },
		{ 191, 199, 53 },
		{ 191, 200, 53 },
		{ 191, 201, 53 },
		{ 191, 202, 53 },
		{ 191, 203, 53 },
		{ 191, 204, 53 },
		{ 191, 205, 53 },
		{ 323, 309, 53 },
		{ 323, 310, 53 },
		{ 323, 311, 53 },
		{ 323, 312, 53 },
		{ 323, 313, 53 },
		{ 323, 314, 53 },
		{ 323, 315, 53 },
		{ 323, 316, 53 },
		{ 323, 317, 53 },
		{ 323, 318, 53 },
		{ 323, 319, 53 },
		{ 323, 320, 53 },
		{ 323, 321, 53 },
		{ 324, 308, 53 },
		{ 324, 309, 53 },
		{ 324, 310, 53 },
		{ 324, 311, 53 },
		{ 324, 312, 53 },
		{ 324, 313, 53 },
		{ 324, 314, 53 },
		{ 324, 315, 53 },
		{ 324, 316, 53 },
		{ 324, 317, 53 },
		{ 324, 318, 53 },
		{ 324, 319, 53 },
		{ 324, 320, 53 },
		{ 324, 321, 53 },
		{ 324, 322, 53 },
		{ 324, 323, 53 },
		{ 325, 309, 53 },
		{ 325, 310, 53 },
		{ 325, 311, 53 },
		{ 325, 312, 53 },
		{ 325, 313, 53 },
		{ 325, 314, 53 },
		{ 325, 315, 53 },
		{ 325, 316, 53 },
		{ 325, 317, 53 },
		{ 325, 318, 53 },
		{ 325, 319, 53 },
		{ 325, 320, 53 },
		{ 325, 321, 53 },
		{ 325, 322, 53 },
		{ 325, 323, 53 },
		{ 325, 324, 53 },
		{ 326, 310, 53 },
		{ 326, 311, 53 },
		{ 326, 312, 53 },
		{ 326, 313, 53 },
		{ 326, 314, 53 },
		{ 326, 315, 53 },
		{ 326, 316, 53 },
		{ 326, 317, 53 },
		{ 326, 318, 53 },
		{ 326, 319, 53 },
		{ 326, 320, 53 },
		{ 326, 321, 53 },
		{ 326, 322, 53 },
		{ 326, 323, 53 },
		{ 326, 324, 53 },
		{ 326, 325, 53 },
		{ 327, 310, 53 },
		{ 327, 311, 53 },
		{ 327, 312, 53 },
		{ 327, 313, 53 },
		{ 327, 314, 53 },
		{ 327, 315, 53 },
		{ 327, 316, 53 },
		{ 327, 317, 53 },
		{ 327, 318, 53 },
		{ 327, 319, 53 },
		{ 327, 320, 53 },
		{ 327, 321, 53 },
		{ 327, 322, 53 },
		{ 327, 323, 53 },
		{ 327, 324, 53 },
		{ 327, 325, 53 },
		{ 327, 326, 53 },
		{ 328, 311, 53 },
		{ 328, 312, 53 },
		{ 328, 313, 53 },
		{ 328, 314, 53 },
		{ 328, 315, 53 },
		{ 328, 316, 53 },
		{ 328, 317, 53 },
		{ 328, 318, 53 },
		{ 328, 319, 53 },
		{ 328, 320, 53 },
		{ 328, 321, 53 },
		{ 328, 322, 53 },
		{ 328, 323, 53 },
		{ 328, 324, 53 },
		{ 328, 325, 53 },
		{ 328, 326, 53 },
		{ 328, 327, 53 },
		{ 329, 312, 53 },
		{ 329, 313, 53 },
		{ 329, 314, 53 },
		{ 329, 315, 53 },
		{ 329, 316, 53 },
		{ 329, 317, 53 },
		{ 329, 318, 53 },
		{ 329, 319, 53 },
		{ 329, 320, 53 },
		{ 329, 321, 53 },
		{ 329, 322, 53 },
		{ 329, 323, 53 },
		{ 329, 324, 53 },
		{ 329, 325, 53 },
		{ 329, 326, 53 },
		{ 329, 327, 53 },
		{ 329, 328, 53 },
		{ 330, 312, 53 },
		{ 330, 313, 53 },
		{ 330, 314, 53 },
		{ 330, 315, 53 },
		{ 330, 316, 53 },
		{ 330, 317, 53 },
		{ 330, 318, 53 },
		{ 330, 319, 53 },
		{ 330, 320, 53 },
		{ 330, 321, 53 },
		{ 330, 322, 53 },
		{ 330, 323, 53 },
		{ 330, 324, 53 },
		{ 330, 325, 53 },
		{ 330, 326, 53 },
		{ 330, 327, 53 },
		{ 330, 328, 53 },
		{ 330, 329, 53 },
		{ 331, 313, 53 },
		{ 331, 314, 53 },
		{ 331, 315, 53 },
		{ 331, 316, 53 },
		{ 331, 317, 53 },
		{ 331, 318, 53 },
		{ 331, 319, 53 },
		{ 331, 320, 53 },
		{ 331, 321, 53 },
		{ 331, 322, 53 },
		{ 331, 323, 53 },
		{ 331, 324, 53 },
		{ 331, 325, 53 },
		{ 331, 326, 53 },
		{ 331, 327, 53 },
		{ 331, 328, 53 },
		{ 331, 329, 53 },
		{ 331, 330, 53 },
		{ 332, 314, 53 },
		{ 332, 315, 53 },
		{ 332, 316, 53 },
		{ 332, 317, 53 },
		{ 332, 318, 53 },
		{ 332, 319, 53 },
		{ 332, 320, 53 },
		{ 332, 321, 53 },
		{ 332, 322, 53 },
		{ 332, 323, 53 },
		{ 332, 324, 53 },
		{ 332, 325, 53 },
		{ 332, 326, 53 },
		{ 332, 327, 53 },
		{ 332, 328, 53 },
		{ 332, 329, 53 },
		{ 332, 330, 53 },
		{ 332, 331, 53 },
		{ 333, 315, 53 },
		{ 333, 316, 53 },
		{ 333, 317, 53 },
		{ 333, 318, 53 },
		{ 333, 319, 53 },
		{ 333, 320, 53 },
		{ 333, 321, 53 },
		{ 333, 322, 53 },
		{ 333, 323, 53 },
		{ 333, 324, 53 },
		{ 333, 325, 53 },
		{ 333, 326, 53 },
		{ 333, 327, 53 },
		{ 333, 328, 53 },
		{ 333, 329, 53 },
		{ 333, 330, 53 },
		{ 333, 331, 53 },
		{ 333, 332, 53 },
		{ 334, 315, 53 },
		{ 334, 316, 53 },
		{ 334, 317, 53 },
		{ 334, 318, 53 },
		{ 334, 319, 53 },
		{ 334, 320, 53 },
		{ 334, 321, 53 },
		{ 334, 322, 53 },
		{ 334, 323, 53 },
		{ 334, 324, 53 },
		{ 334, 325, 53 },
		{ 334, 326, 53 },
		{ 334, 327, 53 },
		{ 334, 328, 53 },
		{ 334, 329, 53 },
		{ 334, 330, 53 },
		{ 334, 331, 53 },
		{ 334, 332, 53 },
		{ 334, 333, 53 },
		{ 335, 316, 53 },
		{ 335, 317, 53 },
		{ 335, 318, 53 },
		{ 335, 319, 53 },
		{ 335, 320, 53 },
		{ 335, 321, 53 },
		{ 335, 322, 53 },
		{ 335, 323, 53 },
		{ 335, 324, 53 },
		{ 335, 325, 53 },
		{ 335, 326, 53 },
		{ 335, 327, 53 },
		{ 335, 328, 53 },
		{ 335, 329, 53 },
		{ 335, 330, 53 },
		{ 335, 331, 53 },
		{ 335, 332, 53 },
		{ 335, 333, 53 },
		{ 335, 334, 53 },
		{ 336, 317, 53 },
		{ 336, 318, 53 },
		{ 336, 319, 53 },
		{ 336, 320, 53 },
		{ 336, 321, 53 },
		{ 336, 322, 53 },
		{ 336, 323, 53 },
		{ 336, 324, 53 },
		{ 336, 325, 53 },
		{ 336, 326, 53 },
		{ 336, 327, 53 },
		{ 336, 328, 53 },
		{ 336, 329, 53 },
		{ 336, 330, 53 },
		{ 336, 331, 53 },
		{ 336, 332, 53 },
		{ 336, 333, 53 },
		{ 336, 334, 53 },
		{ 336, 335, 53 },
		{ 337, 318, 53 },
		{ 337, 319, 53 },
		{ 337, 320, 53 },
		{ 337, 321, 53 },
		{ 337, 322, 53 },
		{ 337, 323, 53 },
		{ 337, 324, 53 },
		{ 337, 325, 53 },
		{ 337, 326, 53 },
		{ 337, 327, 53 },
		{ 337, 328, 53 },
		{ 337, 329, 53 },
		{ 337, 330, 53 },
		{ 337, 331, 53 },
		{ 337, 332, 53 },
		{ 337, 333, 53 },
		{ 337, 334, 53 },
		{ 337, 335, 53 },
		{ 337, 336, 53 },
		{ 338, 318, 53 },
		{ 338, 319, 53 },
		{ 338, 320, 53 },
		{ 338, 321, 53 },
		{ 338, 322, 53 },
		{ 338, 323, 53 },
		{ 338, 324, 53 },
		{ 338, 325, 53 },
		{ 338, 326, 53 },
		{ 338, 327, 53 },
		{ 338, 328, 53 },
		{ 338, 329, 53 },
		{ 338, 330, 53 },
		{ 338, 331, 53 },
		{ 338, 332, 53 },
		{ 338, 333, 53 },
		{ 338, 334, 53 },
		{ 338, 335, 53 },
		{ 338, 336, 53 },
		{ 338, 337, 53 },
		{ 339, 319, 53 },
		{ 339, 320, 53 },
		{ 339, 321, 53 },
		{ 339, 322, 53 },
		{ 339, 323, 53 },
		{ 339, 324, 53 },
		{ 339, 325, 53 },
		{ 339, 326, 53 },
		{ 339, 327, 53 },
		{ 339, 328, 53 },
		{ 339, 329, 53 },
		{ 339, 330, 53 },
		{ 339, 331, 53 },
		{ 339, 332, 53 },
		{ 339, 333, 53 },
		{ 339, 334, 53 },
		{ 339, 335, 53 },
		{ 339, 336, 53 },
		{ 339, 337, 53 },
		{ 339, 338, 53 },
		{ 340, 320, 53 },
		{ 340, 321, 53 },
		{ 340, 322, 53 },
		{ 340, 323, 53 },
		{ 340, 324, 53 },
		{ 340, 325, 53 },
		{ 340, 326, 53 },
		{ 340, 327, 53 },
		{ 340, 328, 53 },
		{ 340, 329, 53 },
		{ 340, 330, 53 },
		{ 340, 331, 53 },
		{ 340, 332, 53 },
		{ 340, 333, 53 },
		{ 340, 334, 53 },
		{ 340, 335, 53 },
		{ 340, 336, 53 },
		{ 340, 337, 53 },
		{ 340, 338, 53 },
		{ 340, 339, 53 },
		{ 341, 321, 53 },
		{ 341, 322, 53 },
		{ 341, 323, 53 },
		{ 341, 324, 53 },
		{ 341, 325, 53 },
		{ 341, 326, 53 },
		{ 341, 327, 53 },
		{ 341, 328, 53 },
		{ 341, 329, 53 },
		{ 341, 330, 53 },
		{ 341, 331, 53 },
		{ 341, 332, 53 },
		{ 341, 333, 53 },
		{ 341, 334, 53 },
		{ 341, 335, 53 },
		{ 341, 336, 53 },
		{ 341, 337, 53 },
		{ 341, 338, 53 },
		{ 341, 339, 53 },
		{ 341, 340, 53 },
		{ 342, 321, 53 },
		{ 342, 322, 53 },
		{ 342, 323, 53 },
		{ 342, 324, 53 },
		{ 342, 325, 53 },
		{ 342, 326, 53 },
		{ 342, 327, 53 },
		{ 342, 328, 53 },
		{ 342, 329, 53 },
		{ 342, 330, 53 },
		{ 342, 331, 53 },
		{ 342, 332, 53 },
		{ 342, 333, 53 },
		{ 342, 334, 53 },
		{ 342, 335, 53 },
		{ 342, 336, 53 },
		{ 342, 337, 53 },
		{ 342, 338, 53 },
		{ 342, 339, 53 },
		{ 342, 340, 53 },
		{ 342, 341, 53 },
		{ 343, 322, 53 },
		{ 343, 323, 53 },
		{ 343, 324, 53 },
		{ 343, 325, 53 },
		{ 343, 326, 53 },
		{ 343, 327, 53 },
		{ 343, 328, 53 },
		{ 343, 329, 53 },
		{ 343, 330, 53 },
		{ 343, 331, 53 },
		{ 343, 332, 53 },
		{ 343, 333, 53 },
		{ 343, 334, 53 },
		{ 343, 335, 53 },
		{ 343, 336, 53 },
		{ 343, 337, 53 },
		{ 343, 338, 53 },
		{ 343, 339, 53 },
		{ 343, 340, 53 },
		{ 343, 341, 53 },
		{ 343, 342, 53 },
		{ 344, 323, 53 },
		{ 344, 324, 53 },
		{ 344, 325, 53 },
		{ 344, 326, 53 },
		{ 344, 327, 53 },
		{ 344, 328, 53 },
		{ 344, 329, 53 },
		{ 344, 330, 53 },
		{ 344, 331, 53 },
		{ 344, 332, 53 },
		{ 344, 333, 53 },
		{ 344, 334, 53 },
		{ 344, 335, 53 },
		{ 344, 336, 53 },
		{ 344, 337, 53 },
		{ 344, 338, 53 },
		{ 344, 339, 53 },
		{ 344, 340, 53 },
		{ 344, 341, 53 },
		{ 344, 342, 53 },
		{ 344, 343, 53 },
		{ 345, 324, 53 },
		{ 345, 325, 53 },
		{ 345, 326, 53 },
		{ 345, 327, 53 },
		{ 345, 328, 53 },
		{ 345, 329, 53 },
		{ 345, 330, 53 },
		{ 345, 331, 53 },
		{ 345, 332, 53 },
		{ 345, 333, 53 },
		{ 345, 334, 53 },
		{ 345, 335, 53 },
		{ 345, 336, 53 },
		{ 345, 337, 53 },
		{ 345, 338, 53 },
		{ 345, 339, 53 },
		{ 345, 340, 53 },
		{ 345, 341, 53 },
		{ 345, 342, 53 },
		{ 345, 343, 53 },
		{ 345, 344, 53 },
		{ 346, 324, 53 },
		{ 346, 325, 53 },
		{ 346, 326, 53 },
		{ 346, 327, 53 },
		{ 346, 328, 53 },
		{ 346, 329, 53 },
		{ 346, 330, 53 },
		{ 346, 331, 53 },
		{ 346, 332, 53 },
		{ 346, 333, 53 },
		{ 346, 334, 53 },
		{ 346, 335, 53 },
		{ 346, 336, 53 },
		{ 346, 337, 53 },
		{ 346, 338, 53 },
		{ 346, 339, 53 },
		{ 346, 340, 53 },
		{ 346, 341, 53 },
		{ 346, 342, 53 },
		{ 346, 343, 53 },
		{ 346, 344, 53 },
		{ 346, 345, 53 },
		{ 347, 325, 53 },
		{ 347, 326, 53 },
		{ 347, 327, 53 },
		{ 347, 328, 53 },
		{ 347, 329, 53 },
		{ 347, 330, 53 },
		{ 347, 331, 53 },
		{ 347, 332, 53 },
		{ 347, 333, 53 },
		{ 347, 334, 53 },
		{ 347, 335, 53 },
		{ 347, 336, 53 },
		{ 347, 337, 53 },
		{ 347, 338, 53 },
		{ 347, 339, 53 },
		{ 347, 340, 53 },
		{ 347, 341, 53 },
		{ 347, 342, 53 },
		{ 347, 343, 53 },
		{ 347, 344, 53 },
		{ 347, 345, 53 },
		{ 347, 346, 53 },
		{ 348, 326, 53 },
		{ 348, 327, 53 },
		{ 348, 328, 53 },
		{ 348, 329, 53 },
		{ 348, 330, 53 },
		{ 348, 331, 53 },
		{ 348, 332, 53 },
		{ 348, 333, 53 },
		{ 348, 334, 53 },
		{ 348, 335, 53 },
		{ 348, 336, 53 },
		{ 348, 337, 53 },
		{ 348, 338, 53 },
		{ 348, 339, 53 },
		{ 348, 340, 53 },
		{ 348, 341, 53 },
		{ 348, 342, 53 },
		{ 348, 343, 53 },
		{ 348, 344, 53 },
		{ 348, 345, 53 },
		{ 348, 346, 53 },
		{ 348, 347, 53 },
		{ 349, 327, 53 },
		{ 349, 328, 53 },
		{ 349, 329, 53 },
		{ 349, 330, 53 },
		{ 349, 331, 53 },
		{ 349, 332, 53 },
		{ 349, 333, 53 },
		{ 349, 334, 53 },
		{ 349, 335, 53 },
		{ 349, 336, 53 },
		{ 349, 337, 53 },
		{ 349, 338, 53 },
		{ 349, 339, 53 },
		{ 349, 340, 53 },
		{ 349, 341, 53 },
		{ 349, 342, 53 },
		{ 349, 343, 53 },
		{ 349, 344, 53 },
		{ 349, 345, 53 },
		{ 349, 346, 53 },
		{ 349, 347, 53 },
		{ 349, 348, 53 },
		{ 350, 327, 53 },
		{ 350, 328, 53 },
		{ 350, 329, 53 },
		{ 350, 330, 53 },
		{ 350, 331, 53 },
		{ 350, 332, 53 },
		{ 350, 333, 53 },
		{ 350, 334, 53 },
		{ 350, 335, 53 },
		{ 350, 336, 53 },
		{ 350, 337, 53 },
		{ 350, 338, 53 },
		{ 350, 339, 53 },
		{ 350, 340, 53 },
		{ 350, 341, 53 },
		{ 350, 342, 53 },
		{ 350, 343, 53 },
		{ 350, 344, 53 },
		{ 350, 345, 53 },
		{ 350, 346, 53 },
		{ 350, 347, 53 },
		{ 350, 348, 53 },
		{ 350, 349, 53 },
		{ 351, 328, 53 },
		{ 351, 329, 53 },
		{ 351, 330, 53 },
		{ 351, 331, 53 },
		{ 351, 332, 53 },
		{ 351, 333, 53 },
		{ 351, 334, 53 },
		{ 351, 335, 53 },
		{ 351, 336, 53 },
		{ 351, 337, 53 },
		{ 351, 338, 53 },
		{ 351, 339, 53 },
		{ 351, 340, 53 },
		{ 351, 341, 53 },
		{ 351, 342, 53 },
		{ 351, 343, 53 },
		{ 351, 344, 53 },
		{ 351, 345, 53 },
		{ 351, 346, 53 },
		{ 351, 347, 53 },
		{ 351, 348, 53 },
		{ 351, 349, 53 },
		{ 351, 350, 53 },
		{ 352, 329, 53 },
		{ 352, 330, 53 },
		{ 352, 331, 53 },
		{ 352, 332, 53 },
		{ 352, 333, 53 },
		{ 352, 334, 53 },
		{ 352, 335, 53 },
		{ 352, 336, 53 },
		{ 352, 337, 53 },
		{ 352, 338, 53 },
		{ 352, 339, 53 },
		{ 352, 340, 53 },
		{ 352, 341, 53 },
		{ 352, 342, 53 },
		{ 352, 343, 53 },
		{ 352, 344, 53 },
		{ 352, 345, 53 },
		{ 352, 346, 53 },
		{ 352, 347, 53 },
		{ 352, 348, 53 },
		{ 352, 349, 53 },
		{ 352, 350, 53 },
		{ 352, 351, 53 },
		{ 353, 330, 53 },
		{ 353, 331, 53 },
		{ 353, 332, 53 },
		{ 353, 333, 53 },
		{ 353, 334, 53 },
		{ 353, 335, 53 },
		{ 353, 336, 53 },
		{ 353, 337, 53 },
		{ 353, 338, 53 },
		{ 353, 339, 53 },
		{ 353, 340, 53 },
		{ 353, 341, 53 },
		{ 353, 342, 53 },
		{ 353, 343, 53 },
		{ 353, 344, 53 },
		{ 353, 345, 53 },
		{ 353, 346, 53 },
		{ 353, 347, 53 },
		{ 353, 348, 53 },
		{ 353, 349, 53 },
		{ 353, 350, 53 },
		{ 353, 351, 53 },
		{ 353, 352, 53 },
		{ 354, 330, 53 },
		{ 354, 331, 53 },
		{ 354, 332, 53 },
		{ 354, 333, 53 },
		{ 354, 334, 53 },
		{ 354, 335, 53 },
		{ 354, 336, 53 },
		{ 354, 337, 53 },
		{ 354, 338, 53 },
		{ 354, 339, 53 },
		{ 354, 340, 53 },
		{ 354, 341, 53 },
		{ 354, 342, 53 },
		{ 354, 343, 53 },
		{ 354, 344, 53 },
		{ 354, 345, 53 },
		{ 354, 346, 53 },
		{ 354, 347, 53 },
		{ 354, 348, 53 },
		{ 354, 349, 53 },
		{ 354, 350, 53 },
		{ 354, 351, 53 },
		{ 354, 352, 53 },
		{ 354, 353, 53 },
		{ 355, 331, 53 },
		{ 355, 332, 53 },
		{ 355, 333, 53 },
		{ 355, 334, 53 },
		{ 355, 335, 53 },
		{ 355, 336, 53 },
		{ 355, 337, 53 },
		{ 355, 338, 53 },
		{ 355, 339, 53 },
		{ 355, 340, 53 },
		{ 355, 341, 53 },
		{ 355, 342, 53 },
		{ 355, 343, 53 },
		{ 355, 344, 53 },
		{ 355, 345, 53 },
		{ 355, 346, 53 },
		{ 355, 347, 53 },
		{ 355, 348, 53 },
		{ 355, 349, 53 },
		{ 355, 350, 53 },
		{ 355, 351, 53 },
		{ 355, 352, 53 },
		{ 355, 353, 53 },
		{ 355, 354, 53 },
		{ 356, 332, 53 },
		{ 356, 333, 53 },
		{ 356, 334, 53 },
		{ 356, 335, 53 },
		{ 356, 336, 53 },
		{ 356, 337, 53 },
		{ 356, 338, 53 },
		{ 356, 339, 53 },
		{ 356, 340, 53 },
		{ 356, 341, 53 },
		{ 356, 342, 53 },
		{ 356, 343, 53 },
		{ 356, 344, 53 },
		{ 356, 345, 53 },
		{ 356, 346, 53 },
		{ 356, 347, 53 },
		{ 356, 348, 53 },
		{ 356, 349, 53 },
		{ 356, 350, 53 },
		{ 356, 351, 53 },
		{ 356, 352, 53 },
		{ 356, 353, 53 },
		{ 356, 354, 53 },
		{ 356, 355, 53 },
		{ 357, 333, 53 },
		{ 357, 334, 53 },
		{ 357, 335, 53 },
		{ 357, 336, 53 },
		{ 357, 337, 53 },
		{ 357, 338, 53 },
		{ 357, 339, 53 },
		{ 357, 340, 53 },
		{ 357, 341, 53 },
		{ 357, 342, 53 },
		{ 357, 343, 53 },
		{ 357, 344, 53 },
		{ 357, 345, 53 },
		{ 357, 346, 53 },
		{ 357, 347, 53 },
		{ 357, 348, 53 },
		{ 357, 349, 53 },
		{ 357, 350, 53 },
		{ 357, 351, 53 },
		{ 357, 352, 53 },
		{ 357, 353, 53 },
		{ 357, 354, 53 },
		{ 357, 355, 53 },
		{ 357, 356, 53 },
		{ 358, 333, 53 },
		{ 358, 334, 53 },
		{ 358, 335, 53 },
		{ 358, 336, 53 },
		{ 358, 337, 53 },
		{ 358, 338, 53 },
		{ 358, 339, 53 },
		{ 358, 340, 53 },
		{ 358, 341, 53 },
		{ 358, 342, 53 },
		{ 358, 343, 53 },
		{ 358, 344, 53 },
		{ 358, 345, 53 },
		{ 358, 346, 53 },
		{ 358, 347, 53 },
		{ 358, 348, 53 },
		{ 358, 349, 53 },
		{ 358, 350, 53 },
		{ 358, 351, 53 },
		{ 358, 352, 53 },
		{ 358, 353, 53 },
		{ 358, 354, 53 },
		{ 358, 355, 53 },
		{ 358, 356, 53 },
		{ 358, 357, 53 },
		{ 359, 334, 53 },
		{ 359, 335, 53 },
		{ 359, 336, 53 },
		{ 359, 337, 53 },
		{ 359, 338, 53 },
		{ 359, 339, 53 },
		{ 359, 340, 53 },
		{ 359, 341, 53 },
		{ 359, 342, 53 },
		{ 359, 343, 53 },
		{ 359, 344, 53 },
		{ 359, 345, 53 },
		{ 359, 346, 53 },
		{ 359, 347, 53 },
		{ 359, 348, 53 },
		{ 359, 349, 53 },
		{ 359, 350, 53 },
		{ 359, 351, 53 },
		{ 359, 352, 53 },
		{ 359, 353, 53 },
		{ 359, 354, 53 },
		{ 359, 355, 53 },
		{ 359, 356, 53 },
		{ 359, 357, 53 },
		{ 359, 358, 53 },
		{ 360, 335, 53 },
		{ 360, 336, 53 },
		{ 360, 337, 53 },
		{ 360, 338, 53 },
		{ 360, 339, 53 },
		{ 360, 340, 53 },
		{ 360, 341, 53 },
		{ 360, 342, 53 },
		{ 360, 343, 53 },
		{ 360, 344, 53 },
		{ 360, 345, 53 },
		{ 360, 346, 53 },
		{ 360, 347, 53 },
		{ 360, 348, 53 },
		{ 360, 349, 53 },
		{ 360, 350, 53 },
		{ 360, 351, 53 },
		{ 360, 352, 53 },
		{ 360, 353, 53 },
		{ 360, 354, 53 },
		{ 360, 355, 53 },
		{ 360, 356, 53 },
		{ 360, 357, 53 },
		{ 360, 358, 53 },
		{ 360, 359, 53 },
		{ 361, 336, 53 },
		{ 361, 337, 53 },
		{ 361, 338, 53 },
		{ 361, 339, 53 },
		{ 361, 340, 53 },
		{ 361, 341, 53 },
		{ 361, 342, 53 },
		{ 361, 343, 53 },
		{ 361, 344, 53 },
		{ 361, 345, 53 },
		{ 361, 346, 53 },
		{ 361, 347, 53 },
		{ 361, 348, 53 },
		{ 361, 349, 53 },
		{ 361, 350, 53 },
		{ 361, 351, 53 },
		{ 361, 352, 53 },
		{ 361, 353, 53 },
		{ 361, 354, 53 },
		{ 361, 355, 53 },
		{ 361, 356, 53 },
		{ 361, 357, 53 },
		{ 361, 358, 53 },
		{ 361, 359, 53 },
		{ 361, 360, 53 },
		{ 362, 336, 53 },
		{ 362, 337, 53 },
		{ 362, 338, 53 },
		{ 362, 339, 53 },
		{ 362, 340, 53 },
		{ 362, 341, 53 },
		{ 362, 342, 53 },
		{ 362, 343, 53 },
		{ 362, 344, 53 },
		{ 362, 345, 53 },
		{ 362, 346, 53 },
		{ 362, 347, 53 },
		{ 362, 348, 53 },
		{ 362, 349, 53 },
		{ 362, 350, 53 },
		{ 362, 351, 53 },
		{ 362, 352, 53 },
		{ 362, 353, 53 },
		{ 362, 354, 53 },
		{ 362, 355, 53 },
		{ 362, 356, 53 },
		{ 362, 357, 53 },
		{ 362, 358, 53 },
		{ 362, 359, 53 },
		{ 362, 360, 53 },
		{ 362, 361, 53 },
		{ 363, 337, 53 },
		{ 363, 338, 53 },
		{ 363, 339, 53 },
		{ 363, 340, 53 },
		{ 363, 341, 53 },
		{ 363, 342, 53 },
		{ 363, 343, 53 },
		{ 363, 344, 53 },
		{ 363, 345, 53 },
		{ 363, 346, 53 },
		{ 363, 347, 53 },
		{ 363, 348, 53 },
		{ 363, 349, 53 },
		{ 363, 350, 53 },
		{ 363, 351, 53 },
		{ 363, 352, 53 },
		{ 363, 353, 53 },
		{ 363, 354, 53 },
		{ 363, 355, 53 },
		{ 363, 356, 53 },
		{ 363, 357, 53 },
		{ 363, 358, 53 },
		{ 363, 359, 53 },
		{ 363, 360, 53 },
		{ 363, 361, 53 },
		{ 363, 362, 53 },
		{ 364, 338, 53 },
		{ 364, 339, 53 },
		{ 364, 340, 53 },
		{ 364, 341, 53 },
		{ 364, 342, 53 },
		{ 364, 343, 53 },
		{ 364, 344, 53 },
		{ 364, 345, 53 },
		{ 364, 346, 53 },
		{ 364, 347, 53 },
		{ 364, 348, 53 },
		{ 364, 349, 53 },
		{ 364, 350, 53 },
		{ 364, 351, 53 },
		{ 364, 352, 53 },
		{ 364, 353, 53 },
		{ 364, 354, 53 },
		{ 364, 355, 53 },
		{ 364, 356, 53 },
		{ 364, 357, 53 },
		{ 364, 358, 53 },
		{ 364, 359, 53 },
		{ 364, 360, 53 },
		{ 364, 361, 53 },
		{ 364, 362, 53 },
		{ 364, 363, 53 },
		{ 365, 339, 53 },
		{ 365, 340, 53 },
		{ 365, 341, 53 },
		{ 365, 342, 53 },
		{ 365, 343, 53 },
		{ 365, 344, 53 },
		{ 365, 345, 53 },
		{ 365, 346, 53 },
		{ 365, 347, 53 },
		{ 365, 348, 53 },
		{ 365, 349, 53 },
		{ 365, 350, 53 },
		{ 365, 351, 53 },
		{ 365, 352, 53 },
		{ 365, 353, 53 },
		{ 365, 354, 53 },
		{ 365, 355, 53 },
		{ 365, 356, 53 },
		{ 365, 357, 53 },
		{ 365, 358, 53 },
		{ 365, 359, 53 },
		{ 365, 360, 53 },
		{ 365, 361, 53 },
		{ 365, 362, 53 },
		{ 365, 363, 53 },
		{ 365, 364, 53 },
		{ 366, 339, 53 },
		{ 366, 340, 53 },
		{ 366, 341, 53 },
		{ 366, 342, 53 },
		{ 366, 343, 53 },
		{ 366, 344, 53 },
		{ 366, 345, 53 },
		{ 366, 346, 53 },
		{ 366, 347, 53 },
		{ 366, 348, 53 },
		{ 366, 349, 53 },
		{ 366, 350, 53 },
		{ 366, 351, 53 },
		{ 366, 352, 53 },
		{ 366, 353, 53 },
		{ 366, 354, 53 },
		{ 366, 355, 53 },
		{ 366, 356, 53 },
		{ 366, 357, 53 },
		{ 366, 358, 53 },
		{ 366, 359, 53 },
		{ 366, 360, 53 },
		{ 366, 361, 53 },
		{ 366, 362, 53 },
		{ 366, 363, 53 },
		{ 366, 364, 53 },
		{ 366, 365, 53 },
		{ 367, 340, 53 },
		{ 367, 341, 53 },
		{ 367, 342, 53 },
		{ 367, 343, 53 },
		{ 367, 344, 53 },
		{ 367, 345, 53 },
		{ 367, 346, 53 },
		{ 367, 347, 53 },
		{ 367, 348, 53 },
		{ 367, 349, 53 },
		{ 367, 350, 53 },
		{ 367, 351, 53 },
		{ 367, 352, 53 },
		{ 367, 353, 53 },
		{ 367, 354, 53 },
		{ 367, 355, 53 },
		{ 367, 356, 53 },
		{ 367, 357, 53 },
		{ 367, 358, 53 },
		{ 367, 359, 53 },
		{ 367, 360, 53 },
		{ 367, 361, 53 },
		{ 367, 362, 53 },
		{ 367, 363, 53 },
		{ 367, 364, 53 },
		{ 367, 365, 53 },
		{ 367, 366, 53 },
		{ 368, 341, 53 },
		{ 368, 342, 53 },
		{ 368, 343, 53 },
		{ 368, 344, 53 },
		{ 368, 345, 53 },
		{ 368, 346, 53 },
		{ 368, 347, 53 },
		{ 368, 348, 53 },
		{ 368, 349, 53 },
		{ 368, 350, 53 },
		{ 368, 351, 53 },
		{ 368, 352, 53 },
		{ 368, 353, 53 },
		{ 368, 354, 53 },
		{ 368, 355, 53 },
		{ 368, 356, 53 },
		{ 368, 357, 53 },
		{ 368, 358, 53 },
		{ 368, 359, 53 },
		{ 368, 360, 53 },
		{ 368, 361, 53 },
		{ 368, 362, 53 },
		{ 368, 363, 53 },
		{ 368, 364, 53 },
		{ 368, 365, 53 },
		{ 368, 366, 53 },
		{ 368, 367, 53 },
		{ 369, 342, 53 },
		{ 369, 343, 53 },
		{ 369, 344, 53 },
		{ 369, 345, 53 },
		{ 369, 346, 53 },
		{ 369, 347, 53 },
		{ 369, 348, 53 },
		{ 369, 349, 53 },
		{ 369, 350, 53 },
		{ 369, 351, 53 },
		{ 369, 352, 53 },
		{ 369, 353, 53 },
		{ 369, 354, 53 },
		{ 369, 355, 53 },
		{ 369, 356, 53 },
		{ 369, 357, 53 },
		{ 369, 358, 53 },
		{ 369, 359, 53 },
		{ 369, 360, 53 },
		{ 369, 361, 53 },
		{ 369, 362, 53 },
		{ 369, 363, 53 },
		{ 369, 364, 53 },
		{ 369, 365, 53 },
		{ 369, 366, 53 },
		{ 369, 367, 53 },
		{ 369, 368, 53 },
		{ 370, 342, 53 },
		{ 370, 343, 53 },
		{ 370, 344, 53 },
		{ 370, 345, 53 },
		{ 370, 346, 53 },
		{ 370, 347, 53 },
		{ 370, 348, 53 },
		{ 370, 349, 53 },
		{ 370, 350, 53 },
		{ 370, 351, 53 },
		{ 370, 352, 53 },
		{ 370, 353, 53 },
		{ 370, 354, 53 },
		{ 370, 355, 53 },
		{ 370, 356, 53 },
		{ 370, 357, 53 },
		{ 370, 358, 53 },
		{ 370, 359, 53 },
		{ 370, 360, 53 },
		{ 370, 361, 53 },
		{ 370, 362, 53 },
		{ 370, 363, 53 },
		{ 370, 364, 53 },
		{ 370, 365, 53 },
		{ 370, 366, 53 },
		{ 370, 367, 53 },
		{ 370, 368, 53 },
		{ 370, 369, 53 },
		{ 371, 343, 53 },
		{ 371, 344, 53 },
		{ 371, 345, 53 },
		{ 371, 346, 53 },
		{ 371, 347, 53 },
		{ 371, 348, 53 },
		{ 371, 349, 53 },
		{ 371, 350, 53 },
		{ 371, 351, 53 },
		{ 371, 352, 53 },
		{ 371, 353, 53 },
		{ 371, 354, 53 },
		{ 371, 355, 53 },
		{ 371, 356, 53 },
		{ 371, 357, 53 },
		{ 371, 358, 53 },
		{ 371, 359, 53 },
		{ 371, 360, 53 },
		{ 371, 361, 53 },
		{ 371, 362, 53 },
		{ 371, 363, 53 },
		{ 371, 364, 53 },
		{ 371, 365, 53 },
		{ 371, 366, 53 },
		{ 371, 367, 53 },
		{ 371, 368, 53 },
		{ 371, 369, 53 },
		{ 371, 370, 53 },
		{ 372, 344, 53 },
		{ 372, 345, 53 },
		{ 372, 346, 53 },
		{ 372, 347, 53 },
		{ 372, 348, 53 },
		{ 372, 349, 53 },
		{ 372, 350, 53 },
		{ 372, 351, 53 },
		{ 372, 352, 53 },
		{ 372, 353, 53 },
		{ 372, 354, 53 },
		{ 372, 355, 53 },
		{ 372, 356, 53 },
		{ 372, 357, 53 },
		{ 372, 358, 53 },
		{ 372, 359, 53 },
		{ 372, 360, 53 },
		{ 372, 361, 53 },
		{ 372, 362, 53 },
		{ 372, 363, 53 },
		{ 372, 364, 53 },
		{ 372, 365, 53 },
		{ 372, 366, 53 },
		{ 372, 367, 53 },
		{ 372, 368, 53 },
		{ 372, 369, 53 },
		{ 372, 370, 53 },
		{ 372, 371, 53 },
		{ 373, 345, 53 },
		{ 373, 346, 53 },
		{ 373, 347, 53 },
		{ 373, 348, 53 },
		{ 373, 349, 53 },
		{ 373, 350, 53 },
		{ 373, 351, 53 },
		{ 373, 352, 53 },
		{ 373, 353, 53 },
		{ 373, 354, 53 },
		{ 373, 355, 53 },
		{ 373, 356, 53 },
		{ 373, 357, 53 },
		{ 373, 358, 53 },
		{ 373, 359, 53 },
		{ 373, 360, 53 },
		{ 373, 361, 53 },
		{ 373, 362, 53 },
		{ 373, 363, 53 },
		{ 373, 364, 53 },
		{ 373, 365, 53 },
		{ 373, 366, 53 },
		{ 373, 367, 53 },
		{ 373, 368, 53 },
		{ 373, 369, 53 },
		{ 373, 370, 53 },
		{ 373, 371, 53 },
		{ 373, 372, 53 },
		{ 374, 345, 53 },
		{ 374, 346, 53 },
		{ 374, 347, 53 },
		{ 374, 348, 53 },
		{ 374, 349, 53 },
		{ 374, 350, 53 },
		{ 374, 351, 53 },
		{ 374, 352, 53 },
		{ 374, 353, 53 },
		{ 374, 354, 53 },
		{ 374, 355, 53 },
		{ 374, 356, 53 },
		{ 374, 357, 53 },
		{ 374, 358, 53 },
		{ 374, 359, 53 },
		{ 374, 360, 53 },
		{ 374, 361, 53 },
		{ 374, 362, 53 },
		{ 374, 363, 53 },
		{ 374, 364, 53 },
		{ 374, 365, 53 },
		{ 374, 366, 53 },
		{ 374, 367, 53 },
		{ 374, 368, 53 },
		{ 374, 369, 53 },
		{ 374, 370, 53 },
		{ 374, 371, 53 },
		{ 374, 372, 53 },
		{ 374, 373, 53 },
		{ 375, 346, 53 },
		{ 375, 347, 53 },
		{ 375, 348, 53 },
		{ 375, 349, 53 },
		{ 375, 350, 53 },
		{ 375, 351, 53 },
		{ 375, 352, 53 },
		{ 375, 353, 53 },
		{ 375, 354, 53 },
		{ 375, 355, 53 },
		{ 375, 356, 53 },
		{ 375, 357, 53 },
		{ 375, 358, 53 },
		{ 375, 359, 53 },
		{ 375, 360, 53 },
		{ 375, 361, 53 },
		{ 375, 362, 53 },
		{ 375, 363, 53 },
		{ 375, 364, 53 },
		{ 375, 365, 53 },
		{ 375, 366, 53 },
		{ 375, 367, 53 },
		{ 375, 368, 53 },
		{ 375, 369, 53 },
		{ 375, 370, 53 },
		{ 375, 371, 53 },
		{ 375, 372, 53 },
		{ 375, 373, 53 },
		{ 375, 374, 53 },
		{ 376, 347, 53 },
		{ 376, 348, 53 },
		{ 376, 349, 53 },
		{ 376, 350, 53 },
		{ 376, 351, 53 },
		{ 376, 352, 53 },
		{ 376, 353, 53 },
		{ 376, 354, 53 },
		{ 376, 355, 53 },
		{ 376, 356, 53 },
		{ 376, 357, 53 },
		{ 376, 358, 53 },
		{ 376, 359, 53 },
		{ 376, 360, 53 },
		{ 376, 361, 53 },
		{ 376, 362, 53 },
		{ 376, 363, 53 },
		{ 376, 364, 53 },
		{ 376, 365, 53 },
		{ 376, 366, 53 },
		{ 376, 367, 53 },
		{ 376, 368, 53 },
		{ 376, 369, 53 },
		{ 376, 370, 53 },
		{ 376, 371, 53 },
		{ 376, 372, 53 },
		{ 376, 373, 53 },
		{ 376, 374, 53 },
		{ 376, 375, 53 },
		{ 377, 348, 53 },
		{ 377, 349, 53 },
		{ 377, 350, 53 },
		{ 377, 351, 53 },
		{ 377, 352, 53 },
		{ 377, 353, 53 },
		{ 377, 354, 53 },
		{ 377, 355, 53 },
		{ 377, 356, 53 },
		{ 377, 357, 53 },
		{ 377, 358, 53 },
		{ 377, 359, 53 },
		{ 377, 360, 53 },
		{ 377, 361, 53 },
		{ 377, 362, 53 },
		{ 377, 363, 53 },
		{ 377, 364, 53 },
		{ 377, 365, 53 },
		{ 377, 366, 53 },
		{ 377, 367, 53 },
		{ 377, 368, 53 },
		{ 377, 369, 53 },
		{ 377, 370, 53 },
		{ 377, 371, 53 },
		{ 377, 372, 53 },
		{ 377, 373, 53 },
		{ 377, 374, 53 },
		{ 377, 375, 53 },
		{ 377, 376, 53 },
		{ 378, 348, 53 },
		{ 378, 349, 53 },
		{ 378, 350, 53 },
		{ 378, 351, 53 },
		{ 378, 352, 53 },
		{ 378, 353, 53 },
		{ 378, 354, 53 },
		{ 378, 355, 53 },
		{ 378, 356, 53 },
		{ 378, 357, 53 },
		{ 378, 358, 53 },
		{ 378, 359, 53 },
		{ 378, 360, 53 },
		{ 378, 361, 53 },
		{ 378, 362, 53 },
		{ 378, 363, 53 },
		{ 378, 364, 53 },
		{ 378, 365, 53 },
		{ 378, 366, 53 },
		{ 378, 367, 53 },
		{ 378, 368, 53 },
		{ 378, 369, 53 },
		{ 378, 370, 53 },
		{ 378, 371, 53 },
		{ 378, 372, 53 },
		{ 378, 373, 53 },
		{ 378, 374, 53 },
		{ 378, 375, 53 },
		{ 378, 376, 53 },
		{ 378, 377, 53 },
		{ 379, 349, 53 },
		{ 379, 350, 53 },
		{ 379, 351, 53 },
		{ 379, 352, 53 },
		{ 379, 353, 53 },
		{ 379, 354, 53 },
		{ 379, 355, 53 },
		{ 379, 356, 53 },
		{ 379, 357, 53 },
		{ 379, 358, 53 },
		{ 379, 359, 53 },
		{ 379, 360, 53 },
		{ 379, 361, 53 },
		{ 379, 362, 53 },
		{ 379, 363, 53 },
		{ 379, 364, 53 },
		{ 379, 365, 53 },
		{ 379, 366, 53 },
		{ 379, 367, 53 },
		{ 379, 368, 53 },
		{ 379, 369, 53 },
		{ 379, 370, 53 },
		{ 379, 371, 53 },
		{ 379, 372, 53 },
		{ 379, 373, 53 },
		{ 379, 374, 53 },
		{ 379, 375, 53 },
		{ 379, 376, 53 },
		{ 379, 377, 53 },
		{ 379, 378, 53 },
		{ 380, 350, 53 },
		{ 380, 351, 53 },
		{ 380, 352, 53 },
		{ 380, 353, 53 },
		{ 380, 354, 53 },
		{ 380, 355, 53 },
		{ 380, 356, 53 },
		{ 380, 357, 53 },
		{ 380, 358, 53 },
		{ 380, 359, 53 },
		{ 380, 360, 53 },
		{ 380, 361, 53 },
		{ 380, 362, 53 },
		{ 380, 363, 53 },
		{ 380, 364, 53 },
		{ 380, 365, 53 },
		{ 380, 366, 53 },
		{ 380, 367, 53 },
		{ 380, 368, 53 },
		{ 380, 369, 53 },
		{ 380, 370, 53 },
		{ 380, 371, 53 },
		{ 380, 372, 53 },
		{ 380, 373, 53 },
		{ 380, 374, 53 },
		{ 380, 375, 53 },
		{ 380, 376, 53 },
		{ 380, 377, 53 },
		{ 380, 378, 53 },
		{ 380, 379, 53 },
		{ 381, 351, 53 },
		{ 381, 352, 53 },
		{ 381, 353, 53 },
		{ 381, 354, 53 },
		{ 381, 355, 53 },
		{ 381, 356, 53 },
		{ 381, 357, 53 },
		{ 381, 358, 53 },
		{ 381, 359, 53 },
		{ 381, 360, 53 },
		{ 381, 361, 53 },
		{ 381, 362, 53 },
		{ 381, 363, 53 },
		{ 381, 364, 53 },
		{ 381, 365, 53 },
		{ 381, 366, 53 },
		{ 381, 367, 53 },
		{ 381, 368, 53 },
		{ 381, 369, 53 },
		{ 381, 370, 53 },
		{ 381, 371, 53 },
		{ 381, 372, 53 },
		{ 381, 373, 53 },
		{ 381, 374, 53 },
		{ 381, 375, 53 },
		{ 381, 376, 53 },
		{ 381, 377, 53 },
		{ 381, 378, 53 },
		{ 381, 379, 53 },
		{ 381, 380, 53 },
		{ 382, 351, 53 },
		{ 382, 352, 53 },
		{ 382, 353, 53 },
		{ 382, 354, 53 },
		{ 382, 355, 53 },
		{ 382, 356, 53 },
		{ 382, 357, 53 },
		{ 382, 358, 53 },
		{ 382, 359, 53 },
		{ 382, 360, 53 },
		{ 382, 361, 53 },
		{ 382, 362, 53 },
		{ 382, 363, 53 },
		{ 382, 364, 53 },
		{ 382, 365, 53 },
		{ 382, 366, 53 },
		{ 382, 367, 53 },
		{ 382, 368, 53 },
		{ 382, 369, 53 },
		{ 382, 370, 53 },
		{ 382, 371, 53 },
		{ 382, 372, 53 },
		{ 382, 373, 53 },
		{ 382, 374, 53 },
		{ 382, 375, 53 },
		{ 382, 376, 53 },
		{ 382, 377, 53 },
		{ 382, 378, 53 },
		{ 382, 379, 53 },
		{ 382, 380, 53 },
		{ 382, 381, 53 },
		{ 383, 352, 53 },
		{ 383, 353, 53 },
		{ 383, 354, 53 },
		{ 383, 355, 53 },
		{ 383, 356, 53 },
		{ 383, 357, 53 },
		{ 383, 358, 53 },
		{ 383, 359, 53 },
		{ 383, 360, 53 },
		{ 383, 361, 53 },
		{ 383, 362, 53 },
		{ 383, 363, 53 },
		{ 383, 364, 53 },
		{ 383, 365, 53 },
		{ 383, 366, 53 },
		{ 383, 367, 53 },
		{ 383, 368, 53 },
		{ 383, 369, 53 },
		{ 383, 370, 53 },
		{ 383, 371, 53 },
		{ 383, 372, 53 },
		{ 383, 373, 53 },
		{ 383, 374, 53 },
		{ 383, 375, 53 },
		{ 383, 376, 53 },
		{ 383, 377, 53 },
		{ 383, 378, 53 },
		{ 383, 379, 53 },
		{ 383, 380, 53 },
		{ 383, 381, 53 },
		{ 383, 382, 53 },
		{ 384, 353, 53 },
		{ 384, 354, 53 },
		{ 384, 355, 53 },
		{ 384, 356, 53 },
		{ 384, 357, 53 },
		{ 384, 358, 53 },
		{ 384, 359, 53 },
		{ 384, 360, 53 },
		{ 384, 361, 53 },
		{ 384, 362, 53 },
		{ 384, 363, 53 },
		{ 384, 364, 53 },
		{ 384, 365, 53 },
		{ 384, 366, 53 },
		{ 384, 367, 53 },
		{ 384, 368, 53 },
		{ 384, 369, 53 },
		{ 384, 370, 53 },
		{ 384, 371, 53 },
		{ 384, 372, 53 },
		{ 384, 373, 53 },
		{ 384, 374, 53 },
		{ 384, 375, 53 },
		{ 384, 376, 53 },
		{ 384, 377, 53 },
		{ 384, 378, 53 },
		{ 384, 379, 53 },
		{ 384, 380, 53 },
		{ 384, 381, 53 },
		{ 384, 382, 53 },
		{ 384, 383, 53 },
		{ 385, 354, 53 },
		{ 385, 355, 53 },
		{ 385, 356, 53 },
		{ 385, 357, 53 },
		{ 385, 358, 53 },
		{ 385, 359, 53 },
		{ 385, 360, 53 },
		{ 385, 361, 53 },
		{ 385, 362, 53 },
		{ 385, 363, 53 },
		{ 385, 364, 53 },
		{ 385, 365, 53 },
		{ 385, 366, 53 },
		{ 385, 367, 53 },
		{ 385, 368, 53 },
		{ 385, 369, 53 },
		{ 385, 370, 53 },
		{ 385, 371, 53 },
		{ 385, 372, 53 },
		{ 385, 373, 53 },
		{ 385, 374, 53 },
		{ 385, 375, 53 },
		{ 385, 376, 53 },
		{ 385, 377, 53 },
		{ 385, 378, 53 },
		{ 385, 379, 53 },
		{ 385, 380, 53 },
		{ 385, 381, 53 },
		{ 385, 382, 53 },
		{ 385, 383, 53 },
		{ 385, 384, 53 },
		{ 386, 354, 53 },
		{ 386, 355, 53 },
		{ 386, 356, 53 },
		{ 386, 357, 53 },
		{ 386, 358, 53 },
		{ 386, 359, 53 },
		{ 386, 360, 53 },
		{ 386, 361, 53 },
		{ 386, 362, 53 },
		{ 386, 363, 53 },
		{ 386, 364, 53 },
		{ 386, 365, 53 },
		{ 386, 366, 53 },
		{ 386, 367, 53 },
		{ 386, 368, 53 },
		{ 386, 369, 53 },
		{ 386, 370, 53 },
		{ 386, 371, 53 },
		{ 386, 372, 53 },
		{ 386, 373, 53 },
		{ 386, 374, 53 },
		{ 386, 375, 53 },
		{ 386, 376, 53 },
		{ 386, 377, 53 },
		{ 386, 378, 53 },
		{ 386, 379, 53 },
		{ 386, 380, 53 },
		{ 386, 381, 53 },
		{ 386, 382, 53 },
		{ 386, 383, 53 },
		{ 386, 384, 53 },
		{ 386, 385, 53 },
		{ 387, 355, 53 },
		{ 387, 356, 53 },
		{ 387, 357, 53 },
		{ 387, 358, 53 },
		{ 387, 359, 53 },
		{ 387, 360, 53 },
		{ 387, 361, 53 },
		{ 387, 362, 53 },
		{ 387, 363, 53 },
		{ 387, 364, 53 },
		{ 387, 365, 53 },
		{ 387, 366, 53 },
		{ 387, 367, 53 },
		{ 387, 368, 53 },
		{ 387, 369, 53 },
		{ 387, 370, 53 },
		{ 387, 371, 53 },
		{ 387, 372, 53 },
		{ 387, 373, 53 },
		{ 387, 374, 53 },
		{ 387, 375, 53 },
		{ 387, 376, 53 },
		{ 387, 377, 53 },
		{ 387, 378, 53 },
		{ 387, 379, 53 },
		{ 387, 380, 53 },
		{ 387, 381, 53 },
		{ 387, 382, 53 },
		{ 387, 383, 53 },
		{ 387, 384, 53 },
		{ 387, 385, 53 },
		{ 387, 386, 53 },
		{ 388, 356, 53 },
		{ 388, 357, 53 },
		{ 388, 358, 53 },
		{ 388, 359, 53 },
		{ 388, 360, 53 },
		{ 388, 361, 53 },
		{ 388, 362, 53 },
		{ 388, 363, 53 },
		{ 388, 364, 53 },
		{ 388, 365, 53 },
		{ 388, 366, 53 },
		{ 388, 367, 53 },
		{ 388, 368, 53 },
		{ 388, 369, 53 },
		{ 388, 370, 53 },
		{ 388, 371, 53 },
		{ 388, 372, 53 },
		{ 388, 373, 53 },
		{ 388, 374, 53 },
		{ 388, 375, 53 },
		{ 388, 376, 53 },
		{ 388, 377, 53 },
		{ 388, 378, 53 },
		{ 388, 379, 53 },
		{ 388, 380, 53 },
		{ 388, 381, 53 },
		{ 388, 382, 53 },
		{ 388, 383, 53 },
		{ 388, 384, 53 },
		{ 388, 385, 53 },
		{ 388, 386, 53 },
		{ 388, 387, 53 },
		{ 389, 357, 53 },
		{ 389, 358, 53 },
		{ 389, 359, 53 },
		{ 389, 360, 53 },
		{ 389, 361, 53 },
		{ 389, 362, 53 },
		{ 389, 363, 53 },
		{ 389, 364, 53 },
		{ 389, 365, 53 },
		{ 389, 366, 53 },
		{ 389, 367, 53 },
		{ 389, 368, 53 },
		{ 389, 369, 53 },
		{ 389, 370, 53 },
		{ 389, 371, 53 },
		{ 389, 372, 53 },
		{ 389, 373, 53 },
		{ 389, 374, 53 },
		{ 389, 375, 53 },
		{ 389, 376, 53 },
		{ 389, 377, 53 },
		{ 389, 378, 53 },
		{ 389, 379, 53 },
		{ 389, 380, 53 },
		{ 389, 381, 53 },
		{ 389, 382, 53 },
		{ 389, 383, 53 },
		{ 389, 384, 53 },
		{ 389, 385, 53 },
		{ 389, 386, 53 },
		{ 389, 387, 53 },
		{ 389, 388, 53 },
		{ 390, 357, 53 },
		{ 390, 358, 53 },
		{ 390, 359, 53 },
		{ 390, 360, 53 },
		{ 390, 361, 53 },
		{ 390, 362, 53 },
		{ 390, 363, 53 },
		{ 390, 364, 53 },
		{ 390, 365, 53 },
		{ 390, 366, 53 },
		{ 390, 367, 53 },
		{ 390, 368, 53 },
		{ 390, 369, 53 },
		{ 390, 370, 53 },
		{ 390, 371, 53 },
		{ 390, 372, 53 },
		{ 390, 373, 53 },
		{ 390, 374, 53 },
		{ 390, 375, 53 },
		{ 390, 376, 53 },
		{ 390, 377, 53 },
		{ 390, 378, 53 },
		{ 390, 379, 53 },
		{ 390, 380, 53 },
		{ 390, 381, 53 },
		{ 390, 382, 53 },
		{ 390, 383, 53 },
		{ 390, 384, 53 },
		{ 390, 385, 53 },
		{ 390, 386, 53 },
		{ 390, 387, 53 },
		{ 390, 388, 53 },
		{ 390, 389, 53 },
		{ 391, 358, 53 },
		{ 391, 359, 53 },
		{ 391, 360, 53 },
		{ 391, 361, 53 },
		{ 391, 362, 53 },
		{ 391, 363, 53 },
		{ 391, 364, 53 },
		{ 391, 365, 53 },
		{ 391, 366, 53 },
		{ 391, 367, 53 },
		{ 391, 368, 53 },
		{ 391, 369, 53 },
		{ 391, 370, 53 },
		{ 391, 371, 53 },
		{ 391, 372, 53 },
		{ 391, 373, 53 },
		{ 391, 374, 53 },
		{ 391, 375, 53 },
		{ 391, 376, 53 },
		{ 391, 377, 53 },
		{ 391, 378, 53 },
		{ 391, 379, 53 },
		{ 391, 380, 53 },
		{ 391, 381, 53 },
		{ 391, 382, 53 },
		{ 391, 383, 53 },
		{ 391, 384, 53 },
		{ 391, 385, 53 },
		{ 391, 386, 53 },
		{ 391, 387, 53 },
		{ 391, 388, 53 },
		{ 391, 389, 53 },
		{ 391, 390, 53 },
		{ 392, 359, 53 },
		{ 392, 360, 53 },
		{ 392, 361, 53 },
		{ 392, 362, 53 },
		{ 392, 363, 53 },
		{ 392, 364, 53 },
		{ 392, 365, 53 },
		{ 392, 366, 53 },
		{ 392, 367, 53 },
		{ 392, 368, 53 },
		{ 392, 369, 53 },
		{ 392, 370, 53 },
		{ 392, 371, 53 },
		{ 392, 372, 53 },
		{ 392, 373, 53 },
		{ 392, 374, 53 },
		{ 392, 375, 53 },
		{ 392, 376, 53 },
		{ 392, 377, 53 },
		{ 392, 378, 53 },
		{ 392, 379, 53 },
		{ 392, 380, 53 },
		{ 392, 381, 53 },
		{ 392, 382, 53 },
		{ 392, 383, 53 },
		{ 392, 384, 53 },
		{ 392, 385, 53 },
		{ 392, 386, 53 },
		{ 392, 387, 53 },
		{ 392, 388, 53 },
		{ 392, 389, 53 },
		{ 392, 390, 53 },
		{ 392, 391, 53 },
		{ 393, 360, 53 },
		{ 393, 361, 53 },
		{ 393, 362, 53 },
		{ 393, 363, 53 },
		{ 393, 364, 53 },
		{ 393, 365, 53 },
		{ 393, 366, 53 },
		{ 393, 367, 53 },
		{ 393, 368, 53 },
		{ 393, 369, 53 },
		{ 393, 370, 53 },
		{ 393, 371, 53 },
		{ 393, 372, 53 },
		{ 393, 373, 53 },
		{ 393, 374, 53 },
		{ 393, 375, 53 },
		{ 393, 376, 53 },
		{ 393, 377, 53 },
		{ 393, 378, 53 },
		{ 393, 379, 53 },
		{ 393, 380, 53 },
		{ 393, 381, 53 },
		{ 393, 382, 53 },
		{ 393, 383, 53 },
		{ 393, 384, 53 },
		{ 393, 385, 53 },
		{ 393, 386, 53 },
		{ 393, 387, 53 },
		{ 393, 388, 53 },
		{ 393, 389, 53 },
		{ 393, 390, 53 },
		{ 393, 391, 53 },
		{ 393, 392, 53 },
		{ 394, 360, 53 },
		{ 394, 361, 53 },
		{ 394, 362, 53 },
		{ 394, 363, 53 },
		{ 394, 364, 53 },
		{ 394, 365, 53 },
		{ 394, 366, 53 },
		{ 394, 367, 53 },
		{ 394, 368, 53 },
		{ 394, 369, 53 },
		{ 394, 370, 53 },
		{ 394, 371, 53 },
		{ 394, 372, 53 },
		{ 394, 373, 53 },
		{ 394, 374, 53 },
		{ 394, 375, 53 },
		{ 394, 376, 53 },
		{ 394, 377, 53 },
		{ 394, 378, 53 },
		{ 394, 379, 53 },
		{ 394, 380, 53 },
		{ 394, 381, 53 },
		{ 394, 382, 53 },
		{ 394, 383, 53 },
		{ 394, 384, 53 },
		{ 394, 385, 53 },
		{ 394, 386, 53 },
		{ 394, 387, 53 },
		{ 394, 388, 53 },
		{ 394, 389, 53 },
		{ 394, 390, 53 },
		{ 394, 391, 53 },
		{ 394, 392, 53 },
		{ 394, 393, 53 },
		{ 395, 361, 53 },
		{ 395, 362, 53 },
		{ 395, 363, 53 },
		{ 395, 364, 53 },
		{ 395, 365, 53 },
		{ 395, 366, 53 },
		{ 395, 367, 53 },
		{ 395, 368, 53 },
		{ 395, 369, 53 },
		{ 395, 370, 53 },
		{ 395, 371, 53 },
		{ 395, 372, 53 },
		{ 395, 373, 53 },
		{ 395, 374, 53 },
		{ 395, 375, 53 },
		{ 395, 376, 53 },
		{ 395, 377, 53 },
		{ 395, 378, 53 },
		{ 395, 379, 53 },
		{ 395, 380, 53 },
		{ 395, 381, 53 },
		{ 395, 382, 53 },
		{ 395, 383, 53 },
		{ 395, 384, 53 },
		{ 395, 385, 53 },
		{ 395, 386, 53 },
		{ 395, 387, 53 },
		{ 395, 388, 53 },
		{ 395, 389, 53 },
		{ 395, 390, 53 },
		{ 395, 391, 53 },
		{ 395, 392, 53 },
		{ 395, 393, 53 },
		{ 395, 394, 53 },
		{ 396, 362, 53 },
		{ 396, 363, 53 },
		{ 396, 364, 53 },
		{ 396, 365, 53 },
		{ 396, 366, 53 },
		{ 396, 367, 53 },
		{ 396, 368, 53 },
		{ 396, 369, 53 },
		{ 396, 370, 53 },
		{ 396, 371, 53 },
		{ 396, 372, 53 },
		{ 396, 373, 53 },
		{ 396, 374, 53 },
		{ 396, 375, 53 },
		{ 396, 376, 53 },
		{ 396, 377, 53 },
		{ 396, 378, 53 },
		{ 396, 379, 53 },
		{ 396, 380, 53 },
		{ 396, 381, 53 },
		{ 396, 382, 53 },
		{ 396, 383, 53 },
		{ 396, 384, 53 },
		{ 396, 385, 53 },
		{ 396, 386, 53 },
		{ 396, 387, 53 },
		{ 396, 388, 53 },
		{ 396, 389, 53 },
		{ 396, 390, 53 },
		{ 396, 391, 53 },
		{ 396, 392, 53 },
		{ 396, 393, 53 },
		{ 396, 394, 53 },
		{ 396, 395, 53 },
		{ 397, 363, 53 },
		{ 397, 364, 53 },
		{ 397, 365, 53 },
		{ 397, 366, 53 },
		{ 397, 367, 53 },
		{ 397, 368, 53 },
		{ 397, 369, 53 },
		{ 397, 370, 53 },
		{ 397, 371, 53 },
		{ 397, 372, 53 },
		{ 397, 373, 53 },
		{ 397, 374, 53 },
		{ 397, 375, 53 },
		{ 397, 376, 53 },
		{ 397, 377, 53 },
		{ 397, 378, 53 },
		{ 397, 379, 53 },
		{ 397, 380, 53 },
		{ 397, 381, 53 },
		{ 397, 382, 53 },
		{ 397, 383, 53 },
		{ 397, 384, 53 },
		{ 397, 385, 53 },
		{ 397, 386, 53 },
		{ 397, 387, 53 },
		{ 397, 388, 53 },
		{ 397, 389, 53 },
		{ 397, 390, 53 },
		{ 397, 391, 53 },
		{ 397, 392, 53 },
		{ 397, 393, 53 },
		{ 397, 394, 53 },
		{ 397, 395, 53 },
		{ 397, 396, 53 },
		{ 398, 363, 53 },
		{ 398, 364, 53 },
		{ 398, 365, 53 },
		{ 398, 366, 53 },
		{ 398, 367, 53 },
		{ 398, 368, 53 },
		{ 398, 369, 53 },
		{ 398, 370, 53 },
		{ 398, 371, 53 },
		{ 398, 372, 53 },
		{ 398, 373, 53 },
		{ 398, 374, 53 },
		{ 398, 375, 53 },
		{ 398, 376, 53 },
		{ 398, 377, 53 },
		{ 398, 378, 53 },
		{ 398, 379, 53 },
		{ 398, 380, 53 },
		{ 398, 381, 53 },
		{ 398, 382, 53 },
		{ 398, 383, 53 },
		{ 398, 384, 53 },
		{ 398, 385, 53 },
		{ 398, 386, 53 },
		{ 398, 387, 53 },
		{ 398, 388, 53 },
		{ 398, 389, 53 },
		{ 398, 390, 53 },
		{ 398, 391, 53 },
		{ 398, 392, 53 },
		{ 398, 393, 53 },
		{ 398, 394, 53 },
		{ 398, 395, 53 },
		{ 398, 396, 53 },
		{ 398, 397, 53 },
		{ 399, 364, 53 },
		{ 399, 365, 53 },
		{ 399, 366, 53 },
		{ 399, 367, 53 },
		{ 399, 368, 53 },
		{ 399, 369, 53 },
		{ 399, 370, 53 },
		{ 399, 371, 53 },
		{ 399, 372, 53 },
		{ 399, 373, 53 },
		{ 399, 374, 53 },
		{ 399, 375, 53 },
		{ 399, 376, 53 },
		{ 399, 377, 53 },
		{ 399, 378, 53 },
		{ 399, 379, 53 },
		{ 399, 380, 53 },
		{ 399, 381, 53 },
		{ 399, 382, 53 },
		{ 399, 383, 53 },
		{ 399, 384, 53 },
		{ 399, 385, 53 },
		{ 399, 386, 53 },
		{ 399, 387, 53 },
		{ 399, 388, 53 },
		{ 399, 389, 53 },
		{ 399, 390, 53 },
		{ 399, 391, 53 },
		{ 399, 392, 53 },
		{ 399, 393, 53 },
		{ 399, 394, 53 },
		{ 399, 395, 53 },
		{ 399, 396, 53 },
		{ 399, 397, 53 },
		{ 399, 398, 53 },
		{ 400, 365, 53 },
		{ 400, 366, 53 },
		{ 400, 367, 53 },
		{ 400, 368, 53 },
		{ 400, 369, 53 },
		{ 400, 370, 53 },
		{ 400, 371, 53 },
		{ 400, 372, 53 },
		{ 400, 373, 53 },
		{ 400, 374, 53 },
		{ 400, 375, 53 },
		{ 400, 376, 53 },
		{ 400, 377, 53 },
		{ 400, 378, 53 },
		{ 400, 379, 53 },
		{ 400, 380, 53 },
		{ 400, 381, 53 },
		{ 400, 382, 53 },
		{ 400, 383, 53 },
		{ 400, 384, 53 },
		{ 400, 385, 53 },
		{ 400, 386, 53 },
		{ 400, 387, 53 },
		{ 400, 388, 53 },
		{ 400, 389, 53 },
		{ 400, 390, 53 },
		{ 400, 391, 53 },
		{ 400, 392, 53 },
		{ 400, 393, 53 },
		{ 400, 394, 53 },
		{ 400, 395, 53 },
		{ 400, 396, 53 },
		{ 400, 397, 53 },
		{ 400, 398, 53 },
		{ 400, 399, 53 },
		{ 401, 366, 53 },
		{ 401, 367, 53 },
		{ 401, 368, 53 },
		{ 401, 369, 53 },
		{ 401, 370, 53 },
		{ 401, 371, 53 },
		{ 401, 372, 53 },
		{ 401, 373, 53 },
		{ 401, 374, 53 },
		{ 401, 375, 53 },
		{ 401, 376, 53 },
		{ 401, 377, 53 },
		{ 401, 378, 53 },
		{ 401, 379, 53 },
		{ 401, 380, 53 },
		{ 401, 381, 53 },
		{ 401, 382, 53 },
		{ 401, 383, 53 },
		{ 401, 384, 53 },
		{ 401, 385, 53 },
		{ 401, 386, 53 },
		{ 401, 387, 53 },
		{ 401, 388, 53 },
		{ 401, 389, 53 },
		{ 401, 390, 53 },
		{ 401, 391, 53 },
		{ 401, 392, 53 },
		{ 401, 393, 53 },
		{ 401, 394, 53 },
		{ 401, 395, 53 },
		{ 401, 396, 53 },
		{ 401, 397, 53 },
		{ 401, 398, 53 },
		{ 401, 399, 53 },
		{ 401, 400, 53 },
		{ 402, 366, 53 },
		{ 402, 367, 53 },
		{ 402, 368, 53 },
		{ 402, 369, 53 },
		{ 402, 370, 53 },
		{ 402, 371, 53 },
		{ 402, 372, 53 },
		{ 402, 373, 53 },
		{ 402, 374, 53 },
		{ 402, 375, 53 },
		{ 402, 376, 53 },
		{ 402, 377, 53 },
		{ 402, 378, 53 },
		{ 402, 379, 53 },
		{ 402, 380, 53 },
		{ 402, 381, 53 },
		{ 402, 382, 53 },
		{ 402, 383, 53 },
		{ 402, 384, 53 },
		{ 402, 385, 53 },
		{ 402, 386, 53 },
		{ 402, 387, 53 },
		{ 402, 388, 53 },
		{ 402, 389, 53 },
		{ 402, 390, 53 },
		{ 402, 391, 53 },
		{ 402, 392, 53 },
		{ 402, 393, 53 },
		{ 402, 394, 53 },
		{ 402, 395, 53 },
		{ 402, 396, 53 },
		{ 402, 397, 53 },
		{ 402, 398, 53 },
		{ 402, 399, 53 },
		{ 402, 400, 53 },
		{ 402, 401, 53 },
		{ 403, 367, 53 },
		{ 403, 368, 53 },
		{ 403, 369, 53 },
		{ 403, 370, 53 },
		{ 403, 371, 53 },
		{ 403, 372, 53 },
		{ 403, 373, 53 },
		{ 403, 374, 53 },
		{ 403, 375, 53 },
		{ 403, 376, 53 },
		{ 403, 377, 53 },
		{ 403, 378, 53 },
		{ 403, 379, 53 },
		{ 403, 380, 53 },
		{ 403, 381, 53 },
		{ 403, 382, 53 },
		{ 403, 383, 53 },
		{ 403, 384, 53 },
		{ 403, 385, 53 },
		{ 403, 386, 53 },
		{ 403, 387, 53 },
		{ 403, 388, 53 },
		{ 403, 389, 53 },
		{ 403, 390, 53 },
		{ 403, 391, 53 },
		{ 403, 392, 53 },
		{ 403, 393, 53 },
		{ 403, 394, 53 },
		{ 403, 395, 53 },
		{ 403, 396, 53 },
		{ 403, 397, 53 },
		{ 403, 398, 53 },
		{ 403, 399, 53 },
		{ 403, 400, 53 },
		{ 403, 401, 53 },
		{ 403, 402, 53 },
		{ 404, 368, 53 },
		{ 404, 369, 53 },
		{ 404, 370, 53 },
		{ 404, 371, 53 },
		{ 404, 372, 53 },
		{ 404, 373, 53 },
		{ 404, 374, 53 },
		{ 404, 375, 53 },
		{ 404, 376, 53 },
		{ 404, 377, 53 },
		{ 404, 378, 53 },
		{ 404, 379, 53 },
		{ 404, 380, 53 },
		{ 404, 381, 53 },
		{ 404, 382, 53 },
		{ 404, 383, 53 },
		{ 404, 384, 53 },
		{ 404, 385, 53 },
		{ 404, 386, 53 },
		{ 404, 387, 53 },
		{ 404, 388, 53 },
		{ 404, 389, 53 },
		{ 404, 390, 53 },
		{ 404, 391, 53 },
		{ 404, 392, 53 },
		{ 404, 393, 53 },
		{ 404, 394, 53 },
		{ 404, 395, 53 },
		{ 404, 396, 53 },
		{ 404, 397, 53 },
		{ 404, 398, 53 },
		{ 404, 399, 53 },
		{ 404, 400, 53 },
		{ 404, 401, 53 },
		{ 404, 402, 53 },
		{ 404, 403, 53 },
		{ 405, 369, 53 },
		{ 405, 370, 53 },
		{ 405, 371, 53 },
		{ 405, 372, 53 },
		{ 405, 373, 53 },
		{ 405, 374, 53 },
		{ 405, 375, 53 },
		{ 405, 376, 53 },
		{ 405, 377, 53 },
		{ 405, 378, 53 },
		{ 405, 379, 53 },
		{ 405, 380, 53 },
		{ 405, 381, 53 },
		{ 405, 382, 53 },
		{ 405, 383, 53 },
		{ 405, 384, 53 },
		{ 405, 385, 53 },
		{ 405, 386, 53 },
		{ 405, 387, 53 },
		{ 405, 388, 53 },
		{ 405, 389, 53 },
		{ 405, 390, 53 },
		{ 405, 391, 53 },
		{ 405, 392, 53 },
		{ 405, 393, 53 },
		{ 405, 394, 53 },
		{ 405, 395, 53 },
		{ 405, 396, 53 },
		{ 405, 397, 53 },
		{ 405, 398, 53 },
		{ 405, 399, 53 },
		{ 405, 400, 53 },
		{ 405, 401, 53 },
		{ 405, 402, 53 },
		{ 405, 403, 53 },
		{ 405, 404, 53 },
		{ 406, 369, 53 },
		{ 406, 370, 53 },
		{ 406, 371, 53 },
		{ 406, 372, 53 },
		{ 406, 373, 53 },
		{ 406, 374, 53 },
		{ 406, 375, 53 },
		{ 406, 376, 53 },
		{ 406, 377, 53 },
		{ 406, 378, 53 },
		{ 406, 379, 53 },
		{ 406, 380, 53 },
		{ 406, 381, 53 },
		{ 406, 382, 53 },
		{ 406, 383, 53 },
		{ 406, 384, 53 },
		{ 406, 385, 53 },
		{ 406, 386, 53 },
		{ 406, 387, 53 },
		{ 406, 388, 53 },
		{ 406, 389, 53 },
		{ 406, 390, 53 },
		{ 406, 391, 53 },
		{ 406, 392, 53 },
		{ 406, 393, 53 },
		{ 406, 394, 53 },
		{ 406, 395, 53 },
		{ 406, 396, 53 },
		{ 406, 397, 53 },
		{ 406, 398, 53 },
		{ 406, 399, 53 },
		{ 406, 400, 53 },
		{ 406, 401, 53 },
		{ 406, 402, 53 },
		{ 406, 403, 53 },
		{ 406, 404, 53 },
		{ 406, 405, 53 },
		{ 407, 370, 53 },
		{ 407, 371, 53 },
		{ 407, 372, 53 },
		{ 407, 373, 53 },
		{ 407, 374, 53 },
		{ 407, 375, 53 },
		{ 407, 376, 53 },
		{ 407, 377, 53 },
		{ 407, 378, 53 },
		{ 407, 379, 53 },
		{ 407, 380, 53 },
		{ 407, 381, 53 },
		{ 407, 382, 53 },
		{ 407, 383, 53 },
		{ 407, 384, 53 },
		{ 407, 385, 53 },
		{ 407, 386, 53 },
		{ 407, 387, 53 },
		{ 407, 388, 53 },
		{ 407, 389, 53 },
		{ 407, 390, 53 },
		{ 407, 391, 53 },
		{ 407, 392, 53 },
		{ 407, 393, 53 },
		{ 407, 394, 53 },
		{ 407, 395, 53 },
		{ 407, 396, 53 },
		{ 407, 397, 53 },
		{ 407, 398, 53 },
		{ 407, 399, 53 },
		{ 407, 400, 53 },
		{ 407, 401, 53 },
		{ 407, 402, 53 },
		{ 407, 403, 53 },
		{ 407, 404, 53 },
		{ 407, 405, 53 },
		{ 407, 406, 53 },
		{ 408, 371, 53 },
		{ 408, 372, 53 },
		{ 408, 373, 53 },
		{ 408, 374, 53 },
		{ 408, 375, 53 },
		{ 408, 376, 53 },
		{ 408, 377, 53 },
		{ 408, 378, 53 },
		{ 408, 379, 53 },
		{ 408, 380, 53 },
		{ 408, 381, 53 },
		{ 408, 382, 53 },
		{ 408, 383, 53 },
		{ 408, 384, 53 },
		{ 408, 385, 53 },
		{ 408, 386, 53 },
		{ 408, 387, 53 },
		{ 408, 388, 53 },
		{ 408, 389, 53 },
		{ 408, 390, 53 },
		{ 408, 391, 53 },
		{ 408, 392, 53 },
		{ 408, 393, 53 },
		{ 408, 394, 53 },
		{ 408, 395, 53 },
		{ 408, 396, 53 },
		{ 408, 397, 53 },
		{ 408, 398, 53 },
		{ 408, 399, 53 },
		{ 408, 400, 53 },
		{ 408, 401, 53 },
		{ 408, 402, 53 },
		{ 408, 403, 53 },
		{ 408, 404, 53 },
		{ 408, 405, 53 },
		{ 408, 406, 53 },
		{ 408, 407, 53 },
		{ 409, 372, 53 },
		{ 409, 373, 53 },
		{ 409, 374, 53 },
		{ 409, 375, 53 },
		{ 409, 376, 53 },
		{ 409, 377, 53 },
		{ 409, 378, 53 },
		{ 409, 379, 53 },
		{ 409, 380, 53 },
		{ 409, 381, 53 },
		{ 409, 382, 53 },
		{ 409, 383, 53 },
		{ 409, 384, 53 },
		{ 409, 385, 53 },
		{ 409, 386, 53 },
		{ 409, 387, 53 },
		{ 409, 388, 53 },
		{ 409, 389, 53 },
		{ 409, 390, 53 },
		{ 409, 391, 53 },
		{ 409, 392, 53 },
		{ 409, 393, 53 },
		{ 409, 394, 53 },
		{ 409, 395, 53 },
		{ 409, 396, 53 },
		{ 409, 397, 53 },
		{ 409, 398, 53 },
		{ 409, 399, 53 },
		{ 409, 400, 53 },
		{ 409, 401, 53 },
		{ 409, 402, 53 },
		{ 409, 403, 53 },
		{ 409, 404, 53 },
		{ 409, 405, 53 },
		{ 409, 406, 53 },
		{ 409, 407, 53 },
		{ 409, 408, 53 },
		{ 410, 372, 53 },
		{ 410, 373, 53 },
		{ 410, 374, 53 },
		{ 410, 375, 53 },
		{ 410, 376, 53 },
		{ 410, 377, 53 },
		{ 410, 378, 53 },
		{ 410, 379, 53 },
		{ 410, 380, 53 },
		{ 410, 381, 53 },
		{ 410, 382, 53 },
		{ 410, 383, 53 },
		{ 410, 384, 53 },
		{ 410, 385, 53 },
		{ 410, 386, 53 },
		{ 410, 387, 53 },
		{ 410, 388, 53 },
		{ 410, 389, 53 },
		{ 410, 390, 53 },
		{ 410, 391, 53 },
		{ 410, 392, 53 },
		{ 410, 393, 53 },
		{ 410, 394, 53 },
		{ 410, 395, 53 },
		{ 410, 396, 53 },
		{ 410, 397, 53 },
		{ 410, 398, 53 },
		{ 410, 399, 53 },
		{ 410, 400, 53 },
		{ 410, 401, 53 },
		{ 410, 402, 53 },
		{ 410, 403, 53 },
		{ 410, 404, 53 },
		{ 410, 405, 53 },
		{ 410, 406, 53 },
		{ 410, 407, 53 },
		{ 410, 408, 53 },
		{ 410, 409, 53 },
		{ 411, 373, 53 },
		{ 411, 374, 53 },
		{ 411, 375, 53 },
		{ 411, 376, 53 },
		{ 411, 377, 53 },
		{ 411, 378, 53 },
		{ 411, 379, 53 },
		{ 411, 380, 53 },
		{ 411, 381, 53 },
		{ 411, 382, 53 },
		{ 411, 383, 53 },
		{ 411, 384, 53 },
		{ 411, 385, 53 },
		{ 411, 386, 53 },
		{ 411, 387, 53 },
		{ 411, 388, 53 },
		{ 411, 389, 53 },
		{ 411, 390, 53 },
		{ 411, 391, 53 },
		{ 411, 392, 53 },
		{ 411, 393, 53 },
		{ 411, 394, 53 },
		{ 411, 395, 53 },
		{ 411, 396, 53 },
		{ 411, 397, 53 },
		{ 411, 398, 53 },
		{ 411, 399, 53 },
		{ 411, 400, 53 },
		{ 411, 401, 53 },
		{ 411, 402, 53 },
		{ 411, 403, 53 },
		{ 411, 404, 53 },
		{ 411, 405, 53 },
		{ 411, 406, 53 },
		{ 411, 407, 53 },
		{ 411, 408, 53 },
		{ 411, 409, 53 },
		{ 411, 410, 53 },
		{ 412, 374, 53 },
		{ 412, 375, 53 },
		{ 412, 376, 53 },
		{ 412, 377, 53 },
		{ 412, 378, 53 },
		{ 412, 379, 53 },
		{ 412, 380, 53 },
		{ 412, 381, 53 },
		{ 412, 382, 53 },
		{ 412, 383, 53 },
		{ 412, 384, 53 },
		{ 412, 385, 53 },
		{ 412, 386, 53 },
		{ 412, 387, 53 },
		{ 412, 388, 53 },
		{ 412, 389, 53 },
		{ 412, 390, 53 },
		{ 412, 391, 53 },
		{ 412, 392, 53 },
		{ 412, 393, 53 },
		{ 412, 394, 53 },
		{ 412, 395, 53 },
		{ 412, 396, 53 },
		{ 412, 397, 53 },
		{ 412, 398, 53 },
		{ 412, 399, 53 },
		{ 412, 400, 53 },
		{ 412, 401, 53 },
		{ 412, 402, 53 },
		{ 412, 403, 53 },
		{ 412, 404, 53 },
		{ 412, 405, 53 },
		{ 412, 406, 53 },
		{ 412, 407, 53 },
		{ 412, 408, 53 },
		{ 412, 409, 53 },
		{ 412, 410, 53 },
		{ 412, 411, 53 },
		{ 413, 375, 53 },
		{ 413, 376, 53 },
		{ 413, 377, 53 },
		{ 413, 378, 53 },
		{ 413, 379, 53 },
		{ 413, 380, 53 },
		{ 413, 381, 53 },
		{ 413, 382, 53 },
		{ 413, 383, 53 },
		{ 413, 384, 53 },
		{ 413, 385, 53 },
		{ 413, 386, 53 },
		{ 413, 387, 53 },
		{ 413, 388, 53 },
		{ 413, 389, 53 },
		{ 413, 390, 53 },
		{ 413, 391, 53 },
		{ 413, 392, 53 },
		{ 413, 393, 53 },
		{ 413, 394, 53 },
		{ 413, 395, 53 },
		{ 413, 396, 53 },
		{ 413, 397, 53 },
		{ 413, 398, 53 },
		{ 413, 399, 53 },
		{ 413, 400, 53 },
		{ 413, 401, 53 },
		{ 413, 402, 53 },
		{ 413, 403, 53 },
		{ 413, 404, 53 },
		{ 413, 405, 53 },
		{ 413, 406, 53 },
		{ 413, 407, 53 },
		{ 413, 408, 53 },
		{ 413, 409, 53 },
		{ 413, 410, 53 },
		{ 413, 411, 53 },
		{ 413, 412, 53 },
		{ 414, 375, 53 },
		{ 414, 376, 53 },
		{ 414, 377, 53 },
		{ 414, 378, 53 },
		{ 414, 379, 53 },
		{ 414, 380, 53 },
		{ 414, 381, 53 },
		{ 414, 382, 53 },
		{ 414, 383, 53 },
		{ 414, 384, 53 },
		{ 414, 385, 53 },
		{ 414, 386, 53 },
		{ 414, 387, 53 },
		{ 414, 388, 53 },
		{ 414, 389, 53 },
		{ 414, 390, 53 },
		{ 414, 391, 53 },
		{ 414, 392, 53 },
		{ 414, 393, 53 },
		{ 414, 394, 53 },
		{ 414, 395, 53 },
		{ 414, 396, 53 },
		{ 414, 397, 53 },
		{ 414, 398, 53 },
		{ 414, 399, 53 },
		{ 414, 400, 53 },
		{ 414, 401, 53 },
		{ 414, 402, 53 },
		{ 414, 403, 53 },
		{ 414, 404, 53 },
		{ 414, 405, 53 },
		{ 414, 406, 53 },
		{ 414, 407, 53 },
		{ 414, 408, 53 },
		{ 414, 409, 53 },
		{ 414, 410, 53 },
		{ 414, 411, 53 },
		{ 414, 412, 53 },
		{ 414, 413, 53 },
		{ 415, 376, 53 },
		{ 415, 377, 53 },
		{ 415, 378, 53 },
		{ 415, 379, 53 },
		{ 415, 380, 53 },
		{ 415, 381, 53 },
		{ 415, 382, 53 },
		{ 415, 383, 53 },
		{ 415, 384, 53 },
		{ 415, 385, 53 },
		{ 415, 386, 53 },
		{ 415, 387, 53 },
		{ 415, 388, 53 },
		{ 415, 389, 53 },
		{ 415, 390, 53 },
		{ 415, 391, 53 },
		{ 415, 392, 53 },
		{ 415, 393, 53 },
		{ 415, 394, 53 },
		{ 415, 395, 53 },
		{ 415, 396, 53 },
		{ 415, 397, 53 },
		{ 415, 398, 53 },
		{ 415, 399, 53 },
		{ 415, 400, 53 },
		{ 415, 401, 53 },
		{ 415, 402, 53 },
		{ 415, 403, 53 },
		{ 415, 404, 53 },
		{ 415, 405, 53 },
		{ 415, 406, 53 },
		{ 415, 407, 53 },
		{ 415, 408, 53 },
		{ 415, 409, 53 },
		{ 415, 410, 53 },
		{ 415, 411, 53 },
		{ 415, 412, 53 },
		{ 415, 413, 53 },
		{ 415, 414, 53 },
		{ 416, 377, 53 },
		{ 416, 378, 53 },
		{ 416, 379, 53 },
		{ 416, 380, 53 },
		{ 416, 381, 53 },
		{ 416, 382, 53 },
		{ 416, 383, 53 },
		{ 416, 384, 53 },
		{ 416, 385, 53 },
		{ 416, 386, 53 },
		{ 416, 387, 53 },
		{ 416, 388, 53 },
		{ 416, 389, 53 },
		{ 416, 390, 53 },
		{ 416, 391, 53 },
		{ 416, 392, 53 },
		{ 416, 393, 53 },
		{ 416, 394, 53 },
		{ 416, 395, 53 },
		{ 416, 396, 53 },
		{ 416, 397, 53 },
		{ 416, 398, 53 },
		{ 416, 399, 53 },
		{ 416, 400, 53 },
		{ 416, 401, 53 },
		{ 416, 402, 53 },
		{ 416, 403, 53 },
		{ 416, 404, 53 },
		{ 416, 405, 53 },
		{ 416, 406, 53 },
		{ 416, 407, 53 },
		{ 416, 408, 53 },
		{ 416, 409, 53 },
		{ 416, 410, 53 },
		{ 416, 411, 53 },
		{ 416, 412, 53 },
		{ 416, 413, 53 },
		{ 416, 414, 53 },
		{ 416, 415, 53 },
		{ 417, 378, 53 },
		{ 417, 379, 53 },
		{ 417, 380, 53 },
		{ 417, 381, 53 },
		{ 417, 382, 53 },
		{ 417, 383, 53 },
		{ 417, 384, 53 },
		{ 417, 385, 53 },
		{ 417, 386, 53 },
		{ 417, 387, 53 },
		{ 417, 388, 53 },
		{ 417, 389, 53 },
		{ 417, 390, 53 },
		{ 417, 391, 53 },
		{ 417, 392, 53 },
		{ 417, 393, 53 },
		{ 417, 394, 53 },
		{ 417, 395, 53 },
		{ 417, 396, 53 },
		{ 417, 397, 53 },
		{ 417, 398, 53 },
		{ 417, 399, 53 },
		{ 417, 400, 53 },
		{ 417, 401, 53 },
		{ 417, 402, 53 },
		{ 417, 403, 53 },
		{ 417, 404, 53 },
		{ 417, 405, 53 },
		{ 417, 406, 53 },
		{ 417, 407, 53 },
		{ 417, 408, 53 },
		{ 417, 409, 53 },
		{ 417, 410, 53 },
		{ 417, 411, 53 },
		{ 417, 412, 53 },
		{ 417, 413, 53 },
		{ 417, 414, 53 },
		{ 417, 415, 53 },
		{ 417, 416, 53 },
		{ 418, 378, 53 },
		{ 418, 379, 53 },
		{ 418, 380, 53 },
		{ 418, 381, 53 },
		{ 418, 382, 53 },
		{ 418, 383, 53 },
		{ 418, 384, 53 },
		{ 418, 385, 53 },
		{ 418, 386, 53 },
		{ 418, 387, 53 },
		{ 418, 388, 53 },
		{ 418, 389, 53 },
		{ 418, 390, 53 },
		{ 418, 391, 53 },
		{ 418, 392, 53 },
		{ 418, 393, 53 },
		{ 418, 394, 53 },
		{ 418, 395, 53 },
		{ 418, 396, 53 },
		{ 418, 397, 53 },
		{ 418, 398, 53 },
		{ 418, 399, 53 },
		{ 418, 400, 53 },
		{ 418, 401, 53 },
		{ 418, 402, 53 },
		{ 418, 403, 53 },
		{ 418, 404, 53 },
		{ 418, 405, 53 },
		{ 418, 406, 53 },
		{ 418, 407, 53 },
		{ 418, 408, 53 },
		{ 418, 409, 53 },
		{ 418, 410, 53 },
		{ 418, 411, 53 },
		{ 418, 412, 53 },
		{ 418, 413, 53 },
		{ 418, 414, 53 },
		{ 418, 415, 53 },
		{ 418, 416, 53 },
		{ 418, 417, 53 },
		{ 419, 379, 53 },
		{ 419, 380, 53 },
		{ 419, 381, 53 },
		{ 419, 382, 53 },
		{ 419, 383, 53 },
		{ 419, 384, 53 },
		{ 419, 385, 53 },
		{ 419, 386, 53 },
		{ 419, 387, 53 },
		{ 419, 388, 53 },
		{ 419, 389, 53 },
		{ 419, 390, 53 },
		{ 419, 391, 53 },
		{ 419, 392, 53 },
		{ 419, 393, 53 },
		{ 419, 394, 53 },
		{ 419, 395, 53 },
		{ 419, 396, 53 },
		{ 419, 397, 53 },
		{ 419, 398, 53 },
		{ 419, 399, 53 },
		{ 419, 400, 53 },
		{ 419, 401, 53 },
		{ 419, 402, 53 },
		{ 419, 403, 53 },
		{ 419, 404, 53 },
		{ 419, 405, 53 },
		{ 419, 406, 53 },
		{ 419, 407, 53 },
		{ 419, 408, 53 },
		{ 419, 409, 53 },
		{ 419, 410, 53 },
		{ 419, 411, 53 },
		{ 419, 412, 53 },
		{ 419, 413, 53 },
		{ 419, 414, 53 },
		{ 419, 415, 53 },
		{ 419, 416, 53 },
		{ 419, 417, 53 },
		{ 419, 418, 53 },
		{ 420, 380, 53 },
		{ 420, 381, 53 },
		{ 420, 382, 53 },
		{ 420, 383, 53 },
		{ 420, 384, 53 },
		{ 420, 385, 53 },
		{ 420, 386, 53 },
		{ 420, 387, 53 },
		{ 420, 388, 53 },
		{ 420, 389, 53 },
		{ 420, 390, 53 },
		{ 420, 391, 53 },
		{ 420, 392, 53 },
		{ 420, 393, 53 },
		{ 420, 394, 53 },
		{ 420, 395, 53 },
		{ 420, 396, 53 },
		{ 420, 397, 53 },
		{ 420, 398, 53 },
		{ 420, 399, 53 },
		{ 420, 400, 53 },
		{ 420, 401, 53 },
		{ 420, 402, 53 },
		{ 420, 403, 53 },
		{ 420, 404, 53 },
		{ 420, 405, 53 },
		{ 420, 406, 53 },
		{ 420, 407, 53 },
		{ 420, 408, 53 },
		{ 420, 409, 53 },
		{ 420, 410, 53 },
		{ 420, 411, 53 },
		{ 420, 412, 53 },
		{ 420, 413, 53 },
		{ 420, 414, 53 },
		{ 420, 415, 53 },
		{ 420, 416, 53 },
		{ 420, 417, 53 },
		{ 420, 418, 53 },
		{ 420, 419, 53 },
		{ 421, 381, 53 },
		{ 421, 382, 53 },
		{ 421, 383, 53 },
		{ 421, 384, 53 },
		{ 421, 385, 53 },
		{ 421, 386, 53 },
		{ 421, 387, 53 },
		{ 421, 388, 53 },
		{ 421, 389, 53 },
		{ 421, 390, 53 },
		{ 421, 391, 53 },
		{ 421, 392, 53 },
		{ 421, 393, 53 },
		{ 421, 394, 53 },
		{ 421, 395, 53 },
		{ 421, 396, 53 },
		{ 421, 397, 53 },
		{ 421, 398, 53 },
		{ 421, 399, 53 },
		{ 421, 400, 53 },
		{ 421, 401, 53 },
		{ 421, 402, 53 },
		{ 421, 403, 53 },
		{ 421, 404, 53 },
		{ 421, 405, 53 },
		{ 421, 406, 53 },
		{ 421, 407, 53 },
		{ 421, 408, 53 },
		{ 421, 409, 53 },
		{ 421, 410, 53 },
		{ 421, 411, 53 },
		{ 421, 412, 53 },
		{ 421, 413, 53 },
		{ 421, 414, 53 },
		{ 421, 415, 53 },
		{ 421, 416, 53 },
		{ 421, 417, 53 },
		{ 421, 418, 53 },
		{ 421, 419, 53 },
		{ 421, 420, 53 },
		{ 422, 381, 53 },
		{ 422, 382, 53 },
		{ 422, 383, 53 },
		{ 422, 384, 53 },
		{ 422, 385, 53 },
		{ 422, 386, 53 },
		{ 422, 387, 53 },
		{ 422, 388, 53 },
		{ 422, 389, 53 },
		{ 422, 390, 53 },
		{ 422, 391, 53 },
		{ 422, 392, 53 },
		{ 422, 393, 53 },
		{ 422, 394, 53 },
		{ 422, 395, 53 },
		{ 422, 396, 53 },
		{ 422, 397, 53 },
		{ 422, 398, 53 },
		{ 422, 399, 53 },
		{ 422, 400, 53 },
		{ 422, 401, 53 },
		{ 422, 402, 53 },
		{ 422, 403, 53 },
		{ 422, 404, 53 },
		{ 422, 405, 53 },
		{ 422, 406, 53 },
		{ 422, 407, 53 },
		{ 422, 408, 53 },
		{ 422, 409, 53 },
		{ 422, 410, 53 },
		{ 422, 411, 53 },
		{ 422, 412, 53 },
		{ 422, 413, 53 },
		{ 422, 414, 53 },
		{ 422, 415, 53 },
		{ 422, 416, 53 },
		{ 422, 417, 53 },
		{ 422, 418, 53 },
		{ 422, 419, 53 },
		{ 422, 420, 53 },
		{ 422, 421, 53 },
		{ 423, 382, 53 },
		{ 423, 383, 53 },
		{ 423, 384, 53 },
		{ 423, 385, 53 },
		{ 423, 386, 53 },
		{ 423, 387, 53 },
		{ 423, 388, 53 },
		{ 423, 389, 53 },
		{ 423, 390, 53 },
		{ 423, 391, 53 },
		{ 423, 392, 53 },
		{ 423, 393, 53 },
		{ 423, 394, 53 },
		{ 423, 395, 53 },
		{ 423, 396, 53 },
		{ 423, 397, 53 },
		{ 423, 398, 53 },
		{ 423, 399, 53 },
		{ 423, 400, 53 },
		{ 423, 401, 53 },
		{ 423, 402, 53 },
		{ 423, 403, 53 },
		{ 423, 404, 53 },
		{ 423, 405, 53 },
		{ 423, 406, 53 },
		{ 423, 407, 53 },
		{ 423, 408, 53 },
		{ 423, 409, 53 },
		{ 423, 410, 53 },
		{ 423, 411, 53 },
		{ 423, 412, 53 },
		{ 423, 413, 53 },
		{ 423, 414, 53 },
		{ 423, 415, 53 },
		{ 423, 416, 53 },
		{ 423, 417, 53 },
		{ 423, 418, 53 },
		{ 423, 419, 53 },
		{ 423, 420, 53 },
		{ 423, 421, 53 },
		{ 423, 422, 53 },
		{ 424, 383, 53 },
		{ 424, 384, 53 },
		{ 424, 385, 53 },
		{ 424, 386, 53 },
		{ 424, 387, 53 },
		{ 424, 388, 53 },
		{ 424, 389, 53 },
		{ 424, 390, 53 },
		{ 424, 391, 53 },
		{ 424, 392, 53 },
		{ 424, 393, 53 },
		{ 424, 394, 53 },
		{ 424, 395, 53 },
		{ 424, 396, 53 },
		{ 424, 397, 53 },
		{ 424, 398, 53 },
		{ 424, 399, 53 },
		{ 424, 400, 53 },
		{ 424, 401, 53 },
		{ 424, 402, 53 },
		{ 424, 403, 53 },
		{ 424, 404, 53 },
		{ 424, 405, 53 },
		{ 424, 406, 53 },
		{ 424, 407, 53 },
		{ 424, 408, 53 },
		{ 424, 409, 53 },
		{ 424, 410, 53 },
		{ 424, 411, 53 },
		{ 424, 412, 53 },
		{ 424, 413, 53 },
		{ 424, 414, 53 },
		{ 424, 415, 53 },
		{ 424, 416, 53 },
		{ 424, 417, 53 },
		{ 424, 418, 53 },
		{ 424, 419, 53 },
		{ 424, 420, 53 },
		{ 424, 421, 53 },
		{ 424, 422, 53 },
		{ 424, 423, 53 },
		{ 425, 384, 53 },
		{ 425, 385, 53 },
		{ 425, 386, 53 },
		{ 425, 387, 53 },
		{ 425, 388, 53 },
		{ 425, 389, 53 },
		{ 425, 390, 53 },
		{ 425, 391, 53 },
		{ 425, 392, 53 },
		{ 425, 393, 53 },
		{ 425, 394, 53 },
		{ 425, 395, 53 },
		{ 425, 396, 53 },
		{ 425, 397, 53 },
		{ 425, 398, 53 },
		{ 425, 399, 53 },
		{ 425, 400, 53 },
		{ 425, 401, 53 },
		{ 425, 402, 53 },
		{ 425, 403, 53 },
		{ 425, 404, 53 },
		{ 425, 405, 53 },
		{ 425, 406, 53 },
		{ 425, 407, 53 },
		{ 425, 408, 53 },
		{ 425, 409, 53 },
		{ 425, 410, 53 },
		{ 425, 411, 53 },
		{ 425, 412, 53 },
		{ 425, 413, 53 },
		{ 425, 414, 53 },
		{ 425, 415, 53 },
		{ 425, 416, 53 },
		{ 425, 417, 53 },
		{ 425, 418, 53 },
		{ 425, 419, 53 },
		{ 425, 420, 53 },
		{ 425, 421, 53 },
		{ 425, 422, 53 },
		{ 425, 423, 53 },
		{ 425, 424, 53 },
		{ 426, 384, 53 },
		{ 426, 385, 53 },
		{ 426, 386, 53 },
		{ 426, 387, 53 },
		{ 426, 388, 53 },
		{ 426, 389, 53 },
		{ 426, 390, 53 },
		{ 426, 391, 53 },
		{ 426, 392, 53 },
		{ 426, 393, 53 },
		{ 426, 394, 53 },
		{ 426, 395, 53 },
		{ 426, 396, 53 },
		{ 426, 397, 53 },
		{ 426, 398, 53 },
		{ 426, 399, 53 },
		{ 426, 400, 53 },
		{ 426, 401, 53 },
		{ 426, 402, 53 },
		{ 426, 403, 53 },
		{ 426, 404, 53 },
		{ 426, 405, 53 },
		{ 426, 406, 53 },
		{ 426, 407, 53 },
		{ 426, 408, 53 },
		{ 426, 409, 53 },
		{ 426, 410, 53 },
		{ 426, 411, 53 },
		{ 426, 412, 53 },
		{ 426, 413, 53 },
		{ 426, 414, 53 },
		{ 426, 415, 53 },
		{ 426, 416, 53 },
		{ 426, 417, 53 },
		{ 426, 418, 53 },
		{ 426, 419, 53 },
		{ 426, 420, 53 },
		{ 426, 421, 53 },
		{ 426, 422, 53 },
		{ 426, 423, 53 },
		{ 426, 424, 53 },
		{ 426, 425, 53 },
		{ 427, 385, 53 },
		{ 427, 386, 53 },
		{ 427, 387, 53 },
		{ 427, 388, 53 },
		{ 427, 389, 53 },
		{ 427, 390, 53 },
		{ 427, 391, 53 },
		{ 427, 392, 53 },
		{ 427, 393, 53 },
		{ 427, 394, 53 },
		{ 427, 395, 53 },
		{ 427, 396, 53 },
		{ 427, 397, 53 },
		{ 427, 398, 53 },
		{ 427, 399, 53 },
		{ 427, 400, 53 },
		{ 427, 401, 53 },
		{ 427, 402, 53 },
		{ 427, 403, 53 },
		{ 427, 404, 53 },
		{ 427, 405, 53 },
		{ 427, 406, 53 },
		{ 427, 407, 53 },
		{ 427, 408, 53 },
		{ 427, 409, 53 },
		{ 427, 410, 53 },
		{ 427, 411, 53 },
		{ 427, 412, 53 },
		{ 427, 413, 53 },
		{ 427, 414, 53 },
		{ 427, 415, 53 },
		{ 427, 416, 53 },
		{ 427, 417, 53 },
		{ 427, 418, 53 },
		{ 427, 419, 53 },
		{ 427, 420, 53 },
		{ 427, 421, 53 },
		{ 427, 422, 53 },
		{ 427, 423, 53 },
		{ 427, 424, 53 },
		{ 427, 425, 53 },
		{ 427, 426, 53 },
		{ 428, 386, 53 },
		{ 428, 387, 53 },
		{ 428, 388, 53 },
		{ 428, 389, 53 },
		{ 428, 390, 53 },
		{ 428, 391, 53 },
		{ 428, 392, 53 },
		{ 428, 393, 53 },
		{ 428, 394, 53 },
		{ 428, 395, 53 },
		{ 428, 396, 53 },
		{ 428, 397, 53 },
		{ 428, 398, 53 },
		{ 428, 399, 53 },
		{ 428, 400, 53 },
		{ 428, 401, 53 },
		{ 428, 402, 53 },
		{ 428, 403, 53 },
		{ 428, 404, 53 },
		{ 428, 405, 53 },
		{ 428, 406, 53 },
		{ 428, 407, 53 },
		{ 428, 408, 53 },
		{ 428, 409, 53 },
		{ 428, 410, 53 },
		{ 428, 411, 53 },
		{ 428, 412, 53 },
		{ 428, 413, 53 },
		{ 428, 414, 53 },
		{ 428, 415, 53 },
		{ 428, 416, 53 },
		{ 428, 417, 53 },
		{ 428, 418, 53 },
		{ 428, 419, 53 },
		{ 428, 420, 53 },
		{ 428, 421, 53 },
		{ 428, 422, 53 },
		{ 428, 423, 53 },
		{ 428, 424, 53 },
		{ 428, 425, 53 },
		{ 428, 426, 53 },
		{ 428, 427, 53 },
		{ 429, 387, 53 },
		{ 429, 388, 53 },
		{ 429, 389, 53 },
		{ 429, 390, 53 },
		{ 429, 391, 53 },
		{ 429, 392, 53 },
		{ 429, 393, 53 },
		{ 429, 394, 53 },
		{ 429, 395, 53 },
		{ 429, 396, 53 },
		{ 429, 397, 53 },
		{ 429, 398, 53 },
		{ 429, 399, 53 },
		{ 429, 400, 53 },
		{ 429, 401, 53 },
		{ 429, 402, 53 },
		{ 429, 403, 53 },
		{ 429, 404, 53 },
		{ 429, 405, 53 },
		{ 429, 406, 53 },
		{ 429, 407, 53 },
		{ 429, 408, 53 },
		{ 429, 409, 53 },
		{ 429, 410, 53 },
		{ 429, 411, 53 },
		{ 429, 412, 53 },
		{ 429, 413, 53 },
		{ 429, 414, 53 },
		{ 429, 415, 53 },
		{ 429, 416, 53 },
		{ 429, 417, 53 },
		{ 429, 418, 53 },
		{ 429, 419, 53 },
		{ 429, 420, 53 },
		{ 429, 421, 53 },
		{ 429, 422, 53 },
		{ 429, 423, 53 },
		{ 429, 424, 53 },
		{ 429, 425, 53 },
		{ 429, 426, 53 },
		{ 429, 427, 53 },
		{ 429, 428, 53 },
		{ 430, 387, 53 },
		{ 430, 388, 53 },
		{ 430, 389, 53 },
		{ 430, 390, 53 },
		{ 430, 391, 53 },
		{ 430, 392, 53 },
		{ 430, 393, 53 },
		{ 430, 394, 53 },
		{ 430, 395, 53 },
		{ 430, 396, 53 },
		{ 430, 397, 53 },
		{ 430, 398, 53 },
		{ 430, 399, 53 },
		{ 430, 400, 53 },
		{ 430, 401, 53 },
		{ 430, 402, 53 },
		{ 430, 403, 53 },
		{ 430, 404, 53 },
		{ 430, 405, 53 },
		{ 430, 406, 53 },
		{ 430, 407, 53 },
		{ 430, 408, 53 },
		{ 430, 409, 53 },
		{ 430, 410, 53 },
		{ 430, 411, 53 },
		{ 430, 412, 53 },
		{ 430, 413, 53 },
		{ 430, 414, 53 },
		{ 430, 415, 53 },
		{ 430, 416, 53 },
		{ 430, 417, 53 },
		{ 430, 418, 53 },
		{ 430, 419, 53 },
		{ 430, 420, 53 },
		{ 430, 421, 53 },
		{ 430, 422, 53 },
		{ 430, 423, 53 },
		{ 430, 424, 53 },
		{ 430, 425, 53 },
		{ 430, 426, 53 },
		{ 430, 427, 53 },
		{ 430, 428, 53 },
		{ 430, 429, 53 },
		{ 431, 388, 53 },
		{ 431, 389, 53 },
		{ 431, 390, 53 },
		{ 431, 391, 53 },
		{ 431, 392, 53 },
		{ 431, 393, 53 },
		{ 431, 394, 53 },
		{ 431, 395, 53 },
		{ 431, 396, 53 },
		{ 431, 397, 53 },
		{ 431, 398, 53 },
		{ 431, 399, 53 },
		{ 431, 400, 53 },
		{ 431, 401, 53 },
		{ 431, 402, 53 },
		{ 431, 403, 53 },
		{ 431, 404, 53 },
		{ 431, 405, 53 },
		{ 431, 406, 53 },
		{ 431, 407, 53 },
		{ 431, 408, 53 },
		{ 431, 409, 53 },
		{ 431, 410, 53 },
		{ 431, 411, 53 },
		{ 431, 412, 53 },
		{ 431, 413, 53 },
		{ 431, 414, 53 },
		{ 431, 415, 53 },
		{ 431, 416, 53 },
		{ 431, 417, 53 },
		{ 431, 418, 53 },
		{ 431, 419, 53 },
		{ 431, 420, 53 },
		{ 431, 421, 53 },
		{ 431, 422, 53 },
		{ 431, 423, 53 },
		{ 431, 424, 53 },
		{ 431, 425, 53 },
		{ 431, 426, 53 },
		{ 431, 427, 53 },
		{ 431, 428, 53 },
		{ 431, 429, 53 },
		{ 431, 430, 53 },
		{ 432, 389, 53 },
		{ 432, 390, 53 },
		{ 432, 391, 53 },
		{ 432, 392, 53 },
		{ 432, 393, 53 },
		{ 432, 394, 53 },
		{ 432, 395, 53 },
		{ 432, 396, 53 },
		{ 432, 397, 53 },
		{ 432, 398, 53 },
		{ 432, 399, 53 },
		{ 432, 400, 53 },
		{ 432, 401, 53 },
		{ 432, 402, 53 },
		{ 432, 403, 53 },
		{ 432, 404, 53 },
		{ 432, 405, 53 },
		{ 432, 406, 53 },
		{ 432, 407, 53 },
		{ 432, 408, 53 },
		{ 432, 409, 53 },
		{ 432, 410, 53 },
		{ 432, 411, 53 },
		{ 432, 412, 53 },
		{ 432, 413, 53 },
		{ 432, 414, 53 },
		{ 432, 415, 53 },
		{ 432, 416, 53 },
		{ 432, 417, 53 },
		{ 432, 418, 53 },
		{ 432, 419, 53 },
		{ 432, 420, 53 },
		{ 432, 421, 53 },
		{ 432, 422, 53 },
		{ 432, 423, 53 },
		{ 432, 424, 53 },
		{ 432, 425, 53 },
		{ 432, 426, 53 },
		{ 432, 427, 53 },
		{ 432, 428, 53 },
		{ 432, 429, 53 },
		{ 432, 430, 53 },
		{ 432, 431, 53 },
		{ 433, 390, 53 },
		{ 433, 391, 53 },
		{ 433, 392, 53 },
		{ 433, 393, 53 },
		{ 433, 394, 53 },
		{ 433, 395, 53 },
		{ 433, 396, 53 },
		{ 433, 397, 53 },
		{ 433, 398, 53 },
		{ 433, 399, 53 },
		{ 433, 400, 53 },
		{ 433, 401, 53 },
		{ 433, 402, 53 },
		{ 433, 403, 53 },
		{ 433, 404, 53 },
		{ 433, 405, 53 },
		{ 433, 406, 53 },
		{ 433, 407, 53 },
		{ 433, 408, 53 },
		{ 433, 409, 53 },
		{ 433, 410, 53 },
		{ 433, 411, 53 },
		{ 433, 412, 53 },
		{ 433, 413, 53 },
		{ 433, 414, 53 },
		{ 433, 415, 53 },
		{ 433, 416, 53 },
		{ 433, 417, 53 },
		{ 433, 418, 53 },
		{ 433, 419, 53 },
		{ 433, 420, 53 },
		{ 433, 421, 53 },
		{ 433, 422, 53 },
		{ 433, 423, 53 },
		{ 433, 424, 53 },
		{ 433, 425, 53 },
		{ 433, 426, 53 },
		{ 433, 427, 53 },
		{ 433, 428, 53 },
		{ 433, 429, 53 },
		{ 433, 430, 53 },
		{ 433, 431, 53 },
		{ 433, 432, 53 },
		{ 434, 390, 53 },
		{ 434, 391, 53 },
		{ 434, 392, 53 },
		{ 434, 393, 53 },
		{ 434, 394, 53 },
		{ 434, 395, 53 },
		{ 434, 396, 53 },
		{ 434, 397, 53 },
		{ 434, 398, 53 },
		{ 434, 399, 53 },
		{ 434, 400, 53 },
		{ 434, 401, 53 },
		{ 434, 402, 53 },
		{ 434, 403, 53 },
		{ 434, 404, 53 },
		{ 434, 405, 53 },
		{ 434, 406, 53 },
		{ 434, 407, 53 },
		{ 434, 408, 53 },
		{ 434, 409, 53 },
		{ 434, 410, 53 },
		{ 434, 411, 53 },
		{ 434, 412, 53 },
		{ 434, 413, 53 },
		{ 434, 414, 53 },
		{ 434, 415, 53 },
		{ 434, 416, 53 },
		{ 434, 417, 53 },
		{ 434, 418, 53 },
		{ 434, 419, 53 },
		{ 434, 420, 53 },
		{ 434, 421, 53 },
		{ 434, 422, 53 },
		{ 434, 423, 53 },
		{ 434, 424, 53 },
		{ 434, 425, 53 },
		{ 434, 426, 53 },
		{ 434, 427, 53 },
		{ 434, 428, 53 },
		{ 434, 429, 53 },
		{ 434, 430, 53 },
		{ 434, 431, 53 },
		{ 434, 432, 53 },
		{ 434, 433, 53 },
		{ 435, 391, 53 },
		{ 435, 392, 53 },
		{ 435, 393, 53 },
		{ 435, 394, 53 },
		{ 435, 395, 53 },
		{ 435, 396, 53 },
		{ 435, 397, 53 },
		{ 435, 398, 53 },
		{ 435, 399, 53 },
		{ 435, 400, 53 },
		{ 435, 401, 53 },
		{ 435, 402, 53 },
		{ 435, 403, 53 },
		{ 435, 404, 53 },
		{ 435, 405, 53 },
		{ 435, 406, 53 },
		{ 435, 407, 53 },
		{ 435, 408, 53 },
		{ 435, 409, 53 },
		{ 435, 410, 53 },
		{ 435, 411, 53 },
		{ 435, 412, 53 },
		{ 435, 413, 53 },
		{ 435, 414, 53 },
		{ 435, 415, 53 },
		{ 435, 416, 53 },
		{ 435, 417, 53 },
		{ 435, 418, 53 },
		{ 435, 419, 53 },
		{ 435, 420, 53 },
		{ 435, 421, 53 },
		{ 435, 422, 53 },
		{ 435, 423, 53 },
		{ 435, 424, 53 },
		{ 435, 425, 53 },
		{ 435, 426, 53 },
		{ 435, 427, 53 },
		{ 435, 428, 53 },
		{ 435, 429, 53 },
		{ 435, 430, 53 },
		{ 435, 431, 53 },
		{ 435, 432, 53 },
		{ 435, 433, 53 },
		{ 435, 434, 53 },
		{ 436, 392, 53 },
		{ 436, 393, 53 },
		{ 436, 394, 53 },
		{ 436, 395, 53 },
		{ 436, 396, 53 },
		{ 436, 397, 53 },
		{ 436, 398, 53 },
		{ 436, 399, 53 },
		{ 436, 400, 53 },
		{ 436, 401, 53 },
		{ 436, 402, 53 },
		{ 436, 403, 53 },
		{ 436, 404, 53 },
		{ 436, 405, 53 },
		{ 436, 406, 53 },
		{ 436, 407, 53 },
		{ 436, 408, 53 },
		{ 436, 409, 53 },
		{ 436, 410, 53 },
		{ 436, 411, 53 },
		{ 436, 412, 53 },
		{ 436, 413, 53 },
		{ 436, 414, 53 },
		{ 436, 415, 53 },
		{ 436, 416, 53 },
		{ 436, 417, 53 },
		{ 436, 418, 53 },
		{ 436, 419, 53 },
		{ 436, 420, 53 },
		{ 436, 421, 53 },
		{ 436, 422, 53 },
		{ 436, 423, 53 },
		{ 436, 424, 53 },
		{ 436, 425, 53 },
		{ 436, 426, 53 },
		{ 436, 427, 53 },
		{ 436, 428, 53 },
		{ 436, 429, 53 },
		{ 436, 430, 53 },
		{ 436, 431, 53 },
		{ 436, 432, 53 },
		{ 436, 433, 53 },
		{ 436, 434, 53 },
		{ 436, 435, 53 },
		{ 437, 393, 53 },
		{ 437, 394, 53 },
		{ 437, 395, 53 },
		{ 437, 396, 53 },
		{ 437, 397, 53 },
		{ 437, 398, 53 },
		{ 437, 399, 53 },
		{ 437, 400, 53 },
		{ 437, 401, 53 },
		{ 437, 402, 53 },
		{ 437, 403, 53 },
		{ 437, 404, 53 },
		{ 437, 405, 53 },
		{ 437, 406, 53 },
		{ 437, 407, 53 },
		{ 437, 408, 53 },
		{ 437, 409, 53 },
		{ 437, 410, 53 },
		{ 437, 411, 53 },
		{ 437, 412, 53 },
		{ 437, 413, 53 },
		{ 437, 414, 53 },
		{ 437, 415, 53 },
		{ 437, 416, 53 },
		{ 437, 417, 53 },
		{ 437, 418, 53 },
		{ 437, 419, 53 },
		{ 437, 420, 53 },
		{ 437, 421, 53 },
		{ 437, 422, 53 },
		{ 437, 423, 53 },
		{ 437, 424, 53 },
		{ 437, 425, 53 },
		{ 437, 426, 53 },
		{ 437, 427, 53 },
		{ 437, 428, 53 },
		{ 437, 429, 53 },
		{ 437, 430, 53 },
		{ 437, 431, 53 },
		{ 437, 432, 53 },
		{ 437, 433, 53 },
		{ 437, 434, 53 },
		{ 437, 435, 53 },
		{ 437, 436, 53 },
		{ 438, 393, 53 },
		{ 438, 394, 53 },
		{ 438, 395, 53 },
		{ 438, 396, 53 },
		{ 438, 397, 53 },
		{ 438, 398, 53 },
		{ 438, 399, 53 },
		{ 438, 400, 53 },
		{ 438, 401, 53 },
		{ 438, 402, 53 },
		{ 438, 403, 53 },
		{ 438, 404, 53 },
		{ 438, 405, 53 },
		{ 438, 406, 53 },
		{ 438, 407, 53 },
		{ 438, 408, 53 },
		{ 438, 409, 53 },
		{ 438, 410, 53 },
		{ 438, 411, 53 },
		{ 438, 412, 53 },
		{ 438, 413, 53 },
		{ 438, 414, 53 },
		{ 438, 415, 53 },
		{ 438, 416, 53 },
		{ 438, 417, 53 },
		{ 438, 418, 53 },
		{ 438, 419, 53 },
		{ 438, 420, 53 },
		{ 438, 421, 53 },
		{ 438, 422, 53 },
		{ 438, 423, 53 },
		{ 438, 424, 53 },
		{ 438, 425, 53 },
		{ 438, 426, 53 },
		{ 438, 427, 53 },
		{ 438, 428, 53 },
		{ 438, 429, 53 },
		{ 438, 430, 53 },
		{ 438, 431, 53 },
		{ 438, 432, 53 },
		{ 438, 433, 53 },
		{ 438, 434, 53 },
		{ 438, 435, 53 },
		{ 438, 436, 53 },
		{ 438, 437, 53 },
		{ 439, 394, 53 },
		{ 439, 395, 53 },
		{ 439, 396, 53 },
		{ 439, 397, 53 },
		{ 439, 398, 53 },
		{ 439, 399, 53 },
		{ 439, 400, 53 },
		{ 439, 401, 53 },
		{ 439, 402, 53 },
		{ 439, 403, 53 },
		{ 439, 404, 53 },
		{ 439, 405, 53 },
		{ 439, 406, 53 },
		{ 439, 407, 53 },
		{ 439, 408, 53 },
		{ 439, 409, 53 },
		{ 439, 410, 53 },
		{ 439, 411, 53 },
		{ 439, 412, 53 },
		{ 439, 413, 53 },
		{ 439, 414, 53 },
		{ 439, 415, 53 },
		{ 439, 416, 53 },
		{ 439, 417, 53 },
		{ 439, 418, 53 },
		{ 439, 419, 53 },
		{ 439, 420, 53 },
		{ 439, 421, 53 },
		{ 439, 422, 53 },
		{ 439, 423, 53 },
		{ 439, 424, 53 },
		{ 439, 425, 53 },
		{ 439, 426, 53 },
		{ 439, 427, 53 },
		{ 439, 428, 53 },
		{ 439, 429, 53 },
		{ 439, 430, 53 },
		{ 439, 431, 53 },
		{ 439, 432, 53 },
		{ 439, 433, 53 },
		{ 439, 434, 53 },
		{ 439, 435, 53 },
		{ 439, 436, 53 },
		{ 439, 437, 53 },
		{ 439, 438, 53 },
		{ 440, 395, 53 },
		{ 440, 396, 53 },
		{ 440, 397, 53 },
		{ 440, 398, 53 },
		{ 440, 399, 53 },
		{ 440, 400, 53 },
		{ 440, 401, 53 },
		{ 440, 402, 53 },
		{ 440, 403, 53 },
		{ 440, 404, 53 },
		{ 440, 405, 53 },
		{ 440, 406, 53 },
		{ 440, 407, 53 },
		{ 440, 408, 53 },
		{ 440, 409, 53 },
		{ 440, 410, 53 },
		{ 440, 411, 53 },
		{ 440, 412, 53 },
		{ 440, 413, 53 },
		{ 440, 414, 53 },
		{ 440, 415, 53 },
		{ 440, 416, 53 },
		{ 440, 417, 53 },
		{ 440, 418, 53 },
		{ 440, 419, 53 },
		{ 440, 420, 53 },
		{ 440, 421, 53 },
		{ 440, 422, 53 },
		{ 440, 423, 53 },
		{ 440, 424, 53 },
		{ 440, 425, 53 },
		{ 440, 426, 53 },
		{ 440, 427, 53 },
		{ 440, 428, 53 },
		{ 440, 429, 53 },
		{ 440, 430, 53 },
		{ 440, 431, 53 },
		{ 440, 432, 53 },
		{ 440, 433, 53 },
		{ 440, 434, 53 },
		{ 440, 435, 53 },
		{ 440, 436, 53 },
		{ 440, 437, 53 },
		{ 440, 438, 53 },
		{ 440, 439, 53 },
		{ 441, 396, 53 },
		{ 441, 397, 53 },
		{ 441, 398, 53 },
		{ 441, 399, 53 },
		{ 441, 400, 53 },
		{ 441, 401, 53 },
		{ 441, 402, 53 },
		{ 441, 403, 53 },
		{ 441, 404, 53 },
		{ 441, 405, 53 },
		{ 441, 406, 53 },
		{ 441, 407, 53 },
		{ 441, 408, 53 },
		{ 441, 409, 53 },
		{ 441, 410, 53 },
		{ 441, 411, 53 },
		{ 441, 412, 53 },
		{ 441, 413, 53 },
		{ 441, 414, 53 },
		{ 441, 415, 53 },
		{ 441, 416, 53 },
		{ 441, 417, 53 },
		{ 441, 418, 53 },
		{ 441, 419, 53 },
		{ 441, 420, 53 },
		{ 441, 421, 53 },
		{ 441, 422, 53 },
		{ 441, 423, 53 },
		{ 441, 424, 53 },
		{ 441, 425, 53 },
		{ 441, 426, 53 },
		{ 441, 427, 53 },
		{ 441, 428, 53 },
		{ 441, 429, 53 },
		{ 441, 430, 53 },
		{ 441, 431, 53 },
		{ 441, 432, 53 },
		{ 441, 433, 53 },
		{ 441, 434, 53 },
		{ 441, 435, 53 },
		{ 441, 436, 53 },
		{ 441, 437, 53 },
		{ 441, 438, 53 },
		{ 441, 439, 53 },
		{ 441, 440, 53 },
		{ 442, 396, 53 },
		{ 442, 397, 53 },
		{ 442, 398, 53 },
		{ 442, 399, 53 },
		{ 442, 400, 53 },
		{ 442, 401, 53 },
		{ 442, 402, 53 },
		{ 442, 403, 53 },
		{ 442, 404, 53 },
		{ 442, 405, 53 },
		{ 442, 406, 53 },
		{ 442, 407, 53 },
		{ 442, 408, 53 },
		{ 442, 409, 53 },
		{ 442, 410, 53 },
		{ 442, 411, 53 },
		{ 442, 412, 53 },
		{ 442, 413, 53 },
		{ 442, 414, 53 },
		{ 442, 415, 53 },
		{ 442, 416, 53 },
		{ 442, 417, 53 },
		{ 442, 418, 53 },
		{ 442, 419, 53 },
		{ 442, 420, 53 },
		{ 442, 421, 53 },
		{ 442, 422, 53 },
		{ 442, 423, 53 },
		{ 442, 424, 53 },
		{ 442, 425, 53 },
		{ 442, 426, 53 },
		{ 442, 427, 53 },
		{ 442, 428, 53 },
		{ 442, 429, 53 },
		{ 442, 430, 53 },
		{ 442, 431, 53 },
		{ 442, 432, 53 },
		{ 442, 433, 53 },
		{ 442, 434, 53 },
		{ 442, 435, 53 },
		{ 442, 436, 53 },
		{ 442, 437, 53 },
		{ 442, 438, 53 },
		{ 442, 439, 53 },
		{ 442, 440, 53 },
		{ 442, 441, 53 },
		{ 443, 397, 53 },
		{ 443, 398, 53 },
		{ 443, 399, 53 },
		{ 443, 400, 53 },
		{ 443, 401, 53 },
		{ 443, 402, 53 },
		{ 443, 403, 53 },
		{ 443, 404, 53 },
		{ 443, 405, 53 },
		{ 443, 406, 53 },
		{ 443, 407, 53 },
		{ 443, 408, 53 },
		{ 443, 409, 53 },
		{ 443, 410, 53 },
		{ 443, 411, 53 },
		{ 443, 412, 53 },
		{ 443, 413, 53 },
		{ 443, 414, 53 },
		{ 443, 415, 53 },
		{ 443, 416, 53 },
		{ 443, 417, 53 },
		{ 443, 418, 53 },
		{ 443, 419, 53 },
		{ 443, 420, 53 },
		{ 443, 421, 53 },
		{ 443, 422, 53 },
		{ 443, 423, 53 },
		{ 443, 424, 53 },
		{ 443, 425, 53 },
		{ 443, 426, 53 },
		{ 443, 427, 53 },
		{ 443, 428, 53 },
		{ 443, 429, 53 },
		{ 443, 430, 53 },
		{ 443, 431, 53 },
		{ 443, 432, 53 },
		{ 443, 433, 53 },
		{ 443, 434, 53 },
		{ 443, 435, 53 },
		{ 443, 436, 53 },
		{ 443, 437, 53 },
		{ 443, 438, 53 },
		{ 443, 439, 53 },
		{ 443, 440, 53 },
		{ 443, 441, 53 },
		{ 443, 442, 53 },
		{ 444, 398, 53 },
		{ 444, 399, 53 },
		{ 444, 400, 53 },
		{ 444, 401, 53 },
		{ 444, 402, 53 },
		{ 444, 403, 53 },
		{ 444, 404, 53 },
		{ 444, 405, 53 },
		{ 444, 406, 53 },
		{ 444, 407, 53 },
		{ 444, 408, 53 },
		{ 444, 409, 53 },
		{ 444, 410, 53 },
		{ 444, 411, 53 },
		{ 444, 412, 53 },
		{ 444, 413, 53 },
		{ 444, 414, 53 },
		{ 444, 415, 53 },
		{ 444, 416, 53 },
		{ 444, 417, 53 },
		{ 444, 418, 53 },
		{ 444, 419, 53 },
		{ 444, 420, 53 },
		{ 444, 421, 53 },
		{ 444, 422, 53 },
		{ 444, 423, 53 },
		{ 444, 424, 53 },
		{ 444, 425, 53 },
		{ 444, 426, 53 },
		{ 444, 427, 53 },
		{ 444, 428, 53 },
		{ 444, 429, 53 },
		{ 444, 430, 53 },
		{ 444, 431, 53 },
		{ 444, 432, 53 },
		{ 444, 433, 53 },
		{ 444, 434, 53 },
		{ 444, 435, 53 },
		{ 444, 436, 53 },
		{ 444, 437, 53 },
		{ 444, 438, 53 },
		{ 444, 439, 53 },
		{ 444, 440, 53 },
		{ 444, 441, 53 },
		{ 444, 442, 53 },
		{ 444, 443, 53 },
		{ 445, 399, 53 },
		{ 445, 400, 53 },
		{ 445, 401, 53 },
		{ 445, 402, 53 },
		{ 445, 403, 53 },
		{ 445, 404, 53 },
		{ 445, 405, 53 },
		{ 445, 406, 53 },
		{ 445, 407, 53 },
		{ 445, 408, 53 },
		{ 445, 409, 53 },
		{ 445, 410, 53 },
		{ 445, 411, 53 },
		{ 445, 412, 53 },
		{ 445, 413, 53 },
		{ 445, 414, 53 },
		{ 445, 415, 53 },
		{ 445, 416, 53 },
		{ 445, 417, 53 },
		{ 445, 418, 53 },
		{ 445, 419, 53 },
		{ 445, 420, 53 },
		{ 445, 421, 53 },
		{ 445, 422, 53 },
		{ 445, 423, 53 },
		{ 445, 424, 53 },
		{ 445, 425, 53 },
		{ 445, 426, 53 },
		{ 445, 427, 53 },
		{ 445, 428, 53 },
		{ 445, 429, 53 },
		{ 445, 430, 53 },
		{ 445, 431, 53 },
		{ 445, 432, 53 },
		{ 445, 433, 53 },
		{ 445, 434, 53 },
		{ 445, 435, 53 },
		{ 445, 436, 53 },
		{ 445, 437, 53 },
		{ 445, 438, 53 },
		{ 445, 439, 53 },
		{ 445, 440, 53 },
		{ 445, 441, 53 },
		{ 445, 442, 53 },
		{ 445, 443, 53 },
		{ 445, 444, 53 },
		{ 446, 399, 53 },
		{ 446, 400, 53 },
		{ 446, 401, 53 },
		{ 446, 402, 53 },
		{ 446, 403, 53 },
		{ 446, 404, 53 },
		{ 446, 405, 53 },
		{ 446, 406, 53 },
		{ 446, 407, 53 },
		{ 446, 408, 53 },
		{ 446, 409, 53 },
		{ 446, 410, 53 },
		{ 446, 411, 53 },
		{ 446, 412, 53 },
		{ 446, 413, 53 },
		{ 446, 414, 53 },
		{ 446, 415, 53 },
		{ 446, 416, 53 },
		{ 446, 417, 53 },
		{ 446, 418, 53 },
		{ 446, 419, 53 },
		{ 446, 420, 53 },
		{ 446, 421, 53 },
		{ 446, 422, 53 },
		{ 446, 423, 53 },
		{ 446, 424, 53 },
		{ 446, 425, 53 },
		{ 446, 426, 53 },
		{ 446, 427, 53 },
		{ 446, 428, 53 },
		{ 446, 429, 53 },
		{ 446, 430, 53 },
		{ 446, 431, 53 },
		{ 446, 432, 53 },
		{ 446, 433, 53 },
		{ 446, 434, 53 },
		{ 446, 435, 53 },
		{ 446, 436, 53 },
		{ 446, 437, 53 },
		{ 446, 438, 53 },
		{ 446, 439, 53 },
		{ 446, 440, 53 },
		{ 446, 441, 53 },
		{ 446, 442, 53 },
		{ 446, 443, 53 },
		{ 446, 444, 53 },
		{ 446, 445, 53 },
		{ 447, 400, 53 },
		{ 447, 401, 53 },
		{ 447, 402, 53 },
		{ 447, 403, 53 },
		{ 447, 404, 53 },
		{ 447, 405, 53 },
		{ 447, 406, 53 },
		{ 447, 407, 53 },
		{ 447, 408, 53 },
		{ 447, 409, 53 },
		{ 447, 410, 53 },
		{ 447, 411, 53 },
		{ 447, 412, 53 },
		{ 447, 413, 53 },
		{ 447, 414, 53 },
		{ 447, 415, 53 },
		{ 447, 416, 53 },
		{ 447, 417, 53 },
		{ 447, 418, 53 },
		{ 447, 419, 53 },
		{ 447, 420, 53 },
		{ 447, 421, 53 },
		{ 447, 422, 53 },
		{ 447, 423, 53 },
		{ 447, 424, 53 },
		{ 447, 425, 53 },
		{ 447, 426, 53 },
		{ 447, 427, 53 },
		{ 447, 428, 53 },
		{ 447, 429, 53 },
		{ 447, 430, 53 },
		{ 447, 431, 53 },
		{ 447, 432, 53 },
		{ 447, 433, 53 },
		{ 447, 434, 53 },
		{ 447, 435, 53 },
		{ 447, 436, 53 },
		{ 447, 437, 53 },
		{ 447, 438, 53 },
		{ 447, 439, 53 },
		{ 447, 440, 53 },
		{ 447, 441, 53 },
		{ 447, 442, 53 },
		{ 447, 443, 53 },
		{ 447, 444, 53 },
		{ 447, 445, 53 },
		{ 447, 446, 53 },
		{ 448, 401, 53 },
		{ 448, 402, 53 },
		{ 448, 403, 53 },
		{ 448, 404, 53 },
		{ 448, 405, 53 },
		{ 448, 406, 53 },
		{ 448, 407, 53 },
		{ 448, 408, 53 },
		{ 448, 409, 53 },
		{ 448, 410, 53 },
		{ 448, 411, 53 },
		{ 448, 412, 53 },
		{ 448, 413, 53 },
		{ 448, 414, 53 },
		{ 448, 415, 53 },
		{ 448, 416, 53 },
		{ 448, 417, 53 },
		{ 448, 418, 53 },
		{ 448, 419, 53 },
		{ 448, 420, 53 },
		{ 448, 421, 53 },
		{ 448, 422, 53 },
		{ 448, 423, 53 },
		{ 448, 424, 53 },
		{ 448, 425, 53 },
		{ 448, 426, 53 },
		{ 448, 427, 53 },
		{ 448, 428, 53 },
		{ 448, 429, 53 },
		{ 448, 430, 53 },
		{ 448, 431, 53 },
		{ 448, 432, 53 },
		{ 448, 433, 53 },
		{ 448, 434, 53 },
		{ 448, 435, 53 },
		{ 448, 436, 53 },
		{ 448, 437, 53 },
		{ 448, 438, 53 },
		{ 448, 439, 53 },
		{ 448, 440, 53 },
		{ 448, 441, 53 },
		{ 448, 442, 53 },
		{ 448, 443, 53 },
		{ 448, 444, 53 },
		{ 448, 445, 53 },
		{ 448, 446, 53 },
		{ 448, 447, 53 },
		{ 449, 402, 53 },
		{ 449, 403, 53 },
		{ 449, 404, 53 },
		{ 449, 405, 53 },
		{ 449, 406, 53 },
		{ 449, 407, 53 },
		{ 449, 408, 53 },
		{ 449, 409, 53 },
		{ 449, 410, 53 },
		{ 449, 411, 53 },
		{ 449, 412, 53 },
		{ 449, 413, 53 },
		{ 449, 414, 53 },
		{ 449, 415, 53 },
		{ 449, 416, 53 },
		{ 449, 417, 53 },
		{ 449, 418, 53 },
		{ 449, 419, 53 },
		{ 449, 420, 53 },
		{ 449, 421, 53 },
		{ 449, 422, 53 },
		{ 449, 423, 53 },
		{ 449, 424, 53 },
		{ 449, 425, 53 },
		{ 449, 426, 53 },
		{ 449, 427, 53 },
		{ 449, 428, 53 },
		{ 449, 429, 53 },
		{ 449, 430, 53 },
		{ 449, 431, 53 },
		{ 449, 432, 53 },
		{ 449, 433, 53 },
		{ 449, 434, 53 },
		{ 449, 435, 53 },
		{ 449, 436, 53 },
		{ 449, 437, 53 },
		{ 449, 438, 53 },
		{ 449, 439, 53 },
		{ 449, 440, 53 },
		{ 449, 441, 53 },
		{ 449, 442, 53 },
		{ 449, 443, 53 },
		{ 449, 444, 53 },
		{ 449, 445, 53 },
		{ 449, 446, 53 },
		{ 449, 447, 53 },
		{ 449, 448, 53 },
		{ 450, 402, 53 },
		{ 450, 403, 53 },
		{ 450, 404, 53 },
		{ 450, 405, 53 },
		{ 450, 406, 53 },
		{ 450, 407, 53 },
		{ 450, 408, 53 },
		{ 450, 409, 53 },
		{ 450, 410, 53 },
		{ 450, 411, 53 },
		{ 450, 412, 53 },
		{ 450, 413, 53 },
		{ 450, 414, 53 },
		{ 450, 415, 53 },
		{ 450, 416, 53 },
		{ 450, 417, 53 },
		{ 450, 418, 53 },
		{ 450, 419, 53 },
		{ 450, 420, 53 },
		{ 450, 421, 53 },
		{ 450, 422, 53 },
		{ 450, 423, 53 },
		{ 450, 424, 53 },
		{ 450, 425, 53 },
		{ 450, 426, 53 },
		{ 450, 427, 53 },
		{ 450, 428, 53 },
		{ 450, 429, 53 },
		{ 450, 430, 53 },
		{ 450, 431, 53 },
		{ 450, 432, 53 },
		{ 450, 433, 53 },
		{ 450, 434, 53 },
		{ 450, 435, 53 },
		{ 450, 436, 53 },
		{ 450, 437, 53 },
		{ 450, 438, 53 },
		{ 450, 439, 53 },
		{ 450, 440, 53 },
		{ 450, 441, 53 },
		{ 450, 442, 53 },
		{ 450, 443, 53 },
		{ 450, 444, 53 },
		{ 450, 445, 53 },
		{ 450, 446, 53 },
		{ 450, 447, 53 },
		{ 450, 448, 53 },
		{ 450, 449, 53 },
		{ 451, 403, 53 },
		{ 451, 404, 53 },
		{ 451, 405, 53 },
		{ 451, 406, 53 },
		{ 451, 407, 53 },
		{ 451, 408, 53 },
		{ 451, 409, 53 },
		{ 451, 410, 53 },
		{ 451, 411, 53 },
		{ 451, 412, 53 },
		{ 451, 413, 53 },
		{ 451, 414, 53 },
		{ 451, 415, 53 },
		{ 451, 416, 53 },
		{ 451, 417, 53 },
		{ 451, 418, 53 },
		{ 451, 419, 53 },
		{ 451, 420, 53 },
		{ 451, 421, 53 },
		{ 451, 422, 53 },
		{ 451, 423, 53 },
		{ 451, 424, 53 },
		{ 451, 425, 53 },
		{ 451, 426, 53 },
		{ 451, 427, 53 },
		{ 451, 428, 53 },
		{ 451, 429, 53 },
		{ 451, 430, 53 },
		{ 451, 431, 53 },
		{ 451, 432, 53 },
		{ 451, 433, 53 },
		{ 451, 434, 53 },
		{ 451, 435, 53 },
		{ 451, 436, 53 },
		{ 451, 437, 53 },
		{ 451, 438, 53 },
		{ 451, 439, 53 },
		{ 451, 440, 53 },
		{ 451, 441, 53 },
		{ 451, 442, 53 },
		{ 451, 443, 53 },
		{ 451, 444, 53 },
		{ 451, 445, 53 },
		{ 451, 446, 53 },
		{ 451, 447, 53 },
		{ 451, 448, 53 },
		{ 451, 449, 53 },
		{ 451, 450, 53 },
		{ 452, 404, 53 },
		{ 452, 405, 53 },
		{ 452, 406, 53 },
		{ 452, 407, 53 },
		{ 452, 408, 53 },
		{ 452, 409, 53 },
		{ 452, 410, 53 },
		{ 452, 411, 53 },
		{ 452, 412, 53 },
		{ 452, 413, 53 },
		{ 452, 414, 53 },
		{ 452, 415, 53 },
		{ 452, 416, 53 },
		{ 452, 417, 53 },
		{ 452, 418, 53 },
		{ 452, 419, 53 },
		{ 452, 420, 53 },
		{ 452, 421, 53 },
		{ 452, 422, 53 },
		{ 452, 423, 53 },
		{ 452, 424, 53 },
		{ 452, 425, 53 },
		{ 452, 426, 53 },
		{ 452, 427, 53 },
		{ 452, 428, 53 },
		{ 452, 429, 53 },
		{ 452, 430, 53 },
		{ 452, 431, 53 },
		{ 452, 432, 53 },
		{ 452, 433, 53 },
		{ 452, 434, 53 },
		{ 452, 435, 53 },
		{ 452, 436, 53 },
		{ 452, 437, 53 },
		{ 452, 438, 53 },
		{ 452, 439, 53 },
		{ 452, 440, 53 },
		{ 452, 441, 53 },
		{ 452, 442, 53 },
		{ 452, 443, 53 },
		{ 452, 444, 53 },
		{ 452, 445, 53 },
		{ 452, 446, 53 },
		{ 452, 447, 53 },
		{ 452, 448, 53 },
		{ 452, 449, 53 },
		{ 452, 450, 53 },
		{ 452, 451, 53 },
		{ 453, 405, 53 },
		{ 453, 406, 53 },
		{ 453, 407, 53 },
		{ 453, 408, 53 },
		{ 453, 409, 53 },
		{ 453, 410, 53 },
		{ 453, 411, 53 },
		{ 453, 412, 53 },
		{ 453, 413, 53 },
		{ 453, 414, 53 },
		{ 453, 415, 53 },
		{ 453, 416, 53 },
		{ 453, 417, 53 },
		{ 453, 418, 53 },
		{ 453, 419, 53 },
		{ 453, 420, 53 },
		{ 453, 421, 53 },
		{ 453, 422, 53 },
		{ 453, 423, 53 },
		{ 453, 424, 53 },
		{ 453, 425, 53 },
		{ 453, 426, 53 },
		{ 453, 427, 53 },
		{ 453, 428, 53 },
		{ 453, 429, 53 },
		{ 453, 430, 53 },
		{ 453, 431, 53 },
		{ 453, 432, 53 },
		{ 453, 433, 53 },
		{ 453, 434, 53 },
		{ 453, 435, 53 },
		{ 453, 436, 53 },
		{ 453, 437, 53 },
		{ 453, 438, 53 },
		{ 453, 439, 53 },
		{ 453, 440, 53 },
		{ 453, 441, 53 },
		{ 453, 442, 53 },
		{ 453, 443, 53 },
		{ 453, 444, 53 },
		{ 453, 445, 53 },
		{ 453, 446, 53 },
		{ 453, 447, 53 },
		{ 453, 448, 53 },
		{ 453, 449, 53 },
		{ 453, 450, 53 },
		{ 453, 451, 53 },
		{ 453, 452, 53 },
		{ 454, 405, 53 },
		{ 454, 406, 53 },
		{ 454, 407, 53 },
		{ 454, 408, 53 },
		{ 454, 409, 53 },
		{ 454, 410, 53 },
		{ 454, 411, 53 },
		{ 454, 412, 53 },
		{ 454, 413, 53 },
		{ 454, 414, 53 },
		{ 454, 415, 53 },
		{ 454, 416, 53 },
		{ 454, 417, 53 },
		{ 454, 418, 53 },
		{ 454, 419, 53 },
		{ 454, 420, 53 },
		{ 454, 421, 53 },
		{ 454, 422, 53 },
		{ 454, 423, 53 },
		{ 454, 424, 53 },
		{ 454, 425, 53 },
		{ 454, 426, 53 },
		{ 454, 427, 53 },
		{ 454, 428, 53 },
		{ 454, 429, 53 },
		{ 454, 430, 53 },
		{ 454, 431, 53 },
		{ 454, 432, 53 },
		{ 454, 433, 53 },
		{ 454, 434, 53 },
		{ 454, 435, 53 },
		{ 454, 436, 53 },
		{ 454, 437, 53 },
		{ 454, 438, 53 },
		{ 454, 439, 53 },
		{ 454, 440, 53 },
		{ 454, 441, 53 },
		{ 454, 442, 53 },
		{ 454, 443, 53 },
		{ 454, 444, 53 },
		{ 454, 445, 53 },
		{ 454, 446, 53 },
		{ 454, 447, 53 },
		{ 454, 448, 53 },
		{ 454, 449, 53 },
		{ 454, 450, 53 },
		{ 454, 451, 53 },
		{ 454, 452, 53 },
		{ 454, 453, 53 },
		{ 455, 406, 53 },
		{ 455, 407, 53 },
		{ 455, 408, 53 },
		{ 455, 409, 53 },
		{ 455, 410, 53 },
		{ 455, 411, 53 },
		{ 455, 412, 53 },
		{ 455, 413, 53 },
		{ 455, 414, 53 },
		{ 455, 415, 53 },
		{ 455, 416, 53 },
		{ 455, 417, 53 },
		{ 455, 418, 53 },
		{ 455, 419, 53 },
		{ 455, 420, 53 },
		{ 455, 421, 53 },
		{ 455, 422, 53 },
		{ 455, 423, 53 },
		{ 455, 424, 53 },
		{ 455, 425, 53 },
		{ 455, 426, 53 },
		{ 455, 427, 53 },
		{ 455, 428, 53 },
		{ 455, 429, 53 },
		{ 455, 430, 53 },
		{ 455, 431, 53 },
		{ 455, 432, 53 },
		{ 455, 433, 53 },
		{ 455, 434, 53 },
		{ 455, 435, 53 },
		{ 455, 436, 53 },
		{ 455, 437, 53 },
		{ 455, 438, 53 },
		{ 455, 439, 53 },
		{ 455, 440, 53 },
		{ 455, 441, 53 },
		{ 455, 442, 53 },
		{ 455, 443, 53 },
		{ 455, 444, 53 },
		{ 455, 445, 53 },
		{ 455, 446, 53 },
		{ 455, 447, 53 },
		{ 455, 448, 53 },
		{ 455, 449, 53 },
		{ 455, 450, 53 },
		{ 455, 451, 53 },
		{ 455, 452, 53 },
		{ 455, 453, 53 },
		{ 455, 454, 53 },
		{ 456, 407, 53 },
		{ 456, 408, 53 },
		{ 456, 409, 53 },
		{ 456, 410, 53 },
		{ 456, 411, 53 },
		{ 456, 412, 53 },
		{ 456, 413, 53 },
		{ 456, 414, 53 },
		{ 456, 415, 53 },
		{ 456, 416, 53 },
		{ 456, 417, 53 },
		{ 456, 418, 53 },
		{ 456, 419, 53 },
		{ 456, 420, 53 },
		{ 456, 421, 53 },
		{ 456, 422, 53 },
		{ 456, 423, 53 },
		{ 456, 424, 53 },
		{ 456, 425, 53 },
		{ 456, 426, 53 },
		{ 456, 427, 53 },
		{ 456, 428, 53 },
		{ 456, 429, 53 },
		{ 456, 430, 53 },
		{ 456, 431, 53 },
		{ 456, 432, 53 },
		{ 456, 433, 53 },
		{ 456, 434, 53 },
		{ 456, 435, 53 },
		{ 456, 436, 53 },
		{ 456, 437, 53 },
		{ 456, 438, 53 },
		{ 456, 439, 53 },
		{ 456, 440, 53 },
		{ 456, 441, 53 },
		{ 456, 442, 53 },
		{ 456, 443, 53 },
		{ 456, 444, 53 },
		{ 456, 445, 53 },
		{ 456, 446, 53 },
		{ 456, 447, 53 },
		{ 456, 448, 53 },
		{ 456, 449, 53 },
		{ 456, 450, 53 },
		{ 456, 451, 53 },
		{ 456, 452, 53 },
		{ 456, 453, 53 },
		{ 456, 454, 53 },
		{ 456, 455, 53 },
		{ 457, 408, 53 },
		{ 457, 409, 53 },
		{ 457, 410, 53 },
		{ 457, 411, 53 },
		{ 457, 412, 53 },
		{ 457, 413, 53 },
		{ 457, 414, 53 },
		{ 457, 415, 53 },
		{ 457, 416, 53 },
		{ 457, 417, 53 },
		{ 457, 418, 53 },
		{ 457, 419, 53 },
		{ 457, 420, 53 },
		{ 457, 421, 53 },
		{ 457, 422, 53 },
		{ 457, 423, 53 },
		{ 457, 424, 53 },
		{ 457, 425, 53 },
		{ 457, 426, 53 },
		{ 457, 427, 53 },
		{ 457, 428, 53 },
		{ 457, 429, 53 },
		{ 457, 430, 53 },
		{ 457, 431, 53 },
		{ 457, 432, 53 },
		{ 457, 433, 53 },
		{ 457, 434, 53 },
		{ 457, 435, 53 },
		{ 457, 436, 53 },
		{ 457, 437, 53 },
		{ 457, 438, 53 },
		{ 457, 439, 53 },
		{ 457, 440, 53 },
		{ 457, 441, 53 },
		{ 457, 442, 53 },
		{ 457, 443, 53 },
		{ 457, 444, 53 },
		{ 457, 445, 53 },
		{ 457, 446, 53 },
		{ 457, 447, 53 },
		{ 457, 448, 53 },
		{ 457, 449, 53 },
		{ 457, 450, 53 },
		{ 457, 451, 53 },
		{ 457, 452, 53 },
		{ 457, 453, 53 },
		{ 457, 454, 53 },
		{ 457, 455, 53 },
		{ 457, 456, 53 },
		{ 458, 408, 53 },
		{ 458, 409, 53 },
		{ 458, 410, 53 },
		{ 458, 411, 53 },
		{ 458, 412, 53 },
		{ 458, 413, 53 },
		{ 458, 414, 53 },
		{ 458, 415, 53 },
		{ 458, 416, 53 },
		{ 458, 417, 53 },
		{ 458, 418, 53 },
		{ 458, 419, 53 },
		{ 458, 420, 53 },
		{ 458, 421, 53 },
		{ 458, 422, 53 },
		{ 458, 423, 53 },
		{ 458, 424, 53 },
		{ 458, 425, 53 },
		{ 458, 426, 53 },
		{ 458, 427, 53 },
		{ 458, 428, 53 },
		{ 458, 429, 53 },
		{ 458, 430, 53 },
		{ 458, 431, 53 },
		{ 458, 432, 53 },
		{ 458, 433, 53 },
		{ 458, 434, 53 },
		{ 458, 435, 53 },
		{ 458, 436, 53 },
		{ 458, 437, 53 },
		{ 458, 438, 53 },
		{ 458, 439, 53 },
		{ 458, 440, 53 },
		{ 458, 441, 53 },
		{ 458, 442, 53 },
		{ 458, 443, 53 },
		{ 458, 444, 53 },
		{ 458, 445, 53 },
		{ 458, 446, 53 },
		{ 458, 447, 53 },
		{ 458, 448, 53 },
		{ 458, 449, 53 },
		{ 458, 450, 53 },
		{ 458, 451, 53 },
		{ 458, 452, 53 },
		{ 458, 453, 53 },
		{ 458, 454, 53 },
		{ 458, 455, 53 },
		{ 458, 456, 53 },
		{ 458, 457, 53 },
		{ 459, 409, 53 },
		{ 459, 410, 53 },
		{ 459, 411, 53 },
		{ 459, 412, 53 },
		{ 459, 413, 53 },
		{ 459, 414, 53 },
		{ 459, 415, 53 },
		{ 459, 416, 53 },
		{ 459, 417, 53 },
		{ 459, 418, 53 },
		{ 459, 419, 53 },
		{ 459, 420, 53 },
		{ 459, 421, 53 },
		{ 459, 422, 53 },
		{ 459, 423, 53 },
		{ 459, 424, 53 },
		{ 459, 425, 53 },
		{ 459, 426, 53 },
		{ 459, 427, 53 },
		{ 459, 428, 53 },
		{ 459, 429, 53 },
		{ 459, 430, 53 },
		{ 459, 431, 53 },
		{ 459, 432, 53 },
		{ 459, 433, 53 },
		{ 459, 434, 53 },
		{ 459, 435, 53 },
		{ 459, 436, 53 },
		{ 459, 437, 53 },
		{ 459, 438, 53 },
		{ 459, 439, 53 },
		{ 459, 440, 53 },
		{ 459, 441, 53 },
		{ 459, 442, 53 },
		{ 459, 443, 53 },
		{ 459, 444, 53 },
		{ 459, 445, 53 },
		{ 459, 446, 53 },
		{ 459, 447, 53 },
		{ 459, 448, 53 },
		{ 459, 449, 53 },
		{ 459, 450, 53 },
		{ 459, 451, 53 },
		{ 459, 452, 53 },
		{ 459, 453, 53 },
		{ 459, 454, 53 },
		{ 459, 455, 53 },
		{ 459, 456, 53 },
		{ 459, 457, 53 },
		{ 459, 458, 53 },
		{ 460, 410, 53 },
		{ 460, 411, 53 },
		{ 460, 412, 53 },
		{ 460, 413, 53 },
		{ 460, 414, 53 },
		{ 460, 415, 53 },
		{ 460, 416, 53 },
		{ 460, 417, 53 },
		{ 460, 418, 53 },
		{ 460, 419, 53 },
		{ 460, 420, 53 },
		{ 460, 421, 53 },
		{ 460, 422, 53 },
		{ 460, 423, 53 },
		{ 460, 424, 53 },
		{ 460, 425, 53 },
		{ 460, 426, 53 },
		{ 460, 427, 53 },
		{ 460, 428, 53 },
		{ 460, 429, 53 },
		{ 460, 430, 53 },
		{ 460, 431, 53 },
		{ 460, 432, 53 },
		{ 460, 433, 53 },
		{ 460, 434, 53 },
		{ 460, 435, 53 },
		{ 460, 436, 53 },
		{ 460, 437, 53 },
		{ 460, 438, 53 },
		{ 460, 439, 53 },
		{ 460, 440, 53 },
		{ 460, 441, 53 },
		{ 460, 442, 53 },
		{ 460, 443, 53 },
		{ 460, 444, 53 },
		{ 460, 445, 53 },
		{ 460, 446, 53 },
		{ 460, 447, 53 },
		{ 460, 448, 53 },
		{ 460, 449, 53 },
		{ 460, 450, 53 },
		{ 460, 451, 53 },
		{ 460, 452, 53 },
		{ 460, 453, 53 },
		{ 460, 454, 53 },
		{ 460, 455, 53 },
		{ 460, 456, 53 },
		{ 460, 457, 53 },
		{ 460, 458, 53 },
		{ 460, 459, 53 },
		{ 461, 411, 53 },
		{ 461, 412, 53 },
		{ 461, 413, 53 },
		{ 461, 414, 53 },
		{ 461, 415, 53 },
		{ 461, 416, 53 },
		{ 461, 417, 53 },
		{ 461, 418, 53 },
		{ 461, 419, 53 },
		{ 461, 420, 53 },
		{ 461, 421, 53 },
		{ 461, 422, 53 },
		{ 461, 423, 53 },
		{ 461, 424, 53 },
		{ 461, 425, 53 },
		{ 461, 426, 53 },
		{ 461, 427, 53 },
		{ 461, 428, 53 },
		{ 461, 429, 53 },
		{ 461, 430, 53 },
		{ 461, 431, 53 },
		{ 461, 432, 53 },
		{ 461, 433, 53 },
		{ 461, 434, 53 },
		{ 461, 435, 53 },
		{ 461, 436, 53 },
		{ 461, 437, 53 },
		{ 461, 438, 53 },
		{ 461, 439, 53 },
		{ 461, 440, 53 },
		{ 461, 441, 53 },
		{ 461, 442, 53 },
		{ 461, 443, 53 },
		{ 461, 444, 53 },
		{ 461, 445, 53 },
		{ 461, 446, 53 },
		{ 461, 447, 53 },
		{ 461, 448, 53 },
		{ 461, 449, 53 },
		{ 461, 450, 53 },
		{ 461, 451, 53 },
		{ 461, 452, 53 },
		{ 461, 453, 53 },
		{ 461, 454, 53 },
		{ 461, 455, 53 },
		{ 461, 456, 53 },
		{ 461, 457, 53 },
		{ 461, 458, 53 },
		{ 461, 459, 53 },
		{ 461, 460, 53 },
		{ 462, 411, 53 },
		{ 462, 412, 53 },
		{ 462, 413, 53 },
		{ 462, 414, 53 },
		{ 462, 415, 53 },
		{ 462, 416, 53 },
		{ 462, 417, 53 },
		{ 462, 418, 53 },
		{ 462, 419, 53 },
		{ 462, 420, 53 },
		{ 462, 421, 53 },
		{ 462, 422, 53 },
		{ 462, 423, 53 },
		{ 462, 424, 53 },
		{ 462, 425, 53 },
		{ 462, 426, 53 },
		{ 462, 427, 53 },
		{ 462, 428, 53 },
		{ 462, 429, 53 },
		{ 462, 430, 53 },
		{ 462, 431, 53 },
		{ 462, 432, 53 },
		{ 462, 433, 53 },
		{ 462, 434, 53 },
		{ 462, 435, 53 },
		{ 462, 436, 53 },
		{ 462, 437, 53 },
		{ 462, 438, 53 },
		{ 462, 439, 53 },
		{ 462, 440, 53 },
		{ 462, 441, 53 },
		{ 462, 442, 53 },
		{ 462, 443, 53 },
		{ 462, 444, 53 },
		{ 462, 445, 53 },
		{ 462, 446, 53 },
		{ 462, 447, 53 },
		{ 462, 448, 53 },
		{ 462, 449, 53 },
		{ 462, 450, 53 },
		{ 462, 451, 53 },
		{ 462, 452, 53 },
		{ 462, 453, 53 },
		{ 462, 454, 53 },
		{ 462, 455, 53 },
		{ 462, 456, 53 },
		{ 462, 457, 53 },
		{ 462, 458, 53 },
		{ 462, 459, 53 },
		{ 462, 460, 53 },
		{ 462, 461, 53 },
		{ 463, 412, 53 },
		{ 463, 413, 53 },
		{ 463, 414, 53 },
		{ 463, 415, 53 },
		{ 463, 416, 53 },
		{ 463, 417, 53 },
		{ 463, 418, 53 },
		{ 463, 419, 53 },
		{ 463, 420, 53 },
		{ 463, 421, 53 },
		{ 463, 422, 53 },
		{ 463, 423, 53 },
		{ 463, 424, 53 },
		{ 463, 425, 53 },
		{ 463, 426, 53 },
		{ 463, 427, 53 },
		{ 463, 428, 53 },
		{ 463, 429, 53 },
		{ 463, 430, 53 },
		{ 463, 431, 53 },
		{ 463, 432, 53 },
		{ 463, 433, 53 },
		{ 463, 434, 53 },
		{ 463, 435, 53 },
		{ 463, 436, 53 },
		{ 463, 437, 53 },
		{ 463, 438, 53 },
		{ 463, 439, 53 },
		{ 463, 440, 53 },
		{ 463, 441, 53 },
		{ 463, 442, 53 },
		{ 463, 443, 53 },
		{ 463, 444, 53 },
		{ 463, 445, 53 },
		{ 463, 446, 53 },
		{ 463, 447, 53 },
		{ 463, 448, 53 },
		{ 463, 449, 53 },
		{ 463, 450, 53 },
		{ 463, 451, 53 },
		{ 463, 452, 53 },
		{ 463, 453, 53 },
		{ 463, 454, 53 },
		{ 463, 455, 53 },
		{ 463, 456, 53 },
		{ 463, 457, 53 },
		{ 463, 458, 53 },
		{ 463, 459, 53 },
		{ 463, 460, 53 },
		{ 463, 461, 53 },
		{ 463, 462, 53 },
		{ 464, 413, 53 },
		{ 464, 414, 53 },
		{ 464, 415, 53 },
		{ 464, 416, 53 },
		{ 464, 417, 53 },
		{ 464, 418, 53 },
		{ 464, 419, 53 },
		{ 464, 420, 53 },
		{ 464, 421, 53 },
		{ 464, 422, 53 },
		{ 464, 423, 53 },
		{ 464, 424, 53 },
		{ 464, 425, 53 },
		{ 464, 426, 53 },
		{ 464, 427, 53 },
		{ 464, 428, 53 },
		{ 464, 429, 53 },
		{ 464, 430, 53 },
		{ 464, 431, 53 },
		{ 464, 432, 53 },
		{ 464, 433, 53 },
		{ 464, 434, 53 },
		{ 464, 435, 53 },
		{ 464, 436, 53 },
		{ 464, 437, 53 },
		{ 464, 438, 53 },
		{ 464, 439, 53 },
		{ 464, 440, 53 },
		{ 464, 441, 53 },
		{ 464, 442, 53 },
		{ 464, 443, 53 },
		{ 464, 444, 53 },
		{ 464, 445, 53 },
		{ 464, 446, 53 },
		{ 464, 447, 53 },
		{ 464, 448, 53 },
		{ 464, 449, 53 },
		{ 464, 450, 53 },
		{ 464, 451, 53 },
		{ 464, 452, 53 },
		{ 464, 453, 53 },
		{ 464, 454, 53 },
		{ 464, 455, 53 },
		{ 464, 456, 53 },
		{ 464, 457, 53 },
		{ 464, 458, 53 },
		{ 464, 459, 53 },
		{ 464, 460, 53 },
		{ 464, 461, 53 },
		{ 464, 462, 53 },
		{ 464, 463, 53 },
		{ 465, 414, 53 },
		{ 465, 415, 53 },
		{ 465, 416, 53 },
		{ 465, 417, 53 },
		{ 465, 418, 53 },
		{ 465, 419, 53 },
		{ 465, 420, 53 },
		{ 465, 421, 53 },
		{ 465, 422, 53 },
		{ 465, 423, 53 },
		{ 465, 424, 53 },
		{ 465, 425, 53 },
		{ 465, 426, 53 },
		{ 465, 427, 53 },
		{ 465, 428, 53 },
		{ 465, 429, 53 },
		{ 465, 430, 53 },
		{ 465, 431, 53 },
		{ 465, 432, 53 },
		{ 465, 433, 53 },
		{ 465, 434, 53 },
		{ 465, 435, 53 },
		{ 465, 436, 53 },
		{ 465, 437, 53 },
		{ 465, 438, 53 },
		{ 465, 439, 53 },
		{ 465, 440, 53 },
		{ 465, 441, 53 },
		{ 465, 442, 53 },
		{ 465, 443, 53 },
		{ 465, 444, 53 },
		{ 465, 445, 53 },
		{ 465, 446, 53 },
		{ 465, 447, 53 },
		{ 465, 448, 53 },
		{ 465, 449, 53 },
		{ 465, 450, 53 },
		{ 465, 451, 53 },
		{ 465, 452, 53 },
		{ 465, 453, 53 },
		{ 465, 454, 53 },
		{ 465, 455, 53 },
		{ 465, 456, 53 },
		{ 465, 457, 53 },
		{ 465, 458, 53 },
		{ 465, 459, 53 },
		{ 465, 460, 53 },
		{ 465, 461, 53 },
		{ 465, 462, 53 },
		{ 465, 463, 53 },
		{ 465, 464, 53 },
		{ 466, 414, 53 },
		{ 466, 415, 53 },
		{ 466, 416, 53 },
		{ 466, 417, 53 },
		{ 466, 418, 53 },
		{ 466, 419, 53 },
		{ 466, 420, 53 },
		{ 466, 421, 53 },
		{ 466, 422, 53 },
		{ 466, 423, 53 },
		{ 466, 424, 53 },
		{ 466, 425, 53 },
		{ 466, 426, 53 },
		{ 466, 427, 53 },
		{ 466, 428, 53 },
		{ 466, 429, 53 },
		{ 466, 430, 53 },
		{ 466, 431, 53 },
		{ 466, 432, 53 },
		{ 466, 433, 53 },
		{ 466, 434, 53 },
		{ 466, 435, 53 },
		{ 466, 436, 53 },
		{ 466, 437, 53 },
		{ 466, 438, 53 },
		{ 466, 439, 53 },
		{ 466, 440, 53 },
		{ 466, 441, 53 },
		{ 466, 442, 53 },
		{ 466, 443, 53 },
		{ 466, 444, 53 },
		{ 466, 445, 53 },
		{ 466, 446, 53 },
		{ 466, 447, 53 },
		{ 466, 448, 53 },
		{ 466, 449, 53 },
		{ 466, 450, 53 },
		{ 466, 451, 53 },
		{ 466, 452, 53 },
		{ 466, 453, 53 },
		{ 466, 454, 53 },
		{ 466, 455, 53 },
		{ 466, 456, 53 },
		{ 466, 457, 53 },
		{ 466, 458, 53 },
		{ 466, 459, 53 },
		{ 466, 460, 53 },
		{ 466, 461, 53 },
		{ 466, 462, 53 },
		{ 466, 463, 53 },
		{ 466, 464, 53 },
		{ 466, 465, 53 },
		{ 467, 415, 53 },
		{ 467, 416, 53 },
		{ 467, 417, 53 },
		{ 467, 418, 53 },
		{ 467, 419, 53 },
		{ 467, 420, 53 },
		{ 467, 421, 53 },
		{ 467, 422, 53 },
		{ 467, 423, 53 },
		{ 467, 424, 53 },
		{ 467, 425, 53 },
		{ 467, 426, 53 },
		{ 467, 427, 53 },
		{ 467, 428, 53 },
		{ 467, 429, 53 },
		{ 467, 430, 53 },
		{ 467, 431, 53 },
		{ 467, 432, 53 },
		{ 467, 433, 53 },
		{ 467, 434, 53 },
		{ 467, 435, 53 },
		{ 467, 436, 53 },
		{ 467, 437, 53 },
		{ 467, 438, 53 },
		{ 467, 439, 53 },
		{ 467, 440, 53 },
		{ 467, 441, 53 },
		{ 467, 442, 53 },
		{ 467, 443, 53 },
		{ 467, 444, 53 },
		{ 467, 445, 53 },
		{ 467, 446, 53 },
		{ 467, 447, 53 },
		{ 467, 448, 53 },
		{ 467, 449, 53 },
		{ 467, 450, 53 },
		{ 467, 451, 53 },
		{ 467, 452, 53 },
		{ 467, 453, 53 },
		{ 467, 454, 53 },
		{ 467, 455, 53 },
		{ 467, 456, 53 },
		{ 467, 457, 53 },
		{ 467, 458, 53 },
		{ 467, 459, 53 },
		{ 467, 460, 53 },
		{ 467, 461, 53 },
		{ 467, 462, 53 },
		{ 467, 463, 53 },
		{ 467, 464, 53 },
		{ 467, 465, 53 },
		{ 467, 466, 53 },
		{ 468, 416, 53 },
		{ 468, 417, 53 },
		{ 468, 418, 53 },
		{ 468, 419, 53 },
		{ 468, 420, 53 },
		{ 468, 421, 53 },
		{ 468, 422, 53 },
		{ 468, 423, 53 },
		{ 468, 424, 53 },
		{ 468, 425, 53 },
		{ 468, 426, 53 },
		{ 468, 427, 53 },
		{ 468, 428, 53 },
		{ 468, 429, 53 },
		{ 468, 430, 53 },
		{ 468, 431, 53 },
		{ 468, 432, 53 },
		{ 468, 433, 53 },
		{ 468, 434, 53 },
		{ 468, 435, 53 },
		{ 468, 436, 53 },
		{ 468, 437, 53 },
		{ 468, 438, 53 },
		{ 468, 439, 53 },
		{ 468, 440, 53 },
		{ 468, 441, 53 },
		{ 468, 442, 53 },
		{ 468, 443, 53 },
		{ 468, 444, 53 },
		{ 468, 445, 53 },
		{ 468, 446, 53 },
		{ 468, 447, 53 },
		{ 468, 448, 53 },
		{ 468, 449, 53 },
		{ 468, 450, 53 },
		{ 468, 451, 53 },
		{ 468, 452, 53 },
		{ 468, 453, 53 },
		{ 468, 454, 53 },
		{ 468, 455, 53 },
		{ 468, 456, 53 },
		{ 468, 457, 53 },
		{ 468, 458, 53 },
		{ 468, 459, 53 },
		{ 468, 460, 53 },
		{ 468, 461, 53 },
		{ 468, 462, 53 },
		{ 468, 463, 53 },
		{ 468, 464, 53 },
		{ 468, 465, 53 },
		{ 468, 466, 53 },
		{ 468, 467, 53 },
		{ 469, 417, 53 },
		{ 469, 418, 53 },
		{ 469, 419, 53 },
		{ 469, 420, 53 },
		{ 469, 421, 53 },
		{ 469, 422, 53 },
		{ 469, 423, 53 },
		{ 469, 424, 53 },
		{ 469, 425, 53 },
		{ 469, 426, 53 },
		{ 469, 427, 53 },
		{ 469, 428, 53 },
		{ 469, 429, 53 },
		{ 469, 430, 53 },
		{ 469, 431, 53 },
		{ 469, 432, 53 },
		{ 469, 433, 53 },
		{ 469, 434, 53 },
		{ 469, 435, 53 },
		{ 469, 436, 53 },
		{ 469, 437, 53 },
		{ 469, 438, 53 },
		{ 469, 439, 53 },
		{ 469, 440, 53 },
		{ 469, 441, 53 },
		{ 469, 442, 53 },
		{ 469, 443, 53 },
		{ 469, 444, 53 },
		{ 469, 445, 53 },
		{ 469, 446, 53 },
		{ 469, 447, 53 },
		{ 469, 448, 53 },
		{ 469, 449, 53 },
		{ 469, 450, 53 },
		{ 469, 451, 53 },
		{ 469, 452, 53 },
		{ 469, 453, 53 },
		{ 469, 454, 53 },
		{ 469, 455, 53 },
		{ 469, 456, 53 },
		{ 469, 457, 53 },
		{ 469, 458, 53 },
		{ 469, 459, 53 },
		{ 469, 460, 53 },
		{ 469, 461, 53 },
		{ 469, 462, 53 },
		{ 469, 463, 53 },
		{ 469, 464, 53 },
		{ 469, 465, 53 },
		{ 469, 466, 53 },
		{ 469, 467, 53 },
		{ 469, 468, 53 },
		{ 470, 417, 53 },
		{ 470, 418, 53 },
		{ 470, 419, 53 },
		{ 470, 420, 53 },
		{ 470, 421, 53 },
		{ 470, 422, 53 },
		{ 470, 423, 53 },
		{ 470, 424, 53 },
		{ 470, 425, 53 },
		{ 470, 426, 53 },
		{ 470, 427, 53 },
		{ 470, 428, 53 },
		{ 470, 429, 53 },
		{ 470, 430, 53 },
		{ 470, 431, 53 },
		{ 470, 432, 53 },
		{ 470, 433, 53 },
		{ 470, 434, 53 },
		{ 470, 435, 53 },
		{ 470, 436, 53 },
		{ 470, 437, 53 },
		{ 470, 438, 53 },
		{ 470, 439, 53 },
		{ 470, 440, 53 },
		{ 470, 441, 53 },
		{ 470, 442, 53 },
		{ 470, 443, 53 },
		{ 470, 444, 53 },
		{ 470, 445, 53 },
		{ 470, 446, 53 },
		{ 470, 447, 53 },
		{ 470, 448, 53 },
		{ 470, 449, 53 },
		{ 470, 450, 53 },
		{ 470, 451, 53 },
		{ 470, 452, 53 },
		{ 470, 453, 53 },
		{ 470, 454, 53 },
		{ 470, 455, 53 },
		{ 470, 456, 53 },
		{ 470, 457, 53 },
		{ 470, 458, 53 },
		{ 470, 459, 53 },
		{ 470, 460, 53 },
		{ 470, 461, 53 },
		{ 470, 462, 53 },
		{ 470, 463, 53 },
		{ 470, 464, 53 },
		{ 470, 465, 53 },
		{ 470, 466, 53 },
		{ 470, 467, 53 },
		{ 470, 468, 53 },
		{ 470, 469, 53 },
		{ 471, 418, 53 },
		{ 471, 419, 53 },
		{ 471, 420, 53 },
		{ 471, 421, 53 },
		{ 471, 422, 53 },
		{ 471, 423, 53 },
		{ 471, 424, 53 },
		{ 471, 425, 53 },
		{ 471, 426, 53 },
		{ 471, 427, 53 },
		{ 471, 428, 53 },
		{ 471, 429, 53 },
		{ 471, 430, 53 },
		{ 471, 431, 53 },
		{ 471, 432, 53 },
		{ 471, 433, 53 },
		{ 471, 434, 53 },
		{ 471, 435, 53 },
		{ 471, 436, 53 },
		{ 471, 437, 53 },
		{ 471, 438, 53 },
		{ 471, 439, 53 },
		{ 471, 440, 53 },
		{ 471, 441, 53 },
		{ 471, 442, 53 },
		{ 471, 443, 53 },
		{ 471, 444, 53 },
		{ 471, 445, 53 },
		{ 471, 446, 53 },
		{ 471, 447, 53 },
		{ 471, 448, 53 },
		{ 471, 449, 53 },
		{ 471, 450, 53 },
		{ 471, 451, 53 },
		{ 471, 452, 53 },
		{ 471, 453, 53 },
		{ 471, 454, 53 },
		{ 471, 455, 53 },
		{ 471, 456, 53 },
		{ 471, 457, 53 },
		{ 471, 458, 53 },
		{ 471, 459, 53 },
		{ 471, 460, 53 },
		{ 471, 461, 53 },
		{ 471, 462, 53 },
		{ 471, 463, 53 },
		{ 471, 464, 53 },
		{ 471, 465, 53 },
		{ 471, 466, 53 },
		{ 471, 467, 53 },
		{ 471, 468, 53 },
		{ 471, 469, 53 },
		{ 471, 470, 53 },
		{ 472, 419, 53 },
		{ 472, 420, 53 },
		{ 472, 421, 53 },
		{ 472, 422, 53 },
		{ 472, 423, 53 },
		{ 472, 424, 53 },
		{ 472, 425, 53 },
		{ 472, 426, 53 },
		{ 472, 427, 53 },
		{ 472, 428, 53 },
		{ 472, 429, 53 },
		{ 472, 430, 53 },
		{ 472, 431, 53 },
		{ 472, 432, 53 },
		{ 472, 433, 53 },
		{ 472, 434, 53 },
		{ 472, 435, 53 },
		{ 472, 436, 53 },
		{ 472, 437, 53 },
		{ 472, 438, 53 },
		{ 472, 439, 53 },
		{ 472, 440, 53 },
		{ 472, 441, 53 },
		{ 472, 442, 53 },
		{ 472, 443, 53 },
		{ 472, 444, 53 },
		{ 472, 445, 53 },
		{ 472, 446, 53 },
		{ 472, 447, 53 },
		{ 472, 448, 53 },
		{ 472, 449, 53 },
		{ 472, 450, 53 },
		{ 472, 451, 53 },
		{ 472, 452, 53 },
		{ 472, 453, 53 },
		{ 472, 454, 53 },
		{ 472, 455, 53 },
		{ 472, 456, 53 },
		{ 472, 457, 53 },
		{ 472, 458, 53 },
		{ 472, 459, 53 },
		{ 472, 460, 53 },
		{ 472, 461, 53 },
		{ 472, 462, 53 },
		{ 472, 463, 53 },
		{ 472, 464, 53 },
		{ 472, 465, 53 },
		{ 472, 466, 53 },
		{ 472, 467, 53 },
		{ 472, 468, 53 },
		{ 472, 469, 53 },
		{ 472, 470, 53 },
		{ 472, 471, 53 },
		{ 473, 420, 53 },
		{ 473, 421, 53 },
		{ 473, 422, 53 },
		{ 473, 423, 53 },
		{ 473, 424, 53 },
		{ 473, 425, 53 },
		{ 473, 426, 53 },
		{ 473, 427, 53 },
		{ 473, 428, 53 },
		{ 473, 429, 53 },
		{ 473, 430, 53 },
		{ 473, 431, 53 },
		{ 473, 432, 53 },
		{ 473, 433, 53 },
		{ 473, 434, 53 },
		{ 473, 435, 53 },
		{ 473, 436, 53 },
		{ 473, 437, 53 },
		{ 473, 438, 53 },
		{ 473, 439, 53 },
		{ 473, 440, 53 },
		{ 473, 441, 53 },
		{ 473, 442, 53 },
		{ 473, 443, 53 },
		{ 473, 444, 53 },
		{ 473, 445, 53 },
		{ 473, 446, 53 },
		{ 473, 447, 53 },
		{ 473, 448, 53 },
		{ 473, 449, 53 },
		{ 473, 450, 53 },
		{ 473, 451, 53 },
		{ 473, 452, 53 },
		{ 473, 453, 53 },
		{ 473, 454, 53 },
		{ 473, 455, 53 },
		{ 473, 456, 53 },
		{ 473, 457, 53 },
		{ 473, 458, 53 },
		{ 473, 459, 53 },
		{ 473, 460, 53 },
		{ 473, 461, 53 },
		{ 473, 462, 53 },
		{ 473, 463, 53 },
		{ 473, 464, 53 },
		{ 473, 465, 53 },
		{ 473, 466, 53 },
		{ 473, 467, 53 },
		{ 473, 468, 53 },
		{ 473, 469, 53 },
		{ 473, 470, 53 },
		{ 473, 471, 53 },
		{ 473, 472, 53 },
		{ 474, 420, 53 },
		{ 474, 421, 53 },
		{ 474, 422, 53 },
		{ 474, 423, 53 },
		{ 474, 424, 53 },
		{ 474, 425, 53 },
		{ 474, 426, 53 },
		{ 474, 427, 53 },
		{ 474, 428, 53 },
		{ 474, 429, 53 },
		{ 474, 430, 53 },
		{ 474, 431, 53 },
		{ 474, 432, 53 },
		{ 474, 433, 53 },
		{ 474, 434, 53 },
		{ 474, 435, 53 },
		{ 474, 436, 53 },
		{ 474, 437, 53 },
		{ 474, 438, 53 },
		{ 474, 439, 53 },
		{ 474, 440, 53 },
		{ 474, 441, 53 },
		{ 474, 442, 53 },
		{ 474, 443, 53 },
		{ 474, 444, 53 },
		{ 474, 445, 53 },
		{ 474, 446, 53 },
		{ 474, 447, 53 },
		{ 474, 448, 53 },
		{ 474, 449, 53 },
		{ 474, 450, 53 },
		{ 474, 451, 53 },
		{ 474, 452, 53 },
		{ 474, 453, 53 },
		{ 474, 454, 53 },
		{ 474, 455, 53 },
		{ 474, 456, 53 },
		{ 474, 457, 53 },
		{ 474, 458, 53 },
		{ 474, 459, 53 },
		{ 474, 460, 53 },
		{ 474, 461, 53 },
		{ 474, 462, 53 },
		{ 474, 463, 53 },
		{ 474, 464, 53 },
		{ 474, 465, 53 },
		{ 474, 466, 53 },
		{ 474, 467, 53 },
		{ 474, 468, 53 },
		{ 474, 469, 53 },
		{ 474, 470, 53 },
		{ 474, 471, 53 },
		{ 474, 472, 53 },
		{ 474, 473, 53 },
		{ 475, 421, 53 },
		{ 475, 422, 53 },
		{ 475, 423, 53 },
		{ 475, 424, 53 },
		{ 475, 425, 53 },
		{ 475, 426, 53 },
		{ 475, 427, 53 },
		{ 475, 428, 53 },
		{ 475, 429, 53 },
		{ 475, 430, 53 },
		{ 475, 431, 53 },
		{ 475, 432, 53 },
		{ 475, 433, 53 },
		{ 475, 434, 53 },
		{ 475, 435, 53 },
		{ 475, 436, 53 },
		{ 475, 437, 53 },
		{ 475, 438, 53 },
		{ 475, 439, 53 },
		{ 475, 440, 53 },
		{ 475, 441, 53 },
		{ 475, 442, 53 },
		{ 475, 443, 53 },
		{ 475, 444, 53 },
		{ 475, 445, 53 },
		{ 475, 446, 53 },
		{ 475, 447, 53 },
		{ 475, 448, 53 },
		{ 475, 449, 53 },
		{ 475, 450, 53 },
		{ 475, 451, 53 },
		{ 475, 452, 53 },
		{ 475, 453, 53 },
		{ 475, 454, 53 },
		{ 475, 455, 53 },
		{ 475, 456, 53 },
		{ 475, 457, 53 },
		{ 475, 458, 53 },
		{ 475, 459, 53 },
		{ 475, 460, 53 },
		{ 475, 461, 53 },
		{ 475, 462, 53 },
		{ 475, 463, 53 },
		{ 475, 464, 53 },
		{ 475, 465, 53 },
		{ 475, 466, 53 },
		{ 475, 467, 53 },
		{ 475, 468, 53 },
		{ 475, 469, 53 },
		{ 475, 470, 53 },
		{ 475, 471, 53 },
		{ 475, 472, 53 },
		{ 475, 473, 53 },
		{ 475, 474, 53 },
		{ 476, 422, 53 },
		{ 476, 423, 53 },
		{ 476, 424, 53 },
		{ 476, 425, 53 },
		{ 476, 426, 53 },
		{ 476, 427, 53 },
		{ 476, 428, 53 },
		{ 476, 429, 53 },
		{ 476, 430, 53 },
		{ 476, 431, 53 },
		{ 476, 432, 53 },
		{ 476, 433, 53 },
		{ 476, 434, 53 },
		{ 476, 435, 53 },
		{ 476, 436, 53 },
		{ 476, 437, 53 },
		{ 476, 438, 53 },
		{ 476, 439, 53 },
		{ 476, 440, 53 },
		{ 476, 441, 53 },
		{ 476, 442, 53 },
		{ 476, 443, 53 },
		{ 476, 444, 53 },
		{ 476, 445, 53 },
		{ 476, 446, 53 },
		{ 476, 447, 53 },
		{ 476, 448, 53 },
		{ 476, 449, 53 },
		{ 476, 450, 53 },
		{ 476, 451, 53 },
		{ 476, 452, 53 },
		{ 476, 453, 53 },
		{ 476, 454, 53 },
		{ 476, 455, 53 },
		{ 476, 456, 53 },
		{ 476, 457, 53 },
		{ 476, 458, 53 },
		{ 476, 459, 53 },
		{ 476, 460, 53 },
		{ 476, 461, 53 },
		{ 476, 462, 53 },
		{ 476, 463, 53 },
		{ 476, 464, 53 },
		{ 476, 465, 53 },
		{ 476, 466, 53 },
		{ 476, 467, 53 },
		{ 476, 468, 53 },
		{ 476, 469, 53 },
		{ 476, 470, 53 },
		{ 476, 471, 53 },
		{ 476, 472, 53 },
		{ 476, 473, 53 },
		{ 476, 474, 53 },
		{ 476, 475, 53 },
		{ 477, 423, 53 },
		{ 477, 424, 53 },
		{ 477, 425, 53 },
		{ 477, 426, 53 },
		{ 477, 427, 53 },
		{ 477, 428, 53 },
		{ 477, 429, 53 },
		{ 477, 430, 53 },
		{ 477, 431, 53 },
		{ 477, 432, 53 },
		{ 477, 433, 53 },
		{ 477, 434, 53 },
		{ 477, 435, 53 },
		{ 477, 436, 53 },
		{ 477, 437, 53 },
		{ 477, 438, 53 },
		{ 477, 439, 53 },
		{ 477, 440, 53 },
		{ 477, 441, 53 },
		{ 477, 442, 53 },
		{ 477, 443, 53 },
		{ 477, 444, 53 },
		{ 477, 445, 53 },
		{ 477, 446, 53 },
		{ 477, 447, 53 },
		{ 477, 448, 53 },
		{ 477, 449, 53 },
		{ 477, 450, 53 },
		{ 477, 451, 53 },
		{ 477, 452, 53 },
		{ 477, 453, 53 },
		{ 477, 454, 53 },
		{ 477, 455, 53 },
		{ 477, 456, 53 },
		{ 477, 457, 53 },
		{ 477, 458, 53 },
		{ 477, 459, 53 },
		{ 477, 460, 53 },
		{ 477, 461, 53 },
		{ 477, 462, 53 },
		{ 477, 463, 53 },
		{ 477, 464, 53 },
		{ 477, 465, 53 },
		{ 477, 466, 53 },
		{ 477, 467, 53 },
		{ 477, 468, 53 },
		{ 477, 469, 53 },
		{ 477, 470, 53 },
		{ 477, 471, 53 },
		{ 477, 472, 53 },
		{ 477, 473, 53 },
		{ 477, 474, 53 },
		{ 477, 475, 53 },
		{ 477, 476, 53 },
		{ 478, 423, 53 },
		{ 478, 424, 53 },
		{ 478, 425, 53 },
		{ 478, 426, 53 },
		{ 478, 427, 53 },
		{ 478, 428, 53 },
		{ 478, 429, 53 },
		{ 478, 430, 53 },
		{ 478, 431, 53 },
		{ 478, 432, 53 },
		{ 478, 433, 53 },
		{ 478, 434, 53 },
		{ 478, 435, 53 },
		{ 478, 436, 53 },
		{ 478, 437, 53 },
		{ 478, 438, 53 },
		{ 478, 439, 53 },
		{ 478, 440, 53 },
		{ 478, 441, 53 },
		{ 478, 442, 53 },
		{ 478, 443, 53 },
		{ 478, 444, 53 },
		{ 478, 445, 53 },
		{ 478, 446, 53 },
		{ 478, 447, 53 },
		{ 478, 448, 53 },
		{ 478, 449, 53 },
		{ 478, 450, 53 },
		{ 478, 451, 53 },
		{ 478, 452, 53 },
		{ 478, 453, 53 },
		{ 478, 454, 53 },
		{ 478, 455, 53 },
		{ 478, 456, 53 },
		{ 478, 457, 53 },
		{ 478, 458, 53 },
		{ 478, 459, 53 },
		{ 478, 460, 53 },
		{ 478, 461, 53 },
		{ 478, 462, 53 },
		{ 478, 463, 53 },
		{ 478, 464, 53 },
		{ 478, 465, 53 },
		{ 478, 466, 53 },
		{ 478, 467, 53 },
		{ 478, 468, 53 },
		{ 478, 469, 53 },
		{ 478, 470, 53 },
		{ 478, 471, 53 },
		{ 478, 472, 53 },
		{ 478, 473, 53 },
		{ 478, 474, 53 },
		{ 478, 475, 53 },
		{ 478, 476, 53 },
		{ 478, 477, 53 },
		{ 479, 424, 53 },
		{ 479, 425, 53 },
		{ 479, 426, 53 },
		{ 479, 427, 53 },
		{ 479, 428, 53 },
		{ 479, 429, 53 },
		{ 479, 430, 53 },
		{ 479, 431, 53 },
		{ 479, 432, 53 },
		{ 479, 433, 53 },
		{ 479, 434, 53 },
		{ 479, 435, 53 },
		{ 479, 436, 53 },
		{ 479, 437, 53 },
		{ 479, 438, 53 },
		{ 479, 439, 53 },
		{ 479, 440, 53 },
		{ 479, 441, 53 },
		{ 479, 442, 53 },
		{ 479, 443, 53 },
		{ 479, 444, 53 },
		{ 479, 445, 53 },
		{ 479, 446, 53 },
		{ 479, 447, 53 },
		{ 479, 448, 53 },
		{ 479, 449, 53 },
		{ 479, 450, 53 },
		{ 479, 451, 53 },
		{ 479, 452, 53 },
		{ 479, 453, 53 },
		{ 479, 454, 53 },
		{ 479, 455, 53 },
		{ 479, 456, 53 },
		{ 479, 457, 53 },
		{ 479, 458, 53 },
		{ 479, 459, 53 },
		{ 479, 460, 53 },
		{ 479, 461, 53 },
		{ 479, 462, 53 },
		{ 479, 463, 53 },
		{ 479, 464, 53 },
		{ 479, 465, 53 },
		{ 479, 466, 53 },
		{ 479, 467, 53 },
		{ 479, 468, 53 },
		{ 479, 469, 53 },
		{ 479, 470, 53 },
		{ 479, 471, 53 },
		{ 479, 472, 53 },
		{ 479, 473, 53 },
		{ 479, 474, 53 },
		{ 479, 475, 53 },
		{ 479, 476, 53 },
		{ 479, 477, 53 },
		{ 479, 478, 53 },
		{ 480, 425, 53 },
		{ 480, 426, 53 },
		{ 480, 427, 53 },
		{ 480, 428, 53 },
		{ 480, 429, 53 },
		{ 480, 430, 53 },
		{ 480, 431, 53 },
		{ 480, 432, 53 },
		{ 480, 433, 53 },
		{ 480, 434, 53 },
		{ 480, 435, 53 },
		{ 480, 436, 53 },
		{ 480, 437, 53 },
		{ 480, 438, 53 },
		{ 480, 439, 53 },
		{ 480, 440, 53 },
		{ 480, 441, 53 },
		{ 480, 442, 53 },
		{ 480, 443, 53 },
		{ 480, 444, 53 },
		{ 480, 445, 53 },
		{ 480, 446, 53 },
		{ 480, 447, 53 },
		{ 480, 448, 53 },
		{ 480, 449, 53 },
		{ 480, 450, 53 },
		{ 480, 451, 53 },
		{ 480, 452, 53 },
		{ 480, 453, 53 },
		{ 480, 454, 53 },
		{ 480, 455, 53 },
		{ 480, 456, 53 },
		{ 480, 457, 53 },
		{ 480, 458, 53 },
		{ 480, 459, 53 },
		{ 480, 460, 53 },
		{ 480, 461, 53 },
		{ 480, 462, 53 },
		{ 480, 463, 53 },
		{ 480, 464, 53 },
		{ 480, 465, 53 },
		{ 480, 466, 53 },
		{ 480, 467, 53 },
		{ 480, 468, 53 },
		{ 480, 469, 53 },
		{ 480, 470, 53 },
		{ 480, 471, 53 },
		{ 480, 472, 53 },
		{ 480, 473, 53 },
		{ 480, 474, 53 },
		{ 480, 475, 53 },
		{ 480, 476, 53 },
		{ 480, 477, 53 },
		{ 480, 478, 53 },
		{ 480, 479, 53 },
		{ 481, 426, 53 },
		{ 481, 427, 53 },
		{ 481, 428, 53 },
		{ 481, 429, 53 },
		{ 481, 430, 53 },
		{ 481, 431, 53 },
		{ 481, 432, 53 },
		{ 481, 433, 53 },
		{ 481, 434, 53 },
		{ 481, 435, 53 },
		{ 481, 436, 53 },
		{ 481, 437, 53 },
		{ 481, 438, 53 },
		{ 481, 439, 53 },
		{ 481, 440, 53 },
		{ 481, 441, 53 },
		{ 481, 442, 53 },
		{ 481, 443, 53 },
		{ 481, 444, 53 },
		{ 481, 445, 53 },
		{ 481, 446, 53 },
		{ 481, 447, 53 },
		{ 481, 448, 53 },
		{ 481, 449, 53 },
		{ 481, 450, 53 },
		{ 481, 451, 53 },
		{ 481, 452, 53 },
		{ 481, 453, 53 },
		{ 481, 454, 53 },
		{ 481, 455, 53 },
		{ 481, 456, 53 },
		{ 481, 457, 53 },
		{ 481, 458, 53 },
		{ 481, 459, 53 },
		{ 481, 460, 53 },
		{ 481, 461, 53 },
		{ 481, 462, 53 },
		{ 481, 463, 53 },
		{ 481, 464, 53 },
		{ 481, 465, 53 },
		{ 481, 466, 53 },
		{ 481, 467, 53 },
		{ 481, 468, 53 },
		{ 481, 469, 53 },
		{ 481, 470, 53 },
		{ 481, 471, 53 },
		{ 481, 472, 53 },
		{ 481, 473, 53 },
		{ 481, 474, 53 },
		{ 481, 475, 53 },
		{ 481, 476, 53 },
		{ 481, 477, 53 },
		{ 481, 478, 53 },
		{ 481, 479, 53 },
		{ 481, 480, 53 },
		{ 482, 426, 53 },
		{ 482, 427, 53 },
		{ 482, 428, 53 },
		{ 482, 429, 53 },
		{ 482, 430, 53 },
		{ 482, 431, 53 },
		{ 482, 432, 53 },
		{ 482, 433, 53 },
		{ 482, 434, 53 },
		{ 482, 435, 53 },
		{ 482, 436, 53 },
		{ 482, 437, 53 },
		{ 482, 438, 53 },
		{ 482, 439, 53 },
		{ 482, 440, 53 },
		{ 482, 441, 53 },
		{ 482, 442, 53 },
		{ 482, 443, 53 },
		{ 482, 444, 53 },
		{ 482, 445, 53 },
		{ 482, 446, 53 },
		{ 482, 447, 53 },
		{ 482, 448, 53 },
		{ 482, 449, 53 },
		{ 482, 450, 53 },
		{ 482, 451, 53 },
		{ 482, 452, 53 },
		{ 482, 453, 53 },
		{ 482, 454, 53 },
		{ 482, 455, 53 },
		{ 482, 456, 53 },
		{ 482, 457, 53 },
		{ 482, 458, 53 },
		{ 482, 459, 53 },
		{ 482, 460, 53 },
		{ 482, 461, 53 },
		{ 482, 462, 53 },
		{ 482, 463, 53 },
		{ 482, 464, 53 },
		{ 482, 465, 53 },
		{ 482, 466, 53 },
		{ 482, 467, 53 },
		{ 482, 468, 53 },
		{ 482, 469, 53 },
		{ 482, 470, 53 },
		{ 482, 471, 53 },
		{ 482, 472, 53 },
		{ 482, 473, 53 },
		{ 482, 474, 53 },
		{ 482, 475, 53 },
		{ 482, 476, 53 },
		{ 482, 477, 53 },
		{ 482, 478, 53 },
		{ 482, 479, 53 },
		{ 482, 480, 53 },
		{ 482, 481, 53 },
		{ 483, 427, 53 },
		{ 483, 428, 53 },
		{ 483, 429, 53 },
		{ 483, 430, 53 },
		{ 483, 431, 53 },
		{ 483, 432, 53 },
		{ 483, 433, 53 },
		{ 483, 434, 53 },
		{ 483, 435, 53 },
		{ 483, 436, 53 },
		{ 483, 437, 53 },
		{ 483, 438, 53 },
		{ 483, 439, 53 },
		{ 483, 440, 53 },
		{ 483, 441, 53 },
		{ 483, 442, 53 },
		{ 483, 443, 53 },
		{ 483, 444, 53 },
		{ 483, 445, 53 },
		{ 483, 446, 53 },
		{ 483, 447, 53 },
		{ 483, 448, 53 },
		{ 483, 449, 53 },
		{ 483, 450, 53 },
		{ 483, 451, 53 },
		{ 483, 452, 53 },
		{ 483, 453, 53 },
		{ 483, 454, 53 },
		{ 483, 455, 53 },
		{ 483, 456, 53 },
		{ 483, 457, 53 },
		{ 483, 458, 53 },
		{ 483, 459, 53 },
		{ 483, 460, 53 },
		{ 483, 461, 53 },
		{ 483, 462, 53 },
		{ 483, 463, 53 },
		{ 483, 464, 53 },
		{ 483, 465, 53 },
		{ 483, 466, 53 },
		{ 483, 467, 53 },
		{ 483, 468, 53 },
		{ 483, 469, 53 },
		{ 483, 470, 53 },
		{ 483, 471, 53 },
		{ 483, 472, 53 },
		{ 483, 473, 53 },
		{ 483, 474, 53 },
		{ 483, 475, 53 },
		{ 483, 476, 53 },
		{ 483, 477, 53 },
		{ 483, 478, 53 },
		{ 483, 479, 53 },
		{ 483, 480, 53 },
		{ 483, 481, 53 },
		{ 483, 482, 53 },
		{ 484, 428, 53 },
		{ 484, 429, 53 },
		{ 484, 430, 53 },
		{ 484, 431, 53 },
		{ 484, 432, 53 },
		{ 484, 433, 53 },
		{ 484, 434, 53 },
		{ 484, 435, 53 },
		{ 484, 436, 53 },
		{ 484, 437, 53 },
		{ 484, 438, 53 },
		{ 484, 439, 53 },
		{ 484, 440, 53 },
		{ 484, 441, 53 },
		{ 484, 442, 53 },
		{ 484, 443, 53 },
		{ 484, 444, 53 },
		{ 484, 445, 53 },
		{ 484, 446, 53 },
		{ 484, 447, 53 },
		{ 484, 448, 53 },
		{ 484, 449, 53 },
		{ 484, 450, 53 },
		{ 484, 451, 53 },
		{ 484, 452, 53 },
		{ 484, 453, 53 },
		{ 484, 454, 53 },
		{ 484, 455, 53 },
		{ 484, 456, 53 },
		{ 484, 457, 53 },
		{ 484, 458, 53 },
		{ 484, 459, 53 },
		{ 484, 460, 53 },
		{ 484, 461, 53 },
		{ 484, 462, 53 },
		{ 484, 463, 53 },
		{ 484, 464, 53 },
		{ 484, 465, 53 },
		{ 484, 466, 53 },
		{ 484, 467, 53 },
		{ 484, 468, 53 },
		{ 484, 469, 53 },
		{ 484, 470, 53 },
		{ 484, 471, 53 },
		{ 484, 472, 53 },
		{ 484, 473, 53 },
		{ 484, 474, 53 },
		{ 484, 475, 53 },
		{ 484, 476, 53 },
		{ 484, 477, 53 },
		{ 484, 478, 53 },
		{ 484, 479, 53 },
		{ 484, 480, 53 },
		{ 484, 481, 53 },
		{ 484, 482, 53 },
		{ 484, 483, 53 },
		{ 485, 429, 53 },
		{ 485, 430, 53 },
		{ 485, 431, 53 },
		{ 485, 432, 53 },
		{ 485, 433, 53 },
		{ 485, 434, 53 },
		{ 485, 435, 53 },
		{ 485, 436, 53 },
		{ 485, 437, 53 },
		{ 485, 438, 53 },
		{ 485, 439, 53 },
		{ 485, 440, 53 },
		{ 485, 441, 53 },
		{ 485, 442, 53 },
		{ 485, 443, 53 },
		{ 485, 444, 53 },
		{ 485, 445, 53 },
		{ 485, 446, 53 },
		{ 485, 447, 53 },
		{ 485, 448, 53 },
		{ 485, 449, 53 },
		{ 485, 450, 53 },
		{ 485, 451, 53 },
		{ 485, 452, 53 },
		{ 485, 453, 53 },
		{ 485, 454, 53 },
		{ 485, 455, 53 },
		{ 485, 456, 53 },
		{ 485, 457, 53 },
		{ 485, 458, 53 },
		{ 485, 459, 53 },
		{ 485, 460, 53 },
		{ 485, 461, 53 },
		{ 485, 462, 53 },
		{ 485, 463, 53 },
		{ 485, 464, 53 },
		{ 485, 465, 53 },
		{ 485, 466, 53 },
		{ 485, 467, 53 },
		{ 485, 468, 53 },
		{ 485, 469, 53 },
		{ 485, 470, 53 },
		{ 485, 471, 53 },
		{ 485, 472, 53 },
		{ 485, 473, 53 },
		{ 485, 474, 53 },
		{ 485, 475, 53 },
		{ 485, 476, 53 },
		{ 485, 477, 53 },
		{ 485, 478, 53 },
		{ 485, 479, 53 },
		{ 485, 480, 53 },
		{ 485, 481, 53 },
		{ 485, 482, 53 },
		{ 485, 483, 53 },
		{ 485, 484, 53 },
		{ 486, 429, 53 },
		{ 486, 430, 53 },
		{ 486, 431, 53 },
		{ 486, 432, 53 },
		{ 486, 433, 53 },
		{ 486, 434, 53 },
		{ 486, 435, 53 },
		{ 486, 436, 53 },
		{ 486, 437, 53 },
		{ 486, 438, 53 },
		{ 486, 439, 53 },
		{ 486, 440, 53 },
		{ 486, 441, 53 },
		{ 486, 442, 53 },
		{ 486, 443, 53 },
		{ 486, 444, 53 },
		{ 486, 445, 53 },
		{ 486, 446, 53 },
		{ 486, 447, 53 },
		{ 486, 448, 53 },
		{ 486, 449, 53 },
		{ 486, 450, 53 },
		{ 486, 451, 53 },
		{ 486, 452, 53 },
		{ 486, 453, 53 },
		{ 486, 454, 53 },
		{ 486, 455, 53 },
		{ 486, 456, 53 },
		{ 486, 457, 53 },
		{ 486, 458, 53 },
		{ 486, 459, 53 },
		{ 486, 460, 53 },
		{ 486, 461, 53 },
		{ 486, 462, 53 },
		{ 486, 463, 53 },
		{ 486, 464, 53 },
		{ 486, 465, 53 },
		{ 486, 466, 53 },
		{ 486, 467, 53 },
		{ 486, 468, 53 },
		{ 486, 469, 53 },
		{ 486, 470, 53 },
		{ 486, 471, 53 },
		{ 486, 472, 53 },
		{ 486, 473, 53 },
		{ 486, 474, 53 },
		{ 486, 475, 53 },
		{ 486, 476, 53 },
		{ 486, 477, 53 },
		{ 486, 478, 53 },
		{ 486, 479, 53 },
		{ 486, 480, 53 },
		{ 486, 481, 53 },
		{ 486, 482, 53 },
		{ 486, 483, 53 },
		{ 486, 484, 53 },
		{ 486, 485, 53 },
		{ 487, 430, 53 },
		{ 487, 431, 53 },
		{ 487, 432, 53 },
		{ 487, 433, 53 },
		{ 487, 434, 53 },
		{ 487, 435, 53 },
		{ 487, 436, 53 },
		{ 487, 437, 53 },
		{ 487, 438, 53 },
		{ 487, 439, 53 },
		{ 487, 440, 53 },
		{ 487, 441, 53 },
		{ 487, 442, 53 },
		{ 487, 443, 53 },
		{ 487, 444, 53 },
		{ 487, 445, 53 },
		{ 487, 446, 53 },
		{ 487, 447, 53 },
		{ 487, 448, 53 },
		{ 487, 449, 53 },
		{ 487, 450, 53 },
		{ 487, 451, 53 },
		{ 487, 452, 53 },
		{ 487, 453, 53 },
		{ 487, 454, 53 },
		{ 487, 455, 53 },
		{ 487, 456, 53 },
		{ 487, 457, 53 },
		{ 487, 458, 53 },
		{ 487, 459, 53 },
		{ 487, 460, 53 },
		{ 487, 461, 53 },
		{ 487, 462, 53 },
		{ 487, 463, 53 },
		{ 487, 464, 53 },
		{ 487, 465, 53 },
		{ 487, 466, 53 },
		{ 487, 467, 53 },
		{ 487, 468, 53 },
		{ 487, 469, 53 },
		{ 487, 470, 53 },
		{ 487, 471, 53 },
		{ 487, 472, 53 },
		{ 487, 473, 53 },
		{ 487, 474, 53 },
		{ 487, 475, 53 },
		{ 487, 476, 53 },
		{ 487, 477, 53 },
		{ 487, 478, 53 },
		{ 487, 479, 53 },
		{ 487, 480, 53 },
		{ 487, 481, 53 },
		{ 487, 482, 53 },
		{ 487, 483, 53 },
		{ 487, 484, 53 },
		{ 487, 485, 53 },
		{ 487, 486, 53 },
		{ 488, 431, 53 },
		{ 488, 432, 53 },
		{ 488, 433, 53 },
		{ 488, 434, 53 },
		{ 488, 435, 53 },
		{ 488, 436, 53 },
		{ 488, 437, 53 },
		{ 488, 438, 53 },
		{ 488, 439, 53 },
		{ 488, 440, 53 },
		{ 488, 441, 53 },
		{ 488, 442, 53 },
		{ 488, 443, 53 },
		{ 488, 444, 53 },
		{ 488, 445, 53 },
		{ 488, 446, 53 },
		{ 488, 447, 53 },
		{ 488, 448, 53 },
		{ 488, 449, 53 },
		{ 488, 450, 53 },
		{ 488, 451, 53 },
		{ 488, 452, 53 },
		{ 488, 453, 53 },
		{ 488, 454, 53 },
		{ 488, 455, 53 },
		{ 488, 456, 53 },
		{ 488, 457, 53 },
		{ 488, 458, 53 },
		{ 488, 459, 53 },
		{ 488, 460, 53 },
		{ 488, 461, 53 },
		{ 488, 462, 53 },
		{ 488, 463, 53 },
		{ 488, 464, 53 },
		{ 488, 465, 53 },
		{ 488, 466, 53 },
		{ 488, 467, 53 },
		{ 488, 468, 53 },
		{ 488, 469, 53 },
		{ 488, 470, 53 },
		{ 488, 471, 53 },
		{ 488, 472, 53 },
		{ 488, 473, 53 },
		{ 488, 474, 53 },
		{ 488, 475, 53 },
		{ 488, 476, 53 },
		{ 488, 477, 53 },
		{ 488, 478, 53 },
		{ 488, 479, 53 },
		{ 488, 480, 53 },
		{ 488, 481, 53 },
		{ 488, 482, 53 },
		{ 488, 483, 53 },
		{ 488, 484, 53 },
		{ 488, 485, 53 },
		{ 488, 486, 53 },
		{ 488, 487, 53 },
		{ 489, 432, 53 },
		{ 489, 433, 53 },
		{ 489, 434, 53 },
		{ 489, 435, 53 },
		{ 489, 436, 53 },
		{ 489, 437, 53 },
		{ 489, 438, 53 },
		{ 489, 439, 53 },
		{ 489, 440, 53 },
		{ 489, 441, 53 },
		{ 489, 442, 53 },
		{ 489, 443, 53 },
		{ 489, 444, 53 },
		{ 489, 445, 53 },
		{ 489, 446, 53 },
		{ 489, 447, 53 },
		{ 489, 448, 53 },
		{ 489, 449, 53 },
		{ 489, 450, 53 },
		{ 489, 451, 53 },
		{ 489, 452, 53 },
		{ 489, 453, 53 },
		{ 489, 454, 53 },
		{ 489, 455, 53 },
		{ 489, 456, 53 },
		{ 489, 457, 53 },
		{ 489, 458, 53 },
		{ 489, 459, 53 },
		{ 489, 460, 53 },
		{ 489, 461, 53 },
		{ 489, 462, 53 },
		{ 489, 463, 53 },
		{ 489, 464, 53 },
		{ 489, 465, 53 },
		{ 489, 466, 53 },
		{ 489, 467, 53 },
		{ 489, 468, 53 },
		{ 489, 469, 53 },
		{ 489, 470, 53 },
		{ 489, 471, 53 },
		{ 489, 472, 53 },
		{ 489, 473, 53 },
		{ 489, 474, 53 },
		{ 489, 475, 53 },
		{ 489, 476, 53 },
		{ 489, 477, 53 },
		{ 489, 478, 53 },
		{ 489, 479, 53 },
		{ 489, 480, 53 },
		{ 489, 481, 53 },
		{ 489, 482, 53 },
		{ 489, 483, 53 },
		{ 489, 484, 53 },
		{ 489, 485, 53 },
		{ 489, 486, 53 },
		{ 489, 487, 53 },
		{ 489, 488, 53 },
		{ 490, 432, 53 },
		{ 490, 433, 53 },
		{ 490, 434, 53 },
		{ 490, 435, 53 },
		{ 490, 436, 53 },
		{ 490, 437, 53 },
		{ 490, 438, 53 },
		{ 490, 439, 53 },
		{ 490, 440, 53 },
		{ 490, 441, 53 },
		{ 490, 442, 53 },
		{ 490, 443, 53 },
		{ 490, 444, 53 },
		{ 490, 445, 53 },
		{ 490, 446, 53 },
		{ 490, 447, 53 },
		{ 490, 448, 53 },
		{ 490, 449, 53 },
		{ 490, 450, 53 },
		{ 490, 451, 53 },
		{ 490, 452, 53 },
		{ 490, 453, 53 },
		{ 490, 454, 53 },
		{ 490, 455, 53 },
		{ 490, 456, 53 },
		{ 490, 457, 53 },
		{ 490, 458, 53 },
		{ 490, 459, 53 },
		{ 490, 460, 53 },
		{ 490, 461, 53 },
		{ 490, 462, 53 },
		{ 490, 463, 53 },
		{ 490, 464, 53 },
		{ 490, 465, 53 },
		{ 490, 466, 53 },
		{ 490, 467, 53 },
		{ 490, 468, 53 },
		{ 490, 469, 53 },
		{ 490, 470, 53 },
		{ 490, 471, 53 },
		{ 490, 472, 53 },
		{ 490, 473, 53 },
		{ 490, 474, 53 },
		{ 490, 475, 53 },
		{ 490, 476, 53 },
		{ 490, 477, 53 },
		{ 490, 478, 53 },
		{ 490, 479, 53 },
		{ 490, 480, 53 },
		{ 490, 481, 53 },
		{ 490, 482, 53 },
		{ 490, 483, 53 },
		{ 490, 484, 53 },
		{ 490, 485, 53 },
		{ 490, 486, 53 },
		{ 490, 487, 53 },
		{ 490, 488, 53 },
		{ 490, 489, 53 },
		{ 491, 433, 53 },
		{ 491, 434, 53 },
		{ 491, 435, 53 },
		{ 491, 436, 53 },
		{ 491, 437, 53 },
		{ 491, 438, 53 },
		{ 491, 439, 53 },
		{ 491, 440, 53 },
		{ 491, 441, 53 },
		{ 491, 442, 53 },
		{ 491, 443, 53 },
		{ 491, 444, 53 },
		{ 491, 445, 53 },
		{ 491, 446, 53 },
		{ 491, 447, 53 },
		{ 491, 448, 53 },
		{ 491, 449, 53 },
		{ 491, 450, 53 },
		{ 491, 451, 53 },
		{ 491, 452, 53 },
		{ 491, 453, 53 },
		{ 491, 454, 53 },
		{ 491, 455, 53 },
		{ 491, 456, 53 },
		{ 491, 457, 53 },
		{ 491, 458, 53 },
		{ 491, 459, 53 },
		{ 491, 460, 53 },
		{ 491, 461, 53 },
		{ 491, 462, 53 },
		{ 491, 463, 53 },
		{ 491, 464, 53 },
		{ 491, 465, 53 },
		{ 491, 466, 53 },
		{ 491, 467, 53 },
		{ 491, 468, 53 },
		{ 491, 469, 53 },
		{ 491, 470, 53 },
		{ 491, 471, 53 },
		{ 491, 472, 53 },
		{ 491, 473, 53 },
		{ 491, 474, 53 },
		{ 491, 475, 53 },
		{ 491, 476, 53 },
		{ 491, 477, 53 },
		{ 491, 478, 53 },
		{ 491, 479, 53 },
		{ 491, 480, 53 },
		{ 491, 481, 53 },
		{ 491, 482, 53 },
		{ 491, 483, 53 },
		{ 491, 484, 53 },
		{ 491, 485, 53 },
		{ 491, 486, 53 },
		{ 491, 487, 53 },
		{ 491, 488, 53 },
		{ 491, 489, 53 },
		{ 491, 490, 53 },
		{ 492, 434, 53 },
		{ 492, 435, 53 },
		{ 492, 436, 53 },
		{ 492, 437, 53 },
		{ 492, 438, 53 },
		{ 492, 439, 53 },
		{ 492, 440, 53 },
		{ 492, 441, 53 },
		{ 492, 442, 53 },
		{ 492, 443, 53 },
		{ 492, 444, 53 },
		{ 492, 445, 53 },
		{ 492, 446, 53 },
		{ 492, 447, 53 },
		{ 492, 448, 53 },
		{ 492, 449, 53 },
		{ 492, 450, 53 },
		{ 492, 451, 53 },
		{ 492, 452, 53 },
		{ 492, 453, 53 },
		{ 492, 454, 53 },
		{ 492, 455, 53 },
		{ 492, 456, 53 },
		{ 492, 457, 53 },
		{ 492, 458, 53 },
		{ 492, 459, 53 },
		{ 492, 460, 53 },
		{ 492, 461, 53 },
		{ 492, 462, 53 },
		{ 492, 463, 53 },
		{ 492, 464, 53 },
		{ 492, 465, 53 },
		{ 492, 466, 53 },
		{ 492, 467, 53 },
		{ 492, 468, 53 },
		{ 492, 469, 53 },
		{ 492, 470, 53 },
		{ 492, 471, 53 },
		{ 492, 472, 53 },
		{ 492, 473, 53 },
		{ 492, 474, 53 },
		{ 492, 475, 53 },
		{ 492, 476, 53 },
		{ 492, 477, 53 },
		{ 492, 478, 53 },
		{ 492, 479, 53 },
		{ 492, 480, 53 },
		{ 492, 481, 53 },
		{ 492, 482, 53 },
		{ 492, 483, 53 },
		{ 492, 484, 53 },
		{ 492, 485, 53 },
		{ 492, 486, 53 },
		{ 492, 487, 53 },
		{ 492, 488, 53 },
		{ 492, 489, 53 },
		{ 492, 490, 53 },
		{ 492, 491, 53 },
		{ 493, 435, 53 },
		{ 493, 436, 53 },
		{ 493, 437, 53 },
		{ 493, 438, 53 },
		{ 493, 439, 53 },
		{ 493, 440, 53 },
		{ 493, 441, 53 },
		{ 493, 442, 53 },
		{ 493, 443, 53 },
		{ 493, 444, 53 },
		{ 493, 445, 53 },
		{ 493, 446, 53 },
		{ 493, 447, 53 },
		{ 493, 448, 53 },
		{ 493, 449, 53 },
		{ 493, 450, 53 },
		{ 493, 451, 53 },
		{ 493, 452, 53 },
		{ 493, 453, 53 },
		{ 493, 454, 53 },
		{ 493, 455, 53 },
		{ 493, 456, 53 },
		{ 493, 457, 53 },
		{ 493, 458, 53 },
		{ 493, 459, 53 },
		{ 493, 460, 53 },
		{ 493, 461, 53 },
		{ 493, 462, 53 },
		{ 493, 463, 53 },
		{ 493, 464, 53 },
		{ 493, 465, 53 },
		{ 493, 466, 53 },
		{ 493, 467, 53 },
		{ 493, 468, 53 },
		{ 493, 469, 53 },
		{ 493, 470, 53 },
		{ 493, 471, 53 },
		{ 493, 472, 53 },
		{ 493, 473, 53 },
		{ 493, 474, 53 },
		{ 493, 475, 53 },
		{ 493, 476, 53 },
		{ 493, 477, 53 },
		{ 493, 478, 53 },
		{ 493, 479, 53 },
		{ 493, 480, 53 },
		{ 493, 481, 53 },
		{ 493, 482, 53 },
		{ 493, 483, 53 },
		{ 493, 484, 53 },
		{ 493, 485, 53 },
		{ 493, 486, 53 },
		{ 493, 487, 53 },
		{ 493, 488, 53 },
		{ 493, 489, 53 },
		{ 493, 490, 53 },
		{ 493, 491, 53 },
		{ 493, 492, 53 },
		{ 494, 435, 53 },
		{ 494, 436, 53 },
		{ 494, 437, 53 },
		{ 494, 438, 53 },
		{ 494, 439, 53 },
		{ 494, 440, 53 },
		{ 494, 441, 53 },
		{ 494, 442, 53 },
		{ 494, 443, 53 },
		{ 494, 444, 53 },
		{ 494, 445, 53 },
		{ 494, 446, 53 },
		{ 494, 447, 53 },
		{ 494, 448, 53 },
		{ 494, 449, 53 },
		{ 494, 450, 53 },
		{ 494, 451, 53 },
		{ 494, 452, 53 },
		{ 494, 453, 53 },
		{ 494, 454, 53 },
		{ 494, 455, 53 },
		{ 494, 456, 53 },
		{ 494, 457, 53 },
		{ 494, 458, 53 },
		{ 494, 459, 53 },
		{ 494, 460, 53 },
		{ 494, 461, 53 },
		{ 494, 462, 53 },
		{ 494, 463, 53 },
		{ 494, 464, 53 },
		{ 494, 465, 53 },
		{ 494, 466, 53 },
		{ 494, 467, 53 },
		{ 494, 468, 53 },
		{ 494, 469, 53 },
		{ 494, 470, 53 },
		{ 494, 471, 53 },
		{ 494, 472, 53 },
		{ 494, 473, 53 },
		{ 494, 474, 53 },
		{ 494, 475, 53 },
		{ 494, 476, 53 },
		{ 494, 477, 53 },
		{ 494, 478, 53 },
		{ 494, 479, 53 },
		{ 494, 480, 53 },
		{ 494, 481, 53 },
		{ 494, 482, 53 },
		{ 494, 483, 53 },
		{ 494, 484, 53 },
		{ 494, 485, 53 },
		{ 494, 486, 53 },
		{ 494, 487, 53 },
		{ 494, 488, 53 },
		{ 494, 489, 53 },
		{ 494, 490, 53 },
		{ 494, 491, 53 },
		{ 494, 492, 53 },
		{ 494, 493, 53 },
		{ 495, 436, 53 },
		{ 495, 437, 53 },
		{ 495, 438, 53 },
		{ 495, 439, 53 },
		{ 495, 440, 53 },
		{ 495, 441, 53 },
		{ 495, 442, 53 },
		{ 495, 443, 53 },
		{ 495, 444, 53 },
		{ 495, 445, 53 },
		{ 495, 446, 53 },
		{ 495, 447, 53 },
		{ 495, 448, 53 },
		{ 495, 449, 53 },
		{ 495, 450, 53 },
		{ 495, 451, 53 },
		{ 495, 452, 53 },
		{ 495, 453, 53 },
		{ 495, 454, 53 },
		{ 495, 455, 53 },
		{ 495, 456, 53 },
		{ 495, 457, 53 },
		{ 495, 458, 53 },
		{ 495, 459, 53 },
		{ 495, 460, 53 },
		{ 495, 461, 53 },
		{ 495, 462, 53 },
		{ 495, 463, 53 },
		{ 495, 464, 53 },
		{ 495, 465, 53 },
		{ 495, 466, 53 },
		{ 495, 467, 53 },
		{ 495, 468, 53 },
		{ 495, 469, 53 },
		{ 495, 470, 53 },
		{ 495, 471, 53 },
		{ 495, 472, 53 },
		{ 495, 473, 53 },
		{ 495, 474, 53 },
		{ 495, 475, 53 },
		{ 495, 476, 53 },
		{ 495, 477, 53 },
		{ 495, 478, 53 },
		{ 495, 479, 53 },
		{ 495, 480, 53 },
		{ 495, 481, 53 },
		{ 495, 482, 53 },
		{ 495, 483, 53 },
		{ 495, 484, 53 },
		{ 495, 485, 53 },
		{ 495, 486, 53 },
		{ 495, 487, 53 },
		{ 495, 488, 53 },
		{ 495, 489, 53 },
		{ 495, 490, 53 },
		{ 495, 491, 53 },
		{ 495, 492, 53 },
		{ 495, 493, 53 },
		{ 495, 494, 53 },
		{ 496, 437, 53 },
		{ 496, 438, 53 },
		{ 496, 439, 53 },
		{ 496, 440, 53 },
		{ 496, 441, 53 },
		{ 496, 442, 53 },
		{ 496, 443, 53 },
		{ 496, 444, 53 },
		{ 496, 445, 53 },
		{ 496, 446, 53 },
		{ 496, 447, 53 },
		{ 496, 448, 53 },
		{ 496, 449, 53 },
		{ 496, 450, 53 },
		{ 496, 451, 53 },
		{ 496, 452, 53 },
		{ 496, 453, 53 },
		{ 496, 454, 53 },
		{ 496, 455, 53 },
		{ 496, 456, 53 },
		{ 496, 457, 53 },
		{ 496, 458, 53 },
		{ 496, 459, 53 },
		{ 496, 460, 53 },
		{ 496, 461, 53 },
		{ 496, 462, 53 },
		{ 496, 463, 53 },
		{ 496, 464, 53 },
		{ 496, 465, 53 },
		{ 496, 466, 53 },
		{ 496, 467, 53 },
		{ 496, 468, 53 },
		{ 496, 469, 53 },
		{ 496, 470, 53 },
		{ 496, 471, 53 },
		{ 496, 472, 53 },
		{ 496, 473, 53 },
		{ 496, 474, 53 },
		{ 496, 475, 53 },
		{ 496, 476, 53 },
		{ 496, 477, 53 },
		{ 496, 478, 53 },
		{ 496, 479, 53 },
		{ 496, 480, 53 },
		{ 496, 481, 53 },
		{ 496, 482, 53 },
		{ 496, 483, 53 },
		{ 496, 484, 53 },
		{ 496, 485, 53 },
		{ 496, 486, 53 },
		{ 496, 487, 53 },
		{ 496, 488, 53 },
		{ 496, 489, 53 },
		{ 496, 490, 53 },
		{ 496, 491, 53 },
		{ 496, 492, 53 },
		{ 496, 493, 53 },
		{ 496, 494, 53 },
		{ 496, 495, 53 },
		{ 497, 438, 53 },
		{ 497, 439, 53 },
		{ 497, 440, 53 },
		{ 497, 441, 53 },
		{ 497, 442, 53 },
		{ 497, 443, 53 },
		{ 497, 444, 53 },
		{ 497, 445, 53 },
		{ 497, 446, 53 },
		{ 497, 447, 53 },
		{ 497, 448, 53 },
		{ 497, 449, 53 },
		{ 497, 450, 53 },
		{ 497, 451, 53 },
		{ 497, 452, 53 },
		{ 497, 453, 53 },
		{ 497, 454, 53 },
		{ 497, 455, 53 },
		{ 497, 456, 53 },
		{ 497, 457, 53 },
		{ 497, 458, 53 },
		{ 497, 459, 53 },
		{ 497, 460, 53 },
		{ 497, 461, 53 },
		{ 497, 462, 53 },
		{ 497, 463, 53 },
		{ 497, 464, 53 },
		{ 497, 465, 53 },
		{ 497, 466, 53 },
		{ 497, 467, 53 },
		{ 497, 468, 53 },
		{ 497, 469, 53 },
		{ 497, 470, 53 },
		{ 497, 471, 53 },
		{ 497, 472, 53 },
		{ 497, 473, 53 },
		{ 497, 474, 53 },
		{ 497, 475, 53 },
		{ 497, 476, 53 },
		{ 497, 477, 53 },
		{ 497, 478, 53 },
		{ 497, 479, 53 },
		{ 497, 480, 53 },
		{ 497, 481, 53 },
		{ 497, 482, 53 },
		{ 497, 483, 53 },
		{ 497, 484, 53 },
		{ 497, 485, 53 },
		{ 497, 486, 53 },
		{ 497, 487, 53 },
		{ 497, 488, 53 },
		{ 497, 489, 53 },
		{ 497, 490, 53 },
		{ 497, 491, 53 },
		{ 497, 492, 53 },
		{ 497, 493, 53 },
		{ 497, 494, 53 },
		{ 497, 495, 53 },
		{ 497, 496, 53 },
		{ 498, 438, 53 },
		{ 498, 439, 53 },
		{ 498, 440, 53 },
		{ 498, 441, 53 },
		{ 498, 442, 53 },
		{ 498, 443, 53 },
		{ 498, 444, 53 },
		{ 498, 445, 53 },
		{ 498, 446, 53 },
		{ 498, 447, 53 },
		{ 498, 448, 53 },
		{ 498, 449, 53 },
		{ 498, 450, 53 },
		{ 498, 451, 53 },
		{ 498, 452, 53 },
		{ 498, 453, 53 },
		{ 498, 454, 53 },
		{ 498, 455, 53 },
		{ 498, 456, 53 },
		{ 498, 457, 53 },
		{ 498, 458, 53 },
		{ 498, 459, 53 },
		{ 498, 460, 53 },
		{ 498, 461, 53 },
		{ 498, 462, 53 },
		{ 498, 463, 53 },
		{ 498, 464, 53 },
		{ 498, 465, 53 },
		{ 498, 466, 53 },
		{ 498, 467, 53 },
		{ 498, 468, 53 },
		{ 498, 469, 53 },
		{ 498, 470, 53 },
		{ 498, 471, 53 },
		{ 498, 472, 53 },
		{ 498, 473, 53 },
		{ 498, 474, 53 },
		{ 498, 475, 53 },
		{ 498, 476, 53 },
		{ 498, 477, 53 },
		{ 498, 478, 53 },
		{ 498, 479, 53 },
		{ 498, 480, 53 },
		{ 498, 481, 53 },
		{ 498, 482, 53 },
		{ 498, 483, 53 },
		{ 498, 484, 53 },
		{ 498, 485, 53 },
		{ 498, 486, 53 },
		{ 498, 487, 53 },
		{ 498, 488, 53 },
		{ 498, 489, 53 },
		{ 498, 490, 53 },
		{ 498, 491, 53 },
		{ 498, 492, 53 },
		{ 498, 493, 53 },
		{ 498, 494, 53 },
		{ 498, 495, 53 },
		{ 498, 496, 53 },
		{ 498, 497, 53 },
		{ 499, 439, 53 },
		{ 499, 440, 53 },
		{ 499, 441, 53 },
		{ 499, 442, 53 },
		{ 499, 443, 53 },
		{ 499, 444, 53 },
		{ 499, 445, 53 },
		{ 499, 446, 53 },
		{ 499, 447, 53 },
		{ 499, 448, 53 },
		{ 499, 449, 53 },
		{ 499, 450, 53 },
		{ 499, 451, 53 },
		{ 499, 452, 53 },
		{ 499, 453, 53 },
		{ 499, 454, 53 },
		{ 499, 455, 53 },
		{ 499, 456, 53 },
		{ 499, 457, 53 },
		{ 499, 458, 53 },
		{ 499, 459, 53 },
		{ 499, 460, 53 },
		{ 499, 461, 53 },
		{ 499, 462, 53 },
		{ 499, 463, 53 },
		{ 499, 464, 53 },
		{ 499, 465, 53 },
		{ 499, 466, 53 },
		{ 499, 467, 53 },
		{ 499, 468, 53 },
		{ 499, 469, 53 },
		{ 499, 470, 53 },
		{ 499, 471, 53 },
		{ 499, 472, 53 },
		{ 499, 473, 53 },
		{ 499, 474, 53 },
		{ 499, 475, 53 },
		{ 499, 476, 53 },
		{ 499, 477, 53 },
		{ 499, 478, 53 },
		{ 499, 479, 53 },
		{ 499, 480, 53 },
		{ 499, 481, 53 },
		{ 499, 482, 53 },
		{ 499, 483, 53 },
		{ 499, 484, 53 },
		{ 499, 485, 53 },
		{ 499, 486, 53 },
		{ 499, 487, 53 },
		{ 499, 488, 53 },
		{ 499, 489, 53 },
		{ 499, 490, 53 },
		{ 499, 491, 53 },
		{ 499, 492, 53 },
		{ 499, 493, 53 },
		{ 499, 494, 53 },
		{ 499, 495, 53 },
		{ 499, 496, 53 },
		{ 499, 497, 53 },
		{ 499, 498, 53 },
		{ 500, 440, 53 },
		{ 500, 441, 53 },
		{ 500, 442, 53 },
		{ 500, 443, 53 },
		{ 500, 444, 53 },
		{ 500, 445, 53 },
		{ 500, 446, 53 },
		{ 500, 447, 53 },
		{ 500, 448, 53 },
		{ 500, 449, 53 },
		{ 500, 450, 53 },
		{ 500, 451, 53 },
		{ 500, 452, 53 },
		{ 500, 453, 53 },
		{ 500, 454, 53 },
		{ 500, 455, 53 },
		{ 500, 456, 53 },
		{ 500, 457, 53 },
		{ 500, 458, 53 },
		{ 500, 459, 53 },
		{ 500, 460, 53 },
		{ 500, 461, 53 },
		{ 500, 462, 53 },
		{ 500, 463, 53 },
		{ 500, 464, 53 },
		{ 500, 465, 53 },
		{ 500, 466, 53 },
		{ 500, 467, 53 },
		{ 500, 468, 53 },
		{ 500, 469, 53 },
		{ 500, 470, 53 },
		{ 500, 471, 53 },
		{ 500, 472, 53 },
		{ 500, 473, 53 },
		{ 500, 474, 53 },
		{ 500, 475, 53 },
		{ 500, 476, 53 },
		{ 500, 477, 53 },
		{ 500, 478, 53 },
		{ 500, 479, 53 },
		{ 500, 480, 53 },
		{ 500, 481, 53 },
		{ 500, 482, 53 },
		{ 500, 483, 53 },
		{ 500, 484, 53 },
		{ 500, 485, 53 },
		{ 500, 486, 53 },
		{ 500, 487, 53 },
		{ 500, 488, 53 },
		{ 500, 489, 53 },
		{ 500, 490, 53 },
		{ 500, 491, 53 },
		{ 500, 492, 53 },
		{ 500, 493, 53 },
		{ 500, 494, 53 },
		{ 500, 495, 53 },
		{ 500, 496, 53 },
		{ 500, 497, 53 },
		{ 500, 498, 53 },
		{ 500, 499, 53 },
		{ 501, 441, 53 },
		{ 501, 442, 53 },
		{ 501, 443, 53 },
		{ 501, 444, 53 },
		{ 501, 445, 53 },
		{ 501, 446, 53 },
		{ 501, 447, 53 },
		{ 501, 448, 53 },
		{ 501, 449, 53 },
		{ 501, 450, 53 },
		{ 501, 451, 53 },
		{ 501, 452, 53 },
		{ 501, 453, 53 },
		{ 501, 454, 53 },
		{ 501, 455, 53 },
		{ 501, 456, 53 },
		{ 501, 457, 53 },
		{ 501, 458, 53 },
		{ 501, 459, 53 },
		{ 501, 460, 53 },
		{ 501, 461, 53 },
		{ 501, 462, 53 },
		{ 501, 463, 53 },
		{ 501, 464, 53 },
		{ 501, 465, 53 },
		{ 501, 466, 53 },
		{ 501, 467, 53 },
		{ 501, 468, 53 },
		{ 501, 469, 53 },
		{ 501, 470, 53 },
		{ 501, 471, 53 },
		{ 501, 472, 53 },
		{ 501, 473, 53 },
		{ 501, 474, 53 },
		{ 501, 475, 53 },
		{ 501, 476, 53 },
		{ 501, 477, 53 },
		{ 501, 478, 53 },
		{ 501, 479, 53 },
		{ 501, 480, 53 },
		{ 501, 481, 53 },
		{ 501, 482, 53 },
		{ 501, 483, 53 },
		{ 501, 484, 53 },
		{ 501, 485, 53 },
		{ 501, 486, 53 },
		{ 501, 487, 53 },
		{ 501, 488, 53 },
		{ 501, 489, 53 },
		{ 501, 490, 53 },
		{ 501, 491, 53 },
		{ 501, 492, 53 },
		{ 501, 493, 53 },
		{ 501, 494, 53 },
		{ 501, 495, 53 },
		{ 501, 496, 53 },
		{ 501, 497, 53 },
		{ 501, 498, 53 },
		{ 501, 499, 53 },
		{ 501, 500, 53 },
		{ 502, 441, 53 },
		{ 502, 442, 53 },
		{ 502, 443, 53 },
		{ 502, 444, 53 },
		{ 502, 445, 53 },
		{ 502, 446, 53 },
		{ 502, 447, 53 },
		{ 502, 448, 53 },
		{ 502, 449, 53 },
		{ 502, 450, 53 },
		{ 502, 451, 53 },
		{ 502, 452, 53 },
		{ 502, 453, 53 },
		{ 502, 454, 53 },
		{ 502, 455, 53 },
		{ 502, 456, 53 },
		{ 502, 457, 53 },
		{ 502, 458, 53 },
		{ 502, 459, 53 },
		{ 502, 460, 53 },
		{ 502, 461, 53 },
		{ 502, 462, 53 },
		{ 502, 463, 53 },
		{ 502, 464, 53 },
		{ 502, 465, 53 },
		{ 502, 466, 53 },
		{ 502, 467, 53 },
		{ 502, 468, 53 },
		{ 502, 469, 53 },
		{ 502, 470, 53 },
		{ 502, 471, 53 },
		{ 502, 472, 53 },
		{ 502, 473, 53 },
		{ 502, 474, 53 },
		{ 502, 475, 53 },
		{ 502, 476, 53 },
		{ 502, 477, 53 },
		{ 502, 478, 53 },
		{ 502, 479, 53 },
		{ 502, 480, 53 },
		{ 502, 481, 53 },
		{ 502, 482, 53 },
		{ 502, 483, 53 },
		{ 502, 484, 53 },
		{ 502, 485, 53 },
		{ 502, 486, 53 },
		{ 502, 487, 53 },
		{ 502, 488, 53 },
		{ 502, 489, 53 },
		{ 502, 490, 53 },
		{ 502, 491, 53 },
		{ 502, 492, 53 },
		{ 502, 493, 53 },
		{ 502, 494, 53 },
		{ 502, 495, 53 },
		{ 502, 496, 53 },
		{ 502, 497, 53 },
		{ 502, 498, 53 },
		{ 502, 499, 53 },
		{ 502, 500, 53 },
		{ 502, 501, 53 },
		{ 503, 442, 53 },
		{ 503, 443, 53 },
		{ 503, 444, 53 },
		{ 503, 445, 53 },
		{ 503, 446, 53 },
		{ 503, 447, 53 },
		{ 503, 448, 53 },
		{ 503, 449, 53 },
		{ 503, 450, 53 },
		{ 503, 451, 53 },
		{ 503, 452, 53 },
		{ 503, 453, 53 },
		{ 503, 454, 53 },
		{ 503, 455, 53 },
		{ 503, 456, 53 },
		{ 503, 457, 53 },
		{ 503, 458, 53 },
		{ 503, 459, 53 },
		{ 503, 460, 53 },
		{ 503, 461, 53 },
		{ 503, 462, 53 },
		{ 503, 463, 53 },
		{ 503, 464, 53 },
		{ 503, 465, 53 },
		{ 503, 466, 53 },
		{ 503, 467, 53 },
		{ 503, 468, 53 },
		{ 503, 469, 53 },
		{ 503, 470, 53 },
		{ 503, 471, 53 },
		{ 503, 472, 53 },
		{ 503, 473, 53 },
		{ 503, 474, 53 },
		{ 503, 475, 53 },
		{ 503, 476, 53 },
		{ 503, 477, 53 },
		{ 503, 478, 53 },
		{ 503, 479, 53 },
		{ 503, 480, 53 },
		{ 503, 481, 53 },
		{ 503, 482, 53 },
		{ 503, 483, 53 },
		{ 503, 484, 53 },
		{ 503, 485, 53 },
		{ 503, 486, 53 },
		{ 503, 487, 53 },
		{ 503, 488, 53 },
		{ 503, 489, 53 },
		{ 503, 490, 53 },
		{ 503, 491, 53 },
		{ 503, 492, 53 },
		{ 503, 493, 53 },
		{ 503, 494, 53 },
		{ 503, 495, 53 },
		{ 503, 496, 53 },
		{ 503, 497, 53 },
		{ 503, 498, 53 },
		{ 503, 499, 53 },
		{ 503, 500, 53 },
		{ 503, 501, 53 },
		{ 503, 502, 53 },
		{ 504, 443, 53 },
		{ 504, 444, 53 },
		{ 504, 445, 53 },
		{ 504, 446, 53 },
		{ 504, 447, 53 },
		{ 504, 448, 53 },
		{ 504, 449, 53 },
		{ 504, 450, 53 },
		{ 504, 451, 53 },
		{ 504, 452, 53 },
		{ 504, 453, 53 },
		{ 504, 454, 53 },
		{ 504, 455, 53 },
		{ 504, 456, 53 },
		{ 504, 457, 53 },
		{ 504, 458, 53 },
		{ 504, 459, 53 },
		{ 504, 460, 53 },
		{ 504, 461, 53 },
		{ 504, 462, 53 },
		{ 504, 463, 53 },
		{ 504, 464, 53 },
		{ 504, 465, 53 },
		{ 504, 466, 53 },
		{ 504, 467, 53 },
		{ 504, 468, 53 },
		{ 504, 469, 53 },
		{ 504, 470, 53 },
		{ 504, 471, 53 },
		{ 504, 472, 53 },
		{ 504, 473, 53 },
		{ 504, 474, 53 },
		{ 504, 475, 53 },
		{ 504, 476, 53 },
		{ 504, 477, 53 },
		{ 504, 478, 53 },
		{ 504, 479, 53 },
		{ 504, 480, 53 },
		{ 504, 481, 53 },
		{ 504, 482, 53 },
		{ 504, 483, 53 },
		{ 504, 484, 53 },
		{ 504, 485, 53 },
		{ 504, 486, 53 },
		{ 504, 487, 53 },
		{ 504, 488, 53 },
		{ 504, 489, 53 },
		{ 504, 490, 53 },
		{ 504, 491, 53 },
		{ 504, 492, 53 },
		{ 504, 493, 53 },
		{ 504, 494, 53 },
		{ 504, 495, 53 },
		{ 504, 496, 53 },
		{ 504, 497, 53 },
		{ 504, 498, 53 },
		{ 504, 499, 53 },
		{ 504, 500, 53 },
		{ 504, 501, 53 },
		{ 504, 502, 53 },
		{ 504, 503, 53 },
		{ 505, 444, 53 },
		{ 505, 445, 53 },
		{ 505, 446, 53 },
		{ 505, 447, 53 },
		{ 505, 448, 53 },
		{ 505, 449, 53 },
		{ 505, 450, 53 },
		{ 505, 451, 53 },
		{ 505, 452, 53 },
		{ 505, 453, 53 },
		{ 505, 454, 53 },
		{ 505, 455, 53 },
		{ 505, 456, 53 },
		{ 505, 457, 53 },
		{ 505, 458, 53 },
		{ 505, 459, 53 },
		{ 505, 460, 53 },
		{ 505, 461, 53 },
		{ 505, 462, 53 },
		{ 505, 463, 53 },
		{ 505, 464, 53 },
		{ 505, 465, 53 },
		{ 505, 466, 53 },
		{ 505, 467, 53 },
		{ 505, 468, 53 },
		{ 505, 469, 53 },
		{ 505, 470, 53 },
		{ 505, 471, 53 },
		{ 505, 472, 53 },
		{ 505, 473, 53 },
		{ 505, 474, 53 },
		{ 505, 475, 53 },
		{ 505, 476, 53 },
		{ 505, 477, 53 },
		{ 505, 478, 53 },
		{ 505, 479, 53 },
		{ 505, 480, 53 },
		{ 505, 481, 53 },
		{ 505, 482, 53 },
		{ 505, 483, 53 },
		{ 505, 484, 53 },
		{ 505, 485, 53 },
		{ 505, 486, 53 },
		{ 505, 487, 53 },
		{ 505, 488, 53 },
		{ 505, 489, 53 },
		{ 505, 490, 53 },
		{ 505, 491, 53 },
		{ 505, 492, 53 },
		{ 505, 493, 53 },
		{ 505, 494, 53 },
		{ 505, 495, 53 },
		{ 505, 496, 53 },
		{ 505, 497, 53 },
		{ 505, 498, 53 },
		{ 505, 499, 53 },
		{ 505, 500, 53 },
		{ 505, 501, 53 },
		{ 505, 502, 53 },
		{ 505, 503, 53 },
		{ 505, 504, 53 },
		{ 506, 444, 53 },
		{ 506, 445, 53 },
		{ 506, 446, 53 },
		{ 506, 447, 53 },
		{ 506, 448, 53 },
		{ 506, 449, 53 },
		{ 506, 450, 53 },
		{ 506, 451, 53 },
		{ 506, 452, 53 },
		{ 506, 453, 53 },
		{ 506, 454, 53 },
		{ 506, 455, 53 },
		{ 506, 456, 53 },
		{ 506, 457, 53 },
		{ 506, 458, 53 },
		{ 506, 459, 53 },
		{ 506, 460, 53 },
		{ 506, 461, 53 },
		{ 506, 462, 53 },
		{ 506, 463, 53 },
		{ 506, 464, 53 },
		{ 506, 465, 53 },
		{ 506, 466, 53 },
		{ 506, 467, 53 },
		{ 506, 468, 53 },
		{ 506, 469, 53 },
		{ 506, 470, 53 },
		{ 506, 471, 53 },
		{ 506, 472, 53 },
		{ 506, 473, 53 },
		{ 506, 474, 53 },
		{ 506, 475, 53 },
		{ 506, 476, 53 },
		{ 506, 477, 53 },
		{ 506, 478, 53 },
		{ 506, 479, 53 },
		{ 506, 480, 53 },
		{ 506, 481, 53 },
		{ 506, 482, 53 },
		{ 506, 483, 53 },
		{ 506, 484, 53 },
		{ 506, 485, 53 },
		{ 506, 486, 53 },
		{ 506, 487, 53 },
		{ 506, 488, 53 },
		{ 506, 489, 53 },
		{ 506, 490, 53 },
		{ 506, 491, 53 },
		{ 506, 492, 53 },
		{ 506, 493, 53 },
		{ 506, 494, 53 },
		{ 506, 495, 53 },
		{ 506, 496, 53 },
		{ 506, 497, 53 },
		{ 506, 498, 53 },
		{ 506, 499, 53 },
		{ 506, 500, 53 },
		{ 506, 501, 53 },
		{ 506, 502, 53 },
		{ 506, 503, 53 },
		{ 506, 504, 53 },
		{ 506, 505, 53 },
		{ 507, 445, 53 },
		{ 507, 446, 53 },
		{ 507, 447, 53 },
		{ 507, 448, 53 },
		{ 507, 449, 53 },
		{ 507, 450, 53 },
		{ 507, 451, 53 },
		{ 507, 452, 53 },
		{ 507, 453, 53 },
		{ 507, 454, 53 },
		{ 507, 455, 53 },
		{ 507, 456, 53 },
		{ 507, 457, 53 },
		{ 507, 458, 53 },
		{ 507, 459, 53 },
		{ 507, 460, 53 },
		{ 507, 461, 53 },
		{ 507, 462, 53 },
		{ 507, 463, 53 },
		{ 507, 464, 53 },
		{ 507, 465, 53 },
		{ 507, 466, 53 },
		{ 507, 467, 53 },
		{ 507, 468, 53 },
		{ 507, 469, 53 },
		{ 507, 470, 53 },
		{ 507, 471, 53 },
		{ 507, 472, 53 },
		{ 507, 473, 53 },
		{ 507, 474, 53 },
		{ 507, 475, 53 },
		{ 507, 476, 53 },
		{ 507, 477, 53 },
		{ 507, 478, 53 },
		{ 507, 479, 53 },
		{ 507, 480, 53 },
		{ 507, 481, 53 },
		{ 507, 482, 53 },
		{ 507, 483, 53 },
		{ 507, 484, 53 },
		{ 507, 485, 53 },
		{ 507, 486, 53 },
		{ 507, 487, 53 },
		{ 507, 488, 53 },
		{ 507, 489, 53 },
		{ 507, 490, 53 },
		{ 507, 491, 53 },
		{ 507, 492, 53 },
		{ 507, 493, 53 },
		{ 507, 494, 53 },
		{ 507, 495, 53 },
		{ 507, 496, 53 },
		{ 507, 497, 53 },
		{ 507, 498, 53 },
		{ 507, 499, 53 },
		{ 507, 500, 53 },
		{ 507, 501, 53 },
		{ 507, 502, 53 },
		{ 507, 503, 53 },
		{ 507, 504, 53 },
		{ 507, 505, 53 },
		{ 507, 506, 53 },
		{ 508, 446, 53 },
		{ 508, 447, 53 },
		{ 508, 448, 53 },
		{ 508, 449, 53 },
		{ 508, 450, 53 },
		{ 508, 451, 53 },
		{ 508, 452, 53 },
		{ 508, 453, 53 },
		{ 508, 454, 53 },
		{ 508, 455, 53 },
		{ 508, 456, 53 },
		{ 508, 457, 53 },
		{ 508, 458, 53 },
		{ 508, 459, 53 },
		{ 508, 460, 53 },
		{ 508, 461, 53 },
		{ 508, 462, 53 },
		{ 508, 463, 53 },
		{ 508, 464, 53 },
		{ 508, 465, 53 },
		{ 508, 466, 53 },
		{ 508, 467, 53 },
		{ 508, 468, 53 },
		{ 508, 469, 53 },
		{ 508, 470, 53 },
		{ 508, 471, 53 },
		{ 508, 472, 53 },
		{ 508, 473, 53 },
		{ 508, 474, 53 },
		{ 508, 475, 53 },
		{ 508, 476, 53 },
		{ 508, 477, 53 },
		{ 508, 478, 53 },
		{ 508, 479, 53 },
		{ 508, 480, 53 },
		{ 508, 481, 53 },
		{ 508, 482, 53 },
		{ 508, 483, 53 },
		{ 508, 484, 53 },
		{ 508, 485, 53 },
		{ 508, 486, 53 },
		{ 508, 487, 53 },
		{ 508, 488, 53 },
		{ 508, 489, 53 },
		{ 508, 490, 53 },
		{ 508, 491, 53 },
		{ 508, 492, 53 },
		{ 508, 493, 53 },
		{ 508, 494, 53 },
		{ 508, 495, 53 },
		{ 508, 496, 53 },
		{ 508, 497, 53 },
		{ 508, 498, 53 },
		{ 508, 499, 53 },
		{ 508, 500, 53 },
		{ 508, 501, 53 },
		{ 508, 502, 53 },
		{ 508, 503, 53 },
		{ 508, 504, 53 },
		{ 508, 505, 53 },
		{ 508, 506, 53 },
		{ 508, 507, 53 },
		{ 509, 447, 53 },
		{ 509, 448, 53 },
		{ 509, 449, 53 },
		{ 509, 450, 53 },
		{ 509, 451, 53 },
		{ 509, 452, 53 },
		{ 509, 453, 53 },
		{ 509, 454, 53 },
		{ 509, 455, 53 },
		{ 509, 456, 53 },
		{ 509, 457, 53 },
		{ 509, 458, 53 },
		{ 509, 459, 53 },
		{ 509, 460, 53 },
		{ 509, 461, 53 },
		{ 509, 462, 53 },
		{ 509, 463, 53 },
		{ 509, 464, 53 },
		{ 509, 465, 53 },
		{ 509, 466, 53 },
		{ 509, 467, 53 },
		{ 509, 468, 53 },
		{ 509, 469, 53 },
		{ 509, 470, 53 },
		{ 509, 471, 53 },
		{ 509, 472, 53 },
		{ 509, 473, 53 },
		{ 509, 474, 53 },
		{ 509, 475, 53 },
		{ 509, 476, 53 },
		{ 509, 477, 53 },
		{ 509, 478, 53 },
		{ 509, 479, 53 },
		{ 509, 480, 53 },
		{ 509, 481, 53 },
		{ 509, 482, 53 },
		{ 509, 483, 53 },
		{ 509, 484, 53 },
		{ 509, 485, 53 },
		{ 509, 486, 53 },
		{ 509, 487, 53 },
		{ 509, 488, 53 },
		{ 509, 489, 53 },
		{ 509, 490, 53 },
		{ 509, 491, 53 },
		{ 509, 492, 53 },
		{ 509, 493, 53 },
		{ 509, 494, 53 },
		{ 509, 495, 53 },
		{ 509, 496, 53 },
		{ 509, 497, 53 },
		{ 509, 498, 53 },
		{ 509, 499, 53 },
		{ 509, 500, 53 },
		{ 509, 501, 53 },
		{ 509, 502, 53 },
		{ 509, 503, 53 },
		{ 509, 504, 53 },
		{ 509, 505, 53 },
		{ 509, 506, 53 },
		{ 509, 507, 53 },
		{ 509, 508, 53 },
		{ 510, 447, 53 },
		{ 510, 448, 53 },
		{ 510, 449, 53 },
		{ 510, 450, 53 },
		{ 510, 451, 53 },
		{ 510, 452, 53 },
		{ 510, 453, 53 },
		{ 510, 454, 53 },
		{ 510, 455, 53 },
		{ 510, 456, 53 },
		{ 510, 457, 53 },
		{ 510, 458, 53 },
		{ 510, 459, 53 },
		{ 510, 460, 53 },
		{ 510, 461, 53 },
		{ 510, 462, 53 },
		{ 510, 463, 53 },
		{ 510, 464, 53 },
		{ 510, 465, 53 },
		{ 510, 466, 53 },
		{ 510, 467, 53 },
		{ 510, 468, 53 },
		{ 510, 469, 53 },
		{ 510, 470, 53 },
		{ 510, 471, 53 },
		{ 510, 472, 53 },
		{ 510, 473, 53 },
		{ 510, 474, 53 },
		{ 510, 475, 53 },
		{ 510, 476, 53 },
		{ 510, 477, 53 },
		{ 510, 478, 53 },
		{ 510, 479, 53 },
		{ 510, 480, 53 },
		{ 510, 481, 53 },
		{ 510, 482, 53 },
		{ 510, 483, 53 },
		{ 510, 484, 53 },
		{ 510, 485, 53 },
		{ 510, 486, 53 },
		{ 510, 487, 53 },
		{ 510, 488, 53 },
		{ 510, 489, 53 },
		{ 510, 490, 53 },
		{ 510, 491, 53 },
		{ 510, 492, 53 },
		{ 510, 493, 53 },
		{ 510, 494, 53 },
		{ 510, 495, 53 },
		{ 510, 496, 53 },
		{ 510, 497, 53 },
		{ 510, 498, 53 },
		{ 510, 499, 53 },
		{ 510, 500, 53 },
		{ 510, 501, 53 },
		{ 510, 502, 53 },
		{ 510, 503, 53 },
		{ 510, 504, 53 },
		{ 510, 505, 53 },
		{ 510, 506, 53 },
		{ 510, 507, 53 },
		{ 510, 508, 53 },
		{ 510, 509, 53 },
		{ 511, 448, 53 },
		{ 511, 449, 53 },
		{ 511, 450, 53 },
		{ 511, 451, 53 },
		{ 511, 452, 53 },
		{ 511, 453, 53 },
		{ 511, 454, 53 },
		{ 511, 455, 53 },
		{ 511, 456, 53 },
		{ 511, 457, 53 },
		{ 511, 458, 53 },
		{ 511, 459, 53 },
		{ 511, 460, 53 },
		{ 511, 461, 53 },
		{ 511, 462, 53 },
		{ 511, 463, 53 },
		{ 511, 464, 53 },
		{ 511, 465, 53 },
		{ 511, 466, 53 },
		{ 511, 467, 53 },
		{ 511, 468, 53 },
		{ 511, 469, 53 },
		{ 511, 470, 53 },
		{ 511, 471, 53 },
		{ 511, 472, 53 },
		{ 511, 473, 53 },
		{ 511, 474, 53 },
		{ 511, 475, 53 },
		{ 511, 476, 53 },
		{ 511, 477, 53 },
		{ 511, 478, 53 },
		{ 511, 479, 53 },
		{ 511, 480, 53 },
		{ 511, 481, 53 },
		{ 511, 482, 53 },
		{ 511, 483, 53 },
		{ 511, 484, 53 },
		{ 511, 485, 53 },
		{ 511, 486, 53 },
		{ 511, 487, 53 },
		{ 511, 488, 53 },
		{ 511, 489, 53 },
		{ 511, 490, 53 },
		{ 511, 491, 53 },
		{ 511, 492, 53 },
		{ 511, 493, 53 },
		{ 511, 494, 53 },
		{ 511, 495, 53 },
		{ 511, 496, 53 },
		{ 511, 497, 53 },
		{ 511, 498, 53 },
		{ 511, 499, 53 },
		{ 511, 500, 53 },
		{ 511, 501, 53 },
		{ 511, 502, 53 },
		{ 511, 503, 53 },
		{ 511, 504, 53 },
		{ 511, 505, 53 },
		{ 511, 506, 53 },
		{ 511, 507, 53 },
		{ 511, 508, 53 },
		{ 511, 509, 53 },
		{ 511, 510, 53 },
		{ 512, 449, 53 },
		{ 512, 450, 53 },
		{ 512, 451, 53 },
		{ 512, 452, 53 },
		{ 512, 453, 53 },
		{ 512, 454, 53 },
		{ 512, 455, 53 },
		{ 512, 456, 53 },
		{ 512, 457, 53 },
		{ 512, 458, 53 },
		{ 512, 459, 53 },
		{ 512, 460, 53 },
		{ 512, 461, 53 },
		{ 512, 462, 53 },
		{ 512, 463, 53 },
		{ 512, 464, 53 },
		{ 512, 465, 53 },
		{ 512, 466, 53 },
		{ 512, 467, 53 },
		{ 512, 468, 53 },
		{ 512, 469, 53 },
		{ 512, 470, 53 },
		{ 512, 471, 53 },
		{ 512, 472, 53 },
		{ 512, 473, 53 },
		{ 512, 474, 53 },
		{ 512, 475, 53 },
		{ 512, 476, 53 },
		{ 512, 477, 53 },
		{ 512, 478, 53 },
		{ 512, 479, 53 },
		{ 512, 480, 53 },
		{ 512, 481, 53 },
		{ 512, 482, 53 },
		{ 512, 483, 53 },
		{ 512, 484, 53 },
		{ 512, 485, 53 },
		{ 512, 486, 53 },
		{ 512, 487, 53 },
		{ 512, 488, 53 },
		{ 512, 489, 53 },
		{ 512, 490, 53 },
		{ 512, 491, 53 },
		{ 512, 492, 53 },
		{ 512, 493, 53 },
		{ 512, 494, 53 },
		{ 512, 495, 53 },
		{ 512, 496, 53 },
		{ 512, 497, 53 },
		{ 512, 498, 53 },
		{ 512, 499, 53 },
		{ 512, 500, 53 },
		{ 512, 501, 53 },
		{ 512, 502, 53 },
		{ 512, 503, 53 },
		{ 512, 504, 53 },
		{ 512, 505, 53 },
		{ 512, 506, 53 },
		{ 512, 507, 53 },
		{ 512, 508, 53 },
		{ 512, 509, 53 },
		{ 512, 510, 53 },
		{ 512, 511, 53 },
		{ 1, 1, 54 },
		{ 1, 2, 54 },
		{ 1, 3, 54 },
		{ 1, 4, 54 },
		{ 1, 5, 54 },
		{ 1, 6, 54 },
		{ 1, 7, 54 },
		{ 1, 8, 54 },
		{ 1, 9, 54 },
		{ 1, 10, 54 },
		{ 1, 11, 54 },
		{ 1, 12, 54 },
		{ 1, 13, 54 },
		{ 1, 14, 54 },
		{ 1, 15, 54 },
		{ 1, 16, 54 },
		{ 1, 17, 54 },
		{ 1, 18, 54 },
		{ 1, 19, 54 },
		{ 1, 20, 54 },
		{ 1, 21, 54 },
		{ 1, 22, 54 },
		{ 1, 23, 54 },
		{ 1, 24, 54 },
		{ 1, 25, 54 },
		{ 1, 26, 54 },
		{ 1, 27, 54 },
		{ 1, 28, 54 },
		{ 1, 29, 54 },
		{ 1, 30, 54 },
		{ 1, 31, 54 },
		{ 1, 32, 54 },
		{ 2, 1, 54 },
		{ 2, 2, 54 },
		{ 2, 3, 54 },
		{ 2, 4, 54 },
		{ 2, 5, 54 },
		{ 2, 6, 54 },
		{ 2, 7, 54 },
		{ 2, 8, 54 },
		{ 2, 9, 54 },
		{ 2, 10, 54 },
		{ 2, 11, 54 },
		{ 2, 12, 54 },
		{ 2, 13, 54 },
		{ 2, 14, 54 },
		{ 2, 15, 54 },
		{ 2, 16, 54 },
		{ 2, 17, 54 },
		{ 2, 18, 54 },
		{ 2, 19, 54 },
		{ 2, 20, 54 },
		{ 2, 21, 54 },
		{ 2, 22, 54 },
		{ 2, 23, 54 },
		{ 2, 24, 54 },
		{ 2, 25, 54 },
		{ 2, 26, 54 },
		{ 2, 27, 54 },
		{ 2, 28, 54 },
		{ 2, 29, 54 },
		{ 2, 30, 54 },
		{ 2, 31, 54 },
		{ 2, 32, 54 },
		{ 2, 33, 54 },
		{ 3, 1, 54 },
		{ 3, 2, 54 },
		{ 3, 3, 54 },
		{ 3, 4, 54 },
		{ 3, 5, 54 },
		{ 3, 6, 54 },
		{ 3, 7, 54 },
		{ 3, 8, 54 },
		{ 3, 9, 54 },
		{ 3, 10, 54 },
		{ 3, 11, 54 },
		{ 3, 12, 54 },
		{ 3, 13, 54 },
		{ 3, 14, 54 },
		{ 3, 15, 54 },
		{ 3, 16, 54 },
		{ 3, 17, 54 },
		{ 3, 18, 54 },
		{ 3, 19, 54 },
		{ 3, 20, 54 },
		{ 3, 21, 54 },
		{ 3, 22, 54 },
		{ 3, 23, 54 },
		{ 3, 24, 54 },
		{ 3, 25, 54 },
		{ 3, 26, 54 },
		{ 3, 27, 54 },
		{ 3, 28, 54 },
		{ 3, 29, 54 },
		{ 3, 30, 54 },
		{ 3, 31, 54 },
		{ 3, 32, 54 },
		{ 3, 33, 54 },
		{ 3, 34, 54 },
		{ 4, 1, 54 },
		{ 4, 2, 54 },
		{ 4, 3, 54 },
		{ 4, 4, 54 },
		{ 4, 5, 54 },
		{ 4, 6, 54 },
		{ 4, 7, 54 },
		{ 4, 8, 54 },
		{ 4, 9, 54 },
		{ 4, 10, 54 },
		{ 4, 11, 54 },
		{ 4, 12, 54 },
		{ 4, 13, 54 },
		{ 4, 14, 54 },
		{ 4, 15, 54 },
		{ 4, 16, 54 },
		{ 4, 17, 54 },
		{ 4, 18, 54 },
		{ 4, 19, 54 },
		{ 4, 20, 54 },
		{ 4, 21, 54 },
		{ 4, 22, 54 },
		{ 4, 23, 54 },
		{ 4, 24, 54 },
		{ 4, 25, 54 },
		{ 4, 26, 54 },
		{ 4, 27, 54 },
		{ 4, 28, 54 },
		{ 4, 29, 54 },
		{ 4, 30, 54 },
		{ 4, 31, 54 },
		{ 4, 32, 54 },
		{ 4, 33, 54 },
		{ 4, 34, 54 },
		{ 4, 35, 54 },
		{ 5, 1, 54 },
		{ 5, 2, 54 },
		{ 5, 3, 54 },
		{ 5, 4, 54 },
		{ 5, 5, 54 },
		{ 5, 6, 54 },
		{ 5, 7, 54 },
		{ 5, 8, 54 },
		{ 5, 9, 54 },
		{ 5, 10, 54 },
		{ 5, 11, 54 },
		{ 5, 12, 54 },
		{ 5, 13, 54 },
		{ 5, 14, 54 },
		{ 5, 15, 54 },
		{ 5, 16, 54 },
		{ 5, 17, 54 },
		{ 5, 18, 54 },
		{ 5, 19, 54 },
		{ 5, 20, 54 },
		{ 5, 21, 54 },
		{ 5, 22, 54 },
		{ 5, 23, 54 },
		{ 5, 24, 54 },
		{ 5, 25, 54 },
		{ 5, 26, 54 },
		{ 5, 27, 54 },
		{ 5, 28, 54 },
		{ 5, 29, 54 },
		{ 5, 30, 54 },
		{ 5, 31, 54 },
		{ 5, 32, 54 },
		{ 5, 33, 54 },
		{ 5, 34, 54 },
		{ 5, 35, 54 },
		{ 5, 36, 54 },
		{ 6, 1, 54 },
		{ 6, 2, 54 },
		{ 6, 3, 54 },
		{ 6, 4, 54 },
		{ 6, 5, 54 },
		{ 6, 6, 54 },
		{ 6, 7, 54 },
		{ 6, 8, 54 },
		{ 6, 9, 54 },
		{ 6, 10, 54 },
		{ 6, 11, 54 },
		{ 6, 12, 54 },
		{ 6, 13, 54 },
		{ 6, 14, 54 },
		{ 6, 15, 54 },
		{ 6, 16, 54 },
		{ 6, 17, 54 },
		{ 6, 18, 54 },
		{ 6, 19, 54 },
		{ 6, 20, 54 },
		{ 6, 21, 54 },
		{ 6, 22, 54 },
		{ 6, 23, 54 },
		{ 6, 24, 54 },
		{ 6, 25, 54 },
		{ 6, 26, 54 },
		{ 6, 27, 54 },
		{ 6, 28, 54 },
		{ 6, 29, 54 },
		{ 6, 30, 54 },
		{ 6, 31, 54 },
		{ 6, 32, 54 },
		{ 6, 33, 54 },
		{ 6, 34, 54 },
		{ 6, 35, 54 },
		{ 6, 36, 54 },
		{ 6, 37, 54 },
		{ 7, 1, 54 },
		{ 7, 2, 54 },
		{ 7, 3, 54 },
		{ 7, 4, 54 },
		{ 7, 5, 54 },
		{ 7, 6, 54 },
		{ 7, 7, 54 },
		{ 7, 8, 54 },
		{ 7, 9, 54 },
		{ 7, 10, 54 },
		{ 7, 11, 54 },
		{ 7, 12, 54 },
		{ 7, 13, 54 },
		{ 7, 14, 54 },
		{ 7, 15, 54 },
		{ 7, 16, 54 },
		{ 7, 17, 54 },
		{ 7, 18, 54 },
		{ 7, 19, 54 },
		{ 7, 20, 54 },
		{ 7, 21, 54 },
		{ 7, 22, 54 },
		{ 7, 23, 54 },
		{ 7, 24, 54 },
		{ 7, 25, 54 },
		{ 7, 26, 54 },
		{ 7, 27, 54 },
		{ 7, 28, 54 },
		{ 7, 29, 54 },
		{ 7, 30, 54 },
		{ 7, 31, 54 },
		{ 7, 32, 54 },
		{ 7, 33, 54 },
		{ 7, 34, 54 },
		{ 7, 35, 54 },
		{ 7, 36, 54 },
		{ 7, 37, 54 },
		{ 7, 38, 54 },
		{ 8, 1, 54 },
		{ 8, 2, 54 },
		{ 8, 3, 54 },
		{ 8, 4, 54 },
		{ 8, 5, 54 },
		{ 8, 6, 54 },
		{ 8, 7, 54 },
		{ 8, 8, 54 },
		{ 8, 9, 54 },
		{ 8, 10, 54 },
		{ 8, 11, 54 },
		{ 8, 12, 54 },
		{ 8, 13, 54 },
		{ 8, 14, 54 },
		{ 8, 15, 54 },
		{ 8, 16, 54 },
		{ 8, 17, 54 },
		{ 8, 18, 54 },
		{ 8, 19, 54 },
		{ 8, 20, 54 },
		{ 8, 21, 54 },
		{ 8, 22, 54 },
		{ 8, 23, 54 },
		{ 8, 24, 54 },
		{ 8, 25, 54 },
		{ 8, 26, 54 },
		{ 8, 27, 54 },
		{ 8, 28, 54 },
		{ 8, 29, 54 },
		{ 8, 30, 54 },
		{ 8, 31, 54 },
		{ 8, 32, 54 },
		{ 8, 33, 54 },
		{ 8, 34, 54 },
		{ 8, 35, 54 },
		{ 8, 36, 54 },
		{ 8, 37, 54 },
		{ 8, 38, 54 },
		{ 8, 39, 54 },
		{ 9, 1, 54 },
		{ 9, 2, 54 },
		{ 9, 3, 54 },
		{ 9, 4, 54 },
		{ 9, 5, 54 },
		{ 9, 6, 54 },
		{ 9, 7, 54 },
		{ 9, 8, 54 },
		{ 9, 9, 54 },
		{ 9, 10, 54 },
		{ 9, 11, 54 },
		{ 9, 12, 54 },
		{ 9, 13, 54 },
		{ 9, 14, 54 },
		{ 9, 15, 54 },
		{ 9, 16, 54 },
		{ 9, 17, 54 },
		{ 9, 18, 54 },
		{ 9, 19, 54 },
		{ 9, 20, 54 },
		{ 9, 21, 54 },
		{ 9, 22, 54 },
		{ 9, 23, 54 },
		{ 9, 24, 54 },
		{ 9, 25, 54 },
		{ 9, 26, 54 },
		{ 9, 27, 54 },
		{ 9, 28, 54 },
		{ 9, 29, 54 },
		{ 9, 30, 54 },
		{ 9, 31, 54 },
		{ 9, 32, 54 },
		{ 9, 33, 54 },
		{ 9, 34, 54 },
		{ 9, 35, 54 },
		{ 9, 36, 54 },
		{ 9, 37, 54 },
		{ 9, 38, 54 },
		{ 9, 39, 54 },
		{ 10, 1, 54 },
		{ 10, 2, 54 },
		{ 10, 3, 54 },
		{ 10, 4, 54 },
		{ 10, 5, 54 },
		{ 10, 6, 54 },
		{ 10, 7, 54 },
		{ 10, 8, 54 },
		{ 10, 9, 54 },
		{ 10, 10, 54 },
		{ 10, 11, 54 },
		{ 10, 12, 54 },
		{ 10, 13, 54 },
		{ 10, 14, 54 },
		{ 10, 15, 54 },
		{ 10, 16, 54 },
		{ 10, 17, 54 },
		{ 10, 18, 54 },
		{ 10, 19, 54 },
		{ 10, 20, 54 },
		{ 10, 21, 54 },
		{ 10, 22, 54 },
		{ 10, 23, 54 },
		{ 10, 24, 54 },
		{ 10, 25, 54 },
		{ 10, 26, 54 },
		{ 10, 27, 54 },
		{ 10, 28, 54 },
		{ 10, 29, 54 },
		{ 10, 30, 54 },
		{ 10, 31, 54 },
		{ 10, 32, 54 },
		{ 10, 33, 54 },
		{ 10, 34, 54 },
		{ 10, 35, 54 },
		{ 10, 36, 54 },
		{ 10, 37, 54 },
		{ 10, 38, 54 },
		{ 10, 39, 54 },
		{ 10, 40, 54 },
		{ 11, 1, 54 },
		{ 11, 2, 54 },
		{ 11, 3, 54 },
		{ 11, 4, 54 },
		{ 11, 5, 54 },
		{ 11, 6, 54 },
		{ 11, 7, 54 },
		{ 11, 8, 54 },
		{ 11, 9, 54 },
		{ 11, 10, 54 },
		{ 11, 11, 54 },
		{ 11, 12, 54 },
		{ 11, 13, 54 },
		{ 11, 14, 54 },
		{ 11, 15, 54 },
		{ 11, 16, 54 },
		{ 11, 17, 54 },
		{ 11, 18, 54 },
		{ 11, 19, 54 },
		{ 11, 20, 54 },
		{ 11, 21, 54 },
		{ 11, 22, 54 },
		{ 11, 23, 54 },
		{ 11, 24, 54 },
		{ 11, 25, 54 },
		{ 11, 26, 54 },
		{ 11, 27, 54 },
		{ 11, 28, 54 },
		{ 11, 29, 54 },
		{ 11, 30, 54 },
		{ 11, 31, 54 },
		{ 11, 32, 54 },
		{ 11, 33, 54 },
		{ 11, 34, 54 },
		{ 11, 35, 54 },
		{ 11, 36, 54 },
		{ 11, 37, 54 },
		{ 11, 38, 54 },
		{ 11, 39, 54 },
		{ 11, 40, 54 },
		{ 11, 41, 54 },
		{ 12, 1, 54 },
		{ 12, 2, 54 },
		{ 12, 3, 54 },
		{ 12, 4, 54 },
		{ 12, 5, 54 },
		{ 12, 6, 54 },
		{ 12, 7, 54 },
		{ 12, 8, 54 },
		{ 12, 9, 54 },
		{ 12, 10, 54 },
		{ 12, 11, 54 },
		{ 12, 12, 54 },
		{ 12, 13, 54 },
		{ 12, 14, 54 },
		{ 12, 15, 54 },
		{ 12, 16, 54 },
		{ 12, 17, 54 },
		{ 12, 18, 54 },
		{ 12, 19, 54 },
		{ 12, 20, 54 },
		{ 12, 21, 54 },
		{ 12, 22, 54 },
		{ 12, 23, 54 },
		{ 12, 24, 54 },
		{ 12, 25, 54 },
		{ 12, 26, 54 },
		{ 12, 27, 54 },
		{ 12, 28, 54 },
		{ 12, 29, 54 },
		{ 12, 30, 54 },
		{ 12, 31, 54 },
		{ 12, 32, 54 },
		{ 12, 33, 54 },
		{ 12, 34, 54 },
		{ 12, 35, 54 },
		{ 12, 36, 54 },
		{ 12, 37, 54 },
		{ 12, 38, 54 },
		{ 12, 39, 54 },
		{ 12, 40, 54 },
		{ 12, 41, 54 },
		{ 12, 42, 54 },
		{ 13, 1, 54 },
		{ 13, 2, 54 },
		{ 13, 3, 54 },
		{ 13, 4, 54 },
		{ 13, 5, 54 },
		{ 13, 6, 54 },
		{ 13, 7, 54 },
		{ 13, 8, 54 },
		{ 13, 9, 54 },
		{ 13, 10, 54 },
		{ 13, 11, 54 },
		{ 13, 12, 54 },
		{ 13, 13, 54 },
		{ 13, 14, 54 },
		{ 13, 15, 54 },
		{ 13, 16, 54 },
		{ 13, 17, 54 },
		{ 13, 18, 54 },
		{ 13, 19, 54 },
		{ 13, 20, 54 },
		{ 13, 21, 54 },
		{ 13, 22, 54 },
		{ 13, 23, 54 },
		{ 13, 24, 54 },
		{ 13, 25, 54 },
		{ 13, 26, 54 },
		{ 13, 27, 54 },
		{ 13, 28, 54 },
		{ 13, 29, 54 },
		{ 13, 30, 54 },
		{ 13, 31, 54 },
		{ 13, 32, 54 },
		{ 13, 33, 54 },
		{ 13, 34, 54 },
		{ 13, 35, 54 },
		{ 13, 36, 54 },
		{ 13, 37, 54 },
		{ 13, 38, 54 },
		{ 13, 39, 54 },
		{ 13, 40, 54 },
		{ 13, 41, 54 },
		{ 13, 42, 54 },
		{ 13, 43, 54 },
		{ 14, 1, 54 },
		{ 14, 2, 54 },
		{ 14, 3, 54 },
		{ 14, 4, 54 },
		{ 14, 5, 54 },
		{ 14, 6, 54 },
		{ 14, 7, 54 },
		{ 14, 8, 54 },
		{ 14, 9, 54 },
		{ 14, 10, 54 },
		{ 14, 11, 54 },
		{ 14, 12, 54 },
		{ 14, 13, 54 },
		{ 14, 14, 54 },
		{ 14, 15, 54 },
		{ 14, 16, 54 },
		{ 14, 17, 54 },
		{ 14, 18, 54 },
		{ 14, 19, 54 },
		{ 14, 20, 54 },
		{ 14, 21, 54 },
		{ 14, 22, 54 },
		{ 14, 23, 54 },
		{ 14, 24, 54 },
		{ 14, 25, 54 },
		{ 14, 26, 54 },
		{ 14, 27, 54 },
		{ 14, 28, 54 },
		{ 14, 29, 54 },
		{ 14, 30, 54 },
		{ 14, 31, 54 },
		{ 14, 32, 54 },
		{ 14, 33, 54 },
		{ 14, 34, 54 },
		{ 14, 35, 54 },
		{ 14, 36, 54 },
		{ 14, 37, 54 },
		{ 14, 38, 54 },
		{ 14, 39, 54 },
		{ 14, 40, 54 },
		{ 14, 41, 54 },
		{ 14, 42, 54 },
		{ 14, 43, 54 },
		{ 14, 44, 54 },
		{ 15, 1, 54 },
		{ 15, 2, 54 },
		{ 15, 3, 54 },
		{ 15, 4, 54 },
		{ 15, 5, 54 },
		{ 15, 6, 54 },
		{ 15, 7, 54 },
		{ 15, 8, 54 },
		{ 15, 9, 54 },
		{ 15, 10, 54 },
		{ 15, 11, 54 },
		{ 15, 12, 54 },
		{ 15, 13, 54 },
		{ 15, 14, 54 },
		{ 15, 15, 54 },
		{ 15, 16, 54 },
		{ 15, 17, 54 },
		{ 15, 18, 54 },
		{ 15, 19, 54 },
		{ 15, 20, 54 },
		{ 15, 21, 54 },
		{ 15, 22, 54 },
		{ 15, 23, 54 },
		{ 15, 24, 54 },
		{ 15, 25, 54 },
		{ 15, 26, 54 },
		{ 15, 27, 54 },
		{ 15, 28, 54 },
		{ 15, 29, 54 },
		{ 15, 30, 54 },
		{ 15, 31, 54 },
		{ 15, 32, 54 },
		{ 15, 33, 54 },
		{ 15, 34, 54 },
		{ 15, 35, 54 },
		{ 15, 36, 54 },
		{ 15, 37, 54 },
		{ 15, 38, 54 },
		{ 15, 39, 54 },
		{ 15, 40, 54 },
		{ 15, 41, 54 },
		{ 15, 42, 54 },
		{ 15, 43, 54 },
		{ 15, 44, 54 },
		{ 15, 45, 54 },
		{ 16, 1, 54 },
		{ 16, 2, 54 },
		{ 16, 3, 54 },
		{ 16, 4, 54 },
		{ 16, 5, 54 },
		{ 16, 6, 54 },
		{ 16, 7, 54 },
		{ 16, 8, 54 },
		{ 16, 9, 54 },
		{ 16, 10, 54 },
		{ 16, 11, 54 },
		{ 16, 12, 54 },
		{ 16, 13, 54 },
		{ 16, 14, 54 },
		{ 16, 15, 54 },
		{ 16, 16, 54 },
		{ 16, 17, 54 },
		{ 16, 18, 54 },
		{ 16, 19, 54 },
		{ 16, 20, 54 },
		{ 16, 21, 54 },
		{ 16, 22, 54 },
		{ 16, 23, 54 },
		{ 16, 24, 54 },
		{ 16, 25, 54 },
		{ 16, 26, 54 },
		{ 16, 27, 54 },
		{ 16, 28, 54 },
		{ 16, 29, 54 },
		{ 16, 30, 54 },
		{ 16, 31, 54 },
		{ 16, 32, 54 },
		{ 16, 33, 54 },
		{ 16, 34, 54 },
		{ 16, 35, 54 },
		{ 16, 36, 54 },
		{ 16, 37, 54 },
		{ 16, 38, 54 },
		{ 16, 39, 54 },
		{ 16, 40, 54 },
		{ 16, 41, 54 },
		{ 16, 42, 54 },
		{ 16, 43, 54 },
		{ 16, 44, 54 },
		{ 16, 45, 54 },
		{ 16, 46, 54 },
		{ 17, 1, 54 },
		{ 17, 2, 54 },
		{ 17, 3, 54 },
		{ 17, 4, 54 },
		{ 17, 5, 54 },
		{ 17, 6, 54 },
		{ 17, 7, 54 },
		{ 17, 8, 54 },
		{ 17, 9, 54 },
		{ 17, 10, 54 },
		{ 17, 11, 54 },
		{ 17, 12, 54 },
		{ 17, 13, 54 },
		{ 17, 14, 54 },
		{ 17, 15, 54 },
		{ 17, 16, 54 },
		{ 17, 17, 54 },
		{ 17, 18, 54 },
		{ 17, 19, 54 },
		{ 17, 20, 54 },
		{ 17, 21, 54 },
		{ 17, 22, 54 },
		{ 17, 23, 54 },
		{ 17, 24, 54 },
		{ 17, 25, 54 },
		{ 17, 26, 54 },
		{ 17, 27, 54 },
		{ 17, 28, 54 },
		{ 17, 29, 54 },
		{ 17, 30, 54 },
		{ 17, 31, 54 },
		{ 17, 32, 54 },
		{ 17, 33, 54 },
		{ 17, 34, 54 },
		{ 17, 35, 54 },
		{ 17, 36, 54 },
		{ 17, 37, 54 },
		{ 17, 38, 54 },
		{ 17, 39, 54 },
		{ 17, 40, 54 },
		{ 17, 41, 54 },
		{ 17, 42, 54 },
		{ 17, 43, 54 },
		{ 17, 44, 54 },
		{ 17, 45, 54 },
		{ 17, 46, 54 },
		{ 18, 1, 54 },
		{ 18, 2, 54 },
		{ 18, 3, 54 },
		{ 18, 4, 54 },
		{ 18, 5, 54 },
		{ 18, 6, 54 },
		{ 18, 7, 54 },
		{ 18, 8, 54 },
		{ 18, 9, 54 },
		{ 18, 10, 54 },
		{ 18, 11, 54 },
		{ 18, 12, 54 },
		{ 18, 13, 54 },
		{ 18, 14, 54 },
		{ 18, 15, 54 },
		{ 18, 16, 54 },
		{ 18, 17, 54 },
		{ 18, 18, 54 },
		{ 18, 19, 54 },
		{ 18, 20, 54 },
		{ 18, 21, 54 },
		{ 18, 22, 54 },
		{ 18, 23, 54 },
		{ 18, 24, 54 },
		{ 18, 25, 54 },
		{ 18, 26, 54 },
		{ 18, 27, 54 },
		{ 18, 28, 54 },
		{ 18, 29, 54 },
		{ 18, 30, 54 },
		{ 18, 31, 54 },
		{ 18, 32, 54 },
		{ 18, 33, 54 },
		{ 18, 34, 54 },
		{ 18, 35, 54 },
		{ 18, 36, 54 },
		{ 18, 37, 54 },
		{ 18, 38, 54 },
		{ 18, 39, 54 },
		{ 18, 40, 54 },
		{ 18, 41, 54 },
		{ 18, 42, 54 },
		{ 18, 43, 54 },
		{ 18, 44, 54 },
		{ 18, 45, 54 },
		{ 18, 46, 54 },
		{ 18, 47, 54 },
		{ 19, 1, 54 },
		{ 19, 2, 54 },
		{ 19, 3, 54 },
		{ 19, 4, 54 },
		{ 19, 5, 54 },
		{ 19, 6, 54 },
		{ 19, 7, 54 },
		{ 19, 8, 54 },
		{ 19, 9, 54 },
		{ 19, 10, 54 },
		{ 19, 11, 54 },
		{ 19, 12, 54 },
		{ 19, 13, 54 },
		{ 19, 14, 54 },
		{ 19, 15, 54 },
		{ 19, 16, 54 },
		{ 19, 17, 54 },
		{ 19, 18, 54 },
		{ 19, 19, 54 },
		{ 19, 20, 54 },
		{ 19, 21, 54 },
		{ 19, 22, 54 },
		{ 19, 23, 54 },
		{ 19, 24, 54 },
		{ 19, 25, 54 },
		{ 19, 26, 54 },
		{ 19, 27, 54 },
		{ 19, 28, 54 },
		{ 19, 29, 54 },
		{ 19, 30, 54 },
		{ 19, 31, 54 },
		{ 19, 32, 54 },
		{ 19, 33, 54 },
		{ 19, 34, 54 },
		{ 19, 35, 54 },
		{ 19, 36, 54 },
		{ 19, 37, 54 },
		{ 19, 38, 54 },
		{ 19, 39, 54 },
		{ 19, 40, 54 },
		{ 19, 41, 54 },
		{ 19, 42, 54 },
		{ 19, 43, 54 },
		{ 19, 44, 54 },
		{ 19, 45, 54 },
		{ 19, 46, 54 },
		{ 19, 47, 54 },
		{ 19, 48, 54 },
		{ 20, 1, 54 },
		{ 20, 2, 54 },
		{ 20, 3, 54 },
		{ 20, 4, 54 },
		{ 20, 5, 54 },
		{ 20, 6, 54 },
		{ 20, 7, 54 },
		{ 20, 8, 54 },
		{ 20, 9, 54 },
		{ 20, 10, 54 },
		{ 20, 11, 54 },
		{ 20, 12, 54 },
		{ 20, 13, 54 },
		{ 20, 14, 54 },
		{ 20, 15, 54 },
		{ 20, 16, 54 },
		{ 20, 17, 54 },
		{ 20, 18, 54 },
		{ 20, 19, 54 },
		{ 20, 20, 54 },
		{ 20, 21, 54 },
		{ 20, 22, 54 },
		{ 20, 23, 54 },
		{ 20, 24, 54 },
		{ 20, 25, 54 },
		{ 20, 26, 54 },
		{ 20, 27, 54 },
		{ 20, 28, 54 },
		{ 20, 29, 54 },
		{ 20, 30, 54 },
		{ 20, 31, 54 },
		{ 20, 32, 54 },
		{ 20, 33, 54 },
		{ 20, 34, 54 },
		{ 20, 35, 54 },
		{ 20, 36, 54 },
		{ 20, 37, 54 },
		{ 20, 38, 54 },
		{ 20, 39, 54 },
		{ 20, 40, 54 },
		{ 20, 41, 54 },
		{ 20, 42, 54 },
		{ 20, 43, 54 },
		{ 20, 44, 54 },
		{ 20, 45, 54 },
		{ 20, 46, 54 },
		{ 20, 47, 54 },
		{ 20, 48, 54 },
		{ 20, 49, 54 },
		{ 21, 1, 54 },
		{ 21, 2, 54 },
		{ 21, 3, 54 },
		{ 21, 4, 54 },
		{ 21, 5, 54 },
		{ 21, 6, 54 },
		{ 21, 7, 54 },
		{ 21, 8, 54 },
		{ 21, 9, 54 },
		{ 21, 10, 54 },
		{ 21, 11, 54 },
		{ 21, 12, 54 },
		{ 21, 13, 54 },
		{ 21, 14, 54 },
		{ 21, 15, 54 },
		{ 21, 16, 54 },
		{ 21, 17, 54 },
		{ 21, 18, 54 },
		{ 21, 19, 54 },
		{ 21, 20, 54 },
		{ 21, 21, 54 },
		{ 21, 22, 54 },
		{ 21, 23, 54 },
		{ 21, 24, 54 },
		{ 21, 25, 54 },
		{ 21, 26, 54 },
		{ 21, 27, 54 },
		{ 21, 28, 54 },
		{ 21, 29, 54 },
		{ 21, 30, 54 },
		{ 21, 31, 54 },
		{ 21, 32, 54 },
		{ 21, 33, 54 },
		{ 21, 34, 54 },
		{ 21, 35, 54 },
		{ 21, 36, 54 },
		{ 21, 37, 54 },
		{ 21, 38, 54 },
		{ 21, 39, 54 },
		{ 21, 40, 54 },
		{ 21, 41, 54 },
		{ 21, 42, 54 },
		{ 21, 43, 54 },
		{ 21, 44, 54 },
		{ 21, 45, 54 },
		{ 21, 46, 54 },
		{ 21, 47, 54 },
		{ 21, 48, 54 },
		{ 21, 49, 54 },
		{ 21, 50, 54 },
		{ 22, 1, 54 },
		{ 22, 2, 54 },
		{ 22, 3, 54 },
		{ 22, 4, 54 },
		{ 22, 5, 54 },
		{ 22, 6, 54 },
		{ 22, 7, 54 },
		{ 22, 8, 54 },
		{ 22, 9, 54 },
		{ 22, 10, 54 },
		{ 22, 11, 54 },
		{ 22, 12, 54 },
		{ 22, 13, 54 },
		{ 22, 14, 54 },
		{ 22, 15, 54 },
		{ 22, 16, 54 },
		{ 22, 17, 54 },
		{ 22, 18, 54 },
		{ 22, 19, 54 },
		{ 22, 20, 54 },
		{ 22, 21, 54 },
		{ 22, 22, 54 },
		{ 22, 23, 54 },
		{ 22, 24, 54 },
		{ 22, 25, 54 },
		{ 22, 26, 54 },
		{ 22, 27, 54 },
		{ 22, 28, 54 },
		{ 22, 29, 54 },
		{ 22, 30, 54 },
		{ 22, 31, 54 },
		{ 22, 32, 54 },
		{ 22, 33, 54 },
		{ 22, 34, 54 },
		{ 22, 35, 54 },
		{ 22, 36, 54 },
		{ 22, 37, 54 },
		{ 22, 38, 54 },
		{ 22, 39, 54 },
		{ 22, 40, 54 },
		{ 22, 41, 54 },
		{ 22, 42, 54 },
		{ 22, 43, 54 },
		{ 22, 44, 54 },
		{ 22, 45, 54 },
		{ 22, 46, 54 },
		{ 22, 47, 54 },
		{ 22, 48, 54 },
		{ 22, 49, 54 },
		{ 22, 50, 54 },
		{ 22, 51, 54 },
		{ 23, 1, 54 },
		{ 23, 2, 54 },
		{ 23, 3, 54 },
		{ 23, 4, 54 },
		{ 23, 5, 54 },
		{ 23, 6, 54 },
		{ 23, 7, 54 },
		{ 23, 8, 54 },
		{ 23, 9, 54 },
		{ 23, 10, 54 },
		{ 23, 11, 54 },
		{ 23, 12, 54 },
		{ 23, 13, 54 },
		{ 23, 14, 54 },
		{ 23, 15, 54 },
		{ 23, 16, 54 },
		{ 23, 17, 54 },
		{ 23, 18, 54 },
		{ 23, 19, 54 },
		{ 23, 20, 54 },
		{ 23, 21, 54 },
		{ 23, 22, 54 },
		{ 23, 23, 54 },
		{ 23, 24, 54 },
		{ 23, 25, 54 },
		{ 23, 26, 54 },
		{ 23, 27, 54 },
		{ 23, 28, 54 },
		{ 23, 29, 54 },
		{ 23, 30, 54 },
		{ 23, 31, 54 },
		{ 23, 32, 54 },
		{ 23, 33, 54 },
		{ 23, 34, 54 },
		{ 23, 35, 54 },
		{ 23, 36, 54 },
		{ 23, 37, 54 },
		{ 23, 38, 54 },
		{ 23, 39, 54 },
		{ 23, 40, 54 },
		{ 23, 41, 54 },
		{ 23, 42, 54 },
		{ 23, 43, 54 },
		{ 23, 44, 54 },
		{ 23, 45, 54 },
		{ 23, 46, 54 },
		{ 23, 47, 54 },
		{ 23, 48, 54 },
		{ 23, 49, 54 },
		{ 23, 50, 54 },
		{ 23, 51, 54 },
		{ 23, 52, 54 },
		{ 24, 1, 54 },
		{ 24, 2, 54 },
		{ 24, 3, 54 },
		{ 24, 4, 54 },
		{ 24, 5, 54 },
		{ 24, 6, 54 },
		{ 24, 7, 54 },
		{ 24, 8, 54 },
		{ 24, 9, 54 },
		{ 24, 10, 54 },
		{ 24, 11, 54 },
		{ 24, 12, 54 },
		{ 24, 13, 54 },
		{ 24, 14, 54 },
		{ 24, 15, 54 },
		{ 24, 16, 54 },
		{ 24, 17, 54 },
		{ 24, 18, 54 },
		{ 24, 19, 54 },
		{ 24, 20, 54 },
		{ 24, 21, 54 },
		{ 24, 22, 54 },
		{ 24, 23, 54 },
		{ 24, 24, 54 },
		{ 24, 25, 54 },
		{ 24, 26, 54 },
		{ 24, 27, 54 },
		{ 24, 28, 54 },
		{ 24, 29, 54 },
		{ 24, 30, 54 },
		{ 24, 31, 54 },
		{ 24, 32, 54 },
		{ 24, 33, 54 },
		{ 24, 34, 54 },
		{ 24, 35, 54 },
		{ 24, 36, 54 },
		{ 24, 37, 54 },
		{ 24, 38, 54 },
		{ 24, 39, 54 },
		{ 24, 40, 54 },
		{ 24, 41, 54 },
		{ 24, 42, 54 },
		{ 24, 43, 54 },
		{ 24, 44, 54 },
		{ 24, 45, 54 },
		{ 24, 46, 54 },
		{ 24, 47, 54 },
		{ 24, 48, 54 },
		{ 24, 49, 54 },
		{ 24, 50, 54 },
		{ 24, 51, 54 },
		{ 24, 52, 54 },
		{ 24, 53, 54 },
		{ 25, 1, 54 },
		{ 25, 2, 54 },
		{ 25, 3, 54 },
		{ 25, 4, 54 },
		{ 25, 5, 54 },
		{ 25, 6, 54 },
		{ 25, 7, 54 },
		{ 25, 8, 54 },
		{ 25, 9, 54 },
		{ 25, 10, 54 },
		{ 25, 11, 54 },
		{ 25, 12, 54 },
		{ 25, 13, 54 },
		{ 25, 14, 54 },
		{ 25, 15, 54 },
		{ 25, 16, 54 },
		{ 25, 17, 54 },
		{ 25, 18, 54 },
		{ 25, 19, 54 },
		{ 25, 20, 54 },
		{ 25, 21, 54 },
		{ 25, 22, 54 },
		{ 25, 23, 54 },
		{ 25, 24, 54 },
		{ 25, 25, 54 },
		{ 25, 26, 54 },
		{ 25, 27, 54 },
		{ 25, 28, 54 },
		{ 25, 29, 54 },
		{ 25, 30, 54 },
		{ 25, 31, 54 },
		{ 25, 32, 54 },
		{ 25, 33, 54 },
		{ 25, 34, 54 },
		{ 25, 35, 54 },
		{ 25, 36, 54 },
		{ 25, 37, 54 },
		{ 25, 38, 54 },
		{ 25, 39, 54 },
		{ 25, 40, 54 },
		{ 25, 41, 54 },
		{ 25, 42, 54 },
		{ 25, 43, 54 },
		{ 25, 44, 54 },
		{ 25, 45, 54 },
		{ 25, 46, 54 },
		{ 25, 47, 54 },
		{ 25, 48, 54 },
		{ 25, 49, 54 },
		{ 25, 50, 54 },
		{ 25, 51, 54 },
		{ 25, 52, 54 },
		{ 25, 53, 54 },
		{ 26, 1, 54 },
		{ 26, 2, 54 },
		{ 26, 3, 54 },
		{ 26, 4, 54 },
		{ 26, 5, 54 },
		{ 26, 6, 54 },
		{ 26, 7, 54 },
		{ 26, 8, 54 },
		{ 26, 9, 54 },
		{ 26, 10, 54 },
		{ 26, 11, 54 },
		{ 26, 12, 54 },
		{ 26, 13, 54 },
		{ 26, 14, 54 },
		{ 26, 15, 54 },
		{ 26, 16, 54 },
		{ 26, 17, 54 },
		{ 26, 18, 54 },
		{ 26, 19, 54 },
		{ 26, 20, 54 },
		{ 26, 21, 54 },
		{ 26, 22, 54 },
		{ 26, 23, 54 },
		{ 26, 24, 54 },
		{ 26, 25, 54 },
		{ 26, 26, 54 },
		{ 26, 27, 54 },
		{ 26, 28, 54 },
		{ 26, 29, 54 },
		{ 26, 30, 54 },
		{ 26, 31, 54 },
		{ 26, 32, 54 },
		{ 26, 33, 54 },
		{ 26, 34, 54 },
		{ 26, 35, 54 },
		{ 26, 36, 54 },
		{ 26, 37, 54 },
		{ 26, 38, 54 },
		{ 26, 39, 54 },
		{ 26, 40, 54 },
		{ 26, 41, 54 },
		{ 26, 42, 54 },
		{ 26, 43, 54 },
		{ 26, 44, 54 },
		{ 26, 45, 54 },
		{ 26, 46, 54 },
		{ 26, 47, 54 },
		{ 26, 48, 54 },
		{ 26, 49, 54 },
		{ 26, 50, 54 },
		{ 26, 51, 54 },
		{ 26, 52, 54 },
		{ 26, 53, 54 },
		{ 26, 54, 54 },
		{ 27, 1, 54 },
		{ 27, 2, 54 },
		{ 27, 3, 54 },
		{ 27, 4, 54 },
		{ 27, 5, 54 },
		{ 27, 6, 54 },
		{ 27, 7, 54 },
		{ 27, 8, 54 },
		{ 27, 9, 54 },
		{ 27, 10, 54 },
		{ 27, 11, 54 },
		{ 27, 12, 54 },
		{ 27, 13, 54 },
		{ 27, 14, 54 },
		{ 27, 15, 54 },
		{ 27, 16, 54 },
		{ 27, 17, 54 },
		{ 27, 18, 54 },
		{ 27, 19, 54 },
		{ 27, 20, 54 },
		{ 27, 21, 54 },
		{ 27, 22, 54 },
		{ 27, 23, 54 },
		{ 27, 24, 54 },
		{ 27, 25, 54 },
		{ 27, 26, 54 },
		{ 27, 27, 54 },
		{ 27, 28, 54 },
		{ 27, 29, 54 },
		{ 27, 30, 54 },
		{ 27, 31, 54 },
		{ 27, 32, 54 },
		{ 27, 33, 54 },
		{ 27, 34, 54 },
		{ 27, 35, 54 },
		{ 27, 36, 54 },
		{ 27, 37, 54 },
		{ 27, 38, 54 },
		{ 27, 39, 54 },
		{ 27, 40, 54 },
		{ 27, 41, 54 },
		{ 27, 42, 54 },
		{ 27, 43, 54 },
		{ 27, 44, 54 },
		{ 27, 45, 54 },
		{ 27, 46, 54 },
		{ 27, 47, 54 },
		{ 27, 48, 54 },
		{ 27, 49, 54 },
		{ 27, 50, 54 },
		{ 27, 51, 54 },
		{ 27, 52, 54 },
		{ 27, 53, 54 },
		{ 27, 54, 54 },
		{ 27, 55, 54 },
		{ 28, 1, 54 },
		{ 28, 2, 54 },
		{ 28, 3, 54 },
		{ 28, 4, 54 },
		{ 28, 5, 54 },
		{ 28, 6, 54 },
		{ 28, 7, 54 },
		{ 28, 8, 54 },
		{ 28, 9, 54 },
		{ 28, 10, 54 },
		{ 28, 11, 54 },
		{ 28, 12, 54 },
		{ 28, 13, 54 },
		{ 28, 14, 54 },
		{ 28, 15, 54 },
		{ 28, 16, 54 },
		{ 28, 17, 54 },
		{ 28, 18, 54 },
		{ 28, 19, 54 },
		{ 28, 20, 54 },
		{ 28, 21, 54 },
		{ 28, 22, 54 },
		{ 28, 23, 54 },
		{ 28, 24, 54 },
		{ 28, 25, 54 },
		{ 28, 26, 54 },
		{ 28, 27, 54 },
		{ 28, 28, 54 },
		{ 28, 29, 54 },
		{ 28, 30, 54 },
		{ 28, 31, 54 },
		{ 28, 32, 54 },
		{ 28, 33, 54 },
		{ 28, 34, 54 },
		{ 28, 35, 54 },
		{ 28, 36, 54 },
		{ 28, 37, 54 },
		{ 28, 38, 54 },
		{ 28, 39, 54 },
		{ 28, 40, 54 },
		{ 28, 41, 54 },
		{ 28, 42, 54 },
		{ 28, 43, 54 },
		{ 28, 44, 54 },
		{ 28, 45, 54 },
		{ 28, 46, 54 },
		{ 28, 47, 54 },
		{ 28, 48, 54 },
		{ 28, 49, 54 },
		{ 28, 50, 54 },
		{ 28, 51, 54 },
		{ 28, 52, 54 },
		{ 28, 53, 54 },
		{ 28, 54, 54 },
		{ 28, 55, 54 },
		{ 28, 56, 54 },
		{ 29, 1, 54 },
		{ 29, 2, 54 },
		{ 29, 3, 54 },
		{ 29, 4, 54 },
		{ 29, 5, 54 },
		{ 29, 6, 54 },
		{ 29, 7, 54 },
		{ 29, 8, 54 },
		{ 29, 9, 54 },
		{ 29, 10, 54 },
		{ 29, 11, 54 },
		{ 29, 12, 54 },
		{ 29, 13, 54 },
		{ 29, 14, 54 },
		{ 29, 15, 54 },
		{ 29, 16, 54 },
		{ 29, 17, 54 },
		{ 29, 18, 54 },
		{ 29, 19, 54 },
		{ 29, 20, 54 },
		{ 29, 21, 54 },
		{ 29, 22, 54 },
		{ 29, 23, 54 },
		{ 29, 24, 54 },
		{ 29, 25, 54 },
		{ 29, 26, 54 },
		{ 29, 27, 54 },
		{ 29, 28, 54 },
		{ 29, 29, 54 },
		{ 29, 30, 54 },
		{ 29, 31, 54 },
		{ 29, 32, 54 },
		{ 29, 33, 54 },
		{ 29, 34, 54 },
		{ 29, 35, 54 },
		{ 29, 36, 54 },
		{ 29, 37, 54 },
		{ 29, 38, 54 },
		{ 29, 39, 54 },
		{ 29, 40, 54 },
		{ 29, 41, 54 },
		{ 29, 42, 54 },
		{ 29, 43, 54 },
		{ 29, 44, 54 },
		{ 29, 45, 54 },
		{ 29, 46, 54 },
		{ 29, 47, 54 },
		{ 29, 48, 54 },
		{ 29, 49, 54 },
		{ 29, 50, 54 },
		{ 29, 51, 54 },
		{ 29, 52, 54 },
		{ 29, 53, 54 },
		{ 29, 54, 54 },
		{ 29, 55, 54 },
		{ 29, 56, 54 },
		{ 29, 57, 54 },
		{ 30, 1, 54 },
		{ 30, 2, 54 },
		{ 30, 3, 54 },
		{ 30, 4, 54 },
		{ 30, 5, 54 },
		{ 30, 6, 54 },
		{ 30, 7, 54 },
		{ 30, 8, 54 },
		{ 30, 9, 54 },
		{ 30, 10, 54 },
		{ 30, 11, 54 },
		{ 30, 12, 54 },
		{ 30, 13, 54 },
		{ 30, 14, 54 },
		{ 30, 15, 54 },
		{ 30, 16, 54 },
		{ 30, 17, 54 },
		{ 30, 18, 54 },
		{ 30, 19, 54 },
		{ 30, 20, 54 },
		{ 30, 21, 54 },
		{ 30, 22, 54 },
		{ 30, 23, 54 },
		{ 30, 24, 54 },
		{ 30, 25, 54 },
		{ 30, 26, 54 },
		{ 30, 27, 54 },
		{ 30, 28, 54 },
		{ 30, 29, 54 },
		{ 30, 30, 54 },
		{ 30, 31, 54 },
		{ 30, 32, 54 },
		{ 30, 33, 54 },
		{ 30, 34, 54 },
		{ 30, 35, 54 },
		{ 30, 36, 54 },
		{ 30, 37, 54 },
		{ 30, 38, 54 },
		{ 30, 39, 54 },
		{ 30, 40, 54 },
		{ 30, 41, 54 },
		{ 30, 42, 54 },
		{ 30, 43, 54 },
		{ 30, 44, 54 },
		{ 30, 45, 54 },
		{ 30, 46, 54 },
		{ 30, 47, 54 },
		{ 30, 48, 54 },
		{ 30, 49, 54 },
		{ 30, 50, 54 },
		{ 30, 51, 54 },
		{ 30, 52, 54 },
		{ 30, 53, 54 },
		{ 30, 54, 54 },
		{ 30, 55, 54 },
		{ 30, 56, 54 },
		{ 30, 57, 54 },
		{ 30, 58, 54 },
		{ 31, 1, 54 },
		{ 31, 2, 54 },
		{ 31, 3, 54 },
		{ 31, 4, 54 },
		{ 31, 5, 54 },
		{ 31, 6, 54 },
		{ 31, 7, 54 },
		{ 31, 8, 54 },
		{ 31, 9, 54 },
		{ 31, 10, 54 },
		{ 31, 11, 54 },
		{ 31, 12, 54 },
		{ 31, 13, 54 },
		{ 31, 14, 54 },
		{ 31, 15, 54 },
		{ 31, 16, 54 },
		{ 31, 17, 54 },
		{ 31, 18, 54 },
		{ 31, 19, 54 },
		{ 31, 20, 54 },
		{ 31, 21, 54 },
		{ 31, 22, 54 },
		{ 31, 23, 54 },
		{ 31, 24, 54 },
		{ 31, 25, 54 },
		{ 31, 26, 54 },
		{ 31, 27, 54 },
		{ 31, 28, 54 },
		{ 31, 29, 54 },
		{ 31, 30, 54 },
		{ 31, 31, 54 },
		{ 31, 32, 54 },
		{ 31, 33, 54 },
		{ 31, 34, 54 },
		{ 31, 35, 54 },
		{ 31, 36, 54 },
		{ 31, 37, 54 },
		{ 31, 38, 54 },
		{ 31, 39, 54 },
		{ 31, 40, 54 },
		{ 31, 41, 54 },
		{ 31, 42, 54 },
		{ 31, 43, 54 },
		{ 31, 44, 54 },
		{ 31, 45, 54 },
		{ 31, 46, 54 },
		{ 31, 47, 54 },
		{ 31, 48, 54 },
		{ 31, 49, 54 },
		{ 31, 50, 54 },
		{ 31, 51, 54 },
		{ 31, 52, 54 },
		{ 31, 53, 54 },
		{ 31, 54, 54 },
		{ 31, 55, 54 },
		{ 31, 56, 54 },
		{ 31, 57, 54 },
		{ 31, 58, 54 },
		{ 31, 59, 54 },
		{ 32, 1, 54 },
		{ 32, 2, 54 },
		{ 32, 3, 54 },
		{ 32, 4, 54 },
		{ 32, 5, 54 },
		{ 32, 6, 54 },
		{ 32, 7, 54 },
		{ 32, 8, 54 },
		{ 32, 9, 54 },
		{ 32, 10, 54 },
		{ 32, 11, 54 },
		{ 32, 12, 54 },
		{ 32, 13, 54 },
		{ 32, 14, 54 },
		{ 32, 15, 54 },
		{ 32, 16, 54 },
		{ 32, 17, 54 },
		{ 32, 18, 54 },
		{ 32, 19, 54 },
		{ 32, 20, 54 },
		{ 32, 21, 54 },
		{ 32, 22, 54 },
		{ 32, 23, 54 },
		{ 32, 24, 54 },
		{ 32, 25, 54 },
		{ 32, 26, 54 },
		{ 32, 27, 54 },
		{ 32, 28, 54 },
		{ 32, 29, 54 },
		{ 32, 30, 54 },
		{ 32, 31, 54 },
		{ 32, 32, 54 },
		{ 32, 33, 54 },
		{ 32, 34, 54 },
		{ 32, 35, 54 },
		{ 32, 36, 54 },
		{ 32, 37, 54 },
		{ 32, 38, 54 },
		{ 32, 39, 54 },
		{ 32, 40, 54 },
		{ 32, 41, 54 },
		{ 32, 42, 54 },
		{ 32, 43, 54 },
		{ 32, 44, 54 },
		{ 32, 45, 54 },
		{ 32, 46, 54 },
		{ 32, 47, 54 },
		{ 32, 48, 54 },
		{ 32, 49, 54 },
		{ 32, 50, 54 },
		{ 32, 51, 54 },
		{ 32, 52, 54 },
		{ 32, 53, 54 },
		{ 32, 54, 54 },
		{ 32, 55, 54 },
		{ 32, 56, 54 },
		{ 32, 57, 54 },
		{ 32, 58, 54 },
		{ 32, 59, 54 },
		{ 32, 60, 54 },
		{ 33, 2, 54 },
		{ 33, 3, 54 },
		{ 33, 4, 54 },
		{ 33, 5, 54 },
		{ 33, 6, 54 },
		{ 33, 7, 54 },
		{ 33, 8, 54 },
		{ 33, 9, 54 },
		{ 33, 10, 54 },
		{ 33, 11, 54 },
		{ 33, 12, 54 },
		{ 33, 13, 54 },
		{ 33, 14, 54 },
		{ 33, 15, 54 },
		{ 33, 16, 54 },
		{ 33, 17, 54 },
		{ 33, 18, 54 },
		{ 33, 19, 54 },
		{ 33, 20, 54 },
		{ 33, 21, 54 },
		{ 33, 22, 54 },
		{ 33, 23, 54 },
		{ 33, 24, 54 },
		{ 33, 25, 54 },
		{ 33, 26, 54 },
		{ 33, 27, 54 },
		{ 33, 28, 54 },
		{ 33, 29, 54 },
		{ 33, 30, 54 },
		{ 33, 31, 54 },
		{ 33, 32, 54 },
		{ 33, 33, 54 },
		{ 33, 34, 54 },
		{ 33, 35, 54 },
		{ 33, 36, 54 },
		{ 33, 37, 54 },
		{ 33, 38, 54 },
		{ 33, 39, 54 },
		{ 33, 40, 54 },
		{ 33, 41, 54 },
		{ 33, 42, 54 },
		{ 33, 43, 54 },
		{ 33, 44, 54 },
		{ 33, 45, 54 },
		{ 33, 46, 54 },
		{ 33, 47, 54 },
		{ 33, 48, 54 },
		{ 33, 49, 54 },
		{ 33, 50, 54 },
		{ 33, 51, 54 },
		{ 33, 52, 54 },
		{ 33, 53, 54 },
		{ 33, 54, 54 },
		{ 33, 55, 54 },
		{ 33, 56, 54 },
		{ 33, 57, 54 },
		{ 33, 58, 54 },
		{ 33, 59, 54 },
		{ 33, 60, 54 },
		{ 34, 3, 54 },
		{ 34, 4, 54 },
		{ 34, 5, 54 },
		{ 34, 6, 54 },
		{ 34, 7, 54 },
		{ 34, 8, 54 },
		{ 34, 9, 54 },
		{ 34, 10, 54 },
		{ 34, 11, 54 },
		{ 34, 12, 54 },
		{ 34, 13, 54 },
		{ 34, 14, 54 },
		{ 34, 15, 54 },
		{ 34, 16, 54 },
		{ 34, 17, 54 },
		{ 34, 18, 54 },
		{ 34, 19, 54 },
		{ 34, 20, 54 },
		{ 34, 21, 54 },
		{ 34, 22, 54 },
		{ 34, 23, 54 },
		{ 34, 24, 54 },
		{ 34, 25, 54 },
		{ 34, 26, 54 },
		{ 34, 27, 54 },
		{ 34, 28, 54 },
		{ 34, 29, 54 },
		{ 34, 30, 54 },
		{ 34, 31, 54 },
		{ 34, 32, 54 },
		{ 34, 33, 54 },
		{ 34, 34, 54 },
		{ 34, 35, 54 },
		{ 34, 36, 54 },
		{ 34, 37, 54 },
		{ 34, 38, 54 },
		{ 34, 39, 54 },
		{ 34, 40, 54 },
		{ 34, 41, 54 },
		{ 34, 42, 54 },
		{ 34, 43, 54 },
		{ 34, 44, 54 },
		{ 34, 45, 54 },
		{ 34, 46, 54 },
		{ 34, 47, 54 },
		{ 34, 48, 54 },
		{ 34, 49, 54 },
		{ 34, 50, 54 },
		{ 34, 51, 54 },
		{ 34, 52, 54 },
		{ 34, 53, 54 },
		{ 34, 54, 54 },
		{ 34, 55, 54 },
		{ 34, 56, 54 },
		{ 34, 57, 54 },
		{ 34, 58, 54 },
		{ 34, 59, 54 },
		{ 34, 60, 54 },
		{ 34, 61, 54 },
		{ 35, 4, 54 },
		{ 35, 5, 54 },
		{ 35, 6, 54 },
		{ 35, 7, 54 },
		{ 35, 8, 54 },
		{ 35, 9, 54 },
		{ 35, 10, 54 },
		{ 35, 11, 54 },
		{ 35, 12, 54 },
		{ 35, 13, 54 },
		{ 35, 14, 54 },
		{ 35, 15, 54 },
		{ 35, 16, 54 },
		{ 35, 17, 54 },
		{ 35, 18, 54 },
		{ 35, 19, 54 },
		{ 35, 20, 54 },
		{ 35, 21, 54 },
		{ 35, 22, 54 },
		{ 35, 23, 54 },
		{ 35, 24, 54 },
		{ 35, 25, 54 },
		{ 35, 26, 54 },
		{ 35, 27, 54 },
		{ 35, 28, 54 },
		{ 35, 29, 54 },
		{ 35, 30, 54 },
		{ 35, 31, 54 },
		{ 35, 32, 54 },
		{ 35, 33, 54 },
		{ 35, 34, 54 },
		{ 35, 35, 54 },
		{ 35, 36, 54 },
		{ 35, 37, 54 },
		{ 35, 38, 54 },
		{ 35, 39, 54 },
		{ 35, 40, 54 },
		{ 35, 41, 54 },
		{ 35, 42, 54 },
		{ 35, 43, 54 },
		{ 35, 44, 54 },
		{ 35, 45, 54 },
		{ 35, 46, 54 },
		{ 35, 47, 54 },
		{ 35, 48, 54 },
		{ 35, 49, 54 },
		{ 35, 50, 54 },
		{ 35, 51, 54 },
		{ 35, 52, 54 },
		{ 35, 53, 54 },
		{ 35, 54, 54 },
		{ 35, 55, 54 },
		{ 35, 56, 54 },
		{ 35, 57, 54 },
		{ 35, 58, 54 },
		{ 35, 59, 54 },
		{ 35, 60, 54 },
		{ 35, 61, 54 },
		{ 35, 62, 54 },
		{ 36, 5, 54 },
		{ 36, 6, 54 },
		{ 36, 7, 54 },
		{ 36, 8, 54 },
		{ 36, 9, 54 },
		{ 36, 10, 54 },
		{ 36, 11, 54 },
		{ 36, 12, 54 },
		{ 36, 13, 54 },
		{ 36, 14, 54 },
		{ 36, 15, 54 },
		{ 36, 16, 54 },
		{ 36, 17, 54 },
		{ 36, 18, 54 },
		{ 36, 19, 54 },
		{ 36, 20, 54 },
		{ 36, 21, 54 },
		{ 36, 22, 54 },
		{ 36, 23, 54 },
		{ 36, 24, 54 },
		{ 36, 25, 54 },
		{ 36, 26, 54 },
		{ 36, 27, 54 },
		{ 36, 28, 54 },
		{ 36, 29, 54 },
		{ 36, 30, 54 },
		{ 36, 31, 54 },
		{ 36, 32, 54 },
		{ 36, 33, 54 },
		{ 36, 34, 54 },
		{ 36, 35, 54 },
		{ 36, 36, 54 },
		{ 36, 37, 54 },
		{ 36, 38, 54 },
		{ 36, 39, 54 },
		{ 36, 40, 54 },
		{ 36, 41, 54 },
		{ 36, 42, 54 },
		{ 36, 43, 54 },
		{ 36, 44, 54 },
		{ 36, 45, 54 },
		{ 36, 46, 54 },
		{ 36, 47, 54 },
		{ 36, 48, 54 },
		{ 36, 49, 54 },
		{ 36, 50, 54 },
		{ 36, 51, 54 },
		{ 36, 52, 54 },
		{ 36, 53, 54 },
		{ 36, 54, 54 },
		{ 36, 55, 54 },
		{ 36, 56, 54 },
		{ 36, 57, 54 },
		{ 36, 58, 54 },
		{ 36, 59, 54 },
		{ 36, 60, 54 },
		{ 36, 61, 54 },
		{ 36, 62, 54 },
		{ 36, 63, 54 },
		{ 37, 6, 54 },
		{ 37, 7, 54 },
		{ 37, 8, 54 },
		{ 37, 9, 54 },
		{ 37, 10, 54 },
		{ 37, 11, 54 },
		{ 37, 12, 54 },
		{ 37, 13, 54 },
		{ 37, 14, 54 },
		{ 37, 15, 54 },
		{ 37, 16, 54 },
		{ 37, 17, 54 },
		{ 37, 18, 54 },
		{ 37, 19, 54 },
		{ 37, 20, 54 },
		{ 37, 21, 54 },
		{ 37, 22, 54 },
		{ 37, 23, 54 },
		{ 37, 24, 54 },
		{ 37, 25, 54 },
		{ 37, 26, 54 },
		{ 37, 27, 54 },
		{ 37, 28, 54 },
		{ 37, 29, 54 },
		{ 37, 30, 54 },
		{ 37, 31, 54 },
		{ 37, 32, 54 },
		{ 37, 33, 54 },
		{ 37, 34, 54 },
		{ 37, 35, 54 },
		{ 37, 36, 54 },
		{ 37, 37, 54 },
		{ 37, 38, 54 },
		{ 37, 39, 54 },
		{ 37, 40, 54 },
		{ 37, 41, 54 },
		{ 37, 42, 54 },
		{ 37, 43, 54 },
		{ 37, 44, 54 },
		{ 37, 45, 54 },
		{ 37, 46, 54 },
		{ 37, 47, 54 },
		{ 37, 48, 54 },
		{ 37, 49, 54 },
		{ 37, 50, 54 },
		{ 37, 51, 54 },
		{ 37, 52, 54 },
		{ 37, 53, 54 },
		{ 37, 54, 54 },
		{ 37, 55, 54 },
		{ 37, 56, 54 },
		{ 37, 57, 54 },
		{ 37, 58, 54 },
		{ 37, 59, 54 },
		{ 37, 60, 54 },
		{ 37, 61, 54 },
		{ 37, 62, 54 },
		{ 37, 63, 54 },
		{ 37, 64, 54 },
		{ 38, 7, 54 },
		{ 38, 8, 54 },
		{ 38, 9, 54 },
		{ 38, 10, 54 },
		{ 38, 11, 54 },
		{ 38, 12, 54 },
		{ 38, 13, 54 },
		{ 38, 14, 54 },
		{ 38, 15, 54 },
		{ 38, 16, 54 },
		{ 38, 17, 54 },
		{ 38, 18, 54 },
		{ 38, 19, 54 },
		{ 38, 20, 54 },
		{ 38, 21, 54 },
		{ 38, 22, 54 },
		{ 38, 23, 54 },
		{ 38, 24, 54 },
		{ 38, 25, 54 },
		{ 38, 26, 54 },
		{ 38, 27, 54 },
		{ 38, 28, 54 },
		{ 38, 29, 54 },
		{ 38, 30, 54 },
		{ 38, 31, 54 },
		{ 38, 32, 54 },
		{ 38, 33, 54 },
		{ 38, 34, 54 },
		{ 38, 35, 54 },
		{ 38, 36, 54 },
		{ 38, 37, 54 },
		{ 38, 38, 54 },
		{ 38, 39, 54 },
		{ 38, 40, 54 },
		{ 38, 41, 54 },
		{ 38, 42, 54 },
		{ 38, 43, 54 },
		{ 38, 44, 54 },
		{ 38, 45, 54 },
		{ 38, 46, 54 },
		{ 38, 47, 54 },
		{ 38, 48, 54 },
		{ 38, 49, 54 },
		{ 38, 50, 54 },
		{ 38, 51, 54 },
		{ 38, 52, 54 },
		{ 38, 53, 54 },
		{ 38, 54, 54 },
		{ 38, 55, 54 },
		{ 38, 56, 54 },
		{ 38, 57, 54 },
		{ 38, 58, 54 },
		{ 38, 59, 54 },
		{ 38, 60, 54 },
		{ 38, 61, 54 },
		{ 38, 62, 54 },
		{ 38, 63, 54 },
		{ 38, 64, 54 },
		{ 38, 65, 54 },
		{ 39, 8, 54 },
		{ 39, 9, 54 },
		{ 39, 10, 54 },
		{ 39, 11, 54 },
		{ 39, 12, 54 },
		{ 39, 13, 54 },
		{ 39, 14, 54 },
		{ 39, 15, 54 },
		{ 39, 16, 54 },
		{ 39, 17, 54 },
		{ 39, 18, 54 },
		{ 39, 19, 54 },
		{ 39, 20, 54 },
		{ 39, 21, 54 },
		{ 39, 22, 54 },
		{ 39, 23, 54 },
		{ 39, 24, 54 },
		{ 39, 25, 54 },
		{ 39, 26, 54 },
		{ 39, 27, 54 },
		{ 39, 28, 54 },
		{ 39, 29, 54 },
		{ 39, 30, 54 },
		{ 39, 31, 54 },
		{ 39, 32, 54 },
		{ 39, 33, 54 },
		{ 39, 34, 54 },
		{ 39, 35, 54 },
		{ 39, 36, 54 },
		{ 39, 37, 54 },
		{ 39, 38, 54 },
		{ 39, 39, 54 },
		{ 39, 40, 54 },
		{ 39, 41, 54 },
		{ 39, 42, 54 },
		{ 39, 43, 54 },
		{ 39, 44, 54 },
		{ 39, 45, 54 },
		{ 39, 46, 54 },
		{ 39, 47, 54 },
		{ 39, 48, 54 },
		{ 39, 49, 54 },
		{ 39, 50, 54 },
		{ 39, 51, 54 },
		{ 39, 52, 54 },
		{ 39, 53, 54 },
		{ 39, 54, 54 },
		{ 39, 55, 54 },
		{ 39, 56, 54 },
		{ 39, 57, 54 },
		{ 39, 58, 54 },
		{ 39, 59, 54 },
		{ 39, 60, 54 },
		{ 39, 61, 54 },
		{ 39, 62, 54 },
		{ 39, 63, 54 },
		{ 39, 64, 54 },
		{ 39, 65, 54 },
		{ 39, 66, 54 },
		{ 40, 10, 54 },
		{ 40, 11, 54 },
		{ 40, 12, 54 },
		{ 40, 13, 54 },
		{ 40, 14, 54 },
		{ 40, 15, 54 },
		{ 40, 16, 54 },
		{ 40, 17, 54 },
		{ 40, 18, 54 },
		{ 40, 19, 54 },
		{ 40, 20, 54 },
		{ 40, 21, 54 },
		{ 40, 22, 54 },
		{ 40, 23, 54 },
		{ 40, 24, 54 },
		{ 40, 25, 54 },
		{ 40, 26, 54 },
		{ 40, 27, 54 },
		{ 40, 28, 54 },
		{ 40, 29, 54 },
		{ 40, 30, 54 },
		{ 40, 31, 54 },
		{ 40, 32, 54 },
		{ 40, 33, 54 },
		{ 40, 34, 54 },
		{ 40, 35, 54 },
		{ 40, 36, 54 },
		{ 40, 37, 54 },
		{ 40, 38, 54 },
		{ 40, 39, 54 },
		{ 40, 40, 54 },
		{ 40, 41, 54 },
		{ 40, 42, 54 },
		{ 40, 43, 54 },
		{ 40, 44, 54 },
		{ 40, 45, 54 },
		{ 40, 46, 54 },
		{ 40, 47, 54 },
		{ 40, 48, 54 },
		{ 40, 49, 54 },
		{ 40, 50, 54 },
		{ 40, 51, 54 },
		{ 40, 52, 54 },
		{ 40, 53, 54 },
		{ 40, 54, 54 },
		{ 40, 55, 54 },
		{ 40, 56, 54 },
		{ 40, 57, 54 },
		{ 40, 58, 54 },
		{ 40, 59, 54 },
		{ 40, 60, 54 },
		{ 40, 61, 54 },
		{ 40, 62, 54 },
		{ 40, 63, 54 },
		{ 40, 64, 54 },
		{ 40, 65, 54 },
		{ 40, 66, 54 },
		{ 40, 67, 54 },
		{ 41, 11, 54 },
		{ 41, 12, 54 },
		{ 41, 13, 54 },
		{ 41, 14, 54 },
		{ 41, 15, 54 },
		{ 41, 16, 54 },
		{ 41, 17, 54 },
		{ 41, 18, 54 },
		{ 41, 19, 54 },
		{ 41, 20, 54 },
		{ 41, 21, 54 },
		{ 41, 22, 54 },
		{ 41, 23, 54 },
		{ 41, 24, 54 },
		{ 41, 25, 54 },
		{ 41, 26, 54 },
		{ 41, 27, 54 },
		{ 41, 28, 54 },
		{ 41, 29, 54 },
		{ 41, 30, 54 },
		{ 41, 31, 54 },
		{ 41, 32, 54 },
		{ 41, 33, 54 },
		{ 41, 34, 54 },
		{ 41, 35, 54 },
		{ 41, 36, 54 },
		{ 41, 37, 54 },
		{ 41, 38, 54 },
		{ 41, 39, 54 },
		{ 41, 40, 54 },
		{ 41, 41, 54 },
		{ 41, 42, 54 },
		{ 41, 43, 54 },
		{ 41, 44, 54 },
		{ 41, 45, 54 },
		{ 41, 46, 54 },
		{ 41, 47, 54 },
		{ 41, 48, 54 },
		{ 41, 49, 54 },
		{ 41, 50, 54 },
		{ 41, 51, 54 },
		{ 41, 52, 54 },
		{ 41, 53, 54 },
		{ 41, 54, 54 },
		{ 41, 55, 54 },
		{ 41, 56, 54 },
		{ 41, 57, 54 },
		{ 41, 58, 54 },
		{ 41, 59, 54 },
		{ 41, 60, 54 },
		{ 41, 61, 54 },
		{ 41, 62, 54 },
		{ 41, 63, 54 },
		{ 41, 64, 54 },
		{ 41, 65, 54 },
		{ 41, 66, 54 },
		{ 41, 67, 54 },
		{ 42, 12, 54 },
		{ 42, 13, 54 },
		{ 42, 14, 54 },
		{ 42, 15, 54 },
		{ 42, 16, 54 },
		{ 42, 17, 54 },
		{ 42, 18, 54 },
		{ 42, 19, 54 },
		{ 42, 20, 54 },
		{ 42, 21, 54 },
		{ 42, 22, 54 },
		{ 42, 23, 54 },
		{ 42, 24, 54 },
		{ 42, 25, 54 },
		{ 42, 26, 54 },
		{ 42, 27, 54 },
		{ 42, 28, 54 },
		{ 42, 29, 54 },
		{ 42, 30, 54 },
		{ 42, 31, 54 },
		{ 42, 32, 54 },
		{ 42, 33, 54 },
		{ 42, 34, 54 },
		{ 42, 35, 54 },
		{ 42, 36, 54 },
		{ 42, 37, 54 },
		{ 42, 38, 54 },
		{ 42, 39, 54 },
		{ 42, 40, 54 },
		{ 42, 41, 54 },
		{ 42, 42, 54 },
		{ 42, 43, 54 },
		{ 42, 44, 54 },
		{ 42, 45, 54 },
		{ 42, 46, 54 },
		{ 42, 47, 54 },
		{ 42, 48, 54 },
		{ 42, 49, 54 },
		{ 42, 50, 54 },
		{ 42, 51, 54 },
		{ 42, 52, 54 },
		{ 42, 53, 54 },
		{ 42, 54, 54 },
		{ 42, 55, 54 },
		{ 42, 56, 54 },
		{ 42, 57, 54 },
		{ 42, 58, 54 },
		{ 42, 59, 54 },
		{ 42, 60, 54 },
		{ 42, 61, 54 },
		{ 42, 62, 54 },
		{ 42, 63, 54 },
		{ 42, 64, 54 },
		{ 42, 65, 54 },
		{ 42, 66, 54 },
		{ 42, 67, 54 },
		{ 42, 68, 54 },
		{ 43, 13, 54 },
		{ 43, 14, 54 },
		{ 43, 15, 54 },
		{ 43, 16, 54 },
		{ 43, 17, 54 },
		{ 43, 18, 54 },
		{ 43, 19, 54 },
		{ 43, 20, 54 },
		{ 43, 21, 54 },
		{ 43, 22, 54 },
		{ 43, 23, 54 },
		{ 43, 24, 54 },
		{ 43, 25, 54 },
		{ 43, 26, 54 },
		{ 43, 27, 54 },
		{ 43, 28, 54 },
		{ 43, 29, 54 },
		{ 43, 30, 54 },
		{ 43, 31, 54 },
		{ 43, 32, 54 },
		{ 43, 33, 54 },
		{ 43, 34, 54 },
		{ 43, 35, 54 },
		{ 43, 36, 54 },
		{ 43, 37, 54 },
		{ 43, 38, 54 },
		{ 43, 39, 54 },
		{ 43, 40, 54 },
		{ 43, 41, 54 },
		{ 43, 42, 54 },
		{ 43, 43, 54 },
		{ 43, 44, 54 },
		{ 43, 45, 54 },
		{ 43, 46, 54 },
		{ 43, 47, 54 },
		{ 43, 48, 54 },
		{ 43, 49, 54 },
		{ 43, 50, 54 },
		{ 43, 51, 54 },
		{ 43, 52, 54 },
		{ 43, 53, 54 },
		{ 43, 54, 54 },
		{ 43, 55, 54 },
		{ 43, 56, 54 },
		{ 43, 57, 54 },
		{ 43, 58, 54 },
		{ 43, 59, 54 },
		{ 43, 60, 54 },
		{ 43, 61, 54 },
		{ 43, 62, 54 },
		{ 43, 63, 54 },
		{ 43, 64, 54 },
		{ 43, 65, 54 },
		{ 43, 66, 54 },
		{ 43, 67, 54 },
		{ 43, 68, 54 },
		{ 43, 69, 54 },
		{ 44, 14, 54 },
		{ 44, 15, 54 },
		{ 44, 16, 54 },
		{ 44, 17, 54 },
		{ 44, 18, 54 },
		{ 44, 19, 54 },
		{ 44, 20, 54 },
		{ 44, 21, 54 },
		{ 44, 22, 54 },
		{ 44, 23, 54 },
		{ 44, 24, 54 },
		{ 44, 25, 54 },
		{ 44, 26, 54 },
		{ 44, 27, 54 },
		{ 44, 28, 54 },
		{ 44, 29, 54 },
		{ 44, 30, 54 },
		{ 44, 31, 54 },
		{ 44, 32, 54 },
		{ 44, 33, 54 },
		{ 44, 34, 54 },
		{ 44, 35, 54 },
		{ 44, 36, 54 },
		{ 44, 37, 54 },
		{ 44, 38, 54 },
		{ 44, 39, 54 },
		{ 44, 40, 54 },
		{ 44, 41, 54 },
		{ 44, 42, 54 },
		{ 44, 43, 54 },
		{ 44, 44, 54 },
		{ 44, 45, 54 },
		{ 44, 46, 54 },
		{ 44, 47, 54 },
		{ 44, 48, 54 },
		{ 44, 49, 54 },
		{ 44, 50, 54 },
		{ 44, 51, 54 },
		{ 44, 52, 54 },
		{ 44, 53, 54 },
		{ 44, 54, 54 },
		{ 44, 55, 54 },
		{ 44, 56, 54 },
		{ 44, 57, 54 },
		{ 44, 58, 54 },
		{ 44, 59, 54 },
		{ 44, 60, 54 },
		{ 44, 61, 54 },
		{ 44, 62, 54 },
		{ 44, 63, 54 },
		{ 44, 64, 54 },
		{ 44, 65, 54 },
		{ 44, 66, 54 },
		{ 44, 67, 54 },
		{ 44, 68, 54 },
		{ 44, 69, 54 },
		{ 44, 70, 54 },
		{ 45, 15, 54 },
		{ 45, 16, 54 },
		{ 45, 17, 54 },
		{ 45, 18, 54 },
		{ 45, 19, 54 },
		{ 45, 20, 54 },
		{ 45, 21, 54 },
		{ 45, 22, 54 },
		{ 45, 23, 54 },
		{ 45, 24, 54 },
		{ 45, 25, 54 },
		{ 45, 26, 54 },
		{ 45, 27, 54 },
		{ 45, 28, 54 },
		{ 45, 29, 54 },
		{ 45, 30, 54 },
		{ 45, 31, 54 },
		{ 45, 32, 54 },
		{ 45, 33, 54 },
		{ 45, 34, 54 },
		{ 45, 35, 54 },
		{ 45, 36, 54 },
		{ 45, 37, 54 },
		{ 45, 38, 54 },
		{ 45, 39, 54 },
		{ 45, 40, 54 },
		{ 45, 41, 54 },
		{ 45, 42, 54 },
		{ 45, 43, 54 },
		{ 45, 44, 54 },
		{ 45, 45, 54 },
		{ 45, 46, 54 },
		{ 45, 47, 54 },
		{ 45, 48, 54 },
		{ 45, 49, 54 },
		{ 45, 50, 54 },
		{ 45, 51, 54 },
		{ 45, 52, 54 },
		{ 45, 53, 54 },
		{ 45, 54, 54 },
		{ 45, 55, 54 },
		{ 45, 56, 54 },
		{ 45, 57, 54 },
		{ 45, 58, 54 },
		{ 45, 59, 54 },
		{ 45, 60, 54 },
		{ 45, 61, 54 },
		{ 45, 62, 54 },
		{ 45, 63, 54 },
		{ 45, 64, 54 },
		{ 45, 65, 54 },
		{ 45, 66, 54 },
		{ 45, 67, 54 },
		{ 45, 68, 54 },
		{ 45, 69, 54 },
		{ 45, 70, 54 },
		{ 45, 71, 54 },
		{ 46, 16, 54 },
		{ 46, 17, 54 },
		{ 46, 18, 54 },
		{ 46, 19, 54 },
		{ 46, 20, 54 },
		{ 46, 21, 54 },
		{ 46, 22, 54 },
		{ 46, 23, 54 },
		{ 46, 24, 54 },
		{ 46, 25, 54 },
		{ 46, 26, 54 },
		{ 46, 27, 54 },
		{ 46, 28, 54 },
		{ 46, 29, 54 },
		{ 46, 30, 54 },
		{ 46, 31, 54 },
		{ 46, 32, 54 },
		{ 46, 33, 54 },
		{ 46, 34, 54 },
		{ 46, 35, 54 },
		{ 46, 36, 54 },
		{ 46, 37, 54 },
		{ 46, 38, 54 },
		{ 46, 39, 54 },
		{ 46, 40, 54 },
		{ 46, 41, 54 },
		{ 46, 42, 54 },
		{ 46, 43, 54 },
		{ 46, 44, 54 },
		{ 46, 45, 54 },
		{ 46, 46, 54 },
		{ 46, 47, 54 },
		{ 46, 48, 54 },
		{ 46, 49, 54 },
		{ 46, 50, 54 },
		{ 46, 51, 54 },
		{ 46, 52, 54 },
		{ 46, 53, 54 },
		{ 46, 54, 54 },
		{ 46, 55, 54 },
		{ 46, 56, 54 },
		{ 46, 57, 54 },
		{ 46, 58, 54 },
		{ 46, 59, 54 },
		{ 46, 60, 54 },
		{ 46, 61, 54 },
		{ 46, 62, 54 },
		{ 46, 63, 54 },
		{ 46, 64, 54 },
		{ 46, 65, 54 },
		{ 46, 66, 54 },
		{ 46, 67, 54 },
		{ 46, 68, 54 },
		{ 46, 69, 54 },
		{ 46, 70, 54 },
		{ 46, 71, 54 },
		{ 46, 72, 54 },
		{ 47, 18, 54 },
		{ 47, 19, 54 },
		{ 47, 20, 54 },
		{ 47, 21, 54 },
		{ 47, 22, 54 },
		{ 47, 23, 54 },
		{ 47, 24, 54 },
		{ 47, 25, 54 },
		{ 47, 26, 54 },
		{ 47, 27, 54 },
		{ 47, 28, 54 },
		{ 47, 29, 54 },
		{ 47, 30, 54 },
		{ 47, 31, 54 },
		{ 47, 32, 54 },
		{ 47, 33, 54 },
		{ 47, 34, 54 },
		{ 47, 35, 54 },
		{ 47, 36, 54 },
		{ 47, 37, 54 },
		{ 47, 38, 54 },
		{ 47, 39, 54 },
		{ 47, 40, 54 },
		{ 47, 41, 54 },
		{ 47, 42, 54 },
		{ 47, 43, 54 },
		{ 47, 44, 54 },
		{ 47, 45, 54 },
		{ 47, 46, 54 },
		{ 47, 47, 54 },
		{ 47, 48, 54 },
		{ 47, 49, 54 },
		{ 47, 50, 54 },
		{ 47, 51, 54 },
		{ 47, 52, 54 },
		{ 47, 53, 54 },
		{ 47, 54, 54 },
		{ 47, 55, 54 },
		{ 47, 56, 54 },
		{ 47, 57, 54 },
		{ 47, 58, 54 },
		{ 47, 59, 54 },
		{ 47, 60, 54 },
		{ 47, 61, 54 },
		{ 47, 62, 54 },
		{ 47, 63, 54 },
		{ 47, 64, 54 },
		{ 47, 65, 54 },
		{ 47, 66, 54 },
		{ 47, 67, 54 },
		{ 47, 68, 54 },
		{ 47, 69, 54 },
		{ 47, 70, 54 },
		{ 47, 71, 54 },
		{ 47, 72, 54 },
		{ 47, 73, 54 },
		{ 48, 19, 54 },
		{ 48, 20, 54 },
		{ 48, 21, 54 },
		{ 48, 22, 54 },
		{ 48, 23, 54 },
		{ 48, 24, 54 },
		{ 48, 25, 54 },
		{ 48, 26, 54 },
		{ 48, 27, 54 },
		{ 48, 28, 54 },
		{ 48, 29, 54 },
		{ 48, 30, 54 },
		{ 48, 31, 54 },
		{ 48, 32, 54 },
		{ 48, 33, 54 },
		{ 48, 34, 54 },
		{ 48, 35, 54 },
		{ 48, 36, 54 },
		{ 48, 37, 54 },
		{ 48, 38, 54 },
		{ 48, 39, 54 },
		{ 48, 40, 54 },
		{ 48, 41, 54 },
		{ 48, 42, 54 },
		{ 48, 43, 54 },
		{ 48, 44, 54 },
		{ 48, 45, 54 },
		{ 48, 46, 54 },
		{ 48, 47, 54 },
		{ 48, 48, 54 },
		{ 48, 49, 54 },
		{ 48, 50, 54 },
		{ 48, 51, 54 },
		{ 48, 52, 54 },
		{ 48, 53, 54 },
		{ 48, 54, 54 },
		{ 48, 55, 54 },
		{ 48, 56, 54 },
		{ 48, 57, 54 },
		{ 48, 58, 54 },
		{ 48, 59, 54 },
		{ 48, 60, 54 },
		{ 48, 61, 54 },
		{ 48, 62, 54 },
		{ 48, 63, 54 },
		{ 48, 64, 54 },
		{ 48, 65, 54 },
		{ 48, 66, 54 },
		{ 48, 67, 54 },
		{ 48, 68, 54 },
		{ 48, 69, 54 },
		{ 48, 70, 54 },
		{ 48, 71, 54 },
		{ 48, 72, 54 },
		{ 48, 73, 54 },
		{ 48, 74, 54 },
		{ 49, 20, 54 },
		{ 49, 21, 54 },
		{ 49, 22, 54 },
		{ 49, 23, 54 },
		{ 49, 24, 54 },
		{ 49, 25, 54 },
		{ 49, 26, 54 },
		{ 49, 27, 54 },
		{ 49, 28, 54 },
		{ 49, 29, 54 },
		{ 49, 30, 54 },
		{ 49, 31, 54 },
		{ 49, 32, 54 },
		{ 49, 33, 54 },
		{ 49, 34, 54 },
		{ 49, 35, 54 },
		{ 49, 36, 54 },
		{ 49, 37, 54 },
		{ 49, 38, 54 },
		{ 49, 39, 54 },
		{ 49, 40, 54 },
		{ 49, 41, 54 },
		{ 49, 42, 54 },
		{ 49, 43, 54 },
		{ 49, 44, 54 },
		{ 49, 45, 54 },
		{ 49, 46, 54 },
		{ 49, 47, 54 },
		{ 49, 48, 54 },
		{ 49, 49, 54 },
		{ 49, 50, 54 },
		{ 49, 51, 54 },
		{ 49, 52, 54 },
		{ 49, 53, 54 },
		{ 49, 54, 54 },
		{ 49, 55, 54 },
		{ 49, 56, 54 },
		{ 49, 57, 54 },
		{ 49, 58, 54 },
		{ 49, 59, 54 },
		{ 49, 60, 54 },
		{ 49, 61, 54 },
		{ 49, 62, 54 },
		{ 49, 63, 54 },
		{ 49, 64, 54 },
		{ 49, 65, 54 },
		{ 49, 66, 54 },
		{ 49, 67, 54 },
		{ 49, 68, 54 },
		{ 49, 69, 54 },
		{ 49, 70, 54 },
		{ 49, 71, 54 },
		{ 49, 72, 54 },
		{ 49, 73, 54 },
		{ 49, 74, 54 },
		{ 50, 21, 54 },
		{ 50, 22, 54 },
		{ 50, 23, 54 },
		{ 50, 24, 54 },
		{ 50, 25, 54 },
		{ 50, 26, 54 },
		{ 50, 27, 54 },
		{ 50, 28, 54 },
		{ 50, 29, 54 },
		{ 50, 30, 54 },
		{ 50, 31, 54 },
		{ 50, 32, 54 },
		{ 50, 33, 54 },
		{ 50, 34, 54 },
		{ 50, 35, 54 },
		{ 50, 36, 54 },
		{ 50, 37, 54 },
		{ 50, 38, 54 },
		{ 50, 39, 54 },
		{ 50, 40, 54 },
		{ 50, 41, 54 },
		{ 50, 42, 54 },
		{ 50, 43, 54 },
		{ 50, 44, 54 },
		{ 50, 45, 54 },
		{ 50, 46, 54 },
		{ 50, 47, 54 },
		{ 50, 48, 54 },
		{ 50, 49, 54 },
		{ 50, 50, 54 },
		{ 50, 51, 54 },
		{ 50, 52, 54 },
		{ 50, 53, 54 },
		{ 50, 54, 54 },
		{ 50, 55, 54 },
		{ 50, 56, 54 },
		{ 50, 57, 54 },
		{ 50, 58, 54 },
		{ 50, 59, 54 },
		{ 50, 60, 54 },
		{ 50, 61, 54 },
		{ 50, 62, 54 },
		{ 50, 63, 54 },
		{ 50, 64, 54 },
		{ 50, 65, 54 },
		{ 50, 66, 54 },
		{ 50, 67, 54 },
		{ 50, 68, 54 },
		{ 50, 69, 54 },
		{ 50, 70, 54 },
		{ 50, 71, 54 },
		{ 50, 72, 54 },
		{ 50, 73, 54 },
		{ 50, 74, 54 },
		{ 50, 75, 54 },
		{ 51, 22, 54 },
		{ 51, 23, 54 },
		{ 51, 24, 54 },
		{ 51, 25, 54 },
		{ 51, 26, 54 },
		{ 51, 27, 54 },
		{ 51, 28, 54 },
		{ 51, 29, 54 },
		{ 51, 30, 54 },
		{ 51, 31, 54 },
		{ 51, 32, 54 },
		{ 51, 33, 54 },
		{ 51, 34, 54 },
		{ 51, 35, 54 },
		{ 51, 36, 54 },
		{ 51, 37, 54 },
		{ 51, 38, 54 },
		{ 51, 39, 54 },
		{ 51, 40, 54 },
		{ 51, 41, 54 },
		{ 51, 42, 54 },
		{ 51, 43, 54 },
		{ 51, 44, 54 },
		{ 51, 45, 54 },
		{ 51, 46, 54 },
		{ 51, 47, 54 },
		{ 51, 48, 54 },
		{ 51, 49, 54 },
		{ 51, 50, 54 },
		{ 51, 51, 54 },
		{ 51, 52, 54 },
		{ 51, 53, 54 },
		{ 51, 54, 54 },
		{ 51, 55, 54 },
		{ 51, 56, 54 },
		{ 51, 57, 54 },
		{ 51, 58, 54 },
		{ 51, 59, 54 },
		{ 51, 60, 54 },
		{ 51, 61, 54 },
		{ 51, 62, 54 },
		{ 51, 63, 54 },
		{ 51, 64, 54 },
		{ 51, 65, 54 },
		{ 51, 66, 54 },
		{ 51, 67, 54 },
		{ 51, 68, 54 },
		{ 51, 69, 54 },
		{ 51, 70, 54 },
		{ 51, 71, 54 },
		{ 51, 72, 54 },
		{ 51, 73, 54 },
		{ 51, 74, 54 },
		{ 51, 75, 54 },
		{ 51, 76, 54 },
		{ 52, 23, 54 },
		{ 52, 24, 54 },
		{ 52, 25, 54 },
		{ 52, 26, 54 },
		{ 52, 27, 54 },
		{ 52, 28, 54 },
		{ 52, 29, 54 },
		{ 52, 30, 54 },
		{ 52, 31, 54 },
		{ 52, 32, 54 },
		{ 52, 33, 54 },
		{ 52, 34, 54 },
		{ 52, 35, 54 },
		{ 52, 36, 54 },
		{ 52, 37, 54 },
		{ 52, 38, 54 },
		{ 52, 39, 54 },
		{ 52, 40, 54 },
		{ 52, 41, 54 },
		{ 52, 42, 54 },
		{ 52, 43, 54 },
		{ 52, 44, 54 },
		{ 52, 45, 54 },
		{ 52, 46, 54 },
		{ 52, 47, 54 },
		{ 52, 48, 54 },
		{ 52, 49, 54 },
		{ 52, 50, 54 },
		{ 52, 51, 54 },
		{ 52, 52, 54 },
		{ 52, 53, 54 },
		{ 52, 54, 54 },
		{ 52, 55, 54 },
		{ 52, 56, 54 },
		{ 52, 57, 54 },
		{ 52, 58, 54 },
		{ 52, 59, 54 },
		{ 52, 60, 54 },
		{ 52, 61, 54 },
		{ 52, 62, 54 },
		{ 52, 63, 54 },
		{ 52, 64, 54 },
		{ 52, 65, 54 },
		{ 52, 66, 54 },
		{ 52, 67, 54 },
		{ 52, 68, 54 },
		{ 52, 69, 54 },
		{ 52, 70, 54 },
		{ 52, 71, 54 },
		{ 52, 72, 54 },
		{ 52, 73, 54 },
		{ 52, 74, 54 },
		{ 52, 75, 54 },
		{ 52, 76, 54 },
		{ 52, 77, 54 },
		{ 53, 24, 54 },
		{ 53, 25, 54 },
		{ 53, 26, 54 },
		{ 53, 27, 54 },
		{ 53, 28, 54 },
		{ 53, 29, 54 },
		{ 53, 30, 54 },
		{ 53, 31, 54 },
		{ 53, 32, 54 },
		{ 53, 33, 54 },
		{ 53, 34, 54 },
		{ 53, 35, 54 },
		{ 53, 36, 54 },
		{ 53, 37, 54 },
		{ 53, 38, 54 },
		{ 53, 39, 54 },
		{ 53, 40, 54 },
		{ 53, 41, 54 },
		{ 53, 42, 54 },
		{ 53, 43, 54 },
		{ 53, 44, 54 },
		{ 53, 45, 54 },
		{ 53, 46, 54 },
		{ 53, 47, 54 },
		{ 53, 48, 54 },
		{ 53, 49, 54 },
		{ 53, 50, 54 },
		{ 53, 51, 54 },
		{ 53, 52, 54 },
		{ 53, 53, 54 },
		{ 53, 54, 54 },
		{ 53, 55, 54 },
		{ 53, 56, 54 },
		{ 53, 57, 54 },
		{ 53, 58, 54 },
		{ 53, 59, 54 },
		{ 53, 60, 54 },
		{ 53, 61, 54 },
		{ 53, 62, 54 },
		{ 53, 63, 54 },
		{ 53, 64, 54 },
		{ 53, 65, 54 },
		{ 53, 66, 54 },
		{ 53, 67, 54 },
		{ 53, 68, 54 },
		{ 53, 69, 54 },
		{ 53, 70, 54 },
		{ 53, 71, 54 },
		{ 53, 72, 54 },
		{ 53, 73, 54 },
		{ 53, 74, 54 },
		{ 53, 75, 54 },
		{ 53, 76, 54 },
		{ 53, 77, 54 },
		{ 53, 78, 54 },
		{ 54, 26, 54 },
		{ 54, 27, 54 },
		{ 54, 28, 54 },
		{ 54, 29, 54 },
		{ 54, 30, 54 },
		{ 54, 31, 54 },
		{ 54, 32, 54 },
		{ 54, 33, 54 },
		{ 54, 34, 54 },
		{ 54, 35, 54 },
		{ 54, 36, 54 },
		{ 54, 37, 54 },
		{ 54, 38, 54 },
		{ 54, 39, 54 },
		{ 54, 40, 54 },
		{ 54, 41, 54 },
		{ 54, 42, 54 },
		{ 54, 43, 54 },
		{ 54, 44, 54 },
		{ 54, 45, 54 },
		{ 54, 46, 54 },
		{ 54, 47, 54 },
		{ 54, 48, 54 },
		{ 54, 49, 54 },
		{ 54, 50, 54 },
		{ 54, 51, 54 },
		{ 54, 52, 54 },
		{ 54, 53, 54 },
		{ 54, 54, 54 },
		{ 54, 55, 54 },
		{ 54, 56, 54 },
		{ 54, 57, 54 },
		{ 54, 58, 54 },
		{ 54, 59, 54 },
		{ 54, 60, 54 },
		{ 54, 61, 54 },
		{ 54, 62, 54 },
		{ 54, 63, 54 },
		{ 54, 64, 54 },
		{ 54, 65, 54 },
		{ 54, 66, 54 },
		{ 54, 67, 54 },
		{ 54, 68, 54 },
		{ 54, 69, 54 },
		{ 54, 70, 54 },
		{ 54, 71, 54 },
		{ 54, 72, 54 },
		{ 54, 73, 54 },
		{ 54, 74, 54 },
		{ 54, 75, 54 },
		{ 54, 76, 54 },
		{ 54, 77, 54 },
		{ 54, 78, 54 },
		{ 54, 79, 54 },
		{ 55, 27, 54 },
		{ 55, 28, 54 },
		{ 55, 29, 54 },
		{ 55, 30, 54 },
		{ 55, 31, 54 },
		{ 55, 32, 54 },
		{ 55, 33, 54 },
		{ 55, 34, 54 },
		{ 55, 35, 54 },
		{ 55, 36, 54 },
		{ 55, 37, 54 },
		{ 55, 38, 54 },
		{ 55, 39, 54 },
		{ 55, 40, 54 },
		{ 55, 41, 54 },
		{ 55, 42, 54 },
		{ 55, 43, 54 },
		{ 55, 44, 54 },
		{ 55, 45, 54 },
		{ 55, 46, 54 },
		{ 55, 47, 54 },
		{ 55, 48, 54 },
		{ 55, 49, 54 },
		{ 55, 50, 54 },
		{ 55, 51, 54 },
		{ 55, 52, 54 },
		{ 55, 53, 54 },
		{ 55, 54, 54 },
		{ 55, 55, 54 },
		{ 55, 56, 54 },
		{ 55, 57, 54 },
		{ 55, 58, 54 },
		{ 55, 59, 54 },
		{ 55, 60, 54 },
		{ 55, 61, 54 },
		{ 55, 62, 54 },
		{ 55, 63, 54 },
		{ 55, 64, 54 },
		{ 55, 65, 54 },
		{ 55, 66, 54 },
		{ 55, 67, 54 },
		{ 55, 68, 54 },
		{ 55, 69, 54 },
		{ 55, 70, 54 },
		{ 55, 71, 54 },
		{ 55, 72, 54 },
		{ 55, 73, 54 },
		{ 55, 74, 54 },
		{ 55, 75, 54 },
		{ 55, 76, 54 },
		{ 55, 77, 54 },
		{ 55, 78, 54 },
		{ 55, 79, 54 },
		{ 55, 80, 54 },
		{ 56, 28, 54 },
		{ 56, 29, 54 },
		{ 56, 30, 54 },
		{ 56, 31, 54 },
		{ 56, 32, 54 },
		{ 56, 33, 54 },
		{ 56, 34, 54 },
		{ 56, 35, 54 },
		{ 56, 36, 54 },
		{ 56, 37, 54 },
		{ 56, 38, 54 },
		{ 56, 39, 54 },
		{ 56, 40, 54 },
		{ 56, 41, 54 },
		{ 56, 42, 54 },
		{ 56, 43, 54 },
		{ 56, 44, 54 },
		{ 56, 45, 54 },
		{ 56, 46, 54 },
		{ 56, 47, 54 },
		{ 56, 48, 54 },
		{ 56, 49, 54 },
		{ 56, 50, 54 },
		{ 56, 51, 54 },
		{ 56, 52, 54 },
		{ 56, 53, 54 },
		{ 56, 54, 54 },
		{ 56, 55, 54 },
		{ 56, 56, 54 },
		{ 56, 57, 54 },
		{ 56, 58, 54 },
		{ 56, 59, 54 },
		{ 56, 60, 54 },
		{ 56, 61, 54 },
		{ 56, 62, 54 },
		{ 56, 63, 54 },
		{ 56, 64, 54 },
		{ 56, 65, 54 },
		{ 56, 66, 54 },
		{ 56, 67, 54 },
		{ 56, 68, 54 },
		{ 56, 69, 54 },
		{ 56, 70, 54 },
		{ 56, 71, 54 },
		{ 56, 72, 54 },
		{ 56, 73, 54 },
		{ 56, 74, 54 },
		{ 56, 75, 54 },
		{ 56, 76, 54 },
		{ 56, 77, 54 },
		{ 56, 78, 54 },
		{ 56, 79, 54 },
		{ 56, 80, 54 },
		{ 56, 81, 54 },
		{ 57, 29, 54 },
		{ 57, 30, 54 },
		{ 57, 31, 54 },
		{ 57, 32, 54 },
		{ 57, 33, 54 },
		{ 57, 34, 54 },
		{ 57, 35, 54 },
		{ 57, 36, 54 },
		{ 57, 37, 54 },
		{ 57, 38, 54 },
		{ 57, 39, 54 },
		{ 57, 40, 54 },
		{ 57, 41, 54 },
		{ 57, 42, 54 },
		{ 57, 43, 54 },
		{ 57, 44, 54 },
		{ 57, 45, 54 },
		{ 57, 46, 54 },
		{ 57, 47, 54 },
		{ 57, 48, 54 },
		{ 57, 49, 54 },
		{ 57, 50, 54 },
		{ 57, 51, 54 },
		{ 57, 52, 54 },
		{ 57, 53, 54 },
		{ 57, 54, 54 },
		{ 57, 55, 54 },
		{ 57, 56, 54 },
		{ 57, 57, 54 },
		{ 57, 58, 54 },
		{ 57, 59, 54 },
		{ 57, 60, 54 },
		{ 57, 61, 54 },
		{ 57, 62, 54 },
		{ 57, 63, 54 },
		{ 57, 64, 54 },
		{ 57, 65, 54 },
		{ 57, 66, 54 },
		{ 57, 67, 54 },
		{ 57, 68, 54 },
		{ 57, 69, 54 },
		{ 57, 70, 54 },
		{ 57, 71, 54 },
		{ 57, 72, 54 },
		{ 57, 73, 54 },
		{ 57, 74, 54 },
		{ 57, 75, 54 },
		{ 57, 76, 54 },
		{ 57, 77, 54 },
		{ 57, 78, 54 },
		{ 57, 79, 54 },
		{ 57, 80, 54 },
		{ 57, 81, 54 },
		{ 58, 30, 54 },
		{ 58, 31, 54 },
		{ 58, 32, 54 },
		{ 58, 33, 54 },
		{ 58, 34, 54 },
		{ 58, 35, 54 },
		{ 58, 36, 54 },
		{ 58, 37, 54 },
		{ 58, 38, 54 },
		{ 58, 39, 54 },
		{ 58, 40, 54 },
		{ 58, 41, 54 },
		{ 58, 42, 54 },
		{ 58, 43, 54 },
		{ 58, 44, 54 },
		{ 58, 45, 54 },
		{ 58, 46, 54 },
		{ 58, 47, 54 },
		{ 58, 48, 54 },
		{ 58, 49, 54 },
		{ 58, 50, 54 },
		{ 58, 51, 54 },
		{ 58, 52, 54 },
		{ 58, 53, 54 },
		{ 58, 54, 54 },
		{ 58, 55, 54 },
		{ 58, 56, 54 },
		{ 58, 57, 54 },
		{ 58, 58, 54 },
		{ 58, 59, 54 },
		{ 58, 60, 54 },
		{ 58, 61, 54 },
		{ 58, 62, 54 },
		{ 58, 63, 54 },
		{ 58, 64, 54 },
		{ 58, 65, 54 },
		{ 58, 66, 54 },
		{ 58, 67, 54 },
		{ 58, 68, 54 },
		{ 58, 69, 54 },
		{ 58, 70, 54 },
		{ 58, 71, 54 },
		{ 58, 72, 54 },
		{ 58, 73, 54 },
		{ 58, 74, 54 },
		{ 58, 75, 54 },
		{ 58, 76, 54 },
		{ 58, 77, 54 },
		{ 58, 78, 54 },
		{ 58, 79, 54 },
		{ 58, 80, 54 },
		{ 58, 81, 54 },
		{ 58, 82, 54 },
		{ 59, 31, 54 },
		{ 59, 32, 54 },
		{ 59, 33, 54 },
		{ 59, 34, 54 },
		{ 59, 35, 54 },
		{ 59, 36, 54 },
		{ 59, 37, 54 },
		{ 59, 38, 54 },
		{ 59, 39, 54 },
		{ 59, 40, 54 },
		{ 59, 41, 54 },
		{ 59, 42, 54 },
		{ 59, 43, 54 },
		{ 59, 44, 54 },
		{ 59, 45, 54 },
		{ 59, 46, 54 },
		{ 59, 47, 54 },
		{ 59, 48, 54 },
		{ 59, 49, 54 },
		{ 59, 50, 54 },
		{ 59, 51, 54 },
		{ 59, 52, 54 },
		{ 59, 53, 54 },
		{ 59, 54, 54 },
		{ 59, 55, 54 },
		{ 59, 56, 54 },
		{ 59, 57, 54 },
		{ 59, 58, 54 },
		{ 59, 59, 54 },
		{ 59, 60, 54 },
		{ 59, 61, 54 },
		{ 59, 62, 54 },
		{ 59, 63, 54 },
		{ 59, 64, 54 },
		{ 59, 65, 54 },
		{ 59, 66, 54 },
		{ 59, 67, 54 },
		{ 59, 68, 54 },
		{ 59, 69, 54 },
		{ 59, 70, 54 },
		{ 59, 71, 54 },
		{ 59, 72, 54 },
		{ 59, 73, 54 },
		{ 59, 74, 54 },
		{ 59, 75, 54 },
		{ 59, 76, 54 },
		{ 59, 77, 54 },
		{ 59, 78, 54 },
		{ 59, 79, 54 },
		{ 59, 80, 54 },
		{ 59, 81, 54 },
		{ 59, 82, 54 },
		{ 59, 83, 54 },
		{ 60, 32, 54 },
		{ 60, 33, 54 },
		{ 60, 34, 54 },
		{ 60, 35, 54 },
		{ 60, 36, 54 },
		{ 60, 37, 54 },
		{ 60, 38, 54 },
		{ 60, 39, 54 },
		{ 60, 40, 54 },
		{ 60, 41, 54 },
		{ 60, 42, 54 },
		{ 60, 43, 54 },
		{ 60, 44, 54 },
		{ 60, 45, 54 },
		{ 60, 46, 54 },
		{ 60, 47, 54 },
		{ 60, 48, 54 },
		{ 60, 49, 54 },
		{ 60, 50, 54 },
		{ 60, 51, 54 },
		{ 60, 52, 54 },
		{ 60, 53, 54 },
		{ 60, 54, 54 },
		{ 60, 55, 54 },
		{ 60, 56, 54 },
		{ 60, 57, 54 },
		{ 60, 58, 54 },
		{ 60, 59, 54 },
		{ 60, 60, 54 },
		{ 60, 61, 54 },
		{ 60, 62, 54 },
		{ 60, 63, 54 },
		{ 60, 64, 54 },
		{ 60, 65, 54 },
		{ 60, 66, 54 },
		{ 60, 67, 54 },
		{ 60, 68, 54 },
		{ 60, 69, 54 },
		{ 60, 70, 54 },
		{ 60, 71, 54 },
		{ 60, 72, 54 },
		{ 60, 73, 54 },
		{ 60, 74, 54 },
		{ 60, 75, 54 },
		{ 60, 76, 54 },
		{ 60, 77, 54 },
		{ 60, 78, 54 },
		{ 60, 79, 54 },
		{ 60, 80, 54 },
		{ 60, 81, 54 },
		{ 60, 82, 54 },
		{ 60, 83, 54 },
		{ 60, 84, 54 },
		{ 61, 34, 54 },
		{ 61, 35, 54 },
		{ 61, 36, 54 },
		{ 61, 37, 54 },
		{ 61, 38, 54 },
		{ 61, 39, 54 },
		{ 61, 40, 54 },
		{ 61, 41, 54 },
		{ 61, 42, 54 },
		{ 61, 43, 54 },
		{ 61, 44, 54 },
		{ 61, 45, 54 },
		{ 61, 46, 54 },
		{ 61, 47, 54 },
		{ 61, 48, 54 },
		{ 61, 49, 54 },
		{ 61, 50, 54 },
		{ 61, 51, 54 },
		{ 61, 52, 54 },
		{ 61, 53, 54 },
		{ 61, 54, 54 },
		{ 61, 55, 54 },
		{ 61, 56, 54 },
		{ 61, 57, 54 },
		{ 61, 58, 54 },
		{ 61, 59, 54 },
		{ 61, 60, 54 },
		{ 61, 61, 54 },
		{ 61, 62, 54 },
		{ 61, 63, 54 },
		{ 61, 64, 54 },
		{ 61, 65, 54 },
		{ 61, 66, 54 },
		{ 61, 67, 54 },
		{ 61, 68, 54 },
		{ 61, 69, 54 },
		{ 61, 70, 54 },
		{ 61, 71, 54 },
		{ 61, 72, 54 },
		{ 61, 73, 54 },
		{ 61, 74, 54 },
		{ 61, 75, 54 },
		{ 61, 76, 54 },
		{ 61, 77, 54 },
		{ 61, 78, 54 },
		{ 61, 79, 54 },
		{ 61, 80, 54 },
		{ 61, 81, 54 },
		{ 61, 82, 54 },
		{ 61, 83, 54 },
		{ 61, 84, 54 },
		{ 61, 85, 54 },
		{ 62, 35, 54 },
		{ 62, 36, 54 },
		{ 62, 37, 54 },
		{ 62, 38, 54 },
		{ 62, 39, 54 },
		{ 62, 40, 54 },
		{ 62, 41, 54 },
		{ 62, 42, 54 },
		{ 62, 43, 54 },
		{ 62, 44, 54 },
		{ 62, 45, 54 },
		{ 62, 46, 54 },
		{ 62, 47, 54 },
		{ 62, 48, 54 },
		{ 62, 49, 54 },
		{ 62, 50, 54 },
		{ 62, 51, 54 },
		{ 62, 52, 54 },
		{ 62, 53, 54 },
		{ 62, 54, 54 },
		{ 62, 55, 54 },
		{ 62, 56, 54 },
		{ 62, 57, 54 },
		{ 62, 58, 54 },
		{ 62, 59, 54 },
		{ 62, 60, 54 },
		{ 62, 61, 54 },
		{ 62, 62, 54 },
		{ 62, 63, 54 },
		{ 62, 64, 54 },
		{ 62, 65, 54 },
		{ 62, 66, 54 },
		{ 62, 67, 54 },
		{ 62, 68, 54 },
		{ 62, 69, 54 },
		{ 62, 70, 54 },
		{ 62, 71, 54 },
		{ 62, 72, 54 },
		{ 62, 73, 54 },
		{ 62, 74, 54 },
		{ 62, 75, 54 },
		{ 62, 76, 54 },
		{ 62, 77, 54 },
		{ 62, 78, 54 },
		{ 62, 79, 54 },
		{ 62, 80, 54 },
		{ 62, 81, 54 },
		{ 62, 82, 54 },
		{ 62, 83, 54 },
		{ 62, 84, 54 },
		{ 62, 85, 54 },
		{ 62, 86, 54 },
		{ 63, 36, 54 },
		{ 63, 37, 54 },
		{ 63, 38, 54 },
		{ 63, 39, 54 },
		{ 63, 40, 54 },
		{ 63, 41, 54 },
		{ 63, 42, 54 },
		{ 63, 43, 54 },
		{ 63, 44, 54 },
		{ 63, 45, 54 },
		{ 63, 46, 54 },
		{ 63, 47, 54 },
		{ 63, 48, 54 },
		{ 63, 49, 54 },
		{ 63, 50, 54 },
		{ 63, 51, 54 },
		{ 63, 52, 54 },
		{ 63, 53, 54 },
		{ 63, 54, 54 },
		{ 63, 55, 54 },
		{ 63, 56, 54 },
		{ 63, 57, 54 },
		{ 63, 58, 54 },
		{ 63, 59, 54 },
		{ 63, 60, 54 },
		{ 63, 61, 54 },
		{ 63, 62, 54 },
		{ 63, 63, 54 },
		{ 63, 64, 54 },
		{ 63, 65, 54 },
		{ 63, 66, 54 },
		{ 63, 67, 54 },
		{ 63, 68, 54 },
		{ 63, 69, 54 },
		{ 63, 70, 54 },
		{ 63, 71, 54 },
		{ 63, 72, 54 },
		{ 63, 73, 54 },
		{ 63, 74, 54 },
		{ 63, 75, 54 },
		{ 63, 76, 54 },
		{ 63, 77, 54 },
		{ 63, 78, 54 },
		{ 63, 79, 54 },
		{ 63, 80, 54 },
		{ 63, 81, 54 },
		{ 63, 82, 54 },
		{ 63, 83, 54 },
		{ 63, 84, 54 },
		{ 63, 85, 54 },
		{ 63, 86, 54 },
		{ 63, 87, 54 },
		{ 64, 37, 54 },
		{ 64, 38, 54 },
		{ 64, 39, 54 },
		{ 64, 40, 54 },
		{ 64, 41, 54 },
		{ 64, 42, 54 },
		{ 64, 43, 54 },
		{ 64, 44, 54 },
		{ 64, 45, 54 },
		{ 64, 46, 54 },
		{ 64, 47, 54 },
		{ 64, 48, 54 },
		{ 64, 49, 54 },
		{ 64, 50, 54 },
		{ 64, 51, 54 },
		{ 64, 52, 54 },
		{ 64, 53, 54 },
		{ 64, 54, 54 },
		{ 64, 55, 54 },
		{ 64, 56, 54 },
		{ 64, 57, 54 },
		{ 64, 58, 54 },
		{ 64, 59, 54 },
		{ 64, 60, 54 },
		{ 64, 61, 54 },
		{ 64, 62, 54 },
		{ 64, 63, 54 },
		{ 64, 64, 54 },
		{ 64, 65, 54 },
		{ 64, 66, 54 },
		{ 64, 67, 54 },
		{ 64, 68, 54 },
		{ 64, 69, 54 },
		{ 64, 70, 54 },
		{ 64, 71, 54 },
		{ 64, 72, 54 },
		{ 64, 73, 54 },
		{ 64, 74, 54 },
		{ 64, 75, 54 },
		{ 64, 76, 54 },
		{ 64, 77, 54 },
		{ 64, 78, 54 },
		{ 64, 79, 54 },
		{ 64, 80, 54 },
		{ 64, 81, 54 },
		{ 64, 82, 54 },
		{ 64, 83, 54 },
		{ 64, 84, 54 },
		{ 64, 85, 54 },
		{ 64, 86, 54 },
		{ 64, 87, 54 },
		{ 64, 88, 54 },
		{ 65, 38, 54 },
		{ 65, 39, 54 },
		{ 65, 40, 54 },
		{ 65, 41, 54 },
		{ 65, 42, 54 },
		{ 65, 43, 54 },
		{ 65, 44, 54 },
		{ 65, 45, 54 },
		{ 65, 46, 54 },
		{ 65, 47, 54 },
		{ 65, 48, 54 },
		{ 65, 49, 54 },
		{ 65, 50, 54 },
		{ 65, 51, 54 },
		{ 65, 52, 54 },
		{ 65, 53, 54 },
		{ 65, 54, 54 },
		{ 65, 55, 54 },
		{ 65, 56, 54 },
		{ 65, 57, 54 },
		{ 65, 58, 54 },
		{ 65, 59, 54 },
		{ 65, 60, 54 },
		{ 65, 61, 54 },
		{ 65, 62, 54 },
		{ 65, 63, 54 },
		{ 65, 64, 54 },
		{ 65, 65, 54 },
		{ 65, 66, 54 },
		{ 65, 67, 54 },
		{ 65, 68, 54 },
		{ 65, 69, 54 },
		{ 65, 70, 54 },
		{ 65, 71, 54 },
		{ 65, 72, 54 },
		{ 65, 73, 54 },
		{ 65, 74, 54 },
		{ 65, 75, 54 },
		{ 65, 76, 54 },
		{ 65, 77, 54 },
		{ 65, 78, 54 },
		{ 65, 79, 54 },
		{ 65, 80, 54 },
		{ 65, 81, 54 },
		{ 65, 82, 54 },
		{ 65, 83, 54 },
		{ 65, 84, 54 },
		{ 65, 85, 54 },
		{ 65, 86, 54 },
		{ 65, 87, 54 },
		{ 65, 88, 54 },
		{ 66, 39, 54 },
		{ 66, 40, 54 },
		{ 66, 41, 54 },
		{ 66, 42, 54 },
		{ 66, 43, 54 },
		{ 66, 44, 54 },
		{ 66, 45, 54 },
		{ 66, 46, 54 },
		{ 66, 47, 54 },
		{ 66, 48, 54 },
		{ 66, 49, 54 },
		{ 66, 50, 54 },
		{ 66, 51, 54 },
		{ 66, 52, 54 },
		{ 66, 53, 54 },
		{ 66, 54, 54 },
		{ 66, 55, 54 },
		{ 66, 56, 54 },
		{ 66, 57, 54 },
		{ 66, 58, 54 },
		{ 66, 59, 54 },
		{ 66, 60, 54 },
		{ 66, 61, 54 },
		{ 66, 62, 54 },
		{ 66, 63, 54 },
		{ 66, 64, 54 },
		{ 66, 65, 54 },
		{ 66, 66, 54 },
		{ 66, 67, 54 },
		{ 66, 68, 54 },
		{ 66, 69, 54 },
		{ 66, 70, 54 },
		{ 66, 71, 54 },
		{ 66, 72, 54 },
		{ 66, 73, 54 },
		{ 66, 74, 54 },
		{ 66, 75, 54 },
		{ 66, 76, 54 },
		{ 66, 77, 54 },
		{ 66, 78, 54 },
		{ 66, 79, 54 },
		{ 66, 80, 54 },
		{ 66, 81, 54 },
		{ 66, 82, 54 },
		{ 66, 83, 54 },
		{ 66, 84, 54 },
		{ 66, 85, 54 },
		{ 66, 86, 54 },
		{ 66, 87, 54 },
		{ 66, 88, 54 },
		{ 66, 89, 54 },
		{ 67, 40, 54 },
		{ 67, 41, 54 },
		{ 67, 42, 54 },
		{ 67, 43, 54 },
		{ 67, 44, 54 },
		{ 67, 45, 54 },
		{ 67, 46, 54 },
		{ 67, 47, 54 },
		{ 67, 48, 54 },
		{ 67, 49, 54 },
		{ 67, 50, 54 },
		{ 67, 51, 54 },
		{ 67, 52, 54 },
		{ 67, 53, 54 },
		{ 67, 54, 54 },
		{ 67, 55, 54 },
		{ 67, 56, 54 },
		{ 67, 57, 54 },
		{ 67, 58, 54 },
		{ 67, 59, 54 },
		{ 67, 60, 54 },
		{ 67, 61, 54 },
		{ 67, 62, 54 },
		{ 67, 63, 54 },
		{ 67, 64, 54 },
		{ 67, 65, 54 },
		{ 67, 66, 54 },
		{ 67, 67, 54 },
		{ 67, 68, 54 },
		{ 67, 69, 54 },
		{ 67, 70, 54 },
		{ 67, 71, 54 },
		{ 67, 72, 54 },
		{ 67, 73, 54 },
		{ 67, 74, 54 },
		{ 67, 75, 54 },
		{ 67, 76, 54 },
		{ 67, 77, 54 },
		{ 67, 78, 54 },
		{ 67, 79, 54 },
		{ 67, 80, 54 },
		{ 67, 81, 54 },
		{ 67, 82, 54 },
		{ 67, 83, 54 },
		{ 67, 84, 54 },
		{ 67, 85, 54 },
		{ 67, 86, 54 },
		{ 67, 87, 54 },
		{ 67, 88, 54 },
		{ 67, 89, 54 },
		{ 67, 90, 54 },
		{ 68, 42, 54 },
		{ 68, 43, 54 },
		{ 68, 44, 54 },
		{ 68, 45, 54 },
		{ 68, 46, 54 },
		{ 68, 47, 54 },
		{ 68, 48, 54 },
		{ 68, 49, 54 },
		{ 68, 50, 54 },
		{ 68, 51, 54 },
		{ 68, 52, 54 },
		{ 68, 53, 54 },
		{ 68, 54, 54 },
		{ 68, 55, 54 },
		{ 68, 56, 54 },
		{ 68, 57, 54 },
		{ 68, 58, 54 },
		{ 68, 59, 54 },
		{ 68, 60, 54 },
		{ 68, 61, 54 },
		{ 68, 62, 54 },
		{ 68, 63, 54 },
		{ 68, 64, 54 },
		{ 68, 65, 54 },
		{ 68, 66, 54 },
		{ 68, 67, 54 },
		{ 68, 68, 54 },
		{ 68, 69, 54 },
		{ 68, 70, 54 },
		{ 68, 71, 54 },
		{ 68, 72, 54 },
		{ 68, 73, 54 },
		{ 68, 74, 54 },
		{ 68, 75, 54 },
		{ 68, 76, 54 },
		{ 68, 77, 54 },
		{ 68, 78, 54 },
		{ 68, 79, 54 },
		{ 68, 80, 54 },
		{ 68, 81, 54 },
		{ 68, 82, 54 },
		{ 68, 83, 54 },
		{ 68, 84, 54 },
		{ 68, 85, 54 },
		{ 68, 86, 54 },
		{ 68, 87, 54 },
		{ 68, 88, 54 },
		{ 68, 89, 54 },
		{ 68, 90, 54 },
		{ 68, 91, 54 },
		{ 69, 43, 54 },
		{ 69, 44, 54 },
		{ 69, 45, 54 },
		{ 69, 46, 54 },
		{ 69, 47, 54 },
		{ 69, 48, 54 },
		{ 69, 49, 54 },
		{ 69, 50, 54 },
		{ 69, 51, 54 },
		{ 69, 52, 54 },
		{ 69, 53, 54 },
		{ 69, 54, 54 },
		{ 69, 55, 54 },
		{ 69, 56, 54 },
		{ 69, 57, 54 },
		{ 69, 58, 54 },
		{ 69, 59, 54 },
		{ 69, 60, 54 },
		{ 69, 61, 54 },
		{ 69, 62, 54 },
		{ 69, 63, 54 },
		{ 69, 64, 54 },
		{ 69, 65, 54 },
		{ 69, 66, 54 },
		{ 69, 67, 54 },
		{ 69, 68, 54 },
		{ 69, 69, 54 },
		{ 69, 70, 54 },
		{ 69, 71, 54 },
		{ 69, 72, 54 },
		{ 69, 73, 54 },
		{ 69, 74, 54 },
		{ 69, 75, 54 },
		{ 69, 76, 54 },
		{ 69, 77, 54 },
		{ 69, 78, 54 },
		{ 69, 79, 54 },
		{ 69, 80, 54 },
		{ 69, 81, 54 },
		{ 69, 82, 54 },
		{ 69, 83, 54 },
		{ 69, 84, 54 },
		{ 69, 85, 54 },
		{ 69, 86, 54 },
		{ 69, 87, 54 },
		{ 69, 88, 54 },
		{ 69, 89, 54 },
		{ 69, 90, 54 },
		{ 69, 91, 54 },
		{ 69, 92, 54 },
		{ 70, 44, 54 },
		{ 70, 45, 54 },
		{ 70, 46, 54 },
		{ 70, 47, 54 },
		{ 70, 48, 54 },
		{ 70, 49, 54 },
		{ 70, 50, 54 },
		{ 70, 51, 54 },
		{ 70, 52, 54 },
		{ 70, 53, 54 },
		{ 70, 54, 54 },
		{ 70, 55, 54 },
		{ 70, 56, 54 },
		{ 70, 57, 54 },
		{ 70, 58, 54 },
		{ 70, 59, 54 },
		{ 70, 60, 54 },
		{ 70, 61, 54 },
		{ 70, 62, 54 },
		{ 70, 63, 54 },
		{ 70, 64, 54 },
		{ 70, 65, 54 },
		{ 70, 66, 54 },
		{ 70, 67, 54 },
		{ 70, 68, 54 },
		{ 70, 69, 54 },
		{ 70, 70, 54 },
		{ 70, 71, 54 },
		{ 70, 72, 54 },
		{ 70, 73, 54 },
		{ 70, 74, 54 },
		{ 70, 75, 54 },
		{ 70, 76, 54 },
		{ 70, 77, 54 },
		{ 70, 78, 54 },
		{ 70, 79, 54 },
		{ 70, 80, 54 },
		{ 70, 81, 54 },
		{ 70, 82, 54 },
		{ 70, 83, 54 },
		{ 70, 84, 54 },
		{ 70, 85, 54 },
		{ 70, 86, 54 },
		{ 70, 87, 54 },
		{ 70, 88, 54 },
		{ 70, 89, 54 },
		{ 70, 90, 54 },
		{ 70, 91, 54 },
		{ 70, 92, 54 },
		{ 70, 93, 54 },
		{ 71, 45, 54 },
		{ 71, 46, 54 },
		{ 71, 47, 54 },
		{ 71, 48, 54 },
		{ 71, 49, 54 },
		{ 71, 50, 54 },
		{ 71, 51, 54 },
		{ 71, 52, 54 },
		{ 71, 53, 54 },
		{ 71, 54, 54 },
		{ 71, 55, 54 },
		{ 71, 56, 54 },
		{ 71, 57, 54 },
		{ 71, 58, 54 },
		{ 71, 59, 54 },
		{ 71, 60, 54 },
		{ 71, 61, 54 },
		{ 71, 62, 54 },
		{ 71, 63, 54 },
		{ 71, 64, 54 },
		{ 71, 65, 54 },
		{ 71, 66, 54 },
		{ 71, 67, 54 },
		{ 71, 68, 54 },
		{ 71, 69, 54 },
		{ 71, 70, 54 },
		{ 71, 71, 54 },
		{ 71, 72, 54 },
		{ 71, 73, 54 },
		{ 71, 74, 54 },
		{ 71, 75, 54 },
		{ 71, 76, 54 },
		{ 71, 77, 54 },
		{ 71, 78, 54 },
		{ 71, 79, 54 },
		{ 71, 80, 54 },
		{ 71, 81, 54 },
		{ 71, 82, 54 },
		{ 71, 83, 54 },
		{ 71, 84, 54 },
		{ 71, 85, 54 },
		{ 71, 86, 54 },
		{ 71, 87, 54 },
		{ 71, 88, 54 },
		{ 71, 89, 54 },
		{ 71, 90, 54 },
		{ 71, 91, 54 },
		{ 71, 92, 54 },
		{ 71, 93, 54 },
		{ 71, 94, 54 },
		{ 72, 46, 54 },
		{ 72, 47, 54 },
		{ 72, 48, 54 },
		{ 72, 49, 54 },
		{ 72, 50, 54 },
		{ 72, 51, 54 },
		{ 72, 52, 54 },
		{ 72, 53, 54 },
		{ 72, 54, 54 },
		{ 72, 55, 54 },
		{ 72, 56, 54 },
		{ 72, 57, 54 },
		{ 72, 58, 54 },
		{ 72, 59, 54 },
		{ 72, 60, 54 },
		{ 72, 61, 54 },
		{ 72, 62, 54 },
		{ 72, 63, 54 },
		{ 72, 64, 54 },
		{ 72, 65, 54 },
		{ 72, 66, 54 },
		{ 72, 67, 54 },
		{ 72, 68, 54 },
		{ 72, 69, 54 },
		{ 72, 70, 54 },
		{ 72, 71, 54 },
		{ 72, 72, 54 },
		{ 72, 73, 54 },
		{ 72, 74, 54 },
		{ 72, 75, 54 },
		{ 72, 76, 54 },
		{ 72, 77, 54 },
		{ 72, 78, 54 },
		{ 72, 79, 54 },
		{ 72, 80, 54 },
		{ 72, 81, 54 },
		{ 72, 82, 54 },
		{ 72, 83, 54 },
		{ 72, 84, 54 },
		{ 72, 85, 54 },
		{ 72, 86, 54 },
		{ 72, 87, 54 },
		{ 72, 88, 54 },
		{ 72, 89, 54 },
		{ 72, 90, 54 },
		{ 72, 91, 54 },
		{ 72, 92, 54 },
		{ 72, 93, 54 },
		{ 72, 94, 54 },
		{ 72, 95, 54 },
		{ 73, 47, 54 },
		{ 73, 48, 54 },
		{ 73, 49, 54 },
		{ 73, 50, 54 },
		{ 73, 51, 54 },
		{ 73, 52, 54 },
		{ 73, 53, 54 },
		{ 73, 54, 54 },
		{ 73, 55, 54 },
		{ 73, 56, 54 },
		{ 73, 57, 54 },
		{ 73, 58, 54 },
		{ 73, 59, 54 },
		{ 73, 60, 54 },
		{ 73, 61, 54 },
		{ 73, 62, 54 },
		{ 73, 63, 54 },
		{ 73, 64, 54 },
		{ 73, 65, 54 },
		{ 73, 66, 54 },
		{ 73, 67, 54 },
		{ 73, 68, 54 },
		{ 73, 69, 54 },
		{ 73, 70, 54 },
		{ 73, 71, 54 },
		{ 73, 72, 54 },
		{ 73, 73, 54 },
		{ 73, 74, 54 },
		{ 73, 75, 54 },
		{ 73, 76, 54 },
		{ 73, 77, 54 },
		{ 73, 78, 54 },
		{ 73, 79, 54 },
		{ 73, 80, 54 },
		{ 73, 81, 54 },
		{ 73, 82, 54 },
		{ 73, 83, 54 },
		{ 73, 84, 54 },
		{ 73, 85, 54 },
		{ 73, 86, 54 },
		{ 73, 87, 54 },
		{ 73, 88, 54 },
		{ 73, 89, 54 },
		{ 73, 90, 54 },
		{ 73, 91, 54 },
		{ 73, 92, 54 },
		{ 73, 93, 54 },
		{ 73, 94, 54 },
		{ 73, 95, 54 },
		{ 74, 48, 54 },
		{ 74, 49, 54 },
		{ 74, 50, 54 },
		{ 74, 51, 54 },
		{ 74, 52, 54 },
		{ 74, 53, 54 },
		{ 74, 54, 54 },
		{ 74, 55, 54 },
		{ 74, 56, 54 },
		{ 74, 57, 54 },
		{ 74, 58, 54 },
		{ 74, 59, 54 },
		{ 74, 60, 54 },
		{ 74, 61, 54 },
		{ 74, 62, 54 },
		{ 74, 63, 54 },
		{ 74, 64, 54 },
		{ 74, 65, 54 },
		{ 74, 66, 54 },
		{ 74, 67, 54 },
		{ 74, 68, 54 },
		{ 74, 69, 54 },
		{ 74, 70, 54 },
		{ 74, 71, 54 },
		{ 74, 72, 54 },
		{ 74, 73, 54 },
		{ 74, 74, 54 },
		{ 74, 75, 54 },
		{ 74, 76, 54 },
		{ 74, 77, 54 },
		{ 74, 78, 54 },
		{ 74, 79, 54 },
		{ 74, 80, 54 },
		{ 74, 81, 54 },
		{ 74, 82, 54 },
		{ 74, 83, 54 },
		{ 74, 84, 54 },
		{ 74, 85, 54 },
		{ 74, 86, 54 },
		{ 74, 87, 54 },
		{ 74, 88, 54 },
		{ 74, 89, 54 },
		{ 74, 90, 54 },
		{ 74, 91, 54 },
		{ 74, 92, 54 },
		{ 74, 93, 54 },
		{ 74, 94, 54 },
		{ 74, 95, 54 },
		{ 74, 96, 54 },
		{ 75, 50, 54 },
		{ 75, 51, 54 },
		{ 75, 52, 54 },
		{ 75, 53, 54 },
		{ 75, 54, 54 },
		{ 75, 55, 54 },
		{ 75, 56, 54 },
		{ 75, 57, 54 },
		{ 75, 58, 54 },
		{ 75, 59, 54 },
		{ 75, 60, 54 },
		{ 75, 61, 54 },
		{ 75, 62, 54 },
		{ 75, 63, 54 },
		{ 75, 64, 54 },
		{ 75, 65, 54 },
		{ 75, 66, 54 },
		{ 75, 67, 54 },
		{ 75, 68, 54 },
		{ 75, 69, 54 },
		{ 75, 70, 54 },
		{ 75, 71, 54 },
		{ 75, 72, 54 },
		{ 75, 73, 54 },
		{ 75, 74, 54 },
		{ 75, 75, 54 },
		{ 75, 76, 54 },
		{ 75, 77, 54 },
		{ 75, 78, 54 },
		{ 75, 79, 54 },
		{ 75, 80, 54 },
		{ 75, 81, 54 },
		{ 75, 82, 54 },
		{ 75, 83, 54 },
		{ 75, 84, 54 },
		{ 75, 85, 54 },
		{ 75, 86, 54 },
		{ 75, 87, 54 },
		{ 75, 88, 54 },
		{ 75, 89, 54 },
		{ 75, 90, 54 },
		{ 75, 91, 54 },
		{ 75, 92, 54 },
		{ 75, 93, 54 },
		{ 75, 94, 54 },
		{ 75, 95, 54 },
		{ 75, 96, 54 },
		{ 75, 97, 54 },
		{ 76, 51, 54 },
		{ 76, 52, 54 },
		{ 76, 53, 54 },
		{ 76, 54, 54 },
		{ 76, 55, 54 },
		{ 76, 56, 54 },
		{ 76, 57, 54 },
		{ 76, 58, 54 },
		{ 76, 59, 54 },
		{ 76, 60, 54 },
		{ 76, 61, 54 },
		{ 76, 62, 54 },
		{ 76, 63, 54 },
		{ 76, 64, 54 },
		{ 76, 65, 54 },
		{ 76, 66, 54 },
		{ 76, 67, 54 },
		{ 76, 68, 54 },
		{ 76, 69, 54 },
		{ 76, 70, 54 },
		{ 76, 71, 54 },
		{ 76, 72, 54 },
		{ 76, 73, 54 },
		{ 76, 74, 54 },
		{ 76, 75, 54 },
		{ 76, 76, 54 },
		{ 76, 77, 54 },
		{ 76, 78, 54 },
		{ 76, 79, 54 },
		{ 76, 80, 54 },
		{ 76, 81, 54 },
		{ 76, 82, 54 },
		{ 76, 83, 54 },
		{ 76, 84, 54 },
		{ 76, 85, 54 },
		{ 76, 86, 54 },
		{ 76, 87, 54 },
		{ 76, 88, 54 },
		{ 76, 89, 54 },
		{ 76, 90, 54 },
		{ 76, 91, 54 },
		{ 76, 92, 54 },
		{ 76, 93, 54 },
		{ 76, 94, 54 },
		{ 76, 95, 54 },
		{ 76, 96, 54 },
		{ 76, 97, 54 },
		{ 76, 98, 54 },
		{ 77, 52, 54 },
		{ 77, 53, 54 },
		{ 77, 54, 54 },
		{ 77, 55, 54 },
		{ 77, 56, 54 },
		{ 77, 57, 54 },
		{ 77, 58, 54 },
		{ 77, 59, 54 },
		{ 77, 60, 54 },
		{ 77, 61, 54 },
		{ 77, 62, 54 },
		{ 77, 63, 54 },
		{ 77, 64, 54 },
		{ 77, 65, 54 },
		{ 77, 66, 54 },
		{ 77, 67, 54 },
		{ 77, 68, 54 },
		{ 77, 69, 54 },
		{ 77, 70, 54 },
		{ 77, 71, 54 },
		{ 77, 72, 54 },
		{ 77, 73, 54 },
		{ 77, 74, 54 },
		{ 77, 75, 54 },
		{ 77, 76, 54 },
		{ 77, 77, 54 },
		{ 77, 78, 54 },
		{ 77, 79, 54 },
		{ 77, 80, 54 },
		{ 77, 81, 54 },
		{ 77, 82, 54 },
		{ 77, 83, 54 },
		{ 77, 84, 54 },
		{ 77, 85, 54 },
		{ 77, 86, 54 },
		{ 77, 87, 54 },
		{ 77, 88, 54 },
		{ 77, 89, 54 },
		{ 77, 90, 54 },
		{ 77, 91, 54 },
		{ 77, 92, 54 },
		{ 77, 93, 54 },
		{ 77, 94, 54 },
		{ 77, 95, 54 },
		{ 77, 96, 54 },
		{ 77, 97, 54 },
		{ 77, 98, 54 },
		{ 77, 99, 54 },
		{ 78, 53, 54 },
		{ 78, 54, 54 },
		{ 78, 55, 54 },
		{ 78, 56, 54 },
		{ 78, 57, 54 },
		{ 78, 58, 54 },
		{ 78, 59, 54 },
		{ 78, 60, 54 },
		{ 78, 61, 54 },
		{ 78, 62, 54 },
		{ 78, 63, 54 },
		{ 78, 64, 54 },
		{ 78, 65, 54 },
		{ 78, 66, 54 },
		{ 78, 67, 54 },
		{ 78, 68, 54 },
		{ 78, 69, 54 },
		{ 78, 70, 54 },
		{ 78, 71, 54 },
		{ 78, 72, 54 },
		{ 78, 73, 54 },
		{ 78, 74, 54 },
		{ 78, 75, 54 },
		{ 78, 76, 54 },
		{ 78, 77, 54 },
		{ 78, 78, 54 },
		{ 78, 79, 54 },
		{ 78, 80, 54 },
		{ 78, 81, 54 },
		{ 78, 82, 54 },
		{ 78, 83, 54 },
		{ 78, 84, 54 },
		{ 78, 85, 54 },
		{ 78, 86, 54 },
		{ 78, 87, 54 },
		{ 78, 88, 54 },
		{ 78, 89, 54 },
		{ 78, 90, 54 },
		{ 78, 91, 54 },
		{ 78, 92, 54 },
		{ 78, 93, 54 },
		{ 78, 94, 54 },
		{ 78, 95, 54 },
		{ 78, 96, 54 },
		{ 78, 97, 54 },
		{ 78, 98, 54 },
		{ 78, 99, 54 },
		{ 78, 100, 54 },
		{ 79, 54, 54 },
		{ 79, 55, 54 },
		{ 79, 56, 54 },
		{ 79, 57, 54 },
		{ 79, 58, 54 },
		{ 79, 59, 54 },
		{ 79, 60, 54 },
		{ 79, 61, 54 },
		{ 79, 62, 54 },
		{ 79, 63, 54 },
		{ 79, 64, 54 },
		{ 79, 65, 54 },
		{ 79, 66, 54 },
		{ 79, 67, 54 },
		{ 79, 68, 54 },
		{ 79, 69, 54 },
		{ 79, 70, 54 },
		{ 79, 71, 54 },
		{ 79, 72, 54 },
		{ 79, 73, 54 },
		{ 79, 74, 54 },
		{ 79, 75, 54 },
		{ 79, 76, 54 },
		{ 79, 77, 54 },
		{ 79, 78, 54 },
		{ 79, 79, 54 },
		{ 79, 80, 54 },
		{ 79, 81, 54 },
		{ 79, 82, 54 },
		{ 79, 83, 54 },
		{ 79, 84, 54 },
		{ 79, 85, 54 },
		{ 79, 86, 54 },
		{ 79, 87, 54 },
		{ 79, 88, 54 },
		{ 79, 89, 54 },
		{ 79, 90, 54 },
		{ 79, 91, 54 },
		{ 79, 92, 54 },
		{ 79, 93, 54 },
		{ 79, 94, 54 },
		{ 79, 95, 54 },
		{ 79, 96, 54 },
		{ 79, 97, 54 },
		{ 79, 98, 54 },
		{ 79, 99, 54 },
		{ 79, 100, 54 },
		{ 79, 101, 54 },
		{ 80, 55, 54 },
		{ 80, 56, 54 },
		{ 80, 57, 54 },
		{ 80, 58, 54 },
		{ 80, 59, 54 },
		{ 80, 60, 54 },
		{ 80, 61, 54 },
		{ 80, 62, 54 },
		{ 80, 63, 54 },
		{ 80, 64, 54 },
		{ 80, 65, 54 },
		{ 80, 66, 54 },
		{ 80, 67, 54 },
		{ 80, 68, 54 },
		{ 80, 69, 54 },
		{ 80, 70, 54 },
		{ 80, 71, 54 },
		{ 80, 72, 54 },
		{ 80, 73, 54 },
		{ 80, 74, 54 },
		{ 80, 75, 54 },
		{ 80, 76, 54 },
		{ 80, 77, 54 },
		{ 80, 78, 54 },
		{ 80, 79, 54 },
		{ 80, 80, 54 },
		{ 80, 81, 54 },
		{ 80, 82, 54 },
		{ 80, 83, 54 },
		{ 80, 84, 54 },
		{ 80, 85, 54 },
		{ 80, 86, 54 },
		{ 80, 87, 54 },
		{ 80, 88, 54 },
		{ 80, 89, 54 },
		{ 80, 90, 54 },
		{ 80, 91, 54 },
		{ 80, 92, 54 },
		{ 80, 93, 54 },
		{ 80, 94, 54 },
		{ 80, 95, 54 },
		{ 80, 96, 54 },
		{ 80, 97, 54 },
		{ 80, 98, 54 },
		{ 80, 99, 54 },
		{ 80, 100, 54 },
		{ 80, 101, 54 },
		{ 80, 102, 54 },
		{ 81, 56, 54 },
		{ 81, 57, 54 },
		{ 81, 58, 54 },
		{ 81, 59, 54 },
		{ 81, 60, 54 },
		{ 81, 61, 54 },
		{ 81, 62, 54 },
		{ 81, 63, 54 },
		{ 81, 64, 54 },
		{ 81, 65, 54 },
		{ 81, 66, 54 },
		{ 81, 67, 54 },
		{ 81, 68, 54 },
		{ 81, 69, 54 },
		{ 81, 70, 54 },
		{ 81, 71, 54 },
		{ 81, 72, 54 },
		{ 81, 73, 54 },
		{ 81, 74, 54 },
		{ 81, 75, 54 },
		{ 81, 76, 54 },
		{ 81, 77, 54 },
		{ 81, 78, 54 },
		{ 81, 79, 54 },
		{ 81, 80, 54 },
		{ 81, 81, 54 },
		{ 81, 82, 54 },
		{ 81, 83, 54 },
		{ 81, 84, 54 },
		{ 81, 85, 54 },
		{ 81, 86, 54 },
		{ 81, 87, 54 },
		{ 81, 88, 54 },
		{ 81, 89, 54 },
		{ 81, 90, 54 },
		{ 81, 91, 54 },
		{ 81, 92, 54 },
		{ 81, 93, 54 },
		{ 81, 94, 54 },
		{ 81, 95, 54 },
		{ 81, 96, 54 },
		{ 81, 97, 54 },
		{ 81, 98, 54 },
		{ 81, 99, 54 },
		{ 81, 100, 54 },
		{ 81, 101, 54 },
		{ 81, 102, 54 },
		{ 82, 58, 54 },
		{ 82, 59, 54 },
		{ 82, 60, 54 },
		{ 82, 61, 54 },
		{ 82, 62, 54 },
		{ 82, 63, 54 },
		{ 82, 64, 54 },
		{ 82, 65, 54 },
		{ 82, 66, 54 },
		{ 82, 67, 54 },
		{ 82, 68, 54 },
		{ 82, 69, 54 },
		{ 82, 70, 54 },
		{ 82, 71, 54 },
		{ 82, 72, 54 },
		{ 82, 73, 54 },
		{ 82, 74, 54 },
		{ 82, 75, 54 },
		{ 82, 76, 54 },
		{ 82, 77, 54 },
		{ 82, 78, 54 },
		{ 82, 79, 54 },
		{ 82, 80, 54 },
		{ 82, 81, 54 },
		{ 82, 82, 54 },
		{ 82, 83, 54 },
		{ 82, 84, 54 },
		{ 82, 85, 54 },
		{ 82, 86, 54 },
		{ 82, 87, 54 },
		{ 82, 88, 54 },
		{ 82, 89, 54 },
		{ 82, 90, 54 },
		{ 82, 91, 54 },
		{ 82, 92, 54 },
		{ 82, 93, 54 },
		{ 82, 94, 54 },
		{ 82, 95, 54 },
		{ 82, 96, 54 },
		{ 82, 97, 54 },
		{ 82, 98, 54 },
		{ 82, 99, 54 },
		{ 82, 100, 54 },
		{ 82, 101, 54 },
		{ 82, 102, 54 },
		{ 82, 103, 54 },
		{ 83, 59, 54 },
		{ 83, 60, 54 },
		{ 83, 61, 54 },
		{ 83, 62, 54 },
		{ 83, 63, 54 },
		{ 83, 64, 54 },
		{ 83, 65, 54 },
		{ 83, 66, 54 },
		{ 83, 67, 54 },
		{ 83, 68, 54 },
		{ 83, 69, 54 },
		{ 83, 70, 54 },
		{ 83, 71, 54 },
		{ 83, 72, 54 },
		{ 83, 73, 54 },
		{ 83, 74, 54 },
		{ 83, 75, 54 },
		{ 83, 76, 54 },
		{ 83, 77, 54 },
		{ 83, 78, 54 },
		{ 83, 79, 54 },
		{ 83, 80, 54 },
		{ 83, 81, 54 },
		{ 83, 82, 54 },
		{ 83, 83, 54 },
		{ 83, 84, 54 },
		{ 83, 85, 54 },
		{ 83, 86, 54 },
		{ 83, 87, 54 },
		{ 83, 88, 54 },
		{ 83, 89, 54 },
		{ 83, 90, 54 },
		{ 83, 91, 54 },
		{ 83, 92, 54 },
		{ 83, 93, 54 },
		{ 83, 94, 54 },
		{ 83, 95, 54 },
		{ 83, 96, 54 },
		{ 83, 97, 54 },
		{ 83, 98, 54 },
		{ 83, 99, 54 },
		{ 83, 100, 54 },
		{ 83, 101, 54 },
		{ 83, 102, 54 },
		{ 83, 103, 54 },
		{ 83, 104, 54 },
		{ 84, 60, 54 },
		{ 84, 61, 54 },
		{ 84, 62, 54 },
		{ 84, 63, 54 },
		{ 84, 64, 54 },
		{ 84, 65, 54 },
		{ 84, 66, 54 },
		{ 84, 67, 54 },
		{ 84, 68, 54 },
		{ 84, 69, 54 },
		{ 84, 70, 54 },
		{ 84, 71, 54 },
		{ 84, 72, 54 },
		{ 84, 73, 54 },
		{ 84, 74, 54 },
		{ 84, 75, 54 },
		{ 84, 76, 54 },
		{ 84, 77, 54 },
		{ 84, 78, 54 },
		{ 84, 79, 54 },
		{ 84, 80, 54 },
		{ 84, 81, 54 },
		{ 84, 82, 54 },
		{ 84, 83, 54 },
		{ 84, 84, 54 },
		{ 84, 85, 54 },
		{ 84, 86, 54 },
		{ 84, 87, 54 },
		{ 84, 88, 54 },
		{ 84, 89, 54 },
		{ 84, 90, 54 },
		{ 84, 91, 54 },
		{ 84, 92, 54 },
		{ 84, 93, 54 },
		{ 84, 94, 54 },
		{ 84, 95, 54 },
		{ 84, 96, 54 },
		{ 84, 97, 54 },
		{ 84, 98, 54 },
		{ 84, 99, 54 },
		{ 84, 100, 54 },
		{ 84, 101, 54 },
		{ 84, 102, 54 },
		{ 84, 103, 54 },
		{ 84, 104, 54 },
		{ 84, 105, 54 },
		{ 85, 61, 54 },
		{ 85, 62, 54 },
		{ 85, 63, 54 },
		{ 85, 64, 54 },
		{ 85, 65, 54 },
		{ 85, 66, 54 },
		{ 85, 67, 54 },
		{ 85, 68, 54 },
		{ 85, 69, 54 },
		{ 85, 70, 54 },
		{ 85, 71, 54 },
		{ 85, 72, 54 },
		{ 85, 73, 54 },
		{ 85, 74, 54 },
		{ 85, 75, 54 },
		{ 85, 76, 54 },
		{ 85, 77, 54 },
		{ 85, 78, 54 },
		{ 85, 79, 54 },
		{ 85, 80, 54 },
		{ 85, 81, 54 },
		{ 85, 82, 54 },
		{ 85, 83, 54 },
		{ 85, 84, 54 },
		{ 85, 85, 54 },
		{ 85, 86, 54 },
		{ 85, 87, 54 },
		{ 85, 88, 54 },
		{ 85, 89, 54 },
		{ 85, 90, 54 },
		{ 85, 91, 54 },
		{ 85, 92, 54 },
		{ 85, 93, 54 },
		{ 85, 94, 54 },
		{ 85, 95, 54 },
		{ 85, 96, 54 },
		{ 85, 97, 54 },
		{ 85, 98, 54 },
		{ 85, 99, 54 },
		{ 85, 100, 54 },
		{ 85, 101, 54 },
		{ 85, 102, 54 },
		{ 85, 103, 54 },
		{ 85, 104, 54 },
		{ 85, 105, 54 },
		{ 85, 106, 54 },
		{ 86, 62, 54 },
		{ 86, 63, 54 },
		{ 86, 64, 54 },
		{ 86, 65, 54 },
		{ 86, 66, 54 },
		{ 86, 67, 54 },
		{ 86, 68, 54 },
		{ 86, 69, 54 },
		{ 86, 70, 54 },
		{ 86, 71, 54 },
		{ 86, 72, 54 },
		{ 86, 73, 54 },
		{ 86, 74, 54 },
		{ 86, 75, 54 },
		{ 86, 76, 54 },
		{ 86, 77, 54 },
		{ 86, 78, 54 },
		{ 86, 79, 54 },
		{ 86, 80, 54 },
		{ 86, 81, 54 },
		{ 86, 82, 54 },
		{ 86, 83, 54 },
		{ 86, 84, 54 },
		{ 86, 85, 54 },
		{ 86, 86, 54 },
		{ 86, 87, 54 },
		{ 86, 88, 54 },
		{ 86, 89, 54 },
		{ 86, 90, 54 },
		{ 86, 91, 54 },
		{ 86, 92, 54 },
		{ 86, 93, 54 },
		{ 86, 94, 54 },
		{ 86, 95, 54 },
		{ 86, 96, 54 },
		{ 86, 97, 54 },
		{ 86, 98, 54 },
		{ 86, 99, 54 },
		{ 86, 100, 54 },
		{ 86, 101, 54 },
		{ 86, 102, 54 },
		{ 86, 103, 54 },
		{ 86, 104, 54 },
		{ 86, 105, 54 },
		{ 86, 106, 54 },
		{ 86, 107, 54 },
		{ 87, 63, 54 },
		{ 87, 64, 54 },
		{ 87, 65, 54 },
		{ 87, 66, 54 },
		{ 87, 67, 54 },
		{ 87, 68, 54 },
		{ 87, 69, 54 },
		{ 87, 70, 54 },
		{ 87, 71, 54 },
		{ 87, 72, 54 },
		{ 87, 73, 54 },
		{ 87, 74, 54 },
		{ 87, 75, 54 },
		{ 87, 76, 54 },
		{ 87, 77, 54 },
		{ 87, 78, 54 },
		{ 87, 79, 54 },
		{ 87, 80, 54 },
		{ 87, 81, 54 },
		{ 87, 82, 54 },
		{ 87, 83, 54 },
		{ 87, 84, 54 },
		{ 87, 85, 54 },
		{ 87, 86, 54 },
		{ 87, 87, 54 },
		{ 87, 88, 54 },
		{ 87, 89, 54 },
		{ 87, 90, 54 },
		{ 87, 91, 54 },
		{ 87, 92, 54 },
		{ 87, 93, 54 },
		{ 87, 94, 54 },
		{ 87, 95, 54 },
		{ 87, 96, 54 },
		{ 87, 97, 54 },
		{ 87, 98, 54 },
		{ 87, 99, 54 },
		{ 87, 100, 54 },
		{ 87, 101, 54 },
		{ 87, 102, 54 },
		{ 87, 103, 54 },
		{ 87, 104, 54 },
		{ 87, 105, 54 },
		{ 87, 106, 54 },
		{ 87, 107, 54 },
		{ 87, 108, 54 },
		{ 88, 64, 54 },
		{ 88, 65, 54 },
		{ 88, 66, 54 },
		{ 88, 67, 54 },
		{ 88, 68, 54 },
		{ 88, 69, 54 },
		{ 88, 70, 54 },
		{ 88, 71, 54 },
		{ 88, 72, 54 },
		{ 88, 73, 54 },
		{ 88, 74, 54 },
		{ 88, 75, 54 },
		{ 88, 76, 54 },
		{ 88, 77, 54 },
		{ 88, 78, 54 },
		{ 88, 79, 54 },
		{ 88, 80, 54 },
		{ 88, 81, 54 },
		{ 88, 82, 54 },
		{ 88, 83, 54 },
		{ 88, 84, 54 },
		{ 88, 85, 54 },
		{ 88, 86, 54 },
		{ 88, 87, 54 },
		{ 88, 88, 54 },
		{ 88, 89, 54 },
		{ 88, 90, 54 },
		{ 88, 91, 54 },
		{ 88, 92, 54 },
		{ 88, 93, 54 },
		{ 88, 94, 54 },
		{ 88, 95, 54 },
		{ 88, 96, 54 },
		{ 88, 97, 54 },
		{ 88, 98, 54 },
		{ 88, 99, 54 },
		{ 88, 100, 54 },
		{ 88, 101, 54 },
		{ 88, 102, 54 },
		{ 88, 103, 54 },
		{ 88, 104, 54 },
		{ 88, 105, 54 },
		{ 88, 106, 54 },
		{ 88, 107, 54 },
		{ 88, 108, 54 },
		{ 88, 109, 54 },
		{ 89, 66, 54 },
		{ 89, 67, 54 },
		{ 89, 68, 54 },
		{ 89, 69, 54 },
		{ 89, 70, 54 },
		{ 89, 71, 54 },
		{ 89, 72, 54 },
		{ 89, 73, 54 },
		{ 89, 74, 54 },
		{ 89, 75, 54 },
		{ 89, 76, 54 },
		{ 89, 77, 54 },
		{ 89, 78, 54 },
		{ 89, 79, 54 },
		{ 89, 80, 54 },
		{ 89, 81, 54 },
		{ 89, 82, 54 },
		{ 89, 83, 54 },
		{ 89, 84, 54 },
		{ 89, 85, 54 },
		{ 89, 86, 54 },
		{ 89, 87, 54 },
		{ 89, 88, 54 },
		{ 89, 89, 54 },
		{ 89, 90, 54 },
		{ 89, 91, 54 },
		{ 89, 92, 54 },
		{ 89, 93, 54 },
		{ 89, 94, 54 },
		{ 89, 95, 54 },
		{ 89, 96, 54 },
		{ 89, 97, 54 },
		{ 89, 98, 54 },
		{ 89, 99, 54 },
		{ 89, 100, 54 },
		{ 89, 101, 54 },
		{ 89, 102, 54 },
		{ 89, 103, 54 },
		{ 89, 104, 54 },
		{ 89, 105, 54 },
		{ 89, 106, 54 },
		{ 89, 107, 54 },
		{ 89, 108, 54 },
		{ 89, 109, 54 },
		{ 90, 67, 54 },
		{ 90, 68, 54 },
		{ 90, 69, 54 },
		{ 90, 70, 54 },
		{ 90, 71, 54 },
		{ 90, 72, 54 },
		{ 90, 73, 54 },
		{ 90, 74, 54 },
		{ 90, 75, 54 },
		{ 90, 76, 54 },
		{ 90, 77, 54 },
		{ 90, 78, 54 },
		{ 90, 79, 54 },
		{ 90, 80, 54 },
		{ 90, 81, 54 },
		{ 90, 82, 54 },
		{ 90, 83, 54 },
		{ 90, 84, 54 },
		{ 90, 85, 54 },
		{ 90, 86, 54 },
		{ 90, 87, 54 },
		{ 90, 88, 54 },
		{ 90, 89, 54 },
		{ 90, 90, 54 },
		{ 90, 91, 54 },
		{ 90, 92, 54 },
		{ 90, 93, 54 },
		{ 90, 94, 54 },
		{ 90, 95, 54 },
		{ 90, 96, 54 },
		{ 90, 97, 54 },
		{ 90, 98, 54 },
		{ 90, 99, 54 },
		{ 90, 100, 54 },
		{ 90, 101, 54 },
		{ 90, 102, 54 },
		{ 90, 103, 54 },
		{ 90, 104, 54 },
		{ 90, 105, 54 },
		{ 90, 106, 54 },
		{ 90, 107, 54 },
		{ 90, 108, 54 },
		{ 90, 109, 54 },
		{ 90, 110, 54 },
		{ 91, 68, 54 },
		{ 91, 69, 54 },
		{ 91, 70, 54 },
		{ 91, 71, 54 },
		{ 91, 72, 54 },
		{ 91, 73, 54 },
		{ 91, 74, 54 },
		{ 91, 75, 54 },
		{ 91, 76, 54 },
		{ 91, 77, 54 },
		{ 91, 78, 54 },
		{ 91, 79, 54 },
		{ 91, 80, 54 },
		{ 91, 81, 54 },
		{ 91, 82, 54 },
		{ 91, 83, 54 },
		{ 91, 84, 54 },
		{ 91, 85, 54 },
		{ 91, 86, 54 },
		{ 91, 87, 54 },
		{ 91, 88, 54 },
		{ 91, 89, 54 },
		{ 91, 90, 54 },
		{ 91, 91, 54 },
		{ 91, 92, 54 },
		{ 91, 93, 54 },
		{ 91, 94, 54 },
		{ 91, 95, 54 },
		{ 91, 96, 54 },
		{ 91, 97, 54 },
		{ 91, 98, 54 },
		{ 91, 99, 54 },
		{ 91, 100, 54 },
		{ 91, 101, 54 },
		{ 91, 102, 54 },
		{ 91, 103, 54 },
		{ 91, 104, 54 },
		{ 91, 105, 54 },
		{ 91, 106, 54 },
		{ 91, 107, 54 },
		{ 91, 108, 54 },
		{ 91, 109, 54 },
		{ 91, 110, 54 },
		{ 91, 111, 54 },
		{ 92, 69, 54 },
		{ 92, 70, 54 },
		{ 92, 71, 54 },
		{ 92, 72, 54 },
		{ 92, 73, 54 },
		{ 92, 74, 54 },
		{ 92, 75, 54 },
		{ 92, 76, 54 },
		{ 92, 77, 54 },
		{ 92, 78, 54 },
		{ 92, 79, 54 },
		{ 92, 80, 54 },
		{ 92, 81, 54 },
		{ 92, 82, 54 },
		{ 92, 83, 54 },
		{ 92, 84, 54 },
		{ 92, 85, 54 },
		{ 92, 86, 54 },
		{ 92, 87, 54 },
		{ 92, 88, 54 },
		{ 92, 89, 54 },
		{ 92, 90, 54 },
		{ 92, 91, 54 },
		{ 92, 92, 54 },
		{ 92, 93, 54 },
		{ 92, 94, 54 },
		{ 92, 95, 54 },
		{ 92, 96, 54 },
		{ 92, 97, 54 },
		{ 92, 98, 54 },
		{ 92, 99, 54 },
		{ 92, 100, 54 },
		{ 92, 101, 54 },
		{ 92, 102, 54 },
		{ 92, 103, 54 },
		{ 92, 104, 54 },
		{ 92, 105, 54 },
		{ 92, 106, 54 },
		{ 92, 107, 54 },
		{ 92, 108, 54 },
		{ 92, 109, 54 },
		{ 92, 110, 54 },
		{ 92, 111, 54 },
		{ 92, 112, 54 },
		{ 93, 70, 54 },
		{ 93, 71, 54 },
		{ 93, 72, 54 },
		{ 93, 73, 54 },
		{ 93, 74, 54 },
		{ 93, 75, 54 },
		{ 93, 76, 54 },
		{ 93, 77, 54 },
		{ 93, 78, 54 },
		{ 93, 79, 54 },
		{ 93, 80, 54 },
		{ 93, 81, 54 },
		{ 93, 82, 54 },
		{ 93, 83, 54 },
		{ 93, 84, 54 },
		{ 93, 85, 54 },
		{ 93, 86, 54 },
		{ 93, 87, 54 },
		{ 93, 88, 54 },
		{ 93, 89, 54 },
		{ 93, 90, 54 },
		{ 93, 91, 54 },
		{ 93, 92, 54 },
		{ 93, 93, 54 },
		{ 93, 94, 54 },
		{ 93, 95, 54 },
		{ 93, 96, 54 },
		{ 93, 97, 54 },
		{ 93, 98, 54 },
		{ 93, 99, 54 },
		{ 93, 100, 54 },
		{ 93, 101, 54 },
		{ 93, 102, 54 },
		{ 93, 103, 54 },
		{ 93, 104, 54 },
		{ 93, 105, 54 },
		{ 93, 106, 54 },
		{ 93, 107, 54 },
		{ 93, 108, 54 },
		{ 93, 109, 54 },
		{ 93, 110, 54 },
		{ 93, 111, 54 },
		{ 93, 112, 54 },
		{ 93, 113, 54 },
		{ 94, 71, 54 },
		{ 94, 72, 54 },
		{ 94, 73, 54 },
		{ 94, 74, 54 },
		{ 94, 75, 54 },
		{ 94, 76, 54 },
		{ 94, 77, 54 },
		{ 94, 78, 54 },
		{ 94, 79, 54 },
		{ 94, 80, 54 },
		{ 94, 81, 54 },
		{ 94, 82, 54 },
		{ 94, 83, 54 },
		{ 94, 84, 54 },
		{ 94, 85, 54 },
		{ 94, 86, 54 },
		{ 94, 87, 54 },
		{ 94, 88, 54 },
		{ 94, 89, 54 },
		{ 94, 90, 54 },
		{ 94, 91, 54 },
		{ 94, 92, 54 },
		{ 94, 93, 54 },
		{ 94, 94, 54 },
		{ 94, 95, 54 },
		{ 94, 96, 54 },
		{ 94, 97, 54 },
		{ 94, 98, 54 },
		{ 94, 99, 54 },
		{ 94, 100, 54 },
		{ 94, 101, 54 },
		{ 94, 102, 54 },
		{ 94, 103, 54 },
		{ 94, 104, 54 },
		{ 94, 105, 54 },
		{ 94, 106, 54 },
		{ 94, 107, 54 },
		{ 94, 108, 54 },
		{ 94, 109, 54 },
		{ 94, 110, 54 },
		{ 94, 111, 54 },
		{ 94, 112, 54 },
		{ 94, 113, 54 },
		{ 94, 114, 54 },
		{ 95, 72, 54 },
		{ 95, 73, 54 },
		{ 95, 74, 54 },
		{ 95, 75, 54 },
		{ 95, 76, 54 },
		{ 95, 77, 54 },
		{ 95, 78, 54 },
		{ 95, 79, 54 },
		{ 95, 80, 54 },
		{ 95, 81, 54 },
		{ 95, 82, 54 },
		{ 95, 83, 54 },
		{ 95, 84, 54 },
		{ 95, 85, 54 },
		{ 95, 86, 54 },
		{ 95, 87, 54 },
		{ 95, 88, 54 },
		{ 95, 89, 54 },
		{ 95, 90, 54 },
		{ 95, 91, 54 },
		{ 95, 92, 54 },
		{ 95, 93, 54 },
		{ 95, 94, 54 },
		{ 95, 95, 54 },
		{ 95, 96, 54 },
		{ 95, 97, 54 },
		{ 95, 98, 54 },
		{ 95, 99, 54 },
		{ 95, 100, 54 },
		{ 95, 101, 54 },
		{ 95, 102, 54 },
		{ 95, 103, 54 },
		{ 95, 104, 54 },
		{ 95, 105, 54 },
		{ 95, 106, 54 },
		{ 95, 107, 54 },
		{ 95, 108, 54 },
		{ 95, 109, 54 },
		{ 95, 110, 54 },
		{ 95, 111, 54 },
		{ 95, 112, 54 },
		{ 95, 113, 54 },
		{ 95, 114, 54 },
		{ 95, 115, 54 },
		{ 96, 74, 54 },
		{ 96, 75, 54 },
		{ 96, 76, 54 },
		{ 96, 77, 54 },
		{ 96, 78, 54 },
		{ 96, 79, 54 },
		{ 96, 80, 54 },
		{ 96, 81, 54 },
		{ 96, 82, 54 },
		{ 96, 83, 54 },
		{ 96, 84, 54 },
		{ 96, 85, 54 },
		{ 96, 86, 54 },
		{ 96, 87, 54 },
		{ 96, 88, 54 },
		{ 96, 89, 54 },
		{ 96, 90, 54 },
		{ 96, 91, 54 },
		{ 96, 92, 54 },
		{ 96, 93, 54 },
		{ 96, 94, 54 },
		{ 96, 95, 54 },
		{ 96, 96, 54 },
		{ 96, 97, 54 },
		{ 96, 98, 54 },
		{ 96, 99, 54 },
		{ 96, 100, 54 },
		{ 96, 101, 54 },
		{ 96, 102, 54 },
		{ 96, 103, 54 },
		{ 96, 104, 54 },
		{ 96, 105, 54 },
		{ 96, 106, 54 },
		{ 96, 107, 54 },
		{ 96, 108, 54 },
		{ 96, 109, 54 },
		{ 96, 110, 54 },
		{ 96, 111, 54 },
		{ 96, 112, 54 },
		{ 96, 113, 54 },
		{ 96, 114, 54 },
		{ 96, 115, 54 },
		{ 96, 116, 54 },
		{ 97, 75, 54 },
		{ 97, 76, 54 },
		{ 97, 77, 54 },
		{ 97, 78, 54 },
		{ 97, 79, 54 },
		{ 97, 80, 54 },
		{ 97, 81, 54 },
		{ 97, 82, 54 },
		{ 97, 83, 54 },
		{ 97, 84, 54 },
		{ 97, 85, 54 },
		{ 97, 86, 54 },
		{ 97, 87, 54 },
		{ 97, 88, 54 },
		{ 97, 89, 54 },
		{ 97, 90, 54 },
		{ 97, 91, 54 },
		{ 97, 92, 54 },
		{ 97, 93, 54 },
		{ 97, 94, 54 },
		{ 97, 95, 54 },
		{ 97, 96, 54 },
		{ 97, 97, 54 },
		{ 97, 98, 54 },
		{ 97, 99, 54 },
		{ 97, 100, 54 },
		{ 97, 101, 54 },
		{ 97, 102, 54 },
		{ 97, 103, 54 },
		{ 97, 104, 54 },
		{ 97, 105, 54 },
		{ 97, 106, 54 },
		{ 97, 107, 54 },
		{ 97, 108, 54 },
		{ 97, 109, 54 },
		{ 97, 110, 54 },
		{ 97, 111, 54 },
		{ 97, 112, 54 },
		{ 97, 113, 54 },
		{ 97, 114, 54 },
		{ 97, 115, 54 },
		{ 97, 116, 54 },
		{ 98, 76, 54 },
		{ 98, 77, 54 },
		{ 98, 78, 54 },
		{ 98, 79, 54 },
		{ 98, 80, 54 },
		{ 98, 81, 54 },
		{ 98, 82, 54 },
		{ 98, 83, 54 },
		{ 98, 84, 54 },
		{ 98, 85, 54 },
		{ 98, 86, 54 },
		{ 98, 87, 54 },
		{ 98, 88, 54 },
		{ 98, 89, 54 },
		{ 98, 90, 54 },
		{ 98, 91, 54 },
		{ 98, 92, 54 },
		{ 98, 93, 54 },
		{ 98, 94, 54 },
		{ 98, 95, 54 },
		{ 98, 96, 54 },
		{ 98, 97, 54 },
		{ 98, 98, 54 },
		{ 98, 99, 54 },
		{ 98, 100, 54 },
		{ 98, 101, 54 },
		{ 98, 102, 54 },
		{ 98, 103, 54 },
		{ 98, 104, 54 },
		{ 98, 105, 54 },
		{ 98, 106, 54 },
		{ 98, 107, 54 },
		{ 98, 108, 54 },
		{ 98, 109, 54 },
		{ 98, 110, 54 },
		{ 98, 111, 54 },
		{ 98, 112, 54 },
		{ 98, 113, 54 },
		{ 98, 114, 54 },
		{ 98, 115, 54 },
		{ 98, 116, 54 },
		{ 98, 117, 54 },
		{ 99, 77, 54 },
		{ 99, 78, 54 },
		{ 99, 79, 54 },
		{ 99, 80, 54 },
		{ 99, 81, 54 },
		{ 99, 82, 54 },
		{ 99, 83, 54 },
		{ 99, 84, 54 },
		{ 99, 85, 54 },
		{ 99, 86, 54 },
		{ 99, 87, 54 },
		{ 99, 88, 54 },
		{ 99, 89, 54 },
		{ 99, 90, 54 },
		{ 99, 91, 54 },
		{ 99, 92, 54 },
		{ 99, 93, 54 },
		{ 99, 94, 54 },
		{ 99, 95, 54 },
		{ 99, 96, 54 },
		{ 99, 97, 54 },
		{ 99, 98, 54 },
		{ 99, 99, 54 },
		{ 99, 100, 54 },
		{ 99, 101, 54 },
		{ 99, 102, 54 },
		{ 99, 103, 54 },
		{ 99, 104, 54 },
		{ 99, 105, 54 },
		{ 99, 106, 54 },
		{ 99, 107, 54 },
		{ 99, 108, 54 },
		{ 99, 109, 54 },
		{ 99, 110, 54 },
		{ 99, 111, 54 },
		{ 99, 112, 54 },
		{ 99, 113, 54 },
		{ 99, 114, 54 },
		{ 99, 115, 54 },
		{ 99, 116, 54 },
		{ 99, 117, 54 },
		{ 99, 118, 54 },
		{ 100, 78, 54 },
		{ 100, 79, 54 },
		{ 100, 80, 54 },
		{ 100, 81, 54 },
		{ 100, 82, 54 },
		{ 100, 83, 54 },
		{ 100, 84, 54 },
		{ 100, 85, 54 },
		{ 100, 86, 54 },
		{ 100, 87, 54 },
		{ 100, 88, 54 },
		{ 100, 89, 54 },
		{ 100, 90, 54 },
		{ 100, 91, 54 },
		{ 100, 92, 54 },
		{ 100, 93, 54 },
		{ 100, 94, 54 },
		{ 100, 95, 54 },
		{ 100, 96, 54 },
		{ 100, 97, 54 },
		{ 100, 98, 54 },
		{ 100, 99, 54 },
		{ 100, 100, 54 },
		{ 100, 101, 54 },
		{ 100, 102, 54 },
		{ 100, 103, 54 },
		{ 100, 104, 54 },
		{ 100, 105, 54 },
		{ 100, 106, 54 },
		{ 100, 107, 54 },
		{ 100, 108, 54 },
		{ 100, 109, 54 },
		{ 100, 110, 54 },
		{ 100, 111, 54 },
		{ 100, 112, 54 },
		{ 100, 113, 54 },
		{ 100, 114, 54 },
		{ 100, 115, 54 },
		{ 100, 116, 54 },
		{ 100, 117, 54 },
		{ 100, 118, 54 },
		{ 100, 119, 54 },
		{ 101, 79, 54 },
		{ 101, 80, 54 },
		{ 101, 81, 54 },
		{ 101, 82, 54 },
		{ 101, 83, 54 },
		{ 101, 84, 54 },
		{ 101, 85, 54 },
		{ 101, 86, 54 },
		{ 101, 87, 54 },
		{ 101, 88, 54 },
		{ 101, 89, 54 },
		{ 101, 90, 54 },
		{ 101, 91, 54 },
		{ 101, 92, 54 },
		{ 101, 93, 54 },
		{ 101, 94, 54 },
		{ 101, 95, 54 },
		{ 101, 96, 54 },
		{ 101, 97, 54 },
		{ 101, 98, 54 },
		{ 101, 99, 54 },
		{ 101, 100, 54 },
		{ 101, 101, 54 },
		{ 101, 102, 54 },
		{ 101, 103, 54 },
		{ 101, 104, 54 },
		{ 101, 105, 54 },
		{ 101, 106, 54 },
		{ 101, 107, 54 },
		{ 101, 108, 54 },
		{ 101, 109, 54 },
		{ 101, 110, 54 },
		{ 101, 111, 54 },
		{ 101, 112, 54 },
		{ 101, 113, 54 },
		{ 101, 114, 54 },
		{ 101, 115, 54 },
		{ 101, 116, 54 },
		{ 101, 117, 54 },
		{ 101, 118, 54 },
		{ 101, 119, 54 },
		{ 101, 120, 54 },
		{ 102, 80, 54 },
		{ 102, 81, 54 },
		{ 102, 82, 54 },
		{ 102, 83, 54 },
		{ 102, 84, 54 },
		{ 102, 85, 54 },
		{ 102, 86, 54 },
		{ 102, 87, 54 },
		{ 102, 88, 54 },
		{ 102, 89, 54 },
		{ 102, 90, 54 },
		{ 102, 91, 54 },
		{ 102, 92, 54 },
		{ 102, 93, 54 },
		{ 102, 94, 54 },
		{ 102, 95, 54 },
		{ 102, 96, 54 },
		{ 102, 97, 54 },
		{ 102, 98, 54 },
		{ 102, 99, 54 },
		{ 102, 100, 54 },
		{ 102, 101, 54 },
		{ 102, 102, 54 },
		{ 102, 103, 54 },
		{ 102, 104, 54 },
		{ 102, 105, 54 },
		{ 102, 106, 54 },
		{ 102, 107, 54 },
		{ 102, 108, 54 },
		{ 102, 109, 54 },
		{ 102, 110, 54 },
		{ 102, 111, 54 },
		{ 102, 112, 54 },
		{ 102, 113, 54 },
		{ 102, 114, 54 },
		{ 102, 115, 54 },
		{ 102, 116, 54 },
		{ 102, 117, 54 },
		{ 102, 118, 54 },
		{ 102, 119, 54 },
		{ 102, 120, 54 },
		{ 102, 121, 54 },
		{ 103, 82, 54 },
		{ 103, 83, 54 },
		{ 103, 84, 54 },
		{ 103, 85, 54 },
		{ 103, 86, 54 },
		{ 103, 87, 54 },
		{ 103, 88, 54 },
		{ 103, 89, 54 },
		{ 103, 90, 54 },
		{ 103, 91, 54 },
		{ 103, 92, 54 },
		{ 103, 93, 54 },
		{ 103, 94, 54 },
		{ 103, 95, 54 },
		{ 103, 96, 54 },
		{ 103, 97, 54 },
		{ 103, 98, 54 },
		{ 103, 99, 54 },
		{ 103, 100, 54 },
		{ 103, 101, 54 },
		{ 103, 102, 54 },
		{ 103, 103, 54 },
		{ 103, 104, 54 },
		{ 103, 105, 54 },
		{ 103, 106, 54 },
		{ 103, 107, 54 },
		{ 103, 108, 54 },
		{ 103, 109, 54 },
		{ 103, 110, 54 },
		{ 103, 111, 54 },
		{ 103, 112, 54 },
		{ 103, 113, 54 },
		{ 103, 114, 54 },
		{ 103, 115, 54 },
		{ 103, 116, 54 },
		{ 103, 117, 54 },
		{ 103, 118, 54 },
		{ 103, 119, 54 },
		{ 103, 120, 54 },
		{ 103, 121, 54 },
		{ 103, 122, 54 },
		{ 104, 83, 54 },
		{ 104, 84, 54 },
		{ 104, 85, 54 },
		{ 104, 86, 54 },
		{ 104, 87, 54 },
		{ 104, 88, 54 },
		{ 104, 89, 54 },
		{ 104, 90, 54 },
		{ 104, 91, 54 },
		{ 104, 92, 54 },
		{ 104, 93, 54 },
		{ 104, 94, 54 },
		{ 104, 95, 54 },
		{ 104, 96, 54 },
		{ 104, 97, 54 },
		{ 104, 98, 54 },
		{ 104, 99, 54 },
		{ 104, 100, 54 },
		{ 104, 101, 54 },
		{ 104, 102, 54 },
		{ 104, 103, 54 },
		{ 104, 104, 54 },
		{ 104, 105, 54 },
		{ 104, 106, 54 },
		{ 104, 107, 54 },
		{ 104, 108, 54 },
		{ 104, 109, 54 },
		{ 104, 110, 54 },
		{ 104, 111, 54 },
		{ 104, 112, 54 },
		{ 104, 113, 54 },
		{ 104, 114, 54 },
		{ 104, 115, 54 },
		{ 104, 116, 54 },
		{ 104, 117, 54 },
		{ 104, 118, 54 },
		{ 104, 119, 54 },
		{ 104, 120, 54 },
		{ 104, 121, 54 },
		{ 104, 122, 54 },
		{ 104, 123, 54 },
		{ 105, 84, 54 },
		{ 105, 85, 54 },
		{ 105, 86, 54 },
		{ 105, 87, 54 },
		{ 105, 88, 54 },
		{ 105, 89, 54 },
		{ 105, 90, 54 },
		{ 105, 91, 54 },
		{ 105, 92, 54 },
		{ 105, 93, 54 },
		{ 105, 94, 54 },
		{ 105, 95, 54 },
		{ 105, 96, 54 },
		{ 105, 97, 54 },
		{ 105, 98, 54 },
		{ 105, 99, 54 },
		{ 105, 100, 54 },
		{ 105, 101, 54 },
		{ 105, 102, 54 },
		{ 105, 103, 54 },
		{ 105, 104, 54 },
		{ 105, 105, 54 },
		{ 105, 106, 54 },
		{ 105, 107, 54 },
		{ 105, 108, 54 },
		{ 105, 109, 54 },
		{ 105, 110, 54 },
		{ 105, 111, 54 },
		{ 105, 112, 54 },
		{ 105, 113, 54 },
		{ 105, 114, 54 },
		{ 105, 115, 54 },
		{ 105, 116, 54 },
		{ 105, 117, 54 },
		{ 105, 118, 54 },
		{ 105, 119, 54 },
		{ 105, 120, 54 },
		{ 105, 121, 54 },
		{ 105, 122, 54 },
		{ 105, 123, 54 },
		{ 106, 85, 54 },
		{ 106, 86, 54 },
		{ 106, 87, 54 },
		{ 106, 88, 54 },
		{ 106, 89, 54 },
		{ 106, 90, 54 },
		{ 106, 91, 54 },
		{ 106, 92, 54 },
		{ 106, 93, 54 },
		{ 106, 94, 54 },
		{ 106, 95, 54 },
		{ 106, 96, 54 },
		{ 106, 97, 54 },
		{ 106, 98, 54 },
		{ 106, 99, 54 },
		{ 106, 100, 54 },
		{ 106, 101, 54 },
		{ 106, 102, 54 },
		{ 106, 103, 54 },
		{ 106, 104, 54 },
		{ 106, 105, 54 },
		{ 106, 106, 54 },
		{ 106, 107, 54 },
		{ 106, 108, 54 },
		{ 106, 109, 54 },
		{ 106, 110, 54 },
		{ 106, 111, 54 },
		{ 106, 112, 54 },
		{ 106, 113, 54 },
		{ 106, 114, 54 },
		{ 106, 115, 54 },
		{ 106, 116, 54 },
		{ 106, 117, 54 },
		{ 106, 118, 54 },
		{ 106, 119, 54 },
		{ 106, 120, 54 },
		{ 106, 121, 54 },
		{ 106, 122, 54 },
		{ 106, 123, 54 },
		{ 106, 124, 54 },
		{ 107, 86, 54 },
		{ 107, 87, 54 },
		{ 107, 88, 54 },
		{ 107, 89, 54 },
		{ 107, 90, 54 },
		{ 107, 91, 54 },
		{ 107, 92, 54 },
		{ 107, 93, 54 },
		{ 107, 94, 54 },
		{ 107, 95, 54 },
		{ 107, 96, 54 },
		{ 107, 97, 54 },
		{ 107, 98, 54 },
		{ 107, 99, 54 },
		{ 107, 100, 54 },
		{ 107, 101, 54 },
		{ 107, 102, 54 },
		{ 107, 103, 54 },
		{ 107, 104, 54 },
		{ 107, 105, 54 },
		{ 107, 106, 54 },
		{ 107, 107, 54 },
		{ 107, 108, 54 },
		{ 107, 109, 54 },
		{ 107, 110, 54 },
		{ 107, 111, 54 },
		{ 107, 112, 54 },
		{ 107, 113, 54 },
		{ 107, 114, 54 },
		{ 107, 115, 54 },
		{ 107, 116, 54 },
		{ 107, 117, 54 },
		{ 107, 118, 54 },
		{ 107, 119, 54 },
		{ 107, 120, 54 },
		{ 107, 121, 54 },
		{ 107, 122, 54 },
		{ 107, 123, 54 },
		{ 107, 124, 54 },
		{ 107, 125, 54 },
		{ 108, 87, 54 },
		{ 108, 88, 54 },
		{ 108, 89, 54 },
		{ 108, 90, 54 },
		{ 108, 91, 54 },
		{ 108, 92, 54 },
		{ 108, 93, 54 },
		{ 108, 94, 54 },
		{ 108, 95, 54 },
		{ 108, 96, 54 },
		{ 108, 97, 54 },
		{ 108, 98, 54 },
		{ 108, 99, 54 },
		{ 108, 100, 54 },
		{ 108, 101, 54 },
		{ 108, 102, 54 },
		{ 108, 103, 54 },
		{ 108, 104, 54 },
		{ 108, 105, 54 },
		{ 108, 106, 54 },
		{ 108, 107, 54 },
		{ 108, 108, 54 },
		{ 108, 109, 54 },
		{ 108, 110, 54 },
		{ 108, 111, 54 },
		{ 108, 112, 54 },
		{ 108, 113, 54 },
		{ 108, 114, 54 },
		{ 108, 115, 54 },
		{ 108, 116, 54 },
		{ 108, 117, 54 },
		{ 108, 118, 54 },
		{ 108, 119, 54 },
		{ 108, 120, 54 },
		{ 108, 121, 54 },
		{ 108, 122, 54 },
		{ 108, 123, 54 },
		{ 108, 124, 54 },
		{ 108, 125, 54 },
		{ 108, 126, 54 },
		{ 109, 88, 54 },
		{ 109, 89, 54 },
		{ 109, 90, 54 },
		{ 109, 91, 54 },
		{ 109, 92, 54 },
		{ 109, 93, 54 },
		{ 109, 94, 54 },
		{ 109, 95, 54 },
		{ 109, 96, 54 },
		{ 109, 97, 54 },
		{ 109, 98, 54 },
		{ 109, 99, 54 },
		{ 109, 100, 54 },
		{ 109, 101, 54 },
		{ 109, 102, 54 },
		{ 109, 103, 54 },
		{ 109, 104, 54 },
		{ 109, 105, 54 },
		{ 109, 106, 54 },
		{ 109, 107, 54 },
		{ 109, 108, 54 },
		{ 109, 109, 54 },
		{ 109, 110, 54 },
		{ 109, 111, 54 },
		{ 109, 112, 54 },
		{ 109, 113, 54 },
		{ 109, 114, 54 },
		{ 109, 115, 54 },
		{ 109, 116, 54 },
		{ 109, 117, 54 },
		{ 109, 118, 54 },
		{ 109, 119, 54 },
		{ 109, 120, 54 },
		{ 109, 121, 54 },
		{ 109, 122, 54 },
		{ 109, 123, 54 },
		{ 109, 124, 54 },
		{ 109, 125, 54 },
		{ 109, 126, 54 },
		{ 109, 127, 54 },
		{ 110, 90, 54 },
		{ 110, 91, 54 },
		{ 110, 92, 54 },
		{ 110, 93, 54 },
		{ 110, 94, 54 },
		{ 110, 95, 54 },
		{ 110, 96, 54 },
		{ 110, 97, 54 },
		{ 110, 98, 54 },
		{ 110, 99, 54 },
		{ 110, 100, 54 },
		{ 110, 101, 54 },
		{ 110, 102, 54 },
		{ 110, 103, 54 },
		{ 110, 104, 54 },
		{ 110, 105, 54 },
		{ 110, 106, 54 },
		{ 110, 107, 54 },
		{ 110, 108, 54 },
		{ 110, 109, 54 },
		{ 110, 110, 54 },
		{ 110, 111, 54 },
		{ 110, 112, 54 },
		{ 110, 113, 54 },
		{ 110, 114, 54 },
		{ 110, 115, 54 },
		{ 110, 116, 54 },
		{ 110, 117, 54 },
		{ 110, 118, 54 },
		{ 110, 119, 54 },
		{ 110, 120, 54 },
		{ 110, 121, 54 },
		{ 110, 122, 54 },
		{ 110, 123, 54 },
		{ 110, 124, 54 },
		{ 110, 125, 54 },
		{ 110, 126, 54 },
		{ 110, 127, 54 },
		{ 110, 128, 54 },
		{ 111, 91, 54 },
		{ 111, 92, 54 },
		{ 111, 93, 54 },
		{ 111, 94, 54 },
		{ 111, 95, 54 },
		{ 111, 96, 54 },
		{ 111, 97, 54 },
		{ 111, 98, 54 },
		{ 111, 99, 54 },
		{ 111, 100, 54 },
		{ 111, 101, 54 },
		{ 111, 102, 54 },
		{ 111, 103, 54 },
		{ 111, 104, 54 },
		{ 111, 105, 54 },
		{ 111, 106, 54 },
		{ 111, 107, 54 },
		{ 111, 108, 54 },
		{ 111, 109, 54 },
		{ 111, 110, 54 },
		{ 111, 111, 54 },
		{ 111, 112, 54 },
		{ 111, 113, 54 },
		{ 111, 114, 54 },
		{ 111, 115, 54 },
		{ 111, 116, 54 },
		{ 111, 117, 54 },
		{ 111, 118, 54 },
		{ 111, 119, 54 },
		{ 111, 120, 54 },
		{ 111, 121, 54 },
		{ 111, 122, 54 },
		{ 111, 123, 54 },
		{ 111, 124, 54 },
		{ 111, 125, 54 },
		{ 111, 126, 54 },
		{ 111, 127, 54 },
		{ 111, 128, 54 },
		{ 111, 129, 54 },
		{ 112, 92, 54 },
		{ 112, 93, 54 },
		{ 112, 94, 54 },
		{ 112, 95, 54 },
		{ 112, 96, 54 },
		{ 112, 97, 54 },
		{ 112, 98, 54 },
		{ 112, 99, 54 },
		{ 112, 100, 54 },
		{ 112, 101, 54 },
		{ 112, 102, 54 },
		{ 112, 103, 54 },
		{ 112, 104, 54 },
		{ 112, 105, 54 },
		{ 112, 106, 54 },
		{ 112, 107, 54 },
		{ 112, 108, 54 },
		{ 112, 109, 54 },
		{ 112, 110, 54 },
		{ 112, 111, 54 },
		{ 112, 112, 54 },
		{ 112, 113, 54 },
		{ 112, 114, 54 },
		{ 112, 115, 54 },
		{ 112, 116, 54 },
		{ 112, 117, 54 },
		{ 112, 118, 54 },
		{ 112, 119, 54 },
		{ 112, 120, 54 },
		{ 112, 121, 54 },
		{ 112, 122, 54 },
		{ 112, 123, 54 },
		{ 112, 124, 54 },
		{ 112, 125, 54 },
		{ 112, 126, 54 },
		{ 112, 127, 54 },
		{ 112, 128, 54 },
		{ 112, 129, 54 },
		{ 112, 130, 54 },
		{ 113, 93, 54 },
		{ 113, 94, 54 },
		{ 113, 95, 54 },
		{ 113, 96, 54 },
		{ 113, 97, 54 },
		{ 113, 98, 54 },
		{ 113, 99, 54 },
		{ 113, 100, 54 },
		{ 113, 101, 54 },
		{ 113, 102, 54 },
		{ 113, 103, 54 },
		{ 113, 104, 54 },
		{ 113, 105, 54 },
		{ 113, 106, 54 },
		{ 113, 107, 54 },
		{ 113, 108, 54 },
		{ 113, 109, 54 },
		{ 113, 110, 54 },
		{ 113, 111, 54 },
		{ 113, 112, 54 },
		{ 113, 113, 54 },
		{ 113, 114, 54 },
		{ 113, 115, 54 },
		{ 113, 116, 54 },
		{ 113, 117, 54 },
		{ 113, 118, 54 },
		{ 113, 119, 54 },
		{ 113, 120, 54 },
		{ 113, 121, 54 },
		{ 113, 122, 54 },
		{ 113, 123, 54 },
		{ 113, 124, 54 },
		{ 113, 125, 54 },
		{ 113, 126, 54 },
		{ 113, 127, 54 },
		{ 113, 128, 54 },
		{ 113, 129, 54 },
		{ 113, 130, 54 },
		{ 114, 94, 54 },
		{ 114, 95, 54 },
		{ 114, 96, 54 },
		{ 114, 97, 54 },
		{ 114, 98, 54 },
		{ 114, 99, 54 },
		{ 114, 100, 54 },
		{ 114, 101, 54 },
		{ 114, 102, 54 },
		{ 114, 103, 54 },
		{ 114, 104, 54 },
		{ 114, 105, 54 },
		{ 114, 106, 54 },
		{ 114, 107, 54 },
		{ 114, 108, 54 },
		{ 114, 109, 54 },
		{ 114, 110, 54 },
		{ 114, 111, 54 },
		{ 114, 112, 54 },
		{ 114, 113, 54 },
		{ 114, 114, 54 },
		{ 114, 115, 54 },
		{ 114, 116, 54 },
		{ 114, 117, 54 },
		{ 114, 118, 54 },
		{ 114, 119, 54 },
		{ 114, 120, 54 },
		{ 114, 121, 54 },
		{ 114, 122, 54 },
		{ 114, 123, 54 },
		{ 114, 124, 54 },
		{ 114, 125, 54 },
		{ 114, 126, 54 },
		{ 114, 127, 54 },
		{ 114, 128, 54 },
		{ 114, 129, 54 },
		{ 114, 130, 54 },
		{ 114, 131, 54 },
		{ 115, 95, 54 },
		{ 115, 96, 54 },
		{ 115, 97, 54 },
		{ 115, 98, 54 },
		{ 115, 99, 54 },
		{ 115, 100, 54 },
		{ 115, 101, 54 },
		{ 115, 102, 54 },
		{ 115, 103, 54 },
		{ 115, 104, 54 },
		{ 115, 105, 54 },
		{ 115, 106, 54 },
		{ 115, 107, 54 },
		{ 115, 108, 54 },
		{ 115, 109, 54 },
		{ 115, 110, 54 },
		{ 115, 111, 54 },
		{ 115, 112, 54 },
		{ 115, 113, 54 },
		{ 115, 114, 54 },
		{ 115, 115, 54 },
		{ 115, 116, 54 },
		{ 115, 117, 54 },
		{ 115, 118, 54 },
		{ 115, 119, 54 },
		{ 115, 120, 54 },
		{ 115, 121, 54 },
		{ 115, 122, 54 },
		{ 115, 123, 54 },
		{ 115, 124, 54 },
		{ 115, 125, 54 },
		{ 115, 126, 54 },
		{ 115, 127, 54 },
		{ 115, 128, 54 },
		{ 115, 129, 54 },
		{ 115, 130, 54 },
		{ 115, 131, 54 },
		{ 115, 132, 54 },
		{ 116, 96, 54 },
		{ 116, 97, 54 },
		{ 116, 98, 54 },
		{ 116, 99, 54 },
		{ 116, 100, 54 },
		{ 116, 101, 54 },
		{ 116, 102, 54 },
		{ 116, 103, 54 },
		{ 116, 104, 54 },
		{ 116, 105, 54 },
		{ 116, 106, 54 },
		{ 116, 107, 54 },
		{ 116, 108, 54 },
		{ 116, 109, 54 },
		{ 116, 110, 54 },
		{ 116, 111, 54 },
		{ 116, 112, 54 },
		{ 116, 113, 54 },
		{ 116, 114, 54 },
		{ 116, 115, 54 },
		{ 116, 116, 54 },
		{ 116, 117, 54 },
		{ 116, 118, 54 },
		{ 116, 119, 54 },
		{ 116, 120, 54 },
		{ 116, 121, 54 },
		{ 116, 122, 54 },
		{ 116, 123, 54 },
		{ 116, 124, 54 },
		{ 116, 125, 54 },
		{ 116, 126, 54 },
		{ 116, 127, 54 },
		{ 116, 128, 54 },
		{ 116, 129, 54 },
		{ 116, 130, 54 },
		{ 116, 131, 54 },
		{ 116, 132, 54 },
		{ 116, 133, 54 },
		{ 117, 98, 54 },
		{ 117, 99, 54 },
		{ 117, 100, 54 },
		{ 117, 101, 54 },
		{ 117, 102, 54 },
		{ 117, 103, 54 },
		{ 117, 104, 54 },
		{ 117, 105, 54 },
		{ 117, 106, 54 },
		{ 117, 107, 54 },
		{ 117, 108, 54 },
		{ 117, 109, 54 },
		{ 117, 110, 54 },
		{ 117, 111, 54 },
		{ 117, 112, 54 },
		{ 117, 113, 54 },
		{ 117, 114, 54 },
		{ 117, 115, 54 },
		{ 117, 116, 54 },
		{ 117, 117, 54 },
		{ 117, 118, 54 },
		{ 117, 119, 54 },
		{ 117, 120, 54 },
		{ 117, 121, 54 },
		{ 117, 122, 54 },
		{ 117, 123, 54 },
		{ 117, 124, 54 },
		{ 117, 125, 54 },
		{ 117, 126, 54 },
		{ 117, 127, 54 },
		{ 117, 128, 54 },
		{ 117, 129, 54 },
		{ 117, 130, 54 },
		{ 117, 131, 54 },
		{ 117, 132, 54 },
		{ 117, 133, 54 },
		{ 117, 134, 54 },
		{ 118, 99, 54 },
		{ 118, 100, 54 },
		{ 118, 101, 54 },
		{ 118, 102, 54 },
		{ 118, 103, 54 },
		{ 118, 104, 54 },
		{ 118, 105, 54 },
		{ 118, 106, 54 },
		{ 118, 107, 54 },
		{ 118, 108, 54 },
		{ 118, 109, 54 },
		{ 118, 110, 54 },
		{ 118, 111, 54 },
		{ 118, 112, 54 },
		{ 118, 113, 54 },
		{ 118, 114, 54 },
		{ 118, 115, 54 },
		{ 118, 116, 54 },
		{ 118, 117, 54 },
		{ 118, 118, 54 },
		{ 118, 119, 54 },
		{ 118, 120, 54 },
		{ 118, 121, 54 },
		{ 118, 122, 54 },
		{ 118, 123, 54 },
		{ 118, 124, 54 },
		{ 118, 125, 54 },
		{ 118, 126, 54 },
		{ 118, 127, 54 },
		{ 118, 128, 54 },
		{ 118, 129, 54 },
		{ 118, 130, 54 },
		{ 118, 131, 54 },
		{ 118, 132, 54 },
		{ 118, 133, 54 },
		{ 118, 134, 54 },
		{ 118, 135, 54 },
		{ 119, 100, 54 },
		{ 119, 101, 54 },
		{ 119, 102, 54 },
		{ 119, 103, 54 },
		{ 119, 104, 54 },
		{ 119, 105, 54 },
		{ 119, 106, 54 },
		{ 119, 107, 54 },
		{ 119, 108, 54 },
		{ 119, 109, 54 },
		{ 119, 110, 54 },
		{ 119, 111, 54 },
		{ 119, 112, 54 },
		{ 119, 113, 54 },
		{ 119, 114, 54 },
		{ 119, 115, 54 },
		{ 119, 116, 54 },
		{ 119, 117, 54 },
		{ 119, 118, 54 },
		{ 119, 119, 54 },
		{ 119, 120, 54 },
		{ 119, 121, 54 },
		{ 119, 122, 54 },
		{ 119, 123, 54 },
		{ 119, 124, 54 },
		{ 119, 125, 54 },
		{ 119, 126, 54 },
		{ 119, 127, 54 },
		{ 119, 128, 54 },
		{ 119, 129, 54 },
		{ 119, 130, 54 },
		{ 119, 131, 54 },
		{ 119, 132, 54 },
		{ 119, 133, 54 },
		{ 119, 134, 54 },
		{ 119, 135, 54 },
		{ 119, 136, 54 },
		{ 120, 101, 54 },
		{ 120, 102, 54 },
		{ 120, 103, 54 },
		{ 120, 104, 54 },
		{ 120, 105, 54 },
		{ 120, 106, 54 },
		{ 120, 107, 54 },
		{ 120, 108, 54 },
		{ 120, 109, 54 },
		{ 120, 110, 54 },
		{ 120, 111, 54 },
		{ 120, 112, 54 },
		{ 120, 113, 54 },
		{ 120, 114, 54 },
		{ 120, 115, 54 },
		{ 120, 116, 54 },
		{ 120, 117, 54 },
		{ 120, 118, 54 },
		{ 120, 119, 54 },
		{ 120, 120, 54 },
		{ 120, 121, 54 },
		{ 120, 122, 54 },
		{ 120, 123, 54 },
		{ 120, 124, 54 },
		{ 120, 125, 54 },
		{ 120, 126, 54 },
		{ 120, 127, 54 },
		{ 120, 128, 54 },
		{ 120, 129, 54 },
		{ 120, 130, 54 },
		{ 120, 131, 54 },
		{ 120, 132, 54 },
		{ 120, 133, 54 },
		{ 120, 134, 54 },
		{ 120, 135, 54 },
		{ 120, 136, 54 },
		{ 120, 137, 54 },
		{ 121, 102, 54 },
		{ 121, 103, 54 },
		{ 121, 104, 54 },
		{ 121, 105, 54 },
		{ 121, 106, 54 },
		{ 121, 107, 54 },
		{ 121, 108, 54 },
		{ 121, 109, 54 },
		{ 121, 110, 54 },
		{ 121, 111, 54 },
		{ 121, 112, 54 },
		{ 121, 113, 54 },
		{ 121, 114, 54 },
		{ 121, 115, 54 },
		{ 121, 116, 54 },
		{ 121, 117, 54 },
		{ 121, 118, 54 },
		{ 121, 119, 54 },
		{ 121, 120, 54 },
		{ 121, 121, 54 },
		{ 121, 122, 54 },
		{ 121, 123, 54 },
		{ 121, 124, 54 },
		{ 121, 125, 54 },
		{ 121, 126, 54 },
		{ 121, 127, 54 },
		{ 121, 128, 54 },
		{ 121, 129, 54 },
		{ 121, 130, 54 },
		{ 121, 131, 54 },
		{ 121, 132, 54 },
		{ 121, 133, 54 },
		{ 121, 134, 54 },
		{ 121, 135, 54 },
		{ 121, 136, 54 },
		{ 121, 137, 54 },
		{ 122, 103, 54 },
		{ 122, 104, 54 },
		{ 122, 105, 54 },
		{ 122, 106, 54 },
		{ 122, 107, 54 },
		{ 122, 108, 54 },
		{ 122, 109, 54 },
		{ 122, 110, 54 },
		{ 122, 111, 54 },
		{ 122, 112, 54 },
		{ 122, 113, 54 },
		{ 122, 114, 54 },
		{ 122, 115, 54 },
		{ 122, 116, 54 },
		{ 122, 117, 54 },
		{ 122, 118, 54 },
		{ 122, 119, 54 },
		{ 122, 120, 54 },
		{ 122, 121, 54 },
		{ 122, 122, 54 },
		{ 122, 123, 54 },
		{ 122, 124, 54 },
		{ 122, 125, 54 },
		{ 122, 126, 54 },
		{ 122, 127, 54 },
		{ 122, 128, 54 },
		{ 122, 129, 54 },
		{ 122, 130, 54 },
		{ 122, 131, 54 },
		{ 122, 132, 54 },
		{ 122, 133, 54 },
		{ 122, 134, 54 },
		{ 122, 135, 54 },
		{ 122, 136, 54 },
		{ 122, 137, 54 },
		{ 122, 138, 54 },
		{ 123, 104, 54 },
		{ 123, 105, 54 },
		{ 123, 106, 54 },
		{ 123, 107, 54 },
		{ 123, 108, 54 },
		{ 123, 109, 54 },
		{ 123, 110, 54 },
		{ 123, 111, 54 },
		{ 123, 112, 54 },
		{ 123, 113, 54 },
		{ 123, 114, 54 },
		{ 123, 115, 54 },
		{ 123, 116, 54 },
		{ 123, 117, 54 },
		{ 123, 118, 54 },
		{ 123, 119, 54 },
		{ 123, 120, 54 },
		{ 123, 121, 54 },
		{ 123, 122, 54 },
		{ 123, 123, 54 },
		{ 123, 124, 54 },
		{ 123, 125, 54 },
		{ 123, 126, 54 },
		{ 123, 127, 54 },
		{ 123, 128, 54 },
		{ 123, 129, 54 },
		{ 123, 130, 54 },
		{ 123, 131, 54 },
		{ 123, 132, 54 },
		{ 123, 133, 54 },
		{ 123, 134, 54 },
		{ 123, 135, 54 },
		{ 123, 136, 54 },
		{ 123, 137, 54 },
		{ 123, 138, 54 },
		{ 123, 139, 54 },
		{ 124, 106, 54 },
		{ 124, 107, 54 },
		{ 124, 108, 54 },
		{ 124, 109, 54 },
		{ 124, 110, 54 },
		{ 124, 111, 54 },
		{ 124, 112, 54 },
		{ 124, 113, 54 },
		{ 124, 114, 54 },
		{ 124, 115, 54 },
		{ 124, 116, 54 },
		{ 124, 117, 54 },
		{ 124, 118, 54 },
		{ 124, 119, 54 },
		{ 124, 120, 54 },
		{ 124, 121, 54 },
		{ 124, 122, 54 },
		{ 124, 123, 54 },
		{ 124, 124, 54 },
		{ 124, 125, 54 },
		{ 124, 126, 54 },
		{ 124, 127, 54 },
		{ 124, 128, 54 },
		{ 124, 129, 54 },
		{ 124, 130, 54 },
		{ 124, 131, 54 },
		{ 124, 132, 54 },
		{ 124, 133, 54 },
		{ 124, 134, 54 },
		{ 124, 135, 54 },
		{ 124, 136, 54 },
		{ 124, 137, 54 },
		{ 124, 138, 54 },
		{ 124, 139, 54 },
		{ 124, 140, 54 },
		{ 125, 107, 54 },
		{ 125, 108, 54 },
		{ 125, 109, 54 },
		{ 125, 110, 54 },
		{ 125, 111, 54 },
		{ 125, 112, 54 },
		{ 125, 113, 54 },
		{ 125, 114, 54 },
		{ 125, 115, 54 },
		{ 125, 116, 54 },
		{ 125, 117, 54 },
		{ 125, 118, 54 },
		{ 125, 119, 54 },
		{ 125, 120, 54 },
		{ 125, 121, 54 },
		{ 125, 122, 54 },
		{ 125, 123, 54 },
		{ 125, 124, 54 },
		{ 125, 125, 54 },
		{ 125, 126, 54 },
		{ 125, 127, 54 },
		{ 125, 128, 54 },
		{ 125, 129, 54 },
		{ 125, 130, 54 },
		{ 125, 131, 54 },
		{ 125, 132, 54 },
		{ 125, 133, 54 },
		{ 125, 134, 54 },
		{ 125, 135, 54 },
		{ 125, 136, 54 },
		{ 125, 137, 54 },
		{ 125, 138, 54 },
		{ 125, 139, 54 },
		{ 125, 140, 54 },
		{ 125, 141, 54 },
		{ 126, 108, 54 },
		{ 126, 109, 54 },
		{ 126, 110, 54 },
		{ 126, 111, 54 },
		{ 126, 112, 54 },
		{ 126, 113, 54 },
		{ 126, 114, 54 },
		{ 126, 115, 54 },
		{ 126, 116, 54 },
		{ 126, 117, 54 },
		{ 126, 118, 54 },
		{ 126, 119, 54 },
		{ 126, 120, 54 },
		{ 126, 121, 54 },
		{ 126, 122, 54 },
		{ 126, 123, 54 },
		{ 126, 124, 54 },
		{ 126, 125, 54 },
		{ 126, 126, 54 },
		{ 126, 127, 54 },
		{ 126, 128, 54 },
		{ 126, 129, 54 },
		{ 126, 130, 54 },
		{ 126, 131, 54 },
		{ 126, 132, 54 },
		{ 126, 133, 54 },
		{ 126, 134, 54 },
		{ 126, 135, 54 },
		{ 126, 136, 54 },
		{ 126, 137, 54 },
		{ 126, 138, 54 },
		{ 126, 139, 54 },
		{ 126, 140, 54 },
		{ 126, 141, 54 },
		{ 126, 142, 54 },
		{ 127, 109, 54 },
		{ 127, 110, 54 },
		{ 127, 111, 54 },
		{ 127, 112, 54 },
		{ 127, 113, 54 },
		{ 127, 114, 54 },
		{ 127, 115, 54 },
		{ 127, 116, 54 },
		{ 127, 117, 54 },
		{ 127, 118, 54 },
		{ 127, 119, 54 },
		{ 127, 120, 54 },
		{ 127, 121, 54 },
		{ 127, 122, 54 },
		{ 127, 123, 54 },
		{ 127, 124, 54 },
		{ 127, 125, 54 },
		{ 127, 126, 54 },
		{ 127, 127, 54 },
		{ 127, 128, 54 },
		{ 127, 129, 54 },
		{ 127, 130, 54 },
		{ 127, 131, 54 },
		{ 127, 132, 54 },
		{ 127, 133, 54 },
		{ 127, 134, 54 },
		{ 127, 135, 54 },
		{ 127, 136, 54 },
		{ 127, 137, 54 },
		{ 127, 138, 54 },
		{ 127, 139, 54 },
		{ 127, 140, 54 },
		{ 127, 141, 54 },
		{ 127, 142, 54 },
		{ 127, 143, 54 },
		{ 128, 110, 54 },
		{ 128, 111, 54 },
		{ 128, 112, 54 },
		{ 128, 113, 54 },
		{ 128, 114, 54 },
		{ 128, 115, 54 },
		{ 128, 116, 54 },
		{ 128, 117, 54 },
		{ 128, 118, 54 },
		{ 128, 119, 54 },
		{ 128, 120, 54 },
		{ 128, 121, 54 },
		{ 128, 122, 54 },
		{ 128, 123, 54 },
		{ 128, 124, 54 },
		{ 128, 125, 54 },
		{ 128, 126, 54 },
		{ 128, 127, 54 },
		{ 128, 128, 54 },
		{ 128, 129, 54 },
		{ 128, 130, 54 },
		{ 128, 131, 54 },
		{ 128, 132, 54 },
		{ 128, 133, 54 },
		{ 128, 134, 54 },
		{ 128, 135, 54 },
		{ 128, 136, 54 },
		{ 128, 137, 54 },
		{ 128, 138, 54 },
		{ 128, 139, 54 },
		{ 128, 140, 54 },
		{ 128, 141, 54 },
		{ 128, 142, 54 },
		{ 128, 143, 54 },
		{ 128, 144, 54 },
		{ 129, 111, 54 },
		{ 129, 112, 54 },
		{ 129, 113, 54 },
		{ 129, 114, 54 },
		{ 129, 115, 54 },
		{ 129, 116, 54 },
		{ 129, 117, 54 },
		{ 129, 118, 54 },
		{ 129, 119, 54 },
		{ 129, 120, 54 },
		{ 129, 121, 54 },
		{ 129, 122, 54 },
		{ 129, 123, 54 },
		{ 129, 124, 54 },
		{ 129, 125, 54 },
		{ 129, 126, 54 },
		{ 129, 127, 54 },
		{ 129, 128, 54 },
		{ 129, 129, 54 },
		{ 129, 130, 54 },
		{ 129, 131, 54 },
		{ 129, 132, 54 },
		{ 129, 133, 54 },
		{ 129, 134, 54 },
		{ 129, 135, 54 },
		{ 129, 136, 54 },
		{ 129, 137, 54 },
		{ 129, 138, 54 },
		{ 129, 139, 54 },
		{ 129, 140, 54 },
		{ 129, 141, 54 },
		{ 129, 142, 54 },
		{ 129, 143, 54 },
		{ 129, 144, 54 },
		{ 130, 112, 54 },
		{ 130, 113, 54 },
		{ 130, 114, 54 },
		{ 130, 115, 54 },
		{ 130, 116, 54 },
		{ 130, 117, 54 },
		{ 130, 118, 54 },
		{ 130, 119, 54 },
		{ 130, 120, 54 },
		{ 130, 121, 54 },
		{ 130, 122, 54 },
		{ 130, 123, 54 },
		{ 130, 124, 54 },
		{ 130, 125, 54 },
		{ 130, 126, 54 },
		{ 130, 127, 54 },
		{ 130, 128, 54 },
		{ 130, 129, 54 },
		{ 130, 130, 54 },
		{ 130, 131, 54 },
		{ 130, 132, 54 },
		{ 130, 133, 54 },
		{ 130, 134, 54 },
		{ 130, 135, 54 },
		{ 130, 136, 54 },
		{ 130, 137, 54 },
		{ 130, 138, 54 },
		{ 130, 139, 54 },
		{ 130, 140, 54 },
		{ 130, 141, 54 },
		{ 130, 142, 54 },
		{ 130, 143, 54 },
		{ 130, 144, 54 },
		{ 130, 145, 54 },
		{ 131, 114, 54 },
		{ 131, 115, 54 },
		{ 131, 116, 54 },
		{ 131, 117, 54 },
		{ 131, 118, 54 },
		{ 131, 119, 54 },
		{ 131, 120, 54 },
		{ 131, 121, 54 },
		{ 131, 122, 54 },
		{ 131, 123, 54 },
		{ 131, 124, 54 },
		{ 131, 125, 54 },
		{ 131, 126, 54 },
		{ 131, 127, 54 },
		{ 131, 128, 54 },
		{ 131, 129, 54 },
		{ 131, 130, 54 },
		{ 131, 131, 54 },
		{ 131, 132, 54 },
		{ 131, 133, 54 },
		{ 131, 134, 54 },
		{ 131, 135, 54 },
		{ 131, 136, 54 },
		{ 131, 137, 54 },
		{ 131, 138, 54 },
		{ 131, 139, 54 },
		{ 131, 140, 54 },
		{ 131, 141, 54 },
		{ 131, 142, 54 },
		{ 131, 143, 54 },
		{ 131, 144, 54 },
		{ 131, 145, 54 },
		{ 131, 146, 54 },
		{ 132, 115, 54 },
		{ 132, 116, 54 },
		{ 132, 117, 54 },
		{ 132, 118, 54 },
		{ 132, 119, 54 },
		{ 132, 120, 54 },
		{ 132, 121, 54 },
		{ 132, 122, 54 },
		{ 132, 123, 54 },
		{ 132, 124, 54 },
		{ 132, 125, 54 },
		{ 132, 126, 54 },
		{ 132, 127, 54 },
		{ 132, 128, 54 },
		{ 132, 129, 54 },
		{ 132, 130, 54 },
		{ 132, 131, 54 },
		{ 132, 132, 54 },
		{ 132, 133, 54 },
		{ 132, 134, 54 },
		{ 132, 135, 54 },
		{ 132, 136, 54 },
		{ 132, 137, 54 },
		{ 132, 138, 54 },
		{ 132, 139, 54 },
		{ 132, 140, 54 },
		{ 132, 141, 54 },
		{ 132, 142, 54 },
		{ 132, 143, 54 },
		{ 132, 144, 54 },
		{ 132, 145, 54 },
		{ 132, 146, 54 },
		{ 132, 147, 54 },
		{ 133, 116, 54 },
		{ 133, 117, 54 },
		{ 133, 118, 54 },
		{ 133, 119, 54 },
		{ 133, 120, 54 },
		{ 133, 121, 54 },
		{ 133, 122, 54 },
		{ 133, 123, 54 },
		{ 133, 124, 54 },
		{ 133, 125, 54 },
		{ 133, 126, 54 },
		{ 133, 127, 54 },
		{ 133, 128, 54 },
		{ 133, 129, 54 },
		{ 133, 130, 54 },
		{ 133, 131, 54 },
		{ 133, 132, 54 },
		{ 133, 133, 54 },
		{ 133, 134, 54 },
		{ 133, 135, 54 },
		{ 133, 136, 54 },
		{ 133, 137, 54 },
		{ 133, 138, 54 },
		{ 133, 139, 54 },
		{ 133, 140, 54 },
		{ 133, 141, 54 },
		{ 133, 142, 54 },
		{ 133, 143, 54 },
		{ 133, 144, 54 },
		{ 133, 145, 54 },
		{ 133, 146, 54 },
		{ 133, 147, 54 },
		{ 133, 148, 54 },
		{ 134, 117, 54 },
		{ 134, 118, 54 },
		{ 134, 119, 54 },
		{ 134, 120, 54 },
		{ 134, 121, 54 },
		{ 134, 122, 54 },
		{ 134, 123, 54 },
		{ 134, 124, 54 },
		{ 134, 125, 54 },
		{ 134, 126, 54 },
		{ 134, 127, 54 },
		{ 134, 128, 54 },
		{ 134, 129, 54 },
		{ 134, 130, 54 },
		{ 134, 131, 54 },
		{ 134, 132, 54 },
		{ 134, 133, 54 },
		{ 134, 134, 54 },
		{ 134, 135, 54 },
		{ 134, 136, 54 },
		{ 134, 137, 54 },
		{ 134, 138, 54 },
		{ 134, 139, 54 },
		{ 134, 140, 54 },
		{ 134, 141, 54 },
		{ 134, 142, 54 },
		{ 134, 143, 54 },
		{ 134, 144, 54 },
		{ 134, 145, 54 },
		{ 134, 146, 54 },
		{ 134, 147, 54 },
		{ 134, 148, 54 },
		{ 134, 149, 54 },
		{ 135, 118, 54 },
		{ 135, 119, 54 },
		{ 135, 120, 54 },
		{ 135, 121, 54 },
		{ 135, 122, 54 },
		{ 135, 123, 54 },
		{ 135, 124, 54 },
		{ 135, 125, 54 },
		{ 135, 126, 54 },
		{ 135, 127, 54 },
		{ 135, 128, 54 },
		{ 135, 129, 54 },
		{ 135, 130, 54 },
		{ 135, 131, 54 },
		{ 135, 132, 54 },
		{ 135, 133, 54 },
		{ 135, 134, 54 },
		{ 135, 135, 54 },
		{ 135, 136, 54 },
		{ 135, 137, 54 },
		{ 135, 138, 54 },
		{ 135, 139, 54 },
		{ 135, 140, 54 },
		{ 135, 141, 54 },
		{ 135, 142, 54 },
		{ 135, 143, 54 },
		{ 135, 144, 54 },
		{ 135, 145, 54 },
		{ 135, 146, 54 },
		{ 135, 147, 54 },
		{ 135, 148, 54 },
		{ 135, 149, 54 },
		{ 135, 150, 54 },
		{ 136, 119, 54 },
		{ 136, 120, 54 },
		{ 136, 121, 54 },
		{ 136, 122, 54 },
		{ 136, 123, 54 },
		{ 136, 124, 54 },
		{ 136, 125, 54 },
		{ 136, 126, 54 },
		{ 136, 127, 54 },
		{ 136, 128, 54 },
		{ 136, 129, 54 },
		{ 136, 130, 54 },
		{ 136, 131, 54 },
		{ 136, 132, 54 },
		{ 136, 133, 54 },
		{ 136, 134, 54 },
		{ 136, 135, 54 },
		{ 136, 136, 54 },
		{ 136, 137, 54 },
		{ 136, 138, 54 },
		{ 136, 139, 54 },
		{ 136, 140, 54 },
		{ 136, 141, 54 },
		{ 136, 142, 54 },
		{ 136, 143, 54 },
		{ 136, 144, 54 },
		{ 136, 145, 54 },
		{ 136, 146, 54 },
		{ 136, 147, 54 },
		{ 136, 148, 54 },
		{ 136, 149, 54 },
		{ 136, 150, 54 },
		{ 136, 151, 54 },
		{ 137, 120, 54 },
		{ 137, 121, 54 },
		{ 137, 122, 54 },
		{ 137, 123, 54 },
		{ 137, 124, 54 },
		{ 137, 125, 54 },
		{ 137, 126, 54 },
		{ 137, 127, 54 },
		{ 137, 128, 54 },
		{ 137, 129, 54 },
		{ 137, 130, 54 },
		{ 137, 131, 54 },
		{ 137, 132, 54 },
		{ 137, 133, 54 },
		{ 137, 134, 54 },
		{ 137, 135, 54 },
		{ 137, 136, 54 },
		{ 137, 137, 54 },
		{ 137, 138, 54 },
		{ 137, 139, 54 },
		{ 137, 140, 54 },
		{ 137, 141, 54 },
		{ 137, 142, 54 },
		{ 137, 143, 54 },
		{ 137, 144, 54 },
		{ 137, 145, 54 },
		{ 137, 146, 54 },
		{ 137, 147, 54 },
		{ 137, 148, 54 },
		{ 137, 149, 54 },
		{ 137, 150, 54 },
		{ 137, 151, 54 },
		{ 138, 122, 54 },
		{ 138, 123, 54 },
		{ 138, 124, 54 },
		{ 138, 125, 54 },
		{ 138, 126, 54 },
		{ 138, 127, 54 },
		{ 138, 128, 54 },
		{ 138, 129, 54 },
		{ 138, 130, 54 },
		{ 138, 131, 54 },
		{ 138, 132, 54 },
		{ 138, 133, 54 },
		{ 138, 134, 54 },
		{ 138, 135, 54 },
		{ 138, 136, 54 },
		{ 138, 137, 54 },
		{ 138, 138, 54 },
		{ 138, 139, 54 },
		{ 138, 140, 54 },
		{ 138, 141, 54 },
		{ 138, 142, 54 },
		{ 138, 143, 54 },
		{ 138, 144, 54 },
		{ 138, 145, 54 },
		{ 138, 146, 54 },
		{ 138, 147, 54 },
		{ 138, 148, 54 },
		{ 138, 149, 54 },
		{ 138, 150, 54 },
		{ 138, 151, 54 },
		{ 138, 152, 54 },
		{ 139, 123, 54 },
		{ 139, 124, 54 },
		{ 139, 125, 54 },
		{ 139, 126, 54 },
		{ 139, 127, 54 },
		{ 139, 128, 54 },
		{ 139, 129, 54 },
		{ 139, 130, 54 },
		{ 139, 131, 54 },
		{ 139, 132, 54 },
		{ 139, 133, 54 },
		{ 139, 134, 54 },
		{ 139, 135, 54 },
		{ 139, 136, 54 },
		{ 139, 137, 54 },
		{ 139, 138, 54 },
		{ 139, 139, 54 },
		{ 139, 140, 54 },
		{ 139, 141, 54 },
		{ 139, 142, 54 },
		{ 139, 143, 54 },
		{ 139, 144, 54 },
		{ 139, 145, 54 },
		{ 139, 146, 54 },
		{ 139, 147, 54 },
		{ 139, 148, 54 },
		{ 139, 149, 54 },
		{ 139, 150, 54 },
		{ 139, 151, 54 },
		{ 139, 152, 54 },
		{ 139, 153, 54 },
		{ 140, 124, 54 },
		{ 140, 125, 54 },
		{ 140, 126, 54 },
		{ 140, 127, 54 },
		{ 140, 128, 54 },
		{ 140, 129, 54 },
		{ 140, 130, 54 },
		{ 140, 131, 54 },
		{ 140, 132, 54 },
		{ 140, 133, 54 },
		{ 140, 134, 54 },
		{ 140, 135, 54 },
		{ 140, 136, 54 },
		{ 140, 137, 54 },
		{ 140, 138, 54 },
		{ 140, 139, 54 },
		{ 140, 140, 54 },
		{ 140, 141, 54 },
		{ 140, 142, 54 },
		{ 140, 143, 54 },
		{ 140, 144, 54 },
		{ 140, 145, 54 },
		{ 140, 146, 54 },
		{ 140, 147, 54 },
		{ 140, 148, 54 },
		{ 140, 149, 54 },
		{ 140, 150, 54 },
		{ 140, 151, 54 },
		{ 140, 152, 54 },
		{ 140, 153, 54 },
		{ 140, 154, 54 },
		{ 141, 125, 54 },
		{ 141, 126, 54 },
		{ 141, 127, 54 },
		{ 141, 128, 54 },
		{ 141, 129, 54 },
		{ 141, 130, 54 },
		{ 141, 131, 54 },
		{ 141, 132, 54 },
		{ 141, 133, 54 },
		{ 141, 134, 54 },
		{ 141, 135, 54 },
		{ 141, 136, 54 },
		{ 141, 137, 54 },
		{ 141, 138, 54 },
		{ 141, 139, 54 },
		{ 141, 140, 54 },
		{ 141, 141, 54 },
		{ 141, 142, 54 },
		{ 141, 143, 54 },
		{ 141, 144, 54 },
		{ 141, 145, 54 },
		{ 141, 146, 54 },
		{ 141, 147, 54 },
		{ 141, 148, 54 },
		{ 141, 149, 54 },
		{ 141, 150, 54 },
		{ 141, 151, 54 },
		{ 141, 152, 54 },
		{ 141, 153, 54 },
		{ 141, 154, 54 },
		{ 141, 155, 54 },
		{ 142, 126, 54 },
		{ 142, 127, 54 },
		{ 142, 128, 54 },
		{ 142, 129, 54 },
		{ 142, 130, 54 },
		{ 142, 131, 54 },
		{ 142, 132, 54 },
		{ 142, 133, 54 },
		{ 142, 134, 54 },
		{ 142, 135, 54 },
		{ 142, 136, 54 },
		{ 142, 137, 54 },
		{ 142, 138, 54 },
		{ 142, 139, 54 },
		{ 142, 140, 54 },
		{ 142, 141, 54 },
		{ 142, 142, 54 },
		{ 142, 143, 54 },
		{ 142, 144, 54 },
		{ 142, 145, 54 },
		{ 142, 146, 54 },
		{ 142, 147, 54 },
		{ 142, 148, 54 },
		{ 142, 149, 54 },
		{ 142, 150, 54 },
		{ 142, 151, 54 },
		{ 142, 152, 54 },
		{ 142, 153, 54 },
		{ 142, 154, 54 },
		{ 142, 155, 54 },
		{ 142, 156, 54 },
		{ 143, 127, 54 },
		{ 143, 128, 54 },
		{ 143, 129, 54 },
		{ 143, 130, 54 },
		{ 143, 131, 54 },
		{ 143, 132, 54 },
		{ 143, 133, 54 },
		{ 143, 134, 54 },
		{ 143, 135, 54 },
		{ 143, 136, 54 },
		{ 143, 137, 54 },
		{ 143, 138, 54 },
		{ 143, 139, 54 },
		{ 143, 140, 54 },
		{ 143, 141, 54 },
		{ 143, 142, 54 },
		{ 143, 143, 54 },
		{ 143, 144, 54 },
		{ 143, 145, 54 },
		{ 143, 146, 54 },
		{ 143, 147, 54 },
		{ 143, 148, 54 },
		{ 143, 149, 54 },
		{ 143, 150, 54 },
		{ 143, 151, 54 },
		{ 143, 152, 54 },
		{ 143, 153, 54 },
		{ 143, 154, 54 },
		{ 143, 155, 54 },
		{ 143, 156, 54 },
		{ 143, 157, 54 },
		{ 144, 128, 54 },
		{ 144, 129, 54 },
		{ 144, 130, 54 },
		{ 144, 131, 54 },
		{ 144, 132, 54 },
		{ 144, 133, 54 },
		{ 144, 134, 54 },
		{ 144, 135, 54 },
		{ 144, 136, 54 },
		{ 144, 137, 54 },
		{ 144, 138, 54 },
		{ 144, 139, 54 },
		{ 144, 140, 54 },
		{ 144, 141, 54 },
		{ 144, 142, 54 },
		{ 144, 143, 54 },
		{ 144, 144, 54 },
		{ 144, 145, 54 },
		{ 144, 146, 54 },
		{ 144, 147, 54 },
		{ 144, 148, 54 },
		{ 144, 149, 54 },
		{ 144, 150, 54 },
		{ 144, 151, 54 },
		{ 144, 152, 54 },
		{ 144, 153, 54 },
		{ 144, 154, 54 },
		{ 144, 155, 54 },
		{ 144, 156, 54 },
		{ 144, 157, 54 },
		{ 144, 158, 54 },
		{ 145, 130, 54 },
		{ 145, 131, 54 },
		{ 145, 132, 54 },
		{ 145, 133, 54 },
		{ 145, 134, 54 },
		{ 145, 135, 54 },
		{ 145, 136, 54 },
		{ 145, 137, 54 },
		{ 145, 138, 54 },
		{ 145, 139, 54 },
		{ 145, 140, 54 },
		{ 145, 141, 54 },
		{ 145, 142, 54 },
		{ 145, 143, 54 },
		{ 145, 144, 54 },
		{ 145, 145, 54 },
		{ 145, 146, 54 },
		{ 145, 147, 54 },
		{ 145, 148, 54 },
		{ 145, 149, 54 },
		{ 145, 150, 54 },
		{ 145, 151, 54 },
		{ 145, 152, 54 },
		{ 145, 153, 54 },
		{ 145, 154, 54 },
		{ 145, 155, 54 },
		{ 145, 156, 54 },
		{ 145, 157, 54 },
		{ 145, 158, 54 },
		{ 146, 131, 54 },
		{ 146, 132, 54 },
		{ 146, 133, 54 },
		{ 146, 134, 54 },
		{ 146, 135, 54 },
		{ 146, 136, 54 },
		{ 146, 137, 54 },
		{ 146, 138, 54 },
		{ 146, 139, 54 },
		{ 146, 140, 54 },
		{ 146, 141, 54 },
		{ 146, 142, 54 },
		{ 146, 143, 54 },
		{ 146, 144, 54 },
		{ 146, 145, 54 },
		{ 146, 146, 54 },
		{ 146, 147, 54 },
		{ 146, 148, 54 },
		{ 146, 149, 54 },
		{ 146, 150, 54 },
		{ 146, 151, 54 },
		{ 146, 152, 54 },
		{ 146, 153, 54 },
		{ 146, 154, 54 },
		{ 146, 155, 54 },
		{ 146, 156, 54 },
		{ 146, 157, 54 },
		{ 146, 158, 54 },
		{ 146, 159, 54 },
		{ 147, 132, 54 },
		{ 147, 133, 54 },
		{ 147, 134, 54 },
		{ 147, 135, 54 },
		{ 147, 136, 54 },
		{ 147, 137, 54 },
		{ 147, 138, 54 },
		{ 147, 139, 54 },
		{ 147, 140, 54 },
		{ 147, 141, 54 },
		{ 147, 142, 54 },
		{ 147, 143, 54 },
		{ 147, 144, 54 },
		{ 147, 145, 54 },
		{ 147, 146, 54 },
		{ 147, 147, 54 },
		{ 147, 148, 54 },
		{ 147, 149, 54 },
		{ 147, 150, 54 },
		{ 147, 151, 54 },
		{ 147, 152, 54 },
		{ 147, 153, 54 },
		{ 147, 154, 54 },
		{ 147, 155, 54 },
		{ 147, 156, 54 },
		{ 147, 157, 54 },
		{ 147, 158, 54 },
		{ 147, 159, 54 },
		{ 147, 160, 54 },
		{ 148, 133, 54 },
		{ 148, 134, 54 },
		{ 148, 135, 54 },
		{ 148, 136, 54 },
		{ 148, 137, 54 },
		{ 148, 138, 54 },
		{ 148, 139, 54 },
		{ 148, 140, 54 },
		{ 148, 141, 54 },
		{ 148, 142, 54 },
		{ 148, 143, 54 },
		{ 148, 144, 54 },
		{ 148, 145, 54 },
		{ 148, 146, 54 },
		{ 148, 147, 54 },
		{ 148, 148, 54 },
		{ 148, 149, 54 },
		{ 148, 150, 54 },
		{ 148, 151, 54 },
		{ 148, 152, 54 },
		{ 148, 153, 54 },
		{ 148, 154, 54 },
		{ 148, 155, 54 },
		{ 148, 156, 54 },
		{ 148, 157, 54 },
		{ 148, 158, 54 },
		{ 148, 159, 54 },
		{ 148, 160, 54 },
		{ 148, 161, 54 },
		{ 149, 134, 54 },
		{ 149, 135, 54 },
		{ 149, 136, 54 },
		{ 149, 137, 54 },
		{ 149, 138, 54 },
		{ 149, 139, 54 },
		{ 149, 140, 54 },
		{ 149, 141, 54 },
		{ 149, 142, 54 },
		{ 149, 143, 54 },
		{ 149, 144, 54 },
		{ 149, 145, 54 },
		{ 149, 146, 54 },
		{ 149, 147, 54 },
		{ 149, 148, 54 },
		{ 149, 149, 54 },
		{ 149, 150, 54 },
		{ 149, 151, 54 },
		{ 149, 152, 54 },
		{ 149, 153, 54 },
		{ 149, 154, 54 },
		{ 149, 155, 54 },
		{ 149, 156, 54 },
		{ 149, 157, 54 },
		{ 149, 158, 54 },
		{ 149, 159, 54 },
		{ 149, 160, 54 },
		{ 149, 161, 54 },
		{ 149, 162, 54 },
		{ 150, 135, 54 },
		{ 150, 136, 54 },
		{ 150, 137, 54 },
		{ 150, 138, 54 },
		{ 150, 139, 54 },
		{ 150, 140, 54 },
		{ 150, 141, 54 },
		{ 150, 142, 54 },
		{ 150, 143, 54 },
		{ 150, 144, 54 },
		{ 150, 145, 54 },
		{ 150, 146, 54 },
		{ 150, 147, 54 },
		{ 150, 148, 54 },
		{ 150, 149, 54 },
		{ 150, 150, 54 },
		{ 150, 151, 54 },
		{ 150, 152, 54 },
		{ 150, 153, 54 },
		{ 150, 154, 54 },
		{ 150, 155, 54 },
		{ 150, 156, 54 },
		{ 150, 157, 54 },
		{ 150, 158, 54 },
		{ 150, 159, 54 },
		{ 150, 160, 54 },
		{ 150, 161, 54 },
		{ 150, 162, 54 },
		{ 150, 163, 54 },
		{ 151, 136, 54 },
		{ 151, 137, 54 },
		{ 151, 138, 54 },
		{ 151, 139, 54 },
		{ 151, 140, 54 },
		{ 151, 141, 54 },
		{ 151, 142, 54 },
		{ 151, 143, 54 },
		{ 151, 144, 54 },
		{ 151, 145, 54 },
		{ 151, 146, 54 },
		{ 151, 147, 54 },
		{ 151, 148, 54 },
		{ 151, 149, 54 },
		{ 151, 150, 54 },
		{ 151, 151, 54 },
		{ 151, 152, 54 },
		{ 151, 153, 54 },
		{ 151, 154, 54 },
		{ 151, 155, 54 },
		{ 151, 156, 54 },
		{ 151, 157, 54 },
		{ 151, 158, 54 },
		{ 151, 159, 54 },
		{ 151, 160, 54 },
		{ 151, 161, 54 },
		{ 151, 162, 54 },
		{ 151, 163, 54 },
		{ 151, 164, 54 },
		{ 152, 138, 54 },
		{ 152, 139, 54 },
		{ 152, 140, 54 },
		{ 152, 141, 54 },
		{ 152, 142, 54 },
		{ 152, 143, 54 },
		{ 152, 144, 54 },
		{ 152, 145, 54 },
		{ 152, 146, 54 },
		{ 152, 147, 54 },
		{ 152, 148, 54 },
		{ 152, 149, 54 },
		{ 152, 150, 54 },
		{ 152, 151, 54 },
		{ 152, 152, 54 },
		{ 152, 153, 54 },
		{ 152, 154, 54 },
		{ 152, 155, 54 },
		{ 152, 156, 54 },
		{ 152, 157, 54 },
		{ 152, 158, 54 },
		{ 152, 159, 54 },
		{ 152, 160, 54 },
		{ 152, 161, 54 },
		{ 152, 162, 54 },
		{ 152, 163, 54 },
		{ 152, 164, 54 },
		{ 152, 165, 54 },
		{ 153, 139, 54 },
		{ 153, 140, 54 },
		{ 153, 141, 54 },
		{ 153, 142, 54 },
		{ 153, 143, 54 },
		{ 153, 144, 54 },
		{ 153, 145, 54 },
		{ 153, 146, 54 },
		{ 153, 147, 54 },
		{ 153, 148, 54 },
		{ 153, 149, 54 },
		{ 153, 150, 54 },
		{ 153, 151, 54 },
		{ 153, 152, 54 },
		{ 153, 153, 54 },
		{ 153, 154, 54 },
		{ 153, 155, 54 },
		{ 153, 156, 54 },
		{ 153, 157, 54 },
		{ 153, 158, 54 },
		{ 153, 159, 54 },
		{ 153, 160, 54 },
		{ 153, 161, 54 },
		{ 153, 162, 54 },
		{ 153, 163, 54 },
		{ 153, 164, 54 },
		{ 153, 165, 54 },
		{ 154, 140, 54 },
		{ 154, 141, 54 },
		{ 154, 142, 54 },
		{ 154, 143, 54 },
		{ 154, 144, 54 },
		{ 154, 145, 54 },
		{ 154, 146, 54 },
		{ 154, 147, 54 },
		{ 154, 148, 54 },
		{ 154, 149, 54 },
		{ 154, 150, 54 },
		{ 154, 151, 54 },
		{ 154, 152, 54 },
		{ 154, 153, 54 },
		{ 154, 154, 54 },
		{ 154, 155, 54 },
		{ 154, 156, 54 },
		{ 154, 157, 54 },
		{ 154, 158, 54 },
		{ 154, 159, 54 },
		{ 154, 160, 54 },
		{ 154, 161, 54 },
		{ 154, 162, 54 },
		{ 154, 163, 54 },
		{ 154, 164, 54 },
		{ 154, 165, 54 },
		{ 154, 166, 54 },
		{ 155, 141, 54 },
		{ 155, 142, 54 },
		{ 155, 143, 54 },
		{ 155, 144, 54 },
		{ 155, 145, 54 },
		{ 155, 146, 54 },
		{ 155, 147, 54 },
		{ 155, 148, 54 },
		{ 155, 149, 54 },
		{ 155, 150, 54 },
		{ 155, 151, 54 },
		{ 155, 152, 54 },
		{ 155, 153, 54 },
		{ 155, 154, 54 },
		{ 155, 155, 54 },
		{ 155, 156, 54 },
		{ 155, 157, 54 },
		{ 155, 158, 54 },
		{ 155, 159, 54 },
		{ 155, 160, 54 },
		{ 155, 161, 54 },
		{ 155, 162, 54 },
		{ 155, 163, 54 },
		{ 155, 164, 54 },
		{ 155, 165, 54 },
		{ 155, 166, 54 },
		{ 155, 167, 54 },
		{ 156, 142, 54 },
		{ 156, 143, 54 },
		{ 156, 144, 54 },
		{ 156, 145, 54 },
		{ 156, 146, 54 },
		{ 156, 147, 54 },
		{ 156, 148, 54 },
		{ 156, 149, 54 },
		{ 156, 150, 54 },
		{ 156, 151, 54 },
		{ 156, 152, 54 },
		{ 156, 153, 54 },
		{ 156, 154, 54 },
		{ 156, 155, 54 },
		{ 156, 156, 54 },
		{ 156, 157, 54 },
		{ 156, 158, 54 },
		{ 156, 159, 54 },
		{ 156, 160, 54 },
		{ 156, 161, 54 },
		{ 156, 162, 54 },
		{ 156, 163, 54 },
		{ 156, 164, 54 },
		{ 156, 165, 54 },
		{ 156, 166, 54 },
		{ 156, 167, 54 },
		{ 156, 168, 54 },
		{ 157, 143, 54 },
		{ 157, 144, 54 },
		{ 157, 145, 54 },
		{ 157, 146, 54 },
		{ 157, 147, 54 },
		{ 157, 148, 54 },
		{ 157, 149, 54 },
		{ 157, 150, 54 },
		{ 157, 151, 54 },
		{ 157, 152, 54 },
		{ 157, 153, 54 },
		{ 157, 154, 54 },
		{ 157, 155, 54 },
		{ 157, 156, 54 },
		{ 157, 157, 54 },
		{ 157, 158, 54 },
		{ 157, 159, 54 },
		{ 157, 160, 54 },
		{ 157, 161, 54 },
		{ 157, 162, 54 },
		{ 157, 163, 54 },
		{ 157, 164, 54 },
		{ 157, 165, 54 },
		{ 157, 166, 54 },
		{ 157, 167, 54 },
		{ 157, 168, 54 },
		{ 157, 169, 54 },
		{ 158, 144, 54 },
		{ 158, 145, 54 },
		{ 158, 146, 54 },
		{ 158, 147, 54 },
		{ 158, 148, 54 },
		{ 158, 149, 54 },
		{ 158, 150, 54 },
		{ 158, 151, 54 },
		{ 158, 152, 54 },
		{ 158, 153, 54 },
		{ 158, 154, 54 },
		{ 158, 155, 54 },
		{ 158, 156, 54 },
		{ 158, 157, 54 },
		{ 158, 158, 54 },
		{ 158, 159, 54 },
		{ 158, 160, 54 },
		{ 158, 161, 54 },
		{ 158, 162, 54 },
		{ 158, 163, 54 },
		{ 158, 164, 54 },
		{ 158, 165, 54 },
		{ 158, 166, 54 },
		{ 158, 167, 54 },
		{ 158, 168, 54 },
		{ 158, 169, 54 },
		{ 158, 170, 54 },
		{ 159, 146, 54 },
		{ 159, 147, 54 },
		{ 159, 148, 54 },
		{ 159, 149, 54 },
		{ 159, 150, 54 },
		{ 159, 151, 54 },
		{ 159, 152, 54 },
		{ 159, 153, 54 },
		{ 159, 154, 54 },
		{ 159, 155, 54 },
		{ 159, 156, 54 },
		{ 159, 157, 54 },
		{ 159, 158, 54 },
		{ 159, 159, 54 },
		{ 159, 160, 54 },
		{ 159, 161, 54 },
		{ 159, 162, 54 },
		{ 159, 163, 54 },
		{ 159, 164, 54 },
		{ 159, 165, 54 },
		{ 159, 166, 54 },
		{ 159, 167, 54 },
		{ 159, 168, 54 },
		{ 159, 169, 54 },
		{ 159, 170, 54 },
		{ 159, 171, 54 },
		{ 160, 147, 54 },
		{ 160, 148, 54 },
		{ 160, 149, 54 },
		{ 160, 150, 54 },
		{ 160, 151, 54 },
		{ 160, 152, 54 },
		{ 160, 153, 54 },
		{ 160, 154, 54 },
		{ 160, 155, 54 },
		{ 160, 156, 54 },
		{ 160, 157, 54 },
		{ 160, 158, 54 },
		{ 160, 159, 54 },
		{ 160, 160, 54 },
		{ 160, 161, 54 },
		{ 160, 162, 54 },
		{ 160, 163, 54 },
		{ 160, 164, 54 },
		{ 160, 165, 54 },
		{ 160, 166, 54 },
		{ 160, 167, 54 },
		{ 160, 168, 54 },
		{ 160, 169, 54 },
		{ 160, 170, 54 },
		{ 160, 171, 54 },
		{ 160, 172, 54 },
		{ 161, 148, 54 },
		{ 161, 149, 54 },
		{ 161, 150, 54 },
		{ 161, 151, 54 },
		{ 161, 152, 54 },
		{ 161, 153, 54 },
		{ 161, 154, 54 },
		{ 161, 155, 54 },
		{ 161, 156, 54 },
		{ 161, 157, 54 },
		{ 161, 158, 54 },
		{ 161, 159, 54 },
		{ 161, 160, 54 },
		{ 161, 161, 54 },
		{ 161, 162, 54 },
		{ 161, 163, 54 },
		{ 161, 164, 54 },
		{ 161, 165, 54 },
		{ 161, 166, 54 },
		{ 161, 167, 54 },
		{ 161, 168, 54 },
		{ 161, 169, 54 },
		{ 161, 170, 54 },
		{ 161, 171, 54 },
		{ 161, 172, 54 },
		{ 162, 149, 54 },
		{ 162, 150, 54 },
		{ 162, 151, 54 },
		{ 162, 152, 54 },
		{ 162, 153, 54 },
		{ 162, 154, 54 },
		{ 162, 155, 54 },
		{ 162, 156, 54 },
		{ 162, 157, 54 },
		{ 162, 158, 54 },
		{ 162, 159, 54 },
		{ 162, 160, 54 },
		{ 162, 161, 54 },
		{ 162, 162, 54 },
		{ 162, 163, 54 },
		{ 162, 164, 54 },
		{ 162, 165, 54 },
		{ 162, 166, 54 },
		{ 162, 167, 54 },
		{ 162, 168, 54 },
		{ 162, 169, 54 },
		{ 162, 170, 54 },
		{ 162, 171, 54 },
		{ 162, 172, 54 },
		{ 162, 173, 54 },
		{ 163, 150, 54 },
		{ 163, 151, 54 },
		{ 163, 152, 54 },
		{ 163, 153, 54 },
		{ 163, 154, 54 },
		{ 163, 155, 54 },
		{ 163, 156, 54 },
		{ 163, 157, 54 },
		{ 163, 158, 54 },
		{ 163, 159, 54 },
		{ 163, 160, 54 },
		{ 163, 161, 54 },
		{ 163, 162, 54 },
		{ 163, 163, 54 },
		{ 163, 164, 54 },
		{ 163, 165, 54 },
		{ 163, 166, 54 },
		{ 163, 167, 54 },
		{ 163, 168, 54 },
		{ 163, 169, 54 },
		{ 163, 170, 54 },
		{ 163, 171, 54 },
		{ 163, 172, 54 },
		{ 163, 173, 54 },
		{ 163, 174, 54 },
		{ 164, 151, 54 },
		{ 164, 152, 54 },
		{ 164, 153, 54 },
		{ 164, 154, 54 },
		{ 164, 155, 54 },
		{ 164, 156, 54 },
		{ 164, 157, 54 },
		{ 164, 158, 54 },
		{ 164, 159, 54 },
		{ 164, 160, 54 },
		{ 164, 161, 54 },
		{ 164, 162, 54 },
		{ 164, 163, 54 },
		{ 164, 164, 54 },
		{ 164, 165, 54 },
		{ 164, 166, 54 },
		{ 164, 167, 54 },
		{ 164, 168, 54 },
		{ 164, 169, 54 },
		{ 164, 170, 54 },
		{ 164, 171, 54 },
		{ 164, 172, 54 },
		{ 164, 173, 54 },
		{ 164, 174, 54 },
		{ 164, 175, 54 },
		{ 165, 152, 54 },
		{ 165, 153, 54 },
		{ 165, 154, 54 },
		{ 165, 155, 54 },
		{ 165, 156, 54 },
		{ 165, 157, 54 },
		{ 165, 158, 54 },
		{ 165, 159, 54 },
		{ 165, 160, 54 },
		{ 165, 161, 54 },
		{ 165, 162, 54 },
		{ 165, 163, 54 },
		{ 165, 164, 54 },
		{ 165, 165, 54 },
		{ 165, 166, 54 },
		{ 165, 167, 54 },
		{ 165, 168, 54 },
		{ 165, 169, 54 },
		{ 165, 170, 54 },
		{ 165, 171, 54 },
		{ 165, 172, 54 },
		{ 165, 173, 54 },
		{ 165, 174, 54 },
		{ 165, 175, 54 },
		{ 165, 176, 54 },
		{ 166, 154, 54 },
		{ 166, 155, 54 },
		{ 166, 156, 54 },
		{ 166, 157, 54 },
		{ 166, 158, 54 },
		{ 166, 159, 54 },
		{ 166, 160, 54 },
		{ 166, 161, 54 },
		{ 166, 162, 54 },
		{ 166, 163, 54 },
		{ 166, 164, 54 },
		{ 166, 165, 54 },
		{ 166, 166, 54 },
		{ 166, 167, 54 },
		{ 166, 168, 54 },
		{ 166, 169, 54 },
		{ 166, 170, 54 },
		{ 166, 171, 54 },
		{ 166, 172, 54 },
		{ 166, 173, 54 },
		{ 166, 174, 54 },
		{ 166, 175, 54 },
		{ 166, 176, 54 },
		{ 166, 177, 54 },
		{ 167, 155, 54 },
		{ 167, 156, 54 },
		{ 167, 157, 54 },
		{ 167, 158, 54 },
		{ 167, 159, 54 },
		{ 167, 160, 54 },
		{ 167, 161, 54 },
		{ 167, 162, 54 },
		{ 167, 163, 54 },
		{ 167, 164, 54 },
		{ 167, 165, 54 },
		{ 167, 166, 54 },
		{ 167, 167, 54 },
		{ 167, 168, 54 },
		{ 167, 169, 54 },
		{ 167, 170, 54 },
		{ 167, 171, 54 },
		{ 167, 172, 54 },
		{ 167, 173, 54 },
		{ 167, 174, 54 },
		{ 167, 175, 54 },
		{ 167, 176, 54 },
		{ 167, 177, 54 },
		{ 167, 178, 54 },
		{ 168, 156, 54 },
		{ 168, 157, 54 },
		{ 168, 158, 54 },
		{ 168, 159, 54 },
		{ 168, 160, 54 },
		{ 168, 161, 54 },
		{ 168, 162, 54 },
		{ 168, 163, 54 },
		{ 168, 164, 54 },
		{ 168, 165, 54 },
		{ 168, 166, 54 },
		{ 168, 167, 54 },
		{ 168, 168, 54 },
		{ 168, 169, 54 },
		{ 168, 170, 54 },
		{ 168, 171, 54 },
		{ 168, 172, 54 },
		{ 168, 173, 54 },
		{ 168, 174, 54 },
		{ 168, 175, 54 },
		{ 168, 176, 54 },
		{ 168, 177, 54 },
		{ 168, 178, 54 },
		{ 168, 179, 54 },
		{ 169, 157, 54 },
		{ 169, 158, 54 },
		{ 169, 159, 54 },
		{ 169, 160, 54 },
		{ 169, 161, 54 },
		{ 169, 162, 54 },
		{ 169, 163, 54 },
		{ 169, 164, 54 },
		{ 169, 165, 54 },
		{ 169, 166, 54 },
		{ 169, 167, 54 },
		{ 169, 168, 54 },
		{ 169, 169, 54 },
		{ 169, 170, 54 },
		{ 169, 171, 54 },
		{ 169, 172, 54 },
		{ 169, 173, 54 },
		{ 169, 174, 54 },
		{ 169, 175, 54 },
		{ 169, 176, 54 },
		{ 169, 177, 54 },
		{ 169, 178, 54 },
		{ 169, 179, 54 },
		{ 170, 158, 54 },
		{ 170, 159, 54 },
		{ 170, 160, 54 },
		{ 170, 161, 54 },
		{ 170, 162, 54 },
		{ 170, 163, 54 },
		{ 170, 164, 54 },
		{ 170, 165, 54 },
		{ 170, 166, 54 },
		{ 170, 167, 54 },
		{ 170, 168, 54 },
		{ 170, 169, 54 },
		{ 170, 170, 54 },
		{ 170, 171, 54 },
		{ 170, 172, 54 },
		{ 170, 173, 54 },
		{ 170, 174, 54 },
		{ 170, 175, 54 },
		{ 170, 176, 54 },
		{ 170, 177, 54 },
		{ 170, 178, 54 },
		{ 170, 179, 54 },
		{ 170, 180, 54 },
		{ 171, 159, 54 },
		{ 171, 160, 54 },
		{ 171, 161, 54 },
		{ 171, 162, 54 },
		{ 171, 163, 54 },
		{ 171, 164, 54 },
		{ 171, 165, 54 },
		{ 171, 166, 54 },
		{ 171, 167, 54 },
		{ 171, 168, 54 },
		{ 171, 169, 54 },
		{ 171, 170, 54 },
		{ 171, 171, 54 },
		{ 171, 172, 54 },
		{ 171, 173, 54 },
		{ 171, 174, 54 },
		{ 171, 175, 54 },
		{ 171, 176, 54 },
		{ 171, 177, 54 },
		{ 171, 178, 54 },
		{ 171, 179, 54 },
		{ 171, 180, 54 },
		{ 171, 181, 54 },
		{ 172, 160, 54 },
		{ 172, 161, 54 },
		{ 172, 162, 54 },
		{ 172, 163, 54 },
		{ 172, 164, 54 },
		{ 172, 165, 54 },
		{ 172, 166, 54 },
		{ 172, 167, 54 },
		{ 172, 168, 54 },
		{ 172, 169, 54 },
		{ 172, 170, 54 },
		{ 172, 171, 54 },
		{ 172, 172, 54 },
		{ 172, 173, 54 },
		{ 172, 174, 54 },
		{ 172, 175, 54 },
		{ 172, 176, 54 },
		{ 172, 177, 54 },
		{ 172, 178, 54 },
		{ 172, 179, 54 },
		{ 172, 180, 54 },
		{ 172, 181, 54 },
		{ 172, 182, 54 },
		{ 173, 162, 54 },
		{ 173, 163, 54 },
		{ 173, 164, 54 },
		{ 173, 165, 54 },
		{ 173, 166, 54 },
		{ 173, 167, 54 },
		{ 173, 168, 54 },
		{ 173, 169, 54 },
		{ 173, 170, 54 },
		{ 173, 171, 54 },
		{ 173, 172, 54 },
		{ 173, 173, 54 },
		{ 173, 174, 54 },
		{ 173, 175, 54 },
		{ 173, 176, 54 },
		{ 173, 177, 54 },
		{ 173, 178, 54 },
		{ 173, 179, 54 },
		{ 173, 180, 54 },
		{ 173, 181, 54 },
		{ 173, 182, 54 },
		{ 173, 183, 54 },
		{ 174, 163, 54 },
		{ 174, 164, 54 },
		{ 174, 165, 54 },
		{ 174, 166, 54 },
		{ 174, 167, 54 },
		{ 174, 168, 54 },
		{ 174, 169, 54 },
		{ 174, 170, 54 },
		{ 174, 171, 54 },
		{ 174, 172, 54 },
		{ 174, 173, 54 },
		{ 174, 174, 54 },
		{ 174, 175, 54 },
		{ 174, 176, 54 },
		{ 174, 177, 54 },
		{ 174, 178, 54 },
		{ 174, 179, 54 },
		{ 174, 180, 54 },
		{ 174, 181, 54 },
		{ 174, 182, 54 },
		{ 174, 183, 54 },
		{ 174, 184, 54 },
		{ 175, 164, 54 },
		{ 175, 165, 54 },
		{ 175, 166, 54 },
		{ 175, 167, 54 },
		{ 175, 168, 54 },
		{ 175, 169, 54 },
		{ 175, 170, 54 },
		{ 175, 171, 54 },
		{ 175, 172, 54 },
		{ 175, 173, 54 },
		{ 175, 174, 54 },
		{ 175, 175, 54 },
		{ 175, 176, 54 },
		{ 175, 177, 54 },
		{ 175, 178, 54 },
		{ 175, 179, 54 },
		{ 175, 180, 54 },
		{ 175, 181, 54 },
		{ 175, 182, 54 },
		{ 175, 183, 54 },
		{ 175, 184, 54 },
		{ 175, 185, 54 },
		{ 176, 165, 54 },
		{ 176, 166, 54 },
		{ 176, 167, 54 },
		{ 176, 168, 54 },
		{ 176, 169, 54 },
		{ 176, 170, 54 },
		{ 176, 171, 54 },
		{ 176, 172, 54 },
		{ 176, 173, 54 },
		{ 176, 174, 54 },
		{ 176, 175, 54 },
		{ 176, 176, 54 },
		{ 176, 177, 54 },
		{ 176, 178, 54 },
		{ 176, 179, 54 },
		{ 176, 180, 54 },
		{ 176, 181, 54 },
		{ 176, 182, 54 },
		{ 176, 183, 54 },
		{ 176, 184, 54 },
		{ 176, 185, 54 },
		{ 176, 186, 54 },
		{ 177, 166, 54 },
		{ 177, 167, 54 },
		{ 177, 168, 54 },
		{ 177, 169, 54 },
		{ 177, 170, 54 },
		{ 177, 171, 54 },
		{ 177, 172, 54 },
		{ 177, 173, 54 },
		{ 177, 174, 54 },
		{ 177, 175, 54 },
		{ 177, 176, 54 },
		{ 177, 177, 54 },
		{ 177, 178, 54 },
		{ 177, 179, 54 },
		{ 177, 180, 54 },
		{ 177, 181, 54 },
		{ 177, 182, 54 },
		{ 177, 183, 54 },
		{ 177, 184, 54 },
		{ 177, 185, 54 },
		{ 177, 186, 54 },
		{ 178, 167, 54 },
		{ 178, 168, 54 },
		{ 178, 169, 54 },
		{ 178, 170, 54 },
		{ 178, 171, 54 },
		{ 178, 172, 54 },
		{ 178, 173, 54 },
		{ 178, 174, 54 },
		{ 178, 175, 54 },
		{ 178, 176, 54 },
		{ 178, 177, 54 },
		{ 178, 178, 54 },
		{ 178, 179, 54 },
		{ 178, 180, 54 },
		{ 178, 181, 54 },
		{ 178, 182, 54 },
		{ 178, 183, 54 },
		{ 178, 184, 54 },
		{ 178, 185, 54 },
		{ 178, 186, 54 },
		{ 178, 187, 54 },
		{ 179, 168, 54 },
		{ 179, 169, 54 },
		{ 179, 170, 54 },
		{ 179, 171, 54 },
		{ 179, 172, 54 },
		{ 179, 173, 54 },
		{ 179, 174, 54 },
		{ 179, 175, 54 },
		{ 179, 176, 54 },
		{ 179, 177, 54 },
		{ 179, 178, 54 },
		{ 179, 179, 54 },
		{ 179, 180, 54 },
		{ 179, 181, 54 },
		{ 179, 182, 54 },
		{ 179, 183, 54 },
		{ 179, 184, 54 },
		{ 179, 185, 54 },
		{ 179, 186, 54 },
		{ 179, 187, 54 },
		{ 179, 188, 54 },
		{ 180, 170, 54 },
		{ 180, 171, 54 },
		{ 180, 172, 54 },
		{ 180, 173, 54 },
		{ 180, 174, 54 },
		{ 180, 175, 54 },
		{ 180, 176, 54 },
		{ 180, 177, 54 },
		{ 180, 178, 54 },
		{ 180, 179, 54 },
		{ 180, 180, 54 },
		{ 180, 181, 54 },
		{ 180, 182, 54 },
		{ 180, 183, 54 },
		{ 180, 184, 54 },
		{ 180, 185, 54 },
		{ 180, 186, 54 },
		{ 180, 187, 54 },
		{ 180, 188, 54 },
		{ 180, 189, 54 },
		{ 181, 171, 54 },
		{ 181, 172, 54 },
		{ 181, 173, 54 },
		{ 181, 174, 54 },
		{ 181, 175, 54 },
		{ 181, 176, 54 },
		{ 181, 177, 54 },
		{ 181, 178, 54 },
		{ 181, 179, 54 },
		{ 181, 180, 54 },
		{ 181, 181, 54 },
		{ 181, 182, 54 },
		{ 181, 183, 54 },
		{ 181, 184, 54 },
		{ 181, 185, 54 },
		{ 181, 186, 54 },
		{ 181, 187, 54 },
		{ 181, 188, 54 },
		{ 181, 189, 54 },
		{ 181, 190, 54 },
		{ 182, 172, 54 },
		{ 182, 173, 54 },
		{ 182, 174, 54 },
		{ 182, 175, 54 },
		{ 182, 176, 54 },
		{ 182, 177, 54 },
		{ 182, 178, 54 },
		{ 182, 179, 54 },
		{ 182, 180, 54 },
		{ 182, 181, 54 },
		{ 182, 182, 54 },
		{ 182, 183, 54 },
		{ 182, 184, 54 },
		{ 182, 185, 54 },
		{ 182, 186, 54 },
		{ 182, 187, 54 },
		{ 182, 188, 54 },
		{ 182, 189, 54 },
		{ 182, 190, 54 },
		{ 182, 191, 54 },
		{ 183, 173, 54 },
		{ 183, 174, 54 },
		{ 183, 175, 54 },
		{ 183, 176, 54 },
		{ 183, 177, 54 },
		{ 183, 178, 54 },
		{ 183, 179, 54 },
		{ 183, 180, 54 },
		{ 183, 181, 54 },
		{ 183, 182, 54 },
		{ 183, 183, 54 },
		{ 183, 184, 54 },
		{ 183, 185, 54 },
		{ 183, 186, 54 },
		{ 183, 187, 54 },
		{ 183, 188, 54 },
		{ 183, 189, 54 },
		{ 183, 190, 54 },
		{ 183, 191, 54 },
		{ 183, 192, 54 },
		{ 184, 174, 54 },
		{ 184, 175, 54 },
		{ 184, 176, 54 },
		{ 184, 177, 54 },
		{ 184, 178, 54 },
		{ 184, 179, 54 },
		{ 184, 180, 54 },
		{ 184, 181, 54 },
		{ 184, 182, 54 },
		{ 184, 183, 54 },
		{ 184, 184, 54 },
		{ 184, 185, 54 },
		{ 184, 186, 54 },
		{ 184, 187, 54 },
		{ 184, 188, 54 },
		{ 184, 189, 54 },
		{ 184, 190, 54 },
		{ 184, 191, 54 },
		{ 184, 192, 54 },
		{ 184, 193, 54 },
		{ 185, 175, 54 },
		{ 185, 176, 54 },
		{ 185, 177, 54 },
		{ 185, 178, 54 },
		{ 185, 179, 54 },
		{ 185, 180, 54 },
		{ 185, 181, 54 },
		{ 185, 182, 54 },
		{ 185, 183, 54 },
		{ 185, 184, 54 },
		{ 185, 185, 54 },
		{ 185, 186, 54 },
		{ 185, 187, 54 },
		{ 185, 188, 54 },
		{ 185, 189, 54 },
		{ 185, 190, 54 },
		{ 185, 191, 54 },
		{ 185, 192, 54 },
		{ 185, 193, 54 },
		{ 186, 176, 54 },
		{ 186, 177, 54 },
		{ 186, 178, 54 },
		{ 186, 179, 54 },
		{ 186, 180, 54 },
		{ 186, 181, 54 },
		{ 186, 182, 54 },
		{ 186, 183, 54 },
		{ 186, 184, 54 },
		{ 186, 185, 54 },
		{ 186, 186, 54 },
		{ 186, 187, 54 },
		{ 186, 188, 54 },
		{ 186, 189, 54 },
		{ 186, 190, 54 },
		{ 186, 191, 54 },
		{ 186, 192, 54 },
		{ 186, 193, 54 },
		{ 186, 194, 54 },
		{ 187, 178, 54 },
		{ 187, 179, 54 },
		{ 187, 180, 54 },
		{ 187, 181, 54 },
		{ 187, 182, 54 },
		{ 187, 183, 54 },
		{ 187, 184, 54 },
		{ 187, 185, 54 },
		{ 187, 186, 54 },
		{ 187, 187, 54 },
		{ 187, 188, 54 },
		{ 187, 189, 54 },
		{ 187, 190, 54 },
		{ 187, 191, 54 },
		{ 187, 192, 54 },
		{ 187, 193, 54 },
		{ 187, 194, 54 },
		{ 187, 195, 54 },
		{ 188, 179, 54 },
		{ 188, 180, 54 },
		{ 188, 181, 54 },
		{ 188, 182, 54 },
		{ 188, 183, 54 },
		{ 188, 184, 54 },
		{ 188, 185, 54 },
		{ 188, 186, 54 },
		{ 188, 187, 54 },
		{ 188, 188, 54 },
		{ 188, 189, 54 },
		{ 188, 190, 54 },
		{ 188, 191, 54 },
		{ 188, 192, 54 },
		{ 188, 193, 54 },
		{ 188, 194, 54 },
		{ 188, 195, 54 },
		{ 188, 196, 54 },
		{ 189, 180, 54 },
		{ 189, 181, 54 },
		{ 189, 182, 54 },
		{ 189, 183, 54 },
		{ 189, 184, 54 },
		{ 189, 185, 54 },
		{ 189, 186, 54 },
		{ 189, 187, 54 },
		{ 189, 188, 54 },
		{ 189, 189, 54 },
		{ 189, 190, 54 },
		{ 189, 191, 54 },
		{ 189, 192, 54 },
		{ 189, 193, 54 },
		{ 189, 194, 54 },
		{ 189, 195, 54 },
		{ 189, 196, 54 },
		{ 189, 197, 54 },
		{ 190, 181, 54 },
		{ 190, 182, 54 },
		{ 190, 183, 54 },
		{ 190, 184, 54 },
		{ 190, 185, 54 },
		{ 190, 186, 54 },
		{ 190, 187, 54 },
		{ 190, 188, 54 },
		{ 190, 189, 54 },
		{ 190, 190, 54 },
		{ 190, 191, 54 },
		{ 190, 192, 54 },
		{ 190, 193, 54 },
		{ 190, 194, 54 },
		{ 190, 195, 54 },
		{ 190, 196, 54 },
		{ 190, 197, 54 },
		{ 191, 182, 54 },
		{ 191, 183, 54 },
		{ 191, 184, 54 },
		{ 191, 185, 54 },
		{ 191, 186, 54 },
		{ 191, 187, 54 },
		{ 191, 188, 54 },
		{ 191, 189, 54 },
		{ 191, 190, 54 },
		{ 191, 191, 54 },
		{ 191, 192, 54 },
		{ 191, 193, 54 },
		{ 191, 194, 54 },
		{ 191, 195, 54 },
		{ 191, 196, 54 },
		{ 191, 197, 54 },
		{ 192, 183, 54 },
		{ 192, 184, 54 },
		{ 192, 185, 54 },
		{ 192, 186, 54 },
		{ 192, 187, 54 },
		{ 192, 188, 54 },
		{ 192, 189, 54 },
		{ 192, 190, 54 },
		{ 192, 191, 54 },
		{ 193, 184, 54 },
		{ 193, 185, 54 },
		{ 193, 186, 54 },
		{ 193, 187, 54 },
		{ 193, 188, 54 },
		{ 193, 189, 54 },
		{ 193, 190, 54 },
		{ 193, 191, 54 },
		{ 194, 186, 54 },
		{ 194, 187, 54 },
		{ 194, 188, 54 },
		{ 194, 189, 54 },
		{ 194, 190, 54 },
		{ 194, 191, 54 },
		{ 195, 187, 54 },
		{ 195, 188, 54 },
		{ 195, 189, 54 },
		{ 195, 190, 54 },
		{ 195, 191, 54 },
		{ 196, 188, 54 },
		{ 196, 189, 54 },
		{ 196, 190, 54 },
		{ 196, 191, 54 },
		{ 197, 189, 54 },
		{ 197, 190, 54 },
		{ 197, 191, 54 },
		{ 317, 323, 54 },
		{ 317, 324, 54 },
		{ 317, 325, 54 },
		{ 318, 323, 54 },
		{ 318, 324, 54 },
		{ 318, 325, 54 },
		{ 318, 326, 54 },
		{ 319, 323, 54 },
		{ 319, 324, 54 },
		{ 319, 325, 54 },
		{ 319, 326, 54 },
		{ 319, 327, 54 },
		{ 320, 323, 54 },
		{ 320, 324, 54 },
		{ 320, 325, 54 },
		{ 320, 326, 54 },
		{ 320, 327, 54 },
		{ 320, 328, 54 },
		{ 321, 323, 54 },
		{ 321, 324, 54 },
		{ 321, 325, 54 },
		{ 321, 326, 54 },
		{ 321, 327, 54 },
		{ 321, 328, 54 },
		{ 321, 329, 54 },
		{ 321, 330, 54 },
		{ 322, 323, 54 },
		{ 322, 324, 54 },
		{ 322, 325, 54 },
		{ 322, 326, 54 },
		{ 322, 327, 54 },
		{ 322, 328, 54 },
		{ 322, 329, 54 },
		{ 322, 330, 54 },
		{ 322, 331, 54 },
		{ 323, 317, 54 },
		{ 323, 318, 54 },
		{ 323, 319, 54 },
		{ 323, 320, 54 },
		{ 323, 321, 54 },
		{ 323, 322, 54 },
		{ 323, 323, 54 },
		{ 323, 324, 54 },
		{ 323, 325, 54 },
		{ 323, 326, 54 },
		{ 323, 327, 54 },
		{ 323, 328, 54 },
		{ 323, 329, 54 },
		{ 323, 330, 54 },
		{ 323, 331, 54 },
		{ 323, 332, 54 },
		{ 324, 317, 54 },
		{ 324, 318, 54 },
		{ 324, 319, 54 },
		{ 324, 320, 54 },
		{ 324, 321, 54 },
		{ 324, 322, 54 },
		{ 324, 323, 54 },
		{ 324, 324, 54 },
		{ 324, 325, 54 },
		{ 324, 326, 54 },
		{ 324, 327, 54 },
		{ 324, 328, 54 },
		{ 324, 329, 54 },
		{ 324, 330, 54 },
		{ 324, 331, 54 },
		{ 324, 332, 54 },
		{ 324, 333, 54 },
		{ 325, 317, 54 },
		{ 325, 318, 54 },
		{ 325, 319, 54 },
		{ 325, 320, 54 },
		{ 325, 321, 54 },
		{ 325, 322, 54 },
		{ 325, 323, 54 },
		{ 325, 324, 54 },
		{ 325, 325, 54 },
		{ 325, 326, 54 },
		{ 325, 327, 54 },
		{ 325, 328, 54 },
		{ 325, 329, 54 },
		{ 325, 330, 54 },
		{ 325, 331, 54 },
		{ 325, 332, 54 },
		{ 325, 333, 54 },
		{ 325, 334, 54 },
		{ 326, 318, 54 },
		{ 326, 319, 54 },
		{ 326, 320, 54 },
		{ 326, 321, 54 },
		{ 326, 322, 54 },
		{ 326, 323, 54 },
		{ 326, 324, 54 },
		{ 326, 325, 54 },
		{ 326, 326, 54 },
		{ 326, 327, 54 },
		{ 326, 328, 54 },
		{ 326, 329, 54 },
		{ 326, 330, 54 },
		{ 326, 331, 54 },
		{ 326, 332, 54 },
		{ 326, 333, 54 },
		{ 326, 334, 54 },
		{ 326, 335, 54 },
		{ 327, 319, 54 },
		{ 327, 320, 54 },
		{ 327, 321, 54 },
		{ 327, 322, 54 },
		{ 327, 323, 54 },
		{ 327, 324, 54 },
		{ 327, 325, 54 },
		{ 327, 326, 54 },
		{ 327, 327, 54 },
		{ 327, 328, 54 },
		{ 327, 329, 54 },
		{ 327, 330, 54 },
		{ 327, 331, 54 },
		{ 327, 332, 54 },
		{ 327, 333, 54 },
		{ 327, 334, 54 },
		{ 327, 335, 54 },
		{ 327, 336, 54 },
		{ 328, 320, 54 },
		{ 328, 321, 54 },
		{ 328, 322, 54 },
		{ 328, 323, 54 },
		{ 328, 324, 54 },
		{ 328, 325, 54 },
		{ 328, 326, 54 },
		{ 328, 327, 54 },
		{ 328, 328, 54 },
		{ 328, 329, 54 },
		{ 328, 330, 54 },
		{ 328, 331, 54 },
		{ 328, 332, 54 },
		{ 328, 333, 54 },
		{ 328, 334, 54 },
		{ 328, 335, 54 },
		{ 328, 336, 54 },
		{ 328, 337, 54 },
		{ 328, 338, 54 },
		{ 329, 321, 54 },
		{ 329, 322, 54 },
		{ 329, 323, 54 },
		{ 329, 324, 54 },
		{ 329, 325, 54 },
		{ 329, 326, 54 },
		{ 329, 327, 54 },
		{ 329, 328, 54 },
		{ 329, 329, 54 },
		{ 329, 330, 54 },
		{ 329, 331, 54 },
		{ 329, 332, 54 },
		{ 329, 333, 54 },
		{ 329, 334, 54 },
		{ 329, 335, 54 },
		{ 329, 336, 54 },
		{ 329, 337, 54 },
		{ 329, 338, 54 },
		{ 329, 339, 54 },
		{ 330, 321, 54 },
		{ 330, 322, 54 },
		{ 330, 323, 54 },
		{ 330, 324, 54 },
		{ 330, 325, 54 },
		{ 330, 326, 54 },
		{ 330, 327, 54 },
		{ 330, 328, 54 },
		{ 330, 329, 54 },
		{ 330, 330, 54 },
		{ 330, 331, 54 },
		{ 330, 332, 54 },
		{ 330, 333, 54 },
		{ 330, 334, 54 },
		{ 330, 335, 54 },
		{ 330, 336, 54 },
		{ 330, 337, 54 },
		{ 330, 338, 54 },
		{ 330, 339, 54 },
		{ 330, 340, 54 },
		{ 331, 322, 54 },
		{ 331, 323, 54 },
		{ 331, 324, 54 },
		{ 331, 325, 54 },
		{ 331, 326, 54 },
		{ 331, 327, 54 },
		{ 331, 328, 54 },
		{ 331, 329, 54 },
		{ 331, 330, 54 },
		{ 331, 331, 54 },
		{ 331, 332, 54 },
		{ 331, 333, 54 },
		{ 331, 334, 54 },
		{ 331, 335, 54 },
		{ 331, 336, 54 },
		{ 331, 337, 54 },
		{ 331, 338, 54 },
		{ 331, 339, 54 },
		{ 331, 340, 54 },
		{ 331, 341, 54 },
		{ 332, 323, 54 },
		{ 332, 324, 54 },
		{ 332, 325, 54 },
		{ 332, 326, 54 },
		{ 332, 327, 54 },
		{ 332, 328, 54 },
		{ 332, 329, 54 },
		{ 332, 330, 54 },
		{ 332, 331, 54 },
		{ 332, 332, 54 },
		{ 332, 333, 54 },
		{ 332, 334, 54 },
		{ 332, 335, 54 },
		{ 332, 336, 54 },
		{ 332, 337, 54 },
		{ 332, 338, 54 },
		{ 332, 339, 54 },
		{ 332, 340, 54 },
		{ 332, 341, 54 },
		{ 332, 342, 54 },
		{ 333, 324, 54 },
		{ 333, 325, 54 },
		{ 333, 326, 54 },
		{ 333, 327, 54 },
		{ 333, 328, 54 },
		{ 333, 329, 54 },
		{ 333, 330, 54 },
		{ 333, 331, 54 },
		{ 333, 332, 54 },
		{ 333, 333, 54 },
		{ 333, 334, 54 },
		{ 333, 335, 54 },
		{ 333, 336, 54 },
		{ 333, 337, 54 },
		{ 333, 338, 54 },
		{ 333, 339, 54 },
		{ 333, 340, 54 },
		{ 333, 341, 54 },
		{ 333, 342, 54 },
		{ 333, 343, 54 },
		{ 334, 325, 54 },
		{ 334, 326, 54 },
		{ 334, 327, 54 },
		{ 334, 328, 54 },
		{ 334, 329, 54 },
		{ 334, 330, 54 },
		{ 334, 331, 54 },
		{ 334, 332, 54 },
		{ 334, 333, 54 },
		{ 334, 334, 54 },
		{ 334, 335, 54 },
		{ 334, 336, 54 },
		{ 334, 337, 54 },
		{ 334, 338, 54 },
		{ 334, 339, 54 },
		{ 334, 340, 54 },
		{ 334, 341, 54 },
		{ 334, 342, 54 },
		{ 334, 343, 54 },
		{ 334, 344, 54 },
		{ 335, 326, 54 },
		{ 335, 327, 54 },
		{ 335, 328, 54 },
		{ 335, 329, 54 },
		{ 335, 330, 54 },
		{ 335, 331, 54 },
		{ 335, 332, 54 },
		{ 335, 333, 54 },
		{ 335, 334, 54 },
		{ 335, 335, 54 },
		{ 335, 336, 54 },
		{ 335, 337, 54 },
		{ 335, 338, 54 },
		{ 335, 339, 54 },
		{ 335, 340, 54 },
		{ 335, 341, 54 },
		{ 335, 342, 54 },
		{ 335, 343, 54 },
		{ 335, 344, 54 },
		{ 335, 345, 54 },
		{ 335, 346, 54 },
		{ 336, 327, 54 },
		{ 336, 328, 54 },
		{ 336, 329, 54 },
		{ 336, 330, 54 },
		{ 336, 331, 54 },
		{ 336, 332, 54 },
		{ 336, 333, 54 },
		{ 336, 334, 54 },
		{ 336, 335, 54 },
		{ 336, 336, 54 },
		{ 336, 337, 54 },
		{ 336, 338, 54 },
		{ 336, 339, 54 },
		{ 336, 340, 54 },
		{ 336, 341, 54 },
		{ 336, 342, 54 },
		{ 336, 343, 54 },
		{ 336, 344, 54 },
		{ 336, 345, 54 },
		{ 336, 346, 54 },
		{ 336, 347, 54 },
		{ 337, 328, 54 },
		{ 337, 329, 54 },
		{ 337, 330, 54 },
		{ 337, 331, 54 },
		{ 337, 332, 54 },
		{ 337, 333, 54 },
		{ 337, 334, 54 },
		{ 337, 335, 54 },
		{ 337, 336, 54 },
		{ 337, 337, 54 },
		{ 337, 338, 54 },
		{ 337, 339, 54 },
		{ 337, 340, 54 },
		{ 337, 341, 54 },
		{ 337, 342, 54 },
		{ 337, 343, 54 },
		{ 337, 344, 54 },
		{ 337, 345, 54 },
		{ 337, 346, 54 },
		{ 337, 347, 54 },
		{ 337, 348, 54 },
		{ 338, 328, 54 },
		{ 338, 329, 54 },
		{ 338, 330, 54 },
		{ 338, 331, 54 },
		{ 338, 332, 54 },
		{ 338, 333, 54 },
		{ 338, 334, 54 },
		{ 338, 335, 54 },
		{ 338, 336, 54 },
		{ 338, 337, 54 },
		{ 338, 338, 54 },
		{ 338, 339, 54 },
		{ 338, 340, 54 },
		{ 338, 341, 54 },
		{ 338, 342, 54 },
		{ 338, 343, 54 },
		{ 338, 344, 54 },
		{ 338, 345, 54 },
		{ 338, 346, 54 },
		{ 338, 347, 54 },
		{ 338, 348, 54 },
		{ 338, 349, 54 },
		{ 339, 329, 54 },
		{ 339, 330, 54 },
		{ 339, 331, 54 },
		{ 339, 332, 54 },
		{ 339, 333, 54 },
		{ 339, 334, 54 },
		{ 339, 335, 54 },
		{ 339, 336, 54 },
		{ 339, 337, 54 },
		{ 339, 338, 54 },
		{ 339, 339, 54 },
		{ 339, 340, 54 },
		{ 339, 341, 54 },
		{ 339, 342, 54 },
		{ 339, 343, 54 },
		{ 339, 344, 54 },
		{ 339, 345, 54 },
		{ 339, 346, 54 },
		{ 339, 347, 54 },
		{ 339, 348, 54 },
		{ 339, 349, 54 },
		{ 339, 350, 54 },
		{ 340, 330, 54 },
		{ 340, 331, 54 },
		{ 340, 332, 54 },
		{ 340, 333, 54 },
		{ 340, 334, 54 },
		{ 340, 335, 54 },
		{ 340, 336, 54 },
		{ 340, 337, 54 },
		{ 340, 338, 54 },
		{ 340, 339, 54 },
		{ 340, 340, 54 },
		{ 340, 341, 54 },
		{ 340, 342, 54 },
		{ 340, 343, 54 },
		{ 340, 344, 54 },
		{ 340, 345, 54 },
		{ 340, 346, 54 },
		{ 340, 347, 54 },
		{ 340, 348, 54 },
		{ 340, 349, 54 },
		{ 340, 350, 54 },
		{ 340, 351, 54 },
		{ 341, 331, 54 },
		{ 341, 332, 54 },
		{ 341, 333, 54 },
		{ 341, 334, 54 },
		{ 341, 335, 54 },
		{ 341, 336, 54 },
		{ 341, 337, 54 },
		{ 341, 338, 54 },
		{ 341, 339, 54 },
		{ 341, 340, 54 },
		{ 341, 341, 54 },
		{ 341, 342, 54 },
		{ 341, 343, 54 },
		{ 341, 344, 54 },
		{ 341, 345, 54 },
		{ 341, 346, 54 },
		{ 341, 347, 54 },
		{ 341, 348, 54 },
		{ 341, 349, 54 },
		{ 341, 350, 54 },
		{ 341, 351, 54 },
		{ 341, 352, 54 },
		{ 342, 332, 54 },
		{ 342, 333, 54 },
		{ 342, 334, 54 },
		{ 342, 335, 54 },
		{ 342, 336, 54 },
		{ 342, 337, 54 },
		{ 342, 338, 54 },
		{ 342, 339, 54 },
		{ 342, 340, 54 },
		{ 342, 341, 54 },
		{ 342, 342, 54 },
		{ 342, 343, 54 },
		{ 342, 344, 54 },
		{ 342, 345, 54 },
		{ 342, 346, 54 },
		{ 342, 347, 54 },
		{ 342, 348, 54 },
		{ 342, 349, 54 },
		{ 342, 350, 54 },
		{ 342, 351, 54 },
		{ 342, 352, 54 },
		{ 342, 353, 54 },
		{ 342, 354, 54 },
		{ 343, 333, 54 },
		{ 343, 334, 54 },
		{ 343, 335, 54 },
		{ 343, 336, 54 },
		{ 343, 337, 54 },
		{ 343, 338, 54 },
		{ 343, 339, 54 },
		{ 343, 340, 54 },
		{ 343, 341, 54 },
		{ 343, 342, 54 },
		{ 343, 343, 54 },
		{ 343, 344, 54 },
		{ 343, 345, 54 },
		{ 343, 346, 54 },
		{ 343, 347, 54 },
		{ 343, 348, 54 },
		{ 343, 349, 54 },
		{ 343, 350, 54 },
		{ 343, 351, 54 },
		{ 343, 352, 54 },
		{ 343, 353, 54 },
		{ 343, 354, 54 },
		{ 343, 355, 54 },
		{ 344, 334, 54 },
		{ 344, 335, 54 },
		{ 344, 336, 54 },
		{ 344, 337, 54 },
		{ 344, 338, 54 },
		{ 344, 339, 54 },
		{ 344, 340, 54 },
		{ 344, 341, 54 },
		{ 344, 342, 54 },
		{ 344, 343, 54 },
		{ 344, 344, 54 },
		{ 344, 345, 54 },
		{ 344, 346, 54 },
		{ 344, 347, 54 },
		{ 344, 348, 54 },
		{ 344, 349, 54 },
		{ 344, 350, 54 },
		{ 344, 351, 54 },
		{ 344, 352, 54 },
		{ 344, 353, 54 },
		{ 344, 354, 54 },
		{ 344, 355, 54 },
		{ 344, 356, 54 },
		{ 345, 335, 54 },
		{ 345, 336, 54 },
		{ 345, 337, 54 },
		{ 345, 338, 54 },
		{ 345, 339, 54 },
		{ 345, 340, 54 },
		{ 345, 341, 54 },
		{ 345, 342, 54 },
		{ 345, 343, 54 },
		{ 345, 344, 54 },
		{ 345, 345, 54 },
		{ 345, 346, 54 },
		{ 345, 347, 54 },
		{ 345, 348, 54 },
		{ 345, 349, 54 },
		{ 345, 350, 54 },
		{ 345, 351, 54 },
		{ 345, 352, 54 },
		{ 345, 353, 54 },
		{ 345, 354, 54 },
		{ 345, 355, 54 },
		{ 345, 356, 54 },
		{ 345, 357, 54 },
		{ 346, 335, 54 },
		{ 346, 336, 54 },
		{ 346, 337, 54 },
		{ 346, 338, 54 },
		{ 346, 339, 54 },
		{ 346, 340, 54 },
		{ 346, 341, 54 },
		{ 346, 342, 54 },
		{ 346, 343, 54 },
		{ 346, 344, 54 },
		{ 346, 345, 54 },
		{ 346, 346, 54 },
		{ 346, 347, 54 },
		{ 346, 348, 54 },
		{ 346, 349, 54 },
		{ 346, 350, 54 },
		{ 346, 351, 54 },
		{ 346, 352, 54 },
		{ 346, 353, 54 },
		{ 346, 354, 54 },
		{ 346, 355, 54 },
		{ 346, 356, 54 },
		{ 346, 357, 54 },
		{ 346, 358, 54 },
		{ 347, 336, 54 },
		{ 347, 337, 54 },
		{ 347, 338, 54 },
		{ 347, 339, 54 },
		{ 347, 340, 54 },
		{ 347, 341, 54 },
		{ 347, 342, 54 },
		{ 347, 343, 54 },
		{ 347, 344, 54 },
		{ 347, 345, 54 },
		{ 347, 346, 54 },
		{ 347, 347, 54 },
		{ 347, 348, 54 },
		{ 347, 349, 54 },
		{ 347, 350, 54 },
		{ 347, 351, 54 },
		{ 347, 352, 54 },
		{ 347, 353, 54 },
		{ 347, 354, 54 },
		{ 347, 355, 54 },
		{ 347, 356, 54 },
		{ 347, 357, 54 },
		{ 347, 358, 54 },
		{ 347, 359, 54 },
		{ 348, 337, 54 },
		{ 348, 338, 54 },
		{ 348, 339, 54 },
		{ 348, 340, 54 },
		{ 348, 341, 54 },
		{ 348, 342, 54 },
		{ 348, 343, 54 },
		{ 348, 344, 54 },
		{ 348, 345, 54 },
		{ 348, 346, 54 },
		{ 348, 347, 54 },
		{ 348, 348, 54 },
		{ 348, 349, 54 },
		{ 348, 350, 54 },
		{ 348, 351, 54 },
		{ 348, 352, 54 },
		{ 348, 353, 54 },
		{ 348, 354, 54 },
		{ 348, 355, 54 },
		{ 348, 356, 54 },
		{ 348, 357, 54 },
		{ 348, 358, 54 },
		{ 348, 359, 54 },
		{ 348, 360, 54 },
		{ 349, 338, 54 },
		{ 349, 339, 54 },
		{ 349, 340, 54 },
		{ 349, 341, 54 },
		{ 349, 342, 54 },
		{ 349, 343, 54 },
		{ 349, 344, 54 },
		{ 349, 345, 54 },
		{ 349, 346, 54 },
		{ 349, 347, 54 },
		{ 349, 348, 54 },
		{ 349, 349, 54 },
		{ 349, 350, 54 },
		{ 349, 351, 54 },
		{ 349, 352, 54 },
		{ 349, 353, 54 },
		{ 349, 354, 54 },
		{ 349, 355, 54 },
		{ 349, 356, 54 },
		{ 349, 357, 54 },
		{ 349, 358, 54 },
		{ 349, 359, 54 },
		{ 349, 360, 54 },
		{ 349, 361, 54 },
		{ 349, 362, 54 },
		{ 350, 339, 54 },
		{ 350, 340, 54 },
		{ 350, 341, 54 },
		{ 350, 342, 54 },
		{ 350, 343, 54 },
		{ 350, 344, 54 },
		{ 350, 345, 54 },
		{ 350, 346, 54 },
		{ 350, 347, 54 },
		{ 350, 348, 54 },
		{ 350, 349, 54 },
		{ 350, 350, 54 },
		{ 350, 351, 54 },
		{ 350, 352, 54 },
		{ 350, 353, 54 },
		{ 350, 354, 54 },
		{ 350, 355, 54 },
		{ 350, 356, 54 },
		{ 350, 357, 54 },
		{ 350, 358, 54 },
		{ 350, 359, 54 },
		{ 350, 360, 54 },
		{ 350, 361, 54 },
		{ 350, 362, 54 },
		{ 350, 363, 54 },
		{ 351, 340, 54 },
		{ 351, 341, 54 },
		{ 351, 342, 54 },
		{ 351, 343, 54 },
		{ 351, 344, 54 },
		{ 351, 345, 54 },
		{ 351, 346, 54 },
		{ 351, 347, 54 },
		{ 351, 348, 54 },
		{ 351, 349, 54 },
		{ 351, 350, 54 },
		{ 351, 351, 54 },
		{ 351, 352, 54 },
		{ 351, 353, 54 },
		{ 351, 354, 54 },
		{ 351, 355, 54 },
		{ 351, 356, 54 },
		{ 351, 357, 54 },
		{ 351, 358, 54 },
		{ 351, 359, 54 },
		{ 351, 360, 54 },
		{ 351, 361, 54 },
		{ 351, 362, 54 },
		{ 351, 363, 54 },
		{ 351, 364, 54 },
		{ 352, 341, 54 },
		{ 352, 342, 54 },
		{ 352, 343, 54 },
		{ 352, 344, 54 },
		{ 352, 345, 54 },
		{ 352, 346, 54 },
		{ 352, 347, 54 },
		{ 352, 348, 54 },
		{ 352, 349, 54 },
		{ 352, 350, 54 },
		{ 352, 351, 54 },
		{ 352, 352, 54 },
		{ 352, 353, 54 },
		{ 352, 354, 54 },
		{ 352, 355, 54 },
		{ 352, 356, 54 },
		{ 352, 357, 54 },
		{ 352, 358, 54 },
		{ 352, 359, 54 },
		{ 352, 360, 54 },
		{ 352, 361, 54 },
		{ 352, 362, 54 },
		{ 352, 363, 54 },
		{ 352, 364, 54 },
		{ 352, 365, 54 },
		{ 353, 342, 54 },
		{ 353, 343, 54 },
		{ 353, 344, 54 },
		{ 353, 345, 54 },
		{ 353, 346, 54 },
		{ 353, 347, 54 },
		{ 353, 348, 54 },
		{ 353, 349, 54 },
		{ 353, 350, 54 },
		{ 353, 351, 54 },
		{ 353, 352, 54 },
		{ 353, 353, 54 },
		{ 353, 354, 54 },
		{ 353, 355, 54 },
		{ 353, 356, 54 },
		{ 353, 357, 54 },
		{ 353, 358, 54 },
		{ 353, 359, 54 },
		{ 353, 360, 54 },
		{ 353, 361, 54 },
		{ 353, 362, 54 },
		{ 353, 363, 54 },
		{ 353, 364, 54 },
		{ 353, 365, 54 },
		{ 353, 366, 54 },
		{ 354, 342, 54 },
		{ 354, 343, 54 },
		{ 354, 344, 54 },
		{ 354, 345, 54 },
		{ 354, 346, 54 },
		{ 354, 347, 54 },
		{ 354, 348, 54 },
		{ 354, 349, 54 },
		{ 354, 350, 54 },
		{ 354, 351, 54 },
		{ 354, 352, 54 },
		{ 354, 353, 54 },
		{ 354, 354, 54 },
		{ 354, 355, 54 },
		{ 354, 356, 54 },
		{ 354, 357, 54 },
		{ 354, 358, 54 },
		{ 354, 359, 54 },
		{ 354, 360, 54 },
		{ 354, 361, 54 },
		{ 354, 362, 54 },
		{ 354, 363, 54 },
		{ 354, 364, 54 },
		{ 354, 365, 54 },
		{ 354, 366, 54 },
		{ 354, 367, 54 },
		{ 355, 343, 54 },
		{ 355, 344, 54 },
		{ 355, 345, 54 },
		{ 355, 346, 54 },
		{ 355, 347, 54 },
		{ 355, 348, 54 },
		{ 355, 349, 54 },
		{ 355, 350, 54 },
		{ 355, 351, 54 },
		{ 355, 352, 54 },
		{ 355, 353, 54 },
		{ 355, 354, 54 },
		{ 355, 355, 54 },
		{ 355, 356, 54 },
		{ 355, 357, 54 },
		{ 355, 358, 54 },
		{ 355, 359, 54 },
		{ 355, 360, 54 },
		{ 355, 361, 54 },
		{ 355, 362, 54 },
		{ 355, 363, 54 },
		{ 355, 364, 54 },
		{ 355, 365, 54 },
		{ 355, 366, 54 },
		{ 355, 367, 54 },
		{ 355, 368, 54 },
		{ 356, 344, 54 },
		{ 356, 345, 54 },
		{ 356, 346, 54 },
		{ 356, 347, 54 },
		{ 356, 348, 54 },
		{ 356, 349, 54 },
		{ 356, 350, 54 },
		{ 356, 351, 54 },
		{ 356, 352, 54 },
		{ 356, 353, 54 },
		{ 356, 354, 54 },
		{ 356, 355, 54 },
		{ 356, 356, 54 },
		{ 356, 357, 54 },
		{ 356, 358, 54 },
		{ 356, 359, 54 },
		{ 356, 360, 54 },
		{ 356, 361, 54 },
		{ 356, 362, 54 },
		{ 356, 363, 54 },
		{ 356, 364, 54 },
		{ 356, 365, 54 },
		{ 356, 366, 54 },
		{ 356, 367, 54 },
		{ 356, 368, 54 },
		{ 356, 369, 54 },
		{ 356, 370, 54 },
		{ 357, 345, 54 },
		{ 357, 346, 54 },
		{ 357, 347, 54 },
		{ 357, 348, 54 },
		{ 357, 349, 54 },
		{ 357, 350, 54 },
		{ 357, 351, 54 },
		{ 357, 352, 54 },
		{ 357, 353, 54 },
		{ 357, 354, 54 },
		{ 357, 355, 54 },
		{ 357, 356, 54 },
		{ 357, 357, 54 },
		{ 357, 358, 54 },
		{ 357, 359, 54 },
		{ 357, 360, 54 },
		{ 357, 361, 54 },
		{ 357, 362, 54 },
		{ 357, 363, 54 },
		{ 357, 364, 54 },
		{ 357, 365, 54 },
		{ 357, 366, 54 },
		{ 357, 367, 54 },
		{ 357, 368, 54 },
		{ 357, 369, 54 },
		{ 357, 370, 54 },
		{ 357, 371, 54 },
		{ 358, 346, 54 },
		{ 358, 347, 54 },
		{ 358, 348, 54 },
		{ 358, 349, 54 },
		{ 358, 350, 54 },
		{ 358, 351, 54 },
		{ 358, 352, 54 },
		{ 358, 353, 54 },
		{ 358, 354, 54 },
		{ 358, 355, 54 },
		{ 358, 356, 54 },
		{ 358, 357, 54 },
		{ 358, 358, 54 },
		{ 358, 359, 54 },
		{ 358, 360, 54 },
		{ 358, 361, 54 },
		{ 358, 362, 54 },
		{ 358, 363, 54 },
		{ 358, 364, 54 },
		{ 358, 365, 54 },
		{ 358, 366, 54 },
		{ 358, 367, 54 },
		{ 358, 368, 54 },
		{ 358, 369, 54 },
		{ 358, 370, 54 },
		{ 358, 371, 54 },
		{ 358, 372, 54 },
		{ 359, 347, 54 },
		{ 359, 348, 54 },
		{ 359, 349, 54 },
		{ 359, 350, 54 },
		{ 359, 351, 54 },
		{ 359, 352, 54 },
		{ 359, 353, 54 },
		{ 359, 354, 54 },
		{ 359, 355, 54 },
		{ 359, 356, 54 },
		{ 359, 357, 54 },
		{ 359, 358, 54 },
		{ 359, 359, 54 },
		{ 359, 360, 54 },
		{ 359, 361, 54 },
		{ 359, 362, 54 },
		{ 359, 363, 54 },
		{ 359, 364, 54 },
		{ 359, 365, 54 },
		{ 359, 366, 54 },
		{ 359, 367, 54 },
		{ 359, 368, 54 },
		{ 359, 369, 54 },
		{ 359, 370, 54 },
		{ 359, 371, 54 },
		{ 359, 372, 54 },
		{ 359, 373, 54 },
		{ 360, 348, 54 },
		{ 360, 349, 54 },
		{ 360, 350, 54 },
		{ 360, 351, 54 },
		{ 360, 352, 54 },
		{ 360, 353, 54 },
		{ 360, 354, 54 },
		{ 360, 355, 54 },
		{ 360, 356, 54 },
		{ 360, 357, 54 },
		{ 360, 358, 54 },
		{ 360, 359, 54 },
		{ 360, 360, 54 },
		{ 360, 361, 54 },
		{ 360, 362, 54 },
		{ 360, 363, 54 },
		{ 360, 364, 54 },
		{ 360, 365, 54 },
		{ 360, 366, 54 },
		{ 360, 367, 54 },
		{ 360, 368, 54 },
		{ 360, 369, 54 },
		{ 360, 370, 54 },
		{ 360, 371, 54 },
		{ 360, 372, 54 },
		{ 360, 373, 54 },
		{ 360, 374, 54 },
		{ 361, 349, 54 },
		{ 361, 350, 54 },
		{ 361, 351, 54 },
		{ 361, 352, 54 },
		{ 361, 353, 54 },
		{ 361, 354, 54 },
		{ 361, 355, 54 },
		{ 361, 356, 54 },
		{ 361, 357, 54 },
		{ 361, 358, 54 },
		{ 361, 359, 54 },
		{ 361, 360, 54 },
		{ 361, 361, 54 },
		{ 361, 362, 54 },
		{ 361, 363, 54 },
		{ 361, 364, 54 },
		{ 361, 365, 54 },
		{ 361, 366, 54 },
		{ 361, 367, 54 },
		{ 361, 368, 54 },
		{ 361, 369, 54 },
		{ 361, 370, 54 },
		{ 361, 371, 54 },
		{ 361, 372, 54 },
		{ 361, 373, 54 },
		{ 361, 374, 54 },
		{ 361, 375, 54 },
		{ 362, 349, 54 },
		{ 362, 350, 54 },
		{ 362, 351, 54 },
		{ 362, 352, 54 },
		{ 362, 353, 54 },
		{ 362, 354, 54 },
		{ 362, 355, 54 },
		{ 362, 356, 54 },
		{ 362, 357, 54 },
		{ 362, 358, 54 },
		{ 362, 359, 54 },
		{ 362, 360, 54 },
		{ 362, 361, 54 },
		{ 362, 362, 54 },
		{ 362, 363, 54 },
		{ 362, 364, 54 },
		{ 362, 365, 54 },
		{ 362, 366, 54 },
		{ 362, 367, 54 },
		{ 362, 368, 54 },
		{ 362, 369, 54 },
		{ 362, 370, 54 },
		{ 362, 371, 54 },
		{ 362, 372, 54 },
		{ 362, 373, 54 },
		{ 362, 374, 54 },
		{ 362, 375, 54 },
		{ 362, 376, 54 },
		{ 363, 350, 54 },
		{ 363, 351, 54 },
		{ 363, 352, 54 },
		{ 363, 353, 54 },
		{ 363, 354, 54 },
		{ 363, 355, 54 },
		{ 363, 356, 54 },
		{ 363, 357, 54 },
		{ 363, 358, 54 },
		{ 363, 359, 54 },
		{ 363, 360, 54 },
		{ 363, 361, 54 },
		{ 363, 362, 54 },
		{ 363, 363, 54 },
		{ 363, 364, 54 },
		{ 363, 365, 54 },
		{ 363, 366, 54 },
		{ 363, 367, 54 },
		{ 363, 368, 54 },
		{ 363, 369, 54 },
		{ 363, 370, 54 },
		{ 363, 371, 54 },
		{ 363, 372, 54 },
		{ 363, 373, 54 },
		{ 363, 374, 54 },
		{ 363, 375, 54 },
		{ 363, 376, 54 },
		{ 363, 377, 54 },
		{ 363, 378, 54 },
		{ 364, 351, 54 },
		{ 364, 352, 54 },
		{ 364, 353, 54 },
		{ 364, 354, 54 },
		{ 364, 355, 54 },
		{ 364, 356, 54 },
		{ 364, 357, 54 },
		{ 364, 358, 54 },
		{ 364, 359, 54 },
		{ 364, 360, 54 },
		{ 364, 361, 54 },
		{ 364, 362, 54 },
		{ 364, 363, 54 },
		{ 364, 364, 54 },
		{ 364, 365, 54 },
		{ 364, 366, 54 },
		{ 364, 367, 54 },
		{ 364, 368, 54 },
		{ 364, 369, 54 },
		{ 364, 370, 54 },
		{ 364, 371, 54 },
		{ 364, 372, 54 },
		{ 364, 373, 54 },
		{ 364, 374, 54 },
		{ 364, 375, 54 },
		{ 364, 376, 54 },
		{ 364, 377, 54 },
		{ 364, 378, 54 },
		{ 364, 379, 54 },
		{ 365, 352, 54 },
		{ 365, 353, 54 },
		{ 365, 354, 54 },
		{ 365, 355, 54 },
		{ 365, 356, 54 },
		{ 365, 357, 54 },
		{ 365, 358, 54 },
		{ 365, 359, 54 },
		{ 365, 360, 54 },
		{ 365, 361, 54 },
		{ 365, 362, 54 },
		{ 365, 363, 54 },
		{ 365, 364, 54 },
		{ 365, 365, 54 },
		{ 365, 366, 54 },
		{ 365, 367, 54 },
		{ 365, 368, 54 },
		{ 365, 369, 54 },
		{ 365, 370, 54 },
		{ 365, 371, 54 },
		{ 365, 372, 54 },
		{ 365, 373, 54 },
		{ 365, 374, 54 },
		{ 365, 375, 54 },
		{ 365, 376, 54 },
		{ 365, 377, 54 },
		{ 365, 378, 54 },
		{ 365, 379, 54 },
		{ 365, 380, 54 },
		{ 366, 353, 54 },
		{ 366, 354, 54 },
		{ 366, 355, 54 },
		{ 366, 356, 54 },
		{ 366, 357, 54 },
		{ 366, 358, 54 },
		{ 366, 359, 54 },
		{ 366, 360, 54 },
		{ 366, 361, 54 },
		{ 366, 362, 54 },
		{ 366, 363, 54 },
		{ 366, 364, 54 },
		{ 366, 365, 54 },
		{ 366, 366, 54 },
		{ 366, 367, 54 },
		{ 366, 368, 54 },
		{ 366, 369, 54 },
		{ 366, 370, 54 },
		{ 366, 371, 54 },
		{ 366, 372, 54 },
		{ 366, 373, 54 },
		{ 366, 374, 54 },
		{ 366, 375, 54 },
		{ 366, 376, 54 },
		{ 366, 377, 54 },
		{ 366, 378, 54 },
		{ 366, 379, 54 },
		{ 366, 380, 54 },
		{ 366, 381, 54 },
		{ 367, 354, 54 },
		{ 367, 355, 54 },
		{ 367, 356, 54 },
		{ 367, 357, 54 },
		{ 367, 358, 54 },
		{ 367, 359, 54 },
		{ 367, 360, 54 },
		{ 367, 361, 54 },
		{ 367, 362, 54 },
		{ 367, 363, 54 },
		{ 367, 364, 54 },
		{ 367, 365, 54 },
		{ 367, 366, 54 },
		{ 367, 367, 54 },
		{ 367, 368, 54 },
		{ 367, 369, 54 },
		{ 367, 370, 54 },
		{ 367, 371, 54 },
		{ 367, 372, 54 },
		{ 367, 373, 54 },
		{ 367, 374, 54 },
		{ 367, 375, 54 },
		{ 367, 376, 54 },
		{ 367, 377, 54 },
		{ 367, 378, 54 },
		{ 367, 379, 54 },
		{ 367, 380, 54 },
		{ 367, 381, 54 },
		{ 367, 382, 54 },
		{ 368, 355, 54 },
		{ 368, 356, 54 },
		{ 368, 357, 54 },
		{ 368, 358, 54 },
		{ 368, 359, 54 },
		{ 368, 360, 54 },
		{ 368, 361, 54 },
		{ 368, 362, 54 },
		{ 368, 363, 54 },
		{ 368, 364, 54 },
		{ 368, 365, 54 },
		{ 368, 366, 54 },
		{ 368, 367, 54 },
		{ 368, 368, 54 },
		{ 368, 369, 54 },
		{ 368, 370, 54 },
		{ 368, 371, 54 },
		{ 368, 372, 54 },
		{ 368, 373, 54 },
		{ 368, 374, 54 },
		{ 368, 375, 54 },
		{ 368, 376, 54 },
		{ 368, 377, 54 },
		{ 368, 378, 54 },
		{ 368, 379, 54 },
		{ 368, 380, 54 },
		{ 368, 381, 54 },
		{ 368, 382, 54 },
		{ 368, 383, 54 },
		{ 369, 356, 54 },
		{ 369, 357, 54 },
		{ 369, 358, 54 },
		{ 369, 359, 54 },
		{ 369, 360, 54 },
		{ 369, 361, 54 },
		{ 369, 362, 54 },
		{ 369, 363, 54 },
		{ 369, 364, 54 },
		{ 369, 365, 54 },
		{ 369, 366, 54 },
		{ 369, 367, 54 },
		{ 369, 368, 54 },
		{ 369, 369, 54 },
		{ 369, 370, 54 },
		{ 369, 371, 54 },
		{ 369, 372, 54 },
		{ 369, 373, 54 },
		{ 369, 374, 54 },
		{ 369, 375, 54 },
		{ 369, 376, 54 },
		{ 369, 377, 54 },
		{ 369, 378, 54 },
		{ 369, 379, 54 },
		{ 369, 380, 54 },
		{ 369, 381, 54 },
		{ 369, 382, 54 },
		{ 369, 383, 54 },
		{ 369, 384, 54 },
		{ 370, 356, 54 },
		{ 370, 357, 54 },
		{ 370, 358, 54 },
		{ 370, 359, 54 },
		{ 370, 360, 54 },
		{ 370, 361, 54 },
		{ 370, 362, 54 },
		{ 370, 363, 54 },
		{ 370, 364, 54 },
		{ 370, 365, 54 },
		{ 370, 366, 54 },
		{ 370, 367, 54 },
		{ 370, 368, 54 },
		{ 370, 369, 54 },
		{ 370, 370, 54 },
		{ 370, 371, 54 },
		{ 370, 372, 54 },
		{ 370, 373, 54 },
		{ 370, 374, 54 },
		{ 370, 375, 54 },
		{ 370, 376, 54 },
		{ 370, 377, 54 },
		{ 370, 378, 54 },
		{ 370, 379, 54 },
		{ 370, 380, 54 },
		{ 370, 381, 54 },
		{ 370, 382, 54 },
		{ 370, 383, 54 },
		{ 370, 384, 54 },
		{ 370, 385, 54 },
		{ 370, 386, 54 },
		{ 371, 357, 54 },
		{ 371, 358, 54 },
		{ 371, 359, 54 },
		{ 371, 360, 54 },
		{ 371, 361, 54 },
		{ 371, 362, 54 },
		{ 371, 363, 54 },
		{ 371, 364, 54 },
		{ 371, 365, 54 },
		{ 371, 366, 54 },
		{ 371, 367, 54 },
		{ 371, 368, 54 },
		{ 371, 369, 54 },
		{ 371, 370, 54 },
		{ 371, 371, 54 },
		{ 371, 372, 54 },
		{ 371, 373, 54 },
		{ 371, 374, 54 },
		{ 371, 375, 54 },
		{ 371, 376, 54 },
		{ 371, 377, 54 },
		{ 371, 378, 54 },
		{ 371, 379, 54 },
		{ 371, 380, 54 },
		{ 371, 381, 54 },
		{ 371, 382, 54 },
		{ 371, 383, 54 },
		{ 371, 384, 54 },
		{ 371, 385, 54 },
		{ 371, 386, 54 },
		{ 371, 387, 54 },
		{ 372, 358, 54 },
		{ 372, 359, 54 },
		{ 372, 360, 54 },
		{ 372, 361, 54 },
		{ 372, 362, 54 },
		{ 372, 363, 54 },
		{ 372, 364, 54 },
		{ 372, 365, 54 },
		{ 372, 366, 54 },
		{ 372, 367, 54 },
		{ 372, 368, 54 },
		{ 372, 369, 54 },
		{ 372, 370, 54 },
		{ 372, 371, 54 },
		{ 372, 372, 54 },
		{ 372, 373, 54 },
		{ 372, 374, 54 },
		{ 372, 375, 54 },
		{ 372, 376, 54 },
		{ 372, 377, 54 },
		{ 372, 378, 54 },
		{ 372, 379, 54 },
		{ 372, 380, 54 },
		{ 372, 381, 54 },
		{ 372, 382, 54 },
		{ 372, 383, 54 },
		{ 372, 384, 54 },
		{ 372, 385, 54 },
		{ 372, 386, 54 },
		{ 372, 387, 54 },
		{ 372, 388, 54 },
		{ 373, 359, 54 },
		{ 373, 360, 54 },
		{ 373, 361, 54 },
		{ 373, 362, 54 },
		{ 373, 363, 54 },
		{ 373, 364, 54 },
		{ 373, 365, 54 },
		{ 373, 366, 54 },
		{ 373, 367, 54 },
		{ 373, 368, 54 },
		{ 373, 369, 54 },
		{ 373, 370, 54 },
		{ 373, 371, 54 },
		{ 373, 372, 54 },
		{ 373, 373, 54 },
		{ 373, 374, 54 },
		{ 373, 375, 54 },
		{ 373, 376, 54 },
		{ 373, 377, 54 },
		{ 373, 378, 54 },
		{ 373, 379, 54 },
		{ 373, 380, 54 },
		{ 373, 381, 54 },
		{ 373, 382, 54 },
		{ 373, 383, 54 },
		{ 373, 384, 54 },
		{ 373, 385, 54 },
		{ 373, 386, 54 },
		{ 373, 387, 54 },
		{ 373, 388, 54 },
		{ 373, 389, 54 },
		{ 374, 360, 54 },
		{ 374, 361, 54 },
		{ 374, 362, 54 },
		{ 374, 363, 54 },
		{ 374, 364, 54 },
		{ 374, 365, 54 },
		{ 374, 366, 54 },
		{ 374, 367, 54 },
		{ 374, 368, 54 },
		{ 374, 369, 54 },
		{ 374, 370, 54 },
		{ 374, 371, 54 },
		{ 374, 372, 54 },
		{ 374, 373, 54 },
		{ 374, 374, 54 },
		{ 374, 375, 54 },
		{ 374, 376, 54 },
		{ 374, 377, 54 },
		{ 374, 378, 54 },
		{ 374, 379, 54 },
		{ 374, 380, 54 },
		{ 374, 381, 54 },
		{ 374, 382, 54 },
		{ 374, 383, 54 },
		{ 374, 384, 54 },
		{ 374, 385, 54 },
		{ 374, 386, 54 },
		{ 374, 387, 54 },
		{ 374, 388, 54 },
		{ 374, 389, 54 },
		{ 374, 390, 54 },
		{ 375, 361, 54 },
		{ 375, 362, 54 },
		{ 375, 363, 54 },
		{ 375, 364, 54 },
		{ 375, 365, 54 },
		{ 375, 366, 54 },
		{ 375, 367, 54 },
		{ 375, 368, 54 },
		{ 375, 369, 54 },
		{ 375, 370, 54 },
		{ 375, 371, 54 },
		{ 375, 372, 54 },
		{ 375, 373, 54 },
		{ 375, 374, 54 },
		{ 375, 375, 54 },
		{ 375, 376, 54 },
		{ 375, 377, 54 },
		{ 375, 378, 54 },
		{ 375, 379, 54 },
		{ 375, 380, 54 },
		{ 375, 381, 54 },
		{ 375, 382, 54 },
		{ 375, 383, 54 },
		{ 375, 384, 54 },
		{ 375, 385, 54 },
		{ 375, 386, 54 },
		{ 375, 387, 54 },
		{ 375, 388, 54 },
		{ 375, 389, 54 },
		{ 375, 390, 54 },
		{ 375, 391, 54 },
		{ 376, 362, 54 },
		{ 376, 363, 54 },
		{ 376, 364, 54 },
		{ 376, 365, 54 },
		{ 376, 366, 54 },
		{ 376, 367, 54 },
		{ 376, 368, 54 },
		{ 376, 369, 54 },
		{ 376, 370, 54 },
		{ 376, 371, 54 },
		{ 376, 372, 54 },
		{ 376, 373, 54 },
		{ 376, 374, 54 },
		{ 376, 375, 54 },
		{ 376, 376, 54 },
		{ 376, 377, 54 },
		{ 376, 378, 54 },
		{ 376, 379, 54 },
		{ 376, 380, 54 },
		{ 376, 381, 54 },
		{ 376, 382, 54 },
		{ 376, 383, 54 },
		{ 376, 384, 54 },
		{ 376, 385, 54 },
		{ 376, 386, 54 },
		{ 376, 387, 54 },
		{ 376, 388, 54 },
		{ 376, 389, 54 },
		{ 376, 390, 54 },
		{ 376, 391, 54 },
		{ 376, 392, 54 },
		{ 377, 363, 54 },
		{ 377, 364, 54 },
		{ 377, 365, 54 },
		{ 377, 366, 54 },
		{ 377, 367, 54 },
		{ 377, 368, 54 },
		{ 377, 369, 54 },
		{ 377, 370, 54 },
		{ 377, 371, 54 },
		{ 377, 372, 54 },
		{ 377, 373, 54 },
		{ 377, 374, 54 },
		{ 377, 375, 54 },
		{ 377, 376, 54 },
		{ 377, 377, 54 },
		{ 377, 378, 54 },
		{ 377, 379, 54 },
		{ 377, 380, 54 },
		{ 377, 381, 54 },
		{ 377, 382, 54 },
		{ 377, 383, 54 },
		{ 377, 384, 54 },
		{ 377, 385, 54 },
		{ 377, 386, 54 },
		{ 377, 387, 54 },
		{ 377, 388, 54 },
		{ 377, 389, 54 },
		{ 377, 390, 54 },
		{ 377, 391, 54 },
		{ 377, 392, 54 },
		{ 377, 393, 54 },
		{ 377, 394, 54 },
		{ 378, 363, 54 },
		{ 378, 364, 54 },
		{ 378, 365, 54 },
		{ 378, 366, 54 },
		{ 378, 367, 54 },
		{ 378, 368, 54 },
		{ 378, 369, 54 },
		{ 378, 370, 54 },
		{ 378, 371, 54 },
		{ 378, 372, 54 },
		{ 378, 373, 54 },
		{ 378, 374, 54 },
		{ 378, 375, 54 },
		{ 378, 376, 54 },
		{ 378, 377, 54 },
		{ 378, 378, 54 },
		{ 378, 379, 54 },
		{ 378, 380, 54 },
		{ 378, 381, 54 },
		{ 378, 382, 54 },
		{ 378, 383, 54 },
		{ 378, 384, 54 },
		{ 378, 385, 54 },
		{ 378, 386, 54 },
		{ 378, 387, 54 },
		{ 378, 388, 54 },
		{ 378, 389, 54 },
		{ 378, 390, 54 },
		{ 378, 391, 54 },
		{ 378, 392, 54 },
		{ 378, 393, 54 },
		{ 378, 394, 54 },
		{ 378, 395, 54 },
		{ 379, 364, 54 },
		{ 379, 365, 54 },
		{ 379, 366, 54 },
		{ 379, 367, 54 },
		{ 379, 368, 54 },
		{ 379, 369, 54 },
		{ 379, 370, 54 },
		{ 379, 371, 54 },
		{ 379, 372, 54 },
		{ 379, 373, 54 },
		{ 379, 374, 54 },
		{ 379, 375, 54 },
		{ 379, 376, 54 },
		{ 379, 377, 54 },
		{ 379, 378, 54 },
		{ 379, 379, 54 },
		{ 379, 380, 54 },
		{ 379, 381, 54 },
		{ 379, 382, 54 },
		{ 379, 383, 54 },
		{ 379, 384, 54 },
		{ 379, 385, 54 },
		{ 379, 386, 54 },
		{ 379, 387, 54 },
		{ 379, 388, 54 },
		{ 379, 389, 54 },
		{ 379, 390, 54 },
		{ 379, 391, 54 },
		{ 379, 392, 54 },
		{ 379, 393, 54 },
		{ 379, 394, 54 },
		{ 379, 395, 54 },
		{ 379, 396, 54 },
		{ 380, 365, 54 },
		{ 380, 366, 54 },
		{ 380, 367, 54 },
		{ 380, 368, 54 },
		{ 380, 369, 54 },
		{ 380, 370, 54 },
		{ 380, 371, 54 },
		{ 380, 372, 54 },
		{ 380, 373, 54 },
		{ 380, 374, 54 },
		{ 380, 375, 54 },
		{ 380, 376, 54 },
		{ 380, 377, 54 },
		{ 380, 378, 54 },
		{ 380, 379, 54 },
		{ 380, 380, 54 },
		{ 380, 381, 54 },
		{ 380, 382, 54 },
		{ 380, 383, 54 },
		{ 380, 384, 54 },
		{ 380, 385, 54 },
		{ 380, 386, 54 },
		{ 380, 387, 54 },
		{ 380, 388, 54 },
		{ 380, 389, 54 },
		{ 380, 390, 54 },
		{ 380, 391, 54 },
		{ 380, 392, 54 },
		{ 380, 393, 54 },
		{ 380, 394, 54 },
		{ 380, 395, 54 },
		{ 380, 396, 54 },
		{ 380, 397, 54 },
		{ 381, 366, 54 },
		{ 381, 367, 54 },
		{ 381, 368, 54 },
		{ 381, 369, 54 },
		{ 381, 370, 54 },
		{ 381, 371, 54 },
		{ 381, 372, 54 },
		{ 381, 373, 54 },
		{ 381, 374, 54 },
		{ 381, 375, 54 },
		{ 381, 376, 54 },
		{ 381, 377, 54 },
		{ 381, 378, 54 },
		{ 381, 379, 54 },
		{ 381, 380, 54 },
		{ 381, 381, 54 },
		{ 381, 382, 54 },
		{ 381, 383, 54 },
		{ 381, 384, 54 },
		{ 381, 385, 54 },
		{ 381, 386, 54 },
		{ 381, 387, 54 },
		{ 381, 388, 54 },
		{ 381, 389, 54 },
		{ 381, 390, 54 },
		{ 381, 391, 54 },
		{ 381, 392, 54 },
		{ 381, 393, 54 },
		{ 381, 394, 54 },
		{ 381, 395, 54 },
		{ 381, 396, 54 },
		{ 381, 397, 54 },
		{ 381, 398, 54 },
		{ 382, 367, 54 },
		{ 382, 368, 54 },
		{ 382, 369, 54 },
		{ 382, 370, 54 },
		{ 382, 371, 54 },
		{ 382, 372, 54 },
		{ 382, 373, 54 },
		{ 382, 374, 54 },
		{ 382, 375, 54 },
		{ 382, 376, 54 },
		{ 382, 377, 54 },
		{ 382, 378, 54 },
		{ 382, 379, 54 },
		{ 382, 380, 54 },
		{ 382, 381, 54 },
		{ 382, 382, 54 },
		{ 382, 383, 54 },
		{ 382, 384, 54 },
		{ 382, 385, 54 },
		{ 382, 386, 54 },
		{ 382, 387, 54 },
		{ 382, 388, 54 },
		{ 382, 389, 54 },
		{ 382, 390, 54 },
		{ 382, 391, 54 },
		{ 382, 392, 54 },
		{ 382, 393, 54 },
		{ 382, 394, 54 },
		{ 382, 395, 54 },
		{ 382, 396, 54 },
		{ 382, 397, 54 },
		{ 382, 398, 54 },
		{ 382, 399, 54 },
		{ 383, 368, 54 },
		{ 383, 369, 54 },
		{ 383, 370, 54 },
		{ 383, 371, 54 },
		{ 383, 372, 54 },
		{ 383, 373, 54 },
		{ 383, 374, 54 },
		{ 383, 375, 54 },
		{ 383, 376, 54 },
		{ 383, 377, 54 },
		{ 383, 378, 54 },
		{ 383, 379, 54 },
		{ 383, 380, 54 },
		{ 383, 381, 54 },
		{ 383, 382, 54 },
		{ 383, 383, 54 },
		{ 383, 384, 54 },
		{ 383, 385, 54 },
		{ 383, 386, 54 },
		{ 383, 387, 54 },
		{ 383, 388, 54 },
		{ 383, 389, 54 },
		{ 383, 390, 54 },
		{ 383, 391, 54 },
		{ 383, 392, 54 },
		{ 383, 393, 54 },
		{ 383, 394, 54 },
		{ 383, 395, 54 },
		{ 383, 396, 54 },
		{ 383, 397, 54 },
		{ 383, 398, 54 },
		{ 383, 399, 54 },
		{ 383, 400, 54 },
		{ 384, 369, 54 },
		{ 384, 370, 54 },
		{ 384, 371, 54 },
		{ 384, 372, 54 },
		{ 384, 373, 54 },
		{ 384, 374, 54 },
		{ 384, 375, 54 },
		{ 384, 376, 54 },
		{ 384, 377, 54 },
		{ 384, 378, 54 },
		{ 384, 379, 54 },
		{ 384, 380, 54 },
		{ 384, 381, 54 },
		{ 384, 382, 54 },
		{ 384, 383, 54 },
		{ 384, 384, 54 },
		{ 384, 385, 54 },
		{ 384, 386, 54 },
		{ 384, 387, 54 },
		{ 384, 388, 54 },
		{ 384, 389, 54 },
		{ 384, 390, 54 },
		{ 384, 391, 54 },
		{ 384, 392, 54 },
		{ 384, 393, 54 },
		{ 384, 394, 54 },
		{ 384, 395, 54 },
		{ 384, 396, 54 },
		{ 384, 397, 54 },
		{ 384, 398, 54 },
		{ 384, 399, 54 },
		{ 384, 400, 54 },
		{ 384, 401, 54 },
		{ 384, 402, 54 },
		{ 385, 370, 54 },
		{ 385, 371, 54 },
		{ 385, 372, 54 },
		{ 385, 373, 54 },
		{ 385, 374, 54 },
		{ 385, 375, 54 },
		{ 385, 376, 54 },
		{ 385, 377, 54 },
		{ 385, 378, 54 },
		{ 385, 379, 54 },
		{ 385, 380, 54 },
		{ 385, 381, 54 },
		{ 385, 382, 54 },
		{ 385, 383, 54 },
		{ 385, 384, 54 },
		{ 385, 385, 54 },
		{ 385, 386, 54 },
		{ 385, 387, 54 },
		{ 385, 388, 54 },
		{ 385, 389, 54 },
		{ 385, 390, 54 },
		{ 385, 391, 54 },
		{ 385, 392, 54 },
		{ 385, 393, 54 },
		{ 385, 394, 54 },
		{ 385, 395, 54 },
		{ 385, 396, 54 },
		{ 385, 397, 54 },
		{ 385, 398, 54 },
		{ 385, 399, 54 },
		{ 385, 400, 54 },
		{ 385, 401, 54 },
		{ 385, 402, 54 },
		{ 385, 403, 54 },
		{ 386, 370, 54 },
		{ 386, 371, 54 },
		{ 386, 372, 54 },
		{ 386, 373, 54 },
		{ 386, 374, 54 },
		{ 386, 375, 54 },
		{ 386, 376, 54 },
		{ 386, 377, 54 },
		{ 386, 378, 54 },
		{ 386, 379, 54 },
		{ 386, 380, 54 },
		{ 386, 381, 54 },
		{ 386, 382, 54 },
		{ 386, 383, 54 },
		{ 386, 384, 54 },
		{ 386, 385, 54 },
		{ 386, 386, 54 },
		{ 386, 387, 54 },
		{ 386, 388, 54 },
		{ 386, 389, 54 },
		{ 386, 390, 54 },
		{ 386, 391, 54 },
		{ 386, 392, 54 },
		{ 386, 393, 54 },
		{ 386, 394, 54 },
		{ 386, 395, 54 },
		{ 386, 396, 54 },
		{ 386, 397, 54 },
		{ 386, 398, 54 },
		{ 386, 399, 54 },
		{ 386, 400, 54 },
		{ 386, 401, 54 },
		{ 386, 402, 54 },
		{ 386, 403, 54 },
		{ 386, 404, 54 },
		{ 387, 371, 54 },
		{ 387, 372, 54 },
		{ 387, 373, 54 },
		{ 387, 374, 54 },
		{ 387, 375, 54 },
		{ 387, 376, 54 },
		{ 387, 377, 54 },
		{ 387, 378, 54 },
		{ 387, 379, 54 },
		{ 387, 380, 54 },
		{ 387, 381, 54 },
		{ 387, 382, 54 },
		{ 387, 383, 54 },
		{ 387, 384, 54 },
		{ 387, 385, 54 },
		{ 387, 386, 54 },
		{ 387, 387, 54 },
		{ 387, 388, 54 },
		{ 387, 389, 54 },
		{ 387, 390, 54 },
		{ 387, 391, 54 },
		{ 387, 392, 54 },
		{ 387, 393, 54 },
		{ 387, 394, 54 },
		{ 387, 395, 54 },
		{ 387, 396, 54 },
		{ 387, 397, 54 },
		{ 387, 398, 54 },
		{ 387, 399, 54 },
		{ 387, 400, 54 },
		{ 387, 401, 54 },
		{ 387, 402, 54 },
		{ 387, 403, 54 },
		{ 387, 404, 54 },
		{ 387, 405, 54 },
		{ 388, 372, 54 },
		{ 388, 373, 54 },
		{ 388, 374, 54 },
		{ 388, 375, 54 },
		{ 388, 376, 54 },
		{ 388, 377, 54 },
		{ 388, 378, 54 },
		{ 388, 379, 54 },
		{ 388, 380, 54 },
		{ 388, 381, 54 },
		{ 388, 382, 54 },
		{ 388, 383, 54 },
		{ 388, 384, 54 },
		{ 388, 385, 54 },
		{ 388, 386, 54 },
		{ 388, 387, 54 },
		{ 388, 388, 54 },
		{ 388, 389, 54 },
		{ 388, 390, 54 },
		{ 388, 391, 54 },
		{ 388, 392, 54 },
		{ 388, 393, 54 },
		{ 388, 394, 54 },
		{ 388, 395, 54 },
		{ 388, 396, 54 },
		{ 388, 397, 54 },
		{ 388, 398, 54 },
		{ 388, 399, 54 },
		{ 388, 400, 54 },
		{ 388, 401, 54 },
		{ 388, 402, 54 },
		{ 388, 403, 54 },
		{ 388, 404, 54 },
		{ 388, 405, 54 },
		{ 388, 406, 54 },
		{ 389, 373, 54 },
		{ 389, 374, 54 },
		{ 389, 375, 54 },
		{ 389, 376, 54 },
		{ 389, 377, 54 },
		{ 389, 378, 54 },
		{ 389, 379, 54 },
		{ 389, 380, 54 },
		{ 389, 381, 54 },
		{ 389, 382, 54 },
		{ 389, 383, 54 },
		{ 389, 384, 54 },
		{ 389, 385, 54 },
		{ 389, 386, 54 },
		{ 389, 387, 54 },
		{ 389, 388, 54 },
		{ 389, 389, 54 },
		{ 389, 390, 54 },
		{ 389, 391, 54 },
		{ 389, 392, 54 },
		{ 389, 393, 54 },
		{ 389, 394, 54 },
		{ 389, 395, 54 },
		{ 389, 396, 54 },
		{ 389, 397, 54 },
		{ 389, 398, 54 },
		{ 389, 399, 54 },
		{ 389, 400, 54 },
		{ 389, 401, 54 },
		{ 389, 402, 54 },
		{ 389, 403, 54 },
		{ 389, 404, 54 },
		{ 389, 405, 54 },
		{ 389, 406, 54 },
		{ 389, 407, 54 },
		{ 390, 374, 54 },
		{ 390, 375, 54 },
		{ 390, 376, 54 },
		{ 390, 377, 54 },
		{ 390, 378, 54 },
		{ 390, 379, 54 },
		{ 390, 380, 54 },
		{ 390, 381, 54 },
		{ 390, 382, 54 },
		{ 390, 383, 54 },
		{ 390, 384, 54 },
		{ 390, 385, 54 },
		{ 390, 386, 54 },
		{ 390, 387, 54 },
		{ 390, 388, 54 },
		{ 390, 389, 54 },
		{ 390, 390, 54 },
		{ 390, 391, 54 },
		{ 390, 392, 54 },
		{ 390, 393, 54 },
		{ 390, 394, 54 },
		{ 390, 395, 54 },
		{ 390, 396, 54 },
		{ 390, 397, 54 },
		{ 390, 398, 54 },
		{ 390, 399, 54 },
		{ 390, 400, 54 },
		{ 390, 401, 54 },
		{ 390, 402, 54 },
		{ 390, 403, 54 },
		{ 390, 404, 54 },
		{ 390, 405, 54 },
		{ 390, 406, 54 },
		{ 390, 407, 54 },
		{ 390, 408, 54 },
		{ 391, 375, 54 },
		{ 391, 376, 54 },
		{ 391, 377, 54 },
		{ 391, 378, 54 },
		{ 391, 379, 54 },
		{ 391, 380, 54 },
		{ 391, 381, 54 },
		{ 391, 382, 54 },
		{ 391, 383, 54 },
		{ 391, 384, 54 },
		{ 391, 385, 54 },
		{ 391, 386, 54 },
		{ 391, 387, 54 },
		{ 391, 388, 54 },
		{ 391, 389, 54 },
		{ 391, 390, 54 },
		{ 391, 391, 54 },
		{ 391, 392, 54 },
		{ 391, 393, 54 },
		{ 391, 394, 54 },
		{ 391, 395, 54 },
		{ 391, 396, 54 },
		{ 391, 397, 54 },
		{ 391, 398, 54 },
		{ 391, 399, 54 },
		{ 391, 400, 54 },
		{ 391, 401, 54 },
		{ 391, 402, 54 },
		{ 391, 403, 54 },
		{ 391, 404, 54 },
		{ 391, 405, 54 },
		{ 391, 406, 54 },
		{ 391, 407, 54 },
		{ 391, 408, 54 },
		{ 391, 409, 54 },
		{ 391, 410, 54 },
		{ 392, 376, 54 },
		{ 392, 377, 54 },
		{ 392, 378, 54 },
		{ 392, 379, 54 },
		{ 392, 380, 54 },
		{ 392, 381, 54 },
		{ 392, 382, 54 },
		{ 392, 383, 54 },
		{ 392, 384, 54 },
		{ 392, 385, 54 },
		{ 392, 386, 54 },
		{ 392, 387, 54 },
		{ 392, 388, 54 },
		{ 392, 389, 54 },
		{ 392, 390, 54 },
		{ 392, 391, 54 },
		{ 392, 392, 54 },
		{ 392, 393, 54 },
		{ 392, 394, 54 },
		{ 392, 395, 54 },
		{ 392, 396, 54 },
		{ 392, 397, 54 },
		{ 392, 398, 54 },
		{ 392, 399, 54 },
		{ 392, 400, 54 },
		{ 392, 401, 54 },
		{ 392, 402, 54 },
		{ 392, 403, 54 },
		{ 392, 404, 54 },
		{ 392, 405, 54 },
		{ 392, 406, 54 },
		{ 392, 407, 54 },
		{ 392, 408, 54 },
		{ 392, 409, 54 },
		{ 392, 410, 54 },
		{ 392, 411, 54 },
		{ 393, 377, 54 },
		{ 393, 378, 54 },
		{ 393, 379, 54 },
		{ 393, 380, 54 },
		{ 393, 381, 54 },
		{ 393, 382, 54 },
		{ 393, 383, 54 },
		{ 393, 384, 54 },
		{ 393, 385, 54 },
		{ 393, 386, 54 },
		{ 393, 387, 54 },
		{ 393, 388, 54 },
		{ 393, 389, 54 },
		{ 393, 390, 54 },
		{ 393, 391, 54 },
		{ 393, 392, 54 },
		{ 393, 393, 54 },
		{ 393, 394, 54 },
		{ 393, 395, 54 },
		{ 393, 396, 54 },
		{ 393, 397, 54 },
		{ 393, 398, 54 },
		{ 393, 399, 54 },
		{ 393, 400, 54 },
		{ 393, 401, 54 },
		{ 393, 402, 54 },
		{ 393, 403, 54 },
		{ 393, 404, 54 },
		{ 393, 405, 54 },
		{ 393, 406, 54 },
		{ 393, 407, 54 },
		{ 393, 408, 54 },
		{ 393, 409, 54 },
		{ 393, 410, 54 },
		{ 393, 411, 54 },
		{ 393, 412, 54 },
		{ 394, 377, 54 },
		{ 394, 378, 54 },
		{ 394, 379, 54 },
		{ 394, 380, 54 },
		{ 394, 381, 54 },
		{ 394, 382, 54 },
		{ 394, 383, 54 },
		{ 394, 384, 54 },
		{ 394, 385, 54 },
		{ 394, 386, 54 },
		{ 394, 387, 54 },
		{ 394, 388, 54 },
		{ 394, 389, 54 },
		{ 394, 390, 54 },
		{ 394, 391, 54 },
		{ 394, 392, 54 },
		{ 394, 393, 54 },
		{ 394, 394, 54 },
		{ 394, 395, 54 },
		{ 394, 396, 54 },
		{ 394, 397, 54 },
		{ 394, 398, 54 },
		{ 394, 399, 54 },
		{ 394, 400, 54 },
		{ 394, 401, 54 },
		{ 394, 402, 54 },
		{ 394, 403, 54 },
		{ 394, 404, 54 },
		{ 394, 405, 54 },
		{ 394, 406, 54 },
		{ 394, 407, 54 },
		{ 394, 408, 54 },
		{ 394, 409, 54 },
		{ 394, 410, 54 },
		{ 394, 411, 54 },
		{ 394, 412, 54 },
		{ 394, 413, 54 },
		{ 395, 378, 54 },
		{ 395, 379, 54 },
		{ 395, 380, 54 },
		{ 395, 381, 54 },
		{ 395, 382, 54 },
		{ 395, 383, 54 },
		{ 395, 384, 54 },
		{ 395, 385, 54 },
		{ 395, 386, 54 },
		{ 395, 387, 54 },
		{ 395, 388, 54 },
		{ 395, 389, 54 },
		{ 395, 390, 54 },
		{ 395, 391, 54 },
		{ 395, 392, 54 },
		{ 395, 393, 54 },
		{ 395, 394, 54 },
		{ 395, 395, 54 },
		{ 395, 396, 54 },
		{ 395, 397, 54 },
		{ 395, 398, 54 },
		{ 395, 399, 54 },
		{ 395, 400, 54 },
		{ 395, 401, 54 },
		{ 395, 402, 54 },
		{ 395, 403, 54 },
		{ 395, 404, 54 },
		{ 395, 405, 54 },
		{ 395, 406, 54 },
		{ 395, 407, 54 },
		{ 395, 408, 54 },
		{ 395, 409, 54 },
		{ 395, 410, 54 },
		{ 395, 411, 54 },
		{ 395, 412, 54 },
		{ 395, 413, 54 },
		{ 395, 414, 54 },
		{ 396, 379, 54 },
		{ 396, 380, 54 },
		{ 396, 381, 54 },
		{ 396, 382, 54 },
		{ 396, 383, 54 },
		{ 396, 384, 54 },
		{ 396, 385, 54 },
		{ 396, 386, 54 },
		{ 396, 387, 54 },
		{ 396, 388, 54 },
		{ 396, 389, 54 },
		{ 396, 390, 54 },
		{ 396, 391, 54 },
		{ 396, 392, 54 },
		{ 396, 393, 54 },
		{ 396, 394, 54 },
		{ 396, 395, 54 },
		{ 396, 396, 54 },
		{ 396, 397, 54 },
		{ 396, 398, 54 },
		{ 396, 399, 54 },
		{ 396, 400, 54 },
		{ 396, 401, 54 },
		{ 396, 402, 54 },
		{ 396, 403, 54 },
		{ 396, 404, 54 },
		{ 396, 405, 54 },
		{ 396, 406, 54 },
		{ 396, 407, 54 },
		{ 396, 408, 54 },
		{ 396, 409, 54 },
		{ 396, 410, 54 },
		{ 396, 411, 54 },
		{ 396, 412, 54 },
		{ 396, 413, 54 },
		{ 396, 414, 54 },
		{ 396, 415, 54 },
		{ 397, 380, 54 },
		{ 397, 381, 54 },
		{ 397, 382, 54 },
		{ 397, 383, 54 },
		{ 397, 384, 54 },
		{ 397, 385, 54 },
		{ 397, 386, 54 },
		{ 397, 387, 54 },
		{ 397, 388, 54 },
		{ 397, 389, 54 },
		{ 397, 390, 54 },
		{ 397, 391, 54 },
		{ 397, 392, 54 },
		{ 397, 393, 54 },
		{ 397, 394, 54 },
		{ 397, 395, 54 },
		{ 397, 396, 54 },
		{ 397, 397, 54 },
		{ 397, 398, 54 },
		{ 397, 399, 54 },
		{ 397, 400, 54 },
		{ 397, 401, 54 },
		{ 397, 402, 54 },
		{ 397, 403, 54 },
		{ 397, 404, 54 },
		{ 397, 405, 54 },
		{ 397, 406, 54 },
		{ 397, 407, 54 },
		{ 397, 408, 54 },
		{ 397, 409, 54 },
		{ 397, 410, 54 },
		{ 397, 411, 54 },
		{ 397, 412, 54 },
		{ 397, 413, 54 },
		{ 397, 414, 54 },
		{ 397, 415, 54 },
		{ 397, 416, 54 },
		{ 398, 381, 54 },
		{ 398, 382, 54 },
		{ 398, 383, 54 },
		{ 398, 384, 54 },
		{ 398, 385, 54 },
		{ 398, 386, 54 },
		{ 398, 387, 54 },
		{ 398, 388, 54 },
		{ 398, 389, 54 },
		{ 398, 390, 54 },
		{ 398, 391, 54 },
		{ 398, 392, 54 },
		{ 398, 393, 54 },
		{ 398, 394, 54 },
		{ 398, 395, 54 },
		{ 398, 396, 54 },
		{ 398, 397, 54 },
		{ 398, 398, 54 },
		{ 398, 399, 54 },
		{ 398, 400, 54 },
		{ 398, 401, 54 },
		{ 398, 402, 54 },
		{ 398, 403, 54 },
		{ 398, 404, 54 },
		{ 398, 405, 54 },
		{ 398, 406, 54 },
		{ 398, 407, 54 },
		{ 398, 408, 54 },
		{ 398, 409, 54 },
		{ 398, 410, 54 },
		{ 398, 411, 54 },
		{ 398, 412, 54 },
		{ 398, 413, 54 },
		{ 398, 414, 54 },
		{ 398, 415, 54 },
		{ 398, 416, 54 },
		{ 398, 417, 54 },
		{ 398, 418, 54 },
		{ 399, 382, 54 },
		{ 399, 383, 54 },
		{ 399, 384, 54 },
		{ 399, 385, 54 },
		{ 399, 386, 54 },
		{ 399, 387, 54 },
		{ 399, 388, 54 },
		{ 399, 389, 54 },
		{ 399, 390, 54 },
		{ 399, 391, 54 },
		{ 399, 392, 54 },
		{ 399, 393, 54 },
		{ 399, 394, 54 },
		{ 399, 395, 54 },
		{ 399, 396, 54 },
		{ 399, 397, 54 },
		{ 399, 398, 54 },
		{ 399, 399, 54 },
		{ 399, 400, 54 },
		{ 399, 401, 54 },
		{ 399, 402, 54 },
		{ 399, 403, 54 },
		{ 399, 404, 54 },
		{ 399, 405, 54 },
		{ 399, 406, 54 },
		{ 399, 407, 54 },
		{ 399, 408, 54 },
		{ 399, 409, 54 },
		{ 399, 410, 54 },
		{ 399, 411, 54 },
		{ 399, 412, 54 },
		{ 399, 413, 54 },
		{ 399, 414, 54 },
		{ 399, 415, 54 },
		{ 399, 416, 54 },
		{ 399, 417, 54 },
		{ 399, 418, 54 },
		{ 399, 419, 54 },
		{ 400, 383, 54 },
		{ 400, 384, 54 },
		{ 400, 385, 54 },
		{ 400, 386, 54 },
		{ 400, 387, 54 },
		{ 400, 388, 54 },
		{ 400, 389, 54 },
		{ 400, 390, 54 },
		{ 400, 391, 54 },
		{ 400, 392, 54 },
		{ 400, 393, 54 },
		{ 400, 394, 54 },
		{ 400, 395, 54 },
		{ 400, 396, 54 },
		{ 400, 397, 54 },
		{ 400, 398, 54 },
		{ 400, 399, 54 },
		{ 400, 400, 54 },
		{ 400, 401, 54 },
		{ 400, 402, 54 },
		{ 400, 403, 54 },
		{ 400, 404, 54 },
		{ 400, 405, 54 },
		{ 400, 406, 54 },
		{ 400, 407, 54 },
		{ 400, 408, 54 },
		{ 400, 409, 54 },
		{ 400, 410, 54 },
		{ 400, 411, 54 },
		{ 400, 412, 54 },
		{ 400, 413, 54 },
		{ 400, 414, 54 },
		{ 400, 415, 54 },
		{ 400, 416, 54 },
		{ 400, 417, 54 },
		{ 400, 418, 54 },
		{ 400, 419, 54 },
		{ 400, 420, 54 },
		{ 401, 384, 54 },
		{ 401, 385, 54 },
		{ 401, 386, 54 },
		{ 401, 387, 54 },
		{ 401, 388, 54 },
		{ 401, 389, 54 },
		{ 401, 390, 54 },
		{ 401, 391, 54 },
		{ 401, 392, 54 },
		{ 401, 393, 54 },
		{ 401, 394, 54 },
		{ 401, 395, 54 },
		{ 401, 396, 54 },
		{ 401, 397, 54 },
		{ 401, 398, 54 },
		{ 401, 399, 54 },
		{ 401, 400, 54 },
		{ 401, 401, 54 },
		{ 401, 402, 54 },
		{ 401, 403, 54 },
		{ 401, 404, 54 },
		{ 401, 405, 54 },
		{ 401, 406, 54 },
		{ 401, 407, 54 },
		{ 401, 408, 54 },
		{ 401, 409, 54 },
		{ 401, 410, 54 },
		{ 401, 411, 54 },
		{ 401, 412, 54 },
		{ 401, 413, 54 },
		{ 401, 414, 54 },
		{ 401, 415, 54 },
		{ 401, 416, 54 },
		{ 401, 417, 54 },
		{ 401, 418, 54 },
		{ 401, 419, 54 },
		{ 401, 420, 54 },
		{ 401, 421, 54 },
		{ 402, 384, 54 },
		{ 402, 385, 54 },
		{ 402, 386, 54 },
		{ 402, 387, 54 },
		{ 402, 388, 54 },
		{ 402, 389, 54 },
		{ 402, 390, 54 },
		{ 402, 391, 54 },
		{ 402, 392, 54 },
		{ 402, 393, 54 },
		{ 402, 394, 54 },
		{ 402, 395, 54 },
		{ 402, 396, 54 },
		{ 402, 397, 54 },
		{ 402, 398, 54 },
		{ 402, 399, 54 },
		{ 402, 400, 54 },
		{ 402, 401, 54 },
		{ 402, 402, 54 },
		{ 402, 403, 54 },
		{ 402, 404, 54 },
		{ 402, 405, 54 },
		{ 402, 406, 54 },
		{ 402, 407, 54 },
		{ 402, 408, 54 },
		{ 402, 409, 54 },
		{ 402, 410, 54 },
		{ 402, 411, 54 },
		{ 402, 412, 54 },
		{ 402, 413, 54 },
		{ 402, 414, 54 },
		{ 402, 415, 54 },
		{ 402, 416, 54 },
		{ 402, 417, 54 },
		{ 402, 418, 54 },
		{ 402, 419, 54 },
		{ 402, 420, 54 },
		{ 402, 421, 54 },
		{ 402, 422, 54 },
		{ 403, 385, 54 },
		{ 403, 386, 54 },
		{ 403, 387, 54 },
		{ 403, 388, 54 },
		{ 403, 389, 54 },
		{ 403, 390, 54 },
		{ 403, 391, 54 },
		{ 403, 392, 54 },
		{ 403, 393, 54 },
		{ 403, 394, 54 },
		{ 403, 395, 54 },
		{ 403, 396, 54 },
		{ 403, 397, 54 },
		{ 403, 398, 54 },
		{ 403, 399, 54 },
		{ 403, 400, 54 },
		{ 403, 401, 54 },
		{ 403, 402, 54 },
		{ 403, 403, 54 },
		{ 403, 404, 54 },
		{ 403, 405, 54 },
		{ 403, 406, 54 },
		{ 403, 407, 54 },
		{ 403, 408, 54 },
		{ 403, 409, 54 },
		{ 403, 410, 54 },
		{ 403, 411, 54 },
		{ 403, 412, 54 },
		{ 403, 413, 54 },
		{ 403, 414, 54 },
		{ 403, 415, 54 },
		{ 403, 416, 54 },
		{ 403, 417, 54 },
		{ 403, 418, 54 },
		{ 403, 419, 54 },
		{ 403, 420, 54 },
		{ 403, 421, 54 },
		{ 403, 422, 54 },
		{ 403, 423, 54 },
		{ 404, 386, 54 },
		{ 404, 387, 54 },
		{ 404, 388, 54 },
		{ 404, 389, 54 },
		{ 404, 390, 54 },
		{ 404, 391, 54 },
		{ 404, 392, 54 },
		{ 404, 393, 54 },
		{ 404, 394, 54 },
		{ 404, 395, 54 },
		{ 404, 396, 54 },
		{ 404, 397, 54 },
		{ 404, 398, 54 },
		{ 404, 399, 54 },
		{ 404, 400, 54 },
		{ 404, 401, 54 },
		{ 404, 402, 54 },
		{ 404, 403, 54 },
		{ 404, 404, 54 },
		{ 404, 405, 54 },
		{ 404, 406, 54 },
		{ 404, 407, 54 },
		{ 404, 408, 54 },
		{ 404, 409, 54 },
		{ 404, 410, 54 },
		{ 404, 411, 54 },
		{ 404, 412, 54 },
		{ 404, 413, 54 },
		{ 404, 414, 54 },
		{ 404, 415, 54 },
		{ 404, 416, 54 },
		{ 404, 417, 54 },
		{ 404, 418, 54 },
		{ 404, 419, 54 },
		{ 404, 420, 54 },
		{ 404, 421, 54 },
		{ 404, 422, 54 },
		{ 404, 423, 54 },
		{ 404, 424, 54 },
		{ 405, 387, 54 },
		{ 405, 388, 54 },
		{ 405, 389, 54 },
		{ 405, 390, 54 },
		{ 405, 391, 54 },
		{ 405, 392, 54 },
		{ 405, 393, 54 },
		{ 405, 394, 54 },
		{ 405, 395, 54 },
		{ 405, 396, 54 },
		{ 405, 397, 54 },
		{ 405, 398, 54 },
		{ 405, 399, 54 },
		{ 405, 400, 54 },
		{ 405, 401, 54 },
		{ 405, 402, 54 },
		{ 405, 403, 54 },
		{ 405, 404, 54 },
		{ 405, 405, 54 },
		{ 405, 406, 54 },
		{ 405, 407, 54 },
		{ 405, 408, 54 },
		{ 405, 409, 54 },
		{ 405, 410, 54 },
		{ 405, 411, 54 },
		{ 405, 412, 54 },
		{ 405, 413, 54 },
		{ 405, 414, 54 },
		{ 405, 415, 54 },
		{ 405, 416, 54 },
		{ 405, 417, 54 },
		{ 405, 418, 54 },
		{ 405, 419, 54 },
		{ 405, 420, 54 },
		{ 405, 421, 54 },
		{ 405, 422, 54 },
		{ 405, 423, 54 },
		{ 405, 424, 54 },
		{ 405, 425, 54 },
		{ 405, 426, 54 },
		{ 406, 388, 54 },
		{ 406, 389, 54 },
		{ 406, 390, 54 },
		{ 406, 391, 54 },
		{ 406, 392, 54 },
		{ 406, 393, 54 },
		{ 406, 394, 54 },
		{ 406, 395, 54 },
		{ 406, 396, 54 },
		{ 406, 397, 54 },
		{ 406, 398, 54 },
		{ 406, 399, 54 },
		{ 406, 400, 54 },
		{ 406, 401, 54 },
		{ 406, 402, 54 },
		{ 406, 403, 54 },
		{ 406, 404, 54 },
		{ 406, 405, 54 },
		{ 406, 406, 54 },
		{ 406, 407, 54 },
		{ 406, 408, 54 },
		{ 406, 409, 54 },
		{ 406, 410, 54 },
		{ 406, 411, 54 },
		{ 406, 412, 54 },
		{ 406, 413, 54 },
		{ 406, 414, 54 },
		{ 406, 415, 54 },
		{ 406, 416, 54 },
		{ 406, 417, 54 },
		{ 406, 418, 54 },
		{ 406, 419, 54 },
		{ 406, 420, 54 },
		{ 406, 421, 54 },
		{ 406, 422, 54 },
		{ 406, 423, 54 },
		{ 406, 424, 54 },
		{ 406, 425, 54 },
		{ 406, 426, 54 },
		{ 406, 427, 54 },
		{ 407, 389, 54 },
		{ 407, 390, 54 },
		{ 407, 391, 54 },
		{ 407, 392, 54 },
		{ 407, 393, 54 },
		{ 407, 394, 54 },
		{ 407, 395, 54 },
		{ 407, 396, 54 },
		{ 407, 397, 54 },
		{ 407, 398, 54 },
		{ 407, 399, 54 },
		{ 407, 400, 54 },
		{ 407, 401, 54 },
		{ 407, 402, 54 },
		{ 407, 403, 54 },
		{ 407, 404, 54 },
		{ 407, 405, 54 },
		{ 407, 406, 54 },
		{ 407, 407, 54 },
		{ 407, 408, 54 },
		{ 407, 409, 54 },
		{ 407, 410, 54 },
		{ 407, 411, 54 },
		{ 407, 412, 54 },
		{ 407, 413, 54 },
		{ 407, 414, 54 },
		{ 407, 415, 54 },
		{ 407, 416, 54 },
		{ 407, 417, 54 },
		{ 407, 418, 54 },
		{ 407, 419, 54 },
		{ 407, 420, 54 },
		{ 407, 421, 54 },
		{ 407, 422, 54 },
		{ 407, 423, 54 },
		{ 407, 424, 54 },
		{ 407, 425, 54 },
		{ 407, 426, 54 },
		{ 407, 427, 54 },
		{ 407, 428, 54 },
		{ 408, 390, 54 },
		{ 408, 391, 54 },
		{ 408, 392, 54 },
		{ 408, 393, 54 },
		{ 408, 394, 54 },
		{ 408, 395, 54 },
		{ 408, 396, 54 },
		{ 408, 397, 54 },
		{ 408, 398, 54 },
		{ 408, 399, 54 },
		{ 408, 400, 54 },
		{ 408, 401, 54 },
		{ 408, 402, 54 },
		{ 408, 403, 54 },
		{ 408, 404, 54 },
		{ 408, 405, 54 },
		{ 408, 406, 54 },
		{ 408, 407, 54 },
		{ 408, 408, 54 },
		{ 408, 409, 54 },
		{ 408, 410, 54 },
		{ 408, 411, 54 },
		{ 408, 412, 54 },
		{ 408, 413, 54 },
		{ 408, 414, 54 },
		{ 408, 415, 54 },
		{ 408, 416, 54 },
		{ 408, 417, 54 },
		{ 408, 418, 54 },
		{ 408, 419, 54 },
		{ 408, 420, 54 },
		{ 408, 421, 54 },
		{ 408, 422, 54 },
		{ 408, 423, 54 },
		{ 408, 424, 54 },
		{ 408, 425, 54 },
		{ 408, 426, 54 },
		{ 408, 427, 54 },
		{ 408, 428, 54 },
		{ 408, 429, 54 },
		{ 409, 391, 54 },
		{ 409, 392, 54 },
		{ 409, 393, 54 },
		{ 409, 394, 54 },
		{ 409, 395, 54 },
		{ 409, 396, 54 },
		{ 409, 397, 54 },
		{ 409, 398, 54 },
		{ 409, 399, 54 },
		{ 409, 400, 54 },
		{ 409, 401, 54 },
		{ 409, 402, 54 },
		{ 409, 403, 54 },
		{ 409, 404, 54 },
		{ 409, 405, 54 },
		{ 409, 406, 54 },
		{ 409, 407, 54 },
		{ 409, 408, 54 },
		{ 409, 409, 54 },
		{ 409, 410, 54 },
		{ 409, 411, 54 },
		{ 409, 412, 54 },
		{ 409, 413, 54 },
		{ 409, 414, 54 },
		{ 409, 415, 54 },
		{ 409, 416, 54 },
		{ 409, 417, 54 },
		{ 409, 418, 54 },
		{ 409, 419, 54 },
		{ 409, 420, 54 },
		{ 409, 421, 54 },
		{ 409, 422, 54 },
		{ 409, 423, 54 },
		{ 409, 424, 54 },
		{ 409, 425, 54 },
		{ 409, 426, 54 },
		{ 409, 427, 54 },
		{ 409, 428, 54 },
		{ 409, 429, 54 },
		{ 409, 430, 54 },
		{ 410, 391, 54 },
		{ 410, 392, 54 },
		{ 410, 393, 54 },
		{ 410, 394, 54 },
		{ 410, 395, 54 },
		{ 410, 396, 54 },
		{ 410, 397, 54 },
		{ 410, 398, 54 },
		{ 410, 399, 54 },
		{ 410, 400, 54 },
		{ 410, 401, 54 },
		{ 410, 402, 54 },
		{ 410, 403, 54 },
		{ 410, 404, 54 },
		{ 410, 405, 54 },
		{ 410, 406, 54 },
		{ 410, 407, 54 },
		{ 410, 408, 54 },
		{ 410, 409, 54 },
		{ 410, 410, 54 },
		{ 410, 411, 54 },
		{ 410, 412, 54 },
		{ 410, 413, 54 },
		{ 410, 414, 54 },
		{ 410, 415, 54 },
		{ 410, 416, 54 },
		{ 410, 417, 54 },
		{ 410, 418, 54 },
		{ 410, 419, 54 },
		{ 410, 420, 54 },
		{ 410, 421, 54 },
		{ 410, 422, 54 },
		{ 410, 423, 54 },
		{ 410, 424, 54 },
		{ 410, 425, 54 },
		{ 410, 426, 54 },
		{ 410, 427, 54 },
		{ 410, 428, 54 },
		{ 410, 429, 54 },
		{ 410, 430, 54 },
		{ 410, 431, 54 },
		{ 411, 392, 54 },
		{ 411, 393, 54 },
		{ 411, 394, 54 },
		{ 411, 395, 54 },
		{ 411, 396, 54 },
		{ 411, 397, 54 },
		{ 411, 398, 54 },
		{ 411, 399, 54 },
		{ 411, 400, 54 },
		{ 411, 401, 54 },
		{ 411, 402, 54 },
		{ 411, 403, 54 },
		{ 411, 404, 54 },
		{ 411, 405, 54 },
		{ 411, 406, 54 },
		{ 411, 407, 54 },
		{ 411, 408, 54 },
		{ 411, 409, 54 },
		{ 411, 410, 54 },
		{ 411, 411, 54 },
		{ 411, 412, 54 },
		{ 411, 413, 54 },
		{ 411, 414, 54 },
		{ 411, 415, 54 },
		{ 411, 416, 54 },
		{ 411, 417, 54 },
		{ 411, 418, 54 },
		{ 411, 419, 54 },
		{ 411, 420, 54 },
		{ 411, 421, 54 },
		{ 411, 422, 54 },
		{ 411, 423, 54 },
		{ 411, 424, 54 },
		{ 411, 425, 54 },
		{ 411, 426, 54 },
		{ 411, 427, 54 },
		{ 411, 428, 54 },
		{ 411, 429, 54 },
		{ 411, 430, 54 },
		{ 411, 431, 54 },
		{ 411, 432, 54 },
		{ 412, 393, 54 },
		{ 412, 394, 54 },
		{ 412, 395, 54 },
		{ 412, 396, 54 },
		{ 412, 397, 54 },
		{ 412, 398, 54 },
		{ 412, 399, 54 },
		{ 412, 400, 54 },
		{ 412, 401, 54 },
		{ 412, 402, 54 },
		{ 412, 403, 54 },
		{ 412, 404, 54 },
		{ 412, 405, 54 },
		{ 412, 406, 54 },
		{ 412, 407, 54 },
		{ 412, 408, 54 },
		{ 412, 409, 54 },
		{ 412, 410, 54 },
		{ 412, 411, 54 },
		{ 412, 412, 54 },
		{ 412, 413, 54 },
		{ 412, 414, 54 },
		{ 412, 415, 54 },
		{ 412, 416, 54 },
		{ 412, 417, 54 },
		{ 412, 418, 54 },
		{ 412, 419, 54 },
		{ 412, 420, 54 },
		{ 412, 421, 54 },
		{ 412, 422, 54 },
		{ 412, 423, 54 },
		{ 412, 424, 54 },
		{ 412, 425, 54 },
		{ 412, 426, 54 },
		{ 412, 427, 54 },
		{ 412, 428, 54 },
		{ 412, 429, 54 },
		{ 412, 430, 54 },
		{ 412, 431, 54 },
		{ 412, 432, 54 },
		{ 412, 433, 54 },
		{ 412, 434, 54 },
		{ 413, 394, 54 },
		{ 413, 395, 54 },
		{ 413, 396, 54 },
		{ 413, 397, 54 },
		{ 413, 398, 54 },
		{ 413, 399, 54 },
		{ 413, 400, 54 },
		{ 413, 401, 54 },
		{ 413, 402, 54 },
		{ 413, 403, 54 },
		{ 413, 404, 54 },
		{ 413, 405, 54 },
		{ 413, 406, 54 },
		{ 413, 407, 54 },
		{ 413, 408, 54 },
		{ 413, 409, 54 },
		{ 413, 410, 54 },
		{ 413, 411, 54 },
		{ 413, 412, 54 },
		{ 413, 413, 54 },
		{ 413, 414, 54 },
		{ 413, 415, 54 },
		{ 413, 416, 54 },
		{ 413, 417, 54 },
		{ 413, 418, 54 },
		{ 413, 419, 54 },
		{ 413, 420, 54 },
		{ 413, 421, 54 },
		{ 413, 422, 54 },
		{ 413, 423, 54 },
		{ 413, 424, 54 },
		{ 413, 425, 54 },
		{ 413, 426, 54 },
		{ 413, 427, 54 },
		{ 413, 428, 54 },
		{ 413, 429, 54 },
		{ 413, 430, 54 },
		{ 413, 431, 54 },
		{ 413, 432, 54 },
		{ 413, 433, 54 },
		{ 413, 434, 54 },
		{ 413, 435, 54 },
		{ 414, 395, 54 },
		{ 414, 396, 54 },
		{ 414, 397, 54 },
		{ 414, 398, 54 },
		{ 414, 399, 54 },
		{ 414, 400, 54 },
		{ 414, 401, 54 },
		{ 414, 402, 54 },
		{ 414, 403, 54 },
		{ 414, 404, 54 },
		{ 414, 405, 54 },
		{ 414, 406, 54 },
		{ 414, 407, 54 },
		{ 414, 408, 54 },
		{ 414, 409, 54 },
		{ 414, 410, 54 },
		{ 414, 411, 54 },
		{ 414, 412, 54 },
		{ 414, 413, 54 },
		{ 414, 414, 54 },
		{ 414, 415, 54 },
		{ 414, 416, 54 },
		{ 414, 417, 54 },
		{ 414, 418, 54 },
		{ 414, 419, 54 },
		{ 414, 420, 54 },
		{ 414, 421, 54 },
		{ 414, 422, 54 },
		{ 414, 423, 54 },
		{ 414, 424, 54 },
		{ 414, 425, 54 },
		{ 414, 426, 54 },
		{ 414, 427, 54 },
		{ 414, 428, 54 },
		{ 414, 429, 54 },
		{ 414, 430, 54 },
		{ 414, 431, 54 },
		{ 414, 432, 54 },
		{ 414, 433, 54 },
		{ 414, 434, 54 },
		{ 414, 435, 54 },
		{ 414, 436, 54 },
		{ 415, 396, 54 },
		{ 415, 397, 54 },
		{ 415, 398, 54 },
		{ 415, 399, 54 },
		{ 415, 400, 54 },
		{ 415, 401, 54 },
		{ 415, 402, 54 },
		{ 415, 403, 54 },
		{ 415, 404, 54 },
		{ 415, 405, 54 },
		{ 415, 406, 54 },
		{ 415, 407, 54 },
		{ 415, 408, 54 },
		{ 415, 409, 54 },
		{ 415, 410, 54 },
		{ 415, 411, 54 },
		{ 415, 412, 54 },
		{ 415, 413, 54 },
		{ 415, 414, 54 },
		{ 415, 415, 54 },
		{ 415, 416, 54 },
		{ 415, 417, 54 },
		{ 415, 418, 54 },
		{ 415, 419, 54 },
		{ 415, 420, 54 },
		{ 415, 421, 54 },
		{ 415, 422, 54 },
		{ 415, 423, 54 },
		{ 415, 424, 54 },
		{ 415, 425, 54 },
		{ 415, 426, 54 },
		{ 415, 427, 54 },
		{ 415, 428, 54 },
		{ 415, 429, 54 },
		{ 415, 430, 54 },
		{ 415, 431, 54 },
		{ 415, 432, 54 },
		{ 415, 433, 54 },
		{ 415, 434, 54 },
		{ 415, 435, 54 },
		{ 415, 436, 54 },
		{ 415, 437, 54 },
		{ 416, 397, 54 },
		{ 416, 398, 54 },
		{ 416, 399, 54 },
		{ 416, 400, 54 },
		{ 416, 401, 54 },
		{ 416, 402, 54 },
		{ 416, 403, 54 },
		{ 416, 404, 54 },
		{ 416, 405, 54 },
		{ 416, 406, 54 },
		{ 416, 407, 54 },
		{ 416, 408, 54 },
		{ 416, 409, 54 },
		{ 416, 410, 54 },
		{ 416, 411, 54 },
		{ 416, 412, 54 },
		{ 416, 413, 54 },
		{ 416, 414, 54 },
		{ 416, 415, 54 },
		{ 416, 416, 54 },
		{ 416, 417, 54 },
		{ 416, 418, 54 },
		{ 416, 419, 54 },
		{ 416, 420, 54 },
		{ 416, 421, 54 },
		{ 416, 422, 54 },
		{ 416, 423, 54 },
		{ 416, 424, 54 },
		{ 416, 425, 54 },
		{ 416, 426, 54 },
		{ 416, 427, 54 },
		{ 416, 428, 54 },
		{ 416, 429, 54 },
		{ 416, 430, 54 },
		{ 416, 431, 54 },
		{ 416, 432, 54 },
		{ 416, 433, 54 },
		{ 416, 434, 54 },
		{ 416, 435, 54 },
		{ 416, 436, 54 },
		{ 416, 437, 54 },
		{ 416, 438, 54 },
		{ 417, 398, 54 },
		{ 417, 399, 54 },
		{ 417, 400, 54 },
		{ 417, 401, 54 },
		{ 417, 402, 54 },
		{ 417, 403, 54 },
		{ 417, 404, 54 },
		{ 417, 405, 54 },
		{ 417, 406, 54 },
		{ 417, 407, 54 },
		{ 417, 408, 54 },
		{ 417, 409, 54 },
		{ 417, 410, 54 },
		{ 417, 411, 54 },
		{ 417, 412, 54 },
		{ 417, 413, 54 },
		{ 417, 414, 54 },
		{ 417, 415, 54 },
		{ 417, 416, 54 },
		{ 417, 417, 54 },
		{ 417, 418, 54 },
		{ 417, 419, 54 },
		{ 417, 420, 54 },
		{ 417, 421, 54 },
		{ 417, 422, 54 },
		{ 417, 423, 54 },
		{ 417, 424, 54 },
		{ 417, 425, 54 },
		{ 417, 426, 54 },
		{ 417, 427, 54 },
		{ 417, 428, 54 },
		{ 417, 429, 54 },
		{ 417, 430, 54 },
		{ 417, 431, 54 },
		{ 417, 432, 54 },
		{ 417, 433, 54 },
		{ 417, 434, 54 },
		{ 417, 435, 54 },
		{ 417, 436, 54 },
		{ 417, 437, 54 },
		{ 417, 438, 54 },
		{ 417, 439, 54 },
		{ 418, 398, 54 },
		{ 418, 399, 54 },
		{ 418, 400, 54 },
		{ 418, 401, 54 },
		{ 418, 402, 54 },
		{ 418, 403, 54 },
		{ 418, 404, 54 },
		{ 418, 405, 54 },
		{ 418, 406, 54 },
		{ 418, 407, 54 },
		{ 418, 408, 54 },
		{ 418, 409, 54 },
		{ 418, 410, 54 },
		{ 418, 411, 54 },
		{ 418, 412, 54 },
		{ 418, 413, 54 },
		{ 418, 414, 54 },
		{ 418, 415, 54 },
		{ 418, 416, 54 },
		{ 418, 417, 54 },
		{ 418, 418, 54 },
		{ 418, 419, 54 },
		{ 418, 420, 54 },
		{ 418, 421, 54 },
		{ 418, 422, 54 },
		{ 418, 423, 54 },
		{ 418, 424, 54 },
		{ 418, 425, 54 },
		{ 418, 426, 54 },
		{ 418, 427, 54 },
		{ 418, 428, 54 },
		{ 418, 429, 54 },
		{ 418, 430, 54 },
		{ 418, 431, 54 },
		{ 418, 432, 54 },
		{ 418, 433, 54 },
		{ 418, 434, 54 },
		{ 418, 435, 54 },
		{ 418, 436, 54 },
		{ 418, 437, 54 },
		{ 418, 438, 54 },
		{ 418, 439, 54 },
		{ 418, 440, 54 },
		{ 419, 399, 54 },
		{ 419, 400, 54 },
		{ 419, 401, 54 },
		{ 419, 402, 54 },
		{ 419, 403, 54 },
		{ 419, 404, 54 },
		{ 419, 405, 54 },
		{ 419, 406, 54 },
		{ 419, 407, 54 },
		{ 419, 408, 54 },
		{ 419, 409, 54 },
		{ 419, 410, 54 },
		{ 419, 411, 54 },
		{ 419, 412, 54 },
		{ 419, 413, 54 },
		{ 419, 414, 54 },
		{ 419, 415, 54 },
		{ 419, 416, 54 },
		{ 419, 417, 54 },
		{ 419, 418, 54 },
		{ 419, 419, 54 },
		{ 419, 420, 54 },
		{ 419, 421, 54 },
		{ 419, 422, 54 },
		{ 419, 423, 54 },
		{ 419, 424, 54 },
		{ 419, 425, 54 },
		{ 419, 426, 54 },
		{ 419, 427, 54 },
		{ 419, 428, 54 },
		{ 419, 429, 54 },
		{ 419, 430, 54 },
		{ 419, 431, 54 },
		{ 419, 432, 54 },
		{ 419, 433, 54 },
		{ 419, 434, 54 },
		{ 419, 435, 54 },
		{ 419, 436, 54 },
		{ 419, 437, 54 },
		{ 419, 438, 54 },
		{ 419, 439, 54 },
		{ 419, 440, 54 },
		{ 419, 441, 54 },
		{ 419, 442, 54 },
		{ 420, 400, 54 },
		{ 420, 401, 54 },
		{ 420, 402, 54 },
		{ 420, 403, 54 },
		{ 420, 404, 54 },
		{ 420, 405, 54 },
		{ 420, 406, 54 },
		{ 420, 407, 54 },
		{ 420, 408, 54 },
		{ 420, 409, 54 },
		{ 420, 410, 54 },
		{ 420, 411, 54 },
		{ 420, 412, 54 },
		{ 420, 413, 54 },
		{ 420, 414, 54 },
		{ 420, 415, 54 },
		{ 420, 416, 54 },
		{ 420, 417, 54 },
		{ 420, 418, 54 },
		{ 420, 419, 54 },
		{ 420, 420, 54 },
		{ 420, 421, 54 },
		{ 420, 422, 54 },
		{ 420, 423, 54 },
		{ 420, 424, 54 },
		{ 420, 425, 54 },
		{ 420, 426, 54 },
		{ 420, 427, 54 },
		{ 420, 428, 54 },
		{ 420, 429, 54 },
		{ 420, 430, 54 },
		{ 420, 431, 54 },
		{ 420, 432, 54 },
		{ 420, 433, 54 },
		{ 420, 434, 54 },
		{ 420, 435, 54 },
		{ 420, 436, 54 },
		{ 420, 437, 54 },
		{ 420, 438, 54 },
		{ 420, 439, 54 },
		{ 420, 440, 54 },
		{ 420, 441, 54 },
		{ 420, 442, 54 },
		{ 420, 443, 54 },
		{ 421, 401, 54 },
		{ 421, 402, 54 },
		{ 421, 403, 54 },
		{ 421, 404, 54 },
		{ 421, 405, 54 },
		{ 421, 406, 54 },
		{ 421, 407, 54 },
		{ 421, 408, 54 },
		{ 421, 409, 54 },
		{ 421, 410, 54 },
		{ 421, 411, 54 },
		{ 421, 412, 54 },
		{ 421, 413, 54 },
		{ 421, 414, 54 },
		{ 421, 415, 54 },
		{ 421, 416, 54 },
		{ 421, 417, 54 },
		{ 421, 418, 54 },
		{ 421, 419, 54 },
		{ 421, 420, 54 },
		{ 421, 421, 54 },
		{ 421, 422, 54 },
		{ 421, 423, 54 },
		{ 421, 424, 54 },
		{ 421, 425, 54 },
		{ 421, 426, 54 },
		{ 421, 427, 54 },
		{ 421, 428, 54 },
		{ 421, 429, 54 },
		{ 421, 430, 54 },
		{ 421, 431, 54 },
		{ 421, 432, 54 },
		{ 421, 433, 54 },
		{ 421, 434, 54 },
		{ 421, 435, 54 },
		{ 421, 436, 54 },
		{ 421, 437, 54 },
		{ 421, 438, 54 },
		{ 421, 439, 54 },
		{ 421, 440, 54 },
		{ 421, 441, 54 },
		{ 421, 442, 54 },
		{ 421, 443, 54 },
		{ 421, 444, 54 },
		{ 422, 402, 54 },
		{ 422, 403, 54 },
		{ 422, 404, 54 },
		{ 422, 405, 54 },
		{ 422, 406, 54 },
		{ 422, 407, 54 },
		{ 422, 408, 54 },
		{ 422, 409, 54 },
		{ 422, 410, 54 },
		{ 422, 411, 54 },
		{ 422, 412, 54 },
		{ 422, 413, 54 },
		{ 422, 414, 54 },
		{ 422, 415, 54 },
		{ 422, 416, 54 },
		{ 422, 417, 54 },
		{ 422, 418, 54 },
		{ 422, 419, 54 },
		{ 422, 420, 54 },
		{ 422, 421, 54 },
		{ 422, 422, 54 },
		{ 422, 423, 54 },
		{ 422, 424, 54 },
		{ 422, 425, 54 },
		{ 422, 426, 54 },
		{ 422, 427, 54 },
		{ 422, 428, 54 },
		{ 422, 429, 54 },
		{ 422, 430, 54 },
		{ 422, 431, 54 },
		{ 422, 432, 54 },
		{ 422, 433, 54 },
		{ 422, 434, 54 },
		{ 422, 435, 54 },
		{ 422, 436, 54 },
		{ 422, 437, 54 },
		{ 422, 438, 54 },
		{ 422, 439, 54 },
		{ 422, 440, 54 },
		{ 422, 441, 54 },
		{ 422, 442, 54 },
		{ 422, 443, 54 },
		{ 422, 444, 54 },
		{ 422, 445, 54 },
		{ 423, 403, 54 },
		{ 423, 404, 54 },
		{ 423, 405, 54 },
		{ 423, 406, 54 },
		{ 423, 407, 54 },
		{ 423, 408, 54 },
		{ 423, 409, 54 },
		{ 423, 410, 54 },
		{ 423, 411, 54 },
		{ 423, 412, 54 },
		{ 423, 413, 54 },
		{ 423, 414, 54 },
		{ 423, 415, 54 },
		{ 423, 416, 54 },
		{ 423, 417, 54 },
		{ 423, 418, 54 },
		{ 423, 419, 54 },
		{ 423, 420, 54 },
		{ 423, 421, 54 },
		{ 423, 422, 54 },
		{ 423, 423, 54 },
		{ 423, 424, 54 },
		{ 423, 425, 54 },
		{ 423, 426, 54 },
		{ 423, 427, 54 },
		{ 423, 428, 54 },
		{ 423, 429, 54 },
		{ 423, 430, 54 },
		{ 423, 431, 54 },
		{ 423, 432, 54 },
		{ 423, 433, 54 },
		{ 423, 434, 54 },
		{ 423, 435, 54 },
		{ 423, 436, 54 },
		{ 423, 437, 54 },
		{ 423, 438, 54 },
		{ 423, 439, 54 },
		{ 423, 440, 54 },
		{ 423, 441, 54 },
		{ 423, 442, 54 },
		{ 423, 443, 54 },
		{ 423, 444, 54 },
		{ 423, 445, 54 },
		{ 423, 446, 54 },
		{ 424, 404, 54 },
		{ 424, 405, 54 },
		{ 424, 406, 54 },
		{ 424, 407, 54 },
		{ 424, 408, 54 },
		{ 424, 409, 54 },
		{ 424, 410, 54 },
		{ 424, 411, 54 },
		{ 424, 412, 54 },
		{ 424, 413, 54 },
		{ 424, 414, 54 },
		{ 424, 415, 54 },
		{ 424, 416, 54 },
		{ 424, 417, 54 },
		{ 424, 418, 54 },
		{ 424, 419, 54 },
		{ 424, 420, 54 },
		{ 424, 421, 54 },
		{ 424, 422, 54 },
		{ 424, 423, 54 },
		{ 424, 424, 54 },
		{ 424, 425, 54 },
		{ 424, 426, 54 },
		{ 424, 427, 54 },
		{ 424, 428, 54 },
		{ 424, 429, 54 },
		{ 424, 430, 54 },
		{ 424, 431, 54 },
		{ 424, 432, 54 },
		{ 424, 433, 54 },
		{ 424, 434, 54 },
		{ 424, 435, 54 },
		{ 424, 436, 54 },
		{ 424, 437, 54 },
		{ 424, 438, 54 },
		{ 424, 439, 54 },
		{ 424, 440, 54 },
		{ 424, 441, 54 },
		{ 424, 442, 54 },
		{ 424, 443, 54 },
		{ 424, 444, 54 },
		{ 424, 445, 54 },
		{ 424, 446, 54 },
		{ 424, 447, 54 },
		{ 425, 405, 54 },
		{ 425, 406, 54 },
		{ 425, 407, 54 },
		{ 425, 408, 54 },
		{ 425, 409, 54 },
		{ 425, 410, 54 },
		{ 425, 411, 54 },
		{ 425, 412, 54 },
		{ 425, 413, 54 },
		{ 425, 414, 54 },
		{ 425, 415, 54 },
		{ 425, 416, 54 },
		{ 425, 417, 54 },
		{ 425, 418, 54 },
		{ 425, 419, 54 },
		{ 425, 420, 54 },
		{ 425, 421, 54 },
		{ 425, 422, 54 },
		{ 425, 423, 54 },
		{ 425, 424, 54 },
		{ 425, 425, 54 },
		{ 425, 426, 54 },
		{ 425, 427, 54 },
		{ 425, 428, 54 },
		{ 425, 429, 54 },
		{ 425, 430, 54 },
		{ 425, 431, 54 },
		{ 425, 432, 54 },
		{ 425, 433, 54 },
		{ 425, 434, 54 },
		{ 425, 435, 54 },
		{ 425, 436, 54 },
		{ 425, 437, 54 },
		{ 425, 438, 54 },
		{ 425, 439, 54 },
		{ 425, 440, 54 },
		{ 425, 441, 54 },
		{ 425, 442, 54 },
		{ 425, 443, 54 },
		{ 425, 444, 54 },
		{ 425, 445, 54 },
		{ 425, 446, 54 },
		{ 425, 447, 54 },
		{ 425, 448, 54 },
		{ 426, 405, 54 },
		{ 426, 406, 54 },
		{ 426, 407, 54 },
		{ 426, 408, 54 },
		{ 426, 409, 54 },
		{ 426, 410, 54 },
		{ 426, 411, 54 },
		{ 426, 412, 54 },
		{ 426, 413, 54 },
		{ 426, 414, 54 },
		{ 426, 415, 54 },
		{ 426, 416, 54 },
		{ 426, 417, 54 },
		{ 426, 418, 54 },
		{ 426, 419, 54 },
		{ 426, 420, 54 },
		{ 426, 421, 54 },
		{ 426, 422, 54 },
		{ 426, 423, 54 },
		{ 426, 424, 54 },
		{ 426, 425, 54 },
		{ 426, 426, 54 },
		{ 426, 427, 54 },
		{ 426, 428, 54 },
		{ 426, 429, 54 },
		{ 426, 430, 54 },
		{ 426, 431, 54 },
		{ 426, 432, 54 },
		{ 426, 433, 54 },
		{ 426, 434, 54 },
		{ 426, 435, 54 },
		{ 426, 436, 54 },
		{ 426, 437, 54 },
		{ 426, 438, 54 },
		{ 426, 439, 54 },
		{ 426, 440, 54 },
		{ 426, 441, 54 },
		{ 426, 442, 54 },
		{ 426, 443, 54 },
		{ 426, 444, 54 },
		{ 426, 445, 54 },
		{ 426, 446, 54 },
		{ 426, 447, 54 },
		{ 426, 448, 54 },
		{ 426, 449, 54 },
		{ 426, 450, 54 },
		{ 427, 406, 54 },
		{ 427, 407, 54 },
		{ 427, 408, 54 },
		{ 427, 409, 54 },
		{ 427, 410, 54 },
		{ 427, 411, 54 },
		{ 427, 412, 54 },
		{ 427, 413, 54 },
		{ 427, 414, 54 },
		{ 427, 415, 54 },
		{ 427, 416, 54 },
		{ 427, 417, 54 },
		{ 427, 418, 54 },
		{ 427, 419, 54 },
		{ 427, 420, 54 },
		{ 427, 421, 54 },
		{ 427, 422, 54 },
		{ 427, 423, 54 },
		{ 427, 424, 54 },
		{ 427, 425, 54 },
		{ 427, 426, 54 },
		{ 427, 427, 54 },
		{ 427, 428, 54 },
		{ 427, 429, 54 },
		{ 427, 430, 54 },
		{ 427, 431, 54 },
		{ 427, 432, 54 },
		{ 427, 433, 54 },
		{ 427, 434, 54 },
		{ 427, 435, 54 },
		{ 427, 436, 54 },
		{ 427, 437, 54 },
		{ 427, 438, 54 },
		{ 427, 439, 54 },
		{ 427, 440, 54 },
		{ 427, 441, 54 },
		{ 427, 442, 54 },
		{ 427, 443, 54 },
		{ 427, 444, 54 },
		{ 427, 445, 54 },
		{ 427, 446, 54 },
		{ 427, 447, 54 },
		{ 427, 448, 54 },
		{ 427, 449, 54 },
		{ 427, 450, 54 },
		{ 427, 451, 54 },
		{ 428, 407, 54 },
		{ 428, 408, 54 },
		{ 428, 409, 54 },
		{ 428, 410, 54 },
		{ 428, 411, 54 },
		{ 428, 412, 54 },
		{ 428, 413, 54 },
		{ 428, 414, 54 },
		{ 428, 415, 54 },
		{ 428, 416, 54 },
		{ 428, 417, 54 },
		{ 428, 418, 54 },
		{ 428, 419, 54 },
		{ 428, 420, 54 },
		{ 428, 421, 54 },
		{ 428, 422, 54 },
		{ 428, 423, 54 },
		{ 428, 424, 54 },
		{ 428, 425, 54 },
		{ 428, 426, 54 },
		{ 428, 427, 54 },
		{ 428, 428, 54 },
		{ 428, 429, 54 },
		{ 428, 430, 54 },
		{ 428, 431, 54 },
		{ 428, 432, 54 },
		{ 428, 433, 54 },
		{ 428, 434, 54 },
		{ 428, 435, 54 },
		{ 428, 436, 54 },
		{ 428, 437, 54 },
		{ 428, 438, 54 },
		{ 428, 439, 54 },
		{ 428, 440, 54 },
		{ 428, 441, 54 },
		{ 428, 442, 54 },
		{ 428, 443, 54 },
		{ 428, 444, 54 },
		{ 428, 445, 54 },
		{ 428, 446, 54 },
		{ 428, 447, 54 },
		{ 428, 448, 54 },
		{ 428, 449, 54 },
		{ 428, 450, 54 },
		{ 428, 451, 54 },
		{ 428, 452, 54 },
		{ 429, 408, 54 },
		{ 429, 409, 54 },
		{ 429, 410, 54 },
		{ 429, 411, 54 },
		{ 429, 412, 54 },
		{ 429, 413, 54 },
		{ 429, 414, 54 },
		{ 429, 415, 54 },
		{ 429, 416, 54 },
		{ 429, 417, 54 },
		{ 429, 418, 54 },
		{ 429, 419, 54 },
		{ 429, 420, 54 },
		{ 429, 421, 54 },
		{ 429, 422, 54 },
		{ 429, 423, 54 },
		{ 429, 424, 54 },
		{ 429, 425, 54 },
		{ 429, 426, 54 },
		{ 429, 427, 54 },
		{ 429, 428, 54 },
		{ 429, 429, 54 },
		{ 429, 430, 54 },
		{ 429, 431, 54 },
		{ 429, 432, 54 },
		{ 429, 433, 54 },
		{ 429, 434, 54 },
		{ 429, 435, 54 },
		{ 429, 436, 54 },
		{ 429, 437, 54 },
		{ 429, 438, 54 },
		{ 429, 439, 54 },
		{ 429, 440, 54 },
		{ 429, 441, 54 },
		{ 429, 442, 54 },
		{ 429, 443, 54 },
		{ 429, 444, 54 },
		{ 429, 445, 54 },
		{ 429, 446, 54 },
		{ 429, 447, 54 },
		{ 429, 448, 54 },
		{ 429, 449, 54 },
		{ 429, 450, 54 },
		{ 429, 451, 54 },
		{ 429, 452, 54 },
		{ 429, 453, 54 },
		{ 430, 409, 54 },
		{ 430, 410, 54 },
		{ 430, 411, 54 },
		{ 430, 412, 54 },
		{ 430, 413, 54 },
		{ 430, 414, 54 },
		{ 430, 415, 54 },
		{ 430, 416, 54 },
		{ 430, 417, 54 },
		{ 430, 418, 54 },
		{ 430, 419, 54 },
		{ 430, 420, 54 },
		{ 430, 421, 54 },
		{ 430, 422, 54 },
		{ 430, 423, 54 },
		{ 430, 424, 54 },
		{ 430, 425, 54 },
		{ 430, 426, 54 },
		{ 430, 427, 54 },
		{ 430, 428, 54 },
		{ 430, 429, 54 },
		{ 430, 430, 54 },
		{ 430, 431, 54 },
		{ 430, 432, 54 },
		{ 430, 433, 54 },
		{ 430, 434, 54 },
		{ 430, 435, 54 },
		{ 430, 436, 54 },
		{ 430, 437, 54 },
		{ 430, 438, 54 },
		{ 430, 439, 54 },
		{ 430, 440, 54 },
		{ 430, 441, 54 },
		{ 430, 442, 54 },
		{ 430, 443, 54 },
		{ 430, 444, 54 },
		{ 430, 445, 54 },
		{ 430, 446, 54 },
		{ 430, 447, 54 },
		{ 430, 448, 54 },
		{ 430, 449, 54 },
		{ 430, 450, 54 },
		{ 430, 451, 54 },
		{ 430, 452, 54 },
		{ 430, 453, 54 },
		{ 430, 454, 54 },
		{ 431, 410, 54 },
		{ 431, 411, 54 },
		{ 431, 412, 54 },
		{ 431, 413, 54 },
		{ 431, 414, 54 },
		{ 431, 415, 54 },
		{ 431, 416, 54 },
		{ 431, 417, 54 },
		{ 431, 418, 54 },
		{ 431, 419, 54 },
		{ 431, 420, 54 },
		{ 431, 421, 54 },
		{ 431, 422, 54 },
		{ 431, 423, 54 },
		{ 431, 424, 54 },
		{ 431, 425, 54 },
		{ 431, 426, 54 },
		{ 431, 427, 54 },
		{ 431, 428, 54 },
		{ 431, 429, 54 },
		{ 431, 430, 54 },
		{ 431, 431, 54 },
		{ 431, 432, 54 },
		{ 431, 433, 54 },
		{ 431, 434, 54 },
		{ 431, 435, 54 },
		{ 431, 436, 54 },
		{ 431, 437, 54 },
		{ 431, 438, 54 },
		{ 431, 439, 54 },
		{ 431, 440, 54 },
		{ 431, 441, 54 },
		{ 431, 442, 54 },
		{ 431, 443, 54 },
		{ 431, 444, 54 },
		{ 431, 445, 54 },
		{ 431, 446, 54 },
		{ 431, 447, 54 },
		{ 431, 448, 54 },
		{ 431, 449, 54 },
		{ 431, 450, 54 },
		{ 431, 451, 54 },
		{ 431, 452, 54 },
		{ 431, 453, 54 },
		{ 431, 454, 54 },
		{ 431, 455, 54 },
		{ 432, 411, 54 },
		{ 432, 412, 54 },
		{ 432, 413, 54 },
		{ 432, 414, 54 },
		{ 432, 415, 54 },
		{ 432, 416, 54 },
		{ 432, 417, 54 },
		{ 432, 418, 54 },
		{ 432, 419, 54 },
		{ 432, 420, 54 },
		{ 432, 421, 54 },
		{ 432, 422, 54 },
		{ 432, 423, 54 },
		{ 432, 424, 54 },
		{ 432, 425, 54 },
		{ 432, 426, 54 },
		{ 432, 427, 54 },
		{ 432, 428, 54 },
		{ 432, 429, 54 },
		{ 432, 430, 54 },
		{ 432, 431, 54 },
		{ 432, 432, 54 },
		{ 432, 433, 54 },
		{ 432, 434, 54 },
		{ 432, 435, 54 },
		{ 432, 436, 54 },
		{ 432, 437, 54 },
		{ 432, 438, 54 },
		{ 432, 439, 54 },
		{ 432, 440, 54 },
		{ 432, 441, 54 },
		{ 432, 442, 54 },
		{ 432, 443, 54 },
		{ 432, 444, 54 },
		{ 432, 445, 54 },
		{ 432, 446, 54 },
		{ 432, 447, 54 },
		{ 432, 448, 54 },
		{ 432, 449, 54 },
		{ 432, 450, 54 },
		{ 432, 451, 54 },
		{ 432, 452, 54 },
		{ 432, 453, 54 },
		{ 432, 454, 54 },
		{ 432, 455, 54 },
		{ 432, 456, 54 },
		{ 433, 412, 54 },
		{ 433, 413, 54 },
		{ 433, 414, 54 },
		{ 433, 415, 54 },
		{ 433, 416, 54 },
		{ 433, 417, 54 },
		{ 433, 418, 54 },
		{ 433, 419, 54 },
		{ 433, 420, 54 },
		{ 433, 421, 54 },
		{ 433, 422, 54 },
		{ 433, 423, 54 },
		{ 433, 424, 54 },
		{ 433, 425, 54 },
		{ 433, 426, 54 },
		{ 433, 427, 54 },
		{ 433, 428, 54 },
		{ 433, 429, 54 },
		{ 433, 430, 54 },
		{ 433, 431, 54 },
		{ 433, 432, 54 },
		{ 433, 433, 54 },
		{ 433, 434, 54 },
		{ 433, 435, 54 },
		{ 433, 436, 54 },
		{ 433, 437, 54 },
		{ 433, 438, 54 },
		{ 433, 439, 54 },
		{ 433, 440, 54 },
		{ 433, 441, 54 },
		{ 433, 442, 54 },
		{ 433, 443, 54 },
		{ 433, 444, 54 },
		{ 433, 445, 54 },
		{ 433, 446, 54 },
		{ 433, 447, 54 },
		{ 433, 448, 54 },
		{ 433, 449, 54 },
		{ 433, 450, 54 },
		{ 433, 451, 54 },
		{ 433, 452, 54 },
		{ 433, 453, 54 },
		{ 433, 454, 54 },
		{ 433, 455, 54 },
		{ 433, 456, 54 },
		{ 433, 457, 54 },
		{ 433, 458, 54 },
		{ 434, 412, 54 },
		{ 434, 413, 54 },
		{ 434, 414, 54 },
		{ 434, 415, 54 },
		{ 434, 416, 54 },
		{ 434, 417, 54 },
		{ 434, 418, 54 },
		{ 434, 419, 54 },
		{ 434, 420, 54 },
		{ 434, 421, 54 },
		{ 434, 422, 54 },
		{ 434, 423, 54 },
		{ 434, 424, 54 },
		{ 434, 425, 54 },
		{ 434, 426, 54 },
		{ 434, 427, 54 },
		{ 434, 428, 54 },
		{ 434, 429, 54 },
		{ 434, 430, 54 },
		{ 434, 431, 54 },
		{ 434, 432, 54 },
		{ 434, 433, 54 },
		{ 434, 434, 54 },
		{ 434, 435, 54 },
		{ 434, 436, 54 },
		{ 434, 437, 54 },
		{ 434, 438, 54 },
		{ 434, 439, 54 },
		{ 434, 440, 54 },
		{ 434, 441, 54 },
		{ 434, 442, 54 },
		{ 434, 443, 54 },
		{ 434, 444, 54 },
		{ 434, 445, 54 },
		{ 434, 446, 54 },
		{ 434, 447, 54 },
		{ 434, 448, 54 },
		{ 434, 449, 54 },
		{ 434, 450, 54 },
		{ 434, 451, 54 },
		{ 434, 452, 54 },
		{ 434, 453, 54 },
		{ 434, 454, 54 },
		{ 434, 455, 54 },
		{ 434, 456, 54 },
		{ 434, 457, 54 },
		{ 434, 458, 54 },
		{ 434, 459, 54 },
		{ 435, 413, 54 },
		{ 435, 414, 54 },
		{ 435, 415, 54 },
		{ 435, 416, 54 },
		{ 435, 417, 54 },
		{ 435, 418, 54 },
		{ 435, 419, 54 },
		{ 435, 420, 54 },
		{ 435, 421, 54 },
		{ 435, 422, 54 },
		{ 435, 423, 54 },
		{ 435, 424, 54 },
		{ 435, 425, 54 },
		{ 435, 426, 54 },
		{ 435, 427, 54 },
		{ 435, 428, 54 },
		{ 435, 429, 54 },
		{ 435, 430, 54 },
		{ 435, 431, 54 },
		{ 435, 432, 54 },
		{ 435, 433, 54 },
		{ 435, 434, 54 },
		{ 435, 435, 54 },
		{ 435, 436, 54 },
		{ 435, 437, 54 },
		{ 435, 438, 54 },
		{ 435, 439, 54 },
		{ 435, 440, 54 },
		{ 435, 441, 54 },
		{ 435, 442, 54 },
		{ 435, 443, 54 },
		{ 435, 444, 54 },
		{ 435, 445, 54 },
		{ 435, 446, 54 },
		{ 435, 447, 54 },
		{ 435, 448, 54 },
		{ 435, 449, 54 },
		{ 435, 450, 54 },
		{ 435, 451, 54 },
		{ 435, 452, 54 },
		{ 435, 453, 54 },
		{ 435, 454, 54 },
		{ 435, 455, 54 },
		{ 435, 456, 54 },
		{ 435, 457, 54 },
		{ 435, 458, 54 },
		{ 435, 459, 54 },
		{ 435, 460, 54 },
		{ 436, 414, 54 },
		{ 436, 415, 54 },
		{ 436, 416, 54 },
		{ 436, 417, 54 },
		{ 436, 418, 54 },
		{ 436, 419, 54 },
		{ 436, 420, 54 },
		{ 436, 421, 54 },
		{ 436, 422, 54 },
		{ 436, 423, 54 },
		{ 436, 424, 54 },
		{ 436, 425, 54 },
		{ 436, 426, 54 },
		{ 436, 427, 54 },
		{ 436, 428, 54 },
		{ 436, 429, 54 },
		{ 436, 430, 54 },
		{ 436, 431, 54 },
		{ 436, 432, 54 },
		{ 436, 433, 54 },
		{ 436, 434, 54 },
		{ 436, 435, 54 },
		{ 436, 436, 54 },
		{ 436, 437, 54 },
		{ 436, 438, 54 },
		{ 436, 439, 54 },
		{ 436, 440, 54 },
		{ 436, 441, 54 },
		{ 436, 442, 54 },
		{ 436, 443, 54 },
		{ 436, 444, 54 },
		{ 436, 445, 54 },
		{ 436, 446, 54 },
		{ 436, 447, 54 },
		{ 436, 448, 54 },
		{ 436, 449, 54 },
		{ 436, 450, 54 },
		{ 436, 451, 54 },
		{ 436, 452, 54 },
		{ 436, 453, 54 },
		{ 436, 454, 54 },
		{ 436, 455, 54 },
		{ 436, 456, 54 },
		{ 436, 457, 54 },
		{ 436, 458, 54 },
		{ 436, 459, 54 },
		{ 436, 460, 54 },
		{ 436, 461, 54 },
		{ 437, 415, 54 },
		{ 437, 416, 54 },
		{ 437, 417, 54 },
		{ 437, 418, 54 },
		{ 437, 419, 54 },
		{ 437, 420, 54 },
		{ 437, 421, 54 },
		{ 437, 422, 54 },
		{ 437, 423, 54 },
		{ 437, 424, 54 },
		{ 437, 425, 54 },
		{ 437, 426, 54 },
		{ 437, 427, 54 },
		{ 437, 428, 54 },
		{ 437, 429, 54 },
		{ 437, 430, 54 },
		{ 437, 431, 54 },
		{ 437, 432, 54 },
		{ 437, 433, 54 },
		{ 437, 434, 54 },
		{ 437, 435, 54 },
		{ 437, 436, 54 },
		{ 437, 437, 54 },
		{ 437, 438, 54 },
		{ 437, 439, 54 },
		{ 437, 440, 54 },
		{ 437, 441, 54 },
		{ 437, 442, 54 },
		{ 437, 443, 54 },
		{ 437, 444, 54 },
		{ 437, 445, 54 },
		{ 437, 446, 54 },
		{ 437, 447, 54 },
		{ 437, 448, 54 },
		{ 437, 449, 54 },
		{ 437, 450, 54 },
		{ 437, 451, 54 },
		{ 437, 452, 54 },
		{ 437, 453, 54 },
		{ 437, 454, 54 },
		{ 437, 455, 54 },
		{ 437, 456, 54 },
		{ 437, 457, 54 },
		{ 437, 458, 54 },
		{ 437, 459, 54 },
		{ 437, 460, 54 },
		{ 437, 461, 54 },
		{ 437, 462, 54 },
		{ 438, 416, 54 },
		{ 438, 417, 54 },
		{ 438, 418, 54 },
		{ 438, 419, 54 },
		{ 438, 420, 54 },
		{ 438, 421, 54 },
		{ 438, 422, 54 },
		{ 438, 423, 54 },
		{ 438, 424, 54 },
		{ 438, 425, 54 },
		{ 438, 426, 54 },
		{ 438, 427, 54 },
		{ 438, 428, 54 },
		{ 438, 429, 54 },
		{ 438, 430, 54 },
		{ 438, 431, 54 },
		{ 438, 432, 54 },
		{ 438, 433, 54 },
		{ 438, 434, 54 },
		{ 438, 435, 54 },
		{ 438, 436, 54 },
		{ 438, 437, 54 },
		{ 438, 438, 54 },
		{ 438, 439, 54 },
		{ 438, 440, 54 },
		{ 438, 441, 54 },
		{ 438, 442, 54 },
		{ 438, 443, 54 },
		{ 438, 444, 54 },
		{ 438, 445, 54 },
		{ 438, 446, 54 },
		{ 438, 447, 54 },
		{ 438, 448, 54 },
		{ 438, 449, 54 },
		{ 438, 450, 54 },
		{ 438, 451, 54 },
		{ 438, 452, 54 },
		{ 438, 453, 54 },
		{ 438, 454, 54 },
		{ 438, 455, 54 },
		{ 438, 456, 54 },
		{ 438, 457, 54 },
		{ 438, 458, 54 },
		{ 438, 459, 54 },
		{ 438, 460, 54 },
		{ 438, 461, 54 },
		{ 438, 462, 54 },
		{ 438, 463, 54 },
		{ 439, 417, 54 },
		{ 439, 418, 54 },
		{ 439, 419, 54 },
		{ 439, 420, 54 },
		{ 439, 421, 54 },
		{ 439, 422, 54 },
		{ 439, 423, 54 },
		{ 439, 424, 54 },
		{ 439, 425, 54 },
		{ 439, 426, 54 },
		{ 439, 427, 54 },
		{ 439, 428, 54 },
		{ 439, 429, 54 },
		{ 439, 430, 54 },
		{ 439, 431, 54 },
		{ 439, 432, 54 },
		{ 439, 433, 54 },
		{ 439, 434, 54 },
		{ 439, 435, 54 },
		{ 439, 436, 54 },
		{ 439, 437, 54 },
		{ 439, 438, 54 },
		{ 439, 439, 54 },
		{ 439, 440, 54 },
		{ 439, 441, 54 },
		{ 439, 442, 54 },
		{ 439, 443, 54 },
		{ 439, 444, 54 },
		{ 439, 445, 54 },
		{ 439, 446, 54 },
		{ 439, 447, 54 },
		{ 439, 448, 54 },
		{ 439, 449, 54 },
		{ 439, 450, 54 },
		{ 439, 451, 54 },
		{ 439, 452, 54 },
		{ 439, 453, 54 },
		{ 439, 454, 54 },
		{ 439, 455, 54 },
		{ 439, 456, 54 },
		{ 439, 457, 54 },
		{ 439, 458, 54 },
		{ 439, 459, 54 },
		{ 439, 460, 54 },
		{ 439, 461, 54 },
		{ 439, 462, 54 },
		{ 439, 463, 54 },
		{ 439, 464, 54 },
		{ 440, 418, 54 },
		{ 440, 419, 54 },
		{ 440, 420, 54 },
		{ 440, 421, 54 },
		{ 440, 422, 54 },
		{ 440, 423, 54 },
		{ 440, 424, 54 },
		{ 440, 425, 54 },
		{ 440, 426, 54 },
		{ 440, 427, 54 },
		{ 440, 428, 54 },
		{ 440, 429, 54 },
		{ 440, 430, 54 },
		{ 440, 431, 54 },
		{ 440, 432, 54 },
		{ 440, 433, 54 },
		{ 440, 434, 54 },
		{ 440, 435, 54 },
		{ 440, 436, 54 },
		{ 440, 437, 54 },
		{ 440, 438, 54 },
		{ 440, 439, 54 },
		{ 440, 440, 54 },
		{ 440, 441, 54 },
		{ 440, 442, 54 },
		{ 440, 443, 54 },
		{ 440, 444, 54 },
		{ 440, 445, 54 },
		{ 440, 446, 54 },
		{ 440, 447, 54 },
		{ 440, 448, 54 },
		{ 440, 449, 54 },
		{ 440, 450, 54 },
		{ 440, 451, 54 },
		{ 440, 452, 54 },
		{ 440, 453, 54 },
		{ 440, 454, 54 },
		{ 440, 455, 54 },
		{ 440, 456, 54 },
		{ 440, 457, 54 },
		{ 440, 458, 54 },
		{ 440, 459, 54 },
		{ 440, 460, 54 },
		{ 440, 461, 54 },
		{ 440, 462, 54 },
		{ 440, 463, 54 },
		{ 440, 464, 54 },
		{ 440, 465, 54 },
		{ 440, 466, 54 },
		{ 441, 419, 54 },
		{ 441, 420, 54 },
		{ 441, 421, 54 },
		{ 441, 422, 54 },
		{ 441, 423, 54 },
		{ 441, 424, 54 },
		{ 441, 425, 54 },
		{ 441, 426, 54 },
		{ 441, 427, 54 },
		{ 441, 428, 54 },
		{ 441, 429, 54 },
		{ 441, 430, 54 },
		{ 441, 431, 54 },
		{ 441, 432, 54 },
		{ 441, 433, 54 },
		{ 441, 434, 54 },
		{ 441, 435, 54 },
		{ 441, 436, 54 },
		{ 441, 437, 54 },
		{ 441, 438, 54 },
		{ 441, 439, 54 },
		{ 441, 440, 54 },
		{ 441, 441, 54 },
		{ 441, 442, 54 },
		{ 441, 443, 54 },
		{ 441, 444, 54 },
		{ 441, 445, 54 },
		{ 441, 446, 54 },
		{ 441, 447, 54 },
		{ 441, 448, 54 },
		{ 441, 449, 54 },
		{ 441, 450, 54 },
		{ 441, 451, 54 },
		{ 441, 452, 54 },
		{ 441, 453, 54 },
		{ 441, 454, 54 },
		{ 441, 455, 54 },
		{ 441, 456, 54 },
		{ 441, 457, 54 },
		{ 441, 458, 54 },
		{ 441, 459, 54 },
		{ 441, 460, 54 },
		{ 441, 461, 54 },
		{ 441, 462, 54 },
		{ 441, 463, 54 },
		{ 441, 464, 54 },
		{ 441, 465, 54 },
		{ 441, 466, 54 },
		{ 441, 467, 54 },
		{ 442, 419, 54 },
		{ 442, 420, 54 },
		{ 442, 421, 54 },
		{ 442, 422, 54 },
		{ 442, 423, 54 },
		{ 442, 424, 54 },
		{ 442, 425, 54 },
		{ 442, 426, 54 },
		{ 442, 427, 54 },
		{ 442, 428, 54 },
		{ 442, 429, 54 },
		{ 442, 430, 54 },
		{ 442, 431, 54 },
		{ 442, 432, 54 },
		{ 442, 433, 54 },
		{ 442, 434, 54 },
		{ 442, 435, 54 },
		{ 442, 436, 54 },
		{ 442, 437, 54 },
		{ 442, 438, 54 },
		{ 442, 439, 54 },
		{ 442, 440, 54 },
		{ 442, 441, 54 },
		{ 442, 442, 54 },
		{ 442, 443, 54 },
		{ 442, 444, 54 },
		{ 442, 445, 54 },
		{ 442, 446, 54 },
		{ 442, 447, 54 },
		{ 442, 448, 54 },
		{ 442, 449, 54 },
		{ 442, 450, 54 },
		{ 442, 451, 54 },
		{ 442, 452, 54 },
		{ 442, 453, 54 },
		{ 442, 454, 54 },
		{ 442, 455, 54 },
		{ 442, 456, 54 },
		{ 442, 457, 54 },
		{ 442, 458, 54 },
		{ 442, 459, 54 },
		{ 442, 460, 54 },
		{ 442, 461, 54 },
		{ 442, 462, 54 },
		{ 442, 463, 54 },
		{ 442, 464, 54 },
		{ 442, 465, 54 },
		{ 442, 466, 54 },
		{ 442, 467, 54 },
		{ 442, 468, 54 },
		{ 443, 420, 54 },
		{ 443, 421, 54 },
		{ 443, 422, 54 },
		{ 443, 423, 54 },
		{ 443, 424, 54 },
		{ 443, 425, 54 },
		{ 443, 426, 54 },
		{ 443, 427, 54 },
		{ 443, 428, 54 },
		{ 443, 429, 54 },
		{ 443, 430, 54 },
		{ 443, 431, 54 },
		{ 443, 432, 54 },
		{ 443, 433, 54 },
		{ 443, 434, 54 },
		{ 443, 435, 54 },
		{ 443, 436, 54 },
		{ 443, 437, 54 },
		{ 443, 438, 54 },
		{ 443, 439, 54 },
		{ 443, 440, 54 },
		{ 443, 441, 54 },
		{ 443, 442, 54 },
		{ 443, 443, 54 },
		{ 443, 444, 54 },
		{ 443, 445, 54 },
		{ 443, 446, 54 },
		{ 443, 447, 54 },
		{ 443, 448, 54 },
		{ 443, 449, 54 },
		{ 443, 450, 54 },
		{ 443, 451, 54 },
		{ 443, 452, 54 },
		{ 443, 453, 54 },
		{ 443, 454, 54 },
		{ 443, 455, 54 },
		{ 443, 456, 54 },
		{ 443, 457, 54 },
		{ 443, 458, 54 },
		{ 443, 459, 54 },
		{ 443, 460, 54 },
		{ 443, 461, 54 },
		{ 443, 462, 54 },
		{ 443, 463, 54 },
		{ 443, 464, 54 },
		{ 443, 465, 54 },
		{ 443, 466, 54 },
		{ 443, 467, 54 },
		{ 443, 468, 54 },
		{ 443, 469, 54 },
		{ 444, 421, 54 },
		{ 444, 422, 54 },
		{ 444, 423, 54 },
		{ 444, 424, 54 },
		{ 444, 425, 54 },
		{ 444, 426, 54 },
		{ 444, 427, 54 },
		{ 444, 428, 54 },
		{ 444, 429, 54 },
		{ 444, 430, 54 },
		{ 444, 431, 54 },
		{ 444, 432, 54 },
		{ 444, 433, 54 },
		{ 444, 434, 54 },
		{ 444, 435, 54 },
		{ 444, 436, 54 },
		{ 444, 437, 54 },
		{ 444, 438, 54 },
		{ 444, 439, 54 },
		{ 444, 440, 54 },
		{ 444, 441, 54 },
		{ 444, 442, 54 },
		{ 444, 443, 54 },
		{ 444, 444, 54 },
		{ 444, 445, 54 },
		{ 444, 446, 54 },
		{ 444, 447, 54 },
		{ 444, 448, 54 },
		{ 444, 449, 54 },
		{ 444, 450, 54 },
		{ 444, 451, 54 },
		{ 444, 452, 54 },
		{ 444, 453, 54 },
		{ 444, 454, 54 },
		{ 444, 455, 54 },
		{ 444, 456, 54 },
		{ 444, 457, 54 },
		{ 444, 458, 54 },
		{ 444, 459, 54 },
		{ 444, 460, 54 },
		{ 444, 461, 54 },
		{ 444, 462, 54 },
		{ 444, 463, 54 },
		{ 444, 464, 54 },
		{ 444, 465, 54 },
		{ 444, 466, 54 },
		{ 444, 467, 54 },
		{ 444, 468, 54 },
		{ 444, 469, 54 },
		{ 444, 470, 54 },
		{ 445, 422, 54 },
		{ 445, 423, 54 },
		{ 445, 424, 54 },
		{ 445, 425, 54 },
		{ 445, 426, 54 },
		{ 445, 427, 54 },
		{ 445, 428, 54 },
		{ 445, 429, 54 },
		{ 445, 430, 54 },
		{ 445, 431, 54 },
		{ 445, 432, 54 },
		{ 445, 433, 54 },
		{ 445, 434, 54 },
		{ 445, 435, 54 },
		{ 445, 436, 54 },
		{ 445, 437, 54 },
		{ 445, 438, 54 },
		{ 445, 439, 54 },
		{ 445, 440, 54 },
		{ 445, 441, 54 },
		{ 445, 442, 54 },
		{ 445, 443, 54 },
		{ 445, 444, 54 },
		{ 445, 445, 54 },
		{ 445, 446, 54 },
		{ 445, 447, 54 },
		{ 445, 448, 54 },
		{ 445, 449, 54 },
		{ 445, 450, 54 },
		{ 445, 451, 54 },
		{ 445, 452, 54 },
		{ 445, 453, 54 },
		{ 445, 454, 54 },
		{ 445, 455, 54 },
		{ 445, 456, 54 },
		{ 445, 457, 54 },
		{ 445, 458, 54 },
		{ 445, 459, 54 },
		{ 445, 460, 54 },
		{ 445, 461, 54 },
		{ 445, 462, 54 },
		{ 445, 463, 54 },
		{ 445, 464, 54 },
		{ 445, 465, 54 },
		{ 445, 466, 54 },
		{ 445, 467, 54 },
		{ 445, 468, 54 },
		{ 445, 469, 54 },
		{ 445, 470, 54 },
		{ 445, 471, 54 },
		{ 446, 423, 54 },
		{ 446, 424, 54 },
		{ 446, 425, 54 },
		{ 446, 426, 54 },
		{ 446, 427, 54 },
		{ 446, 428, 54 },
		{ 446, 429, 54 },
		{ 446, 430, 54 },
		{ 446, 431, 54 },
		{ 446, 432, 54 },
		{ 446, 433, 54 },
		{ 446, 434, 54 },
		{ 446, 435, 54 },
		{ 446, 436, 54 },
		{ 446, 437, 54 },
		{ 446, 438, 54 },
		{ 446, 439, 54 },
		{ 446, 440, 54 },
		{ 446, 441, 54 },
		{ 446, 442, 54 },
		{ 446, 443, 54 },
		{ 446, 444, 54 },
		{ 446, 445, 54 },
		{ 446, 446, 54 },
		{ 446, 447, 54 },
		{ 446, 448, 54 },
		{ 446, 449, 54 },
		{ 446, 450, 54 },
		{ 446, 451, 54 },
		{ 446, 452, 54 },
		{ 446, 453, 54 },
		{ 446, 454, 54 },
		{ 446, 455, 54 },
		{ 446, 456, 54 },
		{ 446, 457, 54 },
		{ 446, 458, 54 },
		{ 446, 459, 54 },
		{ 446, 460, 54 },
		{ 446, 461, 54 },
		{ 446, 462, 54 },
		{ 446, 463, 54 },
		{ 446, 464, 54 },
		{ 446, 465, 54 },
		{ 446, 466, 54 },
		{ 446, 467, 54 },
		{ 446, 468, 54 },
		{ 446, 469, 54 },
		{ 446, 470, 54 },
		{ 446, 471, 54 },
		{ 446, 472, 54 },
		{ 447, 424, 54 },
		{ 447, 425, 54 },
		{ 447, 426, 54 },
		{ 447, 427, 54 },
		{ 447, 428, 54 },
		{ 447, 429, 54 },
		{ 447, 430, 54 },
		{ 447, 431, 54 },
		{ 447, 432, 54 },
		{ 447, 433, 54 },
		{ 447, 434, 54 },
		{ 447, 435, 54 },
		{ 447, 436, 54 },
		{ 447, 437, 54 },
		{ 447, 438, 54 },
		{ 447, 439, 54 },
		{ 447, 440, 54 },
		{ 447, 441, 54 },
		{ 447, 442, 54 },
		{ 447, 443, 54 },
		{ 447, 444, 54 },
		{ 447, 445, 54 },
		{ 447, 446, 54 },
		{ 447, 447, 54 },
		{ 447, 448, 54 },
		{ 447, 449, 54 },
		{ 447, 450, 54 },
		{ 447, 451, 54 },
		{ 447, 452, 54 },
		{ 447, 453, 54 },
		{ 447, 454, 54 },
		{ 447, 455, 54 },
		{ 447, 456, 54 },
		{ 447, 457, 54 },
		{ 447, 458, 54 },
		{ 447, 459, 54 },
		{ 447, 460, 54 },
		{ 447, 461, 54 },
		{ 447, 462, 54 },
		{ 447, 463, 54 },
		{ 447, 464, 54 },
		{ 447, 465, 54 },
		{ 447, 466, 54 },
		{ 447, 467, 54 },
		{ 447, 468, 54 },
		{ 447, 469, 54 },
		{ 447, 470, 54 },
		{ 447, 471, 54 },
		{ 447, 472, 54 },
		{ 447, 473, 54 },
		{ 447, 474, 54 },
		{ 448, 425, 54 },
		{ 448, 426, 54 },
		{ 448, 427, 54 },
		{ 448, 428, 54 },
		{ 448, 429, 54 },
		{ 448, 430, 54 },
		{ 448, 431, 54 },
		{ 448, 432, 54 },
		{ 448, 433, 54 },
		{ 448, 434, 54 },
		{ 448, 435, 54 },
		{ 448, 436, 54 },
		{ 448, 437, 54 },
		{ 448, 438, 54 },
		{ 448, 439, 54 },
		{ 448, 440, 54 },
		{ 448, 441, 54 },
		{ 448, 442, 54 },
		{ 448, 443, 54 },
		{ 448, 444, 54 },
		{ 448, 445, 54 },
		{ 448, 446, 54 },
		{ 448, 447, 54 },
		{ 448, 448, 54 },
		{ 448, 449, 54 },
		{ 448, 450, 54 },
		{ 448, 451, 54 },
		{ 448, 452, 54 },
		{ 448, 453, 54 },
		{ 448, 454, 54 },
		{ 448, 455, 54 },
		{ 448, 456, 54 },
		{ 448, 457, 54 },
		{ 448, 458, 54 },
		{ 448, 459, 54 },
		{ 448, 460, 54 },
		{ 448, 461, 54 },
		{ 448, 462, 54 },
		{ 448, 463, 54 },
		{ 448, 464, 54 },
		{ 448, 465, 54 },
		{ 448, 466, 54 },
		{ 448, 467, 54 },
		{ 448, 468, 54 },
		{ 448, 469, 54 },
		{ 448, 470, 54 },
		{ 448, 471, 54 },
		{ 448, 472, 54 },
		{ 448, 473, 54 },
		{ 448, 474, 54 },
		{ 448, 475, 54 },
		{ 449, 426, 54 },
		{ 449, 427, 54 },
		{ 449, 428, 54 },
		{ 449, 429, 54 },
		{ 449, 430, 54 },
		{ 449, 431, 54 },
		{ 449, 432, 54 },
		{ 449, 433, 54 },
		{ 449, 434, 54 },
		{ 449, 435, 54 },
		{ 449, 436, 54 },
		{ 449, 437, 54 },
		{ 449, 438, 54 },
		{ 449, 439, 54 },
		{ 449, 440, 54 },
		{ 449, 441, 54 },
		{ 449, 442, 54 },
		{ 449, 443, 54 },
		{ 449, 444, 54 },
		{ 449, 445, 54 },
		{ 449, 446, 54 },
		{ 449, 447, 54 },
		{ 449, 448, 54 },
		{ 449, 449, 54 },
		{ 449, 450, 54 },
		{ 449, 451, 54 },
		{ 449, 452, 54 },
		{ 449, 453, 54 },
		{ 449, 454, 54 },
		{ 449, 455, 54 },
		{ 449, 456, 54 },
		{ 449, 457, 54 },
		{ 449, 458, 54 },
		{ 449, 459, 54 },
		{ 449, 460, 54 },
		{ 449, 461, 54 },
		{ 449, 462, 54 },
		{ 449, 463, 54 },
		{ 449, 464, 54 },
		{ 449, 465, 54 },
		{ 449, 466, 54 },
		{ 449, 467, 54 },
		{ 449, 468, 54 },
		{ 449, 469, 54 },
		{ 449, 470, 54 },
		{ 449, 471, 54 },
		{ 449, 472, 54 },
		{ 449, 473, 54 },
		{ 449, 474, 54 },
		{ 449, 475, 54 },
		{ 449, 476, 54 },
		{ 450, 426, 54 },
		{ 450, 427, 54 },
		{ 450, 428, 54 },
		{ 450, 429, 54 },
		{ 450, 430, 54 },
		{ 450, 431, 54 },
		{ 450, 432, 54 },
		{ 450, 433, 54 },
		{ 450, 434, 54 },
		{ 450, 435, 54 },
		{ 450, 436, 54 },
		{ 450, 437, 54 },
		{ 450, 438, 54 },
		{ 450, 439, 54 },
		{ 450, 440, 54 },
		{ 450, 441, 54 },
		{ 450, 442, 54 },
		{ 450, 443, 54 },
		{ 450, 444, 54 },
		{ 450, 445, 54 },
		{ 450, 446, 54 },
		{ 450, 447, 54 },
		{ 450, 448, 54 },
		{ 450, 449, 54 },
		{ 450, 450, 54 },
		{ 450, 451, 54 },
		{ 450, 452, 54 },
		{ 450, 453, 54 },
		{ 450, 454, 54 },
		{ 450, 455, 54 },
		{ 450, 456, 54 },
		{ 450, 457, 54 },
		{ 450, 458, 54 },
		{ 450, 459, 54 },
		{ 450, 460, 54 },
		{ 450, 461, 54 },
		{ 450, 462, 54 },
		{ 450, 463, 54 },
		{ 450, 464, 54 },
		{ 450, 465, 54 },
		{ 450, 466, 54 },
		{ 450, 467, 54 },
		{ 450, 468, 54 },
		{ 450, 469, 54 },
		{ 450, 470, 54 },
		{ 450, 471, 54 },
		{ 450, 472, 54 },
		{ 450, 473, 54 },
		{ 450, 474, 54 },
		{ 450, 475, 54 },
		{ 450, 476, 54 },
		{ 450, 477, 54 },
		{ 451, 427, 54 },
		{ 451, 428, 54 },
		{ 451, 429, 54 },
		{ 451, 430, 54 },
		{ 451, 431, 54 },
		{ 451, 432, 54 },
		{ 451, 433, 54 },
		{ 451, 434, 54 },
		{ 451, 435, 54 },
		{ 451, 436, 54 },
		{ 451, 437, 54 },
		{ 451, 438, 54 },
		{ 451, 439, 54 },
		{ 451, 440, 54 },
		{ 451, 441, 54 },
		{ 451, 442, 54 },
		{ 451, 443, 54 },
		{ 451, 444, 54 },
		{ 451, 445, 54 },
		{ 451, 446, 54 },
		{ 451, 447, 54 },
		{ 451, 448, 54 },
		{ 451, 449, 54 },
		{ 451, 450, 54 },
		{ 451, 451, 54 },
		{ 451, 452, 54 },
		{ 451, 453, 54 },
		{ 451, 454, 54 },
		{ 451, 455, 54 },
		{ 451, 456, 54 },
		{ 451, 457, 54 },
		{ 451, 458, 54 },
		{ 451, 459, 54 },
		{ 451, 460, 54 },
		{ 451, 461, 54 },
		{ 451, 462, 54 },
		{ 451, 463, 54 },
		{ 451, 464, 54 },
		{ 451, 465, 54 },
		{ 451, 466, 54 },
		{ 451, 467, 54 },
		{ 451, 468, 54 },
		{ 451, 469, 54 },
		{ 451, 470, 54 },
		{ 451, 471, 54 },
		{ 451, 472, 54 },
		{ 451, 473, 54 },
		{ 451, 474, 54 },
		{ 451, 475, 54 },
		{ 451, 476, 54 },
		{ 451, 477, 54 },
		{ 451, 478, 54 },
		{ 452, 428, 54 },
		{ 452, 429, 54 },
		{ 452, 430, 54 },
		{ 452, 431, 54 },
		{ 452, 432, 54 },
		{ 452, 433, 54 },
		{ 452, 434, 54 },
		{ 452, 435, 54 },
		{ 452, 436, 54 },
		{ 452, 437, 54 },
		{ 452, 438, 54 },
		{ 452, 439, 54 },
		{ 452, 440, 54 },
		{ 452, 441, 54 },
		{ 452, 442, 54 },
		{ 452, 443, 54 },
		{ 452, 444, 54 },
		{ 452, 445, 54 },
		{ 452, 446, 54 },
		{ 452, 447, 54 },
		{ 452, 448, 54 },
		{ 452, 449, 54 },
		{ 452, 450, 54 },
		{ 452, 451, 54 },
		{ 452, 452, 54 },
		{ 452, 453, 54 },
		{ 452, 454, 54 },
		{ 452, 455, 54 },
		{ 452, 456, 54 },
		{ 452, 457, 54 },
		{ 452, 458, 54 },
		{ 452, 459, 54 },
		{ 452, 460, 54 },
		{ 452, 461, 54 },
		{ 452, 462, 54 },
		{ 452, 463, 54 },
		{ 452, 464, 54 },
		{ 452, 465, 54 },
		{ 452, 466, 54 },
		{ 452, 467, 54 },
		{ 452, 468, 54 },
		{ 452, 469, 54 },
		{ 452, 470, 54 },
		{ 452, 471, 54 },
		{ 452, 472, 54 },
		{ 452, 473, 54 },
		{ 452, 474, 54 },
		{ 452, 475, 54 },
		{ 452, 476, 54 },
		{ 452, 477, 54 },
		{ 452, 478, 54 },
		{ 452, 479, 54 },
		{ 453, 429, 54 },
		{ 453, 430, 54 },
		{ 453, 431, 54 },
		{ 453, 432, 54 },
		{ 453, 433, 54 },
		{ 453, 434, 54 },
		{ 453, 435, 54 },
		{ 453, 436, 54 },
		{ 453, 437, 54 },
		{ 453, 438, 54 },
		{ 453, 439, 54 },
		{ 453, 440, 54 },
		{ 453, 441, 54 },
		{ 453, 442, 54 },
		{ 453, 443, 54 },
		{ 453, 444, 54 },
		{ 453, 445, 54 },
		{ 453, 446, 54 },
		{ 453, 447, 54 },
		{ 453, 448, 54 },
		{ 453, 449, 54 },
		{ 453, 450, 54 },
		{ 453, 451, 54 },
		{ 453, 452, 54 },
		{ 453, 453, 54 },
		{ 453, 454, 54 },
		{ 453, 455, 54 },
		{ 453, 456, 54 },
		{ 453, 457, 54 },
		{ 453, 458, 54 },
		{ 453, 459, 54 },
		{ 453, 460, 54 },
		{ 453, 461, 54 },
		{ 453, 462, 54 },
		{ 453, 463, 54 },
		{ 453, 464, 54 },
		{ 453, 465, 54 },
		{ 453, 466, 54 },
		{ 453, 467, 54 },
		{ 453, 468, 54 },
		{ 453, 469, 54 },
		{ 453, 470, 54 },
		{ 453, 471, 54 },
		{ 453, 472, 54 },
		{ 453, 473, 54 },
		{ 453, 474, 54 },
		{ 453, 475, 54 },
		{ 453, 476, 54 },
		{ 453, 477, 54 },
		{ 453, 478, 54 },
		{ 453, 479, 54 },
		{ 453, 480, 54 },
		{ 454, 430, 54 },
		{ 454, 431, 54 },
		{ 454, 432, 54 },
		{ 454, 433, 54 },
		{ 454, 434, 54 },
		{ 454, 435, 54 },
		{ 454, 436, 54 },
		{ 454, 437, 54 },
		{ 454, 438, 54 },
		{ 454, 439, 54 },
		{ 454, 440, 54 },
		{ 454, 441, 54 },
		{ 454, 442, 54 },
		{ 454, 443, 54 },
		{ 454, 444, 54 },
		{ 454, 445, 54 },
		{ 454, 446, 54 },
		{ 454, 447, 54 },
		{ 454, 448, 54 },
		{ 454, 449, 54 },
		{ 454, 450, 54 },
		{ 454, 451, 54 },
		{ 454, 452, 54 },
		{ 454, 453, 54 },
		{ 454, 454, 54 },
		{ 454, 455, 54 },
		{ 454, 456, 54 },
		{ 454, 457, 54 },
		{ 454, 458, 54 },
		{ 454, 459, 54 },
		{ 454, 460, 54 },
		{ 454, 461, 54 },
		{ 454, 462, 54 },
		{ 454, 463, 54 },
		{ 454, 464, 54 },
		{ 454, 465, 54 },
		{ 454, 466, 54 },
		{ 454, 467, 54 },
		{ 454, 468, 54 },
		{ 454, 469, 54 },
		{ 454, 470, 54 },
		{ 454, 471, 54 },
		{ 454, 472, 54 },
		{ 454, 473, 54 },
		{ 454, 474, 54 },
		{ 454, 475, 54 },
		{ 454, 476, 54 },
		{ 454, 477, 54 },
		{ 454, 478, 54 },
		{ 454, 479, 54 },
		{ 454, 480, 54 },
		{ 454, 481, 54 },
		{ 454, 482, 54 },
		{ 455, 431, 54 },
		{ 455, 432, 54 },
		{ 455, 433, 54 },
		{ 455, 434, 54 },
		{ 455, 435, 54 },
		{ 455, 436, 54 },
		{ 455, 437, 54 },
		{ 455, 438, 54 },
		{ 455, 439, 54 },
		{ 455, 440, 54 },
		{ 455, 441, 54 },
		{ 455, 442, 54 },
		{ 455, 443, 54 },
		{ 455, 444, 54 },
		{ 455, 445, 54 },
		{ 455, 446, 54 },
		{ 455, 447, 54 },
		{ 455, 448, 54 },
		{ 455, 449, 54 },
		{ 455, 450, 54 },
		{ 455, 451, 54 },
		{ 455, 452, 54 },
		{ 455, 453, 54 },
		{ 455, 454, 54 },
		{ 455, 455, 54 },
		{ 455, 456, 54 },
		{ 455, 457, 54 },
		{ 455, 458, 54 },
		{ 455, 459, 54 },
		{ 455, 460, 54 },
		{ 455, 461, 54 },
		{ 455, 462, 54 },
		{ 455, 463, 54 },
		{ 455, 464, 54 },
		{ 455, 465, 54 },
		{ 455, 466, 54 },
		{ 455, 467, 54 },
		{ 455, 468, 54 },
		{ 455, 469, 54 },
		{ 455, 470, 54 },
		{ 455, 471, 54 },
		{ 455, 472, 54 },
		{ 455, 473, 54 },
		{ 455, 474, 54 },
		{ 455, 475, 54 },
		{ 455, 476, 54 },
		{ 455, 477, 54 },
		{ 455, 478, 54 },
		{ 455, 479, 54 },
		{ 455, 480, 54 },
		{ 455, 481, 54 },
		{ 455, 482, 54 },
		{ 455, 483, 54 },
		{ 456, 432, 54 },
		{ 456, 433, 54 },
		{ 456, 434, 54 },
		{ 456, 435, 54 },
		{ 456, 436, 54 },
		{ 456, 437, 54 },
		{ 456, 438, 54 },
		{ 456, 439, 54 },
		{ 456, 440, 54 },
		{ 456, 441, 54 },
		{ 456, 442, 54 },
		{ 456, 443, 54 },
		{ 456, 444, 54 },
		{ 456, 445, 54 },
		{ 456, 446, 54 },
		{ 456, 447, 54 },
		{ 456, 448, 54 },
		{ 456, 449, 54 },
		{ 456, 450, 54 },
		{ 456, 451, 54 },
		{ 456, 452, 54 },
		{ 456, 453, 54 },
		{ 456, 454, 54 },
		{ 456, 455, 54 },
		{ 456, 456, 54 },
		{ 456, 457, 54 },
		{ 456, 458, 54 },
		{ 456, 459, 54 },
		{ 456, 460, 54 },
		{ 456, 461, 54 },
		{ 456, 462, 54 },
		{ 456, 463, 54 },
		{ 456, 464, 54 },
		{ 456, 465, 54 },
		{ 456, 466, 54 },
		{ 456, 467, 54 },
		{ 456, 468, 54 },
		{ 456, 469, 54 },
		{ 456, 470, 54 },
		{ 456, 471, 54 },
		{ 456, 472, 54 },
		{ 456, 473, 54 },
		{ 456, 474, 54 },
		{ 456, 475, 54 },
		{ 456, 476, 54 },
		{ 456, 477, 54 },
		{ 456, 478, 54 },
		{ 456, 479, 54 },
		{ 456, 480, 54 },
		{ 456, 481, 54 },
		{ 456, 482, 54 },
		{ 456, 483, 54 },
		{ 456, 484, 54 },
		{ 457, 433, 54 },
		{ 457, 434, 54 },
		{ 457, 435, 54 },
		{ 457, 436, 54 },
		{ 457, 437, 54 },
		{ 457, 438, 54 },
		{ 457, 439, 54 },
		{ 457, 440, 54 },
		{ 457, 441, 54 },
		{ 457, 442, 54 },
		{ 457, 443, 54 },
		{ 457, 444, 54 },
		{ 457, 445, 54 },
		{ 457, 446, 54 },
		{ 457, 447, 54 },
		{ 457, 448, 54 },
		{ 457, 449, 54 },
		{ 457, 450, 54 },
		{ 457, 451, 54 },
		{ 457, 452, 54 },
		{ 457, 453, 54 },
		{ 457, 454, 54 },
		{ 457, 455, 54 },
		{ 457, 456, 54 },
		{ 457, 457, 54 },
		{ 457, 458, 54 },
		{ 457, 459, 54 },
		{ 457, 460, 54 },
		{ 457, 461, 54 },
		{ 457, 462, 54 },
		{ 457, 463, 54 },
		{ 457, 464, 54 },
		{ 457, 465, 54 },
		{ 457, 466, 54 },
		{ 457, 467, 54 },
		{ 457, 468, 54 },
		{ 457, 469, 54 },
		{ 457, 470, 54 },
		{ 457, 471, 54 },
		{ 457, 472, 54 },
		{ 457, 473, 54 },
		{ 457, 474, 54 },
		{ 457, 475, 54 },
		{ 457, 476, 54 },
		{ 457, 477, 54 },
		{ 457, 478, 54 },
		{ 457, 479, 54 },
		{ 457, 480, 54 },
		{ 457, 481, 54 },
		{ 457, 482, 54 },
		{ 457, 483, 54 },
		{ 457, 484, 54 },
		{ 457, 485, 54 },
		{ 458, 433, 54 },
		{ 458, 434, 54 },
		{ 458, 435, 54 },
		{ 458, 436, 54 },
		{ 458, 437, 54 },
		{ 458, 438, 54 },
		{ 458, 439, 54 },
		{ 458, 440, 54 },
		{ 458, 441, 54 },
		{ 458, 442, 54 },
		{ 458, 443, 54 },
		{ 458, 444, 54 },
		{ 458, 445, 54 },
		{ 458, 446, 54 },
		{ 458, 447, 54 },
		{ 458, 448, 54 },
		{ 458, 449, 54 },
		{ 458, 450, 54 },
		{ 458, 451, 54 },
		{ 458, 452, 54 },
		{ 458, 453, 54 },
		{ 458, 454, 54 },
		{ 458, 455, 54 },
		{ 458, 456, 54 },
		{ 458, 457, 54 },
		{ 458, 458, 54 },
		{ 458, 459, 54 },
		{ 458, 460, 54 },
		{ 458, 461, 54 },
		{ 458, 462, 54 },
		{ 458, 463, 54 },
		{ 458, 464, 54 },
		{ 458, 465, 54 },
		{ 458, 466, 54 },
		{ 458, 467, 54 },
		{ 458, 468, 54 },
		{ 458, 469, 54 },
		{ 458, 470, 54 },
		{ 458, 471, 54 },
		{ 458, 472, 54 },
		{ 458, 473, 54 },
		{ 458, 474, 54 },
		{ 458, 475, 54 },
		{ 458, 476, 54 },
		{ 458, 477, 54 },
		{ 458, 478, 54 },
		{ 458, 479, 54 },
		{ 458, 480, 54 },
		{ 458, 481, 54 },
		{ 458, 482, 54 },
		{ 458, 483, 54 },
		{ 458, 484, 54 },
		{ 458, 485, 54 },
		{ 458, 486, 54 },
		{ 459, 434, 54 },
		{ 459, 435, 54 },
		{ 459, 436, 54 },
		{ 459, 437, 54 },
		{ 459, 438, 54 },
		{ 459, 439, 54 },
		{ 459, 440, 54 },
		{ 459, 441, 54 },
		{ 459, 442, 54 },
		{ 459, 443, 54 },
		{ 459, 444, 54 },
		{ 459, 445, 54 },
		{ 459, 446, 54 },
		{ 459, 447, 54 },
		{ 459, 448, 54 },
		{ 459, 449, 54 },
		{ 459, 450, 54 },
		{ 459, 451, 54 },
		{ 459, 452, 54 },
		{ 459, 453, 54 },
		{ 459, 454, 54 },
		{ 459, 455, 54 },
		{ 459, 456, 54 },
		{ 459, 457, 54 },
		{ 459, 458, 54 },
		{ 459, 459, 54 },
		{ 459, 460, 54 },
		{ 459, 461, 54 },
		{ 459, 462, 54 },
		{ 459, 463, 54 },
		{ 459, 464, 54 },
		{ 459, 465, 54 },
		{ 459, 466, 54 },
		{ 459, 467, 54 },
		{ 459, 468, 54 },
		{ 459, 469, 54 },
		{ 459, 470, 54 },
		{ 459, 471, 54 },
		{ 459, 472, 54 },
		{ 459, 473, 54 },
		{ 459, 474, 54 },
		{ 459, 475, 54 },
		{ 459, 476, 54 },
		{ 459, 477, 54 },
		{ 459, 478, 54 },
		{ 459, 479, 54 },
		{ 459, 480, 54 },
		{ 459, 481, 54 },
		{ 459, 482, 54 },
		{ 459, 483, 54 },
		{ 459, 484, 54 },
		{ 459, 485, 54 },
		{ 459, 486, 54 },
		{ 459, 487, 54 },
		{ 460, 435, 54 },
		{ 460, 436, 54 },
		{ 460, 437, 54 },
		{ 460, 438, 54 },
		{ 460, 439, 54 },
		{ 460, 440, 54 },
		{ 460, 441, 54 },
		{ 460, 442, 54 },
		{ 460, 443, 54 },
		{ 460, 444, 54 },
		{ 460, 445, 54 },
		{ 460, 446, 54 },
		{ 460, 447, 54 },
		{ 460, 448, 54 },
		{ 460, 449, 54 },
		{ 460, 450, 54 },
		{ 460, 451, 54 },
		{ 460, 452, 54 },
		{ 460, 453, 54 },
		{ 460, 454, 54 },
		{ 460, 455, 54 },
		{ 460, 456, 54 },
		{ 460, 457, 54 },
		{ 460, 458, 54 },
		{ 460, 459, 54 },
		{ 460, 460, 54 },
		{ 460, 461, 54 },
		{ 460, 462, 54 },
		{ 460, 463, 54 },
		{ 460, 464, 54 },
		{ 460, 465, 54 },
		{ 460, 466, 54 },
		{ 460, 467, 54 },
		{ 460, 468, 54 },
		{ 460, 469, 54 },
		{ 460, 470, 54 },
		{ 460, 471, 54 },
		{ 460, 472, 54 },
		{ 460, 473, 54 },
		{ 460, 474, 54 },
		{ 460, 475, 54 },
		{ 460, 476, 54 },
		{ 460, 477, 54 },
		{ 460, 478, 54 },
		{ 460, 479, 54 },
		{ 460, 480, 54 },
		{ 460, 481, 54 },
		{ 460, 482, 54 },
		{ 460, 483, 54 },
		{ 460, 484, 54 },
		{ 460, 485, 54 },
		{ 460, 486, 54 },
		{ 460, 487, 54 },
		{ 460, 488, 54 },
		{ 461, 436, 54 },
		{ 461, 437, 54 },
		{ 461, 438, 54 },
		{ 461, 439, 54 },
		{ 461, 440, 54 },
		{ 461, 441, 54 },
		{ 461, 442, 54 },
		{ 461, 443, 54 },
		{ 461, 444, 54 },
		{ 461, 445, 54 },
		{ 461, 446, 54 },
		{ 461, 447, 54 },
		{ 461, 448, 54 },
		{ 461, 449, 54 },
		{ 461, 450, 54 },
		{ 461, 451, 54 },
		{ 461, 452, 54 },
		{ 461, 453, 54 },
		{ 461, 454, 54 },
		{ 461, 455, 54 },
		{ 461, 456, 54 },
		{ 461, 457, 54 },
		{ 461, 458, 54 },
		{ 461, 459, 54 },
		{ 461, 460, 54 },
		{ 461, 461, 54 },
		{ 461, 462, 54 },
		{ 461, 463, 54 },
		{ 461, 464, 54 },
		{ 461, 465, 54 },
		{ 461, 466, 54 },
		{ 461, 467, 54 },
		{ 461, 468, 54 },
		{ 461, 469, 54 },
		{ 461, 470, 54 },
		{ 461, 471, 54 },
		{ 461, 472, 54 },
		{ 461, 473, 54 },
		{ 461, 474, 54 },
		{ 461, 475, 54 },
		{ 461, 476, 54 },
		{ 461, 477, 54 },
		{ 461, 478, 54 },
		{ 461, 479, 54 },
		{ 461, 480, 54 },
		{ 461, 481, 54 },
		{ 461, 482, 54 },
		{ 461, 483, 54 },
		{ 461, 484, 54 },
		{ 461, 485, 54 },
		{ 461, 486, 54 },
		{ 461, 487, 54 },
		{ 461, 488, 54 },
		{ 461, 489, 54 },
		{ 461, 490, 54 },
		{ 462, 437, 54 },
		{ 462, 438, 54 },
		{ 462, 439, 54 },
		{ 462, 440, 54 },
		{ 462, 441, 54 },
		{ 462, 442, 54 },
		{ 462, 443, 54 },
		{ 462, 444, 54 },
		{ 462, 445, 54 },
		{ 462, 446, 54 },
		{ 462, 447, 54 },
		{ 462, 448, 54 },
		{ 462, 449, 54 },
		{ 462, 450, 54 },
		{ 462, 451, 54 },
		{ 462, 452, 54 },
		{ 462, 453, 54 },
		{ 462, 454, 54 },
		{ 462, 455, 54 },
		{ 462, 456, 54 },
		{ 462, 457, 54 },
		{ 462, 458, 54 },
		{ 462, 459, 54 },
		{ 462, 460, 54 },
		{ 462, 461, 54 },
		{ 462, 462, 54 },
		{ 462, 463, 54 },
		{ 462, 464, 54 },
		{ 462, 465, 54 },
		{ 462, 466, 54 },
		{ 462, 467, 54 },
		{ 462, 468, 54 },
		{ 462, 469, 54 },
		{ 462, 470, 54 },
		{ 462, 471, 54 },
		{ 462, 472, 54 },
		{ 462, 473, 54 },
		{ 462, 474, 54 },
		{ 462, 475, 54 },
		{ 462, 476, 54 },
		{ 462, 477, 54 },
		{ 462, 478, 54 },
		{ 462, 479, 54 },
		{ 462, 480, 54 },
		{ 462, 481, 54 },
		{ 462, 482, 54 },
		{ 462, 483, 54 },
		{ 462, 484, 54 },
		{ 462, 485, 54 },
		{ 462, 486, 54 },
		{ 462, 487, 54 },
		{ 462, 488, 54 },
		{ 462, 489, 54 },
		{ 462, 490, 54 },
		{ 462, 491, 54 },
		{ 463, 438, 54 },
		{ 463, 439, 54 },
		{ 463, 440, 54 },
		{ 463, 441, 54 },
		{ 463, 442, 54 },
		{ 463, 443, 54 },
		{ 463, 444, 54 },
		{ 463, 445, 54 },
		{ 463, 446, 54 },
		{ 463, 447, 54 },
		{ 463, 448, 54 },
		{ 463, 449, 54 },
		{ 463, 450, 54 },
		{ 463, 451, 54 },
		{ 463, 452, 54 },
		{ 463, 453, 54 },
		{ 463, 454, 54 },
		{ 463, 455, 54 },
		{ 463, 456, 54 },
		{ 463, 457, 54 },
		{ 463, 458, 54 },
		{ 463, 459, 54 },
		{ 463, 460, 54 },
		{ 463, 461, 54 },
		{ 463, 462, 54 },
		{ 463, 463, 54 },
		{ 463, 464, 54 },
		{ 463, 465, 54 },
		{ 463, 466, 54 },
		{ 463, 467, 54 },
		{ 463, 468, 54 },
		{ 463, 469, 54 },
		{ 463, 470, 54 },
		{ 463, 471, 54 },
		{ 463, 472, 54 },
		{ 463, 473, 54 },
		{ 463, 474, 54 },
		{ 463, 475, 54 },
		{ 463, 476, 54 },
		{ 463, 477, 54 },
		{ 463, 478, 54 },
		{ 463, 479, 54 },
		{ 463, 480, 54 },
		{ 463, 481, 54 },
		{ 463, 482, 54 },
		{ 463, 483, 54 },
		{ 463, 484, 54 },
		{ 463, 485, 54 },
		{ 463, 486, 54 },
		{ 463, 487, 54 },
		{ 463, 488, 54 },
		{ 463, 489, 54 },
		{ 463, 490, 54 },
		{ 463, 491, 54 },
		{ 463, 492, 54 },
		{ 464, 439, 54 },
		{ 464, 440, 54 },
		{ 464, 441, 54 },
		{ 464, 442, 54 },
		{ 464, 443, 54 },
		{ 464, 444, 54 },
		{ 464, 445, 54 },
		{ 464, 446, 54 },
		{ 464, 447, 54 },
		{ 464, 448, 54 },
		{ 464, 449, 54 },
		{ 464, 450, 54 },
		{ 464, 451, 54 },
		{ 464, 452, 54 },
		{ 464, 453, 54 },
		{ 464, 454, 54 },
		{ 464, 455, 54 },
		{ 464, 456, 54 },
		{ 464, 457, 54 },
		{ 464, 458, 54 },
		{ 464, 459, 54 },
		{ 464, 460, 54 },
		{ 464, 461, 54 },
		{ 464, 462, 54 },
		{ 464, 463, 54 },
		{ 464, 464, 54 },
		{ 464, 465, 54 },
		{ 464, 466, 54 },
		{ 464, 467, 54 },
		{ 464, 468, 54 },
		{ 464, 469, 54 },
		{ 464, 470, 54 },
		{ 464, 471, 54 },
		{ 464, 472, 54 },
		{ 464, 473, 54 },
		{ 464, 474, 54 },
		{ 464, 475, 54 },
		{ 464, 476, 54 },
		{ 464, 477, 54 },
		{ 464, 478, 54 },
		{ 464, 479, 54 },
		{ 464, 480, 54 },
		{ 464, 481, 54 },
		{ 464, 482, 54 },
		{ 464, 483, 54 },
		{ 464, 484, 54 },
		{ 464, 485, 54 },
		{ 464, 486, 54 },
		{ 464, 487, 54 },
		{ 464, 488, 54 },
		{ 464, 489, 54 },
		{ 464, 490, 54 },
		{ 464, 491, 54 },
		{ 464, 492, 54 },
		{ 464, 493, 54 },
		{ 465, 440, 54 },
		{ 465, 441, 54 },
		{ 465, 442, 54 },
		{ 465, 443, 54 },
		{ 465, 444, 54 },
		{ 465, 445, 54 },
		{ 465, 446, 54 },
		{ 465, 447, 54 },
		{ 465, 448, 54 },
		{ 465, 449, 54 },
		{ 465, 450, 54 },
		{ 465, 451, 54 },
		{ 465, 452, 54 },
		{ 465, 453, 54 },
		{ 465, 454, 54 },
		{ 465, 455, 54 },
		{ 465, 456, 54 },
		{ 465, 457, 54 },
		{ 465, 458, 54 },
		{ 465, 459, 54 },
		{ 465, 460, 54 },
		{ 465, 461, 54 },
		{ 465, 462, 54 },
		{ 465, 463, 54 },
		{ 465, 464, 54 },
		{ 465, 465, 54 },
		{ 465, 466, 54 },
		{ 465, 467, 54 },
		{ 465, 468, 54 },
		{ 465, 469, 54 },
		{ 465, 470, 54 },
		{ 465, 471, 54 },
		{ 465, 472, 54 },
		{ 465, 473, 54 },
		{ 465, 474, 54 },
		{ 465, 475, 54 },
		{ 465, 476, 54 },
		{ 465, 477, 54 },
		{ 465, 478, 54 },
		{ 465, 479, 54 },
		{ 465, 480, 54 },
		{ 465, 481, 54 },
		{ 465, 482, 54 },
		{ 465, 483, 54 },
		{ 465, 484, 54 },
		{ 465, 485, 54 },
		{ 465, 486, 54 },
		{ 465, 487, 54 },
		{ 465, 488, 54 },
		{ 465, 489, 54 },
		{ 465, 490, 54 },
		{ 465, 491, 54 },
		{ 465, 492, 54 },
		{ 465, 493, 54 },
		{ 465, 494, 54 },
		{ 466, 440, 54 },
		{ 466, 441, 54 },
		{ 466, 442, 54 },
		{ 466, 443, 54 },
		{ 466, 444, 54 },
		{ 466, 445, 54 },
		{ 466, 446, 54 },
		{ 466, 447, 54 },
		{ 466, 448, 54 },
		{ 466, 449, 54 },
		{ 466, 450, 54 },
		{ 466, 451, 54 },
		{ 466, 452, 54 },
		{ 466, 453, 54 },
		{ 466, 454, 54 },
		{ 466, 455, 54 },
		{ 466, 456, 54 },
		{ 466, 457, 54 },
		{ 466, 458, 54 },
		{ 466, 459, 54 },
		{ 466, 460, 54 },
		{ 466, 461, 54 },
		{ 466, 462, 54 },
		{ 466, 463, 54 },
		{ 466, 464, 54 },
		{ 466, 465, 54 },
		{ 466, 466, 54 },
		{ 466, 467, 54 },
		{ 466, 468, 54 },
		{ 466, 469, 54 },
		{ 466, 470, 54 },
		{ 466, 471, 54 },
		{ 466, 472, 54 },
		{ 466, 473, 54 },
		{ 466, 474, 54 },
		{ 466, 475, 54 },
		{ 466, 476, 54 },
		{ 466, 477, 54 },
		{ 466, 478, 54 },
		{ 466, 479, 54 },
		{ 466, 480, 54 },
		{ 466, 481, 54 },
		{ 466, 482, 54 },
		{ 466, 483, 54 },
		{ 466, 484, 54 },
		{ 466, 485, 54 },
		{ 466, 486, 54 },
		{ 466, 487, 54 },
		{ 466, 488, 54 },
		{ 466, 489, 54 },
		{ 466, 490, 54 },
		{ 466, 491, 54 },
		{ 466, 492, 54 },
		{ 466, 493, 54 },
		{ 466, 494, 54 },
		{ 466, 495, 54 },
		{ 467, 441, 54 },
		{ 467, 442, 54 },
		{ 467, 443, 54 },
		{ 467, 444, 54 },
		{ 467, 445, 54 },
		{ 467, 446, 54 },
		{ 467, 447, 54 },
		{ 467, 448, 54 },
		{ 467, 449, 54 },
		{ 467, 450, 54 },
		{ 467, 451, 54 },
		{ 467, 452, 54 },
		{ 467, 453, 54 },
		{ 467, 454, 54 },
		{ 467, 455, 54 },
		{ 467, 456, 54 },
		{ 467, 457, 54 },
		{ 467, 458, 54 },
		{ 467, 459, 54 },
		{ 467, 460, 54 },
		{ 467, 461, 54 },
		{ 467, 462, 54 },
		{ 467, 463, 54 },
		{ 467, 464, 54 },
		{ 467, 465, 54 },
		{ 467, 466, 54 },
		{ 467, 467, 54 },
		{ 467, 468, 54 },
		{ 467, 469, 54 },
		{ 467, 470, 54 },
		{ 467, 471, 54 },
		{ 467, 472, 54 },
		{ 467, 473, 54 },
		{ 467, 474, 54 },
		{ 467, 475, 54 },
		{ 467, 476, 54 },
		{ 467, 477, 54 },
		{ 467, 478, 54 },
		{ 467, 479, 54 },
		{ 467, 480, 54 },
		{ 467, 481, 54 },
		{ 467, 482, 54 },
		{ 467, 483, 54 },
		{ 467, 484, 54 },
		{ 467, 485, 54 },
		{ 467, 486, 54 },
		{ 467, 487, 54 },
		{ 467, 488, 54 },
		{ 467, 489, 54 },
		{ 467, 490, 54 },
		{ 467, 491, 54 },
		{ 467, 492, 54 },
		{ 467, 493, 54 },
		{ 467, 494, 54 },
		{ 467, 495, 54 },
		{ 467, 496, 54 },
		{ 468, 442, 54 },
		{ 468, 443, 54 },
		{ 468, 444, 54 },
		{ 468, 445, 54 },
		{ 468, 446, 54 },
		{ 468, 447, 54 },
		{ 468, 448, 54 },
		{ 468, 449, 54 },
		{ 468, 450, 54 },
		{ 468, 451, 54 },
		{ 468, 452, 54 },
		{ 468, 453, 54 },
		{ 468, 454, 54 },
		{ 468, 455, 54 },
		{ 468, 456, 54 },
		{ 468, 457, 54 },
		{ 468, 458, 54 },
		{ 468, 459, 54 },
		{ 468, 460, 54 },
		{ 468, 461, 54 },
		{ 468, 462, 54 },
		{ 468, 463, 54 },
		{ 468, 464, 54 },
		{ 468, 465, 54 },
		{ 468, 466, 54 },
		{ 468, 467, 54 },
		{ 468, 468, 54 },
		{ 468, 469, 54 },
		{ 468, 470, 54 },
		{ 468, 471, 54 },
		{ 468, 472, 54 },
		{ 468, 473, 54 },
		{ 468, 474, 54 },
		{ 468, 475, 54 },
		{ 468, 476, 54 },
		{ 468, 477, 54 },
		{ 468, 478, 54 },
		{ 468, 479, 54 },
		{ 468, 480, 54 },
		{ 468, 481, 54 },
		{ 468, 482, 54 },
		{ 468, 483, 54 },
		{ 468, 484, 54 },
		{ 468, 485, 54 },
		{ 468, 486, 54 },
		{ 468, 487, 54 },
		{ 468, 488, 54 },
		{ 468, 489, 54 },
		{ 468, 490, 54 },
		{ 468, 491, 54 },
		{ 468, 492, 54 },
		{ 468, 493, 54 },
		{ 468, 494, 54 },
		{ 468, 495, 54 },
		{ 468, 496, 54 },
		{ 468, 497, 54 },
		{ 468, 498, 54 },
		{ 469, 443, 54 },
		{ 469, 444, 54 },
		{ 469, 445, 54 },
		{ 469, 446, 54 },
		{ 469, 447, 54 },
		{ 469, 448, 54 },
		{ 469, 449, 54 },
		{ 469, 450, 54 },
		{ 469, 451, 54 },
		{ 469, 452, 54 },
		{ 469, 453, 54 },
		{ 469, 454, 54 },
		{ 469, 455, 54 },
		{ 469, 456, 54 },
		{ 469, 457, 54 },
		{ 469, 458, 54 },
		{ 469, 459, 54 },
		{ 469, 460, 54 },
		{ 469, 461, 54 },
		{ 469, 462, 54 },
		{ 469, 463, 54 },
		{ 469, 464, 54 },
		{ 469, 465, 54 },
		{ 469, 466, 54 },
		{ 469, 467, 54 },
		{ 469, 468, 54 },
		{ 469, 469, 54 },
		{ 469, 470, 54 },
		{ 469, 471, 54 },
		{ 469, 472, 54 },
		{ 469, 473, 54 },
		{ 469, 474, 54 },
		{ 469, 475, 54 },
		{ 469, 476, 54 },
		{ 469, 477, 54 },
		{ 469, 478, 54 },
		{ 469, 479, 54 },
		{ 469, 480, 54 },
		{ 469, 481, 54 },
		{ 469, 482, 54 },
		{ 469, 483, 54 },
		{ 469, 484, 54 },
		{ 469, 485, 54 },
		{ 469, 486, 54 },
		{ 469, 487, 54 },
		{ 469, 488, 54 },
		{ 469, 489, 54 },
		{ 469, 490, 54 },
		{ 469, 491, 54 },
		{ 469, 492, 54 },
		{ 469, 493, 54 },
		{ 469, 494, 54 },
		{ 469, 495, 54 },
		{ 469, 496, 54 },
		{ 469, 497, 54 },
		{ 469, 498, 54 },
		{ 469, 499, 54 },
		{ 470, 444, 54 },
		{ 470, 445, 54 },
		{ 470, 446, 54 },
		{ 470, 447, 54 },
		{ 470, 448, 54 },
		{ 470, 449, 54 },
		{ 470, 450, 54 },
		{ 470, 451, 54 },
		{ 470, 452, 54 },
		{ 470, 453, 54 },
		{ 470, 454, 54 },
		{ 470, 455, 54 },
		{ 470, 456, 54 },
		{ 470, 457, 54 },
		{ 470, 458, 54 },
		{ 470, 459, 54 },
		{ 470, 460, 54 },
		{ 470, 461, 54 },
		{ 470, 462, 54 },
		{ 470, 463, 54 },
		{ 470, 464, 54 },
		{ 470, 465, 54 },
		{ 470, 466, 54 },
		{ 470, 467, 54 },
		{ 470, 468, 54 },
		{ 470, 469, 54 },
		{ 470, 470, 54 },
		{ 470, 471, 54 },
		{ 470, 472, 54 },
		{ 470, 473, 54 },
		{ 470, 474, 54 },
		{ 470, 475, 54 },
		{ 470, 476, 54 },
		{ 470, 477, 54 },
		{ 470, 478, 54 },
		{ 470, 479, 54 },
		{ 470, 480, 54 },
		{ 470, 481, 54 },
		{ 470, 482, 54 },
		{ 470, 483, 54 },
		{ 470, 484, 54 },
		{ 470, 485, 54 },
		{ 470, 486, 54 },
		{ 470, 487, 54 },
		{ 470, 488, 54 },
		{ 470, 489, 54 },
		{ 470, 490, 54 },
		{ 470, 491, 54 },
		{ 470, 492, 54 },
		{ 470, 493, 54 },
		{ 470, 494, 54 },
		{ 470, 495, 54 },
		{ 470, 496, 54 },
		{ 470, 497, 54 },
		{ 470, 498, 54 },
		{ 470, 499, 54 },
		{ 470, 500, 54 },
		{ 471, 445, 54 },
		{ 471, 446, 54 },
		{ 471, 447, 54 },
		{ 471, 448, 54 },
		{ 471, 449, 54 },
		{ 471, 450, 54 },
		{ 471, 451, 54 },
		{ 471, 452, 54 },
		{ 471, 453, 54 },
		{ 471, 454, 54 },
		{ 471, 455, 54 },
		{ 471, 456, 54 },
		{ 471, 457, 54 },
		{ 471, 458, 54 },
		{ 471, 459, 54 },
		{ 471, 460, 54 },
		{ 471, 461, 54 },
		{ 471, 462, 54 },
		{ 471, 463, 54 },
		{ 471, 464, 54 },
		{ 471, 465, 54 },
		{ 471, 466, 54 },
		{ 471, 467, 54 },
		{ 471, 468, 54 },
		{ 471, 469, 54 },
		{ 471, 470, 54 },
		{ 471, 471, 54 },
		{ 471, 472, 54 },
		{ 471, 473, 54 },
		{ 471, 474, 54 },
		{ 471, 475, 54 },
		{ 471, 476, 54 },
		{ 471, 477, 54 },
		{ 471, 478, 54 },
		{ 471, 479, 54 },
		{ 471, 480, 54 },
		{ 471, 481, 54 },
		{ 471, 482, 54 },
		{ 471, 483, 54 },
		{ 471, 484, 54 },
		{ 471, 485, 54 },
		{ 471, 486, 54 },
		{ 471, 487, 54 },
		{ 471, 488, 54 },
		{ 471, 489, 54 },
		{ 471, 490, 54 },
		{ 471, 491, 54 },
		{ 471, 492, 54 },
		{ 471, 493, 54 },
		{ 471, 494, 54 },
		{ 471, 495, 54 },
		{ 471, 496, 54 },
		{ 471, 497, 54 },
		{ 471, 498, 54 },
		{ 471, 499, 54 },
		{ 471, 500, 54 },
		{ 471, 501, 54 },
		{ 472, 446, 54 },
		{ 472, 447, 54 },
		{ 472, 448, 54 },
		{ 472, 449, 54 },
		{ 472, 450, 54 },
		{ 472, 451, 54 },
		{ 472, 452, 54 },
		{ 472, 453, 54 },
		{ 472, 454, 54 },
		{ 472, 455, 54 },
		{ 472, 456, 54 },
		{ 472, 457, 54 },
		{ 472, 458, 54 },
		{ 472, 459, 54 },
		{ 472, 460, 54 },
		{ 472, 461, 54 },
		{ 472, 462, 54 },
		{ 472, 463, 54 },
		{ 472, 464, 54 },
		{ 472, 465, 54 },
		{ 472, 466, 54 },
		{ 472, 467, 54 },
		{ 472, 468, 54 },
		{ 472, 469, 54 },
		{ 472, 470, 54 },
		{ 472, 471, 54 },
		{ 472, 472, 54 },
		{ 472, 473, 54 },
		{ 472, 474, 54 },
		{ 472, 475, 54 },
		{ 472, 476, 54 },
		{ 472, 477, 54 },
		{ 472, 478, 54 },
		{ 472, 479, 54 },
		{ 472, 480, 54 },
		{ 472, 481, 54 },
		{ 472, 482, 54 },
		{ 472, 483, 54 },
		{ 472, 484, 54 },
		{ 472, 485, 54 },
		{ 472, 486, 54 },
		{ 472, 487, 54 },
		{ 472, 488, 54 },
		{ 472, 489, 54 },
		{ 472, 490, 54 },
		{ 472, 491, 54 },
		{ 472, 492, 54 },
		{ 472, 493, 54 },
		{ 472, 494, 54 },
		{ 472, 495, 54 },
		{ 472, 496, 54 },
		{ 472, 497, 54 },
		{ 472, 498, 54 },
		{ 472, 499, 54 },
		{ 472, 500, 54 },
		{ 472, 501, 54 },
		{ 472, 502, 54 },
		{ 473, 447, 54 },
		{ 473, 448, 54 },
		{ 473, 449, 54 },
		{ 473, 450, 54 },
		{ 473, 451, 54 },
		{ 473, 452, 54 },
		{ 473, 453, 54 },
		{ 473, 454, 54 },
		{ 473, 455, 54 },
		{ 473, 456, 54 },
		{ 473, 457, 54 },
		{ 473, 458, 54 },
		{ 473, 459, 54 },
		{ 473, 460, 54 },
		{ 473, 461, 54 },
		{ 473, 462, 54 },
		{ 473, 463, 54 },
		{ 473, 464, 54 },
		{ 473, 465, 54 },
		{ 473, 466, 54 },
		{ 473, 467, 54 },
		{ 473, 468, 54 },
		{ 473, 469, 54 },
		{ 473, 470, 54 },
		{ 473, 471, 54 },
		{ 473, 472, 54 },
		{ 473, 473, 54 },
		{ 473, 474, 54 },
		{ 473, 475, 54 },
		{ 473, 476, 54 },
		{ 473, 477, 54 },
		{ 473, 478, 54 },
		{ 473, 479, 54 },
		{ 473, 480, 54 },
		{ 473, 481, 54 },
		{ 473, 482, 54 },
		{ 473, 483, 54 },
		{ 473, 484, 54 },
		{ 473, 485, 54 },
		{ 473, 486, 54 },
		{ 473, 487, 54 },
		{ 473, 488, 54 },
		{ 473, 489, 54 },
		{ 473, 490, 54 },
		{ 473, 491, 54 },
		{ 473, 492, 54 },
		{ 473, 493, 54 },
		{ 473, 494, 54 },
		{ 473, 495, 54 },
		{ 473, 496, 54 },
		{ 473, 497, 54 },
		{ 473, 498, 54 },
		{ 473, 499, 54 },
		{ 473, 500, 54 },
		{ 473, 501, 54 },
		{ 473, 502, 54 },
		{ 473, 503, 54 },
		{ 474, 447, 54 },
		{ 474, 448, 54 },
		{ 474, 449, 54 },
		{ 474, 450, 54 },
		{ 474, 451, 54 },
		{ 474, 452, 54 },
		{ 474, 453, 54 },
		{ 474, 454, 54 },
		{ 474, 455, 54 },
		{ 474, 456, 54 },
		{ 474, 457, 54 },
		{ 474, 458, 54 },
		{ 474, 459, 54 },
		{ 474, 460, 54 },
		{ 474, 461, 54 },
		{ 474, 462, 54 },
		{ 474, 463, 54 },
		{ 474, 464, 54 },
		{ 474, 465, 54 },
		{ 474, 466, 54 },
		{ 474, 467, 54 },
		{ 474, 468, 54 },
		{ 474, 469, 54 },
		{ 474, 470, 54 },
		{ 474, 471, 54 },
		{ 474, 472, 54 },
		{ 474, 473, 54 },
		{ 474, 474, 54 },
		{ 474, 475, 54 },
		{ 474, 476, 54 },
		{ 474, 477, 54 },
		{ 474, 478, 54 },
		{ 474, 479, 54 },
		{ 474, 480, 54 },
		{ 474, 481, 54 },
		{ 474, 482, 54 },
		{ 474, 483, 54 },
		{ 474, 484, 54 },
		{ 474, 485, 54 },
		{ 474, 486, 54 },
		{ 474, 487, 54 },
		{ 474, 488, 54 },
		{ 474, 489, 54 },
		{ 474, 490, 54 },
		{ 474, 491, 54 },
		{ 474, 492, 54 },
		{ 474, 493, 54 },
		{ 474, 494, 54 },
		{ 474, 495, 54 },
		{ 474, 496, 54 },
		{ 474, 497, 54 },
		{ 474, 498, 54 },
		{ 474, 499, 54 },
		{ 474, 500, 54 },
		{ 474, 501, 54 },
		{ 474, 502, 54 },
		{ 474, 503, 54 },
		{ 474, 504, 54 },
		{ 475, 448, 54 },
		{ 475, 449, 54 },
		{ 475, 450, 54 },
		{ 475, 451, 54 },
		{ 475, 452, 54 },
		{ 475, 453, 54 },
		{ 475, 454, 54 },
		{ 475, 455, 54 },
		{ 475, 456, 54 },
		{ 475, 457, 54 },
		{ 475, 458, 54 },
		{ 475, 459, 54 },
		{ 475, 460, 54 },
		{ 475, 461, 54 },
		{ 475, 462, 54 },
		{ 475, 463, 54 },
		{ 475, 464, 54 },
		{ 475, 465, 54 },
		{ 475, 466, 54 },
		{ 475, 467, 54 },
		{ 475, 468, 54 },
		{ 475, 469, 54 },
		{ 475, 470, 54 },
		{ 475, 471, 54 },
		{ 475, 472, 54 },
		{ 475, 473, 54 },
		{ 475, 474, 54 },
		{ 475, 475, 54 },
		{ 475, 476, 54 },
		{ 475, 477, 54 },
		{ 475, 478, 54 },
		{ 475, 479, 54 },
		{ 475, 480, 54 },
		{ 475, 481, 54 },
		{ 475, 482, 54 },
		{ 475, 483, 54 },
		{ 475, 484, 54 },
		{ 475, 485, 54 },
		{ 475, 486, 54 },
		{ 475, 487, 54 },
		{ 475, 488, 54 },
		{ 475, 489, 54 },
		{ 475, 490, 54 },
		{ 475, 491, 54 },
		{ 475, 492, 54 },
		{ 475, 493, 54 },
		{ 475, 494, 54 },
		{ 475, 495, 54 },
		{ 475, 496, 54 },
		{ 475, 497, 54 },
		{ 475, 498, 54 },
		{ 475, 499, 54 },
		{ 475, 500, 54 },
		{ 475, 501, 54 },
		{ 475, 502, 54 },
		{ 475, 503, 54 },
		{ 475, 504, 54 },
		{ 475, 505, 54 },
		{ 475, 506, 54 },
		{ 476, 449, 54 },
		{ 476, 450, 54 },
		{ 476, 451, 54 },
		{ 476, 452, 54 },
		{ 476, 453, 54 },
		{ 476, 454, 54 },
		{ 476, 455, 54 },
		{ 476, 456, 54 },
		{ 476, 457, 54 },
		{ 476, 458, 54 },
		{ 476, 459, 54 },
		{ 476, 460, 54 },
		{ 476, 461, 54 },
		{ 476, 462, 54 },
		{ 476, 463, 54 },
		{ 476, 464, 54 },
		{ 476, 465, 54 },
		{ 476, 466, 54 },
		{ 476, 467, 54 },
		{ 476, 468, 54 },
		{ 476, 469, 54 },
		{ 476, 470, 54 },
		{ 476, 471, 54 },
		{ 476, 472, 54 },
		{ 476, 473, 54 },
		{ 476, 474, 54 },
		{ 476, 475, 54 },
		{ 476, 476, 54 },
		{ 476, 477, 54 },
		{ 476, 478, 54 },
		{ 476, 479, 54 },
		{ 476, 480, 54 },
		{ 476, 481, 54 },
		{ 476, 482, 54 },
		{ 476, 483, 54 },
		{ 476, 484, 54 },
		{ 476, 485, 54 },
		{ 476, 486, 54 },
		{ 476, 487, 54 },
		{ 476, 488, 54 },
		{ 476, 489, 54 },
		{ 476, 490, 54 },
		{ 476, 491, 54 },
		{ 476, 492, 54 },
		{ 476, 493, 54 },
		{ 476, 494, 54 },
		{ 476, 495, 54 },
		{ 476, 496, 54 },
		{ 476, 497, 54 },
		{ 476, 498, 54 },
		{ 476, 499, 54 },
		{ 476, 500, 54 },
		{ 476, 501, 54 },
		{ 476, 502, 54 },
		{ 476, 503, 54 },
		{ 476, 504, 54 },
		{ 476, 505, 54 },
		{ 476, 506, 54 },
		{ 476, 507, 54 },
		{ 477, 450, 54 },
		{ 477, 451, 54 },
		{ 477, 452, 54 },
		{ 477, 453, 54 },
		{ 477, 454, 54 },
		{ 477, 455, 54 },
		{ 477, 456, 54 },
		{ 477, 457, 54 },
		{ 477, 458, 54 },
		{ 477, 459, 54 },
		{ 477, 460, 54 },
		{ 477, 461, 54 },
		{ 477, 462, 54 },
		{ 477, 463, 54 },
		{ 477, 464, 54 },
		{ 477, 465, 54 },
		{ 477, 466, 54 },
		{ 477, 467, 54 },
		{ 477, 468, 54 },
		{ 477, 469, 54 },
		{ 477, 470, 54 },
		{ 477, 471, 54 },
		{ 477, 472, 54 },
		{ 477, 473, 54 },
		{ 477, 474, 54 },
		{ 477, 475, 54 },
		{ 477, 476, 54 },
		{ 477, 477, 54 },
		{ 477, 478, 54 },
		{ 477, 479, 54 },
		{ 477, 480, 54 },
		{ 477, 481, 54 },
		{ 477, 482, 54 },
		{ 477, 483, 54 },
		{ 477, 484, 54 },
		{ 477, 485, 54 },
		{ 477, 486, 54 },
		{ 477, 487, 54 },
		{ 477, 488, 54 },
		{ 477, 489, 54 },
		{ 477, 490, 54 },
		{ 477, 491, 54 },
		{ 477, 492, 54 },
		{ 477, 493, 54 },
		{ 477, 494, 54 },
		{ 477, 495, 54 },
		{ 477, 496, 54 },
		{ 477, 497, 54 },
		{ 477, 498, 54 },
		{ 477, 499, 54 },
		{ 477, 500, 54 },
		{ 477, 501, 54 },
		{ 477, 502, 54 },
		{ 477, 503, 54 },
		{ 477, 504, 54 },
		{ 477, 505, 54 },
		{ 477, 506, 54 },
		{ 477, 507, 54 },
		{ 477, 508, 54 },
		{ 478, 451, 54 },
		{ 478, 452, 54 },
		{ 478, 453, 54 },
		{ 478, 454, 54 },
		{ 478, 455, 54 },
		{ 478, 456, 54 },
		{ 478, 457, 54 },
		{ 478, 458, 54 },
		{ 478, 459, 54 },
		{ 478, 460, 54 },
		{ 478, 461, 54 },
		{ 478, 462, 54 },
		{ 478, 463, 54 },
		{ 478, 464, 54 },
		{ 478, 465, 54 },
		{ 478, 466, 54 },
		{ 478, 467, 54 },
		{ 478, 468, 54 },
		{ 478, 469, 54 },
		{ 478, 470, 54 },
		{ 478, 471, 54 },
		{ 478, 472, 54 },
		{ 478, 473, 54 },
		{ 478, 474, 54 },
		{ 478, 475, 54 },
		{ 478, 476, 54 },
		{ 478, 477, 54 },
		{ 478, 478, 54 },
		{ 478, 479, 54 },
		{ 478, 480, 54 },
		{ 478, 481, 54 },
		{ 478, 482, 54 },
		{ 478, 483, 54 },
		{ 478, 484, 54 },
		{ 478, 485, 54 },
		{ 478, 486, 54 },
		{ 478, 487, 54 },
		{ 478, 488, 54 },
		{ 478, 489, 54 },
		{ 478, 490, 54 },
		{ 478, 491, 54 },
		{ 478, 492, 54 },
		{ 478, 493, 54 },
		{ 478, 494, 54 },
		{ 478, 495, 54 },
		{ 478, 496, 54 },
		{ 478, 497, 54 },
		{ 478, 498, 54 },
		{ 478, 499, 54 },
		{ 478, 500, 54 },
		{ 478, 501, 54 },
		{ 478, 502, 54 },
		{ 478, 503, 54 },
		{ 478, 504, 54 },
		{ 478, 505, 54 },
		{ 478, 506, 54 },
		{ 478, 507, 54 },
		{ 478, 508, 54 },
		{ 478, 509, 54 },
		{ 479, 452, 54 },
		{ 479, 453, 54 },
		{ 479, 454, 54 },
		{ 479, 455, 54 },
		{ 479, 456, 54 },
		{ 479, 457, 54 },
		{ 479, 458, 54 },
		{ 479, 459, 54 },
		{ 479, 460, 54 },
		{ 479, 461, 54 },
		{ 479, 462, 54 },
		{ 479, 463, 54 },
		{ 479, 464, 54 },
		{ 479, 465, 54 },
		{ 479, 466, 54 },
		{ 479, 467, 54 },
		{ 479, 468, 54 },
		{ 479, 469, 54 },
		{ 479, 470, 54 },
		{ 479, 471, 54 },
		{ 479, 472, 54 },
		{ 479, 473, 54 },
		{ 479, 474, 54 },
		{ 479, 475, 54 },
		{ 479, 476, 54 },
		{ 479, 477, 54 },
		{ 479, 478, 54 },
		{ 479, 479, 54 },
		{ 479, 480, 54 },
		{ 479, 481, 54 },
		{ 479, 482, 54 },
		{ 479, 483, 54 },
		{ 479, 484, 54 },
		{ 479, 485, 54 },
		{ 479, 486, 54 },
		{ 479, 487, 54 },
		{ 479, 488, 54 },
		{ 479, 489, 54 },
		{ 479, 490, 54 },
		{ 479, 491, 54 },
		{ 479, 492, 54 },
		{ 479, 493, 54 },
		{ 479, 494, 54 },
		{ 479, 495, 54 },
		{ 479, 496, 54 },
		{ 479, 497, 54 },
		{ 479, 498, 54 },
		{ 479, 499, 54 },
		{ 479, 500, 54 },
		{ 479, 501, 54 },
		{ 479, 502, 54 },
		{ 479, 503, 54 },
		{ 479, 504, 54 },
		{ 479, 505, 54 },
		{ 479, 506, 54 },
		{ 479, 507, 54 },
		{ 479, 508, 54 },
		{ 479, 509, 54 },
		{ 479, 510, 54 },
		{ 480, 453, 54 },
		{ 480, 454, 54 },
		{ 480, 455, 54 },
		{ 480, 456, 54 },
		{ 480, 457, 54 },
		{ 480, 458, 54 },
		{ 480, 459, 54 },
		{ 480, 460, 54 },
		{ 480, 461, 54 },
		{ 480, 462, 54 },
		{ 480, 463, 54 },
		{ 480, 464, 54 },
		{ 480, 465, 54 },
		{ 480, 466, 54 },
		{ 480, 467, 54 },
		{ 480, 468, 54 },
		{ 480, 469, 54 },
		{ 480, 470, 54 },
		{ 480, 471, 54 },
		{ 480, 472, 54 },
		{ 480, 473, 54 },
		{ 480, 474, 54 },
		{ 480, 475, 54 },
		{ 480, 476, 54 },
		{ 480, 477, 54 },
		{ 480, 478, 54 },
		{ 480, 479, 54 },
		{ 480, 480, 54 },
		{ 480, 481, 54 },
		{ 480, 482, 54 },
		{ 480, 483, 54 },
		{ 480, 484, 54 },
		{ 480, 485, 54 },
		{ 480, 486, 54 },
		{ 480, 487, 54 },
		{ 480, 488, 54 },
		{ 480, 489, 54 },
		{ 480, 490, 54 },
		{ 480, 491, 54 },
		{ 480, 492, 54 },
		{ 480, 493, 54 },
		{ 480, 494, 54 },
		{ 480, 495, 54 },
		{ 480, 496, 54 },
		{ 480, 497, 54 },
		{ 480, 498, 54 },
		{ 480, 499, 54 },
		{ 480, 500, 54 },
		{ 480, 501, 54 },
		{ 480, 502, 54 },
		{ 480, 503, 54 },
		{ 480, 504, 54 },
		{ 480, 505, 54 },
		{ 480, 506, 54 },
		{ 480, 507, 54 },
		{ 480, 508, 54 },
		{ 480, 509, 54 },
		{ 480, 510, 54 },
		{ 480, 511, 54 },
		{ 481, 454, 54 },
		{ 481, 455, 54 },
		{ 481, 456, 54 },
		{ 481, 457, 54 },
		{ 481, 458, 54 },
		{ 481, 459, 54 },
		{ 481, 460, 54 },
		{ 481, 461, 54 },
		{ 481, 462, 54 },
		{ 481, 463, 54 },
		{ 481, 464, 54 },
		{ 481, 465, 54 },
		{ 481, 466, 54 },
		{ 481, 467, 54 },
		{ 481, 468, 54 },
		{ 481, 469, 54 },
		{ 481, 470, 54 },
		{ 481, 471, 54 },
		{ 481, 472, 54 },
		{ 481, 473, 54 },
		{ 481, 474, 54 },
		{ 481, 475, 54 },
		{ 481, 476, 54 },
		{ 481, 477, 54 },
		{ 481, 478, 54 },
		{ 481, 479, 54 },
		{ 481, 480, 54 },
		{ 481, 481, 54 },
		{ 481, 482, 54 },
		{ 481, 483, 54 },
		{ 481, 484, 54 },
		{ 481, 485, 54 },
		{ 481, 486, 54 },
		{ 481, 487, 54 },
		{ 481, 488, 54 },
		{ 481, 489, 54 },
		{ 481, 490, 54 },
		{ 481, 491, 54 },
		{ 481, 492, 54 },
		{ 481, 493, 54 },
		{ 481, 494, 54 },
		{ 481, 495, 54 },
		{ 481, 496, 54 },
		{ 481, 497, 54 },
		{ 481, 498, 54 },
		{ 481, 499, 54 },
		{ 481, 500, 54 },
		{ 481, 501, 54 },
		{ 481, 502, 54 },
		{ 481, 503, 54 },
		{ 481, 504, 54 },
		{ 481, 505, 54 },
		{ 481, 506, 54 },
		{ 481, 507, 54 },
		{ 481, 508, 54 },
		{ 481, 509, 54 },
		{ 481, 510, 54 },
		{ 481, 511, 54 },
		{ 481, 512, 54 },
		{ 482, 454, 54 },
		{ 482, 455, 54 },
		{ 482, 456, 54 },
		{ 482, 457, 54 },
		{ 482, 458, 54 },
		{ 482, 459, 54 },
		{ 482, 460, 54 },
		{ 482, 461, 54 },
		{ 482, 462, 54 },
		{ 482, 463, 54 },
		{ 482, 464, 54 },
		{ 482, 465, 54 },
		{ 482, 466, 54 },
		{ 482, 467, 54 },
		{ 482, 468, 54 },
		{ 482, 469, 54 },
		{ 482, 470, 54 },
		{ 482, 471, 54 },
		{ 482, 472, 54 },
		{ 482, 473, 54 },
		{ 482, 474, 54 },
		{ 482, 475, 54 },
		{ 482, 476, 54 },
		{ 482, 477, 54 },
		{ 482, 478, 54 },
		{ 482, 479, 54 },
		{ 482, 480, 54 },
		{ 482, 481, 54 },
		{ 482, 482, 54 },
		{ 482, 483, 54 },
		{ 482, 484, 54 },
		{ 482, 485, 54 },
		{ 482, 486, 54 },
		{ 482, 487, 54 },
		{ 482, 488, 54 },
		{ 482, 489, 54 },
		{ 482, 490, 54 },
		{ 482, 491, 54 },
		{ 482, 492, 54 },
		{ 482, 493, 54 },
		{ 482, 494, 54 },
		{ 482, 495, 54 },
		{ 482, 496, 54 },
		{ 482, 497, 54 },
		{ 482, 498, 54 },
		{ 482, 499, 54 },
		{ 482, 500, 54 },
		{ 482, 501, 54 },
		{ 482, 502, 54 },
		{ 482, 503, 54 },
		{ 482, 504, 54 },
		{ 482, 505, 54 },
		{ 482, 506, 54 },
		{ 482, 507, 54 },
		{ 482, 508, 54 },
		{ 482, 509, 54 },
		{ 482, 510, 54 },
		{ 482, 511, 54 },
		{ 482, 512, 54 },
		{ 483, 455, 54 },
		{ 483, 456, 54 },
		{ 483, 457, 54 },
		{ 483, 458, 54 },
		{ 483, 459, 54 },
		{ 483, 460, 54 },
		{ 483, 461, 54 },
		{ 483, 462, 54 },
		{ 483, 463, 54 },
		{ 483, 464, 54 },
		{ 483, 465, 54 },
		{ 483, 466, 54 },
		{ 483, 467, 54 },
		{ 483, 468, 54 },
		{ 483, 469, 54 },
		{ 483, 470, 54 },
		{ 483, 471, 54 },
		{ 483, 472, 54 },
		{ 483, 473, 54 },
		{ 483, 474, 54 },
		{ 483, 475, 54 },
		{ 483, 476, 54 },
		{ 483, 477, 54 },
		{ 483, 478, 54 },
		{ 483, 479, 54 },
		{ 483, 480, 54 },
		{ 483, 481, 54 },
		{ 483, 482, 54 },
		{ 483, 483, 54 },
		{ 483, 484, 54 },
		{ 483, 485, 54 },
		{ 483, 486, 54 },
		{ 483, 487, 54 },
		{ 483, 488, 54 },
		{ 483, 489, 54 },
		{ 483, 490, 54 },
		{ 483, 491, 54 },
		{ 483, 492, 54 },
		{ 483, 493, 54 },
		{ 483, 494, 54 },
		{ 483, 495, 54 },
		{ 483, 496, 54 },
		{ 483, 497, 54 },
		{ 483, 498, 54 },
		{ 483, 499, 54 },
		{ 483, 500, 54 },
		{ 483, 501, 54 },
		{ 483, 502, 54 },
		{ 483, 503, 54 },
		{ 483, 504, 54 },
		{ 483, 505, 54 },
		{ 483, 506, 54 },
		{ 483, 507, 54 },
		{ 483, 508, 54 },
		{ 483, 509, 54 },
		{ 483, 510, 54 },
		{ 483, 511, 54 },
		{ 483, 512, 54 },
		{ 484, 456, 54 },
		{ 484, 457, 54 },
		{ 484, 458, 54 },
		{ 484, 459, 54 },
		{ 484, 460, 54 },
		{ 484, 461, 54 },
		{ 484, 462, 54 },
		{ 484, 463, 54 },
		{ 484, 464, 54 },
		{ 484, 465, 54 },
		{ 484, 466, 54 },
		{ 484, 467, 54 },
		{ 484, 468, 54 },
		{ 484, 469, 54 },
		{ 484, 470, 54 },
		{ 484, 471, 54 },
		{ 484, 472, 54 },
		{ 484, 473, 54 },
		{ 484, 474, 54 },
		{ 484, 475, 54 },
		{ 484, 476, 54 },
		{ 484, 477, 54 },
		{ 484, 478, 54 },
		{ 484, 479, 54 },
		{ 484, 480, 54 },
		{ 484, 481, 54 },
		{ 484, 482, 54 },
		{ 484, 483, 54 },
		{ 484, 484, 54 },
		{ 484, 485, 54 },
		{ 484, 486, 54 },
		{ 484, 487, 54 },
		{ 484, 488, 54 },
		{ 484, 489, 54 },
		{ 484, 490, 54 },
		{ 484, 491, 54 },
		{ 484, 492, 54 },
		{ 484, 493, 54 },
		{ 484, 494, 54 },
		{ 484, 495, 54 },
		{ 484, 496, 54 },
		{ 484, 497, 54 },
		{ 484, 498, 54 },
		{ 484, 499, 54 },
		{ 484, 500, 54 },
		{ 484, 501, 54 },
		{ 484, 502, 54 },
		{ 484, 503, 54 },
		{ 484, 504, 54 },
		{ 484, 505, 54 },
		{ 484, 506, 54 },
		{ 484, 507, 54 },
		{ 484, 508, 54 },
		{ 484, 509, 54 },
		{ 484, 510, 54 },
		{ 484, 511, 54 },
		{ 484, 512, 54 },
		{ 485, 457, 54 },
		{ 485, 458, 54 },
		{ 485, 459, 54 },
		{ 485, 460, 54 },
		{ 485, 461, 54 },
		{ 485, 462, 54 },
		{ 485, 463, 54 },
		{ 485, 464, 54 },
		{ 485, 465, 54 },
		{ 485, 466, 54 },
		{ 485, 467, 54 },
		{ 485, 468, 54 },
		{ 485, 469, 54 },
		{ 485, 470, 54 },
		{ 485, 471, 54 },
		{ 485, 472, 54 },
		{ 485, 473, 54 },
		{ 485, 474, 54 },
		{ 485, 475, 54 },
		{ 485, 476, 54 },
		{ 485, 477, 54 },
		{ 485, 478, 54 },
		{ 485, 479, 54 },
		{ 485, 480, 54 },
		{ 485, 481, 54 },
		{ 485, 482, 54 },
		{ 485, 483, 54 },
		{ 485, 484, 54 },
		{ 485, 485, 54 },
		{ 485, 486, 54 },
		{ 485, 487, 54 },
		{ 485, 488, 54 },
		{ 485, 489, 54 },
		{ 485, 490, 54 },
		{ 485, 491, 54 },
		{ 485, 492, 54 },
		{ 485, 493, 54 },
		{ 485, 494, 54 },
		{ 485, 495, 54 },
		{ 485, 496, 54 },
		{ 485, 497, 54 },
		{ 485, 498, 54 },
		{ 485, 499, 54 },
		{ 485, 500, 54 },
		{ 485, 501, 54 },
		{ 485, 502, 54 },
		{ 485, 503, 54 },
		{ 485, 504, 54 },
		{ 485, 505, 54 },
		{ 485, 506, 54 },
		{ 485, 507, 54 },
		{ 485, 508, 54 },
		{ 485, 509, 54 },
		{ 485, 510, 54 },
		{ 485, 511, 54 },
		{ 485, 512, 54 },
		{ 486, 458, 54 },
		{ 486, 459, 54 },
		{ 486, 460, 54 },
		{ 486, 461, 54 },
		{ 486, 462, 54 },
		{ 486, 463, 54 },
		{ 486, 464, 54 },
		{ 486, 465, 54 },
		{ 486, 466, 54 },
		{ 486, 467, 54 },
		{ 486, 468, 54 },
		{ 486, 469, 54 },
		{ 486, 470, 54 },
		{ 486, 471, 54 },
		{ 486, 472, 54 },
		{ 486, 473, 54 },
		{ 486, 474, 54 },
		{ 486, 475, 54 },
		{ 486, 476, 54 },
		{ 486, 477, 54 },
		{ 486, 478, 54 },
		{ 486, 479, 54 },
		{ 486, 480, 54 },
		{ 486, 481, 54 },
		{ 486, 482, 54 },
		{ 486, 483, 54 },
		{ 486, 484, 54 },
		{ 486, 485, 54 },
		{ 486, 486, 54 },
		{ 486, 487, 54 },
		{ 486, 488, 54 },
		{ 486, 489, 54 },
		{ 486, 490, 54 },
		{ 486, 491, 54 },
		{ 486, 492, 54 },
		{ 486, 493, 54 },
		{ 486, 494, 54 },
		{ 486, 495, 54 },
		{ 486, 496, 54 },
		{ 486, 497, 54 },
		{ 486, 498, 54 },
		{ 486, 499, 54 },
		{ 486, 500, 54 },
		{ 486, 501, 54 },
		{ 486, 502, 54 },
		{ 486, 503, 54 },
		{ 486, 504, 54 },
		{ 486, 505, 54 },
		{ 486, 506, 54 },
		{ 486, 507, 54 },
		{ 486, 508, 54 },
		{ 486, 509, 54 },
		{ 486, 510, 54 },
		{ 486, 511, 54 },
		{ 486, 512, 54 },
		{ 487, 459, 54 },
		{ 487, 460, 54 },
		{ 487, 461, 54 },
		{ 487, 462, 54 },
		{ 487, 463, 54 },
		{ 487, 464, 54 },
		{ 487, 465, 54 },
		{ 487, 466, 54 },
		{ 487, 467, 54 },
		{ 487, 468, 54 },
		{ 487, 469, 54 },
		{ 487, 470, 54 },
		{ 487, 471, 54 },
		{ 487, 472, 54 },
		{ 487, 473, 54 },
		{ 487, 474, 54 },
		{ 487, 475, 54 },
		{ 487, 476, 54 },
		{ 487, 477, 54 },
		{ 487, 478, 54 },
		{ 487, 479, 54 },
		{ 487, 480, 54 },
		{ 487, 481, 54 },
		{ 487, 482, 54 },
		{ 487, 483, 54 },
		{ 487, 484, 54 },
		{ 487, 485, 54 },
		{ 487, 486, 54 },
		{ 487, 487, 54 },
		{ 487, 488, 54 },
		{ 487, 489, 54 },
		{ 487, 490, 54 },
		{ 487, 491, 54 },
		{ 487, 492, 54 },
		{ 487, 493, 54 },
		{ 487, 494, 54 },
		{ 487, 495, 54 },
		{ 487, 496, 54 },
		{ 487, 497, 54 },
		{ 487, 498, 54 },
		{ 487, 499, 54 },
		{ 487, 500, 54 },
		{ 487, 501, 54 },
		{ 487, 502, 54 },
		{ 487, 503, 54 },
		{ 487, 504, 54 },
		{ 487, 505, 54 },
		{ 487, 506, 54 },
		{ 487, 507, 54 },
		{ 487, 508, 54 },
		{ 487, 509, 54 },
		{ 487, 510, 54 },
		{ 487, 511, 54 },
		{ 487, 512, 54 },
		{ 488, 460, 54 },
		{ 488, 461, 54 },
		{ 488, 462, 54 },
		{ 488, 463, 54 },
		{ 488, 464, 54 },
		{ 488, 465, 54 },
		{ 488, 466, 54 },
		{ 488, 467, 54 },
		{ 488, 468, 54 },
		{ 488, 469, 54 },
		{ 488, 470, 54 },
		{ 488, 471, 54 },
		{ 488, 472, 54 },
		{ 488, 473, 54 },
		{ 488, 474, 54 },
		{ 488, 475, 54 },
		{ 488, 476, 54 },
		{ 488, 477, 54 },
		{ 488, 478, 54 },
		{ 488, 479, 54 },
		{ 488, 480, 54 },
		{ 488, 481, 54 },
		{ 488, 482, 54 },
		{ 488, 483, 54 },
		{ 488, 484, 54 },
		{ 488, 485, 54 },
		{ 488, 486, 54 },
		{ 488, 487, 54 },
		{ 488, 488, 54 },
		{ 488, 489, 54 },
		{ 488, 490, 54 },
		{ 488, 491, 54 },
		{ 488, 492, 54 },
		{ 488, 493, 54 },
		{ 488, 494, 54 },
		{ 488, 495, 54 },
		{ 488, 496, 54 },
		{ 488, 497, 54 },
		{ 488, 498, 54 },
		{ 488, 499, 54 },
		{ 488, 500, 54 },
		{ 488, 501, 54 },
		{ 488, 502, 54 },
		{ 488, 503, 54 },
		{ 488, 504, 54 },
		{ 488, 505, 54 },
		{ 488, 506, 54 },
		{ 488, 507, 54 },
		{ 488, 508, 54 },
		{ 488, 509, 54 },
		{ 488, 510, 54 },
		{ 488, 511, 54 },
		{ 488, 512, 54 },
		{ 489, 461, 54 },
		{ 489, 462, 54 },
		{ 489, 463, 54 },
		{ 489, 464, 54 },
		{ 489, 465, 54 },
		{ 489, 466, 54 },
		{ 489, 467, 54 },
		{ 489, 468, 54 },
		{ 489, 469, 54 },
		{ 489, 470, 54 },
		{ 489, 471, 54 },
		{ 489, 472, 54 },
		{ 489, 473, 54 },
		{ 489, 474, 54 },
		{ 489, 475, 54 },
		{ 489, 476, 54 },
		{ 489, 477, 54 },
		{ 489, 478, 54 },
		{ 489, 479, 54 },
		{ 489, 480, 54 },
		{ 489, 481, 54 },
		{ 489, 482, 54 },
		{ 489, 483, 54 },
		{ 489, 484, 54 },
		{ 489, 485, 54 },
		{ 489, 486, 54 },
		{ 489, 487, 54 },
		{ 489, 488, 54 },
		{ 489, 489, 54 },
		{ 489, 490, 54 },
		{ 489, 491, 54 },
		{ 489, 492, 54 },
		{ 489, 493, 54 },
		{ 489, 494, 54 },
		{ 489, 495, 54 },
		{ 489, 496, 54 },
		{ 489, 497, 54 },
		{ 489, 498, 54 },
		{ 489, 499, 54 },
		{ 489, 500, 54 },
		{ 489, 501, 54 },
		{ 489, 502, 54 },
		{ 489, 503, 54 },
		{ 489, 504, 54 },
		{ 489, 505, 54 },
		{ 489, 506, 54 },
		{ 489, 507, 54 },
		{ 489, 508, 54 },
		{ 489, 509, 54 },
		{ 489, 510, 54 },
		{ 489, 511, 54 },
		{ 489, 512, 54 },
		{ 490, 461, 54 },
		{ 490, 462, 54 },
		{ 490, 463, 54 },
		{ 490, 464, 54 },
		{ 490, 465, 54 },
		{ 490, 466, 54 },
		{ 490, 467, 54 },
		{ 490, 468, 54 },
		{ 490, 469, 54 },
		{ 490, 470, 54 },
		{ 490, 471, 54 },
		{ 490, 472, 54 },
		{ 490, 473, 54 },
		{ 490, 474, 54 },
		{ 490, 475, 54 },
		{ 490, 476, 54 },
		{ 490, 477, 54 },
		{ 490, 478, 54 },
		{ 490, 479, 54 },
		{ 490, 480, 54 },
		{ 490, 481, 54 },
		{ 490, 482, 54 },
		{ 490, 483, 54 },
		{ 490, 484, 54 },
		{ 490, 485, 54 },
		{ 490, 486, 54 },
		{ 490, 487, 54 },
		{ 490, 488, 54 },
		{ 490, 489, 54 },
		{ 490, 490, 54 },
		{ 490, 491, 54 },
		{ 490, 492, 54 },
		{ 490, 493, 54 },
		{ 490, 494, 54 },
		{ 490, 495, 54 },
		{ 490, 496, 54 },
		{ 490, 497, 54 },
		{ 490, 498, 54 },
		{ 490, 499, 54 },
		{ 490, 500, 54 },
		{ 490, 501, 54 },
		{ 490, 502, 54 },
		{ 490, 503, 54 },
		{ 490, 504, 54 },
		{ 490, 505, 54 },
		{ 490, 506, 54 },
		{ 490, 507, 54 },
		{ 490, 508, 54 },
		{ 490, 509, 54 },
		{ 490, 510, 54 },
		{ 490, 511, 54 },
		{ 490, 512, 54 },
		{ 491, 462, 54 },
		{ 491, 463, 54 },
		{ 491, 464, 54 },
		{ 491, 465, 54 },
		{ 491, 466, 54 },
		{ 491, 467, 54 },
		{ 491, 468, 54 },
		{ 491, 469, 54 },
		{ 491, 470, 54 },
		{ 491, 471, 54 },
		{ 491, 472, 54 },
		{ 491, 473, 54 },
		{ 491, 474, 54 },
		{ 491, 475, 54 },
		{ 491, 476, 54 },
		{ 491, 477, 54 },
		{ 491, 478, 54 },
		{ 491, 479, 54 },
		{ 491, 480, 54 },
		{ 491, 481, 54 },
		{ 491, 482, 54 },
		{ 491, 483, 54 },
		{ 491, 484, 54 },
		{ 491, 485, 54 },
		{ 491, 486, 54 },
		{ 491, 487, 54 },
		{ 491, 488, 54 },
		{ 491, 489, 54 },
		{ 491, 490, 54 },
		{ 491, 491, 54 },
		{ 491, 492, 54 },
		{ 491, 493, 54 },
		{ 491, 494, 54 },
		{ 491, 495, 54 },
		{ 491, 496, 54 },
		{ 491, 497, 54 },
		{ 491, 498, 54 },
		{ 491, 499, 54 },
		{ 491, 500, 54 },
		{ 491, 501, 54 },
		{ 491, 502, 54 },
		{ 491, 503, 54 },
		{ 491, 504, 54 },
		{ 491, 505, 54 },
		{ 491, 506, 54 },
		{ 491, 507, 54 },
		{ 491, 508, 54 },
		{ 491, 509, 54 },
		{ 491, 510, 54 },
		{ 491, 511, 54 },
		{ 491, 512, 54 },
		{ 492, 463, 54 },
		{ 492, 464, 54 },
		{ 492, 465, 54 },
		{ 492, 466, 54 },
		{ 492, 467, 54 },
		{ 492, 468, 54 },
		{ 492, 469, 54 },
		{ 492, 470, 54 },
		{ 492, 471, 54 },
		{ 492, 472, 54 },
		{ 492, 473, 54 },
		{ 492, 474, 54 },
		{ 492, 475, 54 },
		{ 492, 476, 54 },
		{ 492, 477, 54 },
		{ 492, 478, 54 },
		{ 492, 479, 54 },
		{ 492, 480, 54 },
		{ 492, 481, 54 },
		{ 492, 482, 54 },
		{ 492, 483, 54 },
		{ 492, 484, 54 },
		{ 492, 485, 54 },
		{ 492, 486, 54 },
		{ 492, 487, 54 },
		{ 492, 488, 54 },
		{ 492, 489, 54 },
		{ 492, 490, 54 },
		{ 492, 491, 54 },
		{ 492, 492, 54 },
		{ 492, 493, 54 },
		{ 492, 494, 54 },
		{ 492, 495, 54 },
		{ 492, 496, 54 },
		{ 492, 497, 54 },
		{ 492, 498, 54 },
		{ 492, 499, 54 },
		{ 492, 500, 54 },
		{ 492, 501, 54 },
		{ 492, 502, 54 },
		{ 492, 503, 54 },
		{ 492, 504, 54 },
		{ 492, 505, 54 },
		{ 492, 506, 54 },
		{ 492, 507, 54 },
		{ 492, 508, 54 },
		{ 492, 509, 54 },
		{ 492, 510, 54 },
		{ 492, 511, 54 },
		{ 492, 512, 54 },
		{ 493, 464, 54 },
		{ 493, 465, 54 },
		{ 493, 466, 54 },
		{ 493, 467, 54 },
		{ 493, 468, 54 },
		{ 493, 469, 54 },
		{ 493, 470, 54 },
		{ 493, 471, 54 },
		{ 493, 472, 54 },
		{ 493, 473, 54 },
		{ 493, 474, 54 },
		{ 493, 475, 54 },
		{ 493, 476, 54 },
		{ 493, 477, 54 },
		{ 493, 478, 54 },
		{ 493, 479, 54 },
		{ 493, 480, 54 },
		{ 493, 481, 54 },
		{ 493, 482, 54 },
		{ 493, 483, 54 },
		{ 493, 484, 54 },
		{ 493, 485, 54 },
		{ 493, 486, 54 },
		{ 493, 487, 54 },
		{ 493, 488, 54 },
		{ 493, 489, 54 },
		{ 493, 490, 54 },
		{ 493, 491, 54 },
		{ 493, 492, 54 },
		{ 493, 493, 54 },
		{ 493, 494, 54 },
		{ 493, 495, 54 },
		{ 493, 496, 54 },
		{ 493, 497, 54 },
		{ 493, 498, 54 },
		{ 493, 499, 54 },
		{ 493, 500, 54 },
		{ 493, 501, 54 },
		{ 493, 502, 54 },
		{ 493, 503, 54 },
		{ 493, 504, 54 },
		{ 493, 505, 54 },
		{ 493, 506, 54 },
		{ 493, 507, 54 },
		{ 493, 508, 54 },
		{ 493, 509, 54 },
		{ 493, 510, 54 },
		{ 493, 511, 54 },
		{ 493, 512, 54 },
		{ 494, 465, 54 },
		{ 494, 466, 54 },
		{ 494, 467, 54 },
		{ 494, 468, 54 },
		{ 494, 469, 54 },
		{ 494, 470, 54 },
		{ 494, 471, 54 },
		{ 494, 472, 54 },
		{ 494, 473, 54 },
		{ 494, 474, 54 },
		{ 494, 475, 54 },
		{ 494, 476, 54 },
		{ 494, 477, 54 },
		{ 494, 478, 54 },
		{ 494, 479, 54 },
		{ 494, 480, 54 },
		{ 494, 481, 54 },
		{ 494, 482, 54 },
		{ 494, 483, 54 },
		{ 494, 484, 54 },
		{ 494, 485, 54 },
		{ 494, 486, 54 },
		{ 494, 487, 54 },
		{ 494, 488, 54 },
		{ 494, 489, 54 },
		{ 494, 490, 54 },
		{ 494, 491, 54 },
		{ 494, 492, 54 },
		{ 494, 493, 54 },
		{ 494, 494, 54 },
		{ 494, 495, 54 },
		{ 494, 496, 54 },
		{ 494, 497, 54 },
		{ 494, 498, 54 },
		{ 494, 499, 54 },
		{ 494, 500, 54 },
		{ 494, 501, 54 },
		{ 494, 502, 54 },
		{ 494, 503, 54 },
		{ 494, 504, 54 },
		{ 494, 505, 54 },
		{ 494, 506, 54 },
		{ 494, 507, 54 },
		{ 494, 508, 54 },
		{ 494, 509, 54 },
		{ 494, 510, 54 },
		{ 494, 511, 54 },
		{ 494, 512, 54 },
		{ 495, 466, 54 },
		{ 495, 467, 54 },
		{ 495, 468, 54 },
		{ 495, 469, 54 },
		{ 495, 470, 54 },
		{ 495, 471, 54 },
		{ 495, 472, 54 },
		{ 495, 473, 54 },
		{ 495, 474, 54 },
		{ 495, 475, 54 },
		{ 495, 476, 54 },
		{ 495, 477, 54 },
		{ 495, 478, 54 },
		{ 495, 479, 54 },
		{ 495, 480, 54 },
		{ 495, 481, 54 },
		{ 495, 482, 54 },
		{ 495, 483, 54 },
		{ 495, 484, 54 },
		{ 495, 485, 54 },
		{ 495, 486, 54 },
		{ 495, 487, 54 },
		{ 495, 488, 54 },
		{ 495, 489, 54 },
		{ 495, 490, 54 },
		{ 495, 491, 54 },
		{ 495, 492, 54 },
		{ 495, 493, 54 },
		{ 495, 494, 54 },
		{ 495, 495, 54 },
		{ 495, 496, 54 },
		{ 495, 497, 54 },
		{ 495, 498, 54 },
		{ 495, 499, 54 },
		{ 495, 500, 54 },
		{ 495, 501, 54 },
		{ 495, 502, 54 },
		{ 495, 503, 54 },
		{ 495, 504, 54 },
		{ 495, 505, 54 },
		{ 495, 506, 54 },
		{ 495, 507, 54 },
		{ 495, 508, 54 },
		{ 495, 509, 54 },
		{ 495, 510, 54 },
		{ 495, 511, 54 },
		{ 495, 512, 54 },
		{ 496, 467, 54 },
		{ 496, 468, 54 },
		{ 496, 469, 54 },
		{ 496, 470, 54 },
		{ 496, 471, 54 },
		{ 496, 472, 54 },
		{ 496, 473, 54 },
		{ 496, 474, 54 },
		{ 496, 475, 54 },
		{ 496, 476, 54 },
		{ 496, 477, 54 },
		{ 496, 478, 54 },
		{ 496, 479, 54 },
		{ 496, 480, 54 },
		{ 496, 481, 54 },
		{ 496, 482, 54 },
		{ 496, 483, 54 },
		{ 496, 484, 54 },
		{ 496, 485, 54 },
		{ 496, 486, 54 },
		{ 496, 487, 54 },
		{ 496, 488, 54 },
		{ 496, 489, 54 },
		{ 496, 490, 54 },
		{ 496, 491, 54 },
		{ 496, 492, 54 },
		{ 496, 493, 54 },
		{ 496, 494, 54 },
		{ 496, 495, 54 },
		{ 496, 496, 54 },
		{ 496, 497, 54 },
		{ 496, 498, 54 },
		{ 496, 499, 54 },
		{ 496, 500, 54 },
		{ 496, 501, 54 },
		{ 496, 502, 54 },
		{ 496, 503, 54 },
		{ 496, 504, 54 },
		{ 496, 505, 54 },
		{ 496, 506, 54 },
		{ 496, 507, 54 },
		{ 496, 508, 54 },
		{ 496, 509, 54 },
		{ 496, 510, 54 },
		{ 496, 511, 54 },
		{ 496, 512, 54 },
		{ 497, 468, 54 },
		{ 497, 469, 54 },
		{ 497, 470, 54 },
		{ 497, 471, 54 },
		{ 497, 472, 54 },
		{ 497, 473, 54 },
		{ 497, 474, 54 },
		{ 497, 475, 54 },
		{ 497, 476, 54 },
		{ 497, 477, 54 },
		{ 497, 478, 54 },
		{ 497, 479, 54 },
		{ 497, 480, 54 },
		{ 497, 481, 54 },
		{ 497, 482, 54 },
		{ 497, 483, 54 },
		{ 497, 484, 54 },
		{ 497, 485, 54 },
		{ 497, 486, 54 },
		{ 497, 487, 54 },
		{ 497, 488, 54 },
		{ 497, 489, 54 },
		{ 497, 490, 54 },
		{ 497, 491, 54 },
		{ 497, 492, 54 },
		{ 497, 493, 54 },
		{ 497, 494, 54 },
		{ 497, 495, 54 },
		{ 497, 496, 54 },
		{ 497, 497, 54 },
		{ 497, 498, 54 },
		{ 497, 499, 54 },
		{ 497, 500, 54 },
		{ 497, 501, 54 },
		{ 497, 502, 54 },
		{ 497, 503, 54 },
		{ 497, 504, 54 },
		{ 497, 505, 54 },
		{ 497, 506, 54 },
		{ 497, 507, 54 },
		{ 497, 508, 54 },
		{ 497, 509, 54 },
		{ 497, 510, 54 },
		{ 497, 511, 54 },
		{ 497, 512, 54 },
		{ 498, 468, 54 },
		{ 498, 469, 54 },
		{ 498, 470, 54 },
		{ 498, 471, 54 },
		{ 498, 472, 54 },
		{ 498, 473, 54 },
		{ 498, 474, 54 },
		{ 498, 475, 54 },
		{ 498, 476, 54 },
		{ 498, 477, 54 },
		{ 498, 478, 54 },
		{ 498, 479, 54 },
		{ 498, 480, 54 },
		{ 498, 481, 54 },
		{ 498, 482, 54 },
		{ 498, 483, 54 },
		{ 498, 484, 54 },
		{ 498, 485, 54 },
		{ 498, 486, 54 },
		{ 498, 487, 54 },
		{ 498, 488, 54 },
		{ 498, 489, 54 },
		{ 498, 490, 54 },
		{ 498, 491, 54 },
		{ 498, 492, 54 },
		{ 498, 493, 54 },
		{ 498, 494, 54 },
		{ 498, 495, 54 },
		{ 498, 496, 54 },
		{ 498, 497, 54 },
		{ 498, 498, 54 },
		{ 498, 499, 54 },
		{ 498, 500, 54 },
		{ 498, 501, 54 },
		{ 498, 502, 54 },
		{ 498, 503, 54 },
		{ 498, 504, 54 },
		{ 498, 505, 54 },
		{ 498, 506, 54 },
		{ 498, 507, 54 },
		{ 498, 508, 54 },
		{ 498, 509, 54 },
		{ 498, 510, 54 },
		{ 498, 511, 54 },
		{ 498, 512, 54 },
		{ 499, 469, 54 },
		{ 499, 470, 54 },
		{ 499, 471, 54 },
		{ 499, 472, 54 },
		{ 499, 473, 54 },
		{ 499, 474, 54 },
		{ 499, 475, 54 },
		{ 499, 476, 54 },
		{ 499, 477, 54 },
		{ 499, 478, 54 },
		{ 499, 479, 54 },
		{ 499, 480, 54 },
		{ 499, 481, 54 },
		{ 499, 482, 54 },
		{ 499, 483, 54 },
		{ 499, 484, 54 },
		{ 499, 485, 54 },
		{ 499, 486, 54 },
		{ 499, 487, 54 },
		{ 499, 488, 54 },
		{ 499, 489, 54 },
		{ 499, 490, 54 },
		{ 499, 491, 54 },
		{ 499, 492, 54 },
		{ 499, 493, 54 },
		{ 499, 494, 54 },
		{ 499, 495, 54 },
		{ 499, 496, 54 },
		{ 499, 497, 54 },
		{ 499, 498, 54 },
		{ 499, 499, 54 },
		{ 499, 500, 54 },
		{ 499, 501, 54 },
		{ 499, 502, 54 },
		{ 499, 503, 54 },
		{ 499, 504, 54 },
		{ 499, 505, 54 },
		{ 499, 506, 54 },
		{ 499, 507, 54 },
		{ 499, 508, 54 },
		{ 499, 509, 54 },
		{ 499, 510, 54 },
		{ 499, 511, 54 },
		{ 499, 512, 54 },
		{ 500, 470, 54 },
		{ 500, 471, 54 },
		{ 500, 472, 54 },
		{ 500, 473, 54 },
		{ 500, 474, 54 },
		{ 500, 475, 54 },
		{ 500, 476, 54 },
		{ 500, 477, 54 },
		{ 500, 478, 54 },
		{ 500, 479, 54 },
		{ 500, 480, 54 },
		{ 500, 481, 54 },
		{ 500, 482, 54 },
		{ 500, 483, 54 },
		{ 500, 484, 54 },
		{ 500, 485, 54 },
		{ 500, 486, 54 },
		{ 500, 487, 54 },
		{ 500, 488, 54 },
		{ 500, 489, 54 },
		{ 500, 490, 54 },
		{ 500, 491, 54 },
		{ 500, 492, 54 },
		{ 500, 493, 54 },
		{ 500, 494, 54 },
		{ 500, 495, 54 },
		{ 500, 496, 54 },
		{ 500, 497, 54 },
		{ 500, 498, 54 },
		{ 500, 499, 54 },
		{ 500, 500, 54 },
		{ 500, 501, 54 },
		{ 500, 502, 54 },
		{ 500, 503, 54 },
		{ 500, 504, 54 },
		{ 500, 505, 54 },
		{ 500, 506, 54 },
		{ 500, 507, 54 },
		{ 500, 508, 54 },
		{ 500, 509, 54 },
		{ 500, 510, 54 },
		{ 500, 511, 54 },
		{ 500, 512, 54 },
		{ 501, 471, 54 },
		{ 501, 472, 54 },
		{ 501, 473, 54 },
		{ 501, 474, 54 },
		{ 501, 475, 54 },
		{ 501, 476, 54 },
		{ 501, 477, 54 },
		{ 501, 478, 54 },
		{ 501, 479, 54 },
		{ 501, 480, 54 },
		{ 501, 481, 54 },
		{ 501, 482, 54 },
		{ 501, 483, 54 },
		{ 501, 484, 54 },
		{ 501, 485, 54 },
		{ 501, 486, 54 },
		{ 501, 487, 54 },
		{ 501, 488, 54 },
		{ 501, 489, 54 },
		{ 501, 490, 54 },
		{ 501, 491, 54 },
		{ 501, 492, 54 },
		{ 501, 493, 54 },
		{ 501, 494, 54 },
		{ 501, 495, 54 },
		{ 501, 496, 54 },
		{ 501, 497, 54 },
		{ 501, 498, 54 },
		{ 501, 499, 54 },
		{ 501, 500, 54 },
		{ 501, 501, 54 },
		{ 501, 502, 54 },
		{ 501, 503, 54 },
		{ 501, 504, 54 },
		{ 501, 505, 54 },
		{ 501, 506, 54 },
		{ 501, 507, 54 },
		{ 501, 508, 54 },
		{ 501, 509, 54 },
		{ 501, 510, 54 },
		{ 501, 511, 54 },
		{ 501, 512, 54 },
		{ 502, 472, 54 },
		{ 502, 473, 54 },
		{ 502, 474, 54 },
		{ 502, 475, 54 },
		{ 502, 476, 54 },
		{ 502, 477, 54 },
		{ 502, 478, 54 },
		{ 502, 479, 54 },
		{ 502, 480, 54 },
		{ 502, 481, 54 },
		{ 502, 482, 54 },
		{ 502, 483, 54 },
		{ 502, 484, 54 },
		{ 502, 485, 54 },
		{ 502, 486, 54 },
		{ 502, 487, 54 },
		{ 502, 488, 54 },
		{ 502, 489, 54 },
		{ 502, 490, 54 },
		{ 502, 491, 54 },
		{ 502, 492, 54 },
		{ 502, 493, 54 },
		{ 502, 494, 54 },
		{ 502, 495, 54 },
		{ 502, 496, 54 },
		{ 502, 497, 54 },
		{ 502, 498, 54 },
		{ 502, 499, 54 },
		{ 502, 500, 54 },
		{ 502, 501, 54 },
		{ 502, 502, 54 },
		{ 502, 503, 54 },
		{ 502, 504, 54 },
		{ 502, 505, 54 },
		{ 502, 506, 54 },
		{ 502, 507, 54 },
		{ 502, 508, 54 },
		{ 502, 509, 54 },
		{ 502, 510, 54 },
		{ 502, 511, 54 },
		{ 502, 512, 54 },
		{ 503, 473, 54 },
		{ 503, 474, 54 },
		{ 503, 475, 54 },
		{ 503, 476, 54 },
		{ 503, 477, 54 },
		{ 503, 478, 54 },
		{ 503, 479, 54 },
		{ 503, 480, 54 },
		{ 503, 481, 54 },
		{ 503, 482, 54 },
		{ 503, 483, 54 },
		{ 503, 484, 54 },
		{ 503, 485, 54 },
		{ 503, 486, 54 },
		{ 503, 487, 54 },
		{ 503, 488, 54 },
		{ 503, 489, 54 },
		{ 503, 490, 54 },
		{ 503, 491, 54 },
		{ 503, 492, 54 },
		{ 503, 493, 54 },
		{ 503, 494, 54 },
		{ 503, 495, 54 },
		{ 503, 496, 54 },
		{ 503, 497, 54 },
		{ 503, 498, 54 },
		{ 503, 499, 54 },
		{ 503, 500, 54 },
		{ 503, 501, 54 },
		{ 503, 502, 54 },
		{ 503, 503, 54 },
		{ 503, 504, 54 },
		{ 503, 505, 54 },
		{ 503, 506, 54 },
		{ 503, 507, 54 },
		{ 503, 508, 54 },
		{ 503, 509, 54 },
		{ 503, 510, 54 },
		{ 503, 511, 54 },
		{ 503, 512, 54 },
		{ 504, 474, 54 },
		{ 504, 475, 54 },
		{ 504, 476, 54 },
		{ 504, 477, 54 },
		{ 504, 478, 54 },
		{ 504, 479, 54 },
		{ 504, 480, 54 },
		{ 504, 481, 54 },
		{ 504, 482, 54 },
		{ 504, 483, 54 },
		{ 504, 484, 54 },
		{ 504, 485, 54 },
		{ 504, 486, 54 },
		{ 504, 487, 54 },
		{ 504, 488, 54 },
		{ 504, 489, 54 },
		{ 504, 490, 54 },
		{ 504, 491, 54 },
		{ 504, 492, 54 },
		{ 504, 493, 54 },
		{ 504, 494, 54 },
		{ 504, 495, 54 },
		{ 504, 496, 54 },
		{ 504, 497, 54 },
		{ 504, 498, 54 },
		{ 504, 499, 54 },
		{ 504, 500, 54 },
		{ 504, 501, 54 },
		{ 504, 502, 54 },
		{ 504, 503, 54 },
		{ 504, 504, 54 },
		{ 504, 505, 54 },
		{ 504, 506, 54 },
		{ 504, 507, 54 },
		{ 504, 508, 54 },
		{ 504, 509, 54 },
		{ 504, 510, 54 },
		{ 504, 511, 54 },
		{ 504, 512, 54 },
		{ 505, 475, 54 },
		{ 505, 476, 54 },
		{ 505, 477, 54 },
		{ 505, 478, 54 },
		{ 505, 479, 54 },
		{ 505, 480, 54 },
		{ 505, 481, 54 },
		{ 505, 482, 54 },
		{ 505, 483, 54 },
		{ 505, 484, 54 },
		{ 505, 485, 54 },
		{ 505, 486, 54 },
		{ 505, 487, 54 },
		{ 505, 488, 54 },
		{ 505, 489, 54 },
		{ 505, 490, 54 },
		{ 505, 491, 54 },
		{ 505, 492, 54 },
		{ 505, 493, 54 },
		{ 505, 494, 54 },
		{ 505, 495, 54 },
		{ 505, 496, 54 },
		{ 505, 497, 54 },
		{ 505, 498, 54 },
		{ 505, 499, 54 },
		{ 505, 500, 54 },
		{ 505, 501, 54 },
		{ 505, 502, 54 },
		{ 505, 503, 54 },
		{ 505, 504, 54 },
		{ 505, 505, 54 },
		{ 505, 506, 54 },
		{ 505, 507, 54 },
		{ 505, 508, 54 },
		{ 505, 509, 54 },
		{ 505, 510, 54 },
		{ 505, 511, 54 },
		{ 505, 512, 54 },
		{ 506, 475, 54 },
		{ 506, 476, 54 },
		{ 506, 477, 54 },
		{ 506, 478, 54 },
		{ 506, 479, 54 },
		{ 506, 480, 54 },
		{ 506, 481, 54 },
		{ 506, 482, 54 },
		{ 506, 483, 54 },
		{ 506, 484, 54 },
		{ 506, 485, 54 },
		{ 506, 486, 54 },
		{ 506, 487, 54 },
		{ 506, 488, 54 },
		{ 506, 489, 54 },
		{ 506, 490, 54 },
		{ 506, 491, 54 },
		{ 506, 492, 54 },
		{ 506, 493, 54 },
		{ 506, 494, 54 },
		{ 506, 495, 54 },
		{ 506, 496, 54 },
		{ 506, 497, 54 },
		{ 506, 498, 54 },
		{ 506, 499, 54 },
		{ 506, 500, 54 },
		{ 506, 501, 54 },
		{ 506, 502, 54 },
		{ 506, 503, 54 },
		{ 506, 504, 54 },
		{ 506, 505, 54 },
		{ 506, 506, 54 },
		{ 506, 507, 54 },
		{ 506, 508, 54 },
		{ 506, 509, 54 },
		{ 506, 510, 54 },
		{ 506, 511, 54 },
		{ 506, 512, 54 },
		{ 507, 476, 54 },
		{ 507, 477, 54 },
		{ 507, 478, 54 },
		{ 507, 479, 54 },
		{ 507, 480, 54 },
		{ 507, 481, 54 },
		{ 507, 482, 54 },
		{ 507, 483, 54 },
		{ 507, 484, 54 },
		{ 507, 485, 54 },
		{ 507, 486, 54 },
		{ 507, 487, 54 },
		{ 507, 488, 54 },
		{ 507, 489, 54 },
		{ 507, 490, 54 },
		{ 507, 491, 54 },
		{ 507, 492, 54 },
		{ 507, 493, 54 },
		{ 507, 494, 54 },
		{ 507, 495, 54 },
		{ 507, 496, 54 },
		{ 507, 497, 54 },
		{ 507, 498, 54 },
		{ 507, 499, 54 },
		{ 507, 500, 54 },
		{ 507, 501, 54 },
		{ 507, 502, 54 },
		{ 507, 503, 54 },
		{ 507, 504, 54 },
		{ 507, 505, 54 },
		{ 507, 506, 54 },
		{ 507, 507, 54 },
		{ 507, 508, 54 },
		{ 507, 509, 54 },
		{ 507, 510, 54 },
		{ 507, 511, 54 },
		{ 507, 512, 54 },
		{ 508, 477, 54 },
		{ 508, 478, 54 },
		{ 508, 479, 54 },
		{ 508, 480, 54 },
		{ 508, 481, 54 },
		{ 508, 482, 54 },
		{ 508, 483, 54 },
		{ 508, 484, 54 },
		{ 508, 485, 54 },
		{ 508, 486, 54 },
		{ 508, 487, 54 },
		{ 508, 488, 54 },
		{ 508, 489, 54 },
		{ 508, 490, 54 },
		{ 508, 491, 54 },
		{ 508, 492, 54 },
		{ 508, 493, 54 },
		{ 508, 494, 54 },
		{ 508, 495, 54 },
		{ 508, 496, 54 },
		{ 508, 497, 54 },
		{ 508, 498, 54 },
		{ 508, 499, 54 },
		{ 508, 500, 54 },
		{ 508, 501, 54 },
		{ 508, 502, 54 },
		{ 508, 503, 54 },
		{ 508, 504, 54 },
		{ 508, 505, 54 },
		{ 508, 506, 54 },
		{ 508, 507, 54 },
		{ 508, 508, 54 },
		{ 508, 509, 54 },
		{ 508, 510, 54 },
		{ 508, 511, 54 },
		{ 508, 512, 54 },
		{ 509, 478, 54 },
		{ 509, 479, 54 },
		{ 509, 480, 54 },
		{ 509, 481, 54 },
		{ 509, 482, 54 },
		{ 509, 483, 54 },
		{ 509, 484, 54 },
		{ 509, 485, 54 },
		{ 509, 486, 54 },
		{ 509, 487, 54 },
		{ 509, 488, 54 },
		{ 509, 489, 54 },
		{ 509, 490, 54 },
		{ 509, 491, 54 },
		{ 509, 492, 54 },
		{ 509, 493, 54 },
		{ 509, 494, 54 },
		{ 509, 495, 54 },
		{ 509, 496, 54 },
		{ 509, 497, 54 },
		{ 509, 498, 54 },
		{ 509, 499, 54 },
		{ 509, 500, 54 },
		{ 509, 501, 54 },
		{ 509, 502, 54 },
		{ 509, 503, 54 },
		{ 509, 504, 54 },
		{ 509, 505, 54 },
		{ 509, 506, 54 },
		{ 509, 507, 54 },
		{ 509, 508, 54 },
		{ 509, 509, 54 },
		{ 509, 510, 54 },
		{ 509, 511, 54 },
		{ 509, 512, 54 },
		{ 510, 479, 54 },
		{ 510, 480, 54 },
		{ 510, 481, 54 },
		{ 510, 482, 54 },
		{ 510, 483, 54 },
		{ 510, 484, 54 },
		{ 510, 485, 54 },
		{ 510, 486, 54 },
		{ 510, 487, 54 },
		{ 510, 488, 54 },
		{ 510, 489, 54 },
		{ 510, 490, 54 },
		{ 510, 491, 54 },
		{ 510, 492, 54 },
		{ 510, 493, 54 },
		{ 510, 494, 54 },
		{ 510, 495, 54 },
		{ 510, 496, 54 },
		{ 510, 497, 54 },
		{ 510, 498, 54 },
		{ 510, 499, 54 },
		{ 510, 500, 54 },
		{ 510, 501, 54 },
		{ 510, 502, 54 },
		{ 510, 503, 54 },
		{ 510, 504, 54 },
		{ 510, 505, 54 },
		{ 510, 506, 54 },
		{ 510, 507, 54 },
		{ 510, 508, 54 },
		{ 510, 509, 54 },
		{ 510, 510, 54 },
		{ 510, 511, 54 },
		{ 510, 512, 54 },
		{ 511, 480, 54 },
		{ 511, 481, 54 },
		{ 511, 482, 54 },
		{ 511, 483, 54 },
		{ 511, 484, 54 },
		{ 511, 485, 54 },
		{ 511, 486, 54 },
		{ 511, 487, 54 },
		{ 511, 488, 54 },
		{ 511, 489, 54 },
		{ 511, 490, 54 },
		{ 511, 491, 54 },
		{ 511, 492, 54 },
		{ 511, 493, 54 },
		{ 511, 494, 54 },
		{ 511, 495, 54 },
		{ 511, 496, 54 },
		{ 511, 497, 54 },
		{ 511, 498, 54 },
		{ 511, 499, 54 },
		{ 511, 500, 54 },
		{ 511, 501, 54 },
		{ 511, 502, 54 },
		{ 511, 503, 54 },
		{ 511, 504, 54 },
		{ 511, 505, 54 },
		{ 511, 506, 54 },
		{ 511, 507, 54 },
		{ 511, 508, 54 },
		{ 511, 509, 54 },
		{ 511, 510, 54 },
		{ 511, 511, 54 },
		{ 511, 512, 54 },
		{ 512, 481, 54 },
		{ 512, 482, 54 },
		{ 512, 483, 54 },
		{ 512, 484, 54 },
		{ 512, 485, 54 },
		{ 512, 486, 54 },
		{ 512, 487, 54 },
		{ 512, 488, 54 },
		{ 512, 489, 54 },
		{ 512, 490, 54 },
		{ 512, 491, 54 },
		{ 512, 492, 54 },
		{ 512, 493, 54 },
		{ 512, 494, 54 },
		{ 512, 495, 54 },
		{ 512, 496, 54 },
		{ 512, 497, 54 },
		{ 512, 498, 54 },
		{ 512, 499, 54 },
		{ 512, 500, 54 },
		{ 512, 501, 54 },
		{ 512, 502, 54 },
		{ 512, 503, 54 },
		{ 512, 504, 54 },
		{ 512, 505, 54 },
		{ 512, 506, 54 },
		{ 512, 507, 54 },
		{ 512, 508, 54 },
		{ 512, 509, 54 },
		{ 512, 510, 54 },
		{ 512, 511, 54 },
		{ 512, 512, 54 },
		{ 2, 1, 55 },
		{ 3, 1, 55 },
		{ 3, 2, 55 },
		{ 4, 1, 55 },
		{ 4, 2, 55 },
		{ 4, 3, 55 },
		{ 5, 1, 55 },
		{ 5, 2, 55 },
		{ 5, 3, 55 },
		{ 5, 4, 55 },
		{ 6, 1, 55 },
		{ 6, 2, 55 },
		{ 6, 3, 55 },
		{ 6, 4, 55 },
		{ 6, 5, 55 },
		{ 7, 1, 55 },
		{ 7, 2, 55 },
		{ 7, 3, 55 },
		{ 7, 4, 55 },
		{ 7, 5, 55 },
		{ 7, 6, 55 },
		{ 8, 1, 55 },
		{ 8, 2, 55 },
		{ 8, 3, 55 },
		{ 8, 4, 55 },
		{ 8, 5, 55 },
		{ 8, 6, 55 },
		{ 8, 7, 55 },
		{ 9, 1, 55 },
		{ 9, 2, 55 },
		{ 9, 3, 55 },
		{ 9, 4, 55 },
		{ 9, 5, 55 },
		{ 9, 6, 55 },
		{ 9, 7, 55 },
		{ 9, 8, 55 },
		{ 10, 1, 55 },
		{ 10, 2, 55 },
		{ 10, 3, 55 },
		{ 10, 4, 55 },
		{ 10, 5, 55 },
		{ 10, 6, 55 },
		{ 10, 7, 55 },
		{ 10, 8, 55 },
		{ 10, 9, 55 },
		{ 11, 1, 55 },
		{ 11, 2, 55 },
		{ 11, 3, 55 },
		{ 11, 4, 55 },
		{ 11, 5, 55 },
		{ 11, 6, 55 },
		{ 11, 7, 55 },
		{ 11, 8, 55 },
		{ 11, 9, 55 },
		{ 11, 10, 55 },
		{ 12, 1, 55 },
		{ 12, 2, 55 },
		{ 12, 3, 55 },
		{ 12, 4, 55 },
		{ 12, 5, 55 },
		{ 12, 6, 55 },
		{ 12, 7, 55 },
		{ 12, 8, 55 },
		{ 12, 9, 55 },
		{ 12, 10, 55 },
		{ 12, 11, 55 },
		{ 13, 1, 55 },
		{ 13, 2, 55 },
		{ 13, 3, 55 },
		{ 13, 4, 55 },
		{ 13, 5, 55 },
		{ 13, 6, 55 },
		{ 13, 7, 55 },
		{ 13, 8, 55 },
		{ 13, 9, 55 },
		{ 13, 10, 55 },
		{ 13, 11, 55 },
		{ 13, 12, 55 },
		{ 14, 1, 55 },
		{ 14, 2, 55 },
		{ 14, 3, 55 },
		{ 14, 4, 55 },
		{ 14, 5, 55 },
		{ 14, 6, 55 },
		{ 14, 7, 55 },
		{ 14, 8, 55 },
		{ 14, 9, 55 },
		{ 14, 10, 55 },
		{ 14, 11, 55 },
		{ 14, 12, 55 },
		{ 14, 13, 55 },
		{ 15, 1, 55 },
		{ 15, 2, 55 },
		{ 15, 3, 55 },
		{ 15, 4, 55 },
		{ 15, 5, 55 },
		{ 15, 6, 55 },
		{ 15, 7, 55 },
		{ 15, 8, 55 },
		{ 15, 9, 55 },
		{ 15, 10, 55 },
		{ 15, 11, 55 },
		{ 15, 12, 55 },
		{ 15, 13, 55 },
		{ 15, 14, 55 },
		{ 16, 1, 55 },
		{ 16, 2, 55 },
		{ 16, 3, 55 },
		{ 16, 4, 55 },
		{ 16, 5, 55 },
		{ 16, 6, 55 },
		{ 16, 7, 55 },
		{ 16, 8, 55 },
		{ 16, 9, 55 },
		{ 16, 10, 55 },
		{ 16, 11, 55 },
		{ 16, 12, 55 },
		{ 16, 13, 55 },
		{ 16, 14, 55 },
		{ 16, 15, 55 },
		{ 17, 1, 55 },
		{ 17, 2, 55 },
		{ 17, 3, 55 },
		{ 17, 4, 55 },
		{ 17, 5, 55 },
		{ 17, 6, 55 },
		{ 17, 7, 55 },
		{ 17, 8, 55 },
		{ 17, 9, 55 },
		{ 17, 10, 55 },
		{ 17, 11, 55 },
		{ 17, 12, 55 },
		{ 17, 13, 55 },
		{ 17, 14, 55 },
		{ 17, 15, 55 },
		{ 17, 16, 55 },
		{ 18, 1, 55 },
		{ 18, 2, 55 },
		{ 18, 3, 55 },
		{ 18, 4, 55 },
		{ 18, 5, 55 },
		{ 18, 6, 55 },
		{ 18, 7, 55 },
		{ 18, 8, 55 },
		{ 18, 9, 55 },
		{ 18, 10, 55 },
		{ 18, 11, 55 },
		{ 18, 12, 55 },
		{ 18, 13, 55 },
		{ 18, 14, 55 },
		{ 18, 15, 55 },
		{ 18, 16, 55 },
		{ 18, 17, 55 },
		{ 19, 1, 55 },
		{ 19, 2, 55 },
		{ 19, 3, 55 },
		{ 19, 4, 55 },
		{ 19, 5, 55 },
		{ 19, 6, 55 },
		{ 19, 7, 55 },
		{ 19, 8, 55 },
		{ 19, 9, 55 },
		{ 19, 10, 55 },
		{ 19, 11, 55 },
		{ 19, 12, 55 },
		{ 19, 13, 55 },
		{ 19, 14, 55 },
		{ 19, 15, 55 },
		{ 19, 16, 55 },
		{ 19, 17, 55 },
		{ 19, 18, 55 },
		{ 20, 1, 55 },
		{ 20, 2, 55 },
		{ 20, 3, 55 },
		{ 20, 4, 55 },
		{ 20, 5, 55 },
		{ 20, 6, 55 },
		{ 20, 7, 55 },
		{ 20, 8, 55 },
		{ 20, 9, 55 },
		{ 20, 10, 55 },
		{ 20, 11, 55 },
		{ 20, 12, 55 },
		{ 20, 13, 55 },
		{ 20, 14, 55 },
		{ 20, 15, 55 },
		{ 20, 16, 55 },
		{ 20, 17, 55 },
		{ 20, 18, 55 },
		{ 20, 19, 55 },
		{ 21, 1, 55 },
		{ 21, 2, 55 },
		{ 21, 3, 55 },
		{ 21, 4, 55 },
		{ 21, 5, 55 },
		{ 21, 6, 55 },
		{ 21, 7, 55 },
		{ 21, 8, 55 },
		{ 21, 9, 55 },
		{ 21, 10, 55 },
		{ 21, 11, 55 },
		{ 21, 12, 55 },
		{ 21, 13, 55 },
		{ 21, 14, 55 },
		{ 21, 15, 55 },
		{ 21, 16, 55 },
		{ 21, 17, 55 },
		{ 21, 18, 55 },
		{ 21, 19, 55 },
		{ 21, 20, 55 },
		{ 22, 1, 55 },
		{ 22, 2, 55 },
		{ 22, 3, 55 },
		{ 22, 4, 55 },
		{ 22, 5, 55 },
		{ 22, 6, 55 },
		{ 22, 7, 55 },
		{ 22, 8, 55 },
		{ 22, 9, 55 },
		{ 22, 10, 55 },
		{ 22, 11, 55 },
		{ 22, 12, 55 },
		{ 22, 13, 55 },
		{ 22, 14, 55 },
		{ 22, 15, 55 },
		{ 22, 16, 55 },
		{ 22, 17, 55 },
		{ 22, 18, 55 },
		{ 22, 19, 55 },
		{ 22, 20, 55 },
		{ 22, 21, 55 },
		{ 23, 1, 55 },
		{ 23, 2, 55 },
		{ 23, 3, 55 },
		{ 23, 4, 55 },
		{ 23, 5, 55 },
		{ 23, 6, 55 },
		{ 23, 7, 55 },
		{ 23, 8, 55 },
		{ 23, 9, 55 },
		{ 23, 10, 55 },
		{ 23, 11, 55 },
		{ 23, 12, 55 },
		{ 23, 13, 55 },
		{ 23, 14, 55 },
		{ 23, 15, 55 },
		{ 23, 16, 55 },
		{ 23, 17, 55 },
		{ 23, 18, 55 },
		{ 23, 19, 55 },
		{ 23, 20, 55 },
		{ 23, 21, 55 },
		{ 23, 22, 55 },
		{ 24, 1, 55 },
		{ 24, 2, 55 },
		{ 24, 3, 55 },
		{ 24, 4, 55 },
		{ 24, 5, 55 },
		{ 24, 6, 55 },
		{ 24, 7, 55 },
		{ 24, 8, 55 },
		{ 24, 9, 55 },
		{ 24, 10, 55 },
		{ 24, 11, 55 },
		{ 24, 12, 55 },
		{ 24, 13, 55 },
		{ 24, 14, 55 },
		{ 24, 15, 55 },
		{ 24, 16, 55 },
		{ 24, 17, 55 },
		{ 24, 18, 55 },
		{ 24, 19, 55 },
		{ 24, 20, 55 },
		{ 24, 21, 55 },
		{ 24, 22, 55 },
		{ 24, 23, 55 },
		{ 25, 1, 55 },
		{ 25, 2, 55 },
		{ 25, 3, 55 },
		{ 25, 4, 55 },
		{ 25, 5, 55 },
		{ 25, 6, 55 },
		{ 25, 7, 55 },
		{ 25, 8, 55 },
		{ 25, 9, 55 },
		{ 25, 10, 55 },
		{ 25, 11, 55 },
		{ 25, 12, 55 },
		{ 25, 13, 55 },
		{ 25, 14, 55 },
		{ 25, 15, 55 },
		{ 25, 16, 55 },
		{ 25, 17, 55 },
		{ 25, 18, 55 },
		{ 25, 19, 55 },
		{ 25, 20, 55 },
		{ 25, 21, 55 },
		{ 25, 22, 55 },
		{ 25, 23, 55 },
		{ 25, 24, 55 },
		{ 26, 1, 55 },
		{ 26, 2, 55 },
		{ 26, 3, 55 },
		{ 26, 4, 55 },
		{ 26, 5, 55 },
		{ 26, 6, 55 },
		{ 26, 7, 55 },
		{ 26, 8, 55 },
		{ 26, 9, 55 },
		{ 26, 10, 55 },
		{ 26, 11, 55 },
		{ 26, 12, 55 },
		{ 26, 13, 55 },
		{ 26, 14, 55 },
		{ 26, 15, 55 },
		{ 26, 16, 55 },
		{ 26, 17, 55 },
		{ 26, 18, 55 },
		{ 26, 19, 55 },
		{ 26, 20, 55 },
		{ 26, 21, 55 },
		{ 26, 22, 55 },
		{ 26, 23, 55 },
		{ 26, 24, 55 },
		{ 26, 25, 55 },
		{ 27, 1, 55 },
		{ 27, 2, 55 },
		{ 27, 3, 55 },
		{ 27, 4, 55 },
		{ 27, 5, 55 },
		{ 27, 6, 55 },
		{ 27, 7, 55 },
		{ 27, 8, 55 },
		{ 27, 9, 55 },
		{ 27, 10, 55 },
		{ 27, 11, 55 },
		{ 27, 12, 55 },
		{ 27, 13, 55 },
		{ 27, 14, 55 },
		{ 27, 15, 55 },
		{ 27, 16, 55 },
		{ 27, 17, 55 },
		{ 27, 18, 55 },
		{ 27, 19, 55 },
		{ 27, 20, 55 },
		{ 27, 21, 55 },
		{ 27, 22, 55 },
		{ 27, 23, 55 },
		{ 27, 24, 55 },
		{ 27, 25, 55 },
		{ 27, 26, 55 },
		{ 28, 1, 55 },
		{ 28, 2, 55 },
		{ 28, 3, 55 },
		{ 28, 4, 55 },
		{ 28, 5, 55 },
		{ 28, 6, 55 },
		{ 28, 7, 55 },
		{ 28, 8, 55 },
		{ 28, 9, 55 },
		{ 28, 10, 55 },
		{ 28, 11, 55 },
		{ 28, 12, 55 },
		{ 28, 13, 55 },
		{ 28, 14, 55 },
		{ 28, 15, 55 },
		{ 28, 16, 55 },
		{ 28, 17, 55 },
		{ 28, 18, 55 },
		{ 28, 19, 55 },
		{ 28, 20, 55 },
		{ 28, 21, 55 },
		{ 28, 22, 55 },
		{ 28, 23, 55 },
		{ 28, 24, 55 },
		{ 28, 25, 55 },
		{ 28, 26, 55 },
		{ 28, 27, 55 },
		{ 29, 1, 55 },
		{ 29, 2, 55 },
		{ 29, 3, 55 },
		{ 29, 4, 55 },
		{ 29, 5, 55 },
		{ 29, 6, 55 },
		{ 29, 7, 55 },
		{ 29, 8, 55 },
		{ 29, 9, 55 },
		{ 29, 10, 55 },
		{ 29, 11, 55 },
		{ 29, 12, 55 },
		{ 29, 13, 55 },
		{ 29, 14, 55 },
		{ 29, 15, 55 },
		{ 29, 16, 55 },
		{ 29, 17, 55 },
		{ 29, 18, 55 },
		{ 29, 19, 55 },
		{ 29, 20, 55 },
		{ 29, 21, 55 },
		{ 29, 22, 55 },
		{ 29, 23, 55 },
		{ 29, 24, 55 },
		{ 29, 25, 55 },
		{ 29, 26, 55 },
		{ 29, 27, 55 },
		{ 29, 28, 55 },
		{ 30, 1, 55 },
		{ 30, 2, 55 },
		{ 30, 3, 55 },
		{ 30, 4, 55 },
		{ 30, 5, 55 },
		{ 30, 6, 55 },
		{ 30, 7, 55 },
		{ 30, 8, 55 },
		{ 30, 9, 55 },
		{ 30, 10, 55 },
		{ 30, 11, 55 },
		{ 30, 12, 55 },
		{ 30, 13, 55 },
		{ 30, 14, 55 },
		{ 30, 15, 55 },
		{ 30, 16, 55 },
		{ 30, 17, 55 },
		{ 30, 18, 55 },
		{ 30, 19, 55 },
		{ 30, 20, 55 },
		{ 30, 21, 55 },
		{ 30, 22, 55 },
		{ 30, 23, 55 },
		{ 30, 24, 55 },
		{ 30, 25, 55 },
		{ 30, 26, 55 },
		{ 30, 27, 55 },
		{ 30, 28, 55 },
		{ 30, 29, 55 },
		{ 31, 1, 55 },
		{ 31, 2, 55 },
		{ 31, 3, 55 },
		{ 31, 4, 55 },
		{ 31, 5, 55 },
		{ 31, 6, 55 },
		{ 31, 7, 55 },
		{ 31, 8, 55 },
		{ 31, 9, 55 },
		{ 31, 10, 55 },
		{ 31, 11, 55 },
		{ 31, 12, 55 },
		{ 31, 13, 55 },
		{ 31, 14, 55 },
		{ 31, 15, 55 },
		{ 31, 16, 55 },
		{ 31, 17, 55 },
		{ 31, 18, 55 },
		{ 31, 19, 55 },
		{ 31, 20, 55 },
		{ 31, 21, 55 },
		{ 31, 22, 55 },
		{ 31, 23, 55 },
		{ 31, 24, 55 },
		{ 31, 25, 55 },
		{ 31, 26, 55 },
		{ 31, 27, 55 },
		{ 31, 28, 55 },
		{ 31, 29, 55 },
		{ 31, 30, 55 },
		{ 32, 1, 55 },
		{ 32, 2, 55 },
		{ 32, 3, 55 },
		{ 32, 4, 55 },
		{ 32, 5, 55 },
		{ 32, 6, 55 },
		{ 32, 7, 55 },
		{ 32, 8, 55 },
		{ 32, 9, 55 },
		{ 32, 10, 55 },
		{ 32, 11, 55 },
		{ 32, 12, 55 },
		{ 32, 13, 55 },
		{ 32, 14, 55 },
		{ 32, 15, 55 },
		{ 32, 16, 55 },
		{ 32, 17, 55 },
		{ 32, 18, 55 },
		{ 32, 19, 55 },
		{ 32, 20, 55 },
		{ 32, 21, 55 },
		{ 32, 22, 55 },
		{ 32, 23, 55 },
		{ 32, 24, 55 },
		{ 32, 25, 55 },
		{ 32, 26, 55 },
		{ 32, 27, 55 },
		{ 32, 28, 55 },
		{ 32, 29, 55 },
		{ 32, 30, 55 },
		{ 32, 31, 55 },
		{ 33, 1, 55 },
		{ 33, 2, 55 },
		{ 33, 3, 55 },
		{ 33, 4, 55 },
		{ 33, 5, 55 },
		{ 33, 6, 55 },
		{ 33, 7, 55 },
		{ 33, 8, 55 },
		{ 33, 9, 55 },
		{ 33, 10, 55 },
		{ 33, 11, 55 },
		{ 33, 12, 55 },
		{ 33, 13, 55 },
		{ 33, 14, 55 },
		{ 33, 15, 55 },
		{ 33, 16, 55 },
		{ 33, 17, 55 },
		{ 33, 18, 55 },
		{ 33, 19, 55 },
		{ 33, 20, 55 },
		{ 33, 21, 55 },
		{ 33, 22, 55 },
		{ 33, 23, 55 },
		{ 33, 24, 55 },
		{ 33, 25, 55 },
		{ 33, 26, 55 },
		{ 33, 27, 55 },
		{ 33, 28, 55 },
		{ 33, 29, 55 },
		{ 33, 30, 55 },
		{ 33, 31, 55 },
		{ 33, 32, 55 },
		{ 34, 1, 55 },
		{ 34, 2, 55 },
		{ 34, 3, 55 },
		{ 34, 4, 55 },
		{ 34, 5, 55 },
		{ 34, 6, 55 },
		{ 34, 7, 55 },
		{ 34, 8, 55 },
		{ 34, 9, 55 },
		{ 34, 10, 55 },
		{ 34, 11, 55 },
		{ 34, 12, 55 },
		{ 34, 13, 55 },
		{ 34, 14, 55 },
		{ 34, 15, 55 },
		{ 34, 16, 55 },
		{ 34, 17, 55 },
		{ 34, 18, 55 },
		{ 34, 19, 55 },
		{ 34, 20, 55 },
		{ 34, 21, 55 },
		{ 34, 22, 55 },
		{ 34, 23, 55 },
		{ 34, 24, 55 },
		{ 34, 25, 55 },
		{ 34, 26, 55 },
		{ 34, 27, 55 },
		{ 34, 28, 55 },
		{ 34, 29, 55 },
		{ 34, 30, 55 },
		{ 34, 31, 55 },
		{ 34, 32, 55 },
		{ 34, 33, 55 },
		{ 35, 1, 55 },
		{ 35, 2, 55 },
		{ 35, 3, 55 },
		{ 35, 4, 55 },
		{ 35, 5, 55 },
		{ 35, 6, 55 },
		{ 35, 7, 55 },
		{ 35, 8, 55 },
		{ 35, 9, 55 },
		{ 35, 10, 55 },
		{ 35, 11, 55 },
		{ 35, 12, 55 },
		{ 35, 13, 55 },
		{ 35, 14, 55 },
		{ 35, 15, 55 },
		{ 35, 16, 55 },
		{ 35, 17, 55 },
		{ 35, 18, 55 },
		{ 35, 19, 55 },
		{ 35, 20, 55 },
		{ 35, 21, 55 },
		{ 35, 22, 55 },
		{ 35, 23, 55 },
		{ 35, 24, 55 },
		{ 35, 25, 55 },
		{ 35, 26, 55 },
		{ 35, 27, 55 },
		{ 35, 28, 55 },
		{ 35, 29, 55 },
		{ 35, 30, 55 },
		{ 35, 31, 55 },
		{ 35, 32, 55 },
		{ 35, 33, 55 },
		{ 35, 34, 55 },
		{ 36, 1, 55 },
		{ 36, 2, 55 },
		{ 36, 3, 55 },
		{ 36, 4, 55 },
		{ 36, 5, 55 },
		{ 36, 6, 55 },
		{ 36, 7, 55 },
		{ 36, 8, 55 },
		{ 36, 9, 55 },
		{ 36, 10, 55 },
		{ 36, 11, 55 },
		{ 36, 12, 55 },
		{ 36, 13, 55 },
		{ 36, 14, 55 },
		{ 36, 15, 55 },
		{ 36, 16, 55 },
		{ 36, 17, 55 },
		{ 36, 18, 55 },
		{ 36, 19, 55 },
		{ 36, 20, 55 },
		{ 36, 21, 55 },
		{ 36, 22, 55 },
		{ 36, 23, 55 },
		{ 36, 24, 55 },
		{ 36, 25, 55 },
		{ 36, 26, 55 },
		{ 36, 27, 55 },
		{ 36, 28, 55 },
		{ 36, 29, 55 },
		{ 36, 30, 55 },
		{ 36, 31, 55 },
		{ 36, 32, 55 },
		{ 36, 33, 55 },
		{ 36, 34, 55 },
		{ 36, 35, 55 },
		{ 37, 1, 55 },
		{ 37, 2, 55 },
		{ 37, 3, 55 },
		{ 37, 4, 55 },
		{ 37, 5, 55 },
		{ 37, 6, 55 },
		{ 37, 7, 55 },
		{ 37, 8, 55 },
		{ 37, 9, 55 },
		{ 37, 10, 55 },
		{ 37, 11, 55 },
		{ 37, 12, 55 },
		{ 37, 13, 55 },
		{ 37, 14, 55 },
		{ 37, 15, 55 },
		{ 37, 16, 55 },
		{ 37, 17, 55 },
		{ 37, 18, 55 },
		{ 37, 19, 55 },
		{ 37, 20, 55 },
		{ 37, 21, 55 },
		{ 37, 22, 55 },
		{ 37, 23, 55 },
		{ 37, 24, 55 },
		{ 37, 25, 55 },
		{ 37, 26, 55 },
		{ 37, 27, 55 },
		{ 37, 28, 55 },
		{ 37, 29, 55 },
		{ 37, 30, 55 },
		{ 37, 31, 55 },
		{ 37, 32, 55 },
		{ 37, 33, 55 },
		{ 37, 34, 55 },
		{ 37, 35, 55 },
		{ 37, 36, 55 },
		{ 38, 1, 55 },
		{ 38, 2, 55 },
		{ 38, 3, 55 },
		{ 38, 4, 55 },
		{ 38, 5, 55 },
		{ 38, 6, 55 },
		{ 38, 7, 55 },
		{ 38, 8, 55 },
		{ 38, 9, 55 },
		{ 38, 10, 55 },
		{ 38, 11, 55 },
		{ 38, 12, 55 },
		{ 38, 13, 55 },
		{ 38, 14, 55 },
		{ 38, 15, 55 },
		{ 38, 16, 55 },
		{ 38, 17, 55 },
		{ 38, 18, 55 },
		{ 38, 19, 55 },
		{ 38, 20, 55 },
		{ 38, 21, 55 },
		{ 38, 22, 55 },
		{ 38, 23, 55 },
		{ 38, 24, 55 },
		{ 38, 25, 55 },
		{ 38, 26, 55 },
		{ 38, 27, 55 },
		{ 38, 28, 55 },
		{ 38, 29, 55 },
		{ 38, 30, 55 },
		{ 38, 31, 55 },
		{ 38, 32, 55 },
		{ 38, 33, 55 },
		{ 38, 34, 55 },
		{ 38, 35, 55 },
		{ 38, 36, 55 },
		{ 38, 37, 55 },
		{ 39, 1, 55 },
		{ 39, 2, 55 },
		{ 39, 3, 55 },
		{ 39, 4, 55 },
		{ 39, 5, 55 },
		{ 39, 6, 55 },
		{ 39, 7, 55 },
		{ 39, 8, 55 },
		{ 39, 9, 55 },
		{ 39, 10, 55 },
		{ 39, 11, 55 },
		{ 39, 12, 55 },
		{ 39, 13, 55 },
		{ 39, 14, 55 },
		{ 39, 15, 55 },
		{ 39, 16, 55 },
		{ 39, 17, 55 },
		{ 39, 18, 55 },
		{ 39, 19, 55 },
		{ 39, 20, 55 },
		{ 39, 21, 55 },
		{ 39, 22, 55 },
		{ 39, 23, 55 },
		{ 39, 24, 55 },
		{ 39, 25, 55 },
		{ 39, 26, 55 },
		{ 39, 27, 55 },
		{ 39, 28, 55 },
		{ 39, 29, 55 },
		{ 39, 30, 55 },
		{ 39, 31, 55 },
		{ 39, 32, 55 },
		{ 39, 33, 55 },
		{ 39, 34, 55 },
		{ 39, 35, 55 },
		{ 39, 36, 55 },
		{ 39, 37, 55 },
		{ 39, 38, 55 },
		{ 40, 1, 55 },
		{ 40, 2, 55 },
		{ 40, 3, 55 },
		{ 40, 4, 55 },
		{ 40, 5, 55 },
		{ 40, 6, 55 },
		{ 40, 7, 55 },
		{ 40, 8, 55 },
		{ 40, 9, 55 },
		{ 40, 10, 55 },
		{ 40, 11, 55 },
		{ 40, 12, 55 },
		{ 40, 13, 55 },
		{ 40, 14, 55 },
		{ 40, 15, 55 },
		{ 40, 16, 55 },
		{ 40, 17, 55 },
		{ 40, 18, 55 },
		{ 40, 19, 55 },
		{ 40, 20, 55 },
		{ 40, 21, 55 },
		{ 40, 22, 55 },
		{ 40, 23, 55 },
		{ 40, 24, 55 },
		{ 40, 25, 55 },
		{ 40, 26, 55 },
		{ 40, 27, 55 },
		{ 40, 28, 55 },
		{ 40, 29, 55 },
		{ 40, 30, 55 },
		{ 40, 31, 55 },
		{ 40, 32, 55 },
		{ 40, 33, 55 },
		{ 40, 34, 55 },
		{ 40, 35, 55 },
		{ 40, 36, 55 },
		{ 40, 37, 55 },
		{ 40, 38, 55 },
		{ 40, 39, 55 },
		{ 41, 1, 55 },
		{ 41, 2, 55 },
		{ 41, 3, 55 },
		{ 41, 4, 55 },
		{ 41, 5, 55 },
		{ 41, 6, 55 },
		{ 41, 7, 55 },
		{ 41, 8, 55 },
		{ 41, 9, 55 },
		{ 41, 10, 55 },
		{ 41, 11, 55 },
		{ 41, 12, 55 },
		{ 41, 13, 55 },
		{ 41, 14, 55 },
		{ 41, 15, 55 },
		{ 41, 16, 55 },
		{ 41, 17, 55 },
		{ 41, 18, 55 },
		{ 41, 19, 55 },
		{ 41, 20, 55 },
		{ 41, 21, 55 },
		{ 41, 22, 55 },
		{ 41, 23, 55 },
		{ 41, 24, 55 },
		{ 41, 25, 55 },
		{ 41, 26, 55 },
		{ 41, 27, 55 },
		{ 41, 28, 55 },
		{ 41, 29, 55 },
		{ 41, 30, 55 },
		{ 41, 31, 55 },
		{ 41, 32, 55 },
		{ 41, 33, 55 },
		{ 41, 34, 55 },
		{ 41, 35, 55 },
		{ 41, 36, 55 },
		{ 41, 37, 55 },
		{ 41, 38, 55 },
		{ 41, 39, 55 },
		{ 41, 40, 55 },
		{ 42, 1, 55 },
		{ 42, 2, 55 },
		{ 42, 3, 55 },
		{ 42, 4, 55 },
		{ 42, 5, 55 },
		{ 42, 6, 55 },
		{ 42, 7, 55 },
		{ 42, 8, 55 },
		{ 42, 9, 55 },
		{ 42, 10, 55 },
		{ 42, 11, 55 },
		{ 42, 12, 55 },
		{ 42, 13, 55 },
		{ 42, 14, 55 },
		{ 42, 15, 55 },
		{ 42, 16, 55 },
		{ 42, 17, 55 },
		{ 42, 18, 55 },
		{ 42, 19, 55 },
		{ 42, 20, 55 },
		{ 42, 21, 55 },
		{ 42, 22, 55 },
		{ 42, 23, 55 },
		{ 42, 24, 55 },
		{ 42, 25, 55 },
		{ 42, 26, 55 },
		{ 42, 27, 55 },
		{ 42, 28, 55 },
		{ 42, 29, 55 },
		{ 42, 30, 55 },
		{ 42, 31, 55 },
		{ 42, 32, 55 },
		{ 42, 33, 55 },
		{ 42, 34, 55 },
		{ 42, 35, 55 },
		{ 42, 36, 55 },
		{ 42, 37, 55 },
		{ 42, 38, 55 },
		{ 42, 39, 55 },
		{ 42, 40, 55 },
		{ 42, 41, 55 },
		{ 43, 1, 55 },
		{ 43, 2, 55 },
		{ 43, 3, 55 },
		{ 43, 4, 55 },
		{ 43, 5, 55 },
		{ 43, 6, 55 },
		{ 43, 7, 55 },
		{ 43, 8, 55 },
		{ 43, 9, 55 },
		{ 43, 10, 55 },
		{ 43, 11, 55 },
		{ 43, 12, 55 },
		{ 43, 13, 55 },
		{ 43, 14, 55 },
		{ 43, 15, 55 },
		{ 43, 16, 55 },
		{ 43, 17, 55 },
		{ 43, 18, 55 },
		{ 43, 19, 55 },
		{ 43, 20, 55 },
		{ 43, 21, 55 },
		{ 43, 22, 55 },
		{ 43, 23, 55 },
		{ 43, 24, 55 },
		{ 43, 25, 55 },
		{ 43, 26, 55 },
		{ 43, 27, 55 },
		{ 43, 28, 55 },
		{ 43, 29, 55 },
		{ 43, 30, 55 },
		{ 43, 31, 55 },
		{ 43, 32, 55 },
		{ 43, 33, 55 },
		{ 43, 34, 55 },
		{ 43, 35, 55 },
		{ 43, 36, 55 },
		{ 43, 37, 55 },
		{ 43, 38, 55 },
		{ 43, 39, 55 },
		{ 43, 40, 55 },
		{ 43, 41, 55 },
		{ 43, 42, 55 },
		{ 44, 1, 55 },
		{ 44, 2, 55 },
		{ 44, 3, 55 },
		{ 44, 4, 55 },
		{ 44, 5, 55 },
		{ 44, 6, 55 },
		{ 44, 7, 55 },
		{ 44, 8, 55 },
		{ 44, 9, 55 },
		{ 44, 10, 55 },
		{ 44, 11, 55 },
		{ 44, 12, 55 },
		{ 44, 13, 55 },
		{ 44, 14, 55 },
		{ 44, 15, 55 },
		{ 44, 16, 55 },
		{ 44, 17, 55 },
		{ 44, 18, 55 },
		{ 44, 19, 55 },
		{ 44, 20, 55 },
		{ 44, 21, 55 },
		{ 44, 22, 55 },
		{ 44, 23, 55 },
		{ 44, 24, 55 },
		{ 44, 25, 55 },
		{ 44, 26, 55 },
		{ 44, 27, 55 },
		{ 44, 28, 55 },
		{ 44, 29, 55 },
		{ 44, 30, 55 },
		{ 44, 31, 55 },
		{ 44, 32, 55 },
		{ 44, 33, 55 },
		{ 44, 34, 55 },
		{ 44, 35, 55 },
		{ 44, 36, 55 },
		{ 44, 37, 55 },
		{ 44, 38, 55 },
		{ 44, 39, 55 },
		{ 44, 40, 55 },
		{ 44, 41, 55 },
		{ 44, 42, 55 },
		{ 44, 43, 55 },
		{ 45, 1, 55 },
		{ 45, 2, 55 },
		{ 45, 3, 55 },
		{ 45, 4, 55 },
		{ 45, 5, 55 },
		{ 45, 6, 55 },
		{ 45, 7, 55 },
		{ 45, 8, 55 },
		{ 45, 9, 55 },
		{ 45, 10, 55 },
		{ 45, 11, 55 },
		{ 45, 12, 55 },
		{ 45, 13, 55 },
		{ 45, 14, 55 },
		{ 45, 15, 55 },
		{ 45, 16, 55 },
		{ 45, 17, 55 },
		{ 45, 18, 55 },
		{ 45, 19, 55 },
		{ 45, 20, 55 },
		{ 45, 21, 55 },
		{ 45, 22, 55 },
		{ 45, 23, 55 },
		{ 45, 24, 55 },
		{ 45, 25, 55 },
		{ 45, 26, 55 },
		{ 45, 27, 55 },
		{ 45, 28, 55 },
		{ 45, 29, 55 },
		{ 45, 30, 55 },
		{ 45, 31, 55 },
		{ 45, 32, 55 },
		{ 45, 33, 55 },
		{ 45, 34, 55 },
		{ 45, 35, 55 },
		{ 45, 36, 55 },
		{ 45, 37, 55 },
		{ 45, 38, 55 },
		{ 45, 39, 55 },
		{ 45, 40, 55 },
		{ 45, 41, 55 },
		{ 45, 42, 55 },
		{ 45, 43, 55 },
		{ 45, 44, 55 },
		{ 46, 1, 55 },
		{ 46, 2, 55 },
		{ 46, 3, 55 },
		{ 46, 4, 55 },
		{ 46, 5, 55 },
		{ 46, 6, 55 },
		{ 46, 7, 55 },
		{ 46, 8, 55 },
		{ 46, 9, 55 },
		{ 46, 10, 55 },
		{ 46, 11, 55 },
		{ 46, 12, 55 },
		{ 46, 13, 55 },
		{ 46, 14, 55 },
		{ 46, 15, 55 },
		{ 46, 16, 55 },
		{ 46, 17, 55 },
		{ 46, 18, 55 },
		{ 46, 19, 55 },
		{ 46, 20, 55 },
		{ 46, 21, 55 },
		{ 46, 22, 55 },
		{ 46, 23, 55 },
		{ 46, 24, 55 },
		{ 46, 25, 55 },
		{ 46, 26, 55 },
		{ 46, 27, 55 },
		{ 46, 28, 55 },
		{ 46, 29, 55 },
		{ 46, 30, 55 },
		{ 46, 31, 55 },
		{ 46, 32, 55 },
		{ 46, 33, 55 },
		{ 46, 34, 55 },
		{ 46, 35, 55 },
		{ 46, 36, 55 },
		{ 46, 37, 55 },
		{ 46, 38, 55 },
		{ 46, 39, 55 },
		{ 46, 40, 55 },
		{ 46, 41, 55 },
		{ 46, 42, 55 },
		{ 46, 43, 55 },
		{ 46, 44, 55 },
		{ 46, 45, 55 },
		{ 47, 1, 55 },
		{ 47, 2, 55 },
		{ 47, 3, 55 },
		{ 47, 4, 55 },
		{ 47, 5, 55 },
		{ 47, 6, 55 },
		{ 47, 7, 55 },
		{ 47, 8, 55 },
		{ 47, 9, 55 },
		{ 47, 10, 55 },
		{ 47, 11, 55 },
		{ 47, 12, 55 },
		{ 47, 13, 55 },
		{ 47, 14, 55 },
		{ 47, 15, 55 },
		{ 47, 16, 55 },
		{ 47, 17, 55 },
		{ 47, 18, 55 },
		{ 47, 19, 55 },
		{ 47, 20, 55 },
		{ 47, 21, 55 },
		{ 47, 22, 55 },
		{ 47, 23, 55 },
		{ 47, 24, 55 },
		{ 47, 25, 55 },
		{ 47, 26, 55 },
		{ 47, 27, 55 },
		{ 47, 28, 55 },
		{ 47, 29, 55 },
		{ 47, 30, 55 },
		{ 47, 31, 55 },
		{ 47, 32, 55 },
		{ 47, 33, 55 },
		{ 47, 34, 55 },
		{ 47, 35, 55 },
		{ 47, 36, 55 },
		{ 47, 37, 55 },
		{ 47, 38, 55 },
		{ 47, 39, 55 },
		{ 47, 40, 55 },
		{ 47, 41, 55 },
		{ 47, 42, 55 },
		{ 47, 43, 55 },
		{ 47, 44, 55 },
		{ 47, 45, 55 },
		{ 47, 46, 55 },
		{ 48, 1, 55 },
		{ 48, 2, 55 },
		{ 48, 3, 55 },
		{ 48, 4, 55 },
		{ 48, 5, 55 },
		{ 48, 6, 55 },
		{ 48, 7, 55 },
		{ 48, 8, 55 },
		{ 48, 9, 55 },
		{ 48, 10, 55 },
		{ 48, 11, 55 },
		{ 48, 12, 55 },
		{ 48, 13, 55 },
		{ 48, 14, 55 },
		{ 48, 15, 55 },
		{ 48, 16, 55 },
		{ 48, 17, 55 },
		{ 48, 18, 55 },
		{ 48, 19, 55 },
		{ 48, 20, 55 },
		{ 48, 21, 55 },
		{ 48, 22, 55 },
		{ 48, 23, 55 },
		{ 48, 24, 55 },
		{ 48, 25, 55 },
		{ 48, 26, 55 },
		{ 48, 27, 55 },
		{ 48, 28, 55 },
		{ 48, 29, 55 },
		{ 48, 30, 55 },
		{ 48, 31, 55 },
		{ 48, 32, 55 },
		{ 48, 33, 55 },
		{ 48, 34, 55 },
		{ 48, 35, 55 },
		{ 48, 36, 55 },
		{ 48, 37, 55 },
		{ 48, 38, 55 },
		{ 48, 39, 55 },
		{ 48, 40, 55 },
		{ 48, 41, 55 },
		{ 48, 42, 55 },
		{ 48, 43, 55 },
		{ 48, 44, 55 },
		{ 48, 45, 55 },
		{ 48, 46, 55 },
		{ 48, 47, 55 },
		{ 49, 1, 55 },
		{ 49, 2, 55 },
		{ 49, 3, 55 },
		{ 49, 4, 55 },
		{ 49, 5, 55 },
		{ 49, 6, 55 },
		{ 49, 7, 55 },
		{ 49, 8, 55 },
		{ 49, 9, 55 },
		{ 49, 10, 55 },
		{ 49, 11, 55 },
		{ 49, 12, 55 },
		{ 49, 13, 55 },
		{ 49, 14, 55 },
		{ 49, 15, 55 },
		{ 49, 16, 55 },
		{ 49, 17, 55 },
		{ 49, 18, 55 },
		{ 49, 19, 55 },
		{ 49, 20, 55 },
		{ 49, 21, 55 },
		{ 49, 22, 55 },
		{ 49, 23, 55 },
		{ 49, 24, 55 },
		{ 49, 25, 55 },
		{ 49, 26, 55 },
		{ 49, 27, 55 },
		{ 49, 28, 55 },
		{ 49, 29, 55 },
		{ 49, 30, 55 },
		{ 49, 31, 55 },
		{ 49, 32, 55 },
		{ 49, 33, 55 },
		{ 49, 34, 55 },
		{ 49, 35, 55 },
		{ 49, 36, 55 },
		{ 49, 37, 55 },
		{ 49, 38, 55 },
		{ 49, 39, 55 },
		{ 49, 40, 55 },
		{ 49, 41, 55 },
		{ 49, 42, 55 },
		{ 49, 43, 55 },
		{ 49, 44, 55 },
		{ 49, 45, 55 },
		{ 49, 46, 55 },
		{ 49, 47, 55 },
		{ 49, 48, 55 },
		{ 50, 1, 55 },
		{ 50, 2, 55 },
		{ 50, 3, 55 },
		{ 50, 4, 55 },
		{ 50, 5, 55 },
		{ 50, 6, 55 },
		{ 50, 7, 55 },
		{ 50, 8, 55 },
		{ 50, 9, 55 },
		{ 50, 10, 55 },
		{ 50, 11, 55 },
		{ 50, 12, 55 },
		{ 50, 13, 55 },
		{ 50, 14, 55 },
		{ 50, 15, 55 },
		{ 50, 16, 55 },
		{ 50, 17, 55 },
		{ 50, 18, 55 },
		{ 50, 19, 55 },
		{ 50, 20, 55 },
		{ 50, 21, 55 },
		{ 50, 22, 55 },
		{ 50, 23, 55 },
		{ 50, 24, 55 },
		{ 50, 25, 55 },
		{ 50, 26, 55 },
		{ 50, 27, 55 },
		{ 50, 28, 55 },
		{ 50, 29, 55 },
		{ 50, 30, 55 },
		{ 50, 31, 55 },
		{ 50, 32, 55 },
		{ 50, 33, 55 },
		{ 50, 34, 55 },
		{ 50, 35, 55 },
		{ 50, 36, 55 },
		{ 50, 37, 55 },
		{ 50, 38, 55 },
		{ 50, 39, 55 },
		{ 50, 40, 55 },
		{ 50, 41, 55 },
		{ 50, 42, 55 },
		{ 50, 43, 55 },
		{ 50, 44, 55 },
		{ 50, 45, 55 },
		{ 50, 46, 55 },
		{ 50, 47, 55 },
		{ 50, 48, 55 },
		{ 50, 49, 55 },
		{ 51, 1, 55 },
		{ 51, 2, 55 },
		{ 51, 3, 55 },
		{ 51, 4, 55 },
		{ 51, 5, 55 },
		{ 51, 6, 55 },
		{ 51, 7, 55 },
		{ 51, 8, 55 },
		{ 51, 9, 55 },
		{ 51, 10, 55 },
		{ 51, 11, 55 },
		{ 51, 12, 55 },
		{ 51, 13, 55 },
		{ 51, 14, 55 },
		{ 51, 15, 55 },
		{ 51, 16, 55 },
		{ 51, 17, 55 },
		{ 51, 18, 55 },
		{ 51, 19, 55 },
		{ 51, 20, 55 },
		{ 51, 21, 55 },
		{ 51, 22, 55 },
		{ 51, 23, 55 },
		{ 51, 24, 55 },
		{ 51, 25, 55 },
		{ 51, 26, 55 },
		{ 51, 27, 55 },
		{ 51, 28, 55 },
		{ 51, 29, 55 },
		{ 51, 30, 55 },
		{ 51, 31, 55 },
		{ 51, 32, 55 },
		{ 51, 33, 55 },
		{ 51, 34, 55 },
		{ 51, 35, 55 },
		{ 51, 36, 55 },
		{ 51, 37, 55 },
		{ 51, 38, 55 },
		{ 51, 39, 55 },
		{ 51, 40, 55 },
		{ 51, 41, 55 },
		{ 51, 42, 55 },
		{ 51, 43, 55 },
		{ 51, 44, 55 },
		{ 51, 45, 55 },
		{ 51, 46, 55 },
		{ 51, 47, 55 },
		{ 51, 48, 55 },
		{ 51, 49, 55 },
		{ 51, 50, 55 },
		{ 52, 1, 55 },
		{ 52, 2, 55 },
		{ 52, 3, 55 },
		{ 52, 4, 55 },
		{ 52, 5, 55 },
		{ 52, 6, 55 },
		{ 52, 7, 55 },
		{ 52, 8, 55 },
		{ 52, 9, 55 },
		{ 52, 10, 55 },
		{ 52, 11, 55 },
		{ 52, 12, 55 },
		{ 52, 13, 55 },
		{ 52, 14, 55 },
		{ 52, 15, 55 },
		{ 52, 16, 55 },
		{ 52, 17, 55 },
		{ 52, 18, 55 },
		{ 52, 19, 55 },
		{ 52, 20, 55 },
		{ 52, 21, 55 },
		{ 52, 22, 55 },
		{ 52, 23, 55 },
		{ 52, 24, 55 },
		{ 52, 25, 55 },
		{ 52, 26, 55 },
		{ 52, 27, 55 },
		{ 52, 28, 55 },
		{ 52, 29, 55 },
		{ 52, 30, 55 },
		{ 52, 31, 55 },
		{ 52, 32, 55 },
		{ 52, 33, 55 },
		{ 52, 34, 55 },
		{ 52, 35, 55 },
		{ 52, 36, 55 },
		{ 52, 37, 55 },
		{ 52, 38, 55 },
		{ 52, 39, 55 },
		{ 52, 40, 55 },
		{ 52, 41, 55 },
		{ 52, 42, 55 },
		{ 52, 43, 55 },
		{ 52, 44, 55 },
		{ 52, 45, 55 },
		{ 52, 46, 55 },
		{ 52, 47, 55 },
		{ 52, 48, 55 },
		{ 52, 49, 55 },
		{ 52, 50, 55 },
		{ 52, 51, 55 },
		{ 53, 1, 55 },
		{ 53, 2, 55 },
		{ 53, 3, 55 },
		{ 53, 4, 55 },
		{ 53, 5, 55 },
		{ 53, 6, 55 },
		{ 53, 7, 55 },
		{ 53, 8, 55 },
		{ 53, 9, 55 },
		{ 53, 10, 55 },
		{ 53, 11, 55 },
		{ 53, 12, 55 },
		{ 53, 13, 55 },
		{ 53, 14, 55 },
		{ 53, 15, 55 },
		{ 53, 16, 55 },
		{ 53, 17, 55 },
		{ 53, 18, 55 },
		{ 53, 19, 55 },
		{ 53, 20, 55 },
		{ 53, 21, 55 },
		{ 53, 22, 55 },
		{ 53, 23, 55 },
		{ 53, 24, 55 },
		{ 53, 25, 55 },
		{ 53, 26, 55 },
		{ 53, 27, 55 },
		{ 53, 28, 55 },
		{ 53, 29, 55 },
		{ 53, 30, 55 },
		{ 53, 31, 55 },
		{ 53, 32, 55 },
		{ 53, 33, 55 },
		{ 53, 34, 55 },
		{ 53, 35, 55 },
		{ 53, 36, 55 },
		{ 53, 37, 55 },
		{ 53, 38, 55 },
		{ 53, 39, 55 },
		{ 53, 40, 55 },
		{ 53, 41, 55 },
		{ 53, 42, 55 },
		{ 53, 43, 55 },
		{ 53, 44, 55 },
		{ 53, 45, 55 },
		{ 53, 46, 55 },
		{ 53, 47, 55 },
		{ 53, 48, 55 },
		{ 53, 49, 55 },
		{ 53, 50, 55 },
		{ 53, 51, 55 },
		{ 53, 52, 55 },
		{ 54, 1, 55 },
		{ 54, 2, 55 },
		{ 54, 3, 55 },
		{ 54, 4, 55 },
		{ 54, 5, 55 },
		{ 54, 6, 55 },
		{ 54, 7, 55 },
		{ 54, 8, 55 },
		{ 54, 9, 55 },
		{ 54, 10, 55 },
		{ 54, 11, 55 },
		{ 54, 12, 55 },
		{ 54, 13, 55 },
		{ 54, 14, 55 },
		{ 54, 15, 55 },
		{ 54, 16, 55 },
		{ 54, 17, 55 },
		{ 54, 18, 55 },
		{ 54, 19, 55 },
		{ 54, 20, 55 },
		{ 54, 21, 55 },
		{ 54, 22, 55 },
		{ 54, 23, 55 },
		{ 54, 24, 55 },
		{ 54, 25, 55 },
		{ 54, 26, 55 },
		{ 54, 27, 55 },
		{ 54, 28, 55 },
		{ 54, 29, 55 },
		{ 54, 30, 55 },
		{ 54, 31, 55 },
		{ 54, 32, 55 },
		{ 54, 33, 55 },
		{ 54, 34, 55 },
		{ 54, 35, 55 },
		{ 54, 36, 55 },
		{ 54, 37, 55 },
		{ 54, 38, 55 },
		{ 54, 39, 55 },
		{ 54, 40, 55 },
		{ 54, 41, 55 },
		{ 54, 42, 55 },
		{ 54, 43, 55 },
		{ 54, 44, 55 },
		{ 54, 45, 55 },
		{ 54, 46, 55 },
		{ 54, 47, 55 },
		{ 54, 48, 55 },
		{ 54, 49, 55 },
		{ 54, 50, 55 },
		{ 54, 51, 55 },
		{ 54, 52, 55 },
		{ 54, 53, 55 },
		{ 55, 1, 55 },
		{ 55, 2, 55 },
		{ 55, 3, 55 },
		{ 55, 4, 55 },
		{ 55, 5, 55 },
		{ 55, 6, 55 },
		{ 55, 7, 55 },
		{ 55, 8, 55 },
		{ 55, 9, 55 },
		{ 55, 10, 55 },
		{ 55, 11, 55 },
		{ 55, 12, 55 },
		{ 55, 13, 55 },
		{ 55, 14, 55 },
		{ 55, 15, 55 },
		{ 55, 16, 55 },
		{ 55, 17, 55 },
		{ 55, 18, 55 },
		{ 55, 19, 55 },
		{ 55, 20, 55 },
		{ 55, 21, 55 },
		{ 55, 22, 55 },
		{ 55, 23, 55 },
		{ 55, 24, 55 },
		{ 55, 25, 55 },
		{ 55, 26, 55 },
		{ 55, 27, 55 },
		{ 55, 28, 55 },
		{ 55, 29, 55 },
		{ 55, 30, 55 },
		{ 55, 31, 55 },
		{ 55, 32, 55 },
		{ 55, 33, 55 },
		{ 55, 34, 55 },
		{ 55, 35, 55 },
		{ 55, 36, 55 },
		{ 55, 37, 55 },
		{ 55, 38, 55 },
		{ 55, 39, 55 },
		{ 55, 40, 55 },
		{ 55, 41, 55 },
		{ 55, 42, 55 },
		{ 55, 43, 55 },
		{ 55, 44, 55 },
		{ 55, 45, 55 },
		{ 55, 46, 55 },
		{ 55, 47, 55 },
		{ 55, 48, 55 },
		{ 55, 49, 55 },
		{ 55, 50, 55 },
		{ 55, 51, 55 },
		{ 55, 52, 55 },
		{ 55, 53, 55 },
		{ 55, 54, 55 },
		{ 56, 1, 55 },
		{ 56, 2, 55 },
		{ 56, 3, 55 },
		{ 56, 4, 55 },
		{ 56, 5, 55 },
		{ 56, 6, 55 },
		{ 56, 7, 55 },
		{ 56, 8, 55 },
		{ 56, 9, 55 },
		{ 56, 10, 55 },
		{ 56, 11, 55 },
		{ 56, 12, 55 },
		{ 56, 13, 55 },
		{ 56, 14, 55 },
		{ 56, 15, 55 },
		{ 56, 16, 55 },
		{ 56, 17, 55 },
		{ 56, 18, 55 },
		{ 56, 19, 55 },
		{ 56, 20, 55 },
		{ 56, 21, 55 },
		{ 56, 22, 55 },
		{ 56, 23, 55 },
		{ 56, 24, 55 },
		{ 56, 25, 55 },
		{ 56, 26, 55 },
		{ 56, 27, 55 },
		{ 56, 28, 55 },
		{ 56, 29, 55 },
		{ 56, 30, 55 },
		{ 56, 31, 55 },
		{ 56, 32, 55 },
		{ 56, 33, 55 },
		{ 56, 34, 55 },
		{ 56, 35, 55 },
		{ 56, 36, 55 },
		{ 56, 37, 55 },
		{ 56, 38, 55 },
		{ 56, 39, 55 },
		{ 56, 40, 55 },
		{ 56, 41, 55 },
		{ 56, 42, 55 },
		{ 56, 43, 55 },
		{ 56, 44, 55 },
		{ 56, 45, 55 },
		{ 56, 46, 55 },
		{ 56, 47, 55 },
		{ 56, 48, 55 },
		{ 56, 49, 55 },
		{ 56, 50, 55 },
		{ 56, 51, 55 },
		{ 56, 52, 55 },
		{ 56, 53, 55 },
		{ 56, 54, 55 },
		{ 56, 55, 55 },
		{ 57, 1, 55 },
		{ 57, 2, 55 },
		{ 57, 3, 55 },
		{ 57, 4, 55 },
		{ 57, 5, 55 },
		{ 57, 6, 55 },
		{ 57, 7, 55 },
		{ 57, 8, 55 },
		{ 57, 9, 55 },
		{ 57, 10, 55 },
		{ 57, 11, 55 },
		{ 57, 12, 55 },
		{ 57, 13, 55 },
		{ 57, 14, 55 },
		{ 57, 15, 55 },
		{ 57, 16, 55 },
		{ 57, 17, 55 },
		{ 57, 18, 55 },
		{ 57, 19, 55 },
		{ 57, 20, 55 },
		{ 57, 21, 55 },
		{ 57, 22, 55 },
		{ 57, 23, 55 },
		{ 57, 24, 55 },
		{ 57, 25, 55 },
		{ 57, 26, 55 },
		{ 57, 27, 55 },
		{ 57, 28, 55 },
		{ 57, 29, 55 },
		{ 57, 30, 55 },
		{ 57, 31, 55 },
		{ 57, 32, 55 },
		{ 57, 33, 55 },
		{ 57, 34, 55 },
		{ 57, 35, 55 },
		{ 57, 36, 55 },
		{ 57, 37, 55 },
		{ 57, 38, 55 },
		{ 57, 39, 55 },
		{ 57, 40, 55 },
		{ 57, 41, 55 },
		{ 57, 42, 55 },
		{ 57, 43, 55 },
		{ 57, 44, 55 },
		{ 57, 45, 55 },
		{ 57, 46, 55 },
		{ 57, 47, 55 },
		{ 57, 48, 55 },
		{ 57, 49, 55 },
		{ 57, 50, 55 },
		{ 57, 51, 55 },
		{ 57, 52, 55 },
		{ 57, 53, 55 },
		{ 57, 54, 55 },
		{ 57, 55, 55 },
		{ 57, 56, 55 },
		{ 58, 1, 55 },
		{ 58, 2, 55 },
		{ 58, 3, 55 },
		{ 58, 4, 55 },
		{ 58, 5, 55 },
		{ 58, 6, 55 },
		{ 58, 7, 55 },
		{ 58, 8, 55 },
		{ 58, 9, 55 },
		{ 58, 10, 55 },
		{ 58, 11, 55 },
		{ 58, 12, 55 },
		{ 58, 13, 55 },
		{ 58, 14, 55 },
		{ 58, 15, 55 },
		{ 58, 16, 55 },
		{ 58, 17, 55 },
		{ 58, 18, 55 },
		{ 58, 19, 55 },
		{ 58, 20, 55 },
		{ 58, 21, 55 },
		{ 58, 22, 55 },
		{ 58, 23, 55 },
		{ 58, 24, 55 },
		{ 58, 25, 55 },
		{ 58, 26, 55 },
		{ 58, 27, 55 },
		{ 58, 28, 55 },
		{ 58, 29, 55 },
		{ 58, 30, 55 },
		{ 58, 31, 55 },
		{ 58, 32, 55 },
		{ 58, 33, 55 },
		{ 58, 34, 55 },
		{ 58, 35, 55 },
		{ 58, 36, 55 },
		{ 58, 37, 55 },
		{ 58, 38, 55 },
		{ 58, 39, 55 },
		{ 58, 40, 55 },
		{ 58, 41, 55 },
		{ 58, 42, 55 },
		{ 58, 43, 55 },
		{ 58, 44, 55 },
		{ 58, 45, 55 },
		{ 58, 46, 55 },
		{ 58, 47, 55 },
		{ 58, 48, 55 },
		{ 58, 49, 55 },
		{ 58, 50, 55 },
		{ 58, 51, 55 },
		{ 58, 52, 55 },
		{ 58, 53, 55 },
		{ 58, 54, 55 },
		{ 58, 55, 55 },
		{ 58, 56, 55 },
		{ 58, 57, 55 },
		{ 59, 1, 55 },
		{ 59, 2, 55 },
		{ 59, 3, 55 },
		{ 59, 4, 55 },
		{ 59, 5, 55 },
		{ 59, 6, 55 },
		{ 59, 7, 55 },
		{ 59, 8, 55 },
		{ 59, 9, 55 },
		{ 59, 10, 55 },
		{ 59, 11, 55 },
		{ 59, 12, 55 },
		{ 59, 13, 55 },
		{ 59, 14, 55 },
		{ 59, 15, 55 },
		{ 59, 16, 55 },
		{ 59, 17, 55 },
		{ 59, 18, 55 },
		{ 59, 19, 55 },
		{ 59, 20, 55 },
		{ 59, 21, 55 },
		{ 59, 22, 55 },
		{ 59, 23, 55 },
		{ 59, 24, 55 },
		{ 59, 25, 55 },
		{ 59, 26, 55 },
		{ 59, 27, 55 },
		{ 59, 28, 55 },
		{ 59, 29, 55 },
		{ 59, 30, 55 },
		{ 59, 31, 55 },
		{ 59, 32, 55 },
		{ 59, 33, 55 },
		{ 59, 34, 55 },
		{ 59, 35, 55 },
		{ 59, 36, 55 },
		{ 59, 37, 55 },
		{ 59, 38, 55 },
		{ 59, 39, 55 },
		{ 59, 40, 55 },
		{ 59, 41, 55 },
		{ 59, 42, 55 },
		{ 59, 43, 55 },
		{ 59, 44, 55 },
		{ 59, 45, 55 },
		{ 59, 46, 55 },
		{ 59, 47, 55 },
		{ 59, 48, 55 },
		{ 59, 49, 55 },
		{ 59, 50, 55 },
		{ 59, 51, 55 },
		{ 59, 52, 55 },
		{ 59, 53, 55 },
		{ 59, 54, 55 },
		{ 59, 55, 55 },
		{ 59, 56, 55 },
		{ 59, 57, 55 },
		{ 59, 58, 55 },
		{ 60, 1, 55 },
		{ 60, 2, 55 },
		{ 60, 3, 55 },
		{ 60, 4, 55 },
		{ 60, 5, 55 },
		{ 60, 6, 55 },
		{ 60, 7, 55 },
		{ 60, 8, 55 },
		{ 60, 9, 55 },
		{ 60, 10, 55 },
		{ 60, 11, 55 },
		{ 60, 12, 55 },
		{ 60, 13, 55 },
		{ 60, 14, 55 },
		{ 60, 15, 55 },
		{ 60, 16, 55 },
		{ 60, 17, 55 },
		{ 60, 18, 55 },
		{ 60, 19, 55 },
		{ 60, 20, 55 },
		{ 60, 21, 55 },
		{ 60, 22, 55 },
		{ 60, 23, 55 },
		{ 60, 24, 55 },
		{ 60, 25, 55 },
		{ 60, 26, 55 },
		{ 60, 27, 55 },
		{ 60, 28, 55 },
		{ 60, 29, 55 },
		{ 60, 30, 55 },
		{ 60, 31, 55 },
		{ 60, 32, 55 },
		{ 60, 33, 55 },
		{ 60, 34, 55 },
		{ 60, 35, 55 },
		{ 60, 36, 55 },
		{ 60, 37, 55 },
		{ 60, 38, 55 },
		{ 60, 39, 55 },
		{ 60, 40, 55 },
		{ 60, 41, 55 },
		{ 60, 42, 55 },
		{ 60, 43, 55 },
		{ 60, 44, 55 },
		{ 60, 45, 55 },
		{ 60, 46, 55 },
		{ 60, 47, 55 },
		{ 60, 48, 55 },
		{ 60, 49, 55 },
		{ 60, 50, 55 },
		{ 60, 51, 55 },
		{ 60, 52, 55 },
		{ 60, 53, 55 },
		{ 60, 54, 55 },
		{ 60, 55, 55 },
		{ 60, 56, 55 },
		{ 60, 57, 55 },
		{ 60, 58, 55 },
		{ 60, 59, 55 },
		{ 61, 1, 55 },
		{ 61, 2, 55 },
		{ 61, 3, 55 },
		{ 61, 4, 55 },
		{ 61, 5, 55 },
		{ 61, 6, 55 },
		{ 61, 7, 55 },
		{ 61, 8, 55 },
		{ 61, 9, 55 },
		{ 61, 10, 55 },
		{ 61, 11, 55 },
		{ 61, 12, 55 },
		{ 61, 13, 55 },
		{ 61, 14, 55 },
		{ 61, 15, 55 },
		{ 61, 16, 55 },
		{ 61, 17, 55 },
		{ 61, 18, 55 },
		{ 61, 19, 55 },
		{ 61, 20, 55 },
		{ 61, 21, 55 },
		{ 61, 22, 55 },
		{ 61, 23, 55 },
		{ 61, 24, 55 },
		{ 61, 25, 55 },
		{ 61, 26, 55 },
		{ 61, 27, 55 },
		{ 61, 28, 55 },
		{ 61, 29, 55 },
		{ 61, 30, 55 },
		{ 61, 31, 55 },
		{ 61, 32, 55 },
		{ 61, 33, 55 },
		{ 61, 34, 55 },
		{ 61, 35, 55 },
		{ 61, 36, 55 },
		{ 61, 37, 55 },
		{ 61, 38, 55 },
		{ 61, 39, 55 },
		{ 61, 40, 55 },
		{ 61, 41, 55 },
		{ 61, 42, 55 },
		{ 61, 43, 55 },
		{ 61, 44, 55 },
		{ 61, 45, 55 },
		{ 61, 46, 55 },
		{ 61, 47, 55 },
		{ 61, 48, 55 },
		{ 61, 49, 55 },
		{ 61, 50, 55 },
		{ 61, 51, 55 },
		{ 61, 52, 55 },
		{ 61, 53, 55 },
		{ 61, 54, 55 },
		{ 61, 55, 55 },
		{ 61, 56, 55 },
		{ 61, 57, 55 },
		{ 61, 58, 55 },
		{ 61, 59, 55 },
		{ 61, 60, 55 },
		{ 62, 1, 55 },
		{ 62, 2, 55 },
		{ 62, 3, 55 },
		{ 62, 4, 55 },
		{ 62, 5, 55 },
		{ 62, 6, 55 },
		{ 62, 7, 55 },
		{ 62, 8, 55 },
		{ 62, 9, 55 },
		{ 62, 10, 55 },
		{ 62, 11, 55 },
		{ 62, 12, 55 },
		{ 62, 13, 55 },
		{ 62, 14, 55 },
		{ 62, 15, 55 },
		{ 62, 16, 55 },
		{ 62, 17, 55 },
		{ 62, 18, 55 },
		{ 62, 19, 55 },
		{ 62, 20, 55 },
		{ 62, 21, 55 },
		{ 62, 22, 55 },
		{ 62, 23, 55 },
		{ 62, 24, 55 },
		{ 62, 25, 55 },
		{ 62, 26, 55 },
		{ 62, 27, 55 },
		{ 62, 28, 55 },
		{ 62, 29, 55 },
		{ 62, 30, 55 },
		{ 62, 31, 55 },
		{ 62, 32, 55 },
		{ 62, 33, 55 },
		{ 62, 34, 55 },
		{ 62, 35, 55 },
		{ 62, 36, 55 },
		{ 62, 37, 55 },
		{ 62, 38, 55 },
		{ 62, 39, 55 },
		{ 62, 40, 55 },
		{ 62, 41, 55 },
		{ 62, 42, 55 },
		{ 62, 43, 55 },
		{ 62, 44, 55 },
		{ 62, 45, 55 },
		{ 62, 46, 55 },
		{ 62, 47, 55 },
		{ 62, 48, 55 },
		{ 62, 49, 55 },
		{ 62, 50, 55 },
		{ 62, 51, 55 },
		{ 62, 52, 55 },
		{ 62, 53, 55 },
		{ 62, 54, 55 },
		{ 62, 55, 55 },
		{ 62, 56, 55 },
		{ 62, 57, 55 },
		{ 62, 58, 55 },
		{ 62, 59, 55 },
		{ 62, 60, 55 },
		{ 62, 61, 55 },
		{ 63, 1, 55 },
		{ 63, 2, 55 },
		{ 63, 3, 55 },
		{ 63, 4, 55 },
		{ 63, 5, 55 },
		{ 63, 6, 55 },
		{ 63, 7, 55 },
		{ 63, 8, 55 },
		{ 63, 9, 55 },
		{ 63, 10, 55 },
		{ 63, 11, 55 },
		{ 63, 12, 55 },
		{ 63, 13, 55 },
		{ 63, 14, 55 },
		{ 63, 15, 55 },
		{ 63, 16, 55 },
		{ 63, 17, 55 },
		{ 63, 18, 55 },
		{ 63, 19, 55 },
		{ 63, 20, 55 },
		{ 63, 21, 55 },
		{ 63, 22, 55 },
		{ 63, 23, 55 },
		{ 63, 24, 55 },
		{ 63, 25, 55 },
		{ 63, 26, 55 },
		{ 63, 27, 55 },
		{ 63, 28, 55 },
		{ 63, 29, 55 },
		{ 63, 30, 55 },
		{ 63, 31, 55 },
		{ 63, 32, 55 },
		{ 63, 33, 55 },
		{ 63, 34, 55 },
		{ 63, 35, 55 },
		{ 63, 36, 55 },
		{ 63, 37, 55 },
		{ 63, 38, 55 },
		{ 63, 39, 55 },
		{ 63, 40, 55 },
		{ 63, 41, 55 },
		{ 63, 42, 55 },
		{ 63, 43, 55 },
		{ 63, 44, 55 },
		{ 63, 45, 55 },
		{ 63, 46, 55 },
		{ 63, 47, 55 },
		{ 63, 48, 55 },
		{ 63, 49, 55 },
		{ 63, 50, 55 },
		{ 63, 51, 55 },
		{ 63, 52, 55 },
		{ 63, 53, 55 },
		{ 63, 54, 55 },
		{ 63, 55, 55 },
		{ 63, 56, 55 },
		{ 63, 57, 55 },
		{ 63, 58, 55 },
		{ 63, 59, 55 },
		{ 63, 60, 55 },
		{ 63, 61, 55 },
		{ 63, 62, 55 },
		{ 64, 1, 55 },
		{ 64, 2, 55 },
		{ 64, 3, 55 },
		{ 64, 4, 55 },
		{ 64, 5, 55 },
		{ 64, 6, 55 },
		{ 64, 7, 55 },
		{ 64, 8, 55 },
		{ 64, 9, 55 },
		{ 64, 10, 55 },
		{ 64, 11, 55 },
		{ 64, 12, 55 },
		{ 64, 13, 55 },
		{ 64, 14, 55 },
		{ 64, 15, 55 },
		{ 64, 16, 55 },
		{ 64, 17, 55 },
		{ 64, 18, 55 },
		{ 64, 19, 55 },
		{ 64, 20, 55 },
		{ 64, 21, 55 },
		{ 64, 22, 55 },
		{ 64, 23, 55 },
		{ 64, 24, 55 },
		{ 64, 25, 55 },
		{ 64, 26, 55 },
		{ 64, 27, 55 },
		{ 64, 28, 55 },
		{ 64, 29, 55 },
		{ 64, 30, 55 },
		{ 64, 31, 55 },
		{ 64, 32, 55 },
		{ 64, 33, 55 },
		{ 64, 34, 55 },
		{ 64, 35, 55 },
		{ 64, 36, 55 },
		{ 64, 37, 55 },
		{ 64, 38, 55 },
		{ 64, 39, 55 },
		{ 64, 40, 55 },
		{ 64, 41, 55 },
		{ 64, 42, 55 },
		{ 64, 43, 55 },
		{ 64, 44, 55 },
		{ 64, 45, 55 },
		{ 64, 46, 55 },
		{ 64, 47, 55 },
		{ 64, 48, 55 },
		{ 64, 49, 55 },
		{ 64, 50, 55 },
		{ 64, 51, 55 },
		{ 64, 52, 55 },
		{ 64, 53, 55 },
		{ 64, 54, 55 },
		{ 64, 55, 55 },
		{ 64, 56, 55 },
		{ 64, 57, 55 },
		{ 64, 58, 55 },
		{ 64, 59, 55 },
		{ 64, 60, 55 },
		{ 64, 61, 55 },
		{ 64, 62, 55 },
		{ 64, 63, 55 },
		{ 65, 2, 55 },
		{ 65, 3, 55 },
		{ 65, 4, 55 },
		{ 65, 5, 55 },
		{ 65, 6, 55 },
		{ 65, 7, 55 },
		{ 65, 8, 55 },
		{ 65, 9, 55 },
		{ 65, 10, 55 },
		{ 65, 11, 55 },
		{ 65, 12, 55 },
		{ 65, 13, 55 },
		{ 65, 14, 55 },
		{ 65, 15, 55 },
		{ 65, 16, 55 },
		{ 65, 17, 55 },
		{ 65, 18, 55 },
		{ 65, 19, 55 },
		{ 65, 20, 55 },
		{ 65, 21, 55 },
		{ 65, 22, 55 },
		{ 65, 23, 55 },
		{ 65, 24, 55 },
		{ 65, 25, 55 },
		{ 65, 26, 55 },
		{ 65, 27, 55 },
		{ 65, 28, 55 },
		{ 65, 29, 55 },
		{ 65, 30, 55 },
		{ 65, 31, 55 },
		{ 65, 32, 55 },
		{ 65, 33, 55 },
		{ 65, 34, 55 },
		{ 65, 35, 55 },
		{ 65, 36, 55 },
		{ 65, 37, 55 },
		{ 65, 38, 55 },
		{ 65, 39, 55 },
		{ 65, 40, 55 },
		{ 65, 41, 55 },
		{ 65, 42, 55 },
		{ 65, 43, 55 },
		{ 65, 44, 55 },
		{ 65, 45, 55 },
		{ 65, 46, 55 },
		{ 65, 47, 55 },
		{ 65, 48, 55 },
		{ 65, 49, 55 },
		{ 65, 50, 55 },
		{ 65, 51, 55 },
		{ 65, 52, 55 },
		{ 65, 53, 55 },
		{ 65, 54, 55 },
		{ 65, 55, 55 },
		{ 65, 56, 55 },
		{ 65, 57, 55 },
		{ 65, 58, 55 },
		{ 65, 59, 55 },
		{ 65, 60, 55 },
		{ 65, 61, 55 },
		{ 65, 62, 55 },
		{ 65, 63, 55 },
		{ 65, 64, 55 },
		{ 66, 3, 55 },
		{ 66, 4, 55 },
		{ 66, 5, 55 },
		{ 66, 6, 55 },
		{ 66, 7, 55 },
		{ 66, 8, 55 },
		{ 66, 9, 55 },
		{ 66, 10, 55 },
		{ 66, 11, 55 },
		{ 66, 12, 55 },
		{ 66, 13, 55 },
		{ 66, 14, 55 },
		{ 66, 15, 55 },
		{ 66, 16, 55 },
		{ 66, 17, 55 },
		{ 66, 18, 55 },
		{ 66, 19, 55 },
		{ 66, 20, 55 },
		{ 66, 21, 55 },
		{ 66, 22, 55 },
		{ 66, 23, 55 },
		{ 66, 24, 55 },
		{ 66, 25, 55 },
		{ 66, 26, 55 },
		{ 66, 27, 55 },
		{ 66, 28, 55 },
		{ 66, 29, 55 },
		{ 66, 30, 55 },
		{ 66, 31, 55 },
		{ 66, 32, 55 },
		{ 66, 33, 55 },
		{ 66, 34, 55 },
		{ 66, 35, 55 },
		{ 66, 36, 55 },
		{ 66, 37, 55 },
		{ 66, 38, 55 },
		{ 66, 39, 55 },
		{ 66, 40, 55 },
		{ 66, 41, 55 },
		{ 66, 42, 55 },
		{ 66, 43, 55 },
		{ 66, 44, 55 },
		{ 66, 45, 55 },
		{ 66, 46, 55 },
		{ 66, 47, 55 },
		{ 66, 48, 55 },
		{ 66, 49, 55 },
		{ 66, 50, 55 },
		{ 66, 51, 55 },
		{ 66, 52, 55 },
		{ 66, 53, 55 },
		{ 66, 54, 55 },
		{ 66, 55, 55 },
		{ 66, 56, 55 },
		{ 66, 57, 55 },
		{ 66, 58, 55 },
		{ 66, 59, 55 },
		{ 66, 60, 55 },
		{ 66, 61, 55 },
		{ 66, 62, 55 },
		{ 66, 63, 55 },
		{ 66, 64, 55 },
		{ 66, 65, 55 },
		{ 67, 4, 55 },
		{ 67, 5, 55 },
		{ 67, 6, 55 },
		{ 67, 7, 55 },
		{ 67, 8, 55 },
		{ 67, 9, 55 },
		{ 67, 10, 55 },
		{ 67, 11, 55 },
		{ 67, 12, 55 },
		{ 67, 13, 55 },
		{ 67, 14, 55 },
		{ 67, 15, 55 },
		{ 67, 16, 55 },
		{ 67, 17, 55 },
		{ 67, 18, 55 },
		{ 67, 19, 55 },
		{ 67, 20, 55 },
		{ 67, 21, 55 },
		{ 67, 22, 55 },
		{ 67, 23, 55 },
		{ 67, 24, 55 },
		{ 67, 25, 55 },
		{ 67, 26, 55 },
		{ 67, 27, 55 },
		{ 67, 28, 55 },
		{ 67, 29, 55 },
		{ 67, 30, 55 },
		{ 67, 31, 55 },
		{ 67, 32, 55 },
		{ 67, 33, 55 },
		{ 67, 34, 55 },
		{ 67, 35, 55 },
		{ 67, 36, 55 },
		{ 67, 37, 55 },
		{ 67, 38, 55 },
		{ 67, 39, 55 },
		{ 67, 40, 55 },
		{ 67, 41, 55 },
		{ 67, 42, 55 },
		{ 67, 43, 55 },
		{ 67, 44, 55 },
		{ 67, 45, 55 },
		{ 67, 46, 55 },
		{ 67, 47, 55 },
		{ 67, 48, 55 },
		{ 67, 49, 55 },
		{ 67, 50, 55 },
		{ 67, 51, 55 },
		{ 67, 52, 55 },
		{ 67, 53, 55 },
		{ 67, 54, 55 },
		{ 67, 55, 55 },
		{ 67, 56, 55 },
		{ 67, 57, 55 },
		{ 67, 58, 55 },
		{ 67, 59, 55 },
		{ 67, 60, 55 },
		{ 67, 61, 55 },
		{ 67, 62, 55 },
		{ 67, 63, 55 },
		{ 67, 64, 55 },
		{ 67, 65, 55 },
		{ 67, 66, 55 },
		{ 68, 6, 55 },
		{ 68, 7, 55 },
		{ 68, 8, 55 },
		{ 68, 9, 55 },
		{ 68, 10, 55 },
		{ 68, 11, 55 },
		{ 68, 12, 55 },
		{ 68, 13, 55 },
		{ 68, 14, 55 },
		{ 68, 15, 55 },
		{ 68, 16, 55 },
		{ 68, 17, 55 },
		{ 68, 18, 55 },
		{ 68, 19, 55 },
		{ 68, 20, 55 },
		{ 68, 21, 55 },
		{ 68, 22, 55 },
		{ 68, 23, 55 },
		{ 68, 24, 55 },
		{ 68, 25, 55 },
		{ 68, 26, 55 },
		{ 68, 27, 55 },
		{ 68, 28, 55 },
		{ 68, 29, 55 },
		{ 68, 30, 55 },
		{ 68, 31, 55 },
		{ 68, 32, 55 },
		{ 68, 33, 55 },
		{ 68, 34, 55 },
		{ 68, 35, 55 },
		{ 68, 36, 55 },
		{ 68, 37, 55 },
		{ 68, 38, 55 },
		{ 68, 39, 55 },
		{ 68, 40, 55 },
		{ 68, 41, 55 },
		{ 68, 42, 55 },
		{ 68, 43, 55 },
		{ 68, 44, 55 },
		{ 68, 45, 55 },
		{ 68, 46, 55 },
		{ 68, 47, 55 },
		{ 68, 48, 55 },
		{ 68, 49, 55 },
		{ 68, 50, 55 },
		{ 68, 51, 55 },
		{ 68, 52, 55 },
		{ 68, 53, 55 },
		{ 68, 54, 55 },
		{ 68, 55, 55 },
		{ 68, 56, 55 },
		{ 68, 57, 55 },
		{ 68, 58, 55 },
		{ 68, 59, 55 },
		{ 68, 60, 55 },
		{ 68, 61, 55 },
		{ 68, 62, 55 },
		{ 68, 63, 55 },
		{ 68, 64, 55 },
		{ 68, 65, 55 },
		{ 68, 66, 55 },
		{ 68, 67, 55 },
		{ 69, 7, 55 },
		{ 69, 8, 55 },
		{ 69, 9, 55 },
		{ 69, 10, 55 },
		{ 69, 11, 55 },
		{ 69, 12, 55 },
		{ 69, 13, 55 },
		{ 69, 14, 55 },
		{ 69, 15, 55 },
		{ 69, 16, 55 },
		{ 69, 17, 55 },
		{ 69, 18, 55 },
		{ 69, 19, 55 },
		{ 69, 20, 55 },
		{ 69, 21, 55 },
		{ 69, 22, 55 },
		{ 69, 23, 55 },
		{ 69, 24, 55 },
		{ 69, 25, 55 },
		{ 69, 26, 55 },
		{ 69, 27, 55 },
		{ 69, 28, 55 },
		{ 69, 29, 55 },
		{ 69, 30, 55 },
		{ 69, 31, 55 },
		{ 69, 32, 55 },
		{ 69, 33, 55 },
		{ 69, 34, 55 },
		{ 69, 35, 55 },
		{ 69, 36, 55 },
		{ 69, 37, 55 },
		{ 69, 38, 55 },
		{ 69, 39, 55 },
		{ 69, 40, 55 },
		{ 69, 41, 55 },
		{ 69, 42, 55 },
		{ 69, 43, 55 },
		{ 69, 44, 55 },
		{ 69, 45, 55 },
		{ 69, 46, 55 },
		{ 69, 47, 55 },
		{ 69, 48, 55 },
		{ 69, 49, 55 },
		{ 69, 50, 55 },
		{ 69, 51, 55 },
		{ 69, 52, 55 },
		{ 69, 53, 55 },
		{ 69, 54, 55 },
		{ 69, 55, 55 },
		{ 69, 56, 55 },
		{ 69, 57, 55 },
		{ 69, 58, 55 },
		{ 69, 59, 55 },
		{ 69, 60, 55 },
		{ 69, 61, 55 },
		{ 69, 62, 55 },
		{ 69, 63, 55 },
		{ 69, 64, 55 },
		{ 69, 65, 55 },
		{ 69, 66, 55 },
		{ 69, 67, 55 },
		{ 69, 68, 55 },
		{ 70, 8, 55 },
		{ 70, 9, 55 },
		{ 70, 10, 55 },
		{ 70, 11, 55 },
		{ 70, 12, 55 },
		{ 70, 13, 55 },
		{ 70, 14, 55 },
		{ 70, 15, 55 },
		{ 70, 16, 55 },
		{ 70, 17, 55 },
		{ 70, 18, 55 },
		{ 70, 19, 55 },
		{ 70, 20, 55 },
		{ 70, 21, 55 },
		{ 70, 22, 55 },
		{ 70, 23, 55 },
		{ 70, 24, 55 },
		{ 70, 25, 55 },
		{ 70, 26, 55 },
		{ 70, 27, 55 },
		{ 70, 28, 55 },
		{ 70, 29, 55 },
		{ 70, 30, 55 },
		{ 70, 31, 55 },
		{ 70, 32, 55 },
		{ 70, 33, 55 },
		{ 70, 34, 55 },
		{ 70, 35, 55 },
		{ 70, 36, 55 },
		{ 70, 37, 55 },
		{ 70, 38, 55 },
		{ 70, 39, 55 },
		{ 70, 40, 55 },
		{ 70, 41, 55 },
		{ 70, 42, 55 },
		{ 70, 43, 55 },
		{ 70, 44, 55 },
		{ 70, 45, 55 },
		{ 70, 46, 55 },
		{ 70, 47, 55 },
		{ 70, 48, 55 },
		{ 70, 49, 55 },
		{ 70, 50, 55 },
		{ 70, 51, 55 },
		{ 70, 52, 55 },
		{ 70, 53, 55 },
		{ 70, 54, 55 },
		{ 70, 55, 55 },
		{ 70, 56, 55 },
		{ 70, 57, 55 },
		{ 70, 58, 55 },
		{ 70, 59, 55 },
		{ 70, 60, 55 },
		{ 70, 61, 55 },
		{ 70, 62, 55 },
		{ 70, 63, 55 },
		{ 70, 64, 55 },
		{ 70, 65, 55 },
		{ 70, 66, 55 },
		{ 70, 67, 55 },
		{ 70, 68, 55 },
		{ 70, 69, 55 },
		{ 71, 10, 55 },
		{ 71, 11, 55 },
		{ 71, 12, 55 },
		{ 71, 13, 55 },
		{ 71, 14, 55 },
		{ 71, 15, 55 },
		{ 71, 16, 55 },
		{ 71, 17, 55 },
		{ 71, 18, 55 },
		{ 71, 19, 55 },
		{ 71, 20, 55 },
		{ 71, 21, 55 },
		{ 71, 22, 55 },
		{ 71, 23, 55 },
		{ 71, 24, 55 },
		{ 71, 25, 55 },
		{ 71, 26, 55 },
		{ 71, 27, 55 },
		{ 71, 28, 55 },
		{ 71, 29, 55 },
		{ 71, 30, 55 },
		{ 71, 31, 55 },
		{ 71, 32, 55 },
		{ 71, 33, 55 },
		{ 71, 34, 55 },
		{ 71, 35, 55 },
		{ 71, 36, 55 },
		{ 71, 37, 55 },
		{ 71, 38, 55 },
		{ 71, 39, 55 },
		{ 71, 40, 55 },
		{ 71, 41, 55 },
		{ 71, 42, 55 },
		{ 71, 43, 55 },
		{ 71, 44, 55 },
		{ 71, 45, 55 },
		{ 71, 46, 55 },
		{ 71, 47, 55 },
		{ 71, 48, 55 },
		{ 71, 49, 55 },
		{ 71, 50, 55 },
		{ 71, 51, 55 },
		{ 71, 52, 55 },
		{ 71, 53, 55 },
		{ 71, 54, 55 },
		{ 71, 55, 55 },
		{ 71, 56, 55 },
		{ 71, 57, 55 },
		{ 71, 58, 55 },
		{ 71, 59, 55 },
		{ 71, 60, 55 },
		{ 71, 61, 55 },
		{ 71, 62, 55 },
		{ 71, 63, 55 },
		{ 71, 64, 55 },
		{ 71, 65, 55 },
		{ 71, 66, 55 },
		{ 71, 67, 55 },
		{ 71, 68, 55 },
		{ 71, 69, 55 },
		{ 71, 70, 55 },
		{ 72, 11, 55 },
		{ 72, 12, 55 },
		{ 72, 13, 55 },
		{ 72, 14, 55 },
		{ 72, 15, 55 },
		{ 72, 16, 55 },
		{ 72, 17, 55 },
		{ 72, 18, 55 },
		{ 72, 19, 55 },
		{ 72, 20, 55 },
		{ 72, 21, 55 },
		{ 72, 22, 55 },
		{ 72, 23, 55 },
		{ 72, 24, 55 },
		{ 72, 25, 55 },
		{ 72, 26, 55 },
		{ 72, 27, 55 },
		{ 72, 28, 55 },
		{ 72, 29, 55 },
		{ 72, 30, 55 },
		{ 72, 31, 55 },
		{ 72, 32, 55 },
		{ 72, 33, 55 },
		{ 72, 34, 55 },
		{ 72, 35, 55 },
		{ 72, 36, 55 },
		{ 72, 37, 55 },
		{ 72, 38, 55 },
		{ 72, 39, 55 },
		{ 72, 40, 55 },
		{ 72, 41, 55 },
		{ 72, 42, 55 },
		{ 72, 43, 55 },
		{ 72, 44, 55 },
		{ 72, 45, 55 },
		{ 72, 46, 55 },
		{ 72, 47, 55 },
		{ 72, 48, 55 },
		{ 72, 49, 55 },
		{ 72, 50, 55 },
		{ 72, 51, 55 },
		{ 72, 52, 55 },
		{ 72, 53, 55 },
		{ 72, 54, 55 },
		{ 72, 55, 55 },
		{ 72, 56, 55 },
		{ 72, 57, 55 },
		{ 72, 58, 55 },
		{ 72, 59, 55 },
		{ 72, 60, 55 },
		{ 72, 61, 55 },
		{ 72, 62, 55 },
		{ 72, 63, 55 },
		{ 72, 64, 55 },
		{ 72, 65, 55 },
		{ 72, 66, 55 },
		{ 72, 67, 55 },
		{ 72, 68, 55 },
		{ 72, 69, 55 },
		{ 72, 70, 55 },
		{ 72, 71, 55 },
		{ 73, 12, 55 },
		{ 73, 13, 55 },
		{ 73, 14, 55 },
		{ 73, 15, 55 },
		{ 73, 16, 55 },
		{ 73, 17, 55 },
		{ 73, 18, 55 },
		{ 73, 19, 55 },
		{ 73, 20, 55 },
		{ 73, 21, 55 },
		{ 73, 22, 55 },
		{ 73, 23, 55 },
		{ 73, 24, 55 },
		{ 73, 25, 55 },
		{ 73, 26, 55 },
		{ 73, 27, 55 },
		{ 73, 28, 55 },
		{ 73, 29, 55 },
		{ 73, 30, 55 },
		{ 73, 31, 55 },
		{ 73, 32, 55 },
		{ 73, 33, 55 },
		{ 73, 34, 55 },
		{ 73, 35, 55 },
		{ 73, 36, 55 },
		{ 73, 37, 55 },
		{ 73, 38, 55 },
		{ 73, 39, 55 },
		{ 73, 40, 55 },
		{ 73, 41, 55 },
		{ 73, 42, 55 },
		{ 73, 43, 55 },
		{ 73, 44, 55 },
		{ 73, 45, 55 },
		{ 73, 46, 55 },
		{ 73, 47, 55 },
		{ 73, 48, 55 },
		{ 73, 49, 55 },
		{ 73, 50, 55 },
		{ 73, 51, 55 },
		{ 73, 52, 55 },
		{ 73, 53, 55 },
		{ 73, 54, 55 },
		{ 73, 55, 55 },
		{ 73, 56, 55 },
		{ 73, 57, 55 },
		{ 73, 58, 55 },
		{ 73, 59, 55 },
		{ 73, 60, 55 },
		{ 73, 61, 55 },
		{ 73, 62, 55 },
		{ 73, 63, 55 },
		{ 73, 64, 55 },
		{ 73, 65, 55 },
		{ 73, 66, 55 },
		{ 73, 67, 55 },
		{ 73, 68, 55 },
		{ 73, 69, 55 },
		{ 73, 70, 55 },
		{ 73, 71, 55 },
		{ 73, 72, 55 },
		{ 74, 14, 55 },
		{ 74, 15, 55 },
		{ 74, 16, 55 },
		{ 74, 17, 55 },
		{ 74, 18, 55 },
		{ 74, 19, 55 },
		{ 74, 20, 55 },
		{ 74, 21, 55 },
		{ 74, 22, 55 },
		{ 74, 23, 55 },
		{ 74, 24, 55 },
		{ 74, 25, 55 },
		{ 74, 26, 55 },
		{ 74, 27, 55 },
		{ 74, 28, 55 },
		{ 74, 29, 55 },
		{ 74, 30, 55 },
		{ 74, 31, 55 },
		{ 74, 32, 55 },
		{ 74, 33, 55 },
		{ 74, 34, 55 },
		{ 74, 35, 55 },
		{ 74, 36, 55 },
		{ 74, 37, 55 },
		{ 74, 38, 55 },
		{ 74, 39, 55 },
		{ 74, 40, 55 },
		{ 74, 41, 55 },
		{ 74, 42, 55 },
		{ 74, 43, 55 },
		{ 74, 44, 55 },
		{ 74, 45, 55 },
		{ 74, 46, 55 },
		{ 74, 47, 55 },
		{ 74, 48, 55 },
		{ 74, 49, 55 },
		{ 74, 50, 55 },
		{ 74, 51, 55 },
		{ 74, 52, 55 },
		{ 74, 53, 55 },
		{ 74, 54, 55 },
		{ 74, 55, 55 },
		{ 74, 56, 55 },
		{ 74, 57, 55 },
		{ 74, 58, 55 },
		{ 74, 59, 55 },
		{ 74, 60, 55 },
		{ 74, 61, 55 },
		{ 74, 62, 55 },
		{ 74, 63, 55 },
		{ 74, 64, 55 },
		{ 74, 65, 55 },
		{ 74, 66, 55 },
		{ 74, 67, 55 },
		{ 74, 68, 55 },
		{ 74, 69, 55 },
		{ 74, 70, 55 },
		{ 74, 71, 55 },
		{ 74, 72, 55 },
		{ 74, 73, 55 },
		{ 75, 15, 55 },
		{ 75, 16, 55 },
		{ 75, 17, 55 },
		{ 75, 18, 55 },
		{ 75, 19, 55 },
		{ 75, 20, 55 },
		{ 75, 21, 55 },
		{ 75, 22, 55 },
		{ 75, 23, 55 },
		{ 75, 24, 55 },
		{ 75, 25, 55 },
		{ 75, 26, 55 },
		{ 75, 27, 55 },
		{ 75, 28, 55 },
		{ 75, 29, 55 },
		{ 75, 30, 55 },
		{ 75, 31, 55 },
		{ 75, 32, 55 },
		{ 75, 33, 55 },
		{ 75, 34, 55 },
		{ 75, 35, 55 },
		{ 75, 36, 55 },
		{ 75, 37, 55 },
		{ 75, 38, 55 },
		{ 75, 39, 55 },
		{ 75, 40, 55 },
		{ 75, 41, 55 },
		{ 75, 42, 55 },
		{ 75, 43, 55 },
		{ 75, 44, 55 },
		{ 75, 45, 55 },
		{ 75, 46, 55 },
		{ 75, 47, 55 },
		{ 75, 48, 55 },
		{ 75, 49, 55 },
		{ 75, 50, 55 },
		{ 75, 51, 55 },
		{ 75, 52, 55 },
		{ 75, 53, 55 },
		{ 75, 54, 55 },
		{ 75, 55, 55 },
		{ 75, 56, 55 },
		{ 75, 57, 55 },
		{ 75, 58, 55 },
		{ 75, 59, 55 },
		{ 75, 60, 55 },
		{ 75, 61, 55 },
		{ 75, 62, 55 },
		{ 75, 63, 55 },
		{ 75, 64, 55 },
		{ 75, 65, 55 },
		{ 75, 66, 55 },
		{ 75, 67, 55 },
		{ 75, 68, 55 },
		{ 75, 69, 55 },
		{ 75, 70, 55 },
		{ 75, 71, 55 },
		{ 75, 72, 55 },
		{ 75, 73, 55 },
		{ 75, 74, 55 },
		{ 76, 16, 55 },
		{ 76, 17, 55 },
		{ 76, 18, 55 },
		{ 76, 19, 55 },
		{ 76, 20, 55 },
		{ 76, 21, 55 },
		{ 76, 22, 55 },
		{ 76, 23, 55 },
		{ 76, 24, 55 },
		{ 76, 25, 55 },
		{ 76, 26, 55 },
		{ 76, 27, 55 },
		{ 76, 28, 55 },
		{ 76, 29, 55 },
		{ 76, 30, 55 },
		{ 76, 31, 55 },
		{ 76, 32, 55 },
		{ 76, 33, 55 },
		{ 76, 34, 55 },
		{ 76, 35, 55 },
		{ 76, 36, 55 },
		{ 76, 37, 55 },
		{ 76, 38, 55 },
		{ 76, 39, 55 },
		{ 76, 40, 55 },
		{ 76, 41, 55 },
		{ 76, 42, 55 },
		{ 76, 43, 55 },
		{ 76, 44, 55 },
		{ 76, 45, 55 },
		{ 76, 46, 55 },
		{ 76, 47, 55 },
		{ 76, 48, 55 },
		{ 76, 49, 55 },
		{ 76, 50, 55 },
		{ 76, 51, 55 },
		{ 76, 52, 55 },
		{ 76, 53, 55 },
		{ 76, 54, 55 },
		{ 76, 55, 55 },
		{ 76, 56, 55 },
		{ 76, 57, 55 },
		{ 76, 58, 55 },
		{ 76, 59, 55 },
		{ 76, 60, 55 },
		{ 76, 61, 55 },
		{ 76, 62, 55 },
		{ 76, 63, 55 },
		{ 76, 64, 55 },
		{ 76, 65, 55 },
		{ 76, 66, 55 },
		{ 76, 67, 55 },
		{ 76, 68, 55 },
		{ 76, 69, 55 },
		{ 76, 70, 55 },
		{ 76, 71, 55 },
		{ 76, 72, 55 },
		{ 76, 73, 55 },
		{ 76, 74, 55 },
		{ 76, 75, 55 },
		{ 77, 18, 55 },
		{ 77, 19, 55 },
		{ 77, 20, 55 },
		{ 77, 21, 55 },
		{ 77, 22, 55 },
		{ 77, 23, 55 },
		{ 77, 24, 55 },
		{ 77, 25, 55 },
		{ 77, 26, 55 },
		{ 77, 27, 55 },
		{ 77, 28, 55 },
		{ 77, 29, 55 },
		{ 77, 30, 55 },
		{ 77, 31, 55 },
		{ 77, 32, 55 },
		{ 77, 33, 55 },
		{ 77, 34, 55 },
		{ 77, 35, 55 },
		{ 77, 36, 55 },
		{ 77, 37, 55 },
		{ 77, 38, 55 },
		{ 77, 39, 55 },
		{ 77, 40, 55 },
		{ 77, 41, 55 },
		{ 77, 42, 55 },
		{ 77, 43, 55 },
		{ 77, 44, 55 },
		{ 77, 45, 55 },
		{ 77, 46, 55 },
		{ 77, 47, 55 },
		{ 77, 48, 55 },
		{ 77, 49, 55 },
		{ 77, 50, 55 },
		{ 77, 51, 55 },
		{ 77, 52, 55 },
		{ 77, 53, 55 },
		{ 77, 54, 55 },
		{ 77, 55, 55 },
		{ 77, 56, 55 },
		{ 77, 57, 55 },
		{ 77, 58, 55 },
		{ 77, 59, 55 },
		{ 77, 60, 55 },
		{ 77, 61, 55 },
		{ 77, 62, 55 },
		{ 77, 63, 55 },
		{ 77, 64, 55 },
		{ 77, 65, 55 },
		{ 77, 66, 55 },
		{ 77, 67, 55 },
		{ 77, 68, 55 },
		{ 77, 69, 55 },
		{ 77, 70, 55 },
		{ 77, 71, 55 },
		{ 77, 72, 55 },
		{ 77, 73, 55 },
		{ 77, 74, 55 },
		{ 77, 75, 55 },
		{ 77, 76, 55 },
		{ 78, 19, 55 },
		{ 78, 20, 55 },
		{ 78, 21, 55 },
		{ 78, 22, 55 },
		{ 78, 23, 55 },
		{ 78, 24, 55 },
		{ 78, 25, 55 },
		{ 78, 26, 55 },
		{ 78, 27, 55 },
		{ 78, 28, 55 },
		{ 78, 29, 55 },
		{ 78, 30, 55 },
		{ 78, 31, 55 },
		{ 78, 32, 55 },
		{ 78, 33, 55 },
		{ 78, 34, 55 },
		{ 78, 35, 55 },
		{ 78, 36, 55 },
		{ 78, 37, 55 },
		{ 78, 38, 55 },
		{ 78, 39, 55 },
		{ 78, 40, 55 },
		{ 78, 41, 55 },
		{ 78, 42, 55 },
		{ 78, 43, 55 },
		{ 78, 44, 55 },
		{ 78, 45, 55 },
		{ 78, 46, 55 },
		{ 78, 47, 55 },
		{ 78, 48, 55 },
		{ 78, 49, 55 },
		{ 78, 50, 55 },
		{ 78, 51, 55 },
		{ 78, 52, 55 },
		{ 78, 53, 55 },
		{ 78, 54, 55 },
		{ 78, 55, 55 },
		{ 78, 56, 55 },
		{ 78, 57, 55 },
		{ 78, 58, 55 },
		{ 78, 59, 55 },
		{ 78, 60, 55 },
		{ 78, 61, 55 },
		{ 78, 62, 55 },
		{ 78, 63, 55 },
		{ 78, 64, 55 },
		{ 78, 65, 55 },
		{ 78, 66, 55 },
		{ 78, 67, 55 },
		{ 78, 68, 55 },
		{ 78, 69, 55 },
		{ 78, 70, 55 },
		{ 78, 71, 55 },
		{ 78, 72, 55 },
		{ 78, 73, 55 },
		{ 78, 74, 55 },
		{ 78, 75, 55 },
		{ 78, 76, 55 },
		{ 78, 77, 55 },
		{ 79, 20, 55 },
		{ 79, 21, 55 },
		{ 79, 22, 55 },
		{ 79, 23, 55 },
		{ 79, 24, 55 },
		{ 79, 25, 55 },
		{ 79, 26, 55 },
		{ 79, 27, 55 },
		{ 79, 28, 55 },
		{ 79, 29, 55 },
		{ 79, 30, 55 },
		{ 79, 31, 55 },
		{ 79, 32, 55 },
		{ 79, 33, 55 },
		{ 79, 34, 55 },
		{ 79, 35, 55 },
		{ 79, 36, 55 },
		{ 79, 37, 55 },
		{ 79, 38, 55 },
		{ 79, 39, 55 },
		{ 79, 40, 55 },
		{ 79, 41, 55 },
		{ 79, 42, 55 },
		{ 79, 43, 55 },
		{ 79, 44, 55 },
		{ 79, 45, 55 },
		{ 79, 46, 55 },
		{ 79, 47, 55 },
		{ 79, 48, 55 },
		{ 79, 49, 55 },
		{ 79, 50, 55 },
		{ 79, 51, 55 },
		{ 79, 52, 55 },
		{ 79, 53, 55 },
		{ 79, 54, 55 },
		{ 79, 55, 55 },
		{ 79, 56, 55 },
		{ 79, 57, 55 },
		{ 79, 58, 55 },
		{ 79, 59, 55 },
		{ 79, 60, 55 },
		{ 79, 61, 55 },
		{ 79, 62, 55 },
		{ 79, 63, 55 },
		{ 79, 64, 55 },
		{ 79, 65, 55 },
		{ 79, 66, 55 },
		{ 79, 67, 55 },
		{ 79, 68, 55 },
		{ 79, 69, 55 },
		{ 79, 70, 55 },
		{ 79, 71, 55 },
		{ 79, 72, 55 },
		{ 79, 73, 55 },
		{ 79, 74, 55 },
		{ 79, 75, 55 },
		{ 79, 76, 55 },
		{ 79, 77, 55 },
		{ 79, 78, 55 },
		{ 80, 22, 55 },
		{ 80, 23, 55 },
		{ 80, 24, 55 },
		{ 80, 25, 55 },
		{ 80, 26, 55 },
		{ 80, 27, 55 },
		{ 80, 28, 55 },
		{ 80, 29, 55 },
		{ 80, 30, 55 },
		{ 80, 31, 55 },
		{ 80, 32, 55 },
		{ 80, 33, 55 },
		{ 80, 34, 55 },
		{ 80, 35, 55 },
		{ 80, 36, 55 },
		{ 80, 37, 55 },
		{ 80, 38, 55 },
		{ 80, 39, 55 },
		{ 80, 40, 55 },
		{ 80, 41, 55 },
		{ 80, 42, 55 },
		{ 80, 43, 55 },
		{ 80, 44, 55 },
		{ 80, 45, 55 },
		{ 80, 46, 55 },
		{ 80, 47, 55 },
		{ 80, 48, 55 },
		{ 80, 49, 55 },
		{ 80, 50, 55 },
		{ 80, 51, 55 },
		{ 80, 52, 55 },
		{ 80, 53, 55 },
		{ 80, 54, 55 },
		{ 80, 55, 55 },
		{ 80, 56, 55 },
		{ 80, 57, 55 },
		{ 80, 58, 55 },
		{ 80, 59, 55 },
		{ 80, 60, 55 },
		{ 80, 61, 55 },
		{ 80, 62, 55 },
		{ 80, 63, 55 },
		{ 80, 64, 55 },
		{ 80, 65, 55 },
		{ 80, 66, 55 },
		{ 80, 67, 55 },
		{ 80, 68, 55 },
		{ 80, 69, 55 },
		{ 80, 70, 55 },
		{ 80, 71, 55 },
		{ 80, 72, 55 },
		{ 80, 73, 55 },
		{ 80, 74, 55 },
		{ 80, 75, 55 },
		{ 80, 76, 55 },
		{ 80, 77, 55 },
		{ 80, 78, 55 },
		{ 80, 79, 55 },
		{ 81, 23, 55 },
		{ 81, 24, 55 },
		{ 81, 25, 55 },
		{ 81, 26, 55 },
		{ 81, 27, 55 },
		{ 81, 28, 55 },
		{ 81, 29, 55 },
		{ 81, 30, 55 },
		{ 81, 31, 55 },
		{ 81, 32, 55 },
		{ 81, 33, 55 },
		{ 81, 34, 55 },
		{ 81, 35, 55 },
		{ 81, 36, 55 },
		{ 81, 37, 55 },
		{ 81, 38, 55 },
		{ 81, 39, 55 },
		{ 81, 40, 55 },
		{ 81, 41, 55 },
		{ 81, 42, 55 },
		{ 81, 43, 55 },
		{ 81, 44, 55 },
		{ 81, 45, 55 },
		{ 81, 46, 55 },
		{ 81, 47, 55 },
		{ 81, 48, 55 },
		{ 81, 49, 55 },
		{ 81, 50, 55 },
		{ 81, 51, 55 },
		{ 81, 52, 55 },
		{ 81, 53, 55 },
		{ 81, 54, 55 },
		{ 81, 55, 55 },
		{ 81, 56, 55 },
		{ 81, 57, 55 },
		{ 81, 58, 55 },
		{ 81, 59, 55 },
		{ 81, 60, 55 },
		{ 81, 61, 55 },
		{ 81, 62, 55 },
		{ 81, 63, 55 },
		{ 81, 64, 55 },
		{ 81, 65, 55 },
		{ 81, 66, 55 },
		{ 81, 67, 55 },
		{ 81, 68, 55 },
		{ 81, 69, 55 },
		{ 81, 70, 55 },
		{ 81, 71, 55 },
		{ 81, 72, 55 },
		{ 81, 73, 55 },
		{ 81, 74, 55 },
		{ 81, 75, 55 },
		{ 81, 76, 55 },
		{ 81, 77, 55 },
		{ 81, 78, 55 },
		{ 81, 79, 55 },
		{ 81, 80, 55 },
		{ 82, 24, 55 },
		{ 82, 25, 55 },
		{ 82, 26, 55 },
		{ 82, 27, 55 },
		{ 82, 28, 55 },
		{ 82, 29, 55 },
		{ 82, 30, 55 },
		{ 82, 31, 55 },
		{ 82, 32, 55 },
		{ 82, 33, 55 },
		{ 82, 34, 55 },
		{ 82, 35, 55 },
		{ 82, 36, 55 },
		{ 82, 37, 55 },
		{ 82, 38, 55 },
		{ 82, 39, 55 },
		{ 82, 40, 55 },
		{ 82, 41, 55 },
		{ 82, 42, 55 },
		{ 82, 43, 55 },
		{ 82, 44, 55 },
		{ 82, 45, 55 },
		{ 82, 46, 55 },
		{ 82, 47, 55 },
		{ 82, 48, 55 },
		{ 82, 49, 55 },
		{ 82, 50, 55 },
		{ 82, 51, 55 },
		{ 82, 52, 55 },
		{ 82, 53, 55 },
		{ 82, 54, 55 },
		{ 82, 55, 55 },
		{ 82, 56, 55 },
		{ 82, 57, 55 },
		{ 82, 58, 55 },
		{ 82, 59, 55 },
		{ 82, 60, 55 },
		{ 82, 61, 55 },
		{ 82, 62, 55 },
		{ 82, 63, 55 },
		{ 82, 64, 55 },
		{ 82, 65, 55 },
		{ 82, 66, 55 },
		{ 82, 67, 55 },
		{ 82, 68, 55 },
		{ 82, 69, 55 },
		{ 82, 70, 55 },
		{ 82, 71, 55 },
		{ 82, 72, 55 },
		{ 82, 73, 55 },
		{ 82, 74, 55 },
		{ 82, 75, 55 },
		{ 82, 76, 55 },
		{ 82, 77, 55 },
		{ 82, 78, 55 },
		{ 82, 79, 55 },
		{ 82, 80, 55 },
		{ 82, 81, 55 },
		{ 83, 26, 55 },
		{ 83, 27, 55 },
		{ 83, 28, 55 },
		{ 83, 29, 55 },
		{ 83, 30, 55 },
		{ 83, 31, 55 },
		{ 83, 32, 55 },
		{ 83, 33, 55 },
		{ 83, 34, 55 },
		{ 83, 35, 55 },
		{ 83, 36, 55 },
		{ 83, 37, 55 },
		{ 83, 38, 55 },
		{ 83, 39, 55 },
		{ 83, 40, 55 },
		{ 83, 41, 55 },
		{ 83, 42, 55 },
		{ 83, 43, 55 },
		{ 83, 44, 55 },
		{ 83, 45, 55 },
		{ 83, 46, 55 },
		{ 83, 47, 55 },
		{ 83, 48, 55 },
		{ 83, 49, 55 },
		{ 83, 50, 55 },
		{ 83, 51, 55 },
		{ 83, 52, 55 },
		{ 83, 53, 55 },
		{ 83, 54, 55 },
		{ 83, 55, 55 },
		{ 83, 56, 55 },
		{ 83, 57, 55 },
		{ 83, 58, 55 },
		{ 83, 59, 55 },
		{ 83, 60, 55 },
		{ 83, 61, 55 },
		{ 83, 62, 55 },
		{ 83, 63, 55 },
		{ 83, 64, 55 },
		{ 83, 65, 55 },
		{ 83, 66, 55 },
		{ 83, 67, 55 },
		{ 83, 68, 55 },
		{ 83, 69, 55 },
		{ 83, 70, 55 },
		{ 83, 71, 55 },
		{ 83, 72, 55 },
		{ 83, 73, 55 },
		{ 83, 74, 55 },
		{ 83, 75, 55 },
		{ 83, 76, 55 },
		{ 83, 77, 55 },
		{ 83, 78, 55 },
		{ 83, 79, 55 },
		{ 83, 80, 55 },
		{ 83, 81, 55 },
		{ 83, 82, 55 },
		{ 84, 27, 55 },
		{ 84, 28, 55 },
		{ 84, 29, 55 },
		{ 84, 30, 55 },
		{ 84, 31, 55 },
		{ 84, 32, 55 },
		{ 84, 33, 55 },
		{ 84, 34, 55 },
		{ 84, 35, 55 },
		{ 84, 36, 55 },
		{ 84, 37, 55 },
		{ 84, 38, 55 },
		{ 84, 39, 55 },
		{ 84, 40, 55 },
		{ 84, 41, 55 },
		{ 84, 42, 55 },
		{ 84, 43, 55 },
		{ 84, 44, 55 },
		{ 84, 45, 55 },
		{ 84, 46, 55 },
		{ 84, 47, 55 },
		{ 84, 48, 55 },
		{ 84, 49, 55 },
		{ 84, 50, 55 },
		{ 84, 51, 55 },
		{ 84, 52, 55 },
		{ 84, 53, 55 },
		{ 84, 54, 55 },
		{ 84, 55, 55 },
		{ 84, 56, 55 },
		{ 84, 57, 55 },
		{ 84, 58, 55 },
		{ 84, 59, 55 },
		{ 84, 60, 55 },
		{ 84, 61, 55 },
		{ 84, 62, 55 },
		{ 84, 63, 55 },
		{ 84, 64, 55 },
		{ 84, 65, 55 },
		{ 84, 66, 55 },
		{ 84, 67, 55 },
		{ 84, 68, 55 },
		{ 84, 69, 55 },
		{ 84, 70, 55 },
		{ 84, 71, 55 },
		{ 84, 72, 55 },
		{ 84, 73, 55 },
		{ 84, 74, 55 },
		{ 84, 75, 55 },
		{ 84, 76, 55 },
		{ 84, 77, 55 },
		{ 84, 78, 55 },
		{ 84, 79, 55 },
		{ 84, 80, 55 },
		{ 84, 81, 55 },
		{ 84, 82, 55 },
		{ 84, 83, 55 },
		{ 85, 28, 55 },
		{ 85, 29, 55 },
		{ 85, 30, 55 },
		{ 85, 31, 55 },
		{ 85, 32, 55 },
		{ 85, 33, 55 },
		{ 85, 34, 55 },
		{ 85, 35, 55 },
		{ 85, 36, 55 },
		{ 85, 37, 55 },
		{ 85, 38, 55 },
		{ 85, 39, 55 },
		{ 85, 40, 55 },
		{ 85, 41, 55 },
		{ 85, 42, 55 },
		{ 85, 43, 55 },
		{ 85, 44, 55 },
		{ 85, 45, 55 },
		{ 85, 46, 55 },
		{ 85, 47, 55 },
		{ 85, 48, 55 },
		{ 85, 49, 55 },
		{ 85, 50, 55 },
		{ 85, 51, 55 },
		{ 85, 52, 55 },
		{ 85, 53, 55 },
		{ 85, 54, 55 },
		{ 85, 55, 55 },
		{ 85, 56, 55 },
		{ 85, 57, 55 },
		{ 85, 58, 55 },
		{ 85, 59, 55 },
		{ 85, 60, 55 },
		{ 85, 61, 55 },
		{ 85, 62, 55 },
		{ 85, 63, 55 },
		{ 85, 64, 55 },
		{ 85, 65, 55 },
		{ 85, 66, 55 },
		{ 85, 67, 55 },
		{ 85, 68, 55 },
		{ 85, 69, 55 },
		{ 85, 70, 55 },
		{ 85, 71, 55 },
		{ 85, 72, 55 },
		{ 85, 73, 55 },
		{ 85, 74, 55 },
		{ 85, 75, 55 },
		{ 85, 76, 55 },
		{ 85, 77, 55 },
		{ 85, 78, 55 },
		{ 85, 79, 55 },
		{ 85, 80, 55 },
		{ 85, 81, 55 },
		{ 85, 82, 55 },
		{ 85, 83, 55 },
		{ 85, 84, 55 },
		{ 86, 30, 55 },
		{ 86, 31, 55 },
		{ 86, 32, 55 },
		{ 86, 33, 55 },
		{ 86, 34, 55 },
		{ 86, 35, 55 },
		{ 86, 36, 55 },
		{ 86, 37, 55 },
		{ 86, 38, 55 },
		{ 86, 39, 55 },
		{ 86, 40, 55 },
		{ 86, 41, 55 },
		{ 86, 42, 55 },
		{ 86, 43, 55 },
		{ 86, 44, 55 },
		{ 86, 45, 55 },
		{ 86, 46, 55 },
		{ 86, 47, 55 },
		{ 86, 48, 55 },
		{ 86, 49, 55 },
		{ 86, 50, 55 },
		{ 86, 51, 55 },
		{ 86, 52, 55 },
		{ 86, 53, 55 },
		{ 86, 54, 55 },
		{ 86, 55, 55 },
		{ 86, 56, 55 },
		{ 86, 57, 55 },
		{ 86, 58, 55 },
		{ 86, 59, 55 },
		{ 86, 60, 55 },
		{ 86, 61, 55 },
		{ 86, 62, 55 },
		{ 86, 63, 55 },
		{ 86, 64, 55 },
		{ 86, 65, 55 },
		{ 86, 66, 55 },
		{ 86, 67, 55 },
		{ 86, 68, 55 },
		{ 86, 69, 55 },
		{ 86, 70, 55 },
		{ 86, 71, 55 },
		{ 86, 72, 55 },
		{ 86, 73, 55 },
		{ 86, 74, 55 },
		{ 86, 75, 55 },
		{ 86, 76, 55 },
		{ 86, 77, 55 },
		{ 86, 78, 55 },
		{ 86, 79, 55 },
		{ 86, 80, 55 },
		{ 86, 81, 55 },
		{ 86, 82, 55 },
		{ 86, 83, 55 },
		{ 86, 84, 55 },
		{ 86, 85, 55 },
		{ 87, 31, 55 },
		{ 87, 32, 55 },
		{ 87, 33, 55 },
		{ 87, 34, 55 },
		{ 87, 35, 55 },
		{ 87, 36, 55 },
		{ 87, 37, 55 },
		{ 87, 38, 55 },
		{ 87, 39, 55 },
		{ 87, 40, 55 },
		{ 87, 41, 55 },
		{ 87, 42, 55 },
		{ 87, 43, 55 },
		{ 87, 44, 55 },
		{ 87, 45, 55 },
		{ 87, 46, 55 },
		{ 87, 47, 55 },
		{ 87, 48, 55 },
		{ 87, 49, 55 },
		{ 87, 50, 55 },
		{ 87, 51, 55 },
		{ 87, 52, 55 },
		{ 87, 53, 55 },
		{ 87, 54, 55 },
		{ 87, 55, 55 },
		{ 87, 56, 55 },
		{ 87, 57, 55 },
		{ 87, 58, 55 },
		{ 87, 59, 55 },
		{ 87, 60, 55 },
		{ 87, 61, 55 },
		{ 87, 62, 55 },
		{ 87, 63, 55 },
		{ 87, 64, 55 },
		{ 87, 65, 55 },
		{ 87, 66, 55 },
		{ 87, 67, 55 },
		{ 87, 68, 55 },
		{ 87, 69, 55 },
		{ 87, 70, 55 },
		{ 87, 71, 55 },
		{ 87, 72, 55 },
		{ 87, 73, 55 },
		{ 87, 74, 55 },
		{ 87, 75, 55 },
		{ 87, 76, 55 },
		{ 87, 77, 55 },
		{ 87, 78, 55 },
		{ 87, 79, 55 },
		{ 87, 80, 55 },
		{ 87, 81, 55 },
		{ 87, 82, 55 },
		{ 87, 83, 55 },
		{ 87, 84, 55 },
		{ 87, 85, 55 },
		{ 87, 86, 55 },
		{ 88, 32, 55 },
		{ 88, 33, 55 },
		{ 88, 34, 55 },
		{ 88, 35, 55 },
		{ 88, 36, 55 },
		{ 88, 37, 55 },
		{ 88, 38, 55 },
		{ 88, 39, 55 },
		{ 88, 40, 55 },
		{ 88, 41, 55 },
		{ 88, 42, 55 },
		{ 88, 43, 55 },
		{ 88, 44, 55 },
		{ 88, 45, 55 },
		{ 88, 46, 55 },
		{ 88, 47, 55 },
		{ 88, 48, 55 },
		{ 88, 49, 55 },
		{ 88, 50, 55 },
		{ 88, 51, 55 },
		{ 88, 52, 55 },
		{ 88, 53, 55 },
		{ 88, 54, 55 },
		{ 88, 55, 55 },
		{ 88, 56, 55 },
		{ 88, 57, 55 },
		{ 88, 58, 55 },
		{ 88, 59, 55 },
		{ 88, 60, 55 },
		{ 88, 61, 55 },
		{ 88, 62, 55 },
		{ 88, 63, 55 },
		{ 88, 64, 55 },
		{ 88, 65, 55 },
		{ 88, 66, 55 },
		{ 88, 67, 55 },
		{ 88, 68, 55 },
		{ 88, 69, 55 },
		{ 88, 70, 55 },
		{ 88, 71, 55 },
		{ 88, 72, 55 },
		{ 88, 73, 55 },
		{ 88, 74, 55 },
		{ 88, 75, 55 },
		{ 88, 76, 55 },
		{ 88, 77, 55 },
		{ 88, 78, 55 },
		{ 88, 79, 55 },
		{ 88, 80, 55 },
		{ 88, 81, 55 },
		{ 88, 82, 55 },
		{ 88, 83, 55 },
		{ 88, 84, 55 },
		{ 88, 85, 55 },
		{ 88, 86, 55 },
		{ 88, 87, 55 },
		{ 89, 34, 55 },
		{ 89, 35, 55 },
		{ 89, 36, 55 },
		{ 89, 37, 55 },
		{ 89, 38, 55 },
		{ 89, 39, 55 },
		{ 89, 40, 55 },
		{ 89, 41, 55 },
		{ 89, 42, 55 },
		{ 89, 43, 55 },
		{ 89, 44, 55 },
		{ 89, 45, 55 },
		{ 89, 46, 55 },
		{ 89, 47, 55 },
		{ 89, 48, 55 },
		{ 89, 49, 55 },
		{ 89, 50, 55 },
		{ 89, 51, 55 },
		{ 89, 52, 55 },
		{ 89, 53, 55 },
		{ 89, 54, 55 },
		{ 89, 55, 55 },
		{ 89, 56, 55 },
		{ 89, 57, 55 },
		{ 89, 58, 55 },
		{ 89, 59, 55 },
		{ 89, 60, 55 },
		{ 89, 61, 55 },
		{ 89, 62, 55 },
		{ 89, 63, 55 },
		{ 89, 64, 55 },
		{ 89, 65, 55 },
		{ 89, 66, 55 },
		{ 89, 67, 55 },
		{ 89, 68, 55 },
		{ 89, 69, 55 },
		{ 89, 70, 55 },
		{ 89, 71, 55 },
		{ 89, 72, 55 },
		{ 89, 73, 55 },
		{ 89, 74, 55 },
		{ 89, 75, 55 },
		{ 89, 76, 55 },
		{ 89, 77, 55 },
		{ 89, 78, 55 },
		{ 89, 79, 55 },
		{ 89, 80, 55 },
		{ 89, 81, 55 },
		{ 89, 82, 55 },
		{ 89, 83, 55 },
		{ 89, 84, 55 },
		{ 89, 85, 55 },
		{ 89, 86, 55 },
		{ 89, 87, 55 },
		{ 89, 88, 55 },
		{ 90, 35, 55 },
		{ 90, 36, 55 },
		{ 90, 37, 55 },
		{ 90, 38, 55 },
		{ 90, 39, 55 },
		{ 90, 40, 55 },
		{ 90, 41, 55 },
		{ 90, 42, 55 },
		{ 90, 43, 55 },
		{ 90, 44, 55 },
		{ 90, 45, 55 },
		{ 90, 46, 55 },
		{ 90, 47, 55 },
		{ 90, 48, 55 },
		{ 90, 49, 55 },
		{ 90, 50, 55 },
		{ 90, 51, 55 },
		{ 90, 52, 55 },
		{ 90, 53, 55 },
		{ 90, 54, 55 },
		{ 90, 55, 55 },
		{ 90, 56, 55 },
		{ 90, 57, 55 },
		{ 90, 58, 55 },
		{ 90, 59, 55 },
		{ 90, 60, 55 },
		{ 90, 61, 55 },
		{ 90, 62, 55 },
		{ 90, 63, 55 },
		{ 90, 64, 55 },
		{ 90, 65, 55 },
		{ 90, 66, 55 },
		{ 90, 67, 55 },
		{ 90, 68, 55 },
		{ 90, 69, 55 },
		{ 90, 70, 55 },
		{ 90, 71, 55 },
		{ 90, 72, 55 },
		{ 90, 73, 55 },
		{ 90, 74, 55 },
		{ 90, 75, 55 },
		{ 90, 76, 55 },
		{ 90, 77, 55 },
		{ 90, 78, 55 },
		{ 90, 79, 55 },
		{ 90, 80, 55 },
		{ 90, 81, 55 },
		{ 90, 82, 55 },
		{ 90, 83, 55 },
		{ 90, 84, 55 },
		{ 90, 85, 55 },
		{ 90, 86, 55 },
		{ 90, 87, 55 },
		{ 90, 88, 55 },
		{ 90, 89, 55 },
		{ 91, 36, 55 },
		{ 91, 37, 55 },
		{ 91, 38, 55 },
		{ 91, 39, 55 },
		{ 91, 40, 55 },
		{ 91, 41, 55 },
		{ 91, 42, 55 },
		{ 91, 43, 55 },
		{ 91, 44, 55 },
		{ 91, 45, 55 },
		{ 91, 46, 55 },
		{ 91, 47, 55 },
		{ 91, 48, 55 },
		{ 91, 49, 55 },
		{ 91, 50, 55 },
		{ 91, 51, 55 },
		{ 91, 52, 55 },
		{ 91, 53, 55 },
		{ 91, 54, 55 },
		{ 91, 55, 55 },
		{ 91, 56, 55 },
		{ 91, 57, 55 },
		{ 91, 58, 55 },
		{ 91, 59, 55 },
		{ 91, 60, 55 },
		{ 91, 61, 55 },
		{ 91, 62, 55 },
		{ 91, 63, 55 },
		{ 91, 64, 55 },
		{ 91, 65, 55 },
		{ 91, 66, 55 },
		{ 91, 67, 55 },
		{ 91, 68, 55 },
		{ 91, 69, 55 },
		{ 91, 70, 55 },
		{ 91, 71, 55 },
		{ 91, 72, 55 },
		{ 91, 73, 55 },
		{ 91, 74, 55 },
		{ 91, 75, 55 },
		{ 91, 76, 55 },
		{ 91, 77, 55 },
		{ 91, 78, 55 },
		{ 91, 79, 55 },
		{ 91, 80, 55 },
		{ 91, 81, 55 },
		{ 91, 82, 55 },
		{ 91, 83, 55 },
		{ 91, 84, 55 },
		{ 91, 85, 55 },
		{ 91, 86, 55 },
		{ 91, 87, 55 },
		{ 91, 88, 55 },
		{ 91, 89, 55 },
		{ 91, 90, 55 },
		{ 92, 38, 55 },
		{ 92, 39, 55 },
		{ 92, 40, 55 },
		{ 92, 41, 55 },
		{ 92, 42, 55 },
		{ 92, 43, 55 },
		{ 92, 44, 55 },
		{ 92, 45, 55 },
		{ 92, 46, 55 },
		{ 92, 47, 55 },
		{ 92, 48, 55 },
		{ 92, 49, 55 },
		{ 92, 50, 55 },
		{ 92, 51, 55 },
		{ 92, 52, 55 },
		{ 92, 53, 55 },
		{ 92, 54, 55 },
		{ 92, 55, 55 },
		{ 92, 56, 55 },
		{ 92, 57, 55 },
		{ 92, 58, 55 },
		{ 92, 59, 55 },
		{ 92, 60, 55 },
		{ 92, 61, 55 },
		{ 92, 62, 55 },
		{ 92, 63, 55 },
		{ 92, 64, 55 },
		{ 92, 65, 55 },
		{ 92, 66, 55 },
		{ 92, 67, 55 },
		{ 92, 68, 55 },
		{ 92, 69, 55 },
		{ 92, 70, 55 },
		{ 92, 71, 55 },
		{ 92, 72, 55 },
		{ 92, 73, 55 },
		{ 92, 74, 55 },
		{ 92, 75, 55 },
		{ 92, 76, 55 },
		{ 92, 77, 55 },
		{ 92, 78, 55 },
		{ 92, 79, 55 },
		{ 92, 80, 55 },
		{ 92, 81, 55 },
		{ 92, 82, 55 },
		{ 92, 83, 55 },
		{ 92, 84, 55 },
		{ 92, 85, 55 },
		{ 92, 86, 55 },
		{ 92, 87, 55 },
		{ 92, 88, 55 },
		{ 92, 89, 55 },
		{ 92, 90, 55 },
		{ 92, 91, 55 },
		{ 93, 39, 55 },
		{ 93, 40, 55 },
		{ 93, 41, 55 },
		{ 93, 42, 55 },
		{ 93, 43, 55 },
		{ 93, 44, 55 },
		{ 93, 45, 55 },
		{ 93, 46, 55 },
		{ 93, 47, 55 },
		{ 93, 48, 55 },
		{ 93, 49, 55 },
		{ 93, 50, 55 },
		{ 93, 51, 55 },
		{ 93, 52, 55 },
		{ 93, 53, 55 },
		{ 93, 54, 55 },
		{ 93, 55, 55 },
		{ 93, 56, 55 },
		{ 93, 57, 55 },
		{ 93, 58, 55 },
		{ 93, 59, 55 },
		{ 93, 60, 55 },
		{ 93, 61, 55 },
		{ 93, 62, 55 },
		{ 93, 63, 55 },
		{ 93, 64, 55 },
		{ 93, 65, 55 },
		{ 93, 66, 55 },
		{ 93, 67, 55 },
		{ 93, 68, 55 },
		{ 93, 69, 55 },
		{ 93, 70, 55 },
		{ 93, 71, 55 },
		{ 93, 72, 55 },
		{ 93, 73, 55 },
		{ 93, 74, 55 },
		{ 93, 75, 55 },
		{ 93, 76, 55 },
		{ 93, 77, 55 },
		{ 93, 78, 55 },
		{ 93, 79, 55 },
		{ 93, 80, 55 },
		{ 93, 81, 55 },
		{ 93, 82, 55 },
		{ 93, 83, 55 },
		{ 93, 84, 55 },
		{ 93, 85, 55 },
		{ 93, 86, 55 },
		{ 93, 87, 55 },
		{ 93, 88, 55 },
		{ 93, 89, 55 },
		{ 93, 90, 55 },
		{ 93, 91, 55 },
		{ 93, 92, 55 },
		{ 94, 40, 55 },
		{ 94, 41, 55 },
		{ 94, 42, 55 },
		{ 94, 43, 55 },
		{ 94, 44, 55 },
		{ 94, 45, 55 },
		{ 94, 46, 55 },
		{ 94, 47, 55 },
		{ 94, 48, 55 },
		{ 94, 49, 55 },
		{ 94, 50, 55 },
		{ 94, 51, 55 },
		{ 94, 52, 55 },
		{ 94, 53, 55 },
		{ 94, 54, 55 },
		{ 94, 55, 55 },
		{ 94, 56, 55 },
		{ 94, 57, 55 },
		{ 94, 58, 55 },
		{ 94, 59, 55 },
		{ 94, 60, 55 },
		{ 94, 61, 55 },
		{ 94, 62, 55 },
		{ 94, 63, 55 },
		{ 94, 64, 55 },
		{ 94, 65, 55 },
		{ 94, 66, 55 },
		{ 94, 67, 55 },
		{ 94, 68, 55 },
		{ 94, 69, 55 },
		{ 94, 70, 55 },
		{ 94, 71, 55 },
		{ 94, 72, 55 },
		{ 94, 73, 55 },
		{ 94, 74, 55 },
		{ 94, 75, 55 },
		{ 94, 76, 55 },
		{ 94, 77, 55 },
		{ 94, 78, 55 },
		{ 94, 79, 55 },
		{ 94, 80, 55 },
		{ 94, 81, 55 },
		{ 94, 82, 55 },
		{ 94, 83, 55 },
		{ 94, 84, 55 },
		{ 94, 85, 55 },
		{ 94, 86, 55 },
		{ 94, 87, 55 },
		{ 94, 88, 55 },
		{ 94, 89, 55 },
		{ 94, 90, 55 },
		{ 94, 91, 55 },
		{ 94, 92, 55 },
		{ 94, 93, 55 },
		{ 95, 42, 55 },
		{ 95, 43, 55 },
		{ 95, 44, 55 },
		{ 95, 45, 55 },
		{ 95, 46, 55 },
		{ 95, 47, 55 },
		{ 95, 48, 55 },
		{ 95, 49, 55 },
		{ 95, 50, 55 },
		{ 95, 51, 55 },
		{ 95, 52, 55 },
		{ 95, 53, 55 },
		{ 95, 54, 55 },
		{ 95, 55, 55 },
		{ 95, 56, 55 },
		{ 95, 57, 55 },
		{ 95, 58, 55 },
		{ 95, 59, 55 },
		{ 95, 60, 55 },
		{ 95, 61, 55 },
		{ 95, 62, 55 },
		{ 95, 63, 55 },
		{ 95, 64, 55 },
		{ 95, 65, 55 },
		{ 95, 66, 55 },
		{ 95, 67, 55 },
		{ 95, 68, 55 },
		{ 95, 69, 55 },
		{ 95, 70, 55 },
		{ 95, 71, 55 },
		{ 95, 72, 55 },
		{ 95, 73, 55 },
		{ 95, 74, 55 },
		{ 95, 75, 55 },
		{ 95, 76, 55 },
		{ 95, 77, 55 },
		{ 95, 78, 55 },
		{ 95, 79, 55 },
		{ 95, 80, 55 },
		{ 95, 81, 55 },
		{ 95, 82, 55 },
		{ 95, 83, 55 },
		{ 95, 84, 55 },
		{ 95, 85, 55 },
		{ 95, 86, 55 },
		{ 95, 87, 55 },
		{ 95, 88, 55 },
		{ 95, 89, 55 },
		{ 95, 90, 55 },
		{ 95, 91, 55 },
		{ 95, 92, 55 },
		{ 95, 93, 55 },
		{ 95, 94, 55 },
		{ 96, 43, 55 },
		{ 96, 44, 55 },
		{ 96, 45, 55 },
		{ 96, 46, 55 },
		{ 96, 47, 55 },
		{ 96, 48, 55 },
		{ 96, 49, 55 },
		{ 96, 50, 55 },
		{ 96, 51, 55 },
		{ 96, 52, 55 },
		{ 96, 53, 55 },
		{ 96, 54, 55 },
		{ 96, 55, 55 },
		{ 96, 56, 55 },
		{ 96, 57, 55 },
		{ 96, 58, 55 },
		{ 96, 59, 55 },
		{ 96, 60, 55 },
		{ 96, 61, 55 },
		{ 96, 62, 55 },
		{ 96, 63, 55 },
		{ 96, 64, 55 },
		{ 96, 65, 55 },
		{ 96, 66, 55 },
		{ 96, 67, 55 },
		{ 96, 68, 55 },
		{ 96, 69, 55 },
		{ 96, 70, 55 },
		{ 96, 71, 55 },
		{ 96, 72, 55 },
		{ 96, 73, 55 },
		{ 96, 74, 55 },
		{ 96, 75, 55 },
		{ 96, 76, 55 },
		{ 96, 77, 55 },
		{ 96, 78, 55 },
		{ 96, 79, 55 },
		{ 96, 80, 55 },
		{ 96, 81, 55 },
		{ 96, 82, 55 },
		{ 96, 83, 55 },
		{ 96, 84, 55 },
		{ 96, 85, 55 },
		{ 96, 86, 55 },
		{ 96, 87, 55 },
		{ 96, 88, 55 },
		{ 96, 89, 55 },
		{ 96, 90, 55 },
		{ 96, 91, 55 },
		{ 96, 92, 55 },
		{ 96, 93, 55 },
		{ 96, 94, 55 },
		{ 96, 95, 55 },
		{ 97, 44, 55 },
		{ 97, 45, 55 },
		{ 97, 46, 55 },
		{ 97, 47, 55 },
		{ 97, 48, 55 },
		{ 97, 49, 55 },
		{ 97, 50, 55 },
		{ 97, 51, 55 },
		{ 97, 52, 55 },
		{ 97, 53, 55 },
		{ 97, 54, 55 },
		{ 97, 55, 55 },
		{ 97, 56, 55 },
		{ 97, 57, 55 },
		{ 97, 58, 55 },
		{ 97, 59, 55 },
		{ 97, 60, 55 },
		{ 97, 61, 55 },
		{ 97, 62, 55 },
		{ 97, 63, 55 },
		{ 97, 64, 55 },
		{ 97, 65, 55 },
		{ 97, 66, 55 },
		{ 97, 67, 55 },
		{ 97, 68, 55 },
		{ 97, 69, 55 },
		{ 97, 70, 55 },
		{ 97, 71, 55 },
		{ 97, 72, 55 },
		{ 97, 73, 55 },
		{ 97, 74, 55 },
		{ 97, 75, 55 },
		{ 97, 76, 55 },
		{ 97, 77, 55 },
		{ 97, 78, 55 },
		{ 97, 79, 55 },
		{ 97, 80, 55 },
		{ 97, 81, 55 },
		{ 97, 82, 55 },
		{ 97, 83, 55 },
		{ 97, 84, 55 },
		{ 97, 85, 55 },
		{ 97, 86, 55 },
		{ 97, 87, 55 },
		{ 97, 88, 55 },
		{ 97, 89, 55 },
		{ 97, 90, 55 },
		{ 97, 91, 55 },
		{ 97, 92, 55 },
		{ 97, 93, 55 },
		{ 97, 94, 55 },
		{ 97, 95, 55 },
		{ 97, 96, 55 },
		{ 98, 46, 55 },
		{ 98, 47, 55 },
		{ 98, 48, 55 },
		{ 98, 49, 55 },
		{ 98, 50, 55 },
		{ 98, 51, 55 },
		{ 98, 52, 55 },
		{ 98, 53, 55 },
		{ 98, 54, 55 },
		{ 98, 55, 55 },
		{ 98, 56, 55 },
		{ 98, 57, 55 },
		{ 98, 58, 55 },
		{ 98, 59, 55 },
		{ 98, 60, 55 },
		{ 98, 61, 55 },
		{ 98, 62, 55 },
		{ 98, 63, 55 },
		{ 98, 64, 55 },
		{ 98, 65, 55 },
		{ 98, 66, 55 },
		{ 98, 67, 55 },
		{ 98, 68, 55 },
		{ 98, 69, 55 },
		{ 98, 70, 55 },
		{ 98, 71, 55 },
		{ 98, 72, 55 },
		{ 98, 73, 55 },
		{ 98, 74, 55 },
		{ 98, 75, 55 },
		{ 98, 76, 55 },
		{ 98, 77, 55 },
		{ 98, 78, 55 },
		{ 98, 79, 55 },
		{ 98, 80, 55 },
		{ 98, 81, 55 },
		{ 98, 82, 55 },
		{ 98, 83, 55 },
		{ 98, 84, 55 },
		{ 98, 85, 55 },
		{ 98, 86, 55 },
		{ 98, 87, 55 },
		{ 98, 88, 55 },
		{ 98, 89, 55 },
		{ 98, 90, 55 },
		{ 98, 91, 55 },
		{ 98, 92, 55 },
		{ 98, 93, 55 },
		{ 98, 94, 55 },
		{ 98, 95, 55 },
		{ 98, 96, 55 },
		{ 98, 97, 55 },
		{ 99, 47, 55 },
		{ 99, 48, 55 },
		{ 99, 49, 55 },
		{ 99, 50, 55 },
		{ 99, 51, 55 },
		{ 99, 52, 55 },
		{ 99, 53, 55 },
		{ 99, 54, 55 },
		{ 99, 55, 55 },
		{ 99, 56, 55 },
		{ 99, 57, 55 },
		{ 99, 58, 55 },
		{ 99, 59, 55 },
		{ 99, 60, 55 },
		{ 99, 61, 55 },
		{ 99, 62, 55 },
		{ 99, 63, 55 },
		{ 99, 64, 55 },
		{ 99, 65, 55 },
		{ 99, 66, 55 },
		{ 99, 67, 55 },
		{ 99, 68, 55 },
		{ 99, 69, 55 },
		{ 99, 70, 55 },
		{ 99, 71, 55 },
		{ 99, 72, 55 },
		{ 99, 73, 55 },
		{ 99, 74, 55 },
		{ 99, 75, 55 },
		{ 99, 76, 55 },
		{ 99, 77, 55 },
		{ 99, 78, 55 },
		{ 99, 79, 55 },
		{ 99, 80, 55 },
		{ 99, 81, 55 },
		{ 99, 82, 55 },
		{ 99, 83, 55 },
		{ 99, 84, 55 },
		{ 99, 85, 55 },
		{ 99, 86, 55 },
		{ 99, 87, 55 },
		{ 99, 88, 55 },
		{ 99, 89, 55 },
		{ 99, 90, 55 },
		{ 99, 91, 55 },
		{ 99, 92, 55 },
		{ 99, 93, 55 },
		{ 99, 94, 55 },
		{ 99, 95, 55 },
		{ 99, 96, 55 },
		{ 99, 97, 55 },
		{ 99, 98, 55 },
		{ 100, 48, 55 },
		{ 100, 49, 55 },
		{ 100, 50, 55 },
		{ 100, 51, 55 },
		{ 100, 52, 55 },
		{ 100, 53, 55 },
		{ 100, 54, 55 },
		{ 100, 55, 55 },
		{ 100, 56, 55 },
		{ 100, 57, 55 },
		{ 100, 58, 55 },
		{ 100, 59, 55 },
		{ 100, 60, 55 },
		{ 100, 61, 55 },
		{ 100, 62, 55 },
		{ 100, 63, 55 },
		{ 100, 64, 55 },
		{ 100, 65, 55 },
		{ 100, 66, 55 },
		{ 100, 67, 55 },
		{ 100, 68, 55 },
		{ 100, 69, 55 },
		{ 100, 70, 55 },
		{ 100, 71, 55 },
		{ 100, 72, 55 },
		{ 100, 73, 55 },
		{ 100, 74, 55 },
		{ 100, 75, 55 },
		{ 100, 76, 55 },
		{ 100, 77, 55 },
		{ 100, 78, 55 },
		{ 100, 79, 55 },
		{ 100, 80, 55 },
		{ 100, 81, 55 },
		{ 100, 82, 55 },
		{ 100, 83, 55 },
		{ 100, 84, 55 },
		{ 100, 85, 55 },
		{ 100, 86, 55 },
		{ 100, 87, 55 },
		{ 100, 88, 55 },
		{ 100, 89, 55 },
		{ 100, 90, 55 },
		{ 100, 91, 55 },
		{ 100, 92, 55 },
		{ 100, 93, 55 },
		{ 100, 94, 55 },
		{ 100, 95, 55 },
		{ 100, 96, 55 },
		{ 100, 97, 55 },
		{ 100, 98, 55 },
		{ 100, 99, 55 },
		{ 101, 50, 55 },
		{ 101, 51, 55 },
		{ 101, 52, 55 },
		{ 101, 53, 55 },
		{ 101, 54, 55 },
		{ 101, 55, 55 },
		{ 101, 56, 55 },
		{ 101, 57, 55 },
		{ 101, 58, 55 },
		{ 101, 59, 55 },
		{ 101, 60, 55 },
		{ 101, 61, 55 },
		{ 101, 62, 55 },
		{ 101, 63, 55 },
		{ 101, 64, 55 },
		{ 101, 65, 55 },
		{ 101, 66, 55 },
		{ 101, 67, 55 },
		{ 101, 68, 55 },
		{ 101, 69, 55 },
		{ 101, 70, 55 },
		{ 101, 71, 55 },
		{ 101, 72, 55 },
		{ 101, 73, 55 },
		{ 101, 74, 55 },
		{ 101, 75, 55 },
		{ 101, 76, 55 },
		{ 101, 77, 55 },
		{ 101, 78, 55 },
		{ 101, 79, 55 },
		{ 101, 80, 55 },
		{ 101, 81, 55 },
		{ 101, 82, 55 },
		{ 101, 83, 55 },
		{ 101, 84, 55 },
		{ 101, 85, 55 },
		{ 101, 86, 55 },
		{ 101, 87, 55 },
		{ 101, 88, 55 },
		{ 101, 89, 55 },
		{ 101, 90, 55 },
		{ 101, 91, 55 },
		{ 101, 92, 55 },
		{ 101, 93, 55 },
		{ 101, 94, 55 },
		{ 101, 95, 55 },
		{ 101, 96, 55 },
		{ 101, 97, 55 },
		{ 101, 98, 55 },
		{ 101, 99, 55 },
		{ 101, 100, 55 },
		{ 102, 51, 55 },
		{ 102, 52, 55 },
		{ 102, 53, 55 },
		{ 102, 54, 55 },
		{ 102, 55, 55 },
		{ 102, 56, 55 },
		{ 102, 57, 55 },
		{ 102, 58, 55 },
		{ 102, 59, 55 },
		{ 102, 60, 55 },
		{ 102, 61, 55 },
		{ 102, 62, 55 },
		{ 102, 63, 55 },
		{ 102, 64, 55 },
		{ 102, 65, 55 },
		{ 102, 66, 55 },
		{ 102, 67, 55 },
		{ 102, 68, 55 },
		{ 102, 69, 55 },
		{ 102, 70, 55 },
		{ 102, 71, 55 },
		{ 102, 72, 55 },
		{ 102, 73, 55 },
		{ 102, 74, 55 },
		{ 102, 75, 55 },
		{ 102, 76, 55 },
		{ 102, 77, 55 },
		{ 102, 78, 55 },
		{ 102, 79, 55 },
		{ 102, 80, 55 },
		{ 102, 81, 55 },
		{ 102, 82, 55 },
		{ 102, 83, 55 },
		{ 102, 84, 55 },
		{ 102, 85, 55 },
		{ 102, 86, 55 },
		{ 102, 87, 55 },
		{ 102, 88, 55 },
		{ 102, 89, 55 },
		{ 102, 90, 55 },
		{ 102, 91, 55 },
		{ 102, 92, 55 },
		{ 102, 93, 55 },
		{ 102, 94, 55 },
		{ 102, 95, 55 },
		{ 102, 96, 55 },
		{ 102, 97, 55 },
		{ 102, 98, 55 },
		{ 102, 99, 55 },
		{ 102, 100, 55 },
		{ 102, 101, 55 },
		{ 103, 52, 55 },
		{ 103, 53, 55 },
		{ 103, 54, 55 },
		{ 103, 55, 55 },
		{ 103, 56, 55 },
		{ 103, 57, 55 },
		{ 103, 58, 55 },
		{ 103, 59, 55 },
		{ 103, 60, 55 },
		{ 103, 61, 55 },
		{ 103, 62, 55 },
		{ 103, 63, 55 },
		{ 103, 64, 55 },
		{ 103, 65, 55 },
		{ 103, 66, 55 },
		{ 103, 67, 55 },
		{ 103, 68, 55 },
		{ 103, 69, 55 },
		{ 103, 70, 55 },
		{ 103, 71, 55 },
		{ 103, 72, 55 },
		{ 103, 73, 55 },
		{ 103, 74, 55 },
		{ 103, 75, 55 },
		{ 103, 76, 55 },
		{ 103, 77, 55 },
		{ 103, 78, 55 },
		{ 103, 79, 55 },
		{ 103, 80, 55 },
		{ 103, 81, 55 },
		{ 103, 82, 55 },
		{ 103, 83, 55 },
		{ 103, 84, 55 },
		{ 103, 85, 55 },
		{ 103, 86, 55 },
		{ 103, 87, 55 },
		{ 103, 88, 55 },
		{ 103, 89, 55 },
		{ 103, 90, 55 },
		{ 103, 91, 55 },
		{ 103, 92, 55 },
		{ 103, 93, 55 },
		{ 103, 94, 55 },
		{ 103, 95, 55 },
		{ 103, 96, 55 },
		{ 103, 97, 55 },
		{ 103, 98, 55 },
		{ 103, 99, 55 },
		{ 103, 100, 55 },
		{ 103, 101, 55 },
		{ 103, 102, 55 },
		{ 104, 54, 55 },
		{ 104, 55, 55 },
		{ 104, 56, 55 },
		{ 104, 57, 55 },
		{ 104, 58, 55 },
		{ 104, 59, 55 },
		{ 104, 60, 55 },
		{ 104, 61, 55 },
		{ 104, 62, 55 },
		{ 104, 63, 55 },
		{ 104, 64, 55 },
		{ 104, 65, 55 },
		{ 104, 66, 55 },
		{ 104, 67, 55 },
		{ 104, 68, 55 },
		{ 104, 69, 55 },
		{ 104, 70, 55 },
		{ 104, 71, 55 },
		{ 104, 72, 55 },
		{ 104, 73, 55 },
		{ 104, 74, 55 },
		{ 104, 75, 55 },
		{ 104, 76, 55 },
		{ 104, 77, 55 },
		{ 104, 78, 55 },
		{ 104, 79, 55 },
		{ 104, 80, 55 },
		{ 104, 81, 55 },
		{ 104, 82, 55 },
		{ 104, 83, 55 },
		{ 104, 84, 55 },
		{ 104, 85, 55 },
		{ 104, 86, 55 },
		{ 104, 87, 55 },
		{ 104, 88, 55 },
		{ 104, 89, 55 },
		{ 104, 90, 55 },
		{ 104, 91, 55 },
		{ 104, 92, 55 },
		{ 104, 93, 55 },
		{ 104, 94, 55 },
		{ 104, 95, 55 },
		{ 104, 96, 55 },
		{ 104, 97, 55 },
		{ 104, 98, 55 },
		{ 104, 99, 55 },
		{ 104, 100, 55 },
		{ 104, 101, 55 },
		{ 104, 102, 55 },
		{ 104, 103, 55 },
		{ 105, 55, 55 },
		{ 105, 56, 55 },
		{ 105, 57, 55 },
		{ 105, 58, 55 },
		{ 105, 59, 55 },
		{ 105, 60, 55 },
		{ 105, 61, 55 },
		{ 105, 62, 55 },
		{ 105, 63, 55 },
		{ 105, 64, 55 },
		{ 105, 65, 55 },
		{ 105, 66, 55 },
		{ 105, 67, 55 },
		{ 105, 68, 55 },
		{ 105, 69, 55 },
		{ 105, 70, 55 },
		{ 105, 71, 55 },
		{ 105, 72, 55 },
		{ 105, 73, 55 },
		{ 105, 74, 55 },
		{ 105, 75, 55 },
		{ 105, 76, 55 },
		{ 105, 77, 55 },
		{ 105, 78, 55 },
		{ 105, 79, 55 },
		{ 105, 80, 55 },
		{ 105, 81, 55 },
		{ 105, 82, 55 },
		{ 105, 83, 55 },
		{ 105, 84, 55 },
		{ 105, 85, 55 },
		{ 105, 86, 55 },
		{ 105, 87, 55 },
		{ 105, 88, 55 },
		{ 105, 89, 55 },
		{ 105, 90, 55 },
		{ 105, 91, 55 },
		{ 105, 92, 55 },
		{ 105, 93, 55 },
		{ 105, 94, 55 },
		{ 105, 95, 55 },
		{ 105, 96, 55 },
		{ 105, 97, 55 },
		{ 105, 98, 55 },
		{ 105, 99, 55 },
		{ 105, 100, 55 },
		{ 105, 101, 55 },
		{ 105, 102, 55 },
		{ 105, 103, 55 },
		{ 105, 104, 55 },
		{ 106, 56, 55 },
		{ 106, 57, 55 },
		{ 106, 58, 55 },
		{ 106, 59, 55 },
		{ 106, 60, 55 },
		{ 106, 61, 55 },
		{ 106, 62, 55 },
		{ 106, 63, 55 },
		{ 106, 64, 55 },
		{ 106, 65, 55 },
		{ 106, 66, 55 },
		{ 106, 67, 55 },
		{ 106, 68, 55 },
		{ 106, 69, 55 },
		{ 106, 70, 55 },
		{ 106, 71, 55 },
		{ 106, 72, 55 },
		{ 106, 73, 55 },
		{ 106, 74, 55 },
		{ 106, 75, 55 },
		{ 106, 76, 55 },
		{ 106, 77, 55 },
		{ 106, 78, 55 },
		{ 106, 79, 55 },
		{ 106, 80, 55 },
		{ 106, 81, 55 },
		{ 106, 82, 55 },
		{ 106, 83, 55 },
		{ 106, 84, 55 },
		{ 106, 85, 55 },
		{ 106, 86, 55 },
		{ 106, 87, 55 },
		{ 106, 88, 55 },
		{ 106, 89, 55 },
		{ 106, 90, 55 },
		{ 106, 91, 55 },
		{ 106, 92, 55 },
		{ 106, 93, 55 },
		{ 106, 94, 55 },
		{ 106, 95, 55 },
		{ 106, 96, 55 },
		{ 106, 97, 55 },
		{ 106, 98, 55 },
		{ 106, 99, 55 },
		{ 106, 100, 55 },
		{ 106, 101, 55 },
		{ 106, 102, 55 },
		{ 106, 103, 55 },
		{ 106, 104, 55 },
		{ 106, 105, 55 },
		{ 107, 58, 55 },
		{ 107, 59, 55 },
		{ 107, 60, 55 },
		{ 107, 61, 55 },
		{ 107, 62, 55 },
		{ 107, 63, 55 },
		{ 107, 64, 55 },
		{ 107, 65, 55 },
		{ 107, 66, 55 },
		{ 107, 67, 55 },
		{ 107, 68, 55 },
		{ 107, 69, 55 },
		{ 107, 70, 55 },
		{ 107, 71, 55 },
		{ 107, 72, 55 },
		{ 107, 73, 55 },
		{ 107, 74, 55 },
		{ 107, 75, 55 },
		{ 107, 76, 55 },
		{ 107, 77, 55 },
		{ 107, 78, 55 },
		{ 107, 79, 55 },
		{ 107, 80, 55 },
		{ 107, 81, 55 },
		{ 107, 82, 55 },
		{ 107, 83, 55 },
		{ 107, 84, 55 },
		{ 107, 85, 55 },
		{ 107, 86, 55 },
		{ 107, 87, 55 },
		{ 107, 88, 55 },
		{ 107, 89, 55 },
		{ 107, 90, 55 },
		{ 107, 91, 55 },
		{ 107, 92, 55 },
		{ 107, 93, 55 },
		{ 107, 94, 55 },
		{ 107, 95, 55 },
		{ 107, 96, 55 },
		{ 107, 97, 55 },
		{ 107, 98, 55 },
		{ 107, 99, 55 },
		{ 107, 100, 55 },
		{ 107, 101, 55 },
		{ 107, 102, 55 },
		{ 107, 103, 55 },
		{ 107, 104, 55 },
		{ 107, 105, 55 },
		{ 107, 106, 55 },
		{ 108, 59, 55 },
		{ 108, 60, 55 },
		{ 108, 61, 55 },
		{ 108, 62, 55 },
		{ 108, 63, 55 },
		{ 108, 64, 55 },
		{ 108, 65, 55 },
		{ 108, 66, 55 },
		{ 108, 67, 55 },
		{ 108, 68, 55 },
		{ 108, 69, 55 },
		{ 108, 70, 55 },
		{ 108, 71, 55 },
		{ 108, 72, 55 },
		{ 108, 73, 55 },
		{ 108, 74, 55 },
		{ 108, 75, 55 },
		{ 108, 76, 55 },
		{ 108, 77, 55 },
		{ 108, 78, 55 },
		{ 108, 79, 55 },
		{ 108, 80, 55 },
		{ 108, 81, 55 },
		{ 108, 82, 55 },
		{ 108, 83, 55 },
		{ 108, 84, 55 },
		{ 108, 85, 55 },
		{ 108, 86, 55 },
		{ 108, 87, 55 },
		{ 108, 88, 55 },
		{ 108, 89, 55 },
		{ 108, 90, 55 },
		{ 108, 91, 55 },
		{ 108, 92, 55 },
		{ 108, 93, 55 },
		{ 108, 94, 55 },
		{ 108, 95, 55 },
		{ 108, 96, 55 },
		{ 108, 97, 55 },
		{ 108, 98, 55 },
		{ 108, 99, 55 },
		{ 108, 100, 55 },
		{ 108, 101, 55 },
		{ 108, 102, 55 },
		{ 108, 103, 55 },
		{ 108, 104, 55 },
		{ 108, 105, 55 },
		{ 108, 106, 55 },
		{ 108, 107, 55 },
		{ 109, 60, 55 },
		{ 109, 61, 55 },
		{ 109, 62, 55 },
		{ 109, 63, 55 },
		{ 109, 64, 55 },
		{ 109, 65, 55 },
		{ 109, 66, 55 },
		{ 109, 67, 55 },
		{ 109, 68, 55 },
		{ 109, 69, 55 },
		{ 109, 70, 55 },
		{ 109, 71, 55 },
		{ 109, 72, 55 },
		{ 109, 73, 55 },
		{ 109, 74, 55 },
		{ 109, 75, 55 },
		{ 109, 76, 55 },
		{ 109, 77, 55 },
		{ 109, 78, 55 },
		{ 109, 79, 55 },
		{ 109, 80, 55 },
		{ 109, 81, 55 },
		{ 109, 82, 55 },
		{ 109, 83, 55 },
		{ 109, 84, 55 },
		{ 109, 85, 55 },
		{ 109, 86, 55 },
		{ 109, 87, 55 },
		{ 109, 88, 55 },
		{ 109, 89, 55 },
		{ 109, 90, 55 },
		{ 109, 91, 55 },
		{ 109, 92, 55 },
		{ 109, 93, 55 },
		{ 109, 94, 55 },
		{ 109, 95, 55 },
		{ 109, 96, 55 },
		{ 109, 97, 55 },
		{ 109, 98, 55 },
		{ 109, 99, 55 },
		{ 109, 100, 55 },
		{ 109, 101, 55 },
		{ 109, 102, 55 },
		{ 109, 103, 55 },
		{ 109, 104, 55 },
		{ 109, 105, 55 },
		{ 109, 106, 55 },
		{ 109, 107, 55 },
		{ 109, 108, 55 },
		{ 110, 62, 55 },
		{ 110, 63, 55 },
		{ 110, 64, 55 },
		{ 110, 65, 55 },
		{ 110, 66, 55 },
		{ 110, 67, 55 },
		{ 110, 68, 55 },
		{ 110, 69, 55 },
		{ 110, 70, 55 },
		{ 110, 71, 55 },
		{ 110, 72, 55 },
		{ 110, 73, 55 },
		{ 110, 74, 55 },
		{ 110, 75, 55 },
		{ 110, 76, 55 },
		{ 110, 77, 55 },
		{ 110, 78, 55 },
		{ 110, 79, 55 },
		{ 110, 80, 55 },
		{ 110, 81, 55 },
		{ 110, 82, 55 },
		{ 110, 83, 55 },
		{ 110, 84, 55 },
		{ 110, 85, 55 },
		{ 110, 86, 55 },
		{ 110, 87, 55 },
		{ 110, 88, 55 },
		{ 110, 89, 55 },
		{ 110, 90, 55 },
		{ 110, 91, 55 },
		{ 110, 92, 55 },
		{ 110, 93, 55 },
		{ 110, 94, 55 },
		{ 110, 95, 55 },
		{ 110, 96, 55 },
		{ 110, 97, 55 },
		{ 110, 98, 55 },
		{ 110, 99, 55 },
		{ 110, 100, 55 },
		{ 110, 101, 55 },
		{ 110, 102, 55 },
		{ 110, 103, 55 },
		{ 110, 104, 55 },
		{ 110, 105, 55 },
		{ 110, 106, 55 },
		{ 110, 107, 55 },
		{ 110, 108, 55 },
		{ 110, 109, 55 },
		{ 111, 63, 55 },
		{ 111, 64, 55 },
		{ 111, 65, 55 },
		{ 111, 66, 55 },
		{ 111, 67, 55 },
		{ 111, 68, 55 },
		{ 111, 69, 55 },
		{ 111, 70, 55 },
		{ 111, 71, 55 },
		{ 111, 72, 55 },
		{ 111, 73, 55 },
		{ 111, 74, 55 },
		{ 111, 75, 55 },
		{ 111, 76, 55 },
		{ 111, 77, 55 },
		{ 111, 78, 55 },
		{ 111, 79, 55 },
		{ 111, 80, 55 },
		{ 111, 81, 55 },
		{ 111, 82, 55 },
		{ 111, 83, 55 },
		{ 111, 84, 55 },
		{ 111, 85, 55 },
		{ 111, 86, 55 },
		{ 111, 87, 55 },
		{ 111, 88, 55 },
		{ 111, 89, 55 },
		{ 111, 90, 55 },
		{ 111, 91, 55 },
		{ 111, 92, 55 },
		{ 111, 93, 55 },
		{ 111, 94, 55 },
		{ 111, 95, 55 },
		{ 111, 96, 55 },
		{ 111, 97, 55 },
		{ 111, 98, 55 },
		{ 111, 99, 55 },
		{ 111, 100, 55 },
		{ 111, 101, 55 },
		{ 111, 102, 55 },
		{ 111, 103, 55 },
		{ 111, 104, 55 },
		{ 111, 105, 55 },
		{ 111, 106, 55 },
		{ 111, 107, 55 },
		{ 111, 108, 55 },
		{ 111, 109, 55 },
		{ 111, 110, 55 },
		{ 112, 64, 55 },
		{ 112, 65, 55 },
		{ 112, 66, 55 },
		{ 112, 67, 55 },
		{ 112, 68, 55 },
		{ 112, 69, 55 },
		{ 112, 70, 55 },
		{ 112, 71, 55 },
		{ 112, 72, 55 },
		{ 112, 73, 55 },
		{ 112, 74, 55 },
		{ 112, 75, 55 },
		{ 112, 76, 55 },
		{ 112, 77, 55 },
		{ 112, 78, 55 },
		{ 112, 79, 55 },
		{ 112, 80, 55 },
		{ 112, 81, 55 },
		{ 112, 82, 55 },
		{ 112, 83, 55 },
		{ 112, 84, 55 },
		{ 112, 85, 55 },
		{ 112, 86, 55 },
		{ 112, 87, 55 },
		{ 112, 88, 55 },
		{ 112, 89, 55 },
		{ 112, 90, 55 },
		{ 112, 91, 55 },
		{ 112, 92, 55 },
		{ 112, 93, 55 },
		{ 112, 94, 55 },
		{ 112, 95, 55 },
		{ 112, 96, 55 },
		{ 112, 97, 55 },
		{ 112, 98, 55 },
		{ 112, 99, 55 },
		{ 112, 100, 55 },
		{ 112, 101, 55 },
		{ 112, 102, 55 },
		{ 112, 103, 55 },
		{ 112, 104, 55 },
		{ 112, 105, 55 },
		{ 112, 106, 55 },
		{ 112, 107, 55 },
		{ 112, 108, 55 },
		{ 112, 109, 55 },
		{ 112, 110, 55 },
		{ 112, 111, 55 },
		{ 113, 66, 55 },
		{ 113, 67, 55 },
		{ 113, 68, 55 },
		{ 113, 69, 55 },
		{ 113, 70, 55 },
		{ 113, 71, 55 },
		{ 113, 72, 55 },
		{ 113, 73, 55 },
		{ 113, 74, 55 },
		{ 113, 75, 55 },
		{ 113, 76, 55 },
		{ 113, 77, 55 },
		{ 113, 78, 55 },
		{ 113, 79, 55 },
		{ 113, 80, 55 },
		{ 113, 81, 55 },
		{ 113, 82, 55 },
		{ 113, 83, 55 },
		{ 113, 84, 55 },
		{ 113, 85, 55 },
		{ 113, 86, 55 },
		{ 113, 87, 55 },
		{ 113, 88, 55 },
		{ 113, 89, 55 },
		{ 113, 90, 55 },
		{ 113, 91, 55 },
		{ 113, 92, 55 },
		{ 113, 93, 55 },
		{ 113, 94, 55 },
		{ 113, 95, 55 },
		{ 113, 96, 55 },
		{ 113, 97, 55 },
		{ 113, 98, 55 },
		{ 113, 99, 55 },
		{ 113, 100, 55 },
		{ 113, 101, 55 },
		{ 113, 102, 55 },
		{ 113, 103, 55 },
		{ 113, 104, 55 },
		{ 113, 105, 55 },
		{ 113, 106, 55 },
		{ 113, 107, 55 },
		{ 113, 108, 55 },
		{ 113, 109, 55 },
		{ 113, 110, 55 },
		{ 113, 111, 55 },
		{ 113, 112, 55 },
		{ 114, 67, 55 },
		{ 114, 68, 55 },
		{ 114, 69, 55 },
		{ 114, 70, 55 },
		{ 114, 71, 55 },
		{ 114, 72, 55 },
		{ 114, 73, 55 },
		{ 114, 74, 55 },
		{ 114, 75, 55 },
		{ 114, 76, 55 },
		{ 114, 77, 55 },
		{ 114, 78, 55 },
		{ 114, 79, 55 },
		{ 114, 80, 55 },
		{ 114, 81, 55 },
		{ 114, 82, 55 },
		{ 114, 83, 55 },
		{ 114, 84, 55 },
		{ 114, 85, 55 },
		{ 114, 86, 55 },
		{ 114, 87, 55 },
		{ 114, 88, 55 },
		{ 114, 89, 55 },
		{ 114, 90, 55 },
		{ 114, 91, 55 },
		{ 114, 92, 55 },
		{ 114, 93, 55 },
		{ 114, 94, 55 },
		{ 114, 95, 55 },
		{ 114, 96, 55 },
		{ 114, 97, 55 },
		{ 114, 98, 55 },
		{ 114, 99, 55 },
		{ 114, 100, 55 },
		{ 114, 101, 55 },
		{ 114, 102, 55 },
		{ 114, 103, 55 },
		{ 114, 104, 55 },
		{ 114, 105, 55 },
		{ 114, 106, 55 },
		{ 114, 107, 55 },
		{ 114, 108, 55 },
		{ 114, 109, 55 },
		{ 114, 110, 55 },
		{ 114, 111, 55 },
		{ 114, 112, 55 },
		{ 114, 113, 55 },
		{ 115, 68, 55 },
		{ 115, 69, 55 },
		{ 115, 70, 55 },
		{ 115, 71, 55 },
		{ 115, 72, 55 },
		{ 115, 73, 55 },
		{ 115, 74, 55 },
		{ 115, 75, 55 },
		{ 115, 76, 55 },
		{ 115, 77, 55 },
		{ 115, 78, 55 },
		{ 115, 79, 55 },
		{ 115, 80, 55 },
		{ 115, 81, 55 },
		{ 115, 82, 55 },
		{ 115, 83, 55 },
		{ 115, 84, 55 },
		{ 115, 85, 55 },
		{ 115, 86, 55 },
		{ 115, 87, 55 },
		{ 115, 88, 55 },
		{ 115, 89, 55 },
		{ 115, 90, 55 },
		{ 115, 91, 55 },
		{ 115, 92, 55 },
		{ 115, 93, 55 },
		{ 115, 94, 55 },
		{ 115, 95, 55 },
		{ 115, 96, 55 },
		{ 115, 97, 55 },
		{ 115, 98, 55 },
		{ 115, 99, 55 },
		{ 115, 100, 55 },
		{ 115, 101, 55 },
		{ 115, 102, 55 },
		{ 115, 103, 55 },
		{ 115, 104, 55 },
		{ 115, 105, 55 },
		{ 115, 106, 55 },
		{ 115, 107, 55 },
		{ 115, 108, 55 },
		{ 115, 109, 55 },
		{ 115, 110, 55 },
		{ 115, 111, 55 },
		{ 115, 112, 55 },
		{ 115, 113, 55 },
		{ 115, 114, 55 },
		{ 116, 70, 55 },
		{ 116, 71, 55 },
		{ 116, 72, 55 },
		{ 116, 73, 55 },
		{ 116, 74, 55 },
		{ 116, 75, 55 },
		{ 116, 76, 55 },
		{ 116, 77, 55 },
		{ 116, 78, 55 },
		{ 116, 79, 55 },
		{ 116, 80, 55 },
		{ 116, 81, 55 },
		{ 116, 82, 55 },
		{ 116, 83, 55 },
		{ 116, 84, 55 },
		{ 116, 85, 55 },
		{ 116, 86, 55 },
		{ 116, 87, 55 },
		{ 116, 88, 55 },
		{ 116, 89, 55 },
		{ 116, 90, 55 },
		{ 116, 91, 55 },
		{ 116, 92, 55 },
		{ 116, 93, 55 },
		{ 116, 94, 55 },
		{ 116, 95, 55 },
		{ 116, 96, 55 },
		{ 116, 97, 55 },
		{ 116, 98, 55 },
		{ 116, 99, 55 },
		{ 116, 100, 55 },
		{ 116, 101, 55 },
		{ 116, 102, 55 },
		{ 116, 103, 55 },
		{ 116, 104, 55 },
		{ 116, 105, 55 },
		{ 116, 106, 55 },
		{ 116, 107, 55 },
		{ 116, 108, 55 },
		{ 116, 109, 55 },
		{ 116, 110, 55 },
		{ 116, 111, 55 },
		{ 116, 112, 55 },
		{ 116, 113, 55 },
		{ 116, 114, 55 },
		{ 116, 115, 55 },
		{ 117, 71, 55 },
		{ 117, 72, 55 },
		{ 117, 73, 55 },
		{ 117, 74, 55 },
		{ 117, 75, 55 },
		{ 117, 76, 55 },
		{ 117, 77, 55 },
		{ 117, 78, 55 },
		{ 117, 79, 55 },
		{ 117, 80, 55 },
		{ 117, 81, 55 },
		{ 117, 82, 55 },
		{ 117, 83, 55 },
		{ 117, 84, 55 },
		{ 117, 85, 55 },
		{ 117, 86, 55 },
		{ 117, 87, 55 },
		{ 117, 88, 55 },
		{ 117, 89, 55 },
		{ 117, 90, 55 },
		{ 117, 91, 55 },
		{ 117, 92, 55 },
		{ 117, 93, 55 },
		{ 117, 94, 55 },
		{ 117, 95, 55 },
		{ 117, 96, 55 },
		{ 117, 97, 55 },
		{ 117, 98, 55 },
		{ 117, 99, 55 },
		{ 117, 100, 55 },
		{ 117, 101, 55 },
		{ 117, 102, 55 },
		{ 117, 103, 55 },
		{ 117, 104, 55 },
		{ 117, 105, 55 },
		{ 117, 106, 55 },
		{ 117, 107, 55 },
		{ 117, 108, 55 },
		{ 117, 109, 55 },
		{ 117, 110, 55 },
		{ 117, 111, 55 },
		{ 117, 112, 55 },
		{ 117, 113, 55 },
		{ 117, 114, 55 },
		{ 117, 115, 55 },
		{ 117, 116, 55 },
		{ 118, 72, 55 },
		{ 118, 73, 55 },
		{ 118, 74, 55 },
		{ 118, 75, 55 },
		{ 118, 76, 55 },
		{ 118, 77, 55 },
		{ 118, 78, 55 },
		{ 118, 79, 55 },
		{ 118, 80, 55 },
		{ 118, 81, 55 },
		{ 118, 82, 55 },
		{ 118, 83, 55 },
		{ 118, 84, 55 },
		{ 118, 85, 55 },
		{ 118, 86, 55 },
		{ 118, 87, 55 },
		{ 118, 88, 55 },
		{ 118, 89, 55 },
		{ 118, 90, 55 },
		{ 118, 91, 55 },
		{ 118, 92, 55 },
		{ 118, 93, 55 },
		{ 118, 94, 55 },
		{ 118, 95, 55 },
		{ 118, 96, 55 },
		{ 118, 97, 55 },
		{ 118, 98, 55 },
		{ 118, 99, 55 },
		{ 118, 100, 55 },
		{ 118, 101, 55 },
		{ 118, 102, 55 },
		{ 118, 103, 55 },
		{ 118, 104, 55 },
		{ 118, 105, 55 },
		{ 118, 106, 55 },
		{ 118, 107, 55 },
		{ 118, 108, 55 },
		{ 118, 109, 55 },
		{ 118, 110, 55 },
		{ 118, 111, 55 },
		{ 118, 112, 55 },
		{ 118, 113, 55 },
		{ 118, 114, 55 },
		{ 118, 115, 55 },
		{ 118, 116, 55 },
		{ 118, 117, 55 },
		{ 119, 74, 55 },
		{ 119, 75, 55 },
		{ 119, 76, 55 },
		{ 119, 77, 55 },
		{ 119, 78, 55 },
		{ 119, 79, 55 },
		{ 119, 80, 55 },
		{ 119, 81, 55 },
		{ 119, 82, 55 },
		{ 119, 83, 55 },
		{ 119, 84, 55 },
		{ 119, 85, 55 },
		{ 119, 86, 55 },
		{ 119, 87, 55 },
		{ 119, 88, 55 },
		{ 119, 89, 55 },
		{ 119, 90, 55 },
		{ 119, 91, 55 },
		{ 119, 92, 55 },
		{ 119, 93, 55 },
		{ 119, 94, 55 },
		{ 119, 95, 55 },
		{ 119, 96, 55 },
		{ 119, 97, 55 },
		{ 119, 98, 55 },
		{ 119, 99, 55 },
		{ 119, 100, 55 },
		{ 119, 101, 55 },
		{ 119, 102, 55 },
		{ 119, 103, 55 },
		{ 119, 104, 55 },
		{ 119, 105, 55 },
		{ 119, 106, 55 },
		{ 119, 107, 55 },
		{ 119, 108, 55 },
		{ 119, 109, 55 },
		{ 119, 110, 55 },
		{ 119, 111, 55 },
		{ 119, 112, 55 },
		{ 119, 113, 55 },
		{ 119, 114, 55 },
		{ 119, 115, 55 },
		{ 119, 116, 55 },
		{ 119, 117, 55 },
		{ 119, 118, 55 },
		{ 120, 75, 55 },
		{ 120, 76, 55 },
		{ 120, 77, 55 },
		{ 120, 78, 55 },
		{ 120, 79, 55 },
		{ 120, 80, 55 },
		{ 120, 81, 55 },
		{ 120, 82, 55 },
		{ 120, 83, 55 },
		{ 120, 84, 55 },
		{ 120, 85, 55 },
		{ 120, 86, 55 },
		{ 120, 87, 55 },
		{ 120, 88, 55 },
		{ 120, 89, 55 },
		{ 120, 90, 55 },
		{ 120, 91, 55 },
		{ 120, 92, 55 },
		{ 120, 93, 55 },
		{ 120, 94, 55 },
		{ 120, 95, 55 },
		{ 120, 96, 55 },
		{ 120, 97, 55 },
		{ 120, 98, 55 },
		{ 120, 99, 55 },
		{ 120, 100, 55 },
		{ 120, 101, 55 },
		{ 120, 102, 55 },
		{ 120, 103, 55 },
		{ 120, 104, 55 },
		{ 120, 105, 55 },
		{ 120, 106, 55 },
		{ 120, 107, 55 },
		{ 120, 108, 55 },
		{ 120, 109, 55 },
		{ 120, 110, 55 },
		{ 120, 111, 55 },
		{ 120, 112, 55 },
		{ 120, 113, 55 },
		{ 120, 114, 55 },
		{ 120, 115, 55 },
		{ 120, 116, 55 },
		{ 120, 117, 55 },
		{ 120, 118, 55 },
		{ 120, 119, 55 },
		{ 121, 76, 55 },
		{ 121, 77, 55 },
		{ 121, 78, 55 },
		{ 121, 79, 55 },
		{ 121, 80, 55 },
		{ 121, 81, 55 },
		{ 121, 82, 55 },
		{ 121, 83, 55 },
		{ 121, 84, 55 },
		{ 121, 85, 55 },
		{ 121, 86, 55 },
		{ 121, 87, 55 },
		{ 121, 88, 55 },
		{ 121, 89, 55 },
		{ 121, 90, 55 },
		{ 121, 91, 55 },
		{ 121, 92, 55 },
		{ 121, 93, 55 },
		{ 121, 94, 55 },
		{ 121, 95, 55 },
		{ 121, 96, 55 },
		{ 121, 97, 55 },
		{ 121, 98, 55 },
		{ 121, 99, 55 },
		{ 121, 100, 55 },
		{ 121, 101, 55 },
		{ 121, 102, 55 },
		{ 121, 103, 55 },
		{ 121, 104, 55 },
		{ 121, 105, 55 },
		{ 121, 106, 55 },
		{ 121, 107, 55 },
		{ 121, 108, 55 },
		{ 121, 109, 55 },
		{ 121, 110, 55 },
		{ 121, 111, 55 },
		{ 121, 112, 55 },
		{ 121, 113, 55 },
		{ 121, 114, 55 },
		{ 121, 115, 55 },
		{ 121, 116, 55 },
		{ 121, 117, 55 },
		{ 121, 118, 55 },
		{ 121, 119, 55 },
		{ 121, 120, 55 },
		{ 122, 78, 55 },
		{ 122, 79, 55 },
		{ 122, 80, 55 },
		{ 122, 81, 55 },
		{ 122, 82, 55 },
		{ 122, 83, 55 },
		{ 122, 84, 55 },
		{ 122, 85, 55 },
		{ 122, 86, 55 },
		{ 122, 87, 55 },
		{ 122, 88, 55 },
		{ 122, 89, 55 },
		{ 122, 90, 55 },
		{ 122, 91, 55 },
		{ 122, 92, 55 },
		{ 122, 93, 55 },
		{ 122, 94, 55 },
		{ 122, 95, 55 },
		{ 122, 96, 55 },
		{ 122, 97, 55 },
		{ 122, 98, 55 },
		{ 122, 99, 55 },
		{ 122, 100, 55 },
		{ 122, 101, 55 },
		{ 122, 102, 55 },
		{ 122, 103, 55 },
		{ 122, 104, 55 },
		{ 122, 105, 55 },
		{ 122, 106, 55 },
		{ 122, 107, 55 },
		{ 122, 108, 55 },
		{ 122, 109, 55 },
		{ 122, 110, 55 },
		{ 122, 111, 55 },
		{ 122, 112, 55 },
		{ 122, 113, 55 },
		{ 122, 114, 55 },
		{ 122, 115, 55 },
		{ 122, 116, 55 },
		{ 122, 117, 55 },
		{ 122, 118, 55 },
		{ 122, 119, 55 },
		{ 122, 120, 55 },
		{ 122, 121, 55 },
		{ 123, 79, 55 },
		{ 123, 80, 55 },
		{ 123, 81, 55 },
		{ 123, 82, 55 },
		{ 123, 83, 55 },
		{ 123, 84, 55 },
		{ 123, 85, 55 },
		{ 123, 86, 55 },
		{ 123, 87, 55 },
		{ 123, 88, 55 },
		{ 123, 89, 55 },
		{ 123, 90, 55 },
		{ 123, 91, 55 },
		{ 123, 92, 55 },
		{ 123, 93, 55 },
		{ 123, 94, 55 },
		{ 123, 95, 55 },
		{ 123, 96, 55 },
		{ 123, 97, 55 },
		{ 123, 98, 55 },
		{ 123, 99, 55 },
		{ 123, 100, 55 },
		{ 123, 101, 55 },
		{ 123, 102, 55 },
		{ 123, 103, 55 },
		{ 123, 104, 55 },
		{ 123, 105, 55 },
		{ 123, 106, 55 },
		{ 123, 107, 55 },
		{ 123, 108, 55 },
		{ 123, 109, 55 },
		{ 123, 110, 55 },
		{ 123, 111, 55 },
		{ 123, 112, 55 },
		{ 123, 113, 55 },
		{ 123, 114, 55 },
		{ 123, 115, 55 },
		{ 123, 116, 55 },
		{ 123, 117, 55 },
		{ 123, 118, 55 },
		{ 123, 119, 55 },
		{ 123, 120, 55 },
		{ 123, 121, 55 },
		{ 123, 122, 55 },
		{ 124, 80, 55 },
		{ 124, 81, 55 },
		{ 124, 82, 55 },
		{ 124, 83, 55 },
		{ 124, 84, 55 },
		{ 124, 85, 55 },
		{ 124, 86, 55 },
		{ 124, 87, 55 },
		{ 124, 88, 55 },
		{ 124, 89, 55 },
		{ 124, 90, 55 },
		{ 124, 91, 55 },
		{ 124, 92, 55 },
		{ 124, 93, 55 },
		{ 124, 94, 55 },
		{ 124, 95, 55 },
		{ 124, 96, 55 },
		{ 124, 97, 55 },
		{ 124, 98, 55 },
		{ 124, 99, 55 },
		{ 124, 100, 55 },
		{ 124, 101, 55 },
		{ 124, 102, 55 },
		{ 124, 103, 55 },
		{ 124, 104, 55 },
		{ 124, 105, 55 },
		{ 124, 106, 55 },
		{ 124, 107, 55 },
		{ 124, 108, 55 },
		{ 124, 109, 55 },
		{ 124, 110, 55 },
		{ 124, 111, 55 },
		{ 124, 112, 55 },
		{ 124, 113, 55 },
		{ 124, 114, 55 },
		{ 124, 115, 55 },
		{ 124, 116, 55 },
		{ 124, 117, 55 },
		{ 124, 118, 55 },
		{ 124, 119, 55 },
		{ 124, 120, 55 },
		{ 124, 121, 55 },
		{ 124, 122, 55 },
		{ 124, 123, 55 },
		{ 125, 82, 55 },
		{ 125, 83, 55 },
		{ 125, 84, 55 },
		{ 125, 85, 55 },
		{ 125, 86, 55 },
		{ 125, 87, 55 },
		{ 125, 88, 55 },
		{ 125, 89, 55 },
		{ 125, 90, 55 },
		{ 125, 91, 55 },
		{ 125, 92, 55 },
		{ 125, 93, 55 },
		{ 125, 94, 55 },
		{ 125, 95, 55 },
		{ 125, 96, 55 },
		{ 125, 97, 55 },
		{ 125, 98, 55 },
		{ 125, 99, 55 },
		{ 125, 100, 55 },
		{ 125, 101, 55 },
		{ 125, 102, 55 },
		{ 125, 103, 55 },
		{ 125, 104, 55 },
		{ 125, 105, 55 },
		{ 125, 106, 55 },
		{ 125, 107, 55 },
		{ 125, 108, 55 },
		{ 125, 109, 55 },
		{ 125, 110, 55 },
		{ 125, 111, 55 },
		{ 125, 112, 55 },
		{ 125, 113, 55 },
		{ 125, 114, 55 },
		{ 125, 115, 55 },
		{ 125, 116, 55 },
		{ 125, 117, 55 },
		{ 125, 118, 55 },
		{ 125, 119, 55 },
		{ 125, 120, 55 },
		{ 125, 121, 55 },
		{ 125, 122, 55 },
		{ 125, 123, 55 },
		{ 125, 124, 55 },
		{ 126, 83, 55 },
		{ 126, 84, 55 },
		{ 126, 85, 55 },
		{ 126, 86, 55 },
		{ 126, 87, 55 },
		{ 126, 88, 55 },
		{ 126, 89, 55 },
		{ 126, 90, 55 },
		{ 126, 91, 55 },
		{ 126, 92, 55 },
		{ 126, 93, 55 },
		{ 126, 94, 55 },
		{ 126, 95, 55 },
		{ 126, 96, 55 },
		{ 126, 97, 55 },
		{ 126, 98, 55 },
		{ 126, 99, 55 },
		{ 126, 100, 55 },
		{ 126, 101, 55 },
		{ 126, 102, 55 },
		{ 126, 103, 55 },
		{ 126, 104, 55 },
		{ 126, 105, 55 },
		{ 126, 106, 55 },
		{ 126, 107, 55 },
		{ 126, 108, 55 },
		{ 126, 109, 55 },
		{ 126, 110, 55 },
		{ 126, 111, 55 },
		{ 126, 112, 55 },
		{ 126, 113, 55 },
		{ 126, 114, 55 },
		{ 126, 115, 55 },
		{ 126, 116, 55 },
		{ 126, 117, 55 },
		{ 126, 118, 55 },
		{ 126, 119, 55 },
		{ 126, 120, 55 },
		{ 126, 121, 55 },
		{ 126, 122, 55 },
		{ 126, 123, 55 },
		{ 126, 124, 55 },
		{ 126, 125, 55 },
		{ 127, 84, 55 },
		{ 127, 85, 55 },
		{ 127, 86, 55 },
		{ 127, 87, 55 },
		{ 127, 88, 55 },
		{ 127, 89, 55 },
		{ 127, 90, 55 },
		{ 127, 91, 55 },
		{ 127, 92, 55 },
		{ 127, 93, 55 },
		{ 127, 94, 55 },
		{ 127, 95, 55 },
		{ 127, 96, 55 },
		{ 127, 97, 55 },
		{ 127, 98, 55 },
		{ 127, 99, 55 },
		{ 127, 100, 55 },
		{ 127, 101, 55 },
		{ 127, 102, 55 },
		{ 127, 103, 55 },
		{ 127, 104, 55 },
		{ 127, 105, 55 },
		{ 127, 106, 55 },
		{ 127, 107, 55 },
		{ 127, 108, 55 },
		{ 127, 109, 55 },
		{ 127, 110, 55 },
		{ 127, 111, 55 },
		{ 127, 112, 55 },
		{ 127, 113, 55 },
		{ 127, 114, 55 },
		{ 127, 115, 55 },
		{ 127, 116, 55 },
		{ 127, 117, 55 },
		{ 127, 118, 55 },
		{ 127, 119, 55 },
		{ 127, 120, 55 },
		{ 127, 121, 55 },
		{ 127, 122, 55 },
		{ 127, 123, 55 },
		{ 127, 124, 55 },
		{ 127, 125, 55 },
		{ 127, 126, 55 },
		{ 128, 86, 55 },
		{ 128, 87, 55 },
		{ 128, 88, 55 },
		{ 128, 89, 55 },
		{ 128, 90, 55 },
		{ 128, 91, 55 },
		{ 128, 92, 55 },
		{ 128, 93, 55 },
		{ 128, 94, 55 },
		{ 128, 95, 55 },
		{ 128, 96, 55 },
		{ 128, 97, 55 },
		{ 128, 98, 55 },
		{ 128, 99, 55 },
		{ 128, 100, 55 },
		{ 128, 101, 55 },
		{ 128, 102, 55 },
		{ 128, 103, 55 },
		{ 128, 104, 55 },
		{ 128, 105, 55 },
		{ 128, 106, 55 },
		{ 128, 107, 55 },
		{ 128, 108, 55 },
		{ 128, 109, 55 },
		{ 128, 110, 55 },
		{ 128, 111, 55 },
		{ 128, 112, 55 },
		{ 128, 113, 55 },
		{ 128, 114, 55 },
		{ 128, 115, 55 },
		{ 128, 116, 55 },
		{ 128, 117, 55 },
		{ 128, 118, 55 },
		{ 128, 119, 55 },
		{ 128, 120, 55 },
		{ 128, 121, 55 },
		{ 128, 122, 55 },
		{ 128, 123, 55 },
		{ 128, 124, 55 },
		{ 128, 125, 55 },
		{ 128, 126, 55 },
		{ 128, 127, 55 },
		{ 129, 87, 55 },
		{ 129, 88, 55 },
		{ 129, 89, 55 },
		{ 129, 90, 55 },
		{ 129, 91, 55 },
		{ 129, 92, 55 },
		{ 129, 93, 55 },
		{ 129, 94, 55 },
		{ 129, 95, 55 },
		{ 129, 96, 55 },
		{ 129, 97, 55 },
		{ 129, 98, 55 },
		{ 129, 99, 55 },
		{ 129, 100, 55 },
		{ 129, 101, 55 },
		{ 129, 102, 55 },
		{ 129, 103, 55 },
		{ 129, 104, 55 },
		{ 129, 105, 55 },
		{ 129, 106, 55 },
		{ 129, 107, 55 },
		{ 129, 108, 55 },
		{ 129, 109, 55 },
		{ 129, 110, 55 },
		{ 129, 111, 55 },
		{ 129, 112, 55 },
		{ 129, 113, 55 },
		{ 129, 114, 55 },
		{ 129, 115, 55 },
		{ 129, 116, 55 },
		{ 129, 117, 55 },
		{ 129, 118, 55 },
		{ 129, 119, 55 },
		{ 129, 120, 55 },
		{ 129, 121, 55 },
		{ 129, 122, 55 },
		{ 129, 123, 55 },
		{ 129, 124, 55 },
		{ 129, 125, 55 },
		{ 129, 126, 55 },
		{ 129, 127, 55 },
		{ 129, 128, 55 },
		{ 130, 88, 55 },
		{ 130, 89, 55 },
		{ 130, 90, 55 },
		{ 130, 91, 55 },
		{ 130, 92, 55 },
		{ 130, 93, 55 },
		{ 130, 94, 55 },
		{ 130, 95, 55 },
		{ 130, 96, 55 },
		{ 130, 97, 55 },
		{ 130, 98, 55 },
		{ 130, 99, 55 },
		{ 130, 100, 55 },
		{ 130, 101, 55 },
		{ 130, 102, 55 },
		{ 130, 103, 55 },
		{ 130, 104, 55 },
		{ 130, 105, 55 },
		{ 130, 106, 55 },
		{ 130, 107, 55 },
		{ 130, 108, 55 },
		{ 130, 109, 55 },
		{ 130, 110, 55 },
		{ 130, 111, 55 },
		{ 130, 112, 55 },
		{ 130, 113, 55 },
		{ 130, 114, 55 },
		{ 130, 115, 55 },
		{ 130, 116, 55 },
		{ 130, 117, 55 },
		{ 130, 118, 55 },
		{ 130, 119, 55 },
		{ 130, 120, 55 },
		{ 130, 121, 55 },
		{ 130, 122, 55 },
		{ 130, 123, 55 },
		{ 130, 124, 55 },
		{ 130, 125, 55 },
		{ 130, 126, 55 },
		{ 130, 127, 55 },
		{ 130, 128, 55 },
		{ 130, 129, 55 },
		{ 131, 90, 55 },
		{ 131, 91, 55 },
		{ 131, 92, 55 },
		{ 131, 93, 55 },
		{ 131, 94, 55 },
		{ 131, 95, 55 },
		{ 131, 96, 55 },
		{ 131, 97, 55 },
		{ 131, 98, 55 },
		{ 131, 99, 55 },
		{ 131, 100, 55 },
		{ 131, 101, 55 },
		{ 131, 102, 55 },
		{ 131, 103, 55 },
		{ 131, 104, 55 },
		{ 131, 105, 55 },
		{ 131, 106, 55 },
		{ 131, 107, 55 },
		{ 131, 108, 55 },
		{ 131, 109, 55 },
		{ 131, 110, 55 },
		{ 131, 111, 55 },
		{ 131, 112, 55 },
		{ 131, 113, 55 },
		{ 131, 114, 55 },
		{ 131, 115, 55 },
		{ 131, 116, 55 },
		{ 131, 117, 55 },
		{ 131, 118, 55 },
		{ 131, 119, 55 },
		{ 131, 120, 55 },
		{ 131, 121, 55 },
		{ 131, 122, 55 },
		{ 131, 123, 55 },
		{ 131, 124, 55 },
		{ 131, 125, 55 },
		{ 131, 126, 55 },
		{ 131, 127, 55 },
		{ 131, 128, 55 },
		{ 131, 129, 55 },
		{ 131, 130, 55 },
		{ 132, 91, 55 },
		{ 132, 92, 55 },
		{ 132, 93, 55 },
		{ 132, 94, 55 },
		{ 132, 95, 55 },
		{ 132, 96, 55 },
		{ 132, 97, 55 },
		{ 132, 98, 55 },
		{ 132, 99, 55 },
		{ 132, 100, 55 },
		{ 132, 101, 55 },
		{ 132, 102, 55 },
		{ 132, 103, 55 },
		{ 132, 104, 55 },
		{ 132, 105, 55 },
		{ 132, 106, 55 },
		{ 132, 107, 55 },
		{ 132, 108, 55 },
		{ 132, 109, 55 },
		{ 132, 110, 55 },
		{ 132, 111, 55 },
		{ 132, 112, 55 },
		{ 132, 113, 55 },
		{ 132, 114, 55 },
		{ 132, 115, 55 },
		{ 132, 116, 55 },
		{ 132, 117, 55 },
		{ 132, 118, 55 },
		{ 132, 119, 55 },
		{ 132, 120, 55 },
		{ 132, 121, 55 },
		{ 132, 122, 55 },
		{ 132, 123, 55 },
		{ 132, 124, 55 },
		{ 132, 125, 55 },
		{ 132, 126, 55 },
		{ 132, 127, 55 },
		{ 132, 128, 55 },
		{ 132, 129, 55 },
		{ 132, 130, 55 },
		{ 132, 131, 55 },
		{ 133, 92, 55 },
		{ 133, 93, 55 },
		{ 133, 94, 55 },
		{ 133, 95, 55 },
		{ 133, 96, 55 },
		{ 133, 97, 55 },
		{ 133, 98, 55 },
		{ 133, 99, 55 },
		{ 133, 100, 55 },
		{ 133, 101, 55 },
		{ 133, 102, 55 },
		{ 133, 103, 55 },
		{ 133, 104, 55 },
		{ 133, 105, 55 },
		{ 133, 106, 55 },
		{ 133, 107, 55 },
		{ 133, 108, 55 },
		{ 133, 109, 55 },
		{ 133, 110, 55 },
		{ 133, 111, 55 },
		{ 133, 112, 55 },
		{ 133, 113, 55 },
		{ 133, 114, 55 },
		{ 133, 115, 55 },
		{ 133, 116, 55 },
		{ 133, 117, 55 },
		{ 133, 118, 55 },
		{ 133, 119, 55 },
		{ 133, 120, 55 },
		{ 133, 121, 55 },
		{ 133, 122, 55 },
		{ 133, 123, 55 },
		{ 133, 124, 55 },
		{ 133, 125, 55 },
		{ 133, 126, 55 },
		{ 133, 127, 55 },
		{ 133, 128, 55 },
		{ 133, 129, 55 },
		{ 133, 130, 55 },
		{ 133, 131, 55 },
		{ 133, 132, 55 },
		{ 134, 94, 55 },
		{ 134, 95, 55 },
		{ 134, 96, 55 },
		{ 134, 97, 55 },
		{ 134, 98, 55 },
		{ 134, 99, 55 },
		{ 134, 100, 55 },
		{ 134, 101, 55 },
		{ 134, 102, 55 },
		{ 134, 103, 55 },
		{ 134, 104, 55 },
		{ 134, 105, 55 },
		{ 134, 106, 55 },
		{ 134, 107, 55 },
		{ 134, 108, 55 },
		{ 134, 109, 55 },
		{ 134, 110, 55 },
		{ 134, 111, 55 },
		{ 134, 112, 55 },
		{ 134, 113, 55 },
		{ 134, 114, 55 },
		{ 134, 115, 55 },
		{ 134, 116, 55 },
		{ 134, 117, 55 },
		{ 134, 118, 55 },
		{ 134, 119, 55 },
		{ 134, 120, 55 },
		{ 134, 121, 55 },
		{ 134, 122, 55 },
		{ 134, 123, 55 },
		{ 134, 124, 55 },
		{ 134, 125, 55 },
		{ 134, 126, 55 },
		{ 134, 127, 55 },
		{ 134, 128, 55 },
		{ 134, 129, 55 },
		{ 134, 130, 55 },
		{ 134, 131, 55 },
		{ 134, 132, 55 },
		{ 134, 133, 55 },
		{ 135, 95, 55 },
		{ 135, 96, 55 },
		{ 135, 97, 55 },
		{ 135, 98, 55 },
		{ 135, 99, 55 },
		{ 135, 100, 55 },
		{ 135, 101, 55 },
		{ 135, 102, 55 },
		{ 135, 103, 55 },
		{ 135, 104, 55 },
		{ 135, 105, 55 },
		{ 135, 106, 55 },
		{ 135, 107, 55 },
		{ 135, 108, 55 },
		{ 135, 109, 55 },
		{ 135, 110, 55 },
		{ 135, 111, 55 },
		{ 135, 112, 55 },
		{ 135, 113, 55 },
		{ 135, 114, 55 },
		{ 135, 115, 55 },
		{ 135, 116, 55 },
		{ 135, 117, 55 },
		{ 135, 118, 55 },
		{ 135, 119, 55 },
		{ 135, 120, 55 },
		{ 135, 121, 55 },
		{ 135, 122, 55 },
		{ 135, 123, 55 },
		{ 135, 124, 55 },
		{ 135, 125, 55 },
		{ 135, 126, 55 },
		{ 135, 127, 55 },
		{ 135, 128, 55 },
		{ 135, 129, 55 },
		{ 135, 130, 55 },
		{ 135, 131, 55 },
		{ 135, 132, 55 },
		{ 135, 133, 55 },
		{ 135, 134, 55 },
		{ 136, 96, 55 },
		{ 136, 97, 55 },
		{ 136, 98, 55 },
		{ 136, 99, 55 },
		{ 136, 100, 55 },
		{ 136, 101, 55 },
		{ 136, 102, 55 },
		{ 136, 103, 55 },
		{ 136, 104, 55 },
		{ 136, 105, 55 },
		{ 136, 106, 55 },
		{ 136, 107, 55 },
		{ 136, 108, 55 },
		{ 136, 109, 55 },
		{ 136, 110, 55 },
		{ 136, 111, 55 },
		{ 136, 112, 55 },
		{ 136, 113, 55 },
		{ 136, 114, 55 },
		{ 136, 115, 55 },
		{ 136, 116, 55 },
		{ 136, 117, 55 },
		{ 136, 118, 55 },
		{ 136, 119, 55 },
		{ 136, 120, 55 },
		{ 136, 121, 55 },
		{ 136, 122, 55 },
		{ 136, 123, 55 },
		{ 136, 124, 55 },
		{ 136, 125, 55 },
		{ 136, 126, 55 },
		{ 136, 127, 55 },
		{ 136, 128, 55 },
		{ 136, 129, 55 },
		{ 136, 130, 55 },
		{ 136, 131, 55 },
		{ 136, 132, 55 },
		{ 136, 133, 55 },
		{ 136, 134, 55 },
		{ 136, 135, 55 },
		{ 137, 98, 55 },
		{ 137, 99, 55 },
		{ 137, 100, 55 },
		{ 137, 101, 55 },
		{ 137, 102, 55 },
		{ 137, 103, 55 },
		{ 137, 104, 55 },
		{ 137, 105, 55 },
		{ 137, 106, 55 },
		{ 137, 107, 55 },
		{ 137, 108, 55 },
		{ 137, 109, 55 },
		{ 137, 110, 55 },
		{ 137, 111, 55 },
		{ 137, 112, 55 },
		{ 137, 113, 55 },
		{ 137, 114, 55 },
		{ 137, 115, 55 },
		{ 137, 116, 55 },
		{ 137, 117, 55 },
		{ 137, 118, 55 },
		{ 137, 119, 55 },
		{ 137, 120, 55 },
		{ 137, 121, 55 },
		{ 137, 122, 55 },
		{ 137, 123, 55 },
		{ 137, 124, 55 },
		{ 137, 125, 55 },
		{ 137, 126, 55 },
		{ 137, 127, 55 },
		{ 137, 128, 55 },
		{ 137, 129, 55 },
		{ 137, 130, 55 },
		{ 137, 131, 55 },
		{ 137, 132, 55 },
		{ 137, 133, 55 },
		{ 137, 134, 55 },
		{ 137, 135, 55 },
		{ 137, 136, 55 },
		{ 138, 99, 55 },
		{ 138, 100, 55 },
		{ 138, 101, 55 },
		{ 138, 102, 55 },
		{ 138, 103, 55 },
		{ 138, 104, 55 },
		{ 138, 105, 55 },
		{ 138, 106, 55 },
		{ 138, 107, 55 },
		{ 138, 108, 55 },
		{ 138, 109, 55 },
		{ 138, 110, 55 },
		{ 138, 111, 55 },
		{ 138, 112, 55 },
		{ 138, 113, 55 },
		{ 138, 114, 55 },
		{ 138, 115, 55 },
		{ 138, 116, 55 },
		{ 138, 117, 55 },
		{ 138, 118, 55 },
		{ 138, 119, 55 },
		{ 138, 120, 55 },
		{ 138, 121, 55 },
		{ 138, 122, 55 },
		{ 138, 123, 55 },
		{ 138, 124, 55 },
		{ 138, 125, 55 },
		{ 138, 126, 55 },
		{ 138, 127, 55 },
		{ 138, 128, 55 },
		{ 138, 129, 55 },
		{ 138, 130, 55 },
		{ 138, 131, 55 },
		{ 138, 132, 55 },
		{ 138, 133, 55 },
		{ 138, 134, 55 },
		{ 138, 135, 55 },
		{ 138, 136, 55 },
		{ 138, 137, 55 },
		{ 139, 100, 55 },
		{ 139, 101, 55 },
		{ 139, 102, 55 },
		{ 139, 103, 55 },
		{ 139, 104, 55 },
		{ 139, 105, 55 },
		{ 139, 106, 55 },
		{ 139, 107, 55 },
		{ 139, 108, 55 },
		{ 139, 109, 55 },
		{ 139, 110, 55 },
		{ 139, 111, 55 },
		{ 139, 112, 55 },
		{ 139, 113, 55 },
		{ 139, 114, 55 },
		{ 139, 115, 55 },
		{ 139, 116, 55 },
		{ 139, 117, 55 },
		{ 139, 118, 55 },
		{ 139, 119, 55 },
		{ 139, 120, 55 },
		{ 139, 121, 55 },
		{ 139, 122, 55 },
		{ 139, 123, 55 },
		{ 139, 124, 55 },
		{ 139, 125, 55 },
		{ 139, 126, 55 },
		{ 139, 127, 55 },
		{ 139, 128, 55 },
		{ 139, 129, 55 },
		{ 139, 130, 55 },
		{ 139, 131, 55 },
		{ 139, 132, 55 },
		{ 139, 133, 55 },
		{ 139, 134, 55 },
		{ 139, 135, 55 },
		{ 139, 136, 55 },
		{ 139, 137, 55 },
		{ 139, 138, 55 },
		{ 140, 102, 55 },
		{ 140, 103, 55 },
		{ 140, 104, 55 },
		{ 140, 105, 55 },
		{ 140, 106, 55 },
		{ 140, 107, 55 },
		{ 140, 108, 55 },
		{ 140, 109, 55 },
		{ 140, 110, 55 },
		{ 140, 111, 55 },
		{ 140, 112, 55 },
		{ 140, 113, 55 },
		{ 140, 114, 55 },
		{ 140, 115, 55 },
		{ 140, 116, 55 },
		{ 140, 117, 55 },
		{ 140, 118, 55 },
		{ 140, 119, 55 },
		{ 140, 120, 55 },
		{ 140, 121, 55 },
		{ 140, 122, 55 },
		{ 140, 123, 55 },
		{ 140, 124, 55 },
		{ 140, 125, 55 },
		{ 140, 126, 55 },
		{ 140, 127, 55 },
		{ 140, 128, 55 },
		{ 140, 129, 55 },
		{ 140, 130, 55 },
		{ 140, 131, 55 },
		{ 140, 132, 55 },
		{ 140, 133, 55 },
		{ 140, 134, 55 },
		{ 140, 135, 55 },
		{ 140, 136, 55 },
		{ 140, 137, 55 },
		{ 140, 138, 55 },
		{ 140, 139, 55 },
		{ 141, 103, 55 },
		{ 141, 104, 55 },
		{ 141, 105, 55 },
		{ 141, 106, 55 },
		{ 141, 107, 55 },
		{ 141, 108, 55 },
		{ 141, 109, 55 },
		{ 141, 110, 55 },
		{ 141, 111, 55 },
		{ 141, 112, 55 },
		{ 141, 113, 55 },
		{ 141, 114, 55 },
		{ 141, 115, 55 },
		{ 141, 116, 55 },
		{ 141, 117, 55 },
		{ 141, 118, 55 },
		{ 141, 119, 55 },
		{ 141, 120, 55 },
		{ 141, 121, 55 },
		{ 141, 122, 55 },
		{ 141, 123, 55 },
		{ 141, 124, 55 },
		{ 141, 125, 55 },
		{ 141, 126, 55 },
		{ 141, 127, 55 },
		{ 141, 128, 55 },
		{ 141, 129, 55 },
		{ 141, 130, 55 },
		{ 141, 131, 55 },
		{ 141, 132, 55 },
		{ 141, 133, 55 },
		{ 141, 134, 55 },
		{ 141, 135, 55 },
		{ 141, 136, 55 },
		{ 141, 137, 55 },
		{ 141, 138, 55 },
		{ 141, 139, 55 },
		{ 141, 140, 55 },
		{ 142, 104, 55 },
		{ 142, 105, 55 },
		{ 142, 106, 55 },
		{ 142, 107, 55 },
		{ 142, 108, 55 },
		{ 142, 109, 55 },
		{ 142, 110, 55 },
		{ 142, 111, 55 },
		{ 142, 112, 55 },
		{ 142, 113, 55 },
		{ 142, 114, 55 },
		{ 142, 115, 55 },
		{ 142, 116, 55 },
		{ 142, 117, 55 },
		{ 142, 118, 55 },
		{ 142, 119, 55 },
		{ 142, 120, 55 },
		{ 142, 121, 55 },
		{ 142, 122, 55 },
		{ 142, 123, 55 },
		{ 142, 124, 55 },
		{ 142, 125, 55 },
		{ 142, 126, 55 },
		{ 142, 127, 55 },
		{ 142, 128, 55 },
		{ 142, 129, 55 },
		{ 142, 130, 55 },
		{ 142, 131, 55 },
		{ 142, 132, 55 },
		{ 142, 133, 55 },
		{ 142, 134, 55 },
		{ 142, 135, 55 },
		{ 142, 136, 55 },
		{ 142, 137, 55 },
		{ 142, 138, 55 },
		{ 142, 139, 55 },
		{ 142, 140, 55 },
		{ 142, 141, 55 },
		{ 143, 106, 55 },
		{ 143, 107, 55 },
		{ 143, 108, 55 },
		{ 143, 109, 55 },
		{ 143, 110, 55 },
		{ 143, 111, 55 },
		{ 143, 112, 55 },
		{ 143, 113, 55 },
		{ 143, 114, 55 },
		{ 143, 115, 55 },
		{ 143, 116, 55 },
		{ 143, 117, 55 },
		{ 143, 118, 55 },
		{ 143, 119, 55 },
		{ 143, 120, 55 },
		{ 143, 121, 55 },
		{ 143, 122, 55 },
		{ 143, 123, 55 },
		{ 143, 124, 55 },
		{ 143, 125, 55 },
		{ 143, 126, 55 },
		{ 143, 127, 55 },
		{ 143, 128, 55 },
		{ 143, 129, 55 },
		{ 143, 130, 55 },
		{ 143, 131, 55 },
		{ 143, 132, 55 },
		{ 143, 133, 55 },
		{ 143, 134, 55 },
		{ 143, 135, 55 },
		{ 143, 136, 55 },
		{ 143, 137, 55 },
		{ 143, 138, 55 },
		{ 143, 139, 55 },
		{ 143, 140, 55 },
		{ 143, 141, 55 },
		{ 143, 142, 55 },
		{ 144, 107, 55 },
		{ 144, 108, 55 },
		{ 144, 109, 55 },
		{ 144, 110, 55 },
		{ 144, 111, 55 },
		{ 144, 112, 55 },
		{ 144, 113, 55 },
		{ 144, 114, 55 },
		{ 144, 115, 55 },
		{ 144, 116, 55 },
		{ 144, 117, 55 },
		{ 144, 118, 55 },
		{ 144, 119, 55 },
		{ 144, 120, 55 },
		{ 144, 121, 55 },
		{ 144, 122, 55 },
		{ 144, 123, 55 },
		{ 144, 124, 55 },
		{ 144, 125, 55 },
		{ 144, 126, 55 },
		{ 144, 127, 55 },
		{ 144, 128, 55 },
		{ 144, 129, 55 },
		{ 144, 130, 55 },
		{ 144, 131, 55 },
		{ 144, 132, 55 },
		{ 144, 133, 55 },
		{ 144, 134, 55 },
		{ 144, 135, 55 },
		{ 144, 136, 55 },
		{ 144, 137, 55 },
		{ 144, 138, 55 },
		{ 144, 139, 55 },
		{ 144, 140, 55 },
		{ 144, 141, 55 },
		{ 144, 142, 55 },
		{ 144, 143, 55 },
		{ 145, 108, 55 },
		{ 145, 109, 55 },
		{ 145, 110, 55 },
		{ 145, 111, 55 },
		{ 145, 112, 55 },
		{ 145, 113, 55 },
		{ 145, 114, 55 },
		{ 145, 115, 55 },
		{ 145, 116, 55 },
		{ 145, 117, 55 },
		{ 145, 118, 55 },
		{ 145, 119, 55 },
		{ 145, 120, 55 },
		{ 145, 121, 55 },
		{ 145, 122, 55 },
		{ 145, 123, 55 },
		{ 145, 124, 55 },
		{ 145, 125, 55 },
		{ 145, 126, 55 },
		{ 145, 127, 55 },
		{ 145, 128, 55 },
		{ 145, 129, 55 },
		{ 145, 130, 55 },
		{ 145, 131, 55 },
		{ 145, 132, 55 },
		{ 145, 133, 55 },
		{ 145, 134, 55 },
		{ 145, 135, 55 },
		{ 145, 136, 55 },
		{ 145, 137, 55 },
		{ 145, 138, 55 },
		{ 145, 139, 55 },
		{ 145, 140, 55 },
		{ 145, 141, 55 },
		{ 145, 142, 55 },
		{ 145, 143, 55 },
		{ 145, 144, 55 },
		{ 146, 110, 55 },
		{ 146, 111, 55 },
		{ 146, 112, 55 },
		{ 146, 113, 55 },
		{ 146, 114, 55 },
		{ 146, 115, 55 },
		{ 146, 116, 55 },
		{ 146, 117, 55 },
		{ 146, 118, 55 },
		{ 146, 119, 55 },
		{ 146, 120, 55 },
		{ 146, 121, 55 },
		{ 146, 122, 55 },
		{ 146, 123, 55 },
		{ 146, 124, 55 },
		{ 146, 125, 55 },
		{ 146, 126, 55 },
		{ 146, 127, 55 },
		{ 146, 128, 55 },
		{ 146, 129, 55 },
		{ 146, 130, 55 },
		{ 146, 131, 55 },
		{ 146, 132, 55 },
		{ 146, 133, 55 },
		{ 146, 134, 55 },
		{ 146, 135, 55 },
		{ 146, 136, 55 },
		{ 146, 137, 55 },
		{ 146, 138, 55 },
		{ 146, 139, 55 },
		{ 146, 140, 55 },
		{ 146, 141, 55 },
		{ 146, 142, 55 },
		{ 146, 143, 55 },
		{ 146, 144, 55 },
		{ 146, 145, 55 },
		{ 147, 111, 55 },
		{ 147, 112, 55 },
		{ 147, 113, 55 },
		{ 147, 114, 55 },
		{ 147, 115, 55 },
		{ 147, 116, 55 },
		{ 147, 117, 55 },
		{ 147, 118, 55 },
		{ 147, 119, 55 },
		{ 147, 120, 55 },
		{ 147, 121, 55 },
		{ 147, 122, 55 },
		{ 147, 123, 55 },
		{ 147, 124, 55 },
		{ 147, 125, 55 },
		{ 147, 126, 55 },
		{ 147, 127, 55 },
		{ 147, 128, 55 },
		{ 147, 129, 55 },
		{ 147, 130, 55 },
		{ 147, 131, 55 },
		{ 147, 132, 55 },
		{ 147, 133, 55 },
		{ 147, 134, 55 },
		{ 147, 135, 55 },
		{ 147, 136, 55 },
		{ 147, 137, 55 },
		{ 147, 138, 55 },
		{ 147, 139, 55 },
		{ 147, 140, 55 },
		{ 147, 141, 55 },
		{ 147, 142, 55 },
		{ 147, 143, 55 },
		{ 147, 144, 55 },
		{ 147, 145, 55 },
		{ 147, 146, 55 },
		{ 148, 112, 55 },
		{ 148, 113, 55 },
		{ 148, 114, 55 },
		{ 148, 115, 55 },
		{ 148, 116, 55 },
		{ 148, 117, 55 },
		{ 148, 118, 55 },
		{ 148, 119, 55 },
		{ 148, 120, 55 },
		{ 148, 121, 55 },
		{ 148, 122, 55 },
		{ 148, 123, 55 },
		{ 148, 124, 55 },
		{ 148, 125, 55 },
		{ 148, 126, 55 },
		{ 148, 127, 55 },
		{ 148, 128, 55 },
		{ 148, 129, 55 },
		{ 148, 130, 55 },
		{ 148, 131, 55 },
		{ 148, 132, 55 },
		{ 148, 133, 55 },
		{ 148, 134, 55 },
		{ 148, 135, 55 },
		{ 148, 136, 55 },
		{ 148, 137, 55 },
		{ 148, 138, 55 },
		{ 148, 139, 55 },
		{ 148, 140, 55 },
		{ 148, 141, 55 },
		{ 148, 142, 55 },
		{ 148, 143, 55 },
		{ 148, 144, 55 },
		{ 148, 145, 55 },
		{ 148, 146, 55 },
		{ 148, 147, 55 },
		{ 149, 114, 55 },
		{ 149, 115, 55 },
		{ 149, 116, 55 },
		{ 149, 117, 55 },
		{ 149, 118, 55 },
		{ 149, 119, 55 },
		{ 149, 120, 55 },
		{ 149, 121, 55 },
		{ 149, 122, 55 },
		{ 149, 123, 55 },
		{ 149, 124, 55 },
		{ 149, 125, 55 },
		{ 149, 126, 55 },
		{ 149, 127, 55 },
		{ 149, 128, 55 },
		{ 149, 129, 55 },
		{ 149, 130, 55 },
		{ 149, 131, 55 },
		{ 149, 132, 55 },
		{ 149, 133, 55 },
		{ 149, 134, 55 },
		{ 149, 135, 55 },
		{ 149, 136, 55 },
		{ 149, 137, 55 },
		{ 149, 138, 55 },
		{ 149, 139, 55 },
		{ 149, 140, 55 },
		{ 149, 141, 55 },
		{ 149, 142, 55 },
		{ 149, 143, 55 },
		{ 149, 144, 55 },
		{ 149, 145, 55 },
		{ 149, 146, 55 },
		{ 149, 147, 55 },
		{ 149, 148, 55 },
		{ 150, 115, 55 },
		{ 150, 116, 55 },
		{ 150, 117, 55 },
		{ 150, 118, 55 },
		{ 150, 119, 55 },
		{ 150, 120, 55 },
		{ 150, 121, 55 },
		{ 150, 122, 55 },
		{ 150, 123, 55 },
		{ 150, 124, 55 },
		{ 150, 125, 55 },
		{ 150, 126, 55 },
		{ 150, 127, 55 },
		{ 150, 128, 55 },
		{ 150, 129, 55 },
		{ 150, 130, 55 },
		{ 150, 131, 55 },
		{ 150, 132, 55 },
		{ 150, 133, 55 },
		{ 150, 134, 55 },
		{ 150, 135, 55 },
		{ 150, 136, 55 },
		{ 150, 137, 55 },
		{ 150, 138, 55 },
		{ 150, 139, 55 },
		{ 150, 140, 55 },
		{ 150, 141, 55 },
		{ 150, 142, 55 },
		{ 150, 143, 55 },
		{ 150, 144, 55 },
		{ 150, 145, 55 },
		{ 150, 146, 55 },
		{ 150, 147, 55 },
		{ 150, 148, 55 },
		{ 150, 149, 55 },
		{ 151, 116, 55 },
		{ 151, 117, 55 },
		{ 151, 118, 55 },
		{ 151, 119, 55 },
		{ 151, 120, 55 },
		{ 151, 121, 55 },
		{ 151, 122, 55 },
		{ 151, 123, 55 },
		{ 151, 124, 55 },
		{ 151, 125, 55 },
		{ 151, 126, 55 },
		{ 151, 127, 55 },
		{ 151, 128, 55 },
		{ 151, 129, 55 },
		{ 151, 130, 55 },
		{ 151, 131, 55 },
		{ 151, 132, 55 },
		{ 151, 133, 55 },
		{ 151, 134, 55 },
		{ 151, 135, 55 },
		{ 151, 136, 55 },
		{ 151, 137, 55 },
		{ 151, 138, 55 },
		{ 151, 139, 55 },
		{ 151, 140, 55 },
		{ 151, 141, 55 },
		{ 151, 142, 55 },
		{ 151, 143, 55 },
		{ 151, 144, 55 },
		{ 151, 145, 55 },
		{ 151, 146, 55 },
		{ 151, 147, 55 },
		{ 151, 148, 55 },
		{ 151, 149, 55 },
		{ 151, 150, 55 },
		{ 152, 118, 55 },
		{ 152, 119, 55 },
		{ 152, 120, 55 },
		{ 152, 121, 55 },
		{ 152, 122, 55 },
		{ 152, 123, 55 },
		{ 152, 124, 55 },
		{ 152, 125, 55 },
		{ 152, 126, 55 },
		{ 152, 127, 55 },
		{ 152, 128, 55 },
		{ 152, 129, 55 },
		{ 152, 130, 55 },
		{ 152, 131, 55 },
		{ 152, 132, 55 },
		{ 152, 133, 55 },
		{ 152, 134, 55 },
		{ 152, 135, 55 },
		{ 152, 136, 55 },
		{ 152, 137, 55 },
		{ 152, 138, 55 },
		{ 152, 139, 55 },
		{ 152, 140, 55 },
		{ 152, 141, 55 },
		{ 152, 142, 55 },
		{ 152, 143, 55 },
		{ 152, 144, 55 },
		{ 152, 145, 55 },
		{ 152, 146, 55 },
		{ 152, 147, 55 },
		{ 152, 148, 55 },
		{ 152, 149, 55 },
		{ 152, 150, 55 },
		{ 152, 151, 55 },
		{ 153, 119, 55 },
		{ 153, 120, 55 },
		{ 153, 121, 55 },
		{ 153, 122, 55 },
		{ 153, 123, 55 },
		{ 153, 124, 55 },
		{ 153, 125, 55 },
		{ 153, 126, 55 },
		{ 153, 127, 55 },
		{ 153, 128, 55 },
		{ 153, 129, 55 },
		{ 153, 130, 55 },
		{ 153, 131, 55 },
		{ 153, 132, 55 },
		{ 153, 133, 55 },
		{ 153, 134, 55 },
		{ 153, 135, 55 },
		{ 153, 136, 55 },
		{ 153, 137, 55 },
		{ 153, 138, 55 },
		{ 153, 139, 55 },
		{ 153, 140, 55 },
		{ 153, 141, 55 },
		{ 153, 142, 55 },
		{ 153, 143, 55 },
		{ 153, 144, 55 },
		{ 153, 145, 55 },
		{ 153, 146, 55 },
		{ 153, 147, 55 },
		{ 153, 148, 55 },
		{ 153, 149, 55 },
		{ 153, 150, 55 },
		{ 153, 151, 55 },
		{ 153, 152, 55 },
		{ 154, 120, 55 },
		{ 154, 121, 55 },
		{ 154, 122, 55 },
		{ 154, 123, 55 },
		{ 154, 124, 55 },
		{ 154, 125, 55 },
		{ 154, 126, 55 },
		{ 154, 127, 55 },
		{ 154, 128, 55 },
		{ 154, 129, 55 },
		{ 154, 130, 55 },
		{ 154, 131, 55 },
		{ 154, 132, 55 },
		{ 154, 133, 55 },
		{ 154, 134, 55 },
		{ 154, 135, 55 },
		{ 154, 136, 55 },
		{ 154, 137, 55 },
		{ 154, 138, 55 },
		{ 154, 139, 55 },
		{ 154, 140, 55 },
		{ 154, 141, 55 },
		{ 154, 142, 55 },
		{ 154, 143, 55 },
		{ 154, 144, 55 },
		{ 154, 145, 55 },
		{ 154, 146, 55 },
		{ 154, 147, 55 },
		{ 154, 148, 55 },
		{ 154, 149, 55 },
		{ 154, 150, 55 },
		{ 154, 151, 55 },
		{ 154, 152, 55 },
		{ 154, 153, 55 },
		{ 155, 122, 55 },
		{ 155, 123, 55 },
		{ 155, 124, 55 },
		{ 155, 125, 55 },
		{ 155, 126, 55 },
		{ 155, 127, 55 },
		{ 155, 128, 55 },
		{ 155, 129, 55 },
		{ 155, 130, 55 },
		{ 155, 131, 55 },
		{ 155, 132, 55 },
		{ 155, 133, 55 },
		{ 155, 134, 55 },
		{ 155, 135, 55 },
		{ 155, 136, 55 },
		{ 155, 137, 55 },
		{ 155, 138, 55 },
		{ 155, 139, 55 },
		{ 155, 140, 55 },
		{ 155, 141, 55 },
		{ 155, 142, 55 },
		{ 155, 143, 55 },
		{ 155, 144, 55 },
		{ 155, 145, 55 },
		{ 155, 146, 55 },
		{ 155, 147, 55 },
		{ 155, 148, 55 },
		{ 155, 149, 55 },
		{ 155, 150, 55 },
		{ 155, 151, 55 },
		{ 155, 152, 55 },
		{ 155, 153, 55 },
		{ 155, 154, 55 },
		{ 156, 123, 55 },
		{ 156, 124, 55 },
		{ 156, 125, 55 },
		{ 156, 126, 55 },
		{ 156, 127, 55 },
		{ 156, 128, 55 },
		{ 156, 129, 55 },
		{ 156, 130, 55 },
		{ 156, 131, 55 },
		{ 156, 132, 55 },
		{ 156, 133, 55 },
		{ 156, 134, 55 },
		{ 156, 135, 55 },
		{ 156, 136, 55 },
		{ 156, 137, 55 },
		{ 156, 138, 55 },
		{ 156, 139, 55 },
		{ 156, 140, 55 },
		{ 156, 141, 55 },
		{ 156, 142, 55 },
		{ 156, 143, 55 },
		{ 156, 144, 55 },
		{ 156, 145, 55 },
		{ 156, 146, 55 },
		{ 156, 147, 55 },
		{ 156, 148, 55 },
		{ 156, 149, 55 },
		{ 156, 150, 55 },
		{ 156, 151, 55 },
		{ 156, 152, 55 },
		{ 156, 153, 55 },
		{ 156, 154, 55 },
		{ 156, 155, 55 },
		{ 157, 124, 55 },
		{ 157, 125, 55 },
		{ 157, 126, 55 },
		{ 157, 127, 55 },
		{ 157, 128, 55 },
		{ 157, 129, 55 },
		{ 157, 130, 55 },
		{ 157, 131, 55 },
		{ 157, 132, 55 },
		{ 157, 133, 55 },
		{ 157, 134, 55 },
		{ 157, 135, 55 },
		{ 157, 136, 55 },
		{ 157, 137, 55 },
		{ 157, 138, 55 },
		{ 157, 139, 55 },
		{ 157, 140, 55 },
		{ 157, 141, 55 },
		{ 157, 142, 55 },
		{ 157, 143, 55 },
		{ 157, 144, 55 },
		{ 157, 145, 55 },
		{ 157, 146, 55 },
		{ 157, 147, 55 },
		{ 157, 148, 55 },
		{ 157, 149, 55 },
		{ 157, 150, 55 },
		{ 157, 151, 55 },
		{ 157, 152, 55 },
		{ 157, 153, 55 },
		{ 157, 154, 55 },
		{ 157, 155, 55 },
		{ 157, 156, 55 },
		{ 158, 126, 55 },
		{ 158, 127, 55 },
		{ 158, 128, 55 },
		{ 158, 129, 55 },
		{ 158, 130, 55 },
		{ 158, 131, 55 },
		{ 158, 132, 55 },
		{ 158, 133, 55 },
		{ 158, 134, 55 },
		{ 158, 135, 55 },
		{ 158, 136, 55 },
		{ 158, 137, 55 },
		{ 158, 138, 55 },
		{ 158, 139, 55 },
		{ 158, 140, 55 },
		{ 158, 141, 55 },
		{ 158, 142, 55 },
		{ 158, 143, 55 },
		{ 158, 144, 55 },
		{ 158, 145, 55 },
		{ 158, 146, 55 },
		{ 158, 147, 55 },
		{ 158, 148, 55 },
		{ 158, 149, 55 },
		{ 158, 150, 55 },
		{ 158, 151, 55 },
		{ 158, 152, 55 },
		{ 158, 153, 55 },
		{ 158, 154, 55 },
		{ 158, 155, 55 },
		{ 158, 156, 55 },
		{ 158, 157, 55 },
		{ 159, 127, 55 },
		{ 159, 128, 55 },
		{ 159, 129, 55 },
		{ 159, 130, 55 },
		{ 159, 131, 55 },
		{ 159, 132, 55 },
		{ 159, 133, 55 },
		{ 159, 134, 55 },
		{ 159, 135, 55 },
		{ 159, 136, 55 },
		{ 159, 137, 55 },
		{ 159, 138, 55 },
		{ 159, 139, 55 },
		{ 159, 140, 55 },
		{ 159, 141, 55 },
		{ 159, 142, 55 },
		{ 159, 143, 55 },
		{ 159, 144, 55 },
		{ 159, 145, 55 },
		{ 159, 146, 55 },
		{ 159, 147, 55 },
		{ 159, 148, 55 },
		{ 159, 149, 55 },
		{ 159, 150, 55 },
		{ 159, 151, 55 },
		{ 159, 152, 55 },
		{ 159, 153, 55 },
		{ 159, 154, 55 },
		{ 159, 155, 55 },
		{ 159, 156, 55 },
		{ 159, 157, 55 },
		{ 159, 158, 55 },
		{ 160, 128, 55 },
		{ 160, 129, 55 },
		{ 160, 130, 55 },
		{ 160, 131, 55 },
		{ 160, 132, 55 },
		{ 160, 133, 55 },
		{ 160, 134, 55 },
		{ 160, 135, 55 },
		{ 160, 136, 55 },
		{ 160, 137, 55 },
		{ 160, 138, 55 },
		{ 160, 139, 55 },
		{ 160, 140, 55 },
		{ 160, 141, 55 },
		{ 160, 142, 55 },
		{ 160, 143, 55 },
		{ 160, 144, 55 },
		{ 160, 145, 55 },
		{ 160, 146, 55 },
		{ 160, 147, 55 },
		{ 160, 148, 55 },
		{ 160, 149, 55 },
		{ 160, 150, 55 },
		{ 160, 151, 55 },
		{ 160, 152, 55 },
		{ 160, 153, 55 },
		{ 160, 154, 55 },
		{ 160, 155, 55 },
		{ 160, 156, 55 },
		{ 160, 157, 55 },
		{ 160, 158, 55 },
		{ 160, 159, 55 },
		{ 161, 130, 55 },
		{ 161, 131, 55 },
		{ 161, 132, 55 },
		{ 161, 133, 55 },
		{ 161, 134, 55 },
		{ 161, 135, 55 },
		{ 161, 136, 55 },
		{ 161, 137, 55 },
		{ 161, 138, 55 },
		{ 161, 139, 55 },
		{ 161, 140, 55 },
		{ 161, 141, 55 },
		{ 161, 142, 55 },
		{ 161, 143, 55 },
		{ 161, 144, 55 },
		{ 161, 145, 55 },
		{ 161, 146, 55 },
		{ 161, 147, 55 },
		{ 161, 148, 55 },
		{ 161, 149, 55 },
		{ 161, 150, 55 },
		{ 161, 151, 55 },
		{ 161, 152, 55 },
		{ 161, 153, 55 },
		{ 161, 154, 55 },
		{ 161, 155, 55 },
		{ 161, 156, 55 },
		{ 161, 157, 55 },
		{ 161, 158, 55 },
		{ 161, 159, 55 },
		{ 161, 160, 55 },
		{ 162, 131, 55 },
		{ 162, 132, 55 },
		{ 162, 133, 55 },
		{ 162, 134, 55 },
		{ 162, 135, 55 },
		{ 162, 136, 55 },
		{ 162, 137, 55 },
		{ 162, 138, 55 },
		{ 162, 139, 55 },
		{ 162, 140, 55 },
		{ 162, 141, 55 },
		{ 162, 142, 55 },
		{ 162, 143, 55 },
		{ 162, 144, 55 },
		{ 162, 145, 55 },
		{ 162, 146, 55 },
		{ 162, 147, 55 },
		{ 162, 148, 55 },
		{ 162, 149, 55 },
		{ 162, 150, 55 },
		{ 162, 151, 55 },
		{ 162, 152, 55 },
		{ 162, 153, 55 },
		{ 162, 154, 55 },
		{ 162, 155, 55 },
		{ 162, 156, 55 },
		{ 162, 157, 55 },
		{ 162, 158, 55 },
		{ 162, 159, 55 },
		{ 162, 160, 55 },
		{ 162, 161, 55 },
		{ 163, 132, 55 },
		{ 163, 133, 55 },
		{ 163, 134, 55 },
		{ 163, 135, 55 },
		{ 163, 136, 55 },
		{ 163, 137, 55 },
		{ 163, 138, 55 },
		{ 163, 139, 55 },
		{ 163, 140, 55 },
		{ 163, 141, 55 },
		{ 163, 142, 55 },
		{ 163, 143, 55 },
		{ 163, 144, 55 },
		{ 163, 145, 55 },
		{ 163, 146, 55 },
		{ 163, 147, 55 },
		{ 163, 148, 55 },
		{ 163, 149, 55 },
		{ 163, 150, 55 },
		{ 163, 151, 55 },
		{ 163, 152, 55 },
		{ 163, 153, 55 },
		{ 163, 154, 55 },
		{ 163, 155, 55 },
		{ 163, 156, 55 },
		{ 163, 157, 55 },
		{ 163, 158, 55 },
		{ 163, 159, 55 },
		{ 163, 160, 55 },
		{ 163, 161, 55 },
		{ 163, 162, 55 },
		{ 164, 134, 55 },
		{ 164, 135, 55 },
		{ 164, 136, 55 },
		{ 164, 137, 55 },
		{ 164, 138, 55 },
		{ 164, 139, 55 },
		{ 164, 140, 55 },
		{ 164, 141, 55 },
		{ 164, 142, 55 },
		{ 164, 143, 55 },
		{ 164, 144, 55 },
		{ 164, 145, 55 },
		{ 164, 146, 55 },
		{ 164, 147, 55 },
		{ 164, 148, 55 },
		{ 164, 149, 55 },
		{ 164, 150, 55 },
		{ 164, 151, 55 },
		{ 164, 152, 55 },
		{ 164, 153, 55 },
		{ 164, 154, 55 },
		{ 164, 155, 55 },
		{ 164, 156, 55 },
		{ 164, 157, 55 },
		{ 164, 158, 55 },
		{ 164, 159, 55 },
		{ 164, 160, 55 },
		{ 164, 161, 55 },
		{ 164, 162, 55 },
		{ 164, 163, 55 },
		{ 165, 135, 55 },
		{ 165, 136, 55 },
		{ 165, 137, 55 },
		{ 165, 138, 55 },
		{ 165, 139, 55 },
		{ 165, 140, 55 },
		{ 165, 141, 55 },
		{ 165, 142, 55 },
		{ 165, 143, 55 },
		{ 165, 144, 55 },
		{ 165, 145, 55 },
		{ 165, 146, 55 },
		{ 165, 147, 55 },
		{ 165, 148, 55 },
		{ 165, 149, 55 },
		{ 165, 150, 55 },
		{ 165, 151, 55 },
		{ 165, 152, 55 },
		{ 165, 153, 55 },
		{ 165, 154, 55 },
		{ 165, 155, 55 },
		{ 165, 156, 55 },
		{ 165, 157, 55 },
		{ 165, 158, 55 },
		{ 165, 159, 55 },
		{ 165, 160, 55 },
		{ 165, 161, 55 },
		{ 165, 162, 55 },
		{ 165, 163, 55 },
		{ 165, 164, 55 },
		{ 166, 136, 55 },
		{ 166, 137, 55 },
		{ 166, 138, 55 },
		{ 166, 139, 55 },
		{ 166, 140, 55 },
		{ 166, 141, 55 },
		{ 166, 142, 55 },
		{ 166, 143, 55 },
		{ 166, 144, 55 },
		{ 166, 145, 55 },
		{ 166, 146, 55 },
		{ 166, 147, 55 },
		{ 166, 148, 55 },
		{ 166, 149, 55 },
		{ 166, 150, 55 },
		{ 166, 151, 55 },
		{ 166, 152, 55 },
		{ 166, 153, 55 },
		{ 166, 154, 55 },
		{ 166, 155, 55 },
		{ 166, 156, 55 },
		{ 166, 157, 55 },
		{ 166, 158, 55 },
		{ 166, 159, 55 },
		{ 166, 160, 55 },
		{ 166, 161, 55 },
		{ 166, 162, 55 },
		{ 166, 163, 55 },
		{ 166, 164, 55 },
		{ 166, 165, 55 },
		{ 167, 138, 55 },
		{ 167, 139, 55 },
		{ 167, 140, 55 },
		{ 167, 141, 55 },
		{ 167, 142, 55 },
		{ 167, 143, 55 },
		{ 167, 144, 55 },
		{ 167, 145, 55 },
		{ 167, 146, 55 },
		{ 167, 147, 55 },
		{ 167, 148, 55 },
		{ 167, 149, 55 },
		{ 167, 150, 55 },
		{ 167, 151, 55 },
		{ 167, 152, 55 },
		{ 167, 153, 55 },
		{ 167, 154, 55 },
		{ 167, 155, 55 },
		{ 167, 156, 55 },
		{ 167, 157, 55 },
		{ 167, 158, 55 },
		{ 167, 159, 55 },
		{ 167, 160, 55 },
		{ 167, 161, 55 },
		{ 167, 162, 55 },
		{ 167, 163, 55 },
		{ 167, 164, 55 },
		{ 167, 165, 55 },
		{ 167, 166, 55 },
		{ 168, 139, 55 },
		{ 168, 140, 55 },
		{ 168, 141, 55 },
		{ 168, 142, 55 },
		{ 168, 143, 55 },
		{ 168, 144, 55 },
		{ 168, 145, 55 },
		{ 168, 146, 55 },
		{ 168, 147, 55 },
		{ 168, 148, 55 },
		{ 168, 149, 55 },
		{ 168, 150, 55 },
		{ 168, 151, 55 },
		{ 168, 152, 55 },
		{ 168, 153, 55 },
		{ 168, 154, 55 },
		{ 168, 155, 55 },
		{ 168, 156, 55 },
		{ 168, 157, 55 },
		{ 168, 158, 55 },
		{ 168, 159, 55 },
		{ 168, 160, 55 },
		{ 168, 161, 55 },
		{ 168, 162, 55 },
		{ 168, 163, 55 },
		{ 168, 164, 55 },
		{ 168, 165, 55 },
		{ 168, 166, 55 },
		{ 168, 167, 55 },
		{ 169, 140, 55 },
		{ 169, 141, 55 },
		{ 169, 142, 55 },
		{ 169, 143, 55 },
		{ 169, 144, 55 },
		{ 169, 145, 55 },
		{ 169, 146, 55 },
		{ 169, 147, 55 },
		{ 169, 148, 55 },
		{ 169, 149, 55 },
		{ 169, 150, 55 },
		{ 169, 151, 55 },
		{ 169, 152, 55 },
		{ 169, 153, 55 },
		{ 169, 154, 55 },
		{ 169, 155, 55 },
		{ 169, 156, 55 },
		{ 169, 157, 55 },
		{ 169, 158, 55 },
		{ 169, 159, 55 },
		{ 169, 160, 55 },
		{ 169, 161, 55 },
		{ 169, 162, 55 },
		{ 169, 163, 55 },
		{ 169, 164, 55 },
		{ 169, 165, 55 },
		{ 169, 166, 55 },
		{ 169, 167, 55 },
		{ 169, 168, 55 },
		{ 170, 142, 55 },
		{ 170, 143, 55 },
		{ 170, 144, 55 },
		{ 170, 145, 55 },
		{ 170, 146, 55 },
		{ 170, 147, 55 },
		{ 170, 148, 55 },
		{ 170, 149, 55 },
		{ 170, 150, 55 },
		{ 170, 151, 55 },
		{ 170, 152, 55 },
		{ 170, 153, 55 },
		{ 170, 154, 55 },
		{ 170, 155, 55 },
		{ 170, 156, 55 },
		{ 170, 157, 55 },
		{ 170, 158, 55 },
		{ 170, 159, 55 },
		{ 170, 160, 55 },
		{ 170, 161, 55 },
		{ 170, 162, 55 },
		{ 170, 163, 55 },
		{ 170, 164, 55 },
		{ 170, 165, 55 },
		{ 170, 166, 55 },
		{ 170, 167, 55 },
		{ 170, 168, 55 },
		{ 170, 169, 55 },
		{ 171, 143, 55 },
		{ 171, 144, 55 },
		{ 171, 145, 55 },
		{ 171, 146, 55 },
		{ 171, 147, 55 },
		{ 171, 148, 55 },
		{ 171, 149, 55 },
		{ 171, 150, 55 },
		{ 171, 151, 55 },
		{ 171, 152, 55 },
		{ 171, 153, 55 },
		{ 171, 154, 55 },
		{ 171, 155, 55 },
		{ 171, 156, 55 },
		{ 171, 157, 55 },
		{ 171, 158, 55 },
		{ 171, 159, 55 },
		{ 171, 160, 55 },
		{ 171, 161, 55 },
		{ 171, 162, 55 },
		{ 171, 163, 55 },
		{ 171, 164, 55 },
		{ 171, 165, 55 },
		{ 171, 166, 55 },
		{ 171, 167, 55 },
		{ 171, 168, 55 },
		{ 171, 169, 55 },
		{ 171, 170, 55 },
		{ 172, 144, 55 },
		{ 172, 145, 55 },
		{ 172, 146, 55 },
		{ 172, 147, 55 },
		{ 172, 148, 55 },
		{ 172, 149, 55 },
		{ 172, 150, 55 },
		{ 172, 151, 55 },
		{ 172, 152, 55 },
		{ 172, 153, 55 },
		{ 172, 154, 55 },
		{ 172, 155, 55 },
		{ 172, 156, 55 },
		{ 172, 157, 55 },
		{ 172, 158, 55 },
		{ 172, 159, 55 },
		{ 172, 160, 55 },
		{ 172, 161, 55 },
		{ 172, 162, 55 },
		{ 172, 163, 55 },
		{ 172, 164, 55 },
		{ 172, 165, 55 },
		{ 172, 166, 55 },
		{ 172, 167, 55 },
		{ 172, 168, 55 },
		{ 172, 169, 55 },
		{ 172, 170, 55 },
		{ 172, 171, 55 },
		{ 173, 146, 55 },
		{ 173, 147, 55 },
		{ 173, 148, 55 },
		{ 173, 149, 55 },
		{ 173, 150, 55 },
		{ 173, 151, 55 },
		{ 173, 152, 55 },
		{ 173, 153, 55 },
		{ 173, 154, 55 },
		{ 173, 155, 55 },
		{ 173, 156, 55 },
		{ 173, 157, 55 },
		{ 173, 158, 55 },
		{ 173, 159, 55 },
		{ 173, 160, 55 },
		{ 173, 161, 55 },
		{ 173, 162, 55 },
		{ 173, 163, 55 },
		{ 173, 164, 55 },
		{ 173, 165, 55 },
		{ 173, 166, 55 },
		{ 173, 167, 55 },
		{ 173, 168, 55 },
		{ 173, 169, 55 },
		{ 173, 170, 55 },
		{ 173, 171, 55 },
		{ 173, 172, 55 },
		{ 174, 147, 55 },
		{ 174, 148, 55 },
		{ 174, 149, 55 },
		{ 174, 150, 55 },
		{ 174, 151, 55 },
		{ 174, 152, 55 },
		{ 174, 153, 55 },
		{ 174, 154, 55 },
		{ 174, 155, 55 },
		{ 174, 156, 55 },
		{ 174, 157, 55 },
		{ 174, 158, 55 },
		{ 174, 159, 55 },
		{ 174, 160, 55 },
		{ 174, 161, 55 },
		{ 174, 162, 55 },
		{ 174, 163, 55 },
		{ 174, 164, 55 },
		{ 174, 165, 55 },
		{ 174, 166, 55 },
		{ 174, 167, 55 },
		{ 174, 168, 55 },
		{ 174, 169, 55 },
		{ 174, 170, 55 },
		{ 174, 171, 55 },
		{ 174, 172, 55 },
		{ 174, 173, 55 },
		{ 175, 148, 55 },
		{ 175, 149, 55 },
		{ 175, 150, 55 },
		{ 175, 151, 55 },
		{ 175, 152, 55 },
		{ 175, 153, 55 },
		{ 175, 154, 55 },
		{ 175, 155, 55 },
		{ 175, 156, 55 },
		{ 175, 157, 55 },
		{ 175, 158, 55 },
		{ 175, 159, 55 },
		{ 175, 160, 55 },
		{ 175, 161, 55 },
		{ 175, 162, 55 },
		{ 175, 163, 55 },
		{ 175, 164, 55 },
		{ 175, 165, 55 },
		{ 175, 166, 55 },
		{ 175, 167, 55 },
		{ 175, 168, 55 },
		{ 175, 169, 55 },
		{ 175, 170, 55 },
		{ 175, 171, 55 },
		{ 175, 172, 55 },
		{ 175, 173, 55 },
		{ 175, 174, 55 },
		{ 176, 150, 55 },
		{ 176, 151, 55 },
		{ 176, 152, 55 },
		{ 176, 153, 55 },
		{ 176, 154, 55 },
		{ 176, 155, 55 },
		{ 176, 156, 55 },
		{ 176, 157, 55 },
		{ 176, 158, 55 },
		{ 176, 159, 55 },
		{ 176, 160, 55 },
		{ 176, 161, 55 },
		{ 176, 162, 55 },
		{ 176, 163, 55 },
		{ 176, 164, 55 },
		{ 176, 165, 55 },
		{ 176, 166, 55 },
		{ 176, 167, 55 },
		{ 176, 168, 55 },
		{ 176, 169, 55 },
		{ 176, 170, 55 },
		{ 176, 171, 55 },
		{ 176, 172, 55 },
		{ 176, 173, 55 },
		{ 176, 174, 55 },
		{ 176, 175, 55 },
		{ 177, 151, 55 },
		{ 177, 152, 55 },
		{ 177, 153, 55 },
		{ 177, 154, 55 },
		{ 177, 155, 55 },
		{ 177, 156, 55 },
		{ 177, 157, 55 },
		{ 177, 158, 55 },
		{ 177, 159, 55 },
		{ 177, 160, 55 },
		{ 177, 161, 55 },
		{ 177, 162, 55 },
		{ 177, 163, 55 },
		{ 177, 164, 55 },
		{ 177, 165, 55 },
		{ 177, 166, 55 },
		{ 177, 167, 55 },
		{ 177, 168, 55 },
		{ 177, 169, 55 },
		{ 177, 170, 55 },
		{ 177, 171, 55 },
		{ 177, 172, 55 },
		{ 177, 173, 55 },
		{ 177, 174, 55 },
		{ 177, 175, 55 },
		{ 177, 176, 55 },
		{ 178, 152, 55 },
		{ 178, 153, 55 },
		{ 178, 154, 55 },
		{ 178, 155, 55 },
		{ 178, 156, 55 },
		{ 178, 157, 55 },
		{ 178, 158, 55 },
		{ 178, 159, 55 },
		{ 178, 160, 55 },
		{ 178, 161, 55 },
		{ 178, 162, 55 },
		{ 178, 163, 55 },
		{ 178, 164, 55 },
		{ 178, 165, 55 },
		{ 178, 166, 55 },
		{ 178, 167, 55 },
		{ 178, 168, 55 },
		{ 178, 169, 55 },
		{ 178, 170, 55 },
		{ 178, 171, 55 },
		{ 178, 172, 55 },
		{ 178, 173, 55 },
		{ 178, 174, 55 },
		{ 178, 175, 55 },
		{ 178, 176, 55 },
		{ 178, 177, 55 },
		{ 179, 154, 55 },
		{ 179, 155, 55 },
		{ 179, 156, 55 },
		{ 179, 157, 55 },
		{ 179, 158, 55 },
		{ 179, 159, 55 },
		{ 179, 160, 55 },
		{ 179, 161, 55 },
		{ 179, 162, 55 },
		{ 179, 163, 55 },
		{ 179, 164, 55 },
		{ 179, 165, 55 },
		{ 179, 166, 55 },
		{ 179, 167, 55 },
		{ 179, 168, 55 },
		{ 179, 169, 55 },
		{ 179, 170, 55 },
		{ 179, 171, 55 },
		{ 179, 172, 55 },
		{ 179, 173, 55 },
		{ 179, 174, 55 },
		{ 179, 175, 55 },
		{ 179, 176, 55 },
		{ 179, 177, 55 },
		{ 179, 178, 55 },
		{ 180, 155, 55 },
		{ 180, 156, 55 },
		{ 180, 157, 55 },
		{ 180, 158, 55 },
		{ 180, 159, 55 },
		{ 180, 160, 55 },
		{ 180, 161, 55 },
		{ 180, 162, 55 },
		{ 180, 163, 55 },
		{ 180, 164, 55 },
		{ 180, 165, 55 },
		{ 180, 166, 55 },
		{ 180, 167, 55 },
		{ 180, 168, 55 },
		{ 180, 169, 55 },
		{ 180, 170, 55 },
		{ 180, 171, 55 },
		{ 180, 172, 55 },
		{ 180, 173, 55 },
		{ 180, 174, 55 },
		{ 180, 175, 55 },
		{ 180, 176, 55 },
		{ 180, 177, 55 },
		{ 180, 178, 55 },
		{ 180, 179, 55 },
		{ 181, 156, 55 },
		{ 181, 157, 55 },
		{ 181, 158, 55 },
		{ 181, 159, 55 },
		{ 181, 160, 55 },
		{ 181, 161, 55 },
		{ 181, 162, 55 },
		{ 181, 163, 55 },
		{ 181, 164, 55 },
		{ 181, 165, 55 },
		{ 181, 166, 55 },
		{ 181, 167, 55 },
		{ 181, 168, 55 },
		{ 181, 169, 55 },
		{ 181, 170, 55 },
		{ 181, 171, 55 },
		{ 181, 172, 55 },
		{ 181, 173, 55 },
		{ 181, 174, 55 },
		{ 181, 175, 55 },
		{ 181, 176, 55 },
		{ 181, 177, 55 },
		{ 181, 178, 55 },
		{ 181, 179, 55 },
		{ 181, 180, 55 },
		{ 182, 158, 55 },
		{ 182, 159, 55 },
		{ 182, 160, 55 },
		{ 182, 161, 55 },
		{ 182, 162, 55 },
		{ 182, 163, 55 },
		{ 182, 164, 55 },
		{ 182, 165, 55 },
		{ 182, 166, 55 },
		{ 182, 167, 55 },
		{ 182, 168, 55 },
		{ 182, 169, 55 },
		{ 182, 170, 55 },
		{ 182, 171, 55 },
		{ 182, 172, 55 },
		{ 182, 173, 55 },
		{ 182, 174, 55 },
		{ 182, 175, 55 },
		{ 182, 176, 55 },
		{ 182, 177, 55 },
		{ 182, 178, 55 },
		{ 182, 179, 55 },
		{ 182, 180, 55 },
		{ 182, 181, 55 },
		{ 183, 159, 55 },
		{ 183, 160, 55 },
		{ 183, 161, 55 },
		{ 183, 162, 55 },
		{ 183, 163, 55 },
		{ 183, 164, 55 },
		{ 183, 165, 55 },
		{ 183, 166, 55 },
		{ 183, 167, 55 },
		{ 183, 168, 55 },
		{ 183, 169, 55 },
		{ 183, 170, 55 },
		{ 183, 171, 55 },
		{ 183, 172, 55 },
		{ 183, 173, 55 },
		{ 183, 174, 55 },
		{ 183, 175, 55 },
		{ 183, 176, 55 },
		{ 183, 177, 55 },
		{ 183, 178, 55 },
		{ 183, 179, 55 },
		{ 183, 180, 55 },
		{ 183, 181, 55 },
		{ 183, 182, 55 },
		{ 184, 160, 55 },
		{ 184, 161, 55 },
		{ 184, 162, 55 },
		{ 184, 163, 55 },
		{ 184, 164, 55 },
		{ 184, 165, 55 },
		{ 184, 166, 55 },
		{ 184, 167, 55 },
		{ 184, 168, 55 },
		{ 184, 169, 55 },
		{ 184, 170, 55 },
		{ 184, 171, 55 },
		{ 184, 172, 55 },
		{ 184, 173, 55 },
		{ 184, 174, 55 },
		{ 184, 175, 55 },
		{ 184, 176, 55 },
		{ 184, 177, 55 },
		{ 184, 178, 55 },
		{ 184, 179, 55 },
		{ 184, 180, 55 },
		{ 184, 181, 55 },
		{ 184, 182, 55 },
		{ 184, 183, 55 },
		{ 185, 162, 55 },
		{ 185, 163, 55 },
		{ 185, 164, 55 },
		{ 185, 165, 55 },
		{ 185, 166, 55 },
		{ 185, 167, 55 },
		{ 185, 168, 55 },
		{ 185, 169, 55 },
		{ 185, 170, 55 },
		{ 185, 171, 55 },
		{ 185, 172, 55 },
		{ 185, 173, 55 },
		{ 185, 174, 55 },
		{ 185, 175, 55 },
		{ 185, 176, 55 },
		{ 185, 177, 55 },
		{ 185, 178, 55 },
		{ 185, 179, 55 },
		{ 185, 180, 55 },
		{ 185, 181, 55 },
		{ 185, 182, 55 },
		{ 185, 183, 55 },
		{ 185, 184, 55 },
		{ 186, 163, 55 },
		{ 186, 164, 55 },
		{ 186, 165, 55 },
		{ 186, 166, 55 },
		{ 186, 167, 55 },
		{ 186, 168, 55 },
		{ 186, 169, 55 },
		{ 186, 170, 55 },
		{ 186, 171, 55 },
		{ 186, 172, 55 },
		{ 186, 173, 55 },
		{ 186, 174, 55 },
		{ 186, 175, 55 },
		{ 186, 176, 55 },
		{ 186, 177, 55 },
		{ 186, 178, 55 },
		{ 186, 179, 55 },
		{ 186, 180, 55 },
		{ 186, 181, 55 },
		{ 186, 182, 55 },
		{ 186, 183, 55 },
		{ 186, 184, 55 },
		{ 186, 185, 55 },
		{ 187, 164, 55 },
		{ 187, 165, 55 },
		{ 187, 166, 55 },
		{ 187, 167, 55 },
		{ 187, 168, 55 },
		{ 187, 169, 55 },
		{ 187, 170, 55 },
		{ 187, 171, 55 },
		{ 187, 172, 55 },
		{ 187, 173, 55 },
		{ 187, 174, 55 },
		{ 187, 175, 55 },
		{ 187, 176, 55 },
		{ 187, 177, 55 },
		{ 187, 178, 55 },
		{ 187, 179, 55 },
		{ 187, 180, 55 },
		{ 187, 181, 55 },
		{ 187, 182, 55 },
		{ 187, 183, 55 },
		{ 187, 184, 55 },
		{ 187, 185, 55 },
		{ 187, 186, 55 },
		{ 188, 166, 55 },
		{ 188, 167, 55 },
		{ 188, 168, 55 },
		{ 188, 169, 55 },
		{ 188, 170, 55 },
		{ 188, 171, 55 },
		{ 188, 172, 55 },
		{ 188, 173, 55 },
		{ 188, 174, 55 },
		{ 188, 175, 55 },
		{ 188, 176, 55 },
		{ 188, 177, 55 },
		{ 188, 178, 55 },
		{ 188, 179, 55 },
		{ 188, 180, 55 },
		{ 188, 181, 55 },
		{ 188, 182, 55 },
		{ 188, 183, 55 },
		{ 188, 184, 55 },
		{ 188, 185, 55 },
		{ 188, 186, 55 },
		{ 188, 187, 55 },
		{ 189, 167, 55 },
		{ 189, 168, 55 },
		{ 189, 169, 55 },
		{ 189, 170, 55 },
		{ 189, 171, 55 },
		{ 189, 172, 55 },
		{ 189, 173, 55 },
		{ 189, 174, 55 },
		{ 189, 175, 55 },
		{ 189, 176, 55 },
		{ 189, 177, 55 },
		{ 189, 178, 55 },
		{ 189, 179, 55 },
		{ 189, 180, 55 },
		{ 189, 181, 55 },
		{ 189, 182, 55 },
		{ 189, 183, 55 },
		{ 189, 184, 55 },
		{ 189, 185, 55 },
		{ 189, 186, 55 },
		{ 189, 187, 55 },
		{ 189, 188, 55 },
		{ 190, 168, 55 },
		{ 190, 169, 55 },
		{ 190, 170, 55 },
		{ 190, 171, 55 },
		{ 190, 172, 55 },
		{ 190, 173, 55 },
		{ 190, 174, 55 },
		{ 190, 175, 55 },
		{ 190, 176, 55 },
		{ 190, 177, 55 },
		{ 190, 178, 55 },
		{ 190, 179, 55 },
		{ 190, 180, 55 },
		{ 190, 181, 55 },
		{ 190, 182, 55 },
		{ 190, 183, 55 },
		{ 190, 184, 55 },
		{ 190, 185, 55 },
		{ 190, 186, 55 },
		{ 190, 187, 55 },
		{ 190, 188, 55 },
		{ 190, 189, 55 },
		{ 191, 170, 55 },
		{ 191, 171, 55 },
		{ 191, 172, 55 },
		{ 191, 173, 55 },
		{ 191, 174, 55 },
		{ 191, 175, 55 },
		{ 191, 176, 55 },
		{ 191, 177, 55 },
		{ 191, 178, 55 },
		{ 191, 179, 55 },
		{ 191, 180, 55 },
		{ 191, 181, 55 },
		{ 191, 182, 55 },
		{ 191, 183, 55 },
		{ 191, 184, 55 },
		{ 191, 185, 55 },
		{ 191, 186, 55 },
		{ 191, 187, 55 },
		{ 191, 188, 55 },
		{ 191, 189, 55 },
		{ 191, 190, 55 },
		{ 192, 171, 55 },
		{ 192, 172, 55 },
		{ 192, 173, 55 },
		{ 192, 174, 55 },
		{ 192, 175, 55 },
		{ 192, 176, 55 },
		{ 192, 177, 55 },
		{ 192, 178, 55 },
		{ 192, 179, 55 },
		{ 192, 180, 55 },
		{ 192, 181, 55 },
		{ 192, 182, 55 },
		{ 192, 183, 55 },
		{ 192, 184, 55 },
		{ 192, 185, 55 },
		{ 192, 186, 55 },
		{ 192, 187, 55 },
		{ 192, 188, 55 },
		{ 192, 189, 55 },
		{ 192, 190, 55 },
		{ 193, 172, 55 },
		{ 193, 173, 55 },
		{ 193, 174, 55 },
		{ 193, 175, 55 },
		{ 193, 176, 55 },
		{ 193, 177, 55 },
		{ 193, 178, 55 },
		{ 193, 179, 55 },
		{ 193, 180, 55 },
		{ 193, 181, 55 },
		{ 193, 182, 55 },
		{ 193, 183, 55 },
		{ 193, 184, 55 },
		{ 193, 185, 55 },
		{ 193, 186, 55 },
		{ 193, 187, 55 },
		{ 193, 188, 55 },
		{ 193, 189, 55 },
		{ 193, 190, 55 },
		{ 193, 191, 55 },
		{ 194, 174, 55 },
		{ 194, 175, 55 },
		{ 194, 176, 55 },
		{ 194, 177, 55 },
		{ 194, 178, 55 },
		{ 194, 179, 55 },
		{ 194, 180, 55 },
		{ 194, 181, 55 },
		{ 194, 182, 55 },
		{ 194, 183, 55 },
		{ 194, 184, 55 },
		{ 194, 185, 55 },
		{ 194, 186, 55 },
		{ 194, 187, 55 },
		{ 194, 188, 55 },
		{ 194, 189, 55 },
		{ 194, 190, 55 },
		{ 194, 191, 55 },
		{ 195, 175, 55 },
		{ 195, 176, 55 },
		{ 195, 177, 55 },
		{ 195, 178, 55 },
		{ 195, 179, 55 },
		{ 195, 180, 55 },
		{ 195, 181, 55 },
		{ 195, 182, 55 },
		{ 195, 183, 55 },
		{ 195, 184, 55 },
		{ 195, 185, 55 },
		{ 195, 186, 55 },
		{ 195, 187, 55 },
		{ 195, 188, 55 },
		{ 195, 189, 55 },
		{ 195, 190, 55 },
		{ 195, 191, 55 },
		{ 196, 176, 55 },
		{ 196, 177, 55 },
		{ 196, 178, 55 },
		{ 196, 179, 55 },
		{ 196, 180, 55 },
		{ 196, 181, 55 },
		{ 196, 182, 55 },
		{ 196, 183, 55 },
		{ 196, 184, 55 },
		{ 196, 185, 55 },
		{ 196, 186, 55 },
		{ 196, 187, 55 },
		{ 196, 188, 55 },
		{ 196, 189, 55 },
		{ 196, 190, 55 },
		{ 196, 191, 55 },
		{ 197, 178, 55 },
		{ 197, 179, 55 },
		{ 197, 180, 55 },
		{ 197, 181, 55 },
		{ 197, 182, 55 },
		{ 197, 183, 55 },
		{ 197, 184, 55 },
		{ 197, 185, 55 },
		{ 197, 186, 55 },
		{ 197, 187, 55 },
		{ 197, 188, 55 },
		{ 197, 189, 55 },
		{ 197, 190, 55 },
		{ 197, 191, 55 },
		{ 198, 179, 55 },
		{ 198, 180, 55 },
		{ 198, 181, 55 },
		{ 198, 182, 55 },
		{ 198, 183, 55 },
		{ 198, 184, 55 },
		{ 198, 185, 55 },
		{ 198, 186, 55 },
		{ 198, 187, 55 },
		{ 198, 188, 55 },
		{ 198, 189, 55 },
		{ 198, 190, 55 },
		{ 198, 191, 55 },
		{ 199, 180, 55 },
		{ 199, 181, 55 },
		{ 199, 182, 55 },
		{ 199, 183, 55 },
		{ 199, 184, 55 },
		{ 199, 185, 55 },
		{ 199, 186, 55 },
		{ 199, 187, 55 },
		{ 199, 188, 55 },
		{ 199, 189, 55 },
		{ 199, 190, 55 },
		{ 199, 191, 55 },
		{ 200, 182, 55 },
		{ 200, 183, 55 },
		{ 200, 184, 55 },
		{ 200, 185, 55 },
		{ 200, 186, 55 },
		{ 200, 187, 55 },
		{ 200, 188, 55 },
		{ 200, 189, 55 },
		{ 200, 190, 55 },
		{ 200, 191, 55 },
		{ 201, 183, 55 },
		{ 201, 184, 55 },
		{ 201, 185, 55 },
		{ 201, 186, 55 },
		{ 201, 187, 55 },
		{ 201, 188, 55 },
		{ 201, 189, 55 },
		{ 201, 190, 55 },
		{ 201, 191, 55 },
		{ 202, 184, 55 },
		{ 202, 185, 55 },
		{ 202, 186, 55 },
		{ 202, 187, 55 },
		{ 202, 188, 55 },
		{ 202, 189, 55 },
		{ 202, 190, 55 },
		{ 202, 191, 55 },
		{ 203, 186, 55 },
		{ 203, 187, 55 },
		{ 203, 188, 55 },
		{ 203, 189, 55 },
		{ 203, 190, 55 },
		{ 203, 191, 55 },
		{ 204, 187, 55 },
		{ 204, 188, 55 },
		{ 204, 189, 55 },
		{ 204, 190, 55 },
		{ 204, 191, 55 },
		{ 205, 189, 55 },
		{ 205, 190, 55 },
		{ 205, 191, 55 },
		{ 206, 190, 55 },
		{ 308, 324, 55 },
		{ 309, 323, 55 },
		{ 309, 324, 55 },
		{ 309, 325, 55 },
		{ 310, 323, 55 },
		{ 310, 324, 55 },
		{ 310, 325, 55 },
		{ 310, 326, 55 },
		{ 310, 327, 55 },
		{ 311, 323, 55 },
		{ 311, 324, 55 },
		{ 311, 325, 55 },
		{ 311, 326, 55 },
		{ 311, 327, 55 },
		{ 311, 328, 55 },
		{ 312, 323, 55 },
		{ 312, 324, 55 },
		{ 312, 325, 55 },
		{ 312, 326, 55 },
		{ 312, 327, 55 },
		{ 312, 328, 55 },
		{ 312, 329, 55 },
		{ 312, 330, 55 },
		{ 313, 323, 55 },
		{ 313, 324, 55 },
		{ 313, 325, 55 },
		{ 313, 326, 55 },
		{ 313, 327, 55 },
		{ 313, 328, 55 },
		{ 313, 329, 55 },
		{ 313, 330, 55 },
		{ 313, 331, 55 },
		{ 314, 323, 55 },
		{ 314, 324, 55 },
		{ 314, 325, 55 },
		{ 314, 326, 55 },
		{ 314, 327, 55 },
		{ 314, 328, 55 },
		{ 314, 329, 55 },
		{ 314, 330, 55 },
		{ 314, 331, 55 },
		{ 314, 332, 55 },
		{ 315, 323, 55 },
		{ 315, 324, 55 },
		{ 315, 325, 55 },
		{ 315, 326, 55 },
		{ 315, 327, 55 },
		{ 315, 328, 55 },
		{ 315, 329, 55 },
		{ 315, 330, 55 },
		{ 315, 331, 55 },
		{ 315, 332, 55 },
		{ 315, 333, 55 },
		{ 315, 334, 55 },
		{ 316, 323, 55 },
		{ 316, 324, 55 },
		{ 316, 325, 55 },
		{ 316, 326, 55 },
		{ 316, 327, 55 },
		{ 316, 328, 55 },
		{ 316, 329, 55 },
		{ 316, 330, 55 },
		{ 316, 331, 55 },
		{ 316, 332, 55 },
		{ 316, 333, 55 },
		{ 316, 334, 55 },
		{ 316, 335, 55 },
		{ 317, 323, 55 },
		{ 317, 324, 55 },
		{ 317, 325, 55 },
		{ 317, 326, 55 },
		{ 317, 327, 55 },
		{ 317, 328, 55 },
		{ 317, 329, 55 },
		{ 317, 330, 55 },
		{ 317, 331, 55 },
		{ 317, 332, 55 },
		{ 317, 333, 55 },
		{ 317, 334, 55 },
		{ 317, 335, 55 },
		{ 317, 336, 55 },
		{ 318, 323, 55 },
		{ 318, 324, 55 },
		{ 318, 325, 55 },
		{ 318, 326, 55 },
		{ 318, 327, 55 },
		{ 318, 328, 55 },
		{ 318, 329, 55 },
		{ 318, 330, 55 },
		{ 318, 331, 55 },
		{ 318, 332, 55 },
		{ 318, 333, 55 },
		{ 318, 334, 55 },
		{ 318, 335, 55 },
		{ 318, 336, 55 },
		{ 318, 337, 55 },
		{ 318, 338, 55 },
		{ 319, 323, 55 },
		{ 319, 324, 55 },
		{ 319, 325, 55 },
		{ 319, 326, 55 },
		{ 319, 327, 55 },
		{ 319, 328, 55 },
		{ 319, 329, 55 },
		{ 319, 330, 55 },
		{ 319, 331, 55 },
		{ 319, 332, 55 },
		{ 319, 333, 55 },
		{ 319, 334, 55 },
		{ 319, 335, 55 },
		{ 319, 336, 55 },
		{ 319, 337, 55 },
		{ 319, 338, 55 },
		{ 319, 339, 55 },
		{ 320, 323, 55 },
		{ 320, 324, 55 },
		{ 320, 325, 55 },
		{ 320, 326, 55 },
		{ 320, 327, 55 },
		{ 320, 328, 55 },
		{ 320, 329, 55 },
		{ 320, 330, 55 },
		{ 320, 331, 55 },
		{ 320, 332, 55 },
		{ 320, 333, 55 },
		{ 320, 334, 55 },
		{ 320, 335, 55 },
		{ 320, 336, 55 },
		{ 320, 337, 55 },
		{ 320, 338, 55 },
		{ 320, 339, 55 },
		{ 320, 340, 55 },
		{ 321, 323, 55 },
		{ 321, 324, 55 },
		{ 321, 325, 55 },
		{ 321, 326, 55 },
		{ 321, 327, 55 },
		{ 321, 328, 55 },
		{ 321, 329, 55 },
		{ 321, 330, 55 },
		{ 321, 331, 55 },
		{ 321, 332, 55 },
		{ 321, 333, 55 },
		{ 321, 334, 55 },
		{ 321, 335, 55 },
		{ 321, 336, 55 },
		{ 321, 337, 55 },
		{ 321, 338, 55 },
		{ 321, 339, 55 },
		{ 321, 340, 55 },
		{ 321, 341, 55 },
		{ 321, 342, 55 },
		{ 322, 324, 55 },
		{ 322, 325, 55 },
		{ 322, 326, 55 },
		{ 322, 327, 55 },
		{ 322, 328, 55 },
		{ 322, 329, 55 },
		{ 322, 330, 55 },
		{ 322, 331, 55 },
		{ 322, 332, 55 },
		{ 322, 333, 55 },
		{ 322, 334, 55 },
		{ 322, 335, 55 },
		{ 322, 336, 55 },
		{ 322, 337, 55 },
		{ 322, 338, 55 },
		{ 322, 339, 55 },
		{ 322, 340, 55 },
		{ 322, 341, 55 },
		{ 322, 342, 55 },
		{ 322, 343, 55 },
		{ 323, 324, 55 },
		{ 323, 325, 55 },
		{ 323, 326, 55 },
		{ 323, 327, 55 },
		{ 323, 328, 55 },
		{ 323, 329, 55 },
		{ 323, 330, 55 },
		{ 323, 331, 55 },
		{ 323, 332, 55 },
		{ 323, 333, 55 },
		{ 323, 334, 55 },
		{ 323, 335, 55 },
		{ 323, 336, 55 },
		{ 323, 337, 55 },
		{ 323, 338, 55 },
		{ 323, 339, 55 },
		{ 323, 340, 55 },
		{ 323, 341, 55 },
		{ 323, 342, 55 },
		{ 323, 343, 55 },
		{ 323, 344, 55 },
		{ 324, 325, 55 },
		{ 324, 326, 55 },
		{ 324, 327, 55 },
		{ 324, 328, 55 },
		{ 324, 329, 55 },
		{ 324, 330, 55 },
		{ 324, 331, 55 },
		{ 324, 332, 55 },
		{ 324, 333, 55 },
		{ 324, 334, 55 },
		{ 324, 335, 55 },
		{ 324, 336, 55 },
		{ 324, 337, 55 },
		{ 324, 338, 55 },
		{ 324, 339, 55 },
		{ 324, 340, 55 },
		{ 324, 341, 55 },
		{ 324, 342, 55 },
		{ 324, 343, 55 },
		{ 324, 344, 55 },
		{ 324, 345, 55 },
		{ 324, 346, 55 },
		{ 325, 326, 55 },
		{ 325, 327, 55 },
		{ 325, 328, 55 },
		{ 325, 329, 55 },
		{ 325, 330, 55 },
		{ 325, 331, 55 },
		{ 325, 332, 55 },
		{ 325, 333, 55 },
		{ 325, 334, 55 },
		{ 325, 335, 55 },
		{ 325, 336, 55 },
		{ 325, 337, 55 },
		{ 325, 338, 55 },
		{ 325, 339, 55 },
		{ 325, 340, 55 },
		{ 325, 341, 55 },
		{ 325, 342, 55 },
		{ 325, 343, 55 },
		{ 325, 344, 55 },
		{ 325, 345, 55 },
		{ 325, 346, 55 },
		{ 325, 347, 55 },
		{ 326, 327, 55 },
		{ 326, 328, 55 },
		{ 326, 329, 55 },
		{ 326, 330, 55 },
		{ 326, 331, 55 },
		{ 326, 332, 55 },
		{ 326, 333, 55 },
		{ 326, 334, 55 },
		{ 326, 335, 55 },
		{ 326, 336, 55 },
		{ 326, 337, 55 },
		{ 326, 338, 55 },
		{ 326, 339, 55 },
		{ 326, 340, 55 },
		{ 326, 341, 55 },
		{ 326, 342, 55 },
		{ 326, 343, 55 },
		{ 326, 344, 55 },
		{ 326, 345, 55 },
		{ 326, 346, 55 },
		{ 326, 347, 55 },
		{ 326, 348, 55 },
		{ 327, 328, 55 },
		{ 327, 329, 55 },
		{ 327, 330, 55 },
		{ 327, 331, 55 },
		{ 327, 332, 55 },
		{ 327, 333, 55 },
		{ 327, 334, 55 },
		{ 327, 335, 55 },
		{ 327, 336, 55 },
		{ 327, 337, 55 },
		{ 327, 338, 55 },
		{ 327, 339, 55 },
		{ 327, 340, 55 },
		{ 327, 341, 55 },
		{ 327, 342, 55 },
		{ 327, 343, 55 },
		{ 327, 344, 55 },
		{ 327, 345, 55 },
		{ 327, 346, 55 },
		{ 327, 347, 55 },
		{ 327, 348, 55 },
		{ 327, 349, 55 },
		{ 327, 350, 55 },
		{ 328, 329, 55 },
		{ 328, 330, 55 },
		{ 328, 331, 55 },
		{ 328, 332, 55 },
		{ 328, 333, 55 },
		{ 328, 334, 55 },
		{ 328, 335, 55 },
		{ 328, 336, 55 },
		{ 328, 337, 55 },
		{ 328, 338, 55 },
		{ 328, 339, 55 },
		{ 328, 340, 55 },
		{ 328, 341, 55 },
		{ 328, 342, 55 },
		{ 328, 343, 55 },
		{ 328, 344, 55 },
		{ 328, 345, 55 },
		{ 328, 346, 55 },
		{ 328, 347, 55 },
		{ 328, 348, 55 },
		{ 328, 349, 55 },
		{ 328, 350, 55 },
		{ 328, 351, 55 },
		{ 329, 330, 55 },
		{ 329, 331, 55 },
		{ 329, 332, 55 },
		{ 329, 333, 55 },
		{ 329, 334, 55 },
		{ 329, 335, 55 },
		{ 329, 336, 55 },
		{ 329, 337, 55 },
		{ 329, 338, 55 },
		{ 329, 339, 55 },
		{ 329, 340, 55 },
		{ 329, 341, 55 },
		{ 329, 342, 55 },
		{ 329, 343, 55 },
		{ 329, 344, 55 },
		{ 329, 345, 55 },
		{ 329, 346, 55 },
		{ 329, 347, 55 },
		{ 329, 348, 55 },
		{ 329, 349, 55 },
		{ 329, 350, 55 },
		{ 329, 351, 55 },
		{ 329, 352, 55 },
		{ 330, 331, 55 },
		{ 330, 332, 55 },
		{ 330, 333, 55 },
		{ 330, 334, 55 },
		{ 330, 335, 55 },
		{ 330, 336, 55 },
		{ 330, 337, 55 },
		{ 330, 338, 55 },
		{ 330, 339, 55 },
		{ 330, 340, 55 },
		{ 330, 341, 55 },
		{ 330, 342, 55 },
		{ 330, 343, 55 },
		{ 330, 344, 55 },
		{ 330, 345, 55 },
		{ 330, 346, 55 },
		{ 330, 347, 55 },
		{ 330, 348, 55 },
		{ 330, 349, 55 },
		{ 330, 350, 55 },
		{ 330, 351, 55 },
		{ 330, 352, 55 },
		{ 330, 353, 55 },
		{ 330, 354, 55 },
		{ 331, 332, 55 },
		{ 331, 333, 55 },
		{ 331, 334, 55 },
		{ 331, 335, 55 },
		{ 331, 336, 55 },
		{ 331, 337, 55 },
		{ 331, 338, 55 },
		{ 331, 339, 55 },
		{ 331, 340, 55 },
		{ 331, 341, 55 },
		{ 331, 342, 55 },
		{ 331, 343, 55 },
		{ 331, 344, 55 },
		{ 331, 345, 55 },
		{ 331, 346, 55 },
		{ 331, 347, 55 },
		{ 331, 348, 55 },
		{ 331, 349, 55 },
		{ 331, 350, 55 },
		{ 331, 351, 55 },
		{ 331, 352, 55 },
		{ 331, 353, 55 },
		{ 331, 354, 55 },
		{ 331, 355, 55 },
		{ 332, 333, 55 },
		{ 332, 334, 55 },
		{ 332, 335, 55 },
		{ 332, 336, 55 },
		{ 332, 337, 55 },
		{ 332, 338, 55 },
		{ 332, 339, 55 },
		{ 332, 340, 55 },
		{ 332, 341, 55 },
		{ 332, 342, 55 },
		{ 332, 343, 55 },
		{ 332, 344, 55 },
		{ 332, 345, 55 },
		{ 332, 346, 55 },
		{ 332, 347, 55 },
		{ 332, 348, 55 },
		{ 332, 349, 55 },
		{ 332, 350, 55 },
		{ 332, 351, 55 },
		{ 332, 352, 55 },
		{ 332, 353, 55 },
		{ 332, 354, 55 },
		{ 332, 355, 55 },
		{ 332, 356, 55 },
		{ 333, 334, 55 },
		{ 333, 335, 55 },
		{ 333, 336, 55 },
		{ 333, 337, 55 },
		{ 333, 338, 55 },
		{ 333, 339, 55 },
		{ 333, 340, 55 },
		{ 333, 341, 55 },
		{ 333, 342, 55 },
		{ 333, 343, 55 },
		{ 333, 344, 55 },
		{ 333, 345, 55 },
		{ 333, 346, 55 },
		{ 333, 347, 55 },
		{ 333, 348, 55 },
		{ 333, 349, 55 },
		{ 333, 350, 55 },
		{ 333, 351, 55 },
		{ 333, 352, 55 },
		{ 333, 353, 55 },
		{ 333, 354, 55 },
		{ 333, 355, 55 },
		{ 333, 356, 55 },
		{ 333, 357, 55 },
		{ 333, 358, 55 },
		{ 334, 335, 55 },
		{ 334, 336, 55 },
		{ 334, 337, 55 },
		{ 334, 338, 55 },
		{ 334, 339, 55 },
		{ 334, 340, 55 },
		{ 334, 341, 55 },
		{ 334, 342, 55 },
		{ 334, 343, 55 },
		{ 334, 344, 55 },
		{ 334, 345, 55 },
		{ 334, 346, 55 },
		{ 334, 347, 55 },
		{ 334, 348, 55 },
		{ 334, 349, 55 },
		{ 334, 350, 55 },
		{ 334, 351, 55 },
		{ 334, 352, 55 },
		{ 334, 353, 55 },
		{ 334, 354, 55 },
		{ 334, 355, 55 },
		{ 334, 356, 55 },
		{ 334, 357, 55 },
		{ 334, 358, 55 },
		{ 334, 359, 55 },
		{ 335, 336, 55 },
		{ 335, 337, 55 },
		{ 335, 338, 55 },
		{ 335, 339, 55 },
		{ 335, 340, 55 },
		{ 335, 341, 55 },
		{ 335, 342, 55 },
		{ 335, 343, 55 },
		{ 335, 344, 55 },
		{ 335, 345, 55 },
		{ 335, 346, 55 },
		{ 335, 347, 55 },
		{ 335, 348, 55 },
		{ 335, 349, 55 },
		{ 335, 350, 55 },
		{ 335, 351, 55 },
		{ 335, 352, 55 },
		{ 335, 353, 55 },
		{ 335, 354, 55 },
		{ 335, 355, 55 },
		{ 335, 356, 55 },
		{ 335, 357, 55 },
		{ 335, 358, 55 },
		{ 335, 359, 55 },
		{ 335, 360, 55 },
		{ 336, 337, 55 },
		{ 336, 338, 55 },
		{ 336, 339, 55 },
		{ 336, 340, 55 },
		{ 336, 341, 55 },
		{ 336, 342, 55 },
		{ 336, 343, 55 },
		{ 336, 344, 55 },
		{ 336, 345, 55 },
		{ 336, 346, 55 },
		{ 336, 347, 55 },
		{ 336, 348, 55 },
		{ 336, 349, 55 },
		{ 336, 350, 55 },
		{ 336, 351, 55 },
		{ 336, 352, 55 },
		{ 336, 353, 55 },
		{ 336, 354, 55 },
		{ 336, 355, 55 },
		{ 336, 356, 55 },
		{ 336, 357, 55 },
		{ 336, 358, 55 },
		{ 336, 359, 55 },
		{ 336, 360, 55 },
		{ 336, 361, 55 },
		{ 336, 362, 55 },
		{ 337, 338, 55 },
		{ 337, 339, 55 },
		{ 337, 340, 55 },
		{ 337, 341, 55 },
		{ 337, 342, 55 },
		{ 337, 343, 55 },
		{ 337, 344, 55 },
		{ 337, 345, 55 },
		{ 337, 346, 55 },
		{ 337, 347, 55 },
		{ 337, 348, 55 },
		{ 337, 349, 55 },
		{ 337, 350, 55 },
		{ 337, 351, 55 },
		{ 337, 352, 55 },
		{ 337, 353, 55 },
		{ 337, 354, 55 },
		{ 337, 355, 55 },
		{ 337, 356, 55 },
		{ 337, 357, 55 },
		{ 337, 358, 55 },
		{ 337, 359, 55 },
		{ 337, 360, 55 },
		{ 337, 361, 55 },
		{ 337, 362, 55 },
		{ 337, 363, 55 },
		{ 338, 339, 55 },
		{ 338, 340, 55 },
		{ 338, 341, 55 },
		{ 338, 342, 55 },
		{ 338, 343, 55 },
		{ 338, 344, 55 },
		{ 338, 345, 55 },
		{ 338, 346, 55 },
		{ 338, 347, 55 },
		{ 338, 348, 55 },
		{ 338, 349, 55 },
		{ 338, 350, 55 },
		{ 338, 351, 55 },
		{ 338, 352, 55 },
		{ 338, 353, 55 },
		{ 338, 354, 55 },
		{ 338, 355, 55 },
		{ 338, 356, 55 },
		{ 338, 357, 55 },
		{ 338, 358, 55 },
		{ 338, 359, 55 },
		{ 338, 360, 55 },
		{ 338, 361, 55 },
		{ 338, 362, 55 },
		{ 338, 363, 55 },
		{ 338, 364, 55 },
		{ 339, 340, 55 },
		{ 339, 341, 55 },
		{ 339, 342, 55 },
		{ 339, 343, 55 },
		{ 339, 344, 55 },
		{ 339, 345, 55 },
		{ 339, 346, 55 },
		{ 339, 347, 55 },
		{ 339, 348, 55 },
		{ 339, 349, 55 },
		{ 339, 350, 55 },
		{ 339, 351, 55 },
		{ 339, 352, 55 },
		{ 339, 353, 55 },
		{ 339, 354, 55 },
		{ 339, 355, 55 },
		{ 339, 356, 55 },
		{ 339, 357, 55 },
		{ 339, 358, 55 },
		{ 339, 359, 55 },
		{ 339, 360, 55 },
		{ 339, 361, 55 },
		{ 339, 362, 55 },
		{ 339, 363, 55 },
		{ 339, 364, 55 },
		{ 339, 365, 55 },
		{ 339, 366, 55 },
		{ 340, 341, 55 },
		{ 340, 342, 55 },
		{ 340, 343, 55 },
		{ 340, 344, 55 },
		{ 340, 345, 55 },
		{ 340, 346, 55 },
		{ 340, 347, 55 },
		{ 340, 348, 55 },
		{ 340, 349, 55 },
		{ 340, 350, 55 },
		{ 340, 351, 55 },
		{ 340, 352, 55 },
		{ 340, 353, 55 },
		{ 340, 354, 55 },
		{ 340, 355, 55 },
		{ 340, 356, 55 },
		{ 340, 357, 55 },
		{ 340, 358, 55 },
		{ 340, 359, 55 },
		{ 340, 360, 55 },
		{ 340, 361, 55 },
		{ 340, 362, 55 },
		{ 340, 363, 55 },
		{ 340, 364, 55 },
		{ 340, 365, 55 },
		{ 340, 366, 55 },
		{ 340, 367, 55 },
		{ 341, 342, 55 },
		{ 341, 343, 55 },
		{ 341, 344, 55 },
		{ 341, 345, 55 },
		{ 341, 346, 55 },
		{ 341, 347, 55 },
		{ 341, 348, 55 },
		{ 341, 349, 55 },
		{ 341, 350, 55 },
		{ 341, 351, 55 },
		{ 341, 352, 55 },
		{ 341, 353, 55 },
		{ 341, 354, 55 },
		{ 341, 355, 55 },
		{ 341, 356, 55 },
		{ 341, 357, 55 },
		{ 341, 358, 55 },
		{ 341, 359, 55 },
		{ 341, 360, 55 },
		{ 341, 361, 55 },
		{ 341, 362, 55 },
		{ 341, 363, 55 },
		{ 341, 364, 55 },
		{ 341, 365, 55 },
		{ 341, 366, 55 },
		{ 341, 367, 55 },
		{ 341, 368, 55 },
		{ 342, 343, 55 },
		{ 342, 344, 55 },
		{ 342, 345, 55 },
		{ 342, 346, 55 },
		{ 342, 347, 55 },
		{ 342, 348, 55 },
		{ 342, 349, 55 },
		{ 342, 350, 55 },
		{ 342, 351, 55 },
		{ 342, 352, 55 },
		{ 342, 353, 55 },
		{ 342, 354, 55 },
		{ 342, 355, 55 },
		{ 342, 356, 55 },
		{ 342, 357, 55 },
		{ 342, 358, 55 },
		{ 342, 359, 55 },
		{ 342, 360, 55 },
		{ 342, 361, 55 },
		{ 342, 362, 55 },
		{ 342, 363, 55 },
		{ 342, 364, 55 },
		{ 342, 365, 55 },
		{ 342, 366, 55 },
		{ 342, 367, 55 },
		{ 342, 368, 55 },
		{ 342, 369, 55 },
		{ 342, 370, 55 },
		{ 343, 344, 55 },
		{ 343, 345, 55 },
		{ 343, 346, 55 },
		{ 343, 347, 55 },
		{ 343, 348, 55 },
		{ 343, 349, 55 },
		{ 343, 350, 55 },
		{ 343, 351, 55 },
		{ 343, 352, 55 },
		{ 343, 353, 55 },
		{ 343, 354, 55 },
		{ 343, 355, 55 },
		{ 343, 356, 55 },
		{ 343, 357, 55 },
		{ 343, 358, 55 },
		{ 343, 359, 55 },
		{ 343, 360, 55 },
		{ 343, 361, 55 },
		{ 343, 362, 55 },
		{ 343, 363, 55 },
		{ 343, 364, 55 },
		{ 343, 365, 55 },
		{ 343, 366, 55 },
		{ 343, 367, 55 },
		{ 343, 368, 55 },
		{ 343, 369, 55 },
		{ 343, 370, 55 },
		{ 343, 371, 55 },
		{ 344, 345, 55 },
		{ 344, 346, 55 },
		{ 344, 347, 55 },
		{ 344, 348, 55 },
		{ 344, 349, 55 },
		{ 344, 350, 55 },
		{ 344, 351, 55 },
		{ 344, 352, 55 },
		{ 344, 353, 55 },
		{ 344, 354, 55 },
		{ 344, 355, 55 },
		{ 344, 356, 55 },
		{ 344, 357, 55 },
		{ 344, 358, 55 },
		{ 344, 359, 55 },
		{ 344, 360, 55 },
		{ 344, 361, 55 },
		{ 344, 362, 55 },
		{ 344, 363, 55 },
		{ 344, 364, 55 },
		{ 344, 365, 55 },
		{ 344, 366, 55 },
		{ 344, 367, 55 },
		{ 344, 368, 55 },
		{ 344, 369, 55 },
		{ 344, 370, 55 },
		{ 344, 371, 55 },
		{ 344, 372, 55 },
		{ 345, 346, 55 },
		{ 345, 347, 55 },
		{ 345, 348, 55 },
		{ 345, 349, 55 },
		{ 345, 350, 55 },
		{ 345, 351, 55 },
		{ 345, 352, 55 },
		{ 345, 353, 55 },
		{ 345, 354, 55 },
		{ 345, 355, 55 },
		{ 345, 356, 55 },
		{ 345, 357, 55 },
		{ 345, 358, 55 },
		{ 345, 359, 55 },
		{ 345, 360, 55 },
		{ 345, 361, 55 },
		{ 345, 362, 55 },
		{ 345, 363, 55 },
		{ 345, 364, 55 },
		{ 345, 365, 55 },
		{ 345, 366, 55 },
		{ 345, 367, 55 },
		{ 345, 368, 55 },
		{ 345, 369, 55 },
		{ 345, 370, 55 },
		{ 345, 371, 55 },
		{ 345, 372, 55 },
		{ 345, 373, 55 },
		{ 345, 374, 55 },
		{ 346, 347, 55 },
		{ 346, 348, 55 },
		{ 346, 349, 55 },
		{ 346, 350, 55 },
		{ 346, 351, 55 },
		{ 346, 352, 55 },
		{ 346, 353, 55 },
		{ 346, 354, 55 },
		{ 346, 355, 55 },
		{ 346, 356, 55 },
		{ 346, 357, 55 },
		{ 346, 358, 55 },
		{ 346, 359, 55 },
		{ 346, 360, 55 },
		{ 346, 361, 55 },
		{ 346, 362, 55 },
		{ 346, 363, 55 },
		{ 346, 364, 55 },
		{ 346, 365, 55 },
		{ 346, 366, 55 },
		{ 346, 367, 55 },
		{ 346, 368, 55 },
		{ 346, 369, 55 },
		{ 346, 370, 55 },
		{ 346, 371, 55 },
		{ 346, 372, 55 },
		{ 346, 373, 55 },
		{ 346, 374, 55 },
		{ 346, 375, 55 },
		{ 347, 348, 55 },
		{ 347, 349, 55 },
		{ 347, 350, 55 },
		{ 347, 351, 55 },
		{ 347, 352, 55 },
		{ 347, 353, 55 },
		{ 347, 354, 55 },
		{ 347, 355, 55 },
		{ 347, 356, 55 },
		{ 347, 357, 55 },
		{ 347, 358, 55 },
		{ 347, 359, 55 },
		{ 347, 360, 55 },
		{ 347, 361, 55 },
		{ 347, 362, 55 },
		{ 347, 363, 55 },
		{ 347, 364, 55 },
		{ 347, 365, 55 },
		{ 347, 366, 55 },
		{ 347, 367, 55 },
		{ 347, 368, 55 },
		{ 347, 369, 55 },
		{ 347, 370, 55 },
		{ 347, 371, 55 },
		{ 347, 372, 55 },
		{ 347, 373, 55 },
		{ 347, 374, 55 },
		{ 347, 375, 55 },
		{ 347, 376, 55 },
		{ 348, 349, 55 },
		{ 348, 350, 55 },
		{ 348, 351, 55 },
		{ 348, 352, 55 },
		{ 348, 353, 55 },
		{ 348, 354, 55 },
		{ 348, 355, 55 },
		{ 348, 356, 55 },
		{ 348, 357, 55 },
		{ 348, 358, 55 },
		{ 348, 359, 55 },
		{ 348, 360, 55 },
		{ 348, 361, 55 },
		{ 348, 362, 55 },
		{ 348, 363, 55 },
		{ 348, 364, 55 },
		{ 348, 365, 55 },
		{ 348, 366, 55 },
		{ 348, 367, 55 },
		{ 348, 368, 55 },
		{ 348, 369, 55 },
		{ 348, 370, 55 },
		{ 348, 371, 55 },
		{ 348, 372, 55 },
		{ 348, 373, 55 },
		{ 348, 374, 55 },
		{ 348, 375, 55 },
		{ 348, 376, 55 },
		{ 348, 377, 55 },
		{ 348, 378, 55 },
		{ 349, 350, 55 },
		{ 349, 351, 55 },
		{ 349, 352, 55 },
		{ 349, 353, 55 },
		{ 349, 354, 55 },
		{ 349, 355, 55 },
		{ 349, 356, 55 },
		{ 349, 357, 55 },
		{ 349, 358, 55 },
		{ 349, 359, 55 },
		{ 349, 360, 55 },
		{ 349, 361, 55 },
		{ 349, 362, 55 },
		{ 349, 363, 55 },
		{ 349, 364, 55 },
		{ 349, 365, 55 },
		{ 349, 366, 55 },
		{ 349, 367, 55 },
		{ 349, 368, 55 },
		{ 349, 369, 55 },
		{ 349, 370, 55 },
		{ 349, 371, 55 },
		{ 349, 372, 55 },
		{ 349, 373, 55 },
		{ 349, 374, 55 },
		{ 349, 375, 55 },
		{ 349, 376, 55 },
		{ 349, 377, 55 },
		{ 349, 378, 55 },
		{ 349, 379, 55 },
		{ 350, 351, 55 },
		{ 350, 352, 55 },
		{ 350, 353, 55 },
		{ 350, 354, 55 },
		{ 350, 355, 55 },
		{ 350, 356, 55 },
		{ 350, 357, 55 },
		{ 350, 358, 55 },
		{ 350, 359, 55 },
		{ 350, 360, 55 },
		{ 350, 361, 55 },
		{ 350, 362, 55 },
		{ 350, 363, 55 },
		{ 350, 364, 55 },
		{ 350, 365, 55 },
		{ 350, 366, 55 },
		{ 350, 367, 55 },
		{ 350, 368, 55 },
		{ 350, 369, 55 },
		{ 350, 370, 55 },
		{ 350, 371, 55 },
		{ 350, 372, 55 },
		{ 350, 373, 55 },
		{ 350, 374, 55 },
		{ 350, 375, 55 },
		{ 350, 376, 55 },
		{ 350, 377, 55 },
		{ 350, 378, 55 },
		{ 350, 379, 55 },
		{ 350, 380, 55 },
		{ 351, 352, 55 },
		{ 351, 353, 55 },
		{ 351, 354, 55 },
		{ 351, 355, 55 },
		{ 351, 356, 55 },
		{ 351, 357, 55 },
		{ 351, 358, 55 },
		{ 351, 359, 55 },
		{ 351, 360, 55 },
		{ 351, 361, 55 },
		{ 351, 362, 55 },
		{ 351, 363, 55 },
		{ 351, 364, 55 },
		{ 351, 365, 55 },
		{ 351, 366, 55 },
		{ 351, 367, 55 },
		{ 351, 368, 55 },
		{ 351, 369, 55 },
		{ 351, 370, 55 },
		{ 351, 371, 55 },
		{ 351, 372, 55 },
		{ 351, 373, 55 },
		{ 351, 374, 55 },
		{ 351, 375, 55 },
		{ 351, 376, 55 },
		{ 351, 377, 55 },
		{ 351, 378, 55 },
		{ 351, 379, 55 },
		{ 351, 380, 55 },
		{ 351, 381, 55 },
		{ 351, 382, 55 },
		{ 352, 353, 55 },
		{ 352, 354, 55 },
		{ 352, 355, 55 },
		{ 352, 356, 55 },
		{ 352, 357, 55 },
		{ 352, 358, 55 },
		{ 352, 359, 55 },
		{ 352, 360, 55 },
		{ 352, 361, 55 },
		{ 352, 362, 55 },
		{ 352, 363, 55 },
		{ 352, 364, 55 },
		{ 352, 365, 55 },
		{ 352, 366, 55 },
		{ 352, 367, 55 },
		{ 352, 368, 55 },
		{ 352, 369, 55 },
		{ 352, 370, 55 },
		{ 352, 371, 55 },
		{ 352, 372, 55 },
		{ 352, 373, 55 },
		{ 352, 374, 55 },
		{ 352, 375, 55 },
		{ 352, 376, 55 },
		{ 352, 377, 55 },
		{ 352, 378, 55 },
		{ 352, 379, 55 },
		{ 352, 380, 55 },
		{ 352, 381, 55 },
		{ 352, 382, 55 },
		{ 352, 383, 55 },
		{ 353, 354, 55 },
		{ 353, 355, 55 },
		{ 353, 356, 55 },
		{ 353, 357, 55 },
		{ 353, 358, 55 },
		{ 353, 359, 55 },
		{ 353, 360, 55 },
		{ 353, 361, 55 },
		{ 353, 362, 55 },
		{ 353, 363, 55 },
		{ 353, 364, 55 },
		{ 353, 365, 55 },
		{ 353, 366, 55 },
		{ 353, 367, 55 },
		{ 353, 368, 55 },
		{ 353, 369, 55 },
		{ 353, 370, 55 },
		{ 353, 371, 55 },
		{ 353, 372, 55 },
		{ 353, 373, 55 },
		{ 353, 374, 55 },
		{ 353, 375, 55 },
		{ 353, 376, 55 },
		{ 353, 377, 55 },
		{ 353, 378, 55 },
		{ 353, 379, 55 },
		{ 353, 380, 55 },
		{ 353, 381, 55 },
		{ 353, 382, 55 },
		{ 353, 383, 55 },
		{ 353, 384, 55 },
		{ 354, 355, 55 },
		{ 354, 356, 55 },
		{ 354, 357, 55 },
		{ 354, 358, 55 },
		{ 354, 359, 55 },
		{ 354, 360, 55 },
		{ 354, 361, 55 },
		{ 354, 362, 55 },
		{ 354, 363, 55 },
		{ 354, 364, 55 },
		{ 354, 365, 55 },
		{ 354, 366, 55 },
		{ 354, 367, 55 },
		{ 354, 368, 55 },
		{ 354, 369, 55 },
		{ 354, 370, 55 },
		{ 354, 371, 55 },
		{ 354, 372, 55 },
		{ 354, 373, 55 },
		{ 354, 374, 55 },
		{ 354, 375, 55 },
		{ 354, 376, 55 },
		{ 354, 377, 55 },
		{ 354, 378, 55 },
		{ 354, 379, 55 },
		{ 354, 380, 55 },
		{ 354, 381, 55 },
		{ 354, 382, 55 },
		{ 354, 383, 55 },
		{ 354, 384, 55 },
		{ 354, 385, 55 },
		{ 354, 386, 55 },
		{ 355, 356, 55 },
		{ 355, 357, 55 },
		{ 355, 358, 55 },
		{ 355, 359, 55 },
		{ 355, 360, 55 },
		{ 355, 361, 55 },
		{ 355, 362, 55 },
		{ 355, 363, 55 },
		{ 355, 364, 55 },
		{ 355, 365, 55 },
		{ 355, 366, 55 },
		{ 355, 367, 55 },
		{ 355, 368, 55 },
		{ 355, 369, 55 },
		{ 355, 370, 55 },
		{ 355, 371, 55 },
		{ 355, 372, 55 },
		{ 355, 373, 55 },
		{ 355, 374, 55 },
		{ 355, 375, 55 },
		{ 355, 376, 55 },
		{ 355, 377, 55 },
		{ 355, 378, 55 },
		{ 355, 379, 55 },
		{ 355, 380, 55 },
		{ 355, 381, 55 },
		{ 355, 382, 55 },
		{ 355, 383, 55 },
		{ 355, 384, 55 },
		{ 355, 385, 55 },
		{ 355, 386, 55 },
		{ 355, 387, 55 },
		{ 356, 357, 55 },
		{ 356, 358, 55 },
		{ 356, 359, 55 },
		{ 356, 360, 55 },
		{ 356, 361, 55 },
		{ 356, 362, 55 },
		{ 356, 363, 55 },
		{ 356, 364, 55 },
		{ 356, 365, 55 },
		{ 356, 366, 55 },
		{ 356, 367, 55 },
		{ 356, 368, 55 },
		{ 356, 369, 55 },
		{ 356, 370, 55 },
		{ 356, 371, 55 },
		{ 356, 372, 55 },
		{ 356, 373, 55 },
		{ 356, 374, 55 },
		{ 356, 375, 55 },
		{ 356, 376, 55 },
		{ 356, 377, 55 },
		{ 356, 378, 55 },
		{ 356, 379, 55 },
		{ 356, 380, 55 },
		{ 356, 381, 55 },
		{ 356, 382, 55 },
		{ 356, 383, 55 },
		{ 356, 384, 55 },
		{ 356, 385, 55 },
		{ 356, 386, 55 },
		{ 356, 387, 55 },
		{ 356, 388, 55 },
		{ 357, 358, 55 },
		{ 357, 359, 55 },
		{ 357, 360, 55 },
		{ 357, 361, 55 },
		{ 357, 362, 55 },
		{ 357, 363, 55 },
		{ 357, 364, 55 },
		{ 357, 365, 55 },
		{ 357, 366, 55 },
		{ 357, 367, 55 },
		{ 357, 368, 55 },
		{ 357, 369, 55 },
		{ 357, 370, 55 },
		{ 357, 371, 55 },
		{ 357, 372, 55 },
		{ 357, 373, 55 },
		{ 357, 374, 55 },
		{ 357, 375, 55 },
		{ 357, 376, 55 },
		{ 357, 377, 55 },
		{ 357, 378, 55 },
		{ 357, 379, 55 },
		{ 357, 380, 55 },
		{ 357, 381, 55 },
		{ 357, 382, 55 },
		{ 357, 383, 55 },
		{ 357, 384, 55 },
		{ 357, 385, 55 },
		{ 357, 386, 55 },
		{ 357, 387, 55 },
		{ 357, 388, 55 },
		{ 357, 389, 55 },
		{ 357, 390, 55 },
		{ 358, 359, 55 },
		{ 358, 360, 55 },
		{ 358, 361, 55 },
		{ 358, 362, 55 },
		{ 358, 363, 55 },
		{ 358, 364, 55 },
		{ 358, 365, 55 },
		{ 358, 366, 55 },
		{ 358, 367, 55 },
		{ 358, 368, 55 },
		{ 358, 369, 55 },
		{ 358, 370, 55 },
		{ 358, 371, 55 },
		{ 358, 372, 55 },
		{ 358, 373, 55 },
		{ 358, 374, 55 },
		{ 358, 375, 55 },
		{ 358, 376, 55 },
		{ 358, 377, 55 },
		{ 358, 378, 55 },
		{ 358, 379, 55 },
		{ 358, 380, 55 },
		{ 358, 381, 55 },
		{ 358, 382, 55 },
		{ 358, 383, 55 },
		{ 358, 384, 55 },
		{ 358, 385, 55 },
		{ 358, 386, 55 },
		{ 358, 387, 55 },
		{ 358, 388, 55 },
		{ 358, 389, 55 },
		{ 358, 390, 55 },
		{ 358, 391, 55 },
		{ 359, 360, 55 },
		{ 359, 361, 55 },
		{ 359, 362, 55 },
		{ 359, 363, 55 },
		{ 359, 364, 55 },
		{ 359, 365, 55 },
		{ 359, 366, 55 },
		{ 359, 367, 55 },
		{ 359, 368, 55 },
		{ 359, 369, 55 },
		{ 359, 370, 55 },
		{ 359, 371, 55 },
		{ 359, 372, 55 },
		{ 359, 373, 55 },
		{ 359, 374, 55 },
		{ 359, 375, 55 },
		{ 359, 376, 55 },
		{ 359, 377, 55 },
		{ 359, 378, 55 },
		{ 359, 379, 55 },
		{ 359, 380, 55 },
		{ 359, 381, 55 },
		{ 359, 382, 55 },
		{ 359, 383, 55 },
		{ 359, 384, 55 },
		{ 359, 385, 55 },
		{ 359, 386, 55 },
		{ 359, 387, 55 },
		{ 359, 388, 55 },
		{ 359, 389, 55 },
		{ 359, 390, 55 },
		{ 359, 391, 55 },
		{ 359, 392, 55 },
		{ 360, 361, 55 },
		{ 360, 362, 55 },
		{ 360, 363, 55 },
		{ 360, 364, 55 },
		{ 360, 365, 55 },
		{ 360, 366, 55 },
		{ 360, 367, 55 },
		{ 360, 368, 55 },
		{ 360, 369, 55 },
		{ 360, 370, 55 },
		{ 360, 371, 55 },
		{ 360, 372, 55 },
		{ 360, 373, 55 },
		{ 360, 374, 55 },
		{ 360, 375, 55 },
		{ 360, 376, 55 },
		{ 360, 377, 55 },
		{ 360, 378, 55 },
		{ 360, 379, 55 },
		{ 360, 380, 55 },
		{ 360, 381, 55 },
		{ 360, 382, 55 },
		{ 360, 383, 55 },
		{ 360, 384, 55 },
		{ 360, 385, 55 },
		{ 360, 386, 55 },
		{ 360, 387, 55 },
		{ 360, 388, 55 },
		{ 360, 389, 55 },
		{ 360, 390, 55 },
		{ 360, 391, 55 },
		{ 360, 392, 55 },
		{ 360, 393, 55 },
		{ 360, 394, 55 },
		{ 361, 362, 55 },
		{ 361, 363, 55 },
		{ 361, 364, 55 },
		{ 361, 365, 55 },
		{ 361, 366, 55 },
		{ 361, 367, 55 },
		{ 361, 368, 55 },
		{ 361, 369, 55 },
		{ 361, 370, 55 },
		{ 361, 371, 55 },
		{ 361, 372, 55 },
		{ 361, 373, 55 },
		{ 361, 374, 55 },
		{ 361, 375, 55 },
		{ 361, 376, 55 },
		{ 361, 377, 55 },
		{ 361, 378, 55 },
		{ 361, 379, 55 },
		{ 361, 380, 55 },
		{ 361, 381, 55 },
		{ 361, 382, 55 },
		{ 361, 383, 55 },
		{ 361, 384, 55 },
		{ 361, 385, 55 },
		{ 361, 386, 55 },
		{ 361, 387, 55 },
		{ 361, 388, 55 },
		{ 361, 389, 55 },
		{ 361, 390, 55 },
		{ 361, 391, 55 },
		{ 361, 392, 55 },
		{ 361, 393, 55 },
		{ 361, 394, 55 },
		{ 361, 395, 55 },
		{ 362, 363, 55 },
		{ 362, 364, 55 },
		{ 362, 365, 55 },
		{ 362, 366, 55 },
		{ 362, 367, 55 },
		{ 362, 368, 55 },
		{ 362, 369, 55 },
		{ 362, 370, 55 },
		{ 362, 371, 55 },
		{ 362, 372, 55 },
		{ 362, 373, 55 },
		{ 362, 374, 55 },
		{ 362, 375, 55 },
		{ 362, 376, 55 },
		{ 362, 377, 55 },
		{ 362, 378, 55 },
		{ 362, 379, 55 },
		{ 362, 380, 55 },
		{ 362, 381, 55 },
		{ 362, 382, 55 },
		{ 362, 383, 55 },
		{ 362, 384, 55 },
		{ 362, 385, 55 },
		{ 362, 386, 55 },
		{ 362, 387, 55 },
		{ 362, 388, 55 },
		{ 362, 389, 55 },
		{ 362, 390, 55 },
		{ 362, 391, 55 },
		{ 362, 392, 55 },
		{ 362, 393, 55 },
		{ 362, 394, 55 },
		{ 362, 395, 55 },
		{ 362, 396, 55 },
		{ 363, 364, 55 },
		{ 363, 365, 55 },
		{ 363, 366, 55 },
		{ 363, 367, 55 },
		{ 363, 368, 55 },
		{ 363, 369, 55 },
		{ 363, 370, 55 },
		{ 363, 371, 55 },
		{ 363, 372, 55 },
		{ 363, 373, 55 },
		{ 363, 374, 55 },
		{ 363, 375, 55 },
		{ 363, 376, 55 },
		{ 363, 377, 55 },
		{ 363, 378, 55 },
		{ 363, 379, 55 },
		{ 363, 380, 55 },
		{ 363, 381, 55 },
		{ 363, 382, 55 },
		{ 363, 383, 55 },
		{ 363, 384, 55 },
		{ 363, 385, 55 },
		{ 363, 386, 55 },
		{ 363, 387, 55 },
		{ 363, 388, 55 },
		{ 363, 389, 55 },
		{ 363, 390, 55 },
		{ 363, 391, 55 },
		{ 363, 392, 55 },
		{ 363, 393, 55 },
		{ 363, 394, 55 },
		{ 363, 395, 55 },
		{ 363, 396, 55 },
		{ 363, 397, 55 },
		{ 363, 398, 55 },
		{ 364, 365, 55 },
		{ 364, 366, 55 },
		{ 364, 367, 55 },
		{ 364, 368, 55 },
		{ 364, 369, 55 },
		{ 364, 370, 55 },
		{ 364, 371, 55 },
		{ 364, 372, 55 },
		{ 364, 373, 55 },
		{ 364, 374, 55 },
		{ 364, 375, 55 },
		{ 364, 376, 55 },
		{ 364, 377, 55 },
		{ 364, 378, 55 },
		{ 364, 379, 55 },
		{ 364, 380, 55 },
		{ 364, 381, 55 },
		{ 364, 382, 55 },
		{ 364, 383, 55 },
		{ 364, 384, 55 },
		{ 364, 385, 55 },
		{ 364, 386, 55 },
		{ 364, 387, 55 },
		{ 364, 388, 55 },
		{ 364, 389, 55 },
		{ 364, 390, 55 },
		{ 364, 391, 55 },
		{ 364, 392, 55 },
		{ 364, 393, 55 },
		{ 364, 394, 55 },
		{ 364, 395, 55 },
		{ 364, 396, 55 },
		{ 364, 397, 55 },
		{ 364, 398, 55 },
		{ 364, 399, 55 },
		{ 365, 366, 55 },
		{ 365, 367, 55 },
		{ 365, 368, 55 },
		{ 365, 369, 55 },
		{ 365, 370, 55 },
		{ 365, 371, 55 },
		{ 365, 372, 55 },
		{ 365, 373, 55 },
		{ 365, 374, 55 },
		{ 365, 375, 55 },
		{ 365, 376, 55 },
		{ 365, 377, 55 },
		{ 365, 378, 55 },
		{ 365, 379, 55 },
		{ 365, 380, 55 },
		{ 365, 381, 55 },
		{ 365, 382, 55 },
		{ 365, 383, 55 },
		{ 365, 384, 55 },
		{ 365, 385, 55 },
		{ 365, 386, 55 },
		{ 365, 387, 55 },
		{ 365, 388, 55 },
		{ 365, 389, 55 },
		{ 365, 390, 55 },
		{ 365, 391, 55 },
		{ 365, 392, 55 },
		{ 365, 393, 55 },
		{ 365, 394, 55 },
		{ 365, 395, 55 },
		{ 365, 396, 55 },
		{ 365, 397, 55 },
		{ 365, 398, 55 },
		{ 365, 399, 55 },
		{ 365, 400, 55 },
		{ 366, 367, 55 },
		{ 366, 368, 55 },
		{ 366, 369, 55 },
		{ 366, 370, 55 },
		{ 366, 371, 55 },
		{ 366, 372, 55 },
		{ 366, 373, 55 },
		{ 366, 374, 55 },
		{ 366, 375, 55 },
		{ 366, 376, 55 },
		{ 366, 377, 55 },
		{ 366, 378, 55 },
		{ 366, 379, 55 },
		{ 366, 380, 55 },
		{ 366, 381, 55 },
		{ 366, 382, 55 },
		{ 366, 383, 55 },
		{ 366, 384, 55 },
		{ 366, 385, 55 },
		{ 366, 386, 55 },
		{ 366, 387, 55 },
		{ 366, 388, 55 },
		{ 366, 389, 55 },
		{ 366, 390, 55 },
		{ 366, 391, 55 },
		{ 366, 392, 55 },
		{ 366, 393, 55 },
		{ 366, 394, 55 },
		{ 366, 395, 55 },
		{ 366, 396, 55 },
		{ 366, 397, 55 },
		{ 366, 398, 55 },
		{ 366, 399, 55 },
		{ 366, 400, 55 },
		{ 366, 401, 55 },
		{ 366, 402, 55 },
		{ 367, 368, 55 },
		{ 367, 369, 55 },
		{ 367, 370, 55 },
		{ 367, 371, 55 },
		{ 367, 372, 55 },
		{ 367, 373, 55 },
		{ 367, 374, 55 },
		{ 367, 375, 55 },
		{ 367, 376, 55 },
		{ 367, 377, 55 },
		{ 367, 378, 55 },
		{ 367, 379, 55 },
		{ 367, 380, 55 },
		{ 367, 381, 55 },
		{ 367, 382, 55 },
		{ 367, 383, 55 },
		{ 367, 384, 55 },
		{ 367, 385, 55 },
		{ 367, 386, 55 },
		{ 367, 387, 55 },
		{ 367, 388, 55 },
		{ 367, 389, 55 },
		{ 367, 390, 55 },
		{ 367, 391, 55 },
		{ 367, 392, 55 },
		{ 367, 393, 55 },
		{ 367, 394, 55 },
		{ 367, 395, 55 },
		{ 367, 396, 55 },
		{ 367, 397, 55 },
		{ 367, 398, 55 },
		{ 367, 399, 55 },
		{ 367, 400, 55 },
		{ 367, 401, 55 },
		{ 367, 402, 55 },
		{ 367, 403, 55 },
		{ 368, 369, 55 },
		{ 368, 370, 55 },
		{ 368, 371, 55 },
		{ 368, 372, 55 },
		{ 368, 373, 55 },
		{ 368, 374, 55 },
		{ 368, 375, 55 },
		{ 368, 376, 55 },
		{ 368, 377, 55 },
		{ 368, 378, 55 },
		{ 368, 379, 55 },
		{ 368, 380, 55 },
		{ 368, 381, 55 },
		{ 368, 382, 55 },
		{ 368, 383, 55 },
		{ 368, 384, 55 },
		{ 368, 385, 55 },
		{ 368, 386, 55 },
		{ 368, 387, 55 },
		{ 368, 388, 55 },
		{ 368, 389, 55 },
		{ 368, 390, 55 },
		{ 368, 391, 55 },
		{ 368, 392, 55 },
		{ 368, 393, 55 },
		{ 368, 394, 55 },
		{ 368, 395, 55 },
		{ 368, 396, 55 },
		{ 368, 397, 55 },
		{ 368, 398, 55 },
		{ 368, 399, 55 },
		{ 368, 400, 55 },
		{ 368, 401, 55 },
		{ 368, 402, 55 },
		{ 368, 403, 55 },
		{ 368, 404, 55 },
		{ 369, 370, 55 },
		{ 369, 371, 55 },
		{ 369, 372, 55 },
		{ 369, 373, 55 },
		{ 369, 374, 55 },
		{ 369, 375, 55 },
		{ 369, 376, 55 },
		{ 369, 377, 55 },
		{ 369, 378, 55 },
		{ 369, 379, 55 },
		{ 369, 380, 55 },
		{ 369, 381, 55 },
		{ 369, 382, 55 },
		{ 369, 383, 55 },
		{ 369, 384, 55 },
		{ 369, 385, 55 },
		{ 369, 386, 55 },
		{ 369, 387, 55 },
		{ 369, 388, 55 },
		{ 369, 389, 55 },
		{ 369, 390, 55 },
		{ 369, 391, 55 },
		{ 369, 392, 55 },
		{ 369, 393, 55 },
		{ 369, 394, 55 },
		{ 369, 395, 55 },
		{ 369, 396, 55 },
		{ 369, 397, 55 },
		{ 369, 398, 55 },
		{ 369, 399, 55 },
		{ 369, 400, 55 },
		{ 369, 401, 55 },
		{ 369, 402, 55 },
		{ 369, 403, 55 },
		{ 369, 404, 55 },
		{ 369, 405, 55 },
		{ 369, 406, 55 },
		{ 370, 371, 55 },
		{ 370, 372, 55 },
		{ 370, 373, 55 },
		{ 370, 374, 55 },
		{ 370, 375, 55 },
		{ 370, 376, 55 },
		{ 370, 377, 55 },
		{ 370, 378, 55 },
		{ 370, 379, 55 },
		{ 370, 380, 55 },
		{ 370, 381, 55 },
		{ 370, 382, 55 },
		{ 370, 383, 55 },
		{ 370, 384, 55 },
		{ 370, 385, 55 },
		{ 370, 386, 55 },
		{ 370, 387, 55 },
		{ 370, 388, 55 },
		{ 370, 389, 55 },
		{ 370, 390, 55 },
		{ 370, 391, 55 },
		{ 370, 392, 55 },
		{ 370, 393, 55 },
		{ 370, 394, 55 },
		{ 370, 395, 55 },
		{ 370, 396, 55 },
		{ 370, 397, 55 },
		{ 370, 398, 55 },
		{ 370, 399, 55 },
		{ 370, 400, 55 },
		{ 370, 401, 55 },
		{ 370, 402, 55 },
		{ 370, 403, 55 },
		{ 370, 404, 55 },
		{ 370, 405, 55 },
		{ 370, 406, 55 },
		{ 370, 407, 55 },
		{ 371, 372, 55 },
		{ 371, 373, 55 },
		{ 371, 374, 55 },
		{ 371, 375, 55 },
		{ 371, 376, 55 },
		{ 371, 377, 55 },
		{ 371, 378, 55 },
		{ 371, 379, 55 },
		{ 371, 380, 55 },
		{ 371, 381, 55 },
		{ 371, 382, 55 },
		{ 371, 383, 55 },
		{ 371, 384, 55 },
		{ 371, 385, 55 },
		{ 371, 386, 55 },
		{ 371, 387, 55 },
		{ 371, 388, 55 },
		{ 371, 389, 55 },
		{ 371, 390, 55 },
		{ 371, 391, 55 },
		{ 371, 392, 55 },
		{ 371, 393, 55 },
		{ 371, 394, 55 },
		{ 371, 395, 55 },
		{ 371, 396, 55 },
		{ 371, 397, 55 },
		{ 371, 398, 55 },
		{ 371, 399, 55 },
		{ 371, 400, 55 },
		{ 371, 401, 55 },
		{ 371, 402, 55 },
		{ 371, 403, 55 },
		{ 371, 404, 55 },
		{ 371, 405, 55 },
		{ 371, 406, 55 },
		{ 371, 407, 55 },
		{ 371, 408, 55 },
		{ 372, 373, 55 },
		{ 372, 374, 55 },
		{ 372, 375, 55 },
		{ 372, 376, 55 },
		{ 372, 377, 55 },
		{ 372, 378, 55 },
		{ 372, 379, 55 },
		{ 372, 380, 55 },
		{ 372, 381, 55 },
		{ 372, 382, 55 },
		{ 372, 383, 55 },
		{ 372, 384, 55 },
		{ 372, 385, 55 },
		{ 372, 386, 55 },
		{ 372, 387, 55 },
		{ 372, 388, 55 },
		{ 372, 389, 55 },
		{ 372, 390, 55 },
		{ 372, 391, 55 },
		{ 372, 392, 55 },
		{ 372, 393, 55 },
		{ 372, 394, 55 },
		{ 372, 395, 55 },
		{ 372, 396, 55 },
		{ 372, 397, 55 },
		{ 372, 398, 55 },
		{ 372, 399, 55 },
		{ 372, 400, 55 },
		{ 372, 401, 55 },
		{ 372, 402, 55 },
		{ 372, 403, 55 },
		{ 372, 404, 55 },
		{ 372, 405, 55 },
		{ 372, 406, 55 },
		{ 372, 407, 55 },
		{ 372, 408, 55 },
		{ 372, 409, 55 },
		{ 372, 410, 55 },
		{ 373, 374, 55 },
		{ 373, 375, 55 },
		{ 373, 376, 55 },
		{ 373, 377, 55 },
		{ 373, 378, 55 },
		{ 373, 379, 55 },
		{ 373, 380, 55 },
		{ 373, 381, 55 },
		{ 373, 382, 55 },
		{ 373, 383, 55 },
		{ 373, 384, 55 },
		{ 373, 385, 55 },
		{ 373, 386, 55 },
		{ 373, 387, 55 },
		{ 373, 388, 55 },
		{ 373, 389, 55 },
		{ 373, 390, 55 },
		{ 373, 391, 55 },
		{ 373, 392, 55 },
		{ 373, 393, 55 },
		{ 373, 394, 55 },
		{ 373, 395, 55 },
		{ 373, 396, 55 },
		{ 373, 397, 55 },
		{ 373, 398, 55 },
		{ 373, 399, 55 },
		{ 373, 400, 55 },
		{ 373, 401, 55 },
		{ 373, 402, 55 },
		{ 373, 403, 55 },
		{ 373, 404, 55 },
		{ 373, 405, 55 },
		{ 373, 406, 55 },
		{ 373, 407, 55 },
		{ 373, 408, 55 },
		{ 373, 409, 55 },
		{ 373, 410, 55 },
		{ 373, 411, 55 },
		{ 374, 375, 55 },
		{ 374, 376, 55 },
		{ 374, 377, 55 },
		{ 374, 378, 55 },
		{ 374, 379, 55 },
		{ 374, 380, 55 },
		{ 374, 381, 55 },
		{ 374, 382, 55 },
		{ 374, 383, 55 },
		{ 374, 384, 55 },
		{ 374, 385, 55 },
		{ 374, 386, 55 },
		{ 374, 387, 55 },
		{ 374, 388, 55 },
		{ 374, 389, 55 },
		{ 374, 390, 55 },
		{ 374, 391, 55 },
		{ 374, 392, 55 },
		{ 374, 393, 55 },
		{ 374, 394, 55 },
		{ 374, 395, 55 },
		{ 374, 396, 55 },
		{ 374, 397, 55 },
		{ 374, 398, 55 },
		{ 374, 399, 55 },
		{ 374, 400, 55 },
		{ 374, 401, 55 },
		{ 374, 402, 55 },
		{ 374, 403, 55 },
		{ 374, 404, 55 },
		{ 374, 405, 55 },
		{ 374, 406, 55 },
		{ 374, 407, 55 },
		{ 374, 408, 55 },
		{ 374, 409, 55 },
		{ 374, 410, 55 },
		{ 374, 411, 55 },
		{ 374, 412, 55 },
		{ 375, 376, 55 },
		{ 375, 377, 55 },
		{ 375, 378, 55 },
		{ 375, 379, 55 },
		{ 375, 380, 55 },
		{ 375, 381, 55 },
		{ 375, 382, 55 },
		{ 375, 383, 55 },
		{ 375, 384, 55 },
		{ 375, 385, 55 },
		{ 375, 386, 55 },
		{ 375, 387, 55 },
		{ 375, 388, 55 },
		{ 375, 389, 55 },
		{ 375, 390, 55 },
		{ 375, 391, 55 },
		{ 375, 392, 55 },
		{ 375, 393, 55 },
		{ 375, 394, 55 },
		{ 375, 395, 55 },
		{ 375, 396, 55 },
		{ 375, 397, 55 },
		{ 375, 398, 55 },
		{ 375, 399, 55 },
		{ 375, 400, 55 },
		{ 375, 401, 55 },
		{ 375, 402, 55 },
		{ 375, 403, 55 },
		{ 375, 404, 55 },
		{ 375, 405, 55 },
		{ 375, 406, 55 },
		{ 375, 407, 55 },
		{ 375, 408, 55 },
		{ 375, 409, 55 },
		{ 375, 410, 55 },
		{ 375, 411, 55 },
		{ 375, 412, 55 },
		{ 375, 413, 55 },
		{ 375, 414, 55 },
		{ 376, 377, 55 },
		{ 376, 378, 55 },
		{ 376, 379, 55 },
		{ 376, 380, 55 },
		{ 376, 381, 55 },
		{ 376, 382, 55 },
		{ 376, 383, 55 },
		{ 376, 384, 55 },
		{ 376, 385, 55 },
		{ 376, 386, 55 },
		{ 376, 387, 55 },
		{ 376, 388, 55 },
		{ 376, 389, 55 },
		{ 376, 390, 55 },
		{ 376, 391, 55 },
		{ 376, 392, 55 },
		{ 376, 393, 55 },
		{ 376, 394, 55 },
		{ 376, 395, 55 },
		{ 376, 396, 55 },
		{ 376, 397, 55 },
		{ 376, 398, 55 },
		{ 376, 399, 55 },
		{ 376, 400, 55 },
		{ 376, 401, 55 },
		{ 376, 402, 55 },
		{ 376, 403, 55 },
		{ 376, 404, 55 },
		{ 376, 405, 55 },
		{ 376, 406, 55 },
		{ 376, 407, 55 },
		{ 376, 408, 55 },
		{ 376, 409, 55 },
		{ 376, 410, 55 },
		{ 376, 411, 55 },
		{ 376, 412, 55 },
		{ 376, 413, 55 },
		{ 376, 414, 55 },
		{ 376, 415, 55 },
		{ 377, 378, 55 },
		{ 377, 379, 55 },
		{ 377, 380, 55 },
		{ 377, 381, 55 },
		{ 377, 382, 55 },
		{ 377, 383, 55 },
		{ 377, 384, 55 },
		{ 377, 385, 55 },
		{ 377, 386, 55 },
		{ 377, 387, 55 },
		{ 377, 388, 55 },
		{ 377, 389, 55 },
		{ 377, 390, 55 },
		{ 377, 391, 55 },
		{ 377, 392, 55 },
		{ 377, 393, 55 },
		{ 377, 394, 55 },
		{ 377, 395, 55 },
		{ 377, 396, 55 },
		{ 377, 397, 55 },
		{ 377, 398, 55 },
		{ 377, 399, 55 },
		{ 377, 400, 55 },
		{ 377, 401, 55 },
		{ 377, 402, 55 },
		{ 377, 403, 55 },
		{ 377, 404, 55 },
		{ 377, 405, 55 },
		{ 377, 406, 55 },
		{ 377, 407, 55 },
		{ 377, 408, 55 },
		{ 377, 409, 55 },
		{ 377, 410, 55 },
		{ 377, 411, 55 },
		{ 377, 412, 55 },
		{ 377, 413, 55 },
		{ 377, 414, 55 },
		{ 377, 415, 55 },
		{ 377, 416, 55 },
		{ 378, 379, 55 },
		{ 378, 380, 55 },
		{ 378, 381, 55 },
		{ 378, 382, 55 },
		{ 378, 383, 55 },
		{ 378, 384, 55 },
		{ 378, 385, 55 },
		{ 378, 386, 55 },
		{ 378, 387, 55 },
		{ 378, 388, 55 },
		{ 378, 389, 55 },
		{ 378, 390, 55 },
		{ 378, 391, 55 },
		{ 378, 392, 55 },
		{ 378, 393, 55 },
		{ 378, 394, 55 },
		{ 378, 395, 55 },
		{ 378, 396, 55 },
		{ 378, 397, 55 },
		{ 378, 398, 55 },
		{ 378, 399, 55 },
		{ 378, 400, 55 },
		{ 378, 401, 55 },
		{ 378, 402, 55 },
		{ 378, 403, 55 },
		{ 378, 404, 55 },
		{ 378, 405, 55 },
		{ 378, 406, 55 },
		{ 378, 407, 55 },
		{ 378, 408, 55 },
		{ 378, 409, 55 },
		{ 378, 410, 55 },
		{ 378, 411, 55 },
		{ 378, 412, 55 },
		{ 378, 413, 55 },
		{ 378, 414, 55 },
		{ 378, 415, 55 },
		{ 378, 416, 55 },
		{ 378, 417, 55 },
		{ 378, 418, 55 },
		{ 379, 380, 55 },
		{ 379, 381, 55 },
		{ 379, 382, 55 },
		{ 379, 383, 55 },
		{ 379, 384, 55 },
		{ 379, 385, 55 },
		{ 379, 386, 55 },
		{ 379, 387, 55 },
		{ 379, 388, 55 },
		{ 379, 389, 55 },
		{ 379, 390, 55 },
		{ 379, 391, 55 },
		{ 379, 392, 55 },
		{ 379, 393, 55 },
		{ 379, 394, 55 },
		{ 379, 395, 55 },
		{ 379, 396, 55 },
		{ 379, 397, 55 },
		{ 379, 398, 55 },
		{ 379, 399, 55 },
		{ 379, 400, 55 },
		{ 379, 401, 55 },
		{ 379, 402, 55 },
		{ 379, 403, 55 },
		{ 379, 404, 55 },
		{ 379, 405, 55 },
		{ 379, 406, 55 },
		{ 379, 407, 55 },
		{ 379, 408, 55 },
		{ 379, 409, 55 },
		{ 379, 410, 55 },
		{ 379, 411, 55 },
		{ 379, 412, 55 },
		{ 379, 413, 55 },
		{ 379, 414, 55 },
		{ 379, 415, 55 },
		{ 379, 416, 55 },
		{ 379, 417, 55 },
		{ 379, 418, 55 },
		{ 379, 419, 55 },
		{ 380, 381, 55 },
		{ 380, 382, 55 },
		{ 380, 383, 55 },
		{ 380, 384, 55 },
		{ 380, 385, 55 },
		{ 380, 386, 55 },
		{ 380, 387, 55 },
		{ 380, 388, 55 },
		{ 380, 389, 55 },
		{ 380, 390, 55 },
		{ 380, 391, 55 },
		{ 380, 392, 55 },
		{ 380, 393, 55 },
		{ 380, 394, 55 },
		{ 380, 395, 55 },
		{ 380, 396, 55 },
		{ 380, 397, 55 },
		{ 380, 398, 55 },
		{ 380, 399, 55 },
		{ 380, 400, 55 },
		{ 380, 401, 55 },
		{ 380, 402, 55 },
		{ 380, 403, 55 },
		{ 380, 404, 55 },
		{ 380, 405, 55 },
		{ 380, 406, 55 },
		{ 380, 407, 55 },
		{ 380, 408, 55 },
		{ 380, 409, 55 },
		{ 380, 410, 55 },
		{ 380, 411, 55 },
		{ 380, 412, 55 },
		{ 380, 413, 55 },
		{ 380, 414, 55 },
		{ 380, 415, 55 },
		{ 380, 416, 55 },
		{ 380, 417, 55 },
		{ 380, 418, 55 },
		{ 380, 419, 55 },
		{ 380, 420, 55 },
		{ 381, 382, 55 },
		{ 381, 383, 55 },
		{ 381, 384, 55 },
		{ 381, 385, 55 },
		{ 381, 386, 55 },
		{ 381, 387, 55 },
		{ 381, 388, 55 },
		{ 381, 389, 55 },
		{ 381, 390, 55 },
		{ 381, 391, 55 },
		{ 381, 392, 55 },
		{ 381, 393, 55 },
		{ 381, 394, 55 },
		{ 381, 395, 55 },
		{ 381, 396, 55 },
		{ 381, 397, 55 },
		{ 381, 398, 55 },
		{ 381, 399, 55 },
		{ 381, 400, 55 },
		{ 381, 401, 55 },
		{ 381, 402, 55 },
		{ 381, 403, 55 },
		{ 381, 404, 55 },
		{ 381, 405, 55 },
		{ 381, 406, 55 },
		{ 381, 407, 55 },
		{ 381, 408, 55 },
		{ 381, 409, 55 },
		{ 381, 410, 55 },
		{ 381, 411, 55 },
		{ 381, 412, 55 },
		{ 381, 413, 55 },
		{ 381, 414, 55 },
		{ 381, 415, 55 },
		{ 381, 416, 55 },
		{ 381, 417, 55 },
		{ 381, 418, 55 },
		{ 381, 419, 55 },
		{ 381, 420, 55 },
		{ 381, 421, 55 },
		{ 381, 422, 55 },
		{ 382, 383, 55 },
		{ 382, 384, 55 },
		{ 382, 385, 55 },
		{ 382, 386, 55 },
		{ 382, 387, 55 },
		{ 382, 388, 55 },
		{ 382, 389, 55 },
		{ 382, 390, 55 },
		{ 382, 391, 55 },
		{ 382, 392, 55 },
		{ 382, 393, 55 },
		{ 382, 394, 55 },
		{ 382, 395, 55 },
		{ 382, 396, 55 },
		{ 382, 397, 55 },
		{ 382, 398, 55 },
		{ 382, 399, 55 },
		{ 382, 400, 55 },
		{ 382, 401, 55 },
		{ 382, 402, 55 },
		{ 382, 403, 55 },
		{ 382, 404, 55 },
		{ 382, 405, 55 },
		{ 382, 406, 55 },
		{ 382, 407, 55 },
		{ 382, 408, 55 },
		{ 382, 409, 55 },
		{ 382, 410, 55 },
		{ 382, 411, 55 },
		{ 382, 412, 55 },
		{ 382, 413, 55 },
		{ 382, 414, 55 },
		{ 382, 415, 55 },
		{ 382, 416, 55 },
		{ 382, 417, 55 },
		{ 382, 418, 55 },
		{ 382, 419, 55 },
		{ 382, 420, 55 },
		{ 382, 421, 55 },
		{ 382, 422, 55 },
		{ 382, 423, 55 },
		{ 383, 384, 55 },
		{ 383, 385, 55 },
		{ 383, 386, 55 },
		{ 383, 387, 55 },
		{ 383, 388, 55 },
		{ 383, 389, 55 },
		{ 383, 390, 55 },
		{ 383, 391, 55 },
		{ 383, 392, 55 },
		{ 383, 393, 55 },
		{ 383, 394, 55 },
		{ 383, 395, 55 },
		{ 383, 396, 55 },
		{ 383, 397, 55 },
		{ 383, 398, 55 },
		{ 383, 399, 55 },
		{ 383, 400, 55 },
		{ 383, 401, 55 },
		{ 383, 402, 55 },
		{ 383, 403, 55 },
		{ 383, 404, 55 },
		{ 383, 405, 55 },
		{ 383, 406, 55 },
		{ 383, 407, 55 },
		{ 383, 408, 55 },
		{ 383, 409, 55 },
		{ 383, 410, 55 },
		{ 383, 411, 55 },
		{ 383, 412, 55 },
		{ 383, 413, 55 },
		{ 383, 414, 55 },
		{ 383, 415, 55 },
		{ 383, 416, 55 },
		{ 383, 417, 55 },
		{ 383, 418, 55 },
		{ 383, 419, 55 },
		{ 383, 420, 55 },
		{ 383, 421, 55 },
		{ 383, 422, 55 },
		{ 383, 423, 55 },
		{ 383, 424, 55 },
		{ 384, 385, 55 },
		{ 384, 386, 55 },
		{ 384, 387, 55 },
		{ 384, 388, 55 },
		{ 384, 389, 55 },
		{ 384, 390, 55 },
		{ 384, 391, 55 },
		{ 384, 392, 55 },
		{ 384, 393, 55 },
		{ 384, 394, 55 },
		{ 384, 395, 55 },
		{ 384, 396, 55 },
		{ 384, 397, 55 },
		{ 384, 398, 55 },
		{ 384, 399, 55 },
		{ 384, 400, 55 },
		{ 384, 401, 55 },
		{ 384, 402, 55 },
		{ 384, 403, 55 },
		{ 384, 404, 55 },
		{ 384, 405, 55 },
		{ 384, 406, 55 },
		{ 384, 407, 55 },
		{ 384, 408, 55 },
		{ 384, 409, 55 },
		{ 384, 410, 55 },
		{ 384, 411, 55 },
		{ 384, 412, 55 },
		{ 384, 413, 55 },
		{ 384, 414, 55 },
		{ 384, 415, 55 },
		{ 384, 416, 55 },
		{ 384, 417, 55 },
		{ 384, 418, 55 },
		{ 384, 419, 55 },
		{ 384, 420, 55 },
		{ 384, 421, 55 },
		{ 384, 422, 55 },
		{ 384, 423, 55 },
		{ 384, 424, 55 },
		{ 384, 425, 55 },
		{ 384, 426, 55 },
		{ 385, 386, 55 },
		{ 385, 387, 55 },
		{ 385, 388, 55 },
		{ 385, 389, 55 },
		{ 385, 390, 55 },
		{ 385, 391, 55 },
		{ 385, 392, 55 },
		{ 385, 393, 55 },
		{ 385, 394, 55 },
		{ 385, 395, 55 },
		{ 385, 396, 55 },
		{ 385, 397, 55 },
		{ 385, 398, 55 },
		{ 385, 399, 55 },
		{ 385, 400, 55 },
		{ 385, 401, 55 },
		{ 385, 402, 55 },
		{ 385, 403, 55 },
		{ 385, 404, 55 },
		{ 385, 405, 55 },
		{ 385, 406, 55 },
		{ 385, 407, 55 },
		{ 385, 408, 55 },
		{ 385, 409, 55 },
		{ 385, 410, 55 },
		{ 385, 411, 55 },
		{ 385, 412, 55 },
		{ 385, 413, 55 },
		{ 385, 414, 55 },
		{ 385, 415, 55 },
		{ 385, 416, 55 },
		{ 385, 417, 55 },
		{ 385, 418, 55 },
		{ 385, 419, 55 },
		{ 385, 420, 55 },
		{ 385, 421, 55 },
		{ 385, 422, 55 },
		{ 385, 423, 55 },
		{ 385, 424, 55 },
		{ 385, 425, 55 },
		{ 385, 426, 55 },
		{ 385, 427, 55 },
		{ 386, 387, 55 },
		{ 386, 388, 55 },
		{ 386, 389, 55 },
		{ 386, 390, 55 },
		{ 386, 391, 55 },
		{ 386, 392, 55 },
		{ 386, 393, 55 },
		{ 386, 394, 55 },
		{ 386, 395, 55 },
		{ 386, 396, 55 },
		{ 386, 397, 55 },
		{ 386, 398, 55 },
		{ 386, 399, 55 },
		{ 386, 400, 55 },
		{ 386, 401, 55 },
		{ 386, 402, 55 },
		{ 386, 403, 55 },
		{ 386, 404, 55 },
		{ 386, 405, 55 },
		{ 386, 406, 55 },
		{ 386, 407, 55 },
		{ 386, 408, 55 },
		{ 386, 409, 55 },
		{ 386, 410, 55 },
		{ 386, 411, 55 },
		{ 386, 412, 55 },
		{ 386, 413, 55 },
		{ 386, 414, 55 },
		{ 386, 415, 55 },
		{ 386, 416, 55 },
		{ 386, 417, 55 },
		{ 386, 418, 55 },
		{ 386, 419, 55 },
		{ 386, 420, 55 },
		{ 386, 421, 55 },
		{ 386, 422, 55 },
		{ 386, 423, 55 },
		{ 386, 424, 55 },
		{ 386, 425, 55 },
		{ 386, 426, 55 },
		{ 386, 427, 55 },
		{ 386, 428, 55 },
		{ 387, 388, 55 },
		{ 387, 389, 55 },
		{ 387, 390, 55 },
		{ 387, 391, 55 },
		{ 387, 392, 55 },
		{ 387, 393, 55 },
		{ 387, 394, 55 },
		{ 387, 395, 55 },
		{ 387, 396, 55 },
		{ 387, 397, 55 },
		{ 387, 398, 55 },
		{ 387, 399, 55 },
		{ 387, 400, 55 },
		{ 387, 401, 55 },
		{ 387, 402, 55 },
		{ 387, 403, 55 },
		{ 387, 404, 55 },
		{ 387, 405, 55 },
		{ 387, 406, 55 },
		{ 387, 407, 55 },
		{ 387, 408, 55 },
		{ 387, 409, 55 },
		{ 387, 410, 55 },
		{ 387, 411, 55 },
		{ 387, 412, 55 },
		{ 387, 413, 55 },
		{ 387, 414, 55 },
		{ 387, 415, 55 },
		{ 387, 416, 55 },
		{ 387, 417, 55 },
		{ 387, 418, 55 },
		{ 387, 419, 55 },
		{ 387, 420, 55 },
		{ 387, 421, 55 },
		{ 387, 422, 55 },
		{ 387, 423, 55 },
		{ 387, 424, 55 },
		{ 387, 425, 55 },
		{ 387, 426, 55 },
		{ 387, 427, 55 },
		{ 387, 428, 55 },
		{ 387, 429, 55 },
		{ 387, 430, 55 },
		{ 388, 389, 55 },
		{ 388, 390, 55 },
		{ 388, 391, 55 },
		{ 388, 392, 55 },
		{ 388, 393, 55 },
		{ 388, 394, 55 },
		{ 388, 395, 55 },
		{ 388, 396, 55 },
		{ 388, 397, 55 },
		{ 388, 398, 55 },
		{ 388, 399, 55 },
		{ 388, 400, 55 },
		{ 388, 401, 55 },
		{ 388, 402, 55 },
		{ 388, 403, 55 },
		{ 388, 404, 55 },
		{ 388, 405, 55 },
		{ 388, 406, 55 },
		{ 388, 407, 55 },
		{ 388, 408, 55 },
		{ 388, 409, 55 },
		{ 388, 410, 55 },
		{ 388, 411, 55 },
		{ 388, 412, 55 },
		{ 388, 413, 55 },
		{ 388, 414, 55 },
		{ 388, 415, 55 },
		{ 388, 416, 55 },
		{ 388, 417, 55 },
		{ 388, 418, 55 },
		{ 388, 419, 55 },
		{ 388, 420, 55 },
		{ 388, 421, 55 },
		{ 388, 422, 55 },
		{ 388, 423, 55 },
		{ 388, 424, 55 },
		{ 388, 425, 55 },
		{ 388, 426, 55 },
		{ 388, 427, 55 },
		{ 388, 428, 55 },
		{ 388, 429, 55 },
		{ 388, 430, 55 },
		{ 388, 431, 55 },
		{ 389, 390, 55 },
		{ 389, 391, 55 },
		{ 389, 392, 55 },
		{ 389, 393, 55 },
		{ 389, 394, 55 },
		{ 389, 395, 55 },
		{ 389, 396, 55 },
		{ 389, 397, 55 },
		{ 389, 398, 55 },
		{ 389, 399, 55 },
		{ 389, 400, 55 },
		{ 389, 401, 55 },
		{ 389, 402, 55 },
		{ 389, 403, 55 },
		{ 389, 404, 55 },
		{ 389, 405, 55 },
		{ 389, 406, 55 },
		{ 389, 407, 55 },
		{ 389, 408, 55 },
		{ 389, 409, 55 },
		{ 389, 410, 55 },
		{ 389, 411, 55 },
		{ 389, 412, 55 },
		{ 389, 413, 55 },
		{ 389, 414, 55 },
		{ 389, 415, 55 },
		{ 389, 416, 55 },
		{ 389, 417, 55 },
		{ 389, 418, 55 },
		{ 389, 419, 55 },
		{ 389, 420, 55 },
		{ 389, 421, 55 },
		{ 389, 422, 55 },
		{ 389, 423, 55 },
		{ 389, 424, 55 },
		{ 389, 425, 55 },
		{ 389, 426, 55 },
		{ 389, 427, 55 },
		{ 389, 428, 55 },
		{ 389, 429, 55 },
		{ 389, 430, 55 },
		{ 389, 431, 55 },
		{ 389, 432, 55 },
		{ 390, 391, 55 },
		{ 390, 392, 55 },
		{ 390, 393, 55 },
		{ 390, 394, 55 },
		{ 390, 395, 55 },
		{ 390, 396, 55 },
		{ 390, 397, 55 },
		{ 390, 398, 55 },
		{ 390, 399, 55 },
		{ 390, 400, 55 },
		{ 390, 401, 55 },
		{ 390, 402, 55 },
		{ 390, 403, 55 },
		{ 390, 404, 55 },
		{ 390, 405, 55 },
		{ 390, 406, 55 },
		{ 390, 407, 55 },
		{ 390, 408, 55 },
		{ 390, 409, 55 },
		{ 390, 410, 55 },
		{ 390, 411, 55 },
		{ 390, 412, 55 },
		{ 390, 413, 55 },
		{ 390, 414, 55 },
		{ 390, 415, 55 },
		{ 390, 416, 55 },
		{ 390, 417, 55 },
		{ 390, 418, 55 },
		{ 390, 419, 55 },
		{ 390, 420, 55 },
		{ 390, 421, 55 },
		{ 390, 422, 55 },
		{ 390, 423, 55 },
		{ 390, 424, 55 },
		{ 390, 425, 55 },
		{ 390, 426, 55 },
		{ 390, 427, 55 },
		{ 390, 428, 55 },
		{ 390, 429, 55 },
		{ 390, 430, 55 },
		{ 390, 431, 55 },
		{ 390, 432, 55 },
		{ 390, 433, 55 },
		{ 390, 434, 55 },
		{ 391, 392, 55 },
		{ 391, 393, 55 },
		{ 391, 394, 55 },
		{ 391, 395, 55 },
		{ 391, 396, 55 },
		{ 391, 397, 55 },
		{ 391, 398, 55 },
		{ 391, 399, 55 },
		{ 391, 400, 55 },
		{ 391, 401, 55 },
		{ 391, 402, 55 },
		{ 391, 403, 55 },
		{ 391, 404, 55 },
		{ 391, 405, 55 },
		{ 391, 406, 55 },
		{ 391, 407, 55 },
		{ 391, 408, 55 },
		{ 391, 409, 55 },
		{ 391, 410, 55 },
		{ 391, 411, 55 },
		{ 391, 412, 55 },
		{ 391, 413, 55 },
		{ 391, 414, 55 },
		{ 391, 415, 55 },
		{ 391, 416, 55 },
		{ 391, 417, 55 },
		{ 391, 418, 55 },
		{ 391, 419, 55 },
		{ 391, 420, 55 },
		{ 391, 421, 55 },
		{ 391, 422, 55 },
		{ 391, 423, 55 },
		{ 391, 424, 55 },
		{ 391, 425, 55 },
		{ 391, 426, 55 },
		{ 391, 427, 55 },
		{ 391, 428, 55 },
		{ 391, 429, 55 },
		{ 391, 430, 55 },
		{ 391, 431, 55 },
		{ 391, 432, 55 },
		{ 391, 433, 55 },
		{ 391, 434, 55 },
		{ 391, 435, 55 },
		{ 392, 393, 55 },
		{ 392, 394, 55 },
		{ 392, 395, 55 },
		{ 392, 396, 55 },
		{ 392, 397, 55 },
		{ 392, 398, 55 },
		{ 392, 399, 55 },
		{ 392, 400, 55 },
		{ 392, 401, 55 },
		{ 392, 402, 55 },
		{ 392, 403, 55 },
		{ 392, 404, 55 },
		{ 392, 405, 55 },
		{ 392, 406, 55 },
		{ 392, 407, 55 },
		{ 392, 408, 55 },
		{ 392, 409, 55 },
		{ 392, 410, 55 },
		{ 392, 411, 55 },
		{ 392, 412, 55 },
		{ 392, 413, 55 },
		{ 392, 414, 55 },
		{ 392, 415, 55 },
		{ 392, 416, 55 },
		{ 392, 417, 55 },
		{ 392, 418, 55 },
		{ 392, 419, 55 },
		{ 392, 420, 55 },
		{ 392, 421, 55 },
		{ 392, 422, 55 },
		{ 392, 423, 55 },
		{ 392, 424, 55 },
		{ 392, 425, 55 },
		{ 392, 426, 55 },
		{ 392, 427, 55 },
		{ 392, 428, 55 },
		{ 392, 429, 55 },
		{ 392, 430, 55 },
		{ 392, 431, 55 },
		{ 392, 432, 55 },
		{ 392, 433, 55 },
		{ 392, 434, 55 },
		{ 392, 435, 55 },
		{ 392, 436, 55 },
		{ 393, 394, 55 },
		{ 393, 395, 55 },
		{ 393, 396, 55 },
		{ 393, 397, 55 },
		{ 393, 398, 55 },
		{ 393, 399, 55 },
		{ 393, 400, 55 },
		{ 393, 401, 55 },
		{ 393, 402, 55 },
		{ 393, 403, 55 },
		{ 393, 404, 55 },
		{ 393, 405, 55 },
		{ 393, 406, 55 },
		{ 393, 407, 55 },
		{ 393, 408, 55 },
		{ 393, 409, 55 },
		{ 393, 410, 55 },
		{ 393, 411, 55 },
		{ 393, 412, 55 },
		{ 393, 413, 55 },
		{ 393, 414, 55 },
		{ 393, 415, 55 },
		{ 393, 416, 55 },
		{ 393, 417, 55 },
		{ 393, 418, 55 },
		{ 393, 419, 55 },
		{ 393, 420, 55 },
		{ 393, 421, 55 },
		{ 393, 422, 55 },
		{ 393, 423, 55 },
		{ 393, 424, 55 },
		{ 393, 425, 55 },
		{ 393, 426, 55 },
		{ 393, 427, 55 },
		{ 393, 428, 55 },
		{ 393, 429, 55 },
		{ 393, 430, 55 },
		{ 393, 431, 55 },
		{ 393, 432, 55 },
		{ 393, 433, 55 },
		{ 393, 434, 55 },
		{ 393, 435, 55 },
		{ 393, 436, 55 },
		{ 393, 437, 55 },
		{ 393, 438, 55 },
		{ 394, 395, 55 },
		{ 394, 396, 55 },
		{ 394, 397, 55 },
		{ 394, 398, 55 },
		{ 394, 399, 55 },
		{ 394, 400, 55 },
		{ 394, 401, 55 },
		{ 394, 402, 55 },
		{ 394, 403, 55 },
		{ 394, 404, 55 },
		{ 394, 405, 55 },
		{ 394, 406, 55 },
		{ 394, 407, 55 },
		{ 394, 408, 55 },
		{ 394, 409, 55 },
		{ 394, 410, 55 },
		{ 394, 411, 55 },
		{ 394, 412, 55 },
		{ 394, 413, 55 },
		{ 394, 414, 55 },
		{ 394, 415, 55 },
		{ 394, 416, 55 },
		{ 394, 417, 55 },
		{ 394, 418, 55 },
		{ 394, 419, 55 },
		{ 394, 420, 55 },
		{ 394, 421, 55 },
		{ 394, 422, 55 },
		{ 394, 423, 55 },
		{ 394, 424, 55 },
		{ 394, 425, 55 },
		{ 394, 426, 55 },
		{ 394, 427, 55 },
		{ 394, 428, 55 },
		{ 394, 429, 55 },
		{ 394, 430, 55 },
		{ 394, 431, 55 },
		{ 394, 432, 55 },
		{ 394, 433, 55 },
		{ 394, 434, 55 },
		{ 394, 435, 55 },
		{ 394, 436, 55 },
		{ 394, 437, 55 },
		{ 394, 438, 55 },
		{ 394, 439, 55 },
		{ 395, 396, 55 },
		{ 395, 397, 55 },
		{ 395, 398, 55 },
		{ 395, 399, 55 },
		{ 395, 400, 55 },
		{ 395, 401, 55 },
		{ 395, 402, 55 },
		{ 395, 403, 55 },
		{ 395, 404, 55 },
		{ 395, 405, 55 },
		{ 395, 406, 55 },
		{ 395, 407, 55 },
		{ 395, 408, 55 },
		{ 395, 409, 55 },
		{ 395, 410, 55 },
		{ 395, 411, 55 },
		{ 395, 412, 55 },
		{ 395, 413, 55 },
		{ 395, 414, 55 },
		{ 395, 415, 55 },
		{ 395, 416, 55 },
		{ 395, 417, 55 },
		{ 395, 418, 55 },
		{ 395, 419, 55 },
		{ 395, 420, 55 },
		{ 395, 421, 55 },
		{ 395, 422, 55 },
		{ 395, 423, 55 },
		{ 395, 424, 55 },
		{ 395, 425, 55 },
		{ 395, 426, 55 },
		{ 395, 427, 55 },
		{ 395, 428, 55 },
		{ 395, 429, 55 },
		{ 395, 430, 55 },
		{ 395, 431, 55 },
		{ 395, 432, 55 },
		{ 395, 433, 55 },
		{ 395, 434, 55 },
		{ 395, 435, 55 },
		{ 395, 436, 55 },
		{ 395, 437, 55 },
		{ 395, 438, 55 },
		{ 395, 439, 55 },
		{ 395, 440, 55 },
		{ 396, 397, 55 },
		{ 396, 398, 55 },
		{ 396, 399, 55 },
		{ 396, 400, 55 },
		{ 396, 401, 55 },
		{ 396, 402, 55 },
		{ 396, 403, 55 },
		{ 396, 404, 55 },
		{ 396, 405, 55 },
		{ 396, 406, 55 },
		{ 396, 407, 55 },
		{ 396, 408, 55 },
		{ 396, 409, 55 },
		{ 396, 410, 55 },
		{ 396, 411, 55 },
		{ 396, 412, 55 },
		{ 396, 413, 55 },
		{ 396, 414, 55 },
		{ 396, 415, 55 },
		{ 396, 416, 55 },
		{ 396, 417, 55 },
		{ 396, 418, 55 },
		{ 396, 419, 55 },
		{ 396, 420, 55 },
		{ 396, 421, 55 },
		{ 396, 422, 55 },
		{ 396, 423, 55 },
		{ 396, 424, 55 },
		{ 396, 425, 55 },
		{ 396, 426, 55 },
		{ 396, 427, 55 },
		{ 396, 428, 55 },
		{ 396, 429, 55 },
		{ 396, 430, 55 },
		{ 396, 431, 55 },
		{ 396, 432, 55 },
		{ 396, 433, 55 },
		{ 396, 434, 55 },
		{ 396, 435, 55 },
		{ 396, 436, 55 },
		{ 396, 437, 55 },
		{ 396, 438, 55 },
		{ 396, 439, 55 },
		{ 396, 440, 55 },
		{ 396, 441, 55 },
		{ 396, 442, 55 },
		{ 397, 398, 55 },
		{ 397, 399, 55 },
		{ 397, 400, 55 },
		{ 397, 401, 55 },
		{ 397, 402, 55 },
		{ 397, 403, 55 },
		{ 397, 404, 55 },
		{ 397, 405, 55 },
		{ 397, 406, 55 },
		{ 397, 407, 55 },
		{ 397, 408, 55 },
		{ 397, 409, 55 },
		{ 397, 410, 55 },
		{ 397, 411, 55 },
		{ 397, 412, 55 },
		{ 397, 413, 55 },
		{ 397, 414, 55 },
		{ 397, 415, 55 },
		{ 397, 416, 55 },
		{ 397, 417, 55 },
		{ 397, 418, 55 },
		{ 397, 419, 55 },
		{ 397, 420, 55 },
		{ 397, 421, 55 },
		{ 397, 422, 55 },
		{ 397, 423, 55 },
		{ 397, 424, 55 },
		{ 397, 425, 55 },
		{ 397, 426, 55 },
		{ 397, 427, 55 },
		{ 397, 428, 55 },
		{ 397, 429, 55 },
		{ 397, 430, 55 },
		{ 397, 431, 55 },
		{ 397, 432, 55 },
		{ 397, 433, 55 },
		{ 397, 434, 55 },
		{ 397, 435, 55 },
		{ 397, 436, 55 },
		{ 397, 437, 55 },
		{ 397, 438, 55 },
		{ 397, 439, 55 },
		{ 397, 440, 55 },
		{ 397, 441, 55 },
		{ 397, 442, 55 },
		{ 397, 443, 55 },
		{ 398, 399, 55 },
		{ 398, 400, 55 },
		{ 398, 401, 55 },
		{ 398, 402, 55 },
		{ 398, 403, 55 },
		{ 398, 404, 55 },
		{ 398, 405, 55 },
		{ 398, 406, 55 },
		{ 398, 407, 55 },
		{ 398, 408, 55 },
		{ 398, 409, 55 },
		{ 398, 410, 55 },
		{ 398, 411, 55 },
		{ 398, 412, 55 },
		{ 398, 413, 55 },
		{ 398, 414, 55 },
		{ 398, 415, 55 },
		{ 398, 416, 55 },
		{ 398, 417, 55 },
		{ 398, 418, 55 },
		{ 398, 419, 55 },
		{ 398, 420, 55 },
		{ 398, 421, 55 },
		{ 398, 422, 55 },
		{ 398, 423, 55 },
		{ 398, 424, 55 },
		{ 398, 425, 55 },
		{ 398, 426, 55 },
		{ 398, 427, 55 },
		{ 398, 428, 55 },
		{ 398, 429, 55 },
		{ 398, 430, 55 },
		{ 398, 431, 55 },
		{ 398, 432, 55 },
		{ 398, 433, 55 },
		{ 398, 434, 55 },
		{ 398, 435, 55 },
		{ 398, 436, 55 },
		{ 398, 437, 55 },
		{ 398, 438, 55 },
		{ 398, 439, 55 },
		{ 398, 440, 55 },
		{ 398, 441, 55 },
		{ 398, 442, 55 },
		{ 398, 443, 55 },
		{ 398, 444, 55 },
		{ 399, 400, 55 },
		{ 399, 401, 55 },
		{ 399, 402, 55 },
		{ 399, 403, 55 },
		{ 399, 404, 55 },
		{ 399, 405, 55 },
		{ 399, 406, 55 },
		{ 399, 407, 55 },
		{ 399, 408, 55 },
		{ 399, 409, 55 },
		{ 399, 410, 55 },
		{ 399, 411, 55 },
		{ 399, 412, 55 },
		{ 399, 413, 55 },
		{ 399, 414, 55 },
		{ 399, 415, 55 },
		{ 399, 416, 55 },
		{ 399, 417, 55 },
		{ 399, 418, 55 },
		{ 399, 419, 55 },
		{ 399, 420, 55 },
		{ 399, 421, 55 },
		{ 399, 422, 55 },
		{ 399, 423, 55 },
		{ 399, 424, 55 },
		{ 399, 425, 55 },
		{ 399, 426, 55 },
		{ 399, 427, 55 },
		{ 399, 428, 55 },
		{ 399, 429, 55 },
		{ 399, 430, 55 },
		{ 399, 431, 55 },
		{ 399, 432, 55 },
		{ 399, 433, 55 },
		{ 399, 434, 55 },
		{ 399, 435, 55 },
		{ 399, 436, 55 },
		{ 399, 437, 55 },
		{ 399, 438, 55 },
		{ 399, 439, 55 },
		{ 399, 440, 55 },
		{ 399, 441, 55 },
		{ 399, 442, 55 },
		{ 399, 443, 55 },
		{ 399, 444, 55 },
		{ 399, 445, 55 },
		{ 399, 446, 55 },
		{ 400, 401, 55 },
		{ 400, 402, 55 },
		{ 400, 403, 55 },
		{ 400, 404, 55 },
		{ 400, 405, 55 },
		{ 400, 406, 55 },
		{ 400, 407, 55 },
		{ 400, 408, 55 },
		{ 400, 409, 55 },
		{ 400, 410, 55 },
		{ 400, 411, 55 },
		{ 400, 412, 55 },
		{ 400, 413, 55 },
		{ 400, 414, 55 },
		{ 400, 415, 55 },
		{ 400, 416, 55 },
		{ 400, 417, 55 },
		{ 400, 418, 55 },
		{ 400, 419, 55 },
		{ 400, 420, 55 },
		{ 400, 421, 55 },
		{ 400, 422, 55 },
		{ 400, 423, 55 },
		{ 400, 424, 55 },
		{ 400, 425, 55 },
		{ 400, 426, 55 },
		{ 400, 427, 55 },
		{ 400, 428, 55 },
		{ 400, 429, 55 },
		{ 400, 430, 55 },
		{ 400, 431, 55 },
		{ 400, 432, 55 },
		{ 400, 433, 55 },
		{ 400, 434, 55 },
		{ 400, 435, 55 },
		{ 400, 436, 55 },
		{ 400, 437, 55 },
		{ 400, 438, 55 },
		{ 400, 439, 55 },
		{ 400, 440, 55 },
		{ 400, 441, 55 },
		{ 400, 442, 55 },
		{ 400, 443, 55 },
		{ 400, 444, 55 },
		{ 400, 445, 55 },
		{ 400, 446, 55 },
		{ 400, 447, 55 },
		{ 401, 402, 55 },
		{ 401, 403, 55 },
		{ 401, 404, 55 },
		{ 401, 405, 55 },
		{ 401, 406, 55 },
		{ 401, 407, 55 },
		{ 401, 408, 55 },
		{ 401, 409, 55 },
		{ 401, 410, 55 },
		{ 401, 411, 55 },
		{ 401, 412, 55 },
		{ 401, 413, 55 },
		{ 401, 414, 55 },
		{ 401, 415, 55 },
		{ 401, 416, 55 },
		{ 401, 417, 55 },
		{ 401, 418, 55 },
		{ 401, 419, 55 },
		{ 401, 420, 55 },
		{ 401, 421, 55 },
		{ 401, 422, 55 },
		{ 401, 423, 55 },
		{ 401, 424, 55 },
		{ 401, 425, 55 },
		{ 401, 426, 55 },
		{ 401, 427, 55 },
		{ 401, 428, 55 },
		{ 401, 429, 55 },
		{ 401, 430, 55 },
		{ 401, 431, 55 },
		{ 401, 432, 55 },
		{ 401, 433, 55 },
		{ 401, 434, 55 },
		{ 401, 435, 55 },
		{ 401, 436, 55 },
		{ 401, 437, 55 },
		{ 401, 438, 55 },
		{ 401, 439, 55 },
		{ 401, 440, 55 },
		{ 401, 441, 55 },
		{ 401, 442, 55 },
		{ 401, 443, 55 },
		{ 401, 444, 55 },
		{ 401, 445, 55 },
		{ 401, 446, 55 },
		{ 401, 447, 55 },
		{ 401, 448, 55 },
		{ 402, 403, 55 },
		{ 402, 404, 55 },
		{ 402, 405, 55 },
		{ 402, 406, 55 },
		{ 402, 407, 55 },
		{ 402, 408, 55 },
		{ 402, 409, 55 },
		{ 402, 410, 55 },
		{ 402, 411, 55 },
		{ 402, 412, 55 },
		{ 402, 413, 55 },
		{ 402, 414, 55 },
		{ 402, 415, 55 },
		{ 402, 416, 55 },
		{ 402, 417, 55 },
		{ 402, 418, 55 },
		{ 402, 419, 55 },
		{ 402, 420, 55 },
		{ 402, 421, 55 },
		{ 402, 422, 55 },
		{ 402, 423, 55 },
		{ 402, 424, 55 },
		{ 402, 425, 55 },
		{ 402, 426, 55 },
		{ 402, 427, 55 },
		{ 402, 428, 55 },
		{ 402, 429, 55 },
		{ 402, 430, 55 },
		{ 402, 431, 55 },
		{ 402, 432, 55 },
		{ 402, 433, 55 },
		{ 402, 434, 55 },
		{ 402, 435, 55 },
		{ 402, 436, 55 },
		{ 402, 437, 55 },
		{ 402, 438, 55 },
		{ 402, 439, 55 },
		{ 402, 440, 55 },
		{ 402, 441, 55 },
		{ 402, 442, 55 },
		{ 402, 443, 55 },
		{ 402, 444, 55 },
		{ 402, 445, 55 },
		{ 402, 446, 55 },
		{ 402, 447, 55 },
		{ 402, 448, 55 },
		{ 402, 449, 55 },
		{ 402, 450, 55 },
		{ 403, 404, 55 },
		{ 403, 405, 55 },
		{ 403, 406, 55 },
		{ 403, 407, 55 },
		{ 403, 408, 55 },
		{ 403, 409, 55 },
		{ 403, 410, 55 },
		{ 403, 411, 55 },
		{ 403, 412, 55 },
		{ 403, 413, 55 },
		{ 403, 414, 55 },
		{ 403, 415, 55 },
		{ 403, 416, 55 },
		{ 403, 417, 55 },
		{ 403, 418, 55 },
		{ 403, 419, 55 },
		{ 403, 420, 55 },
		{ 403, 421, 55 },
		{ 403, 422, 55 },
		{ 403, 423, 55 },
		{ 403, 424, 55 },
		{ 403, 425, 55 },
		{ 403, 426, 55 },
		{ 403, 427, 55 },
		{ 403, 428, 55 },
		{ 403, 429, 55 },
		{ 403, 430, 55 },
		{ 403, 431, 55 },
		{ 403, 432, 55 },
		{ 403, 433, 55 },
		{ 403, 434, 55 },
		{ 403, 435, 55 },
		{ 403, 436, 55 },
		{ 403, 437, 55 },
		{ 403, 438, 55 },
		{ 403, 439, 55 },
		{ 403, 440, 55 },
		{ 403, 441, 55 },
		{ 403, 442, 55 },
		{ 403, 443, 55 },
		{ 403, 444, 55 },
		{ 403, 445, 55 },
		{ 403, 446, 55 },
		{ 403, 447, 55 },
		{ 403, 448, 55 },
		{ 403, 449, 55 },
		{ 403, 450, 55 },
		{ 403, 451, 55 },
		{ 404, 405, 55 },
		{ 404, 406, 55 },
		{ 404, 407, 55 },
		{ 404, 408, 55 },
		{ 404, 409, 55 },
		{ 404, 410, 55 },
		{ 404, 411, 55 },
		{ 404, 412, 55 },
		{ 404, 413, 55 },
		{ 404, 414, 55 },
		{ 404, 415, 55 },
		{ 404, 416, 55 },
		{ 404, 417, 55 },
		{ 404, 418, 55 },
		{ 404, 419, 55 },
		{ 404, 420, 55 },
		{ 404, 421, 55 },
		{ 404, 422, 55 },
		{ 404, 423, 55 },
		{ 404, 424, 55 },
		{ 404, 425, 55 },
		{ 404, 426, 55 },
		{ 404, 427, 55 },
		{ 404, 428, 55 },
		{ 404, 429, 55 },
		{ 404, 430, 55 },
		{ 404, 431, 55 },
		{ 404, 432, 55 },
		{ 404, 433, 55 },
		{ 404, 434, 55 },
		{ 404, 435, 55 },
		{ 404, 436, 55 },
		{ 404, 437, 55 },
		{ 404, 438, 55 },
		{ 404, 439, 55 },
		{ 404, 440, 55 },
		{ 404, 441, 55 },
		{ 404, 442, 55 },
		{ 404, 443, 55 },
		{ 404, 444, 55 },
		{ 404, 445, 55 },
		{ 404, 446, 55 },
		{ 404, 447, 55 },
		{ 404, 448, 55 },
		{ 404, 449, 55 },
		{ 404, 450, 55 },
		{ 404, 451, 55 },
		{ 404, 452, 55 },
		{ 405, 406, 55 },
		{ 405, 407, 55 },
		{ 405, 408, 55 },
		{ 405, 409, 55 },
		{ 405, 410, 55 },
		{ 405, 411, 55 },
		{ 405, 412, 55 },
		{ 405, 413, 55 },
		{ 405, 414, 55 },
		{ 405, 415, 55 },
		{ 405, 416, 55 },
		{ 405, 417, 55 },
		{ 405, 418, 55 },
		{ 405, 419, 55 },
		{ 405, 420, 55 },
		{ 405, 421, 55 },
		{ 405, 422, 55 },
		{ 405, 423, 55 },
		{ 405, 424, 55 },
		{ 405, 425, 55 },
		{ 405, 426, 55 },
		{ 405, 427, 55 },
		{ 405, 428, 55 },
		{ 405, 429, 55 },
		{ 405, 430, 55 },
		{ 405, 431, 55 },
		{ 405, 432, 55 },
		{ 405, 433, 55 },
		{ 405, 434, 55 },
		{ 405, 435, 55 },
		{ 405, 436, 55 },
		{ 405, 437, 55 },
		{ 405, 438, 55 },
		{ 405, 439, 55 },
		{ 405, 440, 55 },
		{ 405, 441, 55 },
		{ 405, 442, 55 },
		{ 405, 443, 55 },
		{ 405, 444, 55 },
		{ 405, 445, 55 },
		{ 405, 446, 55 },
		{ 405, 447, 55 },
		{ 405, 448, 55 },
		{ 405, 449, 55 },
		{ 405, 450, 55 },
		{ 405, 451, 55 },
		{ 405, 452, 55 },
		{ 405, 453, 55 },
		{ 405, 454, 55 },
		{ 406, 407, 55 },
		{ 406, 408, 55 },
		{ 406, 409, 55 },
		{ 406, 410, 55 },
		{ 406, 411, 55 },
		{ 406, 412, 55 },
		{ 406, 413, 55 },
		{ 406, 414, 55 },
		{ 406, 415, 55 },
		{ 406, 416, 55 },
		{ 406, 417, 55 },
		{ 406, 418, 55 },
		{ 406, 419, 55 },
		{ 406, 420, 55 },
		{ 406, 421, 55 },
		{ 406, 422, 55 },
		{ 406, 423, 55 },
		{ 406, 424, 55 },
		{ 406, 425, 55 },
		{ 406, 426, 55 },
		{ 406, 427, 55 },
		{ 406, 428, 55 },
		{ 406, 429, 55 },
		{ 406, 430, 55 },
		{ 406, 431, 55 },
		{ 406, 432, 55 },
		{ 406, 433, 55 },
		{ 406, 434, 55 },
		{ 406, 435, 55 },
		{ 406, 436, 55 },
		{ 406, 437, 55 },
		{ 406, 438, 55 },
		{ 406, 439, 55 },
		{ 406, 440, 55 },
		{ 406, 441, 55 },
		{ 406, 442, 55 },
		{ 406, 443, 55 },
		{ 406, 444, 55 },
		{ 406, 445, 55 },
		{ 406, 446, 55 },
		{ 406, 447, 55 },
		{ 406, 448, 55 },
		{ 406, 449, 55 },
		{ 406, 450, 55 },
		{ 406, 451, 55 },
		{ 406, 452, 55 },
		{ 406, 453, 55 },
		{ 406, 454, 55 },
		{ 406, 455, 55 },
		{ 407, 408, 55 },
		{ 407, 409, 55 },
		{ 407, 410, 55 },
		{ 407, 411, 55 },
		{ 407, 412, 55 },
		{ 407, 413, 55 },
		{ 407, 414, 55 },
		{ 407, 415, 55 },
		{ 407, 416, 55 },
		{ 407, 417, 55 },
		{ 407, 418, 55 },
		{ 407, 419, 55 },
		{ 407, 420, 55 },
		{ 407, 421, 55 },
		{ 407, 422, 55 },
		{ 407, 423, 55 },
		{ 407, 424, 55 },
		{ 407, 425, 55 },
		{ 407, 426, 55 },
		{ 407, 427, 55 },
		{ 407, 428, 55 },
		{ 407, 429, 55 },
		{ 407, 430, 55 },
		{ 407, 431, 55 },
		{ 407, 432, 55 },
		{ 407, 433, 55 },
		{ 407, 434, 55 },
		{ 407, 435, 55 },
		{ 407, 436, 55 },
		{ 407, 437, 55 },
		{ 407, 438, 55 },
		{ 407, 439, 55 },
		{ 407, 440, 55 },
		{ 407, 441, 55 },
		{ 407, 442, 55 },
		{ 407, 443, 55 },
		{ 407, 444, 55 },
		{ 407, 445, 55 },
		{ 407, 446, 55 },
		{ 407, 447, 55 },
		{ 407, 448, 55 },
		{ 407, 449, 55 },
		{ 407, 450, 55 },
		{ 407, 451, 55 },
		{ 407, 452, 55 },
		{ 407, 453, 55 },
		{ 407, 454, 55 },
		{ 407, 455, 55 },
		{ 407, 456, 55 },
		{ 408, 409, 55 },
		{ 408, 410, 55 },
		{ 408, 411, 55 },
		{ 408, 412, 55 },
		{ 408, 413, 55 },
		{ 408, 414, 55 },
		{ 408, 415, 55 },
		{ 408, 416, 55 },
		{ 408, 417, 55 },
		{ 408, 418, 55 },
		{ 408, 419, 55 },
		{ 408, 420, 55 },
		{ 408, 421, 55 },
		{ 408, 422, 55 },
		{ 408, 423, 55 },
		{ 408, 424, 55 },
		{ 408, 425, 55 },
		{ 408, 426, 55 },
		{ 408, 427, 55 },
		{ 408, 428, 55 },
		{ 408, 429, 55 },
		{ 408, 430, 55 },
		{ 408, 431, 55 },
		{ 408, 432, 55 },
		{ 408, 433, 55 },
		{ 408, 434, 55 },
		{ 408, 435, 55 },
		{ 408, 436, 55 },
		{ 408, 437, 55 },
		{ 408, 438, 55 },
		{ 408, 439, 55 },
		{ 408, 440, 55 },
		{ 408, 441, 55 },
		{ 408, 442, 55 },
		{ 408, 443, 55 },
		{ 408, 444, 55 },
		{ 408, 445, 55 },
		{ 408, 446, 55 },
		{ 408, 447, 55 },
		{ 408, 448, 55 },
		{ 408, 449, 55 },
		{ 408, 450, 55 },
		{ 408, 451, 55 },
		{ 408, 452, 55 },
		{ 408, 453, 55 },
		{ 408, 454, 55 },
		{ 408, 455, 55 },
		{ 408, 456, 55 },
		{ 408, 457, 55 },
		{ 408, 458, 55 },
		{ 409, 410, 55 },
		{ 409, 411, 55 },
		{ 409, 412, 55 },
		{ 409, 413, 55 },
		{ 409, 414, 55 },
		{ 409, 415, 55 },
		{ 409, 416, 55 },
		{ 409, 417, 55 },
		{ 409, 418, 55 },
		{ 409, 419, 55 },
		{ 409, 420, 55 },
		{ 409, 421, 55 },
		{ 409, 422, 55 },
		{ 409, 423, 55 },
		{ 409, 424, 55 },
		{ 409, 425, 55 },
		{ 409, 426, 55 },
		{ 409, 427, 55 },
		{ 409, 428, 55 },
		{ 409, 429, 55 },
		{ 409, 430, 55 },
		{ 409, 431, 55 },
		{ 409, 432, 55 },
		{ 409, 433, 55 },
		{ 409, 434, 55 },
		{ 409, 435, 55 },
		{ 409, 436, 55 },
		{ 409, 437, 55 },
		{ 409, 438, 55 },
		{ 409, 439, 55 },
		{ 409, 440, 55 },
		{ 409, 441, 55 },
		{ 409, 442, 55 },
		{ 409, 443, 55 },
		{ 409, 444, 55 },
		{ 409, 445, 55 },
		{ 409, 446, 55 },
		{ 409, 447, 55 },
		{ 409, 448, 55 },
		{ 409, 449, 55 },
		{ 409, 450, 55 },
		{ 409, 451, 55 },
		{ 409, 452, 55 },
		{ 409, 453, 55 },
		{ 409, 454, 55 },
		{ 409, 455, 55 },
		{ 409, 456, 55 },
		{ 409, 457, 55 },
		{ 409, 458, 55 },
		{ 409, 459, 55 },
		{ 410, 411, 55 },
		{ 410, 412, 55 },
		{ 410, 413, 55 },
		{ 410, 414, 55 },
		{ 410, 415, 55 },
		{ 410, 416, 55 },
		{ 410, 417, 55 },
		{ 410, 418, 55 },
		{ 410, 419, 55 },
		{ 410, 420, 55 },
		{ 410, 421, 55 },
		{ 410, 422, 55 },
		{ 410, 423, 55 },
		{ 410, 424, 55 },
		{ 410, 425, 55 },
		{ 410, 426, 55 },
		{ 410, 427, 55 },
		{ 410, 428, 55 },
		{ 410, 429, 55 },
		{ 410, 430, 55 },
		{ 410, 431, 55 },
		{ 410, 432, 55 },
		{ 410, 433, 55 },
		{ 410, 434, 55 },
		{ 410, 435, 55 },
		{ 410, 436, 55 },
		{ 410, 437, 55 },
		{ 410, 438, 55 },
		{ 410, 439, 55 },
		{ 410, 440, 55 },
		{ 410, 441, 55 },
		{ 410, 442, 55 },
		{ 410, 443, 55 },
		{ 410, 444, 55 },
		{ 410, 445, 55 },
		{ 410, 446, 55 },
		{ 410, 447, 55 },
		{ 410, 448, 55 },
		{ 410, 449, 55 },
		{ 410, 450, 55 },
		{ 410, 451, 55 },
		{ 410, 452, 55 },
		{ 410, 453, 55 },
		{ 410, 454, 55 },
		{ 410, 455, 55 },
		{ 410, 456, 55 },
		{ 410, 457, 55 },
		{ 410, 458, 55 },
		{ 410, 459, 55 },
		{ 410, 460, 55 },
		{ 411, 412, 55 },
		{ 411, 413, 55 },
		{ 411, 414, 55 },
		{ 411, 415, 55 },
		{ 411, 416, 55 },
		{ 411, 417, 55 },
		{ 411, 418, 55 },
		{ 411, 419, 55 },
		{ 411, 420, 55 },
		{ 411, 421, 55 },
		{ 411, 422, 55 },
		{ 411, 423, 55 },
		{ 411, 424, 55 },
		{ 411, 425, 55 },
		{ 411, 426, 55 },
		{ 411, 427, 55 },
		{ 411, 428, 55 },
		{ 411, 429, 55 },
		{ 411, 430, 55 },
		{ 411, 431, 55 },
		{ 411, 432, 55 },
		{ 411, 433, 55 },
		{ 411, 434, 55 },
		{ 411, 435, 55 },
		{ 411, 436, 55 },
		{ 411, 437, 55 },
		{ 411, 438, 55 },
		{ 411, 439, 55 },
		{ 411, 440, 55 },
		{ 411, 441, 55 },
		{ 411, 442, 55 },
		{ 411, 443, 55 },
		{ 411, 444, 55 },
		{ 411, 445, 55 },
		{ 411, 446, 55 },
		{ 411, 447, 55 },
		{ 411, 448, 55 },
		{ 411, 449, 55 },
		{ 411, 450, 55 },
		{ 411, 451, 55 },
		{ 411, 452, 55 },
		{ 411, 453, 55 },
		{ 411, 454, 55 },
		{ 411, 455, 55 },
		{ 411, 456, 55 },
		{ 411, 457, 55 },
		{ 411, 458, 55 },
		{ 411, 459, 55 },
		{ 411, 460, 55 },
		{ 411, 461, 55 },
		{ 411, 462, 55 },
		{ 412, 413, 55 },
		{ 412, 414, 55 },
		{ 412, 415, 55 },
		{ 412, 416, 55 },
		{ 412, 417, 55 },
		{ 412, 418, 55 },
		{ 412, 419, 55 },
		{ 412, 420, 55 },
		{ 412, 421, 55 },
		{ 412, 422, 55 },
		{ 412, 423, 55 },
		{ 412, 424, 55 },
		{ 412, 425, 55 },
		{ 412, 426, 55 },
		{ 412, 427, 55 },
		{ 412, 428, 55 },
		{ 412, 429, 55 },
		{ 412, 430, 55 },
		{ 412, 431, 55 },
		{ 412, 432, 55 },
		{ 412, 433, 55 },
		{ 412, 434, 55 },
		{ 412, 435, 55 },
		{ 412, 436, 55 },
		{ 412, 437, 55 },
		{ 412, 438, 55 },
		{ 412, 439, 55 },
		{ 412, 440, 55 },
		{ 412, 441, 55 },
		{ 412, 442, 55 },
		{ 412, 443, 55 },
		{ 412, 444, 55 },
		{ 412, 445, 55 },
		{ 412, 446, 55 },
		{ 412, 447, 55 },
		{ 412, 448, 55 },
		{ 412, 449, 55 },
		{ 412, 450, 55 },
		{ 412, 451, 55 },
		{ 412, 452, 55 },
		{ 412, 453, 55 },
		{ 412, 454, 55 },
		{ 412, 455, 55 },
		{ 412, 456, 55 },
		{ 412, 457, 55 },
		{ 412, 458, 55 },
		{ 412, 459, 55 },
		{ 412, 460, 55 },
		{ 412, 461, 55 },
		{ 412, 462, 55 },
		{ 412, 463, 55 },
		{ 413, 414, 55 },
		{ 413, 415, 55 },
		{ 413, 416, 55 },
		{ 413, 417, 55 },
		{ 413, 418, 55 },
		{ 413, 419, 55 },
		{ 413, 420, 55 },
		{ 413, 421, 55 },
		{ 413, 422, 55 },
		{ 413, 423, 55 },
		{ 413, 424, 55 },
		{ 413, 425, 55 },
		{ 413, 426, 55 },
		{ 413, 427, 55 },
		{ 413, 428, 55 },
		{ 413, 429, 55 },
		{ 413, 430, 55 },
		{ 413, 431, 55 },
		{ 413, 432, 55 },
		{ 413, 433, 55 },
		{ 413, 434, 55 },
		{ 413, 435, 55 },
		{ 413, 436, 55 },
		{ 413, 437, 55 },
		{ 413, 438, 55 },
		{ 413, 439, 55 },
		{ 413, 440, 55 },
		{ 413, 441, 55 },
		{ 413, 442, 55 },
		{ 413, 443, 55 },
		{ 413, 444, 55 },
		{ 413, 445, 55 },
		{ 413, 446, 55 },
		{ 413, 447, 55 },
		{ 413, 448, 55 },
		{ 413, 449, 55 },
		{ 413, 450, 55 },
		{ 413, 451, 55 },
		{ 413, 452, 55 },
		{ 413, 453, 55 },
		{ 413, 454, 55 },
		{ 413, 455, 55 },
		{ 413, 456, 55 },
		{ 413, 457, 55 },
		{ 413, 458, 55 },
		{ 413, 459, 55 },
		{ 413, 460, 55 },
		{ 413, 461, 55 },
		{ 413, 462, 55 },
		{ 413, 463, 55 },
		{ 413, 464, 55 },
		{ 414, 415, 55 },
		{ 414, 416, 55 },
		{ 414, 417, 55 },
		{ 414, 418, 55 },
		{ 414, 419, 55 },
		{ 414, 420, 55 },
		{ 414, 421, 55 },
		{ 414, 422, 55 },
		{ 414, 423, 55 },
		{ 414, 424, 55 },
		{ 414, 425, 55 },
		{ 414, 426, 55 },
		{ 414, 427, 55 },
		{ 414, 428, 55 },
		{ 414, 429, 55 },
		{ 414, 430, 55 },
		{ 414, 431, 55 },
		{ 414, 432, 55 },
		{ 414, 433, 55 },
		{ 414, 434, 55 },
		{ 414, 435, 55 },
		{ 414, 436, 55 },
		{ 414, 437, 55 },
		{ 414, 438, 55 },
		{ 414, 439, 55 },
		{ 414, 440, 55 },
		{ 414, 441, 55 },
		{ 414, 442, 55 },
		{ 414, 443, 55 },
		{ 414, 444, 55 },
		{ 414, 445, 55 },
		{ 414, 446, 55 },
		{ 414, 447, 55 },
		{ 414, 448, 55 },
		{ 414, 449, 55 },
		{ 414, 450, 55 },
		{ 414, 451, 55 },
		{ 414, 452, 55 },
		{ 414, 453, 55 },
		{ 414, 454, 55 },
		{ 414, 455, 55 },
		{ 414, 456, 55 },
		{ 414, 457, 55 },
		{ 414, 458, 55 },
		{ 414, 459, 55 },
		{ 414, 460, 55 },
		{ 414, 461, 55 },
		{ 414, 462, 55 },
		{ 414, 463, 55 },
		{ 414, 464, 55 },
		{ 414, 465, 55 },
		{ 414, 466, 55 },
		{ 415, 416, 55 },
		{ 415, 417, 55 },
		{ 415, 418, 55 },
		{ 415, 419, 55 },
		{ 415, 420, 55 },
		{ 415, 421, 55 },
		{ 415, 422, 55 },
		{ 415, 423, 55 },
		{ 415, 424, 55 },
		{ 415, 425, 55 },
		{ 415, 426, 55 },
		{ 415, 427, 55 },
		{ 415, 428, 55 },
		{ 415, 429, 55 },
		{ 415, 430, 55 },
		{ 415, 431, 55 },
		{ 415, 432, 55 },
		{ 415, 433, 55 },
		{ 415, 434, 55 },
		{ 415, 435, 55 },
		{ 415, 436, 55 },
		{ 415, 437, 55 },
		{ 415, 438, 55 },
		{ 415, 439, 55 },
		{ 415, 440, 55 },
		{ 415, 441, 55 },
		{ 415, 442, 55 },
		{ 415, 443, 55 },
		{ 415, 444, 55 },
		{ 415, 445, 55 },
		{ 415, 446, 55 },
		{ 415, 447, 55 },
		{ 415, 448, 55 },
		{ 415, 449, 55 },
		{ 415, 450, 55 },
		{ 415, 451, 55 },
		{ 415, 452, 55 },
		{ 415, 453, 55 },
		{ 415, 454, 55 },
		{ 415, 455, 55 },
		{ 415, 456, 55 },
		{ 415, 457, 55 },
		{ 415, 458, 55 },
		{ 415, 459, 55 },
		{ 415, 460, 55 },
		{ 415, 461, 55 },
		{ 415, 462, 55 },
		{ 415, 463, 55 },
		{ 415, 464, 55 },
		{ 415, 465, 55 },
		{ 415, 466, 55 },
		{ 415, 467, 55 },
		{ 416, 417, 55 },
		{ 416, 418, 55 },
		{ 416, 419, 55 },
		{ 416, 420, 55 },
		{ 416, 421, 55 },
		{ 416, 422, 55 },
		{ 416, 423, 55 },
		{ 416, 424, 55 },
		{ 416, 425, 55 },
		{ 416, 426, 55 },
		{ 416, 427, 55 },
		{ 416, 428, 55 },
		{ 416, 429, 55 },
		{ 416, 430, 55 },
		{ 416, 431, 55 },
		{ 416, 432, 55 },
		{ 416, 433, 55 },
		{ 416, 434, 55 },
		{ 416, 435, 55 },
		{ 416, 436, 55 },
		{ 416, 437, 55 },
		{ 416, 438, 55 },
		{ 416, 439, 55 },
		{ 416, 440, 55 },
		{ 416, 441, 55 },
		{ 416, 442, 55 },
		{ 416, 443, 55 },
		{ 416, 444, 55 },
		{ 416, 445, 55 },
		{ 416, 446, 55 },
		{ 416, 447, 55 },
		{ 416, 448, 55 },
		{ 416, 449, 55 },
		{ 416, 450, 55 },
		{ 416, 451, 55 },
		{ 416, 452, 55 },
		{ 416, 453, 55 },
		{ 416, 454, 55 },
		{ 416, 455, 55 },
		{ 416, 456, 55 },
		{ 416, 457, 55 },
		{ 416, 458, 55 },
		{ 416, 459, 55 },
		{ 416, 460, 55 },
		{ 416, 461, 55 },
		{ 416, 462, 55 },
		{ 416, 463, 55 },
		{ 416, 464, 55 },
		{ 416, 465, 55 },
		{ 416, 466, 55 },
		{ 416, 467, 55 },
		{ 416, 468, 55 },
		{ 417, 418, 55 },
		{ 417, 419, 55 },
		{ 417, 420, 55 },
		{ 417, 421, 55 },
		{ 417, 422, 55 },
		{ 417, 423, 55 },
		{ 417, 424, 55 },
		{ 417, 425, 55 },
		{ 417, 426, 55 },
		{ 417, 427, 55 },
		{ 417, 428, 55 },
		{ 417, 429, 55 },
		{ 417, 430, 55 },
		{ 417, 431, 55 },
		{ 417, 432, 55 },
		{ 417, 433, 55 },
		{ 417, 434, 55 },
		{ 417, 435, 55 },
		{ 417, 436, 55 },
		{ 417, 437, 55 },
		{ 417, 438, 55 },
		{ 417, 439, 55 },
		{ 417, 440, 55 },
		{ 417, 441, 55 },
		{ 417, 442, 55 },
		{ 417, 443, 55 },
		{ 417, 444, 55 },
		{ 417, 445, 55 },
		{ 417, 446, 55 },
		{ 417, 447, 55 },
		{ 417, 448, 55 },
		{ 417, 449, 55 },
		{ 417, 450, 55 },
		{ 417, 451, 55 },
		{ 417, 452, 55 },
		{ 417, 453, 55 },
		{ 417, 454, 55 },
		{ 417, 455, 55 },
		{ 417, 456, 55 },
		{ 417, 457, 55 },
		{ 417, 458, 55 },
		{ 417, 459, 55 },
		{ 417, 460, 55 },
		{ 417, 461, 55 },
		{ 417, 462, 55 },
		{ 417, 463, 55 },
		{ 417, 464, 55 },
		{ 417, 465, 55 },
		{ 417, 466, 55 },
		{ 417, 467, 55 },
		{ 417, 468, 55 },
		{ 417, 469, 55 },
		{ 417, 470, 55 },
		{ 418, 419, 55 },
		{ 418, 420, 55 },
		{ 418, 421, 55 },
		{ 418, 422, 55 },
		{ 418, 423, 55 },
		{ 418, 424, 55 },
		{ 418, 425, 55 },
		{ 418, 426, 55 },
		{ 418, 427, 55 },
		{ 418, 428, 55 },
		{ 418, 429, 55 },
		{ 418, 430, 55 },
		{ 418, 431, 55 },
		{ 418, 432, 55 },
		{ 418, 433, 55 },
		{ 418, 434, 55 },
		{ 418, 435, 55 },
		{ 418, 436, 55 },
		{ 418, 437, 55 },
		{ 418, 438, 55 },
		{ 418, 439, 55 },
		{ 418, 440, 55 },
		{ 418, 441, 55 },
		{ 418, 442, 55 },
		{ 418, 443, 55 },
		{ 418, 444, 55 },
		{ 418, 445, 55 },
		{ 418, 446, 55 },
		{ 418, 447, 55 },
		{ 418, 448, 55 },
		{ 418, 449, 55 },
		{ 418, 450, 55 },
		{ 418, 451, 55 },
		{ 418, 452, 55 },
		{ 418, 453, 55 },
		{ 418, 454, 55 },
		{ 418, 455, 55 },
		{ 418, 456, 55 },
		{ 418, 457, 55 },
		{ 418, 458, 55 },
		{ 418, 459, 55 },
		{ 418, 460, 55 },
		{ 418, 461, 55 },
		{ 418, 462, 55 },
		{ 418, 463, 55 },
		{ 418, 464, 55 },
		{ 418, 465, 55 },
		{ 418, 466, 55 },
		{ 418, 467, 55 },
		{ 418, 468, 55 },
		{ 418, 469, 55 },
		{ 418, 470, 55 },
		{ 418, 471, 55 },
		{ 419, 420, 55 },
		{ 419, 421, 55 },
		{ 419, 422, 55 },
		{ 419, 423, 55 },
		{ 419, 424, 55 },
		{ 419, 425, 55 },
		{ 419, 426, 55 },
		{ 419, 427, 55 },
		{ 419, 428, 55 },
		{ 419, 429, 55 },
		{ 419, 430, 55 },
		{ 419, 431, 55 },
		{ 419, 432, 55 },
		{ 419, 433, 55 },
		{ 419, 434, 55 },
		{ 419, 435, 55 },
		{ 419, 436, 55 },
		{ 419, 437, 55 },
		{ 419, 438, 55 },
		{ 419, 439, 55 },
		{ 419, 440, 55 },
		{ 419, 441, 55 },
		{ 419, 442, 55 },
		{ 419, 443, 55 },
		{ 419, 444, 55 },
		{ 419, 445, 55 },
		{ 419, 446, 55 },
		{ 419, 447, 55 },
		{ 419, 448, 55 },
		{ 419, 449, 55 },
		{ 419, 450, 55 },
		{ 419, 451, 55 },
		{ 419, 452, 55 },
		{ 419, 453, 55 },
		{ 419, 454, 55 },
		{ 419, 455, 55 },
		{ 419, 456, 55 },
		{ 419, 457, 55 },
		{ 419, 458, 55 },
		{ 419, 459, 55 },
		{ 419, 460, 55 },
		{ 419, 461, 55 },
		{ 419, 462, 55 },
		{ 419, 463, 55 },
		{ 419, 464, 55 },
		{ 419, 465, 55 },
		{ 419, 466, 55 },
		{ 419, 467, 55 },
		{ 419, 468, 55 },
		{ 419, 469, 55 },
		{ 419, 470, 55 },
		{ 419, 471, 55 },
		{ 419, 472, 55 },
		{ 420, 421, 55 },
		{ 420, 422, 55 },
		{ 420, 423, 55 },
		{ 420, 424, 55 },
		{ 420, 425, 55 },
		{ 420, 426, 55 },
		{ 420, 427, 55 },
		{ 420, 428, 55 },
		{ 420, 429, 55 },
		{ 420, 430, 55 },
		{ 420, 431, 55 },
		{ 420, 432, 55 },
		{ 420, 433, 55 },
		{ 420, 434, 55 },
		{ 420, 435, 55 },
		{ 420, 436, 55 },
		{ 420, 437, 55 },
		{ 420, 438, 55 },
		{ 420, 439, 55 },
		{ 420, 440, 55 },
		{ 420, 441, 55 },
		{ 420, 442, 55 },
		{ 420, 443, 55 },
		{ 420, 444, 55 },
		{ 420, 445, 55 },
		{ 420, 446, 55 },
		{ 420, 447, 55 },
		{ 420, 448, 55 },
		{ 420, 449, 55 },
		{ 420, 450, 55 },
		{ 420, 451, 55 },
		{ 420, 452, 55 },
		{ 420, 453, 55 },
		{ 420, 454, 55 },
		{ 420, 455, 55 },
		{ 420, 456, 55 },
		{ 420, 457, 55 },
		{ 420, 458, 55 },
		{ 420, 459, 55 },
		{ 420, 460, 55 },
		{ 420, 461, 55 },
		{ 420, 462, 55 },
		{ 420, 463, 55 },
		{ 420, 464, 55 },
		{ 420, 465, 55 },
		{ 420, 466, 55 },
		{ 420, 467, 55 },
		{ 420, 468, 55 },
		{ 420, 469, 55 },
		{ 420, 470, 55 },
		{ 420, 471, 55 },
		{ 420, 472, 55 },
		{ 420, 473, 55 },
		{ 420, 474, 55 },
		{ 421, 422, 55 },
		{ 421, 423, 55 },
		{ 421, 424, 55 },
		{ 421, 425, 55 },
		{ 421, 426, 55 },
		{ 421, 427, 55 },
		{ 421, 428, 55 },
		{ 421, 429, 55 },
		{ 421, 430, 55 },
		{ 421, 431, 55 },
		{ 421, 432, 55 },
		{ 421, 433, 55 },
		{ 421, 434, 55 },
		{ 421, 435, 55 },
		{ 421, 436, 55 },
		{ 421, 437, 55 },
		{ 421, 438, 55 },
		{ 421, 439, 55 },
		{ 421, 440, 55 },
		{ 421, 441, 55 },
		{ 421, 442, 55 },
		{ 421, 443, 55 },
		{ 421, 444, 55 },
		{ 421, 445, 55 },
		{ 421, 446, 55 },
		{ 421, 447, 55 },
		{ 421, 448, 55 },
		{ 421, 449, 55 },
		{ 421, 450, 55 },
		{ 421, 451, 55 },
		{ 421, 452, 55 },
		{ 421, 453, 55 },
		{ 421, 454, 55 },
		{ 421, 455, 55 },
		{ 421, 456, 55 },
		{ 421, 457, 55 },
		{ 421, 458, 55 },
		{ 421, 459, 55 },
		{ 421, 460, 55 },
		{ 421, 461, 55 },
		{ 421, 462, 55 },
		{ 421, 463, 55 },
		{ 421, 464, 55 },
		{ 421, 465, 55 },
		{ 421, 466, 55 },
		{ 421, 467, 55 },
		{ 421, 468, 55 },
		{ 421, 469, 55 },
		{ 421, 470, 55 },
		{ 421, 471, 55 },
		{ 421, 472, 55 },
		{ 421, 473, 55 },
		{ 421, 474, 55 },
		{ 421, 475, 55 },
		{ 422, 423, 55 },
		{ 422, 424, 55 },
		{ 422, 425, 55 },
		{ 422, 426, 55 },
		{ 422, 427, 55 },
		{ 422, 428, 55 },
		{ 422, 429, 55 },
		{ 422, 430, 55 },
		{ 422, 431, 55 },
		{ 422, 432, 55 },
		{ 422, 433, 55 },
		{ 422, 434, 55 },
		{ 422, 435, 55 },
		{ 422, 436, 55 },
		{ 422, 437, 55 },
		{ 422, 438, 55 },
		{ 422, 439, 55 },
		{ 422, 440, 55 },
		{ 422, 441, 55 },
		{ 422, 442, 55 },
		{ 422, 443, 55 },
		{ 422, 444, 55 },
		{ 422, 445, 55 },
		{ 422, 446, 55 },
		{ 422, 447, 55 },
		{ 422, 448, 55 },
		{ 422, 449, 55 },
		{ 422, 450, 55 },
		{ 422, 451, 55 },
		{ 422, 452, 55 },
		{ 422, 453, 55 },
		{ 422, 454, 55 },
		{ 422, 455, 55 },
		{ 422, 456, 55 },
		{ 422, 457, 55 },
		{ 422, 458, 55 },
		{ 422, 459, 55 },
		{ 422, 460, 55 },
		{ 422, 461, 55 },
		{ 422, 462, 55 },
		{ 422, 463, 55 },
		{ 422, 464, 55 },
		{ 422, 465, 55 },
		{ 422, 466, 55 },
		{ 422, 467, 55 },
		{ 422, 468, 55 },
		{ 422, 469, 55 },
		{ 422, 470, 55 },
		{ 422, 471, 55 },
		{ 422, 472, 55 },
		{ 422, 473, 55 },
		{ 422, 474, 55 },
		{ 422, 475, 55 },
		{ 422, 476, 55 },
		{ 423, 424, 55 },
		{ 423, 425, 55 },
		{ 423, 426, 55 },
		{ 423, 427, 55 },
		{ 423, 428, 55 },
		{ 423, 429, 55 },
		{ 423, 430, 55 },
		{ 423, 431, 55 },
		{ 423, 432, 55 },
		{ 423, 433, 55 },
		{ 423, 434, 55 },
		{ 423, 435, 55 },
		{ 423, 436, 55 },
		{ 423, 437, 55 },
		{ 423, 438, 55 },
		{ 423, 439, 55 },
		{ 423, 440, 55 },
		{ 423, 441, 55 },
		{ 423, 442, 55 },
		{ 423, 443, 55 },
		{ 423, 444, 55 },
		{ 423, 445, 55 },
		{ 423, 446, 55 },
		{ 423, 447, 55 },
		{ 423, 448, 55 },
		{ 423, 449, 55 },
		{ 423, 450, 55 },
		{ 423, 451, 55 },
		{ 423, 452, 55 },
		{ 423, 453, 55 },
		{ 423, 454, 55 },
		{ 423, 455, 55 },
		{ 423, 456, 55 },
		{ 423, 457, 55 },
		{ 423, 458, 55 },
		{ 423, 459, 55 },
		{ 423, 460, 55 },
		{ 423, 461, 55 },
		{ 423, 462, 55 },
		{ 423, 463, 55 },
		{ 423, 464, 55 },
		{ 423, 465, 55 },
		{ 423, 466, 55 },
		{ 423, 467, 55 },
		{ 423, 468, 55 },
		{ 423, 469, 55 },
		{ 423, 470, 55 },
		{ 423, 471, 55 },
		{ 423, 472, 55 },
		{ 423, 473, 55 },
		{ 423, 474, 55 },
		{ 423, 475, 55 },
		{ 423, 476, 55 },
		{ 423, 477, 55 },
		{ 423, 478, 55 },
		{ 424, 425, 55 },
		{ 424, 426, 55 },
		{ 424, 427, 55 },
		{ 424, 428, 55 },
		{ 424, 429, 55 },
		{ 424, 430, 55 },
		{ 424, 431, 55 },
		{ 424, 432, 55 },
		{ 424, 433, 55 },
		{ 424, 434, 55 },
		{ 424, 435, 55 },
		{ 424, 436, 55 },
		{ 424, 437, 55 },
		{ 424, 438, 55 },
		{ 424, 439, 55 },
		{ 424, 440, 55 },
		{ 424, 441, 55 },
		{ 424, 442, 55 },
		{ 424, 443, 55 },
		{ 424, 444, 55 },
		{ 424, 445, 55 },
		{ 424, 446, 55 },
		{ 424, 447, 55 },
		{ 424, 448, 55 },
		{ 424, 449, 55 },
		{ 424, 450, 55 },
		{ 424, 451, 55 },
		{ 424, 452, 55 },
		{ 424, 453, 55 },
		{ 424, 454, 55 },
		{ 424, 455, 55 },
		{ 424, 456, 55 },
		{ 424, 457, 55 },
		{ 424, 458, 55 },
		{ 424, 459, 55 },
		{ 424, 460, 55 },
		{ 424, 461, 55 },
		{ 424, 462, 55 },
		{ 424, 463, 55 },
		{ 424, 464, 55 },
		{ 424, 465, 55 },
		{ 424, 466, 55 },
		{ 424, 467, 55 },
		{ 424, 468, 55 },
		{ 424, 469, 55 },
		{ 424, 470, 55 },
		{ 424, 471, 55 },
		{ 424, 472, 55 },
		{ 424, 473, 55 },
		{ 424, 474, 55 },
		{ 424, 475, 55 },
		{ 424, 476, 55 },
		{ 424, 477, 55 },
		{ 424, 478, 55 },
		{ 424, 479, 55 },
		{ 425, 426, 55 },
		{ 425, 427, 55 },
		{ 425, 428, 55 },
		{ 425, 429, 55 },
		{ 425, 430, 55 },
		{ 425, 431, 55 },
		{ 425, 432, 55 },
		{ 425, 433, 55 },
		{ 425, 434, 55 },
		{ 425, 435, 55 },
		{ 425, 436, 55 },
		{ 425, 437, 55 },
		{ 425, 438, 55 },
		{ 425, 439, 55 },
		{ 425, 440, 55 },
		{ 425, 441, 55 },
		{ 425, 442, 55 },
		{ 425, 443, 55 },
		{ 425, 444, 55 },
		{ 425, 445, 55 },
		{ 425, 446, 55 },
		{ 425, 447, 55 },
		{ 425, 448, 55 },
		{ 425, 449, 55 },
		{ 425, 450, 55 },
		{ 425, 451, 55 },
		{ 425, 452, 55 },
		{ 425, 453, 55 },
		{ 425, 454, 55 },
		{ 425, 455, 55 },
		{ 425, 456, 55 },
		{ 425, 457, 55 },
		{ 425, 458, 55 },
		{ 425, 459, 55 },
		{ 425, 460, 55 },
		{ 425, 461, 55 },
		{ 425, 462, 55 },
		{ 425, 463, 55 },
		{ 425, 464, 55 },
		{ 425, 465, 55 },
		{ 425, 466, 55 },
		{ 425, 467, 55 },
		{ 425, 468, 55 },
		{ 425, 469, 55 },
		{ 425, 470, 55 },
		{ 425, 471, 55 },
		{ 425, 472, 55 },
		{ 425, 473, 55 },
		{ 425, 474, 55 },
		{ 425, 475, 55 },
		{ 425, 476, 55 },
		{ 425, 477, 55 },
		{ 425, 478, 55 },
		{ 425, 479, 55 },
		{ 425, 480, 55 },
		{ 426, 427, 55 },
		{ 426, 428, 55 },
		{ 426, 429, 55 },
		{ 426, 430, 55 },
		{ 426, 431, 55 },
		{ 426, 432, 55 },
		{ 426, 433, 55 },
		{ 426, 434, 55 },
		{ 426, 435, 55 },
		{ 426, 436, 55 },
		{ 426, 437, 55 },
		{ 426, 438, 55 },
		{ 426, 439, 55 },
		{ 426, 440, 55 },
		{ 426, 441, 55 },
		{ 426, 442, 55 },
		{ 426, 443, 55 },
		{ 426, 444, 55 },
		{ 426, 445, 55 },
		{ 426, 446, 55 },
		{ 426, 447, 55 },
		{ 426, 448, 55 },
		{ 426, 449, 55 },
		{ 426, 450, 55 },
		{ 426, 451, 55 },
		{ 426, 452, 55 },
		{ 426, 453, 55 },
		{ 426, 454, 55 },
		{ 426, 455, 55 },
		{ 426, 456, 55 },
		{ 426, 457, 55 },
		{ 426, 458, 55 },
		{ 426, 459, 55 },
		{ 426, 460, 55 },
		{ 426, 461, 55 },
		{ 426, 462, 55 },
		{ 426, 463, 55 },
		{ 426, 464, 55 },
		{ 426, 465, 55 },
		{ 426, 466, 55 },
		{ 426, 467, 55 },
		{ 426, 468, 55 },
		{ 426, 469, 55 },
		{ 426, 470, 55 },
		{ 426, 471, 55 },
		{ 426, 472, 55 },
		{ 426, 473, 55 },
		{ 426, 474, 55 },
		{ 426, 475, 55 },
		{ 426, 476, 55 },
		{ 426, 477, 55 },
		{ 426, 478, 55 },
		{ 426, 479, 55 },
		{ 426, 480, 55 },
		{ 426, 481, 55 },
		{ 426, 482, 55 },
		{ 427, 428, 55 },
		{ 427, 429, 55 },
		{ 427, 430, 55 },
		{ 427, 431, 55 },
		{ 427, 432, 55 },
		{ 427, 433, 55 },
		{ 427, 434, 55 },
		{ 427, 435, 55 },
		{ 427, 436, 55 },
		{ 427, 437, 55 },
		{ 427, 438, 55 },
		{ 427, 439, 55 },
		{ 427, 440, 55 },
		{ 427, 441, 55 },
		{ 427, 442, 55 },
		{ 427, 443, 55 },
		{ 427, 444, 55 },
		{ 427, 445, 55 },
		{ 427, 446, 55 },
		{ 427, 447, 55 },
		{ 427, 448, 55 },
		{ 427, 449, 55 },
		{ 427, 450, 55 },
		{ 427, 451, 55 },
		{ 427, 452, 55 },
		{ 427, 453, 55 },
		{ 427, 454, 55 },
		{ 427, 455, 55 },
		{ 427, 456, 55 },
		{ 427, 457, 55 },
		{ 427, 458, 55 },
		{ 427, 459, 55 },
		{ 427, 460, 55 },
		{ 427, 461, 55 },
		{ 427, 462, 55 },
		{ 427, 463, 55 },
		{ 427, 464, 55 },
		{ 427, 465, 55 },
		{ 427, 466, 55 },
		{ 427, 467, 55 },
		{ 427, 468, 55 },
		{ 427, 469, 55 },
		{ 427, 470, 55 },
		{ 427, 471, 55 },
		{ 427, 472, 55 },
		{ 427, 473, 55 },
		{ 427, 474, 55 },
		{ 427, 475, 55 },
		{ 427, 476, 55 },
		{ 427, 477, 55 },
		{ 427, 478, 55 },
		{ 427, 479, 55 },
		{ 427, 480, 55 },
		{ 427, 481, 55 },
		{ 427, 482, 55 },
		{ 427, 483, 55 },
		{ 428, 429, 55 },
		{ 428, 430, 55 },
		{ 428, 431, 55 },
		{ 428, 432, 55 },
		{ 428, 433, 55 },
		{ 428, 434, 55 },
		{ 428, 435, 55 },
		{ 428, 436, 55 },
		{ 428, 437, 55 },
		{ 428, 438, 55 },
		{ 428, 439, 55 },
		{ 428, 440, 55 },
		{ 428, 441, 55 },
		{ 428, 442, 55 },
		{ 428, 443, 55 },
		{ 428, 444, 55 },
		{ 428, 445, 55 },
		{ 428, 446, 55 },
		{ 428, 447, 55 },
		{ 428, 448, 55 },
		{ 428, 449, 55 },
		{ 428, 450, 55 },
		{ 428, 451, 55 },
		{ 428, 452, 55 },
		{ 428, 453, 55 },
		{ 428, 454, 55 },
		{ 428, 455, 55 },
		{ 428, 456, 55 },
		{ 428, 457, 55 },
		{ 428, 458, 55 },
		{ 428, 459, 55 },
		{ 428, 460, 55 },
		{ 428, 461, 55 },
		{ 428, 462, 55 },
		{ 428, 463, 55 },
		{ 428, 464, 55 },
		{ 428, 465, 55 },
		{ 428, 466, 55 },
		{ 428, 467, 55 },
		{ 428, 468, 55 },
		{ 428, 469, 55 },
		{ 428, 470, 55 },
		{ 428, 471, 55 },
		{ 428, 472, 55 },
		{ 428, 473, 55 },
		{ 428, 474, 55 },
		{ 428, 475, 55 },
		{ 428, 476, 55 },
		{ 428, 477, 55 },
		{ 428, 478, 55 },
		{ 428, 479, 55 },
		{ 428, 480, 55 },
		{ 428, 481, 55 },
		{ 428, 482, 55 },
		{ 428, 483, 55 },
		{ 428, 484, 55 },
		{ 429, 430, 55 },
		{ 429, 431, 55 },
		{ 429, 432, 55 },
		{ 429, 433, 55 },
		{ 429, 434, 55 },
		{ 429, 435, 55 },
		{ 429, 436, 55 },
		{ 429, 437, 55 },
		{ 429, 438, 55 },
		{ 429, 439, 55 },
		{ 429, 440, 55 },
		{ 429, 441, 55 },
		{ 429, 442, 55 },
		{ 429, 443, 55 },
		{ 429, 444, 55 },
		{ 429, 445, 55 },
		{ 429, 446, 55 },
		{ 429, 447, 55 },
		{ 429, 448, 55 },
		{ 429, 449, 55 },
		{ 429, 450, 55 },
		{ 429, 451, 55 },
		{ 429, 452, 55 },
		{ 429, 453, 55 },
		{ 429, 454, 55 },
		{ 429, 455, 55 },
		{ 429, 456, 55 },
		{ 429, 457, 55 },
		{ 429, 458, 55 },
		{ 429, 459, 55 },
		{ 429, 460, 55 },
		{ 429, 461, 55 },
		{ 429, 462, 55 },
		{ 429, 463, 55 },
		{ 429, 464, 55 },
		{ 429, 465, 55 },
		{ 429, 466, 55 },
		{ 429, 467, 55 },
		{ 429, 468, 55 },
		{ 429, 469, 55 },
		{ 429, 470, 55 },
		{ 429, 471, 55 },
		{ 429, 472, 55 },
		{ 429, 473, 55 },
		{ 429, 474, 55 },
		{ 429, 475, 55 },
		{ 429, 476, 55 },
		{ 429, 477, 55 },
		{ 429, 478, 55 },
		{ 429, 479, 55 },
		{ 429, 480, 55 },
		{ 429, 481, 55 },
		{ 429, 482, 55 },
		{ 429, 483, 55 },
		{ 429, 484, 55 },
		{ 429, 485, 55 },
		{ 429, 486, 55 },
		{ 430, 431, 55 },
		{ 430, 432, 55 },
		{ 430, 433, 55 },
		{ 430, 434, 55 },
		{ 430, 435, 55 },
		{ 430, 436, 55 },
		{ 430, 437, 55 },
		{ 430, 438, 55 },
		{ 430, 439, 55 },
		{ 430, 440, 55 },
		{ 430, 441, 55 },
		{ 430, 442, 55 },
		{ 430, 443, 55 },
		{ 430, 444, 55 },
		{ 430, 445, 55 },
		{ 430, 446, 55 },
		{ 430, 447, 55 },
		{ 430, 448, 55 },
		{ 430, 449, 55 },
		{ 430, 450, 55 },
		{ 430, 451, 55 },
		{ 430, 452, 55 },
		{ 430, 453, 55 },
		{ 430, 454, 55 },
		{ 430, 455, 55 },
		{ 430, 456, 55 },
		{ 430, 457, 55 },
		{ 430, 458, 55 },
		{ 430, 459, 55 },
		{ 430, 460, 55 },
		{ 430, 461, 55 },
		{ 430, 462, 55 },
		{ 430, 463, 55 },
		{ 430, 464, 55 },
		{ 430, 465, 55 },
		{ 430, 466, 55 },
		{ 430, 467, 55 },
		{ 430, 468, 55 },
		{ 430, 469, 55 },
		{ 430, 470, 55 },
		{ 430, 471, 55 },
		{ 430, 472, 55 },
		{ 430, 473, 55 },
		{ 430, 474, 55 },
		{ 430, 475, 55 },
		{ 430, 476, 55 },
		{ 430, 477, 55 },
		{ 430, 478, 55 },
		{ 430, 479, 55 },
		{ 430, 480, 55 },
		{ 430, 481, 55 },
		{ 430, 482, 55 },
		{ 430, 483, 55 },
		{ 430, 484, 55 },
		{ 430, 485, 55 },
		{ 430, 486, 55 },
		{ 430, 487, 55 },
		{ 431, 432, 55 },
		{ 431, 433, 55 },
		{ 431, 434, 55 },
		{ 431, 435, 55 },
		{ 431, 436, 55 },
		{ 431, 437, 55 },
		{ 431, 438, 55 },
		{ 431, 439, 55 },
		{ 431, 440, 55 },
		{ 431, 441, 55 },
		{ 431, 442, 55 },
		{ 431, 443, 55 },
		{ 431, 444, 55 },
		{ 431, 445, 55 },
		{ 431, 446, 55 },
		{ 431, 447, 55 },
		{ 431, 448, 55 },
		{ 431, 449, 55 },
		{ 431, 450, 55 },
		{ 431, 451, 55 },
		{ 431, 452, 55 },
		{ 431, 453, 55 },
		{ 431, 454, 55 },
		{ 431, 455, 55 },
		{ 431, 456, 55 },
		{ 431, 457, 55 },
		{ 431, 458, 55 },
		{ 431, 459, 55 },
		{ 431, 460, 55 },
		{ 431, 461, 55 },
		{ 431, 462, 55 },
		{ 431, 463, 55 },
		{ 431, 464, 55 },
		{ 431, 465, 55 },
		{ 431, 466, 55 },
		{ 431, 467, 55 },
		{ 431, 468, 55 },
		{ 431, 469, 55 },
		{ 431, 470, 55 },
		{ 431, 471, 55 },
		{ 431, 472, 55 },
		{ 431, 473, 55 },
		{ 431, 474, 55 },
		{ 431, 475, 55 },
		{ 431, 476, 55 },
		{ 431, 477, 55 },
		{ 431, 478, 55 },
		{ 431, 479, 55 },
		{ 431, 480, 55 },
		{ 431, 481, 55 },
		{ 431, 482, 55 },
		{ 431, 483, 55 },
		{ 431, 484, 55 },
		{ 431, 485, 55 },
		{ 431, 486, 55 },
		{ 431, 487, 55 },
		{ 431, 488, 55 },
		{ 432, 433, 55 },
		{ 432, 434, 55 },
		{ 432, 435, 55 },
		{ 432, 436, 55 },
		{ 432, 437, 55 },
		{ 432, 438, 55 },
		{ 432, 439, 55 },
		{ 432, 440, 55 },
		{ 432, 441, 55 },
		{ 432, 442, 55 },
		{ 432, 443, 55 },
		{ 432, 444, 55 },
		{ 432, 445, 55 },
		{ 432, 446, 55 },
		{ 432, 447, 55 },
		{ 432, 448, 55 },
		{ 432, 449, 55 },
		{ 432, 450, 55 },
		{ 432, 451, 55 },
		{ 432, 452, 55 },
		{ 432, 453, 55 },
		{ 432, 454, 55 },
		{ 432, 455, 55 },
		{ 432, 456, 55 },
		{ 432, 457, 55 },
		{ 432, 458, 55 },
		{ 432, 459, 55 },
		{ 432, 460, 55 },
		{ 432, 461, 55 },
		{ 432, 462, 55 },
		{ 432, 463, 55 },
		{ 432, 464, 55 },
		{ 432, 465, 55 },
		{ 432, 466, 55 },
		{ 432, 467, 55 },
		{ 432, 468, 55 },
		{ 432, 469, 55 },
		{ 432, 470, 55 },
		{ 432, 471, 55 },
		{ 432, 472, 55 },
		{ 432, 473, 55 },
		{ 432, 474, 55 },
		{ 432, 475, 55 },
		{ 432, 476, 55 },
		{ 432, 477, 55 },
		{ 432, 478, 55 },
		{ 432, 479, 55 },
		{ 432, 480, 55 },
		{ 432, 481, 55 },
		{ 432, 482, 55 },
		{ 432, 483, 55 },
		{ 432, 484, 55 },
		{ 432, 485, 55 },
		{ 432, 486, 55 },
		{ 432, 487, 55 },
		{ 432, 488, 55 },
		{ 432, 489, 55 },
		{ 432, 490, 55 },
		{ 433, 434, 55 },
		{ 433, 435, 55 },
		{ 433, 436, 55 },
		{ 433, 437, 55 },
		{ 433, 438, 55 },
		{ 433, 439, 55 },
		{ 433, 440, 55 },
		{ 433, 441, 55 },
		{ 433, 442, 55 },
		{ 433, 443, 55 },
		{ 433, 444, 55 },
		{ 433, 445, 55 },
		{ 433, 446, 55 },
		{ 433, 447, 55 },
		{ 433, 448, 55 },
		{ 433, 449, 55 },
		{ 433, 450, 55 },
		{ 433, 451, 55 },
		{ 433, 452, 55 },
		{ 433, 453, 55 },
		{ 433, 454, 55 },
		{ 433, 455, 55 },
		{ 433, 456, 55 },
		{ 433, 457, 55 },
		{ 433, 458, 55 },
		{ 433, 459, 55 },
		{ 433, 460, 55 },
		{ 433, 461, 55 },
		{ 433, 462, 55 },
		{ 433, 463, 55 },
		{ 433, 464, 55 },
		{ 433, 465, 55 },
		{ 433, 466, 55 },
		{ 433, 467, 55 },
		{ 433, 468, 55 },
		{ 433, 469, 55 },
		{ 433, 470, 55 },
		{ 433, 471, 55 },
		{ 433, 472, 55 },
		{ 433, 473, 55 },
		{ 433, 474, 55 },
		{ 433, 475, 55 },
		{ 433, 476, 55 },
		{ 433, 477, 55 },
		{ 433, 478, 55 },
		{ 433, 479, 55 },
		{ 433, 480, 55 },
		{ 433, 481, 55 },
		{ 433, 482, 55 },
		{ 433, 483, 55 },
		{ 433, 484, 55 },
		{ 433, 485, 55 },
		{ 433, 486, 55 },
		{ 433, 487, 55 },
		{ 433, 488, 55 },
		{ 433, 489, 55 },
		{ 433, 490, 55 },
		{ 433, 491, 55 },
		{ 434, 435, 55 },
		{ 434, 436, 55 },
		{ 434, 437, 55 },
		{ 434, 438, 55 },
		{ 434, 439, 55 },
		{ 434, 440, 55 },
		{ 434, 441, 55 },
		{ 434, 442, 55 },
		{ 434, 443, 55 },
		{ 434, 444, 55 },
		{ 434, 445, 55 },
		{ 434, 446, 55 },
		{ 434, 447, 55 },
		{ 434, 448, 55 },
		{ 434, 449, 55 },
		{ 434, 450, 55 },
		{ 434, 451, 55 },
		{ 434, 452, 55 },
		{ 434, 453, 55 },
		{ 434, 454, 55 },
		{ 434, 455, 55 },
		{ 434, 456, 55 },
		{ 434, 457, 55 },
		{ 434, 458, 55 },
		{ 434, 459, 55 },
		{ 434, 460, 55 },
		{ 434, 461, 55 },
		{ 434, 462, 55 },
		{ 434, 463, 55 },
		{ 434, 464, 55 },
		{ 434, 465, 55 },
		{ 434, 466, 55 },
		{ 434, 467, 55 },
		{ 434, 468, 55 },
		{ 434, 469, 55 },
		{ 434, 470, 55 },
		{ 434, 471, 55 },
		{ 434, 472, 55 },
		{ 434, 473, 55 },
		{ 434, 474, 55 },
		{ 434, 475, 55 },
		{ 434, 476, 55 },
		{ 434, 477, 55 },
		{ 434, 478, 55 },
		{ 434, 479, 55 },
		{ 434, 480, 55 },
		{ 434, 481, 55 },
		{ 434, 482, 55 },
		{ 434, 483, 55 },
		{ 434, 484, 55 },
		{ 434, 485, 55 },
		{ 434, 486, 55 },
		{ 434, 487, 55 },
		{ 434, 488, 55 },
		{ 434, 489, 55 },
		{ 434, 490, 55 },
		{ 434, 491, 55 },
		{ 434, 492, 55 },
		{ 435, 436, 55 },
		{ 435, 437, 55 },
		{ 435, 438, 55 },
		{ 435, 439, 55 },
		{ 435, 440, 55 },
		{ 435, 441, 55 },
		{ 435, 442, 55 },
		{ 435, 443, 55 },
		{ 435, 444, 55 },
		{ 435, 445, 55 },
		{ 435, 446, 55 },
		{ 435, 447, 55 },
		{ 435, 448, 55 },
		{ 435, 449, 55 },
		{ 435, 450, 55 },
		{ 435, 451, 55 },
		{ 435, 452, 55 },
		{ 435, 453, 55 },
		{ 435, 454, 55 },
		{ 435, 455, 55 },
		{ 435, 456, 55 },
		{ 435, 457, 55 },
		{ 435, 458, 55 },
		{ 435, 459, 55 },
		{ 435, 460, 55 },
		{ 435, 461, 55 },
		{ 435, 462, 55 },
		{ 435, 463, 55 },
		{ 435, 464, 55 },
		{ 435, 465, 55 },
		{ 435, 466, 55 },
		{ 435, 467, 55 },
		{ 435, 468, 55 },
		{ 435, 469, 55 },
		{ 435, 470, 55 },
		{ 435, 471, 55 },
		{ 435, 472, 55 },
		{ 435, 473, 55 },
		{ 435, 474, 55 },
		{ 435, 475, 55 },
		{ 435, 476, 55 },
		{ 435, 477, 55 },
		{ 435, 478, 55 },
		{ 435, 479, 55 },
		{ 435, 480, 55 },
		{ 435, 481, 55 },
		{ 435, 482, 55 },
		{ 435, 483, 55 },
		{ 435, 484, 55 },
		{ 435, 485, 55 },
		{ 435, 486, 55 },
		{ 435, 487, 55 },
		{ 435, 488, 55 },
		{ 435, 489, 55 },
		{ 435, 490, 55 },
		{ 435, 491, 55 },
		{ 435, 492, 55 },
		{ 435, 493, 55 },
		{ 435, 494, 55 },
		{ 436, 437, 55 },
		{ 436, 438, 55 },
		{ 436, 439, 55 },
		{ 436, 440, 55 },
		{ 436, 441, 55 },
		{ 436, 442, 55 },
		{ 436, 443, 55 },
		{ 436, 444, 55 },
		{ 436, 445, 55 },
		{ 436, 446, 55 },
		{ 436, 447, 55 },
		{ 436, 448, 55 },
		{ 436, 449, 55 },
		{ 436, 450, 55 },
		{ 436, 451, 55 },
		{ 436, 452, 55 },
		{ 436, 453, 55 },
		{ 436, 454, 55 },
		{ 436, 455, 55 },
		{ 436, 456, 55 },
		{ 436, 457, 55 },
		{ 436, 458, 55 },
		{ 436, 459, 55 },
		{ 436, 460, 55 },
		{ 436, 461, 55 },
		{ 436, 462, 55 },
		{ 436, 463, 55 },
		{ 436, 464, 55 },
		{ 436, 465, 55 },
		{ 436, 466, 55 },
		{ 436, 467, 55 },
		{ 436, 468, 55 },
		{ 436, 469, 55 },
		{ 436, 470, 55 },
		{ 436, 471, 55 },
		{ 436, 472, 55 },
		{ 436, 473, 55 },
		{ 436, 474, 55 },
		{ 436, 475, 55 },
		{ 436, 476, 55 },
		{ 436, 477, 55 },
		{ 436, 478, 55 },
		{ 436, 479, 55 },
		{ 436, 480, 55 },
		{ 436, 481, 55 },
		{ 436, 482, 55 },
		{ 436, 483, 55 },
		{ 436, 484, 55 },
		{ 436, 485, 55 },
		{ 436, 486, 55 },
		{ 436, 487, 55 },
		{ 436, 488, 55 },
		{ 436, 489, 55 },
		{ 436, 490, 55 },
		{ 436, 491, 55 },
		{ 436, 492, 55 },
		{ 436, 493, 55 },
		{ 436, 494, 55 },
		{ 436, 495, 55 },
		{ 437, 438, 55 },
		{ 437, 439, 55 },
		{ 437, 440, 55 },
		{ 437, 441, 55 },
		{ 437, 442, 55 },
		{ 437, 443, 55 },
		{ 437, 444, 55 },
		{ 437, 445, 55 },
		{ 437, 446, 55 },
		{ 437, 447, 55 },
		{ 437, 448, 55 },
		{ 437, 449, 55 },
		{ 437, 450, 55 },
		{ 437, 451, 55 },
		{ 437, 452, 55 },
		{ 437, 453, 55 },
		{ 437, 454, 55 },
		{ 437, 455, 55 },
		{ 437, 456, 55 },
		{ 437, 457, 55 },
		{ 437, 458, 55 },
		{ 437, 459, 55 },
		{ 437, 460, 55 },
		{ 437, 461, 55 },
		{ 437, 462, 55 },
		{ 437, 463, 55 },
		{ 437, 464, 55 },
		{ 437, 465, 55 },
		{ 437, 466, 55 },
		{ 437, 467, 55 },
		{ 437, 468, 55 },
		{ 437, 469, 55 },
		{ 437, 470, 55 },
		{ 437, 471, 55 },
		{ 437, 472, 55 },
		{ 437, 473, 55 },
		{ 437, 474, 55 },
		{ 437, 475, 55 },
		{ 437, 476, 55 },
		{ 437, 477, 55 },
		{ 437, 478, 55 },
		{ 437, 479, 55 },
		{ 437, 480, 55 },
		{ 437, 481, 55 },
		{ 437, 482, 55 },
		{ 437, 483, 55 },
		{ 437, 484, 55 },
		{ 437, 485, 55 },
		{ 437, 486, 55 },
		{ 437, 487, 55 },
		{ 437, 488, 55 },
		{ 437, 489, 55 },
		{ 437, 490, 55 },
		{ 437, 491, 55 },
		{ 437, 492, 55 },
		{ 437, 493, 55 },
		{ 437, 494, 55 },
		{ 437, 495, 55 },
		{ 437, 496, 55 },
		{ 438, 439, 55 },
		{ 438, 440, 55 },
		{ 438, 441, 55 },
		{ 438, 442, 55 },
		{ 438, 443, 55 },
		{ 438, 444, 55 },
		{ 438, 445, 55 },
		{ 438, 446, 55 },
		{ 438, 447, 55 },
		{ 438, 448, 55 },
		{ 438, 449, 55 },
		{ 438, 450, 55 },
		{ 438, 451, 55 },
		{ 438, 452, 55 },
		{ 438, 453, 55 },
		{ 438, 454, 55 },
		{ 438, 455, 55 },
		{ 438, 456, 55 },
		{ 438, 457, 55 },
		{ 438, 458, 55 },
		{ 438, 459, 55 },
		{ 438, 460, 55 },
		{ 438, 461, 55 },
		{ 438, 462, 55 },
		{ 438, 463, 55 },
		{ 438, 464, 55 },
		{ 438, 465, 55 },
		{ 438, 466, 55 },
		{ 438, 467, 55 },
		{ 438, 468, 55 },
		{ 438, 469, 55 },
		{ 438, 470, 55 },
		{ 438, 471, 55 },
		{ 438, 472, 55 },
		{ 438, 473, 55 },
		{ 438, 474, 55 },
		{ 438, 475, 55 },
		{ 438, 476, 55 },
		{ 438, 477, 55 },
		{ 438, 478, 55 },
		{ 438, 479, 55 },
		{ 438, 480, 55 },
		{ 438, 481, 55 },
		{ 438, 482, 55 },
		{ 438, 483, 55 },
		{ 438, 484, 55 },
		{ 438, 485, 55 },
		{ 438, 486, 55 },
		{ 438, 487, 55 },
		{ 438, 488, 55 },
		{ 438, 489, 55 },
		{ 438, 490, 55 },
		{ 438, 491, 55 },
		{ 438, 492, 55 },
		{ 438, 493, 55 },
		{ 438, 494, 55 },
		{ 438, 495, 55 },
		{ 438, 496, 55 },
		{ 438, 497, 55 },
		{ 438, 498, 55 },
		{ 439, 440, 55 },
		{ 439, 441, 55 },
		{ 439, 442, 55 },
		{ 439, 443, 55 },
		{ 439, 444, 55 },
		{ 439, 445, 55 },
		{ 439, 446, 55 },
		{ 439, 447, 55 },
		{ 439, 448, 55 },
		{ 439, 449, 55 },
		{ 439, 450, 55 },
		{ 439, 451, 55 },
		{ 439, 452, 55 },
		{ 439, 453, 55 },
		{ 439, 454, 55 },
		{ 439, 455, 55 },
		{ 439, 456, 55 },
		{ 439, 457, 55 },
		{ 439, 458, 55 },
		{ 439, 459, 55 },
		{ 439, 460, 55 },
		{ 439, 461, 55 },
		{ 439, 462, 55 },
		{ 439, 463, 55 },
		{ 439, 464, 55 },
		{ 439, 465, 55 },
		{ 439, 466, 55 },
		{ 439, 467, 55 },
		{ 439, 468, 55 },
		{ 439, 469, 55 },
		{ 439, 470, 55 },
		{ 439, 471, 55 },
		{ 439, 472, 55 },
		{ 439, 473, 55 },
		{ 439, 474, 55 },
		{ 439, 475, 55 },
		{ 439, 476, 55 },
		{ 439, 477, 55 },
		{ 439, 478, 55 },
		{ 439, 479, 55 },
		{ 439, 480, 55 },
		{ 439, 481, 55 },
		{ 439, 482, 55 },
		{ 439, 483, 55 },
		{ 439, 484, 55 },
		{ 439, 485, 55 },
		{ 439, 486, 55 },
		{ 439, 487, 55 },
		{ 439, 488, 55 },
		{ 439, 489, 55 },
		{ 439, 490, 55 },
		{ 439, 491, 55 },
		{ 439, 492, 55 },
		{ 439, 493, 55 },
		{ 439, 494, 55 },
		{ 439, 495, 55 },
		{ 439, 496, 55 },
		{ 439, 497, 55 },
		{ 439, 498, 55 },
		{ 439, 499, 55 },
		{ 440, 441, 55 },
		{ 440, 442, 55 },
		{ 440, 443, 55 },
		{ 440, 444, 55 },
		{ 440, 445, 55 },
		{ 440, 446, 55 },
		{ 440, 447, 55 },
		{ 440, 448, 55 },
		{ 440, 449, 55 },
		{ 440, 450, 55 },
		{ 440, 451, 55 },
		{ 440, 452, 55 },
		{ 440, 453, 55 },
		{ 440, 454, 55 },
		{ 440, 455, 55 },
		{ 440, 456, 55 },
		{ 440, 457, 55 },
		{ 440, 458, 55 },
		{ 440, 459, 55 },
		{ 440, 460, 55 },
		{ 440, 461, 55 },
		{ 440, 462, 55 },
		{ 440, 463, 55 },
		{ 440, 464, 55 },
		{ 440, 465, 55 },
		{ 440, 466, 55 },
		{ 440, 467, 55 },
		{ 440, 468, 55 },
		{ 440, 469, 55 },
		{ 440, 470, 55 },
		{ 440, 471, 55 },
		{ 440, 472, 55 },
		{ 440, 473, 55 },
		{ 440, 474, 55 },
		{ 440, 475, 55 },
		{ 440, 476, 55 },
		{ 440, 477, 55 },
		{ 440, 478, 55 },
		{ 440, 479, 55 },
		{ 440, 480, 55 },
		{ 440, 481, 55 },
		{ 440, 482, 55 },
		{ 440, 483, 55 },
		{ 440, 484, 55 },
		{ 440, 485, 55 },
		{ 440, 486, 55 },
		{ 440, 487, 55 },
		{ 440, 488, 55 },
		{ 440, 489, 55 },
		{ 440, 490, 55 },
		{ 440, 491, 55 },
		{ 440, 492, 55 },
		{ 440, 493, 55 },
		{ 440, 494, 55 },
		{ 440, 495, 55 },
		{ 440, 496, 55 },
		{ 440, 497, 55 },
		{ 440, 498, 55 },
		{ 440, 499, 55 },
		{ 440, 500, 55 },
		{ 441, 442, 55 },
		{ 441, 443, 55 },
		{ 441, 444, 55 },
		{ 441, 445, 55 },
		{ 441, 446, 55 },
		{ 441, 447, 55 },
		{ 441, 448, 55 },
		{ 441, 449, 55 },
		{ 441, 450, 55 },
		{ 441, 451, 55 },
		{ 441, 452, 55 },
		{ 441, 453, 55 },
		{ 441, 454, 55 },
		{ 441, 455, 55 },
		{ 441, 456, 55 },
		{ 441, 457, 55 },
		{ 441, 458, 55 },
		{ 441, 459, 55 },
		{ 441, 460, 55 },
		{ 441, 461, 55 },
		{ 441, 462, 55 },
		{ 441, 463, 55 },
		{ 441, 464, 55 },
		{ 441, 465, 55 },
		{ 441, 466, 55 },
		{ 441, 467, 55 },
		{ 441, 468, 55 },
		{ 441, 469, 55 },
		{ 441, 470, 55 },
		{ 441, 471, 55 },
		{ 441, 472, 55 },
		{ 441, 473, 55 },
		{ 441, 474, 55 },
		{ 441, 475, 55 },
		{ 441, 476, 55 },
		{ 441, 477, 55 },
		{ 441, 478, 55 },
		{ 441, 479, 55 },
		{ 441, 480, 55 },
		{ 441, 481, 55 },
		{ 441, 482, 55 },
		{ 441, 483, 55 },
		{ 441, 484, 55 },
		{ 441, 485, 55 },
		{ 441, 486, 55 },
		{ 441, 487, 55 },
		{ 441, 488, 55 },
		{ 441, 489, 55 },
		{ 441, 490, 55 },
		{ 441, 491, 55 },
		{ 441, 492, 55 },
		{ 441, 493, 55 },
		{ 441, 494, 55 },
		{ 441, 495, 55 },
		{ 441, 496, 55 },
		{ 441, 497, 55 },
		{ 441, 498, 55 },
		{ 441, 499, 55 },
		{ 441, 500, 55 },
		{ 441, 501, 55 },
		{ 441, 502, 55 },
		{ 442, 443, 55 },
		{ 442, 444, 55 },
		{ 442, 445, 55 },
		{ 442, 446, 55 },
		{ 442, 447, 55 },
		{ 442, 448, 55 },
		{ 442, 449, 55 },
		{ 442, 450, 55 },
		{ 442, 451, 55 },
		{ 442, 452, 55 },
		{ 442, 453, 55 },
		{ 442, 454, 55 },
		{ 442, 455, 55 },
		{ 442, 456, 55 },
		{ 442, 457, 55 },
		{ 442, 458, 55 },
		{ 442, 459, 55 },
		{ 442, 460, 55 },
		{ 442, 461, 55 },
		{ 442, 462, 55 },
		{ 442, 463, 55 },
		{ 442, 464, 55 },
		{ 442, 465, 55 },
		{ 442, 466, 55 },
		{ 442, 467, 55 },
		{ 442, 468, 55 },
		{ 442, 469, 55 },
		{ 442, 470, 55 },
		{ 442, 471, 55 },
		{ 442, 472, 55 },
		{ 442, 473, 55 },
		{ 442, 474, 55 },
		{ 442, 475, 55 },
		{ 442, 476, 55 },
		{ 442, 477, 55 },
		{ 442, 478, 55 },
		{ 442, 479, 55 },
		{ 442, 480, 55 },
		{ 442, 481, 55 },
		{ 442, 482, 55 },
		{ 442, 483, 55 },
		{ 442, 484, 55 },
		{ 442, 485, 55 },
		{ 442, 486, 55 },
		{ 442, 487, 55 },
		{ 442, 488, 55 },
		{ 442, 489, 55 },
		{ 442, 490, 55 },
		{ 442, 491, 55 },
		{ 442, 492, 55 },
		{ 442, 493, 55 },
		{ 442, 494, 55 },
		{ 442, 495, 55 },
		{ 442, 496, 55 },
		{ 442, 497, 55 },
		{ 442, 498, 55 },
		{ 442, 499, 55 },
		{ 442, 500, 55 },
		{ 442, 501, 55 },
		{ 442, 502, 55 },
		{ 442, 503, 55 },
		{ 443, 444, 55 },
		{ 443, 445, 55 },
		{ 443, 446, 55 },
		{ 443, 447, 55 },
		{ 443, 448, 55 },
		{ 443, 449, 55 },
		{ 443, 450, 55 },
		{ 443, 451, 55 },
		{ 443, 452, 55 },
		{ 443, 453, 55 },
		{ 443, 454, 55 },
		{ 443, 455, 55 },
		{ 443, 456, 55 },
		{ 443, 457, 55 },
		{ 443, 458, 55 },
		{ 443, 459, 55 },
		{ 443, 460, 55 },
		{ 443, 461, 55 },
		{ 443, 462, 55 },
		{ 443, 463, 55 },
		{ 443, 464, 55 },
		{ 443, 465, 55 },
		{ 443, 466, 55 },
		{ 443, 467, 55 },
		{ 443, 468, 55 },
		{ 443, 469, 55 },
		{ 443, 470, 55 },
		{ 443, 471, 55 },
		{ 443, 472, 55 },
		{ 443, 473, 55 },
		{ 443, 474, 55 },
		{ 443, 475, 55 },
		{ 443, 476, 55 },
		{ 443, 477, 55 },
		{ 443, 478, 55 },
		{ 443, 479, 55 },
		{ 443, 480, 55 },
		{ 443, 481, 55 },
		{ 443, 482, 55 },
		{ 443, 483, 55 },
		{ 443, 484, 55 },
		{ 443, 485, 55 },
		{ 443, 486, 55 },
		{ 443, 487, 55 },
		{ 443, 488, 55 },
		{ 443, 489, 55 },
		{ 443, 490, 55 },
		{ 443, 491, 55 },
		{ 443, 492, 55 },
		{ 443, 493, 55 },
		{ 443, 494, 55 },
		{ 443, 495, 55 },
		{ 443, 496, 55 },
		{ 443, 497, 55 },
		{ 443, 498, 55 },
		{ 443, 499, 55 },
		{ 443, 500, 55 },
		{ 443, 501, 55 },
		{ 443, 502, 55 },
		{ 443, 503, 55 },
		{ 443, 504, 55 },
		{ 444, 445, 55 },
		{ 444, 446, 55 },
		{ 444, 447, 55 },
		{ 444, 448, 55 },
		{ 444, 449, 55 },
		{ 444, 450, 55 },
		{ 444, 451, 55 },
		{ 444, 452, 55 },
		{ 444, 453, 55 },
		{ 444, 454, 55 },
		{ 444, 455, 55 },
		{ 444, 456, 55 },
		{ 444, 457, 55 },
		{ 444, 458, 55 },
		{ 444, 459, 55 },
		{ 444, 460, 55 },
		{ 444, 461, 55 },
		{ 444, 462, 55 },
		{ 444, 463, 55 },
		{ 444, 464, 55 },
		{ 444, 465, 55 },
		{ 444, 466, 55 },
		{ 444, 467, 55 },
		{ 444, 468, 55 },
		{ 444, 469, 55 },
		{ 444, 470, 55 },
		{ 444, 471, 55 },
		{ 444, 472, 55 },
		{ 444, 473, 55 },
		{ 444, 474, 55 },
		{ 444, 475, 55 },
		{ 444, 476, 55 },
		{ 444, 477, 55 },
		{ 444, 478, 55 },
		{ 444, 479, 55 },
		{ 444, 480, 55 },
		{ 444, 481, 55 },
		{ 444, 482, 55 },
		{ 444, 483, 55 },
		{ 444, 484, 55 },
		{ 444, 485, 55 },
		{ 444, 486, 55 },
		{ 444, 487, 55 },
		{ 444, 488, 55 },
		{ 444, 489, 55 },
		{ 444, 490, 55 },
		{ 444, 491, 55 },
		{ 444, 492, 55 },
		{ 444, 493, 55 },
		{ 444, 494, 55 },
		{ 444, 495, 55 },
		{ 444, 496, 55 },
		{ 444, 497, 55 },
		{ 444, 498, 55 },
		{ 444, 499, 55 },
		{ 444, 500, 55 },
		{ 444, 501, 55 },
		{ 444, 502, 55 },
		{ 444, 503, 55 },
		{ 444, 504, 55 },
		{ 444, 505, 55 },
		{ 444, 506, 55 },
		{ 445, 446, 55 },
		{ 445, 447, 55 },
		{ 445, 448, 55 },
		{ 445, 449, 55 },
		{ 445, 450, 55 },
		{ 445, 451, 55 },
		{ 445, 452, 55 },
		{ 445, 453, 55 },
		{ 445, 454, 55 },
		{ 445, 455, 55 },
		{ 445, 456, 55 },
		{ 445, 457, 55 },
		{ 445, 458, 55 },
		{ 445, 459, 55 },
		{ 445, 460, 55 },
		{ 445, 461, 55 },
		{ 445, 462, 55 },
		{ 445, 463, 55 },
		{ 445, 464, 55 },
		{ 445, 465, 55 },
		{ 445, 466, 55 },
		{ 445, 467, 55 },
		{ 445, 468, 55 },
		{ 445, 469, 55 },
		{ 445, 470, 55 },
		{ 445, 471, 55 },
		{ 445, 472, 55 },
		{ 445, 473, 55 },
		{ 445, 474, 55 },
		{ 445, 475, 55 },
		{ 445, 476, 55 },
		{ 445, 477, 55 },
		{ 445, 478, 55 },
		{ 445, 479, 55 },
		{ 445, 480, 55 },
		{ 445, 481, 55 },
		{ 445, 482, 55 },
		{ 445, 483, 55 },
		{ 445, 484, 55 },
		{ 445, 485, 55 },
		{ 445, 486, 55 },
		{ 445, 487, 55 },
		{ 445, 488, 55 },
		{ 445, 489, 55 },
		{ 445, 490, 55 },
		{ 445, 491, 55 },
		{ 445, 492, 55 },
		{ 445, 493, 55 },
		{ 445, 494, 55 },
		{ 445, 495, 55 },
		{ 445, 496, 55 },
		{ 445, 497, 55 },
		{ 445, 498, 55 },
		{ 445, 499, 55 },
		{ 445, 500, 55 },
		{ 445, 501, 55 },
		{ 445, 502, 55 },
		{ 445, 503, 55 },
		{ 445, 504, 55 },
		{ 445, 505, 55 },
		{ 445, 506, 55 },
		{ 445, 507, 55 },
		{ 446, 447, 55 },
		{ 446, 448, 55 },
		{ 446, 449, 55 },
		{ 446, 450, 55 },
		{ 446, 451, 55 },
		{ 446, 452, 55 },
		{ 446, 453, 55 },
		{ 446, 454, 55 },
		{ 446, 455, 55 },
		{ 446, 456, 55 },
		{ 446, 457, 55 },
		{ 446, 458, 55 },
		{ 446, 459, 55 },
		{ 446, 460, 55 },
		{ 446, 461, 55 },
		{ 446, 462, 55 },
		{ 446, 463, 55 },
		{ 446, 464, 55 },
		{ 446, 465, 55 },
		{ 446, 466, 55 },
		{ 446, 467, 55 },
		{ 446, 468, 55 },
		{ 446, 469, 55 },
		{ 446, 470, 55 },
		{ 446, 471, 55 },
		{ 446, 472, 55 },
		{ 446, 473, 55 },
		{ 446, 474, 55 },
		{ 446, 475, 55 },
		{ 446, 476, 55 },
		{ 446, 477, 55 },
		{ 446, 478, 55 },
		{ 446, 479, 55 },
		{ 446, 480, 55 },
		{ 446, 481, 55 },
		{ 446, 482, 55 },
		{ 446, 483, 55 },
		{ 446, 484, 55 },
		{ 446, 485, 55 },
		{ 446, 486, 55 },
		{ 446, 487, 55 },
		{ 446, 488, 55 },
		{ 446, 489, 55 },
		{ 446, 490, 55 },
		{ 446, 491, 55 },
		{ 446, 492, 55 },
		{ 446, 493, 55 },
		{ 446, 494, 55 },
		{ 446, 495, 55 },
		{ 446, 496, 55 },
		{ 446, 497, 55 },
		{ 446, 498, 55 },
		{ 446, 499, 55 },
		{ 446, 500, 55 },
		{ 446, 501, 55 },
		{ 446, 502, 55 },
		{ 446, 503, 55 },
		{ 446, 504, 55 },
		{ 446, 505, 55 },
		{ 446, 506, 55 },
		{ 446, 507, 55 },
		{ 446, 508, 55 },
		{ 447, 448, 55 },
		{ 447, 449, 55 },
		{ 447, 450, 55 },
		{ 447, 451, 55 },
		{ 447, 452, 55 },
		{ 447, 453, 55 },
		{ 447, 454, 55 },
		{ 447, 455, 55 },
		{ 447, 456, 55 },
		{ 447, 457, 55 },
		{ 447, 458, 55 },
		{ 447, 459, 55 },
		{ 447, 460, 55 },
		{ 447, 461, 55 },
		{ 447, 462, 55 },
		{ 447, 463, 55 },
		{ 447, 464, 55 },
		{ 447, 465, 55 },
		{ 447, 466, 55 },
		{ 447, 467, 55 },
		{ 447, 468, 55 },
		{ 447, 469, 55 },
		{ 447, 470, 55 },
		{ 447, 471, 55 },
		{ 447, 472, 55 },
		{ 447, 473, 55 },
		{ 447, 474, 55 },
		{ 447, 475, 55 },
		{ 447, 476, 55 },
		{ 447, 477, 55 },
		{ 447, 478, 55 },
		{ 447, 479, 55 },
		{ 447, 480, 55 },
		{ 447, 481, 55 },
		{ 447, 482, 55 },
		{ 447, 483, 55 },
		{ 447, 484, 55 },
		{ 447, 485, 55 },
		{ 447, 486, 55 },
		{ 447, 487, 55 },
		{ 447, 488, 55 },
		{ 447, 489, 55 },
		{ 447, 490, 55 },
		{ 447, 491, 55 },
		{ 447, 492, 55 },
		{ 447, 493, 55 },
		{ 447, 494, 55 },
		{ 447, 495, 55 },
		{ 447, 496, 55 },
		{ 447, 497, 55 },
		{ 447, 498, 55 },
		{ 447, 499, 55 },
		{ 447, 500, 55 },
		{ 447, 501, 55 },
		{ 447, 502, 55 },
		{ 447, 503, 55 },
		{ 447, 504, 55 },
		{ 447, 505, 55 },
		{ 447, 506, 55 },
		{ 447, 507, 55 },
		{ 447, 508, 55 },
		{ 447, 509, 55 },
		{ 447, 510, 55 },
		{ 448, 449, 55 },
		{ 448, 450, 55 },
		{ 448, 451, 55 },
		{ 448, 452, 55 },
		{ 448, 453, 55 },
		{ 448, 454, 55 },
		{ 448, 455, 55 },
		{ 448, 456, 55 },
		{ 448, 457, 55 },
		{ 448, 458, 55 },
		{ 448, 459, 55 },
		{ 448, 460, 55 },
		{ 448, 461, 55 },
		{ 448, 462, 55 },
		{ 448, 463, 55 },
		{ 448, 464, 55 },
		{ 448, 465, 55 },
		{ 448, 466, 55 },
		{ 448, 467, 55 },
		{ 448, 468, 55 },
		{ 448, 469, 55 },
		{ 448, 470, 55 },
		{ 448, 471, 55 },
		{ 448, 472, 55 },
		{ 448, 473, 55 },
		{ 448, 474, 55 },
		{ 448, 475, 55 },
		{ 448, 476, 55 },
		{ 448, 477, 55 },
		{ 448, 478, 55 },
		{ 448, 479, 55 },
		{ 448, 480, 55 },
		{ 448, 481, 55 },
		{ 448, 482, 55 },
		{ 448, 483, 55 },
		{ 448, 484, 55 },
		{ 448, 485, 55 },
		{ 448, 486, 55 },
		{ 448, 487, 55 },
		{ 448, 488, 55 },
		{ 448, 489, 55 },
		{ 448, 490, 55 },
		{ 448, 491, 55 },
		{ 448, 492, 55 },
		{ 448, 493, 55 },
		{ 448, 494, 55 },
		{ 448, 495, 55 },
		{ 448, 496, 55 },
		{ 448, 497, 55 },
		{ 448, 498, 55 },
		{ 448, 499, 55 },
		{ 448, 500, 55 },
		{ 448, 501, 55 },
		{ 448, 502, 55 },
		{ 448, 503, 55 },
		{ 448, 504, 55 },
		{ 448, 505, 55 },
		{ 448, 506, 55 },
		{ 448, 507, 55 },
		{ 448, 508, 55 },
		{ 448, 509, 55 },
		{ 448, 510, 55 },
		{ 448, 511, 55 },
		{ 449, 450, 55 },
		{ 449, 451, 55 },
		{ 449, 452, 55 },
		{ 449, 453, 55 },
		{ 449, 454, 55 },
		{ 449, 455, 55 },
		{ 449, 456, 55 },
		{ 449, 457, 55 },
		{ 449, 458, 55 },
		{ 449, 459, 55 },
		{ 449, 460, 55 },
		{ 449, 461, 55 },
		{ 449, 462, 55 },
		{ 449, 463, 55 },
		{ 449, 464, 55 },
		{ 449, 465, 55 },
		{ 449, 466, 55 },
		{ 449, 467, 55 },
		{ 449, 468, 55 },
		{ 449, 469, 55 },
		{ 449, 470, 55 },
		{ 449, 471, 55 },
		{ 449, 472, 55 },
		{ 449, 473, 55 },
		{ 449, 474, 55 },
		{ 449, 475, 55 },
		{ 449, 476, 55 },
		{ 449, 477, 55 },
		{ 449, 478, 55 },
		{ 449, 479, 55 },
		{ 449, 480, 55 },
		{ 449, 481, 55 },
		{ 449, 482, 55 },
		{ 449, 483, 55 },
		{ 449, 484, 55 },
		{ 449, 485, 55 },
		{ 449, 486, 55 },
		{ 449, 487, 55 },
		{ 449, 488, 55 },
		{ 449, 489, 55 },
		{ 449, 490, 55 },
		{ 449, 491, 55 },
		{ 449, 492, 55 },
		{ 449, 493, 55 },
		{ 449, 494, 55 },
		{ 449, 495, 55 },
		{ 449, 496, 55 },
		{ 449, 497, 55 },
		{ 449, 498, 55 },
		{ 449, 499, 55 },
		{ 449, 500, 55 },
		{ 449, 501, 55 },
		{ 449, 502, 55 },
		{ 449, 503, 55 },
		{ 449, 504, 55 },
		{ 449, 505, 55 },
		{ 449, 506, 55 },
		{ 449, 507, 55 },
		{ 449, 508, 55 },
		{ 449, 509, 55 },
		{ 449, 510, 55 },
		{ 449, 511, 55 },
		{ 449, 512, 55 },
		{ 450, 451, 55 },
		{ 450, 452, 55 },
		{ 450, 453, 55 },
		{ 450, 454, 55 },
		{ 450, 455, 55 },
		{ 450, 456, 55 },
		{ 450, 457, 55 },
		{ 450, 458, 55 },
		{ 450, 459, 55 },
		{ 450, 460, 55 },
		{ 450, 461, 55 },
		{ 450, 462, 55 },
		{ 450, 463, 55 },
		{ 450, 464, 55 },
		{ 450, 465, 55 },
		{ 450, 466, 55 },
		{ 450, 467, 55 },
		{ 450, 468, 55 },
		{ 450, 469, 55 },
		{ 450, 470, 55 },
		{ 450, 471, 55 },
		{ 450, 472, 55 },
		{ 450, 473, 55 },
		{ 450, 474, 55 },
		{ 450, 475, 55 },
		{ 450, 476, 55 },
		{ 450, 477, 55 },
		{ 450, 478, 55 },
		{ 450, 479, 55 },
		{ 450, 480, 55 },
		{ 450, 481, 55 },
		{ 450, 482, 55 },
		{ 450, 483, 55 },
		{ 450, 484, 55 },
		{ 450, 485, 55 },
		{ 450, 486, 55 },
		{ 450, 487, 55 },
		{ 450, 488, 55 },
		{ 450, 489, 55 },
		{ 450, 490, 55 },
		{ 450, 491, 55 },
		{ 450, 492, 55 },
		{ 450, 493, 55 },
		{ 450, 494, 55 },
		{ 450, 495, 55 },
		{ 450, 496, 55 },
		{ 450, 497, 55 },
		{ 450, 498, 55 },
		{ 450, 499, 55 },
		{ 450, 500, 55 },
		{ 450, 501, 55 },
		{ 450, 502, 55 },
		{ 450, 503, 55 },
		{ 450, 504, 55 },
		{ 450, 505, 55 },
		{ 450, 506, 55 },
		{ 450, 507, 55 },
		{ 450, 508, 55 },
		{ 450, 509, 55 },
		{ 450, 510, 55 },
		{ 450, 511, 55 },
		{ 450, 512, 55 },
		{ 451, 452, 55 },
		{ 451, 453, 55 },
		{ 451, 454, 55 },
		{ 451, 455, 55 },
		{ 451, 456, 55 },
		{ 451, 457, 55 },
		{ 451, 458, 55 },
		{ 451, 459, 55 },
		{ 451, 460, 55 },
		{ 451, 461, 55 },
		{ 451, 462, 55 },
		{ 451, 463, 55 },
		{ 451, 464, 55 },
		{ 451, 465, 55 },
		{ 451, 466, 55 },
		{ 451, 467, 55 },
		{ 451, 468, 55 },
		{ 451, 469, 55 },
		{ 451, 470, 55 },
		{ 451, 471, 55 },
		{ 451, 472, 55 },
		{ 451, 473, 55 },
		{ 451, 474, 55 },
		{ 451, 475, 55 },
		{ 451, 476, 55 },
		{ 451, 477, 55 },
		{ 451, 478, 55 },
		{ 451, 479, 55 },
		{ 451, 480, 55 },
		{ 451, 481, 55 },
		{ 451, 482, 55 },
		{ 451, 483, 55 },
		{ 451, 484, 55 },
		{ 451, 485, 55 },
		{ 451, 486, 55 },
		{ 451, 487, 55 },
		{ 451, 488, 55 },
		{ 451, 489, 55 },
		{ 451, 490, 55 },
		{ 451, 491, 55 },
		{ 451, 492, 55 },
		{ 451, 493, 55 },
		{ 451, 494, 55 },
		{ 451, 495, 55 },
		{ 451, 496, 55 },
		{ 451, 497, 55 },
		{ 451, 498, 55 },
		{ 451, 499, 55 },
		{ 451, 500, 55 },
		{ 451, 501, 55 },
		{ 451, 502, 55 },
		{ 451, 503, 55 },
		{ 451, 504, 55 },
		{ 451, 505, 55 },
		{ 451, 506, 55 },
		{ 451, 507, 55 },
		{ 451, 508, 55 },
		{ 451, 509, 55 },
		{ 451, 510, 55 },
		{ 451, 511, 55 },
		{ 451, 512, 55 },
		{ 452, 453, 55 },
		{ 452, 454, 55 },
		{ 452, 455, 55 },
		{ 452, 456, 55 },
		{ 452, 457, 55 },
		{ 452, 458, 55 },
		{ 452, 459, 55 },
		{ 452, 460, 55 },
		{ 452, 461, 55 },
		{ 452, 462, 55 },
		{ 452, 463, 55 },
		{ 452, 464, 55 },
		{ 452, 465, 55 },
		{ 452, 466, 55 },
		{ 452, 467, 55 },
		{ 452, 468, 55 },
		{ 452, 469, 55 },
		{ 452, 470, 55 },
		{ 452, 471, 55 },
		{ 452, 472, 55 },
		{ 452, 473, 55 },
		{ 452, 474, 55 },
		{ 452, 475, 55 },
		{ 452, 476, 55 },
		{ 452, 477, 55 },
		{ 452, 478, 55 },
		{ 452, 479, 55 },
		{ 452, 480, 55 },
		{ 452, 481, 55 },
		{ 452, 482, 55 },
		{ 452, 483, 55 },
		{ 452, 484, 55 },
		{ 452, 485, 55 },
		{ 452, 486, 55 },
		{ 452, 487, 55 },
		{ 452, 488, 55 },
		{ 452, 489, 55 },
		{ 452, 490, 55 },
		{ 452, 491, 55 },
		{ 452, 492, 55 },
		{ 452, 493, 55 },
		{ 452, 494, 55 },
		{ 452, 495, 55 },
		{ 452, 496, 55 },
		{ 452, 497, 55 },
		{ 452, 498, 55 },
		{ 452, 499, 55 },
		{ 452, 500, 55 },
		{ 452, 501, 55 },
		{ 452, 502, 55 },
		{ 452, 503, 55 },
		{ 452, 504, 55 },
		{ 452, 505, 55 },
		{ 452, 506, 55 },
		{ 452, 507, 55 },
		{ 452, 508, 55 },
		{ 452, 509, 55 },
		{ 452, 510, 55 },
		{ 452, 511, 55 },
		{ 452, 512, 55 },
		{ 453, 454, 55 },
		{ 453, 455, 55 },
		{ 453, 456, 55 },
		{ 453, 457, 55 },
		{ 453, 458, 55 },
		{ 453, 459, 55 },
		{ 453, 460, 55 },
		{ 453, 461, 55 },
		{ 453, 462, 55 },
		{ 453, 463, 55 },
		{ 453, 464, 55 },
		{ 453, 465, 55 },
		{ 453, 466, 55 },
		{ 453, 467, 55 },
		{ 453, 468, 55 },
		{ 453, 469, 55 },
		{ 453, 470, 55 },
		{ 453, 471, 55 },
		{ 453, 472, 55 },
		{ 453, 473, 55 },
		{ 453, 474, 55 },
		{ 453, 475, 55 },
		{ 453, 476, 55 },
		{ 453, 477, 55 },
		{ 453, 478, 55 },
		{ 453, 479, 55 },
		{ 453, 480, 55 },
		{ 453, 481, 55 },
		{ 453, 482, 55 },
		{ 453, 483, 55 },
		{ 453, 484, 55 },
		{ 453, 485, 55 },
		{ 453, 486, 55 },
		{ 453, 487, 55 },
		{ 453, 488, 55 },
		{ 453, 489, 55 },
		{ 453, 490, 55 },
		{ 453, 491, 55 },
		{ 453, 492, 55 },
		{ 453, 493, 55 },
		{ 453, 494, 55 },
		{ 453, 495, 55 },
		{ 453, 496, 55 },
		{ 453, 497, 55 },
		{ 453, 498, 55 },
		{ 453, 499, 55 },
		{ 453, 500, 55 },
		{ 453, 501, 55 },
		{ 453, 502, 55 },
		{ 453, 503, 55 },
		{ 453, 504, 55 },
		{ 453, 505, 55 },
		{ 453, 506, 55 },
		{ 453, 507, 55 },
		{ 453, 508, 55 },
		{ 453, 509, 55 },
		{ 453, 510, 55 },
		{ 453, 511, 55 },
		{ 453, 512, 55 },
		{ 454, 455, 55 },
		{ 454, 456, 55 },
		{ 454, 457, 55 },
		{ 454, 458, 55 },
		{ 454, 459, 55 },
		{ 454, 460, 55 },
		{ 454, 461, 55 },
		{ 454, 462, 55 },
		{ 454, 463, 55 },
		{ 454, 464, 55 },
		{ 454, 465, 55 },
		{ 454, 466, 55 },
		{ 454, 467, 55 },
		{ 454, 468, 55 },
		{ 454, 469, 55 },
		{ 454, 470, 55 },
		{ 454, 471, 55 },
		{ 454, 472, 55 },
		{ 454, 473, 55 },
		{ 454, 474, 55 },
		{ 454, 475, 55 },
		{ 454, 476, 55 },
		{ 454, 477, 55 },
		{ 454, 478, 55 },
		{ 454, 479, 55 },
		{ 454, 480, 55 },
		{ 454, 481, 55 },
		{ 454, 482, 55 },
		{ 454, 483, 55 },
		{ 454, 484, 55 },
		{ 454, 485, 55 },
		{ 454, 486, 55 },
		{ 454, 487, 55 },
		{ 454, 488, 55 },
		{ 454, 489, 55 },
		{ 454, 490, 55 },
		{ 454, 491, 55 },
		{ 454, 492, 55 },
		{ 454, 493, 55 },
		{ 454, 494, 55 },
		{ 454, 495, 55 },
		{ 454, 496, 55 },
		{ 454, 497, 55 },
		{ 454, 498, 55 },
		{ 454, 499, 55 },
		{ 454, 500, 55 },
		{ 454, 501, 55 },
		{ 454, 502, 55 },
		{ 454, 503, 55 },
		{ 454, 504, 55 },
		{ 454, 505, 55 },
		{ 454, 506, 55 },
		{ 454, 507, 55 },
		{ 454, 508, 55 },
		{ 454, 509, 55 },
		{ 454, 510, 55 },
		{ 454, 511, 55 },
		{ 454, 512, 55 },
		{ 455, 456, 55 },
		{ 455, 457, 55 },
		{ 455, 458, 55 },
		{ 455, 459, 55 },
		{ 455, 460, 55 },
		{ 455, 461, 55 },
		{ 455, 462, 55 },
		{ 455, 463, 55 },
		{ 455, 464, 55 },
		{ 455, 465, 55 },
		{ 455, 466, 55 },
		{ 455, 467, 55 },
		{ 455, 468, 55 },
		{ 455, 469, 55 },
		{ 455, 470, 55 },
		{ 455, 471, 55 },
		{ 455, 472, 55 },
		{ 455, 473, 55 },
		{ 455, 474, 55 },
		{ 455, 475, 55 },
		{ 455, 476, 55 },
		{ 455, 477, 55 },
		{ 455, 478, 55 },
		{ 455, 479, 55 },
		{ 455, 480, 55 },
		{ 455, 481, 55 },
		{ 455, 482, 55 },
		{ 455, 483, 55 },
		{ 455, 484, 55 },
		{ 455, 485, 55 },
		{ 455, 486, 55 },
		{ 455, 487, 55 },
		{ 455, 488, 55 },
		{ 455, 489, 55 },
		{ 455, 490, 55 },
		{ 455, 491, 55 },
		{ 455, 492, 55 },
		{ 455, 493, 55 },
		{ 455, 494, 55 },
		{ 455, 495, 55 },
		{ 455, 496, 55 },
		{ 455, 497, 55 },
		{ 455, 498, 55 },
		{ 455, 499, 55 },
		{ 455, 500, 55 },
		{ 455, 501, 55 },
		{ 455, 502, 55 },
		{ 455, 503, 55 },
		{ 455, 504, 55 },
		{ 455, 505, 55 },
		{ 455, 506, 55 },
		{ 455, 507, 55 },
		{ 455, 508, 55 },
		{ 455, 509, 55 },
		{ 455, 510, 55 },
		{ 455, 511, 55 },
		{ 455, 512, 55 },
		{ 456, 457, 55 },
		{ 456, 458, 55 },
		{ 456, 459, 55 },
		{ 456, 460, 55 },
		{ 456, 461, 55 },
		{ 456, 462, 55 },
		{ 456, 463, 55 },
		{ 456, 464, 55 },
		{ 456, 465, 55 },
		{ 456, 466, 55 },
		{ 456, 467, 55 },
		{ 456, 468, 55 },
		{ 456, 469, 55 },
		{ 456, 470, 55 },
		{ 456, 471, 55 },
		{ 456, 472, 55 },
		{ 456, 473, 55 },
		{ 456, 474, 55 },
		{ 456, 475, 55 },
		{ 456, 476, 55 },
		{ 456, 477, 55 },
		{ 456, 478, 55 },
		{ 456, 479, 55 },
		{ 456, 480, 55 },
		{ 456, 481, 55 },
		{ 456, 482, 55 },
		{ 456, 483, 55 },
		{ 456, 484, 55 },
		{ 456, 485, 55 },
		{ 456, 486, 55 },
		{ 456, 487, 55 },
		{ 456, 488, 55 },
		{ 456, 489, 55 },
		{ 456, 490, 55 },
		{ 456, 491, 55 },
		{ 456, 492, 55 },
		{ 456, 493, 55 },
		{ 456, 494, 55 },
		{ 456, 495, 55 },
		{ 456, 496, 55 },
		{ 456, 497, 55 },
		{ 456, 498, 55 },
		{ 456, 499, 55 },
		{ 456, 500, 55 },
		{ 456, 501, 55 },
		{ 456, 502, 55 },
		{ 456, 503, 55 },
		{ 456, 504, 55 },
		{ 456, 505, 55 },
		{ 456, 506, 55 },
		{ 456, 507, 55 },
		{ 456, 508, 55 },
		{ 456, 509, 55 },
		{ 456, 510, 55 },
		{ 456, 511, 55 },
		{ 456, 512, 55 },
		{ 457, 458, 55 },
		{ 457, 459, 55 },
		{ 457, 460, 55 },
		{ 457, 461, 55 },
		{ 457, 462, 55 },
		{ 457, 463, 55 },
		{ 457, 464, 55 },
		{ 457, 465, 55 },
		{ 457, 466, 55 },
		{ 457, 467, 55 },
		{ 457, 468, 55 },
		{ 457, 469, 55 },
		{ 457, 470, 55 },
		{ 457, 471, 55 },
		{ 457, 472, 55 },
		{ 457, 473, 55 },
		{ 457, 474, 55 },
		{ 457, 475, 55 },
		{ 457, 476, 55 },
		{ 457, 477, 55 },
		{ 457, 478, 55 },
		{ 457, 479, 55 },
		{ 457, 480, 55 },
		{ 457, 481, 55 },
		{ 457, 482, 55 },
		{ 457, 483, 55 },
		{ 457, 484, 55 },
		{ 457, 485, 55 },
		{ 457, 486, 55 },
		{ 457, 487, 55 },
		{ 457, 488, 55 },
		{ 457, 489, 55 },
		{ 457, 490, 55 },
		{ 457, 491, 55 },
		{ 457, 492, 55 },
		{ 457, 493, 55 },
		{ 457, 494, 55 },
		{ 457, 495, 55 },
		{ 457, 496, 55 },
		{ 457, 497, 55 },
		{ 457, 498, 55 },
		{ 457, 499, 55 },
		{ 457, 500, 55 },
		{ 457, 501, 55 },
		{ 457, 502, 55 },
		{ 457, 503, 55 },
		{ 457, 504, 55 },
		{ 457, 505, 55 },
		{ 457, 506, 55 },
		{ 457, 507, 55 },
		{ 457, 508, 55 },
		{ 457, 509, 55 },
		{ 457, 510, 55 },
		{ 457, 511, 55 },
		{ 457, 512, 55 },
		{ 458, 459, 55 },
		{ 458, 460, 55 },
		{ 458, 461, 55 },
		{ 458, 462, 55 },
		{ 458, 463, 55 },
		{ 458, 464, 55 },
		{ 458, 465, 55 },
		{ 458, 466, 55 },
		{ 458, 467, 55 },
		{ 458, 468, 55 },
		{ 458, 469, 55 },
		{ 458, 470, 55 },
		{ 458, 471, 55 },
		{ 458, 472, 55 },
		{ 458, 473, 55 },
		{ 458, 474, 55 },
		{ 458, 475, 55 },
		{ 458, 476, 55 },
		{ 458, 477, 55 },
		{ 458, 478, 55 },
		{ 458, 479, 55 },
		{ 458, 480, 55 },
		{ 458, 481, 55 },
		{ 458, 482, 55 },
		{ 458, 483, 55 },
		{ 458, 484, 55 },
		{ 458, 485, 55 },
		{ 458, 486, 55 },
		{ 458, 487, 55 },
		{ 458, 488, 55 },
		{ 458, 489, 55 },
		{ 458, 490, 55 },
		{ 458, 491, 55 },
		{ 458, 492, 55 },
		{ 458, 493, 55 },
		{ 458, 494, 55 },
		{ 458, 495, 55 },
		{ 458, 496, 55 },
		{ 458, 497, 55 },
		{ 458, 498, 55 },
		{ 458, 499, 55 },
		{ 458, 500, 55 },
		{ 458, 501, 55 },
		{ 458, 502, 55 },
		{ 458, 503, 55 },
		{ 458, 504, 55 },
		{ 458, 505, 55 },
		{ 458, 506, 55 },
		{ 458, 507, 55 },
		{ 458, 508, 55 },
		{ 458, 509, 55 },
		{ 458, 510, 55 },
		{ 458, 511, 55 },
		{ 458, 512, 55 },
		{ 459, 460, 55 },
		{ 459, 461, 55 },
		{ 459, 462, 55 },
		{ 459, 463, 55 },
		{ 459, 464, 55 },
		{ 459, 465, 55 },
		{ 459, 466, 55 },
		{ 459, 467, 55 },
		{ 459, 468, 55 },
		{ 459, 469, 55 },
		{ 459, 470, 55 },
		{ 459, 471, 55 },
		{ 459, 472, 55 },
		{ 459, 473, 55 },
		{ 459, 474, 55 },
		{ 459, 475, 55 },
		{ 459, 476, 55 },
		{ 459, 477, 55 },
		{ 459, 478, 55 },
		{ 459, 479, 55 },
		{ 459, 480, 55 },
		{ 459, 481, 55 },
		{ 459, 482, 55 },
		{ 459, 483, 55 },
		{ 459, 484, 55 },
		{ 459, 485, 55 },
		{ 459, 486, 55 },
		{ 459, 487, 55 },
		{ 459, 488, 55 },
		{ 459, 489, 55 },
		{ 459, 490, 55 },
		{ 459, 491, 55 },
		{ 459, 492, 55 },
		{ 459, 493, 55 },
		{ 459, 494, 55 },
		{ 459, 495, 55 },
		{ 459, 496, 55 },
		{ 459, 497, 55 },
		{ 459, 498, 55 },
		{ 459, 499, 55 },
		{ 459, 500, 55 },
		{ 459, 501, 55 },
		{ 459, 502, 55 },
		{ 459, 503, 55 },
		{ 459, 504, 55 },
		{ 459, 505, 55 },
		{ 459, 506, 55 },
		{ 459, 507, 55 },
		{ 459, 508, 55 },
		{ 459, 509, 55 },
		{ 459, 510, 55 },
		{ 459, 511, 55 },
		{ 459, 512, 55 },
		{ 460, 461, 55 },
		{ 460, 462, 55 },
		{ 460, 463, 55 },
		{ 460, 464, 55 },
		{ 460, 465, 55 },
		{ 460, 466, 55 },
		{ 460, 467, 55 },
		{ 460, 468, 55 },
		{ 460, 469, 55 },
		{ 460, 470, 55 },
		{ 460, 471, 55 },
		{ 460, 472, 55 },
		{ 460, 473, 55 },
		{ 460, 474, 55 },
		{ 460, 475, 55 },
		{ 460, 476, 55 },
		{ 460, 477, 55 },
		{ 460, 478, 55 },
		{ 460, 479, 55 },
		{ 460, 480, 55 },
		{ 460, 481, 55 },
		{ 460, 482, 55 },
		{ 460, 483, 55 },
		{ 460, 484, 55 },
		{ 460, 485, 55 },
		{ 460, 486, 55 },
		{ 460, 487, 55 },
		{ 460, 488, 55 },
		{ 460, 489, 55 },
		{ 460, 490, 55 },
		{ 460, 491, 55 },
		{ 460, 492, 55 },
		{ 460, 493, 55 },
		{ 460, 494, 55 },
		{ 460, 495, 55 },
		{ 460, 496, 55 },
		{ 460, 497, 55 },
		{ 460, 498, 55 },
		{ 460, 499, 55 },
		{ 460, 500, 55 },
		{ 460, 501, 55 },
		{ 460, 502, 55 },
		{ 460, 503, 55 },
		{ 460, 504, 55 },
		{ 460, 505, 55 },
		{ 460, 506, 55 },
		{ 460, 507, 55 },
		{ 460, 508, 55 },
		{ 460, 509, 55 },
		{ 460, 510, 55 },
		{ 460, 511, 55 },
		{ 460, 512, 55 },
		{ 461, 462, 55 },
		{ 461, 463, 55 },
		{ 461, 464, 55 },
		{ 461, 465, 55 },
		{ 461, 466, 55 },
		{ 461, 467, 55 },
		{ 461, 468, 55 },
		{ 461, 469, 55 },
		{ 461, 470, 55 },
		{ 461, 471, 55 },
		{ 461, 472, 55 },
		{ 461, 473, 55 },
		{ 461, 474, 55 },
		{ 461, 475, 55 },
		{ 461, 476, 55 },
		{ 461, 477, 55 },
		{ 461, 478, 55 },
		{ 461, 479, 55 },
		{ 461, 480, 55 },
		{ 461, 481, 55 },
		{ 461, 482, 55 },
		{ 461, 483, 55 },
		{ 461, 484, 55 },
		{ 461, 485, 55 },
		{ 461, 486, 55 },
		{ 461, 487, 55 },
		{ 461, 488, 55 },
		{ 461, 489, 55 },
		{ 461, 490, 55 },
		{ 461, 491, 55 },
		{ 461, 492, 55 },
		{ 461, 493, 55 },
		{ 461, 494, 55 },
		{ 461, 495, 55 },
		{ 461, 496, 55 },
		{ 461, 497, 55 },
		{ 461, 498, 55 },
		{ 461, 499, 55 },
		{ 461, 500, 55 },
		{ 461, 501, 55 },
		{ 461, 502, 55 },
		{ 461, 503, 55 },
		{ 461, 504, 55 },
		{ 461, 505, 55 },
		{ 461, 506, 55 },
		{ 461, 507, 55 },
		{ 461, 508, 55 },
		{ 461, 509, 55 },
		{ 461, 510, 55 },
		{ 461, 511, 55 },
		{ 461, 512, 55 },
		{ 462, 463, 55 },
		{ 462, 464, 55 },
		{ 462, 465, 55 },
		{ 462, 466, 55 },
		{ 462, 467, 55 },
		{ 462, 468, 55 },
		{ 462, 469, 55 },
		{ 462, 470, 55 },
		{ 462, 471, 55 },
		{ 462, 472, 55 },
		{ 462, 473, 55 },
		{ 462, 474, 55 },
		{ 462, 475, 55 },
		{ 462, 476, 55 },
		{ 462, 477, 55 },
		{ 462, 478, 55 },
		{ 462, 479, 55 },
		{ 462, 480, 55 },
		{ 462, 481, 55 },
		{ 462, 482, 55 },
		{ 462, 483, 55 },
		{ 462, 484, 55 },
		{ 462, 485, 55 },
		{ 462, 486, 55 },
		{ 462, 487, 55 },
		{ 462, 488, 55 },
		{ 462, 489, 55 },
		{ 462, 490, 55 },
		{ 462, 491, 55 },
		{ 462, 492, 55 },
		{ 462, 493, 55 },
		{ 462, 494, 55 },
		{ 462, 495, 55 },
		{ 462, 496, 55 },
		{ 462, 497, 55 },
		{ 462, 498, 55 },
		{ 462, 499, 55 },
		{ 462, 500, 55 },
		{ 462, 501, 55 },
		{ 462, 502, 55 },
		{ 462, 503, 55 },
		{ 462, 504, 55 },
		{ 462, 505, 55 },
		{ 462, 506, 55 },
		{ 462, 507, 55 },
		{ 462, 508, 55 },
		{ 462, 509, 55 },
		{ 462, 510, 55 },
		{ 462, 511, 55 },
		{ 462, 512, 55 },
		{ 463, 464, 55 },
		{ 463, 465, 55 },
		{ 463, 466, 55 },
		{ 463, 467, 55 },
		{ 463, 468, 55 },
		{ 463, 469, 55 },
		{ 463, 470, 55 },
		{ 463, 471, 55 },
		{ 463, 472, 55 },
		{ 463, 473, 55 },
		{ 463, 474, 55 },
		{ 463, 475, 55 },
		{ 463, 476, 55 },
		{ 463, 477, 55 },
		{ 463, 478, 55 },
		{ 463, 479, 55 },
		{ 463, 480, 55 },
		{ 463, 481, 55 },
		{ 463, 482, 55 },
		{ 463, 483, 55 },
		{ 463, 484, 55 },
		{ 463, 485, 55 },
		{ 463, 486, 55 },
		{ 463, 487, 55 },
		{ 463, 488, 55 },
		{ 463, 489, 55 },
		{ 463, 490, 55 },
		{ 463, 491, 55 },
		{ 463, 492, 55 },
		{ 463, 493, 55 },
		{ 463, 494, 55 },
		{ 463, 495, 55 },
		{ 463, 496, 55 },
		{ 463, 497, 55 },
		{ 463, 498, 55 },
		{ 463, 499, 55 },
		{ 463, 500, 55 },
		{ 463, 501, 55 },
		{ 463, 502, 55 },
		{ 463, 503, 55 },
		{ 463, 504, 55 },
		{ 463, 505, 55 },
		{ 463, 506, 55 },
		{ 463, 507, 55 },
		{ 463, 508, 55 },
		{ 463, 509, 55 },
		{ 463, 510, 55 },
		{ 463, 511, 55 },
		{ 463, 512, 55 },
		{ 464, 465, 55 },
		{ 464, 466, 55 },
		{ 464, 467, 55 },
		{ 464, 468, 55 },
		{ 464, 469, 55 },
		{ 464, 470, 55 },
		{ 464, 471, 55 },
		{ 464, 472, 55 },
		{ 464, 473, 55 },
		{ 464, 474, 55 },
		{ 464, 475, 55 },
		{ 464, 476, 55 },
		{ 464, 477, 55 },
		{ 464, 478, 55 },
		{ 464, 479, 55 },
		{ 464, 480, 55 },
		{ 464, 481, 55 },
		{ 464, 482, 55 },
		{ 464, 483, 55 },
		{ 464, 484, 55 },
		{ 464, 485, 55 },
		{ 464, 486, 55 },
		{ 464, 487, 55 },
		{ 464, 488, 55 },
		{ 464, 489, 55 },
		{ 464, 490, 55 },
		{ 464, 491, 55 },
		{ 464, 492, 55 },
		{ 464, 493, 55 },
		{ 464, 494, 55 },
		{ 464, 495, 55 },
		{ 464, 496, 55 },
		{ 464, 497, 55 },
		{ 464, 498, 55 },
		{ 464, 499, 55 },
		{ 464, 500, 55 },
		{ 464, 501, 55 },
		{ 464, 502, 55 },
		{ 464, 503, 55 },
		{ 464, 504, 55 },
		{ 464, 505, 55 },
		{ 464, 506, 55 },
		{ 464, 507, 55 },
		{ 464, 508, 55 },
		{ 464, 509, 55 },
		{ 464, 510, 55 },
		{ 464, 511, 55 },
		{ 464, 512, 55 },
		{ 465, 466, 55 },
		{ 465, 467, 55 },
		{ 465, 468, 55 },
		{ 465, 469, 55 },
		{ 465, 470, 55 },
		{ 465, 471, 55 },
		{ 465, 472, 55 },
		{ 465, 473, 55 },
		{ 465, 474, 55 },
		{ 465, 475, 55 },
		{ 465, 476, 55 },
		{ 465, 477, 55 },
		{ 465, 478, 55 },
		{ 465, 479, 55 },
		{ 465, 480, 55 },
		{ 465, 481, 55 },
		{ 465, 482, 55 },
		{ 465, 483, 55 },
		{ 465, 484, 55 },
		{ 465, 485, 55 },
		{ 465, 486, 55 },
		{ 465, 487, 55 },
		{ 465, 488, 55 },
		{ 465, 489, 55 },
		{ 465, 490, 55 },
		{ 465, 491, 55 },
		{ 465, 492, 55 },
		{ 465, 493, 55 },
		{ 465, 494, 55 },
		{ 465, 495, 55 },
		{ 465, 496, 55 },
		{ 465, 497, 55 },
		{ 465, 498, 55 },
		{ 465, 499, 55 },
		{ 465, 500, 55 },
		{ 465, 501, 55 },
		{ 465, 502, 55 },
		{ 465, 503, 55 },
		{ 465, 504, 55 },
		{ 465, 505, 55 },
		{ 465, 506, 55 },
		{ 465, 507, 55 },
		{ 465, 508, 55 },
		{ 465, 509, 55 },
		{ 465, 510, 55 },
		{ 465, 511, 55 },
		{ 465, 512, 55 },
		{ 466, 467, 55 },
		{ 466, 468, 55 },
		{ 466, 469, 55 },
		{ 466, 470, 55 },
		{ 466, 471, 55 },
		{ 466, 472, 55 },
		{ 466, 473, 55 },
		{ 466, 474, 55 },
		{ 466, 475, 55 },
		{ 466, 476, 55 },
		{ 466, 477, 55 },
		{ 466, 478, 55 },
		{ 466, 479, 55 },
		{ 466, 480, 55 },
		{ 466, 481, 55 },
		{ 466, 482, 55 },
		{ 466, 483, 55 },
		{ 466, 484, 55 },
		{ 466, 485, 55 },
		{ 466, 486, 55 },
		{ 466, 487, 55 },
		{ 466, 488, 55 },
		{ 466, 489, 55 },
		{ 466, 490, 55 },
		{ 466, 491, 55 },
		{ 466, 492, 55 },
		{ 466, 493, 55 },
		{ 466, 494, 55 },
		{ 466, 495, 55 },
		{ 466, 496, 55 },
		{ 466, 497, 55 },
		{ 466, 498, 55 },
		{ 466, 499, 55 },
		{ 466, 500, 55 },
		{ 466, 501, 55 },
		{ 466, 502, 55 },
		{ 466, 503, 55 },
		{ 466, 504, 55 },
		{ 466, 505, 55 },
		{ 466, 506, 55 },
		{ 466, 507, 55 },
		{ 466, 508, 55 },
		{ 466, 509, 55 },
		{ 466, 510, 55 },
		{ 466, 511, 55 },
		{ 466, 512, 55 },
		{ 467, 468, 55 },
		{ 467, 469, 55 },
		{ 467, 470, 55 },
		{ 467, 471, 55 },
		{ 467, 472, 55 },
		{ 467, 473, 55 },
		{ 467, 474, 55 },
		{ 467, 475, 55 },
		{ 467, 476, 55 },
		{ 467, 477, 55 },
		{ 467, 478, 55 },
		{ 467, 479, 55 },
		{ 467, 480, 55 },
		{ 467, 481, 55 },
		{ 467, 482, 55 },
		{ 467, 483, 55 },
		{ 467, 484, 55 },
		{ 467, 485, 55 },
		{ 467, 486, 55 },
		{ 467, 487, 55 },
		{ 467, 488, 55 },
		{ 467, 489, 55 },
		{ 467, 490, 55 },
		{ 467, 491, 55 },
		{ 467, 492, 55 },
		{ 467, 493, 55 },
		{ 467, 494, 55 },
		{ 467, 495, 55 },
		{ 467, 496, 55 },
		{ 467, 497, 55 },
		{ 467, 498, 55 },
		{ 467, 499, 55 },
		{ 467, 500, 55 },
		{ 467, 501, 55 },
		{ 467, 502, 55 },
		{ 467, 503, 55 },
		{ 467, 504, 55 },
		{ 467, 505, 55 },
		{ 467, 506, 55 },
		{ 467, 507, 55 },
		{ 467, 508, 55 },
		{ 467, 509, 55 },
		{ 467, 510, 55 },
		{ 467, 511, 55 },
		{ 467, 512, 55 },
		{ 468, 469, 55 },
		{ 468, 470, 55 },
		{ 468, 471, 55 },
		{ 468, 472, 55 },
		{ 468, 473, 55 },
		{ 468, 474, 55 },
		{ 468, 475, 55 },
		{ 468, 476, 55 },
		{ 468, 477, 55 },
		{ 468, 478, 55 },
		{ 468, 479, 55 },
		{ 468, 480, 55 },
		{ 468, 481, 55 },
		{ 468, 482, 55 },
		{ 468, 483, 55 },
		{ 468, 484, 55 },
		{ 468, 485, 55 },
		{ 468, 486, 55 },
		{ 468, 487, 55 },
		{ 468, 488, 55 },
		{ 468, 489, 55 },
		{ 468, 490, 55 },
		{ 468, 491, 55 },
		{ 468, 492, 55 },
		{ 468, 493, 55 },
		{ 468, 494, 55 },
		{ 468, 495, 55 },
		{ 468, 496, 55 },
		{ 468, 497, 55 },
		{ 468, 498, 55 },
		{ 468, 499, 55 },
		{ 468, 500, 55 },
		{ 468, 501, 55 },
		{ 468, 502, 55 },
		{ 468, 503, 55 },
		{ 468, 504, 55 },
		{ 468, 505, 55 },
		{ 468, 506, 55 },
		{ 468, 507, 55 },
		{ 468, 508, 55 },
		{ 468, 509, 55 },
		{ 468, 510, 55 },
		{ 468, 511, 55 },
		{ 468, 512, 55 },
		{ 469, 470, 55 },
		{ 469, 471, 55 },
		{ 469, 472, 55 },
		{ 469, 473, 55 },
		{ 469, 474, 55 },
		{ 469, 475, 55 },
		{ 469, 476, 55 },
		{ 469, 477, 55 },
		{ 469, 478, 55 },
		{ 469, 479, 55 },
		{ 469, 480, 55 },
		{ 469, 481, 55 },
		{ 469, 482, 55 },
		{ 469, 483, 55 },
		{ 469, 484, 55 },
		{ 469, 485, 55 },
		{ 469, 486, 55 },
		{ 469, 487, 55 },
		{ 469, 488, 55 },
		{ 469, 489, 55 },
		{ 469, 490, 55 },
		{ 469, 491, 55 },
		{ 469, 492, 55 },
		{ 469, 493, 55 },
		{ 469, 494, 55 },
		{ 469, 495, 55 },
		{ 469, 496, 55 },
		{ 469, 497, 55 },
		{ 469, 498, 55 },
		{ 469, 499, 55 },
		{ 469, 500, 55 },
		{ 469, 501, 55 },
		{ 469, 502, 55 },
		{ 469, 503, 55 },
		{ 469, 504, 55 },
		{ 469, 505, 55 },
		{ 469, 506, 55 },
		{ 469, 507, 55 },
		{ 469, 508, 55 },
		{ 469, 509, 55 },
		{ 469, 510, 55 },
		{ 469, 511, 55 },
		{ 469, 512, 55 },
		{ 470, 471, 55 },
		{ 470, 472, 55 },
		{ 470, 473, 55 },
		{ 470, 474, 55 },
		{ 470, 475, 55 },
		{ 470, 476, 55 },
		{ 470, 477, 55 },
		{ 470, 478, 55 },
		{ 470, 479, 55 },
		{ 470, 480, 55 },
		{ 470, 481, 55 },
		{ 470, 482, 55 },
		{ 470, 483, 55 },
		{ 470, 484, 55 },
		{ 470, 485, 55 },
		{ 470, 486, 55 },
		{ 470, 487, 55 },
		{ 470, 488, 55 },
		{ 470, 489, 55 },
		{ 470, 490, 55 },
		{ 470, 491, 55 },
		{ 470, 492, 55 },
		{ 470, 493, 55 },
		{ 470, 494, 55 },
		{ 470, 495, 55 },
		{ 470, 496, 55 },
		{ 470, 497, 55 },
		{ 470, 498, 55 },
		{ 470, 499, 55 },
		{ 470, 500, 55 },
		{ 470, 501, 55 },
		{ 470, 502, 55 },
		{ 470, 503, 55 },
		{ 470, 504, 55 },
		{ 470, 505, 55 },
		{ 470, 506, 55 },
		{ 470, 507, 55 },
		{ 470, 508, 55 },
		{ 470, 509, 55 },
		{ 470, 510, 55 },
		{ 470, 511, 55 },
		{ 470, 512, 55 },
		{ 471, 472, 55 },
		{ 471, 473, 55 },
		{ 471, 474, 55 },
		{ 471, 475, 55 },
		{ 471, 476, 55 },
		{ 471, 477, 55 },
		{ 471, 478, 55 },
		{ 471, 479, 55 },
		{ 471, 480, 55 },
		{ 471, 481, 55 },
		{ 471, 482, 55 },
		{ 471, 483, 55 },
		{ 471, 484, 55 },
		{ 471, 485, 55 },
		{ 471, 486, 55 },
		{ 471, 487, 55 },
		{ 471, 488, 55 },
		{ 471, 489, 55 },
		{ 471, 490, 55 },
		{ 471, 491, 55 },
		{ 471, 492, 55 },
		{ 471, 493, 55 },
		{ 471, 494, 55 },
		{ 471, 495, 55 },
		{ 471, 496, 55 },
		{ 471, 497, 55 },
		{ 471, 498, 55 },
		{ 471, 499, 55 },
		{ 471, 500, 55 },
		{ 471, 501, 55 },
		{ 471, 502, 55 },
		{ 471, 503, 55 },
		{ 471, 504, 55 },
		{ 471, 505, 55 },
		{ 471, 506, 55 },
		{ 471, 507, 55 },
		{ 471, 508, 55 },
		{ 471, 509, 55 },
		{ 471, 510, 55 },
		{ 471, 511, 55 },
		{ 471, 512, 55 },
		{ 472, 473, 55 },
		{ 472, 474, 55 },
		{ 472, 475, 55 },
		{ 472, 476, 55 },
		{ 472, 477, 55 },
		{ 472, 478, 55 },
		{ 472, 479, 55 },
		{ 472, 480, 55 },
		{ 472, 481, 55 },
		{ 472, 482, 55 },
		{ 472, 483, 55 },
		{ 472, 484, 55 },
		{ 472, 485, 55 },
		{ 472, 486, 55 },
		{ 472, 487, 55 },
		{ 472, 488, 55 },
		{ 472, 489, 55 },
		{ 472, 490, 55 },
		{ 472, 491, 55 },
		{ 472, 492, 55 },
		{ 472, 493, 55 },
		{ 472, 494, 55 },
		{ 472, 495, 55 },
		{ 472, 496, 55 },
		{ 472, 497, 55 },
		{ 472, 498, 55 },
		{ 472, 499, 55 },
		{ 472, 500, 55 },
		{ 472, 501, 55 },
		{ 472, 502, 55 },
		{ 472, 503, 55 },
		{ 472, 504, 55 },
		{ 472, 505, 55 },
		{ 472, 506, 55 },
		{ 472, 507, 55 },
		{ 472, 508, 55 },
		{ 472, 509, 55 },
		{ 472, 510, 55 },
		{ 472, 511, 55 },
		{ 472, 512, 55 },
		{ 473, 474, 55 },
		{ 473, 475, 55 },
		{ 473, 476, 55 },
		{ 473, 477, 55 },
		{ 473, 478, 55 },
		{ 473, 479, 55 },
		{ 473, 480, 55 },
		{ 473, 481, 55 },
		{ 473, 482, 55 },
		{ 473, 483, 55 },
		{ 473, 484, 55 },
		{ 473, 485, 55 },
		{ 473, 486, 55 },
		{ 473, 487, 55 },
		{ 473, 488, 55 },
		{ 473, 489, 55 },
		{ 473, 490, 55 },
		{ 473, 491, 55 },
		{ 473, 492, 55 },
		{ 473, 493, 55 },
		{ 473, 494, 55 },
		{ 473, 495, 55 },
		{ 473, 496, 55 },
		{ 473, 497, 55 },
		{ 473, 498, 55 },
		{ 473, 499, 55 },
		{ 473, 500, 55 },
		{ 473, 501, 55 },
		{ 473, 502, 55 },
		{ 473, 503, 55 },
		{ 473, 504, 55 },
		{ 473, 505, 55 },
		{ 473, 506, 55 },
		{ 473, 507, 55 },
		{ 473, 508, 55 },
		{ 473, 509, 55 },
		{ 473, 510, 55 },
		{ 473, 511, 55 },
		{ 473, 512, 55 },
		{ 474, 475, 55 },
		{ 474, 476, 55 },
		{ 474, 477, 55 },
		{ 474, 478, 55 },
		{ 474, 479, 55 },
		{ 474, 480, 55 },
		{ 474, 481, 55 },
		{ 474, 482, 55 },
		{ 474, 483, 55 },
		{ 474, 484, 55 },
		{ 474, 485, 55 },
		{ 474, 486, 55 },
		{ 474, 487, 55 },
		{ 474, 488, 55 },
		{ 474, 489, 55 },
		{ 474, 490, 55 },
		{ 474, 491, 55 },
		{ 474, 492, 55 },
		{ 474, 493, 55 },
		{ 474, 494, 55 },
		{ 474, 495, 55 },
		{ 474, 496, 55 },
		{ 474, 497, 55 },
		{ 474, 498, 55 },
		{ 474, 499, 55 },
		{ 474, 500, 55 },
		{ 474, 501, 55 },
		{ 474, 502, 55 },
		{ 474, 503, 55 },
		{ 474, 504, 55 },
		{ 474, 505, 55 },
		{ 474, 506, 55 },
		{ 474, 507, 55 },
		{ 474, 508, 55 },
		{ 474, 509, 55 },
		{ 474, 510, 55 },
		{ 474, 511, 55 },
		{ 474, 512, 55 },
		{ 475, 476, 55 },
		{ 475, 477, 55 },
		{ 475, 478, 55 },
		{ 475, 479, 55 },
		{ 475, 480, 55 },
		{ 475, 481, 55 },
		{ 475, 482, 55 },
		{ 475, 483, 55 },
		{ 475, 484, 55 },
		{ 475, 485, 55 },
		{ 475, 486, 55 },
		{ 475, 487, 55 },
		{ 475, 488, 55 },
		{ 475, 489, 55 },
		{ 475, 490, 55 },
		{ 475, 491, 55 },
		{ 475, 492, 55 },
		{ 475, 493, 55 },
		{ 475, 494, 55 },
		{ 475, 495, 55 },
		{ 475, 496, 55 },
		{ 475, 497, 55 },
		{ 475, 498, 55 },
		{ 475, 499, 55 },
		{ 475, 500, 55 },
		{ 475, 501, 55 },
		{ 475, 502, 55 },
		{ 475, 503, 55 },
		{ 475, 504, 55 },
		{ 475, 505, 55 },
		{ 475, 506, 55 },
		{ 475, 507, 55 },
		{ 475, 508, 55 },
		{ 475, 509, 55 },
		{ 475, 510, 55 },
		{ 475, 511, 55 },
		{ 475, 512, 55 },
		{ 476, 477, 55 },
		{ 476, 478, 55 },
		{ 476, 479, 55 },
		{ 476, 480, 55 },
		{ 476, 481, 55 },
		{ 476, 482, 55 },
		{ 476, 483, 55 },
		{ 476, 484, 55 },
		{ 476, 485, 55 },
		{ 476, 486, 55 },
		{ 476, 487, 55 },
		{ 476, 488, 55 },
		{ 476, 489, 55 },
		{ 476, 490, 55 },
		{ 476, 491, 55 },
		{ 476, 492, 55 },
		{ 476, 493, 55 },
		{ 476, 494, 55 },
		{ 476, 495, 55 },
		{ 476, 496, 55 },
		{ 476, 497, 55 },
		{ 476, 498, 55 },
		{ 476, 499, 55 },
		{ 476, 500, 55 },
		{ 476, 501, 55 },
		{ 476, 502, 55 },
		{ 476, 503, 55 },
		{ 476, 504, 55 },
		{ 476, 505, 55 },
		{ 476, 506, 55 },
		{ 476, 507, 55 },
		{ 476, 508, 55 },
		{ 476, 509, 55 },
		{ 476, 510, 55 },
		{ 476, 511, 55 },
		{ 476, 512, 55 },
		{ 477, 478, 55 },
		{ 477, 479, 55 },
		{ 477, 480, 55 },
		{ 477, 481, 55 },
		{ 477, 482, 55 },
		{ 477, 483, 55 },
		{ 477, 484, 55 },
		{ 477, 485, 55 },
		{ 477, 486, 55 },
		{ 477, 487, 55 },
		{ 477, 488, 55 },
		{ 477, 489, 55 },
		{ 477, 490, 55 },
		{ 477, 491, 55 },
		{ 477, 492, 55 },
		{ 477, 493, 55 },
		{ 477, 494, 55 },
		{ 477, 495, 55 },
		{ 477, 496, 55 },
		{ 477, 497, 55 },
		{ 477, 498, 55 },
		{ 477, 499, 55 },
		{ 477, 500, 55 },
		{ 477, 501, 55 },
		{ 477, 502, 55 },
		{ 477, 503, 55 },
		{ 477, 504, 55 },
		{ 477, 505, 55 },
		{ 477, 506, 55 },
		{ 477, 507, 55 },
		{ 477, 508, 55 },
		{ 477, 509, 55 },
		{ 477, 510, 55 },
		{ 477, 511, 55 },
		{ 477, 512, 55 },
		{ 478, 479, 55 },
		{ 478, 480, 55 },
		{ 478, 481, 55 },
		{ 478, 482, 55 },
		{ 478, 483, 55 },
		{ 478, 484, 55 },
		{ 478, 485, 55 },
		{ 478, 486, 55 },
		{ 478, 487, 55 },
		{ 478, 488, 55 },
		{ 478, 489, 55 },
		{ 478, 490, 55 },
		{ 478, 491, 55 },
		{ 478, 492, 55 },
		{ 478, 493, 55 },
		{ 478, 494, 55 },
		{ 478, 495, 55 },
		{ 478, 496, 55 },
		{ 478, 497, 55 },
		{ 478, 498, 55 },
		{ 478, 499, 55 },
		{ 478, 500, 55 },
		{ 478, 501, 55 },
		{ 478, 502, 55 },
		{ 478, 503, 55 },
		{ 478, 504, 55 },
		{ 478, 505, 55 },
		{ 478, 506, 55 },
		{ 478, 507, 55 },
		{ 478, 508, 55 },
		{ 478, 509, 55 },
		{ 478, 510, 55 },
		{ 478, 511, 55 },
		{ 478, 512, 55 },
		{ 479, 480, 55 },
		{ 479, 481, 55 },
		{ 479, 482, 55 },
		{ 479, 483, 55 },
		{ 479, 484, 55 },
		{ 479, 485, 55 },
		{ 479, 486, 55 },
		{ 479, 487, 55 },
		{ 479, 488, 55 },
		{ 479, 489, 55 },
		{ 479, 490, 55 },
		{ 479, 491, 55 },
		{ 479, 492, 55 },
		{ 479, 493, 55 },
		{ 479, 494, 55 },
		{ 479, 495, 55 },
		{ 479, 496, 55 },
		{ 479, 497, 55 },
		{ 479, 498, 55 },
		{ 479, 499, 55 },
		{ 479, 500, 55 },
		{ 479, 501, 55 },
		{ 479, 502, 55 },
		{ 479, 503, 55 },
		{ 479, 504, 55 },
		{ 479, 505, 55 },
		{ 479, 506, 55 },
		{ 479, 507, 55 },
		{ 479, 508, 55 },
		{ 479, 509, 55 },
		{ 479, 510, 55 },
		{ 479, 511, 55 },
		{ 479, 512, 55 },
		{ 480, 481, 55 },
		{ 480, 482, 55 },
		{ 480, 483, 55 },
		{ 480, 484, 55 },
		{ 480, 485, 55 },
		{ 480, 486, 55 },
		{ 480, 487, 55 },
		{ 480, 488, 55 },
		{ 480, 489, 55 },
		{ 480, 490, 55 },
		{ 480, 491, 55 },
		{ 480, 492, 55 },
		{ 480, 493, 55 },
		{ 480, 494, 55 },
		{ 480, 495, 55 },
		{ 480, 496, 55 },
		{ 480, 497, 55 },
		{ 480, 498, 55 },
		{ 480, 499, 55 },
		{ 480, 500, 55 },
		{ 480, 501, 55 },
		{ 480, 502, 55 },
		{ 480, 503, 55 },
		{ 480, 504, 55 },
		{ 480, 505, 55 },
		{ 480, 506, 55 },
		{ 480, 507, 55 },
		{ 480, 508, 55 },
		{ 480, 509, 55 },
		{ 480, 510, 55 },
		{ 480, 511, 55 },
		{ 480, 512, 55 },
		{ 481, 482, 55 },
		{ 481, 483, 55 },
		{ 481, 484, 55 },
		{ 481, 485, 55 },
		{ 481, 486, 55 },
		{ 481, 487, 55 },
		{ 481, 488, 55 },
		{ 481, 489, 55 },
		{ 481, 490, 55 },
		{ 481, 491, 55 },
		{ 481, 492, 55 },
		{ 481, 493, 55 },
		{ 481, 494, 55 },
		{ 481, 495, 55 },
		{ 481, 496, 55 },
		{ 481, 497, 55 },
		{ 481, 498, 55 },
		{ 481, 499, 55 },
		{ 481, 500, 55 },
		{ 481, 501, 55 },
		{ 481, 502, 55 },
		{ 481, 503, 55 },
		{ 481, 504, 55 },
		{ 481, 505, 55 },
		{ 481, 506, 55 },
		{ 481, 507, 55 },
		{ 481, 508, 55 },
		{ 481, 509, 55 },
		{ 481, 510, 55 },
		{ 481, 511, 55 },
		{ 481, 512, 55 },
		{ 482, 483, 55 },
		{ 482, 484, 55 },
		{ 482, 485, 55 },
		{ 482, 486, 55 },
		{ 482, 487, 55 },
		{ 482, 488, 55 },
		{ 482, 489, 55 },
		{ 482, 490, 55 },
		{ 482, 491, 55 },
		{ 482, 492, 55 },
		{ 482, 493, 55 },
		{ 482, 494, 55 },
		{ 482, 495, 55 },
		{ 482, 496, 55 },
		{ 482, 497, 55 },
		{ 482, 498, 55 },
		{ 482, 499, 55 },
		{ 482, 500, 55 },
		{ 482, 501, 55 },
		{ 482, 502, 55 },
		{ 482, 503, 55 },
		{ 482, 504, 55 },
		{ 482, 505, 55 },
		{ 482, 506, 55 },
		{ 482, 507, 55 },
		{ 482, 508, 55 },
		{ 482, 509, 55 },
		{ 482, 510, 55 },
		{ 482, 511, 55 },
		{ 482, 512, 55 },
		{ 483, 484, 55 },
		{ 483, 485, 55 },
		{ 483, 486, 55 },
		{ 483, 487, 55 },
		{ 483, 488, 55 },
		{ 483, 489, 55 },
		{ 483, 490, 55 },
		{ 483, 491, 55 },
		{ 483, 492, 55 },
		{ 483, 493, 55 },
		{ 483, 494, 55 },
		{ 483, 495, 55 },
		{ 483, 496, 55 },
		{ 483, 497, 55 },
		{ 483, 498, 55 },
		{ 483, 499, 55 },
		{ 483, 500, 55 },
		{ 483, 501, 55 },
		{ 483, 502, 55 },
		{ 483, 503, 55 },
		{ 483, 504, 55 },
		{ 483, 505, 55 },
		{ 483, 506, 55 },
		{ 483, 507, 55 },
		{ 483, 508, 55 },
		{ 483, 509, 55 },
		{ 483, 510, 55 },
		{ 483, 511, 55 },
		{ 483, 512, 55 },
		{ 484, 485, 55 },
		{ 484, 486, 55 },
		{ 484, 487, 55 },
		{ 484, 488, 55 },
		{ 484, 489, 55 },
		{ 484, 490, 55 },
		{ 484, 491, 55 },
		{ 484, 492, 55 },
		{ 484, 493, 55 },
		{ 484, 494, 55 },
		{ 484, 495, 55 },
		{ 484, 496, 55 },
		{ 484, 497, 55 },
		{ 484, 498, 55 },
		{ 484, 499, 55 },
		{ 484, 500, 55 },
		{ 484, 501, 55 },
		{ 484, 502, 55 },
		{ 484, 503, 55 },
		{ 484, 504, 55 },
		{ 484, 505, 55 },
		{ 484, 506, 55 },
		{ 484, 507, 55 },
		{ 484, 508, 55 },
		{ 484, 509, 55 },
		{ 484, 510, 55 },
		{ 484, 511, 55 },
		{ 484, 512, 55 },
		{ 485, 486, 55 },
		{ 485, 487, 55 },
		{ 485, 488, 55 },
		{ 485, 489, 55 },
		{ 485, 490, 55 },
		{ 485, 491, 55 },
		{ 485, 492, 55 },
		{ 485, 493, 55 },
		{ 485, 494, 55 },
		{ 485, 495, 55 },
		{ 485, 496, 55 },
		{ 485, 497, 55 },
		{ 485, 498, 55 },
		{ 485, 499, 55 },
		{ 485, 500, 55 },
		{ 485, 501, 55 },
		{ 485, 502, 55 },
		{ 485, 503, 55 },
		{ 485, 504, 55 },
		{ 485, 505, 55 },
		{ 485, 506, 55 },
		{ 485, 507, 55 },
		{ 485, 508, 55 },
		{ 485, 509, 55 },
		{ 485, 510, 55 },
		{ 485, 511, 55 },
		{ 485, 512, 55 },
		{ 486, 487, 55 },
		{ 486, 488, 55 },
		{ 486, 489, 55 },
		{ 486, 490, 55 },
		{ 486, 491, 55 },
		{ 486, 492, 55 },
		{ 486, 493, 55 },
		{ 486, 494, 55 },
		{ 486, 495, 55 },
		{ 486, 496, 55 },
		{ 486, 497, 55 },
		{ 486, 498, 55 },
		{ 486, 499, 55 },
		{ 486, 500, 55 },
		{ 486, 501, 55 },
		{ 486, 502, 55 },
		{ 486, 503, 55 },
		{ 486, 504, 55 },
		{ 486, 505, 55 },
		{ 486, 506, 55 },
		{ 486, 507, 55 },
		{ 486, 508, 55 },
		{ 486, 509, 55 },
		{ 486, 510, 55 },
		{ 486, 511, 55 },
		{ 486, 512, 55 },
		{ 487, 488, 55 },
		{ 487, 489, 55 },
		{ 487, 490, 55 },
		{ 487, 491, 55 },
		{ 487, 492, 55 },
		{ 487, 493, 55 },
		{ 487, 494, 55 },
		{ 487, 495, 55 },
		{ 487, 496, 55 },
		{ 487, 497, 55 },
		{ 487, 498, 55 },
		{ 487, 499, 55 },
		{ 487, 500, 55 },
		{ 487, 501, 55 },
		{ 487, 502, 55 },
		{ 487, 503, 55 },
		{ 487, 504, 55 },
		{ 487, 505, 55 },
		{ 487, 506, 55 },
		{ 487, 507, 55 },
		{ 487, 508, 55 },
		{ 487, 509, 55 },
		{ 487, 510, 55 },
		{ 487, 511, 55 },
		{ 487, 512, 55 },
		{ 488, 489, 55 },
		{ 488, 490, 55 },
		{ 488, 491, 55 },
		{ 488, 492, 55 },
		{ 488, 493, 55 },
		{ 488, 494, 55 },
		{ 488, 495, 55 },
		{ 488, 496, 55 },
		{ 488, 497, 55 },
		{ 488, 498, 55 },
		{ 488, 499, 55 },
		{ 488, 500, 55 },
		{ 488, 501, 55 },
		{ 488, 502, 55 },
		{ 488, 503, 55 },
		{ 488, 504, 55 },
		{ 488, 505, 55 },
		{ 488, 506, 55 },
		{ 488, 507, 55 },
		{ 488, 508, 55 },
		{ 488, 509, 55 },
		{ 488, 510, 55 },
		{ 488, 511, 55 },
		{ 488, 512, 55 },
		{ 489, 490, 55 },
		{ 489, 491, 55 },
		{ 489, 492, 55 },
		{ 489, 493, 55 },
		{ 489, 494, 55 },
		{ 489, 495, 55 },
		{ 489, 496, 55 },
		{ 489, 497, 55 },
		{ 489, 498, 55 },
		{ 489, 499, 55 },
		{ 489, 500, 55 },
		{ 489, 501, 55 },
		{ 489, 502, 55 },
		{ 489, 503, 55 },
		{ 489, 504, 55 },
		{ 489, 505, 55 },
		{ 489, 506, 55 },
		{ 489, 507, 55 },
		{ 489, 508, 55 },
		{ 489, 509, 55 },
		{ 489, 510, 55 },
		{ 489, 511, 55 },
		{ 489, 512, 55 },
		{ 490, 491, 55 },
		{ 490, 492, 55 },
		{ 490, 493, 55 },
		{ 490, 494, 55 },
		{ 490, 495, 55 },
		{ 490, 496, 55 },
		{ 490, 497, 55 },
		{ 490, 498, 55 },
		{ 490, 499, 55 },
		{ 490, 500, 55 },
		{ 490, 501, 55 },
		{ 490, 502, 55 },
		{ 490, 503, 55 },
		{ 490, 504, 55 },
		{ 490, 505, 55 },
		{ 490, 506, 55 },
		{ 490, 507, 55 },
		{ 490, 508, 55 },
		{ 490, 509, 55 },
		{ 490, 510, 55 },
		{ 490, 511, 55 },
		{ 490, 512, 55 },
		{ 491, 492, 55 },
		{ 491, 493, 55 },
		{ 491, 494, 55 },
		{ 491, 495, 55 },
		{ 491, 496, 55 },
		{ 491, 497, 55 },
		{ 491, 498, 55 },
		{ 491, 499, 55 },
		{ 491, 500, 55 },
		{ 491, 501, 55 },
		{ 491, 502, 55 },
		{ 491, 503, 55 },
		{ 491, 504, 55 },
		{ 491, 505, 55 },
		{ 491, 506, 55 },
		{ 491, 507, 55 },
		{ 491, 508, 55 },
		{ 491, 509, 55 },
		{ 491, 510, 55 },
		{ 491, 511, 55 },
		{ 491, 512, 55 },
		{ 492, 493, 55 },
		{ 492, 494, 55 },
		{ 492, 495, 55 },
		{ 492, 496, 55 },
		{ 492, 497, 55 },
		{ 492, 498, 55 },
		{ 492, 499, 55 },
		{ 492, 500, 55 },
		{ 492, 501, 55 },
		{ 492, 502, 55 },
		{ 492, 503, 55 },
		{ 492, 504, 55 },
		{ 492, 505, 55 },
		{ 492, 506, 55 },
		{ 492, 507, 55 },
		{ 492, 508, 55 },
		{ 492, 509, 55 },
		{ 492, 510, 55 },
		{ 492, 511, 55 },
		{ 492, 512, 55 },
		{ 493, 494, 55 },
		{ 493, 495, 55 },
		{ 493, 496, 55 },
		{ 493, 497, 55 },
		{ 493, 498, 55 },
		{ 493, 499, 55 },
		{ 493, 500, 55 },
		{ 493, 501, 55 },
		{ 493, 502, 55 },
		{ 493, 503, 55 },
		{ 493, 504, 55 },
		{ 493, 505, 55 },
		{ 493, 506, 55 },
		{ 493, 507, 55 },
		{ 493, 508, 55 },
		{ 493, 509, 55 },
		{ 493, 510, 55 },
		{ 493, 511, 55 },
		{ 493, 512, 55 },
		{ 494, 495, 55 },
		{ 494, 496, 55 },
		{ 494, 497, 55 },
		{ 494, 498, 55 },
		{ 494, 499, 55 },
		{ 494, 500, 55 },
		{ 494, 501, 55 },
		{ 494, 502, 55 },
		{ 494, 503, 55 },
		{ 494, 504, 55 },
		{ 494, 505, 55 },
		{ 494, 506, 55 },
		{ 494, 507, 55 },
		{ 494, 508, 55 },
		{ 494, 509, 55 },
		{ 494, 510, 55 },
		{ 494, 511, 55 },
		{ 494, 512, 55 },
		{ 495, 496, 55 },
		{ 495, 497, 55 },
		{ 495, 498, 55 },
		{ 495, 499, 55 },
		{ 495, 500, 55 },
		{ 495, 501, 55 },
		{ 495, 502, 55 },
		{ 495, 503, 55 },
		{ 495, 504, 55 },
		{ 495, 505, 55 },
		{ 495, 506, 55 },
		{ 495, 507, 55 },
		{ 495, 508, 55 },
		{ 495, 509, 55 },
		{ 495, 510, 55 },
		{ 495, 511, 55 },
		{ 495, 512, 55 },
		{ 496, 497, 55 },
		{ 496, 498, 55 },
		{ 496, 499, 55 },
		{ 496, 500, 55 },
		{ 496, 501, 55 },
		{ 496, 502, 55 },
		{ 496, 503, 55 },
		{ 496, 504, 55 },
		{ 496, 505, 55 },
		{ 496, 506, 55 },
		{ 496, 507, 55 },
		{ 496, 508, 55 },
		{ 496, 509, 55 },
		{ 496, 510, 55 },
		{ 496, 511, 55 },
		{ 496, 512, 55 },
		{ 497, 498, 55 },
		{ 497, 499, 55 },
		{ 497, 500, 55 },
		{ 497, 501, 55 },
		{ 497, 502, 55 },
		{ 497, 503, 55 },
		{ 497, 504, 55 },
		{ 497, 505, 55 },
		{ 497, 506, 55 },
		{ 497, 507, 55 },
		{ 497, 508, 55 },
		{ 497, 509, 55 },
		{ 497, 510, 55 },
		{ 497, 511, 55 },
		{ 497, 512, 55 },
		{ 498, 499, 55 },
		{ 498, 500, 55 },
		{ 498, 501, 55 },
		{ 498, 502, 55 },
		{ 498, 503, 55 },
		{ 498, 504, 55 },
		{ 498, 505, 55 },
		{ 498, 506, 55 },
		{ 498, 507, 55 },
		{ 498, 508, 55 },
		{ 498, 509, 55 },
		{ 498, 510, 55 },
		{ 498, 511, 55 },
		{ 498, 512, 55 },
		{ 499, 500, 55 },
		{ 499, 501, 55 },
		{ 499, 502, 55 },
		{ 499, 503, 55 },
		{ 499, 504, 55 },
		{ 499, 505, 55 },
		{ 499, 506, 55 },
		{ 499, 507, 55 },
		{ 499, 508, 55 },
		{ 499, 509, 55 },
		{ 499, 510, 55 },
		{ 499, 511, 55 },
		{ 499, 512, 55 },
		{ 500, 501, 55 },
		{ 500, 502, 55 },
		{ 500, 503, 55 },
		{ 500, 504, 55 },
		{ 500, 505, 55 },
		{ 500, 506, 55 },
		{ 500, 507, 55 },
		{ 500, 508, 55 },
		{ 500, 509, 55 },
		{ 500, 510, 55 },
		{ 500, 511, 55 },
		{ 500, 512, 55 },
		{ 501, 502, 55 },
		{ 501, 503, 55 },
		{ 501, 504, 55 },
		{ 501, 505, 55 },
		{ 501, 506, 55 },
		{ 501, 507, 55 },
		{ 501, 508, 55 },
		{ 501, 509, 55 },
		{ 501, 510, 55 },
		{ 501, 511, 55 },
		{ 501, 512, 55 },
		{ 502, 503, 55 },
		{ 502, 504, 55 },
		{ 502, 505, 55 },
		{ 502, 506, 55 },
		{ 502, 507, 55 },
		{ 502, 508, 55 },
		{ 502, 509, 55 },
		{ 502, 510, 55 },
		{ 502, 511, 55 },
		{ 502, 512, 55 },
		{ 503, 504, 55 },
		{ 503, 505, 55 },
		{ 503, 506, 55 },
		{ 503, 507, 55 },
		{ 503, 508, 55 },
		{ 503, 509, 55 },
		{ 503, 510, 55 },
		{ 503, 511, 55 },
		{ 503, 512, 55 },
		{ 504, 505, 55 },
		{ 504, 506, 55 },
		{ 504, 507, 55 },
		{ 504, 508, 55 },
		{ 504, 509, 55 },
		{ 504, 510, 55 },
		{ 504, 511, 55 },
		{ 504, 512, 55 },
		{ 505, 506, 55 },
		{ 505, 507, 55 },
		{ 505, 508, 55 },
		{ 505, 509, 55 },
		{ 505, 510, 55 },
		{ 505, 511, 55 },
		{ 505, 512, 55 },
		{ 506, 507, 55 },
		{ 506, 508, 55 },
		{ 506, 509, 55 },
		{ 506, 510, 55 },
		{ 506, 511, 55 },
		{ 506, 512, 55 },
		{ 507, 508, 55 },
		{ 507, 509, 55 },
		{ 507, 510, 55 },
		{ 507, 511, 55 },
		{ 507, 512, 55 },
		{ 508, 509, 55 },
		{ 508, 510, 55 },
		{ 508, 511, 55 },
		{ 508, 512, 55 },
		{ 509, 510, 55 },
		{ 509, 511, 55 },
		{ 509, 512, 55 },
		{ 510, 511, 55 },
		{ 510, 512, 55 },
		{ 511, 512, 55 },
		{ 34, 1, 56 },
		{ 34, 2, 56 },
		{ 35, 1, 56 },
		{ 35, 2, 56 },
		{ 35, 3, 56 },
		{ 36, 1, 56 },
		{ 36, 2, 56 },
		{ 36, 3, 56 },
		{ 36, 4, 56 },
		{ 37, 1, 56 },
		{ 37, 2, 56 },
		{ 37, 3, 56 },
		{ 37, 4, 56 },
		{ 37, 5, 56 },
		{ 38, 1, 56 },
		{ 38, 2, 56 },
		{ 38, 3, 56 },
		{ 38, 4, 56 },
		{ 38, 5, 56 },
		{ 38, 6, 56 },
		{ 39, 1, 56 },
		{ 39, 2, 56 },
		{ 39, 3, 56 },
		{ 39, 4, 56 },
		{ 39, 5, 56 },
		{ 39, 6, 56 },
		{ 39, 7, 56 },
		{ 40, 1, 56 },
		{ 40, 2, 56 },
		{ 40, 3, 56 },
		{ 40, 4, 56 },
		{ 40, 5, 56 },
		{ 40, 6, 56 },
		{ 40, 7, 56 },
		{ 40, 8, 56 },
		{ 41, 1, 56 },
		{ 41, 2, 56 },
		{ 41, 3, 56 },
		{ 41, 4, 56 },
		{ 41, 5, 56 },
		{ 41, 6, 56 },
		{ 41, 7, 56 },
		{ 41, 8, 56 },
		{ 41, 9, 56 },
		{ 41, 10, 56 },
		{ 42, 1, 56 },
		{ 42, 2, 56 },
		{ 42, 3, 56 },
		{ 42, 4, 56 },
		{ 42, 5, 56 },
		{ 42, 6, 56 },
		{ 42, 7, 56 },
		{ 42, 8, 56 },
		{ 42, 9, 56 },
		{ 42, 10, 56 },
		{ 42, 11, 56 },
		{ 43, 1, 56 },
		{ 43, 2, 56 },
		{ 43, 3, 56 },
		{ 43, 4, 56 },
		{ 43, 5, 56 },
		{ 43, 6, 56 },
		{ 43, 7, 56 },
		{ 43, 8, 56 },
		{ 43, 9, 56 },
		{ 43, 10, 56 },
		{ 43, 11, 56 },
		{ 43, 12, 56 },
		{ 44, 1, 56 },
		{ 44, 2, 56 },
		{ 44, 3, 56 },
		{ 44, 4, 56 },
		{ 44, 5, 56 },
		{ 44, 6, 56 },
		{ 44, 7, 56 },
		{ 44, 8, 56 },
		{ 44, 9, 56 },
		{ 44, 10, 56 },
		{ 44, 11, 56 },
		{ 44, 12, 56 },
		{ 44, 13, 56 },
		{ 45, 1, 56 },
		{ 45, 2, 56 },
		{ 45, 3, 56 },
		{ 45, 4, 56 },
		{ 45, 5, 56 },
		{ 45, 6, 56 },
		{ 45, 7, 56 },
		{ 45, 8, 56 },
		{ 45, 9, 56 },
		{ 45, 10, 56 },
		{ 45, 11, 56 },
		{ 45, 12, 56 },
		{ 45, 13, 56 },
		{ 45, 14, 56 },
		{ 46, 1, 56 },
		{ 46, 2, 56 },
		{ 46, 3, 56 },
		{ 46, 4, 56 },
		{ 46, 5, 56 },
		{ 46, 6, 56 },
		{ 46, 7, 56 },
		{ 46, 8, 56 },
		{ 46, 9, 56 },
		{ 46, 10, 56 },
		{ 46, 11, 56 },
		{ 46, 12, 56 },
		{ 46, 13, 56 },
		{ 46, 14, 56 },
		{ 46, 15, 56 },
		{ 47, 1, 56 },
		{ 47, 2, 56 },
		{ 47, 3, 56 },
		{ 47, 4, 56 },
		{ 47, 5, 56 },
		{ 47, 6, 56 },
		{ 47, 7, 56 },
		{ 47, 8, 56 },
		{ 47, 9, 56 },
		{ 47, 10, 56 },
		{ 47, 11, 56 },
		{ 47, 12, 56 },
		{ 47, 13, 56 },
		{ 47, 14, 56 },
		{ 47, 15, 56 },
		{ 47, 16, 56 },
		{ 48, 1, 56 },
		{ 48, 2, 56 },
		{ 48, 3, 56 },
		{ 48, 4, 56 },
		{ 48, 5, 56 },
		{ 48, 6, 56 },
		{ 48, 7, 56 },
		{ 48, 8, 56 },
		{ 48, 9, 56 },
		{ 48, 10, 56 },
		{ 48, 11, 56 },
		{ 48, 12, 56 },
		{ 48, 13, 56 },
		{ 48, 14, 56 },
		{ 48, 15, 56 },
		{ 48, 16, 56 },
		{ 48, 17, 56 },
		{ 48, 18, 56 },
		{ 49, 1, 56 },
		{ 49, 2, 56 },
		{ 49, 3, 56 },
		{ 49, 4, 56 },
		{ 49, 5, 56 },
		{ 49, 6, 56 },
		{ 49, 7, 56 },
		{ 49, 8, 56 },
		{ 49, 9, 56 },
		{ 49, 10, 56 },
		{ 49, 11, 56 },
		{ 49, 12, 56 },
		{ 49, 13, 56 },
		{ 49, 14, 56 },
		{ 49, 15, 56 },
		{ 49, 16, 56 },
		{ 49, 17, 56 },
		{ 49, 18, 56 },
		{ 49, 19, 56 },
		{ 50, 1, 56 },
		{ 50, 2, 56 },
		{ 50, 3, 56 },
		{ 50, 4, 56 },
		{ 50, 5, 56 },
		{ 50, 6, 56 },
		{ 50, 7, 56 },
		{ 50, 8, 56 },
		{ 50, 9, 56 },
		{ 50, 10, 56 },
		{ 50, 11, 56 },
		{ 50, 12, 56 },
		{ 50, 13, 56 },
		{ 50, 14, 56 },
		{ 50, 15, 56 },
		{ 50, 16, 56 },
		{ 50, 17, 56 },
		{ 50, 18, 56 },
		{ 50, 19, 56 },
		{ 50, 20, 56 },
		{ 51, 1, 56 },
		{ 51, 2, 56 },
		{ 51, 3, 56 },
		{ 51, 4, 56 },
		{ 51, 5, 56 },
		{ 51, 6, 56 },
		{ 51, 7, 56 },
		{ 51, 8, 56 },
		{ 51, 9, 56 },
		{ 51, 10, 56 },
		{ 51, 11, 56 },
		{ 51, 12, 56 },
		{ 51, 13, 56 },
		{ 51, 14, 56 },
		{ 51, 15, 56 },
		{ 51, 16, 56 },
		{ 51, 17, 56 },
		{ 51, 18, 56 },
		{ 51, 19, 56 },
		{ 51, 20, 56 },
		{ 51, 21, 56 },
		{ 52, 1, 56 },
		{ 52, 2, 56 },
		{ 52, 3, 56 },
		{ 52, 4, 56 },
		{ 52, 5, 56 },
		{ 52, 6, 56 },
		{ 52, 7, 56 },
		{ 52, 8, 56 },
		{ 52, 9, 56 },
		{ 52, 10, 56 },
		{ 52, 11, 56 },
		{ 52, 12, 56 },
		{ 52, 13, 56 },
		{ 52, 14, 56 },
		{ 52, 15, 56 },
		{ 52, 16, 56 },
		{ 52, 17, 56 },
		{ 52, 18, 56 },
		{ 52, 19, 56 },
		{ 52, 20, 56 },
		{ 52, 21, 56 },
		{ 52, 22, 56 },
		{ 53, 1, 56 },
		{ 53, 2, 56 },
		{ 53, 3, 56 },
		{ 53, 4, 56 },
		{ 53, 5, 56 },
		{ 53, 6, 56 },
		{ 53, 7, 56 },
		{ 53, 8, 56 },
		{ 53, 9, 56 },
		{ 53, 10, 56 },
		{ 53, 11, 56 },
		{ 53, 12, 56 },
		{ 53, 13, 56 },
		{ 53, 14, 56 },
		{ 53, 15, 56 },
		{ 53, 16, 56 },
		{ 53, 17, 56 },
		{ 53, 18, 56 },
		{ 53, 19, 56 },
		{ 53, 20, 56 },
		{ 53, 21, 56 },
		{ 53, 22, 56 },
		{ 53, 23, 56 },
		{ 54, 1, 56 },
		{ 54, 2, 56 },
		{ 54, 3, 56 },
		{ 54, 4, 56 },
		{ 54, 5, 56 },
		{ 54, 6, 56 },
		{ 54, 7, 56 },
		{ 54, 8, 56 },
		{ 54, 9, 56 },
		{ 54, 10, 56 },
		{ 54, 11, 56 },
		{ 54, 12, 56 },
		{ 54, 13, 56 },
		{ 54, 14, 56 },
		{ 54, 15, 56 },
		{ 54, 16, 56 },
		{ 54, 17, 56 },
		{ 54, 18, 56 },
		{ 54, 19, 56 },
		{ 54, 20, 56 },
		{ 54, 21, 56 },
		{ 54, 22, 56 },
		{ 54, 23, 56 },
		{ 54, 24, 56 },
		{ 55, 1, 56 },
		{ 55, 2, 56 },
		{ 55, 3, 56 },
		{ 55, 4, 56 },
		{ 55, 5, 56 },
		{ 55, 6, 56 },
		{ 55, 7, 56 },
		{ 55, 8, 56 },
		{ 55, 9, 56 },
		{ 55, 10, 56 },
		{ 55, 11, 56 },
		{ 55, 12, 56 },
		{ 55, 13, 56 },
		{ 55, 14, 56 },
		{ 55, 15, 56 },
		{ 55, 16, 56 },
		{ 55, 17, 56 },
		{ 55, 18, 56 },
		{ 55, 19, 56 },
		{ 55, 20, 56 },
		{ 55, 21, 56 },
		{ 55, 22, 56 },
		{ 55, 23, 56 },
		{ 55, 24, 56 },
		{ 55, 25, 56 },
		{ 55, 26, 56 },
		{ 56, 1, 56 },
		{ 56, 2, 56 },
		{ 56, 3, 56 },
		{ 56, 4, 56 },
		{ 56, 5, 56 },
		{ 56, 6, 56 },
		{ 56, 7, 56 },
		{ 56, 8, 56 },
		{ 56, 9, 56 },
		{ 56, 10, 56 },
		{ 56, 11, 56 },
		{ 56, 12, 56 },
		{ 56, 13, 56 },
		{ 56, 14, 56 },
		{ 56, 15, 56 },
		{ 56, 16, 56 },
		{ 56, 17, 56 },
		{ 56, 18, 56 },
		{ 56, 19, 56 },
		{ 56, 20, 56 },
		{ 56, 21, 56 },
		{ 56, 22, 56 },
		{ 56, 23, 56 },
		{ 56, 24, 56 },
		{ 56, 25, 56 },
		{ 56, 26, 56 },
		{ 56, 27, 56 },
		{ 57, 1, 56 },
		{ 57, 2, 56 },
		{ 57, 3, 56 },
		{ 57, 4, 56 },
		{ 57, 5, 56 },
		{ 57, 6, 56 },
		{ 57, 7, 56 },
		{ 57, 8, 56 },
		{ 57, 9, 56 },
		{ 57, 10, 56 },
		{ 57, 11, 56 },
		{ 57, 12, 56 },
		{ 57, 13, 56 },
		{ 57, 14, 56 },
		{ 57, 15, 56 },
		{ 57, 16, 56 },
		{ 57, 17, 56 },
		{ 57, 18, 56 },
		{ 57, 19, 56 },
		{ 57, 20, 56 },
		{ 57, 21, 56 },
		{ 57, 22, 56 },
		{ 57, 23, 56 },
		{ 57, 24, 56 },
		{ 57, 25, 56 },
		{ 57, 26, 56 },
		{ 57, 27, 56 },
		{ 57, 28, 56 },
		{ 58, 1, 56 },
		{ 58, 2, 56 },
		{ 58, 3, 56 },
		{ 58, 4, 56 },
		{ 58, 5, 56 },
		{ 58, 6, 56 },
		{ 58, 7, 56 },
		{ 58, 8, 56 },
		{ 58, 9, 56 },
		{ 58, 10, 56 },
		{ 58, 11, 56 },
		{ 58, 12, 56 },
		{ 58, 13, 56 },
		{ 58, 14, 56 },
		{ 58, 15, 56 },
		{ 58, 16, 56 },
		{ 58, 17, 56 },
		{ 58, 18, 56 },
		{ 58, 19, 56 },
		{ 58, 20, 56 },
		{ 58, 21, 56 },
		{ 58, 22, 56 },
		{ 58, 23, 56 },
		{ 58, 24, 56 },
		{ 58, 25, 56 },
		{ 58, 26, 56 },
		{ 58, 27, 56 },
		{ 58, 28, 56 },
		{ 58, 29, 56 },
		{ 59, 1, 56 },
		{ 59, 2, 56 },
		{ 59, 3, 56 },
		{ 59, 4, 56 },
		{ 59, 5, 56 },
		{ 59, 6, 56 },
		{ 59, 7, 56 },
		{ 59, 8, 56 },
		{ 59, 9, 56 },
		{ 59, 10, 56 },
		{ 59, 11, 56 },
		{ 59, 12, 56 },
		{ 59, 13, 56 },
		{ 59, 14, 56 },
		{ 59, 15, 56 },
		{ 59, 16, 56 },
		{ 59, 17, 56 },
		{ 59, 18, 56 },
		{ 59, 19, 56 },
		{ 59, 20, 56 },
		{ 59, 21, 56 },
		{ 59, 22, 56 },
		{ 59, 23, 56 },
		{ 59, 24, 56 },
		{ 59, 25, 56 },
		{ 59, 26, 56 },
		{ 59, 27, 56 },
		{ 59, 28, 56 },
		{ 59, 29, 56 },
		{ 59, 30, 56 },
		{ 60, 1, 56 },
		{ 60, 2, 56 },
		{ 60, 3, 56 },
		{ 60, 4, 56 },
		{ 60, 5, 56 },
		{ 60, 6, 56 },
		{ 60, 7, 56 },
		{ 60, 8, 56 },
		{ 60, 9, 56 },
		{ 60, 10, 56 },
		{ 60, 11, 56 },
		{ 60, 12, 56 },
		{ 60, 13, 56 },
		{ 60, 14, 56 },
		{ 60, 15, 56 },
		{ 60, 16, 56 },
		{ 60, 17, 56 },
		{ 60, 18, 56 },
		{ 60, 19, 56 },
		{ 60, 20, 56 },
		{ 60, 21, 56 },
		{ 60, 22, 56 },
		{ 60, 23, 56 },
		{ 60, 24, 56 },
		{ 60, 25, 56 },
		{ 60, 26, 56 },
		{ 60, 27, 56 },
		{ 60, 28, 56 },
		{ 60, 29, 56 },
		{ 60, 30, 56 },
		{ 60, 31, 56 },
		{ 61, 1, 56 },
		{ 61, 2, 56 },
		{ 61, 3, 56 },
		{ 61, 4, 56 },
		{ 61, 5, 56 },
		{ 61, 6, 56 },
		{ 61, 7, 56 },
		{ 61, 8, 56 },
		{ 61, 9, 56 },
		{ 61, 10, 56 },
		{ 61, 11, 56 },
		{ 61, 12, 56 },
		{ 61, 13, 56 },
		{ 61, 14, 56 },
		{ 61, 15, 56 },
		{ 61, 16, 56 },
		{ 61, 17, 56 },
		{ 61, 18, 56 },
		{ 61, 19, 56 },
		{ 61, 20, 56 },
		{ 61, 21, 56 },
		{ 61, 22, 56 },
		{ 61, 23, 56 },
		{ 61, 24, 56 },
		{ 61, 25, 56 },
		{ 61, 26, 56 },
		{ 61, 27, 56 },
		{ 61, 28, 56 },
		{ 61, 29, 56 },
		{ 61, 30, 56 },
		{ 61, 31, 56 },
		{ 61, 32, 56 },
		{ 62, 1, 56 },
		{ 62, 2, 56 },
		{ 62, 3, 56 },
		{ 62, 4, 56 },
		{ 62, 5, 56 },
		{ 62, 6, 56 },
		{ 62, 7, 56 },
		{ 62, 8, 56 },
		{ 62, 9, 56 },
		{ 62, 10, 56 },
		{ 62, 11, 56 },
		{ 62, 12, 56 },
		{ 62, 13, 56 },
		{ 62, 14, 56 },
		{ 62, 15, 56 },
		{ 62, 16, 56 },
		{ 62, 17, 56 },
		{ 62, 18, 56 },
		{ 62, 19, 56 },
		{ 62, 20, 56 },
		{ 62, 21, 56 },
		{ 62, 22, 56 },
		{ 62, 23, 56 },
		{ 62, 24, 56 },
		{ 62, 25, 56 },
		{ 62, 26, 56 },
		{ 62, 27, 56 },
		{ 62, 28, 56 },
		{ 62, 29, 56 },
		{ 62, 30, 56 },
		{ 62, 31, 56 },
		{ 62, 32, 56 },
		{ 62, 33, 56 },
		{ 62, 34, 56 },
		{ 63, 1, 56 },
		{ 63, 2, 56 },
		{ 63, 3, 56 },
		{ 63, 4, 56 },
		{ 63, 5, 56 },
		{ 63, 6, 56 },
		{ 63, 7, 56 },
		{ 63, 8, 56 },
		{ 63, 9, 56 },
		{ 63, 10, 56 },
		{ 63, 11, 56 },
		{ 63, 12, 56 },
		{ 63, 13, 56 },
		{ 63, 14, 56 },
		{ 63, 15, 56 },
		{ 63, 16, 56 },
		{ 63, 17, 56 },
		{ 63, 18, 56 },
		{ 63, 19, 56 },
		{ 63, 20, 56 },
		{ 63, 21, 56 },
		{ 63, 22, 56 },
		{ 63, 23, 56 },
		{ 63, 24, 56 },
		{ 63, 25, 56 },
		{ 63, 26, 56 },
		{ 63, 27, 56 },
		{ 63, 28, 56 },
		{ 63, 29, 56 },
		{ 63, 30, 56 },
		{ 63, 31, 56 },
		{ 63, 32, 56 },
		{ 63, 33, 56 },
		{ 63, 34, 56 },
		{ 63, 35, 56 },
		{ 64, 1, 56 },
		{ 64, 2, 56 },
		{ 64, 3, 56 },
		{ 64, 4, 56 },
		{ 64, 5, 56 },
		{ 64, 6, 56 },
		{ 64, 7, 56 },
		{ 64, 8, 56 },
		{ 64, 9, 56 },
		{ 64, 10, 56 },
		{ 64, 11, 56 },
		{ 64, 12, 56 },
		{ 64, 13, 56 },
		{ 64, 14, 56 },
		{ 64, 15, 56 },
		{ 64, 16, 56 },
		{ 64, 17, 56 },
		{ 64, 18, 56 },
		{ 64, 19, 56 },
		{ 64, 20, 56 },
		{ 64, 21, 56 },
		{ 64, 22, 56 },
		{ 64, 23, 56 },
		{ 64, 24, 56 },
		{ 64, 25, 56 },
		{ 64, 26, 56 },
		{ 64, 27, 56 },
		{ 64, 28, 56 },
		{ 64, 29, 56 },
		{ 64, 30, 56 },
		{ 64, 31, 56 },
		{ 64, 32, 56 },
		{ 64, 33, 56 },
		{ 64, 34, 56 },
		{ 64, 35, 56 },
		{ 64, 36, 56 },
		{ 65, 1, 56 },
		{ 65, 2, 56 },
		{ 65, 3, 56 },
		{ 65, 4, 56 },
		{ 65, 5, 56 },
		{ 65, 6, 56 },
		{ 65, 7, 56 },
		{ 65, 8, 56 },
		{ 65, 9, 56 },
		{ 65, 10, 56 },
		{ 65, 11, 56 },
		{ 65, 12, 56 },
		{ 65, 13, 56 },
		{ 65, 14, 56 },
		{ 65, 15, 56 },
		{ 65, 16, 56 },
		{ 65, 17, 56 },
		{ 65, 18, 56 },
		{ 65, 19, 56 },
		{ 65, 20, 56 },
		{ 65, 21, 56 },
		{ 65, 22, 56 },
		{ 65, 23, 56 },
		{ 65, 24, 56 },
		{ 65, 25, 56 },
		{ 65, 26, 56 },
		{ 65, 27, 56 },
		{ 65, 28, 56 },
		{ 65, 29, 56 },
		{ 65, 30, 56 },
		{ 65, 31, 56 },
		{ 65, 32, 56 },
		{ 65, 33, 56 },
		{ 65, 34, 56 },
		{ 65, 35, 56 },
		{ 65, 36, 56 },
		{ 65, 37, 56 },
		{ 66, 1, 56 },
		{ 66, 2, 56 },
		{ 66, 3, 56 },
		{ 66, 4, 56 },
		{ 66, 5, 56 },
		{ 66, 6, 56 },
		{ 66, 7, 56 },
		{ 66, 8, 56 },
		{ 66, 9, 56 },
		{ 66, 10, 56 },
		{ 66, 11, 56 },
		{ 66, 12, 56 },
		{ 66, 13, 56 },
		{ 66, 14, 56 },
		{ 66, 15, 56 },
		{ 66, 16, 56 },
		{ 66, 17, 56 },
		{ 66, 18, 56 },
		{ 66, 19, 56 },
		{ 66, 20, 56 },
		{ 66, 21, 56 },
		{ 66, 22, 56 },
		{ 66, 23, 56 },
		{ 66, 24, 56 },
		{ 66, 25, 56 },
		{ 66, 26, 56 },
		{ 66, 27, 56 },
		{ 66, 28, 56 },
		{ 66, 29, 56 },
		{ 66, 30, 56 },
		{ 66, 31, 56 },
		{ 66, 32, 56 },
		{ 66, 33, 56 },
		{ 66, 34, 56 },
		{ 66, 35, 56 },
		{ 66, 36, 56 },
		{ 66, 37, 56 },
		{ 66, 38, 56 },
		{ 67, 1, 56 },
		{ 67, 2, 56 },
		{ 67, 3, 56 },
		{ 67, 4, 56 },
		{ 67, 5, 56 },
		{ 67, 6, 56 },
		{ 67, 7, 56 },
		{ 67, 8, 56 },
		{ 67, 9, 56 },
		{ 67, 10, 56 },
		{ 67, 11, 56 },
		{ 67, 12, 56 },
		{ 67, 13, 56 },
		{ 67, 14, 56 },
		{ 67, 15, 56 },
		{ 67, 16, 56 },
		{ 67, 17, 56 },
		{ 67, 18, 56 },
		{ 67, 19, 56 },
		{ 67, 20, 56 },
		{ 67, 21, 56 },
		{ 67, 22, 56 },
		{ 67, 23, 56 },
		{ 67, 24, 56 },
		{ 67, 25, 56 },
		{ 67, 26, 56 },
		{ 67, 27, 56 },
		{ 67, 28, 56 },
		{ 67, 29, 56 },
		{ 67, 30, 56 },
		{ 67, 31, 56 },
		{ 67, 32, 56 },
		{ 67, 33, 56 },
		{ 67, 34, 56 },
		{ 67, 35, 56 },
		{ 67, 36, 56 },
		{ 67, 37, 56 },
		{ 67, 38, 56 },
		{ 67, 39, 56 },
		{ 68, 1, 56 },
		{ 68, 2, 56 },
		{ 68, 3, 56 },
		{ 68, 4, 56 },
		{ 68, 5, 56 },
		{ 68, 6, 56 },
		{ 68, 7, 56 },
		{ 68, 8, 56 },
		{ 68, 9, 56 },
		{ 68, 10, 56 },
		{ 68, 11, 56 },
		{ 68, 12, 56 },
		{ 68, 13, 56 },
		{ 68, 14, 56 },
		{ 68, 15, 56 },
		{ 68, 16, 56 },
		{ 68, 17, 56 },
		{ 68, 18, 56 },
		{ 68, 19, 56 },
		{ 68, 20, 56 },
		{ 68, 21, 56 },
		{ 68, 22, 56 },
		{ 68, 23, 56 },
		{ 68, 24, 56 },
		{ 68, 25, 56 },
		{ 68, 26, 56 },
		{ 68, 27, 56 },
		{ 68, 28, 56 },
		{ 68, 29, 56 },
		{ 68, 30, 56 },
		{ 68, 31, 56 },
		{ 68, 32, 56 },
		{ 68, 33, 56 },
		{ 68, 34, 56 },
		{ 68, 35, 56 },
		{ 68, 36, 56 },
		{ 68, 37, 56 },
		{ 68, 38, 56 },
		{ 68, 39, 56 },
		{ 68, 40, 56 },
		{ 69, 1, 56 },
		{ 69, 2, 56 },
		{ 69, 3, 56 },
		{ 69, 4, 56 },
		{ 69, 5, 56 },
		{ 69, 6, 56 },
		{ 69, 7, 56 },
		{ 69, 8, 56 },
		{ 69, 9, 56 },
		{ 69, 10, 56 },
		{ 69, 11, 56 },
		{ 69, 12, 56 },
		{ 69, 13, 56 },
		{ 69, 14, 56 },
		{ 69, 15, 56 },
		{ 69, 16, 56 },
		{ 69, 17, 56 },
		{ 69, 18, 56 },
		{ 69, 19, 56 },
		{ 69, 20, 56 },
		{ 69, 21, 56 },
		{ 69, 22, 56 },
		{ 69, 23, 56 },
		{ 69, 24, 56 },
		{ 69, 25, 56 },
		{ 69, 26, 56 },
		{ 69, 27, 56 },
		{ 69, 28, 56 },
		{ 69, 29, 56 },
		{ 69, 30, 56 },
		{ 69, 31, 56 },
		{ 69, 32, 56 },
		{ 69, 33, 56 },
		{ 69, 34, 56 },
		{ 69, 35, 56 },
		{ 69, 36, 56 },
		{ 69, 37, 56 },
		{ 69, 38, 56 },
		{ 69, 39, 56 },
		{ 69, 40, 56 },
		{ 69, 41, 56 },
		{ 69, 42, 56 },
		{ 70, 1, 56 },
		{ 70, 2, 56 },
		{ 70, 3, 56 },
		{ 70, 4, 56 },
		{ 70, 5, 56 },
		{ 70, 6, 56 },
		{ 70, 7, 56 },
		{ 70, 8, 56 },
		{ 70, 9, 56 },
		{ 70, 10, 56 },
		{ 70, 11, 56 },
		{ 70, 12, 56 },
		{ 70, 13, 56 },
		{ 70, 14, 56 },
		{ 70, 15, 56 },
		{ 70, 16, 56 },
		{ 70, 17, 56 },
		{ 70, 18, 56 },
		{ 70, 19, 56 },
		{ 70, 20, 56 },
		{ 70, 21, 56 },
		{ 70, 22, 56 },
		{ 70, 23, 56 },
		{ 70, 24, 56 },
		{ 70, 25, 56 },
		{ 70, 26, 56 },
		{ 70, 27, 56 },
		{ 70, 28, 56 },
		{ 70, 29, 56 },
		{ 70, 30, 56 },
		{ 70, 31, 56 },
		{ 70, 32, 56 },
		{ 70, 33, 56 },
		{ 70, 34, 56 },
		{ 70, 35, 56 },
		{ 70, 36, 56 },
		{ 70, 37, 56 },
		{ 70, 38, 56 },
		{ 70, 39, 56 },
		{ 70, 40, 56 },
		{ 70, 41, 56 },
		{ 70, 42, 56 },
		{ 70, 43, 56 },
		{ 71, 1, 56 },
		{ 71, 2, 56 },
		{ 71, 3, 56 },
		{ 71, 4, 56 },
		{ 71, 5, 56 },
		{ 71, 6, 56 },
		{ 71, 7, 56 },
		{ 71, 8, 56 },
		{ 71, 9, 56 },
		{ 71, 10, 56 },
		{ 71, 11, 56 },
		{ 71, 12, 56 },
		{ 71, 13, 56 },
		{ 71, 14, 56 },
		{ 71, 15, 56 },
		{ 71, 16, 56 },
		{ 71, 17, 56 },
		{ 71, 18, 56 },
		{ 71, 19, 56 },
		{ 71, 20, 56 },
		{ 71, 21, 56 },
		{ 71, 22, 56 },
		{ 71, 23, 56 },
		{ 71, 24, 56 },
		{ 71, 25, 56 },
		{ 71, 26, 56 },
		{ 71, 27, 56 },
		{ 71, 28, 56 },
		{ 71, 29, 56 },
		{ 71, 30, 56 },
		{ 71, 31, 56 },
		{ 71, 32, 56 },
		{ 71, 33, 56 },
		{ 71, 34, 56 },
		{ 71, 35, 56 },
		{ 71, 36, 56 },
		{ 71, 37, 56 },
		{ 71, 38, 56 },
		{ 71, 39, 56 },
		{ 71, 40, 56 },
		{ 71, 41, 56 },
		{ 71, 42, 56 },
		{ 71, 43, 56 },
		{ 71, 44, 56 },
		{ 72, 1, 56 },
		{ 72, 2, 56 },
		{ 72, 3, 56 },
		{ 72, 4, 56 },
		{ 72, 5, 56 },
		{ 72, 6, 56 },
		{ 72, 7, 56 },
		{ 72, 8, 56 },
		{ 72, 9, 56 },
		{ 72, 10, 56 },
		{ 72, 11, 56 },
		{ 72, 12, 56 },
		{ 72, 13, 56 },
		{ 72, 14, 56 },
		{ 72, 15, 56 },
		{ 72, 16, 56 },
		{ 72, 17, 56 },
		{ 72, 18, 56 },
		{ 72, 19, 56 },
		{ 72, 20, 56 },
		{ 72, 21, 56 },
		{ 72, 22, 56 },
		{ 72, 23, 56 },
		{ 72, 24, 56 },
		{ 72, 25, 56 },
		{ 72, 26, 56 },
		{ 72, 27, 56 },
		{ 72, 28, 56 },
		{ 72, 29, 56 },
		{ 72, 30, 56 },
		{ 72, 31, 56 },
		{ 72, 32, 56 },
		{ 72, 33, 56 },
		{ 72, 34, 56 },
		{ 72, 35, 56 },
		{ 72, 36, 56 },
		{ 72, 37, 56 },
		{ 72, 38, 56 },
		{ 72, 39, 56 },
		{ 72, 40, 56 },
		{ 72, 41, 56 },
		{ 72, 42, 56 },
		{ 72, 43, 56 },
		{ 72, 44, 56 },
		{ 72, 45, 56 },
		{ 73, 1, 56 },
		{ 73, 2, 56 },
		{ 73, 3, 56 },
		{ 73, 4, 56 },
		{ 73, 5, 56 },
		{ 73, 6, 56 },
		{ 73, 7, 56 },
		{ 73, 8, 56 },
		{ 73, 9, 56 },
		{ 73, 10, 56 },
		{ 73, 11, 56 },
		{ 73, 12, 56 },
		{ 73, 13, 56 },
		{ 73, 14, 56 },
		{ 73, 15, 56 },
		{ 73, 16, 56 },
		{ 73, 17, 56 },
		{ 73, 18, 56 },
		{ 73, 19, 56 },
		{ 73, 20, 56 },
		{ 73, 21, 56 },
		{ 73, 22, 56 },
		{ 73, 23, 56 },
		{ 73, 24, 56 },
		{ 73, 25, 56 },
		{ 73, 26, 56 },
		{ 73, 27, 56 },
		{ 73, 28, 56 },
		{ 73, 29, 56 },
		{ 73, 30, 56 },
		{ 73, 31, 56 },
		{ 73, 32, 56 },
		{ 73, 33, 56 },
		{ 73, 34, 56 },
		{ 73, 35, 56 },
		{ 73, 36, 56 },
		{ 73, 37, 56 },
		{ 73, 38, 56 },
		{ 73, 39, 56 },
		{ 73, 40, 56 },
		{ 73, 41, 56 },
		{ 73, 42, 56 },
		{ 73, 43, 56 },
		{ 73, 44, 56 },
		{ 73, 45, 56 },
		{ 73, 46, 56 },
		{ 74, 1, 56 },
		{ 74, 2, 56 },
		{ 74, 3, 56 },
		{ 74, 4, 56 },
		{ 74, 5, 56 },
		{ 74, 6, 56 },
		{ 74, 7, 56 },
		{ 74, 8, 56 },
		{ 74, 9, 56 },
		{ 74, 10, 56 },
		{ 74, 11, 56 },
		{ 74, 12, 56 },
		{ 74, 13, 56 },
		{ 74, 14, 56 },
		{ 74, 15, 56 },
		{ 74, 16, 56 },
		{ 74, 17, 56 },
		{ 74, 18, 56 },
		{ 74, 19, 56 },
		{ 74, 20, 56 },
		{ 74, 21, 56 },
		{ 74, 22, 56 },
		{ 74, 23, 56 },
		{ 74, 24, 56 },
		{ 74, 25, 56 },
		{ 74, 26, 56 },
		{ 74, 27, 56 },
		{ 74, 28, 56 },
		{ 74, 29, 56 },
		{ 74, 30, 56 },
		{ 74, 31, 56 },
		{ 74, 32, 56 },
		{ 74, 33, 56 },
		{ 74, 34, 56 },
		{ 74, 35, 56 },
		{ 74, 36, 56 },
		{ 74, 37, 56 },
		{ 74, 38, 56 },
		{ 74, 39, 56 },
		{ 74, 40, 56 },
		{ 74, 41, 56 },
		{ 74, 42, 56 },
		{ 74, 43, 56 },
		{ 74, 44, 56 },
		{ 74, 45, 56 },
		{ 74, 46, 56 },
		{ 74, 47, 56 },
		{ 75, 1, 56 },
		{ 75, 2, 56 },
		{ 75, 3, 56 },
		{ 75, 4, 56 },
		{ 75, 5, 56 },
		{ 75, 6, 56 },
		{ 75, 7, 56 },
		{ 75, 8, 56 },
		{ 75, 9, 56 },
		{ 75, 10, 56 },
		{ 75, 11, 56 },
		{ 75, 12, 56 },
		{ 75, 13, 56 },
		{ 75, 14, 56 },
		{ 75, 15, 56 },
		{ 75, 16, 56 },
		{ 75, 17, 56 },
		{ 75, 18, 56 },
		{ 75, 19, 56 },
		{ 75, 20, 56 },
		{ 75, 21, 56 },
		{ 75, 22, 56 },
		{ 75, 23, 56 },
		{ 75, 24, 56 },
		{ 75, 25, 56 },
		{ 75, 26, 56 },
		{ 75, 27, 56 },
		{ 75, 28, 56 },
		{ 75, 29, 56 },
		{ 75, 30, 56 },
		{ 75, 31, 56 },
		{ 75, 32, 56 },
		{ 75, 33, 56 },
		{ 75, 34, 56 },
		{ 75, 35, 56 },
		{ 75, 36, 56 },
		{ 75, 37, 56 },
		{ 75, 38, 56 },
		{ 75, 39, 56 },
		{ 75, 40, 56 },
		{ 75, 41, 56 },
		{ 75, 42, 56 },
		{ 75, 43, 56 },
		{ 75, 44, 56 },
		{ 75, 45, 56 },
		{ 75, 46, 56 },
		{ 75, 47, 56 },
		{ 75, 48, 56 },
		{ 76, 1, 56 },
		{ 76, 2, 56 },
		{ 76, 3, 56 },
		{ 76, 4, 56 },
		{ 76, 5, 56 },
		{ 76, 6, 56 },
		{ 76, 7, 56 },
		{ 76, 8, 56 },
		{ 76, 9, 56 },
		{ 76, 10, 56 },
		{ 76, 11, 56 },
		{ 76, 12, 56 },
		{ 76, 13, 56 },
		{ 76, 14, 56 },
		{ 76, 15, 56 },
		{ 76, 16, 56 },
		{ 76, 17, 56 },
		{ 76, 18, 56 },
		{ 76, 19, 56 },
		{ 76, 20, 56 },
		{ 76, 21, 56 },
		{ 76, 22, 56 },
		{ 76, 23, 56 },
		{ 76, 24, 56 },
		{ 76, 25, 56 },
		{ 76, 26, 56 },
		{ 76, 27, 56 },
		{ 76, 28, 56 },
		{ 76, 29, 56 },
		{ 76, 30, 56 },
		{ 76, 31, 56 },
		{ 76, 32, 56 },
		{ 76, 33, 56 },
		{ 76, 34, 56 },
		{ 76, 35, 56 },
		{ 76, 36, 56 },
		{ 76, 37, 56 },
		{ 76, 38, 56 },
		{ 76, 39, 56 },
		{ 76, 40, 56 },
		{ 76, 41, 56 },
		{ 76, 42, 56 },
		{ 76, 43, 56 },
		{ 76, 44, 56 },
		{ 76, 45, 56 },
		{ 76, 46, 56 },
		{ 76, 47, 56 },
		{ 76, 48, 56 },
		{ 76, 49, 56 },
		{ 76, 50, 56 },
		{ 77, 1, 56 },
		{ 77, 2, 56 },
		{ 77, 3, 56 },
		{ 77, 4, 56 },
		{ 77, 5, 56 },
		{ 77, 6, 56 },
		{ 77, 7, 56 },
		{ 77, 8, 56 },
		{ 77, 9, 56 },
		{ 77, 10, 56 },
		{ 77, 11, 56 },
		{ 77, 12, 56 },
		{ 77, 13, 56 },
		{ 77, 14, 56 },
		{ 77, 15, 56 },
		{ 77, 16, 56 },
		{ 77, 17, 56 },
		{ 77, 18, 56 },
		{ 77, 19, 56 },
		{ 77, 20, 56 },
		{ 77, 21, 56 },
		{ 77, 22, 56 },
		{ 77, 23, 56 },
		{ 77, 24, 56 },
		{ 77, 25, 56 },
		{ 77, 26, 56 },
		{ 77, 27, 56 },
		{ 77, 28, 56 },
		{ 77, 29, 56 },
		{ 77, 30, 56 },
		{ 77, 31, 56 },
		{ 77, 32, 56 },
		{ 77, 33, 56 },
		{ 77, 34, 56 },
		{ 77, 35, 56 },
		{ 77, 36, 56 },
		{ 77, 37, 56 },
		{ 77, 38, 56 },
		{ 77, 39, 56 },
		{ 77, 40, 56 },
		{ 77, 41, 56 },
		{ 77, 42, 56 },
		{ 77, 43, 56 },
		{ 77, 44, 56 },
		{ 77, 45, 56 },
		{ 77, 46, 56 },
		{ 77, 47, 56 },
		{ 77, 48, 56 },
		{ 77, 49, 56 },
		{ 77, 50, 56 },
		{ 77, 51, 56 },
		{ 78, 1, 56 },
		{ 78, 2, 56 },
		{ 78, 3, 56 },
		{ 78, 4, 56 },
		{ 78, 5, 56 },
		{ 78, 6, 56 },
		{ 78, 7, 56 },
		{ 78, 8, 56 },
		{ 78, 9, 56 },
		{ 78, 10, 56 },
		{ 78, 11, 56 },
		{ 78, 12, 56 },
		{ 78, 13, 56 },
		{ 78, 14, 56 },
		{ 78, 15, 56 },
		{ 78, 16, 56 },
		{ 78, 17, 56 },
		{ 78, 18, 56 },
		{ 78, 19, 56 },
		{ 78, 20, 56 },
		{ 78, 21, 56 },
		{ 78, 22, 56 },
		{ 78, 23, 56 },
		{ 78, 24, 56 },
		{ 78, 25, 56 },
		{ 78, 26, 56 },
		{ 78, 27, 56 },
		{ 78, 28, 56 },
		{ 78, 29, 56 },
		{ 78, 30, 56 },
		{ 78, 31, 56 },
		{ 78, 32, 56 },
		{ 78, 33, 56 },
		{ 78, 34, 56 },
		{ 78, 35, 56 },
		{ 78, 36, 56 },
		{ 78, 37, 56 },
		{ 78, 38, 56 },
		{ 78, 39, 56 },
		{ 78, 40, 56 },
		{ 78, 41, 56 },
		{ 78, 42, 56 },
		{ 78, 43, 56 },
		{ 78, 44, 56 },
		{ 78, 45, 56 },
		{ 78, 46, 56 },
		{ 78, 47, 56 },
		{ 78, 48, 56 },
		{ 78, 49, 56 },
		{ 78, 50, 56 },
		{ 78, 51, 56 },
		{ 78, 52, 56 },
		{ 79, 1, 56 },
		{ 79, 2, 56 },
		{ 79, 3, 56 },
		{ 79, 4, 56 },
		{ 79, 5, 56 },
		{ 79, 6, 56 },
		{ 79, 7, 56 },
		{ 79, 8, 56 },
		{ 79, 9, 56 },
		{ 79, 10, 56 },
		{ 79, 11, 56 },
		{ 79, 12, 56 },
		{ 79, 13, 56 },
		{ 79, 14, 56 },
		{ 79, 15, 56 },
		{ 79, 16, 56 },
		{ 79, 17, 56 },
		{ 79, 18, 56 },
		{ 79, 19, 56 },
		{ 79, 20, 56 },
		{ 79, 21, 56 },
		{ 79, 22, 56 },
		{ 79, 23, 56 },
		{ 79, 24, 56 },
		{ 79, 25, 56 },
		{ 79, 26, 56 },
		{ 79, 27, 56 },
		{ 79, 28, 56 },
		{ 79, 29, 56 },
		{ 79, 30, 56 },
		{ 79, 31, 56 },
		{ 79, 32, 56 },
		{ 79, 33, 56 },
		{ 79, 34, 56 },
		{ 79, 35, 56 },
		{ 79, 36, 56 },
		{ 79, 37, 56 },
		{ 79, 38, 56 },
		{ 79, 39, 56 },
		{ 79, 40, 56 },
		{ 79, 41, 56 },
		{ 79, 42, 56 },
		{ 79, 43, 56 },
		{ 79, 44, 56 },
		{ 79, 45, 56 },
		{ 79, 46, 56 },
		{ 79, 47, 56 },
		{ 79, 48, 56 },
		{ 79, 49, 56 },
		{ 79, 50, 56 },
		{ 79, 51, 56 },
		{ 79, 52, 56 },
		{ 79, 53, 56 },
		{ 80, 1, 56 },
		{ 80, 2, 56 },
		{ 80, 3, 56 },
		{ 80, 4, 56 },
		{ 80, 5, 56 },
		{ 80, 6, 56 },
		{ 80, 7, 56 },
		{ 80, 8, 56 },
		{ 80, 9, 56 },
		{ 80, 10, 56 },
		{ 80, 11, 56 },
		{ 80, 12, 56 },
		{ 80, 13, 56 },
		{ 80, 14, 56 },
		{ 80, 15, 56 },
		{ 80, 16, 56 },
		{ 80, 17, 56 },
		{ 80, 18, 56 },
		{ 80, 19, 56 },
		{ 80, 20, 56 },
		{ 80, 21, 56 },
		{ 80, 22, 56 },
		{ 80, 23, 56 },
		{ 80, 24, 56 },
		{ 80, 25, 56 },
		{ 80, 26, 56 },
		{ 80, 27, 56 },
		{ 80, 28, 56 },
		{ 80, 29, 56 },
		{ 80, 30, 56 },
		{ 80, 31, 56 },
		{ 80, 32, 56 },
		{ 80, 33, 56 },
		{ 80, 34, 56 },
		{ 80, 35, 56 },
		{ 80, 36, 56 },
		{ 80, 37, 56 },
		{ 80, 38, 56 },
		{ 80, 39, 56 },
		{ 80, 40, 56 },
		{ 80, 41, 56 },
		{ 80, 42, 56 },
		{ 80, 43, 56 },
		{ 80, 44, 56 },
		{ 80, 45, 56 },
		{ 80, 46, 56 },
		{ 80, 47, 56 },
		{ 80, 48, 56 },
		{ 80, 49, 56 },
		{ 80, 50, 56 },
		{ 80, 51, 56 },
		{ 80, 52, 56 },
		{ 80, 53, 56 },
		{ 80, 54, 56 },
		{ 81, 1, 56 },
		{ 81, 2, 56 },
		{ 81, 3, 56 },
		{ 81, 4, 56 },
		{ 81, 5, 56 },
		{ 81, 6, 56 },
		{ 81, 7, 56 },
		{ 81, 8, 56 },
		{ 81, 9, 56 },
		{ 81, 10, 56 },
		{ 81, 11, 56 },
		{ 81, 12, 56 },
		{ 81, 13, 56 },
		{ 81, 14, 56 },
		{ 81, 15, 56 },
		{ 81, 16, 56 },
		{ 81, 17, 56 },
		{ 81, 18, 56 },
		{ 81, 19, 56 },
		{ 81, 20, 56 },
		{ 81, 21, 56 },
		{ 81, 22, 56 },
		{ 81, 23, 56 },
		{ 81, 24, 56 },
		{ 81, 25, 56 },
		{ 81, 26, 56 },
		{ 81, 27, 56 },
		{ 81, 28, 56 },
		{ 81, 29, 56 },
		{ 81, 30, 56 },
		{ 81, 31, 56 },
		{ 81, 32, 56 },
		{ 81, 33, 56 },
		{ 81, 34, 56 },
		{ 81, 35, 56 },
		{ 81, 36, 56 },
		{ 81, 37, 56 },
		{ 81, 38, 56 },
		{ 81, 39, 56 },
		{ 81, 40, 56 },
		{ 81, 41, 56 },
		{ 81, 42, 56 },
		{ 81, 43, 56 },
		{ 81, 44, 56 },
		{ 81, 45, 56 },
		{ 81, 46, 56 },
		{ 81, 47, 56 },
		{ 81, 48, 56 },
		{ 81, 49, 56 },
		{ 81, 50, 56 },
		{ 81, 51, 56 },
		{ 81, 52, 56 },
		{ 81, 53, 56 },
		{ 81, 54, 56 },
		{ 81, 55, 56 },
		{ 82, 1, 56 },
		{ 82, 2, 56 },
		{ 82, 3, 56 },
		{ 82, 4, 56 },
		{ 82, 5, 56 },
		{ 82, 6, 56 },
		{ 82, 7, 56 },
		{ 82, 8, 56 },
		{ 82, 9, 56 },
		{ 82, 10, 56 },
		{ 82, 11, 56 },
		{ 82, 12, 56 },
		{ 82, 13, 56 },
		{ 82, 14, 56 },
		{ 82, 15, 56 },
		{ 82, 16, 56 },
		{ 82, 17, 56 },
		{ 82, 18, 56 },
		{ 82, 19, 56 },
		{ 82, 20, 56 },
		{ 82, 21, 56 },
		{ 82, 22, 56 },
		{ 82, 23, 56 },
		{ 82, 24, 56 },
		{ 82, 25, 56 },
		{ 82, 26, 56 },
		{ 82, 27, 56 },
		{ 82, 28, 56 },
		{ 82, 29, 56 },
		{ 82, 30, 56 },
		{ 82, 31, 56 },
		{ 82, 32, 56 },
		{ 82, 33, 56 },
		{ 82, 34, 56 },
		{ 82, 35, 56 },
		{ 82, 36, 56 },
		{ 82, 37, 56 },
		{ 82, 38, 56 },
		{ 82, 39, 56 },
		{ 82, 40, 56 },
		{ 82, 41, 56 },
		{ 82, 42, 56 },
		{ 82, 43, 56 },
		{ 82, 44, 56 },
		{ 82, 45, 56 },
		{ 82, 46, 56 },
		{ 82, 47, 56 },
		{ 82, 48, 56 },
		{ 82, 49, 56 },
		{ 82, 50, 56 },
		{ 82, 51, 56 },
		{ 82, 52, 56 },
		{ 82, 53, 56 },
		{ 82, 54, 56 },
		{ 82, 55, 56 },
		{ 82, 56, 56 },
		{ 83, 1, 56 },
		{ 83, 2, 56 },
		{ 83, 3, 56 },
		{ 83, 4, 56 },
		{ 83, 5, 56 },
		{ 83, 6, 56 },
		{ 83, 7, 56 },
		{ 83, 8, 56 },
		{ 83, 9, 56 },
		{ 83, 10, 56 },
		{ 83, 11, 56 },
		{ 83, 12, 56 },
		{ 83, 13, 56 },
		{ 83, 14, 56 },
		{ 83, 15, 56 },
		{ 83, 16, 56 },
		{ 83, 17, 56 },
		{ 83, 18, 56 },
		{ 83, 19, 56 },
		{ 83, 20, 56 },
		{ 83, 21, 56 },
		{ 83, 22, 56 },
		{ 83, 23, 56 },
		{ 83, 24, 56 },
		{ 83, 25, 56 },
		{ 83, 26, 56 },
		{ 83, 27, 56 },
		{ 83, 28, 56 },
		{ 83, 29, 56 },
		{ 83, 30, 56 },
		{ 83, 31, 56 },
		{ 83, 32, 56 },
		{ 83, 33, 56 },
		{ 83, 34, 56 },
		{ 83, 35, 56 },
		{ 83, 36, 56 },
		{ 83, 37, 56 },
		{ 83, 38, 56 },
		{ 83, 39, 56 },
		{ 83, 40, 56 },
		{ 83, 41, 56 },
		{ 83, 42, 56 },
		{ 83, 43, 56 },
		{ 83, 44, 56 },
		{ 83, 45, 56 },
		{ 83, 46, 56 },
		{ 83, 47, 56 },
		{ 83, 48, 56 },
		{ 83, 49, 56 },
		{ 83, 50, 56 },
		{ 83, 51, 56 },
		{ 83, 52, 56 },
		{ 83, 53, 56 },
		{ 83, 54, 56 },
		{ 83, 55, 56 },
		{ 83, 56, 56 },
		{ 83, 57, 56 },
		{ 83, 58, 56 },
		{ 84, 1, 56 },
		{ 84, 2, 56 },
		{ 84, 3, 56 },
		{ 84, 4, 56 },
		{ 84, 5, 56 },
		{ 84, 6, 56 },
		{ 84, 7, 56 },
		{ 84, 8, 56 },
		{ 84, 9, 56 },
		{ 84, 10, 56 },
		{ 84, 11, 56 },
		{ 84, 12, 56 },
		{ 84, 13, 56 },
		{ 84, 14, 56 },
		{ 84, 15, 56 },
		{ 84, 16, 56 },
		{ 84, 17, 56 },
		{ 84, 18, 56 },
		{ 84, 19, 56 },
		{ 84, 20, 56 },
		{ 84, 21, 56 },
		{ 84, 22, 56 },
		{ 84, 23, 56 },
		{ 84, 24, 56 },
		{ 84, 25, 56 },
		{ 84, 26, 56 },
		{ 84, 27, 56 },
		{ 84, 28, 56 },
		{ 84, 29, 56 },
		{ 84, 30, 56 },
		{ 84, 31, 56 },
		{ 84, 32, 56 },
		{ 84, 33, 56 },
		{ 84, 34, 56 },
		{ 84, 35, 56 },
		{ 84, 36, 56 },
		{ 84, 37, 56 },
		{ 84, 38, 56 },
		{ 84, 39, 56 },
		{ 84, 40, 56 },
		{ 84, 41, 56 },
		{ 84, 42, 56 },
		{ 84, 43, 56 },
		{ 84, 44, 56 },
		{ 84, 45, 56 },
		{ 84, 46, 56 },
		{ 84, 47, 56 },
		{ 84, 48, 56 },
		{ 84, 49, 56 },
		{ 84, 50, 56 },
		{ 84, 51, 56 },
		{ 84, 52, 56 },
		{ 84, 53, 56 },
		{ 84, 54, 56 },
		{ 84, 55, 56 },
		{ 84, 56, 56 },
		{ 84, 57, 56 },
		{ 84, 58, 56 },
		{ 84, 59, 56 },
		{ 85, 1, 56 },
		{ 85, 2, 56 },
		{ 85, 3, 56 },
		{ 85, 4, 56 },
		{ 85, 5, 56 },
		{ 85, 6, 56 },
		{ 85, 7, 56 },
		{ 85, 8, 56 },
		{ 85, 9, 56 },
		{ 85, 10, 56 },
		{ 85, 11, 56 },
		{ 85, 12, 56 },
		{ 85, 13, 56 },
		{ 85, 14, 56 },
		{ 85, 15, 56 },
		{ 85, 16, 56 },
		{ 85, 17, 56 },
		{ 85, 18, 56 },
		{ 85, 19, 56 },
		{ 85, 20, 56 },
		{ 85, 21, 56 },
		{ 85, 22, 56 },
		{ 85, 23, 56 },
		{ 85, 24, 56 },
		{ 85, 25, 56 },
		{ 85, 26, 56 },
		{ 85, 27, 56 },
		{ 85, 28, 56 },
		{ 85, 29, 56 },
		{ 85, 30, 56 },
		{ 85, 31, 56 },
		{ 85, 32, 56 },
		{ 85, 33, 56 },
		{ 85, 34, 56 },
		{ 85, 35, 56 },
		{ 85, 36, 56 },
		{ 85, 37, 56 },
		{ 85, 38, 56 },
		{ 85, 39, 56 },
		{ 85, 40, 56 },
		{ 85, 41, 56 },
		{ 85, 42, 56 },
		{ 85, 43, 56 },
		{ 85, 44, 56 },
		{ 85, 45, 56 },
		{ 85, 46, 56 },
		{ 85, 47, 56 },
		{ 85, 48, 56 },
		{ 85, 49, 56 },
		{ 85, 50, 56 },
		{ 85, 51, 56 },
		{ 85, 52, 56 },
		{ 85, 53, 56 },
		{ 85, 54, 56 },
		{ 85, 55, 56 },
		{ 85, 56, 56 },
		{ 85, 57, 56 },
		{ 85, 58, 56 },
		{ 85, 59, 56 },
		{ 85, 60, 56 },
		{ 86, 1, 56 },
		{ 86, 2, 56 },
		{ 86, 3, 56 },
		{ 86, 4, 56 },
		{ 86, 5, 56 },
		{ 86, 6, 56 },
		{ 86, 7, 56 },
		{ 86, 8, 56 },
		{ 86, 9, 56 },
		{ 86, 10, 56 },
		{ 86, 11, 56 },
		{ 86, 12, 56 },
		{ 86, 13, 56 },
		{ 86, 14, 56 },
		{ 86, 15, 56 },
		{ 86, 16, 56 },
		{ 86, 17, 56 },
		{ 86, 18, 56 },
		{ 86, 19, 56 },
		{ 86, 20, 56 },
		{ 86, 21, 56 },
		{ 86, 22, 56 },
		{ 86, 23, 56 },
		{ 86, 24, 56 },
		{ 86, 25, 56 },
		{ 86, 26, 56 },
		{ 86, 27, 56 },
		{ 86, 28, 56 },
		{ 86, 29, 56 },
		{ 86, 30, 56 },
		{ 86, 31, 56 },
		{ 86, 32, 56 },
		{ 86, 33, 56 },
		{ 86, 34, 56 },
		{ 86, 35, 56 },
		{ 86, 36, 56 },
		{ 86, 37, 56 },
		{ 86, 38, 56 },
		{ 86, 39, 56 },
		{ 86, 40, 56 },
		{ 86, 41, 56 },
		{ 86, 42, 56 },
		{ 86, 43, 56 },
		{ 86, 44, 56 },
		{ 86, 45, 56 },
		{ 86, 46, 56 },
		{ 86, 47, 56 },
		{ 86, 48, 56 },
		{ 86, 49, 56 },
		{ 86, 50, 56 },
		{ 86, 51, 56 },
		{ 86, 52, 56 },
		{ 86, 53, 56 },
		{ 86, 54, 56 },
		{ 86, 55, 56 },
		{ 86, 56, 56 },
		{ 86, 57, 56 },
		{ 86, 58, 56 },
		{ 86, 59, 56 },
		{ 86, 60, 56 },
		{ 86, 61, 56 },
		{ 87, 1, 56 },
		{ 87, 2, 56 },
		{ 87, 3, 56 },
		{ 87, 4, 56 },
		{ 87, 5, 56 },
		{ 87, 6, 56 },
		{ 87, 7, 56 },
		{ 87, 8, 56 },
		{ 87, 9, 56 },
		{ 87, 10, 56 },
		{ 87, 11, 56 },
		{ 87, 12, 56 },
		{ 87, 13, 56 },
		{ 87, 14, 56 },
		{ 87, 15, 56 },
		{ 87, 16, 56 },
		{ 87, 17, 56 },
		{ 87, 18, 56 },
		{ 87, 19, 56 },
		{ 87, 20, 56 },
		{ 87, 21, 56 },
		{ 87, 22, 56 },
		{ 87, 23, 56 },
		{ 87, 24, 56 },
		{ 87, 25, 56 },
		{ 87, 26, 56 },
		{ 87, 27, 56 },
		{ 87, 28, 56 },
		{ 87, 29, 56 },
		{ 87, 30, 56 },
		{ 87, 31, 56 },
		{ 87, 32, 56 },
		{ 87, 33, 56 },
		{ 87, 34, 56 },
		{ 87, 35, 56 },
		{ 87, 36, 56 },
		{ 87, 37, 56 },
		{ 87, 38, 56 },
		{ 87, 39, 56 },
		{ 87, 40, 56 },
		{ 87, 41, 56 },
		{ 87, 42, 56 },
		{ 87, 43, 56 },
		{ 87, 44, 56 },
		{ 87, 45, 56 },
		{ 87, 46, 56 },
		{ 87, 47, 56 },
		{ 87, 48, 56 },
		{ 87, 49, 56 },
		{ 87, 50, 56 },
		{ 87, 51, 56 },
		{ 87, 52, 56 },
		{ 87, 53, 56 },
		{ 87, 54, 56 },
		{ 87, 55, 56 },
		{ 87, 56, 56 },
		{ 87, 57, 56 },
		{ 87, 58, 56 },
		{ 87, 59, 56 },
		{ 87, 60, 56 },
		{ 87, 61, 56 },
		{ 87, 62, 56 },
		{ 88, 1, 56 },
		{ 88, 2, 56 },
		{ 88, 3, 56 },
		{ 88, 4, 56 },
		{ 88, 5, 56 },
		{ 88, 6, 56 },
		{ 88, 7, 56 },
		{ 88, 8, 56 },
		{ 88, 9, 56 },
		{ 88, 10, 56 },
		{ 88, 11, 56 },
		{ 88, 12, 56 },
		{ 88, 13, 56 },
		{ 88, 14, 56 },
		{ 88, 15, 56 },
		{ 88, 16, 56 },
		{ 88, 17, 56 },
		{ 88, 18, 56 },
		{ 88, 19, 56 },
		{ 88, 20, 56 },
		{ 88, 21, 56 },
		{ 88, 22, 56 },
		{ 88, 23, 56 },
		{ 88, 24, 56 },
		{ 88, 25, 56 },
		{ 88, 26, 56 },
		{ 88, 27, 56 },
		{ 88, 28, 56 },
		{ 88, 29, 56 },
		{ 88, 30, 56 },
		{ 88, 31, 56 },
		{ 88, 32, 56 },
		{ 88, 33, 56 },
		{ 88, 34, 56 },
		{ 88, 35, 56 },
		{ 88, 36, 56 },
		{ 88, 37, 56 },
		{ 88, 38, 56 },
		{ 88, 39, 56 },
		{ 88, 40, 56 },
		{ 88, 41, 56 },
		{ 88, 42, 56 },
		{ 88, 43, 56 },
		{ 88, 44, 56 },
		{ 88, 45, 56 },
		{ 88, 46, 56 },
		{ 88, 47, 56 },
		{ 88, 48, 56 },
		{ 88, 49, 56 },
		{ 88, 50, 56 },
		{ 88, 51, 56 },
		{ 88, 52, 56 },
		{ 88, 53, 56 },
		{ 88, 54, 56 },
		{ 88, 55, 56 },
		{ 88, 56, 56 },
		{ 88, 57, 56 },
		{ 88, 58, 56 },
		{ 88, 59, 56 },
		{ 88, 60, 56 },
		{ 88, 61, 56 },
		{ 88, 62, 56 },
		{ 88, 63, 56 },
		{ 89, 1, 56 },
		{ 89, 2, 56 },
		{ 89, 3, 56 },
		{ 89, 4, 56 },
		{ 89, 5, 56 },
		{ 89, 6, 56 },
		{ 89, 7, 56 },
		{ 89, 8, 56 },
		{ 89, 9, 56 },
		{ 89, 10, 56 },
		{ 89, 11, 56 },
		{ 89, 12, 56 },
		{ 89, 13, 56 },
		{ 89, 14, 56 },
		{ 89, 15, 56 },
		{ 89, 16, 56 },
		{ 89, 17, 56 },
		{ 89, 18, 56 },
		{ 89, 19, 56 },
		{ 89, 20, 56 },
		{ 89, 21, 56 },
		{ 89, 22, 56 },
		{ 89, 23, 56 },
		{ 89, 24, 56 },
		{ 89, 25, 56 },
		{ 89, 26, 56 },
		{ 89, 27, 56 },
		{ 89, 28, 56 },
		{ 89, 29, 56 },
		{ 89, 30, 56 },
		{ 89, 31, 56 },
		{ 89, 32, 56 },
		{ 89, 33, 56 },
		{ 89, 34, 56 },
		{ 89, 35, 56 },
		{ 89, 36, 56 },
		{ 89, 37, 56 },
		{ 89, 38, 56 },
		{ 89, 39, 56 },
		{ 89, 40, 56 },
		{ 89, 41, 56 },
		{ 89, 42, 56 },
		{ 89, 43, 56 },
		{ 89, 44, 56 },
		{ 89, 45, 56 },
		{ 89, 46, 56 },
		{ 89, 47, 56 },
		{ 89, 48, 56 },
		{ 89, 49, 56 },
		{ 89, 50, 56 },
		{ 89, 51, 56 },
		{ 89, 52, 56 },
		{ 89, 53, 56 },
		{ 89, 54, 56 },
		{ 89, 55, 56 },
		{ 89, 56, 56 },
		{ 89, 57, 56 },
		{ 89, 58, 56 },
		{ 89, 59, 56 },
		{ 89, 60, 56 },
		{ 89, 61, 56 },
		{ 89, 62, 56 },
		{ 89, 63, 56 },
		{ 89, 64, 56 },
		{ 90, 1, 56 },
		{ 90, 2, 56 },
		{ 90, 3, 56 },
		{ 90, 4, 56 },
		{ 90, 5, 56 },
		{ 90, 6, 56 },
		{ 90, 7, 56 },
		{ 90, 8, 56 },
		{ 90, 9, 56 },
		{ 90, 10, 56 },
		{ 90, 11, 56 },
		{ 90, 12, 56 },
		{ 90, 13, 56 },
		{ 90, 14, 56 },
		{ 90, 15, 56 },
		{ 90, 16, 56 },
		{ 90, 17, 56 },
		{ 90, 18, 56 },
		{ 90, 19, 56 },
		{ 90, 20, 56 },
		{ 90, 21, 56 },
		{ 90, 22, 56 },
		{ 90, 23, 56 },
		{ 90, 24, 56 },
		{ 90, 25, 56 },
		{ 90, 26, 56 },
		{ 90, 27, 56 },
		{ 90, 28, 56 },
		{ 90, 29, 56 },
		{ 90, 30, 56 },
		{ 90, 31, 56 },
		{ 90, 32, 56 },
		{ 90, 33, 56 },
		{ 90, 34, 56 },
		{ 90, 35, 56 },
		{ 90, 36, 56 },
		{ 90, 37, 56 },
		{ 90, 38, 56 },
		{ 90, 39, 56 },
		{ 90, 40, 56 },
		{ 90, 41, 56 },
		{ 90, 42, 56 },
		{ 90, 43, 56 },
		{ 90, 44, 56 },
		{ 90, 45, 56 },
		{ 90, 46, 56 },
		{ 90, 47, 56 },
		{ 90, 48, 56 },
		{ 90, 49, 56 },
		{ 90, 50, 56 },
		{ 90, 51, 56 },
		{ 90, 52, 56 },
		{ 90, 53, 56 },
		{ 90, 54, 56 },
		{ 90, 55, 56 },
		{ 90, 56, 56 },
		{ 90, 57, 56 },
		{ 90, 58, 56 },
		{ 90, 59, 56 },
		{ 90, 60, 56 },
		{ 90, 61, 56 },
		{ 90, 62, 56 },
		{ 90, 63, 56 },
		{ 90, 64, 56 },
		{ 90, 65, 56 },
		{ 90, 66, 56 },
		{ 91, 1, 56 },
		{ 91, 2, 56 },
		{ 91, 3, 56 },
		{ 91, 4, 56 },
		{ 91, 5, 56 },
		{ 91, 6, 56 },
		{ 91, 7, 56 },
		{ 91, 8, 56 },
		{ 91, 9, 56 },
		{ 91, 10, 56 },
		{ 91, 11, 56 },
		{ 91, 12, 56 },
		{ 91, 13, 56 },
		{ 91, 14, 56 },
		{ 91, 15, 56 },
		{ 91, 16, 56 },
		{ 91, 17, 56 },
		{ 91, 18, 56 },
		{ 91, 19, 56 },
		{ 91, 20, 56 },
		{ 91, 21, 56 },
		{ 91, 22, 56 },
		{ 91, 23, 56 },
		{ 91, 24, 56 },
		{ 91, 25, 56 },
		{ 91, 26, 56 },
		{ 91, 27, 56 },
		{ 91, 28, 56 },
		{ 91, 29, 56 },
		{ 91, 30, 56 },
		{ 91, 31, 56 },
		{ 91, 32, 56 },
		{ 91, 33, 56 },
		{ 91, 34, 56 },
		{ 91, 35, 56 },
		{ 91, 36, 56 },
		{ 91, 37, 56 },
		{ 91, 38, 56 },
		{ 91, 39, 56 },
		{ 91, 40, 56 },
		{ 91, 41, 56 },
		{ 91, 42, 56 },
		{ 91, 43, 56 },
		{ 91, 44, 56 },
		{ 91, 45, 56 },
		{ 91, 46, 56 },
		{ 91, 47, 56 },
		{ 91, 48, 56 },
		{ 91, 49, 56 },
		{ 91, 50, 56 },
		{ 91, 51, 56 },
		{ 91, 52, 56 },
		{ 91, 53, 56 },
		{ 91, 54, 56 },
		{ 91, 55, 56 },
		{ 91, 56, 56 },
		{ 91, 57, 56 },
		{ 91, 58, 56 },
		{ 91, 59, 56 },
		{ 91, 60, 56 },
		{ 91, 61, 56 },
		{ 91, 62, 56 },
		{ 91, 63, 56 },
		{ 91, 64, 56 },
		{ 91, 65, 56 },
		{ 91, 66, 56 },
		{ 91, 67, 56 },
		{ 92, 1, 56 },
		{ 92, 2, 56 },
		{ 92, 3, 56 },
		{ 92, 4, 56 },
		{ 92, 5, 56 },
		{ 92, 6, 56 },
		{ 92, 7, 56 },
		{ 92, 8, 56 },
		{ 92, 9, 56 },
		{ 92, 10, 56 },
		{ 92, 11, 56 },
		{ 92, 12, 56 },
		{ 92, 13, 56 },
		{ 92, 14, 56 },
		{ 92, 15, 56 },
		{ 92, 16, 56 },
		{ 92, 17, 56 },
		{ 92, 18, 56 },
		{ 92, 19, 56 },
		{ 92, 20, 56 },
		{ 92, 21, 56 },
		{ 92, 22, 56 },
		{ 92, 23, 56 },
		{ 92, 24, 56 },
		{ 92, 25, 56 },
		{ 92, 26, 56 },
		{ 92, 27, 56 },
		{ 92, 28, 56 },
		{ 92, 29, 56 },
		{ 92, 30, 56 },
		{ 92, 31, 56 },
		{ 92, 32, 56 },
		{ 92, 33, 56 },
		{ 92, 34, 56 },
		{ 92, 35, 56 },
		{ 92, 36, 56 },
		{ 92, 37, 56 },
		{ 92, 38, 56 },
		{ 92, 39, 56 },
		{ 92, 40, 56 },
		{ 92, 41, 56 },
		{ 92, 42, 56 },
		{ 92, 43, 56 },
		{ 92, 44, 56 },
		{ 92, 45, 56 },
		{ 92, 46, 56 },
		{ 92, 47, 56 },
		{ 92, 48, 56 },
		{ 92, 49, 56 },
		{ 92, 50, 56 },
		{ 92, 51, 56 },
		{ 92, 52, 56 },
		{ 92, 53, 56 },
		{ 92, 54, 56 },
		{ 92, 55, 56 },
		{ 92, 56, 56 },
		{ 92, 57, 56 },
		{ 92, 58, 56 },
		{ 92, 59, 56 },
		{ 92, 60, 56 },
		{ 92, 61, 56 },
		{ 92, 62, 56 },
		{ 92, 63, 56 },
		{ 92, 64, 56 },
		{ 92, 65, 56 },
		{ 92, 66, 56 },
		{ 92, 67, 56 },
		{ 92, 68, 56 },
		{ 93, 1, 56 },
		{ 93, 2, 56 },
		{ 93, 3, 56 },
		{ 93, 4, 56 },
		{ 93, 5, 56 },
		{ 93, 6, 56 },
		{ 93, 7, 56 },
		{ 93, 8, 56 },
		{ 93, 9, 56 },
		{ 93, 10, 56 },
		{ 93, 11, 56 },
		{ 93, 12, 56 },
		{ 93, 13, 56 },
		{ 93, 14, 56 },
		{ 93, 15, 56 },
		{ 93, 16, 56 },
		{ 93, 17, 56 },
		{ 93, 18, 56 },
		{ 93, 19, 56 },
		{ 93, 20, 56 },
		{ 93, 21, 56 },
		{ 93, 22, 56 },
		{ 93, 23, 56 },
		{ 93, 24, 56 },
		{ 93, 25, 56 },
		{ 93, 26, 56 },
		{ 93, 27, 56 },
		{ 93, 28, 56 },
		{ 93, 29, 56 },
		{ 93, 30, 56 },
		{ 93, 31, 56 },
		{ 93, 32, 56 },
		{ 93, 33, 56 },
		{ 93, 34, 56 },
		{ 93, 35, 56 },
		{ 93, 36, 56 },
		{ 93, 37, 56 },
		{ 93, 38, 56 },
		{ 93, 39, 56 },
		{ 93, 40, 56 },
		{ 93, 41, 56 },
		{ 93, 42, 56 },
		{ 93, 43, 56 },
		{ 93, 44, 56 },
		{ 93, 45, 56 },
		{ 93, 46, 56 },
		{ 93, 47, 56 },
		{ 93, 48, 56 },
		{ 93, 49, 56 },
		{ 93, 50, 56 },
		{ 93, 51, 56 },
		{ 93, 52, 56 },
		{ 93, 53, 56 },
		{ 93, 54, 56 },
		{ 93, 55, 56 },
		{ 93, 56, 56 },
		{ 93, 57, 56 },
		{ 93, 58, 56 },
		{ 93, 59, 56 },
		{ 93, 60, 56 },
		{ 93, 61, 56 },
		{ 93, 62, 56 },
		{ 93, 63, 56 },
		{ 93, 64, 56 },
		{ 93, 65, 56 },
		{ 93, 66, 56 },
		{ 93, 67, 56 },
		{ 93, 68, 56 },
		{ 93, 69, 56 },
		{ 94, 1, 56 },
		{ 94, 2, 56 },
		{ 94, 3, 56 },
		{ 94, 4, 56 },
		{ 94, 5, 56 },
		{ 94, 6, 56 },
		{ 94, 7, 56 },
		{ 94, 8, 56 },
		{ 94, 9, 56 },
		{ 94, 10, 56 },
		{ 94, 11, 56 },
		{ 94, 12, 56 },
		{ 94, 13, 56 },
		{ 94, 14, 56 },
		{ 94, 15, 56 },
		{ 94, 16, 56 },
		{ 94, 17, 56 },
		{ 94, 18, 56 },
		{ 94, 19, 56 },
		{ 94, 20, 56 },
		{ 94, 21, 56 },
		{ 94, 22, 56 },
		{ 94, 23, 56 },
		{ 94, 24, 56 },
		{ 94, 25, 56 },
		{ 94, 26, 56 },
		{ 94, 27, 56 },
		{ 94, 28, 56 },
		{ 94, 29, 56 },
		{ 94, 30, 56 },
		{ 94, 31, 56 },
		{ 94, 32, 56 },
		{ 94, 33, 56 },
		{ 94, 34, 56 },
		{ 94, 35, 56 },
		{ 94, 36, 56 },
		{ 94, 37, 56 },
		{ 94, 38, 56 },
		{ 94, 39, 56 },
		{ 94, 40, 56 },
		{ 94, 41, 56 },
		{ 94, 42, 56 },
		{ 94, 43, 56 },
		{ 94, 44, 56 },
		{ 94, 45, 56 },
		{ 94, 46, 56 },
		{ 94, 47, 56 },
		{ 94, 48, 56 },
		{ 94, 49, 56 },
		{ 94, 50, 56 },
		{ 94, 51, 56 },
		{ 94, 52, 56 },
		{ 94, 53, 56 },
		{ 94, 54, 56 },
		{ 94, 55, 56 },
		{ 94, 56, 56 },
		{ 94, 57, 56 },
		{ 94, 58, 56 },
		{ 94, 59, 56 },
		{ 94, 60, 56 },
		{ 94, 61, 56 },
		{ 94, 62, 56 },
		{ 94, 63, 56 },
		{ 94, 64, 56 },
		{ 94, 65, 56 },
		{ 94, 66, 56 },
		{ 94, 67, 56 },
		{ 94, 68, 56 },
		{ 94, 69, 56 },
		{ 94, 70, 56 },
		{ 95, 1, 56 },
		{ 95, 2, 56 },
		{ 95, 3, 56 },
		{ 95, 4, 56 },
		{ 95, 5, 56 },
		{ 95, 6, 56 },
		{ 95, 7, 56 },
		{ 95, 8, 56 },
		{ 95, 9, 56 },
		{ 95, 10, 56 },
		{ 95, 11, 56 },
		{ 95, 12, 56 },
		{ 95, 13, 56 },
		{ 95, 14, 56 },
		{ 95, 15, 56 },
		{ 95, 16, 56 },
		{ 95, 17, 56 },
		{ 95, 18, 56 },
		{ 95, 19, 56 },
		{ 95, 20, 56 },
		{ 95, 21, 56 },
		{ 95, 22, 56 },
		{ 95, 23, 56 },
		{ 95, 24, 56 },
		{ 95, 25, 56 },
		{ 95, 26, 56 },
		{ 95, 27, 56 },
		{ 95, 28, 56 },
		{ 95, 29, 56 },
		{ 95, 30, 56 },
		{ 95, 31, 56 },
		{ 95, 32, 56 },
		{ 95, 33, 56 },
		{ 95, 34, 56 },
		{ 95, 35, 56 },
		{ 95, 36, 56 },
		{ 95, 37, 56 },
		{ 95, 38, 56 },
		{ 95, 39, 56 },
		{ 95, 40, 56 },
		{ 95, 41, 56 },
		{ 95, 42, 56 },
		{ 95, 43, 56 },
		{ 95, 44, 56 },
		{ 95, 45, 56 },
		{ 95, 46, 56 },
		{ 95, 47, 56 },
		{ 95, 48, 56 },
		{ 95, 49, 56 },
		{ 95, 50, 56 },
		{ 95, 51, 56 },
		{ 95, 52, 56 },
		{ 95, 53, 56 },
		{ 95, 54, 56 },
		{ 95, 55, 56 },
		{ 95, 56, 56 },
		{ 95, 57, 56 },
		{ 95, 58, 56 },
		{ 95, 59, 56 },
		{ 95, 60, 56 },
		{ 95, 61, 56 },
		{ 95, 62, 56 },
		{ 95, 63, 56 },
		{ 95, 64, 56 },
		{ 95, 65, 56 },
		{ 95, 66, 56 },
		{ 95, 67, 56 },
		{ 95, 68, 56 },
		{ 95, 69, 56 },
		{ 95, 70, 56 },
		{ 95, 71, 56 },
		{ 96, 1, 56 },
		{ 96, 2, 56 },
		{ 96, 3, 56 },
		{ 96, 4, 56 },
		{ 96, 5, 56 },
		{ 96, 6, 56 },
		{ 96, 7, 56 },
		{ 96, 8, 56 },
		{ 96, 9, 56 },
		{ 96, 10, 56 },
		{ 96, 11, 56 },
		{ 96, 12, 56 },
		{ 96, 13, 56 },
		{ 96, 14, 56 },
		{ 96, 15, 56 },
		{ 96, 16, 56 },
		{ 96, 17, 56 },
		{ 96, 18, 56 },
		{ 96, 19, 56 },
		{ 96, 20, 56 },
		{ 96, 21, 56 },
		{ 96, 22, 56 },
		{ 96, 23, 56 },
		{ 96, 24, 56 },
		{ 96, 25, 56 },
		{ 96, 26, 56 },
		{ 96, 27, 56 },
		{ 96, 28, 56 },
		{ 96, 29, 56 },
		{ 96, 30, 56 },
		{ 96, 31, 56 },
		{ 96, 32, 56 },
		{ 96, 33, 56 },
		{ 96, 34, 56 },
		{ 96, 35, 56 },
		{ 96, 36, 56 },
		{ 96, 37, 56 },
		{ 96, 38, 56 },
		{ 96, 39, 56 },
		{ 96, 40, 56 },
		{ 96, 41, 56 },
		{ 96, 42, 56 },
		{ 96, 43, 56 },
		{ 96, 44, 56 },
		{ 96, 45, 56 },
		{ 96, 46, 56 },
		{ 96, 47, 56 },
		{ 96, 48, 56 },
		{ 96, 49, 56 },
		{ 96, 50, 56 },
		{ 96, 51, 56 },
		{ 96, 52, 56 },
		{ 96, 53, 56 },
		{ 96, 54, 56 },
		{ 96, 55, 56 },
		{ 96, 56, 56 },
		{ 96, 57, 56 },
		{ 96, 58, 56 },
		{ 96, 59, 56 },
		{ 96, 60, 56 },
		{ 96, 61, 56 },
		{ 96, 62, 56 },
		{ 96, 63, 56 },
		{ 96, 64, 56 },
		{ 96, 65, 56 },
		{ 96, 66, 56 },
		{ 96, 67, 56 },
		{ 96, 68, 56 },
		{ 96, 69, 56 },
		{ 96, 70, 56 },
		{ 96, 71, 56 },
		{ 96, 72, 56 },
		{ 97, 2, 56 },
		{ 97, 3, 56 },
		{ 97, 4, 56 },
		{ 97, 5, 56 },
		{ 97, 6, 56 },
		{ 97, 7, 56 },
		{ 97, 8, 56 },
		{ 97, 9, 56 },
		{ 97, 10, 56 },
		{ 97, 11, 56 },
		{ 97, 12, 56 },
		{ 97, 13, 56 },
		{ 97, 14, 56 },
		{ 97, 15, 56 },
		{ 97, 16, 56 },
		{ 97, 17, 56 },
		{ 97, 18, 56 },
		{ 97, 19, 56 },
		{ 97, 20, 56 },
		{ 97, 21, 56 },
		{ 97, 22, 56 },
		{ 97, 23, 56 },
		{ 97, 24, 56 },
		{ 97, 25, 56 },
		{ 97, 26, 56 },
		{ 97, 27, 56 },
		{ 97, 28, 56 },
		{ 97, 29, 56 },
		{ 97, 30, 56 },
		{ 97, 31, 56 },
		{ 97, 32, 56 },
		{ 97, 33, 56 },
		{ 97, 34, 56 },
		{ 97, 35, 56 },
		{ 97, 36, 56 },
		{ 97, 37, 56 },
		{ 97, 38, 56 },
		{ 97, 39, 56 },
		{ 97, 40, 56 },
		{ 97, 41, 56 },
		{ 97, 42, 56 },
		{ 97, 43, 56 },
		{ 97, 44, 56 },
		{ 97, 45, 56 },
		{ 97, 46, 56 },
		{ 97, 47, 56 },
		{ 97, 48, 56 },
		{ 97, 49, 56 },
		{ 97, 50, 56 },
		{ 97, 51, 56 },
		{ 97, 52, 56 },
		{ 97, 53, 56 },
		{ 97, 54, 56 },
		{ 97, 55, 56 },
		{ 97, 56, 56 },
		{ 97, 57, 56 },
		{ 97, 58, 56 },
		{ 97, 59, 56 },
		{ 97, 60, 56 },
		{ 97, 61, 56 },
		{ 97, 62, 56 },
		{ 97, 63, 56 },
		{ 97, 64, 56 },
		{ 97, 65, 56 },
		{ 97, 66, 56 },
		{ 97, 67, 56 },
		{ 97, 68, 56 },
		{ 97, 69, 56 },
		{ 97, 70, 56 },
		{ 97, 71, 56 },
		{ 97, 72, 56 },
		{ 97, 73, 56 },
		{ 97, 74, 56 },
		{ 98, 3, 56 },
		{ 98, 4, 56 },
		{ 98, 5, 56 },
		{ 98, 6, 56 },
		{ 98, 7, 56 },
		{ 98, 8, 56 },
		{ 98, 9, 56 },
		{ 98, 10, 56 },
		{ 98, 11, 56 },
		{ 98, 12, 56 },
		{ 98, 13, 56 },
		{ 98, 14, 56 },
		{ 98, 15, 56 },
		{ 98, 16, 56 },
		{ 98, 17, 56 },
		{ 98, 18, 56 },
		{ 98, 19, 56 },
		{ 98, 20, 56 },
		{ 98, 21, 56 },
		{ 98, 22, 56 },
		{ 98, 23, 56 },
		{ 98, 24, 56 },
		{ 98, 25, 56 },
		{ 98, 26, 56 },
		{ 98, 27, 56 },
		{ 98, 28, 56 },
		{ 98, 29, 56 },
		{ 98, 30, 56 },
		{ 98, 31, 56 },
		{ 98, 32, 56 },
		{ 98, 33, 56 },
		{ 98, 34, 56 },
		{ 98, 35, 56 },
		{ 98, 36, 56 },
		{ 98, 37, 56 },
		{ 98, 38, 56 },
		{ 98, 39, 56 },
		{ 98, 40, 56 },
		{ 98, 41, 56 },
		{ 98, 42, 56 },
		{ 98, 43, 56 },
		{ 98, 44, 56 },
		{ 98, 45, 56 },
		{ 98, 46, 56 },
		{ 98, 47, 56 },
		{ 98, 48, 56 },
		{ 98, 49, 56 },
		{ 98, 50, 56 },
		{ 98, 51, 56 },
		{ 98, 52, 56 },
		{ 98, 53, 56 },
		{ 98, 54, 56 },
		{ 98, 55, 56 },
		{ 98, 56, 56 },
		{ 98, 57, 56 },
		{ 98, 58, 56 },
		{ 98, 59, 56 },
		{ 98, 60, 56 },
		{ 98, 61, 56 },
		{ 98, 62, 56 },
		{ 98, 63, 56 },
		{ 98, 64, 56 },
		{ 98, 65, 56 },
		{ 98, 66, 56 },
		{ 98, 67, 56 },
		{ 98, 68, 56 },
		{ 98, 69, 56 },
		{ 98, 70, 56 },
		{ 98, 71, 56 },
		{ 98, 72, 56 },
		{ 98, 73, 56 },
		{ 98, 74, 56 },
		{ 98, 75, 56 },
		{ 99, 5, 56 },
		{ 99, 6, 56 },
		{ 99, 7, 56 },
		{ 99, 8, 56 },
		{ 99, 9, 56 },
		{ 99, 10, 56 },
		{ 99, 11, 56 },
		{ 99, 12, 56 },
		{ 99, 13, 56 },
		{ 99, 14, 56 },
		{ 99, 15, 56 },
		{ 99, 16, 56 },
		{ 99, 17, 56 },
		{ 99, 18, 56 },
		{ 99, 19, 56 },
		{ 99, 20, 56 },
		{ 99, 21, 56 },
		{ 99, 22, 56 },
		{ 99, 23, 56 },
		{ 99, 24, 56 },
		{ 99, 25, 56 },
		{ 99, 26, 56 },
		{ 99, 27, 56 },
		{ 99, 28, 56 },
		{ 99, 29, 56 },
		{ 99, 30, 56 },
		{ 99, 31, 56 },
		{ 99, 32, 56 },
		{ 99, 33, 56 },
		{ 99, 34, 56 },
		{ 99, 35, 56 },
		{ 99, 36, 56 },
		{ 99, 37, 56 },
		{ 99, 38, 56 },
		{ 99, 39, 56 },
		{ 99, 40, 56 },
		{ 99, 41, 56 },
		{ 99, 42, 56 },
		{ 99, 43, 56 },
		{ 99, 44, 56 },
		{ 99, 45, 56 },
		{ 99, 46, 56 },
		{ 99, 47, 56 },
		{ 99, 48, 56 },
		{ 99, 49, 56 },
		{ 99, 50, 56 },
		{ 99, 51, 56 },
		{ 99, 52, 56 },
		{ 99, 53, 56 },
		{ 99, 54, 56 },
		{ 99, 55, 56 },
		{ 99, 56, 56 },
		{ 99, 57, 56 },
		{ 99, 58, 56 },
		{ 99, 59, 56 },
		{ 99, 60, 56 },
		{ 99, 61, 56 },
		{ 99, 62, 56 },
		{ 99, 63, 56 },
		{ 99, 64, 56 },
		{ 99, 65, 56 },
		{ 99, 66, 56 },
		{ 99, 67, 56 },
		{ 99, 68, 56 },
		{ 99, 69, 56 },
		{ 99, 70, 56 },
		{ 99, 71, 56 },
		{ 99, 72, 56 },
		{ 99, 73, 56 },
		{ 99, 74, 56 },
		{ 99, 75, 56 },
		{ 99, 76, 56 },
		{ 100, 6, 56 },
		{ 100, 7, 56 },
		{ 100, 8, 56 },
		{ 100, 9, 56 },
		{ 100, 10, 56 },
		{ 100, 11, 56 },
		{ 100, 12, 56 },
		{ 100, 13, 56 },
		{ 100, 14, 56 },
		{ 100, 15, 56 },
		{ 100, 16, 56 },
		{ 100, 17, 56 },
		{ 100, 18, 56 },
		{ 100, 19, 56 },
		{ 100, 20, 56 },
		{ 100, 21, 56 },
		{ 100, 22, 56 },
		{ 100, 23, 56 },
		{ 100, 24, 56 },
		{ 100, 25, 56 },
		{ 100, 26, 56 },
		{ 100, 27, 56 },
		{ 100, 28, 56 },
		{ 100, 29, 56 },
		{ 100, 30, 56 },
		{ 100, 31, 56 },
		{ 100, 32, 56 },
		{ 100, 33, 56 },
		{ 100, 34, 56 },
		{ 100, 35, 56 },
		{ 100, 36, 56 },
		{ 100, 37, 56 },
		{ 100, 38, 56 },
		{ 100, 39, 56 },
		{ 100, 40, 56 },
		{ 100, 41, 56 },
		{ 100, 42, 56 },
		{ 100, 43, 56 },
		{ 100, 44, 56 },
		{ 100, 45, 56 },
		{ 100, 46, 56 },
		{ 100, 47, 56 },
		{ 100, 48, 56 },
		{ 100, 49, 56 },
		{ 100, 50, 56 },
		{ 100, 51, 56 },
		{ 100, 52, 56 },
		{ 100, 53, 56 },
		{ 100, 54, 56 },
		{ 100, 55, 56 },
		{ 100, 56, 56 },
		{ 100, 57, 56 },
		{ 100, 58, 56 },
		{ 100, 59, 56 },
		{ 100, 60, 56 },
		{ 100, 61, 56 },
		{ 100, 62, 56 },
		{ 100, 63, 56 },
		{ 100, 64, 56 },
		{ 100, 65, 56 },
		{ 100, 66, 56 },
		{ 100, 67, 56 },
		{ 100, 68, 56 },
		{ 100, 69, 56 },
		{ 100, 70, 56 },
		{ 100, 71, 56 },
		{ 100, 72, 56 },
		{ 100, 73, 56 },
		{ 100, 74, 56 },
		{ 100, 75, 56 },
		{ 100, 76, 56 },
		{ 100, 77, 56 },
		{ 101, 8, 56 },
		{ 101, 9, 56 },
		{ 101, 10, 56 },
		{ 101, 11, 56 },
		{ 101, 12, 56 },
		{ 101, 13, 56 },
		{ 101, 14, 56 },
		{ 101, 15, 56 },
		{ 101, 16, 56 },
		{ 101, 17, 56 },
		{ 101, 18, 56 },
		{ 101, 19, 56 },
		{ 101, 20, 56 },
		{ 101, 21, 56 },
		{ 101, 22, 56 },
		{ 101, 23, 56 },
		{ 101, 24, 56 },
		{ 101, 25, 56 },
		{ 101, 26, 56 },
		{ 101, 27, 56 },
		{ 101, 28, 56 },
		{ 101, 29, 56 },
		{ 101, 30, 56 },
		{ 101, 31, 56 },
		{ 101, 32, 56 },
		{ 101, 33, 56 },
		{ 101, 34, 56 },
		{ 101, 35, 56 },
		{ 101, 36, 56 },
		{ 101, 37, 56 },
		{ 101, 38, 56 },
		{ 101, 39, 56 },
		{ 101, 40, 56 },
		{ 101, 41, 56 },
		{ 101, 42, 56 },
		{ 101, 43, 56 },
		{ 101, 44, 56 },
		{ 101, 45, 56 },
		{ 101, 46, 56 },
		{ 101, 47, 56 },
		{ 101, 48, 56 },
		{ 101, 49, 56 },
		{ 101, 50, 56 },
		{ 101, 51, 56 },
		{ 101, 52, 56 },
		{ 101, 53, 56 },
		{ 101, 54, 56 },
		{ 101, 55, 56 },
		{ 101, 56, 56 },
		{ 101, 57, 56 },
		{ 101, 58, 56 },
		{ 101, 59, 56 },
		{ 101, 60, 56 },
		{ 101, 61, 56 },
		{ 101, 62, 56 },
		{ 101, 63, 56 },
		{ 101, 64, 56 },
		{ 101, 65, 56 },
		{ 101, 66, 56 },
		{ 101, 67, 56 },
		{ 101, 68, 56 },
		{ 101, 69, 56 },
		{ 101, 70, 56 },
		{ 101, 71, 56 },
		{ 101, 72, 56 },
		{ 101, 73, 56 },
		{ 101, 74, 56 },
		{ 101, 75, 56 },
		{ 101, 76, 56 },
		{ 101, 77, 56 },
		{ 101, 78, 56 },
		{ 102, 10, 56 },
		{ 102, 11, 56 },
		{ 102, 12, 56 },
		{ 102, 13, 56 },
		{ 102, 14, 56 },
		{ 102, 15, 56 },
		{ 102, 16, 56 },
		{ 102, 17, 56 },
		{ 102, 18, 56 },
		{ 102, 19, 56 },
		{ 102, 20, 56 },
		{ 102, 21, 56 },
		{ 102, 22, 56 },
		{ 102, 23, 56 },
		{ 102, 24, 56 },
		{ 102, 25, 56 },
		{ 102, 26, 56 },
		{ 102, 27, 56 },
		{ 102, 28, 56 },
		{ 102, 29, 56 },
		{ 102, 30, 56 },
		{ 102, 31, 56 },
		{ 102, 32, 56 },
		{ 102, 33, 56 },
		{ 102, 34, 56 },
		{ 102, 35, 56 },
		{ 102, 36, 56 },
		{ 102, 37, 56 },
		{ 102, 38, 56 },
		{ 102, 39, 56 },
		{ 102, 40, 56 },
		{ 102, 41, 56 },
		{ 102, 42, 56 },
		{ 102, 43, 56 },
		{ 102, 44, 56 },
		{ 102, 45, 56 },
		{ 102, 46, 56 },
		{ 102, 47, 56 },
		{ 102, 48, 56 },
		{ 102, 49, 56 },
		{ 102, 50, 56 },
		{ 102, 51, 56 },
		{ 102, 52, 56 },
		{ 102, 53, 56 },
		{ 102, 54, 56 },
		{ 102, 55, 56 },
		{ 102, 56, 56 },
		{ 102, 57, 56 },
		{ 102, 58, 56 },
		{ 102, 59, 56 },
		{ 102, 60, 56 },
		{ 102, 61, 56 },
		{ 102, 62, 56 },
		{ 102, 63, 56 },
		{ 102, 64, 56 },
		{ 102, 65, 56 },
		{ 102, 66, 56 },
		{ 102, 67, 56 },
		{ 102, 68, 56 },
		{ 102, 69, 56 },
		{ 102, 70, 56 },
		{ 102, 71, 56 },
		{ 102, 72, 56 },
		{ 102, 73, 56 },
		{ 102, 74, 56 },
		{ 102, 75, 56 },
		{ 102, 76, 56 },
		{ 102, 77, 56 },
		{ 102, 78, 56 },
		{ 102, 79, 56 },
		{ 103, 11, 56 },
		{ 103, 12, 56 },
		{ 103, 13, 56 },
		{ 103, 14, 56 },
		{ 103, 15, 56 },
		{ 103, 16, 56 },
		{ 103, 17, 56 },
		{ 103, 18, 56 },
		{ 103, 19, 56 },
		{ 103, 20, 56 },
		{ 103, 21, 56 },
		{ 103, 22, 56 },
		{ 103, 23, 56 },
		{ 103, 24, 56 },
		{ 103, 25, 56 },
		{ 103, 26, 56 },
		{ 103, 27, 56 },
		{ 103, 28, 56 },
		{ 103, 29, 56 },
		{ 103, 30, 56 },
		{ 103, 31, 56 },
		{ 103, 32, 56 },
		{ 103, 33, 56 },
		{ 103, 34, 56 },
		{ 103, 35, 56 },
		{ 103, 36, 56 },
		{ 103, 37, 56 },
		{ 103, 38, 56 },
		{ 103, 39, 56 },
		{ 103, 40, 56 },
		{ 103, 41, 56 },
		{ 103, 42, 56 },
		{ 103, 43, 56 },
		{ 103, 44, 56 },
		{ 103, 45, 56 },
		{ 103, 46, 56 },
		{ 103, 47, 56 },
		{ 103, 48, 56 },
		{ 103, 49, 56 },
		{ 103, 50, 56 },
		{ 103, 51, 56 },
		{ 103, 52, 56 },
		{ 103, 53, 56 },
		{ 103, 54, 56 },
		{ 103, 55, 56 },
		{ 103, 56, 56 },
		{ 103, 57, 56 },
		{ 103, 58, 56 },
		{ 103, 59, 56 },
		{ 103, 60, 56 },
		{ 103, 61, 56 },
		{ 103, 62, 56 },
		{ 103, 63, 56 },
		{ 103, 64, 56 },
		{ 103, 65, 56 },
		{ 103, 66, 56 },
		{ 103, 67, 56 },
		{ 103, 68, 56 },
		{ 103, 69, 56 },
		{ 103, 70, 56 },
		{ 103, 71, 56 },
		{ 103, 72, 56 },
		{ 103, 73, 56 },
		{ 103, 74, 56 },
		{ 103, 75, 56 },
		{ 103, 76, 56 },
		{ 103, 77, 56 },
		{ 103, 78, 56 },
		{ 103, 79, 56 },
		{ 103, 80, 56 },
		{ 104, 13, 56 },
		{ 104, 14, 56 },
		{ 104, 15, 56 },
		{ 104, 16, 56 },
		{ 104, 17, 56 },
		{ 104, 18, 56 },
		{ 104, 19, 56 },
		{ 104, 20, 56 },
		{ 104, 21, 56 },
		{ 104, 22, 56 },
		{ 104, 23, 56 },
		{ 104, 24, 56 },
		{ 104, 25, 56 },
		{ 104, 26, 56 },
		{ 104, 27, 56 },
		{ 104, 28, 56 },
		{ 104, 29, 56 },
		{ 104, 30, 56 },
		{ 104, 31, 56 },
		{ 104, 32, 56 },
		{ 104, 33, 56 },
		{ 104, 34, 56 },
		{ 104, 35, 56 },
		{ 104, 36, 56 },
		{ 104, 37, 56 },
		{ 104, 38, 56 },
		{ 104, 39, 56 },
		{ 104, 40, 56 },
		{ 104, 41, 56 },
		{ 104, 42, 56 },
		{ 104, 43, 56 },
		{ 104, 44, 56 },
		{ 104, 45, 56 },
		{ 104, 46, 56 },
		{ 104, 47, 56 },
		{ 104, 48, 56 },
		{ 104, 49, 56 },
		{ 104, 50, 56 },
		{ 104, 51, 56 },
		{ 104, 52, 56 },
		{ 104, 53, 56 },
		{ 104, 54, 56 },
		{ 104, 55, 56 },
		{ 104, 56, 56 },
		{ 104, 57, 56 },
		{ 104, 58, 56 },
		{ 104, 59, 56 },
		{ 104, 60, 56 },
		{ 104, 61, 56 },
		{ 104, 62, 56 },
		{ 104, 63, 56 },
		{ 104, 64, 56 },
		{ 104, 65, 56 },
		{ 104, 66, 56 },
		{ 104, 67, 56 },
		{ 104, 68, 56 },
		{ 104, 69, 56 },
		{ 104, 70, 56 },
		{ 104, 71, 56 },
		{ 104, 72, 56 },
		{ 104, 73, 56 },
		{ 104, 74, 56 },
		{ 104, 75, 56 },
		{ 104, 76, 56 },
		{ 104, 77, 56 },
		{ 104, 78, 56 },
		{ 104, 79, 56 },
		{ 104, 80, 56 },
		{ 104, 81, 56 },
		{ 104, 82, 56 },
		{ 105, 14, 56 },
		{ 105, 15, 56 },
		{ 105, 16, 56 },
		{ 105, 17, 56 },
		{ 105, 18, 56 },
		{ 105, 19, 56 },
		{ 105, 20, 56 },
		{ 105, 21, 56 },
		{ 105, 22, 56 },
		{ 105, 23, 56 },
		{ 105, 24, 56 },
		{ 105, 25, 56 },
		{ 105, 26, 56 },
		{ 105, 27, 56 },
		{ 105, 28, 56 },
		{ 105, 29, 56 },
		{ 105, 30, 56 },
		{ 105, 31, 56 },
		{ 105, 32, 56 },
		{ 105, 33, 56 },
		{ 105, 34, 56 },
		{ 105, 35, 56 },
		{ 105, 36, 56 },
		{ 105, 37, 56 },
		{ 105, 38, 56 },
		{ 105, 39, 56 },
		{ 105, 40, 56 },
		{ 105, 41, 56 },
		{ 105, 42, 56 },
		{ 105, 43, 56 },
		{ 105, 44, 56 },
		{ 105, 45, 56 },
		{ 105, 46, 56 },
		{ 105, 47, 56 },
		{ 105, 48, 56 },
		{ 105, 49, 56 },
		{ 105, 50, 56 },
		{ 105, 51, 56 },
		{ 105, 52, 56 },
		{ 105, 53, 56 },
		{ 105, 54, 56 },
		{ 105, 55, 56 },
		{ 105, 56, 56 },
		{ 105, 57, 56 },
		{ 105, 58, 56 },
		{ 105, 59, 56 },
		{ 105, 60, 56 },
		{ 105, 61, 56 },
		{ 105, 62, 56 },
		{ 105, 63, 56 },
		{ 105, 64, 56 },
		{ 105, 65, 56 },
		{ 105, 66, 56 },
		{ 105, 67, 56 },
		{ 105, 68, 56 },
		{ 105, 69, 56 },
		{ 105, 70, 56 },
		{ 105, 71, 56 },
		{ 105, 72, 56 },
		{ 105, 73, 56 },
		{ 105, 74, 56 },
		{ 105, 75, 56 },
		{ 105, 76, 56 },
		{ 105, 77, 56 },
		{ 105, 78, 56 },
		{ 105, 79, 56 },
		{ 105, 80, 56 },
		{ 105, 81, 56 },
		{ 105, 82, 56 },
		{ 105, 83, 56 },
		{ 106, 16, 56 },
		{ 106, 17, 56 },
		{ 106, 18, 56 },
		{ 106, 19, 56 },
		{ 106, 20, 56 },
		{ 106, 21, 56 },
		{ 106, 22, 56 },
		{ 106, 23, 56 },
		{ 106, 24, 56 },
		{ 106, 25, 56 },
		{ 106, 26, 56 },
		{ 106, 27, 56 },
		{ 106, 28, 56 },
		{ 106, 29, 56 },
		{ 106, 30, 56 },
		{ 106, 31, 56 },
		{ 106, 32, 56 },
		{ 106, 33, 56 },
		{ 106, 34, 56 },
		{ 106, 35, 56 },
		{ 106, 36, 56 },
		{ 106, 37, 56 },
		{ 106, 38, 56 },
		{ 106, 39, 56 },
		{ 106, 40, 56 },
		{ 106, 41, 56 },
		{ 106, 42, 56 },
		{ 106, 43, 56 },
		{ 106, 44, 56 },
		{ 106, 45, 56 },
		{ 106, 46, 56 },
		{ 106, 47, 56 },
		{ 106, 48, 56 },
		{ 106, 49, 56 },
		{ 106, 50, 56 },
		{ 106, 51, 56 },
		{ 106, 52, 56 },
		{ 106, 53, 56 },
		{ 106, 54, 56 },
		{ 106, 55, 56 },
		{ 106, 56, 56 },
		{ 106, 57, 56 },
		{ 106, 58, 56 },
		{ 106, 59, 56 },
		{ 106, 60, 56 },
		{ 106, 61, 56 },
		{ 106, 62, 56 },
		{ 106, 63, 56 },
		{ 106, 64, 56 },
		{ 106, 65, 56 },
		{ 106, 66, 56 },
		{ 106, 67, 56 },
		{ 106, 68, 56 },
		{ 106, 69, 56 },
		{ 106, 70, 56 },
		{ 106, 71, 56 },
		{ 106, 72, 56 },
		{ 106, 73, 56 },
		{ 106, 74, 56 },
		{ 106, 75, 56 },
		{ 106, 76, 56 },
		{ 106, 77, 56 },
		{ 106, 78, 56 },
		{ 106, 79, 56 },
		{ 106, 80, 56 },
		{ 106, 81, 56 },
		{ 106, 82, 56 },
		{ 106, 83, 56 },
		{ 106, 84, 56 },
		{ 107, 18, 56 },
		{ 107, 19, 56 },
		{ 107, 20, 56 },
		{ 107, 21, 56 },
		{ 107, 22, 56 },
		{ 107, 23, 56 },
		{ 107, 24, 56 },
		{ 107, 25, 56 },
		{ 107, 26, 56 },
		{ 107, 27, 56 },
		{ 107, 28, 56 },
		{ 107, 29, 56 },
		{ 107, 30, 56 },
		{ 107, 31, 56 },
		{ 107, 32, 56 },
		{ 107, 33, 56 },
		{ 107, 34, 56 },
		{ 107, 35, 56 },
		{ 107, 36, 56 },
		{ 107, 37, 56 },
		{ 107, 38, 56 },
		{ 107, 39, 56 },
		{ 107, 40, 56 },
		{ 107, 41, 56 },
		{ 107, 42, 56 },
		{ 107, 43, 56 },
		{ 107, 44, 56 },
		{ 107, 45, 56 },
		{ 107, 46, 56 },
		{ 107, 47, 56 },
		{ 107, 48, 56 },
		{ 107, 49, 56 },
		{ 107, 50, 56 },
		{ 107, 51, 56 },
		{ 107, 52, 56 },
		{ 107, 53, 56 },
		{ 107, 54, 56 },
		{ 107, 55, 56 },
		{ 107, 56, 56 },
		{ 107, 57, 56 },
		{ 107, 58, 56 },
		{ 107, 59, 56 },
		{ 107, 60, 56 },
		{ 107, 61, 56 },
		{ 107, 62, 56 },
		{ 107, 63, 56 },
		{ 107, 64, 56 },
		{ 107, 65, 56 },
		{ 107, 66, 56 },
		{ 107, 67, 56 },
		{ 107, 68, 56 },
		{ 107, 69, 56 },
		{ 107, 70, 56 },
		{ 107, 71, 56 },
		{ 107, 72, 56 },
		{ 107, 73, 56 },
		{ 107, 74, 56 },
		{ 107, 75, 56 },
		{ 107, 76, 56 },
		{ 107, 77, 56 },
		{ 107, 78, 56 },
		{ 107, 79, 56 },
		{ 107, 80, 56 },
		{ 107, 81, 56 },
		{ 107, 82, 56 },
		{ 107, 83, 56 },
		{ 107, 84, 56 },
		{ 107, 85, 56 },
		{ 108, 19, 56 },
		{ 108, 20, 56 },
		{ 108, 21, 56 },
		{ 108, 22, 56 },
		{ 108, 23, 56 },
		{ 108, 24, 56 },
		{ 108, 25, 56 },
		{ 108, 26, 56 },
		{ 108, 27, 56 },
		{ 108, 28, 56 },
		{ 108, 29, 56 },
		{ 108, 30, 56 },
		{ 108, 31, 56 },
		{ 108, 32, 56 },
		{ 108, 33, 56 },
		{ 108, 34, 56 },
		{ 108, 35, 56 },
		{ 108, 36, 56 },
		{ 108, 37, 56 },
		{ 108, 38, 56 },
		{ 108, 39, 56 },
		{ 108, 40, 56 },
		{ 108, 41, 56 },
		{ 108, 42, 56 },
		{ 108, 43, 56 },
		{ 108, 44, 56 },
		{ 108, 45, 56 },
		{ 108, 46, 56 },
		{ 108, 47, 56 },
		{ 108, 48, 56 },
		{ 108, 49, 56 },
		{ 108, 50, 56 },
		{ 108, 51, 56 },
		{ 108, 52, 56 },
		{ 108, 53, 56 },
		{ 108, 54, 56 },
		{ 108, 55, 56 },
		{ 108, 56, 56 },
		{ 108, 57, 56 },
		{ 108, 58, 56 },
		{ 108, 59, 56 },
		{ 108, 60, 56 },
		{ 108, 61, 56 },
		{ 108, 62, 56 },
		{ 108, 63, 56 },
		{ 108, 64, 56 },
		{ 108, 65, 56 },
		{ 108, 66, 56 },
		{ 108, 67, 56 },
		{ 108, 68, 56 },
		{ 108, 69, 56 },
		{ 108, 70, 56 },
		{ 108, 71, 56 },
		{ 108, 72, 56 },
		{ 108, 73, 56 },
		{ 108, 74, 56 },
		{ 108, 75, 56 },
		{ 108, 76, 56 },
		{ 108, 77, 56 },
		{ 108, 78, 56 },
		{ 108, 79, 56 },
		{ 108, 80, 56 },
		{ 108, 81, 56 },
		{ 108, 82, 56 },
		{ 108, 83, 56 },
		{ 108, 84, 56 },
		{ 108, 85, 56 },
		{ 108, 86, 56 },
		{ 109, 21, 56 },
		{ 109, 22, 56 },
		{ 109, 23, 56 },
		{ 109, 24, 56 },
		{ 109, 25, 56 },
		{ 109, 26, 56 },
		{ 109, 27, 56 },
		{ 109, 28, 56 },
		{ 109, 29, 56 },
		{ 109, 30, 56 },
		{ 109, 31, 56 },
		{ 109, 32, 56 },
		{ 109, 33, 56 },
		{ 109, 34, 56 },
		{ 109, 35, 56 },
		{ 109, 36, 56 },
		{ 109, 37, 56 },
		{ 109, 38, 56 },
		{ 109, 39, 56 },
		{ 109, 40, 56 },
		{ 109, 41, 56 },
		{ 109, 42, 56 },
		{ 109, 43, 56 },
		{ 109, 44, 56 },
		{ 109, 45, 56 },
		{ 109, 46, 56 },
		{ 109, 47, 56 },
		{ 109, 48, 56 },
		{ 109, 49, 56 },
		{ 109, 50, 56 },
		{ 109, 51, 56 },
		{ 109, 52, 56 },
		{ 109, 53, 56 },
		{ 109, 54, 56 },
		{ 109, 55, 56 },
		{ 109, 56, 56 },
		{ 109, 57, 56 },
		{ 109, 58, 56 },
		{ 109, 59, 56 },
		{ 109, 60, 56 },
		{ 109, 61, 56 },
		{ 109, 62, 56 },
		{ 109, 63, 56 },
		{ 109, 64, 56 },
		{ 109, 65, 56 },
		{ 109, 66, 56 },
		{ 109, 67, 56 },
		{ 109, 68, 56 },
		{ 109, 69, 56 },
		{ 109, 70, 56 },
		{ 109, 71, 56 },
		{ 109, 72, 56 },
		{ 109, 73, 56 },
		{ 109, 74, 56 },
		{ 109, 75, 56 },
		{ 109, 76, 56 },
		{ 109, 77, 56 },
		{ 109, 78, 56 },
		{ 109, 79, 56 },
		{ 109, 80, 56 },
		{ 109, 81, 56 },
		{ 109, 82, 56 },
		{ 109, 83, 56 },
		{ 109, 84, 56 },
		{ 109, 85, 56 },
		{ 109, 86, 56 },
		{ 109, 87, 56 },
		{ 110, 22, 56 },
		{ 110, 23, 56 },
		{ 110, 24, 56 },
		{ 110, 25, 56 },
		{ 110, 26, 56 },
		{ 110, 27, 56 },
		{ 110, 28, 56 },
		{ 110, 29, 56 },
		{ 110, 30, 56 },
		{ 110, 31, 56 },
		{ 110, 32, 56 },
		{ 110, 33, 56 },
		{ 110, 34, 56 },
		{ 110, 35, 56 },
		{ 110, 36, 56 },
		{ 110, 37, 56 },
		{ 110, 38, 56 },
		{ 110, 39, 56 },
		{ 110, 40, 56 },
		{ 110, 41, 56 },
		{ 110, 42, 56 },
		{ 110, 43, 56 },
		{ 110, 44, 56 },
		{ 110, 45, 56 },
		{ 110, 46, 56 },
		{ 110, 47, 56 },
		{ 110, 48, 56 },
		{ 110, 49, 56 },
		{ 110, 50, 56 },
		{ 110, 51, 56 },
		{ 110, 52, 56 },
		{ 110, 53, 56 },
		{ 110, 54, 56 },
		{ 110, 55, 56 },
		{ 110, 56, 56 },
		{ 110, 57, 56 },
		{ 110, 58, 56 },
		{ 110, 59, 56 },
		{ 110, 60, 56 },
		{ 110, 61, 56 },
		{ 110, 62, 56 },
		{ 110, 63, 56 },
		{ 110, 64, 56 },
		{ 110, 65, 56 },
		{ 110, 66, 56 },
		{ 110, 67, 56 },
		{ 110, 68, 56 },
		{ 110, 69, 56 },
		{ 110, 70, 56 },
		{ 110, 71, 56 },
		{ 110, 72, 56 },
		{ 110, 73, 56 },
		{ 110, 74, 56 },
		{ 110, 75, 56 },
		{ 110, 76, 56 },
		{ 110, 77, 56 },
		{ 110, 78, 56 },
		{ 110, 79, 56 },
		{ 110, 80, 56 },
		{ 110, 81, 56 },
		{ 110, 82, 56 },
		{ 110, 83, 56 },
		{ 110, 84, 56 },
		{ 110, 85, 56 },
		{ 110, 86, 56 },
		{ 110, 87, 56 },
		{ 110, 88, 56 },
		{ 111, 24, 56 },
		{ 111, 25, 56 },
		{ 111, 26, 56 },
		{ 111, 27, 56 },
		{ 111, 28, 56 },
		{ 111, 29, 56 },
		{ 111, 30, 56 },
		{ 111, 31, 56 },
		{ 111, 32, 56 },
		{ 111, 33, 56 },
		{ 111, 34, 56 },
		{ 111, 35, 56 },
		{ 111, 36, 56 },
		{ 111, 37, 56 },
		{ 111, 38, 56 },
		{ 111, 39, 56 },
		{ 111, 40, 56 },
		{ 111, 41, 56 },
		{ 111, 42, 56 },
		{ 111, 43, 56 },
		{ 111, 44, 56 },
		{ 111, 45, 56 },
		{ 111, 46, 56 },
		{ 111, 47, 56 },
		{ 111, 48, 56 },
		{ 111, 49, 56 },
		{ 111, 50, 56 },
		{ 111, 51, 56 },
		{ 111, 52, 56 },
		{ 111, 53, 56 },
		{ 111, 54, 56 },
		{ 111, 55, 56 },
		{ 111, 56, 56 },
		{ 111, 57, 56 },
		{ 111, 58, 56 },
		{ 111, 59, 56 },
		{ 111, 60, 56 },
		{ 111, 61, 56 },
		{ 111, 62, 56 },
		{ 111, 63, 56 },
		{ 111, 64, 56 },
		{ 111, 65, 56 },
		{ 111, 66, 56 },
		{ 111, 67, 56 },
		{ 111, 68, 56 },
		{ 111, 69, 56 },
		{ 111, 70, 56 },
		{ 111, 71, 56 },
		{ 111, 72, 56 },
		{ 111, 73, 56 },
		{ 111, 74, 56 },
		{ 111, 75, 56 },
		{ 111, 76, 56 },
		{ 111, 77, 56 },
		{ 111, 78, 56 },
		{ 111, 79, 56 },
		{ 111, 80, 56 },
		{ 111, 81, 56 },
		{ 111, 82, 56 },
		{ 111, 83, 56 },
		{ 111, 84, 56 },
		{ 111, 85, 56 },
		{ 111, 86, 56 },
		{ 111, 87, 56 },
		{ 111, 88, 56 },
		{ 111, 89, 56 },
		{ 111, 90, 56 },
		{ 112, 26, 56 },
		{ 112, 27, 56 },
		{ 112, 28, 56 },
		{ 112, 29, 56 },
		{ 112, 30, 56 },
		{ 112, 31, 56 },
		{ 112, 32, 56 },
		{ 112, 33, 56 },
		{ 112, 34, 56 },
		{ 112, 35, 56 },
		{ 112, 36, 56 },
		{ 112, 37, 56 },
		{ 112, 38, 56 },
		{ 112, 39, 56 },
		{ 112, 40, 56 },
		{ 112, 41, 56 },
		{ 112, 42, 56 },
		{ 112, 43, 56 },
		{ 112, 44, 56 },
		{ 112, 45, 56 },
		{ 112, 46, 56 },
		{ 112, 47, 56 },
		{ 112, 48, 56 },
		{ 112, 49, 56 },
		{ 112, 50, 56 },
		{ 112, 51, 56 },
		{ 112, 52, 56 },
		{ 112, 53, 56 },
		{ 112, 54, 56 },
		{ 112, 55, 56 },
		{ 112, 56, 56 },
		{ 112, 57, 56 },
		{ 112, 58, 56 },
		{ 112, 59, 56 },
		{ 112, 60, 56 },
		{ 112, 61, 56 },
		{ 112, 62, 56 },
		{ 112, 63, 56 },
		{ 112, 64, 56 },
		{ 112, 65, 56 },
		{ 112, 66, 56 },
		{ 112, 67, 56 },
		{ 112, 68, 56 },
		{ 112, 69, 56 },
		{ 112, 70, 56 },
		{ 112, 71, 56 },
		{ 112, 72, 56 },
		{ 112, 73, 56 },
		{ 112, 74, 56 },
		{ 112, 75, 56 },
		{ 112, 76, 56 },
		{ 112, 77, 56 },
		{ 112, 78, 56 },
		{ 112, 79, 56 },
		{ 112, 80, 56 },
		{ 112, 81, 56 },
		{ 112, 82, 56 },
		{ 112, 83, 56 },
		{ 112, 84, 56 },
		{ 112, 85, 56 },
		{ 112, 86, 56 },
		{ 112, 87, 56 },
		{ 112, 88, 56 },
		{ 112, 89, 56 },
		{ 112, 90, 56 },
		{ 112, 91, 56 },
		{ 113, 27, 56 },
		{ 113, 28, 56 },
		{ 113, 29, 56 },
		{ 113, 30, 56 },
		{ 113, 31, 56 },
		{ 113, 32, 56 },
		{ 113, 33, 56 },
		{ 113, 34, 56 },
		{ 113, 35, 56 },
		{ 113, 36, 56 },
		{ 113, 37, 56 },
		{ 113, 38, 56 },
		{ 113, 39, 56 },
		{ 113, 40, 56 },
		{ 113, 41, 56 },
		{ 113, 42, 56 },
		{ 113, 43, 56 },
		{ 113, 44, 56 },
		{ 113, 45, 56 },
		{ 113, 46, 56 },
		{ 113, 47, 56 },
		{ 113, 48, 56 },
		{ 113, 49, 56 },
		{ 113, 50, 56 },
		{ 113, 51, 56 },
		{ 113, 52, 56 },
		{ 113, 53, 56 },
		{ 113, 54, 56 },
		{ 113, 55, 56 },
		{ 113, 56, 56 },
		{ 113, 57, 56 },
		{ 113, 58, 56 },
		{ 113, 59, 56 },
		{ 113, 60, 56 },
		{ 113, 61, 56 },
		{ 113, 62, 56 },
		{ 113, 63, 56 },
		{ 113, 64, 56 },
		{ 113, 65, 56 },
		{ 113, 66, 56 },
		{ 113, 67, 56 },
		{ 113, 68, 56 },
		{ 113, 69, 56 },
		{ 113, 70, 56 },
		{ 113, 71, 56 },
		{ 113, 72, 56 },
		{ 113, 73, 56 },
		{ 113, 74, 56 },
		{ 113, 75, 56 },
		{ 113, 76, 56 },
		{ 113, 77, 56 },
		{ 113, 78, 56 },
		{ 113, 79, 56 },
		{ 113, 80, 56 },
		{ 113, 81, 56 },
		{ 113, 82, 56 },
		{ 113, 83, 56 },
		{ 113, 84, 56 },
		{ 113, 85, 56 },
		{ 113, 86, 56 },
		{ 113, 87, 56 },
		{ 113, 88, 56 },
		{ 113, 89, 56 },
		{ 113, 90, 56 },
		{ 113, 91, 56 },
		{ 113, 92, 56 },
		{ 114, 29, 56 },
		{ 114, 30, 56 },
		{ 114, 31, 56 },
		{ 114, 32, 56 },
		{ 114, 33, 56 },
		{ 114, 34, 56 },
		{ 114, 35, 56 },
		{ 114, 36, 56 },
		{ 114, 37, 56 },
		{ 114, 38, 56 },
		{ 114, 39, 56 },
		{ 114, 40, 56 },
		{ 114, 41, 56 },
		{ 114, 42, 56 },
		{ 114, 43, 56 },
		{ 114, 44, 56 },
		{ 114, 45, 56 },
		{ 114, 46, 56 },
		{ 114, 47, 56 },
		{ 114, 48, 56 },
		{ 114, 49, 56 },
		{ 114, 50, 56 },
		{ 114, 51, 56 },
		{ 114, 52, 56 },
		{ 114, 53, 56 },
		{ 114, 54, 56 },
		{ 114, 55, 56 },
		{ 114, 56, 56 },
		{ 114, 57, 56 },
		{ 114, 58, 56 },
		{ 114, 59, 56 },
		{ 114, 60, 56 },
		{ 114, 61, 56 },
		{ 114, 62, 56 },
		{ 114, 63, 56 },
		{ 114, 64, 56 },
		{ 114, 65, 56 },
		{ 114, 66, 56 },
		{ 114, 67, 56 },
		{ 114, 68, 56 },
		{ 114, 69, 56 },
		{ 114, 70, 56 },
		{ 114, 71, 56 },
		{ 114, 72, 56 },
		{ 114, 73, 56 },
		{ 114, 74, 56 },
		{ 114, 75, 56 },
		{ 114, 76, 56 },
		{ 114, 77, 56 },
		{ 114, 78, 56 },
		{ 114, 79, 56 },
		{ 114, 80, 56 },
		{ 114, 81, 56 },
		{ 114, 82, 56 },
		{ 114, 83, 56 },
		{ 114, 84, 56 },
		{ 114, 85, 56 },
		{ 114, 86, 56 },
		{ 114, 87, 56 },
		{ 114, 88, 56 },
		{ 114, 89, 56 },
		{ 114, 90, 56 },
		{ 114, 91, 56 },
		{ 114, 92, 56 },
		{ 114, 93, 56 },
		{ 115, 30, 56 },
		{ 115, 31, 56 },
		{ 115, 32, 56 },
		{ 115, 33, 56 },
		{ 115, 34, 56 },
		{ 115, 35, 56 },
		{ 115, 36, 56 },
		{ 115, 37, 56 },
		{ 115, 38, 56 },
		{ 115, 39, 56 },
		{ 115, 40, 56 },
		{ 115, 41, 56 },
		{ 115, 42, 56 },
		{ 115, 43, 56 },
		{ 115, 44, 56 },
		{ 115, 45, 56 },
		{ 115, 46, 56 },
		{ 115, 47, 56 },
		{ 115, 48, 56 },
		{ 115, 49, 56 },
		{ 115, 50, 56 },
		{ 115, 51, 56 },
		{ 115, 52, 56 },
		{ 115, 53, 56 },
		{ 115, 54, 56 },
		{ 115, 55, 56 },
		{ 115, 56, 56 },
		{ 115, 57, 56 },
		{ 115, 58, 56 },
		{ 115, 59, 56 },
		{ 115, 60, 56 },
		{ 115, 61, 56 },
		{ 115, 62, 56 },
		{ 115, 63, 56 },
		{ 115, 64, 56 },
		{ 115, 65, 56 },
		{ 115, 66, 56 },
		{ 115, 67, 56 },
		{ 115, 68, 56 },
		{ 115, 69, 56 },
		{ 115, 70, 56 },
		{ 115, 71, 56 },
		{ 115, 72, 56 },
		{ 115, 73, 56 },
		{ 115, 74, 56 },
		{ 115, 75, 56 },
		{ 115, 76, 56 },
		{ 115, 77, 56 },
		{ 115, 78, 56 },
		{ 115, 79, 56 },
		{ 115, 80, 56 },
		{ 115, 81, 56 },
		{ 115, 82, 56 },
		{ 115, 83, 56 },
		{ 115, 84, 56 },
		{ 115, 85, 56 },
		{ 115, 86, 56 },
		{ 115, 87, 56 },
		{ 115, 88, 56 },
		{ 115, 89, 56 },
		{ 115, 90, 56 },
		{ 115, 91, 56 },
		{ 115, 92, 56 },
		{ 115, 93, 56 },
		{ 115, 94, 56 },
		{ 116, 32, 56 },
		{ 116, 33, 56 },
		{ 116, 34, 56 },
		{ 116, 35, 56 },
		{ 116, 36, 56 },
		{ 116, 37, 56 },
		{ 116, 38, 56 },
		{ 116, 39, 56 },
		{ 116, 40, 56 },
		{ 116, 41, 56 },
		{ 116, 42, 56 },
		{ 116, 43, 56 },
		{ 116, 44, 56 },
		{ 116, 45, 56 },
		{ 116, 46, 56 },
		{ 116, 47, 56 },
		{ 116, 48, 56 },
		{ 116, 49, 56 },
		{ 116, 50, 56 },
		{ 116, 51, 56 },
		{ 116, 52, 56 },
		{ 116, 53, 56 },
		{ 116, 54, 56 },
		{ 116, 55, 56 },
		{ 116, 56, 56 },
		{ 116, 57, 56 },
		{ 116, 58, 56 },
		{ 116, 59, 56 },
		{ 116, 60, 56 },
		{ 116, 61, 56 },
		{ 116, 62, 56 },
		{ 116, 63, 56 },
		{ 116, 64, 56 },
		{ 116, 65, 56 },
		{ 116, 66, 56 },
		{ 116, 67, 56 },
		{ 116, 68, 56 },
		{ 116, 69, 56 },
		{ 116, 70, 56 },
		{ 116, 71, 56 },
		{ 116, 72, 56 },
		{ 116, 73, 56 },
		{ 116, 74, 56 },
		{ 116, 75, 56 },
		{ 116, 76, 56 },
		{ 116, 77, 56 },
		{ 116, 78, 56 },
		{ 116, 79, 56 },
		{ 116, 80, 56 },
		{ 116, 81, 56 },
		{ 116, 82, 56 },
		{ 116, 83, 56 },
		{ 116, 84, 56 },
		{ 116, 85, 56 },
		{ 116, 86, 56 },
		{ 116, 87, 56 },
		{ 116, 88, 56 },
		{ 116, 89, 56 },
		{ 116, 90, 56 },
		{ 116, 91, 56 },
		{ 116, 92, 56 },
		{ 116, 93, 56 },
		{ 116, 94, 56 },
		{ 116, 95, 56 },
		{ 117, 34, 56 },
		{ 117, 35, 56 },
		{ 117, 36, 56 },
		{ 117, 37, 56 },
		{ 117, 38, 56 },
		{ 117, 39, 56 },
		{ 117, 40, 56 },
		{ 117, 41, 56 },
		{ 117, 42, 56 },
		{ 117, 43, 56 },
		{ 117, 44, 56 },
		{ 117, 45, 56 },
		{ 117, 46, 56 },
		{ 117, 47, 56 },
		{ 117, 48, 56 },
		{ 117, 49, 56 },
		{ 117, 50, 56 },
		{ 117, 51, 56 },
		{ 117, 52, 56 },
		{ 117, 53, 56 },
		{ 117, 54, 56 },
		{ 117, 55, 56 },
		{ 117, 56, 56 },
		{ 117, 57, 56 },
		{ 117, 58, 56 },
		{ 117, 59, 56 },
		{ 117, 60, 56 },
		{ 117, 61, 56 },
		{ 117, 62, 56 },
		{ 117, 63, 56 },
		{ 117, 64, 56 },
		{ 117, 65, 56 },
		{ 117, 66, 56 },
		{ 117, 67, 56 },
		{ 117, 68, 56 },
		{ 117, 69, 56 },
		{ 117, 70, 56 },
		{ 117, 71, 56 },
		{ 117, 72, 56 },
		{ 117, 73, 56 },
		{ 117, 74, 56 },
		{ 117, 75, 56 },
		{ 117, 76, 56 },
		{ 117, 77, 56 },
		{ 117, 78, 56 },
		{ 117, 79, 56 },
		{ 117, 80, 56 },
		{ 117, 81, 56 },
		{ 117, 82, 56 },
		{ 117, 83, 56 },
		{ 117, 84, 56 },
		{ 117, 85, 56 },
		{ 117, 86, 56 },
		{ 117, 87, 56 },
		{ 117, 88, 56 },
		{ 117, 89, 56 },
		{ 117, 90, 56 },
		{ 117, 91, 56 },
		{ 117, 92, 56 },
		{ 117, 93, 56 },
		{ 117, 94, 56 },
		{ 117, 95, 56 },
		{ 117, 96, 56 },
		{ 118, 35, 56 },
		{ 118, 36, 56 },
		{ 118, 37, 56 },
		{ 118, 38, 56 },
		{ 118, 39, 56 },
		{ 118, 40, 56 },
		{ 118, 41, 56 },
		{ 118, 42, 56 },
		{ 118, 43, 56 },
		{ 118, 44, 56 },
		{ 118, 45, 56 },
		{ 118, 46, 56 },
		{ 118, 47, 56 },
		{ 118, 48, 56 },
		{ 118, 49, 56 },
		{ 118, 50, 56 },
		{ 118, 51, 56 },
		{ 118, 52, 56 },
		{ 118, 53, 56 },
		{ 118, 54, 56 },
		{ 118, 55, 56 },
		{ 118, 56, 56 },
		{ 118, 57, 56 },
		{ 118, 58, 56 },
		{ 118, 59, 56 },
		{ 118, 60, 56 },
		{ 118, 61, 56 },
		{ 118, 62, 56 },
		{ 118, 63, 56 },
		{ 118, 64, 56 },
		{ 118, 65, 56 },
		{ 118, 66, 56 },
		{ 118, 67, 56 },
		{ 118, 68, 56 },
		{ 118, 69, 56 },
		{ 118, 70, 56 },
		{ 118, 71, 56 },
		{ 118, 72, 56 },
		{ 118, 73, 56 },
		{ 118, 74, 56 },
		{ 118, 75, 56 },
		{ 118, 76, 56 },
		{ 118, 77, 56 },
		{ 118, 78, 56 },
		{ 118, 79, 56 },
		{ 118, 80, 56 },
		{ 118, 81, 56 },
		{ 118, 82, 56 },
		{ 118, 83, 56 },
		{ 118, 84, 56 },
		{ 118, 85, 56 },
		{ 118, 86, 56 },
		{ 118, 87, 56 },
		{ 118, 88, 56 },
		{ 118, 89, 56 },
		{ 118, 90, 56 },
		{ 118, 91, 56 },
		{ 118, 92, 56 },
		{ 118, 93, 56 },
		{ 118, 94, 56 },
		{ 118, 95, 56 },
		{ 118, 96, 56 },
		{ 118, 97, 56 },
		{ 118, 98, 56 },
		{ 119, 37, 56 },
		{ 119, 38, 56 },
		{ 119, 39, 56 },
		{ 119, 40, 56 },
		{ 119, 41, 56 },
		{ 119, 42, 56 },
		{ 119, 43, 56 },
		{ 119, 44, 56 },
		{ 119, 45, 56 },
		{ 119, 46, 56 },
		{ 119, 47, 56 },
		{ 119, 48, 56 },
		{ 119, 49, 56 },
		{ 119, 50, 56 },
		{ 119, 51, 56 },
		{ 119, 52, 56 },
		{ 119, 53, 56 },
		{ 119, 54, 56 },
		{ 119, 55, 56 },
		{ 119, 56, 56 },
		{ 119, 57, 56 },
		{ 119, 58, 56 },
		{ 119, 59, 56 },
		{ 119, 60, 56 },
		{ 119, 61, 56 },
		{ 119, 62, 56 },
		{ 119, 63, 56 },
		{ 119, 64, 56 },
		{ 119, 65, 56 },
		{ 119, 66, 56 },
		{ 119, 67, 56 },
		{ 119, 68, 56 },
		{ 119, 69, 56 },
		{ 119, 70, 56 },
		{ 119, 71, 56 },
		{ 119, 72, 56 },
		{ 119, 73, 56 },
		{ 119, 74, 56 },
		{ 119, 75, 56 },
		{ 119, 76, 56 },
		{ 119, 77, 56 },
		{ 119, 78, 56 },
		{ 119, 79, 56 },
		{ 119, 80, 56 },
		{ 119, 81, 56 },
		{ 119, 82, 56 },
		{ 119, 83, 56 },
		{ 119, 84, 56 },
		{ 119, 85, 56 },
		{ 119, 86, 56 },
		{ 119, 87, 56 },
		{ 119, 88, 56 },
		{ 119, 89, 56 },
		{ 119, 90, 56 },
		{ 119, 91, 56 },
		{ 119, 92, 56 },
		{ 119, 93, 56 },
		{ 119, 94, 56 },
		{ 119, 95, 56 },
		{ 119, 96, 56 },
		{ 119, 97, 56 },
		{ 119, 98, 56 },
		{ 119, 99, 56 },
		{ 120, 38, 56 },
		{ 120, 39, 56 },
		{ 120, 40, 56 },
		{ 120, 41, 56 },
		{ 120, 42, 56 },
		{ 120, 43, 56 },
		{ 120, 44, 56 },
		{ 120, 45, 56 },
		{ 120, 46, 56 },
		{ 120, 47, 56 },
		{ 120, 48, 56 },
		{ 120, 49, 56 },
		{ 120, 50, 56 },
		{ 120, 51, 56 },
		{ 120, 52, 56 },
		{ 120, 53, 56 },
		{ 120, 54, 56 },
		{ 120, 55, 56 },
		{ 120, 56, 56 },
		{ 120, 57, 56 },
		{ 120, 58, 56 },
		{ 120, 59, 56 },
		{ 120, 60, 56 },
		{ 120, 61, 56 },
		{ 120, 62, 56 },
		{ 120, 63, 56 },
		{ 120, 64, 56 },
		{ 120, 65, 56 },
		{ 120, 66, 56 },
		{ 120, 67, 56 },
		{ 120, 68, 56 },
		{ 120, 69, 56 },
		{ 120, 70, 56 },
		{ 120, 71, 56 },
		{ 120, 72, 56 },
		{ 120, 73, 56 },
		{ 120, 74, 56 },
		{ 120, 75, 56 },
		{ 120, 76, 56 },
		{ 120, 77, 56 },
		{ 120, 78, 56 },
		{ 120, 79, 56 },
		{ 120, 80, 56 },
		{ 120, 81, 56 },
		{ 120, 82, 56 },
		{ 120, 83, 56 },
		{ 120, 84, 56 },
		{ 120, 85, 56 },
		{ 120, 86, 56 },
		{ 120, 87, 56 },
		{ 120, 88, 56 },
		{ 120, 89, 56 },
		{ 120, 90, 56 },
		{ 120, 91, 56 },
		{ 120, 92, 56 },
		{ 120, 93, 56 },
		{ 120, 94, 56 },
		{ 120, 95, 56 },
		{ 120, 96, 56 },
		{ 120, 97, 56 },
		{ 120, 98, 56 },
		{ 120, 99, 56 },
		{ 120, 100, 56 },
		{ 121, 40, 56 },
		{ 121, 41, 56 },
		{ 121, 42, 56 },
		{ 121, 43, 56 },
		{ 121, 44, 56 },
		{ 121, 45, 56 },
		{ 121, 46, 56 },
		{ 121, 47, 56 },
		{ 121, 48, 56 },
		{ 121, 49, 56 },
		{ 121, 50, 56 },
		{ 121, 51, 56 },
		{ 121, 52, 56 },
		{ 121, 53, 56 },
		{ 121, 54, 56 },
		{ 121, 55, 56 },
		{ 121, 56, 56 },
		{ 121, 57, 56 },
		{ 121, 58, 56 },
		{ 121, 59, 56 },
		{ 121, 60, 56 },
		{ 121, 61, 56 },
		{ 121, 62, 56 },
		{ 121, 63, 56 },
		{ 121, 64, 56 },
		{ 121, 65, 56 },
		{ 121, 66, 56 },
		{ 121, 67, 56 },
		{ 121, 68, 56 },
		{ 121, 69, 56 },
		{ 121, 70, 56 },
		{ 121, 71, 56 },
		{ 121, 72, 56 },
		{ 121, 73, 56 },
		{ 121, 74, 56 },
		{ 121, 75, 56 },
		{ 121, 76, 56 },
		{ 121, 77, 56 },
		{ 121, 78, 56 },
		{ 121, 79, 56 },
		{ 121, 80, 56 },
		{ 121, 81, 56 },
		{ 121, 82, 56 },
		{ 121, 83, 56 },
		{ 121, 84, 56 },
		{ 121, 85, 56 },
		{ 121, 86, 56 },
		{ 121, 87, 56 },
		{ 121, 88, 56 },
		{ 121, 89, 56 },
		{ 121, 90, 56 },
		{ 121, 91, 56 },
		{ 121, 92, 56 },
		{ 121, 93, 56 },
		{ 121, 94, 56 },
		{ 121, 95, 56 },
		{ 121, 96, 56 },
		{ 121, 97, 56 },
		{ 121, 98, 56 },
		{ 121, 99, 56 },
		{ 121, 100, 56 },
		{ 121, 101, 56 },
		{ 122, 42, 56 },
		{ 122, 43, 56 },
		{ 122, 44, 56 },
		{ 122, 45, 56 },
		{ 122, 46, 56 },
		{ 122, 47, 56 },
		{ 122, 48, 56 },
		{ 122, 49, 56 },
		{ 122, 50, 56 },
		{ 122, 51, 56 },
		{ 122, 52, 56 },
		{ 122, 53, 56 },
		{ 122, 54, 56 },
		{ 122, 55, 56 },
		{ 122, 56, 56 },
		{ 122, 57, 56 },
		{ 122, 58, 56 },
		{ 122, 59, 56 },
		{ 122, 60, 56 },
		{ 122, 61, 56 },
		{ 122, 62, 56 },
		{ 122, 63, 56 },
		{ 122, 64, 56 },
		{ 122, 65, 56 },
		{ 122, 66, 56 },
		{ 122, 67, 56 },
		{ 122, 68, 56 },
		{ 122, 69, 56 },
		{ 122, 70, 56 },
		{ 122, 71, 56 },
		{ 122, 72, 56 },
		{ 122, 73, 56 },
		{ 122, 74, 56 },
		{ 122, 75, 56 },
		{ 122, 76, 56 },
		{ 122, 77, 56 },
		{ 122, 78, 56 },
		{ 122, 79, 56 },
		{ 122, 80, 56 },
		{ 122, 81, 56 },
		{ 122, 82, 56 },
		{ 122, 83, 56 },
		{ 122, 84, 56 },
		{ 122, 85, 56 },
		{ 122, 86, 56 },
		{ 122, 87, 56 },
		{ 122, 88, 56 },
		{ 122, 89, 56 },
		{ 122, 90, 56 },
		{ 122, 91, 56 },
		{ 122, 92, 56 },
		{ 122, 93, 56 },
		{ 122, 94, 56 },
		{ 122, 95, 56 },
		{ 122, 96, 56 },
		{ 122, 97, 56 },
		{ 122, 98, 56 },
		{ 122, 99, 56 },
		{ 122, 100, 56 },
		{ 122, 101, 56 },
		{ 122, 102, 56 },
		{ 123, 43, 56 },
		{ 123, 44, 56 },
		{ 123, 45, 56 },
		{ 123, 46, 56 },
		{ 123, 47, 56 },
		{ 123, 48, 56 },
		{ 123, 49, 56 },
		{ 123, 50, 56 },
		{ 123, 51, 56 },
		{ 123, 52, 56 },
		{ 123, 53, 56 },
		{ 123, 54, 56 },
		{ 123, 55, 56 },
		{ 123, 56, 56 },
		{ 123, 57, 56 },
		{ 123, 58, 56 },
		{ 123, 59, 56 },
		{ 123, 60, 56 },
		{ 123, 61, 56 },
		{ 123, 62, 56 },
		{ 123, 63, 56 },
		{ 123, 64, 56 },
		{ 123, 65, 56 },
		{ 123, 66, 56 },
		{ 123, 67, 56 },
		{ 123, 68, 56 },
		{ 123, 69, 56 },
		{ 123, 70, 56 },
		{ 123, 71, 56 },
		{ 123, 72, 56 },
		{ 123, 73, 56 },
		{ 123, 74, 56 },
		{ 123, 75, 56 },
		{ 123, 76, 56 },
		{ 123, 77, 56 },
		{ 123, 78, 56 },
		{ 123, 79, 56 },
		{ 123, 80, 56 },
		{ 123, 81, 56 },
		{ 123, 82, 56 },
		{ 123, 83, 56 },
		{ 123, 84, 56 },
		{ 123, 85, 56 },
		{ 123, 86, 56 },
		{ 123, 87, 56 },
		{ 123, 88, 56 },
		{ 123, 89, 56 },
		{ 123, 90, 56 },
		{ 123, 91, 56 },
		{ 123, 92, 56 },
		{ 123, 93, 56 },
		{ 123, 94, 56 },
		{ 123, 95, 56 },
		{ 123, 96, 56 },
		{ 123, 97, 56 },
		{ 123, 98, 56 },
		{ 123, 99, 56 },
		{ 123, 100, 56 },
		{ 123, 101, 56 },
		{ 123, 102, 56 },
		{ 123, 103, 56 },
		{ 124, 45, 56 },
		{ 124, 46, 56 },
		{ 124, 47, 56 },
		{ 124, 48, 56 },
		{ 124, 49, 56 },
		{ 124, 50, 56 },
		{ 124, 51, 56 },
		{ 124, 52, 56 },
		{ 124, 53, 56 },
		{ 124, 54, 56 },
		{ 124, 55, 56 },
		{ 124, 56, 56 },
		{ 124, 57, 56 },
		{ 124, 58, 56 },
		{ 124, 59, 56 },
		{ 124, 60, 56 },
		{ 124, 61, 56 },
		{ 124, 62, 56 },
		{ 124, 63, 56 },
		{ 124, 64, 56 },
		{ 124, 65, 56 },
		{ 124, 66, 56 },
		{ 124, 67, 56 },
		{ 124, 68, 56 },
		{ 124, 69, 56 },
		{ 124, 70, 56 },
		{ 124, 71, 56 },
		{ 124, 72, 56 },
		{ 124, 73, 56 },
		{ 124, 74, 56 },
		{ 124, 75, 56 },
		{ 124, 76, 56 },
		{ 124, 77, 56 },
		{ 124, 78, 56 },
		{ 124, 79, 56 },
		{ 124, 80, 56 },
		{ 124, 81, 56 },
		{ 124, 82, 56 },
		{ 124, 83, 56 },
		{ 124, 84, 56 },
		{ 124, 85, 56 },
		{ 124, 86, 56 },
		{ 124, 87, 56 },
		{ 124, 88, 56 },
		{ 124, 89, 56 },
		{ 124, 90, 56 },
		{ 124, 91, 56 },
		{ 124, 92, 56 },
		{ 124, 93, 56 },
		{ 124, 94, 56 },
		{ 124, 95, 56 },
		{ 124, 96, 56 },
		{ 124, 97, 56 },
		{ 124, 98, 56 },
		{ 124, 99, 56 },
		{ 124, 100, 56 },
		{ 124, 101, 56 },
		{ 124, 102, 56 },
		{ 124, 103, 56 },
		{ 124, 104, 56 },
		{ 125, 46, 56 },
		{ 125, 47, 56 },
		{ 125, 48, 56 },
		{ 125, 49, 56 },
		{ 125, 50, 56 },
		{ 125, 51, 56 },
		{ 125, 52, 56 },
		{ 125, 53, 56 },
		{ 125, 54, 56 },
		{ 125, 55, 56 },
		{ 125, 56, 56 },
		{ 125, 57, 56 },
		{ 125, 58, 56 },
		{ 125, 59, 56 },
		{ 125, 60, 56 },
		{ 125, 61, 56 },
		{ 125, 62, 56 },
		{ 125, 63, 56 },
		{ 125, 64, 56 },
		{ 125, 65, 56 },
		{ 125, 66, 56 },
		{ 125, 67, 56 },
		{ 125, 68, 56 },
		{ 125, 69, 56 },
		{ 125, 70, 56 },
		{ 125, 71, 56 },
		{ 125, 72, 56 },
		{ 125, 73, 56 },
		{ 125, 74, 56 },
		{ 125, 75, 56 },
		{ 125, 76, 56 },
		{ 125, 77, 56 },
		{ 125, 78, 56 },
		{ 125, 79, 56 },
		{ 125, 80, 56 },
		{ 125, 81, 56 },
		{ 125, 82, 56 },
		{ 125, 83, 56 },
		{ 125, 84, 56 },
		{ 125, 85, 56 },
		{ 125, 86, 56 },
		{ 125, 87, 56 },
		{ 125, 88, 56 },
		{ 125, 89, 56 },
		{ 125, 90, 56 },
		{ 125, 91, 56 },
		{ 125, 92, 56 },
		{ 125, 93, 56 },
		{ 125, 94, 56 },
		{ 125, 95, 56 },
		{ 125, 96, 56 },
		{ 125, 97, 56 },
		{ 125, 98, 56 },
		{ 125, 99, 56 },
		{ 125, 100, 56 },
		{ 125, 101, 56 },
		{ 125, 102, 56 },
		{ 125, 103, 56 },
		{ 125, 104, 56 },
		{ 125, 105, 56 },
		{ 125, 106, 56 },
		{ 126, 48, 56 },
		{ 126, 49, 56 },
		{ 126, 50, 56 },
		{ 126, 51, 56 },
		{ 126, 52, 56 },
		{ 126, 53, 56 },
		{ 126, 54, 56 },
		{ 126, 55, 56 },
		{ 126, 56, 56 },
		{ 126, 57, 56 },
		{ 126, 58, 56 },
		{ 126, 59, 56 },
		{ 126, 60, 56 },
		{ 126, 61, 56 },
		{ 126, 62, 56 },
		{ 126, 63, 56 },
		{ 126, 64, 56 },
		{ 126, 65, 56 },
		{ 126, 66, 56 },
		{ 126, 67, 56 },
		{ 126, 68, 56 },
		{ 126, 69, 56 },
		{ 126, 70, 56 },
		{ 126, 71, 56 },
		{ 126, 72, 56 },
		{ 126, 73, 56 },
		{ 126, 74, 56 },
		{ 126, 75, 56 },
		{ 126, 76, 56 },
		{ 126, 77, 56 },
		{ 126, 78, 56 },
		{ 126, 79, 56 },
		{ 126, 80, 56 },
		{ 126, 81, 56 },
		{ 126, 82, 56 },
		{ 126, 83, 56 },
		{ 126, 84, 56 },
		{ 126, 85, 56 },
		{ 126, 86, 56 },
		{ 126, 87, 56 },
		{ 126, 88, 56 },
		{ 126, 89, 56 },
		{ 126, 90, 56 },
		{ 126, 91, 56 },
		{ 126, 92, 56 },
		{ 126, 93, 56 },
		{ 126, 94, 56 },
		{ 126, 95, 56 },
		{ 126, 96, 56 },
		{ 126, 97, 56 },
		{ 126, 98, 56 },
		{ 126, 99, 56 },
		{ 126, 100, 56 },
		{ 126, 101, 56 },
		{ 126, 102, 56 },
		{ 126, 103, 56 },
		{ 126, 104, 56 },
		{ 126, 105, 56 },
		{ 126, 106, 56 },
		{ 126, 107, 56 },
		{ 127, 50, 56 },
		{ 127, 51, 56 },
		{ 127, 52, 56 },
		{ 127, 53, 56 },
		{ 127, 54, 56 },
		{ 127, 55, 56 },
		{ 127, 56, 56 },
		{ 127, 57, 56 },
		{ 127, 58, 56 },
		{ 127, 59, 56 },
		{ 127, 60, 56 },
		{ 127, 61, 56 },
		{ 127, 62, 56 },
		{ 127, 63, 56 },
		{ 127, 64, 56 },
		{ 127, 65, 56 },
		{ 127, 66, 56 },
		{ 127, 67, 56 },
		{ 127, 68, 56 },
		{ 127, 69, 56 },
		{ 127, 70, 56 },
		{ 127, 71, 56 },
		{ 127, 72, 56 },
		{ 127, 73, 56 },
		{ 127, 74, 56 },
		{ 127, 75, 56 },
		{ 127, 76, 56 },
		{ 127, 77, 56 },
		{ 127, 78, 56 },
		{ 127, 79, 56 },
		{ 127, 80, 56 },
		{ 127, 81, 56 },
		{ 127, 82, 56 },
		{ 127, 83, 56 },
		{ 127, 84, 56 },
		{ 127, 85, 56 },
		{ 127, 86, 56 },
		{ 127, 87, 56 },
		{ 127, 88, 56 },
		{ 127, 89, 56 },
		{ 127, 90, 56 },
		{ 127, 91, 56 },
		{ 127, 92, 56 },
		{ 127, 93, 56 },
		{ 127, 94, 56 },
		{ 127, 95, 56 },
		{ 127, 96, 56 },
		{ 127, 97, 56 },
		{ 127, 98, 56 },
		{ 127, 99, 56 },
		{ 127, 100, 56 },
		{ 127, 101, 56 },
		{ 127, 102, 56 },
		{ 127, 103, 56 },
		{ 127, 104, 56 },
		{ 127, 105, 56 },
		{ 127, 106, 56 },
		{ 127, 107, 56 },
		{ 127, 108, 56 },
		{ 128, 51, 56 },
		{ 128, 52, 56 },
		{ 128, 53, 56 },
		{ 128, 54, 56 },
		{ 128, 55, 56 },
		{ 128, 56, 56 },
		{ 128, 57, 56 },
		{ 128, 58, 56 },
		{ 128, 59, 56 },
		{ 128, 60, 56 },
		{ 128, 61, 56 },
		{ 128, 62, 56 },
		{ 128, 63, 56 },
		{ 128, 64, 56 },
		{ 128, 65, 56 },
		{ 128, 66, 56 },
		{ 128, 67, 56 },
		{ 128, 68, 56 },
		{ 128, 69, 56 },
		{ 128, 70, 56 },
		{ 128, 71, 56 },
		{ 128, 72, 56 },
		{ 128, 73, 56 },
		{ 128, 74, 56 },
		{ 128, 75, 56 },
		{ 128, 76, 56 },
		{ 128, 77, 56 },
		{ 128, 78, 56 },
		{ 128, 79, 56 },
		{ 128, 80, 56 },
		{ 128, 81, 56 },
		{ 128, 82, 56 },
		{ 128, 83, 56 },
		{ 128, 84, 56 },
		{ 128, 85, 56 },
		{ 128, 86, 56 },
		{ 128, 87, 56 },
		{ 128, 88, 56 },
		{ 128, 89, 56 },
		{ 128, 90, 56 },
		{ 128, 91, 56 },
		{ 128, 92, 56 },
		{ 128, 93, 56 },
		{ 128, 94, 56 },
		{ 128, 95, 56 },
		{ 128, 96, 56 },
		{ 128, 97, 56 },
		{ 128, 98, 56 },
		{ 128, 99, 56 },
		{ 128, 100, 56 },
		{ 128, 101, 56 },
		{ 128, 102, 56 },
		{ 128, 103, 56 },
		{ 128, 104, 56 },
		{ 128, 105, 56 },
		{ 128, 106, 56 },
		{ 128, 107, 56 },
		{ 128, 108, 56 },
		{ 128, 109, 56 },
		{ 129, 53, 56 },
		{ 129, 54, 56 },
		{ 129, 55, 56 },
		{ 129, 56, 56 },
		{ 129, 57, 56 },
		{ 129, 58, 56 },
		{ 129, 59, 56 },
		{ 129, 60, 56 },
		{ 129, 61, 56 },
		{ 129, 62, 56 },
		{ 129, 63, 56 },
		{ 129, 64, 56 },
		{ 129, 65, 56 },
		{ 129, 66, 56 },
		{ 129, 67, 56 },
		{ 129, 68, 56 },
		{ 129, 69, 56 },
		{ 129, 70, 56 },
		{ 129, 71, 56 },
		{ 129, 72, 56 },
		{ 129, 73, 56 },
		{ 129, 74, 56 },
		{ 129, 75, 56 },
		{ 129, 76, 56 },
		{ 129, 77, 56 },
		{ 129, 78, 56 },
		{ 129, 79, 56 },
		{ 129, 80, 56 },
		{ 129, 81, 56 },
		{ 129, 82, 56 },
		{ 129, 83, 56 },
		{ 129, 84, 56 },
		{ 129, 85, 56 },
		{ 129, 86, 56 },
		{ 129, 87, 56 },
		{ 129, 88, 56 },
		{ 129, 89, 56 },
		{ 129, 90, 56 },
		{ 129, 91, 56 },
		{ 129, 92, 56 },
		{ 129, 93, 56 },
		{ 129, 94, 56 },
		{ 129, 95, 56 },
		{ 129, 96, 56 },
		{ 129, 97, 56 },
		{ 129, 98, 56 },
		{ 129, 99, 56 },
		{ 129, 100, 56 },
		{ 129, 101, 56 },
		{ 129, 102, 56 },
		{ 129, 103, 56 },
		{ 129, 104, 56 },
		{ 129, 105, 56 },
		{ 129, 106, 56 },
		{ 129, 107, 56 },
		{ 129, 108, 56 },
		{ 129, 109, 56 },
		{ 129, 110, 56 },
		{ 130, 54, 56 },
		{ 130, 55, 56 },
		{ 130, 56, 56 },
		{ 130, 57, 56 },
		{ 130, 58, 56 },
		{ 130, 59, 56 },
		{ 130, 60, 56 },
		{ 130, 61, 56 },
		{ 130, 62, 56 },
		{ 130, 63, 56 },
		{ 130, 64, 56 },
		{ 130, 65, 56 },
		{ 130, 66, 56 },
		{ 130, 67, 56 },
		{ 130, 68, 56 },
		{ 130, 69, 56 },
		{ 130, 70, 56 },
		{ 130, 71, 56 },
		{ 130, 72, 56 },
		{ 130, 73, 56 },
		{ 130, 74, 56 },
		{ 130, 75, 56 },
		{ 130, 76, 56 },
		{ 130, 77, 56 },
		{ 130, 78, 56 },
		{ 130, 79, 56 },
		{ 130, 80, 56 },
		{ 130, 81, 56 },
		{ 130, 82, 56 },
		{ 130, 83, 56 },
		{ 130, 84, 56 },
		{ 130, 85, 56 },
		{ 130, 86, 56 },
		{ 130, 87, 56 },
		{ 130, 88, 56 },
		{ 130, 89, 56 },
		{ 130, 90, 56 },
		{ 130, 91, 56 },
		{ 130, 92, 56 },
		{ 130, 93, 56 },
		{ 130, 94, 56 },
		{ 130, 95, 56 },
		{ 130, 96, 56 },
		{ 130, 97, 56 },
		{ 130, 98, 56 },
		{ 130, 99, 56 },
		{ 130, 100, 56 },
		{ 130, 101, 56 },
		{ 130, 102, 56 },
		{ 130, 103, 56 },
		{ 130, 104, 56 },
		{ 130, 105, 56 },
		{ 130, 106, 56 },
		{ 130, 107, 56 },
		{ 130, 108, 56 },
		{ 130, 109, 56 },
		{ 130, 110, 56 },
		{ 130, 111, 56 },
		{ 131, 56, 56 },
		{ 131, 57, 56 },
		{ 131, 58, 56 },
		{ 131, 59, 56 },
		{ 131, 60, 56 },
		{ 131, 61, 56 },
		{ 131, 62, 56 },
		{ 131, 63, 56 },
		{ 131, 64, 56 },
		{ 131, 65, 56 },
		{ 131, 66, 56 },
		{ 131, 67, 56 },
		{ 131, 68, 56 },
		{ 131, 69, 56 },
		{ 131, 70, 56 },
		{ 131, 71, 56 },
		{ 131, 72, 56 },
		{ 131, 73, 56 },
		{ 131, 74, 56 },
		{ 131, 75, 56 },
		{ 131, 76, 56 },
		{ 131, 77, 56 },
		{ 131, 78, 56 },
		{ 131, 79, 56 },
		{ 131, 80, 56 },
		{ 131, 81, 56 },
		{ 131, 82, 56 },
		{ 131, 83, 56 },
		{ 131, 84, 56 },
		{ 131, 85, 56 },
		{ 131, 86, 56 },
		{ 131, 87, 56 },
		{ 131, 88, 56 },
		{ 131, 89, 56 },
		{ 131, 90, 56 },
		{ 131, 91, 56 },
		{ 131, 92, 56 },
		{ 131, 93, 56 },
		{ 131, 94, 56 },
		{ 131, 95, 56 },
		{ 131, 96, 56 },
		{ 131, 97, 56 },
		{ 131, 98, 56 },
		{ 131, 99, 56 },
		{ 131, 100, 56 },
		{ 131, 101, 56 },
		{ 131, 102, 56 },
		{ 131, 103, 56 },
		{ 131, 104, 56 },
		{ 131, 105, 56 },
		{ 131, 106, 56 },
		{ 131, 107, 56 },
		{ 131, 108, 56 },
		{ 131, 109, 56 },
		{ 131, 110, 56 },
		{ 131, 111, 56 },
		{ 131, 112, 56 },
		{ 132, 58, 56 },
		{ 132, 59, 56 },
		{ 132, 60, 56 },
		{ 132, 61, 56 },
		{ 132, 62, 56 },
		{ 132, 63, 56 },
		{ 132, 64, 56 },
		{ 132, 65, 56 },
		{ 132, 66, 56 },
		{ 132, 67, 56 },
		{ 132, 68, 56 },
		{ 132, 69, 56 },
		{ 132, 70, 56 },
		{ 132, 71, 56 },
		{ 132, 72, 56 },
		{ 132, 73, 56 },
		{ 132, 74, 56 },
		{ 132, 75, 56 },
		{ 132, 76, 56 },
		{ 132, 77, 56 },
		{ 132, 78, 56 },
		{ 132, 79, 56 },
		{ 132, 80, 56 },
		{ 132, 81, 56 },
		{ 132, 82, 56 },
		{ 132, 83, 56 },
		{ 132, 84, 56 },
		{ 132, 85, 56 },
		{ 132, 86, 56 },
		{ 132, 87, 56 },
		{ 132, 88, 56 },
		{ 132, 89, 56 },
		{ 132, 90, 56 },
		{ 132, 91, 56 },
		{ 132, 92, 56 },
		{ 132, 93, 56 },
		{ 132, 94, 56 },
		{ 132, 95, 56 },
		{ 132, 96, 56 },
		{ 132, 97, 56 },
		{ 132, 98, 56 },
		{ 132, 99, 56 },
		{ 132, 100, 56 },
		{ 132, 101, 56 },
		{ 132, 102, 56 },
		{ 132, 103, 56 },
		{ 132, 104, 56 },
		{ 132, 105, 56 },
		{ 132, 106, 56 },
		{ 132, 107, 56 },
		{ 132, 108, 56 },
		{ 132, 109, 56 },
		{ 132, 110, 56 },
		{ 132, 111, 56 },
		{ 132, 112, 56 },
		{ 132, 113, 56 },
		{ 132, 114, 56 },
		{ 133, 59, 56 },
		{ 133, 60, 56 },
		{ 133, 61, 56 },
		{ 133, 62, 56 },
		{ 133, 63, 56 },
		{ 133, 64, 56 },
		{ 133, 65, 56 },
		{ 133, 66, 56 },
		{ 133, 67, 56 },
		{ 133, 68, 56 },
		{ 133, 69, 56 },
		{ 133, 70, 56 },
		{ 133, 71, 56 },
		{ 133, 72, 56 },
		{ 133, 73, 56 },
		{ 133, 74, 56 },
		{ 133, 75, 56 },
		{ 133, 76, 56 },
		{ 133, 77, 56 },
		{ 133, 78, 56 },
		{ 133, 79, 56 },
		{ 133, 80, 56 },
		{ 133, 81, 56 },
		{ 133, 82, 56 },
		{ 133, 83, 56 },
		{ 133, 84, 56 },
		{ 133, 85, 56 },
		{ 133, 86, 56 },
		{ 133, 87, 56 },
		{ 133, 88, 56 },
		{ 133, 89, 56 },
		{ 133, 90, 56 },
		{ 133, 91, 56 },
		{ 133, 92, 56 },
		{ 133, 93, 56 },
		{ 133, 94, 56 },
		{ 133, 95, 56 },
		{ 133, 96, 56 },
		{ 133, 97, 56 },
		{ 133, 98, 56 },
		{ 133, 99, 56 },
		{ 133, 100, 56 },
		{ 133, 101, 56 },
		{ 133, 102, 56 },
		{ 133, 103, 56 },
		{ 133, 104, 56 },
		{ 133, 105, 56 },
		{ 133, 106, 56 },
		{ 133, 107, 56 },
		{ 133, 108, 56 },
		{ 133, 109, 56 },
		{ 133, 110, 56 },
		{ 133, 111, 56 },
		{ 133, 112, 56 },
		{ 133, 113, 56 },
		{ 133, 114, 56 },
		{ 133, 115, 56 },
		{ 134, 61, 56 },
		{ 134, 62, 56 },
		{ 134, 63, 56 },
		{ 134, 64, 56 },
		{ 134, 65, 56 },
		{ 134, 66, 56 },
		{ 134, 67, 56 },
		{ 134, 68, 56 },
		{ 134, 69, 56 },
		{ 134, 70, 56 },
		{ 134, 71, 56 },
		{ 134, 72, 56 },
		{ 134, 73, 56 },
		{ 134, 74, 56 },
		{ 134, 75, 56 },
		{ 134, 76, 56 },
		{ 134, 77, 56 },
		{ 134, 78, 56 },
		{ 134, 79, 56 },
		{ 134, 80, 56 },
		{ 134, 81, 56 },
		{ 134, 82, 56 },
		{ 134, 83, 56 },
		{ 134, 84, 56 },
		{ 134, 85, 56 },
		{ 134, 86, 56 },
		{ 134, 87, 56 },
		{ 134, 88, 56 },
		{ 134, 89, 56 },
		{ 134, 90, 56 },
		{ 134, 91, 56 },
		{ 134, 92, 56 },
		{ 134, 93, 56 },
		{ 134, 94, 56 },
		{ 134, 95, 56 },
		{ 134, 96, 56 },
		{ 134, 97, 56 },
		{ 134, 98, 56 },
		{ 134, 99, 56 },
		{ 134, 100, 56 },
		{ 134, 101, 56 },
		{ 134, 102, 56 },
		{ 134, 103, 56 },
		{ 134, 104, 56 },
		{ 134, 105, 56 },
		{ 134, 106, 56 },
		{ 134, 107, 56 },
		{ 134, 108, 56 },
		{ 134, 109, 56 },
		{ 134, 110, 56 },
		{ 134, 111, 56 },
		{ 134, 112, 56 },
		{ 134, 113, 56 },
		{ 134, 114, 56 },
		{ 134, 115, 56 },
		{ 134, 116, 56 },
		{ 135, 62, 56 },
		{ 135, 63, 56 },
		{ 135, 64, 56 },
		{ 135, 65, 56 },
		{ 135, 66, 56 },
		{ 135, 67, 56 },
		{ 135, 68, 56 },
		{ 135, 69, 56 },
		{ 135, 70, 56 },
		{ 135, 71, 56 },
		{ 135, 72, 56 },
		{ 135, 73, 56 },
		{ 135, 74, 56 },
		{ 135, 75, 56 },
		{ 135, 76, 56 },
		{ 135, 77, 56 },
		{ 135, 78, 56 },
		{ 135, 79, 56 },
		{ 135, 80, 56 },
		{ 135, 81, 56 },
		{ 135, 82, 56 },
		{ 135, 83, 56 },
		{ 135, 84, 56 },
		{ 135, 85, 56 },
		{ 135, 86, 56 },
		{ 135, 87, 56 },
		{ 135, 88, 56 },
		{ 135, 89, 56 },
		{ 135, 90, 56 },
		{ 135, 91, 56 },
		{ 135, 92, 56 },
		{ 135, 93, 56 },
		{ 135, 94, 56 },
		{ 135, 95, 56 },
		{ 135, 96, 56 },
		{ 135, 97, 56 },
		{ 135, 98, 56 },
		{ 135, 99, 56 },
		{ 135, 100, 56 },
		{ 135, 101, 56 },
		{ 135, 102, 56 },
		{ 135, 103, 56 },
		{ 135, 104, 56 },
		{ 135, 105, 56 },
		{ 135, 106, 56 },
		{ 135, 107, 56 },
		{ 135, 108, 56 },
		{ 135, 109, 56 },
		{ 135, 110, 56 },
		{ 135, 111, 56 },
		{ 135, 112, 56 },
		{ 135, 113, 56 },
		{ 135, 114, 56 },
		{ 135, 115, 56 },
		{ 135, 116, 56 },
		{ 135, 117, 56 },
		{ 136, 64, 56 },
		{ 136, 65, 56 },
		{ 136, 66, 56 },
		{ 136, 67, 56 },
		{ 136, 68, 56 },
		{ 136, 69, 56 },
		{ 136, 70, 56 },
		{ 136, 71, 56 },
		{ 136, 72, 56 },
		{ 136, 73, 56 },
		{ 136, 74, 56 },
		{ 136, 75, 56 },
		{ 136, 76, 56 },
		{ 136, 77, 56 },
		{ 136, 78, 56 },
		{ 136, 79, 56 },
		{ 136, 80, 56 },
		{ 136, 81, 56 },
		{ 136, 82, 56 },
		{ 136, 83, 56 },
		{ 136, 84, 56 },
		{ 136, 85, 56 },
		{ 136, 86, 56 },
		{ 136, 87, 56 },
		{ 136, 88, 56 },
		{ 136, 89, 56 },
		{ 136, 90, 56 },
		{ 136, 91, 56 },
		{ 136, 92, 56 },
		{ 136, 93, 56 },
		{ 136, 94, 56 },
		{ 136, 95, 56 },
		{ 136, 96, 56 },
		{ 136, 97, 56 },
		{ 136, 98, 56 },
		{ 136, 99, 56 },
		{ 136, 100, 56 },
		{ 136, 101, 56 },
		{ 136, 102, 56 },
		{ 136, 103, 56 },
		{ 136, 104, 56 },
		{ 136, 105, 56 },
		{ 136, 106, 56 },
		{ 136, 107, 56 },
		{ 136, 108, 56 },
		{ 136, 109, 56 },
		{ 136, 110, 56 },
		{ 136, 111, 56 },
		{ 136, 112, 56 },
		{ 136, 113, 56 },
		{ 136, 114, 56 },
		{ 136, 115, 56 },
		{ 136, 116, 56 },
		{ 136, 117, 56 },
		{ 136, 118, 56 },
		{ 137, 66, 56 },
		{ 137, 67, 56 },
		{ 137, 68, 56 },
		{ 137, 69, 56 },
		{ 137, 70, 56 },
		{ 137, 71, 56 },
		{ 137, 72, 56 },
		{ 137, 73, 56 },
		{ 137, 74, 56 },
		{ 137, 75, 56 },
		{ 137, 76, 56 },
		{ 137, 77, 56 },
		{ 137, 78, 56 },
		{ 137, 79, 56 },
		{ 137, 80, 56 },
		{ 137, 81, 56 },
		{ 137, 82, 56 },
		{ 137, 83, 56 },
		{ 137, 84, 56 },
		{ 137, 85, 56 },
		{ 137, 86, 56 },
		{ 137, 87, 56 },
		{ 137, 88, 56 },
		{ 137, 89, 56 },
		{ 137, 90, 56 },
		{ 137, 91, 56 },
		{ 137, 92, 56 },
		{ 137, 93, 56 },
		{ 137, 94, 56 },
		{ 137, 95, 56 },
		{ 137, 96, 56 },
		{ 137, 97, 56 },
		{ 137, 98, 56 },
		{ 137, 99, 56 },
		{ 137, 100, 56 },
		{ 137, 101, 56 },
		{ 137, 102, 56 },
		{ 137, 103, 56 },
		{ 137, 104, 56 },
		{ 137, 105, 56 },
		{ 137, 106, 56 },
		{ 137, 107, 56 },
		{ 137, 108, 56 },
		{ 137, 109, 56 },
		{ 137, 110, 56 },
		{ 137, 111, 56 },
		{ 137, 112, 56 },
		{ 137, 113, 56 },
		{ 137, 114, 56 },
		{ 137, 115, 56 },
		{ 137, 116, 56 },
		{ 137, 117, 56 },
		{ 137, 118, 56 },
		{ 137, 119, 56 },
		{ 138, 67, 56 },
		{ 138, 68, 56 },
		{ 138, 69, 56 },
		{ 138, 70, 56 },
		{ 138, 71, 56 },
		{ 138, 72, 56 },
		{ 138, 73, 56 },
		{ 138, 74, 56 },
		{ 138, 75, 56 },
		{ 138, 76, 56 },
		{ 138, 77, 56 },
		{ 138, 78, 56 },
		{ 138, 79, 56 },
		{ 138, 80, 56 },
		{ 138, 81, 56 },
		{ 138, 82, 56 },
		{ 138, 83, 56 },
		{ 138, 84, 56 },
		{ 138, 85, 56 },
		{ 138, 86, 56 },
		{ 138, 87, 56 },
		{ 138, 88, 56 },
		{ 138, 89, 56 },
		{ 138, 90, 56 },
		{ 138, 91, 56 },
		{ 138, 92, 56 },
		{ 138, 93, 56 },
		{ 138, 94, 56 },
		{ 138, 95, 56 },
		{ 138, 96, 56 },
		{ 138, 97, 56 },
		{ 138, 98, 56 },
		{ 138, 99, 56 },
		{ 138, 100, 56 },
		{ 138, 101, 56 },
		{ 138, 102, 56 },
		{ 138, 103, 56 },
		{ 138, 104, 56 },
		{ 138, 105, 56 },
		{ 138, 106, 56 },
		{ 138, 107, 56 },
		{ 138, 108, 56 },
		{ 138, 109, 56 },
		{ 138, 110, 56 },
		{ 138, 111, 56 },
		{ 138, 112, 56 },
		{ 138, 113, 56 },
		{ 138, 114, 56 },
		{ 138, 115, 56 },
		{ 138, 116, 56 },
		{ 138, 117, 56 },
		{ 138, 118, 56 },
		{ 138, 119, 56 },
		{ 138, 120, 56 },
		{ 139, 69, 56 },
		{ 139, 70, 56 },
		{ 139, 71, 56 },
		{ 139, 72, 56 },
		{ 139, 73, 56 },
		{ 139, 74, 56 },
		{ 139, 75, 56 },
		{ 139, 76, 56 },
		{ 139, 77, 56 },
		{ 139, 78, 56 },
		{ 139, 79, 56 },
		{ 139, 80, 56 },
		{ 139, 81, 56 },
		{ 139, 82, 56 },
		{ 139, 83, 56 },
		{ 139, 84, 56 },
		{ 139, 85, 56 },
		{ 139, 86, 56 },
		{ 139, 87, 56 },
		{ 139, 88, 56 },
		{ 139, 89, 56 },
		{ 139, 90, 56 },
		{ 139, 91, 56 },
		{ 139, 92, 56 },
		{ 139, 93, 56 },
		{ 139, 94, 56 },
		{ 139, 95, 56 },
		{ 139, 96, 56 },
		{ 139, 97, 56 },
		{ 139, 98, 56 },
		{ 139, 99, 56 },
		{ 139, 100, 56 },
		{ 139, 101, 56 },
		{ 139, 102, 56 },
		{ 139, 103, 56 },
		{ 139, 104, 56 },
		{ 139, 105, 56 },
		{ 139, 106, 56 },
		{ 139, 107, 56 },
		{ 139, 108, 56 },
		{ 139, 109, 56 },
		{ 139, 110, 56 },
		{ 139, 111, 56 },
		{ 139, 112, 56 },
		{ 139, 113, 56 },
		{ 139, 114, 56 },
		{ 139, 115, 56 },
		{ 139, 116, 56 },
		{ 139, 117, 56 },
		{ 139, 118, 56 },
		{ 139, 119, 56 },
		{ 139, 120, 56 },
		{ 139, 121, 56 },
		{ 139, 122, 56 },
		{ 140, 70, 56 },
		{ 140, 71, 56 },
		{ 140, 72, 56 },
		{ 140, 73, 56 },
		{ 140, 74, 56 },
		{ 140, 75, 56 },
		{ 140, 76, 56 },
		{ 140, 77, 56 },
		{ 140, 78, 56 },
		{ 140, 79, 56 },
		{ 140, 80, 56 },
		{ 140, 81, 56 },
		{ 140, 82, 56 },
		{ 140, 83, 56 },
		{ 140, 84, 56 },
		{ 140, 85, 56 },
		{ 140, 86, 56 },
		{ 140, 87, 56 },
		{ 140, 88, 56 },
		{ 140, 89, 56 },
		{ 140, 90, 56 },
		{ 140, 91, 56 },
		{ 140, 92, 56 },
		{ 140, 93, 56 },
		{ 140, 94, 56 },
		{ 140, 95, 56 },
		{ 140, 96, 56 },
		{ 140, 97, 56 },
		{ 140, 98, 56 },
		{ 140, 99, 56 },
		{ 140, 100, 56 },
		{ 140, 101, 56 },
		{ 140, 102, 56 },
		{ 140, 103, 56 },
		{ 140, 104, 56 },
		{ 140, 105, 56 },
		{ 140, 106, 56 },
		{ 140, 107, 56 },
		{ 140, 108, 56 },
		{ 140, 109, 56 },
		{ 140, 110, 56 },
		{ 140, 111, 56 },
		{ 140, 112, 56 },
		{ 140, 113, 56 },
		{ 140, 114, 56 },
		{ 140, 115, 56 },
		{ 140, 116, 56 },
		{ 140, 117, 56 },
		{ 140, 118, 56 },
		{ 140, 119, 56 },
		{ 140, 120, 56 },
		{ 140, 121, 56 },
		{ 140, 122, 56 },
		{ 140, 123, 56 },
		{ 141, 72, 56 },
		{ 141, 73, 56 },
		{ 141, 74, 56 },
		{ 141, 75, 56 },
		{ 141, 76, 56 },
		{ 141, 77, 56 },
		{ 141, 78, 56 },
		{ 141, 79, 56 },
		{ 141, 80, 56 },
		{ 141, 81, 56 },
		{ 141, 82, 56 },
		{ 141, 83, 56 },
		{ 141, 84, 56 },
		{ 141, 85, 56 },
		{ 141, 86, 56 },
		{ 141, 87, 56 },
		{ 141, 88, 56 },
		{ 141, 89, 56 },
		{ 141, 90, 56 },
		{ 141, 91, 56 },
		{ 141, 92, 56 },
		{ 141, 93, 56 },
		{ 141, 94, 56 },
		{ 141, 95, 56 },
		{ 141, 96, 56 },
		{ 141, 97, 56 },
		{ 141, 98, 56 },
		{ 141, 99, 56 },
		{ 141, 100, 56 },
		{ 141, 101, 56 },
		{ 141, 102, 56 },
		{ 141, 103, 56 },
		{ 141, 104, 56 },
		{ 141, 105, 56 },
		{ 141, 106, 56 },
		{ 141, 107, 56 },
		{ 141, 108, 56 },
		{ 141, 109, 56 },
		{ 141, 110, 56 },
		{ 141, 111, 56 },
		{ 141, 112, 56 },
		{ 141, 113, 56 },
		{ 141, 114, 56 },
		{ 141, 115, 56 },
		{ 141, 116, 56 },
		{ 141, 117, 56 },
		{ 141, 118, 56 },
		{ 141, 119, 56 },
		{ 141, 120, 56 },
		{ 141, 121, 56 },
		{ 141, 122, 56 },
		{ 141, 123, 56 },
		{ 141, 124, 56 },
		{ 142, 74, 56 },
		{ 142, 75, 56 },
		{ 142, 76, 56 },
		{ 142, 77, 56 },
		{ 142, 78, 56 },
		{ 142, 79, 56 },
		{ 142, 80, 56 },
		{ 142, 81, 56 },
		{ 142, 82, 56 },
		{ 142, 83, 56 },
		{ 142, 84, 56 },
		{ 142, 85, 56 },
		{ 142, 86, 56 },
		{ 142, 87, 56 },
		{ 142, 88, 56 },
		{ 142, 89, 56 },
		{ 142, 90, 56 },
		{ 142, 91, 56 },
		{ 142, 92, 56 },
		{ 142, 93, 56 },
		{ 142, 94, 56 },
		{ 142, 95, 56 },
		{ 142, 96, 56 },
		{ 142, 97, 56 },
		{ 142, 98, 56 },
		{ 142, 99, 56 },
		{ 142, 100, 56 },
		{ 142, 101, 56 },
		{ 142, 102, 56 },
		{ 142, 103, 56 },
		{ 142, 104, 56 },
		{ 142, 105, 56 },
		{ 142, 106, 56 },
		{ 142, 107, 56 },
		{ 142, 108, 56 },
		{ 142, 109, 56 },
		{ 142, 110, 56 },
		{ 142, 111, 56 },
		{ 142, 112, 56 },
		{ 142, 113, 56 },
		{ 142, 114, 56 },
		{ 142, 115, 56 },
		{ 142, 116, 56 },
		{ 142, 117, 56 },
		{ 142, 118, 56 },
		{ 142, 119, 56 },
		{ 142, 120, 56 },
		{ 142, 121, 56 },
		{ 142, 122, 56 },
		{ 142, 123, 56 },
		{ 142, 124, 56 },
		{ 142, 125, 56 },
		{ 143, 75, 56 },
		{ 143, 76, 56 },
		{ 143, 77, 56 },
		{ 143, 78, 56 },
		{ 143, 79, 56 },
		{ 143, 80, 56 },
		{ 143, 81, 56 },
		{ 143, 82, 56 },
		{ 143, 83, 56 },
		{ 143, 84, 56 },
		{ 143, 85, 56 },
		{ 143, 86, 56 },
		{ 143, 87, 56 },
		{ 143, 88, 56 },
		{ 143, 89, 56 },
		{ 143, 90, 56 },
		{ 143, 91, 56 },
		{ 143, 92, 56 },
		{ 143, 93, 56 },
		{ 143, 94, 56 },
		{ 143, 95, 56 },
		{ 143, 96, 56 },
		{ 143, 97, 56 },
		{ 143, 98, 56 },
		{ 143, 99, 56 },
		{ 143, 100, 56 },
		{ 143, 101, 56 },
		{ 143, 102, 56 },
		{ 143, 103, 56 },
		{ 143, 104, 56 },
		{ 143, 105, 56 },
		{ 143, 106, 56 },
		{ 143, 107, 56 },
		{ 143, 108, 56 },
		{ 143, 109, 56 },
		{ 143, 110, 56 },
		{ 143, 111, 56 },
		{ 143, 112, 56 },
		{ 143, 113, 56 },
		{ 143, 114, 56 },
		{ 143, 115, 56 },
		{ 143, 116, 56 },
		{ 143, 117, 56 },
		{ 143, 118, 56 },
		{ 143, 119, 56 },
		{ 143, 120, 56 },
		{ 143, 121, 56 },
		{ 143, 122, 56 },
		{ 143, 123, 56 },
		{ 143, 124, 56 },
		{ 143, 125, 56 },
		{ 143, 126, 56 },
		{ 144, 77, 56 },
		{ 144, 78, 56 },
		{ 144, 79, 56 },
		{ 144, 80, 56 },
		{ 144, 81, 56 },
		{ 144, 82, 56 },
		{ 144, 83, 56 },
		{ 144, 84, 56 },
		{ 144, 85, 56 },
		{ 144, 86, 56 },
		{ 144, 87, 56 },
		{ 144, 88, 56 },
		{ 144, 89, 56 },
		{ 144, 90, 56 },
		{ 144, 91, 56 },
		{ 144, 92, 56 },
		{ 144, 93, 56 },
		{ 144, 94, 56 },
		{ 144, 95, 56 },
		{ 144, 96, 56 },
		{ 144, 97, 56 },
		{ 144, 98, 56 },
		{ 144, 99, 56 },
		{ 144, 100, 56 },
		{ 144, 101, 56 },
		{ 144, 102, 56 },
		{ 144, 103, 56 },
		{ 144, 104, 56 },
		{ 144, 105, 56 },
		{ 144, 106, 56 },
		{ 144, 107, 56 },
		{ 144, 108, 56 },
		{ 144, 109, 56 },
		{ 144, 110, 56 },
		{ 144, 111, 56 },
		{ 144, 112, 56 },
		{ 144, 113, 56 },
		{ 144, 114, 56 },
		{ 144, 115, 56 },
		{ 144, 116, 56 },
		{ 144, 117, 56 },
		{ 144, 118, 56 },
		{ 144, 119, 56 },
		{ 144, 120, 56 },
		{ 144, 121, 56 },
		{ 144, 122, 56 },
		{ 144, 123, 56 },
		{ 144, 124, 56 },
		{ 144, 125, 56 },
		{ 144, 126, 56 },
		{ 144, 127, 56 },
		{ 145, 78, 56 },
		{ 145, 79, 56 },
		{ 145, 80, 56 },
		{ 145, 81, 56 },
		{ 145, 82, 56 },
		{ 145, 83, 56 },
		{ 145, 84, 56 },
		{ 145, 85, 56 },
		{ 145, 86, 56 },
		{ 145, 87, 56 },
		{ 145, 88, 56 },
		{ 145, 89, 56 },
		{ 145, 90, 56 },
		{ 145, 91, 56 },
		{ 145, 92, 56 },
		{ 145, 93, 56 },
		{ 145, 94, 56 },
		{ 145, 95, 56 },
		{ 145, 96, 56 },
		{ 145, 97, 56 },
		{ 145, 98, 56 },
		{ 145, 99, 56 },
		{ 145, 100, 56 },
		{ 145, 101, 56 },
		{ 145, 102, 56 },
		{ 145, 103, 56 },
		{ 145, 104, 56 },
		{ 145, 105, 56 },
		{ 145, 106, 56 },
		{ 145, 107, 56 },
		{ 145, 108, 56 },
		{ 145, 109, 56 },
		{ 145, 110, 56 },
		{ 145, 111, 56 },
		{ 145, 112, 56 },
		{ 145, 113, 56 },
		{ 145, 114, 56 },
		{ 145, 115, 56 },
		{ 145, 116, 56 },
		{ 145, 117, 56 },
		{ 145, 118, 56 },
		{ 145, 119, 56 },
		{ 145, 120, 56 },
		{ 145, 121, 56 },
		{ 145, 122, 56 },
		{ 145, 123, 56 },
		{ 145, 124, 56 },
		{ 145, 125, 56 },
		{ 145, 126, 56 },
		{ 145, 127, 56 },
		{ 145, 128, 56 },
		{ 146, 80, 56 },
		{ 146, 81, 56 },
		{ 146, 82, 56 },
		{ 146, 83, 56 },
		{ 146, 84, 56 },
		{ 146, 85, 56 },
		{ 146, 86, 56 },
		{ 146, 87, 56 },
		{ 146, 88, 56 },
		{ 146, 89, 56 },
		{ 146, 90, 56 },
		{ 146, 91, 56 },
		{ 146, 92, 56 },
		{ 146, 93, 56 },
		{ 146, 94, 56 },
		{ 146, 95, 56 },
		{ 146, 96, 56 },
		{ 146, 97, 56 },
		{ 146, 98, 56 },
		{ 146, 99, 56 },
		{ 146, 100, 56 },
		{ 146, 101, 56 },
		{ 146, 102, 56 },
		{ 146, 103, 56 },
		{ 146, 104, 56 },
		{ 146, 105, 56 },
		{ 146, 106, 56 },
		{ 146, 107, 56 },
		{ 146, 108, 56 },
		{ 146, 109, 56 },
		{ 146, 110, 56 },
		{ 146, 111, 56 },
		{ 146, 112, 56 },
		{ 146, 113, 56 },
		{ 146, 114, 56 },
		{ 146, 115, 56 },
		{ 146, 116, 56 },
		{ 146, 117, 56 },
		{ 146, 118, 56 },
		{ 146, 119, 56 },
		{ 146, 120, 56 },
		{ 146, 121, 56 },
		{ 146, 122, 56 },
		{ 146, 123, 56 },
		{ 146, 124, 56 },
		{ 146, 125, 56 },
		{ 146, 126, 56 },
		{ 146, 127, 56 },
		{ 146, 128, 56 },
		{ 146, 129, 56 },
		{ 146, 130, 56 },
		{ 147, 82, 56 },
		{ 147, 83, 56 },
		{ 147, 84, 56 },
		{ 147, 85, 56 },
		{ 147, 86, 56 },
		{ 147, 87, 56 },
		{ 147, 88, 56 },
		{ 147, 89, 56 },
		{ 147, 90, 56 },
		{ 147, 91, 56 },
		{ 147, 92, 56 },
		{ 147, 93, 56 },
		{ 147, 94, 56 },
		{ 147, 95, 56 },
		{ 147, 96, 56 },
		{ 147, 97, 56 },
		{ 147, 98, 56 },
		{ 147, 99, 56 },
		{ 147, 100, 56 },
		{ 147, 101, 56 },
		{ 147, 102, 56 },
		{ 147, 103, 56 },
		{ 147, 104, 56 },
		{ 147, 105, 56 },
		{ 147, 106, 56 },
		{ 147, 107, 56 },
		{ 147, 108, 56 },
		{ 147, 109, 56 },
		{ 147, 110, 56 },
		{ 147, 111, 56 },
		{ 147, 112, 56 },
		{ 147, 113, 56 },
		{ 147, 114, 56 },
		{ 147, 115, 56 },
		{ 147, 116, 56 },
		{ 147, 117, 56 },
		{ 147, 118, 56 },
		{ 147, 119, 56 },
		{ 147, 120, 56 },
		{ 147, 121, 56 },
		{ 147, 122, 56 },
		{ 147, 123, 56 },
		{ 147, 124, 56 },
		{ 147, 125, 56 },
		{ 147, 126, 56 },
		{ 147, 127, 56 },
		{ 147, 128, 56 },
		{ 147, 129, 56 },
		{ 147, 130, 56 },
		{ 147, 131, 56 },
		{ 148, 83, 56 },
		{ 148, 84, 56 },
		{ 148, 85, 56 },
		{ 148, 86, 56 },
		{ 148, 87, 56 },
		{ 148, 88, 56 },
		{ 148, 89, 56 },
		{ 148, 90, 56 },
		{ 148, 91, 56 },
		{ 148, 92, 56 },
		{ 148, 93, 56 },
		{ 148, 94, 56 },
		{ 148, 95, 56 },
		{ 148, 96, 56 },
		{ 148, 97, 56 },
		{ 148, 98, 56 },
		{ 148, 99, 56 },
		{ 148, 100, 56 },
		{ 148, 101, 56 },
		{ 148, 102, 56 },
		{ 148, 103, 56 },
		{ 148, 104, 56 },
		{ 148, 105, 56 },
		{ 148, 106, 56 },
		{ 148, 107, 56 },
		{ 148, 108, 56 },
		{ 148, 109, 56 },
		{ 148, 110, 56 },
		{ 148, 111, 56 },
		{ 148, 112, 56 },
		{ 148, 113, 56 },
		{ 148, 114, 56 },
		{ 148, 115, 56 },
		{ 148, 116, 56 },
		{ 148, 117, 56 },
		{ 148, 118, 56 },
		{ 148, 119, 56 },
		{ 148, 120, 56 },
		{ 148, 121, 56 },
		{ 148, 122, 56 },
		{ 148, 123, 56 },
		{ 148, 124, 56 },
		{ 148, 125, 56 },
		{ 148, 126, 56 },
		{ 148, 127, 56 },
		{ 148, 128, 56 },
		{ 148, 129, 56 },
		{ 148, 130, 56 },
		{ 148, 131, 56 },
		{ 148, 132, 56 },
		{ 149, 85, 56 },
		{ 149, 86, 56 },
		{ 149, 87, 56 },
		{ 149, 88, 56 },
		{ 149, 89, 56 },
		{ 149, 90, 56 },
		{ 149, 91, 56 },
		{ 149, 92, 56 },
		{ 149, 93, 56 },
		{ 149, 94, 56 },
		{ 149, 95, 56 },
		{ 149, 96, 56 },
		{ 149, 97, 56 },
		{ 149, 98, 56 },
		{ 149, 99, 56 },
		{ 149, 100, 56 },
		{ 149, 101, 56 },
		{ 149, 102, 56 },
		{ 149, 103, 56 },
		{ 149, 104, 56 },
		{ 149, 105, 56 },
		{ 149, 106, 56 },
		{ 149, 107, 56 },
		{ 149, 108, 56 },
		{ 149, 109, 56 },
		{ 149, 110, 56 },
		{ 149, 111, 56 },
		{ 149, 112, 56 },
		{ 149, 113, 56 },
		{ 149, 114, 56 },
		{ 149, 115, 56 },
		{ 149, 116, 56 },
		{ 149, 117, 56 },
		{ 149, 118, 56 },
		{ 149, 119, 56 },
		{ 149, 120, 56 },
		{ 149, 121, 56 },
		{ 149, 122, 56 },
		{ 149, 123, 56 },
		{ 149, 124, 56 },
		{ 149, 125, 56 },
		{ 149, 126, 56 },
		{ 149, 127, 56 },
		{ 149, 128, 56 },
		{ 149, 129, 56 },
		{ 149, 130, 56 },
		{ 149, 131, 56 },
		{ 149, 132, 56 },
		{ 149, 133, 56 },
		{ 150, 86, 56 },
		{ 150, 87, 56 },
		{ 150, 88, 56 },
		{ 150, 89, 56 },
		{ 150, 90, 56 },
		{ 150, 91, 56 },
		{ 150, 92, 56 },
		{ 150, 93, 56 },
		{ 150, 94, 56 },
		{ 150, 95, 56 },
		{ 150, 96, 56 },
		{ 150, 97, 56 },
		{ 150, 98, 56 },
		{ 150, 99, 56 },
		{ 150, 100, 56 },
		{ 150, 101, 56 },
		{ 150, 102, 56 },
		{ 150, 103, 56 },
		{ 150, 104, 56 },
		{ 150, 105, 56 },
		{ 150, 106, 56 },
		{ 150, 107, 56 },
		{ 150, 108, 56 },
		{ 150, 109, 56 },
		{ 150, 110, 56 },
		{ 150, 111, 56 },
		{ 150, 112, 56 },
		{ 150, 113, 56 },
		{ 150, 114, 56 },
		{ 150, 115, 56 },
		{ 150, 116, 56 },
		{ 150, 117, 56 },
		{ 150, 118, 56 },
		{ 150, 119, 56 },
		{ 150, 120, 56 },
		{ 150, 121, 56 },
		{ 150, 122, 56 },
		{ 150, 123, 56 },
		{ 150, 124, 56 },
		{ 150, 125, 56 },
		{ 150, 126, 56 },
		{ 150, 127, 56 },
		{ 150, 128, 56 },
		{ 150, 129, 56 },
		{ 150, 130, 56 },
		{ 150, 131, 56 },
		{ 150, 132, 56 },
		{ 150, 133, 56 },
		{ 150, 134, 56 },
		{ 151, 88, 56 },
		{ 151, 89, 56 },
		{ 151, 90, 56 },
		{ 151, 91, 56 },
		{ 151, 92, 56 },
		{ 151, 93, 56 },
		{ 151, 94, 56 },
		{ 151, 95, 56 },
		{ 151, 96, 56 },
		{ 151, 97, 56 },
		{ 151, 98, 56 },
		{ 151, 99, 56 },
		{ 151, 100, 56 },
		{ 151, 101, 56 },
		{ 151, 102, 56 },
		{ 151, 103, 56 },
		{ 151, 104, 56 },
		{ 151, 105, 56 },
		{ 151, 106, 56 },
		{ 151, 107, 56 },
		{ 151, 108, 56 },
		{ 151, 109, 56 },
		{ 151, 110, 56 },
		{ 151, 111, 56 },
		{ 151, 112, 56 },
		{ 151, 113, 56 },
		{ 151, 114, 56 },
		{ 151, 115, 56 },
		{ 151, 116, 56 },
		{ 151, 117, 56 },
		{ 151, 118, 56 },
		{ 151, 119, 56 },
		{ 151, 120, 56 },
		{ 151, 121, 56 },
		{ 151, 122, 56 },
		{ 151, 123, 56 },
		{ 151, 124, 56 },
		{ 151, 125, 56 },
		{ 151, 126, 56 },
		{ 151, 127, 56 },
		{ 151, 128, 56 },
		{ 151, 129, 56 },
		{ 151, 130, 56 },
		{ 151, 131, 56 },
		{ 151, 132, 56 },
		{ 151, 133, 56 },
		{ 151, 134, 56 },
		{ 151, 135, 56 },
		{ 152, 90, 56 },
		{ 152, 91, 56 },
		{ 152, 92, 56 },
		{ 152, 93, 56 },
		{ 152, 94, 56 },
		{ 152, 95, 56 },
		{ 152, 96, 56 },
		{ 152, 97, 56 },
		{ 152, 98, 56 },
		{ 152, 99, 56 },
		{ 152, 100, 56 },
		{ 152, 101, 56 },
		{ 152, 102, 56 },
		{ 152, 103, 56 },
		{ 152, 104, 56 },
		{ 152, 105, 56 },
		{ 152, 106, 56 },
		{ 152, 107, 56 },
		{ 152, 108, 56 },
		{ 152, 109, 56 },
		{ 152, 110, 56 },
		{ 152, 111, 56 },
		{ 152, 112, 56 },
		{ 152, 113, 56 },
		{ 152, 114, 56 },
		{ 152, 115, 56 },
		{ 152, 116, 56 },
		{ 152, 117, 56 },
		{ 152, 118, 56 },
		{ 152, 119, 56 },
		{ 152, 120, 56 },
		{ 152, 121, 56 },
		{ 152, 122, 56 },
		{ 152, 123, 56 },
		{ 152, 124, 56 },
		{ 152, 125, 56 },
		{ 152, 126, 56 },
		{ 152, 127, 56 },
		{ 152, 128, 56 },
		{ 152, 129, 56 },
		{ 152, 130, 56 },
		{ 152, 131, 56 },
		{ 152, 132, 56 },
		{ 152, 133, 56 },
		{ 152, 134, 56 },
		{ 152, 135, 56 },
		{ 152, 136, 56 },
		{ 153, 91, 56 },
		{ 153, 92, 56 },
		{ 153, 93, 56 },
		{ 153, 94, 56 },
		{ 153, 95, 56 },
		{ 153, 96, 56 },
		{ 153, 97, 56 },
		{ 153, 98, 56 },
		{ 153, 99, 56 },
		{ 153, 100, 56 },
		{ 153, 101, 56 },
		{ 153, 102, 56 },
		{ 153, 103, 56 },
		{ 153, 104, 56 },
		{ 153, 105, 56 },
		{ 153, 106, 56 },
		{ 153, 107, 56 },
		{ 153, 108, 56 },
		{ 153, 109, 56 },
		{ 153, 110, 56 },
		{ 153, 111, 56 },
		{ 153, 112, 56 },
		{ 153, 113, 56 },
		{ 153, 114, 56 },
		{ 153, 115, 56 },
		{ 153, 116, 56 },
		{ 153, 117, 56 },
		{ 153, 118, 56 },
		{ 153, 119, 56 },
		{ 153, 120, 56 },
		{ 153, 121, 56 },
		{ 153, 122, 56 },
		{ 153, 123, 56 },
		{ 153, 124, 56 },
		{ 153, 125, 56 },
		{ 153, 126, 56 },
		{ 153, 127, 56 },
		{ 153, 128, 56 },
		{ 153, 129, 56 },
		{ 153, 130, 56 },
		{ 153, 131, 56 },
		{ 153, 132, 56 },
		{ 153, 133, 56 },
		{ 153, 134, 56 },
		{ 153, 135, 56 },
		{ 153, 136, 56 },
		{ 153, 137, 56 },
		{ 153, 138, 56 },
		{ 154, 93, 56 },
		{ 154, 94, 56 },
		{ 154, 95, 56 },
		{ 154, 96, 56 },
		{ 154, 97, 56 },
		{ 154, 98, 56 },
		{ 154, 99, 56 },
		{ 154, 100, 56 },
		{ 154, 101, 56 },
		{ 154, 102, 56 },
		{ 154, 103, 56 },
		{ 154, 104, 56 },
		{ 154, 105, 56 },
		{ 154, 106, 56 },
		{ 154, 107, 56 },
		{ 154, 108, 56 },
		{ 154, 109, 56 },
		{ 154, 110, 56 },
		{ 154, 111, 56 },
		{ 154, 112, 56 },
		{ 154, 113, 56 },
		{ 154, 114, 56 },
		{ 154, 115, 56 },
		{ 154, 116, 56 },
		{ 154, 117, 56 },
		{ 154, 118, 56 },
		{ 154, 119, 56 },
		{ 154, 120, 56 },
		{ 154, 121, 56 },
		{ 154, 122, 56 },
		{ 154, 123, 56 },
		{ 154, 124, 56 },
		{ 154, 125, 56 },
		{ 154, 126, 56 },
		{ 154, 127, 56 },
		{ 154, 128, 56 },
		{ 154, 129, 56 },
		{ 154, 130, 56 },
		{ 154, 131, 56 },
		{ 154, 132, 56 },
		{ 154, 133, 56 },
		{ 154, 134, 56 },
		{ 154, 135, 56 },
		{ 154, 136, 56 },
		{ 154, 137, 56 },
		{ 154, 138, 56 },
		{ 154, 139, 56 },
		{ 155, 94, 56 },
		{ 155, 95, 56 },
		{ 155, 96, 56 },
		{ 155, 97, 56 },
		{ 155, 98, 56 },
		{ 155, 99, 56 },
		{ 155, 100, 56 },
		{ 155, 101, 56 },
		{ 155, 102, 56 },
		{ 155, 103, 56 },
		{ 155, 104, 56 },
		{ 155, 105, 56 },
		{ 155, 106, 56 },
		{ 155, 107, 56 },
		{ 155, 108, 56 },
		{ 155, 109, 56 },
		{ 155, 110, 56 },
		{ 155, 111, 56 },
		{ 155, 112, 56 },
		{ 155, 113, 56 },
		{ 155, 114, 56 },
		{ 155, 115, 56 },
		{ 155, 116, 56 },
		{ 155, 117, 56 },
		{ 155, 118, 56 },
		{ 155, 119, 56 },
		{ 155, 120, 56 },
		{ 155, 121, 56 },
		{ 155, 122, 56 },
		{ 155, 123, 56 },
		{ 155, 124, 56 },
		{ 155, 125, 56 },
		{ 155, 126, 56 },
		{ 155, 127, 56 },
		{ 155, 128, 56 },
		{ 155, 129, 56 },
		{ 155, 130, 56 },
		{ 155, 131, 56 },
		{ 155, 132, 56 },
		{ 155, 133, 56 },
		{ 155, 134, 56 },
		{ 155, 135, 56 },
		{ 155, 136, 56 },
		{ 155, 137, 56 },
		{ 155, 138, 56 },
		{ 155, 139, 56 },
		{ 155, 140, 56 },
		{ 156, 96, 56 },
		{ 156, 97, 56 },
		{ 156, 98, 56 },
		{ 156, 99, 56 },
		{ 156, 100, 56 },
		{ 156, 101, 56 },
		{ 156, 102, 56 },
		{ 156, 103, 56 },
		{ 156, 104, 56 },
		{ 156, 105, 56 },
		{ 156, 106, 56 },
		{ 156, 107, 56 },
		{ 156, 108, 56 },
		{ 156, 109, 56 },
		{ 156, 110, 56 },
		{ 156, 111, 56 },
		{ 156, 112, 56 },
		{ 156, 113, 56 },
		{ 156, 114, 56 },
		{ 156, 115, 56 },
		{ 156, 116, 56 },
		{ 156, 117, 56 },
		{ 156, 118, 56 },
		{ 156, 119, 56 },
		{ 156, 120, 56 },
		{ 156, 121, 56 },
		{ 156, 122, 56 },
		{ 156, 123, 56 },
		{ 156, 124, 56 },
		{ 156, 125, 56 },
		{ 156, 126, 56 },
		{ 156, 127, 56 },
		{ 156, 128, 56 },
		{ 156, 129, 56 },
		{ 156, 130, 56 },
		{ 156, 131, 56 },
		{ 156, 132, 56 },
		{ 156, 133, 56 },
		{ 156, 134, 56 },
		{ 156, 135, 56 },
		{ 156, 136, 56 },
		{ 156, 137, 56 },
		{ 156, 138, 56 },
		{ 156, 139, 56 },
		{ 156, 140, 56 },
		{ 156, 141, 56 },
		{ 157, 98, 56 },
		{ 157, 99, 56 },
		{ 157, 100, 56 },
		{ 157, 101, 56 },
		{ 157, 102, 56 },
		{ 157, 103, 56 },
		{ 157, 104, 56 },
		{ 157, 105, 56 },
		{ 157, 106, 56 },
		{ 157, 107, 56 },
		{ 157, 108, 56 },
		{ 157, 109, 56 },
		{ 157, 110, 56 },
		{ 157, 111, 56 },
		{ 157, 112, 56 },
		{ 157, 113, 56 },
		{ 157, 114, 56 },
		{ 157, 115, 56 },
		{ 157, 116, 56 },
		{ 157, 117, 56 },
		{ 157, 118, 56 },
		{ 157, 119, 56 },
		{ 157, 120, 56 },
		{ 157, 121, 56 },
		{ 157, 122, 56 },
		{ 157, 123, 56 },
		{ 157, 124, 56 },
		{ 157, 125, 56 },
		{ 157, 126, 56 },
		{ 157, 127, 56 },
		{ 157, 128, 56 },
		{ 157, 129, 56 },
		{ 157, 130, 56 },
		{ 157, 131, 56 },
		{ 157, 132, 56 },
		{ 157, 133, 56 },
		{ 157, 134, 56 },
		{ 157, 135, 56 },
		{ 157, 136, 56 },
		{ 157, 137, 56 },
		{ 157, 138, 56 },
		{ 157, 139, 56 },
		{ 157, 140, 56 },
		{ 157, 141, 56 },
		{ 157, 142, 56 },
		{ 158, 99, 56 },
		{ 158, 100, 56 },
		{ 158, 101, 56 },
		{ 158, 102, 56 },
		{ 158, 103, 56 },
		{ 158, 104, 56 },
		{ 158, 105, 56 },
		{ 158, 106, 56 },
		{ 158, 107, 56 },
		{ 158, 108, 56 },
		{ 158, 109, 56 },
		{ 158, 110, 56 },
		{ 158, 111, 56 },
		{ 158, 112, 56 },
		{ 158, 113, 56 },
		{ 158, 114, 56 },
		{ 158, 115, 56 },
		{ 158, 116, 56 },
		{ 158, 117, 56 },
		{ 158, 118, 56 },
		{ 158, 119, 56 },
		{ 158, 120, 56 },
		{ 158, 121, 56 },
		{ 158, 122, 56 },
		{ 158, 123, 56 },
		{ 158, 124, 56 },
		{ 158, 125, 56 },
		{ 158, 126, 56 },
		{ 158, 127, 56 },
		{ 158, 128, 56 },
		{ 158, 129, 56 },
		{ 158, 130, 56 },
		{ 158, 131, 56 },
		{ 158, 132, 56 },
		{ 158, 133, 56 },
		{ 158, 134, 56 },
		{ 158, 135, 56 },
		{ 158, 136, 56 },
		{ 158, 137, 56 },
		{ 158, 138, 56 },
		{ 158, 139, 56 },
		{ 158, 140, 56 },
		{ 158, 141, 56 },
		{ 158, 142, 56 },
		{ 158, 143, 56 },
		{ 159, 101, 56 },
		{ 159, 102, 56 },
		{ 159, 103, 56 },
		{ 159, 104, 56 },
		{ 159, 105, 56 },
		{ 159, 106, 56 },
		{ 159, 107, 56 },
		{ 159, 108, 56 },
		{ 159, 109, 56 },
		{ 159, 110, 56 },
		{ 159, 111, 56 },
		{ 159, 112, 56 },
		{ 159, 113, 56 },
		{ 159, 114, 56 },
		{ 159, 115, 56 },
		{ 159, 116, 56 },
		{ 159, 117, 56 },
		{ 159, 118, 56 },
		{ 159, 119, 56 },
		{ 159, 120, 56 },
		{ 159, 121, 56 },
		{ 159, 122, 56 },
		{ 159, 123, 56 },
		{ 159, 124, 56 },
		{ 159, 125, 56 },
		{ 159, 126, 56 },
		{ 159, 127, 56 },
		{ 159, 128, 56 },
		{ 159, 129, 56 },
		{ 159, 130, 56 },
		{ 159, 131, 56 },
		{ 159, 132, 56 },
		{ 159, 133, 56 },
		{ 159, 134, 56 },
		{ 159, 135, 56 },
		{ 159, 136, 56 },
		{ 159, 137, 56 },
		{ 159, 138, 56 },
		{ 159, 139, 56 },
		{ 159, 140, 56 },
		{ 159, 141, 56 },
		{ 159, 142, 56 },
		{ 159, 143, 56 },
		{ 159, 144, 56 },
		{ 160, 102, 56 },
		{ 160, 103, 56 },
		{ 160, 104, 56 },
		{ 160, 105, 56 },
		{ 160, 106, 56 },
		{ 160, 107, 56 },
		{ 160, 108, 56 },
		{ 160, 109, 56 },
		{ 160, 110, 56 },
		{ 160, 111, 56 },
		{ 160, 112, 56 },
		{ 160, 113, 56 },
		{ 160, 114, 56 },
		{ 160, 115, 56 },
		{ 160, 116, 56 },
		{ 160, 117, 56 },
		{ 160, 118, 56 },
		{ 160, 119, 56 },
		{ 160, 120, 56 },
		{ 160, 121, 56 },
		{ 160, 122, 56 },
		{ 160, 123, 56 },
		{ 160, 124, 56 },
		{ 160, 125, 56 },
		{ 160, 126, 56 },
		{ 160, 127, 56 },
		{ 160, 128, 56 },
		{ 160, 129, 56 },
		{ 160, 130, 56 },
		{ 160, 131, 56 },
		{ 160, 132, 56 },
		{ 160, 133, 56 },
		{ 160, 134, 56 },
		{ 160, 135, 56 },
		{ 160, 136, 56 },
		{ 160, 137, 56 },
		{ 160, 138, 56 },
		{ 160, 139, 56 },
		{ 160, 140, 56 },
		{ 160, 141, 56 },
		{ 160, 142, 56 },
		{ 160, 143, 56 },
		{ 160, 144, 56 },
		{ 160, 145, 56 },
		{ 160, 146, 56 },
		{ 161, 104, 56 },
		{ 161, 105, 56 },
		{ 161, 106, 56 },
		{ 161, 107, 56 },
		{ 161, 108, 56 },
		{ 161, 109, 56 },
		{ 161, 110, 56 },
		{ 161, 111, 56 },
		{ 161, 112, 56 },
		{ 161, 113, 56 },
		{ 161, 114, 56 },
		{ 161, 115, 56 },
		{ 161, 116, 56 },
		{ 161, 117, 56 },
		{ 161, 118, 56 },
		{ 161, 119, 56 },
		{ 161, 120, 56 },
		{ 161, 121, 56 },
		{ 161, 122, 56 },
		{ 161, 123, 56 },
		{ 161, 124, 56 },
		{ 161, 125, 56 },
		{ 161, 126, 56 },
		{ 161, 127, 56 },
		{ 161, 128, 56 },
		{ 161, 129, 56 },
		{ 161, 130, 56 },
		{ 161, 131, 56 },
		{ 161, 132, 56 },
		{ 161, 133, 56 },
		{ 161, 134, 56 },
		{ 161, 135, 56 },
		{ 161, 136, 56 },
		{ 161, 137, 56 },
		{ 161, 138, 56 },
		{ 161, 139, 56 },
		{ 161, 140, 56 },
		{ 161, 141, 56 },
		{ 161, 142, 56 },
		{ 161, 143, 56 },
		{ 161, 144, 56 },
		{ 161, 145, 56 },
		{ 161, 146, 56 },
		{ 161, 147, 56 },
		{ 162, 106, 56 },
		{ 162, 107, 56 },
		{ 162, 108, 56 },
		{ 162, 109, 56 },
		{ 162, 110, 56 },
		{ 162, 111, 56 },
		{ 162, 112, 56 },
		{ 162, 113, 56 },
		{ 162, 114, 56 },
		{ 162, 115, 56 },
		{ 162, 116, 56 },
		{ 162, 117, 56 },
		{ 162, 118, 56 },
		{ 162, 119, 56 },
		{ 162, 120, 56 },
		{ 162, 121, 56 },
		{ 162, 122, 56 },
		{ 162, 123, 56 },
		{ 162, 124, 56 },
		{ 162, 125, 56 },
		{ 162, 126, 56 },
		{ 162, 127, 56 },
		{ 162, 128, 56 },
		{ 162, 129, 56 },
		{ 162, 130, 56 },
		{ 162, 131, 56 },
		{ 162, 132, 56 },
		{ 162, 133, 56 },
		{ 162, 134, 56 },
		{ 162, 135, 56 },
		{ 162, 136, 56 },
		{ 162, 137, 56 },
		{ 162, 138, 56 },
		{ 162, 139, 56 },
		{ 162, 140, 56 },
		{ 162, 141, 56 },
		{ 162, 142, 56 },
		{ 162, 143, 56 },
		{ 162, 144, 56 },
		{ 162, 145, 56 },
		{ 162, 146, 56 },
		{ 162, 147, 56 },
		{ 162, 148, 56 },
		{ 163, 107, 56 },
		{ 163, 108, 56 },
		{ 163, 109, 56 },
		{ 163, 110, 56 },
		{ 163, 111, 56 },
		{ 163, 112, 56 },
		{ 163, 113, 56 },
		{ 163, 114, 56 },
		{ 163, 115, 56 },
		{ 163, 116, 56 },
		{ 163, 117, 56 },
		{ 163, 118, 56 },
		{ 163, 119, 56 },
		{ 163, 120, 56 },
		{ 163, 121, 56 },
		{ 163, 122, 56 },
		{ 163, 123, 56 },
		{ 163, 124, 56 },
		{ 163, 125, 56 },
		{ 163, 126, 56 },
		{ 163, 127, 56 },
		{ 163, 128, 56 },
		{ 163, 129, 56 },
		{ 163, 130, 56 },
		{ 163, 131, 56 },
		{ 163, 132, 56 },
		{ 163, 133, 56 },
		{ 163, 134, 56 },
		{ 163, 135, 56 },
		{ 163, 136, 56 },
		{ 163, 137, 56 },
		{ 163, 138, 56 },
		{ 163, 139, 56 },
		{ 163, 140, 56 },
		{ 163, 141, 56 },
		{ 163, 142, 56 },
		{ 163, 143, 56 },
		{ 163, 144, 56 },
		{ 163, 145, 56 },
		{ 163, 146, 56 },
		{ 163, 147, 56 },
		{ 163, 148, 56 },
		{ 163, 149, 56 },
		{ 164, 109, 56 },
		{ 164, 110, 56 },
		{ 164, 111, 56 },
		{ 164, 112, 56 },
		{ 164, 113, 56 },
		{ 164, 114, 56 },
		{ 164, 115, 56 },
		{ 164, 116, 56 },
		{ 164, 117, 56 },
		{ 164, 118, 56 },
		{ 164, 119, 56 },
		{ 164, 120, 56 },
		{ 164, 121, 56 },
		{ 164, 122, 56 },
		{ 164, 123, 56 },
		{ 164, 124, 56 },
		{ 164, 125, 56 },
		{ 164, 126, 56 },
		{ 164, 127, 56 },
		{ 164, 128, 56 },
		{ 164, 129, 56 },
		{ 164, 130, 56 },
		{ 164, 131, 56 },
		{ 164, 132, 56 },
		{ 164, 133, 56 },
		{ 164, 134, 56 },
		{ 164, 135, 56 },
		{ 164, 136, 56 },
		{ 164, 137, 56 },
		{ 164, 138, 56 },
		{ 164, 139, 56 },
		{ 164, 140, 56 },
		{ 164, 141, 56 },
		{ 164, 142, 56 },
		{ 164, 143, 56 },
		{ 164, 144, 56 },
		{ 164, 145, 56 },
		{ 164, 146, 56 },
		{ 164, 147, 56 },
		{ 164, 148, 56 },
		{ 164, 149, 56 },
		{ 164, 150, 56 },
		{ 165, 110, 56 },
		{ 165, 111, 56 },
		{ 165, 112, 56 },
		{ 165, 113, 56 },
		{ 165, 114, 56 },
		{ 165, 115, 56 },
		{ 165, 116, 56 },
		{ 165, 117, 56 },
		{ 165, 118, 56 },
		{ 165, 119, 56 },
		{ 165, 120, 56 },
		{ 165, 121, 56 },
		{ 165, 122, 56 },
		{ 165, 123, 56 },
		{ 165, 124, 56 },
		{ 165, 125, 56 },
		{ 165, 126, 56 },
		{ 165, 127, 56 },
		{ 165, 128, 56 },
		{ 165, 129, 56 },
		{ 165, 130, 56 },
		{ 165, 131, 56 },
		{ 165, 132, 56 },
		{ 165, 133, 56 },
		{ 165, 134, 56 },
		{ 165, 135, 56 },
		{ 165, 136, 56 },
		{ 165, 137, 56 },
		{ 165, 138, 56 },
		{ 165, 139, 56 },
		{ 165, 140, 56 },
		{ 165, 141, 56 },
		{ 165, 142, 56 },
		{ 165, 143, 56 },
		{ 165, 144, 56 },
		{ 165, 145, 56 },
		{ 165, 146, 56 },
		{ 165, 147, 56 },
		{ 165, 148, 56 },
		{ 165, 149, 56 },
		{ 165, 150, 56 },
		{ 165, 151, 56 },
		{ 166, 112, 56 },
		{ 166, 113, 56 },
		{ 166, 114, 56 },
		{ 166, 115, 56 },
		{ 166, 116, 56 },
		{ 166, 117, 56 },
		{ 166, 118, 56 },
		{ 166, 119, 56 },
		{ 166, 120, 56 },
		{ 166, 121, 56 },
		{ 166, 122, 56 },
		{ 166, 123, 56 },
		{ 166, 124, 56 },
		{ 166, 125, 56 },
		{ 166, 126, 56 },
		{ 166, 127, 56 },
		{ 166, 128, 56 },
		{ 166, 129, 56 },
		{ 166, 130, 56 },
		{ 166, 131, 56 },
		{ 166, 132, 56 },
		{ 166, 133, 56 },
		{ 166, 134, 56 },
		{ 166, 135, 56 },
		{ 166, 136, 56 },
		{ 166, 137, 56 },
		{ 166, 138, 56 },
		{ 166, 139, 56 },
		{ 166, 140, 56 },
		{ 166, 141, 56 },
		{ 166, 142, 56 },
		{ 166, 143, 56 },
		{ 166, 144, 56 },
		{ 166, 145, 56 },
		{ 166, 146, 56 },
		{ 166, 147, 56 },
		{ 166, 148, 56 },
		{ 166, 149, 56 },
		{ 166, 150, 56 },
		{ 166, 151, 56 },
		{ 166, 152, 56 },
		{ 167, 114, 56 },
		{ 167, 115, 56 },
		{ 167, 116, 56 },
		{ 167, 117, 56 },
		{ 167, 118, 56 },
		{ 167, 119, 56 },
		{ 167, 120, 56 },
		{ 167, 121, 56 },
		{ 167, 122, 56 },
		{ 167, 123, 56 },
		{ 167, 124, 56 },
		{ 167, 125, 56 },
		{ 167, 126, 56 },
		{ 167, 127, 56 },
		{ 167, 128, 56 },
		{ 167, 129, 56 },
		{ 167, 130, 56 },
		{ 167, 131, 56 },
		{ 167, 132, 56 },
		{ 167, 133, 56 },
		{ 167, 134, 56 },
		{ 167, 135, 56 },
		{ 167, 136, 56 },
		{ 167, 137, 56 },
		{ 167, 138, 56 },
		{ 167, 139, 56 },
		{ 167, 140, 56 },
		{ 167, 141, 56 },
		{ 167, 142, 56 },
		{ 167, 143, 56 },
		{ 167, 144, 56 },
		{ 167, 145, 56 },
		{ 167, 146, 56 },
		{ 167, 147, 56 },
		{ 167, 148, 56 },
		{ 167, 149, 56 },
		{ 167, 150, 56 },
		{ 167, 151, 56 },
		{ 167, 152, 56 },
		{ 167, 153, 56 },
		{ 167, 154, 56 },
		{ 168, 115, 56 },
		{ 168, 116, 56 },
		{ 168, 117, 56 },
		{ 168, 118, 56 },
		{ 168, 119, 56 },
		{ 168, 120, 56 },
		{ 168, 121, 56 },
		{ 168, 122, 56 },
		{ 168, 123, 56 },
		{ 168, 124, 56 },
		{ 168, 125, 56 },
		{ 168, 126, 56 },
		{ 168, 127, 56 },
		{ 168, 128, 56 },
		{ 168, 129, 56 },
		{ 168, 130, 56 },
		{ 168, 131, 56 },
		{ 168, 132, 56 },
		{ 168, 133, 56 },
		{ 168, 134, 56 },
		{ 168, 135, 56 },
		{ 168, 136, 56 },
		{ 168, 137, 56 },
		{ 168, 138, 56 },
		{ 168, 139, 56 },
		{ 168, 140, 56 },
		{ 168, 141, 56 },
		{ 168, 142, 56 },
		{ 168, 143, 56 },
		{ 168, 144, 56 },
		{ 168, 145, 56 },
		{ 168, 146, 56 },
		{ 168, 147, 56 },
		{ 168, 148, 56 },
		{ 168, 149, 56 },
		{ 168, 150, 56 },
		{ 168, 151, 56 },
		{ 168, 152, 56 },
		{ 168, 153, 56 },
		{ 168, 154, 56 },
		{ 168, 155, 56 },
		{ 169, 117, 56 },
		{ 169, 118, 56 },
		{ 169, 119, 56 },
		{ 169, 120, 56 },
		{ 169, 121, 56 },
		{ 169, 122, 56 },
		{ 169, 123, 56 },
		{ 169, 124, 56 },
		{ 169, 125, 56 },
		{ 169, 126, 56 },
		{ 169, 127, 56 },
		{ 169, 128, 56 },
		{ 169, 129, 56 },
		{ 169, 130, 56 },
		{ 169, 131, 56 },
		{ 169, 132, 56 },
		{ 169, 133, 56 },
		{ 169, 134, 56 },
		{ 169, 135, 56 },
		{ 169, 136, 56 },
		{ 169, 137, 56 },
		{ 169, 138, 56 },
		{ 169, 139, 56 },
		{ 169, 140, 56 },
		{ 169, 141, 56 },
		{ 169, 142, 56 },
		{ 169, 143, 56 },
		{ 169, 144, 56 },
		{ 169, 145, 56 },
		{ 169, 146, 56 },
		{ 169, 147, 56 },
		{ 169, 148, 56 },
		{ 169, 149, 56 },
		{ 169, 150, 56 },
		{ 169, 151, 56 },
		{ 169, 152, 56 },
		{ 169, 153, 56 },
		{ 169, 154, 56 },
		{ 169, 155, 56 },
		{ 169, 156, 56 },
		{ 170, 118, 56 },
		{ 170, 119, 56 },
		{ 170, 120, 56 },
		{ 170, 121, 56 },
		{ 170, 122, 56 },
		{ 170, 123, 56 },
		{ 170, 124, 56 },
		{ 170, 125, 56 },
		{ 170, 126, 56 },
		{ 170, 127, 56 },
		{ 170, 128, 56 },
		{ 170, 129, 56 },
		{ 170, 130, 56 },
		{ 170, 131, 56 },
		{ 170, 132, 56 },
		{ 170, 133, 56 },
		{ 170, 134, 56 },
		{ 170, 135, 56 },
		{ 170, 136, 56 },
		{ 170, 137, 56 },
		{ 170, 138, 56 },
		{ 170, 139, 56 },
		{ 170, 140, 56 },
		{ 170, 141, 56 },
		{ 170, 142, 56 },
		{ 170, 143, 56 },
		{ 170, 144, 56 },
		{ 170, 145, 56 },
		{ 170, 146, 56 },
		{ 170, 147, 56 },
		{ 170, 148, 56 },
		{ 170, 149, 56 },
		{ 170, 150, 56 },
		{ 170, 151, 56 },
		{ 170, 152, 56 },
		{ 170, 153, 56 },
		{ 170, 154, 56 },
		{ 170, 155, 56 },
		{ 170, 156, 56 },
		{ 170, 157, 56 },
		{ 171, 120, 56 },
		{ 171, 121, 56 },
		{ 171, 122, 56 },
		{ 171, 123, 56 },
		{ 171, 124, 56 },
		{ 171, 125, 56 },
		{ 171, 126, 56 },
		{ 171, 127, 56 },
		{ 171, 128, 56 },
		{ 171, 129, 56 },
		{ 171, 130, 56 },
		{ 171, 131, 56 },
		{ 171, 132, 56 },
		{ 171, 133, 56 },
		{ 171, 134, 56 },
		{ 171, 135, 56 },
		{ 171, 136, 56 },
		{ 171, 137, 56 },
		{ 171, 138, 56 },
		{ 171, 139, 56 },
		{ 171, 140, 56 },
		{ 171, 141, 56 },
		{ 171, 142, 56 },
		{ 171, 143, 56 },
		{ 171, 144, 56 },
		{ 171, 145, 56 },
		{ 171, 146, 56 },
		{ 171, 147, 56 },
		{ 171, 148, 56 },
		{ 171, 149, 56 },
		{ 171, 150, 56 },
		{ 171, 151, 56 },
		{ 171, 152, 56 },
		{ 171, 153, 56 },
		{ 171, 154, 56 },
		{ 171, 155, 56 },
		{ 171, 156, 56 },
		{ 171, 157, 56 },
		{ 171, 158, 56 },
		{ 172, 122, 56 },
		{ 172, 123, 56 },
		{ 172, 124, 56 },
		{ 172, 125, 56 },
		{ 172, 126, 56 },
		{ 172, 127, 56 },
		{ 172, 128, 56 },
		{ 172, 129, 56 },
		{ 172, 130, 56 },
		{ 172, 131, 56 },
		{ 172, 132, 56 },
		{ 172, 133, 56 },
		{ 172, 134, 56 },
		{ 172, 135, 56 },
		{ 172, 136, 56 },
		{ 172, 137, 56 },
		{ 172, 138, 56 },
		{ 172, 139, 56 },
		{ 172, 140, 56 },
		{ 172, 141, 56 },
		{ 172, 142, 56 },
		{ 172, 143, 56 },
		{ 172, 144, 56 },
		{ 172, 145, 56 },
		{ 172, 146, 56 },
		{ 172, 147, 56 },
		{ 172, 148, 56 },
		{ 172, 149, 56 },
		{ 172, 150, 56 },
		{ 172, 151, 56 },
		{ 172, 152, 56 },
		{ 172, 153, 56 },
		{ 172, 154, 56 },
		{ 172, 155, 56 },
		{ 172, 156, 56 },
		{ 172, 157, 56 },
		{ 172, 158, 56 },
		{ 172, 159, 56 },
		{ 173, 123, 56 },
		{ 173, 124, 56 },
		{ 173, 125, 56 },
		{ 173, 126, 56 },
		{ 173, 127, 56 },
		{ 173, 128, 56 },
		{ 173, 129, 56 },
		{ 173, 130, 56 },
		{ 173, 131, 56 },
		{ 173, 132, 56 },
		{ 173, 133, 56 },
		{ 173, 134, 56 },
		{ 173, 135, 56 },
		{ 173, 136, 56 },
		{ 173, 137, 56 },
		{ 173, 138, 56 },
		{ 173, 139, 56 },
		{ 173, 140, 56 },
		{ 173, 141, 56 },
		{ 173, 142, 56 },
		{ 173, 143, 56 },
		{ 173, 144, 56 },
		{ 173, 145, 56 },
		{ 173, 146, 56 },
		{ 173, 147, 56 },
		{ 173, 148, 56 },
		{ 173, 149, 56 },
		{ 173, 150, 56 },
		{ 173, 151, 56 },
		{ 173, 152, 56 },
		{ 173, 153, 56 },
		{ 173, 154, 56 },
		{ 173, 155, 56 },
		{ 173, 156, 56 },
		{ 173, 157, 56 },
		{ 173, 158, 56 },
		{ 173, 159, 56 },
		{ 173, 160, 56 },
		{ 174, 125, 56 },
		{ 174, 126, 56 },
		{ 174, 127, 56 },
		{ 174, 128, 56 },
		{ 174, 129, 56 },
		{ 174, 130, 56 },
		{ 174, 131, 56 },
		{ 174, 132, 56 },
		{ 174, 133, 56 },
		{ 174, 134, 56 },
		{ 174, 135, 56 },
		{ 174, 136, 56 },
		{ 174, 137, 56 },
		{ 174, 138, 56 },
		{ 174, 139, 56 },
		{ 174, 140, 56 },
		{ 174, 141, 56 },
		{ 174, 142, 56 },
		{ 174, 143, 56 },
		{ 174, 144, 56 },
		{ 174, 145, 56 },
		{ 174, 146, 56 },
		{ 174, 147, 56 },
		{ 174, 148, 56 },
		{ 174, 149, 56 },
		{ 174, 150, 56 },
		{ 174, 151, 56 },
		{ 174, 152, 56 },
		{ 174, 153, 56 },
		{ 174, 154, 56 },
		{ 174, 155, 56 },
		{ 174, 156, 56 },
		{ 174, 157, 56 },
		{ 174, 158, 56 },
		{ 174, 159, 56 },
		{ 174, 160, 56 },
		{ 174, 161, 56 },
		{ 174, 162, 56 },
		{ 175, 126, 56 },
		{ 175, 127, 56 },
		{ 175, 128, 56 },
		{ 175, 129, 56 },
		{ 175, 130, 56 },
		{ 175, 131, 56 },
		{ 175, 132, 56 },
		{ 175, 133, 56 },
		{ 175, 134, 56 },
		{ 175, 135, 56 },
		{ 175, 136, 56 },
		{ 175, 137, 56 },
		{ 175, 138, 56 },
		{ 175, 139, 56 },
		{ 175, 140, 56 },
		{ 175, 141, 56 },
		{ 175, 142, 56 },
		{ 175, 143, 56 },
		{ 175, 144, 56 },
		{ 175, 145, 56 },
		{ 175, 146, 56 },
		{ 175, 147, 56 },
		{ 175, 148, 56 },
		{ 175, 149, 56 },
		{ 175, 150, 56 },
		{ 175, 151, 56 },
		{ 175, 152, 56 },
		{ 175, 153, 56 },
		{ 175, 154, 56 },
		{ 175, 155, 56 },
		{ 175, 156, 56 },
		{ 175, 157, 56 },
		{ 175, 158, 56 },
		{ 175, 159, 56 },
		{ 175, 160, 56 },
		{ 175, 161, 56 },
		{ 175, 162, 56 },
		{ 175, 163, 56 },
		{ 176, 128, 56 },
		{ 176, 129, 56 },
		{ 176, 130, 56 },
		{ 176, 131, 56 },
		{ 176, 132, 56 },
		{ 176, 133, 56 },
		{ 176, 134, 56 },
		{ 176, 135, 56 },
		{ 176, 136, 56 },
		{ 176, 137, 56 },
		{ 176, 138, 56 },
		{ 176, 139, 56 },
		{ 176, 140, 56 },
		{ 176, 141, 56 },
		{ 176, 142, 56 },
		{ 176, 143, 56 },
		{ 176, 144, 56 },
		{ 176, 145, 56 },
		{ 176, 146, 56 },
		{ 176, 147, 56 },
		{ 176, 148, 56 },
		{ 176, 149, 56 },
		{ 176, 150, 56 },
		{ 176, 151, 56 },
		{ 176, 152, 56 },
		{ 176, 153, 56 },
		{ 176, 154, 56 },
		{ 176, 155, 56 },
		{ 176, 156, 56 },
		{ 176, 157, 56 },
		{ 176, 158, 56 },
		{ 176, 159, 56 },
		{ 176, 160, 56 },
		{ 176, 161, 56 },
		{ 176, 162, 56 },
		{ 176, 163, 56 },
		{ 176, 164, 56 },
		{ 177, 130, 56 },
		{ 177, 131, 56 },
		{ 177, 132, 56 },
		{ 177, 133, 56 },
		{ 177, 134, 56 },
		{ 177, 135, 56 },
		{ 177, 136, 56 },
		{ 177, 137, 56 },
		{ 177, 138, 56 },
		{ 177, 139, 56 },
		{ 177, 140, 56 },
		{ 177, 141, 56 },
		{ 177, 142, 56 },
		{ 177, 143, 56 },
		{ 177, 144, 56 },
		{ 177, 145, 56 },
		{ 177, 146, 56 },
		{ 177, 147, 56 },
		{ 177, 148, 56 },
		{ 177, 149, 56 },
		{ 177, 150, 56 },
		{ 177, 151, 56 },
		{ 177, 152, 56 },
		{ 177, 153, 56 },
		{ 177, 154, 56 },
		{ 177, 155, 56 },
		{ 177, 156, 56 },
		{ 177, 157, 56 },
		{ 177, 158, 56 },
		{ 177, 159, 56 },
		{ 177, 160, 56 },
		{ 177, 161, 56 },
		{ 177, 162, 56 },
		{ 177, 163, 56 },
		{ 177, 164, 56 },
		{ 177, 165, 56 },
		{ 178, 131, 56 },
		{ 178, 132, 56 },
		{ 178, 133, 56 },
		{ 178, 134, 56 },
		{ 178, 135, 56 },
		{ 178, 136, 56 },
		{ 178, 137, 56 },
		{ 178, 138, 56 },
		{ 178, 139, 56 },
		{ 178, 140, 56 },
		{ 178, 141, 56 },
		{ 178, 142, 56 },
		{ 178, 143, 56 },
		{ 178, 144, 56 },
		{ 178, 145, 56 },
		{ 178, 146, 56 },
		{ 178, 147, 56 },
		{ 178, 148, 56 },
		{ 178, 149, 56 },
		{ 178, 150, 56 },
		{ 178, 151, 56 },
		{ 178, 152, 56 },
		{ 178, 153, 56 },
		{ 178, 154, 56 },
		{ 178, 155, 56 },
		{ 178, 156, 56 },
		{ 178, 157, 56 },
		{ 178, 158, 56 },
		{ 178, 159, 56 },
		{ 178, 160, 56 },
		{ 178, 161, 56 },
		{ 178, 162, 56 },
		{ 178, 163, 56 },
		{ 178, 164, 56 },
		{ 178, 165, 56 },
		{ 178, 166, 56 },
		{ 179, 133, 56 },
		{ 179, 134, 56 },
		{ 179, 135, 56 },
		{ 179, 136, 56 },
		{ 179, 137, 56 },
		{ 179, 138, 56 },
		{ 179, 139, 56 },
		{ 179, 140, 56 },
		{ 179, 141, 56 },
		{ 179, 142, 56 },
		{ 179, 143, 56 },
		{ 179, 144, 56 },
		{ 179, 145, 56 },
		{ 179, 146, 56 },
		{ 179, 147, 56 },
		{ 179, 148, 56 },
		{ 179, 149, 56 },
		{ 179, 150, 56 },
		{ 179, 151, 56 },
		{ 179, 152, 56 },
		{ 179, 153, 56 },
		{ 179, 154, 56 },
		{ 179, 155, 56 },
		{ 179, 156, 56 },
		{ 179, 157, 56 },
		{ 179, 158, 56 },
		{ 179, 159, 56 },
		{ 179, 160, 56 },
		{ 179, 161, 56 },
		{ 179, 162, 56 },
		{ 179, 163, 56 },
		{ 179, 164, 56 },
		{ 179, 165, 56 },
		{ 179, 166, 56 },
		{ 179, 167, 56 },
		{ 180, 134, 56 },
		{ 180, 135, 56 },
		{ 180, 136, 56 },
		{ 180, 137, 56 },
		{ 180, 138, 56 },
		{ 180, 139, 56 },
		{ 180, 140, 56 },
		{ 180, 141, 56 },
		{ 180, 142, 56 },
		{ 180, 143, 56 },
		{ 180, 144, 56 },
		{ 180, 145, 56 },
		{ 180, 146, 56 },
		{ 180, 147, 56 },
		{ 180, 148, 56 },
		{ 180, 149, 56 },
		{ 180, 150, 56 },
		{ 180, 151, 56 },
		{ 180, 152, 56 },
		{ 180, 153, 56 },
		{ 180, 154, 56 },
		{ 180, 155, 56 },
		{ 180, 156, 56 },
		{ 180, 157, 56 },
		{ 180, 158, 56 },
		{ 180, 159, 56 },
		{ 180, 160, 56 },
		{ 180, 161, 56 },
		{ 180, 162, 56 },
		{ 180, 163, 56 },
		{ 180, 164, 56 },
		{ 180, 165, 56 },
		{ 180, 166, 56 },
		{ 180, 167, 56 },
		{ 180, 168, 56 },
		{ 181, 136, 56 },
		{ 181, 137, 56 },
		{ 181, 138, 56 },
		{ 181, 139, 56 },
		{ 181, 140, 56 },
		{ 181, 141, 56 },
		{ 181, 142, 56 },
		{ 181, 143, 56 },
		{ 181, 144, 56 },
		{ 181, 145, 56 },
		{ 181, 146, 56 },
		{ 181, 147, 56 },
		{ 181, 148, 56 },
		{ 181, 149, 56 },
		{ 181, 150, 56 },
		{ 181, 151, 56 },
		{ 181, 152, 56 },
		{ 181, 153, 56 },
		{ 181, 154, 56 },
		{ 181, 155, 56 },
		{ 181, 156, 56 },
		{ 181, 157, 56 },
		{ 181, 158, 56 },
		{ 181, 159, 56 },
		{ 181, 160, 56 },
		{ 181, 161, 56 },
		{ 181, 162, 56 },
		{ 181, 163, 56 },
		{ 181, 164, 56 },
		{ 181, 165, 56 },
		{ 181, 166, 56 },
		{ 181, 167, 56 },
		{ 181, 168, 56 },
		{ 181, 169, 56 },
		{ 181, 170, 56 },
		{ 182, 138, 56 },
		{ 182, 139, 56 },
		{ 182, 140, 56 },
		{ 182, 141, 56 },
		{ 182, 142, 56 },
		{ 182, 143, 56 },
		{ 182, 144, 56 },
		{ 182, 145, 56 },
		{ 182, 146, 56 },
		{ 182, 147, 56 },
		{ 182, 148, 56 },
		{ 182, 149, 56 },
		{ 182, 150, 56 },
		{ 182, 151, 56 },
		{ 182, 152, 56 },
		{ 182, 153, 56 },
		{ 182, 154, 56 },
		{ 182, 155, 56 },
		{ 182, 156, 56 },
		{ 182, 157, 56 },
		{ 182, 158, 56 },
		{ 182, 159, 56 },
		{ 182, 160, 56 },
		{ 182, 161, 56 },
		{ 182, 162, 56 },
		{ 182, 163, 56 },
		{ 182, 164, 56 },
		{ 182, 165, 56 },
		{ 182, 166, 56 },
		{ 182, 167, 56 },
		{ 182, 168, 56 },
		{ 182, 169, 56 },
		{ 182, 170, 56 },
		{ 182, 171, 56 },
		{ 183, 139, 56 },
		{ 183, 140, 56 },
		{ 183, 141, 56 },
		{ 183, 142, 56 },
		{ 183, 143, 56 },
		{ 183, 144, 56 },
		{ 183, 145, 56 },
		{ 183, 146, 56 },
		{ 183, 147, 56 },
		{ 183, 148, 56 },
		{ 183, 149, 56 },
		{ 183, 150, 56 },
		{ 183, 151, 56 },
		{ 183, 152, 56 },
		{ 183, 153, 56 },
		{ 183, 154, 56 },
		{ 183, 155, 56 },
		{ 183, 156, 56 },
		{ 183, 157, 56 },
		{ 183, 158, 56 },
		{ 183, 159, 56 },
		{ 183, 160, 56 },
		{ 183, 161, 56 },
		{ 183, 162, 56 },
		{ 183, 163, 56 },
		{ 183, 164, 56 },
		{ 183, 165, 56 },
		{ 183, 166, 56 },
		{ 183, 167, 56 },
		{ 183, 168, 56 },
		{ 183, 169, 56 },
		{ 183, 170, 56 },
		{ 183, 171, 56 },
		{ 183, 172, 56 },
		{ 184, 141, 56 },
		{ 184, 142, 56 },
		{ 184, 143, 56 },
		{ 184, 144, 56 },
		{ 184, 145, 56 },
		{ 184, 146, 56 },
		{ 184, 147, 56 },
		{ 184, 148, 56 },
		{ 184, 149, 56 },
		{ 184, 150, 56 },
		{ 184, 151, 56 },
		{ 184, 152, 56 },
		{ 184, 153, 56 },
		{ 184, 154, 56 },
		{ 184, 155, 56 },
		{ 184, 156, 56 },
		{ 184, 157, 56 },
		{ 184, 158, 56 },
		{ 184, 159, 56 },
		{ 184, 160, 56 },
		{ 184, 161, 56 },
		{ 184, 162, 56 },
		{ 184, 163, 56 },
		{ 184, 164, 56 },
		{ 184, 165, 56 },
		{ 184, 166, 56 },
		{ 184, 167, 56 },
		{ 184, 168, 56 },
		{ 184, 169, 56 },
		{ 184, 170, 56 },
		{ 184, 171, 56 },
		{ 184, 172, 56 },
		{ 184, 173, 56 },
		{ 185, 142, 56 },
		{ 185, 143, 56 },
		{ 185, 144, 56 },
		{ 185, 145, 56 },
		{ 185, 146, 56 },
		{ 185, 147, 56 },
		{ 185, 148, 56 },
		{ 185, 149, 56 },
		{ 185, 150, 56 },
		{ 185, 151, 56 },
		{ 185, 152, 56 },
		{ 185, 153, 56 },
		{ 185, 154, 56 },
		{ 185, 155, 56 },
		{ 185, 156, 56 },
		{ 185, 157, 56 },
		{ 185, 158, 56 },
		{ 185, 159, 56 },
		{ 185, 160, 56 },
		{ 185, 161, 56 },
		{ 185, 162, 56 },
		{ 185, 163, 56 },
		{ 185, 164, 56 },
		{ 185, 165, 56 },
		{ 185, 166, 56 },
		{ 185, 167, 56 },
		{ 185, 168, 56 },
		{ 185, 169, 56 },
		{ 185, 170, 56 },
		{ 185, 171, 56 },
		{ 185, 172, 56 },
		{ 185, 173, 56 },
		{ 185, 174, 56 },
		{ 186, 144, 56 },
		{ 186, 145, 56 },
		{ 186, 146, 56 },
		{ 186, 147, 56 },
		{ 186, 148, 56 },
		{ 186, 149, 56 },
		{ 186, 150, 56 },
		{ 186, 151, 56 },
		{ 186, 152, 56 },
		{ 186, 153, 56 },
		{ 186, 154, 56 },
		{ 186, 155, 56 },
		{ 186, 156, 56 },
		{ 186, 157, 56 },
		{ 186, 158, 56 },
		{ 186, 159, 56 },
		{ 186, 160, 56 },
		{ 186, 161, 56 },
		{ 186, 162, 56 },
		{ 186, 163, 56 },
		{ 186, 164, 56 },
		{ 186, 165, 56 },
		{ 186, 166, 56 },
		{ 186, 167, 56 },
		{ 186, 168, 56 },
		{ 186, 169, 56 },
		{ 186, 170, 56 },
		{ 186, 171, 56 },
		{ 186, 172, 56 },
		{ 186, 173, 56 },
		{ 186, 174, 56 },
		{ 186, 175, 56 },
		{ 187, 146, 56 },
		{ 187, 147, 56 },
		{ 187, 148, 56 },
		{ 187, 149, 56 },
		{ 187, 150, 56 },
		{ 187, 151, 56 },
		{ 187, 152, 56 },
		{ 187, 153, 56 },
		{ 187, 154, 56 },
		{ 187, 155, 56 },
		{ 187, 156, 56 },
		{ 187, 157, 56 },
		{ 187, 158, 56 },
		{ 187, 159, 56 },
		{ 187, 160, 56 },
		{ 187, 161, 56 },
		{ 187, 162, 56 },
		{ 187, 163, 56 },
		{ 187, 164, 56 },
		{ 187, 165, 56 },
		{ 187, 166, 56 },
		{ 187, 167, 56 },
		{ 187, 168, 56 },
		{ 187, 169, 56 },
		{ 187, 170, 56 },
		{ 187, 171, 56 },
		{ 187, 172, 56 },
		{ 187, 173, 56 },
		{ 187, 174, 56 },
		{ 187, 175, 56 },
		{ 187, 176, 56 },
		{ 188, 147, 56 },
		{ 188, 148, 56 },
		{ 188, 149, 56 },
		{ 188, 150, 56 },
		{ 188, 151, 56 },
		{ 188, 152, 56 },
		{ 188, 153, 56 },
		{ 188, 154, 56 },
		{ 188, 155, 56 },
		{ 188, 156, 56 },
		{ 188, 157, 56 },
		{ 188, 158, 56 },
		{ 188, 159, 56 },
		{ 188, 160, 56 },
		{ 188, 161, 56 },
		{ 188, 162, 56 },
		{ 188, 163, 56 },
		{ 188, 164, 56 },
		{ 188, 165, 56 },
		{ 188, 166, 56 },
		{ 188, 167, 56 },
		{ 188, 168, 56 },
		{ 188, 169, 56 },
		{ 188, 170, 56 },
		{ 188, 171, 56 },
		{ 188, 172, 56 },
		{ 188, 173, 56 },
		{ 188, 174, 56 },
		{ 188, 175, 56 },
		{ 188, 176, 56 },
		{ 188, 177, 56 },
		{ 188, 178, 56 },
		{ 189, 149, 56 },
		{ 189, 150, 56 },
		{ 189, 151, 56 },
		{ 189, 152, 56 },
		{ 189, 153, 56 },
		{ 189, 154, 56 },
		{ 189, 155, 56 },
		{ 189, 156, 56 },
		{ 189, 157, 56 },
		{ 189, 158, 56 },
		{ 189, 159, 56 },
		{ 189, 160, 56 },
		{ 189, 161, 56 },
		{ 189, 162, 56 },
		{ 189, 163, 56 },
		{ 189, 164, 56 },
		{ 189, 165, 56 },
		{ 189, 166, 56 },
		{ 189, 167, 56 },
		{ 189, 168, 56 },
		{ 189, 169, 56 },
		{ 189, 170, 56 },
		{ 189, 171, 56 },
		{ 189, 172, 56 },
		{ 189, 173, 56 },
		{ 189, 174, 56 },
		{ 189, 175, 56 },
		{ 189, 176, 56 },
		{ 189, 177, 56 },
		{ 189, 178, 56 },
		{ 189, 179, 56 },
		{ 190, 150, 56 },
		{ 190, 151, 56 },
		{ 190, 152, 56 },
		{ 190, 153, 56 },
		{ 190, 154, 56 },
		{ 190, 155, 56 },
		{ 190, 156, 56 },
		{ 190, 157, 56 },
		{ 190, 158, 56 },
		{ 190, 159, 56 },
		{ 190, 160, 56 },
		{ 190, 161, 56 },
		{ 190, 162, 56 },
		{ 190, 163, 56 },
		{ 190, 164, 56 },
		{ 190, 165, 56 },
		{ 190, 166, 56 },
		{ 190, 167, 56 },
		{ 190, 168, 56 },
		{ 190, 169, 56 },
		{ 190, 170, 56 },
		{ 190, 171, 56 },
		{ 190, 172, 56 },
		{ 190, 173, 56 },
		{ 190, 174, 56 },
		{ 190, 175, 56 },
		{ 190, 176, 56 },
		{ 190, 177, 56 },
		{ 190, 178, 56 },
		{ 190, 179, 56 },
		{ 190, 180, 56 },
		{ 191, 152, 56 },
		{ 191, 153, 56 },
		{ 191, 154, 56 },
		{ 191, 155, 56 },
		{ 191, 156, 56 },
		{ 191, 157, 56 },
		{ 191, 158, 56 },
		{ 191, 159, 56 },
		{ 191, 160, 56 },
		{ 191, 161, 56 },
		{ 191, 162, 56 },
		{ 191, 163, 56 },
		{ 191, 164, 56 },
		{ 191, 165, 56 },
		{ 191, 166, 56 },
		{ 191, 167, 56 },
		{ 191, 168, 56 },
		{ 191, 169, 56 },
		{ 191, 170, 56 },
		{ 191, 171, 56 },
		{ 191, 172, 56 },
		{ 191, 173, 56 },
		{ 191, 174, 56 },
		{ 191, 175, 56 },
		{ 191, 176, 56 },
		{ 191, 177, 56 },
		{ 191, 178, 56 },
		{ 191, 179, 56 },
		{ 191, 180, 56 },
		{ 191, 181, 56 },
		{ 192, 154, 56 },
		{ 192, 155, 56 },
		{ 192, 156, 56 },
		{ 192, 157, 56 },
		{ 192, 158, 56 },
		{ 192, 159, 56 },
		{ 192, 160, 56 },
		{ 192, 161, 56 },
		{ 192, 162, 56 },
		{ 192, 163, 56 },
		{ 192, 164, 56 },
		{ 192, 165, 56 },
		{ 192, 166, 56 },
		{ 192, 167, 56 },
		{ 192, 168, 56 },
		{ 192, 169, 56 },
		{ 192, 170, 56 },
		{ 192, 171, 56 },
		{ 192, 172, 56 },
		{ 192, 173, 56 },
		{ 192, 174, 56 },
		{ 192, 175, 56 },
		{ 192, 176, 56 },
		{ 192, 177, 56 },
		{ 192, 178, 56 },
		{ 192, 179, 56 },
		{ 192, 180, 56 },
		{ 192, 181, 56 },
		{ 192, 182, 56 },
		{ 193, 155, 56 },
		{ 193, 156, 56 },
		{ 193, 157, 56 },
		{ 193, 158, 56 },
		{ 193, 159, 56 },
		{ 193, 160, 56 },
		{ 193, 161, 56 },
		{ 193, 162, 56 },
		{ 193, 163, 56 },
		{ 193, 164, 56 },
		{ 193, 165, 56 },
		{ 193, 166, 56 },
		{ 193, 167, 56 },
		{ 193, 168, 56 },
		{ 193, 169, 56 },
		{ 193, 170, 56 },
		{ 193, 171, 56 },
		{ 193, 172, 56 },
		{ 193, 173, 56 },
		{ 193, 174, 56 },
		{ 193, 175, 56 },
		{ 193, 176, 56 },
		{ 193, 177, 56 },
		{ 193, 178, 56 },
		{ 193, 179, 56 },
		{ 193, 180, 56 },
		{ 193, 181, 56 },
		{ 193, 182, 56 },
		{ 193, 183, 56 },
		{ 194, 157, 56 },
		{ 194, 158, 56 },
		{ 194, 159, 56 },
		{ 194, 160, 56 },
		{ 194, 161, 56 },
		{ 194, 162, 56 },
		{ 194, 163, 56 },
		{ 194, 164, 56 },
		{ 194, 165, 56 },
		{ 194, 166, 56 },
		{ 194, 167, 56 },
		{ 194, 168, 56 },
		{ 194, 169, 56 },
		{ 194, 170, 56 },
		{ 194, 171, 56 },
		{ 194, 172, 56 },
		{ 194, 173, 56 },
		{ 194, 174, 56 },
		{ 194, 175, 56 },
		{ 194, 176, 56 },
		{ 194, 177, 56 },
		{ 194, 178, 56 },
		{ 194, 179, 56 },
		{ 194, 180, 56 },
		{ 194, 181, 56 },
		{ 194, 182, 56 },
		{ 194, 183, 56 },
		{ 194, 184, 56 },
		{ 195, 158, 56 },
		{ 195, 159, 56 },
		{ 195, 160, 56 },
		{ 195, 161, 56 },
		{ 195, 162, 56 },
		{ 195, 163, 56 },
		{ 195, 164, 56 },
		{ 195, 165, 56 },
		{ 195, 166, 56 },
		{ 195, 167, 56 },
		{ 195, 168, 56 },
		{ 195, 169, 56 },
		{ 195, 170, 56 },
		{ 195, 171, 56 },
		{ 195, 172, 56 },
		{ 195, 173, 56 },
		{ 195, 174, 56 },
		{ 195, 175, 56 },
		{ 195, 176, 56 },
		{ 195, 177, 56 },
		{ 195, 178, 56 },
		{ 195, 179, 56 },
		{ 195, 180, 56 },
		{ 195, 181, 56 },
		{ 195, 182, 56 },
		{ 195, 183, 56 },
		{ 195, 184, 56 },
		{ 195, 185, 56 },
		{ 196, 160, 56 },
		{ 196, 161, 56 },
		{ 196, 162, 56 },
		{ 196, 163, 56 },
		{ 196, 164, 56 },
		{ 196, 165, 56 },
		{ 196, 166, 56 },
		{ 196, 167, 56 },
		{ 196, 168, 56 },
		{ 196, 169, 56 },
		{ 196, 170, 56 },
		{ 196, 171, 56 },
		{ 196, 172, 56 },
		{ 196, 173, 56 },
		{ 196, 174, 56 },
		{ 196, 175, 56 },
		{ 196, 176, 56 },
		{ 196, 177, 56 },
		{ 196, 178, 56 },
		{ 196, 179, 56 },
		{ 196, 180, 56 },
		{ 196, 181, 56 },
		{ 196, 182, 56 },
		{ 196, 183, 56 },
		{ 196, 184, 56 },
		{ 196, 185, 56 },
		{ 196, 186, 56 },
		{ 196, 187, 56 },
		{ 197, 162, 56 },
		{ 197, 163, 56 },
		{ 197, 164, 56 },
		{ 197, 165, 56 },
		{ 197, 166, 56 },
		{ 197, 167, 56 },
		{ 197, 168, 56 },
		{ 197, 169, 56 },
		{ 197, 170, 56 },
		{ 197, 171, 56 },
		{ 197, 172, 56 },
		{ 197, 173, 56 },
		{ 197, 174, 56 },
		{ 197, 175, 56 },
		{ 197, 176, 56 },
		{ 197, 177, 56 },
		{ 197, 178, 56 },
		{ 197, 179, 56 },
		{ 197, 180, 56 },
		{ 197, 181, 56 },
		{ 197, 182, 56 },
		{ 197, 183, 56 },
		{ 197, 184, 56 },
		{ 197, 185, 56 },
		{ 197, 186, 56 },
		{ 197, 187, 56 },
		{ 197, 188, 56 },
		{ 198, 163, 56 },
		{ 198, 164, 56 },
		{ 198, 165, 56 },
		{ 198, 166, 56 },
		{ 198, 167, 56 },
		{ 198, 168, 56 },
		{ 198, 169, 56 },
		{ 198, 170, 56 },
		{ 198, 171, 56 },
		{ 198, 172, 56 },
		{ 198, 173, 56 },
		{ 198, 174, 56 },
		{ 198, 175, 56 },
		{ 198, 176, 56 },
		{ 198, 177, 56 },
		{ 198, 178, 56 },
		{ 198, 179, 56 },
		{ 198, 180, 56 },
		{ 198, 181, 56 },
		{ 198, 182, 56 },
		{ 198, 183, 56 },
		{ 198, 184, 56 },
		{ 198, 185, 56 },
		{ 198, 186, 56 },
		{ 198, 187, 56 },
		{ 198, 188, 56 },
		{ 198, 189, 56 },
		{ 199, 165, 56 },
		{ 199, 166, 56 },
		{ 199, 167, 56 },
		{ 199, 168, 56 },
		{ 199, 169, 56 },
		{ 199, 170, 56 },
		{ 199, 171, 56 },
		{ 199, 172, 56 },
		{ 199, 173, 56 },
		{ 199, 174, 56 },
		{ 199, 175, 56 },
		{ 199, 176, 56 },
		{ 199, 177, 56 },
		{ 199, 178, 56 },
		{ 199, 179, 56 },
		{ 199, 180, 56 },
		{ 199, 181, 56 },
		{ 199, 182, 56 },
		{ 199, 183, 56 },
		{ 199, 184, 56 },
		{ 199, 185, 56 },
		{ 199, 186, 56 },
		{ 199, 187, 56 },
		{ 199, 188, 56 },
		{ 199, 189, 56 },
		{ 200, 166, 56 },
		{ 200, 167, 56 },
		{ 200, 168, 56 },
		{ 200, 169, 56 },
		{ 200, 170, 56 },
		{ 200, 171, 56 },
		{ 200, 172, 56 },
		{ 200, 173, 56 },
		{ 200, 174, 56 },
		{ 200, 175, 56 },
		{ 200, 176, 56 },
		{ 200, 177, 56 },
		{ 200, 178, 56 },
		{ 200, 179, 56 },
		{ 200, 180, 56 },
		{ 200, 181, 56 },
		{ 200, 182, 56 },
		{ 200, 183, 56 },
		{ 200, 184, 56 },
		{ 200, 185, 56 },
		{ 200, 186, 56 },
		{ 200, 187, 56 },
		{ 200, 188, 56 },
		{ 200, 189, 56 },
		{ 200, 190, 56 },
		{ 201, 168, 56 },
		{ 201, 169, 56 },
		{ 201, 170, 56 },
		{ 201, 171, 56 },
		{ 201, 172, 56 },
		{ 201, 173, 56 },
		{ 201, 174, 56 },
		{ 201, 175, 56 },
		{ 201, 176, 56 },
		{ 201, 177, 56 },
		{ 201, 178, 56 },
		{ 201, 179, 56 },
		{ 201, 180, 56 },
		{ 201, 181, 56 },
		{ 201, 182, 56 },
		{ 201, 183, 56 },
		{ 201, 184, 56 },
		{ 201, 185, 56 },
		{ 201, 186, 56 },
		{ 201, 187, 56 },
		{ 201, 188, 56 },
		{ 201, 189, 56 },
		{ 201, 190, 56 },
		{ 202, 170, 56 },
		{ 202, 171, 56 },
		{ 202, 172, 56 },
		{ 202, 173, 56 },
		{ 202, 174, 56 },
		{ 202, 175, 56 },
		{ 202, 176, 56 },
		{ 202, 177, 56 },
		{ 202, 178, 56 },
		{ 202, 179, 56 },
		{ 202, 180, 56 },
		{ 202, 181, 56 },
		{ 202, 182, 56 },
		{ 202, 183, 56 },
		{ 202, 184, 56 },
		{ 202, 185, 56 },
		{ 202, 186, 56 },
		{ 202, 187, 56 },
		{ 202, 188, 56 },
		{ 202, 189, 56 },
		{ 202, 190, 56 },
		{ 202, 191, 56 },
		{ 203, 171, 56 },
		{ 203, 172, 56 },
		{ 203, 173, 56 },
		{ 203, 174, 56 },
		{ 203, 175, 56 },
		{ 203, 176, 56 },
		{ 203, 177, 56 },
		{ 203, 178, 56 },
		{ 203, 179, 56 },
		{ 203, 180, 56 },
		{ 203, 181, 56 },
		{ 203, 182, 56 },
		{ 203, 183, 56 },
		{ 203, 184, 56 },
		{ 203, 185, 56 },
		{ 203, 186, 56 },
		{ 203, 187, 56 },
		{ 203, 188, 56 },
		{ 203, 189, 56 },
		{ 203, 190, 56 },
		{ 203, 191, 56 },
		{ 204, 173, 56 },
		{ 204, 174, 56 },
		{ 204, 175, 56 },
		{ 204, 176, 56 },
		{ 204, 177, 56 },
		{ 204, 178, 56 },
		{ 204, 179, 56 },
		{ 204, 180, 56 },
		{ 204, 181, 56 },
		{ 204, 182, 56 },
		{ 204, 183, 56 },
		{ 204, 184, 56 },
		{ 204, 185, 56 },
		{ 204, 186, 56 },
		{ 204, 187, 56 },
		{ 204, 188, 56 },
		{ 204, 189, 56 },
		{ 204, 190, 56 },
		{ 204, 191, 56 },
		{ 205, 174, 56 },
		{ 205, 175, 56 },
		{ 205, 176, 56 },
		{ 205, 177, 56 },
		{ 205, 178, 56 },
		{ 205, 179, 56 },
		{ 205, 180, 56 },
		{ 205, 181, 56 },
		{ 205, 182, 56 },
		{ 205, 183, 56 },
		{ 205, 184, 56 },
		{ 205, 185, 56 },
		{ 205, 186, 56 },
		{ 205, 187, 56 },
		{ 205, 188, 56 },
		{ 205, 189, 56 },
		{ 205, 190, 56 },
		{ 205, 191, 56 },
		{ 206, 176, 56 },
		{ 206, 177, 56 },
		{ 206, 178, 56 },
		{ 206, 179, 56 },
		{ 206, 180, 56 },
		{ 206, 181, 56 },
		{ 206, 182, 56 },
		{ 206, 183, 56 },
		{ 206, 184, 56 },
		{ 206, 185, 56 },
		{ 206, 186, 56 },
		{ 206, 187, 56 },
		{ 206, 188, 56 },
		{ 206, 189, 56 },
		{ 206, 190, 56 },
		{ 206, 191, 56 },
		{ 207, 178, 56 },
		{ 207, 179, 56 },
		{ 207, 180, 56 },
		{ 207, 181, 56 },
		{ 207, 182, 56 },
		{ 207, 183, 56 },
		{ 207, 184, 56 },
		{ 207, 185, 56 },
		{ 207, 186, 56 },
		{ 207, 187, 56 },
		{ 207, 188, 56 },
		{ 207, 189, 56 },
		{ 207, 190, 56 },
		{ 207, 191, 56 },
		{ 208, 179, 56 },
		{ 208, 180, 56 },
		{ 208, 181, 56 },
		{ 208, 182, 56 },
		{ 208, 183, 56 },
		{ 208, 184, 56 },
		{ 208, 185, 56 },
		{ 208, 186, 56 },
		{ 208, 187, 56 },
		{ 208, 188, 56 },
		{ 208, 189, 56 },
		{ 208, 190, 56 },
		{ 208, 191, 56 },
		{ 209, 181, 56 },
		{ 209, 182, 56 },
		{ 209, 183, 56 },
		{ 209, 184, 56 },
		{ 209, 185, 56 },
		{ 209, 186, 56 },
		{ 209, 187, 56 },
		{ 209, 188, 56 },
		{ 209, 189, 56 },
		{ 209, 190, 56 },
		{ 209, 191, 56 },
		{ 210, 182, 56 },
		{ 210, 183, 56 },
		{ 210, 184, 56 },
		{ 210, 185, 56 },
		{ 210, 186, 56 },
		{ 210, 187, 56 },
		{ 210, 188, 56 },
		{ 210, 189, 56 },
		{ 210, 190, 56 },
		{ 210, 191, 56 },
		{ 211, 184, 56 },
		{ 211, 185, 56 },
		{ 211, 186, 56 },
		{ 211, 187, 56 },
		{ 211, 188, 56 },
		{ 211, 189, 56 },
		{ 211, 190, 56 },
		{ 211, 191, 56 },
		{ 212, 186, 56 },
		{ 212, 187, 56 },
		{ 212, 188, 56 },
		{ 212, 189, 56 },
		{ 212, 190, 56 },
		{ 212, 191, 56 },
		{ 213, 187, 56 },
		{ 213, 188, 56 },
		{ 213, 189, 56 },
		{ 213, 190, 56 },
		{ 213, 191, 56 },
		{ 214, 189, 56 },
		{ 214, 190, 56 },
		{ 300, 324, 56 },
		{ 300, 325, 56 },
		{ 301, 323, 56 },
		{ 301, 324, 56 },
		{ 301, 325, 56 },
		{ 301, 326, 56 },
		{ 301, 327, 56 },
		{ 302, 323, 56 },
		{ 302, 324, 56 },
		{ 302, 325, 56 },
		{ 302, 326, 56 },
		{ 302, 327, 56 },
		{ 302, 328, 56 },
		{ 303, 323, 56 },
		{ 303, 324, 56 },
		{ 303, 325, 56 },
		{ 303, 326, 56 },
		{ 303, 327, 56 },
		{ 303, 328, 56 },
		{ 303, 329, 56 },
		{ 303, 330, 56 },
		{ 304, 323, 56 },
		{ 304, 324, 56 },
		{ 304, 325, 56 },
		{ 304, 326, 56 },
		{ 304, 327, 56 },
		{ 304, 328, 56 },
		{ 304, 329, 56 },
		{ 304, 330, 56 },
		{ 304, 331, 56 },
		{ 304, 332, 56 },
		{ 305, 323, 56 },
		{ 305, 324, 56 },
		{ 305, 325, 56 },
		{ 305, 326, 56 },
		{ 305, 327, 56 },
		{ 305, 328, 56 },
		{ 305, 329, 56 },
		{ 305, 330, 56 },
		{ 305, 331, 56 },
		{ 305, 332, 56 },
		{ 305, 333, 56 },
		{ 306, 323, 56 },
		{ 306, 324, 56 },
		{ 306, 325, 56 },
		{ 306, 326, 56 },
		{ 306, 327, 56 },
		{ 306, 328, 56 },
		{ 306, 329, 56 },
		{ 306, 330, 56 },
		{ 306, 331, 56 },
		{ 306, 332, 56 },
		{ 306, 333, 56 },
		{ 306, 334, 56 },
		{ 306, 335, 56 },
		{ 307, 323, 56 },
		{ 307, 324, 56 },
		{ 307, 325, 56 },
		{ 307, 326, 56 },
		{ 307, 327, 56 },
		{ 307, 328, 56 },
		{ 307, 329, 56 },
		{ 307, 330, 56 },
		{ 307, 331, 56 },
		{ 307, 332, 56 },
		{ 307, 333, 56 },
		{ 307, 334, 56 },
		{ 307, 335, 56 },
		{ 307, 336, 56 },
		{ 308, 323, 56 },
		{ 308, 324, 56 },
		{ 308, 325, 56 },
		{ 308, 326, 56 },
		{ 308, 327, 56 },
		{ 308, 328, 56 },
		{ 308, 329, 56 },
		{ 308, 330, 56 },
		{ 308, 331, 56 },
		{ 308, 332, 56 },
		{ 308, 333, 56 },
		{ 308, 334, 56 },
		{ 308, 335, 56 },
		{ 308, 336, 56 },
		{ 308, 337, 56 },
		{ 308, 338, 56 },
		{ 309, 323, 56 },
		{ 309, 324, 56 },
		{ 309, 325, 56 },
		{ 309, 326, 56 },
		{ 309, 327, 56 },
		{ 309, 328, 56 },
		{ 309, 329, 56 },
		{ 309, 330, 56 },
		{ 309, 331, 56 },
		{ 309, 332, 56 },
		{ 309, 333, 56 },
		{ 309, 334, 56 },
		{ 309, 335, 56 },
		{ 309, 336, 56 },
		{ 309, 337, 56 },
		{ 309, 338, 56 },
		{ 309, 339, 56 },
		{ 309, 340, 56 },
		{ 310, 323, 56 },
		{ 310, 324, 56 },
		{ 310, 325, 56 },
		{ 310, 326, 56 },
		{ 310, 327, 56 },
		{ 310, 328, 56 },
		{ 310, 329, 56 },
		{ 310, 330, 56 },
		{ 310, 331, 56 },
		{ 310, 332, 56 },
		{ 310, 333, 56 },
		{ 310, 334, 56 },
		{ 310, 335, 56 },
		{ 310, 336, 56 },
		{ 310, 337, 56 },
		{ 310, 338, 56 },
		{ 310, 339, 56 },
		{ 310, 340, 56 },
		{ 310, 341, 56 },
		{ 311, 323, 56 },
		{ 311, 324, 56 },
		{ 311, 325, 56 },
		{ 311, 326, 56 },
		{ 311, 327, 56 },
		{ 311, 328, 56 },
		{ 311, 329, 56 },
		{ 311, 330, 56 },
		{ 311, 331, 56 },
		{ 311, 332, 56 },
		{ 311, 333, 56 },
		{ 311, 334, 56 },
		{ 311, 335, 56 },
		{ 311, 336, 56 },
		{ 311, 337, 56 },
		{ 311, 338, 56 },
		{ 311, 339, 56 },
		{ 311, 340, 56 },
		{ 311, 341, 56 },
		{ 311, 342, 56 },
		{ 311, 343, 56 },
		{ 312, 323, 56 },
		{ 312, 324, 56 },
		{ 312, 325, 56 },
		{ 312, 326, 56 },
		{ 312, 327, 56 },
		{ 312, 328, 56 },
		{ 312, 329, 56 },
		{ 312, 330, 56 },
		{ 312, 331, 56 },
		{ 312, 332, 56 },
		{ 312, 333, 56 },
		{ 312, 334, 56 },
		{ 312, 335, 56 },
		{ 312, 336, 56 },
		{ 312, 337, 56 },
		{ 312, 338, 56 },
		{ 312, 339, 56 },
		{ 312, 340, 56 },
		{ 312, 341, 56 },
		{ 312, 342, 56 },
		{ 312, 343, 56 },
		{ 312, 344, 56 },
		{ 313, 324, 56 },
		{ 313, 325, 56 },
		{ 313, 326, 56 },
		{ 313, 327, 56 },
		{ 313, 328, 56 },
		{ 313, 329, 56 },
		{ 313, 330, 56 },
		{ 313, 331, 56 },
		{ 313, 332, 56 },
		{ 313, 333, 56 },
		{ 313, 334, 56 },
		{ 313, 335, 56 },
		{ 313, 336, 56 },
		{ 313, 337, 56 },
		{ 313, 338, 56 },
		{ 313, 339, 56 },
		{ 313, 340, 56 },
		{ 313, 341, 56 },
		{ 313, 342, 56 },
		{ 313, 343, 56 },
		{ 313, 344, 56 },
		{ 313, 345, 56 },
		{ 313, 346, 56 },
		{ 314, 324, 56 },
		{ 314, 325, 56 },
		{ 314, 326, 56 },
		{ 314, 327, 56 },
		{ 314, 328, 56 },
		{ 314, 329, 56 },
		{ 314, 330, 56 },
		{ 314, 331, 56 },
		{ 314, 332, 56 },
		{ 314, 333, 56 },
		{ 314, 334, 56 },
		{ 314, 335, 56 },
		{ 314, 336, 56 },
		{ 314, 337, 56 },
		{ 314, 338, 56 },
		{ 314, 339, 56 },
		{ 314, 340, 56 },
		{ 314, 341, 56 },
		{ 314, 342, 56 },
		{ 314, 343, 56 },
		{ 314, 344, 56 },
		{ 314, 345, 56 },
		{ 314, 346, 56 },
		{ 314, 347, 56 },
		{ 314, 348, 56 },
		{ 315, 325, 56 },
		{ 315, 326, 56 },
		{ 315, 327, 56 },
		{ 315, 328, 56 },
		{ 315, 329, 56 },
		{ 315, 330, 56 },
		{ 315, 331, 56 },
		{ 315, 332, 56 },
		{ 315, 333, 56 },
		{ 315, 334, 56 },
		{ 315, 335, 56 },
		{ 315, 336, 56 },
		{ 315, 337, 56 },
		{ 315, 338, 56 },
		{ 315, 339, 56 },
		{ 315, 340, 56 },
		{ 315, 341, 56 },
		{ 315, 342, 56 },
		{ 315, 343, 56 },
		{ 315, 344, 56 },
		{ 315, 345, 56 },
		{ 315, 346, 56 },
		{ 315, 347, 56 },
		{ 315, 348, 56 },
		{ 315, 349, 56 },
		{ 316, 325, 56 },
		{ 316, 326, 56 },
		{ 316, 327, 56 },
		{ 316, 328, 56 },
		{ 316, 329, 56 },
		{ 316, 330, 56 },
		{ 316, 331, 56 },
		{ 316, 332, 56 },
		{ 316, 333, 56 },
		{ 316, 334, 56 },
		{ 316, 335, 56 },
		{ 316, 336, 56 },
		{ 316, 337, 56 },
		{ 316, 338, 56 },
		{ 316, 339, 56 },
		{ 316, 340, 56 },
		{ 316, 341, 56 },
		{ 316, 342, 56 },
		{ 316, 343, 56 },
		{ 316, 344, 56 },
		{ 316, 345, 56 },
		{ 316, 346, 56 },
		{ 316, 347, 56 },
		{ 316, 348, 56 },
		{ 316, 349, 56 },
		{ 316, 350, 56 },
		{ 316, 351, 56 },
		{ 317, 326, 56 },
		{ 317, 327, 56 },
		{ 317, 328, 56 },
		{ 317, 329, 56 },
		{ 317, 330, 56 },
		{ 317, 331, 56 },
		{ 317, 332, 56 },
		{ 317, 333, 56 },
		{ 317, 334, 56 },
		{ 317, 335, 56 },
		{ 317, 336, 56 },
		{ 317, 337, 56 },
		{ 317, 338, 56 },
		{ 317, 339, 56 },
		{ 317, 340, 56 },
		{ 317, 341, 56 },
		{ 317, 342, 56 },
		{ 317, 343, 56 },
		{ 317, 344, 56 },
		{ 317, 345, 56 },
		{ 317, 346, 56 },
		{ 317, 347, 56 },
		{ 317, 348, 56 },
		{ 317, 349, 56 },
		{ 317, 350, 56 },
		{ 317, 351, 56 },
		{ 317, 352, 56 },
		{ 318, 327, 56 },
		{ 318, 328, 56 },
		{ 318, 329, 56 },
		{ 318, 330, 56 },
		{ 318, 331, 56 },
		{ 318, 332, 56 },
		{ 318, 333, 56 },
		{ 318, 334, 56 },
		{ 318, 335, 56 },
		{ 318, 336, 56 },
		{ 318, 337, 56 },
		{ 318, 338, 56 },
		{ 318, 339, 56 },
		{ 318, 340, 56 },
		{ 318, 341, 56 },
		{ 318, 342, 56 },
		{ 318, 343, 56 },
		{ 318, 344, 56 },
		{ 318, 345, 56 },
		{ 318, 346, 56 },
		{ 318, 347, 56 },
		{ 318, 348, 56 },
		{ 318, 349, 56 },
		{ 318, 350, 56 },
		{ 318, 351, 56 },
		{ 318, 352, 56 },
		{ 318, 353, 56 },
		{ 318, 354, 56 },
		{ 319, 329, 56 },
		{ 319, 330, 56 },
		{ 319, 331, 56 },
		{ 319, 332, 56 },
		{ 319, 333, 56 },
		{ 319, 334, 56 },
		{ 319, 335, 56 },
		{ 319, 336, 56 },
		{ 319, 337, 56 },
		{ 319, 338, 56 },
		{ 319, 339, 56 },
		{ 319, 340, 56 },
		{ 319, 341, 56 },
		{ 319, 342, 56 },
		{ 319, 343, 56 },
		{ 319, 344, 56 },
		{ 319, 345, 56 },
		{ 319, 346, 56 },
		{ 319, 347, 56 },
		{ 319, 348, 56 },
		{ 319, 349, 56 },
		{ 319, 350, 56 },
		{ 319, 351, 56 },
		{ 319, 352, 56 },
		{ 319, 353, 56 },
		{ 319, 354, 56 },
		{ 319, 355, 56 },
		{ 319, 356, 56 },
		{ 320, 330, 56 },
		{ 320, 331, 56 },
		{ 320, 332, 56 },
		{ 320, 333, 56 },
		{ 320, 334, 56 },
		{ 320, 335, 56 },
		{ 320, 336, 56 },
		{ 320, 337, 56 },
		{ 320, 338, 56 },
		{ 320, 339, 56 },
		{ 320, 340, 56 },
		{ 320, 341, 56 },
		{ 320, 342, 56 },
		{ 320, 343, 56 },
		{ 320, 344, 56 },
		{ 320, 345, 56 },
		{ 320, 346, 56 },
		{ 320, 347, 56 },
		{ 320, 348, 56 },
		{ 320, 349, 56 },
		{ 320, 350, 56 },
		{ 320, 351, 56 },
		{ 320, 352, 56 },
		{ 320, 353, 56 },
		{ 320, 354, 56 },
		{ 320, 355, 56 },
		{ 320, 356, 56 },
		{ 320, 357, 56 },
		{ 321, 331, 56 },
		{ 321, 332, 56 },
		{ 321, 333, 56 },
		{ 321, 334, 56 },
		{ 321, 335, 56 },
		{ 321, 336, 56 },
		{ 321, 337, 56 },
		{ 321, 338, 56 },
		{ 321, 339, 56 },
		{ 321, 340, 56 },
		{ 321, 341, 56 },
		{ 321, 342, 56 },
		{ 321, 343, 56 },
		{ 321, 344, 56 },
		{ 321, 345, 56 },
		{ 321, 346, 56 },
		{ 321, 347, 56 },
		{ 321, 348, 56 },
		{ 321, 349, 56 },
		{ 321, 350, 56 },
		{ 321, 351, 56 },
		{ 321, 352, 56 },
		{ 321, 353, 56 },
		{ 321, 354, 56 },
		{ 321, 355, 56 },
		{ 321, 356, 56 },
		{ 321, 357, 56 },
		{ 321, 358, 56 },
		{ 321, 359, 56 },
		{ 322, 332, 56 },
		{ 322, 333, 56 },
		{ 322, 334, 56 },
		{ 322, 335, 56 },
		{ 322, 336, 56 },
		{ 322, 337, 56 },
		{ 322, 338, 56 },
		{ 322, 339, 56 },
		{ 322, 340, 56 },
		{ 322, 341, 56 },
		{ 322, 342, 56 },
		{ 322, 343, 56 },
		{ 322, 344, 56 },
		{ 322, 345, 56 },
		{ 322, 346, 56 },
		{ 322, 347, 56 },
		{ 322, 348, 56 },
		{ 322, 349, 56 },
		{ 322, 350, 56 },
		{ 322, 351, 56 },
		{ 322, 352, 56 },
		{ 322, 353, 56 },
		{ 322, 354, 56 },
		{ 322, 355, 56 },
		{ 322, 356, 56 },
		{ 322, 357, 56 },
		{ 322, 358, 56 },
		{ 322, 359, 56 },
		{ 322, 360, 56 },
		{ 323, 333, 56 },
		{ 323, 334, 56 },
		{ 323, 335, 56 },
		{ 323, 336, 56 },
		{ 323, 337, 56 },
		{ 323, 338, 56 },
		{ 323, 339, 56 },
		{ 323, 340, 56 },
		{ 323, 341, 56 },
		{ 323, 342, 56 },
		{ 323, 343, 56 },
		{ 323, 344, 56 },
		{ 323, 345, 56 },
		{ 323, 346, 56 },
		{ 323, 347, 56 },
		{ 323, 348, 56 },
		{ 323, 349, 56 },
		{ 323, 350, 56 },
		{ 323, 351, 56 },
		{ 323, 352, 56 },
		{ 323, 353, 56 },
		{ 323, 354, 56 },
		{ 323, 355, 56 },
		{ 323, 356, 56 },
		{ 323, 357, 56 },
		{ 323, 358, 56 },
		{ 323, 359, 56 },
		{ 323, 360, 56 },
		{ 323, 361, 56 },
		{ 323, 362, 56 },
		{ 324, 334, 56 },
		{ 324, 335, 56 },
		{ 324, 336, 56 },
		{ 324, 337, 56 },
		{ 324, 338, 56 },
		{ 324, 339, 56 },
		{ 324, 340, 56 },
		{ 324, 341, 56 },
		{ 324, 342, 56 },
		{ 324, 343, 56 },
		{ 324, 344, 56 },
		{ 324, 345, 56 },
		{ 324, 346, 56 },
		{ 324, 347, 56 },
		{ 324, 348, 56 },
		{ 324, 349, 56 },
		{ 324, 350, 56 },
		{ 324, 351, 56 },
		{ 324, 352, 56 },
		{ 324, 353, 56 },
		{ 324, 354, 56 },
		{ 324, 355, 56 },
		{ 324, 356, 56 },
		{ 324, 357, 56 },
		{ 324, 358, 56 },
		{ 324, 359, 56 },
		{ 324, 360, 56 },
		{ 324, 361, 56 },
		{ 324, 362, 56 },
		{ 324, 363, 56 },
		{ 324, 364, 56 },
		{ 325, 335, 56 },
		{ 325, 336, 56 },
		{ 325, 337, 56 },
		{ 325, 338, 56 },
		{ 325, 339, 56 },
		{ 325, 340, 56 },
		{ 325, 341, 56 },
		{ 325, 342, 56 },
		{ 325, 343, 56 },
		{ 325, 344, 56 },
		{ 325, 345, 56 },
		{ 325, 346, 56 },
		{ 325, 347, 56 },
		{ 325, 348, 56 },
		{ 325, 349, 56 },
		{ 325, 350, 56 },
		{ 325, 351, 56 },
		{ 325, 352, 56 },
		{ 325, 353, 56 },
		{ 325, 354, 56 },
		{ 325, 355, 56 },
		{ 325, 356, 56 },
		{ 325, 357, 56 },
		{ 325, 358, 56 },
		{ 325, 359, 56 },
		{ 325, 360, 56 },
		{ 325, 361, 56 },
		{ 325, 362, 56 },
		{ 325, 363, 56 },
		{ 325, 364, 56 },
		{ 325, 365, 56 },
		{ 326, 336, 56 },
		{ 326, 337, 56 },
		{ 326, 338, 56 },
		{ 326, 339, 56 },
		{ 326, 340, 56 },
		{ 326, 341, 56 },
		{ 326, 342, 56 },
		{ 326, 343, 56 },
		{ 326, 344, 56 },
		{ 326, 345, 56 },
		{ 326, 346, 56 },
		{ 326, 347, 56 },
		{ 326, 348, 56 },
		{ 326, 349, 56 },
		{ 326, 350, 56 },
		{ 326, 351, 56 },
		{ 326, 352, 56 },
		{ 326, 353, 56 },
		{ 326, 354, 56 },
		{ 326, 355, 56 },
		{ 326, 356, 56 },
		{ 326, 357, 56 },
		{ 326, 358, 56 },
		{ 326, 359, 56 },
		{ 326, 360, 56 },
		{ 326, 361, 56 },
		{ 326, 362, 56 },
		{ 326, 363, 56 },
		{ 326, 364, 56 },
		{ 326, 365, 56 },
		{ 326, 366, 56 },
		{ 326, 367, 56 },
		{ 327, 338, 56 },
		{ 327, 339, 56 },
		{ 327, 340, 56 },
		{ 327, 341, 56 },
		{ 327, 342, 56 },
		{ 327, 343, 56 },
		{ 327, 344, 56 },
		{ 327, 345, 56 },
		{ 327, 346, 56 },
		{ 327, 347, 56 },
		{ 327, 348, 56 },
		{ 327, 349, 56 },
		{ 327, 350, 56 },
		{ 327, 351, 56 },
		{ 327, 352, 56 },
		{ 327, 353, 56 },
		{ 327, 354, 56 },
		{ 327, 355, 56 },
		{ 327, 356, 56 },
		{ 327, 357, 56 },
		{ 327, 358, 56 },
		{ 327, 359, 56 },
		{ 327, 360, 56 },
		{ 327, 361, 56 },
		{ 327, 362, 56 },
		{ 327, 363, 56 },
		{ 327, 364, 56 },
		{ 327, 365, 56 },
		{ 327, 366, 56 },
		{ 327, 367, 56 },
		{ 327, 368, 56 },
		{ 328, 339, 56 },
		{ 328, 340, 56 },
		{ 328, 341, 56 },
		{ 328, 342, 56 },
		{ 328, 343, 56 },
		{ 328, 344, 56 },
		{ 328, 345, 56 },
		{ 328, 346, 56 },
		{ 328, 347, 56 },
		{ 328, 348, 56 },
		{ 328, 349, 56 },
		{ 328, 350, 56 },
		{ 328, 351, 56 },
		{ 328, 352, 56 },
		{ 328, 353, 56 },
		{ 328, 354, 56 },
		{ 328, 355, 56 },
		{ 328, 356, 56 },
		{ 328, 357, 56 },
		{ 328, 358, 56 },
		{ 328, 359, 56 },
		{ 328, 360, 56 },
		{ 328, 361, 56 },
		{ 328, 362, 56 },
		{ 328, 363, 56 },
		{ 328, 364, 56 },
		{ 328, 365, 56 },
		{ 328, 366, 56 },
		{ 328, 367, 56 },
		{ 328, 368, 56 },
		{ 328, 369, 56 },
		{ 328, 370, 56 },
		{ 329, 340, 56 },
		{ 329, 341, 56 },
		{ 329, 342, 56 },
		{ 329, 343, 56 },
		{ 329, 344, 56 },
		{ 329, 345, 56 },
		{ 329, 346, 56 },
		{ 329, 347, 56 },
		{ 329, 348, 56 },
		{ 329, 349, 56 },
		{ 329, 350, 56 },
		{ 329, 351, 56 },
		{ 329, 352, 56 },
		{ 329, 353, 56 },
		{ 329, 354, 56 },
		{ 329, 355, 56 },
		{ 329, 356, 56 },
		{ 329, 357, 56 },
		{ 329, 358, 56 },
		{ 329, 359, 56 },
		{ 329, 360, 56 },
		{ 329, 361, 56 },
		{ 329, 362, 56 },
		{ 329, 363, 56 },
		{ 329, 364, 56 },
		{ 329, 365, 56 },
		{ 329, 366, 56 },
		{ 329, 367, 56 },
		{ 329, 368, 56 },
		{ 329, 369, 56 },
		{ 329, 370, 56 },
		{ 329, 371, 56 },
		{ 329, 372, 56 },
		{ 330, 341, 56 },
		{ 330, 342, 56 },
		{ 330, 343, 56 },
		{ 330, 344, 56 },
		{ 330, 345, 56 },
		{ 330, 346, 56 },
		{ 330, 347, 56 },
		{ 330, 348, 56 },
		{ 330, 349, 56 },
		{ 330, 350, 56 },
		{ 330, 351, 56 },
		{ 330, 352, 56 },
		{ 330, 353, 56 },
		{ 330, 354, 56 },
		{ 330, 355, 56 },
		{ 330, 356, 56 },
		{ 330, 357, 56 },
		{ 330, 358, 56 },
		{ 330, 359, 56 },
		{ 330, 360, 56 },
		{ 330, 361, 56 },
		{ 330, 362, 56 },
		{ 330, 363, 56 },
		{ 330, 364, 56 },
		{ 330, 365, 56 },
		{ 330, 366, 56 },
		{ 330, 367, 56 },
		{ 330, 368, 56 },
		{ 330, 369, 56 },
		{ 330, 370, 56 },
		{ 330, 371, 56 },
		{ 330, 372, 56 },
		{ 330, 373, 56 },
		{ 331, 342, 56 },
		{ 331, 343, 56 },
		{ 331, 344, 56 },
		{ 331, 345, 56 },
		{ 331, 346, 56 },
		{ 331, 347, 56 },
		{ 331, 348, 56 },
		{ 331, 349, 56 },
		{ 331, 350, 56 },
		{ 331, 351, 56 },
		{ 331, 352, 56 },
		{ 331, 353, 56 },
		{ 331, 354, 56 },
		{ 331, 355, 56 },
		{ 331, 356, 56 },
		{ 331, 357, 56 },
		{ 331, 358, 56 },
		{ 331, 359, 56 },
		{ 331, 360, 56 },
		{ 331, 361, 56 },
		{ 331, 362, 56 },
		{ 331, 363, 56 },
		{ 331, 364, 56 },
		{ 331, 365, 56 },
		{ 331, 366, 56 },
		{ 331, 367, 56 },
		{ 331, 368, 56 },
		{ 331, 369, 56 },
		{ 331, 370, 56 },
		{ 331, 371, 56 },
		{ 331, 372, 56 },
		{ 331, 373, 56 },
		{ 331, 374, 56 },
		{ 331, 375, 56 },
		{ 332, 343, 56 },
		{ 332, 344, 56 },
		{ 332, 345, 56 },
		{ 332, 346, 56 },
		{ 332, 347, 56 },
		{ 332, 348, 56 },
		{ 332, 349, 56 },
		{ 332, 350, 56 },
		{ 332, 351, 56 },
		{ 332, 352, 56 },
		{ 332, 353, 56 },
		{ 332, 354, 56 },
		{ 332, 355, 56 },
		{ 332, 356, 56 },
		{ 332, 357, 56 },
		{ 332, 358, 56 },
		{ 332, 359, 56 },
		{ 332, 360, 56 },
		{ 332, 361, 56 },
		{ 332, 362, 56 },
		{ 332, 363, 56 },
		{ 332, 364, 56 },
		{ 332, 365, 56 },
		{ 332, 366, 56 },
		{ 332, 367, 56 },
		{ 332, 368, 56 },
		{ 332, 369, 56 },
		{ 332, 370, 56 },
		{ 332, 371, 56 },
		{ 332, 372, 56 },
		{ 332, 373, 56 },
		{ 332, 374, 56 },
		{ 332, 375, 56 },
		{ 332, 376, 56 },
		{ 333, 344, 56 },
		{ 333, 345, 56 },
		{ 333, 346, 56 },
		{ 333, 347, 56 },
		{ 333, 348, 56 },
		{ 333, 349, 56 },
		{ 333, 350, 56 },
		{ 333, 351, 56 },
		{ 333, 352, 56 },
		{ 333, 353, 56 },
		{ 333, 354, 56 },
		{ 333, 355, 56 },
		{ 333, 356, 56 },
		{ 333, 357, 56 },
		{ 333, 358, 56 },
		{ 333, 359, 56 },
		{ 333, 360, 56 },
		{ 333, 361, 56 },
		{ 333, 362, 56 },
		{ 333, 363, 56 },
		{ 333, 364, 56 },
		{ 333, 365, 56 },
		{ 333, 366, 56 },
		{ 333, 367, 56 },
		{ 333, 368, 56 },
		{ 333, 369, 56 },
		{ 333, 370, 56 },
		{ 333, 371, 56 },
		{ 333, 372, 56 },
		{ 333, 373, 56 },
		{ 333, 374, 56 },
		{ 333, 375, 56 },
		{ 333, 376, 56 },
		{ 333, 377, 56 },
		{ 333, 378, 56 },
		{ 334, 346, 56 },
		{ 334, 347, 56 },
		{ 334, 348, 56 },
		{ 334, 349, 56 },
		{ 334, 350, 56 },
		{ 334, 351, 56 },
		{ 334, 352, 56 },
		{ 334, 353, 56 },
		{ 334, 354, 56 },
		{ 334, 355, 56 },
		{ 334, 356, 56 },
		{ 334, 357, 56 },
		{ 334, 358, 56 },
		{ 334, 359, 56 },
		{ 334, 360, 56 },
		{ 334, 361, 56 },
		{ 334, 362, 56 },
		{ 334, 363, 56 },
		{ 334, 364, 56 },
		{ 334, 365, 56 },
		{ 334, 366, 56 },
		{ 334, 367, 56 },
		{ 334, 368, 56 },
		{ 334, 369, 56 },
		{ 334, 370, 56 },
		{ 334, 371, 56 },
		{ 334, 372, 56 },
		{ 334, 373, 56 },
		{ 334, 374, 56 },
		{ 334, 375, 56 },
		{ 334, 376, 56 },
		{ 334, 377, 56 },
		{ 334, 378, 56 },
		{ 334, 379, 56 },
		{ 334, 380, 56 },
		{ 335, 347, 56 },
		{ 335, 348, 56 },
		{ 335, 349, 56 },
		{ 335, 350, 56 },
		{ 335, 351, 56 },
		{ 335, 352, 56 },
		{ 335, 353, 56 },
		{ 335, 354, 56 },
		{ 335, 355, 56 },
		{ 335, 356, 56 },
		{ 335, 357, 56 },
		{ 335, 358, 56 },
		{ 335, 359, 56 },
		{ 335, 360, 56 },
		{ 335, 361, 56 },
		{ 335, 362, 56 },
		{ 335, 363, 56 },
		{ 335, 364, 56 },
		{ 335, 365, 56 },
		{ 335, 366, 56 },
		{ 335, 367, 56 },
		{ 335, 368, 56 },
		{ 335, 369, 56 },
		{ 335, 370, 56 },
		{ 335, 371, 56 },
		{ 335, 372, 56 },
		{ 335, 373, 56 },
		{ 335, 374, 56 },
		{ 335, 375, 56 },
		{ 335, 376, 56 },
		{ 335, 377, 56 },
		{ 335, 378, 56 },
		{ 335, 379, 56 },
		{ 335, 380, 56 },
		{ 335, 381, 56 },
		{ 336, 348, 56 },
		{ 336, 349, 56 },
		{ 336, 350, 56 },
		{ 336, 351, 56 },
		{ 336, 352, 56 },
		{ 336, 353, 56 },
		{ 336, 354, 56 },
		{ 336, 355, 56 },
		{ 336, 356, 56 },
		{ 336, 357, 56 },
		{ 336, 358, 56 },
		{ 336, 359, 56 },
		{ 336, 360, 56 },
		{ 336, 361, 56 },
		{ 336, 362, 56 },
		{ 336, 363, 56 },
		{ 336, 364, 56 },
		{ 336, 365, 56 },
		{ 336, 366, 56 },
		{ 336, 367, 56 },
		{ 336, 368, 56 },
		{ 336, 369, 56 },
		{ 336, 370, 56 },
		{ 336, 371, 56 },
		{ 336, 372, 56 },
		{ 336, 373, 56 },
		{ 336, 374, 56 },
		{ 336, 375, 56 },
		{ 336, 376, 56 },
		{ 336, 377, 56 },
		{ 336, 378, 56 },
		{ 336, 379, 56 },
		{ 336, 380, 56 },
		{ 336, 381, 56 },
		{ 336, 382, 56 },
		{ 336, 383, 56 },
		{ 337, 349, 56 },
		{ 337, 350, 56 },
		{ 337, 351, 56 },
		{ 337, 352, 56 },
		{ 337, 353, 56 },
		{ 337, 354, 56 },
		{ 337, 355, 56 },
		{ 337, 356, 56 },
		{ 337, 357, 56 },
		{ 337, 358, 56 },
		{ 337, 359, 56 },
		{ 337, 360, 56 },
		{ 337, 361, 56 },
		{ 337, 362, 56 },
		{ 337, 363, 56 },
		{ 337, 364, 56 },
		{ 337, 365, 56 },
		{ 337, 366, 56 },
		{ 337, 367, 56 },
		{ 337, 368, 56 },
		{ 337, 369, 56 },
		{ 337, 370, 56 },
		{ 337, 371, 56 },
		{ 337, 372, 56 },
		{ 337, 373, 56 },
		{ 337, 374, 56 },
		{ 337, 375, 56 },
		{ 337, 376, 56 },
		{ 337, 377, 56 },
		{ 337, 378, 56 },
		{ 337, 379, 56 },
		{ 337, 380, 56 },
		{ 337, 381, 56 },
		{ 337, 382, 56 },
		{ 337, 383, 56 },
		{ 337, 384, 56 },
		{ 338, 350, 56 },
		{ 338, 351, 56 },
		{ 338, 352, 56 },
		{ 338, 353, 56 },
		{ 338, 354, 56 },
		{ 338, 355, 56 },
		{ 338, 356, 56 },
		{ 338, 357, 56 },
		{ 338, 358, 56 },
		{ 338, 359, 56 },
		{ 338, 360, 56 },
		{ 338, 361, 56 },
		{ 338, 362, 56 },
		{ 338, 363, 56 },
		{ 338, 364, 56 },
		{ 338, 365, 56 },
		{ 338, 366, 56 },
		{ 338, 367, 56 },
		{ 338, 368, 56 },
		{ 338, 369, 56 },
		{ 338, 370, 56 },
		{ 338, 371, 56 },
		{ 338, 372, 56 },
		{ 338, 373, 56 },
		{ 338, 374, 56 },
		{ 338, 375, 56 },
		{ 338, 376, 56 },
		{ 338, 377, 56 },
		{ 338, 378, 56 },
		{ 338, 379, 56 },
		{ 338, 380, 56 },
		{ 338, 381, 56 },
		{ 338, 382, 56 },
		{ 338, 383, 56 },
		{ 338, 384, 56 },
		{ 338, 385, 56 },
		{ 338, 386, 56 },
		{ 339, 351, 56 },
		{ 339, 352, 56 },
		{ 339, 353, 56 },
		{ 339, 354, 56 },
		{ 339, 355, 56 },
		{ 339, 356, 56 },
		{ 339, 357, 56 },
		{ 339, 358, 56 },
		{ 339, 359, 56 },
		{ 339, 360, 56 },
		{ 339, 361, 56 },
		{ 339, 362, 56 },
		{ 339, 363, 56 },
		{ 339, 364, 56 },
		{ 339, 365, 56 },
		{ 339, 366, 56 },
		{ 339, 367, 56 },
		{ 339, 368, 56 },
		{ 339, 369, 56 },
		{ 339, 370, 56 },
		{ 339, 371, 56 },
		{ 339, 372, 56 },
		{ 339, 373, 56 },
		{ 339, 374, 56 },
		{ 339, 375, 56 },
		{ 339, 376, 56 },
		{ 339, 377, 56 },
		{ 339, 378, 56 },
		{ 339, 379, 56 },
		{ 339, 380, 56 },
		{ 339, 381, 56 },
		{ 339, 382, 56 },
		{ 339, 383, 56 },
		{ 339, 384, 56 },
		{ 339, 385, 56 },
		{ 339, 386, 56 },
		{ 339, 387, 56 },
		{ 339, 388, 56 },
		{ 340, 352, 56 },
		{ 340, 353, 56 },
		{ 340, 354, 56 },
		{ 340, 355, 56 },
		{ 340, 356, 56 },
		{ 340, 357, 56 },
		{ 340, 358, 56 },
		{ 340, 359, 56 },
		{ 340, 360, 56 },
		{ 340, 361, 56 },
		{ 340, 362, 56 },
		{ 340, 363, 56 },
		{ 340, 364, 56 },
		{ 340, 365, 56 },
		{ 340, 366, 56 },
		{ 340, 367, 56 },
		{ 340, 368, 56 },
		{ 340, 369, 56 },
		{ 340, 370, 56 },
		{ 340, 371, 56 },
		{ 340, 372, 56 },
		{ 340, 373, 56 },
		{ 340, 374, 56 },
		{ 340, 375, 56 },
		{ 340, 376, 56 },
		{ 340, 377, 56 },
		{ 340, 378, 56 },
		{ 340, 379, 56 },
		{ 340, 380, 56 },
		{ 340, 381, 56 },
		{ 340, 382, 56 },
		{ 340, 383, 56 },
		{ 340, 384, 56 },
		{ 340, 385, 56 },
		{ 340, 386, 56 },
		{ 340, 387, 56 },
		{ 340, 388, 56 },
		{ 340, 389, 56 },
		{ 341, 354, 56 },
		{ 341, 355, 56 },
		{ 341, 356, 56 },
		{ 341, 357, 56 },
		{ 341, 358, 56 },
		{ 341, 359, 56 },
		{ 341, 360, 56 },
		{ 341, 361, 56 },
		{ 341, 362, 56 },
		{ 341, 363, 56 },
		{ 341, 364, 56 },
		{ 341, 365, 56 },
		{ 341, 366, 56 },
		{ 341, 367, 56 },
		{ 341, 368, 56 },
		{ 341, 369, 56 },
		{ 341, 370, 56 },
		{ 341, 371, 56 },
		{ 341, 372, 56 },
		{ 341, 373, 56 },
		{ 341, 374, 56 },
		{ 341, 375, 56 },
		{ 341, 376, 56 },
		{ 341, 377, 56 },
		{ 341, 378, 56 },
		{ 341, 379, 56 },
		{ 341, 380, 56 },
		{ 341, 381, 56 },
		{ 341, 382, 56 },
		{ 341, 383, 56 },
		{ 341, 384, 56 },
		{ 341, 385, 56 },
		{ 341, 386, 56 },
		{ 341, 387, 56 },
		{ 341, 388, 56 },
		{ 341, 389, 56 },
		{ 341, 390, 56 },
		{ 341, 391, 56 },
		{ 342, 355, 56 },
		{ 342, 356, 56 },
		{ 342, 357, 56 },
		{ 342, 358, 56 },
		{ 342, 359, 56 },
		{ 342, 360, 56 },
		{ 342, 361, 56 },
		{ 342, 362, 56 },
		{ 342, 363, 56 },
		{ 342, 364, 56 },
		{ 342, 365, 56 },
		{ 342, 366, 56 },
		{ 342, 367, 56 },
		{ 342, 368, 56 },
		{ 342, 369, 56 },
		{ 342, 370, 56 },
		{ 342, 371, 56 },
		{ 342, 372, 56 },
		{ 342, 373, 56 },
		{ 342, 374, 56 },
		{ 342, 375, 56 },
		{ 342, 376, 56 },
		{ 342, 377, 56 },
		{ 342, 378, 56 },
		{ 342, 379, 56 },
		{ 342, 380, 56 },
		{ 342, 381, 56 },
		{ 342, 382, 56 },
		{ 342, 383, 56 },
		{ 342, 384, 56 },
		{ 342, 385, 56 },
		{ 342, 386, 56 },
		{ 342, 387, 56 },
		{ 342, 388, 56 },
		{ 342, 389, 56 },
		{ 342, 390, 56 },
		{ 342, 391, 56 },
		{ 342, 392, 56 },
		{ 343, 356, 56 },
		{ 343, 357, 56 },
		{ 343, 358, 56 },
		{ 343, 359, 56 },
		{ 343, 360, 56 },
		{ 343, 361, 56 },
		{ 343, 362, 56 },
		{ 343, 363, 56 },
		{ 343, 364, 56 },
		{ 343, 365, 56 },
		{ 343, 366, 56 },
		{ 343, 367, 56 },
		{ 343, 368, 56 },
		{ 343, 369, 56 },
		{ 343, 370, 56 },
		{ 343, 371, 56 },
		{ 343, 372, 56 },
		{ 343, 373, 56 },
		{ 343, 374, 56 },
		{ 343, 375, 56 },
		{ 343, 376, 56 },
		{ 343, 377, 56 },
		{ 343, 378, 56 },
		{ 343, 379, 56 },
		{ 343, 380, 56 },
		{ 343, 381, 56 },
		{ 343, 382, 56 },
		{ 343, 383, 56 },
		{ 343, 384, 56 },
		{ 343, 385, 56 },
		{ 343, 386, 56 },
		{ 343, 387, 56 },
		{ 343, 388, 56 },
		{ 343, 389, 56 },
		{ 343, 390, 56 },
		{ 343, 391, 56 },
		{ 343, 392, 56 },
		{ 343, 393, 56 },
		{ 343, 394, 56 },
		{ 344, 357, 56 },
		{ 344, 358, 56 },
		{ 344, 359, 56 },
		{ 344, 360, 56 },
		{ 344, 361, 56 },
		{ 344, 362, 56 },
		{ 344, 363, 56 },
		{ 344, 364, 56 },
		{ 344, 365, 56 },
		{ 344, 366, 56 },
		{ 344, 367, 56 },
		{ 344, 368, 56 },
		{ 344, 369, 56 },
		{ 344, 370, 56 },
		{ 344, 371, 56 },
		{ 344, 372, 56 },
		{ 344, 373, 56 },
		{ 344, 374, 56 },
		{ 344, 375, 56 },
		{ 344, 376, 56 },
		{ 344, 377, 56 },
		{ 344, 378, 56 },
		{ 344, 379, 56 },
		{ 344, 380, 56 },
		{ 344, 381, 56 },
		{ 344, 382, 56 },
		{ 344, 383, 56 },
		{ 344, 384, 56 },
		{ 344, 385, 56 },
		{ 344, 386, 56 },
		{ 344, 387, 56 },
		{ 344, 388, 56 },
		{ 344, 389, 56 },
		{ 344, 390, 56 },
		{ 344, 391, 56 },
		{ 344, 392, 56 },
		{ 344, 393, 56 },
		{ 344, 394, 56 },
		{ 344, 395, 56 },
		{ 344, 396, 56 },
		{ 345, 358, 56 },
		{ 345, 359, 56 },
		{ 345, 360, 56 },
		{ 345, 361, 56 },
		{ 345, 362, 56 },
		{ 345, 363, 56 },
		{ 345, 364, 56 },
		{ 345, 365, 56 },
		{ 345, 366, 56 },
		{ 345, 367, 56 },
		{ 345, 368, 56 },
		{ 345, 369, 56 },
		{ 345, 370, 56 },
		{ 345, 371, 56 },
		{ 345, 372, 56 },
		{ 345, 373, 56 },
		{ 345, 374, 56 },
		{ 345, 375, 56 },
		{ 345, 376, 56 },
		{ 345, 377, 56 },
		{ 345, 378, 56 },
		{ 345, 379, 56 },
		{ 345, 380, 56 },
		{ 345, 381, 56 },
		{ 345, 382, 56 },
		{ 345, 383, 56 },
		{ 345, 384, 56 },
		{ 345, 385, 56 },
		{ 345, 386, 56 },
		{ 345, 387, 56 },
		{ 345, 388, 56 },
		{ 345, 389, 56 },
		{ 345, 390, 56 },
		{ 345, 391, 56 },
		{ 345, 392, 56 },
		{ 345, 393, 56 },
		{ 345, 394, 56 },
		{ 345, 395, 56 },
		{ 345, 396, 56 },
		{ 345, 397, 56 },
		{ 346, 359, 56 },
		{ 346, 360, 56 },
		{ 346, 361, 56 },
		{ 346, 362, 56 },
		{ 346, 363, 56 },
		{ 346, 364, 56 },
		{ 346, 365, 56 },
		{ 346, 366, 56 },
		{ 346, 367, 56 },
		{ 346, 368, 56 },
		{ 346, 369, 56 },
		{ 346, 370, 56 },
		{ 346, 371, 56 },
		{ 346, 372, 56 },
		{ 346, 373, 56 },
		{ 346, 374, 56 },
		{ 346, 375, 56 },
		{ 346, 376, 56 },
		{ 346, 377, 56 },
		{ 346, 378, 56 },
		{ 346, 379, 56 },
		{ 346, 380, 56 },
		{ 346, 381, 56 },
		{ 346, 382, 56 },
		{ 346, 383, 56 },
		{ 346, 384, 56 },
		{ 346, 385, 56 },
		{ 346, 386, 56 },
		{ 346, 387, 56 },
		{ 346, 388, 56 },
		{ 346, 389, 56 },
		{ 346, 390, 56 },
		{ 346, 391, 56 },
		{ 346, 392, 56 },
		{ 346, 393, 56 },
		{ 346, 394, 56 },
		{ 346, 395, 56 },
		{ 346, 396, 56 },
		{ 346, 397, 56 },
		{ 346, 398, 56 },
		{ 346, 399, 56 },
		{ 347, 360, 56 },
		{ 347, 361, 56 },
		{ 347, 362, 56 },
		{ 347, 363, 56 },
		{ 347, 364, 56 },
		{ 347, 365, 56 },
		{ 347, 366, 56 },
		{ 347, 367, 56 },
		{ 347, 368, 56 },
		{ 347, 369, 56 },
		{ 347, 370, 56 },
		{ 347, 371, 56 },
		{ 347, 372, 56 },
		{ 347, 373, 56 },
		{ 347, 374, 56 },
		{ 347, 375, 56 },
		{ 347, 376, 56 },
		{ 347, 377, 56 },
		{ 347, 378, 56 },
		{ 347, 379, 56 },
		{ 347, 380, 56 },
		{ 347, 381, 56 },
		{ 347, 382, 56 },
		{ 347, 383, 56 },
		{ 347, 384, 56 },
		{ 347, 385, 56 },
		{ 347, 386, 56 },
		{ 347, 387, 56 },
		{ 347, 388, 56 },
		{ 347, 389, 56 },
		{ 347, 390, 56 },
		{ 347, 391, 56 },
		{ 347, 392, 56 },
		{ 347, 393, 56 },
		{ 347, 394, 56 },
		{ 347, 395, 56 },
		{ 347, 396, 56 },
		{ 347, 397, 56 },
		{ 347, 398, 56 },
		{ 347, 399, 56 },
		{ 347, 400, 56 },
		{ 348, 362, 56 },
		{ 348, 363, 56 },
		{ 348, 364, 56 },
		{ 348, 365, 56 },
		{ 348, 366, 56 },
		{ 348, 367, 56 },
		{ 348, 368, 56 },
		{ 348, 369, 56 },
		{ 348, 370, 56 },
		{ 348, 371, 56 },
		{ 348, 372, 56 },
		{ 348, 373, 56 },
		{ 348, 374, 56 },
		{ 348, 375, 56 },
		{ 348, 376, 56 },
		{ 348, 377, 56 },
		{ 348, 378, 56 },
		{ 348, 379, 56 },
		{ 348, 380, 56 },
		{ 348, 381, 56 },
		{ 348, 382, 56 },
		{ 348, 383, 56 },
		{ 348, 384, 56 },
		{ 348, 385, 56 },
		{ 348, 386, 56 },
		{ 348, 387, 56 },
		{ 348, 388, 56 },
		{ 348, 389, 56 },
		{ 348, 390, 56 },
		{ 348, 391, 56 },
		{ 348, 392, 56 },
		{ 348, 393, 56 },
		{ 348, 394, 56 },
		{ 348, 395, 56 },
		{ 348, 396, 56 },
		{ 348, 397, 56 },
		{ 348, 398, 56 },
		{ 348, 399, 56 },
		{ 348, 400, 56 },
		{ 348, 401, 56 },
		{ 348, 402, 56 },
		{ 349, 363, 56 },
		{ 349, 364, 56 },
		{ 349, 365, 56 },
		{ 349, 366, 56 },
		{ 349, 367, 56 },
		{ 349, 368, 56 },
		{ 349, 369, 56 },
		{ 349, 370, 56 },
		{ 349, 371, 56 },
		{ 349, 372, 56 },
		{ 349, 373, 56 },
		{ 349, 374, 56 },
		{ 349, 375, 56 },
		{ 349, 376, 56 },
		{ 349, 377, 56 },
		{ 349, 378, 56 },
		{ 349, 379, 56 },
		{ 349, 380, 56 },
		{ 349, 381, 56 },
		{ 349, 382, 56 },
		{ 349, 383, 56 },
		{ 349, 384, 56 },
		{ 349, 385, 56 },
		{ 349, 386, 56 },
		{ 349, 387, 56 },
		{ 349, 388, 56 },
		{ 349, 389, 56 },
		{ 349, 390, 56 },
		{ 349, 391, 56 },
		{ 349, 392, 56 },
		{ 349, 393, 56 },
		{ 349, 394, 56 },
		{ 349, 395, 56 },
		{ 349, 396, 56 },
		{ 349, 397, 56 },
		{ 349, 398, 56 },
		{ 349, 399, 56 },
		{ 349, 400, 56 },
		{ 349, 401, 56 },
		{ 349, 402, 56 },
		{ 349, 403, 56 },
		{ 349, 404, 56 },
		{ 350, 364, 56 },
		{ 350, 365, 56 },
		{ 350, 366, 56 },
		{ 350, 367, 56 },
		{ 350, 368, 56 },
		{ 350, 369, 56 },
		{ 350, 370, 56 },
		{ 350, 371, 56 },
		{ 350, 372, 56 },
		{ 350, 373, 56 },
		{ 350, 374, 56 },
		{ 350, 375, 56 },
		{ 350, 376, 56 },
		{ 350, 377, 56 },
		{ 350, 378, 56 },
		{ 350, 379, 56 },
		{ 350, 380, 56 },
		{ 350, 381, 56 },
		{ 350, 382, 56 },
		{ 350, 383, 56 },
		{ 350, 384, 56 },
		{ 350, 385, 56 },
		{ 350, 386, 56 },
		{ 350, 387, 56 },
		{ 350, 388, 56 },
		{ 350, 389, 56 },
		{ 350, 390, 56 },
		{ 350, 391, 56 },
		{ 350, 392, 56 },
		{ 350, 393, 56 },
		{ 350, 394, 56 },
		{ 350, 395, 56 },
		{ 350, 396, 56 },
		{ 350, 397, 56 },
		{ 350, 398, 56 },
		{ 350, 399, 56 },
		{ 350, 400, 56 },
		{ 350, 401, 56 },
		{ 350, 402, 56 },
		{ 350, 403, 56 },
		{ 350, 404, 56 },
		{ 350, 405, 56 },
		{ 351, 365, 56 },
		{ 351, 366, 56 },
		{ 351, 367, 56 },
		{ 351, 368, 56 },
		{ 351, 369, 56 },
		{ 351, 370, 56 },
		{ 351, 371, 56 },
		{ 351, 372, 56 },
		{ 351, 373, 56 },
		{ 351, 374, 56 },
		{ 351, 375, 56 },
		{ 351, 376, 56 },
		{ 351, 377, 56 },
		{ 351, 378, 56 },
		{ 351, 379, 56 },
		{ 351, 380, 56 },
		{ 351, 381, 56 },
		{ 351, 382, 56 },
		{ 351, 383, 56 },
		{ 351, 384, 56 },
		{ 351, 385, 56 },
		{ 351, 386, 56 },
		{ 351, 387, 56 },
		{ 351, 388, 56 },
		{ 351, 389, 56 },
		{ 351, 390, 56 },
		{ 351, 391, 56 },
		{ 351, 392, 56 },
		{ 351, 393, 56 },
		{ 351, 394, 56 },
		{ 351, 395, 56 },
		{ 351, 396, 56 },
		{ 351, 397, 56 },
		{ 351, 398, 56 },
		{ 351, 399, 56 },
		{ 351, 400, 56 },
		{ 351, 401, 56 },
		{ 351, 402, 56 },
		{ 351, 403, 56 },
		{ 351, 404, 56 },
		{ 351, 405, 56 },
		{ 351, 406, 56 },
		{ 351, 407, 56 },
		{ 352, 366, 56 },
		{ 352, 367, 56 },
		{ 352, 368, 56 },
		{ 352, 369, 56 },
		{ 352, 370, 56 },
		{ 352, 371, 56 },
		{ 352, 372, 56 },
		{ 352, 373, 56 },
		{ 352, 374, 56 },
		{ 352, 375, 56 },
		{ 352, 376, 56 },
		{ 352, 377, 56 },
		{ 352, 378, 56 },
		{ 352, 379, 56 },
		{ 352, 380, 56 },
		{ 352, 381, 56 },
		{ 352, 382, 56 },
		{ 352, 383, 56 },
		{ 352, 384, 56 },
		{ 352, 385, 56 },
		{ 352, 386, 56 },
		{ 352, 387, 56 },
		{ 352, 388, 56 },
		{ 352, 389, 56 },
		{ 352, 390, 56 },
		{ 352, 391, 56 },
		{ 352, 392, 56 },
		{ 352, 393, 56 },
		{ 352, 394, 56 },
		{ 352, 395, 56 },
		{ 352, 396, 56 },
		{ 352, 397, 56 },
		{ 352, 398, 56 },
		{ 352, 399, 56 },
		{ 352, 400, 56 },
		{ 352, 401, 56 },
		{ 352, 402, 56 },
		{ 352, 403, 56 },
		{ 352, 404, 56 },
		{ 352, 405, 56 },
		{ 352, 406, 56 },
		{ 352, 407, 56 },
		{ 352, 408, 56 },
		{ 353, 367, 56 },
		{ 353, 368, 56 },
		{ 353, 369, 56 },
		{ 353, 370, 56 },
		{ 353, 371, 56 },
		{ 353, 372, 56 },
		{ 353, 373, 56 },
		{ 353, 374, 56 },
		{ 353, 375, 56 },
		{ 353, 376, 56 },
		{ 353, 377, 56 },
		{ 353, 378, 56 },
		{ 353, 379, 56 },
		{ 353, 380, 56 },
		{ 353, 381, 56 },
		{ 353, 382, 56 },
		{ 353, 383, 56 },
		{ 353, 384, 56 },
		{ 353, 385, 56 },
		{ 353, 386, 56 },
		{ 353, 387, 56 },
		{ 353, 388, 56 },
		{ 353, 389, 56 },
		{ 353, 390, 56 },
		{ 353, 391, 56 },
		{ 353, 392, 56 },
		{ 353, 393, 56 },
		{ 353, 394, 56 },
		{ 353, 395, 56 },
		{ 353, 396, 56 },
		{ 353, 397, 56 },
		{ 353, 398, 56 },
		{ 353, 399, 56 },
		{ 353, 400, 56 },
		{ 353, 401, 56 },
		{ 353, 402, 56 },
		{ 353, 403, 56 },
		{ 353, 404, 56 },
		{ 353, 405, 56 },
		{ 353, 406, 56 },
		{ 353, 407, 56 },
		{ 353, 408, 56 },
		{ 353, 409, 56 },
		{ 353, 410, 56 },
		{ 354, 368, 56 },
		{ 354, 369, 56 },
		{ 354, 370, 56 },
		{ 354, 371, 56 },
		{ 354, 372, 56 },
		{ 354, 373, 56 },
		{ 354, 374, 56 },
		{ 354, 375, 56 },
		{ 354, 376, 56 },
		{ 354, 377, 56 },
		{ 354, 378, 56 },
		{ 354, 379, 56 },
		{ 354, 380, 56 },
		{ 354, 381, 56 },
		{ 354, 382, 56 },
		{ 354, 383, 56 },
		{ 354, 384, 56 },
		{ 354, 385, 56 },
		{ 354, 386, 56 },
		{ 354, 387, 56 },
		{ 354, 388, 56 },
		{ 354, 389, 56 },
		{ 354, 390, 56 },
		{ 354, 391, 56 },
		{ 354, 392, 56 },
		{ 354, 393, 56 },
		{ 354, 394, 56 },
		{ 354, 395, 56 },
		{ 354, 396, 56 },
		{ 354, 397, 56 },
		{ 354, 398, 56 },
		{ 354, 399, 56 },
		{ 354, 400, 56 },
		{ 354, 401, 56 },
		{ 354, 402, 56 },
		{ 354, 403, 56 },
		{ 354, 404, 56 },
		{ 354, 405, 56 },
		{ 354, 406, 56 },
		{ 354, 407, 56 },
		{ 354, 408, 56 },
		{ 354, 409, 56 },
		{ 354, 410, 56 },
		{ 354, 411, 56 },
		{ 354, 412, 56 },
		{ 355, 370, 56 },
		{ 355, 371, 56 },
		{ 355, 372, 56 },
		{ 355, 373, 56 },
		{ 355, 374, 56 },
		{ 355, 375, 56 },
		{ 355, 376, 56 },
		{ 355, 377, 56 },
		{ 355, 378, 56 },
		{ 355, 379, 56 },
		{ 355, 380, 56 },
		{ 355, 381, 56 },
		{ 355, 382, 56 },
		{ 355, 383, 56 },
		{ 355, 384, 56 },
		{ 355, 385, 56 },
		{ 355, 386, 56 },
		{ 355, 387, 56 },
		{ 355, 388, 56 },
		{ 355, 389, 56 },
		{ 355, 390, 56 },
		{ 355, 391, 56 },
		{ 355, 392, 56 },
		{ 355, 393, 56 },
		{ 355, 394, 56 },
		{ 355, 395, 56 },
		{ 355, 396, 56 },
		{ 355, 397, 56 },
		{ 355, 398, 56 },
		{ 355, 399, 56 },
		{ 355, 400, 56 },
		{ 355, 401, 56 },
		{ 355, 402, 56 },
		{ 355, 403, 56 },
		{ 355, 404, 56 },
		{ 355, 405, 56 },
		{ 355, 406, 56 },
		{ 355, 407, 56 },
		{ 355, 408, 56 },
		{ 355, 409, 56 },
		{ 355, 410, 56 },
		{ 355, 411, 56 },
		{ 355, 412, 56 },
		{ 355, 413, 56 },
		{ 356, 371, 56 },
		{ 356, 372, 56 },
		{ 356, 373, 56 },
		{ 356, 374, 56 },
		{ 356, 375, 56 },
		{ 356, 376, 56 },
		{ 356, 377, 56 },
		{ 356, 378, 56 },
		{ 356, 379, 56 },
		{ 356, 380, 56 },
		{ 356, 381, 56 },
		{ 356, 382, 56 },
		{ 356, 383, 56 },
		{ 356, 384, 56 },
		{ 356, 385, 56 },
		{ 356, 386, 56 },
		{ 356, 387, 56 },
		{ 356, 388, 56 },
		{ 356, 389, 56 },
		{ 356, 390, 56 },
		{ 356, 391, 56 },
		{ 356, 392, 56 },
		{ 356, 393, 56 },
		{ 356, 394, 56 },
		{ 356, 395, 56 },
		{ 356, 396, 56 },
		{ 356, 397, 56 },
		{ 356, 398, 56 },
		{ 356, 399, 56 },
		{ 356, 400, 56 },
		{ 356, 401, 56 },
		{ 356, 402, 56 },
		{ 356, 403, 56 },
		{ 356, 404, 56 },
		{ 356, 405, 56 },
		{ 356, 406, 56 },
		{ 356, 407, 56 },
		{ 356, 408, 56 },
		{ 356, 409, 56 },
		{ 356, 410, 56 },
		{ 356, 411, 56 },
		{ 356, 412, 56 },
		{ 356, 413, 56 },
		{ 356, 414, 56 },
		{ 356, 415, 56 },
		{ 357, 372, 56 },
		{ 357, 373, 56 },
		{ 357, 374, 56 },
		{ 357, 375, 56 },
		{ 357, 376, 56 },
		{ 357, 377, 56 },
		{ 357, 378, 56 },
		{ 357, 379, 56 },
		{ 357, 380, 56 },
		{ 357, 381, 56 },
		{ 357, 382, 56 },
		{ 357, 383, 56 },
		{ 357, 384, 56 },
		{ 357, 385, 56 },
		{ 357, 386, 56 },
		{ 357, 387, 56 },
		{ 357, 388, 56 },
		{ 357, 389, 56 },
		{ 357, 390, 56 },
		{ 357, 391, 56 },
		{ 357, 392, 56 },
		{ 357, 393, 56 },
		{ 357, 394, 56 },
		{ 357, 395, 56 },
		{ 357, 396, 56 },
		{ 357, 397, 56 },
		{ 357, 398, 56 },
		{ 357, 399, 56 },
		{ 357, 400, 56 },
		{ 357, 401, 56 },
		{ 357, 402, 56 },
		{ 357, 403, 56 },
		{ 357, 404, 56 },
		{ 357, 405, 56 },
		{ 357, 406, 56 },
		{ 357, 407, 56 },
		{ 357, 408, 56 },
		{ 357, 409, 56 },
		{ 357, 410, 56 },
		{ 357, 411, 56 },
		{ 357, 412, 56 },
		{ 357, 413, 56 },
		{ 357, 414, 56 },
		{ 357, 415, 56 },
		{ 357, 416, 56 },
		{ 358, 373, 56 },
		{ 358, 374, 56 },
		{ 358, 375, 56 },
		{ 358, 376, 56 },
		{ 358, 377, 56 },
		{ 358, 378, 56 },
		{ 358, 379, 56 },
		{ 358, 380, 56 },
		{ 358, 381, 56 },
		{ 358, 382, 56 },
		{ 358, 383, 56 },
		{ 358, 384, 56 },
		{ 358, 385, 56 },
		{ 358, 386, 56 },
		{ 358, 387, 56 },
		{ 358, 388, 56 },
		{ 358, 389, 56 },
		{ 358, 390, 56 },
		{ 358, 391, 56 },
		{ 358, 392, 56 },
		{ 358, 393, 56 },
		{ 358, 394, 56 },
		{ 358, 395, 56 },
		{ 358, 396, 56 },
		{ 358, 397, 56 },
		{ 358, 398, 56 },
		{ 358, 399, 56 },
		{ 358, 400, 56 },
		{ 358, 401, 56 },
		{ 358, 402, 56 },
		{ 358, 403, 56 },
		{ 358, 404, 56 },
		{ 358, 405, 56 },
		{ 358, 406, 56 },
		{ 358, 407, 56 },
		{ 358, 408, 56 },
		{ 358, 409, 56 },
		{ 358, 410, 56 },
		{ 358, 411, 56 },
		{ 358, 412, 56 },
		{ 358, 413, 56 },
		{ 358, 414, 56 },
		{ 358, 415, 56 },
		{ 358, 416, 56 },
		{ 358, 417, 56 },
		{ 358, 418, 56 },
		{ 359, 374, 56 },
		{ 359, 375, 56 },
		{ 359, 376, 56 },
		{ 359, 377, 56 },
		{ 359, 378, 56 },
		{ 359, 379, 56 },
		{ 359, 380, 56 },
		{ 359, 381, 56 },
		{ 359, 382, 56 },
		{ 359, 383, 56 },
		{ 359, 384, 56 },
		{ 359, 385, 56 },
		{ 359, 386, 56 },
		{ 359, 387, 56 },
		{ 359, 388, 56 },
		{ 359, 389, 56 },
		{ 359, 390, 56 },
		{ 359, 391, 56 },
		{ 359, 392, 56 },
		{ 359, 393, 56 },
		{ 359, 394, 56 },
		{ 359, 395, 56 },
		{ 359, 396, 56 },
		{ 359, 397, 56 },
		{ 359, 398, 56 },
		{ 359, 399, 56 },
		{ 359, 400, 56 },
		{ 359, 401, 56 },
		{ 359, 402, 56 },
		{ 359, 403, 56 },
		{ 359, 404, 56 },
		{ 359, 405, 56 },
		{ 359, 406, 56 },
		{ 359, 407, 56 },
		{ 359, 408, 56 },
		{ 359, 409, 56 },
		{ 359, 410, 56 },
		{ 359, 411, 56 },
		{ 359, 412, 56 },
		{ 359, 413, 56 },
		{ 359, 414, 56 },
		{ 359, 415, 56 },
		{ 359, 416, 56 },
		{ 359, 417, 56 },
		{ 359, 418, 56 },
		{ 359, 419, 56 },
		{ 359, 420, 56 },
		{ 360, 375, 56 },
		{ 360, 376, 56 },
		{ 360, 377, 56 },
		{ 360, 378, 56 },
		{ 360, 379, 56 },
		{ 360, 380, 56 },
		{ 360, 381, 56 },
		{ 360, 382, 56 },
		{ 360, 383, 56 },
		{ 360, 384, 56 },
		{ 360, 385, 56 },
		{ 360, 386, 56 },
		{ 360, 387, 56 },
		{ 360, 388, 56 },
		{ 360, 389, 56 },
		{ 360, 390, 56 },
		{ 360, 391, 56 },
		{ 360, 392, 56 },
		{ 360, 393, 56 },
		{ 360, 394, 56 },
		{ 360, 395, 56 },
		{ 360, 396, 56 },
		{ 360, 397, 56 },
		{ 360, 398, 56 },
		{ 360, 399, 56 },
		{ 360, 400, 56 },
		{ 360, 401, 56 },
		{ 360, 402, 56 },
		{ 360, 403, 56 },
		{ 360, 404, 56 },
		{ 360, 405, 56 },
		{ 360, 406, 56 },
		{ 360, 407, 56 },
		{ 360, 408, 56 },
		{ 360, 409, 56 },
		{ 360, 410, 56 },
		{ 360, 411, 56 },
		{ 360, 412, 56 },
		{ 360, 413, 56 },
		{ 360, 414, 56 },
		{ 360, 415, 56 },
		{ 360, 416, 56 },
		{ 360, 417, 56 },
		{ 360, 418, 56 },
		{ 360, 419, 56 },
		{ 360, 420, 56 },
		{ 360, 421, 56 },
		{ 361, 376, 56 },
		{ 361, 377, 56 },
		{ 361, 378, 56 },
		{ 361, 379, 56 },
		{ 361, 380, 56 },
		{ 361, 381, 56 },
		{ 361, 382, 56 },
		{ 361, 383, 56 },
		{ 361, 384, 56 },
		{ 361, 385, 56 },
		{ 361, 386, 56 },
		{ 361, 387, 56 },
		{ 361, 388, 56 },
		{ 361, 389, 56 },
		{ 361, 390, 56 },
		{ 361, 391, 56 },
		{ 361, 392, 56 },
		{ 361, 393, 56 },
		{ 361, 394, 56 },
		{ 361, 395, 56 },
		{ 361, 396, 56 },
		{ 361, 397, 56 },
		{ 361, 398, 56 },
		{ 361, 399, 56 },
		{ 361, 400, 56 },
		{ 361, 401, 56 },
		{ 361, 402, 56 },
		{ 361, 403, 56 },
		{ 361, 404, 56 },
		{ 361, 405, 56 },
		{ 361, 406, 56 },
		{ 361, 407, 56 },
		{ 361, 408, 56 },
		{ 361, 409, 56 },
		{ 361, 410, 56 },
		{ 361, 411, 56 },
		{ 361, 412, 56 },
		{ 361, 413, 56 },
		{ 361, 414, 56 },
		{ 361, 415, 56 },
		{ 361, 416, 56 },
		{ 361, 417, 56 },
		{ 361, 418, 56 },
		{ 361, 419, 56 },
		{ 361, 420, 56 },
		{ 361, 421, 56 },
		{ 361, 422, 56 },
		{ 361, 423, 56 },
		{ 362, 378, 56 },
		{ 362, 379, 56 },
		{ 362, 380, 56 },
		{ 362, 381, 56 },
		{ 362, 382, 56 },
		{ 362, 383, 56 },
		{ 362, 384, 56 },
		{ 362, 385, 56 },
		{ 362, 386, 56 },
		{ 362, 387, 56 },
		{ 362, 388, 56 },
		{ 362, 389, 56 },
		{ 362, 390, 56 },
		{ 362, 391, 56 },
		{ 362, 392, 56 },
		{ 362, 393, 56 },
		{ 362, 394, 56 },
		{ 362, 395, 56 },
		{ 362, 396, 56 },
		{ 362, 397, 56 },
		{ 362, 398, 56 },
		{ 362, 399, 56 },
		{ 362, 400, 56 },
		{ 362, 401, 56 },
		{ 362, 402, 56 },
		{ 362, 403, 56 },
		{ 362, 404, 56 },
		{ 362, 405, 56 },
		{ 362, 406, 56 },
		{ 362, 407, 56 },
		{ 362, 408, 56 },
		{ 362, 409, 56 },
		{ 362, 410, 56 },
		{ 362, 411, 56 },
		{ 362, 412, 56 },
		{ 362, 413, 56 },
		{ 362, 414, 56 },
		{ 362, 415, 56 },
		{ 362, 416, 56 },
		{ 362, 417, 56 },
		{ 362, 418, 56 },
		{ 362, 419, 56 },
		{ 362, 420, 56 },
		{ 362, 421, 56 },
		{ 362, 422, 56 },
		{ 362, 423, 56 },
		{ 362, 424, 56 },
		{ 363, 379, 56 },
		{ 363, 380, 56 },
		{ 363, 381, 56 },
		{ 363, 382, 56 },
		{ 363, 383, 56 },
		{ 363, 384, 56 },
		{ 363, 385, 56 },
		{ 363, 386, 56 },
		{ 363, 387, 56 },
		{ 363, 388, 56 },
		{ 363, 389, 56 },
		{ 363, 390, 56 },
		{ 363, 391, 56 },
		{ 363, 392, 56 },
		{ 363, 393, 56 },
		{ 363, 394, 56 },
		{ 363, 395, 56 },
		{ 363, 396, 56 },
		{ 363, 397, 56 },
		{ 363, 398, 56 },
		{ 363, 399, 56 },
		{ 363, 400, 56 },
		{ 363, 401, 56 },
		{ 363, 402, 56 },
		{ 363, 403, 56 },
		{ 363, 404, 56 },
		{ 363, 405, 56 },
		{ 363, 406, 56 },
		{ 363, 407, 56 },
		{ 363, 408, 56 },
		{ 363, 409, 56 },
		{ 363, 410, 56 },
		{ 363, 411, 56 },
		{ 363, 412, 56 },
		{ 363, 413, 56 },
		{ 363, 414, 56 },
		{ 363, 415, 56 },
		{ 363, 416, 56 },
		{ 363, 417, 56 },
		{ 363, 418, 56 },
		{ 363, 419, 56 },
		{ 363, 420, 56 },
		{ 363, 421, 56 },
		{ 363, 422, 56 },
		{ 363, 423, 56 },
		{ 363, 424, 56 },
		{ 363, 425, 56 },
		{ 363, 426, 56 },
		{ 364, 380, 56 },
		{ 364, 381, 56 },
		{ 364, 382, 56 },
		{ 364, 383, 56 },
		{ 364, 384, 56 },
		{ 364, 385, 56 },
		{ 364, 386, 56 },
		{ 364, 387, 56 },
		{ 364, 388, 56 },
		{ 364, 389, 56 },
		{ 364, 390, 56 },
		{ 364, 391, 56 },
		{ 364, 392, 56 },
		{ 364, 393, 56 },
		{ 364, 394, 56 },
		{ 364, 395, 56 },
		{ 364, 396, 56 },
		{ 364, 397, 56 },
		{ 364, 398, 56 },
		{ 364, 399, 56 },
		{ 364, 400, 56 },
		{ 364, 401, 56 },
		{ 364, 402, 56 },
		{ 364, 403, 56 },
		{ 364, 404, 56 },
		{ 364, 405, 56 },
		{ 364, 406, 56 },
		{ 364, 407, 56 },
		{ 364, 408, 56 },
		{ 364, 409, 56 },
		{ 364, 410, 56 },
		{ 364, 411, 56 },
		{ 364, 412, 56 },
		{ 364, 413, 56 },
		{ 364, 414, 56 },
		{ 364, 415, 56 },
		{ 364, 416, 56 },
		{ 364, 417, 56 },
		{ 364, 418, 56 },
		{ 364, 419, 56 },
		{ 364, 420, 56 },
		{ 364, 421, 56 },
		{ 364, 422, 56 },
		{ 364, 423, 56 },
		{ 364, 424, 56 },
		{ 364, 425, 56 },
		{ 364, 426, 56 },
		{ 364, 427, 56 },
		{ 364, 428, 56 },
		{ 365, 381, 56 },
		{ 365, 382, 56 },
		{ 365, 383, 56 },
		{ 365, 384, 56 },
		{ 365, 385, 56 },
		{ 365, 386, 56 },
		{ 365, 387, 56 },
		{ 365, 388, 56 },
		{ 365, 389, 56 },
		{ 365, 390, 56 },
		{ 365, 391, 56 },
		{ 365, 392, 56 },
		{ 365, 393, 56 },
		{ 365, 394, 56 },
		{ 365, 395, 56 },
		{ 365, 396, 56 },
		{ 365, 397, 56 },
		{ 365, 398, 56 },
		{ 365, 399, 56 },
		{ 365, 400, 56 },
		{ 365, 401, 56 },
		{ 365, 402, 56 },
		{ 365, 403, 56 },
		{ 365, 404, 56 },
		{ 365, 405, 56 },
		{ 365, 406, 56 },
		{ 365, 407, 56 },
		{ 365, 408, 56 },
		{ 365, 409, 56 },
		{ 365, 410, 56 },
		{ 365, 411, 56 },
		{ 365, 412, 56 },
		{ 365, 413, 56 },
		{ 365, 414, 56 },
		{ 365, 415, 56 },
		{ 365, 416, 56 },
		{ 365, 417, 56 },
		{ 365, 418, 56 },
		{ 365, 419, 56 },
		{ 365, 420, 56 },
		{ 365, 421, 56 },
		{ 365, 422, 56 },
		{ 365, 423, 56 },
		{ 365, 424, 56 },
		{ 365, 425, 56 },
		{ 365, 426, 56 },
		{ 365, 427, 56 },
		{ 365, 428, 56 },
		{ 365, 429, 56 },
		{ 366, 382, 56 },
		{ 366, 383, 56 },
		{ 366, 384, 56 },
		{ 366, 385, 56 },
		{ 366, 386, 56 },
		{ 366, 387, 56 },
		{ 366, 388, 56 },
		{ 366, 389, 56 },
		{ 366, 390, 56 },
		{ 366, 391, 56 },
		{ 366, 392, 56 },
		{ 366, 393, 56 },
		{ 366, 394, 56 },
		{ 366, 395, 56 },
		{ 366, 396, 56 },
		{ 366, 397, 56 },
		{ 366, 398, 56 },
		{ 366, 399, 56 },
		{ 366, 400, 56 },
		{ 366, 401, 56 },
		{ 366, 402, 56 },
		{ 366, 403, 56 },
		{ 366, 404, 56 },
		{ 366, 405, 56 },
		{ 366, 406, 56 },
		{ 366, 407, 56 },
		{ 366, 408, 56 },
		{ 366, 409, 56 },
		{ 366, 410, 56 },
		{ 366, 411, 56 },
		{ 366, 412, 56 },
		{ 366, 413, 56 },
		{ 366, 414, 56 },
		{ 366, 415, 56 },
		{ 366, 416, 56 },
		{ 366, 417, 56 },
		{ 366, 418, 56 },
		{ 366, 419, 56 },
		{ 366, 420, 56 },
		{ 366, 421, 56 },
		{ 366, 422, 56 },
		{ 366, 423, 56 },
		{ 366, 424, 56 },
		{ 366, 425, 56 },
		{ 366, 426, 56 },
		{ 366, 427, 56 },
		{ 366, 428, 56 },
		{ 366, 429, 56 },
		{ 366, 430, 56 },
		{ 366, 431, 56 },
		{ 367, 383, 56 },
		{ 367, 384, 56 },
		{ 367, 385, 56 },
		{ 367, 386, 56 },
		{ 367, 387, 56 },
		{ 367, 388, 56 },
		{ 367, 389, 56 },
		{ 367, 390, 56 },
		{ 367, 391, 56 },
		{ 367, 392, 56 },
		{ 367, 393, 56 },
		{ 367, 394, 56 },
		{ 367, 395, 56 },
		{ 367, 396, 56 },
		{ 367, 397, 56 },
		{ 367, 398, 56 },
		{ 367, 399, 56 },
		{ 367, 400, 56 },
		{ 367, 401, 56 },
		{ 367, 402, 56 },
		{ 367, 403, 56 },
		{ 367, 404, 56 },
		{ 367, 405, 56 },
		{ 367, 406, 56 },
		{ 367, 407, 56 },
		{ 367, 408, 56 },
		{ 367, 409, 56 },
		{ 367, 410, 56 },
		{ 367, 411, 56 },
		{ 367, 412, 56 },
		{ 367, 413, 56 },
		{ 367, 414, 56 },
		{ 367, 415, 56 },
		{ 367, 416, 56 },
		{ 367, 417, 56 },
		{ 367, 418, 56 },
		{ 367, 419, 56 },
		{ 367, 420, 56 },
		{ 367, 421, 56 },
		{ 367, 422, 56 },
		{ 367, 423, 56 },
		{ 367, 424, 56 },
		{ 367, 425, 56 },
		{ 367, 426, 56 },
		{ 367, 427, 56 },
		{ 367, 428, 56 },
		{ 367, 429, 56 },
		{ 367, 430, 56 },
		{ 367, 431, 56 },
		{ 367, 432, 56 },
		{ 368, 384, 56 },
		{ 368, 385, 56 },
		{ 368, 386, 56 },
		{ 368, 387, 56 },
		{ 368, 388, 56 },
		{ 368, 389, 56 },
		{ 368, 390, 56 },
		{ 368, 391, 56 },
		{ 368, 392, 56 },
		{ 368, 393, 56 },
		{ 368, 394, 56 },
		{ 368, 395, 56 },
		{ 368, 396, 56 },
		{ 368, 397, 56 },
		{ 368, 398, 56 },
		{ 368, 399, 56 },
		{ 368, 400, 56 },
		{ 368, 401, 56 },
		{ 368, 402, 56 },
		{ 368, 403, 56 },
		{ 368, 404, 56 },
		{ 368, 405, 56 },
		{ 368, 406, 56 },
		{ 368, 407, 56 },
		{ 368, 408, 56 },
		{ 368, 409, 56 },
		{ 368, 410, 56 },
		{ 368, 411, 56 },
		{ 368, 412, 56 },
		{ 368, 413, 56 },
		{ 368, 414, 56 },
		{ 368, 415, 56 },
		{ 368, 416, 56 },
		{ 368, 417, 56 },
		{ 368, 418, 56 },
		{ 368, 419, 56 },
		{ 368, 420, 56 },
		{ 368, 421, 56 },
		{ 368, 422, 56 },
		{ 368, 423, 56 },
		{ 368, 424, 56 },
		{ 368, 425, 56 },
		{ 368, 426, 56 },
		{ 368, 427, 56 },
		{ 368, 428, 56 },
		{ 368, 429, 56 },
		{ 368, 430, 56 },
		{ 368, 431, 56 },
		{ 368, 432, 56 },
		{ 368, 433, 56 },
		{ 368, 434, 56 },
		{ 369, 386, 56 },
		{ 369, 387, 56 },
		{ 369, 388, 56 },
		{ 369, 389, 56 },
		{ 369, 390, 56 },
		{ 369, 391, 56 },
		{ 369, 392, 56 },
		{ 369, 393, 56 },
		{ 369, 394, 56 },
		{ 369, 395, 56 },
		{ 369, 396, 56 },
		{ 369, 397, 56 },
		{ 369, 398, 56 },
		{ 369, 399, 56 },
		{ 369, 400, 56 },
		{ 369, 401, 56 },
		{ 369, 402, 56 },
		{ 369, 403, 56 },
		{ 369, 404, 56 },
		{ 369, 405, 56 },
		{ 369, 406, 56 },
		{ 369, 407, 56 },
		{ 369, 408, 56 },
		{ 369, 409, 56 },
		{ 369, 410, 56 },
		{ 369, 411, 56 },
		{ 369, 412, 56 },
		{ 369, 413, 56 },
		{ 369, 414, 56 },
		{ 369, 415, 56 },
		{ 369, 416, 56 },
		{ 369, 417, 56 },
		{ 369, 418, 56 },
		{ 369, 419, 56 },
		{ 369, 420, 56 },
		{ 369, 421, 56 },
		{ 369, 422, 56 },
		{ 369, 423, 56 },
		{ 369, 424, 56 },
		{ 369, 425, 56 },
		{ 369, 426, 56 },
		{ 369, 427, 56 },
		{ 369, 428, 56 },
		{ 369, 429, 56 },
		{ 369, 430, 56 },
		{ 369, 431, 56 },
		{ 369, 432, 56 },
		{ 369, 433, 56 },
		{ 369, 434, 56 },
		{ 369, 435, 56 },
		{ 369, 436, 56 },
		{ 370, 387, 56 },
		{ 370, 388, 56 },
		{ 370, 389, 56 },
		{ 370, 390, 56 },
		{ 370, 391, 56 },
		{ 370, 392, 56 },
		{ 370, 393, 56 },
		{ 370, 394, 56 },
		{ 370, 395, 56 },
		{ 370, 396, 56 },
		{ 370, 397, 56 },
		{ 370, 398, 56 },
		{ 370, 399, 56 },
		{ 370, 400, 56 },
		{ 370, 401, 56 },
		{ 370, 402, 56 },
		{ 370, 403, 56 },
		{ 370, 404, 56 },
		{ 370, 405, 56 },
		{ 370, 406, 56 },
		{ 370, 407, 56 },
		{ 370, 408, 56 },
		{ 370, 409, 56 },
		{ 370, 410, 56 },
		{ 370, 411, 56 },
		{ 370, 412, 56 },
		{ 370, 413, 56 },
		{ 370, 414, 56 },
		{ 370, 415, 56 },
		{ 370, 416, 56 },
		{ 370, 417, 56 },
		{ 370, 418, 56 },
		{ 370, 419, 56 },
		{ 370, 420, 56 },
		{ 370, 421, 56 },
		{ 370, 422, 56 },
		{ 370, 423, 56 },
		{ 370, 424, 56 },
		{ 370, 425, 56 },
		{ 370, 426, 56 },
		{ 370, 427, 56 },
		{ 370, 428, 56 },
		{ 370, 429, 56 },
		{ 370, 430, 56 },
		{ 370, 431, 56 },
		{ 370, 432, 56 },
		{ 370, 433, 56 },
		{ 370, 434, 56 },
		{ 370, 435, 56 },
		{ 370, 436, 56 },
		{ 370, 437, 56 },
		{ 371, 388, 56 },
		{ 371, 389, 56 },
		{ 371, 390, 56 },
		{ 371, 391, 56 },
		{ 371, 392, 56 },
		{ 371, 393, 56 },
		{ 371, 394, 56 },
		{ 371, 395, 56 },
		{ 371, 396, 56 },
		{ 371, 397, 56 },
		{ 371, 398, 56 },
		{ 371, 399, 56 },
		{ 371, 400, 56 },
		{ 371, 401, 56 },
		{ 371, 402, 56 },
		{ 371, 403, 56 },
		{ 371, 404, 56 },
		{ 371, 405, 56 },
		{ 371, 406, 56 },
		{ 371, 407, 56 },
		{ 371, 408, 56 },
		{ 371, 409, 56 },
		{ 371, 410, 56 },
		{ 371, 411, 56 },
		{ 371, 412, 56 },
		{ 371, 413, 56 },
		{ 371, 414, 56 },
		{ 371, 415, 56 },
		{ 371, 416, 56 },
		{ 371, 417, 56 },
		{ 371, 418, 56 },
		{ 371, 419, 56 },
		{ 371, 420, 56 },
		{ 371, 421, 56 },
		{ 371, 422, 56 },
		{ 371, 423, 56 },
		{ 371, 424, 56 },
		{ 371, 425, 56 },
		{ 371, 426, 56 },
		{ 371, 427, 56 },
		{ 371, 428, 56 },
		{ 371, 429, 56 },
		{ 371, 430, 56 },
		{ 371, 431, 56 },
		{ 371, 432, 56 },
		{ 371, 433, 56 },
		{ 371, 434, 56 },
		{ 371, 435, 56 },
		{ 371, 436, 56 },
		{ 371, 437, 56 },
		{ 371, 438, 56 },
		{ 371, 439, 56 },
		{ 372, 389, 56 },
		{ 372, 390, 56 },
		{ 372, 391, 56 },
		{ 372, 392, 56 },
		{ 372, 393, 56 },
		{ 372, 394, 56 },
		{ 372, 395, 56 },
		{ 372, 396, 56 },
		{ 372, 397, 56 },
		{ 372, 398, 56 },
		{ 372, 399, 56 },
		{ 372, 400, 56 },
		{ 372, 401, 56 },
		{ 372, 402, 56 },
		{ 372, 403, 56 },
		{ 372, 404, 56 },
		{ 372, 405, 56 },
		{ 372, 406, 56 },
		{ 372, 407, 56 },
		{ 372, 408, 56 },
		{ 372, 409, 56 },
		{ 372, 410, 56 },
		{ 372, 411, 56 },
		{ 372, 412, 56 },
		{ 372, 413, 56 },
		{ 372, 414, 56 },
		{ 372, 415, 56 },
		{ 372, 416, 56 },
		{ 372, 417, 56 },
		{ 372, 418, 56 },
		{ 372, 419, 56 },
		{ 372, 420, 56 },
		{ 372, 421, 56 },
		{ 372, 422, 56 },
		{ 372, 423, 56 },
		{ 372, 424, 56 },
		{ 372, 425, 56 },
		{ 372, 426, 56 },
		{ 372, 427, 56 },
		{ 372, 428, 56 },
		{ 372, 429, 56 },
		{ 372, 430, 56 },
		{ 372, 431, 56 },
		{ 372, 432, 56 },
		{ 372, 433, 56 },
		{ 372, 434, 56 },
		{ 372, 435, 56 },
		{ 372, 436, 56 },
		{ 372, 437, 56 },
		{ 372, 438, 56 },
		{ 372, 439, 56 },
		{ 372, 440, 56 },
		{ 373, 390, 56 },
		{ 373, 391, 56 },
		{ 373, 392, 56 },
		{ 373, 393, 56 },
		{ 373, 394, 56 },
		{ 373, 395, 56 },
		{ 373, 396, 56 },
		{ 373, 397, 56 },
		{ 373, 398, 56 },
		{ 373, 399, 56 },
		{ 373, 400, 56 },
		{ 373, 401, 56 },
		{ 373, 402, 56 },
		{ 373, 403, 56 },
		{ 373, 404, 56 },
		{ 373, 405, 56 },
		{ 373, 406, 56 },
		{ 373, 407, 56 },
		{ 373, 408, 56 },
		{ 373, 409, 56 },
		{ 373, 410, 56 },
		{ 373, 411, 56 },
		{ 373, 412, 56 },
		{ 373, 413, 56 },
		{ 373, 414, 56 },
		{ 373, 415, 56 },
		{ 373, 416, 56 },
		{ 373, 417, 56 },
		{ 373, 418, 56 },
		{ 373, 419, 56 },
		{ 373, 420, 56 },
		{ 373, 421, 56 },
		{ 373, 422, 56 },
		{ 373, 423, 56 },
		{ 373, 424, 56 },
		{ 373, 425, 56 },
		{ 373, 426, 56 },
		{ 373, 427, 56 },
		{ 373, 428, 56 },
		{ 373, 429, 56 },
		{ 373, 430, 56 },
		{ 373, 431, 56 },
		{ 373, 432, 56 },
		{ 373, 433, 56 },
		{ 373, 434, 56 },
		{ 373, 435, 56 },
		{ 373, 436, 56 },
		{ 373, 437, 56 },
		{ 373, 438, 56 },
		{ 373, 439, 56 },
		{ 373, 440, 56 },
		{ 373, 441, 56 },
		{ 373, 442, 56 },
		{ 374, 391, 56 },
		{ 374, 392, 56 },
		{ 374, 393, 56 },
		{ 374, 394, 56 },
		{ 374, 395, 56 },
		{ 374, 396, 56 },
		{ 374, 397, 56 },
		{ 374, 398, 56 },
		{ 374, 399, 56 },
		{ 374, 400, 56 },
		{ 374, 401, 56 },
		{ 374, 402, 56 },
		{ 374, 403, 56 },
		{ 374, 404, 56 },
		{ 374, 405, 56 },
		{ 374, 406, 56 },
		{ 374, 407, 56 },
		{ 374, 408, 56 },
		{ 374, 409, 56 },
		{ 374, 410, 56 },
		{ 374, 411, 56 },
		{ 374, 412, 56 },
		{ 374, 413, 56 },
		{ 374, 414, 56 },
		{ 374, 415, 56 },
		{ 374, 416, 56 },
		{ 374, 417, 56 },
		{ 374, 418, 56 },
		{ 374, 419, 56 },
		{ 374, 420, 56 },
		{ 374, 421, 56 },
		{ 374, 422, 56 },
		{ 374, 423, 56 },
		{ 374, 424, 56 },
		{ 374, 425, 56 },
		{ 374, 426, 56 },
		{ 374, 427, 56 },
		{ 374, 428, 56 },
		{ 374, 429, 56 },
		{ 374, 430, 56 },
		{ 374, 431, 56 },
		{ 374, 432, 56 },
		{ 374, 433, 56 },
		{ 374, 434, 56 },
		{ 374, 435, 56 },
		{ 374, 436, 56 },
		{ 374, 437, 56 },
		{ 374, 438, 56 },
		{ 374, 439, 56 },
		{ 374, 440, 56 },
		{ 374, 441, 56 },
		{ 374, 442, 56 },
		{ 374, 443, 56 },
		{ 374, 444, 56 },
		{ 375, 392, 56 },
		{ 375, 393, 56 },
		{ 375, 394, 56 },
		{ 375, 395, 56 },
		{ 375, 396, 56 },
		{ 375, 397, 56 },
		{ 375, 398, 56 },
		{ 375, 399, 56 },
		{ 375, 400, 56 },
		{ 375, 401, 56 },
		{ 375, 402, 56 },
		{ 375, 403, 56 },
		{ 375, 404, 56 },
		{ 375, 405, 56 },
		{ 375, 406, 56 },
		{ 375, 407, 56 },
		{ 375, 408, 56 },
		{ 375, 409, 56 },
		{ 375, 410, 56 },
		{ 375, 411, 56 },
		{ 375, 412, 56 },
		{ 375, 413, 56 },
		{ 375, 414, 56 },
		{ 375, 415, 56 },
		{ 375, 416, 56 },
		{ 375, 417, 56 },
		{ 375, 418, 56 },
		{ 375, 419, 56 },
		{ 375, 420, 56 },
		{ 375, 421, 56 },
		{ 375, 422, 56 },
		{ 375, 423, 56 },
		{ 375, 424, 56 },
		{ 375, 425, 56 },
		{ 375, 426, 56 },
		{ 375, 427, 56 },
		{ 375, 428, 56 },
		{ 375, 429, 56 },
		{ 375, 430, 56 },
		{ 375, 431, 56 },
		{ 375, 432, 56 },
		{ 375, 433, 56 },
		{ 375, 434, 56 },
		{ 375, 435, 56 },
		{ 375, 436, 56 },
		{ 375, 437, 56 },
		{ 375, 438, 56 },
		{ 375, 439, 56 },
		{ 375, 440, 56 },
		{ 375, 441, 56 },
		{ 375, 442, 56 },
		{ 375, 443, 56 },
		{ 375, 444, 56 },
		{ 375, 445, 56 },
		{ 376, 394, 56 },
		{ 376, 395, 56 },
		{ 376, 396, 56 },
		{ 376, 397, 56 },
		{ 376, 398, 56 },
		{ 376, 399, 56 },
		{ 376, 400, 56 },
		{ 376, 401, 56 },
		{ 376, 402, 56 },
		{ 376, 403, 56 },
		{ 376, 404, 56 },
		{ 376, 405, 56 },
		{ 376, 406, 56 },
		{ 376, 407, 56 },
		{ 376, 408, 56 },
		{ 376, 409, 56 },
		{ 376, 410, 56 },
		{ 376, 411, 56 },
		{ 376, 412, 56 },
		{ 376, 413, 56 },
		{ 376, 414, 56 },
		{ 376, 415, 56 },
		{ 376, 416, 56 },
		{ 376, 417, 56 },
		{ 376, 418, 56 },
		{ 376, 419, 56 },
		{ 376, 420, 56 },
		{ 376, 421, 56 },
		{ 376, 422, 56 },
		{ 376, 423, 56 },
		{ 376, 424, 56 },
		{ 376, 425, 56 },
		{ 376, 426, 56 },
		{ 376, 427, 56 },
		{ 376, 428, 56 },
		{ 376, 429, 56 },
		{ 376, 430, 56 },
		{ 376, 431, 56 },
		{ 376, 432, 56 },
		{ 376, 433, 56 },
		{ 376, 434, 56 },
		{ 376, 435, 56 },
		{ 376, 436, 56 },
		{ 376, 437, 56 },
		{ 376, 438, 56 },
		{ 376, 439, 56 },
		{ 376, 440, 56 },
		{ 376, 441, 56 },
		{ 376, 442, 56 },
		{ 376, 443, 56 },
		{ 376, 444, 56 },
		{ 376, 445, 56 },
		{ 376, 446, 56 },
		{ 376, 447, 56 },
		{ 377, 395, 56 },
		{ 377, 396, 56 },
		{ 377, 397, 56 },
		{ 377, 398, 56 },
		{ 377, 399, 56 },
		{ 377, 400, 56 },
		{ 377, 401, 56 },
		{ 377, 402, 56 },
		{ 377, 403, 56 },
		{ 377, 404, 56 },
		{ 377, 405, 56 },
		{ 377, 406, 56 },
		{ 377, 407, 56 },
		{ 377, 408, 56 },
		{ 377, 409, 56 },
		{ 377, 410, 56 },
		{ 377, 411, 56 },
		{ 377, 412, 56 },
		{ 377, 413, 56 },
		{ 377, 414, 56 },
		{ 377, 415, 56 },
		{ 377, 416, 56 },
		{ 377, 417, 56 },
		{ 377, 418, 56 },
		{ 377, 419, 56 },
		{ 377, 420, 56 },
		{ 377, 421, 56 },
		{ 377, 422, 56 },
		{ 377, 423, 56 },
		{ 377, 424, 56 },
		{ 377, 425, 56 },
		{ 377, 426, 56 },
		{ 377, 427, 56 },
		{ 377, 428, 56 },
		{ 377, 429, 56 },
		{ 377, 430, 56 },
		{ 377, 431, 56 },
		{ 377, 432, 56 },
		{ 377, 433, 56 },
		{ 377, 434, 56 },
		{ 377, 435, 56 },
		{ 377, 436, 56 },
		{ 377, 437, 56 },
		{ 377, 438, 56 },
		{ 377, 439, 56 },
		{ 377, 440, 56 },
		{ 377, 441, 56 },
		{ 377, 442, 56 },
		{ 377, 443, 56 },
		{ 377, 444, 56 },
		{ 377, 445, 56 },
		{ 377, 446, 56 },
		{ 377, 447, 56 },
		{ 377, 448, 56 },
		{ 378, 396, 56 },
		{ 378, 397, 56 },
		{ 378, 398, 56 },
		{ 378, 399, 56 },
		{ 378, 400, 56 },
		{ 378, 401, 56 },
		{ 378, 402, 56 },
		{ 378, 403, 56 },
		{ 378, 404, 56 },
		{ 378, 405, 56 },
		{ 378, 406, 56 },
		{ 378, 407, 56 },
		{ 378, 408, 56 },
		{ 378, 409, 56 },
		{ 378, 410, 56 },
		{ 378, 411, 56 },
		{ 378, 412, 56 },
		{ 378, 413, 56 },
		{ 378, 414, 56 },
		{ 378, 415, 56 },
		{ 378, 416, 56 },
		{ 378, 417, 56 },
		{ 378, 418, 56 },
		{ 378, 419, 56 },
		{ 378, 420, 56 },
		{ 378, 421, 56 },
		{ 378, 422, 56 },
		{ 378, 423, 56 },
		{ 378, 424, 56 },
		{ 378, 425, 56 },
		{ 378, 426, 56 },
		{ 378, 427, 56 },
		{ 378, 428, 56 },
		{ 378, 429, 56 },
		{ 378, 430, 56 },
		{ 378, 431, 56 },
		{ 378, 432, 56 },
		{ 378, 433, 56 },
		{ 378, 434, 56 },
		{ 378, 435, 56 },
		{ 378, 436, 56 },
		{ 378, 437, 56 },
		{ 378, 438, 56 },
		{ 378, 439, 56 },
		{ 378, 440, 56 },
		{ 378, 441, 56 },
		{ 378, 442, 56 },
		{ 378, 443, 56 },
		{ 378, 444, 56 },
		{ 378, 445, 56 },
		{ 378, 446, 56 },
		{ 378, 447, 56 },
		{ 378, 448, 56 },
		{ 378, 449, 56 },
		{ 378, 450, 56 },
		{ 379, 397, 56 },
		{ 379, 398, 56 },
		{ 379, 399, 56 },
		{ 379, 400, 56 },
		{ 379, 401, 56 },
		{ 379, 402, 56 },
		{ 379, 403, 56 },
		{ 379, 404, 56 },
		{ 379, 405, 56 },
		{ 379, 406, 56 },
		{ 379, 407, 56 },
		{ 379, 408, 56 },
		{ 379, 409, 56 },
		{ 379, 410, 56 },
		{ 379, 411, 56 },
		{ 379, 412, 56 },
		{ 379, 413, 56 },
		{ 379, 414, 56 },
		{ 379, 415, 56 },
		{ 379, 416, 56 },
		{ 379, 417, 56 },
		{ 379, 418, 56 },
		{ 379, 419, 56 },
		{ 379, 420, 56 },
		{ 379, 421, 56 },
		{ 379, 422, 56 },
		{ 379, 423, 56 },
		{ 379, 424, 56 },
		{ 379, 425, 56 },
		{ 379, 426, 56 },
		{ 379, 427, 56 },
		{ 379, 428, 56 },
		{ 379, 429, 56 },
		{ 379, 430, 56 },
		{ 379, 431, 56 },
		{ 379, 432, 56 },
		{ 379, 433, 56 },
		{ 379, 434, 56 },
		{ 379, 435, 56 },
		{ 379, 436, 56 },
		{ 379, 437, 56 },
		{ 379, 438, 56 },
		{ 379, 439, 56 },
		{ 379, 440, 56 },
		{ 379, 441, 56 },
		{ 379, 442, 56 },
		{ 379, 443, 56 },
		{ 379, 444, 56 },
		{ 379, 445, 56 },
		{ 379, 446, 56 },
		{ 379, 447, 56 },
		{ 379, 448, 56 },
		{ 379, 449, 56 },
		{ 379, 450, 56 },
		{ 379, 451, 56 },
		{ 379, 452, 56 },
		{ 380, 398, 56 },
		{ 380, 399, 56 },
		{ 380, 400, 56 },
		{ 380, 401, 56 },
		{ 380, 402, 56 },
		{ 380, 403, 56 },
		{ 380, 404, 56 },
		{ 380, 405, 56 },
		{ 380, 406, 56 },
		{ 380, 407, 56 },
		{ 380, 408, 56 },
		{ 380, 409, 56 },
		{ 380, 410, 56 },
		{ 380, 411, 56 },
		{ 380, 412, 56 },
		{ 380, 413, 56 },
		{ 380, 414, 56 },
		{ 380, 415, 56 },
		{ 380, 416, 56 },
		{ 380, 417, 56 },
		{ 380, 418, 56 },
		{ 380, 419, 56 },
		{ 380, 420, 56 },
		{ 380, 421, 56 },
		{ 380, 422, 56 },
		{ 380, 423, 56 },
		{ 380, 424, 56 },
		{ 380, 425, 56 },
		{ 380, 426, 56 },
		{ 380, 427, 56 },
		{ 380, 428, 56 },
		{ 380, 429, 56 },
		{ 380, 430, 56 },
		{ 380, 431, 56 },
		{ 380, 432, 56 },
		{ 380, 433, 56 },
		{ 380, 434, 56 },
		{ 380, 435, 56 },
		{ 380, 436, 56 },
		{ 380, 437, 56 },
		{ 380, 438, 56 },
		{ 380, 439, 56 },
		{ 380, 440, 56 },
		{ 380, 441, 56 },
		{ 380, 442, 56 },
		{ 380, 443, 56 },
		{ 380, 444, 56 },
		{ 380, 445, 56 },
		{ 380, 446, 56 },
		{ 380, 447, 56 },
		{ 380, 448, 56 },
		{ 380, 449, 56 },
		{ 380, 450, 56 },
		{ 380, 451, 56 },
		{ 380, 452, 56 },
		{ 380, 453, 56 },
		{ 381, 399, 56 },
		{ 381, 400, 56 },
		{ 381, 401, 56 },
		{ 381, 402, 56 },
		{ 381, 403, 56 },
		{ 381, 404, 56 },
		{ 381, 405, 56 },
		{ 381, 406, 56 },
		{ 381, 407, 56 },
		{ 381, 408, 56 },
		{ 381, 409, 56 },
		{ 381, 410, 56 },
		{ 381, 411, 56 },
		{ 381, 412, 56 },
		{ 381, 413, 56 },
		{ 381, 414, 56 },
		{ 381, 415, 56 },
		{ 381, 416, 56 },
		{ 381, 417, 56 },
		{ 381, 418, 56 },
		{ 381, 419, 56 },
		{ 381, 420, 56 },
		{ 381, 421, 56 },
		{ 381, 422, 56 },
		{ 381, 423, 56 },
		{ 381, 424, 56 },
		{ 381, 425, 56 },
		{ 381, 426, 56 },
		{ 381, 427, 56 },
		{ 381, 428, 56 },
		{ 381, 429, 56 },
		{ 381, 430, 56 },
		{ 381, 431, 56 },
		{ 381, 432, 56 },
		{ 381, 433, 56 },
		{ 381, 434, 56 },
		{ 381, 435, 56 },
		{ 381, 436, 56 },
		{ 381, 437, 56 },
		{ 381, 438, 56 },
		{ 381, 439, 56 },
		{ 381, 440, 56 },
		{ 381, 441, 56 },
		{ 381, 442, 56 },
		{ 381, 443, 56 },
		{ 381, 444, 56 },
		{ 381, 445, 56 },
		{ 381, 446, 56 },
		{ 381, 447, 56 },
		{ 381, 448, 56 },
		{ 381, 449, 56 },
		{ 381, 450, 56 },
		{ 381, 451, 56 },
		{ 381, 452, 56 },
		{ 381, 453, 56 },
		{ 381, 454, 56 },
		{ 381, 455, 56 },
		{ 382, 400, 56 },
		{ 382, 401, 56 },
		{ 382, 402, 56 },
		{ 382, 403, 56 },
		{ 382, 404, 56 },
		{ 382, 405, 56 },
		{ 382, 406, 56 },
		{ 382, 407, 56 },
		{ 382, 408, 56 },
		{ 382, 409, 56 },
		{ 382, 410, 56 },
		{ 382, 411, 56 },
		{ 382, 412, 56 },
		{ 382, 413, 56 },
		{ 382, 414, 56 },
		{ 382, 415, 56 },
		{ 382, 416, 56 },
		{ 382, 417, 56 },
		{ 382, 418, 56 },
		{ 382, 419, 56 },
		{ 382, 420, 56 },
		{ 382, 421, 56 },
		{ 382, 422, 56 },
		{ 382, 423, 56 },
		{ 382, 424, 56 },
		{ 382, 425, 56 },
		{ 382, 426, 56 },
		{ 382, 427, 56 },
		{ 382, 428, 56 },
		{ 382, 429, 56 },
		{ 382, 430, 56 },
		{ 382, 431, 56 },
		{ 382, 432, 56 },
		{ 382, 433, 56 },
		{ 382, 434, 56 },
		{ 382, 435, 56 },
		{ 382, 436, 56 },
		{ 382, 437, 56 },
		{ 382, 438, 56 },
		{ 382, 439, 56 },
		{ 382, 440, 56 },
		{ 382, 441, 56 },
		{ 382, 442, 56 },
		{ 382, 443, 56 },
		{ 382, 444, 56 },
		{ 382, 445, 56 },
		{ 382, 446, 56 },
		{ 382, 447, 56 },
		{ 382, 448, 56 },
		{ 382, 449, 56 },
		{ 382, 450, 56 },
		{ 382, 451, 56 },
		{ 382, 452, 56 },
		{ 382, 453, 56 },
		{ 382, 454, 56 },
		{ 382, 455, 56 },
		{ 382, 456, 56 },
		{ 383, 402, 56 },
		{ 383, 403, 56 },
		{ 383, 404, 56 },
		{ 383, 405, 56 },
		{ 383, 406, 56 },
		{ 383, 407, 56 },
		{ 383, 408, 56 },
		{ 383, 409, 56 },
		{ 383, 410, 56 },
		{ 383, 411, 56 },
		{ 383, 412, 56 },
		{ 383, 413, 56 },
		{ 383, 414, 56 },
		{ 383, 415, 56 },
		{ 383, 416, 56 },
		{ 383, 417, 56 },
		{ 383, 418, 56 },
		{ 383, 419, 56 },
		{ 383, 420, 56 },
		{ 383, 421, 56 },
		{ 383, 422, 56 },
		{ 383, 423, 56 },
		{ 383, 424, 56 },
		{ 383, 425, 56 },
		{ 383, 426, 56 },
		{ 383, 427, 56 },
		{ 383, 428, 56 },
		{ 383, 429, 56 },
		{ 383, 430, 56 },
		{ 383, 431, 56 },
		{ 383, 432, 56 },
		{ 383, 433, 56 },
		{ 383, 434, 56 },
		{ 383, 435, 56 },
		{ 383, 436, 56 },
		{ 383, 437, 56 },
		{ 383, 438, 56 },
		{ 383, 439, 56 },
		{ 383, 440, 56 },
		{ 383, 441, 56 },
		{ 383, 442, 56 },
		{ 383, 443, 56 },
		{ 383, 444, 56 },
		{ 383, 445, 56 },
		{ 383, 446, 56 },
		{ 383, 447, 56 },
		{ 383, 448, 56 },
		{ 383, 449, 56 },
		{ 383, 450, 56 },
		{ 383, 451, 56 },
		{ 383, 452, 56 },
		{ 383, 453, 56 },
		{ 383, 454, 56 },
		{ 383, 455, 56 },
		{ 383, 456, 56 },
		{ 383, 457, 56 },
		{ 383, 458, 56 },
		{ 384, 403, 56 },
		{ 384, 404, 56 },
		{ 384, 405, 56 },
		{ 384, 406, 56 },
		{ 384, 407, 56 },
		{ 384, 408, 56 },
		{ 384, 409, 56 },
		{ 384, 410, 56 },
		{ 384, 411, 56 },
		{ 384, 412, 56 },
		{ 384, 413, 56 },
		{ 384, 414, 56 },
		{ 384, 415, 56 },
		{ 384, 416, 56 },
		{ 384, 417, 56 },
		{ 384, 418, 56 },
		{ 384, 419, 56 },
		{ 384, 420, 56 },
		{ 384, 421, 56 },
		{ 384, 422, 56 },
		{ 384, 423, 56 },
		{ 384, 424, 56 },
		{ 384, 425, 56 },
		{ 384, 426, 56 },
		{ 384, 427, 56 },
		{ 384, 428, 56 },
		{ 384, 429, 56 },
		{ 384, 430, 56 },
		{ 384, 431, 56 },
		{ 384, 432, 56 },
		{ 384, 433, 56 },
		{ 384, 434, 56 },
		{ 384, 435, 56 },
		{ 384, 436, 56 },
		{ 384, 437, 56 },
		{ 384, 438, 56 },
		{ 384, 439, 56 },
		{ 384, 440, 56 },
		{ 384, 441, 56 },
		{ 384, 442, 56 },
		{ 384, 443, 56 },
		{ 384, 444, 56 },
		{ 384, 445, 56 },
		{ 384, 446, 56 },
		{ 384, 447, 56 },
		{ 384, 448, 56 },
		{ 384, 449, 56 },
		{ 384, 450, 56 },
		{ 384, 451, 56 },
		{ 384, 452, 56 },
		{ 384, 453, 56 },
		{ 384, 454, 56 },
		{ 384, 455, 56 },
		{ 384, 456, 56 },
		{ 384, 457, 56 },
		{ 384, 458, 56 },
		{ 384, 459, 56 },
		{ 384, 460, 56 },
		{ 385, 404, 56 },
		{ 385, 405, 56 },
		{ 385, 406, 56 },
		{ 385, 407, 56 },
		{ 385, 408, 56 },
		{ 385, 409, 56 },
		{ 385, 410, 56 },
		{ 385, 411, 56 },
		{ 385, 412, 56 },
		{ 385, 413, 56 },
		{ 385, 414, 56 },
		{ 385, 415, 56 },
		{ 385, 416, 56 },
		{ 385, 417, 56 },
		{ 385, 418, 56 },
		{ 385, 419, 56 },
		{ 385, 420, 56 },
		{ 385, 421, 56 },
		{ 385, 422, 56 },
		{ 385, 423, 56 },
		{ 385, 424, 56 },
		{ 385, 425, 56 },
		{ 385, 426, 56 },
		{ 385, 427, 56 },
		{ 385, 428, 56 },
		{ 385, 429, 56 },
		{ 385, 430, 56 },
		{ 385, 431, 56 },
		{ 385, 432, 56 },
		{ 385, 433, 56 },
		{ 385, 434, 56 },
		{ 385, 435, 56 },
		{ 385, 436, 56 },
		{ 385, 437, 56 },
		{ 385, 438, 56 },
		{ 385, 439, 56 },
		{ 385, 440, 56 },
		{ 385, 441, 56 },
		{ 385, 442, 56 },
		{ 385, 443, 56 },
		{ 385, 444, 56 },
		{ 385, 445, 56 },
		{ 385, 446, 56 },
		{ 385, 447, 56 },
		{ 385, 448, 56 },
		{ 385, 449, 56 },
		{ 385, 450, 56 },
		{ 385, 451, 56 },
		{ 385, 452, 56 },
		{ 385, 453, 56 },
		{ 385, 454, 56 },
		{ 385, 455, 56 },
		{ 385, 456, 56 },
		{ 385, 457, 56 },
		{ 385, 458, 56 },
		{ 385, 459, 56 },
		{ 385, 460, 56 },
		{ 385, 461, 56 },
		{ 386, 405, 56 },
		{ 386, 406, 56 },
		{ 386, 407, 56 },
		{ 386, 408, 56 },
		{ 386, 409, 56 },
		{ 386, 410, 56 },
		{ 386, 411, 56 },
		{ 386, 412, 56 },
		{ 386, 413, 56 },
		{ 386, 414, 56 },
		{ 386, 415, 56 },
		{ 386, 416, 56 },
		{ 386, 417, 56 },
		{ 386, 418, 56 },
		{ 386, 419, 56 },
		{ 386, 420, 56 },
		{ 386, 421, 56 },
		{ 386, 422, 56 },
		{ 386, 423, 56 },
		{ 386, 424, 56 },
		{ 386, 425, 56 },
		{ 386, 426, 56 },
		{ 386, 427, 56 },
		{ 386, 428, 56 },
		{ 386, 429, 56 },
		{ 386, 430, 56 },
		{ 386, 431, 56 },
		{ 386, 432, 56 },
		{ 386, 433, 56 },
		{ 386, 434, 56 },
		{ 386, 435, 56 },
		{ 386, 436, 56 },
		{ 386, 437, 56 },
		{ 386, 438, 56 },
		{ 386, 439, 56 },
		{ 386, 440, 56 },
		{ 386, 441, 56 },
		{ 386, 442, 56 },
		{ 386, 443, 56 },
		{ 386, 444, 56 },
		{ 386, 445, 56 },
		{ 386, 446, 56 },
		{ 386, 447, 56 },
		{ 386, 448, 56 },
		{ 386, 449, 56 },
		{ 386, 450, 56 },
		{ 386, 451, 56 },
		{ 386, 452, 56 },
		{ 386, 453, 56 },
		{ 386, 454, 56 },
		{ 386, 455, 56 },
		{ 386, 456, 56 },
		{ 386, 457, 56 },
		{ 386, 458, 56 },
		{ 386, 459, 56 },
		{ 386, 460, 56 },
		{ 386, 461, 56 },
		{ 386, 462, 56 },
		{ 386, 463, 56 },
		{ 387, 406, 56 },
		{ 387, 407, 56 },
		{ 387, 408, 56 },
		{ 387, 409, 56 },
		{ 387, 410, 56 },
		{ 387, 411, 56 },
		{ 387, 412, 56 },
		{ 387, 413, 56 },
		{ 387, 414, 56 },
		{ 387, 415, 56 },
		{ 387, 416, 56 },
		{ 387, 417, 56 },
		{ 387, 418, 56 },
		{ 387, 419, 56 },
		{ 387, 420, 56 },
		{ 387, 421, 56 },
		{ 387, 422, 56 },
		{ 387, 423, 56 },
		{ 387, 424, 56 },
		{ 387, 425, 56 },
		{ 387, 426, 56 },
		{ 387, 427, 56 },
		{ 387, 428, 56 },
		{ 387, 429, 56 },
		{ 387, 430, 56 },
		{ 387, 431, 56 },
		{ 387, 432, 56 },
		{ 387, 433, 56 },
		{ 387, 434, 56 },
		{ 387, 435, 56 },
		{ 387, 436, 56 },
		{ 387, 437, 56 },
		{ 387, 438, 56 },
		{ 387, 439, 56 },
		{ 387, 440, 56 },
		{ 387, 441, 56 },
		{ 387, 442, 56 },
		{ 387, 443, 56 },
		{ 387, 444, 56 },
		{ 387, 445, 56 },
		{ 387, 446, 56 },
		{ 387, 447, 56 },
		{ 387, 448, 56 },
		{ 387, 449, 56 },
		{ 387, 450, 56 },
		{ 387, 451, 56 },
		{ 387, 452, 56 },
		{ 387, 453, 56 },
		{ 387, 454, 56 },
		{ 387, 455, 56 },
		{ 387, 456, 56 },
		{ 387, 457, 56 },
		{ 387, 458, 56 },
		{ 387, 459, 56 },
		{ 387, 460, 56 },
		{ 387, 461, 56 },
		{ 387, 462, 56 },
		{ 387, 463, 56 },
		{ 387, 464, 56 },
		{ 388, 407, 56 },
		{ 388, 408, 56 },
		{ 388, 409, 56 },
		{ 388, 410, 56 },
		{ 388, 411, 56 },
		{ 388, 412, 56 },
		{ 388, 413, 56 },
		{ 388, 414, 56 },
		{ 388, 415, 56 },
		{ 388, 416, 56 },
		{ 388, 417, 56 },
		{ 388, 418, 56 },
		{ 388, 419, 56 },
		{ 388, 420, 56 },
		{ 388, 421, 56 },
		{ 388, 422, 56 },
		{ 388, 423, 56 },
		{ 388, 424, 56 },
		{ 388, 425, 56 },
		{ 388, 426, 56 },
		{ 388, 427, 56 },
		{ 388, 428, 56 },
		{ 388, 429, 56 },
		{ 388, 430, 56 },
		{ 388, 431, 56 },
		{ 388, 432, 56 },
		{ 388, 433, 56 },
		{ 388, 434, 56 },
		{ 388, 435, 56 },
		{ 388, 436, 56 },
		{ 388, 437, 56 },
		{ 388, 438, 56 },
		{ 388, 439, 56 },
		{ 388, 440, 56 },
		{ 388, 441, 56 },
		{ 388, 442, 56 },
		{ 388, 443, 56 },
		{ 388, 444, 56 },
		{ 388, 445, 56 },
		{ 388, 446, 56 },
		{ 388, 447, 56 },
		{ 388, 448, 56 },
		{ 388, 449, 56 },
		{ 388, 450, 56 },
		{ 388, 451, 56 },
		{ 388, 452, 56 },
		{ 388, 453, 56 },
		{ 388, 454, 56 },
		{ 388, 455, 56 },
		{ 388, 456, 56 },
		{ 388, 457, 56 },
		{ 388, 458, 56 },
		{ 388, 459, 56 },
		{ 388, 460, 56 },
		{ 388, 461, 56 },
		{ 388, 462, 56 },
		{ 388, 463, 56 },
		{ 388, 464, 56 },
		{ 388, 465, 56 },
		{ 388, 466, 56 },
		{ 389, 408, 56 },
		{ 389, 409, 56 },
		{ 389, 410, 56 },
		{ 389, 411, 56 },
		{ 389, 412, 56 },
		{ 389, 413, 56 },
		{ 389, 414, 56 },
		{ 389, 415, 56 },
		{ 389, 416, 56 },
		{ 389, 417, 56 },
		{ 389, 418, 56 },
		{ 389, 419, 56 },
		{ 389, 420, 56 },
		{ 389, 421, 56 },
		{ 389, 422, 56 },
		{ 389, 423, 56 },
		{ 389, 424, 56 },
		{ 389, 425, 56 },
		{ 389, 426, 56 },
		{ 389, 427, 56 },
		{ 389, 428, 56 },
		{ 389, 429, 56 },
		{ 389, 430, 56 },
		{ 389, 431, 56 },
		{ 389, 432, 56 },
		{ 389, 433, 56 },
		{ 389, 434, 56 },
		{ 389, 435, 56 },
		{ 389, 436, 56 },
		{ 389, 437, 56 },
		{ 389, 438, 56 },
		{ 389, 439, 56 },
		{ 389, 440, 56 },
		{ 389, 441, 56 },
		{ 389, 442, 56 },
		{ 389, 443, 56 },
		{ 389, 444, 56 },
		{ 389, 445, 56 },
		{ 389, 446, 56 },
		{ 389, 447, 56 },
		{ 389, 448, 56 },
		{ 389, 449, 56 },
		{ 389, 450, 56 },
		{ 389, 451, 56 },
		{ 389, 452, 56 },
		{ 389, 453, 56 },
		{ 389, 454, 56 },
		{ 389, 455, 56 },
		{ 389, 456, 56 },
		{ 389, 457, 56 },
		{ 389, 458, 56 },
		{ 389, 459, 56 },
		{ 389, 460, 56 },
		{ 389, 461, 56 },
		{ 389, 462, 56 },
		{ 389, 463, 56 },
		{ 389, 464, 56 },
		{ 389, 465, 56 },
		{ 389, 466, 56 },
		{ 389, 467, 56 },
		{ 389, 468, 56 },
		{ 390, 410, 56 },
		{ 390, 411, 56 },
		{ 390, 412, 56 },
		{ 390, 413, 56 },
		{ 390, 414, 56 },
		{ 390, 415, 56 },
		{ 390, 416, 56 },
		{ 390, 417, 56 },
		{ 390, 418, 56 },
		{ 390, 419, 56 },
		{ 390, 420, 56 },
		{ 390, 421, 56 },
		{ 390, 422, 56 },
		{ 390, 423, 56 },
		{ 390, 424, 56 },
		{ 390, 425, 56 },
		{ 390, 426, 56 },
		{ 390, 427, 56 },
		{ 390, 428, 56 },
		{ 390, 429, 56 },
		{ 390, 430, 56 },
		{ 390, 431, 56 },
		{ 390, 432, 56 },
		{ 390, 433, 56 },
		{ 390, 434, 56 },
		{ 390, 435, 56 },
		{ 390, 436, 56 },
		{ 390, 437, 56 },
		{ 390, 438, 56 },
		{ 390, 439, 56 },
		{ 390, 440, 56 },
		{ 390, 441, 56 },
		{ 390, 442, 56 },
		{ 390, 443, 56 },
		{ 390, 444, 56 },
		{ 390, 445, 56 },
		{ 390, 446, 56 },
		{ 390, 447, 56 },
		{ 390, 448, 56 },
		{ 390, 449, 56 },
		{ 390, 450, 56 },
		{ 390, 451, 56 },
		{ 390, 452, 56 },
		{ 390, 453, 56 },
		{ 390, 454, 56 },
		{ 390, 455, 56 },
		{ 390, 456, 56 },
		{ 390, 457, 56 },
		{ 390, 458, 56 },
		{ 390, 459, 56 },
		{ 390, 460, 56 },
		{ 390, 461, 56 },
		{ 390, 462, 56 },
		{ 390, 463, 56 },
		{ 390, 464, 56 },
		{ 390, 465, 56 },
		{ 390, 466, 56 },
		{ 390, 467, 56 },
		{ 390, 468, 56 },
		{ 390, 469, 56 },
		{ 391, 411, 56 },
		{ 391, 412, 56 },
		{ 391, 413, 56 },
		{ 391, 414, 56 },
		{ 391, 415, 56 },
		{ 391, 416, 56 },
		{ 391, 417, 56 },
		{ 391, 418, 56 },
		{ 391, 419, 56 },
		{ 391, 420, 56 },
		{ 391, 421, 56 },
		{ 391, 422, 56 },
		{ 391, 423, 56 },
		{ 391, 424, 56 },
		{ 391, 425, 56 },
		{ 391, 426, 56 },
		{ 391, 427, 56 },
		{ 391, 428, 56 },
		{ 391, 429, 56 },
		{ 391, 430, 56 },
		{ 391, 431, 56 },
		{ 391, 432, 56 },
		{ 391, 433, 56 },
		{ 391, 434, 56 },
		{ 391, 435, 56 },
		{ 391, 436, 56 },
		{ 391, 437, 56 },
		{ 391, 438, 56 },
		{ 391, 439, 56 },
		{ 391, 440, 56 },
		{ 391, 441, 56 },
		{ 391, 442, 56 },
		{ 391, 443, 56 },
		{ 391, 444, 56 },
		{ 391, 445, 56 },
		{ 391, 446, 56 },
		{ 391, 447, 56 },
		{ 391, 448, 56 },
		{ 391, 449, 56 },
		{ 391, 450, 56 },
		{ 391, 451, 56 },
		{ 391, 452, 56 },
		{ 391, 453, 56 },
		{ 391, 454, 56 },
		{ 391, 455, 56 },
		{ 391, 456, 56 },
		{ 391, 457, 56 },
		{ 391, 458, 56 },
		{ 391, 459, 56 },
		{ 391, 460, 56 },
		{ 391, 461, 56 },
		{ 391, 462, 56 },
		{ 391, 463, 56 },
		{ 391, 464, 56 },
		{ 391, 465, 56 },
		{ 391, 466, 56 },
		{ 391, 467, 56 },
		{ 391, 468, 56 },
		{ 391, 469, 56 },
		{ 391, 470, 56 },
		{ 391, 471, 56 },
		{ 392, 412, 56 },
		{ 392, 413, 56 },
		{ 392, 414, 56 },
		{ 392, 415, 56 },
		{ 392, 416, 56 },
		{ 392, 417, 56 },
		{ 392, 418, 56 },
		{ 392, 419, 56 },
		{ 392, 420, 56 },
		{ 392, 421, 56 },
		{ 392, 422, 56 },
		{ 392, 423, 56 },
		{ 392, 424, 56 },
		{ 392, 425, 56 },
		{ 392, 426, 56 },
		{ 392, 427, 56 },
		{ 392, 428, 56 },
		{ 392, 429, 56 },
		{ 392, 430, 56 },
		{ 392, 431, 56 },
		{ 392, 432, 56 },
		{ 392, 433, 56 },
		{ 392, 434, 56 },
		{ 392, 435, 56 },
		{ 392, 436, 56 },
		{ 392, 437, 56 },
		{ 392, 438, 56 },
		{ 392, 439, 56 },
		{ 392, 440, 56 },
		{ 392, 441, 56 },
		{ 392, 442, 56 },
		{ 392, 443, 56 },
		{ 392, 444, 56 },
		{ 392, 445, 56 },
		{ 392, 446, 56 },
		{ 392, 447, 56 },
		{ 392, 448, 56 },
		{ 392, 449, 56 },
		{ 392, 450, 56 },
		{ 392, 451, 56 },
		{ 392, 452, 56 },
		{ 392, 453, 56 },
		{ 392, 454, 56 },
		{ 392, 455, 56 },
		{ 392, 456, 56 },
		{ 392, 457, 56 },
		{ 392, 458, 56 },
		{ 392, 459, 56 },
		{ 392, 460, 56 },
		{ 392, 461, 56 },
		{ 392, 462, 56 },
		{ 392, 463, 56 },
		{ 392, 464, 56 },
		{ 392, 465, 56 },
		{ 392, 466, 56 },
		{ 392, 467, 56 },
		{ 392, 468, 56 },
		{ 392, 469, 56 },
		{ 392, 470, 56 },
		{ 392, 471, 56 },
		{ 392, 472, 56 },
		{ 393, 413, 56 },
		{ 393, 414, 56 },
		{ 393, 415, 56 },
		{ 393, 416, 56 },
		{ 393, 417, 56 },
		{ 393, 418, 56 },
		{ 393, 419, 56 },
		{ 393, 420, 56 },
		{ 393, 421, 56 },
		{ 393, 422, 56 },
		{ 393, 423, 56 },
		{ 393, 424, 56 },
		{ 393, 425, 56 },
		{ 393, 426, 56 },
		{ 393, 427, 56 },
		{ 393, 428, 56 },
		{ 393, 429, 56 },
		{ 393, 430, 56 },
		{ 393, 431, 56 },
		{ 393, 432, 56 },
		{ 393, 433, 56 },
		{ 393, 434, 56 },
		{ 393, 435, 56 },
		{ 393, 436, 56 },
		{ 393, 437, 56 },
		{ 393, 438, 56 },
		{ 393, 439, 56 },
		{ 393, 440, 56 },
		{ 393, 441, 56 },
		{ 393, 442, 56 },
		{ 393, 443, 56 },
		{ 393, 444, 56 },
		{ 393, 445, 56 },
		{ 393, 446, 56 },
		{ 393, 447, 56 },
		{ 393, 448, 56 },
		{ 393, 449, 56 },
		{ 393, 450, 56 },
		{ 393, 451, 56 },
		{ 393, 452, 56 },
		{ 393, 453, 56 },
		{ 393, 454, 56 },
		{ 393, 455, 56 },
		{ 393, 456, 56 },
		{ 393, 457, 56 },
		{ 393, 458, 56 },
		{ 393, 459, 56 },
		{ 393, 460, 56 },
		{ 393, 461, 56 },
		{ 393, 462, 56 },
		{ 393, 463, 56 },
		{ 393, 464, 56 },
		{ 393, 465, 56 },
		{ 393, 466, 56 },
		{ 393, 467, 56 },
		{ 393, 468, 56 },
		{ 393, 469, 56 },
		{ 393, 470, 56 },
		{ 393, 471, 56 },
		{ 393, 472, 56 },
		{ 393, 473, 56 },
		{ 393, 474, 56 },
		{ 394, 414, 56 },
		{ 394, 415, 56 },
		{ 394, 416, 56 },
		{ 394, 417, 56 },
		{ 394, 418, 56 },
		{ 394, 419, 56 },
		{ 394, 420, 56 },
		{ 394, 421, 56 },
		{ 394, 422, 56 },
		{ 394, 423, 56 },
		{ 394, 424, 56 },
		{ 394, 425, 56 },
		{ 394, 426, 56 },
		{ 394, 427, 56 },
		{ 394, 428, 56 },
		{ 394, 429, 56 },
		{ 394, 430, 56 },
		{ 394, 431, 56 },
		{ 394, 432, 56 },
		{ 394, 433, 56 },
		{ 394, 434, 56 },
		{ 394, 435, 56 },
		{ 394, 436, 56 },
		{ 394, 437, 56 },
		{ 394, 438, 56 },
		{ 394, 439, 56 },
		{ 394, 440, 56 },
		{ 394, 441, 56 },
		{ 394, 442, 56 },
		{ 394, 443, 56 },
		{ 394, 444, 56 },
		{ 394, 445, 56 },
		{ 394, 446, 56 },
		{ 394, 447, 56 },
		{ 394, 448, 56 },
		{ 394, 449, 56 },
		{ 394, 450, 56 },
		{ 394, 451, 56 },
		{ 394, 452, 56 },
		{ 394, 453, 56 },
		{ 394, 454, 56 },
		{ 394, 455, 56 },
		{ 394, 456, 56 },
		{ 394, 457, 56 },
		{ 394, 458, 56 },
		{ 394, 459, 56 },
		{ 394, 460, 56 },
		{ 394, 461, 56 },
		{ 394, 462, 56 },
		{ 394, 463, 56 },
		{ 394, 464, 56 },
		{ 394, 465, 56 },
		{ 394, 466, 56 },
		{ 394, 467, 56 },
		{ 394, 468, 56 },
		{ 394, 469, 56 },
		{ 394, 470, 56 },
		{ 394, 471, 56 },
		{ 394, 472, 56 },
		{ 394, 473, 56 },
		{ 394, 474, 56 },
		{ 394, 475, 56 },
		{ 394, 476, 56 },
		{ 395, 415, 56 },
		{ 395, 416, 56 },
		{ 395, 417, 56 },
		{ 395, 418, 56 },
		{ 395, 419, 56 },
		{ 395, 420, 56 },
		{ 395, 421, 56 },
		{ 395, 422, 56 },
		{ 395, 423, 56 },
		{ 395, 424, 56 },
		{ 395, 425, 56 },
		{ 395, 426, 56 },
		{ 395, 427, 56 },
		{ 395, 428, 56 },
		{ 395, 429, 56 },
		{ 395, 430, 56 },
		{ 395, 431, 56 },
		{ 395, 432, 56 },
		{ 395, 433, 56 },
		{ 395, 434, 56 },
		{ 395, 435, 56 },
		{ 395, 436, 56 },
		{ 395, 437, 56 },
		{ 395, 438, 56 },
		{ 395, 439, 56 },
		{ 395, 440, 56 },
		{ 395, 441, 56 },
		{ 395, 442, 56 },
		{ 395, 443, 56 },
		{ 395, 444, 56 },
		{ 395, 445, 56 },
		{ 395, 446, 56 },
		{ 395, 447, 56 },
		{ 395, 448, 56 },
		{ 395, 449, 56 },
		{ 395, 450, 56 },
		{ 395, 451, 56 },
		{ 395, 452, 56 },
		{ 395, 453, 56 },
		{ 395, 454, 56 },
		{ 395, 455, 56 },
		{ 395, 456, 56 },
		{ 395, 457, 56 },
		{ 395, 458, 56 },
		{ 395, 459, 56 },
		{ 395, 460, 56 },
		{ 395, 461, 56 },
		{ 395, 462, 56 },
		{ 395, 463, 56 },
		{ 395, 464, 56 },
		{ 395, 465, 56 },
		{ 395, 466, 56 },
		{ 395, 467, 56 },
		{ 395, 468, 56 },
		{ 395, 469, 56 },
		{ 395, 470, 56 },
		{ 395, 471, 56 },
		{ 395, 472, 56 },
		{ 395, 473, 56 },
		{ 395, 474, 56 },
		{ 395, 475, 56 },
		{ 395, 476, 56 },
		{ 395, 477, 56 },
		{ 396, 416, 56 },
		{ 396, 417, 56 },
		{ 396, 418, 56 },
		{ 396, 419, 56 },
		{ 396, 420, 56 },
		{ 396, 421, 56 },
		{ 396, 422, 56 },
		{ 396, 423, 56 },
		{ 396, 424, 56 },
		{ 396, 425, 56 },
		{ 396, 426, 56 },
		{ 396, 427, 56 },
		{ 396, 428, 56 },
		{ 396, 429, 56 },
		{ 396, 430, 56 },
		{ 396, 431, 56 },
		{ 396, 432, 56 },
		{ 396, 433, 56 },
		{ 396, 434, 56 },
		{ 396, 435, 56 },
		{ 396, 436, 56 },
		{ 396, 437, 56 },
		{ 396, 438, 56 },
		{ 396, 439, 56 },
		{ 396, 440, 56 },
		{ 396, 441, 56 },
		{ 396, 442, 56 },
		{ 396, 443, 56 },
		{ 396, 444, 56 },
		{ 396, 445, 56 },
		{ 396, 446, 56 },
		{ 396, 447, 56 },
		{ 396, 448, 56 },
		{ 396, 449, 56 },
		{ 396, 450, 56 },
		{ 396, 451, 56 },
		{ 396, 452, 56 },
		{ 396, 453, 56 },
		{ 396, 454, 56 },
		{ 396, 455, 56 },
		{ 396, 456, 56 },
		{ 396, 457, 56 },
		{ 396, 458, 56 },
		{ 396, 459, 56 },
		{ 396, 460, 56 },
		{ 396, 461, 56 },
		{ 396, 462, 56 },
		{ 396, 463, 56 },
		{ 396, 464, 56 },
		{ 396, 465, 56 },
		{ 396, 466, 56 },
		{ 396, 467, 56 },
		{ 396, 468, 56 },
		{ 396, 469, 56 },
		{ 396, 470, 56 },
		{ 396, 471, 56 },
		{ 396, 472, 56 },
		{ 396, 473, 56 },
		{ 396, 474, 56 },
		{ 396, 475, 56 },
		{ 396, 476, 56 },
		{ 396, 477, 56 },
		{ 396, 478, 56 },
		{ 396, 479, 56 },
		{ 397, 418, 56 },
		{ 397, 419, 56 },
		{ 397, 420, 56 },
		{ 397, 421, 56 },
		{ 397, 422, 56 },
		{ 397, 423, 56 },
		{ 397, 424, 56 },
		{ 397, 425, 56 },
		{ 397, 426, 56 },
		{ 397, 427, 56 },
		{ 397, 428, 56 },
		{ 397, 429, 56 },
		{ 397, 430, 56 },
		{ 397, 431, 56 },
		{ 397, 432, 56 },
		{ 397, 433, 56 },
		{ 397, 434, 56 },
		{ 397, 435, 56 },
		{ 397, 436, 56 },
		{ 397, 437, 56 },
		{ 397, 438, 56 },
		{ 397, 439, 56 },
		{ 397, 440, 56 },
		{ 397, 441, 56 },
		{ 397, 442, 56 },
		{ 397, 443, 56 },
		{ 397, 444, 56 },
		{ 397, 445, 56 },
		{ 397, 446, 56 },
		{ 397, 447, 56 },
		{ 397, 448, 56 },
		{ 397, 449, 56 },
		{ 397, 450, 56 },
		{ 397, 451, 56 },
		{ 397, 452, 56 },
		{ 397, 453, 56 },
		{ 397, 454, 56 },
		{ 397, 455, 56 },
		{ 397, 456, 56 },
		{ 397, 457, 56 },
		{ 397, 458, 56 },
		{ 397, 459, 56 },
		{ 397, 460, 56 },
		{ 397, 461, 56 },
		{ 397, 462, 56 },
		{ 397, 463, 56 },
		{ 397, 464, 56 },
		{ 397, 465, 56 },
		{ 397, 466, 56 },
		{ 397, 467, 56 },
		{ 397, 468, 56 },
		{ 397, 469, 56 },
		{ 397, 470, 56 },
		{ 397, 471, 56 },
		{ 397, 472, 56 },
		{ 397, 473, 56 },
		{ 397, 474, 56 },
		{ 397, 475, 56 },
		{ 397, 476, 56 },
		{ 397, 477, 56 },
		{ 397, 478, 56 },
		{ 397, 479, 56 },
		{ 397, 480, 56 },
		{ 398, 419, 56 },
		{ 398, 420, 56 },
		{ 398, 421, 56 },
		{ 398, 422, 56 },
		{ 398, 423, 56 },
		{ 398, 424, 56 },
		{ 398, 425, 56 },
		{ 398, 426, 56 },
		{ 398, 427, 56 },
		{ 398, 428, 56 },
		{ 398, 429, 56 },
		{ 398, 430, 56 },
		{ 398, 431, 56 },
		{ 398, 432, 56 },
		{ 398, 433, 56 },
		{ 398, 434, 56 },
		{ 398, 435, 56 },
		{ 398, 436, 56 },
		{ 398, 437, 56 },
		{ 398, 438, 56 },
		{ 398, 439, 56 },
		{ 398, 440, 56 },
		{ 398, 441, 56 },
		{ 398, 442, 56 },
		{ 398, 443, 56 },
		{ 398, 444, 56 },
		{ 398, 445, 56 },
		{ 398, 446, 56 },
		{ 398, 447, 56 },
		{ 398, 448, 56 },
		{ 398, 449, 56 },
		{ 398, 450, 56 },
		{ 398, 451, 56 },
		{ 398, 452, 56 },
		{ 398, 453, 56 },
		{ 398, 454, 56 },
		{ 398, 455, 56 },
		{ 398, 456, 56 },
		{ 398, 457, 56 },
		{ 398, 458, 56 },
		{ 398, 459, 56 },
		{ 398, 460, 56 },
		{ 398, 461, 56 },
		{ 398, 462, 56 },
		{ 398, 463, 56 },
		{ 398, 464, 56 },
		{ 398, 465, 56 },
		{ 398, 466, 56 },
		{ 398, 467, 56 },
		{ 398, 468, 56 },
		{ 398, 469, 56 },
		{ 398, 470, 56 },
		{ 398, 471, 56 },
		{ 398, 472, 56 },
		{ 398, 473, 56 },
		{ 398, 474, 56 },
		{ 398, 475, 56 },
		{ 398, 476, 56 },
		{ 398, 477, 56 },
		{ 398, 478, 56 },
		{ 398, 479, 56 },
		{ 398, 480, 56 },
		{ 398, 481, 56 },
		{ 398, 482, 56 },
		{ 399, 420, 56 },
		{ 399, 421, 56 },
		{ 399, 422, 56 },
		{ 399, 423, 56 },
		{ 399, 424, 56 },
		{ 399, 425, 56 },
		{ 399, 426, 56 },
		{ 399, 427, 56 },
		{ 399, 428, 56 },
		{ 399, 429, 56 },
		{ 399, 430, 56 },
		{ 399, 431, 56 },
		{ 399, 432, 56 },
		{ 399, 433, 56 },
		{ 399, 434, 56 },
		{ 399, 435, 56 },
		{ 399, 436, 56 },
		{ 399, 437, 56 },
		{ 399, 438, 56 },
		{ 399, 439, 56 },
		{ 399, 440, 56 },
		{ 399, 441, 56 },
		{ 399, 442, 56 },
		{ 399, 443, 56 },
		{ 399, 444, 56 },
		{ 399, 445, 56 },
		{ 399, 446, 56 },
		{ 399, 447, 56 },
		{ 399, 448, 56 },
		{ 399, 449, 56 },
		{ 399, 450, 56 },
		{ 399, 451, 56 },
		{ 399, 452, 56 },
		{ 399, 453, 56 },
		{ 399, 454, 56 },
		{ 399, 455, 56 },
		{ 399, 456, 56 },
		{ 399, 457, 56 },
		{ 399, 458, 56 },
		{ 399, 459, 56 },
		{ 399, 460, 56 },
		{ 399, 461, 56 },
		{ 399, 462, 56 },
		{ 399, 463, 56 },
		{ 399, 464, 56 },
		{ 399, 465, 56 },
		{ 399, 466, 56 },
		{ 399, 467, 56 },
		{ 399, 468, 56 },
		{ 399, 469, 56 },
		{ 399, 470, 56 },
		{ 399, 471, 56 },
		{ 399, 472, 56 },
		{ 399, 473, 56 },
		{ 399, 474, 56 },
		{ 399, 475, 56 },
		{ 399, 476, 56 },
		{ 399, 477, 56 },
		{ 399, 478, 56 },
		{ 399, 479, 56 },
		{ 399, 480, 56 },
		{ 399, 481, 56 },
		{ 399, 482, 56 },
		{ 399, 483, 56 },
		{ 399, 484, 56 },
		{ 400, 421, 56 },
		{ 400, 422, 56 },
		{ 400, 423, 56 },
		{ 400, 424, 56 },
		{ 400, 425, 56 },
		{ 400, 426, 56 },
		{ 400, 427, 56 },
		{ 400, 428, 56 },
		{ 400, 429, 56 },
		{ 400, 430, 56 },
		{ 400, 431, 56 },
		{ 400, 432, 56 },
		{ 400, 433, 56 },
		{ 400, 434, 56 },
		{ 400, 435, 56 },
		{ 400, 436, 56 },
		{ 400, 437, 56 },
		{ 400, 438, 56 },
		{ 400, 439, 56 },
		{ 400, 440, 56 },
		{ 400, 441, 56 },
		{ 400, 442, 56 },
		{ 400, 443, 56 },
		{ 400, 444, 56 },
		{ 400, 445, 56 },
		{ 400, 446, 56 },
		{ 400, 447, 56 },
		{ 400, 448, 56 },
		{ 400, 449, 56 },
		{ 400, 450, 56 },
		{ 400, 451, 56 },
		{ 400, 452, 56 },
		{ 400, 453, 56 },
		{ 400, 454, 56 },
		{ 400, 455, 56 },
		{ 400, 456, 56 },
		{ 400, 457, 56 },
		{ 400, 458, 56 },
		{ 400, 459, 56 },
		{ 400, 460, 56 },
		{ 400, 461, 56 },
		{ 400, 462, 56 },
		{ 400, 463, 56 },
		{ 400, 464, 56 },
		{ 400, 465, 56 },
		{ 400, 466, 56 },
		{ 400, 467, 56 },
		{ 400, 468, 56 },
		{ 400, 469, 56 },
		{ 400, 470, 56 },
		{ 400, 471, 56 },
		{ 400, 472, 56 },
		{ 400, 473, 56 },
		{ 400, 474, 56 },
		{ 400, 475, 56 },
		{ 400, 476, 56 },
		{ 400, 477, 56 },
		{ 400, 478, 56 },
		{ 400, 479, 56 },
		{ 400, 480, 56 },
		{ 400, 481, 56 },
		{ 400, 482, 56 },
		{ 400, 483, 56 },
		{ 400, 484, 56 },
		{ 400, 485, 56 },
		{ 401, 422, 56 },
		{ 401, 423, 56 },
		{ 401, 424, 56 },
		{ 401, 425, 56 },
		{ 401, 426, 56 },
		{ 401, 427, 56 },
		{ 401, 428, 56 },
		{ 401, 429, 56 },
		{ 401, 430, 56 },
		{ 401, 431, 56 },
		{ 401, 432, 56 },
		{ 401, 433, 56 },
		{ 401, 434, 56 },
		{ 401, 435, 56 },
		{ 401, 436, 56 },
		{ 401, 437, 56 },
		{ 401, 438, 56 },
		{ 401, 439, 56 },
		{ 401, 440, 56 },
		{ 401, 441, 56 },
		{ 401, 442, 56 },
		{ 401, 443, 56 },
		{ 401, 444, 56 },
		{ 401, 445, 56 },
		{ 401, 446, 56 },
		{ 401, 447, 56 },
		{ 401, 448, 56 },
		{ 401, 449, 56 },
		{ 401, 450, 56 },
		{ 401, 451, 56 },
		{ 401, 452, 56 },
		{ 401, 453, 56 },
		{ 401, 454, 56 },
		{ 401, 455, 56 },
		{ 401, 456, 56 },
		{ 401, 457, 56 },
		{ 401, 458, 56 },
		{ 401, 459, 56 },
		{ 401, 460, 56 },
		{ 401, 461, 56 },
		{ 401, 462, 56 },
		{ 401, 463, 56 },
		{ 401, 464, 56 },
		{ 401, 465, 56 },
		{ 401, 466, 56 },
		{ 401, 467, 56 },
		{ 401, 468, 56 },
		{ 401, 469, 56 },
		{ 401, 470, 56 },
		{ 401, 471, 56 },
		{ 401, 472, 56 },
		{ 401, 473, 56 },
		{ 401, 474, 56 },
		{ 401, 475, 56 },
		{ 401, 476, 56 },
		{ 401, 477, 56 },
		{ 401, 478, 56 },
		{ 401, 479, 56 },
		{ 401, 480, 56 },
		{ 401, 481, 56 },
		{ 401, 482, 56 },
		{ 401, 483, 56 },
		{ 401, 484, 56 },
		{ 401, 485, 56 },
		{ 401, 486, 56 },
		{ 401, 487, 56 },
		{ 402, 423, 56 },
		{ 402, 424, 56 },
		{ 402, 425, 56 },
		{ 402, 426, 56 },
		{ 402, 427, 56 },
		{ 402, 428, 56 },
		{ 402, 429, 56 },
		{ 402, 430, 56 },
		{ 402, 431, 56 },
		{ 402, 432, 56 },
		{ 402, 433, 56 },
		{ 402, 434, 56 },
		{ 402, 435, 56 },
		{ 402, 436, 56 },
		{ 402, 437, 56 },
		{ 402, 438, 56 },
		{ 402, 439, 56 },
		{ 402, 440, 56 },
		{ 402, 441, 56 },
		{ 402, 442, 56 },
		{ 402, 443, 56 },
		{ 402, 444, 56 },
		{ 402, 445, 56 },
		{ 402, 446, 56 },
		{ 402, 447, 56 },
		{ 402, 448, 56 },
		{ 402, 449, 56 },
		{ 402, 450, 56 },
		{ 402, 451, 56 },
		{ 402, 452, 56 },
		{ 402, 453, 56 },
		{ 402, 454, 56 },
		{ 402, 455, 56 },
		{ 402, 456, 56 },
		{ 402, 457, 56 },
		{ 402, 458, 56 },
		{ 402, 459, 56 },
		{ 402, 460, 56 },
		{ 402, 461, 56 },
		{ 402, 462, 56 },
		{ 402, 463, 56 },
		{ 402, 464, 56 },
		{ 402, 465, 56 },
		{ 402, 466, 56 },
		{ 402, 467, 56 },
		{ 402, 468, 56 },
		{ 402, 469, 56 },
		{ 402, 470, 56 },
		{ 402, 471, 56 },
		{ 402, 472, 56 },
		{ 402, 473, 56 },
		{ 402, 474, 56 },
		{ 402, 475, 56 },
		{ 402, 476, 56 },
		{ 402, 477, 56 },
		{ 402, 478, 56 },
		{ 402, 479, 56 },
		{ 402, 480, 56 },
		{ 402, 481, 56 },
		{ 402, 482, 56 },
		{ 402, 483, 56 },
		{ 402, 484, 56 },
		{ 402, 485, 56 },
		{ 402, 486, 56 },
		{ 402, 487, 56 },
		{ 402, 488, 56 },
		{ 403, 424, 56 },
		{ 403, 425, 56 },
		{ 403, 426, 56 },
		{ 403, 427, 56 },
		{ 403, 428, 56 },
		{ 403, 429, 56 },
		{ 403, 430, 56 },
		{ 403, 431, 56 },
		{ 403, 432, 56 },
		{ 403, 433, 56 },
		{ 403, 434, 56 },
		{ 403, 435, 56 },
		{ 403, 436, 56 },
		{ 403, 437, 56 },
		{ 403, 438, 56 },
		{ 403, 439, 56 },
		{ 403, 440, 56 },
		{ 403, 441, 56 },
		{ 403, 442, 56 },
		{ 403, 443, 56 },
		{ 403, 444, 56 },
		{ 403, 445, 56 },
		{ 403, 446, 56 },
		{ 403, 447, 56 },
		{ 403, 448, 56 },
		{ 403, 449, 56 },
		{ 403, 450, 56 },
		{ 403, 451, 56 },
		{ 403, 452, 56 },
		{ 403, 453, 56 },
		{ 403, 454, 56 },
		{ 403, 455, 56 },
		{ 403, 456, 56 },
		{ 403, 457, 56 },
		{ 403, 458, 56 },
		{ 403, 459, 56 },
		{ 403, 460, 56 },
		{ 403, 461, 56 },
		{ 403, 462, 56 },
		{ 403, 463, 56 },
		{ 403, 464, 56 },
		{ 403, 465, 56 },
		{ 403, 466, 56 },
		{ 403, 467, 56 },
		{ 403, 468, 56 },
		{ 403, 469, 56 },
		{ 403, 470, 56 },
		{ 403, 471, 56 },
		{ 403, 472, 56 },
		{ 403, 473, 56 },
		{ 403, 474, 56 },
		{ 403, 475, 56 },
		{ 403, 476, 56 },
		{ 403, 477, 56 },
		{ 403, 478, 56 },
		{ 403, 479, 56 },
		{ 403, 480, 56 },
		{ 403, 481, 56 },
		{ 403, 482, 56 },
		{ 403, 483, 56 },
		{ 403, 484, 56 },
		{ 403, 485, 56 },
		{ 403, 486, 56 },
		{ 403, 487, 56 },
		{ 403, 488, 56 },
		{ 403, 489, 56 },
		{ 403, 490, 56 },
		{ 404, 426, 56 },
		{ 404, 427, 56 },
		{ 404, 428, 56 },
		{ 404, 429, 56 },
		{ 404, 430, 56 },
		{ 404, 431, 56 },
		{ 404, 432, 56 },
		{ 404, 433, 56 },
		{ 404, 434, 56 },
		{ 404, 435, 56 },
		{ 404, 436, 56 },
		{ 404, 437, 56 },
		{ 404, 438, 56 },
		{ 404, 439, 56 },
		{ 404, 440, 56 },
		{ 404, 441, 56 },
		{ 404, 442, 56 },
		{ 404, 443, 56 },
		{ 404, 444, 56 },
		{ 404, 445, 56 },
		{ 404, 446, 56 },
		{ 404, 447, 56 },
		{ 404, 448, 56 },
		{ 404, 449, 56 },
		{ 404, 450, 56 },
		{ 404, 451, 56 },
		{ 404, 452, 56 },
		{ 404, 453, 56 },
		{ 404, 454, 56 },
		{ 404, 455, 56 },
		{ 404, 456, 56 },
		{ 404, 457, 56 },
		{ 404, 458, 56 },
		{ 404, 459, 56 },
		{ 404, 460, 56 },
		{ 404, 461, 56 },
		{ 404, 462, 56 },
		{ 404, 463, 56 },
		{ 404, 464, 56 },
		{ 404, 465, 56 },
		{ 404, 466, 56 },
		{ 404, 467, 56 },
		{ 404, 468, 56 },
		{ 404, 469, 56 },
		{ 404, 470, 56 },
		{ 404, 471, 56 },
		{ 404, 472, 56 },
		{ 404, 473, 56 },
		{ 404, 474, 56 },
		{ 404, 475, 56 },
		{ 404, 476, 56 },
		{ 404, 477, 56 },
		{ 404, 478, 56 },
		{ 404, 479, 56 },
		{ 404, 480, 56 },
		{ 404, 481, 56 },
		{ 404, 482, 56 },
		{ 404, 483, 56 },
		{ 404, 484, 56 },
		{ 404, 485, 56 },
		{ 404, 486, 56 },
		{ 404, 487, 56 },
		{ 404, 488, 56 },
		{ 404, 489, 56 },
		{ 404, 490, 56 },
		{ 404, 491, 56 },
		{ 404, 492, 56 },
		{ 405, 427, 56 },
		{ 405, 428, 56 },
		{ 405, 429, 56 },
		{ 405, 430, 56 },
		{ 405, 431, 56 },
		{ 405, 432, 56 },
		{ 405, 433, 56 },
		{ 405, 434, 56 },
		{ 405, 435, 56 },
		{ 405, 436, 56 },
		{ 405, 437, 56 },
		{ 405, 438, 56 },
		{ 405, 439, 56 },
		{ 405, 440, 56 },
		{ 405, 441, 56 },
		{ 405, 442, 56 },
		{ 405, 443, 56 },
		{ 405, 444, 56 },
		{ 405, 445, 56 },
		{ 405, 446, 56 },
		{ 405, 447, 56 },
		{ 405, 448, 56 },
		{ 405, 449, 56 },
		{ 405, 450, 56 },
		{ 405, 451, 56 },
		{ 405, 452, 56 },
		{ 405, 453, 56 },
		{ 405, 454, 56 },
		{ 405, 455, 56 },
		{ 405, 456, 56 },
		{ 405, 457, 56 },
		{ 405, 458, 56 },
		{ 405, 459, 56 },
		{ 405, 460, 56 },
		{ 405, 461, 56 },
		{ 405, 462, 56 },
		{ 405, 463, 56 },
		{ 405, 464, 56 },
		{ 405, 465, 56 },
		{ 405, 466, 56 },
		{ 405, 467, 56 },
		{ 405, 468, 56 },
		{ 405, 469, 56 },
		{ 405, 470, 56 },
		{ 405, 471, 56 },
		{ 405, 472, 56 },
		{ 405, 473, 56 },
		{ 405, 474, 56 },
		{ 405, 475, 56 },
		{ 405, 476, 56 },
		{ 405, 477, 56 },
		{ 405, 478, 56 },
		{ 405, 479, 56 },
		{ 405, 480, 56 },
		{ 405, 481, 56 },
		{ 405, 482, 56 },
		{ 405, 483, 56 },
		{ 405, 484, 56 },
		{ 405, 485, 56 },
		{ 405, 486, 56 },
		{ 405, 487, 56 },
		{ 405, 488, 56 },
		{ 405, 489, 56 },
		{ 405, 490, 56 },
		{ 405, 491, 56 },
		{ 405, 492, 56 },
		{ 405, 493, 56 },
		{ 406, 428, 56 },
		{ 406, 429, 56 },
		{ 406, 430, 56 },
		{ 406, 431, 56 },
		{ 406, 432, 56 },
		{ 406, 433, 56 },
		{ 406, 434, 56 },
		{ 406, 435, 56 },
		{ 406, 436, 56 },
		{ 406, 437, 56 },
		{ 406, 438, 56 },
		{ 406, 439, 56 },
		{ 406, 440, 56 },
		{ 406, 441, 56 },
		{ 406, 442, 56 },
		{ 406, 443, 56 },
		{ 406, 444, 56 },
		{ 406, 445, 56 },
		{ 406, 446, 56 },
		{ 406, 447, 56 },
		{ 406, 448, 56 },
		{ 406, 449, 56 },
		{ 406, 450, 56 },
		{ 406, 451, 56 },
		{ 406, 452, 56 },
		{ 406, 453, 56 },
		{ 406, 454, 56 },
		{ 406, 455, 56 },
		{ 406, 456, 56 },
		{ 406, 457, 56 },
		{ 406, 458, 56 },
		{ 406, 459, 56 },
		{ 406, 460, 56 },
		{ 406, 461, 56 },
		{ 406, 462, 56 },
		{ 406, 463, 56 },
		{ 406, 464, 56 },
		{ 406, 465, 56 },
		{ 406, 466, 56 },
		{ 406, 467, 56 },
		{ 406, 468, 56 },
		{ 406, 469, 56 },
		{ 406, 470, 56 },
		{ 406, 471, 56 },
		{ 406, 472, 56 },
		{ 406, 473, 56 },
		{ 406, 474, 56 },
		{ 406, 475, 56 },
		{ 406, 476, 56 },
		{ 406, 477, 56 },
		{ 406, 478, 56 },
		{ 406, 479, 56 },
		{ 406, 480, 56 },
		{ 406, 481, 56 },
		{ 406, 482, 56 },
		{ 406, 483, 56 },
		{ 406, 484, 56 },
		{ 406, 485, 56 },
		{ 406, 486, 56 },
		{ 406, 487, 56 },
		{ 406, 488, 56 },
		{ 406, 489, 56 },
		{ 406, 490, 56 },
		{ 406, 491, 56 },
		{ 406, 492, 56 },
		{ 406, 493, 56 },
		{ 406, 494, 56 },
		{ 406, 495, 56 },
		{ 407, 429, 56 },
		{ 407, 430, 56 },
		{ 407, 431, 56 },
		{ 407, 432, 56 },
		{ 407, 433, 56 },
		{ 407, 434, 56 },
		{ 407, 435, 56 },
		{ 407, 436, 56 },
		{ 407, 437, 56 },
		{ 407, 438, 56 },
		{ 407, 439, 56 },
		{ 407, 440, 56 },
		{ 407, 441, 56 },
		{ 407, 442, 56 },
		{ 407, 443, 56 },
		{ 407, 444, 56 },
		{ 407, 445, 56 },
		{ 407, 446, 56 },
		{ 407, 447, 56 },
		{ 407, 448, 56 },
		{ 407, 449, 56 },
		{ 407, 450, 56 },
		{ 407, 451, 56 },
		{ 407, 452, 56 },
		{ 407, 453, 56 },
		{ 407, 454, 56 },
		{ 407, 455, 56 },
		{ 407, 456, 56 },
		{ 407, 457, 56 },
		{ 407, 458, 56 },
		{ 407, 459, 56 },
		{ 407, 460, 56 },
		{ 407, 461, 56 },
		{ 407, 462, 56 },
		{ 407, 463, 56 },
		{ 407, 464, 56 },
		{ 407, 465, 56 },
		{ 407, 466, 56 },
		{ 407, 467, 56 },
		{ 407, 468, 56 },
		{ 407, 469, 56 },
		{ 407, 470, 56 },
		{ 407, 471, 56 },
		{ 407, 472, 56 },
		{ 407, 473, 56 },
		{ 407, 474, 56 },
		{ 407, 475, 56 },
		{ 407, 476, 56 },
		{ 407, 477, 56 },
		{ 407, 478, 56 },
		{ 407, 479, 56 },
		{ 407, 480, 56 },
		{ 407, 481, 56 },
		{ 407, 482, 56 },
		{ 407, 483, 56 },
		{ 407, 484, 56 },
		{ 407, 485, 56 },
		{ 407, 486, 56 },
		{ 407, 487, 56 },
		{ 407, 488, 56 },
		{ 407, 489, 56 },
		{ 407, 490, 56 },
		{ 407, 491, 56 },
		{ 407, 492, 56 },
		{ 407, 493, 56 },
		{ 407, 494, 56 },
		{ 407, 495, 56 },
		{ 407, 496, 56 },
		{ 408, 430, 56 },
		{ 408, 431, 56 },
		{ 408, 432, 56 },
		{ 408, 433, 56 },
		{ 408, 434, 56 },
		{ 408, 435, 56 },
		{ 408, 436, 56 },
		{ 408, 437, 56 },
		{ 408, 438, 56 },
		{ 408, 439, 56 },
		{ 408, 440, 56 },
		{ 408, 441, 56 },
		{ 408, 442, 56 },
		{ 408, 443, 56 },
		{ 408, 444, 56 },
		{ 408, 445, 56 },
		{ 408, 446, 56 },
		{ 408, 447, 56 },
		{ 408, 448, 56 },
		{ 408, 449, 56 },
		{ 408, 450, 56 },
		{ 408, 451, 56 },
		{ 408, 452, 56 },
		{ 408, 453, 56 },
		{ 408, 454, 56 },
		{ 408, 455, 56 },
		{ 408, 456, 56 },
		{ 408, 457, 56 },
		{ 408, 458, 56 },
		{ 408, 459, 56 },
		{ 408, 460, 56 },
		{ 408, 461, 56 },
		{ 408, 462, 56 },
		{ 408, 463, 56 },
		{ 408, 464, 56 },
		{ 408, 465, 56 },
		{ 408, 466, 56 },
		{ 408, 467, 56 },
		{ 408, 468, 56 },
		{ 408, 469, 56 },
		{ 408, 470, 56 },
		{ 408, 471, 56 },
		{ 408, 472, 56 },
		{ 408, 473, 56 },
		{ 408, 474, 56 },
		{ 408, 475, 56 },
		{ 408, 476, 56 },
		{ 408, 477, 56 },
		{ 408, 478, 56 },
		{ 408, 479, 56 },
		{ 408, 480, 56 },
		{ 408, 481, 56 },
		{ 408, 482, 56 },
		{ 408, 483, 56 },
		{ 408, 484, 56 },
		{ 408, 485, 56 },
		{ 408, 486, 56 },
		{ 408, 487, 56 },
		{ 408, 488, 56 },
		{ 408, 489, 56 },
		{ 408, 490, 56 },
		{ 408, 491, 56 },
		{ 408, 492, 56 },
		{ 408, 493, 56 },
		{ 408, 494, 56 },
		{ 408, 495, 56 },
		{ 408, 496, 56 },
		{ 408, 497, 56 },
		{ 408, 498, 56 },
		{ 409, 431, 56 },
		{ 409, 432, 56 },
		{ 409, 433, 56 },
		{ 409, 434, 56 },
		{ 409, 435, 56 },
		{ 409, 436, 56 },
		{ 409, 437, 56 },
		{ 409, 438, 56 },
		{ 409, 439, 56 },
		{ 409, 440, 56 },
		{ 409, 441, 56 },
		{ 409, 442, 56 },
		{ 409, 443, 56 },
		{ 409, 444, 56 },
		{ 409, 445, 56 },
		{ 409, 446, 56 },
		{ 409, 447, 56 },
		{ 409, 448, 56 },
		{ 409, 449, 56 },
		{ 409, 450, 56 },
		{ 409, 451, 56 },
		{ 409, 452, 56 },
		{ 409, 453, 56 },
		{ 409, 454, 56 },
		{ 409, 455, 56 },
		{ 409, 456, 56 },
		{ 409, 457, 56 },
		{ 409, 458, 56 },
		{ 409, 459, 56 },
		{ 409, 460, 56 },
		{ 409, 461, 56 },
		{ 409, 462, 56 },
		{ 409, 463, 56 },
		{ 409, 464, 56 },
		{ 409, 465, 56 },
		{ 409, 466, 56 },
		{ 409, 467, 56 },
		{ 409, 468, 56 },
		{ 409, 469, 56 },
		{ 409, 470, 56 },
		{ 409, 471, 56 },
		{ 409, 472, 56 },
		{ 409, 473, 56 },
		{ 409, 474, 56 },
		{ 409, 475, 56 },
		{ 409, 476, 56 },
		{ 409, 477, 56 },
		{ 409, 478, 56 },
		{ 409, 479, 56 },
		{ 409, 480, 56 },
		{ 409, 481, 56 },
		{ 409, 482, 56 },
		{ 409, 483, 56 },
		{ 409, 484, 56 },
		{ 409, 485, 56 },
		{ 409, 486, 56 },
		{ 409, 487, 56 },
		{ 409, 488, 56 },
		{ 409, 489, 56 },
		{ 409, 490, 56 },
		{ 409, 491, 56 },
		{ 409, 492, 56 },
		{ 409, 493, 56 },
		{ 409, 494, 56 },
		{ 409, 495, 56 },
		{ 409, 496, 56 },
		{ 409, 497, 56 },
		{ 409, 498, 56 },
		{ 409, 499, 56 },
		{ 409, 500, 56 },
		{ 410, 432, 56 },
		{ 410, 433, 56 },
		{ 410, 434, 56 },
		{ 410, 435, 56 },
		{ 410, 436, 56 },
		{ 410, 437, 56 },
		{ 410, 438, 56 },
		{ 410, 439, 56 },
		{ 410, 440, 56 },
		{ 410, 441, 56 },
		{ 410, 442, 56 },
		{ 410, 443, 56 },
		{ 410, 444, 56 },
		{ 410, 445, 56 },
		{ 410, 446, 56 },
		{ 410, 447, 56 },
		{ 410, 448, 56 },
		{ 410, 449, 56 },
		{ 410, 450, 56 },
		{ 410, 451, 56 },
		{ 410, 452, 56 },
		{ 410, 453, 56 },
		{ 410, 454, 56 },
		{ 410, 455, 56 },
		{ 410, 456, 56 },
		{ 410, 457, 56 },
		{ 410, 458, 56 },
		{ 410, 459, 56 },
		{ 410, 460, 56 },
		{ 410, 461, 56 },
		{ 410, 462, 56 },
		{ 410, 463, 56 },
		{ 410, 464, 56 },
		{ 410, 465, 56 },
		{ 410, 466, 56 },
		{ 410, 467, 56 },
		{ 410, 468, 56 },
		{ 410, 469, 56 },
		{ 410, 470, 56 },
		{ 410, 471, 56 },
		{ 410, 472, 56 },
		{ 410, 473, 56 },
		{ 410, 474, 56 },
		{ 410, 475, 56 },
		{ 410, 476, 56 },
		{ 410, 477, 56 },
		{ 410, 478, 56 },
		{ 410, 479, 56 },
		{ 410, 480, 56 },
		{ 410, 481, 56 },
		{ 410, 482, 56 },
		{ 410, 483, 56 },
		{ 410, 484, 56 },
		{ 410, 485, 56 },
		{ 410, 486, 56 },
		{ 410, 487, 56 },
		{ 410, 488, 56 },
		{ 410, 489, 56 },
		{ 410, 490, 56 },
		{ 410, 491, 56 },
		{ 410, 492, 56 },
		{ 410, 493, 56 },
		{ 410, 494, 56 },
		{ 410, 495, 56 },
		{ 410, 496, 56 },
		{ 410, 497, 56 },
		{ 410, 498, 56 },
		{ 410, 499, 56 },
		{ 410, 500, 56 },
		{ 410, 501, 56 },
		{ 411, 434, 56 },
		{ 411, 435, 56 },
		{ 411, 436, 56 },
		{ 411, 437, 56 },
		{ 411, 438, 56 },
		{ 411, 439, 56 },
		{ 411, 440, 56 },
		{ 411, 441, 56 },
		{ 411, 442, 56 },
		{ 411, 443, 56 },
		{ 411, 444, 56 },
		{ 411, 445, 56 },
		{ 411, 446, 56 },
		{ 411, 447, 56 },
		{ 411, 448, 56 },
		{ 411, 449, 56 },
		{ 411, 450, 56 },
		{ 411, 451, 56 },
		{ 411, 452, 56 },
		{ 411, 453, 56 },
		{ 411, 454, 56 },
		{ 411, 455, 56 },
		{ 411, 456, 56 },
		{ 411, 457, 56 },
		{ 411, 458, 56 },
		{ 411, 459, 56 },
		{ 411, 460, 56 },
		{ 411, 461, 56 },
		{ 411, 462, 56 },
		{ 411, 463, 56 },
		{ 411, 464, 56 },
		{ 411, 465, 56 },
		{ 411, 466, 56 },
		{ 411, 467, 56 },
		{ 411, 468, 56 },
		{ 411, 469, 56 },
		{ 411, 470, 56 },
		{ 411, 471, 56 },
		{ 411, 472, 56 },
		{ 411, 473, 56 },
		{ 411, 474, 56 },
		{ 411, 475, 56 },
		{ 411, 476, 56 },
		{ 411, 477, 56 },
		{ 411, 478, 56 },
		{ 411, 479, 56 },
		{ 411, 480, 56 },
		{ 411, 481, 56 },
		{ 411, 482, 56 },
		{ 411, 483, 56 },
		{ 411, 484, 56 },
		{ 411, 485, 56 },
		{ 411, 486, 56 },
		{ 411, 487, 56 },
		{ 411, 488, 56 },
		{ 411, 489, 56 },
		{ 411, 490, 56 },
		{ 411, 491, 56 },
		{ 411, 492, 56 },
		{ 411, 493, 56 },
		{ 411, 494, 56 },
		{ 411, 495, 56 },
		{ 411, 496, 56 },
		{ 411, 497, 56 },
		{ 411, 498, 56 },
		{ 411, 499, 56 },
		{ 411, 500, 56 },
		{ 411, 501, 56 },
		{ 411, 502, 56 },
		{ 411, 503, 56 },
		{ 412, 435, 56 },
		{ 412, 436, 56 },
		{ 412, 437, 56 },
		{ 412, 438, 56 },
		{ 412, 439, 56 },
		{ 412, 440, 56 },
		{ 412, 441, 56 },
		{ 412, 442, 56 },
		{ 412, 443, 56 },
		{ 412, 444, 56 },
		{ 412, 445, 56 },
		{ 412, 446, 56 },
		{ 412, 447, 56 },
		{ 412, 448, 56 },
		{ 412, 449, 56 },
		{ 412, 450, 56 },
		{ 412, 451, 56 },
		{ 412, 452, 56 },
		{ 412, 453, 56 },
		{ 412, 454, 56 },
		{ 412, 455, 56 },
		{ 412, 456, 56 },
		{ 412, 457, 56 },
		{ 412, 458, 56 },
		{ 412, 459, 56 },
		{ 412, 460, 56 },
		{ 412, 461, 56 },
		{ 412, 462, 56 },
		{ 412, 463, 56 },
		{ 412, 464, 56 },
		{ 412, 465, 56 },
		{ 412, 466, 56 },
		{ 412, 467, 56 },
		{ 412, 468, 56 },
		{ 412, 469, 56 },
		{ 412, 470, 56 },
		{ 412, 471, 56 },
		{ 412, 472, 56 },
		{ 412, 473, 56 },
		{ 412, 474, 56 },
		{ 412, 475, 56 },
		{ 412, 476, 56 },
		{ 412, 477, 56 },
		{ 412, 478, 56 },
		{ 412, 479, 56 },
		{ 412, 480, 56 },
		{ 412, 481, 56 },
		{ 412, 482, 56 },
		{ 412, 483, 56 },
		{ 412, 484, 56 },
		{ 412, 485, 56 },
		{ 412, 486, 56 },
		{ 412, 487, 56 },
		{ 412, 488, 56 },
		{ 412, 489, 56 },
		{ 412, 490, 56 },
		{ 412, 491, 56 },
		{ 412, 492, 56 },
		{ 412, 493, 56 },
		{ 412, 494, 56 },
		{ 412, 495, 56 },
		{ 412, 496, 56 },
		{ 412, 497, 56 },
		{ 412, 498, 56 },
		{ 412, 499, 56 },
		{ 412, 500, 56 },
		{ 412, 501, 56 },
		{ 412, 502, 56 },
		{ 412, 503, 56 },
		{ 412, 504, 56 },
		{ 413, 436, 56 },
		{ 413, 437, 56 },
		{ 413, 438, 56 },
		{ 413, 439, 56 },
		{ 413, 440, 56 },
		{ 413, 441, 56 },
		{ 413, 442, 56 },
		{ 413, 443, 56 },
		{ 413, 444, 56 },
		{ 413, 445, 56 },
		{ 413, 446, 56 },
		{ 413, 447, 56 },
		{ 413, 448, 56 },
		{ 413, 449, 56 },
		{ 413, 450, 56 },
		{ 413, 451, 56 },
		{ 413, 452, 56 },
		{ 413, 453, 56 },
		{ 413, 454, 56 },
		{ 413, 455, 56 },
		{ 413, 456, 56 },
		{ 413, 457, 56 },
		{ 413, 458, 56 },
		{ 413, 459, 56 },
		{ 413, 460, 56 },
		{ 413, 461, 56 },
		{ 413, 462, 56 },
		{ 413, 463, 56 },
		{ 413, 464, 56 },
		{ 413, 465, 56 },
		{ 413, 466, 56 },
		{ 413, 467, 56 },
		{ 413, 468, 56 },
		{ 413, 469, 56 },
		{ 413, 470, 56 },
		{ 413, 471, 56 },
		{ 413, 472, 56 },
		{ 413, 473, 56 },
		{ 413, 474, 56 },
		{ 413, 475, 56 },
		{ 413, 476, 56 },
		{ 413, 477, 56 },
		{ 413, 478, 56 },
		{ 413, 479, 56 },
		{ 413, 480, 56 },
		{ 413, 481, 56 },
		{ 413, 482, 56 },
		{ 413, 483, 56 },
		{ 413, 484, 56 },
		{ 413, 485, 56 },
		{ 413, 486, 56 },
		{ 413, 487, 56 },
		{ 413, 488, 56 },
		{ 413, 489, 56 },
		{ 413, 490, 56 },
		{ 413, 491, 56 },
		{ 413, 492, 56 },
		{ 413, 493, 56 },
		{ 413, 494, 56 },
		{ 413, 495, 56 },
		{ 413, 496, 56 },
		{ 413, 497, 56 },
		{ 413, 498, 56 },
		{ 413, 499, 56 },
		{ 413, 500, 56 },
		{ 413, 501, 56 },
		{ 413, 502, 56 },
		{ 413, 503, 56 },
		{ 413, 504, 56 },
		{ 413, 505, 56 },
		{ 413, 506, 56 },
		{ 414, 437, 56 },
		{ 414, 438, 56 },
		{ 414, 439, 56 },
		{ 414, 440, 56 },
		{ 414, 441, 56 },
		{ 414, 442, 56 },
		{ 414, 443, 56 },
		{ 414, 444, 56 },
		{ 414, 445, 56 },
		{ 414, 446, 56 },
		{ 414, 447, 56 },
		{ 414, 448, 56 },
		{ 414, 449, 56 },
		{ 414, 450, 56 },
		{ 414, 451, 56 },
		{ 414, 452, 56 },
		{ 414, 453, 56 },
		{ 414, 454, 56 },
		{ 414, 455, 56 },
		{ 414, 456, 56 },
		{ 414, 457, 56 },
		{ 414, 458, 56 },
		{ 414, 459, 56 },
		{ 414, 460, 56 },
		{ 414, 461, 56 },
		{ 414, 462, 56 },
		{ 414, 463, 56 },
		{ 414, 464, 56 },
		{ 414, 465, 56 },
		{ 414, 466, 56 },
		{ 414, 467, 56 },
		{ 414, 468, 56 },
		{ 414, 469, 56 },
		{ 414, 470, 56 },
		{ 414, 471, 56 },
		{ 414, 472, 56 },
		{ 414, 473, 56 },
		{ 414, 474, 56 },
		{ 414, 475, 56 },
		{ 414, 476, 56 },
		{ 414, 477, 56 },
		{ 414, 478, 56 },
		{ 414, 479, 56 },
		{ 414, 480, 56 },
		{ 414, 481, 56 },
		{ 414, 482, 56 },
		{ 414, 483, 56 },
		{ 414, 484, 56 },
		{ 414, 485, 56 },
		{ 414, 486, 56 },
		{ 414, 487, 56 },
		{ 414, 488, 56 },
		{ 414, 489, 56 },
		{ 414, 490, 56 },
		{ 414, 491, 56 },
		{ 414, 492, 56 },
		{ 414, 493, 56 },
		{ 414, 494, 56 },
		{ 414, 495, 56 },
		{ 414, 496, 56 },
		{ 414, 497, 56 },
		{ 414, 498, 56 },
		{ 414, 499, 56 },
		{ 414, 500, 56 },
		{ 414, 501, 56 },
		{ 414, 502, 56 },
		{ 414, 503, 56 },
		{ 414, 504, 56 },
		{ 414, 505, 56 },
		{ 414, 506, 56 },
		{ 414, 507, 56 },
		{ 414, 508, 56 },
		{ 415, 438, 56 },
		{ 415, 439, 56 },
		{ 415, 440, 56 },
		{ 415, 441, 56 },
		{ 415, 442, 56 },
		{ 415, 443, 56 },
		{ 415, 444, 56 },
		{ 415, 445, 56 },
		{ 415, 446, 56 },
		{ 415, 447, 56 },
		{ 415, 448, 56 },
		{ 415, 449, 56 },
		{ 415, 450, 56 },
		{ 415, 451, 56 },
		{ 415, 452, 56 },
		{ 415, 453, 56 },
		{ 415, 454, 56 },
		{ 415, 455, 56 },
		{ 415, 456, 56 },
		{ 415, 457, 56 },
		{ 415, 458, 56 },
		{ 415, 459, 56 },
		{ 415, 460, 56 },
		{ 415, 461, 56 },
		{ 415, 462, 56 },
		{ 415, 463, 56 },
		{ 415, 464, 56 },
		{ 415, 465, 56 },
		{ 415, 466, 56 },
		{ 415, 467, 56 },
		{ 415, 468, 56 },
		{ 415, 469, 56 },
		{ 415, 470, 56 },
		{ 415, 471, 56 },
		{ 415, 472, 56 },
		{ 415, 473, 56 },
		{ 415, 474, 56 },
		{ 415, 475, 56 },
		{ 415, 476, 56 },
		{ 415, 477, 56 },
		{ 415, 478, 56 },
		{ 415, 479, 56 },
		{ 415, 480, 56 },
		{ 415, 481, 56 },
		{ 415, 482, 56 },
		{ 415, 483, 56 },
		{ 415, 484, 56 },
		{ 415, 485, 56 },
		{ 415, 486, 56 },
		{ 415, 487, 56 },
		{ 415, 488, 56 },
		{ 415, 489, 56 },
		{ 415, 490, 56 },
		{ 415, 491, 56 },
		{ 415, 492, 56 },
		{ 415, 493, 56 },
		{ 415, 494, 56 },
		{ 415, 495, 56 },
		{ 415, 496, 56 },
		{ 415, 497, 56 },
		{ 415, 498, 56 },
		{ 415, 499, 56 },
		{ 415, 500, 56 },
		{ 415, 501, 56 },
		{ 415, 502, 56 },
		{ 415, 503, 56 },
		{ 415, 504, 56 },
		{ 415, 505, 56 },
		{ 415, 506, 56 },
		{ 415, 507, 56 },
		{ 415, 508, 56 },
		{ 415, 509, 56 },
		{ 416, 439, 56 },
		{ 416, 440, 56 },
		{ 416, 441, 56 },
		{ 416, 442, 56 },
		{ 416, 443, 56 },
		{ 416, 444, 56 },
		{ 416, 445, 56 },
		{ 416, 446, 56 },
		{ 416, 447, 56 },
		{ 416, 448, 56 },
		{ 416, 449, 56 },
		{ 416, 450, 56 },
		{ 416, 451, 56 },
		{ 416, 452, 56 },
		{ 416, 453, 56 },
		{ 416, 454, 56 },
		{ 416, 455, 56 },
		{ 416, 456, 56 },
		{ 416, 457, 56 },
		{ 416, 458, 56 },
		{ 416, 459, 56 },
		{ 416, 460, 56 },
		{ 416, 461, 56 },
		{ 416, 462, 56 },
		{ 416, 463, 56 },
		{ 416, 464, 56 },
		{ 416, 465, 56 },
		{ 416, 466, 56 },
		{ 416, 467, 56 },
		{ 416, 468, 56 },
		{ 416, 469, 56 },
		{ 416, 470, 56 },
		{ 416, 471, 56 },
		{ 416, 472, 56 },
		{ 416, 473, 56 },
		{ 416, 474, 56 },
		{ 416, 475, 56 },
		{ 416, 476, 56 },
		{ 416, 477, 56 },
		{ 416, 478, 56 },
		{ 416, 479, 56 },
		{ 416, 480, 56 },
		{ 416, 481, 56 },
		{ 416, 482, 56 },
		{ 416, 483, 56 },
		{ 416, 484, 56 },
		{ 416, 485, 56 },
		{ 416, 486, 56 },
		{ 416, 487, 56 },
		{ 416, 488, 56 },
		{ 416, 489, 56 },
		{ 416, 490, 56 },
		{ 416, 491, 56 },
		{ 416, 492, 56 },
		{ 416, 493, 56 },
		{ 416, 494, 56 },
		{ 416, 495, 56 },
		{ 416, 496, 56 },
		{ 416, 497, 56 },
		{ 416, 498, 56 },
		{ 416, 499, 56 },
		{ 416, 500, 56 },
		{ 416, 501, 56 },
		{ 416, 502, 56 },
		{ 416, 503, 56 },
		{ 416, 504, 56 },
		{ 416, 505, 56 },
		{ 416, 506, 56 },
		{ 416, 507, 56 },
		{ 416, 508, 56 },
		{ 416, 509, 56 },
		{ 416, 510, 56 },
		{ 416, 511, 56 },
		{ 417, 440, 56 },
		{ 417, 441, 56 },
		{ 417, 442, 56 },
		{ 417, 443, 56 },
		{ 417, 444, 56 },
		{ 417, 445, 56 },
		{ 417, 446, 56 },
		{ 417, 447, 56 },
		{ 417, 448, 56 },
		{ 417, 449, 56 },
		{ 417, 450, 56 },
		{ 417, 451, 56 },
		{ 417, 452, 56 },
		{ 417, 453, 56 },
		{ 417, 454, 56 },
		{ 417, 455, 56 },
		{ 417, 456, 56 },
		{ 417, 457, 56 },
		{ 417, 458, 56 },
		{ 417, 459, 56 },
		{ 417, 460, 56 },
		{ 417, 461, 56 },
		{ 417, 462, 56 },
		{ 417, 463, 56 },
		{ 417, 464, 56 },
		{ 417, 465, 56 },
		{ 417, 466, 56 },
		{ 417, 467, 56 },
		{ 417, 468, 56 },
		{ 417, 469, 56 },
		{ 417, 470, 56 },
		{ 417, 471, 56 },
		{ 417, 472, 56 },
		{ 417, 473, 56 },
		{ 417, 474, 56 },
		{ 417, 475, 56 },
		{ 417, 476, 56 },
		{ 417, 477, 56 },
		{ 417, 478, 56 },
		{ 417, 479, 56 },
		{ 417, 480, 56 },
		{ 417, 481, 56 },
		{ 417, 482, 56 },
		{ 417, 483, 56 },
		{ 417, 484, 56 },
		{ 417, 485, 56 },
		{ 417, 486, 56 },
		{ 417, 487, 56 },
		{ 417, 488, 56 },
		{ 417, 489, 56 },
		{ 417, 490, 56 },
		{ 417, 491, 56 },
		{ 417, 492, 56 },
		{ 417, 493, 56 },
		{ 417, 494, 56 },
		{ 417, 495, 56 },
		{ 417, 496, 56 },
		{ 417, 497, 56 },
		{ 417, 498, 56 },
		{ 417, 499, 56 },
		{ 417, 500, 56 },
		{ 417, 501, 56 },
		{ 417, 502, 56 },
		{ 417, 503, 56 },
		{ 417, 504, 56 },
		{ 417, 505, 56 },
		{ 417, 506, 56 },
		{ 417, 507, 56 },
		{ 417, 508, 56 },
		{ 417, 509, 56 },
		{ 417, 510, 56 },
		{ 417, 511, 56 },
		{ 417, 512, 56 },
		{ 418, 442, 56 },
		{ 418, 443, 56 },
		{ 418, 444, 56 },
		{ 418, 445, 56 },
		{ 418, 446, 56 },
		{ 418, 447, 56 },
		{ 418, 448, 56 },
		{ 418, 449, 56 },
		{ 418, 450, 56 },
		{ 418, 451, 56 },
		{ 418, 452, 56 },
		{ 418, 453, 56 },
		{ 418, 454, 56 },
		{ 418, 455, 56 },
		{ 418, 456, 56 },
		{ 418, 457, 56 },
		{ 418, 458, 56 },
		{ 418, 459, 56 },
		{ 418, 460, 56 },
		{ 418, 461, 56 },
		{ 418, 462, 56 },
		{ 418, 463, 56 },
		{ 418, 464, 56 },
		{ 418, 465, 56 },
		{ 418, 466, 56 },
		{ 418, 467, 56 },
		{ 418, 468, 56 },
		{ 418, 469, 56 },
		{ 418, 470, 56 },
		{ 418, 471, 56 },
		{ 418, 472, 56 },
		{ 418, 473, 56 },
		{ 418, 474, 56 },
		{ 418, 475, 56 },
		{ 418, 476, 56 },
		{ 418, 477, 56 },
		{ 418, 478, 56 },
		{ 418, 479, 56 },
		{ 418, 480, 56 },
		{ 418, 481, 56 },
		{ 418, 482, 56 },
		{ 418, 483, 56 },
		{ 418, 484, 56 },
		{ 418, 485, 56 },
		{ 418, 486, 56 },
		{ 418, 487, 56 },
		{ 418, 488, 56 },
		{ 418, 489, 56 },
		{ 418, 490, 56 },
		{ 418, 491, 56 },
		{ 418, 492, 56 },
		{ 418, 493, 56 },
		{ 418, 494, 56 },
		{ 418, 495, 56 },
		{ 418, 496, 56 },
		{ 418, 497, 56 },
		{ 418, 498, 56 },
		{ 418, 499, 56 },
		{ 418, 500, 56 },
		{ 418, 501, 56 },
		{ 418, 502, 56 },
		{ 418, 503, 56 },
		{ 418, 504, 56 },
		{ 418, 505, 56 },
		{ 418, 506, 56 },
		{ 418, 507, 56 },
		{ 418, 508, 56 },
		{ 418, 509, 56 },
		{ 418, 510, 56 },
		{ 418, 511, 56 },
		{ 418, 512, 56 },
		{ 419, 443, 56 },
		{ 419, 444, 56 },
		{ 419, 445, 56 },
		{ 419, 446, 56 },
		{ 419, 447, 56 },
		{ 419, 448, 56 },
		{ 419, 449, 56 },
		{ 419, 450, 56 },
		{ 419, 451, 56 },
		{ 419, 452, 56 },
		{ 419, 453, 56 },
		{ 419, 454, 56 },
		{ 419, 455, 56 },
		{ 419, 456, 56 },
		{ 419, 457, 56 },
		{ 419, 458, 56 },
		{ 419, 459, 56 },
		{ 419, 460, 56 },
		{ 419, 461, 56 },
		{ 419, 462, 56 },
		{ 419, 463, 56 },
		{ 419, 464, 56 },
		{ 419, 465, 56 },
		{ 419, 466, 56 },
		{ 419, 467, 56 },
		{ 419, 468, 56 },
		{ 419, 469, 56 },
		{ 419, 470, 56 },
		{ 419, 471, 56 },
		{ 419, 472, 56 },
		{ 419, 473, 56 },
		{ 419, 474, 56 },
		{ 419, 475, 56 },
		{ 419, 476, 56 },
		{ 419, 477, 56 },
		{ 419, 478, 56 },
		{ 419, 479, 56 },
		{ 419, 480, 56 },
		{ 419, 481, 56 },
		{ 419, 482, 56 },
		{ 419, 483, 56 },
		{ 419, 484, 56 },
		{ 419, 485, 56 },
		{ 419, 486, 56 },
		{ 419, 487, 56 },
		{ 419, 488, 56 },
		{ 419, 489, 56 },
		{ 419, 490, 56 },
		{ 419, 491, 56 },
		{ 419, 492, 56 },
		{ 419, 493, 56 },
		{ 419, 494, 56 },
		{ 419, 495, 56 },
		{ 419, 496, 56 },
		{ 419, 497, 56 },
		{ 419, 498, 56 },
		{ 419, 499, 56 },
		{ 419, 500, 56 },
		{ 419, 501, 56 },
		{ 419, 502, 56 },
		{ 419, 503, 56 },
		{ 419, 504, 56 },
		{ 419, 505, 56 },
		{ 419, 506, 56 },
		{ 419, 507, 56 },
		{ 419, 508, 56 },
		{ 419, 509, 56 },
		{ 419, 510, 56 },
		{ 419, 511, 56 },
		{ 419, 512, 56 },
		{ 420, 444, 56 },
		{ 420, 445, 56 },
		{ 420, 446, 56 },
		{ 420, 447, 56 },
		{ 420, 448, 56 },
		{ 420, 449, 56 },
		{ 420, 450, 56 },
		{ 420, 451, 56 },
		{ 420, 452, 56 },
		{ 420, 453, 56 },
		{ 420, 454, 56 },
		{ 420, 455, 56 },
		{ 420, 456, 56 },
		{ 420, 457, 56 },
		{ 420, 458, 56 },
		{ 420, 459, 56 },
		{ 420, 460, 56 },
		{ 420, 461, 56 },
		{ 420, 462, 56 },
		{ 420, 463, 56 },
		{ 420, 464, 56 },
		{ 420, 465, 56 },
		{ 420, 466, 56 },
		{ 420, 467, 56 },
		{ 420, 468, 56 },
		{ 420, 469, 56 },
		{ 420, 470, 56 },
		{ 420, 471, 56 },
		{ 420, 472, 56 },
		{ 420, 473, 56 },
		{ 420, 474, 56 },
		{ 420, 475, 56 },
		{ 420, 476, 56 },
		{ 420, 477, 56 },
		{ 420, 478, 56 },
		{ 420, 479, 56 },
		{ 420, 480, 56 },
		{ 420, 481, 56 },
		{ 420, 482, 56 },
		{ 420, 483, 56 },
		{ 420, 484, 56 },
		{ 420, 485, 56 },
		{ 420, 486, 56 },
		{ 420, 487, 56 },
		{ 420, 488, 56 },
		{ 420, 489, 56 },
		{ 420, 490, 56 },
		{ 420, 491, 56 },
		{ 420, 492, 56 },
		{ 420, 493, 56 },
		{ 420, 494, 56 },
		{ 420, 495, 56 },
		{ 420, 496, 56 },
		{ 420, 497, 56 },
		{ 420, 498, 56 },
		{ 420, 499, 56 },
		{ 420, 500, 56 },
		{ 420, 501, 56 },
		{ 420, 502, 56 },
		{ 420, 503, 56 },
		{ 420, 504, 56 },
		{ 420, 505, 56 },
		{ 420, 506, 56 },
		{ 420, 507, 56 },
		{ 420, 508, 56 },
		{ 420, 509, 56 },
		{ 420, 510, 56 },
		{ 420, 511, 56 },
		{ 420, 512, 56 },
		{ 421, 445, 56 },
		{ 421, 446, 56 },
		{ 421, 447, 56 },
		{ 421, 448, 56 },
		{ 421, 449, 56 },
		{ 421, 450, 56 },
		{ 421, 451, 56 },
		{ 421, 452, 56 },
		{ 421, 453, 56 },
		{ 421, 454, 56 },
		{ 421, 455, 56 },
		{ 421, 456, 56 },
		{ 421, 457, 56 },
		{ 421, 458, 56 },
		{ 421, 459, 56 },
		{ 421, 460, 56 },
		{ 421, 461, 56 },
		{ 421, 462, 56 },
		{ 421, 463, 56 },
		{ 421, 464, 56 },
		{ 421, 465, 56 },
		{ 421, 466, 56 },
		{ 421, 467, 56 },
		{ 421, 468, 56 },
		{ 421, 469, 56 },
		{ 421, 470, 56 },
		{ 421, 471, 56 },
		{ 421, 472, 56 },
		{ 421, 473, 56 },
		{ 421, 474, 56 },
		{ 421, 475, 56 },
		{ 421, 476, 56 },
		{ 421, 477, 56 },
		{ 421, 478, 56 },
		{ 421, 479, 56 },
		{ 421, 480, 56 },
		{ 421, 481, 56 },
		{ 421, 482, 56 },
		{ 421, 483, 56 },
		{ 421, 484, 56 },
		{ 421, 485, 56 },
		{ 421, 486, 56 },
		{ 421, 487, 56 },
		{ 421, 488, 56 },
		{ 421, 489, 56 },
		{ 421, 490, 56 },
		{ 421, 491, 56 },
		{ 421, 492, 56 },
		{ 421, 493, 56 },
		{ 421, 494, 56 },
		{ 421, 495, 56 },
		{ 421, 496, 56 },
		{ 421, 497, 56 },
		{ 421, 498, 56 },
		{ 421, 499, 56 },
		{ 421, 500, 56 },
		{ 421, 501, 56 },
		{ 421, 502, 56 },
		{ 421, 503, 56 },
		{ 421, 504, 56 },
		{ 421, 505, 56 },
		{ 421, 506, 56 },
		{ 421, 507, 56 },
		{ 421, 508, 56 },
		{ 421, 509, 56 },
		{ 421, 510, 56 },
		{ 421, 511, 56 },
		{ 421, 512, 56 },
		{ 422, 446, 56 },
		{ 422, 447, 56 },
		{ 422, 448, 56 },
		{ 422, 449, 56 },
		{ 422, 450, 56 },
		{ 422, 451, 56 },
		{ 422, 452, 56 },
		{ 422, 453, 56 },
		{ 422, 454, 56 },
		{ 422, 455, 56 },
		{ 422, 456, 56 },
		{ 422, 457, 56 },
		{ 422, 458, 56 },
		{ 422, 459, 56 },
		{ 422, 460, 56 },
		{ 422, 461, 56 },
		{ 422, 462, 56 },
		{ 422, 463, 56 },
		{ 422, 464, 56 },
		{ 422, 465, 56 },
		{ 422, 466, 56 },
		{ 422, 467, 56 },
		{ 422, 468, 56 },
		{ 422, 469, 56 },
		{ 422, 470, 56 },
		{ 422, 471, 56 },
		{ 422, 472, 56 },
		{ 422, 473, 56 },
		{ 422, 474, 56 },
		{ 422, 475, 56 },
		{ 422, 476, 56 },
		{ 422, 477, 56 },
		{ 422, 478, 56 },
		{ 422, 479, 56 },
		{ 422, 480, 56 },
		{ 422, 481, 56 },
		{ 422, 482, 56 },
		{ 422, 483, 56 },
		{ 422, 484, 56 },
		{ 422, 485, 56 },
		{ 422, 486, 56 },
		{ 422, 487, 56 },
		{ 422, 488, 56 },
		{ 422, 489, 56 },
		{ 422, 490, 56 },
		{ 422, 491, 56 },
		{ 422, 492, 56 },
		{ 422, 493, 56 },
		{ 422, 494, 56 },
		{ 422, 495, 56 },
		{ 422, 496, 56 },
		{ 422, 497, 56 },
		{ 422, 498, 56 },
		{ 422, 499, 56 },
		{ 422, 500, 56 },
		{ 422, 501, 56 },
		{ 422, 502, 56 },
		{ 422, 503, 56 },
		{ 422, 504, 56 },
		{ 422, 505, 56 },
		{ 422, 506, 56 },
		{ 422, 507, 56 },
		{ 422, 508, 56 },
		{ 422, 509, 56 },
		{ 422, 510, 56 },
		{ 422, 511, 56 },
		{ 422, 512, 56 },
		{ 423, 447, 56 },
		{ 423, 448, 56 },
		{ 423, 449, 56 },
		{ 423, 450, 56 },
		{ 423, 451, 56 },
		{ 423, 452, 56 },
		{ 423, 453, 56 },
		{ 423, 454, 56 },
		{ 423, 455, 56 },
		{ 423, 456, 56 },
		{ 423, 457, 56 },
		{ 423, 458, 56 },
		{ 423, 459, 56 },
		{ 423, 460, 56 },
		{ 423, 461, 56 },
		{ 423, 462, 56 },
		{ 423, 463, 56 },
		{ 423, 464, 56 },
		{ 423, 465, 56 },
		{ 423, 466, 56 },
		{ 423, 467, 56 },
		{ 423, 468, 56 },
		{ 423, 469, 56 },
		{ 423, 470, 56 },
		{ 423, 471, 56 },
		{ 423, 472, 56 },
		{ 423, 473, 56 },
		{ 423, 474, 56 },
		{ 423, 475, 56 },
		{ 423, 476, 56 },
		{ 423, 477, 56 },
		{ 423, 478, 56 },
		{ 423, 479, 56 },
		{ 423, 480, 56 },
		{ 423, 481, 56 },
		{ 423, 482, 56 },
		{ 423, 483, 56 },
		{ 423, 484, 56 },
		{ 423, 485, 56 },
		{ 423, 486, 56 },
		{ 423, 487, 56 },
		{ 423, 488, 56 },
		{ 423, 489, 56 },
		{ 423, 490, 56 },
		{ 423, 491, 56 },
		{ 423, 492, 56 },
		{ 423, 493, 56 },
		{ 423, 494, 56 },
		{ 423, 495, 56 },
		{ 423, 496, 56 },
		{ 423, 497, 56 },
		{ 423, 498, 56 },
		{ 423, 499, 56 },
		{ 423, 500, 56 },
		{ 423, 501, 56 },
		{ 423, 502, 56 },
		{ 423, 503, 56 },
		{ 423, 504, 56 },
		{ 423, 505, 56 },
		{ 423, 506, 56 },
		{ 423, 507, 56 },
		{ 423, 508, 56 },
		{ 423, 509, 56 },
		{ 423, 510, 56 },
		{ 423, 511, 56 },
		{ 423, 512, 56 },
		{ 424, 448, 56 },
		{ 424, 449, 56 },
		{ 424, 450, 56 },
		{ 424, 451, 56 },
		{ 424, 452, 56 },
		{ 424, 453, 56 },
		{ 424, 454, 56 },
		{ 424, 455, 56 },
		{ 424, 456, 56 },
		{ 424, 457, 56 },
		{ 424, 458, 56 },
		{ 424, 459, 56 },
		{ 424, 460, 56 },
		{ 424, 461, 56 },
		{ 424, 462, 56 },
		{ 424, 463, 56 },
		{ 424, 464, 56 },
		{ 424, 465, 56 },
		{ 424, 466, 56 },
		{ 424, 467, 56 },
		{ 424, 468, 56 },
		{ 424, 469, 56 },
		{ 424, 470, 56 },
		{ 424, 471, 56 },
		{ 424, 472, 56 },
		{ 424, 473, 56 },
		{ 424, 474, 56 },
		{ 424, 475, 56 },
		{ 424, 476, 56 },
		{ 424, 477, 56 },
		{ 424, 478, 56 },
		{ 424, 479, 56 },
		{ 424, 480, 56 },
		{ 424, 481, 56 },
		{ 424, 482, 56 },
		{ 424, 483, 56 },
		{ 424, 484, 56 },
		{ 424, 485, 56 },
		{ 424, 486, 56 },
		{ 424, 487, 56 },
		{ 424, 488, 56 },
		{ 424, 489, 56 },
		{ 424, 490, 56 },
		{ 424, 491, 56 },
		{ 424, 492, 56 },
		{ 424, 493, 56 },
		{ 424, 494, 56 },
		{ 424, 495, 56 },
		{ 424, 496, 56 },
		{ 424, 497, 56 },
		{ 424, 498, 56 },
		{ 424, 499, 56 },
		{ 424, 500, 56 },
		{ 424, 501, 56 },
		{ 424, 502, 56 },
		{ 424, 503, 56 },
		{ 424, 504, 56 },
		{ 424, 505, 56 },
		{ 424, 506, 56 },
		{ 424, 507, 56 },
		{ 424, 508, 56 },
		{ 424, 509, 56 },
		{ 424, 510, 56 },
		{ 424, 511, 56 },
		{ 424, 512, 56 },
		{ 425, 450, 56 },
		{ 425, 451, 56 },
		{ 425, 452, 56 },
		{ 425, 453, 56 },
		{ 425, 454, 56 },
		{ 425, 455, 56 },
		{ 425, 456, 56 },
		{ 425, 457, 56 },
		{ 425, 458, 56 },
		{ 425, 459, 56 },
		{ 425, 460, 56 },
		{ 425, 461, 56 },
		{ 425, 462, 56 },
		{ 425, 463, 56 },
		{ 425, 464, 56 },
		{ 425, 465, 56 },
		{ 425, 466, 56 },
		{ 425, 467, 56 },
		{ 425, 468, 56 },
		{ 425, 469, 56 },
		{ 425, 470, 56 },
		{ 425, 471, 56 },
		{ 425, 472, 56 },
		{ 425, 473, 56 },
		{ 425, 474, 56 },
		{ 425, 475, 56 },
		{ 425, 476, 56 },
		{ 425, 477, 56 },
		{ 425, 478, 56 },
		{ 425, 479, 56 },
		{ 425, 480, 56 },
		{ 425, 481, 56 },
		{ 425, 482, 56 },
		{ 425, 483, 56 },
		{ 425, 484, 56 },
		{ 425, 485, 56 },
		{ 425, 486, 56 },
		{ 425, 487, 56 },
		{ 425, 488, 56 },
		{ 425, 489, 56 },
		{ 425, 490, 56 },
		{ 425, 491, 56 },
		{ 425, 492, 56 },
		{ 425, 493, 56 },
		{ 425, 494, 56 },
		{ 425, 495, 56 },
		{ 425, 496, 56 },
		{ 425, 497, 56 },
		{ 425, 498, 56 },
		{ 425, 499, 56 },
		{ 425, 500, 56 },
		{ 425, 501, 56 },
		{ 425, 502, 56 },
		{ 425, 503, 56 },
		{ 425, 504, 56 },
		{ 425, 505, 56 },
		{ 425, 506, 56 },
		{ 425, 507, 56 },
		{ 425, 508, 56 },
		{ 425, 509, 56 },
		{ 425, 510, 56 },
		{ 425, 511, 56 },
		{ 425, 512, 56 },
		{ 426, 451, 56 },
		{ 426, 452, 56 },
		{ 426, 453, 56 },
		{ 426, 454, 56 },
		{ 426, 455, 56 },
		{ 426, 456, 56 },
		{ 426, 457, 56 },
		{ 426, 458, 56 },
		{ 426, 459, 56 },
		{ 426, 460, 56 },
		{ 426, 461, 56 },
		{ 426, 462, 56 },
		{ 426, 463, 56 },
		{ 426, 464, 56 },
		{ 426, 465, 56 },
		{ 426, 466, 56 },
		{ 426, 467, 56 },
		{ 426, 468, 56 },
		{ 426, 469, 56 },
		{ 426, 470, 56 },
		{ 426, 471, 56 },
		{ 426, 472, 56 },
		{ 426, 473, 56 },
		{ 426, 474, 56 },
		{ 426, 475, 56 },
		{ 426, 476, 56 },
		{ 426, 477, 56 },
		{ 426, 478, 56 },
		{ 426, 479, 56 },
		{ 426, 480, 56 },
		{ 426, 481, 56 },
		{ 426, 482, 56 },
		{ 426, 483, 56 },
		{ 426, 484, 56 },
		{ 426, 485, 56 },
		{ 426, 486, 56 },
		{ 426, 487, 56 },
		{ 426, 488, 56 },
		{ 426, 489, 56 },
		{ 426, 490, 56 },
		{ 426, 491, 56 },
		{ 426, 492, 56 },
		{ 426, 493, 56 },
		{ 426, 494, 56 },
		{ 426, 495, 56 },
		{ 426, 496, 56 },
		{ 426, 497, 56 },
		{ 426, 498, 56 },
		{ 426, 499, 56 },
		{ 426, 500, 56 },
		{ 426, 501, 56 },
		{ 426, 502, 56 },
		{ 426, 503, 56 },
		{ 426, 504, 56 },
		{ 426, 505, 56 },
		{ 426, 506, 56 },
		{ 426, 507, 56 },
		{ 426, 508, 56 },
		{ 426, 509, 56 },
		{ 426, 510, 56 },
		{ 426, 511, 56 },
		{ 426, 512, 56 },
		{ 427, 452, 56 },
		{ 427, 453, 56 },
		{ 427, 454, 56 },
		{ 427, 455, 56 },
		{ 427, 456, 56 },
		{ 427, 457, 56 },
		{ 427, 458, 56 },
		{ 427, 459, 56 },
		{ 427, 460, 56 },
		{ 427, 461, 56 },
		{ 427, 462, 56 },
		{ 427, 463, 56 },
		{ 427, 464, 56 },
		{ 427, 465, 56 },
		{ 427, 466, 56 },
		{ 427, 467, 56 },
		{ 427, 468, 56 },
		{ 427, 469, 56 },
		{ 427, 470, 56 },
		{ 427, 471, 56 },
		{ 427, 472, 56 },
		{ 427, 473, 56 },
		{ 427, 474, 56 },
		{ 427, 475, 56 },
		{ 427, 476, 56 },
		{ 427, 477, 56 },
		{ 427, 478, 56 },
		{ 427, 479, 56 },
		{ 427, 480, 56 },
		{ 427, 481, 56 },
		{ 427, 482, 56 },
		{ 427, 483, 56 },
		{ 427, 484, 56 },
		{ 427, 485, 56 },
		{ 427, 486, 56 },
		{ 427, 487, 56 },
		{ 427, 488, 56 },
		{ 427, 489, 56 },
		{ 427, 490, 56 },
		{ 427, 491, 56 },
		{ 427, 492, 56 },
		{ 427, 493, 56 },
		{ 427, 494, 56 },
		{ 427, 495, 56 },
		{ 427, 496, 56 },
		{ 427, 497, 56 },
		{ 427, 498, 56 },
		{ 427, 499, 56 },
		{ 427, 500, 56 },
		{ 427, 501, 56 },
		{ 427, 502, 56 },
		{ 427, 503, 56 },
		{ 427, 504, 56 },
		{ 427, 505, 56 },
		{ 427, 506, 56 },
		{ 427, 507, 56 },
		{ 427, 508, 56 },
		{ 427, 509, 56 },
		{ 427, 510, 56 },
		{ 427, 511, 56 },
		{ 427, 512, 56 },
		{ 428, 453, 56 },
		{ 428, 454, 56 },
		{ 428, 455, 56 },
		{ 428, 456, 56 },
		{ 428, 457, 56 },
		{ 428, 458, 56 },
		{ 428, 459, 56 },
		{ 428, 460, 56 },
		{ 428, 461, 56 },
		{ 428, 462, 56 },
		{ 428, 463, 56 },
		{ 428, 464, 56 },
		{ 428, 465, 56 },
		{ 428, 466, 56 },
		{ 428, 467, 56 },
		{ 428, 468, 56 },
		{ 428, 469, 56 },
		{ 428, 470, 56 },
		{ 428, 471, 56 },
		{ 428, 472, 56 },
		{ 428, 473, 56 },
		{ 428, 474, 56 },
		{ 428, 475, 56 },
		{ 428, 476, 56 },
		{ 428, 477, 56 },
		{ 428, 478, 56 },
		{ 428, 479, 56 },
		{ 428, 480, 56 },
		{ 428, 481, 56 },
		{ 428, 482, 56 },
		{ 428, 483, 56 },
		{ 428, 484, 56 },
		{ 428, 485, 56 },
		{ 428, 486, 56 },
		{ 428, 487, 56 },
		{ 428, 488, 56 },
		{ 428, 489, 56 },
		{ 428, 490, 56 },
		{ 428, 491, 56 },
		{ 428, 492, 56 },
		{ 428, 493, 56 },
		{ 428, 494, 56 },
		{ 428, 495, 56 },
		{ 428, 496, 56 },
		{ 428, 497, 56 },
		{ 428, 498, 56 },
		{ 428, 499, 56 },
		{ 428, 500, 56 },
		{ 428, 501, 56 },
		{ 428, 502, 56 },
		{ 428, 503, 56 },
		{ 428, 504, 56 },
		{ 428, 505, 56 },
		{ 428, 506, 56 },
		{ 428, 507, 56 },
		{ 428, 508, 56 },
		{ 428, 509, 56 },
		{ 428, 510, 56 },
		{ 428, 511, 56 },
		{ 428, 512, 56 },
		{ 429, 454, 56 },
		{ 429, 455, 56 },
		{ 429, 456, 56 },
		{ 429, 457, 56 },
		{ 429, 458, 56 },
		{ 429, 459, 56 },
		{ 429, 460, 56 },
		{ 429, 461, 56 },
		{ 429, 462, 56 },
		{ 429, 463, 56 },
		{ 429, 464, 56 },
		{ 429, 465, 56 },
		{ 429, 466, 56 },
		{ 429, 467, 56 },
		{ 429, 468, 56 },
		{ 429, 469, 56 },
		{ 429, 470, 56 },
		{ 429, 471, 56 },
		{ 429, 472, 56 },
		{ 429, 473, 56 },
		{ 429, 474, 56 },
		{ 429, 475, 56 },
		{ 429, 476, 56 },
		{ 429, 477, 56 },
		{ 429, 478, 56 },
		{ 429, 479, 56 },
		{ 429, 480, 56 },
		{ 429, 481, 56 },
		{ 429, 482, 56 },
		{ 429, 483, 56 },
		{ 429, 484, 56 },
		{ 429, 485, 56 },
		{ 429, 486, 56 },
		{ 429, 487, 56 },
		{ 429, 488, 56 },
		{ 429, 489, 56 },
		{ 429, 490, 56 },
		{ 429, 491, 56 },
		{ 429, 492, 56 },
		{ 429, 493, 56 },
		{ 429, 494, 56 },
		{ 429, 495, 56 },
		{ 429, 496, 56 },
		{ 429, 497, 56 },
		{ 429, 498, 56 },
		{ 429, 499, 56 },
		{ 429, 500, 56 },
		{ 429, 501, 56 },
		{ 429, 502, 56 },
		{ 429, 503, 56 },
		{ 429, 504, 56 },
		{ 429, 505, 56 },
		{ 429, 506, 56 },
		{ 429, 507, 56 },
		{ 429, 508, 56 },
		{ 429, 509, 56 },
		{ 429, 510, 56 },
		{ 429, 511, 56 },
		{ 429, 512, 56 },
		{ 430, 455, 56 },
		{ 430, 456, 56 },
		{ 430, 457, 56 },
		{ 430, 458, 56 },
		{ 430, 459, 56 },
		{ 430, 460, 56 },
		{ 430, 461, 56 },
		{ 430, 462, 56 },
		{ 430, 463, 56 },
		{ 430, 464, 56 },
		{ 430, 465, 56 },
		{ 430, 466, 56 },
		{ 430, 467, 56 },
		{ 430, 468, 56 },
		{ 430, 469, 56 },
		{ 430, 470, 56 },
		{ 430, 471, 56 },
		{ 430, 472, 56 },
		{ 430, 473, 56 },
		{ 430, 474, 56 },
		{ 430, 475, 56 },
		{ 430, 476, 56 },
		{ 430, 477, 56 },
		{ 430, 478, 56 },
		{ 430, 479, 56 },
		{ 430, 480, 56 },
		{ 430, 481, 56 },
		{ 430, 482, 56 },
		{ 430, 483, 56 },
		{ 430, 484, 56 },
		{ 430, 485, 56 },
		{ 430, 486, 56 },
		{ 430, 487, 56 },
		{ 430, 488, 56 },
		{ 430, 489, 56 },
		{ 430, 490, 56 },
		{ 430, 491, 56 },
		{ 430, 492, 56 },
		{ 430, 493, 56 },
		{ 430, 494, 56 },
		{ 430, 495, 56 },
		{ 430, 496, 56 },
		{ 430, 497, 56 },
		{ 430, 498, 56 },
		{ 430, 499, 56 },
		{ 430, 500, 56 },
		{ 430, 501, 56 },
		{ 430, 502, 56 },
		{ 430, 503, 56 },
		{ 430, 504, 56 },
		{ 430, 505, 56 },
		{ 430, 506, 56 },
		{ 430, 507, 56 },
		{ 430, 508, 56 },
		{ 430, 509, 56 },
		{ 430, 510, 56 },
		{ 430, 511, 56 },
		{ 430, 512, 56 },
		{ 431, 456, 56 },
		{ 431, 457, 56 },
		{ 431, 458, 56 },
		{ 431, 459, 56 },
		{ 431, 460, 56 },
		{ 431, 461, 56 },
		{ 431, 462, 56 },
		{ 431, 463, 56 },
		{ 431, 464, 56 },
		{ 431, 465, 56 },
		{ 431, 466, 56 },
		{ 431, 467, 56 },
		{ 431, 468, 56 },
		{ 431, 469, 56 },
		{ 431, 470, 56 },
		{ 431, 471, 56 },
		{ 431, 472, 56 },
		{ 431, 473, 56 },
		{ 431, 474, 56 },
		{ 431, 475, 56 },
		{ 431, 476, 56 },
		{ 431, 477, 56 },
		{ 431, 478, 56 },
		{ 431, 479, 56 },
		{ 431, 480, 56 },
		{ 431, 481, 56 },
		{ 431, 482, 56 },
		{ 431, 483, 56 },
		{ 431, 484, 56 },
		{ 431, 485, 56 },
		{ 431, 486, 56 },
		{ 431, 487, 56 },
		{ 431, 488, 56 },
		{ 431, 489, 56 },
		{ 431, 490, 56 },
		{ 431, 491, 56 },
		{ 431, 492, 56 },
		{ 431, 493, 56 },
		{ 431, 494, 56 },
		{ 431, 495, 56 },
		{ 431, 496, 56 },
		{ 431, 497, 56 },
		{ 431, 498, 56 },
		{ 431, 499, 56 },
		{ 431, 500, 56 },
		{ 431, 501, 56 },
		{ 431, 502, 56 },
		{ 431, 503, 56 },
		{ 431, 504, 56 },
		{ 431, 505, 56 },
		{ 431, 506, 56 },
		{ 431, 507, 56 },
		{ 431, 508, 56 },
		{ 431, 509, 56 },
		{ 431, 510, 56 },
		{ 431, 511, 56 },
		{ 431, 512, 56 },
		{ 432, 458, 56 },
		{ 432, 459, 56 },
		{ 432, 460, 56 },
		{ 432, 461, 56 },
		{ 432, 462, 56 },
		{ 432, 463, 56 },
		{ 432, 464, 56 },
		{ 432, 465, 56 },
		{ 432, 466, 56 },
		{ 432, 467, 56 },
		{ 432, 468, 56 },
		{ 432, 469, 56 },
		{ 432, 470, 56 },
		{ 432, 471, 56 },
		{ 432, 472, 56 },
		{ 432, 473, 56 },
		{ 432, 474, 56 },
		{ 432, 475, 56 },
		{ 432, 476, 56 },
		{ 432, 477, 56 },
		{ 432, 478, 56 },
		{ 432, 479, 56 },
		{ 432, 480, 56 },
		{ 432, 481, 56 },
		{ 432, 482, 56 },
		{ 432, 483, 56 },
		{ 432, 484, 56 },
		{ 432, 485, 56 },
		{ 432, 486, 56 },
		{ 432, 487, 56 },
		{ 432, 488, 56 },
		{ 432, 489, 56 },
		{ 432, 490, 56 },
		{ 432, 491, 56 },
		{ 432, 492, 56 },
		{ 432, 493, 56 },
		{ 432, 494, 56 },
		{ 432, 495, 56 },
		{ 432, 496, 56 },
		{ 432, 497, 56 },
		{ 432, 498, 56 },
		{ 432, 499, 56 },
		{ 432, 500, 56 },
		{ 432, 501, 56 },
		{ 432, 502, 56 },
		{ 432, 503, 56 },
		{ 432, 504, 56 },
		{ 432, 505, 56 },
		{ 432, 506, 56 },
		{ 432, 507, 56 },
		{ 432, 508, 56 },
		{ 432, 509, 56 },
		{ 432, 510, 56 },
		{ 432, 511, 56 },
		{ 432, 512, 56 },
		{ 433, 459, 56 },
		{ 433, 460, 56 },
		{ 433, 461, 56 },
		{ 433, 462, 56 },
		{ 433, 463, 56 },
		{ 433, 464, 56 },
		{ 433, 465, 56 },
		{ 433, 466, 56 },
		{ 433, 467, 56 },
		{ 433, 468, 56 },
		{ 433, 469, 56 },
		{ 433, 470, 56 },
		{ 433, 471, 56 },
		{ 433, 472, 56 },
		{ 433, 473, 56 },
		{ 433, 474, 56 },
		{ 433, 475, 56 },
		{ 433, 476, 56 },
		{ 433, 477, 56 },
		{ 433, 478, 56 },
		{ 433, 479, 56 },
		{ 433, 480, 56 },
		{ 433, 481, 56 },
		{ 433, 482, 56 },
		{ 433, 483, 56 },
		{ 433, 484, 56 },
		{ 433, 485, 56 },
		{ 433, 486, 56 },
		{ 433, 487, 56 },
		{ 433, 488, 56 },
		{ 433, 489, 56 },
		{ 433, 490, 56 },
		{ 433, 491, 56 },
		{ 433, 492, 56 },
		{ 433, 493, 56 },
		{ 433, 494, 56 },
		{ 433, 495, 56 },
		{ 433, 496, 56 },
		{ 433, 497, 56 },
		{ 433, 498, 56 },
		{ 433, 499, 56 },
		{ 433, 500, 56 },
		{ 433, 501, 56 },
		{ 433, 502, 56 },
		{ 433, 503, 56 },
		{ 433, 504, 56 },
		{ 433, 505, 56 },
		{ 433, 506, 56 },
		{ 433, 507, 56 },
		{ 433, 508, 56 },
		{ 433, 509, 56 },
		{ 433, 510, 56 },
		{ 433, 511, 56 },
		{ 433, 512, 56 },
		{ 434, 460, 56 },
		{ 434, 461, 56 },
		{ 434, 462, 56 },
		{ 434, 463, 56 },
		{ 434, 464, 56 },
		{ 434, 465, 56 },
		{ 434, 466, 56 },
		{ 434, 467, 56 },
		{ 434, 468, 56 },
		{ 434, 469, 56 },
		{ 434, 470, 56 },
		{ 434, 471, 56 },
		{ 434, 472, 56 },
		{ 434, 473, 56 },
		{ 434, 474, 56 },
		{ 434, 475, 56 },
		{ 434, 476, 56 },
		{ 434, 477, 56 },
		{ 434, 478, 56 },
		{ 434, 479, 56 },
		{ 434, 480, 56 },
		{ 434, 481, 56 },
		{ 434, 482, 56 },
		{ 434, 483, 56 },
		{ 434, 484, 56 },
		{ 434, 485, 56 },
		{ 434, 486, 56 },
		{ 434, 487, 56 },
		{ 434, 488, 56 },
		{ 434, 489, 56 },
		{ 434, 490, 56 },
		{ 434, 491, 56 },
		{ 434, 492, 56 },
		{ 434, 493, 56 },
		{ 434, 494, 56 },
		{ 434, 495, 56 },
		{ 434, 496, 56 },
		{ 434, 497, 56 },
		{ 434, 498, 56 },
		{ 434, 499, 56 },
		{ 434, 500, 56 },
		{ 434, 501, 56 },
		{ 434, 502, 56 },
		{ 434, 503, 56 },
		{ 434, 504, 56 },
		{ 434, 505, 56 },
		{ 434, 506, 56 },
		{ 434, 507, 56 },
		{ 434, 508, 56 },
		{ 434, 509, 56 },
		{ 434, 510, 56 },
		{ 434, 511, 56 },
		{ 434, 512, 56 },
		{ 435, 461, 56 },
		{ 435, 462, 56 },
		{ 435, 463, 56 },
		{ 435, 464, 56 },
		{ 435, 465, 56 },
		{ 435, 466, 56 },
		{ 435, 467, 56 },
		{ 435, 468, 56 },
		{ 435, 469, 56 },
		{ 435, 470, 56 },
		{ 435, 471, 56 },
		{ 435, 472, 56 },
		{ 435, 473, 56 },
		{ 435, 474, 56 },
		{ 435, 475, 56 },
		{ 435, 476, 56 },
		{ 435, 477, 56 },
		{ 435, 478, 56 },
		{ 435, 479, 56 },
		{ 435, 480, 56 },
		{ 435, 481, 56 },
		{ 435, 482, 56 },
		{ 435, 483, 56 },
		{ 435, 484, 56 },
		{ 435, 485, 56 },
		{ 435, 486, 56 },
		{ 435, 487, 56 },
		{ 435, 488, 56 },
		{ 435, 489, 56 },
		{ 435, 490, 56 },
		{ 435, 491, 56 },
		{ 435, 492, 56 },
		{ 435, 493, 56 },
		{ 435, 494, 56 },
		{ 435, 495, 56 },
		{ 435, 496, 56 },
		{ 435, 497, 56 },
		{ 435, 498, 56 },
		{ 435, 499, 56 },
		{ 435, 500, 56 },
		{ 435, 501, 56 },
		{ 435, 502, 56 },
		{ 435, 503, 56 },
		{ 435, 504, 56 },
		{ 435, 505, 56 },
		{ 435, 506, 56 },
		{ 435, 507, 56 },
		{ 435, 508, 56 },
		{ 435, 509, 56 },
		{ 435, 510, 56 },
		{ 435, 511, 56 },
		{ 435, 512, 56 },
		{ 436, 462, 56 },
		{ 436, 463, 56 },
		{ 436, 464, 56 },
		{ 436, 465, 56 },
		{ 436, 466, 56 },
		{ 436, 467, 56 },
		{ 436, 468, 56 },
		{ 436, 469, 56 },
		{ 436, 470, 56 },
		{ 436, 471, 56 },
		{ 436, 472, 56 },
		{ 436, 473, 56 },
		{ 436, 474, 56 },
		{ 436, 475, 56 },
		{ 436, 476, 56 },
		{ 436, 477, 56 },
		{ 436, 478, 56 },
		{ 436, 479, 56 },
		{ 436, 480, 56 },
		{ 436, 481, 56 },
		{ 436, 482, 56 },
		{ 436, 483, 56 },
		{ 436, 484, 56 },
		{ 436, 485, 56 },
		{ 436, 486, 56 },
		{ 436, 487, 56 },
		{ 436, 488, 56 },
		{ 436, 489, 56 },
		{ 436, 490, 56 },
		{ 436, 491, 56 },
		{ 436, 492, 56 },
		{ 436, 493, 56 },
		{ 436, 494, 56 },
		{ 436, 495, 56 },
		{ 436, 496, 56 },
		{ 436, 497, 56 },
		{ 436, 498, 56 },
		{ 436, 499, 56 },
		{ 436, 500, 56 },
		{ 436, 501, 56 },
		{ 436, 502, 56 },
		{ 436, 503, 56 },
		{ 436, 504, 56 },
		{ 436, 505, 56 },
		{ 436, 506, 56 },
		{ 436, 507, 56 },
		{ 436, 508, 56 },
		{ 436, 509, 56 },
		{ 436, 510, 56 },
		{ 436, 511, 56 },
		{ 436, 512, 56 },
		{ 437, 463, 56 },
		{ 437, 464, 56 },
		{ 437, 465, 56 },
		{ 437, 466, 56 },
		{ 437, 467, 56 },
		{ 437, 468, 56 },
		{ 437, 469, 56 },
		{ 437, 470, 56 },
		{ 437, 471, 56 },
		{ 437, 472, 56 },
		{ 437, 473, 56 },
		{ 437, 474, 56 },
		{ 437, 475, 56 },
		{ 437, 476, 56 },
		{ 437, 477, 56 },
		{ 437, 478, 56 },
		{ 437, 479, 56 },
		{ 437, 480, 56 },
		{ 437, 481, 56 },
		{ 437, 482, 56 },
		{ 437, 483, 56 },
		{ 437, 484, 56 },
		{ 437, 485, 56 },
		{ 437, 486, 56 },
		{ 437, 487, 56 },
		{ 437, 488, 56 },
		{ 437, 489, 56 },
		{ 437, 490, 56 },
		{ 437, 491, 56 },
		{ 437, 492, 56 },
		{ 437, 493, 56 },
		{ 437, 494, 56 },
		{ 437, 495, 56 },
		{ 437, 496, 56 },
		{ 437, 497, 56 },
		{ 437, 498, 56 },
		{ 437, 499, 56 },
		{ 437, 500, 56 },
		{ 437, 501, 56 },
		{ 437, 502, 56 },
		{ 437, 503, 56 },
		{ 437, 504, 56 },
		{ 437, 505, 56 },
		{ 437, 506, 56 },
		{ 437, 507, 56 },
		{ 437, 508, 56 },
		{ 437, 509, 56 },
		{ 437, 510, 56 },
		{ 437, 511, 56 },
		{ 437, 512, 56 },
		{ 438, 464, 56 },
		{ 438, 465, 56 },
		{ 438, 466, 56 },
		{ 438, 467, 56 },
		{ 438, 468, 56 },
		{ 438, 469, 56 },
		{ 438, 470, 56 },
		{ 438, 471, 56 },
		{ 438, 472, 56 },
		{ 438, 473, 56 },
		{ 438, 474, 56 },
		{ 438, 475, 56 },
		{ 438, 476, 56 },
		{ 438, 477, 56 },
		{ 438, 478, 56 },
		{ 438, 479, 56 },
		{ 438, 480, 56 },
		{ 438, 481, 56 },
		{ 438, 482, 56 },
		{ 438, 483, 56 },
		{ 438, 484, 56 },
		{ 438, 485, 56 },
		{ 438, 486, 56 },
		{ 438, 487, 56 },
		{ 438, 488, 56 },
		{ 438, 489, 56 },
		{ 438, 490, 56 },
		{ 438, 491, 56 },
		{ 438, 492, 56 },
		{ 438, 493, 56 },
		{ 438, 494, 56 },
		{ 438, 495, 56 },
		{ 438, 496, 56 },
		{ 438, 497, 56 },
		{ 438, 498, 56 },
		{ 438, 499, 56 },
		{ 438, 500, 56 },
		{ 438, 501, 56 },
		{ 438, 502, 56 },
		{ 438, 503, 56 },
		{ 438, 504, 56 },
		{ 438, 505, 56 },
		{ 438, 506, 56 },
		{ 438, 507, 56 },
		{ 438, 508, 56 },
		{ 438, 509, 56 },
		{ 438, 510, 56 },
		{ 438, 511, 56 },
		{ 438, 512, 56 },
		{ 439, 466, 56 },
		{ 439, 467, 56 },
		{ 439, 468, 56 },
		{ 439, 469, 56 },
		{ 439, 470, 56 },
		{ 439, 471, 56 },
		{ 439, 472, 56 },
		{ 439, 473, 56 },
		{ 439, 474, 56 },
		{ 439, 475, 56 },
		{ 439, 476, 56 },
		{ 439, 477, 56 },
		{ 439, 478, 56 },
		{ 439, 479, 56 },
		{ 439, 480, 56 },
		{ 439, 481, 56 },
		{ 439, 482, 56 },
		{ 439, 483, 56 },
		{ 439, 484, 56 },
		{ 439, 485, 56 },
		{ 439, 486, 56 },
		{ 439, 487, 56 },
		{ 439, 488, 56 },
		{ 439, 489, 56 },
		{ 439, 490, 56 },
		{ 439, 491, 56 },
		{ 439, 492, 56 },
		{ 439, 493, 56 },
		{ 439, 494, 56 },
		{ 439, 495, 56 },
		{ 439, 496, 56 },
		{ 439, 497, 56 },
		{ 439, 498, 56 },
		{ 439, 499, 56 },
		{ 439, 500, 56 },
		{ 439, 501, 56 },
		{ 439, 502, 56 },
		{ 439, 503, 56 },
		{ 439, 504, 56 },
		{ 439, 505, 56 },
		{ 439, 506, 56 },
		{ 439, 507, 56 },
		{ 439, 508, 56 },
		{ 439, 509, 56 },
		{ 439, 510, 56 },
		{ 439, 511, 56 },
		{ 439, 512, 56 },
		{ 440, 467, 56 },
		{ 440, 468, 56 },
		{ 440, 469, 56 },
		{ 440, 470, 56 },
		{ 440, 471, 56 },
		{ 440, 472, 56 },
		{ 440, 473, 56 },
		{ 440, 474, 56 },
		{ 440, 475, 56 },
		{ 440, 476, 56 },
		{ 440, 477, 56 },
		{ 440, 478, 56 },
		{ 440, 479, 56 },
		{ 440, 480, 56 },
		{ 440, 481, 56 },
		{ 440, 482, 56 },
		{ 440, 483, 56 },
		{ 440, 484, 56 },
		{ 440, 485, 56 },
		{ 440, 486, 56 },
		{ 440, 487, 56 },
		{ 440, 488, 56 },
		{ 440, 489, 56 },
		{ 440, 490, 56 },
		{ 440, 491, 56 },
		{ 440, 492, 56 },
		{ 440, 493, 56 },
		{ 440, 494, 56 },
		{ 440, 495, 56 },
		{ 440, 496, 56 },
		{ 440, 497, 56 },
		{ 440, 498, 56 },
		{ 440, 499, 56 },
		{ 440, 500, 56 },
		{ 440, 501, 56 },
		{ 440, 502, 56 },
		{ 440, 503, 56 },
		{ 440, 504, 56 },
		{ 440, 505, 56 },
		{ 440, 506, 56 },
		{ 440, 507, 56 },
		{ 440, 508, 56 },
		{ 440, 509, 56 },
		{ 440, 510, 56 },
		{ 440, 511, 56 },
		{ 440, 512, 56 },
		{ 441, 468, 56 },
		{ 441, 469, 56 },
		{ 441, 470, 56 },
		{ 441, 471, 56 },
		{ 441, 472, 56 },
		{ 441, 473, 56 },
		{ 441, 474, 56 },
		{ 441, 475, 56 },
		{ 441, 476, 56 },
		{ 441, 477, 56 },
		{ 441, 478, 56 },
		{ 441, 479, 56 },
		{ 441, 480, 56 },
		{ 441, 481, 56 },
		{ 441, 482, 56 },
		{ 441, 483, 56 },
		{ 441, 484, 56 },
		{ 441, 485, 56 },
		{ 441, 486, 56 },
		{ 441, 487, 56 },
		{ 441, 488, 56 },
		{ 441, 489, 56 },
		{ 441, 490, 56 },
		{ 441, 491, 56 },
		{ 441, 492, 56 },
		{ 441, 493, 56 },
		{ 441, 494, 56 },
		{ 441, 495, 56 },
		{ 441, 496, 56 },
		{ 441, 497, 56 },
		{ 441, 498, 56 },
		{ 441, 499, 56 },
		{ 441, 500, 56 },
		{ 441, 501, 56 },
		{ 441, 502, 56 },
		{ 441, 503, 56 },
		{ 441, 504, 56 },
		{ 441, 505, 56 },
		{ 441, 506, 56 },
		{ 441, 507, 56 },
		{ 441, 508, 56 },
		{ 441, 509, 56 },
		{ 441, 510, 56 },
		{ 441, 511, 56 },
		{ 441, 512, 56 },
		{ 442, 469, 56 },
		{ 442, 470, 56 },
		{ 442, 471, 56 },
		{ 442, 472, 56 },
		{ 442, 473, 56 },
		{ 442, 474, 56 },
		{ 442, 475, 56 },
		{ 442, 476, 56 },
		{ 442, 477, 56 },
		{ 442, 478, 56 },
		{ 442, 479, 56 },
		{ 442, 480, 56 },
		{ 442, 481, 56 },
		{ 442, 482, 56 },
		{ 442, 483, 56 },
		{ 442, 484, 56 },
		{ 442, 485, 56 },
		{ 442, 486, 56 },
		{ 442, 487, 56 },
		{ 442, 488, 56 },
		{ 442, 489, 56 },
		{ 442, 490, 56 },
		{ 442, 491, 56 },
		{ 442, 492, 56 },
		{ 442, 493, 56 },
		{ 442, 494, 56 },
		{ 442, 495, 56 },
		{ 442, 496, 56 },
		{ 442, 497, 56 },
		{ 442, 498, 56 },
		{ 442, 499, 56 },
		{ 442, 500, 56 },
		{ 442, 501, 56 },
		{ 442, 502, 56 },
		{ 442, 503, 56 },
		{ 442, 504, 56 },
		{ 442, 505, 56 },
		{ 442, 506, 56 },
		{ 442, 507, 56 },
		{ 442, 508, 56 },
		{ 442, 509, 56 },
		{ 442, 510, 56 },
		{ 442, 511, 56 },
		{ 442, 512, 56 },
		{ 443, 470, 56 },
		{ 443, 471, 56 },
		{ 443, 472, 56 },
		{ 443, 473, 56 },
		{ 443, 474, 56 },
		{ 443, 475, 56 },
		{ 443, 476, 56 },
		{ 443, 477, 56 },
		{ 443, 478, 56 },
		{ 443, 479, 56 },
		{ 443, 480, 56 },
		{ 443, 481, 56 },
		{ 443, 482, 56 },
		{ 443, 483, 56 },
		{ 443, 484, 56 },
		{ 443, 485, 56 },
		{ 443, 486, 56 },
		{ 443, 487, 56 },
		{ 443, 488, 56 },
		{ 443, 489, 56 },
		{ 443, 490, 56 },
		{ 443, 491, 56 },
		{ 443, 492, 56 },
		{ 443, 493, 56 },
		{ 443, 494, 56 },
		{ 443, 495, 56 },
		{ 443, 496, 56 },
		{ 443, 497, 56 },
		{ 443, 498, 56 },
		{ 443, 499, 56 },
		{ 443, 500, 56 },
		{ 443, 501, 56 },
		{ 443, 502, 56 },
		{ 443, 503, 56 },
		{ 443, 504, 56 },
		{ 443, 505, 56 },
		{ 443, 506, 56 },
		{ 443, 507, 56 },
		{ 443, 508, 56 },
		{ 443, 509, 56 },
		{ 443, 510, 56 },
		{ 443, 511, 56 },
		{ 443, 512, 56 },
		{ 444, 471, 56 },
		{ 444, 472, 56 },
		{ 444, 473, 56 },
		{ 444, 474, 56 },
		{ 444, 475, 56 },
		{ 444, 476, 56 },
		{ 444, 477, 56 },
		{ 444, 478, 56 },
		{ 444, 479, 56 },
		{ 444, 480, 56 },
		{ 444, 481, 56 },
		{ 444, 482, 56 },
		{ 444, 483, 56 },
		{ 444, 484, 56 },
		{ 444, 485, 56 },
		{ 444, 486, 56 },
		{ 444, 487, 56 },
		{ 444, 488, 56 },
		{ 444, 489, 56 },
		{ 444, 490, 56 },
		{ 444, 491, 56 },
		{ 444, 492, 56 },
		{ 444, 493, 56 },
		{ 444, 494, 56 },
		{ 444, 495, 56 },
		{ 444, 496, 56 },
		{ 444, 497, 56 },
		{ 444, 498, 56 },
		{ 444, 499, 56 },
		{ 444, 500, 56 },
		{ 444, 501, 56 },
		{ 444, 502, 56 },
		{ 444, 503, 56 },
		{ 444, 504, 56 },
		{ 444, 505, 56 },
		{ 444, 506, 56 },
		{ 444, 507, 56 },
		{ 444, 508, 56 },
		{ 444, 509, 56 },
		{ 444, 510, 56 },
		{ 444, 511, 56 },
		{ 444, 512, 56 },
		{ 445, 472, 56 },
		{ 445, 473, 56 },
		{ 445, 474, 56 },
		{ 445, 475, 56 },
		{ 445, 476, 56 },
		{ 445, 477, 56 },
		{ 445, 478, 56 },
		{ 445, 479, 56 },
		{ 445, 480, 56 },
		{ 445, 481, 56 },
		{ 445, 482, 56 },
		{ 445, 483, 56 },
		{ 445, 484, 56 },
		{ 445, 485, 56 },
		{ 445, 486, 56 },
		{ 445, 487, 56 },
		{ 445, 488, 56 },
		{ 445, 489, 56 },
		{ 445, 490, 56 },
		{ 445, 491, 56 },
		{ 445, 492, 56 },
		{ 445, 493, 56 },
		{ 445, 494, 56 },
		{ 445, 495, 56 },
		{ 445, 496, 56 },
		{ 445, 497, 56 },
		{ 445, 498, 56 },
		{ 445, 499, 56 },
		{ 445, 500, 56 },
		{ 445, 501, 56 },
		{ 445, 502, 56 },
		{ 445, 503, 56 },
		{ 445, 504, 56 },
		{ 445, 505, 56 },
		{ 445, 506, 56 },
		{ 445, 507, 56 },
		{ 445, 508, 56 },
		{ 445, 509, 56 },
		{ 445, 510, 56 },
		{ 445, 511, 56 },
		{ 445, 512, 56 },
		{ 446, 474, 56 },
		{ 446, 475, 56 },
		{ 446, 476, 56 },
		{ 446, 477, 56 },
		{ 446, 478, 56 },
		{ 446, 479, 56 },
		{ 446, 480, 56 },
		{ 446, 481, 56 },
		{ 446, 482, 56 },
		{ 446, 483, 56 },
		{ 446, 484, 56 },
		{ 446, 485, 56 },
		{ 446, 486, 56 },
		{ 446, 487, 56 },
		{ 446, 488, 56 },
		{ 446, 489, 56 },
		{ 446, 490, 56 },
		{ 446, 491, 56 },
		{ 446, 492, 56 },
		{ 446, 493, 56 },
		{ 446, 494, 56 },
		{ 446, 495, 56 },
		{ 446, 496, 56 },
		{ 446, 497, 56 },
		{ 446, 498, 56 },
		{ 446, 499, 56 },
		{ 446, 500, 56 },
		{ 446, 501, 56 },
		{ 446, 502, 56 },
		{ 446, 503, 56 },
		{ 446, 504, 56 },
		{ 446, 505, 56 },
		{ 446, 506, 56 },
		{ 446, 507, 56 },
		{ 446, 508, 56 },
		{ 446, 509, 56 },
		{ 446, 510, 56 },
		{ 446, 511, 56 },
		{ 446, 512, 56 },
		{ 447, 475, 56 },
		{ 447, 476, 56 },
		{ 447, 477, 56 },
		{ 447, 478, 56 },
		{ 447, 479, 56 },
		{ 447, 480, 56 },
		{ 447, 481, 56 },
		{ 447, 482, 56 },
		{ 447, 483, 56 },
		{ 447, 484, 56 },
		{ 447, 485, 56 },
		{ 447, 486, 56 },
		{ 447, 487, 56 },
		{ 447, 488, 56 },
		{ 447, 489, 56 },
		{ 447, 490, 56 },
		{ 447, 491, 56 },
		{ 447, 492, 56 },
		{ 447, 493, 56 },
		{ 447, 494, 56 },
		{ 447, 495, 56 },
		{ 447, 496, 56 },
		{ 447, 497, 56 },
		{ 447, 498, 56 },
		{ 447, 499, 56 },
		{ 447, 500, 56 },
		{ 447, 501, 56 },
		{ 447, 502, 56 },
		{ 447, 503, 56 },
		{ 447, 504, 56 },
		{ 447, 505, 56 },
		{ 447, 506, 56 },
		{ 447, 507, 56 },
		{ 447, 508, 56 },
		{ 447, 509, 56 },
		{ 447, 510, 56 },
		{ 447, 511, 56 },
		{ 447, 512, 56 },
		{ 448, 476, 56 },
		{ 448, 477, 56 },
		{ 448, 478, 56 },
		{ 448, 479, 56 },
		{ 448, 480, 56 },
		{ 448, 481, 56 },
		{ 448, 482, 56 },
		{ 448, 483, 56 },
		{ 448, 484, 56 },
		{ 448, 485, 56 },
		{ 448, 486, 56 },
		{ 448, 487, 56 },
		{ 448, 488, 56 },
		{ 448, 489, 56 },
		{ 448, 490, 56 },
		{ 448, 491, 56 },
		{ 448, 492, 56 },
		{ 448, 493, 56 },
		{ 448, 494, 56 },
		{ 448, 495, 56 },
		{ 448, 496, 56 },
		{ 448, 497, 56 },
		{ 448, 498, 56 },
		{ 448, 499, 56 },
		{ 448, 500, 56 },
		{ 448, 501, 56 },
		{ 448, 502, 56 },
		{ 448, 503, 56 },
		{ 448, 504, 56 },
		{ 448, 505, 56 },
		{ 448, 506, 56 },
		{ 448, 507, 56 },
		{ 448, 508, 56 },
		{ 448, 509, 56 },
		{ 448, 510, 56 },
		{ 448, 511, 56 },
		{ 448, 512, 56 },
		{ 449, 477, 56 },
		{ 449, 478, 56 },
		{ 449, 479, 56 },
		{ 449, 480, 56 },
		{ 449, 481, 56 },
		{ 449, 482, 56 },
		{ 449, 483, 56 },
		{ 449, 484, 56 },
		{ 449, 485, 56 },
		{ 449, 486, 56 },
		{ 449, 487, 56 },
		{ 449, 488, 56 },
		{ 449, 489, 56 },
		{ 449, 490, 56 },
		{ 449, 491, 56 },
		{ 449, 492, 56 },
		{ 449, 493, 56 },
		{ 449, 494, 56 },
		{ 449, 495, 56 },
		{ 449, 496, 56 },
		{ 449, 497, 56 },
		{ 449, 498, 56 },
		{ 449, 499, 56 },
		{ 449, 500, 56 },
		{ 449, 501, 56 },
		{ 449, 502, 56 },
		{ 449, 503, 56 },
		{ 449, 504, 56 },
		{ 449, 505, 56 },
		{ 449, 506, 56 },
		{ 449, 507, 56 },
		{ 449, 508, 56 },
		{ 449, 509, 56 },
		{ 449, 510, 56 },
		{ 449, 511, 56 },
		{ 449, 512, 56 },
		{ 450, 478, 56 },
		{ 450, 479, 56 },
		{ 450, 480, 56 },
		{ 450, 481, 56 },
		{ 450, 482, 56 },
		{ 450, 483, 56 },
		{ 450, 484, 56 },
		{ 450, 485, 56 },
		{ 450, 486, 56 },
		{ 450, 487, 56 },
		{ 450, 488, 56 },
		{ 450, 489, 56 },
		{ 450, 490, 56 },
		{ 450, 491, 56 },
		{ 450, 492, 56 },
		{ 450, 493, 56 },
		{ 450, 494, 56 },
		{ 450, 495, 56 },
		{ 450, 496, 56 },
		{ 450, 497, 56 },
		{ 450, 498, 56 },
		{ 450, 499, 56 },
		{ 450, 500, 56 },
		{ 450, 501, 56 },
		{ 450, 502, 56 },
		{ 450, 503, 56 },
		{ 450, 504, 56 },
		{ 450, 505, 56 },
		{ 450, 506, 56 },
		{ 450, 507, 56 },
		{ 450, 508, 56 },
		{ 450, 509, 56 },
		{ 450, 510, 56 },
		{ 450, 511, 56 },
		{ 450, 512, 56 },
		{ 451, 479, 56 },
		{ 451, 480, 56 },
		{ 451, 481, 56 },
		{ 451, 482, 56 },
		{ 451, 483, 56 },
		{ 451, 484, 56 },
		{ 451, 485, 56 },
		{ 451, 486, 56 },
		{ 451, 487, 56 },
		{ 451, 488, 56 },
		{ 451, 489, 56 },
		{ 451, 490, 56 },
		{ 451, 491, 56 },
		{ 451, 492, 56 },
		{ 451, 493, 56 },
		{ 451, 494, 56 },
		{ 451, 495, 56 },
		{ 451, 496, 56 },
		{ 451, 497, 56 },
		{ 451, 498, 56 },
		{ 451, 499, 56 },
		{ 451, 500, 56 },
		{ 451, 501, 56 },
		{ 451, 502, 56 },
		{ 451, 503, 56 },
		{ 451, 504, 56 },
		{ 451, 505, 56 },
		{ 451, 506, 56 },
		{ 451, 507, 56 },
		{ 451, 508, 56 },
		{ 451, 509, 56 },
		{ 451, 510, 56 },
		{ 451, 511, 56 },
		{ 451, 512, 56 },
		{ 452, 480, 56 },
		{ 452, 481, 56 },
		{ 452, 482, 56 },
		{ 452, 483, 56 },
		{ 452, 484, 56 },
		{ 452, 485, 56 },
		{ 452, 486, 56 },
		{ 452, 487, 56 },
		{ 452, 488, 56 },
		{ 452, 489, 56 },
		{ 452, 490, 56 },
		{ 452, 491, 56 },
		{ 452, 492, 56 },
		{ 452, 493, 56 },
		{ 452, 494, 56 },
		{ 452, 495, 56 },
		{ 452, 496, 56 },
		{ 452, 497, 56 },
		{ 452, 498, 56 },
		{ 452, 499, 56 },
		{ 452, 500, 56 },
		{ 452, 501, 56 },
		{ 452, 502, 56 },
		{ 452, 503, 56 },
		{ 452, 504, 56 },
		{ 452, 505, 56 },
		{ 452, 506, 56 },
		{ 452, 507, 56 },
		{ 452, 508, 56 },
		{ 452, 509, 56 },
		{ 452, 510, 56 },
		{ 452, 511, 56 },
		{ 452, 512, 56 },
		{ 453, 482, 56 },
		{ 453, 483, 56 },
		{ 453, 484, 56 },
		{ 453, 485, 56 },
		{ 453, 486, 56 },
		{ 453, 487, 56 },
		{ 453, 488, 56 },
		{ 453, 489, 56 },
		{ 453, 490, 56 },
		{ 453, 491, 56 },
		{ 453, 492, 56 },
		{ 453, 493, 56 },
		{ 453, 494, 56 },
		{ 453, 495, 56 },
		{ 453, 496, 56 },
		{ 453, 497, 56 },
		{ 453, 498, 56 },
		{ 453, 499, 56 },
		{ 453, 500, 56 },
		{ 453, 501, 56 },
		{ 453, 502, 56 },
		{ 453, 503, 56 },
		{ 453, 504, 56 },
		{ 453, 505, 56 },
		{ 453, 506, 56 },
		{ 453, 507, 56 },
		{ 453, 508, 56 },
		{ 453, 509, 56 },
		{ 453, 510, 56 },
		{ 453, 511, 56 },
		{ 453, 512, 56 },
		{ 454, 483, 56 },
		{ 454, 484, 56 },
		{ 454, 485, 56 },
		{ 454, 486, 56 },
		{ 454, 487, 56 },
		{ 454, 488, 56 },
		{ 454, 489, 56 },
		{ 454, 490, 56 },
		{ 454, 491, 56 },
		{ 454, 492, 56 },
		{ 454, 493, 56 },
		{ 454, 494, 56 },
		{ 454, 495, 56 },
		{ 454, 496, 56 },
		{ 454, 497, 56 },
		{ 454, 498, 56 },
		{ 454, 499, 56 },
		{ 454, 500, 56 },
		{ 454, 501, 56 },
		{ 454, 502, 56 },
		{ 454, 503, 56 },
		{ 454, 504, 56 },
		{ 454, 505, 56 },
		{ 454, 506, 56 },
		{ 454, 507, 56 },
		{ 454, 508, 56 },
		{ 454, 509, 56 },
		{ 454, 510, 56 },
		{ 454, 511, 56 },
		{ 454, 512, 56 },
		{ 455, 484, 56 },
		{ 455, 485, 56 },
		{ 455, 486, 56 },
		{ 455, 487, 56 },
		{ 455, 488, 56 },
		{ 455, 489, 56 },
		{ 455, 490, 56 },
		{ 455, 491, 56 },
		{ 455, 492, 56 },
		{ 455, 493, 56 },
		{ 455, 494, 56 },
		{ 455, 495, 56 },
		{ 455, 496, 56 },
		{ 455, 497, 56 },
		{ 455, 498, 56 },
		{ 455, 499, 56 },
		{ 455, 500, 56 },
		{ 455, 501, 56 },
		{ 455, 502, 56 },
		{ 455, 503, 56 },
		{ 455, 504, 56 },
		{ 455, 505, 56 },
		{ 455, 506, 56 },
		{ 455, 507, 56 },
		{ 455, 508, 56 },
		{ 455, 509, 56 },
		{ 455, 510, 56 },
		{ 455, 511, 56 },
		{ 455, 512, 56 },
		{ 456, 485, 56 },
		{ 456, 486, 56 },
		{ 456, 487, 56 },
		{ 456, 488, 56 },
		{ 456, 489, 56 },
		{ 456, 490, 56 },
		{ 456, 491, 56 },
		{ 456, 492, 56 },
		{ 456, 493, 56 },
		{ 456, 494, 56 },
		{ 456, 495, 56 },
		{ 456, 496, 56 },
		{ 456, 497, 56 },
		{ 456, 498, 56 },
		{ 456, 499, 56 },
		{ 456, 500, 56 },
		{ 456, 501, 56 },
		{ 456, 502, 56 },
		{ 456, 503, 56 },
		{ 456, 504, 56 },
		{ 456, 505, 56 },
		{ 456, 506, 56 },
		{ 456, 507, 56 },
		{ 456, 508, 56 },
		{ 456, 509, 56 },
		{ 456, 510, 56 },
		{ 456, 511, 56 },
		{ 456, 512, 56 },
		{ 457, 486, 56 },
		{ 457, 487, 56 },
		{ 457, 488, 56 },
		{ 457, 489, 56 },
		{ 457, 490, 56 },
		{ 457, 491, 56 },
		{ 457, 492, 56 },
		{ 457, 493, 56 },
		{ 457, 494, 56 },
		{ 457, 495, 56 },
		{ 457, 496, 56 },
		{ 457, 497, 56 },
		{ 457, 498, 56 },
		{ 457, 499, 56 },
		{ 457, 500, 56 },
		{ 457, 501, 56 },
		{ 457, 502, 56 },
		{ 457, 503, 56 },
		{ 457, 504, 56 },
		{ 457, 505, 56 },
		{ 457, 506, 56 },
		{ 457, 507, 56 },
		{ 457, 508, 56 },
		{ 457, 509, 56 },
		{ 457, 510, 56 },
		{ 457, 511, 56 },
		{ 457, 512, 56 },
		{ 458, 487, 56 },
		{ 458, 488, 56 },
		{ 458, 489, 56 },
		{ 458, 490, 56 },
		{ 458, 491, 56 },
		{ 458, 492, 56 },
		{ 458, 493, 56 },
		{ 458, 494, 56 },
		{ 458, 495, 56 },
		{ 458, 496, 56 },
		{ 458, 497, 56 },
		{ 458, 498, 56 },
		{ 458, 499, 56 },
		{ 458, 500, 56 },
		{ 458, 501, 56 },
		{ 458, 502, 56 },
		{ 458, 503, 56 },
		{ 458, 504, 56 },
		{ 458, 505, 56 },
		{ 458, 506, 56 },
		{ 458, 507, 56 },
		{ 458, 508, 56 },
		{ 458, 509, 56 },
		{ 458, 510, 56 },
		{ 458, 511, 56 },
		{ 458, 512, 56 },
		{ 459, 488, 56 },
		{ 459, 489, 56 },
		{ 459, 490, 56 },
		{ 459, 491, 56 },
		{ 459, 492, 56 },
		{ 459, 493, 56 },
		{ 459, 494, 56 },
		{ 459, 495, 56 },
		{ 459, 496, 56 },
		{ 459, 497, 56 },
		{ 459, 498, 56 },
		{ 459, 499, 56 },
		{ 459, 500, 56 },
		{ 459, 501, 56 },
		{ 459, 502, 56 },
		{ 459, 503, 56 },
		{ 459, 504, 56 },
		{ 459, 505, 56 },
		{ 459, 506, 56 },
		{ 459, 507, 56 },
		{ 459, 508, 56 },
		{ 459, 509, 56 },
		{ 459, 510, 56 },
		{ 459, 511, 56 },
		{ 459, 512, 56 },
		{ 460, 490, 56 },
		{ 460, 491, 56 },
		{ 460, 492, 56 },
		{ 460, 493, 56 },
		{ 460, 494, 56 },
		{ 460, 495, 56 },
		{ 460, 496, 56 },
		{ 460, 497, 56 },
		{ 460, 498, 56 },
		{ 460, 499, 56 },
		{ 460, 500, 56 },
		{ 460, 501, 56 },
		{ 460, 502, 56 },
		{ 460, 503, 56 },
		{ 460, 504, 56 },
		{ 460, 505, 56 },
		{ 460, 506, 56 },
		{ 460, 507, 56 },
		{ 460, 508, 56 },
		{ 460, 509, 56 },
		{ 460, 510, 56 },
		{ 460, 511, 56 },
		{ 460, 512, 56 },
		{ 461, 491, 56 },
		{ 461, 492, 56 },
		{ 461, 493, 56 },
		{ 461, 494, 56 },
		{ 461, 495, 56 },
		{ 461, 496, 56 },
		{ 461, 497, 56 },
		{ 461, 498, 56 },
		{ 461, 499, 56 },
		{ 461, 500, 56 },
		{ 461, 501, 56 },
		{ 461, 502, 56 },
		{ 461, 503, 56 },
		{ 461, 504, 56 },
		{ 461, 505, 56 },
		{ 461, 506, 56 },
		{ 461, 507, 56 },
		{ 461, 508, 56 },
		{ 461, 509, 56 },
		{ 461, 510, 56 },
		{ 461, 511, 56 },
		{ 461, 512, 56 },
		{ 462, 492, 56 },
		{ 462, 493, 56 },
		{ 462, 494, 56 },
		{ 462, 495, 56 },
		{ 462, 496, 56 },
		{ 462, 497, 56 },
		{ 462, 498, 56 },
		{ 462, 499, 56 },
		{ 462, 500, 56 },
		{ 462, 501, 56 },
		{ 462, 502, 56 },
		{ 462, 503, 56 },
		{ 462, 504, 56 },
		{ 462, 505, 56 },
		{ 462, 506, 56 },
		{ 462, 507, 56 },
		{ 462, 508, 56 },
		{ 462, 509, 56 },
		{ 462, 510, 56 },
		{ 462, 511, 56 },
		{ 462, 512, 56 },
		{ 463, 493, 56 },
		{ 463, 494, 56 },
		{ 463, 495, 56 },
		{ 463, 496, 56 },
		{ 463, 497, 56 },
		{ 463, 498, 56 },
		{ 463, 499, 56 },
		{ 463, 500, 56 },
		{ 463, 501, 56 },
		{ 463, 502, 56 },
		{ 463, 503, 56 },
		{ 463, 504, 56 },
		{ 463, 505, 56 },
		{ 463, 506, 56 },
		{ 463, 507, 56 },
		{ 463, 508, 56 },
		{ 463, 509, 56 },
		{ 463, 510, 56 },
		{ 463, 511, 56 },
		{ 463, 512, 56 },
		{ 464, 494, 56 },
		{ 464, 495, 56 },
		{ 464, 496, 56 },
		{ 464, 497, 56 },
		{ 464, 498, 56 },
		{ 464, 499, 56 },
		{ 464, 500, 56 },
		{ 464, 501, 56 },
		{ 464, 502, 56 },
		{ 464, 503, 56 },
		{ 464, 504, 56 },
		{ 464, 505, 56 },
		{ 464, 506, 56 },
		{ 464, 507, 56 },
		{ 464, 508, 56 },
		{ 464, 509, 56 },
		{ 464, 510, 56 },
		{ 464, 511, 56 },
		{ 464, 512, 56 },
		{ 465, 495, 56 },
		{ 465, 496, 56 },
		{ 465, 497, 56 },
		{ 465, 498, 56 },
		{ 465, 499, 56 },
		{ 465, 500, 56 },
		{ 465, 501, 56 },
		{ 465, 502, 56 },
		{ 465, 503, 56 },
		{ 465, 504, 56 },
		{ 465, 505, 56 },
		{ 465, 506, 56 },
		{ 465, 507, 56 },
		{ 465, 508, 56 },
		{ 465, 509, 56 },
		{ 465, 510, 56 },
		{ 465, 511, 56 },
		{ 465, 512, 56 },
		{ 466, 496, 56 },
		{ 466, 497, 56 },
		{ 466, 498, 56 },
		{ 466, 499, 56 },
		{ 466, 500, 56 },
		{ 466, 501, 56 },
		{ 466, 502, 56 },
		{ 466, 503, 56 },
		{ 466, 504, 56 },
		{ 466, 505, 56 },
		{ 466, 506, 56 },
		{ 466, 507, 56 },
		{ 466, 508, 56 },
		{ 466, 509, 56 },
		{ 466, 510, 56 },
		{ 466, 511, 56 },
		{ 466, 512, 56 },
		{ 467, 498, 56 },
		{ 467, 499, 56 },
		{ 467, 500, 56 },
		{ 467, 501, 56 },
		{ 467, 502, 56 },
		{ 467, 503, 56 },
		{ 467, 504, 56 },
		{ 467, 505, 56 },
		{ 467, 506, 56 },
		{ 467, 507, 56 },
		{ 467, 508, 56 },
		{ 467, 509, 56 },
		{ 467, 510, 56 },
		{ 467, 511, 56 },
		{ 467, 512, 56 },
		{ 468, 499, 56 },
		{ 468, 500, 56 },
		{ 468, 501, 56 },
		{ 468, 502, 56 },
		{ 468, 503, 56 },
		{ 468, 504, 56 },
		{ 468, 505, 56 },
		{ 468, 506, 56 },
		{ 468, 507, 56 },
		{ 468, 508, 56 },
		{ 468, 509, 56 },
		{ 468, 510, 56 },
		{ 468, 511, 56 },
		{ 468, 512, 56 },
		{ 469, 500, 56 },
		{ 469, 501, 56 },
		{ 469, 502, 56 },
		{ 469, 503, 56 },
		{ 469, 504, 56 },
		{ 469, 505, 56 },
		{ 469, 506, 56 },
		{ 469, 507, 56 },
		{ 469, 508, 56 },
		{ 469, 509, 56 },
		{ 469, 510, 56 },
		{ 469, 511, 56 },
		{ 469, 512, 56 },
		{ 470, 501, 56 },
		{ 470, 502, 56 },
		{ 470, 503, 56 },
		{ 470, 504, 56 },
		{ 470, 505, 56 },
		{ 470, 506, 56 },
		{ 470, 507, 56 },
		{ 470, 508, 56 },
		{ 470, 509, 56 },
		{ 470, 510, 56 },
		{ 470, 511, 56 },
		{ 470, 512, 56 },
		{ 471, 502, 56 },
		{ 471, 503, 56 },
		{ 471, 504, 56 },
		{ 471, 505, 56 },
		{ 471, 506, 56 },
		{ 471, 507, 56 },
		{ 471, 508, 56 },
		{ 471, 509, 56 },
		{ 471, 510, 56 },
		{ 471, 511, 56 },
		{ 471, 512, 56 },
		{ 472, 503, 56 },
		{ 472, 504, 56 },
		{ 472, 505, 56 },
		{ 472, 506, 56 },
		{ 472, 507, 56 },
		{ 472, 508, 56 },
		{ 472, 509, 56 },
		{ 472, 510, 56 },
		{ 472, 511, 56 },
		{ 472, 512, 56 },
		{ 473, 504, 56 },
		{ 473, 505, 56 },
		{ 473, 506, 56 },
		{ 473, 507, 56 },
		{ 473, 508, 56 },
		{ 473, 509, 56 },
		{ 473, 510, 56 },
		{ 473, 511, 56 },
		{ 473, 512, 56 },
		{ 474, 506, 56 },
		{ 474, 507, 56 },
		{ 474, 508, 56 },
		{ 474, 509, 56 },
		{ 474, 510, 56 },
		{ 474, 511, 56 },
		{ 474, 512, 56 },
		{ 475, 507, 56 },
		{ 475, 508, 56 },
		{ 475, 509, 56 },
		{ 475, 510, 56 },
		{ 475, 511, 56 },
		{ 475, 512, 56 },
		{ 476, 508, 56 },
		{ 476, 509, 56 },
		{ 476, 510, 56 },
		{ 476, 511, 56 },
		{ 476, 512, 56 },
		{ 477, 509, 56 },
		{ 477, 510, 56 },
		{ 477, 511, 56 },
		{ 477, 512, 56 },
		{ 478, 510, 56 },
		{ 478, 511, 56 },
		{ 478, 512, 56 },
		{ 479, 511, 56 },
		{ 479, 512, 56 },
		{ 480, 512, 56 },
		{ 66, 1, 57 },
		{ 66, 2, 57 },
		{ 67, 1, 57 },
		{ 67, 2, 57 },
		{ 67, 3, 57 },
		{ 68, 1, 57 },
		{ 68, 2, 57 },
		{ 68, 3, 57 },
		{ 68, 4, 57 },
		{ 69, 1, 57 },
		{ 69, 2, 57 },
		{ 69, 3, 57 },
		{ 69, 4, 57 },
		{ 69, 5, 57 },
		{ 69, 6, 57 },
		{ 70, 1, 57 },
		{ 70, 2, 57 },
		{ 70, 3, 57 },
		{ 70, 4, 57 },
		{ 70, 5, 57 },
		{ 70, 6, 57 },
		{ 70, 7, 57 },
		{ 71, 1, 57 },
		{ 71, 2, 57 },
		{ 71, 3, 57 },
		{ 71, 4, 57 },
		{ 71, 5, 57 },
		{ 71, 6, 57 },
		{ 71, 7, 57 },
		{ 71, 8, 57 },
		{ 72, 1, 57 },
		{ 72, 2, 57 },
		{ 72, 3, 57 },
		{ 72, 4, 57 },
		{ 72, 5, 57 },
		{ 72, 6, 57 },
		{ 72, 7, 57 },
		{ 72, 8, 57 },
		{ 72, 9, 57 },
		{ 72, 10, 57 },
		{ 73, 1, 57 },
		{ 73, 2, 57 },
		{ 73, 3, 57 },
		{ 73, 4, 57 },
		{ 73, 5, 57 },
		{ 73, 6, 57 },
		{ 73, 7, 57 },
		{ 73, 8, 57 },
		{ 73, 9, 57 },
		{ 73, 10, 57 },
		{ 73, 11, 57 },
		{ 74, 1, 57 },
		{ 74, 2, 57 },
		{ 74, 3, 57 },
		{ 74, 4, 57 },
		{ 74, 5, 57 },
		{ 74, 6, 57 },
		{ 74, 7, 57 },
		{ 74, 8, 57 },
		{ 74, 9, 57 },
		{ 74, 10, 57 },
		{ 74, 11, 57 },
		{ 74, 12, 57 },
		{ 75, 1, 57 },
		{ 75, 2, 57 },
		{ 75, 3, 57 },
		{ 75, 4, 57 },
		{ 75, 5, 57 },
		{ 75, 6, 57 },
		{ 75, 7, 57 },
		{ 75, 8, 57 },
		{ 75, 9, 57 },
		{ 75, 10, 57 },
		{ 75, 11, 57 },
		{ 75, 12, 57 },
		{ 75, 13, 57 },
		{ 75, 14, 57 },
		{ 76, 1, 57 },
		{ 76, 2, 57 },
		{ 76, 3, 57 },
		{ 76, 4, 57 },
		{ 76, 5, 57 },
		{ 76, 6, 57 },
		{ 76, 7, 57 },
		{ 76, 8, 57 },
		{ 76, 9, 57 },
		{ 76, 10, 57 },
		{ 76, 11, 57 },
		{ 76, 12, 57 },
		{ 76, 13, 57 },
		{ 76, 14, 57 },
		{ 76, 15, 57 },
		{ 77, 1, 57 },
		{ 77, 2, 57 },
		{ 77, 3, 57 },
		{ 77, 4, 57 },
		{ 77, 5, 57 },
		{ 77, 6, 57 },
		{ 77, 7, 57 },
		{ 77, 8, 57 },
		{ 77, 9, 57 },
		{ 77, 10, 57 },
		{ 77, 11, 57 },
		{ 77, 12, 57 },
		{ 77, 13, 57 },
		{ 77, 14, 57 },
		{ 77, 15, 57 },
		{ 77, 16, 57 },
		{ 78, 1, 57 },
		{ 78, 2, 57 },
		{ 78, 3, 57 },
		{ 78, 4, 57 },
		{ 78, 5, 57 },
		{ 78, 6, 57 },
		{ 78, 7, 57 },
		{ 78, 8, 57 },
		{ 78, 9, 57 },
		{ 78, 10, 57 },
		{ 78, 11, 57 },
		{ 78, 12, 57 },
		{ 78, 13, 57 },
		{ 78, 14, 57 },
		{ 78, 15, 57 },
		{ 78, 16, 57 },
		{ 78, 17, 57 },
		{ 78, 18, 57 },
		{ 79, 1, 57 },
		{ 79, 2, 57 },
		{ 79, 3, 57 },
		{ 79, 4, 57 },
		{ 79, 5, 57 },
		{ 79, 6, 57 },
		{ 79, 7, 57 },
		{ 79, 8, 57 },
		{ 79, 9, 57 },
		{ 79, 10, 57 },
		{ 79, 11, 57 },
		{ 79, 12, 57 },
		{ 79, 13, 57 },
		{ 79, 14, 57 },
		{ 79, 15, 57 },
		{ 79, 16, 57 },
		{ 79, 17, 57 },
		{ 79, 18, 57 },
		{ 79, 19, 57 },
		{ 80, 1, 57 },
		{ 80, 2, 57 },
		{ 80, 3, 57 },
		{ 80, 4, 57 },
		{ 80, 5, 57 },
		{ 80, 6, 57 },
		{ 80, 7, 57 },
		{ 80, 8, 57 },
		{ 80, 9, 57 },
		{ 80, 10, 57 },
		{ 80, 11, 57 },
		{ 80, 12, 57 },
		{ 80, 13, 57 },
		{ 80, 14, 57 },
		{ 80, 15, 57 },
		{ 80, 16, 57 },
		{ 80, 17, 57 },
		{ 80, 18, 57 },
		{ 80, 19, 57 },
		{ 80, 20, 57 },
		{ 81, 1, 57 },
		{ 81, 2, 57 },
		{ 81, 3, 57 },
		{ 81, 4, 57 },
		{ 81, 5, 57 },
		{ 81, 6, 57 },
		{ 81, 7, 57 },
		{ 81, 8, 57 },
		{ 81, 9, 57 },
		{ 81, 10, 57 },
		{ 81, 11, 57 },
		{ 81, 12, 57 },
		{ 81, 13, 57 },
		{ 81, 14, 57 },
		{ 81, 15, 57 },
		{ 81, 16, 57 },
		{ 81, 17, 57 },
		{ 81, 18, 57 },
		{ 81, 19, 57 },
		{ 81, 20, 57 },
		{ 81, 21, 57 },
		{ 81, 22, 57 },
		{ 82, 1, 57 },
		{ 82, 2, 57 },
		{ 82, 3, 57 },
		{ 82, 4, 57 },
		{ 82, 5, 57 },
		{ 82, 6, 57 },
		{ 82, 7, 57 },
		{ 82, 8, 57 },
		{ 82, 9, 57 },
		{ 82, 10, 57 },
		{ 82, 11, 57 },
		{ 82, 12, 57 },
		{ 82, 13, 57 },
		{ 82, 14, 57 },
		{ 82, 15, 57 },
		{ 82, 16, 57 },
		{ 82, 17, 57 },
		{ 82, 18, 57 },
		{ 82, 19, 57 },
		{ 82, 20, 57 },
		{ 82, 21, 57 },
		{ 82, 22, 57 },
		{ 82, 23, 57 },
		{ 83, 1, 57 },
		{ 83, 2, 57 },
		{ 83, 3, 57 },
		{ 83, 4, 57 },
		{ 83, 5, 57 },
		{ 83, 6, 57 },
		{ 83, 7, 57 },
		{ 83, 8, 57 },
		{ 83, 9, 57 },
		{ 83, 10, 57 },
		{ 83, 11, 57 },
		{ 83, 12, 57 },
		{ 83, 13, 57 },
		{ 83, 14, 57 },
		{ 83, 15, 57 },
		{ 83, 16, 57 },
		{ 83, 17, 57 },
		{ 83, 18, 57 },
		{ 83, 19, 57 },
		{ 83, 20, 57 },
		{ 83, 21, 57 },
		{ 83, 22, 57 },
		{ 83, 23, 57 },
		{ 83, 24, 57 },
		{ 84, 1, 57 },
		{ 84, 2, 57 },
		{ 84, 3, 57 },
		{ 84, 4, 57 },
		{ 84, 5, 57 },
		{ 84, 6, 57 },
		{ 84, 7, 57 },
		{ 84, 8, 57 },
		{ 84, 9, 57 },
		{ 84, 10, 57 },
		{ 84, 11, 57 },
		{ 84, 12, 57 },
		{ 84, 13, 57 },
		{ 84, 14, 57 },
		{ 84, 15, 57 },
		{ 84, 16, 57 },
		{ 84, 17, 57 },
		{ 84, 18, 57 },
		{ 84, 19, 57 },
		{ 84, 20, 57 },
		{ 84, 21, 57 },
		{ 84, 22, 57 },
		{ 84, 23, 57 },
		{ 84, 24, 57 },
		{ 84, 25, 57 },
		{ 84, 26, 57 },
		{ 85, 1, 57 },
		{ 85, 2, 57 },
		{ 85, 3, 57 },
		{ 85, 4, 57 },
		{ 85, 5, 57 },
		{ 85, 6, 57 },
		{ 85, 7, 57 },
		{ 85, 8, 57 },
		{ 85, 9, 57 },
		{ 85, 10, 57 },
		{ 85, 11, 57 },
		{ 85, 12, 57 },
		{ 85, 13, 57 },
		{ 85, 14, 57 },
		{ 85, 15, 57 },
		{ 85, 16, 57 },
		{ 85, 17, 57 },
		{ 85, 18, 57 },
		{ 85, 19, 57 },
		{ 85, 20, 57 },
		{ 85, 21, 57 },
		{ 85, 22, 57 },
		{ 85, 23, 57 },
		{ 85, 24, 57 },
		{ 85, 25, 57 },
		{ 85, 26, 57 },
		{ 85, 27, 57 },
		{ 86, 1, 57 },
		{ 86, 2, 57 },
		{ 86, 3, 57 },
		{ 86, 4, 57 },
		{ 86, 5, 57 },
		{ 86, 6, 57 },
		{ 86, 7, 57 },
		{ 86, 8, 57 },
		{ 86, 9, 57 },
		{ 86, 10, 57 },
		{ 86, 11, 57 },
		{ 86, 12, 57 },
		{ 86, 13, 57 },
		{ 86, 14, 57 },
		{ 86, 15, 57 },
		{ 86, 16, 57 },
		{ 86, 17, 57 },
		{ 86, 18, 57 },
		{ 86, 19, 57 },
		{ 86, 20, 57 },
		{ 86, 21, 57 },
		{ 86, 22, 57 },
		{ 86, 23, 57 },
		{ 86, 24, 57 },
		{ 86, 25, 57 },
		{ 86, 26, 57 },
		{ 86, 27, 57 },
		{ 86, 28, 57 },
		{ 87, 1, 57 },
		{ 87, 2, 57 },
		{ 87, 3, 57 },
		{ 87, 4, 57 },
		{ 87, 5, 57 },
		{ 87, 6, 57 },
		{ 87, 7, 57 },
		{ 87, 8, 57 },
		{ 87, 9, 57 },
		{ 87, 10, 57 },
		{ 87, 11, 57 },
		{ 87, 12, 57 },
		{ 87, 13, 57 },
		{ 87, 14, 57 },
		{ 87, 15, 57 },
		{ 87, 16, 57 },
		{ 87, 17, 57 },
		{ 87, 18, 57 },
		{ 87, 19, 57 },
		{ 87, 20, 57 },
		{ 87, 21, 57 },
		{ 87, 22, 57 },
		{ 87, 23, 57 },
		{ 87, 24, 57 },
		{ 87, 25, 57 },
		{ 87, 26, 57 },
		{ 87, 27, 57 },
		{ 87, 28, 57 },
		{ 87, 29, 57 },
		{ 87, 30, 57 },
		{ 88, 1, 57 },
		{ 88, 2, 57 },
		{ 88, 3, 57 },
		{ 88, 4, 57 },
		{ 88, 5, 57 },
		{ 88, 6, 57 },
		{ 88, 7, 57 },
		{ 88, 8, 57 },
		{ 88, 9, 57 },
		{ 88, 10, 57 },
		{ 88, 11, 57 },
		{ 88, 12, 57 },
		{ 88, 13, 57 },
		{ 88, 14, 57 },
		{ 88, 15, 57 },
		{ 88, 16, 57 },
		{ 88, 17, 57 },
		{ 88, 18, 57 },
		{ 88, 19, 57 },
		{ 88, 20, 57 },
		{ 88, 21, 57 },
		{ 88, 22, 57 },
		{ 88, 23, 57 },
		{ 88, 24, 57 },
		{ 88, 25, 57 },
		{ 88, 26, 57 },
		{ 88, 27, 57 },
		{ 88, 28, 57 },
		{ 88, 29, 57 },
		{ 88, 30, 57 },
		{ 88, 31, 57 },
		{ 89, 1, 57 },
		{ 89, 2, 57 },
		{ 89, 3, 57 },
		{ 89, 4, 57 },
		{ 89, 5, 57 },
		{ 89, 6, 57 },
		{ 89, 7, 57 },
		{ 89, 8, 57 },
		{ 89, 9, 57 },
		{ 89, 10, 57 },
		{ 89, 11, 57 },
		{ 89, 12, 57 },
		{ 89, 13, 57 },
		{ 89, 14, 57 },
		{ 89, 15, 57 },
		{ 89, 16, 57 },
		{ 89, 17, 57 },
		{ 89, 18, 57 },
		{ 89, 19, 57 },
		{ 89, 20, 57 },
		{ 89, 21, 57 },
		{ 89, 22, 57 },
		{ 89, 23, 57 },
		{ 89, 24, 57 },
		{ 89, 25, 57 },
		{ 89, 26, 57 },
		{ 89, 27, 57 },
		{ 89, 28, 57 },
		{ 89, 29, 57 },
		{ 89, 30, 57 },
		{ 89, 31, 57 },
		{ 89, 32, 57 },
		{ 90, 1, 57 },
		{ 90, 2, 57 },
		{ 90, 3, 57 },
		{ 90, 4, 57 },
		{ 90, 5, 57 },
		{ 90, 6, 57 },
		{ 90, 7, 57 },
		{ 90, 8, 57 },
		{ 90, 9, 57 },
		{ 90, 10, 57 },
		{ 90, 11, 57 },
		{ 90, 12, 57 },
		{ 90, 13, 57 },
		{ 90, 14, 57 },
		{ 90, 15, 57 },
		{ 90, 16, 57 },
		{ 90, 17, 57 },
		{ 90, 18, 57 },
		{ 90, 19, 57 },
		{ 90, 20, 57 },
		{ 90, 21, 57 },
		{ 90, 22, 57 },
		{ 90, 23, 57 },
		{ 90, 24, 57 },
		{ 90, 25, 57 },
		{ 90, 26, 57 },
		{ 90, 27, 57 },
		{ 90, 28, 57 },
		{ 90, 29, 57 },
		{ 90, 30, 57 },
		{ 90, 31, 57 },
		{ 90, 32, 57 },
		{ 90, 33, 57 },
		{ 90, 34, 57 },
		{ 91, 1, 57 },
		{ 91, 2, 57 },
		{ 91, 3, 57 },
		{ 91, 4, 57 },
		{ 91, 5, 57 },
		{ 91, 6, 57 },
		{ 91, 7, 57 },
		{ 91, 8, 57 },
		{ 91, 9, 57 },
		{ 91, 10, 57 },
		{ 91, 11, 57 },
		{ 91, 12, 57 },
		{ 91, 13, 57 },
		{ 91, 14, 57 },
		{ 91, 15, 57 },
		{ 91, 16, 57 },
		{ 91, 17, 57 },
		{ 91, 18, 57 },
		{ 91, 19, 57 },
		{ 91, 20, 57 },
		{ 91, 21, 57 },
		{ 91, 22, 57 },
		{ 91, 23, 57 },
		{ 91, 24, 57 },
		{ 91, 25, 57 },
		{ 91, 26, 57 },
		{ 91, 27, 57 },
		{ 91, 28, 57 },
		{ 91, 29, 57 },
		{ 91, 30, 57 },
		{ 91, 31, 57 },
		{ 91, 32, 57 },
		{ 91, 33, 57 },
		{ 91, 34, 57 },
		{ 91, 35, 57 },
		{ 92, 1, 57 },
		{ 92, 2, 57 },
		{ 92, 3, 57 },
		{ 92, 4, 57 },
		{ 92, 5, 57 },
		{ 92, 6, 57 },
		{ 92, 7, 57 },
		{ 92, 8, 57 },
		{ 92, 9, 57 },
		{ 92, 10, 57 },
		{ 92, 11, 57 },
		{ 92, 12, 57 },
		{ 92, 13, 57 },
		{ 92, 14, 57 },
		{ 92, 15, 57 },
		{ 92, 16, 57 },
		{ 92, 17, 57 },
		{ 92, 18, 57 },
		{ 92, 19, 57 },
		{ 92, 20, 57 },
		{ 92, 21, 57 },
		{ 92, 22, 57 },
		{ 92, 23, 57 },
		{ 92, 24, 57 },
		{ 92, 25, 57 },
		{ 92, 26, 57 },
		{ 92, 27, 57 },
		{ 92, 28, 57 },
		{ 92, 29, 57 },
		{ 92, 30, 57 },
		{ 92, 31, 57 },
		{ 92, 32, 57 },
		{ 92, 33, 57 },
		{ 92, 34, 57 },
		{ 92, 35, 57 },
		{ 92, 36, 57 },
		{ 93, 1, 57 },
		{ 93, 2, 57 },
		{ 93, 3, 57 },
		{ 93, 4, 57 },
		{ 93, 5, 57 },
		{ 93, 6, 57 },
		{ 93, 7, 57 },
		{ 93, 8, 57 },
		{ 93, 9, 57 },
		{ 93, 10, 57 },
		{ 93, 11, 57 },
		{ 93, 12, 57 },
		{ 93, 13, 57 },
		{ 93, 14, 57 },
		{ 93, 15, 57 },
		{ 93, 16, 57 },
		{ 93, 17, 57 },
		{ 93, 18, 57 },
		{ 93, 19, 57 },
		{ 93, 20, 57 },
		{ 93, 21, 57 },
		{ 93, 22, 57 },
		{ 93, 23, 57 },
		{ 93, 24, 57 },
		{ 93, 25, 57 },
		{ 93, 26, 57 },
		{ 93, 27, 57 },
		{ 93, 28, 57 },
		{ 93, 29, 57 },
		{ 93, 30, 57 },
		{ 93, 31, 57 },
		{ 93, 32, 57 },
		{ 93, 33, 57 },
		{ 93, 34, 57 },
		{ 93, 35, 57 },
		{ 93, 36, 57 },
		{ 93, 37, 57 },
		{ 93, 38, 57 },
		{ 94, 1, 57 },
		{ 94, 2, 57 },
		{ 94, 3, 57 },
		{ 94, 4, 57 },
		{ 94, 5, 57 },
		{ 94, 6, 57 },
		{ 94, 7, 57 },
		{ 94, 8, 57 },
		{ 94, 9, 57 },
		{ 94, 10, 57 },
		{ 94, 11, 57 },
		{ 94, 12, 57 },
		{ 94, 13, 57 },
		{ 94, 14, 57 },
		{ 94, 15, 57 },
		{ 94, 16, 57 },
		{ 94, 17, 57 },
		{ 94, 18, 57 },
		{ 94, 19, 57 },
		{ 94, 20, 57 },
		{ 94, 21, 57 },
		{ 94, 22, 57 },
		{ 94, 23, 57 },
		{ 94, 24, 57 },
		{ 94, 25, 57 },
		{ 94, 26, 57 },
		{ 94, 27, 57 },
		{ 94, 28, 57 },
		{ 94, 29, 57 },
		{ 94, 30, 57 },
		{ 94, 31, 57 },
		{ 94, 32, 57 },
		{ 94, 33, 57 },
		{ 94, 34, 57 },
		{ 94, 35, 57 },
		{ 94, 36, 57 },
		{ 94, 37, 57 },
		{ 94, 38, 57 },
		{ 94, 39, 57 },
		{ 95, 1, 57 },
		{ 95, 2, 57 },
		{ 95, 3, 57 },
		{ 95, 4, 57 },
		{ 95, 5, 57 },
		{ 95, 6, 57 },
		{ 95, 7, 57 },
		{ 95, 8, 57 },
		{ 95, 9, 57 },
		{ 95, 10, 57 },
		{ 95, 11, 57 },
		{ 95, 12, 57 },
		{ 95, 13, 57 },
		{ 95, 14, 57 },
		{ 95, 15, 57 },
		{ 95, 16, 57 },
		{ 95, 17, 57 },
		{ 95, 18, 57 },
		{ 95, 19, 57 },
		{ 95, 20, 57 },
		{ 95, 21, 57 },
		{ 95, 22, 57 },
		{ 95, 23, 57 },
		{ 95, 24, 57 },
		{ 95, 25, 57 },
		{ 95, 26, 57 },
		{ 95, 27, 57 },
		{ 95, 28, 57 },
		{ 95, 29, 57 },
		{ 95, 30, 57 },
		{ 95, 31, 57 },
		{ 95, 32, 57 },
		{ 95, 33, 57 },
		{ 95, 34, 57 },
		{ 95, 35, 57 },
		{ 95, 36, 57 },
		{ 95, 37, 57 },
		{ 95, 38, 57 },
		{ 95, 39, 57 },
		{ 95, 40, 57 },
		{ 96, 1, 57 },
		{ 96, 2, 57 },
		{ 96, 3, 57 },
		{ 96, 4, 57 },
		{ 96, 5, 57 },
		{ 96, 6, 57 },
		{ 96, 7, 57 },
		{ 96, 8, 57 },
		{ 96, 9, 57 },
		{ 96, 10, 57 },
		{ 96, 11, 57 },
		{ 96, 12, 57 },
		{ 96, 13, 57 },
		{ 96, 14, 57 },
		{ 96, 15, 57 },
		{ 96, 16, 57 },
		{ 96, 17, 57 },
		{ 96, 18, 57 },
		{ 96, 19, 57 },
		{ 96, 20, 57 },
		{ 96, 21, 57 },
		{ 96, 22, 57 },
		{ 96, 23, 57 },
		{ 96, 24, 57 },
		{ 96, 25, 57 },
		{ 96, 26, 57 },
		{ 96, 27, 57 },
		{ 96, 28, 57 },
		{ 96, 29, 57 },
		{ 96, 30, 57 },
		{ 96, 31, 57 },
		{ 96, 32, 57 },
		{ 96, 33, 57 },
		{ 96, 34, 57 },
		{ 96, 35, 57 },
		{ 96, 36, 57 },
		{ 96, 37, 57 },
		{ 96, 38, 57 },
		{ 96, 39, 57 },
		{ 96, 40, 57 },
		{ 96, 41, 57 },
		{ 96, 42, 57 },
		{ 97, 1, 57 },
		{ 97, 2, 57 },
		{ 97, 3, 57 },
		{ 97, 4, 57 },
		{ 97, 5, 57 },
		{ 97, 6, 57 },
		{ 97, 7, 57 },
		{ 97, 8, 57 },
		{ 97, 9, 57 },
		{ 97, 10, 57 },
		{ 97, 11, 57 },
		{ 97, 12, 57 },
		{ 97, 13, 57 },
		{ 97, 14, 57 },
		{ 97, 15, 57 },
		{ 97, 16, 57 },
		{ 97, 17, 57 },
		{ 97, 18, 57 },
		{ 97, 19, 57 },
		{ 97, 20, 57 },
		{ 97, 21, 57 },
		{ 97, 22, 57 },
		{ 97, 23, 57 },
		{ 97, 24, 57 },
		{ 97, 25, 57 },
		{ 97, 26, 57 },
		{ 97, 27, 57 },
		{ 97, 28, 57 },
		{ 97, 29, 57 },
		{ 97, 30, 57 },
		{ 97, 31, 57 },
		{ 97, 32, 57 },
		{ 97, 33, 57 },
		{ 97, 34, 57 },
		{ 97, 35, 57 },
		{ 97, 36, 57 },
		{ 97, 37, 57 },
		{ 97, 38, 57 },
		{ 97, 39, 57 },
		{ 97, 40, 57 },
		{ 97, 41, 57 },
		{ 97, 42, 57 },
		{ 97, 43, 57 },
		{ 98, 1, 57 },
		{ 98, 2, 57 },
		{ 98, 3, 57 },
		{ 98, 4, 57 },
		{ 98, 5, 57 },
		{ 98, 6, 57 },
		{ 98, 7, 57 },
		{ 98, 8, 57 },
		{ 98, 9, 57 },
		{ 98, 10, 57 },
		{ 98, 11, 57 },
		{ 98, 12, 57 },
		{ 98, 13, 57 },
		{ 98, 14, 57 },
		{ 98, 15, 57 },
		{ 98, 16, 57 },
		{ 98, 17, 57 },
		{ 98, 18, 57 },
		{ 98, 19, 57 },
		{ 98, 20, 57 },
		{ 98, 21, 57 },
		{ 98, 22, 57 },
		{ 98, 23, 57 },
		{ 98, 24, 57 },
		{ 98, 25, 57 },
		{ 98, 26, 57 },
		{ 98, 27, 57 },
		{ 98, 28, 57 },
		{ 98, 29, 57 },
		{ 98, 30, 57 },
		{ 98, 31, 57 },
		{ 98, 32, 57 },
		{ 98, 33, 57 },
		{ 98, 34, 57 },
		{ 98, 35, 57 },
		{ 98, 36, 57 },
		{ 98, 37, 57 },
		{ 98, 38, 57 },
		{ 98, 39, 57 },
		{ 98, 40, 57 },
		{ 98, 41, 57 },
		{ 98, 42, 57 },
		{ 98, 43, 57 },
		{ 98, 44, 57 },
		{ 99, 1, 57 },
		{ 99, 2, 57 },
		{ 99, 3, 57 },
		{ 99, 4, 57 },
		{ 99, 5, 57 },
		{ 99, 6, 57 },
		{ 99, 7, 57 },
		{ 99, 8, 57 },
		{ 99, 9, 57 },
		{ 99, 10, 57 },
		{ 99, 11, 57 },
		{ 99, 12, 57 },
		{ 99, 13, 57 },
		{ 99, 14, 57 },
		{ 99, 15, 57 },
		{ 99, 16, 57 },
		{ 99, 17, 57 },
		{ 99, 18, 57 },
		{ 99, 19, 57 },
		{ 99, 20, 57 },
		{ 99, 21, 57 },
		{ 99, 22, 57 },
		{ 99, 23, 57 },
		{ 99, 24, 57 },
		{ 99, 25, 57 },
		{ 99, 26, 57 },
		{ 99, 27, 57 },
		{ 99, 28, 57 },
		{ 99, 29, 57 },
		{ 99, 30, 57 },
		{ 99, 31, 57 },
		{ 99, 32, 57 },
		{ 99, 33, 57 },
		{ 99, 34, 57 },
		{ 99, 35, 57 },
		{ 99, 36, 57 },
		{ 99, 37, 57 },
		{ 99, 38, 57 },
		{ 99, 39, 57 },
		{ 99, 40, 57 },
		{ 99, 41, 57 },
		{ 99, 42, 57 },
		{ 99, 43, 57 },
		{ 99, 44, 57 },
		{ 99, 45, 57 },
		{ 99, 46, 57 },
		{ 100, 1, 57 },
		{ 100, 2, 57 },
		{ 100, 3, 57 },
		{ 100, 4, 57 },
		{ 100, 5, 57 },
		{ 100, 6, 57 },
		{ 100, 7, 57 },
		{ 100, 8, 57 },
		{ 100, 9, 57 },
		{ 100, 10, 57 },
		{ 100, 11, 57 },
		{ 100, 12, 57 },
		{ 100, 13, 57 },
		{ 100, 14, 57 },
		{ 100, 15, 57 },
		{ 100, 16, 57 },
		{ 100, 17, 57 },
		{ 100, 18, 57 },
		{ 100, 19, 57 },
		{ 100, 20, 57 },
		{ 100, 21, 57 },
		{ 100, 22, 57 },
		{ 100, 23, 57 },
		{ 100, 24, 57 },
		{ 100, 25, 57 },
		{ 100, 26, 57 },
		{ 100, 27, 57 },
		{ 100, 28, 57 },
		{ 100, 29, 57 },
		{ 100, 30, 57 },
		{ 100, 31, 57 },
		{ 100, 32, 57 },
		{ 100, 33, 57 },
		{ 100, 34, 57 },
		{ 100, 35, 57 },
		{ 100, 36, 57 },
		{ 100, 37, 57 },
		{ 100, 38, 57 },
		{ 100, 39, 57 },
		{ 100, 40, 57 },
		{ 100, 41, 57 },
		{ 100, 42, 57 },
		{ 100, 43, 57 },
		{ 100, 44, 57 },
		{ 100, 45, 57 },
		{ 100, 46, 57 },
		{ 100, 47, 57 },
		{ 101, 1, 57 },
		{ 101, 2, 57 },
		{ 101, 3, 57 },
		{ 101, 4, 57 },
		{ 101, 5, 57 },
		{ 101, 6, 57 },
		{ 101, 7, 57 },
		{ 101, 8, 57 },
		{ 101, 9, 57 },
		{ 101, 10, 57 },
		{ 101, 11, 57 },
		{ 101, 12, 57 },
		{ 101, 13, 57 },
		{ 101, 14, 57 },
		{ 101, 15, 57 },
		{ 101, 16, 57 },
		{ 101, 17, 57 },
		{ 101, 18, 57 },
		{ 101, 19, 57 },
		{ 101, 20, 57 },
		{ 101, 21, 57 },
		{ 101, 22, 57 },
		{ 101, 23, 57 },
		{ 101, 24, 57 },
		{ 101, 25, 57 },
		{ 101, 26, 57 },
		{ 101, 27, 57 },
		{ 101, 28, 57 },
		{ 101, 29, 57 },
		{ 101, 30, 57 },
		{ 101, 31, 57 },
		{ 101, 32, 57 },
		{ 101, 33, 57 },
		{ 101, 34, 57 },
		{ 101, 35, 57 },
		{ 101, 36, 57 },
		{ 101, 37, 57 },
		{ 101, 38, 57 },
		{ 101, 39, 57 },
		{ 101, 40, 57 },
		{ 101, 41, 57 },
		{ 101, 42, 57 },
		{ 101, 43, 57 },
		{ 101, 44, 57 },
		{ 101, 45, 57 },
		{ 101, 46, 57 },
		{ 101, 47, 57 },
		{ 101, 48, 57 },
		{ 102, 1, 57 },
		{ 102, 2, 57 },
		{ 102, 3, 57 },
		{ 102, 4, 57 },
		{ 102, 5, 57 },
		{ 102, 6, 57 },
		{ 102, 7, 57 },
		{ 102, 8, 57 },
		{ 102, 9, 57 },
		{ 102, 10, 57 },
		{ 102, 11, 57 },
		{ 102, 12, 57 },
		{ 102, 13, 57 },
		{ 102, 14, 57 },
		{ 102, 15, 57 },
		{ 102, 16, 57 },
		{ 102, 17, 57 },
		{ 102, 18, 57 },
		{ 102, 19, 57 },
		{ 102, 20, 57 },
		{ 102, 21, 57 },
		{ 102, 22, 57 },
		{ 102, 23, 57 },
		{ 102, 24, 57 },
		{ 102, 25, 57 },
		{ 102, 26, 57 },
		{ 102, 27, 57 },
		{ 102, 28, 57 },
		{ 102, 29, 57 },
		{ 102, 30, 57 },
		{ 102, 31, 57 },
		{ 102, 32, 57 },
		{ 102, 33, 57 },
		{ 102, 34, 57 },
		{ 102, 35, 57 },
		{ 102, 36, 57 },
		{ 102, 37, 57 },
		{ 102, 38, 57 },
		{ 102, 39, 57 },
		{ 102, 40, 57 },
		{ 102, 41, 57 },
		{ 102, 42, 57 },
		{ 102, 43, 57 },
		{ 102, 44, 57 },
		{ 102, 45, 57 },
		{ 102, 46, 57 },
		{ 102, 47, 57 },
		{ 102, 48, 57 },
		{ 102, 49, 57 },
		{ 102, 50, 57 },
		{ 103, 1, 57 },
		{ 103, 2, 57 },
		{ 103, 3, 57 },
		{ 103, 4, 57 },
		{ 103, 5, 57 },
		{ 103, 6, 57 },
		{ 103, 7, 57 },
		{ 103, 8, 57 },
		{ 103, 9, 57 },
		{ 103, 10, 57 },
		{ 103, 11, 57 },
		{ 103, 12, 57 },
		{ 103, 13, 57 },
		{ 103, 14, 57 },
		{ 103, 15, 57 },
		{ 103, 16, 57 },
		{ 103, 17, 57 },
		{ 103, 18, 57 },
		{ 103, 19, 57 },
		{ 103, 20, 57 },
		{ 103, 21, 57 },
		{ 103, 22, 57 },
		{ 103, 23, 57 },
		{ 103, 24, 57 },
		{ 103, 25, 57 },
		{ 103, 26, 57 },
		{ 103, 27, 57 },
		{ 103, 28, 57 },
		{ 103, 29, 57 },
		{ 103, 30, 57 },
		{ 103, 31, 57 },
		{ 103, 32, 57 },
		{ 103, 33, 57 },
		{ 103, 34, 57 },
		{ 103, 35, 57 },
		{ 103, 36, 57 },
		{ 103, 37, 57 },
		{ 103, 38, 57 },
		{ 103, 39, 57 },
		{ 103, 40, 57 },
		{ 103, 41, 57 },
		{ 103, 42, 57 },
		{ 103, 43, 57 },
		{ 103, 44, 57 },
		{ 103, 45, 57 },
		{ 103, 46, 57 },
		{ 103, 47, 57 },
		{ 103, 48, 57 },
		{ 103, 49, 57 },
		{ 103, 50, 57 },
		{ 103, 51, 57 },
		{ 104, 1, 57 },
		{ 104, 2, 57 },
		{ 104, 3, 57 },
		{ 104, 4, 57 },
		{ 104, 5, 57 },
		{ 104, 6, 57 },
		{ 104, 7, 57 },
		{ 104, 8, 57 },
		{ 104, 9, 57 },
		{ 104, 10, 57 },
		{ 104, 11, 57 },
		{ 104, 12, 57 },
		{ 104, 13, 57 },
		{ 104, 14, 57 },
		{ 104, 15, 57 },
		{ 104, 16, 57 },
		{ 104, 17, 57 },
		{ 104, 18, 57 },
		{ 104, 19, 57 },
		{ 104, 20, 57 },
		{ 104, 21, 57 },
		{ 104, 22, 57 },
		{ 104, 23, 57 },
		{ 104, 24, 57 },
		{ 104, 25, 57 },
		{ 104, 26, 57 },
		{ 104, 27, 57 },
		{ 104, 28, 57 },
		{ 104, 29, 57 },
		{ 104, 30, 57 },
		{ 104, 31, 57 },
		{ 104, 32, 57 },
		{ 104, 33, 57 },
		{ 104, 34, 57 },
		{ 104, 35, 57 },
		{ 104, 36, 57 },
		{ 104, 37, 57 },
		{ 104, 38, 57 },
		{ 104, 39, 57 },
		{ 104, 40, 57 },
		{ 104, 41, 57 },
		{ 104, 42, 57 },
		{ 104, 43, 57 },
		{ 104, 44, 57 },
		{ 104, 45, 57 },
		{ 104, 46, 57 },
		{ 104, 47, 57 },
		{ 104, 48, 57 },
		{ 104, 49, 57 },
		{ 104, 50, 57 },
		{ 104, 51, 57 },
		{ 104, 52, 57 },
		{ 105, 1, 57 },
		{ 105, 2, 57 },
		{ 105, 3, 57 },
		{ 105, 4, 57 },
		{ 105, 5, 57 },
		{ 105, 6, 57 },
		{ 105, 7, 57 },
		{ 105, 8, 57 },
		{ 105, 9, 57 },
		{ 105, 10, 57 },
		{ 105, 11, 57 },
		{ 105, 12, 57 },
		{ 105, 13, 57 },
		{ 105, 14, 57 },
		{ 105, 15, 57 },
		{ 105, 16, 57 },
		{ 105, 17, 57 },
		{ 105, 18, 57 },
		{ 105, 19, 57 },
		{ 105, 20, 57 },
		{ 105, 21, 57 },
		{ 105, 22, 57 },
		{ 105, 23, 57 },
		{ 105, 24, 57 },
		{ 105, 25, 57 },
		{ 105, 26, 57 },
		{ 105, 27, 57 },
		{ 105, 28, 57 },
		{ 105, 29, 57 },
		{ 105, 30, 57 },
		{ 105, 31, 57 },
		{ 105, 32, 57 },
		{ 105, 33, 57 },
		{ 105, 34, 57 },
		{ 105, 35, 57 },
		{ 105, 36, 57 },
		{ 105, 37, 57 },
		{ 105, 38, 57 },
		{ 105, 39, 57 },
		{ 105, 40, 57 },
		{ 105, 41, 57 },
		{ 105, 42, 57 },
		{ 105, 43, 57 },
		{ 105, 44, 57 },
		{ 105, 45, 57 },
		{ 105, 46, 57 },
		{ 105, 47, 57 },
		{ 105, 48, 57 },
		{ 105, 49, 57 },
		{ 105, 50, 57 },
		{ 105, 51, 57 },
		{ 105, 52, 57 },
		{ 105, 53, 57 },
		{ 105, 54, 57 },
		{ 106, 1, 57 },
		{ 106, 2, 57 },
		{ 106, 3, 57 },
		{ 106, 4, 57 },
		{ 106, 5, 57 },
		{ 106, 6, 57 },
		{ 106, 7, 57 },
		{ 106, 8, 57 },
		{ 106, 9, 57 },
		{ 106, 10, 57 },
		{ 106, 11, 57 },
		{ 106, 12, 57 },
		{ 106, 13, 57 },
		{ 106, 14, 57 },
		{ 106, 15, 57 },
		{ 106, 16, 57 },
		{ 106, 17, 57 },
		{ 106, 18, 57 },
		{ 106, 19, 57 },
		{ 106, 20, 57 },
		{ 106, 21, 57 },
		{ 106, 22, 57 },
		{ 106, 23, 57 },
		{ 106, 24, 57 },
		{ 106, 25, 57 },
		{ 106, 26, 57 },
		{ 106, 27, 57 },
		{ 106, 28, 57 },
		{ 106, 29, 57 },
		{ 106, 30, 57 },
		{ 106, 31, 57 },
		{ 106, 32, 57 },
		{ 106, 33, 57 },
		{ 106, 34, 57 },
		{ 106, 35, 57 },
		{ 106, 36, 57 },
		{ 106, 37, 57 },
		{ 106, 38, 57 },
		{ 106, 39, 57 },
		{ 106, 40, 57 },
		{ 106, 41, 57 },
		{ 106, 42, 57 },
		{ 106, 43, 57 },
		{ 106, 44, 57 },
		{ 106, 45, 57 },
		{ 106, 46, 57 },
		{ 106, 47, 57 },
		{ 106, 48, 57 },
		{ 106, 49, 57 },
		{ 106, 50, 57 },
		{ 106, 51, 57 },
		{ 106, 52, 57 },
		{ 106, 53, 57 },
		{ 106, 54, 57 },
		{ 106, 55, 57 },
		{ 107, 1, 57 },
		{ 107, 2, 57 },
		{ 107, 3, 57 },
		{ 107, 4, 57 },
		{ 107, 5, 57 },
		{ 107, 6, 57 },
		{ 107, 7, 57 },
		{ 107, 8, 57 },
		{ 107, 9, 57 },
		{ 107, 10, 57 },
		{ 107, 11, 57 },
		{ 107, 12, 57 },
		{ 107, 13, 57 },
		{ 107, 14, 57 },
		{ 107, 15, 57 },
		{ 107, 16, 57 },
		{ 107, 17, 57 },
		{ 107, 18, 57 },
		{ 107, 19, 57 },
		{ 107, 20, 57 },
		{ 107, 21, 57 },
		{ 107, 22, 57 },
		{ 107, 23, 57 },
		{ 107, 24, 57 },
		{ 107, 25, 57 },
		{ 107, 26, 57 },
		{ 107, 27, 57 },
		{ 107, 28, 57 },
		{ 107, 29, 57 },
		{ 107, 30, 57 },
		{ 107, 31, 57 },
		{ 107, 32, 57 },
		{ 107, 33, 57 },
		{ 107, 34, 57 },
		{ 107, 35, 57 },
		{ 107, 36, 57 },
		{ 107, 37, 57 },
		{ 107, 38, 57 },
		{ 107, 39, 57 },
		{ 107, 40, 57 },
		{ 107, 41, 57 },
		{ 107, 42, 57 },
		{ 107, 43, 57 },
		{ 107, 44, 57 },
		{ 107, 45, 57 },
		{ 107, 46, 57 },
		{ 107, 47, 57 },
		{ 107, 48, 57 },
		{ 107, 49, 57 },
		{ 107, 50, 57 },
		{ 107, 51, 57 },
		{ 107, 52, 57 },
		{ 107, 53, 57 },
		{ 107, 54, 57 },
		{ 107, 55, 57 },
		{ 107, 56, 57 },
		{ 108, 1, 57 },
		{ 108, 2, 57 },
		{ 108, 3, 57 },
		{ 108, 4, 57 },
		{ 108, 5, 57 },
		{ 108, 6, 57 },
		{ 108, 7, 57 },
		{ 108, 8, 57 },
		{ 108, 9, 57 },
		{ 108, 10, 57 },
		{ 108, 11, 57 },
		{ 108, 12, 57 },
		{ 108, 13, 57 },
		{ 108, 14, 57 },
		{ 108, 15, 57 },
		{ 108, 16, 57 },
		{ 108, 17, 57 },
		{ 108, 18, 57 },
		{ 108, 19, 57 },
		{ 108, 20, 57 },
		{ 108, 21, 57 },
		{ 108, 22, 57 },
		{ 108, 23, 57 },
		{ 108, 24, 57 },
		{ 108, 25, 57 },
		{ 108, 26, 57 },
		{ 108, 27, 57 },
		{ 108, 28, 57 },
		{ 108, 29, 57 },
		{ 108, 30, 57 },
		{ 108, 31, 57 },
		{ 108, 32, 57 },
		{ 108, 33, 57 },
		{ 108, 34, 57 },
		{ 108, 35, 57 },
		{ 108, 36, 57 },
		{ 108, 37, 57 },
		{ 108, 38, 57 },
		{ 108, 39, 57 },
		{ 108, 40, 57 },
		{ 108, 41, 57 },
		{ 108, 42, 57 },
		{ 108, 43, 57 },
		{ 108, 44, 57 },
		{ 108, 45, 57 },
		{ 108, 46, 57 },
		{ 108, 47, 57 },
		{ 108, 48, 57 },
		{ 108, 49, 57 },
		{ 108, 50, 57 },
		{ 108, 51, 57 },
		{ 108, 52, 57 },
		{ 108, 53, 57 },
		{ 108, 54, 57 },
		{ 108, 55, 57 },
		{ 108, 56, 57 },
		{ 108, 57, 57 },
		{ 108, 58, 57 },
		{ 109, 1, 57 },
		{ 109, 2, 57 },
		{ 109, 3, 57 },
		{ 109, 4, 57 },
		{ 109, 5, 57 },
		{ 109, 6, 57 },
		{ 109, 7, 57 },
		{ 109, 8, 57 },
		{ 109, 9, 57 },
		{ 109, 10, 57 },
		{ 109, 11, 57 },
		{ 109, 12, 57 },
		{ 109, 13, 57 },
		{ 109, 14, 57 },
		{ 109, 15, 57 },
		{ 109, 16, 57 },
		{ 109, 17, 57 },
		{ 109, 18, 57 },
		{ 109, 19, 57 },
		{ 109, 20, 57 },
		{ 109, 21, 57 },
		{ 109, 22, 57 },
		{ 109, 23, 57 },
		{ 109, 24, 57 },
		{ 109, 25, 57 },
		{ 109, 26, 57 },
		{ 109, 27, 57 },
		{ 109, 28, 57 },
		{ 109, 29, 57 },
		{ 109, 30, 57 },
		{ 109, 31, 57 },
		{ 109, 32, 57 },
		{ 109, 33, 57 },
		{ 109, 34, 57 },
		{ 109, 35, 57 },
		{ 109, 36, 57 },
		{ 109, 37, 57 },
		{ 109, 38, 57 },
		{ 109, 39, 57 },
		{ 109, 40, 57 },
		{ 109, 41, 57 },
		{ 109, 42, 57 },
		{ 109, 43, 57 },
		{ 109, 44, 57 },
		{ 109, 45, 57 },
		{ 109, 46, 57 },
		{ 109, 47, 57 },
		{ 109, 48, 57 },
		{ 109, 49, 57 },
		{ 109, 50, 57 },
		{ 109, 51, 57 },
		{ 109, 52, 57 },
		{ 109, 53, 57 },
		{ 109, 54, 57 },
		{ 109, 55, 57 },
		{ 109, 56, 57 },
		{ 109, 57, 57 },
		{ 109, 58, 57 },
		{ 109, 59, 57 },
		{ 110, 1, 57 },
		{ 110, 2, 57 },
		{ 110, 3, 57 },
		{ 110, 4, 57 },
		{ 110, 5, 57 },
		{ 110, 6, 57 },
		{ 110, 7, 57 },
		{ 110, 8, 57 },
		{ 110, 9, 57 },
		{ 110, 10, 57 },
		{ 110, 11, 57 },
		{ 110, 12, 57 },
		{ 110, 13, 57 },
		{ 110, 14, 57 },
		{ 110, 15, 57 },
		{ 110, 16, 57 },
		{ 110, 17, 57 },
		{ 110, 18, 57 },
		{ 110, 19, 57 },
		{ 110, 20, 57 },
		{ 110, 21, 57 },
		{ 110, 22, 57 },
		{ 110, 23, 57 },
		{ 110, 24, 57 },
		{ 110, 25, 57 },
		{ 110, 26, 57 },
		{ 110, 27, 57 },
		{ 110, 28, 57 },
		{ 110, 29, 57 },
		{ 110, 30, 57 },
		{ 110, 31, 57 },
		{ 110, 32, 57 },
		{ 110, 33, 57 },
		{ 110, 34, 57 },
		{ 110, 35, 57 },
		{ 110, 36, 57 },
		{ 110, 37, 57 },
		{ 110, 38, 57 },
		{ 110, 39, 57 },
		{ 110, 40, 57 },
		{ 110, 41, 57 },
		{ 110, 42, 57 },
		{ 110, 43, 57 },
		{ 110, 44, 57 },
		{ 110, 45, 57 },
		{ 110, 46, 57 },
		{ 110, 47, 57 },
		{ 110, 48, 57 },
		{ 110, 49, 57 },
		{ 110, 50, 57 },
		{ 110, 51, 57 },
		{ 110, 52, 57 },
		{ 110, 53, 57 },
		{ 110, 54, 57 },
		{ 110, 55, 57 },
		{ 110, 56, 57 },
		{ 110, 57, 57 },
		{ 110, 58, 57 },
		{ 110, 59, 57 },
		{ 110, 60, 57 },
		{ 111, 1, 57 },
		{ 111, 2, 57 },
		{ 111, 3, 57 },
		{ 111, 4, 57 },
		{ 111, 5, 57 },
		{ 111, 6, 57 },
		{ 111, 7, 57 },
		{ 111, 8, 57 },
		{ 111, 9, 57 },
		{ 111, 10, 57 },
		{ 111, 11, 57 },
		{ 111, 12, 57 },
		{ 111, 13, 57 },
		{ 111, 14, 57 },
		{ 111, 15, 57 },
		{ 111, 16, 57 },
		{ 111, 17, 57 },
		{ 111, 18, 57 },
		{ 111, 19, 57 },
		{ 111, 20, 57 },
		{ 111, 21, 57 },
		{ 111, 22, 57 },
		{ 111, 23, 57 },
		{ 111, 24, 57 },
		{ 111, 25, 57 },
		{ 111, 26, 57 },
		{ 111, 27, 57 },
		{ 111, 28, 57 },
		{ 111, 29, 57 },
		{ 111, 30, 57 },
		{ 111, 31, 57 },
		{ 111, 32, 57 },
		{ 111, 33, 57 },
		{ 111, 34, 57 },
		{ 111, 35, 57 },
		{ 111, 36, 57 },
		{ 111, 37, 57 },
		{ 111, 38, 57 },
		{ 111, 39, 57 },
		{ 111, 40, 57 },
		{ 111, 41, 57 },
		{ 111, 42, 57 },
		{ 111, 43, 57 },
		{ 111, 44, 57 },
		{ 111, 45, 57 },
		{ 111, 46, 57 },
		{ 111, 47, 57 },
		{ 111, 48, 57 },
		{ 111, 49, 57 },
		{ 111, 50, 57 },
		{ 111, 51, 57 },
		{ 111, 52, 57 },
		{ 111, 53, 57 },
		{ 111, 54, 57 },
		{ 111, 55, 57 },
		{ 111, 56, 57 },
		{ 111, 57, 57 },
		{ 111, 58, 57 },
		{ 111, 59, 57 },
		{ 111, 60, 57 },
		{ 111, 61, 57 },
		{ 111, 62, 57 },
		{ 112, 1, 57 },
		{ 112, 2, 57 },
		{ 112, 3, 57 },
		{ 112, 4, 57 },
		{ 112, 5, 57 },
		{ 112, 6, 57 },
		{ 112, 7, 57 },
		{ 112, 8, 57 },
		{ 112, 9, 57 },
		{ 112, 10, 57 },
		{ 112, 11, 57 },
		{ 112, 12, 57 },
		{ 112, 13, 57 },
		{ 112, 14, 57 },
		{ 112, 15, 57 },
		{ 112, 16, 57 },
		{ 112, 17, 57 },
		{ 112, 18, 57 },
		{ 112, 19, 57 },
		{ 112, 20, 57 },
		{ 112, 21, 57 },
		{ 112, 22, 57 },
		{ 112, 23, 57 },
		{ 112, 24, 57 },
		{ 112, 25, 57 },
		{ 112, 26, 57 },
		{ 112, 27, 57 },
		{ 112, 28, 57 },
		{ 112, 29, 57 },
		{ 112, 30, 57 },
		{ 112, 31, 57 },
		{ 112, 32, 57 },
		{ 112, 33, 57 },
		{ 112, 34, 57 },
		{ 112, 35, 57 },
		{ 112, 36, 57 },
		{ 112, 37, 57 },
		{ 112, 38, 57 },
		{ 112, 39, 57 },
		{ 112, 40, 57 },
		{ 112, 41, 57 },
		{ 112, 42, 57 },
		{ 112, 43, 57 },
		{ 112, 44, 57 },
		{ 112, 45, 57 },
		{ 112, 46, 57 },
		{ 112, 47, 57 },
		{ 112, 48, 57 },
		{ 112, 49, 57 },
		{ 112, 50, 57 },
		{ 112, 51, 57 },
		{ 112, 52, 57 },
		{ 112, 53, 57 },
		{ 112, 54, 57 },
		{ 112, 55, 57 },
		{ 112, 56, 57 },
		{ 112, 57, 57 },
		{ 112, 58, 57 },
		{ 112, 59, 57 },
		{ 112, 60, 57 },
		{ 112, 61, 57 },
		{ 112, 62, 57 },
		{ 112, 63, 57 },
		{ 113, 1, 57 },
		{ 113, 2, 57 },
		{ 113, 3, 57 },
		{ 113, 4, 57 },
		{ 113, 5, 57 },
		{ 113, 6, 57 },
		{ 113, 7, 57 },
		{ 113, 8, 57 },
		{ 113, 9, 57 },
		{ 113, 10, 57 },
		{ 113, 11, 57 },
		{ 113, 12, 57 },
		{ 113, 13, 57 },
		{ 113, 14, 57 },
		{ 113, 15, 57 },
		{ 113, 16, 57 },
		{ 113, 17, 57 },
		{ 113, 18, 57 },
		{ 113, 19, 57 },
		{ 113, 20, 57 },
		{ 113, 21, 57 },
		{ 113, 22, 57 },
		{ 113, 23, 57 },
		{ 113, 24, 57 },
		{ 113, 25, 57 },
		{ 113, 26, 57 },
		{ 113, 27, 57 },
		{ 113, 28, 57 },
		{ 113, 29, 57 },
		{ 113, 30, 57 },
		{ 113, 31, 57 },
		{ 113, 32, 57 },
		{ 113, 33, 57 },
		{ 113, 34, 57 },
		{ 113, 35, 57 },
		{ 113, 36, 57 },
		{ 113, 37, 57 },
		{ 113, 38, 57 },
		{ 113, 39, 57 },
		{ 113, 40, 57 },
		{ 113, 41, 57 },
		{ 113, 42, 57 },
		{ 113, 43, 57 },
		{ 113, 44, 57 },
		{ 113, 45, 57 },
		{ 113, 46, 57 },
		{ 113, 47, 57 },
		{ 113, 48, 57 },
		{ 113, 49, 57 },
		{ 113, 50, 57 },
		{ 113, 51, 57 },
		{ 113, 52, 57 },
		{ 113, 53, 57 },
		{ 113, 54, 57 },
		{ 113, 55, 57 },
		{ 113, 56, 57 },
		{ 113, 57, 57 },
		{ 113, 58, 57 },
		{ 113, 59, 57 },
		{ 113, 60, 57 },
		{ 113, 61, 57 },
		{ 113, 62, 57 },
		{ 113, 63, 57 },
		{ 113, 64, 57 },
		{ 114, 1, 57 },
		{ 114, 2, 57 },
		{ 114, 3, 57 },
		{ 114, 4, 57 },
		{ 114, 5, 57 },
		{ 114, 6, 57 },
		{ 114, 7, 57 },
		{ 114, 8, 57 },
		{ 114, 9, 57 },
		{ 114, 10, 57 },
		{ 114, 11, 57 },
		{ 114, 12, 57 },
		{ 114, 13, 57 },
		{ 114, 14, 57 },
		{ 114, 15, 57 },
		{ 114, 16, 57 },
		{ 114, 17, 57 },
		{ 114, 18, 57 },
		{ 114, 19, 57 },
		{ 114, 20, 57 },
		{ 114, 21, 57 },
		{ 114, 22, 57 },
		{ 114, 23, 57 },
		{ 114, 24, 57 },
		{ 114, 25, 57 },
		{ 114, 26, 57 },
		{ 114, 27, 57 },
		{ 114, 28, 57 },
		{ 114, 29, 57 },
		{ 114, 30, 57 },
		{ 114, 31, 57 },
		{ 114, 32, 57 },
		{ 114, 33, 57 },
		{ 114, 34, 57 },
		{ 114, 35, 57 },
		{ 114, 36, 57 },
		{ 114, 37, 57 },
		{ 114, 38, 57 },
		{ 114, 39, 57 },
		{ 114, 40, 57 },
		{ 114, 41, 57 },
		{ 114, 42, 57 },
		{ 114, 43, 57 },
		{ 114, 44, 57 },
		{ 114, 45, 57 },
		{ 114, 46, 57 },
		{ 114, 47, 57 },
		{ 114, 48, 57 },
		{ 114, 49, 57 },
		{ 114, 50, 57 },
		{ 114, 51, 57 },
		{ 114, 52, 57 },
		{ 114, 53, 57 },
		{ 114, 54, 57 },
		{ 114, 55, 57 },
		{ 114, 56, 57 },
		{ 114, 57, 57 },
		{ 114, 58, 57 },
		{ 114, 59, 57 },
		{ 114, 60, 57 },
		{ 114, 61, 57 },
		{ 114, 62, 57 },
		{ 114, 63, 57 },
		{ 114, 64, 57 },
		{ 114, 65, 57 },
		{ 114, 66, 57 },
		{ 115, 1, 57 },
		{ 115, 2, 57 },
		{ 115, 3, 57 },
		{ 115, 4, 57 },
		{ 115, 5, 57 },
		{ 115, 6, 57 },
		{ 115, 7, 57 },
		{ 115, 8, 57 },
		{ 115, 9, 57 },
		{ 115, 10, 57 },
		{ 115, 11, 57 },
		{ 115, 12, 57 },
		{ 115, 13, 57 },
		{ 115, 14, 57 },
		{ 115, 15, 57 },
		{ 115, 16, 57 },
		{ 115, 17, 57 },
		{ 115, 18, 57 },
		{ 115, 19, 57 },
		{ 115, 20, 57 },
		{ 115, 21, 57 },
		{ 115, 22, 57 },
		{ 115, 23, 57 },
		{ 115, 24, 57 },
		{ 115, 25, 57 },
		{ 115, 26, 57 },
		{ 115, 27, 57 },
		{ 115, 28, 57 },
		{ 115, 29, 57 },
		{ 115, 30, 57 },
		{ 115, 31, 57 },
		{ 115, 32, 57 },
		{ 115, 33, 57 },
		{ 115, 34, 57 },
		{ 115, 35, 57 },
		{ 115, 36, 57 },
		{ 115, 37, 57 },
		{ 115, 38, 57 },
		{ 115, 39, 57 },
		{ 115, 40, 57 },
		{ 115, 41, 57 },
		{ 115, 42, 57 },
		{ 115, 43, 57 },
		{ 115, 44, 57 },
		{ 115, 45, 57 },
		{ 115, 46, 57 },
		{ 115, 47, 57 },
		{ 115, 48, 57 },
		{ 115, 49, 57 },
		{ 115, 50, 57 },
		{ 115, 51, 57 },
		{ 115, 52, 57 },
		{ 115, 53, 57 },
		{ 115, 54, 57 },
		{ 115, 55, 57 },
		{ 115, 56, 57 },
		{ 115, 57, 57 },
		{ 115, 58, 57 },
		{ 115, 59, 57 },
		{ 115, 60, 57 },
		{ 115, 61, 57 },
		{ 115, 62, 57 },
		{ 115, 63, 57 },
		{ 115, 64, 57 },
		{ 115, 65, 57 },
		{ 115, 66, 57 },
		{ 115, 67, 57 },
		{ 116, 1, 57 },
		{ 116, 2, 57 },
		{ 116, 3, 57 },
		{ 116, 4, 57 },
		{ 116, 5, 57 },
		{ 116, 6, 57 },
		{ 116, 7, 57 },
		{ 116, 8, 57 },
		{ 116, 9, 57 },
		{ 116, 10, 57 },
		{ 116, 11, 57 },
		{ 116, 12, 57 },
		{ 116, 13, 57 },
		{ 116, 14, 57 },
		{ 116, 15, 57 },
		{ 116, 16, 57 },
		{ 116, 17, 57 },
		{ 116, 18, 57 },
		{ 116, 19, 57 },
		{ 116, 20, 57 },
		{ 116, 21, 57 },
		{ 116, 22, 57 },
		{ 116, 23, 57 },
		{ 116, 24, 57 },
		{ 116, 25, 57 },
		{ 116, 26, 57 },
		{ 116, 27, 57 },
		{ 116, 28, 57 },
		{ 116, 29, 57 },
		{ 116, 30, 57 },
		{ 116, 31, 57 },
		{ 116, 32, 57 },
		{ 116, 33, 57 },
		{ 116, 34, 57 },
		{ 116, 35, 57 },
		{ 116, 36, 57 },
		{ 116, 37, 57 },
		{ 116, 38, 57 },
		{ 116, 39, 57 },
		{ 116, 40, 57 },
		{ 116, 41, 57 },
		{ 116, 42, 57 },
		{ 116, 43, 57 },
		{ 116, 44, 57 },
		{ 116, 45, 57 },
		{ 116, 46, 57 },
		{ 116, 47, 57 },
		{ 116, 48, 57 },
		{ 116, 49, 57 },
		{ 116, 50, 57 },
		{ 116, 51, 57 },
		{ 116, 52, 57 },
		{ 116, 53, 57 },
		{ 116, 54, 57 },
		{ 116, 55, 57 },
		{ 116, 56, 57 },
		{ 116, 57, 57 },
		{ 116, 58, 57 },
		{ 116, 59, 57 },
		{ 116, 60, 57 },
		{ 116, 61, 57 },
		{ 116, 62, 57 },
		{ 116, 63, 57 },
		{ 116, 64, 57 },
		{ 116, 65, 57 },
		{ 116, 66, 57 },
		{ 116, 67, 57 },
		{ 116, 68, 57 },
		{ 117, 1, 57 },
		{ 117, 2, 57 },
		{ 117, 3, 57 },
		{ 117, 4, 57 },
		{ 117, 5, 57 },
		{ 117, 6, 57 },
		{ 117, 7, 57 },
		{ 117, 8, 57 },
		{ 117, 9, 57 },
		{ 117, 10, 57 },
		{ 117, 11, 57 },
		{ 117, 12, 57 },
		{ 117, 13, 57 },
		{ 117, 14, 57 },
		{ 117, 15, 57 },
		{ 117, 16, 57 },
		{ 117, 17, 57 },
		{ 117, 18, 57 },
		{ 117, 19, 57 },
		{ 117, 20, 57 },
		{ 117, 21, 57 },
		{ 117, 22, 57 },
		{ 117, 23, 57 },
		{ 117, 24, 57 },
		{ 117, 25, 57 },
		{ 117, 26, 57 },
		{ 117, 27, 57 },
		{ 117, 28, 57 },
		{ 117, 29, 57 },
		{ 117, 30, 57 },
		{ 117, 31, 57 },
		{ 117, 32, 57 },
		{ 117, 33, 57 },
		{ 117, 34, 57 },
		{ 117, 35, 57 },
		{ 117, 36, 57 },
		{ 117, 37, 57 },
		{ 117, 38, 57 },
		{ 117, 39, 57 },
		{ 117, 40, 57 },
		{ 117, 41, 57 },
		{ 117, 42, 57 },
		{ 117, 43, 57 },
		{ 117, 44, 57 },
		{ 117, 45, 57 },
		{ 117, 46, 57 },
		{ 117, 47, 57 },
		{ 117, 48, 57 },
		{ 117, 49, 57 },
		{ 117, 50, 57 },
		{ 117, 51, 57 },
		{ 117, 52, 57 },
		{ 117, 53, 57 },
		{ 117, 54, 57 },
		{ 117, 55, 57 },
		{ 117, 56, 57 },
		{ 117, 57, 57 },
		{ 117, 58, 57 },
		{ 117, 59, 57 },
		{ 117, 60, 57 },
		{ 117, 61, 57 },
		{ 117, 62, 57 },
		{ 117, 63, 57 },
		{ 117, 64, 57 },
		{ 117, 65, 57 },
		{ 117, 66, 57 },
		{ 117, 67, 57 },
		{ 117, 68, 57 },
		{ 117, 69, 57 },
		{ 117, 70, 57 },
		{ 118, 1, 57 },
		{ 118, 2, 57 },
		{ 118, 3, 57 },
		{ 118, 4, 57 },
		{ 118, 5, 57 },
		{ 118, 6, 57 },
		{ 118, 7, 57 },
		{ 118, 8, 57 },
		{ 118, 9, 57 },
		{ 118, 10, 57 },
		{ 118, 11, 57 },
		{ 118, 12, 57 },
		{ 118, 13, 57 },
		{ 118, 14, 57 },
		{ 118, 15, 57 },
		{ 118, 16, 57 },
		{ 118, 17, 57 },
		{ 118, 18, 57 },
		{ 118, 19, 57 },
		{ 118, 20, 57 },
		{ 118, 21, 57 },
		{ 118, 22, 57 },
		{ 118, 23, 57 },
		{ 118, 24, 57 },
		{ 118, 25, 57 },
		{ 118, 26, 57 },
		{ 118, 27, 57 },
		{ 118, 28, 57 },
		{ 118, 29, 57 },
		{ 118, 30, 57 },
		{ 118, 31, 57 },
		{ 118, 32, 57 },
		{ 118, 33, 57 },
		{ 118, 34, 57 },
		{ 118, 35, 57 },
		{ 118, 36, 57 },
		{ 118, 37, 57 },
		{ 118, 38, 57 },
		{ 118, 39, 57 },
		{ 118, 40, 57 },
		{ 118, 41, 57 },
		{ 118, 42, 57 },
		{ 118, 43, 57 },
		{ 118, 44, 57 },
		{ 118, 45, 57 },
		{ 118, 46, 57 },
		{ 118, 47, 57 },
		{ 118, 48, 57 },
		{ 118, 49, 57 },
		{ 118, 50, 57 },
		{ 118, 51, 57 },
		{ 118, 52, 57 },
		{ 118, 53, 57 },
		{ 118, 54, 57 },
		{ 118, 55, 57 },
		{ 118, 56, 57 },
		{ 118, 57, 57 },
		{ 118, 58, 57 },
		{ 118, 59, 57 },
		{ 118, 60, 57 },
		{ 118, 61, 57 },
		{ 118, 62, 57 },
		{ 118, 63, 57 },
		{ 118, 64, 57 },
		{ 118, 65, 57 },
		{ 118, 66, 57 },
		{ 118, 67, 57 },
		{ 118, 68, 57 },
		{ 118, 69, 57 },
		{ 118, 70, 57 },
		{ 118, 71, 57 },
		{ 119, 1, 57 },
		{ 119, 2, 57 },
		{ 119, 3, 57 },
		{ 119, 4, 57 },
		{ 119, 5, 57 },
		{ 119, 6, 57 },
		{ 119, 7, 57 },
		{ 119, 8, 57 },
		{ 119, 9, 57 },
		{ 119, 10, 57 },
		{ 119, 11, 57 },
		{ 119, 12, 57 },
		{ 119, 13, 57 },
		{ 119, 14, 57 },
		{ 119, 15, 57 },
		{ 119, 16, 57 },
		{ 119, 17, 57 },
		{ 119, 18, 57 },
		{ 119, 19, 57 },
		{ 119, 20, 57 },
		{ 119, 21, 57 },
		{ 119, 22, 57 },
		{ 119, 23, 57 },
		{ 119, 24, 57 },
		{ 119, 25, 57 },
		{ 119, 26, 57 },
		{ 119, 27, 57 },
		{ 119, 28, 57 },
		{ 119, 29, 57 },
		{ 119, 30, 57 },
		{ 119, 31, 57 },
		{ 119, 32, 57 },
		{ 119, 33, 57 },
		{ 119, 34, 57 },
		{ 119, 35, 57 },
		{ 119, 36, 57 },
		{ 119, 37, 57 },
		{ 119, 38, 57 },
		{ 119, 39, 57 },
		{ 119, 40, 57 },
		{ 119, 41, 57 },
		{ 119, 42, 57 },
		{ 119, 43, 57 },
		{ 119, 44, 57 },
		{ 119, 45, 57 },
		{ 119, 46, 57 },
		{ 119, 47, 57 },
		{ 119, 48, 57 },
		{ 119, 49, 57 },
		{ 119, 50, 57 },
		{ 119, 51, 57 },
		{ 119, 52, 57 },
		{ 119, 53, 57 },
		{ 119, 54, 57 },
		{ 119, 55, 57 },
		{ 119, 56, 57 },
		{ 119, 57, 57 },
		{ 119, 58, 57 },
		{ 119, 59, 57 },
		{ 119, 60, 57 },
		{ 119, 61, 57 },
		{ 119, 62, 57 },
		{ 119, 63, 57 },
		{ 119, 64, 57 },
		{ 119, 65, 57 },
		{ 119, 66, 57 },
		{ 119, 67, 57 },
		{ 119, 68, 57 },
		{ 119, 69, 57 },
		{ 119, 70, 57 },
		{ 119, 71, 57 },
		{ 119, 72, 57 },
		{ 120, 1, 57 },
		{ 120, 2, 57 },
		{ 120, 3, 57 },
		{ 120, 4, 57 },
		{ 120, 5, 57 },
		{ 120, 6, 57 },
		{ 120, 7, 57 },
		{ 120, 8, 57 },
		{ 120, 9, 57 },
		{ 120, 10, 57 },
		{ 120, 11, 57 },
		{ 120, 12, 57 },
		{ 120, 13, 57 },
		{ 120, 14, 57 },
		{ 120, 15, 57 },
		{ 120, 16, 57 },
		{ 120, 17, 57 },
		{ 120, 18, 57 },
		{ 120, 19, 57 },
		{ 120, 20, 57 },
		{ 120, 21, 57 },
		{ 120, 22, 57 },
		{ 120, 23, 57 },
		{ 120, 24, 57 },
		{ 120, 25, 57 },
		{ 120, 26, 57 },
		{ 120, 27, 57 },
		{ 120, 28, 57 },
		{ 120, 29, 57 },
		{ 120, 30, 57 },
		{ 120, 31, 57 },
		{ 120, 32, 57 },
		{ 120, 33, 57 },
		{ 120, 34, 57 },
		{ 120, 35, 57 },
		{ 120, 36, 57 },
		{ 120, 37, 57 },
		{ 120, 38, 57 },
		{ 120, 39, 57 },
		{ 120, 40, 57 },
		{ 120, 41, 57 },
		{ 120, 42, 57 },
		{ 120, 43, 57 },
		{ 120, 44, 57 },
		{ 120, 45, 57 },
		{ 120, 46, 57 },
		{ 120, 47, 57 },
		{ 120, 48, 57 },
		{ 120, 49, 57 },
		{ 120, 50, 57 },
		{ 120, 51, 57 },
		{ 120, 52, 57 },
		{ 120, 53, 57 },
		{ 120, 54, 57 },
		{ 120, 55, 57 },
		{ 120, 56, 57 },
		{ 120, 57, 57 },
		{ 120, 58, 57 },
		{ 120, 59, 57 },
		{ 120, 60, 57 },
		{ 120, 61, 57 },
		{ 120, 62, 57 },
		{ 120, 63, 57 },
		{ 120, 64, 57 },
		{ 120, 65, 57 },
		{ 120, 66, 57 },
		{ 120, 67, 57 },
		{ 120, 68, 57 },
		{ 120, 69, 57 },
		{ 120, 70, 57 },
		{ 120, 71, 57 },
		{ 120, 72, 57 },
		{ 120, 73, 57 },
		{ 120, 74, 57 },
		{ 121, 1, 57 },
		{ 121, 2, 57 },
		{ 121, 3, 57 },
		{ 121, 4, 57 },
		{ 121, 5, 57 },
		{ 121, 6, 57 },
		{ 121, 7, 57 },
		{ 121, 8, 57 },
		{ 121, 9, 57 },
		{ 121, 10, 57 },
		{ 121, 11, 57 },
		{ 121, 12, 57 },
		{ 121, 13, 57 },
		{ 121, 14, 57 },
		{ 121, 15, 57 },
		{ 121, 16, 57 },
		{ 121, 17, 57 },
		{ 121, 18, 57 },
		{ 121, 19, 57 },
		{ 121, 20, 57 },
		{ 121, 21, 57 },
		{ 121, 22, 57 },
		{ 121, 23, 57 },
		{ 121, 24, 57 },
		{ 121, 25, 57 },
		{ 121, 26, 57 },
		{ 121, 27, 57 },
		{ 121, 28, 57 },
		{ 121, 29, 57 },
		{ 121, 30, 57 },
		{ 121, 31, 57 },
		{ 121, 32, 57 },
		{ 121, 33, 57 },
		{ 121, 34, 57 },
		{ 121, 35, 57 },
		{ 121, 36, 57 },
		{ 121, 37, 57 },
		{ 121, 38, 57 },
		{ 121, 39, 57 },
		{ 121, 40, 57 },
		{ 121, 41, 57 },
		{ 121, 42, 57 },
		{ 121, 43, 57 },
		{ 121, 44, 57 },
		{ 121, 45, 57 },
		{ 121, 46, 57 },
		{ 121, 47, 57 },
		{ 121, 48, 57 },
		{ 121, 49, 57 },
		{ 121, 50, 57 },
		{ 121, 51, 57 },
		{ 121, 52, 57 },
		{ 121, 53, 57 },
		{ 121, 54, 57 },
		{ 121, 55, 57 },
		{ 121, 56, 57 },
		{ 121, 57, 57 },
		{ 121, 58, 57 },
		{ 121, 59, 57 },
		{ 121, 60, 57 },
		{ 121, 61, 57 },
		{ 121, 62, 57 },
		{ 121, 63, 57 },
		{ 121, 64, 57 },
		{ 121, 65, 57 },
		{ 121, 66, 57 },
		{ 121, 67, 57 },
		{ 121, 68, 57 },
		{ 121, 69, 57 },
		{ 121, 70, 57 },
		{ 121, 71, 57 },
		{ 121, 72, 57 },
		{ 121, 73, 57 },
		{ 121, 74, 57 },
		{ 121, 75, 57 },
		{ 122, 1, 57 },
		{ 122, 2, 57 },
		{ 122, 3, 57 },
		{ 122, 4, 57 },
		{ 122, 5, 57 },
		{ 122, 6, 57 },
		{ 122, 7, 57 },
		{ 122, 8, 57 },
		{ 122, 9, 57 },
		{ 122, 10, 57 },
		{ 122, 11, 57 },
		{ 122, 12, 57 },
		{ 122, 13, 57 },
		{ 122, 14, 57 },
		{ 122, 15, 57 },
		{ 122, 16, 57 },
		{ 122, 17, 57 },
		{ 122, 18, 57 },
		{ 122, 19, 57 },
		{ 122, 20, 57 },
		{ 122, 21, 57 },
		{ 122, 22, 57 },
		{ 122, 23, 57 },
		{ 122, 24, 57 },
		{ 122, 25, 57 },
		{ 122, 26, 57 },
		{ 122, 27, 57 },
		{ 122, 28, 57 },
		{ 122, 29, 57 },
		{ 122, 30, 57 },
		{ 122, 31, 57 },
		{ 122, 32, 57 },
		{ 122, 33, 57 },
		{ 122, 34, 57 },
		{ 122, 35, 57 },
		{ 122, 36, 57 },
		{ 122, 37, 57 },
		{ 122, 38, 57 },
		{ 122, 39, 57 },
		{ 122, 40, 57 },
		{ 122, 41, 57 },
		{ 122, 42, 57 },
		{ 122, 43, 57 },
		{ 122, 44, 57 },
		{ 122, 45, 57 },
		{ 122, 46, 57 },
		{ 122, 47, 57 },
		{ 122, 48, 57 },
		{ 122, 49, 57 },
		{ 122, 50, 57 },
		{ 122, 51, 57 },
		{ 122, 52, 57 },
		{ 122, 53, 57 },
		{ 122, 54, 57 },
		{ 122, 55, 57 },
		{ 122, 56, 57 },
		{ 122, 57, 57 },
		{ 122, 58, 57 },
		{ 122, 59, 57 },
		{ 122, 60, 57 },
		{ 122, 61, 57 },
		{ 122, 62, 57 },
		{ 122, 63, 57 },
		{ 122, 64, 57 },
		{ 122, 65, 57 },
		{ 122, 66, 57 },
		{ 122, 67, 57 },
		{ 122, 68, 57 },
		{ 122, 69, 57 },
		{ 122, 70, 57 },
		{ 122, 71, 57 },
		{ 122, 72, 57 },
		{ 122, 73, 57 },
		{ 122, 74, 57 },
		{ 122, 75, 57 },
		{ 122, 76, 57 },
		{ 123, 1, 57 },
		{ 123, 2, 57 },
		{ 123, 3, 57 },
		{ 123, 4, 57 },
		{ 123, 5, 57 },
		{ 123, 6, 57 },
		{ 123, 7, 57 },
		{ 123, 8, 57 },
		{ 123, 9, 57 },
		{ 123, 10, 57 },
		{ 123, 11, 57 },
		{ 123, 12, 57 },
		{ 123, 13, 57 },
		{ 123, 14, 57 },
		{ 123, 15, 57 },
		{ 123, 16, 57 },
		{ 123, 17, 57 },
		{ 123, 18, 57 },
		{ 123, 19, 57 },
		{ 123, 20, 57 },
		{ 123, 21, 57 },
		{ 123, 22, 57 },
		{ 123, 23, 57 },
		{ 123, 24, 57 },
		{ 123, 25, 57 },
		{ 123, 26, 57 },
		{ 123, 27, 57 },
		{ 123, 28, 57 },
		{ 123, 29, 57 },
		{ 123, 30, 57 },
		{ 123, 31, 57 },
		{ 123, 32, 57 },
		{ 123, 33, 57 },
		{ 123, 34, 57 },
		{ 123, 35, 57 },
		{ 123, 36, 57 },
		{ 123, 37, 57 },
		{ 123, 38, 57 },
		{ 123, 39, 57 },
		{ 123, 40, 57 },
		{ 123, 41, 57 },
		{ 123, 42, 57 },
		{ 123, 43, 57 },
		{ 123, 44, 57 },
		{ 123, 45, 57 },
		{ 123, 46, 57 },
		{ 123, 47, 57 },
		{ 123, 48, 57 },
		{ 123, 49, 57 },
		{ 123, 50, 57 },
		{ 123, 51, 57 },
		{ 123, 52, 57 },
		{ 123, 53, 57 },
		{ 123, 54, 57 },
		{ 123, 55, 57 },
		{ 123, 56, 57 },
		{ 123, 57, 57 },
		{ 123, 58, 57 },
		{ 123, 59, 57 },
		{ 123, 60, 57 },
		{ 123, 61, 57 },
		{ 123, 62, 57 },
		{ 123, 63, 57 },
		{ 123, 64, 57 },
		{ 123, 65, 57 },
		{ 123, 66, 57 },
		{ 123, 67, 57 },
		{ 123, 68, 57 },
		{ 123, 69, 57 },
		{ 123, 70, 57 },
		{ 123, 71, 57 },
		{ 123, 72, 57 },
		{ 123, 73, 57 },
		{ 123, 74, 57 },
		{ 123, 75, 57 },
		{ 123, 76, 57 },
		{ 123, 77, 57 },
		{ 123, 78, 57 },
		{ 124, 1, 57 },
		{ 124, 2, 57 },
		{ 124, 3, 57 },
		{ 124, 4, 57 },
		{ 124, 5, 57 },
		{ 124, 6, 57 },
		{ 124, 7, 57 },
		{ 124, 8, 57 },
		{ 124, 9, 57 },
		{ 124, 10, 57 },
		{ 124, 11, 57 },
		{ 124, 12, 57 },
		{ 124, 13, 57 },
		{ 124, 14, 57 },
		{ 124, 15, 57 },
		{ 124, 16, 57 },
		{ 124, 17, 57 },
		{ 124, 18, 57 },
		{ 124, 19, 57 },
		{ 124, 20, 57 },
		{ 124, 21, 57 },
		{ 124, 22, 57 },
		{ 124, 23, 57 },
		{ 124, 24, 57 },
		{ 124, 25, 57 },
		{ 124, 26, 57 },
		{ 124, 27, 57 },
		{ 124, 28, 57 },
		{ 124, 29, 57 },
		{ 124, 30, 57 },
		{ 124, 31, 57 },
		{ 124, 32, 57 },
		{ 124, 33, 57 },
		{ 124, 34, 57 },
		{ 124, 35, 57 },
		{ 124, 36, 57 },
		{ 124, 37, 57 },
		{ 124, 38, 57 },
		{ 124, 39, 57 },
		{ 124, 40, 57 },
		{ 124, 41, 57 },
		{ 124, 42, 57 },
		{ 124, 43, 57 },
		{ 124, 44, 57 },
		{ 124, 45, 57 },
		{ 124, 46, 57 },
		{ 124, 47, 57 },
		{ 124, 48, 57 },
		{ 124, 49, 57 },
		{ 124, 50, 57 },
		{ 124, 51, 57 },
		{ 124, 52, 57 },
		{ 124, 53, 57 },
		{ 124, 54, 57 },
		{ 124, 55, 57 },
		{ 124, 56, 57 },
		{ 124, 57, 57 },
		{ 124, 58, 57 },
		{ 124, 59, 57 },
		{ 124, 60, 57 },
		{ 124, 61, 57 },
		{ 124, 62, 57 },
		{ 124, 63, 57 },
		{ 124, 64, 57 },
		{ 124, 65, 57 },
		{ 124, 66, 57 },
		{ 124, 67, 57 },
		{ 124, 68, 57 },
		{ 124, 69, 57 },
		{ 124, 70, 57 },
		{ 124, 71, 57 },
		{ 124, 72, 57 },
		{ 124, 73, 57 },
		{ 124, 74, 57 },
		{ 124, 75, 57 },
		{ 124, 76, 57 },
		{ 124, 77, 57 },
		{ 124, 78, 57 },
		{ 124, 79, 57 },
		{ 125, 1, 57 },
		{ 125, 2, 57 },
		{ 125, 3, 57 },
		{ 125, 4, 57 },
		{ 125, 5, 57 },
		{ 125, 6, 57 },
		{ 125, 7, 57 },
		{ 125, 8, 57 },
		{ 125, 9, 57 },
		{ 125, 10, 57 },
		{ 125, 11, 57 },
		{ 125, 12, 57 },
		{ 125, 13, 57 },
		{ 125, 14, 57 },
		{ 125, 15, 57 },
		{ 125, 16, 57 },
		{ 125, 17, 57 },
		{ 125, 18, 57 },
		{ 125, 19, 57 },
		{ 125, 20, 57 },
		{ 125, 21, 57 },
		{ 125, 22, 57 },
		{ 125, 23, 57 },
		{ 125, 24, 57 },
		{ 125, 25, 57 },
		{ 125, 26, 57 },
		{ 125, 27, 57 },
		{ 125, 28, 57 },
		{ 125, 29, 57 },
		{ 125, 30, 57 },
		{ 125, 31, 57 },
		{ 125, 32, 57 },
		{ 125, 33, 57 },
		{ 125, 34, 57 },
		{ 125, 35, 57 },
		{ 125, 36, 57 },
		{ 125, 37, 57 },
		{ 125, 38, 57 },
		{ 125, 39, 57 },
		{ 125, 40, 57 },
		{ 125, 41, 57 },
		{ 125, 42, 57 },
		{ 125, 43, 57 },
		{ 125, 44, 57 },
		{ 125, 45, 57 },
		{ 125, 46, 57 },
		{ 125, 47, 57 },
		{ 125, 48, 57 },
		{ 125, 49, 57 },
		{ 125, 50, 57 },
		{ 125, 51, 57 },
		{ 125, 52, 57 },
		{ 125, 53, 57 },
		{ 125, 54, 57 },
		{ 125, 55, 57 },
		{ 125, 56, 57 },
		{ 125, 57, 57 },
		{ 125, 58, 57 },
		{ 125, 59, 57 },
		{ 125, 60, 57 },
		{ 125, 61, 57 },
		{ 125, 62, 57 },
		{ 125, 63, 57 },
		{ 125, 64, 57 },
		{ 125, 65, 57 },
		{ 125, 66, 57 },
		{ 125, 67, 57 },
		{ 125, 68, 57 },
		{ 125, 69, 57 },
		{ 125, 70, 57 },
		{ 125, 71, 57 },
		{ 125, 72, 57 },
		{ 125, 73, 57 },
		{ 125, 74, 57 },
		{ 125, 75, 57 },
		{ 125, 76, 57 },
		{ 125, 77, 57 },
		{ 125, 78, 57 },
		{ 125, 79, 57 },
		{ 125, 80, 57 },
		{ 126, 1, 57 },
		{ 126, 2, 57 },
		{ 126, 3, 57 },
		{ 126, 4, 57 },
		{ 126, 5, 57 },
		{ 126, 6, 57 },
		{ 126, 7, 57 },
		{ 126, 8, 57 },
		{ 126, 9, 57 },
		{ 126, 10, 57 },
		{ 126, 11, 57 },
		{ 126, 12, 57 },
		{ 126, 13, 57 },
		{ 126, 14, 57 },
		{ 126, 15, 57 },
		{ 126, 16, 57 },
		{ 126, 17, 57 },
		{ 126, 18, 57 },
		{ 126, 19, 57 },
		{ 126, 20, 57 },
		{ 126, 21, 57 },
		{ 126, 22, 57 },
		{ 126, 23, 57 },
		{ 126, 24, 57 },
		{ 126, 25, 57 },
		{ 126, 26, 57 },
		{ 126, 27, 57 },
		{ 126, 28, 57 },
		{ 126, 29, 57 },
		{ 126, 30, 57 },
		{ 126, 31, 57 },
		{ 126, 32, 57 },
		{ 126, 33, 57 },
		{ 126, 34, 57 },
		{ 126, 35, 57 },
		{ 126, 36, 57 },
		{ 126, 37, 57 },
		{ 126, 38, 57 },
		{ 126, 39, 57 },
		{ 126, 40, 57 },
		{ 126, 41, 57 },
		{ 126, 42, 57 },
		{ 126, 43, 57 },
		{ 126, 44, 57 },
		{ 126, 45, 57 },
		{ 126, 46, 57 },
		{ 126, 47, 57 },
		{ 126, 48, 57 },
		{ 126, 49, 57 },
		{ 126, 50, 57 },
		{ 126, 51, 57 },
		{ 126, 52, 57 },
		{ 126, 53, 57 },
		{ 126, 54, 57 },
		{ 126, 55, 57 },
		{ 126, 56, 57 },
		{ 126, 57, 57 },
		{ 126, 58, 57 },
		{ 126, 59, 57 },
		{ 126, 60, 57 },
		{ 126, 61, 57 },
		{ 126, 62, 57 },
		{ 126, 63, 57 },
		{ 126, 64, 57 },
		{ 126, 65, 57 },
		{ 126, 66, 57 },
		{ 126, 67, 57 },
		{ 126, 68, 57 },
		{ 126, 69, 57 },
		{ 126, 70, 57 },
		{ 126, 71, 57 },
		{ 126, 72, 57 },
		{ 126, 73, 57 },
		{ 126, 74, 57 },
		{ 126, 75, 57 },
		{ 126, 76, 57 },
		{ 126, 77, 57 },
		{ 126, 78, 57 },
		{ 126, 79, 57 },
		{ 126, 80, 57 },
		{ 126, 81, 57 },
		{ 126, 82, 57 },
		{ 127, 1, 57 },
		{ 127, 2, 57 },
		{ 127, 3, 57 },
		{ 127, 4, 57 },
		{ 127, 5, 57 },
		{ 127, 6, 57 },
		{ 127, 7, 57 },
		{ 127, 8, 57 },
		{ 127, 9, 57 },
		{ 127, 10, 57 },
		{ 127, 11, 57 },
		{ 127, 12, 57 },
		{ 127, 13, 57 },
		{ 127, 14, 57 },
		{ 127, 15, 57 },
		{ 127, 16, 57 },
		{ 127, 17, 57 },
		{ 127, 18, 57 },
		{ 127, 19, 57 },
		{ 127, 20, 57 },
		{ 127, 21, 57 },
		{ 127, 22, 57 },
		{ 127, 23, 57 },
		{ 127, 24, 57 },
		{ 127, 25, 57 },
		{ 127, 26, 57 },
		{ 127, 27, 57 },
		{ 127, 28, 57 },
		{ 127, 29, 57 },
		{ 127, 30, 57 },
		{ 127, 31, 57 },
		{ 127, 32, 57 },
		{ 127, 33, 57 },
		{ 127, 34, 57 },
		{ 127, 35, 57 },
		{ 127, 36, 57 },
		{ 127, 37, 57 },
		{ 127, 38, 57 },
		{ 127, 39, 57 },
		{ 127, 40, 57 },
		{ 127, 41, 57 },
		{ 127, 42, 57 },
		{ 127, 43, 57 },
		{ 127, 44, 57 },
		{ 127, 45, 57 },
		{ 127, 46, 57 },
		{ 127, 47, 57 },
		{ 127, 48, 57 },
		{ 127, 49, 57 },
		{ 127, 50, 57 },
		{ 127, 51, 57 },
		{ 127, 52, 57 },
		{ 127, 53, 57 },
		{ 127, 54, 57 },
		{ 127, 55, 57 },
		{ 127, 56, 57 },
		{ 127, 57, 57 },
		{ 127, 58, 57 },
		{ 127, 59, 57 },
		{ 127, 60, 57 },
		{ 127, 61, 57 },
		{ 127, 62, 57 },
		{ 127, 63, 57 },
		{ 127, 64, 57 },
		{ 127, 65, 57 },
		{ 127, 66, 57 },
		{ 127, 67, 57 },
		{ 127, 68, 57 },
		{ 127, 69, 57 },
		{ 127, 70, 57 },
		{ 127, 71, 57 },
		{ 127, 72, 57 },
		{ 127, 73, 57 },
		{ 127, 74, 57 },
		{ 127, 75, 57 },
		{ 127, 76, 57 },
		{ 127, 77, 57 },
		{ 127, 78, 57 },
		{ 127, 79, 57 },
		{ 127, 80, 57 },
		{ 127, 81, 57 },
		{ 127, 82, 57 },
		{ 127, 83, 57 },
		{ 128, 1, 57 },
		{ 128, 2, 57 },
		{ 128, 3, 57 },
		{ 128, 4, 57 },
		{ 128, 5, 57 },
		{ 128, 6, 57 },
		{ 128, 7, 57 },
		{ 128, 8, 57 },
		{ 128, 9, 57 },
		{ 128, 10, 57 },
		{ 128, 11, 57 },
		{ 128, 12, 57 },
		{ 128, 13, 57 },
		{ 128, 14, 57 },
		{ 128, 15, 57 },
		{ 128, 16, 57 },
		{ 128, 17, 57 },
		{ 128, 18, 57 },
		{ 128, 19, 57 },
		{ 128, 20, 57 },
		{ 128, 21, 57 },
		{ 128, 22, 57 },
		{ 128, 23, 57 },
		{ 128, 24, 57 },
		{ 128, 25, 57 },
		{ 128, 26, 57 },
		{ 128, 27, 57 },
		{ 128, 28, 57 },
		{ 128, 29, 57 },
		{ 128, 30, 57 },
		{ 128, 31, 57 },
		{ 128, 32, 57 },
		{ 128, 33, 57 },
		{ 128, 34, 57 },
		{ 128, 35, 57 },
		{ 128, 36, 57 },
		{ 128, 37, 57 },
		{ 128, 38, 57 },
		{ 128, 39, 57 },
		{ 128, 40, 57 },
		{ 128, 41, 57 },
		{ 128, 42, 57 },
		{ 128, 43, 57 },
		{ 128, 44, 57 },
		{ 128, 45, 57 },
		{ 128, 46, 57 },
		{ 128, 47, 57 },
		{ 128, 48, 57 },
		{ 128, 49, 57 },
		{ 128, 50, 57 },
		{ 128, 51, 57 },
		{ 128, 52, 57 },
		{ 128, 53, 57 },
		{ 128, 54, 57 },
		{ 128, 55, 57 },
		{ 128, 56, 57 },
		{ 128, 57, 57 },
		{ 128, 58, 57 },
		{ 128, 59, 57 },
		{ 128, 60, 57 },
		{ 128, 61, 57 },
		{ 128, 62, 57 },
		{ 128, 63, 57 },
		{ 128, 64, 57 },
		{ 128, 65, 57 },
		{ 128, 66, 57 },
		{ 128, 67, 57 },
		{ 128, 68, 57 },
		{ 128, 69, 57 },
		{ 128, 70, 57 },
		{ 128, 71, 57 },
		{ 128, 72, 57 },
		{ 128, 73, 57 },
		{ 128, 74, 57 },
		{ 128, 75, 57 },
		{ 128, 76, 57 },
		{ 128, 77, 57 },
		{ 128, 78, 57 },
		{ 128, 79, 57 },
		{ 128, 80, 57 },
		{ 128, 81, 57 },
		{ 128, 82, 57 },
		{ 128, 83, 57 },
		{ 128, 84, 57 },
		{ 129, 2, 57 },
		{ 129, 3, 57 },
		{ 129, 4, 57 },
		{ 129, 5, 57 },
		{ 129, 6, 57 },
		{ 129, 7, 57 },
		{ 129, 8, 57 },
		{ 129, 9, 57 },
		{ 129, 10, 57 },
		{ 129, 11, 57 },
		{ 129, 12, 57 },
		{ 129, 13, 57 },
		{ 129, 14, 57 },
		{ 129, 15, 57 },
		{ 129, 16, 57 },
		{ 129, 17, 57 },
		{ 129, 18, 57 },
		{ 129, 19, 57 },
		{ 129, 20, 57 },
		{ 129, 21, 57 },
		{ 129, 22, 57 },
		{ 129, 23, 57 },
		{ 129, 24, 57 },
		{ 129, 25, 57 },
		{ 129, 26, 57 },
		{ 129, 27, 57 },
		{ 129, 28, 57 },
		{ 129, 29, 57 },
		{ 129, 30, 57 },
		{ 129, 31, 57 },
		{ 129, 32, 57 },
		{ 129, 33, 57 },
		{ 129, 34, 57 },
		{ 129, 35, 57 },
		{ 129, 36, 57 },
		{ 129, 37, 57 },
		{ 129, 38, 57 },
		{ 129, 39, 57 },
		{ 129, 40, 57 },
		{ 129, 41, 57 },
		{ 129, 42, 57 },
		{ 129, 43, 57 },
		{ 129, 44, 57 },
		{ 129, 45, 57 },
		{ 129, 46, 57 },
		{ 129, 47, 57 },
		{ 129, 48, 57 },
		{ 129, 49, 57 },
		{ 129, 50, 57 },
		{ 129, 51, 57 },
		{ 129, 52, 57 },
		{ 129, 53, 57 },
		{ 129, 54, 57 },
		{ 129, 55, 57 },
		{ 129, 56, 57 },
		{ 129, 57, 57 },
		{ 129, 58, 57 },
		{ 129, 59, 57 },
		{ 129, 60, 57 },
		{ 129, 61, 57 },
		{ 129, 62, 57 },
		{ 129, 63, 57 },
		{ 129, 64, 57 },
		{ 129, 65, 57 },
		{ 129, 66, 57 },
		{ 129, 67, 57 },
		{ 129, 68, 57 },
		{ 129, 69, 57 },
		{ 129, 70, 57 },
		{ 129, 71, 57 },
		{ 129, 72, 57 },
		{ 129, 73, 57 },
		{ 129, 74, 57 },
		{ 129, 75, 57 },
		{ 129, 76, 57 },
		{ 129, 77, 57 },
		{ 129, 78, 57 },
		{ 129, 79, 57 },
		{ 129, 80, 57 },
		{ 129, 81, 57 },
		{ 129, 82, 57 },
		{ 129, 83, 57 },
		{ 129, 84, 57 },
		{ 129, 85, 57 },
		{ 129, 86, 57 },
		{ 130, 4, 57 },
		{ 130, 5, 57 },
		{ 130, 6, 57 },
		{ 130, 7, 57 },
		{ 130, 8, 57 },
		{ 130, 9, 57 },
		{ 130, 10, 57 },
		{ 130, 11, 57 },
		{ 130, 12, 57 },
		{ 130, 13, 57 },
		{ 130, 14, 57 },
		{ 130, 15, 57 },
		{ 130, 16, 57 },
		{ 130, 17, 57 },
		{ 130, 18, 57 },
		{ 130, 19, 57 },
		{ 130, 20, 57 },
		{ 130, 21, 57 },
		{ 130, 22, 57 },
		{ 130, 23, 57 },
		{ 130, 24, 57 },
		{ 130, 25, 57 },
		{ 130, 26, 57 },
		{ 130, 27, 57 },
		{ 130, 28, 57 },
		{ 130, 29, 57 },
		{ 130, 30, 57 },
		{ 130, 31, 57 },
		{ 130, 32, 57 },
		{ 130, 33, 57 },
		{ 130, 34, 57 },
		{ 130, 35, 57 },
		{ 130, 36, 57 },
		{ 130, 37, 57 },
		{ 130, 38, 57 },
		{ 130, 39, 57 },
		{ 130, 40, 57 },
		{ 130, 41, 57 },
		{ 130, 42, 57 },
		{ 130, 43, 57 },
		{ 130, 44, 57 },
		{ 130, 45, 57 },
		{ 130, 46, 57 },
		{ 130, 47, 57 },
		{ 130, 48, 57 },
		{ 130, 49, 57 },
		{ 130, 50, 57 },
		{ 130, 51, 57 },
		{ 130, 52, 57 },
		{ 130, 53, 57 },
		{ 130, 54, 57 },
		{ 130, 55, 57 },
		{ 130, 56, 57 },
		{ 130, 57, 57 },
		{ 130, 58, 57 },
		{ 130, 59, 57 },
		{ 130, 60, 57 },
		{ 130, 61, 57 },
		{ 130, 62, 57 },
		{ 130, 63, 57 },
		{ 130, 64, 57 },
		{ 130, 65, 57 },
		{ 130, 66, 57 },
		{ 130, 67, 57 },
		{ 130, 68, 57 },
		{ 130, 69, 57 },
		{ 130, 70, 57 },
		{ 130, 71, 57 },
		{ 130, 72, 57 },
		{ 130, 73, 57 },
		{ 130, 74, 57 },
		{ 130, 75, 57 },
		{ 130, 76, 57 },
		{ 130, 77, 57 },
		{ 130, 78, 57 },
		{ 130, 79, 57 },
		{ 130, 80, 57 },
		{ 130, 81, 57 },
		{ 130, 82, 57 },
		{ 130, 83, 57 },
		{ 130, 84, 57 },
		{ 130, 85, 57 },
		{ 130, 86, 57 },
		{ 130, 87, 57 },
		{ 131, 6, 57 },
		{ 131, 7, 57 },
		{ 131, 8, 57 },
		{ 131, 9, 57 },
		{ 131, 10, 57 },
		{ 131, 11, 57 },
		{ 131, 12, 57 },
		{ 131, 13, 57 },
		{ 131, 14, 57 },
		{ 131, 15, 57 },
		{ 131, 16, 57 },
		{ 131, 17, 57 },
		{ 131, 18, 57 },
		{ 131, 19, 57 },
		{ 131, 20, 57 },
		{ 131, 21, 57 },
		{ 131, 22, 57 },
		{ 131, 23, 57 },
		{ 131, 24, 57 },
		{ 131, 25, 57 },
		{ 131, 26, 57 },
		{ 131, 27, 57 },
		{ 131, 28, 57 },
		{ 131, 29, 57 },
		{ 131, 30, 57 },
		{ 131, 31, 57 },
		{ 131, 32, 57 },
		{ 131, 33, 57 },
		{ 131, 34, 57 },
		{ 131, 35, 57 },
		{ 131, 36, 57 },
		{ 131, 37, 57 },
		{ 131, 38, 57 },
		{ 131, 39, 57 },
		{ 131, 40, 57 },
		{ 131, 41, 57 },
		{ 131, 42, 57 },
		{ 131, 43, 57 },
		{ 131, 44, 57 },
		{ 131, 45, 57 },
		{ 131, 46, 57 },
		{ 131, 47, 57 },
		{ 131, 48, 57 },
		{ 131, 49, 57 },
		{ 131, 50, 57 },
		{ 131, 51, 57 },
		{ 131, 52, 57 },
		{ 131, 53, 57 },
		{ 131, 54, 57 },
		{ 131, 55, 57 },
		{ 131, 56, 57 },
		{ 131, 57, 57 },
		{ 131, 58, 57 },
		{ 131, 59, 57 },
		{ 131, 60, 57 },
		{ 131, 61, 57 },
		{ 131, 62, 57 },
		{ 131, 63, 57 },
		{ 131, 64, 57 },
		{ 131, 65, 57 },
		{ 131, 66, 57 },
		{ 131, 67, 57 },
		{ 131, 68, 57 },
		{ 131, 69, 57 },
		{ 131, 70, 57 },
		{ 131, 71, 57 },
		{ 131, 72, 57 },
		{ 131, 73, 57 },
		{ 131, 74, 57 },
		{ 131, 75, 57 },
		{ 131, 76, 57 },
		{ 131, 77, 57 },
		{ 131, 78, 57 },
		{ 131, 79, 57 },
		{ 131, 80, 57 },
		{ 131, 81, 57 },
		{ 131, 82, 57 },
		{ 131, 83, 57 },
		{ 131, 84, 57 },
		{ 131, 85, 57 },
		{ 131, 86, 57 },
		{ 131, 87, 57 },
		{ 131, 88, 57 },
		{ 132, 8, 57 },
		{ 132, 9, 57 },
		{ 132, 10, 57 },
		{ 132, 11, 57 },
		{ 132, 12, 57 },
		{ 132, 13, 57 },
		{ 132, 14, 57 },
		{ 132, 15, 57 },
		{ 132, 16, 57 },
		{ 132, 17, 57 },
		{ 132, 18, 57 },
		{ 132, 19, 57 },
		{ 132, 20, 57 },
		{ 132, 21, 57 },
		{ 132, 22, 57 },
		{ 132, 23, 57 },
		{ 132, 24, 57 },
		{ 132, 25, 57 },
		{ 132, 26, 57 },
		{ 132, 27, 57 },
		{ 132, 28, 57 },
		{ 132, 29, 57 },
		{ 132, 30, 57 },
		{ 132, 31, 57 },
		{ 132, 32, 57 },
		{ 132, 33, 57 },
		{ 132, 34, 57 },
		{ 132, 35, 57 },
		{ 132, 36, 57 },
		{ 132, 37, 57 },
		{ 132, 38, 57 },
		{ 132, 39, 57 },
		{ 132, 40, 57 },
		{ 132, 41, 57 },
		{ 132, 42, 57 },
		{ 132, 43, 57 },
		{ 132, 44, 57 },
		{ 132, 45, 57 },
		{ 132, 46, 57 },
		{ 132, 47, 57 },
		{ 132, 48, 57 },
		{ 132, 49, 57 },
		{ 132, 50, 57 },
		{ 132, 51, 57 },
		{ 132, 52, 57 },
		{ 132, 53, 57 },
		{ 132, 54, 57 },
		{ 132, 55, 57 },
		{ 132, 56, 57 },
		{ 132, 57, 57 },
		{ 132, 58, 57 },
		{ 132, 59, 57 },
		{ 132, 60, 57 },
		{ 132, 61, 57 },
		{ 132, 62, 57 },
		{ 132, 63, 57 },
		{ 132, 64, 57 },
		{ 132, 65, 57 },
		{ 132, 66, 57 },
		{ 132, 67, 57 },
		{ 132, 68, 57 },
		{ 132, 69, 57 },
		{ 132, 70, 57 },
		{ 132, 71, 57 },
		{ 132, 72, 57 },
		{ 132, 73, 57 },
		{ 132, 74, 57 },
		{ 132, 75, 57 },
		{ 132, 76, 57 },
		{ 132, 77, 57 },
		{ 132, 78, 57 },
		{ 132, 79, 57 },
		{ 132, 80, 57 },
		{ 132, 81, 57 },
		{ 132, 82, 57 },
		{ 132, 83, 57 },
		{ 132, 84, 57 },
		{ 132, 85, 57 },
		{ 132, 86, 57 },
		{ 132, 87, 57 },
		{ 132, 88, 57 },
		{ 132, 89, 57 },
		{ 132, 90, 57 },
		{ 133, 10, 57 },
		{ 133, 11, 57 },
		{ 133, 12, 57 },
		{ 133, 13, 57 },
		{ 133, 14, 57 },
		{ 133, 15, 57 },
		{ 133, 16, 57 },
		{ 133, 17, 57 },
		{ 133, 18, 57 },
		{ 133, 19, 57 },
		{ 133, 20, 57 },
		{ 133, 21, 57 },
		{ 133, 22, 57 },
		{ 133, 23, 57 },
		{ 133, 24, 57 },
		{ 133, 25, 57 },
		{ 133, 26, 57 },
		{ 133, 27, 57 },
		{ 133, 28, 57 },
		{ 133, 29, 57 },
		{ 133, 30, 57 },
		{ 133, 31, 57 },
		{ 133, 32, 57 },
		{ 133, 33, 57 },
		{ 133, 34, 57 },
		{ 133, 35, 57 },
		{ 133, 36, 57 },
		{ 133, 37, 57 },
		{ 133, 38, 57 },
		{ 133, 39, 57 },
		{ 133, 40, 57 },
		{ 133, 41, 57 },
		{ 133, 42, 57 },
		{ 133, 43, 57 },
		{ 133, 44, 57 },
		{ 133, 45, 57 },
		{ 133, 46, 57 },
		{ 133, 47, 57 },
		{ 133, 48, 57 },
		{ 133, 49, 57 },
		{ 133, 50, 57 },
		{ 133, 51, 57 },
		{ 133, 52, 57 },
		{ 133, 53, 57 },
		{ 133, 54, 57 },
		{ 133, 55, 57 },
		{ 133, 56, 57 },
		{ 133, 57, 57 },
		{ 133, 58, 57 },
		{ 133, 59, 57 },
		{ 133, 60, 57 },
		{ 133, 61, 57 },
		{ 133, 62, 57 },
		{ 133, 63, 57 },
		{ 133, 64, 57 },
		{ 133, 65, 57 },
		{ 133, 66, 57 },
		{ 133, 67, 57 },
		{ 133, 68, 57 },
		{ 133, 69, 57 },
		{ 133, 70, 57 },
		{ 133, 71, 57 },
		{ 133, 72, 57 },
		{ 133, 73, 57 },
		{ 133, 74, 57 },
		{ 133, 75, 57 },
		{ 133, 76, 57 },
		{ 133, 77, 57 },
		{ 133, 78, 57 },
		{ 133, 79, 57 },
		{ 133, 80, 57 },
		{ 133, 81, 57 },
		{ 133, 82, 57 },
		{ 133, 83, 57 },
		{ 133, 84, 57 },
		{ 133, 85, 57 },
		{ 133, 86, 57 },
		{ 133, 87, 57 },
		{ 133, 88, 57 },
		{ 133, 89, 57 },
		{ 133, 90, 57 },
		{ 133, 91, 57 },
		{ 134, 12, 57 },
		{ 134, 13, 57 },
		{ 134, 14, 57 },
		{ 134, 15, 57 },
		{ 134, 16, 57 },
		{ 134, 17, 57 },
		{ 134, 18, 57 },
		{ 134, 19, 57 },
		{ 134, 20, 57 },
		{ 134, 21, 57 },
		{ 134, 22, 57 },
		{ 134, 23, 57 },
		{ 134, 24, 57 },
		{ 134, 25, 57 },
		{ 134, 26, 57 },
		{ 134, 27, 57 },
		{ 134, 28, 57 },
		{ 134, 29, 57 },
		{ 134, 30, 57 },
		{ 134, 31, 57 },
		{ 134, 32, 57 },
		{ 134, 33, 57 },
		{ 134, 34, 57 },
		{ 134, 35, 57 },
		{ 134, 36, 57 },
		{ 134, 37, 57 },
		{ 134, 38, 57 },
		{ 134, 39, 57 },
		{ 134, 40, 57 },
		{ 134, 41, 57 },
		{ 134, 42, 57 },
		{ 134, 43, 57 },
		{ 134, 44, 57 },
		{ 134, 45, 57 },
		{ 134, 46, 57 },
		{ 134, 47, 57 },
		{ 134, 48, 57 },
		{ 134, 49, 57 },
		{ 134, 50, 57 },
		{ 134, 51, 57 },
		{ 134, 52, 57 },
		{ 134, 53, 57 },
		{ 134, 54, 57 },
		{ 134, 55, 57 },
		{ 134, 56, 57 },
		{ 134, 57, 57 },
		{ 134, 58, 57 },
		{ 134, 59, 57 },
		{ 134, 60, 57 },
		{ 134, 61, 57 },
		{ 134, 62, 57 },
		{ 134, 63, 57 },
		{ 134, 64, 57 },
		{ 134, 65, 57 },
		{ 134, 66, 57 },
		{ 134, 67, 57 },
		{ 134, 68, 57 },
		{ 134, 69, 57 },
		{ 134, 70, 57 },
		{ 134, 71, 57 },
		{ 134, 72, 57 },
		{ 134, 73, 57 },
		{ 134, 74, 57 },
		{ 134, 75, 57 },
		{ 134, 76, 57 },
		{ 134, 77, 57 },
		{ 134, 78, 57 },
		{ 134, 79, 57 },
		{ 134, 80, 57 },
		{ 134, 81, 57 },
		{ 134, 82, 57 },
		{ 134, 83, 57 },
		{ 134, 84, 57 },
		{ 134, 85, 57 },
		{ 134, 86, 57 },
		{ 134, 87, 57 },
		{ 134, 88, 57 },
		{ 134, 89, 57 },
		{ 134, 90, 57 },
		{ 134, 91, 57 },
		{ 134, 92, 57 },
		{ 135, 14, 57 },
		{ 135, 15, 57 },
		{ 135, 16, 57 },
		{ 135, 17, 57 },
		{ 135, 18, 57 },
		{ 135, 19, 57 },
		{ 135, 20, 57 },
		{ 135, 21, 57 },
		{ 135, 22, 57 },
		{ 135, 23, 57 },
		{ 135, 24, 57 },
		{ 135, 25, 57 },
		{ 135, 26, 57 },
		{ 135, 27, 57 },
		{ 135, 28, 57 },
		{ 135, 29, 57 },
		{ 135, 30, 57 },
		{ 135, 31, 57 },
		{ 135, 32, 57 },
		{ 135, 33, 57 },
		{ 135, 34, 57 },
		{ 135, 35, 57 },
		{ 135, 36, 57 },
		{ 135, 37, 57 },
		{ 135, 38, 57 },
		{ 135, 39, 57 },
		{ 135, 40, 57 },
		{ 135, 41, 57 },
		{ 135, 42, 57 },
		{ 135, 43, 57 },
		{ 135, 44, 57 },
		{ 135, 45, 57 },
		{ 135, 46, 57 },
		{ 135, 47, 57 },
		{ 135, 48, 57 },
		{ 135, 49, 57 },
		{ 135, 50, 57 },
		{ 135, 51, 57 },
		{ 135, 52, 57 },
		{ 135, 53, 57 },
		{ 135, 54, 57 },
		{ 135, 55, 57 },
		{ 135, 56, 57 },
		{ 135, 57, 57 },
		{ 135, 58, 57 },
		{ 135, 59, 57 },
		{ 135, 60, 57 },
		{ 135, 61, 57 },
		{ 135, 62, 57 },
		{ 135, 63, 57 },
		{ 135, 64, 57 },
		{ 135, 65, 57 },
		{ 135, 66, 57 },
		{ 135, 67, 57 },
		{ 135, 68, 57 },
		{ 135, 69, 57 },
		{ 135, 70, 57 },
		{ 135, 71, 57 },
		{ 135, 72, 57 },
		{ 135, 73, 57 },
		{ 135, 74, 57 },
		{ 135, 75, 57 },
		{ 135, 76, 57 },
		{ 135, 77, 57 },
		{ 135, 78, 57 },
		{ 135, 79, 57 },
		{ 135, 80, 57 },
		{ 135, 81, 57 },
		{ 135, 82, 57 },
		{ 135, 83, 57 },
		{ 135, 84, 57 },
		{ 135, 85, 57 },
		{ 135, 86, 57 },
		{ 135, 87, 57 },
		{ 135, 88, 57 },
		{ 135, 89, 57 },
		{ 135, 90, 57 },
		{ 135, 91, 57 },
		{ 135, 92, 57 },
		{ 135, 93, 57 },
		{ 135, 94, 57 },
		{ 136, 16, 57 },
		{ 136, 17, 57 },
		{ 136, 18, 57 },
		{ 136, 19, 57 },
		{ 136, 20, 57 },
		{ 136, 21, 57 },
		{ 136, 22, 57 },
		{ 136, 23, 57 },
		{ 136, 24, 57 },
		{ 136, 25, 57 },
		{ 136, 26, 57 },
		{ 136, 27, 57 },
		{ 136, 28, 57 },
		{ 136, 29, 57 },
		{ 136, 30, 57 },
		{ 136, 31, 57 },
		{ 136, 32, 57 },
		{ 136, 33, 57 },
		{ 136, 34, 57 },
		{ 136, 35, 57 },
		{ 136, 36, 57 },
		{ 136, 37, 57 },
		{ 136, 38, 57 },
		{ 136, 39, 57 },
		{ 136, 40, 57 },
		{ 136, 41, 57 },
		{ 136, 42, 57 },
		{ 136, 43, 57 },
		{ 136, 44, 57 },
		{ 136, 45, 57 },
		{ 136, 46, 57 },
		{ 136, 47, 57 },
		{ 136, 48, 57 },
		{ 136, 49, 57 },
		{ 136, 50, 57 },
		{ 136, 51, 57 },
		{ 136, 52, 57 },
		{ 136, 53, 57 },
		{ 136, 54, 57 },
		{ 136, 55, 57 },
		{ 136, 56, 57 },
		{ 136, 57, 57 },
		{ 136, 58, 57 },
		{ 136, 59, 57 },
		{ 136, 60, 57 },
		{ 136, 61, 57 },
		{ 136, 62, 57 },
		{ 136, 63, 57 },
		{ 136, 64, 57 },
		{ 136, 65, 57 },
		{ 136, 66, 57 },
		{ 136, 67, 57 },
		{ 136, 68, 57 },
		{ 136, 69, 57 },
		{ 136, 70, 57 },
		{ 136, 71, 57 },
		{ 136, 72, 57 },
		{ 136, 73, 57 },
		{ 136, 74, 57 },
		{ 136, 75, 57 },
		{ 136, 76, 57 },
		{ 136, 77, 57 },
		{ 136, 78, 57 },
		{ 136, 79, 57 },
		{ 136, 80, 57 },
		{ 136, 81, 57 },
		{ 136, 82, 57 },
		{ 136, 83, 57 },
		{ 136, 84, 57 },
		{ 136, 85, 57 },
		{ 136, 86, 57 },
		{ 136, 87, 57 },
		{ 136, 88, 57 },
		{ 136, 89, 57 },
		{ 136, 90, 57 },
		{ 136, 91, 57 },
		{ 136, 92, 57 },
		{ 136, 93, 57 },
		{ 136, 94, 57 },
		{ 136, 95, 57 },
		{ 137, 18, 57 },
		{ 137, 19, 57 },
		{ 137, 20, 57 },
		{ 137, 21, 57 },
		{ 137, 22, 57 },
		{ 137, 23, 57 },
		{ 137, 24, 57 },
		{ 137, 25, 57 },
		{ 137, 26, 57 },
		{ 137, 27, 57 },
		{ 137, 28, 57 },
		{ 137, 29, 57 },
		{ 137, 30, 57 },
		{ 137, 31, 57 },
		{ 137, 32, 57 },
		{ 137, 33, 57 },
		{ 137, 34, 57 },
		{ 137, 35, 57 },
		{ 137, 36, 57 },
		{ 137, 37, 57 },
		{ 137, 38, 57 },
		{ 137, 39, 57 },
		{ 137, 40, 57 },
		{ 137, 41, 57 },
		{ 137, 42, 57 },
		{ 137, 43, 57 },
		{ 137, 44, 57 },
		{ 137, 45, 57 },
		{ 137, 46, 57 },
		{ 137, 47, 57 },
		{ 137, 48, 57 },
		{ 137, 49, 57 },
		{ 137, 50, 57 },
		{ 137, 51, 57 },
		{ 137, 52, 57 },
		{ 137, 53, 57 },
		{ 137, 54, 57 },
		{ 137, 55, 57 },
		{ 137, 56, 57 },
		{ 137, 57, 57 },
		{ 137, 58, 57 },
		{ 137, 59, 57 },
		{ 137, 60, 57 },
		{ 137, 61, 57 },
		{ 137, 62, 57 },
		{ 137, 63, 57 },
		{ 137, 64, 57 },
		{ 137, 65, 57 },
		{ 137, 66, 57 },
		{ 137, 67, 57 },
		{ 137, 68, 57 },
		{ 137, 69, 57 },
		{ 137, 70, 57 },
		{ 137, 71, 57 },
		{ 137, 72, 57 },
		{ 137, 73, 57 },
		{ 137, 74, 57 },
		{ 137, 75, 57 },
		{ 137, 76, 57 },
		{ 137, 77, 57 },
		{ 137, 78, 57 },
		{ 137, 79, 57 },
		{ 137, 80, 57 },
		{ 137, 81, 57 },
		{ 137, 82, 57 },
		{ 137, 83, 57 },
		{ 137, 84, 57 },
		{ 137, 85, 57 },
		{ 137, 86, 57 },
		{ 137, 87, 57 },
		{ 137, 88, 57 },
		{ 137, 89, 57 },
		{ 137, 90, 57 },
		{ 137, 91, 57 },
		{ 137, 92, 57 },
		{ 137, 93, 57 },
		{ 137, 94, 57 },
		{ 137, 95, 57 },
		{ 137, 96, 57 },
		{ 138, 20, 57 },
		{ 138, 21, 57 },
		{ 138, 22, 57 },
		{ 138, 23, 57 },
		{ 138, 24, 57 },
		{ 138, 25, 57 },
		{ 138, 26, 57 },
		{ 138, 27, 57 },
		{ 138, 28, 57 },
		{ 138, 29, 57 },
		{ 138, 30, 57 },
		{ 138, 31, 57 },
		{ 138, 32, 57 },
		{ 138, 33, 57 },
		{ 138, 34, 57 },
		{ 138, 35, 57 },
		{ 138, 36, 57 },
		{ 138, 37, 57 },
		{ 138, 38, 57 },
		{ 138, 39, 57 },
		{ 138, 40, 57 },
		{ 138, 41, 57 },
		{ 138, 42, 57 },
		{ 138, 43, 57 },
		{ 138, 44, 57 },
		{ 138, 45, 57 },
		{ 138, 46, 57 },
		{ 138, 47, 57 },
		{ 138, 48, 57 },
		{ 138, 49, 57 },
		{ 138, 50, 57 },
		{ 138, 51, 57 },
		{ 138, 52, 57 },
		{ 138, 53, 57 },
		{ 138, 54, 57 },
		{ 138, 55, 57 },
		{ 138, 56, 57 },
		{ 138, 57, 57 },
		{ 138, 58, 57 },
		{ 138, 59, 57 },
		{ 138, 60, 57 },
		{ 138, 61, 57 },
		{ 138, 62, 57 },
		{ 138, 63, 57 },
		{ 138, 64, 57 },
		{ 138, 65, 57 },
		{ 138, 66, 57 },
		{ 138, 67, 57 },
		{ 138, 68, 57 },
		{ 138, 69, 57 },
		{ 138, 70, 57 },
		{ 138, 71, 57 },
		{ 138, 72, 57 },
		{ 138, 73, 57 },
		{ 138, 74, 57 },
		{ 138, 75, 57 },
		{ 138, 76, 57 },
		{ 138, 77, 57 },
		{ 138, 78, 57 },
		{ 138, 79, 57 },
		{ 138, 80, 57 },
		{ 138, 81, 57 },
		{ 138, 82, 57 },
		{ 138, 83, 57 },
		{ 138, 84, 57 },
		{ 138, 85, 57 },
		{ 138, 86, 57 },
		{ 138, 87, 57 },
		{ 138, 88, 57 },
		{ 138, 89, 57 },
		{ 138, 90, 57 },
		{ 138, 91, 57 },
		{ 138, 92, 57 },
		{ 138, 93, 57 },
		{ 138, 94, 57 },
		{ 138, 95, 57 },
		{ 138, 96, 57 },
		{ 138, 97, 57 },
		{ 138, 98, 57 },
		{ 139, 22, 57 },
		{ 139, 23, 57 },
		{ 139, 24, 57 },
		{ 139, 25, 57 },
		{ 139, 26, 57 },
		{ 139, 27, 57 },
		{ 139, 28, 57 },
		{ 139, 29, 57 },
		{ 139, 30, 57 },
		{ 139, 31, 57 },
		{ 139, 32, 57 },
		{ 139, 33, 57 },
		{ 139, 34, 57 },
		{ 139, 35, 57 },
		{ 139, 36, 57 },
		{ 139, 37, 57 },
		{ 139, 38, 57 },
		{ 139, 39, 57 },
		{ 139, 40, 57 },
		{ 139, 41, 57 },
		{ 139, 42, 57 },
		{ 139, 43, 57 },
		{ 139, 44, 57 },
		{ 139, 45, 57 },
		{ 139, 46, 57 },
		{ 139, 47, 57 },
		{ 139, 48, 57 },
		{ 139, 49, 57 },
		{ 139, 50, 57 },
		{ 139, 51, 57 },
		{ 139, 52, 57 },
		{ 139, 53, 57 },
		{ 139, 54, 57 },
		{ 139, 55, 57 },
		{ 139, 56, 57 },
		{ 139, 57, 57 },
		{ 139, 58, 57 },
		{ 139, 59, 57 },
		{ 139, 60, 57 },
		{ 139, 61, 57 },
		{ 139, 62, 57 },
		{ 139, 63, 57 },
		{ 139, 64, 57 },
		{ 139, 65, 57 },
		{ 139, 66, 57 },
		{ 139, 67, 57 },
		{ 139, 68, 57 },
		{ 139, 69, 57 },
		{ 139, 70, 57 },
		{ 139, 71, 57 },
		{ 139, 72, 57 },
		{ 139, 73, 57 },
		{ 139, 74, 57 },
		{ 139, 75, 57 },
		{ 139, 76, 57 },
		{ 139, 77, 57 },
		{ 139, 78, 57 },
		{ 139, 79, 57 },
		{ 139, 80, 57 },
		{ 139, 81, 57 },
		{ 139, 82, 57 },
		{ 139, 83, 57 },
		{ 139, 84, 57 },
		{ 139, 85, 57 },
		{ 139, 86, 57 },
		{ 139, 87, 57 },
		{ 139, 88, 57 },
		{ 139, 89, 57 },
		{ 139, 90, 57 },
		{ 139, 91, 57 },
		{ 139, 92, 57 },
		{ 139, 93, 57 },
		{ 139, 94, 57 },
		{ 139, 95, 57 },
		{ 139, 96, 57 },
		{ 139, 97, 57 },
		{ 139, 98, 57 },
		{ 139, 99, 57 },
		{ 140, 24, 57 },
		{ 140, 25, 57 },
		{ 140, 26, 57 },
		{ 140, 27, 57 },
		{ 140, 28, 57 },
		{ 140, 29, 57 },
		{ 140, 30, 57 },
		{ 140, 31, 57 },
		{ 140, 32, 57 },
		{ 140, 33, 57 },
		{ 140, 34, 57 },
		{ 140, 35, 57 },
		{ 140, 36, 57 },
		{ 140, 37, 57 },
		{ 140, 38, 57 },
		{ 140, 39, 57 },
		{ 140, 40, 57 },
		{ 140, 41, 57 },
		{ 140, 42, 57 },
		{ 140, 43, 57 },
		{ 140, 44, 57 },
		{ 140, 45, 57 },
		{ 140, 46, 57 },
		{ 140, 47, 57 },
		{ 140, 48, 57 },
		{ 140, 49, 57 },
		{ 140, 50, 57 },
		{ 140, 51, 57 },
		{ 140, 52, 57 },
		{ 140, 53, 57 },
		{ 140, 54, 57 },
		{ 140, 55, 57 },
		{ 140, 56, 57 },
		{ 140, 57, 57 },
		{ 140, 58, 57 },
		{ 140, 59, 57 },
		{ 140, 60, 57 },
		{ 140, 61, 57 },
		{ 140, 62, 57 },
		{ 140, 63, 57 },
		{ 140, 64, 57 },
		{ 140, 65, 57 },
		{ 140, 66, 57 },
		{ 140, 67, 57 },
		{ 140, 68, 57 },
		{ 140, 69, 57 },
		{ 140, 70, 57 },
		{ 140, 71, 57 },
		{ 140, 72, 57 },
		{ 140, 73, 57 },
		{ 140, 74, 57 },
		{ 140, 75, 57 },
		{ 140, 76, 57 },
		{ 140, 77, 57 },
		{ 140, 78, 57 },
		{ 140, 79, 57 },
		{ 140, 80, 57 },
		{ 140, 81, 57 },
		{ 140, 82, 57 },
		{ 140, 83, 57 },
		{ 140, 84, 57 },
		{ 140, 85, 57 },
		{ 140, 86, 57 },
		{ 140, 87, 57 },
		{ 140, 88, 57 },
		{ 140, 89, 57 },
		{ 140, 90, 57 },
		{ 140, 91, 57 },
		{ 140, 92, 57 },
		{ 140, 93, 57 },
		{ 140, 94, 57 },
		{ 140, 95, 57 },
		{ 140, 96, 57 },
		{ 140, 97, 57 },
		{ 140, 98, 57 },
		{ 140, 99, 57 },
		{ 140, 100, 57 },
		{ 141, 26, 57 },
		{ 141, 27, 57 },
		{ 141, 28, 57 },
		{ 141, 29, 57 },
		{ 141, 30, 57 },
		{ 141, 31, 57 },
		{ 141, 32, 57 },
		{ 141, 33, 57 },
		{ 141, 34, 57 },
		{ 141, 35, 57 },
		{ 141, 36, 57 },
		{ 141, 37, 57 },
		{ 141, 38, 57 },
		{ 141, 39, 57 },
		{ 141, 40, 57 },
		{ 141, 41, 57 },
		{ 141, 42, 57 },
		{ 141, 43, 57 },
		{ 141, 44, 57 },
		{ 141, 45, 57 },
		{ 141, 46, 57 },
		{ 141, 47, 57 },
		{ 141, 48, 57 },
		{ 141, 49, 57 },
		{ 141, 50, 57 },
		{ 141, 51, 57 },
		{ 141, 52, 57 },
		{ 141, 53, 57 },
		{ 141, 54, 57 },
		{ 141, 55, 57 },
		{ 141, 56, 57 },
		{ 141, 57, 57 },
		{ 141, 58, 57 },
		{ 141, 59, 57 },
		{ 141, 60, 57 },
		{ 141, 61, 57 },
		{ 141, 62, 57 },
		{ 141, 63, 57 },
		{ 141, 64, 57 },
		{ 141, 65, 57 },
		{ 141, 66, 57 },
		{ 141, 67, 57 },
		{ 141, 68, 57 },
		{ 141, 69, 57 },
		{ 141, 70, 57 },
		{ 141, 71, 57 },
		{ 141, 72, 57 },
		{ 141, 73, 57 },
		{ 141, 74, 57 },
		{ 141, 75, 57 },
		{ 141, 76, 57 },
		{ 141, 77, 57 },
		{ 141, 78, 57 },
		{ 141, 79, 57 },
		{ 141, 80, 57 },
		{ 141, 81, 57 },
		{ 141, 82, 57 },
		{ 141, 83, 57 },
		{ 141, 84, 57 },
		{ 141, 85, 57 },
		{ 141, 86, 57 },
		{ 141, 87, 57 },
		{ 141, 88, 57 },
		{ 141, 89, 57 },
		{ 141, 90, 57 },
		{ 141, 91, 57 },
		{ 141, 92, 57 },
		{ 141, 93, 57 },
		{ 141, 94, 57 },
		{ 141, 95, 57 },
		{ 141, 96, 57 },
		{ 141, 97, 57 },
		{ 141, 98, 57 },
		{ 141, 99, 57 },
		{ 141, 100, 57 },
		{ 141, 101, 57 },
		{ 141, 102, 57 },
		{ 142, 28, 57 },
		{ 142, 29, 57 },
		{ 142, 30, 57 },
		{ 142, 31, 57 },
		{ 142, 32, 57 },
		{ 142, 33, 57 },
		{ 142, 34, 57 },
		{ 142, 35, 57 },
		{ 142, 36, 57 },
		{ 142, 37, 57 },
		{ 142, 38, 57 },
		{ 142, 39, 57 },
		{ 142, 40, 57 },
		{ 142, 41, 57 },
		{ 142, 42, 57 },
		{ 142, 43, 57 },
		{ 142, 44, 57 },
		{ 142, 45, 57 },
		{ 142, 46, 57 },
		{ 142, 47, 57 },
		{ 142, 48, 57 },
		{ 142, 49, 57 },
		{ 142, 50, 57 },
		{ 142, 51, 57 },
		{ 142, 52, 57 },
		{ 142, 53, 57 },
		{ 142, 54, 57 },
		{ 142, 55, 57 },
		{ 142, 56, 57 },
		{ 142, 57, 57 },
		{ 142, 58, 57 },
		{ 142, 59, 57 },
		{ 142, 60, 57 },
		{ 142, 61, 57 },
		{ 142, 62, 57 },
		{ 142, 63, 57 },
		{ 142, 64, 57 },
		{ 142, 65, 57 },
		{ 142, 66, 57 },
		{ 142, 67, 57 },
		{ 142, 68, 57 },
		{ 142, 69, 57 },
		{ 142, 70, 57 },
		{ 142, 71, 57 },
		{ 142, 72, 57 },
		{ 142, 73, 57 },
		{ 142, 74, 57 },
		{ 142, 75, 57 },
		{ 142, 76, 57 },
		{ 142, 77, 57 },
		{ 142, 78, 57 },
		{ 142, 79, 57 },
		{ 142, 80, 57 },
		{ 142, 81, 57 },
		{ 142, 82, 57 },
		{ 142, 83, 57 },
		{ 142, 84, 57 },
		{ 142, 85, 57 },
		{ 142, 86, 57 },
		{ 142, 87, 57 },
		{ 142, 88, 57 },
		{ 142, 89, 57 },
		{ 142, 90, 57 },
		{ 142, 91, 57 },
		{ 142, 92, 57 },
		{ 142, 93, 57 },
		{ 142, 94, 57 },
		{ 142, 95, 57 },
		{ 142, 96, 57 },
		{ 142, 97, 57 },
		{ 142, 98, 57 },
		{ 142, 99, 57 },
		{ 142, 100, 57 },
		{ 142, 101, 57 },
		{ 142, 102, 57 },
		{ 142, 103, 57 },
		{ 143, 30, 57 },
		{ 143, 31, 57 },
		{ 143, 32, 57 },
		{ 143, 33, 57 },
		{ 143, 34, 57 },
		{ 143, 35, 57 },
		{ 143, 36, 57 },
		{ 143, 37, 57 },
		{ 143, 38, 57 },
		{ 143, 39, 57 },
		{ 143, 40, 57 },
		{ 143, 41, 57 },
		{ 143, 42, 57 },
		{ 143, 43, 57 },
		{ 143, 44, 57 },
		{ 143, 45, 57 },
		{ 143, 46, 57 },
		{ 143, 47, 57 },
		{ 143, 48, 57 },
		{ 143, 49, 57 },
		{ 143, 50, 57 },
		{ 143, 51, 57 },
		{ 143, 52, 57 },
		{ 143, 53, 57 },
		{ 143, 54, 57 },
		{ 143, 55, 57 },
		{ 143, 56, 57 },
		{ 143, 57, 57 },
		{ 143, 58, 57 },
		{ 143, 59, 57 },
		{ 143, 60, 57 },
		{ 143, 61, 57 },
		{ 143, 62, 57 },
		{ 143, 63, 57 },
		{ 143, 64, 57 },
		{ 143, 65, 57 },
		{ 143, 66, 57 },
		{ 143, 67, 57 },
		{ 143, 68, 57 },
		{ 143, 69, 57 },
		{ 143, 70, 57 },
		{ 143, 71, 57 },
		{ 143, 72, 57 },
		{ 143, 73, 57 },
		{ 143, 74, 57 },
		{ 143, 75, 57 },
		{ 143, 76, 57 },
		{ 143, 77, 57 },
		{ 143, 78, 57 },
		{ 143, 79, 57 },
		{ 143, 80, 57 },
		{ 143, 81, 57 },
		{ 143, 82, 57 },
		{ 143, 83, 57 },
		{ 143, 84, 57 },
		{ 143, 85, 57 },
		{ 143, 86, 57 },
		{ 143, 87, 57 },
		{ 143, 88, 57 },
		{ 143, 89, 57 },
		{ 143, 90, 57 },
		{ 143, 91, 57 },
		{ 143, 92, 57 },
		{ 143, 93, 57 },
		{ 143, 94, 57 },
		{ 143, 95, 57 },
		{ 143, 96, 57 },
		{ 143, 97, 57 },
		{ 143, 98, 57 },
		{ 143, 99, 57 },
		{ 143, 100, 57 },
		{ 143, 101, 57 },
		{ 143, 102, 57 },
		{ 143, 103, 57 },
		{ 143, 104, 57 },
		{ 144, 32, 57 },
		{ 144, 33, 57 },
		{ 144, 34, 57 },
		{ 144, 35, 57 },
		{ 144, 36, 57 },
		{ 144, 37, 57 },
		{ 144, 38, 57 },
		{ 144, 39, 57 },
		{ 144, 40, 57 },
		{ 144, 41, 57 },
		{ 144, 42, 57 },
		{ 144, 43, 57 },
		{ 144, 44, 57 },
		{ 144, 45, 57 },
		{ 144, 46, 57 },
		{ 144, 47, 57 },
		{ 144, 48, 57 },
		{ 144, 49, 57 },
		{ 144, 50, 57 },
		{ 144, 51, 57 },
		{ 144, 52, 57 },
		{ 144, 53, 57 },
		{ 144, 54, 57 },
		{ 144, 55, 57 },
		{ 144, 56, 57 },
		{ 144, 57, 57 },
		{ 144, 58, 57 },
		{ 144, 59, 57 },
		{ 144, 60, 57 },
		{ 144, 61, 57 },
		{ 144, 62, 57 },
		{ 144, 63, 57 },
		{ 144, 64, 57 },
		{ 144, 65, 57 },
		{ 144, 66, 57 },
		{ 144, 67, 57 },
		{ 144, 68, 57 },
		{ 144, 69, 57 },
		{ 144, 70, 57 },
		{ 144, 71, 57 },
		{ 144, 72, 57 },
		{ 144, 73, 57 },
		{ 144, 74, 57 },
		{ 144, 75, 57 },
		{ 144, 76, 57 },
		{ 144, 77, 57 },
		{ 144, 78, 57 },
		{ 144, 79, 57 },
		{ 144, 80, 57 },
		{ 144, 81, 57 },
		{ 144, 82, 57 },
		{ 144, 83, 57 },
		{ 144, 84, 57 },
		{ 144, 85, 57 },
		{ 144, 86, 57 },
		{ 144, 87, 57 },
		{ 144, 88, 57 },
		{ 144, 89, 57 },
		{ 144, 90, 57 },
		{ 144, 91, 57 },
		{ 144, 92, 57 },
		{ 144, 93, 57 },
		{ 144, 94, 57 },
		{ 144, 95, 57 },
		{ 144, 96, 57 },
		{ 144, 97, 57 },
		{ 144, 98, 57 },
		{ 144, 99, 57 },
		{ 144, 100, 57 },
		{ 144, 101, 57 },
		{ 144, 102, 57 },
		{ 144, 103, 57 },
		{ 144, 104, 57 },
		{ 144, 105, 57 },
		{ 144, 106, 57 },
		{ 145, 34, 57 },
		{ 145, 35, 57 },
		{ 145, 36, 57 },
		{ 145, 37, 57 },
		{ 145, 38, 57 },
		{ 145, 39, 57 },
		{ 145, 40, 57 },
		{ 145, 41, 57 },
		{ 145, 42, 57 },
		{ 145, 43, 57 },
		{ 145, 44, 57 },
		{ 145, 45, 57 },
		{ 145, 46, 57 },
		{ 145, 47, 57 },
		{ 145, 48, 57 },
		{ 145, 49, 57 },
		{ 145, 50, 57 },
		{ 145, 51, 57 },
		{ 145, 52, 57 },
		{ 145, 53, 57 },
		{ 145, 54, 57 },
		{ 145, 55, 57 },
		{ 145, 56, 57 },
		{ 145, 57, 57 },
		{ 145, 58, 57 },
		{ 145, 59, 57 },
		{ 145, 60, 57 },
		{ 145, 61, 57 },
		{ 145, 62, 57 },
		{ 145, 63, 57 },
		{ 145, 64, 57 },
		{ 145, 65, 57 },
		{ 145, 66, 57 },
		{ 145, 67, 57 },
		{ 145, 68, 57 },
		{ 145, 69, 57 },
		{ 145, 70, 57 },
		{ 145, 71, 57 },
		{ 145, 72, 57 },
		{ 145, 73, 57 },
		{ 145, 74, 57 },
		{ 145, 75, 57 },
		{ 145, 76, 57 },
		{ 145, 77, 57 },
		{ 145, 78, 57 },
		{ 145, 79, 57 },
		{ 145, 80, 57 },
		{ 145, 81, 57 },
		{ 145, 82, 57 },
		{ 145, 83, 57 },
		{ 145, 84, 57 },
		{ 145, 85, 57 },
		{ 145, 86, 57 },
		{ 145, 87, 57 },
		{ 145, 88, 57 },
		{ 145, 89, 57 },
		{ 145, 90, 57 },
		{ 145, 91, 57 },
		{ 145, 92, 57 },
		{ 145, 93, 57 },
		{ 145, 94, 57 },
		{ 145, 95, 57 },
		{ 145, 96, 57 },
		{ 145, 97, 57 },
		{ 145, 98, 57 },
		{ 145, 99, 57 },
		{ 145, 100, 57 },
		{ 145, 101, 57 },
		{ 145, 102, 57 },
		{ 145, 103, 57 },
		{ 145, 104, 57 },
		{ 145, 105, 57 },
		{ 145, 106, 57 },
		{ 145, 107, 57 },
		{ 146, 36, 57 },
		{ 146, 37, 57 },
		{ 146, 38, 57 },
		{ 146, 39, 57 },
		{ 146, 40, 57 },
		{ 146, 41, 57 },
		{ 146, 42, 57 },
		{ 146, 43, 57 },
		{ 146, 44, 57 },
		{ 146, 45, 57 },
		{ 146, 46, 57 },
		{ 146, 47, 57 },
		{ 146, 48, 57 },
		{ 146, 49, 57 },
		{ 146, 50, 57 },
		{ 146, 51, 57 },
		{ 146, 52, 57 },
		{ 146, 53, 57 },
		{ 146, 54, 57 },
		{ 146, 55, 57 },
		{ 146, 56, 57 },
		{ 146, 57, 57 },
		{ 146, 58, 57 },
		{ 146, 59, 57 },
		{ 146, 60, 57 },
		{ 146, 61, 57 },
		{ 146, 62, 57 },
		{ 146, 63, 57 },
		{ 146, 64, 57 },
		{ 146, 65, 57 },
		{ 146, 66, 57 },
		{ 146, 67, 57 },
		{ 146, 68, 57 },
		{ 146, 69, 57 },
		{ 146, 70, 57 },
		{ 146, 71, 57 },
		{ 146, 72, 57 },
		{ 146, 73, 57 },
		{ 146, 74, 57 },
		{ 146, 75, 57 },
		{ 146, 76, 57 },
		{ 146, 77, 57 },
		{ 146, 78, 57 },
		{ 146, 79, 57 },
		{ 146, 80, 57 },
		{ 146, 81, 57 },
		{ 146, 82, 57 },
		{ 146, 83, 57 },
		{ 146, 84, 57 },
		{ 146, 85, 57 },
		{ 146, 86, 57 },
		{ 146, 87, 57 },
		{ 146, 88, 57 },
		{ 146, 89, 57 },
		{ 146, 90, 57 },
		{ 146, 91, 57 },
		{ 146, 92, 57 },
		{ 146, 93, 57 },
		{ 146, 94, 57 },
		{ 146, 95, 57 },
		{ 146, 96, 57 },
		{ 146, 97, 57 },
		{ 146, 98, 57 },
		{ 146, 99, 57 },
		{ 146, 100, 57 },
		{ 146, 101, 57 },
		{ 146, 102, 57 },
		{ 146, 103, 57 },
		{ 146, 104, 57 },
		{ 146, 105, 57 },
		{ 146, 106, 57 },
		{ 146, 107, 57 },
		{ 146, 108, 57 },
		{ 147, 38, 57 },
		{ 147, 39, 57 },
		{ 147, 40, 57 },
		{ 147, 41, 57 },
		{ 147, 42, 57 },
		{ 147, 43, 57 },
		{ 147, 44, 57 },
		{ 147, 45, 57 },
		{ 147, 46, 57 },
		{ 147, 47, 57 },
		{ 147, 48, 57 },
		{ 147, 49, 57 },
		{ 147, 50, 57 },
		{ 147, 51, 57 },
		{ 147, 52, 57 },
		{ 147, 53, 57 },
		{ 147, 54, 57 },
		{ 147, 55, 57 },
		{ 147, 56, 57 },
		{ 147, 57, 57 },
		{ 147, 58, 57 },
		{ 147, 59, 57 },
		{ 147, 60, 57 },
		{ 147, 61, 57 },
		{ 147, 62, 57 },
		{ 147, 63, 57 },
		{ 147, 64, 57 },
		{ 147, 65, 57 },
		{ 147, 66, 57 },
		{ 147, 67, 57 },
		{ 147, 68, 57 },
		{ 147, 69, 57 },
		{ 147, 70, 57 },
		{ 147, 71, 57 },
		{ 147, 72, 57 },
		{ 147, 73, 57 },
		{ 147, 74, 57 },
		{ 147, 75, 57 },
		{ 147, 76, 57 },
		{ 147, 77, 57 },
		{ 147, 78, 57 },
		{ 147, 79, 57 },
		{ 147, 80, 57 },
		{ 147, 81, 57 },
		{ 147, 82, 57 },
		{ 147, 83, 57 },
		{ 147, 84, 57 },
		{ 147, 85, 57 },
		{ 147, 86, 57 },
		{ 147, 87, 57 },
		{ 147, 88, 57 },
		{ 147, 89, 57 },
		{ 147, 90, 57 },
		{ 147, 91, 57 },
		{ 147, 92, 57 },
		{ 147, 93, 57 },
		{ 147, 94, 57 },
		{ 147, 95, 57 },
		{ 147, 96, 57 },
		{ 147, 97, 57 },
		{ 147, 98, 57 },
		{ 147, 99, 57 },
		{ 147, 100, 57 },
		{ 147, 101, 57 },
		{ 147, 102, 57 },
		{ 147, 103, 57 },
		{ 147, 104, 57 },
		{ 147, 105, 57 },
		{ 147, 106, 57 },
		{ 147, 107, 57 },
		{ 147, 108, 57 },
		{ 147, 109, 57 },
		{ 147, 110, 57 },
		{ 148, 40, 57 },
		{ 148, 41, 57 },
		{ 148, 42, 57 },
		{ 148, 43, 57 },
		{ 148, 44, 57 },
		{ 148, 45, 57 },
		{ 148, 46, 57 },
		{ 148, 47, 57 },
		{ 148, 48, 57 },
		{ 148, 49, 57 },
		{ 148, 50, 57 },
		{ 148, 51, 57 },
		{ 148, 52, 57 },
		{ 148, 53, 57 },
		{ 148, 54, 57 },
		{ 148, 55, 57 },
		{ 148, 56, 57 },
		{ 148, 57, 57 },
		{ 148, 58, 57 },
		{ 148, 59, 57 },
		{ 148, 60, 57 },
		{ 148, 61, 57 },
		{ 148, 62, 57 },
		{ 148, 63, 57 },
		{ 148, 64, 57 },
		{ 148, 65, 57 },
		{ 148, 66, 57 },
		{ 148, 67, 57 },
		{ 148, 68, 57 },
		{ 148, 69, 57 },
		{ 148, 70, 57 },
		{ 148, 71, 57 },
		{ 148, 72, 57 },
		{ 148, 73, 57 },
		{ 148, 74, 57 },
		{ 148, 75, 57 },
		{ 148, 76, 57 },
		{ 148, 77, 57 },
		{ 148, 78, 57 },
		{ 148, 79, 57 },
		{ 148, 80, 57 },
		{ 148, 81, 57 },
		{ 148, 82, 57 },
		{ 148, 83, 57 },
		{ 148, 84, 57 },
		{ 148, 85, 57 },
		{ 148, 86, 57 },
		{ 148, 87, 57 },
		{ 148, 88, 57 },
		{ 148, 89, 57 },
		{ 148, 90, 57 },
		{ 148, 91, 57 },
		{ 148, 92, 57 },
		{ 148, 93, 57 },
		{ 148, 94, 57 },
		{ 148, 95, 57 },
		{ 148, 96, 57 },
		{ 148, 97, 57 },
		{ 148, 98, 57 },
		{ 148, 99, 57 },
		{ 148, 100, 57 },
		{ 148, 101, 57 },
		{ 148, 102, 57 },
		{ 148, 103, 57 },
		{ 148, 104, 57 },
		{ 148, 105, 57 },
		{ 148, 106, 57 },
		{ 148, 107, 57 },
		{ 148, 108, 57 },
		{ 148, 109, 57 },
		{ 148, 110, 57 },
		{ 148, 111, 57 },
		{ 149, 42, 57 },
		{ 149, 43, 57 },
		{ 149, 44, 57 },
		{ 149, 45, 57 },
		{ 149, 46, 57 },
		{ 149, 47, 57 },
		{ 149, 48, 57 },
		{ 149, 49, 57 },
		{ 149, 50, 57 },
		{ 149, 51, 57 },
		{ 149, 52, 57 },
		{ 149, 53, 57 },
		{ 149, 54, 57 },
		{ 149, 55, 57 },
		{ 149, 56, 57 },
		{ 149, 57, 57 },
		{ 149, 58, 57 },
		{ 149, 59, 57 },
		{ 149, 60, 57 },
		{ 149, 61, 57 },
		{ 149, 62, 57 },
		{ 149, 63, 57 },
		{ 149, 64, 57 },
		{ 149, 65, 57 },
		{ 149, 66, 57 },
		{ 149, 67, 57 },
		{ 149, 68, 57 },
		{ 149, 69, 57 },
		{ 149, 70, 57 },
		{ 149, 71, 57 },
		{ 149, 72, 57 },
		{ 149, 73, 57 },
		{ 149, 74, 57 },
		{ 149, 75, 57 },
		{ 149, 76, 57 },
		{ 149, 77, 57 },
		{ 149, 78, 57 },
		{ 149, 79, 57 },
		{ 149, 80, 57 },
		{ 149, 81, 57 },
		{ 149, 82, 57 },
		{ 149, 83, 57 },
		{ 149, 84, 57 },
		{ 149, 85, 57 },
		{ 149, 86, 57 },
		{ 149, 87, 57 },
		{ 149, 88, 57 },
		{ 149, 89, 57 },
		{ 149, 90, 57 },
		{ 149, 91, 57 },
		{ 149, 92, 57 },
		{ 149, 93, 57 },
		{ 149, 94, 57 },
		{ 149, 95, 57 },
		{ 149, 96, 57 },
		{ 149, 97, 57 },
		{ 149, 98, 57 },
		{ 149, 99, 57 },
		{ 149, 100, 57 },
		{ 149, 101, 57 },
		{ 149, 102, 57 },
		{ 149, 103, 57 },
		{ 149, 104, 57 },
		{ 149, 105, 57 },
		{ 149, 106, 57 },
		{ 149, 107, 57 },
		{ 149, 108, 57 },
		{ 149, 109, 57 },
		{ 149, 110, 57 },
		{ 149, 111, 57 },
		{ 149, 112, 57 },
		{ 150, 44, 57 },
		{ 150, 45, 57 },
		{ 150, 46, 57 },
		{ 150, 47, 57 },
		{ 150, 48, 57 },
		{ 150, 49, 57 },
		{ 150, 50, 57 },
		{ 150, 51, 57 },
		{ 150, 52, 57 },
		{ 150, 53, 57 },
		{ 150, 54, 57 },
		{ 150, 55, 57 },
		{ 150, 56, 57 },
		{ 150, 57, 57 },
		{ 150, 58, 57 },
		{ 150, 59, 57 },
		{ 150, 60, 57 },
		{ 150, 61, 57 },
		{ 150, 62, 57 },
		{ 150, 63, 57 },
		{ 150, 64, 57 },
		{ 150, 65, 57 },
		{ 150, 66, 57 },
		{ 150, 67, 57 },
		{ 150, 68, 57 },
		{ 150, 69, 57 },
		{ 150, 70, 57 },
		{ 150, 71, 57 },
		{ 150, 72, 57 },
		{ 150, 73, 57 },
		{ 150, 74, 57 },
		{ 150, 75, 57 },
		{ 150, 76, 57 },
		{ 150, 77, 57 },
		{ 150, 78, 57 },
		{ 150, 79, 57 },
		{ 150, 80, 57 },
		{ 150, 81, 57 },
		{ 150, 82, 57 },
		{ 150, 83, 57 },
		{ 150, 84, 57 },
		{ 150, 85, 57 },
		{ 150, 86, 57 },
		{ 150, 87, 57 },
		{ 150, 88, 57 },
		{ 150, 89, 57 },
		{ 150, 90, 57 },
		{ 150, 91, 57 },
		{ 150, 92, 57 },
		{ 150, 93, 57 },
		{ 150, 94, 57 },
		{ 150, 95, 57 },
		{ 150, 96, 57 },
		{ 150, 97, 57 },
		{ 150, 98, 57 },
		{ 150, 99, 57 },
		{ 150, 100, 57 },
		{ 150, 101, 57 },
		{ 150, 102, 57 },
		{ 150, 103, 57 },
		{ 150, 104, 57 },
		{ 150, 105, 57 },
		{ 150, 106, 57 },
		{ 150, 107, 57 },
		{ 150, 108, 57 },
		{ 150, 109, 57 },
		{ 150, 110, 57 },
		{ 150, 111, 57 },
		{ 150, 112, 57 },
		{ 150, 113, 57 },
		{ 150, 114, 57 },
		{ 151, 46, 57 },
		{ 151, 47, 57 },
		{ 151, 48, 57 },
		{ 151, 49, 57 },
		{ 151, 50, 57 },
		{ 151, 51, 57 },
		{ 151, 52, 57 },
		{ 151, 53, 57 },
		{ 151, 54, 57 },
		{ 151, 55, 57 },
		{ 151, 56, 57 },
		{ 151, 57, 57 },
		{ 151, 58, 57 },
		{ 151, 59, 57 },
		{ 151, 60, 57 },
		{ 151, 61, 57 },
		{ 151, 62, 57 },
		{ 151, 63, 57 },
		{ 151, 64, 57 },
		{ 151, 65, 57 },
		{ 151, 66, 57 },
		{ 151, 67, 57 },
		{ 151, 68, 57 },
		{ 151, 69, 57 },
		{ 151, 70, 57 },
		{ 151, 71, 57 },
		{ 151, 72, 57 },
		{ 151, 73, 57 },
		{ 151, 74, 57 },
		{ 151, 75, 57 },
		{ 151, 76, 57 },
		{ 151, 77, 57 },
		{ 151, 78, 57 },
		{ 151, 79, 57 },
		{ 151, 80, 57 },
		{ 151, 81, 57 },
		{ 151, 82, 57 },
		{ 151, 83, 57 },
		{ 151, 84, 57 },
		{ 151, 85, 57 },
		{ 151, 86, 57 },
		{ 151, 87, 57 },
		{ 151, 88, 57 },
		{ 151, 89, 57 },
		{ 151, 90, 57 },
		{ 151, 91, 57 },
		{ 151, 92, 57 },
		{ 151, 93, 57 },
		{ 151, 94, 57 },
		{ 151, 95, 57 },
		{ 151, 96, 57 },
		{ 151, 97, 57 },
		{ 151, 98, 57 },
		{ 151, 99, 57 },
		{ 151, 100, 57 },
		{ 151, 101, 57 },
		{ 151, 102, 57 },
		{ 151, 103, 57 },
		{ 151, 104, 57 },
		{ 151, 105, 57 },
		{ 151, 106, 57 },
		{ 151, 107, 57 },
		{ 151, 108, 57 },
		{ 151, 109, 57 },
		{ 151, 110, 57 },
		{ 151, 111, 57 },
		{ 151, 112, 57 },
		{ 151, 113, 57 },
		{ 151, 114, 57 },
		{ 151, 115, 57 },
		{ 152, 48, 57 },
		{ 152, 49, 57 },
		{ 152, 50, 57 },
		{ 152, 51, 57 },
		{ 152, 52, 57 },
		{ 152, 53, 57 },
		{ 152, 54, 57 },
		{ 152, 55, 57 },
		{ 152, 56, 57 },
		{ 152, 57, 57 },
		{ 152, 58, 57 },
		{ 152, 59, 57 },
		{ 152, 60, 57 },
		{ 152, 61, 57 },
		{ 152, 62, 57 },
		{ 152, 63, 57 },
		{ 152, 64, 57 },
		{ 152, 65, 57 },
		{ 152, 66, 57 },
		{ 152, 67, 57 },
		{ 152, 68, 57 },
		{ 152, 69, 57 },
		{ 152, 70, 57 },
		{ 152, 71, 57 },
		{ 152, 72, 57 },
		{ 152, 73, 57 },
		{ 152, 74, 57 },
		{ 152, 75, 57 },
		{ 152, 76, 57 },
		{ 152, 77, 57 },
		{ 152, 78, 57 },
		{ 152, 79, 57 },
		{ 152, 80, 57 },
		{ 152, 81, 57 },
		{ 152, 82, 57 },
		{ 152, 83, 57 },
		{ 152, 84, 57 },
		{ 152, 85, 57 },
		{ 152, 86, 57 },
		{ 152, 87, 57 },
		{ 152, 88, 57 },
		{ 152, 89, 57 },
		{ 152, 90, 57 },
		{ 152, 91, 57 },
		{ 152, 92, 57 },
		{ 152, 93, 57 },
		{ 152, 94, 57 },
		{ 152, 95, 57 },
		{ 152, 96, 57 },
		{ 152, 97, 57 },
		{ 152, 98, 57 },
		{ 152, 99, 57 },
		{ 152, 100, 57 },
		{ 152, 101, 57 },
		{ 152, 102, 57 },
		{ 152, 103, 57 },
		{ 152, 104, 57 },
		{ 152, 105, 57 },
		{ 152, 106, 57 },
		{ 152, 107, 57 },
		{ 152, 108, 57 },
		{ 152, 109, 57 },
		{ 152, 110, 57 },
		{ 152, 111, 57 },
		{ 152, 112, 57 },
		{ 152, 113, 57 },
		{ 152, 114, 57 },
		{ 152, 115, 57 },
		{ 152, 116, 57 },
		{ 153, 50, 57 },
		{ 153, 51, 57 },
		{ 153, 52, 57 },
		{ 153, 53, 57 },
		{ 153, 54, 57 },
		{ 153, 55, 57 },
		{ 153, 56, 57 },
		{ 153, 57, 57 },
		{ 153, 58, 57 },
		{ 153, 59, 57 },
		{ 153, 60, 57 },
		{ 153, 61, 57 },
		{ 153, 62, 57 },
		{ 153, 63, 57 },
		{ 153, 64, 57 },
		{ 153, 65, 57 },
		{ 153, 66, 57 },
		{ 153, 67, 57 },
		{ 153, 68, 57 },
		{ 153, 69, 57 },
		{ 153, 70, 57 },
		{ 153, 71, 57 },
		{ 153, 72, 57 },
		{ 153, 73, 57 },
		{ 153, 74, 57 },
		{ 153, 75, 57 },
		{ 153, 76, 57 },
		{ 153, 77, 57 },
		{ 153, 78, 57 },
		{ 153, 79, 57 },
		{ 153, 80, 57 },
		{ 153, 81, 57 },
		{ 153, 82, 57 },
		{ 153, 83, 57 },
		{ 153, 84, 57 },
		{ 153, 85, 57 },
		{ 153, 86, 57 },
		{ 153, 87, 57 },
		{ 153, 88, 57 },
		{ 153, 89, 57 },
		{ 153, 90, 57 },
		{ 153, 91, 57 },
		{ 153, 92, 57 },
		{ 153, 93, 57 },
		{ 153, 94, 57 },
		{ 153, 95, 57 },
		{ 153, 96, 57 },
		{ 153, 97, 57 },
		{ 153, 98, 57 },
		{ 153, 99, 57 },
		{ 153, 100, 57 },
		{ 153, 101, 57 },
		{ 153, 102, 57 },
		{ 153, 103, 57 },
		{ 153, 104, 57 },
		{ 153, 105, 57 },
		{ 153, 106, 57 },
		{ 153, 107, 57 },
		{ 153, 108, 57 },
		{ 153, 109, 57 },
		{ 153, 110, 57 },
		{ 153, 111, 57 },
		{ 153, 112, 57 },
		{ 153, 113, 57 },
		{ 153, 114, 57 },
		{ 153, 115, 57 },
		{ 153, 116, 57 },
		{ 153, 117, 57 },
		{ 153, 118, 57 },
		{ 154, 52, 57 },
		{ 154, 53, 57 },
		{ 154, 54, 57 },
		{ 154, 55, 57 },
		{ 154, 56, 57 },
		{ 154, 57, 57 },
		{ 154, 58, 57 },
		{ 154, 59, 57 },
		{ 154, 60, 57 },
		{ 154, 61, 57 },
		{ 154, 62, 57 },
		{ 154, 63, 57 },
		{ 154, 64, 57 },
		{ 154, 65, 57 },
		{ 154, 66, 57 },
		{ 154, 67, 57 },
		{ 154, 68, 57 },
		{ 154, 69, 57 },
		{ 154, 70, 57 },
		{ 154, 71, 57 },
		{ 154, 72, 57 },
		{ 154, 73, 57 },
		{ 154, 74, 57 },
		{ 154, 75, 57 },
		{ 154, 76, 57 },
		{ 154, 77, 57 },
		{ 154, 78, 57 },
		{ 154, 79, 57 },
		{ 154, 80, 57 },
		{ 154, 81, 57 },
		{ 154, 82, 57 },
		{ 154, 83, 57 },
		{ 154, 84, 57 },
		{ 154, 85, 57 },
		{ 154, 86, 57 },
		{ 154, 87, 57 },
		{ 154, 88, 57 },
		{ 154, 89, 57 },
		{ 154, 90, 57 },
		{ 154, 91, 57 },
		{ 154, 92, 57 },
		{ 154, 93, 57 },
		{ 154, 94, 57 },
		{ 154, 95, 57 },
		{ 154, 96, 57 },
		{ 154, 97, 57 },
		{ 154, 98, 57 },
		{ 154, 99, 57 },
		{ 154, 100, 57 },
		{ 154, 101, 57 },
		{ 154, 102, 57 },
		{ 154, 103, 57 },
		{ 154, 104, 57 },
		{ 154, 105, 57 },
		{ 154, 106, 57 },
		{ 154, 107, 57 },
		{ 154, 108, 57 },
		{ 154, 109, 57 },
		{ 154, 110, 57 },
		{ 154, 111, 57 },
		{ 154, 112, 57 },
		{ 154, 113, 57 },
		{ 154, 114, 57 },
		{ 154, 115, 57 },
		{ 154, 116, 57 },
		{ 154, 117, 57 },
		{ 154, 118, 57 },
		{ 154, 119, 57 },
		{ 155, 54, 57 },
		{ 155, 55, 57 },
		{ 155, 56, 57 },
		{ 155, 57, 57 },
		{ 155, 58, 57 },
		{ 155, 59, 57 },
		{ 155, 60, 57 },
		{ 155, 61, 57 },
		{ 155, 62, 57 },
		{ 155, 63, 57 },
		{ 155, 64, 57 },
		{ 155, 65, 57 },
		{ 155, 66, 57 },
		{ 155, 67, 57 },
		{ 155, 68, 57 },
		{ 155, 69, 57 },
		{ 155, 70, 57 },
		{ 155, 71, 57 },
		{ 155, 72, 57 },
		{ 155, 73, 57 },
		{ 155, 74, 57 },
		{ 155, 75, 57 },
		{ 155, 76, 57 },
		{ 155, 77, 57 },
		{ 155, 78, 57 },
		{ 155, 79, 57 },
		{ 155, 80, 57 },
		{ 155, 81, 57 },
		{ 155, 82, 57 },
		{ 155, 83, 57 },
		{ 155, 84, 57 },
		{ 155, 85, 57 },
		{ 155, 86, 57 },
		{ 155, 87, 57 },
		{ 155, 88, 57 },
		{ 155, 89, 57 },
		{ 155, 90, 57 },
		{ 155, 91, 57 },
		{ 155, 92, 57 },
		{ 155, 93, 57 },
		{ 155, 94, 57 },
		{ 155, 95, 57 },
		{ 155, 96, 57 },
		{ 155, 97, 57 },
		{ 155, 98, 57 },
		{ 155, 99, 57 },
		{ 155, 100, 57 },
		{ 155, 101, 57 },
		{ 155, 102, 57 },
		{ 155, 103, 57 },
		{ 155, 104, 57 },
		{ 155, 105, 57 },
		{ 155, 106, 57 },
		{ 155, 107, 57 },
		{ 155, 108, 57 },
		{ 155, 109, 57 },
		{ 155, 110, 57 },
		{ 155, 111, 57 },
		{ 155, 112, 57 },
		{ 155, 113, 57 },
		{ 155, 114, 57 },
		{ 155, 115, 57 },
		{ 155, 116, 57 },
		{ 155, 117, 57 },
		{ 155, 118, 57 },
		{ 155, 119, 57 },
		{ 155, 120, 57 },
		{ 156, 56, 57 },
		{ 156, 57, 57 },
		{ 156, 58, 57 },
		{ 156, 59, 57 },
		{ 156, 60, 57 },
		{ 156, 61, 57 },
		{ 156, 62, 57 },
		{ 156, 63, 57 },
		{ 156, 64, 57 },
		{ 156, 65, 57 },
		{ 156, 66, 57 },
		{ 156, 67, 57 },
		{ 156, 68, 57 },
		{ 156, 69, 57 },
		{ 156, 70, 57 },
		{ 156, 71, 57 },
		{ 156, 72, 57 },
		{ 156, 73, 57 },
		{ 156, 74, 57 },
		{ 156, 75, 57 },
		{ 156, 76, 57 },
		{ 156, 77, 57 },
		{ 156, 78, 57 },
		{ 156, 79, 57 },
		{ 156, 80, 57 },
		{ 156, 81, 57 },
		{ 156, 82, 57 },
		{ 156, 83, 57 },
		{ 156, 84, 57 },
		{ 156, 85, 57 },
		{ 156, 86, 57 },
		{ 156, 87, 57 },
		{ 156, 88, 57 },
		{ 156, 89, 57 },
		{ 156, 90, 57 },
		{ 156, 91, 57 },
		{ 156, 92, 57 },
		{ 156, 93, 57 },
		{ 156, 94, 57 },
		{ 156, 95, 57 },
		{ 156, 96, 57 },
		{ 156, 97, 57 },
		{ 156, 98, 57 },
		{ 156, 99, 57 },
		{ 156, 100, 57 },
		{ 156, 101, 57 },
		{ 156, 102, 57 },
		{ 156, 103, 57 },
		{ 156, 104, 57 },
		{ 156, 105, 57 },
		{ 156, 106, 57 },
		{ 156, 107, 57 },
		{ 156, 108, 57 },
		{ 156, 109, 57 },
		{ 156, 110, 57 },
		{ 156, 111, 57 },
		{ 156, 112, 57 },
		{ 156, 113, 57 },
		{ 156, 114, 57 },
		{ 156, 115, 57 },
		{ 156, 116, 57 },
		{ 156, 117, 57 },
		{ 156, 118, 57 },
		{ 156, 119, 57 },
		{ 156, 120, 57 },
		{ 156, 121, 57 },
		{ 156, 122, 57 },
		{ 157, 58, 57 },
		{ 157, 59, 57 },
		{ 157, 60, 57 },
		{ 157, 61, 57 },
		{ 157, 62, 57 },
		{ 157, 63, 57 },
		{ 157, 64, 57 },
		{ 157, 65, 57 },
		{ 157, 66, 57 },
		{ 157, 67, 57 },
		{ 157, 68, 57 },
		{ 157, 69, 57 },
		{ 157, 70, 57 },
		{ 157, 71, 57 },
		{ 157, 72, 57 },
		{ 157, 73, 57 },
		{ 157, 74, 57 },
		{ 157, 75, 57 },
		{ 157, 76, 57 },
		{ 157, 77, 57 },
		{ 157, 78, 57 },
		{ 157, 79, 57 },
		{ 157, 80, 57 },
		{ 157, 81, 57 },
		{ 157, 82, 57 },
		{ 157, 83, 57 },
		{ 157, 84, 57 },
		{ 157, 85, 57 },
		{ 157, 86, 57 },
		{ 157, 87, 57 },
		{ 157, 88, 57 },
		{ 157, 89, 57 },
		{ 157, 90, 57 },
		{ 157, 91, 57 },
		{ 157, 92, 57 },
		{ 157, 93, 57 },
		{ 157, 94, 57 },
		{ 157, 95, 57 },
		{ 157, 96, 57 },
		{ 157, 97, 57 },
		{ 157, 98, 57 },
		{ 157, 99, 57 },
		{ 157, 100, 57 },
		{ 157, 101, 57 },
		{ 157, 102, 57 },
		{ 157, 103, 57 },
		{ 157, 104, 57 },
		{ 157, 105, 57 },
		{ 157, 106, 57 },
		{ 157, 107, 57 },
		{ 157, 108, 57 },
		{ 157, 109, 57 },
		{ 157, 110, 57 },
		{ 157, 111, 57 },
		{ 157, 112, 57 },
		{ 157, 113, 57 },
		{ 157, 114, 57 },
		{ 157, 115, 57 },
		{ 157, 116, 57 },
		{ 157, 117, 57 },
		{ 157, 118, 57 },
		{ 157, 119, 57 },
		{ 157, 120, 57 },
		{ 157, 121, 57 },
		{ 157, 122, 57 },
		{ 157, 123, 57 },
		{ 158, 60, 57 },
		{ 158, 61, 57 },
		{ 158, 62, 57 },
		{ 158, 63, 57 },
		{ 158, 64, 57 },
		{ 158, 65, 57 },
		{ 158, 66, 57 },
		{ 158, 67, 57 },
		{ 158, 68, 57 },
		{ 158, 69, 57 },
		{ 158, 70, 57 },
		{ 158, 71, 57 },
		{ 158, 72, 57 },
		{ 158, 73, 57 },
		{ 158, 74, 57 },
		{ 158, 75, 57 },
		{ 158, 76, 57 },
		{ 158, 77, 57 },
		{ 158, 78, 57 },
		{ 158, 79, 57 },
		{ 158, 80, 57 },
		{ 158, 81, 57 },
		{ 158, 82, 57 },
		{ 158, 83, 57 },
		{ 158, 84, 57 },
		{ 158, 85, 57 },
		{ 158, 86, 57 },
		{ 158, 87, 57 },
		{ 158, 88, 57 },
		{ 158, 89, 57 },
		{ 158, 90, 57 },
		{ 158, 91, 57 },
		{ 158, 92, 57 },
		{ 158, 93, 57 },
		{ 158, 94, 57 },
		{ 158, 95, 57 },
		{ 158, 96, 57 },
		{ 158, 97, 57 },
		{ 158, 98, 57 },
		{ 158, 99, 57 },
		{ 158, 100, 57 },
		{ 158, 101, 57 },
		{ 158, 102, 57 },
		{ 158, 103, 57 },
		{ 158, 104, 57 },
		{ 158, 105, 57 },
		{ 158, 106, 57 },
		{ 158, 107, 57 },
		{ 158, 108, 57 },
		{ 158, 109, 57 },
		{ 158, 110, 57 },
		{ 158, 111, 57 },
		{ 158, 112, 57 },
		{ 158, 113, 57 },
		{ 158, 114, 57 },
		{ 158, 115, 57 },
		{ 158, 116, 57 },
		{ 158, 117, 57 },
		{ 158, 118, 57 },
		{ 158, 119, 57 },
		{ 158, 120, 57 },
		{ 158, 121, 57 },
		{ 158, 122, 57 },
		{ 158, 123, 57 },
		{ 158, 124, 57 },
		{ 159, 62, 57 },
		{ 159, 63, 57 },
		{ 159, 64, 57 },
		{ 159, 65, 57 },
		{ 159, 66, 57 },
		{ 159, 67, 57 },
		{ 159, 68, 57 },
		{ 159, 69, 57 },
		{ 159, 70, 57 },
		{ 159, 71, 57 },
		{ 159, 72, 57 },
		{ 159, 73, 57 },
		{ 159, 74, 57 },
		{ 159, 75, 57 },
		{ 159, 76, 57 },
		{ 159, 77, 57 },
		{ 159, 78, 57 },
		{ 159, 79, 57 },
		{ 159, 80, 57 },
		{ 159, 81, 57 },
		{ 159, 82, 57 },
		{ 159, 83, 57 },
		{ 159, 84, 57 },
		{ 159, 85, 57 },
		{ 159, 86, 57 },
		{ 159, 87, 57 },
		{ 159, 88, 57 },
		{ 159, 89, 57 },
		{ 159, 90, 57 },
		{ 159, 91, 57 },
		{ 159, 92, 57 },
		{ 159, 93, 57 },
		{ 159, 94, 57 },
		{ 159, 95, 57 },
		{ 159, 96, 57 },
		{ 159, 97, 57 },
		{ 159, 98, 57 },
		{ 159, 99, 57 },
		{ 159, 100, 57 },
		{ 159, 101, 57 },
		{ 159, 102, 57 },
		{ 159, 103, 57 },
		{ 159, 104, 57 },
		{ 159, 105, 57 },
		{ 159, 106, 57 },
		{ 159, 107, 57 },
		{ 159, 108, 57 },
		{ 159, 109, 57 },
		{ 159, 110, 57 },
		{ 159, 111, 57 },
		{ 159, 112, 57 },
		{ 159, 113, 57 },
		{ 159, 114, 57 },
		{ 159, 115, 57 },
		{ 159, 116, 57 },
		{ 159, 117, 57 },
		{ 159, 118, 57 },
		{ 159, 119, 57 },
		{ 159, 120, 57 },
		{ 159, 121, 57 },
		{ 159, 122, 57 },
		{ 159, 123, 57 },
		{ 159, 124, 57 },
		{ 159, 125, 57 },
		{ 159, 126, 57 },
		{ 160, 64, 57 },
		{ 160, 65, 57 },
		{ 160, 66, 57 },
		{ 160, 67, 57 },
		{ 160, 68, 57 },
		{ 160, 69, 57 },
		{ 160, 70, 57 },
		{ 160, 71, 57 },
		{ 160, 72, 57 },
		{ 160, 73, 57 },
		{ 160, 74, 57 },
		{ 160, 75, 57 },
		{ 160, 76, 57 },
		{ 160, 77, 57 },
		{ 160, 78, 57 },
		{ 160, 79, 57 },
		{ 160, 80, 57 },
		{ 160, 81, 57 },
		{ 160, 82, 57 },
		{ 160, 83, 57 },
		{ 160, 84, 57 },
		{ 160, 85, 57 },
		{ 160, 86, 57 },
		{ 160, 87, 57 },
		{ 160, 88, 57 },
		{ 160, 89, 57 },
		{ 160, 90, 57 },
		{ 160, 91, 57 },
		{ 160, 92, 57 },
		{ 160, 93, 57 },
		{ 160, 94, 57 },
		{ 160, 95, 57 },
		{ 160, 96, 57 },
		{ 160, 97, 57 },
		{ 160, 98, 57 },
		{ 160, 99, 57 },
		{ 160, 100, 57 },
		{ 160, 101, 57 },
		{ 160, 102, 57 },
		{ 160, 103, 57 },
		{ 160, 104, 57 },
		{ 160, 105, 57 },
		{ 160, 106, 57 },
		{ 160, 107, 57 },
		{ 160, 108, 57 },
		{ 160, 109, 57 },
		{ 160, 110, 57 },
		{ 160, 111, 57 },
		{ 160, 112, 57 },
		{ 160, 113, 57 },
		{ 160, 114, 57 },
		{ 160, 115, 57 },
		{ 160, 116, 57 },
		{ 160, 117, 57 },
		{ 160, 118, 57 },
		{ 160, 119, 57 },
		{ 160, 120, 57 },
		{ 160, 121, 57 },
		{ 160, 122, 57 },
		{ 160, 123, 57 },
		{ 160, 124, 57 },
		{ 160, 125, 57 },
		{ 160, 126, 57 },
		{ 160, 127, 57 },
		{ 161, 66, 57 },
		{ 161, 67, 57 },
		{ 161, 68, 57 },
		{ 161, 69, 57 },
		{ 161, 70, 57 },
		{ 161, 71, 57 },
		{ 161, 72, 57 },
		{ 161, 73, 57 },
		{ 161, 74, 57 },
		{ 161, 75, 57 },
		{ 161, 76, 57 },
		{ 161, 77, 57 },
		{ 161, 78, 57 },
		{ 161, 79, 57 },
		{ 161, 80, 57 },
		{ 161, 81, 57 },
		{ 161, 82, 57 },
		{ 161, 83, 57 },
		{ 161, 84, 57 },
		{ 161, 85, 57 },
		{ 161, 86, 57 },
		{ 161, 87, 57 },
		{ 161, 88, 57 },
		{ 161, 89, 57 },
		{ 161, 90, 57 },
		{ 161, 91, 57 },
		{ 161, 92, 57 },
		{ 161, 93, 57 },
		{ 161, 94, 57 },
		{ 161, 95, 57 },
		{ 161, 96, 57 },
		{ 161, 97, 57 },
		{ 161, 98, 57 },
		{ 161, 99, 57 },
		{ 161, 100, 57 },
		{ 161, 101, 57 },
		{ 161, 102, 57 },
		{ 161, 103, 57 },
		{ 161, 104, 57 },
		{ 161, 105, 57 },
		{ 161, 106, 57 },
		{ 161, 107, 57 },
		{ 161, 108, 57 },
		{ 161, 109, 57 },
		{ 161, 110, 57 },
		{ 161, 111, 57 },
		{ 161, 112, 57 },
		{ 161, 113, 57 },
		{ 161, 114, 57 },
		{ 161, 115, 57 },
		{ 161, 116, 57 },
		{ 161, 117, 57 },
		{ 161, 118, 57 },
		{ 161, 119, 57 },
		{ 161, 120, 57 },
		{ 161, 121, 57 },
		{ 161, 122, 57 },
		{ 161, 123, 57 },
		{ 161, 124, 57 },
		{ 161, 125, 57 },
		{ 161, 126, 57 },
		{ 161, 127, 57 },
		{ 161, 128, 57 },
		{ 162, 68, 57 },
		{ 162, 69, 57 },
		{ 162, 70, 57 },
		{ 162, 71, 57 },
		{ 162, 72, 57 },
		{ 162, 73, 57 },
		{ 162, 74, 57 },
		{ 162, 75, 57 },
		{ 162, 76, 57 },
		{ 162, 77, 57 },
		{ 162, 78, 57 },
		{ 162, 79, 57 },
		{ 162, 80, 57 },
		{ 162, 81, 57 },
		{ 162, 82, 57 },
		{ 162, 83, 57 },
		{ 162, 84, 57 },
		{ 162, 85, 57 },
		{ 162, 86, 57 },
		{ 162, 87, 57 },
		{ 162, 88, 57 },
		{ 162, 89, 57 },
		{ 162, 90, 57 },
		{ 162, 91, 57 },
		{ 162, 92, 57 },
		{ 162, 93, 57 },
		{ 162, 94, 57 },
		{ 162, 95, 57 },
		{ 162, 96, 57 },
		{ 162, 97, 57 },
		{ 162, 98, 57 },
		{ 162, 99, 57 },
		{ 162, 100, 57 },
		{ 162, 101, 57 },
		{ 162, 102, 57 },
		{ 162, 103, 57 },
		{ 162, 104, 57 },
		{ 162, 105, 57 },
		{ 162, 106, 57 },
		{ 162, 107, 57 },
		{ 162, 108, 57 },
		{ 162, 109, 57 },
		{ 162, 110, 57 },
		{ 162, 111, 57 },
		{ 162, 112, 57 },
		{ 162, 113, 57 },
		{ 162, 114, 57 },
		{ 162, 115, 57 },
		{ 162, 116, 57 },
		{ 162, 117, 57 },
		{ 162, 118, 57 },
		{ 162, 119, 57 },
		{ 162, 120, 57 },
		{ 162, 121, 57 },
		{ 162, 122, 57 },
		{ 162, 123, 57 },
		{ 162, 124, 57 },
		{ 162, 125, 57 },
		{ 162, 126, 57 },
		{ 162, 127, 57 },
		{ 162, 128, 57 },
		{ 162, 129, 57 },
		{ 162, 130, 57 },
		{ 163, 70, 57 },
		{ 163, 71, 57 },
		{ 163, 72, 57 },
		{ 163, 73, 57 },
		{ 163, 74, 57 },
		{ 163, 75, 57 },
		{ 163, 76, 57 },
		{ 163, 77, 57 },
		{ 163, 78, 57 },
		{ 163, 79, 57 },
		{ 163, 80, 57 },
		{ 163, 81, 57 },
		{ 163, 82, 57 },
		{ 163, 83, 57 },
		{ 163, 84, 57 },
		{ 163, 85, 57 },
		{ 163, 86, 57 },
		{ 163, 87, 57 },
		{ 163, 88, 57 },
		{ 163, 89, 57 },
		{ 163, 90, 57 },
		{ 163, 91, 57 },
		{ 163, 92, 57 },
		{ 163, 93, 57 },
		{ 163, 94, 57 },
		{ 163, 95, 57 },
		{ 163, 96, 57 },
		{ 163, 97, 57 },
		{ 163, 98, 57 },
		{ 163, 99, 57 },
		{ 163, 100, 57 },
		{ 163, 101, 57 },
		{ 163, 102, 57 },
		{ 163, 103, 57 },
		{ 163, 104, 57 },
		{ 163, 105, 57 },
		{ 163, 106, 57 },
		{ 163, 107, 57 },
		{ 163, 108, 57 },
		{ 163, 109, 57 },
		{ 163, 110, 57 },
		{ 163, 111, 57 },
		{ 163, 112, 57 },
		{ 163, 113, 57 },
		{ 163, 114, 57 },
		{ 163, 115, 57 },
		{ 163, 116, 57 },
		{ 163, 117, 57 },
		{ 163, 118, 57 },
		{ 163, 119, 57 },
		{ 163, 120, 57 },
		{ 163, 121, 57 },
		{ 163, 122, 57 },
		{ 163, 123, 57 },
		{ 163, 124, 57 },
		{ 163, 125, 57 },
		{ 163, 126, 57 },
		{ 163, 127, 57 },
		{ 163, 128, 57 },
		{ 163, 129, 57 },
		{ 163, 130, 57 },
		{ 163, 131, 57 },
		{ 164, 72, 57 },
		{ 164, 73, 57 },
		{ 164, 74, 57 },
		{ 164, 75, 57 },
		{ 164, 76, 57 },
		{ 164, 77, 57 },
		{ 164, 78, 57 },
		{ 164, 79, 57 },
		{ 164, 80, 57 },
		{ 164, 81, 57 },
		{ 164, 82, 57 },
		{ 164, 83, 57 },
		{ 164, 84, 57 },
		{ 164, 85, 57 },
		{ 164, 86, 57 },
		{ 164, 87, 57 },
		{ 164, 88, 57 },
		{ 164, 89, 57 },
		{ 164, 90, 57 },
		{ 164, 91, 57 },
		{ 164, 92, 57 },
		{ 164, 93, 57 },
		{ 164, 94, 57 },
		{ 164, 95, 57 },
		{ 164, 96, 57 },
		{ 164, 97, 57 },
		{ 164, 98, 57 },
		{ 164, 99, 57 },
		{ 164, 100, 57 },
		{ 164, 101, 57 },
		{ 164, 102, 57 },
		{ 164, 103, 57 },
		{ 164, 104, 57 },
		{ 164, 105, 57 },
		{ 164, 106, 57 },
		{ 164, 107, 57 },
		{ 164, 108, 57 },
		{ 164, 109, 57 },
		{ 164, 110, 57 },
		{ 164, 111, 57 },
		{ 164, 112, 57 },
		{ 164, 113, 57 },
		{ 164, 114, 57 },
		{ 164, 115, 57 },
		{ 164, 116, 57 },
		{ 164, 117, 57 },
		{ 164, 118, 57 },
		{ 164, 119, 57 },
		{ 164, 120, 57 },
		{ 164, 121, 57 },
		{ 164, 122, 57 },
		{ 164, 123, 57 },
		{ 164, 124, 57 },
		{ 164, 125, 57 },
		{ 164, 126, 57 },
		{ 164, 127, 57 },
		{ 164, 128, 57 },
		{ 164, 129, 57 },
		{ 164, 130, 57 },
		{ 164, 131, 57 },
		{ 164, 132, 57 },
		{ 165, 74, 57 },
		{ 165, 75, 57 },
		{ 165, 76, 57 },
		{ 165, 77, 57 },
		{ 165, 78, 57 },
		{ 165, 79, 57 },
		{ 165, 80, 57 },
		{ 165, 81, 57 },
		{ 165, 82, 57 },
		{ 165, 83, 57 },
		{ 165, 84, 57 },
		{ 165, 85, 57 },
		{ 165, 86, 57 },
		{ 165, 87, 57 },
		{ 165, 88, 57 },
		{ 165, 89, 57 },
		{ 165, 90, 57 },
		{ 165, 91, 57 },
		{ 165, 92, 57 },
		{ 165, 93, 57 },
		{ 165, 94, 57 },
		{ 165, 95, 57 },
		{ 165, 96, 57 },
		{ 165, 97, 57 },
		{ 165, 98, 57 },
		{ 165, 99, 57 },
		{ 165, 100, 57 },
		{ 165, 101, 57 },
		{ 165, 102, 57 },
		{ 165, 103, 57 },
		{ 165, 104, 57 },
		{ 165, 105, 57 },
		{ 165, 106, 57 },
		{ 165, 107, 57 },
		{ 165, 108, 57 },
		{ 165, 109, 57 },
		{ 165, 110, 57 },
		{ 165, 111, 57 },
		{ 165, 112, 57 },
		{ 165, 113, 57 },
		{ 165, 114, 57 },
		{ 165, 115, 57 },
		{ 165, 116, 57 },
		{ 165, 117, 57 },
		{ 165, 118, 57 },
		{ 165, 119, 57 },
		{ 165, 120, 57 },
		{ 165, 121, 57 },
		{ 165, 122, 57 },
		{ 165, 123, 57 },
		{ 165, 124, 57 },
		{ 165, 125, 57 },
		{ 165, 126, 57 },
		{ 165, 127, 57 },
		{ 165, 128, 57 },
		{ 165, 129, 57 },
		{ 165, 130, 57 },
		{ 165, 131, 57 },
		{ 165, 132, 57 },
		{ 165, 133, 57 },
		{ 165, 134, 57 },
		{ 166, 76, 57 },
		{ 166, 77, 57 },
		{ 166, 78, 57 },
		{ 166, 79, 57 },
		{ 166, 80, 57 },
		{ 166, 81, 57 },
		{ 166, 82, 57 },
		{ 166, 83, 57 },
		{ 166, 84, 57 },
		{ 166, 85, 57 },
		{ 166, 86, 57 },
		{ 166, 87, 57 },
		{ 166, 88, 57 },
		{ 166, 89, 57 },
		{ 166, 90, 57 },
		{ 166, 91, 57 },
		{ 166, 92, 57 },
		{ 166, 93, 57 },
		{ 166, 94, 57 },
		{ 166, 95, 57 },
		{ 166, 96, 57 },
		{ 166, 97, 57 },
		{ 166, 98, 57 },
		{ 166, 99, 57 },
		{ 166, 100, 57 },
		{ 166, 101, 57 },
		{ 166, 102, 57 },
		{ 166, 103, 57 },
		{ 166, 104, 57 },
		{ 166, 105, 57 },
		{ 166, 106, 57 },
		{ 166, 107, 57 },
		{ 166, 108, 57 },
		{ 166, 109, 57 },
		{ 166, 110, 57 },
		{ 166, 111, 57 },
		{ 166, 112, 57 },
		{ 166, 113, 57 },
		{ 166, 114, 57 },
		{ 166, 115, 57 },
		{ 166, 116, 57 },
		{ 166, 117, 57 },
		{ 166, 118, 57 },
		{ 166, 119, 57 },
		{ 166, 120, 57 },
		{ 166, 121, 57 },
		{ 166, 122, 57 },
		{ 166, 123, 57 },
		{ 166, 124, 57 },
		{ 166, 125, 57 },
		{ 166, 126, 57 },
		{ 166, 127, 57 },
		{ 166, 128, 57 },
		{ 166, 129, 57 },
		{ 166, 130, 57 },
		{ 166, 131, 57 },
		{ 166, 132, 57 },
		{ 166, 133, 57 },
		{ 166, 134, 57 },
		{ 166, 135, 57 },
		{ 167, 78, 57 },
		{ 167, 79, 57 },
		{ 167, 80, 57 },
		{ 167, 81, 57 },
		{ 167, 82, 57 },
		{ 167, 83, 57 },
		{ 167, 84, 57 },
		{ 167, 85, 57 },
		{ 167, 86, 57 },
		{ 167, 87, 57 },
		{ 167, 88, 57 },
		{ 167, 89, 57 },
		{ 167, 90, 57 },
		{ 167, 91, 57 },
		{ 167, 92, 57 },
		{ 167, 93, 57 },
		{ 167, 94, 57 },
		{ 167, 95, 57 },
		{ 167, 96, 57 },
		{ 167, 97, 57 },
		{ 167, 98, 57 },
		{ 167, 99, 57 },
		{ 167, 100, 57 },
		{ 167, 101, 57 },
		{ 167, 102, 57 },
		{ 167, 103, 57 },
		{ 167, 104, 57 },
		{ 167, 105, 57 },
		{ 167, 106, 57 },
		{ 167, 107, 57 },
		{ 167, 108, 57 },
		{ 167, 109, 57 },
		{ 167, 110, 57 },
		{ 167, 111, 57 },
		{ 167, 112, 57 },
		{ 167, 113, 57 },
		{ 167, 114, 57 },
		{ 167, 115, 57 },
		{ 167, 116, 57 },
		{ 167, 117, 57 },
		{ 167, 118, 57 },
		{ 167, 119, 57 },
		{ 167, 120, 57 },
		{ 167, 121, 57 },
		{ 167, 122, 57 },
		{ 167, 123, 57 },
		{ 167, 124, 57 },
		{ 167, 125, 57 },
		{ 167, 126, 57 },
		{ 167, 127, 57 },
		{ 167, 128, 57 },
		{ 167, 129, 57 },
		{ 167, 130, 57 },
		{ 167, 131, 57 },
		{ 167, 132, 57 },
		{ 167, 133, 57 },
		{ 167, 134, 57 },
		{ 167, 135, 57 },
		{ 167, 136, 57 },
		{ 168, 80, 57 },
		{ 168, 81, 57 },
		{ 168, 82, 57 },
		{ 168, 83, 57 },
		{ 168, 84, 57 },
		{ 168, 85, 57 },
		{ 168, 86, 57 },
		{ 168, 87, 57 },
		{ 168, 88, 57 },
		{ 168, 89, 57 },
		{ 168, 90, 57 },
		{ 168, 91, 57 },
		{ 168, 92, 57 },
		{ 168, 93, 57 },
		{ 168, 94, 57 },
		{ 168, 95, 57 },
		{ 168, 96, 57 },
		{ 168, 97, 57 },
		{ 168, 98, 57 },
		{ 168, 99, 57 },
		{ 168, 100, 57 },
		{ 168, 101, 57 },
		{ 168, 102, 57 },
		{ 168, 103, 57 },
		{ 168, 104, 57 },
		{ 168, 105, 57 },
		{ 168, 106, 57 },
		{ 168, 107, 57 },
		{ 168, 108, 57 },
		{ 168, 109, 57 },
		{ 168, 110, 57 },
		{ 168, 111, 57 },
		{ 168, 112, 57 },
		{ 168, 113, 57 },
		{ 168, 114, 57 },
		{ 168, 115, 57 },
		{ 168, 116, 57 },
		{ 168, 117, 57 },
		{ 168, 118, 57 },
		{ 168, 119, 57 },
		{ 168, 120, 57 },
		{ 168, 121, 57 },
		{ 168, 122, 57 },
		{ 168, 123, 57 },
		{ 168, 124, 57 },
		{ 168, 125, 57 },
		{ 168, 126, 57 },
		{ 168, 127, 57 },
		{ 168, 128, 57 },
		{ 168, 129, 57 },
		{ 168, 130, 57 },
		{ 168, 131, 57 },
		{ 168, 132, 57 },
		{ 168, 133, 57 },
		{ 168, 134, 57 },
		{ 168, 135, 57 },
		{ 168, 136, 57 },
		{ 168, 137, 57 },
		{ 168, 138, 57 },
		{ 169, 82, 57 },
		{ 169, 83, 57 },
		{ 169, 84, 57 },
		{ 169, 85, 57 },
		{ 169, 86, 57 },
		{ 169, 87, 57 },
		{ 169, 88, 57 },
		{ 169, 89, 57 },
		{ 169, 90, 57 },
		{ 169, 91, 57 },
		{ 169, 92, 57 },
		{ 169, 93, 57 },
		{ 169, 94, 57 },
		{ 169, 95, 57 },
		{ 169, 96, 57 },
		{ 169, 97, 57 },
		{ 169, 98, 57 },
		{ 169, 99, 57 },
		{ 169, 100, 57 },
		{ 169, 101, 57 },
		{ 169, 102, 57 },
		{ 169, 103, 57 },
		{ 169, 104, 57 },
		{ 169, 105, 57 },
		{ 169, 106, 57 },
		{ 169, 107, 57 },
		{ 169, 108, 57 },
		{ 169, 109, 57 },
		{ 169, 110, 57 },
		{ 169, 111, 57 },
		{ 169, 112, 57 },
		{ 169, 113, 57 },
		{ 169, 114, 57 },
		{ 169, 115, 57 },
		{ 169, 116, 57 },
		{ 169, 117, 57 },
		{ 169, 118, 57 },
		{ 169, 119, 57 },
		{ 169, 120, 57 },
		{ 169, 121, 57 },
		{ 169, 122, 57 },
		{ 169, 123, 57 },
		{ 169, 124, 57 },
		{ 169, 125, 57 },
		{ 169, 126, 57 },
		{ 169, 127, 57 },
		{ 169, 128, 57 },
		{ 169, 129, 57 },
		{ 169, 130, 57 },
		{ 169, 131, 57 },
		{ 169, 132, 57 },
		{ 169, 133, 57 },
		{ 169, 134, 57 },
		{ 169, 135, 57 },
		{ 169, 136, 57 },
		{ 169, 137, 57 },
		{ 169, 138, 57 },
		{ 169, 139, 57 },
		{ 170, 84, 57 },
		{ 170, 85, 57 },
		{ 170, 86, 57 },
		{ 170, 87, 57 },
		{ 170, 88, 57 },
		{ 170, 89, 57 },
		{ 170, 90, 57 },
		{ 170, 91, 57 },
		{ 170, 92, 57 },
		{ 170, 93, 57 },
		{ 170, 94, 57 },
		{ 170, 95, 57 },
		{ 170, 96, 57 },
		{ 170, 97, 57 },
		{ 170, 98, 57 },
		{ 170, 99, 57 },
		{ 170, 100, 57 },
		{ 170, 101, 57 },
		{ 170, 102, 57 },
		{ 170, 103, 57 },
		{ 170, 104, 57 },
		{ 170, 105, 57 },
		{ 170, 106, 57 },
		{ 170, 107, 57 },
		{ 170, 108, 57 },
		{ 170, 109, 57 },
		{ 170, 110, 57 },
		{ 170, 111, 57 },
		{ 170, 112, 57 },
		{ 170, 113, 57 },
		{ 170, 114, 57 },
		{ 170, 115, 57 },
		{ 170, 116, 57 },
		{ 170, 117, 57 },
		{ 170, 118, 57 },
		{ 170, 119, 57 },
		{ 170, 120, 57 },
		{ 170, 121, 57 },
		{ 170, 122, 57 },
		{ 170, 123, 57 },
		{ 170, 124, 57 },
		{ 170, 125, 57 },
		{ 170, 126, 57 },
		{ 170, 127, 57 },
		{ 170, 128, 57 },
		{ 170, 129, 57 },
		{ 170, 130, 57 },
		{ 170, 131, 57 },
		{ 170, 132, 57 },
		{ 170, 133, 57 },
		{ 170, 134, 57 },
		{ 170, 135, 57 },
		{ 170, 136, 57 },
		{ 170, 137, 57 },
		{ 170, 138, 57 },
		{ 170, 139, 57 },
		{ 170, 140, 57 },
		{ 171, 86, 57 },
		{ 171, 87, 57 },
		{ 171, 88, 57 },
		{ 171, 89, 57 },
		{ 171, 90, 57 },
		{ 171, 91, 57 },
		{ 171, 92, 57 },
		{ 171, 93, 57 },
		{ 171, 94, 57 },
		{ 171, 95, 57 },
		{ 171, 96, 57 },
		{ 171, 97, 57 },
		{ 171, 98, 57 },
		{ 171, 99, 57 },
		{ 171, 100, 57 },
		{ 171, 101, 57 },
		{ 171, 102, 57 },
		{ 171, 103, 57 },
		{ 171, 104, 57 },
		{ 171, 105, 57 },
		{ 171, 106, 57 },
		{ 171, 107, 57 },
		{ 171, 108, 57 },
		{ 171, 109, 57 },
		{ 171, 110, 57 },
		{ 171, 111, 57 },
		{ 171, 112, 57 },
		{ 171, 113, 57 },
		{ 171, 114, 57 },
		{ 171, 115, 57 },
		{ 171, 116, 57 },
		{ 171, 117, 57 },
		{ 171, 118, 57 },
		{ 171, 119, 57 },
		{ 171, 120, 57 },
		{ 171, 121, 57 },
		{ 171, 122, 57 },
		{ 171, 123, 57 },
		{ 171, 124, 57 },
		{ 171, 125, 57 },
		{ 171, 126, 57 },
		{ 171, 127, 57 },
		{ 171, 128, 57 },
		{ 171, 129, 57 },
		{ 171, 130, 57 },
		{ 171, 131, 57 },
		{ 171, 132, 57 },
		{ 171, 133, 57 },
		{ 171, 134, 57 },
		{ 171, 135, 57 },
		{ 171, 136, 57 },
		{ 171, 137, 57 },
		{ 171, 138, 57 },
		{ 171, 139, 57 },
		{ 171, 140, 57 },
		{ 171, 141, 57 },
		{ 171, 142, 57 },
		{ 172, 88, 57 },
		{ 172, 89, 57 },
		{ 172, 90, 57 },
		{ 172, 91, 57 },
		{ 172, 92, 57 },
		{ 172, 93, 57 },
		{ 172, 94, 57 },
		{ 172, 95, 57 },
		{ 172, 96, 57 },
		{ 172, 97, 57 },
		{ 172, 98, 57 },
		{ 172, 99, 57 },
		{ 172, 100, 57 },
		{ 172, 101, 57 },
		{ 172, 102, 57 },
		{ 172, 103, 57 },
		{ 172, 104, 57 },
		{ 172, 105, 57 },
		{ 172, 106, 57 },
		{ 172, 107, 57 },
		{ 172, 108, 57 },
		{ 172, 109, 57 },
		{ 172, 110, 57 },
		{ 172, 111, 57 },
		{ 172, 112, 57 },
		{ 172, 113, 57 },
		{ 172, 114, 57 },
		{ 172, 115, 57 },
		{ 172, 116, 57 },
		{ 172, 117, 57 },
		{ 172, 118, 57 },
		{ 172, 119, 57 },
		{ 172, 120, 57 },
		{ 172, 121, 57 },
		{ 172, 122, 57 },
		{ 172, 123, 57 },
		{ 172, 124, 57 },
		{ 172, 125, 57 },
		{ 172, 126, 57 },
		{ 172, 127, 57 },
		{ 172, 128, 57 },
		{ 172, 129, 57 },
		{ 172, 130, 57 },
		{ 172, 131, 57 },
		{ 172, 132, 57 },
		{ 172, 133, 57 },
		{ 172, 134, 57 },
		{ 172, 135, 57 },
		{ 172, 136, 57 },
		{ 172, 137, 57 },
		{ 172, 138, 57 },
		{ 172, 139, 57 },
		{ 172, 140, 57 },
		{ 172, 141, 57 },
		{ 172, 142, 57 },
		{ 172, 143, 57 },
		{ 173, 90, 57 },
		{ 173, 91, 57 },
		{ 173, 92, 57 },
		{ 173, 93, 57 },
		{ 173, 94, 57 },
		{ 173, 95, 57 },
		{ 173, 96, 57 },
		{ 173, 97, 57 },
		{ 173, 98, 57 },
		{ 173, 99, 57 },
		{ 173, 100, 57 },
		{ 173, 101, 57 },
		{ 173, 102, 57 },
		{ 173, 103, 57 },
		{ 173, 104, 57 },
		{ 173, 105, 57 },
		{ 173, 106, 57 },
		{ 173, 107, 57 },
		{ 173, 108, 57 },
		{ 173, 109, 57 },
		{ 173, 110, 57 },
		{ 173, 111, 57 },
		{ 173, 112, 57 },
		{ 173, 113, 57 },
		{ 173, 114, 57 },
		{ 173, 115, 57 },
		{ 173, 116, 57 },
		{ 173, 117, 57 },
		{ 173, 118, 57 },
		{ 173, 119, 57 },
		{ 173, 120, 57 },
		{ 173, 121, 57 },
		{ 173, 122, 57 },
		{ 173, 123, 57 },
		{ 173, 124, 57 },
		{ 173, 125, 57 },
		{ 173, 126, 57 },
		{ 173, 127, 57 },
		{ 173, 128, 57 },
		{ 173, 129, 57 },
		{ 173, 130, 57 },
		{ 173, 131, 57 },
		{ 173, 132, 57 },
		{ 173, 133, 57 },
		{ 173, 134, 57 },
		{ 173, 135, 57 },
		{ 173, 136, 57 },
		{ 173, 137, 57 },
		{ 173, 138, 57 },
		{ 173, 139, 57 },
		{ 173, 140, 57 },
		{ 173, 141, 57 },
		{ 173, 142, 57 },
		{ 173, 143, 57 },
		{ 173, 144, 57 },
		{ 174, 92, 57 },
		{ 174, 93, 57 },
		{ 174, 94, 57 },
		{ 174, 95, 57 },
		{ 174, 96, 57 },
		{ 174, 97, 57 },
		{ 174, 98, 57 },
		{ 174, 99, 57 },
		{ 174, 100, 57 },
		{ 174, 101, 57 },
		{ 174, 102, 57 },
		{ 174, 103, 57 },
		{ 174, 104, 57 },
		{ 174, 105, 57 },
		{ 174, 106, 57 },
		{ 174, 107, 57 },
		{ 174, 108, 57 },
		{ 174, 109, 57 },
		{ 174, 110, 57 },
		{ 174, 111, 57 },
		{ 174, 112, 57 },
		{ 174, 113, 57 },
		{ 174, 114, 57 },
		{ 174, 115, 57 },
		{ 174, 116, 57 },
		{ 174, 117, 57 },
		{ 174, 118, 57 },
		{ 174, 119, 57 },
		{ 174, 120, 57 },
		{ 174, 121, 57 },
		{ 174, 122, 57 },
		{ 174, 123, 57 },
		{ 174, 124, 57 },
		{ 174, 125, 57 },
		{ 174, 126, 57 },
		{ 174, 127, 57 },
		{ 174, 128, 57 },
		{ 174, 129, 57 },
		{ 174, 130, 57 },
		{ 174, 131, 57 },
		{ 174, 132, 57 },
		{ 174, 133, 57 },
		{ 174, 134, 57 },
		{ 174, 135, 57 },
		{ 174, 136, 57 },
		{ 174, 137, 57 },
		{ 174, 138, 57 },
		{ 174, 139, 57 },
		{ 174, 140, 57 },
		{ 174, 141, 57 },
		{ 174, 142, 57 },
		{ 174, 143, 57 },
		{ 174, 144, 57 },
		{ 174, 145, 57 },
		{ 174, 146, 57 },
		{ 175, 94, 57 },
		{ 175, 95, 57 },
		{ 175, 96, 57 },
		{ 175, 97, 57 },
		{ 175, 98, 57 },
		{ 175, 99, 57 },
		{ 175, 100, 57 },
		{ 175, 101, 57 },
		{ 175, 102, 57 },
		{ 175, 103, 57 },
		{ 175, 104, 57 },
		{ 175, 105, 57 },
		{ 175, 106, 57 },
		{ 175, 107, 57 },
		{ 175, 108, 57 },
		{ 175, 109, 57 },
		{ 175, 110, 57 },
		{ 175, 111, 57 },
		{ 175, 112, 57 },
		{ 175, 113, 57 },
		{ 175, 114, 57 },
		{ 175, 115, 57 },
		{ 175, 116, 57 },
		{ 175, 117, 57 },
		{ 175, 118, 57 },
		{ 175, 119, 57 },
		{ 175, 120, 57 },
		{ 175, 121, 57 },
		{ 175, 122, 57 },
		{ 175, 123, 57 },
		{ 175, 124, 57 },
		{ 175, 125, 57 },
		{ 175, 126, 57 },
		{ 175, 127, 57 },
		{ 175, 128, 57 },
		{ 175, 129, 57 },
		{ 175, 130, 57 },
		{ 175, 131, 57 },
		{ 175, 132, 57 },
		{ 175, 133, 57 },
		{ 175, 134, 57 },
		{ 175, 135, 57 },
		{ 175, 136, 57 },
		{ 175, 137, 57 },
		{ 175, 138, 57 },
		{ 175, 139, 57 },
		{ 175, 140, 57 },
		{ 175, 141, 57 },
		{ 175, 142, 57 },
		{ 175, 143, 57 },
		{ 175, 144, 57 },
		{ 175, 145, 57 },
		{ 175, 146, 57 },
		{ 175, 147, 57 },
		{ 176, 96, 57 },
		{ 176, 97, 57 },
		{ 176, 98, 57 },
		{ 176, 99, 57 },
		{ 176, 100, 57 },
		{ 176, 101, 57 },
		{ 176, 102, 57 },
		{ 176, 103, 57 },
		{ 176, 104, 57 },
		{ 176, 105, 57 },
		{ 176, 106, 57 },
		{ 176, 107, 57 },
		{ 176, 108, 57 },
		{ 176, 109, 57 },
		{ 176, 110, 57 },
		{ 176, 111, 57 },
		{ 176, 112, 57 },
		{ 176, 113, 57 },
		{ 176, 114, 57 },
		{ 176, 115, 57 },
		{ 176, 116, 57 },
		{ 176, 117, 57 },
		{ 176, 118, 57 },
		{ 176, 119, 57 },
		{ 176, 120, 57 },
		{ 176, 121, 57 },
		{ 176, 122, 57 },
		{ 176, 123, 57 },
		{ 176, 124, 57 },
		{ 176, 125, 57 },
		{ 176, 126, 57 },
		{ 176, 127, 57 },
		{ 176, 128, 57 },
		{ 176, 129, 57 },
		{ 176, 130, 57 },
		{ 176, 131, 57 },
		{ 176, 132, 57 },
		{ 176, 133, 57 },
		{ 176, 134, 57 },
		{ 176, 135, 57 },
		{ 176, 136, 57 },
		{ 176, 137, 57 },
		{ 176, 138, 57 },
		{ 176, 139, 57 },
		{ 176, 140, 57 },
		{ 176, 141, 57 },
		{ 176, 142, 57 },
		{ 176, 143, 57 },
		{ 176, 144, 57 },
		{ 176, 145, 57 },
		{ 176, 146, 57 },
		{ 176, 147, 57 },
		{ 176, 148, 57 },
		{ 177, 98, 57 },
		{ 177, 99, 57 },
		{ 177, 100, 57 },
		{ 177, 101, 57 },
		{ 177, 102, 57 },
		{ 177, 103, 57 },
		{ 177, 104, 57 },
		{ 177, 105, 57 },
		{ 177, 106, 57 },
		{ 177, 107, 57 },
		{ 177, 108, 57 },
		{ 177, 109, 57 },
		{ 177, 110, 57 },
		{ 177, 111, 57 },
		{ 177, 112, 57 },
		{ 177, 113, 57 },
		{ 177, 114, 57 },
		{ 177, 115, 57 },
		{ 177, 116, 57 },
		{ 177, 117, 57 },
		{ 177, 118, 57 },
		{ 177, 119, 57 },
		{ 177, 120, 57 },
		{ 177, 121, 57 },
		{ 177, 122, 57 },
		{ 177, 123, 57 },
		{ 177, 124, 57 },
		{ 177, 125, 57 },
		{ 177, 126, 57 },
		{ 177, 127, 57 },
		{ 177, 128, 57 },
		{ 177, 129, 57 },
		{ 177, 130, 57 },
		{ 177, 131, 57 },
		{ 177, 132, 57 },
		{ 177, 133, 57 },
		{ 177, 134, 57 },
		{ 177, 135, 57 },
		{ 177, 136, 57 },
		{ 177, 137, 57 },
		{ 177, 138, 57 },
		{ 177, 139, 57 },
		{ 177, 140, 57 },
		{ 177, 141, 57 },
		{ 177, 142, 57 },
		{ 177, 143, 57 },
		{ 177, 144, 57 },
		{ 177, 145, 57 },
		{ 177, 146, 57 },
		{ 177, 147, 57 },
		{ 177, 148, 57 },
		{ 177, 149, 57 },
		{ 177, 150, 57 },
		{ 178, 100, 57 },
		{ 178, 101, 57 },
		{ 178, 102, 57 },
		{ 178, 103, 57 },
		{ 178, 104, 57 },
		{ 178, 105, 57 },
		{ 178, 106, 57 },
		{ 178, 107, 57 },
		{ 178, 108, 57 },
		{ 178, 109, 57 },
		{ 178, 110, 57 },
		{ 178, 111, 57 },
		{ 178, 112, 57 },
		{ 178, 113, 57 },
		{ 178, 114, 57 },
		{ 178, 115, 57 },
		{ 178, 116, 57 },
		{ 178, 117, 57 },
		{ 178, 118, 57 },
		{ 178, 119, 57 },
		{ 178, 120, 57 },
		{ 178, 121, 57 },
		{ 178, 122, 57 },
		{ 178, 123, 57 },
		{ 178, 124, 57 },
		{ 178, 125, 57 },
		{ 178, 126, 57 },
		{ 178, 127, 57 },
		{ 178, 128, 57 },
		{ 178, 129, 57 },
		{ 178, 130, 57 },
		{ 178, 131, 57 },
		{ 178, 132, 57 },
		{ 178, 133, 57 },
		{ 178, 134, 57 },
		{ 178, 135, 57 },
		{ 178, 136, 57 },
		{ 178, 137, 57 },
		{ 178, 138, 57 },
		{ 178, 139, 57 },
		{ 178, 140, 57 },
		{ 178, 141, 57 },
		{ 178, 142, 57 },
		{ 178, 143, 57 },
		{ 178, 144, 57 },
		{ 178, 145, 57 },
		{ 178, 146, 57 },
		{ 178, 147, 57 },
		{ 178, 148, 57 },
		{ 178, 149, 57 },
		{ 178, 150, 57 },
		{ 178, 151, 57 },
		{ 179, 102, 57 },
		{ 179, 103, 57 },
		{ 179, 104, 57 },
		{ 179, 105, 57 },
		{ 179, 106, 57 },
		{ 179, 107, 57 },
		{ 179, 108, 57 },
		{ 179, 109, 57 },
		{ 179, 110, 57 },
		{ 179, 111, 57 },
		{ 179, 112, 57 },
		{ 179, 113, 57 },
		{ 179, 114, 57 },
		{ 179, 115, 57 },
		{ 179, 116, 57 },
		{ 179, 117, 57 },
		{ 179, 118, 57 },
		{ 179, 119, 57 },
		{ 179, 120, 57 },
		{ 179, 121, 57 },
		{ 179, 122, 57 },
		{ 179, 123, 57 },
		{ 179, 124, 57 },
		{ 179, 125, 57 },
		{ 179, 126, 57 },
		{ 179, 127, 57 },
		{ 179, 128, 57 },
		{ 179, 129, 57 },
		{ 179, 130, 57 },
		{ 179, 131, 57 },
		{ 179, 132, 57 },
		{ 179, 133, 57 },
		{ 179, 134, 57 },
		{ 179, 135, 57 },
		{ 179, 136, 57 },
		{ 179, 137, 57 },
		{ 179, 138, 57 },
		{ 179, 139, 57 },
		{ 179, 140, 57 },
		{ 179, 141, 57 },
		{ 179, 142, 57 },
		{ 179, 143, 57 },
		{ 179, 144, 57 },
		{ 179, 145, 57 },
		{ 179, 146, 57 },
		{ 179, 147, 57 },
		{ 179, 148, 57 },
		{ 179, 149, 57 },
		{ 179, 150, 57 },
		{ 179, 151, 57 },
		{ 179, 152, 57 },
		{ 180, 104, 57 },
		{ 180, 105, 57 },
		{ 180, 106, 57 },
		{ 180, 107, 57 },
		{ 180, 108, 57 },
		{ 180, 109, 57 },
		{ 180, 110, 57 },
		{ 180, 111, 57 },
		{ 180, 112, 57 },
		{ 180, 113, 57 },
		{ 180, 114, 57 },
		{ 180, 115, 57 },
		{ 180, 116, 57 },
		{ 180, 117, 57 },
		{ 180, 118, 57 },
		{ 180, 119, 57 },
		{ 180, 120, 57 },
		{ 180, 121, 57 },
		{ 180, 122, 57 },
		{ 180, 123, 57 },
		{ 180, 124, 57 },
		{ 180, 125, 57 },
		{ 180, 126, 57 },
		{ 180, 127, 57 },
		{ 180, 128, 57 },
		{ 180, 129, 57 },
		{ 180, 130, 57 },
		{ 180, 131, 57 },
		{ 180, 132, 57 },
		{ 180, 133, 57 },
		{ 180, 134, 57 },
		{ 180, 135, 57 },
		{ 180, 136, 57 },
		{ 180, 137, 57 },
		{ 180, 138, 57 },
		{ 180, 139, 57 },
		{ 180, 140, 57 },
		{ 180, 141, 57 },
		{ 180, 142, 57 },
		{ 180, 143, 57 },
		{ 180, 144, 57 },
		{ 180, 145, 57 },
		{ 180, 146, 57 },
		{ 180, 147, 57 },
		{ 180, 148, 57 },
		{ 180, 149, 57 },
		{ 180, 150, 57 },
		{ 180, 151, 57 },
		{ 180, 152, 57 },
		{ 180, 153, 57 },
		{ 180, 154, 57 },
		{ 181, 106, 57 },
		{ 181, 107, 57 },
		{ 181, 108, 57 },
		{ 181, 109, 57 },
		{ 181, 110, 57 },
		{ 181, 111, 57 },
		{ 181, 112, 57 },
		{ 181, 113, 57 },
		{ 181, 114, 57 },
		{ 181, 115, 57 },
		{ 181, 116, 57 },
		{ 181, 117, 57 },
		{ 181, 118, 57 },
		{ 181, 119, 57 },
		{ 181, 120, 57 },
		{ 181, 121, 57 },
		{ 181, 122, 57 },
		{ 181, 123, 57 },
		{ 181, 124, 57 },
		{ 181, 125, 57 },
		{ 181, 126, 57 },
		{ 181, 127, 57 },
		{ 181, 128, 57 },
		{ 181, 129, 57 },
		{ 181, 130, 57 },
		{ 181, 131, 57 },
		{ 181, 132, 57 },
		{ 181, 133, 57 },
		{ 181, 134, 57 },
		{ 181, 135, 57 },
		{ 181, 136, 57 },
		{ 181, 137, 57 },
		{ 181, 138, 57 },
		{ 181, 139, 57 },
		{ 181, 140, 57 },
		{ 181, 141, 57 },
		{ 181, 142, 57 },
		{ 181, 143, 57 },
		{ 181, 144, 57 },
		{ 181, 145, 57 },
		{ 181, 146, 57 },
		{ 181, 147, 57 },
		{ 181, 148, 57 },
		{ 181, 149, 57 },
		{ 181, 150, 57 },
		{ 181, 151, 57 },
		{ 181, 152, 57 },
		{ 181, 153, 57 },
		{ 181, 154, 57 },
		{ 181, 155, 57 },
		{ 182, 108, 57 },
		{ 182, 109, 57 },
		{ 182, 110, 57 },
		{ 182, 111, 57 },
		{ 182, 112, 57 },
		{ 182, 113, 57 },
		{ 182, 114, 57 },
		{ 182, 115, 57 },
		{ 182, 116, 57 },
		{ 182, 117, 57 },
		{ 182, 118, 57 },
		{ 182, 119, 57 },
		{ 182, 120, 57 },
		{ 182, 121, 57 },
		{ 182, 122, 57 },
		{ 182, 123, 57 },
		{ 182, 124, 57 },
		{ 182, 125, 57 },
		{ 182, 126, 57 },
		{ 182, 127, 57 },
		{ 182, 128, 57 },
		{ 182, 129, 57 },
		{ 182, 130, 57 },
		{ 182, 131, 57 },
		{ 182, 132, 57 },
		{ 182, 133, 57 },
		{ 182, 134, 57 },
		{ 182, 135, 57 },
		{ 182, 136, 57 },
		{ 182, 137, 57 },
		{ 182, 138, 57 },
		{ 182, 139, 57 },
		{ 182, 140, 57 },
		{ 182, 141, 57 },
		{ 182, 142, 57 },
		{ 182, 143, 57 },
		{ 182, 144, 57 },
		{ 182, 145, 57 },
		{ 182, 146, 57 },
		{ 182, 147, 57 },
		{ 182, 148, 57 },
		{ 182, 149, 57 },
		{ 182, 150, 57 },
		{ 182, 151, 57 },
		{ 182, 152, 57 },
		{ 182, 153, 57 },
		{ 182, 154, 57 },
		{ 182, 155, 57 },
		{ 182, 156, 57 },
		{ 183, 110, 57 },
		{ 183, 111, 57 },
		{ 183, 112, 57 },
		{ 183, 113, 57 },
		{ 183, 114, 57 },
		{ 183, 115, 57 },
		{ 183, 116, 57 },
		{ 183, 117, 57 },
		{ 183, 118, 57 },
		{ 183, 119, 57 },
		{ 183, 120, 57 },
		{ 183, 121, 57 },
		{ 183, 122, 57 },
		{ 183, 123, 57 },
		{ 183, 124, 57 },
		{ 183, 125, 57 },
		{ 183, 126, 57 },
		{ 183, 127, 57 },
		{ 183, 128, 57 },
		{ 183, 129, 57 },
		{ 183, 130, 57 },
		{ 183, 131, 57 },
		{ 183, 132, 57 },
		{ 183, 133, 57 },
		{ 183, 134, 57 },
		{ 183, 135, 57 },
		{ 183, 136, 57 },
		{ 183, 137, 57 },
		{ 183, 138, 57 },
		{ 183, 139, 57 },
		{ 183, 140, 57 },
		{ 183, 141, 57 },
		{ 183, 142, 57 },
		{ 183, 143, 57 },
		{ 183, 144, 57 },
		{ 183, 145, 57 },
		{ 183, 146, 57 },
		{ 183, 147, 57 },
		{ 183, 148, 57 },
		{ 183, 149, 57 },
		{ 183, 150, 57 },
		{ 183, 151, 57 },
		{ 183, 152, 57 },
		{ 183, 153, 57 },
		{ 183, 154, 57 },
		{ 183, 155, 57 },
		{ 183, 156, 57 },
		{ 183, 157, 57 },
		{ 183, 158, 57 },
		{ 184, 112, 57 },
		{ 184, 113, 57 },
		{ 184, 114, 57 },
		{ 184, 115, 57 },
		{ 184, 116, 57 },
		{ 184, 117, 57 },
		{ 184, 118, 57 },
		{ 184, 119, 57 },
		{ 184, 120, 57 },
		{ 184, 121, 57 },
		{ 184, 122, 57 },
		{ 184, 123, 57 },
		{ 184, 124, 57 },
		{ 184, 125, 57 },
		{ 184, 126, 57 },
		{ 184, 127, 57 },
		{ 184, 128, 57 },
		{ 184, 129, 57 },
		{ 184, 130, 57 },
		{ 184, 131, 57 },
		{ 184, 132, 57 },
		{ 184, 133, 57 },
		{ 184, 134, 57 },
		{ 184, 135, 57 },
		{ 184, 136, 57 },
		{ 184, 137, 57 },
		{ 184, 138, 57 },
		{ 184, 139, 57 },
		{ 184, 140, 57 },
		{ 184, 141, 57 },
		{ 184, 142, 57 },
		{ 184, 143, 57 },
		{ 184, 144, 57 },
		{ 184, 145, 57 },
		{ 184, 146, 57 },
		{ 184, 147, 57 },
		{ 184, 148, 57 },
		{ 184, 149, 57 },
		{ 184, 150, 57 },
		{ 184, 151, 57 },
		{ 184, 152, 57 },
		{ 184, 153, 57 },
		{ 184, 154, 57 },
		{ 184, 155, 57 },
		{ 184, 156, 57 },
		{ 184, 157, 57 },
		{ 184, 158, 57 },
		{ 184, 159, 57 },
		{ 185, 114, 57 },
		{ 185, 115, 57 },
		{ 185, 116, 57 },
		{ 185, 117, 57 },
		{ 185, 118, 57 },
		{ 185, 119, 57 },
		{ 185, 120, 57 },
		{ 185, 121, 57 },
		{ 185, 122, 57 },
		{ 185, 123, 57 },
		{ 185, 124, 57 },
		{ 185, 125, 57 },
		{ 185, 126, 57 },
		{ 185, 127, 57 },
		{ 185, 128, 57 },
		{ 185, 129, 57 },
		{ 185, 130, 57 },
		{ 185, 131, 57 },
		{ 185, 132, 57 },
		{ 185, 133, 57 },
		{ 185, 134, 57 },
		{ 185, 135, 57 },
		{ 185, 136, 57 },
		{ 185, 137, 57 },
		{ 185, 138, 57 },
		{ 185, 139, 57 },
		{ 185, 140, 57 },
		{ 185, 141, 57 },
		{ 185, 142, 57 },
		{ 185, 143, 57 },
		{ 185, 144, 57 },
		{ 185, 145, 57 },
		{ 185, 146, 57 },
		{ 185, 147, 57 },
		{ 185, 148, 57 },
		{ 185, 149, 57 },
		{ 185, 150, 57 },
		{ 185, 151, 57 },
		{ 185, 152, 57 },
		{ 185, 153, 57 },
		{ 185, 154, 57 },
		{ 185, 155, 57 },
		{ 185, 156, 57 },
		{ 185, 157, 57 },
		{ 185, 158, 57 },
		{ 185, 159, 57 },
		{ 185, 160, 57 },
		{ 186, 116, 57 },
		{ 186, 117, 57 },
		{ 186, 118, 57 },
		{ 186, 119, 57 },
		{ 186, 120, 57 },
		{ 186, 121, 57 },
		{ 186, 122, 57 },
		{ 186, 123, 57 },
		{ 186, 124, 57 },
		{ 186, 125, 57 },
		{ 186, 126, 57 },
		{ 186, 127, 57 },
		{ 186, 128, 57 },
		{ 186, 129, 57 },
		{ 186, 130, 57 },
		{ 186, 131, 57 },
		{ 186, 132, 57 },
		{ 186, 133, 57 },
		{ 186, 134, 57 },
		{ 186, 135, 57 },
		{ 186, 136, 57 },
		{ 186, 137, 57 },
		{ 186, 138, 57 },
		{ 186, 139, 57 },
		{ 186, 140, 57 },
		{ 186, 141, 57 },
		{ 186, 142, 57 },
		{ 186, 143, 57 },
		{ 186, 144, 57 },
		{ 186, 145, 57 },
		{ 186, 146, 57 },
		{ 186, 147, 57 },
		{ 186, 148, 57 },
		{ 186, 149, 57 },
		{ 186, 150, 57 },
		{ 186, 151, 57 },
		{ 186, 152, 57 },
		{ 186, 153, 57 },
		{ 186, 154, 57 },
		{ 186, 155, 57 },
		{ 186, 156, 57 },
		{ 186, 157, 57 },
		{ 186, 158, 57 },
		{ 186, 159, 57 },
		{ 186, 160, 57 },
		{ 186, 161, 57 },
		{ 186, 162, 57 },
		{ 187, 118, 57 },
		{ 187, 119, 57 },
		{ 187, 120, 57 },
		{ 187, 121, 57 },
		{ 187, 122, 57 },
		{ 187, 123, 57 },
		{ 187, 124, 57 },
		{ 187, 125, 57 },
		{ 187, 126, 57 },
		{ 187, 127, 57 },
		{ 187, 128, 57 },
		{ 187, 129, 57 },
		{ 187, 130, 57 },
		{ 187, 131, 57 },
		{ 187, 132, 57 },
		{ 187, 133, 57 },
		{ 187, 134, 57 },
		{ 187, 135, 57 },
		{ 187, 136, 57 },
		{ 187, 137, 57 },
		{ 187, 138, 57 },
		{ 187, 139, 57 },
		{ 187, 140, 57 },
		{ 187, 141, 57 },
		{ 187, 142, 57 },
		{ 187, 143, 57 },
		{ 187, 144, 57 },
		{ 187, 145, 57 },
		{ 187, 146, 57 },
		{ 187, 147, 57 },
		{ 187, 148, 57 },
		{ 187, 149, 57 },
		{ 187, 150, 57 },
		{ 187, 151, 57 },
		{ 187, 152, 57 },
		{ 187, 153, 57 },
		{ 187, 154, 57 },
		{ 187, 155, 57 },
		{ 187, 156, 57 },
		{ 187, 157, 57 },
		{ 187, 158, 57 },
		{ 187, 159, 57 },
		{ 187, 160, 57 },
		{ 187, 161, 57 },
		{ 187, 162, 57 },
		{ 187, 163, 57 },
		{ 188, 120, 57 },
		{ 188, 121, 57 },
		{ 188, 122, 57 },
		{ 188, 123, 57 },
		{ 188, 124, 57 },
		{ 188, 125, 57 },
		{ 188, 126, 57 },
		{ 188, 127, 57 },
		{ 188, 128, 57 },
		{ 188, 129, 57 },
		{ 188, 130, 57 },
		{ 188, 131, 57 },
		{ 188, 132, 57 },
		{ 188, 133, 57 },
		{ 188, 134, 57 },
		{ 188, 135, 57 },
		{ 188, 136, 57 },
		{ 188, 137, 57 },
		{ 188, 138, 57 },
		{ 188, 139, 57 },
		{ 188, 140, 57 },
		{ 188, 141, 57 },
		{ 188, 142, 57 },
		{ 188, 143, 57 },
		{ 188, 144, 57 },
		{ 188, 145, 57 },
		{ 188, 146, 57 },
		{ 188, 147, 57 },
		{ 188, 148, 57 },
		{ 188, 149, 57 },
		{ 188, 150, 57 },
		{ 188, 151, 57 },
		{ 188, 152, 57 },
		{ 188, 153, 57 },
		{ 188, 154, 57 },
		{ 188, 155, 57 },
		{ 188, 156, 57 },
		{ 188, 157, 57 },
		{ 188, 158, 57 },
		{ 188, 159, 57 },
		{ 188, 160, 57 },
		{ 188, 161, 57 },
		{ 188, 162, 57 },
		{ 188, 163, 57 },
		{ 188, 164, 57 },
		{ 189, 122, 57 },
		{ 189, 123, 57 },
		{ 189, 124, 57 },
		{ 189, 125, 57 },
		{ 189, 126, 57 },
		{ 189, 127, 57 },
		{ 189, 128, 57 },
		{ 189, 129, 57 },
		{ 189, 130, 57 },
		{ 189, 131, 57 },
		{ 189, 132, 57 },
		{ 189, 133, 57 },
		{ 189, 134, 57 },
		{ 189, 135, 57 },
		{ 189, 136, 57 },
		{ 189, 137, 57 },
		{ 189, 138, 57 },
		{ 189, 139, 57 },
		{ 189, 140, 57 },
		{ 189, 141, 57 },
		{ 189, 142, 57 },
		{ 189, 143, 57 },
		{ 189, 144, 57 },
		{ 189, 145, 57 },
		{ 189, 146, 57 },
		{ 189, 147, 57 },
		{ 189, 148, 57 },
		{ 189, 149, 57 },
		{ 189, 150, 57 },
		{ 189, 151, 57 },
		{ 189, 152, 57 },
		{ 189, 153, 57 },
		{ 189, 154, 57 },
		{ 189, 155, 57 },
		{ 189, 156, 57 },
		{ 189, 157, 57 },
		{ 189, 158, 57 },
		{ 189, 159, 57 },
		{ 189, 160, 57 },
		{ 189, 161, 57 },
		{ 189, 162, 57 },
		{ 189, 163, 57 },
		{ 189, 164, 57 },
		{ 189, 165, 57 },
		{ 189, 166, 57 },
		{ 190, 124, 57 },
		{ 190, 125, 57 },
		{ 190, 126, 57 },
		{ 190, 127, 57 },
		{ 190, 128, 57 },
		{ 190, 129, 57 },
		{ 190, 130, 57 },
		{ 190, 131, 57 },
		{ 190, 132, 57 },
		{ 190, 133, 57 },
		{ 190, 134, 57 },
		{ 190, 135, 57 },
		{ 190, 136, 57 },
		{ 190, 137, 57 },
		{ 190, 138, 57 },
		{ 190, 139, 57 },
		{ 190, 140, 57 },
		{ 190, 141, 57 },
		{ 190, 142, 57 },
		{ 190, 143, 57 },
		{ 190, 144, 57 },
		{ 190, 145, 57 },
		{ 190, 146, 57 },
		{ 190, 147, 57 },
		{ 190, 148, 57 },
		{ 190, 149, 57 },
		{ 190, 150, 57 },
		{ 190, 151, 57 },
		{ 190, 152, 57 },
		{ 190, 153, 57 },
		{ 190, 154, 57 },
		{ 190, 155, 57 },
		{ 190, 156, 57 },
		{ 190, 157, 57 },
		{ 190, 158, 57 },
		{ 190, 159, 57 },
		{ 190, 160, 57 },
		{ 190, 161, 57 },
		{ 190, 162, 57 },
		{ 190, 163, 57 },
		{ 190, 164, 57 },
		{ 190, 165, 57 },
		{ 190, 166, 57 },
		{ 190, 167, 57 },
		{ 191, 126, 57 },
		{ 191, 127, 57 },
		{ 191, 128, 57 },
		{ 191, 129, 57 },
		{ 191, 130, 57 },
		{ 191, 131, 57 },
		{ 191, 132, 57 },
		{ 191, 133, 57 },
		{ 191, 134, 57 },
		{ 191, 135, 57 },
		{ 191, 136, 57 },
		{ 191, 137, 57 },
		{ 191, 138, 57 },
		{ 191, 139, 57 },
		{ 191, 140, 57 },
		{ 191, 141, 57 },
		{ 191, 142, 57 },
		{ 191, 143, 57 },
		{ 191, 144, 57 },
		{ 191, 145, 57 },
		{ 191, 146, 57 },
		{ 191, 147, 57 },
		{ 191, 148, 57 },
		{ 191, 149, 57 },
		{ 191, 150, 57 },
		{ 191, 151, 57 },
		{ 191, 152, 57 },
		{ 191, 153, 57 },
		{ 191, 154, 57 },
		{ 191, 155, 57 },
		{ 191, 156, 57 },
		{ 191, 157, 57 },
		{ 191, 158, 57 },
		{ 191, 159, 57 },
		{ 191, 160, 57 },
		{ 191, 161, 57 },
		{ 191, 162, 57 },
		{ 191, 163, 57 },
		{ 191, 164, 57 },
		{ 191, 165, 57 },
		{ 191, 166, 57 },
		{ 191, 167, 57 },
		{ 191, 168, 57 },
		{ 192, 128, 57 },
		{ 192, 129, 57 },
		{ 192, 130, 57 },
		{ 192, 131, 57 },
		{ 192, 132, 57 },
		{ 192, 133, 57 },
		{ 192, 134, 57 },
		{ 192, 135, 57 },
		{ 192, 136, 57 },
		{ 192, 137, 57 },
		{ 192, 138, 57 },
		{ 192, 139, 57 },
		{ 192, 140, 57 },
		{ 192, 141, 57 },
		{ 192, 142, 57 },
		{ 192, 143, 57 },
		{ 192, 144, 57 },
		{ 192, 145, 57 },
		{ 192, 146, 57 },
		{ 192, 147, 57 },
		{ 192, 148, 57 },
		{ 192, 149, 57 },
		{ 192, 150, 57 },
		{ 192, 151, 57 },
		{ 192, 152, 57 },
		{ 192, 153, 57 },
		{ 192, 154, 57 },
		{ 192, 155, 57 },
		{ 192, 156, 57 },
		{ 192, 157, 57 },
		{ 192, 158, 57 },
		{ 192, 159, 57 },
		{ 192, 160, 57 },
		{ 192, 161, 57 },
		{ 192, 162, 57 },
		{ 192, 163, 57 },
		{ 192, 164, 57 },
		{ 192, 165, 57 },
		{ 192, 166, 57 },
		{ 192, 167, 57 },
		{ 192, 168, 57 },
		{ 192, 169, 57 },
		{ 192, 170, 57 },
		{ 193, 130, 57 },
		{ 193, 131, 57 },
		{ 193, 132, 57 },
		{ 193, 133, 57 },
		{ 193, 134, 57 },
		{ 193, 135, 57 },
		{ 193, 136, 57 },
		{ 193, 137, 57 },
		{ 193, 138, 57 },
		{ 193, 139, 57 },
		{ 193, 140, 57 },
		{ 193, 141, 57 },
		{ 193, 142, 57 },
		{ 193, 143, 57 },
		{ 193, 144, 57 },
		{ 193, 145, 57 },
		{ 193, 146, 57 },
		{ 193, 147, 57 },
		{ 193, 148, 57 },
		{ 193, 149, 57 },
		{ 193, 150, 57 },
		{ 193, 151, 57 },
		{ 193, 152, 57 },
		{ 193, 153, 57 },
		{ 193, 154, 57 },
		{ 193, 155, 57 },
		{ 193, 156, 57 },
		{ 193, 157, 57 },
		{ 193, 158, 57 },
		{ 193, 159, 57 },
		{ 193, 160, 57 },
		{ 193, 161, 57 },
		{ 193, 162, 57 },
		{ 193, 163, 57 },
		{ 193, 164, 57 },
		{ 193, 165, 57 },
		{ 193, 166, 57 },
		{ 193, 167, 57 },
		{ 193, 168, 57 },
		{ 193, 169, 57 },
		{ 193, 170, 57 },
		{ 193, 171, 57 },
		{ 194, 132, 57 },
		{ 194, 133, 57 },
		{ 194, 134, 57 },
		{ 194, 135, 57 },
		{ 194, 136, 57 },
		{ 194, 137, 57 },
		{ 194, 138, 57 },
		{ 194, 139, 57 },
		{ 194, 140, 57 },
		{ 194, 141, 57 },
		{ 194, 142, 57 },
		{ 194, 143, 57 },
		{ 194, 144, 57 },
		{ 194, 145, 57 },
		{ 194, 146, 57 },
		{ 194, 147, 57 },
		{ 194, 148, 57 },
		{ 194, 149, 57 },
		{ 194, 150, 57 },
		{ 194, 151, 57 },
		{ 194, 152, 57 },
		{ 194, 153, 57 },
		{ 194, 154, 57 },
		{ 194, 155, 57 },
		{ 194, 156, 57 },
		{ 194, 157, 57 },
		{ 194, 158, 57 },
		{ 194, 159, 57 },
		{ 194, 160, 57 },
		{ 194, 161, 57 },
		{ 194, 162, 57 },
		{ 194, 163, 57 },
		{ 194, 164, 57 },
		{ 194, 165, 57 },
		{ 194, 166, 57 },
		{ 194, 167, 57 },
		{ 194, 168, 57 },
		{ 194, 169, 57 },
		{ 194, 170, 57 },
		{ 194, 171, 57 },
		{ 194, 172, 57 },
		{ 195, 134, 57 },
		{ 195, 135, 57 },
		{ 195, 136, 57 },
		{ 195, 137, 57 },
		{ 195, 138, 57 },
		{ 195, 139, 57 },
		{ 195, 140, 57 },
		{ 195, 141, 57 },
		{ 195, 142, 57 },
		{ 195, 143, 57 },
		{ 195, 144, 57 },
		{ 195, 145, 57 },
		{ 195, 146, 57 },
		{ 195, 147, 57 },
		{ 195, 148, 57 },
		{ 195, 149, 57 },
		{ 195, 150, 57 },
		{ 195, 151, 57 },
		{ 195, 152, 57 },
		{ 195, 153, 57 },
		{ 195, 154, 57 },
		{ 195, 155, 57 },
		{ 195, 156, 57 },
		{ 195, 157, 57 },
		{ 195, 158, 57 },
		{ 195, 159, 57 },
		{ 195, 160, 57 },
		{ 195, 161, 57 },
		{ 195, 162, 57 },
		{ 195, 163, 57 },
		{ 195, 164, 57 },
		{ 195, 165, 57 },
		{ 195, 166, 57 },
		{ 195, 167, 57 },
		{ 195, 168, 57 },
		{ 195, 169, 57 },
		{ 195, 170, 57 },
		{ 195, 171, 57 },
		{ 195, 172, 57 },
		{ 195, 173, 57 },
		{ 195, 174, 57 },
		{ 196, 136, 57 },
		{ 196, 137, 57 },
		{ 196, 138, 57 },
		{ 196, 139, 57 },
		{ 196, 140, 57 },
		{ 196, 141, 57 },
		{ 196, 142, 57 },
		{ 196, 143, 57 },
		{ 196, 144, 57 },
		{ 196, 145, 57 },
		{ 196, 146, 57 },
		{ 196, 147, 57 },
		{ 196, 148, 57 },
		{ 196, 149, 57 },
		{ 196, 150, 57 },
		{ 196, 151, 57 },
		{ 196, 152, 57 },
		{ 196, 153, 57 },
		{ 196, 154, 57 },
		{ 196, 155, 57 },
		{ 196, 156, 57 },
		{ 196, 157, 57 },
		{ 196, 158, 57 },
		{ 196, 159, 57 },
		{ 196, 160, 57 },
		{ 196, 161, 57 },
		{ 196, 162, 57 },
		{ 196, 163, 57 },
		{ 196, 164, 57 },
		{ 196, 165, 57 },
		{ 196, 166, 57 },
		{ 196, 167, 57 },
		{ 196, 168, 57 },
		{ 196, 169, 57 },
		{ 196, 170, 57 },
		{ 196, 171, 57 },
		{ 196, 172, 57 },
		{ 196, 173, 57 },
		{ 196, 174, 57 },
		{ 196, 175, 57 },
		{ 197, 138, 57 },
		{ 197, 139, 57 },
		{ 197, 140, 57 },
		{ 197, 141, 57 },
		{ 197, 142, 57 },
		{ 197, 143, 57 },
		{ 197, 144, 57 },
		{ 197, 145, 57 },
		{ 197, 146, 57 },
		{ 197, 147, 57 },
		{ 197, 148, 57 },
		{ 197, 149, 57 },
		{ 197, 150, 57 },
		{ 197, 151, 57 },
		{ 197, 152, 57 },
		{ 197, 153, 57 },
		{ 197, 154, 57 },
		{ 197, 155, 57 },
		{ 197, 156, 57 },
		{ 197, 157, 57 },
		{ 197, 158, 57 },
		{ 197, 159, 57 },
		{ 197, 160, 57 },
		{ 197, 161, 57 },
		{ 197, 162, 57 },
		{ 197, 163, 57 },
		{ 197, 164, 57 },
		{ 197, 165, 57 },
		{ 197, 166, 57 },
		{ 197, 167, 57 },
		{ 197, 168, 57 },
		{ 197, 169, 57 },
		{ 197, 170, 57 },
		{ 197, 171, 57 },
		{ 197, 172, 57 },
		{ 197, 173, 57 },
		{ 197, 174, 57 },
		{ 197, 175, 57 },
		{ 197, 176, 57 },
		{ 198, 140, 57 },
		{ 198, 141, 57 },
		{ 198, 142, 57 },
		{ 198, 143, 57 },
		{ 198, 144, 57 },
		{ 198, 145, 57 },
		{ 198, 146, 57 },
		{ 198, 147, 57 },
		{ 198, 148, 57 },
		{ 198, 149, 57 },
		{ 198, 150, 57 },
		{ 198, 151, 57 },
		{ 198, 152, 57 },
		{ 198, 153, 57 },
		{ 198, 154, 57 },
		{ 198, 155, 57 },
		{ 198, 156, 57 },
		{ 198, 157, 57 },
		{ 198, 158, 57 },
		{ 198, 159, 57 },
		{ 198, 160, 57 },
		{ 198, 161, 57 },
		{ 198, 162, 57 },
		{ 198, 163, 57 },
		{ 198, 164, 57 },
		{ 198, 165, 57 },
		{ 198, 166, 57 },
		{ 198, 167, 57 },
		{ 198, 168, 57 },
		{ 198, 169, 57 },
		{ 198, 170, 57 },
		{ 198, 171, 57 },
		{ 198, 172, 57 },
		{ 198, 173, 57 },
		{ 198, 174, 57 },
		{ 198, 175, 57 },
		{ 198, 176, 57 },
		{ 198, 177, 57 },
		{ 198, 178, 57 },
		{ 199, 142, 57 },
		{ 199, 143, 57 },
		{ 199, 144, 57 },
		{ 199, 145, 57 },
		{ 199, 146, 57 },
		{ 199, 147, 57 },
		{ 199, 148, 57 },
		{ 199, 149, 57 },
		{ 199, 150, 57 },
		{ 199, 151, 57 },
		{ 199, 152, 57 },
		{ 199, 153, 57 },
		{ 199, 154, 57 },
		{ 199, 155, 57 },
		{ 199, 156, 57 },
		{ 199, 157, 57 },
		{ 199, 158, 57 },
		{ 199, 159, 57 },
		{ 199, 160, 57 },
		{ 199, 161, 57 },
		{ 199, 162, 57 },
		{ 199, 163, 57 },
		{ 199, 164, 57 },
		{ 199, 165, 57 },
		{ 199, 166, 57 },
		{ 199, 167, 57 },
		{ 199, 168, 57 },
		{ 199, 169, 57 },
		{ 199, 170, 57 },
		{ 199, 171, 57 },
		{ 199, 172, 57 },
		{ 199, 173, 57 },
		{ 199, 174, 57 },
		{ 199, 175, 57 },
		{ 199, 176, 57 },
		{ 199, 177, 57 },
		{ 199, 178, 57 },
		{ 199, 179, 57 },
		{ 200, 144, 57 },
		{ 200, 145, 57 },
		{ 200, 146, 57 },
		{ 200, 147, 57 },
		{ 200, 148, 57 },
		{ 200, 149, 57 },
		{ 200, 150, 57 },
		{ 200, 151, 57 },
		{ 200, 152, 57 },
		{ 200, 153, 57 },
		{ 200, 154, 57 },
		{ 200, 155, 57 },
		{ 200, 156, 57 },
		{ 200, 157, 57 },
		{ 200, 158, 57 },
		{ 200, 159, 57 },
		{ 200, 160, 57 },
		{ 200, 161, 57 },
		{ 200, 162, 57 },
		{ 200, 163, 57 },
		{ 200, 164, 57 },
		{ 200, 165, 57 },
		{ 200, 166, 57 },
		{ 200, 167, 57 },
		{ 200, 168, 57 },
		{ 200, 169, 57 },
		{ 200, 170, 57 },
		{ 200, 171, 57 },
		{ 200, 172, 57 },
		{ 200, 173, 57 },
		{ 200, 174, 57 },
		{ 200, 175, 57 },
		{ 200, 176, 57 },
		{ 200, 177, 57 },
		{ 200, 178, 57 },
		{ 200, 179, 57 },
		{ 200, 180, 57 },
		{ 201, 146, 57 },
		{ 201, 147, 57 },
		{ 201, 148, 57 },
		{ 201, 149, 57 },
		{ 201, 150, 57 },
		{ 201, 151, 57 },
		{ 201, 152, 57 },
		{ 201, 153, 57 },
		{ 201, 154, 57 },
		{ 201, 155, 57 },
		{ 201, 156, 57 },
		{ 201, 157, 57 },
		{ 201, 158, 57 },
		{ 201, 159, 57 },
		{ 201, 160, 57 },
		{ 201, 161, 57 },
		{ 201, 162, 57 },
		{ 201, 163, 57 },
		{ 201, 164, 57 },
		{ 201, 165, 57 },
		{ 201, 166, 57 },
		{ 201, 167, 57 },
		{ 201, 168, 57 },
		{ 201, 169, 57 },
		{ 201, 170, 57 },
		{ 201, 171, 57 },
		{ 201, 172, 57 },
		{ 201, 173, 57 },
		{ 201, 174, 57 },
		{ 201, 175, 57 },
		{ 201, 176, 57 },
		{ 201, 177, 57 },
		{ 201, 178, 57 },
		{ 201, 179, 57 },
		{ 201, 180, 57 },
		{ 201, 181, 57 },
		{ 201, 182, 57 },
		{ 202, 148, 57 },
		{ 202, 149, 57 },
		{ 202, 150, 57 },
		{ 202, 151, 57 },
		{ 202, 152, 57 },
		{ 202, 153, 57 },
		{ 202, 154, 57 },
		{ 202, 155, 57 },
		{ 202, 156, 57 },
		{ 202, 157, 57 },
		{ 202, 158, 57 },
		{ 202, 159, 57 },
		{ 202, 160, 57 },
		{ 202, 161, 57 },
		{ 202, 162, 57 },
		{ 202, 163, 57 },
		{ 202, 164, 57 },
		{ 202, 165, 57 },
		{ 202, 166, 57 },
		{ 202, 167, 57 },
		{ 202, 168, 57 },
		{ 202, 169, 57 },
		{ 202, 170, 57 },
		{ 202, 171, 57 },
		{ 202, 172, 57 },
		{ 202, 173, 57 },
		{ 202, 174, 57 },
		{ 202, 175, 57 },
		{ 202, 176, 57 },
		{ 202, 177, 57 },
		{ 202, 178, 57 },
		{ 202, 179, 57 },
		{ 202, 180, 57 },
		{ 202, 181, 57 },
		{ 202, 182, 57 },
		{ 202, 183, 57 },
		{ 203, 150, 57 },
		{ 203, 151, 57 },
		{ 203, 152, 57 },
		{ 203, 153, 57 },
		{ 203, 154, 57 },
		{ 203, 155, 57 },
		{ 203, 156, 57 },
		{ 203, 157, 57 },
		{ 203, 158, 57 },
		{ 203, 159, 57 },
		{ 203, 160, 57 },
		{ 203, 161, 57 },
		{ 203, 162, 57 },
		{ 203, 163, 57 },
		{ 203, 164, 57 },
		{ 203, 165, 57 },
		{ 203, 166, 57 },
		{ 203, 167, 57 },
		{ 203, 168, 57 },
		{ 203, 169, 57 },
		{ 203, 170, 57 },
		{ 203, 171, 57 },
		{ 203, 172, 57 },
		{ 203, 173, 57 },
		{ 203, 174, 57 },
		{ 203, 175, 57 },
		{ 203, 176, 57 },
		{ 203, 177, 57 },
		{ 203, 178, 57 },
		{ 203, 179, 57 },
		{ 203, 180, 57 },
		{ 203, 181, 57 },
		{ 203, 182, 57 },
		{ 203, 183, 57 },
		{ 203, 184, 57 },
		{ 204, 152, 57 },
		{ 204, 153, 57 },
		{ 204, 154, 57 },
		{ 204, 155, 57 },
		{ 204, 156, 57 },
		{ 204, 157, 57 },
		{ 204, 158, 57 },
		{ 204, 159, 57 },
		{ 204, 160, 57 },
		{ 204, 161, 57 },
		{ 204, 162, 57 },
		{ 204, 163, 57 },
		{ 204, 164, 57 },
		{ 204, 165, 57 },
		{ 204, 166, 57 },
		{ 204, 167, 57 },
		{ 204, 168, 57 },
		{ 204, 169, 57 },
		{ 204, 170, 57 },
		{ 204, 171, 57 },
		{ 204, 172, 57 },
		{ 204, 173, 57 },
		{ 204, 174, 57 },
		{ 204, 175, 57 },
		{ 204, 176, 57 },
		{ 204, 177, 57 },
		{ 204, 178, 57 },
		{ 204, 179, 57 },
		{ 204, 180, 57 },
		{ 204, 181, 57 },
		{ 204, 182, 57 },
		{ 204, 183, 57 },
		{ 204, 184, 57 },
		{ 204, 185, 57 },
		{ 204, 186, 57 },
		{ 205, 154, 57 },
		{ 205, 155, 57 },
		{ 205, 156, 57 },
		{ 205, 157, 57 },
		{ 205, 158, 57 },
		{ 205, 159, 57 },
		{ 205, 160, 57 },
		{ 205, 161, 57 },
		{ 205, 162, 57 },
		{ 205, 163, 57 },
		{ 205, 164, 57 },
		{ 205, 165, 57 },
		{ 205, 166, 57 },
		{ 205, 167, 57 },
		{ 205, 168, 57 },
		{ 205, 169, 57 },
		{ 205, 170, 57 },
		{ 205, 171, 57 },
		{ 205, 172, 57 },
		{ 205, 173, 57 },
		{ 205, 174, 57 },
		{ 205, 175, 57 },
		{ 205, 176, 57 },
		{ 205, 177, 57 },
		{ 205, 178, 57 },
		{ 205, 179, 57 },
		{ 205, 180, 57 },
		{ 205, 181, 57 },
		{ 205, 182, 57 },
		{ 205, 183, 57 },
		{ 205, 184, 57 },
		{ 205, 185, 57 },
		{ 205, 186, 57 },
		{ 205, 187, 57 },
		{ 206, 156, 57 },
		{ 206, 157, 57 },
		{ 206, 158, 57 },
		{ 206, 159, 57 },
		{ 206, 160, 57 },
		{ 206, 161, 57 },
		{ 206, 162, 57 },
		{ 206, 163, 57 },
		{ 206, 164, 57 },
		{ 206, 165, 57 },
		{ 206, 166, 57 },
		{ 206, 167, 57 },
		{ 206, 168, 57 },
		{ 206, 169, 57 },
		{ 206, 170, 57 },
		{ 206, 171, 57 },
		{ 206, 172, 57 },
		{ 206, 173, 57 },
		{ 206, 174, 57 },
		{ 206, 175, 57 },
		{ 206, 176, 57 },
		{ 206, 177, 57 },
		{ 206, 178, 57 },
		{ 206, 179, 57 },
		{ 206, 180, 57 },
		{ 206, 181, 57 },
		{ 206, 182, 57 },
		{ 206, 183, 57 },
		{ 206, 184, 57 },
		{ 206, 185, 57 },
		{ 206, 186, 57 },
		{ 206, 187, 57 },
		{ 206, 188, 57 },
		{ 207, 158, 57 },
		{ 207, 159, 57 },
		{ 207, 160, 57 },
		{ 207, 161, 57 },
		{ 207, 162, 57 },
		{ 207, 163, 57 },
		{ 207, 164, 57 },
		{ 207, 165, 57 },
		{ 207, 166, 57 },
		{ 207, 167, 57 },
		{ 207, 168, 57 },
		{ 207, 169, 57 },
		{ 207, 170, 57 },
		{ 207, 171, 57 },
		{ 207, 172, 57 },
		{ 207, 173, 57 },
		{ 207, 174, 57 },
		{ 207, 175, 57 },
		{ 207, 176, 57 },
		{ 207, 177, 57 },
		{ 207, 178, 57 },
		{ 207, 179, 57 },
		{ 207, 180, 57 },
		{ 207, 181, 57 },
		{ 207, 182, 57 },
		{ 207, 183, 57 },
		{ 207, 184, 57 },
		{ 207, 185, 57 },
		{ 207, 186, 57 },
		{ 207, 187, 57 },
		{ 207, 188, 57 },
		{ 207, 189, 57 },
		{ 208, 160, 57 },
		{ 208, 161, 57 },
		{ 208, 162, 57 },
		{ 208, 163, 57 },
		{ 208, 164, 57 },
		{ 208, 165, 57 },
		{ 208, 166, 57 },
		{ 208, 167, 57 },
		{ 208, 168, 57 },
		{ 208, 169, 57 },
		{ 208, 170, 57 },
		{ 208, 171, 57 },
		{ 208, 172, 57 },
		{ 208, 173, 57 },
		{ 208, 174, 57 },
		{ 208, 175, 57 },
		{ 208, 176, 57 },
		{ 208, 177, 57 },
		{ 208, 178, 57 },
		{ 208, 179, 57 },
		{ 208, 180, 57 },
		{ 208, 181, 57 },
		{ 208, 182, 57 },
		{ 208, 183, 57 },
		{ 208, 184, 57 },
		{ 208, 185, 57 },
		{ 208, 186, 57 },
		{ 208, 187, 57 },
		{ 208, 188, 57 },
		{ 208, 189, 57 },
		{ 208, 190, 57 },
		{ 209, 162, 57 },
		{ 209, 163, 57 },
		{ 209, 164, 57 },
		{ 209, 165, 57 },
		{ 209, 166, 57 },
		{ 209, 167, 57 },
		{ 209, 168, 57 },
		{ 209, 169, 57 },
		{ 209, 170, 57 },
		{ 209, 171, 57 },
		{ 209, 172, 57 },
		{ 209, 173, 57 },
		{ 209, 174, 57 },
		{ 209, 175, 57 },
		{ 209, 176, 57 },
		{ 209, 177, 57 },
		{ 209, 178, 57 },
		{ 209, 179, 57 },
		{ 209, 180, 57 },
		{ 209, 181, 57 },
		{ 209, 182, 57 },
		{ 209, 183, 57 },
		{ 209, 184, 57 },
		{ 209, 185, 57 },
		{ 209, 186, 57 },
		{ 209, 187, 57 },
		{ 209, 188, 57 },
		{ 209, 189, 57 },
		{ 209, 190, 57 },
		{ 210, 164, 57 },
		{ 210, 165, 57 },
		{ 210, 166, 57 },
		{ 210, 167, 57 },
		{ 210, 168, 57 },
		{ 210, 169, 57 },
		{ 210, 170, 57 },
		{ 210, 171, 57 },
		{ 210, 172, 57 },
		{ 210, 173, 57 },
		{ 210, 174, 57 },
		{ 210, 175, 57 },
		{ 210, 176, 57 },
		{ 210, 177, 57 },
		{ 210, 178, 57 },
		{ 210, 179, 57 },
		{ 210, 180, 57 },
		{ 210, 181, 57 },
		{ 210, 182, 57 },
		{ 210, 183, 57 },
		{ 210, 184, 57 },
		{ 210, 185, 57 },
		{ 210, 186, 57 },
		{ 210, 187, 57 },
		{ 210, 188, 57 },
		{ 210, 189, 57 },
		{ 210, 190, 57 },
		{ 210, 191, 57 },
		{ 211, 166, 57 },
		{ 211, 167, 57 },
		{ 211, 168, 57 },
		{ 211, 169, 57 },
		{ 211, 170, 57 },
		{ 211, 171, 57 },
		{ 211, 172, 57 },
		{ 211, 173, 57 },
		{ 211, 174, 57 },
		{ 211, 175, 57 },
		{ 211, 176, 57 },
		{ 211, 177, 57 },
		{ 211, 178, 57 },
		{ 211, 179, 57 },
		{ 211, 180, 57 },
		{ 211, 181, 57 },
		{ 211, 182, 57 },
		{ 211, 183, 57 },
		{ 211, 184, 57 },
		{ 211, 185, 57 },
		{ 211, 186, 57 },
		{ 211, 187, 57 },
		{ 211, 188, 57 },
		{ 211, 189, 57 },
		{ 211, 190, 57 },
		{ 211, 191, 57 },
		{ 212, 168, 57 },
		{ 212, 169, 57 },
		{ 212, 170, 57 },
		{ 212, 171, 57 },
		{ 212, 172, 57 },
		{ 212, 173, 57 },
		{ 212, 174, 57 },
		{ 212, 175, 57 },
		{ 212, 176, 57 },
		{ 212, 177, 57 },
		{ 212, 178, 57 },
		{ 212, 179, 57 },
		{ 212, 180, 57 },
		{ 212, 181, 57 },
		{ 212, 182, 57 },
		{ 212, 183, 57 },
		{ 212, 184, 57 },
		{ 212, 185, 57 },
		{ 212, 186, 57 },
		{ 212, 187, 57 },
		{ 212, 188, 57 },
		{ 212, 189, 57 },
		{ 212, 190, 57 },
		{ 212, 191, 57 },
		{ 213, 170, 57 },
		{ 213, 171, 57 },
		{ 213, 172, 57 },
		{ 213, 173, 57 },
		{ 213, 174, 57 },
		{ 213, 175, 57 },
		{ 213, 176, 57 },
		{ 213, 177, 57 },
		{ 213, 178, 57 },
		{ 213, 179, 57 },
		{ 213, 180, 57 },
		{ 213, 181, 57 },
		{ 213, 182, 57 },
		{ 213, 183, 57 },
		{ 213, 184, 57 },
		{ 213, 185, 57 },
		{ 213, 186, 57 },
		{ 213, 187, 57 },
		{ 213, 188, 57 },
		{ 213, 189, 57 },
		{ 213, 190, 57 },
		{ 213, 191, 57 },
		{ 214, 172, 57 },
		{ 214, 173, 57 },
		{ 214, 174, 57 },
		{ 214, 175, 57 },
		{ 214, 176, 57 },
		{ 214, 177, 57 },
		{ 214, 178, 57 },
		{ 214, 179, 57 },
		{ 214, 180, 57 },
		{ 214, 181, 57 },
		{ 214, 182, 57 },
		{ 214, 183, 57 },
		{ 214, 184, 57 },
		{ 214, 185, 57 },
		{ 214, 186, 57 },
		{ 214, 187, 57 },
		{ 214, 188, 57 },
		{ 214, 189, 57 },
		{ 214, 190, 57 },
		{ 214, 191, 57 },
		{ 215, 174, 57 },
		{ 215, 175, 57 },
		{ 215, 176, 57 },
		{ 215, 177, 57 },
		{ 215, 178, 57 },
		{ 215, 179, 57 },
		{ 215, 180, 57 },
		{ 215, 181, 57 },
		{ 215, 182, 57 },
		{ 215, 183, 57 },
		{ 215, 184, 57 },
		{ 215, 185, 57 },
		{ 215, 186, 57 },
		{ 215, 187, 57 },
		{ 215, 188, 57 },
		{ 215, 189, 57 },
		{ 215, 190, 57 },
		{ 215, 191, 57 },
		{ 216, 176, 57 },
		{ 216, 177, 57 },
		{ 216, 178, 57 },
		{ 216, 179, 57 },
		{ 216, 180, 57 },
		{ 216, 181, 57 },
		{ 216, 182, 57 },
		{ 216, 183, 57 },
		{ 216, 184, 57 },
		{ 216, 185, 57 },
		{ 216, 186, 57 },
		{ 216, 187, 57 },
		{ 216, 188, 57 },
		{ 216, 189, 57 },
		{ 216, 190, 57 },
		{ 216, 191, 57 },
		{ 217, 178, 57 },
		{ 217, 179, 57 },
		{ 217, 180, 57 },
		{ 217, 181, 57 },
		{ 217, 182, 57 },
		{ 217, 183, 57 },
		{ 217, 184, 57 },
		{ 217, 185, 57 },
		{ 217, 186, 57 },
		{ 217, 187, 57 },
		{ 217, 188, 57 },
		{ 217, 189, 57 },
		{ 217, 190, 57 },
		{ 217, 191, 57 },
		{ 218, 180, 57 },
		{ 218, 181, 57 },
		{ 218, 182, 57 },
		{ 218, 183, 57 },
		{ 218, 184, 57 },
		{ 218, 185, 57 },
		{ 218, 186, 57 },
		{ 218, 187, 57 },
		{ 218, 188, 57 },
		{ 218, 189, 57 },
		{ 218, 190, 57 },
		{ 218, 191, 57 },
		{ 219, 182, 57 },
		{ 219, 183, 57 },
		{ 219, 184, 57 },
		{ 219, 185, 57 },
		{ 219, 186, 57 },
		{ 219, 187, 57 },
		{ 219, 188, 57 },
		{ 219, 189, 57 },
		{ 219, 190, 57 },
		{ 219, 191, 57 },
		{ 220, 184, 57 },
		{ 220, 185, 57 },
		{ 220, 186, 57 },
		{ 220, 187, 57 },
		{ 220, 188, 57 },
		{ 220, 189, 57 },
		{ 220, 190, 57 },
		{ 220, 191, 57 },
		{ 221, 186, 57 },
		{ 221, 187, 57 },
		{ 221, 188, 57 },
		{ 221, 189, 57 },
		{ 221, 190, 57 },
		{ 221, 191, 57 },
		{ 222, 188, 57 },
		{ 222, 189, 57 },
		{ 222, 190, 57 },
		{ 222, 191, 57 },
		{ 292, 323, 57 },
		{ 292, 324, 57 },
		{ 292, 325, 57 },
		{ 292, 326, 57 },
		{ 293, 323, 57 },
		{ 293, 324, 57 },
		{ 293, 325, 57 },
		{ 293, 326, 57 },
		{ 293, 327, 57 },
		{ 293, 328, 57 },
		{ 294, 323, 57 },
		{ 294, 324, 57 },
		{ 294, 325, 57 },
		{ 294, 326, 57 },
		{ 294, 327, 57 },
		{ 294, 328, 57 },
		{ 294, 329, 57 },
		{ 294, 330, 57 },
		{ 295, 323, 57 },
		{ 295, 324, 57 },
		{ 295, 325, 57 },
		{ 295, 326, 57 },
		{ 295, 327, 57 },
		{ 295, 328, 57 },
		{ 295, 329, 57 },
		{ 295, 330, 57 },
		{ 295, 331, 57 },
		{ 295, 332, 57 },
		{ 296, 323, 57 },
		{ 296, 324, 57 },
		{ 296, 325, 57 },
		{ 296, 326, 57 },
		{ 296, 327, 57 },
		{ 296, 328, 57 },
		{ 296, 329, 57 },
		{ 296, 330, 57 },
		{ 296, 331, 57 },
		{ 296, 332, 57 },
		{ 296, 333, 57 },
		{ 296, 334, 57 },
		{ 297, 323, 57 },
		{ 297, 324, 57 },
		{ 297, 325, 57 },
		{ 297, 326, 57 },
		{ 297, 327, 57 },
		{ 297, 328, 57 },
		{ 297, 329, 57 },
		{ 297, 330, 57 },
		{ 297, 331, 57 },
		{ 297, 332, 57 },
		{ 297, 333, 57 },
		{ 297, 334, 57 },
		{ 297, 335, 57 },
		{ 297, 336, 57 },
		{ 298, 323, 57 },
		{ 298, 324, 57 },
		{ 298, 325, 57 },
		{ 298, 326, 57 },
		{ 298, 327, 57 },
		{ 298, 328, 57 },
		{ 298, 329, 57 },
		{ 298, 330, 57 },
		{ 298, 331, 57 },
		{ 298, 332, 57 },
		{ 298, 333, 57 },
		{ 298, 334, 57 },
		{ 298, 335, 57 },
		{ 298, 336, 57 },
		{ 298, 337, 57 },
		{ 298, 338, 57 },
		{ 299, 323, 57 },
		{ 299, 324, 57 },
		{ 299, 325, 57 },
		{ 299, 326, 57 },
		{ 299, 327, 57 },
		{ 299, 328, 57 },
		{ 299, 329, 57 },
		{ 299, 330, 57 },
		{ 299, 331, 57 },
		{ 299, 332, 57 },
		{ 299, 333, 57 },
		{ 299, 334, 57 },
		{ 299, 335, 57 },
		{ 299, 336, 57 },
		{ 299, 337, 57 },
		{ 299, 338, 57 },
		{ 299, 339, 57 },
		{ 299, 340, 57 },
		{ 300, 323, 57 },
		{ 300, 324, 57 },
		{ 300, 325, 57 },
		{ 300, 326, 57 },
		{ 300, 327, 57 },
		{ 300, 328, 57 },
		{ 300, 329, 57 },
		{ 300, 330, 57 },
		{ 300, 331, 57 },
		{ 300, 332, 57 },
		{ 300, 333, 57 },
		{ 300, 334, 57 },
		{ 300, 335, 57 },
		{ 300, 336, 57 },
		{ 300, 337, 57 },
		{ 300, 338, 57 },
		{ 300, 339, 57 },
		{ 300, 340, 57 },
		{ 300, 341, 57 },
		{ 300, 342, 57 },
		{ 301, 323, 57 },
		{ 301, 324, 57 },
		{ 301, 325, 57 },
		{ 301, 326, 57 },
		{ 301, 327, 57 },
		{ 301, 328, 57 },
		{ 301, 329, 57 },
		{ 301, 330, 57 },
		{ 301, 331, 57 },
		{ 301, 332, 57 },
		{ 301, 333, 57 },
		{ 301, 334, 57 },
		{ 301, 335, 57 },
		{ 301, 336, 57 },
		{ 301, 337, 57 },
		{ 301, 338, 57 },
		{ 301, 339, 57 },
		{ 301, 340, 57 },
		{ 301, 341, 57 },
		{ 301, 342, 57 },
		{ 301, 343, 57 },
		{ 301, 344, 57 },
		{ 302, 323, 57 },
		{ 302, 324, 57 },
		{ 302, 325, 57 },
		{ 302, 326, 57 },
		{ 302, 327, 57 },
		{ 302, 328, 57 },
		{ 302, 329, 57 },
		{ 302, 330, 57 },
		{ 302, 331, 57 },
		{ 302, 332, 57 },
		{ 302, 333, 57 },
		{ 302, 334, 57 },
		{ 302, 335, 57 },
		{ 302, 336, 57 },
		{ 302, 337, 57 },
		{ 302, 338, 57 },
		{ 302, 339, 57 },
		{ 302, 340, 57 },
		{ 302, 341, 57 },
		{ 302, 342, 57 },
		{ 302, 343, 57 },
		{ 302, 344, 57 },
		{ 302, 345, 57 },
		{ 302, 346, 57 },
		{ 303, 323, 57 },
		{ 303, 324, 57 },
		{ 303, 325, 57 },
		{ 303, 326, 57 },
		{ 303, 327, 57 },
		{ 303, 328, 57 },
		{ 303, 329, 57 },
		{ 303, 330, 57 },
		{ 303, 331, 57 },
		{ 303, 332, 57 },
		{ 303, 333, 57 },
		{ 303, 334, 57 },
		{ 303, 335, 57 },
		{ 303, 336, 57 },
		{ 303, 337, 57 },
		{ 303, 338, 57 },
		{ 303, 339, 57 },
		{ 303, 340, 57 },
		{ 303, 341, 57 },
		{ 303, 342, 57 },
		{ 303, 343, 57 },
		{ 303, 344, 57 },
		{ 303, 345, 57 },
		{ 303, 346, 57 },
		{ 303, 347, 57 },
		{ 303, 348, 57 },
		{ 304, 323, 57 },
		{ 304, 324, 57 },
		{ 304, 325, 57 },
		{ 304, 326, 57 },
		{ 304, 327, 57 },
		{ 304, 328, 57 },
		{ 304, 329, 57 },
		{ 304, 330, 57 },
		{ 304, 331, 57 },
		{ 304, 332, 57 },
		{ 304, 333, 57 },
		{ 304, 334, 57 },
		{ 304, 335, 57 },
		{ 304, 336, 57 },
		{ 304, 337, 57 },
		{ 304, 338, 57 },
		{ 304, 339, 57 },
		{ 304, 340, 57 },
		{ 304, 341, 57 },
		{ 304, 342, 57 },
		{ 304, 343, 57 },
		{ 304, 344, 57 },
		{ 304, 345, 57 },
		{ 304, 346, 57 },
		{ 304, 347, 57 },
		{ 304, 348, 57 },
		{ 304, 349, 57 },
		{ 304, 350, 57 },
		{ 305, 324, 57 },
		{ 305, 325, 57 },
		{ 305, 326, 57 },
		{ 305, 327, 57 },
		{ 305, 328, 57 },
		{ 305, 329, 57 },
		{ 305, 330, 57 },
		{ 305, 331, 57 },
		{ 305, 332, 57 },
		{ 305, 333, 57 },
		{ 305, 334, 57 },
		{ 305, 335, 57 },
		{ 305, 336, 57 },
		{ 305, 337, 57 },
		{ 305, 338, 57 },
		{ 305, 339, 57 },
		{ 305, 340, 57 },
		{ 305, 341, 57 },
		{ 305, 342, 57 },
		{ 305, 343, 57 },
		{ 305, 344, 57 },
		{ 305, 345, 57 },
		{ 305, 346, 57 },
		{ 305, 347, 57 },
		{ 305, 348, 57 },
		{ 305, 349, 57 },
		{ 305, 350, 57 },
		{ 305, 351, 57 },
		{ 305, 352, 57 },
		{ 306, 324, 57 },
		{ 306, 325, 57 },
		{ 306, 326, 57 },
		{ 306, 327, 57 },
		{ 306, 328, 57 },
		{ 306, 329, 57 },
		{ 306, 330, 57 },
		{ 306, 331, 57 },
		{ 306, 332, 57 },
		{ 306, 333, 57 },
		{ 306, 334, 57 },
		{ 306, 335, 57 },
		{ 306, 336, 57 },
		{ 306, 337, 57 },
		{ 306, 338, 57 },
		{ 306, 339, 57 },
		{ 306, 340, 57 },
		{ 306, 341, 57 },
		{ 306, 342, 57 },
		{ 306, 343, 57 },
		{ 306, 344, 57 },
		{ 306, 345, 57 },
		{ 306, 346, 57 },
		{ 306, 347, 57 },
		{ 306, 348, 57 },
		{ 306, 349, 57 },
		{ 306, 350, 57 },
		{ 306, 351, 57 },
		{ 306, 352, 57 },
		{ 306, 353, 57 },
		{ 306, 354, 57 },
		{ 307, 325, 57 },
		{ 307, 326, 57 },
		{ 307, 327, 57 },
		{ 307, 328, 57 },
		{ 307, 329, 57 },
		{ 307, 330, 57 },
		{ 307, 331, 57 },
		{ 307, 332, 57 },
		{ 307, 333, 57 },
		{ 307, 334, 57 },
		{ 307, 335, 57 },
		{ 307, 336, 57 },
		{ 307, 337, 57 },
		{ 307, 338, 57 },
		{ 307, 339, 57 },
		{ 307, 340, 57 },
		{ 307, 341, 57 },
		{ 307, 342, 57 },
		{ 307, 343, 57 },
		{ 307, 344, 57 },
		{ 307, 345, 57 },
		{ 307, 346, 57 },
		{ 307, 347, 57 },
		{ 307, 348, 57 },
		{ 307, 349, 57 },
		{ 307, 350, 57 },
		{ 307, 351, 57 },
		{ 307, 352, 57 },
		{ 307, 353, 57 },
		{ 307, 354, 57 },
		{ 307, 355, 57 },
		{ 307, 356, 57 },
		{ 308, 326, 57 },
		{ 308, 327, 57 },
		{ 308, 328, 57 },
		{ 308, 329, 57 },
		{ 308, 330, 57 },
		{ 308, 331, 57 },
		{ 308, 332, 57 },
		{ 308, 333, 57 },
		{ 308, 334, 57 },
		{ 308, 335, 57 },
		{ 308, 336, 57 },
		{ 308, 337, 57 },
		{ 308, 338, 57 },
		{ 308, 339, 57 },
		{ 308, 340, 57 },
		{ 308, 341, 57 },
		{ 308, 342, 57 },
		{ 308, 343, 57 },
		{ 308, 344, 57 },
		{ 308, 345, 57 },
		{ 308, 346, 57 },
		{ 308, 347, 57 },
		{ 308, 348, 57 },
		{ 308, 349, 57 },
		{ 308, 350, 57 },
		{ 308, 351, 57 },
		{ 308, 352, 57 },
		{ 308, 353, 57 },
		{ 308, 354, 57 },
		{ 308, 355, 57 },
		{ 308, 356, 57 },
		{ 308, 357, 57 },
		{ 308, 358, 57 },
		{ 309, 327, 57 },
		{ 309, 328, 57 },
		{ 309, 329, 57 },
		{ 309, 330, 57 },
		{ 309, 331, 57 },
		{ 309, 332, 57 },
		{ 309, 333, 57 },
		{ 309, 334, 57 },
		{ 309, 335, 57 },
		{ 309, 336, 57 },
		{ 309, 337, 57 },
		{ 309, 338, 57 },
		{ 309, 339, 57 },
		{ 309, 340, 57 },
		{ 309, 341, 57 },
		{ 309, 342, 57 },
		{ 309, 343, 57 },
		{ 309, 344, 57 },
		{ 309, 345, 57 },
		{ 309, 346, 57 },
		{ 309, 347, 57 },
		{ 309, 348, 57 },
		{ 309, 349, 57 },
		{ 309, 350, 57 },
		{ 309, 351, 57 },
		{ 309, 352, 57 },
		{ 309, 353, 57 },
		{ 309, 354, 57 },
		{ 309, 355, 57 },
		{ 309, 356, 57 },
		{ 309, 357, 57 },
		{ 309, 358, 57 },
		{ 309, 359, 57 },
		{ 309, 360, 57 },
		{ 310, 328, 57 },
		{ 310, 329, 57 },
		{ 310, 330, 57 },
		{ 310, 331, 57 },
		{ 310, 332, 57 },
		{ 310, 333, 57 },
		{ 310, 334, 57 },
		{ 310, 335, 57 },
		{ 310, 336, 57 },
		{ 310, 337, 57 },
		{ 310, 338, 57 },
		{ 310, 339, 57 },
		{ 310, 340, 57 },
		{ 310, 341, 57 },
		{ 310, 342, 57 },
		{ 310, 343, 57 },
		{ 310, 344, 57 },
		{ 310, 345, 57 },
		{ 310, 346, 57 },
		{ 310, 347, 57 },
		{ 310, 348, 57 },
		{ 310, 349, 57 },
		{ 310, 350, 57 },
		{ 310, 351, 57 },
		{ 310, 352, 57 },
		{ 310, 353, 57 },
		{ 310, 354, 57 },
		{ 310, 355, 57 },
		{ 310, 356, 57 },
		{ 310, 357, 57 },
		{ 310, 358, 57 },
		{ 310, 359, 57 },
		{ 310, 360, 57 },
		{ 310, 361, 57 },
		{ 310, 362, 57 },
		{ 311, 330, 57 },
		{ 311, 331, 57 },
		{ 311, 332, 57 },
		{ 311, 333, 57 },
		{ 311, 334, 57 },
		{ 311, 335, 57 },
		{ 311, 336, 57 },
		{ 311, 337, 57 },
		{ 311, 338, 57 },
		{ 311, 339, 57 },
		{ 311, 340, 57 },
		{ 311, 341, 57 },
		{ 311, 342, 57 },
		{ 311, 343, 57 },
		{ 311, 344, 57 },
		{ 311, 345, 57 },
		{ 311, 346, 57 },
		{ 311, 347, 57 },
		{ 311, 348, 57 },
		{ 311, 349, 57 },
		{ 311, 350, 57 },
		{ 311, 351, 57 },
		{ 311, 352, 57 },
		{ 311, 353, 57 },
		{ 311, 354, 57 },
		{ 311, 355, 57 },
		{ 311, 356, 57 },
		{ 311, 357, 57 },
		{ 311, 358, 57 },
		{ 311, 359, 57 },
		{ 311, 360, 57 },
		{ 311, 361, 57 },
		{ 311, 362, 57 },
		{ 311, 363, 57 },
		{ 311, 364, 57 },
		{ 312, 331, 57 },
		{ 312, 332, 57 },
		{ 312, 333, 57 },
		{ 312, 334, 57 },
		{ 312, 335, 57 },
		{ 312, 336, 57 },
		{ 312, 337, 57 },
		{ 312, 338, 57 },
		{ 312, 339, 57 },
		{ 312, 340, 57 },
		{ 312, 341, 57 },
		{ 312, 342, 57 },
		{ 312, 343, 57 },
		{ 312, 344, 57 },
		{ 312, 345, 57 },
		{ 312, 346, 57 },
		{ 312, 347, 57 },
		{ 312, 348, 57 },
		{ 312, 349, 57 },
		{ 312, 350, 57 },
		{ 312, 351, 57 },
		{ 312, 352, 57 },
		{ 312, 353, 57 },
		{ 312, 354, 57 },
		{ 312, 355, 57 },
		{ 312, 356, 57 },
		{ 312, 357, 57 },
		{ 312, 358, 57 },
		{ 312, 359, 57 },
		{ 312, 360, 57 },
		{ 312, 361, 57 },
		{ 312, 362, 57 },
		{ 312, 363, 57 },
		{ 312, 364, 57 },
		{ 312, 365, 57 },
		{ 312, 366, 57 },
		{ 313, 332, 57 },
		{ 313, 333, 57 },
		{ 313, 334, 57 },
		{ 313, 335, 57 },
		{ 313, 336, 57 },
		{ 313, 337, 57 },
		{ 313, 338, 57 },
		{ 313, 339, 57 },
		{ 313, 340, 57 },
		{ 313, 341, 57 },
		{ 313, 342, 57 },
		{ 313, 343, 57 },
		{ 313, 344, 57 },
		{ 313, 345, 57 },
		{ 313, 346, 57 },
		{ 313, 347, 57 },
		{ 313, 348, 57 },
		{ 313, 349, 57 },
		{ 313, 350, 57 },
		{ 313, 351, 57 },
		{ 313, 352, 57 },
		{ 313, 353, 57 },
		{ 313, 354, 57 },
		{ 313, 355, 57 },
		{ 313, 356, 57 },
		{ 313, 357, 57 },
		{ 313, 358, 57 },
		{ 313, 359, 57 },
		{ 313, 360, 57 },
		{ 313, 361, 57 },
		{ 313, 362, 57 },
		{ 313, 363, 57 },
		{ 313, 364, 57 },
		{ 313, 365, 57 },
		{ 313, 366, 57 },
		{ 313, 367, 57 },
		{ 313, 368, 57 },
		{ 314, 334, 57 },
		{ 314, 335, 57 },
		{ 314, 336, 57 },
		{ 314, 337, 57 },
		{ 314, 338, 57 },
		{ 314, 339, 57 },
		{ 314, 340, 57 },
		{ 314, 341, 57 },
		{ 314, 342, 57 },
		{ 314, 343, 57 },
		{ 314, 344, 57 },
		{ 314, 345, 57 },
		{ 314, 346, 57 },
		{ 314, 347, 57 },
		{ 314, 348, 57 },
		{ 314, 349, 57 },
		{ 314, 350, 57 },
		{ 314, 351, 57 },
		{ 314, 352, 57 },
		{ 314, 353, 57 },
		{ 314, 354, 57 },
		{ 314, 355, 57 },
		{ 314, 356, 57 },
		{ 314, 357, 57 },
		{ 314, 358, 57 },
		{ 314, 359, 57 },
		{ 314, 360, 57 },
		{ 314, 361, 57 },
		{ 314, 362, 57 },
		{ 314, 363, 57 },
		{ 314, 364, 57 },
		{ 314, 365, 57 },
		{ 314, 366, 57 },
		{ 314, 367, 57 },
		{ 314, 368, 57 },
		{ 314, 369, 57 },
		{ 314, 370, 57 },
		{ 315, 335, 57 },
		{ 315, 336, 57 },
		{ 315, 337, 57 },
		{ 315, 338, 57 },
		{ 315, 339, 57 },
		{ 315, 340, 57 },
		{ 315, 341, 57 },
		{ 315, 342, 57 },
		{ 315, 343, 57 },
		{ 315, 344, 57 },
		{ 315, 345, 57 },
		{ 315, 346, 57 },
		{ 315, 347, 57 },
		{ 315, 348, 57 },
		{ 315, 349, 57 },
		{ 315, 350, 57 },
		{ 315, 351, 57 },
		{ 315, 352, 57 },
		{ 315, 353, 57 },
		{ 315, 354, 57 },
		{ 315, 355, 57 },
		{ 315, 356, 57 },
		{ 315, 357, 57 },
		{ 315, 358, 57 },
		{ 315, 359, 57 },
		{ 315, 360, 57 },
		{ 315, 361, 57 },
		{ 315, 362, 57 },
		{ 315, 363, 57 },
		{ 315, 364, 57 },
		{ 315, 365, 57 },
		{ 315, 366, 57 },
		{ 315, 367, 57 },
		{ 315, 368, 57 },
		{ 315, 369, 57 },
		{ 315, 370, 57 },
		{ 315, 371, 57 },
		{ 315, 372, 57 },
		{ 316, 336, 57 },
		{ 316, 337, 57 },
		{ 316, 338, 57 },
		{ 316, 339, 57 },
		{ 316, 340, 57 },
		{ 316, 341, 57 },
		{ 316, 342, 57 },
		{ 316, 343, 57 },
		{ 316, 344, 57 },
		{ 316, 345, 57 },
		{ 316, 346, 57 },
		{ 316, 347, 57 },
		{ 316, 348, 57 },
		{ 316, 349, 57 },
		{ 316, 350, 57 },
		{ 316, 351, 57 },
		{ 316, 352, 57 },
		{ 316, 353, 57 },
		{ 316, 354, 57 },
		{ 316, 355, 57 },
		{ 316, 356, 57 },
		{ 316, 357, 57 },
		{ 316, 358, 57 },
		{ 316, 359, 57 },
		{ 316, 360, 57 },
		{ 316, 361, 57 },
		{ 316, 362, 57 },
		{ 316, 363, 57 },
		{ 316, 364, 57 },
		{ 316, 365, 57 },
		{ 316, 366, 57 },
		{ 316, 367, 57 },
		{ 316, 368, 57 },
		{ 316, 369, 57 },
		{ 316, 370, 57 },
		{ 316, 371, 57 },
		{ 316, 372, 57 },
		{ 316, 373, 57 },
		{ 316, 374, 57 },
		{ 317, 338, 57 },
		{ 317, 339, 57 },
		{ 317, 340, 57 },
		{ 317, 341, 57 },
		{ 317, 342, 57 },
		{ 317, 343, 57 },
		{ 317, 344, 57 },
		{ 317, 345, 57 },
		{ 317, 346, 57 },
		{ 317, 347, 57 },
		{ 317, 348, 57 },
		{ 317, 349, 57 },
		{ 317, 350, 57 },
		{ 317, 351, 57 },
		{ 317, 352, 57 },
		{ 317, 353, 57 },
		{ 317, 354, 57 },
		{ 317, 355, 57 },
		{ 317, 356, 57 },
		{ 317, 357, 57 },
		{ 317, 358, 57 },
		{ 317, 359, 57 },
		{ 317, 360, 57 },
		{ 317, 361, 57 },
		{ 317, 362, 57 },
		{ 317, 363, 57 },
		{ 317, 364, 57 },
		{ 317, 365, 57 },
		{ 317, 366, 57 },
		{ 317, 367, 57 },
		{ 317, 368, 57 },
		{ 317, 369, 57 },
		{ 317, 370, 57 },
		{ 317, 371, 57 },
		{ 317, 372, 57 },
		{ 317, 373, 57 },
		{ 317, 374, 57 },
		{ 317, 375, 57 },
		{ 317, 376, 57 },
		{ 318, 339, 57 },
		{ 318, 340, 57 },
		{ 318, 341, 57 },
		{ 318, 342, 57 },
		{ 318, 343, 57 },
		{ 318, 344, 57 },
		{ 318, 345, 57 },
		{ 318, 346, 57 },
		{ 318, 347, 57 },
		{ 318, 348, 57 },
		{ 318, 349, 57 },
		{ 318, 350, 57 },
		{ 318, 351, 57 },
		{ 318, 352, 57 },
		{ 318, 353, 57 },
		{ 318, 354, 57 },
		{ 318, 355, 57 },
		{ 318, 356, 57 },
		{ 318, 357, 57 },
		{ 318, 358, 57 },
		{ 318, 359, 57 },
		{ 318, 360, 57 },
		{ 318, 361, 57 },
		{ 318, 362, 57 },
		{ 318, 363, 57 },
		{ 318, 364, 57 },
		{ 318, 365, 57 },
		{ 318, 366, 57 },
		{ 318, 367, 57 },
		{ 318, 368, 57 },
		{ 318, 369, 57 },
		{ 318, 370, 57 },
		{ 318, 371, 57 },
		{ 318, 372, 57 },
		{ 318, 373, 57 },
		{ 318, 374, 57 },
		{ 318, 375, 57 },
		{ 318, 376, 57 },
		{ 318, 377, 57 },
		{ 318, 378, 57 },
		{ 319, 340, 57 },
		{ 319, 341, 57 },
		{ 319, 342, 57 },
		{ 319, 343, 57 },
		{ 319, 344, 57 },
		{ 319, 345, 57 },
		{ 319, 346, 57 },
		{ 319, 347, 57 },
		{ 319, 348, 57 },
		{ 319, 349, 57 },
		{ 319, 350, 57 },
		{ 319, 351, 57 },
		{ 319, 352, 57 },
		{ 319, 353, 57 },
		{ 319, 354, 57 },
		{ 319, 355, 57 },
		{ 319, 356, 57 },
		{ 319, 357, 57 },
		{ 319, 358, 57 },
		{ 319, 359, 57 },
		{ 319, 360, 57 },
		{ 319, 361, 57 },
		{ 319, 362, 57 },
		{ 319, 363, 57 },
		{ 319, 364, 57 },
		{ 319, 365, 57 },
		{ 319, 366, 57 },
		{ 319, 367, 57 },
		{ 319, 368, 57 },
		{ 319, 369, 57 },
		{ 319, 370, 57 },
		{ 319, 371, 57 },
		{ 319, 372, 57 },
		{ 319, 373, 57 },
		{ 319, 374, 57 },
		{ 319, 375, 57 },
		{ 319, 376, 57 },
		{ 319, 377, 57 },
		{ 319, 378, 57 },
		{ 319, 379, 57 },
		{ 319, 380, 57 },
		{ 320, 342, 57 },
		{ 320, 343, 57 },
		{ 320, 344, 57 },
		{ 320, 345, 57 },
		{ 320, 346, 57 },
		{ 320, 347, 57 },
		{ 320, 348, 57 },
		{ 320, 349, 57 },
		{ 320, 350, 57 },
		{ 320, 351, 57 },
		{ 320, 352, 57 },
		{ 320, 353, 57 },
		{ 320, 354, 57 },
		{ 320, 355, 57 },
		{ 320, 356, 57 },
		{ 320, 357, 57 },
		{ 320, 358, 57 },
		{ 320, 359, 57 },
		{ 320, 360, 57 },
		{ 320, 361, 57 },
		{ 320, 362, 57 },
		{ 320, 363, 57 },
		{ 320, 364, 57 },
		{ 320, 365, 57 },
		{ 320, 366, 57 },
		{ 320, 367, 57 },
		{ 320, 368, 57 },
		{ 320, 369, 57 },
		{ 320, 370, 57 },
		{ 320, 371, 57 },
		{ 320, 372, 57 },
		{ 320, 373, 57 },
		{ 320, 374, 57 },
		{ 320, 375, 57 },
		{ 320, 376, 57 },
		{ 320, 377, 57 },
		{ 320, 378, 57 },
		{ 320, 379, 57 },
		{ 320, 380, 57 },
		{ 320, 381, 57 },
		{ 320, 382, 57 },
		{ 321, 343, 57 },
		{ 321, 344, 57 },
		{ 321, 345, 57 },
		{ 321, 346, 57 },
		{ 321, 347, 57 },
		{ 321, 348, 57 },
		{ 321, 349, 57 },
		{ 321, 350, 57 },
		{ 321, 351, 57 },
		{ 321, 352, 57 },
		{ 321, 353, 57 },
		{ 321, 354, 57 },
		{ 321, 355, 57 },
		{ 321, 356, 57 },
		{ 321, 357, 57 },
		{ 321, 358, 57 },
		{ 321, 359, 57 },
		{ 321, 360, 57 },
		{ 321, 361, 57 },
		{ 321, 362, 57 },
		{ 321, 363, 57 },
		{ 321, 364, 57 },
		{ 321, 365, 57 },
		{ 321, 366, 57 },
		{ 321, 367, 57 },
		{ 321, 368, 57 },
		{ 321, 369, 57 },
		{ 321, 370, 57 },
		{ 321, 371, 57 },
		{ 321, 372, 57 },
		{ 321, 373, 57 },
		{ 321, 374, 57 },
		{ 321, 375, 57 },
		{ 321, 376, 57 },
		{ 321, 377, 57 },
		{ 321, 378, 57 },
		{ 321, 379, 57 },
		{ 321, 380, 57 },
		{ 321, 381, 57 },
		{ 321, 382, 57 },
		{ 321, 383, 57 },
		{ 321, 384, 57 },
		{ 322, 344, 57 },
		{ 322, 345, 57 },
		{ 322, 346, 57 },
		{ 322, 347, 57 },
		{ 322, 348, 57 },
		{ 322, 349, 57 },
		{ 322, 350, 57 },
		{ 322, 351, 57 },
		{ 322, 352, 57 },
		{ 322, 353, 57 },
		{ 322, 354, 57 },
		{ 322, 355, 57 },
		{ 322, 356, 57 },
		{ 322, 357, 57 },
		{ 322, 358, 57 },
		{ 322, 359, 57 },
		{ 322, 360, 57 },
		{ 322, 361, 57 },
		{ 322, 362, 57 },
		{ 322, 363, 57 },
		{ 322, 364, 57 },
		{ 322, 365, 57 },
		{ 322, 366, 57 },
		{ 322, 367, 57 },
		{ 322, 368, 57 },
		{ 322, 369, 57 },
		{ 322, 370, 57 },
		{ 322, 371, 57 },
		{ 322, 372, 57 },
		{ 322, 373, 57 },
		{ 322, 374, 57 },
		{ 322, 375, 57 },
		{ 322, 376, 57 },
		{ 322, 377, 57 },
		{ 322, 378, 57 },
		{ 322, 379, 57 },
		{ 322, 380, 57 },
		{ 322, 381, 57 },
		{ 322, 382, 57 },
		{ 322, 383, 57 },
		{ 322, 384, 57 },
		{ 322, 385, 57 },
		{ 322, 386, 57 },
		{ 323, 346, 57 },
		{ 323, 347, 57 },
		{ 323, 348, 57 },
		{ 323, 349, 57 },
		{ 323, 350, 57 },
		{ 323, 351, 57 },
		{ 323, 352, 57 },
		{ 323, 353, 57 },
		{ 323, 354, 57 },
		{ 323, 355, 57 },
		{ 323, 356, 57 },
		{ 323, 357, 57 },
		{ 323, 358, 57 },
		{ 323, 359, 57 },
		{ 323, 360, 57 },
		{ 323, 361, 57 },
		{ 323, 362, 57 },
		{ 323, 363, 57 },
		{ 323, 364, 57 },
		{ 323, 365, 57 },
		{ 323, 366, 57 },
		{ 323, 367, 57 },
		{ 323, 368, 57 },
		{ 323, 369, 57 },
		{ 323, 370, 57 },
		{ 323, 371, 57 },
		{ 323, 372, 57 },
		{ 323, 373, 57 },
		{ 323, 374, 57 },
		{ 323, 375, 57 },
		{ 323, 376, 57 },
		{ 323, 377, 57 },
		{ 323, 378, 57 },
		{ 323, 379, 57 },
		{ 323, 380, 57 },
		{ 323, 381, 57 },
		{ 323, 382, 57 },
		{ 323, 383, 57 },
		{ 323, 384, 57 },
		{ 323, 385, 57 },
		{ 323, 386, 57 },
		{ 323, 387, 57 },
		{ 323, 388, 57 },
		{ 324, 347, 57 },
		{ 324, 348, 57 },
		{ 324, 349, 57 },
		{ 324, 350, 57 },
		{ 324, 351, 57 },
		{ 324, 352, 57 },
		{ 324, 353, 57 },
		{ 324, 354, 57 },
		{ 324, 355, 57 },
		{ 324, 356, 57 },
		{ 324, 357, 57 },
		{ 324, 358, 57 },
		{ 324, 359, 57 },
		{ 324, 360, 57 },
		{ 324, 361, 57 },
		{ 324, 362, 57 },
		{ 324, 363, 57 },
		{ 324, 364, 57 },
		{ 324, 365, 57 },
		{ 324, 366, 57 },
		{ 324, 367, 57 },
		{ 324, 368, 57 },
		{ 324, 369, 57 },
		{ 324, 370, 57 },
		{ 324, 371, 57 },
		{ 324, 372, 57 },
		{ 324, 373, 57 },
		{ 324, 374, 57 },
		{ 324, 375, 57 },
		{ 324, 376, 57 },
		{ 324, 377, 57 },
		{ 324, 378, 57 },
		{ 324, 379, 57 },
		{ 324, 380, 57 },
		{ 324, 381, 57 },
		{ 324, 382, 57 },
		{ 324, 383, 57 },
		{ 324, 384, 57 },
		{ 324, 385, 57 },
		{ 324, 386, 57 },
		{ 324, 387, 57 },
		{ 324, 388, 57 },
		{ 324, 389, 57 },
		{ 324, 390, 57 },
		{ 325, 348, 57 },
		{ 325, 349, 57 },
		{ 325, 350, 57 },
		{ 325, 351, 57 },
		{ 325, 352, 57 },
		{ 325, 353, 57 },
		{ 325, 354, 57 },
		{ 325, 355, 57 },
		{ 325, 356, 57 },
		{ 325, 357, 57 },
		{ 325, 358, 57 },
		{ 325, 359, 57 },
		{ 325, 360, 57 },
		{ 325, 361, 57 },
		{ 325, 362, 57 },
		{ 325, 363, 57 },
		{ 325, 364, 57 },
		{ 325, 365, 57 },
		{ 325, 366, 57 },
		{ 325, 367, 57 },
		{ 325, 368, 57 },
		{ 325, 369, 57 },
		{ 325, 370, 57 },
		{ 325, 371, 57 },
		{ 325, 372, 57 },
		{ 325, 373, 57 },
		{ 325, 374, 57 },
		{ 325, 375, 57 },
		{ 325, 376, 57 },
		{ 325, 377, 57 },
		{ 325, 378, 57 },
		{ 325, 379, 57 },
		{ 325, 380, 57 },
		{ 325, 381, 57 },
		{ 325, 382, 57 },
		{ 325, 383, 57 },
		{ 325, 384, 57 },
		{ 325, 385, 57 },
		{ 325, 386, 57 },
		{ 325, 387, 57 },
		{ 325, 388, 57 },
		{ 325, 389, 57 },
		{ 325, 390, 57 },
		{ 325, 391, 57 },
		{ 325, 392, 57 },
		{ 326, 350, 57 },
		{ 326, 351, 57 },
		{ 326, 352, 57 },
		{ 326, 353, 57 },
		{ 326, 354, 57 },
		{ 326, 355, 57 },
		{ 326, 356, 57 },
		{ 326, 357, 57 },
		{ 326, 358, 57 },
		{ 326, 359, 57 },
		{ 326, 360, 57 },
		{ 326, 361, 57 },
		{ 326, 362, 57 },
		{ 326, 363, 57 },
		{ 326, 364, 57 },
		{ 326, 365, 57 },
		{ 326, 366, 57 },
		{ 326, 367, 57 },
		{ 326, 368, 57 },
		{ 326, 369, 57 },
		{ 326, 370, 57 },
		{ 326, 371, 57 },
		{ 326, 372, 57 },
		{ 326, 373, 57 },
		{ 326, 374, 57 },
		{ 326, 375, 57 },
		{ 326, 376, 57 },
		{ 326, 377, 57 },
		{ 326, 378, 57 },
		{ 326, 379, 57 },
		{ 326, 380, 57 },
		{ 326, 381, 57 },
		{ 326, 382, 57 },
		{ 326, 383, 57 },
		{ 326, 384, 57 },
		{ 326, 385, 57 },
		{ 326, 386, 57 },
		{ 326, 387, 57 },
		{ 326, 388, 57 },
		{ 326, 389, 57 },
		{ 326, 390, 57 },
		{ 326, 391, 57 },
		{ 326, 392, 57 },
		{ 326, 393, 57 },
		{ 326, 394, 57 },
		{ 327, 351, 57 },
		{ 327, 352, 57 },
		{ 327, 353, 57 },
		{ 327, 354, 57 },
		{ 327, 355, 57 },
		{ 327, 356, 57 },
		{ 327, 357, 57 },
		{ 327, 358, 57 },
		{ 327, 359, 57 },
		{ 327, 360, 57 },
		{ 327, 361, 57 },
		{ 327, 362, 57 },
		{ 327, 363, 57 },
		{ 327, 364, 57 },
		{ 327, 365, 57 },
		{ 327, 366, 57 },
		{ 327, 367, 57 },
		{ 327, 368, 57 },
		{ 327, 369, 57 },
		{ 327, 370, 57 },
		{ 327, 371, 57 },
		{ 327, 372, 57 },
		{ 327, 373, 57 },
		{ 327, 374, 57 },
		{ 327, 375, 57 },
		{ 327, 376, 57 },
		{ 327, 377, 57 },
		{ 327, 378, 57 },
		{ 327, 379, 57 },
		{ 327, 380, 57 },
		{ 327, 381, 57 },
		{ 327, 382, 57 },
		{ 327, 383, 57 },
		{ 327, 384, 57 },
		{ 327, 385, 57 },
		{ 327, 386, 57 },
		{ 327, 387, 57 },
		{ 327, 388, 57 },
		{ 327, 389, 57 },
		{ 327, 390, 57 },
		{ 327, 391, 57 },
		{ 327, 392, 57 },
		{ 327, 393, 57 },
		{ 327, 394, 57 },
		{ 327, 395, 57 },
		{ 327, 396, 57 },
		{ 328, 352, 57 },
		{ 328, 353, 57 },
		{ 328, 354, 57 },
		{ 328, 355, 57 },
		{ 328, 356, 57 },
		{ 328, 357, 57 },
		{ 328, 358, 57 },
		{ 328, 359, 57 },
		{ 328, 360, 57 },
		{ 328, 361, 57 },
		{ 328, 362, 57 },
		{ 328, 363, 57 },
		{ 328, 364, 57 },
		{ 328, 365, 57 },
		{ 328, 366, 57 },
		{ 328, 367, 57 },
		{ 328, 368, 57 },
		{ 328, 369, 57 },
		{ 328, 370, 57 },
		{ 328, 371, 57 },
		{ 328, 372, 57 },
		{ 328, 373, 57 },
		{ 328, 374, 57 },
		{ 328, 375, 57 },
		{ 328, 376, 57 },
		{ 328, 377, 57 },
		{ 328, 378, 57 },
		{ 328, 379, 57 },
		{ 328, 380, 57 },
		{ 328, 381, 57 },
		{ 328, 382, 57 },
		{ 328, 383, 57 },
		{ 328, 384, 57 },
		{ 328, 385, 57 },
		{ 328, 386, 57 },
		{ 328, 387, 57 },
		{ 328, 388, 57 },
		{ 328, 389, 57 },
		{ 328, 390, 57 },
		{ 328, 391, 57 },
		{ 328, 392, 57 },
		{ 328, 393, 57 },
		{ 328, 394, 57 },
		{ 328, 395, 57 },
		{ 328, 396, 57 },
		{ 328, 397, 57 },
		{ 328, 398, 57 },
		{ 329, 354, 57 },
		{ 329, 355, 57 },
		{ 329, 356, 57 },
		{ 329, 357, 57 },
		{ 329, 358, 57 },
		{ 329, 359, 57 },
		{ 329, 360, 57 },
		{ 329, 361, 57 },
		{ 329, 362, 57 },
		{ 329, 363, 57 },
		{ 329, 364, 57 },
		{ 329, 365, 57 },
		{ 329, 366, 57 },
		{ 329, 367, 57 },
		{ 329, 368, 57 },
		{ 329, 369, 57 },
		{ 329, 370, 57 },
		{ 329, 371, 57 },
		{ 329, 372, 57 },
		{ 329, 373, 57 },
		{ 329, 374, 57 },
		{ 329, 375, 57 },
		{ 329, 376, 57 },
		{ 329, 377, 57 },
		{ 329, 378, 57 },
		{ 329, 379, 57 },
		{ 329, 380, 57 },
		{ 329, 381, 57 },
		{ 329, 382, 57 },
		{ 329, 383, 57 },
		{ 329, 384, 57 },
		{ 329, 385, 57 },
		{ 329, 386, 57 },
		{ 329, 387, 57 },
		{ 329, 388, 57 },
		{ 329, 389, 57 },
		{ 329, 390, 57 },
		{ 329, 391, 57 },
		{ 329, 392, 57 },
		{ 329, 393, 57 },
		{ 329, 394, 57 },
		{ 329, 395, 57 },
		{ 329, 396, 57 },
		{ 329, 397, 57 },
		{ 329, 398, 57 },
		{ 329, 399, 57 },
		{ 329, 400, 57 },
		{ 330, 355, 57 },
		{ 330, 356, 57 },
		{ 330, 357, 57 },
		{ 330, 358, 57 },
		{ 330, 359, 57 },
		{ 330, 360, 57 },
		{ 330, 361, 57 },
		{ 330, 362, 57 },
		{ 330, 363, 57 },
		{ 330, 364, 57 },
		{ 330, 365, 57 },
		{ 330, 366, 57 },
		{ 330, 367, 57 },
		{ 330, 368, 57 },
		{ 330, 369, 57 },
		{ 330, 370, 57 },
		{ 330, 371, 57 },
		{ 330, 372, 57 },
		{ 330, 373, 57 },
		{ 330, 374, 57 },
		{ 330, 375, 57 },
		{ 330, 376, 57 },
		{ 330, 377, 57 },
		{ 330, 378, 57 },
		{ 330, 379, 57 },
		{ 330, 380, 57 },
		{ 330, 381, 57 },
		{ 330, 382, 57 },
		{ 330, 383, 57 },
		{ 330, 384, 57 },
		{ 330, 385, 57 },
		{ 330, 386, 57 },
		{ 330, 387, 57 },
		{ 330, 388, 57 },
		{ 330, 389, 57 },
		{ 330, 390, 57 },
		{ 330, 391, 57 },
		{ 330, 392, 57 },
		{ 330, 393, 57 },
		{ 330, 394, 57 },
		{ 330, 395, 57 },
		{ 330, 396, 57 },
		{ 330, 397, 57 },
		{ 330, 398, 57 },
		{ 330, 399, 57 },
		{ 330, 400, 57 },
		{ 330, 401, 57 },
		{ 330, 402, 57 },
		{ 331, 356, 57 },
		{ 331, 357, 57 },
		{ 331, 358, 57 },
		{ 331, 359, 57 },
		{ 331, 360, 57 },
		{ 331, 361, 57 },
		{ 331, 362, 57 },
		{ 331, 363, 57 },
		{ 331, 364, 57 },
		{ 331, 365, 57 },
		{ 331, 366, 57 },
		{ 331, 367, 57 },
		{ 331, 368, 57 },
		{ 331, 369, 57 },
		{ 331, 370, 57 },
		{ 331, 371, 57 },
		{ 331, 372, 57 },
		{ 331, 373, 57 },
		{ 331, 374, 57 },
		{ 331, 375, 57 },
		{ 331, 376, 57 },
		{ 331, 377, 57 },
		{ 331, 378, 57 },
		{ 331, 379, 57 },
		{ 331, 380, 57 },
		{ 331, 381, 57 },
		{ 331, 382, 57 },
		{ 331, 383, 57 },
		{ 331, 384, 57 },
		{ 331, 385, 57 },
		{ 331, 386, 57 },
		{ 331, 387, 57 },
		{ 331, 388, 57 },
		{ 331, 389, 57 },
		{ 331, 390, 57 },
		{ 331, 391, 57 },
		{ 331, 392, 57 },
		{ 331, 393, 57 },
		{ 331, 394, 57 },
		{ 331, 395, 57 },
		{ 331, 396, 57 },
		{ 331, 397, 57 },
		{ 331, 398, 57 },
		{ 331, 399, 57 },
		{ 331, 400, 57 },
		{ 331, 401, 57 },
		{ 331, 402, 57 },
		{ 331, 403, 57 },
		{ 331, 404, 57 },
		{ 332, 358, 57 },
		{ 332, 359, 57 },
		{ 332, 360, 57 },
		{ 332, 361, 57 },
		{ 332, 362, 57 },
		{ 332, 363, 57 },
		{ 332, 364, 57 },
		{ 332, 365, 57 },
		{ 332, 366, 57 },
		{ 332, 367, 57 },
		{ 332, 368, 57 },
		{ 332, 369, 57 },
		{ 332, 370, 57 },
		{ 332, 371, 57 },
		{ 332, 372, 57 },
		{ 332, 373, 57 },
		{ 332, 374, 57 },
		{ 332, 375, 57 },
		{ 332, 376, 57 },
		{ 332, 377, 57 },
		{ 332, 378, 57 },
		{ 332, 379, 57 },
		{ 332, 380, 57 },
		{ 332, 381, 57 },
		{ 332, 382, 57 },
		{ 332, 383, 57 },
		{ 332, 384, 57 },
		{ 332, 385, 57 },
		{ 332, 386, 57 },
		{ 332, 387, 57 },
		{ 332, 388, 57 },
		{ 332, 389, 57 },
		{ 332, 390, 57 },
		{ 332, 391, 57 },
		{ 332, 392, 57 },
		{ 332, 393, 57 },
		{ 332, 394, 57 },
		{ 332, 395, 57 },
		{ 332, 396, 57 },
		{ 332, 397, 57 },
		{ 332, 398, 57 },
		{ 332, 399, 57 },
		{ 332, 400, 57 },
		{ 332, 401, 57 },
		{ 332, 402, 57 },
		{ 332, 403, 57 },
		{ 332, 404, 57 },
		{ 332, 405, 57 },
		{ 332, 406, 57 },
		{ 333, 359, 57 },
		{ 333, 360, 57 },
		{ 333, 361, 57 },
		{ 333, 362, 57 },
		{ 333, 363, 57 },
		{ 333, 364, 57 },
		{ 333, 365, 57 },
		{ 333, 366, 57 },
		{ 333, 367, 57 },
		{ 333, 368, 57 },
		{ 333, 369, 57 },
		{ 333, 370, 57 },
		{ 333, 371, 57 },
		{ 333, 372, 57 },
		{ 333, 373, 57 },
		{ 333, 374, 57 },
		{ 333, 375, 57 },
		{ 333, 376, 57 },
		{ 333, 377, 57 },
		{ 333, 378, 57 },
		{ 333, 379, 57 },
		{ 333, 380, 57 },
		{ 333, 381, 57 },
		{ 333, 382, 57 },
		{ 333, 383, 57 },
		{ 333, 384, 57 },
		{ 333, 385, 57 },
		{ 333, 386, 57 },
		{ 333, 387, 57 },
		{ 333, 388, 57 },
		{ 333, 389, 57 },
		{ 333, 390, 57 },
		{ 333, 391, 57 },
		{ 333, 392, 57 },
		{ 333, 393, 57 },
		{ 333, 394, 57 },
		{ 333, 395, 57 },
		{ 333, 396, 57 },
		{ 333, 397, 57 },
		{ 333, 398, 57 },
		{ 333, 399, 57 },
		{ 333, 400, 57 },
		{ 333, 401, 57 },
		{ 333, 402, 57 },
		{ 333, 403, 57 },
		{ 333, 404, 57 },
		{ 333, 405, 57 },
		{ 333, 406, 57 },
		{ 333, 407, 57 },
		{ 333, 408, 57 },
		{ 334, 360, 57 },
		{ 334, 361, 57 },
		{ 334, 362, 57 },
		{ 334, 363, 57 },
		{ 334, 364, 57 },
		{ 334, 365, 57 },
		{ 334, 366, 57 },
		{ 334, 367, 57 },
		{ 334, 368, 57 },
		{ 334, 369, 57 },
		{ 334, 370, 57 },
		{ 334, 371, 57 },
		{ 334, 372, 57 },
		{ 334, 373, 57 },
		{ 334, 374, 57 },
		{ 334, 375, 57 },
		{ 334, 376, 57 },
		{ 334, 377, 57 },
		{ 334, 378, 57 },
		{ 334, 379, 57 },
		{ 334, 380, 57 },
		{ 334, 381, 57 },
		{ 334, 382, 57 },
		{ 334, 383, 57 },
		{ 334, 384, 57 },
		{ 334, 385, 57 },
		{ 334, 386, 57 },
		{ 334, 387, 57 },
		{ 334, 388, 57 },
		{ 334, 389, 57 },
		{ 334, 390, 57 },
		{ 334, 391, 57 },
		{ 334, 392, 57 },
		{ 334, 393, 57 },
		{ 334, 394, 57 },
		{ 334, 395, 57 },
		{ 334, 396, 57 },
		{ 334, 397, 57 },
		{ 334, 398, 57 },
		{ 334, 399, 57 },
		{ 334, 400, 57 },
		{ 334, 401, 57 },
		{ 334, 402, 57 },
		{ 334, 403, 57 },
		{ 334, 404, 57 },
		{ 334, 405, 57 },
		{ 334, 406, 57 },
		{ 334, 407, 57 },
		{ 334, 408, 57 },
		{ 334, 409, 57 },
		{ 334, 410, 57 },
		{ 335, 362, 57 },
		{ 335, 363, 57 },
		{ 335, 364, 57 },
		{ 335, 365, 57 },
		{ 335, 366, 57 },
		{ 335, 367, 57 },
		{ 335, 368, 57 },
		{ 335, 369, 57 },
		{ 335, 370, 57 },
		{ 335, 371, 57 },
		{ 335, 372, 57 },
		{ 335, 373, 57 },
		{ 335, 374, 57 },
		{ 335, 375, 57 },
		{ 335, 376, 57 },
		{ 335, 377, 57 },
		{ 335, 378, 57 },
		{ 335, 379, 57 },
		{ 335, 380, 57 },
		{ 335, 381, 57 },
		{ 335, 382, 57 },
		{ 335, 383, 57 },
		{ 335, 384, 57 },
		{ 335, 385, 57 },
		{ 335, 386, 57 },
		{ 335, 387, 57 },
		{ 335, 388, 57 },
		{ 335, 389, 57 },
		{ 335, 390, 57 },
		{ 335, 391, 57 },
		{ 335, 392, 57 },
		{ 335, 393, 57 },
		{ 335, 394, 57 },
		{ 335, 395, 57 },
		{ 335, 396, 57 },
		{ 335, 397, 57 },
		{ 335, 398, 57 },
		{ 335, 399, 57 },
		{ 335, 400, 57 },
		{ 335, 401, 57 },
		{ 335, 402, 57 },
		{ 335, 403, 57 },
		{ 335, 404, 57 },
		{ 335, 405, 57 },
		{ 335, 406, 57 },
		{ 335, 407, 57 },
		{ 335, 408, 57 },
		{ 335, 409, 57 },
		{ 335, 410, 57 },
		{ 335, 411, 57 },
		{ 335, 412, 57 },
		{ 336, 363, 57 },
		{ 336, 364, 57 },
		{ 336, 365, 57 },
		{ 336, 366, 57 },
		{ 336, 367, 57 },
		{ 336, 368, 57 },
		{ 336, 369, 57 },
		{ 336, 370, 57 },
		{ 336, 371, 57 },
		{ 336, 372, 57 },
		{ 336, 373, 57 },
		{ 336, 374, 57 },
		{ 336, 375, 57 },
		{ 336, 376, 57 },
		{ 336, 377, 57 },
		{ 336, 378, 57 },
		{ 336, 379, 57 },
		{ 336, 380, 57 },
		{ 336, 381, 57 },
		{ 336, 382, 57 },
		{ 336, 383, 57 },
		{ 336, 384, 57 },
		{ 336, 385, 57 },
		{ 336, 386, 57 },
		{ 336, 387, 57 },
		{ 336, 388, 57 },
		{ 336, 389, 57 },
		{ 336, 390, 57 },
		{ 336, 391, 57 },
		{ 336, 392, 57 },
		{ 336, 393, 57 },
		{ 336, 394, 57 },
		{ 336, 395, 57 },
		{ 336, 396, 57 },
		{ 336, 397, 57 },
		{ 336, 398, 57 },
		{ 336, 399, 57 },
		{ 336, 400, 57 },
		{ 336, 401, 57 },
		{ 336, 402, 57 },
		{ 336, 403, 57 },
		{ 336, 404, 57 },
		{ 336, 405, 57 },
		{ 336, 406, 57 },
		{ 336, 407, 57 },
		{ 336, 408, 57 },
		{ 336, 409, 57 },
		{ 336, 410, 57 },
		{ 336, 411, 57 },
		{ 336, 412, 57 },
		{ 336, 413, 57 },
		{ 336, 414, 57 },
		{ 337, 364, 57 },
		{ 337, 365, 57 },
		{ 337, 366, 57 },
		{ 337, 367, 57 },
		{ 337, 368, 57 },
		{ 337, 369, 57 },
		{ 337, 370, 57 },
		{ 337, 371, 57 },
		{ 337, 372, 57 },
		{ 337, 373, 57 },
		{ 337, 374, 57 },
		{ 337, 375, 57 },
		{ 337, 376, 57 },
		{ 337, 377, 57 },
		{ 337, 378, 57 },
		{ 337, 379, 57 },
		{ 337, 380, 57 },
		{ 337, 381, 57 },
		{ 337, 382, 57 },
		{ 337, 383, 57 },
		{ 337, 384, 57 },
		{ 337, 385, 57 },
		{ 337, 386, 57 },
		{ 337, 387, 57 },
		{ 337, 388, 57 },
		{ 337, 389, 57 },
		{ 337, 390, 57 },
		{ 337, 391, 57 },
		{ 337, 392, 57 },
		{ 337, 393, 57 },
		{ 337, 394, 57 },
		{ 337, 395, 57 },
		{ 337, 396, 57 },
		{ 337, 397, 57 },
		{ 337, 398, 57 },
		{ 337, 399, 57 },
		{ 337, 400, 57 },
		{ 337, 401, 57 },
		{ 337, 402, 57 },
		{ 337, 403, 57 },
		{ 337, 404, 57 },
		{ 337, 405, 57 },
		{ 337, 406, 57 },
		{ 337, 407, 57 },
		{ 337, 408, 57 },
		{ 337, 409, 57 },
		{ 337, 410, 57 },
		{ 337, 411, 57 },
		{ 337, 412, 57 },
		{ 337, 413, 57 },
		{ 337, 414, 57 },
		{ 337, 415, 57 },
		{ 337, 416, 57 },
		{ 338, 366, 57 },
		{ 338, 367, 57 },
		{ 338, 368, 57 },
		{ 338, 369, 57 },
		{ 338, 370, 57 },
		{ 338, 371, 57 },
		{ 338, 372, 57 },
		{ 338, 373, 57 },
		{ 338, 374, 57 },
		{ 338, 375, 57 },
		{ 338, 376, 57 },
		{ 338, 377, 57 },
		{ 338, 378, 57 },
		{ 338, 379, 57 },
		{ 338, 380, 57 },
		{ 338, 381, 57 },
		{ 338, 382, 57 },
		{ 338, 383, 57 },
		{ 338, 384, 57 },
		{ 338, 385, 57 },
		{ 338, 386, 57 },
		{ 338, 387, 57 },
		{ 338, 388, 57 },
		{ 338, 389, 57 },
		{ 338, 390, 57 },
		{ 338, 391, 57 },
		{ 338, 392, 57 },
		{ 338, 393, 57 },
		{ 338, 394, 57 },
		{ 338, 395, 57 },
		{ 338, 396, 57 },
		{ 338, 397, 57 },
		{ 338, 398, 57 },
		{ 338, 399, 57 },
		{ 338, 400, 57 },
		{ 338, 401, 57 },
		{ 338, 402, 57 },
		{ 338, 403, 57 },
		{ 338, 404, 57 },
		{ 338, 405, 57 },
		{ 338, 406, 57 },
		{ 338, 407, 57 },
		{ 338, 408, 57 },
		{ 338, 409, 57 },
		{ 338, 410, 57 },
		{ 338, 411, 57 },
		{ 338, 412, 57 },
		{ 338, 413, 57 },
		{ 338, 414, 57 },
		{ 338, 415, 57 },
		{ 338, 416, 57 },
		{ 338, 417, 57 },
		{ 338, 418, 57 },
		{ 339, 367, 57 },
		{ 339, 368, 57 },
		{ 339, 369, 57 },
		{ 339, 370, 57 },
		{ 339, 371, 57 },
		{ 339, 372, 57 },
		{ 339, 373, 57 },
		{ 339, 374, 57 },
		{ 339, 375, 57 },
		{ 339, 376, 57 },
		{ 339, 377, 57 },
		{ 339, 378, 57 },
		{ 339, 379, 57 },
		{ 339, 380, 57 },
		{ 339, 381, 57 },
		{ 339, 382, 57 },
		{ 339, 383, 57 },
		{ 339, 384, 57 },
		{ 339, 385, 57 },
		{ 339, 386, 57 },
		{ 339, 387, 57 },
		{ 339, 388, 57 },
		{ 339, 389, 57 },
		{ 339, 390, 57 },
		{ 339, 391, 57 },
		{ 339, 392, 57 },
		{ 339, 393, 57 },
		{ 339, 394, 57 },
		{ 339, 395, 57 },
		{ 339, 396, 57 },
		{ 339, 397, 57 },
		{ 339, 398, 57 },
		{ 339, 399, 57 },
		{ 339, 400, 57 },
		{ 339, 401, 57 },
		{ 339, 402, 57 },
		{ 339, 403, 57 },
		{ 339, 404, 57 },
		{ 339, 405, 57 },
		{ 339, 406, 57 },
		{ 339, 407, 57 },
		{ 339, 408, 57 },
		{ 339, 409, 57 },
		{ 339, 410, 57 },
		{ 339, 411, 57 },
		{ 339, 412, 57 },
		{ 339, 413, 57 },
		{ 339, 414, 57 },
		{ 339, 415, 57 },
		{ 339, 416, 57 },
		{ 339, 417, 57 },
		{ 339, 418, 57 },
		{ 339, 419, 57 },
		{ 339, 420, 57 },
		{ 340, 368, 57 },
		{ 340, 369, 57 },
		{ 340, 370, 57 },
		{ 340, 371, 57 },
		{ 340, 372, 57 },
		{ 340, 373, 57 },
		{ 340, 374, 57 },
		{ 340, 375, 57 },
		{ 340, 376, 57 },
		{ 340, 377, 57 },
		{ 340, 378, 57 },
		{ 340, 379, 57 },
		{ 340, 380, 57 },
		{ 340, 381, 57 },
		{ 340, 382, 57 },
		{ 340, 383, 57 },
		{ 340, 384, 57 },
		{ 340, 385, 57 },
		{ 340, 386, 57 },
		{ 340, 387, 57 },
		{ 340, 388, 57 },
		{ 340, 389, 57 },
		{ 340, 390, 57 },
		{ 340, 391, 57 },
		{ 340, 392, 57 },
		{ 340, 393, 57 },
		{ 340, 394, 57 },
		{ 340, 395, 57 },
		{ 340, 396, 57 },
		{ 340, 397, 57 },
		{ 340, 398, 57 },
		{ 340, 399, 57 },
		{ 340, 400, 57 },
		{ 340, 401, 57 },
		{ 340, 402, 57 },
		{ 340, 403, 57 },
		{ 340, 404, 57 },
		{ 340, 405, 57 },
		{ 340, 406, 57 },
		{ 340, 407, 57 },
		{ 340, 408, 57 },
		{ 340, 409, 57 },
		{ 340, 410, 57 },
		{ 340, 411, 57 },
		{ 340, 412, 57 },
		{ 340, 413, 57 },
		{ 340, 414, 57 },
		{ 340, 415, 57 },
		{ 340, 416, 57 },
		{ 340, 417, 57 },
		{ 340, 418, 57 },
		{ 340, 419, 57 },
		{ 340, 420, 57 },
		{ 340, 421, 57 },
		{ 340, 422, 57 },
		{ 341, 370, 57 },
		{ 341, 371, 57 },
		{ 341, 372, 57 },
		{ 341, 373, 57 },
		{ 341, 374, 57 },
		{ 341, 375, 57 },
		{ 341, 376, 57 },
		{ 341, 377, 57 },
		{ 341, 378, 57 },
		{ 341, 379, 57 },
		{ 341, 380, 57 },
		{ 341, 381, 57 },
		{ 341, 382, 57 },
		{ 341, 383, 57 },
		{ 341, 384, 57 },
		{ 341, 385, 57 },
		{ 341, 386, 57 },
		{ 341, 387, 57 },
		{ 341, 388, 57 },
		{ 341, 389, 57 },
		{ 341, 390, 57 },
		{ 341, 391, 57 },
		{ 341, 392, 57 },
		{ 341, 393, 57 },
		{ 341, 394, 57 },
		{ 341, 395, 57 },
		{ 341, 396, 57 },
		{ 341, 397, 57 },
		{ 341, 398, 57 },
		{ 341, 399, 57 },
		{ 341, 400, 57 },
		{ 341, 401, 57 },
		{ 341, 402, 57 },
		{ 341, 403, 57 },
		{ 341, 404, 57 },
		{ 341, 405, 57 },
		{ 341, 406, 57 },
		{ 341, 407, 57 },
		{ 341, 408, 57 },
		{ 341, 409, 57 },
		{ 341, 410, 57 },
		{ 341, 411, 57 },
		{ 341, 412, 57 },
		{ 341, 413, 57 },
		{ 341, 414, 57 },
		{ 341, 415, 57 },
		{ 341, 416, 57 },
		{ 341, 417, 57 },
		{ 341, 418, 57 },
		{ 341, 419, 57 },
		{ 341, 420, 57 },
		{ 341, 421, 57 },
		{ 341, 422, 57 },
		{ 341, 423, 57 },
		{ 341, 424, 57 },
		{ 342, 371, 57 },
		{ 342, 372, 57 },
		{ 342, 373, 57 },
		{ 342, 374, 57 },
		{ 342, 375, 57 },
		{ 342, 376, 57 },
		{ 342, 377, 57 },
		{ 342, 378, 57 },
		{ 342, 379, 57 },
		{ 342, 380, 57 },
		{ 342, 381, 57 },
		{ 342, 382, 57 },
		{ 342, 383, 57 },
		{ 342, 384, 57 },
		{ 342, 385, 57 },
		{ 342, 386, 57 },
		{ 342, 387, 57 },
		{ 342, 388, 57 },
		{ 342, 389, 57 },
		{ 342, 390, 57 },
		{ 342, 391, 57 },
		{ 342, 392, 57 },
		{ 342, 393, 57 },
		{ 342, 394, 57 },
		{ 342, 395, 57 },
		{ 342, 396, 57 },
		{ 342, 397, 57 },
		{ 342, 398, 57 },
		{ 342, 399, 57 },
		{ 342, 400, 57 },
		{ 342, 401, 57 },
		{ 342, 402, 57 },
		{ 342, 403, 57 },
		{ 342, 404, 57 },
		{ 342, 405, 57 },
		{ 342, 406, 57 },
		{ 342, 407, 57 },
		{ 342, 408, 57 },
		{ 342, 409, 57 },
		{ 342, 410, 57 },
		{ 342, 411, 57 },
		{ 342, 412, 57 },
		{ 342, 413, 57 },
		{ 342, 414, 57 },
		{ 342, 415, 57 },
		{ 342, 416, 57 },
		{ 342, 417, 57 },
		{ 342, 418, 57 },
		{ 342, 419, 57 },
		{ 342, 420, 57 },
		{ 342, 421, 57 },
		{ 342, 422, 57 },
		{ 342, 423, 57 },
		{ 342, 424, 57 },
		{ 342, 425, 57 },
		{ 342, 426, 57 },
		{ 343, 372, 57 },
		{ 343, 373, 57 },
		{ 343, 374, 57 },
		{ 343, 375, 57 },
		{ 343, 376, 57 },
		{ 343, 377, 57 },
		{ 343, 378, 57 },
		{ 343, 379, 57 },
		{ 343, 380, 57 },
		{ 343, 381, 57 },
		{ 343, 382, 57 },
		{ 343, 383, 57 },
		{ 343, 384, 57 },
		{ 343, 385, 57 },
		{ 343, 386, 57 },
		{ 343, 387, 57 },
		{ 343, 388, 57 },
		{ 343, 389, 57 },
		{ 343, 390, 57 },
		{ 343, 391, 57 },
		{ 343, 392, 57 },
		{ 343, 393, 57 },
		{ 343, 394, 57 },
		{ 343, 395, 57 },
		{ 343, 396, 57 },
		{ 343, 397, 57 },
		{ 343, 398, 57 },
		{ 343, 399, 57 },
		{ 343, 400, 57 },
		{ 343, 401, 57 },
		{ 343, 402, 57 },
		{ 343, 403, 57 },
		{ 343, 404, 57 },
		{ 343, 405, 57 },
		{ 343, 406, 57 },
		{ 343, 407, 57 },
		{ 343, 408, 57 },
		{ 343, 409, 57 },
		{ 343, 410, 57 },
		{ 343, 411, 57 },
		{ 343, 412, 57 },
		{ 343, 413, 57 },
		{ 343, 414, 57 },
		{ 343, 415, 57 },
		{ 343, 416, 57 },
		{ 343, 417, 57 },
		{ 343, 418, 57 },
		{ 343, 419, 57 },
		{ 343, 420, 57 },
		{ 343, 421, 57 },
		{ 343, 422, 57 },
		{ 343, 423, 57 },
		{ 343, 424, 57 },
		{ 343, 425, 57 },
		{ 343, 426, 57 },
		{ 343, 427, 57 },
		{ 343, 428, 57 },
		{ 344, 374, 57 },
		{ 344, 375, 57 },
		{ 344, 376, 57 },
		{ 344, 377, 57 },
		{ 344, 378, 57 },
		{ 344, 379, 57 },
		{ 344, 380, 57 },
		{ 344, 381, 57 },
		{ 344, 382, 57 },
		{ 344, 383, 57 },
		{ 344, 384, 57 },
		{ 344, 385, 57 },
		{ 344, 386, 57 },
		{ 344, 387, 57 },
		{ 344, 388, 57 },
		{ 344, 389, 57 },
		{ 344, 390, 57 },
		{ 344, 391, 57 },
		{ 344, 392, 57 },
		{ 344, 393, 57 },
		{ 344, 394, 57 },
		{ 344, 395, 57 },
		{ 344, 396, 57 },
		{ 344, 397, 57 },
		{ 344, 398, 57 },
		{ 344, 399, 57 },
		{ 344, 400, 57 },
		{ 344, 401, 57 },
		{ 344, 402, 57 },
		{ 344, 403, 57 },
		{ 344, 404, 57 },
		{ 344, 405, 57 },
		{ 344, 406, 57 },
		{ 344, 407, 57 },
		{ 344, 408, 57 },
		{ 344, 409, 57 },
		{ 344, 410, 57 },
		{ 344, 411, 57 },
		{ 344, 412, 57 },
		{ 344, 413, 57 },
		{ 344, 414, 57 },
		{ 344, 415, 57 },
		{ 344, 416, 57 },
		{ 344, 417, 57 },
		{ 344, 418, 57 },
		{ 344, 419, 57 },
		{ 344, 420, 57 },
		{ 344, 421, 57 },
		{ 344, 422, 57 },
		{ 344, 423, 57 },
		{ 344, 424, 57 },
		{ 344, 425, 57 },
		{ 344, 426, 57 },
		{ 344, 427, 57 },
		{ 344, 428, 57 },
		{ 344, 429, 57 },
		{ 344, 430, 57 },
		{ 345, 375, 57 },
		{ 345, 376, 57 },
		{ 345, 377, 57 },
		{ 345, 378, 57 },
		{ 345, 379, 57 },
		{ 345, 380, 57 },
		{ 345, 381, 57 },
		{ 345, 382, 57 },
		{ 345, 383, 57 },
		{ 345, 384, 57 },
		{ 345, 385, 57 },
		{ 345, 386, 57 },
		{ 345, 387, 57 },
		{ 345, 388, 57 },
		{ 345, 389, 57 },
		{ 345, 390, 57 },
		{ 345, 391, 57 },
		{ 345, 392, 57 },
		{ 345, 393, 57 },
		{ 345, 394, 57 },
		{ 345, 395, 57 },
		{ 345, 396, 57 },
		{ 345, 397, 57 },
		{ 345, 398, 57 },
		{ 345, 399, 57 },
		{ 345, 400, 57 },
		{ 345, 401, 57 },
		{ 345, 402, 57 },
		{ 345, 403, 57 },
		{ 345, 404, 57 },
		{ 345, 405, 57 },
		{ 345, 406, 57 },
		{ 345, 407, 57 },
		{ 345, 408, 57 },
		{ 345, 409, 57 },
		{ 345, 410, 57 },
		{ 345, 411, 57 },
		{ 345, 412, 57 },
		{ 345, 413, 57 },
		{ 345, 414, 57 },
		{ 345, 415, 57 },
		{ 345, 416, 57 },
		{ 345, 417, 57 },
		{ 345, 418, 57 },
		{ 345, 419, 57 },
		{ 345, 420, 57 },
		{ 345, 421, 57 },
		{ 345, 422, 57 },
		{ 345, 423, 57 },
		{ 345, 424, 57 },
		{ 345, 425, 57 },
		{ 345, 426, 57 },
		{ 345, 427, 57 },
		{ 345, 428, 57 },
		{ 345, 429, 57 },
		{ 345, 430, 57 },
		{ 345, 431, 57 },
		{ 345, 432, 57 },
		{ 346, 376, 57 },
		{ 346, 377, 57 },
		{ 346, 378, 57 },
		{ 346, 379, 57 },
		{ 346, 380, 57 },
		{ 346, 381, 57 },
		{ 346, 382, 57 },
		{ 346, 383, 57 },
		{ 346, 384, 57 },
		{ 346, 385, 57 },
		{ 346, 386, 57 },
		{ 346, 387, 57 },
		{ 346, 388, 57 },
		{ 346, 389, 57 },
		{ 346, 390, 57 },
		{ 346, 391, 57 },
		{ 346, 392, 57 },
		{ 346, 393, 57 },
		{ 346, 394, 57 },
		{ 346, 395, 57 },
		{ 346, 396, 57 },
		{ 346, 397, 57 },
		{ 346, 398, 57 },
		{ 346, 399, 57 },
		{ 346, 400, 57 },
		{ 346, 401, 57 },
		{ 346, 402, 57 },
		{ 346, 403, 57 },
		{ 346, 404, 57 },
		{ 346, 405, 57 },
		{ 346, 406, 57 },
		{ 346, 407, 57 },
		{ 346, 408, 57 },
		{ 346, 409, 57 },
		{ 346, 410, 57 },
		{ 346, 411, 57 },
		{ 346, 412, 57 },
		{ 346, 413, 57 },
		{ 346, 414, 57 },
		{ 346, 415, 57 },
		{ 346, 416, 57 },
		{ 346, 417, 57 },
		{ 346, 418, 57 },
		{ 346, 419, 57 },
		{ 346, 420, 57 },
		{ 346, 421, 57 },
		{ 346, 422, 57 },
		{ 346, 423, 57 },
		{ 346, 424, 57 },
		{ 346, 425, 57 },
		{ 346, 426, 57 },
		{ 346, 427, 57 },
		{ 346, 428, 57 },
		{ 346, 429, 57 },
		{ 346, 430, 57 },
		{ 346, 431, 57 },
		{ 346, 432, 57 },
		{ 346, 433, 57 },
		{ 346, 434, 57 },
		{ 347, 378, 57 },
		{ 347, 379, 57 },
		{ 347, 380, 57 },
		{ 347, 381, 57 },
		{ 347, 382, 57 },
		{ 347, 383, 57 },
		{ 347, 384, 57 },
		{ 347, 385, 57 },
		{ 347, 386, 57 },
		{ 347, 387, 57 },
		{ 347, 388, 57 },
		{ 347, 389, 57 },
		{ 347, 390, 57 },
		{ 347, 391, 57 },
		{ 347, 392, 57 },
		{ 347, 393, 57 },
		{ 347, 394, 57 },
		{ 347, 395, 57 },
		{ 347, 396, 57 },
		{ 347, 397, 57 },
		{ 347, 398, 57 },
		{ 347, 399, 57 },
		{ 347, 400, 57 },
		{ 347, 401, 57 },
		{ 347, 402, 57 },
		{ 347, 403, 57 },
		{ 347, 404, 57 },
		{ 347, 405, 57 },
		{ 347, 406, 57 },
		{ 347, 407, 57 },
		{ 347, 408, 57 },
		{ 347, 409, 57 },
		{ 347, 410, 57 },
		{ 347, 411, 57 },
		{ 347, 412, 57 },
		{ 347, 413, 57 },
		{ 347, 414, 57 },
		{ 347, 415, 57 },
		{ 347, 416, 57 },
		{ 347, 417, 57 },
		{ 347, 418, 57 },
		{ 347, 419, 57 },
		{ 347, 420, 57 },
		{ 347, 421, 57 },
		{ 347, 422, 57 },
		{ 347, 423, 57 },
		{ 347, 424, 57 },
		{ 347, 425, 57 },
		{ 347, 426, 57 },
		{ 347, 427, 57 },
		{ 347, 428, 57 },
		{ 347, 429, 57 },
		{ 347, 430, 57 },
		{ 347, 431, 57 },
		{ 347, 432, 57 },
		{ 347, 433, 57 },
		{ 347, 434, 57 },
		{ 347, 435, 57 },
		{ 347, 436, 57 },
		{ 348, 379, 57 },
		{ 348, 380, 57 },
		{ 348, 381, 57 },
		{ 348, 382, 57 },
		{ 348, 383, 57 },
		{ 348, 384, 57 },
		{ 348, 385, 57 },
		{ 348, 386, 57 },
		{ 348, 387, 57 },
		{ 348, 388, 57 },
		{ 348, 389, 57 },
		{ 348, 390, 57 },
		{ 348, 391, 57 },
		{ 348, 392, 57 },
		{ 348, 393, 57 },
		{ 348, 394, 57 },
		{ 348, 395, 57 },
		{ 348, 396, 57 },
		{ 348, 397, 57 },
		{ 348, 398, 57 },
		{ 348, 399, 57 },
		{ 348, 400, 57 },
		{ 348, 401, 57 },
		{ 348, 402, 57 },
		{ 348, 403, 57 },
		{ 348, 404, 57 },
		{ 348, 405, 57 },
		{ 348, 406, 57 },
		{ 348, 407, 57 },
		{ 348, 408, 57 },
		{ 348, 409, 57 },
		{ 348, 410, 57 },
		{ 348, 411, 57 },
		{ 348, 412, 57 },
		{ 348, 413, 57 },
		{ 348, 414, 57 },
		{ 348, 415, 57 },
		{ 348, 416, 57 },
		{ 348, 417, 57 },
		{ 348, 418, 57 },
		{ 348, 419, 57 },
		{ 348, 420, 57 },
		{ 348, 421, 57 },
		{ 348, 422, 57 },
		{ 348, 423, 57 },
		{ 348, 424, 57 },
		{ 348, 425, 57 },
		{ 348, 426, 57 },
		{ 348, 427, 57 },
		{ 348, 428, 57 },
		{ 348, 429, 57 },
		{ 348, 430, 57 },
		{ 348, 431, 57 },
		{ 348, 432, 57 },
		{ 348, 433, 57 },
		{ 348, 434, 57 },
		{ 348, 435, 57 },
		{ 348, 436, 57 },
		{ 348, 437, 57 },
		{ 348, 438, 57 },
		{ 349, 380, 57 },
		{ 349, 381, 57 },
		{ 349, 382, 57 },
		{ 349, 383, 57 },
		{ 349, 384, 57 },
		{ 349, 385, 57 },
		{ 349, 386, 57 },
		{ 349, 387, 57 },
		{ 349, 388, 57 },
		{ 349, 389, 57 },
		{ 349, 390, 57 },
		{ 349, 391, 57 },
		{ 349, 392, 57 },
		{ 349, 393, 57 },
		{ 349, 394, 57 },
		{ 349, 395, 57 },
		{ 349, 396, 57 },
		{ 349, 397, 57 },
		{ 349, 398, 57 },
		{ 349, 399, 57 },
		{ 349, 400, 57 },
		{ 349, 401, 57 },
		{ 349, 402, 57 },
		{ 349, 403, 57 },
		{ 349, 404, 57 },
		{ 349, 405, 57 },
		{ 349, 406, 57 },
		{ 349, 407, 57 },
		{ 349, 408, 57 },
		{ 349, 409, 57 },
		{ 349, 410, 57 },
		{ 349, 411, 57 },
		{ 349, 412, 57 },
		{ 349, 413, 57 },
		{ 349, 414, 57 },
		{ 349, 415, 57 },
		{ 349, 416, 57 },
		{ 349, 417, 57 },
		{ 349, 418, 57 },
		{ 349, 419, 57 },
		{ 349, 420, 57 },
		{ 349, 421, 57 },
		{ 349, 422, 57 },
		{ 349, 423, 57 },
		{ 349, 424, 57 },
		{ 349, 425, 57 },
		{ 349, 426, 57 },
		{ 349, 427, 57 },
		{ 349, 428, 57 },
		{ 349, 429, 57 },
		{ 349, 430, 57 },
		{ 349, 431, 57 },
		{ 349, 432, 57 },
		{ 349, 433, 57 },
		{ 349, 434, 57 },
		{ 349, 435, 57 },
		{ 349, 436, 57 },
		{ 349, 437, 57 },
		{ 349, 438, 57 },
		{ 349, 439, 57 },
		{ 349, 440, 57 },
		{ 350, 382, 57 },
		{ 350, 383, 57 },
		{ 350, 384, 57 },
		{ 350, 385, 57 },
		{ 350, 386, 57 },
		{ 350, 387, 57 },
		{ 350, 388, 57 },
		{ 350, 389, 57 },
		{ 350, 390, 57 },
		{ 350, 391, 57 },
		{ 350, 392, 57 },
		{ 350, 393, 57 },
		{ 350, 394, 57 },
		{ 350, 395, 57 },
		{ 350, 396, 57 },
		{ 350, 397, 57 },
		{ 350, 398, 57 },
		{ 350, 399, 57 },
		{ 350, 400, 57 },
		{ 350, 401, 57 },
		{ 350, 402, 57 },
		{ 350, 403, 57 },
		{ 350, 404, 57 },
		{ 350, 405, 57 },
		{ 350, 406, 57 },
		{ 350, 407, 57 },
		{ 350, 408, 57 },
		{ 350, 409, 57 },
		{ 350, 410, 57 },
		{ 350, 411, 57 },
		{ 350, 412, 57 },
		{ 350, 413, 57 },
		{ 350, 414, 57 },
		{ 350, 415, 57 },
		{ 350, 416, 57 },
		{ 350, 417, 57 },
		{ 350, 418, 57 },
		{ 350, 419, 57 },
		{ 350, 420, 57 },
		{ 350, 421, 57 },
		{ 350, 422, 57 },
		{ 350, 423, 57 },
		{ 350, 424, 57 },
		{ 350, 425, 57 },
		{ 350, 426, 57 },
		{ 350, 427, 57 },
		{ 350, 428, 57 },
		{ 350, 429, 57 },
		{ 350, 430, 57 },
		{ 350, 431, 57 },
		{ 350, 432, 57 },
		{ 350, 433, 57 },
		{ 350, 434, 57 },
		{ 350, 435, 57 },
		{ 350, 436, 57 },
		{ 350, 437, 57 },
		{ 350, 438, 57 },
		{ 350, 439, 57 },
		{ 350, 440, 57 },
		{ 350, 441, 57 },
		{ 350, 442, 57 },
		{ 351, 383, 57 },
		{ 351, 384, 57 },
		{ 351, 385, 57 },
		{ 351, 386, 57 },
		{ 351, 387, 57 },
		{ 351, 388, 57 },
		{ 351, 389, 57 },
		{ 351, 390, 57 },
		{ 351, 391, 57 },
		{ 351, 392, 57 },
		{ 351, 393, 57 },
		{ 351, 394, 57 },
		{ 351, 395, 57 },
		{ 351, 396, 57 },
		{ 351, 397, 57 },
		{ 351, 398, 57 },
		{ 351, 399, 57 },
		{ 351, 400, 57 },
		{ 351, 401, 57 },
		{ 351, 402, 57 },
		{ 351, 403, 57 },
		{ 351, 404, 57 },
		{ 351, 405, 57 },
		{ 351, 406, 57 },
		{ 351, 407, 57 },
		{ 351, 408, 57 },
		{ 351, 409, 57 },
		{ 351, 410, 57 },
		{ 351, 411, 57 },
		{ 351, 412, 57 },
		{ 351, 413, 57 },
		{ 351, 414, 57 },
		{ 351, 415, 57 },
		{ 351, 416, 57 },
		{ 351, 417, 57 },
		{ 351, 418, 57 },
		{ 351, 419, 57 },
		{ 351, 420, 57 },
		{ 351, 421, 57 },
		{ 351, 422, 57 },
		{ 351, 423, 57 },
		{ 351, 424, 57 },
		{ 351, 425, 57 },
		{ 351, 426, 57 },
		{ 351, 427, 57 },
		{ 351, 428, 57 },
		{ 351, 429, 57 },
		{ 351, 430, 57 },
		{ 351, 431, 57 },
		{ 351, 432, 57 },
		{ 351, 433, 57 },
		{ 351, 434, 57 },
		{ 351, 435, 57 },
		{ 351, 436, 57 },
		{ 351, 437, 57 },
		{ 351, 438, 57 },
		{ 351, 439, 57 },
		{ 351, 440, 57 },
		{ 351, 441, 57 },
		{ 351, 442, 57 },
		{ 351, 443, 57 },
		{ 351, 444, 57 },
		{ 352, 384, 57 },
		{ 352, 385, 57 },
		{ 352, 386, 57 },
		{ 352, 387, 57 },
		{ 352, 388, 57 },
		{ 352, 389, 57 },
		{ 352, 390, 57 },
		{ 352, 391, 57 },
		{ 352, 392, 57 },
		{ 352, 393, 57 },
		{ 352, 394, 57 },
		{ 352, 395, 57 },
		{ 352, 396, 57 },
		{ 352, 397, 57 },
		{ 352, 398, 57 },
		{ 352, 399, 57 },
		{ 352, 400, 57 },
		{ 352, 401, 57 },
		{ 352, 402, 57 },
		{ 352, 403, 57 },
		{ 352, 404, 57 },
		{ 352, 405, 57 },
		{ 352, 406, 57 },
		{ 352, 407, 57 },
		{ 352, 408, 57 },
		{ 352, 409, 57 },
		{ 352, 410, 57 },
		{ 352, 411, 57 },
		{ 352, 412, 57 },
		{ 352, 413, 57 },
		{ 352, 414, 57 },
		{ 352, 415, 57 },
		{ 352, 416, 57 },
		{ 352, 417, 57 },
		{ 352, 418, 57 },
		{ 352, 419, 57 },
		{ 352, 420, 57 },
		{ 352, 421, 57 },
		{ 352, 422, 57 },
		{ 352, 423, 57 },
		{ 352, 424, 57 },
		{ 352, 425, 57 },
		{ 352, 426, 57 },
		{ 352, 427, 57 },
		{ 352, 428, 57 },
		{ 352, 429, 57 },
		{ 352, 430, 57 },
		{ 352, 431, 57 },
		{ 352, 432, 57 },
		{ 352, 433, 57 },
		{ 352, 434, 57 },
		{ 352, 435, 57 },
		{ 352, 436, 57 },
		{ 352, 437, 57 },
		{ 352, 438, 57 },
		{ 352, 439, 57 },
		{ 352, 440, 57 },
		{ 352, 441, 57 },
		{ 352, 442, 57 },
		{ 352, 443, 57 },
		{ 352, 444, 57 },
		{ 352, 445, 57 },
		{ 352, 446, 57 },
		{ 353, 386, 57 },
		{ 353, 387, 57 },
		{ 353, 388, 57 },
		{ 353, 389, 57 },
		{ 353, 390, 57 },
		{ 353, 391, 57 },
		{ 353, 392, 57 },
		{ 353, 393, 57 },
		{ 353, 394, 57 },
		{ 353, 395, 57 },
		{ 353, 396, 57 },
		{ 353, 397, 57 },
		{ 353, 398, 57 },
		{ 353, 399, 57 },
		{ 353, 400, 57 },
		{ 353, 401, 57 },
		{ 353, 402, 57 },
		{ 353, 403, 57 },
		{ 353, 404, 57 },
		{ 353, 405, 57 },
		{ 353, 406, 57 },
		{ 353, 407, 57 },
		{ 353, 408, 57 },
		{ 353, 409, 57 },
		{ 353, 410, 57 },
		{ 353, 411, 57 },
		{ 353, 412, 57 },
		{ 353, 413, 57 },
		{ 353, 414, 57 },
		{ 353, 415, 57 },
		{ 353, 416, 57 },
		{ 353, 417, 57 },
		{ 353, 418, 57 },
		{ 353, 419, 57 },
		{ 353, 420, 57 },
		{ 353, 421, 57 },
		{ 353, 422, 57 },
		{ 353, 423, 57 },
		{ 353, 424, 57 },
		{ 353, 425, 57 },
		{ 353, 426, 57 },
		{ 353, 427, 57 },
		{ 353, 428, 57 },
		{ 353, 429, 57 },
		{ 353, 430, 57 },
		{ 353, 431, 57 },
		{ 353, 432, 57 },
		{ 353, 433, 57 },
		{ 353, 434, 57 },
		{ 353, 435, 57 },
		{ 353, 436, 57 },
		{ 353, 437, 57 },
		{ 353, 438, 57 },
		{ 353, 439, 57 },
		{ 353, 440, 57 },
		{ 353, 441, 57 },
		{ 353, 442, 57 },
		{ 353, 443, 57 },
		{ 353, 444, 57 },
		{ 353, 445, 57 },
		{ 353, 446, 57 },
		{ 353, 447, 57 },
		{ 353, 448, 57 },
		{ 354, 387, 57 },
		{ 354, 388, 57 },
		{ 354, 389, 57 },
		{ 354, 390, 57 },
		{ 354, 391, 57 },
		{ 354, 392, 57 },
		{ 354, 393, 57 },
		{ 354, 394, 57 },
		{ 354, 395, 57 },
		{ 354, 396, 57 },
		{ 354, 397, 57 },
		{ 354, 398, 57 },
		{ 354, 399, 57 },
		{ 354, 400, 57 },
		{ 354, 401, 57 },
		{ 354, 402, 57 },
		{ 354, 403, 57 },
		{ 354, 404, 57 },
		{ 354, 405, 57 },
		{ 354, 406, 57 },
		{ 354, 407, 57 },
		{ 354, 408, 57 },
		{ 354, 409, 57 },
		{ 354, 410, 57 },
		{ 354, 411, 57 },
		{ 354, 412, 57 },
		{ 354, 413, 57 },
		{ 354, 414, 57 },
		{ 354, 415, 57 },
		{ 354, 416, 57 },
		{ 354, 417, 57 },
		{ 354, 418, 57 },
		{ 354, 419, 57 },
		{ 354, 420, 57 },
		{ 354, 421, 57 },
		{ 354, 422, 57 },
		{ 354, 423, 57 },
		{ 354, 424, 57 },
		{ 354, 425, 57 },
		{ 354, 426, 57 },
		{ 354, 427, 57 },
		{ 354, 428, 57 },
		{ 354, 429, 57 },
		{ 354, 430, 57 },
		{ 354, 431, 57 },
		{ 354, 432, 57 },
		{ 354, 433, 57 },
		{ 354, 434, 57 },
		{ 354, 435, 57 },
		{ 354, 436, 57 },
		{ 354, 437, 57 },
		{ 354, 438, 57 },
		{ 354, 439, 57 },
		{ 354, 440, 57 },
		{ 354, 441, 57 },
		{ 354, 442, 57 },
		{ 354, 443, 57 },
		{ 354, 444, 57 },
		{ 354, 445, 57 },
		{ 354, 446, 57 },
		{ 354, 447, 57 },
		{ 354, 448, 57 },
		{ 354, 449, 57 },
		{ 354, 450, 57 },
		{ 355, 388, 57 },
		{ 355, 389, 57 },
		{ 355, 390, 57 },
		{ 355, 391, 57 },
		{ 355, 392, 57 },
		{ 355, 393, 57 },
		{ 355, 394, 57 },
		{ 355, 395, 57 },
		{ 355, 396, 57 },
		{ 355, 397, 57 },
		{ 355, 398, 57 },
		{ 355, 399, 57 },
		{ 355, 400, 57 },
		{ 355, 401, 57 },
		{ 355, 402, 57 },
		{ 355, 403, 57 },
		{ 355, 404, 57 },
		{ 355, 405, 57 },
		{ 355, 406, 57 },
		{ 355, 407, 57 },
		{ 355, 408, 57 },
		{ 355, 409, 57 },
		{ 355, 410, 57 },
		{ 355, 411, 57 },
		{ 355, 412, 57 },
		{ 355, 413, 57 },
		{ 355, 414, 57 },
		{ 355, 415, 57 },
		{ 355, 416, 57 },
		{ 355, 417, 57 },
		{ 355, 418, 57 },
		{ 355, 419, 57 },
		{ 355, 420, 57 },
		{ 355, 421, 57 },
		{ 355, 422, 57 },
		{ 355, 423, 57 },
		{ 355, 424, 57 },
		{ 355, 425, 57 },
		{ 355, 426, 57 },
		{ 355, 427, 57 },
		{ 355, 428, 57 },
		{ 355, 429, 57 },
		{ 355, 430, 57 },
		{ 355, 431, 57 },
		{ 355, 432, 57 },
		{ 355, 433, 57 },
		{ 355, 434, 57 },
		{ 355, 435, 57 },
		{ 355, 436, 57 },
		{ 355, 437, 57 },
		{ 355, 438, 57 },
		{ 355, 439, 57 },
		{ 355, 440, 57 },
		{ 355, 441, 57 },
		{ 355, 442, 57 },
		{ 355, 443, 57 },
		{ 355, 444, 57 },
		{ 355, 445, 57 },
		{ 355, 446, 57 },
		{ 355, 447, 57 },
		{ 355, 448, 57 },
		{ 355, 449, 57 },
		{ 355, 450, 57 },
		{ 355, 451, 57 },
		{ 355, 452, 57 },
		{ 356, 390, 57 },
		{ 356, 391, 57 },
		{ 356, 392, 57 },
		{ 356, 393, 57 },
		{ 356, 394, 57 },
		{ 356, 395, 57 },
		{ 356, 396, 57 },
		{ 356, 397, 57 },
		{ 356, 398, 57 },
		{ 356, 399, 57 },
		{ 356, 400, 57 },
		{ 356, 401, 57 },
		{ 356, 402, 57 },
		{ 356, 403, 57 },
		{ 356, 404, 57 },
		{ 356, 405, 57 },
		{ 356, 406, 57 },
		{ 356, 407, 57 },
		{ 356, 408, 57 },
		{ 356, 409, 57 },
		{ 356, 410, 57 },
		{ 356, 411, 57 },
		{ 356, 412, 57 },
		{ 356, 413, 57 },
		{ 356, 414, 57 },
		{ 356, 415, 57 },
		{ 356, 416, 57 },
		{ 356, 417, 57 },
		{ 356, 418, 57 },
		{ 356, 419, 57 },
		{ 356, 420, 57 },
		{ 356, 421, 57 },
		{ 356, 422, 57 },
		{ 356, 423, 57 },
		{ 356, 424, 57 },
		{ 356, 425, 57 },
		{ 356, 426, 57 },
		{ 356, 427, 57 },
		{ 356, 428, 57 },
		{ 356, 429, 57 },
		{ 356, 430, 57 },
		{ 356, 431, 57 },
		{ 356, 432, 57 },
		{ 356, 433, 57 },
		{ 356, 434, 57 },
		{ 356, 435, 57 },
		{ 356, 436, 57 },
		{ 356, 437, 57 },
		{ 356, 438, 57 },
		{ 356, 439, 57 },
		{ 356, 440, 57 },
		{ 356, 441, 57 },
		{ 356, 442, 57 },
		{ 356, 443, 57 },
		{ 356, 444, 57 },
		{ 356, 445, 57 },
		{ 356, 446, 57 },
		{ 356, 447, 57 },
		{ 356, 448, 57 },
		{ 356, 449, 57 },
		{ 356, 450, 57 },
		{ 356, 451, 57 },
		{ 356, 452, 57 },
		{ 356, 453, 57 },
		{ 356, 454, 57 },
		{ 357, 391, 57 },
		{ 357, 392, 57 },
		{ 357, 393, 57 },
		{ 357, 394, 57 },
		{ 357, 395, 57 },
		{ 357, 396, 57 },
		{ 357, 397, 57 },
		{ 357, 398, 57 },
		{ 357, 399, 57 },
		{ 357, 400, 57 },
		{ 357, 401, 57 },
		{ 357, 402, 57 },
		{ 357, 403, 57 },
		{ 357, 404, 57 },
		{ 357, 405, 57 },
		{ 357, 406, 57 },
		{ 357, 407, 57 },
		{ 357, 408, 57 },
		{ 357, 409, 57 },
		{ 357, 410, 57 },
		{ 357, 411, 57 },
		{ 357, 412, 57 },
		{ 357, 413, 57 },
		{ 357, 414, 57 },
		{ 357, 415, 57 },
		{ 357, 416, 57 },
		{ 357, 417, 57 },
		{ 357, 418, 57 },
		{ 357, 419, 57 },
		{ 357, 420, 57 },
		{ 357, 421, 57 },
		{ 357, 422, 57 },
		{ 357, 423, 57 },
		{ 357, 424, 57 },
		{ 357, 425, 57 },
		{ 357, 426, 57 },
		{ 357, 427, 57 },
		{ 357, 428, 57 },
		{ 357, 429, 57 },
		{ 357, 430, 57 },
		{ 357, 431, 57 },
		{ 357, 432, 57 },
		{ 357, 433, 57 },
		{ 357, 434, 57 },
		{ 357, 435, 57 },
		{ 357, 436, 57 },
		{ 357, 437, 57 },
		{ 357, 438, 57 },
		{ 357, 439, 57 },
		{ 357, 440, 57 },
		{ 357, 441, 57 },
		{ 357, 442, 57 },
		{ 357, 443, 57 },
		{ 357, 444, 57 },
		{ 357, 445, 57 },
		{ 357, 446, 57 },
		{ 357, 447, 57 },
		{ 357, 448, 57 },
		{ 357, 449, 57 },
		{ 357, 450, 57 },
		{ 357, 451, 57 },
		{ 357, 452, 57 },
		{ 357, 453, 57 },
		{ 357, 454, 57 },
		{ 357, 455, 57 },
		{ 357, 456, 57 },
		{ 358, 392, 57 },
		{ 358, 393, 57 },
		{ 358, 394, 57 },
		{ 358, 395, 57 },
		{ 358, 396, 57 },
		{ 358, 397, 57 },
		{ 358, 398, 57 },
		{ 358, 399, 57 },
		{ 358, 400, 57 },
		{ 358, 401, 57 },
		{ 358, 402, 57 },
		{ 358, 403, 57 },
		{ 358, 404, 57 },
		{ 358, 405, 57 },
		{ 358, 406, 57 },
		{ 358, 407, 57 },
		{ 358, 408, 57 },
		{ 358, 409, 57 },
		{ 358, 410, 57 },
		{ 358, 411, 57 },
		{ 358, 412, 57 },
		{ 358, 413, 57 },
		{ 358, 414, 57 },
		{ 358, 415, 57 },
		{ 358, 416, 57 },
		{ 358, 417, 57 },
		{ 358, 418, 57 },
		{ 358, 419, 57 },
		{ 358, 420, 57 },
		{ 358, 421, 57 },
		{ 358, 422, 57 },
		{ 358, 423, 57 },
		{ 358, 424, 57 },
		{ 358, 425, 57 },
		{ 358, 426, 57 },
		{ 358, 427, 57 },
		{ 358, 428, 57 },
		{ 358, 429, 57 },
		{ 358, 430, 57 },
		{ 358, 431, 57 },
		{ 358, 432, 57 },
		{ 358, 433, 57 },
		{ 358, 434, 57 },
		{ 358, 435, 57 },
		{ 358, 436, 57 },
		{ 358, 437, 57 },
		{ 358, 438, 57 },
		{ 358, 439, 57 },
		{ 358, 440, 57 },
		{ 358, 441, 57 },
		{ 358, 442, 57 },
		{ 358, 443, 57 },
		{ 358, 444, 57 },
		{ 358, 445, 57 },
		{ 358, 446, 57 },
		{ 358, 447, 57 },
		{ 358, 448, 57 },
		{ 358, 449, 57 },
		{ 358, 450, 57 },
		{ 358, 451, 57 },
		{ 358, 452, 57 },
		{ 358, 453, 57 },
		{ 358, 454, 57 },
		{ 358, 455, 57 },
		{ 358, 456, 57 },
		{ 358, 457, 57 },
		{ 358, 458, 57 },
		{ 359, 394, 57 },
		{ 359, 395, 57 },
		{ 359, 396, 57 },
		{ 359, 397, 57 },
		{ 359, 398, 57 },
		{ 359, 399, 57 },
		{ 359, 400, 57 },
		{ 359, 401, 57 },
		{ 359, 402, 57 },
		{ 359, 403, 57 },
		{ 359, 404, 57 },
		{ 359, 405, 57 },
		{ 359, 406, 57 },
		{ 359, 407, 57 },
		{ 359, 408, 57 },
		{ 359, 409, 57 },
		{ 359, 410, 57 },
		{ 359, 411, 57 },
		{ 359, 412, 57 },
		{ 359, 413, 57 },
		{ 359, 414, 57 },
		{ 359, 415, 57 },
		{ 359, 416, 57 },
		{ 359, 417, 57 },
		{ 359, 418, 57 },
		{ 359, 419, 57 },
		{ 359, 420, 57 },
		{ 359, 421, 57 },
		{ 359, 422, 57 },
		{ 359, 423, 57 },
		{ 359, 424, 57 },
		{ 359, 425, 57 },
		{ 359, 426, 57 },
		{ 359, 427, 57 },
		{ 359, 428, 57 },
		{ 359, 429, 57 },
		{ 359, 430, 57 },
		{ 359, 431, 57 },
		{ 359, 432, 57 },
		{ 359, 433, 57 },
		{ 359, 434, 57 },
		{ 359, 435, 57 },
		{ 359, 436, 57 },
		{ 359, 437, 57 },
		{ 359, 438, 57 },
		{ 359, 439, 57 },
		{ 359, 440, 57 },
		{ 359, 441, 57 },
		{ 359, 442, 57 },
		{ 359, 443, 57 },
		{ 359, 444, 57 },
		{ 359, 445, 57 },
		{ 359, 446, 57 },
		{ 359, 447, 57 },
		{ 359, 448, 57 },
		{ 359, 449, 57 },
		{ 359, 450, 57 },
		{ 359, 451, 57 },
		{ 359, 452, 57 },
		{ 359, 453, 57 },
		{ 359, 454, 57 },
		{ 359, 455, 57 },
		{ 359, 456, 57 },
		{ 359, 457, 57 },
		{ 359, 458, 57 },
		{ 359, 459, 57 },
		{ 359, 460, 57 },
		{ 360, 395, 57 },
		{ 360, 396, 57 },
		{ 360, 397, 57 },
		{ 360, 398, 57 },
		{ 360, 399, 57 },
		{ 360, 400, 57 },
		{ 360, 401, 57 },
		{ 360, 402, 57 },
		{ 360, 403, 57 },
		{ 360, 404, 57 },
		{ 360, 405, 57 },
		{ 360, 406, 57 },
		{ 360, 407, 57 },
		{ 360, 408, 57 },
		{ 360, 409, 57 },
		{ 360, 410, 57 },
		{ 360, 411, 57 },
		{ 360, 412, 57 },
		{ 360, 413, 57 },
		{ 360, 414, 57 },
		{ 360, 415, 57 },
		{ 360, 416, 57 },
		{ 360, 417, 57 },
		{ 360, 418, 57 },
		{ 360, 419, 57 },
		{ 360, 420, 57 },
		{ 360, 421, 57 },
		{ 360, 422, 57 },
		{ 360, 423, 57 },
		{ 360, 424, 57 },
		{ 360, 425, 57 },
		{ 360, 426, 57 },
		{ 360, 427, 57 },
		{ 360, 428, 57 },
		{ 360, 429, 57 },
		{ 360, 430, 57 },
		{ 360, 431, 57 },
		{ 360, 432, 57 },
		{ 360, 433, 57 },
		{ 360, 434, 57 },
		{ 360, 435, 57 },
		{ 360, 436, 57 },
		{ 360, 437, 57 },
		{ 360, 438, 57 },
		{ 360, 439, 57 },
		{ 360, 440, 57 },
		{ 360, 441, 57 },
		{ 360, 442, 57 },
		{ 360, 443, 57 },
		{ 360, 444, 57 },
		{ 360, 445, 57 },
		{ 360, 446, 57 },
		{ 360, 447, 57 },
		{ 360, 448, 57 },
		{ 360, 449, 57 },
		{ 360, 450, 57 },
		{ 360, 451, 57 },
		{ 360, 452, 57 },
		{ 360, 453, 57 },
		{ 360, 454, 57 },
		{ 360, 455, 57 },
		{ 360, 456, 57 },
		{ 360, 457, 57 },
		{ 360, 458, 57 },
		{ 360, 459, 57 },
		{ 360, 460, 57 },
		{ 360, 461, 57 },
		{ 360, 462, 57 },
		{ 361, 396, 57 },
		{ 361, 397, 57 },
		{ 361, 398, 57 },
		{ 361, 399, 57 },
		{ 361, 400, 57 },
		{ 361, 401, 57 },
		{ 361, 402, 57 },
		{ 361, 403, 57 },
		{ 361, 404, 57 },
		{ 361, 405, 57 },
		{ 361, 406, 57 },
		{ 361, 407, 57 },
		{ 361, 408, 57 },
		{ 361, 409, 57 },
		{ 361, 410, 57 },
		{ 361, 411, 57 },
		{ 361, 412, 57 },
		{ 361, 413, 57 },
		{ 361, 414, 57 },
		{ 361, 415, 57 },
		{ 361, 416, 57 },
		{ 361, 417, 57 },
		{ 361, 418, 57 },
		{ 361, 419, 57 },
		{ 361, 420, 57 },
		{ 361, 421, 57 },
		{ 361, 422, 57 },
		{ 361, 423, 57 },
		{ 361, 424, 57 },
		{ 361, 425, 57 },
		{ 361, 426, 57 },
		{ 361, 427, 57 },
		{ 361, 428, 57 },
		{ 361, 429, 57 },
		{ 361, 430, 57 },
		{ 361, 431, 57 },
		{ 361, 432, 57 },
		{ 361, 433, 57 },
		{ 361, 434, 57 },
		{ 361, 435, 57 },
		{ 361, 436, 57 },
		{ 361, 437, 57 },
		{ 361, 438, 57 },
		{ 361, 439, 57 },
		{ 361, 440, 57 },
		{ 361, 441, 57 },
		{ 361, 442, 57 },
		{ 361, 443, 57 },
		{ 361, 444, 57 },
		{ 361, 445, 57 },
		{ 361, 446, 57 },
		{ 361, 447, 57 },
		{ 361, 448, 57 },
		{ 361, 449, 57 },
		{ 361, 450, 57 },
		{ 361, 451, 57 },
		{ 361, 452, 57 },
		{ 361, 453, 57 },
		{ 361, 454, 57 },
		{ 361, 455, 57 },
		{ 361, 456, 57 },
		{ 361, 457, 57 },
		{ 361, 458, 57 },
		{ 361, 459, 57 },
		{ 361, 460, 57 },
		{ 361, 461, 57 },
		{ 361, 462, 57 },
		{ 361, 463, 57 },
		{ 361, 464, 57 },
		{ 362, 398, 57 },
		{ 362, 399, 57 },
		{ 362, 400, 57 },
		{ 362, 401, 57 },
		{ 362, 402, 57 },
		{ 362, 403, 57 },
		{ 362, 404, 57 },
		{ 362, 405, 57 },
		{ 362, 406, 57 },
		{ 362, 407, 57 },
		{ 362, 408, 57 },
		{ 362, 409, 57 },
		{ 362, 410, 57 },
		{ 362, 411, 57 },
		{ 362, 412, 57 },
		{ 362, 413, 57 },
		{ 362, 414, 57 },
		{ 362, 415, 57 },
		{ 362, 416, 57 },
		{ 362, 417, 57 },
		{ 362, 418, 57 },
		{ 362, 419, 57 },
		{ 362, 420, 57 },
		{ 362, 421, 57 },
		{ 362, 422, 57 },
		{ 362, 423, 57 },
		{ 362, 424, 57 },
		{ 362, 425, 57 },
		{ 362, 426, 57 },
		{ 362, 427, 57 },
		{ 362, 428, 57 },
		{ 362, 429, 57 },
		{ 362, 430, 57 },
		{ 362, 431, 57 },
		{ 362, 432, 57 },
		{ 362, 433, 57 },
		{ 362, 434, 57 },
		{ 362, 435, 57 },
		{ 362, 436, 57 },
		{ 362, 437, 57 },
		{ 362, 438, 57 },
		{ 362, 439, 57 },
		{ 362, 440, 57 },
		{ 362, 441, 57 },
		{ 362, 442, 57 },
		{ 362, 443, 57 },
		{ 362, 444, 57 },
		{ 362, 445, 57 },
		{ 362, 446, 57 },
		{ 362, 447, 57 },
		{ 362, 448, 57 },
		{ 362, 449, 57 },
		{ 362, 450, 57 },
		{ 362, 451, 57 },
		{ 362, 452, 57 },
		{ 362, 453, 57 },
		{ 362, 454, 57 },
		{ 362, 455, 57 },
		{ 362, 456, 57 },
		{ 362, 457, 57 },
		{ 362, 458, 57 },
		{ 362, 459, 57 },
		{ 362, 460, 57 },
		{ 362, 461, 57 },
		{ 362, 462, 57 },
		{ 362, 463, 57 },
		{ 362, 464, 57 },
		{ 362, 465, 57 },
		{ 362, 466, 57 },
		{ 363, 399, 57 },
		{ 363, 400, 57 },
		{ 363, 401, 57 },
		{ 363, 402, 57 },
		{ 363, 403, 57 },
		{ 363, 404, 57 },
		{ 363, 405, 57 },
		{ 363, 406, 57 },
		{ 363, 407, 57 },
		{ 363, 408, 57 },
		{ 363, 409, 57 },
		{ 363, 410, 57 },
		{ 363, 411, 57 },
		{ 363, 412, 57 },
		{ 363, 413, 57 },
		{ 363, 414, 57 },
		{ 363, 415, 57 },
		{ 363, 416, 57 },
		{ 363, 417, 57 },
		{ 363, 418, 57 },
		{ 363, 419, 57 },
		{ 363, 420, 57 },
		{ 363, 421, 57 },
		{ 363, 422, 57 },
		{ 363, 423, 57 },
		{ 363, 424, 57 },
		{ 363, 425, 57 },
		{ 363, 426, 57 },
		{ 363, 427, 57 },
		{ 363, 428, 57 },
		{ 363, 429, 57 },
		{ 363, 430, 57 },
		{ 363, 431, 57 },
		{ 363, 432, 57 },
		{ 363, 433, 57 },
		{ 363, 434, 57 },
		{ 363, 435, 57 },
		{ 363, 436, 57 },
		{ 363, 437, 57 },
		{ 363, 438, 57 },
		{ 363, 439, 57 },
		{ 363, 440, 57 },
		{ 363, 441, 57 },
		{ 363, 442, 57 },
		{ 363, 443, 57 },
		{ 363, 444, 57 },
		{ 363, 445, 57 },
		{ 363, 446, 57 },
		{ 363, 447, 57 },
		{ 363, 448, 57 },
		{ 363, 449, 57 },
		{ 363, 450, 57 },
		{ 363, 451, 57 },
		{ 363, 452, 57 },
		{ 363, 453, 57 },
		{ 363, 454, 57 },
		{ 363, 455, 57 },
		{ 363, 456, 57 },
		{ 363, 457, 57 },
		{ 363, 458, 57 },
		{ 363, 459, 57 },
		{ 363, 460, 57 },
		{ 363, 461, 57 },
		{ 363, 462, 57 },
		{ 363, 463, 57 },
		{ 363, 464, 57 },
		{ 363, 465, 57 },
		{ 363, 466, 57 },
		{ 363, 467, 57 },
		{ 363, 468, 57 },
		{ 364, 400, 57 },
		{ 364, 401, 57 },
		{ 364, 402, 57 },
		{ 364, 403, 57 },
		{ 364, 404, 57 },
		{ 364, 405, 57 },
		{ 364, 406, 57 },
		{ 364, 407, 57 },
		{ 364, 408, 57 },
		{ 364, 409, 57 },
		{ 364, 410, 57 },
		{ 364, 411, 57 },
		{ 364, 412, 57 },
		{ 364, 413, 57 },
		{ 364, 414, 57 },
		{ 364, 415, 57 },
		{ 364, 416, 57 },
		{ 364, 417, 57 },
		{ 364, 418, 57 },
		{ 364, 419, 57 },
		{ 364, 420, 57 },
		{ 364, 421, 57 },
		{ 364, 422, 57 },
		{ 364, 423, 57 },
		{ 364, 424, 57 },
		{ 364, 425, 57 },
		{ 364, 426, 57 },
		{ 364, 427, 57 },
		{ 364, 428, 57 },
		{ 364, 429, 57 },
		{ 364, 430, 57 },
		{ 364, 431, 57 },
		{ 364, 432, 57 },
		{ 364, 433, 57 },
		{ 364, 434, 57 },
		{ 364, 435, 57 },
		{ 364, 436, 57 },
		{ 364, 437, 57 },
		{ 364, 438, 57 },
		{ 364, 439, 57 },
		{ 364, 440, 57 },
		{ 364, 441, 57 },
		{ 364, 442, 57 },
		{ 364, 443, 57 },
		{ 364, 444, 57 },
		{ 364, 445, 57 },
		{ 364, 446, 57 },
		{ 364, 447, 57 },
		{ 364, 448, 57 },
		{ 364, 449, 57 },
		{ 364, 450, 57 },
		{ 364, 451, 57 },
		{ 364, 452, 57 },
		{ 364, 453, 57 },
		{ 364, 454, 57 },
		{ 364, 455, 57 },
		{ 364, 456, 57 },
		{ 364, 457, 57 },
		{ 364, 458, 57 },
		{ 364, 459, 57 },
		{ 364, 460, 57 },
		{ 364, 461, 57 },
		{ 364, 462, 57 },
		{ 364, 463, 57 },
		{ 364, 464, 57 },
		{ 364, 465, 57 },
		{ 364, 466, 57 },
		{ 364, 467, 57 },
		{ 364, 468, 57 },
		{ 364, 469, 57 },
		{ 364, 470, 57 },
		{ 365, 402, 57 },
		{ 365, 403, 57 },
		{ 365, 404, 57 },
		{ 365, 405, 57 },
		{ 365, 406, 57 },
		{ 365, 407, 57 },
		{ 365, 408, 57 },
		{ 365, 409, 57 },
		{ 365, 410, 57 },
		{ 365, 411, 57 },
		{ 365, 412, 57 },
		{ 365, 413, 57 },
		{ 365, 414, 57 },
		{ 365, 415, 57 },
		{ 365, 416, 57 },
		{ 365, 417, 57 },
		{ 365, 418, 57 },
		{ 365, 419, 57 },
		{ 365, 420, 57 },
		{ 365, 421, 57 },
		{ 365, 422, 57 },
		{ 365, 423, 57 },
		{ 365, 424, 57 },
		{ 365, 425, 57 },
		{ 365, 426, 57 },
		{ 365, 427, 57 },
		{ 365, 428, 57 },
		{ 365, 429, 57 },
		{ 365, 430, 57 },
		{ 365, 431, 57 },
		{ 365, 432, 57 },
		{ 365, 433, 57 },
		{ 365, 434, 57 },
		{ 365, 435, 57 },
		{ 365, 436, 57 },
		{ 365, 437, 57 },
		{ 365, 438, 57 },
		{ 365, 439, 57 },
		{ 365, 440, 57 },
		{ 365, 441, 57 },
		{ 365, 442, 57 },
		{ 365, 443, 57 },
		{ 365, 444, 57 },
		{ 365, 445, 57 },
		{ 365, 446, 57 },
		{ 365, 447, 57 },
		{ 365, 448, 57 },
		{ 365, 449, 57 },
		{ 365, 450, 57 },
		{ 365, 451, 57 },
		{ 365, 452, 57 },
		{ 365, 453, 57 },
		{ 365, 454, 57 },
		{ 365, 455, 57 },
		{ 365, 456, 57 },
		{ 365, 457, 57 },
		{ 365, 458, 57 },
		{ 365, 459, 57 },
		{ 365, 460, 57 },
		{ 365, 461, 57 },
		{ 365, 462, 57 },
		{ 365, 463, 57 },
		{ 365, 464, 57 },
		{ 365, 465, 57 },
		{ 365, 466, 57 },
		{ 365, 467, 57 },
		{ 365, 468, 57 },
		{ 365, 469, 57 },
		{ 365, 470, 57 },
		{ 365, 471, 57 },
		{ 365, 472, 57 },
		{ 366, 403, 57 },
		{ 366, 404, 57 },
		{ 366, 405, 57 },
		{ 366, 406, 57 },
		{ 366, 407, 57 },
		{ 366, 408, 57 },
		{ 366, 409, 57 },
		{ 366, 410, 57 },
		{ 366, 411, 57 },
		{ 366, 412, 57 },
		{ 366, 413, 57 },
		{ 366, 414, 57 },
		{ 366, 415, 57 },
		{ 366, 416, 57 },
		{ 366, 417, 57 },
		{ 366, 418, 57 },
		{ 366, 419, 57 },
		{ 366, 420, 57 },
		{ 366, 421, 57 },
		{ 366, 422, 57 },
		{ 366, 423, 57 },
		{ 366, 424, 57 },
		{ 366, 425, 57 },
		{ 366, 426, 57 },
		{ 366, 427, 57 },
		{ 366, 428, 57 },
		{ 366, 429, 57 },
		{ 366, 430, 57 },
		{ 366, 431, 57 },
		{ 366, 432, 57 },
		{ 366, 433, 57 },
		{ 366, 434, 57 },
		{ 366, 435, 57 },
		{ 366, 436, 57 },
		{ 366, 437, 57 },
		{ 366, 438, 57 },
		{ 366, 439, 57 },
		{ 366, 440, 57 },
		{ 366, 441, 57 },
		{ 366, 442, 57 },
		{ 366, 443, 57 },
		{ 366, 444, 57 },
		{ 366, 445, 57 },
		{ 366, 446, 57 },
		{ 366, 447, 57 },
		{ 366, 448, 57 },
		{ 366, 449, 57 },
		{ 366, 450, 57 },
		{ 366, 451, 57 },
		{ 366, 452, 57 },
		{ 366, 453, 57 },
		{ 366, 454, 57 },
		{ 366, 455, 57 },
		{ 366, 456, 57 },
		{ 366, 457, 57 },
		{ 366, 458, 57 },
		{ 366, 459, 57 },
		{ 366, 460, 57 },
		{ 366, 461, 57 },
		{ 366, 462, 57 },
		{ 366, 463, 57 },
		{ 366, 464, 57 },
		{ 366, 465, 57 },
		{ 366, 466, 57 },
		{ 366, 467, 57 },
		{ 366, 468, 57 },
		{ 366, 469, 57 },
		{ 366, 470, 57 },
		{ 366, 471, 57 },
		{ 366, 472, 57 },
		{ 366, 473, 57 },
		{ 366, 474, 57 },
		{ 367, 404, 57 },
		{ 367, 405, 57 },
		{ 367, 406, 57 },
		{ 367, 407, 57 },
		{ 367, 408, 57 },
		{ 367, 409, 57 },
		{ 367, 410, 57 },
		{ 367, 411, 57 },
		{ 367, 412, 57 },
		{ 367, 413, 57 },
		{ 367, 414, 57 },
		{ 367, 415, 57 },
		{ 367, 416, 57 },
		{ 367, 417, 57 },
		{ 367, 418, 57 },
		{ 367, 419, 57 },
		{ 367, 420, 57 },
		{ 367, 421, 57 },
		{ 367, 422, 57 },
		{ 367, 423, 57 },
		{ 367, 424, 57 },
		{ 367, 425, 57 },
		{ 367, 426, 57 },
		{ 367, 427, 57 },
		{ 367, 428, 57 },
		{ 367, 429, 57 },
		{ 367, 430, 57 },
		{ 367, 431, 57 },
		{ 367, 432, 57 },
		{ 367, 433, 57 },
		{ 367, 434, 57 },
		{ 367, 435, 57 },
		{ 367, 436, 57 },
		{ 367, 437, 57 },
		{ 367, 438, 57 },
		{ 367, 439, 57 },
		{ 367, 440, 57 },
		{ 367, 441, 57 },
		{ 367, 442, 57 },
		{ 367, 443, 57 },
		{ 367, 444, 57 },
		{ 367, 445, 57 },
		{ 367, 446, 57 },
		{ 367, 447, 57 },
		{ 367, 448, 57 },
		{ 367, 449, 57 },
		{ 367, 450, 57 },
		{ 367, 451, 57 },
		{ 367, 452, 57 },
		{ 367, 453, 57 },
		{ 367, 454, 57 },
		{ 367, 455, 57 },
		{ 367, 456, 57 },
		{ 367, 457, 57 },
		{ 367, 458, 57 },
		{ 367, 459, 57 },
		{ 367, 460, 57 },
		{ 367, 461, 57 },
		{ 367, 462, 57 },
		{ 367, 463, 57 },
		{ 367, 464, 57 },
		{ 367, 465, 57 },
		{ 367, 466, 57 },
		{ 367, 467, 57 },
		{ 367, 468, 57 },
		{ 367, 469, 57 },
		{ 367, 470, 57 },
		{ 367, 471, 57 },
		{ 367, 472, 57 },
		{ 367, 473, 57 },
		{ 367, 474, 57 },
		{ 367, 475, 57 },
		{ 367, 476, 57 },
		{ 368, 406, 57 },
		{ 368, 407, 57 },
		{ 368, 408, 57 },
		{ 368, 409, 57 },
		{ 368, 410, 57 },
		{ 368, 411, 57 },
		{ 368, 412, 57 },
		{ 368, 413, 57 },
		{ 368, 414, 57 },
		{ 368, 415, 57 },
		{ 368, 416, 57 },
		{ 368, 417, 57 },
		{ 368, 418, 57 },
		{ 368, 419, 57 },
		{ 368, 420, 57 },
		{ 368, 421, 57 },
		{ 368, 422, 57 },
		{ 368, 423, 57 },
		{ 368, 424, 57 },
		{ 368, 425, 57 },
		{ 368, 426, 57 },
		{ 368, 427, 57 },
		{ 368, 428, 57 },
		{ 368, 429, 57 },
		{ 368, 430, 57 },
		{ 368, 431, 57 },
		{ 368, 432, 57 },
		{ 368, 433, 57 },
		{ 368, 434, 57 },
		{ 368, 435, 57 },
		{ 368, 436, 57 },
		{ 368, 437, 57 },
		{ 368, 438, 57 },
		{ 368, 439, 57 },
		{ 368, 440, 57 },
		{ 368, 441, 57 },
		{ 368, 442, 57 },
		{ 368, 443, 57 },
		{ 368, 444, 57 },
		{ 368, 445, 57 },
		{ 368, 446, 57 },
		{ 368, 447, 57 },
		{ 368, 448, 57 },
		{ 368, 449, 57 },
		{ 368, 450, 57 },
		{ 368, 451, 57 },
		{ 368, 452, 57 },
		{ 368, 453, 57 },
		{ 368, 454, 57 },
		{ 368, 455, 57 },
		{ 368, 456, 57 },
		{ 368, 457, 57 },
		{ 368, 458, 57 },
		{ 368, 459, 57 },
		{ 368, 460, 57 },
		{ 368, 461, 57 },
		{ 368, 462, 57 },
		{ 368, 463, 57 },
		{ 368, 464, 57 },
		{ 368, 465, 57 },
		{ 368, 466, 57 },
		{ 368, 467, 57 },
		{ 368, 468, 57 },
		{ 368, 469, 57 },
		{ 368, 470, 57 },
		{ 368, 471, 57 },
		{ 368, 472, 57 },
		{ 368, 473, 57 },
		{ 368, 474, 57 },
		{ 368, 475, 57 },
		{ 368, 476, 57 },
		{ 368, 477, 57 },
		{ 368, 478, 57 },
		{ 369, 407, 57 },
		{ 369, 408, 57 },
		{ 369, 409, 57 },
		{ 369, 410, 57 },
		{ 369, 411, 57 },
		{ 369, 412, 57 },
		{ 369, 413, 57 },
		{ 369, 414, 57 },
		{ 369, 415, 57 },
		{ 369, 416, 57 },
		{ 369, 417, 57 },
		{ 369, 418, 57 },
		{ 369, 419, 57 },
		{ 369, 420, 57 },
		{ 369, 421, 57 },
		{ 369, 422, 57 },
		{ 369, 423, 57 },
		{ 369, 424, 57 },
		{ 369, 425, 57 },
		{ 369, 426, 57 },
		{ 369, 427, 57 },
		{ 369, 428, 57 },
		{ 369, 429, 57 },
		{ 369, 430, 57 },
		{ 369, 431, 57 },
		{ 369, 432, 57 },
		{ 369, 433, 57 },
		{ 369, 434, 57 },
		{ 369, 435, 57 },
		{ 369, 436, 57 },
		{ 369, 437, 57 },
		{ 369, 438, 57 },
		{ 369, 439, 57 },
		{ 369, 440, 57 },
		{ 369, 441, 57 },
		{ 369, 442, 57 },
		{ 369, 443, 57 },
		{ 369, 444, 57 },
		{ 369, 445, 57 },
		{ 369, 446, 57 },
		{ 369, 447, 57 },
		{ 369, 448, 57 },
		{ 369, 449, 57 },
		{ 369, 450, 57 },
		{ 369, 451, 57 },
		{ 369, 452, 57 },
		{ 369, 453, 57 },
		{ 369, 454, 57 },
		{ 369, 455, 57 },
		{ 369, 456, 57 },
		{ 369, 457, 57 },
		{ 369, 458, 57 },
		{ 369, 459, 57 },
		{ 369, 460, 57 },
		{ 369, 461, 57 },
		{ 369, 462, 57 },
		{ 369, 463, 57 },
		{ 369, 464, 57 },
		{ 369, 465, 57 },
		{ 369, 466, 57 },
		{ 369, 467, 57 },
		{ 369, 468, 57 },
		{ 369, 469, 57 },
		{ 369, 470, 57 },
		{ 369, 471, 57 },
		{ 369, 472, 57 },
		{ 369, 473, 57 },
		{ 369, 474, 57 },
		{ 369, 475, 57 },
		{ 369, 476, 57 },
		{ 369, 477, 57 },
		{ 369, 478, 57 },
		{ 369, 479, 57 },
		{ 369, 480, 57 },
		{ 370, 408, 57 },
		{ 370, 409, 57 },
		{ 370, 410, 57 },
		{ 370, 411, 57 },
		{ 370, 412, 57 },
		{ 370, 413, 57 },
		{ 370, 414, 57 },
		{ 370, 415, 57 },
		{ 370, 416, 57 },
		{ 370, 417, 57 },
		{ 370, 418, 57 },
		{ 370, 419, 57 },
		{ 370, 420, 57 },
		{ 370, 421, 57 },
		{ 370, 422, 57 },
		{ 370, 423, 57 },
		{ 370, 424, 57 },
		{ 370, 425, 57 },
		{ 370, 426, 57 },
		{ 370, 427, 57 },
		{ 370, 428, 57 },
		{ 370, 429, 57 },
		{ 370, 430, 57 },
		{ 370, 431, 57 },
		{ 370, 432, 57 },
		{ 370, 433, 57 },
		{ 370, 434, 57 },
		{ 370, 435, 57 },
		{ 370, 436, 57 },
		{ 370, 437, 57 },
		{ 370, 438, 57 },
		{ 370, 439, 57 },
		{ 370, 440, 57 },
		{ 370, 441, 57 },
		{ 370, 442, 57 },
		{ 370, 443, 57 },
		{ 370, 444, 57 },
		{ 370, 445, 57 },
		{ 370, 446, 57 },
		{ 370, 447, 57 },
		{ 370, 448, 57 },
		{ 370, 449, 57 },
		{ 370, 450, 57 },
		{ 370, 451, 57 },
		{ 370, 452, 57 },
		{ 370, 453, 57 },
		{ 370, 454, 57 },
		{ 370, 455, 57 },
		{ 370, 456, 57 },
		{ 370, 457, 57 },
		{ 370, 458, 57 },
		{ 370, 459, 57 },
		{ 370, 460, 57 },
		{ 370, 461, 57 },
		{ 370, 462, 57 },
		{ 370, 463, 57 },
		{ 370, 464, 57 },
		{ 370, 465, 57 },
		{ 370, 466, 57 },
		{ 370, 467, 57 },
		{ 370, 468, 57 },
		{ 370, 469, 57 },
		{ 370, 470, 57 },
		{ 370, 471, 57 },
		{ 370, 472, 57 },
		{ 370, 473, 57 },
		{ 370, 474, 57 },
		{ 370, 475, 57 },
		{ 370, 476, 57 },
		{ 370, 477, 57 },
		{ 370, 478, 57 },
		{ 370, 479, 57 },
		{ 370, 480, 57 },
		{ 370, 481, 57 },
		{ 370, 482, 57 },
		{ 371, 410, 57 },
		{ 371, 411, 57 },
		{ 371, 412, 57 },
		{ 371, 413, 57 },
		{ 371, 414, 57 },
		{ 371, 415, 57 },
		{ 371, 416, 57 },
		{ 371, 417, 57 },
		{ 371, 418, 57 },
		{ 371, 419, 57 },
		{ 371, 420, 57 },
		{ 371, 421, 57 },
		{ 371, 422, 57 },
		{ 371, 423, 57 },
		{ 371, 424, 57 },
		{ 371, 425, 57 },
		{ 371, 426, 57 },
		{ 371, 427, 57 },
		{ 371, 428, 57 },
		{ 371, 429, 57 },
		{ 371, 430, 57 },
		{ 371, 431, 57 },
		{ 371, 432, 57 },
		{ 371, 433, 57 },
		{ 371, 434, 57 },
		{ 371, 435, 57 },
		{ 371, 436, 57 },
		{ 371, 437, 57 },
		{ 371, 438, 57 },
		{ 371, 439, 57 },
		{ 371, 440, 57 },
		{ 371, 441, 57 },
		{ 371, 442, 57 },
		{ 371, 443, 57 },
		{ 371, 444, 57 },
		{ 371, 445, 57 },
		{ 371, 446, 57 },
		{ 371, 447, 57 },
		{ 371, 448, 57 },
		{ 371, 449, 57 },
		{ 371, 450, 57 },
		{ 371, 451, 57 },
		{ 371, 452, 57 },
		{ 371, 453, 57 },
		{ 371, 454, 57 },
		{ 371, 455, 57 },
		{ 371, 456, 57 },
		{ 371, 457, 57 },
		{ 371, 458, 57 },
		{ 371, 459, 57 },
		{ 371, 460, 57 },
		{ 371, 461, 57 },
		{ 371, 462, 57 },
		{ 371, 463, 57 },
		{ 371, 464, 57 },
		{ 371, 465, 57 },
		{ 371, 466, 57 },
		{ 371, 467, 57 },
		{ 371, 468, 57 },
		{ 371, 469, 57 },
		{ 371, 470, 57 },
		{ 371, 471, 57 },
		{ 371, 472, 57 },
		{ 371, 473, 57 },
		{ 371, 474, 57 },
		{ 371, 475, 57 },
		{ 371, 476, 57 },
		{ 371, 477, 57 },
		{ 371, 478, 57 },
		{ 371, 479, 57 },
		{ 371, 480, 57 },
		{ 371, 481, 57 },
		{ 371, 482, 57 },
		{ 371, 483, 57 },
		{ 371, 484, 57 },
		{ 372, 411, 57 },
		{ 372, 412, 57 },
		{ 372, 413, 57 },
		{ 372, 414, 57 },
		{ 372, 415, 57 },
		{ 372, 416, 57 },
		{ 372, 417, 57 },
		{ 372, 418, 57 },
		{ 372, 419, 57 },
		{ 372, 420, 57 },
		{ 372, 421, 57 },
		{ 372, 422, 57 },
		{ 372, 423, 57 },
		{ 372, 424, 57 },
		{ 372, 425, 57 },
		{ 372, 426, 57 },
		{ 372, 427, 57 },
		{ 372, 428, 57 },
		{ 372, 429, 57 },
		{ 372, 430, 57 },
		{ 372, 431, 57 },
		{ 372, 432, 57 },
		{ 372, 433, 57 },
		{ 372, 434, 57 },
		{ 372, 435, 57 },
		{ 372, 436, 57 },
		{ 372, 437, 57 },
		{ 372, 438, 57 },
		{ 372, 439, 57 },
		{ 372, 440, 57 },
		{ 372, 441, 57 },
		{ 372, 442, 57 },
		{ 372, 443, 57 },
		{ 372, 444, 57 },
		{ 372, 445, 57 },
		{ 372, 446, 57 },
		{ 372, 447, 57 },
		{ 372, 448, 57 },
		{ 372, 449, 57 },
		{ 372, 450, 57 },
		{ 372, 451, 57 },
		{ 372, 452, 57 },
		{ 372, 453, 57 },
		{ 372, 454, 57 },
		{ 372, 455, 57 },
		{ 372, 456, 57 },
		{ 372, 457, 57 },
		{ 372, 458, 57 },
		{ 372, 459, 57 },
		{ 372, 460, 57 },
		{ 372, 461, 57 },
		{ 372, 462, 57 },
		{ 372, 463, 57 },
		{ 372, 464, 57 },
		{ 372, 465, 57 },
		{ 372, 466, 57 },
		{ 372, 467, 57 },
		{ 372, 468, 57 },
		{ 372, 469, 57 },
		{ 372, 470, 57 },
		{ 372, 471, 57 },
		{ 372, 472, 57 },
		{ 372, 473, 57 },
		{ 372, 474, 57 },
		{ 372, 475, 57 },
		{ 372, 476, 57 },
		{ 372, 477, 57 },
		{ 372, 478, 57 },
		{ 372, 479, 57 },
		{ 372, 480, 57 },
		{ 372, 481, 57 },
		{ 372, 482, 57 },
		{ 372, 483, 57 },
		{ 372, 484, 57 },
		{ 372, 485, 57 },
		{ 372, 486, 57 },
		{ 373, 412, 57 },
		{ 373, 413, 57 },
		{ 373, 414, 57 },
		{ 373, 415, 57 },
		{ 373, 416, 57 },
		{ 373, 417, 57 },
		{ 373, 418, 57 },
		{ 373, 419, 57 },
		{ 373, 420, 57 },
		{ 373, 421, 57 },
		{ 373, 422, 57 },
		{ 373, 423, 57 },
		{ 373, 424, 57 },
		{ 373, 425, 57 },
		{ 373, 426, 57 },
		{ 373, 427, 57 },
		{ 373, 428, 57 },
		{ 373, 429, 57 },
		{ 373, 430, 57 },
		{ 373, 431, 57 },
		{ 373, 432, 57 },
		{ 373, 433, 57 },
		{ 373, 434, 57 },
		{ 373, 435, 57 },
		{ 373, 436, 57 },
		{ 373, 437, 57 },
		{ 373, 438, 57 },
		{ 373, 439, 57 },
		{ 373, 440, 57 },
		{ 373, 441, 57 },
		{ 373, 442, 57 },
		{ 373, 443, 57 },
		{ 373, 444, 57 },
		{ 373, 445, 57 },
		{ 373, 446, 57 },
		{ 373, 447, 57 },
		{ 373, 448, 57 },
		{ 373, 449, 57 },
		{ 373, 450, 57 },
		{ 373, 451, 57 },
		{ 373, 452, 57 },
		{ 373, 453, 57 },
		{ 373, 454, 57 },
		{ 373, 455, 57 },
		{ 373, 456, 57 },
		{ 373, 457, 57 },
		{ 373, 458, 57 },
		{ 373, 459, 57 },
		{ 373, 460, 57 },
		{ 373, 461, 57 },
		{ 373, 462, 57 },
		{ 373, 463, 57 },
		{ 373, 464, 57 },
		{ 373, 465, 57 },
		{ 373, 466, 57 },
		{ 373, 467, 57 },
		{ 373, 468, 57 },
		{ 373, 469, 57 },
		{ 373, 470, 57 },
		{ 373, 471, 57 },
		{ 373, 472, 57 },
		{ 373, 473, 57 },
		{ 373, 474, 57 },
		{ 373, 475, 57 },
		{ 373, 476, 57 },
		{ 373, 477, 57 },
		{ 373, 478, 57 },
		{ 373, 479, 57 },
		{ 373, 480, 57 },
		{ 373, 481, 57 },
		{ 373, 482, 57 },
		{ 373, 483, 57 },
		{ 373, 484, 57 },
		{ 373, 485, 57 },
		{ 373, 486, 57 },
		{ 373, 487, 57 },
		{ 373, 488, 57 },
		{ 374, 414, 57 },
		{ 374, 415, 57 },
		{ 374, 416, 57 },
		{ 374, 417, 57 },
		{ 374, 418, 57 },
		{ 374, 419, 57 },
		{ 374, 420, 57 },
		{ 374, 421, 57 },
		{ 374, 422, 57 },
		{ 374, 423, 57 },
		{ 374, 424, 57 },
		{ 374, 425, 57 },
		{ 374, 426, 57 },
		{ 374, 427, 57 },
		{ 374, 428, 57 },
		{ 374, 429, 57 },
		{ 374, 430, 57 },
		{ 374, 431, 57 },
		{ 374, 432, 57 },
		{ 374, 433, 57 },
		{ 374, 434, 57 },
		{ 374, 435, 57 },
		{ 374, 436, 57 },
		{ 374, 437, 57 },
		{ 374, 438, 57 },
		{ 374, 439, 57 },
		{ 374, 440, 57 },
		{ 374, 441, 57 },
		{ 374, 442, 57 },
		{ 374, 443, 57 },
		{ 374, 444, 57 },
		{ 374, 445, 57 },
		{ 374, 446, 57 },
		{ 374, 447, 57 },
		{ 374, 448, 57 },
		{ 374, 449, 57 },
		{ 374, 450, 57 },
		{ 374, 451, 57 },
		{ 374, 452, 57 },
		{ 374, 453, 57 },
		{ 374, 454, 57 },
		{ 374, 455, 57 },
		{ 374, 456, 57 },
		{ 374, 457, 57 },
		{ 374, 458, 57 },
		{ 374, 459, 57 },
		{ 374, 460, 57 },
		{ 374, 461, 57 },
		{ 374, 462, 57 },
		{ 374, 463, 57 },
		{ 374, 464, 57 },
		{ 374, 465, 57 },
		{ 374, 466, 57 },
		{ 374, 467, 57 },
		{ 374, 468, 57 },
		{ 374, 469, 57 },
		{ 374, 470, 57 },
		{ 374, 471, 57 },
		{ 374, 472, 57 },
		{ 374, 473, 57 },
		{ 374, 474, 57 },
		{ 374, 475, 57 },
		{ 374, 476, 57 },
		{ 374, 477, 57 },
		{ 374, 478, 57 },
		{ 374, 479, 57 },
		{ 374, 480, 57 },
		{ 374, 481, 57 },
		{ 374, 482, 57 },
		{ 374, 483, 57 },
		{ 374, 484, 57 },
		{ 374, 485, 57 },
		{ 374, 486, 57 },
		{ 374, 487, 57 },
		{ 374, 488, 57 },
		{ 374, 489, 57 },
		{ 374, 490, 57 },
		{ 375, 415, 57 },
		{ 375, 416, 57 },
		{ 375, 417, 57 },
		{ 375, 418, 57 },
		{ 375, 419, 57 },
		{ 375, 420, 57 },
		{ 375, 421, 57 },
		{ 375, 422, 57 },
		{ 375, 423, 57 },
		{ 375, 424, 57 },
		{ 375, 425, 57 },
		{ 375, 426, 57 },
		{ 375, 427, 57 },
		{ 375, 428, 57 },
		{ 375, 429, 57 },
		{ 375, 430, 57 },
		{ 375, 431, 57 },
		{ 375, 432, 57 },
		{ 375, 433, 57 },
		{ 375, 434, 57 },
		{ 375, 435, 57 },
		{ 375, 436, 57 },
		{ 375, 437, 57 },
		{ 375, 438, 57 },
		{ 375, 439, 57 },
		{ 375, 440, 57 },
		{ 375, 441, 57 },
		{ 375, 442, 57 },
		{ 375, 443, 57 },
		{ 375, 444, 57 },
		{ 375, 445, 57 },
		{ 375, 446, 57 },
		{ 375, 447, 57 },
		{ 375, 448, 57 },
		{ 375, 449, 57 },
		{ 375, 450, 57 },
		{ 375, 451, 57 },
		{ 375, 452, 57 },
		{ 375, 453, 57 },
		{ 375, 454, 57 },
		{ 375, 455, 57 },
		{ 375, 456, 57 },
		{ 375, 457, 57 },
		{ 375, 458, 57 },
		{ 375, 459, 57 },
		{ 375, 460, 57 },
		{ 375, 461, 57 },
		{ 375, 462, 57 },
		{ 375, 463, 57 },
		{ 375, 464, 57 },
		{ 375, 465, 57 },
		{ 375, 466, 57 },
		{ 375, 467, 57 },
		{ 375, 468, 57 },
		{ 375, 469, 57 },
		{ 375, 470, 57 },
		{ 375, 471, 57 },
		{ 375, 472, 57 },
		{ 375, 473, 57 },
		{ 375, 474, 57 },
		{ 375, 475, 57 },
		{ 375, 476, 57 },
		{ 375, 477, 57 },
		{ 375, 478, 57 },
		{ 375, 479, 57 },
		{ 375, 480, 57 },
		{ 375, 481, 57 },
		{ 375, 482, 57 },
		{ 375, 483, 57 },
		{ 375, 484, 57 },
		{ 375, 485, 57 },
		{ 375, 486, 57 },
		{ 375, 487, 57 },
		{ 375, 488, 57 },
		{ 375, 489, 57 },
		{ 375, 490, 57 },
		{ 375, 491, 57 },
		{ 375, 492, 57 },
		{ 376, 416, 57 },
		{ 376, 417, 57 },
		{ 376, 418, 57 },
		{ 376, 419, 57 },
		{ 376, 420, 57 },
		{ 376, 421, 57 },
		{ 376, 422, 57 },
		{ 376, 423, 57 },
		{ 376, 424, 57 },
		{ 376, 425, 57 },
		{ 376, 426, 57 },
		{ 376, 427, 57 },
		{ 376, 428, 57 },
		{ 376, 429, 57 },
		{ 376, 430, 57 },
		{ 376, 431, 57 },
		{ 376, 432, 57 },
		{ 376, 433, 57 },
		{ 376, 434, 57 },
		{ 376, 435, 57 },
		{ 376, 436, 57 },
		{ 376, 437, 57 },
		{ 376, 438, 57 },
		{ 376, 439, 57 },
		{ 376, 440, 57 },
		{ 376, 441, 57 },
		{ 376, 442, 57 },
		{ 376, 443, 57 },
		{ 376, 444, 57 },
		{ 376, 445, 57 },
		{ 376, 446, 57 },
		{ 376, 447, 57 },
		{ 376, 448, 57 },
		{ 376, 449, 57 },
		{ 376, 450, 57 },
		{ 376, 451, 57 },
		{ 376, 452, 57 },
		{ 376, 453, 57 },
		{ 376, 454, 57 },
		{ 376, 455, 57 },
		{ 376, 456, 57 },
		{ 376, 457, 57 },
		{ 376, 458, 57 },
		{ 376, 459, 57 },
		{ 376, 460, 57 },
		{ 376, 461, 57 },
		{ 376, 462, 57 },
		{ 376, 463, 57 },
		{ 376, 464, 57 },
		{ 376, 465, 57 },
		{ 376, 466, 57 },
		{ 376, 467, 57 },
		{ 376, 468, 57 },
		{ 376, 469, 57 },
		{ 376, 470, 57 },
		{ 376, 471, 57 },
		{ 376, 472, 57 },
		{ 376, 473, 57 },
		{ 376, 474, 57 },
		{ 376, 475, 57 },
		{ 376, 476, 57 },
		{ 376, 477, 57 },
		{ 376, 478, 57 },
		{ 376, 479, 57 },
		{ 376, 480, 57 },
		{ 376, 481, 57 },
		{ 376, 482, 57 },
		{ 376, 483, 57 },
		{ 376, 484, 57 },
		{ 376, 485, 57 },
		{ 376, 486, 57 },
		{ 376, 487, 57 },
		{ 376, 488, 57 },
		{ 376, 489, 57 },
		{ 376, 490, 57 },
		{ 376, 491, 57 },
		{ 376, 492, 57 },
		{ 376, 493, 57 },
		{ 376, 494, 57 },
		{ 377, 418, 57 },
		{ 377, 419, 57 },
		{ 377, 420, 57 },
		{ 377, 421, 57 },
		{ 377, 422, 57 },
		{ 377, 423, 57 },
		{ 377, 424, 57 },
		{ 377, 425, 57 },
		{ 377, 426, 57 },
		{ 377, 427, 57 },
		{ 377, 428, 57 },
		{ 377, 429, 57 },
		{ 377, 430, 57 },
		{ 377, 431, 57 },
		{ 377, 432, 57 },
		{ 377, 433, 57 },
		{ 377, 434, 57 },
		{ 377, 435, 57 },
		{ 377, 436, 57 },
		{ 377, 437, 57 },
		{ 377, 438, 57 },
		{ 377, 439, 57 },
		{ 377, 440, 57 },
		{ 377, 441, 57 },
		{ 377, 442, 57 },
		{ 377, 443, 57 },
		{ 377, 444, 57 },
		{ 377, 445, 57 },
		{ 377, 446, 57 },
		{ 377, 447, 57 },
		{ 377, 448, 57 },
		{ 377, 449, 57 },
		{ 377, 450, 57 },
		{ 377, 451, 57 },
		{ 377, 452, 57 },
		{ 377, 453, 57 },
		{ 377, 454, 57 },
		{ 377, 455, 57 },
		{ 377, 456, 57 },
		{ 377, 457, 57 },
		{ 377, 458, 57 },
		{ 377, 459, 57 },
		{ 377, 460, 57 },
		{ 377, 461, 57 },
		{ 377, 462, 57 },
		{ 377, 463, 57 },
		{ 377, 464, 57 },
		{ 377, 465, 57 },
		{ 377, 466, 57 },
		{ 377, 467, 57 },
		{ 377, 468, 57 },
		{ 377, 469, 57 },
		{ 377, 470, 57 },
		{ 377, 471, 57 },
		{ 377, 472, 57 },
		{ 377, 473, 57 },
		{ 377, 474, 57 },
		{ 377, 475, 57 },
		{ 377, 476, 57 },
		{ 377, 477, 57 },
		{ 377, 478, 57 },
		{ 377, 479, 57 },
		{ 377, 480, 57 },
		{ 377, 481, 57 },
		{ 377, 482, 57 },
		{ 377, 483, 57 },
		{ 377, 484, 57 },
		{ 377, 485, 57 },
		{ 377, 486, 57 },
		{ 377, 487, 57 },
		{ 377, 488, 57 },
		{ 377, 489, 57 },
		{ 377, 490, 57 },
		{ 377, 491, 57 },
		{ 377, 492, 57 },
		{ 377, 493, 57 },
		{ 377, 494, 57 },
		{ 377, 495, 57 },
		{ 377, 496, 57 },
		{ 378, 419, 57 },
		{ 378, 420, 57 },
		{ 378, 421, 57 },
		{ 378, 422, 57 },
		{ 378, 423, 57 },
		{ 378, 424, 57 },
		{ 378, 425, 57 },
		{ 378, 426, 57 },
		{ 378, 427, 57 },
		{ 378, 428, 57 },
		{ 378, 429, 57 },
		{ 378, 430, 57 },
		{ 378, 431, 57 },
		{ 378, 432, 57 },
		{ 378, 433, 57 },
		{ 378, 434, 57 },
		{ 378, 435, 57 },
		{ 378, 436, 57 },
		{ 378, 437, 57 },
		{ 378, 438, 57 },
		{ 378, 439, 57 },
		{ 378, 440, 57 },
		{ 378, 441, 57 },
		{ 378, 442, 57 },
		{ 378, 443, 57 },
		{ 378, 444, 57 },
		{ 378, 445, 57 },
		{ 378, 446, 57 },
		{ 378, 447, 57 },
		{ 378, 448, 57 },
		{ 378, 449, 57 },
		{ 378, 450, 57 },
		{ 378, 451, 57 },
		{ 378, 452, 57 },
		{ 378, 453, 57 },
		{ 378, 454, 57 },
		{ 378, 455, 57 },
		{ 378, 456, 57 },
		{ 378, 457, 57 },
		{ 378, 458, 57 },
		{ 378, 459, 57 },
		{ 378, 460, 57 },
		{ 378, 461, 57 },
		{ 378, 462, 57 },
		{ 378, 463, 57 },
		{ 378, 464, 57 },
		{ 378, 465, 57 },
		{ 378, 466, 57 },
		{ 378, 467, 57 },
		{ 378, 468, 57 },
		{ 378, 469, 57 },
		{ 378, 470, 57 },
		{ 378, 471, 57 },
		{ 378, 472, 57 },
		{ 378, 473, 57 },
		{ 378, 474, 57 },
		{ 378, 475, 57 },
		{ 378, 476, 57 },
		{ 378, 477, 57 },
		{ 378, 478, 57 },
		{ 378, 479, 57 },
		{ 378, 480, 57 },
		{ 378, 481, 57 },
		{ 378, 482, 57 },
		{ 378, 483, 57 },
		{ 378, 484, 57 },
		{ 378, 485, 57 },
		{ 378, 486, 57 },
		{ 378, 487, 57 },
		{ 378, 488, 57 },
		{ 378, 489, 57 },
		{ 378, 490, 57 },
		{ 378, 491, 57 },
		{ 378, 492, 57 },
		{ 378, 493, 57 },
		{ 378, 494, 57 },
		{ 378, 495, 57 },
		{ 378, 496, 57 },
		{ 378, 497, 57 },
		{ 378, 498, 57 },
		{ 379, 420, 57 },
		{ 379, 421, 57 },
		{ 379, 422, 57 },
		{ 379, 423, 57 },
		{ 379, 424, 57 },
		{ 379, 425, 57 },
		{ 379, 426, 57 },
		{ 379, 427, 57 },
		{ 379, 428, 57 },
		{ 379, 429, 57 },
		{ 379, 430, 57 },
		{ 379, 431, 57 },
		{ 379, 432, 57 },
		{ 379, 433, 57 },
		{ 379, 434, 57 },
		{ 379, 435, 57 },
		{ 379, 436, 57 },
		{ 379, 437, 57 },
		{ 379, 438, 57 },
		{ 379, 439, 57 },
		{ 379, 440, 57 },
		{ 379, 441, 57 },
		{ 379, 442, 57 },
		{ 379, 443, 57 },
		{ 379, 444, 57 },
		{ 379, 445, 57 },
		{ 379, 446, 57 },
		{ 379, 447, 57 },
		{ 379, 448, 57 },
		{ 379, 449, 57 },
		{ 379, 450, 57 },
		{ 379, 451, 57 },
		{ 379, 452, 57 },
		{ 379, 453, 57 },
		{ 379, 454, 57 },
		{ 379, 455, 57 },
		{ 379, 456, 57 },
		{ 379, 457, 57 },
		{ 379, 458, 57 },
		{ 379, 459, 57 },
		{ 379, 460, 57 },
		{ 379, 461, 57 },
		{ 379, 462, 57 },
		{ 379, 463, 57 },
		{ 379, 464, 57 },
		{ 379, 465, 57 },
		{ 379, 466, 57 },
		{ 379, 467, 57 },
		{ 379, 468, 57 },
		{ 379, 469, 57 },
		{ 379, 470, 57 },
		{ 379, 471, 57 },
		{ 379, 472, 57 },
		{ 379, 473, 57 },
		{ 379, 474, 57 },
		{ 379, 475, 57 },
		{ 379, 476, 57 },
		{ 379, 477, 57 },
		{ 379, 478, 57 },
		{ 379, 479, 57 },
		{ 379, 480, 57 },
		{ 379, 481, 57 },
		{ 379, 482, 57 },
		{ 379, 483, 57 },
		{ 379, 484, 57 },
		{ 379, 485, 57 },
		{ 379, 486, 57 },
		{ 379, 487, 57 },
		{ 379, 488, 57 },
		{ 379, 489, 57 },
		{ 379, 490, 57 },
		{ 379, 491, 57 },
		{ 379, 492, 57 },
		{ 379, 493, 57 },
		{ 379, 494, 57 },
		{ 379, 495, 57 },
		{ 379, 496, 57 },
		{ 379, 497, 57 },
		{ 379, 498, 57 },
		{ 379, 499, 57 },
		{ 379, 500, 57 },
		{ 380, 422, 57 },
		{ 380, 423, 57 },
		{ 380, 424, 57 },
		{ 380, 425, 57 },
		{ 380, 426, 57 },
		{ 380, 427, 57 },
		{ 380, 428, 57 },
		{ 380, 429, 57 },
		{ 380, 430, 57 },
		{ 380, 431, 57 },
		{ 380, 432, 57 },
		{ 380, 433, 57 },
		{ 380, 434, 57 },
		{ 380, 435, 57 },
		{ 380, 436, 57 },
		{ 380, 437, 57 },
		{ 380, 438, 57 },
		{ 380, 439, 57 },
		{ 380, 440, 57 },
		{ 380, 441, 57 },
		{ 380, 442, 57 },
		{ 380, 443, 57 },
		{ 380, 444, 57 },
		{ 380, 445, 57 },
		{ 380, 446, 57 },
		{ 380, 447, 57 },
		{ 380, 448, 57 },
		{ 380, 449, 57 },
		{ 380, 450, 57 },
		{ 380, 451, 57 },
		{ 380, 452, 57 },
		{ 380, 453, 57 },
		{ 380, 454, 57 },
		{ 380, 455, 57 },
		{ 380, 456, 57 },
		{ 380, 457, 57 },
		{ 380, 458, 57 },
		{ 380, 459, 57 },
		{ 380, 460, 57 },
		{ 380, 461, 57 },
		{ 380, 462, 57 },
		{ 380, 463, 57 },
		{ 380, 464, 57 },
		{ 380, 465, 57 },
		{ 380, 466, 57 },
		{ 380, 467, 57 },
		{ 380, 468, 57 },
		{ 380, 469, 57 },
		{ 380, 470, 57 },
		{ 380, 471, 57 },
		{ 380, 472, 57 },
		{ 380, 473, 57 },
		{ 380, 474, 57 },
		{ 380, 475, 57 },
		{ 380, 476, 57 },
		{ 380, 477, 57 },
		{ 380, 478, 57 },
		{ 380, 479, 57 },
		{ 380, 480, 57 },
		{ 380, 481, 57 },
		{ 380, 482, 57 },
		{ 380, 483, 57 },
		{ 380, 484, 57 },
		{ 380, 485, 57 },
		{ 380, 486, 57 },
		{ 380, 487, 57 },
		{ 380, 488, 57 },
		{ 380, 489, 57 },
		{ 380, 490, 57 },
		{ 380, 491, 57 },
		{ 380, 492, 57 },
		{ 380, 493, 57 },
		{ 380, 494, 57 },
		{ 380, 495, 57 },
		{ 380, 496, 57 },
		{ 380, 497, 57 },
		{ 380, 498, 57 },
		{ 380, 499, 57 },
		{ 380, 500, 57 },
		{ 380, 501, 57 },
		{ 380, 502, 57 },
		{ 381, 423, 57 },
		{ 381, 424, 57 },
		{ 381, 425, 57 },
		{ 381, 426, 57 },
		{ 381, 427, 57 },
		{ 381, 428, 57 },
		{ 381, 429, 57 },
		{ 381, 430, 57 },
		{ 381, 431, 57 },
		{ 381, 432, 57 },
		{ 381, 433, 57 },
		{ 381, 434, 57 },
		{ 381, 435, 57 },
		{ 381, 436, 57 },
		{ 381, 437, 57 },
		{ 381, 438, 57 },
		{ 381, 439, 57 },
		{ 381, 440, 57 },
		{ 381, 441, 57 },
		{ 381, 442, 57 },
		{ 381, 443, 57 },
		{ 381, 444, 57 },
		{ 381, 445, 57 },
		{ 381, 446, 57 },
		{ 381, 447, 57 },
		{ 381, 448, 57 },
		{ 381, 449, 57 },
		{ 381, 450, 57 },
		{ 381, 451, 57 },
		{ 381, 452, 57 },
		{ 381, 453, 57 },
		{ 381, 454, 57 },
		{ 381, 455, 57 },
		{ 381, 456, 57 },
		{ 381, 457, 57 },
		{ 381, 458, 57 },
		{ 381, 459, 57 },
		{ 381, 460, 57 },
		{ 381, 461, 57 },
		{ 381, 462, 57 },
		{ 381, 463, 57 },
		{ 381, 464, 57 },
		{ 381, 465, 57 },
		{ 381, 466, 57 },
		{ 381, 467, 57 },
		{ 381, 468, 57 },
		{ 381, 469, 57 },
		{ 381, 470, 57 },
		{ 381, 471, 57 },
		{ 381, 472, 57 },
		{ 381, 473, 57 },
		{ 381, 474, 57 },
		{ 381, 475, 57 },
		{ 381, 476, 57 },
		{ 381, 477, 57 },
		{ 381, 478, 57 },
		{ 381, 479, 57 },
		{ 381, 480, 57 },
		{ 381, 481, 57 },
		{ 381, 482, 57 },
		{ 381, 483, 57 },
		{ 381, 484, 57 },
		{ 381, 485, 57 },
		{ 381, 486, 57 },
		{ 381, 487, 57 },
		{ 381, 488, 57 },
		{ 381, 489, 57 },
		{ 381, 490, 57 },
		{ 381, 491, 57 },
		{ 381, 492, 57 },
		{ 381, 493, 57 },
		{ 381, 494, 57 },
		{ 381, 495, 57 },
		{ 381, 496, 57 },
		{ 381, 497, 57 },
		{ 381, 498, 57 },
		{ 381, 499, 57 },
		{ 381, 500, 57 },
		{ 381, 501, 57 },
		{ 381, 502, 57 },
		{ 381, 503, 57 },
		{ 381, 504, 57 },
		{ 382, 424, 57 },
		{ 382, 425, 57 },
		{ 382, 426, 57 },
		{ 382, 427, 57 },
		{ 382, 428, 57 },
		{ 382, 429, 57 },
		{ 382, 430, 57 },
		{ 382, 431, 57 },
		{ 382, 432, 57 },
		{ 382, 433, 57 },
		{ 382, 434, 57 },
		{ 382, 435, 57 },
		{ 382, 436, 57 },
		{ 382, 437, 57 },
		{ 382, 438, 57 },
		{ 382, 439, 57 },
		{ 382, 440, 57 },
		{ 382, 441, 57 },
		{ 382, 442, 57 },
		{ 382, 443, 57 },
		{ 382, 444, 57 },
		{ 382, 445, 57 },
		{ 382, 446, 57 },
		{ 382, 447, 57 },
		{ 382, 448, 57 },
		{ 382, 449, 57 },
		{ 382, 450, 57 },
		{ 382, 451, 57 },
		{ 382, 452, 57 },
		{ 382, 453, 57 },
		{ 382, 454, 57 },
		{ 382, 455, 57 },
		{ 382, 456, 57 },
		{ 382, 457, 57 },
		{ 382, 458, 57 },
		{ 382, 459, 57 },
		{ 382, 460, 57 },
		{ 382, 461, 57 },
		{ 382, 462, 57 },
		{ 382, 463, 57 },
		{ 382, 464, 57 },
		{ 382, 465, 57 },
		{ 382, 466, 57 },
		{ 382, 467, 57 },
		{ 382, 468, 57 },
		{ 382, 469, 57 },
		{ 382, 470, 57 },
		{ 382, 471, 57 },
		{ 382, 472, 57 },
		{ 382, 473, 57 },
		{ 382, 474, 57 },
		{ 382, 475, 57 },
		{ 382, 476, 57 },
		{ 382, 477, 57 },
		{ 382, 478, 57 },
		{ 382, 479, 57 },
		{ 382, 480, 57 },
		{ 382, 481, 57 },
		{ 382, 482, 57 },
		{ 382, 483, 57 },
		{ 382, 484, 57 },
		{ 382, 485, 57 },
		{ 382, 486, 57 },
		{ 382, 487, 57 },
		{ 382, 488, 57 },
		{ 382, 489, 57 },
		{ 382, 490, 57 },
		{ 382, 491, 57 },
		{ 382, 492, 57 },
		{ 382, 493, 57 },
		{ 382, 494, 57 },
		{ 382, 495, 57 },
		{ 382, 496, 57 },
		{ 382, 497, 57 },
		{ 382, 498, 57 },
		{ 382, 499, 57 },
		{ 382, 500, 57 },
		{ 382, 501, 57 },
		{ 382, 502, 57 },
		{ 382, 503, 57 },
		{ 382, 504, 57 },
		{ 382, 505, 57 },
		{ 382, 506, 57 },
		{ 383, 426, 57 },
		{ 383, 427, 57 },
		{ 383, 428, 57 },
		{ 383, 429, 57 },
		{ 383, 430, 57 },
		{ 383, 431, 57 },
		{ 383, 432, 57 },
		{ 383, 433, 57 },
		{ 383, 434, 57 },
		{ 383, 435, 57 },
		{ 383, 436, 57 },
		{ 383, 437, 57 },
		{ 383, 438, 57 },
		{ 383, 439, 57 },
		{ 383, 440, 57 },
		{ 383, 441, 57 },
		{ 383, 442, 57 },
		{ 383, 443, 57 },
		{ 383, 444, 57 },
		{ 383, 445, 57 },
		{ 383, 446, 57 },
		{ 383, 447, 57 },
		{ 383, 448, 57 },
		{ 383, 449, 57 },
		{ 383, 450, 57 },
		{ 383, 451, 57 },
		{ 383, 452, 57 },
		{ 383, 453, 57 },
		{ 383, 454, 57 },
		{ 383, 455, 57 },
		{ 383, 456, 57 },
		{ 383, 457, 57 },
		{ 383, 458, 57 },
		{ 383, 459, 57 },
		{ 383, 460, 57 },
		{ 383, 461, 57 },
		{ 383, 462, 57 },
		{ 383, 463, 57 },
		{ 383, 464, 57 },
		{ 383, 465, 57 },
		{ 383, 466, 57 },
		{ 383, 467, 57 },
		{ 383, 468, 57 },
		{ 383, 469, 57 },
		{ 383, 470, 57 },
		{ 383, 471, 57 },
		{ 383, 472, 57 },
		{ 383, 473, 57 },
		{ 383, 474, 57 },
		{ 383, 475, 57 },
		{ 383, 476, 57 },
		{ 383, 477, 57 },
		{ 383, 478, 57 },
		{ 383, 479, 57 },
		{ 383, 480, 57 },
		{ 383, 481, 57 },
		{ 383, 482, 57 },
		{ 383, 483, 57 },
		{ 383, 484, 57 },
		{ 383, 485, 57 },
		{ 383, 486, 57 },
		{ 383, 487, 57 },
		{ 383, 488, 57 },
		{ 383, 489, 57 },
		{ 383, 490, 57 },
		{ 383, 491, 57 },
		{ 383, 492, 57 },
		{ 383, 493, 57 },
		{ 383, 494, 57 },
		{ 383, 495, 57 },
		{ 383, 496, 57 },
		{ 383, 497, 57 },
		{ 383, 498, 57 },
		{ 383, 499, 57 },
		{ 383, 500, 57 },
		{ 383, 501, 57 },
		{ 383, 502, 57 },
		{ 383, 503, 57 },
		{ 383, 504, 57 },
		{ 383, 505, 57 },
		{ 383, 506, 57 },
		{ 383, 507, 57 },
		{ 383, 508, 57 },
		{ 384, 427, 57 },
		{ 384, 428, 57 },
		{ 384, 429, 57 },
		{ 384, 430, 57 },
		{ 384, 431, 57 },
		{ 384, 432, 57 },
		{ 384, 433, 57 },
		{ 384, 434, 57 },
		{ 384, 435, 57 },
		{ 384, 436, 57 },
		{ 384, 437, 57 },
		{ 384, 438, 57 },
		{ 384, 439, 57 },
		{ 384, 440, 57 },
		{ 384, 441, 57 },
		{ 384, 442, 57 },
		{ 384, 443, 57 },
		{ 384, 444, 57 },
		{ 384, 445, 57 },
		{ 384, 446, 57 },
		{ 384, 447, 57 },
		{ 384, 448, 57 },
		{ 384, 449, 57 },
		{ 384, 450, 57 },
		{ 384, 451, 57 },
		{ 384, 452, 57 },
		{ 384, 453, 57 },
		{ 384, 454, 57 },
		{ 384, 455, 57 },
		{ 384, 456, 57 },
		{ 384, 457, 57 },
		{ 384, 458, 57 },
		{ 384, 459, 57 },
		{ 384, 460, 57 },
		{ 384, 461, 57 },
		{ 384, 462, 57 },
		{ 384, 463, 57 },
		{ 384, 464, 57 },
		{ 384, 465, 57 },
		{ 384, 466, 57 },
		{ 384, 467, 57 },
		{ 384, 468, 57 },
		{ 384, 469, 57 },
		{ 384, 470, 57 },
		{ 384, 471, 57 },
		{ 384, 472, 57 },
		{ 384, 473, 57 },
		{ 384, 474, 57 },
		{ 384, 475, 57 },
		{ 384, 476, 57 },
		{ 384, 477, 57 },
		{ 384, 478, 57 },
		{ 384, 479, 57 },
		{ 384, 480, 57 },
		{ 384, 481, 57 },
		{ 384, 482, 57 },
		{ 384, 483, 57 },
		{ 384, 484, 57 },
		{ 384, 485, 57 },
		{ 384, 486, 57 },
		{ 384, 487, 57 },
		{ 384, 488, 57 },
		{ 384, 489, 57 },
		{ 384, 490, 57 },
		{ 384, 491, 57 },
		{ 384, 492, 57 },
		{ 384, 493, 57 },
		{ 384, 494, 57 },
		{ 384, 495, 57 },
		{ 384, 496, 57 },
		{ 384, 497, 57 },
		{ 384, 498, 57 },
		{ 384, 499, 57 },
		{ 384, 500, 57 },
		{ 384, 501, 57 },
		{ 384, 502, 57 },
		{ 384, 503, 57 },
		{ 384, 504, 57 },
		{ 384, 505, 57 },
		{ 384, 506, 57 },
		{ 384, 507, 57 },
		{ 384, 508, 57 },
		{ 384, 509, 57 },
		{ 384, 510, 57 },
		{ 385, 428, 57 },
		{ 385, 429, 57 },
		{ 385, 430, 57 },
		{ 385, 431, 57 },
		{ 385, 432, 57 },
		{ 385, 433, 57 },
		{ 385, 434, 57 },
		{ 385, 435, 57 },
		{ 385, 436, 57 },
		{ 385, 437, 57 },
		{ 385, 438, 57 },
		{ 385, 439, 57 },
		{ 385, 440, 57 },
		{ 385, 441, 57 },
		{ 385, 442, 57 },
		{ 385, 443, 57 },
		{ 385, 444, 57 },
		{ 385, 445, 57 },
		{ 385, 446, 57 },
		{ 385, 447, 57 },
		{ 385, 448, 57 },
		{ 385, 449, 57 },
		{ 385, 450, 57 },
		{ 385, 451, 57 },
		{ 385, 452, 57 },
		{ 385, 453, 57 },
		{ 385, 454, 57 },
		{ 385, 455, 57 },
		{ 385, 456, 57 },
		{ 385, 457, 57 },
		{ 385, 458, 57 },
		{ 385, 459, 57 },
		{ 385, 460, 57 },
		{ 385, 461, 57 },
		{ 385, 462, 57 },
		{ 385, 463, 57 },
		{ 385, 464, 57 },
		{ 385, 465, 57 },
		{ 385, 466, 57 },
		{ 385, 467, 57 },
		{ 385, 468, 57 },
		{ 385, 469, 57 },
		{ 385, 470, 57 },
		{ 385, 471, 57 },
		{ 385, 472, 57 },
		{ 385, 473, 57 },
		{ 385, 474, 57 },
		{ 385, 475, 57 },
		{ 385, 476, 57 },
		{ 385, 477, 57 },
		{ 385, 478, 57 },
		{ 385, 479, 57 },
		{ 385, 480, 57 },
		{ 385, 481, 57 },
		{ 385, 482, 57 },
		{ 385, 483, 57 },
		{ 385, 484, 57 },
		{ 385, 485, 57 },
		{ 385, 486, 57 },
		{ 385, 487, 57 },
		{ 385, 488, 57 },
		{ 385, 489, 57 },
		{ 385, 490, 57 },
		{ 385, 491, 57 },
		{ 385, 492, 57 },
		{ 385, 493, 57 },
		{ 385, 494, 57 },
		{ 385, 495, 57 },
		{ 385, 496, 57 },
		{ 385, 497, 57 },
		{ 385, 498, 57 },
		{ 385, 499, 57 },
		{ 385, 500, 57 },
		{ 385, 501, 57 },
		{ 385, 502, 57 },
		{ 385, 503, 57 },
		{ 385, 504, 57 },
		{ 385, 505, 57 },
		{ 385, 506, 57 },
		{ 385, 507, 57 },
		{ 385, 508, 57 },
		{ 385, 509, 57 },
		{ 385, 510, 57 },
		{ 385, 511, 57 },
		{ 385, 512, 57 },
		{ 386, 430, 57 },
		{ 386, 431, 57 },
		{ 386, 432, 57 },
		{ 386, 433, 57 },
		{ 386, 434, 57 },
		{ 386, 435, 57 },
		{ 386, 436, 57 },
		{ 386, 437, 57 },
		{ 386, 438, 57 },
		{ 386, 439, 57 },
		{ 386, 440, 57 },
		{ 386, 441, 57 },
		{ 386, 442, 57 },
		{ 386, 443, 57 },
		{ 386, 444, 57 },
		{ 386, 445, 57 },
		{ 386, 446, 57 },
		{ 386, 447, 57 },
		{ 386, 448, 57 },
		{ 386, 449, 57 },
		{ 386, 450, 57 },
		{ 386, 451, 57 },
		{ 386, 452, 57 },
		{ 386, 453, 57 },
		{ 386, 454, 57 },
		{ 386, 455, 57 },
		{ 386, 456, 57 },
		{ 386, 457, 57 },
		{ 386, 458, 57 },
		{ 386, 459, 57 },
		{ 386, 460, 57 },
		{ 386, 461, 57 },
		{ 386, 462, 57 },
		{ 386, 463, 57 },
		{ 386, 464, 57 },
		{ 386, 465, 57 },
		{ 386, 466, 57 },
		{ 386, 467, 57 },
		{ 386, 468, 57 },
		{ 386, 469, 57 },
		{ 386, 470, 57 },
		{ 386, 471, 57 },
		{ 386, 472, 57 },
		{ 386, 473, 57 },
		{ 386, 474, 57 },
		{ 386, 475, 57 },
		{ 386, 476, 57 },
		{ 386, 477, 57 },
		{ 386, 478, 57 },
		{ 386, 479, 57 },
		{ 386, 480, 57 },
		{ 386, 481, 57 },
		{ 386, 482, 57 },
		{ 386, 483, 57 },
		{ 386, 484, 57 },
		{ 386, 485, 57 },
		{ 386, 486, 57 },
		{ 386, 487, 57 },
		{ 386, 488, 57 },
		{ 386, 489, 57 },
		{ 386, 490, 57 },
		{ 386, 491, 57 },
		{ 386, 492, 57 },
		{ 386, 493, 57 },
		{ 386, 494, 57 },
		{ 386, 495, 57 },
		{ 386, 496, 57 },
		{ 386, 497, 57 },
		{ 386, 498, 57 },
		{ 386, 499, 57 },
		{ 386, 500, 57 },
		{ 386, 501, 57 },
		{ 386, 502, 57 },
		{ 386, 503, 57 },
		{ 386, 504, 57 },
		{ 386, 505, 57 },
		{ 386, 506, 57 },
		{ 386, 507, 57 },
		{ 386, 508, 57 },
		{ 386, 509, 57 },
		{ 386, 510, 57 },
		{ 386, 511, 57 },
		{ 386, 512, 57 },
		{ 387, 431, 57 },
		{ 387, 432, 57 },
		{ 387, 433, 57 },
		{ 387, 434, 57 },
		{ 387, 435, 57 },
		{ 387, 436, 57 },
		{ 387, 437, 57 },
		{ 387, 438, 57 },
		{ 387, 439, 57 },
		{ 387, 440, 57 },
		{ 387, 441, 57 },
		{ 387, 442, 57 },
		{ 387, 443, 57 },
		{ 387, 444, 57 },
		{ 387, 445, 57 },
		{ 387, 446, 57 },
		{ 387, 447, 57 },
		{ 387, 448, 57 },
		{ 387, 449, 57 },
		{ 387, 450, 57 },
		{ 387, 451, 57 },
		{ 387, 452, 57 },
		{ 387, 453, 57 },
		{ 387, 454, 57 },
		{ 387, 455, 57 },
		{ 387, 456, 57 },
		{ 387, 457, 57 },
		{ 387, 458, 57 },
		{ 387, 459, 57 },
		{ 387, 460, 57 },
		{ 387, 461, 57 },
		{ 387, 462, 57 },
		{ 387, 463, 57 },
		{ 387, 464, 57 },
		{ 387, 465, 57 },
		{ 387, 466, 57 },
		{ 387, 467, 57 },
		{ 387, 468, 57 },
		{ 387, 469, 57 },
		{ 387, 470, 57 },
		{ 387, 471, 57 },
		{ 387, 472, 57 },
		{ 387, 473, 57 },
		{ 387, 474, 57 },
		{ 387, 475, 57 },
		{ 387, 476, 57 },
		{ 387, 477, 57 },
		{ 387, 478, 57 },
		{ 387, 479, 57 },
		{ 387, 480, 57 },
		{ 387, 481, 57 },
		{ 387, 482, 57 },
		{ 387, 483, 57 },
		{ 387, 484, 57 },
		{ 387, 485, 57 },
		{ 387, 486, 57 },
		{ 387, 487, 57 },
		{ 387, 488, 57 },
		{ 387, 489, 57 },
		{ 387, 490, 57 },
		{ 387, 491, 57 },
		{ 387, 492, 57 },
		{ 387, 493, 57 },
		{ 387, 494, 57 },
		{ 387, 495, 57 },
		{ 387, 496, 57 },
		{ 387, 497, 57 },
		{ 387, 498, 57 },
		{ 387, 499, 57 },
		{ 387, 500, 57 },
		{ 387, 501, 57 },
		{ 387, 502, 57 },
		{ 387, 503, 57 },
		{ 387, 504, 57 },
		{ 387, 505, 57 },
		{ 387, 506, 57 },
		{ 387, 507, 57 },
		{ 387, 508, 57 },
		{ 387, 509, 57 },
		{ 387, 510, 57 },
		{ 387, 511, 57 },
		{ 387, 512, 57 },
		{ 388, 432, 57 },
		{ 388, 433, 57 },
		{ 388, 434, 57 },
		{ 388, 435, 57 },
		{ 388, 436, 57 },
		{ 388, 437, 57 },
		{ 388, 438, 57 },
		{ 388, 439, 57 },
		{ 388, 440, 57 },
		{ 388, 441, 57 },
		{ 388, 442, 57 },
		{ 388, 443, 57 },
		{ 388, 444, 57 },
		{ 388, 445, 57 },
		{ 388, 446, 57 },
		{ 388, 447, 57 },
		{ 388, 448, 57 },
		{ 388, 449, 57 },
		{ 388, 450, 57 },
		{ 388, 451, 57 },
		{ 388, 452, 57 },
		{ 388, 453, 57 },
		{ 388, 454, 57 },
		{ 388, 455, 57 },
		{ 388, 456, 57 },
		{ 388, 457, 57 },
		{ 388, 458, 57 },
		{ 388, 459, 57 },
		{ 388, 460, 57 },
		{ 388, 461, 57 },
		{ 388, 462, 57 },
		{ 388, 463, 57 },
		{ 388, 464, 57 },
		{ 388, 465, 57 },
		{ 388, 466, 57 },
		{ 388, 467, 57 },
		{ 388, 468, 57 },
		{ 388, 469, 57 },
		{ 388, 470, 57 },
		{ 388, 471, 57 },
		{ 388, 472, 57 },
		{ 388, 473, 57 },
		{ 388, 474, 57 },
		{ 388, 475, 57 },
		{ 388, 476, 57 },
		{ 388, 477, 57 },
		{ 388, 478, 57 },
		{ 388, 479, 57 },
		{ 388, 480, 57 },
		{ 388, 481, 57 },
		{ 388, 482, 57 },
		{ 388, 483, 57 },
		{ 388, 484, 57 },
		{ 388, 485, 57 },
		{ 388, 486, 57 },
		{ 388, 487, 57 },
		{ 388, 488, 57 },
		{ 388, 489, 57 },
		{ 388, 490, 57 },
		{ 388, 491, 57 },
		{ 388, 492, 57 },
		{ 388, 493, 57 },
		{ 388, 494, 57 },
		{ 388, 495, 57 },
		{ 388, 496, 57 },
		{ 388, 497, 57 },
		{ 388, 498, 57 },
		{ 388, 499, 57 },
		{ 388, 500, 57 },
		{ 388, 501, 57 },
		{ 388, 502, 57 },
		{ 388, 503, 57 },
		{ 388, 504, 57 },
		{ 388, 505, 57 },
		{ 388, 506, 57 },
		{ 388, 507, 57 },
		{ 388, 508, 57 },
		{ 388, 509, 57 },
		{ 388, 510, 57 },
		{ 388, 511, 57 },
		{ 388, 512, 57 },
		{ 389, 434, 57 },
		{ 389, 435, 57 },
		{ 389, 436, 57 },
		{ 389, 437, 57 },
		{ 389, 438, 57 },
		{ 389, 439, 57 },
		{ 389, 440, 57 },
		{ 389, 441, 57 },
		{ 389, 442, 57 },
		{ 389, 443, 57 },
		{ 389, 444, 57 },
		{ 389, 445, 57 },
		{ 389, 446, 57 },
		{ 389, 447, 57 },
		{ 389, 448, 57 },
		{ 389, 449, 57 },
		{ 389, 450, 57 },
		{ 389, 451, 57 },
		{ 389, 452, 57 },
		{ 389, 453, 57 },
		{ 389, 454, 57 },
		{ 389, 455, 57 },
		{ 389, 456, 57 },
		{ 389, 457, 57 },
		{ 389, 458, 57 },
		{ 389, 459, 57 },
		{ 389, 460, 57 },
		{ 389, 461, 57 },
		{ 389, 462, 57 },
		{ 389, 463, 57 },
		{ 389, 464, 57 },
		{ 389, 465, 57 },
		{ 389, 466, 57 },
		{ 389, 467, 57 },
		{ 389, 468, 57 },
		{ 389, 469, 57 },
		{ 389, 470, 57 },
		{ 389, 471, 57 },
		{ 389, 472, 57 },
		{ 389, 473, 57 },
		{ 389, 474, 57 },
		{ 389, 475, 57 },
		{ 389, 476, 57 },
		{ 389, 477, 57 },
		{ 389, 478, 57 },
		{ 389, 479, 57 },
		{ 389, 480, 57 },
		{ 389, 481, 57 },
		{ 389, 482, 57 },
		{ 389, 483, 57 },
		{ 389, 484, 57 },
		{ 389, 485, 57 },
		{ 389, 486, 57 },
		{ 389, 487, 57 },
		{ 389, 488, 57 },
		{ 389, 489, 57 },
		{ 389, 490, 57 },
		{ 389, 491, 57 },
		{ 389, 492, 57 },
		{ 389, 493, 57 },
		{ 389, 494, 57 },
		{ 389, 495, 57 },
		{ 389, 496, 57 },
		{ 389, 497, 57 },
		{ 389, 498, 57 },
		{ 389, 499, 57 },
		{ 389, 500, 57 },
		{ 389, 501, 57 },
		{ 389, 502, 57 },
		{ 389, 503, 57 },
		{ 389, 504, 57 },
		{ 389, 505, 57 },
		{ 389, 506, 57 },
		{ 389, 507, 57 },
		{ 389, 508, 57 },
		{ 389, 509, 57 },
		{ 389, 510, 57 },
		{ 389, 511, 57 },
		{ 389, 512, 57 },
		{ 390, 435, 57 },
		{ 390, 436, 57 },
		{ 390, 437, 57 },
		{ 390, 438, 57 },
		{ 390, 439, 57 },
		{ 390, 440, 57 },
		{ 390, 441, 57 },
		{ 390, 442, 57 },
		{ 390, 443, 57 },
		{ 390, 444, 57 },
		{ 390, 445, 57 },
		{ 390, 446, 57 },
		{ 390, 447, 57 },
		{ 390, 448, 57 },
		{ 390, 449, 57 },
		{ 390, 450, 57 },
		{ 390, 451, 57 },
		{ 390, 452, 57 },
		{ 390, 453, 57 },
		{ 390, 454, 57 },
		{ 390, 455, 57 },
		{ 390, 456, 57 },
		{ 390, 457, 57 },
		{ 390, 458, 57 },
		{ 390, 459, 57 },
		{ 390, 460, 57 },
		{ 390, 461, 57 },
		{ 390, 462, 57 },
		{ 390, 463, 57 },
		{ 390, 464, 57 },
		{ 390, 465, 57 },
		{ 390, 466, 57 },
		{ 390, 467, 57 },
		{ 390, 468, 57 },
		{ 390, 469, 57 },
		{ 390, 470, 57 },
		{ 390, 471, 57 },
		{ 390, 472, 57 },
		{ 390, 473, 57 },
		{ 390, 474, 57 },
		{ 390, 475, 57 },
		{ 390, 476, 57 },
		{ 390, 477, 57 },
		{ 390, 478, 57 },
		{ 390, 479, 57 },
		{ 390, 480, 57 },
		{ 390, 481, 57 },
		{ 390, 482, 57 },
		{ 390, 483, 57 },
		{ 390, 484, 57 },
		{ 390, 485, 57 },
		{ 390, 486, 57 },
		{ 390, 487, 57 },
		{ 390, 488, 57 },
		{ 390, 489, 57 },
		{ 390, 490, 57 },
		{ 390, 491, 57 },
		{ 390, 492, 57 },
		{ 390, 493, 57 },
		{ 390, 494, 57 },
		{ 390, 495, 57 },
		{ 390, 496, 57 },
		{ 390, 497, 57 },
		{ 390, 498, 57 },
		{ 390, 499, 57 },
		{ 390, 500, 57 },
		{ 390, 501, 57 },
		{ 390, 502, 57 },
		{ 390, 503, 57 },
		{ 390, 504, 57 },
		{ 390, 505, 57 },
		{ 390, 506, 57 },
		{ 390, 507, 57 },
		{ 390, 508, 57 },
		{ 390, 509, 57 },
		{ 390, 510, 57 },
		{ 390, 511, 57 },
		{ 390, 512, 57 },
		{ 391, 436, 57 },
		{ 391, 437, 57 },
		{ 391, 438, 57 },
		{ 391, 439, 57 },
		{ 391, 440, 57 },
		{ 391, 441, 57 },
		{ 391, 442, 57 },
		{ 391, 443, 57 },
		{ 391, 444, 57 },
		{ 391, 445, 57 },
		{ 391, 446, 57 },
		{ 391, 447, 57 },
		{ 391, 448, 57 },
		{ 391, 449, 57 },
		{ 391, 450, 57 },
		{ 391, 451, 57 },
		{ 391, 452, 57 },
		{ 391, 453, 57 },
		{ 391, 454, 57 },
		{ 391, 455, 57 },
		{ 391, 456, 57 },
		{ 391, 457, 57 },
		{ 391, 458, 57 },
		{ 391, 459, 57 },
		{ 391, 460, 57 },
		{ 391, 461, 57 },
		{ 391, 462, 57 },
		{ 391, 463, 57 },
		{ 391, 464, 57 },
		{ 391, 465, 57 },
		{ 391, 466, 57 },
		{ 391, 467, 57 },
		{ 391, 468, 57 },
		{ 391, 469, 57 },
		{ 391, 470, 57 },
		{ 391, 471, 57 },
		{ 391, 472, 57 },
		{ 391, 473, 57 },
		{ 391, 474, 57 },
		{ 391, 475, 57 },
		{ 391, 476, 57 },
		{ 391, 477, 57 },
		{ 391, 478, 57 },
		{ 391, 479, 57 },
		{ 391, 480, 57 },
		{ 391, 481, 57 },
		{ 391, 482, 57 },
		{ 391, 483, 57 },
		{ 391, 484, 57 },
		{ 391, 485, 57 },
		{ 391, 486, 57 },
		{ 391, 487, 57 },
		{ 391, 488, 57 },
		{ 391, 489, 57 },
		{ 391, 490, 57 },
		{ 391, 491, 57 },
		{ 391, 492, 57 },
		{ 391, 493, 57 },
		{ 391, 494, 57 },
		{ 391, 495, 57 },
		{ 391, 496, 57 },
		{ 391, 497, 57 },
		{ 391, 498, 57 },
		{ 391, 499, 57 },
		{ 391, 500, 57 },
		{ 391, 501, 57 },
		{ 391, 502, 57 },
		{ 391, 503, 57 },
		{ 391, 504, 57 },
		{ 391, 505, 57 },
		{ 391, 506, 57 },
		{ 391, 507, 57 },
		{ 391, 508, 57 },
		{ 391, 509, 57 },
		{ 391, 510, 57 },
		{ 391, 511, 57 },
		{ 391, 512, 57 },
		{ 392, 438, 57 },
		{ 392, 439, 57 },
		{ 392, 440, 57 },
		{ 392, 441, 57 },
		{ 392, 442, 57 },
		{ 392, 443, 57 },
		{ 392, 444, 57 },
		{ 392, 445, 57 },
		{ 392, 446, 57 },
		{ 392, 447, 57 },
		{ 392, 448, 57 },
		{ 392, 449, 57 },
		{ 392, 450, 57 },
		{ 392, 451, 57 },
		{ 392, 452, 57 },
		{ 392, 453, 57 },
		{ 392, 454, 57 },
		{ 392, 455, 57 },
		{ 392, 456, 57 },
		{ 392, 457, 57 },
		{ 392, 458, 57 },
		{ 392, 459, 57 },
		{ 392, 460, 57 },
		{ 392, 461, 57 },
		{ 392, 462, 57 },
		{ 392, 463, 57 },
		{ 392, 464, 57 },
		{ 392, 465, 57 },
		{ 392, 466, 57 },
		{ 392, 467, 57 },
		{ 392, 468, 57 },
		{ 392, 469, 57 },
		{ 392, 470, 57 },
		{ 392, 471, 57 },
		{ 392, 472, 57 },
		{ 392, 473, 57 },
		{ 392, 474, 57 },
		{ 392, 475, 57 },
		{ 392, 476, 57 },
		{ 392, 477, 57 },
		{ 392, 478, 57 },
		{ 392, 479, 57 },
		{ 392, 480, 57 },
		{ 392, 481, 57 },
		{ 392, 482, 57 },
		{ 392, 483, 57 },
		{ 392, 484, 57 },
		{ 392, 485, 57 },
		{ 392, 486, 57 },
		{ 392, 487, 57 },
		{ 392, 488, 57 },
		{ 392, 489, 57 },
		{ 392, 490, 57 },
		{ 392, 491, 57 },
		{ 392, 492, 57 },
		{ 392, 493, 57 },
		{ 392, 494, 57 },
		{ 392, 495, 57 },
		{ 392, 496, 57 },
		{ 392, 497, 57 },
		{ 392, 498, 57 },
		{ 392, 499, 57 },
		{ 392, 500, 57 },
		{ 392, 501, 57 },
		{ 392, 502, 57 },
		{ 392, 503, 57 },
		{ 392, 504, 57 },
		{ 392, 505, 57 },
		{ 392, 506, 57 },
		{ 392, 507, 57 },
		{ 392, 508, 57 },
		{ 392, 509, 57 },
		{ 392, 510, 57 },
		{ 392, 511, 57 },
		{ 392, 512, 57 },
		{ 393, 439, 57 },
		{ 393, 440, 57 },
		{ 393, 441, 57 },
		{ 393, 442, 57 },
		{ 393, 443, 57 },
		{ 393, 444, 57 },
		{ 393, 445, 57 },
		{ 393, 446, 57 },
		{ 393, 447, 57 },
		{ 393, 448, 57 },
		{ 393, 449, 57 },
		{ 393, 450, 57 },
		{ 393, 451, 57 },
		{ 393, 452, 57 },
		{ 393, 453, 57 },
		{ 393, 454, 57 },
		{ 393, 455, 57 },
		{ 393, 456, 57 },
		{ 393, 457, 57 },
		{ 393, 458, 57 },
		{ 393, 459, 57 },
		{ 393, 460, 57 },
		{ 393, 461, 57 },
		{ 393, 462, 57 },
		{ 393, 463, 57 },
		{ 393, 464, 57 },
		{ 393, 465, 57 },
		{ 393, 466, 57 },
		{ 393, 467, 57 },
		{ 393, 468, 57 },
		{ 393, 469, 57 },
		{ 393, 470, 57 },
		{ 393, 471, 57 },
		{ 393, 472, 57 },
		{ 393, 473, 57 },
		{ 393, 474, 57 },
		{ 393, 475, 57 },
		{ 393, 476, 57 },
		{ 393, 477, 57 },
		{ 393, 478, 57 },
		{ 393, 479, 57 },
		{ 393, 480, 57 },
		{ 393, 481, 57 },
		{ 393, 482, 57 },
		{ 393, 483, 57 },
		{ 393, 484, 57 },
		{ 393, 485, 57 },
		{ 393, 486, 57 },
		{ 393, 487, 57 },
		{ 393, 488, 57 },
		{ 393, 489, 57 },
		{ 393, 490, 57 },
		{ 393, 491, 57 },
		{ 393, 492, 57 },
		{ 393, 493, 57 },
		{ 393, 494, 57 },
		{ 393, 495, 57 },
		{ 393, 496, 57 },
		{ 393, 497, 57 },
		{ 393, 498, 57 },
		{ 393, 499, 57 },
		{ 393, 500, 57 },
		{ 393, 501, 57 },
		{ 393, 502, 57 },
		{ 393, 503, 57 },
		{ 393, 504, 57 },
		{ 393, 505, 57 },
		{ 393, 506, 57 },
		{ 393, 507, 57 },
		{ 393, 508, 57 },
		{ 393, 509, 57 },
		{ 393, 510, 57 },
		{ 393, 511, 57 },
		{ 393, 512, 57 },
		{ 394, 440, 57 },
		{ 394, 441, 57 },
		{ 394, 442, 57 },
		{ 394, 443, 57 },
		{ 394, 444, 57 },
		{ 394, 445, 57 },
		{ 394, 446, 57 },
		{ 394, 447, 57 },
		{ 394, 448, 57 },
		{ 394, 449, 57 },
		{ 394, 450, 57 },
		{ 394, 451, 57 },
		{ 394, 452, 57 },
		{ 394, 453, 57 },
		{ 394, 454, 57 },
		{ 394, 455, 57 },
		{ 394, 456, 57 },
		{ 394, 457, 57 },
		{ 394, 458, 57 },
		{ 394, 459, 57 },
		{ 394, 460, 57 },
		{ 394, 461, 57 },
		{ 394, 462, 57 },
		{ 394, 463, 57 },
		{ 394, 464, 57 },
		{ 394, 465, 57 },
		{ 394, 466, 57 },
		{ 394, 467, 57 },
		{ 394, 468, 57 },
		{ 394, 469, 57 },
		{ 394, 470, 57 },
		{ 394, 471, 57 },
		{ 394, 472, 57 },
		{ 394, 473, 57 },
		{ 394, 474, 57 },
		{ 394, 475, 57 },
		{ 394, 476, 57 },
		{ 394, 477, 57 },
		{ 394, 478, 57 },
		{ 394, 479, 57 },
		{ 394, 480, 57 },
		{ 394, 481, 57 },
		{ 394, 482, 57 },
		{ 394, 483, 57 },
		{ 394, 484, 57 },
		{ 394, 485, 57 },
		{ 394, 486, 57 },
		{ 394, 487, 57 },
		{ 394, 488, 57 },
		{ 394, 489, 57 },
		{ 394, 490, 57 },
		{ 394, 491, 57 },
		{ 394, 492, 57 },
		{ 394, 493, 57 },
		{ 394, 494, 57 },
		{ 394, 495, 57 },
		{ 394, 496, 57 },
		{ 394, 497, 57 },
		{ 394, 498, 57 },
		{ 394, 499, 57 },
		{ 394, 500, 57 },
		{ 394, 501, 57 },
		{ 394, 502, 57 },
		{ 394, 503, 57 },
		{ 394, 504, 57 },
		{ 394, 505, 57 },
		{ 394, 506, 57 },
		{ 394, 507, 57 },
		{ 394, 508, 57 },
		{ 394, 509, 57 },
		{ 394, 510, 57 },
		{ 394, 511, 57 },
		{ 394, 512, 57 },
		{ 395, 442, 57 },
		{ 395, 443, 57 },
		{ 395, 444, 57 },
		{ 395, 445, 57 },
		{ 395, 446, 57 },
		{ 395, 447, 57 },
		{ 395, 448, 57 },
		{ 395, 449, 57 },
		{ 395, 450, 57 },
		{ 395, 451, 57 },
		{ 395, 452, 57 },
		{ 395, 453, 57 },
		{ 395, 454, 57 },
		{ 395, 455, 57 },
		{ 395, 456, 57 },
		{ 395, 457, 57 },
		{ 395, 458, 57 },
		{ 395, 459, 57 },
		{ 395, 460, 57 },
		{ 395, 461, 57 },
		{ 395, 462, 57 },
		{ 395, 463, 57 },
		{ 395, 464, 57 },
		{ 395, 465, 57 },
		{ 395, 466, 57 },
		{ 395, 467, 57 },
		{ 395, 468, 57 },
		{ 395, 469, 57 },
		{ 395, 470, 57 },
		{ 395, 471, 57 },
		{ 395, 472, 57 },
		{ 395, 473, 57 },
		{ 395, 474, 57 },
		{ 395, 475, 57 },
		{ 395, 476, 57 },
		{ 395, 477, 57 },
		{ 395, 478, 57 },
		{ 395, 479, 57 },
		{ 395, 480, 57 },
		{ 395, 481, 57 },
		{ 395, 482, 57 },
		{ 395, 483, 57 },
		{ 395, 484, 57 },
		{ 395, 485, 57 },
		{ 395, 486, 57 },
		{ 395, 487, 57 },
		{ 395, 488, 57 },
		{ 395, 489, 57 },
		{ 395, 490, 57 },
		{ 395, 491, 57 },
		{ 395, 492, 57 },
		{ 395, 493, 57 },
		{ 395, 494, 57 },
		{ 395, 495, 57 },
		{ 395, 496, 57 },
		{ 395, 497, 57 },
		{ 395, 498, 57 },
		{ 395, 499, 57 },
		{ 395, 500, 57 },
		{ 395, 501, 57 },
		{ 395, 502, 57 },
		{ 395, 503, 57 },
		{ 395, 504, 57 },
		{ 395, 505, 57 },
		{ 395, 506, 57 },
		{ 395, 507, 57 },
		{ 395, 508, 57 },
		{ 395, 509, 57 },
		{ 395, 510, 57 },
		{ 395, 511, 57 },
		{ 395, 512, 57 },
		{ 396, 443, 57 },
		{ 396, 444, 57 },
		{ 396, 445, 57 },
		{ 396, 446, 57 },
		{ 396, 447, 57 },
		{ 396, 448, 57 },
		{ 396, 449, 57 },
		{ 396, 450, 57 },
		{ 396, 451, 57 },
		{ 396, 452, 57 },
		{ 396, 453, 57 },
		{ 396, 454, 57 },
		{ 396, 455, 57 },
		{ 396, 456, 57 },
		{ 396, 457, 57 },
		{ 396, 458, 57 },
		{ 396, 459, 57 },
		{ 396, 460, 57 },
		{ 396, 461, 57 },
		{ 396, 462, 57 },
		{ 396, 463, 57 },
		{ 396, 464, 57 },
		{ 396, 465, 57 },
		{ 396, 466, 57 },
		{ 396, 467, 57 },
		{ 396, 468, 57 },
		{ 396, 469, 57 },
		{ 396, 470, 57 },
		{ 396, 471, 57 },
		{ 396, 472, 57 },
		{ 396, 473, 57 },
		{ 396, 474, 57 },
		{ 396, 475, 57 },
		{ 396, 476, 57 },
		{ 396, 477, 57 },
		{ 396, 478, 57 },
		{ 396, 479, 57 },
		{ 396, 480, 57 },
		{ 396, 481, 57 },
		{ 396, 482, 57 },
		{ 396, 483, 57 },
		{ 396, 484, 57 },
		{ 396, 485, 57 },
		{ 396, 486, 57 },
		{ 396, 487, 57 },
		{ 396, 488, 57 },
		{ 396, 489, 57 },
		{ 396, 490, 57 },
		{ 396, 491, 57 },
		{ 396, 492, 57 },
		{ 396, 493, 57 },
		{ 396, 494, 57 },
		{ 396, 495, 57 },
		{ 396, 496, 57 },
		{ 396, 497, 57 },
		{ 396, 498, 57 },
		{ 396, 499, 57 },
		{ 396, 500, 57 },
		{ 396, 501, 57 },
		{ 396, 502, 57 },
		{ 396, 503, 57 },
		{ 396, 504, 57 },
		{ 396, 505, 57 },
		{ 396, 506, 57 },
		{ 396, 507, 57 },
		{ 396, 508, 57 },
		{ 396, 509, 57 },
		{ 396, 510, 57 },
		{ 396, 511, 57 },
		{ 396, 512, 57 },
		{ 397, 444, 57 },
		{ 397, 445, 57 },
		{ 397, 446, 57 },
		{ 397, 447, 57 },
		{ 397, 448, 57 },
		{ 397, 449, 57 },
		{ 397, 450, 57 },
		{ 397, 451, 57 },
		{ 397, 452, 57 },
		{ 397, 453, 57 },
		{ 397, 454, 57 },
		{ 397, 455, 57 },
		{ 397, 456, 57 },
		{ 397, 457, 57 },
		{ 397, 458, 57 },
		{ 397, 459, 57 },
		{ 397, 460, 57 },
		{ 397, 461, 57 },
		{ 397, 462, 57 },
		{ 397, 463, 57 },
		{ 397, 464, 57 },
		{ 397, 465, 57 },
		{ 397, 466, 57 },
		{ 397, 467, 57 },
		{ 397, 468, 57 },
		{ 397, 469, 57 },
		{ 397, 470, 57 },
		{ 397, 471, 57 },
		{ 397, 472, 57 },
		{ 397, 473, 57 },
		{ 397, 474, 57 },
		{ 397, 475, 57 },
		{ 397, 476, 57 },
		{ 397, 477, 57 },
		{ 397, 478, 57 },
		{ 397, 479, 57 },
		{ 397, 480, 57 },
		{ 397, 481, 57 },
		{ 397, 482, 57 },
		{ 397, 483, 57 },
		{ 397, 484, 57 },
		{ 397, 485, 57 },
		{ 397, 486, 57 },
		{ 397, 487, 57 },
		{ 397, 488, 57 },
		{ 397, 489, 57 },
		{ 397, 490, 57 },
		{ 397, 491, 57 },
		{ 397, 492, 57 },
		{ 397, 493, 57 },
		{ 397, 494, 57 },
		{ 397, 495, 57 },
		{ 397, 496, 57 },
		{ 397, 497, 57 },
		{ 397, 498, 57 },
		{ 397, 499, 57 },
		{ 397, 500, 57 },
		{ 397, 501, 57 },
		{ 397, 502, 57 },
		{ 397, 503, 57 },
		{ 397, 504, 57 },
		{ 397, 505, 57 },
		{ 397, 506, 57 },
		{ 397, 507, 57 },
		{ 397, 508, 57 },
		{ 397, 509, 57 },
		{ 397, 510, 57 },
		{ 397, 511, 57 },
		{ 397, 512, 57 },
		{ 398, 446, 57 },
		{ 398, 447, 57 },
		{ 398, 448, 57 },
		{ 398, 449, 57 },
		{ 398, 450, 57 },
		{ 398, 451, 57 },
		{ 398, 452, 57 },
		{ 398, 453, 57 },
		{ 398, 454, 57 },
		{ 398, 455, 57 },
		{ 398, 456, 57 },
		{ 398, 457, 57 },
		{ 398, 458, 57 },
		{ 398, 459, 57 },
		{ 398, 460, 57 },
		{ 398, 461, 57 },
		{ 398, 462, 57 },
		{ 398, 463, 57 },
		{ 398, 464, 57 },
		{ 398, 465, 57 },
		{ 398, 466, 57 },
		{ 398, 467, 57 },
		{ 398, 468, 57 },
		{ 398, 469, 57 },
		{ 398, 470, 57 },
		{ 398, 471, 57 },
		{ 398, 472, 57 },
		{ 398, 473, 57 },
		{ 398, 474, 57 },
		{ 398, 475, 57 },
		{ 398, 476, 57 },
		{ 398, 477, 57 },
		{ 398, 478, 57 },
		{ 398, 479, 57 },
		{ 398, 480, 57 },
		{ 398, 481, 57 },
		{ 398, 482, 57 },
		{ 398, 483, 57 },
		{ 398, 484, 57 },
		{ 398, 485, 57 },
		{ 398, 486, 57 },
		{ 398, 487, 57 },
		{ 398, 488, 57 },
		{ 398, 489, 57 },
		{ 398, 490, 57 },
		{ 398, 491, 57 },
		{ 398, 492, 57 },
		{ 398, 493, 57 },
		{ 398, 494, 57 },
		{ 398, 495, 57 },
		{ 398, 496, 57 },
		{ 398, 497, 57 },
		{ 398, 498, 57 },
		{ 398, 499, 57 },
		{ 398, 500, 57 },
		{ 398, 501, 57 },
		{ 398, 502, 57 },
		{ 398, 503, 57 },
		{ 398, 504, 57 },
		{ 398, 505, 57 },
		{ 398, 506, 57 },
		{ 398, 507, 57 },
		{ 398, 508, 57 },
		{ 398, 509, 57 },
		{ 398, 510, 57 },
		{ 398, 511, 57 },
		{ 398, 512, 57 },
		{ 399, 447, 57 },
		{ 399, 448, 57 },
		{ 399, 449, 57 },
		{ 399, 450, 57 },
		{ 399, 451, 57 },
		{ 399, 452, 57 },
		{ 399, 453, 57 },
		{ 399, 454, 57 },
		{ 399, 455, 57 },
		{ 399, 456, 57 },
		{ 399, 457, 57 },
		{ 399, 458, 57 },
		{ 399, 459, 57 },
		{ 399, 460, 57 },
		{ 399, 461, 57 },
		{ 399, 462, 57 },
		{ 399, 463, 57 },
		{ 399, 464, 57 },
		{ 399, 465, 57 },
		{ 399, 466, 57 },
		{ 399, 467, 57 },
		{ 399, 468, 57 },
		{ 399, 469, 57 },
		{ 399, 470, 57 },
		{ 399, 471, 57 },
		{ 399, 472, 57 },
		{ 399, 473, 57 },
		{ 399, 474, 57 },
		{ 399, 475, 57 },
		{ 399, 476, 57 },
		{ 399, 477, 57 },
		{ 399, 478, 57 },
		{ 399, 479, 57 },
		{ 399, 480, 57 },
		{ 399, 481, 57 },
		{ 399, 482, 57 },
		{ 399, 483, 57 },
		{ 399, 484, 57 },
		{ 399, 485, 57 },
		{ 399, 486, 57 },
		{ 399, 487, 57 },
		{ 399, 488, 57 },
		{ 399, 489, 57 },
		{ 399, 490, 57 },
		{ 399, 491, 57 },
		{ 399, 492, 57 },
		{ 399, 493, 57 },
		{ 399, 494, 57 },
		{ 399, 495, 57 },
		{ 399, 496, 57 },
		{ 399, 497, 57 },
		{ 399, 498, 57 },
		{ 399, 499, 57 },
		{ 399, 500, 57 },
		{ 399, 501, 57 },
		{ 399, 502, 57 },
		{ 399, 503, 57 },
		{ 399, 504, 57 },
		{ 399, 505, 57 },
		{ 399, 506, 57 },
		{ 399, 507, 57 },
		{ 399, 508, 57 },
		{ 399, 509, 57 },
		{ 399, 510, 57 },
		{ 399, 511, 57 },
		{ 399, 512, 57 },
		{ 400, 448, 57 },
		{ 400, 449, 57 },
		{ 400, 450, 57 },
		{ 400, 451, 57 },
		{ 400, 452, 57 },
		{ 400, 453, 57 },
		{ 400, 454, 57 },
		{ 400, 455, 57 },
		{ 400, 456, 57 },
		{ 400, 457, 57 },
		{ 400, 458, 57 },
		{ 400, 459, 57 },
		{ 400, 460, 57 },
		{ 400, 461, 57 },
		{ 400, 462, 57 },
		{ 400, 463, 57 },
		{ 400, 464, 57 },
		{ 400, 465, 57 },
		{ 400, 466, 57 },
		{ 400, 467, 57 },
		{ 400, 468, 57 },
		{ 400, 469, 57 },
		{ 400, 470, 57 },
		{ 400, 471, 57 },
		{ 400, 472, 57 },
		{ 400, 473, 57 },
		{ 400, 474, 57 },
		{ 400, 475, 57 },
		{ 400, 476, 57 },
		{ 400, 477, 57 },
		{ 400, 478, 57 },
		{ 400, 479, 57 },
		{ 400, 480, 57 },
		{ 400, 481, 57 },
		{ 400, 482, 57 },
		{ 400, 483, 57 },
		{ 400, 484, 57 },
		{ 400, 485, 57 },
		{ 400, 486, 57 },
		{ 400, 487, 57 },
		{ 400, 488, 57 },
		{ 400, 489, 57 },
		{ 400, 490, 57 },
		{ 400, 491, 57 },
		{ 400, 492, 57 },
		{ 400, 493, 57 },
		{ 400, 494, 57 },
		{ 400, 495, 57 },
		{ 400, 496, 57 },
		{ 400, 497, 57 },
		{ 400, 498, 57 },
		{ 400, 499, 57 },
		{ 400, 500, 57 },
		{ 400, 501, 57 },
		{ 400, 502, 57 },
		{ 400, 503, 57 },
		{ 400, 504, 57 },
		{ 400, 505, 57 },
		{ 400, 506, 57 },
		{ 400, 507, 57 },
		{ 400, 508, 57 },
		{ 400, 509, 57 },
		{ 400, 510, 57 },
		{ 400, 511, 57 },
		{ 400, 512, 57 },
		{ 401, 450, 57 },
		{ 401, 451, 57 },
		{ 401, 452, 57 },
		{ 401, 453, 57 },
		{ 401, 454, 57 },
		{ 401, 455, 57 },
		{ 401, 456, 57 },
		{ 401, 457, 57 },
		{ 401, 458, 57 },
		{ 401, 459, 57 },
		{ 401, 460, 57 },
		{ 401, 461, 57 },
		{ 401, 462, 57 },
		{ 401, 463, 57 },
		{ 401, 464, 57 },
		{ 401, 465, 57 },
		{ 401, 466, 57 },
		{ 401, 467, 57 },
		{ 401, 468, 57 },
		{ 401, 469, 57 },
		{ 401, 470, 57 },
		{ 401, 471, 57 },
		{ 401, 472, 57 },
		{ 401, 473, 57 },
		{ 401, 474, 57 },
		{ 401, 475, 57 },
		{ 401, 476, 57 },
		{ 401, 477, 57 },
		{ 401, 478, 57 },
		{ 401, 479, 57 },
		{ 401, 480, 57 },
		{ 401, 481, 57 },
		{ 401, 482, 57 },
		{ 401, 483, 57 },
		{ 401, 484, 57 },
		{ 401, 485, 57 },
		{ 401, 486, 57 },
		{ 401, 487, 57 },
		{ 401, 488, 57 },
		{ 401, 489, 57 },
		{ 401, 490, 57 },
		{ 401, 491, 57 },
		{ 401, 492, 57 },
		{ 401, 493, 57 },
		{ 401, 494, 57 },
		{ 401, 495, 57 },
		{ 401, 496, 57 },
		{ 401, 497, 57 },
		{ 401, 498, 57 },
		{ 401, 499, 57 },
		{ 401, 500, 57 },
		{ 401, 501, 57 },
		{ 401, 502, 57 },
		{ 401, 503, 57 },
		{ 401, 504, 57 },
		{ 401, 505, 57 },
		{ 401, 506, 57 },
		{ 401, 507, 57 },
		{ 401, 508, 57 },
		{ 401, 509, 57 },
		{ 401, 510, 57 },
		{ 401, 511, 57 },
		{ 401, 512, 57 },
		{ 402, 451, 57 },
		{ 402, 452, 57 },
		{ 402, 453, 57 },
		{ 402, 454, 57 },
		{ 402, 455, 57 },
		{ 402, 456, 57 },
		{ 402, 457, 57 },
		{ 402, 458, 57 },
		{ 402, 459, 57 },
		{ 402, 460, 57 },
		{ 402, 461, 57 },
		{ 402, 462, 57 },
		{ 402, 463, 57 },
		{ 402, 464, 57 },
		{ 402, 465, 57 },
		{ 402, 466, 57 },
		{ 402, 467, 57 },
		{ 402, 468, 57 },
		{ 402, 469, 57 },
		{ 402, 470, 57 },
		{ 402, 471, 57 },
		{ 402, 472, 57 },
		{ 402, 473, 57 },
		{ 402, 474, 57 },
		{ 402, 475, 57 },
		{ 402, 476, 57 },
		{ 402, 477, 57 },
		{ 402, 478, 57 },
		{ 402, 479, 57 },
		{ 402, 480, 57 },
		{ 402, 481, 57 },
		{ 402, 482, 57 },
		{ 402, 483, 57 },
		{ 402, 484, 57 },
		{ 402, 485, 57 },
		{ 402, 486, 57 },
		{ 402, 487, 57 },
		{ 402, 488, 57 },
		{ 402, 489, 57 },
		{ 402, 490, 57 },
		{ 402, 491, 57 },
		{ 402, 492, 57 },
		{ 402, 493, 57 },
		{ 402, 494, 57 },
		{ 402, 495, 57 },
		{ 402, 496, 57 },
		{ 402, 497, 57 },
		{ 402, 498, 57 },
		{ 402, 499, 57 },
		{ 402, 500, 57 },
		{ 402, 501, 57 },
		{ 402, 502, 57 },
		{ 402, 503, 57 },
		{ 402, 504, 57 },
		{ 402, 505, 57 },
		{ 402, 506, 57 },
		{ 402, 507, 57 },
		{ 402, 508, 57 },
		{ 402, 509, 57 },
		{ 402, 510, 57 },
		{ 402, 511, 57 },
		{ 402, 512, 57 },
		{ 403, 452, 57 },
		{ 403, 453, 57 },
		{ 403, 454, 57 },
		{ 403, 455, 57 },
		{ 403, 456, 57 },
		{ 403, 457, 57 },
		{ 403, 458, 57 },
		{ 403, 459, 57 },
		{ 403, 460, 57 },
		{ 403, 461, 57 },
		{ 403, 462, 57 },
		{ 403, 463, 57 },
		{ 403, 464, 57 },
		{ 403, 465, 57 },
		{ 403, 466, 57 },
		{ 403, 467, 57 },
		{ 403, 468, 57 },
		{ 403, 469, 57 },
		{ 403, 470, 57 },
		{ 403, 471, 57 },
		{ 403, 472, 57 },
		{ 403, 473, 57 },
		{ 403, 474, 57 },
		{ 403, 475, 57 },
		{ 403, 476, 57 },
		{ 403, 477, 57 },
		{ 403, 478, 57 },
		{ 403, 479, 57 },
		{ 403, 480, 57 },
		{ 403, 481, 57 },
		{ 403, 482, 57 },
		{ 403, 483, 57 },
		{ 403, 484, 57 },
		{ 403, 485, 57 },
		{ 403, 486, 57 },
		{ 403, 487, 57 },
		{ 403, 488, 57 },
		{ 403, 489, 57 },
		{ 403, 490, 57 },
		{ 403, 491, 57 },
		{ 403, 492, 57 },
		{ 403, 493, 57 },
		{ 403, 494, 57 },
		{ 403, 495, 57 },
		{ 403, 496, 57 },
		{ 403, 497, 57 },
		{ 403, 498, 57 },
		{ 403, 499, 57 },
		{ 403, 500, 57 },
		{ 403, 501, 57 },
		{ 403, 502, 57 },
		{ 403, 503, 57 },
		{ 403, 504, 57 },
		{ 403, 505, 57 },
		{ 403, 506, 57 },
		{ 403, 507, 57 },
		{ 403, 508, 57 },
		{ 403, 509, 57 },
		{ 403, 510, 57 },
		{ 403, 511, 57 },
		{ 403, 512, 57 },
		{ 404, 454, 57 },
		{ 404, 455, 57 },
		{ 404, 456, 57 },
		{ 404, 457, 57 },
		{ 404, 458, 57 },
		{ 404, 459, 57 },
		{ 404, 460, 57 },
		{ 404, 461, 57 },
		{ 404, 462, 57 },
		{ 404, 463, 57 },
		{ 404, 464, 57 },
		{ 404, 465, 57 },
		{ 404, 466, 57 },
		{ 404, 467, 57 },
		{ 404, 468, 57 },
		{ 404, 469, 57 },
		{ 404, 470, 57 },
		{ 404, 471, 57 },
		{ 404, 472, 57 },
		{ 404, 473, 57 },
		{ 404, 474, 57 },
		{ 404, 475, 57 },
		{ 404, 476, 57 },
		{ 404, 477, 57 },
		{ 404, 478, 57 },
		{ 404, 479, 57 },
		{ 404, 480, 57 },
		{ 404, 481, 57 },
		{ 404, 482, 57 },
		{ 404, 483, 57 },
		{ 404, 484, 57 },
		{ 404, 485, 57 },
		{ 404, 486, 57 },
		{ 404, 487, 57 },
		{ 404, 488, 57 },
		{ 404, 489, 57 },
		{ 404, 490, 57 },
		{ 404, 491, 57 },
		{ 404, 492, 57 },
		{ 404, 493, 57 },
		{ 404, 494, 57 },
		{ 404, 495, 57 },
		{ 404, 496, 57 },
		{ 404, 497, 57 },
		{ 404, 498, 57 },
		{ 404, 499, 57 },
		{ 404, 500, 57 },
		{ 404, 501, 57 },
		{ 404, 502, 57 },
		{ 404, 503, 57 },
		{ 404, 504, 57 },
		{ 404, 505, 57 },
		{ 404, 506, 57 },
		{ 404, 507, 57 },
		{ 404, 508, 57 },
		{ 404, 509, 57 },
		{ 404, 510, 57 },
		{ 404, 511, 57 },
		{ 404, 512, 57 },
		{ 405, 455, 57 },
		{ 405, 456, 57 },
		{ 405, 457, 57 },
		{ 405, 458, 57 },
		{ 405, 459, 57 },
		{ 405, 460, 57 },
		{ 405, 461, 57 },
		{ 405, 462, 57 },
		{ 405, 463, 57 },
		{ 405, 464, 57 },
		{ 405, 465, 57 },
		{ 405, 466, 57 },
		{ 405, 467, 57 },
		{ 405, 468, 57 },
		{ 405, 469, 57 },
		{ 405, 470, 57 },
		{ 405, 471, 57 },
		{ 405, 472, 57 },
		{ 405, 473, 57 },
		{ 405, 474, 57 },
		{ 405, 475, 57 },
		{ 405, 476, 57 },
		{ 405, 477, 57 },
		{ 405, 478, 57 },
		{ 405, 479, 57 },
		{ 405, 480, 57 },
		{ 405, 481, 57 },
		{ 405, 482, 57 },
		{ 405, 483, 57 },
		{ 405, 484, 57 },
		{ 405, 485, 57 },
		{ 405, 486, 57 },
		{ 405, 487, 57 },
		{ 405, 488, 57 },
		{ 405, 489, 57 },
		{ 405, 490, 57 },
		{ 405, 491, 57 },
		{ 405, 492, 57 },
		{ 405, 493, 57 },
		{ 405, 494, 57 },
		{ 405, 495, 57 },
		{ 405, 496, 57 },
		{ 405, 497, 57 },
		{ 405, 498, 57 },
		{ 405, 499, 57 },
		{ 405, 500, 57 },
		{ 405, 501, 57 },
		{ 405, 502, 57 },
		{ 405, 503, 57 },
		{ 405, 504, 57 },
		{ 405, 505, 57 },
		{ 405, 506, 57 },
		{ 405, 507, 57 },
		{ 405, 508, 57 },
		{ 405, 509, 57 },
		{ 405, 510, 57 },
		{ 405, 511, 57 },
		{ 405, 512, 57 },
		{ 406, 456, 57 },
		{ 406, 457, 57 },
		{ 406, 458, 57 },
		{ 406, 459, 57 },
		{ 406, 460, 57 },
		{ 406, 461, 57 },
		{ 406, 462, 57 },
		{ 406, 463, 57 },
		{ 406, 464, 57 },
		{ 406, 465, 57 },
		{ 406, 466, 57 },
		{ 406, 467, 57 },
		{ 406, 468, 57 },
		{ 406, 469, 57 },
		{ 406, 470, 57 },
		{ 406, 471, 57 },
		{ 406, 472, 57 },
		{ 406, 473, 57 },
		{ 406, 474, 57 },
		{ 406, 475, 57 },
		{ 406, 476, 57 },
		{ 406, 477, 57 },
		{ 406, 478, 57 },
		{ 406, 479, 57 },
		{ 406, 480, 57 },
		{ 406, 481, 57 },
		{ 406, 482, 57 },
		{ 406, 483, 57 },
		{ 406, 484, 57 },
		{ 406, 485, 57 },
		{ 406, 486, 57 },
		{ 406, 487, 57 },
		{ 406, 488, 57 },
		{ 406, 489, 57 },
		{ 406, 490, 57 },
		{ 406, 491, 57 },
		{ 406, 492, 57 },
		{ 406, 493, 57 },
		{ 406, 494, 57 },
		{ 406, 495, 57 },
		{ 406, 496, 57 },
		{ 406, 497, 57 },
		{ 406, 498, 57 },
		{ 406, 499, 57 },
		{ 406, 500, 57 },
		{ 406, 501, 57 },
		{ 406, 502, 57 },
		{ 406, 503, 57 },
		{ 406, 504, 57 },
		{ 406, 505, 57 },
		{ 406, 506, 57 },
		{ 406, 507, 57 },
		{ 406, 508, 57 },
		{ 406, 509, 57 },
		{ 406, 510, 57 },
		{ 406, 511, 57 },
		{ 406, 512, 57 },
		{ 407, 458, 57 },
		{ 407, 459, 57 },
		{ 407, 460, 57 },
		{ 407, 461, 57 },
		{ 407, 462, 57 },
		{ 407, 463, 57 },
		{ 407, 464, 57 },
		{ 407, 465, 57 },
		{ 407, 466, 57 },
		{ 407, 467, 57 },
		{ 407, 468, 57 },
		{ 407, 469, 57 },
		{ 407, 470, 57 },
		{ 407, 471, 57 },
		{ 407, 472, 57 },
		{ 407, 473, 57 },
		{ 407, 474, 57 },
		{ 407, 475, 57 },
		{ 407, 476, 57 },
		{ 407, 477, 57 },
		{ 407, 478, 57 },
		{ 407, 479, 57 },
		{ 407, 480, 57 },
		{ 407, 481, 57 },
		{ 407, 482, 57 },
		{ 407, 483, 57 },
		{ 407, 484, 57 },
		{ 407, 485, 57 },
		{ 407, 486, 57 },
		{ 407, 487, 57 },
		{ 407, 488, 57 },
		{ 407, 489, 57 },
		{ 407, 490, 57 },
		{ 407, 491, 57 },
		{ 407, 492, 57 },
		{ 407, 493, 57 },
		{ 407, 494, 57 },
		{ 407, 495, 57 },
		{ 407, 496, 57 },
		{ 407, 497, 57 },
		{ 407, 498, 57 },
		{ 407, 499, 57 },
		{ 407, 500, 57 },
		{ 407, 501, 57 },
		{ 407, 502, 57 },
		{ 407, 503, 57 },
		{ 407, 504, 57 },
		{ 407, 505, 57 },
		{ 407, 506, 57 },
		{ 407, 507, 57 },
		{ 407, 508, 57 },
		{ 407, 509, 57 },
		{ 407, 510, 57 },
		{ 407, 511, 57 },
		{ 407, 512, 57 },
		{ 408, 459, 57 },
		{ 408, 460, 57 },
		{ 408, 461, 57 },
		{ 408, 462, 57 },
		{ 408, 463, 57 },
		{ 408, 464, 57 },
		{ 408, 465, 57 },
		{ 408, 466, 57 },
		{ 408, 467, 57 },
		{ 408, 468, 57 },
		{ 408, 469, 57 },
		{ 408, 470, 57 },
		{ 408, 471, 57 },
		{ 408, 472, 57 },
		{ 408, 473, 57 },
		{ 408, 474, 57 },
		{ 408, 475, 57 },
		{ 408, 476, 57 },
		{ 408, 477, 57 },
		{ 408, 478, 57 },
		{ 408, 479, 57 },
		{ 408, 480, 57 },
		{ 408, 481, 57 },
		{ 408, 482, 57 },
		{ 408, 483, 57 },
		{ 408, 484, 57 },
		{ 408, 485, 57 },
		{ 408, 486, 57 },
		{ 408, 487, 57 },
		{ 408, 488, 57 },
		{ 408, 489, 57 },
		{ 408, 490, 57 },
		{ 408, 491, 57 },
		{ 408, 492, 57 },
		{ 408, 493, 57 },
		{ 408, 494, 57 },
		{ 408, 495, 57 },
		{ 408, 496, 57 },
		{ 408, 497, 57 },
		{ 408, 498, 57 },
		{ 408, 499, 57 },
		{ 408, 500, 57 },
		{ 408, 501, 57 },
		{ 408, 502, 57 },
		{ 408, 503, 57 },
		{ 408, 504, 57 },
		{ 408, 505, 57 },
		{ 408, 506, 57 },
		{ 408, 507, 57 },
		{ 408, 508, 57 },
		{ 408, 509, 57 },
		{ 408, 510, 57 },
		{ 408, 511, 57 },
		{ 408, 512, 57 },
		{ 409, 460, 57 },
		{ 409, 461, 57 },
		{ 409, 462, 57 },
		{ 409, 463, 57 },
		{ 409, 464, 57 },
		{ 409, 465, 57 },
		{ 409, 466, 57 },
		{ 409, 467, 57 },
		{ 409, 468, 57 },
		{ 409, 469, 57 },
		{ 409, 470, 57 },
		{ 409, 471, 57 },
		{ 409, 472, 57 },
		{ 409, 473, 57 },
		{ 409, 474, 57 },
		{ 409, 475, 57 },
		{ 409, 476, 57 },
		{ 409, 477, 57 },
		{ 409, 478, 57 },
		{ 409, 479, 57 },
		{ 409, 480, 57 },
		{ 409, 481, 57 },
		{ 409, 482, 57 },
		{ 409, 483, 57 },
		{ 409, 484, 57 },
		{ 409, 485, 57 },
		{ 409, 486, 57 },
		{ 409, 487, 57 },
		{ 409, 488, 57 },
		{ 409, 489, 57 },
		{ 409, 490, 57 },
		{ 409, 491, 57 },
		{ 409, 492, 57 },
		{ 409, 493, 57 },
		{ 409, 494, 57 },
		{ 409, 495, 57 },
		{ 409, 496, 57 },
		{ 409, 497, 57 },
		{ 409, 498, 57 },
		{ 409, 499, 57 },
		{ 409, 500, 57 },
		{ 409, 501, 57 },
		{ 409, 502, 57 },
		{ 409, 503, 57 },
		{ 409, 504, 57 },
		{ 409, 505, 57 },
		{ 409, 506, 57 },
		{ 409, 507, 57 },
		{ 409, 508, 57 },
		{ 409, 509, 57 },
		{ 409, 510, 57 },
		{ 409, 511, 57 },
		{ 409, 512, 57 },
		{ 410, 462, 57 },
		{ 410, 463, 57 },
		{ 410, 464, 57 },
		{ 410, 465, 57 },
		{ 410, 466, 57 },
		{ 410, 467, 57 },
		{ 410, 468, 57 },
		{ 410, 469, 57 },
		{ 410, 470, 57 },
		{ 410, 471, 57 },
		{ 410, 472, 57 },
		{ 410, 473, 57 },
		{ 410, 474, 57 },
		{ 410, 475, 57 },
		{ 410, 476, 57 },
		{ 410, 477, 57 },
		{ 410, 478, 57 },
		{ 410, 479, 57 },
		{ 410, 480, 57 },
		{ 410, 481, 57 },
		{ 410, 482, 57 },
		{ 410, 483, 57 },
		{ 410, 484, 57 },
		{ 410, 485, 57 },
		{ 410, 486, 57 },
		{ 410, 487, 57 },
		{ 410, 488, 57 },
		{ 410, 489, 57 },
		{ 410, 490, 57 },
		{ 410, 491, 57 },
		{ 410, 492, 57 },
		{ 410, 493, 57 },
		{ 410, 494, 57 },
		{ 410, 495, 57 },
		{ 410, 496, 57 },
		{ 410, 497, 57 },
		{ 410, 498, 57 },
		{ 410, 499, 57 },
		{ 410, 500, 57 },
		{ 410, 501, 57 },
		{ 410, 502, 57 },
		{ 410, 503, 57 },
		{ 410, 504, 57 },
		{ 410, 505, 57 },
		{ 410, 506, 57 },
		{ 410, 507, 57 },
		{ 410, 508, 57 },
		{ 410, 509, 57 },
		{ 410, 510, 57 },
		{ 410, 511, 57 },
		{ 410, 512, 57 },
		{ 411, 463, 57 },
		{ 411, 464, 57 },
		{ 411, 465, 57 },
		{ 411, 466, 57 },
		{ 411, 467, 57 },
		{ 411, 468, 57 },
		{ 411, 469, 57 },
		{ 411, 470, 57 },
		{ 411, 471, 57 },
		{ 411, 472, 57 },
		{ 411, 473, 57 },
		{ 411, 474, 57 },
		{ 411, 475, 57 },
		{ 411, 476, 57 },
		{ 411, 477, 57 },
		{ 411, 478, 57 },
		{ 411, 479, 57 },
		{ 411, 480, 57 },
		{ 411, 481, 57 },
		{ 411, 482, 57 },
		{ 411, 483, 57 },
		{ 411, 484, 57 },
		{ 411, 485, 57 },
		{ 411, 486, 57 },
		{ 411, 487, 57 },
		{ 411, 488, 57 },
		{ 411, 489, 57 },
		{ 411, 490, 57 },
		{ 411, 491, 57 },
		{ 411, 492, 57 },
		{ 411, 493, 57 },
		{ 411, 494, 57 },
		{ 411, 495, 57 },
		{ 411, 496, 57 },
		{ 411, 497, 57 },
		{ 411, 498, 57 },
		{ 411, 499, 57 },
		{ 411, 500, 57 },
		{ 411, 501, 57 },
		{ 411, 502, 57 },
		{ 411, 503, 57 },
		{ 411, 504, 57 },
		{ 411, 505, 57 },
		{ 411, 506, 57 },
		{ 411, 507, 57 },
		{ 411, 508, 57 },
		{ 411, 509, 57 },
		{ 411, 510, 57 },
		{ 411, 511, 57 },
		{ 411, 512, 57 },
		{ 412, 464, 57 },
		{ 412, 465, 57 },
		{ 412, 466, 57 },
		{ 412, 467, 57 },
		{ 412, 468, 57 },
		{ 412, 469, 57 },
		{ 412, 470, 57 },
		{ 412, 471, 57 },
		{ 412, 472, 57 },
		{ 412, 473, 57 },
		{ 412, 474, 57 },
		{ 412, 475, 57 },
		{ 412, 476, 57 },
		{ 412, 477, 57 },
		{ 412, 478, 57 },
		{ 412, 479, 57 },
		{ 412, 480, 57 },
		{ 412, 481, 57 },
		{ 412, 482, 57 },
		{ 412, 483, 57 },
		{ 412, 484, 57 },
		{ 412, 485, 57 },
		{ 412, 486, 57 },
		{ 412, 487, 57 },
		{ 412, 488, 57 },
		{ 412, 489, 57 },
		{ 412, 490, 57 },
		{ 412, 491, 57 },
		{ 412, 492, 57 },
		{ 412, 493, 57 },
		{ 412, 494, 57 },
		{ 412, 495, 57 },
		{ 412, 496, 57 },
		{ 412, 497, 57 },
		{ 412, 498, 57 },
		{ 412, 499, 57 },
		{ 412, 500, 57 },
		{ 412, 501, 57 },
		{ 412, 502, 57 },
		{ 412, 503, 57 },
		{ 412, 504, 57 },
		{ 412, 505, 57 },
		{ 412, 506, 57 },
		{ 412, 507, 57 },
		{ 412, 508, 57 },
		{ 412, 509, 57 },
		{ 412, 510, 57 },
		{ 412, 511, 57 },
		{ 412, 512, 57 },
		{ 413, 466, 57 },
		{ 413, 467, 57 },
		{ 413, 468, 57 },
		{ 413, 469, 57 },
		{ 413, 470, 57 },
		{ 413, 471, 57 },
		{ 413, 472, 57 },
		{ 413, 473, 57 },
		{ 413, 474, 57 },
		{ 413, 475, 57 },
		{ 413, 476, 57 },
		{ 413, 477, 57 },
		{ 413, 478, 57 },
		{ 413, 479, 57 },
		{ 413, 480, 57 },
		{ 413, 481, 57 },
		{ 413, 482, 57 },
		{ 413, 483, 57 },
		{ 413, 484, 57 },
		{ 413, 485, 57 },
		{ 413, 486, 57 },
		{ 413, 487, 57 },
		{ 413, 488, 57 },
		{ 413, 489, 57 },
		{ 413, 490, 57 },
		{ 413, 491, 57 },
		{ 413, 492, 57 },
		{ 413, 493, 57 },
		{ 413, 494, 57 },
		{ 413, 495, 57 },
		{ 413, 496, 57 },
		{ 413, 497, 57 },
		{ 413, 498, 57 },
		{ 413, 499, 57 },
		{ 413, 500, 57 },
		{ 413, 501, 57 },
		{ 413, 502, 57 },
		{ 413, 503, 57 },
		{ 413, 504, 57 },
		{ 413, 505, 57 },
		{ 413, 506, 57 },
		{ 413, 507, 57 },
		{ 413, 508, 57 },
		{ 413, 509, 57 },
		{ 413, 510, 57 },
		{ 413, 511, 57 },
		{ 413, 512, 57 },
		{ 414, 467, 57 },
		{ 414, 468, 57 },
		{ 414, 469, 57 },
		{ 414, 470, 57 },
		{ 414, 471, 57 },
		{ 414, 472, 57 },
		{ 414, 473, 57 },
		{ 414, 474, 57 },
		{ 414, 475, 57 },
		{ 414, 476, 57 },
		{ 414, 477, 57 },
		{ 414, 478, 57 },
		{ 414, 479, 57 },
		{ 414, 480, 57 },
		{ 414, 481, 57 },
		{ 414, 482, 57 },
		{ 414, 483, 57 },
		{ 414, 484, 57 },
		{ 414, 485, 57 },
		{ 414, 486, 57 },
		{ 414, 487, 57 },
		{ 414, 488, 57 },
		{ 414, 489, 57 },
		{ 414, 490, 57 },
		{ 414, 491, 57 },
		{ 414, 492, 57 },
		{ 414, 493, 57 },
		{ 414, 494, 57 },
		{ 414, 495, 57 },
		{ 414, 496, 57 },
		{ 414, 497, 57 },
		{ 414, 498, 57 },
		{ 414, 499, 57 },
		{ 414, 500, 57 },
		{ 414, 501, 57 },
		{ 414, 502, 57 },
		{ 414, 503, 57 },
		{ 414, 504, 57 },
		{ 414, 505, 57 },
		{ 414, 506, 57 },
		{ 414, 507, 57 },
		{ 414, 508, 57 },
		{ 414, 509, 57 },
		{ 414, 510, 57 },
		{ 414, 511, 57 },
		{ 414, 512, 57 },
		{ 415, 468, 57 },
		{ 415, 469, 57 },
		{ 415, 470, 57 },
		{ 415, 471, 57 },
		{ 415, 472, 57 },
		{ 415, 473, 57 },
		{ 415, 474, 57 },
		{ 415, 475, 57 },
		{ 415, 476, 57 },
		{ 415, 477, 57 },
		{ 415, 478, 57 },
		{ 415, 479, 57 },
		{ 415, 480, 57 },
		{ 415, 481, 57 },
		{ 415, 482, 57 },
		{ 415, 483, 57 },
		{ 415, 484, 57 },
		{ 415, 485, 57 },
		{ 415, 486, 57 },
		{ 415, 487, 57 },
		{ 415, 488, 57 },
		{ 415, 489, 57 },
		{ 415, 490, 57 },
		{ 415, 491, 57 },
		{ 415, 492, 57 },
		{ 415, 493, 57 },
		{ 415, 494, 57 },
		{ 415, 495, 57 },
		{ 415, 496, 57 },
		{ 415, 497, 57 },
		{ 415, 498, 57 },
		{ 415, 499, 57 },
		{ 415, 500, 57 },
		{ 415, 501, 57 },
		{ 415, 502, 57 },
		{ 415, 503, 57 },
		{ 415, 504, 57 },
		{ 415, 505, 57 },
		{ 415, 506, 57 },
		{ 415, 507, 57 },
		{ 415, 508, 57 },
		{ 415, 509, 57 },
		{ 415, 510, 57 },
		{ 415, 511, 57 },
		{ 415, 512, 57 },
		{ 416, 470, 57 },
		{ 416, 471, 57 },
		{ 416, 472, 57 },
		{ 416, 473, 57 },
		{ 416, 474, 57 },
		{ 416, 475, 57 },
		{ 416, 476, 57 },
		{ 416, 477, 57 },
		{ 416, 478, 57 },
		{ 416, 479, 57 },
		{ 416, 480, 57 },
		{ 416, 481, 57 },
		{ 416, 482, 57 },
		{ 416, 483, 57 },
		{ 416, 484, 57 },
		{ 416, 485, 57 },
		{ 416, 486, 57 },
		{ 416, 487, 57 },
		{ 416, 488, 57 },
		{ 416, 489, 57 },
		{ 416, 490, 57 },
		{ 416, 491, 57 },
		{ 416, 492, 57 },
		{ 416, 493, 57 },
		{ 416, 494, 57 },
		{ 416, 495, 57 },
		{ 416, 496, 57 },
		{ 416, 497, 57 },
		{ 416, 498, 57 },
		{ 416, 499, 57 },
		{ 416, 500, 57 },
		{ 416, 501, 57 },
		{ 416, 502, 57 },
		{ 416, 503, 57 },
		{ 416, 504, 57 },
		{ 416, 505, 57 },
		{ 416, 506, 57 },
		{ 416, 507, 57 },
		{ 416, 508, 57 },
		{ 416, 509, 57 },
		{ 416, 510, 57 },
		{ 416, 511, 57 },
		{ 416, 512, 57 },
		{ 417, 471, 57 },
		{ 417, 472, 57 },
		{ 417, 473, 57 },
		{ 417, 474, 57 },
		{ 417, 475, 57 },
		{ 417, 476, 57 },
		{ 417, 477, 57 },
		{ 417, 478, 57 },
		{ 417, 479, 57 },
		{ 417, 480, 57 },
		{ 417, 481, 57 },
		{ 417, 482, 57 },
		{ 417, 483, 57 },
		{ 417, 484, 57 },
		{ 417, 485, 57 },
		{ 417, 486, 57 },
		{ 417, 487, 57 },
		{ 417, 488, 57 },
		{ 417, 489, 57 },
		{ 417, 490, 57 },
		{ 417, 491, 57 },
		{ 417, 492, 57 },
		{ 417, 493, 57 },
		{ 417, 494, 57 },
		{ 417, 495, 57 },
		{ 417, 496, 57 },
		{ 417, 497, 57 },
		{ 417, 498, 57 },
		{ 417, 499, 57 },
		{ 417, 500, 57 },
		{ 417, 501, 57 },
		{ 417, 502, 57 },
		{ 417, 503, 57 },
		{ 417, 504, 57 },
		{ 417, 505, 57 },
		{ 417, 506, 57 },
		{ 417, 507, 57 },
		{ 417, 508, 57 },
		{ 417, 509, 57 },
		{ 417, 510, 57 },
		{ 417, 511, 57 },
		{ 417, 512, 57 },
		{ 418, 472, 57 },
		{ 418, 473, 57 },
		{ 418, 474, 57 },
		{ 418, 475, 57 },
		{ 418, 476, 57 },
		{ 418, 477, 57 },
		{ 418, 478, 57 },
		{ 418, 479, 57 },
		{ 418, 480, 57 },
		{ 418, 481, 57 },
		{ 418, 482, 57 },
		{ 418, 483, 57 },
		{ 418, 484, 57 },
		{ 418, 485, 57 },
		{ 418, 486, 57 },
		{ 418, 487, 57 },
		{ 418, 488, 57 },
		{ 418, 489, 57 },
		{ 418, 490, 57 },
		{ 418, 491, 57 },
		{ 418, 492, 57 },
		{ 418, 493, 57 },
		{ 418, 494, 57 },
		{ 418, 495, 57 },
		{ 418, 496, 57 },
		{ 418, 497, 57 },
		{ 418, 498, 57 },
		{ 418, 499, 57 },
		{ 418, 500, 57 },
		{ 418, 501, 57 },
		{ 418, 502, 57 },
		{ 418, 503, 57 },
		{ 418, 504, 57 },
		{ 418, 505, 57 },
		{ 418, 506, 57 },
		{ 418, 507, 57 },
		{ 418, 508, 57 },
		{ 418, 509, 57 },
		{ 418, 510, 57 },
		{ 418, 511, 57 },
		{ 418, 512, 57 },
		{ 419, 474, 57 },
		{ 419, 475, 57 },
		{ 419, 476, 57 },
		{ 419, 477, 57 },
		{ 419, 478, 57 },
		{ 419, 479, 57 },
		{ 419, 480, 57 },
		{ 419, 481, 57 },
		{ 419, 482, 57 },
		{ 419, 483, 57 },
		{ 419, 484, 57 },
		{ 419, 485, 57 },
		{ 419, 486, 57 },
		{ 419, 487, 57 },
		{ 419, 488, 57 },
		{ 419, 489, 57 },
		{ 419, 490, 57 },
		{ 419, 491, 57 },
		{ 419, 492, 57 },
		{ 419, 493, 57 },
		{ 419, 494, 57 },
		{ 419, 495, 57 },
		{ 419, 496, 57 },
		{ 419, 497, 57 },
		{ 419, 498, 57 },
		{ 419, 499, 57 },
		{ 419, 500, 57 },
		{ 419, 501, 57 },
		{ 419, 502, 57 },
		{ 419, 503, 57 },
		{ 419, 504, 57 },
		{ 419, 505, 57 },
		{ 419, 506, 57 },
		{ 419, 507, 57 },
		{ 419, 508, 57 },
		{ 419, 509, 57 },
		{ 419, 510, 57 },
		{ 419, 511, 57 },
		{ 419, 512, 57 },
		{ 420, 475, 57 },
		{ 420, 476, 57 },
		{ 420, 477, 57 },
		{ 420, 478, 57 },
		{ 420, 479, 57 },
		{ 420, 480, 57 },
		{ 420, 481, 57 },
		{ 420, 482, 57 },
		{ 420, 483, 57 },
		{ 420, 484, 57 },
		{ 420, 485, 57 },
		{ 420, 486, 57 },
		{ 420, 487, 57 },
		{ 420, 488, 57 },
		{ 420, 489, 57 },
		{ 420, 490, 57 },
		{ 420, 491, 57 },
		{ 420, 492, 57 },
		{ 420, 493, 57 },
		{ 420, 494, 57 },
		{ 420, 495, 57 },
		{ 420, 496, 57 },
		{ 420, 497, 57 },
		{ 420, 498, 57 },
		{ 420, 499, 57 },
		{ 420, 500, 57 },
		{ 420, 501, 57 },
		{ 420, 502, 57 },
		{ 420, 503, 57 },
		{ 420, 504, 57 },
		{ 420, 505, 57 },
		{ 420, 506, 57 },
		{ 420, 507, 57 },
		{ 420, 508, 57 },
		{ 420, 509, 57 },
		{ 420, 510, 57 },
		{ 420, 511, 57 },
		{ 420, 512, 57 },
		{ 421, 476, 57 },
		{ 421, 477, 57 },
		{ 421, 478, 57 },
		{ 421, 479, 57 },
		{ 421, 480, 57 },
		{ 421, 481, 57 },
		{ 421, 482, 57 },
		{ 421, 483, 57 },
		{ 421, 484, 57 },
		{ 421, 485, 57 },
		{ 421, 486, 57 },
		{ 421, 487, 57 },
		{ 421, 488, 57 },
		{ 421, 489, 57 },
		{ 421, 490, 57 },
		{ 421, 491, 57 },
		{ 421, 492, 57 },
		{ 421, 493, 57 },
		{ 421, 494, 57 },
		{ 421, 495, 57 },
		{ 421, 496, 57 },
		{ 421, 497, 57 },
		{ 421, 498, 57 },
		{ 421, 499, 57 },
		{ 421, 500, 57 },
		{ 421, 501, 57 },
		{ 421, 502, 57 },
		{ 421, 503, 57 },
		{ 421, 504, 57 },
		{ 421, 505, 57 },
		{ 421, 506, 57 },
		{ 421, 507, 57 },
		{ 421, 508, 57 },
		{ 421, 509, 57 },
		{ 421, 510, 57 },
		{ 421, 511, 57 },
		{ 421, 512, 57 },
		{ 422, 478, 57 },
		{ 422, 479, 57 },
		{ 422, 480, 57 },
		{ 422, 481, 57 },
		{ 422, 482, 57 },
		{ 422, 483, 57 },
		{ 422, 484, 57 },
		{ 422, 485, 57 },
		{ 422, 486, 57 },
		{ 422, 487, 57 },
		{ 422, 488, 57 },
		{ 422, 489, 57 },
		{ 422, 490, 57 },
		{ 422, 491, 57 },
		{ 422, 492, 57 },
		{ 422, 493, 57 },
		{ 422, 494, 57 },
		{ 422, 495, 57 },
		{ 422, 496, 57 },
		{ 422, 497, 57 },
		{ 422, 498, 57 },
		{ 422, 499, 57 },
		{ 422, 500, 57 },
		{ 422, 501, 57 },
		{ 422, 502, 57 },
		{ 422, 503, 57 },
		{ 422, 504, 57 },
		{ 422, 505, 57 },
		{ 422, 506, 57 },
		{ 422, 507, 57 },
		{ 422, 508, 57 },
		{ 422, 509, 57 },
		{ 422, 510, 57 },
		{ 422, 511, 57 },
		{ 422, 512, 57 },
		{ 423, 479, 57 },
		{ 423, 480, 57 },
		{ 423, 481, 57 },
		{ 423, 482, 57 },
		{ 423, 483, 57 },
		{ 423, 484, 57 },
		{ 423, 485, 57 },
		{ 423, 486, 57 },
		{ 423, 487, 57 },
		{ 423, 488, 57 },
		{ 423, 489, 57 },
		{ 423, 490, 57 },
		{ 423, 491, 57 },
		{ 423, 492, 57 },
		{ 423, 493, 57 },
		{ 423, 494, 57 },
		{ 423, 495, 57 },
		{ 423, 496, 57 },
		{ 423, 497, 57 },
		{ 423, 498, 57 },
		{ 423, 499, 57 },
		{ 423, 500, 57 },
		{ 423, 501, 57 },
		{ 423, 502, 57 },
		{ 423, 503, 57 },
		{ 423, 504, 57 },
		{ 423, 505, 57 },
		{ 423, 506, 57 },
		{ 423, 507, 57 },
		{ 423, 508, 57 },
		{ 423, 509, 57 },
		{ 423, 510, 57 },
		{ 423, 511, 57 },
		{ 423, 512, 57 },
		{ 424, 480, 57 },
		{ 424, 481, 57 },
		{ 424, 482, 57 },
		{ 424, 483, 57 },
		{ 424, 484, 57 },
		{ 424, 485, 57 },
		{ 424, 486, 57 },
		{ 424, 487, 57 },
		{ 424, 488, 57 },
		{ 424, 489, 57 },
		{ 424, 490, 57 },
		{ 424, 491, 57 },
		{ 424, 492, 57 },
		{ 424, 493, 57 },
		{ 424, 494, 57 },
		{ 424, 495, 57 },
		{ 424, 496, 57 },
		{ 424, 497, 57 },
		{ 424, 498, 57 },
		{ 424, 499, 57 },
		{ 424, 500, 57 },
		{ 424, 501, 57 },
		{ 424, 502, 57 },
		{ 424, 503, 57 },
		{ 424, 504, 57 },
		{ 424, 505, 57 },
		{ 424, 506, 57 },
		{ 424, 507, 57 },
		{ 424, 508, 57 },
		{ 424, 509, 57 },
		{ 424, 510, 57 },
		{ 424, 511, 57 },
		{ 424, 512, 57 },
		{ 425, 482, 57 },
		{ 425, 483, 57 },
		{ 425, 484, 57 },
		{ 425, 485, 57 },
		{ 425, 486, 57 },
		{ 425, 487, 57 },
		{ 425, 488, 57 },
		{ 425, 489, 57 },
		{ 425, 490, 57 },
		{ 425, 491, 57 },
		{ 425, 492, 57 },
		{ 425, 493, 57 },
		{ 425, 494, 57 },
		{ 425, 495, 57 },
		{ 425, 496, 57 },
		{ 425, 497, 57 },
		{ 425, 498, 57 },
		{ 425, 499, 57 },
		{ 425, 500, 57 },
		{ 425, 501, 57 },
		{ 425, 502, 57 },
		{ 425, 503, 57 },
		{ 425, 504, 57 },
		{ 425, 505, 57 },
		{ 425, 506, 57 },
		{ 425, 507, 57 },
		{ 425, 508, 57 },
		{ 425, 509, 57 },
		{ 425, 510, 57 },
		{ 425, 511, 57 },
		{ 425, 512, 57 },
		{ 426, 483, 57 },
		{ 426, 484, 57 },
		{ 426, 485, 57 },
		{ 426, 486, 57 },
		{ 426, 487, 57 },
		{ 426, 488, 57 },
		{ 426, 489, 57 },
		{ 426, 490, 57 },
		{ 426, 491, 57 },
		{ 426, 492, 57 },
		{ 426, 493, 57 },
		{ 426, 494, 57 },
		{ 426, 495, 57 },
		{ 426, 496, 57 },
		{ 426, 497, 57 },
		{ 426, 498, 57 },
		{ 426, 499, 57 },
		{ 426, 500, 57 },
		{ 426, 501, 57 },
		{ 426, 502, 57 },
		{ 426, 503, 57 },
		{ 426, 504, 57 },
		{ 426, 505, 57 },
		{ 426, 506, 57 },
		{ 426, 507, 57 },
		{ 426, 508, 57 },
		{ 426, 509, 57 },
		{ 426, 510, 57 },
		{ 426, 511, 57 },
		{ 426, 512, 57 },
		{ 427, 484, 57 },
		{ 427, 485, 57 },
		{ 427, 486, 57 },
		{ 427, 487, 57 },
		{ 427, 488, 57 },
		{ 427, 489, 57 },
		{ 427, 490, 57 },
		{ 427, 491, 57 },
		{ 427, 492, 57 },
		{ 427, 493, 57 },
		{ 427, 494, 57 },
		{ 427, 495, 57 },
		{ 427, 496, 57 },
		{ 427, 497, 57 },
		{ 427, 498, 57 },
		{ 427, 499, 57 },
		{ 427, 500, 57 },
		{ 427, 501, 57 },
		{ 427, 502, 57 },
		{ 427, 503, 57 },
		{ 427, 504, 57 },
		{ 427, 505, 57 },
		{ 427, 506, 57 },
		{ 427, 507, 57 },
		{ 427, 508, 57 },
		{ 427, 509, 57 },
		{ 427, 510, 57 },
		{ 427, 511, 57 },
		{ 427, 512, 57 },
		{ 428, 486, 57 },
		{ 428, 487, 57 },
		{ 428, 488, 57 },
		{ 428, 489, 57 },
		{ 428, 490, 57 },
		{ 428, 491, 57 },
		{ 428, 492, 57 },
		{ 428, 493, 57 },
		{ 428, 494, 57 },
		{ 428, 495, 57 },
		{ 428, 496, 57 },
		{ 428, 497, 57 },
		{ 428, 498, 57 },
		{ 428, 499, 57 },
		{ 428, 500, 57 },
		{ 428, 501, 57 },
		{ 428, 502, 57 },
		{ 428, 503, 57 },
		{ 428, 504, 57 },
		{ 428, 505, 57 },
		{ 428, 506, 57 },
		{ 428, 507, 57 },
		{ 428, 508, 57 },
		{ 428, 509, 57 },
		{ 428, 510, 57 },
		{ 428, 511, 57 },
		{ 428, 512, 57 },
		{ 429, 487, 57 },
		{ 429, 488, 57 },
		{ 429, 489, 57 },
		{ 429, 490, 57 },
		{ 429, 491, 57 },
		{ 429, 492, 57 },
		{ 429, 493, 57 },
		{ 429, 494, 57 },
		{ 429, 495, 57 },
		{ 429, 496, 57 },
		{ 429, 497, 57 },
		{ 429, 498, 57 },
		{ 429, 499, 57 },
		{ 429, 500, 57 },
		{ 429, 501, 57 },
		{ 429, 502, 57 },
		{ 429, 503, 57 },
		{ 429, 504, 57 },
		{ 429, 505, 57 },
		{ 429, 506, 57 },
		{ 429, 507, 57 },
		{ 429, 508, 57 },
		{ 429, 509, 57 },
		{ 429, 510, 57 },
		{ 429, 511, 57 },
		{ 429, 512, 57 },
		{ 430, 488, 57 },
		{ 430, 489, 57 },
		{ 430, 490, 57 },
		{ 430, 491, 57 },
		{ 430, 492, 57 },
		{ 430, 493, 57 },
		{ 430, 494, 57 },
		{ 430, 495, 57 },
		{ 430, 496, 57 },
		{ 430, 497, 57 },
		{ 430, 498, 57 },
		{ 430, 499, 57 },
		{ 430, 500, 57 },
		{ 430, 501, 57 },
		{ 430, 502, 57 },
		{ 430, 503, 57 },
		{ 430, 504, 57 },
		{ 430, 505, 57 },
		{ 430, 506, 57 },
		{ 430, 507, 57 },
		{ 430, 508, 57 },
		{ 430, 509, 57 },
		{ 430, 510, 57 },
		{ 430, 511, 57 },
		{ 430, 512, 57 },
		{ 431, 490, 57 },
		{ 431, 491, 57 },
		{ 431, 492, 57 },
		{ 431, 493, 57 },
		{ 431, 494, 57 },
		{ 431, 495, 57 },
		{ 431, 496, 57 },
		{ 431, 497, 57 },
		{ 431, 498, 57 },
		{ 431, 499, 57 },
		{ 431, 500, 57 },
		{ 431, 501, 57 },
		{ 431, 502, 57 },
		{ 431, 503, 57 },
		{ 431, 504, 57 },
		{ 431, 505, 57 },
		{ 431, 506, 57 },
		{ 431, 507, 57 },
		{ 431, 508, 57 },
		{ 431, 509, 57 },
		{ 431, 510, 57 },
		{ 431, 511, 57 },
		{ 431, 512, 57 },
		{ 432, 491, 57 },
		{ 432, 492, 57 },
		{ 432, 493, 57 },
		{ 432, 494, 57 },
		{ 432, 495, 57 },
		{ 432, 496, 57 },
		{ 432, 497, 57 },
		{ 432, 498, 57 },
		{ 432, 499, 57 },
		{ 432, 500, 57 },
		{ 432, 501, 57 },
		{ 432, 502, 57 },
		{ 432, 503, 57 },
		{ 432, 504, 57 },
		{ 432, 505, 57 },
		{ 432, 506, 57 },
		{ 432, 507, 57 },
		{ 432, 508, 57 },
		{ 432, 509, 57 },
		{ 432, 510, 57 },
		{ 432, 511, 57 },
		{ 432, 512, 57 },
		{ 433, 492, 57 },
		{ 433, 493, 57 },
		{ 433, 494, 57 },
		{ 433, 495, 57 },
		{ 433, 496, 57 },
		{ 433, 497, 57 },
		{ 433, 498, 57 },
		{ 433, 499, 57 },
		{ 433, 500, 57 },
		{ 433, 501, 57 },
		{ 433, 502, 57 },
		{ 433, 503, 57 },
		{ 433, 504, 57 },
		{ 433, 505, 57 },
		{ 433, 506, 57 },
		{ 433, 507, 57 },
		{ 433, 508, 57 },
		{ 433, 509, 57 },
		{ 433, 510, 57 },
		{ 433, 511, 57 },
		{ 433, 512, 57 },
		{ 434, 494, 57 },
		{ 434, 495, 57 },
		{ 434, 496, 57 },
		{ 434, 497, 57 },
		{ 434, 498, 57 },
		{ 434, 499, 57 },
		{ 434, 500, 57 },
		{ 434, 501, 57 },
		{ 434, 502, 57 },
		{ 434, 503, 57 },
		{ 434, 504, 57 },
		{ 434, 505, 57 },
		{ 434, 506, 57 },
		{ 434, 507, 57 },
		{ 434, 508, 57 },
		{ 434, 509, 57 },
		{ 434, 510, 57 },
		{ 434, 511, 57 },
		{ 434, 512, 57 },
		{ 435, 495, 57 },
		{ 435, 496, 57 },
		{ 435, 497, 57 },
		{ 435, 498, 57 },
		{ 435, 499, 57 },
		{ 435, 500, 57 },
		{ 435, 501, 57 },
		{ 435, 502, 57 },
		{ 435, 503, 57 },
		{ 435, 504, 57 },
		{ 435, 505, 57 },
		{ 435, 506, 57 },
		{ 435, 507, 57 },
		{ 435, 508, 57 },
		{ 435, 509, 57 },
		{ 435, 510, 57 },
		{ 435, 511, 57 },
		{ 435, 512, 57 },
		{ 436, 496, 57 },
		{ 436, 497, 57 },
		{ 436, 498, 57 },
		{ 436, 499, 57 },
		{ 436, 500, 57 },
		{ 436, 501, 57 },
		{ 436, 502, 57 },
		{ 436, 503, 57 },
		{ 436, 504, 57 },
		{ 436, 505, 57 },
		{ 436, 506, 57 },
		{ 436, 507, 57 },
		{ 436, 508, 57 },
		{ 436, 509, 57 },
		{ 436, 510, 57 },
		{ 436, 511, 57 },
		{ 436, 512, 57 },
		{ 437, 498, 57 },
		{ 437, 499, 57 },
		{ 437, 500, 57 },
		{ 437, 501, 57 },
		{ 437, 502, 57 },
		{ 437, 503, 57 },
		{ 437, 504, 57 },
		{ 437, 505, 57 },
		{ 437, 506, 57 },
		{ 437, 507, 57 },
		{ 437, 508, 57 },
		{ 437, 509, 57 },
		{ 437, 510, 57 },
		{ 437, 511, 57 },
		{ 437, 512, 57 },
		{ 438, 499, 57 },
		{ 438, 500, 57 },
		{ 438, 501, 57 },
		{ 438, 502, 57 },
		{ 438, 503, 57 },
		{ 438, 504, 57 },
		{ 438, 505, 57 },
		{ 438, 506, 57 },
		{ 438, 507, 57 },
		{ 438, 508, 57 },
		{ 438, 509, 57 },
		{ 438, 510, 57 },
		{ 438, 511, 57 },
		{ 438, 512, 57 },
		{ 439, 500, 57 },
		{ 439, 501, 57 },
		{ 439, 502, 57 },
		{ 439, 503, 57 },
		{ 439, 504, 57 },
		{ 439, 505, 57 },
		{ 439, 506, 57 },
		{ 439, 507, 57 },
		{ 439, 508, 57 },
		{ 439, 509, 57 },
		{ 439, 510, 57 },
		{ 439, 511, 57 },
		{ 439, 512, 57 },
		{ 440, 502, 57 },
		{ 440, 503, 57 },
		{ 440, 504, 57 },
		{ 440, 505, 57 },
		{ 440, 506, 57 },
		{ 440, 507, 57 },
		{ 440, 508, 57 },
		{ 440, 509, 57 },
		{ 440, 510, 57 },
		{ 440, 511, 57 },
		{ 440, 512, 57 },
		{ 441, 503, 57 },
		{ 441, 504, 57 },
		{ 441, 505, 57 },
		{ 441, 506, 57 },
		{ 441, 507, 57 },
		{ 441, 508, 57 },
		{ 441, 509, 57 },
		{ 441, 510, 57 },
		{ 441, 511, 57 },
		{ 441, 512, 57 },
		{ 442, 504, 57 },
		{ 442, 505, 57 },
		{ 442, 506, 57 },
		{ 442, 507, 57 },
		{ 442, 508, 57 },
		{ 442, 509, 57 },
		{ 442, 510, 57 },
		{ 442, 511, 57 },
		{ 442, 512, 57 },
		{ 443, 506, 57 },
		{ 443, 507, 57 },
		{ 443, 508, 57 },
		{ 443, 509, 57 },
		{ 443, 510, 57 },
		{ 443, 511, 57 },
		{ 443, 512, 57 },
		{ 444, 507, 57 },
		{ 444, 508, 57 },
		{ 444, 509, 57 },
		{ 444, 510, 57 },
		{ 444, 511, 57 },
		{ 444, 512, 57 },
		{ 445, 508, 57 },
		{ 445, 509, 57 },
		{ 445, 510, 57 },
		{ 445, 511, 57 },
		{ 445, 512, 57 },
		{ 446, 510, 57 },
		{ 446, 511, 57 },
		{ 446, 512, 57 },
		{ 447, 511, 57 },
		{ 447, 512, 57 },
		{ 448, 512, 57 },
		{ 98, 1, 58 },
		{ 98, 2, 58 },
		{ 99, 1, 58 },
		{ 99, 2, 58 },
		{ 99, 3, 58 },
		{ 99, 4, 58 },
		{ 100, 1, 58 },
		{ 100, 2, 58 },
		{ 100, 3, 58 },
		{ 100, 4, 58 },
		{ 100, 5, 58 },
		{ 101, 1, 58 },
		{ 101, 2, 58 },
		{ 101, 3, 58 },
		{ 101, 4, 58 },
		{ 101, 5, 58 },
		{ 101, 6, 58 },
		{ 101, 7, 58 },
		{ 102, 1, 58 },
		{ 102, 2, 58 },
		{ 102, 3, 58 },
		{ 102, 4, 58 },
		{ 102, 5, 58 },
		{ 102, 6, 58 },
		{ 102, 7, 58 },
		{ 102, 8, 58 },
		{ 103, 1, 58 },
		{ 103, 2, 58 },
		{ 103, 3, 58 },
		{ 103, 4, 58 },
		{ 103, 5, 58 },
		{ 103, 6, 58 },
		{ 103, 7, 58 },
		{ 103, 8, 58 },
		{ 103, 9, 58 },
		{ 103, 10, 58 },
		{ 104, 1, 58 },
		{ 104, 2, 58 },
		{ 104, 3, 58 },
		{ 104, 4, 58 },
		{ 104, 5, 58 },
		{ 104, 6, 58 },
		{ 104, 7, 58 },
		{ 104, 8, 58 },
		{ 104, 9, 58 },
		{ 104, 10, 58 },
		{ 104, 11, 58 },
		{ 104, 12, 58 },
		{ 105, 1, 58 },
		{ 105, 2, 58 },
		{ 105, 3, 58 },
		{ 105, 4, 58 },
		{ 105, 5, 58 },
		{ 105, 6, 58 },
		{ 105, 7, 58 },
		{ 105, 8, 58 },
		{ 105, 9, 58 },
		{ 105, 10, 58 },
		{ 105, 11, 58 },
		{ 105, 12, 58 },
		{ 105, 13, 58 },
		{ 106, 1, 58 },
		{ 106, 2, 58 },
		{ 106, 3, 58 },
		{ 106, 4, 58 },
		{ 106, 5, 58 },
		{ 106, 6, 58 },
		{ 106, 7, 58 },
		{ 106, 8, 58 },
		{ 106, 9, 58 },
		{ 106, 10, 58 },
		{ 106, 11, 58 },
		{ 106, 12, 58 },
		{ 106, 13, 58 },
		{ 106, 14, 58 },
		{ 106, 15, 58 },
		{ 107, 1, 58 },
		{ 107, 2, 58 },
		{ 107, 3, 58 },
		{ 107, 4, 58 },
		{ 107, 5, 58 },
		{ 107, 6, 58 },
		{ 107, 7, 58 },
		{ 107, 8, 58 },
		{ 107, 9, 58 },
		{ 107, 10, 58 },
		{ 107, 11, 58 },
		{ 107, 12, 58 },
		{ 107, 13, 58 },
		{ 107, 14, 58 },
		{ 107, 15, 58 },
		{ 107, 16, 58 },
		{ 108, 1, 58 },
		{ 108, 2, 58 },
		{ 108, 3, 58 },
		{ 108, 4, 58 },
		{ 108, 5, 58 },
		{ 108, 6, 58 },
		{ 108, 7, 58 },
		{ 108, 8, 58 },
		{ 108, 9, 58 },
		{ 108, 10, 58 },
		{ 108, 11, 58 },
		{ 108, 12, 58 },
		{ 108, 13, 58 },
		{ 108, 14, 58 },
		{ 108, 15, 58 },
		{ 108, 16, 58 },
		{ 108, 17, 58 },
		{ 108, 18, 58 },
		{ 109, 1, 58 },
		{ 109, 2, 58 },
		{ 109, 3, 58 },
		{ 109, 4, 58 },
		{ 109, 5, 58 },
		{ 109, 6, 58 },
		{ 109, 7, 58 },
		{ 109, 8, 58 },
		{ 109, 9, 58 },
		{ 109, 10, 58 },
		{ 109, 11, 58 },
		{ 109, 12, 58 },
		{ 109, 13, 58 },
		{ 109, 14, 58 },
		{ 109, 15, 58 },
		{ 109, 16, 58 },
		{ 109, 17, 58 },
		{ 109, 18, 58 },
		{ 109, 19, 58 },
		{ 109, 20, 58 },
		{ 110, 1, 58 },
		{ 110, 2, 58 },
		{ 110, 3, 58 },
		{ 110, 4, 58 },
		{ 110, 5, 58 },
		{ 110, 6, 58 },
		{ 110, 7, 58 },
		{ 110, 8, 58 },
		{ 110, 9, 58 },
		{ 110, 10, 58 },
		{ 110, 11, 58 },
		{ 110, 12, 58 },
		{ 110, 13, 58 },
		{ 110, 14, 58 },
		{ 110, 15, 58 },
		{ 110, 16, 58 },
		{ 110, 17, 58 },
		{ 110, 18, 58 },
		{ 110, 19, 58 },
		{ 110, 20, 58 },
		{ 110, 21, 58 },
		{ 111, 1, 58 },
		{ 111, 2, 58 },
		{ 111, 3, 58 },
		{ 111, 4, 58 },
		{ 111, 5, 58 },
		{ 111, 6, 58 },
		{ 111, 7, 58 },
		{ 111, 8, 58 },
		{ 111, 9, 58 },
		{ 111, 10, 58 },
		{ 111, 11, 58 },
		{ 111, 12, 58 },
		{ 111, 13, 58 },
		{ 111, 14, 58 },
		{ 111, 15, 58 },
		{ 111, 16, 58 },
		{ 111, 17, 58 },
		{ 111, 18, 58 },
		{ 111, 19, 58 },
		{ 111, 20, 58 },
		{ 111, 21, 58 },
		{ 111, 22, 58 },
		{ 111, 23, 58 },
		{ 112, 1, 58 },
		{ 112, 2, 58 },
		{ 112, 3, 58 },
		{ 112, 4, 58 },
		{ 112, 5, 58 },
		{ 112, 6, 58 },
		{ 112, 7, 58 },
		{ 112, 8, 58 },
		{ 112, 9, 58 },
		{ 112, 10, 58 },
		{ 112, 11, 58 },
		{ 112, 12, 58 },
		{ 112, 13, 58 },
		{ 112, 14, 58 },
		{ 112, 15, 58 },
		{ 112, 16, 58 },
		{ 112, 17, 58 },
		{ 112, 18, 58 },
		{ 112, 19, 58 },
		{ 112, 20, 58 },
		{ 112, 21, 58 },
		{ 112, 22, 58 },
		{ 112, 23, 58 },
		{ 112, 24, 58 },
		{ 113, 1, 58 },
		{ 113, 2, 58 },
		{ 113, 3, 58 },
		{ 113, 4, 58 },
		{ 113, 5, 58 },
		{ 113, 6, 58 },
		{ 113, 7, 58 },
		{ 113, 8, 58 },
		{ 113, 9, 58 },
		{ 113, 10, 58 },
		{ 113, 11, 58 },
		{ 113, 12, 58 },
		{ 113, 13, 58 },
		{ 113, 14, 58 },
		{ 113, 15, 58 },
		{ 113, 16, 58 },
		{ 113, 17, 58 },
		{ 113, 18, 58 },
		{ 113, 19, 58 },
		{ 113, 20, 58 },
		{ 113, 21, 58 },
		{ 113, 22, 58 },
		{ 113, 23, 58 },
		{ 113, 24, 58 },
		{ 113, 25, 58 },
		{ 113, 26, 58 },
		{ 114, 1, 58 },
		{ 114, 2, 58 },
		{ 114, 3, 58 },
		{ 114, 4, 58 },
		{ 114, 5, 58 },
		{ 114, 6, 58 },
		{ 114, 7, 58 },
		{ 114, 8, 58 },
		{ 114, 9, 58 },
		{ 114, 10, 58 },
		{ 114, 11, 58 },
		{ 114, 12, 58 },
		{ 114, 13, 58 },
		{ 114, 14, 58 },
		{ 114, 15, 58 },
		{ 114, 16, 58 },
		{ 114, 17, 58 },
		{ 114, 18, 58 },
		{ 114, 19, 58 },
		{ 114, 20, 58 },
		{ 114, 21, 58 },
		{ 114, 22, 58 },
		{ 114, 23, 58 },
		{ 114, 24, 58 },
		{ 114, 25, 58 },
		{ 114, 26, 58 },
		{ 114, 27, 58 },
		{ 114, 28, 58 },
		{ 115, 1, 58 },
		{ 115, 2, 58 },
		{ 115, 3, 58 },
		{ 115, 4, 58 },
		{ 115, 5, 58 },
		{ 115, 6, 58 },
		{ 115, 7, 58 },
		{ 115, 8, 58 },
		{ 115, 9, 58 },
		{ 115, 10, 58 },
		{ 115, 11, 58 },
		{ 115, 12, 58 },
		{ 115, 13, 58 },
		{ 115, 14, 58 },
		{ 115, 15, 58 },
		{ 115, 16, 58 },
		{ 115, 17, 58 },
		{ 115, 18, 58 },
		{ 115, 19, 58 },
		{ 115, 20, 58 },
		{ 115, 21, 58 },
		{ 115, 22, 58 },
		{ 115, 23, 58 },
		{ 115, 24, 58 },
		{ 115, 25, 58 },
		{ 115, 26, 58 },
		{ 115, 27, 58 },
		{ 115, 28, 58 },
		{ 115, 29, 58 },
		{ 116, 1, 58 },
		{ 116, 2, 58 },
		{ 116, 3, 58 },
		{ 116, 4, 58 },
		{ 116, 5, 58 },
		{ 116, 6, 58 },
		{ 116, 7, 58 },
		{ 116, 8, 58 },
		{ 116, 9, 58 },
		{ 116, 10, 58 },
		{ 116, 11, 58 },
		{ 116, 12, 58 },
		{ 116, 13, 58 },
		{ 116, 14, 58 },
		{ 116, 15, 58 },
		{ 116, 16, 58 },
		{ 116, 17, 58 },
		{ 116, 18, 58 },
		{ 116, 19, 58 },
		{ 116, 20, 58 },
		{ 116, 21, 58 },
		{ 116, 22, 58 },
		{ 116, 23, 58 },
		{ 116, 24, 58 },
		{ 116, 25, 58 },
		{ 116, 26, 58 },
		{ 116, 27, 58 },
		{ 116, 28, 58 },
		{ 116, 29, 58 },
		{ 116, 30, 58 },
		{ 116, 31, 58 },
		{ 117, 1, 58 },
		{ 117, 2, 58 },
		{ 117, 3, 58 },
		{ 117, 4, 58 },
		{ 117, 5, 58 },
		{ 117, 6, 58 },
		{ 117, 7, 58 },
		{ 117, 8, 58 },
		{ 117, 9, 58 },
		{ 117, 10, 58 },
		{ 117, 11, 58 },
		{ 117, 12, 58 },
		{ 117, 13, 58 },
		{ 117, 14, 58 },
		{ 117, 15, 58 },
		{ 117, 16, 58 },
		{ 117, 17, 58 },
		{ 117, 18, 58 },
		{ 117, 19, 58 },
		{ 117, 20, 58 },
		{ 117, 21, 58 },
		{ 117, 22, 58 },
		{ 117, 23, 58 },
		{ 117, 24, 58 },
		{ 117, 25, 58 },
		{ 117, 26, 58 },
		{ 117, 27, 58 },
		{ 117, 28, 58 },
		{ 117, 29, 58 },
		{ 117, 30, 58 },
		{ 117, 31, 58 },
		{ 117, 32, 58 },
		{ 118, 1, 58 },
		{ 118, 2, 58 },
		{ 118, 3, 58 },
		{ 118, 4, 58 },
		{ 118, 5, 58 },
		{ 118, 6, 58 },
		{ 118, 7, 58 },
		{ 118, 8, 58 },
		{ 118, 9, 58 },
		{ 118, 10, 58 },
		{ 118, 11, 58 },
		{ 118, 12, 58 },
		{ 118, 13, 58 },
		{ 118, 14, 58 },
		{ 118, 15, 58 },
		{ 118, 16, 58 },
		{ 118, 17, 58 },
		{ 118, 18, 58 },
		{ 118, 19, 58 },
		{ 118, 20, 58 },
		{ 118, 21, 58 },
		{ 118, 22, 58 },
		{ 118, 23, 58 },
		{ 118, 24, 58 },
		{ 118, 25, 58 },
		{ 118, 26, 58 },
		{ 118, 27, 58 },
		{ 118, 28, 58 },
		{ 118, 29, 58 },
		{ 118, 30, 58 },
		{ 118, 31, 58 },
		{ 118, 32, 58 },
		{ 118, 33, 58 },
		{ 118, 34, 58 },
		{ 119, 1, 58 },
		{ 119, 2, 58 },
		{ 119, 3, 58 },
		{ 119, 4, 58 },
		{ 119, 5, 58 },
		{ 119, 6, 58 },
		{ 119, 7, 58 },
		{ 119, 8, 58 },
		{ 119, 9, 58 },
		{ 119, 10, 58 },
		{ 119, 11, 58 },
		{ 119, 12, 58 },
		{ 119, 13, 58 },
		{ 119, 14, 58 },
		{ 119, 15, 58 },
		{ 119, 16, 58 },
		{ 119, 17, 58 },
		{ 119, 18, 58 },
		{ 119, 19, 58 },
		{ 119, 20, 58 },
		{ 119, 21, 58 },
		{ 119, 22, 58 },
		{ 119, 23, 58 },
		{ 119, 24, 58 },
		{ 119, 25, 58 },
		{ 119, 26, 58 },
		{ 119, 27, 58 },
		{ 119, 28, 58 },
		{ 119, 29, 58 },
		{ 119, 30, 58 },
		{ 119, 31, 58 },
		{ 119, 32, 58 },
		{ 119, 33, 58 },
		{ 119, 34, 58 },
		{ 119, 35, 58 },
		{ 119, 36, 58 },
		{ 120, 1, 58 },
		{ 120, 2, 58 },
		{ 120, 3, 58 },
		{ 120, 4, 58 },
		{ 120, 5, 58 },
		{ 120, 6, 58 },
		{ 120, 7, 58 },
		{ 120, 8, 58 },
		{ 120, 9, 58 },
		{ 120, 10, 58 },
		{ 120, 11, 58 },
		{ 120, 12, 58 },
		{ 120, 13, 58 },
		{ 120, 14, 58 },
		{ 120, 15, 58 },
		{ 120, 16, 58 },
		{ 120, 17, 58 },
		{ 120, 18, 58 },
		{ 120, 19, 58 },
		{ 120, 20, 58 },
		{ 120, 21, 58 },
		{ 120, 22, 58 },
		{ 120, 23, 58 },
		{ 120, 24, 58 },
		{ 120, 25, 58 },
		{ 120, 26, 58 },
		{ 120, 27, 58 },
		{ 120, 28, 58 },
		{ 120, 29, 58 },
		{ 120, 30, 58 },
		{ 120, 31, 58 },
		{ 120, 32, 58 },
		{ 120, 33, 58 },
		{ 120, 34, 58 },
		{ 120, 35, 58 },
		{ 120, 36, 58 },
		{ 120, 37, 58 },
		{ 121, 1, 58 },
		{ 121, 2, 58 },
		{ 121, 3, 58 },
		{ 121, 4, 58 },
		{ 121, 5, 58 },
		{ 121, 6, 58 },
		{ 121, 7, 58 },
		{ 121, 8, 58 },
		{ 121, 9, 58 },
		{ 121, 10, 58 },
		{ 121, 11, 58 },
		{ 121, 12, 58 },
		{ 121, 13, 58 },
		{ 121, 14, 58 },
		{ 121, 15, 58 },
		{ 121, 16, 58 },
		{ 121, 17, 58 },
		{ 121, 18, 58 },
		{ 121, 19, 58 },
		{ 121, 20, 58 },
		{ 121, 21, 58 },
		{ 121, 22, 58 },
		{ 121, 23, 58 },
		{ 121, 24, 58 },
		{ 121, 25, 58 },
		{ 121, 26, 58 },
		{ 121, 27, 58 },
		{ 121, 28, 58 },
		{ 121, 29, 58 },
		{ 121, 30, 58 },
		{ 121, 31, 58 },
		{ 121, 32, 58 },
		{ 121, 33, 58 },
		{ 121, 34, 58 },
		{ 121, 35, 58 },
		{ 121, 36, 58 },
		{ 121, 37, 58 },
		{ 121, 38, 58 },
		{ 121, 39, 58 },
		{ 122, 1, 58 },
		{ 122, 2, 58 },
		{ 122, 3, 58 },
		{ 122, 4, 58 },
		{ 122, 5, 58 },
		{ 122, 6, 58 },
		{ 122, 7, 58 },
		{ 122, 8, 58 },
		{ 122, 9, 58 },
		{ 122, 10, 58 },
		{ 122, 11, 58 },
		{ 122, 12, 58 },
		{ 122, 13, 58 },
		{ 122, 14, 58 },
		{ 122, 15, 58 },
		{ 122, 16, 58 },
		{ 122, 17, 58 },
		{ 122, 18, 58 },
		{ 122, 19, 58 },
		{ 122, 20, 58 },
		{ 122, 21, 58 },
		{ 122, 22, 58 },
		{ 122, 23, 58 },
		{ 122, 24, 58 },
		{ 122, 25, 58 },
		{ 122, 26, 58 },
		{ 122, 27, 58 },
		{ 122, 28, 58 },
		{ 122, 29, 58 },
		{ 122, 30, 58 },
		{ 122, 31, 58 },
		{ 122, 32, 58 },
		{ 122, 33, 58 },
		{ 122, 34, 58 },
		{ 122, 35, 58 },
		{ 122, 36, 58 },
		{ 122, 37, 58 },
		{ 122, 38, 58 },
		{ 122, 39, 58 },
		{ 122, 40, 58 },
		{ 123, 1, 58 },
		{ 123, 2, 58 },
		{ 123, 3, 58 },
		{ 123, 4, 58 },
		{ 123, 5, 58 },
		{ 123, 6, 58 },
		{ 123, 7, 58 },
		{ 123, 8, 58 },
		{ 123, 9, 58 },
		{ 123, 10, 58 },
		{ 123, 11, 58 },
		{ 123, 12, 58 },
		{ 123, 13, 58 },
		{ 123, 14, 58 },
		{ 123, 15, 58 },
		{ 123, 16, 58 },
		{ 123, 17, 58 },
		{ 123, 18, 58 },
		{ 123, 19, 58 },
		{ 123, 20, 58 },
		{ 123, 21, 58 },
		{ 123, 22, 58 },
		{ 123, 23, 58 },
		{ 123, 24, 58 },
		{ 123, 25, 58 },
		{ 123, 26, 58 },
		{ 123, 27, 58 },
		{ 123, 28, 58 },
		{ 123, 29, 58 },
		{ 123, 30, 58 },
		{ 123, 31, 58 },
		{ 123, 32, 58 },
		{ 123, 33, 58 },
		{ 123, 34, 58 },
		{ 123, 35, 58 },
		{ 123, 36, 58 },
		{ 123, 37, 58 },
		{ 123, 38, 58 },
		{ 123, 39, 58 },
		{ 123, 40, 58 },
		{ 123, 41, 58 },
		{ 123, 42, 58 },
		{ 124, 1, 58 },
		{ 124, 2, 58 },
		{ 124, 3, 58 },
		{ 124, 4, 58 },
		{ 124, 5, 58 },
		{ 124, 6, 58 },
		{ 124, 7, 58 },
		{ 124, 8, 58 },
		{ 124, 9, 58 },
		{ 124, 10, 58 },
		{ 124, 11, 58 },
		{ 124, 12, 58 },
		{ 124, 13, 58 },
		{ 124, 14, 58 },
		{ 124, 15, 58 },
		{ 124, 16, 58 },
		{ 124, 17, 58 },
		{ 124, 18, 58 },
		{ 124, 19, 58 },
		{ 124, 20, 58 },
		{ 124, 21, 58 },
		{ 124, 22, 58 },
		{ 124, 23, 58 },
		{ 124, 24, 58 },
		{ 124, 25, 58 },
		{ 124, 26, 58 },
		{ 124, 27, 58 },
		{ 124, 28, 58 },
		{ 124, 29, 58 },
		{ 124, 30, 58 },
		{ 124, 31, 58 },
		{ 124, 32, 58 },
		{ 124, 33, 58 },
		{ 124, 34, 58 },
		{ 124, 35, 58 },
		{ 124, 36, 58 },
		{ 124, 37, 58 },
		{ 124, 38, 58 },
		{ 124, 39, 58 },
		{ 124, 40, 58 },
		{ 124, 41, 58 },
		{ 124, 42, 58 },
		{ 124, 43, 58 },
		{ 124, 44, 58 },
		{ 125, 1, 58 },
		{ 125, 2, 58 },
		{ 125, 3, 58 },
		{ 125, 4, 58 },
		{ 125, 5, 58 },
		{ 125, 6, 58 },
		{ 125, 7, 58 },
		{ 125, 8, 58 },
		{ 125, 9, 58 },
		{ 125, 10, 58 },
		{ 125, 11, 58 },
		{ 125, 12, 58 },
		{ 125, 13, 58 },
		{ 125, 14, 58 },
		{ 125, 15, 58 },
		{ 125, 16, 58 },
		{ 125, 17, 58 },
		{ 125, 18, 58 },
		{ 125, 19, 58 },
		{ 125, 20, 58 },
		{ 125, 21, 58 },
		{ 125, 22, 58 },
		{ 125, 23, 58 },
		{ 125, 24, 58 },
		{ 125, 25, 58 },
		{ 125, 26, 58 },
		{ 125, 27, 58 },
		{ 125, 28, 58 },
		{ 125, 29, 58 },
		{ 125, 30, 58 },
		{ 125, 31, 58 },
		{ 125, 32, 58 },
		{ 125, 33, 58 },
		{ 125, 34, 58 },
		{ 125, 35, 58 },
		{ 125, 36, 58 },
		{ 125, 37, 58 },
		{ 125, 38, 58 },
		{ 125, 39, 58 },
		{ 125, 40, 58 },
		{ 125, 41, 58 },
		{ 125, 42, 58 },
		{ 125, 43, 58 },
		{ 125, 44, 58 },
		{ 125, 45, 58 },
		{ 126, 1, 58 },
		{ 126, 2, 58 },
		{ 126, 3, 58 },
		{ 126, 4, 58 },
		{ 126, 5, 58 },
		{ 126, 6, 58 },
		{ 126, 7, 58 },
		{ 126, 8, 58 },
		{ 126, 9, 58 },
		{ 126, 10, 58 },
		{ 126, 11, 58 },
		{ 126, 12, 58 },
		{ 126, 13, 58 },
		{ 126, 14, 58 },
		{ 126, 15, 58 },
		{ 126, 16, 58 },
		{ 126, 17, 58 },
		{ 126, 18, 58 },
		{ 126, 19, 58 },
		{ 126, 20, 58 },
		{ 126, 21, 58 },
		{ 126, 22, 58 },
		{ 126, 23, 58 },
		{ 126, 24, 58 },
		{ 126, 25, 58 },
		{ 126, 26, 58 },
		{ 126, 27, 58 },
		{ 126, 28, 58 },
		{ 126, 29, 58 },
		{ 126, 30, 58 },
		{ 126, 31, 58 },
		{ 126, 32, 58 },
		{ 126, 33, 58 },
		{ 126, 34, 58 },
		{ 126, 35, 58 },
		{ 126, 36, 58 },
		{ 126, 37, 58 },
		{ 126, 38, 58 },
		{ 126, 39, 58 },
		{ 126, 40, 58 },
		{ 126, 41, 58 },
		{ 126, 42, 58 },
		{ 126, 43, 58 },
		{ 126, 44, 58 },
		{ 126, 45, 58 },
		{ 126, 46, 58 },
		{ 126, 47, 58 },
		{ 127, 1, 58 },
		{ 127, 2, 58 },
		{ 127, 3, 58 },
		{ 127, 4, 58 },
		{ 127, 5, 58 },
		{ 127, 6, 58 },
		{ 127, 7, 58 },
		{ 127, 8, 58 },
		{ 127, 9, 58 },
		{ 127, 10, 58 },
		{ 127, 11, 58 },
		{ 127, 12, 58 },
		{ 127, 13, 58 },
		{ 127, 14, 58 },
		{ 127, 15, 58 },
		{ 127, 16, 58 },
		{ 127, 17, 58 },
		{ 127, 18, 58 },
		{ 127, 19, 58 },
		{ 127, 20, 58 },
		{ 127, 21, 58 },
		{ 127, 22, 58 },
		{ 127, 23, 58 },
		{ 127, 24, 58 },
		{ 127, 25, 58 },
		{ 127, 26, 58 },
		{ 127, 27, 58 },
		{ 127, 28, 58 },
		{ 127, 29, 58 },
		{ 127, 30, 58 },
		{ 127, 31, 58 },
		{ 127, 32, 58 },
		{ 127, 33, 58 },
		{ 127, 34, 58 },
		{ 127, 35, 58 },
		{ 127, 36, 58 },
		{ 127, 37, 58 },
		{ 127, 38, 58 },
		{ 127, 39, 58 },
		{ 127, 40, 58 },
		{ 127, 41, 58 },
		{ 127, 42, 58 },
		{ 127, 43, 58 },
		{ 127, 44, 58 },
		{ 127, 45, 58 },
		{ 127, 46, 58 },
		{ 127, 47, 58 },
		{ 127, 48, 58 },
		{ 128, 1, 58 },
		{ 128, 2, 58 },
		{ 128, 3, 58 },
		{ 128, 4, 58 },
		{ 128, 5, 58 },
		{ 128, 6, 58 },
		{ 128, 7, 58 },
		{ 128, 8, 58 },
		{ 128, 9, 58 },
		{ 128, 10, 58 },
		{ 128, 11, 58 },
		{ 128, 12, 58 },
		{ 128, 13, 58 },
		{ 128, 14, 58 },
		{ 128, 15, 58 },
		{ 128, 16, 58 },
		{ 128, 17, 58 },
		{ 128, 18, 58 },
		{ 128, 19, 58 },
		{ 128, 20, 58 },
		{ 128, 21, 58 },
		{ 128, 22, 58 },
		{ 128, 23, 58 },
		{ 128, 24, 58 },
		{ 128, 25, 58 },
		{ 128, 26, 58 },
		{ 128, 27, 58 },
		{ 128, 28, 58 },
		{ 128, 29, 58 },
		{ 128, 30, 58 },
		{ 128, 31, 58 },
		{ 128, 32, 58 },
		{ 128, 33, 58 },
		{ 128, 34, 58 },
		{ 128, 35, 58 },
		{ 128, 36, 58 },
		{ 128, 37, 58 },
		{ 128, 38, 58 },
		{ 128, 39, 58 },
		{ 128, 40, 58 },
		{ 128, 41, 58 },
		{ 128, 42, 58 },
		{ 128, 43, 58 },
		{ 128, 44, 58 },
		{ 128, 45, 58 },
		{ 128, 46, 58 },
		{ 128, 47, 58 },
		{ 128, 48, 58 },
		{ 128, 49, 58 },
		{ 128, 50, 58 },
		{ 129, 1, 58 },
		{ 129, 2, 58 },
		{ 129, 3, 58 },
		{ 129, 4, 58 },
		{ 129, 5, 58 },
		{ 129, 6, 58 },
		{ 129, 7, 58 },
		{ 129, 8, 58 },
		{ 129, 9, 58 },
		{ 129, 10, 58 },
		{ 129, 11, 58 },
		{ 129, 12, 58 },
		{ 129, 13, 58 },
		{ 129, 14, 58 },
		{ 129, 15, 58 },
		{ 129, 16, 58 },
		{ 129, 17, 58 },
		{ 129, 18, 58 },
		{ 129, 19, 58 },
		{ 129, 20, 58 },
		{ 129, 21, 58 },
		{ 129, 22, 58 },
		{ 129, 23, 58 },
		{ 129, 24, 58 },
		{ 129, 25, 58 },
		{ 129, 26, 58 },
		{ 129, 27, 58 },
		{ 129, 28, 58 },
		{ 129, 29, 58 },
		{ 129, 30, 58 },
		{ 129, 31, 58 },
		{ 129, 32, 58 },
		{ 129, 33, 58 },
		{ 129, 34, 58 },
		{ 129, 35, 58 },
		{ 129, 36, 58 },
		{ 129, 37, 58 },
		{ 129, 38, 58 },
		{ 129, 39, 58 },
		{ 129, 40, 58 },
		{ 129, 41, 58 },
		{ 129, 42, 58 },
		{ 129, 43, 58 },
		{ 129, 44, 58 },
		{ 129, 45, 58 },
		{ 129, 46, 58 },
		{ 129, 47, 58 },
		{ 129, 48, 58 },
		{ 129, 49, 58 },
		{ 129, 50, 58 },
		{ 129, 51, 58 },
		{ 129, 52, 58 },
		{ 130, 1, 58 },
		{ 130, 2, 58 },
		{ 130, 3, 58 },
		{ 130, 4, 58 },
		{ 130, 5, 58 },
		{ 130, 6, 58 },
		{ 130, 7, 58 },
		{ 130, 8, 58 },
		{ 130, 9, 58 },
		{ 130, 10, 58 },
		{ 130, 11, 58 },
		{ 130, 12, 58 },
		{ 130, 13, 58 },
		{ 130, 14, 58 },
		{ 130, 15, 58 },
		{ 130, 16, 58 },
		{ 130, 17, 58 },
		{ 130, 18, 58 },
		{ 130, 19, 58 },
		{ 130, 20, 58 },
		{ 130, 21, 58 },
		{ 130, 22, 58 },
		{ 130, 23, 58 },
		{ 130, 24, 58 },
		{ 130, 25, 58 },
		{ 130, 26, 58 },
		{ 130, 27, 58 },
		{ 130, 28, 58 },
		{ 130, 29, 58 },
		{ 130, 30, 58 },
		{ 130, 31, 58 },
		{ 130, 32, 58 },
		{ 130, 33, 58 },
		{ 130, 34, 58 },
		{ 130, 35, 58 },
		{ 130, 36, 58 },
		{ 130, 37, 58 },
		{ 130, 38, 58 },
		{ 130, 39, 58 },
		{ 130, 40, 58 },
		{ 130, 41, 58 },
		{ 130, 42, 58 },
		{ 130, 43, 58 },
		{ 130, 44, 58 },
		{ 130, 45, 58 },
		{ 130, 46, 58 },
		{ 130, 47, 58 },
		{ 130, 48, 58 },
		{ 130, 49, 58 },
		{ 130, 50, 58 },
		{ 130, 51, 58 },
		{ 130, 52, 58 },
		{ 130, 53, 58 },
		{ 131, 1, 58 },
		{ 131, 2, 58 },
		{ 131, 3, 58 },
		{ 131, 4, 58 },
		{ 131, 5, 58 },
		{ 131, 6, 58 },
		{ 131, 7, 58 },
		{ 131, 8, 58 },
		{ 131, 9, 58 },
		{ 131, 10, 58 },
		{ 131, 11, 58 },
		{ 131, 12, 58 },
		{ 131, 13, 58 },
		{ 131, 14, 58 },
		{ 131, 15, 58 },
		{ 131, 16, 58 },
		{ 131, 17, 58 },
		{ 131, 18, 58 },
		{ 131, 19, 58 },
		{ 131, 20, 58 },
		{ 131, 21, 58 },
		{ 131, 22, 58 },
		{ 131, 23, 58 },
		{ 131, 24, 58 },
		{ 131, 25, 58 },
		{ 131, 26, 58 },
		{ 131, 27, 58 },
		{ 131, 28, 58 },
		{ 131, 29, 58 },
		{ 131, 30, 58 },
		{ 131, 31, 58 },
		{ 131, 32, 58 },
		{ 131, 33, 58 },
		{ 131, 34, 58 },
		{ 131, 35, 58 },
		{ 131, 36, 58 },
		{ 131, 37, 58 },
		{ 131, 38, 58 },
		{ 131, 39, 58 },
		{ 131, 40, 58 },
		{ 131, 41, 58 },
		{ 131, 42, 58 },
		{ 131, 43, 58 },
		{ 131, 44, 58 },
		{ 131, 45, 58 },
		{ 131, 46, 58 },
		{ 131, 47, 58 },
		{ 131, 48, 58 },
		{ 131, 49, 58 },
		{ 131, 50, 58 },
		{ 131, 51, 58 },
		{ 131, 52, 58 },
		{ 131, 53, 58 },
		{ 131, 54, 58 },
		{ 131, 55, 58 },
		{ 132, 1, 58 },
		{ 132, 2, 58 },
		{ 132, 3, 58 },
		{ 132, 4, 58 },
		{ 132, 5, 58 },
		{ 132, 6, 58 },
		{ 132, 7, 58 },
		{ 132, 8, 58 },
		{ 132, 9, 58 },
		{ 132, 10, 58 },
		{ 132, 11, 58 },
		{ 132, 12, 58 },
		{ 132, 13, 58 },
		{ 132, 14, 58 },
		{ 132, 15, 58 },
		{ 132, 16, 58 },
		{ 132, 17, 58 },
		{ 132, 18, 58 },
		{ 132, 19, 58 },
		{ 132, 20, 58 },
		{ 132, 21, 58 },
		{ 132, 22, 58 },
		{ 132, 23, 58 },
		{ 132, 24, 58 },
		{ 132, 25, 58 },
		{ 132, 26, 58 },
		{ 132, 27, 58 },
		{ 132, 28, 58 },
		{ 132, 29, 58 },
		{ 132, 30, 58 },
		{ 132, 31, 58 },
		{ 132, 32, 58 },
		{ 132, 33, 58 },
		{ 132, 34, 58 },
		{ 132, 35, 58 },
		{ 132, 36, 58 },
		{ 132, 37, 58 },
		{ 132, 38, 58 },
		{ 132, 39, 58 },
		{ 132, 40, 58 },
		{ 132, 41, 58 },
		{ 132, 42, 58 },
		{ 132, 43, 58 },
		{ 132, 44, 58 },
		{ 132, 45, 58 },
		{ 132, 46, 58 },
		{ 132, 47, 58 },
		{ 132, 48, 58 },
		{ 132, 49, 58 },
		{ 132, 50, 58 },
		{ 132, 51, 58 },
		{ 132, 52, 58 },
		{ 132, 53, 58 },
		{ 132, 54, 58 },
		{ 132, 55, 58 },
		{ 132, 56, 58 },
		{ 133, 1, 58 },
		{ 133, 2, 58 },
		{ 133, 3, 58 },
		{ 133, 4, 58 },
		{ 133, 5, 58 },
		{ 133, 6, 58 },
		{ 133, 7, 58 },
		{ 133, 8, 58 },
		{ 133, 9, 58 },
		{ 133, 10, 58 },
		{ 133, 11, 58 },
		{ 133, 12, 58 },
		{ 133, 13, 58 },
		{ 133, 14, 58 },
		{ 133, 15, 58 },
		{ 133, 16, 58 },
		{ 133, 17, 58 },
		{ 133, 18, 58 },
		{ 133, 19, 58 },
		{ 133, 20, 58 },
		{ 133, 21, 58 },
		{ 133, 22, 58 },
		{ 133, 23, 58 },
		{ 133, 24, 58 },
		{ 133, 25, 58 },
		{ 133, 26, 58 },
		{ 133, 27, 58 },
		{ 133, 28, 58 },
		{ 133, 29, 58 },
		{ 133, 30, 58 },
		{ 133, 31, 58 },
		{ 133, 32, 58 },
		{ 133, 33, 58 },
		{ 133, 34, 58 },
		{ 133, 35, 58 },
		{ 133, 36, 58 },
		{ 133, 37, 58 },
		{ 133, 38, 58 },
		{ 133, 39, 58 },
		{ 133, 40, 58 },
		{ 133, 41, 58 },
		{ 133, 42, 58 },
		{ 133, 43, 58 },
		{ 133, 44, 58 },
		{ 133, 45, 58 },
		{ 133, 46, 58 },
		{ 133, 47, 58 },
		{ 133, 48, 58 },
		{ 133, 49, 58 },
		{ 133, 50, 58 },
		{ 133, 51, 58 },
		{ 133, 52, 58 },
		{ 133, 53, 58 },
		{ 133, 54, 58 },
		{ 133, 55, 58 },
		{ 133, 56, 58 },
		{ 133, 57, 58 },
		{ 133, 58, 58 },
		{ 134, 1, 58 },
		{ 134, 2, 58 },
		{ 134, 3, 58 },
		{ 134, 4, 58 },
		{ 134, 5, 58 },
		{ 134, 6, 58 },
		{ 134, 7, 58 },
		{ 134, 8, 58 },
		{ 134, 9, 58 },
		{ 134, 10, 58 },
		{ 134, 11, 58 },
		{ 134, 12, 58 },
		{ 134, 13, 58 },
		{ 134, 14, 58 },
		{ 134, 15, 58 },
		{ 134, 16, 58 },
		{ 134, 17, 58 },
		{ 134, 18, 58 },
		{ 134, 19, 58 },
		{ 134, 20, 58 },
		{ 134, 21, 58 },
		{ 134, 22, 58 },
		{ 134, 23, 58 },
		{ 134, 24, 58 },
		{ 134, 25, 58 },
		{ 134, 26, 58 },
		{ 134, 27, 58 },
		{ 134, 28, 58 },
		{ 134, 29, 58 },
		{ 134, 30, 58 },
		{ 134, 31, 58 },
		{ 134, 32, 58 },
		{ 134, 33, 58 },
		{ 134, 34, 58 },
		{ 134, 35, 58 },
		{ 134, 36, 58 },
		{ 134, 37, 58 },
		{ 134, 38, 58 },
		{ 134, 39, 58 },
		{ 134, 40, 58 },
		{ 134, 41, 58 },
		{ 134, 42, 58 },
		{ 134, 43, 58 },
		{ 134, 44, 58 },
		{ 134, 45, 58 },
		{ 134, 46, 58 },
		{ 134, 47, 58 },
		{ 134, 48, 58 },
		{ 134, 49, 58 },
		{ 134, 50, 58 },
		{ 134, 51, 58 },
		{ 134, 52, 58 },
		{ 134, 53, 58 },
		{ 134, 54, 58 },
		{ 134, 55, 58 },
		{ 134, 56, 58 },
		{ 134, 57, 58 },
		{ 134, 58, 58 },
		{ 134, 59, 58 },
		{ 134, 60, 58 },
		{ 135, 1, 58 },
		{ 135, 2, 58 },
		{ 135, 3, 58 },
		{ 135, 4, 58 },
		{ 135, 5, 58 },
		{ 135, 6, 58 },
		{ 135, 7, 58 },
		{ 135, 8, 58 },
		{ 135, 9, 58 },
		{ 135, 10, 58 },
		{ 135, 11, 58 },
		{ 135, 12, 58 },
		{ 135, 13, 58 },
		{ 135, 14, 58 },
		{ 135, 15, 58 },
		{ 135, 16, 58 },
		{ 135, 17, 58 },
		{ 135, 18, 58 },
		{ 135, 19, 58 },
		{ 135, 20, 58 },
		{ 135, 21, 58 },
		{ 135, 22, 58 },
		{ 135, 23, 58 },
		{ 135, 24, 58 },
		{ 135, 25, 58 },
		{ 135, 26, 58 },
		{ 135, 27, 58 },
		{ 135, 28, 58 },
		{ 135, 29, 58 },
		{ 135, 30, 58 },
		{ 135, 31, 58 },
		{ 135, 32, 58 },
		{ 135, 33, 58 },
		{ 135, 34, 58 },
		{ 135, 35, 58 },
		{ 135, 36, 58 },
		{ 135, 37, 58 },
		{ 135, 38, 58 },
		{ 135, 39, 58 },
		{ 135, 40, 58 },
		{ 135, 41, 58 },
		{ 135, 42, 58 },
		{ 135, 43, 58 },
		{ 135, 44, 58 },
		{ 135, 45, 58 },
		{ 135, 46, 58 },
		{ 135, 47, 58 },
		{ 135, 48, 58 },
		{ 135, 49, 58 },
		{ 135, 50, 58 },
		{ 135, 51, 58 },
		{ 135, 52, 58 },
		{ 135, 53, 58 },
		{ 135, 54, 58 },
		{ 135, 55, 58 },
		{ 135, 56, 58 },
		{ 135, 57, 58 },
		{ 135, 58, 58 },
		{ 135, 59, 58 },
		{ 135, 60, 58 },
		{ 135, 61, 58 },
		{ 136, 1, 58 },
		{ 136, 2, 58 },
		{ 136, 3, 58 },
		{ 136, 4, 58 },
		{ 136, 5, 58 },
		{ 136, 6, 58 },
		{ 136, 7, 58 },
		{ 136, 8, 58 },
		{ 136, 9, 58 },
		{ 136, 10, 58 },
		{ 136, 11, 58 },
		{ 136, 12, 58 },
		{ 136, 13, 58 },
		{ 136, 14, 58 },
		{ 136, 15, 58 },
		{ 136, 16, 58 },
		{ 136, 17, 58 },
		{ 136, 18, 58 },
		{ 136, 19, 58 },
		{ 136, 20, 58 },
		{ 136, 21, 58 },
		{ 136, 22, 58 },
		{ 136, 23, 58 },
		{ 136, 24, 58 },
		{ 136, 25, 58 },
		{ 136, 26, 58 },
		{ 136, 27, 58 },
		{ 136, 28, 58 },
		{ 136, 29, 58 },
		{ 136, 30, 58 },
		{ 136, 31, 58 },
		{ 136, 32, 58 },
		{ 136, 33, 58 },
		{ 136, 34, 58 },
		{ 136, 35, 58 },
		{ 136, 36, 58 },
		{ 136, 37, 58 },
		{ 136, 38, 58 },
		{ 136, 39, 58 },
		{ 136, 40, 58 },
		{ 136, 41, 58 },
		{ 136, 42, 58 },
		{ 136, 43, 58 },
		{ 136, 44, 58 },
		{ 136, 45, 58 },
		{ 136, 46, 58 },
		{ 136, 47, 58 },
		{ 136, 48, 58 },
		{ 136, 49, 58 },
		{ 136, 50, 58 },
		{ 136, 51, 58 },
		{ 136, 52, 58 },
		{ 136, 53, 58 },
		{ 136, 54, 58 },
		{ 136, 55, 58 },
		{ 136, 56, 58 },
		{ 136, 57, 58 },
		{ 136, 58, 58 },
		{ 136, 59, 58 },
		{ 136, 60, 58 },
		{ 136, 61, 58 },
		{ 136, 62, 58 },
		{ 136, 63, 58 },
		{ 137, 1, 58 },
		{ 137, 2, 58 },
		{ 137, 3, 58 },
		{ 137, 4, 58 },
		{ 137, 5, 58 },
		{ 137, 6, 58 },
		{ 137, 7, 58 },
		{ 137, 8, 58 },
		{ 137, 9, 58 },
		{ 137, 10, 58 },
		{ 137, 11, 58 },
		{ 137, 12, 58 },
		{ 137, 13, 58 },
		{ 137, 14, 58 },
		{ 137, 15, 58 },
		{ 137, 16, 58 },
		{ 137, 17, 58 },
		{ 137, 18, 58 },
		{ 137, 19, 58 },
		{ 137, 20, 58 },
		{ 137, 21, 58 },
		{ 137, 22, 58 },
		{ 137, 23, 58 },
		{ 137, 24, 58 },
		{ 137, 25, 58 },
		{ 137, 26, 58 },
		{ 137, 27, 58 },
		{ 137, 28, 58 },
		{ 137, 29, 58 },
		{ 137, 30, 58 },
		{ 137, 31, 58 },
		{ 137, 32, 58 },
		{ 137, 33, 58 },
		{ 137, 34, 58 },
		{ 137, 35, 58 },
		{ 137, 36, 58 },
		{ 137, 37, 58 },
		{ 137, 38, 58 },
		{ 137, 39, 58 },
		{ 137, 40, 58 },
		{ 137, 41, 58 },
		{ 137, 42, 58 },
		{ 137, 43, 58 },
		{ 137, 44, 58 },
		{ 137, 45, 58 },
		{ 137, 46, 58 },
		{ 137, 47, 58 },
		{ 137, 48, 58 },
		{ 137, 49, 58 },
		{ 137, 50, 58 },
		{ 137, 51, 58 },
		{ 137, 52, 58 },
		{ 137, 53, 58 },
		{ 137, 54, 58 },
		{ 137, 55, 58 },
		{ 137, 56, 58 },
		{ 137, 57, 58 },
		{ 137, 58, 58 },
		{ 137, 59, 58 },
		{ 137, 60, 58 },
		{ 137, 61, 58 },
		{ 137, 62, 58 },
		{ 137, 63, 58 },
		{ 137, 64, 58 },
		{ 138, 1, 58 },
		{ 138, 2, 58 },
		{ 138, 3, 58 },
		{ 138, 4, 58 },
		{ 138, 5, 58 },
		{ 138, 6, 58 },
		{ 138, 7, 58 },
		{ 138, 8, 58 },
		{ 138, 9, 58 },
		{ 138, 10, 58 },
		{ 138, 11, 58 },
		{ 138, 12, 58 },
		{ 138, 13, 58 },
		{ 138, 14, 58 },
		{ 138, 15, 58 },
		{ 138, 16, 58 },
		{ 138, 17, 58 },
		{ 138, 18, 58 },
		{ 138, 19, 58 },
		{ 138, 20, 58 },
		{ 138, 21, 58 },
		{ 138, 22, 58 },
		{ 138, 23, 58 },
		{ 138, 24, 58 },
		{ 138, 25, 58 },
		{ 138, 26, 58 },
		{ 138, 27, 58 },
		{ 138, 28, 58 },
		{ 138, 29, 58 },
		{ 138, 30, 58 },
		{ 138, 31, 58 },
		{ 138, 32, 58 },
		{ 138, 33, 58 },
		{ 138, 34, 58 },
		{ 138, 35, 58 },
		{ 138, 36, 58 },
		{ 138, 37, 58 },
		{ 138, 38, 58 },
		{ 138, 39, 58 },
		{ 138, 40, 58 },
		{ 138, 41, 58 },
		{ 138, 42, 58 },
		{ 138, 43, 58 },
		{ 138, 44, 58 },
		{ 138, 45, 58 },
		{ 138, 46, 58 },
		{ 138, 47, 58 },
		{ 138, 48, 58 },
		{ 138, 49, 58 },
		{ 138, 50, 58 },
		{ 138, 51, 58 },
		{ 138, 52, 58 },
		{ 138, 53, 58 },
		{ 138, 54, 58 },
		{ 138, 55, 58 },
		{ 138, 56, 58 },
		{ 138, 57, 58 },
		{ 138, 58, 58 },
		{ 138, 59, 58 },
		{ 138, 60, 58 },
		{ 138, 61, 58 },
		{ 138, 62, 58 },
		{ 138, 63, 58 },
		{ 138, 64, 58 },
		{ 138, 65, 58 },
		{ 138, 66, 58 },
		{ 139, 1, 58 },
		{ 139, 2, 58 },
		{ 139, 3, 58 },
		{ 139, 4, 58 },
		{ 139, 5, 58 },
		{ 139, 6, 58 },
		{ 139, 7, 58 },
		{ 139, 8, 58 },
		{ 139, 9, 58 },
		{ 139, 10, 58 },
		{ 139, 11, 58 },
		{ 139, 12, 58 },
		{ 139, 13, 58 },
		{ 139, 14, 58 },
		{ 139, 15, 58 },
		{ 139, 16, 58 },
		{ 139, 17, 58 },
		{ 139, 18, 58 },
		{ 139, 19, 58 },
		{ 139, 20, 58 },
		{ 139, 21, 58 },
		{ 139, 22, 58 },
		{ 139, 23, 58 },
		{ 139, 24, 58 },
		{ 139, 25, 58 },
		{ 139, 26, 58 },
		{ 139, 27, 58 },
		{ 139, 28, 58 },
		{ 139, 29, 58 },
		{ 139, 30, 58 },
		{ 139, 31, 58 },
		{ 139, 32, 58 },
		{ 139, 33, 58 },
		{ 139, 34, 58 },
		{ 139, 35, 58 },
		{ 139, 36, 58 },
		{ 139, 37, 58 },
		{ 139, 38, 58 },
		{ 139, 39, 58 },
		{ 139, 40, 58 },
		{ 139, 41, 58 },
		{ 139, 42, 58 },
		{ 139, 43, 58 },
		{ 139, 44, 58 },
		{ 139, 45, 58 },
		{ 139, 46, 58 },
		{ 139, 47, 58 },
		{ 139, 48, 58 },
		{ 139, 49, 58 },
		{ 139, 50, 58 },
		{ 139, 51, 58 },
		{ 139, 52, 58 },
		{ 139, 53, 58 },
		{ 139, 54, 58 },
		{ 139, 55, 58 },
		{ 139, 56, 58 },
		{ 139, 57, 58 },
		{ 139, 58, 58 },
		{ 139, 59, 58 },
		{ 139, 60, 58 },
		{ 139, 61, 58 },
		{ 139, 62, 58 },
		{ 139, 63, 58 },
		{ 139, 64, 58 },
		{ 139, 65, 58 },
		{ 139, 66, 58 },
		{ 139, 67, 58 },
		{ 139, 68, 58 },
		{ 140, 1, 58 },
		{ 140, 2, 58 },
		{ 140, 3, 58 },
		{ 140, 4, 58 },
		{ 140, 5, 58 },
		{ 140, 6, 58 },
		{ 140, 7, 58 },
		{ 140, 8, 58 },
		{ 140, 9, 58 },
		{ 140, 10, 58 },
		{ 140, 11, 58 },
		{ 140, 12, 58 },
		{ 140, 13, 58 },
		{ 140, 14, 58 },
		{ 140, 15, 58 },
		{ 140, 16, 58 },
		{ 140, 17, 58 },
		{ 140, 18, 58 },
		{ 140, 19, 58 },
		{ 140, 20, 58 },
		{ 140, 21, 58 },
		{ 140, 22, 58 },
		{ 140, 23, 58 },
		{ 140, 24, 58 },
		{ 140, 25, 58 },
		{ 140, 26, 58 },
		{ 140, 27, 58 },
		{ 140, 28, 58 },
		{ 140, 29, 58 },
		{ 140, 30, 58 },
		{ 140, 31, 58 },
		{ 140, 32, 58 },
		{ 140, 33, 58 },
		{ 140, 34, 58 },
		{ 140, 35, 58 },
		{ 140, 36, 58 },
		{ 140, 37, 58 },
		{ 140, 38, 58 },
		{ 140, 39, 58 },
		{ 140, 40, 58 },
		{ 140, 41, 58 },
		{ 140, 42, 58 },
		{ 140, 43, 58 },
		{ 140, 44, 58 },
		{ 140, 45, 58 },
		{ 140, 46, 58 },
		{ 140, 47, 58 },
		{ 140, 48, 58 },
		{ 140, 49, 58 },
		{ 140, 50, 58 },
		{ 140, 51, 58 },
		{ 140, 52, 58 },
		{ 140, 53, 58 },
		{ 140, 54, 58 },
		{ 140, 55, 58 },
		{ 140, 56, 58 },
		{ 140, 57, 58 },
		{ 140, 58, 58 },
		{ 140, 59, 58 },
		{ 140, 60, 58 },
		{ 140, 61, 58 },
		{ 140, 62, 58 },
		{ 140, 63, 58 },
		{ 140, 64, 58 },
		{ 140, 65, 58 },
		{ 140, 66, 58 },
		{ 140, 67, 58 },
		{ 140, 68, 58 },
		{ 140, 69, 58 },
		{ 141, 1, 58 },
		{ 141, 2, 58 },
		{ 141, 3, 58 },
		{ 141, 4, 58 },
		{ 141, 5, 58 },
		{ 141, 6, 58 },
		{ 141, 7, 58 },
		{ 141, 8, 58 },
		{ 141, 9, 58 },
		{ 141, 10, 58 },
		{ 141, 11, 58 },
		{ 141, 12, 58 },
		{ 141, 13, 58 },
		{ 141, 14, 58 },
		{ 141, 15, 58 },
		{ 141, 16, 58 },
		{ 141, 17, 58 },
		{ 141, 18, 58 },
		{ 141, 19, 58 },
		{ 141, 20, 58 },
		{ 141, 21, 58 },
		{ 141, 22, 58 },
		{ 141, 23, 58 },
		{ 141, 24, 58 },
		{ 141, 25, 58 },
		{ 141, 26, 58 },
		{ 141, 27, 58 },
		{ 141, 28, 58 },
		{ 141, 29, 58 },
		{ 141, 30, 58 },
		{ 141, 31, 58 },
		{ 141, 32, 58 },
		{ 141, 33, 58 },
		{ 141, 34, 58 },
		{ 141, 35, 58 },
		{ 141, 36, 58 },
		{ 141, 37, 58 },
		{ 141, 38, 58 },
		{ 141, 39, 58 },
		{ 141, 40, 58 },
		{ 141, 41, 58 },
		{ 141, 42, 58 },
		{ 141, 43, 58 },
		{ 141, 44, 58 },
		{ 141, 45, 58 },
		{ 141, 46, 58 },
		{ 141, 47, 58 },
		{ 141, 48, 58 },
		{ 141, 49, 58 },
		{ 141, 50, 58 },
		{ 141, 51, 58 },
		{ 141, 52, 58 },
		{ 141, 53, 58 },
		{ 141, 54, 58 },
		{ 141, 55, 58 },
		{ 141, 56, 58 },
		{ 141, 57, 58 },
		{ 141, 58, 58 },
		{ 141, 59, 58 },
		{ 141, 60, 58 },
		{ 141, 61, 58 },
		{ 141, 62, 58 },
		{ 141, 63, 58 },
		{ 141, 64, 58 },
		{ 141, 65, 58 },
		{ 141, 66, 58 },
		{ 141, 67, 58 },
		{ 141, 68, 58 },
		{ 141, 69, 58 },
		{ 141, 70, 58 },
		{ 141, 71, 58 },
		{ 142, 1, 58 },
		{ 142, 2, 58 },
		{ 142, 3, 58 },
		{ 142, 4, 58 },
		{ 142, 5, 58 },
		{ 142, 6, 58 },
		{ 142, 7, 58 },
		{ 142, 8, 58 },
		{ 142, 9, 58 },
		{ 142, 10, 58 },
		{ 142, 11, 58 },
		{ 142, 12, 58 },
		{ 142, 13, 58 },
		{ 142, 14, 58 },
		{ 142, 15, 58 },
		{ 142, 16, 58 },
		{ 142, 17, 58 },
		{ 142, 18, 58 },
		{ 142, 19, 58 },
		{ 142, 20, 58 },
		{ 142, 21, 58 },
		{ 142, 22, 58 },
		{ 142, 23, 58 },
		{ 142, 24, 58 },
		{ 142, 25, 58 },
		{ 142, 26, 58 },
		{ 142, 27, 58 },
		{ 142, 28, 58 },
		{ 142, 29, 58 },
		{ 142, 30, 58 },
		{ 142, 31, 58 },
		{ 142, 32, 58 },
		{ 142, 33, 58 },
		{ 142, 34, 58 },
		{ 142, 35, 58 },
		{ 142, 36, 58 },
		{ 142, 37, 58 },
		{ 142, 38, 58 },
		{ 142, 39, 58 },
		{ 142, 40, 58 },
		{ 142, 41, 58 },
		{ 142, 42, 58 },
		{ 142, 43, 58 },
		{ 142, 44, 58 },
		{ 142, 45, 58 },
		{ 142, 46, 58 },
		{ 142, 47, 58 },
		{ 142, 48, 58 },
		{ 142, 49, 58 },
		{ 142, 50, 58 },
		{ 142, 51, 58 },
		{ 142, 52, 58 },
		{ 142, 53, 58 },
		{ 142, 54, 58 },
		{ 142, 55, 58 },
		{ 142, 56, 58 },
		{ 142, 57, 58 },
		{ 142, 58, 58 },
		{ 142, 59, 58 },
		{ 142, 60, 58 },
		{ 142, 61, 58 },
		{ 142, 62, 58 },
		{ 142, 63, 58 },
		{ 142, 64, 58 },
		{ 142, 65, 58 },
		{ 142, 66, 58 },
		{ 142, 67, 58 },
		{ 142, 68, 58 },
		{ 142, 69, 58 },
		{ 142, 70, 58 },
		{ 142, 71, 58 },
		{ 142, 72, 58 },
		{ 143, 1, 58 },
		{ 143, 2, 58 },
		{ 143, 3, 58 },
		{ 143, 4, 58 },
		{ 143, 5, 58 },
		{ 143, 6, 58 },
		{ 143, 7, 58 },
		{ 143, 8, 58 },
		{ 143, 9, 58 },
		{ 143, 10, 58 },
		{ 143, 11, 58 },
		{ 143, 12, 58 },
		{ 143, 13, 58 },
		{ 143, 14, 58 },
		{ 143, 15, 58 },
		{ 143, 16, 58 },
		{ 143, 17, 58 },
		{ 143, 18, 58 },
		{ 143, 19, 58 },
		{ 143, 20, 58 },
		{ 143, 21, 58 },
		{ 143, 22, 58 },
		{ 143, 23, 58 },
		{ 143, 24, 58 },
		{ 143, 25, 58 },
		{ 143, 26, 58 },
		{ 143, 27, 58 },
		{ 143, 28, 58 },
		{ 143, 29, 58 },
		{ 143, 30, 58 },
		{ 143, 31, 58 },
		{ 143, 32, 58 },
		{ 143, 33, 58 },
		{ 143, 34, 58 },
		{ 143, 35, 58 },
		{ 143, 36, 58 },
		{ 143, 37, 58 },
		{ 143, 38, 58 },
		{ 143, 39, 58 },
		{ 143, 40, 58 },
		{ 143, 41, 58 },
		{ 143, 42, 58 },
		{ 143, 43, 58 },
		{ 143, 44, 58 },
		{ 143, 45, 58 },
		{ 143, 46, 58 },
		{ 143, 47, 58 },
		{ 143, 48, 58 },
		{ 143, 49, 58 },
		{ 143, 50, 58 },
		{ 143, 51, 58 },
		{ 143, 52, 58 },
		{ 143, 53, 58 },
		{ 143, 54, 58 },
		{ 143, 55, 58 },
		{ 143, 56, 58 },
		{ 143, 57, 58 },
		{ 143, 58, 58 },
		{ 143, 59, 58 },
		{ 143, 60, 58 },
		{ 143, 61, 58 },
		{ 143, 62, 58 },
		{ 143, 63, 58 },
		{ 143, 64, 58 },
		{ 143, 65, 58 },
		{ 143, 66, 58 },
		{ 143, 67, 58 },
		{ 143, 68, 58 },
		{ 143, 69, 58 },
		{ 143, 70, 58 },
		{ 143, 71, 58 },
		{ 143, 72, 58 },
		{ 143, 73, 58 },
		{ 143, 74, 58 },
		{ 144, 1, 58 },
		{ 144, 2, 58 },
		{ 144, 3, 58 },
		{ 144, 4, 58 },
		{ 144, 5, 58 },
		{ 144, 6, 58 },
		{ 144, 7, 58 },
		{ 144, 8, 58 },
		{ 144, 9, 58 },
		{ 144, 10, 58 },
		{ 144, 11, 58 },
		{ 144, 12, 58 },
		{ 144, 13, 58 },
		{ 144, 14, 58 },
		{ 144, 15, 58 },
		{ 144, 16, 58 },
		{ 144, 17, 58 },
		{ 144, 18, 58 },
		{ 144, 19, 58 },
		{ 144, 20, 58 },
		{ 144, 21, 58 },
		{ 144, 22, 58 },
		{ 144, 23, 58 },
		{ 144, 24, 58 },
		{ 144, 25, 58 },
		{ 144, 26, 58 },
		{ 144, 27, 58 },
		{ 144, 28, 58 },
		{ 144, 29, 58 },
		{ 144, 30, 58 },
		{ 144, 31, 58 },
		{ 144, 32, 58 },
		{ 144, 33, 58 },
		{ 144, 34, 58 },
		{ 144, 35, 58 },
		{ 144, 36, 58 },
		{ 144, 37, 58 },
		{ 144, 38, 58 },
		{ 144, 39, 58 },
		{ 144, 40, 58 },
		{ 144, 41, 58 },
		{ 144, 42, 58 },
		{ 144, 43, 58 },
		{ 144, 44, 58 },
		{ 144, 45, 58 },
		{ 144, 46, 58 },
		{ 144, 47, 58 },
		{ 144, 48, 58 },
		{ 144, 49, 58 },
		{ 144, 50, 58 },
		{ 144, 51, 58 },
		{ 144, 52, 58 },
		{ 144, 53, 58 },
		{ 144, 54, 58 },
		{ 144, 55, 58 },
		{ 144, 56, 58 },
		{ 144, 57, 58 },
		{ 144, 58, 58 },
		{ 144, 59, 58 },
		{ 144, 60, 58 },
		{ 144, 61, 58 },
		{ 144, 62, 58 },
		{ 144, 63, 58 },
		{ 144, 64, 58 },
		{ 144, 65, 58 },
		{ 144, 66, 58 },
		{ 144, 67, 58 },
		{ 144, 68, 58 },
		{ 144, 69, 58 },
		{ 144, 70, 58 },
		{ 144, 71, 58 },
		{ 144, 72, 58 },
		{ 144, 73, 58 },
		{ 144, 74, 58 },
		{ 144, 75, 58 },
		{ 144, 76, 58 },
		{ 145, 1, 58 },
		{ 145, 2, 58 },
		{ 145, 3, 58 },
		{ 145, 4, 58 },
		{ 145, 5, 58 },
		{ 145, 6, 58 },
		{ 145, 7, 58 },
		{ 145, 8, 58 },
		{ 145, 9, 58 },
		{ 145, 10, 58 },
		{ 145, 11, 58 },
		{ 145, 12, 58 },
		{ 145, 13, 58 },
		{ 145, 14, 58 },
		{ 145, 15, 58 },
		{ 145, 16, 58 },
		{ 145, 17, 58 },
		{ 145, 18, 58 },
		{ 145, 19, 58 },
		{ 145, 20, 58 },
		{ 145, 21, 58 },
		{ 145, 22, 58 },
		{ 145, 23, 58 },
		{ 145, 24, 58 },
		{ 145, 25, 58 },
		{ 145, 26, 58 },
		{ 145, 27, 58 },
		{ 145, 28, 58 },
		{ 145, 29, 58 },
		{ 145, 30, 58 },
		{ 145, 31, 58 },
		{ 145, 32, 58 },
		{ 145, 33, 58 },
		{ 145, 34, 58 },
		{ 145, 35, 58 },
		{ 145, 36, 58 },
		{ 145, 37, 58 },
		{ 145, 38, 58 },
		{ 145, 39, 58 },
		{ 145, 40, 58 },
		{ 145, 41, 58 },
		{ 145, 42, 58 },
		{ 145, 43, 58 },
		{ 145, 44, 58 },
		{ 145, 45, 58 },
		{ 145, 46, 58 },
		{ 145, 47, 58 },
		{ 145, 48, 58 },
		{ 145, 49, 58 },
		{ 145, 50, 58 },
		{ 145, 51, 58 },
		{ 145, 52, 58 },
		{ 145, 53, 58 },
		{ 145, 54, 58 },
		{ 145, 55, 58 },
		{ 145, 56, 58 },
		{ 145, 57, 58 },
		{ 145, 58, 58 },
		{ 145, 59, 58 },
		{ 145, 60, 58 },
		{ 145, 61, 58 },
		{ 145, 62, 58 },
		{ 145, 63, 58 },
		{ 145, 64, 58 },
		{ 145, 65, 58 },
		{ 145, 66, 58 },
		{ 145, 67, 58 },
		{ 145, 68, 58 },
		{ 145, 69, 58 },
		{ 145, 70, 58 },
		{ 145, 71, 58 },
		{ 145, 72, 58 },
		{ 145, 73, 58 },
		{ 145, 74, 58 },
		{ 145, 75, 58 },
		{ 145, 76, 58 },
		{ 145, 77, 58 },
		{ 146, 1, 58 },
		{ 146, 2, 58 },
		{ 146, 3, 58 },
		{ 146, 4, 58 },
		{ 146, 5, 58 },
		{ 146, 6, 58 },
		{ 146, 7, 58 },
		{ 146, 8, 58 },
		{ 146, 9, 58 },
		{ 146, 10, 58 },
		{ 146, 11, 58 },
		{ 146, 12, 58 },
		{ 146, 13, 58 },
		{ 146, 14, 58 },
		{ 146, 15, 58 },
		{ 146, 16, 58 },
		{ 146, 17, 58 },
		{ 146, 18, 58 },
		{ 146, 19, 58 },
		{ 146, 20, 58 },
		{ 146, 21, 58 },
		{ 146, 22, 58 },
		{ 146, 23, 58 },
		{ 146, 24, 58 },
		{ 146, 25, 58 },
		{ 146, 26, 58 },
		{ 146, 27, 58 },
		{ 146, 28, 58 },
		{ 146, 29, 58 },
		{ 146, 30, 58 },
		{ 146, 31, 58 },
		{ 146, 32, 58 },
		{ 146, 33, 58 },
		{ 146, 34, 58 },
		{ 146, 35, 58 },
		{ 146, 36, 58 },
		{ 146, 37, 58 },
		{ 146, 38, 58 },
		{ 146, 39, 58 },
		{ 146, 40, 58 },
		{ 146, 41, 58 },
		{ 146, 42, 58 },
		{ 146, 43, 58 },
		{ 146, 44, 58 },
		{ 146, 45, 58 },
		{ 146, 46, 58 },
		{ 146, 47, 58 },
		{ 146, 48, 58 },
		{ 146, 49, 58 },
		{ 146, 50, 58 },
		{ 146, 51, 58 },
		{ 146, 52, 58 },
		{ 146, 53, 58 },
		{ 146, 54, 58 },
		{ 146, 55, 58 },
		{ 146, 56, 58 },
		{ 146, 57, 58 },
		{ 146, 58, 58 },
		{ 146, 59, 58 },
		{ 146, 60, 58 },
		{ 146, 61, 58 },
		{ 146, 62, 58 },
		{ 146, 63, 58 },
		{ 146, 64, 58 },
		{ 146, 65, 58 },
		{ 146, 66, 58 },
		{ 146, 67, 58 },
		{ 146, 68, 58 },
		{ 146, 69, 58 },
		{ 146, 70, 58 },
		{ 146, 71, 58 },
		{ 146, 72, 58 },
		{ 146, 73, 58 },
		{ 146, 74, 58 },
		{ 146, 75, 58 },
		{ 146, 76, 58 },
		{ 146, 77, 58 },
		{ 146, 78, 58 },
		{ 146, 79, 58 },
		{ 147, 1, 58 },
		{ 147, 2, 58 },
		{ 147, 3, 58 },
		{ 147, 4, 58 },
		{ 147, 5, 58 },
		{ 147, 6, 58 },
		{ 147, 7, 58 },
		{ 147, 8, 58 },
		{ 147, 9, 58 },
		{ 147, 10, 58 },
		{ 147, 11, 58 },
		{ 147, 12, 58 },
		{ 147, 13, 58 },
		{ 147, 14, 58 },
		{ 147, 15, 58 },
		{ 147, 16, 58 },
		{ 147, 17, 58 },
		{ 147, 18, 58 },
		{ 147, 19, 58 },
		{ 147, 20, 58 },
		{ 147, 21, 58 },
		{ 147, 22, 58 },
		{ 147, 23, 58 },
		{ 147, 24, 58 },
		{ 147, 25, 58 },
		{ 147, 26, 58 },
		{ 147, 27, 58 },
		{ 147, 28, 58 },
		{ 147, 29, 58 },
		{ 147, 30, 58 },
		{ 147, 31, 58 },
		{ 147, 32, 58 },
		{ 147, 33, 58 },
		{ 147, 34, 58 },
		{ 147, 35, 58 },
		{ 147, 36, 58 },
		{ 147, 37, 58 },
		{ 147, 38, 58 },
		{ 147, 39, 58 },
		{ 147, 40, 58 },
		{ 147, 41, 58 },
		{ 147, 42, 58 },
		{ 147, 43, 58 },
		{ 147, 44, 58 },
		{ 147, 45, 58 },
		{ 147, 46, 58 },
		{ 147, 47, 58 },
		{ 147, 48, 58 },
		{ 147, 49, 58 },
		{ 147, 50, 58 },
		{ 147, 51, 58 },
		{ 147, 52, 58 },
		{ 147, 53, 58 },
		{ 147, 54, 58 },
		{ 147, 55, 58 },
		{ 147, 56, 58 },
		{ 147, 57, 58 },
		{ 147, 58, 58 },
		{ 147, 59, 58 },
		{ 147, 60, 58 },
		{ 147, 61, 58 },
		{ 147, 62, 58 },
		{ 147, 63, 58 },
		{ 147, 64, 58 },
		{ 147, 65, 58 },
		{ 147, 66, 58 },
		{ 147, 67, 58 },
		{ 147, 68, 58 },
		{ 147, 69, 58 },
		{ 147, 70, 58 },
		{ 147, 71, 58 },
		{ 147, 72, 58 },
		{ 147, 73, 58 },
		{ 147, 74, 58 },
		{ 147, 75, 58 },
		{ 147, 76, 58 },
		{ 147, 77, 58 },
		{ 147, 78, 58 },
		{ 147, 79, 58 },
		{ 147, 80, 58 },
		{ 148, 1, 58 },
		{ 148, 2, 58 },
		{ 148, 3, 58 },
		{ 148, 4, 58 },
		{ 148, 5, 58 },
		{ 148, 6, 58 },
		{ 148, 7, 58 },
		{ 148, 8, 58 },
		{ 148, 9, 58 },
		{ 148, 10, 58 },
		{ 148, 11, 58 },
		{ 148, 12, 58 },
		{ 148, 13, 58 },
		{ 148, 14, 58 },
		{ 148, 15, 58 },
		{ 148, 16, 58 },
		{ 148, 17, 58 },
		{ 148, 18, 58 },
		{ 148, 19, 58 },
		{ 148, 20, 58 },
		{ 148, 21, 58 },
		{ 148, 22, 58 },
		{ 148, 23, 58 },
		{ 148, 24, 58 },
		{ 148, 25, 58 },
		{ 148, 26, 58 },
		{ 148, 27, 58 },
		{ 148, 28, 58 },
		{ 148, 29, 58 },
		{ 148, 30, 58 },
		{ 148, 31, 58 },
		{ 148, 32, 58 },
		{ 148, 33, 58 },
		{ 148, 34, 58 },
		{ 148, 35, 58 },
		{ 148, 36, 58 },
		{ 148, 37, 58 },
		{ 148, 38, 58 },
		{ 148, 39, 58 },
		{ 148, 40, 58 },
		{ 148, 41, 58 },
		{ 148, 42, 58 },
		{ 148, 43, 58 },
		{ 148, 44, 58 },
		{ 148, 45, 58 },
		{ 148, 46, 58 },
		{ 148, 47, 58 },
		{ 148, 48, 58 },
		{ 148, 49, 58 },
		{ 148, 50, 58 },
		{ 148, 51, 58 },
		{ 148, 52, 58 },
		{ 148, 53, 58 },
		{ 148, 54, 58 },
		{ 148, 55, 58 },
		{ 148, 56, 58 },
		{ 148, 57, 58 },
		{ 148, 58, 58 },
		{ 148, 59, 58 },
		{ 148, 60, 58 },
		{ 148, 61, 58 },
		{ 148, 62, 58 },
		{ 148, 63, 58 },
		{ 148, 64, 58 },
		{ 148, 65, 58 },
		{ 148, 66, 58 },
		{ 148, 67, 58 },
		{ 148, 68, 58 },
		{ 148, 69, 58 },
		{ 148, 70, 58 },
		{ 148, 71, 58 },
		{ 148, 72, 58 },
		{ 148, 73, 58 },
		{ 148, 74, 58 },
		{ 148, 75, 58 },
		{ 148, 76, 58 },
		{ 148, 77, 58 },
		{ 148, 78, 58 },
		{ 148, 79, 58 },
		{ 148, 80, 58 },
		{ 148, 81, 58 },
		{ 148, 82, 58 },
		{ 149, 1, 58 },
		{ 149, 2, 58 },
		{ 149, 3, 58 },
		{ 149, 4, 58 },
		{ 149, 5, 58 },
		{ 149, 6, 58 },
		{ 149, 7, 58 },
		{ 149, 8, 58 },
		{ 149, 9, 58 },
		{ 149, 10, 58 },
		{ 149, 11, 58 },
		{ 149, 12, 58 },
		{ 149, 13, 58 },
		{ 149, 14, 58 },
		{ 149, 15, 58 },
		{ 149, 16, 58 },
		{ 149, 17, 58 },
		{ 149, 18, 58 },
		{ 149, 19, 58 },
		{ 149, 20, 58 },
		{ 149, 21, 58 },
		{ 149, 22, 58 },
		{ 149, 23, 58 },
		{ 149, 24, 58 },
		{ 149, 25, 58 },
		{ 149, 26, 58 },
		{ 149, 27, 58 },
		{ 149, 28, 58 },
		{ 149, 29, 58 },
		{ 149, 30, 58 },
		{ 149, 31, 58 },
		{ 149, 32, 58 },
		{ 149, 33, 58 },
		{ 149, 34, 58 },
		{ 149, 35, 58 },
		{ 149, 36, 58 },
		{ 149, 37, 58 },
		{ 149, 38, 58 },
		{ 149, 39, 58 },
		{ 149, 40, 58 },
		{ 149, 41, 58 },
		{ 149, 42, 58 },
		{ 149, 43, 58 },
		{ 149, 44, 58 },
		{ 149, 45, 58 },
		{ 149, 46, 58 },
		{ 149, 47, 58 },
		{ 149, 48, 58 },
		{ 149, 49, 58 },
		{ 149, 50, 58 },
		{ 149, 51, 58 },
		{ 149, 52, 58 },
		{ 149, 53, 58 },
		{ 149, 54, 58 },
		{ 149, 55, 58 },
		{ 149, 56, 58 },
		{ 149, 57, 58 },
		{ 149, 58, 58 },
		{ 149, 59, 58 },
		{ 149, 60, 58 },
		{ 149, 61, 58 },
		{ 149, 62, 58 },
		{ 149, 63, 58 },
		{ 149, 64, 58 },
		{ 149, 65, 58 },
		{ 149, 66, 58 },
		{ 149, 67, 58 },
		{ 149, 68, 58 },
		{ 149, 69, 58 },
		{ 149, 70, 58 },
		{ 149, 71, 58 },
		{ 149, 72, 58 },
		{ 149, 73, 58 },
		{ 149, 74, 58 },
		{ 149, 75, 58 },
		{ 149, 76, 58 },
		{ 149, 77, 58 },
		{ 149, 78, 58 },
		{ 149, 79, 58 },
		{ 149, 80, 58 },
		{ 149, 81, 58 },
		{ 149, 82, 58 },
		{ 149, 83, 58 },
		{ 149, 84, 58 },
		{ 150, 1, 58 },
		{ 150, 2, 58 },
		{ 150, 3, 58 },
		{ 150, 4, 58 },
		{ 150, 5, 58 },
		{ 150, 6, 58 },
		{ 150, 7, 58 },
		{ 150, 8, 58 },
		{ 150, 9, 58 },
		{ 150, 10, 58 },
		{ 150, 11, 58 },
		{ 150, 12, 58 },
		{ 150, 13, 58 },
		{ 150, 14, 58 },
		{ 150, 15, 58 },
		{ 150, 16, 58 },
		{ 150, 17, 58 },
		{ 150, 18, 58 },
		{ 150, 19, 58 },
		{ 150, 20, 58 },
		{ 150, 21, 58 },
		{ 150, 22, 58 },
		{ 150, 23, 58 },
		{ 150, 24, 58 },
		{ 150, 25, 58 },
		{ 150, 26, 58 },
		{ 150, 27, 58 },
		{ 150, 28, 58 },
		{ 150, 29, 58 },
		{ 150, 30, 58 },
		{ 150, 31, 58 },
		{ 150, 32, 58 },
		{ 150, 33, 58 },
		{ 150, 34, 58 },
		{ 150, 35, 58 },
		{ 150, 36, 58 },
		{ 150, 37, 58 },
		{ 150, 38, 58 },
		{ 150, 39, 58 },
		{ 150, 40, 58 },
		{ 150, 41, 58 },
		{ 150, 42, 58 },
		{ 150, 43, 58 },
		{ 150, 44, 58 },
		{ 150, 45, 58 },
		{ 150, 46, 58 },
		{ 150, 47, 58 },
		{ 150, 48, 58 },
		{ 150, 49, 58 },
		{ 150, 50, 58 },
		{ 150, 51, 58 },
		{ 150, 52, 58 },
		{ 150, 53, 58 },
		{ 150, 54, 58 },
		{ 150, 55, 58 },
		{ 150, 56, 58 },
		{ 150, 57, 58 },
		{ 150, 58, 58 },
		{ 150, 59, 58 },
		{ 150, 60, 58 },
		{ 150, 61, 58 },
		{ 150, 62, 58 },
		{ 150, 63, 58 },
		{ 150, 64, 58 },
		{ 150, 65, 58 },
		{ 150, 66, 58 },
		{ 150, 67, 58 },
		{ 150, 68, 58 },
		{ 150, 69, 58 },
		{ 150, 70, 58 },
		{ 150, 71, 58 },
		{ 150, 72, 58 },
		{ 150, 73, 58 },
		{ 150, 74, 58 },
		{ 150, 75, 58 },
		{ 150, 76, 58 },
		{ 150, 77, 58 },
		{ 150, 78, 58 },
		{ 150, 79, 58 },
		{ 150, 80, 58 },
		{ 150, 81, 58 },
		{ 150, 82, 58 },
		{ 150, 83, 58 },
		{ 150, 84, 58 },
		{ 150, 85, 58 },
		{ 151, 1, 58 },
		{ 151, 2, 58 },
		{ 151, 3, 58 },
		{ 151, 4, 58 },
		{ 151, 5, 58 },
		{ 151, 6, 58 },
		{ 151, 7, 58 },
		{ 151, 8, 58 },
		{ 151, 9, 58 },
		{ 151, 10, 58 },
		{ 151, 11, 58 },
		{ 151, 12, 58 },
		{ 151, 13, 58 },
		{ 151, 14, 58 },
		{ 151, 15, 58 },
		{ 151, 16, 58 },
		{ 151, 17, 58 },
		{ 151, 18, 58 },
		{ 151, 19, 58 },
		{ 151, 20, 58 },
		{ 151, 21, 58 },
		{ 151, 22, 58 },
		{ 151, 23, 58 },
		{ 151, 24, 58 },
		{ 151, 25, 58 },
		{ 151, 26, 58 },
		{ 151, 27, 58 },
		{ 151, 28, 58 },
		{ 151, 29, 58 },
		{ 151, 30, 58 },
		{ 151, 31, 58 },
		{ 151, 32, 58 },
		{ 151, 33, 58 },
		{ 151, 34, 58 },
		{ 151, 35, 58 },
		{ 151, 36, 58 },
		{ 151, 37, 58 },
		{ 151, 38, 58 },
		{ 151, 39, 58 },
		{ 151, 40, 58 },
		{ 151, 41, 58 },
		{ 151, 42, 58 },
		{ 151, 43, 58 },
		{ 151, 44, 58 },
		{ 151, 45, 58 },
		{ 151, 46, 58 },
		{ 151, 47, 58 },
		{ 151, 48, 58 },
		{ 151, 49, 58 },
		{ 151, 50, 58 },
		{ 151, 51, 58 },
		{ 151, 52, 58 },
		{ 151, 53, 58 },
		{ 151, 54, 58 },
		{ 151, 55, 58 },
		{ 151, 56, 58 },
		{ 151, 57, 58 },
		{ 151, 58, 58 },
		{ 151, 59, 58 },
		{ 151, 60, 58 },
		{ 151, 61, 58 },
		{ 151, 62, 58 },
		{ 151, 63, 58 },
		{ 151, 64, 58 },
		{ 151, 65, 58 },
		{ 151, 66, 58 },
		{ 151, 67, 58 },
		{ 151, 68, 58 },
		{ 151, 69, 58 },
		{ 151, 70, 58 },
		{ 151, 71, 58 },
		{ 151, 72, 58 },
		{ 151, 73, 58 },
		{ 151, 74, 58 },
		{ 151, 75, 58 },
		{ 151, 76, 58 },
		{ 151, 77, 58 },
		{ 151, 78, 58 },
		{ 151, 79, 58 },
		{ 151, 80, 58 },
		{ 151, 81, 58 },
		{ 151, 82, 58 },
		{ 151, 83, 58 },
		{ 151, 84, 58 },
		{ 151, 85, 58 },
		{ 151, 86, 58 },
		{ 151, 87, 58 },
		{ 152, 1, 58 },
		{ 152, 2, 58 },
		{ 152, 3, 58 },
		{ 152, 4, 58 },
		{ 152, 5, 58 },
		{ 152, 6, 58 },
		{ 152, 7, 58 },
		{ 152, 8, 58 },
		{ 152, 9, 58 },
		{ 152, 10, 58 },
		{ 152, 11, 58 },
		{ 152, 12, 58 },
		{ 152, 13, 58 },
		{ 152, 14, 58 },
		{ 152, 15, 58 },
		{ 152, 16, 58 },
		{ 152, 17, 58 },
		{ 152, 18, 58 },
		{ 152, 19, 58 },
		{ 152, 20, 58 },
		{ 152, 21, 58 },
		{ 152, 22, 58 },
		{ 152, 23, 58 },
		{ 152, 24, 58 },
		{ 152, 25, 58 },
		{ 152, 26, 58 },
		{ 152, 27, 58 },
		{ 152, 28, 58 },
		{ 152, 29, 58 },
		{ 152, 30, 58 },
		{ 152, 31, 58 },
		{ 152, 32, 58 },
		{ 152, 33, 58 },
		{ 152, 34, 58 },
		{ 152, 35, 58 },
		{ 152, 36, 58 },
		{ 152, 37, 58 },
		{ 152, 38, 58 },
		{ 152, 39, 58 },
		{ 152, 40, 58 },
		{ 152, 41, 58 },
		{ 152, 42, 58 },
		{ 152, 43, 58 },
		{ 152, 44, 58 },
		{ 152, 45, 58 },
		{ 152, 46, 58 },
		{ 152, 47, 58 },
		{ 152, 48, 58 },
		{ 152, 49, 58 },
		{ 152, 50, 58 },
		{ 152, 51, 58 },
		{ 152, 52, 58 },
		{ 152, 53, 58 },
		{ 152, 54, 58 },
		{ 152, 55, 58 },
		{ 152, 56, 58 },
		{ 152, 57, 58 },
		{ 152, 58, 58 },
		{ 152, 59, 58 },
		{ 152, 60, 58 },
		{ 152, 61, 58 },
		{ 152, 62, 58 },
		{ 152, 63, 58 },
		{ 152, 64, 58 },
		{ 152, 65, 58 },
		{ 152, 66, 58 },
		{ 152, 67, 58 },
		{ 152, 68, 58 },
		{ 152, 69, 58 },
		{ 152, 70, 58 },
		{ 152, 71, 58 },
		{ 152, 72, 58 },
		{ 152, 73, 58 },
		{ 152, 74, 58 },
		{ 152, 75, 58 },
		{ 152, 76, 58 },
		{ 152, 77, 58 },
		{ 152, 78, 58 },
		{ 152, 79, 58 },
		{ 152, 80, 58 },
		{ 152, 81, 58 },
		{ 152, 82, 58 },
		{ 152, 83, 58 },
		{ 152, 84, 58 },
		{ 152, 85, 58 },
		{ 152, 86, 58 },
		{ 152, 87, 58 },
		{ 152, 88, 58 },
		{ 153, 1, 58 },
		{ 153, 2, 58 },
		{ 153, 3, 58 },
		{ 153, 4, 58 },
		{ 153, 5, 58 },
		{ 153, 6, 58 },
		{ 153, 7, 58 },
		{ 153, 8, 58 },
		{ 153, 9, 58 },
		{ 153, 10, 58 },
		{ 153, 11, 58 },
		{ 153, 12, 58 },
		{ 153, 13, 58 },
		{ 153, 14, 58 },
		{ 153, 15, 58 },
		{ 153, 16, 58 },
		{ 153, 17, 58 },
		{ 153, 18, 58 },
		{ 153, 19, 58 },
		{ 153, 20, 58 },
		{ 153, 21, 58 },
		{ 153, 22, 58 },
		{ 153, 23, 58 },
		{ 153, 24, 58 },
		{ 153, 25, 58 },
		{ 153, 26, 58 },
		{ 153, 27, 58 },
		{ 153, 28, 58 },
		{ 153, 29, 58 },
		{ 153, 30, 58 },
		{ 153, 31, 58 },
		{ 153, 32, 58 },
		{ 153, 33, 58 },
		{ 153, 34, 58 },
		{ 153, 35, 58 },
		{ 153, 36, 58 },
		{ 153, 37, 58 },
		{ 153, 38, 58 },
		{ 153, 39, 58 },
		{ 153, 40, 58 },
		{ 153, 41, 58 },
		{ 153, 42, 58 },
		{ 153, 43, 58 },
		{ 153, 44, 58 },
		{ 153, 45, 58 },
		{ 153, 46, 58 },
		{ 153, 47, 58 },
		{ 153, 48, 58 },
		{ 153, 49, 58 },
		{ 153, 50, 58 },
		{ 153, 51, 58 },
		{ 153, 52, 58 },
		{ 153, 53, 58 },
		{ 153, 54, 58 },
		{ 153, 55, 58 },
		{ 153, 56, 58 },
		{ 153, 57, 58 },
		{ 153, 58, 58 },
		{ 153, 59, 58 },
		{ 153, 60, 58 },
		{ 153, 61, 58 },
		{ 153, 62, 58 },
		{ 153, 63, 58 },
		{ 153, 64, 58 },
		{ 153, 65, 58 },
		{ 153, 66, 58 },
		{ 153, 67, 58 },
		{ 153, 68, 58 },
		{ 153, 69, 58 },
		{ 153, 70, 58 },
		{ 153, 71, 58 },
		{ 153, 72, 58 },
		{ 153, 73, 58 },
		{ 153, 74, 58 },
		{ 153, 75, 58 },
		{ 153, 76, 58 },
		{ 153, 77, 58 },
		{ 153, 78, 58 },
		{ 153, 79, 58 },
		{ 153, 80, 58 },
		{ 153, 81, 58 },
		{ 153, 82, 58 },
		{ 153, 83, 58 },
		{ 153, 84, 58 },
		{ 153, 85, 58 },
		{ 153, 86, 58 },
		{ 153, 87, 58 },
		{ 153, 88, 58 },
		{ 153, 89, 58 },
		{ 153, 90, 58 },
		{ 154, 1, 58 },
		{ 154, 2, 58 },
		{ 154, 3, 58 },
		{ 154, 4, 58 },
		{ 154, 5, 58 },
		{ 154, 6, 58 },
		{ 154, 7, 58 },
		{ 154, 8, 58 },
		{ 154, 9, 58 },
		{ 154, 10, 58 },
		{ 154, 11, 58 },
		{ 154, 12, 58 },
		{ 154, 13, 58 },
		{ 154, 14, 58 },
		{ 154, 15, 58 },
		{ 154, 16, 58 },
		{ 154, 17, 58 },
		{ 154, 18, 58 },
		{ 154, 19, 58 },
		{ 154, 20, 58 },
		{ 154, 21, 58 },
		{ 154, 22, 58 },
		{ 154, 23, 58 },
		{ 154, 24, 58 },
		{ 154, 25, 58 },
		{ 154, 26, 58 },
		{ 154, 27, 58 },
		{ 154, 28, 58 },
		{ 154, 29, 58 },
		{ 154, 30, 58 },
		{ 154, 31, 58 },
		{ 154, 32, 58 },
		{ 154, 33, 58 },
		{ 154, 34, 58 },
		{ 154, 35, 58 },
		{ 154, 36, 58 },
		{ 154, 37, 58 },
		{ 154, 38, 58 },
		{ 154, 39, 58 },
		{ 154, 40, 58 },
		{ 154, 41, 58 },
		{ 154, 42, 58 },
		{ 154, 43, 58 },
		{ 154, 44, 58 },
		{ 154, 45, 58 },
		{ 154, 46, 58 },
		{ 154, 47, 58 },
		{ 154, 48, 58 },
		{ 154, 49, 58 },
		{ 154, 50, 58 },
		{ 154, 51, 58 },
		{ 154, 52, 58 },
		{ 154, 53, 58 },
		{ 154, 54, 58 },
		{ 154, 55, 58 },
		{ 154, 56, 58 },
		{ 154, 57, 58 },
		{ 154, 58, 58 },
		{ 154, 59, 58 },
		{ 154, 60, 58 },
		{ 154, 61, 58 },
		{ 154, 62, 58 },
		{ 154, 63, 58 },
		{ 154, 64, 58 },
		{ 154, 65, 58 },
		{ 154, 66, 58 },
		{ 154, 67, 58 },
		{ 154, 68, 58 },
		{ 154, 69, 58 },
		{ 154, 70, 58 },
		{ 154, 71, 58 },
		{ 154, 72, 58 },
		{ 154, 73, 58 },
		{ 154, 74, 58 },
		{ 154, 75, 58 },
		{ 154, 76, 58 },
		{ 154, 77, 58 },
		{ 154, 78, 58 },
		{ 154, 79, 58 },
		{ 154, 80, 58 },
		{ 154, 81, 58 },
		{ 154, 82, 58 },
		{ 154, 83, 58 },
		{ 154, 84, 58 },
		{ 154, 85, 58 },
		{ 154, 86, 58 },
		{ 154, 87, 58 },
		{ 154, 88, 58 },
		{ 154, 89, 58 },
		{ 154, 90, 58 },
		{ 154, 91, 58 },
		{ 154, 92, 58 },
		{ 155, 1, 58 },
		{ 155, 2, 58 },
		{ 155, 3, 58 },
		{ 155, 4, 58 },
		{ 155, 5, 58 },
		{ 155, 6, 58 },
		{ 155, 7, 58 },
		{ 155, 8, 58 },
		{ 155, 9, 58 },
		{ 155, 10, 58 },
		{ 155, 11, 58 },
		{ 155, 12, 58 },
		{ 155, 13, 58 },
		{ 155, 14, 58 },
		{ 155, 15, 58 },
		{ 155, 16, 58 },
		{ 155, 17, 58 },
		{ 155, 18, 58 },
		{ 155, 19, 58 },
		{ 155, 20, 58 },
		{ 155, 21, 58 },
		{ 155, 22, 58 },
		{ 155, 23, 58 },
		{ 155, 24, 58 },
		{ 155, 25, 58 },
		{ 155, 26, 58 },
		{ 155, 27, 58 },
		{ 155, 28, 58 },
		{ 155, 29, 58 },
		{ 155, 30, 58 },
		{ 155, 31, 58 },
		{ 155, 32, 58 },
		{ 155, 33, 58 },
		{ 155, 34, 58 },
		{ 155, 35, 58 },
		{ 155, 36, 58 },
		{ 155, 37, 58 },
		{ 155, 38, 58 },
		{ 155, 39, 58 },
		{ 155, 40, 58 },
		{ 155, 41, 58 },
		{ 155, 42, 58 },
		{ 155, 43, 58 },
		{ 155, 44, 58 },
		{ 155, 45, 58 },
		{ 155, 46, 58 },
		{ 155, 47, 58 },
		{ 155, 48, 58 },
		{ 155, 49, 58 },
		{ 155, 50, 58 },
		{ 155, 51, 58 },
		{ 155, 52, 58 },
		{ 155, 53, 58 },
		{ 155, 54, 58 },
		{ 155, 55, 58 },
		{ 155, 56, 58 },
		{ 155, 57, 58 },
		{ 155, 58, 58 },
		{ 155, 59, 58 },
		{ 155, 60, 58 },
		{ 155, 61, 58 },
		{ 155, 62, 58 },
		{ 155, 63, 58 },
		{ 155, 64, 58 },
		{ 155, 65, 58 },
		{ 155, 66, 58 },
		{ 155, 67, 58 },
		{ 155, 68, 58 },
		{ 155, 69, 58 },
		{ 155, 70, 58 },
		{ 155, 71, 58 },
		{ 155, 72, 58 },
		{ 155, 73, 58 },
		{ 155, 74, 58 },
		{ 155, 75, 58 },
		{ 155, 76, 58 },
		{ 155, 77, 58 },
		{ 155, 78, 58 },
		{ 155, 79, 58 },
		{ 155, 80, 58 },
		{ 155, 81, 58 },
		{ 155, 82, 58 },
		{ 155, 83, 58 },
		{ 155, 84, 58 },
		{ 155, 85, 58 },
		{ 155, 86, 58 },
		{ 155, 87, 58 },
		{ 155, 88, 58 },
		{ 155, 89, 58 },
		{ 155, 90, 58 },
		{ 155, 91, 58 },
		{ 155, 92, 58 },
		{ 155, 93, 58 },
		{ 156, 1, 58 },
		{ 156, 2, 58 },
		{ 156, 3, 58 },
		{ 156, 4, 58 },
		{ 156, 5, 58 },
		{ 156, 6, 58 },
		{ 156, 7, 58 },
		{ 156, 8, 58 },
		{ 156, 9, 58 },
		{ 156, 10, 58 },
		{ 156, 11, 58 },
		{ 156, 12, 58 },
		{ 156, 13, 58 },
		{ 156, 14, 58 },
		{ 156, 15, 58 },
		{ 156, 16, 58 },
		{ 156, 17, 58 },
		{ 156, 18, 58 },
		{ 156, 19, 58 },
		{ 156, 20, 58 },
		{ 156, 21, 58 },
		{ 156, 22, 58 },
		{ 156, 23, 58 },
		{ 156, 24, 58 },
		{ 156, 25, 58 },
		{ 156, 26, 58 },
		{ 156, 27, 58 },
		{ 156, 28, 58 },
		{ 156, 29, 58 },
		{ 156, 30, 58 },
		{ 156, 31, 58 },
		{ 156, 32, 58 },
		{ 156, 33, 58 },
		{ 156, 34, 58 },
		{ 156, 35, 58 },
		{ 156, 36, 58 },
		{ 156, 37, 58 },
		{ 156, 38, 58 },
		{ 156, 39, 58 },
		{ 156, 40, 58 },
		{ 156, 41, 58 },
		{ 156, 42, 58 },
		{ 156, 43, 58 },
		{ 156, 44, 58 },
		{ 156, 45, 58 },
		{ 156, 46, 58 },
		{ 156, 47, 58 },
		{ 156, 48, 58 },
		{ 156, 49, 58 },
		{ 156, 50, 58 },
		{ 156, 51, 58 },
		{ 156, 52, 58 },
		{ 156, 53, 58 },
		{ 156, 54, 58 },
		{ 156, 55, 58 },
		{ 156, 56, 58 },
		{ 156, 57, 58 },
		{ 156, 58, 58 },
		{ 156, 59, 58 },
		{ 156, 60, 58 },
		{ 156, 61, 58 },
		{ 156, 62, 58 },
		{ 156, 63, 58 },
		{ 156, 64, 58 },
		{ 156, 65, 58 },
		{ 156, 66, 58 },
		{ 156, 67, 58 },
		{ 156, 68, 58 },
		{ 156, 69, 58 },
		{ 156, 70, 58 },
		{ 156, 71, 58 },
		{ 156, 72, 58 },
		{ 156, 73, 58 },
		{ 156, 74, 58 },
		{ 156, 75, 58 },
		{ 156, 76, 58 },
		{ 156, 77, 58 },
		{ 156, 78, 58 },
		{ 156, 79, 58 },
		{ 156, 80, 58 },
		{ 156, 81, 58 },
		{ 156, 82, 58 },
		{ 156, 83, 58 },
		{ 156, 84, 58 },
		{ 156, 85, 58 },
		{ 156, 86, 58 },
		{ 156, 87, 58 },
		{ 156, 88, 58 },
		{ 156, 89, 58 },
		{ 156, 90, 58 },
		{ 156, 91, 58 },
		{ 156, 92, 58 },
		{ 156, 93, 58 },
		{ 156, 94, 58 },
		{ 156, 95, 58 },
		{ 157, 1, 58 },
		{ 157, 2, 58 },
		{ 157, 3, 58 },
		{ 157, 4, 58 },
		{ 157, 5, 58 },
		{ 157, 6, 58 },
		{ 157, 7, 58 },
		{ 157, 8, 58 },
		{ 157, 9, 58 },
		{ 157, 10, 58 },
		{ 157, 11, 58 },
		{ 157, 12, 58 },
		{ 157, 13, 58 },
		{ 157, 14, 58 },
		{ 157, 15, 58 },
		{ 157, 16, 58 },
		{ 157, 17, 58 },
		{ 157, 18, 58 },
		{ 157, 19, 58 },
		{ 157, 20, 58 },
		{ 157, 21, 58 },
		{ 157, 22, 58 },
		{ 157, 23, 58 },
		{ 157, 24, 58 },
		{ 157, 25, 58 },
		{ 157, 26, 58 },
		{ 157, 27, 58 },
		{ 157, 28, 58 },
		{ 157, 29, 58 },
		{ 157, 30, 58 },
		{ 157, 31, 58 },
		{ 157, 32, 58 },
		{ 157, 33, 58 },
		{ 157, 34, 58 },
		{ 157, 35, 58 },
		{ 157, 36, 58 },
		{ 157, 37, 58 },
		{ 157, 38, 58 },
		{ 157, 39, 58 },
		{ 157, 40, 58 },
		{ 157, 41, 58 },
		{ 157, 42, 58 },
		{ 157, 43, 58 },
		{ 157, 44, 58 },
		{ 157, 45, 58 },
		{ 157, 46, 58 },
		{ 157, 47, 58 },
		{ 157, 48, 58 },
		{ 157, 49, 58 },
		{ 157, 50, 58 },
		{ 157, 51, 58 },
		{ 157, 52, 58 },
		{ 157, 53, 58 },
		{ 157, 54, 58 },
		{ 157, 55, 58 },
		{ 157, 56, 58 },
		{ 157, 57, 58 },
		{ 157, 58, 58 },
		{ 157, 59, 58 },
		{ 157, 60, 58 },
		{ 157, 61, 58 },
		{ 157, 62, 58 },
		{ 157, 63, 58 },
		{ 157, 64, 58 },
		{ 157, 65, 58 },
		{ 157, 66, 58 },
		{ 157, 67, 58 },
		{ 157, 68, 58 },
		{ 157, 69, 58 },
		{ 157, 70, 58 },
		{ 157, 71, 58 },
		{ 157, 72, 58 },
		{ 157, 73, 58 },
		{ 157, 74, 58 },
		{ 157, 75, 58 },
		{ 157, 76, 58 },
		{ 157, 77, 58 },
		{ 157, 78, 58 },
		{ 157, 79, 58 },
		{ 157, 80, 58 },
		{ 157, 81, 58 },
		{ 157, 82, 58 },
		{ 157, 83, 58 },
		{ 157, 84, 58 },
		{ 157, 85, 58 },
		{ 157, 86, 58 },
		{ 157, 87, 58 },
		{ 157, 88, 58 },
		{ 157, 89, 58 },
		{ 157, 90, 58 },
		{ 157, 91, 58 },
		{ 157, 92, 58 },
		{ 157, 93, 58 },
		{ 157, 94, 58 },
		{ 157, 95, 58 },
		{ 157, 96, 58 },
		{ 158, 1, 58 },
		{ 158, 2, 58 },
		{ 158, 3, 58 },
		{ 158, 4, 58 },
		{ 158, 5, 58 },
		{ 158, 6, 58 },
		{ 158, 7, 58 },
		{ 158, 8, 58 },
		{ 158, 9, 58 },
		{ 158, 10, 58 },
		{ 158, 11, 58 },
		{ 158, 12, 58 },
		{ 158, 13, 58 },
		{ 158, 14, 58 },
		{ 158, 15, 58 },
		{ 158, 16, 58 },
		{ 158, 17, 58 },
		{ 158, 18, 58 },
		{ 158, 19, 58 },
		{ 158, 20, 58 },
		{ 158, 21, 58 },
		{ 158, 22, 58 },
		{ 158, 23, 58 },
		{ 158, 24, 58 },
		{ 158, 25, 58 },
		{ 158, 26, 58 },
		{ 158, 27, 58 },
		{ 158, 28, 58 },
		{ 158, 29, 58 },
		{ 158, 30, 58 },
		{ 158, 31, 58 },
		{ 158, 32, 58 },
		{ 158, 33, 58 },
		{ 158, 34, 58 },
		{ 158, 35, 58 },
		{ 158, 36, 58 },
		{ 158, 37, 58 },
		{ 158, 38, 58 },
		{ 158, 39, 58 },
		{ 158, 40, 58 },
		{ 158, 41, 58 },
		{ 158, 42, 58 },
		{ 158, 43, 58 },
		{ 158, 44, 58 },
		{ 158, 45, 58 },
		{ 158, 46, 58 },
		{ 158, 47, 58 },
		{ 158, 48, 58 },
		{ 158, 49, 58 },
		{ 158, 50, 58 },
		{ 158, 51, 58 },
		{ 158, 52, 58 },
		{ 158, 53, 58 },
		{ 158, 54, 58 },
		{ 158, 55, 58 },
		{ 158, 56, 58 },
		{ 158, 57, 58 },
		{ 158, 58, 58 },
		{ 158, 59, 58 },
		{ 158, 60, 58 },
		{ 158, 61, 58 },
		{ 158, 62, 58 },
		{ 158, 63, 58 },
		{ 158, 64, 58 },
		{ 158, 65, 58 },
		{ 158, 66, 58 },
		{ 158, 67, 58 },
		{ 158, 68, 58 },
		{ 158, 69, 58 },
		{ 158, 70, 58 },
		{ 158, 71, 58 },
		{ 158, 72, 58 },
		{ 158, 73, 58 },
		{ 158, 74, 58 },
		{ 158, 75, 58 },
		{ 158, 76, 58 },
		{ 158, 77, 58 },
		{ 158, 78, 58 },
		{ 158, 79, 58 },
		{ 158, 80, 58 },
		{ 158, 81, 58 },
		{ 158, 82, 58 },
		{ 158, 83, 58 },
		{ 158, 84, 58 },
		{ 158, 85, 58 },
		{ 158, 86, 58 },
		{ 158, 87, 58 },
		{ 158, 88, 58 },
		{ 158, 89, 58 },
		{ 158, 90, 58 },
		{ 158, 91, 58 },
		{ 158, 92, 58 },
		{ 158, 93, 58 },
		{ 158, 94, 58 },
		{ 158, 95, 58 },
		{ 158, 96, 58 },
		{ 158, 97, 58 },
		{ 158, 98, 58 },
		{ 159, 1, 58 },
		{ 159, 2, 58 },
		{ 159, 3, 58 },
		{ 159, 4, 58 },
		{ 159, 5, 58 },
		{ 159, 6, 58 },
		{ 159, 7, 58 },
		{ 159, 8, 58 },
		{ 159, 9, 58 },
		{ 159, 10, 58 },
		{ 159, 11, 58 },
		{ 159, 12, 58 },
		{ 159, 13, 58 },
		{ 159, 14, 58 },
		{ 159, 15, 58 },
		{ 159, 16, 58 },
		{ 159, 17, 58 },
		{ 159, 18, 58 },
		{ 159, 19, 58 },
		{ 159, 20, 58 },
		{ 159, 21, 58 },
		{ 159, 22, 58 },
		{ 159, 23, 58 },
		{ 159, 24, 58 },
		{ 159, 25, 58 },
		{ 159, 26, 58 },
		{ 159, 27, 58 },
		{ 159, 28, 58 },
		{ 159, 29, 58 },
		{ 159, 30, 58 },
		{ 159, 31, 58 },
		{ 159, 32, 58 },
		{ 159, 33, 58 },
		{ 159, 34, 58 },
		{ 159, 35, 58 },
		{ 159, 36, 58 },
		{ 159, 37, 58 },
		{ 159, 38, 58 },
		{ 159, 39, 58 },
		{ 159, 40, 58 },
		{ 159, 41, 58 },
		{ 159, 42, 58 },
		{ 159, 43, 58 },
		{ 159, 44, 58 },
		{ 159, 45, 58 },
		{ 159, 46, 58 },
		{ 159, 47, 58 },
		{ 159, 48, 58 },
		{ 159, 49, 58 },
		{ 159, 50, 58 },
		{ 159, 51, 58 },
		{ 159, 52, 58 },
		{ 159, 53, 58 },
		{ 159, 54, 58 },
		{ 159, 55, 58 },
		{ 159, 56, 58 },
		{ 159, 57, 58 },
		{ 159, 58, 58 },
		{ 159, 59, 58 },
		{ 159, 60, 58 },
		{ 159, 61, 58 },
		{ 159, 62, 58 },
		{ 159, 63, 58 },
		{ 159, 64, 58 },
		{ 159, 65, 58 },
		{ 159, 66, 58 },
		{ 159, 67, 58 },
		{ 159, 68, 58 },
		{ 159, 69, 58 },
		{ 159, 70, 58 },
		{ 159, 71, 58 },
		{ 159, 72, 58 },
		{ 159, 73, 58 },
		{ 159, 74, 58 },
		{ 159, 75, 58 },
		{ 159, 76, 58 },
		{ 159, 77, 58 },
		{ 159, 78, 58 },
		{ 159, 79, 58 },
		{ 159, 80, 58 },
		{ 159, 81, 58 },
		{ 159, 82, 58 },
		{ 159, 83, 58 },
		{ 159, 84, 58 },
		{ 159, 85, 58 },
		{ 159, 86, 58 },
		{ 159, 87, 58 },
		{ 159, 88, 58 },
		{ 159, 89, 58 },
		{ 159, 90, 58 },
		{ 159, 91, 58 },
		{ 159, 92, 58 },
		{ 159, 93, 58 },
		{ 159, 94, 58 },
		{ 159, 95, 58 },
		{ 159, 96, 58 },
		{ 159, 97, 58 },
		{ 159, 98, 58 },
		{ 159, 99, 58 },
		{ 159, 100, 58 },
		{ 160, 1, 58 },
		{ 160, 2, 58 },
		{ 160, 3, 58 },
		{ 160, 4, 58 },
		{ 160, 5, 58 },
		{ 160, 6, 58 },
		{ 160, 7, 58 },
		{ 160, 8, 58 },
		{ 160, 9, 58 },
		{ 160, 10, 58 },
		{ 160, 11, 58 },
		{ 160, 12, 58 },
		{ 160, 13, 58 },
		{ 160, 14, 58 },
		{ 160, 15, 58 },
		{ 160, 16, 58 },
		{ 160, 17, 58 },
		{ 160, 18, 58 },
		{ 160, 19, 58 },
		{ 160, 20, 58 },
		{ 160, 21, 58 },
		{ 160, 22, 58 },
		{ 160, 23, 58 },
		{ 160, 24, 58 },
		{ 160, 25, 58 },
		{ 160, 26, 58 },
		{ 160, 27, 58 },
		{ 160, 28, 58 },
		{ 160, 29, 58 },
		{ 160, 30, 58 },
		{ 160, 31, 58 },
		{ 160, 32, 58 },
		{ 160, 33, 58 },
		{ 160, 34, 58 },
		{ 160, 35, 58 },
		{ 160, 36, 58 },
		{ 160, 37, 58 },
		{ 160, 38, 58 },
		{ 160, 39, 58 },
		{ 160, 40, 58 },
		{ 160, 41, 58 },
		{ 160, 42, 58 },
		{ 160, 43, 58 },
		{ 160, 44, 58 },
		{ 160, 45, 58 },
		{ 160, 46, 58 },
		{ 160, 47, 58 },
		{ 160, 48, 58 },
		{ 160, 49, 58 },
		{ 160, 50, 58 },
		{ 160, 51, 58 },
		{ 160, 52, 58 },
		{ 160, 53, 58 },
		{ 160, 54, 58 },
		{ 160, 55, 58 },
		{ 160, 56, 58 },
		{ 160, 57, 58 },
		{ 160, 58, 58 },
		{ 160, 59, 58 },
		{ 160, 60, 58 },
		{ 160, 61, 58 },
		{ 160, 62, 58 },
		{ 160, 63, 58 },
		{ 160, 64, 58 },
		{ 160, 65, 58 },
		{ 160, 66, 58 },
		{ 160, 67, 58 },
		{ 160, 68, 58 },
		{ 160, 69, 58 },
		{ 160, 70, 58 },
		{ 160, 71, 58 },
		{ 160, 72, 58 },
		{ 160, 73, 58 },
		{ 160, 74, 58 },
		{ 160, 75, 58 },
		{ 160, 76, 58 },
		{ 160, 77, 58 },
		{ 160, 78, 58 },
		{ 160, 79, 58 },
		{ 160, 80, 58 },
		{ 160, 81, 58 },
		{ 160, 82, 58 },
		{ 160, 83, 58 },
		{ 160, 84, 58 },
		{ 160, 85, 58 },
		{ 160, 86, 58 },
		{ 160, 87, 58 },
		{ 160, 88, 58 },
		{ 160, 89, 58 },
		{ 160, 90, 58 },
		{ 160, 91, 58 },
		{ 160, 92, 58 },
		{ 160, 93, 58 },
		{ 160, 94, 58 },
		{ 160, 95, 58 },
		{ 160, 96, 58 },
		{ 160, 97, 58 },
		{ 160, 98, 58 },
		{ 160, 99, 58 },
		{ 160, 100, 58 },
		{ 160, 101, 58 },
		{ 161, 2, 58 },
		{ 161, 3, 58 },
		{ 161, 4, 58 },
		{ 161, 5, 58 },
		{ 161, 6, 58 },
		{ 161, 7, 58 },
		{ 161, 8, 58 },
		{ 161, 9, 58 },
		{ 161, 10, 58 },
		{ 161, 11, 58 },
		{ 161, 12, 58 },
		{ 161, 13, 58 },
		{ 161, 14, 58 },
		{ 161, 15, 58 },
		{ 161, 16, 58 },
		{ 161, 17, 58 },
		{ 161, 18, 58 },
		{ 161, 19, 58 },
		{ 161, 20, 58 },
		{ 161, 21, 58 },
		{ 161, 22, 58 },
		{ 161, 23, 58 },
		{ 161, 24, 58 },
		{ 161, 25, 58 },
		{ 161, 26, 58 },
		{ 161, 27, 58 },
		{ 161, 28, 58 },
		{ 161, 29, 58 },
		{ 161, 30, 58 },
		{ 161, 31, 58 },
		{ 161, 32, 58 },
		{ 161, 33, 58 },
		{ 161, 34, 58 },
		{ 161, 35, 58 },
		{ 161, 36, 58 },
		{ 161, 37, 58 },
		{ 161, 38, 58 },
		{ 161, 39, 58 },
		{ 161, 40, 58 },
		{ 161, 41, 58 },
		{ 161, 42, 58 },
		{ 161, 43, 58 },
		{ 161, 44, 58 },
		{ 161, 45, 58 },
		{ 161, 46, 58 },
		{ 161, 47, 58 },
		{ 161, 48, 58 },
		{ 161, 49, 58 },
		{ 161, 50, 58 },
		{ 161, 51, 58 },
		{ 161, 52, 58 },
		{ 161, 53, 58 },
		{ 161, 54, 58 },
		{ 161, 55, 58 },
		{ 161, 56, 58 },
		{ 161, 57, 58 },
		{ 161, 58, 58 },
		{ 161, 59, 58 },
		{ 161, 60, 58 },
		{ 161, 61, 58 },
		{ 161, 62, 58 },
		{ 161, 63, 58 },
		{ 161, 64, 58 },
		{ 161, 65, 58 },
		{ 161, 66, 58 },
		{ 161, 67, 58 },
		{ 161, 68, 58 },
		{ 161, 69, 58 },
		{ 161, 70, 58 },
		{ 161, 71, 58 },
		{ 161, 72, 58 },
		{ 161, 73, 58 },
		{ 161, 74, 58 },
		{ 161, 75, 58 },
		{ 161, 76, 58 },
		{ 161, 77, 58 },
		{ 161, 78, 58 },
		{ 161, 79, 58 },
		{ 161, 80, 58 },
		{ 161, 81, 58 },
		{ 161, 82, 58 },
		{ 161, 83, 58 },
		{ 161, 84, 58 },
		{ 161, 85, 58 },
		{ 161, 86, 58 },
		{ 161, 87, 58 },
		{ 161, 88, 58 },
		{ 161, 89, 58 },
		{ 161, 90, 58 },
		{ 161, 91, 58 },
		{ 161, 92, 58 },
		{ 161, 93, 58 },
		{ 161, 94, 58 },
		{ 161, 95, 58 },
		{ 161, 96, 58 },
		{ 161, 97, 58 },
		{ 161, 98, 58 },
		{ 161, 99, 58 },
		{ 161, 100, 58 },
		{ 161, 101, 58 },
		{ 161, 102, 58 },
		{ 161, 103, 58 },
		{ 162, 4, 58 },
		{ 162, 5, 58 },
		{ 162, 6, 58 },
		{ 162, 7, 58 },
		{ 162, 8, 58 },
		{ 162, 9, 58 },
		{ 162, 10, 58 },
		{ 162, 11, 58 },
		{ 162, 12, 58 },
		{ 162, 13, 58 },
		{ 162, 14, 58 },
		{ 162, 15, 58 },
		{ 162, 16, 58 },
		{ 162, 17, 58 },
		{ 162, 18, 58 },
		{ 162, 19, 58 },
		{ 162, 20, 58 },
		{ 162, 21, 58 },
		{ 162, 22, 58 },
		{ 162, 23, 58 },
		{ 162, 24, 58 },
		{ 162, 25, 58 },
		{ 162, 26, 58 },
		{ 162, 27, 58 },
		{ 162, 28, 58 },
		{ 162, 29, 58 },
		{ 162, 30, 58 },
		{ 162, 31, 58 },
		{ 162, 32, 58 },
		{ 162, 33, 58 },
		{ 162, 34, 58 },
		{ 162, 35, 58 },
		{ 162, 36, 58 },
		{ 162, 37, 58 },
		{ 162, 38, 58 },
		{ 162, 39, 58 },
		{ 162, 40, 58 },
		{ 162, 41, 58 },
		{ 162, 42, 58 },
		{ 162, 43, 58 },
		{ 162, 44, 58 },
		{ 162, 45, 58 },
		{ 162, 46, 58 },
		{ 162, 47, 58 },
		{ 162, 48, 58 },
		{ 162, 49, 58 },
		{ 162, 50, 58 },
		{ 162, 51, 58 },
		{ 162, 52, 58 },
		{ 162, 53, 58 },
		{ 162, 54, 58 },
		{ 162, 55, 58 },
		{ 162, 56, 58 },
		{ 162, 57, 58 },
		{ 162, 58, 58 },
		{ 162, 59, 58 },
		{ 162, 60, 58 },
		{ 162, 61, 58 },
		{ 162, 62, 58 },
		{ 162, 63, 58 },
		{ 162, 64, 58 },
		{ 162, 65, 58 },
		{ 162, 66, 58 },
		{ 162, 67, 58 },
		{ 162, 68, 58 },
		{ 162, 69, 58 },
		{ 162, 70, 58 },
		{ 162, 71, 58 },
		{ 162, 72, 58 },
		{ 162, 73, 58 },
		{ 162, 74, 58 },
		{ 162, 75, 58 },
		{ 162, 76, 58 },
		{ 162, 77, 58 },
		{ 162, 78, 58 },
		{ 162, 79, 58 },
		{ 162, 80, 58 },
		{ 162, 81, 58 },
		{ 162, 82, 58 },
		{ 162, 83, 58 },
		{ 162, 84, 58 },
		{ 162, 85, 58 },
		{ 162, 86, 58 },
		{ 162, 87, 58 },
		{ 162, 88, 58 },
		{ 162, 89, 58 },
		{ 162, 90, 58 },
		{ 162, 91, 58 },
		{ 162, 92, 58 },
		{ 162, 93, 58 },
		{ 162, 94, 58 },
		{ 162, 95, 58 },
		{ 162, 96, 58 },
		{ 162, 97, 58 },
		{ 162, 98, 58 },
		{ 162, 99, 58 },
		{ 162, 100, 58 },
		{ 162, 101, 58 },
		{ 162, 102, 58 },
		{ 162, 103, 58 },
		{ 162, 104, 58 },
		{ 163, 7, 58 },
		{ 163, 8, 58 },
		{ 163, 9, 58 },
		{ 163, 10, 58 },
		{ 163, 11, 58 },
		{ 163, 12, 58 },
		{ 163, 13, 58 },
		{ 163, 14, 58 },
		{ 163, 15, 58 },
		{ 163, 16, 58 },
		{ 163, 17, 58 },
		{ 163, 18, 58 },
		{ 163, 19, 58 },
		{ 163, 20, 58 },
		{ 163, 21, 58 },
		{ 163, 22, 58 },
		{ 163, 23, 58 },
		{ 163, 24, 58 },
		{ 163, 25, 58 },
		{ 163, 26, 58 },
		{ 163, 27, 58 },
		{ 163, 28, 58 },
		{ 163, 29, 58 },
		{ 163, 30, 58 },
		{ 163, 31, 58 },
		{ 163, 32, 58 },
		{ 163, 33, 58 },
		{ 163, 34, 58 },
		{ 163, 35, 58 },
		{ 163, 36, 58 },
		{ 163, 37, 58 },
		{ 163, 38, 58 },
		{ 163, 39, 58 },
		{ 163, 40, 58 },
		{ 163, 41, 58 },
		{ 163, 42, 58 },
		{ 163, 43, 58 },
		{ 163, 44, 58 },
		{ 163, 45, 58 },
		{ 163, 46, 58 },
		{ 163, 47, 58 },
		{ 163, 48, 58 },
		{ 163, 49, 58 },
		{ 163, 50, 58 },
		{ 163, 51, 58 },
		{ 163, 52, 58 },
		{ 163, 53, 58 },
		{ 163, 54, 58 },
		{ 163, 55, 58 },
		{ 163, 56, 58 },
		{ 163, 57, 58 },
		{ 163, 58, 58 },
		{ 163, 59, 58 },
		{ 163, 60, 58 },
		{ 163, 61, 58 },
		{ 163, 62, 58 },
		{ 163, 63, 58 },
		{ 163, 64, 58 },
		{ 163, 65, 58 },
		{ 163, 66, 58 },
		{ 163, 67, 58 },
		{ 163, 68, 58 },
		{ 163, 69, 58 },
		{ 163, 70, 58 },
		{ 163, 71, 58 },
		{ 163, 72, 58 },
		{ 163, 73, 58 },
		{ 163, 74, 58 },
		{ 163, 75, 58 },
		{ 163, 76, 58 },
		{ 163, 77, 58 },
		{ 163, 78, 58 },
		{ 163, 79, 58 },
		{ 163, 80, 58 },
		{ 163, 81, 58 },
		{ 163, 82, 58 },
		{ 163, 83, 58 },
		{ 163, 84, 58 },
		{ 163, 85, 58 },
		{ 163, 86, 58 },
		{ 163, 87, 58 },
		{ 163, 88, 58 },
		{ 163, 89, 58 },
		{ 163, 90, 58 },
		{ 163, 91, 58 },
		{ 163, 92, 58 },
		{ 163, 93, 58 },
		{ 163, 94, 58 },
		{ 163, 95, 58 },
		{ 163, 96, 58 },
		{ 163, 97, 58 },
		{ 163, 98, 58 },
		{ 163, 99, 58 },
		{ 163, 100, 58 },
		{ 163, 101, 58 },
		{ 163, 102, 58 },
		{ 163, 103, 58 },
		{ 163, 104, 58 },
		{ 163, 105, 58 },
		{ 163, 106, 58 },
		{ 164, 10, 58 },
		{ 164, 11, 58 },
		{ 164, 12, 58 },
		{ 164, 13, 58 },
		{ 164, 14, 58 },
		{ 164, 15, 58 },
		{ 164, 16, 58 },
		{ 164, 17, 58 },
		{ 164, 18, 58 },
		{ 164, 19, 58 },
		{ 164, 20, 58 },
		{ 164, 21, 58 },
		{ 164, 22, 58 },
		{ 164, 23, 58 },
		{ 164, 24, 58 },
		{ 164, 25, 58 },
		{ 164, 26, 58 },
		{ 164, 27, 58 },
		{ 164, 28, 58 },
		{ 164, 29, 58 },
		{ 164, 30, 58 },
		{ 164, 31, 58 },
		{ 164, 32, 58 },
		{ 164, 33, 58 },
		{ 164, 34, 58 },
		{ 164, 35, 58 },
		{ 164, 36, 58 },
		{ 164, 37, 58 },
		{ 164, 38, 58 },
		{ 164, 39, 58 },
		{ 164, 40, 58 },
		{ 164, 41, 58 },
		{ 164, 42, 58 },
		{ 164, 43, 58 },
		{ 164, 44, 58 },
		{ 164, 45, 58 },
		{ 164, 46, 58 },
		{ 164, 47, 58 },
		{ 164, 48, 58 },
		{ 164, 49, 58 },
		{ 164, 50, 58 },
		{ 164, 51, 58 },
		{ 164, 52, 58 },
		{ 164, 53, 58 },
		{ 164, 54, 58 },
		{ 164, 55, 58 },
		{ 164, 56, 58 },
		{ 164, 57, 58 },
		{ 164, 58, 58 },
		{ 164, 59, 58 },
		{ 164, 60, 58 },
		{ 164, 61, 58 },
		{ 164, 62, 58 },
		{ 164, 63, 58 },
		{ 164, 64, 58 },
		{ 164, 65, 58 },
		{ 164, 66, 58 },
		{ 164, 67, 58 },
		{ 164, 68, 58 },
		{ 164, 69, 58 },
		{ 164, 70, 58 },
		{ 164, 71, 58 },
		{ 164, 72, 58 },
		{ 164, 73, 58 },
		{ 164, 74, 58 },
		{ 164, 75, 58 },
		{ 164, 76, 58 },
		{ 164, 77, 58 },
		{ 164, 78, 58 },
		{ 164, 79, 58 },
		{ 164, 80, 58 },
		{ 164, 81, 58 },
		{ 164, 82, 58 },
		{ 164, 83, 58 },
		{ 164, 84, 58 },
		{ 164, 85, 58 },
		{ 164, 86, 58 },
		{ 164, 87, 58 },
		{ 164, 88, 58 },
		{ 164, 89, 58 },
		{ 164, 90, 58 },
		{ 164, 91, 58 },
		{ 164, 92, 58 },
		{ 164, 93, 58 },
		{ 164, 94, 58 },
		{ 164, 95, 58 },
		{ 164, 96, 58 },
		{ 164, 97, 58 },
		{ 164, 98, 58 },
		{ 164, 99, 58 },
		{ 164, 100, 58 },
		{ 164, 101, 58 },
		{ 164, 102, 58 },
		{ 164, 103, 58 },
		{ 164, 104, 58 },
		{ 164, 105, 58 },
		{ 164, 106, 58 },
		{ 164, 107, 58 },
		{ 164, 108, 58 },
		{ 165, 12, 58 },
		{ 165, 13, 58 },
		{ 165, 14, 58 },
		{ 165, 15, 58 },
		{ 165, 16, 58 },
		{ 165, 17, 58 },
		{ 165, 18, 58 },
		{ 165, 19, 58 },
		{ 165, 20, 58 },
		{ 165, 21, 58 },
		{ 165, 22, 58 },
		{ 165, 23, 58 },
		{ 165, 24, 58 },
		{ 165, 25, 58 },
		{ 165, 26, 58 },
		{ 165, 27, 58 },
		{ 165, 28, 58 },
		{ 165, 29, 58 },
		{ 165, 30, 58 },
		{ 165, 31, 58 },
		{ 165, 32, 58 },
		{ 165, 33, 58 },
		{ 165, 34, 58 },
		{ 165, 35, 58 },
		{ 165, 36, 58 },
		{ 165, 37, 58 },
		{ 165, 38, 58 },
		{ 165, 39, 58 },
		{ 165, 40, 58 },
		{ 165, 41, 58 },
		{ 165, 42, 58 },
		{ 165, 43, 58 },
		{ 165, 44, 58 },
		{ 165, 45, 58 },
		{ 165, 46, 58 },
		{ 165, 47, 58 },
		{ 165, 48, 58 },
		{ 165, 49, 58 },
		{ 165, 50, 58 },
		{ 165, 51, 58 },
		{ 165, 52, 58 },
		{ 165, 53, 58 },
		{ 165, 54, 58 },
		{ 165, 55, 58 },
		{ 165, 56, 58 },
		{ 165, 57, 58 },
		{ 165, 58, 58 },
		{ 165, 59, 58 },
		{ 165, 60, 58 },
		{ 165, 61, 58 },
		{ 165, 62, 58 },
		{ 165, 63, 58 },
		{ 165, 64, 58 },
		{ 165, 65, 58 },
		{ 165, 66, 58 },
		{ 165, 67, 58 },
		{ 165, 68, 58 },
		{ 165, 69, 58 },
		{ 165, 70, 58 },
		{ 165, 71, 58 },
		{ 165, 72, 58 },
		{ 165, 73, 58 },
		{ 165, 74, 58 },
		{ 165, 75, 58 },
		{ 165, 76, 58 },
		{ 165, 77, 58 },
		{ 165, 78, 58 },
		{ 165, 79, 58 },
		{ 165, 80, 58 },
		{ 165, 81, 58 },
		{ 165, 82, 58 },
		{ 165, 83, 58 },
		{ 165, 84, 58 },
		{ 165, 85, 58 },
		{ 165, 86, 58 },
		{ 165, 87, 58 },
		{ 165, 88, 58 },
		{ 165, 89, 58 },
		{ 165, 90, 58 },
		{ 165, 91, 58 },
		{ 165, 92, 58 },
		{ 165, 93, 58 },
		{ 165, 94, 58 },
		{ 165, 95, 58 },
		{ 165, 96, 58 },
		{ 165, 97, 58 },
		{ 165, 98, 58 },
		{ 165, 99, 58 },
		{ 165, 100, 58 },
		{ 165, 101, 58 },
		{ 165, 102, 58 },
		{ 165, 103, 58 },
		{ 165, 104, 58 },
		{ 165, 105, 58 },
		{ 165, 106, 58 },
		{ 165, 107, 58 },
		{ 165, 108, 58 },
		{ 165, 109, 58 },
		{ 166, 15, 58 },
		{ 166, 16, 58 },
		{ 166, 17, 58 },
		{ 166, 18, 58 },
		{ 166, 19, 58 },
		{ 166, 20, 58 },
		{ 166, 21, 58 },
		{ 166, 22, 58 },
		{ 166, 23, 58 },
		{ 166, 24, 58 },
		{ 166, 25, 58 },
		{ 166, 26, 58 },
		{ 166, 27, 58 },
		{ 166, 28, 58 },
		{ 166, 29, 58 },
		{ 166, 30, 58 },
		{ 166, 31, 58 },
		{ 166, 32, 58 },
		{ 166, 33, 58 },
		{ 166, 34, 58 },
		{ 166, 35, 58 },
		{ 166, 36, 58 },
		{ 166, 37, 58 },
		{ 166, 38, 58 },
		{ 166, 39, 58 },
		{ 166, 40, 58 },
		{ 166, 41, 58 },
		{ 166, 42, 58 },
		{ 166, 43, 58 },
		{ 166, 44, 58 },
		{ 166, 45, 58 },
		{ 166, 46, 58 },
		{ 166, 47, 58 },
		{ 166, 48, 58 },
		{ 166, 49, 58 },
		{ 166, 50, 58 },
		{ 166, 51, 58 },
		{ 166, 52, 58 },
		{ 166, 53, 58 },
		{ 166, 54, 58 },
		{ 166, 55, 58 },
		{ 166, 56, 58 },
		{ 166, 57, 58 },
		{ 166, 58, 58 },
		{ 166, 59, 58 },
		{ 166, 60, 58 },
		{ 166, 61, 58 },
		{ 166, 62, 58 },
		{ 166, 63, 58 },
		{ 166, 64, 58 },
		{ 166, 65, 58 },
		{ 166, 66, 58 },
		{ 166, 67, 58 },
		{ 166, 68, 58 },
		{ 166, 69, 58 },
		{ 166, 70, 58 },
		{ 166, 71, 58 },
		{ 166, 72, 58 },
		{ 166, 73, 58 },
		{ 166, 74, 58 },
		{ 166, 75, 58 },
		{ 166, 76, 58 },
		{ 166, 77, 58 },
		{ 166, 78, 58 },
		{ 166, 79, 58 },
		{ 166, 80, 58 },
		{ 166, 81, 58 },
		{ 166, 82, 58 },
		{ 166, 83, 58 },
		{ 166, 84, 58 },
		{ 166, 85, 58 },
		{ 166, 86, 58 },
		{ 166, 87, 58 },
		{ 166, 88, 58 },
		{ 166, 89, 58 },
		{ 166, 90, 58 },
		{ 166, 91, 58 },
		{ 166, 92, 58 },
		{ 166, 93, 58 },
		{ 166, 94, 58 },
		{ 166, 95, 58 },
		{ 166, 96, 58 },
		{ 166, 97, 58 },
		{ 166, 98, 58 },
		{ 166, 99, 58 },
		{ 166, 100, 58 },
		{ 166, 101, 58 },
		{ 166, 102, 58 },
		{ 166, 103, 58 },
		{ 166, 104, 58 },
		{ 166, 105, 58 },
		{ 166, 106, 58 },
		{ 166, 107, 58 },
		{ 166, 108, 58 },
		{ 166, 109, 58 },
		{ 166, 110, 58 },
		{ 166, 111, 58 },
		{ 167, 18, 58 },
		{ 167, 19, 58 },
		{ 167, 20, 58 },
		{ 167, 21, 58 },
		{ 167, 22, 58 },
		{ 167, 23, 58 },
		{ 167, 24, 58 },
		{ 167, 25, 58 },
		{ 167, 26, 58 },
		{ 167, 27, 58 },
		{ 167, 28, 58 },
		{ 167, 29, 58 },
		{ 167, 30, 58 },
		{ 167, 31, 58 },
		{ 167, 32, 58 },
		{ 167, 33, 58 },
		{ 167, 34, 58 },
		{ 167, 35, 58 },
		{ 167, 36, 58 },
		{ 167, 37, 58 },
		{ 167, 38, 58 },
		{ 167, 39, 58 },
		{ 167, 40, 58 },
		{ 167, 41, 58 },
		{ 167, 42, 58 },
		{ 167, 43, 58 },
		{ 167, 44, 58 },
		{ 167, 45, 58 },
		{ 167, 46, 58 },
		{ 167, 47, 58 },
		{ 167, 48, 58 },
		{ 167, 49, 58 },
		{ 167, 50, 58 },
		{ 167, 51, 58 },
		{ 167, 52, 58 },
		{ 167, 53, 58 },
		{ 167, 54, 58 },
		{ 167, 55, 58 },
		{ 167, 56, 58 },
		{ 167, 57, 58 },
		{ 167, 58, 58 },
		{ 167, 59, 58 },
		{ 167, 60, 58 },
		{ 167, 61, 58 },
		{ 167, 62, 58 },
		{ 167, 63, 58 },
		{ 167, 64, 58 },
		{ 167, 65, 58 },
		{ 167, 66, 58 },
		{ 167, 67, 58 },
		{ 167, 68, 58 },
		{ 167, 69, 58 },
		{ 167, 70, 58 },
		{ 167, 71, 58 },
		{ 167, 72, 58 },
		{ 167, 73, 58 },
		{ 167, 74, 58 },
		{ 167, 75, 58 },
		{ 167, 76, 58 },
		{ 167, 77, 58 },
		{ 167, 78, 58 },
		{ 167, 79, 58 },
		{ 167, 80, 58 },
		{ 167, 81, 58 },
		{ 167, 82, 58 },
		{ 167, 83, 58 },
		{ 167, 84, 58 },
		{ 167, 85, 58 },
		{ 167, 86, 58 },
		{ 167, 87, 58 },
		{ 167, 88, 58 },
		{ 167, 89, 58 },
		{ 167, 90, 58 },
		{ 167, 91, 58 },
		{ 167, 92, 58 },
		{ 167, 93, 58 },
		{ 167, 94, 58 },
		{ 167, 95, 58 },
		{ 167, 96, 58 },
		{ 167, 97, 58 },
		{ 167, 98, 58 },
		{ 167, 99, 58 },
		{ 167, 100, 58 },
		{ 167, 101, 58 },
		{ 167, 102, 58 },
		{ 167, 103, 58 },
		{ 167, 104, 58 },
		{ 167, 105, 58 },
		{ 167, 106, 58 },
		{ 167, 107, 58 },
		{ 167, 108, 58 },
		{ 167, 109, 58 },
		{ 167, 110, 58 },
		{ 167, 111, 58 },
		{ 167, 112, 58 },
		{ 168, 20, 58 },
		{ 168, 21, 58 },
		{ 168, 22, 58 },
		{ 168, 23, 58 },
		{ 168, 24, 58 },
		{ 168, 25, 58 },
		{ 168, 26, 58 },
		{ 168, 27, 58 },
		{ 168, 28, 58 },
		{ 168, 29, 58 },
		{ 168, 30, 58 },
		{ 168, 31, 58 },
		{ 168, 32, 58 },
		{ 168, 33, 58 },
		{ 168, 34, 58 },
		{ 168, 35, 58 },
		{ 168, 36, 58 },
		{ 168, 37, 58 },
		{ 168, 38, 58 },
		{ 168, 39, 58 },
		{ 168, 40, 58 },
		{ 168, 41, 58 },
		{ 168, 42, 58 },
		{ 168, 43, 58 },
		{ 168, 44, 58 },
		{ 168, 45, 58 },
		{ 168, 46, 58 },
		{ 168, 47, 58 },
		{ 168, 48, 58 },
		{ 168, 49, 58 },
		{ 168, 50, 58 },
		{ 168, 51, 58 },
		{ 168, 52, 58 },
		{ 168, 53, 58 },
		{ 168, 54, 58 },
		{ 168, 55, 58 },
		{ 168, 56, 58 },
		{ 168, 57, 58 },
		{ 168, 58, 58 },
		{ 168, 59, 58 },
		{ 168, 60, 58 },
		{ 168, 61, 58 },
		{ 168, 62, 58 },
		{ 168, 63, 58 },
		{ 168, 64, 58 },
		{ 168, 65, 58 },
		{ 168, 66, 58 },
		{ 168, 67, 58 },
		{ 168, 68, 58 },
		{ 168, 69, 58 },
		{ 168, 70, 58 },
		{ 168, 71, 58 },
		{ 168, 72, 58 },
		{ 168, 73, 58 },
		{ 168, 74, 58 },
		{ 168, 75, 58 },
		{ 168, 76, 58 },
		{ 168, 77, 58 },
		{ 168, 78, 58 },
		{ 168, 79, 58 },
		{ 168, 80, 58 },
		{ 168, 81, 58 },
		{ 168, 82, 58 },
		{ 168, 83, 58 },
		{ 168, 84, 58 },
		{ 168, 85, 58 },
		{ 168, 86, 58 },
		{ 168, 87, 58 },
		{ 168, 88, 58 },
		{ 168, 89, 58 },
		{ 168, 90, 58 },
		{ 168, 91, 58 },
		{ 168, 92, 58 },
		{ 168, 93, 58 },
		{ 168, 94, 58 },
		{ 168, 95, 58 },
		{ 168, 96, 58 },
		{ 168, 97, 58 },
		{ 168, 98, 58 },
		{ 168, 99, 58 },
		{ 168, 100, 58 },
		{ 168, 101, 58 },
		{ 168, 102, 58 },
		{ 168, 103, 58 },
		{ 168, 104, 58 },
		{ 168, 105, 58 },
		{ 168, 106, 58 },
		{ 168, 107, 58 },
		{ 168, 108, 58 },
		{ 168, 109, 58 },
		{ 168, 110, 58 },
		{ 168, 111, 58 },
		{ 168, 112, 58 },
		{ 168, 113, 58 },
		{ 168, 114, 58 },
		{ 169, 23, 58 },
		{ 169, 24, 58 },
		{ 169, 25, 58 },
		{ 169, 26, 58 },
		{ 169, 27, 58 },
		{ 169, 28, 58 },
		{ 169, 29, 58 },
		{ 169, 30, 58 },
		{ 169, 31, 58 },
		{ 169, 32, 58 },
		{ 169, 33, 58 },
		{ 169, 34, 58 },
		{ 169, 35, 58 },
		{ 169, 36, 58 },
		{ 169, 37, 58 },
		{ 169, 38, 58 },
		{ 169, 39, 58 },
		{ 169, 40, 58 },
		{ 169, 41, 58 },
		{ 169, 42, 58 },
		{ 169, 43, 58 },
		{ 169, 44, 58 },
		{ 169, 45, 58 },
		{ 169, 46, 58 },
		{ 169, 47, 58 },
		{ 169, 48, 58 },
		{ 169, 49, 58 },
		{ 169, 50, 58 },
		{ 169, 51, 58 },
		{ 169, 52, 58 },
		{ 169, 53, 58 },
		{ 169, 54, 58 },
		{ 169, 55, 58 },
		{ 169, 56, 58 },
		{ 169, 57, 58 },
		{ 169, 58, 58 },
		{ 169, 59, 58 },
		{ 169, 60, 58 },
		{ 169, 61, 58 },
		{ 169, 62, 58 },
		{ 169, 63, 58 },
		{ 169, 64, 58 },
		{ 169, 65, 58 },
		{ 169, 66, 58 },
		{ 169, 67, 58 },
		{ 169, 68, 58 },
		{ 169, 69, 58 },
		{ 169, 70, 58 },
		{ 169, 71, 58 },
		{ 169, 72, 58 },
		{ 169, 73, 58 },
		{ 169, 74, 58 },
		{ 169, 75, 58 },
		{ 169, 76, 58 },
		{ 169, 77, 58 },
		{ 169, 78, 58 },
		{ 169, 79, 58 },
		{ 169, 80, 58 },
		{ 169, 81, 58 },
		{ 169, 82, 58 },
		{ 169, 83, 58 },
		{ 169, 84, 58 },
		{ 169, 85, 58 },
		{ 169, 86, 58 },
		{ 169, 87, 58 },
		{ 169, 88, 58 },
		{ 169, 89, 58 },
		{ 169, 90, 58 },
		{ 169, 91, 58 },
		{ 169, 92, 58 },
		{ 169, 93, 58 },
		{ 169, 94, 58 },
		{ 169, 95, 58 },
		{ 169, 96, 58 },
		{ 169, 97, 58 },
		{ 169, 98, 58 },
		{ 169, 99, 58 },
		{ 169, 100, 58 },
		{ 169, 101, 58 },
		{ 169, 102, 58 },
		{ 169, 103, 58 },
		{ 169, 104, 58 },
		{ 169, 105, 58 },
		{ 169, 106, 58 },
		{ 169, 107, 58 },
		{ 169, 108, 58 },
		{ 169, 109, 58 },
		{ 169, 110, 58 },
		{ 169, 111, 58 },
		{ 169, 112, 58 },
		{ 169, 113, 58 },
		{ 169, 114, 58 },
		{ 169, 115, 58 },
		{ 169, 116, 58 },
		{ 170, 26, 58 },
		{ 170, 27, 58 },
		{ 170, 28, 58 },
		{ 170, 29, 58 },
		{ 170, 30, 58 },
		{ 170, 31, 58 },
		{ 170, 32, 58 },
		{ 170, 33, 58 },
		{ 170, 34, 58 },
		{ 170, 35, 58 },
		{ 170, 36, 58 },
		{ 170, 37, 58 },
		{ 170, 38, 58 },
		{ 170, 39, 58 },
		{ 170, 40, 58 },
		{ 170, 41, 58 },
		{ 170, 42, 58 },
		{ 170, 43, 58 },
		{ 170, 44, 58 },
		{ 170, 45, 58 },
		{ 170, 46, 58 },
		{ 170, 47, 58 },
		{ 170, 48, 58 },
		{ 170, 49, 58 },
		{ 170, 50, 58 },
		{ 170, 51, 58 },
		{ 170, 52, 58 },
		{ 170, 53, 58 },
		{ 170, 54, 58 },
		{ 170, 55, 58 },
		{ 170, 56, 58 },
		{ 170, 57, 58 },
		{ 170, 58, 58 },
		{ 170, 59, 58 },
		{ 170, 60, 58 },
		{ 170, 61, 58 },
		{ 170, 62, 58 },
		{ 170, 63, 58 },
		{ 170, 64, 58 },
		{ 170, 65, 58 },
		{ 170, 66, 58 },
		{ 170, 67, 58 },
		{ 170, 68, 58 },
		{ 170, 69, 58 },
		{ 170, 70, 58 },
		{ 170, 71, 58 },
		{ 170, 72, 58 },
		{ 170, 73, 58 },
		{ 170, 74, 58 },
		{ 170, 75, 58 },
		{ 170, 76, 58 },
		{ 170, 77, 58 },
		{ 170, 78, 58 },
		{ 170, 79, 58 },
		{ 170, 80, 58 },
		{ 170, 81, 58 },
		{ 170, 82, 58 },
		{ 170, 83, 58 },
		{ 170, 84, 58 },
		{ 170, 85, 58 },
		{ 170, 86, 58 },
		{ 170, 87, 58 },
		{ 170, 88, 58 },
		{ 170, 89, 58 },
		{ 170, 90, 58 },
		{ 170, 91, 58 },
		{ 170, 92, 58 },
		{ 170, 93, 58 },
		{ 170, 94, 58 },
		{ 170, 95, 58 },
		{ 170, 96, 58 },
		{ 170, 97, 58 },
		{ 170, 98, 58 },
		{ 170, 99, 58 },
		{ 170, 100, 58 },
		{ 170, 101, 58 },
		{ 170, 102, 58 },
		{ 170, 103, 58 },
		{ 170, 104, 58 },
		{ 170, 105, 58 },
		{ 170, 106, 58 },
		{ 170, 107, 58 },
		{ 170, 108, 58 },
		{ 170, 109, 58 },
		{ 170, 110, 58 },
		{ 170, 111, 58 },
		{ 170, 112, 58 },
		{ 170, 113, 58 },
		{ 170, 114, 58 },
		{ 170, 115, 58 },
		{ 170, 116, 58 },
		{ 170, 117, 58 },
		{ 171, 28, 58 },
		{ 171, 29, 58 },
		{ 171, 30, 58 },
		{ 171, 31, 58 },
		{ 171, 32, 58 },
		{ 171, 33, 58 },
		{ 171, 34, 58 },
		{ 171, 35, 58 },
		{ 171, 36, 58 },
		{ 171, 37, 58 },
		{ 171, 38, 58 },
		{ 171, 39, 58 },
		{ 171, 40, 58 },
		{ 171, 41, 58 },
		{ 171, 42, 58 },
		{ 171, 43, 58 },
		{ 171, 44, 58 },
		{ 171, 45, 58 },
		{ 171, 46, 58 },
		{ 171, 47, 58 },
		{ 171, 48, 58 },
		{ 171, 49, 58 },
		{ 171, 50, 58 },
		{ 171, 51, 58 },
		{ 171, 52, 58 },
		{ 171, 53, 58 },
		{ 171, 54, 58 },
		{ 171, 55, 58 },
		{ 171, 56, 58 },
		{ 171, 57, 58 },
		{ 171, 58, 58 },
		{ 171, 59, 58 },
		{ 171, 60, 58 },
		{ 171, 61, 58 },
		{ 171, 62, 58 },
		{ 171, 63, 58 },
		{ 171, 64, 58 },
		{ 171, 65, 58 },
		{ 171, 66, 58 },
		{ 171, 67, 58 },
		{ 171, 68, 58 },
		{ 171, 69, 58 },
		{ 171, 70, 58 },
		{ 171, 71, 58 },
		{ 171, 72, 58 },
		{ 171, 73, 58 },
		{ 171, 74, 58 },
		{ 171, 75, 58 },
		{ 171, 76, 58 },
		{ 171, 77, 58 },
		{ 171, 78, 58 },
		{ 171, 79, 58 },
		{ 171, 80, 58 },
		{ 171, 81, 58 },
		{ 171, 82, 58 },
		{ 171, 83, 58 },
		{ 171, 84, 58 },
		{ 171, 85, 58 },
		{ 171, 86, 58 },
		{ 171, 87, 58 },
		{ 171, 88, 58 },
		{ 171, 89, 58 },
		{ 171, 90, 58 },
		{ 171, 91, 58 },
		{ 171, 92, 58 },
		{ 171, 93, 58 },
		{ 171, 94, 58 },
		{ 171, 95, 58 },
		{ 171, 96, 58 },
		{ 171, 97, 58 },
		{ 171, 98, 58 },
		{ 171, 99, 58 },
		{ 171, 100, 58 },
		{ 171, 101, 58 },
		{ 171, 102, 58 },
		{ 171, 103, 58 },
		{ 171, 104, 58 },
		{ 171, 105, 58 },
		{ 171, 106, 58 },
		{ 171, 107, 58 },
		{ 171, 108, 58 },
		{ 171, 109, 58 },
		{ 171, 110, 58 },
		{ 171, 111, 58 },
		{ 171, 112, 58 },
		{ 171, 113, 58 },
		{ 171, 114, 58 },
		{ 171, 115, 58 },
		{ 171, 116, 58 },
		{ 171, 117, 58 },
		{ 171, 118, 58 },
		{ 171, 119, 58 },
		{ 172, 31, 58 },
		{ 172, 32, 58 },
		{ 172, 33, 58 },
		{ 172, 34, 58 },
		{ 172, 35, 58 },
		{ 172, 36, 58 },
		{ 172, 37, 58 },
		{ 172, 38, 58 },
		{ 172, 39, 58 },
		{ 172, 40, 58 },
		{ 172, 41, 58 },
		{ 172, 42, 58 },
		{ 172, 43, 58 },
		{ 172, 44, 58 },
		{ 172, 45, 58 },
		{ 172, 46, 58 },
		{ 172, 47, 58 },
		{ 172, 48, 58 },
		{ 172, 49, 58 },
		{ 172, 50, 58 },
		{ 172, 51, 58 },
		{ 172, 52, 58 },
		{ 172, 53, 58 },
		{ 172, 54, 58 },
		{ 172, 55, 58 },
		{ 172, 56, 58 },
		{ 172, 57, 58 },
		{ 172, 58, 58 },
		{ 172, 59, 58 },
		{ 172, 60, 58 },
		{ 172, 61, 58 },
		{ 172, 62, 58 },
		{ 172, 63, 58 },
		{ 172, 64, 58 },
		{ 172, 65, 58 },
		{ 172, 66, 58 },
		{ 172, 67, 58 },
		{ 172, 68, 58 },
		{ 172, 69, 58 },
		{ 172, 70, 58 },
		{ 172, 71, 58 },
		{ 172, 72, 58 },
		{ 172, 73, 58 },
		{ 172, 74, 58 },
		{ 172, 75, 58 },
		{ 172, 76, 58 },
		{ 172, 77, 58 },
		{ 172, 78, 58 },
		{ 172, 79, 58 },
		{ 172, 80, 58 },
		{ 172, 81, 58 },
		{ 172, 82, 58 },
		{ 172, 83, 58 },
		{ 172, 84, 58 },
		{ 172, 85, 58 },
		{ 172, 86, 58 },
		{ 172, 87, 58 },
		{ 172, 88, 58 },
		{ 172, 89, 58 },
		{ 172, 90, 58 },
		{ 172, 91, 58 },
		{ 172, 92, 58 },
		{ 172, 93, 58 },
		{ 172, 94, 58 },
		{ 172, 95, 58 },
		{ 172, 96, 58 },
		{ 172, 97, 58 },
		{ 172, 98, 58 },
		{ 172, 99, 58 },
		{ 172, 100, 58 },
		{ 172, 101, 58 },
		{ 172, 102, 58 },
		{ 172, 103, 58 },
		{ 172, 104, 58 },
		{ 172, 105, 58 },
		{ 172, 106, 58 },
		{ 172, 107, 58 },
		{ 172, 108, 58 },
		{ 172, 109, 58 },
		{ 172, 110, 58 },
		{ 172, 111, 58 },
		{ 172, 112, 58 },
		{ 172, 113, 58 },
		{ 172, 114, 58 },
		{ 172, 115, 58 },
		{ 172, 116, 58 },
		{ 172, 117, 58 },
		{ 172, 118, 58 },
		{ 172, 119, 58 },
		{ 172, 120, 58 },
		{ 173, 34, 58 },
		{ 173, 35, 58 },
		{ 173, 36, 58 },
		{ 173, 37, 58 },
		{ 173, 38, 58 },
		{ 173, 39, 58 },
		{ 173, 40, 58 },
		{ 173, 41, 58 },
		{ 173, 42, 58 },
		{ 173, 43, 58 },
		{ 173, 44, 58 },
		{ 173, 45, 58 },
		{ 173, 46, 58 },
		{ 173, 47, 58 },
		{ 173, 48, 58 },
		{ 173, 49, 58 },
		{ 173, 50, 58 },
		{ 173, 51, 58 },
		{ 173, 52, 58 },
		{ 173, 53, 58 },
		{ 173, 54, 58 },
		{ 173, 55, 58 },
		{ 173, 56, 58 },
		{ 173, 57, 58 },
		{ 173, 58, 58 },
		{ 173, 59, 58 },
		{ 173, 60, 58 },
		{ 173, 61, 58 },
		{ 173, 62, 58 },
		{ 173, 63, 58 },
		{ 173, 64, 58 },
		{ 173, 65, 58 },
		{ 173, 66, 58 },
		{ 173, 67, 58 },
		{ 173, 68, 58 },
		{ 173, 69, 58 },
		{ 173, 70, 58 },
		{ 173, 71, 58 },
		{ 173, 72, 58 },
		{ 173, 73, 58 },
		{ 173, 74, 58 },
		{ 173, 75, 58 },
		{ 173, 76, 58 },
		{ 173, 77, 58 },
		{ 173, 78, 58 },
		{ 173, 79, 58 },
		{ 173, 80, 58 },
		{ 173, 81, 58 },
		{ 173, 82, 58 },
		{ 173, 83, 58 },
		{ 173, 84, 58 },
		{ 173, 85, 58 },
		{ 173, 86, 58 },
		{ 173, 87, 58 },
		{ 173, 88, 58 },
		{ 173, 89, 58 },
		{ 173, 90, 58 },
		{ 173, 91, 58 },
		{ 173, 92, 58 },
		{ 173, 93, 58 },
		{ 173, 94, 58 },
		{ 173, 95, 58 },
		{ 173, 96, 58 },
		{ 173, 97, 58 },
		{ 173, 98, 58 },
		{ 173, 99, 58 },
		{ 173, 100, 58 },
		{ 173, 101, 58 },
		{ 173, 102, 58 },
		{ 173, 103, 58 },
		{ 173, 104, 58 },
		{ 173, 105, 58 },
		{ 173, 106, 58 },
		{ 173, 107, 58 },
		{ 173, 108, 58 },
		{ 173, 109, 58 },
		{ 173, 110, 58 },
		{ 173, 111, 58 },
		{ 173, 112, 58 },
		{ 173, 113, 58 },
		{ 173, 114, 58 },
		{ 173, 115, 58 },
		{ 173, 116, 58 },
		{ 173, 117, 58 },
		{ 173, 118, 58 },
		{ 173, 119, 58 },
		{ 173, 120, 58 },
		{ 173, 121, 58 },
		{ 173, 122, 58 },
		{ 174, 36, 58 },
		{ 174, 37, 58 },
		{ 174, 38, 58 },
		{ 174, 39, 58 },
		{ 174, 40, 58 },
		{ 174, 41, 58 },
		{ 174, 42, 58 },
		{ 174, 43, 58 },
		{ 174, 44, 58 },
		{ 174, 45, 58 },
		{ 174, 46, 58 },
		{ 174, 47, 58 },
		{ 174, 48, 58 },
		{ 174, 49, 58 },
		{ 174, 50, 58 },
		{ 174, 51, 58 },
		{ 174, 52, 58 },
		{ 174, 53, 58 },
		{ 174, 54, 58 },
		{ 174, 55, 58 },
		{ 174, 56, 58 },
		{ 174, 57, 58 },
		{ 174, 58, 58 },
		{ 174, 59, 58 },
		{ 174, 60, 58 },
		{ 174, 61, 58 },
		{ 174, 62, 58 },
		{ 174, 63, 58 },
		{ 174, 64, 58 },
		{ 174, 65, 58 },
		{ 174, 66, 58 },
		{ 174, 67, 58 },
		{ 174, 68, 58 },
		{ 174, 69, 58 },
		{ 174, 70, 58 },
		{ 174, 71, 58 },
		{ 174, 72, 58 },
		{ 174, 73, 58 },
		{ 174, 74, 58 },
		{ 174, 75, 58 },
		{ 174, 76, 58 },
		{ 174, 77, 58 },
		{ 174, 78, 58 },
		{ 174, 79, 58 },
		{ 174, 80, 58 },
		{ 174, 81, 58 },
		{ 174, 82, 58 },
		{ 174, 83, 58 },
		{ 174, 84, 58 },
		{ 174, 85, 58 },
		{ 174, 86, 58 },
		{ 174, 87, 58 },
		{ 174, 88, 58 },
		{ 174, 89, 58 },
		{ 174, 90, 58 },
		{ 174, 91, 58 },
		{ 174, 92, 58 },
		{ 174, 93, 58 },
		{ 174, 94, 58 },
		{ 174, 95, 58 },
		{ 174, 96, 58 },
		{ 174, 97, 58 },
		{ 174, 98, 58 },
		{ 174, 99, 58 },
		{ 174, 100, 58 },
		{ 174, 101, 58 },
		{ 174, 102, 58 },
		{ 174, 103, 58 },
		{ 174, 104, 58 },
		{ 174, 105, 58 },
		{ 174, 106, 58 },
		{ 174, 107, 58 },
		{ 174, 108, 58 },
		{ 174, 109, 58 },
		{ 174, 110, 58 },
		{ 174, 111, 58 },
		{ 174, 112, 58 },
		{ 174, 113, 58 },
		{ 174, 114, 58 },
		{ 174, 115, 58 },
		{ 174, 116, 58 },
		{ 174, 117, 58 },
		{ 174, 118, 58 },
		{ 174, 119, 58 },
		{ 174, 120, 58 },
		{ 174, 121, 58 },
		{ 174, 122, 58 },
		{ 174, 123, 58 },
		{ 174, 124, 58 },
		{ 175, 39, 58 },
		{ 175, 40, 58 },
		{ 175, 41, 58 },
		{ 175, 42, 58 },
		{ 175, 43, 58 },
		{ 175, 44, 58 },
		{ 175, 45, 58 },
		{ 175, 46, 58 },
		{ 175, 47, 58 },
		{ 175, 48, 58 },
		{ 175, 49, 58 },
		{ 175, 50, 58 },
		{ 175, 51, 58 },
		{ 175, 52, 58 },
		{ 175, 53, 58 },
		{ 175, 54, 58 },
		{ 175, 55, 58 },
		{ 175, 56, 58 },
		{ 175, 57, 58 },
		{ 175, 58, 58 },
		{ 175, 59, 58 },
		{ 175, 60, 58 },
		{ 175, 61, 58 },
		{ 175, 62, 58 },
		{ 175, 63, 58 },
		{ 175, 64, 58 },
		{ 175, 65, 58 },
		{ 175, 66, 58 },
		{ 175, 67, 58 },
		{ 175, 68, 58 },
		{ 175, 69, 58 },
		{ 175, 70, 58 },
		{ 175, 71, 58 },
		{ 175, 72, 58 },
		{ 175, 73, 58 },
		{ 175, 74, 58 },
		{ 175, 75, 58 },
		{ 175, 76, 58 },
		{ 175, 77, 58 },
		{ 175, 78, 58 },
		{ 175, 79, 58 },
		{ 175, 80, 58 },
		{ 175, 81, 58 },
		{ 175, 82, 58 },
		{ 175, 83, 58 },
		{ 175, 84, 58 },
		{ 175, 85, 58 },
		{ 175, 86, 58 },
		{ 175, 87, 58 },
		{ 175, 88, 58 },
		{ 175, 89, 58 },
		{ 175, 90, 58 },
		{ 175, 91, 58 },
		{ 175, 92, 58 },
		{ 175, 93, 58 },
		{ 175, 94, 58 },
		{ 175, 95, 58 },
		{ 175, 96, 58 },
		{ 175, 97, 58 },
		{ 175, 98, 58 },
		{ 175, 99, 58 },
		{ 175, 100, 58 },
		{ 175, 101, 58 },
		{ 175, 102, 58 },
		{ 175, 103, 58 },
		{ 175, 104, 58 },
		{ 175, 105, 58 },
		{ 175, 106, 58 },
		{ 175, 107, 58 },
		{ 175, 108, 58 },
		{ 175, 109, 58 },
		{ 175, 110, 58 },
		{ 175, 111, 58 },
		{ 175, 112, 58 },
		{ 175, 113, 58 },
		{ 175, 114, 58 },
		{ 175, 115, 58 },
		{ 175, 116, 58 },
		{ 175, 117, 58 },
		{ 175, 118, 58 },
		{ 175, 119, 58 },
		{ 175, 120, 58 },
		{ 175, 121, 58 },
		{ 175, 122, 58 },
		{ 175, 123, 58 },
		{ 175, 124, 58 },
		{ 175, 125, 58 },
		{ 176, 42, 58 },
		{ 176, 43, 58 },
		{ 176, 44, 58 },
		{ 176, 45, 58 },
		{ 176, 46, 58 },
		{ 176, 47, 58 },
		{ 176, 48, 58 },
		{ 176, 49, 58 },
		{ 176, 50, 58 },
		{ 176, 51, 58 },
		{ 176, 52, 58 },
		{ 176, 53, 58 },
		{ 176, 54, 58 },
		{ 176, 55, 58 },
		{ 176, 56, 58 },
		{ 176, 57, 58 },
		{ 176, 58, 58 },
		{ 176, 59, 58 },
		{ 176, 60, 58 },
		{ 176, 61, 58 },
		{ 176, 62, 58 },
		{ 176, 63, 58 },
		{ 176, 64, 58 },
		{ 176, 65, 58 },
		{ 176, 66, 58 },
		{ 176, 67, 58 },
		{ 176, 68, 58 },
		{ 176, 69, 58 },
		{ 176, 70, 58 },
		{ 176, 71, 58 },
		{ 176, 72, 58 },
		{ 176, 73, 58 },
		{ 176, 74, 58 },
		{ 176, 75, 58 },
		{ 176, 76, 58 },
		{ 176, 77, 58 },
		{ 176, 78, 58 },
		{ 176, 79, 58 },
		{ 176, 80, 58 },
		{ 176, 81, 58 },
		{ 176, 82, 58 },
		{ 176, 83, 58 },
		{ 176, 84, 58 },
		{ 176, 85, 58 },
		{ 176, 86, 58 },
		{ 176, 87, 58 },
		{ 176, 88, 58 },
		{ 176, 89, 58 },
		{ 176, 90, 58 },
		{ 176, 91, 58 },
		{ 176, 92, 58 },
		{ 176, 93, 58 },
		{ 176, 94, 58 },
		{ 176, 95, 58 },
		{ 176, 96, 58 },
		{ 176, 97, 58 },
		{ 176, 98, 58 },
		{ 176, 99, 58 },
		{ 176, 100, 58 },
		{ 176, 101, 58 },
		{ 176, 102, 58 },
		{ 176, 103, 58 },
		{ 176, 104, 58 },
		{ 176, 105, 58 },
		{ 176, 106, 58 },
		{ 176, 107, 58 },
		{ 176, 108, 58 },
		{ 176, 109, 58 },
		{ 176, 110, 58 },
		{ 176, 111, 58 },
		{ 176, 112, 58 },
		{ 176, 113, 58 },
		{ 176, 114, 58 },
		{ 176, 115, 58 },
		{ 176, 116, 58 },
		{ 176, 117, 58 },
		{ 176, 118, 58 },
		{ 176, 119, 58 },
		{ 176, 120, 58 },
		{ 176, 121, 58 },
		{ 176, 122, 58 },
		{ 176, 123, 58 },
		{ 176, 124, 58 },
		{ 176, 125, 58 },
		{ 176, 126, 58 },
		{ 176, 127, 58 },
		{ 177, 44, 58 },
		{ 177, 45, 58 },
		{ 177, 46, 58 },
		{ 177, 47, 58 },
		{ 177, 48, 58 },
		{ 177, 49, 58 },
		{ 177, 50, 58 },
		{ 177, 51, 58 },
		{ 177, 52, 58 },
		{ 177, 53, 58 },
		{ 177, 54, 58 },
		{ 177, 55, 58 },
		{ 177, 56, 58 },
		{ 177, 57, 58 },
		{ 177, 58, 58 },
		{ 177, 59, 58 },
		{ 177, 60, 58 },
		{ 177, 61, 58 },
		{ 177, 62, 58 },
		{ 177, 63, 58 },
		{ 177, 64, 58 },
		{ 177, 65, 58 },
		{ 177, 66, 58 },
		{ 177, 67, 58 },
		{ 177, 68, 58 },
		{ 177, 69, 58 },
		{ 177, 70, 58 },
		{ 177, 71, 58 },
		{ 177, 72, 58 },
		{ 177, 73, 58 },
		{ 177, 74, 58 },
		{ 177, 75, 58 },
		{ 177, 76, 58 },
		{ 177, 77, 58 },
		{ 177, 78, 58 },
		{ 177, 79, 58 },
		{ 177, 80, 58 },
		{ 177, 81, 58 },
		{ 177, 82, 58 },
		{ 177, 83, 58 },
		{ 177, 84, 58 },
		{ 177, 85, 58 },
		{ 177, 86, 58 },
		{ 177, 87, 58 },
		{ 177, 88, 58 },
		{ 177, 89, 58 },
		{ 177, 90, 58 },
		{ 177, 91, 58 },
		{ 177, 92, 58 },
		{ 177, 93, 58 },
		{ 177, 94, 58 },
		{ 177, 95, 58 },
		{ 177, 96, 58 },
		{ 177, 97, 58 },
		{ 177, 98, 58 },
		{ 177, 99, 58 },
		{ 177, 100, 58 },
		{ 177, 101, 58 },
		{ 177, 102, 58 },
		{ 177, 103, 58 },
		{ 177, 104, 58 },
		{ 177, 105, 58 },
		{ 177, 106, 58 },
		{ 177, 107, 58 },
		{ 177, 108, 58 },
		{ 177, 109, 58 },
		{ 177, 110, 58 },
		{ 177, 111, 58 },
		{ 177, 112, 58 },
		{ 177, 113, 58 },
		{ 177, 114, 58 },
		{ 177, 115, 58 },
		{ 177, 116, 58 },
		{ 177, 117, 58 },
		{ 177, 118, 58 },
		{ 177, 119, 58 },
		{ 177, 120, 58 },
		{ 177, 121, 58 },
		{ 177, 122, 58 },
		{ 177, 123, 58 },
		{ 177, 124, 58 },
		{ 177, 125, 58 },
		{ 177, 126, 58 },
		{ 177, 127, 58 },
		{ 177, 128, 58 },
		{ 178, 47, 58 },
		{ 178, 48, 58 },
		{ 178, 49, 58 },
		{ 178, 50, 58 },
		{ 178, 51, 58 },
		{ 178, 52, 58 },
		{ 178, 53, 58 },
		{ 178, 54, 58 },
		{ 178, 55, 58 },
		{ 178, 56, 58 },
		{ 178, 57, 58 },
		{ 178, 58, 58 },
		{ 178, 59, 58 },
		{ 178, 60, 58 },
		{ 178, 61, 58 },
		{ 178, 62, 58 },
		{ 178, 63, 58 },
		{ 178, 64, 58 },
		{ 178, 65, 58 },
		{ 178, 66, 58 },
		{ 178, 67, 58 },
		{ 178, 68, 58 },
		{ 178, 69, 58 },
		{ 178, 70, 58 },
		{ 178, 71, 58 },
		{ 178, 72, 58 },
		{ 178, 73, 58 },
		{ 178, 74, 58 },
		{ 178, 75, 58 },
		{ 178, 76, 58 },
		{ 178, 77, 58 },
		{ 178, 78, 58 },
		{ 178, 79, 58 },
		{ 178, 80, 58 },
		{ 178, 81, 58 },
		{ 178, 82, 58 },
		{ 178, 83, 58 },
		{ 178, 84, 58 },
		{ 178, 85, 58 },
		{ 178, 86, 58 },
		{ 178, 87, 58 },
		{ 178, 88, 58 },
		{ 178, 89, 58 },
		{ 178, 90, 58 },
		{ 178, 91, 58 },
		{ 178, 92, 58 },
		{ 178, 93, 58 },
		{ 178, 94, 58 },
		{ 178, 95, 58 },
		{ 178, 96, 58 },
		{ 178, 97, 58 },
		{ 178, 98, 58 },
		{ 178, 99, 58 },
		{ 178, 100, 58 },
		{ 178, 101, 58 },
		{ 178, 102, 58 },
		{ 178, 103, 58 },
		{ 178, 104, 58 },
		{ 178, 105, 58 },
		{ 178, 106, 58 },
		{ 178, 107, 58 },
		{ 178, 108, 58 },
		{ 178, 109, 58 },
		{ 178, 110, 58 },
		{ 178, 111, 58 },
		{ 178, 112, 58 },
		{ 178, 113, 58 },
		{ 178, 114, 58 },
		{ 178, 115, 58 },
		{ 178, 116, 58 },
		{ 178, 117, 58 },
		{ 178, 118, 58 },
		{ 178, 119, 58 },
		{ 178, 120, 58 },
		{ 178, 121, 58 },
		{ 178, 122, 58 },
		{ 178, 123, 58 },
		{ 178, 124, 58 },
		{ 178, 125, 58 },
		{ 178, 126, 58 },
		{ 178, 127, 58 },
		{ 178, 128, 58 },
		{ 178, 129, 58 },
		{ 178, 130, 58 },
		{ 179, 50, 58 },
		{ 179, 51, 58 },
		{ 179, 52, 58 },
		{ 179, 53, 58 },
		{ 179, 54, 58 },
		{ 179, 55, 58 },
		{ 179, 56, 58 },
		{ 179, 57, 58 },
		{ 179, 58, 58 },
		{ 179, 59, 58 },
		{ 179, 60, 58 },
		{ 179, 61, 58 },
		{ 179, 62, 58 },
		{ 179, 63, 58 },
		{ 179, 64, 58 },
		{ 179, 65, 58 },
		{ 179, 66, 58 },
		{ 179, 67, 58 },
		{ 179, 68, 58 },
		{ 179, 69, 58 },
		{ 179, 70, 58 },
		{ 179, 71, 58 },
		{ 179, 72, 58 },
		{ 179, 73, 58 },
		{ 179, 74, 58 },
		{ 179, 75, 58 },
		{ 179, 76, 58 },
		{ 179, 77, 58 },
		{ 179, 78, 58 },
		{ 179, 79, 58 },
		{ 179, 80, 58 },
		{ 179, 81, 58 },
		{ 179, 82, 58 },
		{ 179, 83, 58 },
		{ 179, 84, 58 },
		{ 179, 85, 58 },
		{ 179, 86, 58 },
		{ 179, 87, 58 },
		{ 179, 88, 58 },
		{ 179, 89, 58 },
		{ 179, 90, 58 },
		{ 179, 91, 58 },
		{ 179, 92, 58 },
		{ 179, 93, 58 },
		{ 179, 94, 58 },
		{ 179, 95, 58 },
		{ 179, 96, 58 },
		{ 179, 97, 58 },
		{ 179, 98, 58 },
		{ 179, 99, 58 },
		{ 179, 100, 58 },
		{ 179, 101, 58 },
		{ 179, 102, 58 },
		{ 179, 103, 58 },
		{ 179, 104, 58 },
		{ 179, 105, 58 },
		{ 179, 106, 58 },
		{ 179, 107, 58 },
		{ 179, 108, 58 },
		{ 179, 109, 58 },
		{ 179, 110, 58 },
		{ 179, 111, 58 },
		{ 179, 112, 58 },
		{ 179, 113, 58 },
		{ 179, 114, 58 },
		{ 179, 115, 58 },
		{ 179, 116, 58 },
		{ 179, 117, 58 },
		{ 179, 118, 58 },
		{ 179, 119, 58 },
		{ 179, 120, 58 },
		{ 179, 121, 58 },
		{ 179, 122, 58 },
		{ 179, 123, 58 },
		{ 179, 124, 58 },
		{ 179, 125, 58 },
		{ 179, 126, 58 },
		{ 179, 127, 58 },
		{ 179, 128, 58 },
		{ 179, 129, 58 },
		{ 179, 130, 58 },
		{ 179, 131, 58 },
		{ 179, 132, 58 },
		{ 180, 52, 58 },
		{ 180, 53, 58 },
		{ 180, 54, 58 },
		{ 180, 55, 58 },
		{ 180, 56, 58 },
		{ 180, 57, 58 },
		{ 180, 58, 58 },
		{ 180, 59, 58 },
		{ 180, 60, 58 },
		{ 180, 61, 58 },
		{ 180, 62, 58 },
		{ 180, 63, 58 },
		{ 180, 64, 58 },
		{ 180, 65, 58 },
		{ 180, 66, 58 },
		{ 180, 67, 58 },
		{ 180, 68, 58 },
		{ 180, 69, 58 },
		{ 180, 70, 58 },
		{ 180, 71, 58 },
		{ 180, 72, 58 },
		{ 180, 73, 58 },
		{ 180, 74, 58 },
		{ 180, 75, 58 },
		{ 180, 76, 58 },
		{ 180, 77, 58 },
		{ 180, 78, 58 },
		{ 180, 79, 58 },
		{ 180, 80, 58 },
		{ 180, 81, 58 },
		{ 180, 82, 58 },
		{ 180, 83, 58 },
		{ 180, 84, 58 },
		{ 180, 85, 58 },
		{ 180, 86, 58 },
		{ 180, 87, 58 },
		{ 180, 88, 58 },
		{ 180, 89, 58 },
		{ 180, 90, 58 },
		{ 180, 91, 58 },
		{ 180, 92, 58 },
		{ 180, 93, 58 },
		{ 180, 94, 58 },
		{ 180, 95, 58 },
		{ 180, 96, 58 },
		{ 180, 97, 58 },
		{ 180, 98, 58 },
		{ 180, 99, 58 },
		{ 180, 100, 58 },
		{ 180, 101, 58 },
		{ 180, 102, 58 },
		{ 180, 103, 58 },
		{ 180, 104, 58 },
		{ 180, 105, 58 },
		{ 180, 106, 58 },
		{ 180, 107, 58 },
		{ 180, 108, 58 },
		{ 180, 109, 58 },
		{ 180, 110, 58 },
		{ 180, 111, 58 },
		{ 180, 112, 58 },
		{ 180, 113, 58 },
		{ 180, 114, 58 },
		{ 180, 115, 58 },
		{ 180, 116, 58 },
		{ 180, 117, 58 },
		{ 180, 118, 58 },
		{ 180, 119, 58 },
		{ 180, 120, 58 },
		{ 180, 121, 58 },
		{ 180, 122, 58 },
		{ 180, 123, 58 },
		{ 180, 124, 58 },
		{ 180, 125, 58 },
		{ 180, 126, 58 },
		{ 180, 127, 58 },
		{ 180, 128, 58 },
		{ 180, 129, 58 },
		{ 180, 130, 58 },
		{ 180, 131, 58 },
		{ 180, 132, 58 },
		{ 180, 133, 58 },
		{ 181, 55, 58 },
		{ 181, 56, 58 },
		{ 181, 57, 58 },
		{ 181, 58, 58 },
		{ 181, 59, 58 },
		{ 181, 60, 58 },
		{ 181, 61, 58 },
		{ 181, 62, 58 },
		{ 181, 63, 58 },
		{ 181, 64, 58 },
		{ 181, 65, 58 },
		{ 181, 66, 58 },
		{ 181, 67, 58 },
		{ 181, 68, 58 },
		{ 181, 69, 58 },
		{ 181, 70, 58 },
		{ 181, 71, 58 },
		{ 181, 72, 58 },
		{ 181, 73, 58 },
		{ 181, 74, 58 },
		{ 181, 75, 58 },
		{ 181, 76, 58 },
		{ 181, 77, 58 },
		{ 181, 78, 58 },
		{ 181, 79, 58 },
		{ 181, 80, 58 },
		{ 181, 81, 58 },
		{ 181, 82, 58 },
		{ 181, 83, 58 },
		{ 181, 84, 58 },
		{ 181, 85, 58 },
		{ 181, 86, 58 },
		{ 181, 87, 58 },
		{ 181, 88, 58 },
		{ 181, 89, 58 },
		{ 181, 90, 58 },
		{ 181, 91, 58 },
		{ 181, 92, 58 },
		{ 181, 93, 58 },
		{ 181, 94, 58 },
		{ 181, 95, 58 },
		{ 181, 96, 58 },
		{ 181, 97, 58 },
		{ 181, 98, 58 },
		{ 181, 99, 58 },
		{ 181, 100, 58 },
		{ 181, 101, 58 },
		{ 181, 102, 58 },
		{ 181, 103, 58 },
		{ 181, 104, 58 },
		{ 181, 105, 58 },
		{ 181, 106, 58 },
		{ 181, 107, 58 },
		{ 181, 108, 58 },
		{ 181, 109, 58 },
		{ 181, 110, 58 },
		{ 181, 111, 58 },
		{ 181, 112, 58 },
		{ 181, 113, 58 },
		{ 181, 114, 58 },
		{ 181, 115, 58 },
		{ 181, 116, 58 },
		{ 181, 117, 58 },
		{ 181, 118, 58 },
		{ 181, 119, 58 },
		{ 181, 120, 58 },
		{ 181, 121, 58 },
		{ 181, 122, 58 },
		{ 181, 123, 58 },
		{ 181, 124, 58 },
		{ 181, 125, 58 },
		{ 181, 126, 58 },
		{ 181, 127, 58 },
		{ 181, 128, 58 },
		{ 181, 129, 58 },
		{ 181, 130, 58 },
		{ 181, 131, 58 },
		{ 181, 132, 58 },
		{ 181, 133, 58 },
		{ 181, 134, 58 },
		{ 181, 135, 58 },
		{ 182, 58, 58 },
		{ 182, 59, 58 },
		{ 182, 60, 58 },
		{ 182, 61, 58 },
		{ 182, 62, 58 },
		{ 182, 63, 58 },
		{ 182, 64, 58 },
		{ 182, 65, 58 },
		{ 182, 66, 58 },
		{ 182, 67, 58 },
		{ 182, 68, 58 },
		{ 182, 69, 58 },
		{ 182, 70, 58 },
		{ 182, 71, 58 },
		{ 182, 72, 58 },
		{ 182, 73, 58 },
		{ 182, 74, 58 },
		{ 182, 75, 58 },
		{ 182, 76, 58 },
		{ 182, 77, 58 },
		{ 182, 78, 58 },
		{ 182, 79, 58 },
		{ 182, 80, 58 },
		{ 182, 81, 58 },
		{ 182, 82, 58 },
		{ 182, 83, 58 },
		{ 182, 84, 58 },
		{ 182, 85, 58 },
		{ 182, 86, 58 },
		{ 182, 87, 58 },
		{ 182, 88, 58 },
		{ 182, 89, 58 },
		{ 182, 90, 58 },
		{ 182, 91, 58 },
		{ 182, 92, 58 },
		{ 182, 93, 58 },
		{ 182, 94, 58 },
		{ 182, 95, 58 },
		{ 182, 96, 58 },
		{ 182, 97, 58 },
		{ 182, 98, 58 },
		{ 182, 99, 58 },
		{ 182, 100, 58 },
		{ 182, 101, 58 },
		{ 182, 102, 58 },
		{ 182, 103, 58 },
		{ 182, 104, 58 },
		{ 182, 105, 58 },
		{ 182, 106, 58 },
		{ 182, 107, 58 },
		{ 182, 108, 58 },
		{ 182, 109, 58 },
		{ 182, 110, 58 },
		{ 182, 111, 58 },
		{ 182, 112, 58 },
		{ 182, 113, 58 },
		{ 182, 114, 58 },
		{ 182, 115, 58 },
		{ 182, 116, 58 },
		{ 182, 117, 58 },
		{ 182, 118, 58 },
		{ 182, 119, 58 },
		{ 182, 120, 58 },
		{ 182, 121, 58 },
		{ 182, 122, 58 },
		{ 182, 123, 58 },
		{ 182, 124, 58 },
		{ 182, 125, 58 },
		{ 182, 126, 58 },
		{ 182, 127, 58 },
		{ 182, 128, 58 },
		{ 182, 129, 58 },
		{ 182, 130, 58 },
		{ 182, 131, 58 },
		{ 182, 132, 58 },
		{ 182, 133, 58 },
		{ 182, 134, 58 },
		{ 182, 135, 58 },
		{ 182, 136, 58 },
		{ 183, 60, 58 },
		{ 183, 61, 58 },
		{ 183, 62, 58 },
		{ 183, 63, 58 },
		{ 183, 64, 58 },
		{ 183, 65, 58 },
		{ 183, 66, 58 },
		{ 183, 67, 58 },
		{ 183, 68, 58 },
		{ 183, 69, 58 },
		{ 183, 70, 58 },
		{ 183, 71, 58 },
		{ 183, 72, 58 },
		{ 183, 73, 58 },
		{ 183, 74, 58 },
		{ 183, 75, 58 },
		{ 183, 76, 58 },
		{ 183, 77, 58 },
		{ 183, 78, 58 },
		{ 183, 79, 58 },
		{ 183, 80, 58 },
		{ 183, 81, 58 },
		{ 183, 82, 58 },
		{ 183, 83, 58 },
		{ 183, 84, 58 },
		{ 183, 85, 58 },
		{ 183, 86, 58 },
		{ 183, 87, 58 },
		{ 183, 88, 58 },
		{ 183, 89, 58 },
		{ 183, 90, 58 },
		{ 183, 91, 58 },
		{ 183, 92, 58 },
		{ 183, 93, 58 },
		{ 183, 94, 58 },
		{ 183, 95, 58 },
		{ 183, 96, 58 },
		{ 183, 97, 58 },
		{ 183, 98, 58 },
		{ 183, 99, 58 },
		{ 183, 100, 58 },
		{ 183, 101, 58 },
		{ 183, 102, 58 },
		{ 183, 103, 58 },
		{ 183, 104, 58 },
		{ 183, 105, 58 },
		{ 183, 106, 58 },
		{ 183, 107, 58 },
		{ 183, 108, 58 },
		{ 183, 109, 58 },
		{ 183, 110, 58 },
		{ 183, 111, 58 },
		{ 183, 112, 58 },
		{ 183, 113, 58 },
		{ 183, 114, 58 },
		{ 183, 115, 58 },
		{ 183, 116, 58 },
		{ 183, 117, 58 },
		{ 183, 118, 58 },
		{ 183, 119, 58 },
		{ 183, 120, 58 },
		{ 183, 121, 58 },
		{ 183, 122, 58 },
		{ 183, 123, 58 },
		{ 183, 124, 58 },
		{ 183, 125, 58 },
		{ 183, 126, 58 },
		{ 183, 127, 58 },
		{ 183, 128, 58 },
		{ 183, 129, 58 },
		{ 183, 130, 58 },
		{ 183, 131, 58 },
		{ 183, 132, 58 },
		{ 183, 133, 58 },
		{ 183, 134, 58 },
		{ 183, 135, 58 },
		{ 183, 136, 58 },
		{ 183, 137, 58 },
		{ 183, 138, 58 },
		{ 184, 63, 58 },
		{ 184, 64, 58 },
		{ 184, 65, 58 },
		{ 184, 66, 58 },
		{ 184, 67, 58 },
		{ 184, 68, 58 },
		{ 184, 69, 58 },
		{ 184, 70, 58 },
		{ 184, 71, 58 },
		{ 184, 72, 58 },
		{ 184, 73, 58 },
		{ 184, 74, 58 },
		{ 184, 75, 58 },
		{ 184, 76, 58 },
		{ 184, 77, 58 },
		{ 184, 78, 58 },
		{ 184, 79, 58 },
		{ 184, 80, 58 },
		{ 184, 81, 58 },
		{ 184, 82, 58 },
		{ 184, 83, 58 },
		{ 184, 84, 58 },
		{ 184, 85, 58 },
		{ 184, 86, 58 },
		{ 184, 87, 58 },
		{ 184, 88, 58 },
		{ 184, 89, 58 },
		{ 184, 90, 58 },
		{ 184, 91, 58 },
		{ 184, 92, 58 },
		{ 184, 93, 58 },
		{ 184, 94, 58 },
		{ 184, 95, 58 },
		{ 184, 96, 58 },
		{ 184, 97, 58 },
		{ 184, 98, 58 },
		{ 184, 99, 58 },
		{ 184, 100, 58 },
		{ 184, 101, 58 },
		{ 184, 102, 58 },
		{ 184, 103, 58 },
		{ 184, 104, 58 },
		{ 184, 105, 58 },
		{ 184, 106, 58 },
		{ 184, 107, 58 },
		{ 184, 108, 58 },
		{ 184, 109, 58 },
		{ 184, 110, 58 },
		{ 184, 111, 58 },
		{ 184, 112, 58 },
		{ 184, 113, 58 },
		{ 184, 114, 58 },
		{ 184, 115, 58 },
		{ 184, 116, 58 },
		{ 184, 117, 58 },
		{ 184, 118, 58 },
		{ 184, 119, 58 },
		{ 184, 120, 58 },
		{ 184, 121, 58 },
		{ 184, 122, 58 },
		{ 184, 123, 58 },
		{ 184, 124, 58 },
		{ 184, 125, 58 },
		{ 184, 126, 58 },
		{ 184, 127, 58 },
		{ 184, 128, 58 },
		{ 184, 129, 58 },
		{ 184, 130, 58 },
		{ 184, 131, 58 },
		{ 184, 132, 58 },
		{ 184, 133, 58 },
		{ 184, 134, 58 },
		{ 184, 135, 58 },
		{ 184, 136, 58 },
		{ 184, 137, 58 },
		{ 184, 138, 58 },
		{ 184, 139, 58 },
		{ 184, 140, 58 },
		{ 185, 66, 58 },
		{ 185, 67, 58 },
		{ 185, 68, 58 },
		{ 185, 69, 58 },
		{ 185, 70, 58 },
		{ 185, 71, 58 },
		{ 185, 72, 58 },
		{ 185, 73, 58 },
		{ 185, 74, 58 },
		{ 185, 75, 58 },
		{ 185, 76, 58 },
		{ 185, 77, 58 },
		{ 185, 78, 58 },
		{ 185, 79, 58 },
		{ 185, 80, 58 },
		{ 185, 81, 58 },
		{ 185, 82, 58 },
		{ 185, 83, 58 },
		{ 185, 84, 58 },
		{ 185, 85, 58 },
		{ 185, 86, 58 },
		{ 185, 87, 58 },
		{ 185, 88, 58 },
		{ 185, 89, 58 },
		{ 185, 90, 58 },
		{ 185, 91, 58 },
		{ 185, 92, 58 },
		{ 185, 93, 58 },
		{ 185, 94, 58 },
		{ 185, 95, 58 },
		{ 185, 96, 58 },
		{ 185, 97, 58 },
		{ 185, 98, 58 },
		{ 185, 99, 58 },
		{ 185, 100, 58 },
		{ 185, 101, 58 },
		{ 185, 102, 58 },
		{ 185, 103, 58 },
		{ 185, 104, 58 },
		{ 185, 105, 58 },
		{ 185, 106, 58 },
		{ 185, 107, 58 },
		{ 185, 108, 58 },
		{ 185, 109, 58 },
		{ 185, 110, 58 },
		{ 185, 111, 58 },
		{ 185, 112, 58 },
		{ 185, 113, 58 },
		{ 185, 114, 58 },
		{ 185, 115, 58 },
		{ 185, 116, 58 },
		{ 185, 117, 58 },
		{ 185, 118, 58 },
		{ 185, 119, 58 },
		{ 185, 120, 58 },
		{ 185, 121, 58 },
		{ 185, 122, 58 },
		{ 185, 123, 58 },
		{ 185, 124, 58 },
		{ 185, 125, 58 },
		{ 185, 126, 58 },
		{ 185, 127, 58 },
		{ 185, 128, 58 },
		{ 185, 129, 58 },
		{ 185, 130, 58 },
		{ 185, 131, 58 },
		{ 185, 132, 58 },
		{ 185, 133, 58 },
		{ 185, 134, 58 },
		{ 185, 135, 58 },
		{ 185, 136, 58 },
		{ 185, 137, 58 },
		{ 185, 138, 58 },
		{ 185, 139, 58 },
		{ 185, 140, 58 },
		{ 185, 141, 58 },
		{ 186, 68, 58 },
		{ 186, 69, 58 },
		{ 186, 70, 58 },
		{ 186, 71, 58 },
		{ 186, 72, 58 },
		{ 186, 73, 58 },
		{ 186, 74, 58 },
		{ 186, 75, 58 },
		{ 186, 76, 58 },
		{ 186, 77, 58 },
		{ 186, 78, 58 },
		{ 186, 79, 58 },
		{ 186, 80, 58 },
		{ 186, 81, 58 },
		{ 186, 82, 58 },
		{ 186, 83, 58 },
		{ 186, 84, 58 },
		{ 186, 85, 58 },
		{ 186, 86, 58 },
		{ 186, 87, 58 },
		{ 186, 88, 58 },
		{ 186, 89, 58 },
		{ 186, 90, 58 },
		{ 186, 91, 58 },
		{ 186, 92, 58 },
		{ 186, 93, 58 },
		{ 186, 94, 58 },
		{ 186, 95, 58 },
		{ 186, 96, 58 },
		{ 186, 97, 58 },
		{ 186, 98, 58 },
		{ 186, 99, 58 },
		{ 186, 100, 58 },
		{ 186, 101, 58 },
		{ 186, 102, 58 },
		{ 186, 103, 58 },
		{ 186, 104, 58 },
		{ 186, 105, 58 },
		{ 186, 106, 58 },
		{ 186, 107, 58 },
		{ 186, 108, 58 },
		{ 186, 109, 58 },
		{ 186, 110, 58 },
		{ 186, 111, 58 },
		{ 186, 112, 58 },
		{ 186, 113, 58 },
		{ 186, 114, 58 },
		{ 186, 115, 58 },
		{ 186, 116, 58 },
		{ 186, 117, 58 },
		{ 186, 118, 58 },
		{ 186, 119, 58 },
		{ 186, 120, 58 },
		{ 186, 121, 58 },
		{ 186, 122, 58 },
		{ 186, 123, 58 },
		{ 186, 124, 58 },
		{ 186, 125, 58 },
		{ 186, 126, 58 },
		{ 186, 127, 58 },
		{ 186, 128, 58 },
		{ 186, 129, 58 },
		{ 186, 130, 58 },
		{ 186, 131, 58 },
		{ 186, 132, 58 },
		{ 186, 133, 58 },
		{ 186, 134, 58 },
		{ 186, 135, 58 },
		{ 186, 136, 58 },
		{ 186, 137, 58 },
		{ 186, 138, 58 },
		{ 186, 139, 58 },
		{ 186, 140, 58 },
		{ 186, 141, 58 },
		{ 186, 142, 58 },
		{ 186, 143, 58 },
		{ 187, 71, 58 },
		{ 187, 72, 58 },
		{ 187, 73, 58 },
		{ 187, 74, 58 },
		{ 187, 75, 58 },
		{ 187, 76, 58 },
		{ 187, 77, 58 },
		{ 187, 78, 58 },
		{ 187, 79, 58 },
		{ 187, 80, 58 },
		{ 187, 81, 58 },
		{ 187, 82, 58 },
		{ 187, 83, 58 },
		{ 187, 84, 58 },
		{ 187, 85, 58 },
		{ 187, 86, 58 },
		{ 187, 87, 58 },
		{ 187, 88, 58 },
		{ 187, 89, 58 },
		{ 187, 90, 58 },
		{ 187, 91, 58 },
		{ 187, 92, 58 },
		{ 187, 93, 58 },
		{ 187, 94, 58 },
		{ 187, 95, 58 },
		{ 187, 96, 58 },
		{ 187, 97, 58 },
		{ 187, 98, 58 },
		{ 187, 99, 58 },
		{ 187, 100, 58 },
		{ 187, 101, 58 },
		{ 187, 102, 58 },
		{ 187, 103, 58 },
		{ 187, 104, 58 },
		{ 187, 105, 58 },
		{ 187, 106, 58 },
		{ 187, 107, 58 },
		{ 187, 108, 58 },
		{ 187, 109, 58 },
		{ 187, 110, 58 },
		{ 187, 111, 58 },
		{ 187, 112, 58 },
		{ 187, 113, 58 },
		{ 187, 114, 58 },
		{ 187, 115, 58 },
		{ 187, 116, 58 },
		{ 187, 117, 58 },
		{ 187, 118, 58 },
		{ 187, 119, 58 },
		{ 187, 120, 58 },
		{ 187, 121, 58 },
		{ 187, 122, 58 },
		{ 187, 123, 58 },
		{ 187, 124, 58 },
		{ 187, 125, 58 },
		{ 187, 126, 58 },
		{ 187, 127, 58 },
		{ 187, 128, 58 },
		{ 187, 129, 58 },
		{ 187, 130, 58 },
		{ 187, 131, 58 },
		{ 187, 132, 58 },
		{ 187, 133, 58 },
		{ 187, 134, 58 },
		{ 187, 135, 58 },
		{ 187, 136, 58 },
		{ 187, 137, 58 },
		{ 187, 138, 58 },
		{ 187, 139, 58 },
		{ 187, 140, 58 },
		{ 187, 141, 58 },
		{ 187, 142, 58 },
		{ 187, 143, 58 },
		{ 187, 144, 58 },
		{ 188, 74, 58 },
		{ 188, 75, 58 },
		{ 188, 76, 58 },
		{ 188, 77, 58 },
		{ 188, 78, 58 },
		{ 188, 79, 58 },
		{ 188, 80, 58 },
		{ 188, 81, 58 },
		{ 188, 82, 58 },
		{ 188, 83, 58 },
		{ 188, 84, 58 },
		{ 188, 85, 58 },
		{ 188, 86, 58 },
		{ 188, 87, 58 },
		{ 188, 88, 58 },
		{ 188, 89, 58 },
		{ 188, 90, 58 },
		{ 188, 91, 58 },
		{ 188, 92, 58 },
		{ 188, 93, 58 },
		{ 188, 94, 58 },
		{ 188, 95, 58 },
		{ 188, 96, 58 },
		{ 188, 97, 58 },
		{ 188, 98, 58 },
		{ 188, 99, 58 },
		{ 188, 100, 58 },
		{ 188, 101, 58 },
		{ 188, 102, 58 },
		{ 188, 103, 58 },
		{ 188, 104, 58 },
		{ 188, 105, 58 },
		{ 188, 106, 58 },
		{ 188, 107, 58 },
		{ 188, 108, 58 },
		{ 188, 109, 58 },
		{ 188, 110, 58 },
		{ 188, 111, 58 },
		{ 188, 112, 58 },
		{ 188, 113, 58 },
		{ 188, 114, 58 },
		{ 188, 115, 58 },
		{ 188, 116, 58 },
		{ 188, 117, 58 },
		{ 188, 118, 58 },
		{ 188, 119, 58 },
		{ 188, 120, 58 },
		{ 188, 121, 58 },
		{ 188, 122, 58 },
		{ 188, 123, 58 },
		{ 188, 124, 58 },
		{ 188, 125, 58 },
		{ 188, 126, 58 },
		{ 188, 127, 58 },
		{ 188, 128, 58 },
		{ 188, 129, 58 },
		{ 188, 130, 58 },
		{ 188, 131, 58 },
		{ 188, 132, 58 },
		{ 188, 133, 58 },
		{ 188, 134, 58 },
		{ 188, 135, 58 },
		{ 188, 136, 58 },
		{ 188, 137, 58 },
		{ 188, 138, 58 },
		{ 188, 139, 58 },
		{ 188, 140, 58 },
		{ 188, 141, 58 },
		{ 188, 142, 58 },
		{ 188, 143, 58 },
		{ 188, 144, 58 },
		{ 188, 145, 58 },
		{ 188, 146, 58 },
		{ 189, 76, 58 },
		{ 189, 77, 58 },
		{ 189, 78, 58 },
		{ 189, 79, 58 },
		{ 189, 80, 58 },
		{ 189, 81, 58 },
		{ 189, 82, 58 },
		{ 189, 83, 58 },
		{ 189, 84, 58 },
		{ 189, 85, 58 },
		{ 189, 86, 58 },
		{ 189, 87, 58 },
		{ 189, 88, 58 },
		{ 189, 89, 58 },
		{ 189, 90, 58 },
		{ 189, 91, 58 },
		{ 189, 92, 58 },
		{ 189, 93, 58 },
		{ 189, 94, 58 },
		{ 189, 95, 58 },
		{ 189, 96, 58 },
		{ 189, 97, 58 },
		{ 189, 98, 58 },
		{ 189, 99, 58 },
		{ 189, 100, 58 },
		{ 189, 101, 58 },
		{ 189, 102, 58 },
		{ 189, 103, 58 },
		{ 189, 104, 58 },
		{ 189, 105, 58 },
		{ 189, 106, 58 },
		{ 189, 107, 58 },
		{ 189, 108, 58 },
		{ 189, 109, 58 },
		{ 189, 110, 58 },
		{ 189, 111, 58 },
		{ 189, 112, 58 },
		{ 189, 113, 58 },
		{ 189, 114, 58 },
		{ 189, 115, 58 },
		{ 189, 116, 58 },
		{ 189, 117, 58 },
		{ 189, 118, 58 },
		{ 189, 119, 58 },
		{ 189, 120, 58 },
		{ 189, 121, 58 },
		{ 189, 122, 58 },
		{ 189, 123, 58 },
		{ 189, 124, 58 },
		{ 189, 125, 58 },
		{ 189, 126, 58 },
		{ 189, 127, 58 },
		{ 189, 128, 58 },
		{ 189, 129, 58 },
		{ 189, 130, 58 },
		{ 189, 131, 58 },
		{ 189, 132, 58 },
		{ 189, 133, 58 },
		{ 189, 134, 58 },
		{ 189, 135, 58 },
		{ 189, 136, 58 },
		{ 189, 137, 58 },
		{ 189, 138, 58 },
		{ 189, 139, 58 },
		{ 189, 140, 58 },
		{ 189, 141, 58 },
		{ 189, 142, 58 },
		{ 189, 143, 58 },
		{ 189, 144, 58 },
		{ 189, 145, 58 },
		{ 189, 146, 58 },
		{ 189, 147, 58 },
		{ 189, 148, 58 },
		{ 190, 79, 58 },
		{ 190, 80, 58 },
		{ 190, 81, 58 },
		{ 190, 82, 58 },
		{ 190, 83, 58 },
		{ 190, 84, 58 },
		{ 190, 85, 58 },
		{ 190, 86, 58 },
		{ 190, 87, 58 },
		{ 190, 88, 58 },
		{ 190, 89, 58 },
		{ 190, 90, 58 },
		{ 190, 91, 58 },
		{ 190, 92, 58 },
		{ 190, 93, 58 },
		{ 190, 94, 58 },
		{ 190, 95, 58 },
		{ 190, 96, 58 },
		{ 190, 97, 58 },
		{ 190, 98, 58 },
		{ 190, 99, 58 },
		{ 190, 100, 58 },
		{ 190, 101, 58 },
		{ 190, 102, 58 },
		{ 190, 103, 58 },
		{ 190, 104, 58 },
		{ 190, 105, 58 },
		{ 190, 106, 58 },
		{ 190, 107, 58 },
		{ 190, 108, 58 },
		{ 190, 109, 58 },
		{ 190, 110, 58 },
		{ 190, 111, 58 },
		{ 190, 112, 58 },
		{ 190, 113, 58 },
		{ 190, 114, 58 },
		{ 190, 115, 58 },
		{ 190, 116, 58 },
		{ 190, 117, 58 },
		{ 190, 118, 58 },
		{ 190, 119, 58 },
		{ 190, 120, 58 },
		{ 190, 121, 58 },
		{ 190, 122, 58 },
		{ 190, 123, 58 },
		{ 190, 124, 58 },
		{ 190, 125, 58 },
		{ 190, 126, 58 },
		{ 190, 127, 58 },
		{ 190, 128, 58 },
		{ 190, 129, 58 },
		{ 190, 130, 58 },
		{ 190, 131, 58 },
		{ 190, 132, 58 },
		{ 190, 133, 58 },
		{ 190, 134, 58 },
		{ 190, 135, 58 },
		{ 190, 136, 58 },
		{ 190, 137, 58 },
		{ 190, 138, 58 },
		{ 190, 139, 58 },
		{ 190, 140, 58 },
		{ 190, 141, 58 },
		{ 190, 142, 58 },
		{ 190, 143, 58 },
		{ 190, 144, 58 },
		{ 190, 145, 58 },
		{ 190, 146, 58 },
		{ 190, 147, 58 },
		{ 190, 148, 58 },
		{ 190, 149, 58 },
		{ 191, 82, 58 },
		{ 191, 83, 58 },
		{ 191, 84, 58 },
		{ 191, 85, 58 },
		{ 191, 86, 58 },
		{ 191, 87, 58 },
		{ 191, 88, 58 },
		{ 191, 89, 58 },
		{ 191, 90, 58 },
		{ 191, 91, 58 },
		{ 191, 92, 58 },
		{ 191, 93, 58 },
		{ 191, 94, 58 },
		{ 191, 95, 58 },
		{ 191, 96, 58 },
		{ 191, 97, 58 },
		{ 191, 98, 58 },
		{ 191, 99, 58 },
		{ 191, 100, 58 },
		{ 191, 101, 58 },
		{ 191, 102, 58 },
		{ 191, 103, 58 },
		{ 191, 104, 58 },
		{ 191, 105, 58 },
		{ 191, 106, 58 },
		{ 191, 107, 58 },
		{ 191, 108, 58 },
		{ 191, 109, 58 },
		{ 191, 110, 58 },
		{ 191, 111, 58 },
		{ 191, 112, 58 },
		{ 191, 113, 58 },
		{ 191, 114, 58 },
		{ 191, 115, 58 },
		{ 191, 116, 58 },
		{ 191, 117, 58 },
		{ 191, 118, 58 },
		{ 191, 119, 58 },
		{ 191, 120, 58 },
		{ 191, 121, 58 },
		{ 191, 122, 58 },
		{ 191, 123, 58 },
		{ 191, 124, 58 },
		{ 191, 125, 58 },
		{ 191, 126, 58 },
		{ 191, 127, 58 },
		{ 191, 128, 58 },
		{ 191, 129, 58 },
		{ 191, 130, 58 },
		{ 191, 131, 58 },
		{ 191, 132, 58 },
		{ 191, 133, 58 },
		{ 191, 134, 58 },
		{ 191, 135, 58 },
		{ 191, 136, 58 },
		{ 191, 137, 58 },
		{ 191, 138, 58 },
		{ 191, 139, 58 },
		{ 191, 140, 58 },
		{ 191, 141, 58 },
		{ 191, 142, 58 },
		{ 191, 143, 58 },
		{ 191, 144, 58 },
		{ 191, 145, 58 },
		{ 191, 146, 58 },
		{ 191, 147, 58 },
		{ 191, 148, 58 },
		{ 191, 149, 58 },
		{ 191, 150, 58 },
		{ 191, 151, 58 },
		{ 192, 84, 58 },
		{ 192, 85, 58 },
		{ 192, 86, 58 },
		{ 192, 87, 58 },
		{ 192, 88, 58 },
		{ 192, 89, 58 },
		{ 192, 90, 58 },
		{ 192, 91, 58 },
		{ 192, 92, 58 },
		{ 192, 93, 58 },
		{ 192, 94, 58 },
		{ 192, 95, 58 },
		{ 192, 96, 58 },
		{ 192, 97, 58 },
		{ 192, 98, 58 },
		{ 192, 99, 58 },
		{ 192, 100, 58 },
		{ 192, 101, 58 },
		{ 192, 102, 58 },
		{ 192, 103, 58 },
		{ 192, 104, 58 },
		{ 192, 105, 58 },
		{ 192, 106, 58 },
		{ 192, 107, 58 },
		{ 192, 108, 58 },
		{ 192, 109, 58 },
		{ 192, 110, 58 },
		{ 192, 111, 58 },
		{ 192, 112, 58 },
		{ 192, 113, 58 },
		{ 192, 114, 58 },
		{ 192, 115, 58 },
		{ 192, 116, 58 },
		{ 192, 117, 58 },
		{ 192, 118, 58 },
		{ 192, 119, 58 },
		{ 192, 120, 58 },
		{ 192, 121, 58 },
		{ 192, 122, 58 },
		{ 192, 123, 58 },
		{ 192, 124, 58 },
		{ 192, 125, 58 },
		{ 192, 126, 58 },
		{ 192, 127, 58 },
		{ 192, 128, 58 },
		{ 192, 129, 58 },
		{ 192, 130, 58 },
		{ 192, 131, 58 },
		{ 192, 132, 58 },
		{ 192, 133, 58 },
		{ 192, 134, 58 },
		{ 192, 135, 58 },
		{ 192, 136, 58 },
		{ 192, 137, 58 },
		{ 192, 138, 58 },
		{ 192, 139, 58 },
		{ 192, 140, 58 },
		{ 192, 141, 58 },
		{ 192, 142, 58 },
		{ 192, 143, 58 },
		{ 192, 144, 58 },
		{ 192, 145, 58 },
		{ 192, 146, 58 },
		{ 192, 147, 58 },
		{ 192, 148, 58 },
		{ 192, 149, 58 },
		{ 192, 150, 58 },
		{ 192, 151, 58 },
		{ 192, 152, 58 },
		{ 193, 87, 58 },
		{ 193, 88, 58 },
		{ 193, 89, 58 },
		{ 193, 90, 58 },
		{ 193, 91, 58 },
		{ 193, 92, 58 },
		{ 193, 93, 58 },
		{ 193, 94, 58 },
		{ 193, 95, 58 },
		{ 193, 96, 58 },
		{ 193, 97, 58 },
		{ 193, 98, 58 },
		{ 193, 99, 58 },
		{ 193, 100, 58 },
		{ 193, 101, 58 },
		{ 193, 102, 58 },
		{ 193, 103, 58 },
		{ 193, 104, 58 },
		{ 193, 105, 58 },
		{ 193, 106, 58 },
		{ 193, 107, 58 },
		{ 193, 108, 58 },
		{ 193, 109, 58 },
		{ 193, 110, 58 },
		{ 193, 111, 58 },
		{ 193, 112, 58 },
		{ 193, 113, 58 },
		{ 193, 114, 58 },
		{ 193, 115, 58 },
		{ 193, 116, 58 },
		{ 193, 117, 58 },
		{ 193, 118, 58 },
		{ 193, 119, 58 },
		{ 193, 120, 58 },
		{ 193, 121, 58 },
		{ 193, 122, 58 },
		{ 193, 123, 58 },
		{ 193, 124, 58 },
		{ 193, 125, 58 },
		{ 193, 126, 58 },
		{ 193, 127, 58 },
		{ 193, 128, 58 },
		{ 193, 129, 58 },
		{ 193, 130, 58 },
		{ 193, 131, 58 },
		{ 193, 132, 58 },
		{ 193, 133, 58 },
		{ 193, 134, 58 },
		{ 193, 135, 58 },
		{ 193, 136, 58 },
		{ 193, 137, 58 },
		{ 193, 138, 58 },
		{ 193, 139, 58 },
		{ 193, 140, 58 },
		{ 193, 141, 58 },
		{ 193, 142, 58 },
		{ 193, 143, 58 },
		{ 193, 144, 58 },
		{ 193, 145, 58 },
		{ 193, 146, 58 },
		{ 193, 147, 58 },
		{ 193, 148, 58 },
		{ 193, 149, 58 },
		{ 193, 150, 58 },
		{ 193, 151, 58 },
		{ 193, 152, 58 },
		{ 193, 153, 58 },
		{ 193, 154, 58 },
		{ 194, 90, 58 },
		{ 194, 91, 58 },
		{ 194, 92, 58 },
		{ 194, 93, 58 },
		{ 194, 94, 58 },
		{ 194, 95, 58 },
		{ 194, 96, 58 },
		{ 194, 97, 58 },
		{ 194, 98, 58 },
		{ 194, 99, 58 },
		{ 194, 100, 58 },
		{ 194, 101, 58 },
		{ 194, 102, 58 },
		{ 194, 103, 58 },
		{ 194, 104, 58 },
		{ 194, 105, 58 },
		{ 194, 106, 58 },
		{ 194, 107, 58 },
		{ 194, 108, 58 },
		{ 194, 109, 58 },
		{ 194, 110, 58 },
		{ 194, 111, 58 },
		{ 194, 112, 58 },
		{ 194, 113, 58 },
		{ 194, 114, 58 },
		{ 194, 115, 58 },
		{ 194, 116, 58 },
		{ 194, 117, 58 },
		{ 194, 118, 58 },
		{ 194, 119, 58 },
		{ 194, 120, 58 },
		{ 194, 121, 58 },
		{ 194, 122, 58 },
		{ 194, 123, 58 },
		{ 194, 124, 58 },
		{ 194, 125, 58 },
		{ 194, 126, 58 },
		{ 194, 127, 58 },
		{ 194, 128, 58 },
		{ 194, 129, 58 },
		{ 194, 130, 58 },
		{ 194, 131, 58 },
		{ 194, 132, 58 },
		{ 194, 133, 58 },
		{ 194, 134, 58 },
		{ 194, 135, 58 },
		{ 194, 136, 58 },
		{ 194, 137, 58 },
		{ 194, 138, 58 },
		{ 194, 139, 58 },
		{ 194, 140, 58 },
		{ 194, 141, 58 },
		{ 194, 142, 58 },
		{ 194, 143, 58 },
		{ 194, 144, 58 },
		{ 194, 145, 58 },
		{ 194, 146, 58 },
		{ 194, 147, 58 },
		{ 194, 148, 58 },
		{ 194, 149, 58 },
		{ 194, 150, 58 },
		{ 194, 151, 58 },
		{ 194, 152, 58 },
		{ 194, 153, 58 },
		{ 194, 154, 58 },
		{ 194, 155, 58 },
		{ 194, 156, 58 },
		{ 195, 92, 58 },
		{ 195, 93, 58 },
		{ 195, 94, 58 },
		{ 195, 95, 58 },
		{ 195, 96, 58 },
		{ 195, 97, 58 },
		{ 195, 98, 58 },
		{ 195, 99, 58 },
		{ 195, 100, 58 },
		{ 195, 101, 58 },
		{ 195, 102, 58 },
		{ 195, 103, 58 },
		{ 195, 104, 58 },
		{ 195, 105, 58 },
		{ 195, 106, 58 },
		{ 195, 107, 58 },
		{ 195, 108, 58 },
		{ 195, 109, 58 },
		{ 195, 110, 58 },
		{ 195, 111, 58 },
		{ 195, 112, 58 },
		{ 195, 113, 58 },
		{ 195, 114, 58 },
		{ 195, 115, 58 },
		{ 195, 116, 58 },
		{ 195, 117, 58 },
		{ 195, 118, 58 },
		{ 195, 119, 58 },
		{ 195, 120, 58 },
		{ 195, 121, 58 },
		{ 195, 122, 58 },
		{ 195, 123, 58 },
		{ 195, 124, 58 },
		{ 195, 125, 58 },
		{ 195, 126, 58 },
		{ 195, 127, 58 },
		{ 195, 128, 58 },
		{ 195, 129, 58 },
		{ 195, 130, 58 },
		{ 195, 131, 58 },
		{ 195, 132, 58 },
		{ 195, 133, 58 },
		{ 195, 134, 58 },
		{ 195, 135, 58 },
		{ 195, 136, 58 },
		{ 195, 137, 58 },
		{ 195, 138, 58 },
		{ 195, 139, 58 },
		{ 195, 140, 58 },
		{ 195, 141, 58 },
		{ 195, 142, 58 },
		{ 195, 143, 58 },
		{ 195, 144, 58 },
		{ 195, 145, 58 },
		{ 195, 146, 58 },
		{ 195, 147, 58 },
		{ 195, 148, 58 },
		{ 195, 149, 58 },
		{ 195, 150, 58 },
		{ 195, 151, 58 },
		{ 195, 152, 58 },
		{ 195, 153, 58 },
		{ 195, 154, 58 },
		{ 195, 155, 58 },
		{ 195, 156, 58 },
		{ 195, 157, 58 },
		{ 196, 95, 58 },
		{ 196, 96, 58 },
		{ 196, 97, 58 },
		{ 196, 98, 58 },
		{ 196, 99, 58 },
		{ 196, 100, 58 },
		{ 196, 101, 58 },
		{ 196, 102, 58 },
		{ 196, 103, 58 },
		{ 196, 104, 58 },
		{ 196, 105, 58 },
		{ 196, 106, 58 },
		{ 196, 107, 58 },
		{ 196, 108, 58 },
		{ 196, 109, 58 },
		{ 196, 110, 58 },
		{ 196, 111, 58 },
		{ 196, 112, 58 },
		{ 196, 113, 58 },
		{ 196, 114, 58 },
		{ 196, 115, 58 },
		{ 196, 116, 58 },
		{ 196, 117, 58 },
		{ 196, 118, 58 },
		{ 196, 119, 58 },
		{ 196, 120, 58 },
		{ 196, 121, 58 },
		{ 196, 122, 58 },
		{ 196, 123, 58 },
		{ 196, 124, 58 },
		{ 196, 125, 58 },
		{ 196, 126, 58 },
		{ 196, 127, 58 },
		{ 196, 128, 58 },
		{ 196, 129, 58 },
		{ 196, 130, 58 },
		{ 196, 131, 58 },
		{ 196, 132, 58 },
		{ 196, 133, 58 },
		{ 196, 134, 58 },
		{ 196, 135, 58 },
		{ 196, 136, 58 },
		{ 196, 137, 58 },
		{ 196, 138, 58 },
		{ 196, 139, 58 },
		{ 196, 140, 58 },
		{ 196, 141, 58 },
		{ 196, 142, 58 },
		{ 196, 143, 58 },
		{ 196, 144, 58 },
		{ 196, 145, 58 },
		{ 196, 146, 58 },
		{ 196, 147, 58 },
		{ 196, 148, 58 },
		{ 196, 149, 58 },
		{ 196, 150, 58 },
		{ 196, 151, 58 },
		{ 196, 152, 58 },
		{ 196, 153, 58 },
		{ 196, 154, 58 },
		{ 196, 155, 58 },
		{ 196, 156, 58 },
		{ 196, 157, 58 },
		{ 196, 158, 58 },
		{ 196, 159, 58 },
		{ 197, 98, 58 },
		{ 197, 99, 58 },
		{ 197, 100, 58 },
		{ 197, 101, 58 },
		{ 197, 102, 58 },
		{ 197, 103, 58 },
		{ 197, 104, 58 },
		{ 197, 105, 58 },
		{ 197, 106, 58 },
		{ 197, 107, 58 },
		{ 197, 108, 58 },
		{ 197, 109, 58 },
		{ 197, 110, 58 },
		{ 197, 111, 58 },
		{ 197, 112, 58 },
		{ 197, 113, 58 },
		{ 197, 114, 58 },
		{ 197, 115, 58 },
		{ 197, 116, 58 },
		{ 197, 117, 58 },
		{ 197, 118, 58 },
		{ 197, 119, 58 },
		{ 197, 120, 58 },
		{ 197, 121, 58 },
		{ 197, 122, 58 },
		{ 197, 123, 58 },
		{ 197, 124, 58 },
		{ 197, 125, 58 },
		{ 197, 126, 58 },
		{ 197, 127, 58 },
		{ 197, 128, 58 },
		{ 197, 129, 58 },
		{ 197, 130, 58 },
		{ 197, 131, 58 },
		{ 197, 132, 58 },
		{ 197, 133, 58 },
		{ 197, 134, 58 },
		{ 197, 135, 58 },
		{ 197, 136, 58 },
		{ 197, 137, 58 },
		{ 197, 138, 58 },
		{ 197, 139, 58 },
		{ 197, 140, 58 },
		{ 197, 141, 58 },
		{ 197, 142, 58 },
		{ 197, 143, 58 },
		{ 197, 144, 58 },
		{ 197, 145, 58 },
		{ 197, 146, 58 },
		{ 197, 147, 58 },
		{ 197, 148, 58 },
		{ 197, 149, 58 },
		{ 197, 150, 58 },
		{ 197, 151, 58 },
		{ 197, 152, 58 },
		{ 197, 153, 58 },
		{ 197, 154, 58 },
		{ 197, 155, 58 },
		{ 197, 156, 58 },
		{ 197, 157, 58 },
		{ 197, 158, 58 },
		{ 197, 159, 58 },
		{ 197, 160, 58 },
		{ 198, 100, 58 },
		{ 198, 101, 58 },
		{ 198, 102, 58 },
		{ 198, 103, 58 },
		{ 198, 104, 58 },
		{ 198, 105, 58 },
		{ 198, 106, 58 },
		{ 198, 107, 58 },
		{ 198, 108, 58 },
		{ 198, 109, 58 },
		{ 198, 110, 58 },
		{ 198, 111, 58 },
		{ 198, 112, 58 },
		{ 198, 113, 58 },
		{ 198, 114, 58 },
		{ 198, 115, 58 },
		{ 198, 116, 58 },
		{ 198, 117, 58 },
		{ 198, 118, 58 },
		{ 198, 119, 58 },
		{ 198, 120, 58 },
		{ 198, 121, 58 },
		{ 198, 122, 58 },
		{ 198, 123, 58 },
		{ 198, 124, 58 },
		{ 198, 125, 58 },
		{ 198, 126, 58 },
		{ 198, 127, 58 },
		{ 198, 128, 58 },
		{ 198, 129, 58 },
		{ 198, 130, 58 },
		{ 198, 131, 58 },
		{ 198, 132, 58 },
		{ 198, 133, 58 },
		{ 198, 134, 58 },
		{ 198, 135, 58 },
		{ 198, 136, 58 },
		{ 198, 137, 58 },
		{ 198, 138, 58 },
		{ 198, 139, 58 },
		{ 198, 140, 58 },
		{ 198, 141, 58 },
		{ 198, 142, 58 },
		{ 198, 143, 58 },
		{ 198, 144, 58 },
		{ 198, 145, 58 },
		{ 198, 146, 58 },
		{ 198, 147, 58 },
		{ 198, 148, 58 },
		{ 198, 149, 58 },
		{ 198, 150, 58 },
		{ 198, 151, 58 },
		{ 198, 152, 58 },
		{ 198, 153, 58 },
		{ 198, 154, 58 },
		{ 198, 155, 58 },
		{ 198, 156, 58 },
		{ 198, 157, 58 },
		{ 198, 158, 58 },
		{ 198, 159, 58 },
		{ 198, 160, 58 },
		{ 198, 161, 58 },
		{ 198, 162, 58 },
		{ 199, 103, 58 },
		{ 199, 104, 58 },
		{ 199, 105, 58 },
		{ 199, 106, 58 },
		{ 199, 107, 58 },
		{ 199, 108, 58 },
		{ 199, 109, 58 },
		{ 199, 110, 58 },
		{ 199, 111, 58 },
		{ 199, 112, 58 },
		{ 199, 113, 58 },
		{ 199, 114, 58 },
		{ 199, 115, 58 },
		{ 199, 116, 58 },
		{ 199, 117, 58 },
		{ 199, 118, 58 },
		{ 199, 119, 58 },
		{ 199, 120, 58 },
		{ 199, 121, 58 },
		{ 199, 122, 58 },
		{ 199, 123, 58 },
		{ 199, 124, 58 },
		{ 199, 125, 58 },
		{ 199, 126, 58 },
		{ 199, 127, 58 },
		{ 199, 128, 58 },
		{ 199, 129, 58 },
		{ 199, 130, 58 },
		{ 199, 131, 58 },
		{ 199, 132, 58 },
		{ 199, 133, 58 },
		{ 199, 134, 58 },
		{ 199, 135, 58 },
		{ 199, 136, 58 },
		{ 199, 137, 58 },
		{ 199, 138, 58 },
		{ 199, 139, 58 },
		{ 199, 140, 58 },
		{ 199, 141, 58 },
		{ 199, 142, 58 },
		{ 199, 143, 58 },
		{ 199, 144, 58 },
		{ 199, 145, 58 },
		{ 199, 146, 58 },
		{ 199, 147, 58 },
		{ 199, 148, 58 },
		{ 199, 149, 58 },
		{ 199, 150, 58 },
		{ 199, 151, 58 },
		{ 199, 152, 58 },
		{ 199, 153, 58 },
		{ 199, 154, 58 },
		{ 199, 155, 58 },
		{ 199, 156, 58 },
		{ 199, 157, 58 },
		{ 199, 158, 58 },
		{ 199, 159, 58 },
		{ 199, 160, 58 },
		{ 199, 161, 58 },
		{ 199, 162, 58 },
		{ 199, 163, 58 },
		{ 199, 164, 58 },
		{ 200, 106, 58 },
		{ 200, 107, 58 },
		{ 200, 108, 58 },
		{ 200, 109, 58 },
		{ 200, 110, 58 },
		{ 200, 111, 58 },
		{ 200, 112, 58 },
		{ 200, 113, 58 },
		{ 200, 114, 58 },
		{ 200, 115, 58 },
		{ 200, 116, 58 },
		{ 200, 117, 58 },
		{ 200, 118, 58 },
		{ 200, 119, 58 },
		{ 200, 120, 58 },
		{ 200, 121, 58 },
		{ 200, 122, 58 },
		{ 200, 123, 58 },
		{ 200, 124, 58 },
		{ 200, 125, 58 },
		{ 200, 126, 58 },
		{ 200, 127, 58 },
		{ 200, 128, 58 },
		{ 200, 129, 58 },
		{ 200, 130, 58 },
		{ 200, 131, 58 },
		{ 200, 132, 58 },
		{ 200, 133, 58 },
		{ 200, 134, 58 },
		{ 200, 135, 58 },
		{ 200, 136, 58 },
		{ 200, 137, 58 },
		{ 200, 138, 58 },
		{ 200, 139, 58 },
		{ 200, 140, 58 },
		{ 200, 141, 58 },
		{ 200, 142, 58 },
		{ 200, 143, 58 },
		{ 200, 144, 58 },
		{ 200, 145, 58 },
		{ 200, 146, 58 },
		{ 200, 147, 58 },
		{ 200, 148, 58 },
		{ 200, 149, 58 },
		{ 200, 150, 58 },
		{ 200, 151, 58 },
		{ 200, 152, 58 },
		{ 200, 153, 58 },
		{ 200, 154, 58 },
		{ 200, 155, 58 },
		{ 200, 156, 58 },
		{ 200, 157, 58 },
		{ 200, 158, 58 },
		{ 200, 159, 58 },
		{ 200, 160, 58 },
		{ 200, 161, 58 },
		{ 200, 162, 58 },
		{ 200, 163, 58 },
		{ 200, 164, 58 },
		{ 200, 165, 58 },
		{ 201, 108, 58 },
		{ 201, 109, 58 },
		{ 201, 110, 58 },
		{ 201, 111, 58 },
		{ 201, 112, 58 },
		{ 201, 113, 58 },
		{ 201, 114, 58 },
		{ 201, 115, 58 },
		{ 201, 116, 58 },
		{ 201, 117, 58 },
		{ 201, 118, 58 },
		{ 201, 119, 58 },
		{ 201, 120, 58 },
		{ 201, 121, 58 },
		{ 201, 122, 58 },
		{ 201, 123, 58 },
		{ 201, 124, 58 },
		{ 201, 125, 58 },
		{ 201, 126, 58 },
		{ 201, 127, 58 },
		{ 201, 128, 58 },
		{ 201, 129, 58 },
		{ 201, 130, 58 },
		{ 201, 131, 58 },
		{ 201, 132, 58 },
		{ 201, 133, 58 },
		{ 201, 134, 58 },
		{ 201, 135, 58 },
		{ 201, 136, 58 },
		{ 201, 137, 58 },
		{ 201, 138, 58 },
		{ 201, 139, 58 },
		{ 201, 140, 58 },
		{ 201, 141, 58 },
		{ 201, 142, 58 },
		{ 201, 143, 58 },
		{ 201, 144, 58 },
		{ 201, 145, 58 },
		{ 201, 146, 58 },
		{ 201, 147, 58 },
		{ 201, 148, 58 },
		{ 201, 149, 58 },
		{ 201, 150, 58 },
		{ 201, 151, 58 },
		{ 201, 152, 58 },
		{ 201, 153, 58 },
		{ 201, 154, 58 },
		{ 201, 155, 58 },
		{ 201, 156, 58 },
		{ 201, 157, 58 },
		{ 201, 158, 58 },
		{ 201, 159, 58 },
		{ 201, 160, 58 },
		{ 201, 161, 58 },
		{ 201, 162, 58 },
		{ 201, 163, 58 },
		{ 201, 164, 58 },
		{ 201, 165, 58 },
		{ 201, 166, 58 },
		{ 201, 167, 58 },
		{ 202, 111, 58 },
		{ 202, 112, 58 },
		{ 202, 113, 58 },
		{ 202, 114, 58 },
		{ 202, 115, 58 },
		{ 202, 116, 58 },
		{ 202, 117, 58 },
		{ 202, 118, 58 },
		{ 202, 119, 58 },
		{ 202, 120, 58 },
		{ 202, 121, 58 },
		{ 202, 122, 58 },
		{ 202, 123, 58 },
		{ 202, 124, 58 },
		{ 202, 125, 58 },
		{ 202, 126, 58 },
		{ 202, 127, 58 },
		{ 202, 128, 58 },
		{ 202, 129, 58 },
		{ 202, 130, 58 },
		{ 202, 131, 58 },
		{ 202, 132, 58 },
		{ 202, 133, 58 },
		{ 202, 134, 58 },
		{ 202, 135, 58 },
		{ 202, 136, 58 },
		{ 202, 137, 58 },
		{ 202, 138, 58 },
		{ 202, 139, 58 },
		{ 202, 140, 58 },
		{ 202, 141, 58 },
		{ 202, 142, 58 },
		{ 202, 143, 58 },
		{ 202, 144, 58 },
		{ 202, 145, 58 },
		{ 202, 146, 58 },
		{ 202, 147, 58 },
		{ 202, 148, 58 },
		{ 202, 149, 58 },
		{ 202, 150, 58 },
		{ 202, 151, 58 },
		{ 202, 152, 58 },
		{ 202, 153, 58 },
		{ 202, 154, 58 },
		{ 202, 155, 58 },
		{ 202, 156, 58 },
		{ 202, 157, 58 },
		{ 202, 158, 58 },
		{ 202, 159, 58 },
		{ 202, 160, 58 },
		{ 202, 161, 58 },
		{ 202, 162, 58 },
		{ 202, 163, 58 },
		{ 202, 164, 58 },
		{ 202, 165, 58 },
		{ 202, 166, 58 },
		{ 202, 167, 58 },
		{ 202, 168, 58 },
		{ 203, 114, 58 },
		{ 203, 115, 58 },
		{ 203, 116, 58 },
		{ 203, 117, 58 },
		{ 203, 118, 58 },
		{ 203, 119, 58 },
		{ 203, 120, 58 },
		{ 203, 121, 58 },
		{ 203, 122, 58 },
		{ 203, 123, 58 },
		{ 203, 124, 58 },
		{ 203, 125, 58 },
		{ 203, 126, 58 },
		{ 203, 127, 58 },
		{ 203, 128, 58 },
		{ 203, 129, 58 },
		{ 203, 130, 58 },
		{ 203, 131, 58 },
		{ 203, 132, 58 },
		{ 203, 133, 58 },
		{ 203, 134, 58 },
		{ 203, 135, 58 },
		{ 203, 136, 58 },
		{ 203, 137, 58 },
		{ 203, 138, 58 },
		{ 203, 139, 58 },
		{ 203, 140, 58 },
		{ 203, 141, 58 },
		{ 203, 142, 58 },
		{ 203, 143, 58 },
		{ 203, 144, 58 },
		{ 203, 145, 58 },
		{ 203, 146, 58 },
		{ 203, 147, 58 },
		{ 203, 148, 58 },
		{ 203, 149, 58 },
		{ 203, 150, 58 },
		{ 203, 151, 58 },
		{ 203, 152, 58 },
		{ 203, 153, 58 },
		{ 203, 154, 58 },
		{ 203, 155, 58 },
		{ 203, 156, 58 },
		{ 203, 157, 58 },
		{ 203, 158, 58 },
		{ 203, 159, 58 },
		{ 203, 160, 58 },
		{ 203, 161, 58 },
		{ 203, 162, 58 },
		{ 203, 163, 58 },
		{ 203, 164, 58 },
		{ 203, 165, 58 },
		{ 203, 166, 58 },
		{ 203, 167, 58 },
		{ 203, 168, 58 },
		{ 203, 169, 58 },
		{ 203, 170, 58 },
		{ 204, 116, 58 },
		{ 204, 117, 58 },
		{ 204, 118, 58 },
		{ 204, 119, 58 },
		{ 204, 120, 58 },
		{ 204, 121, 58 },
		{ 204, 122, 58 },
		{ 204, 123, 58 },
		{ 204, 124, 58 },
		{ 204, 125, 58 },
		{ 204, 126, 58 },
		{ 204, 127, 58 },
		{ 204, 128, 58 },
		{ 204, 129, 58 },
		{ 204, 130, 58 },
		{ 204, 131, 58 },
		{ 204, 132, 58 },
		{ 204, 133, 58 },
		{ 204, 134, 58 },
		{ 204, 135, 58 },
		{ 204, 136, 58 },
		{ 204, 137, 58 },
		{ 204, 138, 58 },
		{ 204, 139, 58 },
		{ 204, 140, 58 },
		{ 204, 141, 58 },
		{ 204, 142, 58 },
		{ 204, 143, 58 },
		{ 204, 144, 58 },
		{ 204, 145, 58 },
		{ 204, 146, 58 },
		{ 204, 147, 58 },
		{ 204, 148, 58 },
		{ 204, 149, 58 },
		{ 204, 150, 58 },
		{ 204, 151, 58 },
		{ 204, 152, 58 },
		{ 204, 153, 58 },
		{ 204, 154, 58 },
		{ 204, 155, 58 },
		{ 204, 156, 58 },
		{ 204, 157, 58 },
		{ 204, 158, 58 },
		{ 204, 159, 58 },
		{ 204, 160, 58 },
		{ 204, 161, 58 },
		{ 204, 162, 58 },
		{ 204, 163, 58 },
		{ 204, 164, 58 },
		{ 204, 165, 58 },
		{ 204, 166, 58 },
		{ 204, 167, 58 },
		{ 204, 168, 58 },
		{ 204, 169, 58 },
		{ 204, 170, 58 },
		{ 204, 171, 58 },
		{ 204, 172, 58 },
		{ 205, 119, 58 },
		{ 205, 120, 58 },
		{ 205, 121, 58 },
		{ 205, 122, 58 },
		{ 205, 123, 58 },
		{ 205, 124, 58 },
		{ 205, 125, 58 },
		{ 205, 126, 58 },
		{ 205, 127, 58 },
		{ 205, 128, 58 },
		{ 205, 129, 58 },
		{ 205, 130, 58 },
		{ 205, 131, 58 },
		{ 205, 132, 58 },
		{ 205, 133, 58 },
		{ 205, 134, 58 },
		{ 205, 135, 58 },
		{ 205, 136, 58 },
		{ 205, 137, 58 },
		{ 205, 138, 58 },
		{ 205, 139, 58 },
		{ 205, 140, 58 },
		{ 205, 141, 58 },
		{ 205, 142, 58 },
		{ 205, 143, 58 },
		{ 205, 144, 58 },
		{ 205, 145, 58 },
		{ 205, 146, 58 },
		{ 205, 147, 58 },
		{ 205, 148, 58 },
		{ 205, 149, 58 },
		{ 205, 150, 58 },
		{ 205, 151, 58 },
		{ 205, 152, 58 },
		{ 205, 153, 58 },
		{ 205, 154, 58 },
		{ 205, 155, 58 },
		{ 205, 156, 58 },
		{ 205, 157, 58 },
		{ 205, 158, 58 },
		{ 205, 159, 58 },
		{ 205, 160, 58 },
		{ 205, 161, 58 },
		{ 205, 162, 58 },
		{ 205, 163, 58 },
		{ 205, 164, 58 },
		{ 205, 165, 58 },
		{ 205, 166, 58 },
		{ 205, 167, 58 },
		{ 205, 168, 58 },
		{ 205, 169, 58 },
		{ 205, 170, 58 },
		{ 205, 171, 58 },
		{ 205, 172, 58 },
		{ 205, 173, 58 },
		{ 206, 122, 58 },
		{ 206, 123, 58 },
		{ 206, 124, 58 },
		{ 206, 125, 58 },
		{ 206, 126, 58 },
		{ 206, 127, 58 },
		{ 206, 128, 58 },
		{ 206, 129, 58 },
		{ 206, 130, 58 },
		{ 206, 131, 58 },
		{ 206, 132, 58 },
		{ 206, 133, 58 },
		{ 206, 134, 58 },
		{ 206, 135, 58 },
		{ 206, 136, 58 },
		{ 206, 137, 58 },
		{ 206, 138, 58 },
		{ 206, 139, 58 },
		{ 206, 140, 58 },
		{ 206, 141, 58 },
		{ 206, 142, 58 },
		{ 206, 143, 58 },
		{ 206, 144, 58 },
		{ 206, 145, 58 },
		{ 206, 146, 58 },
		{ 206, 147, 58 },
		{ 206, 148, 58 },
		{ 206, 149, 58 },
		{ 206, 150, 58 },
		{ 206, 151, 58 },
		{ 206, 152, 58 },
		{ 206, 153, 58 },
		{ 206, 154, 58 },
		{ 206, 155, 58 },
		{ 206, 156, 58 },
		{ 206, 157, 58 },
		{ 206, 158, 58 },
		{ 206, 159, 58 },
		{ 206, 160, 58 },
		{ 206, 161, 58 },
		{ 206, 162, 58 },
		{ 206, 163, 58 },
		{ 206, 164, 58 },
		{ 206, 165, 58 },
		{ 206, 166, 58 },
		{ 206, 167, 58 },
		{ 206, 168, 58 },
		{ 206, 169, 58 },
		{ 206, 170, 58 },
		{ 206, 171, 58 },
		{ 206, 172, 58 },
		{ 206, 173, 58 },
		{ 206, 174, 58 },
		{ 206, 175, 58 },
		{ 207, 124, 58 },
		{ 207, 125, 58 },
		{ 207, 126, 58 },
		{ 207, 127, 58 },
		{ 207, 128, 58 },
		{ 207, 129, 58 },
		{ 207, 130, 58 },
		{ 207, 131, 58 },
		{ 207, 132, 58 },
		{ 207, 133, 58 },
		{ 207, 134, 58 },
		{ 207, 135, 58 },
		{ 207, 136, 58 },
		{ 207, 137, 58 },
		{ 207, 138, 58 },
		{ 207, 139, 58 },
		{ 207, 140, 58 },
		{ 207, 141, 58 },
		{ 207, 142, 58 },
		{ 207, 143, 58 },
		{ 207, 144, 58 },
		{ 207, 145, 58 },
		{ 207, 146, 58 },
		{ 207, 147, 58 },
		{ 207, 148, 58 },
		{ 207, 149, 58 },
		{ 207, 150, 58 },
		{ 207, 151, 58 },
		{ 207, 152, 58 },
		{ 207, 153, 58 },
		{ 207, 154, 58 },
		{ 207, 155, 58 },
		{ 207, 156, 58 },
		{ 207, 157, 58 },
		{ 207, 158, 58 },
		{ 207, 159, 58 },
		{ 207, 160, 58 },
		{ 207, 161, 58 },
		{ 207, 162, 58 },
		{ 207, 163, 58 },
		{ 207, 164, 58 },
		{ 207, 165, 58 },
		{ 207, 166, 58 },
		{ 207, 167, 58 },
		{ 207, 168, 58 },
		{ 207, 169, 58 },
		{ 207, 170, 58 },
		{ 207, 171, 58 },
		{ 207, 172, 58 },
		{ 207, 173, 58 },
		{ 207, 174, 58 },
		{ 207, 175, 58 },
		{ 207, 176, 58 },
		{ 208, 127, 58 },
		{ 208, 128, 58 },
		{ 208, 129, 58 },
		{ 208, 130, 58 },
		{ 208, 131, 58 },
		{ 208, 132, 58 },
		{ 208, 133, 58 },
		{ 208, 134, 58 },
		{ 208, 135, 58 },
		{ 208, 136, 58 },
		{ 208, 137, 58 },
		{ 208, 138, 58 },
		{ 208, 139, 58 },
		{ 208, 140, 58 },
		{ 208, 141, 58 },
		{ 208, 142, 58 },
		{ 208, 143, 58 },
		{ 208, 144, 58 },
		{ 208, 145, 58 },
		{ 208, 146, 58 },
		{ 208, 147, 58 },
		{ 208, 148, 58 },
		{ 208, 149, 58 },
		{ 208, 150, 58 },
		{ 208, 151, 58 },
		{ 208, 152, 58 },
		{ 208, 153, 58 },
		{ 208, 154, 58 },
		{ 208, 155, 58 },
		{ 208, 156, 58 },
		{ 208, 157, 58 },
		{ 208, 158, 58 },
		{ 208, 159, 58 },
		{ 208, 160, 58 },
		{ 208, 161, 58 },
		{ 208, 162, 58 },
		{ 208, 163, 58 },
		{ 208, 164, 58 },
		{ 208, 165, 58 },
		{ 208, 166, 58 },
		{ 208, 167, 58 },
		{ 208, 168, 58 },
		{ 208, 169, 58 },
		{ 208, 170, 58 },
		{ 208, 171, 58 },
		{ 208, 172, 58 },
		{ 208, 173, 58 },
		{ 208, 174, 58 },
		{ 208, 175, 58 },
		{ 208, 176, 58 },
		{ 208, 177, 58 },
		{ 208, 178, 58 },
		{ 209, 130, 58 },
		{ 209, 131, 58 },
		{ 209, 132, 58 },
		{ 209, 133, 58 },
		{ 209, 134, 58 },
		{ 209, 135, 58 },
		{ 209, 136, 58 },
		{ 209, 137, 58 },
		{ 209, 138, 58 },
		{ 209, 139, 58 },
		{ 209, 140, 58 },
		{ 209, 141, 58 },
		{ 209, 142, 58 },
		{ 209, 143, 58 },
		{ 209, 144, 58 },
		{ 209, 145, 58 },
		{ 209, 146, 58 },
		{ 209, 147, 58 },
		{ 209, 148, 58 },
		{ 209, 149, 58 },
		{ 209, 150, 58 },
		{ 209, 151, 58 },
		{ 209, 152, 58 },
		{ 209, 153, 58 },
		{ 209, 154, 58 },
		{ 209, 155, 58 },
		{ 209, 156, 58 },
		{ 209, 157, 58 },
		{ 209, 158, 58 },
		{ 209, 159, 58 },
		{ 209, 160, 58 },
		{ 209, 161, 58 },
		{ 209, 162, 58 },
		{ 209, 163, 58 },
		{ 209, 164, 58 },
		{ 209, 165, 58 },
		{ 209, 166, 58 },
		{ 209, 167, 58 },
		{ 209, 168, 58 },
		{ 209, 169, 58 },
		{ 209, 170, 58 },
		{ 209, 171, 58 },
		{ 209, 172, 58 },
		{ 209, 173, 58 },
		{ 209, 174, 58 },
		{ 209, 175, 58 },
		{ 209, 176, 58 },
		{ 209, 177, 58 },
		{ 209, 178, 58 },
		{ 209, 179, 58 },
		{ 209, 180, 58 },
		{ 210, 132, 58 },
		{ 210, 133, 58 },
		{ 210, 134, 58 },
		{ 210, 135, 58 },
		{ 210, 136, 58 },
		{ 210, 137, 58 },
		{ 210, 138, 58 },
		{ 210, 139, 58 },
		{ 210, 140, 58 },
		{ 210, 141, 58 },
		{ 210, 142, 58 },
		{ 210, 143, 58 },
		{ 210, 144, 58 },
		{ 210, 145, 58 },
		{ 210, 146, 58 },
		{ 210, 147, 58 },
		{ 210, 148, 58 },
		{ 210, 149, 58 },
		{ 210, 150, 58 },
		{ 210, 151, 58 },
		{ 210, 152, 58 },
		{ 210, 153, 58 },
		{ 210, 154, 58 },
		{ 210, 155, 58 },
		{ 210, 156, 58 },
		{ 210, 157, 58 },
		{ 210, 158, 58 },
		{ 210, 159, 58 },
		{ 210, 160, 58 },
		{ 210, 161, 58 },
		{ 210, 162, 58 },
		{ 210, 163, 58 },
		{ 210, 164, 58 },
		{ 210, 165, 58 },
		{ 210, 166, 58 },
		{ 210, 167, 58 },
		{ 210, 168, 58 },
		{ 210, 169, 58 },
		{ 210, 170, 58 },
		{ 210, 171, 58 },
		{ 210, 172, 58 },
		{ 210, 173, 58 },
		{ 210, 174, 58 },
		{ 210, 175, 58 },
		{ 210, 176, 58 },
		{ 210, 177, 58 },
		{ 210, 178, 58 },
		{ 210, 179, 58 },
		{ 210, 180, 58 },
		{ 210, 181, 58 },
		{ 211, 135, 58 },
		{ 211, 136, 58 },
		{ 211, 137, 58 },
		{ 211, 138, 58 },
		{ 211, 139, 58 },
		{ 211, 140, 58 },
		{ 211, 141, 58 },
		{ 211, 142, 58 },
		{ 211, 143, 58 },
		{ 211, 144, 58 },
		{ 211, 145, 58 },
		{ 211, 146, 58 },
		{ 211, 147, 58 },
		{ 211, 148, 58 },
		{ 211, 149, 58 },
		{ 211, 150, 58 },
		{ 211, 151, 58 },
		{ 211, 152, 58 },
		{ 211, 153, 58 },
		{ 211, 154, 58 },
		{ 211, 155, 58 },
		{ 211, 156, 58 },
		{ 211, 157, 58 },
		{ 211, 158, 58 },
		{ 211, 159, 58 },
		{ 211, 160, 58 },
		{ 211, 161, 58 },
		{ 211, 162, 58 },
		{ 211, 163, 58 },
		{ 211, 164, 58 },
		{ 211, 165, 58 },
		{ 211, 166, 58 },
		{ 211, 167, 58 },
		{ 211, 168, 58 },
		{ 211, 169, 58 },
		{ 211, 170, 58 },
		{ 211, 171, 58 },
		{ 211, 172, 58 },
		{ 211, 173, 58 },
		{ 211, 174, 58 },
		{ 211, 175, 58 },
		{ 211, 176, 58 },
		{ 211, 177, 58 },
		{ 211, 178, 58 },
		{ 211, 179, 58 },
		{ 211, 180, 58 },
		{ 211, 181, 58 },
		{ 211, 182, 58 },
		{ 211, 183, 58 },
		{ 212, 138, 58 },
		{ 212, 139, 58 },
		{ 212, 140, 58 },
		{ 212, 141, 58 },
		{ 212, 142, 58 },
		{ 212, 143, 58 },
		{ 212, 144, 58 },
		{ 212, 145, 58 },
		{ 212, 146, 58 },
		{ 212, 147, 58 },
		{ 212, 148, 58 },
		{ 212, 149, 58 },
		{ 212, 150, 58 },
		{ 212, 151, 58 },
		{ 212, 152, 58 },
		{ 212, 153, 58 },
		{ 212, 154, 58 },
		{ 212, 155, 58 },
		{ 212, 156, 58 },
		{ 212, 157, 58 },
		{ 212, 158, 58 },
		{ 212, 159, 58 },
		{ 212, 160, 58 },
		{ 212, 161, 58 },
		{ 212, 162, 58 },
		{ 212, 163, 58 },
		{ 212, 164, 58 },
		{ 212, 165, 58 },
		{ 212, 166, 58 },
		{ 212, 167, 58 },
		{ 212, 168, 58 },
		{ 212, 169, 58 },
		{ 212, 170, 58 },
		{ 212, 171, 58 },
		{ 212, 172, 58 },
		{ 212, 173, 58 },
		{ 212, 174, 58 },
		{ 212, 175, 58 },
		{ 212, 176, 58 },
		{ 212, 177, 58 },
		{ 212, 178, 58 },
		{ 212, 179, 58 },
		{ 212, 180, 58 },
		{ 212, 181, 58 },
		{ 212, 182, 58 },
		{ 212, 183, 58 },
		{ 212, 184, 58 },
		{ 213, 140, 58 },
		{ 213, 141, 58 },
		{ 213, 142, 58 },
		{ 213, 143, 58 },
		{ 213, 144, 58 },
		{ 213, 145, 58 },
		{ 213, 146, 58 },
		{ 213, 147, 58 },
		{ 213, 148, 58 },
		{ 213, 149, 58 },
		{ 213, 150, 58 },
		{ 213, 151, 58 },
		{ 213, 152, 58 },
		{ 213, 153, 58 },
		{ 213, 154, 58 },
		{ 213, 155, 58 },
		{ 213, 156, 58 },
		{ 213, 157, 58 },
		{ 213, 158, 58 },
		{ 213, 159, 58 },
		{ 213, 160, 58 },
		{ 213, 161, 58 },
		{ 213, 162, 58 },
		{ 213, 163, 58 },
		{ 213, 164, 58 },
		{ 213, 165, 58 },
		{ 213, 166, 58 },
		{ 213, 167, 58 },
		{ 213, 168, 58 },
		{ 213, 169, 58 },
		{ 213, 170, 58 },
		{ 213, 171, 58 },
		{ 213, 172, 58 },
		{ 213, 173, 58 },
		{ 213, 174, 58 },
		{ 213, 175, 58 },
		{ 213, 176, 58 },
		{ 213, 177, 58 },
		{ 213, 178, 58 },
		{ 213, 179, 58 },
		{ 213, 180, 58 },
		{ 213, 181, 58 },
		{ 213, 182, 58 },
		{ 213, 183, 58 },
		{ 213, 184, 58 },
		{ 213, 185, 58 },
		{ 213, 186, 58 },
		{ 214, 143, 58 },
		{ 214, 144, 58 },
		{ 214, 145, 58 },
		{ 214, 146, 58 },
		{ 214, 147, 58 },
		{ 214, 148, 58 },
		{ 214, 149, 58 },
		{ 214, 150, 58 },
		{ 214, 151, 58 },
		{ 214, 152, 58 },
		{ 214, 153, 58 },
		{ 214, 154, 58 },
		{ 214, 155, 58 },
		{ 214, 156, 58 },
		{ 214, 157, 58 },
		{ 214, 158, 58 },
		{ 214, 159, 58 },
		{ 214, 160, 58 },
		{ 214, 161, 58 },
		{ 214, 162, 58 },
		{ 214, 163, 58 },
		{ 214, 164, 58 },
		{ 214, 165, 58 },
		{ 214, 166, 58 },
		{ 214, 167, 58 },
		{ 214, 168, 58 },
		{ 214, 169, 58 },
		{ 214, 170, 58 },
		{ 214, 171, 58 },
		{ 214, 172, 58 },
		{ 214, 173, 58 },
		{ 214, 174, 58 },
		{ 214, 175, 58 },
		{ 214, 176, 58 },
		{ 214, 177, 58 },
		{ 214, 178, 58 },
		{ 214, 179, 58 },
		{ 214, 180, 58 },
		{ 214, 181, 58 },
		{ 214, 182, 58 },
		{ 214, 183, 58 },
		{ 214, 184, 58 },
		{ 214, 185, 58 },
		{ 214, 186, 58 },
		{ 214, 187, 58 },
		{ 215, 146, 58 },
		{ 215, 147, 58 },
		{ 215, 148, 58 },
		{ 215, 149, 58 },
		{ 215, 150, 58 },
		{ 215, 151, 58 },
		{ 215, 152, 58 },
		{ 215, 153, 58 },
		{ 215, 154, 58 },
		{ 215, 155, 58 },
		{ 215, 156, 58 },
		{ 215, 157, 58 },
		{ 215, 158, 58 },
		{ 215, 159, 58 },
		{ 215, 160, 58 },
		{ 215, 161, 58 },
		{ 215, 162, 58 },
		{ 215, 163, 58 },
		{ 215, 164, 58 },
		{ 215, 165, 58 },
		{ 215, 166, 58 },
		{ 215, 167, 58 },
		{ 215, 168, 58 },
		{ 215, 169, 58 },
		{ 215, 170, 58 },
		{ 215, 171, 58 },
		{ 215, 172, 58 },
		{ 215, 173, 58 },
		{ 215, 174, 58 },
		{ 215, 175, 58 },
		{ 215, 176, 58 },
		{ 215, 177, 58 },
		{ 215, 178, 58 },
		{ 215, 179, 58 },
		{ 215, 180, 58 },
		{ 215, 181, 58 },
		{ 215, 182, 58 },
		{ 215, 183, 58 },
		{ 215, 184, 58 },
		{ 215, 185, 58 },
		{ 215, 186, 58 },
		{ 215, 187, 58 },
		{ 215, 188, 58 },
		{ 215, 189, 58 },
		{ 216, 148, 58 },
		{ 216, 149, 58 },
		{ 216, 150, 58 },
		{ 216, 151, 58 },
		{ 216, 152, 58 },
		{ 216, 153, 58 },
		{ 216, 154, 58 },
		{ 216, 155, 58 },
		{ 216, 156, 58 },
		{ 216, 157, 58 },
		{ 216, 158, 58 },
		{ 216, 159, 58 },
		{ 216, 160, 58 },
		{ 216, 161, 58 },
		{ 216, 162, 58 },
		{ 216, 163, 58 },
		{ 216, 164, 58 },
		{ 216, 165, 58 },
		{ 216, 166, 58 },
		{ 216, 167, 58 },
		{ 216, 168, 58 },
		{ 216, 169, 58 },
		{ 216, 170, 58 },
		{ 216, 171, 58 },
		{ 216, 172, 58 },
		{ 216, 173, 58 },
		{ 216, 174, 58 },
		{ 216, 175, 58 },
		{ 216, 176, 58 },
		{ 216, 177, 58 },
		{ 216, 178, 58 },
		{ 216, 179, 58 },
		{ 216, 180, 58 },
		{ 216, 181, 58 },
		{ 216, 182, 58 },
		{ 216, 183, 58 },
		{ 216, 184, 58 },
		{ 216, 185, 58 },
		{ 216, 186, 58 },
		{ 216, 187, 58 },
		{ 216, 188, 58 },
		{ 216, 189, 58 },
		{ 216, 190, 58 },
		{ 217, 151, 58 },
		{ 217, 152, 58 },
		{ 217, 153, 58 },
		{ 217, 154, 58 },
		{ 217, 155, 58 },
		{ 217, 156, 58 },
		{ 217, 157, 58 },
		{ 217, 158, 58 },
		{ 217, 159, 58 },
		{ 217, 160, 58 },
		{ 217, 161, 58 },
		{ 217, 162, 58 },
		{ 217, 163, 58 },
		{ 217, 164, 58 },
		{ 217, 165, 58 },
		{ 217, 166, 58 },
		{ 217, 167, 58 },
		{ 217, 168, 58 },
		{ 217, 169, 58 },
		{ 217, 170, 58 },
		{ 217, 171, 58 },
		{ 217, 172, 58 },
		{ 217, 173, 58 },
		{ 217, 174, 58 },
		{ 217, 175, 58 },
		{ 217, 176, 58 },
		{ 217, 177, 58 },
		{ 217, 178, 58 },
		{ 217, 179, 58 },
		{ 217, 180, 58 },
		{ 217, 181, 58 },
		{ 217, 182, 58 },
		{ 217, 183, 58 },
		{ 217, 184, 58 },
		{ 217, 185, 58 },
		{ 217, 186, 58 },
		{ 217, 187, 58 },
		{ 217, 188, 58 },
		{ 217, 189, 58 },
		{ 217, 190, 58 },
		{ 218, 154, 58 },
		{ 218, 155, 58 },
		{ 218, 156, 58 },
		{ 218, 157, 58 },
		{ 218, 158, 58 },
		{ 218, 159, 58 },
		{ 218, 160, 58 },
		{ 218, 161, 58 },
		{ 218, 162, 58 },
		{ 218, 163, 58 },
		{ 218, 164, 58 },
		{ 218, 165, 58 },
		{ 218, 166, 58 },
		{ 218, 167, 58 },
		{ 218, 168, 58 },
		{ 218, 169, 58 },
		{ 218, 170, 58 },
		{ 218, 171, 58 },
		{ 218, 172, 58 },
		{ 218, 173, 58 },
		{ 218, 174, 58 },
		{ 218, 175, 58 },
		{ 218, 176, 58 },
		{ 218, 177, 58 },
		{ 218, 178, 58 },
		{ 218, 179, 58 },
		{ 218, 180, 58 },
		{ 218, 181, 58 },
		{ 218, 182, 58 },
		{ 218, 183, 58 },
		{ 218, 184, 58 },
		{ 218, 185, 58 },
		{ 218, 186, 58 },
		{ 218, 187, 58 },
		{ 218, 188, 58 },
		{ 218, 189, 58 },
		{ 218, 190, 58 },
		{ 218, 191, 58 },
		{ 219, 156, 58 },
		{ 219, 157, 58 },
		{ 219, 158, 58 },
		{ 219, 159, 58 },
		{ 219, 160, 58 },
		{ 219, 161, 58 },
		{ 219, 162, 58 },
		{ 219, 163, 58 },
		{ 219, 164, 58 },
		{ 219, 165, 58 },
		{ 219, 166, 58 },
		{ 219, 167, 58 },
		{ 219, 168, 58 },
		{ 219, 169, 58 },
		{ 219, 170, 58 },
		{ 219, 171, 58 },
		{ 219, 172, 58 },
		{ 219, 173, 58 },
		{ 219, 174, 58 },
		{ 219, 175, 58 },
		{ 219, 176, 58 },
		{ 219, 177, 58 },
		{ 219, 178, 58 },
		{ 219, 179, 58 },
		{ 219, 180, 58 },
		{ 219, 181, 58 },
		{ 219, 182, 58 },
		{ 219, 183, 58 },
		{ 219, 184, 58 },
		{ 219, 185, 58 },
		{ 219, 186, 58 },
		{ 219, 187, 58 },
		{ 219, 188, 58 },
		{ 219, 189, 58 },
		{ 219, 190, 58 },
		{ 219, 191, 58 },
		{ 220, 159, 58 },
		{ 220, 160, 58 },
		{ 220, 161, 58 },
		{ 220, 162, 58 },
		{ 220, 163, 58 },
		{ 220, 164, 58 },
		{ 220, 165, 58 },
		{ 220, 166, 58 },
		{ 220, 167, 58 },
		{ 220, 168, 58 },
		{ 220, 169, 58 },
		{ 220, 170, 58 },
		{ 220, 171, 58 },
		{ 220, 172, 58 },
		{ 220, 173, 58 },
		{ 220, 174, 58 },
		{ 220, 175, 58 },
		{ 220, 176, 58 },
		{ 220, 177, 58 },
		{ 220, 178, 58 },
		{ 220, 179, 58 },
		{ 220, 180, 58 },
		{ 220, 181, 58 },
		{ 220, 182, 58 },
		{ 220, 183, 58 },
		{ 220, 184, 58 },
		{ 220, 185, 58 },
		{ 220, 186, 58 },
		{ 220, 187, 58 },
		{ 220, 188, 58 },
		{ 220, 189, 58 },
		{ 220, 190, 58 },
		{ 220, 191, 58 },
		{ 221, 162, 58 },
		{ 221, 163, 58 },
		{ 221, 164, 58 },
		{ 221, 165, 58 },
		{ 221, 166, 58 },
		{ 221, 167, 58 },
		{ 221, 168, 58 },
		{ 221, 169, 58 },
		{ 221, 170, 58 },
		{ 221, 171, 58 },
		{ 221, 172, 58 },
		{ 221, 173, 58 },
		{ 221, 174, 58 },
		{ 221, 175, 58 },
		{ 221, 176, 58 },
		{ 221, 177, 58 },
		{ 221, 178, 58 },
		{ 221, 179, 58 },
		{ 221, 180, 58 },
		{ 221, 181, 58 },
		{ 221, 182, 58 },
		{ 221, 183, 58 },
		{ 221, 184, 58 },
		{ 221, 185, 58 },
		{ 221, 186, 58 },
		{ 221, 187, 58 },
		{ 221, 188, 58 },
		{ 221, 189, 58 },
		{ 221, 190, 58 },
		{ 221, 191, 58 },
		{ 222, 164, 58 },
		{ 222, 165, 58 },
		{ 222, 166, 58 },
		{ 222, 167, 58 },
		{ 222, 168, 58 },
		{ 222, 169, 58 },
		{ 222, 170, 58 },
		{ 222, 171, 58 },
		{ 222, 172, 58 },
		{ 222, 173, 58 },
		{ 222, 174, 58 },
		{ 222, 175, 58 },
		{ 222, 176, 58 },
		{ 222, 177, 58 },
		{ 222, 178, 58 },
		{ 222, 179, 58 },
		{ 222, 180, 58 },
		{ 222, 181, 58 },
		{ 222, 182, 58 },
		{ 222, 183, 58 },
		{ 222, 184, 58 },
		{ 222, 185, 58 },
		{ 222, 186, 58 },
		{ 222, 187, 58 },
		{ 222, 188, 58 },
		{ 222, 189, 58 },
		{ 222, 190, 58 },
		{ 222, 191, 58 },
		{ 223, 167, 58 },
		{ 223, 168, 58 },
		{ 223, 169, 58 },
		{ 223, 170, 58 },
		{ 223, 171, 58 },
		{ 223, 172, 58 },
		{ 223, 173, 58 },
		{ 223, 174, 58 },
		{ 223, 175, 58 },
		{ 223, 176, 58 },
		{ 223, 177, 58 },
		{ 223, 178, 58 },
		{ 223, 179, 58 },
		{ 223, 180, 58 },
		{ 223, 181, 58 },
		{ 223, 182, 58 },
		{ 223, 183, 58 },
		{ 223, 184, 58 },
		{ 223, 185, 58 },
		{ 223, 186, 58 },
		{ 223, 187, 58 },
		{ 223, 188, 58 },
		{ 223, 189, 58 },
		{ 223, 190, 58 },
		{ 223, 191, 58 },
		{ 224, 170, 58 },
		{ 224, 171, 58 },
		{ 224, 172, 58 },
		{ 224, 173, 58 },
		{ 224, 174, 58 },
		{ 224, 175, 58 },
		{ 224, 176, 58 },
		{ 224, 177, 58 },
		{ 224, 178, 58 },
		{ 224, 179, 58 },
		{ 224, 180, 58 },
		{ 224, 181, 58 },
		{ 224, 182, 58 },
		{ 224, 183, 58 },
		{ 224, 184, 58 },
		{ 224, 185, 58 },
		{ 224, 186, 58 },
		{ 224, 187, 58 },
		{ 224, 188, 58 },
		{ 224, 189, 58 },
		{ 224, 190, 58 },
		{ 224, 191, 58 },
		{ 225, 172, 58 },
		{ 225, 173, 58 },
		{ 225, 174, 58 },
		{ 225, 175, 58 },
		{ 225, 176, 58 },
		{ 225, 177, 58 },
		{ 225, 178, 58 },
		{ 225, 179, 58 },
		{ 225, 180, 58 },
		{ 225, 181, 58 },
		{ 225, 182, 58 },
		{ 225, 183, 58 },
		{ 225, 184, 58 },
		{ 225, 185, 58 },
		{ 225, 186, 58 },
		{ 225, 187, 58 },
		{ 225, 188, 58 },
		{ 225, 189, 58 },
		{ 225, 190, 58 },
		{ 225, 191, 58 },
		{ 226, 175, 58 },
		{ 226, 176, 58 },
		{ 226, 177, 58 },
		{ 226, 178, 58 },
		{ 226, 179, 58 },
		{ 226, 180, 58 },
		{ 226, 181, 58 },
		{ 226, 182, 58 },
		{ 226, 183, 58 },
		{ 226, 184, 58 },
		{ 226, 185, 58 },
		{ 226, 186, 58 },
		{ 226, 187, 58 },
		{ 226, 188, 58 },
		{ 226, 189, 58 },
		{ 226, 190, 58 },
		{ 226, 191, 58 },
		{ 227, 178, 58 },
		{ 227, 179, 58 },
		{ 227, 180, 58 },
		{ 227, 181, 58 },
		{ 227, 182, 58 },
		{ 227, 183, 58 },
		{ 227, 184, 58 },
		{ 227, 185, 58 },
		{ 227, 186, 58 },
		{ 227, 187, 58 },
		{ 227, 188, 58 },
		{ 227, 189, 58 },
		{ 227, 190, 58 },
		{ 227, 191, 58 },
		{ 228, 180, 58 },
		{ 228, 181, 58 },
		{ 228, 182, 58 },
		{ 228, 183, 58 },
		{ 228, 184, 58 },
		{ 228, 185, 58 },
		{ 228, 186, 58 },
		{ 228, 187, 58 },
		{ 228, 188, 58 },
		{ 228, 189, 58 },
		{ 228, 190, 58 },
		{ 228, 191, 58 },
		{ 229, 183, 58 },
		{ 229, 184, 58 },
		{ 229, 185, 58 },
		{ 229, 186, 58 },
		{ 229, 187, 58 },
		{ 229, 188, 58 },
		{ 229, 189, 58 },
		{ 229, 190, 58 },
		{ 229, 191, 58 },
		{ 230, 186, 58 },
		{ 230, 187, 58 },
		{ 230, 188, 58 },
		{ 230, 189, 58 },
		{ 230, 190, 58 },
		{ 230, 191, 58 },
		{ 231, 189, 58 },
		{ 231, 190, 58 },
		{ 283, 324, 58 },
		{ 283, 325, 58 },
		{ 284, 323, 58 },
		{ 284, 324, 58 },
		{ 284, 325, 58 },
		{ 284, 326, 58 },
		{ 284, 327, 58 },
		{ 284, 328, 58 },
		{ 285, 323, 58 },
		{ 285, 324, 58 },
		{ 285, 325, 58 },
		{ 285, 326, 58 },
		{ 285, 327, 58 },
		{ 285, 328, 58 },
		{ 285, 329, 58 },
		{ 285, 330, 58 },
		{ 285, 331, 58 },
		{ 286, 323, 58 },
		{ 286, 324, 58 },
		{ 286, 325, 58 },
		{ 286, 326, 58 },
		{ 286, 327, 58 },
		{ 286, 328, 58 },
		{ 286, 329, 58 },
		{ 286, 330, 58 },
		{ 286, 331, 58 },
		{ 286, 332, 58 },
		{ 286, 333, 58 },
		{ 286, 334, 58 },
		{ 287, 323, 58 },
		{ 287, 324, 58 },
		{ 287, 325, 58 },
		{ 287, 326, 58 },
		{ 287, 327, 58 },
		{ 287, 328, 58 },
		{ 287, 329, 58 },
		{ 287, 330, 58 },
		{ 287, 331, 58 },
		{ 287, 332, 58 },
		{ 287, 333, 58 },
		{ 287, 334, 58 },
		{ 287, 335, 58 },
		{ 287, 336, 58 },
		{ 288, 323, 58 },
		{ 288, 324, 58 },
		{ 288, 325, 58 },
		{ 288, 326, 58 },
		{ 288, 327, 58 },
		{ 288, 328, 58 },
		{ 288, 329, 58 },
		{ 288, 330, 58 },
		{ 288, 331, 58 },
		{ 288, 332, 58 },
		{ 288, 333, 58 },
		{ 288, 334, 58 },
		{ 288, 335, 58 },
		{ 288, 336, 58 },
		{ 288, 337, 58 },
		{ 288, 338, 58 },
		{ 288, 339, 58 },
		{ 289, 323, 58 },
		{ 289, 324, 58 },
		{ 289, 325, 58 },
		{ 289, 326, 58 },
		{ 289, 327, 58 },
		{ 289, 328, 58 },
		{ 289, 329, 58 },
		{ 289, 330, 58 },
		{ 289, 331, 58 },
		{ 289, 332, 58 },
		{ 289, 333, 58 },
		{ 289, 334, 58 },
		{ 289, 335, 58 },
		{ 289, 336, 58 },
		{ 289, 337, 58 },
		{ 289, 338, 58 },
		{ 289, 339, 58 },
		{ 289, 340, 58 },
		{ 289, 341, 58 },
		{ 289, 342, 58 },
		{ 290, 323, 58 },
		{ 290, 324, 58 },
		{ 290, 325, 58 },
		{ 290, 326, 58 },
		{ 290, 327, 58 },
		{ 290, 328, 58 },
		{ 290, 329, 58 },
		{ 290, 330, 58 },
		{ 290, 331, 58 },
		{ 290, 332, 58 },
		{ 290, 333, 58 },
		{ 290, 334, 58 },
		{ 290, 335, 58 },
		{ 290, 336, 58 },
		{ 290, 337, 58 },
		{ 290, 338, 58 },
		{ 290, 339, 58 },
		{ 290, 340, 58 },
		{ 290, 341, 58 },
		{ 290, 342, 58 },
		{ 290, 343, 58 },
		{ 290, 344, 58 },
		{ 291, 323, 58 },
		{ 291, 324, 58 },
		{ 291, 325, 58 },
		{ 291, 326, 58 },
		{ 291, 327, 58 },
		{ 291, 328, 58 },
		{ 291, 329, 58 },
		{ 291, 330, 58 },
		{ 291, 331, 58 },
		{ 291, 332, 58 },
		{ 291, 333, 58 },
		{ 291, 334, 58 },
		{ 291, 335, 58 },
		{ 291, 336, 58 },
		{ 291, 337, 58 },
		{ 291, 338, 58 },
		{ 291, 339, 58 },
		{ 291, 340, 58 },
		{ 291, 341, 58 },
		{ 291, 342, 58 },
		{ 291, 343, 58 },
		{ 291, 344, 58 },
		{ 291, 345, 58 },
		{ 291, 346, 58 },
		{ 291, 347, 58 },
		{ 292, 323, 58 },
		{ 292, 324, 58 },
		{ 292, 325, 58 },
		{ 292, 326, 58 },
		{ 292, 327, 58 },
		{ 292, 328, 58 },
		{ 292, 329, 58 },
		{ 292, 330, 58 },
		{ 292, 331, 58 },
		{ 292, 332, 58 },
		{ 292, 333, 58 },
		{ 292, 334, 58 },
		{ 292, 335, 58 },
		{ 292, 336, 58 },
		{ 292, 337, 58 },
		{ 292, 338, 58 },
		{ 292, 339, 58 },
		{ 292, 340, 58 },
		{ 292, 341, 58 },
		{ 292, 342, 58 },
		{ 292, 343, 58 },
		{ 292, 344, 58 },
		{ 292, 345, 58 },
		{ 292, 346, 58 },
		{ 292, 347, 58 },
		{ 292, 348, 58 },
		{ 292, 349, 58 },
		{ 292, 350, 58 },
		{ 293, 323, 58 },
		{ 293, 324, 58 },
		{ 293, 325, 58 },
		{ 293, 326, 58 },
		{ 293, 327, 58 },
		{ 293, 328, 58 },
		{ 293, 329, 58 },
		{ 293, 330, 58 },
		{ 293, 331, 58 },
		{ 293, 332, 58 },
		{ 293, 333, 58 },
		{ 293, 334, 58 },
		{ 293, 335, 58 },
		{ 293, 336, 58 },
		{ 293, 337, 58 },
		{ 293, 338, 58 },
		{ 293, 339, 58 },
		{ 293, 340, 58 },
		{ 293, 341, 58 },
		{ 293, 342, 58 },
		{ 293, 343, 58 },
		{ 293, 344, 58 },
		{ 293, 345, 58 },
		{ 293, 346, 58 },
		{ 293, 347, 58 },
		{ 293, 348, 58 },
		{ 293, 349, 58 },
		{ 293, 350, 58 },
		{ 293, 351, 58 },
		{ 293, 352, 58 },
		{ 294, 323, 58 },
		{ 294, 324, 58 },
		{ 294, 325, 58 },
		{ 294, 326, 58 },
		{ 294, 327, 58 },
		{ 294, 328, 58 },
		{ 294, 329, 58 },
		{ 294, 330, 58 },
		{ 294, 331, 58 },
		{ 294, 332, 58 },
		{ 294, 333, 58 },
		{ 294, 334, 58 },
		{ 294, 335, 58 },
		{ 294, 336, 58 },
		{ 294, 337, 58 },
		{ 294, 338, 58 },
		{ 294, 339, 58 },
		{ 294, 340, 58 },
		{ 294, 341, 58 },
		{ 294, 342, 58 },
		{ 294, 343, 58 },
		{ 294, 344, 58 },
		{ 294, 345, 58 },
		{ 294, 346, 58 },
		{ 294, 347, 58 },
		{ 294, 348, 58 },
		{ 294, 349, 58 },
		{ 294, 350, 58 },
		{ 294, 351, 58 },
		{ 294, 352, 58 },
		{ 294, 353, 58 },
		{ 294, 354, 58 },
		{ 294, 355, 58 },
		{ 295, 323, 58 },
		{ 295, 324, 58 },
		{ 295, 325, 58 },
		{ 295, 326, 58 },
		{ 295, 327, 58 },
		{ 295, 328, 58 },
		{ 295, 329, 58 },
		{ 295, 330, 58 },
		{ 295, 331, 58 },
		{ 295, 332, 58 },
		{ 295, 333, 58 },
		{ 295, 334, 58 },
		{ 295, 335, 58 },
		{ 295, 336, 58 },
		{ 295, 337, 58 },
		{ 295, 338, 58 },
		{ 295, 339, 58 },
		{ 295, 340, 58 },
		{ 295, 341, 58 },
		{ 295, 342, 58 },
		{ 295, 343, 58 },
		{ 295, 344, 58 },
		{ 295, 345, 58 },
		{ 295, 346, 58 },
		{ 295, 347, 58 },
		{ 295, 348, 58 },
		{ 295, 349, 58 },
		{ 295, 350, 58 },
		{ 295, 351, 58 },
		{ 295, 352, 58 },
		{ 295, 353, 58 },
		{ 295, 354, 58 },
		{ 295, 355, 58 },
		{ 295, 356, 58 },
		{ 295, 357, 58 },
		{ 295, 358, 58 },
		{ 296, 323, 58 },
		{ 296, 324, 58 },
		{ 296, 325, 58 },
		{ 296, 326, 58 },
		{ 296, 327, 58 },
		{ 296, 328, 58 },
		{ 296, 329, 58 },
		{ 296, 330, 58 },
		{ 296, 331, 58 },
		{ 296, 332, 58 },
		{ 296, 333, 58 },
		{ 296, 334, 58 },
		{ 296, 335, 58 },
		{ 296, 336, 58 },
		{ 296, 337, 58 },
		{ 296, 338, 58 },
		{ 296, 339, 58 },
		{ 296, 340, 58 },
		{ 296, 341, 58 },
		{ 296, 342, 58 },
		{ 296, 343, 58 },
		{ 296, 344, 58 },
		{ 296, 345, 58 },
		{ 296, 346, 58 },
		{ 296, 347, 58 },
		{ 296, 348, 58 },
		{ 296, 349, 58 },
		{ 296, 350, 58 },
		{ 296, 351, 58 },
		{ 296, 352, 58 },
		{ 296, 353, 58 },
		{ 296, 354, 58 },
		{ 296, 355, 58 },
		{ 296, 356, 58 },
		{ 296, 357, 58 },
		{ 296, 358, 58 },
		{ 296, 359, 58 },
		{ 296, 360, 58 },
		{ 297, 324, 58 },
		{ 297, 325, 58 },
		{ 297, 326, 58 },
		{ 297, 327, 58 },
		{ 297, 328, 58 },
		{ 297, 329, 58 },
		{ 297, 330, 58 },
		{ 297, 331, 58 },
		{ 297, 332, 58 },
		{ 297, 333, 58 },
		{ 297, 334, 58 },
		{ 297, 335, 58 },
		{ 297, 336, 58 },
		{ 297, 337, 58 },
		{ 297, 338, 58 },
		{ 297, 339, 58 },
		{ 297, 340, 58 },
		{ 297, 341, 58 },
		{ 297, 342, 58 },
		{ 297, 343, 58 },
		{ 297, 344, 58 },
		{ 297, 345, 58 },
		{ 297, 346, 58 },
		{ 297, 347, 58 },
		{ 297, 348, 58 },
		{ 297, 349, 58 },
		{ 297, 350, 58 },
		{ 297, 351, 58 },
		{ 297, 352, 58 },
		{ 297, 353, 58 },
		{ 297, 354, 58 },
		{ 297, 355, 58 },
		{ 297, 356, 58 },
		{ 297, 357, 58 },
		{ 297, 358, 58 },
		{ 297, 359, 58 },
		{ 297, 360, 58 },
		{ 297, 361, 58 },
		{ 297, 362, 58 },
		{ 297, 363, 58 },
		{ 298, 324, 58 },
		{ 298, 325, 58 },
		{ 298, 326, 58 },
		{ 298, 327, 58 },
		{ 298, 328, 58 },
		{ 298, 329, 58 },
		{ 298, 330, 58 },
		{ 298, 331, 58 },
		{ 298, 332, 58 },
		{ 298, 333, 58 },
		{ 298, 334, 58 },
		{ 298, 335, 58 },
		{ 298, 336, 58 },
		{ 298, 337, 58 },
		{ 298, 338, 58 },
		{ 298, 339, 58 },
		{ 298, 340, 58 },
		{ 298, 341, 58 },
		{ 298, 342, 58 },
		{ 298, 343, 58 },
		{ 298, 344, 58 },
		{ 298, 345, 58 },
		{ 298, 346, 58 },
		{ 298, 347, 58 },
		{ 298, 348, 58 },
		{ 298, 349, 58 },
		{ 298, 350, 58 },
		{ 298, 351, 58 },
		{ 298, 352, 58 },
		{ 298, 353, 58 },
		{ 298, 354, 58 },
		{ 298, 355, 58 },
		{ 298, 356, 58 },
		{ 298, 357, 58 },
		{ 298, 358, 58 },
		{ 298, 359, 58 },
		{ 298, 360, 58 },
		{ 298, 361, 58 },
		{ 298, 362, 58 },
		{ 298, 363, 58 },
		{ 298, 364, 58 },
		{ 298, 365, 58 },
		{ 298, 366, 58 },
		{ 299, 325, 58 },
		{ 299, 326, 58 },
		{ 299, 327, 58 },
		{ 299, 328, 58 },
		{ 299, 329, 58 },
		{ 299, 330, 58 },
		{ 299, 331, 58 },
		{ 299, 332, 58 },
		{ 299, 333, 58 },
		{ 299, 334, 58 },
		{ 299, 335, 58 },
		{ 299, 336, 58 },
		{ 299, 337, 58 },
		{ 299, 338, 58 },
		{ 299, 339, 58 },
		{ 299, 340, 58 },
		{ 299, 341, 58 },
		{ 299, 342, 58 },
		{ 299, 343, 58 },
		{ 299, 344, 58 },
		{ 299, 345, 58 },
		{ 299, 346, 58 },
		{ 299, 347, 58 },
		{ 299, 348, 58 },
		{ 299, 349, 58 },
		{ 299, 350, 58 },
		{ 299, 351, 58 },
		{ 299, 352, 58 },
		{ 299, 353, 58 },
		{ 299, 354, 58 },
		{ 299, 355, 58 },
		{ 299, 356, 58 },
		{ 299, 357, 58 },
		{ 299, 358, 58 },
		{ 299, 359, 58 },
		{ 299, 360, 58 },
		{ 299, 361, 58 },
		{ 299, 362, 58 },
		{ 299, 363, 58 },
		{ 299, 364, 58 },
		{ 299, 365, 58 },
		{ 299, 366, 58 },
		{ 299, 367, 58 },
		{ 299, 368, 58 },
		{ 300, 327, 58 },
		{ 300, 328, 58 },
		{ 300, 329, 58 },
		{ 300, 330, 58 },
		{ 300, 331, 58 },
		{ 300, 332, 58 },
		{ 300, 333, 58 },
		{ 300, 334, 58 },
		{ 300, 335, 58 },
		{ 300, 336, 58 },
		{ 300, 337, 58 },
		{ 300, 338, 58 },
		{ 300, 339, 58 },
		{ 300, 340, 58 },
		{ 300, 341, 58 },
		{ 300, 342, 58 },
		{ 300, 343, 58 },
		{ 300, 344, 58 },
		{ 300, 345, 58 },
		{ 300, 346, 58 },
		{ 300, 347, 58 },
		{ 300, 348, 58 },
		{ 300, 349, 58 },
		{ 300, 350, 58 },
		{ 300, 351, 58 },
		{ 300, 352, 58 },
		{ 300, 353, 58 },
		{ 300, 354, 58 },
		{ 300, 355, 58 },
		{ 300, 356, 58 },
		{ 300, 357, 58 },
		{ 300, 358, 58 },
		{ 300, 359, 58 },
		{ 300, 360, 58 },
		{ 300, 361, 58 },
		{ 300, 362, 58 },
		{ 300, 363, 58 },
		{ 300, 364, 58 },
		{ 300, 365, 58 },
		{ 300, 366, 58 },
		{ 300, 367, 58 },
		{ 300, 368, 58 },
		{ 300, 369, 58 },
		{ 300, 370, 58 },
		{ 300, 371, 58 },
		{ 301, 328, 58 },
		{ 301, 329, 58 },
		{ 301, 330, 58 },
		{ 301, 331, 58 },
		{ 301, 332, 58 },
		{ 301, 333, 58 },
		{ 301, 334, 58 },
		{ 301, 335, 58 },
		{ 301, 336, 58 },
		{ 301, 337, 58 },
		{ 301, 338, 58 },
		{ 301, 339, 58 },
		{ 301, 340, 58 },
		{ 301, 341, 58 },
		{ 301, 342, 58 },
		{ 301, 343, 58 },
		{ 301, 344, 58 },
		{ 301, 345, 58 },
		{ 301, 346, 58 },
		{ 301, 347, 58 },
		{ 301, 348, 58 },
		{ 301, 349, 58 },
		{ 301, 350, 58 },
		{ 301, 351, 58 },
		{ 301, 352, 58 },
		{ 301, 353, 58 },
		{ 301, 354, 58 },
		{ 301, 355, 58 },
		{ 301, 356, 58 },
		{ 301, 357, 58 },
		{ 301, 358, 58 },
		{ 301, 359, 58 },
		{ 301, 360, 58 },
		{ 301, 361, 58 },
		{ 301, 362, 58 },
		{ 301, 363, 58 },
		{ 301, 364, 58 },
		{ 301, 365, 58 },
		{ 301, 366, 58 },
		{ 301, 367, 58 },
		{ 301, 368, 58 },
		{ 301, 369, 58 },
		{ 301, 370, 58 },
		{ 301, 371, 58 },
		{ 301, 372, 58 },
		{ 301, 373, 58 },
		{ 301, 374, 58 },
		{ 302, 330, 58 },
		{ 302, 331, 58 },
		{ 302, 332, 58 },
		{ 302, 333, 58 },
		{ 302, 334, 58 },
		{ 302, 335, 58 },
		{ 302, 336, 58 },
		{ 302, 337, 58 },
		{ 302, 338, 58 },
		{ 302, 339, 58 },
		{ 302, 340, 58 },
		{ 302, 341, 58 },
		{ 302, 342, 58 },
		{ 302, 343, 58 },
		{ 302, 344, 58 },
		{ 302, 345, 58 },
		{ 302, 346, 58 },
		{ 302, 347, 58 },
		{ 302, 348, 58 },
		{ 302, 349, 58 },
		{ 302, 350, 58 },
		{ 302, 351, 58 },
		{ 302, 352, 58 },
		{ 302, 353, 58 },
		{ 302, 354, 58 },
		{ 302, 355, 58 },
		{ 302, 356, 58 },
		{ 302, 357, 58 },
		{ 302, 358, 58 },
		{ 302, 359, 58 },
		{ 302, 360, 58 },
		{ 302, 361, 58 },
		{ 302, 362, 58 },
		{ 302, 363, 58 },
		{ 302, 364, 58 },
		{ 302, 365, 58 },
		{ 302, 366, 58 },
		{ 302, 367, 58 },
		{ 302, 368, 58 },
		{ 302, 369, 58 },
		{ 302, 370, 58 },
		{ 302, 371, 58 },
		{ 302, 372, 58 },
		{ 302, 373, 58 },
		{ 302, 374, 58 },
		{ 302, 375, 58 },
		{ 302, 376, 58 },
		{ 303, 331, 58 },
		{ 303, 332, 58 },
		{ 303, 333, 58 },
		{ 303, 334, 58 },
		{ 303, 335, 58 },
		{ 303, 336, 58 },
		{ 303, 337, 58 },
		{ 303, 338, 58 },
		{ 303, 339, 58 },
		{ 303, 340, 58 },
		{ 303, 341, 58 },
		{ 303, 342, 58 },
		{ 303, 343, 58 },
		{ 303, 344, 58 },
		{ 303, 345, 58 },
		{ 303, 346, 58 },
		{ 303, 347, 58 },
		{ 303, 348, 58 },
		{ 303, 349, 58 },
		{ 303, 350, 58 },
		{ 303, 351, 58 },
		{ 303, 352, 58 },
		{ 303, 353, 58 },
		{ 303, 354, 58 },
		{ 303, 355, 58 },
		{ 303, 356, 58 },
		{ 303, 357, 58 },
		{ 303, 358, 58 },
		{ 303, 359, 58 },
		{ 303, 360, 58 },
		{ 303, 361, 58 },
		{ 303, 362, 58 },
		{ 303, 363, 58 },
		{ 303, 364, 58 },
		{ 303, 365, 58 },
		{ 303, 366, 58 },
		{ 303, 367, 58 },
		{ 303, 368, 58 },
		{ 303, 369, 58 },
		{ 303, 370, 58 },
		{ 303, 371, 58 },
		{ 303, 372, 58 },
		{ 303, 373, 58 },
		{ 303, 374, 58 },
		{ 303, 375, 58 },
		{ 303, 376, 58 },
		{ 303, 377, 58 },
		{ 303, 378, 58 },
		{ 303, 379, 58 },
		{ 304, 333, 58 },
		{ 304, 334, 58 },
		{ 304, 335, 58 },
		{ 304, 336, 58 },
		{ 304, 337, 58 },
		{ 304, 338, 58 },
		{ 304, 339, 58 },
		{ 304, 340, 58 },
		{ 304, 341, 58 },
		{ 304, 342, 58 },
		{ 304, 343, 58 },
		{ 304, 344, 58 },
		{ 304, 345, 58 },
		{ 304, 346, 58 },
		{ 304, 347, 58 },
		{ 304, 348, 58 },
		{ 304, 349, 58 },
		{ 304, 350, 58 },
		{ 304, 351, 58 },
		{ 304, 352, 58 },
		{ 304, 353, 58 },
		{ 304, 354, 58 },
		{ 304, 355, 58 },
		{ 304, 356, 58 },
		{ 304, 357, 58 },
		{ 304, 358, 58 },
		{ 304, 359, 58 },
		{ 304, 360, 58 },
		{ 304, 361, 58 },
		{ 304, 362, 58 },
		{ 304, 363, 58 },
		{ 304, 364, 58 },
		{ 304, 365, 58 },
		{ 304, 366, 58 },
		{ 304, 367, 58 },
		{ 304, 368, 58 },
		{ 304, 369, 58 },
		{ 304, 370, 58 },
		{ 304, 371, 58 },
		{ 304, 372, 58 },
		{ 304, 373, 58 },
		{ 304, 374, 58 },
		{ 304, 375, 58 },
		{ 304, 376, 58 },
		{ 304, 377, 58 },
		{ 304, 378, 58 },
		{ 304, 379, 58 },
		{ 304, 380, 58 },
		{ 304, 381, 58 },
		{ 304, 382, 58 },
		{ 305, 334, 58 },
		{ 305, 335, 58 },
		{ 305, 336, 58 },
		{ 305, 337, 58 },
		{ 305, 338, 58 },
		{ 305, 339, 58 },
		{ 305, 340, 58 },
		{ 305, 341, 58 },
		{ 305, 342, 58 },
		{ 305, 343, 58 },
		{ 305, 344, 58 },
		{ 305, 345, 58 },
		{ 305, 346, 58 },
		{ 305, 347, 58 },
		{ 305, 348, 58 },
		{ 305, 349, 58 },
		{ 305, 350, 58 },
		{ 305, 351, 58 },
		{ 305, 352, 58 },
		{ 305, 353, 58 },
		{ 305, 354, 58 },
		{ 305, 355, 58 },
		{ 305, 356, 58 },
		{ 305, 357, 58 },
		{ 305, 358, 58 },
		{ 305, 359, 58 },
		{ 305, 360, 58 },
		{ 305, 361, 58 },
		{ 305, 362, 58 },
		{ 305, 363, 58 },
		{ 305, 364, 58 },
		{ 305, 365, 58 },
		{ 305, 366, 58 },
		{ 305, 367, 58 },
		{ 305, 368, 58 },
		{ 305, 369, 58 },
		{ 305, 370, 58 },
		{ 305, 371, 58 },
		{ 305, 372, 58 },
		{ 305, 373, 58 },
		{ 305, 374, 58 },
		{ 305, 375, 58 },
		{ 305, 376, 58 },
		{ 305, 377, 58 },
		{ 305, 378, 58 },
		{ 305, 379, 58 },
		{ 305, 380, 58 },
		{ 305, 381, 58 },
		{ 305, 382, 58 },
		{ 305, 383, 58 },
		{ 305, 384, 58 },
		{ 306, 336, 58 },
		{ 306, 337, 58 },
		{ 306, 338, 58 },
		{ 306, 339, 58 },
		{ 306, 340, 58 },
		{ 306, 341, 58 },
		{ 306, 342, 58 },
		{ 306, 343, 58 },
		{ 306, 344, 58 },
		{ 306, 345, 58 },
		{ 306, 346, 58 },
		{ 306, 347, 58 },
		{ 306, 348, 58 },
		{ 306, 349, 58 },
		{ 306, 350, 58 },
		{ 306, 351, 58 },
		{ 306, 352, 58 },
		{ 306, 353, 58 },
		{ 306, 354, 58 },
		{ 306, 355, 58 },
		{ 306, 356, 58 },
		{ 306, 357, 58 },
		{ 306, 358, 58 },
		{ 306, 359, 58 },
		{ 306, 360, 58 },
		{ 306, 361, 58 },
		{ 306, 362, 58 },
		{ 306, 363, 58 },
		{ 306, 364, 58 },
		{ 306, 365, 58 },
		{ 306, 366, 58 },
		{ 306, 367, 58 },
		{ 306, 368, 58 },
		{ 306, 369, 58 },
		{ 306, 370, 58 },
		{ 306, 371, 58 },
		{ 306, 372, 58 },
		{ 306, 373, 58 },
		{ 306, 374, 58 },
		{ 306, 375, 58 },
		{ 306, 376, 58 },
		{ 306, 377, 58 },
		{ 306, 378, 58 },
		{ 306, 379, 58 },
		{ 306, 380, 58 },
		{ 306, 381, 58 },
		{ 306, 382, 58 },
		{ 306, 383, 58 },
		{ 306, 384, 58 },
		{ 306, 385, 58 },
		{ 306, 386, 58 },
		{ 306, 387, 58 },
		{ 307, 338, 58 },
		{ 307, 339, 58 },
		{ 307, 340, 58 },
		{ 307, 341, 58 },
		{ 307, 342, 58 },
		{ 307, 343, 58 },
		{ 307, 344, 58 },
		{ 307, 345, 58 },
		{ 307, 346, 58 },
		{ 307, 347, 58 },
		{ 307, 348, 58 },
		{ 307, 349, 58 },
		{ 307, 350, 58 },
		{ 307, 351, 58 },
		{ 307, 352, 58 },
		{ 307, 353, 58 },
		{ 307, 354, 58 },
		{ 307, 355, 58 },
		{ 307, 356, 58 },
		{ 307, 357, 58 },
		{ 307, 358, 58 },
		{ 307, 359, 58 },
		{ 307, 360, 58 },
		{ 307, 361, 58 },
		{ 307, 362, 58 },
		{ 307, 363, 58 },
		{ 307, 364, 58 },
		{ 307, 365, 58 },
		{ 307, 366, 58 },
		{ 307, 367, 58 },
		{ 307, 368, 58 },
		{ 307, 369, 58 },
		{ 307, 370, 58 },
		{ 307, 371, 58 },
		{ 307, 372, 58 },
		{ 307, 373, 58 },
		{ 307, 374, 58 },
		{ 307, 375, 58 },
		{ 307, 376, 58 },
		{ 307, 377, 58 },
		{ 307, 378, 58 },
		{ 307, 379, 58 },
		{ 307, 380, 58 },
		{ 307, 381, 58 },
		{ 307, 382, 58 },
		{ 307, 383, 58 },
		{ 307, 384, 58 },
		{ 307, 385, 58 },
		{ 307, 386, 58 },
		{ 307, 387, 58 },
		{ 307, 388, 58 },
		{ 307, 389, 58 },
		{ 307, 390, 58 },
		{ 308, 339, 58 },
		{ 308, 340, 58 },
		{ 308, 341, 58 },
		{ 308, 342, 58 },
		{ 308, 343, 58 },
		{ 308, 344, 58 },
		{ 308, 345, 58 },
		{ 308, 346, 58 },
		{ 308, 347, 58 },
		{ 308, 348, 58 },
		{ 308, 349, 58 },
		{ 308, 350, 58 },
		{ 308, 351, 58 },
		{ 308, 352, 58 },
		{ 308, 353, 58 },
		{ 308, 354, 58 },
		{ 308, 355, 58 },
		{ 308, 356, 58 },
		{ 308, 357, 58 },
		{ 308, 358, 58 },
		{ 308, 359, 58 },
		{ 308, 360, 58 },
		{ 308, 361, 58 },
		{ 308, 362, 58 },
		{ 308, 363, 58 },
		{ 308, 364, 58 },
		{ 308, 365, 58 },
		{ 308, 366, 58 },
		{ 308, 367, 58 },
		{ 308, 368, 58 },
		{ 308, 369, 58 },
		{ 308, 370, 58 },
		{ 308, 371, 58 },
		{ 308, 372, 58 },
		{ 308, 373, 58 },
		{ 308, 374, 58 },
		{ 308, 375, 58 },
		{ 308, 376, 58 },
		{ 308, 377, 58 },
		{ 308, 378, 58 },
		{ 308, 379, 58 },
		{ 308, 380, 58 },
		{ 308, 381, 58 },
		{ 308, 382, 58 },
		{ 308, 383, 58 },
		{ 308, 384, 58 },
		{ 308, 385, 58 },
		{ 308, 386, 58 },
		{ 308, 387, 58 },
		{ 308, 388, 58 },
		{ 308, 389, 58 },
		{ 308, 390, 58 },
		{ 308, 391, 58 },
		{ 308, 392, 58 },
		{ 309, 341, 58 },
		{ 309, 342, 58 },
		{ 309, 343, 58 },
		{ 309, 344, 58 },
		{ 309, 345, 58 },
		{ 309, 346, 58 },
		{ 309, 347, 58 },
		{ 309, 348, 58 },
		{ 309, 349, 58 },
		{ 309, 350, 58 },
		{ 309, 351, 58 },
		{ 309, 352, 58 },
		{ 309, 353, 58 },
		{ 309, 354, 58 },
		{ 309, 355, 58 },
		{ 309, 356, 58 },
		{ 309, 357, 58 },
		{ 309, 358, 58 },
		{ 309, 359, 58 },
		{ 309, 360, 58 },
		{ 309, 361, 58 },
		{ 309, 362, 58 },
		{ 309, 363, 58 },
		{ 309, 364, 58 },
		{ 309, 365, 58 },
		{ 309, 366, 58 },
		{ 309, 367, 58 },
		{ 309, 368, 58 },
		{ 309, 369, 58 },
		{ 309, 370, 58 },
		{ 309, 371, 58 },
		{ 309, 372, 58 },
		{ 309, 373, 58 },
		{ 309, 374, 58 },
		{ 309, 375, 58 },
		{ 309, 376, 58 },
		{ 309, 377, 58 },
		{ 309, 378, 58 },
		{ 309, 379, 58 },
		{ 309, 380, 58 },
		{ 309, 381, 58 },
		{ 309, 382, 58 },
		{ 309, 383, 58 },
		{ 309, 384, 58 },
		{ 309, 385, 58 },
		{ 309, 386, 58 },
		{ 309, 387, 58 },
		{ 309, 388, 58 },
		{ 309, 389, 58 },
		{ 309, 390, 58 },
		{ 309, 391, 58 },
		{ 309, 392, 58 },
		{ 309, 393, 58 },
		{ 309, 394, 58 },
		{ 309, 395, 58 },
		{ 310, 342, 58 },
		{ 310, 343, 58 },
		{ 310, 344, 58 },
		{ 310, 345, 58 },
		{ 310, 346, 58 },
		{ 310, 347, 58 },
		{ 310, 348, 58 },
		{ 310, 349, 58 },
		{ 310, 350, 58 },
		{ 310, 351, 58 },
		{ 310, 352, 58 },
		{ 310, 353, 58 },
		{ 310, 354, 58 },
		{ 310, 355, 58 },
		{ 310, 356, 58 },
		{ 310, 357, 58 },
		{ 310, 358, 58 },
		{ 310, 359, 58 },
		{ 310, 360, 58 },
		{ 310, 361, 58 },
		{ 310, 362, 58 },
		{ 310, 363, 58 },
		{ 310, 364, 58 },
		{ 310, 365, 58 },
		{ 310, 366, 58 },
		{ 310, 367, 58 },
		{ 310, 368, 58 },
		{ 310, 369, 58 },
		{ 310, 370, 58 },
		{ 310, 371, 58 },
		{ 310, 372, 58 },
		{ 310, 373, 58 },
		{ 310, 374, 58 },
		{ 310, 375, 58 },
		{ 310, 376, 58 },
		{ 310, 377, 58 },
		{ 310, 378, 58 },
		{ 310, 379, 58 },
		{ 310, 380, 58 },
		{ 310, 381, 58 },
		{ 310, 382, 58 },
		{ 310, 383, 58 },
		{ 310, 384, 58 },
		{ 310, 385, 58 },
		{ 310, 386, 58 },
		{ 310, 387, 58 },
		{ 310, 388, 58 },
		{ 310, 389, 58 },
		{ 310, 390, 58 },
		{ 310, 391, 58 },
		{ 310, 392, 58 },
		{ 310, 393, 58 },
		{ 310, 394, 58 },
		{ 310, 395, 58 },
		{ 310, 396, 58 },
		{ 310, 397, 58 },
		{ 310, 398, 58 },
		{ 311, 344, 58 },
		{ 311, 345, 58 },
		{ 311, 346, 58 },
		{ 311, 347, 58 },
		{ 311, 348, 58 },
		{ 311, 349, 58 },
		{ 311, 350, 58 },
		{ 311, 351, 58 },
		{ 311, 352, 58 },
		{ 311, 353, 58 },
		{ 311, 354, 58 },
		{ 311, 355, 58 },
		{ 311, 356, 58 },
		{ 311, 357, 58 },
		{ 311, 358, 58 },
		{ 311, 359, 58 },
		{ 311, 360, 58 },
		{ 311, 361, 58 },
		{ 311, 362, 58 },
		{ 311, 363, 58 },
		{ 311, 364, 58 },
		{ 311, 365, 58 },
		{ 311, 366, 58 },
		{ 311, 367, 58 },
		{ 311, 368, 58 },
		{ 311, 369, 58 },
		{ 311, 370, 58 },
		{ 311, 371, 58 },
		{ 311, 372, 58 },
		{ 311, 373, 58 },
		{ 311, 374, 58 },
		{ 311, 375, 58 },
		{ 311, 376, 58 },
		{ 311, 377, 58 },
		{ 311, 378, 58 },
		{ 311, 379, 58 },
		{ 311, 380, 58 },
		{ 311, 381, 58 },
		{ 311, 382, 58 },
		{ 311, 383, 58 },
		{ 311, 384, 58 },
		{ 311, 385, 58 },
		{ 311, 386, 58 },
		{ 311, 387, 58 },
		{ 311, 388, 58 },
		{ 311, 389, 58 },
		{ 311, 390, 58 },
		{ 311, 391, 58 },
		{ 311, 392, 58 },
		{ 311, 393, 58 },
		{ 311, 394, 58 },
		{ 311, 395, 58 },
		{ 311, 396, 58 },
		{ 311, 397, 58 },
		{ 311, 398, 58 },
		{ 311, 399, 58 },
		{ 311, 400, 58 },
		{ 312, 346, 58 },
		{ 312, 347, 58 },
		{ 312, 348, 58 },
		{ 312, 349, 58 },
		{ 312, 350, 58 },
		{ 312, 351, 58 },
		{ 312, 352, 58 },
		{ 312, 353, 58 },
		{ 312, 354, 58 },
		{ 312, 355, 58 },
		{ 312, 356, 58 },
		{ 312, 357, 58 },
		{ 312, 358, 58 },
		{ 312, 359, 58 },
		{ 312, 360, 58 },
		{ 312, 361, 58 },
		{ 312, 362, 58 },
		{ 312, 363, 58 },
		{ 312, 364, 58 },
		{ 312, 365, 58 },
		{ 312, 366, 58 },
		{ 312, 367, 58 },
		{ 312, 368, 58 },
		{ 312, 369, 58 },
		{ 312, 370, 58 },
		{ 312, 371, 58 },
		{ 312, 372, 58 },
		{ 312, 373, 58 },
		{ 312, 374, 58 },
		{ 312, 375, 58 },
		{ 312, 376, 58 },
		{ 312, 377, 58 },
		{ 312, 378, 58 },
		{ 312, 379, 58 },
		{ 312, 380, 58 },
		{ 312, 381, 58 },
		{ 312, 382, 58 },
		{ 312, 383, 58 },
		{ 312, 384, 58 },
		{ 312, 385, 58 },
		{ 312, 386, 58 },
		{ 312, 387, 58 },
		{ 312, 388, 58 },
		{ 312, 389, 58 },
		{ 312, 390, 58 },
		{ 312, 391, 58 },
		{ 312, 392, 58 },
		{ 312, 393, 58 },
		{ 312, 394, 58 },
		{ 312, 395, 58 },
		{ 312, 396, 58 },
		{ 312, 397, 58 },
		{ 312, 398, 58 },
		{ 312, 399, 58 },
		{ 312, 400, 58 },
		{ 312, 401, 58 },
		{ 312, 402, 58 },
		{ 312, 403, 58 },
		{ 313, 347, 58 },
		{ 313, 348, 58 },
		{ 313, 349, 58 },
		{ 313, 350, 58 },
		{ 313, 351, 58 },
		{ 313, 352, 58 },
		{ 313, 353, 58 },
		{ 313, 354, 58 },
		{ 313, 355, 58 },
		{ 313, 356, 58 },
		{ 313, 357, 58 },
		{ 313, 358, 58 },
		{ 313, 359, 58 },
		{ 313, 360, 58 },
		{ 313, 361, 58 },
		{ 313, 362, 58 },
		{ 313, 363, 58 },
		{ 313, 364, 58 },
		{ 313, 365, 58 },
		{ 313, 366, 58 },
		{ 313, 367, 58 },
		{ 313, 368, 58 },
		{ 313, 369, 58 },
		{ 313, 370, 58 },
		{ 313, 371, 58 },
		{ 313, 372, 58 },
		{ 313, 373, 58 },
		{ 313, 374, 58 },
		{ 313, 375, 58 },
		{ 313, 376, 58 },
		{ 313, 377, 58 },
		{ 313, 378, 58 },
		{ 313, 379, 58 },
		{ 313, 380, 58 },
		{ 313, 381, 58 },
		{ 313, 382, 58 },
		{ 313, 383, 58 },
		{ 313, 384, 58 },
		{ 313, 385, 58 },
		{ 313, 386, 58 },
		{ 313, 387, 58 },
		{ 313, 388, 58 },
		{ 313, 389, 58 },
		{ 313, 390, 58 },
		{ 313, 391, 58 },
		{ 313, 392, 58 },
		{ 313, 393, 58 },
		{ 313, 394, 58 },
		{ 313, 395, 58 },
		{ 313, 396, 58 },
		{ 313, 397, 58 },
		{ 313, 398, 58 },
		{ 313, 399, 58 },
		{ 313, 400, 58 },
		{ 313, 401, 58 },
		{ 313, 402, 58 },
		{ 313, 403, 58 },
		{ 313, 404, 58 },
		{ 313, 405, 58 },
		{ 313, 406, 58 },
		{ 314, 349, 58 },
		{ 314, 350, 58 },
		{ 314, 351, 58 },
		{ 314, 352, 58 },
		{ 314, 353, 58 },
		{ 314, 354, 58 },
		{ 314, 355, 58 },
		{ 314, 356, 58 },
		{ 314, 357, 58 },
		{ 314, 358, 58 },
		{ 314, 359, 58 },
		{ 314, 360, 58 },
		{ 314, 361, 58 },
		{ 314, 362, 58 },
		{ 314, 363, 58 },
		{ 314, 364, 58 },
		{ 314, 365, 58 },
		{ 314, 366, 58 },
		{ 314, 367, 58 },
		{ 314, 368, 58 },
		{ 314, 369, 58 },
		{ 314, 370, 58 },
		{ 314, 371, 58 },
		{ 314, 372, 58 },
		{ 314, 373, 58 },
		{ 314, 374, 58 },
		{ 314, 375, 58 },
		{ 314, 376, 58 },
		{ 314, 377, 58 },
		{ 314, 378, 58 },
		{ 314, 379, 58 },
		{ 314, 380, 58 },
		{ 314, 381, 58 },
		{ 314, 382, 58 },
		{ 314, 383, 58 },
		{ 314, 384, 58 },
		{ 314, 385, 58 },
		{ 314, 386, 58 },
		{ 314, 387, 58 },
		{ 314, 388, 58 },
		{ 314, 389, 58 },
		{ 314, 390, 58 },
		{ 314, 391, 58 },
		{ 314, 392, 58 },
		{ 314, 393, 58 },
		{ 314, 394, 58 },
		{ 314, 395, 58 },
		{ 314, 396, 58 },
		{ 314, 397, 58 },
		{ 314, 398, 58 },
		{ 314, 399, 58 },
		{ 314, 400, 58 },
		{ 314, 401, 58 },
		{ 314, 402, 58 },
		{ 314, 403, 58 },
		{ 314, 404, 58 },
		{ 314, 405, 58 },
		{ 314, 406, 58 },
		{ 314, 407, 58 },
		{ 314, 408, 58 },
		{ 315, 350, 58 },
		{ 315, 351, 58 },
		{ 315, 352, 58 },
		{ 315, 353, 58 },
		{ 315, 354, 58 },
		{ 315, 355, 58 },
		{ 315, 356, 58 },
		{ 315, 357, 58 },
		{ 315, 358, 58 },
		{ 315, 359, 58 },
		{ 315, 360, 58 },
		{ 315, 361, 58 },
		{ 315, 362, 58 },
		{ 315, 363, 58 },
		{ 315, 364, 58 },
		{ 315, 365, 58 },
		{ 315, 366, 58 },
		{ 315, 367, 58 },
		{ 315, 368, 58 },
		{ 315, 369, 58 },
		{ 315, 370, 58 },
		{ 315, 371, 58 },
		{ 315, 372, 58 },
		{ 315, 373, 58 },
		{ 315, 374, 58 },
		{ 315, 375, 58 },
		{ 315, 376, 58 },
		{ 315, 377, 58 },
		{ 315, 378, 58 },
		{ 315, 379, 58 },
		{ 315, 380, 58 },
		{ 315, 381, 58 },
		{ 315, 382, 58 },
		{ 315, 383, 58 },
		{ 315, 384, 58 },
		{ 315, 385, 58 },
		{ 315, 386, 58 },
		{ 315, 387, 58 },
		{ 315, 388, 58 },
		{ 315, 389, 58 },
		{ 315, 390, 58 },
		{ 315, 391, 58 },
		{ 315, 392, 58 },
		{ 315, 393, 58 },
		{ 315, 394, 58 },
		{ 315, 395, 58 },
		{ 315, 396, 58 },
		{ 315, 397, 58 },
		{ 315, 398, 58 },
		{ 315, 399, 58 },
		{ 315, 400, 58 },
		{ 315, 401, 58 },
		{ 315, 402, 58 },
		{ 315, 403, 58 },
		{ 315, 404, 58 },
		{ 315, 405, 58 },
		{ 315, 406, 58 },
		{ 315, 407, 58 },
		{ 315, 408, 58 },
		{ 315, 409, 58 },
		{ 315, 410, 58 },
		{ 315, 411, 58 },
		{ 316, 352, 58 },
		{ 316, 353, 58 },
		{ 316, 354, 58 },
		{ 316, 355, 58 },
		{ 316, 356, 58 },
		{ 316, 357, 58 },
		{ 316, 358, 58 },
		{ 316, 359, 58 },
		{ 316, 360, 58 },
		{ 316, 361, 58 },
		{ 316, 362, 58 },
		{ 316, 363, 58 },
		{ 316, 364, 58 },
		{ 316, 365, 58 },
		{ 316, 366, 58 },
		{ 316, 367, 58 },
		{ 316, 368, 58 },
		{ 316, 369, 58 },
		{ 316, 370, 58 },
		{ 316, 371, 58 },
		{ 316, 372, 58 },
		{ 316, 373, 58 },
		{ 316, 374, 58 },
		{ 316, 375, 58 },
		{ 316, 376, 58 },
		{ 316, 377, 58 },
		{ 316, 378, 58 },
		{ 316, 379, 58 },
		{ 316, 380, 58 },
		{ 316, 381, 58 },
		{ 316, 382, 58 },
		{ 316, 383, 58 },
		{ 316, 384, 58 },
		{ 316, 385, 58 },
		{ 316, 386, 58 },
		{ 316, 387, 58 },
		{ 316, 388, 58 },
		{ 316, 389, 58 },
		{ 316, 390, 58 },
		{ 316, 391, 58 },
		{ 316, 392, 58 },
		{ 316, 393, 58 },
		{ 316, 394, 58 },
		{ 316, 395, 58 },
		{ 316, 396, 58 },
		{ 316, 397, 58 },
		{ 316, 398, 58 },
		{ 316, 399, 58 },
		{ 316, 400, 58 },
		{ 316, 401, 58 },
		{ 316, 402, 58 },
		{ 316, 403, 58 },
		{ 316, 404, 58 },
		{ 316, 405, 58 },
		{ 316, 406, 58 },
		{ 316, 407, 58 },
		{ 316, 408, 58 },
		{ 316, 409, 58 },
		{ 316, 410, 58 },
		{ 316, 411, 58 },
		{ 316, 412, 58 },
		{ 316, 413, 58 },
		{ 316, 414, 58 },
		{ 317, 354, 58 },
		{ 317, 355, 58 },
		{ 317, 356, 58 },
		{ 317, 357, 58 },
		{ 317, 358, 58 },
		{ 317, 359, 58 },
		{ 317, 360, 58 },
		{ 317, 361, 58 },
		{ 317, 362, 58 },
		{ 317, 363, 58 },
		{ 317, 364, 58 },
		{ 317, 365, 58 },
		{ 317, 366, 58 },
		{ 317, 367, 58 },
		{ 317, 368, 58 },
		{ 317, 369, 58 },
		{ 317, 370, 58 },
		{ 317, 371, 58 },
		{ 317, 372, 58 },
		{ 317, 373, 58 },
		{ 317, 374, 58 },
		{ 317, 375, 58 },
		{ 317, 376, 58 },
		{ 317, 377, 58 },
		{ 317, 378, 58 },
		{ 317, 379, 58 },
		{ 317, 380, 58 },
		{ 317, 381, 58 },
		{ 317, 382, 58 },
		{ 317, 383, 58 },
		{ 317, 384, 58 },
		{ 317, 385, 58 },
		{ 317, 386, 58 },
		{ 317, 387, 58 },
		{ 317, 388, 58 },
		{ 317, 389, 58 },
		{ 317, 390, 58 },
		{ 317, 391, 58 },
		{ 317, 392, 58 },
		{ 317, 393, 58 },
		{ 317, 394, 58 },
		{ 317, 395, 58 },
		{ 317, 396, 58 },
		{ 317, 397, 58 },
		{ 317, 398, 58 },
		{ 317, 399, 58 },
		{ 317, 400, 58 },
		{ 317, 401, 58 },
		{ 317, 402, 58 },
		{ 317, 403, 58 },
		{ 317, 404, 58 },
		{ 317, 405, 58 },
		{ 317, 406, 58 },
		{ 317, 407, 58 },
		{ 317, 408, 58 },
		{ 317, 409, 58 },
		{ 317, 410, 58 },
		{ 317, 411, 58 },
		{ 317, 412, 58 },
		{ 317, 413, 58 },
		{ 317, 414, 58 },
		{ 317, 415, 58 },
		{ 317, 416, 58 },
		{ 318, 355, 58 },
		{ 318, 356, 58 },
		{ 318, 357, 58 },
		{ 318, 358, 58 },
		{ 318, 359, 58 },
		{ 318, 360, 58 },
		{ 318, 361, 58 },
		{ 318, 362, 58 },
		{ 318, 363, 58 },
		{ 318, 364, 58 },
		{ 318, 365, 58 },
		{ 318, 366, 58 },
		{ 318, 367, 58 },
		{ 318, 368, 58 },
		{ 318, 369, 58 },
		{ 318, 370, 58 },
		{ 318, 371, 58 },
		{ 318, 372, 58 },
		{ 318, 373, 58 },
		{ 318, 374, 58 },
		{ 318, 375, 58 },
		{ 318, 376, 58 },
		{ 318, 377, 58 },
		{ 318, 378, 58 },
		{ 318, 379, 58 },
		{ 318, 380, 58 },
		{ 318, 381, 58 },
		{ 318, 382, 58 },
		{ 318, 383, 58 },
		{ 318, 384, 58 },
		{ 318, 385, 58 },
		{ 318, 386, 58 },
		{ 318, 387, 58 },
		{ 318, 388, 58 },
		{ 318, 389, 58 },
		{ 318, 390, 58 },
		{ 318, 391, 58 },
		{ 318, 392, 58 },
		{ 318, 393, 58 },
		{ 318, 394, 58 },
		{ 318, 395, 58 },
		{ 318, 396, 58 },
		{ 318, 397, 58 },
		{ 318, 398, 58 },
		{ 318, 399, 58 },
		{ 318, 400, 58 },
		{ 318, 401, 58 },
		{ 318, 402, 58 },
		{ 318, 403, 58 },
		{ 318, 404, 58 },
		{ 318, 405, 58 },
		{ 318, 406, 58 },
		{ 318, 407, 58 },
		{ 318, 408, 58 },
		{ 318, 409, 58 },
		{ 318, 410, 58 },
		{ 318, 411, 58 },
		{ 318, 412, 58 },
		{ 318, 413, 58 },
		{ 318, 414, 58 },
		{ 318, 415, 58 },
		{ 318, 416, 58 },
		{ 318, 417, 58 },
		{ 318, 418, 58 },
		{ 318, 419, 58 },
		{ 319, 357, 58 },
		{ 319, 358, 58 },
		{ 319, 359, 58 },
		{ 319, 360, 58 },
		{ 319, 361, 58 },
		{ 319, 362, 58 },
		{ 319, 363, 58 },
		{ 319, 364, 58 },
		{ 319, 365, 58 },
		{ 319, 366, 58 },
		{ 319, 367, 58 },
		{ 319, 368, 58 },
		{ 319, 369, 58 },
		{ 319, 370, 58 },
		{ 319, 371, 58 },
		{ 319, 372, 58 },
		{ 319, 373, 58 },
		{ 319, 374, 58 },
		{ 319, 375, 58 },
		{ 319, 376, 58 },
		{ 319, 377, 58 },
		{ 319, 378, 58 },
		{ 319, 379, 58 },
		{ 319, 380, 58 },
		{ 319, 381, 58 },
		{ 319, 382, 58 },
		{ 319, 383, 58 },
		{ 319, 384, 58 },
		{ 319, 385, 58 },
		{ 319, 386, 58 },
		{ 319, 387, 58 },
		{ 319, 388, 58 },
		{ 319, 389, 58 },
		{ 319, 390, 58 },
		{ 319, 391, 58 },
		{ 319, 392, 58 },
		{ 319, 393, 58 },
		{ 319, 394, 58 },
		{ 319, 395, 58 },
		{ 319, 396, 58 },
		{ 319, 397, 58 },
		{ 319, 398, 58 },
		{ 319, 399, 58 },
		{ 319, 400, 58 },
		{ 319, 401, 58 },
		{ 319, 402, 58 },
		{ 319, 403, 58 },
		{ 319, 404, 58 },
		{ 319, 405, 58 },
		{ 319, 406, 58 },
		{ 319, 407, 58 },
		{ 319, 408, 58 },
		{ 319, 409, 58 },
		{ 319, 410, 58 },
		{ 319, 411, 58 },
		{ 319, 412, 58 },
		{ 319, 413, 58 },
		{ 319, 414, 58 },
		{ 319, 415, 58 },
		{ 319, 416, 58 },
		{ 319, 417, 58 },
		{ 319, 418, 58 },
		{ 319, 419, 58 },
		{ 319, 420, 58 },
		{ 319, 421, 58 },
		{ 319, 422, 58 },
		{ 320, 358, 58 },
		{ 320, 359, 58 },
		{ 320, 360, 58 },
		{ 320, 361, 58 },
		{ 320, 362, 58 },
		{ 320, 363, 58 },
		{ 320, 364, 58 },
		{ 320, 365, 58 },
		{ 320, 366, 58 },
		{ 320, 367, 58 },
		{ 320, 368, 58 },
		{ 320, 369, 58 },
		{ 320, 370, 58 },
		{ 320, 371, 58 },
		{ 320, 372, 58 },
		{ 320, 373, 58 },
		{ 320, 374, 58 },
		{ 320, 375, 58 },
		{ 320, 376, 58 },
		{ 320, 377, 58 },
		{ 320, 378, 58 },
		{ 320, 379, 58 },
		{ 320, 380, 58 },
		{ 320, 381, 58 },
		{ 320, 382, 58 },
		{ 320, 383, 58 },
		{ 320, 384, 58 },
		{ 320, 385, 58 },
		{ 320, 386, 58 },
		{ 320, 387, 58 },
		{ 320, 388, 58 },
		{ 320, 389, 58 },
		{ 320, 390, 58 },
		{ 320, 391, 58 },
		{ 320, 392, 58 },
		{ 320, 393, 58 },
		{ 320, 394, 58 },
		{ 320, 395, 58 },
		{ 320, 396, 58 },
		{ 320, 397, 58 },
		{ 320, 398, 58 },
		{ 320, 399, 58 },
		{ 320, 400, 58 },
		{ 320, 401, 58 },
		{ 320, 402, 58 },
		{ 320, 403, 58 },
		{ 320, 404, 58 },
		{ 320, 405, 58 },
		{ 320, 406, 58 },
		{ 320, 407, 58 },
		{ 320, 408, 58 },
		{ 320, 409, 58 },
		{ 320, 410, 58 },
		{ 320, 411, 58 },
		{ 320, 412, 58 },
		{ 320, 413, 58 },
		{ 320, 414, 58 },
		{ 320, 415, 58 },
		{ 320, 416, 58 },
		{ 320, 417, 58 },
		{ 320, 418, 58 },
		{ 320, 419, 58 },
		{ 320, 420, 58 },
		{ 320, 421, 58 },
		{ 320, 422, 58 },
		{ 320, 423, 58 },
		{ 320, 424, 58 },
		{ 321, 360, 58 },
		{ 321, 361, 58 },
		{ 321, 362, 58 },
		{ 321, 363, 58 },
		{ 321, 364, 58 },
		{ 321, 365, 58 },
		{ 321, 366, 58 },
		{ 321, 367, 58 },
		{ 321, 368, 58 },
		{ 321, 369, 58 },
		{ 321, 370, 58 },
		{ 321, 371, 58 },
		{ 321, 372, 58 },
		{ 321, 373, 58 },
		{ 321, 374, 58 },
		{ 321, 375, 58 },
		{ 321, 376, 58 },
		{ 321, 377, 58 },
		{ 321, 378, 58 },
		{ 321, 379, 58 },
		{ 321, 380, 58 },
		{ 321, 381, 58 },
		{ 321, 382, 58 },
		{ 321, 383, 58 },
		{ 321, 384, 58 },
		{ 321, 385, 58 },
		{ 321, 386, 58 },
		{ 321, 387, 58 },
		{ 321, 388, 58 },
		{ 321, 389, 58 },
		{ 321, 390, 58 },
		{ 321, 391, 58 },
		{ 321, 392, 58 },
		{ 321, 393, 58 },
		{ 321, 394, 58 },
		{ 321, 395, 58 },
		{ 321, 396, 58 },
		{ 321, 397, 58 },
		{ 321, 398, 58 },
		{ 321, 399, 58 },
		{ 321, 400, 58 },
		{ 321, 401, 58 },
		{ 321, 402, 58 },
		{ 321, 403, 58 },
		{ 321, 404, 58 },
		{ 321, 405, 58 },
		{ 321, 406, 58 },
		{ 321, 407, 58 },
		{ 321, 408, 58 },
		{ 321, 409, 58 },
		{ 321, 410, 58 },
		{ 321, 411, 58 },
		{ 321, 412, 58 },
		{ 321, 413, 58 },
		{ 321, 414, 58 },
		{ 321, 415, 58 },
		{ 321, 416, 58 },
		{ 321, 417, 58 },
		{ 321, 418, 58 },
		{ 321, 419, 58 },
		{ 321, 420, 58 },
		{ 321, 421, 58 },
		{ 321, 422, 58 },
		{ 321, 423, 58 },
		{ 321, 424, 58 },
		{ 321, 425, 58 },
		{ 321, 426, 58 },
		{ 321, 427, 58 },
		{ 322, 362, 58 },
		{ 322, 363, 58 },
		{ 322, 364, 58 },
		{ 322, 365, 58 },
		{ 322, 366, 58 },
		{ 322, 367, 58 },
		{ 322, 368, 58 },
		{ 322, 369, 58 },
		{ 322, 370, 58 },
		{ 322, 371, 58 },
		{ 322, 372, 58 },
		{ 322, 373, 58 },
		{ 322, 374, 58 },
		{ 322, 375, 58 },
		{ 322, 376, 58 },
		{ 322, 377, 58 },
		{ 322, 378, 58 },
		{ 322, 379, 58 },
		{ 322, 380, 58 },
		{ 322, 381, 58 },
		{ 322, 382, 58 },
		{ 322, 383, 58 },
		{ 322, 384, 58 },
		{ 322, 385, 58 },
		{ 322, 386, 58 },
		{ 322, 387, 58 },
		{ 322, 388, 58 },
		{ 322, 389, 58 },
		{ 322, 390, 58 },
		{ 322, 391, 58 },
		{ 322, 392, 58 },
		{ 322, 393, 58 },
		{ 322, 394, 58 },
		{ 322, 395, 58 },
		{ 322, 396, 58 },
		{ 322, 397, 58 },
		{ 322, 398, 58 },
		{ 322, 399, 58 },
		{ 322, 400, 58 },
		{ 322, 401, 58 },
		{ 322, 402, 58 },
		{ 322, 403, 58 },
		{ 322, 404, 58 },
		{ 322, 405, 58 },
		{ 322, 406, 58 },
		{ 322, 407, 58 },
		{ 322, 408, 58 },
		{ 322, 409, 58 },
		{ 322, 410, 58 },
		{ 322, 411, 58 },
		{ 322, 412, 58 },
		{ 322, 413, 58 },
		{ 322, 414, 58 },
		{ 322, 415, 58 },
		{ 322, 416, 58 },
		{ 322, 417, 58 },
		{ 322, 418, 58 },
		{ 322, 419, 58 },
		{ 322, 420, 58 },
		{ 322, 421, 58 },
		{ 322, 422, 58 },
		{ 322, 423, 58 },
		{ 322, 424, 58 },
		{ 322, 425, 58 },
		{ 322, 426, 58 },
		{ 322, 427, 58 },
		{ 322, 428, 58 },
		{ 322, 429, 58 },
		{ 322, 430, 58 },
		{ 323, 363, 58 },
		{ 323, 364, 58 },
		{ 323, 365, 58 },
		{ 323, 366, 58 },
		{ 323, 367, 58 },
		{ 323, 368, 58 },
		{ 323, 369, 58 },
		{ 323, 370, 58 },
		{ 323, 371, 58 },
		{ 323, 372, 58 },
		{ 323, 373, 58 },
		{ 323, 374, 58 },
		{ 323, 375, 58 },
		{ 323, 376, 58 },
		{ 323, 377, 58 },
		{ 323, 378, 58 },
		{ 323, 379, 58 },
		{ 323, 380, 58 },
		{ 323, 381, 58 },
		{ 323, 382, 58 },
		{ 323, 383, 58 },
		{ 323, 384, 58 },
		{ 323, 385, 58 },
		{ 323, 386, 58 },
		{ 323, 387, 58 },
		{ 323, 388, 58 },
		{ 323, 389, 58 },
		{ 323, 390, 58 },
		{ 323, 391, 58 },
		{ 323, 392, 58 },
		{ 323, 393, 58 },
		{ 323, 394, 58 },
		{ 323, 395, 58 },
		{ 323, 396, 58 },
		{ 323, 397, 58 },
		{ 323, 398, 58 },
		{ 323, 399, 58 },
		{ 323, 400, 58 },
		{ 323, 401, 58 },
		{ 323, 402, 58 },
		{ 323, 403, 58 },
		{ 323, 404, 58 },
		{ 323, 405, 58 },
		{ 323, 406, 58 },
		{ 323, 407, 58 },
		{ 323, 408, 58 },
		{ 323, 409, 58 },
		{ 323, 410, 58 },
		{ 323, 411, 58 },
		{ 323, 412, 58 },
		{ 323, 413, 58 },
		{ 323, 414, 58 },
		{ 323, 415, 58 },
		{ 323, 416, 58 },
		{ 323, 417, 58 },
		{ 323, 418, 58 },
		{ 323, 419, 58 },
		{ 323, 420, 58 },
		{ 323, 421, 58 },
		{ 323, 422, 58 },
		{ 323, 423, 58 },
		{ 323, 424, 58 },
		{ 323, 425, 58 },
		{ 323, 426, 58 },
		{ 323, 427, 58 },
		{ 323, 428, 58 },
		{ 323, 429, 58 },
		{ 323, 430, 58 },
		{ 323, 431, 58 },
		{ 323, 432, 58 },
		{ 324, 365, 58 },
		{ 324, 366, 58 },
		{ 324, 367, 58 },
		{ 324, 368, 58 },
		{ 324, 369, 58 },
		{ 324, 370, 58 },
		{ 324, 371, 58 },
		{ 324, 372, 58 },
		{ 324, 373, 58 },
		{ 324, 374, 58 },
		{ 324, 375, 58 },
		{ 324, 376, 58 },
		{ 324, 377, 58 },
		{ 324, 378, 58 },
		{ 324, 379, 58 },
		{ 324, 380, 58 },
		{ 324, 381, 58 },
		{ 324, 382, 58 },
		{ 324, 383, 58 },
		{ 324, 384, 58 },
		{ 324, 385, 58 },
		{ 324, 386, 58 },
		{ 324, 387, 58 },
		{ 324, 388, 58 },
		{ 324, 389, 58 },
		{ 324, 390, 58 },
		{ 324, 391, 58 },
		{ 324, 392, 58 },
		{ 324, 393, 58 },
		{ 324, 394, 58 },
		{ 324, 395, 58 },
		{ 324, 396, 58 },
		{ 324, 397, 58 },
		{ 324, 398, 58 },
		{ 324, 399, 58 },
		{ 324, 400, 58 },
		{ 324, 401, 58 },
		{ 324, 402, 58 },
		{ 324, 403, 58 },
		{ 324, 404, 58 },
		{ 324, 405, 58 },
		{ 324, 406, 58 },
		{ 324, 407, 58 },
		{ 324, 408, 58 },
		{ 324, 409, 58 },
		{ 324, 410, 58 },
		{ 324, 411, 58 },
		{ 324, 412, 58 },
		{ 324, 413, 58 },
		{ 324, 414, 58 },
		{ 324, 415, 58 },
		{ 324, 416, 58 },
		{ 324, 417, 58 },
		{ 324, 418, 58 },
		{ 324, 419, 58 },
		{ 324, 420, 58 },
		{ 324, 421, 58 },
		{ 324, 422, 58 },
		{ 324, 423, 58 },
		{ 324, 424, 58 },
		{ 324, 425, 58 },
		{ 324, 426, 58 },
		{ 324, 427, 58 },
		{ 324, 428, 58 },
		{ 324, 429, 58 },
		{ 324, 430, 58 },
		{ 324, 431, 58 },
		{ 324, 432, 58 },
		{ 324, 433, 58 },
		{ 324, 434, 58 },
		{ 324, 435, 58 },
		{ 325, 366, 58 },
		{ 325, 367, 58 },
		{ 325, 368, 58 },
		{ 325, 369, 58 },
		{ 325, 370, 58 },
		{ 325, 371, 58 },
		{ 325, 372, 58 },
		{ 325, 373, 58 },
		{ 325, 374, 58 },
		{ 325, 375, 58 },
		{ 325, 376, 58 },
		{ 325, 377, 58 },
		{ 325, 378, 58 },
		{ 325, 379, 58 },
		{ 325, 380, 58 },
		{ 325, 381, 58 },
		{ 325, 382, 58 },
		{ 325, 383, 58 },
		{ 325, 384, 58 },
		{ 325, 385, 58 },
		{ 325, 386, 58 },
		{ 325, 387, 58 },
		{ 325, 388, 58 },
		{ 325, 389, 58 },
		{ 325, 390, 58 },
		{ 325, 391, 58 },
		{ 325, 392, 58 },
		{ 325, 393, 58 },
		{ 325, 394, 58 },
		{ 325, 395, 58 },
		{ 325, 396, 58 },
		{ 325, 397, 58 },
		{ 325, 398, 58 },
		{ 325, 399, 58 },
		{ 325, 400, 58 },
		{ 325, 401, 58 },
		{ 325, 402, 58 },
		{ 325, 403, 58 },
		{ 325, 404, 58 },
		{ 325, 405, 58 },
		{ 325, 406, 58 },
		{ 325, 407, 58 },
		{ 325, 408, 58 },
		{ 325, 409, 58 },
		{ 325, 410, 58 },
		{ 325, 411, 58 },
		{ 325, 412, 58 },
		{ 325, 413, 58 },
		{ 325, 414, 58 },
		{ 325, 415, 58 },
		{ 325, 416, 58 },
		{ 325, 417, 58 },
		{ 325, 418, 58 },
		{ 325, 419, 58 },
		{ 325, 420, 58 },
		{ 325, 421, 58 },
		{ 325, 422, 58 },
		{ 325, 423, 58 },
		{ 325, 424, 58 },
		{ 325, 425, 58 },
		{ 325, 426, 58 },
		{ 325, 427, 58 },
		{ 325, 428, 58 },
		{ 325, 429, 58 },
		{ 325, 430, 58 },
		{ 325, 431, 58 },
		{ 325, 432, 58 },
		{ 325, 433, 58 },
		{ 325, 434, 58 },
		{ 325, 435, 58 },
		{ 325, 436, 58 },
		{ 325, 437, 58 },
		{ 325, 438, 58 },
		{ 326, 368, 58 },
		{ 326, 369, 58 },
		{ 326, 370, 58 },
		{ 326, 371, 58 },
		{ 326, 372, 58 },
		{ 326, 373, 58 },
		{ 326, 374, 58 },
		{ 326, 375, 58 },
		{ 326, 376, 58 },
		{ 326, 377, 58 },
		{ 326, 378, 58 },
		{ 326, 379, 58 },
		{ 326, 380, 58 },
		{ 326, 381, 58 },
		{ 326, 382, 58 },
		{ 326, 383, 58 },
		{ 326, 384, 58 },
		{ 326, 385, 58 },
		{ 326, 386, 58 },
		{ 326, 387, 58 },
		{ 326, 388, 58 },
		{ 326, 389, 58 },
		{ 326, 390, 58 },
		{ 326, 391, 58 },
		{ 326, 392, 58 },
		{ 326, 393, 58 },
		{ 326, 394, 58 },
		{ 326, 395, 58 },
		{ 326, 396, 58 },
		{ 326, 397, 58 },
		{ 326, 398, 58 },
		{ 326, 399, 58 },
		{ 326, 400, 58 },
		{ 326, 401, 58 },
		{ 326, 402, 58 },
		{ 326, 403, 58 },
		{ 326, 404, 58 },
		{ 326, 405, 58 },
		{ 326, 406, 58 },
		{ 326, 407, 58 },
		{ 326, 408, 58 },
		{ 326, 409, 58 },
		{ 326, 410, 58 },
		{ 326, 411, 58 },
		{ 326, 412, 58 },
		{ 326, 413, 58 },
		{ 326, 414, 58 },
		{ 326, 415, 58 },
		{ 326, 416, 58 },
		{ 326, 417, 58 },
		{ 326, 418, 58 },
		{ 326, 419, 58 },
		{ 326, 420, 58 },
		{ 326, 421, 58 },
		{ 326, 422, 58 },
		{ 326, 423, 58 },
		{ 326, 424, 58 },
		{ 326, 425, 58 },
		{ 326, 426, 58 },
		{ 326, 427, 58 },
		{ 326, 428, 58 },
		{ 326, 429, 58 },
		{ 326, 430, 58 },
		{ 326, 431, 58 },
		{ 326, 432, 58 },
		{ 326, 433, 58 },
		{ 326, 434, 58 },
		{ 326, 435, 58 },
		{ 326, 436, 58 },
		{ 326, 437, 58 },
		{ 326, 438, 58 },
		{ 326, 439, 58 },
		{ 326, 440, 58 },
		{ 327, 370, 58 },
		{ 327, 371, 58 },
		{ 327, 372, 58 },
		{ 327, 373, 58 },
		{ 327, 374, 58 },
		{ 327, 375, 58 },
		{ 327, 376, 58 },
		{ 327, 377, 58 },
		{ 327, 378, 58 },
		{ 327, 379, 58 },
		{ 327, 380, 58 },
		{ 327, 381, 58 },
		{ 327, 382, 58 },
		{ 327, 383, 58 },
		{ 327, 384, 58 },
		{ 327, 385, 58 },
		{ 327, 386, 58 },
		{ 327, 387, 58 },
		{ 327, 388, 58 },
		{ 327, 389, 58 },
		{ 327, 390, 58 },
		{ 327, 391, 58 },
		{ 327, 392, 58 },
		{ 327, 393, 58 },
		{ 327, 394, 58 },
		{ 327, 395, 58 },
		{ 327, 396, 58 },
		{ 327, 397, 58 },
		{ 327, 398, 58 },
		{ 327, 399, 58 },
		{ 327, 400, 58 },
		{ 327, 401, 58 },
		{ 327, 402, 58 },
		{ 327, 403, 58 },
		{ 327, 404, 58 },
		{ 327, 405, 58 },
		{ 327, 406, 58 },
		{ 327, 407, 58 },
		{ 327, 408, 58 },
		{ 327, 409, 58 },
		{ 327, 410, 58 },
		{ 327, 411, 58 },
		{ 327, 412, 58 },
		{ 327, 413, 58 },
		{ 327, 414, 58 },
		{ 327, 415, 58 },
		{ 327, 416, 58 },
		{ 327, 417, 58 },
		{ 327, 418, 58 },
		{ 327, 419, 58 },
		{ 327, 420, 58 },
		{ 327, 421, 58 },
		{ 327, 422, 58 },
		{ 327, 423, 58 },
		{ 327, 424, 58 },
		{ 327, 425, 58 },
		{ 327, 426, 58 },
		{ 327, 427, 58 },
		{ 327, 428, 58 },
		{ 327, 429, 58 },
		{ 327, 430, 58 },
		{ 327, 431, 58 },
		{ 327, 432, 58 },
		{ 327, 433, 58 },
		{ 327, 434, 58 },
		{ 327, 435, 58 },
		{ 327, 436, 58 },
		{ 327, 437, 58 },
		{ 327, 438, 58 },
		{ 327, 439, 58 },
		{ 327, 440, 58 },
		{ 327, 441, 58 },
		{ 327, 442, 58 },
		{ 327, 443, 58 },
		{ 328, 371, 58 },
		{ 328, 372, 58 },
		{ 328, 373, 58 },
		{ 328, 374, 58 },
		{ 328, 375, 58 },
		{ 328, 376, 58 },
		{ 328, 377, 58 },
		{ 328, 378, 58 },
		{ 328, 379, 58 },
		{ 328, 380, 58 },
		{ 328, 381, 58 },
		{ 328, 382, 58 },
		{ 328, 383, 58 },
		{ 328, 384, 58 },
		{ 328, 385, 58 },
		{ 328, 386, 58 },
		{ 328, 387, 58 },
		{ 328, 388, 58 },
		{ 328, 389, 58 },
		{ 328, 390, 58 },
		{ 328, 391, 58 },
		{ 328, 392, 58 },
		{ 328, 393, 58 },
		{ 328, 394, 58 },
		{ 328, 395, 58 },
		{ 328, 396, 58 },
		{ 328, 397, 58 },
		{ 328, 398, 58 },
		{ 328, 399, 58 },
		{ 328, 400, 58 },
		{ 328, 401, 58 },
		{ 328, 402, 58 },
		{ 328, 403, 58 },
		{ 328, 404, 58 },
		{ 328, 405, 58 },
		{ 328, 406, 58 },
		{ 328, 407, 58 },
		{ 328, 408, 58 },
		{ 328, 409, 58 },
		{ 328, 410, 58 },
		{ 328, 411, 58 },
		{ 328, 412, 58 },
		{ 328, 413, 58 },
		{ 328, 414, 58 },
		{ 328, 415, 58 },
		{ 328, 416, 58 },
		{ 328, 417, 58 },
		{ 328, 418, 58 },
		{ 328, 419, 58 },
		{ 328, 420, 58 },
		{ 328, 421, 58 },
		{ 328, 422, 58 },
		{ 328, 423, 58 },
		{ 328, 424, 58 },
		{ 328, 425, 58 },
		{ 328, 426, 58 },
		{ 328, 427, 58 },
		{ 328, 428, 58 },
		{ 328, 429, 58 },
		{ 328, 430, 58 },
		{ 328, 431, 58 },
		{ 328, 432, 58 },
		{ 328, 433, 58 },
		{ 328, 434, 58 },
		{ 328, 435, 58 },
		{ 328, 436, 58 },
		{ 328, 437, 58 },
		{ 328, 438, 58 },
		{ 328, 439, 58 },
		{ 328, 440, 58 },
		{ 328, 441, 58 },
		{ 328, 442, 58 },
		{ 328, 443, 58 },
		{ 328, 444, 58 },
		{ 328, 445, 58 },
		{ 328, 446, 58 },
		{ 329, 373, 58 },
		{ 329, 374, 58 },
		{ 329, 375, 58 },
		{ 329, 376, 58 },
		{ 329, 377, 58 },
		{ 329, 378, 58 },
		{ 329, 379, 58 },
		{ 329, 380, 58 },
		{ 329, 381, 58 },
		{ 329, 382, 58 },
		{ 329, 383, 58 },
		{ 329, 384, 58 },
		{ 329, 385, 58 },
		{ 329, 386, 58 },
		{ 329, 387, 58 },
		{ 329, 388, 58 },
		{ 329, 389, 58 },
		{ 329, 390, 58 },
		{ 329, 391, 58 },
		{ 329, 392, 58 },
		{ 329, 393, 58 },
		{ 329, 394, 58 },
		{ 329, 395, 58 },
		{ 329, 396, 58 },
		{ 329, 397, 58 },
		{ 329, 398, 58 },
		{ 329, 399, 58 },
		{ 329, 400, 58 },
		{ 329, 401, 58 },
		{ 329, 402, 58 },
		{ 329, 403, 58 },
		{ 329, 404, 58 },
		{ 329, 405, 58 },
		{ 329, 406, 58 },
		{ 329, 407, 58 },
		{ 329, 408, 58 },
		{ 329, 409, 58 },
		{ 329, 410, 58 },
		{ 329, 411, 58 },
		{ 329, 412, 58 },
		{ 329, 413, 58 },
		{ 329, 414, 58 },
		{ 329, 415, 58 },
		{ 329, 416, 58 },
		{ 329, 417, 58 },
		{ 329, 418, 58 },
		{ 329, 419, 58 },
		{ 329, 420, 58 },
		{ 329, 421, 58 },
		{ 329, 422, 58 },
		{ 329, 423, 58 },
		{ 329, 424, 58 },
		{ 329, 425, 58 },
		{ 329, 426, 58 },
		{ 329, 427, 58 },
		{ 329, 428, 58 },
		{ 329, 429, 58 },
		{ 329, 430, 58 },
		{ 329, 431, 58 },
		{ 329, 432, 58 },
		{ 329, 433, 58 },
		{ 329, 434, 58 },
		{ 329, 435, 58 },
		{ 329, 436, 58 },
		{ 329, 437, 58 },
		{ 329, 438, 58 },
		{ 329, 439, 58 },
		{ 329, 440, 58 },
		{ 329, 441, 58 },
		{ 329, 442, 58 },
		{ 329, 443, 58 },
		{ 329, 444, 58 },
		{ 329, 445, 58 },
		{ 329, 446, 58 },
		{ 329, 447, 58 },
		{ 329, 448, 58 },
		{ 330, 374, 58 },
		{ 330, 375, 58 },
		{ 330, 376, 58 },
		{ 330, 377, 58 },
		{ 330, 378, 58 },
		{ 330, 379, 58 },
		{ 330, 380, 58 },
		{ 330, 381, 58 },
		{ 330, 382, 58 },
		{ 330, 383, 58 },
		{ 330, 384, 58 },
		{ 330, 385, 58 },
		{ 330, 386, 58 },
		{ 330, 387, 58 },
		{ 330, 388, 58 },
		{ 330, 389, 58 },
		{ 330, 390, 58 },
		{ 330, 391, 58 },
		{ 330, 392, 58 },
		{ 330, 393, 58 },
		{ 330, 394, 58 },
		{ 330, 395, 58 },
		{ 330, 396, 58 },
		{ 330, 397, 58 },
		{ 330, 398, 58 },
		{ 330, 399, 58 },
		{ 330, 400, 58 },
		{ 330, 401, 58 },
		{ 330, 402, 58 },
		{ 330, 403, 58 },
		{ 330, 404, 58 },
		{ 330, 405, 58 },
		{ 330, 406, 58 },
		{ 330, 407, 58 },
		{ 330, 408, 58 },
		{ 330, 409, 58 },
		{ 330, 410, 58 },
		{ 330, 411, 58 },
		{ 330, 412, 58 },
		{ 330, 413, 58 },
		{ 330, 414, 58 },
		{ 330, 415, 58 },
		{ 330, 416, 58 },
		{ 330, 417, 58 },
		{ 330, 418, 58 },
		{ 330, 419, 58 },
		{ 330, 420, 58 },
		{ 330, 421, 58 },
		{ 330, 422, 58 },
		{ 330, 423, 58 },
		{ 330, 424, 58 },
		{ 330, 425, 58 },
		{ 330, 426, 58 },
		{ 330, 427, 58 },
		{ 330, 428, 58 },
		{ 330, 429, 58 },
		{ 330, 430, 58 },
		{ 330, 431, 58 },
		{ 330, 432, 58 },
		{ 330, 433, 58 },
		{ 330, 434, 58 },
		{ 330, 435, 58 },
		{ 330, 436, 58 },
		{ 330, 437, 58 },
		{ 330, 438, 58 },
		{ 330, 439, 58 },
		{ 330, 440, 58 },
		{ 330, 441, 58 },
		{ 330, 442, 58 },
		{ 330, 443, 58 },
		{ 330, 444, 58 },
		{ 330, 445, 58 },
		{ 330, 446, 58 },
		{ 330, 447, 58 },
		{ 330, 448, 58 },
		{ 330, 449, 58 },
		{ 330, 450, 58 },
		{ 330, 451, 58 },
		{ 331, 376, 58 },
		{ 331, 377, 58 },
		{ 331, 378, 58 },
		{ 331, 379, 58 },
		{ 331, 380, 58 },
		{ 331, 381, 58 },
		{ 331, 382, 58 },
		{ 331, 383, 58 },
		{ 331, 384, 58 },
		{ 331, 385, 58 },
		{ 331, 386, 58 },
		{ 331, 387, 58 },
		{ 331, 388, 58 },
		{ 331, 389, 58 },
		{ 331, 390, 58 },
		{ 331, 391, 58 },
		{ 331, 392, 58 },
		{ 331, 393, 58 },
		{ 331, 394, 58 },
		{ 331, 395, 58 },
		{ 331, 396, 58 },
		{ 331, 397, 58 },
		{ 331, 398, 58 },
		{ 331, 399, 58 },
		{ 331, 400, 58 },
		{ 331, 401, 58 },
		{ 331, 402, 58 },
		{ 331, 403, 58 },
		{ 331, 404, 58 },
		{ 331, 405, 58 },
		{ 331, 406, 58 },
		{ 331, 407, 58 },
		{ 331, 408, 58 },
		{ 331, 409, 58 },
		{ 331, 410, 58 },
		{ 331, 411, 58 },
		{ 331, 412, 58 },
		{ 331, 413, 58 },
		{ 331, 414, 58 },
		{ 331, 415, 58 },
		{ 331, 416, 58 },
		{ 331, 417, 58 },
		{ 331, 418, 58 },
		{ 331, 419, 58 },
		{ 331, 420, 58 },
		{ 331, 421, 58 },
		{ 331, 422, 58 },
		{ 331, 423, 58 },
		{ 331, 424, 58 },
		{ 331, 425, 58 },
		{ 331, 426, 58 },
		{ 331, 427, 58 },
		{ 331, 428, 58 },
		{ 331, 429, 58 },
		{ 331, 430, 58 },
		{ 331, 431, 58 },
		{ 331, 432, 58 },
		{ 331, 433, 58 },
		{ 331, 434, 58 },
		{ 331, 435, 58 },
		{ 331, 436, 58 },
		{ 331, 437, 58 },
		{ 331, 438, 58 },
		{ 331, 439, 58 },
		{ 331, 440, 58 },
		{ 331, 441, 58 },
		{ 331, 442, 58 },
		{ 331, 443, 58 },
		{ 331, 444, 58 },
		{ 331, 445, 58 },
		{ 331, 446, 58 },
		{ 331, 447, 58 },
		{ 331, 448, 58 },
		{ 331, 449, 58 },
		{ 331, 450, 58 },
		{ 331, 451, 58 },
		{ 331, 452, 58 },
		{ 331, 453, 58 },
		{ 331, 454, 58 },
		{ 332, 378, 58 },
		{ 332, 379, 58 },
		{ 332, 380, 58 },
		{ 332, 381, 58 },
		{ 332, 382, 58 },
		{ 332, 383, 58 },
		{ 332, 384, 58 },
		{ 332, 385, 58 },
		{ 332, 386, 58 },
		{ 332, 387, 58 },
		{ 332, 388, 58 },
		{ 332, 389, 58 },
		{ 332, 390, 58 },
		{ 332, 391, 58 },
		{ 332, 392, 58 },
		{ 332, 393, 58 },
		{ 332, 394, 58 },
		{ 332, 395, 58 },
		{ 332, 396, 58 },
		{ 332, 397, 58 },
		{ 332, 398, 58 },
		{ 332, 399, 58 },
		{ 332, 400, 58 },
		{ 332, 401, 58 },
		{ 332, 402, 58 },
		{ 332, 403, 58 },
		{ 332, 404, 58 },
		{ 332, 405, 58 },
		{ 332, 406, 58 },
		{ 332, 407, 58 },
		{ 332, 408, 58 },
		{ 332, 409, 58 },
		{ 332, 410, 58 },
		{ 332, 411, 58 },
		{ 332, 412, 58 },
		{ 332, 413, 58 },
		{ 332, 414, 58 },
		{ 332, 415, 58 },
		{ 332, 416, 58 },
		{ 332, 417, 58 },
		{ 332, 418, 58 },
		{ 332, 419, 58 },
		{ 332, 420, 58 },
		{ 332, 421, 58 },
		{ 332, 422, 58 },
		{ 332, 423, 58 },
		{ 332, 424, 58 },
		{ 332, 425, 58 },
		{ 332, 426, 58 },
		{ 332, 427, 58 },
		{ 332, 428, 58 },
		{ 332, 429, 58 },
		{ 332, 430, 58 },
		{ 332, 431, 58 },
		{ 332, 432, 58 },
		{ 332, 433, 58 },
		{ 332, 434, 58 },
		{ 332, 435, 58 },
		{ 332, 436, 58 },
		{ 332, 437, 58 },
		{ 332, 438, 58 },
		{ 332, 439, 58 },
		{ 332, 440, 58 },
		{ 332, 441, 58 },
		{ 332, 442, 58 },
		{ 332, 443, 58 },
		{ 332, 444, 58 },
		{ 332, 445, 58 },
		{ 332, 446, 58 },
		{ 332, 447, 58 },
		{ 332, 448, 58 },
		{ 332, 449, 58 },
		{ 332, 450, 58 },
		{ 332, 451, 58 },
		{ 332, 452, 58 },
		{ 332, 453, 58 },
		{ 332, 454, 58 },
		{ 332, 455, 58 },
		{ 332, 456, 58 },
		{ 333, 379, 58 },
		{ 333, 380, 58 },
		{ 333, 381, 58 },
		{ 333, 382, 58 },
		{ 333, 383, 58 },
		{ 333, 384, 58 },
		{ 333, 385, 58 },
		{ 333, 386, 58 },
		{ 333, 387, 58 },
		{ 333, 388, 58 },
		{ 333, 389, 58 },
		{ 333, 390, 58 },
		{ 333, 391, 58 },
		{ 333, 392, 58 },
		{ 333, 393, 58 },
		{ 333, 394, 58 },
		{ 333, 395, 58 },
		{ 333, 396, 58 },
		{ 333, 397, 58 },
		{ 333, 398, 58 },
		{ 333, 399, 58 },
		{ 333, 400, 58 },
		{ 333, 401, 58 },
		{ 333, 402, 58 },
		{ 333, 403, 58 },
		{ 333, 404, 58 },
		{ 333, 405, 58 },
		{ 333, 406, 58 },
		{ 333, 407, 58 },
		{ 333, 408, 58 },
		{ 333, 409, 58 },
		{ 333, 410, 58 },
		{ 333, 411, 58 },
		{ 333, 412, 58 },
		{ 333, 413, 58 },
		{ 333, 414, 58 },
		{ 333, 415, 58 },
		{ 333, 416, 58 },
		{ 333, 417, 58 },
		{ 333, 418, 58 },
		{ 333, 419, 58 },
		{ 333, 420, 58 },
		{ 333, 421, 58 },
		{ 333, 422, 58 },
		{ 333, 423, 58 },
		{ 333, 424, 58 },
		{ 333, 425, 58 },
		{ 333, 426, 58 },
		{ 333, 427, 58 },
		{ 333, 428, 58 },
		{ 333, 429, 58 },
		{ 333, 430, 58 },
		{ 333, 431, 58 },
		{ 333, 432, 58 },
		{ 333, 433, 58 },
		{ 333, 434, 58 },
		{ 333, 435, 58 },
		{ 333, 436, 58 },
		{ 333, 437, 58 },
		{ 333, 438, 58 },
		{ 333, 439, 58 },
		{ 333, 440, 58 },
		{ 333, 441, 58 },
		{ 333, 442, 58 },
		{ 333, 443, 58 },
		{ 333, 444, 58 },
		{ 333, 445, 58 },
		{ 333, 446, 58 },
		{ 333, 447, 58 },
		{ 333, 448, 58 },
		{ 333, 449, 58 },
		{ 333, 450, 58 },
		{ 333, 451, 58 },
		{ 333, 452, 58 },
		{ 333, 453, 58 },
		{ 333, 454, 58 },
		{ 333, 455, 58 },
		{ 333, 456, 58 },
		{ 333, 457, 58 },
		{ 333, 458, 58 },
		{ 333, 459, 58 },
		{ 334, 381, 58 },
		{ 334, 382, 58 },
		{ 334, 383, 58 },
		{ 334, 384, 58 },
		{ 334, 385, 58 },
		{ 334, 386, 58 },
		{ 334, 387, 58 },
		{ 334, 388, 58 },
		{ 334, 389, 58 },
		{ 334, 390, 58 },
		{ 334, 391, 58 },
		{ 334, 392, 58 },
		{ 334, 393, 58 },
		{ 334, 394, 58 },
		{ 334, 395, 58 },
		{ 334, 396, 58 },
		{ 334, 397, 58 },
		{ 334, 398, 58 },
		{ 334, 399, 58 },
		{ 334, 400, 58 },
		{ 334, 401, 58 },
		{ 334, 402, 58 },
		{ 334, 403, 58 },
		{ 334, 404, 58 },
		{ 334, 405, 58 },
		{ 334, 406, 58 },
		{ 334, 407, 58 },
		{ 334, 408, 58 },
		{ 334, 409, 58 },
		{ 334, 410, 58 },
		{ 334, 411, 58 },
		{ 334, 412, 58 },
		{ 334, 413, 58 },
		{ 334, 414, 58 },
		{ 334, 415, 58 },
		{ 334, 416, 58 },
		{ 334, 417, 58 },
		{ 334, 418, 58 },
		{ 334, 419, 58 },
		{ 334, 420, 58 },
		{ 334, 421, 58 },
		{ 334, 422, 58 },
		{ 334, 423, 58 },
		{ 334, 424, 58 },
		{ 334, 425, 58 },
		{ 334, 426, 58 },
		{ 334, 427, 58 },
		{ 334, 428, 58 },
		{ 334, 429, 58 },
		{ 334, 430, 58 },
		{ 334, 431, 58 },
		{ 334, 432, 58 },
		{ 334, 433, 58 },
		{ 334, 434, 58 },
		{ 334, 435, 58 },
		{ 334, 436, 58 },
		{ 334, 437, 58 },
		{ 334, 438, 58 },
		{ 334, 439, 58 },
		{ 334, 440, 58 },
		{ 334, 441, 58 },
		{ 334, 442, 58 },
		{ 334, 443, 58 },
		{ 334, 444, 58 },
		{ 334, 445, 58 },
		{ 334, 446, 58 },
		{ 334, 447, 58 },
		{ 334, 448, 58 },
		{ 334, 449, 58 },
		{ 334, 450, 58 },
		{ 334, 451, 58 },
		{ 334, 452, 58 },
		{ 334, 453, 58 },
		{ 334, 454, 58 },
		{ 334, 455, 58 },
		{ 334, 456, 58 },
		{ 334, 457, 58 },
		{ 334, 458, 58 },
		{ 334, 459, 58 },
		{ 334, 460, 58 },
		{ 334, 461, 58 },
		{ 334, 462, 58 },
		{ 335, 382, 58 },
		{ 335, 383, 58 },
		{ 335, 384, 58 },
		{ 335, 385, 58 },
		{ 335, 386, 58 },
		{ 335, 387, 58 },
		{ 335, 388, 58 },
		{ 335, 389, 58 },
		{ 335, 390, 58 },
		{ 335, 391, 58 },
		{ 335, 392, 58 },
		{ 335, 393, 58 },
		{ 335, 394, 58 },
		{ 335, 395, 58 },
		{ 335, 396, 58 },
		{ 335, 397, 58 },
		{ 335, 398, 58 },
		{ 335, 399, 58 },
		{ 335, 400, 58 },
		{ 335, 401, 58 },
		{ 335, 402, 58 },
		{ 335, 403, 58 },
		{ 335, 404, 58 },
		{ 335, 405, 58 },
		{ 335, 406, 58 },
		{ 335, 407, 58 },
		{ 335, 408, 58 },
		{ 335, 409, 58 },
		{ 335, 410, 58 },
		{ 335, 411, 58 },
		{ 335, 412, 58 },
		{ 335, 413, 58 },
		{ 335, 414, 58 },
		{ 335, 415, 58 },
		{ 335, 416, 58 },
		{ 335, 417, 58 },
		{ 335, 418, 58 },
		{ 335, 419, 58 },
		{ 335, 420, 58 },
		{ 335, 421, 58 },
		{ 335, 422, 58 },
		{ 335, 423, 58 },
		{ 335, 424, 58 },
		{ 335, 425, 58 },
		{ 335, 426, 58 },
		{ 335, 427, 58 },
		{ 335, 428, 58 },
		{ 335, 429, 58 },
		{ 335, 430, 58 },
		{ 335, 431, 58 },
		{ 335, 432, 58 },
		{ 335, 433, 58 },
		{ 335, 434, 58 },
		{ 335, 435, 58 },
		{ 335, 436, 58 },
		{ 335, 437, 58 },
		{ 335, 438, 58 },
		{ 335, 439, 58 },
		{ 335, 440, 58 },
		{ 335, 441, 58 },
		{ 335, 442, 58 },
		{ 335, 443, 58 },
		{ 335, 444, 58 },
		{ 335, 445, 58 },
		{ 335, 446, 58 },
		{ 335, 447, 58 },
		{ 335, 448, 58 },
		{ 335, 449, 58 },
		{ 335, 450, 58 },
		{ 335, 451, 58 },
		{ 335, 452, 58 },
		{ 335, 453, 58 },
		{ 335, 454, 58 },
		{ 335, 455, 58 },
		{ 335, 456, 58 },
		{ 335, 457, 58 },
		{ 335, 458, 58 },
		{ 335, 459, 58 },
		{ 335, 460, 58 },
		{ 335, 461, 58 },
		{ 335, 462, 58 },
		{ 335, 463, 58 },
		{ 335, 464, 58 },
		{ 336, 384, 58 },
		{ 336, 385, 58 },
		{ 336, 386, 58 },
		{ 336, 387, 58 },
		{ 336, 388, 58 },
		{ 336, 389, 58 },
		{ 336, 390, 58 },
		{ 336, 391, 58 },
		{ 336, 392, 58 },
		{ 336, 393, 58 },
		{ 336, 394, 58 },
		{ 336, 395, 58 },
		{ 336, 396, 58 },
		{ 336, 397, 58 },
		{ 336, 398, 58 },
		{ 336, 399, 58 },
		{ 336, 400, 58 },
		{ 336, 401, 58 },
		{ 336, 402, 58 },
		{ 336, 403, 58 },
		{ 336, 404, 58 },
		{ 336, 405, 58 },
		{ 336, 406, 58 },
		{ 336, 407, 58 },
		{ 336, 408, 58 },
		{ 336, 409, 58 },
		{ 336, 410, 58 },
		{ 336, 411, 58 },
		{ 336, 412, 58 },
		{ 336, 413, 58 },
		{ 336, 414, 58 },
		{ 336, 415, 58 },
		{ 336, 416, 58 },
		{ 336, 417, 58 },
		{ 336, 418, 58 },
		{ 336, 419, 58 },
		{ 336, 420, 58 },
		{ 336, 421, 58 },
		{ 336, 422, 58 },
		{ 336, 423, 58 },
		{ 336, 424, 58 },
		{ 336, 425, 58 },
		{ 336, 426, 58 },
		{ 336, 427, 58 },
		{ 336, 428, 58 },
		{ 336, 429, 58 },
		{ 336, 430, 58 },
		{ 336, 431, 58 },
		{ 336, 432, 58 },
		{ 336, 433, 58 },
		{ 336, 434, 58 },
		{ 336, 435, 58 },
		{ 336, 436, 58 },
		{ 336, 437, 58 },
		{ 336, 438, 58 },
		{ 336, 439, 58 },
		{ 336, 440, 58 },
		{ 336, 441, 58 },
		{ 336, 442, 58 },
		{ 336, 443, 58 },
		{ 336, 444, 58 },
		{ 336, 445, 58 },
		{ 336, 446, 58 },
		{ 336, 447, 58 },
		{ 336, 448, 58 },
		{ 336, 449, 58 },
		{ 336, 450, 58 },
		{ 336, 451, 58 },
		{ 336, 452, 58 },
		{ 336, 453, 58 },
		{ 336, 454, 58 },
		{ 336, 455, 58 },
		{ 336, 456, 58 },
		{ 336, 457, 58 },
		{ 336, 458, 58 },
		{ 336, 459, 58 },
		{ 336, 460, 58 },
		{ 336, 461, 58 },
		{ 336, 462, 58 },
		{ 336, 463, 58 },
		{ 336, 464, 58 },
		{ 336, 465, 58 },
		{ 336, 466, 58 },
		{ 336, 467, 58 },
		{ 337, 386, 58 },
		{ 337, 387, 58 },
		{ 337, 388, 58 },
		{ 337, 389, 58 },
		{ 337, 390, 58 },
		{ 337, 391, 58 },
		{ 337, 392, 58 },
		{ 337, 393, 58 },
		{ 337, 394, 58 },
		{ 337, 395, 58 },
		{ 337, 396, 58 },
		{ 337, 397, 58 },
		{ 337, 398, 58 },
		{ 337, 399, 58 },
		{ 337, 400, 58 },
		{ 337, 401, 58 },
		{ 337, 402, 58 },
		{ 337, 403, 58 },
		{ 337, 404, 58 },
		{ 337, 405, 58 },
		{ 337, 406, 58 },
		{ 337, 407, 58 },
		{ 337, 408, 58 },
		{ 337, 409, 58 },
		{ 337, 410, 58 },
		{ 337, 411, 58 },
		{ 337, 412, 58 },
		{ 337, 413, 58 },
		{ 337, 414, 58 },
		{ 337, 415, 58 },
		{ 337, 416, 58 },
		{ 337, 417, 58 },
		{ 337, 418, 58 },
		{ 337, 419, 58 },
		{ 337, 420, 58 },
		{ 337, 421, 58 },
		{ 337, 422, 58 },
		{ 337, 423, 58 },
		{ 337, 424, 58 },
		{ 337, 425, 58 },
		{ 337, 426, 58 },
		{ 337, 427, 58 },
		{ 337, 428, 58 },
		{ 337, 429, 58 },
		{ 337, 430, 58 },
		{ 337, 431, 58 },
		{ 337, 432, 58 },
		{ 337, 433, 58 },
		{ 337, 434, 58 },
		{ 337, 435, 58 },
		{ 337, 436, 58 },
		{ 337, 437, 58 },
		{ 337, 438, 58 },
		{ 337, 439, 58 },
		{ 337, 440, 58 },
		{ 337, 441, 58 },
		{ 337, 442, 58 },
		{ 337, 443, 58 },
		{ 337, 444, 58 },
		{ 337, 445, 58 },
		{ 337, 446, 58 },
		{ 337, 447, 58 },
		{ 337, 448, 58 },
		{ 337, 449, 58 },
		{ 337, 450, 58 },
		{ 337, 451, 58 },
		{ 337, 452, 58 },
		{ 337, 453, 58 },
		{ 337, 454, 58 },
		{ 337, 455, 58 },
		{ 337, 456, 58 },
		{ 337, 457, 58 },
		{ 337, 458, 58 },
		{ 337, 459, 58 },
		{ 337, 460, 58 },
		{ 337, 461, 58 },
		{ 337, 462, 58 },
		{ 337, 463, 58 },
		{ 337, 464, 58 },
		{ 337, 465, 58 },
		{ 337, 466, 58 },
		{ 337, 467, 58 },
		{ 337, 468, 58 },
		{ 337, 469, 58 },
		{ 337, 470, 58 },
		{ 338, 387, 58 },
		{ 338, 388, 58 },
		{ 338, 389, 58 },
		{ 338, 390, 58 },
		{ 338, 391, 58 },
		{ 338, 392, 58 },
		{ 338, 393, 58 },
		{ 338, 394, 58 },
		{ 338, 395, 58 },
		{ 338, 396, 58 },
		{ 338, 397, 58 },
		{ 338, 398, 58 },
		{ 338, 399, 58 },
		{ 338, 400, 58 },
		{ 338, 401, 58 },
		{ 338, 402, 58 },
		{ 338, 403, 58 },
		{ 338, 404, 58 },
		{ 338, 405, 58 },
		{ 338, 406, 58 },
		{ 338, 407, 58 },
		{ 338, 408, 58 },
		{ 338, 409, 58 },
		{ 338, 410, 58 },
		{ 338, 411, 58 },
		{ 338, 412, 58 },
		{ 338, 413, 58 },
		{ 338, 414, 58 },
		{ 338, 415, 58 },
		{ 338, 416, 58 },
		{ 338, 417, 58 },
		{ 338, 418, 58 },
		{ 338, 419, 58 },
		{ 338, 420, 58 },
		{ 338, 421, 58 },
		{ 338, 422, 58 },
		{ 338, 423, 58 },
		{ 338, 424, 58 },
		{ 338, 425, 58 },
		{ 338, 426, 58 },
		{ 338, 427, 58 },
		{ 338, 428, 58 },
		{ 338, 429, 58 },
		{ 338, 430, 58 },
		{ 338, 431, 58 },
		{ 338, 432, 58 },
		{ 338, 433, 58 },
		{ 338, 434, 58 },
		{ 338, 435, 58 },
		{ 338, 436, 58 },
		{ 338, 437, 58 },
		{ 338, 438, 58 },
		{ 338, 439, 58 },
		{ 338, 440, 58 },
		{ 338, 441, 58 },
		{ 338, 442, 58 },
		{ 338, 443, 58 },
		{ 338, 444, 58 },
		{ 338, 445, 58 },
		{ 338, 446, 58 },
		{ 338, 447, 58 },
		{ 338, 448, 58 },
		{ 338, 449, 58 },
		{ 338, 450, 58 },
		{ 338, 451, 58 },
		{ 338, 452, 58 },
		{ 338, 453, 58 },
		{ 338, 454, 58 },
		{ 338, 455, 58 },
		{ 338, 456, 58 },
		{ 338, 457, 58 },
		{ 338, 458, 58 },
		{ 338, 459, 58 },
		{ 338, 460, 58 },
		{ 338, 461, 58 },
		{ 338, 462, 58 },
		{ 338, 463, 58 },
		{ 338, 464, 58 },
		{ 338, 465, 58 },
		{ 338, 466, 58 },
		{ 338, 467, 58 },
		{ 338, 468, 58 },
		{ 338, 469, 58 },
		{ 338, 470, 58 },
		{ 338, 471, 58 },
		{ 338, 472, 58 },
		{ 339, 389, 58 },
		{ 339, 390, 58 },
		{ 339, 391, 58 },
		{ 339, 392, 58 },
		{ 339, 393, 58 },
		{ 339, 394, 58 },
		{ 339, 395, 58 },
		{ 339, 396, 58 },
		{ 339, 397, 58 },
		{ 339, 398, 58 },
		{ 339, 399, 58 },
		{ 339, 400, 58 },
		{ 339, 401, 58 },
		{ 339, 402, 58 },
		{ 339, 403, 58 },
		{ 339, 404, 58 },
		{ 339, 405, 58 },
		{ 339, 406, 58 },
		{ 339, 407, 58 },
		{ 339, 408, 58 },
		{ 339, 409, 58 },
		{ 339, 410, 58 },
		{ 339, 411, 58 },
		{ 339, 412, 58 },
		{ 339, 413, 58 },
		{ 339, 414, 58 },
		{ 339, 415, 58 },
		{ 339, 416, 58 },
		{ 339, 417, 58 },
		{ 339, 418, 58 },
		{ 339, 419, 58 },
		{ 339, 420, 58 },
		{ 339, 421, 58 },
		{ 339, 422, 58 },
		{ 339, 423, 58 },
		{ 339, 424, 58 },
		{ 339, 425, 58 },
		{ 339, 426, 58 },
		{ 339, 427, 58 },
		{ 339, 428, 58 },
		{ 339, 429, 58 },
		{ 339, 430, 58 },
		{ 339, 431, 58 },
		{ 339, 432, 58 },
		{ 339, 433, 58 },
		{ 339, 434, 58 },
		{ 339, 435, 58 },
		{ 339, 436, 58 },
		{ 339, 437, 58 },
		{ 339, 438, 58 },
		{ 339, 439, 58 },
		{ 339, 440, 58 },
		{ 339, 441, 58 },
		{ 339, 442, 58 },
		{ 339, 443, 58 },
		{ 339, 444, 58 },
		{ 339, 445, 58 },
		{ 339, 446, 58 },
		{ 339, 447, 58 },
		{ 339, 448, 58 },
		{ 339, 449, 58 },
		{ 339, 450, 58 },
		{ 339, 451, 58 },
		{ 339, 452, 58 },
		{ 339, 453, 58 },
		{ 339, 454, 58 },
		{ 339, 455, 58 },
		{ 339, 456, 58 },
		{ 339, 457, 58 },
		{ 339, 458, 58 },
		{ 339, 459, 58 },
		{ 339, 460, 58 },
		{ 339, 461, 58 },
		{ 339, 462, 58 },
		{ 339, 463, 58 },
		{ 339, 464, 58 },
		{ 339, 465, 58 },
		{ 339, 466, 58 },
		{ 339, 467, 58 },
		{ 339, 468, 58 },
		{ 339, 469, 58 },
		{ 339, 470, 58 },
		{ 339, 471, 58 },
		{ 339, 472, 58 },
		{ 339, 473, 58 },
		{ 339, 474, 58 },
		{ 339, 475, 58 },
		{ 340, 390, 58 },
		{ 340, 391, 58 },
		{ 340, 392, 58 },
		{ 340, 393, 58 },
		{ 340, 394, 58 },
		{ 340, 395, 58 },
		{ 340, 396, 58 },
		{ 340, 397, 58 },
		{ 340, 398, 58 },
		{ 340, 399, 58 },
		{ 340, 400, 58 },
		{ 340, 401, 58 },
		{ 340, 402, 58 },
		{ 340, 403, 58 },
		{ 340, 404, 58 },
		{ 340, 405, 58 },
		{ 340, 406, 58 },
		{ 340, 407, 58 },
		{ 340, 408, 58 },
		{ 340, 409, 58 },
		{ 340, 410, 58 },
		{ 340, 411, 58 },
		{ 340, 412, 58 },
		{ 340, 413, 58 },
		{ 340, 414, 58 },
		{ 340, 415, 58 },
		{ 340, 416, 58 },
		{ 340, 417, 58 },
		{ 340, 418, 58 },
		{ 340, 419, 58 },
		{ 340, 420, 58 },
		{ 340, 421, 58 },
		{ 340, 422, 58 },
		{ 340, 423, 58 },
		{ 340, 424, 58 },
		{ 340, 425, 58 },
		{ 340, 426, 58 },
		{ 340, 427, 58 },
		{ 340, 428, 58 },
		{ 340, 429, 58 },
		{ 340, 430, 58 },
		{ 340, 431, 58 },
		{ 340, 432, 58 },
		{ 340, 433, 58 },
		{ 340, 434, 58 },
		{ 340, 435, 58 },
		{ 340, 436, 58 },
		{ 340, 437, 58 },
		{ 340, 438, 58 },
		{ 340, 439, 58 },
		{ 340, 440, 58 },
		{ 340, 441, 58 },
		{ 340, 442, 58 },
		{ 340, 443, 58 },
		{ 340, 444, 58 },
		{ 340, 445, 58 },
		{ 340, 446, 58 },
		{ 340, 447, 58 },
		{ 340, 448, 58 },
		{ 340, 449, 58 },
		{ 340, 450, 58 },
		{ 340, 451, 58 },
		{ 340, 452, 58 },
		{ 340, 453, 58 },
		{ 340, 454, 58 },
		{ 340, 455, 58 },
		{ 340, 456, 58 },
		{ 340, 457, 58 },
		{ 340, 458, 58 },
		{ 340, 459, 58 },
		{ 340, 460, 58 },
		{ 340, 461, 58 },
		{ 340, 462, 58 },
		{ 340, 463, 58 },
		{ 340, 464, 58 },
		{ 340, 465, 58 },
		{ 340, 466, 58 },
		{ 340, 467, 58 },
		{ 340, 468, 58 },
		{ 340, 469, 58 },
		{ 340, 470, 58 },
		{ 340, 471, 58 },
		{ 340, 472, 58 },
		{ 340, 473, 58 },
		{ 340, 474, 58 },
		{ 340, 475, 58 },
		{ 340, 476, 58 },
		{ 340, 477, 58 },
		{ 340, 478, 58 },
		{ 341, 392, 58 },
		{ 341, 393, 58 },
		{ 341, 394, 58 },
		{ 341, 395, 58 },
		{ 341, 396, 58 },
		{ 341, 397, 58 },
		{ 341, 398, 58 },
		{ 341, 399, 58 },
		{ 341, 400, 58 },
		{ 341, 401, 58 },
		{ 341, 402, 58 },
		{ 341, 403, 58 },
		{ 341, 404, 58 },
		{ 341, 405, 58 },
		{ 341, 406, 58 },
		{ 341, 407, 58 },
		{ 341, 408, 58 },
		{ 341, 409, 58 },
		{ 341, 410, 58 },
		{ 341, 411, 58 },
		{ 341, 412, 58 },
		{ 341, 413, 58 },
		{ 341, 414, 58 },
		{ 341, 415, 58 },
		{ 341, 416, 58 },
		{ 341, 417, 58 },
		{ 341, 418, 58 },
		{ 341, 419, 58 },
		{ 341, 420, 58 },
		{ 341, 421, 58 },
		{ 341, 422, 58 },
		{ 341, 423, 58 },
		{ 341, 424, 58 },
		{ 341, 425, 58 },
		{ 341, 426, 58 },
		{ 341, 427, 58 },
		{ 341, 428, 58 },
		{ 341, 429, 58 },
		{ 341, 430, 58 },
		{ 341, 431, 58 },
		{ 341, 432, 58 },
		{ 341, 433, 58 },
		{ 341, 434, 58 },
		{ 341, 435, 58 },
		{ 341, 436, 58 },
		{ 341, 437, 58 },
		{ 341, 438, 58 },
		{ 341, 439, 58 },
		{ 341, 440, 58 },
		{ 341, 441, 58 },
		{ 341, 442, 58 },
		{ 341, 443, 58 },
		{ 341, 444, 58 },
		{ 341, 445, 58 },
		{ 341, 446, 58 },
		{ 341, 447, 58 },
		{ 341, 448, 58 },
		{ 341, 449, 58 },
		{ 341, 450, 58 },
		{ 341, 451, 58 },
		{ 341, 452, 58 },
		{ 341, 453, 58 },
		{ 341, 454, 58 },
		{ 341, 455, 58 },
		{ 341, 456, 58 },
		{ 341, 457, 58 },
		{ 341, 458, 58 },
		{ 341, 459, 58 },
		{ 341, 460, 58 },
		{ 341, 461, 58 },
		{ 341, 462, 58 },
		{ 341, 463, 58 },
		{ 341, 464, 58 },
		{ 341, 465, 58 },
		{ 341, 466, 58 },
		{ 341, 467, 58 },
		{ 341, 468, 58 },
		{ 341, 469, 58 },
		{ 341, 470, 58 },
		{ 341, 471, 58 },
		{ 341, 472, 58 },
		{ 341, 473, 58 },
		{ 341, 474, 58 },
		{ 341, 475, 58 },
		{ 341, 476, 58 },
		{ 341, 477, 58 },
		{ 341, 478, 58 },
		{ 341, 479, 58 },
		{ 341, 480, 58 },
		{ 342, 394, 58 },
		{ 342, 395, 58 },
		{ 342, 396, 58 },
		{ 342, 397, 58 },
		{ 342, 398, 58 },
		{ 342, 399, 58 },
		{ 342, 400, 58 },
		{ 342, 401, 58 },
		{ 342, 402, 58 },
		{ 342, 403, 58 },
		{ 342, 404, 58 },
		{ 342, 405, 58 },
		{ 342, 406, 58 },
		{ 342, 407, 58 },
		{ 342, 408, 58 },
		{ 342, 409, 58 },
		{ 342, 410, 58 },
		{ 342, 411, 58 },
		{ 342, 412, 58 },
		{ 342, 413, 58 },
		{ 342, 414, 58 },
		{ 342, 415, 58 },
		{ 342, 416, 58 },
		{ 342, 417, 58 },
		{ 342, 418, 58 },
		{ 342, 419, 58 },
		{ 342, 420, 58 },
		{ 342, 421, 58 },
		{ 342, 422, 58 },
		{ 342, 423, 58 },
		{ 342, 424, 58 },
		{ 342, 425, 58 },
		{ 342, 426, 58 },
		{ 342, 427, 58 },
		{ 342, 428, 58 },
		{ 342, 429, 58 },
		{ 342, 430, 58 },
		{ 342, 431, 58 },
		{ 342, 432, 58 },
		{ 342, 433, 58 },
		{ 342, 434, 58 },
		{ 342, 435, 58 },
		{ 342, 436, 58 },
		{ 342, 437, 58 },
		{ 342, 438, 58 },
		{ 342, 439, 58 },
		{ 342, 440, 58 },
		{ 342, 441, 58 },
		{ 342, 442, 58 },
		{ 342, 443, 58 },
		{ 342, 444, 58 },
		{ 342, 445, 58 },
		{ 342, 446, 58 },
		{ 342, 447, 58 },
		{ 342, 448, 58 },
		{ 342, 449, 58 },
		{ 342, 450, 58 },
		{ 342, 451, 58 },
		{ 342, 452, 58 },
		{ 342, 453, 58 },
		{ 342, 454, 58 },
		{ 342, 455, 58 },
		{ 342, 456, 58 },
		{ 342, 457, 58 },
		{ 342, 458, 58 },
		{ 342, 459, 58 },
		{ 342, 460, 58 },
		{ 342, 461, 58 },
		{ 342, 462, 58 },
		{ 342, 463, 58 },
		{ 342, 464, 58 },
		{ 342, 465, 58 },
		{ 342, 466, 58 },
		{ 342, 467, 58 },
		{ 342, 468, 58 },
		{ 342, 469, 58 },
		{ 342, 470, 58 },
		{ 342, 471, 58 },
		{ 342, 472, 58 },
		{ 342, 473, 58 },
		{ 342, 474, 58 },
		{ 342, 475, 58 },
		{ 342, 476, 58 },
		{ 342, 477, 58 },
		{ 342, 478, 58 },
		{ 342, 479, 58 },
		{ 342, 480, 58 },
		{ 342, 481, 58 },
		{ 342, 482, 58 },
		{ 342, 483, 58 },
		{ 343, 395, 58 },
		{ 343, 396, 58 },
		{ 343, 397, 58 },
		{ 343, 398, 58 },
		{ 343, 399, 58 },
		{ 343, 400, 58 },
		{ 343, 401, 58 },
		{ 343, 402, 58 },
		{ 343, 403, 58 },
		{ 343, 404, 58 },
		{ 343, 405, 58 },
		{ 343, 406, 58 },
		{ 343, 407, 58 },
		{ 343, 408, 58 },
		{ 343, 409, 58 },
		{ 343, 410, 58 },
		{ 343, 411, 58 },
		{ 343, 412, 58 },
		{ 343, 413, 58 },
		{ 343, 414, 58 },
		{ 343, 415, 58 },
		{ 343, 416, 58 },
		{ 343, 417, 58 },
		{ 343, 418, 58 },
		{ 343, 419, 58 },
		{ 343, 420, 58 },
		{ 343, 421, 58 },
		{ 343, 422, 58 },
		{ 343, 423, 58 },
		{ 343, 424, 58 },
		{ 343, 425, 58 },
		{ 343, 426, 58 },
		{ 343, 427, 58 },
		{ 343, 428, 58 },
		{ 343, 429, 58 },
		{ 343, 430, 58 },
		{ 343, 431, 58 },
		{ 343, 432, 58 },
		{ 343, 433, 58 },
		{ 343, 434, 58 },
		{ 343, 435, 58 },
		{ 343, 436, 58 },
		{ 343, 437, 58 },
		{ 343, 438, 58 },
		{ 343, 439, 58 },
		{ 343, 440, 58 },
		{ 343, 441, 58 },
		{ 343, 442, 58 },
		{ 343, 443, 58 },
		{ 343, 444, 58 },
		{ 343, 445, 58 },
		{ 343, 446, 58 },
		{ 343, 447, 58 },
		{ 343, 448, 58 },
		{ 343, 449, 58 },
		{ 343, 450, 58 },
		{ 343, 451, 58 },
		{ 343, 452, 58 },
		{ 343, 453, 58 },
		{ 343, 454, 58 },
		{ 343, 455, 58 },
		{ 343, 456, 58 },
		{ 343, 457, 58 },
		{ 343, 458, 58 },
		{ 343, 459, 58 },
		{ 343, 460, 58 },
		{ 343, 461, 58 },
		{ 343, 462, 58 },
		{ 343, 463, 58 },
		{ 343, 464, 58 },
		{ 343, 465, 58 },
		{ 343, 466, 58 },
		{ 343, 467, 58 },
		{ 343, 468, 58 },
		{ 343, 469, 58 },
		{ 343, 470, 58 },
		{ 343, 471, 58 },
		{ 343, 472, 58 },
		{ 343, 473, 58 },
		{ 343, 474, 58 },
		{ 343, 475, 58 },
		{ 343, 476, 58 },
		{ 343, 477, 58 },
		{ 343, 478, 58 },
		{ 343, 479, 58 },
		{ 343, 480, 58 },
		{ 343, 481, 58 },
		{ 343, 482, 58 },
		{ 343, 483, 58 },
		{ 343, 484, 58 },
		{ 343, 485, 58 },
		{ 343, 486, 58 },
		{ 344, 397, 58 },
		{ 344, 398, 58 },
		{ 344, 399, 58 },
		{ 344, 400, 58 },
		{ 344, 401, 58 },
		{ 344, 402, 58 },
		{ 344, 403, 58 },
		{ 344, 404, 58 },
		{ 344, 405, 58 },
		{ 344, 406, 58 },
		{ 344, 407, 58 },
		{ 344, 408, 58 },
		{ 344, 409, 58 },
		{ 344, 410, 58 },
		{ 344, 411, 58 },
		{ 344, 412, 58 },
		{ 344, 413, 58 },
		{ 344, 414, 58 },
		{ 344, 415, 58 },
		{ 344, 416, 58 },
		{ 344, 417, 58 },
		{ 344, 418, 58 },
		{ 344, 419, 58 },
		{ 344, 420, 58 },
		{ 344, 421, 58 },
		{ 344, 422, 58 },
		{ 344, 423, 58 },
		{ 344, 424, 58 },
		{ 344, 425, 58 },
		{ 344, 426, 58 },
		{ 344, 427, 58 },
		{ 344, 428, 58 },
		{ 344, 429, 58 },
		{ 344, 430, 58 },
		{ 344, 431, 58 },
		{ 344, 432, 58 },
		{ 344, 433, 58 },
		{ 344, 434, 58 },
		{ 344, 435, 58 },
		{ 344, 436, 58 },
		{ 344, 437, 58 },
		{ 344, 438, 58 },
		{ 344, 439, 58 },
		{ 344, 440, 58 },
		{ 344, 441, 58 },
		{ 344, 442, 58 },
		{ 344, 443, 58 },
		{ 344, 444, 58 },
		{ 344, 445, 58 },
		{ 344, 446, 58 },
		{ 344, 447, 58 },
		{ 344, 448, 58 },
		{ 344, 449, 58 },
		{ 344, 450, 58 },
		{ 344, 451, 58 },
		{ 344, 452, 58 },
		{ 344, 453, 58 },
		{ 344, 454, 58 },
		{ 344, 455, 58 },
		{ 344, 456, 58 },
		{ 344, 457, 58 },
		{ 344, 458, 58 },
		{ 344, 459, 58 },
		{ 344, 460, 58 },
		{ 344, 461, 58 },
		{ 344, 462, 58 },
		{ 344, 463, 58 },
		{ 344, 464, 58 },
		{ 344, 465, 58 },
		{ 344, 466, 58 },
		{ 344, 467, 58 },
		{ 344, 468, 58 },
		{ 344, 469, 58 },
		{ 344, 470, 58 },
		{ 344, 471, 58 },
		{ 344, 472, 58 },
		{ 344, 473, 58 },
		{ 344, 474, 58 },
		{ 344, 475, 58 },
		{ 344, 476, 58 },
		{ 344, 477, 58 },
		{ 344, 478, 58 },
		{ 344, 479, 58 },
		{ 344, 480, 58 },
		{ 344, 481, 58 },
		{ 344, 482, 58 },
		{ 344, 483, 58 },
		{ 344, 484, 58 },
		{ 344, 485, 58 },
		{ 344, 486, 58 },
		{ 344, 487, 58 },
		{ 344, 488, 58 },
		{ 345, 398, 58 },
		{ 345, 399, 58 },
		{ 345, 400, 58 },
		{ 345, 401, 58 },
		{ 345, 402, 58 },
		{ 345, 403, 58 },
		{ 345, 404, 58 },
		{ 345, 405, 58 },
		{ 345, 406, 58 },
		{ 345, 407, 58 },
		{ 345, 408, 58 },
		{ 345, 409, 58 },
		{ 345, 410, 58 },
		{ 345, 411, 58 },
		{ 345, 412, 58 },
		{ 345, 413, 58 },
		{ 345, 414, 58 },
		{ 345, 415, 58 },
		{ 345, 416, 58 },
		{ 345, 417, 58 },
		{ 345, 418, 58 },
		{ 345, 419, 58 },
		{ 345, 420, 58 },
		{ 345, 421, 58 },
		{ 345, 422, 58 },
		{ 345, 423, 58 },
		{ 345, 424, 58 },
		{ 345, 425, 58 },
		{ 345, 426, 58 },
		{ 345, 427, 58 },
		{ 345, 428, 58 },
		{ 345, 429, 58 },
		{ 345, 430, 58 },
		{ 345, 431, 58 },
		{ 345, 432, 58 },
		{ 345, 433, 58 },
		{ 345, 434, 58 },
		{ 345, 435, 58 },
		{ 345, 436, 58 },
		{ 345, 437, 58 },
		{ 345, 438, 58 },
		{ 345, 439, 58 },
		{ 345, 440, 58 },
		{ 345, 441, 58 },
		{ 345, 442, 58 },
		{ 345, 443, 58 },
		{ 345, 444, 58 },
		{ 345, 445, 58 },
		{ 345, 446, 58 },
		{ 345, 447, 58 },
		{ 345, 448, 58 },
		{ 345, 449, 58 },
		{ 345, 450, 58 },
		{ 345, 451, 58 },
		{ 345, 452, 58 },
		{ 345, 453, 58 },
		{ 345, 454, 58 },
		{ 345, 455, 58 },
		{ 345, 456, 58 },
		{ 345, 457, 58 },
		{ 345, 458, 58 },
		{ 345, 459, 58 },
		{ 345, 460, 58 },
		{ 345, 461, 58 },
		{ 345, 462, 58 },
		{ 345, 463, 58 },
		{ 345, 464, 58 },
		{ 345, 465, 58 },
		{ 345, 466, 58 },
		{ 345, 467, 58 },
		{ 345, 468, 58 },
		{ 345, 469, 58 },
		{ 345, 470, 58 },
		{ 345, 471, 58 },
		{ 345, 472, 58 },
		{ 345, 473, 58 },
		{ 345, 474, 58 },
		{ 345, 475, 58 },
		{ 345, 476, 58 },
		{ 345, 477, 58 },
		{ 345, 478, 58 },
		{ 345, 479, 58 },
		{ 345, 480, 58 },
		{ 345, 481, 58 },
		{ 345, 482, 58 },
		{ 345, 483, 58 },
		{ 345, 484, 58 },
		{ 345, 485, 58 },
		{ 345, 486, 58 },
		{ 345, 487, 58 },
		{ 345, 488, 58 },
		{ 345, 489, 58 },
		{ 345, 490, 58 },
		{ 345, 491, 58 },
		{ 346, 400, 58 },
		{ 346, 401, 58 },
		{ 346, 402, 58 },
		{ 346, 403, 58 },
		{ 346, 404, 58 },
		{ 346, 405, 58 },
		{ 346, 406, 58 },
		{ 346, 407, 58 },
		{ 346, 408, 58 },
		{ 346, 409, 58 },
		{ 346, 410, 58 },
		{ 346, 411, 58 },
		{ 346, 412, 58 },
		{ 346, 413, 58 },
		{ 346, 414, 58 },
		{ 346, 415, 58 },
		{ 346, 416, 58 },
		{ 346, 417, 58 },
		{ 346, 418, 58 },
		{ 346, 419, 58 },
		{ 346, 420, 58 },
		{ 346, 421, 58 },
		{ 346, 422, 58 },
		{ 346, 423, 58 },
		{ 346, 424, 58 },
		{ 346, 425, 58 },
		{ 346, 426, 58 },
		{ 346, 427, 58 },
		{ 346, 428, 58 },
		{ 346, 429, 58 },
		{ 346, 430, 58 },
		{ 346, 431, 58 },
		{ 346, 432, 58 },
		{ 346, 433, 58 },
		{ 346, 434, 58 },
		{ 346, 435, 58 },
		{ 346, 436, 58 },
		{ 346, 437, 58 },
		{ 346, 438, 58 },
		{ 346, 439, 58 },
		{ 346, 440, 58 },
		{ 346, 441, 58 },
		{ 346, 442, 58 },
		{ 346, 443, 58 },
		{ 346, 444, 58 },
		{ 346, 445, 58 },
		{ 346, 446, 58 },
		{ 346, 447, 58 },
		{ 346, 448, 58 },
		{ 346, 449, 58 },
		{ 346, 450, 58 },
		{ 346, 451, 58 },
		{ 346, 452, 58 },
		{ 346, 453, 58 },
		{ 346, 454, 58 },
		{ 346, 455, 58 },
		{ 346, 456, 58 },
		{ 346, 457, 58 },
		{ 346, 458, 58 },
		{ 346, 459, 58 },
		{ 346, 460, 58 },
		{ 346, 461, 58 },
		{ 346, 462, 58 },
		{ 346, 463, 58 },
		{ 346, 464, 58 },
		{ 346, 465, 58 },
		{ 346, 466, 58 },
		{ 346, 467, 58 },
		{ 346, 468, 58 },
		{ 346, 469, 58 },
		{ 346, 470, 58 },
		{ 346, 471, 58 },
		{ 346, 472, 58 },
		{ 346, 473, 58 },
		{ 346, 474, 58 },
		{ 346, 475, 58 },
		{ 346, 476, 58 },
		{ 346, 477, 58 },
		{ 346, 478, 58 },
		{ 346, 479, 58 },
		{ 346, 480, 58 },
		{ 346, 481, 58 },
		{ 346, 482, 58 },
		{ 346, 483, 58 },
		{ 346, 484, 58 },
		{ 346, 485, 58 },
		{ 346, 486, 58 },
		{ 346, 487, 58 },
		{ 346, 488, 58 },
		{ 346, 489, 58 },
		{ 346, 490, 58 },
		{ 346, 491, 58 },
		{ 346, 492, 58 },
		{ 346, 493, 58 },
		{ 346, 494, 58 },
		{ 347, 402, 58 },
		{ 347, 403, 58 },
		{ 347, 404, 58 },
		{ 347, 405, 58 },
		{ 347, 406, 58 },
		{ 347, 407, 58 },
		{ 347, 408, 58 },
		{ 347, 409, 58 },
		{ 347, 410, 58 },
		{ 347, 411, 58 },
		{ 347, 412, 58 },
		{ 347, 413, 58 },
		{ 347, 414, 58 },
		{ 347, 415, 58 },
		{ 347, 416, 58 },
		{ 347, 417, 58 },
		{ 347, 418, 58 },
		{ 347, 419, 58 },
		{ 347, 420, 58 },
		{ 347, 421, 58 },
		{ 347, 422, 58 },
		{ 347, 423, 58 },
		{ 347, 424, 58 },
		{ 347, 425, 58 },
		{ 347, 426, 58 },
		{ 347, 427, 58 },
		{ 347, 428, 58 },
		{ 347, 429, 58 },
		{ 347, 430, 58 },
		{ 347, 431, 58 },
		{ 347, 432, 58 },
		{ 347, 433, 58 },
		{ 347, 434, 58 },
		{ 347, 435, 58 },
		{ 347, 436, 58 },
		{ 347, 437, 58 },
		{ 347, 438, 58 },
		{ 347, 439, 58 },
		{ 347, 440, 58 },
		{ 347, 441, 58 },
		{ 347, 442, 58 },
		{ 347, 443, 58 },
		{ 347, 444, 58 },
		{ 347, 445, 58 },
		{ 347, 446, 58 },
		{ 347, 447, 58 },
		{ 347, 448, 58 },
		{ 347, 449, 58 },
		{ 347, 450, 58 },
		{ 347, 451, 58 },
		{ 347, 452, 58 },
		{ 347, 453, 58 },
		{ 347, 454, 58 },
		{ 347, 455, 58 },
		{ 347, 456, 58 },
		{ 347, 457, 58 },
		{ 347, 458, 58 },
		{ 347, 459, 58 },
		{ 347, 460, 58 },
		{ 347, 461, 58 },
		{ 347, 462, 58 },
		{ 347, 463, 58 },
		{ 347, 464, 58 },
		{ 347, 465, 58 },
		{ 347, 466, 58 },
		{ 347, 467, 58 },
		{ 347, 468, 58 },
		{ 347, 469, 58 },
		{ 347, 470, 58 },
		{ 347, 471, 58 },
		{ 347, 472, 58 },
		{ 347, 473, 58 },
		{ 347, 474, 58 },
		{ 347, 475, 58 },
		{ 347, 476, 58 },
		{ 347, 477, 58 },
		{ 347, 478, 58 },
		{ 347, 479, 58 },
		{ 347, 480, 58 },
		{ 347, 481, 58 },
		{ 347, 482, 58 },
		{ 347, 483, 58 },
		{ 347, 484, 58 },
		{ 347, 485, 58 },
		{ 347, 486, 58 },
		{ 347, 487, 58 },
		{ 347, 488, 58 },
		{ 347, 489, 58 },
		{ 347, 490, 58 },
		{ 347, 491, 58 },
		{ 347, 492, 58 },
		{ 347, 493, 58 },
		{ 347, 494, 58 },
		{ 347, 495, 58 },
		{ 347, 496, 58 },
		{ 348, 403, 58 },
		{ 348, 404, 58 },
		{ 348, 405, 58 },
		{ 348, 406, 58 },
		{ 348, 407, 58 },
		{ 348, 408, 58 },
		{ 348, 409, 58 },
		{ 348, 410, 58 },
		{ 348, 411, 58 },
		{ 348, 412, 58 },
		{ 348, 413, 58 },
		{ 348, 414, 58 },
		{ 348, 415, 58 },
		{ 348, 416, 58 },
		{ 348, 417, 58 },
		{ 348, 418, 58 },
		{ 348, 419, 58 },
		{ 348, 420, 58 },
		{ 348, 421, 58 },
		{ 348, 422, 58 },
		{ 348, 423, 58 },
		{ 348, 424, 58 },
		{ 348, 425, 58 },
		{ 348, 426, 58 },
		{ 348, 427, 58 },
		{ 348, 428, 58 },
		{ 348, 429, 58 },
		{ 348, 430, 58 },
		{ 348, 431, 58 },
		{ 348, 432, 58 },
		{ 348, 433, 58 },
		{ 348, 434, 58 },
		{ 348, 435, 58 },
		{ 348, 436, 58 },
		{ 348, 437, 58 },
		{ 348, 438, 58 },
		{ 348, 439, 58 },
		{ 348, 440, 58 },
		{ 348, 441, 58 },
		{ 348, 442, 58 },
		{ 348, 443, 58 },
		{ 348, 444, 58 },
		{ 348, 445, 58 },
		{ 348, 446, 58 },
		{ 348, 447, 58 },
		{ 348, 448, 58 },
		{ 348, 449, 58 },
		{ 348, 450, 58 },
		{ 348, 451, 58 },
		{ 348, 452, 58 },
		{ 348, 453, 58 },
		{ 348, 454, 58 },
		{ 348, 455, 58 },
		{ 348, 456, 58 },
		{ 348, 457, 58 },
		{ 348, 458, 58 },
		{ 348, 459, 58 },
		{ 348, 460, 58 },
		{ 348, 461, 58 },
		{ 348, 462, 58 },
		{ 348, 463, 58 },
		{ 348, 464, 58 },
		{ 348, 465, 58 },
		{ 348, 466, 58 },
		{ 348, 467, 58 },
		{ 348, 468, 58 },
		{ 348, 469, 58 },
		{ 348, 470, 58 },
		{ 348, 471, 58 },
		{ 348, 472, 58 },
		{ 348, 473, 58 },
		{ 348, 474, 58 },
		{ 348, 475, 58 },
		{ 348, 476, 58 },
		{ 348, 477, 58 },
		{ 348, 478, 58 },
		{ 348, 479, 58 },
		{ 348, 480, 58 },
		{ 348, 481, 58 },
		{ 348, 482, 58 },
		{ 348, 483, 58 },
		{ 348, 484, 58 },
		{ 348, 485, 58 },
		{ 348, 486, 58 },
		{ 348, 487, 58 },
		{ 348, 488, 58 },
		{ 348, 489, 58 },
		{ 348, 490, 58 },
		{ 348, 491, 58 },
		{ 348, 492, 58 },
		{ 348, 493, 58 },
		{ 348, 494, 58 },
		{ 348, 495, 58 },
		{ 348, 496, 58 },
		{ 348, 497, 58 },
		{ 348, 498, 58 },
		{ 348, 499, 58 },
		{ 349, 405, 58 },
		{ 349, 406, 58 },
		{ 349, 407, 58 },
		{ 349, 408, 58 },
		{ 349, 409, 58 },
		{ 349, 410, 58 },
		{ 349, 411, 58 },
		{ 349, 412, 58 },
		{ 349, 413, 58 },
		{ 349, 414, 58 },
		{ 349, 415, 58 },
		{ 349, 416, 58 },
		{ 349, 417, 58 },
		{ 349, 418, 58 },
		{ 349, 419, 58 },
		{ 349, 420, 58 },
		{ 349, 421, 58 },
		{ 349, 422, 58 },
		{ 349, 423, 58 },
		{ 349, 424, 58 },
		{ 349, 425, 58 },
		{ 349, 426, 58 },
		{ 349, 427, 58 },
		{ 349, 428, 58 },
		{ 349, 429, 58 },
		{ 349, 430, 58 },
		{ 349, 431, 58 },
		{ 349, 432, 58 },
		{ 349, 433, 58 },
		{ 349, 434, 58 },
		{ 349, 435, 58 },
		{ 349, 436, 58 },
		{ 349, 437, 58 },
		{ 349, 438, 58 },
		{ 349, 439, 58 },
		{ 349, 440, 58 },
		{ 349, 441, 58 },
		{ 349, 442, 58 },
		{ 349, 443, 58 },
		{ 349, 444, 58 },
		{ 349, 445, 58 },
		{ 349, 446, 58 },
		{ 349, 447, 58 },
		{ 349, 448, 58 },
		{ 349, 449, 58 },
		{ 349, 450, 58 },
		{ 349, 451, 58 },
		{ 349, 452, 58 },
		{ 349, 453, 58 },
		{ 349, 454, 58 },
		{ 349, 455, 58 },
		{ 349, 456, 58 },
		{ 349, 457, 58 },
		{ 349, 458, 58 },
		{ 349, 459, 58 },
		{ 349, 460, 58 },
		{ 349, 461, 58 },
		{ 349, 462, 58 },
		{ 349, 463, 58 },
		{ 349, 464, 58 },
		{ 349, 465, 58 },
		{ 349, 466, 58 },
		{ 349, 467, 58 },
		{ 349, 468, 58 },
		{ 349, 469, 58 },
		{ 349, 470, 58 },
		{ 349, 471, 58 },
		{ 349, 472, 58 },
		{ 349, 473, 58 },
		{ 349, 474, 58 },
		{ 349, 475, 58 },
		{ 349, 476, 58 },
		{ 349, 477, 58 },
		{ 349, 478, 58 },
		{ 349, 479, 58 },
		{ 349, 480, 58 },
		{ 349, 481, 58 },
		{ 349, 482, 58 },
		{ 349, 483, 58 },
		{ 349, 484, 58 },
		{ 349, 485, 58 },
		{ 349, 486, 58 },
		{ 349, 487, 58 },
		{ 349, 488, 58 },
		{ 349, 489, 58 },
		{ 349, 490, 58 },
		{ 349, 491, 58 },
		{ 349, 492, 58 },
		{ 349, 493, 58 },
		{ 349, 494, 58 },
		{ 349, 495, 58 },
		{ 349, 496, 58 },
		{ 349, 497, 58 },
		{ 349, 498, 58 },
		{ 349, 499, 58 },
		{ 349, 500, 58 },
		{ 349, 501, 58 },
		{ 349, 502, 58 },
		{ 350, 406, 58 },
		{ 350, 407, 58 },
		{ 350, 408, 58 },
		{ 350, 409, 58 },
		{ 350, 410, 58 },
		{ 350, 411, 58 },
		{ 350, 412, 58 },
		{ 350, 413, 58 },
		{ 350, 414, 58 },
		{ 350, 415, 58 },
		{ 350, 416, 58 },
		{ 350, 417, 58 },
		{ 350, 418, 58 },
		{ 350, 419, 58 },
		{ 350, 420, 58 },
		{ 350, 421, 58 },
		{ 350, 422, 58 },
		{ 350, 423, 58 },
		{ 350, 424, 58 },
		{ 350, 425, 58 },
		{ 350, 426, 58 },
		{ 350, 427, 58 },
		{ 350, 428, 58 },
		{ 350, 429, 58 },
		{ 350, 430, 58 },
		{ 350, 431, 58 },
		{ 350, 432, 58 },
		{ 350, 433, 58 },
		{ 350, 434, 58 },
		{ 350, 435, 58 },
		{ 350, 436, 58 },
		{ 350, 437, 58 },
		{ 350, 438, 58 },
		{ 350, 439, 58 },
		{ 350, 440, 58 },
		{ 350, 441, 58 },
		{ 350, 442, 58 },
		{ 350, 443, 58 },
		{ 350, 444, 58 },
		{ 350, 445, 58 },
		{ 350, 446, 58 },
		{ 350, 447, 58 },
		{ 350, 448, 58 },
		{ 350, 449, 58 },
		{ 350, 450, 58 },
		{ 350, 451, 58 },
		{ 350, 452, 58 },
		{ 350, 453, 58 },
		{ 350, 454, 58 },
		{ 350, 455, 58 },
		{ 350, 456, 58 },
		{ 350, 457, 58 },
		{ 350, 458, 58 },
		{ 350, 459, 58 },
		{ 350, 460, 58 },
		{ 350, 461, 58 },
		{ 350, 462, 58 },
		{ 350, 463, 58 },
		{ 350, 464, 58 },
		{ 350, 465, 58 },
		{ 350, 466, 58 },
		{ 350, 467, 58 },
		{ 350, 468, 58 },
		{ 350, 469, 58 },
		{ 350, 470, 58 },
		{ 350, 471, 58 },
		{ 350, 472, 58 },
		{ 350, 473, 58 },
		{ 350, 474, 58 },
		{ 350, 475, 58 },
		{ 350, 476, 58 },
		{ 350, 477, 58 },
		{ 350, 478, 58 },
		{ 350, 479, 58 },
		{ 350, 480, 58 },
		{ 350, 481, 58 },
		{ 350, 482, 58 },
		{ 350, 483, 58 },
		{ 350, 484, 58 },
		{ 350, 485, 58 },
		{ 350, 486, 58 },
		{ 350, 487, 58 },
		{ 350, 488, 58 },
		{ 350, 489, 58 },
		{ 350, 490, 58 },
		{ 350, 491, 58 },
		{ 350, 492, 58 },
		{ 350, 493, 58 },
		{ 350, 494, 58 },
		{ 350, 495, 58 },
		{ 350, 496, 58 },
		{ 350, 497, 58 },
		{ 350, 498, 58 },
		{ 350, 499, 58 },
		{ 350, 500, 58 },
		{ 350, 501, 58 },
		{ 350, 502, 58 },
		{ 350, 503, 58 },
		{ 350, 504, 58 },
		{ 351, 408, 58 },
		{ 351, 409, 58 },
		{ 351, 410, 58 },
		{ 351, 411, 58 },
		{ 351, 412, 58 },
		{ 351, 413, 58 },
		{ 351, 414, 58 },
		{ 351, 415, 58 },
		{ 351, 416, 58 },
		{ 351, 417, 58 },
		{ 351, 418, 58 },
		{ 351, 419, 58 },
		{ 351, 420, 58 },
		{ 351, 421, 58 },
		{ 351, 422, 58 },
		{ 351, 423, 58 },
		{ 351, 424, 58 },
		{ 351, 425, 58 },
		{ 351, 426, 58 },
		{ 351, 427, 58 },
		{ 351, 428, 58 },
		{ 351, 429, 58 },
		{ 351, 430, 58 },
		{ 351, 431, 58 },
		{ 351, 432, 58 },
		{ 351, 433, 58 },
		{ 351, 434, 58 },
		{ 351, 435, 58 },
		{ 351, 436, 58 },
		{ 351, 437, 58 },
		{ 351, 438, 58 },
		{ 351, 439, 58 },
		{ 351, 440, 58 },
		{ 351, 441, 58 },
		{ 351, 442, 58 },
		{ 351, 443, 58 },
		{ 351, 444, 58 },
		{ 351, 445, 58 },
		{ 351, 446, 58 },
		{ 351, 447, 58 },
		{ 351, 448, 58 },
		{ 351, 449, 58 },
		{ 351, 450, 58 },
		{ 351, 451, 58 },
		{ 351, 452, 58 },
		{ 351, 453, 58 },
		{ 351, 454, 58 },
		{ 351, 455, 58 },
		{ 351, 456, 58 },
		{ 351, 457, 58 },
		{ 351, 458, 58 },
		{ 351, 459, 58 },
		{ 351, 460, 58 },
		{ 351, 461, 58 },
		{ 351, 462, 58 },
		{ 351, 463, 58 },
		{ 351, 464, 58 },
		{ 351, 465, 58 },
		{ 351, 466, 58 },
		{ 351, 467, 58 },
		{ 351, 468, 58 },
		{ 351, 469, 58 },
		{ 351, 470, 58 },
		{ 351, 471, 58 },
		{ 351, 472, 58 },
		{ 351, 473, 58 },
		{ 351, 474, 58 },
		{ 351, 475, 58 },
		{ 351, 476, 58 },
		{ 351, 477, 58 },
		{ 351, 478, 58 },
		{ 351, 479, 58 },
		{ 351, 480, 58 },
		{ 351, 481, 58 },
		{ 351, 482, 58 },
		{ 351, 483, 58 },
		{ 351, 484, 58 },
		{ 351, 485, 58 },
		{ 351, 486, 58 },
		{ 351, 487, 58 },
		{ 351, 488, 58 },
		{ 351, 489, 58 },
		{ 351, 490, 58 },
		{ 351, 491, 58 },
		{ 351, 492, 58 },
		{ 351, 493, 58 },
		{ 351, 494, 58 },
		{ 351, 495, 58 },
		{ 351, 496, 58 },
		{ 351, 497, 58 },
		{ 351, 498, 58 },
		{ 351, 499, 58 },
		{ 351, 500, 58 },
		{ 351, 501, 58 },
		{ 351, 502, 58 },
		{ 351, 503, 58 },
		{ 351, 504, 58 },
		{ 351, 505, 58 },
		{ 351, 506, 58 },
		{ 351, 507, 58 },
		{ 352, 410, 58 },
		{ 352, 411, 58 },
		{ 352, 412, 58 },
		{ 352, 413, 58 },
		{ 352, 414, 58 },
		{ 352, 415, 58 },
		{ 352, 416, 58 },
		{ 352, 417, 58 },
		{ 352, 418, 58 },
		{ 352, 419, 58 },
		{ 352, 420, 58 },
		{ 352, 421, 58 },
		{ 352, 422, 58 },
		{ 352, 423, 58 },
		{ 352, 424, 58 },
		{ 352, 425, 58 },
		{ 352, 426, 58 },
		{ 352, 427, 58 },
		{ 352, 428, 58 },
		{ 352, 429, 58 },
		{ 352, 430, 58 },
		{ 352, 431, 58 },
		{ 352, 432, 58 },
		{ 352, 433, 58 },
		{ 352, 434, 58 },
		{ 352, 435, 58 },
		{ 352, 436, 58 },
		{ 352, 437, 58 },
		{ 352, 438, 58 },
		{ 352, 439, 58 },
		{ 352, 440, 58 },
		{ 352, 441, 58 },
		{ 352, 442, 58 },
		{ 352, 443, 58 },
		{ 352, 444, 58 },
		{ 352, 445, 58 },
		{ 352, 446, 58 },
		{ 352, 447, 58 },
		{ 352, 448, 58 },
		{ 352, 449, 58 },
		{ 352, 450, 58 },
		{ 352, 451, 58 },
		{ 352, 452, 58 },
		{ 352, 453, 58 },
		{ 352, 454, 58 },
		{ 352, 455, 58 },
		{ 352, 456, 58 },
		{ 352, 457, 58 },
		{ 352, 458, 58 },
		{ 352, 459, 58 },
		{ 352, 460, 58 },
		{ 352, 461, 58 },
		{ 352, 462, 58 },
		{ 352, 463, 58 },
		{ 352, 464, 58 },
		{ 352, 465, 58 },
		{ 352, 466, 58 },
		{ 352, 467, 58 },
		{ 352, 468, 58 },
		{ 352, 469, 58 },
		{ 352, 470, 58 },
		{ 352, 471, 58 },
		{ 352, 472, 58 },
		{ 352, 473, 58 },
		{ 352, 474, 58 },
		{ 352, 475, 58 },
		{ 352, 476, 58 },
		{ 352, 477, 58 },
		{ 352, 478, 58 },
		{ 352, 479, 58 },
		{ 352, 480, 58 },
		{ 352, 481, 58 },
		{ 352, 482, 58 },
		{ 352, 483, 58 },
		{ 352, 484, 58 },
		{ 352, 485, 58 },
		{ 352, 486, 58 },
		{ 352, 487, 58 },
		{ 352, 488, 58 },
		{ 352, 489, 58 },
		{ 352, 490, 58 },
		{ 352, 491, 58 },
		{ 352, 492, 58 },
		{ 352, 493, 58 },
		{ 352, 494, 58 },
		{ 352, 495, 58 },
		{ 352, 496, 58 },
		{ 352, 497, 58 },
		{ 352, 498, 58 },
		{ 352, 499, 58 },
		{ 352, 500, 58 },
		{ 352, 501, 58 },
		{ 352, 502, 58 },
		{ 352, 503, 58 },
		{ 352, 504, 58 },
		{ 352, 505, 58 },
		{ 352, 506, 58 },
		{ 352, 507, 58 },
		{ 352, 508, 58 },
		{ 352, 509, 58 },
		{ 352, 510, 58 },
		{ 353, 411, 58 },
		{ 353, 412, 58 },
		{ 353, 413, 58 },
		{ 353, 414, 58 },
		{ 353, 415, 58 },
		{ 353, 416, 58 },
		{ 353, 417, 58 },
		{ 353, 418, 58 },
		{ 353, 419, 58 },
		{ 353, 420, 58 },
		{ 353, 421, 58 },
		{ 353, 422, 58 },
		{ 353, 423, 58 },
		{ 353, 424, 58 },
		{ 353, 425, 58 },
		{ 353, 426, 58 },
		{ 353, 427, 58 },
		{ 353, 428, 58 },
		{ 353, 429, 58 },
		{ 353, 430, 58 },
		{ 353, 431, 58 },
		{ 353, 432, 58 },
		{ 353, 433, 58 },
		{ 353, 434, 58 },
		{ 353, 435, 58 },
		{ 353, 436, 58 },
		{ 353, 437, 58 },
		{ 353, 438, 58 },
		{ 353, 439, 58 },
		{ 353, 440, 58 },
		{ 353, 441, 58 },
		{ 353, 442, 58 },
		{ 353, 443, 58 },
		{ 353, 444, 58 },
		{ 353, 445, 58 },
		{ 353, 446, 58 },
		{ 353, 447, 58 },
		{ 353, 448, 58 },
		{ 353, 449, 58 },
		{ 353, 450, 58 },
		{ 353, 451, 58 },
		{ 353, 452, 58 },
		{ 353, 453, 58 },
		{ 353, 454, 58 },
		{ 353, 455, 58 },
		{ 353, 456, 58 },
		{ 353, 457, 58 },
		{ 353, 458, 58 },
		{ 353, 459, 58 },
		{ 353, 460, 58 },
		{ 353, 461, 58 },
		{ 353, 462, 58 },
		{ 353, 463, 58 },
		{ 353, 464, 58 },
		{ 353, 465, 58 },
		{ 353, 466, 58 },
		{ 353, 467, 58 },
		{ 353, 468, 58 },
		{ 353, 469, 58 },
		{ 353, 470, 58 },
		{ 353, 471, 58 },
		{ 353, 472, 58 },
		{ 353, 473, 58 },
		{ 353, 474, 58 },
		{ 353, 475, 58 },
		{ 353, 476, 58 },
		{ 353, 477, 58 },
		{ 353, 478, 58 },
		{ 353, 479, 58 },
		{ 353, 480, 58 },
		{ 353, 481, 58 },
		{ 353, 482, 58 },
		{ 353, 483, 58 },
		{ 353, 484, 58 },
		{ 353, 485, 58 },
		{ 353, 486, 58 },
		{ 353, 487, 58 },
		{ 353, 488, 58 },
		{ 353, 489, 58 },
		{ 353, 490, 58 },
		{ 353, 491, 58 },
		{ 353, 492, 58 },
		{ 353, 493, 58 },
		{ 353, 494, 58 },
		{ 353, 495, 58 },
		{ 353, 496, 58 },
		{ 353, 497, 58 },
		{ 353, 498, 58 },
		{ 353, 499, 58 },
		{ 353, 500, 58 },
		{ 353, 501, 58 },
		{ 353, 502, 58 },
		{ 353, 503, 58 },
		{ 353, 504, 58 },
		{ 353, 505, 58 },
		{ 353, 506, 58 },
		{ 353, 507, 58 },
		{ 353, 508, 58 },
		{ 353, 509, 58 },
		{ 353, 510, 58 },
		{ 353, 511, 58 },
		{ 353, 512, 58 },
		{ 354, 413, 58 },
		{ 354, 414, 58 },
		{ 354, 415, 58 },
		{ 354, 416, 58 },
		{ 354, 417, 58 },
		{ 354, 418, 58 },
		{ 354, 419, 58 },
		{ 354, 420, 58 },
		{ 354, 421, 58 },
		{ 354, 422, 58 },
		{ 354, 423, 58 },
		{ 354, 424, 58 },
		{ 354, 425, 58 },
		{ 354, 426, 58 },
		{ 354, 427, 58 },
		{ 354, 428, 58 },
		{ 354, 429, 58 },
		{ 354, 430, 58 },
		{ 354, 431, 58 },
		{ 354, 432, 58 },
		{ 354, 433, 58 },
		{ 354, 434, 58 },
		{ 354, 435, 58 },
		{ 354, 436, 58 },
		{ 354, 437, 58 },
		{ 354, 438, 58 },
		{ 354, 439, 58 },
		{ 354, 440, 58 },
		{ 354, 441, 58 },
		{ 354, 442, 58 },
		{ 354, 443, 58 },
		{ 354, 444, 58 },
		{ 354, 445, 58 },
		{ 354, 446, 58 },
		{ 354, 447, 58 },
		{ 354, 448, 58 },
		{ 354, 449, 58 },
		{ 354, 450, 58 },
		{ 354, 451, 58 },
		{ 354, 452, 58 },
		{ 354, 453, 58 },
		{ 354, 454, 58 },
		{ 354, 455, 58 },
		{ 354, 456, 58 },
		{ 354, 457, 58 },
		{ 354, 458, 58 },
		{ 354, 459, 58 },
		{ 354, 460, 58 },
		{ 354, 461, 58 },
		{ 354, 462, 58 },
		{ 354, 463, 58 },
		{ 354, 464, 58 },
		{ 354, 465, 58 },
		{ 354, 466, 58 },
		{ 354, 467, 58 },
		{ 354, 468, 58 },
		{ 354, 469, 58 },
		{ 354, 470, 58 },
		{ 354, 471, 58 },
		{ 354, 472, 58 },
		{ 354, 473, 58 },
		{ 354, 474, 58 },
		{ 354, 475, 58 },
		{ 354, 476, 58 },
		{ 354, 477, 58 },
		{ 354, 478, 58 },
		{ 354, 479, 58 },
		{ 354, 480, 58 },
		{ 354, 481, 58 },
		{ 354, 482, 58 },
		{ 354, 483, 58 },
		{ 354, 484, 58 },
		{ 354, 485, 58 },
		{ 354, 486, 58 },
		{ 354, 487, 58 },
		{ 354, 488, 58 },
		{ 354, 489, 58 },
		{ 354, 490, 58 },
		{ 354, 491, 58 },
		{ 354, 492, 58 },
		{ 354, 493, 58 },
		{ 354, 494, 58 },
		{ 354, 495, 58 },
		{ 354, 496, 58 },
		{ 354, 497, 58 },
		{ 354, 498, 58 },
		{ 354, 499, 58 },
		{ 354, 500, 58 },
		{ 354, 501, 58 },
		{ 354, 502, 58 },
		{ 354, 503, 58 },
		{ 354, 504, 58 },
		{ 354, 505, 58 },
		{ 354, 506, 58 },
		{ 354, 507, 58 },
		{ 354, 508, 58 },
		{ 354, 509, 58 },
		{ 354, 510, 58 },
		{ 354, 511, 58 },
		{ 354, 512, 58 },
		{ 355, 414, 58 },
		{ 355, 415, 58 },
		{ 355, 416, 58 },
		{ 355, 417, 58 },
		{ 355, 418, 58 },
		{ 355, 419, 58 },
		{ 355, 420, 58 },
		{ 355, 421, 58 },
		{ 355, 422, 58 },
		{ 355, 423, 58 },
		{ 355, 424, 58 },
		{ 355, 425, 58 },
		{ 355, 426, 58 },
		{ 355, 427, 58 },
		{ 355, 428, 58 },
		{ 355, 429, 58 },
		{ 355, 430, 58 },
		{ 355, 431, 58 },
		{ 355, 432, 58 },
		{ 355, 433, 58 },
		{ 355, 434, 58 },
		{ 355, 435, 58 },
		{ 355, 436, 58 },
		{ 355, 437, 58 },
		{ 355, 438, 58 },
		{ 355, 439, 58 },
		{ 355, 440, 58 },
		{ 355, 441, 58 },
		{ 355, 442, 58 },
		{ 355, 443, 58 },
		{ 355, 444, 58 },
		{ 355, 445, 58 },
		{ 355, 446, 58 },
		{ 355, 447, 58 },
		{ 355, 448, 58 },
		{ 355, 449, 58 },
		{ 355, 450, 58 },
		{ 355, 451, 58 },
		{ 355, 452, 58 },
		{ 355, 453, 58 },
		{ 355, 454, 58 },
		{ 355, 455, 58 },
		{ 355, 456, 58 },
		{ 355, 457, 58 },
		{ 355, 458, 58 },
		{ 355, 459, 58 },
		{ 355, 460, 58 },
		{ 355, 461, 58 },
		{ 355, 462, 58 },
		{ 355, 463, 58 },
		{ 355, 464, 58 },
		{ 355, 465, 58 },
		{ 355, 466, 58 },
		{ 355, 467, 58 },
		{ 355, 468, 58 },
		{ 355, 469, 58 },
		{ 355, 470, 58 },
		{ 355, 471, 58 },
		{ 355, 472, 58 },
		{ 355, 473, 58 },
		{ 355, 474, 58 },
		{ 355, 475, 58 },
		{ 355, 476, 58 },
		{ 355, 477, 58 },
		{ 355, 478, 58 },
		{ 355, 479, 58 },
		{ 355, 480, 58 },
		{ 355, 481, 58 },
		{ 355, 482, 58 },
		{ 355, 483, 58 },
		{ 355, 484, 58 },
		{ 355, 485, 58 },
		{ 355, 486, 58 },
		{ 355, 487, 58 },
		{ 355, 488, 58 },
		{ 355, 489, 58 },
		{ 355, 490, 58 },
		{ 355, 491, 58 },
		{ 355, 492, 58 },
		{ 355, 493, 58 },
		{ 355, 494, 58 },
		{ 355, 495, 58 },
		{ 355, 496, 58 },
		{ 355, 497, 58 },
		{ 355, 498, 58 },
		{ 355, 499, 58 },
		{ 355, 500, 58 },
		{ 355, 501, 58 },
		{ 355, 502, 58 },
		{ 355, 503, 58 },
		{ 355, 504, 58 },
		{ 355, 505, 58 },
		{ 355, 506, 58 },
		{ 355, 507, 58 },
		{ 355, 508, 58 },
		{ 355, 509, 58 },
		{ 355, 510, 58 },
		{ 355, 511, 58 },
		{ 355, 512, 58 },
		{ 356, 416, 58 },
		{ 356, 417, 58 },
		{ 356, 418, 58 },
		{ 356, 419, 58 },
		{ 356, 420, 58 },
		{ 356, 421, 58 },
		{ 356, 422, 58 },
		{ 356, 423, 58 },
		{ 356, 424, 58 },
		{ 356, 425, 58 },
		{ 356, 426, 58 },
		{ 356, 427, 58 },
		{ 356, 428, 58 },
		{ 356, 429, 58 },
		{ 356, 430, 58 },
		{ 356, 431, 58 },
		{ 356, 432, 58 },
		{ 356, 433, 58 },
		{ 356, 434, 58 },
		{ 356, 435, 58 },
		{ 356, 436, 58 },
		{ 356, 437, 58 },
		{ 356, 438, 58 },
		{ 356, 439, 58 },
		{ 356, 440, 58 },
		{ 356, 441, 58 },
		{ 356, 442, 58 },
		{ 356, 443, 58 },
		{ 356, 444, 58 },
		{ 356, 445, 58 },
		{ 356, 446, 58 },
		{ 356, 447, 58 },
		{ 356, 448, 58 },
		{ 356, 449, 58 },
		{ 356, 450, 58 },
		{ 356, 451, 58 },
		{ 356, 452, 58 },
		{ 356, 453, 58 },
		{ 356, 454, 58 },
		{ 356, 455, 58 },
		{ 356, 456, 58 },
		{ 356, 457, 58 },
		{ 356, 458, 58 },
		{ 356, 459, 58 },
		{ 356, 460, 58 },
		{ 356, 461, 58 },
		{ 356, 462, 58 },
		{ 356, 463, 58 },
		{ 356, 464, 58 },
		{ 356, 465, 58 },
		{ 356, 466, 58 },
		{ 356, 467, 58 },
		{ 356, 468, 58 },
		{ 356, 469, 58 },
		{ 356, 470, 58 },
		{ 356, 471, 58 },
		{ 356, 472, 58 },
		{ 356, 473, 58 },
		{ 356, 474, 58 },
		{ 356, 475, 58 },
		{ 356, 476, 58 },
		{ 356, 477, 58 },
		{ 356, 478, 58 },
		{ 356, 479, 58 },
		{ 356, 480, 58 },
		{ 356, 481, 58 },
		{ 356, 482, 58 },
		{ 356, 483, 58 },
		{ 356, 484, 58 },
		{ 356, 485, 58 },
		{ 356, 486, 58 },
		{ 356, 487, 58 },
		{ 356, 488, 58 },
		{ 356, 489, 58 },
		{ 356, 490, 58 },
		{ 356, 491, 58 },
		{ 356, 492, 58 },
		{ 356, 493, 58 },
		{ 356, 494, 58 },
		{ 356, 495, 58 },
		{ 356, 496, 58 },
		{ 356, 497, 58 },
		{ 356, 498, 58 },
		{ 356, 499, 58 },
		{ 356, 500, 58 },
		{ 356, 501, 58 },
		{ 356, 502, 58 },
		{ 356, 503, 58 },
		{ 356, 504, 58 },
		{ 356, 505, 58 },
		{ 356, 506, 58 },
		{ 356, 507, 58 },
		{ 356, 508, 58 },
		{ 356, 509, 58 },
		{ 356, 510, 58 },
		{ 356, 511, 58 },
		{ 356, 512, 58 },
		{ 357, 418, 58 },
		{ 357, 419, 58 },
		{ 357, 420, 58 },
		{ 357, 421, 58 },
		{ 357, 422, 58 },
		{ 357, 423, 58 },
		{ 357, 424, 58 },
		{ 357, 425, 58 },
		{ 357, 426, 58 },
		{ 357, 427, 58 },
		{ 357, 428, 58 },
		{ 357, 429, 58 },
		{ 357, 430, 58 },
		{ 357, 431, 58 },
		{ 357, 432, 58 },
		{ 357, 433, 58 },
		{ 357, 434, 58 },
		{ 357, 435, 58 },
		{ 357, 436, 58 },
		{ 357, 437, 58 },
		{ 357, 438, 58 },
		{ 357, 439, 58 },
		{ 357, 440, 58 },
		{ 357, 441, 58 },
		{ 357, 442, 58 },
		{ 357, 443, 58 },
		{ 357, 444, 58 },
		{ 357, 445, 58 },
		{ 357, 446, 58 },
		{ 357, 447, 58 },
		{ 357, 448, 58 },
		{ 357, 449, 58 },
		{ 357, 450, 58 },
		{ 357, 451, 58 },
		{ 357, 452, 58 },
		{ 357, 453, 58 },
		{ 357, 454, 58 },
		{ 357, 455, 58 },
		{ 357, 456, 58 },
		{ 357, 457, 58 },
		{ 357, 458, 58 },
		{ 357, 459, 58 },
		{ 357, 460, 58 },
		{ 357, 461, 58 },
		{ 357, 462, 58 },
		{ 357, 463, 58 },
		{ 357, 464, 58 },
		{ 357, 465, 58 },
		{ 357, 466, 58 },
		{ 357, 467, 58 },
		{ 357, 468, 58 },
		{ 357, 469, 58 },
		{ 357, 470, 58 },
		{ 357, 471, 58 },
		{ 357, 472, 58 },
		{ 357, 473, 58 },
		{ 357, 474, 58 },
		{ 357, 475, 58 },
		{ 357, 476, 58 },
		{ 357, 477, 58 },
		{ 357, 478, 58 },
		{ 357, 479, 58 },
		{ 357, 480, 58 },
		{ 357, 481, 58 },
		{ 357, 482, 58 },
		{ 357, 483, 58 },
		{ 357, 484, 58 },
		{ 357, 485, 58 },
		{ 357, 486, 58 },
		{ 357, 487, 58 },
		{ 357, 488, 58 },
		{ 357, 489, 58 },
		{ 357, 490, 58 },
		{ 357, 491, 58 },
		{ 357, 492, 58 },
		{ 357, 493, 58 },
		{ 357, 494, 58 },
		{ 357, 495, 58 },
		{ 357, 496, 58 },
		{ 357, 497, 58 },
		{ 357, 498, 58 },
		{ 357, 499, 58 },
		{ 357, 500, 58 },
		{ 357, 501, 58 },
		{ 357, 502, 58 },
		{ 357, 503, 58 },
		{ 357, 504, 58 },
		{ 357, 505, 58 },
		{ 357, 506, 58 },
		{ 357, 507, 58 },
		{ 357, 508, 58 },
		{ 357, 509, 58 },
		{ 357, 510, 58 },
		{ 357, 511, 58 },
		{ 357, 512, 58 },
		{ 358, 419, 58 },
		{ 358, 420, 58 },
		{ 358, 421, 58 },
		{ 358, 422, 58 },
		{ 358, 423, 58 },
		{ 358, 424, 58 },
		{ 358, 425, 58 },
		{ 358, 426, 58 },
		{ 358, 427, 58 },
		{ 358, 428, 58 },
		{ 358, 429, 58 },
		{ 358, 430, 58 },
		{ 358, 431, 58 },
		{ 358, 432, 58 },
		{ 358, 433, 58 },
		{ 358, 434, 58 },
		{ 358, 435, 58 },
		{ 358, 436, 58 },
		{ 358, 437, 58 },
		{ 358, 438, 58 },
		{ 358, 439, 58 },
		{ 358, 440, 58 },
		{ 358, 441, 58 },
		{ 358, 442, 58 },
		{ 358, 443, 58 },
		{ 358, 444, 58 },
		{ 358, 445, 58 },
		{ 358, 446, 58 },
		{ 358, 447, 58 },
		{ 358, 448, 58 },
		{ 358, 449, 58 },
		{ 358, 450, 58 },
		{ 358, 451, 58 },
		{ 358, 452, 58 },
		{ 358, 453, 58 },
		{ 358, 454, 58 },
		{ 358, 455, 58 },
		{ 358, 456, 58 },
		{ 358, 457, 58 },
		{ 358, 458, 58 },
		{ 358, 459, 58 },
		{ 358, 460, 58 },
		{ 358, 461, 58 },
		{ 358, 462, 58 },
		{ 358, 463, 58 },
		{ 358, 464, 58 },
		{ 358, 465, 58 },
		{ 358, 466, 58 },
		{ 358, 467, 58 },
		{ 358, 468, 58 },
		{ 358, 469, 58 },
		{ 358, 470, 58 },
		{ 358, 471, 58 },
		{ 358, 472, 58 },
		{ 358, 473, 58 },
		{ 358, 474, 58 },
		{ 358, 475, 58 },
		{ 358, 476, 58 },
		{ 358, 477, 58 },
		{ 358, 478, 58 },
		{ 358, 479, 58 },
		{ 358, 480, 58 },
		{ 358, 481, 58 },
		{ 358, 482, 58 },
		{ 358, 483, 58 },
		{ 358, 484, 58 },
		{ 358, 485, 58 },
		{ 358, 486, 58 },
		{ 358, 487, 58 },
		{ 358, 488, 58 },
		{ 358, 489, 58 },
		{ 358, 490, 58 },
		{ 358, 491, 58 },
		{ 358, 492, 58 },
		{ 358, 493, 58 },
		{ 358, 494, 58 },
		{ 358, 495, 58 },
		{ 358, 496, 58 },
		{ 358, 497, 58 },
		{ 358, 498, 58 },
		{ 358, 499, 58 },
		{ 358, 500, 58 },
		{ 358, 501, 58 },
		{ 358, 502, 58 },
		{ 358, 503, 58 },
		{ 358, 504, 58 },
		{ 358, 505, 58 },
		{ 358, 506, 58 },
		{ 358, 507, 58 },
		{ 358, 508, 58 },
		{ 358, 509, 58 },
		{ 358, 510, 58 },
		{ 358, 511, 58 },
		{ 358, 512, 58 },
		{ 359, 421, 58 },
		{ 359, 422, 58 },
		{ 359, 423, 58 },
		{ 359, 424, 58 },
		{ 359, 425, 58 },
		{ 359, 426, 58 },
		{ 359, 427, 58 },
		{ 359, 428, 58 },
		{ 359, 429, 58 },
		{ 359, 430, 58 },
		{ 359, 431, 58 },
		{ 359, 432, 58 },
		{ 359, 433, 58 },
		{ 359, 434, 58 },
		{ 359, 435, 58 },
		{ 359, 436, 58 },
		{ 359, 437, 58 },
		{ 359, 438, 58 },
		{ 359, 439, 58 },
		{ 359, 440, 58 },
		{ 359, 441, 58 },
		{ 359, 442, 58 },
		{ 359, 443, 58 },
		{ 359, 444, 58 },
		{ 359, 445, 58 },
		{ 359, 446, 58 },
		{ 359, 447, 58 },
		{ 359, 448, 58 },
		{ 359, 449, 58 },
		{ 359, 450, 58 },
		{ 359, 451, 58 },
		{ 359, 452, 58 },
		{ 359, 453, 58 },
		{ 359, 454, 58 },
		{ 359, 455, 58 },
		{ 359, 456, 58 },
		{ 359, 457, 58 },
		{ 359, 458, 58 },
		{ 359, 459, 58 },
		{ 359, 460, 58 },
		{ 359, 461, 58 },
		{ 359, 462, 58 },
		{ 359, 463, 58 },
		{ 359, 464, 58 },
		{ 359, 465, 58 },
		{ 359, 466, 58 },
		{ 359, 467, 58 },
		{ 359, 468, 58 },
		{ 359, 469, 58 },
		{ 359, 470, 58 },
		{ 359, 471, 58 },
		{ 359, 472, 58 },
		{ 359, 473, 58 },
		{ 359, 474, 58 },
		{ 359, 475, 58 },
		{ 359, 476, 58 },
		{ 359, 477, 58 },
		{ 359, 478, 58 },
		{ 359, 479, 58 },
		{ 359, 480, 58 },
		{ 359, 481, 58 },
		{ 359, 482, 58 },
		{ 359, 483, 58 },
		{ 359, 484, 58 },
		{ 359, 485, 58 },
		{ 359, 486, 58 },
		{ 359, 487, 58 },
		{ 359, 488, 58 },
		{ 359, 489, 58 },
		{ 359, 490, 58 },
		{ 359, 491, 58 },
		{ 359, 492, 58 },
		{ 359, 493, 58 },
		{ 359, 494, 58 },
		{ 359, 495, 58 },
		{ 359, 496, 58 },
		{ 359, 497, 58 },
		{ 359, 498, 58 },
		{ 359, 499, 58 },
		{ 359, 500, 58 },
		{ 359, 501, 58 },
		{ 359, 502, 58 },
		{ 359, 503, 58 },
		{ 359, 504, 58 },
		{ 359, 505, 58 },
		{ 359, 506, 58 },
		{ 359, 507, 58 },
		{ 359, 508, 58 },
		{ 359, 509, 58 },
		{ 359, 510, 58 },
		{ 359, 511, 58 },
		{ 359, 512, 58 },
		{ 360, 422, 58 },
		{ 360, 423, 58 },
		{ 360, 424, 58 },
		{ 360, 425, 58 },
		{ 360, 426, 58 },
		{ 360, 427, 58 },
		{ 360, 428, 58 },
		{ 360, 429, 58 },
		{ 360, 430, 58 },
		{ 360, 431, 58 },
		{ 360, 432, 58 },
		{ 360, 433, 58 },
		{ 360, 434, 58 },
		{ 360, 435, 58 },
		{ 360, 436, 58 },
		{ 360, 437, 58 },
		{ 360, 438, 58 },
		{ 360, 439, 58 },
		{ 360, 440, 58 },
		{ 360, 441, 58 },
		{ 360, 442, 58 },
		{ 360, 443, 58 },
		{ 360, 444, 58 },
		{ 360, 445, 58 },
		{ 360, 446, 58 },
		{ 360, 447, 58 },
		{ 360, 448, 58 },
		{ 360, 449, 58 },
		{ 360, 450, 58 },
		{ 360, 451, 58 },
		{ 360, 452, 58 },
		{ 360, 453, 58 },
		{ 360, 454, 58 },
		{ 360, 455, 58 },
		{ 360, 456, 58 },
		{ 360, 457, 58 },
		{ 360, 458, 58 },
		{ 360, 459, 58 },
		{ 360, 460, 58 },
		{ 360, 461, 58 },
		{ 360, 462, 58 },
		{ 360, 463, 58 },
		{ 360, 464, 58 },
		{ 360, 465, 58 },
		{ 360, 466, 58 },
		{ 360, 467, 58 },
		{ 360, 468, 58 },
		{ 360, 469, 58 },
		{ 360, 470, 58 },
		{ 360, 471, 58 },
		{ 360, 472, 58 },
		{ 360, 473, 58 },
		{ 360, 474, 58 },
		{ 360, 475, 58 },
		{ 360, 476, 58 },
		{ 360, 477, 58 },
		{ 360, 478, 58 },
		{ 360, 479, 58 },
		{ 360, 480, 58 },
		{ 360, 481, 58 },
		{ 360, 482, 58 },
		{ 360, 483, 58 },
		{ 360, 484, 58 },
		{ 360, 485, 58 },
		{ 360, 486, 58 },
		{ 360, 487, 58 },
		{ 360, 488, 58 },
		{ 360, 489, 58 },
		{ 360, 490, 58 },
		{ 360, 491, 58 },
		{ 360, 492, 58 },
		{ 360, 493, 58 },
		{ 360, 494, 58 },
		{ 360, 495, 58 },
		{ 360, 496, 58 },
		{ 360, 497, 58 },
		{ 360, 498, 58 },
		{ 360, 499, 58 },
		{ 360, 500, 58 },
		{ 360, 501, 58 },
		{ 360, 502, 58 },
		{ 360, 503, 58 },
		{ 360, 504, 58 },
		{ 360, 505, 58 },
		{ 360, 506, 58 },
		{ 360, 507, 58 },
		{ 360, 508, 58 },
		{ 360, 509, 58 },
		{ 360, 510, 58 },
		{ 360, 511, 58 },
		{ 360, 512, 58 },
		{ 361, 424, 58 },
		{ 361, 425, 58 },
		{ 361, 426, 58 },
		{ 361, 427, 58 },
		{ 361, 428, 58 },
		{ 361, 429, 58 },
		{ 361, 430, 58 },
		{ 361, 431, 58 },
		{ 361, 432, 58 },
		{ 361, 433, 58 },
		{ 361, 434, 58 },
		{ 361, 435, 58 },
		{ 361, 436, 58 },
		{ 361, 437, 58 },
		{ 361, 438, 58 },
		{ 361, 439, 58 },
		{ 361, 440, 58 },
		{ 361, 441, 58 },
		{ 361, 442, 58 },
		{ 361, 443, 58 },
		{ 361, 444, 58 },
		{ 361, 445, 58 },
		{ 361, 446, 58 },
		{ 361, 447, 58 },
		{ 361, 448, 58 },
		{ 361, 449, 58 },
		{ 361, 450, 58 },
		{ 361, 451, 58 },
		{ 361, 452, 58 },
		{ 361, 453, 58 },
		{ 361, 454, 58 },
		{ 361, 455, 58 },
		{ 361, 456, 58 },
		{ 361, 457, 58 },
		{ 361, 458, 58 },
		{ 361, 459, 58 },
		{ 361, 460, 58 },
		{ 361, 461, 58 },
		{ 361, 462, 58 },
		{ 361, 463, 58 },
		{ 361, 464, 58 },
		{ 361, 465, 58 },
		{ 361, 466, 58 },
		{ 361, 467, 58 },
		{ 361, 468, 58 },
		{ 361, 469, 58 },
		{ 361, 470, 58 },
		{ 361, 471, 58 },
		{ 361, 472, 58 },
		{ 361, 473, 58 },
		{ 361, 474, 58 },
		{ 361, 475, 58 },
		{ 361, 476, 58 },
		{ 361, 477, 58 },
		{ 361, 478, 58 },
		{ 361, 479, 58 },
		{ 361, 480, 58 },
		{ 361, 481, 58 },
		{ 361, 482, 58 },
		{ 361, 483, 58 },
		{ 361, 484, 58 },
		{ 361, 485, 58 },
		{ 361, 486, 58 },
		{ 361, 487, 58 },
		{ 361, 488, 58 },
		{ 361, 489, 58 },
		{ 361, 490, 58 },
		{ 361, 491, 58 },
		{ 361, 492, 58 },
		{ 361, 493, 58 },
		{ 361, 494, 58 },
		{ 361, 495, 58 },
		{ 361, 496, 58 },
		{ 361, 497, 58 },
		{ 361, 498, 58 },
		{ 361, 499, 58 },
		{ 361, 500, 58 },
		{ 361, 501, 58 },
		{ 361, 502, 58 },
		{ 361, 503, 58 },
		{ 361, 504, 58 },
		{ 361, 505, 58 },
		{ 361, 506, 58 },
		{ 361, 507, 58 },
		{ 361, 508, 58 },
		{ 361, 509, 58 },
		{ 361, 510, 58 },
		{ 361, 511, 58 },
		{ 361, 512, 58 },
		{ 362, 426, 58 },
		{ 362, 427, 58 },
		{ 362, 428, 58 },
		{ 362, 429, 58 },
		{ 362, 430, 58 },
		{ 362, 431, 58 },
		{ 362, 432, 58 },
		{ 362, 433, 58 },
		{ 362, 434, 58 },
		{ 362, 435, 58 },
		{ 362, 436, 58 },
		{ 362, 437, 58 },
		{ 362, 438, 58 },
		{ 362, 439, 58 },
		{ 362, 440, 58 },
		{ 362, 441, 58 },
		{ 362, 442, 58 },
		{ 362, 443, 58 },
		{ 362, 444, 58 },
		{ 362, 445, 58 },
		{ 362, 446, 58 },
		{ 362, 447, 58 },
		{ 362, 448, 58 },
		{ 362, 449, 58 },
		{ 362, 450, 58 },
		{ 362, 451, 58 },
		{ 362, 452, 58 },
		{ 362, 453, 58 },
		{ 362, 454, 58 },
		{ 362, 455, 58 },
		{ 362, 456, 58 },
		{ 362, 457, 58 },
		{ 362, 458, 58 },
		{ 362, 459, 58 },
		{ 362, 460, 58 },
		{ 362, 461, 58 },
		{ 362, 462, 58 },
		{ 362, 463, 58 },
		{ 362, 464, 58 },
		{ 362, 465, 58 },
		{ 362, 466, 58 },
		{ 362, 467, 58 },
		{ 362, 468, 58 },
		{ 362, 469, 58 },
		{ 362, 470, 58 },
		{ 362, 471, 58 },
		{ 362, 472, 58 },
		{ 362, 473, 58 },
		{ 362, 474, 58 },
		{ 362, 475, 58 },
		{ 362, 476, 58 },
		{ 362, 477, 58 },
		{ 362, 478, 58 },
		{ 362, 479, 58 },
		{ 362, 480, 58 },
		{ 362, 481, 58 },
		{ 362, 482, 58 },
		{ 362, 483, 58 },
		{ 362, 484, 58 },
		{ 362, 485, 58 },
		{ 362, 486, 58 },
		{ 362, 487, 58 },
		{ 362, 488, 58 },
		{ 362, 489, 58 },
		{ 362, 490, 58 },
		{ 362, 491, 58 },
		{ 362, 492, 58 },
		{ 362, 493, 58 },
		{ 362, 494, 58 },
		{ 362, 495, 58 },
		{ 362, 496, 58 },
		{ 362, 497, 58 },
		{ 362, 498, 58 },
		{ 362, 499, 58 },
		{ 362, 500, 58 },
		{ 362, 501, 58 },
		{ 362, 502, 58 },
		{ 362, 503, 58 },
		{ 362, 504, 58 },
		{ 362, 505, 58 },
		{ 362, 506, 58 },
		{ 362, 507, 58 },
		{ 362, 508, 58 },
		{ 362, 509, 58 },
		{ 362, 510, 58 },
		{ 362, 511, 58 },
		{ 362, 512, 58 },
		{ 363, 427, 58 },
		{ 363, 428, 58 },
		{ 363, 429, 58 },
		{ 363, 430, 58 },
		{ 363, 431, 58 },
		{ 363, 432, 58 },
		{ 363, 433, 58 },
		{ 363, 434, 58 },
		{ 363, 435, 58 },
		{ 363, 436, 58 },
		{ 363, 437, 58 },
		{ 363, 438, 58 },
		{ 363, 439, 58 },
		{ 363, 440, 58 },
		{ 363, 441, 58 },
		{ 363, 442, 58 },
		{ 363, 443, 58 },
		{ 363, 444, 58 },
		{ 363, 445, 58 },
		{ 363, 446, 58 },
		{ 363, 447, 58 },
		{ 363, 448, 58 },
		{ 363, 449, 58 },
		{ 363, 450, 58 },
		{ 363, 451, 58 },
		{ 363, 452, 58 },
		{ 363, 453, 58 },
		{ 363, 454, 58 },
		{ 363, 455, 58 },
		{ 363, 456, 58 },
		{ 363, 457, 58 },
		{ 363, 458, 58 },
		{ 363, 459, 58 },
		{ 363, 460, 58 },
		{ 363, 461, 58 },
		{ 363, 462, 58 },
		{ 363, 463, 58 },
		{ 363, 464, 58 },
		{ 363, 465, 58 },
		{ 363, 466, 58 },
		{ 363, 467, 58 },
		{ 363, 468, 58 },
		{ 363, 469, 58 },
		{ 363, 470, 58 },
		{ 363, 471, 58 },
		{ 363, 472, 58 },
		{ 363, 473, 58 },
		{ 363, 474, 58 },
		{ 363, 475, 58 },
		{ 363, 476, 58 },
		{ 363, 477, 58 },
		{ 363, 478, 58 },
		{ 363, 479, 58 },
		{ 363, 480, 58 },
		{ 363, 481, 58 },
		{ 363, 482, 58 },
		{ 363, 483, 58 },
		{ 363, 484, 58 },
		{ 363, 485, 58 },
		{ 363, 486, 58 },
		{ 363, 487, 58 },
		{ 363, 488, 58 },
		{ 363, 489, 58 },
		{ 363, 490, 58 },
		{ 363, 491, 58 },
		{ 363, 492, 58 },
		{ 363, 493, 58 },
		{ 363, 494, 58 },
		{ 363, 495, 58 },
		{ 363, 496, 58 },
		{ 363, 497, 58 },
		{ 363, 498, 58 },
		{ 363, 499, 58 },
		{ 363, 500, 58 },
		{ 363, 501, 58 },
		{ 363, 502, 58 },
		{ 363, 503, 58 },
		{ 363, 504, 58 },
		{ 363, 505, 58 },
		{ 363, 506, 58 },
		{ 363, 507, 58 },
		{ 363, 508, 58 },
		{ 363, 509, 58 },
		{ 363, 510, 58 },
		{ 363, 511, 58 },
		{ 363, 512, 58 },
		{ 364, 429, 58 },
		{ 364, 430, 58 },
		{ 364, 431, 58 },
		{ 364, 432, 58 },
		{ 364, 433, 58 },
		{ 364, 434, 58 },
		{ 364, 435, 58 },
		{ 364, 436, 58 },
		{ 364, 437, 58 },
		{ 364, 438, 58 },
		{ 364, 439, 58 },
		{ 364, 440, 58 },
		{ 364, 441, 58 },
		{ 364, 442, 58 },
		{ 364, 443, 58 },
		{ 364, 444, 58 },
		{ 364, 445, 58 },
		{ 364, 446, 58 },
		{ 364, 447, 58 },
		{ 364, 448, 58 },
		{ 364, 449, 58 },
		{ 364, 450, 58 },
		{ 364, 451, 58 },
		{ 364, 452, 58 },
		{ 364, 453, 58 },
		{ 364, 454, 58 },
		{ 364, 455, 58 },
		{ 364, 456, 58 },
		{ 364, 457, 58 },
		{ 364, 458, 58 },
		{ 364, 459, 58 },
		{ 364, 460, 58 },
		{ 364, 461, 58 },
		{ 364, 462, 58 },
		{ 364, 463, 58 },
		{ 364, 464, 58 },
		{ 364, 465, 58 },
		{ 364, 466, 58 },
		{ 364, 467, 58 },
		{ 364, 468, 58 },
		{ 364, 469, 58 },
		{ 364, 470, 58 },
		{ 364, 471, 58 },
		{ 364, 472, 58 },
		{ 364, 473, 58 },
		{ 364, 474, 58 },
		{ 364, 475, 58 },
		{ 364, 476, 58 },
		{ 364, 477, 58 },
		{ 364, 478, 58 },
		{ 364, 479, 58 },
		{ 364, 480, 58 },
		{ 364, 481, 58 },
		{ 364, 482, 58 },
		{ 364, 483, 58 },
		{ 364, 484, 58 },
		{ 364, 485, 58 },
		{ 364, 486, 58 },
		{ 364, 487, 58 },
		{ 364, 488, 58 },
		{ 364, 489, 58 },
		{ 364, 490, 58 },
		{ 364, 491, 58 },
		{ 364, 492, 58 },
		{ 364, 493, 58 },
		{ 364, 494, 58 },
		{ 364, 495, 58 },
		{ 364, 496, 58 },
		{ 364, 497, 58 },
		{ 364, 498, 58 },
		{ 364, 499, 58 },
		{ 364, 500, 58 },
		{ 364, 501, 58 },
		{ 364, 502, 58 },
		{ 364, 503, 58 },
		{ 364, 504, 58 },
		{ 364, 505, 58 },
		{ 364, 506, 58 },
		{ 364, 507, 58 },
		{ 364, 508, 58 },
		{ 364, 509, 58 },
		{ 364, 510, 58 },
		{ 364, 511, 58 },
		{ 364, 512, 58 },
		{ 365, 430, 58 },
		{ 365, 431, 58 },
		{ 365, 432, 58 },
		{ 365, 433, 58 },
		{ 365, 434, 58 },
		{ 365, 435, 58 },
		{ 365, 436, 58 },
		{ 365, 437, 58 },
		{ 365, 438, 58 },
		{ 365, 439, 58 },
		{ 365, 440, 58 },
		{ 365, 441, 58 },
		{ 365, 442, 58 },
		{ 365, 443, 58 },
		{ 365, 444, 58 },
		{ 365, 445, 58 },
		{ 365, 446, 58 },
		{ 365, 447, 58 },
		{ 365, 448, 58 },
		{ 365, 449, 58 },
		{ 365, 450, 58 },
		{ 365, 451, 58 },
		{ 365, 452, 58 },
		{ 365, 453, 58 },
		{ 365, 454, 58 },
		{ 365, 455, 58 },
		{ 365, 456, 58 },
		{ 365, 457, 58 },
		{ 365, 458, 58 },
		{ 365, 459, 58 },
		{ 365, 460, 58 },
		{ 365, 461, 58 },
		{ 365, 462, 58 },
		{ 365, 463, 58 },
		{ 365, 464, 58 },
		{ 365, 465, 58 },
		{ 365, 466, 58 },
		{ 365, 467, 58 },
		{ 365, 468, 58 },
		{ 365, 469, 58 },
		{ 365, 470, 58 },
		{ 365, 471, 58 },
		{ 365, 472, 58 },
		{ 365, 473, 58 },
		{ 365, 474, 58 },
		{ 365, 475, 58 },
		{ 365, 476, 58 },
		{ 365, 477, 58 },
		{ 365, 478, 58 },
		{ 365, 479, 58 },
		{ 365, 480, 58 },
		{ 365, 481, 58 },
		{ 365, 482, 58 },
		{ 365, 483, 58 },
		{ 365, 484, 58 },
		{ 365, 485, 58 },
		{ 365, 486, 58 },
		{ 365, 487, 58 },
		{ 365, 488, 58 },
		{ 365, 489, 58 },
		{ 365, 490, 58 },
		{ 365, 491, 58 },
		{ 365, 492, 58 },
		{ 365, 493, 58 },
		{ 365, 494, 58 },
		{ 365, 495, 58 },
		{ 365, 496, 58 },
		{ 365, 497, 58 },
		{ 365, 498, 58 },
		{ 365, 499, 58 },
		{ 365, 500, 58 },
		{ 365, 501, 58 },
		{ 365, 502, 58 },
		{ 365, 503, 58 },
		{ 365, 504, 58 },
		{ 365, 505, 58 },
		{ 365, 506, 58 },
		{ 365, 507, 58 },
		{ 365, 508, 58 },
		{ 365, 509, 58 },
		{ 365, 510, 58 },
		{ 365, 511, 58 },
		{ 365, 512, 58 },
		{ 366, 432, 58 },
		{ 366, 433, 58 },
		{ 366, 434, 58 },
		{ 366, 435, 58 },
		{ 366, 436, 58 },
		{ 366, 437, 58 },
		{ 366, 438, 58 },
		{ 366, 439, 58 },
		{ 366, 440, 58 },
		{ 366, 441, 58 },
		{ 366, 442, 58 },
		{ 366, 443, 58 },
		{ 366, 444, 58 },
		{ 366, 445, 58 },
		{ 366, 446, 58 },
		{ 366, 447, 58 },
		{ 366, 448, 58 },
		{ 366, 449, 58 },
		{ 366, 450, 58 },
		{ 366, 451, 58 },
		{ 366, 452, 58 },
		{ 366, 453, 58 },
		{ 366, 454, 58 },
		{ 366, 455, 58 },
		{ 366, 456, 58 },
		{ 366, 457, 58 },
		{ 366, 458, 58 },
		{ 366, 459, 58 },
		{ 366, 460, 58 },
		{ 366, 461, 58 },
		{ 366, 462, 58 },
		{ 366, 463, 58 },
		{ 366, 464, 58 },
		{ 366, 465, 58 },
		{ 366, 466, 58 },
		{ 366, 467, 58 },
		{ 366, 468, 58 },
		{ 366, 469, 58 },
		{ 366, 470, 58 },
		{ 366, 471, 58 },
		{ 366, 472, 58 },
		{ 366, 473, 58 },
		{ 366, 474, 58 },
		{ 366, 475, 58 },
		{ 366, 476, 58 },
		{ 366, 477, 58 },
		{ 366, 478, 58 },
		{ 366, 479, 58 },
		{ 366, 480, 58 },
		{ 366, 481, 58 },
		{ 366, 482, 58 },
		{ 366, 483, 58 },
		{ 366, 484, 58 },
		{ 366, 485, 58 },
		{ 366, 486, 58 },
		{ 366, 487, 58 },
		{ 366, 488, 58 },
		{ 366, 489, 58 },
		{ 366, 490, 58 },
		{ 366, 491, 58 },
		{ 366, 492, 58 },
		{ 366, 493, 58 },
		{ 366, 494, 58 },
		{ 366, 495, 58 },
		{ 366, 496, 58 },
		{ 366, 497, 58 },
		{ 366, 498, 58 },
		{ 366, 499, 58 },
		{ 366, 500, 58 },
		{ 366, 501, 58 },
		{ 366, 502, 58 },
		{ 366, 503, 58 },
		{ 366, 504, 58 },
		{ 366, 505, 58 },
		{ 366, 506, 58 },
		{ 366, 507, 58 },
		{ 366, 508, 58 },
		{ 366, 509, 58 },
		{ 366, 510, 58 },
		{ 366, 511, 58 },
		{ 366, 512, 58 },
		{ 367, 434, 58 },
		{ 367, 435, 58 },
		{ 367, 436, 58 },
		{ 367, 437, 58 },
		{ 367, 438, 58 },
		{ 367, 439, 58 },
		{ 367, 440, 58 },
		{ 367, 441, 58 },
		{ 367, 442, 58 },
		{ 367, 443, 58 },
		{ 367, 444, 58 },
		{ 367, 445, 58 },
		{ 367, 446, 58 },
		{ 367, 447, 58 },
		{ 367, 448, 58 },
		{ 367, 449, 58 },
		{ 367, 450, 58 },
		{ 367, 451, 58 },
		{ 367, 452, 58 },
		{ 367, 453, 58 },
		{ 367, 454, 58 },
		{ 367, 455, 58 },
		{ 367, 456, 58 },
		{ 367, 457, 58 },
		{ 367, 458, 58 },
		{ 367, 459, 58 },
		{ 367, 460, 58 },
		{ 367, 461, 58 },
		{ 367, 462, 58 },
		{ 367, 463, 58 },
		{ 367, 464, 58 },
		{ 367, 465, 58 },
		{ 367, 466, 58 },
		{ 367, 467, 58 },
		{ 367, 468, 58 },
		{ 367, 469, 58 },
		{ 367, 470, 58 },
		{ 367, 471, 58 },
		{ 367, 472, 58 },
		{ 367, 473, 58 },
		{ 367, 474, 58 },
		{ 367, 475, 58 },
		{ 367, 476, 58 },
		{ 367, 477, 58 },
		{ 367, 478, 58 },
		{ 367, 479, 58 },
		{ 367, 480, 58 },
		{ 367, 481, 58 },
		{ 367, 482, 58 },
		{ 367, 483, 58 },
		{ 367, 484, 58 },
		{ 367, 485, 58 },
		{ 367, 486, 58 },
		{ 367, 487, 58 },
		{ 367, 488, 58 },
		{ 367, 489, 58 },
		{ 367, 490, 58 },
		{ 367, 491, 58 },
		{ 367, 492, 58 },
		{ 367, 493, 58 },
		{ 367, 494, 58 },
		{ 367, 495, 58 },
		{ 367, 496, 58 },
		{ 367, 497, 58 },
		{ 367, 498, 58 },
		{ 367, 499, 58 },
		{ 367, 500, 58 },
		{ 367, 501, 58 },
		{ 367, 502, 58 },
		{ 367, 503, 58 },
		{ 367, 504, 58 },
		{ 367, 505, 58 },
		{ 367, 506, 58 },
		{ 367, 507, 58 },
		{ 367, 508, 58 },
		{ 367, 509, 58 },
		{ 367, 510, 58 },
		{ 367, 511, 58 },
		{ 367, 512, 58 },
		{ 368, 435, 58 },
		{ 368, 436, 58 },
		{ 368, 437, 58 },
		{ 368, 438, 58 },
		{ 368, 439, 58 },
		{ 368, 440, 58 },
		{ 368, 441, 58 },
		{ 368, 442, 58 },
		{ 368, 443, 58 },
		{ 368, 444, 58 },
		{ 368, 445, 58 },
		{ 368, 446, 58 },
		{ 368, 447, 58 },
		{ 368, 448, 58 },
		{ 368, 449, 58 },
		{ 368, 450, 58 },
		{ 368, 451, 58 },
		{ 368, 452, 58 },
		{ 368, 453, 58 },
		{ 368, 454, 58 },
		{ 368, 455, 58 },
		{ 368, 456, 58 },
		{ 368, 457, 58 },
		{ 368, 458, 58 },
		{ 368, 459, 58 },
		{ 368, 460, 58 },
		{ 368, 461, 58 },
		{ 368, 462, 58 },
		{ 368, 463, 58 },
		{ 368, 464, 58 },
		{ 368, 465, 58 },
		{ 368, 466, 58 },
		{ 368, 467, 58 },
		{ 368, 468, 58 },
		{ 368, 469, 58 },
		{ 368, 470, 58 },
		{ 368, 471, 58 },
		{ 368, 472, 58 },
		{ 368, 473, 58 },
		{ 368, 474, 58 },
		{ 368, 475, 58 },
		{ 368, 476, 58 },
		{ 368, 477, 58 },
		{ 368, 478, 58 },
		{ 368, 479, 58 },
		{ 368, 480, 58 },
		{ 368, 481, 58 },
		{ 368, 482, 58 },
		{ 368, 483, 58 },
		{ 368, 484, 58 },
		{ 368, 485, 58 },
		{ 368, 486, 58 },
		{ 368, 487, 58 },
		{ 368, 488, 58 },
		{ 368, 489, 58 },
		{ 368, 490, 58 },
		{ 368, 491, 58 },
		{ 368, 492, 58 },
		{ 368, 493, 58 },
		{ 368, 494, 58 },
		{ 368, 495, 58 },
		{ 368, 496, 58 },
		{ 368, 497, 58 },
		{ 368, 498, 58 },
		{ 368, 499, 58 },
		{ 368, 500, 58 },
		{ 368, 501, 58 },
		{ 368, 502, 58 },
		{ 368, 503, 58 },
		{ 368, 504, 58 },
		{ 368, 505, 58 },
		{ 368, 506, 58 },
		{ 368, 507, 58 },
		{ 368, 508, 58 },
		{ 368, 509, 58 },
		{ 368, 510, 58 },
		{ 368, 511, 58 },
		{ 368, 512, 58 },
		{ 369, 437, 58 },
		{ 369, 438, 58 },
		{ 369, 439, 58 },
		{ 369, 440, 58 },
		{ 369, 441, 58 },
		{ 369, 442, 58 },
		{ 369, 443, 58 },
		{ 369, 444, 58 },
		{ 369, 445, 58 },
		{ 369, 446, 58 },
		{ 369, 447, 58 },
		{ 369, 448, 58 },
		{ 369, 449, 58 },
		{ 369, 450, 58 },
		{ 369, 451, 58 },
		{ 369, 452, 58 },
		{ 369, 453, 58 },
		{ 369, 454, 58 },
		{ 369, 455, 58 },
		{ 369, 456, 58 },
		{ 369, 457, 58 },
		{ 369, 458, 58 },
		{ 369, 459, 58 },
		{ 369, 460, 58 },
		{ 369, 461, 58 },
		{ 369, 462, 58 },
		{ 369, 463, 58 },
		{ 369, 464, 58 },
		{ 369, 465, 58 },
		{ 369, 466, 58 },
		{ 369, 467, 58 },
		{ 369, 468, 58 },
		{ 369, 469, 58 },
		{ 369, 470, 58 },
		{ 369, 471, 58 },
		{ 369, 472, 58 },
		{ 369, 473, 58 },
		{ 369, 474, 58 },
		{ 369, 475, 58 },
		{ 369, 476, 58 },
		{ 369, 477, 58 },
		{ 369, 478, 58 },
		{ 369, 479, 58 },
		{ 369, 480, 58 },
		{ 369, 481, 58 },
		{ 369, 482, 58 },
		{ 369, 483, 58 },
		{ 369, 484, 58 },
		{ 369, 485, 58 },
		{ 369, 486, 58 },
		{ 369, 487, 58 },
		{ 369, 488, 58 },
		{ 369, 489, 58 },
		{ 369, 490, 58 },
		{ 369, 491, 58 },
		{ 369, 492, 58 },
		{ 369, 493, 58 },
		{ 369, 494, 58 },
		{ 369, 495, 58 },
		{ 369, 496, 58 },
		{ 369, 497, 58 },
		{ 369, 498, 58 },
		{ 369, 499, 58 },
		{ 369, 500, 58 },
		{ 369, 501, 58 },
		{ 369, 502, 58 },
		{ 369, 503, 58 },
		{ 369, 504, 58 },
		{ 369, 505, 58 },
		{ 369, 506, 58 },
		{ 369, 507, 58 },
		{ 369, 508, 58 },
		{ 369, 509, 58 },
		{ 369, 510, 58 },
		{ 369, 511, 58 },
		{ 369, 512, 58 },
		{ 370, 438, 58 },
		{ 370, 439, 58 },
		{ 370, 440, 58 },
		{ 370, 441, 58 },
		{ 370, 442, 58 },
		{ 370, 443, 58 },
		{ 370, 444, 58 },
		{ 370, 445, 58 },
		{ 370, 446, 58 },
		{ 370, 447, 58 },
		{ 370, 448, 58 },
		{ 370, 449, 58 },
		{ 370, 450, 58 },
		{ 370, 451, 58 },
		{ 370, 452, 58 },
		{ 370, 453, 58 },
		{ 370, 454, 58 },
		{ 370, 455, 58 },
		{ 370, 456, 58 },
		{ 370, 457, 58 },
		{ 370, 458, 58 },
		{ 370, 459, 58 },
		{ 370, 460, 58 },
		{ 370, 461, 58 },
		{ 370, 462, 58 },
		{ 370, 463, 58 },
		{ 370, 464, 58 },
		{ 370, 465, 58 },
		{ 370, 466, 58 },
		{ 370, 467, 58 },
		{ 370, 468, 58 },
		{ 370, 469, 58 },
		{ 370, 470, 58 },
		{ 370, 471, 58 },
		{ 370, 472, 58 },
		{ 370, 473, 58 },
		{ 370, 474, 58 },
		{ 370, 475, 58 },
		{ 370, 476, 58 },
		{ 370, 477, 58 },
		{ 370, 478, 58 },
		{ 370, 479, 58 },
		{ 370, 480, 58 },
		{ 370, 481, 58 },
		{ 370, 482, 58 },
		{ 370, 483, 58 },
		{ 370, 484, 58 },
		{ 370, 485, 58 },
		{ 370, 486, 58 },
		{ 370, 487, 58 },
		{ 370, 488, 58 },
		{ 370, 489, 58 },
		{ 370, 490, 58 },
		{ 370, 491, 58 },
		{ 370, 492, 58 },
		{ 370, 493, 58 },
		{ 370, 494, 58 },
		{ 370, 495, 58 },
		{ 370, 496, 58 },
		{ 370, 497, 58 },
		{ 370, 498, 58 },
		{ 370, 499, 58 },
		{ 370, 500, 58 },
		{ 370, 501, 58 },
		{ 370, 502, 58 },
		{ 370, 503, 58 },
		{ 370, 504, 58 },
		{ 370, 505, 58 },
		{ 370, 506, 58 },
		{ 370, 507, 58 },
		{ 370, 508, 58 },
		{ 370, 509, 58 },
		{ 370, 510, 58 },
		{ 370, 511, 58 },
		{ 370, 512, 58 },
		{ 371, 440, 58 },
		{ 371, 441, 58 },
		{ 371, 442, 58 },
		{ 371, 443, 58 },
		{ 371, 444, 58 },
		{ 371, 445, 58 },
		{ 371, 446, 58 },
		{ 371, 447, 58 },
		{ 371, 448, 58 },
		{ 371, 449, 58 },
		{ 371, 450, 58 },
		{ 371, 451, 58 },
		{ 371, 452, 58 },
		{ 371, 453, 58 },
		{ 371, 454, 58 },
		{ 371, 455, 58 },
		{ 371, 456, 58 },
		{ 371, 457, 58 },
		{ 371, 458, 58 },
		{ 371, 459, 58 },
		{ 371, 460, 58 },
		{ 371, 461, 58 },
		{ 371, 462, 58 },
		{ 371, 463, 58 },
		{ 371, 464, 58 },
		{ 371, 465, 58 },
		{ 371, 466, 58 },
		{ 371, 467, 58 },
		{ 371, 468, 58 },
		{ 371, 469, 58 },
		{ 371, 470, 58 },
		{ 371, 471, 58 },
		{ 371, 472, 58 },
		{ 371, 473, 58 },
		{ 371, 474, 58 },
		{ 371, 475, 58 },
		{ 371, 476, 58 },
		{ 371, 477, 58 },
		{ 371, 478, 58 },
		{ 371, 479, 58 },
		{ 371, 480, 58 },
		{ 371, 481, 58 },
		{ 371, 482, 58 },
		{ 371, 483, 58 },
		{ 371, 484, 58 },
		{ 371, 485, 58 },
		{ 371, 486, 58 },
		{ 371, 487, 58 },
		{ 371, 488, 58 },
		{ 371, 489, 58 },
		{ 371, 490, 58 },
		{ 371, 491, 58 },
		{ 371, 492, 58 },
		{ 371, 493, 58 },
		{ 371, 494, 58 },
		{ 371, 495, 58 },
		{ 371, 496, 58 },
		{ 371, 497, 58 },
		{ 371, 498, 58 },
		{ 371, 499, 58 },
		{ 371, 500, 58 },
		{ 371, 501, 58 },
		{ 371, 502, 58 },
		{ 371, 503, 58 },
		{ 371, 504, 58 },
		{ 371, 505, 58 },
		{ 371, 506, 58 },
		{ 371, 507, 58 },
		{ 371, 508, 58 },
		{ 371, 509, 58 },
		{ 371, 510, 58 },
		{ 371, 511, 58 },
		{ 371, 512, 58 },
		{ 372, 442, 58 },
		{ 372, 443, 58 },
		{ 372, 444, 58 },
		{ 372, 445, 58 },
		{ 372, 446, 58 },
		{ 372, 447, 58 },
		{ 372, 448, 58 },
		{ 372, 449, 58 },
		{ 372, 450, 58 },
		{ 372, 451, 58 },
		{ 372, 452, 58 },
		{ 372, 453, 58 },
		{ 372, 454, 58 },
		{ 372, 455, 58 },
		{ 372, 456, 58 },
		{ 372, 457, 58 },
		{ 372, 458, 58 },
		{ 372, 459, 58 },
		{ 372, 460, 58 },
		{ 372, 461, 58 },
		{ 372, 462, 58 },
		{ 372, 463, 58 },
		{ 372, 464, 58 },
		{ 372, 465, 58 },
		{ 372, 466, 58 },
		{ 372, 467, 58 },
		{ 372, 468, 58 },
		{ 372, 469, 58 },
		{ 372, 470, 58 },
		{ 372, 471, 58 },
		{ 372, 472, 58 },
		{ 372, 473, 58 },
		{ 372, 474, 58 },
		{ 372, 475, 58 },
		{ 372, 476, 58 },
		{ 372, 477, 58 },
		{ 372, 478, 58 },
		{ 372, 479, 58 },
		{ 372, 480, 58 },
		{ 372, 481, 58 },
		{ 372, 482, 58 },
		{ 372, 483, 58 },
		{ 372, 484, 58 },
		{ 372, 485, 58 },
		{ 372, 486, 58 },
		{ 372, 487, 58 },
		{ 372, 488, 58 },
		{ 372, 489, 58 },
		{ 372, 490, 58 },
		{ 372, 491, 58 },
		{ 372, 492, 58 },
		{ 372, 493, 58 },
		{ 372, 494, 58 },
		{ 372, 495, 58 },
		{ 372, 496, 58 },
		{ 372, 497, 58 },
		{ 372, 498, 58 },
		{ 372, 499, 58 },
		{ 372, 500, 58 },
		{ 372, 501, 58 },
		{ 372, 502, 58 },
		{ 372, 503, 58 },
		{ 372, 504, 58 },
		{ 372, 505, 58 },
		{ 372, 506, 58 },
		{ 372, 507, 58 },
		{ 372, 508, 58 },
		{ 372, 509, 58 },
		{ 372, 510, 58 },
		{ 372, 511, 58 },
		{ 372, 512, 58 },
		{ 373, 443, 58 },
		{ 373, 444, 58 },
		{ 373, 445, 58 },
		{ 373, 446, 58 },
		{ 373, 447, 58 },
		{ 373, 448, 58 },
		{ 373, 449, 58 },
		{ 373, 450, 58 },
		{ 373, 451, 58 },
		{ 373, 452, 58 },
		{ 373, 453, 58 },
		{ 373, 454, 58 },
		{ 373, 455, 58 },
		{ 373, 456, 58 },
		{ 373, 457, 58 },
		{ 373, 458, 58 },
		{ 373, 459, 58 },
		{ 373, 460, 58 },
		{ 373, 461, 58 },
		{ 373, 462, 58 },
		{ 373, 463, 58 },
		{ 373, 464, 58 },
		{ 373, 465, 58 },
		{ 373, 466, 58 },
		{ 373, 467, 58 },
		{ 373, 468, 58 },
		{ 373, 469, 58 },
		{ 373, 470, 58 },
		{ 373, 471, 58 },
		{ 373, 472, 58 },
		{ 373, 473, 58 },
		{ 373, 474, 58 },
		{ 373, 475, 58 },
		{ 373, 476, 58 },
		{ 373, 477, 58 },
		{ 373, 478, 58 },
		{ 373, 479, 58 },
		{ 373, 480, 58 },
		{ 373, 481, 58 },
		{ 373, 482, 58 },
		{ 373, 483, 58 },
		{ 373, 484, 58 },
		{ 373, 485, 58 },
		{ 373, 486, 58 },
		{ 373, 487, 58 },
		{ 373, 488, 58 },
		{ 373, 489, 58 },
		{ 373, 490, 58 },
		{ 373, 491, 58 },
		{ 373, 492, 58 },
		{ 373, 493, 58 },
		{ 373, 494, 58 },
		{ 373, 495, 58 },
		{ 373, 496, 58 },
		{ 373, 497, 58 },
		{ 373, 498, 58 },
		{ 373, 499, 58 },
		{ 373, 500, 58 },
		{ 373, 501, 58 },
		{ 373, 502, 58 },
		{ 373, 503, 58 },
		{ 373, 504, 58 },
		{ 373, 505, 58 },
		{ 373, 506, 58 },
		{ 373, 507, 58 },
		{ 373, 508, 58 },
		{ 373, 509, 58 },
		{ 373, 510, 58 },
		{ 373, 511, 58 },
		{ 373, 512, 58 },
		{ 374, 445, 58 },
		{ 374, 446, 58 },
		{ 374, 447, 58 },
		{ 374, 448, 58 },
		{ 374, 449, 58 },
		{ 374, 450, 58 },
		{ 374, 451, 58 },
		{ 374, 452, 58 },
		{ 374, 453, 58 },
		{ 374, 454, 58 },
		{ 374, 455, 58 },
		{ 374, 456, 58 },
		{ 374, 457, 58 },
		{ 374, 458, 58 },
		{ 374, 459, 58 },
		{ 374, 460, 58 },
		{ 374, 461, 58 },
		{ 374, 462, 58 },
		{ 374, 463, 58 },
		{ 374, 464, 58 },
		{ 374, 465, 58 },
		{ 374, 466, 58 },
		{ 374, 467, 58 },
		{ 374, 468, 58 },
		{ 374, 469, 58 },
		{ 374, 470, 58 },
		{ 374, 471, 58 },
		{ 374, 472, 58 },
		{ 374, 473, 58 },
		{ 374, 474, 58 },
		{ 374, 475, 58 },
		{ 374, 476, 58 },
		{ 374, 477, 58 },
		{ 374, 478, 58 },
		{ 374, 479, 58 },
		{ 374, 480, 58 },
		{ 374, 481, 58 },
		{ 374, 482, 58 },
		{ 374, 483, 58 },
		{ 374, 484, 58 },
		{ 374, 485, 58 },
		{ 374, 486, 58 },
		{ 374, 487, 58 },
		{ 374, 488, 58 },
		{ 374, 489, 58 },
		{ 374, 490, 58 },
		{ 374, 491, 58 },
		{ 374, 492, 58 },
		{ 374, 493, 58 },
		{ 374, 494, 58 },
		{ 374, 495, 58 },
		{ 374, 496, 58 },
		{ 374, 497, 58 },
		{ 374, 498, 58 },
		{ 374, 499, 58 },
		{ 374, 500, 58 },
		{ 374, 501, 58 },
		{ 374, 502, 58 },
		{ 374, 503, 58 },
		{ 374, 504, 58 },
		{ 374, 505, 58 },
		{ 374, 506, 58 },
		{ 374, 507, 58 },
		{ 374, 508, 58 },
		{ 374, 509, 58 },
		{ 374, 510, 58 },
		{ 374, 511, 58 },
		{ 374, 512, 58 },
		{ 375, 446, 58 },
		{ 375, 447, 58 },
		{ 375, 448, 58 },
		{ 375, 449, 58 },
		{ 375, 450, 58 },
		{ 375, 451, 58 },
		{ 375, 452, 58 },
		{ 375, 453, 58 },
		{ 375, 454, 58 },
		{ 375, 455, 58 },
		{ 375, 456, 58 },
		{ 375, 457, 58 },
		{ 375, 458, 58 },
		{ 375, 459, 58 },
		{ 375, 460, 58 },
		{ 375, 461, 58 },
		{ 375, 462, 58 },
		{ 375, 463, 58 },
		{ 375, 464, 58 },
		{ 375, 465, 58 },
		{ 375, 466, 58 },
		{ 375, 467, 58 },
		{ 375, 468, 58 },
		{ 375, 469, 58 },
		{ 375, 470, 58 },
		{ 375, 471, 58 },
		{ 375, 472, 58 },
		{ 375, 473, 58 },
		{ 375, 474, 58 },
		{ 375, 475, 58 },
		{ 375, 476, 58 },
		{ 375, 477, 58 },
		{ 375, 478, 58 },
		{ 375, 479, 58 },
		{ 375, 480, 58 },
		{ 375, 481, 58 },
		{ 375, 482, 58 },
		{ 375, 483, 58 },
		{ 375, 484, 58 },
		{ 375, 485, 58 },
		{ 375, 486, 58 },
		{ 375, 487, 58 },
		{ 375, 488, 58 },
		{ 375, 489, 58 },
		{ 375, 490, 58 },
		{ 375, 491, 58 },
		{ 375, 492, 58 },
		{ 375, 493, 58 },
		{ 375, 494, 58 },
		{ 375, 495, 58 },
		{ 375, 496, 58 },
		{ 375, 497, 58 },
		{ 375, 498, 58 },
		{ 375, 499, 58 },
		{ 375, 500, 58 },
		{ 375, 501, 58 },
		{ 375, 502, 58 },
		{ 375, 503, 58 },
		{ 375, 504, 58 },
		{ 375, 505, 58 },
		{ 375, 506, 58 },
		{ 375, 507, 58 },
		{ 375, 508, 58 },
		{ 375, 509, 58 },
		{ 375, 510, 58 },
		{ 375, 511, 58 },
		{ 375, 512, 58 },
		{ 376, 448, 58 },
		{ 376, 449, 58 },
		{ 376, 450, 58 },
		{ 376, 451, 58 },
		{ 376, 452, 58 },
		{ 376, 453, 58 },
		{ 376, 454, 58 },
		{ 376, 455, 58 },
		{ 376, 456, 58 },
		{ 376, 457, 58 },
		{ 376, 458, 58 },
		{ 376, 459, 58 },
		{ 376, 460, 58 },
		{ 376, 461, 58 },
		{ 376, 462, 58 },
		{ 376, 463, 58 },
		{ 376, 464, 58 },
		{ 376, 465, 58 },
		{ 376, 466, 58 },
		{ 376, 467, 58 },
		{ 376, 468, 58 },
		{ 376, 469, 58 },
		{ 376, 470, 58 },
		{ 376, 471, 58 },
		{ 376, 472, 58 },
		{ 376, 473, 58 },
		{ 376, 474, 58 },
		{ 376, 475, 58 },
		{ 376, 476, 58 },
		{ 376, 477, 58 },
		{ 376, 478, 58 },
		{ 376, 479, 58 },
		{ 376, 480, 58 },
		{ 376, 481, 58 },
		{ 376, 482, 58 },
		{ 376, 483, 58 },
		{ 376, 484, 58 },
		{ 376, 485, 58 },
		{ 376, 486, 58 },
		{ 376, 487, 58 },
		{ 376, 488, 58 },
		{ 376, 489, 58 },
		{ 376, 490, 58 },
		{ 376, 491, 58 },
		{ 376, 492, 58 },
		{ 376, 493, 58 },
		{ 376, 494, 58 },
		{ 376, 495, 58 },
		{ 376, 496, 58 },
		{ 376, 497, 58 },
		{ 376, 498, 58 },
		{ 376, 499, 58 },
		{ 376, 500, 58 },
		{ 376, 501, 58 },
		{ 376, 502, 58 },
		{ 376, 503, 58 },
		{ 376, 504, 58 },
		{ 376, 505, 58 },
		{ 376, 506, 58 },
		{ 376, 507, 58 },
		{ 376, 508, 58 },
		{ 376, 509, 58 },
		{ 376, 510, 58 },
		{ 376, 511, 58 },
		{ 376, 512, 58 },
		{ 377, 450, 58 },
		{ 377, 451, 58 },
		{ 377, 452, 58 },
		{ 377, 453, 58 },
		{ 377, 454, 58 },
		{ 377, 455, 58 },
		{ 377, 456, 58 },
		{ 377, 457, 58 },
		{ 377, 458, 58 },
		{ 377, 459, 58 },
		{ 377, 460, 58 },
		{ 377, 461, 58 },
		{ 377, 462, 58 },
		{ 377, 463, 58 },
		{ 377, 464, 58 },
		{ 377, 465, 58 },
		{ 377, 466, 58 },
		{ 377, 467, 58 },
		{ 377, 468, 58 },
		{ 377, 469, 58 },
		{ 377, 470, 58 },
		{ 377, 471, 58 },
		{ 377, 472, 58 },
		{ 377, 473, 58 },
		{ 377, 474, 58 },
		{ 377, 475, 58 },
		{ 377, 476, 58 },
		{ 377, 477, 58 },
		{ 377, 478, 58 },
		{ 377, 479, 58 },
		{ 377, 480, 58 },
		{ 377, 481, 58 },
		{ 377, 482, 58 },
		{ 377, 483, 58 },
		{ 377, 484, 58 },
		{ 377, 485, 58 },
		{ 377, 486, 58 },
		{ 377, 487, 58 },
		{ 377, 488, 58 },
		{ 377, 489, 58 },
		{ 377, 490, 58 },
		{ 377, 491, 58 },
		{ 377, 492, 58 },
		{ 377, 493, 58 },
		{ 377, 494, 58 },
		{ 377, 495, 58 },
		{ 377, 496, 58 },
		{ 377, 497, 58 },
		{ 377, 498, 58 },
		{ 377, 499, 58 },
		{ 377, 500, 58 },
		{ 377, 501, 58 },
		{ 377, 502, 58 },
		{ 377, 503, 58 },
		{ 377, 504, 58 },
		{ 377, 505, 58 },
		{ 377, 506, 58 },
		{ 377, 507, 58 },
		{ 377, 508, 58 },
		{ 377, 509, 58 },
		{ 377, 510, 58 },
		{ 377, 511, 58 },
		{ 377, 512, 58 },
		{ 378, 451, 58 },
		{ 378, 452, 58 },
		{ 378, 453, 58 },
		{ 378, 454, 58 },
		{ 378, 455, 58 },
		{ 378, 456, 58 },
		{ 378, 457, 58 },
		{ 378, 458, 58 },
		{ 378, 459, 58 },
		{ 378, 460, 58 },
		{ 378, 461, 58 },
		{ 378, 462, 58 },
		{ 378, 463, 58 },
		{ 378, 464, 58 },
		{ 378, 465, 58 },
		{ 378, 466, 58 },
		{ 378, 467, 58 },
		{ 378, 468, 58 },
		{ 378, 469, 58 },
		{ 378, 470, 58 },
		{ 378, 471, 58 },
		{ 378, 472, 58 },
		{ 378, 473, 58 },
		{ 378, 474, 58 },
		{ 378, 475, 58 },
		{ 378, 476, 58 },
		{ 378, 477, 58 },
		{ 378, 478, 58 },
		{ 378, 479, 58 },
		{ 378, 480, 58 },
		{ 378, 481, 58 },
		{ 378, 482, 58 },
		{ 378, 483, 58 },
		{ 378, 484, 58 },
		{ 378, 485, 58 },
		{ 378, 486, 58 },
		{ 378, 487, 58 },
		{ 378, 488, 58 },
		{ 378, 489, 58 },
		{ 378, 490, 58 },
		{ 378, 491, 58 },
		{ 378, 492, 58 },
		{ 378, 493, 58 },
		{ 378, 494, 58 },
		{ 378, 495, 58 },
		{ 378, 496, 58 },
		{ 378, 497, 58 },
		{ 378, 498, 58 },
		{ 378, 499, 58 },
		{ 378, 500, 58 },
		{ 378, 501, 58 },
		{ 378, 502, 58 },
		{ 378, 503, 58 },
		{ 378, 504, 58 },
		{ 378, 505, 58 },
		{ 378, 506, 58 },
		{ 378, 507, 58 },
		{ 378, 508, 58 },
		{ 378, 509, 58 },
		{ 378, 510, 58 },
		{ 378, 511, 58 },
		{ 378, 512, 58 },
		{ 379, 453, 58 },
		{ 379, 454, 58 },
		{ 379, 455, 58 },
		{ 379, 456, 58 },
		{ 379, 457, 58 },
		{ 379, 458, 58 },
		{ 379, 459, 58 },
		{ 379, 460, 58 },
		{ 379, 461, 58 },
		{ 379, 462, 58 },
		{ 379, 463, 58 },
		{ 379, 464, 58 },
		{ 379, 465, 58 },
		{ 379, 466, 58 },
		{ 379, 467, 58 },
		{ 379, 468, 58 },
		{ 379, 469, 58 },
		{ 379, 470, 58 },
		{ 379, 471, 58 },
		{ 379, 472, 58 },
		{ 379, 473, 58 },
		{ 379, 474, 58 },
		{ 379, 475, 58 },
		{ 379, 476, 58 },
		{ 379, 477, 58 },
		{ 379, 478, 58 },
		{ 379, 479, 58 },
		{ 379, 480, 58 },
		{ 379, 481, 58 },
		{ 379, 482, 58 },
		{ 379, 483, 58 },
		{ 379, 484, 58 },
		{ 379, 485, 58 },
		{ 379, 486, 58 },
		{ 379, 487, 58 },
		{ 379, 488, 58 },
		{ 379, 489, 58 },
		{ 379, 490, 58 },
		{ 379, 491, 58 },
		{ 379, 492, 58 },
		{ 379, 493, 58 },
		{ 379, 494, 58 },
		{ 379, 495, 58 },
		{ 379, 496, 58 },
		{ 379, 497, 58 },
		{ 379, 498, 58 },
		{ 379, 499, 58 },
		{ 379, 500, 58 },
		{ 379, 501, 58 },
		{ 379, 502, 58 },
		{ 379, 503, 58 },
		{ 379, 504, 58 },
		{ 379, 505, 58 },
		{ 379, 506, 58 },
		{ 379, 507, 58 },
		{ 379, 508, 58 },
		{ 379, 509, 58 },
		{ 379, 510, 58 },
		{ 379, 511, 58 },
		{ 379, 512, 58 },
		{ 380, 454, 58 },
		{ 380, 455, 58 },
		{ 380, 456, 58 },
		{ 380, 457, 58 },
		{ 380, 458, 58 },
		{ 380, 459, 58 },
		{ 380, 460, 58 },
		{ 380, 461, 58 },
		{ 380, 462, 58 },
		{ 380, 463, 58 },
		{ 380, 464, 58 },
		{ 380, 465, 58 },
		{ 380, 466, 58 },
		{ 380, 467, 58 },
		{ 380, 468, 58 },
		{ 380, 469, 58 },
		{ 380, 470, 58 },
		{ 380, 471, 58 },
		{ 380, 472, 58 },
		{ 380, 473, 58 },
		{ 380, 474, 58 },
		{ 380, 475, 58 },
		{ 380, 476, 58 },
		{ 380, 477, 58 },
		{ 380, 478, 58 },
		{ 380, 479, 58 },
		{ 380, 480, 58 },
		{ 380, 481, 58 },
		{ 380, 482, 58 },
		{ 380, 483, 58 },
		{ 380, 484, 58 },
		{ 380, 485, 58 },
		{ 380, 486, 58 },
		{ 380, 487, 58 },
		{ 380, 488, 58 },
		{ 380, 489, 58 },
		{ 380, 490, 58 },
		{ 380, 491, 58 },
		{ 380, 492, 58 },
		{ 380, 493, 58 },
		{ 380, 494, 58 },
		{ 380, 495, 58 },
		{ 380, 496, 58 },
		{ 380, 497, 58 },
		{ 380, 498, 58 },
		{ 380, 499, 58 },
		{ 380, 500, 58 },
		{ 380, 501, 58 },
		{ 380, 502, 58 },
		{ 380, 503, 58 },
		{ 380, 504, 58 },
		{ 380, 505, 58 },
		{ 380, 506, 58 },
		{ 380, 507, 58 },
		{ 380, 508, 58 },
		{ 380, 509, 58 },
		{ 380, 510, 58 },
		{ 380, 511, 58 },
		{ 380, 512, 58 },
		{ 381, 456, 58 },
		{ 381, 457, 58 },
		{ 381, 458, 58 },
		{ 381, 459, 58 },
		{ 381, 460, 58 },
		{ 381, 461, 58 },
		{ 381, 462, 58 },
		{ 381, 463, 58 },
		{ 381, 464, 58 },
		{ 381, 465, 58 },
		{ 381, 466, 58 },
		{ 381, 467, 58 },
		{ 381, 468, 58 },
		{ 381, 469, 58 },
		{ 381, 470, 58 },
		{ 381, 471, 58 },
		{ 381, 472, 58 },
		{ 381, 473, 58 },
		{ 381, 474, 58 },
		{ 381, 475, 58 },
		{ 381, 476, 58 },
		{ 381, 477, 58 },
		{ 381, 478, 58 },
		{ 381, 479, 58 },
		{ 381, 480, 58 },
		{ 381, 481, 58 },
		{ 381, 482, 58 },
		{ 381, 483, 58 },
		{ 381, 484, 58 },
		{ 381, 485, 58 },
		{ 381, 486, 58 },
		{ 381, 487, 58 },
		{ 381, 488, 58 },
		{ 381, 489, 58 },
		{ 381, 490, 58 },
		{ 381, 491, 58 },
		{ 381, 492, 58 },
		{ 381, 493, 58 },
		{ 381, 494, 58 },
		{ 381, 495, 58 },
		{ 381, 496, 58 },
		{ 381, 497, 58 },
		{ 381, 498, 58 },
		{ 381, 499, 58 },
		{ 381, 500, 58 },
		{ 381, 501, 58 },
		{ 381, 502, 58 },
		{ 381, 503, 58 },
		{ 381, 504, 58 },
		{ 381, 505, 58 },
		{ 381, 506, 58 },
		{ 381, 507, 58 },
		{ 381, 508, 58 },
		{ 381, 509, 58 },
		{ 381, 510, 58 },
		{ 381, 511, 58 },
		{ 381, 512, 58 },
		{ 382, 458, 58 },
		{ 382, 459, 58 },
		{ 382, 460, 58 },
		{ 382, 461, 58 },
		{ 382, 462, 58 },
		{ 382, 463, 58 },
		{ 382, 464, 58 },
		{ 382, 465, 58 },
		{ 382, 466, 58 },
		{ 382, 467, 58 },
		{ 382, 468, 58 },
		{ 382, 469, 58 },
		{ 382, 470, 58 },
		{ 382, 471, 58 },
		{ 382, 472, 58 },
		{ 382, 473, 58 },
		{ 382, 474, 58 },
		{ 382, 475, 58 },
		{ 382, 476, 58 },
		{ 382, 477, 58 },
		{ 382, 478, 58 },
		{ 382, 479, 58 },
		{ 382, 480, 58 },
		{ 382, 481, 58 },
		{ 382, 482, 58 },
		{ 382, 483, 58 },
		{ 382, 484, 58 },
		{ 382, 485, 58 },
		{ 382, 486, 58 },
		{ 382, 487, 58 },
		{ 382, 488, 58 },
		{ 382, 489, 58 },
		{ 382, 490, 58 },
		{ 382, 491, 58 },
		{ 382, 492, 58 },
		{ 382, 493, 58 },
		{ 382, 494, 58 },
		{ 382, 495, 58 },
		{ 382, 496, 58 },
		{ 382, 497, 58 },
		{ 382, 498, 58 },
		{ 382, 499, 58 },
		{ 382, 500, 58 },
		{ 382, 501, 58 },
		{ 382, 502, 58 },
		{ 382, 503, 58 },
		{ 382, 504, 58 },
		{ 382, 505, 58 },
		{ 382, 506, 58 },
		{ 382, 507, 58 },
		{ 382, 508, 58 },
		{ 382, 509, 58 },
		{ 382, 510, 58 },
		{ 382, 511, 58 },
		{ 382, 512, 58 },
		{ 383, 459, 58 },
		{ 383, 460, 58 },
		{ 383, 461, 58 },
		{ 383, 462, 58 },
		{ 383, 463, 58 },
		{ 383, 464, 58 },
		{ 383, 465, 58 },
		{ 383, 466, 58 },
		{ 383, 467, 58 },
		{ 383, 468, 58 },
		{ 383, 469, 58 },
		{ 383, 470, 58 },
		{ 383, 471, 58 },
		{ 383, 472, 58 },
		{ 383, 473, 58 },
		{ 383, 474, 58 },
		{ 383, 475, 58 },
		{ 383, 476, 58 },
		{ 383, 477, 58 },
		{ 383, 478, 58 },
		{ 383, 479, 58 },
		{ 383, 480, 58 },
		{ 383, 481, 58 },
		{ 383, 482, 58 },
		{ 383, 483, 58 },
		{ 383, 484, 58 },
		{ 383, 485, 58 },
		{ 383, 486, 58 },
		{ 383, 487, 58 },
		{ 383, 488, 58 },
		{ 383, 489, 58 },
		{ 383, 490, 58 },
		{ 383, 491, 58 },
		{ 383, 492, 58 },
		{ 383, 493, 58 },
		{ 383, 494, 58 },
		{ 383, 495, 58 },
		{ 383, 496, 58 },
		{ 383, 497, 58 },
		{ 383, 498, 58 },
		{ 383, 499, 58 },
		{ 383, 500, 58 },
		{ 383, 501, 58 },
		{ 383, 502, 58 },
		{ 383, 503, 58 },
		{ 383, 504, 58 },
		{ 383, 505, 58 },
		{ 383, 506, 58 },
		{ 383, 507, 58 },
		{ 383, 508, 58 },
		{ 383, 509, 58 },
		{ 383, 510, 58 },
		{ 383, 511, 58 },
		{ 383, 512, 58 },
		{ 384, 461, 58 },
		{ 384, 462, 58 },
		{ 384, 463, 58 },
		{ 384, 464, 58 },
		{ 384, 465, 58 },
		{ 384, 466, 58 },
		{ 384, 467, 58 },
		{ 384, 468, 58 },
		{ 384, 469, 58 },
		{ 384, 470, 58 },
		{ 384, 471, 58 },
		{ 384, 472, 58 },
		{ 384, 473, 58 },
		{ 384, 474, 58 },
		{ 384, 475, 58 },
		{ 384, 476, 58 },
		{ 384, 477, 58 },
		{ 384, 478, 58 },
		{ 384, 479, 58 },
		{ 384, 480, 58 },
		{ 384, 481, 58 },
		{ 384, 482, 58 },
		{ 384, 483, 58 },
		{ 384, 484, 58 },
		{ 384, 485, 58 },
		{ 384, 486, 58 },
		{ 384, 487, 58 },
		{ 384, 488, 58 },
		{ 384, 489, 58 },
		{ 384, 490, 58 },
		{ 384, 491, 58 },
		{ 384, 492, 58 },
		{ 384, 493, 58 },
		{ 384, 494, 58 },
		{ 384, 495, 58 },
		{ 384, 496, 58 },
		{ 384, 497, 58 },
		{ 384, 498, 58 },
		{ 384, 499, 58 },
		{ 384, 500, 58 },
		{ 384, 501, 58 },
		{ 384, 502, 58 },
		{ 384, 503, 58 },
		{ 384, 504, 58 },
		{ 384, 505, 58 },
		{ 384, 506, 58 },
		{ 384, 507, 58 },
		{ 384, 508, 58 },
		{ 384, 509, 58 },
		{ 384, 510, 58 },
		{ 384, 511, 58 },
		{ 384, 512, 58 },
		{ 385, 462, 58 },
		{ 385, 463, 58 },
		{ 385, 464, 58 },
		{ 385, 465, 58 },
		{ 385, 466, 58 },
		{ 385, 467, 58 },
		{ 385, 468, 58 },
		{ 385, 469, 58 },
		{ 385, 470, 58 },
		{ 385, 471, 58 },
		{ 385, 472, 58 },
		{ 385, 473, 58 },
		{ 385, 474, 58 },
		{ 385, 475, 58 },
		{ 385, 476, 58 },
		{ 385, 477, 58 },
		{ 385, 478, 58 },
		{ 385, 479, 58 },
		{ 385, 480, 58 },
		{ 385, 481, 58 },
		{ 385, 482, 58 },
		{ 385, 483, 58 },
		{ 385, 484, 58 },
		{ 385, 485, 58 },
		{ 385, 486, 58 },
		{ 385, 487, 58 },
		{ 385, 488, 58 },
		{ 385, 489, 58 },
		{ 385, 490, 58 },
		{ 385, 491, 58 },
		{ 385, 492, 58 },
		{ 385, 493, 58 },
		{ 385, 494, 58 },
		{ 385, 495, 58 },
		{ 385, 496, 58 },
		{ 385, 497, 58 },
		{ 385, 498, 58 },
		{ 385, 499, 58 },
		{ 385, 500, 58 },
		{ 385, 501, 58 },
		{ 385, 502, 58 },
		{ 385, 503, 58 },
		{ 385, 504, 58 },
		{ 385, 505, 58 },
		{ 385, 506, 58 },
		{ 385, 507, 58 },
		{ 385, 508, 58 },
		{ 385, 509, 58 },
		{ 385, 510, 58 },
		{ 385, 511, 58 },
		{ 385, 512, 58 },
		{ 386, 464, 58 },
		{ 386, 465, 58 },
		{ 386, 466, 58 },
		{ 386, 467, 58 },
		{ 386, 468, 58 },
		{ 386, 469, 58 },
		{ 386, 470, 58 },
		{ 386, 471, 58 },
		{ 386, 472, 58 },
		{ 386, 473, 58 },
		{ 386, 474, 58 },
		{ 386, 475, 58 },
		{ 386, 476, 58 },
		{ 386, 477, 58 },
		{ 386, 478, 58 },
		{ 386, 479, 58 },
		{ 386, 480, 58 },
		{ 386, 481, 58 },
		{ 386, 482, 58 },
		{ 386, 483, 58 },
		{ 386, 484, 58 },
		{ 386, 485, 58 },
		{ 386, 486, 58 },
		{ 386, 487, 58 },
		{ 386, 488, 58 },
		{ 386, 489, 58 },
		{ 386, 490, 58 },
		{ 386, 491, 58 },
		{ 386, 492, 58 },
		{ 386, 493, 58 },
		{ 386, 494, 58 },
		{ 386, 495, 58 },
		{ 386, 496, 58 },
		{ 386, 497, 58 },
		{ 386, 498, 58 },
		{ 386, 499, 58 },
		{ 386, 500, 58 },
		{ 386, 501, 58 },
		{ 386, 502, 58 },
		{ 386, 503, 58 },
		{ 386, 504, 58 },
		{ 386, 505, 58 },
		{ 386, 506, 58 },
		{ 386, 507, 58 },
		{ 386, 508, 58 },
		{ 386, 509, 58 },
		{ 386, 510, 58 },
		{ 386, 511, 58 },
		{ 386, 512, 58 },
		{ 387, 466, 58 },
		{ 387, 467, 58 },
		{ 387, 468, 58 },
		{ 387, 469, 58 },
		{ 387, 470, 58 },
		{ 387, 471, 58 },
		{ 387, 472, 58 },
		{ 387, 473, 58 },
		{ 387, 474, 58 },
		{ 387, 475, 58 },
		{ 387, 476, 58 },
		{ 387, 477, 58 },
		{ 387, 478, 58 },
		{ 387, 479, 58 },
		{ 387, 480, 58 },
		{ 387, 481, 58 },
		{ 387, 482, 58 },
		{ 387, 483, 58 },
		{ 387, 484, 58 },
		{ 387, 485, 58 },
		{ 387, 486, 58 },
		{ 387, 487, 58 },
		{ 387, 488, 58 },
		{ 387, 489, 58 },
		{ 387, 490, 58 },
		{ 387, 491, 58 },
		{ 387, 492, 58 },
		{ 387, 493, 58 },
		{ 387, 494, 58 },
		{ 387, 495, 58 },
		{ 387, 496, 58 },
		{ 387, 497, 58 },
		{ 387, 498, 58 },
		{ 387, 499, 58 },
		{ 387, 500, 58 },
		{ 387, 501, 58 },
		{ 387, 502, 58 },
		{ 387, 503, 58 },
		{ 387, 504, 58 },
		{ 387, 505, 58 },
		{ 387, 506, 58 },
		{ 387, 507, 58 },
		{ 387, 508, 58 },
		{ 387, 509, 58 },
		{ 387, 510, 58 },
		{ 387, 511, 58 },
		{ 387, 512, 58 },
		{ 388, 467, 58 },
		{ 388, 468, 58 },
		{ 388, 469, 58 },
		{ 388, 470, 58 },
		{ 388, 471, 58 },
		{ 388, 472, 58 },
		{ 388, 473, 58 },
		{ 388, 474, 58 },
		{ 388, 475, 58 },
		{ 388, 476, 58 },
		{ 388, 477, 58 },
		{ 388, 478, 58 },
		{ 388, 479, 58 },
		{ 388, 480, 58 },
		{ 388, 481, 58 },
		{ 388, 482, 58 },
		{ 388, 483, 58 },
		{ 388, 484, 58 },
		{ 388, 485, 58 },
		{ 388, 486, 58 },
		{ 388, 487, 58 },
		{ 388, 488, 58 },
		{ 388, 489, 58 },
		{ 388, 490, 58 },
		{ 388, 491, 58 },
		{ 388, 492, 58 },
		{ 388, 493, 58 },
		{ 388, 494, 58 },
		{ 388, 495, 58 },
		{ 388, 496, 58 },
		{ 388, 497, 58 },
		{ 388, 498, 58 },
		{ 388, 499, 58 },
		{ 388, 500, 58 },
		{ 388, 501, 58 },
		{ 388, 502, 58 },
		{ 388, 503, 58 },
		{ 388, 504, 58 },
		{ 388, 505, 58 },
		{ 388, 506, 58 },
		{ 388, 507, 58 },
		{ 388, 508, 58 },
		{ 388, 509, 58 },
		{ 388, 510, 58 },
		{ 388, 511, 58 },
		{ 388, 512, 58 },
		{ 389, 469, 58 },
		{ 389, 470, 58 },
		{ 389, 471, 58 },
		{ 389, 472, 58 },
		{ 389, 473, 58 },
		{ 389, 474, 58 },
		{ 389, 475, 58 },
		{ 389, 476, 58 },
		{ 389, 477, 58 },
		{ 389, 478, 58 },
		{ 389, 479, 58 },
		{ 389, 480, 58 },
		{ 389, 481, 58 },
		{ 389, 482, 58 },
		{ 389, 483, 58 },
		{ 389, 484, 58 },
		{ 389, 485, 58 },
		{ 389, 486, 58 },
		{ 389, 487, 58 },
		{ 389, 488, 58 },
		{ 389, 489, 58 },
		{ 389, 490, 58 },
		{ 389, 491, 58 },
		{ 389, 492, 58 },
		{ 389, 493, 58 },
		{ 389, 494, 58 },
		{ 389, 495, 58 },
		{ 389, 496, 58 },
		{ 389, 497, 58 },
		{ 389, 498, 58 },
		{ 389, 499, 58 },
		{ 389, 500, 58 },
		{ 389, 501, 58 },
		{ 389, 502, 58 },
		{ 389, 503, 58 },
		{ 389, 504, 58 },
		{ 389, 505, 58 },
		{ 389, 506, 58 },
		{ 389, 507, 58 },
		{ 389, 508, 58 },
		{ 389, 509, 58 },
		{ 389, 510, 58 },
		{ 389, 511, 58 },
		{ 389, 512, 58 },
		{ 390, 470, 58 },
		{ 390, 471, 58 },
		{ 390, 472, 58 },
		{ 390, 473, 58 },
		{ 390, 474, 58 },
		{ 390, 475, 58 },
		{ 390, 476, 58 },
		{ 390, 477, 58 },
		{ 390, 478, 58 },
		{ 390, 479, 58 },
		{ 390, 480, 58 },
		{ 390, 481, 58 },
		{ 390, 482, 58 },
		{ 390, 483, 58 },
		{ 390, 484, 58 },
		{ 390, 485, 58 },
		{ 390, 486, 58 },
		{ 390, 487, 58 },
		{ 390, 488, 58 },
		{ 390, 489, 58 },
		{ 390, 490, 58 },
		{ 390, 491, 58 },
		{ 390, 492, 58 },
		{ 390, 493, 58 },
		{ 390, 494, 58 },
		{ 390, 495, 58 },
		{ 390, 496, 58 },
		{ 390, 497, 58 },
		{ 390, 498, 58 },
		{ 390, 499, 58 },
		{ 390, 500, 58 },
		{ 390, 501, 58 },
		{ 390, 502, 58 },
		{ 390, 503, 58 },
		{ 390, 504, 58 },
		{ 390, 505, 58 },
		{ 390, 506, 58 },
		{ 390, 507, 58 },
		{ 390, 508, 58 },
		{ 390, 509, 58 },
		{ 390, 510, 58 },
		{ 390, 511, 58 },
		{ 390, 512, 58 },
		{ 391, 472, 58 },
		{ 391, 473, 58 },
		{ 391, 474, 58 },
		{ 391, 475, 58 },
		{ 391, 476, 58 },
		{ 391, 477, 58 },
		{ 391, 478, 58 },
		{ 391, 479, 58 },
		{ 391, 480, 58 },
		{ 391, 481, 58 },
		{ 391, 482, 58 },
		{ 391, 483, 58 },
		{ 391, 484, 58 },
		{ 391, 485, 58 },
		{ 391, 486, 58 },
		{ 391, 487, 58 },
		{ 391, 488, 58 },
		{ 391, 489, 58 },
		{ 391, 490, 58 },
		{ 391, 491, 58 },
		{ 391, 492, 58 },
		{ 391, 493, 58 },
		{ 391, 494, 58 },
		{ 391, 495, 58 },
		{ 391, 496, 58 },
		{ 391, 497, 58 },
		{ 391, 498, 58 },
		{ 391, 499, 58 },
		{ 391, 500, 58 },
		{ 391, 501, 58 },
		{ 391, 502, 58 },
		{ 391, 503, 58 },
		{ 391, 504, 58 },
		{ 391, 505, 58 },
		{ 391, 506, 58 },
		{ 391, 507, 58 },
		{ 391, 508, 58 },
		{ 391, 509, 58 },
		{ 391, 510, 58 },
		{ 391, 511, 58 },
		{ 391, 512, 58 },
		{ 392, 474, 58 },
		{ 392, 475, 58 },
		{ 392, 476, 58 },
		{ 392, 477, 58 },
		{ 392, 478, 58 },
		{ 392, 479, 58 },
		{ 392, 480, 58 },
		{ 392, 481, 58 },
		{ 392, 482, 58 },
		{ 392, 483, 58 },
		{ 392, 484, 58 },
		{ 392, 485, 58 },
		{ 392, 486, 58 },
		{ 392, 487, 58 },
		{ 392, 488, 58 },
		{ 392, 489, 58 },
		{ 392, 490, 58 },
		{ 392, 491, 58 },
		{ 392, 492, 58 },
		{ 392, 493, 58 },
		{ 392, 494, 58 },
		{ 392, 495, 58 },
		{ 392, 496, 58 },
		{ 392, 497, 58 },
		{ 392, 498, 58 },
		{ 392, 499, 58 },
		{ 392, 500, 58 },
		{ 392, 501, 58 },
		{ 392, 502, 58 },
		{ 392, 503, 58 },
		{ 392, 504, 58 },
		{ 392, 505, 58 },
		{ 392, 506, 58 },
		{ 392, 507, 58 },
		{ 392, 508, 58 },
		{ 392, 509, 58 },
		{ 392, 510, 58 },
		{ 392, 511, 58 },
		{ 392, 512, 58 },
		{ 393, 475, 58 },
		{ 393, 476, 58 },
		{ 393, 477, 58 },
		{ 393, 478, 58 },
		{ 393, 479, 58 },
		{ 393, 480, 58 },
		{ 393, 481, 58 },
		{ 393, 482, 58 },
		{ 393, 483, 58 },
		{ 393, 484, 58 },
		{ 393, 485, 58 },
		{ 393, 486, 58 },
		{ 393, 487, 58 },
		{ 393, 488, 58 },
		{ 393, 489, 58 },
		{ 393, 490, 58 },
		{ 393, 491, 58 },
		{ 393, 492, 58 },
		{ 393, 493, 58 },
		{ 393, 494, 58 },
		{ 393, 495, 58 },
		{ 393, 496, 58 },
		{ 393, 497, 58 },
		{ 393, 498, 58 },
		{ 393, 499, 58 },
		{ 393, 500, 58 },
		{ 393, 501, 58 },
		{ 393, 502, 58 },
		{ 393, 503, 58 },
		{ 393, 504, 58 },
		{ 393, 505, 58 },
		{ 393, 506, 58 },
		{ 393, 507, 58 },
		{ 393, 508, 58 },
		{ 393, 509, 58 },
		{ 393, 510, 58 },
		{ 393, 511, 58 },
		{ 393, 512, 58 },
		{ 394, 477, 58 },
		{ 394, 478, 58 },
		{ 394, 479, 58 },
		{ 394, 480, 58 },
		{ 394, 481, 58 },
		{ 394, 482, 58 },
		{ 394, 483, 58 },
		{ 394, 484, 58 },
		{ 394, 485, 58 },
		{ 394, 486, 58 },
		{ 394, 487, 58 },
		{ 394, 488, 58 },
		{ 394, 489, 58 },
		{ 394, 490, 58 },
		{ 394, 491, 58 },
		{ 394, 492, 58 },
		{ 394, 493, 58 },
		{ 394, 494, 58 },
		{ 394, 495, 58 },
		{ 394, 496, 58 },
		{ 394, 497, 58 },
		{ 394, 498, 58 },
		{ 394, 499, 58 },
		{ 394, 500, 58 },
		{ 394, 501, 58 },
		{ 394, 502, 58 },
		{ 394, 503, 58 },
		{ 394, 504, 58 },
		{ 394, 505, 58 },
		{ 394, 506, 58 },
		{ 394, 507, 58 },
		{ 394, 508, 58 },
		{ 394, 509, 58 },
		{ 394, 510, 58 },
		{ 394, 511, 58 },
		{ 394, 512, 58 },
		{ 395, 478, 58 },
		{ 395, 479, 58 },
		{ 395, 480, 58 },
		{ 395, 481, 58 },
		{ 395, 482, 58 },
		{ 395, 483, 58 },
		{ 395, 484, 58 },
		{ 395, 485, 58 },
		{ 395, 486, 58 },
		{ 395, 487, 58 },
		{ 395, 488, 58 },
		{ 395, 489, 58 },
		{ 395, 490, 58 },
		{ 395, 491, 58 },
		{ 395, 492, 58 },
		{ 395, 493, 58 },
		{ 395, 494, 58 },
		{ 395, 495, 58 },
		{ 395, 496, 58 },
		{ 395, 497, 58 },
		{ 395, 498, 58 },
		{ 395, 499, 58 },
		{ 395, 500, 58 },
		{ 395, 501, 58 },
		{ 395, 502, 58 },
		{ 395, 503, 58 },
		{ 395, 504, 58 },
		{ 395, 505, 58 },
		{ 395, 506, 58 },
		{ 395, 507, 58 },
		{ 395, 508, 58 },
		{ 395, 509, 58 },
		{ 395, 510, 58 },
		{ 395, 511, 58 },
		{ 395, 512, 58 },
		{ 396, 480, 58 },
		{ 396, 481, 58 },
		{ 396, 482, 58 },
		{ 396, 483, 58 },
		{ 396, 484, 58 },
		{ 396, 485, 58 },
		{ 396, 486, 58 },
		{ 396, 487, 58 },
		{ 396, 488, 58 },
		{ 396, 489, 58 },
		{ 396, 490, 58 },
		{ 396, 491, 58 },
		{ 396, 492, 58 },
		{ 396, 493, 58 },
		{ 396, 494, 58 },
		{ 396, 495, 58 },
		{ 396, 496, 58 },
		{ 396, 497, 58 },
		{ 396, 498, 58 },
		{ 396, 499, 58 },
		{ 396, 500, 58 },
		{ 396, 501, 58 },
		{ 396, 502, 58 },
		{ 396, 503, 58 },
		{ 396, 504, 58 },
		{ 396, 505, 58 },
		{ 396, 506, 58 },
		{ 396, 507, 58 },
		{ 396, 508, 58 },
		{ 396, 509, 58 },
		{ 396, 510, 58 },
		{ 396, 511, 58 },
		{ 396, 512, 58 },
		{ 397, 482, 58 },
		{ 397, 483, 58 },
		{ 397, 484, 58 },
		{ 397, 485, 58 },
		{ 397, 486, 58 },
		{ 397, 487, 58 },
		{ 397, 488, 58 },
		{ 397, 489, 58 },
		{ 397, 490, 58 },
		{ 397, 491, 58 },
		{ 397, 492, 58 },
		{ 397, 493, 58 },
		{ 397, 494, 58 },
		{ 397, 495, 58 },
		{ 397, 496, 58 },
		{ 397, 497, 58 },
		{ 397, 498, 58 },
		{ 397, 499, 58 },
		{ 397, 500, 58 },
		{ 397, 501, 58 },
		{ 397, 502, 58 },
		{ 397, 503, 58 },
		{ 397, 504, 58 },
		{ 397, 505, 58 },
		{ 397, 506, 58 },
		{ 397, 507, 58 },
		{ 397, 508, 58 },
		{ 397, 509, 58 },
		{ 397, 510, 58 },
		{ 397, 511, 58 },
		{ 397, 512, 58 },
		{ 398, 483, 58 },
		{ 398, 484, 58 },
		{ 398, 485, 58 },
		{ 398, 486, 58 },
		{ 398, 487, 58 },
		{ 398, 488, 58 },
		{ 398, 489, 58 },
		{ 398, 490, 58 },
		{ 398, 491, 58 },
		{ 398, 492, 58 },
		{ 398, 493, 58 },
		{ 398, 494, 58 },
		{ 398, 495, 58 },
		{ 398, 496, 58 },
		{ 398, 497, 58 },
		{ 398, 498, 58 },
		{ 398, 499, 58 },
		{ 398, 500, 58 },
		{ 398, 501, 58 },
		{ 398, 502, 58 },
		{ 398, 503, 58 },
		{ 398, 504, 58 },
		{ 398, 505, 58 },
		{ 398, 506, 58 },
		{ 398, 507, 58 },
		{ 398, 508, 58 },
		{ 398, 509, 58 },
		{ 398, 510, 58 },
		{ 398, 511, 58 },
		{ 398, 512, 58 },
		{ 399, 485, 58 },
		{ 399, 486, 58 },
		{ 399, 487, 58 },
		{ 399, 488, 58 },
		{ 399, 489, 58 },
		{ 399, 490, 58 },
		{ 399, 491, 58 },
		{ 399, 492, 58 },
		{ 399, 493, 58 },
		{ 399, 494, 58 },
		{ 399, 495, 58 },
		{ 399, 496, 58 },
		{ 399, 497, 58 },
		{ 399, 498, 58 },
		{ 399, 499, 58 },
		{ 399, 500, 58 },
		{ 399, 501, 58 },
		{ 399, 502, 58 },
		{ 399, 503, 58 },
		{ 399, 504, 58 },
		{ 399, 505, 58 },
		{ 399, 506, 58 },
		{ 399, 507, 58 },
		{ 399, 508, 58 },
		{ 399, 509, 58 },
		{ 399, 510, 58 },
		{ 399, 511, 58 },
		{ 399, 512, 58 },
		{ 400, 486, 58 },
		{ 400, 487, 58 },
		{ 400, 488, 58 },
		{ 400, 489, 58 },
		{ 400, 490, 58 },
		{ 400, 491, 58 },
		{ 400, 492, 58 },
		{ 400, 493, 58 },
		{ 400, 494, 58 },
		{ 400, 495, 58 },
		{ 400, 496, 58 },
		{ 400, 497, 58 },
		{ 400, 498, 58 },
		{ 400, 499, 58 },
		{ 400, 500, 58 },
		{ 400, 501, 58 },
		{ 400, 502, 58 },
		{ 400, 503, 58 },
		{ 400, 504, 58 },
		{ 400, 505, 58 },
		{ 400, 506, 58 },
		{ 400, 507, 58 },
		{ 400, 508, 58 },
		{ 400, 509, 58 },
		{ 400, 510, 58 },
		{ 400, 511, 58 },
		{ 400, 512, 58 },
		{ 401, 488, 58 },
		{ 401, 489, 58 },
		{ 401, 490, 58 },
		{ 401, 491, 58 },
		{ 401, 492, 58 },
		{ 401, 493, 58 },
		{ 401, 494, 58 },
		{ 401, 495, 58 },
		{ 401, 496, 58 },
		{ 401, 497, 58 },
		{ 401, 498, 58 },
		{ 401, 499, 58 },
		{ 401, 500, 58 },
		{ 401, 501, 58 },
		{ 401, 502, 58 },
		{ 401, 503, 58 },
		{ 401, 504, 58 },
		{ 401, 505, 58 },
		{ 401, 506, 58 },
		{ 401, 507, 58 },
		{ 401, 508, 58 },
		{ 401, 509, 58 },
		{ 401, 510, 58 },
		{ 401, 511, 58 },
		{ 401, 512, 58 },
		{ 402, 490, 58 },
		{ 402, 491, 58 },
		{ 402, 492, 58 },
		{ 402, 493, 58 },
		{ 402, 494, 58 },
		{ 402, 495, 58 },
		{ 402, 496, 58 },
		{ 402, 497, 58 },
		{ 402, 498, 58 },
		{ 402, 499, 58 },
		{ 402, 500, 58 },
		{ 402, 501, 58 },
		{ 402, 502, 58 },
		{ 402, 503, 58 },
		{ 402, 504, 58 },
		{ 402, 505, 58 },
		{ 402, 506, 58 },
		{ 402, 507, 58 },
		{ 402, 508, 58 },
		{ 402, 509, 58 },
		{ 402, 510, 58 },
		{ 402, 511, 58 },
		{ 402, 512, 58 },
		{ 403, 491, 58 },
		{ 403, 492, 58 },
		{ 403, 493, 58 },
		{ 403, 494, 58 },
		{ 403, 495, 58 },
		{ 403, 496, 58 },
		{ 403, 497, 58 },
		{ 403, 498, 58 },
		{ 403, 499, 58 },
		{ 403, 500, 58 },
		{ 403, 501, 58 },
		{ 403, 502, 58 },
		{ 403, 503, 58 },
		{ 403, 504, 58 },
		{ 403, 505, 58 },
		{ 403, 506, 58 },
		{ 403, 507, 58 },
		{ 403, 508, 58 },
		{ 403, 509, 58 },
		{ 403, 510, 58 },
		{ 403, 511, 58 },
		{ 403, 512, 58 },
		{ 404, 493, 58 },
		{ 404, 494, 58 },
		{ 404, 495, 58 },
		{ 404, 496, 58 },
		{ 404, 497, 58 },
		{ 404, 498, 58 },
		{ 404, 499, 58 },
		{ 404, 500, 58 },
		{ 404, 501, 58 },
		{ 404, 502, 58 },
		{ 404, 503, 58 },
		{ 404, 504, 58 },
		{ 404, 505, 58 },
		{ 404, 506, 58 },
		{ 404, 507, 58 },
		{ 404, 508, 58 },
		{ 404, 509, 58 },
		{ 404, 510, 58 },
		{ 404, 511, 58 },
		{ 404, 512, 58 },
		{ 405, 494, 58 },
		{ 405, 495, 58 },
		{ 405, 496, 58 },
		{ 405, 497, 58 },
		{ 405, 498, 58 },
		{ 405, 499, 58 },
		{ 405, 500, 58 },
		{ 405, 501, 58 },
		{ 405, 502, 58 },
		{ 405, 503, 58 },
		{ 405, 504, 58 },
		{ 405, 505, 58 },
		{ 405, 506, 58 },
		{ 405, 507, 58 },
		{ 405, 508, 58 },
		{ 405, 509, 58 },
		{ 405, 510, 58 },
		{ 405, 511, 58 },
		{ 405, 512, 58 },
		{ 406, 496, 58 },
		{ 406, 497, 58 },
		{ 406, 498, 58 },
		{ 406, 499, 58 },
		{ 406, 500, 58 },
		{ 406, 501, 58 },
		{ 406, 502, 58 },
		{ 406, 503, 58 },
		{ 406, 504, 58 },
		{ 406, 505, 58 },
		{ 406, 506, 58 },
		{ 406, 507, 58 },
		{ 406, 508, 58 },
		{ 406, 509, 58 },
		{ 406, 510, 58 },
		{ 406, 511, 58 },
		{ 406, 512, 58 },
		{ 407, 498, 58 },
		{ 407, 499, 58 },
		{ 407, 500, 58 },
		{ 407, 501, 58 },
		{ 407, 502, 58 },
		{ 407, 503, 58 },
		{ 407, 504, 58 },
		{ 407, 505, 58 },
		{ 407, 506, 58 },
		{ 407, 507, 58 },
		{ 407, 508, 58 },
		{ 407, 509, 58 },
		{ 407, 510, 58 },
		{ 407, 511, 58 },
		{ 407, 512, 58 },
		{ 408, 499, 58 },
		{ 408, 500, 58 },
		{ 408, 501, 58 },
		{ 408, 502, 58 },
		{ 408, 503, 58 },
		{ 408, 504, 58 },
		{ 408, 505, 58 },
		{ 408, 506, 58 },
		{ 408, 507, 58 },
		{ 408, 508, 58 },
		{ 408, 509, 58 },
		{ 408, 510, 58 },
		{ 408, 511, 58 },
		{ 408, 512, 58 },
		{ 409, 501, 58 },
		{ 409, 502, 58 },
		{ 409, 503, 58 },
		{ 409, 504, 58 },
		{ 409, 505, 58 },
		{ 409, 506, 58 },
		{ 409, 507, 58 },
		{ 409, 508, 58 },
		{ 409, 509, 58 },
		{ 409, 510, 58 },
		{ 409, 511, 58 },
		{ 409, 512, 58 },
		{ 410, 502, 58 },
		{ 410, 503, 58 },
		{ 410, 504, 58 },
		{ 410, 505, 58 },
		{ 410, 506, 58 },
		{ 410, 507, 58 },
		{ 410, 508, 58 },
		{ 410, 509, 58 },
		{ 410, 510, 58 },
		{ 410, 511, 58 },
		{ 410, 512, 58 },
		{ 411, 504, 58 },
		{ 411, 505, 58 },
		{ 411, 506, 58 },
		{ 411, 507, 58 },
		{ 411, 508, 58 },
		{ 411, 509, 58 },
		{ 411, 510, 58 },
		{ 411, 511, 58 },
		{ 411, 512, 58 },
		{ 412, 506, 58 },
		{ 412, 507, 58 },
		{ 412, 508, 58 },
		{ 412, 509, 58 },
		{ 412, 510, 58 },
		{ 412, 511, 58 },
		{ 412, 512, 58 },
		{ 413, 507, 58 },
		{ 413, 508, 58 },
		{ 413, 509, 58 },
		{ 413, 510, 58 },
		{ 413, 511, 58 },
		{ 413, 512, 58 },
		{ 414, 509, 58 },
		{ 414, 510, 58 },
		{ 414, 511, 58 },
		{ 414, 512, 58 },
		{ 415, 510, 58 },
		{ 415, 511, 58 },
		{ 415, 512, 58 },
		{ 416, 512, 58 },
		{ 130, 1, 59 },
		{ 130, 2, 59 },
		{ 131, 1, 59 },
		{ 131, 2, 59 },
		{ 131, 3, 59 },
		{ 131, 4, 59 },
		{ 132, 1, 59 },
		{ 132, 2, 59 },
		{ 132, 3, 59 },
		{ 132, 4, 59 },
		{ 132, 5, 59 },
		{ 132, 6, 59 },
		{ 133, 1, 59 },
		{ 133, 2, 59 },
		{ 133, 3, 59 },
		{ 133, 4, 59 },
		{ 133, 5, 59 },
		{ 133, 6, 59 },
		{ 133, 7, 59 },
		{ 133, 8, 59 },
		{ 134, 1, 59 },
		{ 134, 2, 59 },
		{ 134, 3, 59 },
		{ 134, 4, 59 },
		{ 134, 5, 59 },
		{ 134, 6, 59 },
		{ 134, 7, 59 },
		{ 134, 8, 59 },
		{ 134, 9, 59 },
		{ 134, 10, 59 },
		{ 135, 1, 59 },
		{ 135, 2, 59 },
		{ 135, 3, 59 },
		{ 135, 4, 59 },
		{ 135, 5, 59 },
		{ 135, 6, 59 },
		{ 135, 7, 59 },
		{ 135, 8, 59 },
		{ 135, 9, 59 },
		{ 135, 10, 59 },
		{ 135, 11, 59 },
		{ 135, 12, 59 },
		{ 136, 1, 59 },
		{ 136, 2, 59 },
		{ 136, 3, 59 },
		{ 136, 4, 59 },
		{ 136, 5, 59 },
		{ 136, 6, 59 },
		{ 136, 7, 59 },
		{ 136, 8, 59 },
		{ 136, 9, 59 },
		{ 136, 10, 59 },
		{ 136, 11, 59 },
		{ 136, 12, 59 },
		{ 136, 13, 59 },
		{ 136, 14, 59 },
		{ 137, 1, 59 },
		{ 137, 2, 59 },
		{ 137, 3, 59 },
		{ 137, 4, 59 },
		{ 137, 5, 59 },
		{ 137, 6, 59 },
		{ 137, 7, 59 },
		{ 137, 8, 59 },
		{ 137, 9, 59 },
		{ 137, 10, 59 },
		{ 137, 11, 59 },
		{ 137, 12, 59 },
		{ 137, 13, 59 },
		{ 137, 14, 59 },
		{ 137, 15, 59 },
		{ 137, 16, 59 },
		{ 138, 1, 59 },
		{ 138, 2, 59 },
		{ 138, 3, 59 },
		{ 138, 4, 59 },
		{ 138, 5, 59 },
		{ 138, 6, 59 },
		{ 138, 7, 59 },
		{ 138, 8, 59 },
		{ 138, 9, 59 },
		{ 138, 10, 59 },
		{ 138, 11, 59 },
		{ 138, 12, 59 },
		{ 138, 13, 59 },
		{ 138, 14, 59 },
		{ 138, 15, 59 },
		{ 138, 16, 59 },
		{ 138, 17, 59 },
		{ 138, 18, 59 },
		{ 139, 1, 59 },
		{ 139, 2, 59 },
		{ 139, 3, 59 },
		{ 139, 4, 59 },
		{ 139, 5, 59 },
		{ 139, 6, 59 },
		{ 139, 7, 59 },
		{ 139, 8, 59 },
		{ 139, 9, 59 },
		{ 139, 10, 59 },
		{ 139, 11, 59 },
		{ 139, 12, 59 },
		{ 139, 13, 59 },
		{ 139, 14, 59 },
		{ 139, 15, 59 },
		{ 139, 16, 59 },
		{ 139, 17, 59 },
		{ 139, 18, 59 },
		{ 139, 19, 59 },
		{ 139, 20, 59 },
		{ 140, 1, 59 },
		{ 140, 2, 59 },
		{ 140, 3, 59 },
		{ 140, 4, 59 },
		{ 140, 5, 59 },
		{ 140, 6, 59 },
		{ 140, 7, 59 },
		{ 140, 8, 59 },
		{ 140, 9, 59 },
		{ 140, 10, 59 },
		{ 140, 11, 59 },
		{ 140, 12, 59 },
		{ 140, 13, 59 },
		{ 140, 14, 59 },
		{ 140, 15, 59 },
		{ 140, 16, 59 },
		{ 140, 17, 59 },
		{ 140, 18, 59 },
		{ 140, 19, 59 },
		{ 140, 20, 59 },
		{ 140, 21, 59 },
		{ 140, 22, 59 },
		{ 141, 1, 59 },
		{ 141, 2, 59 },
		{ 141, 3, 59 },
		{ 141, 4, 59 },
		{ 141, 5, 59 },
		{ 141, 6, 59 },
		{ 141, 7, 59 },
		{ 141, 8, 59 },
		{ 141, 9, 59 },
		{ 141, 10, 59 },
		{ 141, 11, 59 },
		{ 141, 12, 59 },
		{ 141, 13, 59 },
		{ 141, 14, 59 },
		{ 141, 15, 59 },
		{ 141, 16, 59 },
		{ 141, 17, 59 },
		{ 141, 18, 59 },
		{ 141, 19, 59 },
		{ 141, 20, 59 },
		{ 141, 21, 59 },
		{ 141, 22, 59 },
		{ 141, 23, 59 },
		{ 141, 24, 59 },
		{ 142, 1, 59 },
		{ 142, 2, 59 },
		{ 142, 3, 59 },
		{ 142, 4, 59 },
		{ 142, 5, 59 },
		{ 142, 6, 59 },
		{ 142, 7, 59 },
		{ 142, 8, 59 },
		{ 142, 9, 59 },
		{ 142, 10, 59 },
		{ 142, 11, 59 },
		{ 142, 12, 59 },
		{ 142, 13, 59 },
		{ 142, 14, 59 },
		{ 142, 15, 59 },
		{ 142, 16, 59 },
		{ 142, 17, 59 },
		{ 142, 18, 59 },
		{ 142, 19, 59 },
		{ 142, 20, 59 },
		{ 142, 21, 59 },
		{ 142, 22, 59 },
		{ 142, 23, 59 },
		{ 142, 24, 59 },
		{ 142, 25, 59 },
		{ 142, 26, 59 },
		{ 143, 1, 59 },
		{ 143, 2, 59 },
		{ 143, 3, 59 },
		{ 143, 4, 59 },
		{ 143, 5, 59 },
		{ 143, 6, 59 },
		{ 143, 7, 59 },
		{ 143, 8, 59 },
		{ 143, 9, 59 },
		{ 143, 10, 59 },
		{ 143, 11, 59 },
		{ 143, 12, 59 },
		{ 143, 13, 59 },
		{ 143, 14, 59 },
		{ 143, 15, 59 },
		{ 143, 16, 59 },
		{ 143, 17, 59 },
		{ 143, 18, 59 },
		{ 143, 19, 59 },
		{ 143, 20, 59 },
		{ 143, 21, 59 },
		{ 143, 22, 59 },
		{ 143, 23, 59 },
		{ 143, 24, 59 },
		{ 143, 25, 59 },
		{ 143, 26, 59 },
		{ 143, 27, 59 },
		{ 143, 28, 59 },
		{ 144, 1, 59 },
		{ 144, 2, 59 },
		{ 144, 3, 59 },
		{ 144, 4, 59 },
		{ 144, 5, 59 },
		{ 144, 6, 59 },
		{ 144, 7, 59 },
		{ 144, 8, 59 },
		{ 144, 9, 59 },
		{ 144, 10, 59 },
		{ 144, 11, 59 },
		{ 144, 12, 59 },
		{ 144, 13, 59 },
		{ 144, 14, 59 },
		{ 144, 15, 59 },
		{ 144, 16, 59 },
		{ 144, 17, 59 },
		{ 144, 18, 59 },
		{ 144, 19, 59 },
		{ 144, 20, 59 },
		{ 144, 21, 59 },
		{ 144, 22, 59 },
		{ 144, 23, 59 },
		{ 144, 24, 59 },
		{ 144, 25, 59 },
		{ 144, 26, 59 },
		{ 144, 27, 59 },
		{ 144, 28, 59 },
		{ 144, 29, 59 },
		{ 144, 30, 59 },
		{ 145, 1, 59 },
		{ 145, 2, 59 },
		{ 145, 3, 59 },
		{ 145, 4, 59 },
		{ 145, 5, 59 },
		{ 145, 6, 59 },
		{ 145, 7, 59 },
		{ 145, 8, 59 },
		{ 145, 9, 59 },
		{ 145, 10, 59 },
		{ 145, 11, 59 },
		{ 145, 12, 59 },
		{ 145, 13, 59 },
		{ 145, 14, 59 },
		{ 145, 15, 59 },
		{ 145, 16, 59 },
		{ 145, 17, 59 },
		{ 145, 18, 59 },
		{ 145, 19, 59 },
		{ 145, 20, 59 },
		{ 145, 21, 59 },
		{ 145, 22, 59 },
		{ 145, 23, 59 },
		{ 145, 24, 59 },
		{ 145, 25, 59 },
		{ 145, 26, 59 },
		{ 145, 27, 59 },
		{ 145, 28, 59 },
		{ 145, 29, 59 },
		{ 145, 30, 59 },
		{ 145, 31, 59 },
		{ 145, 32, 59 },
		{ 146, 1, 59 },
		{ 146, 2, 59 },
		{ 146, 3, 59 },
		{ 146, 4, 59 },
		{ 146, 5, 59 },
		{ 146, 6, 59 },
		{ 146, 7, 59 },
		{ 146, 8, 59 },
		{ 146, 9, 59 },
		{ 146, 10, 59 },
		{ 146, 11, 59 },
		{ 146, 12, 59 },
		{ 146, 13, 59 },
		{ 146, 14, 59 },
		{ 146, 15, 59 },
		{ 146, 16, 59 },
		{ 146, 17, 59 },
		{ 146, 18, 59 },
		{ 146, 19, 59 },
		{ 146, 20, 59 },
		{ 146, 21, 59 },
		{ 146, 22, 59 },
		{ 146, 23, 59 },
		{ 146, 24, 59 },
		{ 146, 25, 59 },
		{ 146, 26, 59 },
		{ 146, 27, 59 },
		{ 146, 28, 59 },
		{ 146, 29, 59 },
		{ 146, 30, 59 },
		{ 146, 31, 59 },
		{ 146, 32, 59 },
		{ 146, 33, 59 },
		{ 146, 34, 59 },
		{ 147, 1, 59 },
		{ 147, 2, 59 },
		{ 147, 3, 59 },
		{ 147, 4, 59 },
		{ 147, 5, 59 },
		{ 147, 6, 59 },
		{ 147, 7, 59 },
		{ 147, 8, 59 },
		{ 147, 9, 59 },
		{ 147, 10, 59 },
		{ 147, 11, 59 },
		{ 147, 12, 59 },
		{ 147, 13, 59 },
		{ 147, 14, 59 },
		{ 147, 15, 59 },
		{ 147, 16, 59 },
		{ 147, 17, 59 },
		{ 147, 18, 59 },
		{ 147, 19, 59 },
		{ 147, 20, 59 },
		{ 147, 21, 59 },
		{ 147, 22, 59 },
		{ 147, 23, 59 },
		{ 147, 24, 59 },
		{ 147, 25, 59 },
		{ 147, 26, 59 },
		{ 147, 27, 59 },
		{ 147, 28, 59 },
		{ 147, 29, 59 },
		{ 147, 30, 59 },
		{ 147, 31, 59 },
		{ 147, 32, 59 },
		{ 147, 33, 59 },
		{ 147, 34, 59 },
		{ 147, 35, 59 },
		{ 147, 36, 59 },
		{ 148, 1, 59 },
		{ 148, 2, 59 },
		{ 148, 3, 59 },
		{ 148, 4, 59 },
		{ 148, 5, 59 },
		{ 148, 6, 59 },
		{ 148, 7, 59 },
		{ 148, 8, 59 },
		{ 148, 9, 59 },
		{ 148, 10, 59 },
		{ 148, 11, 59 },
		{ 148, 12, 59 },
		{ 148, 13, 59 },
		{ 148, 14, 59 },
		{ 148, 15, 59 },
		{ 148, 16, 59 },
		{ 148, 17, 59 },
		{ 148, 18, 59 },
		{ 148, 19, 59 },
		{ 148, 20, 59 },
		{ 148, 21, 59 },
		{ 148, 22, 59 },
		{ 148, 23, 59 },
		{ 148, 24, 59 },
		{ 148, 25, 59 },
		{ 148, 26, 59 },
		{ 148, 27, 59 },
		{ 148, 28, 59 },
		{ 148, 29, 59 },
		{ 148, 30, 59 },
		{ 148, 31, 59 },
		{ 148, 32, 59 },
		{ 148, 33, 59 },
		{ 148, 34, 59 },
		{ 148, 35, 59 },
		{ 148, 36, 59 },
		{ 148, 37, 59 },
		{ 148, 38, 59 },
		{ 149, 1, 59 },
		{ 149, 2, 59 },
		{ 149, 3, 59 },
		{ 149, 4, 59 },
		{ 149, 5, 59 },
		{ 149, 6, 59 },
		{ 149, 7, 59 },
		{ 149, 8, 59 },
		{ 149, 9, 59 },
		{ 149, 10, 59 },
		{ 149, 11, 59 },
		{ 149, 12, 59 },
		{ 149, 13, 59 },
		{ 149, 14, 59 },
		{ 149, 15, 59 },
		{ 149, 16, 59 },
		{ 149, 17, 59 },
		{ 149, 18, 59 },
		{ 149, 19, 59 },
		{ 149, 20, 59 },
		{ 149, 21, 59 },
		{ 149, 22, 59 },
		{ 149, 23, 59 },
		{ 149, 24, 59 },
		{ 149, 25, 59 },
		{ 149, 26, 59 },
		{ 149, 27, 59 },
		{ 149, 28, 59 },
		{ 149, 29, 59 },
		{ 149, 30, 59 },
		{ 149, 31, 59 },
		{ 149, 32, 59 },
		{ 149, 33, 59 },
		{ 149, 34, 59 },
		{ 149, 35, 59 },
		{ 149, 36, 59 },
		{ 149, 37, 59 },
		{ 149, 38, 59 },
		{ 149, 39, 59 },
		{ 149, 40, 59 },
		{ 150, 1, 59 },
		{ 150, 2, 59 },
		{ 150, 3, 59 },
		{ 150, 4, 59 },
		{ 150, 5, 59 },
		{ 150, 6, 59 },
		{ 150, 7, 59 },
		{ 150, 8, 59 },
		{ 150, 9, 59 },
		{ 150, 10, 59 },
		{ 150, 11, 59 },
		{ 150, 12, 59 },
		{ 150, 13, 59 },
		{ 150, 14, 59 },
		{ 150, 15, 59 },
		{ 150, 16, 59 },
		{ 150, 17, 59 },
		{ 150, 18, 59 },
		{ 150, 19, 59 },
		{ 150, 20, 59 },
		{ 150, 21, 59 },
		{ 150, 22, 59 },
		{ 150, 23, 59 },
		{ 150, 24, 59 },
		{ 150, 25, 59 },
		{ 150, 26, 59 },
		{ 150, 27, 59 },
		{ 150, 28, 59 },
		{ 150, 29, 59 },
		{ 150, 30, 59 },
		{ 150, 31, 59 },
		{ 150, 32, 59 },
		{ 150, 33, 59 },
		{ 150, 34, 59 },
		{ 150, 35, 59 },
		{ 150, 36, 59 },
		{ 150, 37, 59 },
		{ 150, 38, 59 },
		{ 150, 39, 59 },
		{ 150, 40, 59 },
		{ 150, 41, 59 },
		{ 150, 42, 59 },
		{ 151, 1, 59 },
		{ 151, 2, 59 },
		{ 151, 3, 59 },
		{ 151, 4, 59 },
		{ 151, 5, 59 },
		{ 151, 6, 59 },
		{ 151, 7, 59 },
		{ 151, 8, 59 },
		{ 151, 9, 59 },
		{ 151, 10, 59 },
		{ 151, 11, 59 },
		{ 151, 12, 59 },
		{ 151, 13, 59 },
		{ 151, 14, 59 },
		{ 151, 15, 59 },
		{ 151, 16, 59 },
		{ 151, 17, 59 },
		{ 151, 18, 59 },
		{ 151, 19, 59 },
		{ 151, 20, 59 },
		{ 151, 21, 59 },
		{ 151, 22, 59 },
		{ 151, 23, 59 },
		{ 151, 24, 59 },
		{ 151, 25, 59 },
		{ 151, 26, 59 },
		{ 151, 27, 59 },
		{ 151, 28, 59 },
		{ 151, 29, 59 },
		{ 151, 30, 59 },
		{ 151, 31, 59 },
		{ 151, 32, 59 },
		{ 151, 33, 59 },
		{ 151, 34, 59 },
		{ 151, 35, 59 },
		{ 151, 36, 59 },
		{ 151, 37, 59 },
		{ 151, 38, 59 },
		{ 151, 39, 59 },
		{ 151, 40, 59 },
		{ 151, 41, 59 },
		{ 151, 42, 59 },
		{ 151, 43, 59 },
		{ 151, 44, 59 },
		{ 152, 1, 59 },
		{ 152, 2, 59 },
		{ 152, 3, 59 },
		{ 152, 4, 59 },
		{ 152, 5, 59 },
		{ 152, 6, 59 },
		{ 152, 7, 59 },
		{ 152, 8, 59 },
		{ 152, 9, 59 },
		{ 152, 10, 59 },
		{ 152, 11, 59 },
		{ 152, 12, 59 },
		{ 152, 13, 59 },
		{ 152, 14, 59 },
		{ 152, 15, 59 },
		{ 152, 16, 59 },
		{ 152, 17, 59 },
		{ 152, 18, 59 },
		{ 152, 19, 59 },
		{ 152, 20, 59 },
		{ 152, 21, 59 },
		{ 152, 22, 59 },
		{ 152, 23, 59 },
		{ 152, 24, 59 },
		{ 152, 25, 59 },
		{ 152, 26, 59 },
		{ 152, 27, 59 },
		{ 152, 28, 59 },
		{ 152, 29, 59 },
		{ 152, 30, 59 },
		{ 152, 31, 59 },
		{ 152, 32, 59 },
		{ 152, 33, 59 },
		{ 152, 34, 59 },
		{ 152, 35, 59 },
		{ 152, 36, 59 },
		{ 152, 37, 59 },
		{ 152, 38, 59 },
		{ 152, 39, 59 },
		{ 152, 40, 59 },
		{ 152, 41, 59 },
		{ 152, 42, 59 },
		{ 152, 43, 59 },
		{ 152, 44, 59 },
		{ 152, 45, 59 },
		{ 152, 46, 59 },
		{ 153, 1, 59 },
		{ 153, 2, 59 },
		{ 153, 3, 59 },
		{ 153, 4, 59 },
		{ 153, 5, 59 },
		{ 153, 6, 59 },
		{ 153, 7, 59 },
		{ 153, 8, 59 },
		{ 153, 9, 59 },
		{ 153, 10, 59 },
		{ 153, 11, 59 },
		{ 153, 12, 59 },
		{ 153, 13, 59 },
		{ 153, 14, 59 },
		{ 153, 15, 59 },
		{ 153, 16, 59 },
		{ 153, 17, 59 },
		{ 153, 18, 59 },
		{ 153, 19, 59 },
		{ 153, 20, 59 },
		{ 153, 21, 59 },
		{ 153, 22, 59 },
		{ 153, 23, 59 },
		{ 153, 24, 59 },
		{ 153, 25, 59 },
		{ 153, 26, 59 },
		{ 153, 27, 59 },
		{ 153, 28, 59 },
		{ 153, 29, 59 },
		{ 153, 30, 59 },
		{ 153, 31, 59 },
		{ 153, 32, 59 },
		{ 153, 33, 59 },
		{ 153, 34, 59 },
		{ 153, 35, 59 },
		{ 153, 36, 59 },
		{ 153, 37, 59 },
		{ 153, 38, 59 },
		{ 153, 39, 59 },
		{ 153, 40, 59 },
		{ 153, 41, 59 },
		{ 153, 42, 59 },
		{ 153, 43, 59 },
		{ 153, 44, 59 },
		{ 153, 45, 59 },
		{ 153, 46, 59 },
		{ 153, 47, 59 },
		{ 153, 48, 59 },
		{ 154, 1, 59 },
		{ 154, 2, 59 },
		{ 154, 3, 59 },
		{ 154, 4, 59 },
		{ 154, 5, 59 },
		{ 154, 6, 59 },
		{ 154, 7, 59 },
		{ 154, 8, 59 },
		{ 154, 9, 59 },
		{ 154, 10, 59 },
		{ 154, 11, 59 },
		{ 154, 12, 59 },
		{ 154, 13, 59 },
		{ 154, 14, 59 },
		{ 154, 15, 59 },
		{ 154, 16, 59 },
		{ 154, 17, 59 },
		{ 154, 18, 59 },
		{ 154, 19, 59 },
		{ 154, 20, 59 },
		{ 154, 21, 59 },
		{ 154, 22, 59 },
		{ 154, 23, 59 },
		{ 154, 24, 59 },
		{ 154, 25, 59 },
		{ 154, 26, 59 },
		{ 154, 27, 59 },
		{ 154, 28, 59 },
		{ 154, 29, 59 },
		{ 154, 30, 59 },
		{ 154, 31, 59 },
		{ 154, 32, 59 },
		{ 154, 33, 59 },
		{ 154, 34, 59 },
		{ 154, 35, 59 },
		{ 154, 36, 59 },
		{ 154, 37, 59 },
		{ 154, 38, 59 },
		{ 154, 39, 59 },
		{ 154, 40, 59 },
		{ 154, 41, 59 },
		{ 154, 42, 59 },
		{ 154, 43, 59 },
		{ 154, 44, 59 },
		{ 154, 45, 59 },
		{ 154, 46, 59 },
		{ 154, 47, 59 },
		{ 154, 48, 59 },
		{ 154, 49, 59 },
		{ 154, 50, 59 },
		{ 155, 1, 59 },
		{ 155, 2, 59 },
		{ 155, 3, 59 },
		{ 155, 4, 59 },
		{ 155, 5, 59 },
		{ 155, 6, 59 },
		{ 155, 7, 59 },
		{ 155, 8, 59 },
		{ 155, 9, 59 },
		{ 155, 10, 59 },
		{ 155, 11, 59 },
		{ 155, 12, 59 },
		{ 155, 13, 59 },
		{ 155, 14, 59 },
		{ 155, 15, 59 },
		{ 155, 16, 59 },
		{ 155, 17, 59 },
		{ 155, 18, 59 },
		{ 155, 19, 59 },
		{ 155, 20, 59 },
		{ 155, 21, 59 },
		{ 155, 22, 59 },
		{ 155, 23, 59 },
		{ 155, 24, 59 },
		{ 155, 25, 59 },
		{ 155, 26, 59 },
		{ 155, 27, 59 },
		{ 155, 28, 59 },
		{ 155, 29, 59 },
		{ 155, 30, 59 },
		{ 155, 31, 59 },
		{ 155, 32, 59 },
		{ 155, 33, 59 },
		{ 155, 34, 59 },
		{ 155, 35, 59 },
		{ 155, 36, 59 },
		{ 155, 37, 59 },
		{ 155, 38, 59 },
		{ 155, 39, 59 },
		{ 155, 40, 59 },
		{ 155, 41, 59 },
		{ 155, 42, 59 },
		{ 155, 43, 59 },
		{ 155, 44, 59 },
		{ 155, 45, 59 },
		{ 155, 46, 59 },
		{ 155, 47, 59 },
		{ 155, 48, 59 },
		{ 155, 49, 59 },
		{ 155, 50, 59 },
		{ 155, 51, 59 },
		{ 155, 52, 59 },
		{ 156, 1, 59 },
		{ 156, 2, 59 },
		{ 156, 3, 59 },
		{ 156, 4, 59 },
		{ 156, 5, 59 },
		{ 156, 6, 59 },
		{ 156, 7, 59 },
		{ 156, 8, 59 },
		{ 156, 9, 59 },
		{ 156, 10, 59 },
		{ 156, 11, 59 },
		{ 156, 12, 59 },
		{ 156, 13, 59 },
		{ 156, 14, 59 },
		{ 156, 15, 59 },
		{ 156, 16, 59 },
		{ 156, 17, 59 },
		{ 156, 18, 59 },
		{ 156, 19, 59 },
		{ 156, 20, 59 },
		{ 156, 21, 59 },
		{ 156, 22, 59 },
		{ 156, 23, 59 },
		{ 156, 24, 59 },
		{ 156, 25, 59 },
		{ 156, 26, 59 },
		{ 156, 27, 59 },
		{ 156, 28, 59 },
		{ 156, 29, 59 },
		{ 156, 30, 59 },
		{ 156, 31, 59 },
		{ 156, 32, 59 },
		{ 156, 33, 59 },
		{ 156, 34, 59 },
		{ 156, 35, 59 },
		{ 156, 36, 59 },
		{ 156, 37, 59 },
		{ 156, 38, 59 },
		{ 156, 39, 59 },
		{ 156, 40, 59 },
		{ 156, 41, 59 },
		{ 156, 42, 59 },
		{ 156, 43, 59 },
		{ 156, 44, 59 },
		{ 156, 45, 59 },
		{ 156, 46, 59 },
		{ 156, 47, 59 },
		{ 156, 48, 59 },
		{ 156, 49, 59 },
		{ 156, 50, 59 },
		{ 156, 51, 59 },
		{ 156, 52, 59 },
		{ 156, 53, 59 },
		{ 156, 54, 59 },
		{ 157, 1, 59 },
		{ 157, 2, 59 },
		{ 157, 3, 59 },
		{ 157, 4, 59 },
		{ 157, 5, 59 },
		{ 157, 6, 59 },
		{ 157, 7, 59 },
		{ 157, 8, 59 },
		{ 157, 9, 59 },
		{ 157, 10, 59 },
		{ 157, 11, 59 },
		{ 157, 12, 59 },
		{ 157, 13, 59 },
		{ 157, 14, 59 },
		{ 157, 15, 59 },
		{ 157, 16, 59 },
		{ 157, 17, 59 },
		{ 157, 18, 59 },
		{ 157, 19, 59 },
		{ 157, 20, 59 },
		{ 157, 21, 59 },
		{ 157, 22, 59 },
		{ 157, 23, 59 },
		{ 157, 24, 59 },
		{ 157, 25, 59 },
		{ 157, 26, 59 },
		{ 157, 27, 59 },
		{ 157, 28, 59 },
		{ 157, 29, 59 },
		{ 157, 30, 59 },
		{ 157, 31, 59 },
		{ 157, 32, 59 },
		{ 157, 33, 59 },
		{ 157, 34, 59 },
		{ 157, 35, 59 },
		{ 157, 36, 59 },
		{ 157, 37, 59 },
		{ 157, 38, 59 },
		{ 157, 39, 59 },
		{ 157, 40, 59 },
		{ 157, 41, 59 },
		{ 157, 42, 59 },
		{ 157, 43, 59 },
		{ 157, 44, 59 },
		{ 157, 45, 59 },
		{ 157, 46, 59 },
		{ 157, 47, 59 },
		{ 157, 48, 59 },
		{ 157, 49, 59 },
		{ 157, 50, 59 },
		{ 157, 51, 59 },
		{ 157, 52, 59 },
		{ 157, 53, 59 },
		{ 157, 54, 59 },
		{ 157, 55, 59 },
		{ 157, 56, 59 },
		{ 158, 1, 59 },
		{ 158, 2, 59 },
		{ 158, 3, 59 },
		{ 158, 4, 59 },
		{ 158, 5, 59 },
		{ 158, 6, 59 },
		{ 158, 7, 59 },
		{ 158, 8, 59 },
		{ 158, 9, 59 },
		{ 158, 10, 59 },
		{ 158, 11, 59 },
		{ 158, 12, 59 },
		{ 158, 13, 59 },
		{ 158, 14, 59 },
		{ 158, 15, 59 },
		{ 158, 16, 59 },
		{ 158, 17, 59 },
		{ 158, 18, 59 },
		{ 158, 19, 59 },
		{ 158, 20, 59 },
		{ 158, 21, 59 },
		{ 158, 22, 59 },
		{ 158, 23, 59 },
		{ 158, 24, 59 },
		{ 158, 25, 59 },
		{ 158, 26, 59 },
		{ 158, 27, 59 },
		{ 158, 28, 59 },
		{ 158, 29, 59 },
		{ 158, 30, 59 },
		{ 158, 31, 59 },
		{ 158, 32, 59 },
		{ 158, 33, 59 },
		{ 158, 34, 59 },
		{ 158, 35, 59 },
		{ 158, 36, 59 },
		{ 158, 37, 59 },
		{ 158, 38, 59 },
		{ 158, 39, 59 },
		{ 158, 40, 59 },
		{ 158, 41, 59 },
		{ 158, 42, 59 },
		{ 158, 43, 59 },
		{ 158, 44, 59 },
		{ 158, 45, 59 },
		{ 158, 46, 59 },
		{ 158, 47, 59 },
		{ 158, 48, 59 },
		{ 158, 49, 59 },
		{ 158, 50, 59 },
		{ 158, 51, 59 },
		{ 158, 52, 59 },
		{ 158, 53, 59 },
		{ 158, 54, 59 },
		{ 158, 55, 59 },
		{ 158, 56, 59 },
		{ 158, 57, 59 },
		{ 158, 58, 59 },
		{ 159, 1, 59 },
		{ 159, 2, 59 },
		{ 159, 3, 59 },
		{ 159, 4, 59 },
		{ 159, 5, 59 },
		{ 159, 6, 59 },
		{ 159, 7, 59 },
		{ 159, 8, 59 },
		{ 159, 9, 59 },
		{ 159, 10, 59 },
		{ 159, 11, 59 },
		{ 159, 12, 59 },
		{ 159, 13, 59 },
		{ 159, 14, 59 },
		{ 159, 15, 59 },
		{ 159, 16, 59 },
		{ 159, 17, 59 },
		{ 159, 18, 59 },
		{ 159, 19, 59 },
		{ 159, 20, 59 },
		{ 159, 21, 59 },
		{ 159, 22, 59 },
		{ 159, 23, 59 },
		{ 159, 24, 59 },
		{ 159, 25, 59 },
		{ 159, 26, 59 },
		{ 159, 27, 59 },
		{ 159, 28, 59 },
		{ 159, 29, 59 },
		{ 159, 30, 59 },
		{ 159, 31, 59 },
		{ 159, 32, 59 },
		{ 159, 33, 59 },
		{ 159, 34, 59 },
		{ 159, 35, 59 },
		{ 159, 36, 59 },
		{ 159, 37, 59 },
		{ 159, 38, 59 },
		{ 159, 39, 59 },
		{ 159, 40, 59 },
		{ 159, 41, 59 },
		{ 159, 42, 59 },
		{ 159, 43, 59 },
		{ 159, 44, 59 },
		{ 159, 45, 59 },
		{ 159, 46, 59 },
		{ 159, 47, 59 },
		{ 159, 48, 59 },
		{ 159, 49, 59 },
		{ 159, 50, 59 },
		{ 159, 51, 59 },
		{ 159, 52, 59 },
		{ 159, 53, 59 },
		{ 159, 54, 59 },
		{ 159, 55, 59 },
		{ 159, 56, 59 },
		{ 159, 57, 59 },
		{ 159, 58, 59 },
		{ 159, 59, 59 },
		{ 159, 60, 59 },
		{ 160, 1, 59 },
		{ 160, 2, 59 },
		{ 160, 3, 59 },
		{ 160, 4, 59 },
		{ 160, 5, 59 },
		{ 160, 6, 59 },
		{ 160, 7, 59 },
		{ 160, 8, 59 },
		{ 160, 9, 59 },
		{ 160, 10, 59 },
		{ 160, 11, 59 },
		{ 160, 12, 59 },
		{ 160, 13, 59 },
		{ 160, 14, 59 },
		{ 160, 15, 59 },
		{ 160, 16, 59 },
		{ 160, 17, 59 },
		{ 160, 18, 59 },
		{ 160, 19, 59 },
		{ 160, 20, 59 },
		{ 160, 21, 59 },
		{ 160, 22, 59 },
		{ 160, 23, 59 },
		{ 160, 24, 59 },
		{ 160, 25, 59 },
		{ 160, 26, 59 },
		{ 160, 27, 59 },
		{ 160, 28, 59 },
		{ 160, 29, 59 },
		{ 160, 30, 59 },
		{ 160, 31, 59 },
		{ 160, 32, 59 },
		{ 160, 33, 59 },
		{ 160, 34, 59 },
		{ 160, 35, 59 },
		{ 160, 36, 59 },
		{ 160, 37, 59 },
		{ 160, 38, 59 },
		{ 160, 39, 59 },
		{ 160, 40, 59 },
		{ 160, 41, 59 },
		{ 160, 42, 59 },
		{ 160, 43, 59 },
		{ 160, 44, 59 },
		{ 160, 45, 59 },
		{ 160, 46, 59 },
		{ 160, 47, 59 },
		{ 160, 48, 59 },
		{ 160, 49, 59 },
		{ 160, 50, 59 },
		{ 160, 51, 59 },
		{ 160, 52, 59 },
		{ 160, 53, 59 },
		{ 160, 54, 59 },
		{ 160, 55, 59 },
		{ 160, 56, 59 },
		{ 160, 57, 59 },
		{ 160, 58, 59 },
		{ 160, 59, 59 },
		{ 160, 60, 59 },
		{ 160, 61, 59 },
		{ 160, 62, 59 },
		{ 161, 1, 59 },
		{ 161, 2, 59 },
		{ 161, 3, 59 },
		{ 161, 4, 59 },
		{ 161, 5, 59 },
		{ 161, 6, 59 },
		{ 161, 7, 59 },
		{ 161, 8, 59 },
		{ 161, 9, 59 },
		{ 161, 10, 59 },
		{ 161, 11, 59 },
		{ 161, 12, 59 },
		{ 161, 13, 59 },
		{ 161, 14, 59 },
		{ 161, 15, 59 },
		{ 161, 16, 59 },
		{ 161, 17, 59 },
		{ 161, 18, 59 },
		{ 161, 19, 59 },
		{ 161, 20, 59 },
		{ 161, 21, 59 },
		{ 161, 22, 59 },
		{ 161, 23, 59 },
		{ 161, 24, 59 },
		{ 161, 25, 59 },
		{ 161, 26, 59 },
		{ 161, 27, 59 },
		{ 161, 28, 59 },
		{ 161, 29, 59 },
		{ 161, 30, 59 },
		{ 161, 31, 59 },
		{ 161, 32, 59 },
		{ 161, 33, 59 },
		{ 161, 34, 59 },
		{ 161, 35, 59 },
		{ 161, 36, 59 },
		{ 161, 37, 59 },
		{ 161, 38, 59 },
		{ 161, 39, 59 },
		{ 161, 40, 59 },
		{ 161, 41, 59 },
		{ 161, 42, 59 },
		{ 161, 43, 59 },
		{ 161, 44, 59 },
		{ 161, 45, 59 },
		{ 161, 46, 59 },
		{ 161, 47, 59 },
		{ 161, 48, 59 },
		{ 161, 49, 59 },
		{ 161, 50, 59 },
		{ 161, 51, 59 },
		{ 161, 52, 59 },
		{ 161, 53, 59 },
		{ 161, 54, 59 },
		{ 161, 55, 59 },
		{ 161, 56, 59 },
		{ 161, 57, 59 },
		{ 161, 58, 59 },
		{ 161, 59, 59 },
		{ 161, 60, 59 },
		{ 161, 61, 59 },
		{ 161, 62, 59 },
		{ 161, 63, 59 },
		{ 161, 64, 59 },
		{ 162, 1, 59 },
		{ 162, 2, 59 },
		{ 162, 3, 59 },
		{ 162, 4, 59 },
		{ 162, 5, 59 },
		{ 162, 6, 59 },
		{ 162, 7, 59 },
		{ 162, 8, 59 },
		{ 162, 9, 59 },
		{ 162, 10, 59 },
		{ 162, 11, 59 },
		{ 162, 12, 59 },
		{ 162, 13, 59 },
		{ 162, 14, 59 },
		{ 162, 15, 59 },
		{ 162, 16, 59 },
		{ 162, 17, 59 },
		{ 162, 18, 59 },
		{ 162, 19, 59 },
		{ 162, 20, 59 },
		{ 162, 21, 59 },
		{ 162, 22, 59 },
		{ 162, 23, 59 },
		{ 162, 24, 59 },
		{ 162, 25, 59 },
		{ 162, 26, 59 },
		{ 162, 27, 59 },
		{ 162, 28, 59 },
		{ 162, 29, 59 },
		{ 162, 30, 59 },
		{ 162, 31, 59 },
		{ 162, 32, 59 },
		{ 162, 33, 59 },
		{ 162, 34, 59 },
		{ 162, 35, 59 },
		{ 162, 36, 59 },
		{ 162, 37, 59 },
		{ 162, 38, 59 },
		{ 162, 39, 59 },
		{ 162, 40, 59 },
		{ 162, 41, 59 },
		{ 162, 42, 59 },
		{ 162, 43, 59 },
		{ 162, 44, 59 },
		{ 162, 45, 59 },
		{ 162, 46, 59 },
		{ 162, 47, 59 },
		{ 162, 48, 59 },
		{ 162, 49, 59 },
		{ 162, 50, 59 },
		{ 162, 51, 59 },
		{ 162, 52, 59 },
		{ 162, 53, 59 },
		{ 162, 54, 59 },
		{ 162, 55, 59 },
		{ 162, 56, 59 },
		{ 162, 57, 59 },
		{ 162, 58, 59 },
		{ 162, 59, 59 },
		{ 162, 60, 59 },
		{ 162, 61, 59 },
		{ 162, 62, 59 },
		{ 162, 63, 59 },
		{ 162, 64, 59 },
		{ 162, 65, 59 },
		{ 162, 66, 59 },
		{ 163, 1, 59 },
		{ 163, 2, 59 },
		{ 163, 3, 59 },
		{ 163, 4, 59 },
		{ 163, 5, 59 },
		{ 163, 6, 59 },
		{ 163, 7, 59 },
		{ 163, 8, 59 },
		{ 163, 9, 59 },
		{ 163, 10, 59 },
		{ 163, 11, 59 },
		{ 163, 12, 59 },
		{ 163, 13, 59 },
		{ 163, 14, 59 },
		{ 163, 15, 59 },
		{ 163, 16, 59 },
		{ 163, 17, 59 },
		{ 163, 18, 59 },
		{ 163, 19, 59 },
		{ 163, 20, 59 },
		{ 163, 21, 59 },
		{ 163, 22, 59 },
		{ 163, 23, 59 },
		{ 163, 24, 59 },
		{ 163, 25, 59 },
		{ 163, 26, 59 },
		{ 163, 27, 59 },
		{ 163, 28, 59 },
		{ 163, 29, 59 },
		{ 163, 30, 59 },
		{ 163, 31, 59 },
		{ 163, 32, 59 },
		{ 163, 33, 59 },
		{ 163, 34, 59 },
		{ 163, 35, 59 },
		{ 163, 36, 59 },
		{ 163, 37, 59 },
		{ 163, 38, 59 },
		{ 163, 39, 59 },
		{ 163, 40, 59 },
		{ 163, 41, 59 },
		{ 163, 42, 59 },
		{ 163, 43, 59 },
		{ 163, 44, 59 },
		{ 163, 45, 59 },
		{ 163, 46, 59 },
		{ 163, 47, 59 },
		{ 163, 48, 59 },
		{ 163, 49, 59 },
		{ 163, 50, 59 },
		{ 163, 51, 59 },
		{ 163, 52, 59 },
		{ 163, 53, 59 },
		{ 163, 54, 59 },
		{ 163, 55, 59 },
		{ 163, 56, 59 },
		{ 163, 57, 59 },
		{ 163, 58, 59 },
		{ 163, 59, 59 },
		{ 163, 60, 59 },
		{ 163, 61, 59 },
		{ 163, 62, 59 },
		{ 163, 63, 59 },
		{ 163, 64, 59 },
		{ 163, 65, 59 },
		{ 163, 66, 59 },
		{ 163, 67, 59 },
		{ 163, 68, 59 },
		{ 164, 1, 59 },
		{ 164, 2, 59 },
		{ 164, 3, 59 },
		{ 164, 4, 59 },
		{ 164, 5, 59 },
		{ 164, 6, 59 },
		{ 164, 7, 59 },
		{ 164, 8, 59 },
		{ 164, 9, 59 },
		{ 164, 10, 59 },
		{ 164, 11, 59 },
		{ 164, 12, 59 },
		{ 164, 13, 59 },
		{ 164, 14, 59 },
		{ 164, 15, 59 },
		{ 164, 16, 59 },
		{ 164, 17, 59 },
		{ 164, 18, 59 },
		{ 164, 19, 59 },
		{ 164, 20, 59 },
		{ 164, 21, 59 },
		{ 164, 22, 59 },
		{ 164, 23, 59 },
		{ 164, 24, 59 },
		{ 164, 25, 59 },
		{ 164, 26, 59 },
		{ 164, 27, 59 },
		{ 164, 28, 59 },
		{ 164, 29, 59 },
		{ 164, 30, 59 },
		{ 164, 31, 59 },
		{ 164, 32, 59 },
		{ 164, 33, 59 },
		{ 164, 34, 59 },
		{ 164, 35, 59 },
		{ 164, 36, 59 },
		{ 164, 37, 59 },
		{ 164, 38, 59 },
		{ 164, 39, 59 },
		{ 164, 40, 59 },
		{ 164, 41, 59 },
		{ 164, 42, 59 },
		{ 164, 43, 59 },
		{ 164, 44, 59 },
		{ 164, 45, 59 },
		{ 164, 46, 59 },
		{ 164, 47, 59 },
		{ 164, 48, 59 },
		{ 164, 49, 59 },
		{ 164, 50, 59 },
		{ 164, 51, 59 },
		{ 164, 52, 59 },
		{ 164, 53, 59 },
		{ 164, 54, 59 },
		{ 164, 55, 59 },
		{ 164, 56, 59 },
		{ 164, 57, 59 },
		{ 164, 58, 59 },
		{ 164, 59, 59 },
		{ 164, 60, 59 },
		{ 164, 61, 59 },
		{ 164, 62, 59 },
		{ 164, 63, 59 },
		{ 164, 64, 59 },
		{ 164, 65, 59 },
		{ 164, 66, 59 },
		{ 164, 67, 59 },
		{ 164, 68, 59 },
		{ 164, 69, 59 },
		{ 164, 70, 59 },
		{ 165, 1, 59 },
		{ 165, 2, 59 },
		{ 165, 3, 59 },
		{ 165, 4, 59 },
		{ 165, 5, 59 },
		{ 165, 6, 59 },
		{ 165, 7, 59 },
		{ 165, 8, 59 },
		{ 165, 9, 59 },
		{ 165, 10, 59 },
		{ 165, 11, 59 },
		{ 165, 12, 59 },
		{ 165, 13, 59 },
		{ 165, 14, 59 },
		{ 165, 15, 59 },
		{ 165, 16, 59 },
		{ 165, 17, 59 },
		{ 165, 18, 59 },
		{ 165, 19, 59 },
		{ 165, 20, 59 },
		{ 165, 21, 59 },
		{ 165, 22, 59 },
		{ 165, 23, 59 },
		{ 165, 24, 59 },
		{ 165, 25, 59 },
		{ 165, 26, 59 },
		{ 165, 27, 59 },
		{ 165, 28, 59 },
		{ 165, 29, 59 },
		{ 165, 30, 59 },
		{ 165, 31, 59 },
		{ 165, 32, 59 },
		{ 165, 33, 59 },
		{ 165, 34, 59 },
		{ 165, 35, 59 },
		{ 165, 36, 59 },
		{ 165, 37, 59 },
		{ 165, 38, 59 },
		{ 165, 39, 59 },
		{ 165, 40, 59 },
		{ 165, 41, 59 },
		{ 165, 42, 59 },
		{ 165, 43, 59 },
		{ 165, 44, 59 },
		{ 165, 45, 59 },
		{ 165, 46, 59 },
		{ 165, 47, 59 },
		{ 165, 48, 59 },
		{ 165, 49, 59 },
		{ 165, 50, 59 },
		{ 165, 51, 59 },
		{ 165, 52, 59 },
		{ 165, 53, 59 },
		{ 165, 54, 59 },
		{ 165, 55, 59 },
		{ 165, 56, 59 },
		{ 165, 57, 59 },
		{ 165, 58, 59 },
		{ 165, 59, 59 },
		{ 165, 60, 59 },
		{ 165, 61, 59 },
		{ 165, 62, 59 },
		{ 165, 63, 59 },
		{ 165, 64, 59 },
		{ 165, 65, 59 },
		{ 165, 66, 59 },
		{ 165, 67, 59 },
		{ 165, 68, 59 },
		{ 165, 69, 59 },
		{ 165, 70, 59 },
		{ 165, 71, 59 },
		{ 165, 72, 59 },
		{ 166, 1, 59 },
		{ 166, 2, 59 },
		{ 166, 3, 59 },
		{ 166, 4, 59 },
		{ 166, 5, 59 },
		{ 166, 6, 59 },
		{ 166, 7, 59 },
		{ 166, 8, 59 },
		{ 166, 9, 59 },
		{ 166, 10, 59 },
		{ 166, 11, 59 },
		{ 166, 12, 59 },
		{ 166, 13, 59 },
		{ 166, 14, 59 },
		{ 166, 15, 59 },
		{ 166, 16, 59 },
		{ 166, 17, 59 },
		{ 166, 18, 59 },
		{ 166, 19, 59 },
		{ 166, 20, 59 },
		{ 166, 21, 59 },
		{ 166, 22, 59 },
		{ 166, 23, 59 },
		{ 166, 24, 59 },
		{ 166, 25, 59 },
		{ 166, 26, 59 },
		{ 166, 27, 59 },
		{ 166, 28, 59 },
		{ 166, 29, 59 },
		{ 166, 30, 59 },
		{ 166, 31, 59 },
		{ 166, 32, 59 },
		{ 166, 33, 59 },
		{ 166, 34, 59 },
		{ 166, 35, 59 },
		{ 166, 36, 59 },
		{ 166, 37, 59 },
		{ 166, 38, 59 },
		{ 166, 39, 59 },
		{ 166, 40, 59 },
		{ 166, 41, 59 },
		{ 166, 42, 59 },
		{ 166, 43, 59 },
		{ 166, 44, 59 },
		{ 166, 45, 59 },
		{ 166, 46, 59 },
		{ 166, 47, 59 },
		{ 166, 48, 59 },
		{ 166, 49, 59 },
		{ 166, 50, 59 },
		{ 166, 51, 59 },
		{ 166, 52, 59 },
		{ 166, 53, 59 },
		{ 166, 54, 59 },
		{ 166, 55, 59 },
		{ 166, 56, 59 },
		{ 166, 57, 59 },
		{ 166, 58, 59 },
		{ 166, 59, 59 },
		{ 166, 60, 59 },
		{ 166, 61, 59 },
		{ 166, 62, 59 },
		{ 166, 63, 59 },
		{ 166, 64, 59 },
		{ 166, 65, 59 },
		{ 166, 66, 59 },
		{ 166, 67, 59 },
		{ 166, 68, 59 },
		{ 166, 69, 59 },
		{ 166, 70, 59 },
		{ 166, 71, 59 },
		{ 166, 72, 59 },
		{ 166, 73, 59 },
		{ 166, 74, 59 },
		{ 167, 1, 59 },
		{ 167, 2, 59 },
		{ 167, 3, 59 },
		{ 167, 4, 59 },
		{ 167, 5, 59 },
		{ 167, 6, 59 },
		{ 167, 7, 59 },
		{ 167, 8, 59 },
		{ 167, 9, 59 },
		{ 167, 10, 59 },
		{ 167, 11, 59 },
		{ 167, 12, 59 },
		{ 167, 13, 59 },
		{ 167, 14, 59 },
		{ 167, 15, 59 },
		{ 167, 16, 59 },
		{ 167, 17, 59 },
		{ 167, 18, 59 },
		{ 167, 19, 59 },
		{ 167, 20, 59 },
		{ 167, 21, 59 },
		{ 167, 22, 59 },
		{ 167, 23, 59 },
		{ 167, 24, 59 },
		{ 167, 25, 59 },
		{ 167, 26, 59 },
		{ 167, 27, 59 },
		{ 167, 28, 59 },
		{ 167, 29, 59 },
		{ 167, 30, 59 },
		{ 167, 31, 59 },
		{ 167, 32, 59 },
		{ 167, 33, 59 },
		{ 167, 34, 59 },
		{ 167, 35, 59 },
		{ 167, 36, 59 },
		{ 167, 37, 59 },
		{ 167, 38, 59 },
		{ 167, 39, 59 },
		{ 167, 40, 59 },
		{ 167, 41, 59 },
		{ 167, 42, 59 },
		{ 167, 43, 59 },
		{ 167, 44, 59 },
		{ 167, 45, 59 },
		{ 167, 46, 59 },
		{ 167, 47, 59 },
		{ 167, 48, 59 },
		{ 167, 49, 59 },
		{ 167, 50, 59 },
		{ 167, 51, 59 },
		{ 167, 52, 59 },
		{ 167, 53, 59 },
		{ 167, 54, 59 },
		{ 167, 55, 59 },
		{ 167, 56, 59 },
		{ 167, 57, 59 },
		{ 167, 58, 59 },
		{ 167, 59, 59 },
		{ 167, 60, 59 },
		{ 167, 61, 59 },
		{ 167, 62, 59 },
		{ 167, 63, 59 },
		{ 167, 64, 59 },
		{ 167, 65, 59 },
		{ 167, 66, 59 },
		{ 167, 67, 59 },
		{ 167, 68, 59 },
		{ 167, 69, 59 },
		{ 167, 70, 59 },
		{ 167, 71, 59 },
		{ 167, 72, 59 },
		{ 167, 73, 59 },
		{ 167, 74, 59 },
		{ 167, 75, 59 },
		{ 167, 76, 59 },
		{ 168, 1, 59 },
		{ 168, 2, 59 },
		{ 168, 3, 59 },
		{ 168, 4, 59 },
		{ 168, 5, 59 },
		{ 168, 6, 59 },
		{ 168, 7, 59 },
		{ 168, 8, 59 },
		{ 168, 9, 59 },
		{ 168, 10, 59 },
		{ 168, 11, 59 },
		{ 168, 12, 59 },
		{ 168, 13, 59 },
		{ 168, 14, 59 },
		{ 168, 15, 59 },
		{ 168, 16, 59 },
		{ 168, 17, 59 },
		{ 168, 18, 59 },
		{ 168, 19, 59 },
		{ 168, 20, 59 },
		{ 168, 21, 59 },
		{ 168, 22, 59 },
		{ 168, 23, 59 },
		{ 168, 24, 59 },
		{ 168, 25, 59 },
		{ 168, 26, 59 },
		{ 168, 27, 59 },
		{ 168, 28, 59 },
		{ 168, 29, 59 },
		{ 168, 30, 59 },
		{ 168, 31, 59 },
		{ 168, 32, 59 },
		{ 168, 33, 59 },
		{ 168, 34, 59 },
		{ 168, 35, 59 },
		{ 168, 36, 59 },
		{ 168, 37, 59 },
		{ 168, 38, 59 },
		{ 168, 39, 59 },
		{ 168, 40, 59 },
		{ 168, 41, 59 },
		{ 168, 42, 59 },
		{ 168, 43, 59 },
		{ 168, 44, 59 },
		{ 168, 45, 59 },
		{ 168, 46, 59 },
		{ 168, 47, 59 },
		{ 168, 48, 59 },
		{ 168, 49, 59 },
		{ 168, 50, 59 },
		{ 168, 51, 59 },
		{ 168, 52, 59 },
		{ 168, 53, 59 },
		{ 168, 54, 59 },
		{ 168, 55, 59 },
		{ 168, 56, 59 },
		{ 168, 57, 59 },
		{ 168, 58, 59 },
		{ 168, 59, 59 },
		{ 168, 60, 59 },
		{ 168, 61, 59 },
		{ 168, 62, 59 },
		{ 168, 63, 59 },
		{ 168, 64, 59 },
		{ 168, 65, 59 },
		{ 168, 66, 59 },
		{ 168, 67, 59 },
		{ 168, 68, 59 },
		{ 168, 69, 59 },
		{ 168, 70, 59 },
		{ 168, 71, 59 },
		{ 168, 72, 59 },
		{ 168, 73, 59 },
		{ 168, 74, 59 },
		{ 168, 75, 59 },
		{ 168, 76, 59 },
		{ 168, 77, 59 },
		{ 168, 78, 59 },
		{ 169, 1, 59 },
		{ 169, 2, 59 },
		{ 169, 3, 59 },
		{ 169, 4, 59 },
		{ 169, 5, 59 },
		{ 169, 6, 59 },
		{ 169, 7, 59 },
		{ 169, 8, 59 },
		{ 169, 9, 59 },
		{ 169, 10, 59 },
		{ 169, 11, 59 },
		{ 169, 12, 59 },
		{ 169, 13, 59 },
		{ 169, 14, 59 },
		{ 169, 15, 59 },
		{ 169, 16, 59 },
		{ 169, 17, 59 },
		{ 169, 18, 59 },
		{ 169, 19, 59 },
		{ 169, 20, 59 },
		{ 169, 21, 59 },
		{ 169, 22, 59 },
		{ 169, 23, 59 },
		{ 169, 24, 59 },
		{ 169, 25, 59 },
		{ 169, 26, 59 },
		{ 169, 27, 59 },
		{ 169, 28, 59 },
		{ 169, 29, 59 },
		{ 169, 30, 59 },
		{ 169, 31, 59 },
		{ 169, 32, 59 },
		{ 169, 33, 59 },
		{ 169, 34, 59 },
		{ 169, 35, 59 },
		{ 169, 36, 59 },
		{ 169, 37, 59 },
		{ 169, 38, 59 },
		{ 169, 39, 59 },
		{ 169, 40, 59 },
		{ 169, 41, 59 },
		{ 169, 42, 59 },
		{ 169, 43, 59 },
		{ 169, 44, 59 },
		{ 169, 45, 59 },
		{ 169, 46, 59 },
		{ 169, 47, 59 },
		{ 169, 48, 59 },
		{ 169, 49, 59 },
		{ 169, 50, 59 },
		{ 169, 51, 59 },
		{ 169, 52, 59 },
		{ 169, 53, 59 },
		{ 169, 54, 59 },
		{ 169, 55, 59 },
		{ 169, 56, 59 },
		{ 169, 57, 59 },
		{ 169, 58, 59 },
		{ 169, 59, 59 },
		{ 169, 60, 59 },
		{ 169, 61, 59 },
		{ 169, 62, 59 },
		{ 169, 63, 59 },
		{ 169, 64, 59 },
		{ 169, 65, 59 },
		{ 169, 66, 59 },
		{ 169, 67, 59 },
		{ 169, 68, 59 },
		{ 169, 69, 59 },
		{ 169, 70, 59 },
		{ 169, 71, 59 },
		{ 169, 72, 59 },
		{ 169, 73, 59 },
		{ 169, 74, 59 },
		{ 169, 75, 59 },
		{ 169, 76, 59 },
		{ 169, 77, 59 },
		{ 169, 78, 59 },
		{ 169, 79, 59 },
		{ 169, 80, 59 },
		{ 170, 1, 59 },
		{ 170, 2, 59 },
		{ 170, 3, 59 },
		{ 170, 4, 59 },
		{ 170, 5, 59 },
		{ 170, 6, 59 },
		{ 170, 7, 59 },
		{ 170, 8, 59 },
		{ 170, 9, 59 },
		{ 170, 10, 59 },
		{ 170, 11, 59 },
		{ 170, 12, 59 },
		{ 170, 13, 59 },
		{ 170, 14, 59 },
		{ 170, 15, 59 },
		{ 170, 16, 59 },
		{ 170, 17, 59 },
		{ 170, 18, 59 },
		{ 170, 19, 59 },
		{ 170, 20, 59 },
		{ 170, 21, 59 },
		{ 170, 22, 59 },
		{ 170, 23, 59 },
		{ 170, 24, 59 },
		{ 170, 25, 59 },
		{ 170, 26, 59 },
		{ 170, 27, 59 },
		{ 170, 28, 59 },
		{ 170, 29, 59 },
		{ 170, 30, 59 },
		{ 170, 31, 59 },
		{ 170, 32, 59 },
		{ 170, 33, 59 },
		{ 170, 34, 59 },
		{ 170, 35, 59 },
		{ 170, 36, 59 },
		{ 170, 37, 59 },
		{ 170, 38, 59 },
		{ 170, 39, 59 },
		{ 170, 40, 59 },
		{ 170, 41, 59 },
		{ 170, 42, 59 },
		{ 170, 43, 59 },
		{ 170, 44, 59 },
		{ 170, 45, 59 },
		{ 170, 46, 59 },
		{ 170, 47, 59 },
		{ 170, 48, 59 },
		{ 170, 49, 59 },
		{ 170, 50, 59 },
		{ 170, 51, 59 },
		{ 170, 52, 59 },
		{ 170, 53, 59 },
		{ 170, 54, 59 },
		{ 170, 55, 59 },
		{ 170, 56, 59 },
		{ 170, 57, 59 },
		{ 170, 58, 59 },
		{ 170, 59, 59 },
		{ 170, 60, 59 },
		{ 170, 61, 59 },
		{ 170, 62, 59 },
		{ 170, 63, 59 },
		{ 170, 64, 59 },
		{ 170, 65, 59 },
		{ 170, 66, 59 },
		{ 170, 67, 59 },
		{ 170, 68, 59 },
		{ 170, 69, 59 },
		{ 170, 70, 59 },
		{ 170, 71, 59 },
		{ 170, 72, 59 },
		{ 170, 73, 59 },
		{ 170, 74, 59 },
		{ 170, 75, 59 },
		{ 170, 76, 59 },
		{ 170, 77, 59 },
		{ 170, 78, 59 },
		{ 170, 79, 59 },
		{ 170, 80, 59 },
		{ 170, 81, 59 },
		{ 170, 82, 59 },
		{ 171, 1, 59 },
		{ 171, 2, 59 },
		{ 171, 3, 59 },
		{ 171, 4, 59 },
		{ 171, 5, 59 },
		{ 171, 6, 59 },
		{ 171, 7, 59 },
		{ 171, 8, 59 },
		{ 171, 9, 59 },
		{ 171, 10, 59 },
		{ 171, 11, 59 },
		{ 171, 12, 59 },
		{ 171, 13, 59 },
		{ 171, 14, 59 },
		{ 171, 15, 59 },
		{ 171, 16, 59 },
		{ 171, 17, 59 },
		{ 171, 18, 59 },
		{ 171, 19, 59 },
		{ 171, 20, 59 },
		{ 171, 21, 59 },
		{ 171, 22, 59 },
		{ 171, 23, 59 },
		{ 171, 24, 59 },
		{ 171, 25, 59 },
		{ 171, 26, 59 },
		{ 171, 27, 59 },
		{ 171, 28, 59 },
		{ 171, 29, 59 },
		{ 171, 30, 59 },
		{ 171, 31, 59 },
		{ 171, 32, 59 },
		{ 171, 33, 59 },
		{ 171, 34, 59 },
		{ 171, 35, 59 },
		{ 171, 36, 59 },
		{ 171, 37, 59 },
		{ 171, 38, 59 },
		{ 171, 39, 59 },
		{ 171, 40, 59 },
		{ 171, 41, 59 },
		{ 171, 42, 59 },
		{ 171, 43, 59 },
		{ 171, 44, 59 },
		{ 171, 45, 59 },
		{ 171, 46, 59 },
		{ 171, 47, 59 },
		{ 171, 48, 59 },
		{ 171, 49, 59 },
		{ 171, 50, 59 },
		{ 171, 51, 59 },
		{ 171, 52, 59 },
		{ 171, 53, 59 },
		{ 171, 54, 59 },
		{ 171, 55, 59 },
		{ 171, 56, 59 },
		{ 171, 57, 59 },
		{ 171, 58, 59 },
		{ 171, 59, 59 },
		{ 171, 60, 59 },
		{ 171, 61, 59 },
		{ 171, 62, 59 },
		{ 171, 63, 59 },
		{ 171, 64, 59 },
		{ 171, 65, 59 },
		{ 171, 66, 59 },
		{ 171, 67, 59 },
		{ 171, 68, 59 },
		{ 171, 69, 59 },
		{ 171, 70, 59 },
		{ 171, 71, 59 },
		{ 171, 72, 59 },
		{ 171, 73, 59 },
		{ 171, 74, 59 },
		{ 171, 75, 59 },
		{ 171, 76, 59 },
		{ 171, 77, 59 },
		{ 171, 78, 59 },
		{ 171, 79, 59 },
		{ 171, 80, 59 },
		{ 171, 81, 59 },
		{ 171, 82, 59 },
		{ 171, 83, 59 },
		{ 171, 84, 59 },
		{ 172, 1, 59 },
		{ 172, 2, 59 },
		{ 172, 3, 59 },
		{ 172, 4, 59 },
		{ 172, 5, 59 },
		{ 172, 6, 59 },
		{ 172, 7, 59 },
		{ 172, 8, 59 },
		{ 172, 9, 59 },
		{ 172, 10, 59 },
		{ 172, 11, 59 },
		{ 172, 12, 59 },
		{ 172, 13, 59 },
		{ 172, 14, 59 },
		{ 172, 15, 59 },
		{ 172, 16, 59 },
		{ 172, 17, 59 },
		{ 172, 18, 59 },
		{ 172, 19, 59 },
		{ 172, 20, 59 },
		{ 172, 21, 59 },
		{ 172, 22, 59 },
		{ 172, 23, 59 },
		{ 172, 24, 59 },
		{ 172, 25, 59 },
		{ 172, 26, 59 },
		{ 172, 27, 59 },
		{ 172, 28, 59 },
		{ 172, 29, 59 },
		{ 172, 30, 59 },
		{ 172, 31, 59 },
		{ 172, 32, 59 },
		{ 172, 33, 59 },
		{ 172, 34, 59 },
		{ 172, 35, 59 },
		{ 172, 36, 59 },
		{ 172, 37, 59 },
		{ 172, 38, 59 },
		{ 172, 39, 59 },
		{ 172, 40, 59 },
		{ 172, 41, 59 },
		{ 172, 42, 59 },
		{ 172, 43, 59 },
		{ 172, 44, 59 },
		{ 172, 45, 59 },
		{ 172, 46, 59 },
		{ 172, 47, 59 },
		{ 172, 48, 59 },
		{ 172, 49, 59 },
		{ 172, 50, 59 },
		{ 172, 51, 59 },
		{ 172, 52, 59 },
		{ 172, 53, 59 },
		{ 172, 54, 59 },
		{ 172, 55, 59 },
		{ 172, 56, 59 },
		{ 172, 57, 59 },
		{ 172, 58, 59 },
		{ 172, 59, 59 },
		{ 172, 60, 59 },
		{ 172, 61, 59 },
		{ 172, 62, 59 },
		{ 172, 63, 59 },
		{ 172, 64, 59 },
		{ 172, 65, 59 },
		{ 172, 66, 59 },
		{ 172, 67, 59 },
		{ 172, 68, 59 },
		{ 172, 69, 59 },
		{ 172, 70, 59 },
		{ 172, 71, 59 },
		{ 172, 72, 59 },
		{ 172, 73, 59 },
		{ 172, 74, 59 },
		{ 172, 75, 59 },
		{ 172, 76, 59 },
		{ 172, 77, 59 },
		{ 172, 78, 59 },
		{ 172, 79, 59 },
		{ 172, 80, 59 },
		{ 172, 81, 59 },
		{ 172, 82, 59 },
		{ 172, 83, 59 },
		{ 172, 84, 59 },
		{ 172, 85, 59 },
		{ 172, 86, 59 },
		{ 173, 1, 59 },
		{ 173, 2, 59 },
		{ 173, 3, 59 },
		{ 173, 4, 59 },
		{ 173, 5, 59 },
		{ 173, 6, 59 },
		{ 173, 7, 59 },
		{ 173, 8, 59 },
		{ 173, 9, 59 },
		{ 173, 10, 59 },
		{ 173, 11, 59 },
		{ 173, 12, 59 },
		{ 173, 13, 59 },
		{ 173, 14, 59 },
		{ 173, 15, 59 },
		{ 173, 16, 59 },
		{ 173, 17, 59 },
		{ 173, 18, 59 },
		{ 173, 19, 59 },
		{ 173, 20, 59 },
		{ 173, 21, 59 },
		{ 173, 22, 59 },
		{ 173, 23, 59 },
		{ 173, 24, 59 },
		{ 173, 25, 59 },
		{ 173, 26, 59 },
		{ 173, 27, 59 },
		{ 173, 28, 59 },
		{ 173, 29, 59 },
		{ 173, 30, 59 },
		{ 173, 31, 59 },
		{ 173, 32, 59 },
		{ 173, 33, 59 },
		{ 173, 34, 59 },
		{ 173, 35, 59 },
		{ 173, 36, 59 },
		{ 173, 37, 59 },
		{ 173, 38, 59 },
		{ 173, 39, 59 },
		{ 173, 40, 59 },
		{ 173, 41, 59 },
		{ 173, 42, 59 },
		{ 173, 43, 59 },
		{ 173, 44, 59 },
		{ 173, 45, 59 },
		{ 173, 46, 59 },
		{ 173, 47, 59 },
		{ 173, 48, 59 },
		{ 173, 49, 59 },
		{ 173, 50, 59 },
		{ 173, 51, 59 },
		{ 173, 52, 59 },
		{ 173, 53, 59 },
		{ 173, 54, 59 },
		{ 173, 55, 59 },
		{ 173, 56, 59 },
		{ 173, 57, 59 },
		{ 173, 58, 59 },
		{ 173, 59, 59 },
		{ 173, 60, 59 },
		{ 173, 61, 59 },
		{ 173, 62, 59 },
		{ 173, 63, 59 },
		{ 173, 64, 59 },
		{ 173, 65, 59 },
		{ 173, 66, 59 },
		{ 173, 67, 59 },
		{ 173, 68, 59 },
		{ 173, 69, 59 },
		{ 173, 70, 59 },
		{ 173, 71, 59 },
		{ 173, 72, 59 },
		{ 173, 73, 59 },
		{ 173, 74, 59 },
		{ 173, 75, 59 },
		{ 173, 76, 59 },
		{ 173, 77, 59 },
		{ 173, 78, 59 },
		{ 173, 79, 59 },
		{ 173, 80, 59 },
		{ 173, 81, 59 },
		{ 173, 82, 59 },
		{ 173, 83, 59 },
		{ 173, 84, 59 },
		{ 173, 85, 59 },
		{ 173, 86, 59 },
		{ 173, 87, 59 },
		{ 173, 88, 59 },
		{ 174, 1, 59 },
		{ 174, 2, 59 },
		{ 174, 3, 59 },
		{ 174, 4, 59 },
		{ 174, 5, 59 },
		{ 174, 6, 59 },
		{ 174, 7, 59 },
		{ 174, 8, 59 },
		{ 174, 9, 59 },
		{ 174, 10, 59 },
		{ 174, 11, 59 },
		{ 174, 12, 59 },
		{ 174, 13, 59 },
		{ 174, 14, 59 },
		{ 174, 15, 59 },
		{ 174, 16, 59 },
		{ 174, 17, 59 },
		{ 174, 18, 59 },
		{ 174, 19, 59 },
		{ 174, 20, 59 },
		{ 174, 21, 59 },
		{ 174, 22, 59 },
		{ 174, 23, 59 },
		{ 174, 24, 59 },
		{ 174, 25, 59 },
		{ 174, 26, 59 },
		{ 174, 27, 59 },
		{ 174, 28, 59 },
		{ 174, 29, 59 },
		{ 174, 30, 59 },
		{ 174, 31, 59 },
		{ 174, 32, 59 },
		{ 174, 33, 59 },
		{ 174, 34, 59 },
		{ 174, 35, 59 },
		{ 174, 36, 59 },
		{ 174, 37, 59 },
		{ 174, 38, 59 },
		{ 174, 39, 59 },
		{ 174, 40, 59 },
		{ 174, 41, 59 },
		{ 174, 42, 59 },
		{ 174, 43, 59 },
		{ 174, 44, 59 },
		{ 174, 45, 59 },
		{ 174, 46, 59 },
		{ 174, 47, 59 },
		{ 174, 48, 59 },
		{ 174, 49, 59 },
		{ 174, 50, 59 },
		{ 174, 51, 59 },
		{ 174, 52, 59 },
		{ 174, 53, 59 },
		{ 174, 54, 59 },
		{ 174, 55, 59 },
		{ 174, 56, 59 },
		{ 174, 57, 59 },
		{ 174, 58, 59 },
		{ 174, 59, 59 },
		{ 174, 60, 59 },
		{ 174, 61, 59 },
		{ 174, 62, 59 },
		{ 174, 63, 59 },
		{ 174, 64, 59 },
		{ 174, 65, 59 },
		{ 174, 66, 59 },
		{ 174, 67, 59 },
		{ 174, 68, 59 },
		{ 174, 69, 59 },
		{ 174, 70, 59 },
		{ 174, 71, 59 },
		{ 174, 72, 59 },
		{ 174, 73, 59 },
		{ 174, 74, 59 },
		{ 174, 75, 59 },
		{ 174, 76, 59 },
		{ 174, 77, 59 },
		{ 174, 78, 59 },
		{ 174, 79, 59 },
		{ 174, 80, 59 },
		{ 174, 81, 59 },
		{ 174, 82, 59 },
		{ 174, 83, 59 },
		{ 174, 84, 59 },
		{ 174, 85, 59 },
		{ 174, 86, 59 },
		{ 174, 87, 59 },
		{ 174, 88, 59 },
		{ 174, 89, 59 },
		{ 174, 90, 59 },
		{ 175, 1, 59 },
		{ 175, 2, 59 },
		{ 175, 3, 59 },
		{ 175, 4, 59 },
		{ 175, 5, 59 },
		{ 175, 6, 59 },
		{ 175, 7, 59 },
		{ 175, 8, 59 },
		{ 175, 9, 59 },
		{ 175, 10, 59 },
		{ 175, 11, 59 },
		{ 175, 12, 59 },
		{ 175, 13, 59 },
		{ 175, 14, 59 },
		{ 175, 15, 59 },
		{ 175, 16, 59 },
		{ 175, 17, 59 },
		{ 175, 18, 59 },
		{ 175, 19, 59 },
		{ 175, 20, 59 },
		{ 175, 21, 59 },
		{ 175, 22, 59 },
		{ 175, 23, 59 },
		{ 175, 24, 59 },
		{ 175, 25, 59 },
		{ 175, 26, 59 },
		{ 175, 27, 59 },
		{ 175, 28, 59 },
		{ 175, 29, 59 },
		{ 175, 30, 59 },
		{ 175, 31, 59 },
		{ 175, 32, 59 },
		{ 175, 33, 59 },
		{ 175, 34, 59 },
		{ 175, 35, 59 },
		{ 175, 36, 59 },
		{ 175, 37, 59 },
		{ 175, 38, 59 },
		{ 175, 39, 59 },
		{ 175, 40, 59 },
		{ 175, 41, 59 },
		{ 175, 42, 59 },
		{ 175, 43, 59 },
		{ 175, 44, 59 },
		{ 175, 45, 59 },
		{ 175, 46, 59 },
		{ 175, 47, 59 },
		{ 175, 48, 59 },
		{ 175, 49, 59 },
		{ 175, 50, 59 },
		{ 175, 51, 59 },
		{ 175, 52, 59 },
		{ 175, 53, 59 },
		{ 175, 54, 59 },
		{ 175, 55, 59 },
		{ 175, 56, 59 },
		{ 175, 57, 59 },
		{ 175, 58, 59 },
		{ 175, 59, 59 },
		{ 175, 60, 59 },
		{ 175, 61, 59 },
		{ 175, 62, 59 },
		{ 175, 63, 59 },
		{ 175, 64, 59 },
		{ 175, 65, 59 },
		{ 175, 66, 59 },
		{ 175, 67, 59 },
		{ 175, 68, 59 },
		{ 175, 69, 59 },
		{ 175, 70, 59 },
		{ 175, 71, 59 },
		{ 175, 72, 59 },
		{ 175, 73, 59 },
		{ 175, 74, 59 },
		{ 175, 75, 59 },
		{ 175, 76, 59 },
		{ 175, 77, 59 },
		{ 175, 78, 59 },
		{ 175, 79, 59 },
		{ 175, 80, 59 },
		{ 175, 81, 59 },
		{ 175, 82, 59 },
		{ 175, 83, 59 },
		{ 175, 84, 59 },
		{ 175, 85, 59 },
		{ 175, 86, 59 },
		{ 175, 87, 59 },
		{ 175, 88, 59 },
		{ 175, 89, 59 },
		{ 175, 90, 59 },
		{ 175, 91, 59 },
		{ 175, 92, 59 },
		{ 176, 1, 59 },
		{ 176, 2, 59 },
		{ 176, 3, 59 },
		{ 176, 4, 59 },
		{ 176, 5, 59 },
		{ 176, 6, 59 },
		{ 176, 7, 59 },
		{ 176, 8, 59 },
		{ 176, 9, 59 },
		{ 176, 10, 59 },
		{ 176, 11, 59 },
		{ 176, 12, 59 },
		{ 176, 13, 59 },
		{ 176, 14, 59 },
		{ 176, 15, 59 },
		{ 176, 16, 59 },
		{ 176, 17, 59 },
		{ 176, 18, 59 },
		{ 176, 19, 59 },
		{ 176, 20, 59 },
		{ 176, 21, 59 },
		{ 176, 22, 59 },
		{ 176, 23, 59 },
		{ 176, 24, 59 },
		{ 176, 25, 59 },
		{ 176, 26, 59 },
		{ 176, 27, 59 },
		{ 176, 28, 59 },
		{ 176, 29, 59 },
		{ 176, 30, 59 },
		{ 176, 31, 59 },
		{ 176, 32, 59 },
		{ 176, 33, 59 },
		{ 176, 34, 59 },
		{ 176, 35, 59 },
		{ 176, 36, 59 },
		{ 176, 37, 59 },
		{ 176, 38, 59 },
		{ 176, 39, 59 },
		{ 176, 40, 59 },
		{ 176, 41, 59 },
		{ 176, 42, 59 },
		{ 176, 43, 59 },
		{ 176, 44, 59 },
		{ 176, 45, 59 },
		{ 176, 46, 59 },
		{ 176, 47, 59 },
		{ 176, 48, 59 },
		{ 176, 49, 59 },
		{ 176, 50, 59 },
		{ 176, 51, 59 },
		{ 176, 52, 59 },
		{ 176, 53, 59 },
		{ 176, 54, 59 },
		{ 176, 55, 59 },
		{ 176, 56, 59 },
		{ 176, 57, 59 },
		{ 176, 58, 59 },
		{ 176, 59, 59 },
		{ 176, 60, 59 },
		{ 176, 61, 59 },
		{ 176, 62, 59 },
		{ 176, 63, 59 },
		{ 176, 64, 59 },
		{ 176, 65, 59 },
		{ 176, 66, 59 },
		{ 176, 67, 59 },
		{ 176, 68, 59 },
		{ 176, 69, 59 },
		{ 176, 70, 59 },
		{ 176, 71, 59 },
		{ 176, 72, 59 },
		{ 176, 73, 59 },
		{ 176, 74, 59 },
		{ 176, 75, 59 },
		{ 176, 76, 59 },
		{ 176, 77, 59 },
		{ 176, 78, 59 },
		{ 176, 79, 59 },
		{ 176, 80, 59 },
		{ 176, 81, 59 },
		{ 176, 82, 59 },
		{ 176, 83, 59 },
		{ 176, 84, 59 },
		{ 176, 85, 59 },
		{ 176, 86, 59 },
		{ 176, 87, 59 },
		{ 176, 88, 59 },
		{ 176, 89, 59 },
		{ 176, 90, 59 },
		{ 176, 91, 59 },
		{ 176, 92, 59 },
		{ 176, 93, 59 },
		{ 176, 94, 59 },
		{ 177, 1, 59 },
		{ 177, 2, 59 },
		{ 177, 3, 59 },
		{ 177, 4, 59 },
		{ 177, 5, 59 },
		{ 177, 6, 59 },
		{ 177, 7, 59 },
		{ 177, 8, 59 },
		{ 177, 9, 59 },
		{ 177, 10, 59 },
		{ 177, 11, 59 },
		{ 177, 12, 59 },
		{ 177, 13, 59 },
		{ 177, 14, 59 },
		{ 177, 15, 59 },
		{ 177, 16, 59 },
		{ 177, 17, 59 },
		{ 177, 18, 59 },
		{ 177, 19, 59 },
		{ 177, 20, 59 },
		{ 177, 21, 59 },
		{ 177, 22, 59 },
		{ 177, 23, 59 },
		{ 177, 24, 59 },
		{ 177, 25, 59 },
		{ 177, 26, 59 },
		{ 177, 27, 59 },
		{ 177, 28, 59 },
		{ 177, 29, 59 },
		{ 177, 30, 59 },
		{ 177, 31, 59 },
		{ 177, 32, 59 },
		{ 177, 33, 59 },
		{ 177, 34, 59 },
		{ 177, 35, 59 },
		{ 177, 36, 59 },
		{ 177, 37, 59 },
		{ 177, 38, 59 },
		{ 177, 39, 59 },
		{ 177, 40, 59 },
		{ 177, 41, 59 },
		{ 177, 42, 59 },
		{ 177, 43, 59 },
		{ 177, 44, 59 },
		{ 177, 45, 59 },
		{ 177, 46, 59 },
		{ 177, 47, 59 },
		{ 177, 48, 59 },
		{ 177, 49, 59 },
		{ 177, 50, 59 },
		{ 177, 51, 59 },
		{ 177, 52, 59 },
		{ 177, 53, 59 },
		{ 177, 54, 59 },
		{ 177, 55, 59 },
		{ 177, 56, 59 },
		{ 177, 57, 59 },
		{ 177, 58, 59 },
		{ 177, 59, 59 },
		{ 177, 60, 59 },
		{ 177, 61, 59 },
		{ 177, 62, 59 },
		{ 177, 63, 59 },
		{ 177, 64, 59 },
		{ 177, 65, 59 },
		{ 177, 66, 59 },
		{ 177, 67, 59 },
		{ 177, 68, 59 },
		{ 177, 69, 59 },
		{ 177, 70, 59 },
		{ 177, 71, 59 },
		{ 177, 72, 59 },
		{ 177, 73, 59 },
		{ 177, 74, 59 },
		{ 177, 75, 59 },
		{ 177, 76, 59 },
		{ 177, 77, 59 },
		{ 177, 78, 59 },
		{ 177, 79, 59 },
		{ 177, 80, 59 },
		{ 177, 81, 59 },
		{ 177, 82, 59 },
		{ 177, 83, 59 },
		{ 177, 84, 59 },
		{ 177, 85, 59 },
		{ 177, 86, 59 },
		{ 177, 87, 59 },
		{ 177, 88, 59 },
		{ 177, 89, 59 },
		{ 177, 90, 59 },
		{ 177, 91, 59 },
		{ 177, 92, 59 },
		{ 177, 93, 59 },
		{ 177, 94, 59 },
		{ 177, 95, 59 },
		{ 177, 96, 59 },
		{ 178, 1, 59 },
		{ 178, 2, 59 },
		{ 178, 3, 59 },
		{ 178, 4, 59 },
		{ 178, 5, 59 },
		{ 178, 6, 59 },
		{ 178, 7, 59 },
		{ 178, 8, 59 },
		{ 178, 9, 59 },
		{ 178, 10, 59 },
		{ 178, 11, 59 },
		{ 178, 12, 59 },
		{ 178, 13, 59 },
		{ 178, 14, 59 },
		{ 178, 15, 59 },
		{ 178, 16, 59 },
		{ 178, 17, 59 },
		{ 178, 18, 59 },
		{ 178, 19, 59 },
		{ 178, 20, 59 },
		{ 178, 21, 59 },
		{ 178, 22, 59 },
		{ 178, 23, 59 },
		{ 178, 24, 59 },
		{ 178, 25, 59 },
		{ 178, 26, 59 },
		{ 178, 27, 59 },
		{ 178, 28, 59 },
		{ 178, 29, 59 },
		{ 178, 30, 59 },
		{ 178, 31, 59 },
		{ 178, 32, 59 },
		{ 178, 33, 59 },
		{ 178, 34, 59 },
		{ 178, 35, 59 },
		{ 178, 36, 59 },
		{ 178, 37, 59 },
		{ 178, 38, 59 },
		{ 178, 39, 59 },
		{ 178, 40, 59 },
		{ 178, 41, 59 },
		{ 178, 42, 59 },
		{ 178, 43, 59 },
		{ 178, 44, 59 },
		{ 178, 45, 59 },
		{ 178, 46, 59 },
		{ 178, 47, 59 },
		{ 178, 48, 59 },
		{ 178, 49, 59 },
		{ 178, 50, 59 },
		{ 178, 51, 59 },
		{ 178, 52, 59 },
		{ 178, 53, 59 },
		{ 178, 54, 59 },
		{ 178, 55, 59 },
		{ 178, 56, 59 },
		{ 178, 57, 59 },
		{ 178, 58, 59 },
		{ 178, 59, 59 },
		{ 178, 60, 59 },
		{ 178, 61, 59 },
		{ 178, 62, 59 },
		{ 178, 63, 59 },
		{ 178, 64, 59 },
		{ 178, 65, 59 },
		{ 178, 66, 59 },
		{ 178, 67, 59 },
		{ 178, 68, 59 },
		{ 178, 69, 59 },
		{ 178, 70, 59 },
		{ 178, 71, 59 },
		{ 178, 72, 59 },
		{ 178, 73, 59 },
		{ 178, 74, 59 },
		{ 178, 75, 59 },
		{ 178, 76, 59 },
		{ 178, 77, 59 },
		{ 178, 78, 59 },
		{ 178, 79, 59 },
		{ 178, 80, 59 },
		{ 178, 81, 59 },
		{ 178, 82, 59 },
		{ 178, 83, 59 },
		{ 178, 84, 59 },
		{ 178, 85, 59 },
		{ 178, 86, 59 },
		{ 178, 87, 59 },
		{ 178, 88, 59 },
		{ 178, 89, 59 },
		{ 178, 90, 59 },
		{ 178, 91, 59 },
		{ 178, 92, 59 },
		{ 178, 93, 59 },
		{ 178, 94, 59 },
		{ 178, 95, 59 },
		{ 178, 96, 59 },
		{ 178, 97, 59 },
		{ 178, 98, 59 },
		{ 179, 1, 59 },
		{ 179, 2, 59 },
		{ 179, 3, 59 },
		{ 179, 4, 59 },
		{ 179, 5, 59 },
		{ 179, 6, 59 },
		{ 179, 7, 59 },
		{ 179, 8, 59 },
		{ 179, 9, 59 },
		{ 179, 10, 59 },
		{ 179, 11, 59 },
		{ 179, 12, 59 },
		{ 179, 13, 59 },
		{ 179, 14, 59 },
		{ 179, 15, 59 },
		{ 179, 16, 59 },
		{ 179, 17, 59 },
		{ 179, 18, 59 },
		{ 179, 19, 59 },
		{ 179, 20, 59 },
		{ 179, 21, 59 },
		{ 179, 22, 59 },
		{ 179, 23, 59 },
		{ 179, 24, 59 },
		{ 179, 25, 59 },
		{ 179, 26, 59 },
		{ 179, 27, 59 },
		{ 179, 28, 59 },
		{ 179, 29, 59 },
		{ 179, 30, 59 },
		{ 179, 31, 59 },
		{ 179, 32, 59 },
		{ 179, 33, 59 },
		{ 179, 34, 59 },
		{ 179, 35, 59 },
		{ 179, 36, 59 },
		{ 179, 37, 59 },
		{ 179, 38, 59 },
		{ 179, 39, 59 },
		{ 179, 40, 59 },
		{ 179, 41, 59 },
		{ 179, 42, 59 },
		{ 179, 43, 59 },
		{ 179, 44, 59 },
		{ 179, 45, 59 },
		{ 179, 46, 59 },
		{ 179, 47, 59 },
		{ 179, 48, 59 },
		{ 179, 49, 59 },
		{ 179, 50, 59 },
		{ 179, 51, 59 },
		{ 179, 52, 59 },
		{ 179, 53, 59 },
		{ 179, 54, 59 },
		{ 179, 55, 59 },
		{ 179, 56, 59 },
		{ 179, 57, 59 },
		{ 179, 58, 59 },
		{ 179, 59, 59 },
		{ 179, 60, 59 },
		{ 179, 61, 59 },
		{ 179, 62, 59 },
		{ 179, 63, 59 },
		{ 179, 64, 59 },
		{ 179, 65, 59 },
		{ 179, 66, 59 },
		{ 179, 67, 59 },
		{ 179, 68, 59 },
		{ 179, 69, 59 },
		{ 179, 70, 59 },
		{ 179, 71, 59 },
		{ 179, 72, 59 },
		{ 179, 73, 59 },
		{ 179, 74, 59 },
		{ 179, 75, 59 },
		{ 179, 76, 59 },
		{ 179, 77, 59 },
		{ 179, 78, 59 },
		{ 179, 79, 59 },
		{ 179, 80, 59 },
		{ 179, 81, 59 },
		{ 179, 82, 59 },
		{ 179, 83, 59 },
		{ 179, 84, 59 },
		{ 179, 85, 59 },
		{ 179, 86, 59 },
		{ 179, 87, 59 },
		{ 179, 88, 59 },
		{ 179, 89, 59 },
		{ 179, 90, 59 },
		{ 179, 91, 59 },
		{ 179, 92, 59 },
		{ 179, 93, 59 },
		{ 179, 94, 59 },
		{ 179, 95, 59 },
		{ 179, 96, 59 },
		{ 179, 97, 59 },
		{ 179, 98, 59 },
		{ 179, 99, 59 },
		{ 179, 100, 59 },
		{ 180, 1, 59 },
		{ 180, 2, 59 },
		{ 180, 3, 59 },
		{ 180, 4, 59 },
		{ 180, 5, 59 },
		{ 180, 6, 59 },
		{ 180, 7, 59 },
		{ 180, 8, 59 },
		{ 180, 9, 59 },
		{ 180, 10, 59 },
		{ 180, 11, 59 },
		{ 180, 12, 59 },
		{ 180, 13, 59 },
		{ 180, 14, 59 },
		{ 180, 15, 59 },
		{ 180, 16, 59 },
		{ 180, 17, 59 },
		{ 180, 18, 59 },
		{ 180, 19, 59 },
		{ 180, 20, 59 },
		{ 180, 21, 59 },
		{ 180, 22, 59 },
		{ 180, 23, 59 },
		{ 180, 24, 59 },
		{ 180, 25, 59 },
		{ 180, 26, 59 },
		{ 180, 27, 59 },
		{ 180, 28, 59 },
		{ 180, 29, 59 },
		{ 180, 30, 59 },
		{ 180, 31, 59 },
		{ 180, 32, 59 },
		{ 180, 33, 59 },
		{ 180, 34, 59 },
		{ 180, 35, 59 },
		{ 180, 36, 59 },
		{ 180, 37, 59 },
		{ 180, 38, 59 },
		{ 180, 39, 59 },
		{ 180, 40, 59 },
		{ 180, 41, 59 },
		{ 180, 42, 59 },
		{ 180, 43, 59 },
		{ 180, 44, 59 },
		{ 180, 45, 59 },
		{ 180, 46, 59 },
		{ 180, 47, 59 },
		{ 180, 48, 59 },
		{ 180, 49, 59 },
		{ 180, 50, 59 },
		{ 180, 51, 59 },
		{ 180, 52, 59 },
		{ 180, 53, 59 },
		{ 180, 54, 59 },
		{ 180, 55, 59 },
		{ 180, 56, 59 },
		{ 180, 57, 59 },
		{ 180, 58, 59 },
		{ 180, 59, 59 },
		{ 180, 60, 59 },
		{ 180, 61, 59 },
		{ 180, 62, 59 },
		{ 180, 63, 59 },
		{ 180, 64, 59 },
		{ 180, 65, 59 },
		{ 180, 66, 59 },
		{ 180, 67, 59 },
		{ 180, 68, 59 },
		{ 180, 69, 59 },
		{ 180, 70, 59 },
		{ 180, 71, 59 },
		{ 180, 72, 59 },
		{ 180, 73, 59 },
		{ 180, 74, 59 },
		{ 180, 75, 59 },
		{ 180, 76, 59 },
		{ 180, 77, 59 },
		{ 180, 78, 59 },
		{ 180, 79, 59 },
		{ 180, 80, 59 },
		{ 180, 81, 59 },
		{ 180, 82, 59 },
		{ 180, 83, 59 },
		{ 180, 84, 59 },
		{ 180, 85, 59 },
		{ 180, 86, 59 },
		{ 180, 87, 59 },
		{ 180, 88, 59 },
		{ 180, 89, 59 },
		{ 180, 90, 59 },
		{ 180, 91, 59 },
		{ 180, 92, 59 },
		{ 180, 93, 59 },
		{ 180, 94, 59 },
		{ 180, 95, 59 },
		{ 180, 96, 59 },
		{ 180, 97, 59 },
		{ 180, 98, 59 },
		{ 180, 99, 59 },
		{ 180, 100, 59 },
		{ 180, 101, 59 },
		{ 180, 102, 59 },
		{ 181, 1, 59 },
		{ 181, 2, 59 },
		{ 181, 3, 59 },
		{ 181, 4, 59 },
		{ 181, 5, 59 },
		{ 181, 6, 59 },
		{ 181, 7, 59 },
		{ 181, 8, 59 },
		{ 181, 9, 59 },
		{ 181, 10, 59 },
		{ 181, 11, 59 },
		{ 181, 12, 59 },
		{ 181, 13, 59 },
		{ 181, 14, 59 },
		{ 181, 15, 59 },
		{ 181, 16, 59 },
		{ 181, 17, 59 },
		{ 181, 18, 59 },
		{ 181, 19, 59 },
		{ 181, 20, 59 },
		{ 181, 21, 59 },
		{ 181, 22, 59 },
		{ 181, 23, 59 },
		{ 181, 24, 59 },
		{ 181, 25, 59 },
		{ 181, 26, 59 },
		{ 181, 27, 59 },
		{ 181, 28, 59 },
		{ 181, 29, 59 },
		{ 181, 30, 59 },
		{ 181, 31, 59 },
		{ 181, 32, 59 },
		{ 181, 33, 59 },
		{ 181, 34, 59 },
		{ 181, 35, 59 },
		{ 181, 36, 59 },
		{ 181, 37, 59 },
		{ 181, 38, 59 },
		{ 181, 39, 59 },
		{ 181, 40, 59 },
		{ 181, 41, 59 },
		{ 181, 42, 59 },
		{ 181, 43, 59 },
		{ 181, 44, 59 },
		{ 181, 45, 59 },
		{ 181, 46, 59 },
		{ 181, 47, 59 },
		{ 181, 48, 59 },
		{ 181, 49, 59 },
		{ 181, 50, 59 },
		{ 181, 51, 59 },
		{ 181, 52, 59 },
		{ 181, 53, 59 },
		{ 181, 54, 59 },
		{ 181, 55, 59 },
		{ 181, 56, 59 },
		{ 181, 57, 59 },
		{ 181, 58, 59 },
		{ 181, 59, 59 },
		{ 181, 60, 59 },
		{ 181, 61, 59 },
		{ 181, 62, 59 },
		{ 181, 63, 59 },
		{ 181, 64, 59 },
		{ 181, 65, 59 },
		{ 181, 66, 59 },
		{ 181, 67, 59 },
		{ 181, 68, 59 },
		{ 181, 69, 59 },
		{ 181, 70, 59 },
		{ 181, 71, 59 },
		{ 181, 72, 59 },
		{ 181, 73, 59 },
		{ 181, 74, 59 },
		{ 181, 75, 59 },
		{ 181, 76, 59 },
		{ 181, 77, 59 },
		{ 181, 78, 59 },
		{ 181, 79, 59 },
		{ 181, 80, 59 },
		{ 181, 81, 59 },
		{ 181, 82, 59 },
		{ 181, 83, 59 },
		{ 181, 84, 59 },
		{ 181, 85, 59 },
		{ 181, 86, 59 },
		{ 181, 87, 59 },
		{ 181, 88, 59 },
		{ 181, 89, 59 },
		{ 181, 90, 59 },
		{ 181, 91, 59 },
		{ 181, 92, 59 },
		{ 181, 93, 59 },
		{ 181, 94, 59 },
		{ 181, 95, 59 },
		{ 181, 96, 59 },
		{ 181, 97, 59 },
		{ 181, 98, 59 },
		{ 181, 99, 59 },
		{ 181, 100, 59 },
		{ 181, 101, 59 },
		{ 181, 102, 59 },
		{ 181, 103, 59 },
		{ 181, 104, 59 },
		{ 182, 1, 59 },
		{ 182, 2, 59 },
		{ 182, 3, 59 },
		{ 182, 4, 59 },
		{ 182, 5, 59 },
		{ 182, 6, 59 },
		{ 182, 7, 59 },
		{ 182, 8, 59 },
		{ 182, 9, 59 },
		{ 182, 10, 59 },
		{ 182, 11, 59 },
		{ 182, 12, 59 },
		{ 182, 13, 59 },
		{ 182, 14, 59 },
		{ 182, 15, 59 },
		{ 182, 16, 59 },
		{ 182, 17, 59 },
		{ 182, 18, 59 },
		{ 182, 19, 59 },
		{ 182, 20, 59 },
		{ 182, 21, 59 },
		{ 182, 22, 59 },
		{ 182, 23, 59 },
		{ 182, 24, 59 },
		{ 182, 25, 59 },
		{ 182, 26, 59 },
		{ 182, 27, 59 },
		{ 182, 28, 59 },
		{ 182, 29, 59 },
		{ 182, 30, 59 },
		{ 182, 31, 59 },
		{ 182, 32, 59 },
		{ 182, 33, 59 },
		{ 182, 34, 59 },
		{ 182, 35, 59 },
		{ 182, 36, 59 },
		{ 182, 37, 59 },
		{ 182, 38, 59 },
		{ 182, 39, 59 },
		{ 182, 40, 59 },
		{ 182, 41, 59 },
		{ 182, 42, 59 },
		{ 182, 43, 59 },
		{ 182, 44, 59 },
		{ 182, 45, 59 },
		{ 182, 46, 59 },
		{ 182, 47, 59 },
		{ 182, 48, 59 },
		{ 182, 49, 59 },
		{ 182, 50, 59 },
		{ 182, 51, 59 },
		{ 182, 52, 59 },
		{ 182, 53, 59 },
		{ 182, 54, 59 },
		{ 182, 55, 59 },
		{ 182, 56, 59 },
		{ 182, 57, 59 },
		{ 182, 58, 59 },
		{ 182, 59, 59 },
		{ 182, 60, 59 },
		{ 182, 61, 59 },
		{ 182, 62, 59 },
		{ 182, 63, 59 },
		{ 182, 64, 59 },
		{ 182, 65, 59 },
		{ 182, 66, 59 },
		{ 182, 67, 59 },
		{ 182, 68, 59 },
		{ 182, 69, 59 },
		{ 182, 70, 59 },
		{ 182, 71, 59 },
		{ 182, 72, 59 },
		{ 182, 73, 59 },
		{ 182, 74, 59 },
		{ 182, 75, 59 },
		{ 182, 76, 59 },
		{ 182, 77, 59 },
		{ 182, 78, 59 },
		{ 182, 79, 59 },
		{ 182, 80, 59 },
		{ 182, 81, 59 },
		{ 182, 82, 59 },
		{ 182, 83, 59 },
		{ 182, 84, 59 },
		{ 182, 85, 59 },
		{ 182, 86, 59 },
		{ 182, 87, 59 },
		{ 182, 88, 59 },
		{ 182, 89, 59 },
		{ 182, 90, 59 },
		{ 182, 91, 59 },
		{ 182, 92, 59 },
		{ 182, 93, 59 },
		{ 182, 94, 59 },
		{ 182, 95, 59 },
		{ 182, 96, 59 },
		{ 182, 97, 59 },
		{ 182, 98, 59 },
		{ 182, 99, 59 },
		{ 182, 100, 59 },
		{ 182, 101, 59 },
		{ 182, 102, 59 },
		{ 182, 103, 59 },
		{ 182, 104, 59 },
		{ 182, 105, 59 },
		{ 182, 106, 59 },
		{ 183, 1, 59 },
		{ 183, 2, 59 },
		{ 183, 3, 59 },
		{ 183, 4, 59 },
		{ 183, 5, 59 },
		{ 183, 6, 59 },
		{ 183, 7, 59 },
		{ 183, 8, 59 },
		{ 183, 9, 59 },
		{ 183, 10, 59 },
		{ 183, 11, 59 },
		{ 183, 12, 59 },
		{ 183, 13, 59 },
		{ 183, 14, 59 },
		{ 183, 15, 59 },
		{ 183, 16, 59 },
		{ 183, 17, 59 },
		{ 183, 18, 59 },
		{ 183, 19, 59 },
		{ 183, 20, 59 },
		{ 183, 21, 59 },
		{ 183, 22, 59 },
		{ 183, 23, 59 },
		{ 183, 24, 59 },
		{ 183, 25, 59 },
		{ 183, 26, 59 },
		{ 183, 27, 59 },
		{ 183, 28, 59 },
		{ 183, 29, 59 },
		{ 183, 30, 59 },
		{ 183, 31, 59 },
		{ 183, 32, 59 },
		{ 183, 33, 59 },
		{ 183, 34, 59 },
		{ 183, 35, 59 },
		{ 183, 36, 59 },
		{ 183, 37, 59 },
		{ 183, 38, 59 },
		{ 183, 39, 59 },
		{ 183, 40, 59 },
		{ 183, 41, 59 },
		{ 183, 42, 59 },
		{ 183, 43, 59 },
		{ 183, 44, 59 },
		{ 183, 45, 59 },
		{ 183, 46, 59 },
		{ 183, 47, 59 },
		{ 183, 48, 59 },
		{ 183, 49, 59 },
		{ 183, 50, 59 },
		{ 183, 51, 59 },
		{ 183, 52, 59 },
		{ 183, 53, 59 },
		{ 183, 54, 59 },
		{ 183, 55, 59 },
		{ 183, 56, 59 },
		{ 183, 57, 59 },
		{ 183, 58, 59 },
		{ 183, 59, 59 },
		{ 183, 60, 59 },
		{ 183, 61, 59 },
		{ 183, 62, 59 },
		{ 183, 63, 59 },
		{ 183, 64, 59 },
		{ 183, 65, 59 },
		{ 183, 66, 59 },
		{ 183, 67, 59 },
		{ 183, 68, 59 },
		{ 183, 69, 59 },
		{ 183, 70, 59 },
		{ 183, 71, 59 },
		{ 183, 72, 59 },
		{ 183, 73, 59 },
		{ 183, 74, 59 },
		{ 183, 75, 59 },
		{ 183, 76, 59 },
		{ 183, 77, 59 },
		{ 183, 78, 59 },
		{ 183, 79, 59 },
		{ 183, 80, 59 },
		{ 183, 81, 59 },
		{ 183, 82, 59 },
		{ 183, 83, 59 },
		{ 183, 84, 59 },
		{ 183, 85, 59 },
		{ 183, 86, 59 },
		{ 183, 87, 59 },
		{ 183, 88, 59 },
		{ 183, 89, 59 },
		{ 183, 90, 59 },
		{ 183, 91, 59 },
		{ 183, 92, 59 },
		{ 183, 93, 59 },
		{ 183, 94, 59 },
		{ 183, 95, 59 },
		{ 183, 96, 59 },
		{ 183, 97, 59 },
		{ 183, 98, 59 },
		{ 183, 99, 59 },
		{ 183, 100, 59 },
		{ 183, 101, 59 },
		{ 183, 102, 59 },
		{ 183, 103, 59 },
		{ 183, 104, 59 },
		{ 183, 105, 59 },
		{ 183, 106, 59 },
		{ 183, 107, 59 },
		{ 183, 108, 59 },
		{ 184, 1, 59 },
		{ 184, 2, 59 },
		{ 184, 3, 59 },
		{ 184, 4, 59 },
		{ 184, 5, 59 },
		{ 184, 6, 59 },
		{ 184, 7, 59 },
		{ 184, 8, 59 },
		{ 184, 9, 59 },
		{ 184, 10, 59 },
		{ 184, 11, 59 },
		{ 184, 12, 59 },
		{ 184, 13, 59 },
		{ 184, 14, 59 },
		{ 184, 15, 59 },
		{ 184, 16, 59 },
		{ 184, 17, 59 },
		{ 184, 18, 59 },
		{ 184, 19, 59 },
		{ 184, 20, 59 },
		{ 184, 21, 59 },
		{ 184, 22, 59 },
		{ 184, 23, 59 },
		{ 184, 24, 59 },
		{ 184, 25, 59 },
		{ 184, 26, 59 },
		{ 184, 27, 59 },
		{ 184, 28, 59 },
		{ 184, 29, 59 },
		{ 184, 30, 59 },
		{ 184, 31, 59 },
		{ 184, 32, 59 },
		{ 184, 33, 59 },
		{ 184, 34, 59 },
		{ 184, 35, 59 },
		{ 184, 36, 59 },
		{ 184, 37, 59 },
		{ 184, 38, 59 },
		{ 184, 39, 59 },
		{ 184, 40, 59 },
		{ 184, 41, 59 },
		{ 184, 42, 59 },
		{ 184, 43, 59 },
		{ 184, 44, 59 },
		{ 184, 45, 59 },
		{ 184, 46, 59 },
		{ 184, 47, 59 },
		{ 184, 48, 59 },
		{ 184, 49, 59 },
		{ 184, 50, 59 },
		{ 184, 51, 59 },
		{ 184, 52, 59 },
		{ 184, 53, 59 },
		{ 184, 54, 59 },
		{ 184, 55, 59 },
		{ 184, 56, 59 },
		{ 184, 57, 59 },
		{ 184, 58, 59 },
		{ 184, 59, 59 },
		{ 184, 60, 59 },
		{ 184, 61, 59 },
		{ 184, 62, 59 },
		{ 184, 63, 59 },
		{ 184, 64, 59 },
		{ 184, 65, 59 },
		{ 184, 66, 59 },
		{ 184, 67, 59 },
		{ 184, 68, 59 },
		{ 184, 69, 59 },
		{ 184, 70, 59 },
		{ 184, 71, 59 },
		{ 184, 72, 59 },
		{ 184, 73, 59 },
		{ 184, 74, 59 },
		{ 184, 75, 59 },
		{ 184, 76, 59 },
		{ 184, 77, 59 },
		{ 184, 78, 59 },
		{ 184, 79, 59 },
		{ 184, 80, 59 },
		{ 184, 81, 59 },
		{ 184, 82, 59 },
		{ 184, 83, 59 },
		{ 184, 84, 59 },
		{ 184, 85, 59 },
		{ 184, 86, 59 },
		{ 184, 87, 59 },
		{ 184, 88, 59 },
		{ 184, 89, 59 },
		{ 184, 90, 59 },
		{ 184, 91, 59 },
		{ 184, 92, 59 },
		{ 184, 93, 59 },
		{ 184, 94, 59 },
		{ 184, 95, 59 },
		{ 184, 96, 59 },
		{ 184, 97, 59 },
		{ 184, 98, 59 },
		{ 184, 99, 59 },
		{ 184, 100, 59 },
		{ 184, 101, 59 },
		{ 184, 102, 59 },
		{ 184, 103, 59 },
		{ 184, 104, 59 },
		{ 184, 105, 59 },
		{ 184, 106, 59 },
		{ 184, 107, 59 },
		{ 184, 108, 59 },
		{ 184, 109, 59 },
		{ 184, 110, 59 },
		{ 185, 1, 59 },
		{ 185, 2, 59 },
		{ 185, 3, 59 },
		{ 185, 4, 59 },
		{ 185, 5, 59 },
		{ 185, 6, 59 },
		{ 185, 7, 59 },
		{ 185, 8, 59 },
		{ 185, 9, 59 },
		{ 185, 10, 59 },
		{ 185, 11, 59 },
		{ 185, 12, 59 },
		{ 185, 13, 59 },
		{ 185, 14, 59 },
		{ 185, 15, 59 },
		{ 185, 16, 59 },
		{ 185, 17, 59 },
		{ 185, 18, 59 },
		{ 185, 19, 59 },
		{ 185, 20, 59 },
		{ 185, 21, 59 },
		{ 185, 22, 59 },
		{ 185, 23, 59 },
		{ 185, 24, 59 },
		{ 185, 25, 59 },
		{ 185, 26, 59 },
		{ 185, 27, 59 },
		{ 185, 28, 59 },
		{ 185, 29, 59 },
		{ 185, 30, 59 },
		{ 185, 31, 59 },
		{ 185, 32, 59 },
		{ 185, 33, 59 },
		{ 185, 34, 59 },
		{ 185, 35, 59 },
		{ 185, 36, 59 },
		{ 185, 37, 59 },
		{ 185, 38, 59 },
		{ 185, 39, 59 },
		{ 185, 40, 59 },
		{ 185, 41, 59 },
		{ 185, 42, 59 },
		{ 185, 43, 59 },
		{ 185, 44, 59 },
		{ 185, 45, 59 },
		{ 185, 46, 59 },
		{ 185, 47, 59 },
		{ 185, 48, 59 },
		{ 185, 49, 59 },
		{ 185, 50, 59 },
		{ 185, 51, 59 },
		{ 185, 52, 59 },
		{ 185, 53, 59 },
		{ 185, 54, 59 },
		{ 185, 55, 59 },
		{ 185, 56, 59 },
		{ 185, 57, 59 },
		{ 185, 58, 59 },
		{ 185, 59, 59 },
		{ 185, 60, 59 },
		{ 185, 61, 59 },
		{ 185, 62, 59 },
		{ 185, 63, 59 },
		{ 185, 64, 59 },
		{ 185, 65, 59 },
		{ 185, 66, 59 },
		{ 185, 67, 59 },
		{ 185, 68, 59 },
		{ 185, 69, 59 },
		{ 185, 70, 59 },
		{ 185, 71, 59 },
		{ 185, 72, 59 },
		{ 185, 73, 59 },
		{ 185, 74, 59 },
		{ 185, 75, 59 },
		{ 185, 76, 59 },
		{ 185, 77, 59 },
		{ 185, 78, 59 },
		{ 185, 79, 59 },
		{ 185, 80, 59 },
		{ 185, 81, 59 },
		{ 185, 82, 59 },
		{ 185, 83, 59 },
		{ 185, 84, 59 },
		{ 185, 85, 59 },
		{ 185, 86, 59 },
		{ 185, 87, 59 },
		{ 185, 88, 59 },
		{ 185, 89, 59 },
		{ 185, 90, 59 },
		{ 185, 91, 59 },
		{ 185, 92, 59 },
		{ 185, 93, 59 },
		{ 185, 94, 59 },
		{ 185, 95, 59 },
		{ 185, 96, 59 },
		{ 185, 97, 59 },
		{ 185, 98, 59 },
		{ 185, 99, 59 },
		{ 185, 100, 59 },
		{ 185, 101, 59 },
		{ 185, 102, 59 },
		{ 185, 103, 59 },
		{ 185, 104, 59 },
		{ 185, 105, 59 },
		{ 185, 106, 59 },
		{ 185, 107, 59 },
		{ 185, 108, 59 },
		{ 185, 109, 59 },
		{ 185, 110, 59 },
		{ 185, 111, 59 },
		{ 185, 112, 59 },
		{ 186, 1, 59 },
		{ 186, 2, 59 },
		{ 186, 3, 59 },
		{ 186, 4, 59 },
		{ 186, 5, 59 },
		{ 186, 6, 59 },
		{ 186, 7, 59 },
		{ 186, 8, 59 },
		{ 186, 9, 59 },
		{ 186, 10, 59 },
		{ 186, 11, 59 },
		{ 186, 12, 59 },
		{ 186, 13, 59 },
		{ 186, 14, 59 },
		{ 186, 15, 59 },
		{ 186, 16, 59 },
		{ 186, 17, 59 },
		{ 186, 18, 59 },
		{ 186, 19, 59 },
		{ 186, 20, 59 },
		{ 186, 21, 59 },
		{ 186, 22, 59 },
		{ 186, 23, 59 },
		{ 186, 24, 59 },
		{ 186, 25, 59 },
		{ 186, 26, 59 },
		{ 186, 27, 59 },
		{ 186, 28, 59 },
		{ 186, 29, 59 },
		{ 186, 30, 59 },
		{ 186, 31, 59 },
		{ 186, 32, 59 },
		{ 186, 33, 59 },
		{ 186, 34, 59 },
		{ 186, 35, 59 },
		{ 186, 36, 59 },
		{ 186, 37, 59 },
		{ 186, 38, 59 },
		{ 186, 39, 59 },
		{ 186, 40, 59 },
		{ 186, 41, 59 },
		{ 186, 42, 59 },
		{ 186, 43, 59 },
		{ 186, 44, 59 },
		{ 186, 45, 59 },
		{ 186, 46, 59 },
		{ 186, 47, 59 },
		{ 186, 48, 59 },
		{ 186, 49, 59 },
		{ 186, 50, 59 },
		{ 186, 51, 59 },
		{ 186, 52, 59 },
		{ 186, 53, 59 },
		{ 186, 54, 59 },
		{ 186, 55, 59 },
		{ 186, 56, 59 },
		{ 186, 57, 59 },
		{ 186, 58, 59 },
		{ 186, 59, 59 },
		{ 186, 60, 59 },
		{ 186, 61, 59 },
		{ 186, 62, 59 },
		{ 186, 63, 59 },
		{ 186, 64, 59 },
		{ 186, 65, 59 },
		{ 186, 66, 59 },
		{ 186, 67, 59 },
		{ 186, 68, 59 },
		{ 186, 69, 59 },
		{ 186, 70, 59 },
		{ 186, 71, 59 },
		{ 186, 72, 59 },
		{ 186, 73, 59 },
		{ 186, 74, 59 },
		{ 186, 75, 59 },
		{ 186, 76, 59 },
		{ 186, 77, 59 },
		{ 186, 78, 59 },
		{ 186, 79, 59 },
		{ 186, 80, 59 },
		{ 186, 81, 59 },
		{ 186, 82, 59 },
		{ 186, 83, 59 },
		{ 186, 84, 59 },
		{ 186, 85, 59 },
		{ 186, 86, 59 },
		{ 186, 87, 59 },
		{ 186, 88, 59 },
		{ 186, 89, 59 },
		{ 186, 90, 59 },
		{ 186, 91, 59 },
		{ 186, 92, 59 },
		{ 186, 93, 59 },
		{ 186, 94, 59 },
		{ 186, 95, 59 },
		{ 186, 96, 59 },
		{ 186, 97, 59 },
		{ 186, 98, 59 },
		{ 186, 99, 59 },
		{ 186, 100, 59 },
		{ 186, 101, 59 },
		{ 186, 102, 59 },
		{ 186, 103, 59 },
		{ 186, 104, 59 },
		{ 186, 105, 59 },
		{ 186, 106, 59 },
		{ 186, 107, 59 },
		{ 186, 108, 59 },
		{ 186, 109, 59 },
		{ 186, 110, 59 },
		{ 186, 111, 59 },
		{ 186, 112, 59 },
		{ 186, 113, 59 },
		{ 186, 114, 59 },
		{ 187, 1, 59 },
		{ 187, 2, 59 },
		{ 187, 3, 59 },
		{ 187, 4, 59 },
		{ 187, 5, 59 },
		{ 187, 6, 59 },
		{ 187, 7, 59 },
		{ 187, 8, 59 },
		{ 187, 9, 59 },
		{ 187, 10, 59 },
		{ 187, 11, 59 },
		{ 187, 12, 59 },
		{ 187, 13, 59 },
		{ 187, 14, 59 },
		{ 187, 15, 59 },
		{ 187, 16, 59 },
		{ 187, 17, 59 },
		{ 187, 18, 59 },
		{ 187, 19, 59 },
		{ 187, 20, 59 },
		{ 187, 21, 59 },
		{ 187, 22, 59 },
		{ 187, 23, 59 },
		{ 187, 24, 59 },
		{ 187, 25, 59 },
		{ 187, 26, 59 },
		{ 187, 27, 59 },
		{ 187, 28, 59 },
		{ 187, 29, 59 },
		{ 187, 30, 59 },
		{ 187, 31, 59 },
		{ 187, 32, 59 },
		{ 187, 33, 59 },
		{ 187, 34, 59 },
		{ 187, 35, 59 },
		{ 187, 36, 59 },
		{ 187, 37, 59 },
		{ 187, 38, 59 },
		{ 187, 39, 59 },
		{ 187, 40, 59 },
		{ 187, 41, 59 },
		{ 187, 42, 59 },
		{ 187, 43, 59 },
		{ 187, 44, 59 },
		{ 187, 45, 59 },
		{ 187, 46, 59 },
		{ 187, 47, 59 },
		{ 187, 48, 59 },
		{ 187, 49, 59 },
		{ 187, 50, 59 },
		{ 187, 51, 59 },
		{ 187, 52, 59 },
		{ 187, 53, 59 },
		{ 187, 54, 59 },
		{ 187, 55, 59 },
		{ 187, 56, 59 },
		{ 187, 57, 59 },
		{ 187, 58, 59 },
		{ 187, 59, 59 },
		{ 187, 60, 59 },
		{ 187, 61, 59 },
		{ 187, 62, 59 },
		{ 187, 63, 59 },
		{ 187, 64, 59 },
		{ 187, 65, 59 },
		{ 187, 66, 59 },
		{ 187, 67, 59 },
		{ 187, 68, 59 },
		{ 187, 69, 59 },
		{ 187, 70, 59 },
		{ 187, 71, 59 },
		{ 187, 72, 59 },
		{ 187, 73, 59 },
		{ 187, 74, 59 },
		{ 187, 75, 59 },
		{ 187, 76, 59 },
		{ 187, 77, 59 },
		{ 187, 78, 59 },
		{ 187, 79, 59 },
		{ 187, 80, 59 },
		{ 187, 81, 59 },
		{ 187, 82, 59 },
		{ 187, 83, 59 },
		{ 187, 84, 59 },
		{ 187, 85, 59 },
		{ 187, 86, 59 },
		{ 187, 87, 59 },
		{ 187, 88, 59 },
		{ 187, 89, 59 },
		{ 187, 90, 59 },
		{ 187, 91, 59 },
		{ 187, 92, 59 },
		{ 187, 93, 59 },
		{ 187, 94, 59 },
		{ 187, 95, 59 },
		{ 187, 96, 59 },
		{ 187, 97, 59 },
		{ 187, 98, 59 },
		{ 187, 99, 59 },
		{ 187, 100, 59 },
		{ 187, 101, 59 },
		{ 187, 102, 59 },
		{ 187, 103, 59 },
		{ 187, 104, 59 },
		{ 187, 105, 59 },
		{ 187, 106, 59 },
		{ 187, 107, 59 },
		{ 187, 108, 59 },
		{ 187, 109, 59 },
		{ 187, 110, 59 },
		{ 187, 111, 59 },
		{ 187, 112, 59 },
		{ 187, 113, 59 },
		{ 187, 114, 59 },
		{ 187, 115, 59 },
		{ 187, 116, 59 },
		{ 188, 1, 59 },
		{ 188, 2, 59 },
		{ 188, 3, 59 },
		{ 188, 4, 59 },
		{ 188, 5, 59 },
		{ 188, 6, 59 },
		{ 188, 7, 59 },
		{ 188, 8, 59 },
		{ 188, 9, 59 },
		{ 188, 10, 59 },
		{ 188, 11, 59 },
		{ 188, 12, 59 },
		{ 188, 13, 59 },
		{ 188, 14, 59 },
		{ 188, 15, 59 },
		{ 188, 16, 59 },
		{ 188, 17, 59 },
		{ 188, 18, 59 },
		{ 188, 19, 59 },
		{ 188, 20, 59 },
		{ 188, 21, 59 },
		{ 188, 22, 59 },
		{ 188, 23, 59 },
		{ 188, 24, 59 },
		{ 188, 25, 59 },
		{ 188, 26, 59 },
		{ 188, 27, 59 },
		{ 188, 28, 59 },
		{ 188, 29, 59 },
		{ 188, 30, 59 },
		{ 188, 31, 59 },
		{ 188, 32, 59 },
		{ 188, 33, 59 },
		{ 188, 34, 59 },
		{ 188, 35, 59 },
		{ 188, 36, 59 },
		{ 188, 37, 59 },
		{ 188, 38, 59 },
		{ 188, 39, 59 },
		{ 188, 40, 59 },
		{ 188, 41, 59 },
		{ 188, 42, 59 },
		{ 188, 43, 59 },
		{ 188, 44, 59 },
		{ 188, 45, 59 },
		{ 188, 46, 59 },
		{ 188, 47, 59 },
		{ 188, 48, 59 },
		{ 188, 49, 59 },
		{ 188, 50, 59 },
		{ 188, 51, 59 },
		{ 188, 52, 59 },
		{ 188, 53, 59 },
		{ 188, 54, 59 },
		{ 188, 55, 59 },
		{ 188, 56, 59 },
		{ 188, 57, 59 },
		{ 188, 58, 59 },
		{ 188, 59, 59 },
		{ 188, 60, 59 },
		{ 188, 61, 59 },
		{ 188, 62, 59 },
		{ 188, 63, 59 },
		{ 188, 64, 59 },
		{ 188, 65, 59 },
		{ 188, 66, 59 },
		{ 188, 67, 59 },
		{ 188, 68, 59 },
		{ 188, 69, 59 },
		{ 188, 70, 59 },
		{ 188, 71, 59 },
		{ 188, 72, 59 },
		{ 188, 73, 59 },
		{ 188, 74, 59 },
		{ 188, 75, 59 },
		{ 188, 76, 59 },
		{ 188, 77, 59 },
		{ 188, 78, 59 },
		{ 188, 79, 59 },
		{ 188, 80, 59 },
		{ 188, 81, 59 },
		{ 188, 82, 59 },
		{ 188, 83, 59 },
		{ 188, 84, 59 },
		{ 188, 85, 59 },
		{ 188, 86, 59 },
		{ 188, 87, 59 },
		{ 188, 88, 59 },
		{ 188, 89, 59 },
		{ 188, 90, 59 },
		{ 188, 91, 59 },
		{ 188, 92, 59 },
		{ 188, 93, 59 },
		{ 188, 94, 59 },
		{ 188, 95, 59 },
		{ 188, 96, 59 },
		{ 188, 97, 59 },
		{ 188, 98, 59 },
		{ 188, 99, 59 },
		{ 188, 100, 59 },
		{ 188, 101, 59 },
		{ 188, 102, 59 },
		{ 188, 103, 59 },
		{ 188, 104, 59 },
		{ 188, 105, 59 },
		{ 188, 106, 59 },
		{ 188, 107, 59 },
		{ 188, 108, 59 },
		{ 188, 109, 59 },
		{ 188, 110, 59 },
		{ 188, 111, 59 },
		{ 188, 112, 59 },
		{ 188, 113, 59 },
		{ 188, 114, 59 },
		{ 188, 115, 59 },
		{ 188, 116, 59 },
		{ 188, 117, 59 },
		{ 188, 118, 59 },
		{ 189, 1, 59 },
		{ 189, 2, 59 },
		{ 189, 3, 59 },
		{ 189, 4, 59 },
		{ 189, 5, 59 },
		{ 189, 6, 59 },
		{ 189, 7, 59 },
		{ 189, 8, 59 },
		{ 189, 9, 59 },
		{ 189, 10, 59 },
		{ 189, 11, 59 },
		{ 189, 12, 59 },
		{ 189, 13, 59 },
		{ 189, 14, 59 },
		{ 189, 15, 59 },
		{ 189, 16, 59 },
		{ 189, 17, 59 },
		{ 189, 18, 59 },
		{ 189, 19, 59 },
		{ 189, 20, 59 },
		{ 189, 21, 59 },
		{ 189, 22, 59 },
		{ 189, 23, 59 },
		{ 189, 24, 59 },
		{ 189, 25, 59 },
		{ 189, 26, 59 },
		{ 189, 27, 59 },
		{ 189, 28, 59 },
		{ 189, 29, 59 },
		{ 189, 30, 59 },
		{ 189, 31, 59 },
		{ 189, 32, 59 },
		{ 189, 33, 59 },
		{ 189, 34, 59 },
		{ 189, 35, 59 },
		{ 189, 36, 59 },
		{ 189, 37, 59 },
		{ 189, 38, 59 },
		{ 189, 39, 59 },
		{ 189, 40, 59 },
		{ 189, 41, 59 },
		{ 189, 42, 59 },
		{ 189, 43, 59 },
		{ 189, 44, 59 },
		{ 189, 45, 59 },
		{ 189, 46, 59 },
		{ 189, 47, 59 },
		{ 189, 48, 59 },
		{ 189, 49, 59 },
		{ 189, 50, 59 },
		{ 189, 51, 59 },
		{ 189, 52, 59 },
		{ 189, 53, 59 },
		{ 189, 54, 59 },
		{ 189, 55, 59 },
		{ 189, 56, 59 },
		{ 189, 57, 59 },
		{ 189, 58, 59 },
		{ 189, 59, 59 },
		{ 189, 60, 59 },
		{ 189, 61, 59 },
		{ 189, 62, 59 },
		{ 189, 63, 59 },
		{ 189, 64, 59 },
		{ 189, 65, 59 },
		{ 189, 66, 59 },
		{ 189, 67, 59 },
		{ 189, 68, 59 },
		{ 189, 69, 59 },
		{ 189, 70, 59 },
		{ 189, 71, 59 },
		{ 189, 72, 59 },
		{ 189, 73, 59 },
		{ 189, 74, 59 },
		{ 189, 75, 59 },
		{ 189, 76, 59 },
		{ 189, 77, 59 },
		{ 189, 78, 59 },
		{ 189, 79, 59 },
		{ 189, 80, 59 },
		{ 189, 81, 59 },
		{ 189, 82, 59 },
		{ 189, 83, 59 },
		{ 189, 84, 59 },
		{ 189, 85, 59 },
		{ 189, 86, 59 },
		{ 189, 87, 59 },
		{ 189, 88, 59 },
		{ 189, 89, 59 },
		{ 189, 90, 59 },
		{ 189, 91, 59 },
		{ 189, 92, 59 },
		{ 189, 93, 59 },
		{ 189, 94, 59 },
		{ 189, 95, 59 },
		{ 189, 96, 59 },
		{ 189, 97, 59 },
		{ 189, 98, 59 },
		{ 189, 99, 59 },
		{ 189, 100, 59 },
		{ 189, 101, 59 },
		{ 189, 102, 59 },
		{ 189, 103, 59 },
		{ 189, 104, 59 },
		{ 189, 105, 59 },
		{ 189, 106, 59 },
		{ 189, 107, 59 },
		{ 189, 108, 59 },
		{ 189, 109, 59 },
		{ 189, 110, 59 },
		{ 189, 111, 59 },
		{ 189, 112, 59 },
		{ 189, 113, 59 },
		{ 189, 114, 59 },
		{ 189, 115, 59 },
		{ 189, 116, 59 },
		{ 189, 117, 59 },
		{ 189, 118, 59 },
		{ 189, 119, 59 },
		{ 189, 120, 59 },
		{ 190, 1, 59 },
		{ 190, 2, 59 },
		{ 190, 3, 59 },
		{ 190, 4, 59 },
		{ 190, 5, 59 },
		{ 190, 6, 59 },
		{ 190, 7, 59 },
		{ 190, 8, 59 },
		{ 190, 9, 59 },
		{ 190, 10, 59 },
		{ 190, 11, 59 },
		{ 190, 12, 59 },
		{ 190, 13, 59 },
		{ 190, 14, 59 },
		{ 190, 15, 59 },
		{ 190, 16, 59 },
		{ 190, 17, 59 },
		{ 190, 18, 59 },
		{ 190, 19, 59 },
		{ 190, 20, 59 },
		{ 190, 21, 59 },
		{ 190, 22, 59 },
		{ 190, 23, 59 },
		{ 190, 24, 59 },
		{ 190, 25, 59 },
		{ 190, 26, 59 },
		{ 190, 27, 59 },
		{ 190, 28, 59 },
		{ 190, 29, 59 },
		{ 190, 30, 59 },
		{ 190, 31, 59 },
		{ 190, 32, 59 },
		{ 190, 33, 59 },
		{ 190, 34, 59 },
		{ 190, 35, 59 },
		{ 190, 36, 59 },
		{ 190, 37, 59 },
		{ 190, 38, 59 },
		{ 190, 39, 59 },
		{ 190, 40, 59 },
		{ 190, 41, 59 },
		{ 190, 42, 59 },
		{ 190, 43, 59 },
		{ 190, 44, 59 },
		{ 190, 45, 59 },
		{ 190, 46, 59 },
		{ 190, 47, 59 },
		{ 190, 48, 59 },
		{ 190, 49, 59 },
		{ 190, 50, 59 },
		{ 190, 51, 59 },
		{ 190, 52, 59 },
		{ 190, 53, 59 },
		{ 190, 54, 59 },
		{ 190, 55, 59 },
		{ 190, 56, 59 },
		{ 190, 57, 59 },
		{ 190, 58, 59 },
		{ 190, 59, 59 },
		{ 190, 60, 59 },
		{ 190, 61, 59 },
		{ 190, 62, 59 },
		{ 190, 63, 59 },
		{ 190, 64, 59 },
		{ 190, 65, 59 },
		{ 190, 66, 59 },
		{ 190, 67, 59 },
		{ 190, 68, 59 },
		{ 190, 69, 59 },
		{ 190, 70, 59 },
		{ 190, 71, 59 },
		{ 190, 72, 59 },
		{ 190, 73, 59 },
		{ 190, 74, 59 },
		{ 190, 75, 59 },
		{ 190, 76, 59 },
		{ 190, 77, 59 },
		{ 190, 78, 59 },
		{ 190, 79, 59 },
		{ 190, 80, 59 },
		{ 190, 81, 59 },
		{ 190, 82, 59 },
		{ 190, 83, 59 },
		{ 190, 84, 59 },
		{ 190, 85, 59 },
		{ 190, 86, 59 },
		{ 190, 87, 59 },
		{ 190, 88, 59 },
		{ 190, 89, 59 },
		{ 190, 90, 59 },
		{ 190, 91, 59 },
		{ 190, 92, 59 },
		{ 190, 93, 59 },
		{ 190, 94, 59 },
		{ 190, 95, 59 },
		{ 190, 96, 59 },
		{ 190, 97, 59 },
		{ 190, 98, 59 },
		{ 190, 99, 59 },
		{ 190, 100, 59 },
		{ 190, 101, 59 },
		{ 190, 102, 59 },
		{ 190, 103, 59 },
		{ 190, 104, 59 },
		{ 190, 105, 59 },
		{ 190, 106, 59 },
		{ 190, 107, 59 },
		{ 190, 108, 59 },
		{ 190, 109, 59 },
		{ 190, 110, 59 },
		{ 190, 111, 59 },
		{ 190, 112, 59 },
		{ 190, 113, 59 },
		{ 190, 114, 59 },
		{ 190, 115, 59 },
		{ 190, 116, 59 },
		{ 190, 117, 59 },
		{ 190, 118, 59 },
		{ 190, 119, 59 },
		{ 190, 120, 59 },
		{ 190, 121, 59 },
		{ 190, 122, 59 },
		{ 191, 1, 59 },
		{ 191, 2, 59 },
		{ 191, 3, 59 },
		{ 191, 4, 59 },
		{ 191, 5, 59 },
		{ 191, 6, 59 },
		{ 191, 7, 59 },
		{ 191, 8, 59 },
		{ 191, 9, 59 },
		{ 191, 10, 59 },
		{ 191, 11, 59 },
		{ 191, 12, 59 },
		{ 191, 13, 59 },
		{ 191, 14, 59 },
		{ 191, 15, 59 },
		{ 191, 16, 59 },
		{ 191, 17, 59 },
		{ 191, 18, 59 },
		{ 191, 19, 59 },
		{ 191, 20, 59 },
		{ 191, 21, 59 },
		{ 191, 22, 59 },
		{ 191, 23, 59 },
		{ 191, 24, 59 },
		{ 191, 25, 59 },
		{ 191, 26, 59 },
		{ 191, 27, 59 },
		{ 191, 28, 59 },
		{ 191, 29, 59 },
		{ 191, 30, 59 },
		{ 191, 31, 59 },
		{ 191, 32, 59 },
		{ 191, 33, 59 },
		{ 191, 34, 59 },
		{ 191, 35, 59 },
		{ 191, 36, 59 },
		{ 191, 37, 59 },
		{ 191, 38, 59 },
		{ 191, 39, 59 },
		{ 191, 40, 59 },
		{ 191, 41, 59 },
		{ 191, 42, 59 },
		{ 191, 43, 59 },
		{ 191, 44, 59 },
		{ 191, 45, 59 },
		{ 191, 46, 59 },
		{ 191, 47, 59 },
		{ 191, 48, 59 },
		{ 191, 49, 59 },
		{ 191, 50, 59 },
		{ 191, 51, 59 },
		{ 191, 52, 59 },
		{ 191, 53, 59 },
		{ 191, 54, 59 },
		{ 191, 55, 59 },
		{ 191, 56, 59 },
		{ 191, 57, 59 },
		{ 191, 58, 59 },
		{ 191, 59, 59 },
		{ 191, 60, 59 },
		{ 191, 61, 59 },
		{ 191, 62, 59 },
		{ 191, 63, 59 },
		{ 191, 64, 59 },
		{ 191, 65, 59 },
		{ 191, 66, 59 },
		{ 191, 67, 59 },
		{ 191, 68, 59 },
		{ 191, 69, 59 },
		{ 191, 70, 59 },
		{ 191, 71, 59 },
		{ 191, 72, 59 },
		{ 191, 73, 59 },
		{ 191, 74, 59 },
		{ 191, 75, 59 },
		{ 191, 76, 59 },
		{ 191, 77, 59 },
		{ 191, 78, 59 },
		{ 191, 79, 59 },
		{ 191, 80, 59 },
		{ 191, 81, 59 },
		{ 191, 82, 59 },
		{ 191, 83, 59 },
		{ 191, 84, 59 },
		{ 191, 85, 59 },
		{ 191, 86, 59 },
		{ 191, 87, 59 },
		{ 191, 88, 59 },
		{ 191, 89, 59 },
		{ 191, 90, 59 },
		{ 191, 91, 59 },
		{ 191, 92, 59 },
		{ 191, 93, 59 },
		{ 191, 94, 59 },
		{ 191, 95, 59 },
		{ 191, 96, 59 },
		{ 191, 97, 59 },
		{ 191, 98, 59 },
		{ 191, 99, 59 },
		{ 191, 100, 59 },
		{ 191, 101, 59 },
		{ 191, 102, 59 },
		{ 191, 103, 59 },
		{ 191, 104, 59 },
		{ 191, 105, 59 },
		{ 191, 106, 59 },
		{ 191, 107, 59 },
		{ 191, 108, 59 },
		{ 191, 109, 59 },
		{ 191, 110, 59 },
		{ 191, 111, 59 },
		{ 191, 112, 59 },
		{ 191, 113, 59 },
		{ 191, 114, 59 },
		{ 191, 115, 59 },
		{ 191, 116, 59 },
		{ 191, 117, 59 },
		{ 191, 118, 59 },
		{ 191, 119, 59 },
		{ 191, 120, 59 },
		{ 191, 121, 59 },
		{ 191, 122, 59 },
		{ 191, 123, 59 },
		{ 191, 124, 59 },
		{ 192, 1, 59 },
		{ 192, 2, 59 },
		{ 192, 3, 59 },
		{ 192, 4, 59 },
		{ 192, 5, 59 },
		{ 192, 6, 59 },
		{ 192, 7, 59 },
		{ 192, 8, 59 },
		{ 192, 9, 59 },
		{ 192, 10, 59 },
		{ 192, 11, 59 },
		{ 192, 12, 59 },
		{ 192, 13, 59 },
		{ 192, 14, 59 },
		{ 192, 15, 59 },
		{ 192, 16, 59 },
		{ 192, 17, 59 },
		{ 192, 18, 59 },
		{ 192, 19, 59 },
		{ 192, 20, 59 },
		{ 192, 21, 59 },
		{ 192, 22, 59 },
		{ 192, 23, 59 },
		{ 192, 24, 59 },
		{ 192, 25, 59 },
		{ 192, 26, 59 },
		{ 192, 27, 59 },
		{ 192, 28, 59 },
		{ 192, 29, 59 },
		{ 192, 30, 59 },
		{ 192, 31, 59 },
		{ 192, 32, 59 },
		{ 192, 33, 59 },
		{ 192, 34, 59 },
		{ 192, 35, 59 },
		{ 192, 36, 59 },
		{ 192, 37, 59 },
		{ 192, 38, 59 },
		{ 192, 39, 59 },
		{ 192, 40, 59 },
		{ 192, 41, 59 },
		{ 192, 42, 59 },
		{ 192, 43, 59 },
		{ 192, 44, 59 },
		{ 192, 45, 59 },
		{ 192, 46, 59 },
		{ 192, 47, 59 },
		{ 192, 48, 59 },
		{ 192, 49, 59 },
		{ 192, 50, 59 },
		{ 192, 51, 59 },
		{ 192, 52, 59 },
		{ 192, 53, 59 },
		{ 192, 54, 59 },
		{ 192, 55, 59 },
		{ 192, 56, 59 },
		{ 192, 57, 59 },
		{ 192, 58, 59 },
		{ 192, 59, 59 },
		{ 192, 60, 59 },
		{ 192, 61, 59 },
		{ 192, 62, 59 },
		{ 192, 63, 59 },
		{ 192, 64, 59 },
		{ 192, 65, 59 },
		{ 192, 66, 59 },
		{ 192, 67, 59 },
		{ 192, 68, 59 },
		{ 192, 69, 59 },
		{ 192, 70, 59 },
		{ 192, 71, 59 },
		{ 192, 72, 59 },
		{ 192, 73, 59 },
		{ 192, 74, 59 },
		{ 192, 75, 59 },
		{ 192, 76, 59 },
		{ 192, 77, 59 },
		{ 192, 78, 59 },
		{ 192, 79, 59 },
		{ 192, 80, 59 },
		{ 192, 81, 59 },
		{ 192, 82, 59 },
		{ 192, 83, 59 },
		{ 192, 84, 59 },
		{ 192, 85, 59 },
		{ 192, 86, 59 },
		{ 192, 87, 59 },
		{ 192, 88, 59 },
		{ 192, 89, 59 },
		{ 192, 90, 59 },
		{ 192, 91, 59 },
		{ 192, 92, 59 },
		{ 192, 93, 59 },
		{ 192, 94, 59 },
		{ 192, 95, 59 },
		{ 192, 96, 59 },
		{ 192, 97, 59 },
		{ 192, 98, 59 },
		{ 192, 99, 59 },
		{ 192, 100, 59 },
		{ 192, 101, 59 },
		{ 192, 102, 59 },
		{ 192, 103, 59 },
		{ 192, 104, 59 },
		{ 192, 105, 59 },
		{ 192, 106, 59 },
		{ 192, 107, 59 },
		{ 192, 108, 59 },
		{ 192, 109, 59 },
		{ 192, 110, 59 },
		{ 192, 111, 59 },
		{ 192, 112, 59 },
		{ 192, 113, 59 },
		{ 192, 114, 59 },
		{ 192, 115, 59 },
		{ 192, 116, 59 },
		{ 192, 117, 59 },
		{ 192, 118, 59 },
		{ 192, 119, 59 },
		{ 192, 120, 59 },
		{ 192, 121, 59 },
		{ 192, 122, 59 },
		{ 192, 123, 59 },
		{ 192, 124, 59 },
		{ 192, 125, 59 },
		{ 192, 126, 59 },
		{ 193, 2, 59 },
		{ 193, 3, 59 },
		{ 193, 4, 59 },
		{ 193, 5, 59 },
		{ 193, 6, 59 },
		{ 193, 7, 59 },
		{ 193, 8, 59 },
		{ 193, 9, 59 },
		{ 193, 10, 59 },
		{ 193, 11, 59 },
		{ 193, 12, 59 },
		{ 193, 13, 59 },
		{ 193, 14, 59 },
		{ 193, 15, 59 },
		{ 193, 16, 59 },
		{ 193, 17, 59 },
		{ 193, 18, 59 },
		{ 193, 19, 59 },
		{ 193, 20, 59 },
		{ 193, 21, 59 },
		{ 193, 22, 59 },
		{ 193, 23, 59 },
		{ 193, 24, 59 },
		{ 193, 25, 59 },
		{ 193, 26, 59 },
		{ 193, 27, 59 },
		{ 193, 28, 59 },
		{ 193, 29, 59 },
		{ 193, 30, 59 },
		{ 193, 31, 59 },
		{ 193, 32, 59 },
		{ 193, 33, 59 },
		{ 193, 34, 59 },
		{ 193, 35, 59 },
		{ 193, 36, 59 },
		{ 193, 37, 59 },
		{ 193, 38, 59 },
		{ 193, 39, 59 },
		{ 193, 40, 59 },
		{ 193, 41, 59 },
		{ 193, 42, 59 },
		{ 193, 43, 59 },
		{ 193, 44, 59 },
		{ 193, 45, 59 },
		{ 193, 46, 59 },
		{ 193, 47, 59 },
		{ 193, 48, 59 },
		{ 193, 49, 59 },
		{ 193, 50, 59 },
		{ 193, 51, 59 },
		{ 193, 52, 59 },
		{ 193, 53, 59 },
		{ 193, 54, 59 },
		{ 193, 55, 59 },
		{ 193, 56, 59 },
		{ 193, 57, 59 },
		{ 193, 58, 59 },
		{ 193, 59, 59 },
		{ 193, 60, 59 },
		{ 193, 61, 59 },
		{ 193, 62, 59 },
		{ 193, 63, 59 },
		{ 193, 64, 59 },
		{ 193, 65, 59 },
		{ 193, 66, 59 },
		{ 193, 67, 59 },
		{ 193, 68, 59 },
		{ 193, 69, 59 },
		{ 193, 70, 59 },
		{ 193, 71, 59 },
		{ 193, 72, 59 },
		{ 193, 73, 59 },
		{ 193, 74, 59 },
		{ 193, 75, 59 },
		{ 193, 76, 59 },
		{ 193, 77, 59 },
		{ 193, 78, 59 },
		{ 193, 79, 59 },
		{ 193, 80, 59 },
		{ 193, 81, 59 },
		{ 193, 82, 59 },
		{ 193, 83, 59 },
		{ 193, 84, 59 },
		{ 193, 85, 59 },
		{ 193, 86, 59 },
		{ 193, 87, 59 },
		{ 193, 88, 59 },
		{ 193, 89, 59 },
		{ 193, 90, 59 },
		{ 193, 91, 59 },
		{ 193, 92, 59 },
		{ 193, 93, 59 },
		{ 193, 94, 59 },
		{ 193, 95, 59 },
		{ 193, 96, 59 },
		{ 193, 97, 59 },
		{ 193, 98, 59 },
		{ 193, 99, 59 },
		{ 193, 100, 59 },
		{ 193, 101, 59 },
		{ 193, 102, 59 },
		{ 193, 103, 59 },
		{ 193, 104, 59 },
		{ 193, 105, 59 },
		{ 193, 106, 59 },
		{ 193, 107, 59 },
		{ 193, 108, 59 },
		{ 193, 109, 59 },
		{ 193, 110, 59 },
		{ 193, 111, 59 },
		{ 193, 112, 59 },
		{ 193, 113, 59 },
		{ 193, 114, 59 },
		{ 193, 115, 59 },
		{ 193, 116, 59 },
		{ 193, 117, 59 },
		{ 193, 118, 59 },
		{ 193, 119, 59 },
		{ 193, 120, 59 },
		{ 193, 121, 59 },
		{ 193, 122, 59 },
		{ 193, 123, 59 },
		{ 193, 124, 59 },
		{ 193, 125, 59 },
		{ 193, 126, 59 },
		{ 193, 127, 59 },
		{ 193, 128, 59 },
		{ 194, 6, 59 },
		{ 194, 7, 59 },
		{ 194, 8, 59 },
		{ 194, 9, 59 },
		{ 194, 10, 59 },
		{ 194, 11, 59 },
		{ 194, 12, 59 },
		{ 194, 13, 59 },
		{ 194, 14, 59 },
		{ 194, 15, 59 },
		{ 194, 16, 59 },
		{ 194, 17, 59 },
		{ 194, 18, 59 },
		{ 194, 19, 59 },
		{ 194, 20, 59 },
		{ 194, 21, 59 },
		{ 194, 22, 59 },
		{ 194, 23, 59 },
		{ 194, 24, 59 },
		{ 194, 25, 59 },
		{ 194, 26, 59 },
		{ 194, 27, 59 },
		{ 194, 28, 59 },
		{ 194, 29, 59 },
		{ 194, 30, 59 },
		{ 194, 31, 59 },
		{ 194, 32, 59 },
		{ 194, 33, 59 },
		{ 194, 34, 59 },
		{ 194, 35, 59 },
		{ 194, 36, 59 },
		{ 194, 37, 59 },
		{ 194, 38, 59 },
		{ 194, 39, 59 },
		{ 194, 40, 59 },
		{ 194, 41, 59 },
		{ 194, 42, 59 },
		{ 194, 43, 59 },
		{ 194, 44, 59 },
		{ 194, 45, 59 },
		{ 194, 46, 59 },
		{ 194, 47, 59 },
		{ 194, 48, 59 },
		{ 194, 49, 59 },
		{ 194, 50, 59 },
		{ 194, 51, 59 },
		{ 194, 52, 59 },
		{ 194, 53, 59 },
		{ 194, 54, 59 },
		{ 194, 55, 59 },
		{ 194, 56, 59 },
		{ 194, 57, 59 },
		{ 194, 58, 59 },
		{ 194, 59, 59 },
		{ 194, 60, 59 },
		{ 194, 61, 59 },
		{ 194, 62, 59 },
		{ 194, 63, 59 },
		{ 194, 64, 59 },
		{ 194, 65, 59 },
		{ 194, 66, 59 },
		{ 194, 67, 59 },
		{ 194, 68, 59 },
		{ 194, 69, 59 },
		{ 194, 70, 59 },
		{ 194, 71, 59 },
		{ 194, 72, 59 },
		{ 194, 73, 59 },
		{ 194, 74, 59 },
		{ 194, 75, 59 },
		{ 194, 76, 59 },
		{ 194, 77, 59 },
		{ 194, 78, 59 },
		{ 194, 79, 59 },
		{ 194, 80, 59 },
		{ 194, 81, 59 },
		{ 194, 82, 59 },
		{ 194, 83, 59 },
		{ 194, 84, 59 },
		{ 194, 85, 59 },
		{ 194, 86, 59 },
		{ 194, 87, 59 },
		{ 194, 88, 59 },
		{ 194, 89, 59 },
		{ 194, 90, 59 },
		{ 194, 91, 59 },
		{ 194, 92, 59 },
		{ 194, 93, 59 },
		{ 194, 94, 59 },
		{ 194, 95, 59 },
		{ 194, 96, 59 },
		{ 194, 97, 59 },
		{ 194, 98, 59 },
		{ 194, 99, 59 },
		{ 194, 100, 59 },
		{ 194, 101, 59 },
		{ 194, 102, 59 },
		{ 194, 103, 59 },
		{ 194, 104, 59 },
		{ 194, 105, 59 },
		{ 194, 106, 59 },
		{ 194, 107, 59 },
		{ 194, 108, 59 },
		{ 194, 109, 59 },
		{ 194, 110, 59 },
		{ 194, 111, 59 },
		{ 194, 112, 59 },
		{ 194, 113, 59 },
		{ 194, 114, 59 },
		{ 194, 115, 59 },
		{ 194, 116, 59 },
		{ 194, 117, 59 },
		{ 194, 118, 59 },
		{ 194, 119, 59 },
		{ 194, 120, 59 },
		{ 194, 121, 59 },
		{ 194, 122, 59 },
		{ 194, 123, 59 },
		{ 194, 124, 59 },
		{ 194, 125, 59 },
		{ 194, 126, 59 },
		{ 194, 127, 59 },
		{ 194, 128, 59 },
		{ 194, 129, 59 },
		{ 194, 130, 59 },
		{ 195, 10, 59 },
		{ 195, 11, 59 },
		{ 195, 12, 59 },
		{ 195, 13, 59 },
		{ 195, 14, 59 },
		{ 195, 15, 59 },
		{ 195, 16, 59 },
		{ 195, 17, 59 },
		{ 195, 18, 59 },
		{ 195, 19, 59 },
		{ 195, 20, 59 },
		{ 195, 21, 59 },
		{ 195, 22, 59 },
		{ 195, 23, 59 },
		{ 195, 24, 59 },
		{ 195, 25, 59 },
		{ 195, 26, 59 },
		{ 195, 27, 59 },
		{ 195, 28, 59 },
		{ 195, 29, 59 },
		{ 195, 30, 59 },
		{ 195, 31, 59 },
		{ 195, 32, 59 },
		{ 195, 33, 59 },
		{ 195, 34, 59 },
		{ 195, 35, 59 },
		{ 195, 36, 59 },
		{ 195, 37, 59 },
		{ 195, 38, 59 },
		{ 195, 39, 59 },
		{ 195, 40, 59 },
		{ 195, 41, 59 },
		{ 195, 42, 59 },
		{ 195, 43, 59 },
		{ 195, 44, 59 },
		{ 195, 45, 59 },
		{ 195, 46, 59 },
		{ 195, 47, 59 },
		{ 195, 48, 59 },
		{ 195, 49, 59 },
		{ 195, 50, 59 },
		{ 195, 51, 59 },
		{ 195, 52, 59 },
		{ 195, 53, 59 },
		{ 195, 54, 59 },
		{ 195, 55, 59 },
		{ 195, 56, 59 },
		{ 195, 57, 59 },
		{ 195, 58, 59 },
		{ 195, 59, 59 },
		{ 195, 60, 59 },
		{ 195, 61, 59 },
		{ 195, 62, 59 },
		{ 195, 63, 59 },
		{ 195, 64, 59 },
		{ 195, 65, 59 },
		{ 195, 66, 59 },
		{ 195, 67, 59 },
		{ 195, 68, 59 },
		{ 195, 69, 59 },
		{ 195, 70, 59 },
		{ 195, 71, 59 },
		{ 195, 72, 59 },
		{ 195, 73, 59 },
		{ 195, 74, 59 },
		{ 195, 75, 59 },
		{ 195, 76, 59 },
		{ 195, 77, 59 },
		{ 195, 78, 59 },
		{ 195, 79, 59 },
		{ 195, 80, 59 },
		{ 195, 81, 59 },
		{ 195, 82, 59 },
		{ 195, 83, 59 },
		{ 195, 84, 59 },
		{ 195, 85, 59 },
		{ 195, 86, 59 },
		{ 195, 87, 59 },
		{ 195, 88, 59 },
		{ 195, 89, 59 },
		{ 195, 90, 59 },
		{ 195, 91, 59 },
		{ 195, 92, 59 },
		{ 195, 93, 59 },
		{ 195, 94, 59 },
		{ 195, 95, 59 },
		{ 195, 96, 59 },
		{ 195, 97, 59 },
		{ 195, 98, 59 },
		{ 195, 99, 59 },
		{ 195, 100, 59 },
		{ 195, 101, 59 },
		{ 195, 102, 59 },
		{ 195, 103, 59 },
		{ 195, 104, 59 },
		{ 195, 105, 59 },
		{ 195, 106, 59 },
		{ 195, 107, 59 },
		{ 195, 108, 59 },
		{ 195, 109, 59 },
		{ 195, 110, 59 },
		{ 195, 111, 59 },
		{ 195, 112, 59 },
		{ 195, 113, 59 },
		{ 195, 114, 59 },
		{ 195, 115, 59 },
		{ 195, 116, 59 },
		{ 195, 117, 59 },
		{ 195, 118, 59 },
		{ 195, 119, 59 },
		{ 195, 120, 59 },
		{ 195, 121, 59 },
		{ 195, 122, 59 },
		{ 195, 123, 59 },
		{ 195, 124, 59 },
		{ 195, 125, 59 },
		{ 195, 126, 59 },
		{ 195, 127, 59 },
		{ 195, 128, 59 },
		{ 195, 129, 59 },
		{ 195, 130, 59 },
		{ 195, 131, 59 },
		{ 195, 132, 59 },
		{ 196, 14, 59 },
		{ 196, 15, 59 },
		{ 196, 16, 59 },
		{ 196, 17, 59 },
		{ 196, 18, 59 },
		{ 196, 19, 59 },
		{ 196, 20, 59 },
		{ 196, 21, 59 },
		{ 196, 22, 59 },
		{ 196, 23, 59 },
		{ 196, 24, 59 },
		{ 196, 25, 59 },
		{ 196, 26, 59 },
		{ 196, 27, 59 },
		{ 196, 28, 59 },
		{ 196, 29, 59 },
		{ 196, 30, 59 },
		{ 196, 31, 59 },
		{ 196, 32, 59 },
		{ 196, 33, 59 },
		{ 196, 34, 59 },
		{ 196, 35, 59 },
		{ 196, 36, 59 },
		{ 196, 37, 59 },
		{ 196, 38, 59 },
		{ 196, 39, 59 },
		{ 196, 40, 59 },
		{ 196, 41, 59 },
		{ 196, 42, 59 },
		{ 196, 43, 59 },
		{ 196, 44, 59 },
		{ 196, 45, 59 },
		{ 196, 46, 59 },
		{ 196, 47, 59 },
		{ 196, 48, 59 },
		{ 196, 49, 59 },
		{ 196, 50, 59 },
		{ 196, 51, 59 },
		{ 196, 52, 59 },
		{ 196, 53, 59 },
		{ 196, 54, 59 },
		{ 196, 55, 59 },
		{ 196, 56, 59 },
		{ 196, 57, 59 },
		{ 196, 58, 59 },
		{ 196, 59, 59 },
		{ 196, 60, 59 },
		{ 196, 61, 59 },
		{ 196, 62, 59 },
		{ 196, 63, 59 },
		{ 196, 64, 59 },
		{ 196, 65, 59 },
		{ 196, 66, 59 },
		{ 196, 67, 59 },
		{ 196, 68, 59 },
		{ 196, 69, 59 },
		{ 196, 70, 59 },
		{ 196, 71, 59 },
		{ 196, 72, 59 },
		{ 196, 73, 59 },
		{ 196, 74, 59 },
		{ 196, 75, 59 },
		{ 196, 76, 59 },
		{ 196, 77, 59 },
		{ 196, 78, 59 },
		{ 196, 79, 59 },
		{ 196, 80, 59 },
		{ 196, 81, 59 },
		{ 196, 82, 59 },
		{ 196, 83, 59 },
		{ 196, 84, 59 },
		{ 196, 85, 59 },
		{ 196, 86, 59 },
		{ 196, 87, 59 },
		{ 196, 88, 59 },
		{ 196, 89, 59 },
		{ 196, 90, 59 },
		{ 196, 91, 59 },
		{ 196, 92, 59 },
		{ 196, 93, 59 },
		{ 196, 94, 59 },
		{ 196, 95, 59 },
		{ 196, 96, 59 },
		{ 196, 97, 59 },
		{ 196, 98, 59 },
		{ 196, 99, 59 },
		{ 196, 100, 59 },
		{ 196, 101, 59 },
		{ 196, 102, 59 },
		{ 196, 103, 59 },
		{ 196, 104, 59 },
		{ 196, 105, 59 },
		{ 196, 106, 59 },
		{ 196, 107, 59 },
		{ 196, 108, 59 },
		{ 196, 109, 59 },
		{ 196, 110, 59 },
		{ 196, 111, 59 },
		{ 196, 112, 59 },
		{ 196, 113, 59 },
		{ 196, 114, 59 },
		{ 196, 115, 59 },
		{ 196, 116, 59 },
		{ 196, 117, 59 },
		{ 196, 118, 59 },
		{ 196, 119, 59 },
		{ 196, 120, 59 },
		{ 196, 121, 59 },
		{ 196, 122, 59 },
		{ 196, 123, 59 },
		{ 196, 124, 59 },
		{ 196, 125, 59 },
		{ 196, 126, 59 },
		{ 196, 127, 59 },
		{ 196, 128, 59 },
		{ 196, 129, 59 },
		{ 196, 130, 59 },
		{ 196, 131, 59 },
		{ 196, 132, 59 },
		{ 196, 133, 59 },
		{ 196, 134, 59 },
		{ 197, 18, 59 },
		{ 197, 19, 59 },
		{ 197, 20, 59 },
		{ 197, 21, 59 },
		{ 197, 22, 59 },
		{ 197, 23, 59 },
		{ 197, 24, 59 },
		{ 197, 25, 59 },
		{ 197, 26, 59 },
		{ 197, 27, 59 },
		{ 197, 28, 59 },
		{ 197, 29, 59 },
		{ 197, 30, 59 },
		{ 197, 31, 59 },
		{ 197, 32, 59 },
		{ 197, 33, 59 },
		{ 197, 34, 59 },
		{ 197, 35, 59 },
		{ 197, 36, 59 },
		{ 197, 37, 59 },
		{ 197, 38, 59 },
		{ 197, 39, 59 },
		{ 197, 40, 59 },
		{ 197, 41, 59 },
		{ 197, 42, 59 },
		{ 197, 43, 59 },
		{ 197, 44, 59 },
		{ 197, 45, 59 },
		{ 197, 46, 59 },
		{ 197, 47, 59 },
		{ 197, 48, 59 },
		{ 197, 49, 59 },
		{ 197, 50, 59 },
		{ 197, 51, 59 },
		{ 197, 52, 59 },
		{ 197, 53, 59 },
		{ 197, 54, 59 },
		{ 197, 55, 59 },
		{ 197, 56, 59 },
		{ 197, 57, 59 },
		{ 197, 58, 59 },
		{ 197, 59, 59 },
		{ 197, 60, 59 },
		{ 197, 61, 59 },
		{ 197, 62, 59 },
		{ 197, 63, 59 },
		{ 197, 64, 59 },
		{ 197, 65, 59 },
		{ 197, 66, 59 },
		{ 197, 67, 59 },
		{ 197, 68, 59 },
		{ 197, 69, 59 },
		{ 197, 70, 59 },
		{ 197, 71, 59 },
		{ 197, 72, 59 },
		{ 197, 73, 59 },
		{ 197, 74, 59 },
		{ 197, 75, 59 },
		{ 197, 76, 59 },
		{ 197, 77, 59 },
		{ 197, 78, 59 },
		{ 197, 79, 59 },
		{ 197, 80, 59 },
		{ 197, 81, 59 },
		{ 197, 82, 59 },
		{ 197, 83, 59 },
		{ 197, 84, 59 },
		{ 197, 85, 59 },
		{ 197, 86, 59 },
		{ 197, 87, 59 },
		{ 197, 88, 59 },
		{ 197, 89, 59 },
		{ 197, 90, 59 },
		{ 197, 91, 59 },
		{ 197, 92, 59 },
		{ 197, 93, 59 },
		{ 197, 94, 59 },
		{ 197, 95, 59 },
		{ 197, 96, 59 },
		{ 197, 97, 59 },
		{ 197, 98, 59 },
		{ 197, 99, 59 },
		{ 197, 100, 59 },
		{ 197, 101, 59 },
		{ 197, 102, 59 },
		{ 197, 103, 59 },
		{ 197, 104, 59 },
		{ 197, 105, 59 },
		{ 197, 106, 59 },
		{ 197, 107, 59 },
		{ 197, 108, 59 },
		{ 197, 109, 59 },
		{ 197, 110, 59 },
		{ 197, 111, 59 },
		{ 197, 112, 59 },
		{ 197, 113, 59 },
		{ 197, 114, 59 },
		{ 197, 115, 59 },
		{ 197, 116, 59 },
		{ 197, 117, 59 },
		{ 197, 118, 59 },
		{ 197, 119, 59 },
		{ 197, 120, 59 },
		{ 197, 121, 59 },
		{ 197, 122, 59 },
		{ 197, 123, 59 },
		{ 197, 124, 59 },
		{ 197, 125, 59 },
		{ 197, 126, 59 },
		{ 197, 127, 59 },
		{ 197, 128, 59 },
		{ 197, 129, 59 },
		{ 197, 130, 59 },
		{ 197, 131, 59 },
		{ 197, 132, 59 },
		{ 197, 133, 59 },
		{ 197, 134, 59 },
		{ 197, 135, 59 },
		{ 197, 136, 59 },
		{ 198, 22, 59 },
		{ 198, 23, 59 },
		{ 198, 24, 59 },
		{ 198, 25, 59 },
		{ 198, 26, 59 },
		{ 198, 27, 59 },
		{ 198, 28, 59 },
		{ 198, 29, 59 },
		{ 198, 30, 59 },
		{ 198, 31, 59 },
		{ 198, 32, 59 },
		{ 198, 33, 59 },
		{ 198, 34, 59 },
		{ 198, 35, 59 },
		{ 198, 36, 59 },
		{ 198, 37, 59 },
		{ 198, 38, 59 },
		{ 198, 39, 59 },
		{ 198, 40, 59 },
		{ 198, 41, 59 },
		{ 198, 42, 59 },
		{ 198, 43, 59 },
		{ 198, 44, 59 },
		{ 198, 45, 59 },
		{ 198, 46, 59 },
		{ 198, 47, 59 },
		{ 198, 48, 59 },
		{ 198, 49, 59 },
		{ 198, 50, 59 },
		{ 198, 51, 59 },
		{ 198, 52, 59 },
		{ 198, 53, 59 },
		{ 198, 54, 59 },
		{ 198, 55, 59 },
		{ 198, 56, 59 },
		{ 198, 57, 59 },
		{ 198, 58, 59 },
		{ 198, 59, 59 },
		{ 198, 60, 59 },
		{ 198, 61, 59 },
		{ 198, 62, 59 },
		{ 198, 63, 59 },
		{ 198, 64, 59 },
		{ 198, 65, 59 },
		{ 198, 66, 59 },
		{ 198, 67, 59 },
		{ 198, 68, 59 },
		{ 198, 69, 59 },
		{ 198, 70, 59 },
		{ 198, 71, 59 },
		{ 198, 72, 59 },
		{ 198, 73, 59 },
		{ 198, 74, 59 },
		{ 198, 75, 59 },
		{ 198, 76, 59 },
		{ 198, 77, 59 },
		{ 198, 78, 59 },
		{ 198, 79, 59 },
		{ 198, 80, 59 },
		{ 198, 81, 59 },
		{ 198, 82, 59 },
		{ 198, 83, 59 },
		{ 198, 84, 59 },
		{ 198, 85, 59 },
		{ 198, 86, 59 },
		{ 198, 87, 59 },
		{ 198, 88, 59 },
		{ 198, 89, 59 },
		{ 198, 90, 59 },
		{ 198, 91, 59 },
		{ 198, 92, 59 },
		{ 198, 93, 59 },
		{ 198, 94, 59 },
		{ 198, 95, 59 },
		{ 198, 96, 59 },
		{ 198, 97, 59 },
		{ 198, 98, 59 },
		{ 198, 99, 59 },
		{ 198, 100, 59 },
		{ 198, 101, 59 },
		{ 198, 102, 59 },
		{ 198, 103, 59 },
		{ 198, 104, 59 },
		{ 198, 105, 59 },
		{ 198, 106, 59 },
		{ 198, 107, 59 },
		{ 198, 108, 59 },
		{ 198, 109, 59 },
		{ 198, 110, 59 },
		{ 198, 111, 59 },
		{ 198, 112, 59 },
		{ 198, 113, 59 },
		{ 198, 114, 59 },
		{ 198, 115, 59 },
		{ 198, 116, 59 },
		{ 198, 117, 59 },
		{ 198, 118, 59 },
		{ 198, 119, 59 },
		{ 198, 120, 59 },
		{ 198, 121, 59 },
		{ 198, 122, 59 },
		{ 198, 123, 59 },
		{ 198, 124, 59 },
		{ 198, 125, 59 },
		{ 198, 126, 59 },
		{ 198, 127, 59 },
		{ 198, 128, 59 },
		{ 198, 129, 59 },
		{ 198, 130, 59 },
		{ 198, 131, 59 },
		{ 198, 132, 59 },
		{ 198, 133, 59 },
		{ 198, 134, 59 },
		{ 198, 135, 59 },
		{ 198, 136, 59 },
		{ 198, 137, 59 },
		{ 198, 138, 59 },
		{ 199, 26, 59 },
		{ 199, 27, 59 },
		{ 199, 28, 59 },
		{ 199, 29, 59 },
		{ 199, 30, 59 },
		{ 199, 31, 59 },
		{ 199, 32, 59 },
		{ 199, 33, 59 },
		{ 199, 34, 59 },
		{ 199, 35, 59 },
		{ 199, 36, 59 },
		{ 199, 37, 59 },
		{ 199, 38, 59 },
		{ 199, 39, 59 },
		{ 199, 40, 59 },
		{ 199, 41, 59 },
		{ 199, 42, 59 },
		{ 199, 43, 59 },
		{ 199, 44, 59 },
		{ 199, 45, 59 },
		{ 199, 46, 59 },
		{ 199, 47, 59 },
		{ 199, 48, 59 },
		{ 199, 49, 59 },
		{ 199, 50, 59 },
		{ 199, 51, 59 },
		{ 199, 52, 59 },
		{ 199, 53, 59 },
		{ 199, 54, 59 },
		{ 199, 55, 59 },
		{ 199, 56, 59 },
		{ 199, 57, 59 },
		{ 199, 58, 59 },
		{ 199, 59, 59 },
		{ 199, 60, 59 },
		{ 199, 61, 59 },
		{ 199, 62, 59 },
		{ 199, 63, 59 },
		{ 199, 64, 59 },
		{ 199, 65, 59 },
		{ 199, 66, 59 },
		{ 199, 67, 59 },
		{ 199, 68, 59 },
		{ 199, 69, 59 },
		{ 199, 70, 59 },
		{ 199, 71, 59 },
		{ 199, 72, 59 },
		{ 199, 73, 59 },
		{ 199, 74, 59 },
		{ 199, 75, 59 },
		{ 199, 76, 59 },
		{ 199, 77, 59 },
		{ 199, 78, 59 },
		{ 199, 79, 59 },
		{ 199, 80, 59 },
		{ 199, 81, 59 },
		{ 199, 82, 59 },
		{ 199, 83, 59 },
		{ 199, 84, 59 },
		{ 199, 85, 59 },
		{ 199, 86, 59 },
		{ 199, 87, 59 },
		{ 199, 88, 59 },
		{ 199, 89, 59 },
		{ 199, 90, 59 },
		{ 199, 91, 59 },
		{ 199, 92, 59 },
		{ 199, 93, 59 },
		{ 199, 94, 59 },
		{ 199, 95, 59 },
		{ 199, 96, 59 },
		{ 199, 97, 59 },
		{ 199, 98, 59 },
		{ 199, 99, 59 },
		{ 199, 100, 59 },
		{ 199, 101, 59 },
		{ 199, 102, 59 },
		{ 199, 103, 59 },
		{ 199, 104, 59 },
		{ 199, 105, 59 },
		{ 199, 106, 59 },
		{ 199, 107, 59 },
		{ 199, 108, 59 },
		{ 199, 109, 59 },
		{ 199, 110, 59 },
		{ 199, 111, 59 },
		{ 199, 112, 59 },
		{ 199, 113, 59 },
		{ 199, 114, 59 },
		{ 199, 115, 59 },
		{ 199, 116, 59 },
		{ 199, 117, 59 },
		{ 199, 118, 59 },
		{ 199, 119, 59 },
		{ 199, 120, 59 },
		{ 199, 121, 59 },
		{ 199, 122, 59 },
		{ 199, 123, 59 },
		{ 199, 124, 59 },
		{ 199, 125, 59 },
		{ 199, 126, 59 },
		{ 199, 127, 59 },
		{ 199, 128, 59 },
		{ 199, 129, 59 },
		{ 199, 130, 59 },
		{ 199, 131, 59 },
		{ 199, 132, 59 },
		{ 199, 133, 59 },
		{ 199, 134, 59 },
		{ 199, 135, 59 },
		{ 199, 136, 59 },
		{ 199, 137, 59 },
		{ 199, 138, 59 },
		{ 199, 139, 59 },
		{ 199, 140, 59 },
		{ 200, 30, 59 },
		{ 200, 31, 59 },
		{ 200, 32, 59 },
		{ 200, 33, 59 },
		{ 200, 34, 59 },
		{ 200, 35, 59 },
		{ 200, 36, 59 },
		{ 200, 37, 59 },
		{ 200, 38, 59 },
		{ 200, 39, 59 },
		{ 200, 40, 59 },
		{ 200, 41, 59 },
		{ 200, 42, 59 },
		{ 200, 43, 59 },
		{ 200, 44, 59 },
		{ 200, 45, 59 },
		{ 200, 46, 59 },
		{ 200, 47, 59 },
		{ 200, 48, 59 },
		{ 200, 49, 59 },
		{ 200, 50, 59 },
		{ 200, 51, 59 },
		{ 200, 52, 59 },
		{ 200, 53, 59 },
		{ 200, 54, 59 },
		{ 200, 55, 59 },
		{ 200, 56, 59 },
		{ 200, 57, 59 },
		{ 200, 58, 59 },
		{ 200, 59, 59 },
		{ 200, 60, 59 },
		{ 200, 61, 59 },
		{ 200, 62, 59 },
		{ 200, 63, 59 },
		{ 200, 64, 59 },
		{ 200, 65, 59 },
		{ 200, 66, 59 },
		{ 200, 67, 59 },
		{ 200, 68, 59 },
		{ 200, 69, 59 },
		{ 200, 70, 59 },
		{ 200, 71, 59 },
		{ 200, 72, 59 },
		{ 200, 73, 59 },
		{ 200, 74, 59 },
		{ 200, 75, 59 },
		{ 200, 76, 59 },
		{ 200, 77, 59 },
		{ 200, 78, 59 },
		{ 200, 79, 59 },
		{ 200, 80, 59 },
		{ 200, 81, 59 },
		{ 200, 82, 59 },
		{ 200, 83, 59 },
		{ 200, 84, 59 },
		{ 200, 85, 59 },
		{ 200, 86, 59 },
		{ 200, 87, 59 },
		{ 200, 88, 59 },
		{ 200, 89, 59 },
		{ 200, 90, 59 },
		{ 200, 91, 59 },
		{ 200, 92, 59 },
		{ 200, 93, 59 },
		{ 200, 94, 59 },
		{ 200, 95, 59 },
		{ 200, 96, 59 },
		{ 200, 97, 59 },
		{ 200, 98, 59 },
		{ 200, 99, 59 },
		{ 200, 100, 59 },
		{ 200, 101, 59 },
		{ 200, 102, 59 },
		{ 200, 103, 59 },
		{ 200, 104, 59 },
		{ 200, 105, 59 },
		{ 200, 106, 59 },
		{ 200, 107, 59 },
		{ 200, 108, 59 },
		{ 200, 109, 59 },
		{ 200, 110, 59 },
		{ 200, 111, 59 },
		{ 200, 112, 59 },
		{ 200, 113, 59 },
		{ 200, 114, 59 },
		{ 200, 115, 59 },
		{ 200, 116, 59 },
		{ 200, 117, 59 },
		{ 200, 118, 59 },
		{ 200, 119, 59 },
		{ 200, 120, 59 },
		{ 200, 121, 59 },
		{ 200, 122, 59 },
		{ 200, 123, 59 },
		{ 200, 124, 59 },
		{ 200, 125, 59 },
		{ 200, 126, 59 },
		{ 200, 127, 59 },
		{ 200, 128, 59 },
		{ 200, 129, 59 },
		{ 200, 130, 59 },
		{ 200, 131, 59 },
		{ 200, 132, 59 },
		{ 200, 133, 59 },
		{ 200, 134, 59 },
		{ 200, 135, 59 },
		{ 200, 136, 59 },
		{ 200, 137, 59 },
		{ 200, 138, 59 },
		{ 200, 139, 59 },
		{ 200, 140, 59 },
		{ 200, 141, 59 },
		{ 200, 142, 59 },
		{ 201, 34, 59 },
		{ 201, 35, 59 },
		{ 201, 36, 59 },
		{ 201, 37, 59 },
		{ 201, 38, 59 },
		{ 201, 39, 59 },
		{ 201, 40, 59 },
		{ 201, 41, 59 },
		{ 201, 42, 59 },
		{ 201, 43, 59 },
		{ 201, 44, 59 },
		{ 201, 45, 59 },
		{ 201, 46, 59 },
		{ 201, 47, 59 },
		{ 201, 48, 59 },
		{ 201, 49, 59 },
		{ 201, 50, 59 },
		{ 201, 51, 59 },
		{ 201, 52, 59 },
		{ 201, 53, 59 },
		{ 201, 54, 59 },
		{ 201, 55, 59 },
		{ 201, 56, 59 },
		{ 201, 57, 59 },
		{ 201, 58, 59 },
		{ 201, 59, 59 },
		{ 201, 60, 59 },
		{ 201, 61, 59 },
		{ 201, 62, 59 },
		{ 201, 63, 59 },
		{ 201, 64, 59 },
		{ 201, 65, 59 },
		{ 201, 66, 59 },
		{ 201, 67, 59 },
		{ 201, 68, 59 },
		{ 201, 69, 59 },
		{ 201, 70, 59 },
		{ 201, 71, 59 },
		{ 201, 72, 59 },
		{ 201, 73, 59 },
		{ 201, 74, 59 },
		{ 201, 75, 59 },
		{ 201, 76, 59 },
		{ 201, 77, 59 },
		{ 201, 78, 59 },
		{ 201, 79, 59 },
		{ 201, 80, 59 },
		{ 201, 81, 59 },
		{ 201, 82, 59 },
		{ 201, 83, 59 },
		{ 201, 84, 59 },
		{ 201, 85, 59 },
		{ 201, 86, 59 },
		{ 201, 87, 59 },
		{ 201, 88, 59 },
		{ 201, 89, 59 },
		{ 201, 90, 59 },
		{ 201, 91, 59 },
		{ 201, 92, 59 },
		{ 201, 93, 59 },
		{ 201, 94, 59 },
		{ 201, 95, 59 },
		{ 201, 96, 59 },
		{ 201, 97, 59 },
		{ 201, 98, 59 },
		{ 201, 99, 59 },
		{ 201, 100, 59 },
		{ 201, 101, 59 },
		{ 201, 102, 59 },
		{ 201, 103, 59 },
		{ 201, 104, 59 },
		{ 201, 105, 59 },
		{ 201, 106, 59 },
		{ 201, 107, 59 },
		{ 201, 108, 59 },
		{ 201, 109, 59 },
		{ 201, 110, 59 },
		{ 201, 111, 59 },
		{ 201, 112, 59 },
		{ 201, 113, 59 },
		{ 201, 114, 59 },
		{ 201, 115, 59 },
		{ 201, 116, 59 },
		{ 201, 117, 59 },
		{ 201, 118, 59 },
		{ 201, 119, 59 },
		{ 201, 120, 59 },
		{ 201, 121, 59 },
		{ 201, 122, 59 },
		{ 201, 123, 59 },
		{ 201, 124, 59 },
		{ 201, 125, 59 },
		{ 201, 126, 59 },
		{ 201, 127, 59 },
		{ 201, 128, 59 },
		{ 201, 129, 59 },
		{ 201, 130, 59 },
		{ 201, 131, 59 },
		{ 201, 132, 59 },
		{ 201, 133, 59 },
		{ 201, 134, 59 },
		{ 201, 135, 59 },
		{ 201, 136, 59 },
		{ 201, 137, 59 },
		{ 201, 138, 59 },
		{ 201, 139, 59 },
		{ 201, 140, 59 },
		{ 201, 141, 59 },
		{ 201, 142, 59 },
		{ 201, 143, 59 },
		{ 201, 144, 59 },
		{ 202, 38, 59 },
		{ 202, 39, 59 },
		{ 202, 40, 59 },
		{ 202, 41, 59 },
		{ 202, 42, 59 },
		{ 202, 43, 59 },
		{ 202, 44, 59 },
		{ 202, 45, 59 },
		{ 202, 46, 59 },
		{ 202, 47, 59 },
		{ 202, 48, 59 },
		{ 202, 49, 59 },
		{ 202, 50, 59 },
		{ 202, 51, 59 },
		{ 202, 52, 59 },
		{ 202, 53, 59 },
		{ 202, 54, 59 },
		{ 202, 55, 59 },
		{ 202, 56, 59 },
		{ 202, 57, 59 },
		{ 202, 58, 59 },
		{ 202, 59, 59 },
		{ 202, 60, 59 },
		{ 202, 61, 59 },
		{ 202, 62, 59 },
		{ 202, 63, 59 },
		{ 202, 64, 59 },
		{ 202, 65, 59 },
		{ 202, 66, 59 },
		{ 202, 67, 59 },
		{ 202, 68, 59 },
		{ 202, 69, 59 },
		{ 202, 70, 59 },
		{ 202, 71, 59 },
		{ 202, 72, 59 },
		{ 202, 73, 59 },
		{ 202, 74, 59 },
		{ 202, 75, 59 },
		{ 202, 76, 59 },
		{ 202, 77, 59 },
		{ 202, 78, 59 },
		{ 202, 79, 59 },
		{ 202, 80, 59 },
		{ 202, 81, 59 },
		{ 202, 82, 59 },
		{ 202, 83, 59 },
		{ 202, 84, 59 },
		{ 202, 85, 59 },
		{ 202, 86, 59 },
		{ 202, 87, 59 },
		{ 202, 88, 59 },
		{ 202, 89, 59 },
		{ 202, 90, 59 },
		{ 202, 91, 59 },
		{ 202, 92, 59 },
		{ 202, 93, 59 },
		{ 202, 94, 59 },
		{ 202, 95, 59 },
		{ 202, 96, 59 },
		{ 202, 97, 59 },
		{ 202, 98, 59 },
		{ 202, 99, 59 },
		{ 202, 100, 59 },
		{ 202, 101, 59 },
		{ 202, 102, 59 },
		{ 202, 103, 59 },
		{ 202, 104, 59 },
		{ 202, 105, 59 },
		{ 202, 106, 59 },
		{ 202, 107, 59 },
		{ 202, 108, 59 },
		{ 202, 109, 59 },
		{ 202, 110, 59 },
		{ 202, 111, 59 },
		{ 202, 112, 59 },
		{ 202, 113, 59 },
		{ 202, 114, 59 },
		{ 202, 115, 59 },
		{ 202, 116, 59 },
		{ 202, 117, 59 },
		{ 202, 118, 59 },
		{ 202, 119, 59 },
		{ 202, 120, 59 },
		{ 202, 121, 59 },
		{ 202, 122, 59 },
		{ 202, 123, 59 },
		{ 202, 124, 59 },
		{ 202, 125, 59 },
		{ 202, 126, 59 },
		{ 202, 127, 59 },
		{ 202, 128, 59 },
		{ 202, 129, 59 },
		{ 202, 130, 59 },
		{ 202, 131, 59 },
		{ 202, 132, 59 },
		{ 202, 133, 59 },
		{ 202, 134, 59 },
		{ 202, 135, 59 },
		{ 202, 136, 59 },
		{ 202, 137, 59 },
		{ 202, 138, 59 },
		{ 202, 139, 59 },
		{ 202, 140, 59 },
		{ 202, 141, 59 },
		{ 202, 142, 59 },
		{ 202, 143, 59 },
		{ 202, 144, 59 },
		{ 202, 145, 59 },
		{ 202, 146, 59 },
		{ 203, 42, 59 },
		{ 203, 43, 59 },
		{ 203, 44, 59 },
		{ 203, 45, 59 },
		{ 203, 46, 59 },
		{ 203, 47, 59 },
		{ 203, 48, 59 },
		{ 203, 49, 59 },
		{ 203, 50, 59 },
		{ 203, 51, 59 },
		{ 203, 52, 59 },
		{ 203, 53, 59 },
		{ 203, 54, 59 },
		{ 203, 55, 59 },
		{ 203, 56, 59 },
		{ 203, 57, 59 },
		{ 203, 58, 59 },
		{ 203, 59, 59 },
		{ 203, 60, 59 },
		{ 203, 61, 59 },
		{ 203, 62, 59 },
		{ 203, 63, 59 },
		{ 203, 64, 59 },
		{ 203, 65, 59 },
		{ 203, 66, 59 },
		{ 203, 67, 59 },
		{ 203, 68, 59 },
		{ 203, 69, 59 },
		{ 203, 70, 59 },
		{ 203, 71, 59 },
		{ 203, 72, 59 },
		{ 203, 73, 59 },
		{ 203, 74, 59 },
		{ 203, 75, 59 },
		{ 203, 76, 59 },
		{ 203, 77, 59 },
		{ 203, 78, 59 },
		{ 203, 79, 59 },
		{ 203, 80, 59 },
		{ 203, 81, 59 },
		{ 203, 82, 59 },
		{ 203, 83, 59 },
		{ 203, 84, 59 },
		{ 203, 85, 59 },
		{ 203, 86, 59 },
		{ 203, 87, 59 },
		{ 203, 88, 59 },
		{ 203, 89, 59 },
		{ 203, 90, 59 },
		{ 203, 91, 59 },
		{ 203, 92, 59 },
		{ 203, 93, 59 },
		{ 203, 94, 59 },
		{ 203, 95, 59 },
		{ 203, 96, 59 },
		{ 203, 97, 59 },
		{ 203, 98, 59 },
		{ 203, 99, 59 },
		{ 203, 100, 59 },
		{ 203, 101, 59 },
		{ 203, 102, 59 },
		{ 203, 103, 59 },
		{ 203, 104, 59 },
		{ 203, 105, 59 },
		{ 203, 106, 59 },
		{ 203, 107, 59 },
		{ 203, 108, 59 },
		{ 203, 109, 59 },
		{ 203, 110, 59 },
		{ 203, 111, 59 },
		{ 203, 112, 59 },
		{ 203, 113, 59 },
		{ 203, 114, 59 },
		{ 203, 115, 59 },
		{ 203, 116, 59 },
		{ 203, 117, 59 },
		{ 203, 118, 59 },
		{ 203, 119, 59 },
		{ 203, 120, 59 },
		{ 203, 121, 59 },
		{ 203, 122, 59 },
		{ 203, 123, 59 },
		{ 203, 124, 59 },
		{ 203, 125, 59 },
		{ 203, 126, 59 },
		{ 203, 127, 59 },
		{ 203, 128, 59 },
		{ 203, 129, 59 },
		{ 203, 130, 59 },
		{ 203, 131, 59 },
		{ 203, 132, 59 },
		{ 203, 133, 59 },
		{ 203, 134, 59 },
		{ 203, 135, 59 },
		{ 203, 136, 59 },
		{ 203, 137, 59 },
		{ 203, 138, 59 },
		{ 203, 139, 59 },
		{ 203, 140, 59 },
		{ 203, 141, 59 },
		{ 203, 142, 59 },
		{ 203, 143, 59 },
		{ 203, 144, 59 },
		{ 203, 145, 59 },
		{ 203, 146, 59 },
		{ 203, 147, 59 },
		{ 203, 148, 59 },
		{ 204, 46, 59 },
		{ 204, 47, 59 },
		{ 204, 48, 59 },
		{ 204, 49, 59 },
		{ 204, 50, 59 },
		{ 204, 51, 59 },
		{ 204, 52, 59 },
		{ 204, 53, 59 },
		{ 204, 54, 59 },
		{ 204, 55, 59 },
		{ 204, 56, 59 },
		{ 204, 57, 59 },
		{ 204, 58, 59 },
		{ 204, 59, 59 },
		{ 204, 60, 59 },
		{ 204, 61, 59 },
		{ 204, 62, 59 },
		{ 204, 63, 59 },
		{ 204, 64, 59 },
		{ 204, 65, 59 },
		{ 204, 66, 59 },
		{ 204, 67, 59 },
		{ 204, 68, 59 },
		{ 204, 69, 59 },
		{ 204, 70, 59 },
		{ 204, 71, 59 },
		{ 204, 72, 59 },
		{ 204, 73, 59 },
		{ 204, 74, 59 },
		{ 204, 75, 59 },
		{ 204, 76, 59 },
		{ 204, 77, 59 },
		{ 204, 78, 59 },
		{ 204, 79, 59 },
		{ 204, 80, 59 },
		{ 204, 81, 59 },
		{ 204, 82, 59 },
		{ 204, 83, 59 },
		{ 204, 84, 59 },
		{ 204, 85, 59 },
		{ 204, 86, 59 },
		{ 204, 87, 59 },
		{ 204, 88, 59 },
		{ 204, 89, 59 },
		{ 204, 90, 59 },
		{ 204, 91, 59 },
		{ 204, 92, 59 },
		{ 204, 93, 59 },
		{ 204, 94, 59 },
		{ 204, 95, 59 },
		{ 204, 96, 59 },
		{ 204, 97, 59 },
		{ 204, 98, 59 },
		{ 204, 99, 59 },
		{ 204, 100, 59 },
		{ 204, 101, 59 },
		{ 204, 102, 59 },
		{ 204, 103, 59 },
		{ 204, 104, 59 },
		{ 204, 105, 59 },
		{ 204, 106, 59 },
		{ 204, 107, 59 },
		{ 204, 108, 59 },
		{ 204, 109, 59 },
		{ 204, 110, 59 },
		{ 204, 111, 59 },
		{ 204, 112, 59 },
		{ 204, 113, 59 },
		{ 204, 114, 59 },
		{ 204, 115, 59 },
		{ 204, 116, 59 },
		{ 204, 117, 59 },
		{ 204, 118, 59 },
		{ 204, 119, 59 },
		{ 204, 120, 59 },
		{ 204, 121, 59 },
		{ 204, 122, 59 },
		{ 204, 123, 59 },
		{ 204, 124, 59 },
		{ 204, 125, 59 },
		{ 204, 126, 59 },
		{ 204, 127, 59 },
		{ 204, 128, 59 },
		{ 204, 129, 59 },
		{ 204, 130, 59 },
		{ 204, 131, 59 },
		{ 204, 132, 59 },
		{ 204, 133, 59 },
		{ 204, 134, 59 },
		{ 204, 135, 59 },
		{ 204, 136, 59 },
		{ 204, 137, 59 },
		{ 204, 138, 59 },
		{ 204, 139, 59 },
		{ 204, 140, 59 },
		{ 204, 141, 59 },
		{ 204, 142, 59 },
		{ 204, 143, 59 },
		{ 204, 144, 59 },
		{ 204, 145, 59 },
		{ 204, 146, 59 },
		{ 204, 147, 59 },
		{ 204, 148, 59 },
		{ 204, 149, 59 },
		{ 204, 150, 59 },
		{ 205, 50, 59 },
		{ 205, 51, 59 },
		{ 205, 52, 59 },
		{ 205, 53, 59 },
		{ 205, 54, 59 },
		{ 205, 55, 59 },
		{ 205, 56, 59 },
		{ 205, 57, 59 },
		{ 205, 58, 59 },
		{ 205, 59, 59 },
		{ 205, 60, 59 },
		{ 205, 61, 59 },
		{ 205, 62, 59 },
		{ 205, 63, 59 },
		{ 205, 64, 59 },
		{ 205, 65, 59 },
		{ 205, 66, 59 },
		{ 205, 67, 59 },
		{ 205, 68, 59 },
		{ 205, 69, 59 },
		{ 205, 70, 59 },
		{ 205, 71, 59 },
		{ 205, 72, 59 },
		{ 205, 73, 59 },
		{ 205, 74, 59 },
		{ 205, 75, 59 },
		{ 205, 76, 59 },
		{ 205, 77, 59 },
		{ 205, 78, 59 },
		{ 205, 79, 59 },
		{ 205, 80, 59 },
		{ 205, 81, 59 },
		{ 205, 82, 59 },
		{ 205, 83, 59 },
		{ 205, 84, 59 },
		{ 205, 85, 59 },
		{ 205, 86, 59 },
		{ 205, 87, 59 },
		{ 205, 88, 59 },
		{ 205, 89, 59 },
		{ 205, 90, 59 },
		{ 205, 91, 59 },
		{ 205, 92, 59 },
		{ 205, 93, 59 },
		{ 205, 94, 59 },
		{ 205, 95, 59 },
		{ 205, 96, 59 },
		{ 205, 97, 59 },
		{ 205, 98, 59 },
		{ 205, 99, 59 },
		{ 205, 100, 59 },
		{ 205, 101, 59 },
		{ 205, 102, 59 },
		{ 205, 103, 59 },
		{ 205, 104, 59 },
		{ 205, 105, 59 },
		{ 205, 106, 59 },
		{ 205, 107, 59 },
		{ 205, 108, 59 },
		{ 205, 109, 59 },
		{ 205, 110, 59 },
		{ 205, 111, 59 },
		{ 205, 112, 59 },
		{ 205, 113, 59 },
		{ 205, 114, 59 },
		{ 205, 115, 59 },
		{ 205, 116, 59 },
		{ 205, 117, 59 },
		{ 205, 118, 59 },
		{ 205, 119, 59 },
		{ 205, 120, 59 },
		{ 205, 121, 59 },
		{ 205, 122, 59 },
		{ 205, 123, 59 },
		{ 205, 124, 59 },
		{ 205, 125, 59 },
		{ 205, 126, 59 },
		{ 205, 127, 59 },
		{ 205, 128, 59 },
		{ 205, 129, 59 },
		{ 205, 130, 59 },
		{ 205, 131, 59 },
		{ 205, 132, 59 },
		{ 205, 133, 59 },
		{ 205, 134, 59 },
		{ 205, 135, 59 },
		{ 205, 136, 59 },
		{ 205, 137, 59 },
		{ 205, 138, 59 },
		{ 205, 139, 59 },
		{ 205, 140, 59 },
		{ 205, 141, 59 },
		{ 205, 142, 59 },
		{ 205, 143, 59 },
		{ 205, 144, 59 },
		{ 205, 145, 59 },
		{ 205, 146, 59 },
		{ 205, 147, 59 },
		{ 205, 148, 59 },
		{ 205, 149, 59 },
		{ 205, 150, 59 },
		{ 205, 151, 59 },
		{ 205, 152, 59 },
		{ 206, 54, 59 },
		{ 206, 55, 59 },
		{ 206, 56, 59 },
		{ 206, 57, 59 },
		{ 206, 58, 59 },
		{ 206, 59, 59 },
		{ 206, 60, 59 },
		{ 206, 61, 59 },
		{ 206, 62, 59 },
		{ 206, 63, 59 },
		{ 206, 64, 59 },
		{ 206, 65, 59 },
		{ 206, 66, 59 },
		{ 206, 67, 59 },
		{ 206, 68, 59 },
		{ 206, 69, 59 },
		{ 206, 70, 59 },
		{ 206, 71, 59 },
		{ 206, 72, 59 },
		{ 206, 73, 59 },
		{ 206, 74, 59 },
		{ 206, 75, 59 },
		{ 206, 76, 59 },
		{ 206, 77, 59 },
		{ 206, 78, 59 },
		{ 206, 79, 59 },
		{ 206, 80, 59 },
		{ 206, 81, 59 },
		{ 206, 82, 59 },
		{ 206, 83, 59 },
		{ 206, 84, 59 },
		{ 206, 85, 59 },
		{ 206, 86, 59 },
		{ 206, 87, 59 },
		{ 206, 88, 59 },
		{ 206, 89, 59 },
		{ 206, 90, 59 },
		{ 206, 91, 59 },
		{ 206, 92, 59 },
		{ 206, 93, 59 },
		{ 206, 94, 59 },
		{ 206, 95, 59 },
		{ 206, 96, 59 },
		{ 206, 97, 59 },
		{ 206, 98, 59 },
		{ 206, 99, 59 },
		{ 206, 100, 59 },
		{ 206, 101, 59 },
		{ 206, 102, 59 },
		{ 206, 103, 59 },
		{ 206, 104, 59 },
		{ 206, 105, 59 },
		{ 206, 106, 59 },
		{ 206, 107, 59 },
		{ 206, 108, 59 },
		{ 206, 109, 59 },
		{ 206, 110, 59 },
		{ 206, 111, 59 },
		{ 206, 112, 59 },
		{ 206, 113, 59 },
		{ 206, 114, 59 },
		{ 206, 115, 59 },
		{ 206, 116, 59 },
		{ 206, 117, 59 },
		{ 206, 118, 59 },
		{ 206, 119, 59 },
		{ 206, 120, 59 },
		{ 206, 121, 59 },
		{ 206, 122, 59 },
		{ 206, 123, 59 },
		{ 206, 124, 59 },
		{ 206, 125, 59 },
		{ 206, 126, 59 },
		{ 206, 127, 59 },
		{ 206, 128, 59 },
		{ 206, 129, 59 },
		{ 206, 130, 59 },
		{ 206, 131, 59 },
		{ 206, 132, 59 },
		{ 206, 133, 59 },
		{ 206, 134, 59 },
		{ 206, 135, 59 },
		{ 206, 136, 59 },
		{ 206, 137, 59 },
		{ 206, 138, 59 },
		{ 206, 139, 59 },
		{ 206, 140, 59 },
		{ 206, 141, 59 },
		{ 206, 142, 59 },
		{ 206, 143, 59 },
		{ 206, 144, 59 },
		{ 206, 145, 59 },
		{ 206, 146, 59 },
		{ 206, 147, 59 },
		{ 206, 148, 59 },
		{ 206, 149, 59 },
		{ 206, 150, 59 },
		{ 206, 151, 59 },
		{ 206, 152, 59 },
		{ 206, 153, 59 },
		{ 206, 154, 59 },
		{ 207, 58, 59 },
		{ 207, 59, 59 },
		{ 207, 60, 59 },
		{ 207, 61, 59 },
		{ 207, 62, 59 },
		{ 207, 63, 59 },
		{ 207, 64, 59 },
		{ 207, 65, 59 },
		{ 207, 66, 59 },
		{ 207, 67, 59 },
		{ 207, 68, 59 },
		{ 207, 69, 59 },
		{ 207, 70, 59 },
		{ 207, 71, 59 },
		{ 207, 72, 59 },
		{ 207, 73, 59 },
		{ 207, 74, 59 },
		{ 207, 75, 59 },
		{ 207, 76, 59 },
		{ 207, 77, 59 },
		{ 207, 78, 59 },
		{ 207, 79, 59 },
		{ 207, 80, 59 },
		{ 207, 81, 59 },
		{ 207, 82, 59 },
		{ 207, 83, 59 },
		{ 207, 84, 59 },
		{ 207, 85, 59 },
		{ 207, 86, 59 },
		{ 207, 87, 59 },
		{ 207, 88, 59 },
		{ 207, 89, 59 },
		{ 207, 90, 59 },
		{ 207, 91, 59 },
		{ 207, 92, 59 },
		{ 207, 93, 59 },
		{ 207, 94, 59 },
		{ 207, 95, 59 },
		{ 207, 96, 59 },
		{ 207, 97, 59 },
		{ 207, 98, 59 },
		{ 207, 99, 59 },
		{ 207, 100, 59 },
		{ 207, 101, 59 },
		{ 207, 102, 59 },
		{ 207, 103, 59 },
		{ 207, 104, 59 },
		{ 207, 105, 59 },
		{ 207, 106, 59 },
		{ 207, 107, 59 },
		{ 207, 108, 59 },
		{ 207, 109, 59 },
		{ 207, 110, 59 },
		{ 207, 111, 59 },
		{ 207, 112, 59 },
		{ 207, 113, 59 },
		{ 207, 114, 59 },
		{ 207, 115, 59 },
		{ 207, 116, 59 },
		{ 207, 117, 59 },
		{ 207, 118, 59 },
		{ 207, 119, 59 },
		{ 207, 120, 59 },
		{ 207, 121, 59 },
		{ 207, 122, 59 },
		{ 207, 123, 59 },
		{ 207, 124, 59 },
		{ 207, 125, 59 },
		{ 207, 126, 59 },
		{ 207, 127, 59 },
		{ 207, 128, 59 },
		{ 207, 129, 59 },
		{ 207, 130, 59 },
		{ 207, 131, 59 },
		{ 207, 132, 59 },
		{ 207, 133, 59 },
		{ 207, 134, 59 },
		{ 207, 135, 59 },
		{ 207, 136, 59 },
		{ 207, 137, 59 },
		{ 207, 138, 59 },
		{ 207, 139, 59 },
		{ 207, 140, 59 },
		{ 207, 141, 59 },
		{ 207, 142, 59 },
		{ 207, 143, 59 },
		{ 207, 144, 59 },
		{ 207, 145, 59 },
		{ 207, 146, 59 },
		{ 207, 147, 59 },
		{ 207, 148, 59 },
		{ 207, 149, 59 },
		{ 207, 150, 59 },
		{ 207, 151, 59 },
		{ 207, 152, 59 },
		{ 207, 153, 59 },
		{ 207, 154, 59 },
		{ 207, 155, 59 },
		{ 207, 156, 59 },
		{ 208, 62, 59 },
		{ 208, 63, 59 },
		{ 208, 64, 59 },
		{ 208, 65, 59 },
		{ 208, 66, 59 },
		{ 208, 67, 59 },
		{ 208, 68, 59 },
		{ 208, 69, 59 },
		{ 208, 70, 59 },
		{ 208, 71, 59 },
		{ 208, 72, 59 },
		{ 208, 73, 59 },
		{ 208, 74, 59 },
		{ 208, 75, 59 },
		{ 208, 76, 59 },
		{ 208, 77, 59 },
		{ 208, 78, 59 },
		{ 208, 79, 59 },
		{ 208, 80, 59 },
		{ 208, 81, 59 },
		{ 208, 82, 59 },
		{ 208, 83, 59 },
		{ 208, 84, 59 },
		{ 208, 85, 59 },
		{ 208, 86, 59 },
		{ 208, 87, 59 },
		{ 208, 88, 59 },
		{ 208, 89, 59 },
		{ 208, 90, 59 },
		{ 208, 91, 59 },
		{ 208, 92, 59 },
		{ 208, 93, 59 },
		{ 208, 94, 59 },
		{ 208, 95, 59 },
		{ 208, 96, 59 },
		{ 208, 97, 59 },
		{ 208, 98, 59 },
		{ 208, 99, 59 },
		{ 208, 100, 59 },
		{ 208, 101, 59 },
		{ 208, 102, 59 },
		{ 208, 103, 59 },
		{ 208, 104, 59 },
		{ 208, 105, 59 },
		{ 208, 106, 59 },
		{ 208, 107, 59 },
		{ 208, 108, 59 },
		{ 208, 109, 59 },
		{ 208, 110, 59 },
		{ 208, 111, 59 },
		{ 208, 112, 59 },
		{ 208, 113, 59 },
		{ 208, 114, 59 },
		{ 208, 115, 59 },
		{ 208, 116, 59 },
		{ 208, 117, 59 },
		{ 208, 118, 59 },
		{ 208, 119, 59 },
		{ 208, 120, 59 },
		{ 208, 121, 59 },
		{ 208, 122, 59 },
		{ 208, 123, 59 },
		{ 208, 124, 59 },
		{ 208, 125, 59 },
		{ 208, 126, 59 },
		{ 208, 127, 59 },
		{ 208, 128, 59 },
		{ 208, 129, 59 },
		{ 208, 130, 59 },
		{ 208, 131, 59 },
		{ 208, 132, 59 },
		{ 208, 133, 59 },
		{ 208, 134, 59 },
		{ 208, 135, 59 },
		{ 208, 136, 59 },
		{ 208, 137, 59 },
		{ 208, 138, 59 },
		{ 208, 139, 59 },
		{ 208, 140, 59 },
		{ 208, 141, 59 },
		{ 208, 142, 59 },
		{ 208, 143, 59 },
		{ 208, 144, 59 },
		{ 208, 145, 59 },
		{ 208, 146, 59 },
		{ 208, 147, 59 },
		{ 208, 148, 59 },
		{ 208, 149, 59 },
		{ 208, 150, 59 },
		{ 208, 151, 59 },
		{ 208, 152, 59 },
		{ 208, 153, 59 },
		{ 208, 154, 59 },
		{ 208, 155, 59 },
		{ 208, 156, 59 },
		{ 208, 157, 59 },
		{ 208, 158, 59 },
		{ 209, 66, 59 },
		{ 209, 67, 59 },
		{ 209, 68, 59 },
		{ 209, 69, 59 },
		{ 209, 70, 59 },
		{ 209, 71, 59 },
		{ 209, 72, 59 },
		{ 209, 73, 59 },
		{ 209, 74, 59 },
		{ 209, 75, 59 },
		{ 209, 76, 59 },
		{ 209, 77, 59 },
		{ 209, 78, 59 },
		{ 209, 79, 59 },
		{ 209, 80, 59 },
		{ 209, 81, 59 },
		{ 209, 82, 59 },
		{ 209, 83, 59 },
		{ 209, 84, 59 },
		{ 209, 85, 59 },
		{ 209, 86, 59 },
		{ 209, 87, 59 },
		{ 209, 88, 59 },
		{ 209, 89, 59 },
		{ 209, 90, 59 },
		{ 209, 91, 59 },
		{ 209, 92, 59 },
		{ 209, 93, 59 },
		{ 209, 94, 59 },
		{ 209, 95, 59 },
		{ 209, 96, 59 },
		{ 209, 97, 59 },
		{ 209, 98, 59 },
		{ 209, 99, 59 },
		{ 209, 100, 59 },
		{ 209, 101, 59 },
		{ 209, 102, 59 },
		{ 209, 103, 59 },
		{ 209, 104, 59 },
		{ 209, 105, 59 },
		{ 209, 106, 59 },
		{ 209, 107, 59 },
		{ 209, 108, 59 },
		{ 209, 109, 59 },
		{ 209, 110, 59 },
		{ 209, 111, 59 },
		{ 209, 112, 59 },
		{ 209, 113, 59 },
		{ 209, 114, 59 },
		{ 209, 115, 59 },
		{ 209, 116, 59 },
		{ 209, 117, 59 },
		{ 209, 118, 59 },
		{ 209, 119, 59 },
		{ 209, 120, 59 },
		{ 209, 121, 59 },
		{ 209, 122, 59 },
		{ 209, 123, 59 },
		{ 209, 124, 59 },
		{ 209, 125, 59 },
		{ 209, 126, 59 },
		{ 209, 127, 59 },
		{ 209, 128, 59 },
		{ 209, 129, 59 },
		{ 209, 130, 59 },
		{ 209, 131, 59 },
		{ 209, 132, 59 },
		{ 209, 133, 59 },
		{ 209, 134, 59 },
		{ 209, 135, 59 },
		{ 209, 136, 59 },
		{ 209, 137, 59 },
		{ 209, 138, 59 },
		{ 209, 139, 59 },
		{ 209, 140, 59 },
		{ 209, 141, 59 },
		{ 209, 142, 59 },
		{ 209, 143, 59 },
		{ 209, 144, 59 },
		{ 209, 145, 59 },
		{ 209, 146, 59 },
		{ 209, 147, 59 },
		{ 209, 148, 59 },
		{ 209, 149, 59 },
		{ 209, 150, 59 },
		{ 209, 151, 59 },
		{ 209, 152, 59 },
		{ 209, 153, 59 },
		{ 209, 154, 59 },
		{ 209, 155, 59 },
		{ 209, 156, 59 },
		{ 209, 157, 59 },
		{ 209, 158, 59 },
		{ 209, 159, 59 },
		{ 209, 160, 59 },
		{ 210, 70, 59 },
		{ 210, 71, 59 },
		{ 210, 72, 59 },
		{ 210, 73, 59 },
		{ 210, 74, 59 },
		{ 210, 75, 59 },
		{ 210, 76, 59 },
		{ 210, 77, 59 },
		{ 210, 78, 59 },
		{ 210, 79, 59 },
		{ 210, 80, 59 },
		{ 210, 81, 59 },
		{ 210, 82, 59 },
		{ 210, 83, 59 },
		{ 210, 84, 59 },
		{ 210, 85, 59 },
		{ 210, 86, 59 },
		{ 210, 87, 59 },
		{ 210, 88, 59 },
		{ 210, 89, 59 },
		{ 210, 90, 59 },
		{ 210, 91, 59 },
		{ 210, 92, 59 },
		{ 210, 93, 59 },
		{ 210, 94, 59 },
		{ 210, 95, 59 },
		{ 210, 96, 59 },
		{ 210, 97, 59 },
		{ 210, 98, 59 },
		{ 210, 99, 59 },
		{ 210, 100, 59 },
		{ 210, 101, 59 },
		{ 210, 102, 59 },
		{ 210, 103, 59 },
		{ 210, 104, 59 },
		{ 210, 105, 59 },
		{ 210, 106, 59 },
		{ 210, 107, 59 },
		{ 210, 108, 59 },
		{ 210, 109, 59 },
		{ 210, 110, 59 },
		{ 210, 111, 59 },
		{ 210, 112, 59 },
		{ 210, 113, 59 },
		{ 210, 114, 59 },
		{ 210, 115, 59 },
		{ 210, 116, 59 },
		{ 210, 117, 59 },
		{ 210, 118, 59 },
		{ 210, 119, 59 },
		{ 210, 120, 59 },
		{ 210, 121, 59 },
		{ 210, 122, 59 },
		{ 210, 123, 59 },
		{ 210, 124, 59 },
		{ 210, 125, 59 },
		{ 210, 126, 59 },
		{ 210, 127, 59 },
		{ 210, 128, 59 },
		{ 210, 129, 59 },
		{ 210, 130, 59 },
		{ 210, 131, 59 },
		{ 210, 132, 59 },
		{ 210, 133, 59 },
		{ 210, 134, 59 },
		{ 210, 135, 59 },
		{ 210, 136, 59 },
		{ 210, 137, 59 },
		{ 210, 138, 59 },
		{ 210, 139, 59 },
		{ 210, 140, 59 },
		{ 210, 141, 59 },
		{ 210, 142, 59 },
		{ 210, 143, 59 },
		{ 210, 144, 59 },
		{ 210, 145, 59 },
		{ 210, 146, 59 },
		{ 210, 147, 59 },
		{ 210, 148, 59 },
		{ 210, 149, 59 },
		{ 210, 150, 59 },
		{ 210, 151, 59 },
		{ 210, 152, 59 },
		{ 210, 153, 59 },
		{ 210, 154, 59 },
		{ 210, 155, 59 },
		{ 210, 156, 59 },
		{ 210, 157, 59 },
		{ 210, 158, 59 },
		{ 210, 159, 59 },
		{ 210, 160, 59 },
		{ 210, 161, 59 },
		{ 210, 162, 59 },
		{ 211, 74, 59 },
		{ 211, 75, 59 },
		{ 211, 76, 59 },
		{ 211, 77, 59 },
		{ 211, 78, 59 },
		{ 211, 79, 59 },
		{ 211, 80, 59 },
		{ 211, 81, 59 },
		{ 211, 82, 59 },
		{ 211, 83, 59 },
		{ 211, 84, 59 },
		{ 211, 85, 59 },
		{ 211, 86, 59 },
		{ 211, 87, 59 },
		{ 211, 88, 59 },
		{ 211, 89, 59 },
		{ 211, 90, 59 },
		{ 211, 91, 59 },
		{ 211, 92, 59 },
		{ 211, 93, 59 },
		{ 211, 94, 59 },
		{ 211, 95, 59 },
		{ 211, 96, 59 },
		{ 211, 97, 59 },
		{ 211, 98, 59 },
		{ 211, 99, 59 },
		{ 211, 100, 59 },
		{ 211, 101, 59 },
		{ 211, 102, 59 },
		{ 211, 103, 59 },
		{ 211, 104, 59 },
		{ 211, 105, 59 },
		{ 211, 106, 59 },
		{ 211, 107, 59 },
		{ 211, 108, 59 },
		{ 211, 109, 59 },
		{ 211, 110, 59 },
		{ 211, 111, 59 },
		{ 211, 112, 59 },
		{ 211, 113, 59 },
		{ 211, 114, 59 },
		{ 211, 115, 59 },
		{ 211, 116, 59 },
		{ 211, 117, 59 },
		{ 211, 118, 59 },
		{ 211, 119, 59 },
		{ 211, 120, 59 },
		{ 211, 121, 59 },
		{ 211, 122, 59 },
		{ 211, 123, 59 },
		{ 211, 124, 59 },
		{ 211, 125, 59 },
		{ 211, 126, 59 },
		{ 211, 127, 59 },
		{ 211, 128, 59 },
		{ 211, 129, 59 },
		{ 211, 130, 59 },
		{ 211, 131, 59 },
		{ 211, 132, 59 },
		{ 211, 133, 59 },
		{ 211, 134, 59 },
		{ 211, 135, 59 },
		{ 211, 136, 59 },
		{ 211, 137, 59 },
		{ 211, 138, 59 },
		{ 211, 139, 59 },
		{ 211, 140, 59 },
		{ 211, 141, 59 },
		{ 211, 142, 59 },
		{ 211, 143, 59 },
		{ 211, 144, 59 },
		{ 211, 145, 59 },
		{ 211, 146, 59 },
		{ 211, 147, 59 },
		{ 211, 148, 59 },
		{ 211, 149, 59 },
		{ 211, 150, 59 },
		{ 211, 151, 59 },
		{ 211, 152, 59 },
		{ 211, 153, 59 },
		{ 211, 154, 59 },
		{ 211, 155, 59 },
		{ 211, 156, 59 },
		{ 211, 157, 59 },
		{ 211, 158, 59 },
		{ 211, 159, 59 },
		{ 211, 160, 59 },
		{ 211, 161, 59 },
		{ 211, 162, 59 },
		{ 211, 163, 59 },
		{ 211, 164, 59 },
		{ 212, 78, 59 },
		{ 212, 79, 59 },
		{ 212, 80, 59 },
		{ 212, 81, 59 },
		{ 212, 82, 59 },
		{ 212, 83, 59 },
		{ 212, 84, 59 },
		{ 212, 85, 59 },
		{ 212, 86, 59 },
		{ 212, 87, 59 },
		{ 212, 88, 59 },
		{ 212, 89, 59 },
		{ 212, 90, 59 },
		{ 212, 91, 59 },
		{ 212, 92, 59 },
		{ 212, 93, 59 },
		{ 212, 94, 59 },
		{ 212, 95, 59 },
		{ 212, 96, 59 },
		{ 212, 97, 59 },
		{ 212, 98, 59 },
		{ 212, 99, 59 },
		{ 212, 100, 59 },
		{ 212, 101, 59 },
		{ 212, 102, 59 },
		{ 212, 103, 59 },
		{ 212, 104, 59 },
		{ 212, 105, 59 },
		{ 212, 106, 59 },
		{ 212, 107, 59 },
		{ 212, 108, 59 },
		{ 212, 109, 59 },
		{ 212, 110, 59 },
		{ 212, 111, 59 },
		{ 212, 112, 59 },
		{ 212, 113, 59 },
		{ 212, 114, 59 },
		{ 212, 115, 59 },
		{ 212, 116, 59 },
		{ 212, 117, 59 },
		{ 212, 118, 59 },
		{ 212, 119, 59 },
		{ 212, 120, 59 },
		{ 212, 121, 59 },
		{ 212, 122, 59 },
		{ 212, 123, 59 },
		{ 212, 124, 59 },
		{ 212, 125, 59 },
		{ 212, 126, 59 },
		{ 212, 127, 59 },
		{ 212, 128, 59 },
		{ 212, 129, 59 },
		{ 212, 130, 59 },
		{ 212, 131, 59 },
		{ 212, 132, 59 },
		{ 212, 133, 59 },
		{ 212, 134, 59 },
		{ 212, 135, 59 },
		{ 212, 136, 59 },
		{ 212, 137, 59 },
		{ 212, 138, 59 },
		{ 212, 139, 59 },
		{ 212, 140, 59 },
		{ 212, 141, 59 },
		{ 212, 142, 59 },
		{ 212, 143, 59 },
		{ 212, 144, 59 },
		{ 212, 145, 59 },
		{ 212, 146, 59 },
		{ 212, 147, 59 },
		{ 212, 148, 59 },
		{ 212, 149, 59 },
		{ 212, 150, 59 },
		{ 212, 151, 59 },
		{ 212, 152, 59 },
		{ 212, 153, 59 },
		{ 212, 154, 59 },
		{ 212, 155, 59 },
		{ 212, 156, 59 },
		{ 212, 157, 59 },
		{ 212, 158, 59 },
		{ 212, 159, 59 },
		{ 212, 160, 59 },
		{ 212, 161, 59 },
		{ 212, 162, 59 },
		{ 212, 163, 59 },
		{ 212, 164, 59 },
		{ 212, 165, 59 },
		{ 212, 166, 59 },
		{ 213, 82, 59 },
		{ 213, 83, 59 },
		{ 213, 84, 59 },
		{ 213, 85, 59 },
		{ 213, 86, 59 },
		{ 213, 87, 59 },
		{ 213, 88, 59 },
		{ 213, 89, 59 },
		{ 213, 90, 59 },
		{ 213, 91, 59 },
		{ 213, 92, 59 },
		{ 213, 93, 59 },
		{ 213, 94, 59 },
		{ 213, 95, 59 },
		{ 213, 96, 59 },
		{ 213, 97, 59 },
		{ 213, 98, 59 },
		{ 213, 99, 59 },
		{ 213, 100, 59 },
		{ 213, 101, 59 },
		{ 213, 102, 59 },
		{ 213, 103, 59 },
		{ 213, 104, 59 },
		{ 213, 105, 59 },
		{ 213, 106, 59 },
		{ 213, 107, 59 },
		{ 213, 108, 59 },
		{ 213, 109, 59 },
		{ 213, 110, 59 },
		{ 213, 111, 59 },
		{ 213, 112, 59 },
		{ 213, 113, 59 },
		{ 213, 114, 59 },
		{ 213, 115, 59 },
		{ 213, 116, 59 },
		{ 213, 117, 59 },
		{ 213, 118, 59 },
		{ 213, 119, 59 },
		{ 213, 120, 59 },
		{ 213, 121, 59 },
		{ 213, 122, 59 },
		{ 213, 123, 59 },
		{ 213, 124, 59 },
		{ 213, 125, 59 },
		{ 213, 126, 59 },
		{ 213, 127, 59 },
		{ 213, 128, 59 },
		{ 213, 129, 59 },
		{ 213, 130, 59 },
		{ 213, 131, 59 },
		{ 213, 132, 59 },
		{ 213, 133, 59 },
		{ 213, 134, 59 },
		{ 213, 135, 59 },
		{ 213, 136, 59 },
		{ 213, 137, 59 },
		{ 213, 138, 59 },
		{ 213, 139, 59 },
		{ 213, 140, 59 },
		{ 213, 141, 59 },
		{ 213, 142, 59 },
		{ 213, 143, 59 },
		{ 213, 144, 59 },
		{ 213, 145, 59 },
		{ 213, 146, 59 },
		{ 213, 147, 59 },
		{ 213, 148, 59 },
		{ 213, 149, 59 },
		{ 213, 150, 59 },
		{ 213, 151, 59 },
		{ 213, 152, 59 },
		{ 213, 153, 59 },
		{ 213, 154, 59 },
		{ 213, 155, 59 },
		{ 213, 156, 59 },
		{ 213, 157, 59 },
		{ 213, 158, 59 },
		{ 213, 159, 59 },
		{ 213, 160, 59 },
		{ 213, 161, 59 },
		{ 213, 162, 59 },
		{ 213, 163, 59 },
		{ 213, 164, 59 },
		{ 213, 165, 59 },
		{ 213, 166, 59 },
		{ 213, 167, 59 },
		{ 213, 168, 59 },
		{ 214, 86, 59 },
		{ 214, 87, 59 },
		{ 214, 88, 59 },
		{ 214, 89, 59 },
		{ 214, 90, 59 },
		{ 214, 91, 59 },
		{ 214, 92, 59 },
		{ 214, 93, 59 },
		{ 214, 94, 59 },
		{ 214, 95, 59 },
		{ 214, 96, 59 },
		{ 214, 97, 59 },
		{ 214, 98, 59 },
		{ 214, 99, 59 },
		{ 214, 100, 59 },
		{ 214, 101, 59 },
		{ 214, 102, 59 },
		{ 214, 103, 59 },
		{ 214, 104, 59 },
		{ 214, 105, 59 },
		{ 214, 106, 59 },
		{ 214, 107, 59 },
		{ 214, 108, 59 },
		{ 214, 109, 59 },
		{ 214, 110, 59 },
		{ 214, 111, 59 },
		{ 214, 112, 59 },
		{ 214, 113, 59 },
		{ 214, 114, 59 },
		{ 214, 115, 59 },
		{ 214, 116, 59 },
		{ 214, 117, 59 },
		{ 214, 118, 59 },
		{ 214, 119, 59 },
		{ 214, 120, 59 },
		{ 214, 121, 59 },
		{ 214, 122, 59 },
		{ 214, 123, 59 },
		{ 214, 124, 59 },
		{ 214, 125, 59 },
		{ 214, 126, 59 },
		{ 214, 127, 59 },
		{ 214, 128, 59 },
		{ 214, 129, 59 },
		{ 214, 130, 59 },
		{ 214, 131, 59 },
		{ 214, 132, 59 },
		{ 214, 133, 59 },
		{ 214, 134, 59 },
		{ 214, 135, 59 },
		{ 214, 136, 59 },
		{ 214, 137, 59 },
		{ 214, 138, 59 },
		{ 214, 139, 59 },
		{ 214, 140, 59 },
		{ 214, 141, 59 },
		{ 214, 142, 59 },
		{ 214, 143, 59 },
		{ 214, 144, 59 },
		{ 214, 145, 59 },
		{ 214, 146, 59 },
		{ 214, 147, 59 },
		{ 214, 148, 59 },
		{ 214, 149, 59 },
		{ 214, 150, 59 },
		{ 214, 151, 59 },
		{ 214, 152, 59 },
		{ 214, 153, 59 },
		{ 214, 154, 59 },
		{ 214, 155, 59 },
		{ 214, 156, 59 },
		{ 214, 157, 59 },
		{ 214, 158, 59 },
		{ 214, 159, 59 },
		{ 214, 160, 59 },
		{ 214, 161, 59 },
		{ 214, 162, 59 },
		{ 214, 163, 59 },
		{ 214, 164, 59 },
		{ 214, 165, 59 },
		{ 214, 166, 59 },
		{ 214, 167, 59 },
		{ 214, 168, 59 },
		{ 214, 169, 59 },
		{ 214, 170, 59 },
		{ 215, 90, 59 },
		{ 215, 91, 59 },
		{ 215, 92, 59 },
		{ 215, 93, 59 },
		{ 215, 94, 59 },
		{ 215, 95, 59 },
		{ 215, 96, 59 },
		{ 215, 97, 59 },
		{ 215, 98, 59 },
		{ 215, 99, 59 },
		{ 215, 100, 59 },
		{ 215, 101, 59 },
		{ 215, 102, 59 },
		{ 215, 103, 59 },
		{ 215, 104, 59 },
		{ 215, 105, 59 },
		{ 215, 106, 59 },
		{ 215, 107, 59 },
		{ 215, 108, 59 },
		{ 215, 109, 59 },
		{ 215, 110, 59 },
		{ 215, 111, 59 },
		{ 215, 112, 59 },
		{ 215, 113, 59 },
		{ 215, 114, 59 },
		{ 215, 115, 59 },
		{ 215, 116, 59 },
		{ 215, 117, 59 },
		{ 215, 118, 59 },
		{ 215, 119, 59 },
		{ 215, 120, 59 },
		{ 215, 121, 59 },
		{ 215, 122, 59 },
		{ 215, 123, 59 },
		{ 215, 124, 59 },
		{ 215, 125, 59 },
		{ 215, 126, 59 },
		{ 215, 127, 59 },
		{ 215, 128, 59 },
		{ 215, 129, 59 },
		{ 215, 130, 59 },
		{ 215, 131, 59 },
		{ 215, 132, 59 },
		{ 215, 133, 59 },
		{ 215, 134, 59 },
		{ 215, 135, 59 },
		{ 215, 136, 59 },
		{ 215, 137, 59 },
		{ 215, 138, 59 },
		{ 215, 139, 59 },
		{ 215, 140, 59 },
		{ 215, 141, 59 },
		{ 215, 142, 59 },
		{ 215, 143, 59 },
		{ 215, 144, 59 },
		{ 215, 145, 59 },
		{ 215, 146, 59 },
		{ 215, 147, 59 },
		{ 215, 148, 59 },
		{ 215, 149, 59 },
		{ 215, 150, 59 },
		{ 215, 151, 59 },
		{ 215, 152, 59 },
		{ 215, 153, 59 },
		{ 215, 154, 59 },
		{ 215, 155, 59 },
		{ 215, 156, 59 },
		{ 215, 157, 59 },
		{ 215, 158, 59 },
		{ 215, 159, 59 },
		{ 215, 160, 59 },
		{ 215, 161, 59 },
		{ 215, 162, 59 },
		{ 215, 163, 59 },
		{ 215, 164, 59 },
		{ 215, 165, 59 },
		{ 215, 166, 59 },
		{ 215, 167, 59 },
		{ 215, 168, 59 },
		{ 215, 169, 59 },
		{ 215, 170, 59 },
		{ 215, 171, 59 },
		{ 215, 172, 59 },
		{ 216, 94, 59 },
		{ 216, 95, 59 },
		{ 216, 96, 59 },
		{ 216, 97, 59 },
		{ 216, 98, 59 },
		{ 216, 99, 59 },
		{ 216, 100, 59 },
		{ 216, 101, 59 },
		{ 216, 102, 59 },
		{ 216, 103, 59 },
		{ 216, 104, 59 },
		{ 216, 105, 59 },
		{ 216, 106, 59 },
		{ 216, 107, 59 },
		{ 216, 108, 59 },
		{ 216, 109, 59 },
		{ 216, 110, 59 },
		{ 216, 111, 59 },
		{ 216, 112, 59 },
		{ 216, 113, 59 },
		{ 216, 114, 59 },
		{ 216, 115, 59 },
		{ 216, 116, 59 },
		{ 216, 117, 59 },
		{ 216, 118, 59 },
		{ 216, 119, 59 },
		{ 216, 120, 59 },
		{ 216, 121, 59 },
		{ 216, 122, 59 },
		{ 216, 123, 59 },
		{ 216, 124, 59 },
		{ 216, 125, 59 },
		{ 216, 126, 59 },
		{ 216, 127, 59 },
		{ 216, 128, 59 },
		{ 216, 129, 59 },
		{ 216, 130, 59 },
		{ 216, 131, 59 },
		{ 216, 132, 59 },
		{ 216, 133, 59 },
		{ 216, 134, 59 },
		{ 216, 135, 59 },
		{ 216, 136, 59 },
		{ 216, 137, 59 },
		{ 216, 138, 59 },
		{ 216, 139, 59 },
		{ 216, 140, 59 },
		{ 216, 141, 59 },
		{ 216, 142, 59 },
		{ 216, 143, 59 },
		{ 216, 144, 59 },
		{ 216, 145, 59 },
		{ 216, 146, 59 },
		{ 216, 147, 59 },
		{ 216, 148, 59 },
		{ 216, 149, 59 },
		{ 216, 150, 59 },
		{ 216, 151, 59 },
		{ 216, 152, 59 },
		{ 216, 153, 59 },
		{ 216, 154, 59 },
		{ 216, 155, 59 },
		{ 216, 156, 59 },
		{ 216, 157, 59 },
		{ 216, 158, 59 },
		{ 216, 159, 59 },
		{ 216, 160, 59 },
		{ 216, 161, 59 },
		{ 216, 162, 59 },
		{ 216, 163, 59 },
		{ 216, 164, 59 },
		{ 216, 165, 59 },
		{ 216, 166, 59 },
		{ 216, 167, 59 },
		{ 216, 168, 59 },
		{ 216, 169, 59 },
		{ 216, 170, 59 },
		{ 216, 171, 59 },
		{ 216, 172, 59 },
		{ 216, 173, 59 },
		{ 216, 174, 59 },
		{ 217, 98, 59 },
		{ 217, 99, 59 },
		{ 217, 100, 59 },
		{ 217, 101, 59 },
		{ 217, 102, 59 },
		{ 217, 103, 59 },
		{ 217, 104, 59 },
		{ 217, 105, 59 },
		{ 217, 106, 59 },
		{ 217, 107, 59 },
		{ 217, 108, 59 },
		{ 217, 109, 59 },
		{ 217, 110, 59 },
		{ 217, 111, 59 },
		{ 217, 112, 59 },
		{ 217, 113, 59 },
		{ 217, 114, 59 },
		{ 217, 115, 59 },
		{ 217, 116, 59 },
		{ 217, 117, 59 },
		{ 217, 118, 59 },
		{ 217, 119, 59 },
		{ 217, 120, 59 },
		{ 217, 121, 59 },
		{ 217, 122, 59 },
		{ 217, 123, 59 },
		{ 217, 124, 59 },
		{ 217, 125, 59 },
		{ 217, 126, 59 },
		{ 217, 127, 59 },
		{ 217, 128, 59 },
		{ 217, 129, 59 },
		{ 217, 130, 59 },
		{ 217, 131, 59 },
		{ 217, 132, 59 },
		{ 217, 133, 59 },
		{ 217, 134, 59 },
		{ 217, 135, 59 },
		{ 217, 136, 59 },
		{ 217, 137, 59 },
		{ 217, 138, 59 },
		{ 217, 139, 59 },
		{ 217, 140, 59 },
		{ 217, 141, 59 },
		{ 217, 142, 59 },
		{ 217, 143, 59 },
		{ 217, 144, 59 },
		{ 217, 145, 59 },
		{ 217, 146, 59 },
		{ 217, 147, 59 },
		{ 217, 148, 59 },
		{ 217, 149, 59 },
		{ 217, 150, 59 },
		{ 217, 151, 59 },
		{ 217, 152, 59 },
		{ 217, 153, 59 },
		{ 217, 154, 59 },
		{ 217, 155, 59 },
		{ 217, 156, 59 },
		{ 217, 157, 59 },
		{ 217, 158, 59 },
		{ 217, 159, 59 },
		{ 217, 160, 59 },
		{ 217, 161, 59 },
		{ 217, 162, 59 },
		{ 217, 163, 59 },
		{ 217, 164, 59 },
		{ 217, 165, 59 },
		{ 217, 166, 59 },
		{ 217, 167, 59 },
		{ 217, 168, 59 },
		{ 217, 169, 59 },
		{ 217, 170, 59 },
		{ 217, 171, 59 },
		{ 217, 172, 59 },
		{ 217, 173, 59 },
		{ 217, 174, 59 },
		{ 217, 175, 59 },
		{ 217, 176, 59 },
		{ 218, 102, 59 },
		{ 218, 103, 59 },
		{ 218, 104, 59 },
		{ 218, 105, 59 },
		{ 218, 106, 59 },
		{ 218, 107, 59 },
		{ 218, 108, 59 },
		{ 218, 109, 59 },
		{ 218, 110, 59 },
		{ 218, 111, 59 },
		{ 218, 112, 59 },
		{ 218, 113, 59 },
		{ 218, 114, 59 },
		{ 218, 115, 59 },
		{ 218, 116, 59 },
		{ 218, 117, 59 },
		{ 218, 118, 59 },
		{ 218, 119, 59 },
		{ 218, 120, 59 },
		{ 218, 121, 59 },
		{ 218, 122, 59 },
		{ 218, 123, 59 },
		{ 218, 124, 59 },
		{ 218, 125, 59 },
		{ 218, 126, 59 },
		{ 218, 127, 59 },
		{ 218, 128, 59 },
		{ 218, 129, 59 },
		{ 218, 130, 59 },
		{ 218, 131, 59 },
		{ 218, 132, 59 },
		{ 218, 133, 59 },
		{ 218, 134, 59 },
		{ 218, 135, 59 },
		{ 218, 136, 59 },
		{ 218, 137, 59 },
		{ 218, 138, 59 },
		{ 218, 139, 59 },
		{ 218, 140, 59 },
		{ 218, 141, 59 },
		{ 218, 142, 59 },
		{ 218, 143, 59 },
		{ 218, 144, 59 },
		{ 218, 145, 59 },
		{ 218, 146, 59 },
		{ 218, 147, 59 },
		{ 218, 148, 59 },
		{ 218, 149, 59 },
		{ 218, 150, 59 },
		{ 218, 151, 59 },
		{ 218, 152, 59 },
		{ 218, 153, 59 },
		{ 218, 154, 59 },
		{ 218, 155, 59 },
		{ 218, 156, 59 },
		{ 218, 157, 59 },
		{ 218, 158, 59 },
		{ 218, 159, 59 },
		{ 218, 160, 59 },
		{ 218, 161, 59 },
		{ 218, 162, 59 },
		{ 218, 163, 59 },
		{ 218, 164, 59 },
		{ 218, 165, 59 },
		{ 218, 166, 59 },
		{ 218, 167, 59 },
		{ 218, 168, 59 },
		{ 218, 169, 59 },
		{ 218, 170, 59 },
		{ 218, 171, 59 },
		{ 218, 172, 59 },
		{ 218, 173, 59 },
		{ 218, 174, 59 },
		{ 218, 175, 59 },
		{ 218, 176, 59 },
		{ 218, 177, 59 },
		{ 218, 178, 59 },
		{ 219, 106, 59 },
		{ 219, 107, 59 },
		{ 219, 108, 59 },
		{ 219, 109, 59 },
		{ 219, 110, 59 },
		{ 219, 111, 59 },
		{ 219, 112, 59 },
		{ 219, 113, 59 },
		{ 219, 114, 59 },
		{ 219, 115, 59 },
		{ 219, 116, 59 },
		{ 219, 117, 59 },
		{ 219, 118, 59 },
		{ 219, 119, 59 },
		{ 219, 120, 59 },
		{ 219, 121, 59 },
		{ 219, 122, 59 },
		{ 219, 123, 59 },
		{ 219, 124, 59 },
		{ 219, 125, 59 },
		{ 219, 126, 59 },
		{ 219, 127, 59 },
		{ 219, 128, 59 },
		{ 219, 129, 59 },
		{ 219, 130, 59 },
		{ 219, 131, 59 },
		{ 219, 132, 59 },
		{ 219, 133, 59 },
		{ 219, 134, 59 },
		{ 219, 135, 59 },
		{ 219, 136, 59 },
		{ 219, 137, 59 },
		{ 219, 138, 59 },
		{ 219, 139, 59 },
		{ 219, 140, 59 },
		{ 219, 141, 59 },
		{ 219, 142, 59 },
		{ 219, 143, 59 },
		{ 219, 144, 59 },
		{ 219, 145, 59 },
		{ 219, 146, 59 },
		{ 219, 147, 59 },
		{ 219, 148, 59 },
		{ 219, 149, 59 },
		{ 219, 150, 59 },
		{ 219, 151, 59 },
		{ 219, 152, 59 },
		{ 219, 153, 59 },
		{ 219, 154, 59 },
		{ 219, 155, 59 },
		{ 219, 156, 59 },
		{ 219, 157, 59 },
		{ 219, 158, 59 },
		{ 219, 159, 59 },
		{ 219, 160, 59 },
		{ 219, 161, 59 },
		{ 219, 162, 59 },
		{ 219, 163, 59 },
		{ 219, 164, 59 },
		{ 219, 165, 59 },
		{ 219, 166, 59 },
		{ 219, 167, 59 },
		{ 219, 168, 59 },
		{ 219, 169, 59 },
		{ 219, 170, 59 },
		{ 219, 171, 59 },
		{ 219, 172, 59 },
		{ 219, 173, 59 },
		{ 219, 174, 59 },
		{ 219, 175, 59 },
		{ 219, 176, 59 },
		{ 219, 177, 59 },
		{ 219, 178, 59 },
		{ 219, 179, 59 },
		{ 219, 180, 59 },
		{ 220, 110, 59 },
		{ 220, 111, 59 },
		{ 220, 112, 59 },
		{ 220, 113, 59 },
		{ 220, 114, 59 },
		{ 220, 115, 59 },
		{ 220, 116, 59 },
		{ 220, 117, 59 },
		{ 220, 118, 59 },
		{ 220, 119, 59 },
		{ 220, 120, 59 },
		{ 220, 121, 59 },
		{ 220, 122, 59 },
		{ 220, 123, 59 },
		{ 220, 124, 59 },
		{ 220, 125, 59 },
		{ 220, 126, 59 },
		{ 220, 127, 59 },
		{ 220, 128, 59 },
		{ 220, 129, 59 },
		{ 220, 130, 59 },
		{ 220, 131, 59 },
		{ 220, 132, 59 },
		{ 220, 133, 59 },
		{ 220, 134, 59 },
		{ 220, 135, 59 },
		{ 220, 136, 59 },
		{ 220, 137, 59 },
		{ 220, 138, 59 },
		{ 220, 139, 59 },
		{ 220, 140, 59 },
		{ 220, 141, 59 },
		{ 220, 142, 59 },
		{ 220, 143, 59 },
		{ 220, 144, 59 },
		{ 220, 145, 59 },
		{ 220, 146, 59 },
		{ 220, 147, 59 },
		{ 220, 148, 59 },
		{ 220, 149, 59 },
		{ 220, 150, 59 },
		{ 220, 151, 59 },
		{ 220, 152, 59 },
		{ 220, 153, 59 },
		{ 220, 154, 59 },
		{ 220, 155, 59 },
		{ 220, 156, 59 },
		{ 220, 157, 59 },
		{ 220, 158, 59 },
		{ 220, 159, 59 },
		{ 220, 160, 59 },
		{ 220, 161, 59 },
		{ 220, 162, 59 },
		{ 220, 163, 59 },
		{ 220, 164, 59 },
		{ 220, 165, 59 },
		{ 220, 166, 59 },
		{ 220, 167, 59 },
		{ 220, 168, 59 },
		{ 220, 169, 59 },
		{ 220, 170, 59 },
		{ 220, 171, 59 },
		{ 220, 172, 59 },
		{ 220, 173, 59 },
		{ 220, 174, 59 },
		{ 220, 175, 59 },
		{ 220, 176, 59 },
		{ 220, 177, 59 },
		{ 220, 178, 59 },
		{ 220, 179, 59 },
		{ 220, 180, 59 },
		{ 220, 181, 59 },
		{ 220, 182, 59 },
		{ 221, 114, 59 },
		{ 221, 115, 59 },
		{ 221, 116, 59 },
		{ 221, 117, 59 },
		{ 221, 118, 59 },
		{ 221, 119, 59 },
		{ 221, 120, 59 },
		{ 221, 121, 59 },
		{ 221, 122, 59 },
		{ 221, 123, 59 },
		{ 221, 124, 59 },
		{ 221, 125, 59 },
		{ 221, 126, 59 },
		{ 221, 127, 59 },
		{ 221, 128, 59 },
		{ 221, 129, 59 },
		{ 221, 130, 59 },
		{ 221, 131, 59 },
		{ 221, 132, 59 },
		{ 221, 133, 59 },
		{ 221, 134, 59 },
		{ 221, 135, 59 },
		{ 221, 136, 59 },
		{ 221, 137, 59 },
		{ 221, 138, 59 },
		{ 221, 139, 59 },
		{ 221, 140, 59 },
		{ 221, 141, 59 },
		{ 221, 142, 59 },
		{ 221, 143, 59 },
		{ 221, 144, 59 },
		{ 221, 145, 59 },
		{ 221, 146, 59 },
		{ 221, 147, 59 },
		{ 221, 148, 59 },
		{ 221, 149, 59 },
		{ 221, 150, 59 },
		{ 221, 151, 59 },
		{ 221, 152, 59 },
		{ 221, 153, 59 },
		{ 221, 154, 59 },
		{ 221, 155, 59 },
		{ 221, 156, 59 },
		{ 221, 157, 59 },
		{ 221, 158, 59 },
		{ 221, 159, 59 },
		{ 221, 160, 59 },
		{ 221, 161, 59 },
		{ 221, 162, 59 },
		{ 221, 163, 59 },
		{ 221, 164, 59 },
		{ 221, 165, 59 },
		{ 221, 166, 59 },
		{ 221, 167, 59 },
		{ 221, 168, 59 },
		{ 221, 169, 59 },
		{ 221, 170, 59 },
		{ 221, 171, 59 },
		{ 221, 172, 59 },
		{ 221, 173, 59 },
		{ 221, 174, 59 },
		{ 221, 175, 59 },
		{ 221, 176, 59 },
		{ 221, 177, 59 },
		{ 221, 178, 59 },
		{ 221, 179, 59 },
		{ 221, 180, 59 },
		{ 221, 181, 59 },
		{ 221, 182, 59 },
		{ 221, 183, 59 },
		{ 221, 184, 59 },
		{ 222, 118, 59 },
		{ 222, 119, 59 },
		{ 222, 120, 59 },
		{ 222, 121, 59 },
		{ 222, 122, 59 },
		{ 222, 123, 59 },
		{ 222, 124, 59 },
		{ 222, 125, 59 },
		{ 222, 126, 59 },
		{ 222, 127, 59 },
		{ 222, 128, 59 },
		{ 222, 129, 59 },
		{ 222, 130, 59 },
		{ 222, 131, 59 },
		{ 222, 132, 59 },
		{ 222, 133, 59 },
		{ 222, 134, 59 },
		{ 222, 135, 59 },
		{ 222, 136, 59 },
		{ 222, 137, 59 },
		{ 222, 138, 59 },
		{ 222, 139, 59 },
		{ 222, 140, 59 },
		{ 222, 141, 59 },
		{ 222, 142, 59 },
		{ 222, 143, 59 },
		{ 222, 144, 59 },
		{ 222, 145, 59 },
		{ 222, 146, 59 },
		{ 222, 147, 59 },
		{ 222, 148, 59 },
		{ 222, 149, 59 },
		{ 222, 150, 59 },
		{ 222, 151, 59 },
		{ 222, 152, 59 },
		{ 222, 153, 59 },
		{ 222, 154, 59 },
		{ 222, 155, 59 },
		{ 222, 156, 59 },
		{ 222, 157, 59 },
		{ 222, 158, 59 },
		{ 222, 159, 59 },
		{ 222, 160, 59 },
		{ 222, 161, 59 },
		{ 222, 162, 59 },
		{ 222, 163, 59 },
		{ 222, 164, 59 },
		{ 222, 165, 59 },
		{ 222, 166, 59 },
		{ 222, 167, 59 },
		{ 222, 168, 59 },
		{ 222, 169, 59 },
		{ 222, 170, 59 },
		{ 222, 171, 59 },
		{ 222, 172, 59 },
		{ 222, 173, 59 },
		{ 222, 174, 59 },
		{ 222, 175, 59 },
		{ 222, 176, 59 },
		{ 222, 177, 59 },
		{ 222, 178, 59 },
		{ 222, 179, 59 },
		{ 222, 180, 59 },
		{ 222, 181, 59 },
		{ 222, 182, 59 },
		{ 222, 183, 59 },
		{ 222, 184, 59 },
		{ 222, 185, 59 },
		{ 222, 186, 59 },
		{ 223, 122, 59 },
		{ 223, 123, 59 },
		{ 223, 124, 59 },
		{ 223, 125, 59 },
		{ 223, 126, 59 },
		{ 223, 127, 59 },
		{ 223, 128, 59 },
		{ 223, 129, 59 },
		{ 223, 130, 59 },
		{ 223, 131, 59 },
		{ 223, 132, 59 },
		{ 223, 133, 59 },
		{ 223, 134, 59 },
		{ 223, 135, 59 },
		{ 223, 136, 59 },
		{ 223, 137, 59 },
		{ 223, 138, 59 },
		{ 223, 139, 59 },
		{ 223, 140, 59 },
		{ 223, 141, 59 },
		{ 223, 142, 59 },
		{ 223, 143, 59 },
		{ 223, 144, 59 },
		{ 223, 145, 59 },
		{ 223, 146, 59 },
		{ 223, 147, 59 },
		{ 223, 148, 59 },
		{ 223, 149, 59 },
		{ 223, 150, 59 },
		{ 223, 151, 59 },
		{ 223, 152, 59 },
		{ 223, 153, 59 },
		{ 223, 154, 59 },
		{ 223, 155, 59 },
		{ 223, 156, 59 },
		{ 223, 157, 59 },
		{ 223, 158, 59 },
		{ 223, 159, 59 },
		{ 223, 160, 59 },
		{ 223, 161, 59 },
		{ 223, 162, 59 },
		{ 223, 163, 59 },
		{ 223, 164, 59 },
		{ 223, 165, 59 },
		{ 223, 166, 59 },
		{ 223, 167, 59 },
		{ 223, 168, 59 },
		{ 223, 169, 59 },
		{ 223, 170, 59 },
		{ 223, 171, 59 },
		{ 223, 172, 59 },
		{ 223, 173, 59 },
		{ 223, 174, 59 },
		{ 223, 175, 59 },
		{ 223, 176, 59 },
		{ 223, 177, 59 },
		{ 223, 178, 59 },
		{ 223, 179, 59 },
		{ 223, 180, 59 },
		{ 223, 181, 59 },
		{ 223, 182, 59 },
		{ 223, 183, 59 },
		{ 223, 184, 59 },
		{ 223, 185, 59 },
		{ 223, 186, 59 },
		{ 223, 187, 59 },
		{ 223, 188, 59 },
		{ 224, 126, 59 },
		{ 224, 127, 59 },
		{ 224, 128, 59 },
		{ 224, 129, 59 },
		{ 224, 130, 59 },
		{ 224, 131, 59 },
		{ 224, 132, 59 },
		{ 224, 133, 59 },
		{ 224, 134, 59 },
		{ 224, 135, 59 },
		{ 224, 136, 59 },
		{ 224, 137, 59 },
		{ 224, 138, 59 },
		{ 224, 139, 59 },
		{ 224, 140, 59 },
		{ 224, 141, 59 },
		{ 224, 142, 59 },
		{ 224, 143, 59 },
		{ 224, 144, 59 },
		{ 224, 145, 59 },
		{ 224, 146, 59 },
		{ 224, 147, 59 },
		{ 224, 148, 59 },
		{ 224, 149, 59 },
		{ 224, 150, 59 },
		{ 224, 151, 59 },
		{ 224, 152, 59 },
		{ 224, 153, 59 },
		{ 224, 154, 59 },
		{ 224, 155, 59 },
		{ 224, 156, 59 },
		{ 224, 157, 59 },
		{ 224, 158, 59 },
		{ 224, 159, 59 },
		{ 224, 160, 59 },
		{ 224, 161, 59 },
		{ 224, 162, 59 },
		{ 224, 163, 59 },
		{ 224, 164, 59 },
		{ 224, 165, 59 },
		{ 224, 166, 59 },
		{ 224, 167, 59 },
		{ 224, 168, 59 },
		{ 224, 169, 59 },
		{ 224, 170, 59 },
		{ 224, 171, 59 },
		{ 224, 172, 59 },
		{ 224, 173, 59 },
		{ 224, 174, 59 },
		{ 224, 175, 59 },
		{ 224, 176, 59 },
		{ 224, 177, 59 },
		{ 224, 178, 59 },
		{ 224, 179, 59 },
		{ 224, 180, 59 },
		{ 224, 181, 59 },
		{ 224, 182, 59 },
		{ 224, 183, 59 },
		{ 224, 184, 59 },
		{ 224, 185, 59 },
		{ 224, 186, 59 },
		{ 224, 187, 59 },
		{ 224, 188, 59 },
		{ 224, 189, 59 },
		{ 225, 130, 59 },
		{ 225, 131, 59 },
		{ 225, 132, 59 },
		{ 225, 133, 59 },
		{ 225, 134, 59 },
		{ 225, 135, 59 },
		{ 225, 136, 59 },
		{ 225, 137, 59 },
		{ 225, 138, 59 },
		{ 225, 139, 59 },
		{ 225, 140, 59 },
		{ 225, 141, 59 },
		{ 225, 142, 59 },
		{ 225, 143, 59 },
		{ 225, 144, 59 },
		{ 225, 145, 59 },
		{ 225, 146, 59 },
		{ 225, 147, 59 },
		{ 225, 148, 59 },
		{ 225, 149, 59 },
		{ 225, 150, 59 },
		{ 225, 151, 59 },
		{ 225, 152, 59 },
		{ 225, 153, 59 },
		{ 225, 154, 59 },
		{ 225, 155, 59 },
		{ 225, 156, 59 },
		{ 225, 157, 59 },
		{ 225, 158, 59 },
		{ 225, 159, 59 },
		{ 225, 160, 59 },
		{ 225, 161, 59 },
		{ 225, 162, 59 },
		{ 225, 163, 59 },
		{ 225, 164, 59 },
		{ 225, 165, 59 },
		{ 225, 166, 59 },
		{ 225, 167, 59 },
		{ 225, 168, 59 },
		{ 225, 169, 59 },
		{ 225, 170, 59 },
		{ 225, 171, 59 },
		{ 225, 172, 59 },
		{ 225, 173, 59 },
		{ 225, 174, 59 },
		{ 225, 175, 59 },
		{ 225, 176, 59 },
		{ 225, 177, 59 },
		{ 225, 178, 59 },
		{ 225, 179, 59 },
		{ 225, 180, 59 },
		{ 225, 181, 59 },
		{ 225, 182, 59 },
		{ 225, 183, 59 },
		{ 225, 184, 59 },
		{ 225, 185, 59 },
		{ 225, 186, 59 },
		{ 225, 187, 59 },
		{ 225, 188, 59 },
		{ 225, 189, 59 },
		{ 225, 190, 59 },
		{ 226, 134, 59 },
		{ 226, 135, 59 },
		{ 226, 136, 59 },
		{ 226, 137, 59 },
		{ 226, 138, 59 },
		{ 226, 139, 59 },
		{ 226, 140, 59 },
		{ 226, 141, 59 },
		{ 226, 142, 59 },
		{ 226, 143, 59 },
		{ 226, 144, 59 },
		{ 226, 145, 59 },
		{ 226, 146, 59 },
		{ 226, 147, 59 },
		{ 226, 148, 59 },
		{ 226, 149, 59 },
		{ 226, 150, 59 },
		{ 226, 151, 59 },
		{ 226, 152, 59 },
		{ 226, 153, 59 },
		{ 226, 154, 59 },
		{ 226, 155, 59 },
		{ 226, 156, 59 },
		{ 226, 157, 59 },
		{ 226, 158, 59 },
		{ 226, 159, 59 },
		{ 226, 160, 59 },
		{ 226, 161, 59 },
		{ 226, 162, 59 },
		{ 226, 163, 59 },
		{ 226, 164, 59 },
		{ 226, 165, 59 },
		{ 226, 166, 59 },
		{ 226, 167, 59 },
		{ 226, 168, 59 },
		{ 226, 169, 59 },
		{ 226, 170, 59 },
		{ 226, 171, 59 },
		{ 226, 172, 59 },
		{ 226, 173, 59 },
		{ 226, 174, 59 },
		{ 226, 175, 59 },
		{ 226, 176, 59 },
		{ 226, 177, 59 },
		{ 226, 178, 59 },
		{ 226, 179, 59 },
		{ 226, 180, 59 },
		{ 226, 181, 59 },
		{ 226, 182, 59 },
		{ 226, 183, 59 },
		{ 226, 184, 59 },
		{ 226, 185, 59 },
		{ 226, 186, 59 },
		{ 226, 187, 59 },
		{ 226, 188, 59 },
		{ 226, 189, 59 },
		{ 226, 190, 59 },
		{ 226, 191, 59 },
		{ 227, 138, 59 },
		{ 227, 139, 59 },
		{ 227, 140, 59 },
		{ 227, 141, 59 },
		{ 227, 142, 59 },
		{ 227, 143, 59 },
		{ 227, 144, 59 },
		{ 227, 145, 59 },
		{ 227, 146, 59 },
		{ 227, 147, 59 },
		{ 227, 148, 59 },
		{ 227, 149, 59 },
		{ 227, 150, 59 },
		{ 227, 151, 59 },
		{ 227, 152, 59 },
		{ 227, 153, 59 },
		{ 227, 154, 59 },
		{ 227, 155, 59 },
		{ 227, 156, 59 },
		{ 227, 157, 59 },
		{ 227, 158, 59 },
		{ 227, 159, 59 },
		{ 227, 160, 59 },
		{ 227, 161, 59 },
		{ 227, 162, 59 },
		{ 227, 163, 59 },
		{ 227, 164, 59 },
		{ 227, 165, 59 },
		{ 227, 166, 59 },
		{ 227, 167, 59 },
		{ 227, 168, 59 },
		{ 227, 169, 59 },
		{ 227, 170, 59 },
		{ 227, 171, 59 },
		{ 227, 172, 59 },
		{ 227, 173, 59 },
		{ 227, 174, 59 },
		{ 227, 175, 59 },
		{ 227, 176, 59 },
		{ 227, 177, 59 },
		{ 227, 178, 59 },
		{ 227, 179, 59 },
		{ 227, 180, 59 },
		{ 227, 181, 59 },
		{ 227, 182, 59 },
		{ 227, 183, 59 },
		{ 227, 184, 59 },
		{ 227, 185, 59 },
		{ 227, 186, 59 },
		{ 227, 187, 59 },
		{ 227, 188, 59 },
		{ 227, 189, 59 },
		{ 227, 190, 59 },
		{ 227, 191, 59 },
		{ 228, 142, 59 },
		{ 228, 143, 59 },
		{ 228, 144, 59 },
		{ 228, 145, 59 },
		{ 228, 146, 59 },
		{ 228, 147, 59 },
		{ 228, 148, 59 },
		{ 228, 149, 59 },
		{ 228, 150, 59 },
		{ 228, 151, 59 },
		{ 228, 152, 59 },
		{ 228, 153, 59 },
		{ 228, 154, 59 },
		{ 228, 155, 59 },
		{ 228, 156, 59 },
		{ 228, 157, 59 },
		{ 228, 158, 59 },
		{ 228, 159, 59 },
		{ 228, 160, 59 },
		{ 228, 161, 59 },
		{ 228, 162, 59 },
		{ 228, 163, 59 },
		{ 228, 164, 59 },
		{ 228, 165, 59 },
		{ 228, 166, 59 },
		{ 228, 167, 59 },
		{ 228, 168, 59 },
		{ 228, 169, 59 },
		{ 228, 170, 59 },
		{ 228, 171, 59 },
		{ 228, 172, 59 },
		{ 228, 173, 59 },
		{ 228, 174, 59 },
		{ 228, 175, 59 },
		{ 228, 176, 59 },
		{ 228, 177, 59 },
		{ 228, 178, 59 },
		{ 228, 179, 59 },
		{ 228, 180, 59 },
		{ 228, 181, 59 },
		{ 228, 182, 59 },
		{ 228, 183, 59 },
		{ 228, 184, 59 },
		{ 228, 185, 59 },
		{ 228, 186, 59 },
		{ 228, 187, 59 },
		{ 228, 188, 59 },
		{ 228, 189, 59 },
		{ 228, 190, 59 },
		{ 228, 191, 59 },
		{ 229, 146, 59 },
		{ 229, 147, 59 },
		{ 229, 148, 59 },
		{ 229, 149, 59 },
		{ 229, 150, 59 },
		{ 229, 151, 59 },
		{ 229, 152, 59 },
		{ 229, 153, 59 },
		{ 229, 154, 59 },
		{ 229, 155, 59 },
		{ 229, 156, 59 },
		{ 229, 157, 59 },
		{ 229, 158, 59 },
		{ 229, 159, 59 },
		{ 229, 160, 59 },
		{ 229, 161, 59 },
		{ 229, 162, 59 },
		{ 229, 163, 59 },
		{ 229, 164, 59 },
		{ 229, 165, 59 },
		{ 229, 166, 59 },
		{ 229, 167, 59 },
		{ 229, 168, 59 },
		{ 229, 169, 59 },
		{ 229, 170, 59 },
		{ 229, 171, 59 },
		{ 229, 172, 59 },
		{ 229, 173, 59 },
		{ 229, 174, 59 },
		{ 229, 175, 59 },
		{ 229, 176, 59 },
		{ 229, 177, 59 },
		{ 229, 178, 59 },
		{ 229, 179, 59 },
		{ 229, 180, 59 },
		{ 229, 181, 59 },
		{ 229, 182, 59 },
		{ 229, 183, 59 },
		{ 229, 184, 59 },
		{ 229, 185, 59 },
		{ 229, 186, 59 },
		{ 229, 187, 59 },
		{ 229, 188, 59 },
		{ 229, 189, 59 },
		{ 229, 190, 59 },
		{ 229, 191, 59 },
		{ 230, 150, 59 },
		{ 230, 151, 59 },
		{ 230, 152, 59 },
		{ 230, 153, 59 },
		{ 230, 154, 59 },
		{ 230, 155, 59 },
		{ 230, 156, 59 },
		{ 230, 157, 59 },
		{ 230, 158, 59 },
		{ 230, 159, 59 },
		{ 230, 160, 59 },
		{ 230, 161, 59 },
		{ 230, 162, 59 },
		{ 230, 163, 59 },
		{ 230, 164, 59 },
		{ 230, 165, 59 },
		{ 230, 166, 59 },
		{ 230, 167, 59 },
		{ 230, 168, 59 },
		{ 230, 169, 59 },
		{ 230, 170, 59 },
		{ 230, 171, 59 },
		{ 230, 172, 59 },
		{ 230, 173, 59 },
		{ 230, 174, 59 },
		{ 230, 175, 59 },
		{ 230, 176, 59 },
		{ 230, 177, 59 },
		{ 230, 178, 59 },
		{ 230, 179, 59 },
		{ 230, 180, 59 },
		{ 230, 181, 59 },
		{ 230, 182, 59 },
		{ 230, 183, 59 },
		{ 230, 184, 59 },
		{ 230, 185, 59 },
		{ 230, 186, 59 },
		{ 230, 187, 59 },
		{ 230, 188, 59 },
		{ 230, 189, 59 },
		{ 230, 190, 59 },
		{ 230, 191, 59 },
		{ 231, 154, 59 },
		{ 231, 155, 59 },
		{ 231, 156, 59 },
		{ 231, 157, 59 },
		{ 231, 158, 59 },
		{ 231, 159, 59 },
		{ 231, 160, 59 },
		{ 231, 161, 59 },
		{ 231, 162, 59 },
		{ 231, 163, 59 },
		{ 231, 164, 59 },
		{ 231, 165, 59 },
		{ 231, 166, 59 },
		{ 231, 167, 59 },
		{ 231, 168, 59 },
		{ 231, 169, 59 },
		{ 231, 170, 59 },
		{ 231, 171, 59 },
		{ 231, 172, 59 },
		{ 231, 173, 59 },
		{ 231, 174, 59 },
		{ 231, 175, 59 },
		{ 231, 176, 59 },
		{ 231, 177, 59 },
		{ 231, 178, 59 },
		{ 231, 179, 59 },
		{ 231, 180, 59 },
		{ 231, 181, 59 },
		{ 231, 182, 59 },
		{ 231, 183, 59 },
		{ 231, 184, 59 },
		{ 231, 185, 59 },
		{ 231, 186, 59 },
		{ 231, 187, 59 },
		{ 231, 188, 59 },
		{ 231, 189, 59 },
		{ 231, 190, 59 },
		{ 231, 191, 59 },
		{ 232, 158, 59 },
		{ 232, 159, 59 },
		{ 232, 160, 59 },
		{ 232, 161, 59 },
		{ 232, 162, 59 },
		{ 232, 163, 59 },
		{ 232, 164, 59 },
		{ 232, 165, 59 },
		{ 232, 166, 59 },
		{ 232, 167, 59 },
		{ 232, 168, 59 },
		{ 232, 169, 59 },
		{ 232, 170, 59 },
		{ 232, 171, 59 },
		{ 232, 172, 59 },
		{ 232, 173, 59 },
		{ 232, 174, 59 },
		{ 232, 175, 59 },
		{ 232, 176, 59 },
		{ 232, 177, 59 },
		{ 232, 178, 59 },
		{ 232, 179, 59 },
		{ 232, 180, 59 },
		{ 232, 181, 59 },
		{ 232, 182, 59 },
		{ 232, 183, 59 },
		{ 232, 184, 59 },
		{ 232, 185, 59 },
		{ 232, 186, 59 },
		{ 232, 187, 59 },
		{ 232, 188, 59 },
		{ 232, 189, 59 },
		{ 232, 190, 59 },
		{ 232, 191, 59 },
		{ 233, 162, 59 },
		{ 233, 163, 59 },
		{ 233, 164, 59 },
		{ 233, 165, 59 },
		{ 233, 166, 59 },
		{ 233, 167, 59 },
		{ 233, 168, 59 },
		{ 233, 169, 59 },
		{ 233, 170, 59 },
		{ 233, 171, 59 },
		{ 233, 172, 59 },
		{ 233, 173, 59 },
		{ 233, 174, 59 },
		{ 233, 175, 59 },
		{ 233, 176, 59 },
		{ 233, 177, 59 },
		{ 233, 178, 59 },
		{ 233, 179, 59 },
		{ 233, 180, 59 },
		{ 233, 181, 59 },
		{ 233, 182, 59 },
		{ 233, 183, 59 },
		{ 233, 184, 59 },
		{ 233, 185, 59 },
		{ 233, 186, 59 },
		{ 233, 187, 59 },
		{ 233, 188, 59 },
		{ 233, 189, 59 },
		{ 233, 190, 59 },
		{ 233, 191, 59 },
		{ 234, 166, 59 },
		{ 234, 167, 59 },
		{ 234, 168, 59 },
		{ 234, 169, 59 },
		{ 234, 170, 59 },
		{ 234, 171, 59 },
		{ 234, 172, 59 },
		{ 234, 173, 59 },
		{ 234, 174, 59 },
		{ 234, 175, 59 },
		{ 234, 176, 59 },
		{ 234, 177, 59 },
		{ 234, 178, 59 },
		{ 234, 179, 59 },
		{ 234, 180, 59 },
		{ 234, 181, 59 },
		{ 234, 182, 59 },
		{ 234, 183, 59 },
		{ 234, 184, 59 },
		{ 234, 185, 59 },
		{ 234, 186, 59 },
		{ 234, 187, 59 },
		{ 234, 188, 59 },
		{ 234, 189, 59 },
		{ 234, 190, 59 },
		{ 234, 191, 59 },
		{ 235, 170, 59 },
		{ 235, 171, 59 },
		{ 235, 172, 59 },
		{ 235, 173, 59 },
		{ 235, 174, 59 },
		{ 235, 175, 59 },
		{ 235, 176, 59 },
		{ 235, 177, 59 },
		{ 235, 178, 59 },
		{ 235, 179, 59 },
		{ 235, 180, 59 },
		{ 235, 181, 59 },
		{ 235, 182, 59 },
		{ 235, 183, 59 },
		{ 235, 184, 59 },
		{ 235, 185, 59 },
		{ 235, 186, 59 },
		{ 235, 187, 59 },
		{ 235, 188, 59 },
		{ 235, 189, 59 },
		{ 235, 190, 59 },
		{ 235, 191, 59 },
		{ 236, 174, 59 },
		{ 236, 175, 59 },
		{ 236, 176, 59 },
		{ 236, 177, 59 },
		{ 236, 178, 59 },
		{ 236, 179, 59 },
		{ 236, 180, 59 },
		{ 236, 181, 59 },
		{ 236, 182, 59 },
		{ 236, 183, 59 },
		{ 236, 184, 59 },
		{ 236, 185, 59 },
		{ 236, 186, 59 },
		{ 236, 187, 59 },
		{ 236, 188, 59 },
		{ 236, 189, 59 },
		{ 236, 190, 59 },
		{ 236, 191, 59 },
		{ 237, 178, 59 },
		{ 237, 179, 59 },
		{ 237, 180, 59 },
		{ 237, 181, 59 },
		{ 237, 182, 59 },
		{ 237, 183, 59 },
		{ 237, 184, 59 },
		{ 237, 185, 59 },
		{ 237, 186, 59 },
		{ 237, 187, 59 },
		{ 237, 188, 59 },
		{ 237, 189, 59 },
		{ 237, 190, 59 },
		{ 237, 191, 59 },
		{ 238, 182, 59 },
		{ 238, 183, 59 },
		{ 238, 184, 59 },
		{ 238, 185, 59 },
		{ 238, 186, 59 },
		{ 238, 187, 59 },
		{ 238, 188, 59 },
		{ 238, 189, 59 },
		{ 238, 190, 59 },
		{ 238, 191, 59 },
		{ 239, 186, 59 },
		{ 239, 187, 59 },
		{ 239, 188, 59 },
		{ 239, 189, 59 },
		{ 239, 190, 59 },
		{ 275, 324, 59 },
		{ 275, 325, 59 },
		{ 275, 326, 59 },
		{ 275, 327, 59 },
		{ 275, 328, 59 },
		{ 276, 323, 59 },
		{ 276, 324, 59 },
		{ 276, 325, 59 },
		{ 276, 326, 59 },
		{ 276, 327, 59 },
		{ 276, 328, 59 },
		{ 276, 329, 59 },
		{ 276, 330, 59 },
		{ 276, 331, 59 },
		{ 276, 332, 59 },
		{ 277, 323, 59 },
		{ 277, 324, 59 },
		{ 277, 325, 59 },
		{ 277, 326, 59 },
		{ 277, 327, 59 },
		{ 277, 328, 59 },
		{ 277, 329, 59 },
		{ 277, 330, 59 },
		{ 277, 331, 59 },
		{ 277, 332, 59 },
		{ 277, 333, 59 },
		{ 277, 334, 59 },
		{ 277, 335, 59 },
		{ 277, 336, 59 },
		{ 278, 323, 59 },
		{ 278, 324, 59 },
		{ 278, 325, 59 },
		{ 278, 326, 59 },
		{ 278, 327, 59 },
		{ 278, 328, 59 },
		{ 278, 329, 59 },
		{ 278, 330, 59 },
		{ 278, 331, 59 },
		{ 278, 332, 59 },
		{ 278, 333, 59 },
		{ 278, 334, 59 },
		{ 278, 335, 59 },
		{ 278, 336, 59 },
		{ 278, 337, 59 },
		{ 278, 338, 59 },
		{ 278, 339, 59 },
		{ 278, 340, 59 },
		{ 279, 323, 59 },
		{ 279, 324, 59 },
		{ 279, 325, 59 },
		{ 279, 326, 59 },
		{ 279, 327, 59 },
		{ 279, 328, 59 },
		{ 279, 329, 59 },
		{ 279, 330, 59 },
		{ 279, 331, 59 },
		{ 279, 332, 59 },
		{ 279, 333, 59 },
		{ 279, 334, 59 },
		{ 279, 335, 59 },
		{ 279, 336, 59 },
		{ 279, 337, 59 },
		{ 279, 338, 59 },
		{ 279, 339, 59 },
		{ 279, 340, 59 },
		{ 279, 341, 59 },
		{ 279, 342, 59 },
		{ 279, 343, 59 },
		{ 279, 344, 59 },
		{ 280, 323, 59 },
		{ 280, 324, 59 },
		{ 280, 325, 59 },
		{ 280, 326, 59 },
		{ 280, 327, 59 },
		{ 280, 328, 59 },
		{ 280, 329, 59 },
		{ 280, 330, 59 },
		{ 280, 331, 59 },
		{ 280, 332, 59 },
		{ 280, 333, 59 },
		{ 280, 334, 59 },
		{ 280, 335, 59 },
		{ 280, 336, 59 },
		{ 280, 337, 59 },
		{ 280, 338, 59 },
		{ 280, 339, 59 },
		{ 280, 340, 59 },
		{ 280, 341, 59 },
		{ 280, 342, 59 },
		{ 280, 343, 59 },
		{ 280, 344, 59 },
		{ 280, 345, 59 },
		{ 280, 346, 59 },
		{ 280, 347, 59 },
		{ 280, 348, 59 },
		{ 281, 323, 59 },
		{ 281, 324, 59 },
		{ 281, 325, 59 },
		{ 281, 326, 59 },
		{ 281, 327, 59 },
		{ 281, 328, 59 },
		{ 281, 329, 59 },
		{ 281, 330, 59 },
		{ 281, 331, 59 },
		{ 281, 332, 59 },
		{ 281, 333, 59 },
		{ 281, 334, 59 },
		{ 281, 335, 59 },
		{ 281, 336, 59 },
		{ 281, 337, 59 },
		{ 281, 338, 59 },
		{ 281, 339, 59 },
		{ 281, 340, 59 },
		{ 281, 341, 59 },
		{ 281, 342, 59 },
		{ 281, 343, 59 },
		{ 281, 344, 59 },
		{ 281, 345, 59 },
		{ 281, 346, 59 },
		{ 281, 347, 59 },
		{ 281, 348, 59 },
		{ 281, 349, 59 },
		{ 281, 350, 59 },
		{ 281, 351, 59 },
		{ 281, 352, 59 },
		{ 282, 323, 59 },
		{ 282, 324, 59 },
		{ 282, 325, 59 },
		{ 282, 326, 59 },
		{ 282, 327, 59 },
		{ 282, 328, 59 },
		{ 282, 329, 59 },
		{ 282, 330, 59 },
		{ 282, 331, 59 },
		{ 282, 332, 59 },
		{ 282, 333, 59 },
		{ 282, 334, 59 },
		{ 282, 335, 59 },
		{ 282, 336, 59 },
		{ 282, 337, 59 },
		{ 282, 338, 59 },
		{ 282, 339, 59 },
		{ 282, 340, 59 },
		{ 282, 341, 59 },
		{ 282, 342, 59 },
		{ 282, 343, 59 },
		{ 282, 344, 59 },
		{ 282, 345, 59 },
		{ 282, 346, 59 },
		{ 282, 347, 59 },
		{ 282, 348, 59 },
		{ 282, 349, 59 },
		{ 282, 350, 59 },
		{ 282, 351, 59 },
		{ 282, 352, 59 },
		{ 282, 353, 59 },
		{ 282, 354, 59 },
		{ 282, 355, 59 },
		{ 282, 356, 59 },
		{ 283, 323, 59 },
		{ 283, 324, 59 },
		{ 283, 325, 59 },
		{ 283, 326, 59 },
		{ 283, 327, 59 },
		{ 283, 328, 59 },
		{ 283, 329, 59 },
		{ 283, 330, 59 },
		{ 283, 331, 59 },
		{ 283, 332, 59 },
		{ 283, 333, 59 },
		{ 283, 334, 59 },
		{ 283, 335, 59 },
		{ 283, 336, 59 },
		{ 283, 337, 59 },
		{ 283, 338, 59 },
		{ 283, 339, 59 },
		{ 283, 340, 59 },
		{ 283, 341, 59 },
		{ 283, 342, 59 },
		{ 283, 343, 59 },
		{ 283, 344, 59 },
		{ 283, 345, 59 },
		{ 283, 346, 59 },
		{ 283, 347, 59 },
		{ 283, 348, 59 },
		{ 283, 349, 59 },
		{ 283, 350, 59 },
		{ 283, 351, 59 },
		{ 283, 352, 59 },
		{ 283, 353, 59 },
		{ 283, 354, 59 },
		{ 283, 355, 59 },
		{ 283, 356, 59 },
		{ 283, 357, 59 },
		{ 283, 358, 59 },
		{ 283, 359, 59 },
		{ 283, 360, 59 },
		{ 284, 323, 59 },
		{ 284, 324, 59 },
		{ 284, 325, 59 },
		{ 284, 326, 59 },
		{ 284, 327, 59 },
		{ 284, 328, 59 },
		{ 284, 329, 59 },
		{ 284, 330, 59 },
		{ 284, 331, 59 },
		{ 284, 332, 59 },
		{ 284, 333, 59 },
		{ 284, 334, 59 },
		{ 284, 335, 59 },
		{ 284, 336, 59 },
		{ 284, 337, 59 },
		{ 284, 338, 59 },
		{ 284, 339, 59 },
		{ 284, 340, 59 },
		{ 284, 341, 59 },
		{ 284, 342, 59 },
		{ 284, 343, 59 },
		{ 284, 344, 59 },
		{ 284, 345, 59 },
		{ 284, 346, 59 },
		{ 284, 347, 59 },
		{ 284, 348, 59 },
		{ 284, 349, 59 },
		{ 284, 350, 59 },
		{ 284, 351, 59 },
		{ 284, 352, 59 },
		{ 284, 353, 59 },
		{ 284, 354, 59 },
		{ 284, 355, 59 },
		{ 284, 356, 59 },
		{ 284, 357, 59 },
		{ 284, 358, 59 },
		{ 284, 359, 59 },
		{ 284, 360, 59 },
		{ 284, 361, 59 },
		{ 284, 362, 59 },
		{ 284, 363, 59 },
		{ 284, 364, 59 },
		{ 285, 323, 59 },
		{ 285, 324, 59 },
		{ 285, 325, 59 },
		{ 285, 326, 59 },
		{ 285, 327, 59 },
		{ 285, 328, 59 },
		{ 285, 329, 59 },
		{ 285, 330, 59 },
		{ 285, 331, 59 },
		{ 285, 332, 59 },
		{ 285, 333, 59 },
		{ 285, 334, 59 },
		{ 285, 335, 59 },
		{ 285, 336, 59 },
		{ 285, 337, 59 },
		{ 285, 338, 59 },
		{ 285, 339, 59 },
		{ 285, 340, 59 },
		{ 285, 341, 59 },
		{ 285, 342, 59 },
		{ 285, 343, 59 },
		{ 285, 344, 59 },
		{ 285, 345, 59 },
		{ 285, 346, 59 },
		{ 285, 347, 59 },
		{ 285, 348, 59 },
		{ 285, 349, 59 },
		{ 285, 350, 59 },
		{ 285, 351, 59 },
		{ 285, 352, 59 },
		{ 285, 353, 59 },
		{ 285, 354, 59 },
		{ 285, 355, 59 },
		{ 285, 356, 59 },
		{ 285, 357, 59 },
		{ 285, 358, 59 },
		{ 285, 359, 59 },
		{ 285, 360, 59 },
		{ 285, 361, 59 },
		{ 285, 362, 59 },
		{ 285, 363, 59 },
		{ 285, 364, 59 },
		{ 285, 365, 59 },
		{ 285, 366, 59 },
		{ 285, 367, 59 },
		{ 285, 368, 59 },
		{ 286, 323, 59 },
		{ 286, 324, 59 },
		{ 286, 325, 59 },
		{ 286, 326, 59 },
		{ 286, 327, 59 },
		{ 286, 328, 59 },
		{ 286, 329, 59 },
		{ 286, 330, 59 },
		{ 286, 331, 59 },
		{ 286, 332, 59 },
		{ 286, 333, 59 },
		{ 286, 334, 59 },
		{ 286, 335, 59 },
		{ 286, 336, 59 },
		{ 286, 337, 59 },
		{ 286, 338, 59 },
		{ 286, 339, 59 },
		{ 286, 340, 59 },
		{ 286, 341, 59 },
		{ 286, 342, 59 },
		{ 286, 343, 59 },
		{ 286, 344, 59 },
		{ 286, 345, 59 },
		{ 286, 346, 59 },
		{ 286, 347, 59 },
		{ 286, 348, 59 },
		{ 286, 349, 59 },
		{ 286, 350, 59 },
		{ 286, 351, 59 },
		{ 286, 352, 59 },
		{ 286, 353, 59 },
		{ 286, 354, 59 },
		{ 286, 355, 59 },
		{ 286, 356, 59 },
		{ 286, 357, 59 },
		{ 286, 358, 59 },
		{ 286, 359, 59 },
		{ 286, 360, 59 },
		{ 286, 361, 59 },
		{ 286, 362, 59 },
		{ 286, 363, 59 },
		{ 286, 364, 59 },
		{ 286, 365, 59 },
		{ 286, 366, 59 },
		{ 286, 367, 59 },
		{ 286, 368, 59 },
		{ 286, 369, 59 },
		{ 286, 370, 59 },
		{ 286, 371, 59 },
		{ 286, 372, 59 },
		{ 287, 323, 59 },
		{ 287, 324, 59 },
		{ 287, 325, 59 },
		{ 287, 326, 59 },
		{ 287, 327, 59 },
		{ 287, 328, 59 },
		{ 287, 329, 59 },
		{ 287, 330, 59 },
		{ 287, 331, 59 },
		{ 287, 332, 59 },
		{ 287, 333, 59 },
		{ 287, 334, 59 },
		{ 287, 335, 59 },
		{ 287, 336, 59 },
		{ 287, 337, 59 },
		{ 287, 338, 59 },
		{ 287, 339, 59 },
		{ 287, 340, 59 },
		{ 287, 341, 59 },
		{ 287, 342, 59 },
		{ 287, 343, 59 },
		{ 287, 344, 59 },
		{ 287, 345, 59 },
		{ 287, 346, 59 },
		{ 287, 347, 59 },
		{ 287, 348, 59 },
		{ 287, 349, 59 },
		{ 287, 350, 59 },
		{ 287, 351, 59 },
		{ 287, 352, 59 },
		{ 287, 353, 59 },
		{ 287, 354, 59 },
		{ 287, 355, 59 },
		{ 287, 356, 59 },
		{ 287, 357, 59 },
		{ 287, 358, 59 },
		{ 287, 359, 59 },
		{ 287, 360, 59 },
		{ 287, 361, 59 },
		{ 287, 362, 59 },
		{ 287, 363, 59 },
		{ 287, 364, 59 },
		{ 287, 365, 59 },
		{ 287, 366, 59 },
		{ 287, 367, 59 },
		{ 287, 368, 59 },
		{ 287, 369, 59 },
		{ 287, 370, 59 },
		{ 287, 371, 59 },
		{ 287, 372, 59 },
		{ 287, 373, 59 },
		{ 287, 374, 59 },
		{ 287, 375, 59 },
		{ 287, 376, 59 },
		{ 288, 323, 59 },
		{ 288, 324, 59 },
		{ 288, 325, 59 },
		{ 288, 326, 59 },
		{ 288, 327, 59 },
		{ 288, 328, 59 },
		{ 288, 329, 59 },
		{ 288, 330, 59 },
		{ 288, 331, 59 },
		{ 288, 332, 59 },
		{ 288, 333, 59 },
		{ 288, 334, 59 },
		{ 288, 335, 59 },
		{ 288, 336, 59 },
		{ 288, 337, 59 },
		{ 288, 338, 59 },
		{ 288, 339, 59 },
		{ 288, 340, 59 },
		{ 288, 341, 59 },
		{ 288, 342, 59 },
		{ 288, 343, 59 },
		{ 288, 344, 59 },
		{ 288, 345, 59 },
		{ 288, 346, 59 },
		{ 288, 347, 59 },
		{ 288, 348, 59 },
		{ 288, 349, 59 },
		{ 288, 350, 59 },
		{ 288, 351, 59 },
		{ 288, 352, 59 },
		{ 288, 353, 59 },
		{ 288, 354, 59 },
		{ 288, 355, 59 },
		{ 288, 356, 59 },
		{ 288, 357, 59 },
		{ 288, 358, 59 },
		{ 288, 359, 59 },
		{ 288, 360, 59 },
		{ 288, 361, 59 },
		{ 288, 362, 59 },
		{ 288, 363, 59 },
		{ 288, 364, 59 },
		{ 288, 365, 59 },
		{ 288, 366, 59 },
		{ 288, 367, 59 },
		{ 288, 368, 59 },
		{ 288, 369, 59 },
		{ 288, 370, 59 },
		{ 288, 371, 59 },
		{ 288, 372, 59 },
		{ 288, 373, 59 },
		{ 288, 374, 59 },
		{ 288, 375, 59 },
		{ 288, 376, 59 },
		{ 288, 377, 59 },
		{ 288, 378, 59 },
		{ 288, 379, 59 },
		{ 288, 380, 59 },
		{ 289, 324, 59 },
		{ 289, 325, 59 },
		{ 289, 326, 59 },
		{ 289, 327, 59 },
		{ 289, 328, 59 },
		{ 289, 329, 59 },
		{ 289, 330, 59 },
		{ 289, 331, 59 },
		{ 289, 332, 59 },
		{ 289, 333, 59 },
		{ 289, 334, 59 },
		{ 289, 335, 59 },
		{ 289, 336, 59 },
		{ 289, 337, 59 },
		{ 289, 338, 59 },
		{ 289, 339, 59 },
		{ 289, 340, 59 },
		{ 289, 341, 59 },
		{ 289, 342, 59 },
		{ 289, 343, 59 },
		{ 289, 344, 59 },
		{ 289, 345, 59 },
		{ 289, 346, 59 },
		{ 289, 347, 59 },
		{ 289, 348, 59 },
		{ 289, 349, 59 },
		{ 289, 350, 59 },
		{ 289, 351, 59 },
		{ 289, 352, 59 },
		{ 289, 353, 59 },
		{ 289, 354, 59 },
		{ 289, 355, 59 },
		{ 289, 356, 59 },
		{ 289, 357, 59 },
		{ 289, 358, 59 },
		{ 289, 359, 59 },
		{ 289, 360, 59 },
		{ 289, 361, 59 },
		{ 289, 362, 59 },
		{ 289, 363, 59 },
		{ 289, 364, 59 },
		{ 289, 365, 59 },
		{ 289, 366, 59 },
		{ 289, 367, 59 },
		{ 289, 368, 59 },
		{ 289, 369, 59 },
		{ 289, 370, 59 },
		{ 289, 371, 59 },
		{ 289, 372, 59 },
		{ 289, 373, 59 },
		{ 289, 374, 59 },
		{ 289, 375, 59 },
		{ 289, 376, 59 },
		{ 289, 377, 59 },
		{ 289, 378, 59 },
		{ 289, 379, 59 },
		{ 289, 380, 59 },
		{ 289, 381, 59 },
		{ 289, 382, 59 },
		{ 289, 383, 59 },
		{ 289, 384, 59 },
		{ 290, 325, 59 },
		{ 290, 326, 59 },
		{ 290, 327, 59 },
		{ 290, 328, 59 },
		{ 290, 329, 59 },
		{ 290, 330, 59 },
		{ 290, 331, 59 },
		{ 290, 332, 59 },
		{ 290, 333, 59 },
		{ 290, 334, 59 },
		{ 290, 335, 59 },
		{ 290, 336, 59 },
		{ 290, 337, 59 },
		{ 290, 338, 59 },
		{ 290, 339, 59 },
		{ 290, 340, 59 },
		{ 290, 341, 59 },
		{ 290, 342, 59 },
		{ 290, 343, 59 },
		{ 290, 344, 59 },
		{ 290, 345, 59 },
		{ 290, 346, 59 },
		{ 290, 347, 59 },
		{ 290, 348, 59 },
		{ 290, 349, 59 },
		{ 290, 350, 59 },
		{ 290, 351, 59 },
		{ 290, 352, 59 },
		{ 290, 353, 59 },
		{ 290, 354, 59 },
		{ 290, 355, 59 },
		{ 290, 356, 59 },
		{ 290, 357, 59 },
		{ 290, 358, 59 },
		{ 290, 359, 59 },
		{ 290, 360, 59 },
		{ 290, 361, 59 },
		{ 290, 362, 59 },
		{ 290, 363, 59 },
		{ 290, 364, 59 },
		{ 290, 365, 59 },
		{ 290, 366, 59 },
		{ 290, 367, 59 },
		{ 290, 368, 59 },
		{ 290, 369, 59 },
		{ 290, 370, 59 },
		{ 290, 371, 59 },
		{ 290, 372, 59 },
		{ 290, 373, 59 },
		{ 290, 374, 59 },
		{ 290, 375, 59 },
		{ 290, 376, 59 },
		{ 290, 377, 59 },
		{ 290, 378, 59 },
		{ 290, 379, 59 },
		{ 290, 380, 59 },
		{ 290, 381, 59 },
		{ 290, 382, 59 },
		{ 290, 383, 59 },
		{ 290, 384, 59 },
		{ 290, 385, 59 },
		{ 290, 386, 59 },
		{ 290, 387, 59 },
		{ 290, 388, 59 },
		{ 291, 326, 59 },
		{ 291, 327, 59 },
		{ 291, 328, 59 },
		{ 291, 329, 59 },
		{ 291, 330, 59 },
		{ 291, 331, 59 },
		{ 291, 332, 59 },
		{ 291, 333, 59 },
		{ 291, 334, 59 },
		{ 291, 335, 59 },
		{ 291, 336, 59 },
		{ 291, 337, 59 },
		{ 291, 338, 59 },
		{ 291, 339, 59 },
		{ 291, 340, 59 },
		{ 291, 341, 59 },
		{ 291, 342, 59 },
		{ 291, 343, 59 },
		{ 291, 344, 59 },
		{ 291, 345, 59 },
		{ 291, 346, 59 },
		{ 291, 347, 59 },
		{ 291, 348, 59 },
		{ 291, 349, 59 },
		{ 291, 350, 59 },
		{ 291, 351, 59 },
		{ 291, 352, 59 },
		{ 291, 353, 59 },
		{ 291, 354, 59 },
		{ 291, 355, 59 },
		{ 291, 356, 59 },
		{ 291, 357, 59 },
		{ 291, 358, 59 },
		{ 291, 359, 59 },
		{ 291, 360, 59 },
		{ 291, 361, 59 },
		{ 291, 362, 59 },
		{ 291, 363, 59 },
		{ 291, 364, 59 },
		{ 291, 365, 59 },
		{ 291, 366, 59 },
		{ 291, 367, 59 },
		{ 291, 368, 59 },
		{ 291, 369, 59 },
		{ 291, 370, 59 },
		{ 291, 371, 59 },
		{ 291, 372, 59 },
		{ 291, 373, 59 },
		{ 291, 374, 59 },
		{ 291, 375, 59 },
		{ 291, 376, 59 },
		{ 291, 377, 59 },
		{ 291, 378, 59 },
		{ 291, 379, 59 },
		{ 291, 380, 59 },
		{ 291, 381, 59 },
		{ 291, 382, 59 },
		{ 291, 383, 59 },
		{ 291, 384, 59 },
		{ 291, 385, 59 },
		{ 291, 386, 59 },
		{ 291, 387, 59 },
		{ 291, 388, 59 },
		{ 291, 389, 59 },
		{ 291, 390, 59 },
		{ 291, 391, 59 },
		{ 291, 392, 59 },
		{ 292, 328, 59 },
		{ 292, 329, 59 },
		{ 292, 330, 59 },
		{ 292, 331, 59 },
		{ 292, 332, 59 },
		{ 292, 333, 59 },
		{ 292, 334, 59 },
		{ 292, 335, 59 },
		{ 292, 336, 59 },
		{ 292, 337, 59 },
		{ 292, 338, 59 },
		{ 292, 339, 59 },
		{ 292, 340, 59 },
		{ 292, 341, 59 },
		{ 292, 342, 59 },
		{ 292, 343, 59 },
		{ 292, 344, 59 },
		{ 292, 345, 59 },
		{ 292, 346, 59 },
		{ 292, 347, 59 },
		{ 292, 348, 59 },
		{ 292, 349, 59 },
		{ 292, 350, 59 },
		{ 292, 351, 59 },
		{ 292, 352, 59 },
		{ 292, 353, 59 },
		{ 292, 354, 59 },
		{ 292, 355, 59 },
		{ 292, 356, 59 },
		{ 292, 357, 59 },
		{ 292, 358, 59 },
		{ 292, 359, 59 },
		{ 292, 360, 59 },
		{ 292, 361, 59 },
		{ 292, 362, 59 },
		{ 292, 363, 59 },
		{ 292, 364, 59 },
		{ 292, 365, 59 },
		{ 292, 366, 59 },
		{ 292, 367, 59 },
		{ 292, 368, 59 },
		{ 292, 369, 59 },
		{ 292, 370, 59 },
		{ 292, 371, 59 },
		{ 292, 372, 59 },
		{ 292, 373, 59 },
		{ 292, 374, 59 },
		{ 292, 375, 59 },
		{ 292, 376, 59 },
		{ 292, 377, 59 },
		{ 292, 378, 59 },
		{ 292, 379, 59 },
		{ 292, 380, 59 },
		{ 292, 381, 59 },
		{ 292, 382, 59 },
		{ 292, 383, 59 },
		{ 292, 384, 59 },
		{ 292, 385, 59 },
		{ 292, 386, 59 },
		{ 292, 387, 59 },
		{ 292, 388, 59 },
		{ 292, 389, 59 },
		{ 292, 390, 59 },
		{ 292, 391, 59 },
		{ 292, 392, 59 },
		{ 292, 393, 59 },
		{ 292, 394, 59 },
		{ 292, 395, 59 },
		{ 292, 396, 59 },
		{ 293, 330, 59 },
		{ 293, 331, 59 },
		{ 293, 332, 59 },
		{ 293, 333, 59 },
		{ 293, 334, 59 },
		{ 293, 335, 59 },
		{ 293, 336, 59 },
		{ 293, 337, 59 },
		{ 293, 338, 59 },
		{ 293, 339, 59 },
		{ 293, 340, 59 },
		{ 293, 341, 59 },
		{ 293, 342, 59 },
		{ 293, 343, 59 },
		{ 293, 344, 59 },
		{ 293, 345, 59 },
		{ 293, 346, 59 },
		{ 293, 347, 59 },
		{ 293, 348, 59 },
		{ 293, 349, 59 },
		{ 293, 350, 59 },
		{ 293, 351, 59 },
		{ 293, 352, 59 },
		{ 293, 353, 59 },
		{ 293, 354, 59 },
		{ 293, 355, 59 },
		{ 293, 356, 59 },
		{ 293, 357, 59 },
		{ 293, 358, 59 },
		{ 293, 359, 59 },
		{ 293, 360, 59 },
		{ 293, 361, 59 },
		{ 293, 362, 59 },
		{ 293, 363, 59 },
		{ 293, 364, 59 },
		{ 293, 365, 59 },
		{ 293, 366, 59 },
		{ 293, 367, 59 },
		{ 293, 368, 59 },
		{ 293, 369, 59 },
		{ 293, 370, 59 },
		{ 293, 371, 59 },
		{ 293, 372, 59 },
		{ 293, 373, 59 },
		{ 293, 374, 59 },
		{ 293, 375, 59 },
		{ 293, 376, 59 },
		{ 293, 377, 59 },
		{ 293, 378, 59 },
		{ 293, 379, 59 },
		{ 293, 380, 59 },
		{ 293, 381, 59 },
		{ 293, 382, 59 },
		{ 293, 383, 59 },
		{ 293, 384, 59 },
		{ 293, 385, 59 },
		{ 293, 386, 59 },
		{ 293, 387, 59 },
		{ 293, 388, 59 },
		{ 293, 389, 59 },
		{ 293, 390, 59 },
		{ 293, 391, 59 },
		{ 293, 392, 59 },
		{ 293, 393, 59 },
		{ 293, 394, 59 },
		{ 293, 395, 59 },
		{ 293, 396, 59 },
		{ 293, 397, 59 },
		{ 293, 398, 59 },
		{ 293, 399, 59 },
		{ 293, 400, 59 },
		{ 294, 332, 59 },
		{ 294, 333, 59 },
		{ 294, 334, 59 },
		{ 294, 335, 59 },
		{ 294, 336, 59 },
		{ 294, 337, 59 },
		{ 294, 338, 59 },
		{ 294, 339, 59 },
		{ 294, 340, 59 },
		{ 294, 341, 59 },
		{ 294, 342, 59 },
		{ 294, 343, 59 },
		{ 294, 344, 59 },
		{ 294, 345, 59 },
		{ 294, 346, 59 },
		{ 294, 347, 59 },
		{ 294, 348, 59 },
		{ 294, 349, 59 },
		{ 294, 350, 59 },
		{ 294, 351, 59 },
		{ 294, 352, 59 },
		{ 294, 353, 59 },
		{ 294, 354, 59 },
		{ 294, 355, 59 },
		{ 294, 356, 59 },
		{ 294, 357, 59 },
		{ 294, 358, 59 },
		{ 294, 359, 59 },
		{ 294, 360, 59 },
		{ 294, 361, 59 },
		{ 294, 362, 59 },
		{ 294, 363, 59 },
		{ 294, 364, 59 },
		{ 294, 365, 59 },
		{ 294, 366, 59 },
		{ 294, 367, 59 },
		{ 294, 368, 59 },
		{ 294, 369, 59 },
		{ 294, 370, 59 },
		{ 294, 371, 59 },
		{ 294, 372, 59 },
		{ 294, 373, 59 },
		{ 294, 374, 59 },
		{ 294, 375, 59 },
		{ 294, 376, 59 },
		{ 294, 377, 59 },
		{ 294, 378, 59 },
		{ 294, 379, 59 },
		{ 294, 380, 59 },
		{ 294, 381, 59 },
		{ 294, 382, 59 },
		{ 294, 383, 59 },
		{ 294, 384, 59 },
		{ 294, 385, 59 },
		{ 294, 386, 59 },
		{ 294, 387, 59 },
		{ 294, 388, 59 },
		{ 294, 389, 59 },
		{ 294, 390, 59 },
		{ 294, 391, 59 },
		{ 294, 392, 59 },
		{ 294, 393, 59 },
		{ 294, 394, 59 },
		{ 294, 395, 59 },
		{ 294, 396, 59 },
		{ 294, 397, 59 },
		{ 294, 398, 59 },
		{ 294, 399, 59 },
		{ 294, 400, 59 },
		{ 294, 401, 59 },
		{ 294, 402, 59 },
		{ 294, 403, 59 },
		{ 294, 404, 59 },
		{ 295, 334, 59 },
		{ 295, 335, 59 },
		{ 295, 336, 59 },
		{ 295, 337, 59 },
		{ 295, 338, 59 },
		{ 295, 339, 59 },
		{ 295, 340, 59 },
		{ 295, 341, 59 },
		{ 295, 342, 59 },
		{ 295, 343, 59 },
		{ 295, 344, 59 },
		{ 295, 345, 59 },
		{ 295, 346, 59 },
		{ 295, 347, 59 },
		{ 295, 348, 59 },
		{ 295, 349, 59 },
		{ 295, 350, 59 },
		{ 295, 351, 59 },
		{ 295, 352, 59 },
		{ 295, 353, 59 },
		{ 295, 354, 59 },
		{ 295, 355, 59 },
		{ 295, 356, 59 },
		{ 295, 357, 59 },
		{ 295, 358, 59 },
		{ 295, 359, 59 },
		{ 295, 360, 59 },
		{ 295, 361, 59 },
		{ 295, 362, 59 },
		{ 295, 363, 59 },
		{ 295, 364, 59 },
		{ 295, 365, 59 },
		{ 295, 366, 59 },
		{ 295, 367, 59 },
		{ 295, 368, 59 },
		{ 295, 369, 59 },
		{ 295, 370, 59 },
		{ 295, 371, 59 },
		{ 295, 372, 59 },
		{ 295, 373, 59 },
		{ 295, 374, 59 },
		{ 295, 375, 59 },
		{ 295, 376, 59 },
		{ 295, 377, 59 },
		{ 295, 378, 59 },
		{ 295, 379, 59 },
		{ 295, 380, 59 },
		{ 295, 381, 59 },
		{ 295, 382, 59 },
		{ 295, 383, 59 },
		{ 295, 384, 59 },
		{ 295, 385, 59 },
		{ 295, 386, 59 },
		{ 295, 387, 59 },
		{ 295, 388, 59 },
		{ 295, 389, 59 },
		{ 295, 390, 59 },
		{ 295, 391, 59 },
		{ 295, 392, 59 },
		{ 295, 393, 59 },
		{ 295, 394, 59 },
		{ 295, 395, 59 },
		{ 295, 396, 59 },
		{ 295, 397, 59 },
		{ 295, 398, 59 },
		{ 295, 399, 59 },
		{ 295, 400, 59 },
		{ 295, 401, 59 },
		{ 295, 402, 59 },
		{ 295, 403, 59 },
		{ 295, 404, 59 },
		{ 295, 405, 59 },
		{ 295, 406, 59 },
		{ 295, 407, 59 },
		{ 295, 408, 59 },
		{ 296, 336, 59 },
		{ 296, 337, 59 },
		{ 296, 338, 59 },
		{ 296, 339, 59 },
		{ 296, 340, 59 },
		{ 296, 341, 59 },
		{ 296, 342, 59 },
		{ 296, 343, 59 },
		{ 296, 344, 59 },
		{ 296, 345, 59 },
		{ 296, 346, 59 },
		{ 296, 347, 59 },
		{ 296, 348, 59 },
		{ 296, 349, 59 },
		{ 296, 350, 59 },
		{ 296, 351, 59 },
		{ 296, 352, 59 },
		{ 296, 353, 59 },
		{ 296, 354, 59 },
		{ 296, 355, 59 },
		{ 296, 356, 59 },
		{ 296, 357, 59 },
		{ 296, 358, 59 },
		{ 296, 359, 59 },
		{ 296, 360, 59 },
		{ 296, 361, 59 },
		{ 296, 362, 59 },
		{ 296, 363, 59 },
		{ 296, 364, 59 },
		{ 296, 365, 59 },
		{ 296, 366, 59 },
		{ 296, 367, 59 },
		{ 296, 368, 59 },
		{ 296, 369, 59 },
		{ 296, 370, 59 },
		{ 296, 371, 59 },
		{ 296, 372, 59 },
		{ 296, 373, 59 },
		{ 296, 374, 59 },
		{ 296, 375, 59 },
		{ 296, 376, 59 },
		{ 296, 377, 59 },
		{ 296, 378, 59 },
		{ 296, 379, 59 },
		{ 296, 380, 59 },
		{ 296, 381, 59 },
		{ 296, 382, 59 },
		{ 296, 383, 59 },
		{ 296, 384, 59 },
		{ 296, 385, 59 },
		{ 296, 386, 59 },
		{ 296, 387, 59 },
		{ 296, 388, 59 },
		{ 296, 389, 59 },
		{ 296, 390, 59 },
		{ 296, 391, 59 },
		{ 296, 392, 59 },
		{ 296, 393, 59 },
		{ 296, 394, 59 },
		{ 296, 395, 59 },
		{ 296, 396, 59 },
		{ 296, 397, 59 },
		{ 296, 398, 59 },
		{ 296, 399, 59 },
		{ 296, 400, 59 },
		{ 296, 401, 59 },
		{ 296, 402, 59 },
		{ 296, 403, 59 },
		{ 296, 404, 59 },
		{ 296, 405, 59 },
		{ 296, 406, 59 },
		{ 296, 407, 59 },
		{ 296, 408, 59 },
		{ 296, 409, 59 },
		{ 296, 410, 59 },
		{ 296, 411, 59 },
		{ 296, 412, 59 },
		{ 297, 338, 59 },
		{ 297, 339, 59 },
		{ 297, 340, 59 },
		{ 297, 341, 59 },
		{ 297, 342, 59 },
		{ 297, 343, 59 },
		{ 297, 344, 59 },
		{ 297, 345, 59 },
		{ 297, 346, 59 },
		{ 297, 347, 59 },
		{ 297, 348, 59 },
		{ 297, 349, 59 },
		{ 297, 350, 59 },
		{ 297, 351, 59 },
		{ 297, 352, 59 },
		{ 297, 353, 59 },
		{ 297, 354, 59 },
		{ 297, 355, 59 },
		{ 297, 356, 59 },
		{ 297, 357, 59 },
		{ 297, 358, 59 },
		{ 297, 359, 59 },
		{ 297, 360, 59 },
		{ 297, 361, 59 },
		{ 297, 362, 59 },
		{ 297, 363, 59 },
		{ 297, 364, 59 },
		{ 297, 365, 59 },
		{ 297, 366, 59 },
		{ 297, 367, 59 },
		{ 297, 368, 59 },
		{ 297, 369, 59 },
		{ 297, 370, 59 },
		{ 297, 371, 59 },
		{ 297, 372, 59 },
		{ 297, 373, 59 },
		{ 297, 374, 59 },
		{ 297, 375, 59 },
		{ 297, 376, 59 },
		{ 297, 377, 59 },
		{ 297, 378, 59 },
		{ 297, 379, 59 },
		{ 297, 380, 59 },
		{ 297, 381, 59 },
		{ 297, 382, 59 },
		{ 297, 383, 59 },
		{ 297, 384, 59 },
		{ 297, 385, 59 },
		{ 297, 386, 59 },
		{ 297, 387, 59 },
		{ 297, 388, 59 },
		{ 297, 389, 59 },
		{ 297, 390, 59 },
		{ 297, 391, 59 },
		{ 297, 392, 59 },
		{ 297, 393, 59 },
		{ 297, 394, 59 },
		{ 297, 395, 59 },
		{ 297, 396, 59 },
		{ 297, 397, 59 },
		{ 297, 398, 59 },
		{ 297, 399, 59 },
		{ 297, 400, 59 },
		{ 297, 401, 59 },
		{ 297, 402, 59 },
		{ 297, 403, 59 },
		{ 297, 404, 59 },
		{ 297, 405, 59 },
		{ 297, 406, 59 },
		{ 297, 407, 59 },
		{ 297, 408, 59 },
		{ 297, 409, 59 },
		{ 297, 410, 59 },
		{ 297, 411, 59 },
		{ 297, 412, 59 },
		{ 297, 413, 59 },
		{ 297, 414, 59 },
		{ 297, 415, 59 },
		{ 297, 416, 59 },
		{ 298, 340, 59 },
		{ 298, 341, 59 },
		{ 298, 342, 59 },
		{ 298, 343, 59 },
		{ 298, 344, 59 },
		{ 298, 345, 59 },
		{ 298, 346, 59 },
		{ 298, 347, 59 },
		{ 298, 348, 59 },
		{ 298, 349, 59 },
		{ 298, 350, 59 },
		{ 298, 351, 59 },
		{ 298, 352, 59 },
		{ 298, 353, 59 },
		{ 298, 354, 59 },
		{ 298, 355, 59 },
		{ 298, 356, 59 },
		{ 298, 357, 59 },
		{ 298, 358, 59 },
		{ 298, 359, 59 },
		{ 298, 360, 59 },
		{ 298, 361, 59 },
		{ 298, 362, 59 },
		{ 298, 363, 59 },
		{ 298, 364, 59 },
		{ 298, 365, 59 },
		{ 298, 366, 59 },
		{ 298, 367, 59 },
		{ 298, 368, 59 },
		{ 298, 369, 59 },
		{ 298, 370, 59 },
		{ 298, 371, 59 },
		{ 298, 372, 59 },
		{ 298, 373, 59 },
		{ 298, 374, 59 },
		{ 298, 375, 59 },
		{ 298, 376, 59 },
		{ 298, 377, 59 },
		{ 298, 378, 59 },
		{ 298, 379, 59 },
		{ 298, 380, 59 },
		{ 298, 381, 59 },
		{ 298, 382, 59 },
		{ 298, 383, 59 },
		{ 298, 384, 59 },
		{ 298, 385, 59 },
		{ 298, 386, 59 },
		{ 298, 387, 59 },
		{ 298, 388, 59 },
		{ 298, 389, 59 },
		{ 298, 390, 59 },
		{ 298, 391, 59 },
		{ 298, 392, 59 },
		{ 298, 393, 59 },
		{ 298, 394, 59 },
		{ 298, 395, 59 },
		{ 298, 396, 59 },
		{ 298, 397, 59 },
		{ 298, 398, 59 },
		{ 298, 399, 59 },
		{ 298, 400, 59 },
		{ 298, 401, 59 },
		{ 298, 402, 59 },
		{ 298, 403, 59 },
		{ 298, 404, 59 },
		{ 298, 405, 59 },
		{ 298, 406, 59 },
		{ 298, 407, 59 },
		{ 298, 408, 59 },
		{ 298, 409, 59 },
		{ 298, 410, 59 },
		{ 298, 411, 59 },
		{ 298, 412, 59 },
		{ 298, 413, 59 },
		{ 298, 414, 59 },
		{ 298, 415, 59 },
		{ 298, 416, 59 },
		{ 298, 417, 59 },
		{ 298, 418, 59 },
		{ 298, 419, 59 },
		{ 298, 420, 59 },
		{ 299, 342, 59 },
		{ 299, 343, 59 },
		{ 299, 344, 59 },
		{ 299, 345, 59 },
		{ 299, 346, 59 },
		{ 299, 347, 59 },
		{ 299, 348, 59 },
		{ 299, 349, 59 },
		{ 299, 350, 59 },
		{ 299, 351, 59 },
		{ 299, 352, 59 },
		{ 299, 353, 59 },
		{ 299, 354, 59 },
		{ 299, 355, 59 },
		{ 299, 356, 59 },
		{ 299, 357, 59 },
		{ 299, 358, 59 },
		{ 299, 359, 59 },
		{ 299, 360, 59 },
		{ 299, 361, 59 },
		{ 299, 362, 59 },
		{ 299, 363, 59 },
		{ 299, 364, 59 },
		{ 299, 365, 59 },
		{ 299, 366, 59 },
		{ 299, 367, 59 },
		{ 299, 368, 59 },
		{ 299, 369, 59 },
		{ 299, 370, 59 },
		{ 299, 371, 59 },
		{ 299, 372, 59 },
		{ 299, 373, 59 },
		{ 299, 374, 59 },
		{ 299, 375, 59 },
		{ 299, 376, 59 },
		{ 299, 377, 59 },
		{ 299, 378, 59 },
		{ 299, 379, 59 },
		{ 299, 380, 59 },
		{ 299, 381, 59 },
		{ 299, 382, 59 },
		{ 299, 383, 59 },
		{ 299, 384, 59 },
		{ 299, 385, 59 },
		{ 299, 386, 59 },
		{ 299, 387, 59 },
		{ 299, 388, 59 },
		{ 299, 389, 59 },
		{ 299, 390, 59 },
		{ 299, 391, 59 },
		{ 299, 392, 59 },
		{ 299, 393, 59 },
		{ 299, 394, 59 },
		{ 299, 395, 59 },
		{ 299, 396, 59 },
		{ 299, 397, 59 },
		{ 299, 398, 59 },
		{ 299, 399, 59 },
		{ 299, 400, 59 },
		{ 299, 401, 59 },
		{ 299, 402, 59 },
		{ 299, 403, 59 },
		{ 299, 404, 59 },
		{ 299, 405, 59 },
		{ 299, 406, 59 },
		{ 299, 407, 59 },
		{ 299, 408, 59 },
		{ 299, 409, 59 },
		{ 299, 410, 59 },
		{ 299, 411, 59 },
		{ 299, 412, 59 },
		{ 299, 413, 59 },
		{ 299, 414, 59 },
		{ 299, 415, 59 },
		{ 299, 416, 59 },
		{ 299, 417, 59 },
		{ 299, 418, 59 },
		{ 299, 419, 59 },
		{ 299, 420, 59 },
		{ 299, 421, 59 },
		{ 299, 422, 59 },
		{ 299, 423, 59 },
		{ 299, 424, 59 },
		{ 300, 344, 59 },
		{ 300, 345, 59 },
		{ 300, 346, 59 },
		{ 300, 347, 59 },
		{ 300, 348, 59 },
		{ 300, 349, 59 },
		{ 300, 350, 59 },
		{ 300, 351, 59 },
		{ 300, 352, 59 },
		{ 300, 353, 59 },
		{ 300, 354, 59 },
		{ 300, 355, 59 },
		{ 300, 356, 59 },
		{ 300, 357, 59 },
		{ 300, 358, 59 },
		{ 300, 359, 59 },
		{ 300, 360, 59 },
		{ 300, 361, 59 },
		{ 300, 362, 59 },
		{ 300, 363, 59 },
		{ 300, 364, 59 },
		{ 300, 365, 59 },
		{ 300, 366, 59 },
		{ 300, 367, 59 },
		{ 300, 368, 59 },
		{ 300, 369, 59 },
		{ 300, 370, 59 },
		{ 300, 371, 59 },
		{ 300, 372, 59 },
		{ 300, 373, 59 },
		{ 300, 374, 59 },
		{ 300, 375, 59 },
		{ 300, 376, 59 },
		{ 300, 377, 59 },
		{ 300, 378, 59 },
		{ 300, 379, 59 },
		{ 300, 380, 59 },
		{ 300, 381, 59 },
		{ 300, 382, 59 },
		{ 300, 383, 59 },
		{ 300, 384, 59 },
		{ 300, 385, 59 },
		{ 300, 386, 59 },
		{ 300, 387, 59 },
		{ 300, 388, 59 },
		{ 300, 389, 59 },
		{ 300, 390, 59 },
		{ 300, 391, 59 },
		{ 300, 392, 59 },
		{ 300, 393, 59 },
		{ 300, 394, 59 },
		{ 300, 395, 59 },
		{ 300, 396, 59 },
		{ 300, 397, 59 },
		{ 300, 398, 59 },
		{ 300, 399, 59 },
		{ 300, 400, 59 },
		{ 300, 401, 59 },
		{ 300, 402, 59 },
		{ 300, 403, 59 },
		{ 300, 404, 59 },
		{ 300, 405, 59 },
		{ 300, 406, 59 },
		{ 300, 407, 59 },
		{ 300, 408, 59 },
		{ 300, 409, 59 },
		{ 300, 410, 59 },
		{ 300, 411, 59 },
		{ 300, 412, 59 },
		{ 300, 413, 59 },
		{ 300, 414, 59 },
		{ 300, 415, 59 },
		{ 300, 416, 59 },
		{ 300, 417, 59 },
		{ 300, 418, 59 },
		{ 300, 419, 59 },
		{ 300, 420, 59 },
		{ 300, 421, 59 },
		{ 300, 422, 59 },
		{ 300, 423, 59 },
		{ 300, 424, 59 },
		{ 300, 425, 59 },
		{ 300, 426, 59 },
		{ 300, 427, 59 },
		{ 300, 428, 59 },
		{ 301, 346, 59 },
		{ 301, 347, 59 },
		{ 301, 348, 59 },
		{ 301, 349, 59 },
		{ 301, 350, 59 },
		{ 301, 351, 59 },
		{ 301, 352, 59 },
		{ 301, 353, 59 },
		{ 301, 354, 59 },
		{ 301, 355, 59 },
		{ 301, 356, 59 },
		{ 301, 357, 59 },
		{ 301, 358, 59 },
		{ 301, 359, 59 },
		{ 301, 360, 59 },
		{ 301, 361, 59 },
		{ 301, 362, 59 },
		{ 301, 363, 59 },
		{ 301, 364, 59 },
		{ 301, 365, 59 },
		{ 301, 366, 59 },
		{ 301, 367, 59 },
		{ 301, 368, 59 },
		{ 301, 369, 59 },
		{ 301, 370, 59 },
		{ 301, 371, 59 },
		{ 301, 372, 59 },
		{ 301, 373, 59 },
		{ 301, 374, 59 },
		{ 301, 375, 59 },
		{ 301, 376, 59 },
		{ 301, 377, 59 },
		{ 301, 378, 59 },
		{ 301, 379, 59 },
		{ 301, 380, 59 },
		{ 301, 381, 59 },
		{ 301, 382, 59 },
		{ 301, 383, 59 },
		{ 301, 384, 59 },
		{ 301, 385, 59 },
		{ 301, 386, 59 },
		{ 301, 387, 59 },
		{ 301, 388, 59 },
		{ 301, 389, 59 },
		{ 301, 390, 59 },
		{ 301, 391, 59 },
		{ 301, 392, 59 },
		{ 301, 393, 59 },
		{ 301, 394, 59 },
		{ 301, 395, 59 },
		{ 301, 396, 59 },
		{ 301, 397, 59 },
		{ 301, 398, 59 },
		{ 301, 399, 59 },
		{ 301, 400, 59 },
		{ 301, 401, 59 },
		{ 301, 402, 59 },
		{ 301, 403, 59 },
		{ 301, 404, 59 },
		{ 301, 405, 59 },
		{ 301, 406, 59 },
		{ 301, 407, 59 },
		{ 301, 408, 59 },
		{ 301, 409, 59 },
		{ 301, 410, 59 },
		{ 301, 411, 59 },
		{ 301, 412, 59 },
		{ 301, 413, 59 },
		{ 301, 414, 59 },
		{ 301, 415, 59 },
		{ 301, 416, 59 },
		{ 301, 417, 59 },
		{ 301, 418, 59 },
		{ 301, 419, 59 },
		{ 301, 420, 59 },
		{ 301, 421, 59 },
		{ 301, 422, 59 },
		{ 301, 423, 59 },
		{ 301, 424, 59 },
		{ 301, 425, 59 },
		{ 301, 426, 59 },
		{ 301, 427, 59 },
		{ 301, 428, 59 },
		{ 301, 429, 59 },
		{ 301, 430, 59 },
		{ 301, 431, 59 },
		{ 301, 432, 59 },
		{ 302, 348, 59 },
		{ 302, 349, 59 },
		{ 302, 350, 59 },
		{ 302, 351, 59 },
		{ 302, 352, 59 },
		{ 302, 353, 59 },
		{ 302, 354, 59 },
		{ 302, 355, 59 },
		{ 302, 356, 59 },
		{ 302, 357, 59 },
		{ 302, 358, 59 },
		{ 302, 359, 59 },
		{ 302, 360, 59 },
		{ 302, 361, 59 },
		{ 302, 362, 59 },
		{ 302, 363, 59 },
		{ 302, 364, 59 },
		{ 302, 365, 59 },
		{ 302, 366, 59 },
		{ 302, 367, 59 },
		{ 302, 368, 59 },
		{ 302, 369, 59 },
		{ 302, 370, 59 },
		{ 302, 371, 59 },
		{ 302, 372, 59 },
		{ 302, 373, 59 },
		{ 302, 374, 59 },
		{ 302, 375, 59 },
		{ 302, 376, 59 },
		{ 302, 377, 59 },
		{ 302, 378, 59 },
		{ 302, 379, 59 },
		{ 302, 380, 59 },
		{ 302, 381, 59 },
		{ 302, 382, 59 },
		{ 302, 383, 59 },
		{ 302, 384, 59 },
		{ 302, 385, 59 },
		{ 302, 386, 59 },
		{ 302, 387, 59 },
		{ 302, 388, 59 },
		{ 302, 389, 59 },
		{ 302, 390, 59 },
		{ 302, 391, 59 },
		{ 302, 392, 59 },
		{ 302, 393, 59 },
		{ 302, 394, 59 },
		{ 302, 395, 59 },
		{ 302, 396, 59 },
		{ 302, 397, 59 },
		{ 302, 398, 59 },
		{ 302, 399, 59 },
		{ 302, 400, 59 },
		{ 302, 401, 59 },
		{ 302, 402, 59 },
		{ 302, 403, 59 },
		{ 302, 404, 59 },
		{ 302, 405, 59 },
		{ 302, 406, 59 },
		{ 302, 407, 59 },
		{ 302, 408, 59 },
		{ 302, 409, 59 },
		{ 302, 410, 59 },
		{ 302, 411, 59 },
		{ 302, 412, 59 },
		{ 302, 413, 59 },
		{ 302, 414, 59 },
		{ 302, 415, 59 },
		{ 302, 416, 59 },
		{ 302, 417, 59 },
		{ 302, 418, 59 },
		{ 302, 419, 59 },
		{ 302, 420, 59 },
		{ 302, 421, 59 },
		{ 302, 422, 59 },
		{ 302, 423, 59 },
		{ 302, 424, 59 },
		{ 302, 425, 59 },
		{ 302, 426, 59 },
		{ 302, 427, 59 },
		{ 302, 428, 59 },
		{ 302, 429, 59 },
		{ 302, 430, 59 },
		{ 302, 431, 59 },
		{ 302, 432, 59 },
		{ 302, 433, 59 },
		{ 302, 434, 59 },
		{ 302, 435, 59 },
		{ 302, 436, 59 },
		{ 303, 350, 59 },
		{ 303, 351, 59 },
		{ 303, 352, 59 },
		{ 303, 353, 59 },
		{ 303, 354, 59 },
		{ 303, 355, 59 },
		{ 303, 356, 59 },
		{ 303, 357, 59 },
		{ 303, 358, 59 },
		{ 303, 359, 59 },
		{ 303, 360, 59 },
		{ 303, 361, 59 },
		{ 303, 362, 59 },
		{ 303, 363, 59 },
		{ 303, 364, 59 },
		{ 303, 365, 59 },
		{ 303, 366, 59 },
		{ 303, 367, 59 },
		{ 303, 368, 59 },
		{ 303, 369, 59 },
		{ 303, 370, 59 },
		{ 303, 371, 59 },
		{ 303, 372, 59 },
		{ 303, 373, 59 },
		{ 303, 374, 59 },
		{ 303, 375, 59 },
		{ 303, 376, 59 },
		{ 303, 377, 59 },
		{ 303, 378, 59 },
		{ 303, 379, 59 },
		{ 303, 380, 59 },
		{ 303, 381, 59 },
		{ 303, 382, 59 },
		{ 303, 383, 59 },
		{ 303, 384, 59 },
		{ 303, 385, 59 },
		{ 303, 386, 59 },
		{ 303, 387, 59 },
		{ 303, 388, 59 },
		{ 303, 389, 59 },
		{ 303, 390, 59 },
		{ 303, 391, 59 },
		{ 303, 392, 59 },
		{ 303, 393, 59 },
		{ 303, 394, 59 },
		{ 303, 395, 59 },
		{ 303, 396, 59 },
		{ 303, 397, 59 },
		{ 303, 398, 59 },
		{ 303, 399, 59 },
		{ 303, 400, 59 },
		{ 303, 401, 59 },
		{ 303, 402, 59 },
		{ 303, 403, 59 },
		{ 303, 404, 59 },
		{ 303, 405, 59 },
		{ 303, 406, 59 },
		{ 303, 407, 59 },
		{ 303, 408, 59 },
		{ 303, 409, 59 },
		{ 303, 410, 59 },
		{ 303, 411, 59 },
		{ 303, 412, 59 },
		{ 303, 413, 59 },
		{ 303, 414, 59 },
		{ 303, 415, 59 },
		{ 303, 416, 59 },
		{ 303, 417, 59 },
		{ 303, 418, 59 },
		{ 303, 419, 59 },
		{ 303, 420, 59 },
		{ 303, 421, 59 },
		{ 303, 422, 59 },
		{ 303, 423, 59 },
		{ 303, 424, 59 },
		{ 303, 425, 59 },
		{ 303, 426, 59 },
		{ 303, 427, 59 },
		{ 303, 428, 59 },
		{ 303, 429, 59 },
		{ 303, 430, 59 },
		{ 303, 431, 59 },
		{ 303, 432, 59 },
		{ 303, 433, 59 },
		{ 303, 434, 59 },
		{ 303, 435, 59 },
		{ 303, 436, 59 },
		{ 303, 437, 59 },
		{ 303, 438, 59 },
		{ 303, 439, 59 },
		{ 303, 440, 59 },
		{ 304, 352, 59 },
		{ 304, 353, 59 },
		{ 304, 354, 59 },
		{ 304, 355, 59 },
		{ 304, 356, 59 },
		{ 304, 357, 59 },
		{ 304, 358, 59 },
		{ 304, 359, 59 },
		{ 304, 360, 59 },
		{ 304, 361, 59 },
		{ 304, 362, 59 },
		{ 304, 363, 59 },
		{ 304, 364, 59 },
		{ 304, 365, 59 },
		{ 304, 366, 59 },
		{ 304, 367, 59 },
		{ 304, 368, 59 },
		{ 304, 369, 59 },
		{ 304, 370, 59 },
		{ 304, 371, 59 },
		{ 304, 372, 59 },
		{ 304, 373, 59 },
		{ 304, 374, 59 },
		{ 304, 375, 59 },
		{ 304, 376, 59 },
		{ 304, 377, 59 },
		{ 304, 378, 59 },
		{ 304, 379, 59 },
		{ 304, 380, 59 },
		{ 304, 381, 59 },
		{ 304, 382, 59 },
		{ 304, 383, 59 },
		{ 304, 384, 59 },
		{ 304, 385, 59 },
		{ 304, 386, 59 },
		{ 304, 387, 59 },
		{ 304, 388, 59 },
		{ 304, 389, 59 },
		{ 304, 390, 59 },
		{ 304, 391, 59 },
		{ 304, 392, 59 },
		{ 304, 393, 59 },
		{ 304, 394, 59 },
		{ 304, 395, 59 },
		{ 304, 396, 59 },
		{ 304, 397, 59 },
		{ 304, 398, 59 },
		{ 304, 399, 59 },
		{ 304, 400, 59 },
		{ 304, 401, 59 },
		{ 304, 402, 59 },
		{ 304, 403, 59 },
		{ 304, 404, 59 },
		{ 304, 405, 59 },
		{ 304, 406, 59 },
		{ 304, 407, 59 },
		{ 304, 408, 59 },
		{ 304, 409, 59 },
		{ 304, 410, 59 },
		{ 304, 411, 59 },
		{ 304, 412, 59 },
		{ 304, 413, 59 },
		{ 304, 414, 59 },
		{ 304, 415, 59 },
		{ 304, 416, 59 },
		{ 304, 417, 59 },
		{ 304, 418, 59 },
		{ 304, 419, 59 },
		{ 304, 420, 59 },
		{ 304, 421, 59 },
		{ 304, 422, 59 },
		{ 304, 423, 59 },
		{ 304, 424, 59 },
		{ 304, 425, 59 },
		{ 304, 426, 59 },
		{ 304, 427, 59 },
		{ 304, 428, 59 },
		{ 304, 429, 59 },
		{ 304, 430, 59 },
		{ 304, 431, 59 },
		{ 304, 432, 59 },
		{ 304, 433, 59 },
		{ 304, 434, 59 },
		{ 304, 435, 59 },
		{ 304, 436, 59 },
		{ 304, 437, 59 },
		{ 304, 438, 59 },
		{ 304, 439, 59 },
		{ 304, 440, 59 },
		{ 304, 441, 59 },
		{ 304, 442, 59 },
		{ 304, 443, 59 },
		{ 304, 444, 59 },
		{ 305, 354, 59 },
		{ 305, 355, 59 },
		{ 305, 356, 59 },
		{ 305, 357, 59 },
		{ 305, 358, 59 },
		{ 305, 359, 59 },
		{ 305, 360, 59 },
		{ 305, 361, 59 },
		{ 305, 362, 59 },
		{ 305, 363, 59 },
		{ 305, 364, 59 },
		{ 305, 365, 59 },
		{ 305, 366, 59 },
		{ 305, 367, 59 },
		{ 305, 368, 59 },
		{ 305, 369, 59 },
		{ 305, 370, 59 },
		{ 305, 371, 59 },
		{ 305, 372, 59 },
		{ 305, 373, 59 },
		{ 305, 374, 59 },
		{ 305, 375, 59 },
		{ 305, 376, 59 },
		{ 305, 377, 59 },
		{ 305, 378, 59 },
		{ 305, 379, 59 },
		{ 305, 380, 59 },
		{ 305, 381, 59 },
		{ 305, 382, 59 },
		{ 305, 383, 59 },
		{ 305, 384, 59 },
		{ 305, 385, 59 },
		{ 305, 386, 59 },
		{ 305, 387, 59 },
		{ 305, 388, 59 },
		{ 305, 389, 59 },
		{ 305, 390, 59 },
		{ 305, 391, 59 },
		{ 305, 392, 59 },
		{ 305, 393, 59 },
		{ 305, 394, 59 },
		{ 305, 395, 59 },
		{ 305, 396, 59 },
		{ 305, 397, 59 },
		{ 305, 398, 59 },
		{ 305, 399, 59 },
		{ 305, 400, 59 },
		{ 305, 401, 59 },
		{ 305, 402, 59 },
		{ 305, 403, 59 },
		{ 305, 404, 59 },
		{ 305, 405, 59 },
		{ 305, 406, 59 },
		{ 305, 407, 59 },
		{ 305, 408, 59 },
		{ 305, 409, 59 },
		{ 305, 410, 59 },
		{ 305, 411, 59 },
		{ 305, 412, 59 },
		{ 305, 413, 59 },
		{ 305, 414, 59 },
		{ 305, 415, 59 },
		{ 305, 416, 59 },
		{ 305, 417, 59 },
		{ 305, 418, 59 },
		{ 305, 419, 59 },
		{ 305, 420, 59 },
		{ 305, 421, 59 },
		{ 305, 422, 59 },
		{ 305, 423, 59 },
		{ 305, 424, 59 },
		{ 305, 425, 59 },
		{ 305, 426, 59 },
		{ 305, 427, 59 },
		{ 305, 428, 59 },
		{ 305, 429, 59 },
		{ 305, 430, 59 },
		{ 305, 431, 59 },
		{ 305, 432, 59 },
		{ 305, 433, 59 },
		{ 305, 434, 59 },
		{ 305, 435, 59 },
		{ 305, 436, 59 },
		{ 305, 437, 59 },
		{ 305, 438, 59 },
		{ 305, 439, 59 },
		{ 305, 440, 59 },
		{ 305, 441, 59 },
		{ 305, 442, 59 },
		{ 305, 443, 59 },
		{ 305, 444, 59 },
		{ 305, 445, 59 },
		{ 305, 446, 59 },
		{ 305, 447, 59 },
		{ 305, 448, 59 },
		{ 306, 356, 59 },
		{ 306, 357, 59 },
		{ 306, 358, 59 },
		{ 306, 359, 59 },
		{ 306, 360, 59 },
		{ 306, 361, 59 },
		{ 306, 362, 59 },
		{ 306, 363, 59 },
		{ 306, 364, 59 },
		{ 306, 365, 59 },
		{ 306, 366, 59 },
		{ 306, 367, 59 },
		{ 306, 368, 59 },
		{ 306, 369, 59 },
		{ 306, 370, 59 },
		{ 306, 371, 59 },
		{ 306, 372, 59 },
		{ 306, 373, 59 },
		{ 306, 374, 59 },
		{ 306, 375, 59 },
		{ 306, 376, 59 },
		{ 306, 377, 59 },
		{ 306, 378, 59 },
		{ 306, 379, 59 },
		{ 306, 380, 59 },
		{ 306, 381, 59 },
		{ 306, 382, 59 },
		{ 306, 383, 59 },
		{ 306, 384, 59 },
		{ 306, 385, 59 },
		{ 306, 386, 59 },
		{ 306, 387, 59 },
		{ 306, 388, 59 },
		{ 306, 389, 59 },
		{ 306, 390, 59 },
		{ 306, 391, 59 },
		{ 306, 392, 59 },
		{ 306, 393, 59 },
		{ 306, 394, 59 },
		{ 306, 395, 59 },
		{ 306, 396, 59 },
		{ 306, 397, 59 },
		{ 306, 398, 59 },
		{ 306, 399, 59 },
		{ 306, 400, 59 },
		{ 306, 401, 59 },
		{ 306, 402, 59 },
		{ 306, 403, 59 },
		{ 306, 404, 59 },
		{ 306, 405, 59 },
		{ 306, 406, 59 },
		{ 306, 407, 59 },
		{ 306, 408, 59 },
		{ 306, 409, 59 },
		{ 306, 410, 59 },
		{ 306, 411, 59 },
		{ 306, 412, 59 },
		{ 306, 413, 59 },
		{ 306, 414, 59 },
		{ 306, 415, 59 },
		{ 306, 416, 59 },
		{ 306, 417, 59 },
		{ 306, 418, 59 },
		{ 306, 419, 59 },
		{ 306, 420, 59 },
		{ 306, 421, 59 },
		{ 306, 422, 59 },
		{ 306, 423, 59 },
		{ 306, 424, 59 },
		{ 306, 425, 59 },
		{ 306, 426, 59 },
		{ 306, 427, 59 },
		{ 306, 428, 59 },
		{ 306, 429, 59 },
		{ 306, 430, 59 },
		{ 306, 431, 59 },
		{ 306, 432, 59 },
		{ 306, 433, 59 },
		{ 306, 434, 59 },
		{ 306, 435, 59 },
		{ 306, 436, 59 },
		{ 306, 437, 59 },
		{ 306, 438, 59 },
		{ 306, 439, 59 },
		{ 306, 440, 59 },
		{ 306, 441, 59 },
		{ 306, 442, 59 },
		{ 306, 443, 59 },
		{ 306, 444, 59 },
		{ 306, 445, 59 },
		{ 306, 446, 59 },
		{ 306, 447, 59 },
		{ 306, 448, 59 },
		{ 306, 449, 59 },
		{ 306, 450, 59 },
		{ 306, 451, 59 },
		{ 306, 452, 59 },
		{ 307, 358, 59 },
		{ 307, 359, 59 },
		{ 307, 360, 59 },
		{ 307, 361, 59 },
		{ 307, 362, 59 },
		{ 307, 363, 59 },
		{ 307, 364, 59 },
		{ 307, 365, 59 },
		{ 307, 366, 59 },
		{ 307, 367, 59 },
		{ 307, 368, 59 },
		{ 307, 369, 59 },
		{ 307, 370, 59 },
		{ 307, 371, 59 },
		{ 307, 372, 59 },
		{ 307, 373, 59 },
		{ 307, 374, 59 },
		{ 307, 375, 59 },
		{ 307, 376, 59 },
		{ 307, 377, 59 },
		{ 307, 378, 59 },
		{ 307, 379, 59 },
		{ 307, 380, 59 },
		{ 307, 381, 59 },
		{ 307, 382, 59 },
		{ 307, 383, 59 },
		{ 307, 384, 59 },
		{ 307, 385, 59 },
		{ 307, 386, 59 },
		{ 307, 387, 59 },
		{ 307, 388, 59 },
		{ 307, 389, 59 },
		{ 307, 390, 59 },
		{ 307, 391, 59 },
		{ 307, 392, 59 },
		{ 307, 393, 59 },
		{ 307, 394, 59 },
		{ 307, 395, 59 },
		{ 307, 396, 59 },
		{ 307, 397, 59 },
		{ 307, 398, 59 },
		{ 307, 399, 59 },
		{ 307, 400, 59 },
		{ 307, 401, 59 },
		{ 307, 402, 59 },
		{ 307, 403, 59 },
		{ 307, 404, 59 },
		{ 307, 405, 59 },
		{ 307, 406, 59 },
		{ 307, 407, 59 },
		{ 307, 408, 59 },
		{ 307, 409, 59 },
		{ 307, 410, 59 },
		{ 307, 411, 59 },
		{ 307, 412, 59 },
		{ 307, 413, 59 },
		{ 307, 414, 59 },
		{ 307, 415, 59 },
		{ 307, 416, 59 },
		{ 307, 417, 59 },
		{ 307, 418, 59 },
		{ 307, 419, 59 },
		{ 307, 420, 59 },
		{ 307, 421, 59 },
		{ 307, 422, 59 },
		{ 307, 423, 59 },
		{ 307, 424, 59 },
		{ 307, 425, 59 },
		{ 307, 426, 59 },
		{ 307, 427, 59 },
		{ 307, 428, 59 },
		{ 307, 429, 59 },
		{ 307, 430, 59 },
		{ 307, 431, 59 },
		{ 307, 432, 59 },
		{ 307, 433, 59 },
		{ 307, 434, 59 },
		{ 307, 435, 59 },
		{ 307, 436, 59 },
		{ 307, 437, 59 },
		{ 307, 438, 59 },
		{ 307, 439, 59 },
		{ 307, 440, 59 },
		{ 307, 441, 59 },
		{ 307, 442, 59 },
		{ 307, 443, 59 },
		{ 307, 444, 59 },
		{ 307, 445, 59 },
		{ 307, 446, 59 },
		{ 307, 447, 59 },
		{ 307, 448, 59 },
		{ 307, 449, 59 },
		{ 307, 450, 59 },
		{ 307, 451, 59 },
		{ 307, 452, 59 },
		{ 307, 453, 59 },
		{ 307, 454, 59 },
		{ 307, 455, 59 },
		{ 307, 456, 59 },
		{ 308, 360, 59 },
		{ 308, 361, 59 },
		{ 308, 362, 59 },
		{ 308, 363, 59 },
		{ 308, 364, 59 },
		{ 308, 365, 59 },
		{ 308, 366, 59 },
		{ 308, 367, 59 },
		{ 308, 368, 59 },
		{ 308, 369, 59 },
		{ 308, 370, 59 },
		{ 308, 371, 59 },
		{ 308, 372, 59 },
		{ 308, 373, 59 },
		{ 308, 374, 59 },
		{ 308, 375, 59 },
		{ 308, 376, 59 },
		{ 308, 377, 59 },
		{ 308, 378, 59 },
		{ 308, 379, 59 },
		{ 308, 380, 59 },
		{ 308, 381, 59 },
		{ 308, 382, 59 },
		{ 308, 383, 59 },
		{ 308, 384, 59 },
		{ 308, 385, 59 },
		{ 308, 386, 59 },
		{ 308, 387, 59 },
		{ 308, 388, 59 },
		{ 308, 389, 59 },
		{ 308, 390, 59 },
		{ 308, 391, 59 },
		{ 308, 392, 59 },
		{ 308, 393, 59 },
		{ 308, 394, 59 },
		{ 308, 395, 59 },
		{ 308, 396, 59 },
		{ 308, 397, 59 },
		{ 308, 398, 59 },
		{ 308, 399, 59 },
		{ 308, 400, 59 },
		{ 308, 401, 59 },
		{ 308, 402, 59 },
		{ 308, 403, 59 },
		{ 308, 404, 59 },
		{ 308, 405, 59 },
		{ 308, 406, 59 },
		{ 308, 407, 59 },
		{ 308, 408, 59 },
		{ 308, 409, 59 },
		{ 308, 410, 59 },
		{ 308, 411, 59 },
		{ 308, 412, 59 },
		{ 308, 413, 59 },
		{ 308, 414, 59 },
		{ 308, 415, 59 },
		{ 308, 416, 59 },
		{ 308, 417, 59 },
		{ 308, 418, 59 },
		{ 308, 419, 59 },
		{ 308, 420, 59 },
		{ 308, 421, 59 },
		{ 308, 422, 59 },
		{ 308, 423, 59 },
		{ 308, 424, 59 },
		{ 308, 425, 59 },
		{ 308, 426, 59 },
		{ 308, 427, 59 },
		{ 308, 428, 59 },
		{ 308, 429, 59 },
		{ 308, 430, 59 },
		{ 308, 431, 59 },
		{ 308, 432, 59 },
		{ 308, 433, 59 },
		{ 308, 434, 59 },
		{ 308, 435, 59 },
		{ 308, 436, 59 },
		{ 308, 437, 59 },
		{ 308, 438, 59 },
		{ 308, 439, 59 },
		{ 308, 440, 59 },
		{ 308, 441, 59 },
		{ 308, 442, 59 },
		{ 308, 443, 59 },
		{ 308, 444, 59 },
		{ 308, 445, 59 },
		{ 308, 446, 59 },
		{ 308, 447, 59 },
		{ 308, 448, 59 },
		{ 308, 449, 59 },
		{ 308, 450, 59 },
		{ 308, 451, 59 },
		{ 308, 452, 59 },
		{ 308, 453, 59 },
		{ 308, 454, 59 },
		{ 308, 455, 59 },
		{ 308, 456, 59 },
		{ 308, 457, 59 },
		{ 308, 458, 59 },
		{ 308, 459, 59 },
		{ 308, 460, 59 },
		{ 309, 362, 59 },
		{ 309, 363, 59 },
		{ 309, 364, 59 },
		{ 309, 365, 59 },
		{ 309, 366, 59 },
		{ 309, 367, 59 },
		{ 309, 368, 59 },
		{ 309, 369, 59 },
		{ 309, 370, 59 },
		{ 309, 371, 59 },
		{ 309, 372, 59 },
		{ 309, 373, 59 },
		{ 309, 374, 59 },
		{ 309, 375, 59 },
		{ 309, 376, 59 },
		{ 309, 377, 59 },
		{ 309, 378, 59 },
		{ 309, 379, 59 },
		{ 309, 380, 59 },
		{ 309, 381, 59 },
		{ 309, 382, 59 },
		{ 309, 383, 59 },
		{ 309, 384, 59 },
		{ 309, 385, 59 },
		{ 309, 386, 59 },
		{ 309, 387, 59 },
		{ 309, 388, 59 },
		{ 309, 389, 59 },
		{ 309, 390, 59 },
		{ 309, 391, 59 },
		{ 309, 392, 59 },
		{ 309, 393, 59 },
		{ 309, 394, 59 },
		{ 309, 395, 59 },
		{ 309, 396, 59 },
		{ 309, 397, 59 },
		{ 309, 398, 59 },
		{ 309, 399, 59 },
		{ 309, 400, 59 },
		{ 309, 401, 59 },
		{ 309, 402, 59 },
		{ 309, 403, 59 },
		{ 309, 404, 59 },
		{ 309, 405, 59 },
		{ 309, 406, 59 },
		{ 309, 407, 59 },
		{ 309, 408, 59 },
		{ 309, 409, 59 },
		{ 309, 410, 59 },
		{ 309, 411, 59 },
		{ 309, 412, 59 },
		{ 309, 413, 59 },
		{ 309, 414, 59 },
		{ 309, 415, 59 },
		{ 309, 416, 59 },
		{ 309, 417, 59 },
		{ 309, 418, 59 },
		{ 309, 419, 59 },
		{ 309, 420, 59 },
		{ 309, 421, 59 },
		{ 309, 422, 59 },
		{ 309, 423, 59 },
		{ 309, 424, 59 },
		{ 309, 425, 59 },
		{ 309, 426, 59 },
		{ 309, 427, 59 },
		{ 309, 428, 59 },
		{ 309, 429, 59 },
		{ 309, 430, 59 },
		{ 309, 431, 59 },
		{ 309, 432, 59 },
		{ 309, 433, 59 },
		{ 309, 434, 59 },
		{ 309, 435, 59 },
		{ 309, 436, 59 },
		{ 309, 437, 59 },
		{ 309, 438, 59 },
		{ 309, 439, 59 },
		{ 309, 440, 59 },
		{ 309, 441, 59 },
		{ 309, 442, 59 },
		{ 309, 443, 59 },
		{ 309, 444, 59 },
		{ 309, 445, 59 },
		{ 309, 446, 59 },
		{ 309, 447, 59 },
		{ 309, 448, 59 },
		{ 309, 449, 59 },
		{ 309, 450, 59 },
		{ 309, 451, 59 },
		{ 309, 452, 59 },
		{ 309, 453, 59 },
		{ 309, 454, 59 },
		{ 309, 455, 59 },
		{ 309, 456, 59 },
		{ 309, 457, 59 },
		{ 309, 458, 59 },
		{ 309, 459, 59 },
		{ 309, 460, 59 },
		{ 309, 461, 59 },
		{ 309, 462, 59 },
		{ 309, 463, 59 },
		{ 309, 464, 59 },
		{ 310, 364, 59 },
		{ 310, 365, 59 },
		{ 310, 366, 59 },
		{ 310, 367, 59 },
		{ 310, 368, 59 },
		{ 310, 369, 59 },
		{ 310, 370, 59 },
		{ 310, 371, 59 },
		{ 310, 372, 59 },
		{ 310, 373, 59 },
		{ 310, 374, 59 },
		{ 310, 375, 59 },
		{ 310, 376, 59 },
		{ 310, 377, 59 },
		{ 310, 378, 59 },
		{ 310, 379, 59 },
		{ 310, 380, 59 },
		{ 310, 381, 59 },
		{ 310, 382, 59 },
		{ 310, 383, 59 },
		{ 310, 384, 59 },
		{ 310, 385, 59 },
		{ 310, 386, 59 },
		{ 310, 387, 59 },
		{ 310, 388, 59 },
		{ 310, 389, 59 },
		{ 310, 390, 59 },
		{ 310, 391, 59 },
		{ 310, 392, 59 },
		{ 310, 393, 59 },
		{ 310, 394, 59 },
		{ 310, 395, 59 },
		{ 310, 396, 59 },
		{ 310, 397, 59 },
		{ 310, 398, 59 },
		{ 310, 399, 59 },
		{ 310, 400, 59 },
		{ 310, 401, 59 },
		{ 310, 402, 59 },
		{ 310, 403, 59 },
		{ 310, 404, 59 },
		{ 310, 405, 59 },
		{ 310, 406, 59 },
		{ 310, 407, 59 },
		{ 310, 408, 59 },
		{ 310, 409, 59 },
		{ 310, 410, 59 },
		{ 310, 411, 59 },
		{ 310, 412, 59 },
		{ 310, 413, 59 },
		{ 310, 414, 59 },
		{ 310, 415, 59 },
		{ 310, 416, 59 },
		{ 310, 417, 59 },
		{ 310, 418, 59 },
		{ 310, 419, 59 },
		{ 310, 420, 59 },
		{ 310, 421, 59 },
		{ 310, 422, 59 },
		{ 310, 423, 59 },
		{ 310, 424, 59 },
		{ 310, 425, 59 },
		{ 310, 426, 59 },
		{ 310, 427, 59 },
		{ 310, 428, 59 },
		{ 310, 429, 59 },
		{ 310, 430, 59 },
		{ 310, 431, 59 },
		{ 310, 432, 59 },
		{ 310, 433, 59 },
		{ 310, 434, 59 },
		{ 310, 435, 59 },
		{ 310, 436, 59 },
		{ 310, 437, 59 },
		{ 310, 438, 59 },
		{ 310, 439, 59 },
		{ 310, 440, 59 },
		{ 310, 441, 59 },
		{ 310, 442, 59 },
		{ 310, 443, 59 },
		{ 310, 444, 59 },
		{ 310, 445, 59 },
		{ 310, 446, 59 },
		{ 310, 447, 59 },
		{ 310, 448, 59 },
		{ 310, 449, 59 },
		{ 310, 450, 59 },
		{ 310, 451, 59 },
		{ 310, 452, 59 },
		{ 310, 453, 59 },
		{ 310, 454, 59 },
		{ 310, 455, 59 },
		{ 310, 456, 59 },
		{ 310, 457, 59 },
		{ 310, 458, 59 },
		{ 310, 459, 59 },
		{ 310, 460, 59 },
		{ 310, 461, 59 },
		{ 310, 462, 59 },
		{ 310, 463, 59 },
		{ 310, 464, 59 },
		{ 310, 465, 59 },
		{ 310, 466, 59 },
		{ 310, 467, 59 },
		{ 310, 468, 59 },
		{ 311, 366, 59 },
		{ 311, 367, 59 },
		{ 311, 368, 59 },
		{ 311, 369, 59 },
		{ 311, 370, 59 },
		{ 311, 371, 59 },
		{ 311, 372, 59 },
		{ 311, 373, 59 },
		{ 311, 374, 59 },
		{ 311, 375, 59 },
		{ 311, 376, 59 },
		{ 311, 377, 59 },
		{ 311, 378, 59 },
		{ 311, 379, 59 },
		{ 311, 380, 59 },
		{ 311, 381, 59 },
		{ 311, 382, 59 },
		{ 311, 383, 59 },
		{ 311, 384, 59 },
		{ 311, 385, 59 },
		{ 311, 386, 59 },
		{ 311, 387, 59 },
		{ 311, 388, 59 },
		{ 311, 389, 59 },
		{ 311, 390, 59 },
		{ 311, 391, 59 },
		{ 311, 392, 59 },
		{ 311, 393, 59 },
		{ 311, 394, 59 },
		{ 311, 395, 59 },
		{ 311, 396, 59 },
		{ 311, 397, 59 },
		{ 311, 398, 59 },
		{ 311, 399, 59 },
		{ 311, 400, 59 },
		{ 311, 401, 59 },
		{ 311, 402, 59 },
		{ 311, 403, 59 },
		{ 311, 404, 59 },
		{ 311, 405, 59 },
		{ 311, 406, 59 },
		{ 311, 407, 59 },
		{ 311, 408, 59 },
		{ 311, 409, 59 },
		{ 311, 410, 59 },
		{ 311, 411, 59 },
		{ 311, 412, 59 },
		{ 311, 413, 59 },
		{ 311, 414, 59 },
		{ 311, 415, 59 },
		{ 311, 416, 59 },
		{ 311, 417, 59 },
		{ 311, 418, 59 },
		{ 311, 419, 59 },
		{ 311, 420, 59 },
		{ 311, 421, 59 },
		{ 311, 422, 59 },
		{ 311, 423, 59 },
		{ 311, 424, 59 },
		{ 311, 425, 59 },
		{ 311, 426, 59 },
		{ 311, 427, 59 },
		{ 311, 428, 59 },
		{ 311, 429, 59 },
		{ 311, 430, 59 },
		{ 311, 431, 59 },
		{ 311, 432, 59 },
		{ 311, 433, 59 },
		{ 311, 434, 59 },
		{ 311, 435, 59 },
		{ 311, 436, 59 },
		{ 311, 437, 59 },
		{ 311, 438, 59 },
		{ 311, 439, 59 },
		{ 311, 440, 59 },
		{ 311, 441, 59 },
		{ 311, 442, 59 },
		{ 311, 443, 59 },
		{ 311, 444, 59 },
		{ 311, 445, 59 },
		{ 311, 446, 59 },
		{ 311, 447, 59 },
		{ 311, 448, 59 },
		{ 311, 449, 59 },
		{ 311, 450, 59 },
		{ 311, 451, 59 },
		{ 311, 452, 59 },
		{ 311, 453, 59 },
		{ 311, 454, 59 },
		{ 311, 455, 59 },
		{ 311, 456, 59 },
		{ 311, 457, 59 },
		{ 311, 458, 59 },
		{ 311, 459, 59 },
		{ 311, 460, 59 },
		{ 311, 461, 59 },
		{ 311, 462, 59 },
		{ 311, 463, 59 },
		{ 311, 464, 59 },
		{ 311, 465, 59 },
		{ 311, 466, 59 },
		{ 311, 467, 59 },
		{ 311, 468, 59 },
		{ 311, 469, 59 },
		{ 311, 470, 59 },
		{ 311, 471, 59 },
		{ 311, 472, 59 },
		{ 312, 368, 59 },
		{ 312, 369, 59 },
		{ 312, 370, 59 },
		{ 312, 371, 59 },
		{ 312, 372, 59 },
		{ 312, 373, 59 },
		{ 312, 374, 59 },
		{ 312, 375, 59 },
		{ 312, 376, 59 },
		{ 312, 377, 59 },
		{ 312, 378, 59 },
		{ 312, 379, 59 },
		{ 312, 380, 59 },
		{ 312, 381, 59 },
		{ 312, 382, 59 },
		{ 312, 383, 59 },
		{ 312, 384, 59 },
		{ 312, 385, 59 },
		{ 312, 386, 59 },
		{ 312, 387, 59 },
		{ 312, 388, 59 },
		{ 312, 389, 59 },
		{ 312, 390, 59 },
		{ 312, 391, 59 },
		{ 312, 392, 59 },
		{ 312, 393, 59 },
		{ 312, 394, 59 },
		{ 312, 395, 59 },
		{ 312, 396, 59 },
		{ 312, 397, 59 },
		{ 312, 398, 59 },
		{ 312, 399, 59 },
		{ 312, 400, 59 },
		{ 312, 401, 59 },
		{ 312, 402, 59 },
		{ 312, 403, 59 },
		{ 312, 404, 59 },
		{ 312, 405, 59 },
		{ 312, 406, 59 },
		{ 312, 407, 59 },
		{ 312, 408, 59 },
		{ 312, 409, 59 },
		{ 312, 410, 59 },
		{ 312, 411, 59 },
		{ 312, 412, 59 },
		{ 312, 413, 59 },
		{ 312, 414, 59 },
		{ 312, 415, 59 },
		{ 312, 416, 59 },
		{ 312, 417, 59 },
		{ 312, 418, 59 },
		{ 312, 419, 59 },
		{ 312, 420, 59 },
		{ 312, 421, 59 },
		{ 312, 422, 59 },
		{ 312, 423, 59 },
		{ 312, 424, 59 },
		{ 312, 425, 59 },
		{ 312, 426, 59 },
		{ 312, 427, 59 },
		{ 312, 428, 59 },
		{ 312, 429, 59 },
		{ 312, 430, 59 },
		{ 312, 431, 59 },
		{ 312, 432, 59 },
		{ 312, 433, 59 },
		{ 312, 434, 59 },
		{ 312, 435, 59 },
		{ 312, 436, 59 },
		{ 312, 437, 59 },
		{ 312, 438, 59 },
		{ 312, 439, 59 },
		{ 312, 440, 59 },
		{ 312, 441, 59 },
		{ 312, 442, 59 },
		{ 312, 443, 59 },
		{ 312, 444, 59 },
		{ 312, 445, 59 },
		{ 312, 446, 59 },
		{ 312, 447, 59 },
		{ 312, 448, 59 },
		{ 312, 449, 59 },
		{ 312, 450, 59 },
		{ 312, 451, 59 },
		{ 312, 452, 59 },
		{ 312, 453, 59 },
		{ 312, 454, 59 },
		{ 312, 455, 59 },
		{ 312, 456, 59 },
		{ 312, 457, 59 },
		{ 312, 458, 59 },
		{ 312, 459, 59 },
		{ 312, 460, 59 },
		{ 312, 461, 59 },
		{ 312, 462, 59 },
		{ 312, 463, 59 },
		{ 312, 464, 59 },
		{ 312, 465, 59 },
		{ 312, 466, 59 },
		{ 312, 467, 59 },
		{ 312, 468, 59 },
		{ 312, 469, 59 },
		{ 312, 470, 59 },
		{ 312, 471, 59 },
		{ 312, 472, 59 },
		{ 312, 473, 59 },
		{ 312, 474, 59 },
		{ 312, 475, 59 },
		{ 312, 476, 59 },
		{ 313, 370, 59 },
		{ 313, 371, 59 },
		{ 313, 372, 59 },
		{ 313, 373, 59 },
		{ 313, 374, 59 },
		{ 313, 375, 59 },
		{ 313, 376, 59 },
		{ 313, 377, 59 },
		{ 313, 378, 59 },
		{ 313, 379, 59 },
		{ 313, 380, 59 },
		{ 313, 381, 59 },
		{ 313, 382, 59 },
		{ 313, 383, 59 },
		{ 313, 384, 59 },
		{ 313, 385, 59 },
		{ 313, 386, 59 },
		{ 313, 387, 59 },
		{ 313, 388, 59 },
		{ 313, 389, 59 },
		{ 313, 390, 59 },
		{ 313, 391, 59 },
		{ 313, 392, 59 },
		{ 313, 393, 59 },
		{ 313, 394, 59 },
		{ 313, 395, 59 },
		{ 313, 396, 59 },
		{ 313, 397, 59 },
		{ 313, 398, 59 },
		{ 313, 399, 59 },
		{ 313, 400, 59 },
		{ 313, 401, 59 },
		{ 313, 402, 59 },
		{ 313, 403, 59 },
		{ 313, 404, 59 },
		{ 313, 405, 59 },
		{ 313, 406, 59 },
		{ 313, 407, 59 },
		{ 313, 408, 59 },
		{ 313, 409, 59 },
		{ 313, 410, 59 },
		{ 313, 411, 59 },
		{ 313, 412, 59 },
		{ 313, 413, 59 },
		{ 313, 414, 59 },
		{ 313, 415, 59 },
		{ 313, 416, 59 },
		{ 313, 417, 59 },
		{ 313, 418, 59 },
		{ 313, 419, 59 },
		{ 313, 420, 59 },
		{ 313, 421, 59 },
		{ 313, 422, 59 },
		{ 313, 423, 59 },
		{ 313, 424, 59 },
		{ 313, 425, 59 },
		{ 313, 426, 59 },
		{ 313, 427, 59 },
		{ 313, 428, 59 },
		{ 313, 429, 59 },
		{ 313, 430, 59 },
		{ 313, 431, 59 },
		{ 313, 432, 59 },
		{ 313, 433, 59 },
		{ 313, 434, 59 },
		{ 313, 435, 59 },
		{ 313, 436, 59 },
		{ 313, 437, 59 },
		{ 313, 438, 59 },
		{ 313, 439, 59 },
		{ 313, 440, 59 },
		{ 313, 441, 59 },
		{ 313, 442, 59 },
		{ 313, 443, 59 },
		{ 313, 444, 59 },
		{ 313, 445, 59 },
		{ 313, 446, 59 },
		{ 313, 447, 59 },
		{ 313, 448, 59 },
		{ 313, 449, 59 },
		{ 313, 450, 59 },
		{ 313, 451, 59 },
		{ 313, 452, 59 },
		{ 313, 453, 59 },
		{ 313, 454, 59 },
		{ 313, 455, 59 },
		{ 313, 456, 59 },
		{ 313, 457, 59 },
		{ 313, 458, 59 },
		{ 313, 459, 59 },
		{ 313, 460, 59 },
		{ 313, 461, 59 },
		{ 313, 462, 59 },
		{ 313, 463, 59 },
		{ 313, 464, 59 },
		{ 313, 465, 59 },
		{ 313, 466, 59 },
		{ 313, 467, 59 },
		{ 313, 468, 59 },
		{ 313, 469, 59 },
		{ 313, 470, 59 },
		{ 313, 471, 59 },
		{ 313, 472, 59 },
		{ 313, 473, 59 },
		{ 313, 474, 59 },
		{ 313, 475, 59 },
		{ 313, 476, 59 },
		{ 313, 477, 59 },
		{ 313, 478, 59 },
		{ 313, 479, 59 },
		{ 313, 480, 59 },
		{ 314, 372, 59 },
		{ 314, 373, 59 },
		{ 314, 374, 59 },
		{ 314, 375, 59 },
		{ 314, 376, 59 },
		{ 314, 377, 59 },
		{ 314, 378, 59 },
		{ 314, 379, 59 },
		{ 314, 380, 59 },
		{ 314, 381, 59 },
		{ 314, 382, 59 },
		{ 314, 383, 59 },
		{ 314, 384, 59 },
		{ 314, 385, 59 },
		{ 314, 386, 59 },
		{ 314, 387, 59 },
		{ 314, 388, 59 },
		{ 314, 389, 59 },
		{ 314, 390, 59 },
		{ 314, 391, 59 },
		{ 314, 392, 59 },
		{ 314, 393, 59 },
		{ 314, 394, 59 },
		{ 314, 395, 59 },
		{ 314, 396, 59 },
		{ 314, 397, 59 },
		{ 314, 398, 59 },
		{ 314, 399, 59 },
		{ 314, 400, 59 },
		{ 314, 401, 59 },
		{ 314, 402, 59 },
		{ 314, 403, 59 },
		{ 314, 404, 59 },
		{ 314, 405, 59 },
		{ 314, 406, 59 },
		{ 314, 407, 59 },
		{ 314, 408, 59 },
		{ 314, 409, 59 },
		{ 314, 410, 59 },
		{ 314, 411, 59 },
		{ 314, 412, 59 },
		{ 314, 413, 59 },
		{ 314, 414, 59 },
		{ 314, 415, 59 },
		{ 314, 416, 59 },
		{ 314, 417, 59 },
		{ 314, 418, 59 },
		{ 314, 419, 59 },
		{ 314, 420, 59 },
		{ 314, 421, 59 },
		{ 314, 422, 59 },
		{ 314, 423, 59 },
		{ 314, 424, 59 },
		{ 314, 425, 59 },
		{ 314, 426, 59 },
		{ 314, 427, 59 },
		{ 314, 428, 59 },
		{ 314, 429, 59 },
		{ 314, 430, 59 },
		{ 314, 431, 59 },
		{ 314, 432, 59 },
		{ 314, 433, 59 },
		{ 314, 434, 59 },
		{ 314, 435, 59 },
		{ 314, 436, 59 },
		{ 314, 437, 59 },
		{ 314, 438, 59 },
		{ 314, 439, 59 },
		{ 314, 440, 59 },
		{ 314, 441, 59 },
		{ 314, 442, 59 },
		{ 314, 443, 59 },
		{ 314, 444, 59 },
		{ 314, 445, 59 },
		{ 314, 446, 59 },
		{ 314, 447, 59 },
		{ 314, 448, 59 },
		{ 314, 449, 59 },
		{ 314, 450, 59 },
		{ 314, 451, 59 },
		{ 314, 452, 59 },
		{ 314, 453, 59 },
		{ 314, 454, 59 },
		{ 314, 455, 59 },
		{ 314, 456, 59 },
		{ 314, 457, 59 },
		{ 314, 458, 59 },
		{ 314, 459, 59 },
		{ 314, 460, 59 },
		{ 314, 461, 59 },
		{ 314, 462, 59 },
		{ 314, 463, 59 },
		{ 314, 464, 59 },
		{ 314, 465, 59 },
		{ 314, 466, 59 },
		{ 314, 467, 59 },
		{ 314, 468, 59 },
		{ 314, 469, 59 },
		{ 314, 470, 59 },
		{ 314, 471, 59 },
		{ 314, 472, 59 },
		{ 314, 473, 59 },
		{ 314, 474, 59 },
		{ 314, 475, 59 },
		{ 314, 476, 59 },
		{ 314, 477, 59 },
		{ 314, 478, 59 },
		{ 314, 479, 59 },
		{ 314, 480, 59 },
		{ 314, 481, 59 },
		{ 314, 482, 59 },
		{ 314, 483, 59 },
		{ 314, 484, 59 },
		{ 315, 374, 59 },
		{ 315, 375, 59 },
		{ 315, 376, 59 },
		{ 315, 377, 59 },
		{ 315, 378, 59 },
		{ 315, 379, 59 },
		{ 315, 380, 59 },
		{ 315, 381, 59 },
		{ 315, 382, 59 },
		{ 315, 383, 59 },
		{ 315, 384, 59 },
		{ 315, 385, 59 },
		{ 315, 386, 59 },
		{ 315, 387, 59 },
		{ 315, 388, 59 },
		{ 315, 389, 59 },
		{ 315, 390, 59 },
		{ 315, 391, 59 },
		{ 315, 392, 59 },
		{ 315, 393, 59 },
		{ 315, 394, 59 },
		{ 315, 395, 59 },
		{ 315, 396, 59 },
		{ 315, 397, 59 },
		{ 315, 398, 59 },
		{ 315, 399, 59 },
		{ 315, 400, 59 },
		{ 315, 401, 59 },
		{ 315, 402, 59 },
		{ 315, 403, 59 },
		{ 315, 404, 59 },
		{ 315, 405, 59 },
		{ 315, 406, 59 },
		{ 315, 407, 59 },
		{ 315, 408, 59 },
		{ 315, 409, 59 },
		{ 315, 410, 59 },
		{ 315, 411, 59 },
		{ 315, 412, 59 },
		{ 315, 413, 59 },
		{ 315, 414, 59 },
		{ 315, 415, 59 },
		{ 315, 416, 59 },
		{ 315, 417, 59 },
		{ 315, 418, 59 },
		{ 315, 419, 59 },
		{ 315, 420, 59 },
		{ 315, 421, 59 },
		{ 315, 422, 59 },
		{ 315, 423, 59 },
		{ 315, 424, 59 },
		{ 315, 425, 59 },
		{ 315, 426, 59 },
		{ 315, 427, 59 },
		{ 315, 428, 59 },
		{ 315, 429, 59 },
		{ 315, 430, 59 },
		{ 315, 431, 59 },
		{ 315, 432, 59 },
		{ 315, 433, 59 },
		{ 315, 434, 59 },
		{ 315, 435, 59 },
		{ 315, 436, 59 },
		{ 315, 437, 59 },
		{ 315, 438, 59 },
		{ 315, 439, 59 },
		{ 315, 440, 59 },
		{ 315, 441, 59 },
		{ 315, 442, 59 },
		{ 315, 443, 59 },
		{ 315, 444, 59 },
		{ 315, 445, 59 },
		{ 315, 446, 59 },
		{ 315, 447, 59 },
		{ 315, 448, 59 },
		{ 315, 449, 59 },
		{ 315, 450, 59 },
		{ 315, 451, 59 },
		{ 315, 452, 59 },
		{ 315, 453, 59 },
		{ 315, 454, 59 },
		{ 315, 455, 59 },
		{ 315, 456, 59 },
		{ 315, 457, 59 },
		{ 315, 458, 59 },
		{ 315, 459, 59 },
		{ 315, 460, 59 },
		{ 315, 461, 59 },
		{ 315, 462, 59 },
		{ 315, 463, 59 },
		{ 315, 464, 59 },
		{ 315, 465, 59 },
		{ 315, 466, 59 },
		{ 315, 467, 59 },
		{ 315, 468, 59 },
		{ 315, 469, 59 },
		{ 315, 470, 59 },
		{ 315, 471, 59 },
		{ 315, 472, 59 },
		{ 315, 473, 59 },
		{ 315, 474, 59 },
		{ 315, 475, 59 },
		{ 315, 476, 59 },
		{ 315, 477, 59 },
		{ 315, 478, 59 },
		{ 315, 479, 59 },
		{ 315, 480, 59 },
		{ 315, 481, 59 },
		{ 315, 482, 59 },
		{ 315, 483, 59 },
		{ 315, 484, 59 },
		{ 315, 485, 59 },
		{ 315, 486, 59 },
		{ 315, 487, 59 },
		{ 315, 488, 59 },
		{ 316, 376, 59 },
		{ 316, 377, 59 },
		{ 316, 378, 59 },
		{ 316, 379, 59 },
		{ 316, 380, 59 },
		{ 316, 381, 59 },
		{ 316, 382, 59 },
		{ 316, 383, 59 },
		{ 316, 384, 59 },
		{ 316, 385, 59 },
		{ 316, 386, 59 },
		{ 316, 387, 59 },
		{ 316, 388, 59 },
		{ 316, 389, 59 },
		{ 316, 390, 59 },
		{ 316, 391, 59 },
		{ 316, 392, 59 },
		{ 316, 393, 59 },
		{ 316, 394, 59 },
		{ 316, 395, 59 },
		{ 316, 396, 59 },
		{ 316, 397, 59 },
		{ 316, 398, 59 },
		{ 316, 399, 59 },
		{ 316, 400, 59 },
		{ 316, 401, 59 },
		{ 316, 402, 59 },
		{ 316, 403, 59 },
		{ 316, 404, 59 },
		{ 316, 405, 59 },
		{ 316, 406, 59 },
		{ 316, 407, 59 },
		{ 316, 408, 59 },
		{ 316, 409, 59 },
		{ 316, 410, 59 },
		{ 316, 411, 59 },
		{ 316, 412, 59 },
		{ 316, 413, 59 },
		{ 316, 414, 59 },
		{ 316, 415, 59 },
		{ 316, 416, 59 },
		{ 316, 417, 59 },
		{ 316, 418, 59 },
		{ 316, 419, 59 },
		{ 316, 420, 59 },
		{ 316, 421, 59 },
		{ 316, 422, 59 },
		{ 316, 423, 59 },
		{ 316, 424, 59 },
		{ 316, 425, 59 },
		{ 316, 426, 59 },
		{ 316, 427, 59 },
		{ 316, 428, 59 },
		{ 316, 429, 59 },
		{ 316, 430, 59 },
		{ 316, 431, 59 },
		{ 316, 432, 59 },
		{ 316, 433, 59 },
		{ 316, 434, 59 },
		{ 316, 435, 59 },
		{ 316, 436, 59 },
		{ 316, 437, 59 },
		{ 316, 438, 59 },
		{ 316, 439, 59 },
		{ 316, 440, 59 },
		{ 316, 441, 59 },
		{ 316, 442, 59 },
		{ 316, 443, 59 },
		{ 316, 444, 59 },
		{ 316, 445, 59 },
		{ 316, 446, 59 },
		{ 316, 447, 59 },
		{ 316, 448, 59 },
		{ 316, 449, 59 },
		{ 316, 450, 59 },
		{ 316, 451, 59 },
		{ 316, 452, 59 },
		{ 316, 453, 59 },
		{ 316, 454, 59 },
		{ 316, 455, 59 },
		{ 316, 456, 59 },
		{ 316, 457, 59 },
		{ 316, 458, 59 },
		{ 316, 459, 59 },
		{ 316, 460, 59 },
		{ 316, 461, 59 },
		{ 316, 462, 59 },
		{ 316, 463, 59 },
		{ 316, 464, 59 },
		{ 316, 465, 59 },
		{ 316, 466, 59 },
		{ 316, 467, 59 },
		{ 316, 468, 59 },
		{ 316, 469, 59 },
		{ 316, 470, 59 },
		{ 316, 471, 59 },
		{ 316, 472, 59 },
		{ 316, 473, 59 },
		{ 316, 474, 59 },
		{ 316, 475, 59 },
		{ 316, 476, 59 },
		{ 316, 477, 59 },
		{ 316, 478, 59 },
		{ 316, 479, 59 },
		{ 316, 480, 59 },
		{ 316, 481, 59 },
		{ 316, 482, 59 },
		{ 316, 483, 59 },
		{ 316, 484, 59 },
		{ 316, 485, 59 },
		{ 316, 486, 59 },
		{ 316, 487, 59 },
		{ 316, 488, 59 },
		{ 316, 489, 59 },
		{ 316, 490, 59 },
		{ 316, 491, 59 },
		{ 316, 492, 59 },
		{ 317, 378, 59 },
		{ 317, 379, 59 },
		{ 317, 380, 59 },
		{ 317, 381, 59 },
		{ 317, 382, 59 },
		{ 317, 383, 59 },
		{ 317, 384, 59 },
		{ 317, 385, 59 },
		{ 317, 386, 59 },
		{ 317, 387, 59 },
		{ 317, 388, 59 },
		{ 317, 389, 59 },
		{ 317, 390, 59 },
		{ 317, 391, 59 },
		{ 317, 392, 59 },
		{ 317, 393, 59 },
		{ 317, 394, 59 },
		{ 317, 395, 59 },
		{ 317, 396, 59 },
		{ 317, 397, 59 },
		{ 317, 398, 59 },
		{ 317, 399, 59 },
		{ 317, 400, 59 },
		{ 317, 401, 59 },
		{ 317, 402, 59 },
		{ 317, 403, 59 },
		{ 317, 404, 59 },
		{ 317, 405, 59 },
		{ 317, 406, 59 },
		{ 317, 407, 59 },
		{ 317, 408, 59 },
		{ 317, 409, 59 },
		{ 317, 410, 59 },
		{ 317, 411, 59 },
		{ 317, 412, 59 },
		{ 317, 413, 59 },
		{ 317, 414, 59 },
		{ 317, 415, 59 },
		{ 317, 416, 59 },
		{ 317, 417, 59 },
		{ 317, 418, 59 },
		{ 317, 419, 59 },
		{ 317, 420, 59 },
		{ 317, 421, 59 },
		{ 317, 422, 59 },
		{ 317, 423, 59 },
		{ 317, 424, 59 },
		{ 317, 425, 59 },
		{ 317, 426, 59 },
		{ 317, 427, 59 },
		{ 317, 428, 59 },
		{ 317, 429, 59 },
		{ 317, 430, 59 },
		{ 317, 431, 59 },
		{ 317, 432, 59 },
		{ 317, 433, 59 },
		{ 317, 434, 59 },
		{ 317, 435, 59 },
		{ 317, 436, 59 },
		{ 317, 437, 59 },
		{ 317, 438, 59 },
		{ 317, 439, 59 },
		{ 317, 440, 59 },
		{ 317, 441, 59 },
		{ 317, 442, 59 },
		{ 317, 443, 59 },
		{ 317, 444, 59 },
		{ 317, 445, 59 },
		{ 317, 446, 59 },
		{ 317, 447, 59 },
		{ 317, 448, 59 },
		{ 317, 449, 59 },
		{ 317, 450, 59 },
		{ 317, 451, 59 },
		{ 317, 452, 59 },
		{ 317, 453, 59 },
		{ 317, 454, 59 },
		{ 317, 455, 59 },
		{ 317, 456, 59 },
		{ 317, 457, 59 },
		{ 317, 458, 59 },
		{ 317, 459, 59 },
		{ 317, 460, 59 },
		{ 317, 461, 59 },
		{ 317, 462, 59 },
		{ 317, 463, 59 },
		{ 317, 464, 59 },
		{ 317, 465, 59 },
		{ 317, 466, 59 },
		{ 317, 467, 59 },
		{ 317, 468, 59 },
		{ 317, 469, 59 },
		{ 317, 470, 59 },
		{ 317, 471, 59 },
		{ 317, 472, 59 },
		{ 317, 473, 59 },
		{ 317, 474, 59 },
		{ 317, 475, 59 },
		{ 317, 476, 59 },
		{ 317, 477, 59 },
		{ 317, 478, 59 },
		{ 317, 479, 59 },
		{ 317, 480, 59 },
		{ 317, 481, 59 },
		{ 317, 482, 59 },
		{ 317, 483, 59 },
		{ 317, 484, 59 },
		{ 317, 485, 59 },
		{ 317, 486, 59 },
		{ 317, 487, 59 },
		{ 317, 488, 59 },
		{ 317, 489, 59 },
		{ 317, 490, 59 },
		{ 317, 491, 59 },
		{ 317, 492, 59 },
		{ 317, 493, 59 },
		{ 317, 494, 59 },
		{ 317, 495, 59 },
		{ 317, 496, 59 },
		{ 318, 380, 59 },
		{ 318, 381, 59 },
		{ 318, 382, 59 },
		{ 318, 383, 59 },
		{ 318, 384, 59 },
		{ 318, 385, 59 },
		{ 318, 386, 59 },
		{ 318, 387, 59 },
		{ 318, 388, 59 },
		{ 318, 389, 59 },
		{ 318, 390, 59 },
		{ 318, 391, 59 },
		{ 318, 392, 59 },
		{ 318, 393, 59 },
		{ 318, 394, 59 },
		{ 318, 395, 59 },
		{ 318, 396, 59 },
		{ 318, 397, 59 },
		{ 318, 398, 59 },
		{ 318, 399, 59 },
		{ 318, 400, 59 },
		{ 318, 401, 59 },
		{ 318, 402, 59 },
		{ 318, 403, 59 },
		{ 318, 404, 59 },
		{ 318, 405, 59 },
		{ 318, 406, 59 },
		{ 318, 407, 59 },
		{ 318, 408, 59 },
		{ 318, 409, 59 },
		{ 318, 410, 59 },
		{ 318, 411, 59 },
		{ 318, 412, 59 },
		{ 318, 413, 59 },
		{ 318, 414, 59 },
		{ 318, 415, 59 },
		{ 318, 416, 59 },
		{ 318, 417, 59 },
		{ 318, 418, 59 },
		{ 318, 419, 59 },
		{ 318, 420, 59 },
		{ 318, 421, 59 },
		{ 318, 422, 59 },
		{ 318, 423, 59 },
		{ 318, 424, 59 },
		{ 318, 425, 59 },
		{ 318, 426, 59 },
		{ 318, 427, 59 },
		{ 318, 428, 59 },
		{ 318, 429, 59 },
		{ 318, 430, 59 },
		{ 318, 431, 59 },
		{ 318, 432, 59 },
		{ 318, 433, 59 },
		{ 318, 434, 59 },
		{ 318, 435, 59 },
		{ 318, 436, 59 },
		{ 318, 437, 59 },
		{ 318, 438, 59 },
		{ 318, 439, 59 },
		{ 318, 440, 59 },
		{ 318, 441, 59 },
		{ 318, 442, 59 },
		{ 318, 443, 59 },
		{ 318, 444, 59 },
		{ 318, 445, 59 },
		{ 318, 446, 59 },
		{ 318, 447, 59 },
		{ 318, 448, 59 },
		{ 318, 449, 59 },
		{ 318, 450, 59 },
		{ 318, 451, 59 },
		{ 318, 452, 59 },
		{ 318, 453, 59 },
		{ 318, 454, 59 },
		{ 318, 455, 59 },
		{ 318, 456, 59 },
		{ 318, 457, 59 },
		{ 318, 458, 59 },
		{ 318, 459, 59 },
		{ 318, 460, 59 },
		{ 318, 461, 59 },
		{ 318, 462, 59 },
		{ 318, 463, 59 },
		{ 318, 464, 59 },
		{ 318, 465, 59 },
		{ 318, 466, 59 },
		{ 318, 467, 59 },
		{ 318, 468, 59 },
		{ 318, 469, 59 },
		{ 318, 470, 59 },
		{ 318, 471, 59 },
		{ 318, 472, 59 },
		{ 318, 473, 59 },
		{ 318, 474, 59 },
		{ 318, 475, 59 },
		{ 318, 476, 59 },
		{ 318, 477, 59 },
		{ 318, 478, 59 },
		{ 318, 479, 59 },
		{ 318, 480, 59 },
		{ 318, 481, 59 },
		{ 318, 482, 59 },
		{ 318, 483, 59 },
		{ 318, 484, 59 },
		{ 318, 485, 59 },
		{ 318, 486, 59 },
		{ 318, 487, 59 },
		{ 318, 488, 59 },
		{ 318, 489, 59 },
		{ 318, 490, 59 },
		{ 318, 491, 59 },
		{ 318, 492, 59 },
		{ 318, 493, 59 },
		{ 318, 494, 59 },
		{ 318, 495, 59 },
		{ 318, 496, 59 },
		{ 318, 497, 59 },
		{ 318, 498, 59 },
		{ 318, 499, 59 },
		{ 318, 500, 59 },
		{ 319, 382, 59 },
		{ 319, 383, 59 },
		{ 319, 384, 59 },
		{ 319, 385, 59 },
		{ 319, 386, 59 },
		{ 319, 387, 59 },
		{ 319, 388, 59 },
		{ 319, 389, 59 },
		{ 319, 390, 59 },
		{ 319, 391, 59 },
		{ 319, 392, 59 },
		{ 319, 393, 59 },
		{ 319, 394, 59 },
		{ 319, 395, 59 },
		{ 319, 396, 59 },
		{ 319, 397, 59 },
		{ 319, 398, 59 },
		{ 319, 399, 59 },
		{ 319, 400, 59 },
		{ 319, 401, 59 },
		{ 319, 402, 59 },
		{ 319, 403, 59 },
		{ 319, 404, 59 },
		{ 319, 405, 59 },
		{ 319, 406, 59 },
		{ 319, 407, 59 },
		{ 319, 408, 59 },
		{ 319, 409, 59 },
		{ 319, 410, 59 },
		{ 319, 411, 59 },
		{ 319, 412, 59 },
		{ 319, 413, 59 },
		{ 319, 414, 59 },
		{ 319, 415, 59 },
		{ 319, 416, 59 },
		{ 319, 417, 59 },
		{ 319, 418, 59 },
		{ 319, 419, 59 },
		{ 319, 420, 59 },
		{ 319, 421, 59 },
		{ 319, 422, 59 },
		{ 319, 423, 59 },
		{ 319, 424, 59 },
		{ 319, 425, 59 },
		{ 319, 426, 59 },
		{ 319, 427, 59 },
		{ 319, 428, 59 },
		{ 319, 429, 59 },
		{ 319, 430, 59 },
		{ 319, 431, 59 },
		{ 319, 432, 59 },
		{ 319, 433, 59 },
		{ 319, 434, 59 },
		{ 319, 435, 59 },
		{ 319, 436, 59 },
		{ 319, 437, 59 },
		{ 319, 438, 59 },
		{ 319, 439, 59 },
		{ 319, 440, 59 },
		{ 319, 441, 59 },
		{ 319, 442, 59 },
		{ 319, 443, 59 },
		{ 319, 444, 59 },
		{ 319, 445, 59 },
		{ 319, 446, 59 },
		{ 319, 447, 59 },
		{ 319, 448, 59 },
		{ 319, 449, 59 },
		{ 319, 450, 59 },
		{ 319, 451, 59 },
		{ 319, 452, 59 },
		{ 319, 453, 59 },
		{ 319, 454, 59 },
		{ 319, 455, 59 },
		{ 319, 456, 59 },
		{ 319, 457, 59 },
		{ 319, 458, 59 },
		{ 319, 459, 59 },
		{ 319, 460, 59 },
		{ 319, 461, 59 },
		{ 319, 462, 59 },
		{ 319, 463, 59 },
		{ 319, 464, 59 },
		{ 319, 465, 59 },
		{ 319, 466, 59 },
		{ 319, 467, 59 },
		{ 319, 468, 59 },
		{ 319, 469, 59 },
		{ 319, 470, 59 },
		{ 319, 471, 59 },
		{ 319, 472, 59 },
		{ 319, 473, 59 },
		{ 319, 474, 59 },
		{ 319, 475, 59 },
		{ 319, 476, 59 },
		{ 319, 477, 59 },
		{ 319, 478, 59 },
		{ 319, 479, 59 },
		{ 319, 480, 59 },
		{ 319, 481, 59 },
		{ 319, 482, 59 },
		{ 319, 483, 59 },
		{ 319, 484, 59 },
		{ 319, 485, 59 },
		{ 319, 486, 59 },
		{ 319, 487, 59 },
		{ 319, 488, 59 },
		{ 319, 489, 59 },
		{ 319, 490, 59 },
		{ 319, 491, 59 },
		{ 319, 492, 59 },
		{ 319, 493, 59 },
		{ 319, 494, 59 },
		{ 319, 495, 59 },
		{ 319, 496, 59 },
		{ 319, 497, 59 },
		{ 319, 498, 59 },
		{ 319, 499, 59 },
		{ 319, 500, 59 },
		{ 319, 501, 59 },
		{ 319, 502, 59 },
		{ 319, 503, 59 },
		{ 319, 504, 59 },
		{ 320, 384, 59 },
		{ 320, 385, 59 },
		{ 320, 386, 59 },
		{ 320, 387, 59 },
		{ 320, 388, 59 },
		{ 320, 389, 59 },
		{ 320, 390, 59 },
		{ 320, 391, 59 },
		{ 320, 392, 59 },
		{ 320, 393, 59 },
		{ 320, 394, 59 },
		{ 320, 395, 59 },
		{ 320, 396, 59 },
		{ 320, 397, 59 },
		{ 320, 398, 59 },
		{ 320, 399, 59 },
		{ 320, 400, 59 },
		{ 320, 401, 59 },
		{ 320, 402, 59 },
		{ 320, 403, 59 },
		{ 320, 404, 59 },
		{ 320, 405, 59 },
		{ 320, 406, 59 },
		{ 320, 407, 59 },
		{ 320, 408, 59 },
		{ 320, 409, 59 },
		{ 320, 410, 59 },
		{ 320, 411, 59 },
		{ 320, 412, 59 },
		{ 320, 413, 59 },
		{ 320, 414, 59 },
		{ 320, 415, 59 },
		{ 320, 416, 59 },
		{ 320, 417, 59 },
		{ 320, 418, 59 },
		{ 320, 419, 59 },
		{ 320, 420, 59 },
		{ 320, 421, 59 },
		{ 320, 422, 59 },
		{ 320, 423, 59 },
		{ 320, 424, 59 },
		{ 320, 425, 59 },
		{ 320, 426, 59 },
		{ 320, 427, 59 },
		{ 320, 428, 59 },
		{ 320, 429, 59 },
		{ 320, 430, 59 },
		{ 320, 431, 59 },
		{ 320, 432, 59 },
		{ 320, 433, 59 },
		{ 320, 434, 59 },
		{ 320, 435, 59 },
		{ 320, 436, 59 },
		{ 320, 437, 59 },
		{ 320, 438, 59 },
		{ 320, 439, 59 },
		{ 320, 440, 59 },
		{ 320, 441, 59 },
		{ 320, 442, 59 },
		{ 320, 443, 59 },
		{ 320, 444, 59 },
		{ 320, 445, 59 },
		{ 320, 446, 59 },
		{ 320, 447, 59 },
		{ 320, 448, 59 },
		{ 320, 449, 59 },
		{ 320, 450, 59 },
		{ 320, 451, 59 },
		{ 320, 452, 59 },
		{ 320, 453, 59 },
		{ 320, 454, 59 },
		{ 320, 455, 59 },
		{ 320, 456, 59 },
		{ 320, 457, 59 },
		{ 320, 458, 59 },
		{ 320, 459, 59 },
		{ 320, 460, 59 },
		{ 320, 461, 59 },
		{ 320, 462, 59 },
		{ 320, 463, 59 },
		{ 320, 464, 59 },
		{ 320, 465, 59 },
		{ 320, 466, 59 },
		{ 320, 467, 59 },
		{ 320, 468, 59 },
		{ 320, 469, 59 },
		{ 320, 470, 59 },
		{ 320, 471, 59 },
		{ 320, 472, 59 },
		{ 320, 473, 59 },
		{ 320, 474, 59 },
		{ 320, 475, 59 },
		{ 320, 476, 59 },
		{ 320, 477, 59 },
		{ 320, 478, 59 },
		{ 320, 479, 59 },
		{ 320, 480, 59 },
		{ 320, 481, 59 },
		{ 320, 482, 59 },
		{ 320, 483, 59 },
		{ 320, 484, 59 },
		{ 320, 485, 59 },
		{ 320, 486, 59 },
		{ 320, 487, 59 },
		{ 320, 488, 59 },
		{ 320, 489, 59 },
		{ 320, 490, 59 },
		{ 320, 491, 59 },
		{ 320, 492, 59 },
		{ 320, 493, 59 },
		{ 320, 494, 59 },
		{ 320, 495, 59 },
		{ 320, 496, 59 },
		{ 320, 497, 59 },
		{ 320, 498, 59 },
		{ 320, 499, 59 },
		{ 320, 500, 59 },
		{ 320, 501, 59 },
		{ 320, 502, 59 },
		{ 320, 503, 59 },
		{ 320, 504, 59 },
		{ 320, 505, 59 },
		{ 320, 506, 59 },
		{ 320, 507, 59 },
		{ 320, 508, 59 },
		{ 321, 386, 59 },
		{ 321, 387, 59 },
		{ 321, 388, 59 },
		{ 321, 389, 59 },
		{ 321, 390, 59 },
		{ 321, 391, 59 },
		{ 321, 392, 59 },
		{ 321, 393, 59 },
		{ 321, 394, 59 },
		{ 321, 395, 59 },
		{ 321, 396, 59 },
		{ 321, 397, 59 },
		{ 321, 398, 59 },
		{ 321, 399, 59 },
		{ 321, 400, 59 },
		{ 321, 401, 59 },
		{ 321, 402, 59 },
		{ 321, 403, 59 },
		{ 321, 404, 59 },
		{ 321, 405, 59 },
		{ 321, 406, 59 },
		{ 321, 407, 59 },
		{ 321, 408, 59 },
		{ 321, 409, 59 },
		{ 321, 410, 59 },
		{ 321, 411, 59 },
		{ 321, 412, 59 },
		{ 321, 413, 59 },
		{ 321, 414, 59 },
		{ 321, 415, 59 },
		{ 321, 416, 59 },
		{ 321, 417, 59 },
		{ 321, 418, 59 },
		{ 321, 419, 59 },
		{ 321, 420, 59 },
		{ 321, 421, 59 },
		{ 321, 422, 59 },
		{ 321, 423, 59 },
		{ 321, 424, 59 },
		{ 321, 425, 59 },
		{ 321, 426, 59 },
		{ 321, 427, 59 },
		{ 321, 428, 59 },
		{ 321, 429, 59 },
		{ 321, 430, 59 },
		{ 321, 431, 59 },
		{ 321, 432, 59 },
		{ 321, 433, 59 },
		{ 321, 434, 59 },
		{ 321, 435, 59 },
		{ 321, 436, 59 },
		{ 321, 437, 59 },
		{ 321, 438, 59 },
		{ 321, 439, 59 },
		{ 321, 440, 59 },
		{ 321, 441, 59 },
		{ 321, 442, 59 },
		{ 321, 443, 59 },
		{ 321, 444, 59 },
		{ 321, 445, 59 },
		{ 321, 446, 59 },
		{ 321, 447, 59 },
		{ 321, 448, 59 },
		{ 321, 449, 59 },
		{ 321, 450, 59 },
		{ 321, 451, 59 },
		{ 321, 452, 59 },
		{ 321, 453, 59 },
		{ 321, 454, 59 },
		{ 321, 455, 59 },
		{ 321, 456, 59 },
		{ 321, 457, 59 },
		{ 321, 458, 59 },
		{ 321, 459, 59 },
		{ 321, 460, 59 },
		{ 321, 461, 59 },
		{ 321, 462, 59 },
		{ 321, 463, 59 },
		{ 321, 464, 59 },
		{ 321, 465, 59 },
		{ 321, 466, 59 },
		{ 321, 467, 59 },
		{ 321, 468, 59 },
		{ 321, 469, 59 },
		{ 321, 470, 59 },
		{ 321, 471, 59 },
		{ 321, 472, 59 },
		{ 321, 473, 59 },
		{ 321, 474, 59 },
		{ 321, 475, 59 },
		{ 321, 476, 59 },
		{ 321, 477, 59 },
		{ 321, 478, 59 },
		{ 321, 479, 59 },
		{ 321, 480, 59 },
		{ 321, 481, 59 },
		{ 321, 482, 59 },
		{ 321, 483, 59 },
		{ 321, 484, 59 },
		{ 321, 485, 59 },
		{ 321, 486, 59 },
		{ 321, 487, 59 },
		{ 321, 488, 59 },
		{ 321, 489, 59 },
		{ 321, 490, 59 },
		{ 321, 491, 59 },
		{ 321, 492, 59 },
		{ 321, 493, 59 },
		{ 321, 494, 59 },
		{ 321, 495, 59 },
		{ 321, 496, 59 },
		{ 321, 497, 59 },
		{ 321, 498, 59 },
		{ 321, 499, 59 },
		{ 321, 500, 59 },
		{ 321, 501, 59 },
		{ 321, 502, 59 },
		{ 321, 503, 59 },
		{ 321, 504, 59 },
		{ 321, 505, 59 },
		{ 321, 506, 59 },
		{ 321, 507, 59 },
		{ 321, 508, 59 },
		{ 321, 509, 59 },
		{ 321, 510, 59 },
		{ 321, 511, 59 },
		{ 321, 512, 59 },
		{ 322, 388, 59 },
		{ 322, 389, 59 },
		{ 322, 390, 59 },
		{ 322, 391, 59 },
		{ 322, 392, 59 },
		{ 322, 393, 59 },
		{ 322, 394, 59 },
		{ 322, 395, 59 },
		{ 322, 396, 59 },
		{ 322, 397, 59 },
		{ 322, 398, 59 },
		{ 322, 399, 59 },
		{ 322, 400, 59 },
		{ 322, 401, 59 },
		{ 322, 402, 59 },
		{ 322, 403, 59 },
		{ 322, 404, 59 },
		{ 322, 405, 59 },
		{ 322, 406, 59 },
		{ 322, 407, 59 },
		{ 322, 408, 59 },
		{ 322, 409, 59 },
		{ 322, 410, 59 },
		{ 322, 411, 59 },
		{ 322, 412, 59 },
		{ 322, 413, 59 },
		{ 322, 414, 59 },
		{ 322, 415, 59 },
		{ 322, 416, 59 },
		{ 322, 417, 59 },
		{ 322, 418, 59 },
		{ 322, 419, 59 },
		{ 322, 420, 59 },
		{ 322, 421, 59 },
		{ 322, 422, 59 },
		{ 322, 423, 59 },
		{ 322, 424, 59 },
		{ 322, 425, 59 },
		{ 322, 426, 59 },
		{ 322, 427, 59 },
		{ 322, 428, 59 },
		{ 322, 429, 59 },
		{ 322, 430, 59 },
		{ 322, 431, 59 },
		{ 322, 432, 59 },
		{ 322, 433, 59 },
		{ 322, 434, 59 },
		{ 322, 435, 59 },
		{ 322, 436, 59 },
		{ 322, 437, 59 },
		{ 322, 438, 59 },
		{ 322, 439, 59 },
		{ 322, 440, 59 },
		{ 322, 441, 59 },
		{ 322, 442, 59 },
		{ 322, 443, 59 },
		{ 322, 444, 59 },
		{ 322, 445, 59 },
		{ 322, 446, 59 },
		{ 322, 447, 59 },
		{ 322, 448, 59 },
		{ 322, 449, 59 },
		{ 322, 450, 59 },
		{ 322, 451, 59 },
		{ 322, 452, 59 },
		{ 322, 453, 59 },
		{ 322, 454, 59 },
		{ 322, 455, 59 },
		{ 322, 456, 59 },
		{ 322, 457, 59 },
		{ 322, 458, 59 },
		{ 322, 459, 59 },
		{ 322, 460, 59 },
		{ 322, 461, 59 },
		{ 322, 462, 59 },
		{ 322, 463, 59 },
		{ 322, 464, 59 },
		{ 322, 465, 59 },
		{ 322, 466, 59 },
		{ 322, 467, 59 },
		{ 322, 468, 59 },
		{ 322, 469, 59 },
		{ 322, 470, 59 },
		{ 322, 471, 59 },
		{ 322, 472, 59 },
		{ 322, 473, 59 },
		{ 322, 474, 59 },
		{ 322, 475, 59 },
		{ 322, 476, 59 },
		{ 322, 477, 59 },
		{ 322, 478, 59 },
		{ 322, 479, 59 },
		{ 322, 480, 59 },
		{ 322, 481, 59 },
		{ 322, 482, 59 },
		{ 322, 483, 59 },
		{ 322, 484, 59 },
		{ 322, 485, 59 },
		{ 322, 486, 59 },
		{ 322, 487, 59 },
		{ 322, 488, 59 },
		{ 322, 489, 59 },
		{ 322, 490, 59 },
		{ 322, 491, 59 },
		{ 322, 492, 59 },
		{ 322, 493, 59 },
		{ 322, 494, 59 },
		{ 322, 495, 59 },
		{ 322, 496, 59 },
		{ 322, 497, 59 },
		{ 322, 498, 59 },
		{ 322, 499, 59 },
		{ 322, 500, 59 },
		{ 322, 501, 59 },
		{ 322, 502, 59 },
		{ 322, 503, 59 },
		{ 322, 504, 59 },
		{ 322, 505, 59 },
		{ 322, 506, 59 },
		{ 322, 507, 59 },
		{ 322, 508, 59 },
		{ 322, 509, 59 },
		{ 322, 510, 59 },
		{ 322, 511, 59 },
		{ 322, 512, 59 },
		{ 323, 390, 59 },
		{ 323, 391, 59 },
		{ 323, 392, 59 },
		{ 323, 393, 59 },
		{ 323, 394, 59 },
		{ 323, 395, 59 },
		{ 323, 396, 59 },
		{ 323, 397, 59 },
		{ 323, 398, 59 },
		{ 323, 399, 59 },
		{ 323, 400, 59 },
		{ 323, 401, 59 },
		{ 323, 402, 59 },
		{ 323, 403, 59 },
		{ 323, 404, 59 },
		{ 323, 405, 59 },
		{ 323, 406, 59 },
		{ 323, 407, 59 },
		{ 323, 408, 59 },
		{ 323, 409, 59 },
		{ 323, 410, 59 },
		{ 323, 411, 59 },
		{ 323, 412, 59 },
		{ 323, 413, 59 },
		{ 323, 414, 59 },
		{ 323, 415, 59 },
		{ 323, 416, 59 },
		{ 323, 417, 59 },
		{ 323, 418, 59 },
		{ 323, 419, 59 },
		{ 323, 420, 59 },
		{ 323, 421, 59 },
		{ 323, 422, 59 },
		{ 323, 423, 59 },
		{ 323, 424, 59 },
		{ 323, 425, 59 },
		{ 323, 426, 59 },
		{ 323, 427, 59 },
		{ 323, 428, 59 },
		{ 323, 429, 59 },
		{ 323, 430, 59 },
		{ 323, 431, 59 },
		{ 323, 432, 59 },
		{ 323, 433, 59 },
		{ 323, 434, 59 },
		{ 323, 435, 59 },
		{ 323, 436, 59 },
		{ 323, 437, 59 },
		{ 323, 438, 59 },
		{ 323, 439, 59 },
		{ 323, 440, 59 },
		{ 323, 441, 59 },
		{ 323, 442, 59 },
		{ 323, 443, 59 },
		{ 323, 444, 59 },
		{ 323, 445, 59 },
		{ 323, 446, 59 },
		{ 323, 447, 59 },
		{ 323, 448, 59 },
		{ 323, 449, 59 },
		{ 323, 450, 59 },
		{ 323, 451, 59 },
		{ 323, 452, 59 },
		{ 323, 453, 59 },
		{ 323, 454, 59 },
		{ 323, 455, 59 },
		{ 323, 456, 59 },
		{ 323, 457, 59 },
		{ 323, 458, 59 },
		{ 323, 459, 59 },
		{ 323, 460, 59 },
		{ 323, 461, 59 },
		{ 323, 462, 59 },
		{ 323, 463, 59 },
		{ 323, 464, 59 },
		{ 323, 465, 59 },
		{ 323, 466, 59 },
		{ 323, 467, 59 },
		{ 323, 468, 59 },
		{ 323, 469, 59 },
		{ 323, 470, 59 },
		{ 323, 471, 59 },
		{ 323, 472, 59 },
		{ 323, 473, 59 },
		{ 323, 474, 59 },
		{ 323, 475, 59 },
		{ 323, 476, 59 },
		{ 323, 477, 59 },
		{ 323, 478, 59 },
		{ 323, 479, 59 },
		{ 323, 480, 59 },
		{ 323, 481, 59 },
		{ 323, 482, 59 },
		{ 323, 483, 59 },
		{ 323, 484, 59 },
		{ 323, 485, 59 },
		{ 323, 486, 59 },
		{ 323, 487, 59 },
		{ 323, 488, 59 },
		{ 323, 489, 59 },
		{ 323, 490, 59 },
		{ 323, 491, 59 },
		{ 323, 492, 59 },
		{ 323, 493, 59 },
		{ 323, 494, 59 },
		{ 323, 495, 59 },
		{ 323, 496, 59 },
		{ 323, 497, 59 },
		{ 323, 498, 59 },
		{ 323, 499, 59 },
		{ 323, 500, 59 },
		{ 323, 501, 59 },
		{ 323, 502, 59 },
		{ 323, 503, 59 },
		{ 323, 504, 59 },
		{ 323, 505, 59 },
		{ 323, 506, 59 },
		{ 323, 507, 59 },
		{ 323, 508, 59 },
		{ 323, 509, 59 },
		{ 323, 510, 59 },
		{ 323, 511, 59 },
		{ 323, 512, 59 },
		{ 324, 392, 59 },
		{ 324, 393, 59 },
		{ 324, 394, 59 },
		{ 324, 395, 59 },
		{ 324, 396, 59 },
		{ 324, 397, 59 },
		{ 324, 398, 59 },
		{ 324, 399, 59 },
		{ 324, 400, 59 },
		{ 324, 401, 59 },
		{ 324, 402, 59 },
		{ 324, 403, 59 },
		{ 324, 404, 59 },
		{ 324, 405, 59 },
		{ 324, 406, 59 },
		{ 324, 407, 59 },
		{ 324, 408, 59 },
		{ 324, 409, 59 },
		{ 324, 410, 59 },
		{ 324, 411, 59 },
		{ 324, 412, 59 },
		{ 324, 413, 59 },
		{ 324, 414, 59 },
		{ 324, 415, 59 },
		{ 324, 416, 59 },
		{ 324, 417, 59 },
		{ 324, 418, 59 },
		{ 324, 419, 59 },
		{ 324, 420, 59 },
		{ 324, 421, 59 },
		{ 324, 422, 59 },
		{ 324, 423, 59 },
		{ 324, 424, 59 },
		{ 324, 425, 59 },
		{ 324, 426, 59 },
		{ 324, 427, 59 },
		{ 324, 428, 59 },
		{ 324, 429, 59 },
		{ 324, 430, 59 },
		{ 324, 431, 59 },
		{ 324, 432, 59 },
		{ 324, 433, 59 },
		{ 324, 434, 59 },
		{ 324, 435, 59 },
		{ 324, 436, 59 },
		{ 324, 437, 59 },
		{ 324, 438, 59 },
		{ 324, 439, 59 },
		{ 324, 440, 59 },
		{ 324, 441, 59 },
		{ 324, 442, 59 },
		{ 324, 443, 59 },
		{ 324, 444, 59 },
		{ 324, 445, 59 },
		{ 324, 446, 59 },
		{ 324, 447, 59 },
		{ 324, 448, 59 },
		{ 324, 449, 59 },
		{ 324, 450, 59 },
		{ 324, 451, 59 },
		{ 324, 452, 59 },
		{ 324, 453, 59 },
		{ 324, 454, 59 },
		{ 324, 455, 59 },
		{ 324, 456, 59 },
		{ 324, 457, 59 },
		{ 324, 458, 59 },
		{ 324, 459, 59 },
		{ 324, 460, 59 },
		{ 324, 461, 59 },
		{ 324, 462, 59 },
		{ 324, 463, 59 },
		{ 324, 464, 59 },
		{ 324, 465, 59 },
		{ 324, 466, 59 },
		{ 324, 467, 59 },
		{ 324, 468, 59 },
		{ 324, 469, 59 },
		{ 324, 470, 59 },
		{ 324, 471, 59 },
		{ 324, 472, 59 },
		{ 324, 473, 59 },
		{ 324, 474, 59 },
		{ 324, 475, 59 },
		{ 324, 476, 59 },
		{ 324, 477, 59 },
		{ 324, 478, 59 },
		{ 324, 479, 59 },
		{ 324, 480, 59 },
		{ 324, 481, 59 },
		{ 324, 482, 59 },
		{ 324, 483, 59 },
		{ 324, 484, 59 },
		{ 324, 485, 59 },
		{ 324, 486, 59 },
		{ 324, 487, 59 },
		{ 324, 488, 59 },
		{ 324, 489, 59 },
		{ 324, 490, 59 },
		{ 324, 491, 59 },
		{ 324, 492, 59 },
		{ 324, 493, 59 },
		{ 324, 494, 59 },
		{ 324, 495, 59 },
		{ 324, 496, 59 },
		{ 324, 497, 59 },
		{ 324, 498, 59 },
		{ 324, 499, 59 },
		{ 324, 500, 59 },
		{ 324, 501, 59 },
		{ 324, 502, 59 },
		{ 324, 503, 59 },
		{ 324, 504, 59 },
		{ 324, 505, 59 },
		{ 324, 506, 59 },
		{ 324, 507, 59 },
		{ 324, 508, 59 },
		{ 324, 509, 59 },
		{ 324, 510, 59 },
		{ 324, 511, 59 },
		{ 324, 512, 59 },
		{ 325, 394, 59 },
		{ 325, 395, 59 },
		{ 325, 396, 59 },
		{ 325, 397, 59 },
		{ 325, 398, 59 },
		{ 325, 399, 59 },
		{ 325, 400, 59 },
		{ 325, 401, 59 },
		{ 325, 402, 59 },
		{ 325, 403, 59 },
		{ 325, 404, 59 },
		{ 325, 405, 59 },
		{ 325, 406, 59 },
		{ 325, 407, 59 },
		{ 325, 408, 59 },
		{ 325, 409, 59 },
		{ 325, 410, 59 },
		{ 325, 411, 59 },
		{ 325, 412, 59 },
		{ 325, 413, 59 },
		{ 325, 414, 59 },
		{ 325, 415, 59 },
		{ 325, 416, 59 },
		{ 325, 417, 59 },
		{ 325, 418, 59 },
		{ 325, 419, 59 },
		{ 325, 420, 59 },
		{ 325, 421, 59 },
		{ 325, 422, 59 },
		{ 325, 423, 59 },
		{ 325, 424, 59 },
		{ 325, 425, 59 },
		{ 325, 426, 59 },
		{ 325, 427, 59 },
		{ 325, 428, 59 },
		{ 325, 429, 59 },
		{ 325, 430, 59 },
		{ 325, 431, 59 },
		{ 325, 432, 59 },
		{ 325, 433, 59 },
		{ 325, 434, 59 },
		{ 325, 435, 59 },
		{ 325, 436, 59 },
		{ 325, 437, 59 },
		{ 325, 438, 59 },
		{ 325, 439, 59 },
		{ 325, 440, 59 },
		{ 325, 441, 59 },
		{ 325, 442, 59 },
		{ 325, 443, 59 },
		{ 325, 444, 59 },
		{ 325, 445, 59 },
		{ 325, 446, 59 },
		{ 325, 447, 59 },
		{ 325, 448, 59 },
		{ 325, 449, 59 },
		{ 325, 450, 59 },
		{ 325, 451, 59 },
		{ 325, 452, 59 },
		{ 325, 453, 59 },
		{ 325, 454, 59 },
		{ 325, 455, 59 },
		{ 325, 456, 59 },
		{ 325, 457, 59 },
		{ 325, 458, 59 },
		{ 325, 459, 59 },
		{ 325, 460, 59 },
		{ 325, 461, 59 },
		{ 325, 462, 59 },
		{ 325, 463, 59 },
		{ 325, 464, 59 },
		{ 325, 465, 59 },
		{ 325, 466, 59 },
		{ 325, 467, 59 },
		{ 325, 468, 59 },
		{ 325, 469, 59 },
		{ 325, 470, 59 },
		{ 325, 471, 59 },
		{ 325, 472, 59 },
		{ 325, 473, 59 },
		{ 325, 474, 59 },
		{ 325, 475, 59 },
		{ 325, 476, 59 },
		{ 325, 477, 59 },
		{ 325, 478, 59 },
		{ 325, 479, 59 },
		{ 325, 480, 59 },
		{ 325, 481, 59 },
		{ 325, 482, 59 },
		{ 325, 483, 59 },
		{ 325, 484, 59 },
		{ 325, 485, 59 },
		{ 325, 486, 59 },
		{ 325, 487, 59 },
		{ 325, 488, 59 },
		{ 325, 489, 59 },
		{ 325, 490, 59 },
		{ 325, 491, 59 },
		{ 325, 492, 59 },
		{ 325, 493, 59 },
		{ 325, 494, 59 },
		{ 325, 495, 59 },
		{ 325, 496, 59 },
		{ 325, 497, 59 },
		{ 325, 498, 59 },
		{ 325, 499, 59 },
		{ 325, 500, 59 },
		{ 325, 501, 59 },
		{ 325, 502, 59 },
		{ 325, 503, 59 },
		{ 325, 504, 59 },
		{ 325, 505, 59 },
		{ 325, 506, 59 },
		{ 325, 507, 59 },
		{ 325, 508, 59 },
		{ 325, 509, 59 },
		{ 325, 510, 59 },
		{ 325, 511, 59 },
		{ 325, 512, 59 },
		{ 326, 396, 59 },
		{ 326, 397, 59 },
		{ 326, 398, 59 },
		{ 326, 399, 59 },
		{ 326, 400, 59 },
		{ 326, 401, 59 },
		{ 326, 402, 59 },
		{ 326, 403, 59 },
		{ 326, 404, 59 },
		{ 326, 405, 59 },
		{ 326, 406, 59 },
		{ 326, 407, 59 },
		{ 326, 408, 59 },
		{ 326, 409, 59 },
		{ 326, 410, 59 },
		{ 326, 411, 59 },
		{ 326, 412, 59 },
		{ 326, 413, 59 },
		{ 326, 414, 59 },
		{ 326, 415, 59 },
		{ 326, 416, 59 },
		{ 326, 417, 59 },
		{ 326, 418, 59 },
		{ 326, 419, 59 },
		{ 326, 420, 59 },
		{ 326, 421, 59 },
		{ 326, 422, 59 },
		{ 326, 423, 59 },
		{ 326, 424, 59 },
		{ 326, 425, 59 },
		{ 326, 426, 59 },
		{ 326, 427, 59 },
		{ 326, 428, 59 },
		{ 326, 429, 59 },
		{ 326, 430, 59 },
		{ 326, 431, 59 },
		{ 326, 432, 59 },
		{ 326, 433, 59 },
		{ 326, 434, 59 },
		{ 326, 435, 59 },
		{ 326, 436, 59 },
		{ 326, 437, 59 },
		{ 326, 438, 59 },
		{ 326, 439, 59 },
		{ 326, 440, 59 },
		{ 326, 441, 59 },
		{ 326, 442, 59 },
		{ 326, 443, 59 },
		{ 326, 444, 59 },
		{ 326, 445, 59 },
		{ 326, 446, 59 },
		{ 326, 447, 59 },
		{ 326, 448, 59 },
		{ 326, 449, 59 },
		{ 326, 450, 59 },
		{ 326, 451, 59 },
		{ 326, 452, 59 },
		{ 326, 453, 59 },
		{ 326, 454, 59 },
		{ 326, 455, 59 },
		{ 326, 456, 59 },
		{ 326, 457, 59 },
		{ 326, 458, 59 },
		{ 326, 459, 59 },
		{ 326, 460, 59 },
		{ 326, 461, 59 },
		{ 326, 462, 59 },
		{ 326, 463, 59 },
		{ 326, 464, 59 },
		{ 326, 465, 59 },
		{ 326, 466, 59 },
		{ 326, 467, 59 },
		{ 326, 468, 59 },
		{ 326, 469, 59 },
		{ 326, 470, 59 },
		{ 326, 471, 59 },
		{ 326, 472, 59 },
		{ 326, 473, 59 },
		{ 326, 474, 59 },
		{ 326, 475, 59 },
		{ 326, 476, 59 },
		{ 326, 477, 59 },
		{ 326, 478, 59 },
		{ 326, 479, 59 },
		{ 326, 480, 59 },
		{ 326, 481, 59 },
		{ 326, 482, 59 },
		{ 326, 483, 59 },
		{ 326, 484, 59 },
		{ 326, 485, 59 },
		{ 326, 486, 59 },
		{ 326, 487, 59 },
		{ 326, 488, 59 },
		{ 326, 489, 59 },
		{ 326, 490, 59 },
		{ 326, 491, 59 },
		{ 326, 492, 59 },
		{ 326, 493, 59 },
		{ 326, 494, 59 },
		{ 326, 495, 59 },
		{ 326, 496, 59 },
		{ 326, 497, 59 },
		{ 326, 498, 59 },
		{ 326, 499, 59 },
		{ 326, 500, 59 },
		{ 326, 501, 59 },
		{ 326, 502, 59 },
		{ 326, 503, 59 },
		{ 326, 504, 59 },
		{ 326, 505, 59 },
		{ 326, 506, 59 },
		{ 326, 507, 59 },
		{ 326, 508, 59 },
		{ 326, 509, 59 },
		{ 326, 510, 59 },
		{ 326, 511, 59 },
		{ 326, 512, 59 },
		{ 327, 398, 59 },
		{ 327, 399, 59 },
		{ 327, 400, 59 },
		{ 327, 401, 59 },
		{ 327, 402, 59 },
		{ 327, 403, 59 },
		{ 327, 404, 59 },
		{ 327, 405, 59 },
		{ 327, 406, 59 },
		{ 327, 407, 59 },
		{ 327, 408, 59 },
		{ 327, 409, 59 },
		{ 327, 410, 59 },
		{ 327, 411, 59 },
		{ 327, 412, 59 },
		{ 327, 413, 59 },
		{ 327, 414, 59 },
		{ 327, 415, 59 },
		{ 327, 416, 59 },
		{ 327, 417, 59 },
		{ 327, 418, 59 },
		{ 327, 419, 59 },
		{ 327, 420, 59 },
		{ 327, 421, 59 },
		{ 327, 422, 59 },
		{ 327, 423, 59 },
		{ 327, 424, 59 },
		{ 327, 425, 59 },
		{ 327, 426, 59 },
		{ 327, 427, 59 },
		{ 327, 428, 59 },
		{ 327, 429, 59 },
		{ 327, 430, 59 },
		{ 327, 431, 59 },
		{ 327, 432, 59 },
		{ 327, 433, 59 },
		{ 327, 434, 59 },
		{ 327, 435, 59 },
		{ 327, 436, 59 },
		{ 327, 437, 59 },
		{ 327, 438, 59 },
		{ 327, 439, 59 },
		{ 327, 440, 59 },
		{ 327, 441, 59 },
		{ 327, 442, 59 },
		{ 327, 443, 59 },
		{ 327, 444, 59 },
		{ 327, 445, 59 },
		{ 327, 446, 59 },
		{ 327, 447, 59 },
		{ 327, 448, 59 },
		{ 327, 449, 59 },
		{ 327, 450, 59 },
		{ 327, 451, 59 },
		{ 327, 452, 59 },
		{ 327, 453, 59 },
		{ 327, 454, 59 },
		{ 327, 455, 59 },
		{ 327, 456, 59 },
		{ 327, 457, 59 },
		{ 327, 458, 59 },
		{ 327, 459, 59 },
		{ 327, 460, 59 },
		{ 327, 461, 59 },
		{ 327, 462, 59 },
		{ 327, 463, 59 },
		{ 327, 464, 59 },
		{ 327, 465, 59 },
		{ 327, 466, 59 },
		{ 327, 467, 59 },
		{ 327, 468, 59 },
		{ 327, 469, 59 },
		{ 327, 470, 59 },
		{ 327, 471, 59 },
		{ 327, 472, 59 },
		{ 327, 473, 59 },
		{ 327, 474, 59 },
		{ 327, 475, 59 },
		{ 327, 476, 59 },
		{ 327, 477, 59 },
		{ 327, 478, 59 },
		{ 327, 479, 59 },
		{ 327, 480, 59 },
		{ 327, 481, 59 },
		{ 327, 482, 59 },
		{ 327, 483, 59 },
		{ 327, 484, 59 },
		{ 327, 485, 59 },
		{ 327, 486, 59 },
		{ 327, 487, 59 },
		{ 327, 488, 59 },
		{ 327, 489, 59 },
		{ 327, 490, 59 },
		{ 327, 491, 59 },
		{ 327, 492, 59 },
		{ 327, 493, 59 },
		{ 327, 494, 59 },
		{ 327, 495, 59 },
		{ 327, 496, 59 },
		{ 327, 497, 59 },
		{ 327, 498, 59 },
		{ 327, 499, 59 },
		{ 327, 500, 59 },
		{ 327, 501, 59 },
		{ 327, 502, 59 },
		{ 327, 503, 59 },
		{ 327, 504, 59 },
		{ 327, 505, 59 },
		{ 327, 506, 59 },
		{ 327, 507, 59 },
		{ 327, 508, 59 },
		{ 327, 509, 59 },
		{ 327, 510, 59 },
		{ 327, 511, 59 },
		{ 327, 512, 59 },
		{ 328, 400, 59 },
		{ 328, 401, 59 },
		{ 328, 402, 59 },
		{ 328, 403, 59 },
		{ 328, 404, 59 },
		{ 328, 405, 59 },
		{ 328, 406, 59 },
		{ 328, 407, 59 },
		{ 328, 408, 59 },
		{ 328, 409, 59 },
		{ 328, 410, 59 },
		{ 328, 411, 59 },
		{ 328, 412, 59 },
		{ 328, 413, 59 },
		{ 328, 414, 59 },
		{ 328, 415, 59 },
		{ 328, 416, 59 },
		{ 328, 417, 59 },
		{ 328, 418, 59 },
		{ 328, 419, 59 },
		{ 328, 420, 59 },
		{ 328, 421, 59 },
		{ 328, 422, 59 },
		{ 328, 423, 59 },
		{ 328, 424, 59 },
		{ 328, 425, 59 },
		{ 328, 426, 59 },
		{ 328, 427, 59 },
		{ 328, 428, 59 },
		{ 328, 429, 59 },
		{ 328, 430, 59 },
		{ 328, 431, 59 },
		{ 328, 432, 59 },
		{ 328, 433, 59 },
		{ 328, 434, 59 },
		{ 328, 435, 59 },
		{ 328, 436, 59 },
		{ 328, 437, 59 },
		{ 328, 438, 59 },
		{ 328, 439, 59 },
		{ 328, 440, 59 },
		{ 328, 441, 59 },
		{ 328, 442, 59 },
		{ 328, 443, 59 },
		{ 328, 444, 59 },
		{ 328, 445, 59 },
		{ 328, 446, 59 },
		{ 328, 447, 59 },
		{ 328, 448, 59 },
		{ 328, 449, 59 },
		{ 328, 450, 59 },
		{ 328, 451, 59 },
		{ 328, 452, 59 },
		{ 328, 453, 59 },
		{ 328, 454, 59 },
		{ 328, 455, 59 },
		{ 328, 456, 59 },
		{ 328, 457, 59 },
		{ 328, 458, 59 },
		{ 328, 459, 59 },
		{ 328, 460, 59 },
		{ 328, 461, 59 },
		{ 328, 462, 59 },
		{ 328, 463, 59 },
		{ 328, 464, 59 },
		{ 328, 465, 59 },
		{ 328, 466, 59 },
		{ 328, 467, 59 },
		{ 328, 468, 59 },
		{ 328, 469, 59 },
		{ 328, 470, 59 },
		{ 328, 471, 59 },
		{ 328, 472, 59 },
		{ 328, 473, 59 },
		{ 328, 474, 59 },
		{ 328, 475, 59 },
		{ 328, 476, 59 },
		{ 328, 477, 59 },
		{ 328, 478, 59 },
		{ 328, 479, 59 },
		{ 328, 480, 59 },
		{ 328, 481, 59 },
		{ 328, 482, 59 },
		{ 328, 483, 59 },
		{ 328, 484, 59 },
		{ 328, 485, 59 },
		{ 328, 486, 59 },
		{ 328, 487, 59 },
		{ 328, 488, 59 },
		{ 328, 489, 59 },
		{ 328, 490, 59 },
		{ 328, 491, 59 },
		{ 328, 492, 59 },
		{ 328, 493, 59 },
		{ 328, 494, 59 },
		{ 328, 495, 59 },
		{ 328, 496, 59 },
		{ 328, 497, 59 },
		{ 328, 498, 59 },
		{ 328, 499, 59 },
		{ 328, 500, 59 },
		{ 328, 501, 59 },
		{ 328, 502, 59 },
		{ 328, 503, 59 },
		{ 328, 504, 59 },
		{ 328, 505, 59 },
		{ 328, 506, 59 },
		{ 328, 507, 59 },
		{ 328, 508, 59 },
		{ 328, 509, 59 },
		{ 328, 510, 59 },
		{ 328, 511, 59 },
		{ 328, 512, 59 },
		{ 329, 402, 59 },
		{ 329, 403, 59 },
		{ 329, 404, 59 },
		{ 329, 405, 59 },
		{ 329, 406, 59 },
		{ 329, 407, 59 },
		{ 329, 408, 59 },
		{ 329, 409, 59 },
		{ 329, 410, 59 },
		{ 329, 411, 59 },
		{ 329, 412, 59 },
		{ 329, 413, 59 },
		{ 329, 414, 59 },
		{ 329, 415, 59 },
		{ 329, 416, 59 },
		{ 329, 417, 59 },
		{ 329, 418, 59 },
		{ 329, 419, 59 },
		{ 329, 420, 59 },
		{ 329, 421, 59 },
		{ 329, 422, 59 },
		{ 329, 423, 59 },
		{ 329, 424, 59 },
		{ 329, 425, 59 },
		{ 329, 426, 59 },
		{ 329, 427, 59 },
		{ 329, 428, 59 },
		{ 329, 429, 59 },
		{ 329, 430, 59 },
		{ 329, 431, 59 },
		{ 329, 432, 59 },
		{ 329, 433, 59 },
		{ 329, 434, 59 },
		{ 329, 435, 59 },
		{ 329, 436, 59 },
		{ 329, 437, 59 },
		{ 329, 438, 59 },
		{ 329, 439, 59 },
		{ 329, 440, 59 },
		{ 329, 441, 59 },
		{ 329, 442, 59 },
		{ 329, 443, 59 },
		{ 329, 444, 59 },
		{ 329, 445, 59 },
		{ 329, 446, 59 },
		{ 329, 447, 59 },
		{ 329, 448, 59 },
		{ 329, 449, 59 },
		{ 329, 450, 59 },
		{ 329, 451, 59 },
		{ 329, 452, 59 },
		{ 329, 453, 59 },
		{ 329, 454, 59 },
		{ 329, 455, 59 },
		{ 329, 456, 59 },
		{ 329, 457, 59 },
		{ 329, 458, 59 },
		{ 329, 459, 59 },
		{ 329, 460, 59 },
		{ 329, 461, 59 },
		{ 329, 462, 59 },
		{ 329, 463, 59 },
		{ 329, 464, 59 },
		{ 329, 465, 59 },
		{ 329, 466, 59 },
		{ 329, 467, 59 },
		{ 329, 468, 59 },
		{ 329, 469, 59 },
		{ 329, 470, 59 },
		{ 329, 471, 59 },
		{ 329, 472, 59 },
		{ 329, 473, 59 },
		{ 329, 474, 59 },
		{ 329, 475, 59 },
		{ 329, 476, 59 },
		{ 329, 477, 59 },
		{ 329, 478, 59 },
		{ 329, 479, 59 },
		{ 329, 480, 59 },
		{ 329, 481, 59 },
		{ 329, 482, 59 },
		{ 329, 483, 59 },
		{ 329, 484, 59 },
		{ 329, 485, 59 },
		{ 329, 486, 59 },
		{ 329, 487, 59 },
		{ 329, 488, 59 },
		{ 329, 489, 59 },
		{ 329, 490, 59 },
		{ 329, 491, 59 },
		{ 329, 492, 59 },
		{ 329, 493, 59 },
		{ 329, 494, 59 },
		{ 329, 495, 59 },
		{ 329, 496, 59 },
		{ 329, 497, 59 },
		{ 329, 498, 59 },
		{ 329, 499, 59 },
		{ 329, 500, 59 },
		{ 329, 501, 59 },
		{ 329, 502, 59 },
		{ 329, 503, 59 },
		{ 329, 504, 59 },
		{ 329, 505, 59 },
		{ 329, 506, 59 },
		{ 329, 507, 59 },
		{ 329, 508, 59 },
		{ 329, 509, 59 },
		{ 329, 510, 59 },
		{ 329, 511, 59 },
		{ 329, 512, 59 },
		{ 330, 404, 59 },
		{ 330, 405, 59 },
		{ 330, 406, 59 },
		{ 330, 407, 59 },
		{ 330, 408, 59 },
		{ 330, 409, 59 },
		{ 330, 410, 59 },
		{ 330, 411, 59 },
		{ 330, 412, 59 },
		{ 330, 413, 59 },
		{ 330, 414, 59 },
		{ 330, 415, 59 },
		{ 330, 416, 59 },
		{ 330, 417, 59 },
		{ 330, 418, 59 },
		{ 330, 419, 59 },
		{ 330, 420, 59 },
		{ 330, 421, 59 },
		{ 330, 422, 59 },
		{ 330, 423, 59 },
		{ 330, 424, 59 },
		{ 330, 425, 59 },
		{ 330, 426, 59 },
		{ 330, 427, 59 },
		{ 330, 428, 59 },
		{ 330, 429, 59 },
		{ 330, 430, 59 },
		{ 330, 431, 59 },
		{ 330, 432, 59 },
		{ 330, 433, 59 },
		{ 330, 434, 59 },
		{ 330, 435, 59 },
		{ 330, 436, 59 },
		{ 330, 437, 59 },
		{ 330, 438, 59 },
		{ 330, 439, 59 },
		{ 330, 440, 59 },
		{ 330, 441, 59 },
		{ 330, 442, 59 },
		{ 330, 443, 59 },
		{ 330, 444, 59 },
		{ 330, 445, 59 },
		{ 330, 446, 59 },
		{ 330, 447, 59 },
		{ 330, 448, 59 },
		{ 330, 449, 59 },
		{ 330, 450, 59 },
		{ 330, 451, 59 },
		{ 330, 452, 59 },
		{ 330, 453, 59 },
		{ 330, 454, 59 },
		{ 330, 455, 59 },
		{ 330, 456, 59 },
		{ 330, 457, 59 },
		{ 330, 458, 59 },
		{ 330, 459, 59 },
		{ 330, 460, 59 },
		{ 330, 461, 59 },
		{ 330, 462, 59 },
		{ 330, 463, 59 },
		{ 330, 464, 59 },
		{ 330, 465, 59 },
		{ 330, 466, 59 },
		{ 330, 467, 59 },
		{ 330, 468, 59 },
		{ 330, 469, 59 },
		{ 330, 470, 59 },
		{ 330, 471, 59 },
		{ 330, 472, 59 },
		{ 330, 473, 59 },
		{ 330, 474, 59 },
		{ 330, 475, 59 },
		{ 330, 476, 59 },
		{ 330, 477, 59 },
		{ 330, 478, 59 },
		{ 330, 479, 59 },
		{ 330, 480, 59 },
		{ 330, 481, 59 },
		{ 330, 482, 59 },
		{ 330, 483, 59 },
		{ 330, 484, 59 },
		{ 330, 485, 59 },
		{ 330, 486, 59 },
		{ 330, 487, 59 },
		{ 330, 488, 59 },
		{ 330, 489, 59 },
		{ 330, 490, 59 },
		{ 330, 491, 59 },
		{ 330, 492, 59 },
		{ 330, 493, 59 },
		{ 330, 494, 59 },
		{ 330, 495, 59 },
		{ 330, 496, 59 },
		{ 330, 497, 59 },
		{ 330, 498, 59 },
		{ 330, 499, 59 },
		{ 330, 500, 59 },
		{ 330, 501, 59 },
		{ 330, 502, 59 },
		{ 330, 503, 59 },
		{ 330, 504, 59 },
		{ 330, 505, 59 },
		{ 330, 506, 59 },
		{ 330, 507, 59 },
		{ 330, 508, 59 },
		{ 330, 509, 59 },
		{ 330, 510, 59 },
		{ 330, 511, 59 },
		{ 330, 512, 59 },
		{ 331, 406, 59 },
		{ 331, 407, 59 },
		{ 331, 408, 59 },
		{ 331, 409, 59 },
		{ 331, 410, 59 },
		{ 331, 411, 59 },
		{ 331, 412, 59 },
		{ 331, 413, 59 },
		{ 331, 414, 59 },
		{ 331, 415, 59 },
		{ 331, 416, 59 },
		{ 331, 417, 59 },
		{ 331, 418, 59 },
		{ 331, 419, 59 },
		{ 331, 420, 59 },
		{ 331, 421, 59 },
		{ 331, 422, 59 },
		{ 331, 423, 59 },
		{ 331, 424, 59 },
		{ 331, 425, 59 },
		{ 331, 426, 59 },
		{ 331, 427, 59 },
		{ 331, 428, 59 },
		{ 331, 429, 59 },
		{ 331, 430, 59 },
		{ 331, 431, 59 },
		{ 331, 432, 59 },
		{ 331, 433, 59 },
		{ 331, 434, 59 },
		{ 331, 435, 59 },
		{ 331, 436, 59 },
		{ 331, 437, 59 },
		{ 331, 438, 59 },
		{ 331, 439, 59 },
		{ 331, 440, 59 },
		{ 331, 441, 59 },
		{ 331, 442, 59 },
		{ 331, 443, 59 },
		{ 331, 444, 59 },
		{ 331, 445, 59 },
		{ 331, 446, 59 },
		{ 331, 447, 59 },
		{ 331, 448, 59 },
		{ 331, 449, 59 },
		{ 331, 450, 59 },
		{ 331, 451, 59 },
		{ 331, 452, 59 },
		{ 331, 453, 59 },
		{ 331, 454, 59 },
		{ 331, 455, 59 },
		{ 331, 456, 59 },
		{ 331, 457, 59 },
		{ 331, 458, 59 },
		{ 331, 459, 59 },
		{ 331, 460, 59 },
		{ 331, 461, 59 },
		{ 331, 462, 59 },
		{ 331, 463, 59 },
		{ 331, 464, 59 },
		{ 331, 465, 59 },
		{ 331, 466, 59 },
		{ 331, 467, 59 },
		{ 331, 468, 59 },
		{ 331, 469, 59 },
		{ 331, 470, 59 },
		{ 331, 471, 59 },
		{ 331, 472, 59 },
		{ 331, 473, 59 },
		{ 331, 474, 59 },
		{ 331, 475, 59 },
		{ 331, 476, 59 },
		{ 331, 477, 59 },
		{ 331, 478, 59 },
		{ 331, 479, 59 },
		{ 331, 480, 59 },
		{ 331, 481, 59 },
		{ 331, 482, 59 },
		{ 331, 483, 59 },
		{ 331, 484, 59 },
		{ 331, 485, 59 },
		{ 331, 486, 59 },
		{ 331, 487, 59 },
		{ 331, 488, 59 },
		{ 331, 489, 59 },
		{ 331, 490, 59 },
		{ 331, 491, 59 },
		{ 331, 492, 59 },
		{ 331, 493, 59 },
		{ 331, 494, 59 },
		{ 331, 495, 59 },
		{ 331, 496, 59 },
		{ 331, 497, 59 },
		{ 331, 498, 59 },
		{ 331, 499, 59 },
		{ 331, 500, 59 },
		{ 331, 501, 59 },
		{ 331, 502, 59 },
		{ 331, 503, 59 },
		{ 331, 504, 59 },
		{ 331, 505, 59 },
		{ 331, 506, 59 },
		{ 331, 507, 59 },
		{ 331, 508, 59 },
		{ 331, 509, 59 },
		{ 331, 510, 59 },
		{ 331, 511, 59 },
		{ 331, 512, 59 },
		{ 332, 408, 59 },
		{ 332, 409, 59 },
		{ 332, 410, 59 },
		{ 332, 411, 59 },
		{ 332, 412, 59 },
		{ 332, 413, 59 },
		{ 332, 414, 59 },
		{ 332, 415, 59 },
		{ 332, 416, 59 },
		{ 332, 417, 59 },
		{ 332, 418, 59 },
		{ 332, 419, 59 },
		{ 332, 420, 59 },
		{ 332, 421, 59 },
		{ 332, 422, 59 },
		{ 332, 423, 59 },
		{ 332, 424, 59 },
		{ 332, 425, 59 },
		{ 332, 426, 59 },
		{ 332, 427, 59 },
		{ 332, 428, 59 },
		{ 332, 429, 59 },
		{ 332, 430, 59 },
		{ 332, 431, 59 },
		{ 332, 432, 59 },
		{ 332, 433, 59 },
		{ 332, 434, 59 },
		{ 332, 435, 59 },
		{ 332, 436, 59 },
		{ 332, 437, 59 },
		{ 332, 438, 59 },
		{ 332, 439, 59 },
		{ 332, 440, 59 },
		{ 332, 441, 59 },
		{ 332, 442, 59 },
		{ 332, 443, 59 },
		{ 332, 444, 59 },
		{ 332, 445, 59 },
		{ 332, 446, 59 },
		{ 332, 447, 59 },
		{ 332, 448, 59 },
		{ 332, 449, 59 },
		{ 332, 450, 59 },
		{ 332, 451, 59 },
		{ 332, 452, 59 },
		{ 332, 453, 59 },
		{ 332, 454, 59 },
		{ 332, 455, 59 },
		{ 332, 456, 59 },
		{ 332, 457, 59 },
		{ 332, 458, 59 },
		{ 332, 459, 59 },
		{ 332, 460, 59 },
		{ 332, 461, 59 },
		{ 332, 462, 59 },
		{ 332, 463, 59 },
		{ 332, 464, 59 },
		{ 332, 465, 59 },
		{ 332, 466, 59 },
		{ 332, 467, 59 },
		{ 332, 468, 59 },
		{ 332, 469, 59 },
		{ 332, 470, 59 },
		{ 332, 471, 59 },
		{ 332, 472, 59 },
		{ 332, 473, 59 },
		{ 332, 474, 59 },
		{ 332, 475, 59 },
		{ 332, 476, 59 },
		{ 332, 477, 59 },
		{ 332, 478, 59 },
		{ 332, 479, 59 },
		{ 332, 480, 59 },
		{ 332, 481, 59 },
		{ 332, 482, 59 },
		{ 332, 483, 59 },
		{ 332, 484, 59 },
		{ 332, 485, 59 },
		{ 332, 486, 59 },
		{ 332, 487, 59 },
		{ 332, 488, 59 },
		{ 332, 489, 59 },
		{ 332, 490, 59 },
		{ 332, 491, 59 },
		{ 332, 492, 59 },
		{ 332, 493, 59 },
		{ 332, 494, 59 },
		{ 332, 495, 59 },
		{ 332, 496, 59 },
		{ 332, 497, 59 },
		{ 332, 498, 59 },
		{ 332, 499, 59 },
		{ 332, 500, 59 },
		{ 332, 501, 59 },
		{ 332, 502, 59 },
		{ 332, 503, 59 },
		{ 332, 504, 59 },
		{ 332, 505, 59 },
		{ 332, 506, 59 },
		{ 332, 507, 59 },
		{ 332, 508, 59 },
		{ 332, 509, 59 },
		{ 332, 510, 59 },
		{ 332, 511, 59 },
		{ 332, 512, 59 },
		{ 333, 410, 59 },
		{ 333, 411, 59 },
		{ 333, 412, 59 },
		{ 333, 413, 59 },
		{ 333, 414, 59 },
		{ 333, 415, 59 },
		{ 333, 416, 59 },
		{ 333, 417, 59 },
		{ 333, 418, 59 },
		{ 333, 419, 59 },
		{ 333, 420, 59 },
		{ 333, 421, 59 },
		{ 333, 422, 59 },
		{ 333, 423, 59 },
		{ 333, 424, 59 },
		{ 333, 425, 59 },
		{ 333, 426, 59 },
		{ 333, 427, 59 },
		{ 333, 428, 59 },
		{ 333, 429, 59 },
		{ 333, 430, 59 },
		{ 333, 431, 59 },
		{ 333, 432, 59 },
		{ 333, 433, 59 },
		{ 333, 434, 59 },
		{ 333, 435, 59 },
		{ 333, 436, 59 },
		{ 333, 437, 59 },
		{ 333, 438, 59 },
		{ 333, 439, 59 },
		{ 333, 440, 59 },
		{ 333, 441, 59 },
		{ 333, 442, 59 },
		{ 333, 443, 59 },
		{ 333, 444, 59 },
		{ 333, 445, 59 },
		{ 333, 446, 59 },
		{ 333, 447, 59 },
		{ 333, 448, 59 },
		{ 333, 449, 59 },
		{ 333, 450, 59 },
		{ 333, 451, 59 },
		{ 333, 452, 59 },
		{ 333, 453, 59 },
		{ 333, 454, 59 },
		{ 333, 455, 59 },
		{ 333, 456, 59 },
		{ 333, 457, 59 },
		{ 333, 458, 59 },
		{ 333, 459, 59 },
		{ 333, 460, 59 },
		{ 333, 461, 59 },
		{ 333, 462, 59 },
		{ 333, 463, 59 },
		{ 333, 464, 59 },
		{ 333, 465, 59 },
		{ 333, 466, 59 },
		{ 333, 467, 59 },
		{ 333, 468, 59 },
		{ 333, 469, 59 },
		{ 333, 470, 59 },
		{ 333, 471, 59 },
		{ 333, 472, 59 },
		{ 333, 473, 59 },
		{ 333, 474, 59 },
		{ 333, 475, 59 },
		{ 333, 476, 59 },
		{ 333, 477, 59 },
		{ 333, 478, 59 },
		{ 333, 479, 59 },
		{ 333, 480, 59 },
		{ 333, 481, 59 },
		{ 333, 482, 59 },
		{ 333, 483, 59 },
		{ 333, 484, 59 },
		{ 333, 485, 59 },
		{ 333, 486, 59 },
		{ 333, 487, 59 },
		{ 333, 488, 59 },
		{ 333, 489, 59 },
		{ 333, 490, 59 },
		{ 333, 491, 59 },
		{ 333, 492, 59 },
		{ 333, 493, 59 },
		{ 333, 494, 59 },
		{ 333, 495, 59 },
		{ 333, 496, 59 },
		{ 333, 497, 59 },
		{ 333, 498, 59 },
		{ 333, 499, 59 },
		{ 333, 500, 59 },
		{ 333, 501, 59 },
		{ 333, 502, 59 },
		{ 333, 503, 59 },
		{ 333, 504, 59 },
		{ 333, 505, 59 },
		{ 333, 506, 59 },
		{ 333, 507, 59 },
		{ 333, 508, 59 },
		{ 333, 509, 59 },
		{ 333, 510, 59 },
		{ 333, 511, 59 },
		{ 333, 512, 59 },
		{ 334, 412, 59 },
		{ 334, 413, 59 },
		{ 334, 414, 59 },
		{ 334, 415, 59 },
		{ 334, 416, 59 },
		{ 334, 417, 59 },
		{ 334, 418, 59 },
		{ 334, 419, 59 },
		{ 334, 420, 59 },
		{ 334, 421, 59 },
		{ 334, 422, 59 },
		{ 334, 423, 59 },
		{ 334, 424, 59 },
		{ 334, 425, 59 },
		{ 334, 426, 59 },
		{ 334, 427, 59 },
		{ 334, 428, 59 },
		{ 334, 429, 59 },
		{ 334, 430, 59 },
		{ 334, 431, 59 },
		{ 334, 432, 59 },
		{ 334, 433, 59 },
		{ 334, 434, 59 },
		{ 334, 435, 59 },
		{ 334, 436, 59 },
		{ 334, 437, 59 },
		{ 334, 438, 59 },
		{ 334, 439, 59 },
		{ 334, 440, 59 },
		{ 334, 441, 59 },
		{ 334, 442, 59 },
		{ 334, 443, 59 },
		{ 334, 444, 59 },
		{ 334, 445, 59 },
		{ 334, 446, 59 },
		{ 334, 447, 59 },
		{ 334, 448, 59 },
		{ 334, 449, 59 },
		{ 334, 450, 59 },
		{ 334, 451, 59 },
		{ 334, 452, 59 },
		{ 334, 453, 59 },
		{ 334, 454, 59 },
		{ 334, 455, 59 },
		{ 334, 456, 59 },
		{ 334, 457, 59 },
		{ 334, 458, 59 },
		{ 334, 459, 59 },
		{ 334, 460, 59 },
		{ 334, 461, 59 },
		{ 334, 462, 59 },
		{ 334, 463, 59 },
		{ 334, 464, 59 },
		{ 334, 465, 59 },
		{ 334, 466, 59 },
		{ 334, 467, 59 },
		{ 334, 468, 59 },
		{ 334, 469, 59 },
		{ 334, 470, 59 },
		{ 334, 471, 59 },
		{ 334, 472, 59 },
		{ 334, 473, 59 },
		{ 334, 474, 59 },
		{ 334, 475, 59 },
		{ 334, 476, 59 },
		{ 334, 477, 59 },
		{ 334, 478, 59 },
		{ 334, 479, 59 },
		{ 334, 480, 59 },
		{ 334, 481, 59 },
		{ 334, 482, 59 },
		{ 334, 483, 59 },
		{ 334, 484, 59 },
		{ 334, 485, 59 },
		{ 334, 486, 59 },
		{ 334, 487, 59 },
		{ 334, 488, 59 },
		{ 334, 489, 59 },
		{ 334, 490, 59 },
		{ 334, 491, 59 },
		{ 334, 492, 59 },
		{ 334, 493, 59 },
		{ 334, 494, 59 },
		{ 334, 495, 59 },
		{ 334, 496, 59 },
		{ 334, 497, 59 },
		{ 334, 498, 59 },
		{ 334, 499, 59 },
		{ 334, 500, 59 },
		{ 334, 501, 59 },
		{ 334, 502, 59 },
		{ 334, 503, 59 },
		{ 334, 504, 59 },
		{ 334, 505, 59 },
		{ 334, 506, 59 },
		{ 334, 507, 59 },
		{ 334, 508, 59 },
		{ 334, 509, 59 },
		{ 334, 510, 59 },
		{ 334, 511, 59 },
		{ 334, 512, 59 },
		{ 335, 414, 59 },
		{ 335, 415, 59 },
		{ 335, 416, 59 },
		{ 335, 417, 59 },
		{ 335, 418, 59 },
		{ 335, 419, 59 },
		{ 335, 420, 59 },
		{ 335, 421, 59 },
		{ 335, 422, 59 },
		{ 335, 423, 59 },
		{ 335, 424, 59 },
		{ 335, 425, 59 },
		{ 335, 426, 59 },
		{ 335, 427, 59 },
		{ 335, 428, 59 },
		{ 335, 429, 59 },
		{ 335, 430, 59 },
		{ 335, 431, 59 },
		{ 335, 432, 59 },
		{ 335, 433, 59 },
		{ 335, 434, 59 },
		{ 335, 435, 59 },
		{ 335, 436, 59 },
		{ 335, 437, 59 },
		{ 335, 438, 59 },
		{ 335, 439, 59 },
		{ 335, 440, 59 },
		{ 335, 441, 59 },
		{ 335, 442, 59 },
		{ 335, 443, 59 },
		{ 335, 444, 59 },
		{ 335, 445, 59 },
		{ 335, 446, 59 },
		{ 335, 447, 59 },
		{ 335, 448, 59 },
		{ 335, 449, 59 },
		{ 335, 450, 59 },
		{ 335, 451, 59 },
		{ 335, 452, 59 },
		{ 335, 453, 59 },
		{ 335, 454, 59 },
		{ 335, 455, 59 },
		{ 335, 456, 59 },
		{ 335, 457, 59 },
		{ 335, 458, 59 },
		{ 335, 459, 59 },
		{ 335, 460, 59 },
		{ 335, 461, 59 },
		{ 335, 462, 59 },
		{ 335, 463, 59 },
		{ 335, 464, 59 },
		{ 335, 465, 59 },
		{ 335, 466, 59 },
		{ 335, 467, 59 },
		{ 335, 468, 59 },
		{ 335, 469, 59 },
		{ 335, 470, 59 },
		{ 335, 471, 59 },
		{ 335, 472, 59 },
		{ 335, 473, 59 },
		{ 335, 474, 59 },
		{ 335, 475, 59 },
		{ 335, 476, 59 },
		{ 335, 477, 59 },
		{ 335, 478, 59 },
		{ 335, 479, 59 },
		{ 335, 480, 59 },
		{ 335, 481, 59 },
		{ 335, 482, 59 },
		{ 335, 483, 59 },
		{ 335, 484, 59 },
		{ 335, 485, 59 },
		{ 335, 486, 59 },
		{ 335, 487, 59 },
		{ 335, 488, 59 },
		{ 335, 489, 59 },
		{ 335, 490, 59 },
		{ 335, 491, 59 },
		{ 335, 492, 59 },
		{ 335, 493, 59 },
		{ 335, 494, 59 },
		{ 335, 495, 59 },
		{ 335, 496, 59 },
		{ 335, 497, 59 },
		{ 335, 498, 59 },
		{ 335, 499, 59 },
		{ 335, 500, 59 },
		{ 335, 501, 59 },
		{ 335, 502, 59 },
		{ 335, 503, 59 },
		{ 335, 504, 59 },
		{ 335, 505, 59 },
		{ 335, 506, 59 },
		{ 335, 507, 59 },
		{ 335, 508, 59 },
		{ 335, 509, 59 },
		{ 335, 510, 59 },
		{ 335, 511, 59 },
		{ 335, 512, 59 },
		{ 336, 416, 59 },
		{ 336, 417, 59 },
		{ 336, 418, 59 },
		{ 336, 419, 59 },
		{ 336, 420, 59 },
		{ 336, 421, 59 },
		{ 336, 422, 59 },
		{ 336, 423, 59 },
		{ 336, 424, 59 },
		{ 336, 425, 59 },
		{ 336, 426, 59 },
		{ 336, 427, 59 },
		{ 336, 428, 59 },
		{ 336, 429, 59 },
		{ 336, 430, 59 },
		{ 336, 431, 59 },
		{ 336, 432, 59 },
		{ 336, 433, 59 },
		{ 336, 434, 59 },
		{ 336, 435, 59 },
		{ 336, 436, 59 },
		{ 336, 437, 59 },
		{ 336, 438, 59 },
		{ 336, 439, 59 },
		{ 336, 440, 59 },
		{ 336, 441, 59 },
		{ 336, 442, 59 },
		{ 336, 443, 59 },
		{ 336, 444, 59 },
		{ 336, 445, 59 },
		{ 336, 446, 59 },
		{ 336, 447, 59 },
		{ 336, 448, 59 },
		{ 336, 449, 59 },
		{ 336, 450, 59 },
		{ 336, 451, 59 },
		{ 336, 452, 59 },
		{ 336, 453, 59 },
		{ 336, 454, 59 },
		{ 336, 455, 59 },
		{ 336, 456, 59 },
		{ 336, 457, 59 },
		{ 336, 458, 59 },
		{ 336, 459, 59 },
		{ 336, 460, 59 },
		{ 336, 461, 59 },
		{ 336, 462, 59 },
		{ 336, 463, 59 },
		{ 336, 464, 59 },
		{ 336, 465, 59 },
		{ 336, 466, 59 },
		{ 336, 467, 59 },
		{ 336, 468, 59 },
		{ 336, 469, 59 },
		{ 336, 470, 59 },
		{ 336, 471, 59 },
		{ 336, 472, 59 },
		{ 336, 473, 59 },
		{ 336, 474, 59 },
		{ 336, 475, 59 },
		{ 336, 476, 59 },
		{ 336, 477, 59 },
		{ 336, 478, 59 },
		{ 336, 479, 59 },
		{ 336, 480, 59 },
		{ 336, 481, 59 },
		{ 336, 482, 59 },
		{ 336, 483, 59 },
		{ 336, 484, 59 },
		{ 336, 485, 59 },
		{ 336, 486, 59 },
		{ 336, 487, 59 },
		{ 336, 488, 59 },
		{ 336, 489, 59 },
		{ 336, 490, 59 },
		{ 336, 491, 59 },
		{ 336, 492, 59 },
		{ 336, 493, 59 },
		{ 336, 494, 59 },
		{ 336, 495, 59 },
		{ 336, 496, 59 },
		{ 336, 497, 59 },
		{ 336, 498, 59 },
		{ 336, 499, 59 },
		{ 336, 500, 59 },
		{ 336, 501, 59 },
		{ 336, 502, 59 },
		{ 336, 503, 59 },
		{ 336, 504, 59 },
		{ 336, 505, 59 },
		{ 336, 506, 59 },
		{ 336, 507, 59 },
		{ 336, 508, 59 },
		{ 336, 509, 59 },
		{ 336, 510, 59 },
		{ 336, 511, 59 },
		{ 336, 512, 59 },
		{ 337, 418, 59 },
		{ 337, 419, 59 },
		{ 337, 420, 59 },
		{ 337, 421, 59 },
		{ 337, 422, 59 },
		{ 337, 423, 59 },
		{ 337, 424, 59 },
		{ 337, 425, 59 },
		{ 337, 426, 59 },
		{ 337, 427, 59 },
		{ 337, 428, 59 },
		{ 337, 429, 59 },
		{ 337, 430, 59 },
		{ 337, 431, 59 },
		{ 337, 432, 59 },
		{ 337, 433, 59 },
		{ 337, 434, 59 },
		{ 337, 435, 59 },
		{ 337, 436, 59 },
		{ 337, 437, 59 },
		{ 337, 438, 59 },
		{ 337, 439, 59 },
		{ 337, 440, 59 },
		{ 337, 441, 59 },
		{ 337, 442, 59 },
		{ 337, 443, 59 },
		{ 337, 444, 59 },
		{ 337, 445, 59 },
		{ 337, 446, 59 },
		{ 337, 447, 59 },
		{ 337, 448, 59 },
		{ 337, 449, 59 },
		{ 337, 450, 59 },
		{ 337, 451, 59 },
		{ 337, 452, 59 },
		{ 337, 453, 59 },
		{ 337, 454, 59 },
		{ 337, 455, 59 },
		{ 337, 456, 59 },
		{ 337, 457, 59 },
		{ 337, 458, 59 },
		{ 337, 459, 59 },
		{ 337, 460, 59 },
		{ 337, 461, 59 },
		{ 337, 462, 59 },
		{ 337, 463, 59 },
		{ 337, 464, 59 },
		{ 337, 465, 59 },
		{ 337, 466, 59 },
		{ 337, 467, 59 },
		{ 337, 468, 59 },
		{ 337, 469, 59 },
		{ 337, 470, 59 },
		{ 337, 471, 59 },
		{ 337, 472, 59 },
		{ 337, 473, 59 },
		{ 337, 474, 59 },
		{ 337, 475, 59 },
		{ 337, 476, 59 },
		{ 337, 477, 59 },
		{ 337, 478, 59 },
		{ 337, 479, 59 },
		{ 337, 480, 59 },
		{ 337, 481, 59 },
		{ 337, 482, 59 },
		{ 337, 483, 59 },
		{ 337, 484, 59 },
		{ 337, 485, 59 },
		{ 337, 486, 59 },
		{ 337, 487, 59 },
		{ 337, 488, 59 },
		{ 337, 489, 59 },
		{ 337, 490, 59 },
		{ 337, 491, 59 },
		{ 337, 492, 59 },
		{ 337, 493, 59 },
		{ 337, 494, 59 },
		{ 337, 495, 59 },
		{ 337, 496, 59 },
		{ 337, 497, 59 },
		{ 337, 498, 59 },
		{ 337, 499, 59 },
		{ 337, 500, 59 },
		{ 337, 501, 59 },
		{ 337, 502, 59 },
		{ 337, 503, 59 },
		{ 337, 504, 59 },
		{ 337, 505, 59 },
		{ 337, 506, 59 },
		{ 337, 507, 59 },
		{ 337, 508, 59 },
		{ 337, 509, 59 },
		{ 337, 510, 59 },
		{ 337, 511, 59 },
		{ 337, 512, 59 },
		{ 338, 420, 59 },
		{ 338, 421, 59 },
		{ 338, 422, 59 },
		{ 338, 423, 59 },
		{ 338, 424, 59 },
		{ 338, 425, 59 },
		{ 338, 426, 59 },
		{ 338, 427, 59 },
		{ 338, 428, 59 },
		{ 338, 429, 59 },
		{ 338, 430, 59 },
		{ 338, 431, 59 },
		{ 338, 432, 59 },
		{ 338, 433, 59 },
		{ 338, 434, 59 },
		{ 338, 435, 59 },
		{ 338, 436, 59 },
		{ 338, 437, 59 },
		{ 338, 438, 59 },
		{ 338, 439, 59 },
		{ 338, 440, 59 },
		{ 338, 441, 59 },
		{ 338, 442, 59 },
		{ 338, 443, 59 },
		{ 338, 444, 59 },
		{ 338, 445, 59 },
		{ 338, 446, 59 },
		{ 338, 447, 59 },
		{ 338, 448, 59 },
		{ 338, 449, 59 },
		{ 338, 450, 59 },
		{ 338, 451, 59 },
		{ 338, 452, 59 },
		{ 338, 453, 59 },
		{ 338, 454, 59 },
		{ 338, 455, 59 },
		{ 338, 456, 59 },
		{ 338, 457, 59 },
		{ 338, 458, 59 },
		{ 338, 459, 59 },
		{ 338, 460, 59 },
		{ 338, 461, 59 },
		{ 338, 462, 59 },
		{ 338, 463, 59 },
		{ 338, 464, 59 },
		{ 338, 465, 59 },
		{ 338, 466, 59 },
		{ 338, 467, 59 },
		{ 338, 468, 59 },
		{ 338, 469, 59 },
		{ 338, 470, 59 },
		{ 338, 471, 59 },
		{ 338, 472, 59 },
		{ 338, 473, 59 },
		{ 338, 474, 59 },
		{ 338, 475, 59 },
		{ 338, 476, 59 },
		{ 338, 477, 59 },
		{ 338, 478, 59 },
		{ 338, 479, 59 },
		{ 338, 480, 59 },
		{ 338, 481, 59 },
		{ 338, 482, 59 },
		{ 338, 483, 59 },
		{ 338, 484, 59 },
		{ 338, 485, 59 },
		{ 338, 486, 59 },
		{ 338, 487, 59 },
		{ 338, 488, 59 },
		{ 338, 489, 59 },
		{ 338, 490, 59 },
		{ 338, 491, 59 },
		{ 338, 492, 59 },
		{ 338, 493, 59 },
		{ 338, 494, 59 },
		{ 338, 495, 59 },
		{ 338, 496, 59 },
		{ 338, 497, 59 },
		{ 338, 498, 59 },
		{ 338, 499, 59 },
		{ 338, 500, 59 },
		{ 338, 501, 59 },
		{ 338, 502, 59 },
		{ 338, 503, 59 },
		{ 338, 504, 59 },
		{ 338, 505, 59 },
		{ 338, 506, 59 },
		{ 338, 507, 59 },
		{ 338, 508, 59 },
		{ 338, 509, 59 },
		{ 338, 510, 59 },
		{ 338, 511, 59 },
		{ 338, 512, 59 },
		{ 339, 422, 59 },
		{ 339, 423, 59 },
		{ 339, 424, 59 },
		{ 339, 425, 59 },
		{ 339, 426, 59 },
		{ 339, 427, 59 },
		{ 339, 428, 59 },
		{ 339, 429, 59 },
		{ 339, 430, 59 },
		{ 339, 431, 59 },
		{ 339, 432, 59 },
		{ 339, 433, 59 },
		{ 339, 434, 59 },
		{ 339, 435, 59 },
		{ 339, 436, 59 },
		{ 339, 437, 59 },
		{ 339, 438, 59 },
		{ 339, 439, 59 },
		{ 339, 440, 59 },
		{ 339, 441, 59 },
		{ 339, 442, 59 },
		{ 339, 443, 59 },
		{ 339, 444, 59 },
		{ 339, 445, 59 },
		{ 339, 446, 59 },
		{ 339, 447, 59 },
		{ 339, 448, 59 },
		{ 339, 449, 59 },
		{ 339, 450, 59 },
		{ 339, 451, 59 },
		{ 339, 452, 59 },
		{ 339, 453, 59 },
		{ 339, 454, 59 },
		{ 339, 455, 59 },
		{ 339, 456, 59 },
		{ 339, 457, 59 },
		{ 339, 458, 59 },
		{ 339, 459, 59 },
		{ 339, 460, 59 },
		{ 339, 461, 59 },
		{ 339, 462, 59 },
		{ 339, 463, 59 },
		{ 339, 464, 59 },
		{ 339, 465, 59 },
		{ 339, 466, 59 },
		{ 339, 467, 59 },
		{ 339, 468, 59 },
		{ 339, 469, 59 },
		{ 339, 470, 59 },
		{ 339, 471, 59 },
		{ 339, 472, 59 },
		{ 339, 473, 59 },
		{ 339, 474, 59 },
		{ 339, 475, 59 },
		{ 339, 476, 59 },
		{ 339, 477, 59 },
		{ 339, 478, 59 },
		{ 339, 479, 59 },
		{ 339, 480, 59 },
		{ 339, 481, 59 },
		{ 339, 482, 59 },
		{ 339, 483, 59 },
		{ 339, 484, 59 },
		{ 339, 485, 59 },
		{ 339, 486, 59 },
		{ 339, 487, 59 },
		{ 339, 488, 59 },
		{ 339, 489, 59 },
		{ 339, 490, 59 },
		{ 339, 491, 59 },
		{ 339, 492, 59 },
		{ 339, 493, 59 },
		{ 339, 494, 59 },
		{ 339, 495, 59 },
		{ 339, 496, 59 },
		{ 339, 497, 59 },
		{ 339, 498, 59 },
		{ 339, 499, 59 },
		{ 339, 500, 59 },
		{ 339, 501, 59 },
		{ 339, 502, 59 },
		{ 339, 503, 59 },
		{ 339, 504, 59 },
		{ 339, 505, 59 },
		{ 339, 506, 59 },
		{ 339, 507, 59 },
		{ 339, 508, 59 },
		{ 339, 509, 59 },
		{ 339, 510, 59 },
		{ 339, 511, 59 },
		{ 339, 512, 59 },
		{ 340, 424, 59 },
		{ 340, 425, 59 },
		{ 340, 426, 59 },
		{ 340, 427, 59 },
		{ 340, 428, 59 },
		{ 340, 429, 59 },
		{ 340, 430, 59 },
		{ 340, 431, 59 },
		{ 340, 432, 59 },
		{ 340, 433, 59 },
		{ 340, 434, 59 },
		{ 340, 435, 59 },
		{ 340, 436, 59 },
		{ 340, 437, 59 },
		{ 340, 438, 59 },
		{ 340, 439, 59 },
		{ 340, 440, 59 },
		{ 340, 441, 59 },
		{ 340, 442, 59 },
		{ 340, 443, 59 },
		{ 340, 444, 59 },
		{ 340, 445, 59 },
		{ 340, 446, 59 },
		{ 340, 447, 59 },
		{ 340, 448, 59 },
		{ 340, 449, 59 },
		{ 340, 450, 59 },
		{ 340, 451, 59 },
		{ 340, 452, 59 },
		{ 340, 453, 59 },
		{ 340, 454, 59 },
		{ 340, 455, 59 },
		{ 340, 456, 59 },
		{ 340, 457, 59 },
		{ 340, 458, 59 },
		{ 340, 459, 59 },
		{ 340, 460, 59 },
		{ 340, 461, 59 },
		{ 340, 462, 59 },
		{ 340, 463, 59 },
		{ 340, 464, 59 },
		{ 340, 465, 59 },
		{ 340, 466, 59 },
		{ 340, 467, 59 },
		{ 340, 468, 59 },
		{ 340, 469, 59 },
		{ 340, 470, 59 },
		{ 340, 471, 59 },
		{ 340, 472, 59 },
		{ 340, 473, 59 },
		{ 340, 474, 59 },
		{ 340, 475, 59 },
		{ 340, 476, 59 },
		{ 340, 477, 59 },
		{ 340, 478, 59 },
		{ 340, 479, 59 },
		{ 340, 480, 59 },
		{ 340, 481, 59 },
		{ 340, 482, 59 },
		{ 340, 483, 59 },
		{ 340, 484, 59 },
		{ 340, 485, 59 },
		{ 340, 486, 59 },
		{ 340, 487, 59 },
		{ 340, 488, 59 },
		{ 340, 489, 59 },
		{ 340, 490, 59 },
		{ 340, 491, 59 },
		{ 340, 492, 59 },
		{ 340, 493, 59 },
		{ 340, 494, 59 },
		{ 340, 495, 59 },
		{ 340, 496, 59 },
		{ 340, 497, 59 },
		{ 340, 498, 59 },
		{ 340, 499, 59 },
		{ 340, 500, 59 },
		{ 340, 501, 59 },
		{ 340, 502, 59 },
		{ 340, 503, 59 },
		{ 340, 504, 59 },
		{ 340, 505, 59 },
		{ 340, 506, 59 },
		{ 340, 507, 59 },
		{ 340, 508, 59 },
		{ 340, 509, 59 },
		{ 340, 510, 59 },
		{ 340, 511, 59 },
		{ 340, 512, 59 },
		{ 341, 426, 59 },
		{ 341, 427, 59 },
		{ 341, 428, 59 },
		{ 341, 429, 59 },
		{ 341, 430, 59 },
		{ 341, 431, 59 },
		{ 341, 432, 59 },
		{ 341, 433, 59 },
		{ 341, 434, 59 },
		{ 341, 435, 59 },
		{ 341, 436, 59 },
		{ 341, 437, 59 },
		{ 341, 438, 59 },
		{ 341, 439, 59 },
		{ 341, 440, 59 },
		{ 341, 441, 59 },
		{ 341, 442, 59 },
		{ 341, 443, 59 },
		{ 341, 444, 59 },
		{ 341, 445, 59 },
		{ 341, 446, 59 },
		{ 341, 447, 59 },
		{ 341, 448, 59 },
		{ 341, 449, 59 },
		{ 341, 450, 59 },
		{ 341, 451, 59 },
		{ 341, 452, 59 },
		{ 341, 453, 59 },
		{ 341, 454, 59 },
		{ 341, 455, 59 },
		{ 341, 456, 59 },
		{ 341, 457, 59 },
		{ 341, 458, 59 },
		{ 341, 459, 59 },
		{ 341, 460, 59 },
		{ 341, 461, 59 },
		{ 341, 462, 59 },
		{ 341, 463, 59 },
		{ 341, 464, 59 },
		{ 341, 465, 59 },
		{ 341, 466, 59 },
		{ 341, 467, 59 },
		{ 341, 468, 59 },
		{ 341, 469, 59 },
		{ 341, 470, 59 },
		{ 341, 471, 59 },
		{ 341, 472, 59 },
		{ 341, 473, 59 },
		{ 341, 474, 59 },
		{ 341, 475, 59 },
		{ 341, 476, 59 },
		{ 341, 477, 59 },
		{ 341, 478, 59 },
		{ 341, 479, 59 },
		{ 341, 480, 59 },
		{ 341, 481, 59 },
		{ 341, 482, 59 },
		{ 341, 483, 59 },
		{ 341, 484, 59 },
		{ 341, 485, 59 },
		{ 341, 486, 59 },
		{ 341, 487, 59 },
		{ 341, 488, 59 },
		{ 341, 489, 59 },
		{ 341, 490, 59 },
		{ 341, 491, 59 },
		{ 341, 492, 59 },
		{ 341, 493, 59 },
		{ 341, 494, 59 },
		{ 341, 495, 59 },
		{ 341, 496, 59 },
		{ 341, 497, 59 },
		{ 341, 498, 59 },
		{ 341, 499, 59 },
		{ 341, 500, 59 },
		{ 341, 501, 59 },
		{ 341, 502, 59 },
		{ 341, 503, 59 },
		{ 341, 504, 59 },
		{ 341, 505, 59 },
		{ 341, 506, 59 },
		{ 341, 507, 59 },
		{ 341, 508, 59 },
		{ 341, 509, 59 },
		{ 341, 510, 59 },
		{ 341, 511, 59 },
		{ 341, 512, 59 },
		{ 342, 428, 59 },
		{ 342, 429, 59 },
		{ 342, 430, 59 },
		{ 342, 431, 59 },
		{ 342, 432, 59 },
		{ 342, 433, 59 },
		{ 342, 434, 59 },
		{ 342, 435, 59 },
		{ 342, 436, 59 },
		{ 342, 437, 59 },
		{ 342, 438, 59 },
		{ 342, 439, 59 },
		{ 342, 440, 59 },
		{ 342, 441, 59 },
		{ 342, 442, 59 },
		{ 342, 443, 59 },
		{ 342, 444, 59 },
		{ 342, 445, 59 },
		{ 342, 446, 59 },
		{ 342, 447, 59 },
		{ 342, 448, 59 },
		{ 342, 449, 59 },
		{ 342, 450, 59 },
		{ 342, 451, 59 },
		{ 342, 452, 59 },
		{ 342, 453, 59 },
		{ 342, 454, 59 },
		{ 342, 455, 59 },
		{ 342, 456, 59 },
		{ 342, 457, 59 },
		{ 342, 458, 59 },
		{ 342, 459, 59 },
		{ 342, 460, 59 },
		{ 342, 461, 59 },
		{ 342, 462, 59 },
		{ 342, 463, 59 },
		{ 342, 464, 59 },
		{ 342, 465, 59 },
		{ 342, 466, 59 },
		{ 342, 467, 59 },
		{ 342, 468, 59 },
		{ 342, 469, 59 },
		{ 342, 470, 59 },
		{ 342, 471, 59 },
		{ 342, 472, 59 },
		{ 342, 473, 59 },
		{ 342, 474, 59 },
		{ 342, 475, 59 },
		{ 342, 476, 59 },
		{ 342, 477, 59 },
		{ 342, 478, 59 },
		{ 342, 479, 59 },
		{ 342, 480, 59 },
		{ 342, 481, 59 },
		{ 342, 482, 59 },
		{ 342, 483, 59 },
		{ 342, 484, 59 },
		{ 342, 485, 59 },
		{ 342, 486, 59 },
		{ 342, 487, 59 },
		{ 342, 488, 59 },
		{ 342, 489, 59 },
		{ 342, 490, 59 },
		{ 342, 491, 59 },
		{ 342, 492, 59 },
		{ 342, 493, 59 },
		{ 342, 494, 59 },
		{ 342, 495, 59 },
		{ 342, 496, 59 },
		{ 342, 497, 59 },
		{ 342, 498, 59 },
		{ 342, 499, 59 },
		{ 342, 500, 59 },
		{ 342, 501, 59 },
		{ 342, 502, 59 },
		{ 342, 503, 59 },
		{ 342, 504, 59 },
		{ 342, 505, 59 },
		{ 342, 506, 59 },
		{ 342, 507, 59 },
		{ 342, 508, 59 },
		{ 342, 509, 59 },
		{ 342, 510, 59 },
		{ 342, 511, 59 },
		{ 342, 512, 59 },
		{ 343, 430, 59 },
		{ 343, 431, 59 },
		{ 343, 432, 59 },
		{ 343, 433, 59 },
		{ 343, 434, 59 },
		{ 343, 435, 59 },
		{ 343, 436, 59 },
		{ 343, 437, 59 },
		{ 343, 438, 59 },
		{ 343, 439, 59 },
		{ 343, 440, 59 },
		{ 343, 441, 59 },
		{ 343, 442, 59 },
		{ 343, 443, 59 },
		{ 343, 444, 59 },
		{ 343, 445, 59 },
		{ 343, 446, 59 },
		{ 343, 447, 59 },
		{ 343, 448, 59 },
		{ 343, 449, 59 },
		{ 343, 450, 59 },
		{ 343, 451, 59 },
		{ 343, 452, 59 },
		{ 343, 453, 59 },
		{ 343, 454, 59 },
		{ 343, 455, 59 },
		{ 343, 456, 59 },
		{ 343, 457, 59 },
		{ 343, 458, 59 },
		{ 343, 459, 59 },
		{ 343, 460, 59 },
		{ 343, 461, 59 },
		{ 343, 462, 59 },
		{ 343, 463, 59 },
		{ 343, 464, 59 },
		{ 343, 465, 59 },
		{ 343, 466, 59 },
		{ 343, 467, 59 },
		{ 343, 468, 59 },
		{ 343, 469, 59 },
		{ 343, 470, 59 },
		{ 343, 471, 59 },
		{ 343, 472, 59 },
		{ 343, 473, 59 },
		{ 343, 474, 59 },
		{ 343, 475, 59 },
		{ 343, 476, 59 },
		{ 343, 477, 59 },
		{ 343, 478, 59 },
		{ 343, 479, 59 },
		{ 343, 480, 59 },
		{ 343, 481, 59 },
		{ 343, 482, 59 },
		{ 343, 483, 59 },
		{ 343, 484, 59 },
		{ 343, 485, 59 },
		{ 343, 486, 59 },
		{ 343, 487, 59 },
		{ 343, 488, 59 },
		{ 343, 489, 59 },
		{ 343, 490, 59 },
		{ 343, 491, 59 },
		{ 343, 492, 59 },
		{ 343, 493, 59 },
		{ 343, 494, 59 },
		{ 343, 495, 59 },
		{ 343, 496, 59 },
		{ 343, 497, 59 },
		{ 343, 498, 59 },
		{ 343, 499, 59 },
		{ 343, 500, 59 },
		{ 343, 501, 59 },
		{ 343, 502, 59 },
		{ 343, 503, 59 },
		{ 343, 504, 59 },
		{ 343, 505, 59 },
		{ 343, 506, 59 },
		{ 343, 507, 59 },
		{ 343, 508, 59 },
		{ 343, 509, 59 },
		{ 343, 510, 59 },
		{ 343, 511, 59 },
		{ 343, 512, 59 },
		{ 344, 432, 59 },
		{ 344, 433, 59 },
		{ 344, 434, 59 },
		{ 344, 435, 59 },
		{ 344, 436, 59 },
		{ 344, 437, 59 },
		{ 344, 438, 59 },
		{ 344, 439, 59 },
		{ 344, 440, 59 },
		{ 344, 441, 59 },
		{ 344, 442, 59 },
		{ 344, 443, 59 },
		{ 344, 444, 59 },
		{ 344, 445, 59 },
		{ 344, 446, 59 },
		{ 344, 447, 59 },
		{ 344, 448, 59 },
		{ 344, 449, 59 },
		{ 344, 450, 59 },
		{ 344, 451, 59 },
		{ 344, 452, 59 },
		{ 344, 453, 59 },
		{ 344, 454, 59 },
		{ 344, 455, 59 },
		{ 344, 456, 59 },
		{ 344, 457, 59 },
		{ 344, 458, 59 },
		{ 344, 459, 59 },
		{ 344, 460, 59 },
		{ 344, 461, 59 },
		{ 344, 462, 59 },
		{ 344, 463, 59 },
		{ 344, 464, 59 },
		{ 344, 465, 59 },
		{ 344, 466, 59 },
		{ 344, 467, 59 },
		{ 344, 468, 59 },
		{ 344, 469, 59 },
		{ 344, 470, 59 },
		{ 344, 471, 59 },
		{ 344, 472, 59 },
		{ 344, 473, 59 },
		{ 344, 474, 59 },
		{ 344, 475, 59 },
		{ 344, 476, 59 },
		{ 344, 477, 59 },
		{ 344, 478, 59 },
		{ 344, 479, 59 },
		{ 344, 480, 59 },
		{ 344, 481, 59 },
		{ 344, 482, 59 },
		{ 344, 483, 59 },
		{ 344, 484, 59 },
		{ 344, 485, 59 },
		{ 344, 486, 59 },
		{ 344, 487, 59 },
		{ 344, 488, 59 },
		{ 344, 489, 59 },
		{ 344, 490, 59 },
		{ 344, 491, 59 },
		{ 344, 492, 59 },
		{ 344, 493, 59 },
		{ 344, 494, 59 },
		{ 344, 495, 59 },
		{ 344, 496, 59 },
		{ 344, 497, 59 },
		{ 344, 498, 59 },
		{ 344, 499, 59 },
		{ 344, 500, 59 },
		{ 344, 501, 59 },
		{ 344, 502, 59 },
		{ 344, 503, 59 },
		{ 344, 504, 59 },
		{ 344, 505, 59 },
		{ 344, 506, 59 },
		{ 344, 507, 59 },
		{ 344, 508, 59 },
		{ 344, 509, 59 },
		{ 344, 510, 59 },
		{ 344, 511, 59 },
		{ 344, 512, 59 },
		{ 345, 434, 59 },
		{ 345, 435, 59 },
		{ 345, 436, 59 },
		{ 345, 437, 59 },
		{ 345, 438, 59 },
		{ 345, 439, 59 },
		{ 345, 440, 59 },
		{ 345, 441, 59 },
		{ 345, 442, 59 },
		{ 345, 443, 59 },
		{ 345, 444, 59 },
		{ 345, 445, 59 },
		{ 345, 446, 59 },
		{ 345, 447, 59 },
		{ 345, 448, 59 },
		{ 345, 449, 59 },
		{ 345, 450, 59 },
		{ 345, 451, 59 },
		{ 345, 452, 59 },
		{ 345, 453, 59 },
		{ 345, 454, 59 },
		{ 345, 455, 59 },
		{ 345, 456, 59 },
		{ 345, 457, 59 },
		{ 345, 458, 59 },
		{ 345, 459, 59 },
		{ 345, 460, 59 },
		{ 345, 461, 59 },
		{ 345, 462, 59 },
		{ 345, 463, 59 },
		{ 345, 464, 59 },
		{ 345, 465, 59 },
		{ 345, 466, 59 },
		{ 345, 467, 59 },
		{ 345, 468, 59 },
		{ 345, 469, 59 },
		{ 345, 470, 59 },
		{ 345, 471, 59 },
		{ 345, 472, 59 },
		{ 345, 473, 59 },
		{ 345, 474, 59 },
		{ 345, 475, 59 },
		{ 345, 476, 59 },
		{ 345, 477, 59 },
		{ 345, 478, 59 },
		{ 345, 479, 59 },
		{ 345, 480, 59 },
		{ 345, 481, 59 },
		{ 345, 482, 59 },
		{ 345, 483, 59 },
		{ 345, 484, 59 },
		{ 345, 485, 59 },
		{ 345, 486, 59 },
		{ 345, 487, 59 },
		{ 345, 488, 59 },
		{ 345, 489, 59 },
		{ 345, 490, 59 },
		{ 345, 491, 59 },
		{ 345, 492, 59 },
		{ 345, 493, 59 },
		{ 345, 494, 59 },
		{ 345, 495, 59 },
		{ 345, 496, 59 },
		{ 345, 497, 59 },
		{ 345, 498, 59 },
		{ 345, 499, 59 },
		{ 345, 500, 59 },
		{ 345, 501, 59 },
		{ 345, 502, 59 },
		{ 345, 503, 59 },
		{ 345, 504, 59 },
		{ 345, 505, 59 },
		{ 345, 506, 59 },
		{ 345, 507, 59 },
		{ 345, 508, 59 },
		{ 345, 509, 59 },
		{ 345, 510, 59 },
		{ 345, 511, 59 },
		{ 345, 512, 59 },
		{ 346, 436, 59 },
		{ 346, 437, 59 },
		{ 346, 438, 59 },
		{ 346, 439, 59 },
		{ 346, 440, 59 },
		{ 346, 441, 59 },
		{ 346, 442, 59 },
		{ 346, 443, 59 },
		{ 346, 444, 59 },
		{ 346, 445, 59 },
		{ 346, 446, 59 },
		{ 346, 447, 59 },
		{ 346, 448, 59 },
		{ 346, 449, 59 },
		{ 346, 450, 59 },
		{ 346, 451, 59 },
		{ 346, 452, 59 },
		{ 346, 453, 59 },
		{ 346, 454, 59 },
		{ 346, 455, 59 },
		{ 346, 456, 59 },
		{ 346, 457, 59 },
		{ 346, 458, 59 },
		{ 346, 459, 59 },
		{ 346, 460, 59 },
		{ 346, 461, 59 },
		{ 346, 462, 59 },
		{ 346, 463, 59 },
		{ 346, 464, 59 },
		{ 346, 465, 59 },
		{ 346, 466, 59 },
		{ 346, 467, 59 },
		{ 346, 468, 59 },
		{ 346, 469, 59 },
		{ 346, 470, 59 },
		{ 346, 471, 59 },
		{ 346, 472, 59 },
		{ 346, 473, 59 },
		{ 346, 474, 59 },
		{ 346, 475, 59 },
		{ 346, 476, 59 },
		{ 346, 477, 59 },
		{ 346, 478, 59 },
		{ 346, 479, 59 },
		{ 346, 480, 59 },
		{ 346, 481, 59 },
		{ 346, 482, 59 },
		{ 346, 483, 59 },
		{ 346, 484, 59 },
		{ 346, 485, 59 },
		{ 346, 486, 59 },
		{ 346, 487, 59 },
		{ 346, 488, 59 },
		{ 346, 489, 59 },
		{ 346, 490, 59 },
		{ 346, 491, 59 },
		{ 346, 492, 59 },
		{ 346, 493, 59 },
		{ 346, 494, 59 },
		{ 346, 495, 59 },
		{ 346, 496, 59 },
		{ 346, 497, 59 },
		{ 346, 498, 59 },
		{ 346, 499, 59 },
		{ 346, 500, 59 },
		{ 346, 501, 59 },
		{ 346, 502, 59 },
		{ 346, 503, 59 },
		{ 346, 504, 59 },
		{ 346, 505, 59 },
		{ 346, 506, 59 },
		{ 346, 507, 59 },
		{ 346, 508, 59 },
		{ 346, 509, 59 },
		{ 346, 510, 59 },
		{ 346, 511, 59 },
		{ 346, 512, 59 },
		{ 347, 438, 59 },
		{ 347, 439, 59 },
		{ 347, 440, 59 },
		{ 347, 441, 59 },
		{ 347, 442, 59 },
		{ 347, 443, 59 },
		{ 347, 444, 59 },
		{ 347, 445, 59 },
		{ 347, 446, 59 },
		{ 347, 447, 59 },
		{ 347, 448, 59 },
		{ 347, 449, 59 },
		{ 347, 450, 59 },
		{ 347, 451, 59 },
		{ 347, 452, 59 },
		{ 347, 453, 59 },
		{ 347, 454, 59 },
		{ 347, 455, 59 },
		{ 347, 456, 59 },
		{ 347, 457, 59 },
		{ 347, 458, 59 },
		{ 347, 459, 59 },
		{ 347, 460, 59 },
		{ 347, 461, 59 },
		{ 347, 462, 59 },
		{ 347, 463, 59 },
		{ 347, 464, 59 },
		{ 347, 465, 59 },
		{ 347, 466, 59 },
		{ 347, 467, 59 },
		{ 347, 468, 59 },
		{ 347, 469, 59 },
		{ 347, 470, 59 },
		{ 347, 471, 59 },
		{ 347, 472, 59 },
		{ 347, 473, 59 },
		{ 347, 474, 59 },
		{ 347, 475, 59 },
		{ 347, 476, 59 },
		{ 347, 477, 59 },
		{ 347, 478, 59 },
		{ 347, 479, 59 },
		{ 347, 480, 59 },
		{ 347, 481, 59 },
		{ 347, 482, 59 },
		{ 347, 483, 59 },
		{ 347, 484, 59 },
		{ 347, 485, 59 },
		{ 347, 486, 59 },
		{ 347, 487, 59 },
		{ 347, 488, 59 },
		{ 347, 489, 59 },
		{ 347, 490, 59 },
		{ 347, 491, 59 },
		{ 347, 492, 59 },
		{ 347, 493, 59 },
		{ 347, 494, 59 },
		{ 347, 495, 59 },
		{ 347, 496, 59 },
		{ 347, 497, 59 },
		{ 347, 498, 59 },
		{ 347, 499, 59 },
		{ 347, 500, 59 },
		{ 347, 501, 59 },
		{ 347, 502, 59 },
		{ 347, 503, 59 },
		{ 347, 504, 59 },
		{ 347, 505, 59 },
		{ 347, 506, 59 },
		{ 347, 507, 59 },
		{ 347, 508, 59 },
		{ 347, 509, 59 },
		{ 347, 510, 59 },
		{ 347, 511, 59 },
		{ 347, 512, 59 },
		{ 348, 440, 59 },
		{ 348, 441, 59 },
		{ 348, 442, 59 },
		{ 348, 443, 59 },
		{ 348, 444, 59 },
		{ 348, 445, 59 },
		{ 348, 446, 59 },
		{ 348, 447, 59 },
		{ 348, 448, 59 },
		{ 348, 449, 59 },
		{ 348, 450, 59 },
		{ 348, 451, 59 },
		{ 348, 452, 59 },
		{ 348, 453, 59 },
		{ 348, 454, 59 },
		{ 348, 455, 59 },
		{ 348, 456, 59 },
		{ 348, 457, 59 },
		{ 348, 458, 59 },
		{ 348, 459, 59 },
		{ 348, 460, 59 },
		{ 348, 461, 59 },
		{ 348, 462, 59 },
		{ 348, 463, 59 },
		{ 348, 464, 59 },
		{ 348, 465, 59 },
		{ 348, 466, 59 },
		{ 348, 467, 59 },
		{ 348, 468, 59 },
		{ 348, 469, 59 },
		{ 348, 470, 59 },
		{ 348, 471, 59 },
		{ 348, 472, 59 },
		{ 348, 473, 59 },
		{ 348, 474, 59 },
		{ 348, 475, 59 },
		{ 348, 476, 59 },
		{ 348, 477, 59 },
		{ 348, 478, 59 },
		{ 348, 479, 59 },
		{ 348, 480, 59 },
		{ 348, 481, 59 },
		{ 348, 482, 59 },
		{ 348, 483, 59 },
		{ 348, 484, 59 },
		{ 348, 485, 59 },
		{ 348, 486, 59 },
		{ 348, 487, 59 },
		{ 348, 488, 59 },
		{ 348, 489, 59 },
		{ 348, 490, 59 },
		{ 348, 491, 59 },
		{ 348, 492, 59 },
		{ 348, 493, 59 },
		{ 348, 494, 59 },
		{ 348, 495, 59 },
		{ 348, 496, 59 },
		{ 348, 497, 59 },
		{ 348, 498, 59 },
		{ 348, 499, 59 },
		{ 348, 500, 59 },
		{ 348, 501, 59 },
		{ 348, 502, 59 },
		{ 348, 503, 59 },
		{ 348, 504, 59 },
		{ 348, 505, 59 },
		{ 348, 506, 59 },
		{ 348, 507, 59 },
		{ 348, 508, 59 },
		{ 348, 509, 59 },
		{ 348, 510, 59 },
		{ 348, 511, 59 },
		{ 348, 512, 59 },
		{ 349, 442, 59 },
		{ 349, 443, 59 },
		{ 349, 444, 59 },
		{ 349, 445, 59 },
		{ 349, 446, 59 },
		{ 349, 447, 59 },
		{ 349, 448, 59 },
		{ 349, 449, 59 },
		{ 349, 450, 59 },
		{ 349, 451, 59 },
		{ 349, 452, 59 },
		{ 349, 453, 59 },
		{ 349, 454, 59 },
		{ 349, 455, 59 },
		{ 349, 456, 59 },
		{ 349, 457, 59 },
		{ 349, 458, 59 },
		{ 349, 459, 59 },
		{ 349, 460, 59 },
		{ 349, 461, 59 },
		{ 349, 462, 59 },
		{ 349, 463, 59 },
		{ 349, 464, 59 },
		{ 349, 465, 59 },
		{ 349, 466, 59 },
		{ 349, 467, 59 },
		{ 349, 468, 59 },
		{ 349, 469, 59 },
		{ 349, 470, 59 },
		{ 349, 471, 59 },
		{ 349, 472, 59 },
		{ 349, 473, 59 },
		{ 349, 474, 59 },
		{ 349, 475, 59 },
		{ 349, 476, 59 },
		{ 349, 477, 59 },
		{ 349, 478, 59 },
		{ 349, 479, 59 },
		{ 349, 480, 59 },
		{ 349, 481, 59 },
		{ 349, 482, 59 },
		{ 349, 483, 59 },
		{ 349, 484, 59 },
		{ 349, 485, 59 },
		{ 349, 486, 59 },
		{ 349, 487, 59 },
		{ 349, 488, 59 },
		{ 349, 489, 59 },
		{ 349, 490, 59 },
		{ 349, 491, 59 },
		{ 349, 492, 59 },
		{ 349, 493, 59 },
		{ 349, 494, 59 },
		{ 349, 495, 59 },
		{ 349, 496, 59 },
		{ 349, 497, 59 },
		{ 349, 498, 59 },
		{ 349, 499, 59 },
		{ 349, 500, 59 },
		{ 349, 501, 59 },
		{ 349, 502, 59 },
		{ 349, 503, 59 },
		{ 349, 504, 59 },
		{ 349, 505, 59 },
		{ 349, 506, 59 },
		{ 349, 507, 59 },
		{ 349, 508, 59 },
		{ 349, 509, 59 },
		{ 349, 510, 59 },
		{ 349, 511, 59 },
		{ 349, 512, 59 },
		{ 350, 444, 59 },
		{ 350, 445, 59 },
		{ 350, 446, 59 },
		{ 350, 447, 59 },
		{ 350, 448, 59 },
		{ 350, 449, 59 },
		{ 350, 450, 59 },
		{ 350, 451, 59 },
		{ 350, 452, 59 },
		{ 350, 453, 59 },
		{ 350, 454, 59 },
		{ 350, 455, 59 },
		{ 350, 456, 59 },
		{ 350, 457, 59 },
		{ 350, 458, 59 },
		{ 350, 459, 59 },
		{ 350, 460, 59 },
		{ 350, 461, 59 },
		{ 350, 462, 59 },
		{ 350, 463, 59 },
		{ 350, 464, 59 },
		{ 350, 465, 59 },
		{ 350, 466, 59 },
		{ 350, 467, 59 },
		{ 350, 468, 59 },
		{ 350, 469, 59 },
		{ 350, 470, 59 },
		{ 350, 471, 59 },
		{ 350, 472, 59 },
		{ 350, 473, 59 },
		{ 350, 474, 59 },
		{ 350, 475, 59 },
		{ 350, 476, 59 },
		{ 350, 477, 59 },
		{ 350, 478, 59 },
		{ 350, 479, 59 },
		{ 350, 480, 59 },
		{ 350, 481, 59 },
		{ 350, 482, 59 },
		{ 350, 483, 59 },
		{ 350, 484, 59 },
		{ 350, 485, 59 },
		{ 350, 486, 59 },
		{ 350, 487, 59 },
		{ 350, 488, 59 },
		{ 350, 489, 59 },
		{ 350, 490, 59 },
		{ 350, 491, 59 },
		{ 350, 492, 59 },
		{ 350, 493, 59 },
		{ 350, 494, 59 },
		{ 350, 495, 59 },
		{ 350, 496, 59 },
		{ 350, 497, 59 },
		{ 350, 498, 59 },
		{ 350, 499, 59 },
		{ 350, 500, 59 },
		{ 350, 501, 59 },
		{ 350, 502, 59 },
		{ 350, 503, 59 },
		{ 350, 504, 59 },
		{ 350, 505, 59 },
		{ 350, 506, 59 },
		{ 350, 507, 59 },
		{ 350, 508, 59 },
		{ 350, 509, 59 },
		{ 350, 510, 59 },
		{ 350, 511, 59 },
		{ 350, 512, 59 },
		{ 351, 446, 59 },
		{ 351, 447, 59 },
		{ 351, 448, 59 },
		{ 351, 449, 59 },
		{ 351, 450, 59 },
		{ 351, 451, 59 },
		{ 351, 452, 59 },
		{ 351, 453, 59 },
		{ 351, 454, 59 },
		{ 351, 455, 59 },
		{ 351, 456, 59 },
		{ 351, 457, 59 },
		{ 351, 458, 59 },
		{ 351, 459, 59 },
		{ 351, 460, 59 },
		{ 351, 461, 59 },
		{ 351, 462, 59 },
		{ 351, 463, 59 },
		{ 351, 464, 59 },
		{ 351, 465, 59 },
		{ 351, 466, 59 },
		{ 351, 467, 59 },
		{ 351, 468, 59 },
		{ 351, 469, 59 },
		{ 351, 470, 59 },
		{ 351, 471, 59 },
		{ 351, 472, 59 },
		{ 351, 473, 59 },
		{ 351, 474, 59 },
		{ 351, 475, 59 },
		{ 351, 476, 59 },
		{ 351, 477, 59 },
		{ 351, 478, 59 },
		{ 351, 479, 59 },
		{ 351, 480, 59 },
		{ 351, 481, 59 },
		{ 351, 482, 59 },
		{ 351, 483, 59 },
		{ 351, 484, 59 },
		{ 351, 485, 59 },
		{ 351, 486, 59 },
		{ 351, 487, 59 },
		{ 351, 488, 59 },
		{ 351, 489, 59 },
		{ 351, 490, 59 },
		{ 351, 491, 59 },
		{ 351, 492, 59 },
		{ 351, 493, 59 },
		{ 351, 494, 59 },
		{ 351, 495, 59 },
		{ 351, 496, 59 },
		{ 351, 497, 59 },
		{ 351, 498, 59 },
		{ 351, 499, 59 },
		{ 351, 500, 59 },
		{ 351, 501, 59 },
		{ 351, 502, 59 },
		{ 351, 503, 59 },
		{ 351, 504, 59 },
		{ 351, 505, 59 },
		{ 351, 506, 59 },
		{ 351, 507, 59 },
		{ 351, 508, 59 },
		{ 351, 509, 59 },
		{ 351, 510, 59 },
		{ 351, 511, 59 },
		{ 351, 512, 59 },
		{ 352, 448, 59 },
		{ 352, 449, 59 },
		{ 352, 450, 59 },
		{ 352, 451, 59 },
		{ 352, 452, 59 },
		{ 352, 453, 59 },
		{ 352, 454, 59 },
		{ 352, 455, 59 },
		{ 352, 456, 59 },
		{ 352, 457, 59 },
		{ 352, 458, 59 },
		{ 352, 459, 59 },
		{ 352, 460, 59 },
		{ 352, 461, 59 },
		{ 352, 462, 59 },
		{ 352, 463, 59 },
		{ 352, 464, 59 },
		{ 352, 465, 59 },
		{ 352, 466, 59 },
		{ 352, 467, 59 },
		{ 352, 468, 59 },
		{ 352, 469, 59 },
		{ 352, 470, 59 },
		{ 352, 471, 59 },
		{ 352, 472, 59 },
		{ 352, 473, 59 },
		{ 352, 474, 59 },
		{ 352, 475, 59 },
		{ 352, 476, 59 },
		{ 352, 477, 59 },
		{ 352, 478, 59 },
		{ 352, 479, 59 },
		{ 352, 480, 59 },
		{ 352, 481, 59 },
		{ 352, 482, 59 },
		{ 352, 483, 59 },
		{ 352, 484, 59 },
		{ 352, 485, 59 },
		{ 352, 486, 59 },
		{ 352, 487, 59 },
		{ 352, 488, 59 },
		{ 352, 489, 59 },
		{ 352, 490, 59 },
		{ 352, 491, 59 },
		{ 352, 492, 59 },
		{ 352, 493, 59 },
		{ 352, 494, 59 },
		{ 352, 495, 59 },
		{ 352, 496, 59 },
		{ 352, 497, 59 },
		{ 352, 498, 59 },
		{ 352, 499, 59 },
		{ 352, 500, 59 },
		{ 352, 501, 59 },
		{ 352, 502, 59 },
		{ 352, 503, 59 },
		{ 352, 504, 59 },
		{ 352, 505, 59 },
		{ 352, 506, 59 },
		{ 352, 507, 59 },
		{ 352, 508, 59 },
		{ 352, 509, 59 },
		{ 352, 510, 59 },
		{ 352, 511, 59 },
		{ 352, 512, 59 },
		{ 353, 450, 59 },
		{ 353, 451, 59 },
		{ 353, 452, 59 },
		{ 353, 453, 59 },
		{ 353, 454, 59 },
		{ 353, 455, 59 },
		{ 353, 456, 59 },
		{ 353, 457, 59 },
		{ 353, 458, 59 },
		{ 353, 459, 59 },
		{ 353, 460, 59 },
		{ 353, 461, 59 },
		{ 353, 462, 59 },
		{ 353, 463, 59 },
		{ 353, 464, 59 },
		{ 353, 465, 59 },
		{ 353, 466, 59 },
		{ 353, 467, 59 },
		{ 353, 468, 59 },
		{ 353, 469, 59 },
		{ 353, 470, 59 },
		{ 353, 471, 59 },
		{ 353, 472, 59 },
		{ 353, 473, 59 },
		{ 353, 474, 59 },
		{ 353, 475, 59 },
		{ 353, 476, 59 },
		{ 353, 477, 59 },
		{ 353, 478, 59 },
		{ 353, 479, 59 },
		{ 353, 480, 59 },
		{ 353, 481, 59 },
		{ 353, 482, 59 },
		{ 353, 483, 59 },
		{ 353, 484, 59 },
		{ 353, 485, 59 },
		{ 353, 486, 59 },
		{ 353, 487, 59 },
		{ 353, 488, 59 },
		{ 353, 489, 59 },
		{ 353, 490, 59 },
		{ 353, 491, 59 },
		{ 353, 492, 59 },
		{ 353, 493, 59 },
		{ 353, 494, 59 },
		{ 353, 495, 59 },
		{ 353, 496, 59 },
		{ 353, 497, 59 },
		{ 353, 498, 59 },
		{ 353, 499, 59 },
		{ 353, 500, 59 },
		{ 353, 501, 59 },
		{ 353, 502, 59 },
		{ 353, 503, 59 },
		{ 353, 504, 59 },
		{ 353, 505, 59 },
		{ 353, 506, 59 },
		{ 353, 507, 59 },
		{ 353, 508, 59 },
		{ 353, 509, 59 },
		{ 353, 510, 59 },
		{ 353, 511, 59 },
		{ 353, 512, 59 },
		{ 354, 452, 59 },
		{ 354, 453, 59 },
		{ 354, 454, 59 },
		{ 354, 455, 59 },
		{ 354, 456, 59 },
		{ 354, 457, 59 },
		{ 354, 458, 59 },
		{ 354, 459, 59 },
		{ 354, 460, 59 },
		{ 354, 461, 59 },
		{ 354, 462, 59 },
		{ 354, 463, 59 },
		{ 354, 464, 59 },
		{ 354, 465, 59 },
		{ 354, 466, 59 },
		{ 354, 467, 59 },
		{ 354, 468, 59 },
		{ 354, 469, 59 },
		{ 354, 470, 59 },
		{ 354, 471, 59 },
		{ 354, 472, 59 },
		{ 354, 473, 59 },
		{ 354, 474, 59 },
		{ 354, 475, 59 },
		{ 354, 476, 59 },
		{ 354, 477, 59 },
		{ 354, 478, 59 },
		{ 354, 479, 59 },
		{ 354, 480, 59 },
		{ 354, 481, 59 },
		{ 354, 482, 59 },
		{ 354, 483, 59 },
		{ 354, 484, 59 },
		{ 354, 485, 59 },
		{ 354, 486, 59 },
		{ 354, 487, 59 },
		{ 354, 488, 59 },
		{ 354, 489, 59 },
		{ 354, 490, 59 },
		{ 354, 491, 59 },
		{ 354, 492, 59 },
		{ 354, 493, 59 },
		{ 354, 494, 59 },
		{ 354, 495, 59 },
		{ 354, 496, 59 },
		{ 354, 497, 59 },
		{ 354, 498, 59 },
		{ 354, 499, 59 },
		{ 354, 500, 59 },
		{ 354, 501, 59 },
		{ 354, 502, 59 },
		{ 354, 503, 59 },
		{ 354, 504, 59 },
		{ 354, 505, 59 },
		{ 354, 506, 59 },
		{ 354, 507, 59 },
		{ 354, 508, 59 },
		{ 354, 509, 59 },
		{ 354, 510, 59 },
		{ 354, 511, 59 },
		{ 354, 512, 59 },
		{ 355, 454, 59 },
		{ 355, 455, 59 },
		{ 355, 456, 59 },
		{ 355, 457, 59 },
		{ 355, 458, 59 },
		{ 355, 459, 59 },
		{ 355, 460, 59 },
		{ 355, 461, 59 },
		{ 355, 462, 59 },
		{ 355, 463, 59 },
		{ 355, 464, 59 },
		{ 355, 465, 59 },
		{ 355, 466, 59 },
		{ 355, 467, 59 },
		{ 355, 468, 59 },
		{ 355, 469, 59 },
		{ 355, 470, 59 },
		{ 355, 471, 59 },
		{ 355, 472, 59 },
		{ 355, 473, 59 },
		{ 355, 474, 59 },
		{ 355, 475, 59 },
		{ 355, 476, 59 },
		{ 355, 477, 59 },
		{ 355, 478, 59 },
		{ 355, 479, 59 },
		{ 355, 480, 59 },
		{ 355, 481, 59 },
		{ 355, 482, 59 },
		{ 355, 483, 59 },
		{ 355, 484, 59 },
		{ 355, 485, 59 },
		{ 355, 486, 59 },
		{ 355, 487, 59 },
		{ 355, 488, 59 },
		{ 355, 489, 59 },
		{ 355, 490, 59 },
		{ 355, 491, 59 },
		{ 355, 492, 59 },
		{ 355, 493, 59 },
		{ 355, 494, 59 },
		{ 355, 495, 59 },
		{ 355, 496, 59 },
		{ 355, 497, 59 },
		{ 355, 498, 59 },
		{ 355, 499, 59 },
		{ 355, 500, 59 },
		{ 355, 501, 59 },
		{ 355, 502, 59 },
		{ 355, 503, 59 },
		{ 355, 504, 59 },
		{ 355, 505, 59 },
		{ 355, 506, 59 },
		{ 355, 507, 59 },
		{ 355, 508, 59 },
		{ 355, 509, 59 },
		{ 355, 510, 59 },
		{ 355, 511, 59 },
		{ 355, 512, 59 },
		{ 356, 456, 59 },
		{ 356, 457, 59 },
		{ 356, 458, 59 },
		{ 356, 459, 59 },
		{ 356, 460, 59 },
		{ 356, 461, 59 },
		{ 356, 462, 59 },
		{ 356, 463, 59 },
		{ 356, 464, 59 },
		{ 356, 465, 59 },
		{ 356, 466, 59 },
		{ 356, 467, 59 },
		{ 356, 468, 59 },
		{ 356, 469, 59 },
		{ 356, 470, 59 },
		{ 356, 471, 59 },
		{ 356, 472, 59 },
		{ 356, 473, 59 },
		{ 356, 474, 59 },
		{ 356, 475, 59 },
		{ 356, 476, 59 },
		{ 356, 477, 59 },
		{ 356, 478, 59 },
		{ 356, 479, 59 },
		{ 356, 480, 59 },
		{ 356, 481, 59 },
		{ 356, 482, 59 },
		{ 356, 483, 59 },
		{ 356, 484, 59 },
		{ 356, 485, 59 },
		{ 356, 486, 59 },
		{ 356, 487, 59 },
		{ 356, 488, 59 },
		{ 356, 489, 59 },
		{ 356, 490, 59 },
		{ 356, 491, 59 },
		{ 356, 492, 59 },
		{ 356, 493, 59 },
		{ 356, 494, 59 },
		{ 356, 495, 59 },
		{ 356, 496, 59 },
		{ 356, 497, 59 },
		{ 356, 498, 59 },
		{ 356, 499, 59 },
		{ 356, 500, 59 },
		{ 356, 501, 59 },
		{ 356, 502, 59 },
		{ 356, 503, 59 },
		{ 356, 504, 59 },
		{ 356, 505, 59 },
		{ 356, 506, 59 },
		{ 356, 507, 59 },
		{ 356, 508, 59 },
		{ 356, 509, 59 },
		{ 356, 510, 59 },
		{ 356, 511, 59 },
		{ 356, 512, 59 },
		{ 357, 458, 59 },
		{ 357, 459, 59 },
		{ 357, 460, 59 },
		{ 357, 461, 59 },
		{ 357, 462, 59 },
		{ 357, 463, 59 },
		{ 357, 464, 59 },
		{ 357, 465, 59 },
		{ 357, 466, 59 },
		{ 357, 467, 59 },
		{ 357, 468, 59 },
		{ 357, 469, 59 },
		{ 357, 470, 59 },
		{ 357, 471, 59 },
		{ 357, 472, 59 },
		{ 357, 473, 59 },
		{ 357, 474, 59 },
		{ 357, 475, 59 },
		{ 357, 476, 59 },
		{ 357, 477, 59 },
		{ 357, 478, 59 },
		{ 357, 479, 59 },
		{ 357, 480, 59 },
		{ 357, 481, 59 },
		{ 357, 482, 59 },
		{ 357, 483, 59 },
		{ 357, 484, 59 },
		{ 357, 485, 59 },
		{ 357, 486, 59 },
		{ 357, 487, 59 },
		{ 357, 488, 59 },
		{ 357, 489, 59 },
		{ 357, 490, 59 },
		{ 357, 491, 59 },
		{ 357, 492, 59 },
		{ 357, 493, 59 },
		{ 357, 494, 59 },
		{ 357, 495, 59 },
		{ 357, 496, 59 },
		{ 357, 497, 59 },
		{ 357, 498, 59 },
		{ 357, 499, 59 },
		{ 357, 500, 59 },
		{ 357, 501, 59 },
		{ 357, 502, 59 },
		{ 357, 503, 59 },
		{ 357, 504, 59 },
		{ 357, 505, 59 },
		{ 357, 506, 59 },
		{ 357, 507, 59 },
		{ 357, 508, 59 },
		{ 357, 509, 59 },
		{ 357, 510, 59 },
		{ 357, 511, 59 },
		{ 357, 512, 59 },
		{ 358, 460, 59 },
		{ 358, 461, 59 },
		{ 358, 462, 59 },
		{ 358, 463, 59 },
		{ 358, 464, 59 },
		{ 358, 465, 59 },
		{ 358, 466, 59 },
		{ 358, 467, 59 },
		{ 358, 468, 59 },
		{ 358, 469, 59 },
		{ 358, 470, 59 },
		{ 358, 471, 59 },
		{ 358, 472, 59 },
		{ 358, 473, 59 },
		{ 358, 474, 59 },
		{ 358, 475, 59 },
		{ 358, 476, 59 },
		{ 358, 477, 59 },
		{ 358, 478, 59 },
		{ 358, 479, 59 },
		{ 358, 480, 59 },
		{ 358, 481, 59 },
		{ 358, 482, 59 },
		{ 358, 483, 59 },
		{ 358, 484, 59 },
		{ 358, 485, 59 },
		{ 358, 486, 59 },
		{ 358, 487, 59 },
		{ 358, 488, 59 },
		{ 358, 489, 59 },
		{ 358, 490, 59 },
		{ 358, 491, 59 },
		{ 358, 492, 59 },
		{ 358, 493, 59 },
		{ 358, 494, 59 },
		{ 358, 495, 59 },
		{ 358, 496, 59 },
		{ 358, 497, 59 },
		{ 358, 498, 59 },
		{ 358, 499, 59 },
		{ 358, 500, 59 },
		{ 358, 501, 59 },
		{ 358, 502, 59 },
		{ 358, 503, 59 },
		{ 358, 504, 59 },
		{ 358, 505, 59 },
		{ 358, 506, 59 },
		{ 358, 507, 59 },
		{ 358, 508, 59 },
		{ 358, 509, 59 },
		{ 358, 510, 59 },
		{ 358, 511, 59 },
		{ 358, 512, 59 },
		{ 359, 462, 59 },
		{ 359, 463, 59 },
		{ 359, 464, 59 },
		{ 359, 465, 59 },
		{ 359, 466, 59 },
		{ 359, 467, 59 },
		{ 359, 468, 59 },
		{ 359, 469, 59 },
		{ 359, 470, 59 },
		{ 359, 471, 59 },
		{ 359, 472, 59 },
		{ 359, 473, 59 },
		{ 359, 474, 59 },
		{ 359, 475, 59 },
		{ 359, 476, 59 },
		{ 359, 477, 59 },
		{ 359, 478, 59 },
		{ 359, 479, 59 },
		{ 359, 480, 59 },
		{ 359, 481, 59 },
		{ 359, 482, 59 },
		{ 359, 483, 59 },
		{ 359, 484, 59 },
		{ 359, 485, 59 },
		{ 359, 486, 59 },
		{ 359, 487, 59 },
		{ 359, 488, 59 },
		{ 359, 489, 59 },
		{ 359, 490, 59 },
		{ 359, 491, 59 },
		{ 359, 492, 59 },
		{ 359, 493, 59 },
		{ 359, 494, 59 },
		{ 359, 495, 59 },
		{ 359, 496, 59 },
		{ 359, 497, 59 },
		{ 359, 498, 59 },
		{ 359, 499, 59 },
		{ 359, 500, 59 },
		{ 359, 501, 59 },
		{ 359, 502, 59 },
		{ 359, 503, 59 },
		{ 359, 504, 59 },
		{ 359, 505, 59 },
		{ 359, 506, 59 },
		{ 359, 507, 59 },
		{ 359, 508, 59 },
		{ 359, 509, 59 },
		{ 359, 510, 59 },
		{ 359, 511, 59 },
		{ 359, 512, 59 },
		{ 360, 464, 59 },
		{ 360, 465, 59 },
		{ 360, 466, 59 },
		{ 360, 467, 59 },
		{ 360, 468, 59 },
		{ 360, 469, 59 },
		{ 360, 470, 59 },
		{ 360, 471, 59 },
		{ 360, 472, 59 },
		{ 360, 473, 59 },
		{ 360, 474, 59 },
		{ 360, 475, 59 },
		{ 360, 476, 59 },
		{ 360, 477, 59 },
		{ 360, 478, 59 },
		{ 360, 479, 59 },
		{ 360, 480, 59 },
		{ 360, 481, 59 },
		{ 360, 482, 59 },
		{ 360, 483, 59 },
		{ 360, 484, 59 },
		{ 360, 485, 59 },
		{ 360, 486, 59 },
		{ 360, 487, 59 },
		{ 360, 488, 59 },
		{ 360, 489, 59 },
		{ 360, 490, 59 },
		{ 360, 491, 59 },
		{ 360, 492, 59 },
		{ 360, 493, 59 },
		{ 360, 494, 59 },
		{ 360, 495, 59 },
		{ 360, 496, 59 },
		{ 360, 497, 59 },
		{ 360, 498, 59 },
		{ 360, 499, 59 },
		{ 360, 500, 59 },
		{ 360, 501, 59 },
		{ 360, 502, 59 },
		{ 360, 503, 59 },
		{ 360, 504, 59 },
		{ 360, 505, 59 },
		{ 360, 506, 59 },
		{ 360, 507, 59 },
		{ 360, 508, 59 },
		{ 360, 509, 59 },
		{ 360, 510, 59 },
		{ 360, 511, 59 },
		{ 360, 512, 59 },
		{ 361, 466, 59 },
		{ 361, 467, 59 },
		{ 361, 468, 59 },
		{ 361, 469, 59 },
		{ 361, 470, 59 },
		{ 361, 471, 59 },
		{ 361, 472, 59 },
		{ 361, 473, 59 },
		{ 361, 474, 59 },
		{ 361, 475, 59 },
		{ 361, 476, 59 },
		{ 361, 477, 59 },
		{ 361, 478, 59 },
		{ 361, 479, 59 },
		{ 361, 480, 59 },
		{ 361, 481, 59 },
		{ 361, 482, 59 },
		{ 361, 483, 59 },
		{ 361, 484, 59 },
		{ 361, 485, 59 },
		{ 361, 486, 59 },
		{ 361, 487, 59 },
		{ 361, 488, 59 },
		{ 361, 489, 59 },
		{ 361, 490, 59 },
		{ 361, 491, 59 },
		{ 361, 492, 59 },
		{ 361, 493, 59 },
		{ 361, 494, 59 },
		{ 361, 495, 59 },
		{ 361, 496, 59 },
		{ 361, 497, 59 },
		{ 361, 498, 59 },
		{ 361, 499, 59 },
		{ 361, 500, 59 },
		{ 361, 501, 59 },
		{ 361, 502, 59 },
		{ 361, 503, 59 },
		{ 361, 504, 59 },
		{ 361, 505, 59 },
		{ 361, 506, 59 },
		{ 361, 507, 59 },
		{ 361, 508, 59 },
		{ 361, 509, 59 },
		{ 361, 510, 59 },
		{ 361, 511, 59 },
		{ 361, 512, 59 },
		{ 362, 468, 59 },
		{ 362, 469, 59 },
		{ 362, 470, 59 },
		{ 362, 471, 59 },
		{ 362, 472, 59 },
		{ 362, 473, 59 },
		{ 362, 474, 59 },
		{ 362, 475, 59 },
		{ 362, 476, 59 },
		{ 362, 477, 59 },
		{ 362, 478, 59 },
		{ 362, 479, 59 },
		{ 362, 480, 59 },
		{ 362, 481, 59 },
		{ 362, 482, 59 },
		{ 362, 483, 59 },
		{ 362, 484, 59 },
		{ 362, 485, 59 },
		{ 362, 486, 59 },
		{ 362, 487, 59 },
		{ 362, 488, 59 },
		{ 362, 489, 59 },
		{ 362, 490, 59 },
		{ 362, 491, 59 },
		{ 362, 492, 59 },
		{ 362, 493, 59 },
		{ 362, 494, 59 },
		{ 362, 495, 59 },
		{ 362, 496, 59 },
		{ 362, 497, 59 },
		{ 362, 498, 59 },
		{ 362, 499, 59 },
		{ 362, 500, 59 },
		{ 362, 501, 59 },
		{ 362, 502, 59 },
		{ 362, 503, 59 },
		{ 362, 504, 59 },
		{ 362, 505, 59 },
		{ 362, 506, 59 },
		{ 362, 507, 59 },
		{ 362, 508, 59 },
		{ 362, 509, 59 },
		{ 362, 510, 59 },
		{ 362, 511, 59 },
		{ 362, 512, 59 },
		{ 363, 470, 59 },
		{ 363, 471, 59 },
		{ 363, 472, 59 },
		{ 363, 473, 59 },
		{ 363, 474, 59 },
		{ 363, 475, 59 },
		{ 363, 476, 59 },
		{ 363, 477, 59 },
		{ 363, 478, 59 },
		{ 363, 479, 59 },
		{ 363, 480, 59 },
		{ 363, 481, 59 },
		{ 363, 482, 59 },
		{ 363, 483, 59 },
		{ 363, 484, 59 },
		{ 363, 485, 59 },
		{ 363, 486, 59 },
		{ 363, 487, 59 },
		{ 363, 488, 59 },
		{ 363, 489, 59 },
		{ 363, 490, 59 },
		{ 363, 491, 59 },
		{ 363, 492, 59 },
		{ 363, 493, 59 },
		{ 363, 494, 59 },
		{ 363, 495, 59 },
		{ 363, 496, 59 },
		{ 363, 497, 59 },
		{ 363, 498, 59 },
		{ 363, 499, 59 },
		{ 363, 500, 59 },
		{ 363, 501, 59 },
		{ 363, 502, 59 },
		{ 363, 503, 59 },
		{ 363, 504, 59 },
		{ 363, 505, 59 },
		{ 363, 506, 59 },
		{ 363, 507, 59 },
		{ 363, 508, 59 },
		{ 363, 509, 59 },
		{ 363, 510, 59 },
		{ 363, 511, 59 },
		{ 363, 512, 59 },
		{ 364, 472, 59 },
		{ 364, 473, 59 },
		{ 364, 474, 59 },
		{ 364, 475, 59 },
		{ 364, 476, 59 },
		{ 364, 477, 59 },
		{ 364, 478, 59 },
		{ 364, 479, 59 },
		{ 364, 480, 59 },
		{ 364, 481, 59 },
		{ 364, 482, 59 },
		{ 364, 483, 59 },
		{ 364, 484, 59 },
		{ 364, 485, 59 },
		{ 364, 486, 59 },
		{ 364, 487, 59 },
		{ 364, 488, 59 },
		{ 364, 489, 59 },
		{ 364, 490, 59 },
		{ 364, 491, 59 },
		{ 364, 492, 59 },
		{ 364, 493, 59 },
		{ 364, 494, 59 },
		{ 364, 495, 59 },
		{ 364, 496, 59 },
		{ 364, 497, 59 },
		{ 364, 498, 59 },
		{ 364, 499, 59 },
		{ 364, 500, 59 },
		{ 364, 501, 59 },
		{ 364, 502, 59 },
		{ 364, 503, 59 },
		{ 364, 504, 59 },
		{ 364, 505, 59 },
		{ 364, 506, 59 },
		{ 364, 507, 59 },
		{ 364, 508, 59 },
		{ 364, 509, 59 },
		{ 364, 510, 59 },
		{ 364, 511, 59 },
		{ 364, 512, 59 },
		{ 365, 474, 59 },
		{ 365, 475, 59 },
		{ 365, 476, 59 },
		{ 365, 477, 59 },
		{ 365, 478, 59 },
		{ 365, 479, 59 },
		{ 365, 480, 59 },
		{ 365, 481, 59 },
		{ 365, 482, 59 },
		{ 365, 483, 59 },
		{ 365, 484, 59 },
		{ 365, 485, 59 },
		{ 365, 486, 59 },
		{ 365, 487, 59 },
		{ 365, 488, 59 },
		{ 365, 489, 59 },
		{ 365, 490, 59 },
		{ 365, 491, 59 },
		{ 365, 492, 59 },
		{ 365, 493, 59 },
		{ 365, 494, 59 },
		{ 365, 495, 59 },
		{ 365, 496, 59 },
		{ 365, 497, 59 },
		{ 365, 498, 59 },
		{ 365, 499, 59 },
		{ 365, 500, 59 },
		{ 365, 501, 59 },
		{ 365, 502, 59 },
		{ 365, 503, 59 },
		{ 365, 504, 59 },
		{ 365, 505, 59 },
		{ 365, 506, 59 },
		{ 365, 507, 59 },
		{ 365, 508, 59 },
		{ 365, 509, 59 },
		{ 365, 510, 59 },
		{ 365, 511, 59 },
		{ 365, 512, 59 },
		{ 366, 476, 59 },
		{ 366, 477, 59 },
		{ 366, 478, 59 },
		{ 366, 479, 59 },
		{ 366, 480, 59 },
		{ 366, 481, 59 },
		{ 366, 482, 59 },
		{ 366, 483, 59 },
		{ 366, 484, 59 },
		{ 366, 485, 59 },
		{ 366, 486, 59 },
		{ 366, 487, 59 },
		{ 366, 488, 59 },
		{ 366, 489, 59 },
		{ 366, 490, 59 },
		{ 366, 491, 59 },
		{ 366, 492, 59 },
		{ 366, 493, 59 },
		{ 366, 494, 59 },
		{ 366, 495, 59 },
		{ 366, 496, 59 },
		{ 366, 497, 59 },
		{ 366, 498, 59 },
		{ 366, 499, 59 },
		{ 366, 500, 59 },
		{ 366, 501, 59 },
		{ 366, 502, 59 },
		{ 366, 503, 59 },
		{ 366, 504, 59 },
		{ 366, 505, 59 },
		{ 366, 506, 59 },
		{ 366, 507, 59 },
		{ 366, 508, 59 },
		{ 366, 509, 59 },
		{ 366, 510, 59 },
		{ 366, 511, 59 },
		{ 366, 512, 59 },
		{ 367, 478, 59 },
		{ 367, 479, 59 },
		{ 367, 480, 59 },
		{ 367, 481, 59 },
		{ 367, 482, 59 },
		{ 367, 483, 59 },
		{ 367, 484, 59 },
		{ 367, 485, 59 },
		{ 367, 486, 59 },
		{ 367, 487, 59 },
		{ 367, 488, 59 },
		{ 367, 489, 59 },
		{ 367, 490, 59 },
		{ 367, 491, 59 },
		{ 367, 492, 59 },
		{ 367, 493, 59 },
		{ 367, 494, 59 },
		{ 367, 495, 59 },
		{ 367, 496, 59 },
		{ 367, 497, 59 },
		{ 367, 498, 59 },
		{ 367, 499, 59 },
		{ 367, 500, 59 },
		{ 367, 501, 59 },
		{ 367, 502, 59 },
		{ 367, 503, 59 },
		{ 367, 504, 59 },
		{ 367, 505, 59 },
		{ 367, 506, 59 },
		{ 367, 507, 59 },
		{ 367, 508, 59 },
		{ 367, 509, 59 },
		{ 367, 510, 59 },
		{ 367, 511, 59 },
		{ 367, 512, 59 },
		{ 368, 480, 59 },
		{ 368, 481, 59 },
		{ 368, 482, 59 },
		{ 368, 483, 59 },
		{ 368, 484, 59 },
		{ 368, 485, 59 },
		{ 368, 486, 59 },
		{ 368, 487, 59 },
		{ 368, 488, 59 },
		{ 368, 489, 59 },
		{ 368, 490, 59 },
		{ 368, 491, 59 },
		{ 368, 492, 59 },
		{ 368, 493, 59 },
		{ 368, 494, 59 },
		{ 368, 495, 59 },
		{ 368, 496, 59 },
		{ 368, 497, 59 },
		{ 368, 498, 59 },
		{ 368, 499, 59 },
		{ 368, 500, 59 },
		{ 368, 501, 59 },
		{ 368, 502, 59 },
		{ 368, 503, 59 },
		{ 368, 504, 59 },
		{ 368, 505, 59 },
		{ 368, 506, 59 },
		{ 368, 507, 59 },
		{ 368, 508, 59 },
		{ 368, 509, 59 },
		{ 368, 510, 59 },
		{ 368, 511, 59 },
		{ 368, 512, 59 },
		{ 369, 482, 59 },
		{ 369, 483, 59 },
		{ 369, 484, 59 },
		{ 369, 485, 59 },
		{ 369, 486, 59 },
		{ 369, 487, 59 },
		{ 369, 488, 59 },
		{ 369, 489, 59 },
		{ 369, 490, 59 },
		{ 369, 491, 59 },
		{ 369, 492, 59 },
		{ 369, 493, 59 },
		{ 369, 494, 59 },
		{ 369, 495, 59 },
		{ 369, 496, 59 },
		{ 369, 497, 59 },
		{ 369, 498, 59 },
		{ 369, 499, 59 },
		{ 369, 500, 59 },
		{ 369, 501, 59 },
		{ 369, 502, 59 },
		{ 369, 503, 59 },
		{ 369, 504, 59 },
		{ 369, 505, 59 },
		{ 369, 506, 59 },
		{ 369, 507, 59 },
		{ 369, 508, 59 },
		{ 369, 509, 59 },
		{ 369, 510, 59 },
		{ 369, 511, 59 },
		{ 369, 512, 59 },
		{ 370, 484, 59 },
		{ 370, 485, 59 },
		{ 370, 486, 59 },
		{ 370, 487, 59 },
		{ 370, 488, 59 },
		{ 370, 489, 59 },
		{ 370, 490, 59 },
		{ 370, 491, 59 },
		{ 370, 492, 59 },
		{ 370, 493, 59 },
		{ 370, 494, 59 },
		{ 370, 495, 59 },
		{ 370, 496, 59 },
		{ 370, 497, 59 },
		{ 370, 498, 59 },
		{ 370, 499, 59 },
		{ 370, 500, 59 },
		{ 370, 501, 59 },
		{ 370, 502, 59 },
		{ 370, 503, 59 },
		{ 370, 504, 59 },
		{ 370, 505, 59 },
		{ 370, 506, 59 },
		{ 370, 507, 59 },
		{ 370, 508, 59 },
		{ 370, 509, 59 },
		{ 370, 510, 59 },
		{ 370, 511, 59 },
		{ 370, 512, 59 },
		{ 371, 486, 59 },
		{ 371, 487, 59 },
		{ 371, 488, 59 },
		{ 371, 489, 59 },
		{ 371, 490, 59 },
		{ 371, 491, 59 },
		{ 371, 492, 59 },
		{ 371, 493, 59 },
		{ 371, 494, 59 },
		{ 371, 495, 59 },
		{ 371, 496, 59 },
		{ 371, 497, 59 },
		{ 371, 498, 59 },
		{ 371, 499, 59 },
		{ 371, 500, 59 },
		{ 371, 501, 59 },
		{ 371, 502, 59 },
		{ 371, 503, 59 },
		{ 371, 504, 59 },
		{ 371, 505, 59 },
		{ 371, 506, 59 },
		{ 371, 507, 59 },
		{ 371, 508, 59 },
		{ 371, 509, 59 },
		{ 371, 510, 59 },
		{ 371, 511, 59 },
		{ 371, 512, 59 },
		{ 372, 488, 59 },
		{ 372, 489, 59 },
		{ 372, 490, 59 },
		{ 372, 491, 59 },
		{ 372, 492, 59 },
		{ 372, 493, 59 },
		{ 372, 494, 59 },
		{ 372, 495, 59 },
		{ 372, 496, 59 },
		{ 372, 497, 59 },
		{ 372, 498, 59 },
		{ 372, 499, 59 },
		{ 372, 500, 59 },
		{ 372, 501, 59 },
		{ 372, 502, 59 },
		{ 372, 503, 59 },
		{ 372, 504, 59 },
		{ 372, 505, 59 },
		{ 372, 506, 59 },
		{ 372, 507, 59 },
		{ 372, 508, 59 },
		{ 372, 509, 59 },
		{ 372, 510, 59 },
		{ 372, 511, 59 },
		{ 372, 512, 59 },
		{ 373, 490, 59 },
		{ 373, 491, 59 },
		{ 373, 492, 59 },
		{ 373, 493, 59 },
		{ 373, 494, 59 },
		{ 373, 495, 59 },
		{ 373, 496, 59 },
		{ 373, 497, 59 },
		{ 373, 498, 59 },
		{ 373, 499, 59 },
		{ 373, 500, 59 },
		{ 373, 501, 59 },
		{ 373, 502, 59 },
		{ 373, 503, 59 },
		{ 373, 504, 59 },
		{ 373, 505, 59 },
		{ 373, 506, 59 },
		{ 373, 507, 59 },
		{ 373, 508, 59 },
		{ 373, 509, 59 },
		{ 373, 510, 59 },
		{ 373, 511, 59 },
		{ 373, 512, 59 },
		{ 374, 492, 59 },
		{ 374, 493, 59 },
		{ 374, 494, 59 },
		{ 374, 495, 59 },
		{ 374, 496, 59 },
		{ 374, 497, 59 },
		{ 374, 498, 59 },
		{ 374, 499, 59 },
		{ 374, 500, 59 },
		{ 374, 501, 59 },
		{ 374, 502, 59 },
		{ 374, 503, 59 },
		{ 374, 504, 59 },
		{ 374, 505, 59 },
		{ 374, 506, 59 },
		{ 374, 507, 59 },
		{ 374, 508, 59 },
		{ 374, 509, 59 },
		{ 374, 510, 59 },
		{ 374, 511, 59 },
		{ 374, 512, 59 },
		{ 375, 494, 59 },
		{ 375, 495, 59 },
		{ 375, 496, 59 },
		{ 375, 497, 59 },
		{ 375, 498, 59 },
		{ 375, 499, 59 },
		{ 375, 500, 59 },
		{ 375, 501, 59 },
		{ 375, 502, 59 },
		{ 375, 503, 59 },
		{ 375, 504, 59 },
		{ 375, 505, 59 },
		{ 375, 506, 59 },
		{ 375, 507, 59 },
		{ 375, 508, 59 },
		{ 375, 509, 59 },
		{ 375, 510, 59 },
		{ 375, 511, 59 },
		{ 375, 512, 59 },
		{ 376, 496, 59 },
		{ 376, 497, 59 },
		{ 376, 498, 59 },
		{ 376, 499, 59 },
		{ 376, 500, 59 },
		{ 376, 501, 59 },
		{ 376, 502, 59 },
		{ 376, 503, 59 },
		{ 376, 504, 59 },
		{ 376, 505, 59 },
		{ 376, 506, 59 },
		{ 376, 507, 59 },
		{ 376, 508, 59 },
		{ 376, 509, 59 },
		{ 376, 510, 59 },
		{ 376, 511, 59 },
		{ 376, 512, 59 },
		{ 377, 498, 59 },
		{ 377, 499, 59 },
		{ 377, 500, 59 },
		{ 377, 501, 59 },
		{ 377, 502, 59 },
		{ 377, 503, 59 },
		{ 377, 504, 59 },
		{ 377, 505, 59 },
		{ 377, 506, 59 },
		{ 377, 507, 59 },
		{ 377, 508, 59 },
		{ 377, 509, 59 },
		{ 377, 510, 59 },
		{ 377, 511, 59 },
		{ 377, 512, 59 },
		{ 378, 500, 59 },
		{ 378, 501, 59 },
		{ 378, 502, 59 },
		{ 378, 503, 59 },
		{ 378, 504, 59 },
		{ 378, 505, 59 },
		{ 378, 506, 59 },
		{ 378, 507, 59 },
		{ 378, 508, 59 },
		{ 378, 509, 59 },
		{ 378, 510, 59 },
		{ 378, 511, 59 },
		{ 378, 512, 59 },
		{ 379, 502, 59 },
		{ 379, 503, 59 },
		{ 379, 504, 59 },
		{ 379, 505, 59 },
		{ 379, 506, 59 },
		{ 379, 507, 59 },
		{ 379, 508, 59 },
		{ 379, 509, 59 },
		{ 379, 510, 59 },
		{ 379, 511, 59 },
		{ 379, 512, 59 },
		{ 380, 504, 59 },
		{ 380, 505, 59 },
		{ 380, 506, 59 },
		{ 380, 507, 59 },
		{ 380, 508, 59 },
		{ 380, 509, 59 },
		{ 380, 510, 59 },
		{ 380, 511, 59 },
		{ 380, 512, 59 },
		{ 381, 506, 59 },
		{ 381, 507, 59 },
		{ 381, 508, 59 },
		{ 381, 509, 59 },
		{ 381, 510, 59 },
		{ 381, 511, 59 },
		{ 381, 512, 59 },
		{ 382, 508, 59 },
		{ 382, 509, 59 },
		{ 382, 510, 59 },
		{ 382, 511, 59 },
		{ 382, 512, 59 },
		{ 383, 510, 59 },
		{ 383, 511, 59 },
		{ 383, 512, 59 },
		{ 384, 512, 59 },
		{ 162, 1, 60 },
		{ 162, 2, 60 },
		{ 162, 3, 60 },
		{ 163, 1, 60 },
		{ 163, 2, 60 },
		{ 163, 3, 60 },
		{ 163, 4, 60 },
		{ 163, 5, 60 },
		{ 163, 6, 60 },
		{ 164, 1, 60 },
		{ 164, 2, 60 },
		{ 164, 3, 60 },
		{ 164, 4, 60 },
		{ 164, 5, 60 },
		{ 164, 6, 60 },
		{ 164, 7, 60 },
		{ 164, 8, 60 },
		{ 165, 1, 60 },
		{ 165, 2, 60 },
		{ 165, 3, 60 },
		{ 165, 4, 60 },
		{ 165, 5, 60 },
		{ 165, 6, 60 },
		{ 165, 7, 60 },
		{ 165, 8, 60 },
		{ 165, 9, 60 },
		{ 165, 10, 60 },
		{ 165, 11, 60 },
		{ 166, 1, 60 },
		{ 166, 2, 60 },
		{ 166, 3, 60 },
		{ 166, 4, 60 },
		{ 166, 5, 60 },
		{ 166, 6, 60 },
		{ 166, 7, 60 },
		{ 166, 8, 60 },
		{ 166, 9, 60 },
		{ 166, 10, 60 },
		{ 166, 11, 60 },
		{ 166, 12, 60 },
		{ 166, 13, 60 },
		{ 166, 14, 60 },
		{ 167, 1, 60 },
		{ 167, 2, 60 },
		{ 167, 3, 60 },
		{ 167, 4, 60 },
		{ 167, 5, 60 },
		{ 167, 6, 60 },
		{ 167, 7, 60 },
		{ 167, 8, 60 },
		{ 167, 9, 60 },
		{ 167, 10, 60 },
		{ 167, 11, 60 },
		{ 167, 12, 60 },
		{ 167, 13, 60 },
		{ 167, 14, 60 },
		{ 167, 15, 60 },
		{ 167, 16, 60 },
		{ 168, 1, 60 },
		{ 168, 2, 60 },
		{ 168, 3, 60 },
		{ 168, 4, 60 },
		{ 168, 5, 60 },
		{ 168, 6, 60 },
		{ 168, 7, 60 },
		{ 168, 8, 60 },
		{ 168, 9, 60 },
		{ 168, 10, 60 },
		{ 168, 11, 60 },
		{ 168, 12, 60 },
		{ 168, 13, 60 },
		{ 168, 14, 60 },
		{ 168, 15, 60 },
		{ 168, 16, 60 },
		{ 168, 17, 60 },
		{ 168, 18, 60 },
		{ 168, 19, 60 },
		{ 169, 1, 60 },
		{ 169, 2, 60 },
		{ 169, 3, 60 },
		{ 169, 4, 60 },
		{ 169, 5, 60 },
		{ 169, 6, 60 },
		{ 169, 7, 60 },
		{ 169, 8, 60 },
		{ 169, 9, 60 },
		{ 169, 10, 60 },
		{ 169, 11, 60 },
		{ 169, 12, 60 },
		{ 169, 13, 60 },
		{ 169, 14, 60 },
		{ 169, 15, 60 },
		{ 169, 16, 60 },
		{ 169, 17, 60 },
		{ 169, 18, 60 },
		{ 169, 19, 60 },
		{ 169, 20, 60 },
		{ 169, 21, 60 },
		{ 169, 22, 60 },
		{ 170, 1, 60 },
		{ 170, 2, 60 },
		{ 170, 3, 60 },
		{ 170, 4, 60 },
		{ 170, 5, 60 },
		{ 170, 6, 60 },
		{ 170, 7, 60 },
		{ 170, 8, 60 },
		{ 170, 9, 60 },
		{ 170, 10, 60 },
		{ 170, 11, 60 },
		{ 170, 12, 60 },
		{ 170, 13, 60 },
		{ 170, 14, 60 },
		{ 170, 15, 60 },
		{ 170, 16, 60 },
		{ 170, 17, 60 },
		{ 170, 18, 60 },
		{ 170, 19, 60 },
		{ 170, 20, 60 },
		{ 170, 21, 60 },
		{ 170, 22, 60 },
		{ 170, 23, 60 },
		{ 170, 24, 60 },
		{ 171, 1, 60 },
		{ 171, 2, 60 },
		{ 171, 3, 60 },
		{ 171, 4, 60 },
		{ 171, 5, 60 },
		{ 171, 6, 60 },
		{ 171, 7, 60 },
		{ 171, 8, 60 },
		{ 171, 9, 60 },
		{ 171, 10, 60 },
		{ 171, 11, 60 },
		{ 171, 12, 60 },
		{ 171, 13, 60 },
		{ 171, 14, 60 },
		{ 171, 15, 60 },
		{ 171, 16, 60 },
		{ 171, 17, 60 },
		{ 171, 18, 60 },
		{ 171, 19, 60 },
		{ 171, 20, 60 },
		{ 171, 21, 60 },
		{ 171, 22, 60 },
		{ 171, 23, 60 },
		{ 171, 24, 60 },
		{ 171, 25, 60 },
		{ 171, 26, 60 },
		{ 171, 27, 60 },
		{ 172, 1, 60 },
		{ 172, 2, 60 },
		{ 172, 3, 60 },
		{ 172, 4, 60 },
		{ 172, 5, 60 },
		{ 172, 6, 60 },
		{ 172, 7, 60 },
		{ 172, 8, 60 },
		{ 172, 9, 60 },
		{ 172, 10, 60 },
		{ 172, 11, 60 },
		{ 172, 12, 60 },
		{ 172, 13, 60 },
		{ 172, 14, 60 },
		{ 172, 15, 60 },
		{ 172, 16, 60 },
		{ 172, 17, 60 },
		{ 172, 18, 60 },
		{ 172, 19, 60 },
		{ 172, 20, 60 },
		{ 172, 21, 60 },
		{ 172, 22, 60 },
		{ 172, 23, 60 },
		{ 172, 24, 60 },
		{ 172, 25, 60 },
		{ 172, 26, 60 },
		{ 172, 27, 60 },
		{ 172, 28, 60 },
		{ 172, 29, 60 },
		{ 172, 30, 60 },
		{ 173, 1, 60 },
		{ 173, 2, 60 },
		{ 173, 3, 60 },
		{ 173, 4, 60 },
		{ 173, 5, 60 },
		{ 173, 6, 60 },
		{ 173, 7, 60 },
		{ 173, 8, 60 },
		{ 173, 9, 60 },
		{ 173, 10, 60 },
		{ 173, 11, 60 },
		{ 173, 12, 60 },
		{ 173, 13, 60 },
		{ 173, 14, 60 },
		{ 173, 15, 60 },
		{ 173, 16, 60 },
		{ 173, 17, 60 },
		{ 173, 18, 60 },
		{ 173, 19, 60 },
		{ 173, 20, 60 },
		{ 173, 21, 60 },
		{ 173, 22, 60 },
		{ 173, 23, 60 },
		{ 173, 24, 60 },
		{ 173, 25, 60 },
		{ 173, 26, 60 },
		{ 173, 27, 60 },
		{ 173, 28, 60 },
		{ 173, 29, 60 },
		{ 173, 30, 60 },
		{ 173, 31, 60 },
		{ 173, 32, 60 },
		{ 174, 1, 60 },
		{ 174, 2, 60 },
		{ 174, 3, 60 },
		{ 174, 4, 60 },
		{ 174, 5, 60 },
		{ 174, 6, 60 },
		{ 174, 7, 60 },
		{ 174, 8, 60 },
		{ 174, 9, 60 },
		{ 174, 10, 60 },
		{ 174, 11, 60 },
		{ 174, 12, 60 },
		{ 174, 13, 60 },
		{ 174, 14, 60 },
		{ 174, 15, 60 },
		{ 174, 16, 60 },
		{ 174, 17, 60 },
		{ 174, 18, 60 },
		{ 174, 19, 60 },
		{ 174, 20, 60 },
		{ 174, 21, 60 },
		{ 174, 22, 60 },
		{ 174, 23, 60 },
		{ 174, 24, 60 },
		{ 174, 25, 60 },
		{ 174, 26, 60 },
		{ 174, 27, 60 },
		{ 174, 28, 60 },
		{ 174, 29, 60 },
		{ 174, 30, 60 },
		{ 174, 31, 60 },
		{ 174, 32, 60 },
		{ 174, 33, 60 },
		{ 174, 34, 60 },
		{ 174, 35, 60 },
		{ 175, 1, 60 },
		{ 175, 2, 60 },
		{ 175, 3, 60 },
		{ 175, 4, 60 },
		{ 175, 5, 60 },
		{ 175, 6, 60 },
		{ 175, 7, 60 },
		{ 175, 8, 60 },
		{ 175, 9, 60 },
		{ 175, 10, 60 },
		{ 175, 11, 60 },
		{ 175, 12, 60 },
		{ 175, 13, 60 },
		{ 175, 14, 60 },
		{ 175, 15, 60 },
		{ 175, 16, 60 },
		{ 175, 17, 60 },
		{ 175, 18, 60 },
		{ 175, 19, 60 },
		{ 175, 20, 60 },
		{ 175, 21, 60 },
		{ 175, 22, 60 },
		{ 175, 23, 60 },
		{ 175, 24, 60 },
		{ 175, 25, 60 },
		{ 175, 26, 60 },
		{ 175, 27, 60 },
		{ 175, 28, 60 },
		{ 175, 29, 60 },
		{ 175, 30, 60 },
		{ 175, 31, 60 },
		{ 175, 32, 60 },
		{ 175, 33, 60 },
		{ 175, 34, 60 },
		{ 175, 35, 60 },
		{ 175, 36, 60 },
		{ 175, 37, 60 },
		{ 175, 38, 60 },
		{ 176, 1, 60 },
		{ 176, 2, 60 },
		{ 176, 3, 60 },
		{ 176, 4, 60 },
		{ 176, 5, 60 },
		{ 176, 6, 60 },
		{ 176, 7, 60 },
		{ 176, 8, 60 },
		{ 176, 9, 60 },
		{ 176, 10, 60 },
		{ 176, 11, 60 },
		{ 176, 12, 60 },
		{ 176, 13, 60 },
		{ 176, 14, 60 },
		{ 176, 15, 60 },
		{ 176, 16, 60 },
		{ 176, 17, 60 },
		{ 176, 18, 60 },
		{ 176, 19, 60 },
		{ 176, 20, 60 },
		{ 176, 21, 60 },
		{ 176, 22, 60 },
		{ 176, 23, 60 },
		{ 176, 24, 60 },
		{ 176, 25, 60 },
		{ 176, 26, 60 },
		{ 176, 27, 60 },
		{ 176, 28, 60 },
		{ 176, 29, 60 },
		{ 176, 30, 60 },
		{ 176, 31, 60 },
		{ 176, 32, 60 },
		{ 176, 33, 60 },
		{ 176, 34, 60 },
		{ 176, 35, 60 },
		{ 176, 36, 60 },
		{ 176, 37, 60 },
		{ 176, 38, 60 },
		{ 176, 39, 60 },
		{ 176, 40, 60 },
		{ 177, 1, 60 },
		{ 177, 2, 60 },
		{ 177, 3, 60 },
		{ 177, 4, 60 },
		{ 177, 5, 60 },
		{ 177, 6, 60 },
		{ 177, 7, 60 },
		{ 177, 8, 60 },
		{ 177, 9, 60 },
		{ 177, 10, 60 },
		{ 177, 11, 60 },
		{ 177, 12, 60 },
		{ 177, 13, 60 },
		{ 177, 14, 60 },
		{ 177, 15, 60 },
		{ 177, 16, 60 },
		{ 177, 17, 60 },
		{ 177, 18, 60 },
		{ 177, 19, 60 },
		{ 177, 20, 60 },
		{ 177, 21, 60 },
		{ 177, 22, 60 },
		{ 177, 23, 60 },
		{ 177, 24, 60 },
		{ 177, 25, 60 },
		{ 177, 26, 60 },
		{ 177, 27, 60 },
		{ 177, 28, 60 },
		{ 177, 29, 60 },
		{ 177, 30, 60 },
		{ 177, 31, 60 },
		{ 177, 32, 60 },
		{ 177, 33, 60 },
		{ 177, 34, 60 },
		{ 177, 35, 60 },
		{ 177, 36, 60 },
		{ 177, 37, 60 },
		{ 177, 38, 60 },
		{ 177, 39, 60 },
		{ 177, 40, 60 },
		{ 177, 41, 60 },
		{ 177, 42, 60 },
		{ 177, 43, 60 },
		{ 178, 1, 60 },
		{ 178, 2, 60 },
		{ 178, 3, 60 },
		{ 178, 4, 60 },
		{ 178, 5, 60 },
		{ 178, 6, 60 },
		{ 178, 7, 60 },
		{ 178, 8, 60 },
		{ 178, 9, 60 },
		{ 178, 10, 60 },
		{ 178, 11, 60 },
		{ 178, 12, 60 },
		{ 178, 13, 60 },
		{ 178, 14, 60 },
		{ 178, 15, 60 },
		{ 178, 16, 60 },
		{ 178, 17, 60 },
		{ 178, 18, 60 },
		{ 178, 19, 60 },
		{ 178, 20, 60 },
		{ 178, 21, 60 },
		{ 178, 22, 60 },
		{ 178, 23, 60 },
		{ 178, 24, 60 },
		{ 178, 25, 60 },
		{ 178, 26, 60 },
		{ 178, 27, 60 },
		{ 178, 28, 60 },
		{ 178, 29, 60 },
		{ 178, 30, 60 },
		{ 178, 31, 60 },
		{ 178, 32, 60 },
		{ 178, 33, 60 },
		{ 178, 34, 60 },
		{ 178, 35, 60 },
		{ 178, 36, 60 },
		{ 178, 37, 60 },
		{ 178, 38, 60 },
		{ 178, 39, 60 },
		{ 178, 40, 60 },
		{ 178, 41, 60 },
		{ 178, 42, 60 },
		{ 178, 43, 60 },
		{ 178, 44, 60 },
		{ 178, 45, 60 },
		{ 178, 46, 60 },
		{ 179, 1, 60 },
		{ 179, 2, 60 },
		{ 179, 3, 60 },
		{ 179, 4, 60 },
		{ 179, 5, 60 },
		{ 179, 6, 60 },
		{ 179, 7, 60 },
		{ 179, 8, 60 },
		{ 179, 9, 60 },
		{ 179, 10, 60 },
		{ 179, 11, 60 },
		{ 179, 12, 60 },
		{ 179, 13, 60 },
		{ 179, 14, 60 },
		{ 179, 15, 60 },
		{ 179, 16, 60 },
		{ 179, 17, 60 },
		{ 179, 18, 60 },
		{ 179, 19, 60 },
		{ 179, 20, 60 },
		{ 179, 21, 60 },
		{ 179, 22, 60 },
		{ 179, 23, 60 },
		{ 179, 24, 60 },
		{ 179, 25, 60 },
		{ 179, 26, 60 },
		{ 179, 27, 60 },
		{ 179, 28, 60 },
		{ 179, 29, 60 },
		{ 179, 30, 60 },
		{ 179, 31, 60 },
		{ 179, 32, 60 },
		{ 179, 33, 60 },
		{ 179, 34, 60 },
		{ 179, 35, 60 },
		{ 179, 36, 60 },
		{ 179, 37, 60 },
		{ 179, 38, 60 },
		{ 179, 39, 60 },
		{ 179, 40, 60 },
		{ 179, 41, 60 },
		{ 179, 42, 60 },
		{ 179, 43, 60 },
		{ 179, 44, 60 },
		{ 179, 45, 60 },
		{ 179, 46, 60 },
		{ 179, 47, 60 },
		{ 179, 48, 60 },
		{ 180, 1, 60 },
		{ 180, 2, 60 },
		{ 180, 3, 60 },
		{ 180, 4, 60 },
		{ 180, 5, 60 },
		{ 180, 6, 60 },
		{ 180, 7, 60 },
		{ 180, 8, 60 },
		{ 180, 9, 60 },
		{ 180, 10, 60 },
		{ 180, 11, 60 },
		{ 180, 12, 60 },
		{ 180, 13, 60 },
		{ 180, 14, 60 },
		{ 180, 15, 60 },
		{ 180, 16, 60 },
		{ 180, 17, 60 },
		{ 180, 18, 60 },
		{ 180, 19, 60 },
		{ 180, 20, 60 },
		{ 180, 21, 60 },
		{ 180, 22, 60 },
		{ 180, 23, 60 },
		{ 180, 24, 60 },
		{ 180, 25, 60 },
		{ 180, 26, 60 },
		{ 180, 27, 60 },
		{ 180, 28, 60 },
		{ 180, 29, 60 },
		{ 180, 30, 60 },
		{ 180, 31, 60 },
		{ 180, 32, 60 },
		{ 180, 33, 60 },
		{ 180, 34, 60 },
		{ 180, 35, 60 },
		{ 180, 36, 60 },
		{ 180, 37, 60 },
		{ 180, 38, 60 },
		{ 180, 39, 60 },
		{ 180, 40, 60 },
		{ 180, 41, 60 },
		{ 180, 42, 60 },
		{ 180, 43, 60 },
		{ 180, 44, 60 },
		{ 180, 45, 60 },
		{ 180, 46, 60 },
		{ 180, 47, 60 },
		{ 180, 48, 60 },
		{ 180, 49, 60 },
		{ 180, 50, 60 },
		{ 180, 51, 60 },
		{ 181, 1, 60 },
		{ 181, 2, 60 },
		{ 181, 3, 60 },
		{ 181, 4, 60 },
		{ 181, 5, 60 },
		{ 181, 6, 60 },
		{ 181, 7, 60 },
		{ 181, 8, 60 },
		{ 181, 9, 60 },
		{ 181, 10, 60 },
		{ 181, 11, 60 },
		{ 181, 12, 60 },
		{ 181, 13, 60 },
		{ 181, 14, 60 },
		{ 181, 15, 60 },
		{ 181, 16, 60 },
		{ 181, 17, 60 },
		{ 181, 18, 60 },
		{ 181, 19, 60 },
		{ 181, 20, 60 },
		{ 181, 21, 60 },
		{ 181, 22, 60 },
		{ 181, 23, 60 },
		{ 181, 24, 60 },
		{ 181, 25, 60 },
		{ 181, 26, 60 },
		{ 181, 27, 60 },
		{ 181, 28, 60 },
		{ 181, 29, 60 },
		{ 181, 30, 60 },
		{ 181, 31, 60 },
		{ 181, 32, 60 },
		{ 181, 33, 60 },
		{ 181, 34, 60 },
		{ 181, 35, 60 },
		{ 181, 36, 60 },
		{ 181, 37, 60 },
		{ 181, 38, 60 },
		{ 181, 39, 60 },
		{ 181, 40, 60 },
		{ 181, 41, 60 },
		{ 181, 42, 60 },
		{ 181, 43, 60 },
		{ 181, 44, 60 },
		{ 181, 45, 60 },
		{ 181, 46, 60 },
		{ 181, 47, 60 },
		{ 181, 48, 60 },
		{ 181, 49, 60 },
		{ 181, 50, 60 },
		{ 181, 51, 60 },
		{ 181, 52, 60 },
		{ 181, 53, 60 },
		{ 181, 54, 60 },
		{ 182, 1, 60 },
		{ 182, 2, 60 },
		{ 182, 3, 60 },
		{ 182, 4, 60 },
		{ 182, 5, 60 },
		{ 182, 6, 60 },
		{ 182, 7, 60 },
		{ 182, 8, 60 },
		{ 182, 9, 60 },
		{ 182, 10, 60 },
		{ 182, 11, 60 },
		{ 182, 12, 60 },
		{ 182, 13, 60 },
		{ 182, 14, 60 },
		{ 182, 15, 60 },
		{ 182, 16, 60 },
		{ 182, 17, 60 },
		{ 182, 18, 60 },
		{ 182, 19, 60 },
		{ 182, 20, 60 },
		{ 182, 21, 60 },
		{ 182, 22, 60 },
		{ 182, 23, 60 },
		{ 182, 24, 60 },
		{ 182, 25, 60 },
		{ 182, 26, 60 },
		{ 182, 27, 60 },
		{ 182, 28, 60 },
		{ 182, 29, 60 },
		{ 182, 30, 60 },
		{ 182, 31, 60 },
		{ 182, 32, 60 },
		{ 182, 33, 60 },
		{ 182, 34, 60 },
		{ 182, 35, 60 },
		{ 182, 36, 60 },
		{ 182, 37, 60 },
		{ 182, 38, 60 },
		{ 182, 39, 60 },
		{ 182, 40, 60 },
		{ 182, 41, 60 },
		{ 182, 42, 60 },
		{ 182, 43, 60 },
		{ 182, 44, 60 },
		{ 182, 45, 60 },
		{ 182, 46, 60 },
		{ 182, 47, 60 },
		{ 182, 48, 60 },
		{ 182, 49, 60 },
		{ 182, 50, 60 },
		{ 182, 51, 60 },
		{ 182, 52, 60 },
		{ 182, 53, 60 },
		{ 182, 54, 60 },
		{ 182, 55, 60 },
		{ 182, 56, 60 },
		{ 183, 1, 60 },
		{ 183, 2, 60 },
		{ 183, 3, 60 },
		{ 183, 4, 60 },
		{ 183, 5, 60 },
		{ 183, 6, 60 },
		{ 183, 7, 60 },
		{ 183, 8, 60 },
		{ 183, 9, 60 },
		{ 183, 10, 60 },
		{ 183, 11, 60 },
		{ 183, 12, 60 },
		{ 183, 13, 60 },
		{ 183, 14, 60 },
		{ 183, 15, 60 },
		{ 183, 16, 60 },
		{ 183, 17, 60 },
		{ 183, 18, 60 },
		{ 183, 19, 60 },
		{ 183, 20, 60 },
		{ 183, 21, 60 },
		{ 183, 22, 60 },
		{ 183, 23, 60 },
		{ 183, 24, 60 },
		{ 183, 25, 60 },
		{ 183, 26, 60 },
		{ 183, 27, 60 },
		{ 183, 28, 60 },
		{ 183, 29, 60 },
		{ 183, 30, 60 },
		{ 183, 31, 60 },
		{ 183, 32, 60 },
		{ 183, 33, 60 },
		{ 183, 34, 60 },
		{ 183, 35, 60 },
		{ 183, 36, 60 },
		{ 183, 37, 60 },
		{ 183, 38, 60 },
		{ 183, 39, 60 },
		{ 183, 40, 60 },
		{ 183, 41, 60 },
		{ 183, 42, 60 },
		{ 183, 43, 60 },
		{ 183, 44, 60 },
		{ 183, 45, 60 },
		{ 183, 46, 60 },
		{ 183, 47, 60 },
		{ 183, 48, 60 },
		{ 183, 49, 60 },
		{ 183, 50, 60 },
		{ 183, 51, 60 },
		{ 183, 52, 60 },
		{ 183, 53, 60 },
		{ 183, 54, 60 },
		{ 183, 55, 60 },
		{ 183, 56, 60 },
		{ 183, 57, 60 },
		{ 183, 58, 60 },
		{ 183, 59, 60 },
		{ 184, 1, 60 },
		{ 184, 2, 60 },
		{ 184, 3, 60 },
		{ 184, 4, 60 },
		{ 184, 5, 60 },
		{ 184, 6, 60 },
		{ 184, 7, 60 },
		{ 184, 8, 60 },
		{ 184, 9, 60 },
		{ 184, 10, 60 },
		{ 184, 11, 60 },
		{ 184, 12, 60 },
		{ 184, 13, 60 },
		{ 184, 14, 60 },
		{ 184, 15, 60 },
		{ 184, 16, 60 },
		{ 184, 17, 60 },
		{ 184, 18, 60 },
		{ 184, 19, 60 },
		{ 184, 20, 60 },
		{ 184, 21, 60 },
		{ 184, 22, 60 },
		{ 184, 23, 60 },
		{ 184, 24, 60 },
		{ 184, 25, 60 },
		{ 184, 26, 60 },
		{ 184, 27, 60 },
		{ 184, 28, 60 },
		{ 184, 29, 60 },
		{ 184, 30, 60 },
		{ 184, 31, 60 },
		{ 184, 32, 60 },
		{ 184, 33, 60 },
		{ 184, 34, 60 },
		{ 184, 35, 60 },
		{ 184, 36, 60 },
		{ 184, 37, 60 },
		{ 184, 38, 60 },
		{ 184, 39, 60 },
		{ 184, 40, 60 },
		{ 184, 41, 60 },
		{ 184, 42, 60 },
		{ 184, 43, 60 },
		{ 184, 44, 60 },
		{ 184, 45, 60 },
		{ 184, 46, 60 },
		{ 184, 47, 60 },
		{ 184, 48, 60 },
		{ 184, 49, 60 },
		{ 184, 50, 60 },
		{ 184, 51, 60 },
		{ 184, 52, 60 },
		{ 184, 53, 60 },
		{ 184, 54, 60 },
		{ 184, 55, 60 },
		{ 184, 56, 60 },
		{ 184, 57, 60 },
		{ 184, 58, 60 },
		{ 184, 59, 60 },
		{ 184, 60, 60 },
		{ 184, 61, 60 },
		{ 184, 62, 60 },
		{ 185, 1, 60 },
		{ 185, 2, 60 },
		{ 185, 3, 60 },
		{ 185, 4, 60 },
		{ 185, 5, 60 },
		{ 185, 6, 60 },
		{ 185, 7, 60 },
		{ 185, 8, 60 },
		{ 185, 9, 60 },
		{ 185, 10, 60 },
		{ 185, 11, 60 },
		{ 185, 12, 60 },
		{ 185, 13, 60 },
		{ 185, 14, 60 },
		{ 185, 15, 60 },
		{ 185, 16, 60 },
		{ 185, 17, 60 },
		{ 185, 18, 60 },
		{ 185, 19, 60 },
		{ 185, 20, 60 },
		{ 185, 21, 60 },
		{ 185, 22, 60 },
		{ 185, 23, 60 },
		{ 185, 24, 60 },
		{ 185, 25, 60 },
		{ 185, 26, 60 },
		{ 185, 27, 60 },
		{ 185, 28, 60 },
		{ 185, 29, 60 },
		{ 185, 30, 60 },
		{ 185, 31, 60 },
		{ 185, 32, 60 },
		{ 185, 33, 60 },
		{ 185, 34, 60 },
		{ 185, 35, 60 },
		{ 185, 36, 60 },
		{ 185, 37, 60 },
		{ 185, 38, 60 },
		{ 185, 39, 60 },
		{ 185, 40, 60 },
		{ 185, 41, 60 },
		{ 185, 42, 60 },
		{ 185, 43, 60 },
		{ 185, 44, 60 },
		{ 185, 45, 60 },
		{ 185, 46, 60 },
		{ 185, 47, 60 },
		{ 185, 48, 60 },
		{ 185, 49, 60 },
		{ 185, 50, 60 },
		{ 185, 51, 60 },
		{ 185, 52, 60 },
		{ 185, 53, 60 },
		{ 185, 54, 60 },
		{ 185, 55, 60 },
		{ 185, 56, 60 },
		{ 185, 57, 60 },
		{ 185, 58, 60 },
		{ 185, 59, 60 },
		{ 185, 60, 60 },
		{ 185, 61, 60 },
		{ 185, 62, 60 },
		{ 185, 63, 60 },
		{ 185, 64, 60 },
		{ 186, 1, 60 },
		{ 186, 2, 60 },
		{ 186, 3, 60 },
		{ 186, 4, 60 },
		{ 186, 5, 60 },
		{ 186, 6, 60 },
		{ 186, 7, 60 },
		{ 186, 8, 60 },
		{ 186, 9, 60 },
		{ 186, 10, 60 },
		{ 186, 11, 60 },
		{ 186, 12, 60 },
		{ 186, 13, 60 },
		{ 186, 14, 60 },
		{ 186, 15, 60 },
		{ 186, 16, 60 },
		{ 186, 17, 60 },
		{ 186, 18, 60 },
		{ 186, 19, 60 },
		{ 186, 20, 60 },
		{ 186, 21, 60 },
		{ 186, 22, 60 },
		{ 186, 23, 60 },
		{ 186, 24, 60 },
		{ 186, 25, 60 },
		{ 186, 26, 60 },
		{ 186, 27, 60 },
		{ 186, 28, 60 },
		{ 186, 29, 60 },
		{ 186, 30, 60 },
		{ 186, 31, 60 },
		{ 186, 32, 60 },
		{ 186, 33, 60 },
		{ 186, 34, 60 },
		{ 186, 35, 60 },
		{ 186, 36, 60 },
		{ 186, 37, 60 },
		{ 186, 38, 60 },
		{ 186, 39, 60 },
		{ 186, 40, 60 },
		{ 186, 41, 60 },
		{ 186, 42, 60 },
		{ 186, 43, 60 },
		{ 186, 44, 60 },
		{ 186, 45, 60 },
		{ 186, 46, 60 },
		{ 186, 47, 60 },
		{ 186, 48, 60 },
		{ 186, 49, 60 },
		{ 186, 50, 60 },
		{ 186, 51, 60 },
		{ 186, 52, 60 },
		{ 186, 53, 60 },
		{ 186, 54, 60 },
		{ 186, 55, 60 },
		{ 186, 56, 60 },
		{ 186, 57, 60 },
		{ 186, 58, 60 },
		{ 186, 59, 60 },
		{ 186, 60, 60 },
		{ 186, 61, 60 },
		{ 186, 62, 60 },
		{ 186, 63, 60 },
		{ 186, 64, 60 },
		{ 186, 65, 60 },
		{ 186, 66, 60 },
		{ 186, 67, 60 },
		{ 187, 1, 60 },
		{ 187, 2, 60 },
		{ 187, 3, 60 },
		{ 187, 4, 60 },
		{ 187, 5, 60 },
		{ 187, 6, 60 },
		{ 187, 7, 60 },
		{ 187, 8, 60 },
		{ 187, 9, 60 },
		{ 187, 10, 60 },
		{ 187, 11, 60 },
		{ 187, 12, 60 },
		{ 187, 13, 60 },
		{ 187, 14, 60 },
		{ 187, 15, 60 },
		{ 187, 16, 60 },
		{ 187, 17, 60 },
		{ 187, 18, 60 },
		{ 187, 19, 60 },
		{ 187, 20, 60 },
		{ 187, 21, 60 },
		{ 187, 22, 60 },
		{ 187, 23, 60 },
		{ 187, 24, 60 },
		{ 187, 25, 60 },
		{ 187, 26, 60 },
		{ 187, 27, 60 },
		{ 187, 28, 60 },
		{ 187, 29, 60 },
		{ 187, 30, 60 },
		{ 187, 31, 60 },
		{ 187, 32, 60 },
		{ 187, 33, 60 },
		{ 187, 34, 60 },
		{ 187, 35, 60 },
		{ 187, 36, 60 },
		{ 187, 37, 60 },
		{ 187, 38, 60 },
		{ 187, 39, 60 },
		{ 187, 40, 60 },
		{ 187, 41, 60 },
		{ 187, 42, 60 },
		{ 187, 43, 60 },
		{ 187, 44, 60 },
		{ 187, 45, 60 },
		{ 187, 46, 60 },
		{ 187, 47, 60 },
		{ 187, 48, 60 },
		{ 187, 49, 60 },
		{ 187, 50, 60 },
		{ 187, 51, 60 },
		{ 187, 52, 60 },
		{ 187, 53, 60 },
		{ 187, 54, 60 },
		{ 187, 55, 60 },
		{ 187, 56, 60 },
		{ 187, 57, 60 },
		{ 187, 58, 60 },
		{ 187, 59, 60 },
		{ 187, 60, 60 },
		{ 187, 61, 60 },
		{ 187, 62, 60 },
		{ 187, 63, 60 },
		{ 187, 64, 60 },
		{ 187, 65, 60 },
		{ 187, 66, 60 },
		{ 187, 67, 60 },
		{ 187, 68, 60 },
		{ 187, 69, 60 },
		{ 187, 70, 60 },
		{ 188, 1, 60 },
		{ 188, 2, 60 },
		{ 188, 3, 60 },
		{ 188, 4, 60 },
		{ 188, 5, 60 },
		{ 188, 6, 60 },
		{ 188, 7, 60 },
		{ 188, 8, 60 },
		{ 188, 9, 60 },
		{ 188, 10, 60 },
		{ 188, 11, 60 },
		{ 188, 12, 60 },
		{ 188, 13, 60 },
		{ 188, 14, 60 },
		{ 188, 15, 60 },
		{ 188, 16, 60 },
		{ 188, 17, 60 },
		{ 188, 18, 60 },
		{ 188, 19, 60 },
		{ 188, 20, 60 },
		{ 188, 21, 60 },
		{ 188, 22, 60 },
		{ 188, 23, 60 },
		{ 188, 24, 60 },
		{ 188, 25, 60 },
		{ 188, 26, 60 },
		{ 188, 27, 60 },
		{ 188, 28, 60 },
		{ 188, 29, 60 },
		{ 188, 30, 60 },
		{ 188, 31, 60 },
		{ 188, 32, 60 },
		{ 188, 33, 60 },
		{ 188, 34, 60 },
		{ 188, 35, 60 },
		{ 188, 36, 60 },
		{ 188, 37, 60 },
		{ 188, 38, 60 },
		{ 188, 39, 60 },
		{ 188, 40, 60 },
		{ 188, 41, 60 },
		{ 188, 42, 60 },
		{ 188, 43, 60 },
		{ 188, 44, 60 },
		{ 188, 45, 60 },
		{ 188, 46, 60 },
		{ 188, 47, 60 },
		{ 188, 48, 60 },
		{ 188, 49, 60 },
		{ 188, 50, 60 },
		{ 188, 51, 60 },
		{ 188, 52, 60 },
		{ 188, 53, 60 },
		{ 188, 54, 60 },
		{ 188, 55, 60 },
		{ 188, 56, 60 },
		{ 188, 57, 60 },
		{ 188, 58, 60 },
		{ 188, 59, 60 },
		{ 188, 60, 60 },
		{ 188, 61, 60 },
		{ 188, 62, 60 },
		{ 188, 63, 60 },
		{ 188, 64, 60 },
		{ 188, 65, 60 },
		{ 188, 66, 60 },
		{ 188, 67, 60 },
		{ 188, 68, 60 },
		{ 188, 69, 60 },
		{ 188, 70, 60 },
		{ 188, 71, 60 },
		{ 188, 72, 60 },
		{ 189, 1, 60 },
		{ 189, 2, 60 },
		{ 189, 3, 60 },
		{ 189, 4, 60 },
		{ 189, 5, 60 },
		{ 189, 6, 60 },
		{ 189, 7, 60 },
		{ 189, 8, 60 },
		{ 189, 9, 60 },
		{ 189, 10, 60 },
		{ 189, 11, 60 },
		{ 189, 12, 60 },
		{ 189, 13, 60 },
		{ 189, 14, 60 },
		{ 189, 15, 60 },
		{ 189, 16, 60 },
		{ 189, 17, 60 },
		{ 189, 18, 60 },
		{ 189, 19, 60 },
		{ 189, 20, 60 },
		{ 189, 21, 60 },
		{ 189, 22, 60 },
		{ 189, 23, 60 },
		{ 189, 24, 60 },
		{ 189, 25, 60 },
		{ 189, 26, 60 },
		{ 189, 27, 60 },
		{ 189, 28, 60 },
		{ 189, 29, 60 },
		{ 189, 30, 60 },
		{ 189, 31, 60 },
		{ 189, 32, 60 },
		{ 189, 33, 60 },
		{ 189, 34, 60 },
		{ 189, 35, 60 },
		{ 189, 36, 60 },
		{ 189, 37, 60 },
		{ 189, 38, 60 },
		{ 189, 39, 60 },
		{ 189, 40, 60 },
		{ 189, 41, 60 },
		{ 189, 42, 60 },
		{ 189, 43, 60 },
		{ 189, 44, 60 },
		{ 189, 45, 60 },
		{ 189, 46, 60 },
		{ 189, 47, 60 },
		{ 189, 48, 60 },
		{ 189, 49, 60 },
		{ 189, 50, 60 },
		{ 189, 51, 60 },
		{ 189, 52, 60 },
		{ 189, 53, 60 },
		{ 189, 54, 60 },
		{ 189, 55, 60 },
		{ 189, 56, 60 },
		{ 189, 57, 60 },
		{ 189, 58, 60 },
		{ 189, 59, 60 },
		{ 189, 60, 60 },
		{ 189, 61, 60 },
		{ 189, 62, 60 },
		{ 189, 63, 60 },
		{ 189, 64, 60 },
		{ 189, 65, 60 },
		{ 189, 66, 60 },
		{ 189, 67, 60 },
		{ 189, 68, 60 },
		{ 189, 69, 60 },
		{ 189, 70, 60 },
		{ 189, 71, 60 },
		{ 189, 72, 60 },
		{ 189, 73, 60 },
		{ 189, 74, 60 },
		{ 189, 75, 60 },
		{ 190, 1, 60 },
		{ 190, 2, 60 },
		{ 190, 3, 60 },
		{ 190, 4, 60 },
		{ 190, 5, 60 },
		{ 190, 6, 60 },
		{ 190, 7, 60 },
		{ 190, 8, 60 },
		{ 190, 9, 60 },
		{ 190, 10, 60 },
		{ 190, 11, 60 },
		{ 190, 12, 60 },
		{ 190, 13, 60 },
		{ 190, 14, 60 },
		{ 190, 15, 60 },
		{ 190, 16, 60 },
		{ 190, 17, 60 },
		{ 190, 18, 60 },
		{ 190, 19, 60 },
		{ 190, 20, 60 },
		{ 190, 21, 60 },
		{ 190, 22, 60 },
		{ 190, 23, 60 },
		{ 190, 24, 60 },
		{ 190, 25, 60 },
		{ 190, 26, 60 },
		{ 190, 27, 60 },
		{ 190, 28, 60 },
		{ 190, 29, 60 },
		{ 190, 30, 60 },
		{ 190, 31, 60 },
		{ 190, 32, 60 },
		{ 190, 33, 60 },
		{ 190, 34, 60 },
		{ 190, 35, 60 },
		{ 190, 36, 60 },
		{ 190, 37, 60 },
		{ 190, 38, 60 },
		{ 190, 39, 60 },
		{ 190, 40, 60 },
		{ 190, 41, 60 },
		{ 190, 42, 60 },
		{ 190, 43, 60 },
		{ 190, 44, 60 },
		{ 190, 45, 60 },
		{ 190, 46, 60 },
		{ 190, 47, 60 },
		{ 190, 48, 60 },
		{ 190, 49, 60 },
		{ 190, 50, 60 },
		{ 190, 51, 60 },
		{ 190, 52, 60 },
		{ 190, 53, 60 },
		{ 190, 54, 60 },
		{ 190, 55, 60 },
		{ 190, 56, 60 },
		{ 190, 57, 60 },
		{ 190, 58, 60 },
		{ 190, 59, 60 },
		{ 190, 60, 60 },
		{ 190, 61, 60 },
		{ 190, 62, 60 },
		{ 190, 63, 60 },
		{ 190, 64, 60 },
		{ 190, 65, 60 },
		{ 190, 66, 60 },
		{ 190, 67, 60 },
		{ 190, 68, 60 },
		{ 190, 69, 60 },
		{ 190, 70, 60 },
		{ 190, 71, 60 },
		{ 190, 72, 60 },
		{ 190, 73, 60 },
		{ 190, 74, 60 },
		{ 190, 75, 60 },
		{ 190, 76, 60 },
		{ 190, 77, 60 },
		{ 190, 78, 60 },
		{ 191, 1, 60 },
		{ 191, 2, 60 },
		{ 191, 3, 60 },
		{ 191, 4, 60 },
		{ 191, 5, 60 },
		{ 191, 6, 60 },
		{ 191, 7, 60 },
		{ 191, 8, 60 },
		{ 191, 9, 60 },
		{ 191, 10, 60 },
		{ 191, 11, 60 },
		{ 191, 12, 60 },
		{ 191, 13, 60 },
		{ 191, 14, 60 },
		{ 191, 15, 60 },
		{ 191, 16, 60 },
		{ 191, 17, 60 },
		{ 191, 18, 60 },
		{ 191, 19, 60 },
		{ 191, 20, 60 },
		{ 191, 21, 60 },
		{ 191, 22, 60 },
		{ 191, 23, 60 },
		{ 191, 24, 60 },
		{ 191, 25, 60 },
		{ 191, 26, 60 },
		{ 191, 27, 60 },
		{ 191, 28, 60 },
		{ 191, 29, 60 },
		{ 191, 30, 60 },
		{ 191, 31, 60 },
		{ 191, 32, 60 },
		{ 191, 33, 60 },
		{ 191, 34, 60 },
		{ 191, 35, 60 },
		{ 191, 36, 60 },
		{ 191, 37, 60 },
		{ 191, 38, 60 },
		{ 191, 39, 60 },
		{ 191, 40, 60 },
		{ 191, 41, 60 },
		{ 191, 42, 60 },
		{ 191, 43, 60 },
		{ 191, 44, 60 },
		{ 191, 45, 60 },
		{ 191, 46, 60 },
		{ 191, 47, 60 },
		{ 191, 48, 60 },
		{ 191, 49, 60 },
		{ 191, 50, 60 },
		{ 191, 51, 60 },
		{ 191, 52, 60 },
		{ 191, 53, 60 },
		{ 191, 54, 60 },
		{ 191, 55, 60 },
		{ 191, 56, 60 },
		{ 191, 57, 60 },
		{ 191, 58, 60 },
		{ 191, 59, 60 },
		{ 191, 60, 60 },
		{ 191, 61, 60 },
		{ 191, 62, 60 },
		{ 191, 63, 60 },
		{ 191, 64, 60 },
		{ 191, 65, 60 },
		{ 191, 66, 60 },
		{ 191, 67, 60 },
		{ 191, 68, 60 },
		{ 191, 69, 60 },
		{ 191, 70, 60 },
		{ 191, 71, 60 },
		{ 191, 72, 60 },
		{ 191, 73, 60 },
		{ 191, 74, 60 },
		{ 191, 75, 60 },
		{ 191, 76, 60 },
		{ 191, 77, 60 },
		{ 191, 78, 60 },
		{ 191, 79, 60 },
		{ 191, 80, 60 },
		{ 192, 1, 60 },
		{ 192, 2, 60 },
		{ 192, 3, 60 },
		{ 192, 4, 60 },
		{ 192, 5, 60 },
		{ 192, 6, 60 },
		{ 192, 7, 60 },
		{ 192, 8, 60 },
		{ 192, 9, 60 },
		{ 192, 10, 60 },
		{ 192, 11, 60 },
		{ 192, 12, 60 },
		{ 192, 13, 60 },
		{ 192, 14, 60 },
		{ 192, 15, 60 },
		{ 192, 16, 60 },
		{ 192, 17, 60 },
		{ 192, 18, 60 },
		{ 192, 19, 60 },
		{ 192, 20, 60 },
		{ 192, 21, 60 },
		{ 192, 22, 60 },
		{ 192, 23, 60 },
		{ 192, 24, 60 },
		{ 192, 25, 60 },
		{ 192, 26, 60 },
		{ 192, 27, 60 },
		{ 192, 28, 60 },
		{ 192, 29, 60 },
		{ 192, 30, 60 },
		{ 192, 31, 60 },
		{ 192, 32, 60 },
		{ 192, 33, 60 },
		{ 192, 34, 60 },
		{ 192, 35, 60 },
		{ 192, 36, 60 },
		{ 192, 37, 60 },
		{ 192, 38, 60 },
		{ 192, 39, 60 },
		{ 192, 40, 60 },
		{ 192, 41, 60 },
		{ 192, 42, 60 },
		{ 192, 43, 60 },
		{ 192, 44, 60 },
		{ 192, 45, 60 },
		{ 192, 46, 60 },
		{ 192, 47, 60 },
		{ 192, 48, 60 },
		{ 192, 49, 60 },
		{ 192, 50, 60 },
		{ 192, 51, 60 },
		{ 192, 52, 60 },
		{ 192, 53, 60 },
		{ 192, 54, 60 },
		{ 192, 55, 60 },
		{ 192, 56, 60 },
		{ 192, 57, 60 },
		{ 192, 58, 60 },
		{ 192, 59, 60 },
		{ 192, 60, 60 },
		{ 192, 61, 60 },
		{ 192, 62, 60 },
		{ 192, 63, 60 },
		{ 192, 64, 60 },
		{ 192, 65, 60 },
		{ 192, 66, 60 },
		{ 192, 67, 60 },
		{ 192, 68, 60 },
		{ 192, 69, 60 },
		{ 192, 70, 60 },
		{ 192, 71, 60 },
		{ 192, 72, 60 },
		{ 192, 73, 60 },
		{ 192, 74, 60 },
		{ 192, 75, 60 },
		{ 192, 76, 60 },
		{ 192, 77, 60 },
		{ 192, 78, 60 },
		{ 192, 79, 60 },
		{ 192, 80, 60 },
		{ 192, 81, 60 },
		{ 192, 82, 60 },
		{ 192, 83, 60 },
		{ 193, 1, 60 },
		{ 193, 2, 60 },
		{ 193, 3, 60 },
		{ 193, 4, 60 },
		{ 193, 5, 60 },
		{ 193, 6, 60 },
		{ 193, 7, 60 },
		{ 193, 8, 60 },
		{ 193, 9, 60 },
		{ 193, 10, 60 },
		{ 193, 11, 60 },
		{ 193, 12, 60 },
		{ 193, 13, 60 },
		{ 193, 14, 60 },
		{ 193, 15, 60 },
		{ 193, 16, 60 },
		{ 193, 17, 60 },
		{ 193, 18, 60 },
		{ 193, 19, 60 },
		{ 193, 20, 60 },
		{ 193, 21, 60 },
		{ 193, 22, 60 },
		{ 193, 23, 60 },
		{ 193, 24, 60 },
		{ 193, 25, 60 },
		{ 193, 26, 60 },
		{ 193, 27, 60 },
		{ 193, 28, 60 },
		{ 193, 29, 60 },
		{ 193, 30, 60 },
		{ 193, 31, 60 },
		{ 193, 32, 60 },
		{ 193, 33, 60 },
		{ 193, 34, 60 },
		{ 193, 35, 60 },
		{ 193, 36, 60 },
		{ 193, 37, 60 },
		{ 193, 38, 60 },
		{ 193, 39, 60 },
		{ 193, 40, 60 },
		{ 193, 41, 60 },
		{ 193, 42, 60 },
		{ 193, 43, 60 },
		{ 193, 44, 60 },
		{ 193, 45, 60 },
		{ 193, 46, 60 },
		{ 193, 47, 60 },
		{ 193, 48, 60 },
		{ 193, 49, 60 },
		{ 193, 50, 60 },
		{ 193, 51, 60 },
		{ 193, 52, 60 },
		{ 193, 53, 60 },
		{ 193, 54, 60 },
		{ 193, 55, 60 },
		{ 193, 56, 60 },
		{ 193, 57, 60 },
		{ 193, 58, 60 },
		{ 193, 59, 60 },
		{ 193, 60, 60 },
		{ 193, 61, 60 },
		{ 193, 62, 60 },
		{ 193, 63, 60 },
		{ 193, 64, 60 },
		{ 193, 65, 60 },
		{ 193, 66, 60 },
		{ 193, 67, 60 },
		{ 193, 68, 60 },
		{ 193, 69, 60 },
		{ 193, 70, 60 },
		{ 193, 71, 60 },
		{ 193, 72, 60 },
		{ 193, 73, 60 },
		{ 193, 74, 60 },
		{ 193, 75, 60 },
		{ 193, 76, 60 },
		{ 193, 77, 60 },
		{ 193, 78, 60 },
		{ 193, 79, 60 },
		{ 193, 80, 60 },
		{ 193, 81, 60 },
		{ 193, 82, 60 },
		{ 193, 83, 60 },
		{ 193, 84, 60 },
		{ 193, 85, 60 },
		{ 193, 86, 60 },
		{ 194, 1, 60 },
		{ 194, 2, 60 },
		{ 194, 3, 60 },
		{ 194, 4, 60 },
		{ 194, 5, 60 },
		{ 194, 6, 60 },
		{ 194, 7, 60 },
		{ 194, 8, 60 },
		{ 194, 9, 60 },
		{ 194, 10, 60 },
		{ 194, 11, 60 },
		{ 194, 12, 60 },
		{ 194, 13, 60 },
		{ 194, 14, 60 },
		{ 194, 15, 60 },
		{ 194, 16, 60 },
		{ 194, 17, 60 },
		{ 194, 18, 60 },
		{ 194, 19, 60 },
		{ 194, 20, 60 },
		{ 194, 21, 60 },
		{ 194, 22, 60 },
		{ 194, 23, 60 },
		{ 194, 24, 60 },
		{ 194, 25, 60 },
		{ 194, 26, 60 },
		{ 194, 27, 60 },
		{ 194, 28, 60 },
		{ 194, 29, 60 },
		{ 194, 30, 60 },
		{ 194, 31, 60 },
		{ 194, 32, 60 },
		{ 194, 33, 60 },
		{ 194, 34, 60 },
		{ 194, 35, 60 },
		{ 194, 36, 60 },
		{ 194, 37, 60 },
		{ 194, 38, 60 },
		{ 194, 39, 60 },
		{ 194, 40, 60 },
		{ 194, 41, 60 },
		{ 194, 42, 60 },
		{ 194, 43, 60 },
		{ 194, 44, 60 },
		{ 194, 45, 60 },
		{ 194, 46, 60 },
		{ 194, 47, 60 },
		{ 194, 48, 60 },
		{ 194, 49, 60 },
		{ 194, 50, 60 },
		{ 194, 51, 60 },
		{ 194, 52, 60 },
		{ 194, 53, 60 },
		{ 194, 54, 60 },
		{ 194, 55, 60 },
		{ 194, 56, 60 },
		{ 194, 57, 60 },
		{ 194, 58, 60 },
		{ 194, 59, 60 },
		{ 194, 60, 60 },
		{ 194, 61, 60 },
		{ 194, 62, 60 },
		{ 194, 63, 60 },
		{ 194, 64, 60 },
		{ 194, 65, 60 },
		{ 194, 66, 60 },
		{ 194, 67, 60 },
		{ 194, 68, 60 },
		{ 194, 69, 60 },
		{ 194, 70, 60 },
		{ 194, 71, 60 },
		{ 194, 72, 60 },
		{ 194, 73, 60 },
		{ 194, 74, 60 },
		{ 194, 75, 60 },
		{ 194, 76, 60 },
		{ 194, 77, 60 },
		{ 194, 78, 60 },
		{ 194, 79, 60 },
		{ 194, 80, 60 },
		{ 194, 81, 60 },
		{ 194, 82, 60 },
		{ 194, 83, 60 },
		{ 194, 84, 60 },
		{ 194, 85, 60 },
		{ 194, 86, 60 },
		{ 194, 87, 60 },
		{ 194, 88, 60 },
		{ 195, 1, 60 },
		{ 195, 2, 60 },
		{ 195, 3, 60 },
		{ 195, 4, 60 },
		{ 195, 5, 60 },
		{ 195, 6, 60 },
		{ 195, 7, 60 },
		{ 195, 8, 60 },
		{ 195, 9, 60 },
		{ 195, 10, 60 },
		{ 195, 11, 60 },
		{ 195, 12, 60 },
		{ 195, 13, 60 },
		{ 195, 14, 60 },
		{ 195, 15, 60 },
		{ 195, 16, 60 },
		{ 195, 17, 60 },
		{ 195, 18, 60 },
		{ 195, 19, 60 },
		{ 195, 20, 60 },
		{ 195, 21, 60 },
		{ 195, 22, 60 },
		{ 195, 23, 60 },
		{ 195, 24, 60 },
		{ 195, 25, 60 },
		{ 195, 26, 60 },
		{ 195, 27, 60 },
		{ 195, 28, 60 },
		{ 195, 29, 60 },
		{ 195, 30, 60 },
		{ 195, 31, 60 },
		{ 195, 32, 60 },
		{ 195, 33, 60 },
		{ 195, 34, 60 },
		{ 195, 35, 60 },
		{ 195, 36, 60 },
		{ 195, 37, 60 },
		{ 195, 38, 60 },
		{ 195, 39, 60 },
		{ 195, 40, 60 },
		{ 195, 41, 60 },
		{ 195, 42, 60 },
		{ 195, 43, 60 },
		{ 195, 44, 60 },
		{ 195, 45, 60 },
		{ 195, 46, 60 },
		{ 195, 47, 60 },
		{ 195, 48, 60 },
		{ 195, 49, 60 },
		{ 195, 50, 60 },
		{ 195, 51, 60 },
		{ 195, 52, 60 },
		{ 195, 53, 60 },
		{ 195, 54, 60 },
		{ 195, 55, 60 },
		{ 195, 56, 60 },
		{ 195, 57, 60 },
		{ 195, 58, 60 },
		{ 195, 59, 60 },
		{ 195, 60, 60 },
		{ 195, 61, 60 },
		{ 195, 62, 60 },
		{ 195, 63, 60 },
		{ 195, 64, 60 },
		{ 195, 65, 60 },
		{ 195, 66, 60 },
		{ 195, 67, 60 },
		{ 195, 68, 60 },
		{ 195, 69, 60 },
		{ 195, 70, 60 },
		{ 195, 71, 60 },
		{ 195, 72, 60 },
		{ 195, 73, 60 },
		{ 195, 74, 60 },
		{ 195, 75, 60 },
		{ 195, 76, 60 },
		{ 195, 77, 60 },
		{ 195, 78, 60 },
		{ 195, 79, 60 },
		{ 195, 80, 60 },
		{ 195, 81, 60 },
		{ 195, 82, 60 },
		{ 195, 83, 60 },
		{ 195, 84, 60 },
		{ 195, 85, 60 },
		{ 195, 86, 60 },
		{ 195, 87, 60 },
		{ 195, 88, 60 },
		{ 195, 89, 60 },
		{ 195, 90, 60 },
		{ 195, 91, 60 },
		{ 196, 1, 60 },
		{ 196, 2, 60 },
		{ 196, 3, 60 },
		{ 196, 4, 60 },
		{ 196, 5, 60 },
		{ 196, 6, 60 },
		{ 196, 7, 60 },
		{ 196, 8, 60 },
		{ 196, 9, 60 },
		{ 196, 10, 60 },
		{ 196, 11, 60 },
		{ 196, 12, 60 },
		{ 196, 13, 60 },
		{ 196, 14, 60 },
		{ 196, 15, 60 },
		{ 196, 16, 60 },
		{ 196, 17, 60 },
		{ 196, 18, 60 },
		{ 196, 19, 60 },
		{ 196, 20, 60 },
		{ 196, 21, 60 },
		{ 196, 22, 60 },
		{ 196, 23, 60 },
		{ 196, 24, 60 },
		{ 196, 25, 60 },
		{ 196, 26, 60 },
		{ 196, 27, 60 },
		{ 196, 28, 60 },
		{ 196, 29, 60 },
		{ 196, 30, 60 },
		{ 196, 31, 60 },
		{ 196, 32, 60 },
		{ 196, 33, 60 },
		{ 196, 34, 60 },
		{ 196, 35, 60 },
		{ 196, 36, 60 },
		{ 196, 37, 60 },
		{ 196, 38, 60 },
		{ 196, 39, 60 },
		{ 196, 40, 60 },
		{ 196, 41, 60 },
		{ 196, 42, 60 },
		{ 196, 43, 60 },
		{ 196, 44, 60 },
		{ 196, 45, 60 },
		{ 196, 46, 60 },
		{ 196, 47, 60 },
		{ 196, 48, 60 },
		{ 196, 49, 60 },
		{ 196, 50, 60 },
		{ 196, 51, 60 },
		{ 196, 52, 60 },
		{ 196, 53, 60 },
		{ 196, 54, 60 },
		{ 196, 55, 60 },
		{ 196, 56, 60 },
		{ 196, 57, 60 },
		{ 196, 58, 60 },
		{ 196, 59, 60 },
		{ 196, 60, 60 },
		{ 196, 61, 60 },
		{ 196, 62, 60 },
		{ 196, 63, 60 },
		{ 196, 64, 60 },
		{ 196, 65, 60 },
		{ 196, 66, 60 },
		{ 196, 67, 60 },
		{ 196, 68, 60 },
		{ 196, 69, 60 },
		{ 196, 70, 60 },
		{ 196, 71, 60 },
		{ 196, 72, 60 },
		{ 196, 73, 60 },
		{ 196, 74, 60 },
		{ 196, 75, 60 },
		{ 196, 76, 60 },
		{ 196, 77, 60 },
		{ 196, 78, 60 },
		{ 196, 79, 60 },
		{ 196, 80, 60 },
		{ 196, 81, 60 },
		{ 196, 82, 60 },
		{ 196, 83, 60 },
		{ 196, 84, 60 },
		{ 196, 85, 60 },
		{ 196, 86, 60 },
		{ 196, 87, 60 },
		{ 196, 88, 60 },
		{ 196, 89, 60 },
		{ 196, 90, 60 },
		{ 196, 91, 60 },
		{ 196, 92, 60 },
		{ 196, 93, 60 },
		{ 196, 94, 60 },
		{ 197, 1, 60 },
		{ 197, 2, 60 },
		{ 197, 3, 60 },
		{ 197, 4, 60 },
		{ 197, 5, 60 },
		{ 197, 6, 60 },
		{ 197, 7, 60 },
		{ 197, 8, 60 },
		{ 197, 9, 60 },
		{ 197, 10, 60 },
		{ 197, 11, 60 },
		{ 197, 12, 60 },
		{ 197, 13, 60 },
		{ 197, 14, 60 },
		{ 197, 15, 60 },
		{ 197, 16, 60 },
		{ 197, 17, 60 },
		{ 197, 18, 60 },
		{ 197, 19, 60 },
		{ 197, 20, 60 },
		{ 197, 21, 60 },
		{ 197, 22, 60 },
		{ 197, 23, 60 },
		{ 197, 24, 60 },
		{ 197, 25, 60 },
		{ 197, 26, 60 },
		{ 197, 27, 60 },
		{ 197, 28, 60 },
		{ 197, 29, 60 },
		{ 197, 30, 60 },
		{ 197, 31, 60 },
		{ 197, 32, 60 },
		{ 197, 33, 60 },
		{ 197, 34, 60 },
		{ 197, 35, 60 },
		{ 197, 36, 60 },
		{ 197, 37, 60 },
		{ 197, 38, 60 },
		{ 197, 39, 60 },
		{ 197, 40, 60 },
		{ 197, 41, 60 },
		{ 197, 42, 60 },
		{ 197, 43, 60 },
		{ 197, 44, 60 },
		{ 197, 45, 60 },
		{ 197, 46, 60 },
		{ 197, 47, 60 },
		{ 197, 48, 60 },
		{ 197, 49, 60 },
		{ 197, 50, 60 },
		{ 197, 51, 60 },
		{ 197, 52, 60 },
		{ 197, 53, 60 },
		{ 197, 54, 60 },
		{ 197, 55, 60 },
		{ 197, 56, 60 },
		{ 197, 57, 60 },
		{ 197, 58, 60 },
		{ 197, 59, 60 },
		{ 197, 60, 60 },
		{ 197, 61, 60 },
		{ 197, 62, 60 },
		{ 197, 63, 60 },
		{ 197, 64, 60 },
		{ 197, 65, 60 },
		{ 197, 66, 60 },
		{ 197, 67, 60 },
		{ 197, 68, 60 },
		{ 197, 69, 60 },
		{ 197, 70, 60 },
		{ 197, 71, 60 },
		{ 197, 72, 60 },
		{ 197, 73, 60 },
		{ 197, 74, 60 },
		{ 197, 75, 60 },
		{ 197, 76, 60 },
		{ 197, 77, 60 },
		{ 197, 78, 60 },
		{ 197, 79, 60 },
		{ 197, 80, 60 },
		{ 197, 81, 60 },
		{ 197, 82, 60 },
		{ 197, 83, 60 },
		{ 197, 84, 60 },
		{ 197, 85, 60 },
		{ 197, 86, 60 },
		{ 197, 87, 60 },
		{ 197, 88, 60 },
		{ 197, 89, 60 },
		{ 197, 90, 60 },
		{ 197, 91, 60 },
		{ 197, 92, 60 },
		{ 197, 93, 60 },
		{ 197, 94, 60 },
		{ 197, 95, 60 },
		{ 197, 96, 60 },
		{ 198, 1, 60 },
		{ 198, 2, 60 },
		{ 198, 3, 60 },
		{ 198, 4, 60 },
		{ 198, 5, 60 },
		{ 198, 6, 60 },
		{ 198, 7, 60 },
		{ 198, 8, 60 },
		{ 198, 9, 60 },
		{ 198, 10, 60 },
		{ 198, 11, 60 },
		{ 198, 12, 60 },
		{ 198, 13, 60 },
		{ 198, 14, 60 },
		{ 198, 15, 60 },
		{ 198, 16, 60 },
		{ 198, 17, 60 },
		{ 198, 18, 60 },
		{ 198, 19, 60 },
		{ 198, 20, 60 },
		{ 198, 21, 60 },
		{ 198, 22, 60 },
		{ 198, 23, 60 },
		{ 198, 24, 60 },
		{ 198, 25, 60 },
		{ 198, 26, 60 },
		{ 198, 27, 60 },
		{ 198, 28, 60 },
		{ 198, 29, 60 },
		{ 198, 30, 60 },
		{ 198, 31, 60 },
		{ 198, 32, 60 },
		{ 198, 33, 60 },
		{ 198, 34, 60 },
		{ 198, 35, 60 },
		{ 198, 36, 60 },
		{ 198, 37, 60 },
		{ 198, 38, 60 },
		{ 198, 39, 60 },
		{ 198, 40, 60 },
		{ 198, 41, 60 },
		{ 198, 42, 60 },
		{ 198, 43, 60 },
		{ 198, 44, 60 },
		{ 198, 45, 60 },
		{ 198, 46, 60 },
		{ 198, 47, 60 },
		{ 198, 48, 60 },
		{ 198, 49, 60 },
		{ 198, 50, 60 },
		{ 198, 51, 60 },
		{ 198, 52, 60 },
		{ 198, 53, 60 },
		{ 198, 54, 60 },
		{ 198, 55, 60 },
		{ 198, 56, 60 },
		{ 198, 57, 60 },
		{ 198, 58, 60 },
		{ 198, 59, 60 },
		{ 198, 60, 60 },
		{ 198, 61, 60 },
		{ 198, 62, 60 },
		{ 198, 63, 60 },
		{ 198, 64, 60 },
		{ 198, 65, 60 },
		{ 198, 66, 60 },
		{ 198, 67, 60 },
		{ 198, 68, 60 },
		{ 198, 69, 60 },
		{ 198, 70, 60 },
		{ 198, 71, 60 },
		{ 198, 72, 60 },
		{ 198, 73, 60 },
		{ 198, 74, 60 },
		{ 198, 75, 60 },
		{ 198, 76, 60 },
		{ 198, 77, 60 },
		{ 198, 78, 60 },
		{ 198, 79, 60 },
		{ 198, 80, 60 },
		{ 198, 81, 60 },
		{ 198, 82, 60 },
		{ 198, 83, 60 },
		{ 198, 84, 60 },
		{ 198, 85, 60 },
		{ 198, 86, 60 },
		{ 198, 87, 60 },
		{ 198, 88, 60 },
		{ 198, 89, 60 },
		{ 198, 90, 60 },
		{ 198, 91, 60 },
		{ 198, 92, 60 },
		{ 198, 93, 60 },
		{ 198, 94, 60 },
		{ 198, 95, 60 },
		{ 198, 96, 60 },
		{ 198, 97, 60 },
		{ 198, 98, 60 },
		{ 198, 99, 60 },
		{ 199, 1, 60 },
		{ 199, 2, 60 },
		{ 199, 3, 60 },
		{ 199, 4, 60 },
		{ 199, 5, 60 },
		{ 199, 6, 60 },
		{ 199, 7, 60 },
		{ 199, 8, 60 },
		{ 199, 9, 60 },
		{ 199, 10, 60 },
		{ 199, 11, 60 },
		{ 199, 12, 60 },
		{ 199, 13, 60 },
		{ 199, 14, 60 },
		{ 199, 15, 60 },
		{ 199, 16, 60 },
		{ 199, 17, 60 },
		{ 199, 18, 60 },
		{ 199, 19, 60 },
		{ 199, 20, 60 },
		{ 199, 21, 60 },
		{ 199, 22, 60 },
		{ 199, 23, 60 },
		{ 199, 24, 60 },
		{ 199, 25, 60 },
		{ 199, 26, 60 },
		{ 199, 27, 60 },
		{ 199, 28, 60 },
		{ 199, 29, 60 },
		{ 199, 30, 60 },
		{ 199, 31, 60 },
		{ 199, 32, 60 },
		{ 199, 33, 60 },
		{ 199, 34, 60 },
		{ 199, 35, 60 },
		{ 199, 36, 60 },
		{ 199, 37, 60 },
		{ 199, 38, 60 },
		{ 199, 39, 60 },
		{ 199, 40, 60 },
		{ 199, 41, 60 },
		{ 199, 42, 60 },
		{ 199, 43, 60 },
		{ 199, 44, 60 },
		{ 199, 45, 60 },
		{ 199, 46, 60 },
		{ 199, 47, 60 },
		{ 199, 48, 60 },
		{ 199, 49, 60 },
		{ 199, 50, 60 },
		{ 199, 51, 60 },
		{ 199, 52, 60 },
		{ 199, 53, 60 },
		{ 199, 54, 60 },
		{ 199, 55, 60 },
		{ 199, 56, 60 },
		{ 199, 57, 60 },
		{ 199, 58, 60 },
		{ 199, 59, 60 },
		{ 199, 60, 60 },
		{ 199, 61, 60 },
		{ 199, 62, 60 },
		{ 199, 63, 60 },
		{ 199, 64, 60 },
		{ 199, 65, 60 },
		{ 199, 66, 60 },
		{ 199, 67, 60 },
		{ 199, 68, 60 },
		{ 199, 69, 60 },
		{ 199, 70, 60 },
		{ 199, 71, 60 },
		{ 199, 72, 60 },
		{ 199, 73, 60 },
		{ 199, 74, 60 },
		{ 199, 75, 60 },
		{ 199, 76, 60 },
		{ 199, 77, 60 },
		{ 199, 78, 60 },
		{ 199, 79, 60 },
		{ 199, 80, 60 },
		{ 199, 81, 60 },
		{ 199, 82, 60 },
		{ 199, 83, 60 },
		{ 199, 84, 60 },
		{ 199, 85, 60 },
		{ 199, 86, 60 },
		{ 199, 87, 60 },
		{ 199, 88, 60 },
		{ 199, 89, 60 },
		{ 199, 90, 60 },
		{ 199, 91, 60 },
		{ 199, 92, 60 },
		{ 199, 93, 60 },
		{ 199, 94, 60 },
		{ 199, 95, 60 },
		{ 199, 96, 60 },
		{ 199, 97, 60 },
		{ 199, 98, 60 },
		{ 199, 99, 60 },
		{ 199, 100, 60 },
		{ 199, 101, 60 },
		{ 199, 102, 60 },
		{ 200, 1, 60 },
		{ 200, 2, 60 },
		{ 200, 3, 60 },
		{ 200, 4, 60 },
		{ 200, 5, 60 },
		{ 200, 6, 60 },
		{ 200, 7, 60 },
		{ 200, 8, 60 },
		{ 200, 9, 60 },
		{ 200, 10, 60 },
		{ 200, 11, 60 },
		{ 200, 12, 60 },
		{ 200, 13, 60 },
		{ 200, 14, 60 },
		{ 200, 15, 60 },
		{ 200, 16, 60 },
		{ 200, 17, 60 },
		{ 200, 18, 60 },
		{ 200, 19, 60 },
		{ 200, 20, 60 },
		{ 200, 21, 60 },
		{ 200, 22, 60 },
		{ 200, 23, 60 },
		{ 200, 24, 60 },
		{ 200, 25, 60 },
		{ 200, 26, 60 },
		{ 200, 27, 60 },
		{ 200, 28, 60 },
		{ 200, 29, 60 },
		{ 200, 30, 60 },
		{ 200, 31, 60 },
		{ 200, 32, 60 },
		{ 200, 33, 60 },
		{ 200, 34, 60 },
		{ 200, 35, 60 },
		{ 200, 36, 60 },
		{ 200, 37, 60 },
		{ 200, 38, 60 },
		{ 200, 39, 60 },
		{ 200, 40, 60 },
		{ 200, 41, 60 },
		{ 200, 42, 60 },
		{ 200, 43, 60 },
		{ 200, 44, 60 },
		{ 200, 45, 60 },
		{ 200, 46, 60 },
		{ 200, 47, 60 },
		{ 200, 48, 60 },
		{ 200, 49, 60 },
		{ 200, 50, 60 },
		{ 200, 51, 60 },
		{ 200, 52, 60 },
		{ 200, 53, 60 },
		{ 200, 54, 60 },
		{ 200, 55, 60 },
		{ 200, 56, 60 },
		{ 200, 57, 60 },
		{ 200, 58, 60 },
		{ 200, 59, 60 },
		{ 200, 60, 60 },
		{ 200, 61, 60 },
		{ 200, 62, 60 },
		{ 200, 63, 60 },
		{ 200, 64, 60 },
		{ 200, 65, 60 },
		{ 200, 66, 60 },
		{ 200, 67, 60 },
		{ 200, 68, 60 },
		{ 200, 69, 60 },
		{ 200, 70, 60 },
		{ 200, 71, 60 },
		{ 200, 72, 60 },
		{ 200, 73, 60 },
		{ 200, 74, 60 },
		{ 200, 75, 60 },
		{ 200, 76, 60 },
		{ 200, 77, 60 },
		{ 200, 78, 60 },
		{ 200, 79, 60 },
		{ 200, 80, 60 },
		{ 200, 81, 60 },
		{ 200, 82, 60 },
		{ 200, 83, 60 },
		{ 200, 84, 60 },
		{ 200, 85, 60 },
		{ 200, 86, 60 },
		{ 200, 87, 60 },
		{ 200, 88, 60 },
		{ 200, 89, 60 },
		{ 200, 90, 60 },
		{ 200, 91, 60 },
		{ 200, 92, 60 },
		{ 200, 93, 60 },
		{ 200, 94, 60 },
		{ 200, 95, 60 },
		{ 200, 96, 60 },
		{ 200, 97, 60 },
		{ 200, 98, 60 },
		{ 200, 99, 60 },
		{ 200, 100, 60 },
		{ 200, 101, 60 },
		{ 200, 102, 60 },
		{ 200, 103, 60 },
		{ 200, 104, 60 },
		{ 201, 1, 60 },
		{ 201, 2, 60 },
		{ 201, 3, 60 },
		{ 201, 4, 60 },
		{ 201, 5, 60 },
		{ 201, 6, 60 },
		{ 201, 7, 60 },
		{ 201, 8, 60 },
		{ 201, 9, 60 },
		{ 201, 10, 60 },
		{ 201, 11, 60 },
		{ 201, 12, 60 },
		{ 201, 13, 60 },
		{ 201, 14, 60 },
		{ 201, 15, 60 },
		{ 201, 16, 60 },
		{ 201, 17, 60 },
		{ 201, 18, 60 },
		{ 201, 19, 60 },
		{ 201, 20, 60 },
		{ 201, 21, 60 },
		{ 201, 22, 60 },
		{ 201, 23, 60 },
		{ 201, 24, 60 },
		{ 201, 25, 60 },
		{ 201, 26, 60 },
		{ 201, 27, 60 },
		{ 201, 28, 60 },
		{ 201, 29, 60 },
		{ 201, 30, 60 },
		{ 201, 31, 60 },
		{ 201, 32, 60 },
		{ 201, 33, 60 },
		{ 201, 34, 60 },
		{ 201, 35, 60 },
		{ 201, 36, 60 },
		{ 201, 37, 60 },
		{ 201, 38, 60 },
		{ 201, 39, 60 },
		{ 201, 40, 60 },
		{ 201, 41, 60 },
		{ 201, 42, 60 },
		{ 201, 43, 60 },
		{ 201, 44, 60 },
		{ 201, 45, 60 },
		{ 201, 46, 60 },
		{ 201, 47, 60 },
		{ 201, 48, 60 },
		{ 201, 49, 60 },
		{ 201, 50, 60 },
		{ 201, 51, 60 },
		{ 201, 52, 60 },
		{ 201, 53, 60 },
		{ 201, 54, 60 },
		{ 201, 55, 60 },
		{ 201, 56, 60 },
		{ 201, 57, 60 },
		{ 201, 58, 60 },
		{ 201, 59, 60 },
		{ 201, 60, 60 },
		{ 201, 61, 60 },
		{ 201, 62, 60 },
		{ 201, 63, 60 },
		{ 201, 64, 60 },
		{ 201, 65, 60 },
		{ 201, 66, 60 },
		{ 201, 67, 60 },
		{ 201, 68, 60 },
		{ 201, 69, 60 },
		{ 201, 70, 60 },
		{ 201, 71, 60 },
		{ 201, 72, 60 },
		{ 201, 73, 60 },
		{ 201, 74, 60 },
		{ 201, 75, 60 },
		{ 201, 76, 60 },
		{ 201, 77, 60 },
		{ 201, 78, 60 },
		{ 201, 79, 60 },
		{ 201, 80, 60 },
		{ 201, 81, 60 },
		{ 201, 82, 60 },
		{ 201, 83, 60 },
		{ 201, 84, 60 },
		{ 201, 85, 60 },
		{ 201, 86, 60 },
		{ 201, 87, 60 },
		{ 201, 88, 60 },
		{ 201, 89, 60 },
		{ 201, 90, 60 },
		{ 201, 91, 60 },
		{ 201, 92, 60 },
		{ 201, 93, 60 },
		{ 201, 94, 60 },
		{ 201, 95, 60 },
		{ 201, 96, 60 },
		{ 201, 97, 60 },
		{ 201, 98, 60 },
		{ 201, 99, 60 },
		{ 201, 100, 60 },
		{ 201, 101, 60 },
		{ 201, 102, 60 },
		{ 201, 103, 60 },
		{ 201, 104, 60 },
		{ 201, 105, 60 },
		{ 201, 106, 60 },
		{ 201, 107, 60 },
		{ 202, 1, 60 },
		{ 202, 2, 60 },
		{ 202, 3, 60 },
		{ 202, 4, 60 },
		{ 202, 5, 60 },
		{ 202, 6, 60 },
		{ 202, 7, 60 },
		{ 202, 8, 60 },
		{ 202, 9, 60 },
		{ 202, 10, 60 },
		{ 202, 11, 60 },
		{ 202, 12, 60 },
		{ 202, 13, 60 },
		{ 202, 14, 60 },
		{ 202, 15, 60 },
		{ 202, 16, 60 },
		{ 202, 17, 60 },
		{ 202, 18, 60 },
		{ 202, 19, 60 },
		{ 202, 20, 60 },
		{ 202, 21, 60 },
		{ 202, 22, 60 },
		{ 202, 23, 60 },
		{ 202, 24, 60 },
		{ 202, 25, 60 },
		{ 202, 26, 60 },
		{ 202, 27, 60 },
		{ 202, 28, 60 },
		{ 202, 29, 60 },
		{ 202, 30, 60 },
		{ 202, 31, 60 },
		{ 202, 32, 60 },
		{ 202, 33, 60 },
		{ 202, 34, 60 },
		{ 202, 35, 60 },
		{ 202, 36, 60 },
		{ 202, 37, 60 },
		{ 202, 38, 60 },
		{ 202, 39, 60 },
		{ 202, 40, 60 },
		{ 202, 41, 60 },
		{ 202, 42, 60 },
		{ 202, 43, 60 },
		{ 202, 44, 60 },
		{ 202, 45, 60 },
		{ 202, 46, 60 },
		{ 202, 47, 60 },
		{ 202, 48, 60 },
		{ 202, 49, 60 },
		{ 202, 50, 60 },
		{ 202, 51, 60 },
		{ 202, 52, 60 },
		{ 202, 53, 60 },
		{ 202, 54, 60 },
		{ 202, 55, 60 },
		{ 202, 56, 60 },
		{ 202, 57, 60 },
		{ 202, 58, 60 },
		{ 202, 59, 60 },
		{ 202, 60, 60 },
		{ 202, 61, 60 },
		{ 202, 62, 60 },
		{ 202, 63, 60 },
		{ 202, 64, 60 },
		{ 202, 65, 60 },
		{ 202, 66, 60 },
		{ 202, 67, 60 },
		{ 202, 68, 60 },
		{ 202, 69, 60 },
		{ 202, 70, 60 },
		{ 202, 71, 60 },
		{ 202, 72, 60 },
		{ 202, 73, 60 },
		{ 202, 74, 60 },
		{ 202, 75, 60 },
		{ 202, 76, 60 },
		{ 202, 77, 60 },
		{ 202, 78, 60 },
		{ 202, 79, 60 },
		{ 202, 80, 60 },
		{ 202, 81, 60 },
		{ 202, 82, 60 },
		{ 202, 83, 60 },
		{ 202, 84, 60 },
		{ 202, 85, 60 },
		{ 202, 86, 60 },
		{ 202, 87, 60 },
		{ 202, 88, 60 },
		{ 202, 89, 60 },
		{ 202, 90, 60 },
		{ 202, 91, 60 },
		{ 202, 92, 60 },
		{ 202, 93, 60 },
		{ 202, 94, 60 },
		{ 202, 95, 60 },
		{ 202, 96, 60 },
		{ 202, 97, 60 },
		{ 202, 98, 60 },
		{ 202, 99, 60 },
		{ 202, 100, 60 },
		{ 202, 101, 60 },
		{ 202, 102, 60 },
		{ 202, 103, 60 },
		{ 202, 104, 60 },
		{ 202, 105, 60 },
		{ 202, 106, 60 },
		{ 202, 107, 60 },
		{ 202, 108, 60 },
		{ 202, 109, 60 },
		{ 202, 110, 60 },
		{ 203, 1, 60 },
		{ 203, 2, 60 },
		{ 203, 3, 60 },
		{ 203, 4, 60 },
		{ 203, 5, 60 },
		{ 203, 6, 60 },
		{ 203, 7, 60 },
		{ 203, 8, 60 },
		{ 203, 9, 60 },
		{ 203, 10, 60 },
		{ 203, 11, 60 },
		{ 203, 12, 60 },
		{ 203, 13, 60 },
		{ 203, 14, 60 },
		{ 203, 15, 60 },
		{ 203, 16, 60 },
		{ 203, 17, 60 },
		{ 203, 18, 60 },
		{ 203, 19, 60 },
		{ 203, 20, 60 },
		{ 203, 21, 60 },
		{ 203, 22, 60 },
		{ 203, 23, 60 },
		{ 203, 24, 60 },
		{ 203, 25, 60 },
		{ 203, 26, 60 },
		{ 203, 27, 60 },
		{ 203, 28, 60 },
		{ 203, 29, 60 },
		{ 203, 30, 60 },
		{ 203, 31, 60 },
		{ 203, 32, 60 },
		{ 203, 33, 60 },
		{ 203, 34, 60 },
		{ 203, 35, 60 },
		{ 203, 36, 60 },
		{ 203, 37, 60 },
		{ 203, 38, 60 },
		{ 203, 39, 60 },
		{ 203, 40, 60 },
		{ 203, 41, 60 },
		{ 203, 42, 60 },
		{ 203, 43, 60 },
		{ 203, 44, 60 },
		{ 203, 45, 60 },
		{ 203, 46, 60 },
		{ 203, 47, 60 },
		{ 203, 48, 60 },
		{ 203, 49, 60 },
		{ 203, 50, 60 },
		{ 203, 51, 60 },
		{ 203, 52, 60 },
		{ 203, 53, 60 },
		{ 203, 54, 60 },
		{ 203, 55, 60 },
		{ 203, 56, 60 },
		{ 203, 57, 60 },
		{ 203, 58, 60 },
		{ 203, 59, 60 },
		{ 203, 60, 60 },
		{ 203, 61, 60 },
		{ 203, 62, 60 },
		{ 203, 63, 60 },
		{ 203, 64, 60 },
		{ 203, 65, 60 },
		{ 203, 66, 60 },
		{ 203, 67, 60 },
		{ 203, 68, 60 },
		{ 203, 69, 60 },
		{ 203, 70, 60 },
		{ 203, 71, 60 },
		{ 203, 72, 60 },
		{ 203, 73, 60 },
		{ 203, 74, 60 },
		{ 203, 75, 60 },
		{ 203, 76, 60 },
		{ 203, 77, 60 },
		{ 203, 78, 60 },
		{ 203, 79, 60 },
		{ 203, 80, 60 },
		{ 203, 81, 60 },
		{ 203, 82, 60 },
		{ 203, 83, 60 },
		{ 203, 84, 60 },
		{ 203, 85, 60 },
		{ 203, 86, 60 },
		{ 203, 87, 60 },
		{ 203, 88, 60 },
		{ 203, 89, 60 },
		{ 203, 90, 60 },
		{ 203, 91, 60 },
		{ 203, 92, 60 },
		{ 203, 93, 60 },
		{ 203, 94, 60 },
		{ 203, 95, 60 },
		{ 203, 96, 60 },
		{ 203, 97, 60 },
		{ 203, 98, 60 },
		{ 203, 99, 60 },
		{ 203, 100, 60 },
		{ 203, 101, 60 },
		{ 203, 102, 60 },
		{ 203, 103, 60 },
		{ 203, 104, 60 },
		{ 203, 105, 60 },
		{ 203, 106, 60 },
		{ 203, 107, 60 },
		{ 203, 108, 60 },
		{ 203, 109, 60 },
		{ 203, 110, 60 },
		{ 203, 111, 60 },
		{ 203, 112, 60 },
		{ 204, 1, 60 },
		{ 204, 2, 60 },
		{ 204, 3, 60 },
		{ 204, 4, 60 },
		{ 204, 5, 60 },
		{ 204, 6, 60 },
		{ 204, 7, 60 },
		{ 204, 8, 60 },
		{ 204, 9, 60 },
		{ 204, 10, 60 },
		{ 204, 11, 60 },
		{ 204, 12, 60 },
		{ 204, 13, 60 },
		{ 204, 14, 60 },
		{ 204, 15, 60 },
		{ 204, 16, 60 },
		{ 204, 17, 60 },
		{ 204, 18, 60 },
		{ 204, 19, 60 },
		{ 204, 20, 60 },
		{ 204, 21, 60 },
		{ 204, 22, 60 },
		{ 204, 23, 60 },
		{ 204, 24, 60 },
		{ 204, 25, 60 },
		{ 204, 26, 60 },
		{ 204, 27, 60 },
		{ 204, 28, 60 },
		{ 204, 29, 60 },
		{ 204, 30, 60 },
		{ 204, 31, 60 },
		{ 204, 32, 60 },
		{ 204, 33, 60 },
		{ 204, 34, 60 },
		{ 204, 35, 60 },
		{ 204, 36, 60 },
		{ 204, 37, 60 },
		{ 204, 38, 60 },
		{ 204, 39, 60 },
		{ 204, 40, 60 },
		{ 204, 41, 60 },
		{ 204, 42, 60 },
		{ 204, 43, 60 },
		{ 204, 44, 60 },
		{ 204, 45, 60 },
		{ 204, 46, 60 },
		{ 204, 47, 60 },
		{ 204, 48, 60 },
		{ 204, 49, 60 },
		{ 204, 50, 60 },
		{ 204, 51, 60 },
		{ 204, 52, 60 },
		{ 204, 53, 60 },
		{ 204, 54, 60 },
		{ 204, 55, 60 },
		{ 204, 56, 60 },
		{ 204, 57, 60 },
		{ 204, 58, 60 },
		{ 204, 59, 60 },
		{ 204, 60, 60 },
		{ 204, 61, 60 },
		{ 204, 62, 60 },
		{ 204, 63, 60 },
		{ 204, 64, 60 },
		{ 204, 65, 60 },
		{ 204, 66, 60 },
		{ 204, 67, 60 },
		{ 204, 68, 60 },
		{ 204, 69, 60 },
		{ 204, 70, 60 },
		{ 204, 71, 60 },
		{ 204, 72, 60 },
		{ 204, 73, 60 },
		{ 204, 74, 60 },
		{ 204, 75, 60 },
		{ 204, 76, 60 },
		{ 204, 77, 60 },
		{ 204, 78, 60 },
		{ 204, 79, 60 },
		{ 204, 80, 60 },
		{ 204, 81, 60 },
		{ 204, 82, 60 },
		{ 204, 83, 60 },
		{ 204, 84, 60 },
		{ 204, 85, 60 },
		{ 204, 86, 60 },
		{ 204, 87, 60 },
		{ 204, 88, 60 },
		{ 204, 89, 60 },
		{ 204, 90, 60 },
		{ 204, 91, 60 },
		{ 204, 92, 60 },
		{ 204, 93, 60 },
		{ 204, 94, 60 },
		{ 204, 95, 60 },
		{ 204, 96, 60 },
		{ 204, 97, 60 },
		{ 204, 98, 60 },
		{ 204, 99, 60 },
		{ 204, 100, 60 },
		{ 204, 101, 60 },
		{ 204, 102, 60 },
		{ 204, 103, 60 },
		{ 204, 104, 60 },
		{ 204, 105, 60 },
		{ 204, 106, 60 },
		{ 204, 107, 60 },
		{ 204, 108, 60 },
		{ 204, 109, 60 },
		{ 204, 110, 60 },
		{ 204, 111, 60 },
		{ 204, 112, 60 },
		{ 204, 113, 60 },
		{ 204, 114, 60 },
		{ 204, 115, 60 },
		{ 205, 1, 60 },
		{ 205, 2, 60 },
		{ 205, 3, 60 },
		{ 205, 4, 60 },
		{ 205, 5, 60 },
		{ 205, 6, 60 },
		{ 205, 7, 60 },
		{ 205, 8, 60 },
		{ 205, 9, 60 },
		{ 205, 10, 60 },
		{ 205, 11, 60 },
		{ 205, 12, 60 },
		{ 205, 13, 60 },
		{ 205, 14, 60 },
		{ 205, 15, 60 },
		{ 205, 16, 60 },
		{ 205, 17, 60 },
		{ 205, 18, 60 },
		{ 205, 19, 60 },
		{ 205, 20, 60 },
		{ 205, 21, 60 },
		{ 205, 22, 60 },
		{ 205, 23, 60 },
		{ 205, 24, 60 },
		{ 205, 25, 60 },
		{ 205, 26, 60 },
		{ 205, 27, 60 },
		{ 205, 28, 60 },
		{ 205, 29, 60 },
		{ 205, 30, 60 },
		{ 205, 31, 60 },
		{ 205, 32, 60 },
		{ 205, 33, 60 },
		{ 205, 34, 60 },
		{ 205, 35, 60 },
		{ 205, 36, 60 },
		{ 205, 37, 60 },
		{ 205, 38, 60 },
		{ 205, 39, 60 },
		{ 205, 40, 60 },
		{ 205, 41, 60 },
		{ 205, 42, 60 },
		{ 205, 43, 60 },
		{ 205, 44, 60 },
		{ 205, 45, 60 },
		{ 205, 46, 60 },
		{ 205, 47, 60 },
		{ 205, 48, 60 },
		{ 205, 49, 60 },
		{ 205, 50, 60 },
		{ 205, 51, 60 },
		{ 205, 52, 60 },
		{ 205, 53, 60 },
		{ 205, 54, 60 },
		{ 205, 55, 60 },
		{ 205, 56, 60 },
		{ 205, 57, 60 },
		{ 205, 58, 60 },
		{ 205, 59, 60 },
		{ 205, 60, 60 },
		{ 205, 61, 60 },
		{ 205, 62, 60 },
		{ 205, 63, 60 },
		{ 205, 64, 60 },
		{ 205, 65, 60 },
		{ 205, 66, 60 },
		{ 205, 67, 60 },
		{ 205, 68, 60 },
		{ 205, 69, 60 },
		{ 205, 70, 60 },
		{ 205, 71, 60 },
		{ 205, 72, 60 },
		{ 205, 73, 60 },
		{ 205, 74, 60 },
		{ 205, 75, 60 },
		{ 205, 76, 60 },
		{ 205, 77, 60 },
		{ 205, 78, 60 },
		{ 205, 79, 60 },
		{ 205, 80, 60 },
		{ 205, 81, 60 },
		{ 205, 82, 60 },
		{ 205, 83, 60 },
		{ 205, 84, 60 },
		{ 205, 85, 60 },
		{ 205, 86, 60 },
		{ 205, 87, 60 },
		{ 205, 88, 60 },
		{ 205, 89, 60 },
		{ 205, 90, 60 },
		{ 205, 91, 60 },
		{ 205, 92, 60 },
		{ 205, 93, 60 },
		{ 205, 94, 60 },
		{ 205, 95, 60 },
		{ 205, 96, 60 },
		{ 205, 97, 60 },
		{ 205, 98, 60 },
		{ 205, 99, 60 },
		{ 205, 100, 60 },
		{ 205, 101, 60 },
		{ 205, 102, 60 },
		{ 205, 103, 60 },
		{ 205, 104, 60 },
		{ 205, 105, 60 },
		{ 205, 106, 60 },
		{ 205, 107, 60 },
		{ 205, 108, 60 },
		{ 205, 109, 60 },
		{ 205, 110, 60 },
		{ 205, 111, 60 },
		{ 205, 112, 60 },
		{ 205, 113, 60 },
		{ 205, 114, 60 },
		{ 205, 115, 60 },
		{ 205, 116, 60 },
		{ 205, 117, 60 },
		{ 205, 118, 60 },
		{ 206, 1, 60 },
		{ 206, 2, 60 },
		{ 206, 3, 60 },
		{ 206, 4, 60 },
		{ 206, 5, 60 },
		{ 206, 6, 60 },
		{ 206, 7, 60 },
		{ 206, 8, 60 },
		{ 206, 9, 60 },
		{ 206, 10, 60 },
		{ 206, 11, 60 },
		{ 206, 12, 60 },
		{ 206, 13, 60 },
		{ 206, 14, 60 },
		{ 206, 15, 60 },
		{ 206, 16, 60 },
		{ 206, 17, 60 },
		{ 206, 18, 60 },
		{ 206, 19, 60 },
		{ 206, 20, 60 },
		{ 206, 21, 60 },
		{ 206, 22, 60 },
		{ 206, 23, 60 },
		{ 206, 24, 60 },
		{ 206, 25, 60 },
		{ 206, 26, 60 },
		{ 206, 27, 60 },
		{ 206, 28, 60 },
		{ 206, 29, 60 },
		{ 206, 30, 60 },
		{ 206, 31, 60 },
		{ 206, 32, 60 },
		{ 206, 33, 60 },
		{ 206, 34, 60 },
		{ 206, 35, 60 },
		{ 206, 36, 60 },
		{ 206, 37, 60 },
		{ 206, 38, 60 },
		{ 206, 39, 60 },
		{ 206, 40, 60 },
		{ 206, 41, 60 },
		{ 206, 42, 60 },
		{ 206, 43, 60 },
		{ 206, 44, 60 },
		{ 206, 45, 60 },
		{ 206, 46, 60 },
		{ 206, 47, 60 },
		{ 206, 48, 60 },
		{ 206, 49, 60 },
		{ 206, 50, 60 },
		{ 206, 51, 60 },
		{ 206, 52, 60 },
		{ 206, 53, 60 },
		{ 206, 54, 60 },
		{ 206, 55, 60 },
		{ 206, 56, 60 },
		{ 206, 57, 60 },
		{ 206, 58, 60 },
		{ 206, 59, 60 },
		{ 206, 60, 60 },
		{ 206, 61, 60 },
		{ 206, 62, 60 },
		{ 206, 63, 60 },
		{ 206, 64, 60 },
		{ 206, 65, 60 },
		{ 206, 66, 60 },
		{ 206, 67, 60 },
		{ 206, 68, 60 },
		{ 206, 69, 60 },
		{ 206, 70, 60 },
		{ 206, 71, 60 },
		{ 206, 72, 60 },
		{ 206, 73, 60 },
		{ 206, 74, 60 },
		{ 206, 75, 60 },
		{ 206, 76, 60 },
		{ 206, 77, 60 },
		{ 206, 78, 60 },
		{ 206, 79, 60 },
		{ 206, 80, 60 },
		{ 206, 81, 60 },
		{ 206, 82, 60 },
		{ 206, 83, 60 },
		{ 206, 84, 60 },
		{ 206, 85, 60 },
		{ 206, 86, 60 },
		{ 206, 87, 60 },
		{ 206, 88, 60 },
		{ 206, 89, 60 },
		{ 206, 90, 60 },
		{ 206, 91, 60 },
		{ 206, 92, 60 },
		{ 206, 93, 60 },
		{ 206, 94, 60 },
		{ 206, 95, 60 },
		{ 206, 96, 60 },
		{ 206, 97, 60 },
		{ 206, 98, 60 },
		{ 206, 99, 60 },
		{ 206, 100, 60 },
		{ 206, 101, 60 },
		{ 206, 102, 60 },
		{ 206, 103, 60 },
		{ 206, 104, 60 },
		{ 206, 105, 60 },
		{ 206, 106, 60 },
		{ 206, 107, 60 },
		{ 206, 108, 60 },
		{ 206, 109, 60 },
		{ 206, 110, 60 },
		{ 206, 111, 60 },
		{ 206, 112, 60 },
		{ 206, 113, 60 },
		{ 206, 114, 60 },
		{ 206, 115, 60 },
		{ 206, 116, 60 },
		{ 206, 117, 60 },
		{ 206, 118, 60 },
		{ 206, 119, 60 },
		{ 206, 120, 60 },
		{ 207, 1, 60 },
		{ 207, 2, 60 },
		{ 207, 3, 60 },
		{ 207, 4, 60 },
		{ 207, 5, 60 },
		{ 207, 6, 60 },
		{ 207, 7, 60 },
		{ 207, 8, 60 },
		{ 207, 9, 60 },
		{ 207, 10, 60 },
		{ 207, 11, 60 },
		{ 207, 12, 60 },
		{ 207, 13, 60 },
		{ 207, 14, 60 },
		{ 207, 15, 60 },
		{ 207, 16, 60 },
		{ 207, 17, 60 },
		{ 207, 18, 60 },
		{ 207, 19, 60 },
		{ 207, 20, 60 },
		{ 207, 21, 60 },
		{ 207, 22, 60 },
		{ 207, 23, 60 },
		{ 207, 24, 60 },
		{ 207, 25, 60 },
		{ 207, 26, 60 },
		{ 207, 27, 60 },
		{ 207, 28, 60 },
		{ 207, 29, 60 },
		{ 207, 30, 60 },
		{ 207, 31, 60 },
		{ 207, 32, 60 },
		{ 207, 33, 60 },
		{ 207, 34, 60 },
		{ 207, 35, 60 },
		{ 207, 36, 60 },
		{ 207, 37, 60 },
		{ 207, 38, 60 },
		{ 207, 39, 60 },
		{ 207, 40, 60 },
		{ 207, 41, 60 },
		{ 207, 42, 60 },
		{ 207, 43, 60 },
		{ 207, 44, 60 },
		{ 207, 45, 60 },
		{ 207, 46, 60 },
		{ 207, 47, 60 },
		{ 207, 48, 60 },
		{ 207, 49, 60 },
		{ 207, 50, 60 },
		{ 207, 51, 60 },
		{ 207, 52, 60 },
		{ 207, 53, 60 },
		{ 207, 54, 60 },
		{ 207, 55, 60 },
		{ 207, 56, 60 },
		{ 207, 57, 60 },
		{ 207, 58, 60 },
		{ 207, 59, 60 },
		{ 207, 60, 60 },
		{ 207, 61, 60 },
		{ 207, 62, 60 },
		{ 207, 63, 60 },
		{ 207, 64, 60 },
		{ 207, 65, 60 },
		{ 207, 66, 60 },
		{ 207, 67, 60 },
		{ 207, 68, 60 },
		{ 207, 69, 60 },
		{ 207, 70, 60 },
		{ 207, 71, 60 },
		{ 207, 72, 60 },
		{ 207, 73, 60 },
		{ 207, 74, 60 },
		{ 207, 75, 60 },
		{ 207, 76, 60 },
		{ 207, 77, 60 },
		{ 207, 78, 60 },
		{ 207, 79, 60 },
		{ 207, 80, 60 },
		{ 207, 81, 60 },
		{ 207, 82, 60 },
		{ 207, 83, 60 },
		{ 207, 84, 60 },
		{ 207, 85, 60 },
		{ 207, 86, 60 },
		{ 207, 87, 60 },
		{ 207, 88, 60 },
		{ 207, 89, 60 },
		{ 207, 90, 60 },
		{ 207, 91, 60 },
		{ 207, 92, 60 },
		{ 207, 93, 60 },
		{ 207, 94, 60 },
		{ 207, 95, 60 },
		{ 207, 96, 60 },
		{ 207, 97, 60 },
		{ 207, 98, 60 },
		{ 207, 99, 60 },
		{ 207, 100, 60 },
		{ 207, 101, 60 },
		{ 207, 102, 60 },
		{ 207, 103, 60 },
		{ 207, 104, 60 },
		{ 207, 105, 60 },
		{ 207, 106, 60 },
		{ 207, 107, 60 },
		{ 207, 108, 60 },
		{ 207, 109, 60 },
		{ 207, 110, 60 },
		{ 207, 111, 60 },
		{ 207, 112, 60 },
		{ 207, 113, 60 },
		{ 207, 114, 60 },
		{ 207, 115, 60 },
		{ 207, 116, 60 },
		{ 207, 117, 60 },
		{ 207, 118, 60 },
		{ 207, 119, 60 },
		{ 207, 120, 60 },
		{ 207, 121, 60 },
		{ 207, 122, 60 },
		{ 207, 123, 60 },
		{ 208, 1, 60 },
		{ 208, 2, 60 },
		{ 208, 3, 60 },
		{ 208, 4, 60 },
		{ 208, 5, 60 },
		{ 208, 6, 60 },
		{ 208, 7, 60 },
		{ 208, 8, 60 },
		{ 208, 9, 60 },
		{ 208, 10, 60 },
		{ 208, 11, 60 },
		{ 208, 12, 60 },
		{ 208, 13, 60 },
		{ 208, 14, 60 },
		{ 208, 15, 60 },
		{ 208, 16, 60 },
		{ 208, 17, 60 },
		{ 208, 18, 60 },
		{ 208, 19, 60 },
		{ 208, 20, 60 },
		{ 208, 21, 60 },
		{ 208, 22, 60 },
		{ 208, 23, 60 },
		{ 208, 24, 60 },
		{ 208, 25, 60 },
		{ 208, 26, 60 },
		{ 208, 27, 60 },
		{ 208, 28, 60 },
		{ 208, 29, 60 },
		{ 208, 30, 60 },
		{ 208, 31, 60 },
		{ 208, 32, 60 },
		{ 208, 33, 60 },
		{ 208, 34, 60 },
		{ 208, 35, 60 },
		{ 208, 36, 60 },
		{ 208, 37, 60 },
		{ 208, 38, 60 },
		{ 208, 39, 60 },
		{ 208, 40, 60 },
		{ 208, 41, 60 },
		{ 208, 42, 60 },
		{ 208, 43, 60 },
		{ 208, 44, 60 },
		{ 208, 45, 60 },
		{ 208, 46, 60 },
		{ 208, 47, 60 },
		{ 208, 48, 60 },
		{ 208, 49, 60 },
		{ 208, 50, 60 },
		{ 208, 51, 60 },
		{ 208, 52, 60 },
		{ 208, 53, 60 },
		{ 208, 54, 60 },
		{ 208, 55, 60 },
		{ 208, 56, 60 },
		{ 208, 57, 60 },
		{ 208, 58, 60 },
		{ 208, 59, 60 },
		{ 208, 60, 60 },
		{ 208, 61, 60 },
		{ 208, 62, 60 },
		{ 208, 63, 60 },
		{ 208, 64, 60 },
		{ 208, 65, 60 },
		{ 208, 66, 60 },
		{ 208, 67, 60 },
		{ 208, 68, 60 },
		{ 208, 69, 60 },
		{ 208, 70, 60 },
		{ 208, 71, 60 },
		{ 208, 72, 60 },
		{ 208, 73, 60 },
		{ 208, 74, 60 },
		{ 208, 75, 60 },
		{ 208, 76, 60 },
		{ 208, 77, 60 },
		{ 208, 78, 60 },
		{ 208, 79, 60 },
		{ 208, 80, 60 },
		{ 208, 81, 60 },
		{ 208, 82, 60 },
		{ 208, 83, 60 },
		{ 208, 84, 60 },
		{ 208, 85, 60 },
		{ 208, 86, 60 },
		{ 208, 87, 60 },
		{ 208, 88, 60 },
		{ 208, 89, 60 },
		{ 208, 90, 60 },
		{ 208, 91, 60 },
		{ 208, 92, 60 },
		{ 208, 93, 60 },
		{ 208, 94, 60 },
		{ 208, 95, 60 },
		{ 208, 96, 60 },
		{ 208, 97, 60 },
		{ 208, 98, 60 },
		{ 208, 99, 60 },
		{ 208, 100, 60 },
		{ 208, 101, 60 },
		{ 208, 102, 60 },
		{ 208, 103, 60 },
		{ 208, 104, 60 },
		{ 208, 105, 60 },
		{ 208, 106, 60 },
		{ 208, 107, 60 },
		{ 208, 108, 60 },
		{ 208, 109, 60 },
		{ 208, 110, 60 },
		{ 208, 111, 60 },
		{ 208, 112, 60 },
		{ 208, 113, 60 },
		{ 208, 114, 60 },
		{ 208, 115, 60 },
		{ 208, 116, 60 },
		{ 208, 117, 60 },
		{ 208, 118, 60 },
		{ 208, 119, 60 },
		{ 208, 120, 60 },
		{ 208, 121, 60 },
		{ 208, 122, 60 },
		{ 208, 123, 60 },
		{ 208, 124, 60 },
		{ 208, 125, 60 },
		{ 208, 126, 60 },
		{ 209, 1, 60 },
		{ 209, 2, 60 },
		{ 209, 3, 60 },
		{ 209, 4, 60 },
		{ 209, 5, 60 },
		{ 209, 6, 60 },
		{ 209, 7, 60 },
		{ 209, 8, 60 },
		{ 209, 9, 60 },
		{ 209, 10, 60 },
		{ 209, 11, 60 },
		{ 209, 12, 60 },
		{ 209, 13, 60 },
		{ 209, 14, 60 },
		{ 209, 15, 60 },
		{ 209, 16, 60 },
		{ 209, 17, 60 },
		{ 209, 18, 60 },
		{ 209, 19, 60 },
		{ 209, 20, 60 },
		{ 209, 21, 60 },
		{ 209, 22, 60 },
		{ 209, 23, 60 },
		{ 209, 24, 60 },
		{ 209, 25, 60 },
		{ 209, 26, 60 },
		{ 209, 27, 60 },
		{ 209, 28, 60 },
		{ 209, 29, 60 },
		{ 209, 30, 60 },
		{ 209, 31, 60 },
		{ 209, 32, 60 },
		{ 209, 33, 60 },
		{ 209, 34, 60 },
		{ 209, 35, 60 },
		{ 209, 36, 60 },
		{ 209, 37, 60 },
		{ 209, 38, 60 },
		{ 209, 39, 60 },
		{ 209, 40, 60 },
		{ 209, 41, 60 },
		{ 209, 42, 60 },
		{ 209, 43, 60 },
		{ 209, 44, 60 },
		{ 209, 45, 60 },
		{ 209, 46, 60 },
		{ 209, 47, 60 },
		{ 209, 48, 60 },
		{ 209, 49, 60 },
		{ 209, 50, 60 },
		{ 209, 51, 60 },
		{ 209, 52, 60 },
		{ 209, 53, 60 },
		{ 209, 54, 60 },
		{ 209, 55, 60 },
		{ 209, 56, 60 },
		{ 209, 57, 60 },
		{ 209, 58, 60 },
		{ 209, 59, 60 },
		{ 209, 60, 60 },
		{ 209, 61, 60 },
		{ 209, 62, 60 },
		{ 209, 63, 60 },
		{ 209, 64, 60 },
		{ 209, 65, 60 },
		{ 209, 66, 60 },
		{ 209, 67, 60 },
		{ 209, 68, 60 },
		{ 209, 69, 60 },
		{ 209, 70, 60 },
		{ 209, 71, 60 },
		{ 209, 72, 60 },
		{ 209, 73, 60 },
		{ 209, 74, 60 },
		{ 209, 75, 60 },
		{ 209, 76, 60 },
		{ 209, 77, 60 },
		{ 209, 78, 60 },
		{ 209, 79, 60 },
		{ 209, 80, 60 },
		{ 209, 81, 60 },
		{ 209, 82, 60 },
		{ 209, 83, 60 },
		{ 209, 84, 60 },
		{ 209, 85, 60 },
		{ 209, 86, 60 },
		{ 209, 87, 60 },
		{ 209, 88, 60 },
		{ 209, 89, 60 },
		{ 209, 90, 60 },
		{ 209, 91, 60 },
		{ 209, 92, 60 },
		{ 209, 93, 60 },
		{ 209, 94, 60 },
		{ 209, 95, 60 },
		{ 209, 96, 60 },
		{ 209, 97, 60 },
		{ 209, 98, 60 },
		{ 209, 99, 60 },
		{ 209, 100, 60 },
		{ 209, 101, 60 },
		{ 209, 102, 60 },
		{ 209, 103, 60 },
		{ 209, 104, 60 },
		{ 209, 105, 60 },
		{ 209, 106, 60 },
		{ 209, 107, 60 },
		{ 209, 108, 60 },
		{ 209, 109, 60 },
		{ 209, 110, 60 },
		{ 209, 111, 60 },
		{ 209, 112, 60 },
		{ 209, 113, 60 },
		{ 209, 114, 60 },
		{ 209, 115, 60 },
		{ 209, 116, 60 },
		{ 209, 117, 60 },
		{ 209, 118, 60 },
		{ 209, 119, 60 },
		{ 209, 120, 60 },
		{ 209, 121, 60 },
		{ 209, 122, 60 },
		{ 209, 123, 60 },
		{ 209, 124, 60 },
		{ 209, 125, 60 },
		{ 209, 126, 60 },
		{ 209, 127, 60 },
		{ 209, 128, 60 },
		{ 210, 1, 60 },
		{ 210, 2, 60 },
		{ 210, 3, 60 },
		{ 210, 4, 60 },
		{ 210, 5, 60 },
		{ 210, 6, 60 },
		{ 210, 7, 60 },
		{ 210, 8, 60 },
		{ 210, 9, 60 },
		{ 210, 10, 60 },
		{ 210, 11, 60 },
		{ 210, 12, 60 },
		{ 210, 13, 60 },
		{ 210, 14, 60 },
		{ 210, 15, 60 },
		{ 210, 16, 60 },
		{ 210, 17, 60 },
		{ 210, 18, 60 },
		{ 210, 19, 60 },
		{ 210, 20, 60 },
		{ 210, 21, 60 },
		{ 210, 22, 60 },
		{ 210, 23, 60 },
		{ 210, 24, 60 },
		{ 210, 25, 60 },
		{ 210, 26, 60 },
		{ 210, 27, 60 },
		{ 210, 28, 60 },
		{ 210, 29, 60 },
		{ 210, 30, 60 },
		{ 210, 31, 60 },
		{ 210, 32, 60 },
		{ 210, 33, 60 },
		{ 210, 34, 60 },
		{ 210, 35, 60 },
		{ 210, 36, 60 },
		{ 210, 37, 60 },
		{ 210, 38, 60 },
		{ 210, 39, 60 },
		{ 210, 40, 60 },
		{ 210, 41, 60 },
		{ 210, 42, 60 },
		{ 210, 43, 60 },
		{ 210, 44, 60 },
		{ 210, 45, 60 },
		{ 210, 46, 60 },
		{ 210, 47, 60 },
		{ 210, 48, 60 },
		{ 210, 49, 60 },
		{ 210, 50, 60 },
		{ 210, 51, 60 },
		{ 210, 52, 60 },
		{ 210, 53, 60 },
		{ 210, 54, 60 },
		{ 210, 55, 60 },
		{ 210, 56, 60 },
		{ 210, 57, 60 },
		{ 210, 58, 60 },
		{ 210, 59, 60 },
		{ 210, 60, 60 },
		{ 210, 61, 60 },
		{ 210, 62, 60 },
		{ 210, 63, 60 },
		{ 210, 64, 60 },
		{ 210, 65, 60 },
		{ 210, 66, 60 },
		{ 210, 67, 60 },
		{ 210, 68, 60 },
		{ 210, 69, 60 },
		{ 210, 70, 60 },
		{ 210, 71, 60 },
		{ 210, 72, 60 },
		{ 210, 73, 60 },
		{ 210, 74, 60 },
		{ 210, 75, 60 },
		{ 210, 76, 60 },
		{ 210, 77, 60 },
		{ 210, 78, 60 },
		{ 210, 79, 60 },
		{ 210, 80, 60 },
		{ 210, 81, 60 },
		{ 210, 82, 60 },
		{ 210, 83, 60 },
		{ 210, 84, 60 },
		{ 210, 85, 60 },
		{ 210, 86, 60 },
		{ 210, 87, 60 },
		{ 210, 88, 60 },
		{ 210, 89, 60 },
		{ 210, 90, 60 },
		{ 210, 91, 60 },
		{ 210, 92, 60 },
		{ 210, 93, 60 },
		{ 210, 94, 60 },
		{ 210, 95, 60 },
		{ 210, 96, 60 },
		{ 210, 97, 60 },
		{ 210, 98, 60 },
		{ 210, 99, 60 },
		{ 210, 100, 60 },
		{ 210, 101, 60 },
		{ 210, 102, 60 },
		{ 210, 103, 60 },
		{ 210, 104, 60 },
		{ 210, 105, 60 },
		{ 210, 106, 60 },
		{ 210, 107, 60 },
		{ 210, 108, 60 },
		{ 210, 109, 60 },
		{ 210, 110, 60 },
		{ 210, 111, 60 },
		{ 210, 112, 60 },
		{ 210, 113, 60 },
		{ 210, 114, 60 },
		{ 210, 115, 60 },
		{ 210, 116, 60 },
		{ 210, 117, 60 },
		{ 210, 118, 60 },
		{ 210, 119, 60 },
		{ 210, 120, 60 },
		{ 210, 121, 60 },
		{ 210, 122, 60 },
		{ 210, 123, 60 },
		{ 210, 124, 60 },
		{ 210, 125, 60 },
		{ 210, 126, 60 },
		{ 210, 127, 60 },
		{ 210, 128, 60 },
		{ 210, 129, 60 },
		{ 210, 130, 60 },
		{ 210, 131, 60 },
		{ 211, 1, 60 },
		{ 211, 2, 60 },
		{ 211, 3, 60 },
		{ 211, 4, 60 },
		{ 211, 5, 60 },
		{ 211, 6, 60 },
		{ 211, 7, 60 },
		{ 211, 8, 60 },
		{ 211, 9, 60 },
		{ 211, 10, 60 },
		{ 211, 11, 60 },
		{ 211, 12, 60 },
		{ 211, 13, 60 },
		{ 211, 14, 60 },
		{ 211, 15, 60 },
		{ 211, 16, 60 },
		{ 211, 17, 60 },
		{ 211, 18, 60 },
		{ 211, 19, 60 },
		{ 211, 20, 60 },
		{ 211, 21, 60 },
		{ 211, 22, 60 },
		{ 211, 23, 60 },
		{ 211, 24, 60 },
		{ 211, 25, 60 },
		{ 211, 26, 60 },
		{ 211, 27, 60 },
		{ 211, 28, 60 },
		{ 211, 29, 60 },
		{ 211, 30, 60 },
		{ 211, 31, 60 },
		{ 211, 32, 60 },
		{ 211, 33, 60 },
		{ 211, 34, 60 },
		{ 211, 35, 60 },
		{ 211, 36, 60 },
		{ 211, 37, 60 },
		{ 211, 38, 60 },
		{ 211, 39, 60 },
		{ 211, 40, 60 },
		{ 211, 41, 60 },
		{ 211, 42, 60 },
		{ 211, 43, 60 },
		{ 211, 44, 60 },
		{ 211, 45, 60 },
		{ 211, 46, 60 },
		{ 211, 47, 60 },
		{ 211, 48, 60 },
		{ 211, 49, 60 },
		{ 211, 50, 60 },
		{ 211, 51, 60 },
		{ 211, 52, 60 },
		{ 211, 53, 60 },
		{ 211, 54, 60 },
		{ 211, 55, 60 },
		{ 211, 56, 60 },
		{ 211, 57, 60 },
		{ 211, 58, 60 },
		{ 211, 59, 60 },
		{ 211, 60, 60 },
		{ 211, 61, 60 },
		{ 211, 62, 60 },
		{ 211, 63, 60 },
		{ 211, 64, 60 },
		{ 211, 65, 60 },
		{ 211, 66, 60 },
		{ 211, 67, 60 },
		{ 211, 68, 60 },
		{ 211, 69, 60 },
		{ 211, 70, 60 },
		{ 211, 71, 60 },
		{ 211, 72, 60 },
		{ 211, 73, 60 },
		{ 211, 74, 60 },
		{ 211, 75, 60 },
		{ 211, 76, 60 },
		{ 211, 77, 60 },
		{ 211, 78, 60 },
		{ 211, 79, 60 },
		{ 211, 80, 60 },
		{ 211, 81, 60 },
		{ 211, 82, 60 },
		{ 211, 83, 60 },
		{ 211, 84, 60 },
		{ 211, 85, 60 },
		{ 211, 86, 60 },
		{ 211, 87, 60 },
		{ 211, 88, 60 },
		{ 211, 89, 60 },
		{ 211, 90, 60 },
		{ 211, 91, 60 },
		{ 211, 92, 60 },
		{ 211, 93, 60 },
		{ 211, 94, 60 },
		{ 211, 95, 60 },
		{ 211, 96, 60 },
		{ 211, 97, 60 },
		{ 211, 98, 60 },
		{ 211, 99, 60 },
		{ 211, 100, 60 },
		{ 211, 101, 60 },
		{ 211, 102, 60 },
		{ 211, 103, 60 },
		{ 211, 104, 60 },
		{ 211, 105, 60 },
		{ 211, 106, 60 },
		{ 211, 107, 60 },
		{ 211, 108, 60 },
		{ 211, 109, 60 },
		{ 211, 110, 60 },
		{ 211, 111, 60 },
		{ 211, 112, 60 },
		{ 211, 113, 60 },
		{ 211, 114, 60 },
		{ 211, 115, 60 },
		{ 211, 116, 60 },
		{ 211, 117, 60 },
		{ 211, 118, 60 },
		{ 211, 119, 60 },
		{ 211, 120, 60 },
		{ 211, 121, 60 },
		{ 211, 122, 60 },
		{ 211, 123, 60 },
		{ 211, 124, 60 },
		{ 211, 125, 60 },
		{ 211, 126, 60 },
		{ 211, 127, 60 },
		{ 211, 128, 60 },
		{ 211, 129, 60 },
		{ 211, 130, 60 },
		{ 211, 131, 60 },
		{ 211, 132, 60 },
		{ 211, 133, 60 },
		{ 211, 134, 60 },
		{ 212, 1, 60 },
		{ 212, 2, 60 },
		{ 212, 3, 60 },
		{ 212, 4, 60 },
		{ 212, 5, 60 },
		{ 212, 6, 60 },
		{ 212, 7, 60 },
		{ 212, 8, 60 },
		{ 212, 9, 60 },
		{ 212, 10, 60 },
		{ 212, 11, 60 },
		{ 212, 12, 60 },
		{ 212, 13, 60 },
		{ 212, 14, 60 },
		{ 212, 15, 60 },
		{ 212, 16, 60 },
		{ 212, 17, 60 },
		{ 212, 18, 60 },
		{ 212, 19, 60 },
		{ 212, 20, 60 },
		{ 212, 21, 60 },
		{ 212, 22, 60 },
		{ 212, 23, 60 },
		{ 212, 24, 60 },
		{ 212, 25, 60 },
		{ 212, 26, 60 },
		{ 212, 27, 60 },
		{ 212, 28, 60 },
		{ 212, 29, 60 },
		{ 212, 30, 60 },
		{ 212, 31, 60 },
		{ 212, 32, 60 },
		{ 212, 33, 60 },
		{ 212, 34, 60 },
		{ 212, 35, 60 },
		{ 212, 36, 60 },
		{ 212, 37, 60 },
		{ 212, 38, 60 },
		{ 212, 39, 60 },
		{ 212, 40, 60 },
		{ 212, 41, 60 },
		{ 212, 42, 60 },
		{ 212, 43, 60 },
		{ 212, 44, 60 },
		{ 212, 45, 60 },
		{ 212, 46, 60 },
		{ 212, 47, 60 },
		{ 212, 48, 60 },
		{ 212, 49, 60 },
		{ 212, 50, 60 },
		{ 212, 51, 60 },
		{ 212, 52, 60 },
		{ 212, 53, 60 },
		{ 212, 54, 60 },
		{ 212, 55, 60 },
		{ 212, 56, 60 },
		{ 212, 57, 60 },
		{ 212, 58, 60 },
		{ 212, 59, 60 },
		{ 212, 60, 60 },
		{ 212, 61, 60 },
		{ 212, 62, 60 },
		{ 212, 63, 60 },
		{ 212, 64, 60 },
		{ 212, 65, 60 },
		{ 212, 66, 60 },
		{ 212, 67, 60 },
		{ 212, 68, 60 },
		{ 212, 69, 60 },
		{ 212, 70, 60 },
		{ 212, 71, 60 },
		{ 212, 72, 60 },
		{ 212, 73, 60 },
		{ 212, 74, 60 },
		{ 212, 75, 60 },
		{ 212, 76, 60 },
		{ 212, 77, 60 },
		{ 212, 78, 60 },
		{ 212, 79, 60 },
		{ 212, 80, 60 },
		{ 212, 81, 60 },
		{ 212, 82, 60 },
		{ 212, 83, 60 },
		{ 212, 84, 60 },
		{ 212, 85, 60 },
		{ 212, 86, 60 },
		{ 212, 87, 60 },
		{ 212, 88, 60 },
		{ 212, 89, 60 },
		{ 212, 90, 60 },
		{ 212, 91, 60 },
		{ 212, 92, 60 },
		{ 212, 93, 60 },
		{ 212, 94, 60 },
		{ 212, 95, 60 },
		{ 212, 96, 60 },
		{ 212, 97, 60 },
		{ 212, 98, 60 },
		{ 212, 99, 60 },
		{ 212, 100, 60 },
		{ 212, 101, 60 },
		{ 212, 102, 60 },
		{ 212, 103, 60 },
		{ 212, 104, 60 },
		{ 212, 105, 60 },
		{ 212, 106, 60 },
		{ 212, 107, 60 },
		{ 212, 108, 60 },
		{ 212, 109, 60 },
		{ 212, 110, 60 },
		{ 212, 111, 60 },
		{ 212, 112, 60 },
		{ 212, 113, 60 },
		{ 212, 114, 60 },
		{ 212, 115, 60 },
		{ 212, 116, 60 },
		{ 212, 117, 60 },
		{ 212, 118, 60 },
		{ 212, 119, 60 },
		{ 212, 120, 60 },
		{ 212, 121, 60 },
		{ 212, 122, 60 },
		{ 212, 123, 60 },
		{ 212, 124, 60 },
		{ 212, 125, 60 },
		{ 212, 126, 60 },
		{ 212, 127, 60 },
		{ 212, 128, 60 },
		{ 212, 129, 60 },
		{ 212, 130, 60 },
		{ 212, 131, 60 },
		{ 212, 132, 60 },
		{ 212, 133, 60 },
		{ 212, 134, 60 },
		{ 212, 135, 60 },
		{ 212, 136, 60 },
		{ 213, 1, 60 },
		{ 213, 2, 60 },
		{ 213, 3, 60 },
		{ 213, 4, 60 },
		{ 213, 5, 60 },
		{ 213, 6, 60 },
		{ 213, 7, 60 },
		{ 213, 8, 60 },
		{ 213, 9, 60 },
		{ 213, 10, 60 },
		{ 213, 11, 60 },
		{ 213, 12, 60 },
		{ 213, 13, 60 },
		{ 213, 14, 60 },
		{ 213, 15, 60 },
		{ 213, 16, 60 },
		{ 213, 17, 60 },
		{ 213, 18, 60 },
		{ 213, 19, 60 },
		{ 213, 20, 60 },
		{ 213, 21, 60 },
		{ 213, 22, 60 },
		{ 213, 23, 60 },
		{ 213, 24, 60 },
		{ 213, 25, 60 },
		{ 213, 26, 60 },
		{ 213, 27, 60 },
		{ 213, 28, 60 },
		{ 213, 29, 60 },
		{ 213, 30, 60 },
		{ 213, 31, 60 },
		{ 213, 32, 60 },
		{ 213, 33, 60 },
		{ 213, 34, 60 },
		{ 213, 35, 60 },
		{ 213, 36, 60 },
		{ 213, 37, 60 },
		{ 213, 38, 60 },
		{ 213, 39, 60 },
		{ 213, 40, 60 },
		{ 213, 41, 60 },
		{ 213, 42, 60 },
		{ 213, 43, 60 },
		{ 213, 44, 60 },
		{ 213, 45, 60 },
		{ 213, 46, 60 },
		{ 213, 47, 60 },
		{ 213, 48, 60 },
		{ 213, 49, 60 },
		{ 213, 50, 60 },
		{ 213, 51, 60 },
		{ 213, 52, 60 },
		{ 213, 53, 60 },
		{ 213, 54, 60 },
		{ 213, 55, 60 },
		{ 213, 56, 60 },
		{ 213, 57, 60 },
		{ 213, 58, 60 },
		{ 213, 59, 60 },
		{ 213, 60, 60 },
		{ 213, 61, 60 },
		{ 213, 62, 60 },
		{ 213, 63, 60 },
		{ 213, 64, 60 },
		{ 213, 65, 60 },
		{ 213, 66, 60 },
		{ 213, 67, 60 },
		{ 213, 68, 60 },
		{ 213, 69, 60 },
		{ 213, 70, 60 },
		{ 213, 71, 60 },
		{ 213, 72, 60 },
		{ 213, 73, 60 },
		{ 213, 74, 60 },
		{ 213, 75, 60 },
		{ 213, 76, 60 },
		{ 213, 77, 60 },
		{ 213, 78, 60 },
		{ 213, 79, 60 },
		{ 213, 80, 60 },
		{ 213, 81, 60 },
		{ 213, 82, 60 },
		{ 213, 83, 60 },
		{ 213, 84, 60 },
		{ 213, 85, 60 },
		{ 213, 86, 60 },
		{ 213, 87, 60 },
		{ 213, 88, 60 },
		{ 213, 89, 60 },
		{ 213, 90, 60 },
		{ 213, 91, 60 },
		{ 213, 92, 60 },
		{ 213, 93, 60 },
		{ 213, 94, 60 },
		{ 213, 95, 60 },
		{ 213, 96, 60 },
		{ 213, 97, 60 },
		{ 213, 98, 60 },
		{ 213, 99, 60 },
		{ 213, 100, 60 },
		{ 213, 101, 60 },
		{ 213, 102, 60 },
		{ 213, 103, 60 },
		{ 213, 104, 60 },
		{ 213, 105, 60 },
		{ 213, 106, 60 },
		{ 213, 107, 60 },
		{ 213, 108, 60 },
		{ 213, 109, 60 },
		{ 213, 110, 60 },
		{ 213, 111, 60 },
		{ 213, 112, 60 },
		{ 213, 113, 60 },
		{ 213, 114, 60 },
		{ 213, 115, 60 },
		{ 213, 116, 60 },
		{ 213, 117, 60 },
		{ 213, 118, 60 },
		{ 213, 119, 60 },
		{ 213, 120, 60 },
		{ 213, 121, 60 },
		{ 213, 122, 60 },
		{ 213, 123, 60 },
		{ 213, 124, 60 },
		{ 213, 125, 60 },
		{ 213, 126, 60 },
		{ 213, 127, 60 },
		{ 213, 128, 60 },
		{ 213, 129, 60 },
		{ 213, 130, 60 },
		{ 213, 131, 60 },
		{ 213, 132, 60 },
		{ 213, 133, 60 },
		{ 213, 134, 60 },
		{ 213, 135, 60 },
		{ 213, 136, 60 },
		{ 213, 137, 60 },
		{ 213, 138, 60 },
		{ 213, 139, 60 },
		{ 214, 1, 60 },
		{ 214, 2, 60 },
		{ 214, 3, 60 },
		{ 214, 4, 60 },
		{ 214, 5, 60 },
		{ 214, 6, 60 },
		{ 214, 7, 60 },
		{ 214, 8, 60 },
		{ 214, 9, 60 },
		{ 214, 10, 60 },
		{ 214, 11, 60 },
		{ 214, 12, 60 },
		{ 214, 13, 60 },
		{ 214, 14, 60 },
		{ 214, 15, 60 },
		{ 214, 16, 60 },
		{ 214, 17, 60 },
		{ 214, 18, 60 },
		{ 214, 19, 60 },
		{ 214, 20, 60 },
		{ 214, 21, 60 },
		{ 214, 22, 60 },
		{ 214, 23, 60 },
		{ 214, 24, 60 },
		{ 214, 25, 60 },
		{ 214, 26, 60 },
		{ 214, 27, 60 },
		{ 214, 28, 60 },
		{ 214, 29, 60 },
		{ 214, 30, 60 },
		{ 214, 31, 60 },
		{ 214, 32, 60 },
		{ 214, 33, 60 },
		{ 214, 34, 60 },
		{ 214, 35, 60 },
		{ 214, 36, 60 },
		{ 214, 37, 60 },
		{ 214, 38, 60 },
		{ 214, 39, 60 },
		{ 214, 40, 60 },
		{ 214, 41, 60 },
		{ 214, 42, 60 },
		{ 214, 43, 60 },
		{ 214, 44, 60 },
		{ 214, 45, 60 },
		{ 214, 46, 60 },
		{ 214, 47, 60 },
		{ 214, 48, 60 },
		{ 214, 49, 60 },
		{ 214, 50, 60 },
		{ 214, 51, 60 },
		{ 214, 52, 60 },
		{ 214, 53, 60 },
		{ 214, 54, 60 },
		{ 214, 55, 60 },
		{ 214, 56, 60 },
		{ 214, 57, 60 },
		{ 214, 58, 60 },
		{ 214, 59, 60 },
		{ 214, 60, 60 },
		{ 214, 61, 60 },
		{ 214, 62, 60 },
		{ 214, 63, 60 },
		{ 214, 64, 60 },
		{ 214, 65, 60 },
		{ 214, 66, 60 },
		{ 214, 67, 60 },
		{ 214, 68, 60 },
		{ 214, 69, 60 },
		{ 214, 70, 60 },
		{ 214, 71, 60 },
		{ 214, 72, 60 },
		{ 214, 73, 60 },
		{ 214, 74, 60 },
		{ 214, 75, 60 },
		{ 214, 76, 60 },
		{ 214, 77, 60 },
		{ 214, 78, 60 },
		{ 214, 79, 60 },
		{ 214, 80, 60 },
		{ 214, 81, 60 },
		{ 214, 82, 60 },
		{ 214, 83, 60 },
		{ 214, 84, 60 },
		{ 214, 85, 60 },
		{ 214, 86, 60 },
		{ 214, 87, 60 },
		{ 214, 88, 60 },
		{ 214, 89, 60 },
		{ 214, 90, 60 },
		{ 214, 91, 60 },
		{ 214, 92, 60 },
		{ 214, 93, 60 },
		{ 214, 94, 60 },
		{ 214, 95, 60 },
		{ 214, 96, 60 },
		{ 214, 97, 60 },
		{ 214, 98, 60 },
		{ 214, 99, 60 },
		{ 214, 100, 60 },
		{ 214, 101, 60 },
		{ 214, 102, 60 },
		{ 214, 103, 60 },
		{ 214, 104, 60 },
		{ 214, 105, 60 },
		{ 214, 106, 60 },
		{ 214, 107, 60 },
		{ 214, 108, 60 },
		{ 214, 109, 60 },
		{ 214, 110, 60 },
		{ 214, 111, 60 },
		{ 214, 112, 60 },
		{ 214, 113, 60 },
		{ 214, 114, 60 },
		{ 214, 115, 60 },
		{ 214, 116, 60 },
		{ 214, 117, 60 },
		{ 214, 118, 60 },
		{ 214, 119, 60 },
		{ 214, 120, 60 },
		{ 214, 121, 60 },
		{ 214, 122, 60 },
		{ 214, 123, 60 },
		{ 214, 124, 60 },
		{ 214, 125, 60 },
		{ 214, 126, 60 },
		{ 214, 127, 60 },
		{ 214, 128, 60 },
		{ 214, 129, 60 },
		{ 214, 130, 60 },
		{ 214, 131, 60 },
		{ 214, 132, 60 },
		{ 214, 133, 60 },
		{ 214, 134, 60 },
		{ 214, 135, 60 },
		{ 214, 136, 60 },
		{ 214, 137, 60 },
		{ 214, 138, 60 },
		{ 214, 139, 60 },
		{ 214, 140, 60 },
		{ 214, 141, 60 },
		{ 214, 142, 60 },
		{ 215, 1, 60 },
		{ 215, 2, 60 },
		{ 215, 3, 60 },
		{ 215, 4, 60 },
		{ 215, 5, 60 },
		{ 215, 6, 60 },
		{ 215, 7, 60 },
		{ 215, 8, 60 },
		{ 215, 9, 60 },
		{ 215, 10, 60 },
		{ 215, 11, 60 },
		{ 215, 12, 60 },
		{ 215, 13, 60 },
		{ 215, 14, 60 },
		{ 215, 15, 60 },
		{ 215, 16, 60 },
		{ 215, 17, 60 },
		{ 215, 18, 60 },
		{ 215, 19, 60 },
		{ 215, 20, 60 },
		{ 215, 21, 60 },
		{ 215, 22, 60 },
		{ 215, 23, 60 },
		{ 215, 24, 60 },
		{ 215, 25, 60 },
		{ 215, 26, 60 },
		{ 215, 27, 60 },
		{ 215, 28, 60 },
		{ 215, 29, 60 },
		{ 215, 30, 60 },
		{ 215, 31, 60 },
		{ 215, 32, 60 },
		{ 215, 33, 60 },
		{ 215, 34, 60 },
		{ 215, 35, 60 },
		{ 215, 36, 60 },
		{ 215, 37, 60 },
		{ 215, 38, 60 },
		{ 215, 39, 60 },
		{ 215, 40, 60 },
		{ 215, 41, 60 },
		{ 215, 42, 60 },
		{ 215, 43, 60 },
		{ 215, 44, 60 },
		{ 215, 45, 60 },
		{ 215, 46, 60 },
		{ 215, 47, 60 },
		{ 215, 48, 60 },
		{ 215, 49, 60 },
		{ 215, 50, 60 },
		{ 215, 51, 60 },
		{ 215, 52, 60 },
		{ 215, 53, 60 },
		{ 215, 54, 60 },
		{ 215, 55, 60 },
		{ 215, 56, 60 },
		{ 215, 57, 60 },
		{ 215, 58, 60 },
		{ 215, 59, 60 },
		{ 215, 60, 60 },
		{ 215, 61, 60 },
		{ 215, 62, 60 },
		{ 215, 63, 60 },
		{ 215, 64, 60 },
		{ 215, 65, 60 },
		{ 215, 66, 60 },
		{ 215, 67, 60 },
		{ 215, 68, 60 },
		{ 215, 69, 60 },
		{ 215, 70, 60 },
		{ 215, 71, 60 },
		{ 215, 72, 60 },
		{ 215, 73, 60 },
		{ 215, 74, 60 },
		{ 215, 75, 60 },
		{ 215, 76, 60 },
		{ 215, 77, 60 },
		{ 215, 78, 60 },
		{ 215, 79, 60 },
		{ 215, 80, 60 },
		{ 215, 81, 60 },
		{ 215, 82, 60 },
		{ 215, 83, 60 },
		{ 215, 84, 60 },
		{ 215, 85, 60 },
		{ 215, 86, 60 },
		{ 215, 87, 60 },
		{ 215, 88, 60 },
		{ 215, 89, 60 },
		{ 215, 90, 60 },
		{ 215, 91, 60 },
		{ 215, 92, 60 },
		{ 215, 93, 60 },
		{ 215, 94, 60 },
		{ 215, 95, 60 },
		{ 215, 96, 60 },
		{ 215, 97, 60 },
		{ 215, 98, 60 },
		{ 215, 99, 60 },
		{ 215, 100, 60 },
		{ 215, 101, 60 },
		{ 215, 102, 60 },
		{ 215, 103, 60 },
		{ 215, 104, 60 },
		{ 215, 105, 60 },
		{ 215, 106, 60 },
		{ 215, 107, 60 },
		{ 215, 108, 60 },
		{ 215, 109, 60 },
		{ 215, 110, 60 },
		{ 215, 111, 60 },
		{ 215, 112, 60 },
		{ 215, 113, 60 },
		{ 215, 114, 60 },
		{ 215, 115, 60 },
		{ 215, 116, 60 },
		{ 215, 117, 60 },
		{ 215, 118, 60 },
		{ 215, 119, 60 },
		{ 215, 120, 60 },
		{ 215, 121, 60 },
		{ 215, 122, 60 },
		{ 215, 123, 60 },
		{ 215, 124, 60 },
		{ 215, 125, 60 },
		{ 215, 126, 60 },
		{ 215, 127, 60 },
		{ 215, 128, 60 },
		{ 215, 129, 60 },
		{ 215, 130, 60 },
		{ 215, 131, 60 },
		{ 215, 132, 60 },
		{ 215, 133, 60 },
		{ 215, 134, 60 },
		{ 215, 135, 60 },
		{ 215, 136, 60 },
		{ 215, 137, 60 },
		{ 215, 138, 60 },
		{ 215, 139, 60 },
		{ 215, 140, 60 },
		{ 215, 141, 60 },
		{ 215, 142, 60 },
		{ 215, 143, 60 },
		{ 215, 144, 60 },
		{ 216, 1, 60 },
		{ 216, 2, 60 },
		{ 216, 3, 60 },
		{ 216, 4, 60 },
		{ 216, 5, 60 },
		{ 216, 6, 60 },
		{ 216, 7, 60 },
		{ 216, 8, 60 },
		{ 216, 9, 60 },
		{ 216, 10, 60 },
		{ 216, 11, 60 },
		{ 216, 12, 60 },
		{ 216, 13, 60 },
		{ 216, 14, 60 },
		{ 216, 15, 60 },
		{ 216, 16, 60 },
		{ 216, 17, 60 },
		{ 216, 18, 60 },
		{ 216, 19, 60 },
		{ 216, 20, 60 },
		{ 216, 21, 60 },
		{ 216, 22, 60 },
		{ 216, 23, 60 },
		{ 216, 24, 60 },
		{ 216, 25, 60 },
		{ 216, 26, 60 },
		{ 216, 27, 60 },
		{ 216, 28, 60 },
		{ 216, 29, 60 },
		{ 216, 30, 60 },
		{ 216, 31, 60 },
		{ 216, 32, 60 },
		{ 216, 33, 60 },
		{ 216, 34, 60 },
		{ 216, 35, 60 },
		{ 216, 36, 60 },
		{ 216, 37, 60 },
		{ 216, 38, 60 },
		{ 216, 39, 60 },
		{ 216, 40, 60 },
		{ 216, 41, 60 },
		{ 216, 42, 60 },
		{ 216, 43, 60 },
		{ 216, 44, 60 },
		{ 216, 45, 60 },
		{ 216, 46, 60 },
		{ 216, 47, 60 },
		{ 216, 48, 60 },
		{ 216, 49, 60 },
		{ 216, 50, 60 },
		{ 216, 51, 60 },
		{ 216, 52, 60 },
		{ 216, 53, 60 },
		{ 216, 54, 60 },
		{ 216, 55, 60 },
		{ 216, 56, 60 },
		{ 216, 57, 60 },
		{ 216, 58, 60 },
		{ 216, 59, 60 },
		{ 216, 60, 60 },
		{ 216, 61, 60 },
		{ 216, 62, 60 },
		{ 216, 63, 60 },
		{ 216, 64, 60 },
		{ 216, 65, 60 },
		{ 216, 66, 60 },
		{ 216, 67, 60 },
		{ 216, 68, 60 },
		{ 216, 69, 60 },
		{ 216, 70, 60 },
		{ 216, 71, 60 },
		{ 216, 72, 60 },
		{ 216, 73, 60 },
		{ 216, 74, 60 },
		{ 216, 75, 60 },
		{ 216, 76, 60 },
		{ 216, 77, 60 },
		{ 216, 78, 60 },
		{ 216, 79, 60 },
		{ 216, 80, 60 },
		{ 216, 81, 60 },
		{ 216, 82, 60 },
		{ 216, 83, 60 },
		{ 216, 84, 60 },
		{ 216, 85, 60 },
		{ 216, 86, 60 },
		{ 216, 87, 60 },
		{ 216, 88, 60 },
		{ 216, 89, 60 },
		{ 216, 90, 60 },
		{ 216, 91, 60 },
		{ 216, 92, 60 },
		{ 216, 93, 60 },
		{ 216, 94, 60 },
		{ 216, 95, 60 },
		{ 216, 96, 60 },
		{ 216, 97, 60 },
		{ 216, 98, 60 },
		{ 216, 99, 60 },
		{ 216, 100, 60 },
		{ 216, 101, 60 },
		{ 216, 102, 60 },
		{ 216, 103, 60 },
		{ 216, 104, 60 },
		{ 216, 105, 60 },
		{ 216, 106, 60 },
		{ 216, 107, 60 },
		{ 216, 108, 60 },
		{ 216, 109, 60 },
		{ 216, 110, 60 },
		{ 216, 111, 60 },
		{ 216, 112, 60 },
		{ 216, 113, 60 },
		{ 216, 114, 60 },
		{ 216, 115, 60 },
		{ 216, 116, 60 },
		{ 216, 117, 60 },
		{ 216, 118, 60 },
		{ 216, 119, 60 },
		{ 216, 120, 60 },
		{ 216, 121, 60 },
		{ 216, 122, 60 },
		{ 216, 123, 60 },
		{ 216, 124, 60 },
		{ 216, 125, 60 },
		{ 216, 126, 60 },
		{ 216, 127, 60 },
		{ 216, 128, 60 },
		{ 216, 129, 60 },
		{ 216, 130, 60 },
		{ 216, 131, 60 },
		{ 216, 132, 60 },
		{ 216, 133, 60 },
		{ 216, 134, 60 },
		{ 216, 135, 60 },
		{ 216, 136, 60 },
		{ 216, 137, 60 },
		{ 216, 138, 60 },
		{ 216, 139, 60 },
		{ 216, 140, 60 },
		{ 216, 141, 60 },
		{ 216, 142, 60 },
		{ 216, 143, 60 },
		{ 216, 144, 60 },
		{ 216, 145, 60 },
		{ 216, 146, 60 },
		{ 216, 147, 60 },
		{ 217, 1, 60 },
		{ 217, 2, 60 },
		{ 217, 3, 60 },
		{ 217, 4, 60 },
		{ 217, 5, 60 },
		{ 217, 6, 60 },
		{ 217, 7, 60 },
		{ 217, 8, 60 },
		{ 217, 9, 60 },
		{ 217, 10, 60 },
		{ 217, 11, 60 },
		{ 217, 12, 60 },
		{ 217, 13, 60 },
		{ 217, 14, 60 },
		{ 217, 15, 60 },
		{ 217, 16, 60 },
		{ 217, 17, 60 },
		{ 217, 18, 60 },
		{ 217, 19, 60 },
		{ 217, 20, 60 },
		{ 217, 21, 60 },
		{ 217, 22, 60 },
		{ 217, 23, 60 },
		{ 217, 24, 60 },
		{ 217, 25, 60 },
		{ 217, 26, 60 },
		{ 217, 27, 60 },
		{ 217, 28, 60 },
		{ 217, 29, 60 },
		{ 217, 30, 60 },
		{ 217, 31, 60 },
		{ 217, 32, 60 },
		{ 217, 33, 60 },
		{ 217, 34, 60 },
		{ 217, 35, 60 },
		{ 217, 36, 60 },
		{ 217, 37, 60 },
		{ 217, 38, 60 },
		{ 217, 39, 60 },
		{ 217, 40, 60 },
		{ 217, 41, 60 },
		{ 217, 42, 60 },
		{ 217, 43, 60 },
		{ 217, 44, 60 },
		{ 217, 45, 60 },
		{ 217, 46, 60 },
		{ 217, 47, 60 },
		{ 217, 48, 60 },
		{ 217, 49, 60 },
		{ 217, 50, 60 },
		{ 217, 51, 60 },
		{ 217, 52, 60 },
		{ 217, 53, 60 },
		{ 217, 54, 60 },
		{ 217, 55, 60 },
		{ 217, 56, 60 },
		{ 217, 57, 60 },
		{ 217, 58, 60 },
		{ 217, 59, 60 },
		{ 217, 60, 60 },
		{ 217, 61, 60 },
		{ 217, 62, 60 },
		{ 217, 63, 60 },
		{ 217, 64, 60 },
		{ 217, 65, 60 },
		{ 217, 66, 60 },
		{ 217, 67, 60 },
		{ 217, 68, 60 },
		{ 217, 69, 60 },
		{ 217, 70, 60 },
		{ 217, 71, 60 },
		{ 217, 72, 60 },
		{ 217, 73, 60 },
		{ 217, 74, 60 },
		{ 217, 75, 60 },
		{ 217, 76, 60 },
		{ 217, 77, 60 },
		{ 217, 78, 60 },
		{ 217, 79, 60 },
		{ 217, 80, 60 },
		{ 217, 81, 60 },
		{ 217, 82, 60 },
		{ 217, 83, 60 },
		{ 217, 84, 60 },
		{ 217, 85, 60 },
		{ 217, 86, 60 },
		{ 217, 87, 60 },
		{ 217, 88, 60 },
		{ 217, 89, 60 },
		{ 217, 90, 60 },
		{ 217, 91, 60 },
		{ 217, 92, 60 },
		{ 217, 93, 60 },
		{ 217, 94, 60 },
		{ 217, 95, 60 },
		{ 217, 96, 60 },
		{ 217, 97, 60 },
		{ 217, 98, 60 },
		{ 217, 99, 60 },
		{ 217, 100, 60 },
		{ 217, 101, 60 },
		{ 217, 102, 60 },
		{ 217, 103, 60 },
		{ 217, 104, 60 },
		{ 217, 105, 60 },
		{ 217, 106, 60 },
		{ 217, 107, 60 },
		{ 217, 108, 60 },
		{ 217, 109, 60 },
		{ 217, 110, 60 },
		{ 217, 111, 60 },
		{ 217, 112, 60 },
		{ 217, 113, 60 },
		{ 217, 114, 60 },
		{ 217, 115, 60 },
		{ 217, 116, 60 },
		{ 217, 117, 60 },
		{ 217, 118, 60 },
		{ 217, 119, 60 },
		{ 217, 120, 60 },
		{ 217, 121, 60 },
		{ 217, 122, 60 },
		{ 217, 123, 60 },
		{ 217, 124, 60 },
		{ 217, 125, 60 },
		{ 217, 126, 60 },
		{ 217, 127, 60 },
		{ 217, 128, 60 },
		{ 217, 129, 60 },
		{ 217, 130, 60 },
		{ 217, 131, 60 },
		{ 217, 132, 60 },
		{ 217, 133, 60 },
		{ 217, 134, 60 },
		{ 217, 135, 60 },
		{ 217, 136, 60 },
		{ 217, 137, 60 },
		{ 217, 138, 60 },
		{ 217, 139, 60 },
		{ 217, 140, 60 },
		{ 217, 141, 60 },
		{ 217, 142, 60 },
		{ 217, 143, 60 },
		{ 217, 144, 60 },
		{ 217, 145, 60 },
		{ 217, 146, 60 },
		{ 217, 147, 60 },
		{ 217, 148, 60 },
		{ 217, 149, 60 },
		{ 217, 150, 60 },
		{ 218, 1, 60 },
		{ 218, 2, 60 },
		{ 218, 3, 60 },
		{ 218, 4, 60 },
		{ 218, 5, 60 },
		{ 218, 6, 60 },
		{ 218, 7, 60 },
		{ 218, 8, 60 },
		{ 218, 9, 60 },
		{ 218, 10, 60 },
		{ 218, 11, 60 },
		{ 218, 12, 60 },
		{ 218, 13, 60 },
		{ 218, 14, 60 },
		{ 218, 15, 60 },
		{ 218, 16, 60 },
		{ 218, 17, 60 },
		{ 218, 18, 60 },
		{ 218, 19, 60 },
		{ 218, 20, 60 },
		{ 218, 21, 60 },
		{ 218, 22, 60 },
		{ 218, 23, 60 },
		{ 218, 24, 60 },
		{ 218, 25, 60 },
		{ 218, 26, 60 },
		{ 218, 27, 60 },
		{ 218, 28, 60 },
		{ 218, 29, 60 },
		{ 218, 30, 60 },
		{ 218, 31, 60 },
		{ 218, 32, 60 },
		{ 218, 33, 60 },
		{ 218, 34, 60 },
		{ 218, 35, 60 },
		{ 218, 36, 60 },
		{ 218, 37, 60 },
		{ 218, 38, 60 },
		{ 218, 39, 60 },
		{ 218, 40, 60 },
		{ 218, 41, 60 },
		{ 218, 42, 60 },
		{ 218, 43, 60 },
		{ 218, 44, 60 },
		{ 218, 45, 60 },
		{ 218, 46, 60 },
		{ 218, 47, 60 },
		{ 218, 48, 60 },
		{ 218, 49, 60 },
		{ 218, 50, 60 },
		{ 218, 51, 60 },
		{ 218, 52, 60 },
		{ 218, 53, 60 },
		{ 218, 54, 60 },
		{ 218, 55, 60 },
		{ 218, 56, 60 },
		{ 218, 57, 60 },
		{ 218, 58, 60 },
		{ 218, 59, 60 },
		{ 218, 60, 60 },
		{ 218, 61, 60 },
		{ 218, 62, 60 },
		{ 218, 63, 60 },
		{ 218, 64, 60 },
		{ 218, 65, 60 },
		{ 218, 66, 60 },
		{ 218, 67, 60 },
		{ 218, 68, 60 },
		{ 218, 69, 60 },
		{ 218, 70, 60 },
		{ 218, 71, 60 },
		{ 218, 72, 60 },
		{ 218, 73, 60 },
		{ 218, 74, 60 },
		{ 218, 75, 60 },
		{ 218, 76, 60 },
		{ 218, 77, 60 },
		{ 218, 78, 60 },
		{ 218, 79, 60 },
		{ 218, 80, 60 },
		{ 218, 81, 60 },
		{ 218, 82, 60 },
		{ 218, 83, 60 },
		{ 218, 84, 60 },
		{ 218, 85, 60 },
		{ 218, 86, 60 },
		{ 218, 87, 60 },
		{ 218, 88, 60 },
		{ 218, 89, 60 },
		{ 218, 90, 60 },
		{ 218, 91, 60 },
		{ 218, 92, 60 },
		{ 218, 93, 60 },
		{ 218, 94, 60 },
		{ 218, 95, 60 },
		{ 218, 96, 60 },
		{ 218, 97, 60 },
		{ 218, 98, 60 },
		{ 218, 99, 60 },
		{ 218, 100, 60 },
		{ 218, 101, 60 },
		{ 218, 102, 60 },
		{ 218, 103, 60 },
		{ 218, 104, 60 },
		{ 218, 105, 60 },
		{ 218, 106, 60 },
		{ 218, 107, 60 },
		{ 218, 108, 60 },
		{ 218, 109, 60 },
		{ 218, 110, 60 },
		{ 218, 111, 60 },
		{ 218, 112, 60 },
		{ 218, 113, 60 },
		{ 218, 114, 60 },
		{ 218, 115, 60 },
		{ 218, 116, 60 },
		{ 218, 117, 60 },
		{ 218, 118, 60 },
		{ 218, 119, 60 },
		{ 218, 120, 60 },
		{ 218, 121, 60 },
		{ 218, 122, 60 },
		{ 218, 123, 60 },
		{ 218, 124, 60 },
		{ 218, 125, 60 },
		{ 218, 126, 60 },
		{ 218, 127, 60 },
		{ 218, 128, 60 },
		{ 218, 129, 60 },
		{ 218, 130, 60 },
		{ 218, 131, 60 },
		{ 218, 132, 60 },
		{ 218, 133, 60 },
		{ 218, 134, 60 },
		{ 218, 135, 60 },
		{ 218, 136, 60 },
		{ 218, 137, 60 },
		{ 218, 138, 60 },
		{ 218, 139, 60 },
		{ 218, 140, 60 },
		{ 218, 141, 60 },
		{ 218, 142, 60 },
		{ 218, 143, 60 },
		{ 218, 144, 60 },
		{ 218, 145, 60 },
		{ 218, 146, 60 },
		{ 218, 147, 60 },
		{ 218, 148, 60 },
		{ 218, 149, 60 },
		{ 218, 150, 60 },
		{ 218, 151, 60 },
		{ 218, 152, 60 },
		{ 219, 1, 60 },
		{ 219, 2, 60 },
		{ 219, 3, 60 },
		{ 219, 4, 60 },
		{ 219, 5, 60 },
		{ 219, 6, 60 },
		{ 219, 7, 60 },
		{ 219, 8, 60 },
		{ 219, 9, 60 },
		{ 219, 10, 60 },
		{ 219, 11, 60 },
		{ 219, 12, 60 },
		{ 219, 13, 60 },
		{ 219, 14, 60 },
		{ 219, 15, 60 },
		{ 219, 16, 60 },
		{ 219, 17, 60 },
		{ 219, 18, 60 },
		{ 219, 19, 60 },
		{ 219, 20, 60 },
		{ 219, 21, 60 },
		{ 219, 22, 60 },
		{ 219, 23, 60 },
		{ 219, 24, 60 },
		{ 219, 25, 60 },
		{ 219, 26, 60 },
		{ 219, 27, 60 },
		{ 219, 28, 60 },
		{ 219, 29, 60 },
		{ 219, 30, 60 },
		{ 219, 31, 60 },
		{ 219, 32, 60 },
		{ 219, 33, 60 },
		{ 219, 34, 60 },
		{ 219, 35, 60 },
		{ 219, 36, 60 },
		{ 219, 37, 60 },
		{ 219, 38, 60 },
		{ 219, 39, 60 },
		{ 219, 40, 60 },
		{ 219, 41, 60 },
		{ 219, 42, 60 },
		{ 219, 43, 60 },
		{ 219, 44, 60 },
		{ 219, 45, 60 },
		{ 219, 46, 60 },
		{ 219, 47, 60 },
		{ 219, 48, 60 },
		{ 219, 49, 60 },
		{ 219, 50, 60 },
		{ 219, 51, 60 },
		{ 219, 52, 60 },
		{ 219, 53, 60 },
		{ 219, 54, 60 },
		{ 219, 55, 60 },
		{ 219, 56, 60 },
		{ 219, 57, 60 },
		{ 219, 58, 60 },
		{ 219, 59, 60 },
		{ 219, 60, 60 },
		{ 219, 61, 60 },
		{ 219, 62, 60 },
		{ 219, 63, 60 },
		{ 219, 64, 60 },
		{ 219, 65, 60 },
		{ 219, 66, 60 },
		{ 219, 67, 60 },
		{ 219, 68, 60 },
		{ 219, 69, 60 },
		{ 219, 70, 60 },
		{ 219, 71, 60 },
		{ 219, 72, 60 },
		{ 219, 73, 60 },
		{ 219, 74, 60 },
		{ 219, 75, 60 },
		{ 219, 76, 60 },
		{ 219, 77, 60 },
		{ 219, 78, 60 },
		{ 219, 79, 60 },
		{ 219, 80, 60 },
		{ 219, 81, 60 },
		{ 219, 82, 60 },
		{ 219, 83, 60 },
		{ 219, 84, 60 },
		{ 219, 85, 60 },
		{ 219, 86, 60 },
		{ 219, 87, 60 },
		{ 219, 88, 60 },
		{ 219, 89, 60 },
		{ 219, 90, 60 },
		{ 219, 91, 60 },
		{ 219, 92, 60 },
		{ 219, 93, 60 },
		{ 219, 94, 60 },
		{ 219, 95, 60 },
		{ 219, 96, 60 },
		{ 219, 97, 60 },
		{ 219, 98, 60 },
		{ 219, 99, 60 },
		{ 219, 100, 60 },
		{ 219, 101, 60 },
		{ 219, 102, 60 },
		{ 219, 103, 60 },
		{ 219, 104, 60 },
		{ 219, 105, 60 },
		{ 219, 106, 60 },
		{ 219, 107, 60 },
		{ 219, 108, 60 },
		{ 219, 109, 60 },
		{ 219, 110, 60 },
		{ 219, 111, 60 },
		{ 219, 112, 60 },
		{ 219, 113, 60 },
		{ 219, 114, 60 },
		{ 219, 115, 60 },
		{ 219, 116, 60 },
		{ 219, 117, 60 },
		{ 219, 118, 60 },
		{ 219, 119, 60 },
		{ 219, 120, 60 },
		{ 219, 121, 60 },
		{ 219, 122, 60 },
		{ 219, 123, 60 },
		{ 219, 124, 60 },
		{ 219, 125, 60 },
		{ 219, 126, 60 },
		{ 219, 127, 60 },
		{ 219, 128, 60 },
		{ 219, 129, 60 },
		{ 219, 130, 60 },
		{ 219, 131, 60 },
		{ 219, 132, 60 },
		{ 219, 133, 60 },
		{ 219, 134, 60 },
		{ 219, 135, 60 },
		{ 219, 136, 60 },
		{ 219, 137, 60 },
		{ 219, 138, 60 },
		{ 219, 139, 60 },
		{ 219, 140, 60 },
		{ 219, 141, 60 },
		{ 219, 142, 60 },
		{ 219, 143, 60 },
		{ 219, 144, 60 },
		{ 219, 145, 60 },
		{ 219, 146, 60 },
		{ 219, 147, 60 },
		{ 219, 148, 60 },
		{ 219, 149, 60 },
		{ 219, 150, 60 },
		{ 219, 151, 60 },
		{ 219, 152, 60 },
		{ 219, 153, 60 },
		{ 219, 154, 60 },
		{ 219, 155, 60 },
		{ 220, 1, 60 },
		{ 220, 2, 60 },
		{ 220, 3, 60 },
		{ 220, 4, 60 },
		{ 220, 5, 60 },
		{ 220, 6, 60 },
		{ 220, 7, 60 },
		{ 220, 8, 60 },
		{ 220, 9, 60 },
		{ 220, 10, 60 },
		{ 220, 11, 60 },
		{ 220, 12, 60 },
		{ 220, 13, 60 },
		{ 220, 14, 60 },
		{ 220, 15, 60 },
		{ 220, 16, 60 },
		{ 220, 17, 60 },
		{ 220, 18, 60 },
		{ 220, 19, 60 },
		{ 220, 20, 60 },
		{ 220, 21, 60 },
		{ 220, 22, 60 },
		{ 220, 23, 60 },
		{ 220, 24, 60 },
		{ 220, 25, 60 },
		{ 220, 26, 60 },
		{ 220, 27, 60 },
		{ 220, 28, 60 },
		{ 220, 29, 60 },
		{ 220, 30, 60 },
		{ 220, 31, 60 },
		{ 220, 32, 60 },
		{ 220, 33, 60 },
		{ 220, 34, 60 },
		{ 220, 35, 60 },
		{ 220, 36, 60 },
		{ 220, 37, 60 },
		{ 220, 38, 60 },
		{ 220, 39, 60 },
		{ 220, 40, 60 },
		{ 220, 41, 60 },
		{ 220, 42, 60 },
		{ 220, 43, 60 },
		{ 220, 44, 60 },
		{ 220, 45, 60 },
		{ 220, 46, 60 },
		{ 220, 47, 60 },
		{ 220, 48, 60 },
		{ 220, 49, 60 },
		{ 220, 50, 60 },
		{ 220, 51, 60 },
		{ 220, 52, 60 },
		{ 220, 53, 60 },
		{ 220, 54, 60 },
		{ 220, 55, 60 },
		{ 220, 56, 60 },
		{ 220, 57, 60 },
		{ 220, 58, 60 },
		{ 220, 59, 60 },
		{ 220, 60, 60 },
		{ 220, 61, 60 },
		{ 220, 62, 60 },
		{ 220, 63, 60 },
		{ 220, 64, 60 },
		{ 220, 65, 60 },
		{ 220, 66, 60 },
		{ 220, 67, 60 },
		{ 220, 68, 60 },
		{ 220, 69, 60 },
		{ 220, 70, 60 },
		{ 220, 71, 60 },
		{ 220, 72, 60 },
		{ 220, 73, 60 },
		{ 220, 74, 60 },
		{ 220, 75, 60 },
		{ 220, 76, 60 },
		{ 220, 77, 60 },
		{ 220, 78, 60 },
		{ 220, 79, 60 },
		{ 220, 80, 60 },
		{ 220, 81, 60 },
		{ 220, 82, 60 },
		{ 220, 83, 60 },
		{ 220, 84, 60 },
		{ 220, 85, 60 },
		{ 220, 86, 60 },
		{ 220, 87, 60 },
		{ 220, 88, 60 },
		{ 220, 89, 60 },
		{ 220, 90, 60 },
		{ 220, 91, 60 },
		{ 220, 92, 60 },
		{ 220, 93, 60 },
		{ 220, 94, 60 },
		{ 220, 95, 60 },
		{ 220, 96, 60 },
		{ 220, 97, 60 },
		{ 220, 98, 60 },
		{ 220, 99, 60 },
		{ 220, 100, 60 },
		{ 220, 101, 60 },
		{ 220, 102, 60 },
		{ 220, 103, 60 },
		{ 220, 104, 60 },
		{ 220, 105, 60 },
		{ 220, 106, 60 },
		{ 220, 107, 60 },
		{ 220, 108, 60 },
		{ 220, 109, 60 },
		{ 220, 110, 60 },
		{ 220, 111, 60 },
		{ 220, 112, 60 },
		{ 220, 113, 60 },
		{ 220, 114, 60 },
		{ 220, 115, 60 },
		{ 220, 116, 60 },
		{ 220, 117, 60 },
		{ 220, 118, 60 },
		{ 220, 119, 60 },
		{ 220, 120, 60 },
		{ 220, 121, 60 },
		{ 220, 122, 60 },
		{ 220, 123, 60 },
		{ 220, 124, 60 },
		{ 220, 125, 60 },
		{ 220, 126, 60 },
		{ 220, 127, 60 },
		{ 220, 128, 60 },
		{ 220, 129, 60 },
		{ 220, 130, 60 },
		{ 220, 131, 60 },
		{ 220, 132, 60 },
		{ 220, 133, 60 },
		{ 220, 134, 60 },
		{ 220, 135, 60 },
		{ 220, 136, 60 },
		{ 220, 137, 60 },
		{ 220, 138, 60 },
		{ 220, 139, 60 },
		{ 220, 140, 60 },
		{ 220, 141, 60 },
		{ 220, 142, 60 },
		{ 220, 143, 60 },
		{ 220, 144, 60 },
		{ 220, 145, 60 },
		{ 220, 146, 60 },
		{ 220, 147, 60 },
		{ 220, 148, 60 },
		{ 220, 149, 60 },
		{ 220, 150, 60 },
		{ 220, 151, 60 },
		{ 220, 152, 60 },
		{ 220, 153, 60 },
		{ 220, 154, 60 },
		{ 220, 155, 60 },
		{ 220, 156, 60 },
		{ 220, 157, 60 },
		{ 220, 158, 60 },
		{ 221, 1, 60 },
		{ 221, 2, 60 },
		{ 221, 3, 60 },
		{ 221, 4, 60 },
		{ 221, 5, 60 },
		{ 221, 6, 60 },
		{ 221, 7, 60 },
		{ 221, 8, 60 },
		{ 221, 9, 60 },
		{ 221, 10, 60 },
		{ 221, 11, 60 },
		{ 221, 12, 60 },
		{ 221, 13, 60 },
		{ 221, 14, 60 },
		{ 221, 15, 60 },
		{ 221, 16, 60 },
		{ 221, 17, 60 },
		{ 221, 18, 60 },
		{ 221, 19, 60 },
		{ 221, 20, 60 },
		{ 221, 21, 60 },
		{ 221, 22, 60 },
		{ 221, 23, 60 },
		{ 221, 24, 60 },
		{ 221, 25, 60 },
		{ 221, 26, 60 },
		{ 221, 27, 60 },
		{ 221, 28, 60 },
		{ 221, 29, 60 },
		{ 221, 30, 60 },
		{ 221, 31, 60 },
		{ 221, 32, 60 },
		{ 221, 33, 60 },
		{ 221, 34, 60 },
		{ 221, 35, 60 },
		{ 221, 36, 60 },
		{ 221, 37, 60 },
		{ 221, 38, 60 },
		{ 221, 39, 60 },
		{ 221, 40, 60 },
		{ 221, 41, 60 },
		{ 221, 42, 60 },
		{ 221, 43, 60 },
		{ 221, 44, 60 },
		{ 221, 45, 60 },
		{ 221, 46, 60 },
		{ 221, 47, 60 },
		{ 221, 48, 60 },
		{ 221, 49, 60 },
		{ 221, 50, 60 },
		{ 221, 51, 60 },
		{ 221, 52, 60 },
		{ 221, 53, 60 },
		{ 221, 54, 60 },
		{ 221, 55, 60 },
		{ 221, 56, 60 },
		{ 221, 57, 60 },
		{ 221, 58, 60 },
		{ 221, 59, 60 },
		{ 221, 60, 60 },
		{ 221, 61, 60 },
		{ 221, 62, 60 },
		{ 221, 63, 60 },
		{ 221, 64, 60 },
		{ 221, 65, 60 },
		{ 221, 66, 60 },
		{ 221, 67, 60 },
		{ 221, 68, 60 },
		{ 221, 69, 60 },
		{ 221, 70, 60 },
		{ 221, 71, 60 },
		{ 221, 72, 60 },
		{ 221, 73, 60 },
		{ 221, 74, 60 },
		{ 221, 75, 60 },
		{ 221, 76, 60 },
		{ 221, 77, 60 },
		{ 221, 78, 60 },
		{ 221, 79, 60 },
		{ 221, 80, 60 },
		{ 221, 81, 60 },
		{ 221, 82, 60 },
		{ 221, 83, 60 },
		{ 221, 84, 60 },
		{ 221, 85, 60 },
		{ 221, 86, 60 },
		{ 221, 87, 60 },
		{ 221, 88, 60 },
		{ 221, 89, 60 },
		{ 221, 90, 60 },
		{ 221, 91, 60 },
		{ 221, 92, 60 },
		{ 221, 93, 60 },
		{ 221, 94, 60 },
		{ 221, 95, 60 },
		{ 221, 96, 60 },
		{ 221, 97, 60 },
		{ 221, 98, 60 },
		{ 221, 99, 60 },
		{ 221, 100, 60 },
		{ 221, 101, 60 },
		{ 221, 102, 60 },
		{ 221, 103, 60 },
		{ 221, 104, 60 },
		{ 221, 105, 60 },
		{ 221, 106, 60 },
		{ 221, 107, 60 },
		{ 221, 108, 60 },
		{ 221, 109, 60 },
		{ 221, 110, 60 },
		{ 221, 111, 60 },
		{ 221, 112, 60 },
		{ 221, 113, 60 },
		{ 221, 114, 60 },
		{ 221, 115, 60 },
		{ 221, 116, 60 },
		{ 221, 117, 60 },
		{ 221, 118, 60 },
		{ 221, 119, 60 },
		{ 221, 120, 60 },
		{ 221, 121, 60 },
		{ 221, 122, 60 },
		{ 221, 123, 60 },
		{ 221, 124, 60 },
		{ 221, 125, 60 },
		{ 221, 126, 60 },
		{ 221, 127, 60 },
		{ 221, 128, 60 },
		{ 221, 129, 60 },
		{ 221, 130, 60 },
		{ 221, 131, 60 },
		{ 221, 132, 60 },
		{ 221, 133, 60 },
		{ 221, 134, 60 },
		{ 221, 135, 60 },
		{ 221, 136, 60 },
		{ 221, 137, 60 },
		{ 221, 138, 60 },
		{ 221, 139, 60 },
		{ 221, 140, 60 },
		{ 221, 141, 60 },
		{ 221, 142, 60 },
		{ 221, 143, 60 },
		{ 221, 144, 60 },
		{ 221, 145, 60 },
		{ 221, 146, 60 },
		{ 221, 147, 60 },
		{ 221, 148, 60 },
		{ 221, 149, 60 },
		{ 221, 150, 60 },
		{ 221, 151, 60 },
		{ 221, 152, 60 },
		{ 221, 153, 60 },
		{ 221, 154, 60 },
		{ 221, 155, 60 },
		{ 221, 156, 60 },
		{ 221, 157, 60 },
		{ 221, 158, 60 },
		{ 221, 159, 60 },
		{ 221, 160, 60 },
		{ 222, 1, 60 },
		{ 222, 2, 60 },
		{ 222, 3, 60 },
		{ 222, 4, 60 },
		{ 222, 5, 60 },
		{ 222, 6, 60 },
		{ 222, 7, 60 },
		{ 222, 8, 60 },
		{ 222, 9, 60 },
		{ 222, 10, 60 },
		{ 222, 11, 60 },
		{ 222, 12, 60 },
		{ 222, 13, 60 },
		{ 222, 14, 60 },
		{ 222, 15, 60 },
		{ 222, 16, 60 },
		{ 222, 17, 60 },
		{ 222, 18, 60 },
		{ 222, 19, 60 },
		{ 222, 20, 60 },
		{ 222, 21, 60 },
		{ 222, 22, 60 },
		{ 222, 23, 60 },
		{ 222, 24, 60 },
		{ 222, 25, 60 },
		{ 222, 26, 60 },
		{ 222, 27, 60 },
		{ 222, 28, 60 },
		{ 222, 29, 60 },
		{ 222, 30, 60 },
		{ 222, 31, 60 },
		{ 222, 32, 60 },
		{ 222, 33, 60 },
		{ 222, 34, 60 },
		{ 222, 35, 60 },
		{ 222, 36, 60 },
		{ 222, 37, 60 },
		{ 222, 38, 60 },
		{ 222, 39, 60 },
		{ 222, 40, 60 },
		{ 222, 41, 60 },
		{ 222, 42, 60 },
		{ 222, 43, 60 },
		{ 222, 44, 60 },
		{ 222, 45, 60 },
		{ 222, 46, 60 },
		{ 222, 47, 60 },
		{ 222, 48, 60 },
		{ 222, 49, 60 },
		{ 222, 50, 60 },
		{ 222, 51, 60 },
		{ 222, 52, 60 },
		{ 222, 53, 60 },
		{ 222, 54, 60 },
		{ 222, 55, 60 },
		{ 222, 56, 60 },
		{ 222, 57, 60 },
		{ 222, 58, 60 },
		{ 222, 59, 60 },
		{ 222, 60, 60 },
		{ 222, 61, 60 },
		{ 222, 62, 60 },
		{ 222, 63, 60 },
		{ 222, 64, 60 },
		{ 222, 65, 60 },
		{ 222, 66, 60 },
		{ 222, 67, 60 },
		{ 222, 68, 60 },
		{ 222, 69, 60 },
		{ 222, 70, 60 },
		{ 222, 71, 60 },
		{ 222, 72, 60 },
		{ 222, 73, 60 },
		{ 222, 74, 60 },
		{ 222, 75, 60 },
		{ 222, 76, 60 },
		{ 222, 77, 60 },
		{ 222, 78, 60 },
		{ 222, 79, 60 },
		{ 222, 80, 60 },
		{ 222, 81, 60 },
		{ 222, 82, 60 },
		{ 222, 83, 60 },
		{ 222, 84, 60 },
		{ 222, 85, 60 },
		{ 222, 86, 60 },
		{ 222, 87, 60 },
		{ 222, 88, 60 },
		{ 222, 89, 60 },
		{ 222, 90, 60 },
		{ 222, 91, 60 },
		{ 222, 92, 60 },
		{ 222, 93, 60 },
		{ 222, 94, 60 },
		{ 222, 95, 60 },
		{ 222, 96, 60 },
		{ 222, 97, 60 },
		{ 222, 98, 60 },
		{ 222, 99, 60 },
		{ 222, 100, 60 },
		{ 222, 101, 60 },
		{ 222, 102, 60 },
		{ 222, 103, 60 },
		{ 222, 104, 60 },
		{ 222, 105, 60 },
		{ 222, 106, 60 },
		{ 222, 107, 60 },
		{ 222, 108, 60 },
		{ 222, 109, 60 },
		{ 222, 110, 60 },
		{ 222, 111, 60 },
		{ 222, 112, 60 },
		{ 222, 113, 60 },
		{ 222, 114, 60 },
		{ 222, 115, 60 },
		{ 222, 116, 60 },
		{ 222, 117, 60 },
		{ 222, 118, 60 },
		{ 222, 119, 60 },
		{ 222, 120, 60 },
		{ 222, 121, 60 },
		{ 222, 122, 60 },
		{ 222, 123, 60 },
		{ 222, 124, 60 },
		{ 222, 125, 60 },
		{ 222, 126, 60 },
		{ 222, 127, 60 },
		{ 222, 128, 60 },
		{ 222, 129, 60 },
		{ 222, 130, 60 },
		{ 222, 131, 60 },
		{ 222, 132, 60 },
		{ 222, 133, 60 },
		{ 222, 134, 60 },
		{ 222, 135, 60 },
		{ 222, 136, 60 },
		{ 222, 137, 60 },
		{ 222, 138, 60 },
		{ 222, 139, 60 },
		{ 222, 140, 60 },
		{ 222, 141, 60 },
		{ 222, 142, 60 },
		{ 222, 143, 60 },
		{ 222, 144, 60 },
		{ 222, 145, 60 },
		{ 222, 146, 60 },
		{ 222, 147, 60 },
		{ 222, 148, 60 },
		{ 222, 149, 60 },
		{ 222, 150, 60 },
		{ 222, 151, 60 },
		{ 222, 152, 60 },
		{ 222, 153, 60 },
		{ 222, 154, 60 },
		{ 222, 155, 60 },
		{ 222, 156, 60 },
		{ 222, 157, 60 },
		{ 222, 158, 60 },
		{ 222, 159, 60 },
		{ 222, 160, 60 },
		{ 222, 161, 60 },
		{ 222, 162, 60 },
		{ 222, 163, 60 },
		{ 223, 1, 60 },
		{ 223, 2, 60 },
		{ 223, 3, 60 },
		{ 223, 4, 60 },
		{ 223, 5, 60 },
		{ 223, 6, 60 },
		{ 223, 7, 60 },
		{ 223, 8, 60 },
		{ 223, 9, 60 },
		{ 223, 10, 60 },
		{ 223, 11, 60 },
		{ 223, 12, 60 },
		{ 223, 13, 60 },
		{ 223, 14, 60 },
		{ 223, 15, 60 },
		{ 223, 16, 60 },
		{ 223, 17, 60 },
		{ 223, 18, 60 },
		{ 223, 19, 60 },
		{ 223, 20, 60 },
		{ 223, 21, 60 },
		{ 223, 22, 60 },
		{ 223, 23, 60 },
		{ 223, 24, 60 },
		{ 223, 25, 60 },
		{ 223, 26, 60 },
		{ 223, 27, 60 },
		{ 223, 28, 60 },
		{ 223, 29, 60 },
		{ 223, 30, 60 },
		{ 223, 31, 60 },
		{ 223, 32, 60 },
		{ 223, 33, 60 },
		{ 223, 34, 60 },
		{ 223, 35, 60 },
		{ 223, 36, 60 },
		{ 223, 37, 60 },
		{ 223, 38, 60 },
		{ 223, 39, 60 },
		{ 223, 40, 60 },
		{ 223, 41, 60 },
		{ 223, 42, 60 },
		{ 223, 43, 60 },
		{ 223, 44, 60 },
		{ 223, 45, 60 },
		{ 223, 46, 60 },
		{ 223, 47, 60 },
		{ 223, 48, 60 },
		{ 223, 49, 60 },
		{ 223, 50, 60 },
		{ 223, 51, 60 },
		{ 223, 52, 60 },
		{ 223, 53, 60 },
		{ 223, 54, 60 },
		{ 223, 55, 60 },
		{ 223, 56, 60 },
		{ 223, 57, 60 },
		{ 223, 58, 60 },
		{ 223, 59, 60 },
		{ 223, 60, 60 },
		{ 223, 61, 60 },
		{ 223, 62, 60 },
		{ 223, 63, 60 },
		{ 223, 64, 60 },
		{ 223, 65, 60 },
		{ 223, 66, 60 },
		{ 223, 67, 60 },
		{ 223, 68, 60 },
		{ 223, 69, 60 },
		{ 223, 70, 60 },
		{ 223, 71, 60 },
		{ 223, 72, 60 },
		{ 223, 73, 60 },
		{ 223, 74, 60 },
		{ 223, 75, 60 },
		{ 223, 76, 60 },
		{ 223, 77, 60 },
		{ 223, 78, 60 },
		{ 223, 79, 60 },
		{ 223, 80, 60 },
		{ 223, 81, 60 },
		{ 223, 82, 60 },
		{ 223, 83, 60 },
		{ 223, 84, 60 },
		{ 223, 85, 60 },
		{ 223, 86, 60 },
		{ 223, 87, 60 },
		{ 223, 88, 60 },
		{ 223, 89, 60 },
		{ 223, 90, 60 },
		{ 223, 91, 60 },
		{ 223, 92, 60 },
		{ 223, 93, 60 },
		{ 223, 94, 60 },
		{ 223, 95, 60 },
		{ 223, 96, 60 },
		{ 223, 97, 60 },
		{ 223, 98, 60 },
		{ 223, 99, 60 },
		{ 223, 100, 60 },
		{ 223, 101, 60 },
		{ 223, 102, 60 },
		{ 223, 103, 60 },
		{ 223, 104, 60 },
		{ 223, 105, 60 },
		{ 223, 106, 60 },
		{ 223, 107, 60 },
		{ 223, 108, 60 },
		{ 223, 109, 60 },
		{ 223, 110, 60 },
		{ 223, 111, 60 },
		{ 223, 112, 60 },
		{ 223, 113, 60 },
		{ 223, 114, 60 },
		{ 223, 115, 60 },
		{ 223, 116, 60 },
		{ 223, 117, 60 },
		{ 223, 118, 60 },
		{ 223, 119, 60 },
		{ 223, 120, 60 },
		{ 223, 121, 60 },
		{ 223, 122, 60 },
		{ 223, 123, 60 },
		{ 223, 124, 60 },
		{ 223, 125, 60 },
		{ 223, 126, 60 },
		{ 223, 127, 60 },
		{ 223, 128, 60 },
		{ 223, 129, 60 },
		{ 223, 130, 60 },
		{ 223, 131, 60 },
		{ 223, 132, 60 },
		{ 223, 133, 60 },
		{ 223, 134, 60 },
		{ 223, 135, 60 },
		{ 223, 136, 60 },
		{ 223, 137, 60 },
		{ 223, 138, 60 },
		{ 223, 139, 60 },
		{ 223, 140, 60 },
		{ 223, 141, 60 },
		{ 223, 142, 60 },
		{ 223, 143, 60 },
		{ 223, 144, 60 },
		{ 223, 145, 60 },
		{ 223, 146, 60 },
		{ 223, 147, 60 },
		{ 223, 148, 60 },
		{ 223, 149, 60 },
		{ 223, 150, 60 },
		{ 223, 151, 60 },
		{ 223, 152, 60 },
		{ 223, 153, 60 },
		{ 223, 154, 60 },
		{ 223, 155, 60 },
		{ 223, 156, 60 },
		{ 223, 157, 60 },
		{ 223, 158, 60 },
		{ 223, 159, 60 },
		{ 223, 160, 60 },
		{ 223, 161, 60 },
		{ 223, 162, 60 },
		{ 223, 163, 60 },
		{ 223, 164, 60 },
		{ 223, 165, 60 },
		{ 223, 166, 60 },
		{ 224, 1, 60 },
		{ 224, 2, 60 },
		{ 224, 3, 60 },
		{ 224, 4, 60 },
		{ 224, 5, 60 },
		{ 224, 6, 60 },
		{ 224, 7, 60 },
		{ 224, 8, 60 },
		{ 224, 9, 60 },
		{ 224, 10, 60 },
		{ 224, 11, 60 },
		{ 224, 12, 60 },
		{ 224, 13, 60 },
		{ 224, 14, 60 },
		{ 224, 15, 60 },
		{ 224, 16, 60 },
		{ 224, 17, 60 },
		{ 224, 18, 60 },
		{ 224, 19, 60 },
		{ 224, 20, 60 },
		{ 224, 21, 60 },
		{ 224, 22, 60 },
		{ 224, 23, 60 },
		{ 224, 24, 60 },
		{ 224, 25, 60 },
		{ 224, 26, 60 },
		{ 224, 27, 60 },
		{ 224, 28, 60 },
		{ 224, 29, 60 },
		{ 224, 30, 60 },
		{ 224, 31, 60 },
		{ 224, 32, 60 },
		{ 224, 33, 60 },
		{ 224, 34, 60 },
		{ 224, 35, 60 },
		{ 224, 36, 60 },
		{ 224, 37, 60 },
		{ 224, 38, 60 },
		{ 224, 39, 60 },
		{ 224, 40, 60 },
		{ 224, 41, 60 },
		{ 224, 42, 60 },
		{ 224, 43, 60 },
		{ 224, 44, 60 },
		{ 224, 45, 60 },
		{ 224, 46, 60 },
		{ 224, 47, 60 },
		{ 224, 48, 60 },
		{ 224, 49, 60 },
		{ 224, 50, 60 },
		{ 224, 51, 60 },
		{ 224, 52, 60 },
		{ 224, 53, 60 },
		{ 224, 54, 60 },
		{ 224, 55, 60 },
		{ 224, 56, 60 },
		{ 224, 57, 60 },
		{ 224, 58, 60 },
		{ 224, 59, 60 },
		{ 224, 60, 60 },
		{ 224, 61, 60 },
		{ 224, 62, 60 },
		{ 224, 63, 60 },
		{ 224, 64, 60 },
		{ 224, 65, 60 },
		{ 224, 66, 60 },
		{ 224, 67, 60 },
		{ 224, 68, 60 },
		{ 224, 69, 60 },
		{ 224, 70, 60 },
		{ 224, 71, 60 },
		{ 224, 72, 60 },
		{ 224, 73, 60 },
		{ 224, 74, 60 },
		{ 224, 75, 60 },
		{ 224, 76, 60 },
		{ 224, 77, 60 },
		{ 224, 78, 60 },
		{ 224, 79, 60 },
		{ 224, 80, 60 },
		{ 224, 81, 60 },
		{ 224, 82, 60 },
		{ 224, 83, 60 },
		{ 224, 84, 60 },
		{ 224, 85, 60 },
		{ 224, 86, 60 },
		{ 224, 87, 60 },
		{ 224, 88, 60 },
		{ 224, 89, 60 },
		{ 224, 90, 60 },
		{ 224, 91, 60 },
		{ 224, 92, 60 },
		{ 224, 93, 60 },
		{ 224, 94, 60 },
		{ 224, 95, 60 },
		{ 224, 96, 60 },
		{ 224, 97, 60 },
		{ 224, 98, 60 },
		{ 224, 99, 60 },
		{ 224, 100, 60 },
		{ 224, 101, 60 },
		{ 224, 102, 60 },
		{ 224, 103, 60 },
		{ 224, 104, 60 },
		{ 224, 105, 60 },
		{ 224, 106, 60 },
		{ 224, 107, 60 },
		{ 224, 108, 60 },
		{ 224, 109, 60 },
		{ 224, 110, 60 },
		{ 224, 111, 60 },
		{ 224, 112, 60 },
		{ 224, 113, 60 },
		{ 224, 114, 60 },
		{ 224, 115, 60 },
		{ 224, 116, 60 },
		{ 224, 117, 60 },
		{ 224, 118, 60 },
		{ 224, 119, 60 },
		{ 224, 120, 60 },
		{ 224, 121, 60 },
		{ 224, 122, 60 },
		{ 224, 123, 60 },
		{ 224, 124, 60 },
		{ 224, 125, 60 },
		{ 224, 126, 60 },
		{ 224, 127, 60 },
		{ 224, 128, 60 },
		{ 224, 129, 60 },
		{ 224, 130, 60 },
		{ 224, 131, 60 },
		{ 224, 132, 60 },
		{ 224, 133, 60 },
		{ 224, 134, 60 },
		{ 224, 135, 60 },
		{ 224, 136, 60 },
		{ 224, 137, 60 },
		{ 224, 138, 60 },
		{ 224, 139, 60 },
		{ 224, 140, 60 },
		{ 224, 141, 60 },
		{ 224, 142, 60 },
		{ 224, 143, 60 },
		{ 224, 144, 60 },
		{ 224, 145, 60 },
		{ 224, 146, 60 },
		{ 224, 147, 60 },
		{ 224, 148, 60 },
		{ 224, 149, 60 },
		{ 224, 150, 60 },
		{ 224, 151, 60 },
		{ 224, 152, 60 },
		{ 224, 153, 60 },
		{ 224, 154, 60 },
		{ 224, 155, 60 },
		{ 224, 156, 60 },
		{ 224, 157, 60 },
		{ 224, 158, 60 },
		{ 224, 159, 60 },
		{ 224, 160, 60 },
		{ 224, 161, 60 },
		{ 224, 162, 60 },
		{ 224, 163, 60 },
		{ 224, 164, 60 },
		{ 224, 165, 60 },
		{ 224, 166, 60 },
		{ 224, 167, 60 },
		{ 224, 168, 60 },
		{ 225, 2, 60 },
		{ 225, 3, 60 },
		{ 225, 4, 60 },
		{ 225, 5, 60 },
		{ 225, 6, 60 },
		{ 225, 7, 60 },
		{ 225, 8, 60 },
		{ 225, 9, 60 },
		{ 225, 10, 60 },
		{ 225, 11, 60 },
		{ 225, 12, 60 },
		{ 225, 13, 60 },
		{ 225, 14, 60 },
		{ 225, 15, 60 },
		{ 225, 16, 60 },
		{ 225, 17, 60 },
		{ 225, 18, 60 },
		{ 225, 19, 60 },
		{ 225, 20, 60 },
		{ 225, 21, 60 },
		{ 225, 22, 60 },
		{ 225, 23, 60 },
		{ 225, 24, 60 },
		{ 225, 25, 60 },
		{ 225, 26, 60 },
		{ 225, 27, 60 },
		{ 225, 28, 60 },
		{ 225, 29, 60 },
		{ 225, 30, 60 },
		{ 225, 31, 60 },
		{ 225, 32, 60 },
		{ 225, 33, 60 },
		{ 225, 34, 60 },
		{ 225, 35, 60 },
		{ 225, 36, 60 },
		{ 225, 37, 60 },
		{ 225, 38, 60 },
		{ 225, 39, 60 },
		{ 225, 40, 60 },
		{ 225, 41, 60 },
		{ 225, 42, 60 },
		{ 225, 43, 60 },
		{ 225, 44, 60 },
		{ 225, 45, 60 },
		{ 225, 46, 60 },
		{ 225, 47, 60 },
		{ 225, 48, 60 },
		{ 225, 49, 60 },
		{ 225, 50, 60 },
		{ 225, 51, 60 },
		{ 225, 52, 60 },
		{ 225, 53, 60 },
		{ 225, 54, 60 },
		{ 225, 55, 60 },
		{ 225, 56, 60 },
		{ 225, 57, 60 },
		{ 225, 58, 60 },
		{ 225, 59, 60 },
		{ 225, 60, 60 },
		{ 225, 61, 60 },
		{ 225, 62, 60 },
		{ 225, 63, 60 },
		{ 225, 64, 60 },
		{ 225, 65, 60 },
		{ 225, 66, 60 },
		{ 225, 67, 60 },
		{ 225, 68, 60 },
		{ 225, 69, 60 },
		{ 225, 70, 60 },
		{ 225, 71, 60 },
		{ 225, 72, 60 },
		{ 225, 73, 60 },
		{ 225, 74, 60 },
		{ 225, 75, 60 },
		{ 225, 76, 60 },
		{ 225, 77, 60 },
		{ 225, 78, 60 },
		{ 225, 79, 60 },
		{ 225, 80, 60 },
		{ 225, 81, 60 },
		{ 225, 82, 60 },
		{ 225, 83, 60 },
		{ 225, 84, 60 },
		{ 225, 85, 60 },
		{ 225, 86, 60 },
		{ 225, 87, 60 },
		{ 225, 88, 60 },
		{ 225, 89, 60 },
		{ 225, 90, 60 },
		{ 225, 91, 60 },
		{ 225, 92, 60 },
		{ 225, 93, 60 },
		{ 225, 94, 60 },
		{ 225, 95, 60 },
		{ 225, 96, 60 },
		{ 225, 97, 60 },
		{ 225, 98, 60 },
		{ 225, 99, 60 },
		{ 225, 100, 60 },
		{ 225, 101, 60 },
		{ 225, 102, 60 },
		{ 225, 103, 60 },
		{ 225, 104, 60 },
		{ 225, 105, 60 },
		{ 225, 106, 60 },
		{ 225, 107, 60 },
		{ 225, 108, 60 },
		{ 225, 109, 60 },
		{ 225, 110, 60 },
		{ 225, 111, 60 },
		{ 225, 112, 60 },
		{ 225, 113, 60 },
		{ 225, 114, 60 },
		{ 225, 115, 60 },
		{ 225, 116, 60 },
		{ 225, 117, 60 },
		{ 225, 118, 60 },
		{ 225, 119, 60 },
		{ 225, 120, 60 },
		{ 225, 121, 60 },
		{ 225, 122, 60 },
		{ 225, 123, 60 },
		{ 225, 124, 60 },
		{ 225, 125, 60 },
		{ 225, 126, 60 },
		{ 225, 127, 60 },
		{ 225, 128, 60 },
		{ 225, 129, 60 },
		{ 225, 130, 60 },
		{ 225, 131, 60 },
		{ 225, 132, 60 },
		{ 225, 133, 60 },
		{ 225, 134, 60 },
		{ 225, 135, 60 },
		{ 225, 136, 60 },
		{ 225, 137, 60 },
		{ 225, 138, 60 },
		{ 225, 139, 60 },
		{ 225, 140, 60 },
		{ 225, 141, 60 },
		{ 225, 142, 60 },
		{ 225, 143, 60 },
		{ 225, 144, 60 },
		{ 225, 145, 60 },
		{ 225, 146, 60 },
		{ 225, 147, 60 },
		{ 225, 148, 60 },
		{ 225, 149, 60 },
		{ 225, 150, 60 },
		{ 225, 151, 60 },
		{ 225, 152, 60 },
		{ 225, 153, 60 },
		{ 225, 154, 60 },
		{ 225, 155, 60 },
		{ 225, 156, 60 },
		{ 225, 157, 60 },
		{ 225, 158, 60 },
		{ 225, 159, 60 },
		{ 225, 160, 60 },
		{ 225, 161, 60 },
		{ 225, 162, 60 },
		{ 225, 163, 60 },
		{ 225, 164, 60 },
		{ 225, 165, 60 },
		{ 225, 166, 60 },
		{ 225, 167, 60 },
		{ 225, 168, 60 },
		{ 225, 169, 60 },
		{ 225, 170, 60 },
		{ 225, 171, 60 },
		{ 226, 10, 60 },
		{ 226, 11, 60 },
		{ 226, 12, 60 },
		{ 226, 13, 60 },
		{ 226, 14, 60 },
		{ 226, 15, 60 },
		{ 226, 16, 60 },
		{ 226, 17, 60 },
		{ 226, 18, 60 },
		{ 226, 19, 60 },
		{ 226, 20, 60 },
		{ 226, 21, 60 },
		{ 226, 22, 60 },
		{ 226, 23, 60 },
		{ 226, 24, 60 },
		{ 226, 25, 60 },
		{ 226, 26, 60 },
		{ 226, 27, 60 },
		{ 226, 28, 60 },
		{ 226, 29, 60 },
		{ 226, 30, 60 },
		{ 226, 31, 60 },
		{ 226, 32, 60 },
		{ 226, 33, 60 },
		{ 226, 34, 60 },
		{ 226, 35, 60 },
		{ 226, 36, 60 },
		{ 226, 37, 60 },
		{ 226, 38, 60 },
		{ 226, 39, 60 },
		{ 226, 40, 60 },
		{ 226, 41, 60 },
		{ 226, 42, 60 },
		{ 226, 43, 60 },
		{ 226, 44, 60 },
		{ 226, 45, 60 },
		{ 226, 46, 60 },
		{ 226, 47, 60 },
		{ 226, 48, 60 },
		{ 226, 49, 60 },
		{ 226, 50, 60 },
		{ 226, 51, 60 },
		{ 226, 52, 60 },
		{ 226, 53, 60 },
		{ 226, 54, 60 },
		{ 226, 55, 60 },
		{ 226, 56, 60 },
		{ 226, 57, 60 },
		{ 226, 58, 60 },
		{ 226, 59, 60 },
		{ 226, 60, 60 },
		{ 226, 61, 60 },
		{ 226, 62, 60 },
		{ 226, 63, 60 },
		{ 226, 64, 60 },
		{ 226, 65, 60 },
		{ 226, 66, 60 },
		{ 226, 67, 60 },
		{ 226, 68, 60 },
		{ 226, 69, 60 },
		{ 226, 70, 60 },
		{ 226, 71, 60 },
		{ 226, 72, 60 },
		{ 226, 73, 60 },
		{ 226, 74, 60 },
		{ 226, 75, 60 },
		{ 226, 76, 60 },
		{ 226, 77, 60 },
		{ 226, 78, 60 },
		{ 226, 79, 60 },
		{ 226, 80, 60 },
		{ 226, 81, 60 },
		{ 226, 82, 60 },
		{ 226, 83, 60 },
		{ 226, 84, 60 },
		{ 226, 85, 60 },
		{ 226, 86, 60 },
		{ 226, 87, 60 },
		{ 226, 88, 60 },
		{ 226, 89, 60 },
		{ 226, 90, 60 },
		{ 226, 91, 60 },
		{ 226, 92, 60 },
		{ 226, 93, 60 },
		{ 226, 94, 60 },
		{ 226, 95, 60 },
		{ 226, 96, 60 },
		{ 226, 97, 60 },
		{ 226, 98, 60 },
		{ 226, 99, 60 },
		{ 226, 100, 60 },
		{ 226, 101, 60 },
		{ 226, 102, 60 },
		{ 226, 103, 60 },
		{ 226, 104, 60 },
		{ 226, 105, 60 },
		{ 226, 106, 60 },
		{ 226, 107, 60 },
		{ 226, 108, 60 },
		{ 226, 109, 60 },
		{ 226, 110, 60 },
		{ 226, 111, 60 },
		{ 226, 112, 60 },
		{ 226, 113, 60 },
		{ 226, 114, 60 },
		{ 226, 115, 60 },
		{ 226, 116, 60 },
		{ 226, 117, 60 },
		{ 226, 118, 60 },
		{ 226, 119, 60 },
		{ 226, 120, 60 },
		{ 226, 121, 60 },
		{ 226, 122, 60 },
		{ 226, 123, 60 },
		{ 226, 124, 60 },
		{ 226, 125, 60 },
		{ 226, 126, 60 },
		{ 226, 127, 60 },
		{ 226, 128, 60 },
		{ 226, 129, 60 },
		{ 226, 130, 60 },
		{ 226, 131, 60 },
		{ 226, 132, 60 },
		{ 226, 133, 60 },
		{ 226, 134, 60 },
		{ 226, 135, 60 },
		{ 226, 136, 60 },
		{ 226, 137, 60 },
		{ 226, 138, 60 },
		{ 226, 139, 60 },
		{ 226, 140, 60 },
		{ 226, 141, 60 },
		{ 226, 142, 60 },
		{ 226, 143, 60 },
		{ 226, 144, 60 },
		{ 226, 145, 60 },
		{ 226, 146, 60 },
		{ 226, 147, 60 },
		{ 226, 148, 60 },
		{ 226, 149, 60 },
		{ 226, 150, 60 },
		{ 226, 151, 60 },
		{ 226, 152, 60 },
		{ 226, 153, 60 },
		{ 226, 154, 60 },
		{ 226, 155, 60 },
		{ 226, 156, 60 },
		{ 226, 157, 60 },
		{ 226, 158, 60 },
		{ 226, 159, 60 },
		{ 226, 160, 60 },
		{ 226, 161, 60 },
		{ 226, 162, 60 },
		{ 226, 163, 60 },
		{ 226, 164, 60 },
		{ 226, 165, 60 },
		{ 226, 166, 60 },
		{ 226, 167, 60 },
		{ 226, 168, 60 },
		{ 226, 169, 60 },
		{ 226, 170, 60 },
		{ 226, 171, 60 },
		{ 226, 172, 60 },
		{ 226, 173, 60 },
		{ 226, 174, 60 },
		{ 227, 18, 60 },
		{ 227, 19, 60 },
		{ 227, 20, 60 },
		{ 227, 21, 60 },
		{ 227, 22, 60 },
		{ 227, 23, 60 },
		{ 227, 24, 60 },
		{ 227, 25, 60 },
		{ 227, 26, 60 },
		{ 227, 27, 60 },
		{ 227, 28, 60 },
		{ 227, 29, 60 },
		{ 227, 30, 60 },
		{ 227, 31, 60 },
		{ 227, 32, 60 },
		{ 227, 33, 60 },
		{ 227, 34, 60 },
		{ 227, 35, 60 },
		{ 227, 36, 60 },
		{ 227, 37, 60 },
		{ 227, 38, 60 },
		{ 227, 39, 60 },
		{ 227, 40, 60 },
		{ 227, 41, 60 },
		{ 227, 42, 60 },
		{ 227, 43, 60 },
		{ 227, 44, 60 },
		{ 227, 45, 60 },
		{ 227, 46, 60 },
		{ 227, 47, 60 },
		{ 227, 48, 60 },
		{ 227, 49, 60 },
		{ 227, 50, 60 },
		{ 227, 51, 60 },
		{ 227, 52, 60 },
		{ 227, 53, 60 },
		{ 227, 54, 60 },
		{ 227, 55, 60 },
		{ 227, 56, 60 },
		{ 227, 57, 60 },
		{ 227, 58, 60 },
		{ 227, 59, 60 },
		{ 227, 60, 60 },
		{ 227, 61, 60 },
		{ 227, 62, 60 },
		{ 227, 63, 60 },
		{ 227, 64, 60 },
		{ 227, 65, 60 },
		{ 227, 66, 60 },
		{ 227, 67, 60 },
		{ 227, 68, 60 },
		{ 227, 69, 60 },
		{ 227, 70, 60 },
		{ 227, 71, 60 },
		{ 227, 72, 60 },
		{ 227, 73, 60 },
		{ 227, 74, 60 },
		{ 227, 75, 60 },
		{ 227, 76, 60 },
		{ 227, 77, 60 },
		{ 227, 78, 60 },
		{ 227, 79, 60 },
		{ 227, 80, 60 },
		{ 227, 81, 60 },
		{ 227, 82, 60 },
		{ 227, 83, 60 },
		{ 227, 84, 60 },
		{ 227, 85, 60 },
		{ 227, 86, 60 },
		{ 227, 87, 60 },
		{ 227, 88, 60 },
		{ 227, 89, 60 },
		{ 227, 90, 60 },
		{ 227, 91, 60 },
		{ 227, 92, 60 },
		{ 227, 93, 60 },
		{ 227, 94, 60 },
		{ 227, 95, 60 },
		{ 227, 96, 60 },
		{ 227, 97, 60 },
		{ 227, 98, 60 },
		{ 227, 99, 60 },
		{ 227, 100, 60 },
		{ 227, 101, 60 },
		{ 227, 102, 60 },
		{ 227, 103, 60 },
		{ 227, 104, 60 },
		{ 227, 105, 60 },
		{ 227, 106, 60 },
		{ 227, 107, 60 },
		{ 227, 108, 60 },
		{ 227, 109, 60 },
		{ 227, 110, 60 },
		{ 227, 111, 60 },
		{ 227, 112, 60 },
		{ 227, 113, 60 },
		{ 227, 114, 60 },
		{ 227, 115, 60 },
		{ 227, 116, 60 },
		{ 227, 117, 60 },
		{ 227, 118, 60 },
		{ 227, 119, 60 },
		{ 227, 120, 60 },
		{ 227, 121, 60 },
		{ 227, 122, 60 },
		{ 227, 123, 60 },
		{ 227, 124, 60 },
		{ 227, 125, 60 },
		{ 227, 126, 60 },
		{ 227, 127, 60 },
		{ 227, 128, 60 },
		{ 227, 129, 60 },
		{ 227, 130, 60 },
		{ 227, 131, 60 },
		{ 227, 132, 60 },
		{ 227, 133, 60 },
		{ 227, 134, 60 },
		{ 227, 135, 60 },
		{ 227, 136, 60 },
		{ 227, 137, 60 },
		{ 227, 138, 60 },
		{ 227, 139, 60 },
		{ 227, 140, 60 },
		{ 227, 141, 60 },
		{ 227, 142, 60 },
		{ 227, 143, 60 },
		{ 227, 144, 60 },
		{ 227, 145, 60 },
		{ 227, 146, 60 },
		{ 227, 147, 60 },
		{ 227, 148, 60 },
		{ 227, 149, 60 },
		{ 227, 150, 60 },
		{ 227, 151, 60 },
		{ 227, 152, 60 },
		{ 227, 153, 60 },
		{ 227, 154, 60 },
		{ 227, 155, 60 },
		{ 227, 156, 60 },
		{ 227, 157, 60 },
		{ 227, 158, 60 },
		{ 227, 159, 60 },
		{ 227, 160, 60 },
		{ 227, 161, 60 },
		{ 227, 162, 60 },
		{ 227, 163, 60 },
		{ 227, 164, 60 },
		{ 227, 165, 60 },
		{ 227, 166, 60 },
		{ 227, 167, 60 },
		{ 227, 168, 60 },
		{ 227, 169, 60 },
		{ 227, 170, 60 },
		{ 227, 171, 60 },
		{ 227, 172, 60 },
		{ 227, 173, 60 },
		{ 227, 174, 60 },
		{ 227, 175, 60 },
		{ 227, 176, 60 },
		{ 228, 26, 60 },
		{ 228, 27, 60 },
		{ 228, 28, 60 },
		{ 228, 29, 60 },
		{ 228, 30, 60 },
		{ 228, 31, 60 },
		{ 228, 32, 60 },
		{ 228, 33, 60 },
		{ 228, 34, 60 },
		{ 228, 35, 60 },
		{ 228, 36, 60 },
		{ 228, 37, 60 },
		{ 228, 38, 60 },
		{ 228, 39, 60 },
		{ 228, 40, 60 },
		{ 228, 41, 60 },
		{ 228, 42, 60 },
		{ 228, 43, 60 },
		{ 228, 44, 60 },
		{ 228, 45, 60 },
		{ 228, 46, 60 },
		{ 228, 47, 60 },
		{ 228, 48, 60 },
		{ 228, 49, 60 },
		{ 228, 50, 60 },
		{ 228, 51, 60 },
		{ 228, 52, 60 },
		{ 228, 53, 60 },
		{ 228, 54, 60 },
		{ 228, 55, 60 },
		{ 228, 56, 60 },
		{ 228, 57, 60 },
		{ 228, 58, 60 },
		{ 228, 59, 60 },
		{ 228, 60, 60 },
		{ 228, 61, 60 },
		{ 228, 62, 60 },
		{ 228, 63, 60 },
		{ 228, 64, 60 },
		{ 228, 65, 60 },
		{ 228, 66, 60 },
		{ 228, 67, 60 },
		{ 228, 68, 60 },
		{ 228, 69, 60 },
		{ 228, 70, 60 },
		{ 228, 71, 60 },
		{ 228, 72, 60 },
		{ 228, 73, 60 },
		{ 228, 74, 60 },
		{ 228, 75, 60 },
		{ 228, 76, 60 },
		{ 228, 77, 60 },
		{ 228, 78, 60 },
		{ 228, 79, 60 },
		{ 228, 80, 60 },
		{ 228, 81, 60 },
		{ 228, 82, 60 },
		{ 228, 83, 60 },
		{ 228, 84, 60 },
		{ 228, 85, 60 },
		{ 228, 86, 60 },
		{ 228, 87, 60 },
		{ 228, 88, 60 },
		{ 228, 89, 60 },
		{ 228, 90, 60 },
		{ 228, 91, 60 },
		{ 228, 92, 60 },
		{ 228, 93, 60 },
		{ 228, 94, 60 },
		{ 228, 95, 60 },
		{ 228, 96, 60 },
		{ 228, 97, 60 },
		{ 228, 98, 60 },
		{ 228, 99, 60 },
		{ 228, 100, 60 },
		{ 228, 101, 60 },
		{ 228, 102, 60 },
		{ 228, 103, 60 },
		{ 228, 104, 60 },
		{ 228, 105, 60 },
		{ 228, 106, 60 },
		{ 228, 107, 60 },
		{ 228, 108, 60 },
		{ 228, 109, 60 },
		{ 228, 110, 60 },
		{ 228, 111, 60 },
		{ 228, 112, 60 },
		{ 228, 113, 60 },
		{ 228, 114, 60 },
		{ 228, 115, 60 },
		{ 228, 116, 60 },
		{ 228, 117, 60 },
		{ 228, 118, 60 },
		{ 228, 119, 60 },
		{ 228, 120, 60 },
		{ 228, 121, 60 },
		{ 228, 122, 60 },
		{ 228, 123, 60 },
		{ 228, 124, 60 },
		{ 228, 125, 60 },
		{ 228, 126, 60 },
		{ 228, 127, 60 },
		{ 228, 128, 60 },
		{ 228, 129, 60 },
		{ 228, 130, 60 },
		{ 228, 131, 60 },
		{ 228, 132, 60 },
		{ 228, 133, 60 },
		{ 228, 134, 60 },
		{ 228, 135, 60 },
		{ 228, 136, 60 },
		{ 228, 137, 60 },
		{ 228, 138, 60 },
		{ 228, 139, 60 },
		{ 228, 140, 60 },
		{ 228, 141, 60 },
		{ 228, 142, 60 },
		{ 228, 143, 60 },
		{ 228, 144, 60 },
		{ 228, 145, 60 },
		{ 228, 146, 60 },
		{ 228, 147, 60 },
		{ 228, 148, 60 },
		{ 228, 149, 60 },
		{ 228, 150, 60 },
		{ 228, 151, 60 },
		{ 228, 152, 60 },
		{ 228, 153, 60 },
		{ 228, 154, 60 },
		{ 228, 155, 60 },
		{ 228, 156, 60 },
		{ 228, 157, 60 },
		{ 228, 158, 60 },
		{ 228, 159, 60 },
		{ 228, 160, 60 },
		{ 228, 161, 60 },
		{ 228, 162, 60 },
		{ 228, 163, 60 },
		{ 228, 164, 60 },
		{ 228, 165, 60 },
		{ 228, 166, 60 },
		{ 228, 167, 60 },
		{ 228, 168, 60 },
		{ 228, 169, 60 },
		{ 228, 170, 60 },
		{ 228, 171, 60 },
		{ 228, 172, 60 },
		{ 228, 173, 60 },
		{ 228, 174, 60 },
		{ 228, 175, 60 },
		{ 228, 176, 60 },
		{ 228, 177, 60 },
		{ 228, 178, 60 },
		{ 228, 179, 60 },
		{ 229, 34, 60 },
		{ 229, 35, 60 },
		{ 229, 36, 60 },
		{ 229, 37, 60 },
		{ 229, 38, 60 },
		{ 229, 39, 60 },
		{ 229, 40, 60 },
		{ 229, 41, 60 },
		{ 229, 42, 60 },
		{ 229, 43, 60 },
		{ 229, 44, 60 },
		{ 229, 45, 60 },
		{ 229, 46, 60 },
		{ 229, 47, 60 },
		{ 229, 48, 60 },
		{ 229, 49, 60 },
		{ 229, 50, 60 },
		{ 229, 51, 60 },
		{ 229, 52, 60 },
		{ 229, 53, 60 },
		{ 229, 54, 60 },
		{ 229, 55, 60 },
		{ 229, 56, 60 },
		{ 229, 57, 60 },
		{ 229, 58, 60 },
		{ 229, 59, 60 },
		{ 229, 60, 60 },
		{ 229, 61, 60 },
		{ 229, 62, 60 },
		{ 229, 63, 60 },
		{ 229, 64, 60 },
		{ 229, 65, 60 },
		{ 229, 66, 60 },
		{ 229, 67, 60 },
		{ 229, 68, 60 },
		{ 229, 69, 60 },
		{ 229, 70, 60 },
		{ 229, 71, 60 },
		{ 229, 72, 60 },
		{ 229, 73, 60 },
		{ 229, 74, 60 },
		{ 229, 75, 60 },
		{ 229, 76, 60 },
		{ 229, 77, 60 },
		{ 229, 78, 60 },
		{ 229, 79, 60 },
		{ 229, 80, 60 },
		{ 229, 81, 60 },
		{ 229, 82, 60 },
		{ 229, 83, 60 },
		{ 229, 84, 60 },
		{ 229, 85, 60 },
		{ 229, 86, 60 },
		{ 229, 87, 60 },
		{ 229, 88, 60 },
		{ 229, 89, 60 },
		{ 229, 90, 60 },
		{ 229, 91, 60 },
		{ 229, 92, 60 },
		{ 229, 93, 60 },
		{ 229, 94, 60 },
		{ 229, 95, 60 },
		{ 229, 96, 60 },
		{ 229, 97, 60 },
		{ 229, 98, 60 },
		{ 229, 99, 60 },
		{ 229, 100, 60 },
		{ 229, 101, 60 },
		{ 229, 102, 60 },
		{ 229, 103, 60 },
		{ 229, 104, 60 },
		{ 229, 105, 60 },
		{ 229, 106, 60 },
		{ 229, 107, 60 },
		{ 229, 108, 60 },
		{ 229, 109, 60 },
		{ 229, 110, 60 },
		{ 229, 111, 60 },
		{ 229, 112, 60 },
		{ 229, 113, 60 },
		{ 229, 114, 60 },
		{ 229, 115, 60 },
		{ 229, 116, 60 },
		{ 229, 117, 60 },
		{ 229, 118, 60 },
		{ 229, 119, 60 },
		{ 229, 120, 60 },
		{ 229, 121, 60 },
		{ 229, 122, 60 },
		{ 229, 123, 60 },
		{ 229, 124, 60 },
		{ 229, 125, 60 },
		{ 229, 126, 60 },
		{ 229, 127, 60 },
		{ 229, 128, 60 },
		{ 229, 129, 60 },
		{ 229, 130, 60 },
		{ 229, 131, 60 },
		{ 229, 132, 60 },
		{ 229, 133, 60 },
		{ 229, 134, 60 },
		{ 229, 135, 60 },
		{ 229, 136, 60 },
		{ 229, 137, 60 },
		{ 229, 138, 60 },
		{ 229, 139, 60 },
		{ 229, 140, 60 },
		{ 229, 141, 60 },
		{ 229, 142, 60 },
		{ 229, 143, 60 },
		{ 229, 144, 60 },
		{ 229, 145, 60 },
		{ 229, 146, 60 },
		{ 229, 147, 60 },
		{ 229, 148, 60 },
		{ 229, 149, 60 },
		{ 229, 150, 60 },
		{ 229, 151, 60 },
		{ 229, 152, 60 },
		{ 229, 153, 60 },
		{ 229, 154, 60 },
		{ 229, 155, 60 },
		{ 229, 156, 60 },
		{ 229, 157, 60 },
		{ 229, 158, 60 },
		{ 229, 159, 60 },
		{ 229, 160, 60 },
		{ 229, 161, 60 },
		{ 229, 162, 60 },
		{ 229, 163, 60 },
		{ 229, 164, 60 },
		{ 229, 165, 60 },
		{ 229, 166, 60 },
		{ 229, 167, 60 },
		{ 229, 168, 60 },
		{ 229, 169, 60 },
		{ 229, 170, 60 },
		{ 229, 171, 60 },
		{ 229, 172, 60 },
		{ 229, 173, 60 },
		{ 229, 174, 60 },
		{ 229, 175, 60 },
		{ 229, 176, 60 },
		{ 229, 177, 60 },
		{ 229, 178, 60 },
		{ 229, 179, 60 },
		{ 229, 180, 60 },
		{ 229, 181, 60 },
		{ 229, 182, 60 },
		{ 230, 42, 60 },
		{ 230, 43, 60 },
		{ 230, 44, 60 },
		{ 230, 45, 60 },
		{ 230, 46, 60 },
		{ 230, 47, 60 },
		{ 230, 48, 60 },
		{ 230, 49, 60 },
		{ 230, 50, 60 },
		{ 230, 51, 60 },
		{ 230, 52, 60 },
		{ 230, 53, 60 },
		{ 230, 54, 60 },
		{ 230, 55, 60 },
		{ 230, 56, 60 },
		{ 230, 57, 60 },
		{ 230, 58, 60 },
		{ 230, 59, 60 },
		{ 230, 60, 60 },
		{ 230, 61, 60 },
		{ 230, 62, 60 },
		{ 230, 63, 60 },
		{ 230, 64, 60 },
		{ 230, 65, 60 },
		{ 230, 66, 60 },
		{ 230, 67, 60 },
		{ 230, 68, 60 },
		{ 230, 69, 60 },
		{ 230, 70, 60 },
		{ 230, 71, 60 },
		{ 230, 72, 60 },
		{ 230, 73, 60 },
		{ 230, 74, 60 },
		{ 230, 75, 60 },
		{ 230, 76, 60 },
		{ 230, 77, 60 },
		{ 230, 78, 60 },
		{ 230, 79, 60 },
		{ 230, 80, 60 },
		{ 230, 81, 60 },
		{ 230, 82, 60 },
		{ 230, 83, 60 },
		{ 230, 84, 60 },
		{ 230, 85, 60 },
		{ 230, 86, 60 },
		{ 230, 87, 60 },
		{ 230, 88, 60 },
		{ 230, 89, 60 },
		{ 230, 90, 60 },
		{ 230, 91, 60 },
		{ 230, 92, 60 },
		{ 230, 93, 60 },
		{ 230, 94, 60 },
		{ 230, 95, 60 },
		{ 230, 96, 60 },
		{ 230, 97, 60 },
		{ 230, 98, 60 },
		{ 230, 99, 60 },
		{ 230, 100, 60 },
		{ 230, 101, 60 },
		{ 230, 102, 60 },
		{ 230, 103, 60 },
		{ 230, 104, 60 },
		{ 230, 105, 60 },
		{ 230, 106, 60 },
		{ 230, 107, 60 },
		{ 230, 108, 60 },
		{ 230, 109, 60 },
		{ 230, 110, 60 },
		{ 230, 111, 60 },
		{ 230, 112, 60 },
		{ 230, 113, 60 },
		{ 230, 114, 60 },
		{ 230, 115, 60 },
		{ 230, 116, 60 },
		{ 230, 117, 60 },
		{ 230, 118, 60 },
		{ 230, 119, 60 },
		{ 230, 120, 60 },
		{ 230, 121, 60 },
		{ 230, 122, 60 },
		{ 230, 123, 60 },
		{ 230, 124, 60 },
		{ 230, 125, 60 },
		{ 230, 126, 60 },
		{ 230, 127, 60 },
		{ 230, 128, 60 },
		{ 230, 129, 60 },
		{ 230, 130, 60 },
		{ 230, 131, 60 },
		{ 230, 132, 60 },
		{ 230, 133, 60 },
		{ 230, 134, 60 },
		{ 230, 135, 60 },
		{ 230, 136, 60 },
		{ 230, 137, 60 },
		{ 230, 138, 60 },
		{ 230, 139, 60 },
		{ 230, 140, 60 },
		{ 230, 141, 60 },
		{ 230, 142, 60 },
		{ 230, 143, 60 },
		{ 230, 144, 60 },
		{ 230, 145, 60 },
		{ 230, 146, 60 },
		{ 230, 147, 60 },
		{ 230, 148, 60 },
		{ 230, 149, 60 },
		{ 230, 150, 60 },
		{ 230, 151, 60 },
		{ 230, 152, 60 },
		{ 230, 153, 60 },
		{ 230, 154, 60 },
		{ 230, 155, 60 },
		{ 230, 156, 60 },
		{ 230, 157, 60 },
		{ 230, 158, 60 },
		{ 230, 159, 60 },
		{ 230, 160, 60 },
		{ 230, 161, 60 },
		{ 230, 162, 60 },
		{ 230, 163, 60 },
		{ 230, 164, 60 },
		{ 230, 165, 60 },
		{ 230, 166, 60 },
		{ 230, 167, 60 },
		{ 230, 168, 60 },
		{ 230, 169, 60 },
		{ 230, 170, 60 },
		{ 230, 171, 60 },
		{ 230, 172, 60 },
		{ 230, 173, 60 },
		{ 230, 174, 60 },
		{ 230, 175, 60 },
		{ 230, 176, 60 },
		{ 230, 177, 60 },
		{ 230, 178, 60 },
		{ 230, 179, 60 },
		{ 230, 180, 60 },
		{ 230, 181, 60 },
		{ 230, 182, 60 },
		{ 230, 183, 60 },
		{ 230, 184, 60 },
		{ 231, 50, 60 },
		{ 231, 51, 60 },
		{ 231, 52, 60 },
		{ 231, 53, 60 },
		{ 231, 54, 60 },
		{ 231, 55, 60 },
		{ 231, 56, 60 },
		{ 231, 57, 60 },
		{ 231, 58, 60 },
		{ 231, 59, 60 },
		{ 231, 60, 60 },
		{ 231, 61, 60 },
		{ 231, 62, 60 },
		{ 231, 63, 60 },
		{ 231, 64, 60 },
		{ 231, 65, 60 },
		{ 231, 66, 60 },
		{ 231, 67, 60 },
		{ 231, 68, 60 },
		{ 231, 69, 60 },
		{ 231, 70, 60 },
		{ 231, 71, 60 },
		{ 231, 72, 60 },
		{ 231, 73, 60 },
		{ 231, 74, 60 },
		{ 231, 75, 60 },
		{ 231, 76, 60 },
		{ 231, 77, 60 },
		{ 231, 78, 60 },
		{ 231, 79, 60 },
		{ 231, 80, 60 },
		{ 231, 81, 60 },
		{ 231, 82, 60 },
		{ 231, 83, 60 },
		{ 231, 84, 60 },
		{ 231, 85, 60 },
		{ 231, 86, 60 },
		{ 231, 87, 60 },
		{ 231, 88, 60 },
		{ 231, 89, 60 },
		{ 231, 90, 60 },
		{ 231, 91, 60 },
		{ 231, 92, 60 },
		{ 231, 93, 60 },
		{ 231, 94, 60 },
		{ 231, 95, 60 },
		{ 231, 96, 60 },
		{ 231, 97, 60 },
		{ 231, 98, 60 },
		{ 231, 99, 60 },
		{ 231, 100, 60 },
		{ 231, 101, 60 },
		{ 231, 102, 60 },
		{ 231, 103, 60 },
		{ 231, 104, 60 },
		{ 231, 105, 60 },
		{ 231, 106, 60 },
		{ 231, 107, 60 },
		{ 231, 108, 60 },
		{ 231, 109, 60 },
		{ 231, 110, 60 },
		{ 231, 111, 60 },
		{ 231, 112, 60 },
		{ 231, 113, 60 },
		{ 231, 114, 60 },
		{ 231, 115, 60 },
		{ 231, 116, 60 },
		{ 231, 117, 60 },
		{ 231, 118, 60 },
		{ 231, 119, 60 },
		{ 231, 120, 60 },
		{ 231, 121, 60 },
		{ 231, 122, 60 },
		{ 231, 123, 60 },
		{ 231, 124, 60 },
		{ 231, 125, 60 },
		{ 231, 126, 60 },
		{ 231, 127, 60 },
		{ 231, 128, 60 },
		{ 231, 129, 60 },
		{ 231, 130, 60 },
		{ 231, 131, 60 },
		{ 231, 132, 60 },
		{ 231, 133, 60 },
		{ 231, 134, 60 },
		{ 231, 135, 60 },
		{ 231, 136, 60 },
		{ 231, 137, 60 },
		{ 231, 138, 60 },
		{ 231, 139, 60 },
		{ 231, 140, 60 },
		{ 231, 141, 60 },
		{ 231, 142, 60 },
		{ 231, 143, 60 },
		{ 231, 144, 60 },
		{ 231, 145, 60 },
		{ 231, 146, 60 },
		{ 231, 147, 60 },
		{ 231, 148, 60 },
		{ 231, 149, 60 },
		{ 231, 150, 60 },
		{ 231, 151, 60 },
		{ 231, 152, 60 },
		{ 231, 153, 60 },
		{ 231, 154, 60 },
		{ 231, 155, 60 },
		{ 231, 156, 60 },
		{ 231, 157, 60 },
		{ 231, 158, 60 },
		{ 231, 159, 60 },
		{ 231, 160, 60 },
		{ 231, 161, 60 },
		{ 231, 162, 60 },
		{ 231, 163, 60 },
		{ 231, 164, 60 },
		{ 231, 165, 60 },
		{ 231, 166, 60 },
		{ 231, 167, 60 },
		{ 231, 168, 60 },
		{ 231, 169, 60 },
		{ 231, 170, 60 },
		{ 231, 171, 60 },
		{ 231, 172, 60 },
		{ 231, 173, 60 },
		{ 231, 174, 60 },
		{ 231, 175, 60 },
		{ 231, 176, 60 },
		{ 231, 177, 60 },
		{ 231, 178, 60 },
		{ 231, 179, 60 },
		{ 231, 180, 60 },
		{ 231, 181, 60 },
		{ 231, 182, 60 },
		{ 231, 183, 60 },
		{ 231, 184, 60 },
		{ 231, 185, 60 },
		{ 231, 186, 60 },
		{ 231, 187, 60 },
		{ 232, 58, 60 },
		{ 232, 59, 60 },
		{ 232, 60, 60 },
		{ 232, 61, 60 },
		{ 232, 62, 60 },
		{ 232, 63, 60 },
		{ 232, 64, 60 },
		{ 232, 65, 60 },
		{ 232, 66, 60 },
		{ 232, 67, 60 },
		{ 232, 68, 60 },
		{ 232, 69, 60 },
		{ 232, 70, 60 },
		{ 232, 71, 60 },
		{ 232, 72, 60 },
		{ 232, 73, 60 },
		{ 232, 74, 60 },
		{ 232, 75, 60 },
		{ 232, 76, 60 },
		{ 232, 77, 60 },
		{ 232, 78, 60 },
		{ 232, 79, 60 },
		{ 232, 80, 60 },
		{ 232, 81, 60 },
		{ 232, 82, 60 },
		{ 232, 83, 60 },
		{ 232, 84, 60 },
		{ 232, 85, 60 },
		{ 232, 86, 60 },
		{ 232, 87, 60 },
		{ 232, 88, 60 },
		{ 232, 89, 60 },
		{ 232, 90, 60 },
		{ 232, 91, 60 },
		{ 232, 92, 60 },
		{ 232, 93, 60 },
		{ 232, 94, 60 },
		{ 232, 95, 60 },
		{ 232, 96, 60 },
		{ 232, 97, 60 },
		{ 232, 98, 60 },
		{ 232, 99, 60 },
		{ 232, 100, 60 },
		{ 232, 101, 60 },
		{ 232, 102, 60 },
		{ 232, 103, 60 },
		{ 232, 104, 60 },
		{ 232, 105, 60 },
		{ 232, 106, 60 },
		{ 232, 107, 60 },
		{ 232, 108, 60 },
		{ 232, 109, 60 },
		{ 232, 110, 60 },
		{ 232, 111, 60 },
		{ 232, 112, 60 },
		{ 232, 113, 60 },
		{ 232, 114, 60 },
		{ 232, 115, 60 },
		{ 232, 116, 60 },
		{ 232, 117, 60 },
		{ 232, 118, 60 },
		{ 232, 119, 60 },
		{ 232, 120, 60 },
		{ 232, 121, 60 },
		{ 232, 122, 60 },
		{ 232, 123, 60 },
		{ 232, 124, 60 },
		{ 232, 125, 60 },
		{ 232, 126, 60 },
		{ 232, 127, 60 },
		{ 232, 128, 60 },
		{ 232, 129, 60 },
		{ 232, 130, 60 },
		{ 232, 131, 60 },
		{ 232, 132, 60 },
		{ 232, 133, 60 },
		{ 232, 134, 60 },
		{ 232, 135, 60 },
		{ 232, 136, 60 },
		{ 232, 137, 60 },
		{ 232, 138, 60 },
		{ 232, 139, 60 },
		{ 232, 140, 60 },
		{ 232, 141, 60 },
		{ 232, 142, 60 },
		{ 232, 143, 60 },
		{ 232, 144, 60 },
		{ 232, 145, 60 },
		{ 232, 146, 60 },
		{ 232, 147, 60 },
		{ 232, 148, 60 },
		{ 232, 149, 60 },
		{ 232, 150, 60 },
		{ 232, 151, 60 },
		{ 232, 152, 60 },
		{ 232, 153, 60 },
		{ 232, 154, 60 },
		{ 232, 155, 60 },
		{ 232, 156, 60 },
		{ 232, 157, 60 },
		{ 232, 158, 60 },
		{ 232, 159, 60 },
		{ 232, 160, 60 },
		{ 232, 161, 60 },
		{ 232, 162, 60 },
		{ 232, 163, 60 },
		{ 232, 164, 60 },
		{ 232, 165, 60 },
		{ 232, 166, 60 },
		{ 232, 167, 60 },
		{ 232, 168, 60 },
		{ 232, 169, 60 },
		{ 232, 170, 60 },
		{ 232, 171, 60 },
		{ 232, 172, 60 },
		{ 232, 173, 60 },
		{ 232, 174, 60 },
		{ 232, 175, 60 },
		{ 232, 176, 60 },
		{ 232, 177, 60 },
		{ 232, 178, 60 },
		{ 232, 179, 60 },
		{ 232, 180, 60 },
		{ 232, 181, 60 },
		{ 232, 182, 60 },
		{ 232, 183, 60 },
		{ 232, 184, 60 },
		{ 232, 185, 60 },
		{ 232, 186, 60 },
		{ 232, 187, 60 },
		{ 232, 188, 60 },
		{ 232, 189, 60 },
		{ 233, 66, 60 },
		{ 233, 67, 60 },
		{ 233, 68, 60 },
		{ 233, 69, 60 },
		{ 233, 70, 60 },
		{ 233, 71, 60 },
		{ 233, 72, 60 },
		{ 233, 73, 60 },
		{ 233, 74, 60 },
		{ 233, 75, 60 },
		{ 233, 76, 60 },
		{ 233, 77, 60 },
		{ 233, 78, 60 },
		{ 233, 79, 60 },
		{ 233, 80, 60 },
		{ 233, 81, 60 },
		{ 233, 82, 60 },
		{ 233, 83, 60 },
		{ 233, 84, 60 },
		{ 233, 85, 60 },
		{ 233, 86, 60 },
		{ 233, 87, 60 },
		{ 233, 88, 60 },
		{ 233, 89, 60 },
		{ 233, 90, 60 },
		{ 233, 91, 60 },
		{ 233, 92, 60 },
		{ 233, 93, 60 },
		{ 233, 94, 60 },
		{ 233, 95, 60 },
		{ 233, 96, 60 },
		{ 233, 97, 60 },
		{ 233, 98, 60 },
		{ 233, 99, 60 },
		{ 233, 100, 60 },
		{ 233, 101, 60 },
		{ 233, 102, 60 },
		{ 233, 103, 60 },
		{ 233, 104, 60 },
		{ 233, 105, 60 },
		{ 233, 106, 60 },
		{ 233, 107, 60 },
		{ 233, 108, 60 },
		{ 233, 109, 60 },
		{ 233, 110, 60 },
		{ 233, 111, 60 },
		{ 233, 112, 60 },
		{ 233, 113, 60 },
		{ 233, 114, 60 },
		{ 233, 115, 60 },
		{ 233, 116, 60 },
		{ 233, 117, 60 },
		{ 233, 118, 60 },
		{ 233, 119, 60 },
		{ 233, 120, 60 },
		{ 233, 121, 60 },
		{ 233, 122, 60 },
		{ 233, 123, 60 },
		{ 233, 124, 60 },
		{ 233, 125, 60 },
		{ 233, 126, 60 },
		{ 233, 127, 60 },
		{ 233, 128, 60 },
		{ 233, 129, 60 },
		{ 233, 130, 60 },
		{ 233, 131, 60 },
		{ 233, 132, 60 },
		{ 233, 133, 60 },
		{ 233, 134, 60 },
		{ 233, 135, 60 },
		{ 233, 136, 60 },
		{ 233, 137, 60 },
		{ 233, 138, 60 },
		{ 233, 139, 60 },
		{ 233, 140, 60 },
		{ 233, 141, 60 },
		{ 233, 142, 60 },
		{ 233, 143, 60 },
		{ 233, 144, 60 },
		{ 233, 145, 60 },
		{ 233, 146, 60 },
		{ 233, 147, 60 },
		{ 233, 148, 60 },
		{ 233, 149, 60 },
		{ 233, 150, 60 },
		{ 233, 151, 60 },
		{ 233, 152, 60 },
		{ 233, 153, 60 },
		{ 233, 154, 60 },
		{ 233, 155, 60 },
		{ 233, 156, 60 },
		{ 233, 157, 60 },
		{ 233, 158, 60 },
		{ 233, 159, 60 },
		{ 233, 160, 60 },
		{ 233, 161, 60 },
		{ 233, 162, 60 },
		{ 233, 163, 60 },
		{ 233, 164, 60 },
		{ 233, 165, 60 },
		{ 233, 166, 60 },
		{ 233, 167, 60 },
		{ 233, 168, 60 },
		{ 233, 169, 60 },
		{ 233, 170, 60 },
		{ 233, 171, 60 },
		{ 233, 172, 60 },
		{ 233, 173, 60 },
		{ 233, 174, 60 },
		{ 233, 175, 60 },
		{ 233, 176, 60 },
		{ 233, 177, 60 },
		{ 233, 178, 60 },
		{ 233, 179, 60 },
		{ 233, 180, 60 },
		{ 233, 181, 60 },
		{ 233, 182, 60 },
		{ 233, 183, 60 },
		{ 233, 184, 60 },
		{ 233, 185, 60 },
		{ 233, 186, 60 },
		{ 233, 187, 60 },
		{ 233, 188, 60 },
		{ 233, 189, 60 },
		{ 233, 190, 60 },
		{ 234, 74, 60 },
		{ 234, 75, 60 },
		{ 234, 76, 60 },
		{ 234, 77, 60 },
		{ 234, 78, 60 },
		{ 234, 79, 60 },
		{ 234, 80, 60 },
		{ 234, 81, 60 },
		{ 234, 82, 60 },
		{ 234, 83, 60 },
		{ 234, 84, 60 },
		{ 234, 85, 60 },
		{ 234, 86, 60 },
		{ 234, 87, 60 },
		{ 234, 88, 60 },
		{ 234, 89, 60 },
		{ 234, 90, 60 },
		{ 234, 91, 60 },
		{ 234, 92, 60 },
		{ 234, 93, 60 },
		{ 234, 94, 60 },
		{ 234, 95, 60 },
		{ 234, 96, 60 },
		{ 234, 97, 60 },
		{ 234, 98, 60 },
		{ 234, 99, 60 },
		{ 234, 100, 60 },
		{ 234, 101, 60 },
		{ 234, 102, 60 },
		{ 234, 103, 60 },
		{ 234, 104, 60 },
		{ 234, 105, 60 },
		{ 234, 106, 60 },
		{ 234, 107, 60 },
		{ 234, 108, 60 },
		{ 234, 109, 60 },
		{ 234, 110, 60 },
		{ 234, 111, 60 },
		{ 234, 112, 60 },
		{ 234, 113, 60 },
		{ 234, 114, 60 },
		{ 234, 115, 60 },
		{ 234, 116, 60 },
		{ 234, 117, 60 },
		{ 234, 118, 60 },
		{ 234, 119, 60 },
		{ 234, 120, 60 },
		{ 234, 121, 60 },
		{ 234, 122, 60 },
		{ 234, 123, 60 },
		{ 234, 124, 60 },
		{ 234, 125, 60 },
		{ 234, 126, 60 },
		{ 234, 127, 60 },
		{ 234, 128, 60 },
		{ 234, 129, 60 },
		{ 234, 130, 60 },
		{ 234, 131, 60 },
		{ 234, 132, 60 },
		{ 234, 133, 60 },
		{ 234, 134, 60 },
		{ 234, 135, 60 },
		{ 234, 136, 60 },
		{ 234, 137, 60 },
		{ 234, 138, 60 },
		{ 234, 139, 60 },
		{ 234, 140, 60 },
		{ 234, 141, 60 },
		{ 234, 142, 60 },
		{ 234, 143, 60 },
		{ 234, 144, 60 },
		{ 234, 145, 60 },
		{ 234, 146, 60 },
		{ 234, 147, 60 },
		{ 234, 148, 60 },
		{ 234, 149, 60 },
		{ 234, 150, 60 },
		{ 234, 151, 60 },
		{ 234, 152, 60 },
		{ 234, 153, 60 },
		{ 234, 154, 60 },
		{ 234, 155, 60 },
		{ 234, 156, 60 },
		{ 234, 157, 60 },
		{ 234, 158, 60 },
		{ 234, 159, 60 },
		{ 234, 160, 60 },
		{ 234, 161, 60 },
		{ 234, 162, 60 },
		{ 234, 163, 60 },
		{ 234, 164, 60 },
		{ 234, 165, 60 },
		{ 234, 166, 60 },
		{ 234, 167, 60 },
		{ 234, 168, 60 },
		{ 234, 169, 60 },
		{ 234, 170, 60 },
		{ 234, 171, 60 },
		{ 234, 172, 60 },
		{ 234, 173, 60 },
		{ 234, 174, 60 },
		{ 234, 175, 60 },
		{ 234, 176, 60 },
		{ 234, 177, 60 },
		{ 234, 178, 60 },
		{ 234, 179, 60 },
		{ 234, 180, 60 },
		{ 234, 181, 60 },
		{ 234, 182, 60 },
		{ 234, 183, 60 },
		{ 234, 184, 60 },
		{ 234, 185, 60 },
		{ 234, 186, 60 },
		{ 234, 187, 60 },
		{ 234, 188, 60 },
		{ 234, 189, 60 },
		{ 234, 190, 60 },
		{ 235, 82, 60 },
		{ 235, 83, 60 },
		{ 235, 84, 60 },
		{ 235, 85, 60 },
		{ 235, 86, 60 },
		{ 235, 87, 60 },
		{ 235, 88, 60 },
		{ 235, 89, 60 },
		{ 235, 90, 60 },
		{ 235, 91, 60 },
		{ 235, 92, 60 },
		{ 235, 93, 60 },
		{ 235, 94, 60 },
		{ 235, 95, 60 },
		{ 235, 96, 60 },
		{ 235, 97, 60 },
		{ 235, 98, 60 },
		{ 235, 99, 60 },
		{ 235, 100, 60 },
		{ 235, 101, 60 },
		{ 235, 102, 60 },
		{ 235, 103, 60 },
		{ 235, 104, 60 },
		{ 235, 105, 60 },
		{ 235, 106, 60 },
		{ 235, 107, 60 },
		{ 235, 108, 60 },
		{ 235, 109, 60 },
		{ 235, 110, 60 },
		{ 235, 111, 60 },
		{ 235, 112, 60 },
		{ 235, 113, 60 },
		{ 235, 114, 60 },
		{ 235, 115, 60 },
		{ 235, 116, 60 },
		{ 235, 117, 60 },
		{ 235, 118, 60 },
		{ 235, 119, 60 },
		{ 235, 120, 60 },
		{ 235, 121, 60 },
		{ 235, 122, 60 },
		{ 235, 123, 60 },
		{ 235, 124, 60 },
		{ 235, 125, 60 },
		{ 235, 126, 60 },
		{ 235, 127, 60 },
		{ 235, 128, 60 },
		{ 235, 129, 60 },
		{ 235, 130, 60 },
		{ 235, 131, 60 },
		{ 235, 132, 60 },
		{ 235, 133, 60 },
		{ 235, 134, 60 },
		{ 235, 135, 60 },
		{ 235, 136, 60 },
		{ 235, 137, 60 },
		{ 235, 138, 60 },
		{ 235, 139, 60 },
		{ 235, 140, 60 },
		{ 235, 141, 60 },
		{ 235, 142, 60 },
		{ 235, 143, 60 },
		{ 235, 144, 60 },
		{ 235, 145, 60 },
		{ 235, 146, 60 },
		{ 235, 147, 60 },
		{ 235, 148, 60 },
		{ 235, 149, 60 },
		{ 235, 150, 60 },
		{ 235, 151, 60 },
		{ 235, 152, 60 },
		{ 235, 153, 60 },
		{ 235, 154, 60 },
		{ 235, 155, 60 },
		{ 235, 156, 60 },
		{ 235, 157, 60 },
		{ 235, 158, 60 },
		{ 235, 159, 60 },
		{ 235, 160, 60 },
		{ 235, 161, 60 },
		{ 235, 162, 60 },
		{ 235, 163, 60 },
		{ 235, 164, 60 },
		{ 235, 165, 60 },
		{ 235, 166, 60 },
		{ 235, 167, 60 },
		{ 235, 168, 60 },
		{ 235, 169, 60 },
		{ 235, 170, 60 },
		{ 235, 171, 60 },
		{ 235, 172, 60 },
		{ 235, 173, 60 },
		{ 235, 174, 60 },
		{ 235, 175, 60 },
		{ 235, 176, 60 },
		{ 235, 177, 60 },
		{ 235, 178, 60 },
		{ 235, 179, 60 },
		{ 235, 180, 60 },
		{ 235, 181, 60 },
		{ 235, 182, 60 },
		{ 235, 183, 60 },
		{ 235, 184, 60 },
		{ 235, 185, 60 },
		{ 235, 186, 60 },
		{ 235, 187, 60 },
		{ 235, 188, 60 },
		{ 235, 189, 60 },
		{ 235, 190, 60 },
		{ 235, 191, 60 },
		{ 236, 90, 60 },
		{ 236, 91, 60 },
		{ 236, 92, 60 },
		{ 236, 93, 60 },
		{ 236, 94, 60 },
		{ 236, 95, 60 },
		{ 236, 96, 60 },
		{ 236, 97, 60 },
		{ 236, 98, 60 },
		{ 236, 99, 60 },
		{ 236, 100, 60 },
		{ 236, 101, 60 },
		{ 236, 102, 60 },
		{ 236, 103, 60 },
		{ 236, 104, 60 },
		{ 236, 105, 60 },
		{ 236, 106, 60 },
		{ 236, 107, 60 },
		{ 236, 108, 60 },
		{ 236, 109, 60 },
		{ 236, 110, 60 },
		{ 236, 111, 60 },
		{ 236, 112, 60 },
		{ 236, 113, 60 },
		{ 236, 114, 60 },
		{ 236, 115, 60 },
		{ 236, 116, 60 },
		{ 236, 117, 60 },
		{ 236, 118, 60 },
		{ 236, 119, 60 },
		{ 236, 120, 60 },
		{ 236, 121, 60 },
		{ 236, 122, 60 },
		{ 236, 123, 60 },
		{ 236, 124, 60 },
		{ 236, 125, 60 },
		{ 236, 126, 60 },
		{ 236, 127, 60 },
		{ 236, 128, 60 },
		{ 236, 129, 60 },
		{ 236, 130, 60 },
		{ 236, 131, 60 },
		{ 236, 132, 60 },
		{ 236, 133, 60 },
		{ 236, 134, 60 },
		{ 236, 135, 60 },
		{ 236, 136, 60 },
		{ 236, 137, 60 },
		{ 236, 138, 60 },
		{ 236, 139, 60 },
		{ 236, 140, 60 },
		{ 236, 141, 60 },
		{ 236, 142, 60 },
		{ 236, 143, 60 },
		{ 236, 144, 60 },
		{ 236, 145, 60 },
		{ 236, 146, 60 },
		{ 236, 147, 60 },
		{ 236, 148, 60 },
		{ 236, 149, 60 },
		{ 236, 150, 60 },
		{ 236, 151, 60 },
		{ 236, 152, 60 },
		{ 236, 153, 60 },
		{ 236, 154, 60 },
		{ 236, 155, 60 },
		{ 236, 156, 60 },
		{ 236, 157, 60 },
		{ 236, 158, 60 },
		{ 236, 159, 60 },
		{ 236, 160, 60 },
		{ 236, 161, 60 },
		{ 236, 162, 60 },
		{ 236, 163, 60 },
		{ 236, 164, 60 },
		{ 236, 165, 60 },
		{ 236, 166, 60 },
		{ 236, 167, 60 },
		{ 236, 168, 60 },
		{ 236, 169, 60 },
		{ 236, 170, 60 },
		{ 236, 171, 60 },
		{ 236, 172, 60 },
		{ 236, 173, 60 },
		{ 236, 174, 60 },
		{ 236, 175, 60 },
		{ 236, 176, 60 },
		{ 236, 177, 60 },
		{ 236, 178, 60 },
		{ 236, 179, 60 },
		{ 236, 180, 60 },
		{ 236, 181, 60 },
		{ 236, 182, 60 },
		{ 236, 183, 60 },
		{ 236, 184, 60 },
		{ 236, 185, 60 },
		{ 236, 186, 60 },
		{ 236, 187, 60 },
		{ 236, 188, 60 },
		{ 236, 189, 60 },
		{ 236, 190, 60 },
		{ 236, 191, 60 },
		{ 237, 98, 60 },
		{ 237, 99, 60 },
		{ 237, 100, 60 },
		{ 237, 101, 60 },
		{ 237, 102, 60 },
		{ 237, 103, 60 },
		{ 237, 104, 60 },
		{ 237, 105, 60 },
		{ 237, 106, 60 },
		{ 237, 107, 60 },
		{ 237, 108, 60 },
		{ 237, 109, 60 },
		{ 237, 110, 60 },
		{ 237, 111, 60 },
		{ 237, 112, 60 },
		{ 237, 113, 60 },
		{ 237, 114, 60 },
		{ 237, 115, 60 },
		{ 237, 116, 60 },
		{ 237, 117, 60 },
		{ 237, 118, 60 },
		{ 237, 119, 60 },
		{ 237, 120, 60 },
		{ 237, 121, 60 },
		{ 237, 122, 60 },
		{ 237, 123, 60 },
		{ 237, 124, 60 },
		{ 237, 125, 60 },
		{ 237, 126, 60 },
		{ 237, 127, 60 },
		{ 237, 128, 60 },
		{ 237, 129, 60 },
		{ 237, 130, 60 },
		{ 237, 131, 60 },
		{ 237, 132, 60 },
		{ 237, 133, 60 },
		{ 237, 134, 60 },
		{ 237, 135, 60 },
		{ 237, 136, 60 },
		{ 237, 137, 60 },
		{ 237, 138, 60 },
		{ 237, 139, 60 },
		{ 237, 140, 60 },
		{ 237, 141, 60 },
		{ 237, 142, 60 },
		{ 237, 143, 60 },
		{ 237, 144, 60 },
		{ 237, 145, 60 },
		{ 237, 146, 60 },
		{ 237, 147, 60 },
		{ 237, 148, 60 },
		{ 237, 149, 60 },
		{ 237, 150, 60 },
		{ 237, 151, 60 },
		{ 237, 152, 60 },
		{ 237, 153, 60 },
		{ 237, 154, 60 },
		{ 237, 155, 60 },
		{ 237, 156, 60 },
		{ 237, 157, 60 },
		{ 237, 158, 60 },
		{ 237, 159, 60 },
		{ 237, 160, 60 },
		{ 237, 161, 60 },
		{ 237, 162, 60 },
		{ 237, 163, 60 },
		{ 237, 164, 60 },
		{ 237, 165, 60 },
		{ 237, 166, 60 },
		{ 237, 167, 60 },
		{ 237, 168, 60 },
		{ 237, 169, 60 },
		{ 237, 170, 60 },
		{ 237, 171, 60 },
		{ 237, 172, 60 },
		{ 237, 173, 60 },
		{ 237, 174, 60 },
		{ 237, 175, 60 },
		{ 237, 176, 60 },
		{ 237, 177, 60 },
		{ 237, 178, 60 },
		{ 237, 179, 60 },
		{ 237, 180, 60 },
		{ 237, 181, 60 },
		{ 237, 182, 60 },
		{ 237, 183, 60 },
		{ 237, 184, 60 },
		{ 237, 185, 60 },
		{ 237, 186, 60 },
		{ 237, 187, 60 },
		{ 237, 188, 60 },
		{ 237, 189, 60 },
		{ 237, 190, 60 },
		{ 237, 191, 60 },
		{ 238, 106, 60 },
		{ 238, 107, 60 },
		{ 238, 108, 60 },
		{ 238, 109, 60 },
		{ 238, 110, 60 },
		{ 238, 111, 60 },
		{ 238, 112, 60 },
		{ 238, 113, 60 },
		{ 238, 114, 60 },
		{ 238, 115, 60 },
		{ 238, 116, 60 },
		{ 238, 117, 60 },
		{ 238, 118, 60 },
		{ 238, 119, 60 },
		{ 238, 120, 60 },
		{ 238, 121, 60 },
		{ 238, 122, 60 },
		{ 238, 123, 60 },
		{ 238, 124, 60 },
		{ 238, 125, 60 },
		{ 238, 126, 60 },
		{ 238, 127, 60 },
		{ 238, 128, 60 },
		{ 238, 129, 60 },
		{ 238, 130, 60 },
		{ 238, 131, 60 },
		{ 238, 132, 60 },
		{ 238, 133, 60 },
		{ 238, 134, 60 },
		{ 238, 135, 60 },
		{ 238, 136, 60 },
		{ 238, 137, 60 },
		{ 238, 138, 60 },
		{ 238, 139, 60 },
		{ 238, 140, 60 },
		{ 238, 141, 60 },
		{ 238, 142, 60 },
		{ 238, 143, 60 },
		{ 238, 144, 60 },
		{ 238, 145, 60 },
		{ 238, 146, 60 },
		{ 238, 147, 60 },
		{ 238, 148, 60 },
		{ 238, 149, 60 },
		{ 238, 150, 60 },
		{ 238, 151, 60 },
		{ 238, 152, 60 },
		{ 238, 153, 60 },
		{ 238, 154, 60 },
		{ 238, 155, 60 },
		{ 238, 156, 60 },
		{ 238, 157, 60 },
		{ 238, 158, 60 },
		{ 238, 159, 60 },
		{ 238, 160, 60 },
		{ 238, 161, 60 },
		{ 238, 162, 60 },
		{ 238, 163, 60 },
		{ 238, 164, 60 },
		{ 238, 165, 60 },
		{ 238, 166, 60 },
		{ 238, 167, 60 },
		{ 238, 168, 60 },
		{ 238, 169, 60 },
		{ 238, 170, 60 },
		{ 238, 171, 60 },
		{ 238, 172, 60 },
		{ 238, 173, 60 },
		{ 238, 174, 60 },
		{ 238, 175, 60 },
		{ 238, 176, 60 },
		{ 238, 177, 60 },
		{ 238, 178, 60 },
		{ 238, 179, 60 },
		{ 238, 180, 60 },
		{ 238, 181, 60 },
		{ 238, 182, 60 },
		{ 238, 183, 60 },
		{ 238, 184, 60 },
		{ 238, 185, 60 },
		{ 238, 186, 60 },
		{ 238, 187, 60 },
		{ 238, 188, 60 },
		{ 238, 189, 60 },
		{ 238, 190, 60 },
		{ 238, 191, 60 },
		{ 239, 114, 60 },
		{ 239, 115, 60 },
		{ 239, 116, 60 },
		{ 239, 117, 60 },
		{ 239, 118, 60 },
		{ 239, 119, 60 },
		{ 239, 120, 60 },
		{ 239, 121, 60 },
		{ 239, 122, 60 },
		{ 239, 123, 60 },
		{ 239, 124, 60 },
		{ 239, 125, 60 },
		{ 239, 126, 60 },
		{ 239, 127, 60 },
		{ 239, 128, 60 },
		{ 239, 129, 60 },
		{ 239, 130, 60 },
		{ 239, 131, 60 },
		{ 239, 132, 60 },
		{ 239, 133, 60 },
		{ 239, 134, 60 },
		{ 239, 135, 60 },
		{ 239, 136, 60 },
		{ 239, 137, 60 },
		{ 239, 138, 60 },
		{ 239, 139, 60 },
		{ 239, 140, 60 },
		{ 239, 141, 60 },
		{ 239, 142, 60 },
		{ 239, 143, 60 },
		{ 239, 144, 60 },
		{ 239, 145, 60 },
		{ 239, 146, 60 },
		{ 239, 147, 60 },
		{ 239, 148, 60 },
		{ 239, 149, 60 },
		{ 239, 150, 60 },
		{ 239, 151, 60 },
		{ 239, 152, 60 },
		{ 239, 153, 60 },
		{ 239, 154, 60 },
		{ 239, 155, 60 },
		{ 239, 156, 60 },
		{ 239, 157, 60 },
		{ 239, 158, 60 },
		{ 239, 159, 60 },
		{ 239, 160, 60 },
		{ 239, 161, 60 },
		{ 239, 162, 60 },
		{ 239, 163, 60 },
		{ 239, 164, 60 },
		{ 239, 165, 60 },
		{ 239, 166, 60 },
		{ 239, 167, 60 },
		{ 239, 168, 60 },
		{ 239, 169, 60 },
		{ 239, 170, 60 },
		{ 239, 171, 60 },
		{ 239, 172, 60 },
		{ 239, 173, 60 },
		{ 239, 174, 60 },
		{ 239, 175, 60 },
		{ 239, 176, 60 },
		{ 239, 177, 60 },
		{ 239, 178, 60 },
		{ 239, 179, 60 },
		{ 239, 180, 60 },
		{ 239, 181, 60 },
		{ 239, 182, 60 },
		{ 239, 183, 60 },
		{ 239, 184, 60 },
		{ 239, 185, 60 },
		{ 239, 186, 60 },
		{ 239, 187, 60 },
		{ 239, 188, 60 },
		{ 239, 189, 60 },
		{ 239, 190, 60 },
		{ 239, 191, 60 },
		{ 240, 122, 60 },
		{ 240, 123, 60 },
		{ 240, 124, 60 },
		{ 240, 125, 60 },
		{ 240, 126, 60 },
		{ 240, 127, 60 },
		{ 240, 128, 60 },
		{ 240, 129, 60 },
		{ 240, 130, 60 },
		{ 240, 131, 60 },
		{ 240, 132, 60 },
		{ 240, 133, 60 },
		{ 240, 134, 60 },
		{ 240, 135, 60 },
		{ 240, 136, 60 },
		{ 240, 137, 60 },
		{ 240, 138, 60 },
		{ 240, 139, 60 },
		{ 240, 140, 60 },
		{ 240, 141, 60 },
		{ 240, 142, 60 },
		{ 240, 143, 60 },
		{ 240, 144, 60 },
		{ 240, 145, 60 },
		{ 240, 146, 60 },
		{ 240, 147, 60 },
		{ 240, 148, 60 },
		{ 240, 149, 60 },
		{ 240, 150, 60 },
		{ 240, 151, 60 },
		{ 240, 152, 60 },
		{ 240, 153, 60 },
		{ 240, 154, 60 },
		{ 240, 155, 60 },
		{ 240, 156, 60 },
		{ 240, 157, 60 },
		{ 240, 158, 60 },
		{ 240, 159, 60 },
		{ 240, 160, 60 },
		{ 240, 161, 60 },
		{ 240, 162, 60 },
		{ 240, 163, 60 },
		{ 240, 164, 60 },
		{ 240, 165, 60 },
		{ 240, 166, 60 },
		{ 240, 167, 60 },
		{ 240, 168, 60 },
		{ 240, 169, 60 },
		{ 240, 170, 60 },
		{ 240, 171, 60 },
		{ 240, 172, 60 },
		{ 240, 173, 60 },
		{ 240, 174, 60 },
		{ 240, 175, 60 },
		{ 240, 176, 60 },
		{ 240, 177, 60 },
		{ 240, 178, 60 },
		{ 240, 179, 60 },
		{ 240, 180, 60 },
		{ 240, 181, 60 },
		{ 240, 182, 60 },
		{ 240, 183, 60 },
		{ 240, 184, 60 },
		{ 240, 185, 60 },
		{ 240, 186, 60 },
		{ 240, 187, 60 },
		{ 240, 188, 60 },
		{ 240, 189, 60 },
		{ 240, 190, 60 },
		{ 240, 191, 60 },
		{ 241, 130, 60 },
		{ 241, 131, 60 },
		{ 241, 132, 60 },
		{ 241, 133, 60 },
		{ 241, 134, 60 },
		{ 241, 135, 60 },
		{ 241, 136, 60 },
		{ 241, 137, 60 },
		{ 241, 138, 60 },
		{ 241, 139, 60 },
		{ 241, 140, 60 },
		{ 241, 141, 60 },
		{ 241, 142, 60 },
		{ 241, 143, 60 },
		{ 241, 144, 60 },
		{ 241, 145, 60 },
		{ 241, 146, 60 },
		{ 241, 147, 60 },
		{ 241, 148, 60 },
		{ 241, 149, 60 },
		{ 241, 150, 60 },
		{ 241, 151, 60 },
		{ 241, 152, 60 },
		{ 241, 153, 60 },
		{ 241, 154, 60 },
		{ 241, 155, 60 },
		{ 241, 156, 60 },
		{ 241, 157, 60 },
		{ 241, 158, 60 },
		{ 241, 159, 60 },
		{ 241, 160, 60 },
		{ 241, 161, 60 },
		{ 241, 162, 60 },
		{ 241, 163, 60 },
		{ 241, 164, 60 },
		{ 241, 165, 60 },
		{ 241, 166, 60 },
		{ 241, 167, 60 },
		{ 241, 168, 60 },
		{ 241, 169, 60 },
		{ 241, 170, 60 },
		{ 241, 171, 60 },
		{ 241, 172, 60 },
		{ 241, 173, 60 },
		{ 241, 174, 60 },
		{ 241, 175, 60 },
		{ 241, 176, 60 },
		{ 241, 177, 60 },
		{ 241, 178, 60 },
		{ 241, 179, 60 },
		{ 241, 180, 60 },
		{ 241, 181, 60 },
		{ 241, 182, 60 },
		{ 241, 183, 60 },
		{ 241, 184, 60 },
		{ 241, 185, 60 },
		{ 241, 186, 60 },
		{ 241, 187, 60 },
		{ 241, 188, 60 },
		{ 241, 189, 60 },
		{ 241, 190, 60 },
		{ 241, 191, 60 },
		{ 242, 138, 60 },
		{ 242, 139, 60 },
		{ 242, 140, 60 },
		{ 242, 141, 60 },
		{ 242, 142, 60 },
		{ 242, 143, 60 },
		{ 242, 144, 60 },
		{ 242, 145, 60 },
		{ 242, 146, 60 },
		{ 242, 147, 60 },
		{ 242, 148, 60 },
		{ 242, 149, 60 },
		{ 242, 150, 60 },
		{ 242, 151, 60 },
		{ 242, 152, 60 },
		{ 242, 153, 60 },
		{ 242, 154, 60 },
		{ 242, 155, 60 },
		{ 242, 156, 60 },
		{ 242, 157, 60 },
		{ 242, 158, 60 },
		{ 242, 159, 60 },
		{ 242, 160, 60 },
		{ 242, 161, 60 },
		{ 242, 162, 60 },
		{ 242, 163, 60 },
		{ 242, 164, 60 },
		{ 242, 165, 60 },
		{ 242, 166, 60 },
		{ 242, 167, 60 },
		{ 242, 168, 60 },
		{ 242, 169, 60 },
		{ 242, 170, 60 },
		{ 242, 171, 60 },
		{ 242, 172, 60 },
		{ 242, 173, 60 },
		{ 242, 174, 60 },
		{ 242, 175, 60 },
		{ 242, 176, 60 },
		{ 242, 177, 60 },
		{ 242, 178, 60 },
		{ 242, 179, 60 },
		{ 242, 180, 60 },
		{ 242, 181, 60 },
		{ 242, 182, 60 },
		{ 242, 183, 60 },
		{ 242, 184, 60 },
		{ 242, 185, 60 },
		{ 242, 186, 60 },
		{ 242, 187, 60 },
		{ 242, 188, 60 },
		{ 242, 189, 60 },
		{ 242, 190, 60 },
		{ 242, 191, 60 },
		{ 243, 146, 60 },
		{ 243, 147, 60 },
		{ 243, 148, 60 },
		{ 243, 149, 60 },
		{ 243, 150, 60 },
		{ 243, 151, 60 },
		{ 243, 152, 60 },
		{ 243, 153, 60 },
		{ 243, 154, 60 },
		{ 243, 155, 60 },
		{ 243, 156, 60 },
		{ 243, 157, 60 },
		{ 243, 158, 60 },
		{ 243, 159, 60 },
		{ 243, 160, 60 },
		{ 243, 161, 60 },
		{ 243, 162, 60 },
		{ 243, 163, 60 },
		{ 243, 164, 60 },
		{ 243, 165, 60 },
		{ 243, 166, 60 },
		{ 243, 167, 60 },
		{ 243, 168, 60 },
		{ 243, 169, 60 },
		{ 243, 170, 60 },
		{ 243, 171, 60 },
		{ 243, 172, 60 },
		{ 243, 173, 60 },
		{ 243, 174, 60 },
		{ 243, 175, 60 },
		{ 243, 176, 60 },
		{ 243, 177, 60 },
		{ 243, 178, 60 },
		{ 243, 179, 60 },
		{ 243, 180, 60 },
		{ 243, 181, 60 },
		{ 243, 182, 60 },
		{ 243, 183, 60 },
		{ 243, 184, 60 },
		{ 243, 185, 60 },
		{ 243, 186, 60 },
		{ 243, 187, 60 },
		{ 243, 188, 60 },
		{ 243, 189, 60 },
		{ 243, 190, 60 },
		{ 243, 191, 60 },
		{ 244, 154, 60 },
		{ 244, 155, 60 },
		{ 244, 156, 60 },
		{ 244, 157, 60 },
		{ 244, 158, 60 },
		{ 244, 159, 60 },
		{ 244, 160, 60 },
		{ 244, 161, 60 },
		{ 244, 162, 60 },
		{ 244, 163, 60 },
		{ 244, 164, 60 },
		{ 244, 165, 60 },
		{ 244, 166, 60 },
		{ 244, 167, 60 },
		{ 244, 168, 60 },
		{ 244, 169, 60 },
		{ 244, 170, 60 },
		{ 244, 171, 60 },
		{ 244, 172, 60 },
		{ 244, 173, 60 },
		{ 244, 174, 60 },
		{ 244, 175, 60 },
		{ 244, 176, 60 },
		{ 244, 177, 60 },
		{ 244, 178, 60 },
		{ 244, 179, 60 },
		{ 244, 180, 60 },
		{ 244, 181, 60 },
		{ 244, 182, 60 },
		{ 244, 183, 60 },
		{ 244, 184, 60 },
		{ 244, 185, 60 },
		{ 244, 186, 60 },
		{ 244, 187, 60 },
		{ 244, 188, 60 },
		{ 244, 189, 60 },
		{ 244, 190, 60 },
		{ 244, 191, 60 },
		{ 245, 162, 60 },
		{ 245, 163, 60 },
		{ 245, 164, 60 },
		{ 245, 165, 60 },
		{ 245, 166, 60 },
		{ 245, 167, 60 },
		{ 245, 168, 60 },
		{ 245, 169, 60 },
		{ 245, 170, 60 },
		{ 245, 171, 60 },
		{ 245, 172, 60 },
		{ 245, 173, 60 },
		{ 245, 174, 60 },
		{ 245, 175, 60 },
		{ 245, 176, 60 },
		{ 245, 177, 60 },
		{ 245, 178, 60 },
		{ 245, 179, 60 },
		{ 245, 180, 60 },
		{ 245, 181, 60 },
		{ 245, 182, 60 },
		{ 245, 183, 60 },
		{ 245, 184, 60 },
		{ 245, 185, 60 },
		{ 245, 186, 60 },
		{ 245, 187, 60 },
		{ 245, 188, 60 },
		{ 245, 189, 60 },
		{ 245, 190, 60 },
		{ 245, 191, 60 },
		{ 246, 170, 60 },
		{ 246, 171, 60 },
		{ 246, 172, 60 },
		{ 246, 173, 60 },
		{ 246, 174, 60 },
		{ 246, 175, 60 },
		{ 246, 176, 60 },
		{ 246, 177, 60 },
		{ 246, 178, 60 },
		{ 246, 179, 60 },
		{ 246, 180, 60 },
		{ 246, 181, 60 },
		{ 246, 182, 60 },
		{ 246, 183, 60 },
		{ 246, 184, 60 },
		{ 246, 185, 60 },
		{ 246, 186, 60 },
		{ 246, 187, 60 },
		{ 246, 188, 60 },
		{ 246, 189, 60 },
		{ 246, 190, 60 },
		{ 246, 191, 60 },
		{ 247, 178, 60 },
		{ 247, 179, 60 },
		{ 247, 180, 60 },
		{ 247, 181, 60 },
		{ 247, 182, 60 },
		{ 247, 183, 60 },
		{ 247, 184, 60 },
		{ 247, 185, 60 },
		{ 247, 186, 60 },
		{ 247, 187, 60 },
		{ 247, 188, 60 },
		{ 247, 189, 60 },
		{ 247, 190, 60 },
		{ 248, 187, 60 },
		{ 248, 188, 60 },
		{ 248, 189, 60 },
		{ 266, 325, 60 },
		{ 266, 326, 60 },
		{ 266, 327, 60 },
		{ 267, 324, 60 },
		{ 267, 325, 60 },
		{ 267, 326, 60 },
		{ 267, 327, 60 },
		{ 267, 328, 60 },
		{ 267, 329, 60 },
		{ 267, 330, 60 },
		{ 267, 331, 60 },
		{ 267, 332, 60 },
		{ 267, 333, 60 },
		{ 267, 334, 60 },
		{ 267, 335, 60 },
		{ 267, 336, 60 },
		{ 268, 323, 60 },
		{ 268, 324, 60 },
		{ 268, 325, 60 },
		{ 268, 326, 60 },
		{ 268, 327, 60 },
		{ 268, 328, 60 },
		{ 268, 329, 60 },
		{ 268, 330, 60 },
		{ 268, 331, 60 },
		{ 268, 332, 60 },
		{ 268, 333, 60 },
		{ 268, 334, 60 },
		{ 268, 335, 60 },
		{ 268, 336, 60 },
		{ 268, 337, 60 },
		{ 268, 338, 60 },
		{ 268, 339, 60 },
		{ 268, 340, 60 },
		{ 268, 341, 60 },
		{ 268, 342, 60 },
		{ 268, 343, 60 },
		{ 268, 344, 60 },
		{ 269, 323, 60 },
		{ 269, 324, 60 },
		{ 269, 325, 60 },
		{ 269, 326, 60 },
		{ 269, 327, 60 },
		{ 269, 328, 60 },
		{ 269, 329, 60 },
		{ 269, 330, 60 },
		{ 269, 331, 60 },
		{ 269, 332, 60 },
		{ 269, 333, 60 },
		{ 269, 334, 60 },
		{ 269, 335, 60 },
		{ 269, 336, 60 },
		{ 269, 337, 60 },
		{ 269, 338, 60 },
		{ 269, 339, 60 },
		{ 269, 340, 60 },
		{ 269, 341, 60 },
		{ 269, 342, 60 },
		{ 269, 343, 60 },
		{ 269, 344, 60 },
		{ 269, 345, 60 },
		{ 269, 346, 60 },
		{ 269, 347, 60 },
		{ 269, 348, 60 },
		{ 269, 349, 60 },
		{ 269, 350, 60 },
		{ 269, 351, 60 },
		{ 269, 352, 60 },
		{ 270, 323, 60 },
		{ 270, 324, 60 },
		{ 270, 325, 60 },
		{ 270, 326, 60 },
		{ 270, 327, 60 },
		{ 270, 328, 60 },
		{ 270, 329, 60 },
		{ 270, 330, 60 },
		{ 270, 331, 60 },
		{ 270, 332, 60 },
		{ 270, 333, 60 },
		{ 270, 334, 60 },
		{ 270, 335, 60 },
		{ 270, 336, 60 },
		{ 270, 337, 60 },
		{ 270, 338, 60 },
		{ 270, 339, 60 },
		{ 270, 340, 60 },
		{ 270, 341, 60 },
		{ 270, 342, 60 },
		{ 270, 343, 60 },
		{ 270, 344, 60 },
		{ 270, 345, 60 },
		{ 270, 346, 60 },
		{ 270, 347, 60 },
		{ 270, 348, 60 },
		{ 270, 349, 60 },
		{ 270, 350, 60 },
		{ 270, 351, 60 },
		{ 270, 352, 60 },
		{ 270, 353, 60 },
		{ 270, 354, 60 },
		{ 270, 355, 60 },
		{ 270, 356, 60 },
		{ 270, 357, 60 },
		{ 270, 358, 60 },
		{ 270, 359, 60 },
		{ 270, 360, 60 },
		{ 271, 323, 60 },
		{ 271, 324, 60 },
		{ 271, 325, 60 },
		{ 271, 326, 60 },
		{ 271, 327, 60 },
		{ 271, 328, 60 },
		{ 271, 329, 60 },
		{ 271, 330, 60 },
		{ 271, 331, 60 },
		{ 271, 332, 60 },
		{ 271, 333, 60 },
		{ 271, 334, 60 },
		{ 271, 335, 60 },
		{ 271, 336, 60 },
		{ 271, 337, 60 },
		{ 271, 338, 60 },
		{ 271, 339, 60 },
		{ 271, 340, 60 },
		{ 271, 341, 60 },
		{ 271, 342, 60 },
		{ 271, 343, 60 },
		{ 271, 344, 60 },
		{ 271, 345, 60 },
		{ 271, 346, 60 },
		{ 271, 347, 60 },
		{ 271, 348, 60 },
		{ 271, 349, 60 },
		{ 271, 350, 60 },
		{ 271, 351, 60 },
		{ 271, 352, 60 },
		{ 271, 353, 60 },
		{ 271, 354, 60 },
		{ 271, 355, 60 },
		{ 271, 356, 60 },
		{ 271, 357, 60 },
		{ 271, 358, 60 },
		{ 271, 359, 60 },
		{ 271, 360, 60 },
		{ 271, 361, 60 },
		{ 271, 362, 60 },
		{ 271, 363, 60 },
		{ 271, 364, 60 },
		{ 271, 365, 60 },
		{ 271, 366, 60 },
		{ 271, 367, 60 },
		{ 271, 368, 60 },
		{ 272, 323, 60 },
		{ 272, 324, 60 },
		{ 272, 325, 60 },
		{ 272, 326, 60 },
		{ 272, 327, 60 },
		{ 272, 328, 60 },
		{ 272, 329, 60 },
		{ 272, 330, 60 },
		{ 272, 331, 60 },
		{ 272, 332, 60 },
		{ 272, 333, 60 },
		{ 272, 334, 60 },
		{ 272, 335, 60 },
		{ 272, 336, 60 },
		{ 272, 337, 60 },
		{ 272, 338, 60 },
		{ 272, 339, 60 },
		{ 272, 340, 60 },
		{ 272, 341, 60 },
		{ 272, 342, 60 },
		{ 272, 343, 60 },
		{ 272, 344, 60 },
		{ 272, 345, 60 },
		{ 272, 346, 60 },
		{ 272, 347, 60 },
		{ 272, 348, 60 },
		{ 272, 349, 60 },
		{ 272, 350, 60 },
		{ 272, 351, 60 },
		{ 272, 352, 60 },
		{ 272, 353, 60 },
		{ 272, 354, 60 },
		{ 272, 355, 60 },
		{ 272, 356, 60 },
		{ 272, 357, 60 },
		{ 272, 358, 60 },
		{ 272, 359, 60 },
		{ 272, 360, 60 },
		{ 272, 361, 60 },
		{ 272, 362, 60 },
		{ 272, 363, 60 },
		{ 272, 364, 60 },
		{ 272, 365, 60 },
		{ 272, 366, 60 },
		{ 272, 367, 60 },
		{ 272, 368, 60 },
		{ 272, 369, 60 },
		{ 272, 370, 60 },
		{ 272, 371, 60 },
		{ 272, 372, 60 },
		{ 272, 373, 60 },
		{ 272, 374, 60 },
		{ 272, 375, 60 },
		{ 272, 376, 60 },
		{ 273, 323, 60 },
		{ 273, 324, 60 },
		{ 273, 325, 60 },
		{ 273, 326, 60 },
		{ 273, 327, 60 },
		{ 273, 328, 60 },
		{ 273, 329, 60 },
		{ 273, 330, 60 },
		{ 273, 331, 60 },
		{ 273, 332, 60 },
		{ 273, 333, 60 },
		{ 273, 334, 60 },
		{ 273, 335, 60 },
		{ 273, 336, 60 },
		{ 273, 337, 60 },
		{ 273, 338, 60 },
		{ 273, 339, 60 },
		{ 273, 340, 60 },
		{ 273, 341, 60 },
		{ 273, 342, 60 },
		{ 273, 343, 60 },
		{ 273, 344, 60 },
		{ 273, 345, 60 },
		{ 273, 346, 60 },
		{ 273, 347, 60 },
		{ 273, 348, 60 },
		{ 273, 349, 60 },
		{ 273, 350, 60 },
		{ 273, 351, 60 },
		{ 273, 352, 60 },
		{ 273, 353, 60 },
		{ 273, 354, 60 },
		{ 273, 355, 60 },
		{ 273, 356, 60 },
		{ 273, 357, 60 },
		{ 273, 358, 60 },
		{ 273, 359, 60 },
		{ 273, 360, 60 },
		{ 273, 361, 60 },
		{ 273, 362, 60 },
		{ 273, 363, 60 },
		{ 273, 364, 60 },
		{ 273, 365, 60 },
		{ 273, 366, 60 },
		{ 273, 367, 60 },
		{ 273, 368, 60 },
		{ 273, 369, 60 },
		{ 273, 370, 60 },
		{ 273, 371, 60 },
		{ 273, 372, 60 },
		{ 273, 373, 60 },
		{ 273, 374, 60 },
		{ 273, 375, 60 },
		{ 273, 376, 60 },
		{ 273, 377, 60 },
		{ 273, 378, 60 },
		{ 273, 379, 60 },
		{ 273, 380, 60 },
		{ 273, 381, 60 },
		{ 273, 382, 60 },
		{ 273, 383, 60 },
		{ 273, 384, 60 },
		{ 274, 323, 60 },
		{ 274, 324, 60 },
		{ 274, 325, 60 },
		{ 274, 326, 60 },
		{ 274, 327, 60 },
		{ 274, 328, 60 },
		{ 274, 329, 60 },
		{ 274, 330, 60 },
		{ 274, 331, 60 },
		{ 274, 332, 60 },
		{ 274, 333, 60 },
		{ 274, 334, 60 },
		{ 274, 335, 60 },
		{ 274, 336, 60 },
		{ 274, 337, 60 },
		{ 274, 338, 60 },
		{ 274, 339, 60 },
		{ 274, 340, 60 },
		{ 274, 341, 60 },
		{ 274, 342, 60 },
		{ 274, 343, 60 },
		{ 274, 344, 60 },
		{ 274, 345, 60 },
		{ 274, 346, 60 },
		{ 274, 347, 60 },
		{ 274, 348, 60 },
		{ 274, 349, 60 },
		{ 274, 350, 60 },
		{ 274, 351, 60 },
		{ 274, 352, 60 },
		{ 274, 353, 60 },
		{ 274, 354, 60 },
		{ 274, 355, 60 },
		{ 274, 356, 60 },
		{ 274, 357, 60 },
		{ 274, 358, 60 },
		{ 274, 359, 60 },
		{ 274, 360, 60 },
		{ 274, 361, 60 },
		{ 274, 362, 60 },
		{ 274, 363, 60 },
		{ 274, 364, 60 },
		{ 274, 365, 60 },
		{ 274, 366, 60 },
		{ 274, 367, 60 },
		{ 274, 368, 60 },
		{ 274, 369, 60 },
		{ 274, 370, 60 },
		{ 274, 371, 60 },
		{ 274, 372, 60 },
		{ 274, 373, 60 },
		{ 274, 374, 60 },
		{ 274, 375, 60 },
		{ 274, 376, 60 },
		{ 274, 377, 60 },
		{ 274, 378, 60 },
		{ 274, 379, 60 },
		{ 274, 380, 60 },
		{ 274, 381, 60 },
		{ 274, 382, 60 },
		{ 274, 383, 60 },
		{ 274, 384, 60 },
		{ 274, 385, 60 },
		{ 274, 386, 60 },
		{ 274, 387, 60 },
		{ 274, 388, 60 },
		{ 274, 389, 60 },
		{ 274, 390, 60 },
		{ 274, 391, 60 },
		{ 274, 392, 60 },
		{ 275, 323, 60 },
		{ 275, 324, 60 },
		{ 275, 325, 60 },
		{ 275, 326, 60 },
		{ 275, 327, 60 },
		{ 275, 328, 60 },
		{ 275, 329, 60 },
		{ 275, 330, 60 },
		{ 275, 331, 60 },
		{ 275, 332, 60 },
		{ 275, 333, 60 },
		{ 275, 334, 60 },
		{ 275, 335, 60 },
		{ 275, 336, 60 },
		{ 275, 337, 60 },
		{ 275, 338, 60 },
		{ 275, 339, 60 },
		{ 275, 340, 60 },
		{ 275, 341, 60 },
		{ 275, 342, 60 },
		{ 275, 343, 60 },
		{ 275, 344, 60 },
		{ 275, 345, 60 },
		{ 275, 346, 60 },
		{ 275, 347, 60 },
		{ 275, 348, 60 },
		{ 275, 349, 60 },
		{ 275, 350, 60 },
		{ 275, 351, 60 },
		{ 275, 352, 60 },
		{ 275, 353, 60 },
		{ 275, 354, 60 },
		{ 275, 355, 60 },
		{ 275, 356, 60 },
		{ 275, 357, 60 },
		{ 275, 358, 60 },
		{ 275, 359, 60 },
		{ 275, 360, 60 },
		{ 275, 361, 60 },
		{ 275, 362, 60 },
		{ 275, 363, 60 },
		{ 275, 364, 60 },
		{ 275, 365, 60 },
		{ 275, 366, 60 },
		{ 275, 367, 60 },
		{ 275, 368, 60 },
		{ 275, 369, 60 },
		{ 275, 370, 60 },
		{ 275, 371, 60 },
		{ 275, 372, 60 },
		{ 275, 373, 60 },
		{ 275, 374, 60 },
		{ 275, 375, 60 },
		{ 275, 376, 60 },
		{ 275, 377, 60 },
		{ 275, 378, 60 },
		{ 275, 379, 60 },
		{ 275, 380, 60 },
		{ 275, 381, 60 },
		{ 275, 382, 60 },
		{ 275, 383, 60 },
		{ 275, 384, 60 },
		{ 275, 385, 60 },
		{ 275, 386, 60 },
		{ 275, 387, 60 },
		{ 275, 388, 60 },
		{ 275, 389, 60 },
		{ 275, 390, 60 },
		{ 275, 391, 60 },
		{ 275, 392, 60 },
		{ 275, 393, 60 },
		{ 275, 394, 60 },
		{ 275, 395, 60 },
		{ 275, 396, 60 },
		{ 275, 397, 60 },
		{ 275, 398, 60 },
		{ 275, 399, 60 },
		{ 275, 400, 60 },
		{ 276, 323, 60 },
		{ 276, 324, 60 },
		{ 276, 325, 60 },
		{ 276, 326, 60 },
		{ 276, 327, 60 },
		{ 276, 328, 60 },
		{ 276, 329, 60 },
		{ 276, 330, 60 },
		{ 276, 331, 60 },
		{ 276, 332, 60 },
		{ 276, 333, 60 },
		{ 276, 334, 60 },
		{ 276, 335, 60 },
		{ 276, 336, 60 },
		{ 276, 337, 60 },
		{ 276, 338, 60 },
		{ 276, 339, 60 },
		{ 276, 340, 60 },
		{ 276, 341, 60 },
		{ 276, 342, 60 },
		{ 276, 343, 60 },
		{ 276, 344, 60 },
		{ 276, 345, 60 },
		{ 276, 346, 60 },
		{ 276, 347, 60 },
		{ 276, 348, 60 },
		{ 276, 349, 60 },
		{ 276, 350, 60 },
		{ 276, 351, 60 },
		{ 276, 352, 60 },
		{ 276, 353, 60 },
		{ 276, 354, 60 },
		{ 276, 355, 60 },
		{ 276, 356, 60 },
		{ 276, 357, 60 },
		{ 276, 358, 60 },
		{ 276, 359, 60 },
		{ 276, 360, 60 },
		{ 276, 361, 60 },
		{ 276, 362, 60 },
		{ 276, 363, 60 },
		{ 276, 364, 60 },
		{ 276, 365, 60 },
		{ 276, 366, 60 },
		{ 276, 367, 60 },
		{ 276, 368, 60 },
		{ 276, 369, 60 },
		{ 276, 370, 60 },
		{ 276, 371, 60 },
		{ 276, 372, 60 },
		{ 276, 373, 60 },
		{ 276, 374, 60 },
		{ 276, 375, 60 },
		{ 276, 376, 60 },
		{ 276, 377, 60 },
		{ 276, 378, 60 },
		{ 276, 379, 60 },
		{ 276, 380, 60 },
		{ 276, 381, 60 },
		{ 276, 382, 60 },
		{ 276, 383, 60 },
		{ 276, 384, 60 },
		{ 276, 385, 60 },
		{ 276, 386, 60 },
		{ 276, 387, 60 },
		{ 276, 388, 60 },
		{ 276, 389, 60 },
		{ 276, 390, 60 },
		{ 276, 391, 60 },
		{ 276, 392, 60 },
		{ 276, 393, 60 },
		{ 276, 394, 60 },
		{ 276, 395, 60 },
		{ 276, 396, 60 },
		{ 276, 397, 60 },
		{ 276, 398, 60 },
		{ 276, 399, 60 },
		{ 276, 400, 60 },
		{ 276, 401, 60 },
		{ 276, 402, 60 },
		{ 276, 403, 60 },
		{ 276, 404, 60 },
		{ 276, 405, 60 },
		{ 276, 406, 60 },
		{ 276, 407, 60 },
		{ 276, 408, 60 },
		{ 277, 323, 60 },
		{ 277, 324, 60 },
		{ 277, 325, 60 },
		{ 277, 326, 60 },
		{ 277, 327, 60 },
		{ 277, 328, 60 },
		{ 277, 329, 60 },
		{ 277, 330, 60 },
		{ 277, 331, 60 },
		{ 277, 332, 60 },
		{ 277, 333, 60 },
		{ 277, 334, 60 },
		{ 277, 335, 60 },
		{ 277, 336, 60 },
		{ 277, 337, 60 },
		{ 277, 338, 60 },
		{ 277, 339, 60 },
		{ 277, 340, 60 },
		{ 277, 341, 60 },
		{ 277, 342, 60 },
		{ 277, 343, 60 },
		{ 277, 344, 60 },
		{ 277, 345, 60 },
		{ 277, 346, 60 },
		{ 277, 347, 60 },
		{ 277, 348, 60 },
		{ 277, 349, 60 },
		{ 277, 350, 60 },
		{ 277, 351, 60 },
		{ 277, 352, 60 },
		{ 277, 353, 60 },
		{ 277, 354, 60 },
		{ 277, 355, 60 },
		{ 277, 356, 60 },
		{ 277, 357, 60 },
		{ 277, 358, 60 },
		{ 277, 359, 60 },
		{ 277, 360, 60 },
		{ 277, 361, 60 },
		{ 277, 362, 60 },
		{ 277, 363, 60 },
		{ 277, 364, 60 },
		{ 277, 365, 60 },
		{ 277, 366, 60 },
		{ 277, 367, 60 },
		{ 277, 368, 60 },
		{ 277, 369, 60 },
		{ 277, 370, 60 },
		{ 277, 371, 60 },
		{ 277, 372, 60 },
		{ 277, 373, 60 },
		{ 277, 374, 60 },
		{ 277, 375, 60 },
		{ 277, 376, 60 },
		{ 277, 377, 60 },
		{ 277, 378, 60 },
		{ 277, 379, 60 },
		{ 277, 380, 60 },
		{ 277, 381, 60 },
		{ 277, 382, 60 },
		{ 277, 383, 60 },
		{ 277, 384, 60 },
		{ 277, 385, 60 },
		{ 277, 386, 60 },
		{ 277, 387, 60 },
		{ 277, 388, 60 },
		{ 277, 389, 60 },
		{ 277, 390, 60 },
		{ 277, 391, 60 },
		{ 277, 392, 60 },
		{ 277, 393, 60 },
		{ 277, 394, 60 },
		{ 277, 395, 60 },
		{ 277, 396, 60 },
		{ 277, 397, 60 },
		{ 277, 398, 60 },
		{ 277, 399, 60 },
		{ 277, 400, 60 },
		{ 277, 401, 60 },
		{ 277, 402, 60 },
		{ 277, 403, 60 },
		{ 277, 404, 60 },
		{ 277, 405, 60 },
		{ 277, 406, 60 },
		{ 277, 407, 60 },
		{ 277, 408, 60 },
		{ 277, 409, 60 },
		{ 277, 410, 60 },
		{ 277, 411, 60 },
		{ 277, 412, 60 },
		{ 277, 413, 60 },
		{ 277, 414, 60 },
		{ 277, 415, 60 },
		{ 277, 416, 60 },
		{ 278, 323, 60 },
		{ 278, 324, 60 },
		{ 278, 325, 60 },
		{ 278, 326, 60 },
		{ 278, 327, 60 },
		{ 278, 328, 60 },
		{ 278, 329, 60 },
		{ 278, 330, 60 },
		{ 278, 331, 60 },
		{ 278, 332, 60 },
		{ 278, 333, 60 },
		{ 278, 334, 60 },
		{ 278, 335, 60 },
		{ 278, 336, 60 },
		{ 278, 337, 60 },
		{ 278, 338, 60 },
		{ 278, 339, 60 },
		{ 278, 340, 60 },
		{ 278, 341, 60 },
		{ 278, 342, 60 },
		{ 278, 343, 60 },
		{ 278, 344, 60 },
		{ 278, 345, 60 },
		{ 278, 346, 60 },
		{ 278, 347, 60 },
		{ 278, 348, 60 },
		{ 278, 349, 60 },
		{ 278, 350, 60 },
		{ 278, 351, 60 },
		{ 278, 352, 60 },
		{ 278, 353, 60 },
		{ 278, 354, 60 },
		{ 278, 355, 60 },
		{ 278, 356, 60 },
		{ 278, 357, 60 },
		{ 278, 358, 60 },
		{ 278, 359, 60 },
		{ 278, 360, 60 },
		{ 278, 361, 60 },
		{ 278, 362, 60 },
		{ 278, 363, 60 },
		{ 278, 364, 60 },
		{ 278, 365, 60 },
		{ 278, 366, 60 },
		{ 278, 367, 60 },
		{ 278, 368, 60 },
		{ 278, 369, 60 },
		{ 278, 370, 60 },
		{ 278, 371, 60 },
		{ 278, 372, 60 },
		{ 278, 373, 60 },
		{ 278, 374, 60 },
		{ 278, 375, 60 },
		{ 278, 376, 60 },
		{ 278, 377, 60 },
		{ 278, 378, 60 },
		{ 278, 379, 60 },
		{ 278, 380, 60 },
		{ 278, 381, 60 },
		{ 278, 382, 60 },
		{ 278, 383, 60 },
		{ 278, 384, 60 },
		{ 278, 385, 60 },
		{ 278, 386, 60 },
		{ 278, 387, 60 },
		{ 278, 388, 60 },
		{ 278, 389, 60 },
		{ 278, 390, 60 },
		{ 278, 391, 60 },
		{ 278, 392, 60 },
		{ 278, 393, 60 },
		{ 278, 394, 60 },
		{ 278, 395, 60 },
		{ 278, 396, 60 },
		{ 278, 397, 60 },
		{ 278, 398, 60 },
		{ 278, 399, 60 },
		{ 278, 400, 60 },
		{ 278, 401, 60 },
		{ 278, 402, 60 },
		{ 278, 403, 60 },
		{ 278, 404, 60 },
		{ 278, 405, 60 },
		{ 278, 406, 60 },
		{ 278, 407, 60 },
		{ 278, 408, 60 },
		{ 278, 409, 60 },
		{ 278, 410, 60 },
		{ 278, 411, 60 },
		{ 278, 412, 60 },
		{ 278, 413, 60 },
		{ 278, 414, 60 },
		{ 278, 415, 60 },
		{ 278, 416, 60 },
		{ 278, 417, 60 },
		{ 278, 418, 60 },
		{ 278, 419, 60 },
		{ 278, 420, 60 },
		{ 278, 421, 60 },
		{ 278, 422, 60 },
		{ 278, 423, 60 },
		{ 278, 424, 60 },
		{ 279, 323, 60 },
		{ 279, 324, 60 },
		{ 279, 325, 60 },
		{ 279, 326, 60 },
		{ 279, 327, 60 },
		{ 279, 328, 60 },
		{ 279, 329, 60 },
		{ 279, 330, 60 },
		{ 279, 331, 60 },
		{ 279, 332, 60 },
		{ 279, 333, 60 },
		{ 279, 334, 60 },
		{ 279, 335, 60 },
		{ 279, 336, 60 },
		{ 279, 337, 60 },
		{ 279, 338, 60 },
		{ 279, 339, 60 },
		{ 279, 340, 60 },
		{ 279, 341, 60 },
		{ 279, 342, 60 },
		{ 279, 343, 60 },
		{ 279, 344, 60 },
		{ 279, 345, 60 },
		{ 279, 346, 60 },
		{ 279, 347, 60 },
		{ 279, 348, 60 },
		{ 279, 349, 60 },
		{ 279, 350, 60 },
		{ 279, 351, 60 },
		{ 279, 352, 60 },
		{ 279, 353, 60 },
		{ 279, 354, 60 },
		{ 279, 355, 60 },
		{ 279, 356, 60 },
		{ 279, 357, 60 },
		{ 279, 358, 60 },
		{ 279, 359, 60 },
		{ 279, 360, 60 },
		{ 279, 361, 60 },
		{ 279, 362, 60 },
		{ 279, 363, 60 },
		{ 279, 364, 60 },
		{ 279, 365, 60 },
		{ 279, 366, 60 },
		{ 279, 367, 60 },
		{ 279, 368, 60 },
		{ 279, 369, 60 },
		{ 279, 370, 60 },
		{ 279, 371, 60 },
		{ 279, 372, 60 },
		{ 279, 373, 60 },
		{ 279, 374, 60 },
		{ 279, 375, 60 },
		{ 279, 376, 60 },
		{ 279, 377, 60 },
		{ 279, 378, 60 },
		{ 279, 379, 60 },
		{ 279, 380, 60 },
		{ 279, 381, 60 },
		{ 279, 382, 60 },
		{ 279, 383, 60 },
		{ 279, 384, 60 },
		{ 279, 385, 60 },
		{ 279, 386, 60 },
		{ 279, 387, 60 },
		{ 279, 388, 60 },
		{ 279, 389, 60 },
		{ 279, 390, 60 },
		{ 279, 391, 60 },
		{ 279, 392, 60 },
		{ 279, 393, 60 },
		{ 279, 394, 60 },
		{ 279, 395, 60 },
		{ 279, 396, 60 },
		{ 279, 397, 60 },
		{ 279, 398, 60 },
		{ 279, 399, 60 },
		{ 279, 400, 60 },
		{ 279, 401, 60 },
		{ 279, 402, 60 },
		{ 279, 403, 60 },
		{ 279, 404, 60 },
		{ 279, 405, 60 },
		{ 279, 406, 60 },
		{ 279, 407, 60 },
		{ 279, 408, 60 },
		{ 279, 409, 60 },
		{ 279, 410, 60 },
		{ 279, 411, 60 },
		{ 279, 412, 60 },
		{ 279, 413, 60 },
		{ 279, 414, 60 },
		{ 279, 415, 60 },
		{ 279, 416, 60 },
		{ 279, 417, 60 },
		{ 279, 418, 60 },
		{ 279, 419, 60 },
		{ 279, 420, 60 },
		{ 279, 421, 60 },
		{ 279, 422, 60 },
		{ 279, 423, 60 },
		{ 279, 424, 60 },
		{ 279, 425, 60 },
		{ 279, 426, 60 },
		{ 279, 427, 60 },
		{ 279, 428, 60 },
		{ 279, 429, 60 },
		{ 279, 430, 60 },
		{ 279, 431, 60 },
		{ 279, 432, 60 },
		{ 280, 324, 60 },
		{ 280, 325, 60 },
		{ 280, 326, 60 },
		{ 280, 327, 60 },
		{ 280, 328, 60 },
		{ 280, 329, 60 },
		{ 280, 330, 60 },
		{ 280, 331, 60 },
		{ 280, 332, 60 },
		{ 280, 333, 60 },
		{ 280, 334, 60 },
		{ 280, 335, 60 },
		{ 280, 336, 60 },
		{ 280, 337, 60 },
		{ 280, 338, 60 },
		{ 280, 339, 60 },
		{ 280, 340, 60 },
		{ 280, 341, 60 },
		{ 280, 342, 60 },
		{ 280, 343, 60 },
		{ 280, 344, 60 },
		{ 280, 345, 60 },
		{ 280, 346, 60 },
		{ 280, 347, 60 },
		{ 280, 348, 60 },
		{ 280, 349, 60 },
		{ 280, 350, 60 },
		{ 280, 351, 60 },
		{ 280, 352, 60 },
		{ 280, 353, 60 },
		{ 280, 354, 60 },
		{ 280, 355, 60 },
		{ 280, 356, 60 },
		{ 280, 357, 60 },
		{ 280, 358, 60 },
		{ 280, 359, 60 },
		{ 280, 360, 60 },
		{ 280, 361, 60 },
		{ 280, 362, 60 },
		{ 280, 363, 60 },
		{ 280, 364, 60 },
		{ 280, 365, 60 },
		{ 280, 366, 60 },
		{ 280, 367, 60 },
		{ 280, 368, 60 },
		{ 280, 369, 60 },
		{ 280, 370, 60 },
		{ 280, 371, 60 },
		{ 280, 372, 60 },
		{ 280, 373, 60 },
		{ 280, 374, 60 },
		{ 280, 375, 60 },
		{ 280, 376, 60 },
		{ 280, 377, 60 },
		{ 280, 378, 60 },
		{ 280, 379, 60 },
		{ 280, 380, 60 },
		{ 280, 381, 60 },
		{ 280, 382, 60 },
		{ 280, 383, 60 },
		{ 280, 384, 60 },
		{ 280, 385, 60 },
		{ 280, 386, 60 },
		{ 280, 387, 60 },
		{ 280, 388, 60 },
		{ 280, 389, 60 },
		{ 280, 390, 60 },
		{ 280, 391, 60 },
		{ 280, 392, 60 },
		{ 280, 393, 60 },
		{ 280, 394, 60 },
		{ 280, 395, 60 },
		{ 280, 396, 60 },
		{ 280, 397, 60 },
		{ 280, 398, 60 },
		{ 280, 399, 60 },
		{ 280, 400, 60 },
		{ 280, 401, 60 },
		{ 280, 402, 60 },
		{ 280, 403, 60 },
		{ 280, 404, 60 },
		{ 280, 405, 60 },
		{ 280, 406, 60 },
		{ 280, 407, 60 },
		{ 280, 408, 60 },
		{ 280, 409, 60 },
		{ 280, 410, 60 },
		{ 280, 411, 60 },
		{ 280, 412, 60 },
		{ 280, 413, 60 },
		{ 280, 414, 60 },
		{ 280, 415, 60 },
		{ 280, 416, 60 },
		{ 280, 417, 60 },
		{ 280, 418, 60 },
		{ 280, 419, 60 },
		{ 280, 420, 60 },
		{ 280, 421, 60 },
		{ 280, 422, 60 },
		{ 280, 423, 60 },
		{ 280, 424, 60 },
		{ 280, 425, 60 },
		{ 280, 426, 60 },
		{ 280, 427, 60 },
		{ 280, 428, 60 },
		{ 280, 429, 60 },
		{ 280, 430, 60 },
		{ 280, 431, 60 },
		{ 280, 432, 60 },
		{ 280, 433, 60 },
		{ 280, 434, 60 },
		{ 280, 435, 60 },
		{ 280, 436, 60 },
		{ 280, 437, 60 },
		{ 280, 438, 60 },
		{ 280, 439, 60 },
		{ 280, 440, 60 },
		{ 281, 324, 60 },
		{ 281, 325, 60 },
		{ 281, 326, 60 },
		{ 281, 327, 60 },
		{ 281, 328, 60 },
		{ 281, 329, 60 },
		{ 281, 330, 60 },
		{ 281, 331, 60 },
		{ 281, 332, 60 },
		{ 281, 333, 60 },
		{ 281, 334, 60 },
		{ 281, 335, 60 },
		{ 281, 336, 60 },
		{ 281, 337, 60 },
		{ 281, 338, 60 },
		{ 281, 339, 60 },
		{ 281, 340, 60 },
		{ 281, 341, 60 },
		{ 281, 342, 60 },
		{ 281, 343, 60 },
		{ 281, 344, 60 },
		{ 281, 345, 60 },
		{ 281, 346, 60 },
		{ 281, 347, 60 },
		{ 281, 348, 60 },
		{ 281, 349, 60 },
		{ 281, 350, 60 },
		{ 281, 351, 60 },
		{ 281, 352, 60 },
		{ 281, 353, 60 },
		{ 281, 354, 60 },
		{ 281, 355, 60 },
		{ 281, 356, 60 },
		{ 281, 357, 60 },
		{ 281, 358, 60 },
		{ 281, 359, 60 },
		{ 281, 360, 60 },
		{ 281, 361, 60 },
		{ 281, 362, 60 },
		{ 281, 363, 60 },
		{ 281, 364, 60 },
		{ 281, 365, 60 },
		{ 281, 366, 60 },
		{ 281, 367, 60 },
		{ 281, 368, 60 },
		{ 281, 369, 60 },
		{ 281, 370, 60 },
		{ 281, 371, 60 },
		{ 281, 372, 60 },
		{ 281, 373, 60 },
		{ 281, 374, 60 },
		{ 281, 375, 60 },
		{ 281, 376, 60 },
		{ 281, 377, 60 },
		{ 281, 378, 60 },
		{ 281, 379, 60 },
		{ 281, 380, 60 },
		{ 281, 381, 60 },
		{ 281, 382, 60 },
		{ 281, 383, 60 },
		{ 281, 384, 60 },
		{ 281, 385, 60 },
		{ 281, 386, 60 },
		{ 281, 387, 60 },
		{ 281, 388, 60 },
		{ 281, 389, 60 },
		{ 281, 390, 60 },
		{ 281, 391, 60 },
		{ 281, 392, 60 },
		{ 281, 393, 60 },
		{ 281, 394, 60 },
		{ 281, 395, 60 },
		{ 281, 396, 60 },
		{ 281, 397, 60 },
		{ 281, 398, 60 },
		{ 281, 399, 60 },
		{ 281, 400, 60 },
		{ 281, 401, 60 },
		{ 281, 402, 60 },
		{ 281, 403, 60 },
		{ 281, 404, 60 },
		{ 281, 405, 60 },
		{ 281, 406, 60 },
		{ 281, 407, 60 },
		{ 281, 408, 60 },
		{ 281, 409, 60 },
		{ 281, 410, 60 },
		{ 281, 411, 60 },
		{ 281, 412, 60 },
		{ 281, 413, 60 },
		{ 281, 414, 60 },
		{ 281, 415, 60 },
		{ 281, 416, 60 },
		{ 281, 417, 60 },
		{ 281, 418, 60 },
		{ 281, 419, 60 },
		{ 281, 420, 60 },
		{ 281, 421, 60 },
		{ 281, 422, 60 },
		{ 281, 423, 60 },
		{ 281, 424, 60 },
		{ 281, 425, 60 },
		{ 281, 426, 60 },
		{ 281, 427, 60 },
		{ 281, 428, 60 },
		{ 281, 429, 60 },
		{ 281, 430, 60 },
		{ 281, 431, 60 },
		{ 281, 432, 60 },
		{ 281, 433, 60 },
		{ 281, 434, 60 },
		{ 281, 435, 60 },
		{ 281, 436, 60 },
		{ 281, 437, 60 },
		{ 281, 438, 60 },
		{ 281, 439, 60 },
		{ 281, 440, 60 },
		{ 281, 441, 60 },
		{ 281, 442, 60 },
		{ 281, 443, 60 },
		{ 281, 444, 60 },
		{ 281, 445, 60 },
		{ 281, 446, 60 },
		{ 281, 447, 60 },
		{ 281, 448, 60 },
		{ 282, 325, 60 },
		{ 282, 326, 60 },
		{ 282, 327, 60 },
		{ 282, 328, 60 },
		{ 282, 329, 60 },
		{ 282, 330, 60 },
		{ 282, 331, 60 },
		{ 282, 332, 60 },
		{ 282, 333, 60 },
		{ 282, 334, 60 },
		{ 282, 335, 60 },
		{ 282, 336, 60 },
		{ 282, 337, 60 },
		{ 282, 338, 60 },
		{ 282, 339, 60 },
		{ 282, 340, 60 },
		{ 282, 341, 60 },
		{ 282, 342, 60 },
		{ 282, 343, 60 },
		{ 282, 344, 60 },
		{ 282, 345, 60 },
		{ 282, 346, 60 },
		{ 282, 347, 60 },
		{ 282, 348, 60 },
		{ 282, 349, 60 },
		{ 282, 350, 60 },
		{ 282, 351, 60 },
		{ 282, 352, 60 },
		{ 282, 353, 60 },
		{ 282, 354, 60 },
		{ 282, 355, 60 },
		{ 282, 356, 60 },
		{ 282, 357, 60 },
		{ 282, 358, 60 },
		{ 282, 359, 60 },
		{ 282, 360, 60 },
		{ 282, 361, 60 },
		{ 282, 362, 60 },
		{ 282, 363, 60 },
		{ 282, 364, 60 },
		{ 282, 365, 60 },
		{ 282, 366, 60 },
		{ 282, 367, 60 },
		{ 282, 368, 60 },
		{ 282, 369, 60 },
		{ 282, 370, 60 },
		{ 282, 371, 60 },
		{ 282, 372, 60 },
		{ 282, 373, 60 },
		{ 282, 374, 60 },
		{ 282, 375, 60 },
		{ 282, 376, 60 },
		{ 282, 377, 60 },
		{ 282, 378, 60 },
		{ 282, 379, 60 },
		{ 282, 380, 60 },
		{ 282, 381, 60 },
		{ 282, 382, 60 },
		{ 282, 383, 60 },
		{ 282, 384, 60 },
		{ 282, 385, 60 },
		{ 282, 386, 60 },
		{ 282, 387, 60 },
		{ 282, 388, 60 },
		{ 282, 389, 60 },
		{ 282, 390, 60 },
		{ 282, 391, 60 },
		{ 282, 392, 60 },
		{ 282, 393, 60 },
		{ 282, 394, 60 },
		{ 282, 395, 60 },
		{ 282, 396, 60 },
		{ 282, 397, 60 },
		{ 282, 398, 60 },
		{ 282, 399, 60 },
		{ 282, 400, 60 },
		{ 282, 401, 60 },
		{ 282, 402, 60 },
		{ 282, 403, 60 },
		{ 282, 404, 60 },
		{ 282, 405, 60 },
		{ 282, 406, 60 },
		{ 282, 407, 60 },
		{ 282, 408, 60 },
		{ 282, 409, 60 },
		{ 282, 410, 60 },
		{ 282, 411, 60 },
		{ 282, 412, 60 },
		{ 282, 413, 60 },
		{ 282, 414, 60 },
		{ 282, 415, 60 },
		{ 282, 416, 60 },
		{ 282, 417, 60 },
		{ 282, 418, 60 },
		{ 282, 419, 60 },
		{ 282, 420, 60 },
		{ 282, 421, 60 },
		{ 282, 422, 60 },
		{ 282, 423, 60 },
		{ 282, 424, 60 },
		{ 282, 425, 60 },
		{ 282, 426, 60 },
		{ 282, 427, 60 },
		{ 282, 428, 60 },
		{ 282, 429, 60 },
		{ 282, 430, 60 },
		{ 282, 431, 60 },
		{ 282, 432, 60 },
		{ 282, 433, 60 },
		{ 282, 434, 60 },
		{ 282, 435, 60 },
		{ 282, 436, 60 },
		{ 282, 437, 60 },
		{ 282, 438, 60 },
		{ 282, 439, 60 },
		{ 282, 440, 60 },
		{ 282, 441, 60 },
		{ 282, 442, 60 },
		{ 282, 443, 60 },
		{ 282, 444, 60 },
		{ 282, 445, 60 },
		{ 282, 446, 60 },
		{ 282, 447, 60 },
		{ 282, 448, 60 },
		{ 282, 449, 60 },
		{ 282, 450, 60 },
		{ 282, 451, 60 },
		{ 282, 452, 60 },
		{ 282, 453, 60 },
		{ 282, 454, 60 },
		{ 282, 455, 60 },
		{ 282, 456, 60 },
		{ 283, 327, 60 },
		{ 283, 328, 60 },
		{ 283, 329, 60 },
		{ 283, 330, 60 },
		{ 283, 331, 60 },
		{ 283, 332, 60 },
		{ 283, 333, 60 },
		{ 283, 334, 60 },
		{ 283, 335, 60 },
		{ 283, 336, 60 },
		{ 283, 337, 60 },
		{ 283, 338, 60 },
		{ 283, 339, 60 },
		{ 283, 340, 60 },
		{ 283, 341, 60 },
		{ 283, 342, 60 },
		{ 283, 343, 60 },
		{ 283, 344, 60 },
		{ 283, 345, 60 },
		{ 283, 346, 60 },
		{ 283, 347, 60 },
		{ 283, 348, 60 },
		{ 283, 349, 60 },
		{ 283, 350, 60 },
		{ 283, 351, 60 },
		{ 283, 352, 60 },
		{ 283, 353, 60 },
		{ 283, 354, 60 },
		{ 283, 355, 60 },
		{ 283, 356, 60 },
		{ 283, 357, 60 },
		{ 283, 358, 60 },
		{ 283, 359, 60 },
		{ 283, 360, 60 },
		{ 283, 361, 60 },
		{ 283, 362, 60 },
		{ 283, 363, 60 },
		{ 283, 364, 60 },
		{ 283, 365, 60 },
		{ 283, 366, 60 },
		{ 283, 367, 60 },
		{ 283, 368, 60 },
		{ 283, 369, 60 },
		{ 283, 370, 60 },
		{ 283, 371, 60 },
		{ 283, 372, 60 },
		{ 283, 373, 60 },
		{ 283, 374, 60 },
		{ 283, 375, 60 },
		{ 283, 376, 60 },
		{ 283, 377, 60 },
		{ 283, 378, 60 },
		{ 283, 379, 60 },
		{ 283, 380, 60 },
		{ 283, 381, 60 },
		{ 283, 382, 60 },
		{ 283, 383, 60 },
		{ 283, 384, 60 },
		{ 283, 385, 60 },
		{ 283, 386, 60 },
		{ 283, 387, 60 },
		{ 283, 388, 60 },
		{ 283, 389, 60 },
		{ 283, 390, 60 },
		{ 283, 391, 60 },
		{ 283, 392, 60 },
		{ 283, 393, 60 },
		{ 283, 394, 60 },
		{ 283, 395, 60 },
		{ 283, 396, 60 },
		{ 283, 397, 60 },
		{ 283, 398, 60 },
		{ 283, 399, 60 },
		{ 283, 400, 60 },
		{ 283, 401, 60 },
		{ 283, 402, 60 },
		{ 283, 403, 60 },
		{ 283, 404, 60 },
		{ 283, 405, 60 },
		{ 283, 406, 60 },
		{ 283, 407, 60 },
		{ 283, 408, 60 },
		{ 283, 409, 60 },
		{ 283, 410, 60 },
		{ 283, 411, 60 },
		{ 283, 412, 60 },
		{ 283, 413, 60 },
		{ 283, 414, 60 },
		{ 283, 415, 60 },
		{ 283, 416, 60 },
		{ 283, 417, 60 },
		{ 283, 418, 60 },
		{ 283, 419, 60 },
		{ 283, 420, 60 },
		{ 283, 421, 60 },
		{ 283, 422, 60 },
		{ 283, 423, 60 },
		{ 283, 424, 60 },
		{ 283, 425, 60 },
		{ 283, 426, 60 },
		{ 283, 427, 60 },
		{ 283, 428, 60 },
		{ 283, 429, 60 },
		{ 283, 430, 60 },
		{ 283, 431, 60 },
		{ 283, 432, 60 },
		{ 283, 433, 60 },
		{ 283, 434, 60 },
		{ 283, 435, 60 },
		{ 283, 436, 60 },
		{ 283, 437, 60 },
		{ 283, 438, 60 },
		{ 283, 439, 60 },
		{ 283, 440, 60 },
		{ 283, 441, 60 },
		{ 283, 442, 60 },
		{ 283, 443, 60 },
		{ 283, 444, 60 },
		{ 283, 445, 60 },
		{ 283, 446, 60 },
		{ 283, 447, 60 },
		{ 283, 448, 60 },
		{ 283, 449, 60 },
		{ 283, 450, 60 },
		{ 283, 451, 60 },
		{ 283, 452, 60 },
		{ 283, 453, 60 },
		{ 283, 454, 60 },
		{ 283, 455, 60 },
		{ 283, 456, 60 },
		{ 283, 457, 60 },
		{ 283, 458, 60 },
		{ 283, 459, 60 },
		{ 283, 460, 60 },
		{ 283, 461, 60 },
		{ 283, 462, 60 },
		{ 283, 463, 60 },
		{ 283, 464, 60 },
		{ 284, 330, 60 },
		{ 284, 331, 60 },
		{ 284, 332, 60 },
		{ 284, 333, 60 },
		{ 284, 334, 60 },
		{ 284, 335, 60 },
		{ 284, 336, 60 },
		{ 284, 337, 60 },
		{ 284, 338, 60 },
		{ 284, 339, 60 },
		{ 284, 340, 60 },
		{ 284, 341, 60 },
		{ 284, 342, 60 },
		{ 284, 343, 60 },
		{ 284, 344, 60 },
		{ 284, 345, 60 },
		{ 284, 346, 60 },
		{ 284, 347, 60 },
		{ 284, 348, 60 },
		{ 284, 349, 60 },
		{ 284, 350, 60 },
		{ 284, 351, 60 },
		{ 284, 352, 60 },
		{ 284, 353, 60 },
		{ 284, 354, 60 },
		{ 284, 355, 60 },
		{ 284, 356, 60 },
		{ 284, 357, 60 },
		{ 284, 358, 60 },
		{ 284, 359, 60 },
		{ 284, 360, 60 },
		{ 284, 361, 60 },
		{ 284, 362, 60 },
		{ 284, 363, 60 },
		{ 284, 364, 60 },
		{ 284, 365, 60 },
		{ 284, 366, 60 },
		{ 284, 367, 60 },
		{ 284, 368, 60 },
		{ 284, 369, 60 },
		{ 284, 370, 60 },
		{ 284, 371, 60 },
		{ 284, 372, 60 },
		{ 284, 373, 60 },
		{ 284, 374, 60 },
		{ 284, 375, 60 },
		{ 284, 376, 60 },
		{ 284, 377, 60 },
		{ 284, 378, 60 },
		{ 284, 379, 60 },
		{ 284, 380, 60 },
		{ 284, 381, 60 },
		{ 284, 382, 60 },
		{ 284, 383, 60 },
		{ 284, 384, 60 },
		{ 284, 385, 60 },
		{ 284, 386, 60 },
		{ 284, 387, 60 },
		{ 284, 388, 60 },
		{ 284, 389, 60 },
		{ 284, 390, 60 },
		{ 284, 391, 60 },
		{ 284, 392, 60 },
		{ 284, 393, 60 },
		{ 284, 394, 60 },
		{ 284, 395, 60 },
		{ 284, 396, 60 },
		{ 284, 397, 60 },
		{ 284, 398, 60 },
		{ 284, 399, 60 },
		{ 284, 400, 60 },
		{ 284, 401, 60 },
		{ 284, 402, 60 },
		{ 284, 403, 60 },
		{ 284, 404, 60 },
		{ 284, 405, 60 },
		{ 284, 406, 60 },
		{ 284, 407, 60 },
		{ 284, 408, 60 },
		{ 284, 409, 60 },
		{ 284, 410, 60 },
		{ 284, 411, 60 },
		{ 284, 412, 60 },
		{ 284, 413, 60 },
		{ 284, 414, 60 },
		{ 284, 415, 60 },
		{ 284, 416, 60 },
		{ 284, 417, 60 },
		{ 284, 418, 60 },
		{ 284, 419, 60 },
		{ 284, 420, 60 },
		{ 284, 421, 60 },
		{ 284, 422, 60 },
		{ 284, 423, 60 },
		{ 284, 424, 60 },
		{ 284, 425, 60 },
		{ 284, 426, 60 },
		{ 284, 427, 60 },
		{ 284, 428, 60 },
		{ 284, 429, 60 },
		{ 284, 430, 60 },
		{ 284, 431, 60 },
		{ 284, 432, 60 },
		{ 284, 433, 60 },
		{ 284, 434, 60 },
		{ 284, 435, 60 },
		{ 284, 436, 60 },
		{ 284, 437, 60 },
		{ 284, 438, 60 },
		{ 284, 439, 60 },
		{ 284, 440, 60 },
		{ 284, 441, 60 },
		{ 284, 442, 60 },
		{ 284, 443, 60 },
		{ 284, 444, 60 },
		{ 284, 445, 60 },
		{ 284, 446, 60 },
		{ 284, 447, 60 },
		{ 284, 448, 60 },
		{ 284, 449, 60 },
		{ 284, 450, 60 },
		{ 284, 451, 60 },
		{ 284, 452, 60 },
		{ 284, 453, 60 },
		{ 284, 454, 60 },
		{ 284, 455, 60 },
		{ 284, 456, 60 },
		{ 284, 457, 60 },
		{ 284, 458, 60 },
		{ 284, 459, 60 },
		{ 284, 460, 60 },
		{ 284, 461, 60 },
		{ 284, 462, 60 },
		{ 284, 463, 60 },
		{ 284, 464, 60 },
		{ 284, 465, 60 },
		{ 284, 466, 60 },
		{ 284, 467, 60 },
		{ 284, 468, 60 },
		{ 284, 469, 60 },
		{ 284, 470, 60 },
		{ 284, 471, 60 },
		{ 284, 472, 60 },
		{ 285, 332, 60 },
		{ 285, 333, 60 },
		{ 285, 334, 60 },
		{ 285, 335, 60 },
		{ 285, 336, 60 },
		{ 285, 337, 60 },
		{ 285, 338, 60 },
		{ 285, 339, 60 },
		{ 285, 340, 60 },
		{ 285, 341, 60 },
		{ 285, 342, 60 },
		{ 285, 343, 60 },
		{ 285, 344, 60 },
		{ 285, 345, 60 },
		{ 285, 346, 60 },
		{ 285, 347, 60 },
		{ 285, 348, 60 },
		{ 285, 349, 60 },
		{ 285, 350, 60 },
		{ 285, 351, 60 },
		{ 285, 352, 60 },
		{ 285, 353, 60 },
		{ 285, 354, 60 },
		{ 285, 355, 60 },
		{ 285, 356, 60 },
		{ 285, 357, 60 },
		{ 285, 358, 60 },
		{ 285, 359, 60 },
		{ 285, 360, 60 },
		{ 285, 361, 60 },
		{ 285, 362, 60 },
		{ 285, 363, 60 },
		{ 285, 364, 60 },
		{ 285, 365, 60 },
		{ 285, 366, 60 },
		{ 285, 367, 60 },
		{ 285, 368, 60 },
		{ 285, 369, 60 },
		{ 285, 370, 60 },
		{ 285, 371, 60 },
		{ 285, 372, 60 },
		{ 285, 373, 60 },
		{ 285, 374, 60 },
		{ 285, 375, 60 },
		{ 285, 376, 60 },
		{ 285, 377, 60 },
		{ 285, 378, 60 },
		{ 285, 379, 60 },
		{ 285, 380, 60 },
		{ 285, 381, 60 },
		{ 285, 382, 60 },
		{ 285, 383, 60 },
		{ 285, 384, 60 },
		{ 285, 385, 60 },
		{ 285, 386, 60 },
		{ 285, 387, 60 },
		{ 285, 388, 60 },
		{ 285, 389, 60 },
		{ 285, 390, 60 },
		{ 285, 391, 60 },
		{ 285, 392, 60 },
		{ 285, 393, 60 },
		{ 285, 394, 60 },
		{ 285, 395, 60 },
		{ 285, 396, 60 },
		{ 285, 397, 60 },
		{ 285, 398, 60 },
		{ 285, 399, 60 },
		{ 285, 400, 60 },
		{ 285, 401, 60 },
		{ 285, 402, 60 },
		{ 285, 403, 60 },
		{ 285, 404, 60 },
		{ 285, 405, 60 },
		{ 285, 406, 60 },
		{ 285, 407, 60 },
		{ 285, 408, 60 },
		{ 285, 409, 60 },
		{ 285, 410, 60 },
		{ 285, 411, 60 },
		{ 285, 412, 60 },
		{ 285, 413, 60 },
		{ 285, 414, 60 },
		{ 285, 415, 60 },
		{ 285, 416, 60 },
		{ 285, 417, 60 },
		{ 285, 418, 60 },
		{ 285, 419, 60 },
		{ 285, 420, 60 },
		{ 285, 421, 60 },
		{ 285, 422, 60 },
		{ 285, 423, 60 },
		{ 285, 424, 60 },
		{ 285, 425, 60 },
		{ 285, 426, 60 },
		{ 285, 427, 60 },
		{ 285, 428, 60 },
		{ 285, 429, 60 },
		{ 285, 430, 60 },
		{ 285, 431, 60 },
		{ 285, 432, 60 },
		{ 285, 433, 60 },
		{ 285, 434, 60 },
		{ 285, 435, 60 },
		{ 285, 436, 60 },
		{ 285, 437, 60 },
		{ 285, 438, 60 },
		{ 285, 439, 60 },
		{ 285, 440, 60 },
		{ 285, 441, 60 },
		{ 285, 442, 60 },
		{ 285, 443, 60 },
		{ 285, 444, 60 },
		{ 285, 445, 60 },
		{ 285, 446, 60 },
		{ 285, 447, 60 },
		{ 285, 448, 60 },
		{ 285, 449, 60 },
		{ 285, 450, 60 },
		{ 285, 451, 60 },
		{ 285, 452, 60 },
		{ 285, 453, 60 },
		{ 285, 454, 60 },
		{ 285, 455, 60 },
		{ 285, 456, 60 },
		{ 285, 457, 60 },
		{ 285, 458, 60 },
		{ 285, 459, 60 },
		{ 285, 460, 60 },
		{ 285, 461, 60 },
		{ 285, 462, 60 },
		{ 285, 463, 60 },
		{ 285, 464, 60 },
		{ 285, 465, 60 },
		{ 285, 466, 60 },
		{ 285, 467, 60 },
		{ 285, 468, 60 },
		{ 285, 469, 60 },
		{ 285, 470, 60 },
		{ 285, 471, 60 },
		{ 285, 472, 60 },
		{ 285, 473, 60 },
		{ 285, 474, 60 },
		{ 285, 475, 60 },
		{ 285, 476, 60 },
		{ 285, 477, 60 },
		{ 285, 478, 60 },
		{ 285, 479, 60 },
		{ 285, 480, 60 },
		{ 286, 335, 60 },
		{ 286, 336, 60 },
		{ 286, 337, 60 },
		{ 286, 338, 60 },
		{ 286, 339, 60 },
		{ 286, 340, 60 },
		{ 286, 341, 60 },
		{ 286, 342, 60 },
		{ 286, 343, 60 },
		{ 286, 344, 60 },
		{ 286, 345, 60 },
		{ 286, 346, 60 },
		{ 286, 347, 60 },
		{ 286, 348, 60 },
		{ 286, 349, 60 },
		{ 286, 350, 60 },
		{ 286, 351, 60 },
		{ 286, 352, 60 },
		{ 286, 353, 60 },
		{ 286, 354, 60 },
		{ 286, 355, 60 },
		{ 286, 356, 60 },
		{ 286, 357, 60 },
		{ 286, 358, 60 },
		{ 286, 359, 60 },
		{ 286, 360, 60 },
		{ 286, 361, 60 },
		{ 286, 362, 60 },
		{ 286, 363, 60 },
		{ 286, 364, 60 },
		{ 286, 365, 60 },
		{ 286, 366, 60 },
		{ 286, 367, 60 },
		{ 286, 368, 60 },
		{ 286, 369, 60 },
		{ 286, 370, 60 },
		{ 286, 371, 60 },
		{ 286, 372, 60 },
		{ 286, 373, 60 },
		{ 286, 374, 60 },
		{ 286, 375, 60 },
		{ 286, 376, 60 },
		{ 286, 377, 60 },
		{ 286, 378, 60 },
		{ 286, 379, 60 },
		{ 286, 380, 60 },
		{ 286, 381, 60 },
		{ 286, 382, 60 },
		{ 286, 383, 60 },
		{ 286, 384, 60 },
		{ 286, 385, 60 },
		{ 286, 386, 60 },
		{ 286, 387, 60 },
		{ 286, 388, 60 },
		{ 286, 389, 60 },
		{ 286, 390, 60 },
		{ 286, 391, 60 },
		{ 286, 392, 60 },
		{ 286, 393, 60 },
		{ 286, 394, 60 },
		{ 286, 395, 60 },
		{ 286, 396, 60 },
		{ 286, 397, 60 },
		{ 286, 398, 60 },
		{ 286, 399, 60 },
		{ 286, 400, 60 },
		{ 286, 401, 60 },
		{ 286, 402, 60 },
		{ 286, 403, 60 },
		{ 286, 404, 60 },
		{ 286, 405, 60 },
		{ 286, 406, 60 },
		{ 286, 407, 60 },
		{ 286, 408, 60 },
		{ 286, 409, 60 },
		{ 286, 410, 60 },
		{ 286, 411, 60 },
		{ 286, 412, 60 },
		{ 286, 413, 60 },
		{ 286, 414, 60 },
		{ 286, 415, 60 },
		{ 286, 416, 60 },
		{ 286, 417, 60 },
		{ 286, 418, 60 },
		{ 286, 419, 60 },
		{ 286, 420, 60 },
		{ 286, 421, 60 },
		{ 286, 422, 60 },
		{ 286, 423, 60 },
		{ 286, 424, 60 },
		{ 286, 425, 60 },
		{ 286, 426, 60 },
		{ 286, 427, 60 },
		{ 286, 428, 60 },
		{ 286, 429, 60 },
		{ 286, 430, 60 },
		{ 286, 431, 60 },
		{ 286, 432, 60 },
		{ 286, 433, 60 },
		{ 286, 434, 60 },
		{ 286, 435, 60 },
		{ 286, 436, 60 },
		{ 286, 437, 60 },
		{ 286, 438, 60 },
		{ 286, 439, 60 },
		{ 286, 440, 60 },
		{ 286, 441, 60 },
		{ 286, 442, 60 },
		{ 286, 443, 60 },
		{ 286, 444, 60 },
		{ 286, 445, 60 },
		{ 286, 446, 60 },
		{ 286, 447, 60 },
		{ 286, 448, 60 },
		{ 286, 449, 60 },
		{ 286, 450, 60 },
		{ 286, 451, 60 },
		{ 286, 452, 60 },
		{ 286, 453, 60 },
		{ 286, 454, 60 },
		{ 286, 455, 60 },
		{ 286, 456, 60 },
		{ 286, 457, 60 },
		{ 286, 458, 60 },
		{ 286, 459, 60 },
		{ 286, 460, 60 },
		{ 286, 461, 60 },
		{ 286, 462, 60 },
		{ 286, 463, 60 },
		{ 286, 464, 60 },
		{ 286, 465, 60 },
		{ 286, 466, 60 },
		{ 286, 467, 60 },
		{ 286, 468, 60 },
		{ 286, 469, 60 },
		{ 286, 470, 60 },
		{ 286, 471, 60 },
		{ 286, 472, 60 },
		{ 286, 473, 60 },
		{ 286, 474, 60 },
		{ 286, 475, 60 },
		{ 286, 476, 60 },
		{ 286, 477, 60 },
		{ 286, 478, 60 },
		{ 286, 479, 60 },
		{ 286, 480, 60 },
		{ 286, 481, 60 },
		{ 286, 482, 60 },
		{ 286, 483, 60 },
		{ 286, 484, 60 },
		{ 286, 485, 60 },
		{ 286, 486, 60 },
		{ 286, 487, 60 },
		{ 286, 488, 60 },
		{ 287, 338, 60 },
		{ 287, 339, 60 },
		{ 287, 340, 60 },
		{ 287, 341, 60 },
		{ 287, 342, 60 },
		{ 287, 343, 60 },
		{ 287, 344, 60 },
		{ 287, 345, 60 },
		{ 287, 346, 60 },
		{ 287, 347, 60 },
		{ 287, 348, 60 },
		{ 287, 349, 60 },
		{ 287, 350, 60 },
		{ 287, 351, 60 },
		{ 287, 352, 60 },
		{ 287, 353, 60 },
		{ 287, 354, 60 },
		{ 287, 355, 60 },
		{ 287, 356, 60 },
		{ 287, 357, 60 },
		{ 287, 358, 60 },
		{ 287, 359, 60 },
		{ 287, 360, 60 },
		{ 287, 361, 60 },
		{ 287, 362, 60 },
		{ 287, 363, 60 },
		{ 287, 364, 60 },
		{ 287, 365, 60 },
		{ 287, 366, 60 },
		{ 287, 367, 60 },
		{ 287, 368, 60 },
		{ 287, 369, 60 },
		{ 287, 370, 60 },
		{ 287, 371, 60 },
		{ 287, 372, 60 },
		{ 287, 373, 60 },
		{ 287, 374, 60 },
		{ 287, 375, 60 },
		{ 287, 376, 60 },
		{ 287, 377, 60 },
		{ 287, 378, 60 },
		{ 287, 379, 60 },
		{ 287, 380, 60 },
		{ 287, 381, 60 },
		{ 287, 382, 60 },
		{ 287, 383, 60 },
		{ 287, 384, 60 },
		{ 287, 385, 60 },
		{ 287, 386, 60 },
		{ 287, 387, 60 },
		{ 287, 388, 60 },
		{ 287, 389, 60 },
		{ 287, 390, 60 },
		{ 287, 391, 60 },
		{ 287, 392, 60 },
		{ 287, 393, 60 },
		{ 287, 394, 60 },
		{ 287, 395, 60 },
		{ 287, 396, 60 },
		{ 287, 397, 60 },
		{ 287, 398, 60 },
		{ 287, 399, 60 },
		{ 287, 400, 60 },
		{ 287, 401, 60 },
		{ 287, 402, 60 },
		{ 287, 403, 60 },
		{ 287, 404, 60 },
		{ 287, 405, 60 },
		{ 287, 406, 60 },
		{ 287, 407, 60 },
		{ 287, 408, 60 },
		{ 287, 409, 60 },
		{ 287, 410, 60 },
		{ 287, 411, 60 },
		{ 287, 412, 60 },
		{ 287, 413, 60 },
		{ 287, 414, 60 },
		{ 287, 415, 60 },
		{ 287, 416, 60 },
		{ 287, 417, 60 },
		{ 287, 418, 60 },
		{ 287, 419, 60 },
		{ 287, 420, 60 },
		{ 287, 421, 60 },
		{ 287, 422, 60 },
		{ 287, 423, 60 },
		{ 287, 424, 60 },
		{ 287, 425, 60 },
		{ 287, 426, 60 },
		{ 287, 427, 60 },
		{ 287, 428, 60 },
		{ 287, 429, 60 },
		{ 287, 430, 60 },
		{ 287, 431, 60 },
		{ 287, 432, 60 },
		{ 287, 433, 60 },
		{ 287, 434, 60 },
		{ 287, 435, 60 },
		{ 287, 436, 60 },
		{ 287, 437, 60 },
		{ 287, 438, 60 },
		{ 287, 439, 60 },
		{ 287, 440, 60 },
		{ 287, 441, 60 },
		{ 287, 442, 60 },
		{ 287, 443, 60 },
		{ 287, 444, 60 },
		{ 287, 445, 60 },
		{ 287, 446, 60 },
		{ 287, 447, 60 },
		{ 287, 448, 60 },
		{ 287, 449, 60 },
		{ 287, 450, 60 },
		{ 287, 451, 60 },
		{ 287, 452, 60 },
		{ 287, 453, 60 },
		{ 287, 454, 60 },
		{ 287, 455, 60 },
		{ 287, 456, 60 },
		{ 287, 457, 60 },
		{ 287, 458, 60 },
		{ 287, 459, 60 },
		{ 287, 460, 60 },
		{ 287, 461, 60 },
		{ 287, 462, 60 },
		{ 287, 463, 60 },
		{ 287, 464, 60 },
		{ 287, 465, 60 },
		{ 287, 466, 60 },
		{ 287, 467, 60 },
		{ 287, 468, 60 },
		{ 287, 469, 60 },
		{ 287, 470, 60 },
		{ 287, 471, 60 },
		{ 287, 472, 60 },
		{ 287, 473, 60 },
		{ 287, 474, 60 },
		{ 287, 475, 60 },
		{ 287, 476, 60 },
		{ 287, 477, 60 },
		{ 287, 478, 60 },
		{ 287, 479, 60 },
		{ 287, 480, 60 },
		{ 287, 481, 60 },
		{ 287, 482, 60 },
		{ 287, 483, 60 },
		{ 287, 484, 60 },
		{ 287, 485, 60 },
		{ 287, 486, 60 },
		{ 287, 487, 60 },
		{ 287, 488, 60 },
		{ 287, 489, 60 },
		{ 287, 490, 60 },
		{ 287, 491, 60 },
		{ 287, 492, 60 },
		{ 287, 493, 60 },
		{ 287, 494, 60 },
		{ 287, 495, 60 },
		{ 287, 496, 60 },
		{ 288, 340, 60 },
		{ 288, 341, 60 },
		{ 288, 342, 60 },
		{ 288, 343, 60 },
		{ 288, 344, 60 },
		{ 288, 345, 60 },
		{ 288, 346, 60 },
		{ 288, 347, 60 },
		{ 288, 348, 60 },
		{ 288, 349, 60 },
		{ 288, 350, 60 },
		{ 288, 351, 60 },
		{ 288, 352, 60 },
		{ 288, 353, 60 },
		{ 288, 354, 60 },
		{ 288, 355, 60 },
		{ 288, 356, 60 },
		{ 288, 357, 60 },
		{ 288, 358, 60 },
		{ 288, 359, 60 },
		{ 288, 360, 60 },
		{ 288, 361, 60 },
		{ 288, 362, 60 },
		{ 288, 363, 60 },
		{ 288, 364, 60 },
		{ 288, 365, 60 },
		{ 288, 366, 60 },
		{ 288, 367, 60 },
		{ 288, 368, 60 },
		{ 288, 369, 60 },
		{ 288, 370, 60 },
		{ 288, 371, 60 },
		{ 288, 372, 60 },
		{ 288, 373, 60 },
		{ 288, 374, 60 },
		{ 288, 375, 60 },
		{ 288, 376, 60 },
		{ 288, 377, 60 },
		{ 288, 378, 60 },
		{ 288, 379, 60 },
		{ 288, 380, 60 },
		{ 288, 381, 60 },
		{ 288, 382, 60 },
		{ 288, 383, 60 },
		{ 288, 384, 60 },
		{ 288, 385, 60 },
		{ 288, 386, 60 },
		{ 288, 387, 60 },
		{ 288, 388, 60 },
		{ 288, 389, 60 },
		{ 288, 390, 60 },
		{ 288, 391, 60 },
		{ 288, 392, 60 },
		{ 288, 393, 60 },
		{ 288, 394, 60 },
		{ 288, 395, 60 },
		{ 288, 396, 60 },
		{ 288, 397, 60 },
		{ 288, 398, 60 },
		{ 288, 399, 60 },
		{ 288, 400, 60 },
		{ 288, 401, 60 },
		{ 288, 402, 60 },
		{ 288, 403, 60 },
		{ 288, 404, 60 },
		{ 288, 405, 60 },
		{ 288, 406, 60 },
		{ 288, 407, 60 },
		{ 288, 408, 60 },
		{ 288, 409, 60 },
		{ 288, 410, 60 },
		{ 288, 411, 60 },
		{ 288, 412, 60 },
		{ 288, 413, 60 },
		{ 288, 414, 60 },
		{ 288, 415, 60 },
		{ 288, 416, 60 },
		{ 288, 417, 60 },
		{ 288, 418, 60 },
		{ 288, 419, 60 },
		{ 288, 420, 60 },
		{ 288, 421, 60 },
		{ 288, 422, 60 },
		{ 288, 423, 60 },
		{ 288, 424, 60 },
		{ 288, 425, 60 },
		{ 288, 426, 60 },
		{ 288, 427, 60 },
		{ 288, 428, 60 },
		{ 288, 429, 60 },
		{ 288, 430, 60 },
		{ 288, 431, 60 },
		{ 288, 432, 60 },
		{ 288, 433, 60 },
		{ 288, 434, 60 },
		{ 288, 435, 60 },
		{ 288, 436, 60 },
		{ 288, 437, 60 },
		{ 288, 438, 60 },
		{ 288, 439, 60 },
		{ 288, 440, 60 },
		{ 288, 441, 60 },
		{ 288, 442, 60 },
		{ 288, 443, 60 },
		{ 288, 444, 60 },
		{ 288, 445, 60 },
		{ 288, 446, 60 },
		{ 288, 447, 60 },
		{ 288, 448, 60 },
		{ 288, 449, 60 },
		{ 288, 450, 60 },
		{ 288, 451, 60 },
		{ 288, 452, 60 },
		{ 288, 453, 60 },
		{ 288, 454, 60 },
		{ 288, 455, 60 },
		{ 288, 456, 60 },
		{ 288, 457, 60 },
		{ 288, 458, 60 },
		{ 288, 459, 60 },
		{ 288, 460, 60 },
		{ 288, 461, 60 },
		{ 288, 462, 60 },
		{ 288, 463, 60 },
		{ 288, 464, 60 },
		{ 288, 465, 60 },
		{ 288, 466, 60 },
		{ 288, 467, 60 },
		{ 288, 468, 60 },
		{ 288, 469, 60 },
		{ 288, 470, 60 },
		{ 288, 471, 60 },
		{ 288, 472, 60 },
		{ 288, 473, 60 },
		{ 288, 474, 60 },
		{ 288, 475, 60 },
		{ 288, 476, 60 },
		{ 288, 477, 60 },
		{ 288, 478, 60 },
		{ 288, 479, 60 },
		{ 288, 480, 60 },
		{ 288, 481, 60 },
		{ 288, 482, 60 },
		{ 288, 483, 60 },
		{ 288, 484, 60 },
		{ 288, 485, 60 },
		{ 288, 486, 60 },
		{ 288, 487, 60 },
		{ 288, 488, 60 },
		{ 288, 489, 60 },
		{ 288, 490, 60 },
		{ 288, 491, 60 },
		{ 288, 492, 60 },
		{ 288, 493, 60 },
		{ 288, 494, 60 },
		{ 288, 495, 60 },
		{ 288, 496, 60 },
		{ 288, 497, 60 },
		{ 288, 498, 60 },
		{ 288, 499, 60 },
		{ 288, 500, 60 },
		{ 288, 501, 60 },
		{ 288, 502, 60 },
		{ 288, 503, 60 },
		{ 288, 504, 60 },
		{ 289, 343, 60 },
		{ 289, 344, 60 },
		{ 289, 345, 60 },
		{ 289, 346, 60 },
		{ 289, 347, 60 },
		{ 289, 348, 60 },
		{ 289, 349, 60 },
		{ 289, 350, 60 },
		{ 289, 351, 60 },
		{ 289, 352, 60 },
		{ 289, 353, 60 },
		{ 289, 354, 60 },
		{ 289, 355, 60 },
		{ 289, 356, 60 },
		{ 289, 357, 60 },
		{ 289, 358, 60 },
		{ 289, 359, 60 },
		{ 289, 360, 60 },
		{ 289, 361, 60 },
		{ 289, 362, 60 },
		{ 289, 363, 60 },
		{ 289, 364, 60 },
		{ 289, 365, 60 },
		{ 289, 366, 60 },
		{ 289, 367, 60 },
		{ 289, 368, 60 },
		{ 289, 369, 60 },
		{ 289, 370, 60 },
		{ 289, 371, 60 },
		{ 289, 372, 60 },
		{ 289, 373, 60 },
		{ 289, 374, 60 },
		{ 289, 375, 60 },
		{ 289, 376, 60 },
		{ 289, 377, 60 },
		{ 289, 378, 60 },
		{ 289, 379, 60 },
		{ 289, 380, 60 },
		{ 289, 381, 60 },
		{ 289, 382, 60 },
		{ 289, 383, 60 },
		{ 289, 384, 60 },
		{ 289, 385, 60 },
		{ 289, 386, 60 },
		{ 289, 387, 60 },
		{ 289, 388, 60 },
		{ 289, 389, 60 },
		{ 289, 390, 60 },
		{ 289, 391, 60 },
		{ 289, 392, 60 },
		{ 289, 393, 60 },
		{ 289, 394, 60 },
		{ 289, 395, 60 },
		{ 289, 396, 60 },
		{ 289, 397, 60 },
		{ 289, 398, 60 },
		{ 289, 399, 60 },
		{ 289, 400, 60 },
		{ 289, 401, 60 },
		{ 289, 402, 60 },
		{ 289, 403, 60 },
		{ 289, 404, 60 },
		{ 289, 405, 60 },
		{ 289, 406, 60 },
		{ 289, 407, 60 },
		{ 289, 408, 60 },
		{ 289, 409, 60 },
		{ 289, 410, 60 },
		{ 289, 411, 60 },
		{ 289, 412, 60 },
		{ 289, 413, 60 },
		{ 289, 414, 60 },
		{ 289, 415, 60 },
		{ 289, 416, 60 },
		{ 289, 417, 60 },
		{ 289, 418, 60 },
		{ 289, 419, 60 },
		{ 289, 420, 60 },
		{ 289, 421, 60 },
		{ 289, 422, 60 },
		{ 289, 423, 60 },
		{ 289, 424, 60 },
		{ 289, 425, 60 },
		{ 289, 426, 60 },
		{ 289, 427, 60 },
		{ 289, 428, 60 },
		{ 289, 429, 60 },
		{ 289, 430, 60 },
		{ 289, 431, 60 },
		{ 289, 432, 60 },
		{ 289, 433, 60 },
		{ 289, 434, 60 },
		{ 289, 435, 60 },
		{ 289, 436, 60 },
		{ 289, 437, 60 },
		{ 289, 438, 60 },
		{ 289, 439, 60 },
		{ 289, 440, 60 },
		{ 289, 441, 60 },
		{ 289, 442, 60 },
		{ 289, 443, 60 },
		{ 289, 444, 60 },
		{ 289, 445, 60 },
		{ 289, 446, 60 },
		{ 289, 447, 60 },
		{ 289, 448, 60 },
		{ 289, 449, 60 },
		{ 289, 450, 60 },
		{ 289, 451, 60 },
		{ 289, 452, 60 },
		{ 289, 453, 60 },
		{ 289, 454, 60 },
		{ 289, 455, 60 },
		{ 289, 456, 60 },
		{ 289, 457, 60 },
		{ 289, 458, 60 },
		{ 289, 459, 60 },
		{ 289, 460, 60 },
		{ 289, 461, 60 },
		{ 289, 462, 60 },
		{ 289, 463, 60 },
		{ 289, 464, 60 },
		{ 289, 465, 60 },
		{ 289, 466, 60 },
		{ 289, 467, 60 },
		{ 289, 468, 60 },
		{ 289, 469, 60 },
		{ 289, 470, 60 },
		{ 289, 471, 60 },
		{ 289, 472, 60 },
		{ 289, 473, 60 },
		{ 289, 474, 60 },
		{ 289, 475, 60 },
		{ 289, 476, 60 },
		{ 289, 477, 60 },
		{ 289, 478, 60 },
		{ 289, 479, 60 },
		{ 289, 480, 60 },
		{ 289, 481, 60 },
		{ 289, 482, 60 },
		{ 289, 483, 60 },
		{ 289, 484, 60 },
		{ 289, 485, 60 },
		{ 289, 486, 60 },
		{ 289, 487, 60 },
		{ 289, 488, 60 },
		{ 289, 489, 60 },
		{ 289, 490, 60 },
		{ 289, 491, 60 },
		{ 289, 492, 60 },
		{ 289, 493, 60 },
		{ 289, 494, 60 },
		{ 289, 495, 60 },
		{ 289, 496, 60 },
		{ 289, 497, 60 },
		{ 289, 498, 60 },
		{ 289, 499, 60 },
		{ 289, 500, 60 },
		{ 289, 501, 60 },
		{ 289, 502, 60 },
		{ 289, 503, 60 },
		{ 289, 504, 60 },
		{ 289, 505, 60 },
		{ 289, 506, 60 },
		{ 289, 507, 60 },
		{ 289, 508, 60 },
		{ 289, 509, 60 },
		{ 289, 510, 60 },
		{ 289, 511, 60 },
		{ 289, 512, 60 },
		{ 290, 346, 60 },
		{ 290, 347, 60 },
		{ 290, 348, 60 },
		{ 290, 349, 60 },
		{ 290, 350, 60 },
		{ 290, 351, 60 },
		{ 290, 352, 60 },
		{ 290, 353, 60 },
		{ 290, 354, 60 },
		{ 290, 355, 60 },
		{ 290, 356, 60 },
		{ 290, 357, 60 },
		{ 290, 358, 60 },
		{ 290, 359, 60 },
		{ 290, 360, 60 },
		{ 290, 361, 60 },
		{ 290, 362, 60 },
		{ 290, 363, 60 },
		{ 290, 364, 60 },
		{ 290, 365, 60 },
		{ 290, 366, 60 },
		{ 290, 367, 60 },
		{ 290, 368, 60 },
		{ 290, 369, 60 },
		{ 290, 370, 60 },
		{ 290, 371, 60 },
		{ 290, 372, 60 },
		{ 290, 373, 60 },
		{ 290, 374, 60 },
		{ 290, 375, 60 },
		{ 290, 376, 60 },
		{ 290, 377, 60 },
		{ 290, 378, 60 },
		{ 290, 379, 60 },
		{ 290, 380, 60 },
		{ 290, 381, 60 },
		{ 290, 382, 60 },
		{ 290, 383, 60 },
		{ 290, 384, 60 },
		{ 290, 385, 60 },
		{ 290, 386, 60 },
		{ 290, 387, 60 },
		{ 290, 388, 60 },
		{ 290, 389, 60 },
		{ 290, 390, 60 },
		{ 290, 391, 60 },
		{ 290, 392, 60 },
		{ 290, 393, 60 },
		{ 290, 394, 60 },
		{ 290, 395, 60 },
		{ 290, 396, 60 },
		{ 290, 397, 60 },
		{ 290, 398, 60 },
		{ 290, 399, 60 },
		{ 290, 400, 60 },
		{ 290, 401, 60 },
		{ 290, 402, 60 },
		{ 290, 403, 60 },
		{ 290, 404, 60 },
		{ 290, 405, 60 },
		{ 290, 406, 60 },
		{ 290, 407, 60 },
		{ 290, 408, 60 },
		{ 290, 409, 60 },
		{ 290, 410, 60 },
		{ 290, 411, 60 },
		{ 290, 412, 60 },
		{ 290, 413, 60 },
		{ 290, 414, 60 },
		{ 290, 415, 60 },
		{ 290, 416, 60 },
		{ 290, 417, 60 },
		{ 290, 418, 60 },
		{ 290, 419, 60 },
		{ 290, 420, 60 },
		{ 290, 421, 60 },
		{ 290, 422, 60 },
		{ 290, 423, 60 },
		{ 290, 424, 60 },
		{ 290, 425, 60 },
		{ 290, 426, 60 },
		{ 290, 427, 60 },
		{ 290, 428, 60 },
		{ 290, 429, 60 },
		{ 290, 430, 60 },
		{ 290, 431, 60 },
		{ 290, 432, 60 },
		{ 290, 433, 60 },
		{ 290, 434, 60 },
		{ 290, 435, 60 },
		{ 290, 436, 60 },
		{ 290, 437, 60 },
		{ 290, 438, 60 },
		{ 290, 439, 60 },
		{ 290, 440, 60 },
		{ 290, 441, 60 },
		{ 290, 442, 60 },
		{ 290, 443, 60 },
		{ 290, 444, 60 },
		{ 290, 445, 60 },
		{ 290, 446, 60 },
		{ 290, 447, 60 },
		{ 290, 448, 60 },
		{ 290, 449, 60 },
		{ 290, 450, 60 },
		{ 290, 451, 60 },
		{ 290, 452, 60 },
		{ 290, 453, 60 },
		{ 290, 454, 60 },
		{ 290, 455, 60 },
		{ 290, 456, 60 },
		{ 290, 457, 60 },
		{ 290, 458, 60 },
		{ 290, 459, 60 },
		{ 290, 460, 60 },
		{ 290, 461, 60 },
		{ 290, 462, 60 },
		{ 290, 463, 60 },
		{ 290, 464, 60 },
		{ 290, 465, 60 },
		{ 290, 466, 60 },
		{ 290, 467, 60 },
		{ 290, 468, 60 },
		{ 290, 469, 60 },
		{ 290, 470, 60 },
		{ 290, 471, 60 },
		{ 290, 472, 60 },
		{ 290, 473, 60 },
		{ 290, 474, 60 },
		{ 290, 475, 60 },
		{ 290, 476, 60 },
		{ 290, 477, 60 },
		{ 290, 478, 60 },
		{ 290, 479, 60 },
		{ 290, 480, 60 },
		{ 290, 481, 60 },
		{ 290, 482, 60 },
		{ 290, 483, 60 },
		{ 290, 484, 60 },
		{ 290, 485, 60 },
		{ 290, 486, 60 },
		{ 290, 487, 60 },
		{ 290, 488, 60 },
		{ 290, 489, 60 },
		{ 290, 490, 60 },
		{ 290, 491, 60 },
		{ 290, 492, 60 },
		{ 290, 493, 60 },
		{ 290, 494, 60 },
		{ 290, 495, 60 },
		{ 290, 496, 60 },
		{ 290, 497, 60 },
		{ 290, 498, 60 },
		{ 290, 499, 60 },
		{ 290, 500, 60 },
		{ 290, 501, 60 },
		{ 290, 502, 60 },
		{ 290, 503, 60 },
		{ 290, 504, 60 },
		{ 290, 505, 60 },
		{ 290, 506, 60 },
		{ 290, 507, 60 },
		{ 290, 508, 60 },
		{ 290, 509, 60 },
		{ 290, 510, 60 },
		{ 290, 511, 60 },
		{ 290, 512, 60 },
		{ 291, 348, 60 },
		{ 291, 349, 60 },
		{ 291, 350, 60 },
		{ 291, 351, 60 },
		{ 291, 352, 60 },
		{ 291, 353, 60 },
		{ 291, 354, 60 },
		{ 291, 355, 60 },
		{ 291, 356, 60 },
		{ 291, 357, 60 },
		{ 291, 358, 60 },
		{ 291, 359, 60 },
		{ 291, 360, 60 },
		{ 291, 361, 60 },
		{ 291, 362, 60 },
		{ 291, 363, 60 },
		{ 291, 364, 60 },
		{ 291, 365, 60 },
		{ 291, 366, 60 },
		{ 291, 367, 60 },
		{ 291, 368, 60 },
		{ 291, 369, 60 },
		{ 291, 370, 60 },
		{ 291, 371, 60 },
		{ 291, 372, 60 },
		{ 291, 373, 60 },
		{ 291, 374, 60 },
		{ 291, 375, 60 },
		{ 291, 376, 60 },
		{ 291, 377, 60 },
		{ 291, 378, 60 },
		{ 291, 379, 60 },
		{ 291, 380, 60 },
		{ 291, 381, 60 },
		{ 291, 382, 60 },
		{ 291, 383, 60 },
		{ 291, 384, 60 },
		{ 291, 385, 60 },
		{ 291, 386, 60 },
		{ 291, 387, 60 },
		{ 291, 388, 60 },
		{ 291, 389, 60 },
		{ 291, 390, 60 },
		{ 291, 391, 60 },
		{ 291, 392, 60 },
		{ 291, 393, 60 },
		{ 291, 394, 60 },
		{ 291, 395, 60 },
		{ 291, 396, 60 },
		{ 291, 397, 60 },
		{ 291, 398, 60 },
		{ 291, 399, 60 },
		{ 291, 400, 60 },
		{ 291, 401, 60 },
		{ 291, 402, 60 },
		{ 291, 403, 60 },
		{ 291, 404, 60 },
		{ 291, 405, 60 },
		{ 291, 406, 60 },
		{ 291, 407, 60 },
		{ 291, 408, 60 },
		{ 291, 409, 60 },
		{ 291, 410, 60 },
		{ 291, 411, 60 },
		{ 291, 412, 60 },
		{ 291, 413, 60 },
		{ 291, 414, 60 },
		{ 291, 415, 60 },
		{ 291, 416, 60 },
		{ 291, 417, 60 },
		{ 291, 418, 60 },
		{ 291, 419, 60 },
		{ 291, 420, 60 },
		{ 291, 421, 60 },
		{ 291, 422, 60 },
		{ 291, 423, 60 },
		{ 291, 424, 60 },
		{ 291, 425, 60 },
		{ 291, 426, 60 },
		{ 291, 427, 60 },
		{ 291, 428, 60 },
		{ 291, 429, 60 },
		{ 291, 430, 60 },
		{ 291, 431, 60 },
		{ 291, 432, 60 },
		{ 291, 433, 60 },
		{ 291, 434, 60 },
		{ 291, 435, 60 },
		{ 291, 436, 60 },
		{ 291, 437, 60 },
		{ 291, 438, 60 },
		{ 291, 439, 60 },
		{ 291, 440, 60 },
		{ 291, 441, 60 },
		{ 291, 442, 60 },
		{ 291, 443, 60 },
		{ 291, 444, 60 },
		{ 291, 445, 60 },
		{ 291, 446, 60 },
		{ 291, 447, 60 },
		{ 291, 448, 60 },
		{ 291, 449, 60 },
		{ 291, 450, 60 },
		{ 291, 451, 60 },
		{ 291, 452, 60 },
		{ 291, 453, 60 },
		{ 291, 454, 60 },
		{ 291, 455, 60 },
		{ 291, 456, 60 },
		{ 291, 457, 60 },
		{ 291, 458, 60 },
		{ 291, 459, 60 },
		{ 291, 460, 60 },
		{ 291, 461, 60 },
		{ 291, 462, 60 },
		{ 291, 463, 60 },
		{ 291, 464, 60 },
		{ 291, 465, 60 },
		{ 291, 466, 60 },
		{ 291, 467, 60 },
		{ 291, 468, 60 },
		{ 291, 469, 60 },
		{ 291, 470, 60 },
		{ 291, 471, 60 },
		{ 291, 472, 60 },
		{ 291, 473, 60 },
		{ 291, 474, 60 },
		{ 291, 475, 60 },
		{ 291, 476, 60 },
		{ 291, 477, 60 },
		{ 291, 478, 60 },
		{ 291, 479, 60 },
		{ 291, 480, 60 },
		{ 291, 481, 60 },
		{ 291, 482, 60 },
		{ 291, 483, 60 },
		{ 291, 484, 60 },
		{ 291, 485, 60 },
		{ 291, 486, 60 },
		{ 291, 487, 60 },
		{ 291, 488, 60 },
		{ 291, 489, 60 },
		{ 291, 490, 60 },
		{ 291, 491, 60 },
		{ 291, 492, 60 },
		{ 291, 493, 60 },
		{ 291, 494, 60 },
		{ 291, 495, 60 },
		{ 291, 496, 60 },
		{ 291, 497, 60 },
		{ 291, 498, 60 },
		{ 291, 499, 60 },
		{ 291, 500, 60 },
		{ 291, 501, 60 },
		{ 291, 502, 60 },
		{ 291, 503, 60 },
		{ 291, 504, 60 },
		{ 291, 505, 60 },
		{ 291, 506, 60 },
		{ 291, 507, 60 },
		{ 291, 508, 60 },
		{ 291, 509, 60 },
		{ 291, 510, 60 },
		{ 291, 511, 60 },
		{ 291, 512, 60 },
		{ 292, 351, 60 },
		{ 292, 352, 60 },
		{ 292, 353, 60 },
		{ 292, 354, 60 },
		{ 292, 355, 60 },
		{ 292, 356, 60 },
		{ 292, 357, 60 },
		{ 292, 358, 60 },
		{ 292, 359, 60 },
		{ 292, 360, 60 },
		{ 292, 361, 60 },
		{ 292, 362, 60 },
		{ 292, 363, 60 },
		{ 292, 364, 60 },
		{ 292, 365, 60 },
		{ 292, 366, 60 },
		{ 292, 367, 60 },
		{ 292, 368, 60 },
		{ 292, 369, 60 },
		{ 292, 370, 60 },
		{ 292, 371, 60 },
		{ 292, 372, 60 },
		{ 292, 373, 60 },
		{ 292, 374, 60 },
		{ 292, 375, 60 },
		{ 292, 376, 60 },
		{ 292, 377, 60 },
		{ 292, 378, 60 },
		{ 292, 379, 60 },
		{ 292, 380, 60 },
		{ 292, 381, 60 },
		{ 292, 382, 60 },
		{ 292, 383, 60 },
		{ 292, 384, 60 },
		{ 292, 385, 60 },
		{ 292, 386, 60 },
		{ 292, 387, 60 },
		{ 292, 388, 60 },
		{ 292, 389, 60 },
		{ 292, 390, 60 },
		{ 292, 391, 60 },
		{ 292, 392, 60 },
		{ 292, 393, 60 },
		{ 292, 394, 60 },
		{ 292, 395, 60 },
		{ 292, 396, 60 },
		{ 292, 397, 60 },
		{ 292, 398, 60 },
		{ 292, 399, 60 },
		{ 292, 400, 60 },
		{ 292, 401, 60 },
		{ 292, 402, 60 },
		{ 292, 403, 60 },
		{ 292, 404, 60 },
		{ 292, 405, 60 },
		{ 292, 406, 60 },
		{ 292, 407, 60 },
		{ 292, 408, 60 },
		{ 292, 409, 60 },
		{ 292, 410, 60 },
		{ 292, 411, 60 },
		{ 292, 412, 60 },
		{ 292, 413, 60 },
		{ 292, 414, 60 },
		{ 292, 415, 60 },
		{ 292, 416, 60 },
		{ 292, 417, 60 },
		{ 292, 418, 60 },
		{ 292, 419, 60 },
		{ 292, 420, 60 },
		{ 292, 421, 60 },
		{ 292, 422, 60 },
		{ 292, 423, 60 },
		{ 292, 424, 60 },
		{ 292, 425, 60 },
		{ 292, 426, 60 },
		{ 292, 427, 60 },
		{ 292, 428, 60 },
		{ 292, 429, 60 },
		{ 292, 430, 60 },
		{ 292, 431, 60 },
		{ 292, 432, 60 },
		{ 292, 433, 60 },
		{ 292, 434, 60 },
		{ 292, 435, 60 },
		{ 292, 436, 60 },
		{ 292, 437, 60 },
		{ 292, 438, 60 },
		{ 292, 439, 60 },
		{ 292, 440, 60 },
		{ 292, 441, 60 },
		{ 292, 442, 60 },
		{ 292, 443, 60 },
		{ 292, 444, 60 },
		{ 292, 445, 60 },
		{ 292, 446, 60 },
		{ 292, 447, 60 },
		{ 292, 448, 60 },
		{ 292, 449, 60 },
		{ 292, 450, 60 },
		{ 292, 451, 60 },
		{ 292, 452, 60 },
		{ 292, 453, 60 },
		{ 292, 454, 60 },
		{ 292, 455, 60 },
		{ 292, 456, 60 },
		{ 292, 457, 60 },
		{ 292, 458, 60 },
		{ 292, 459, 60 },
		{ 292, 460, 60 },
		{ 292, 461, 60 },
		{ 292, 462, 60 },
		{ 292, 463, 60 },
		{ 292, 464, 60 },
		{ 292, 465, 60 },
		{ 292, 466, 60 },
		{ 292, 467, 60 },
		{ 292, 468, 60 },
		{ 292, 469, 60 },
		{ 292, 470, 60 },
		{ 292, 471, 60 },
		{ 292, 472, 60 },
		{ 292, 473, 60 },
		{ 292, 474, 60 },
		{ 292, 475, 60 },
		{ 292, 476, 60 },
		{ 292, 477, 60 },
		{ 292, 478, 60 },
		{ 292, 479, 60 },
		{ 292, 480, 60 },
		{ 292, 481, 60 },
		{ 292, 482, 60 },
		{ 292, 483, 60 },
		{ 292, 484, 60 },
		{ 292, 485, 60 },
		{ 292, 486, 60 },
		{ 292, 487, 60 },
		{ 292, 488, 60 },
		{ 292, 489, 60 },
		{ 292, 490, 60 },
		{ 292, 491, 60 },
		{ 292, 492, 60 },
		{ 292, 493, 60 },
		{ 292, 494, 60 },
		{ 292, 495, 60 },
		{ 292, 496, 60 },
		{ 292, 497, 60 },
		{ 292, 498, 60 },
		{ 292, 499, 60 },
		{ 292, 500, 60 },
		{ 292, 501, 60 },
		{ 292, 502, 60 },
		{ 292, 503, 60 },
		{ 292, 504, 60 },
		{ 292, 505, 60 },
		{ 292, 506, 60 },
		{ 292, 507, 60 },
		{ 292, 508, 60 },
		{ 292, 509, 60 },
		{ 292, 510, 60 },
		{ 292, 511, 60 },
		{ 292, 512, 60 },
		{ 293, 354, 60 },
		{ 293, 355, 60 },
		{ 293, 356, 60 },
		{ 293, 357, 60 },
		{ 293, 358, 60 },
		{ 293, 359, 60 },
		{ 293, 360, 60 },
		{ 293, 361, 60 },
		{ 293, 362, 60 },
		{ 293, 363, 60 },
		{ 293, 364, 60 },
		{ 293, 365, 60 },
		{ 293, 366, 60 },
		{ 293, 367, 60 },
		{ 293, 368, 60 },
		{ 293, 369, 60 },
		{ 293, 370, 60 },
		{ 293, 371, 60 },
		{ 293, 372, 60 },
		{ 293, 373, 60 },
		{ 293, 374, 60 },
		{ 293, 375, 60 },
		{ 293, 376, 60 },
		{ 293, 377, 60 },
		{ 293, 378, 60 },
		{ 293, 379, 60 },
		{ 293, 380, 60 },
		{ 293, 381, 60 },
		{ 293, 382, 60 },
		{ 293, 383, 60 },
		{ 293, 384, 60 },
		{ 293, 385, 60 },
		{ 293, 386, 60 },
		{ 293, 387, 60 },
		{ 293, 388, 60 },
		{ 293, 389, 60 },
		{ 293, 390, 60 },
		{ 293, 391, 60 },
		{ 293, 392, 60 },
		{ 293, 393, 60 },
		{ 293, 394, 60 },
		{ 293, 395, 60 },
		{ 293, 396, 60 },
		{ 293, 397, 60 },
		{ 293, 398, 60 },
		{ 293, 399, 60 },
		{ 293, 400, 60 },
		{ 293, 401, 60 },
		{ 293, 402, 60 },
		{ 293, 403, 60 },
		{ 293, 404, 60 },
		{ 293, 405, 60 },
		{ 293, 406, 60 },
		{ 293, 407, 60 },
		{ 293, 408, 60 },
		{ 293, 409, 60 },
		{ 293, 410, 60 },
		{ 293, 411, 60 },
		{ 293, 412, 60 },
		{ 293, 413, 60 },
		{ 293, 414, 60 },
		{ 293, 415, 60 },
		{ 293, 416, 60 },
		{ 293, 417, 60 },
		{ 293, 418, 60 },
		{ 293, 419, 60 },
		{ 293, 420, 60 },
		{ 293, 421, 60 },
		{ 293, 422, 60 },
		{ 293, 423, 60 },
		{ 293, 424, 60 },
		{ 293, 425, 60 },
		{ 293, 426, 60 },
		{ 293, 427, 60 },
		{ 293, 428, 60 },
		{ 293, 429, 60 },
		{ 293, 430, 60 },
		{ 293, 431, 60 },
		{ 293, 432, 60 },
		{ 293, 433, 60 },
		{ 293, 434, 60 },
		{ 293, 435, 60 },
		{ 293, 436, 60 },
		{ 293, 437, 60 },
		{ 293, 438, 60 },
		{ 293, 439, 60 },
		{ 293, 440, 60 },
		{ 293, 441, 60 },
		{ 293, 442, 60 },
		{ 293, 443, 60 },
		{ 293, 444, 60 },
		{ 293, 445, 60 },
		{ 293, 446, 60 },
		{ 293, 447, 60 },
		{ 293, 448, 60 },
		{ 293, 449, 60 },
		{ 293, 450, 60 },
		{ 293, 451, 60 },
		{ 293, 452, 60 },
		{ 293, 453, 60 },
		{ 293, 454, 60 },
		{ 293, 455, 60 },
		{ 293, 456, 60 },
		{ 293, 457, 60 },
		{ 293, 458, 60 },
		{ 293, 459, 60 },
		{ 293, 460, 60 },
		{ 293, 461, 60 },
		{ 293, 462, 60 },
		{ 293, 463, 60 },
		{ 293, 464, 60 },
		{ 293, 465, 60 },
		{ 293, 466, 60 },
		{ 293, 467, 60 },
		{ 293, 468, 60 },
		{ 293, 469, 60 },
		{ 293, 470, 60 },
		{ 293, 471, 60 },
		{ 293, 472, 60 },
		{ 293, 473, 60 },
		{ 293, 474, 60 },
		{ 293, 475, 60 },
		{ 293, 476, 60 },
		{ 293, 477, 60 },
		{ 293, 478, 60 },
		{ 293, 479, 60 },
		{ 293, 480, 60 },
		{ 293, 481, 60 },
		{ 293, 482, 60 },
		{ 293, 483, 60 },
		{ 293, 484, 60 },
		{ 293, 485, 60 },
		{ 293, 486, 60 },
		{ 293, 487, 60 },
		{ 293, 488, 60 },
		{ 293, 489, 60 },
		{ 293, 490, 60 },
		{ 293, 491, 60 },
		{ 293, 492, 60 },
		{ 293, 493, 60 },
		{ 293, 494, 60 },
		{ 293, 495, 60 },
		{ 293, 496, 60 },
		{ 293, 497, 60 },
		{ 293, 498, 60 },
		{ 293, 499, 60 },
		{ 293, 500, 60 },
		{ 293, 501, 60 },
		{ 293, 502, 60 },
		{ 293, 503, 60 },
		{ 293, 504, 60 },
		{ 293, 505, 60 },
		{ 293, 506, 60 },
		{ 293, 507, 60 },
		{ 293, 508, 60 },
		{ 293, 509, 60 },
		{ 293, 510, 60 },
		{ 293, 511, 60 },
		{ 293, 512, 60 },
		{ 294, 356, 60 },
		{ 294, 357, 60 },
		{ 294, 358, 60 },
		{ 294, 359, 60 },
		{ 294, 360, 60 },
		{ 294, 361, 60 },
		{ 294, 362, 60 },
		{ 294, 363, 60 },
		{ 294, 364, 60 },
		{ 294, 365, 60 },
		{ 294, 366, 60 },
		{ 294, 367, 60 },
		{ 294, 368, 60 },
		{ 294, 369, 60 },
		{ 294, 370, 60 },
		{ 294, 371, 60 },
		{ 294, 372, 60 },
		{ 294, 373, 60 },
		{ 294, 374, 60 },
		{ 294, 375, 60 },
		{ 294, 376, 60 },
		{ 294, 377, 60 },
		{ 294, 378, 60 },
		{ 294, 379, 60 },
		{ 294, 380, 60 },
		{ 294, 381, 60 },
		{ 294, 382, 60 },
		{ 294, 383, 60 },
		{ 294, 384, 60 },
		{ 294, 385, 60 },
		{ 294, 386, 60 },
		{ 294, 387, 60 },
		{ 294, 388, 60 },
		{ 294, 389, 60 },
		{ 294, 390, 60 },
		{ 294, 391, 60 },
		{ 294, 392, 60 },
		{ 294, 393, 60 },
		{ 294, 394, 60 },
		{ 294, 395, 60 },
		{ 294, 396, 60 },
		{ 294, 397, 60 },
		{ 294, 398, 60 },
		{ 294, 399, 60 },
		{ 294, 400, 60 },
		{ 294, 401, 60 },
		{ 294, 402, 60 },
		{ 294, 403, 60 },
		{ 294, 404, 60 },
		{ 294, 405, 60 },
		{ 294, 406, 60 },
		{ 294, 407, 60 },
		{ 294, 408, 60 },
		{ 294, 409, 60 },
		{ 294, 410, 60 },
		{ 294, 411, 60 },
		{ 294, 412, 60 },
		{ 294, 413, 60 },
		{ 294, 414, 60 },
		{ 294, 415, 60 },
		{ 294, 416, 60 },
		{ 294, 417, 60 },
		{ 294, 418, 60 },
		{ 294, 419, 60 },
		{ 294, 420, 60 },
		{ 294, 421, 60 },
		{ 294, 422, 60 },
		{ 294, 423, 60 },
		{ 294, 424, 60 },
		{ 294, 425, 60 },
		{ 294, 426, 60 },
		{ 294, 427, 60 },
		{ 294, 428, 60 },
		{ 294, 429, 60 },
		{ 294, 430, 60 },
		{ 294, 431, 60 },
		{ 294, 432, 60 },
		{ 294, 433, 60 },
		{ 294, 434, 60 },
		{ 294, 435, 60 },
		{ 294, 436, 60 },
		{ 294, 437, 60 },
		{ 294, 438, 60 },
		{ 294, 439, 60 },
		{ 294, 440, 60 },
		{ 294, 441, 60 },
		{ 294, 442, 60 },
		{ 294, 443, 60 },
		{ 294, 444, 60 },
		{ 294, 445, 60 },
		{ 294, 446, 60 },
		{ 294, 447, 60 },
		{ 294, 448, 60 },
		{ 294, 449, 60 },
		{ 294, 450, 60 },
		{ 294, 451, 60 },
		{ 294, 452, 60 },
		{ 294, 453, 60 },
		{ 294, 454, 60 },
		{ 294, 455, 60 },
		{ 294, 456, 60 },
		{ 294, 457, 60 },
		{ 294, 458, 60 },
		{ 294, 459, 60 },
		{ 294, 460, 60 },
		{ 294, 461, 60 },
		{ 294, 462, 60 },
		{ 294, 463, 60 },
		{ 294, 464, 60 },
		{ 294, 465, 60 },
		{ 294, 466, 60 },
		{ 294, 467, 60 },
		{ 294, 468, 60 },
		{ 294, 469, 60 },
		{ 294, 470, 60 },
		{ 294, 471, 60 },
		{ 294, 472, 60 },
		{ 294, 473, 60 },
		{ 294, 474, 60 },
		{ 294, 475, 60 },
		{ 294, 476, 60 },
		{ 294, 477, 60 },
		{ 294, 478, 60 },
		{ 294, 479, 60 },
		{ 294, 480, 60 },
		{ 294, 481, 60 },
		{ 294, 482, 60 },
		{ 294, 483, 60 },
		{ 294, 484, 60 },
		{ 294, 485, 60 },
		{ 294, 486, 60 },
		{ 294, 487, 60 },
		{ 294, 488, 60 },
		{ 294, 489, 60 },
		{ 294, 490, 60 },
		{ 294, 491, 60 },
		{ 294, 492, 60 },
		{ 294, 493, 60 },
		{ 294, 494, 60 },
		{ 294, 495, 60 },
		{ 294, 496, 60 },
		{ 294, 497, 60 },
		{ 294, 498, 60 },
		{ 294, 499, 60 },
		{ 294, 500, 60 },
		{ 294, 501, 60 },
		{ 294, 502, 60 },
		{ 294, 503, 60 },
		{ 294, 504, 60 },
		{ 294, 505, 60 },
		{ 294, 506, 60 },
		{ 294, 507, 60 },
		{ 294, 508, 60 },
		{ 294, 509, 60 },
		{ 294, 510, 60 },
		{ 294, 511, 60 },
		{ 294, 512, 60 },
		{ 295, 359, 60 },
		{ 295, 360, 60 },
		{ 295, 361, 60 },
		{ 295, 362, 60 },
		{ 295, 363, 60 },
		{ 295, 364, 60 },
		{ 295, 365, 60 },
		{ 295, 366, 60 },
		{ 295, 367, 60 },
		{ 295, 368, 60 },
		{ 295, 369, 60 },
		{ 295, 370, 60 },
		{ 295, 371, 60 },
		{ 295, 372, 60 },
		{ 295, 373, 60 },
		{ 295, 374, 60 },
		{ 295, 375, 60 },
		{ 295, 376, 60 },
		{ 295, 377, 60 },
		{ 295, 378, 60 },
		{ 295, 379, 60 },
		{ 295, 380, 60 },
		{ 295, 381, 60 },
		{ 295, 382, 60 },
		{ 295, 383, 60 },
		{ 295, 384, 60 },
		{ 295, 385, 60 },
		{ 295, 386, 60 },
		{ 295, 387, 60 },
		{ 295, 388, 60 },
		{ 295, 389, 60 },
		{ 295, 390, 60 },
		{ 295, 391, 60 },
		{ 295, 392, 60 },
		{ 295, 393, 60 },
		{ 295, 394, 60 },
		{ 295, 395, 60 },
		{ 295, 396, 60 },
		{ 295, 397, 60 },
		{ 295, 398, 60 },
		{ 295, 399, 60 },
		{ 295, 400, 60 },
		{ 295, 401, 60 },
		{ 295, 402, 60 },
		{ 295, 403, 60 },
		{ 295, 404, 60 },
		{ 295, 405, 60 },
		{ 295, 406, 60 },
		{ 295, 407, 60 },
		{ 295, 408, 60 },
		{ 295, 409, 60 },
		{ 295, 410, 60 },
		{ 295, 411, 60 },
		{ 295, 412, 60 },
		{ 295, 413, 60 },
		{ 295, 414, 60 },
		{ 295, 415, 60 },
		{ 295, 416, 60 },
		{ 295, 417, 60 },
		{ 295, 418, 60 },
		{ 295, 419, 60 },
		{ 295, 420, 60 },
		{ 295, 421, 60 },
		{ 295, 422, 60 },
		{ 295, 423, 60 },
		{ 295, 424, 60 },
		{ 295, 425, 60 },
		{ 295, 426, 60 },
		{ 295, 427, 60 },
		{ 295, 428, 60 },
		{ 295, 429, 60 },
		{ 295, 430, 60 },
		{ 295, 431, 60 },
		{ 295, 432, 60 },
		{ 295, 433, 60 },
		{ 295, 434, 60 },
		{ 295, 435, 60 },
		{ 295, 436, 60 },
		{ 295, 437, 60 },
		{ 295, 438, 60 },
		{ 295, 439, 60 },
		{ 295, 440, 60 },
		{ 295, 441, 60 },
		{ 295, 442, 60 },
		{ 295, 443, 60 },
		{ 295, 444, 60 },
		{ 295, 445, 60 },
		{ 295, 446, 60 },
		{ 295, 447, 60 },
		{ 295, 448, 60 },
		{ 295, 449, 60 },
		{ 295, 450, 60 },
		{ 295, 451, 60 },
		{ 295, 452, 60 },
		{ 295, 453, 60 },
		{ 295, 454, 60 },
		{ 295, 455, 60 },
		{ 295, 456, 60 },
		{ 295, 457, 60 },
		{ 295, 458, 60 },
		{ 295, 459, 60 },
		{ 295, 460, 60 },
		{ 295, 461, 60 },
		{ 295, 462, 60 },
		{ 295, 463, 60 },
		{ 295, 464, 60 },
		{ 295, 465, 60 },
		{ 295, 466, 60 },
		{ 295, 467, 60 },
		{ 295, 468, 60 },
		{ 295, 469, 60 },
		{ 295, 470, 60 },
		{ 295, 471, 60 },
		{ 295, 472, 60 },
		{ 295, 473, 60 },
		{ 295, 474, 60 },
		{ 295, 475, 60 },
		{ 295, 476, 60 },
		{ 295, 477, 60 },
		{ 295, 478, 60 },
		{ 295, 479, 60 },
		{ 295, 480, 60 },
		{ 295, 481, 60 },
		{ 295, 482, 60 },
		{ 295, 483, 60 },
		{ 295, 484, 60 },
		{ 295, 485, 60 },
		{ 295, 486, 60 },
		{ 295, 487, 60 },
		{ 295, 488, 60 },
		{ 295, 489, 60 },
		{ 295, 490, 60 },
		{ 295, 491, 60 },
		{ 295, 492, 60 },
		{ 295, 493, 60 },
		{ 295, 494, 60 },
		{ 295, 495, 60 },
		{ 295, 496, 60 },
		{ 295, 497, 60 },
		{ 295, 498, 60 },
		{ 295, 499, 60 },
		{ 295, 500, 60 },
		{ 295, 501, 60 },
		{ 295, 502, 60 },
		{ 295, 503, 60 },
		{ 295, 504, 60 },
		{ 295, 505, 60 },
		{ 295, 506, 60 },
		{ 295, 507, 60 },
		{ 295, 508, 60 },
		{ 295, 509, 60 },
		{ 295, 510, 60 },
		{ 295, 511, 60 },
		{ 295, 512, 60 },
		{ 296, 362, 60 },
		{ 296, 363, 60 },
		{ 296, 364, 60 },
		{ 296, 365, 60 },
		{ 296, 366, 60 },
		{ 296, 367, 60 },
		{ 296, 368, 60 },
		{ 296, 369, 60 },
		{ 296, 370, 60 },
		{ 296, 371, 60 },
		{ 296, 372, 60 },
		{ 296, 373, 60 },
		{ 296, 374, 60 },
		{ 296, 375, 60 },
		{ 296, 376, 60 },
		{ 296, 377, 60 },
		{ 296, 378, 60 },
		{ 296, 379, 60 },
		{ 296, 380, 60 },
		{ 296, 381, 60 },
		{ 296, 382, 60 },
		{ 296, 383, 60 },
		{ 296, 384, 60 },
		{ 296, 385, 60 },
		{ 296, 386, 60 },
		{ 296, 387, 60 },
		{ 296, 388, 60 },
		{ 296, 389, 60 },
		{ 296, 390, 60 },
		{ 296, 391, 60 },
		{ 296, 392, 60 },
		{ 296, 393, 60 },
		{ 296, 394, 60 },
		{ 296, 395, 60 },
		{ 296, 396, 60 },
		{ 296, 397, 60 },
		{ 296, 398, 60 },
		{ 296, 399, 60 },
		{ 296, 400, 60 },
		{ 296, 401, 60 },
		{ 296, 402, 60 },
		{ 296, 403, 60 },
		{ 296, 404, 60 },
		{ 296, 405, 60 },
		{ 296, 406, 60 },
		{ 296, 407, 60 },
		{ 296, 408, 60 },
		{ 296, 409, 60 },
		{ 296, 410, 60 },
		{ 296, 411, 60 },
		{ 296, 412, 60 },
		{ 296, 413, 60 },
		{ 296, 414, 60 },
		{ 296, 415, 60 },
		{ 296, 416, 60 },
		{ 296, 417, 60 },
		{ 296, 418, 60 },
		{ 296, 419, 60 },
		{ 296, 420, 60 },
		{ 296, 421, 60 },
		{ 296, 422, 60 },
		{ 296, 423, 60 },
		{ 296, 424, 60 },
		{ 296, 425, 60 },
		{ 296, 426, 60 },
		{ 296, 427, 60 },
		{ 296, 428, 60 },
		{ 296, 429, 60 },
		{ 296, 430, 60 },
		{ 296, 431, 60 },
		{ 296, 432, 60 },
		{ 296, 433, 60 },
		{ 296, 434, 60 },
		{ 296, 435, 60 },
		{ 296, 436, 60 },
		{ 296, 437, 60 },
		{ 296, 438, 60 },
		{ 296, 439, 60 },
		{ 296, 440, 60 },
		{ 296, 441, 60 },
		{ 296, 442, 60 },
		{ 296, 443, 60 },
		{ 296, 444, 60 },
		{ 296, 445, 60 },
		{ 296, 446, 60 },
		{ 296, 447, 60 },
		{ 296, 448, 60 },
		{ 296, 449, 60 },
		{ 296, 450, 60 },
		{ 296, 451, 60 },
		{ 296, 452, 60 },
		{ 296, 453, 60 },
		{ 296, 454, 60 },
		{ 296, 455, 60 },
		{ 296, 456, 60 },
		{ 296, 457, 60 },
		{ 296, 458, 60 },
		{ 296, 459, 60 },
		{ 296, 460, 60 },
		{ 296, 461, 60 },
		{ 296, 462, 60 },
		{ 296, 463, 60 },
		{ 296, 464, 60 },
		{ 296, 465, 60 },
		{ 296, 466, 60 },
		{ 296, 467, 60 },
		{ 296, 468, 60 },
		{ 296, 469, 60 },
		{ 296, 470, 60 },
		{ 296, 471, 60 },
		{ 296, 472, 60 },
		{ 296, 473, 60 },
		{ 296, 474, 60 },
		{ 296, 475, 60 },
		{ 296, 476, 60 },
		{ 296, 477, 60 },
		{ 296, 478, 60 },
		{ 296, 479, 60 },
		{ 296, 480, 60 },
		{ 296, 481, 60 },
		{ 296, 482, 60 },
		{ 296, 483, 60 },
		{ 296, 484, 60 },
		{ 296, 485, 60 },
		{ 296, 486, 60 },
		{ 296, 487, 60 },
		{ 296, 488, 60 },
		{ 296, 489, 60 },
		{ 296, 490, 60 },
		{ 296, 491, 60 },
		{ 296, 492, 60 },
		{ 296, 493, 60 },
		{ 296, 494, 60 },
		{ 296, 495, 60 },
		{ 296, 496, 60 },
		{ 296, 497, 60 },
		{ 296, 498, 60 },
		{ 296, 499, 60 },
		{ 296, 500, 60 },
		{ 296, 501, 60 },
		{ 296, 502, 60 },
		{ 296, 503, 60 },
		{ 296, 504, 60 },
		{ 296, 505, 60 },
		{ 296, 506, 60 },
		{ 296, 507, 60 },
		{ 296, 508, 60 },
		{ 296, 509, 60 },
		{ 296, 510, 60 },
		{ 296, 511, 60 },
		{ 296, 512, 60 },
		{ 297, 364, 60 },
		{ 297, 365, 60 },
		{ 297, 366, 60 },
		{ 297, 367, 60 },
		{ 297, 368, 60 },
		{ 297, 369, 60 },
		{ 297, 370, 60 },
		{ 297, 371, 60 },
		{ 297, 372, 60 },
		{ 297, 373, 60 },
		{ 297, 374, 60 },
		{ 297, 375, 60 },
		{ 297, 376, 60 },
		{ 297, 377, 60 },
		{ 297, 378, 60 },
		{ 297, 379, 60 },
		{ 297, 380, 60 },
		{ 297, 381, 60 },
		{ 297, 382, 60 },
		{ 297, 383, 60 },
		{ 297, 384, 60 },
		{ 297, 385, 60 },
		{ 297, 386, 60 },
		{ 297, 387, 60 },
		{ 297, 388, 60 },
		{ 297, 389, 60 },
		{ 297, 390, 60 },
		{ 297, 391, 60 },
		{ 297, 392, 60 },
		{ 297, 393, 60 },
		{ 297, 394, 60 },
		{ 297, 395, 60 },
		{ 297, 396, 60 },
		{ 297, 397, 60 },
		{ 297, 398, 60 },
		{ 297, 399, 60 },
		{ 297, 400, 60 },
		{ 297, 401, 60 },
		{ 297, 402, 60 },
		{ 297, 403, 60 },
		{ 297, 404, 60 },
		{ 297, 405, 60 },
		{ 297, 406, 60 },
		{ 297, 407, 60 },
		{ 297, 408, 60 },
		{ 297, 409, 60 },
		{ 297, 410, 60 },
		{ 297, 411, 60 },
		{ 297, 412, 60 },
		{ 297, 413, 60 },
		{ 297, 414, 60 },
		{ 297, 415, 60 },
		{ 297, 416, 60 },
		{ 297, 417, 60 },
		{ 297, 418, 60 },
		{ 297, 419, 60 },
		{ 297, 420, 60 },
		{ 297, 421, 60 },
		{ 297, 422, 60 },
		{ 297, 423, 60 },
		{ 297, 424, 60 },
		{ 297, 425, 60 },
		{ 297, 426, 60 },
		{ 297, 427, 60 },
		{ 297, 428, 60 },
		{ 297, 429, 60 },
		{ 297, 430, 60 },
		{ 297, 431, 60 },
		{ 297, 432, 60 },
		{ 297, 433, 60 },
		{ 297, 434, 60 },
		{ 297, 435, 60 },
		{ 297, 436, 60 },
		{ 297, 437, 60 },
		{ 297, 438, 60 },
		{ 297, 439, 60 },
		{ 297, 440, 60 },
		{ 297, 441, 60 },
		{ 297, 442, 60 },
		{ 297, 443, 60 },
		{ 297, 444, 60 },
		{ 297, 445, 60 },
		{ 297, 446, 60 },
		{ 297, 447, 60 },
		{ 297, 448, 60 },
		{ 297, 449, 60 },
		{ 297, 450, 60 },
		{ 297, 451, 60 },
		{ 297, 452, 60 },
		{ 297, 453, 60 },
		{ 297, 454, 60 },
		{ 297, 455, 60 },
		{ 297, 456, 60 },
		{ 297, 457, 60 },
		{ 297, 458, 60 },
		{ 297, 459, 60 },
		{ 297, 460, 60 },
		{ 297, 461, 60 },
		{ 297, 462, 60 },
		{ 297, 463, 60 },
		{ 297, 464, 60 },
		{ 297, 465, 60 },
		{ 297, 466, 60 },
		{ 297, 467, 60 },
		{ 297, 468, 60 },
		{ 297, 469, 60 },
		{ 297, 470, 60 },
		{ 297, 471, 60 },
		{ 297, 472, 60 },
		{ 297, 473, 60 },
		{ 297, 474, 60 },
		{ 297, 475, 60 },
		{ 297, 476, 60 },
		{ 297, 477, 60 },
		{ 297, 478, 60 },
		{ 297, 479, 60 },
		{ 297, 480, 60 },
		{ 297, 481, 60 },
		{ 297, 482, 60 },
		{ 297, 483, 60 },
		{ 297, 484, 60 },
		{ 297, 485, 60 },
		{ 297, 486, 60 },
		{ 297, 487, 60 },
		{ 297, 488, 60 },
		{ 297, 489, 60 },
		{ 297, 490, 60 },
		{ 297, 491, 60 },
		{ 297, 492, 60 },
		{ 297, 493, 60 },
		{ 297, 494, 60 },
		{ 297, 495, 60 },
		{ 297, 496, 60 },
		{ 297, 497, 60 },
		{ 297, 498, 60 },
		{ 297, 499, 60 },
		{ 297, 500, 60 },
		{ 297, 501, 60 },
		{ 297, 502, 60 },
		{ 297, 503, 60 },
		{ 297, 504, 60 },
		{ 297, 505, 60 },
		{ 297, 506, 60 },
		{ 297, 507, 60 },
		{ 297, 508, 60 },
		{ 297, 509, 60 },
		{ 297, 510, 60 },
		{ 297, 511, 60 },
		{ 297, 512, 60 },
		{ 298, 367, 60 },
		{ 298, 368, 60 },
		{ 298, 369, 60 },
		{ 298, 370, 60 },
		{ 298, 371, 60 },
		{ 298, 372, 60 },
		{ 298, 373, 60 },
		{ 298, 374, 60 },
		{ 298, 375, 60 },
		{ 298, 376, 60 },
		{ 298, 377, 60 },
		{ 298, 378, 60 },
		{ 298, 379, 60 },
		{ 298, 380, 60 },
		{ 298, 381, 60 },
		{ 298, 382, 60 },
		{ 298, 383, 60 },
		{ 298, 384, 60 },
		{ 298, 385, 60 },
		{ 298, 386, 60 },
		{ 298, 387, 60 },
		{ 298, 388, 60 },
		{ 298, 389, 60 },
		{ 298, 390, 60 },
		{ 298, 391, 60 },
		{ 298, 392, 60 },
		{ 298, 393, 60 },
		{ 298, 394, 60 },
		{ 298, 395, 60 },
		{ 298, 396, 60 },
		{ 298, 397, 60 },
		{ 298, 398, 60 },
		{ 298, 399, 60 },
		{ 298, 400, 60 },
		{ 298, 401, 60 },
		{ 298, 402, 60 },
		{ 298, 403, 60 },
		{ 298, 404, 60 },
		{ 298, 405, 60 },
		{ 298, 406, 60 },
		{ 298, 407, 60 },
		{ 298, 408, 60 },
		{ 298, 409, 60 },
		{ 298, 410, 60 },
		{ 298, 411, 60 },
		{ 298, 412, 60 },
		{ 298, 413, 60 },
		{ 298, 414, 60 },
		{ 298, 415, 60 },
		{ 298, 416, 60 },
		{ 298, 417, 60 },
		{ 298, 418, 60 },
		{ 298, 419, 60 },
		{ 298, 420, 60 },
		{ 298, 421, 60 },
		{ 298, 422, 60 },
		{ 298, 423, 60 },
		{ 298, 424, 60 },
		{ 298, 425, 60 },
		{ 298, 426, 60 },
		{ 298, 427, 60 },
		{ 298, 428, 60 },
		{ 298, 429, 60 },
		{ 298, 430, 60 },
		{ 298, 431, 60 },
		{ 298, 432, 60 },
		{ 298, 433, 60 },
		{ 298, 434, 60 },
		{ 298, 435, 60 },
		{ 298, 436, 60 },
		{ 298, 437, 60 },
		{ 298, 438, 60 },
		{ 298, 439, 60 },
		{ 298, 440, 60 },
		{ 298, 441, 60 },
		{ 298, 442, 60 },
		{ 298, 443, 60 },
		{ 298, 444, 60 },
		{ 298, 445, 60 },
		{ 298, 446, 60 },
		{ 298, 447, 60 },
		{ 298, 448, 60 },
		{ 298, 449, 60 },
		{ 298, 450, 60 },
		{ 298, 451, 60 },
		{ 298, 452, 60 },
		{ 298, 453, 60 },
		{ 298, 454, 60 },
		{ 298, 455, 60 },
		{ 298, 456, 60 },
		{ 298, 457, 60 },
		{ 298, 458, 60 },
		{ 298, 459, 60 },
		{ 298, 460, 60 },
		{ 298, 461, 60 },
		{ 298, 462, 60 },
		{ 298, 463, 60 },
		{ 298, 464, 60 },
		{ 298, 465, 60 },
		{ 298, 466, 60 },
		{ 298, 467, 60 },
		{ 298, 468, 60 },
		{ 298, 469, 60 },
		{ 298, 470, 60 },
		{ 298, 471, 60 },
		{ 298, 472, 60 },
		{ 298, 473, 60 },
		{ 298, 474, 60 },
		{ 298, 475, 60 },
		{ 298, 476, 60 },
		{ 298, 477, 60 },
		{ 298, 478, 60 },
		{ 298, 479, 60 },
		{ 298, 480, 60 },
		{ 298, 481, 60 },
		{ 298, 482, 60 },
		{ 298, 483, 60 },
		{ 298, 484, 60 },
		{ 298, 485, 60 },
		{ 298, 486, 60 },
		{ 298, 487, 60 },
		{ 298, 488, 60 },
		{ 298, 489, 60 },
		{ 298, 490, 60 },
		{ 298, 491, 60 },
		{ 298, 492, 60 },
		{ 298, 493, 60 },
		{ 298, 494, 60 },
		{ 298, 495, 60 },
		{ 298, 496, 60 },
		{ 298, 497, 60 },
		{ 298, 498, 60 },
		{ 298, 499, 60 },
		{ 298, 500, 60 },
		{ 298, 501, 60 },
		{ 298, 502, 60 },
		{ 298, 503, 60 },
		{ 298, 504, 60 },
		{ 298, 505, 60 },
		{ 298, 506, 60 },
		{ 298, 507, 60 },
		{ 298, 508, 60 },
		{ 298, 509, 60 },
		{ 298, 510, 60 },
		{ 298, 511, 60 },
		{ 298, 512, 60 },
		{ 299, 370, 60 },
		{ 299, 371, 60 },
		{ 299, 372, 60 },
		{ 299, 373, 60 },
		{ 299, 374, 60 },
		{ 299, 375, 60 },
		{ 299, 376, 60 },
		{ 299, 377, 60 },
		{ 299, 378, 60 },
		{ 299, 379, 60 },
		{ 299, 380, 60 },
		{ 299, 381, 60 },
		{ 299, 382, 60 },
		{ 299, 383, 60 },
		{ 299, 384, 60 },
		{ 299, 385, 60 },
		{ 299, 386, 60 },
		{ 299, 387, 60 },
		{ 299, 388, 60 },
		{ 299, 389, 60 },
		{ 299, 390, 60 },
		{ 299, 391, 60 },
		{ 299, 392, 60 },
		{ 299, 393, 60 },
		{ 299, 394, 60 },
		{ 299, 395, 60 },
		{ 299, 396, 60 },
		{ 299, 397, 60 },
		{ 299, 398, 60 },
		{ 299, 399, 60 },
		{ 299, 400, 60 },
		{ 299, 401, 60 },
		{ 299, 402, 60 },
		{ 299, 403, 60 },
		{ 299, 404, 60 },
		{ 299, 405, 60 },
		{ 299, 406, 60 },
		{ 299, 407, 60 },
		{ 299, 408, 60 },
		{ 299, 409, 60 },
		{ 299, 410, 60 },
		{ 299, 411, 60 },
		{ 299, 412, 60 },
		{ 299, 413, 60 },
		{ 299, 414, 60 },
		{ 299, 415, 60 },
		{ 299, 416, 60 },
		{ 299, 417, 60 },
		{ 299, 418, 60 },
		{ 299, 419, 60 },
		{ 299, 420, 60 },
		{ 299, 421, 60 },
		{ 299, 422, 60 },
		{ 299, 423, 60 },
		{ 299, 424, 60 },
		{ 299, 425, 60 },
		{ 299, 426, 60 },
		{ 299, 427, 60 },
		{ 299, 428, 60 },
		{ 299, 429, 60 },
		{ 299, 430, 60 },
		{ 299, 431, 60 },
		{ 299, 432, 60 },
		{ 299, 433, 60 },
		{ 299, 434, 60 },
		{ 299, 435, 60 },
		{ 299, 436, 60 },
		{ 299, 437, 60 },
		{ 299, 438, 60 },
		{ 299, 439, 60 },
		{ 299, 440, 60 },
		{ 299, 441, 60 },
		{ 299, 442, 60 },
		{ 299, 443, 60 },
		{ 299, 444, 60 },
		{ 299, 445, 60 },
		{ 299, 446, 60 },
		{ 299, 447, 60 },
		{ 299, 448, 60 },
		{ 299, 449, 60 },
		{ 299, 450, 60 },
		{ 299, 451, 60 },
		{ 299, 452, 60 },
		{ 299, 453, 60 },
		{ 299, 454, 60 },
		{ 299, 455, 60 },
		{ 299, 456, 60 },
		{ 299, 457, 60 },
		{ 299, 458, 60 },
		{ 299, 459, 60 },
		{ 299, 460, 60 },
		{ 299, 461, 60 },
		{ 299, 462, 60 },
		{ 299, 463, 60 },
		{ 299, 464, 60 },
		{ 299, 465, 60 },
		{ 299, 466, 60 },
		{ 299, 467, 60 },
		{ 299, 468, 60 },
		{ 299, 469, 60 },
		{ 299, 470, 60 },
		{ 299, 471, 60 },
		{ 299, 472, 60 },
		{ 299, 473, 60 },
		{ 299, 474, 60 },
		{ 299, 475, 60 },
		{ 299, 476, 60 },
		{ 299, 477, 60 },
		{ 299, 478, 60 },
		{ 299, 479, 60 },
		{ 299, 480, 60 },
		{ 299, 481, 60 },
		{ 299, 482, 60 },
		{ 299, 483, 60 },
		{ 299, 484, 60 },
		{ 299, 485, 60 },
		{ 299, 486, 60 },
		{ 299, 487, 60 },
		{ 299, 488, 60 },
		{ 299, 489, 60 },
		{ 299, 490, 60 },
		{ 299, 491, 60 },
		{ 299, 492, 60 },
		{ 299, 493, 60 },
		{ 299, 494, 60 },
		{ 299, 495, 60 },
		{ 299, 496, 60 },
		{ 299, 497, 60 },
		{ 299, 498, 60 },
		{ 299, 499, 60 },
		{ 299, 500, 60 },
		{ 299, 501, 60 },
		{ 299, 502, 60 },
		{ 299, 503, 60 },
		{ 299, 504, 60 },
		{ 299, 505, 60 },
		{ 299, 506, 60 },
		{ 299, 507, 60 },
		{ 299, 508, 60 },
		{ 299, 509, 60 },
		{ 299, 510, 60 },
		{ 299, 511, 60 },
		{ 299, 512, 60 },
		{ 300, 372, 60 },
		{ 300, 373, 60 },
		{ 300, 374, 60 },
		{ 300, 375, 60 },
		{ 300, 376, 60 },
		{ 300, 377, 60 },
		{ 300, 378, 60 },
		{ 300, 379, 60 },
		{ 300, 380, 60 },
		{ 300, 381, 60 },
		{ 300, 382, 60 },
		{ 300, 383, 60 },
		{ 300, 384, 60 },
		{ 300, 385, 60 },
		{ 300, 386, 60 },
		{ 300, 387, 60 },
		{ 300, 388, 60 },
		{ 300, 389, 60 },
		{ 300, 390, 60 },
		{ 300, 391, 60 },
		{ 300, 392, 60 },
		{ 300, 393, 60 },
		{ 300, 394, 60 },
		{ 300, 395, 60 },
		{ 300, 396, 60 },
		{ 300, 397, 60 },
		{ 300, 398, 60 },
		{ 300, 399, 60 },
		{ 300, 400, 60 },
		{ 300, 401, 60 },
		{ 300, 402, 60 },
		{ 300, 403, 60 },
		{ 300, 404, 60 },
		{ 300, 405, 60 },
		{ 300, 406, 60 },
		{ 300, 407, 60 },
		{ 300, 408, 60 },
		{ 300, 409, 60 },
		{ 300, 410, 60 },
		{ 300, 411, 60 },
		{ 300, 412, 60 },
		{ 300, 413, 60 },
		{ 300, 414, 60 },
		{ 300, 415, 60 },
		{ 300, 416, 60 },
		{ 300, 417, 60 },
		{ 300, 418, 60 },
		{ 300, 419, 60 },
		{ 300, 420, 60 },
		{ 300, 421, 60 },
		{ 300, 422, 60 },
		{ 300, 423, 60 },
		{ 300, 424, 60 },
		{ 300, 425, 60 },
		{ 300, 426, 60 },
		{ 300, 427, 60 },
		{ 300, 428, 60 },
		{ 300, 429, 60 },
		{ 300, 430, 60 },
		{ 300, 431, 60 },
		{ 300, 432, 60 },
		{ 300, 433, 60 },
		{ 300, 434, 60 },
		{ 300, 435, 60 },
		{ 300, 436, 60 },
		{ 300, 437, 60 },
		{ 300, 438, 60 },
		{ 300, 439, 60 },
		{ 300, 440, 60 },
		{ 300, 441, 60 },
		{ 300, 442, 60 },
		{ 300, 443, 60 },
		{ 300, 444, 60 },
		{ 300, 445, 60 },
		{ 300, 446, 60 },
		{ 300, 447, 60 },
		{ 300, 448, 60 },
		{ 300, 449, 60 },
		{ 300, 450, 60 },
		{ 300, 451, 60 },
		{ 300, 452, 60 },
		{ 300, 453, 60 },
		{ 300, 454, 60 },
		{ 300, 455, 60 },
		{ 300, 456, 60 },
		{ 300, 457, 60 },
		{ 300, 458, 60 },
		{ 300, 459, 60 },
		{ 300, 460, 60 },
		{ 300, 461, 60 },
		{ 300, 462, 60 },
		{ 300, 463, 60 },
		{ 300, 464, 60 },
		{ 300, 465, 60 },
		{ 300, 466, 60 },
		{ 300, 467, 60 },
		{ 300, 468, 60 },
		{ 300, 469, 60 },
		{ 300, 470, 60 },
		{ 300, 471, 60 },
		{ 300, 472, 60 },
		{ 300, 473, 60 },
		{ 300, 474, 60 },
		{ 300, 475, 60 },
		{ 300, 476, 60 },
		{ 300, 477, 60 },
		{ 300, 478, 60 },
		{ 300, 479, 60 },
		{ 300, 480, 60 },
		{ 300, 481, 60 },
		{ 300, 482, 60 },
		{ 300, 483, 60 },
		{ 300, 484, 60 },
		{ 300, 485, 60 },
		{ 300, 486, 60 },
		{ 300, 487, 60 },
		{ 300, 488, 60 },
		{ 300, 489, 60 },
		{ 300, 490, 60 },
		{ 300, 491, 60 },
		{ 300, 492, 60 },
		{ 300, 493, 60 },
		{ 300, 494, 60 },
		{ 300, 495, 60 },
		{ 300, 496, 60 },
		{ 300, 497, 60 },
		{ 300, 498, 60 },
		{ 300, 499, 60 },
		{ 300, 500, 60 },
		{ 300, 501, 60 },
		{ 300, 502, 60 },
		{ 300, 503, 60 },
		{ 300, 504, 60 },
		{ 300, 505, 60 },
		{ 300, 506, 60 },
		{ 300, 507, 60 },
		{ 300, 508, 60 },
		{ 300, 509, 60 },
		{ 300, 510, 60 },
		{ 300, 511, 60 },
		{ 300, 512, 60 },
		{ 301, 375, 60 },
		{ 301, 376, 60 },
		{ 301, 377, 60 },
		{ 301, 378, 60 },
		{ 301, 379, 60 },
		{ 301, 380, 60 },
		{ 301, 381, 60 },
		{ 301, 382, 60 },
		{ 301, 383, 60 },
		{ 301, 384, 60 },
		{ 301, 385, 60 },
		{ 301, 386, 60 },
		{ 301, 387, 60 },
		{ 301, 388, 60 },
		{ 301, 389, 60 },
		{ 301, 390, 60 },
		{ 301, 391, 60 },
		{ 301, 392, 60 },
		{ 301, 393, 60 },
		{ 301, 394, 60 },
		{ 301, 395, 60 },
		{ 301, 396, 60 },
		{ 301, 397, 60 },
		{ 301, 398, 60 },
		{ 301, 399, 60 },
		{ 301, 400, 60 },
		{ 301, 401, 60 },
		{ 301, 402, 60 },
		{ 301, 403, 60 },
		{ 301, 404, 60 },
		{ 301, 405, 60 },
		{ 301, 406, 60 },
		{ 301, 407, 60 },
		{ 301, 408, 60 },
		{ 301, 409, 60 },
		{ 301, 410, 60 },
		{ 301, 411, 60 },
		{ 301, 412, 60 },
		{ 301, 413, 60 },
		{ 301, 414, 60 },
		{ 301, 415, 60 },
		{ 301, 416, 60 },
		{ 301, 417, 60 },
		{ 301, 418, 60 },
		{ 301, 419, 60 },
		{ 301, 420, 60 },
		{ 301, 421, 60 },
		{ 301, 422, 60 },
		{ 301, 423, 60 },
		{ 301, 424, 60 },
		{ 301, 425, 60 },
		{ 301, 426, 60 },
		{ 301, 427, 60 },
		{ 301, 428, 60 },
		{ 301, 429, 60 },
		{ 301, 430, 60 },
		{ 301, 431, 60 },
		{ 301, 432, 60 },
		{ 301, 433, 60 },
		{ 301, 434, 60 },
		{ 301, 435, 60 },
		{ 301, 436, 60 },
		{ 301, 437, 60 },
		{ 301, 438, 60 },
		{ 301, 439, 60 },
		{ 301, 440, 60 },
		{ 301, 441, 60 },
		{ 301, 442, 60 },
		{ 301, 443, 60 },
		{ 301, 444, 60 },
		{ 301, 445, 60 },
		{ 301, 446, 60 },
		{ 301, 447, 60 },
		{ 301, 448, 60 },
		{ 301, 449, 60 },
		{ 301, 450, 60 },
		{ 301, 451, 60 },
		{ 301, 452, 60 },
		{ 301, 453, 60 },
		{ 301, 454, 60 },
		{ 301, 455, 60 },
		{ 301, 456, 60 },
		{ 301, 457, 60 },
		{ 301, 458, 60 },
		{ 301, 459, 60 },
		{ 301, 460, 60 },
		{ 301, 461, 60 },
		{ 301, 462, 60 },
		{ 301, 463, 60 },
		{ 301, 464, 60 },
		{ 301, 465, 60 },
		{ 301, 466, 60 },
		{ 301, 467, 60 },
		{ 301, 468, 60 },
		{ 301, 469, 60 },
		{ 301, 470, 60 },
		{ 301, 471, 60 },
		{ 301, 472, 60 },
		{ 301, 473, 60 },
		{ 301, 474, 60 },
		{ 301, 475, 60 },
		{ 301, 476, 60 },
		{ 301, 477, 60 },
		{ 301, 478, 60 },
		{ 301, 479, 60 },
		{ 301, 480, 60 },
		{ 301, 481, 60 },
		{ 301, 482, 60 },
		{ 301, 483, 60 },
		{ 301, 484, 60 },
		{ 301, 485, 60 },
		{ 301, 486, 60 },
		{ 301, 487, 60 },
		{ 301, 488, 60 },
		{ 301, 489, 60 },
		{ 301, 490, 60 },
		{ 301, 491, 60 },
		{ 301, 492, 60 },
		{ 301, 493, 60 },
		{ 301, 494, 60 },
		{ 301, 495, 60 },
		{ 301, 496, 60 },
		{ 301, 497, 60 },
		{ 301, 498, 60 },
		{ 301, 499, 60 },
		{ 301, 500, 60 },
		{ 301, 501, 60 },
		{ 301, 502, 60 },
		{ 301, 503, 60 },
		{ 301, 504, 60 },
		{ 301, 505, 60 },
		{ 301, 506, 60 },
		{ 301, 507, 60 },
		{ 301, 508, 60 },
		{ 301, 509, 60 },
		{ 301, 510, 60 },
		{ 301, 511, 60 },
		{ 301, 512, 60 },
		{ 302, 378, 60 },
		{ 302, 379, 60 },
		{ 302, 380, 60 },
		{ 302, 381, 60 },
		{ 302, 382, 60 },
		{ 302, 383, 60 },
		{ 302, 384, 60 },
		{ 302, 385, 60 },
		{ 302, 386, 60 },
		{ 302, 387, 60 },
		{ 302, 388, 60 },
		{ 302, 389, 60 },
		{ 302, 390, 60 },
		{ 302, 391, 60 },
		{ 302, 392, 60 },
		{ 302, 393, 60 },
		{ 302, 394, 60 },
		{ 302, 395, 60 },
		{ 302, 396, 60 },
		{ 302, 397, 60 },
		{ 302, 398, 60 },
		{ 302, 399, 60 },
		{ 302, 400, 60 },
		{ 302, 401, 60 },
		{ 302, 402, 60 },
		{ 302, 403, 60 },
		{ 302, 404, 60 },
		{ 302, 405, 60 },
		{ 302, 406, 60 },
		{ 302, 407, 60 },
		{ 302, 408, 60 },
		{ 302, 409, 60 },
		{ 302, 410, 60 },
		{ 302, 411, 60 },
		{ 302, 412, 60 },
		{ 302, 413, 60 },
		{ 302, 414, 60 },
		{ 302, 415, 60 },
		{ 302, 416, 60 },
		{ 302, 417, 60 },
		{ 302, 418, 60 },
		{ 302, 419, 60 },
		{ 302, 420, 60 },
		{ 302, 421, 60 },
		{ 302, 422, 60 },
		{ 302, 423, 60 },
		{ 302, 424, 60 },
		{ 302, 425, 60 },
		{ 302, 426, 60 },
		{ 302, 427, 60 },
		{ 302, 428, 60 },
		{ 302, 429, 60 },
		{ 302, 430, 60 },
		{ 302, 431, 60 },
		{ 302, 432, 60 },
		{ 302, 433, 60 },
		{ 302, 434, 60 },
		{ 302, 435, 60 },
		{ 302, 436, 60 },
		{ 302, 437, 60 },
		{ 302, 438, 60 },
		{ 302, 439, 60 },
		{ 302, 440, 60 },
		{ 302, 441, 60 },
		{ 302, 442, 60 },
		{ 302, 443, 60 },
		{ 302, 444, 60 },
		{ 302, 445, 60 },
		{ 302, 446, 60 },
		{ 302, 447, 60 },
		{ 302, 448, 60 },
		{ 302, 449, 60 },
		{ 302, 450, 60 },
		{ 302, 451, 60 },
		{ 302, 452, 60 },
		{ 302, 453, 60 },
		{ 302, 454, 60 },
		{ 302, 455, 60 },
		{ 302, 456, 60 },
		{ 302, 457, 60 },
		{ 302, 458, 60 },
		{ 302, 459, 60 },
		{ 302, 460, 60 },
		{ 302, 461, 60 },
		{ 302, 462, 60 },
		{ 302, 463, 60 },
		{ 302, 464, 60 },
		{ 302, 465, 60 },
		{ 302, 466, 60 },
		{ 302, 467, 60 },
		{ 302, 468, 60 },
		{ 302, 469, 60 },
		{ 302, 470, 60 },
		{ 302, 471, 60 },
		{ 302, 472, 60 },
		{ 302, 473, 60 },
		{ 302, 474, 60 },
		{ 302, 475, 60 },
		{ 302, 476, 60 },
		{ 302, 477, 60 },
		{ 302, 478, 60 },
		{ 302, 479, 60 },
		{ 302, 480, 60 },
		{ 302, 481, 60 },
		{ 302, 482, 60 },
		{ 302, 483, 60 },
		{ 302, 484, 60 },
		{ 302, 485, 60 },
		{ 302, 486, 60 },
		{ 302, 487, 60 },
		{ 302, 488, 60 },
		{ 302, 489, 60 },
		{ 302, 490, 60 },
		{ 302, 491, 60 },
		{ 302, 492, 60 },
		{ 302, 493, 60 },
		{ 302, 494, 60 },
		{ 302, 495, 60 },
		{ 302, 496, 60 },
		{ 302, 497, 60 },
		{ 302, 498, 60 },
		{ 302, 499, 60 },
		{ 302, 500, 60 },
		{ 302, 501, 60 },
		{ 302, 502, 60 },
		{ 302, 503, 60 },
		{ 302, 504, 60 },
		{ 302, 505, 60 },
		{ 302, 506, 60 },
		{ 302, 507, 60 },
		{ 302, 508, 60 },
		{ 302, 509, 60 },
		{ 302, 510, 60 },
		{ 302, 511, 60 },
		{ 302, 512, 60 },
		{ 303, 380, 60 },
		{ 303, 381, 60 },
		{ 303, 382, 60 },
		{ 303, 383, 60 },
		{ 303, 384, 60 },
		{ 303, 385, 60 },
		{ 303, 386, 60 },
		{ 303, 387, 60 },
		{ 303, 388, 60 },
		{ 303, 389, 60 },
		{ 303, 390, 60 },
		{ 303, 391, 60 },
		{ 303, 392, 60 },
		{ 303, 393, 60 },
		{ 303, 394, 60 },
		{ 303, 395, 60 },
		{ 303, 396, 60 },
		{ 303, 397, 60 },
		{ 303, 398, 60 },
		{ 303, 399, 60 },
		{ 303, 400, 60 },
		{ 303, 401, 60 },
		{ 303, 402, 60 },
		{ 303, 403, 60 },
		{ 303, 404, 60 },
		{ 303, 405, 60 },
		{ 303, 406, 60 },
		{ 303, 407, 60 },
		{ 303, 408, 60 },
		{ 303, 409, 60 },
		{ 303, 410, 60 },
		{ 303, 411, 60 },
		{ 303, 412, 60 },
		{ 303, 413, 60 },
		{ 303, 414, 60 },
		{ 303, 415, 60 },
		{ 303, 416, 60 },
		{ 303, 417, 60 },
		{ 303, 418, 60 },
		{ 303, 419, 60 },
		{ 303, 420, 60 },
		{ 303, 421, 60 },
		{ 303, 422, 60 },
		{ 303, 423, 60 },
		{ 303, 424, 60 },
		{ 303, 425, 60 },
		{ 303, 426, 60 },
		{ 303, 427, 60 },
		{ 303, 428, 60 },
		{ 303, 429, 60 },
		{ 303, 430, 60 },
		{ 303, 431, 60 },
		{ 303, 432, 60 },
		{ 303, 433, 60 },
		{ 303, 434, 60 },
		{ 303, 435, 60 },
		{ 303, 436, 60 },
		{ 303, 437, 60 },
		{ 303, 438, 60 },
		{ 303, 439, 60 },
		{ 303, 440, 60 },
		{ 303, 441, 60 },
		{ 303, 442, 60 },
		{ 303, 443, 60 },
		{ 303, 444, 60 },
		{ 303, 445, 60 },
		{ 303, 446, 60 },
		{ 303, 447, 60 },
		{ 303, 448, 60 },
		{ 303, 449, 60 },
		{ 303, 450, 60 },
		{ 303, 451, 60 },
		{ 303, 452, 60 },
		{ 303, 453, 60 },
		{ 303, 454, 60 },
		{ 303, 455, 60 },
		{ 303, 456, 60 },
		{ 303, 457, 60 },
		{ 303, 458, 60 },
		{ 303, 459, 60 },
		{ 303, 460, 60 },
		{ 303, 461, 60 },
		{ 303, 462, 60 },
		{ 303, 463, 60 },
		{ 303, 464, 60 },
		{ 303, 465, 60 },
		{ 303, 466, 60 },
		{ 303, 467, 60 },
		{ 303, 468, 60 },
		{ 303, 469, 60 },
		{ 303, 470, 60 },
		{ 303, 471, 60 },
		{ 303, 472, 60 },
		{ 303, 473, 60 },
		{ 303, 474, 60 },
		{ 303, 475, 60 },
		{ 303, 476, 60 },
		{ 303, 477, 60 },
		{ 303, 478, 60 },
		{ 303, 479, 60 },
		{ 303, 480, 60 },
		{ 303, 481, 60 },
		{ 303, 482, 60 },
		{ 303, 483, 60 },
		{ 303, 484, 60 },
		{ 303, 485, 60 },
		{ 303, 486, 60 },
		{ 303, 487, 60 },
		{ 303, 488, 60 },
		{ 303, 489, 60 },
		{ 303, 490, 60 },
		{ 303, 491, 60 },
		{ 303, 492, 60 },
		{ 303, 493, 60 },
		{ 303, 494, 60 },
		{ 303, 495, 60 },
		{ 303, 496, 60 },
		{ 303, 497, 60 },
		{ 303, 498, 60 },
		{ 303, 499, 60 },
		{ 303, 500, 60 },
		{ 303, 501, 60 },
		{ 303, 502, 60 },
		{ 303, 503, 60 },
		{ 303, 504, 60 },
		{ 303, 505, 60 },
		{ 303, 506, 60 },
		{ 303, 507, 60 },
		{ 303, 508, 60 },
		{ 303, 509, 60 },
		{ 303, 510, 60 },
		{ 303, 511, 60 },
		{ 303, 512, 60 },
		{ 304, 383, 60 },
		{ 304, 384, 60 },
		{ 304, 385, 60 },
		{ 304, 386, 60 },
		{ 304, 387, 60 },
		{ 304, 388, 60 },
		{ 304, 389, 60 },
		{ 304, 390, 60 },
		{ 304, 391, 60 },
		{ 304, 392, 60 },
		{ 304, 393, 60 },
		{ 304, 394, 60 },
		{ 304, 395, 60 },
		{ 304, 396, 60 },
		{ 304, 397, 60 },
		{ 304, 398, 60 },
		{ 304, 399, 60 },
		{ 304, 400, 60 },
		{ 304, 401, 60 },
		{ 304, 402, 60 },
		{ 304, 403, 60 },
		{ 304, 404, 60 },
		{ 304, 405, 60 },
		{ 304, 406, 60 },
		{ 304, 407, 60 },
		{ 304, 408, 60 },
		{ 304, 409, 60 },
		{ 304, 410, 60 },
		{ 304, 411, 60 },
		{ 304, 412, 60 },
		{ 304, 413, 60 },
		{ 304, 414, 60 },
		{ 304, 415, 60 },
		{ 304, 416, 60 },
		{ 304, 417, 60 },
		{ 304, 418, 60 },
		{ 304, 419, 60 },
		{ 304, 420, 60 },
		{ 304, 421, 60 },
		{ 304, 422, 60 },
		{ 304, 423, 60 },
		{ 304, 424, 60 },
		{ 304, 425, 60 },
		{ 304, 426, 60 },
		{ 304, 427, 60 },
		{ 304, 428, 60 },
		{ 304, 429, 60 },
		{ 304, 430, 60 },
		{ 304, 431, 60 },
		{ 304, 432, 60 },
		{ 304, 433, 60 },
		{ 304, 434, 60 },
		{ 304, 435, 60 },
		{ 304, 436, 60 },
		{ 304, 437, 60 },
		{ 304, 438, 60 },
		{ 304, 439, 60 },
		{ 304, 440, 60 },
		{ 304, 441, 60 },
		{ 304, 442, 60 },
		{ 304, 443, 60 },
		{ 304, 444, 60 },
		{ 304, 445, 60 },
		{ 304, 446, 60 },
		{ 304, 447, 60 },
		{ 304, 448, 60 },
		{ 304, 449, 60 },
		{ 304, 450, 60 },
		{ 304, 451, 60 },
		{ 304, 452, 60 },
		{ 304, 453, 60 },
		{ 304, 454, 60 },
		{ 304, 455, 60 },
		{ 304, 456, 60 },
		{ 304, 457, 60 },
		{ 304, 458, 60 },
		{ 304, 459, 60 },
		{ 304, 460, 60 },
		{ 304, 461, 60 },
		{ 304, 462, 60 },
		{ 304, 463, 60 },
		{ 304, 464, 60 },
		{ 304, 465, 60 },
		{ 304, 466, 60 },
		{ 304, 467, 60 },
		{ 304, 468, 60 },
		{ 304, 469, 60 },
		{ 304, 470, 60 },
		{ 304, 471, 60 },
		{ 304, 472, 60 },
		{ 304, 473, 60 },
		{ 304, 474, 60 },
		{ 304, 475, 60 },
		{ 304, 476, 60 },
		{ 304, 477, 60 },
		{ 304, 478, 60 },
		{ 304, 479, 60 },
		{ 304, 480, 60 },
		{ 304, 481, 60 },
		{ 304, 482, 60 },
		{ 304, 483, 60 },
		{ 304, 484, 60 },
		{ 304, 485, 60 },
		{ 304, 486, 60 },
		{ 304, 487, 60 },
		{ 304, 488, 60 },
		{ 304, 489, 60 },
		{ 304, 490, 60 },
		{ 304, 491, 60 },
		{ 304, 492, 60 },
		{ 304, 493, 60 },
		{ 304, 494, 60 },
		{ 304, 495, 60 },
		{ 304, 496, 60 },
		{ 304, 497, 60 },
		{ 304, 498, 60 },
		{ 304, 499, 60 },
		{ 304, 500, 60 },
		{ 304, 501, 60 },
		{ 304, 502, 60 },
		{ 304, 503, 60 },
		{ 304, 504, 60 },
		{ 304, 505, 60 },
		{ 304, 506, 60 },
		{ 304, 507, 60 },
		{ 304, 508, 60 },
		{ 304, 509, 60 },
		{ 304, 510, 60 },
		{ 304, 511, 60 },
		{ 304, 512, 60 },
		{ 305, 386, 60 },
		{ 305, 387, 60 },
		{ 305, 388, 60 },
		{ 305, 389, 60 },
		{ 305, 390, 60 },
		{ 305, 391, 60 },
		{ 305, 392, 60 },
		{ 305, 393, 60 },
		{ 305, 394, 60 },
		{ 305, 395, 60 },
		{ 305, 396, 60 },
		{ 305, 397, 60 },
		{ 305, 398, 60 },
		{ 305, 399, 60 },
		{ 305, 400, 60 },
		{ 305, 401, 60 },
		{ 305, 402, 60 },
		{ 305, 403, 60 },
		{ 305, 404, 60 },
		{ 305, 405, 60 },
		{ 305, 406, 60 },
		{ 305, 407, 60 },
		{ 305, 408, 60 },
		{ 305, 409, 60 },
		{ 305, 410, 60 },
		{ 305, 411, 60 },
		{ 305, 412, 60 },
		{ 305, 413, 60 },
		{ 305, 414, 60 },
		{ 305, 415, 60 },
		{ 305, 416, 60 },
		{ 305, 417, 60 },
		{ 305, 418, 60 },
		{ 305, 419, 60 },
		{ 305, 420, 60 },
		{ 305, 421, 60 },
		{ 305, 422, 60 },
		{ 305, 423, 60 },
		{ 305, 424, 60 },
		{ 305, 425, 60 },
		{ 305, 426, 60 },
		{ 305, 427, 60 },
		{ 305, 428, 60 },
		{ 305, 429, 60 },
		{ 305, 430, 60 },
		{ 305, 431, 60 },
		{ 305, 432, 60 },
		{ 305, 433, 60 },
		{ 305, 434, 60 },
		{ 305, 435, 60 },
		{ 305, 436, 60 },
		{ 305, 437, 60 },
		{ 305, 438, 60 },
		{ 305, 439, 60 },
		{ 305, 440, 60 },
		{ 305, 441, 60 },
		{ 305, 442, 60 },
		{ 305, 443, 60 },
		{ 305, 444, 60 },
		{ 305, 445, 60 },
		{ 305, 446, 60 },
		{ 305, 447, 60 },
		{ 305, 448, 60 },
		{ 305, 449, 60 },
		{ 305, 450, 60 },
		{ 305, 451, 60 },
		{ 305, 452, 60 },
		{ 305, 453, 60 },
		{ 305, 454, 60 },
		{ 305, 455, 60 },
		{ 305, 456, 60 },
		{ 305, 457, 60 },
		{ 305, 458, 60 },
		{ 305, 459, 60 },
		{ 305, 460, 60 },
		{ 305, 461, 60 },
		{ 305, 462, 60 },
		{ 305, 463, 60 },
		{ 305, 464, 60 },
		{ 305, 465, 60 },
		{ 305, 466, 60 },
		{ 305, 467, 60 },
		{ 305, 468, 60 },
		{ 305, 469, 60 },
		{ 305, 470, 60 },
		{ 305, 471, 60 },
		{ 305, 472, 60 },
		{ 305, 473, 60 },
		{ 305, 474, 60 },
		{ 305, 475, 60 },
		{ 305, 476, 60 },
		{ 305, 477, 60 },
		{ 305, 478, 60 },
		{ 305, 479, 60 },
		{ 305, 480, 60 },
		{ 305, 481, 60 },
		{ 305, 482, 60 },
		{ 305, 483, 60 },
		{ 305, 484, 60 },
		{ 305, 485, 60 },
		{ 305, 486, 60 },
		{ 305, 487, 60 },
		{ 305, 488, 60 },
		{ 305, 489, 60 },
		{ 305, 490, 60 },
		{ 305, 491, 60 },
		{ 305, 492, 60 },
		{ 305, 493, 60 },
		{ 305, 494, 60 },
		{ 305, 495, 60 },
		{ 305, 496, 60 },
		{ 305, 497, 60 },
		{ 305, 498, 60 },
		{ 305, 499, 60 },
		{ 305, 500, 60 },
		{ 305, 501, 60 },
		{ 305, 502, 60 },
		{ 305, 503, 60 },
		{ 305, 504, 60 },
		{ 305, 505, 60 },
		{ 305, 506, 60 },
		{ 305, 507, 60 },
		{ 305, 508, 60 },
		{ 305, 509, 60 },
		{ 305, 510, 60 },
		{ 305, 511, 60 },
		{ 305, 512, 60 },
		{ 306, 388, 60 },
		{ 306, 389, 60 },
		{ 306, 390, 60 },
		{ 306, 391, 60 },
		{ 306, 392, 60 },
		{ 306, 393, 60 },
		{ 306, 394, 60 },
		{ 306, 395, 60 },
		{ 306, 396, 60 },
		{ 306, 397, 60 },
		{ 306, 398, 60 },
		{ 306, 399, 60 },
		{ 306, 400, 60 },
		{ 306, 401, 60 },
		{ 306, 402, 60 },
		{ 306, 403, 60 },
		{ 306, 404, 60 },
		{ 306, 405, 60 },
		{ 306, 406, 60 },
		{ 306, 407, 60 },
		{ 306, 408, 60 },
		{ 306, 409, 60 },
		{ 306, 410, 60 },
		{ 306, 411, 60 },
		{ 306, 412, 60 },
		{ 306, 413, 60 },
		{ 306, 414, 60 },
		{ 306, 415, 60 },
		{ 306, 416, 60 },
		{ 306, 417, 60 },
		{ 306, 418, 60 },
		{ 306, 419, 60 },
		{ 306, 420, 60 },
		{ 306, 421, 60 },
		{ 306, 422, 60 },
		{ 306, 423, 60 },
		{ 306, 424, 60 },
		{ 306, 425, 60 },
		{ 306, 426, 60 },
		{ 306, 427, 60 },
		{ 306, 428, 60 },
		{ 306, 429, 60 },
		{ 306, 430, 60 },
		{ 306, 431, 60 },
		{ 306, 432, 60 },
		{ 306, 433, 60 },
		{ 306, 434, 60 },
		{ 306, 435, 60 },
		{ 306, 436, 60 },
		{ 306, 437, 60 },
		{ 306, 438, 60 },
		{ 306, 439, 60 },
		{ 306, 440, 60 },
		{ 306, 441, 60 },
		{ 306, 442, 60 },
		{ 306, 443, 60 },
		{ 306, 444, 60 },
		{ 306, 445, 60 },
		{ 306, 446, 60 },
		{ 306, 447, 60 },
		{ 306, 448, 60 },
		{ 306, 449, 60 },
		{ 306, 450, 60 },
		{ 306, 451, 60 },
		{ 306, 452, 60 },
		{ 306, 453, 60 },
		{ 306, 454, 60 },
		{ 306, 455, 60 },
		{ 306, 456, 60 },
		{ 306, 457, 60 },
		{ 306, 458, 60 },
		{ 306, 459, 60 },
		{ 306, 460, 60 },
		{ 306, 461, 60 },
		{ 306, 462, 60 },
		{ 306, 463, 60 },
		{ 306, 464, 60 },
		{ 306, 465, 60 },
		{ 306, 466, 60 },
		{ 306, 467, 60 },
		{ 306, 468, 60 },
		{ 306, 469, 60 },
		{ 306, 470, 60 },
		{ 306, 471, 60 },
		{ 306, 472, 60 },
		{ 306, 473, 60 },
		{ 306, 474, 60 },
		{ 306, 475, 60 },
		{ 306, 476, 60 },
		{ 306, 477, 60 },
		{ 306, 478, 60 },
		{ 306, 479, 60 },
		{ 306, 480, 60 },
		{ 306, 481, 60 },
		{ 306, 482, 60 },
		{ 306, 483, 60 },
		{ 306, 484, 60 },
		{ 306, 485, 60 },
		{ 306, 486, 60 },
		{ 306, 487, 60 },
		{ 306, 488, 60 },
		{ 306, 489, 60 },
		{ 306, 490, 60 },
		{ 306, 491, 60 },
		{ 306, 492, 60 },
		{ 306, 493, 60 },
		{ 306, 494, 60 },
		{ 306, 495, 60 },
		{ 306, 496, 60 },
		{ 306, 497, 60 },
		{ 306, 498, 60 },
		{ 306, 499, 60 },
		{ 306, 500, 60 },
		{ 306, 501, 60 },
		{ 306, 502, 60 },
		{ 306, 503, 60 },
		{ 306, 504, 60 },
		{ 306, 505, 60 },
		{ 306, 506, 60 },
		{ 306, 507, 60 },
		{ 306, 508, 60 },
		{ 306, 509, 60 },
		{ 306, 510, 60 },
		{ 306, 511, 60 },
		{ 306, 512, 60 },
		{ 307, 391, 60 },
		{ 307, 392, 60 },
		{ 307, 393, 60 },
		{ 307, 394, 60 },
		{ 307, 395, 60 },
		{ 307, 396, 60 },
		{ 307, 397, 60 },
		{ 307, 398, 60 },
		{ 307, 399, 60 },
		{ 307, 400, 60 },
		{ 307, 401, 60 },
		{ 307, 402, 60 },
		{ 307, 403, 60 },
		{ 307, 404, 60 },
		{ 307, 405, 60 },
		{ 307, 406, 60 },
		{ 307, 407, 60 },
		{ 307, 408, 60 },
		{ 307, 409, 60 },
		{ 307, 410, 60 },
		{ 307, 411, 60 },
		{ 307, 412, 60 },
		{ 307, 413, 60 },
		{ 307, 414, 60 },
		{ 307, 415, 60 },
		{ 307, 416, 60 },
		{ 307, 417, 60 },
		{ 307, 418, 60 },
		{ 307, 419, 60 },
		{ 307, 420, 60 },
		{ 307, 421, 60 },
		{ 307, 422, 60 },
		{ 307, 423, 60 },
		{ 307, 424, 60 },
		{ 307, 425, 60 },
		{ 307, 426, 60 },
		{ 307, 427, 60 },
		{ 307, 428, 60 },
		{ 307, 429, 60 },
		{ 307, 430, 60 },
		{ 307, 431, 60 },
		{ 307, 432, 60 },
		{ 307, 433, 60 },
		{ 307, 434, 60 },
		{ 307, 435, 60 },
		{ 307, 436, 60 },
		{ 307, 437, 60 },
		{ 307, 438, 60 },
		{ 307, 439, 60 },
		{ 307, 440, 60 },
		{ 307, 441, 60 },
		{ 307, 442, 60 },
		{ 307, 443, 60 },
		{ 307, 444, 60 },
		{ 307, 445, 60 },
		{ 307, 446, 60 },
		{ 307, 447, 60 },
		{ 307, 448, 60 },
		{ 307, 449, 60 },
		{ 307, 450, 60 },
		{ 307, 451, 60 },
		{ 307, 452, 60 },
		{ 307, 453, 60 },
		{ 307, 454, 60 },
		{ 307, 455, 60 },
		{ 307, 456, 60 },
		{ 307, 457, 60 },
		{ 307, 458, 60 },
		{ 307, 459, 60 },
		{ 307, 460, 60 },
		{ 307, 461, 60 },
		{ 307, 462, 60 },
		{ 307, 463, 60 },
		{ 307, 464, 60 },
		{ 307, 465, 60 },
		{ 307, 466, 60 },
		{ 307, 467, 60 },
		{ 307, 468, 60 },
		{ 307, 469, 60 },
		{ 307, 470, 60 },
		{ 307, 471, 60 },
		{ 307, 472, 60 },
		{ 307, 473, 60 },
		{ 307, 474, 60 },
		{ 307, 475, 60 },
		{ 307, 476, 60 },
		{ 307, 477, 60 },
		{ 307, 478, 60 },
		{ 307, 479, 60 },
		{ 307, 480, 60 },
		{ 307, 481, 60 },
		{ 307, 482, 60 },
		{ 307, 483, 60 },
		{ 307, 484, 60 },
		{ 307, 485, 60 },
		{ 307, 486, 60 },
		{ 307, 487, 60 },
		{ 307, 488, 60 },
		{ 307, 489, 60 },
		{ 307, 490, 60 },
		{ 307, 491, 60 },
		{ 307, 492, 60 },
		{ 307, 493, 60 },
		{ 307, 494, 60 },
		{ 307, 495, 60 },
		{ 307, 496, 60 },
		{ 307, 497, 60 },
		{ 307, 498, 60 },
		{ 307, 499, 60 },
		{ 307, 500, 60 },
		{ 307, 501, 60 },
		{ 307, 502, 60 },
		{ 307, 503, 60 },
		{ 307, 504, 60 },
		{ 307, 505, 60 },
		{ 307, 506, 60 },
		{ 307, 507, 60 },
		{ 307, 508, 60 },
		{ 307, 509, 60 },
		{ 307, 510, 60 },
		{ 307, 511, 60 },
		{ 307, 512, 60 },
		{ 308, 394, 60 },
		{ 308, 395, 60 },
		{ 308, 396, 60 },
		{ 308, 397, 60 },
		{ 308, 398, 60 },
		{ 308, 399, 60 },
		{ 308, 400, 60 },
		{ 308, 401, 60 },
		{ 308, 402, 60 },
		{ 308, 403, 60 },
		{ 308, 404, 60 },
		{ 308, 405, 60 },
		{ 308, 406, 60 },
		{ 308, 407, 60 },
		{ 308, 408, 60 },
		{ 308, 409, 60 },
		{ 308, 410, 60 },
		{ 308, 411, 60 },
		{ 308, 412, 60 },
		{ 308, 413, 60 },
		{ 308, 414, 60 },
		{ 308, 415, 60 },
		{ 308, 416, 60 },
		{ 308, 417, 60 },
		{ 308, 418, 60 },
		{ 308, 419, 60 },
		{ 308, 420, 60 },
		{ 308, 421, 60 },
		{ 308, 422, 60 },
		{ 308, 423, 60 },
		{ 308, 424, 60 },
		{ 308, 425, 60 },
		{ 308, 426, 60 },
		{ 308, 427, 60 },
		{ 308, 428, 60 },
		{ 308, 429, 60 },
		{ 308, 430, 60 },
		{ 308, 431, 60 },
		{ 308, 432, 60 },
		{ 308, 433, 60 },
		{ 308, 434, 60 },
		{ 308, 435, 60 },
		{ 308, 436, 60 },
		{ 308, 437, 60 },
		{ 308, 438, 60 },
		{ 308, 439, 60 },
		{ 308, 440, 60 },
		{ 308, 441, 60 },
		{ 308, 442, 60 },
		{ 308, 443, 60 },
		{ 308, 444, 60 },
		{ 308, 445, 60 },
		{ 308, 446, 60 },
		{ 308, 447, 60 },
		{ 308, 448, 60 },
		{ 308, 449, 60 },
		{ 308, 450, 60 },
		{ 308, 451, 60 },
		{ 308, 452, 60 },
		{ 308, 453, 60 },
		{ 308, 454, 60 },
		{ 308, 455, 60 },
		{ 308, 456, 60 },
		{ 308, 457, 60 },
		{ 308, 458, 60 },
		{ 308, 459, 60 },
		{ 308, 460, 60 },
		{ 308, 461, 60 },
		{ 308, 462, 60 },
		{ 308, 463, 60 },
		{ 308, 464, 60 },
		{ 308, 465, 60 },
		{ 308, 466, 60 },
		{ 308, 467, 60 },
		{ 308, 468, 60 },
		{ 308, 469, 60 },
		{ 308, 470, 60 },
		{ 308, 471, 60 },
		{ 308, 472, 60 },
		{ 308, 473, 60 },
		{ 308, 474, 60 },
		{ 308, 475, 60 },
		{ 308, 476, 60 },
		{ 308, 477, 60 },
		{ 308, 478, 60 },
		{ 308, 479, 60 },
		{ 308, 480, 60 },
		{ 308, 481, 60 },
		{ 308, 482, 60 },
		{ 308, 483, 60 },
		{ 308, 484, 60 },
		{ 308, 485, 60 },
		{ 308, 486, 60 },
		{ 308, 487, 60 },
		{ 308, 488, 60 },
		{ 308, 489, 60 },
		{ 308, 490, 60 },
		{ 308, 491, 60 },
		{ 308, 492, 60 },
		{ 308, 493, 60 },
		{ 308, 494, 60 },
		{ 308, 495, 60 },
		{ 308, 496, 60 },
		{ 308, 497, 60 },
		{ 308, 498, 60 },
		{ 308, 499, 60 },
		{ 308, 500, 60 },
		{ 308, 501, 60 },
		{ 308, 502, 60 },
		{ 308, 503, 60 },
		{ 308, 504, 60 },
		{ 308, 505, 60 },
		{ 308, 506, 60 },
		{ 308, 507, 60 },
		{ 308, 508, 60 },
		{ 308, 509, 60 },
		{ 308, 510, 60 },
		{ 308, 511, 60 },
		{ 308, 512, 60 },
		{ 309, 396, 60 },
		{ 309, 397, 60 },
		{ 309, 398, 60 },
		{ 309, 399, 60 },
		{ 309, 400, 60 },
		{ 309, 401, 60 },
		{ 309, 402, 60 },
		{ 309, 403, 60 },
		{ 309, 404, 60 },
		{ 309, 405, 60 },
		{ 309, 406, 60 },
		{ 309, 407, 60 },
		{ 309, 408, 60 },
		{ 309, 409, 60 },
		{ 309, 410, 60 },
		{ 309, 411, 60 },
		{ 309, 412, 60 },
		{ 309, 413, 60 },
		{ 309, 414, 60 },
		{ 309, 415, 60 },
		{ 309, 416, 60 },
		{ 309, 417, 60 },
		{ 309, 418, 60 },
		{ 309, 419, 60 },
		{ 309, 420, 60 },
		{ 309, 421, 60 },
		{ 309, 422, 60 },
		{ 309, 423, 60 },
		{ 309, 424, 60 },
		{ 309, 425, 60 },
		{ 309, 426, 60 },
		{ 309, 427, 60 },
		{ 309, 428, 60 },
		{ 309, 429, 60 },
		{ 309, 430, 60 },
		{ 309, 431, 60 },
		{ 309, 432, 60 },
		{ 309, 433, 60 },
		{ 309, 434, 60 },
		{ 309, 435, 60 },
		{ 309, 436, 60 },
		{ 309, 437, 60 },
		{ 309, 438, 60 },
		{ 309, 439, 60 },
		{ 309, 440, 60 },
		{ 309, 441, 60 },
		{ 309, 442, 60 },
		{ 309, 443, 60 },
		{ 309, 444, 60 },
		{ 309, 445, 60 },
		{ 309, 446, 60 },
		{ 309, 447, 60 },
		{ 309, 448, 60 },
		{ 309, 449, 60 },
		{ 309, 450, 60 },
		{ 309, 451, 60 },
		{ 309, 452, 60 },
		{ 309, 453, 60 },
		{ 309, 454, 60 },
		{ 309, 455, 60 },
		{ 309, 456, 60 },
		{ 309, 457, 60 },
		{ 309, 458, 60 },
		{ 309, 459, 60 },
		{ 309, 460, 60 },
		{ 309, 461, 60 },
		{ 309, 462, 60 },
		{ 309, 463, 60 },
		{ 309, 464, 60 },
		{ 309, 465, 60 },
		{ 309, 466, 60 },
		{ 309, 467, 60 },
		{ 309, 468, 60 },
		{ 309, 469, 60 },
		{ 309, 470, 60 },
		{ 309, 471, 60 },
		{ 309, 472, 60 },
		{ 309, 473, 60 },
		{ 309, 474, 60 },
		{ 309, 475, 60 },
		{ 309, 476, 60 },
		{ 309, 477, 60 },
		{ 309, 478, 60 },
		{ 309, 479, 60 },
		{ 309, 480, 60 },
		{ 309, 481, 60 },
		{ 309, 482, 60 },
		{ 309, 483, 60 },
		{ 309, 484, 60 },
		{ 309, 485, 60 },
		{ 309, 486, 60 },
		{ 309, 487, 60 },
		{ 309, 488, 60 },
		{ 309, 489, 60 },
		{ 309, 490, 60 },
		{ 309, 491, 60 },
		{ 309, 492, 60 },
		{ 309, 493, 60 },
		{ 309, 494, 60 },
		{ 309, 495, 60 },
		{ 309, 496, 60 },
		{ 309, 497, 60 },
		{ 309, 498, 60 },
		{ 309, 499, 60 },
		{ 309, 500, 60 },
		{ 309, 501, 60 },
		{ 309, 502, 60 },
		{ 309, 503, 60 },
		{ 309, 504, 60 },
		{ 309, 505, 60 },
		{ 309, 506, 60 },
		{ 309, 507, 60 },
		{ 309, 508, 60 },
		{ 309, 509, 60 },
		{ 309, 510, 60 },
		{ 309, 511, 60 },
		{ 309, 512, 60 },
		{ 310, 399, 60 },
		{ 310, 400, 60 },
		{ 310, 401, 60 },
		{ 310, 402, 60 },
		{ 310, 403, 60 },
		{ 310, 404, 60 },
		{ 310, 405, 60 },
		{ 310, 406, 60 },
		{ 310, 407, 60 },
		{ 310, 408, 60 },
		{ 310, 409, 60 },
		{ 310, 410, 60 },
		{ 310, 411, 60 },
		{ 310, 412, 60 },
		{ 310, 413, 60 },
		{ 310, 414, 60 },
		{ 310, 415, 60 },
		{ 310, 416, 60 },
		{ 310, 417, 60 },
		{ 310, 418, 60 },
		{ 310, 419, 60 },
		{ 310, 420, 60 },
		{ 310, 421, 60 },
		{ 310, 422, 60 },
		{ 310, 423, 60 },
		{ 310, 424, 60 },
		{ 310, 425, 60 },
		{ 310, 426, 60 },
		{ 310, 427, 60 },
		{ 310, 428, 60 },
		{ 310, 429, 60 },
		{ 310, 430, 60 },
		{ 310, 431, 60 },
		{ 310, 432, 60 },
		{ 310, 433, 60 },
		{ 310, 434, 60 },
		{ 310, 435, 60 },
		{ 310, 436, 60 },
		{ 310, 437, 60 },
		{ 310, 438, 60 },
		{ 310, 439, 60 },
		{ 310, 440, 60 },
		{ 310, 441, 60 },
		{ 310, 442, 60 },
		{ 310, 443, 60 },
		{ 310, 444, 60 },
		{ 310, 445, 60 },
		{ 310, 446, 60 },
		{ 310, 447, 60 },
		{ 310, 448, 60 },
		{ 310, 449, 60 },
		{ 310, 450, 60 },
		{ 310, 451, 60 },
		{ 310, 452, 60 },
		{ 310, 453, 60 },
		{ 310, 454, 60 },
		{ 310, 455, 60 },
		{ 310, 456, 60 },
		{ 310, 457, 60 },
		{ 310, 458, 60 },
		{ 310, 459, 60 },
		{ 310, 460, 60 },
		{ 310, 461, 60 },
		{ 310, 462, 60 },
		{ 310, 463, 60 },
		{ 310, 464, 60 },
		{ 310, 465, 60 },
		{ 310, 466, 60 },
		{ 310, 467, 60 },
		{ 310, 468, 60 },
		{ 310, 469, 60 },
		{ 310, 470, 60 },
		{ 310, 471, 60 },
		{ 310, 472, 60 },
		{ 310, 473, 60 },
		{ 310, 474, 60 },
		{ 310, 475, 60 },
		{ 310, 476, 60 },
		{ 310, 477, 60 },
		{ 310, 478, 60 },
		{ 310, 479, 60 },
		{ 310, 480, 60 },
		{ 310, 481, 60 },
		{ 310, 482, 60 },
		{ 310, 483, 60 },
		{ 310, 484, 60 },
		{ 310, 485, 60 },
		{ 310, 486, 60 },
		{ 310, 487, 60 },
		{ 310, 488, 60 },
		{ 310, 489, 60 },
		{ 310, 490, 60 },
		{ 310, 491, 60 },
		{ 310, 492, 60 },
		{ 310, 493, 60 },
		{ 310, 494, 60 },
		{ 310, 495, 60 },
		{ 310, 496, 60 },
		{ 310, 497, 60 },
		{ 310, 498, 60 },
		{ 310, 499, 60 },
		{ 310, 500, 60 },
		{ 310, 501, 60 },
		{ 310, 502, 60 },
		{ 310, 503, 60 },
		{ 310, 504, 60 },
		{ 310, 505, 60 },
		{ 310, 506, 60 },
		{ 310, 507, 60 },
		{ 310, 508, 60 },
		{ 310, 509, 60 },
		{ 310, 510, 60 },
		{ 310, 511, 60 },
		{ 310, 512, 60 },
		{ 311, 402, 60 },
		{ 311, 403, 60 },
		{ 311, 404, 60 },
		{ 311, 405, 60 },
		{ 311, 406, 60 },
		{ 311, 407, 60 },
		{ 311, 408, 60 },
		{ 311, 409, 60 },
		{ 311, 410, 60 },
		{ 311, 411, 60 },
		{ 311, 412, 60 },
		{ 311, 413, 60 },
		{ 311, 414, 60 },
		{ 311, 415, 60 },
		{ 311, 416, 60 },
		{ 311, 417, 60 },
		{ 311, 418, 60 },
		{ 311, 419, 60 },
		{ 311, 420, 60 },
		{ 311, 421, 60 },
		{ 311, 422, 60 },
		{ 311, 423, 60 },
		{ 311, 424, 60 },
		{ 311, 425, 60 },
		{ 311, 426, 60 },
		{ 311, 427, 60 },
		{ 311, 428, 60 },
		{ 311, 429, 60 },
		{ 311, 430, 60 },
		{ 311, 431, 60 },
		{ 311, 432, 60 },
		{ 311, 433, 60 },
		{ 311, 434, 60 },
		{ 311, 435, 60 },
		{ 311, 436, 60 },
		{ 311, 437, 60 },
		{ 311, 438, 60 },
		{ 311, 439, 60 },
		{ 311, 440, 60 },
		{ 311, 441, 60 },
		{ 311, 442, 60 },
		{ 311, 443, 60 },
		{ 311, 444, 60 },
		{ 311, 445, 60 },
		{ 311, 446, 60 },
		{ 311, 447, 60 },
		{ 311, 448, 60 },
		{ 311, 449, 60 },
		{ 311, 450, 60 },
		{ 311, 451, 60 },
		{ 311, 452, 60 },
		{ 311, 453, 60 },
		{ 311, 454, 60 },
		{ 311, 455, 60 },
		{ 311, 456, 60 },
		{ 311, 457, 60 },
		{ 311, 458, 60 },
		{ 311, 459, 60 },
		{ 311, 460, 60 },
		{ 311, 461, 60 },
		{ 311, 462, 60 },
		{ 311, 463, 60 },
		{ 311, 464, 60 },
		{ 311, 465, 60 },
		{ 311, 466, 60 },
		{ 311, 467, 60 },
		{ 311, 468, 60 },
		{ 311, 469, 60 },
		{ 311, 470, 60 },
		{ 311, 471, 60 },
		{ 311, 472, 60 },
		{ 311, 473, 60 },
		{ 311, 474, 60 },
		{ 311, 475, 60 },
		{ 311, 476, 60 },
		{ 311, 477, 60 },
		{ 311, 478, 60 },
		{ 311, 479, 60 },
		{ 311, 480, 60 },
		{ 311, 481, 60 },
		{ 311, 482, 60 },
		{ 311, 483, 60 },
		{ 311, 484, 60 },
		{ 311, 485, 60 },
		{ 311, 486, 60 },
		{ 311, 487, 60 },
		{ 311, 488, 60 },
		{ 311, 489, 60 },
		{ 311, 490, 60 },
		{ 311, 491, 60 },
		{ 311, 492, 60 },
		{ 311, 493, 60 },
		{ 311, 494, 60 },
		{ 311, 495, 60 },
		{ 311, 496, 60 },
		{ 311, 497, 60 },
		{ 311, 498, 60 },
		{ 311, 499, 60 },
		{ 311, 500, 60 },
		{ 311, 501, 60 },
		{ 311, 502, 60 },
		{ 311, 503, 60 },
		{ 311, 504, 60 },
		{ 311, 505, 60 },
		{ 311, 506, 60 },
		{ 311, 507, 60 },
		{ 311, 508, 60 },
		{ 311, 509, 60 },
		{ 311, 510, 60 },
		{ 311, 511, 60 },
		{ 311, 512, 60 },
		{ 312, 404, 60 },
		{ 312, 405, 60 },
		{ 312, 406, 60 },
		{ 312, 407, 60 },
		{ 312, 408, 60 },
		{ 312, 409, 60 },
		{ 312, 410, 60 },
		{ 312, 411, 60 },
		{ 312, 412, 60 },
		{ 312, 413, 60 },
		{ 312, 414, 60 },
		{ 312, 415, 60 },
		{ 312, 416, 60 },
		{ 312, 417, 60 },
		{ 312, 418, 60 },
		{ 312, 419, 60 },
		{ 312, 420, 60 },
		{ 312, 421, 60 },
		{ 312, 422, 60 },
		{ 312, 423, 60 },
		{ 312, 424, 60 },
		{ 312, 425, 60 },
		{ 312, 426, 60 },
		{ 312, 427, 60 },
		{ 312, 428, 60 },
		{ 312, 429, 60 },
		{ 312, 430, 60 },
		{ 312, 431, 60 },
		{ 312, 432, 60 },
		{ 312, 433, 60 },
		{ 312, 434, 60 },
		{ 312, 435, 60 },
		{ 312, 436, 60 },
		{ 312, 437, 60 },
		{ 312, 438, 60 },
		{ 312, 439, 60 },
		{ 312, 440, 60 },
		{ 312, 441, 60 },
		{ 312, 442, 60 },
		{ 312, 443, 60 },
		{ 312, 444, 60 },
		{ 312, 445, 60 },
		{ 312, 446, 60 },
		{ 312, 447, 60 },
		{ 312, 448, 60 },
		{ 312, 449, 60 },
		{ 312, 450, 60 },
		{ 312, 451, 60 },
		{ 312, 452, 60 },
		{ 312, 453, 60 },
		{ 312, 454, 60 },
		{ 312, 455, 60 },
		{ 312, 456, 60 },
		{ 312, 457, 60 },
		{ 312, 458, 60 },
		{ 312, 459, 60 },
		{ 312, 460, 60 },
		{ 312, 461, 60 },
		{ 312, 462, 60 },
		{ 312, 463, 60 },
		{ 312, 464, 60 },
		{ 312, 465, 60 },
		{ 312, 466, 60 },
		{ 312, 467, 60 },
		{ 312, 468, 60 },
		{ 312, 469, 60 },
		{ 312, 470, 60 },
		{ 312, 471, 60 },
		{ 312, 472, 60 },
		{ 312, 473, 60 },
		{ 312, 474, 60 },
		{ 312, 475, 60 },
		{ 312, 476, 60 },
		{ 312, 477, 60 },
		{ 312, 478, 60 },
		{ 312, 479, 60 },
		{ 312, 480, 60 },
		{ 312, 481, 60 },
		{ 312, 482, 60 },
		{ 312, 483, 60 },
		{ 312, 484, 60 },
		{ 312, 485, 60 },
		{ 312, 486, 60 },
		{ 312, 487, 60 },
		{ 312, 488, 60 },
		{ 312, 489, 60 },
		{ 312, 490, 60 },
		{ 312, 491, 60 },
		{ 312, 492, 60 },
		{ 312, 493, 60 },
		{ 312, 494, 60 },
		{ 312, 495, 60 },
		{ 312, 496, 60 },
		{ 312, 497, 60 },
		{ 312, 498, 60 },
		{ 312, 499, 60 },
		{ 312, 500, 60 },
		{ 312, 501, 60 },
		{ 312, 502, 60 },
		{ 312, 503, 60 },
		{ 312, 504, 60 },
		{ 312, 505, 60 },
		{ 312, 506, 60 },
		{ 312, 507, 60 },
		{ 312, 508, 60 },
		{ 312, 509, 60 },
		{ 312, 510, 60 },
		{ 312, 511, 60 },
		{ 312, 512, 60 },
		{ 313, 407, 60 },
		{ 313, 408, 60 },
		{ 313, 409, 60 },
		{ 313, 410, 60 },
		{ 313, 411, 60 },
		{ 313, 412, 60 },
		{ 313, 413, 60 },
		{ 313, 414, 60 },
		{ 313, 415, 60 },
		{ 313, 416, 60 },
		{ 313, 417, 60 },
		{ 313, 418, 60 },
		{ 313, 419, 60 },
		{ 313, 420, 60 },
		{ 313, 421, 60 },
		{ 313, 422, 60 },
		{ 313, 423, 60 },
		{ 313, 424, 60 },
		{ 313, 425, 60 },
		{ 313, 426, 60 },
		{ 313, 427, 60 },
		{ 313, 428, 60 },
		{ 313, 429, 60 },
		{ 313, 430, 60 },
		{ 313, 431, 60 },
		{ 313, 432, 60 },
		{ 313, 433, 60 },
		{ 313, 434, 60 },
		{ 313, 435, 60 },
		{ 313, 436, 60 },
		{ 313, 437, 60 },
		{ 313, 438, 60 },
		{ 313, 439, 60 },
		{ 313, 440, 60 },
		{ 313, 441, 60 },
		{ 313, 442, 60 },
		{ 313, 443, 60 },
		{ 313, 444, 60 },
		{ 313, 445, 60 },
		{ 313, 446, 60 },
		{ 313, 447, 60 },
		{ 313, 448, 60 },
		{ 313, 449, 60 },
		{ 313, 450, 60 },
		{ 313, 451, 60 },
		{ 313, 452, 60 },
		{ 313, 453, 60 },
		{ 313, 454, 60 },
		{ 313, 455, 60 },
		{ 313, 456, 60 },
		{ 313, 457, 60 },
		{ 313, 458, 60 },
		{ 313, 459, 60 },
		{ 313, 460, 60 },
		{ 313, 461, 60 },
		{ 313, 462, 60 },
		{ 313, 463, 60 },
		{ 313, 464, 60 },
		{ 313, 465, 60 },
		{ 313, 466, 60 },
		{ 313, 467, 60 },
		{ 313, 468, 60 },
		{ 313, 469, 60 },
		{ 313, 470, 60 },
		{ 313, 471, 60 },
		{ 313, 472, 60 },
		{ 313, 473, 60 },
		{ 313, 474, 60 },
		{ 313, 475, 60 },
		{ 313, 476, 60 },
		{ 313, 477, 60 },
		{ 313, 478, 60 },
		{ 313, 479, 60 },
		{ 313, 480, 60 },
		{ 313, 481, 60 },
		{ 313, 482, 60 },
		{ 313, 483, 60 },
		{ 313, 484, 60 },
		{ 313, 485, 60 },
		{ 313, 486, 60 },
		{ 313, 487, 60 },
		{ 313, 488, 60 },
		{ 313, 489, 60 },
		{ 313, 490, 60 },
		{ 313, 491, 60 },
		{ 313, 492, 60 },
		{ 313, 493, 60 },
		{ 313, 494, 60 },
		{ 313, 495, 60 },
		{ 313, 496, 60 },
		{ 313, 497, 60 },
		{ 313, 498, 60 },
		{ 313, 499, 60 },
		{ 313, 500, 60 },
		{ 313, 501, 60 },
		{ 313, 502, 60 },
		{ 313, 503, 60 },
		{ 313, 504, 60 },
		{ 313, 505, 60 },
		{ 313, 506, 60 },
		{ 313, 507, 60 },
		{ 313, 508, 60 },
		{ 313, 509, 60 },
		{ 313, 510, 60 },
		{ 313, 511, 60 },
		{ 313, 512, 60 },
		{ 314, 410, 60 },
		{ 314, 411, 60 },
		{ 314, 412, 60 },
		{ 314, 413, 60 },
		{ 314, 414, 60 },
		{ 314, 415, 60 },
		{ 314, 416, 60 },
		{ 314, 417, 60 },
		{ 314, 418, 60 },
		{ 314, 419, 60 },
		{ 314, 420, 60 },
		{ 314, 421, 60 },
		{ 314, 422, 60 },
		{ 314, 423, 60 },
		{ 314, 424, 60 },
		{ 314, 425, 60 },
		{ 314, 426, 60 },
		{ 314, 427, 60 },
		{ 314, 428, 60 },
		{ 314, 429, 60 },
		{ 314, 430, 60 },
		{ 314, 431, 60 },
		{ 314, 432, 60 },
		{ 314, 433, 60 },
		{ 314, 434, 60 },
		{ 314, 435, 60 },
		{ 314, 436, 60 },
		{ 314, 437, 60 },
		{ 314, 438, 60 },
		{ 314, 439, 60 },
		{ 314, 440, 60 },
		{ 314, 441, 60 },
		{ 314, 442, 60 },
		{ 314, 443, 60 },
		{ 314, 444, 60 },
		{ 314, 445, 60 },
		{ 314, 446, 60 },
		{ 314, 447, 60 },
		{ 314, 448, 60 },
		{ 314, 449, 60 },
		{ 314, 450, 60 },
		{ 314, 451, 60 },
		{ 314, 452, 60 },
		{ 314, 453, 60 },
		{ 314, 454, 60 },
		{ 314, 455, 60 },
		{ 314, 456, 60 },
		{ 314, 457, 60 },
		{ 314, 458, 60 },
		{ 314, 459, 60 },
		{ 314, 460, 60 },
		{ 314, 461, 60 },
		{ 314, 462, 60 },
		{ 314, 463, 60 },
		{ 314, 464, 60 },
		{ 314, 465, 60 },
		{ 314, 466, 60 },
		{ 314, 467, 60 },
		{ 314, 468, 60 },
		{ 314, 469, 60 },
		{ 314, 470, 60 },
		{ 314, 471, 60 },
		{ 314, 472, 60 },
		{ 314, 473, 60 },
		{ 314, 474, 60 },
		{ 314, 475, 60 },
		{ 314, 476, 60 },
		{ 314, 477, 60 },
		{ 314, 478, 60 },
		{ 314, 479, 60 },
		{ 314, 480, 60 },
		{ 314, 481, 60 },
		{ 314, 482, 60 },
		{ 314, 483, 60 },
		{ 314, 484, 60 },
		{ 314, 485, 60 },
		{ 314, 486, 60 },
		{ 314, 487, 60 },
		{ 314, 488, 60 },
		{ 314, 489, 60 },
		{ 314, 490, 60 },
		{ 314, 491, 60 },
		{ 314, 492, 60 },
		{ 314, 493, 60 },
		{ 314, 494, 60 },
		{ 314, 495, 60 },
		{ 314, 496, 60 },
		{ 314, 497, 60 },
		{ 314, 498, 60 },
		{ 314, 499, 60 },
		{ 314, 500, 60 },
		{ 314, 501, 60 },
		{ 314, 502, 60 },
		{ 314, 503, 60 },
		{ 314, 504, 60 },
		{ 314, 505, 60 },
		{ 314, 506, 60 },
		{ 314, 507, 60 },
		{ 314, 508, 60 },
		{ 314, 509, 60 },
		{ 314, 510, 60 },
		{ 314, 511, 60 },
		{ 314, 512, 60 },
		{ 315, 412, 60 },
		{ 315, 413, 60 },
		{ 315, 414, 60 },
		{ 315, 415, 60 },
		{ 315, 416, 60 },
		{ 315, 417, 60 },
		{ 315, 418, 60 },
		{ 315, 419, 60 },
		{ 315, 420, 60 },
		{ 315, 421, 60 },
		{ 315, 422, 60 },
		{ 315, 423, 60 },
		{ 315, 424, 60 },
		{ 315, 425, 60 },
		{ 315, 426, 60 },
		{ 315, 427, 60 },
		{ 315, 428, 60 },
		{ 315, 429, 60 },
		{ 315, 430, 60 },
		{ 315, 431, 60 },
		{ 315, 432, 60 },
		{ 315, 433, 60 },
		{ 315, 434, 60 },
		{ 315, 435, 60 },
		{ 315, 436, 60 },
		{ 315, 437, 60 },
		{ 315, 438, 60 },
		{ 315, 439, 60 },
		{ 315, 440, 60 },
		{ 315, 441, 60 },
		{ 315, 442, 60 },
		{ 315, 443, 60 },
		{ 315, 444, 60 },
		{ 315, 445, 60 },
		{ 315, 446, 60 },
		{ 315, 447, 60 },
		{ 315, 448, 60 },
		{ 315, 449, 60 },
		{ 315, 450, 60 },
		{ 315, 451, 60 },
		{ 315, 452, 60 },
		{ 315, 453, 60 },
		{ 315, 454, 60 },
		{ 315, 455, 60 },
		{ 315, 456, 60 },
		{ 315, 457, 60 },
		{ 315, 458, 60 },
		{ 315, 459, 60 },
		{ 315, 460, 60 },
		{ 315, 461, 60 },
		{ 315, 462, 60 },
		{ 315, 463, 60 },
		{ 315, 464, 60 },
		{ 315, 465, 60 },
		{ 315, 466, 60 },
		{ 315, 467, 60 },
		{ 315, 468, 60 },
		{ 315, 469, 60 },
		{ 315, 470, 60 },
		{ 315, 471, 60 },
		{ 315, 472, 60 },
		{ 315, 473, 60 },
		{ 315, 474, 60 },
		{ 315, 475, 60 },
		{ 315, 476, 60 },
		{ 315, 477, 60 },
		{ 315, 478, 60 },
		{ 315, 479, 60 },
		{ 315, 480, 60 },
		{ 315, 481, 60 },
		{ 315, 482, 60 },
		{ 315, 483, 60 },
		{ 315, 484, 60 },
		{ 315, 485, 60 },
		{ 315, 486, 60 },
		{ 315, 487, 60 },
		{ 315, 488, 60 },
		{ 315, 489, 60 },
		{ 315, 490, 60 },
		{ 315, 491, 60 },
		{ 315, 492, 60 },
		{ 315, 493, 60 },
		{ 315, 494, 60 },
		{ 315, 495, 60 },
		{ 315, 496, 60 },
		{ 315, 497, 60 },
		{ 315, 498, 60 },
		{ 315, 499, 60 },
		{ 315, 500, 60 },
		{ 315, 501, 60 },
		{ 315, 502, 60 },
		{ 315, 503, 60 },
		{ 315, 504, 60 },
		{ 315, 505, 60 },
		{ 315, 506, 60 },
		{ 315, 507, 60 },
		{ 315, 508, 60 },
		{ 315, 509, 60 },
		{ 315, 510, 60 },
		{ 315, 511, 60 },
		{ 315, 512, 60 },
		{ 316, 415, 60 },
		{ 316, 416, 60 },
		{ 316, 417, 60 },
		{ 316, 418, 60 },
		{ 316, 419, 60 },
		{ 316, 420, 60 },
		{ 316, 421, 60 },
		{ 316, 422, 60 },
		{ 316, 423, 60 },
		{ 316, 424, 60 },
		{ 316, 425, 60 },
		{ 316, 426, 60 },
		{ 316, 427, 60 },
		{ 316, 428, 60 },
		{ 316, 429, 60 },
		{ 316, 430, 60 },
		{ 316, 431, 60 },
		{ 316, 432, 60 },
		{ 316, 433, 60 },
		{ 316, 434, 60 },
		{ 316, 435, 60 },
		{ 316, 436, 60 },
		{ 316, 437, 60 },
		{ 316, 438, 60 },
		{ 316, 439, 60 },
		{ 316, 440, 60 },
		{ 316, 441, 60 },
		{ 316, 442, 60 },
		{ 316, 443, 60 },
		{ 316, 444, 60 },
		{ 316, 445, 60 },
		{ 316, 446, 60 },
		{ 316, 447, 60 },
		{ 316, 448, 60 },
		{ 316, 449, 60 },
		{ 316, 450, 60 },
		{ 316, 451, 60 },
		{ 316, 452, 60 },
		{ 316, 453, 60 },
		{ 316, 454, 60 },
		{ 316, 455, 60 },
		{ 316, 456, 60 },
		{ 316, 457, 60 },
		{ 316, 458, 60 },
		{ 316, 459, 60 },
		{ 316, 460, 60 },
		{ 316, 461, 60 },
		{ 316, 462, 60 },
		{ 316, 463, 60 },
		{ 316, 464, 60 },
		{ 316, 465, 60 },
		{ 316, 466, 60 },
		{ 316, 467, 60 },
		{ 316, 468, 60 },
		{ 316, 469, 60 },
		{ 316, 470, 60 },
		{ 316, 471, 60 },
		{ 316, 472, 60 },
		{ 316, 473, 60 },
		{ 316, 474, 60 },
		{ 316, 475, 60 },
		{ 316, 476, 60 },
		{ 316, 477, 60 },
		{ 316, 478, 60 },
		{ 316, 479, 60 },
		{ 316, 480, 60 },
		{ 316, 481, 60 },
		{ 316, 482, 60 },
		{ 316, 483, 60 },
		{ 316, 484, 60 },
		{ 316, 485, 60 },
		{ 316, 486, 60 },
		{ 316, 487, 60 },
		{ 316, 488, 60 },
		{ 316, 489, 60 },
		{ 316, 490, 60 },
		{ 316, 491, 60 },
		{ 316, 492, 60 },
		{ 316, 493, 60 },
		{ 316, 494, 60 },
		{ 316, 495, 60 },
		{ 316, 496, 60 },
		{ 316, 497, 60 },
		{ 316, 498, 60 },
		{ 316, 499, 60 },
		{ 316, 500, 60 },
		{ 316, 501, 60 },
		{ 316, 502, 60 },
		{ 316, 503, 60 },
		{ 316, 504, 60 },
		{ 316, 505, 60 },
		{ 316, 506, 60 },
		{ 316, 507, 60 },
		{ 316, 508, 60 },
		{ 316, 509, 60 },
		{ 316, 510, 60 },
		{ 316, 511, 60 },
		{ 316, 512, 60 },
		{ 317, 418, 60 },
		{ 317, 419, 60 },
		{ 317, 420, 60 },
		{ 317, 421, 60 },
		{ 317, 422, 60 },
		{ 317, 423, 60 },
		{ 317, 424, 60 },
		{ 317, 425, 60 },
		{ 317, 426, 60 },
		{ 317, 427, 60 },
		{ 317, 428, 60 },
		{ 317, 429, 60 },
		{ 317, 430, 60 },
		{ 317, 431, 60 },
		{ 317, 432, 60 },
		{ 317, 433, 60 },
		{ 317, 434, 60 },
		{ 317, 435, 60 },
		{ 317, 436, 60 },
		{ 317, 437, 60 },
		{ 317, 438, 60 },
		{ 317, 439, 60 },
		{ 317, 440, 60 },
		{ 317, 441, 60 },
		{ 317, 442, 60 },
		{ 317, 443, 60 },
		{ 317, 444, 60 },
		{ 317, 445, 60 },
		{ 317, 446, 60 },
		{ 317, 447, 60 },
		{ 317, 448, 60 },
		{ 317, 449, 60 },
		{ 317, 450, 60 },
		{ 317, 451, 60 },
		{ 317, 452, 60 },
		{ 317, 453, 60 },
		{ 317, 454, 60 },
		{ 317, 455, 60 },
		{ 317, 456, 60 },
		{ 317, 457, 60 },
		{ 317, 458, 60 },
		{ 317, 459, 60 },
		{ 317, 460, 60 },
		{ 317, 461, 60 },
		{ 317, 462, 60 },
		{ 317, 463, 60 },
		{ 317, 464, 60 },
		{ 317, 465, 60 },
		{ 317, 466, 60 },
		{ 317, 467, 60 },
		{ 317, 468, 60 },
		{ 317, 469, 60 },
		{ 317, 470, 60 },
		{ 317, 471, 60 },
		{ 317, 472, 60 },
		{ 317, 473, 60 },
		{ 317, 474, 60 },
		{ 317, 475, 60 },
		{ 317, 476, 60 },
		{ 317, 477, 60 },
		{ 317, 478, 60 },
		{ 317, 479, 60 },
		{ 317, 480, 60 },
		{ 317, 481, 60 },
		{ 317, 482, 60 },
		{ 317, 483, 60 },
		{ 317, 484, 60 },
		{ 317, 485, 60 },
		{ 317, 486, 60 },
		{ 317, 487, 60 },
		{ 317, 488, 60 },
		{ 317, 489, 60 },
		{ 317, 490, 60 },
		{ 317, 491, 60 },
		{ 317, 492, 60 },
		{ 317, 493, 60 },
		{ 317, 494, 60 },
		{ 317, 495, 60 },
		{ 317, 496, 60 },
		{ 317, 497, 60 },
		{ 317, 498, 60 },
		{ 317, 499, 60 },
		{ 317, 500, 60 },
		{ 317, 501, 60 },
		{ 317, 502, 60 },
		{ 317, 503, 60 },
		{ 317, 504, 60 },
		{ 317, 505, 60 },
		{ 317, 506, 60 },
		{ 317, 507, 60 },
		{ 317, 508, 60 },
		{ 317, 509, 60 },
		{ 317, 510, 60 },
		{ 317, 511, 60 },
		{ 317, 512, 60 },
		{ 318, 420, 60 },
		{ 318, 421, 60 },
		{ 318, 422, 60 },
		{ 318, 423, 60 },
		{ 318, 424, 60 },
		{ 318, 425, 60 },
		{ 318, 426, 60 },
		{ 318, 427, 60 },
		{ 318, 428, 60 },
		{ 318, 429, 60 },
		{ 318, 430, 60 },
		{ 318, 431, 60 },
		{ 318, 432, 60 },
		{ 318, 433, 60 },
		{ 318, 434, 60 },
		{ 318, 435, 60 },
		{ 318, 436, 60 },
		{ 318, 437, 60 },
		{ 318, 438, 60 },
		{ 318, 439, 60 },
		{ 318, 440, 60 },
		{ 318, 441, 60 },
		{ 318, 442, 60 },
		{ 318, 443, 60 },
		{ 318, 444, 60 },
		{ 318, 445, 60 },
		{ 318, 446, 60 },
		{ 318, 447, 60 },
		{ 318, 448, 60 },
		{ 318, 449, 60 },
		{ 318, 450, 60 },
		{ 318, 451, 60 },
		{ 318, 452, 60 },
		{ 318, 453, 60 },
		{ 318, 454, 60 },
		{ 318, 455, 60 },
		{ 318, 456, 60 },
		{ 318, 457, 60 },
		{ 318, 458, 60 },
		{ 318, 459, 60 },
		{ 318, 460, 60 },
		{ 318, 461, 60 },
		{ 318, 462, 60 },
		{ 318, 463, 60 },
		{ 318, 464, 60 },
		{ 318, 465, 60 },
		{ 318, 466, 60 },
		{ 318, 467, 60 },
		{ 318, 468, 60 },
		{ 318, 469, 60 },
		{ 318, 470, 60 },
		{ 318, 471, 60 },
		{ 318, 472, 60 },
		{ 318, 473, 60 },
		{ 318, 474, 60 },
		{ 318, 475, 60 },
		{ 318, 476, 60 },
		{ 318, 477, 60 },
		{ 318, 478, 60 },
		{ 318, 479, 60 },
		{ 318, 480, 60 },
		{ 318, 481, 60 },
		{ 318, 482, 60 },
		{ 318, 483, 60 },
		{ 318, 484, 60 },
		{ 318, 485, 60 },
		{ 318, 486, 60 },
		{ 318, 487, 60 },
		{ 318, 488, 60 },
		{ 318, 489, 60 },
		{ 318, 490, 60 },
		{ 318, 491, 60 },
		{ 318, 492, 60 },
		{ 318, 493, 60 },
		{ 318, 494, 60 },
		{ 318, 495, 60 },
		{ 318, 496, 60 },
		{ 318, 497, 60 },
		{ 318, 498, 60 },
		{ 318, 499, 60 },
		{ 318, 500, 60 },
		{ 318, 501, 60 },
		{ 318, 502, 60 },
		{ 318, 503, 60 },
		{ 318, 504, 60 },
		{ 318, 505, 60 },
		{ 318, 506, 60 },
		{ 318, 507, 60 },
		{ 318, 508, 60 },
		{ 318, 509, 60 },
		{ 318, 510, 60 },
		{ 318, 511, 60 },
		{ 318, 512, 60 },
		{ 319, 423, 60 },
		{ 319, 424, 60 },
		{ 319, 425, 60 },
		{ 319, 426, 60 },
		{ 319, 427, 60 },
		{ 319, 428, 60 },
		{ 319, 429, 60 },
		{ 319, 430, 60 },
		{ 319, 431, 60 },
		{ 319, 432, 60 },
		{ 319, 433, 60 },
		{ 319, 434, 60 },
		{ 319, 435, 60 },
		{ 319, 436, 60 },
		{ 319, 437, 60 },
		{ 319, 438, 60 },
		{ 319, 439, 60 },
		{ 319, 440, 60 },
		{ 319, 441, 60 },
		{ 319, 442, 60 },
		{ 319, 443, 60 },
		{ 319, 444, 60 },
		{ 319, 445, 60 },
		{ 319, 446, 60 },
		{ 319, 447, 60 },
		{ 319, 448, 60 },
		{ 319, 449, 60 },
		{ 319, 450, 60 },
		{ 319, 451, 60 },
		{ 319, 452, 60 },
		{ 319, 453, 60 },
		{ 319, 454, 60 },
		{ 319, 455, 60 },
		{ 319, 456, 60 },
		{ 319, 457, 60 },
		{ 319, 458, 60 },
		{ 319, 459, 60 },
		{ 319, 460, 60 },
		{ 319, 461, 60 },
		{ 319, 462, 60 },
		{ 319, 463, 60 },
		{ 319, 464, 60 },
		{ 319, 465, 60 },
		{ 319, 466, 60 },
		{ 319, 467, 60 },
		{ 319, 468, 60 },
		{ 319, 469, 60 },
		{ 319, 470, 60 },
		{ 319, 471, 60 },
		{ 319, 472, 60 },
		{ 319, 473, 60 },
		{ 319, 474, 60 },
		{ 319, 475, 60 },
		{ 319, 476, 60 },
		{ 319, 477, 60 },
		{ 319, 478, 60 },
		{ 319, 479, 60 },
		{ 319, 480, 60 },
		{ 319, 481, 60 },
		{ 319, 482, 60 },
		{ 319, 483, 60 },
		{ 319, 484, 60 },
		{ 319, 485, 60 },
		{ 319, 486, 60 },
		{ 319, 487, 60 },
		{ 319, 488, 60 },
		{ 319, 489, 60 },
		{ 319, 490, 60 },
		{ 319, 491, 60 },
		{ 319, 492, 60 },
		{ 319, 493, 60 },
		{ 319, 494, 60 },
		{ 319, 495, 60 },
		{ 319, 496, 60 },
		{ 319, 497, 60 },
		{ 319, 498, 60 },
		{ 319, 499, 60 },
		{ 319, 500, 60 },
		{ 319, 501, 60 },
		{ 319, 502, 60 },
		{ 319, 503, 60 },
		{ 319, 504, 60 },
		{ 319, 505, 60 },
		{ 319, 506, 60 },
		{ 319, 507, 60 },
		{ 319, 508, 60 },
		{ 319, 509, 60 },
		{ 319, 510, 60 },
		{ 319, 511, 60 },
		{ 319, 512, 60 },
		{ 320, 426, 60 },
		{ 320, 427, 60 },
		{ 320, 428, 60 },
		{ 320, 429, 60 },
		{ 320, 430, 60 },
		{ 320, 431, 60 },
		{ 320, 432, 60 },
		{ 320, 433, 60 },
		{ 320, 434, 60 },
		{ 320, 435, 60 },
		{ 320, 436, 60 },
		{ 320, 437, 60 },
		{ 320, 438, 60 },
		{ 320, 439, 60 },
		{ 320, 440, 60 },
		{ 320, 441, 60 },
		{ 320, 442, 60 },
		{ 320, 443, 60 },
		{ 320, 444, 60 },
		{ 320, 445, 60 },
		{ 320, 446, 60 },
		{ 320, 447, 60 },
		{ 320, 448, 60 },
		{ 320, 449, 60 },
		{ 320, 450, 60 },
		{ 320, 451, 60 },
		{ 320, 452, 60 },
		{ 320, 453, 60 },
		{ 320, 454, 60 },
		{ 320, 455, 60 },
		{ 320, 456, 60 },
		{ 320, 457, 60 },
		{ 320, 458, 60 },
		{ 320, 459, 60 },
		{ 320, 460, 60 },
		{ 320, 461, 60 },
		{ 320, 462, 60 },
		{ 320, 463, 60 },
		{ 320, 464, 60 },
		{ 320, 465, 60 },
		{ 320, 466, 60 },
		{ 320, 467, 60 },
		{ 320, 468, 60 },
		{ 320, 469, 60 },
		{ 320, 470, 60 },
		{ 320, 471, 60 },
		{ 320, 472, 60 },
		{ 320, 473, 60 },
		{ 320, 474, 60 },
		{ 320, 475, 60 },
		{ 320, 476, 60 },
		{ 320, 477, 60 },
		{ 320, 478, 60 },
		{ 320, 479, 60 },
		{ 320, 480, 60 },
		{ 320, 481, 60 },
		{ 320, 482, 60 },
		{ 320, 483, 60 },
		{ 320, 484, 60 },
		{ 320, 485, 60 },
		{ 320, 486, 60 },
		{ 320, 487, 60 },
		{ 320, 488, 60 },
		{ 320, 489, 60 },
		{ 320, 490, 60 },
		{ 320, 491, 60 },
		{ 320, 492, 60 },
		{ 320, 493, 60 },
		{ 320, 494, 60 },
		{ 320, 495, 60 },
		{ 320, 496, 60 },
		{ 320, 497, 60 },
		{ 320, 498, 60 },
		{ 320, 499, 60 },
		{ 320, 500, 60 },
		{ 320, 501, 60 },
		{ 320, 502, 60 },
		{ 320, 503, 60 },
		{ 320, 504, 60 },
		{ 320, 505, 60 },
		{ 320, 506, 60 },
		{ 320, 507, 60 },
		{ 320, 508, 60 },
		{ 320, 509, 60 },
		{ 320, 510, 60 },
		{ 320, 511, 60 },
		{ 320, 512, 60 },
		{ 321, 428, 60 },
		{ 321, 429, 60 },
		{ 321, 430, 60 },
		{ 321, 431, 60 },
		{ 321, 432, 60 },
		{ 321, 433, 60 },
		{ 321, 434, 60 },
		{ 321, 435, 60 },
		{ 321, 436, 60 },
		{ 321, 437, 60 },
		{ 321, 438, 60 },
		{ 321, 439, 60 },
		{ 321, 440, 60 },
		{ 321, 441, 60 },
		{ 321, 442, 60 },
		{ 321, 443, 60 },
		{ 321, 444, 60 },
		{ 321, 445, 60 },
		{ 321, 446, 60 },
		{ 321, 447, 60 },
		{ 321, 448, 60 },
		{ 321, 449, 60 },
		{ 321, 450, 60 },
		{ 321, 451, 60 },
		{ 321, 452, 60 },
		{ 321, 453, 60 },
		{ 321, 454, 60 },
		{ 321, 455, 60 },
		{ 321, 456, 60 },
		{ 321, 457, 60 },
		{ 321, 458, 60 },
		{ 321, 459, 60 },
		{ 321, 460, 60 },
		{ 321, 461, 60 },
		{ 321, 462, 60 },
		{ 321, 463, 60 },
		{ 321, 464, 60 },
		{ 321, 465, 60 },
		{ 321, 466, 60 },
		{ 321, 467, 60 },
		{ 321, 468, 60 },
		{ 321, 469, 60 },
		{ 321, 470, 60 },
		{ 321, 471, 60 },
		{ 321, 472, 60 },
		{ 321, 473, 60 },
		{ 321, 474, 60 },
		{ 321, 475, 60 },
		{ 321, 476, 60 },
		{ 321, 477, 60 },
		{ 321, 478, 60 },
		{ 321, 479, 60 },
		{ 321, 480, 60 },
		{ 321, 481, 60 },
		{ 321, 482, 60 },
		{ 321, 483, 60 },
		{ 321, 484, 60 },
		{ 321, 485, 60 },
		{ 321, 486, 60 },
		{ 321, 487, 60 },
		{ 321, 488, 60 },
		{ 321, 489, 60 },
		{ 321, 490, 60 },
		{ 321, 491, 60 },
		{ 321, 492, 60 },
		{ 321, 493, 60 },
		{ 321, 494, 60 },
		{ 321, 495, 60 },
		{ 321, 496, 60 },
		{ 321, 497, 60 },
		{ 321, 498, 60 },
		{ 321, 499, 60 },
		{ 321, 500, 60 },
		{ 321, 501, 60 },
		{ 321, 502, 60 },
		{ 321, 503, 60 },
		{ 321, 504, 60 },
		{ 321, 505, 60 },
		{ 321, 506, 60 },
		{ 321, 507, 60 },
		{ 321, 508, 60 },
		{ 321, 509, 60 },
		{ 321, 510, 60 },
		{ 321, 511, 60 },
		{ 321, 512, 60 },
		{ 322, 431, 60 },
		{ 322, 432, 60 },
		{ 322, 433, 60 },
		{ 322, 434, 60 },
		{ 322, 435, 60 },
		{ 322, 436, 60 },
		{ 322, 437, 60 },
		{ 322, 438, 60 },
		{ 322, 439, 60 },
		{ 322, 440, 60 },
		{ 322, 441, 60 },
		{ 322, 442, 60 },
		{ 322, 443, 60 },
		{ 322, 444, 60 },
		{ 322, 445, 60 },
		{ 322, 446, 60 },
		{ 322, 447, 60 },
		{ 322, 448, 60 },
		{ 322, 449, 60 },
		{ 322, 450, 60 },
		{ 322, 451, 60 },
		{ 322, 452, 60 },
		{ 322, 453, 60 },
		{ 322, 454, 60 },
		{ 322, 455, 60 },
		{ 322, 456, 60 },
		{ 322, 457, 60 },
		{ 322, 458, 60 },
		{ 322, 459, 60 },
		{ 322, 460, 60 },
		{ 322, 461, 60 },
		{ 322, 462, 60 },
		{ 322, 463, 60 },
		{ 322, 464, 60 },
		{ 322, 465, 60 },
		{ 322, 466, 60 },
		{ 322, 467, 60 },
		{ 322, 468, 60 },
		{ 322, 469, 60 },
		{ 322, 470, 60 },
		{ 322, 471, 60 },
		{ 322, 472, 60 },
		{ 322, 473, 60 },
		{ 322, 474, 60 },
		{ 322, 475, 60 },
		{ 322, 476, 60 },
		{ 322, 477, 60 },
		{ 322, 478, 60 },
		{ 322, 479, 60 },
		{ 322, 480, 60 },
		{ 322, 481, 60 },
		{ 322, 482, 60 },
		{ 322, 483, 60 },
		{ 322, 484, 60 },
		{ 322, 485, 60 },
		{ 322, 486, 60 },
		{ 322, 487, 60 },
		{ 322, 488, 60 },
		{ 322, 489, 60 },
		{ 322, 490, 60 },
		{ 322, 491, 60 },
		{ 322, 492, 60 },
		{ 322, 493, 60 },
		{ 322, 494, 60 },
		{ 322, 495, 60 },
		{ 322, 496, 60 },
		{ 322, 497, 60 },
		{ 322, 498, 60 },
		{ 322, 499, 60 },
		{ 322, 500, 60 },
		{ 322, 501, 60 },
		{ 322, 502, 60 },
		{ 322, 503, 60 },
		{ 322, 504, 60 },
		{ 322, 505, 60 },
		{ 322, 506, 60 },
		{ 322, 507, 60 },
		{ 322, 508, 60 },
		{ 322, 509, 60 },
		{ 322, 510, 60 },
		{ 322, 511, 60 },
		{ 322, 512, 60 },
		{ 323, 434, 60 },
		{ 323, 435, 60 },
		{ 323, 436, 60 },
		{ 323, 437, 60 },
		{ 323, 438, 60 },
		{ 323, 439, 60 },
		{ 323, 440, 60 },
		{ 323, 441, 60 },
		{ 323, 442, 60 },
		{ 323, 443, 60 },
		{ 323, 444, 60 },
		{ 323, 445, 60 },
		{ 323, 446, 60 },
		{ 323, 447, 60 },
		{ 323, 448, 60 },
		{ 323, 449, 60 },
		{ 323, 450, 60 },
		{ 323, 451, 60 },
		{ 323, 452, 60 },
		{ 323, 453, 60 },
		{ 323, 454, 60 },
		{ 323, 455, 60 },
		{ 323, 456, 60 },
		{ 323, 457, 60 },
		{ 323, 458, 60 },
		{ 323, 459, 60 },
		{ 323, 460, 60 },
		{ 323, 461, 60 },
		{ 323, 462, 60 },
		{ 323, 463, 60 },
		{ 323, 464, 60 },
		{ 323, 465, 60 },
		{ 323, 466, 60 },
		{ 323, 467, 60 },
		{ 323, 468, 60 },
		{ 323, 469, 60 },
		{ 323, 470, 60 },
		{ 323, 471, 60 },
		{ 323, 472, 60 },
		{ 323, 473, 60 },
		{ 323, 474, 60 },
		{ 323, 475, 60 },
		{ 323, 476, 60 },
		{ 323, 477, 60 },
		{ 323, 478, 60 },
		{ 323, 479, 60 },
		{ 323, 480, 60 },
		{ 323, 481, 60 },
		{ 323, 482, 60 },
		{ 323, 483, 60 },
		{ 323, 484, 60 },
		{ 323, 485, 60 },
		{ 323, 486, 60 },
		{ 323, 487, 60 },
		{ 323, 488, 60 },
		{ 323, 489, 60 },
		{ 323, 490, 60 },
		{ 323, 491, 60 },
		{ 323, 492, 60 },
		{ 323, 493, 60 },
		{ 323, 494, 60 },
		{ 323, 495, 60 },
		{ 323, 496, 60 },
		{ 323, 497, 60 },
		{ 323, 498, 60 },
		{ 323, 499, 60 },
		{ 323, 500, 60 },
		{ 323, 501, 60 },
		{ 323, 502, 60 },
		{ 323, 503, 60 },
		{ 323, 504, 60 },
		{ 323, 505, 60 },
		{ 323, 506, 60 },
		{ 323, 507, 60 },
		{ 323, 508, 60 },
		{ 323, 509, 60 },
		{ 323, 510, 60 },
		{ 323, 511, 60 },
		{ 323, 512, 60 },
		{ 324, 436, 60 },
		{ 324, 437, 60 },
		{ 324, 438, 60 },
		{ 324, 439, 60 },
		{ 324, 440, 60 },
		{ 324, 441, 60 },
		{ 324, 442, 60 },
		{ 324, 443, 60 },
		{ 324, 444, 60 },
		{ 324, 445, 60 },
		{ 324, 446, 60 },
		{ 324, 447, 60 },
		{ 324, 448, 60 },
		{ 324, 449, 60 },
		{ 324, 450, 60 },
		{ 324, 451, 60 },
		{ 324, 452, 60 },
		{ 324, 453, 60 },
		{ 324, 454, 60 },
		{ 324, 455, 60 },
		{ 324, 456, 60 },
		{ 324, 457, 60 },
		{ 324, 458, 60 },
		{ 324, 459, 60 },
		{ 324, 460, 60 },
		{ 324, 461, 60 },
		{ 324, 462, 60 },
		{ 324, 463, 60 },
		{ 324, 464, 60 },
		{ 324, 465, 60 },
		{ 324, 466, 60 },
		{ 324, 467, 60 },
		{ 324, 468, 60 },
		{ 324, 469, 60 },
		{ 324, 470, 60 },
		{ 324, 471, 60 },
		{ 324, 472, 60 },
		{ 324, 473, 60 },
		{ 324, 474, 60 },
		{ 324, 475, 60 },
		{ 324, 476, 60 },
		{ 324, 477, 60 },
		{ 324, 478, 60 },
		{ 324, 479, 60 },
		{ 324, 480, 60 },
		{ 324, 481, 60 },
		{ 324, 482, 60 },
		{ 324, 483, 60 },
		{ 324, 484, 60 },
		{ 324, 485, 60 },
		{ 324, 486, 60 },
		{ 324, 487, 60 },
		{ 324, 488, 60 },
		{ 324, 489, 60 },
		{ 324, 490, 60 },
		{ 324, 491, 60 },
		{ 324, 492, 60 },
		{ 324, 493, 60 },
		{ 324, 494, 60 },
		{ 324, 495, 60 },
		{ 324, 496, 60 },
		{ 324, 497, 60 },
		{ 324, 498, 60 },
		{ 324, 499, 60 },
		{ 324, 500, 60 },
		{ 324, 501, 60 },
		{ 324, 502, 60 },
		{ 324, 503, 60 },
		{ 324, 504, 60 },
		{ 324, 505, 60 },
		{ 324, 506, 60 },
		{ 324, 507, 60 },
		{ 324, 508, 60 },
		{ 324, 509, 60 },
		{ 324, 510, 60 },
		{ 324, 511, 60 },
		{ 324, 512, 60 },
		{ 325, 439, 60 },
		{ 325, 440, 60 },
		{ 325, 441, 60 },
		{ 325, 442, 60 },
		{ 325, 443, 60 },
		{ 325, 444, 60 },
		{ 325, 445, 60 },
		{ 325, 446, 60 },
		{ 325, 447, 60 },
		{ 325, 448, 60 },
		{ 325, 449, 60 },
		{ 325, 450, 60 },
		{ 325, 451, 60 },
		{ 325, 452, 60 },
		{ 325, 453, 60 },
		{ 325, 454, 60 },
		{ 325, 455, 60 },
		{ 325, 456, 60 },
		{ 325, 457, 60 },
		{ 325, 458, 60 },
		{ 325, 459, 60 },
		{ 325, 460, 60 },
		{ 325, 461, 60 },
		{ 325, 462, 60 },
		{ 325, 463, 60 },
		{ 325, 464, 60 },
		{ 325, 465, 60 },
		{ 325, 466, 60 },
		{ 325, 467, 60 },
		{ 325, 468, 60 },
		{ 325, 469, 60 },
		{ 325, 470, 60 },
		{ 325, 471, 60 },
		{ 325, 472, 60 },
		{ 325, 473, 60 },
		{ 325, 474, 60 },
		{ 325, 475, 60 },
		{ 325, 476, 60 },
		{ 325, 477, 60 },
		{ 325, 478, 60 },
		{ 325, 479, 60 },
		{ 325, 480, 60 },
		{ 325, 481, 60 },
		{ 325, 482, 60 },
		{ 325, 483, 60 },
		{ 325, 484, 60 },
		{ 325, 485, 60 },
		{ 325, 486, 60 },
		{ 325, 487, 60 },
		{ 325, 488, 60 },
		{ 325, 489, 60 },
		{ 325, 490, 60 },
		{ 325, 491, 60 },
		{ 325, 492, 60 },
		{ 325, 493, 60 },
		{ 325, 494, 60 },
		{ 325, 495, 60 },
		{ 325, 496, 60 },
		{ 325, 497, 60 },
		{ 325, 498, 60 },
		{ 325, 499, 60 },
		{ 325, 500, 60 },
		{ 325, 501, 60 },
		{ 325, 502, 60 },
		{ 325, 503, 60 },
		{ 325, 504, 60 },
		{ 325, 505, 60 },
		{ 325, 506, 60 },
		{ 325, 507, 60 },
		{ 325, 508, 60 },
		{ 325, 509, 60 },
		{ 325, 510, 60 },
		{ 325, 511, 60 },
		{ 325, 512, 60 },
		{ 326, 442, 60 },
		{ 326, 443, 60 },
		{ 326, 444, 60 },
		{ 326, 445, 60 },
		{ 326, 446, 60 },
		{ 326, 447, 60 },
		{ 326, 448, 60 },
		{ 326, 449, 60 },
		{ 326, 450, 60 },
		{ 326, 451, 60 },
		{ 326, 452, 60 },
		{ 326, 453, 60 },
		{ 326, 454, 60 },
		{ 326, 455, 60 },
		{ 326, 456, 60 },
		{ 326, 457, 60 },
		{ 326, 458, 60 },
		{ 326, 459, 60 },
		{ 326, 460, 60 },
		{ 326, 461, 60 },
		{ 326, 462, 60 },
		{ 326, 463, 60 },
		{ 326, 464, 60 },
		{ 326, 465, 60 },
		{ 326, 466, 60 },
		{ 326, 467, 60 },
		{ 326, 468, 60 },
		{ 326, 469, 60 },
		{ 326, 470, 60 },
		{ 326, 471, 60 },
		{ 326, 472, 60 },
		{ 326, 473, 60 },
		{ 326, 474, 60 },
		{ 326, 475, 60 },
		{ 326, 476, 60 },
		{ 326, 477, 60 },
		{ 326, 478, 60 },
		{ 326, 479, 60 },
		{ 326, 480, 60 },
		{ 326, 481, 60 },
		{ 326, 482, 60 },
		{ 326, 483, 60 },
		{ 326, 484, 60 },
		{ 326, 485, 60 },
		{ 326, 486, 60 },
		{ 326, 487, 60 },
		{ 326, 488, 60 },
		{ 326, 489, 60 },
		{ 326, 490, 60 },
		{ 326, 491, 60 },
		{ 326, 492, 60 },
		{ 326, 493, 60 },
		{ 326, 494, 60 },
		{ 326, 495, 60 },
		{ 326, 496, 60 },
		{ 326, 497, 60 },
		{ 326, 498, 60 },
		{ 326, 499, 60 },
		{ 326, 500, 60 },
		{ 326, 501, 60 },
		{ 326, 502, 60 },
		{ 326, 503, 60 },
		{ 326, 504, 60 },
		{ 326, 505, 60 },
		{ 326, 506, 60 },
		{ 326, 507, 60 },
		{ 326, 508, 60 },
		{ 326, 509, 60 },
		{ 326, 510, 60 },
		{ 326, 511, 60 },
		{ 326, 512, 60 },
		{ 327, 444, 60 },
		{ 327, 445, 60 },
		{ 327, 446, 60 },
		{ 327, 447, 60 },
		{ 327, 448, 60 },
		{ 327, 449, 60 },
		{ 327, 450, 60 },
		{ 327, 451, 60 },
		{ 327, 452, 60 },
		{ 327, 453, 60 },
		{ 327, 454, 60 },
		{ 327, 455, 60 },
		{ 327, 456, 60 },
		{ 327, 457, 60 },
		{ 327, 458, 60 },
		{ 327, 459, 60 },
		{ 327, 460, 60 },
		{ 327, 461, 60 },
		{ 327, 462, 60 },
		{ 327, 463, 60 },
		{ 327, 464, 60 },
		{ 327, 465, 60 },
		{ 327, 466, 60 },
		{ 327, 467, 60 },
		{ 327, 468, 60 },
		{ 327, 469, 60 },
		{ 327, 470, 60 },
		{ 327, 471, 60 },
		{ 327, 472, 60 },
		{ 327, 473, 60 },
		{ 327, 474, 60 },
		{ 327, 475, 60 },
		{ 327, 476, 60 },
		{ 327, 477, 60 },
		{ 327, 478, 60 },
		{ 327, 479, 60 },
		{ 327, 480, 60 },
		{ 327, 481, 60 },
		{ 327, 482, 60 },
		{ 327, 483, 60 },
		{ 327, 484, 60 },
		{ 327, 485, 60 },
		{ 327, 486, 60 },
		{ 327, 487, 60 },
		{ 327, 488, 60 },
		{ 327, 489, 60 },
		{ 327, 490, 60 },
		{ 327, 491, 60 },
		{ 327, 492, 60 },
		{ 327, 493, 60 },
		{ 327, 494, 60 },
		{ 327, 495, 60 },
		{ 327, 496, 60 },
		{ 327, 497, 60 },
		{ 327, 498, 60 },
		{ 327, 499, 60 },
		{ 327, 500, 60 },
		{ 327, 501, 60 },
		{ 327, 502, 60 },
		{ 327, 503, 60 },
		{ 327, 504, 60 },
		{ 327, 505, 60 },
		{ 327, 506, 60 },
		{ 327, 507, 60 },
		{ 327, 508, 60 },
		{ 327, 509, 60 },
		{ 327, 510, 60 },
		{ 327, 511, 60 },
		{ 327, 512, 60 },
		{ 328, 447, 60 },
		{ 328, 448, 60 },
		{ 328, 449, 60 },
		{ 328, 450, 60 },
		{ 328, 451, 60 },
		{ 328, 452, 60 },
		{ 328, 453, 60 },
		{ 328, 454, 60 },
		{ 328, 455, 60 },
		{ 328, 456, 60 },
		{ 328, 457, 60 },
		{ 328, 458, 60 },
		{ 328, 459, 60 },
		{ 328, 460, 60 },
		{ 328, 461, 60 },
		{ 328, 462, 60 },
		{ 328, 463, 60 },
		{ 328, 464, 60 },
		{ 328, 465, 60 },
		{ 328, 466, 60 },
		{ 328, 467, 60 },
		{ 328, 468, 60 },
		{ 328, 469, 60 },
		{ 328, 470, 60 },
		{ 328, 471, 60 },
		{ 328, 472, 60 },
		{ 328, 473, 60 },
		{ 328, 474, 60 },
		{ 328, 475, 60 },
		{ 328, 476, 60 },
		{ 328, 477, 60 },
		{ 328, 478, 60 },
		{ 328, 479, 60 },
		{ 328, 480, 60 },
		{ 328, 481, 60 },
		{ 328, 482, 60 },
		{ 328, 483, 60 },
		{ 328, 484, 60 },
		{ 328, 485, 60 },
		{ 328, 486, 60 },
		{ 328, 487, 60 },
		{ 328, 488, 60 },
		{ 328, 489, 60 },
		{ 328, 490, 60 },
		{ 328, 491, 60 },
		{ 328, 492, 60 },
		{ 328, 493, 60 },
		{ 328, 494, 60 },
		{ 328, 495, 60 },
		{ 328, 496, 60 },
		{ 328, 497, 60 },
		{ 328, 498, 60 },
		{ 328, 499, 60 },
		{ 328, 500, 60 },
		{ 328, 501, 60 },
		{ 328, 502, 60 },
		{ 328, 503, 60 },
		{ 328, 504, 60 },
		{ 328, 505, 60 },
		{ 328, 506, 60 },
		{ 328, 507, 60 },
		{ 328, 508, 60 },
		{ 328, 509, 60 },
		{ 328, 510, 60 },
		{ 328, 511, 60 },
		{ 328, 512, 60 },
		{ 329, 450, 60 },
		{ 329, 451, 60 },
		{ 329, 452, 60 },
		{ 329, 453, 60 },
		{ 329, 454, 60 },
		{ 329, 455, 60 },
		{ 329, 456, 60 },
		{ 329, 457, 60 },
		{ 329, 458, 60 },
		{ 329, 459, 60 },
		{ 329, 460, 60 },
		{ 329, 461, 60 },
		{ 329, 462, 60 },
		{ 329, 463, 60 },
		{ 329, 464, 60 },
		{ 329, 465, 60 },
		{ 329, 466, 60 },
		{ 329, 467, 60 },
		{ 329, 468, 60 },
		{ 329, 469, 60 },
		{ 329, 470, 60 },
		{ 329, 471, 60 },
		{ 329, 472, 60 },
		{ 329, 473, 60 },
		{ 329, 474, 60 },
		{ 329, 475, 60 },
		{ 329, 476, 60 },
		{ 329, 477, 60 },
		{ 329, 478, 60 },
		{ 329, 479, 60 },
		{ 329, 480, 60 },
		{ 329, 481, 60 },
		{ 329, 482, 60 },
		{ 329, 483, 60 },
		{ 329, 484, 60 },
		{ 329, 485, 60 },
		{ 329, 486, 60 },
		{ 329, 487, 60 },
		{ 329, 488, 60 },
		{ 329, 489, 60 },
		{ 329, 490, 60 },
		{ 329, 491, 60 },
		{ 329, 492, 60 },
		{ 329, 493, 60 },
		{ 329, 494, 60 },
		{ 329, 495, 60 },
		{ 329, 496, 60 },
		{ 329, 497, 60 },
		{ 329, 498, 60 },
		{ 329, 499, 60 },
		{ 329, 500, 60 },
		{ 329, 501, 60 },
		{ 329, 502, 60 },
		{ 329, 503, 60 },
		{ 329, 504, 60 },
		{ 329, 505, 60 },
		{ 329, 506, 60 },
		{ 329, 507, 60 },
		{ 329, 508, 60 },
		{ 329, 509, 60 },
		{ 329, 510, 60 },
		{ 329, 511, 60 },
		{ 329, 512, 60 },
		{ 330, 452, 60 },
		{ 330, 453, 60 },
		{ 330, 454, 60 },
		{ 330, 455, 60 },
		{ 330, 456, 60 },
		{ 330, 457, 60 },
		{ 330, 458, 60 },
		{ 330, 459, 60 },
		{ 330, 460, 60 },
		{ 330, 461, 60 },
		{ 330, 462, 60 },
		{ 330, 463, 60 },
		{ 330, 464, 60 },
		{ 330, 465, 60 },
		{ 330, 466, 60 },
		{ 330, 467, 60 },
		{ 330, 468, 60 },
		{ 330, 469, 60 },
		{ 330, 470, 60 },
		{ 330, 471, 60 },
		{ 330, 472, 60 },
		{ 330, 473, 60 },
		{ 330, 474, 60 },
		{ 330, 475, 60 },
		{ 330, 476, 60 },
		{ 330, 477, 60 },
		{ 330, 478, 60 },
		{ 330, 479, 60 },
		{ 330, 480, 60 },
		{ 330, 481, 60 },
		{ 330, 482, 60 },
		{ 330, 483, 60 },
		{ 330, 484, 60 },
		{ 330, 485, 60 },
		{ 330, 486, 60 },
		{ 330, 487, 60 },
		{ 330, 488, 60 },
		{ 330, 489, 60 },
		{ 330, 490, 60 },
		{ 330, 491, 60 },
		{ 330, 492, 60 },
		{ 330, 493, 60 },
		{ 330, 494, 60 },
		{ 330, 495, 60 },
		{ 330, 496, 60 },
		{ 330, 497, 60 },
		{ 330, 498, 60 },
		{ 330, 499, 60 },
		{ 330, 500, 60 },
		{ 330, 501, 60 },
		{ 330, 502, 60 },
		{ 330, 503, 60 },
		{ 330, 504, 60 },
		{ 330, 505, 60 },
		{ 330, 506, 60 },
		{ 330, 507, 60 },
		{ 330, 508, 60 },
		{ 330, 509, 60 },
		{ 330, 510, 60 },
		{ 330, 511, 60 },
		{ 330, 512, 60 },
		{ 331, 455, 60 },
		{ 331, 456, 60 },
		{ 331, 457, 60 },
		{ 331, 458, 60 },
		{ 331, 459, 60 },
		{ 331, 460, 60 },
		{ 331, 461, 60 },
		{ 331, 462, 60 },
		{ 331, 463, 60 },
		{ 331, 464, 60 },
		{ 331, 465, 60 },
		{ 331, 466, 60 },
		{ 331, 467, 60 },
		{ 331, 468, 60 },
		{ 331, 469, 60 },
		{ 331, 470, 60 },
		{ 331, 471, 60 },
		{ 331, 472, 60 },
		{ 331, 473, 60 },
		{ 331, 474, 60 },
		{ 331, 475, 60 },
		{ 331, 476, 60 },
		{ 331, 477, 60 },
		{ 331, 478, 60 },
		{ 331, 479, 60 },
		{ 331, 480, 60 },
		{ 331, 481, 60 },
		{ 331, 482, 60 },
		{ 331, 483, 60 },
		{ 331, 484, 60 },
		{ 331, 485, 60 },
		{ 331, 486, 60 },
		{ 331, 487, 60 },
		{ 331, 488, 60 },
		{ 331, 489, 60 },
		{ 331, 490, 60 },
		{ 331, 491, 60 },
		{ 331, 492, 60 },
		{ 331, 493, 60 },
		{ 331, 494, 60 },
		{ 331, 495, 60 },
		{ 331, 496, 60 },
		{ 331, 497, 60 },
		{ 331, 498, 60 },
		{ 331, 499, 60 },
		{ 331, 500, 60 },
		{ 331, 501, 60 },
		{ 331, 502, 60 },
		{ 331, 503, 60 },
		{ 331, 504, 60 },
		{ 331, 505, 60 },
		{ 331, 506, 60 },
		{ 331, 507, 60 },
		{ 331, 508, 60 },
		{ 331, 509, 60 },
		{ 331, 510, 60 },
		{ 331, 511, 60 },
		{ 331, 512, 60 },
		{ 332, 458, 60 },
		{ 332, 459, 60 },
		{ 332, 460, 60 },
		{ 332, 461, 60 },
		{ 332, 462, 60 },
		{ 332, 463, 60 },
		{ 332, 464, 60 },
		{ 332, 465, 60 },
		{ 332, 466, 60 },
		{ 332, 467, 60 },
		{ 332, 468, 60 },
		{ 332, 469, 60 },
		{ 332, 470, 60 },
		{ 332, 471, 60 },
		{ 332, 472, 60 },
		{ 332, 473, 60 },
		{ 332, 474, 60 },
		{ 332, 475, 60 },
		{ 332, 476, 60 },
		{ 332, 477, 60 },
		{ 332, 478, 60 },
		{ 332, 479, 60 },
		{ 332, 480, 60 },
		{ 332, 481, 60 },
		{ 332, 482, 60 },
		{ 332, 483, 60 },
		{ 332, 484, 60 },
		{ 332, 485, 60 },
		{ 332, 486, 60 },
		{ 332, 487, 60 },
		{ 332, 488, 60 },
		{ 332, 489, 60 },
		{ 332, 490, 60 },
		{ 332, 491, 60 },
		{ 332, 492, 60 },
		{ 332, 493, 60 },
		{ 332, 494, 60 },
		{ 332, 495, 60 },
		{ 332, 496, 60 },
		{ 332, 497, 60 },
		{ 332, 498, 60 },
		{ 332, 499, 60 },
		{ 332, 500, 60 },
		{ 332, 501, 60 },
		{ 332, 502, 60 },
		{ 332, 503, 60 },
		{ 332, 504, 60 },
		{ 332, 505, 60 },
		{ 332, 506, 60 },
		{ 332, 507, 60 },
		{ 332, 508, 60 },
		{ 332, 509, 60 },
		{ 332, 510, 60 },
		{ 332, 511, 60 },
		{ 332, 512, 60 },
		{ 333, 460, 60 },
		{ 333, 461, 60 },
		{ 333, 462, 60 },
		{ 333, 463, 60 },
		{ 333, 464, 60 },
		{ 333, 465, 60 },
		{ 333, 466, 60 },
		{ 333, 467, 60 },
		{ 333, 468, 60 },
		{ 333, 469, 60 },
		{ 333, 470, 60 },
		{ 333, 471, 60 },
		{ 333, 472, 60 },
		{ 333, 473, 60 },
		{ 333, 474, 60 },
		{ 333, 475, 60 },
		{ 333, 476, 60 },
		{ 333, 477, 60 },
		{ 333, 478, 60 },
		{ 333, 479, 60 },
		{ 333, 480, 60 },
		{ 333, 481, 60 },
		{ 333, 482, 60 },
		{ 333, 483, 60 },
		{ 333, 484, 60 },
		{ 333, 485, 60 },
		{ 333, 486, 60 },
		{ 333, 487, 60 },
		{ 333, 488, 60 },
		{ 333, 489, 60 },
		{ 333, 490, 60 },
		{ 333, 491, 60 },
		{ 333, 492, 60 },
		{ 333, 493, 60 },
		{ 333, 494, 60 },
		{ 333, 495, 60 },
		{ 333, 496, 60 },
		{ 333, 497, 60 },
		{ 333, 498, 60 },
		{ 333, 499, 60 },
		{ 333, 500, 60 },
		{ 333, 501, 60 },
		{ 333, 502, 60 },
		{ 333, 503, 60 },
		{ 333, 504, 60 },
		{ 333, 505, 60 },
		{ 333, 506, 60 },
		{ 333, 507, 60 },
		{ 333, 508, 60 },
		{ 333, 509, 60 },
		{ 333, 510, 60 },
		{ 333, 511, 60 },
		{ 333, 512, 60 },
		{ 334, 463, 60 },
		{ 334, 464, 60 },
		{ 334, 465, 60 },
		{ 334, 466, 60 },
		{ 334, 467, 60 },
		{ 334, 468, 60 },
		{ 334, 469, 60 },
		{ 334, 470, 60 },
		{ 334, 471, 60 },
		{ 334, 472, 60 },
		{ 334, 473, 60 },
		{ 334, 474, 60 },
		{ 334, 475, 60 },
		{ 334, 476, 60 },
		{ 334, 477, 60 },
		{ 334, 478, 60 },
		{ 334, 479, 60 },
		{ 334, 480, 60 },
		{ 334, 481, 60 },
		{ 334, 482, 60 },
		{ 334, 483, 60 },
		{ 334, 484, 60 },
		{ 334, 485, 60 },
		{ 334, 486, 60 },
		{ 334, 487, 60 },
		{ 334, 488, 60 },
		{ 334, 489, 60 },
		{ 334, 490, 60 },
		{ 334, 491, 60 },
		{ 334, 492, 60 },
		{ 334, 493, 60 },
		{ 334, 494, 60 },
		{ 334, 495, 60 },
		{ 334, 496, 60 },
		{ 334, 497, 60 },
		{ 334, 498, 60 },
		{ 334, 499, 60 },
		{ 334, 500, 60 },
		{ 334, 501, 60 },
		{ 334, 502, 60 },
		{ 334, 503, 60 },
		{ 334, 504, 60 },
		{ 334, 505, 60 },
		{ 334, 506, 60 },
		{ 334, 507, 60 },
		{ 334, 508, 60 },
		{ 334, 509, 60 },
		{ 334, 510, 60 },
		{ 334, 511, 60 },
		{ 334, 512, 60 },
		{ 335, 466, 60 },
		{ 335, 467, 60 },
		{ 335, 468, 60 },
		{ 335, 469, 60 },
		{ 335, 470, 60 },
		{ 335, 471, 60 },
		{ 335, 472, 60 },
		{ 335, 473, 60 },
		{ 335, 474, 60 },
		{ 335, 475, 60 },
		{ 335, 476, 60 },
		{ 335, 477, 60 },
		{ 335, 478, 60 },
		{ 335, 479, 60 },
		{ 335, 480, 60 },
		{ 335, 481, 60 },
		{ 335, 482, 60 },
		{ 335, 483, 60 },
		{ 335, 484, 60 },
		{ 335, 485, 60 },
		{ 335, 486, 60 },
		{ 335, 487, 60 },
		{ 335, 488, 60 },
		{ 335, 489, 60 },
		{ 335, 490, 60 },
		{ 335, 491, 60 },
		{ 335, 492, 60 },
		{ 335, 493, 60 },
		{ 335, 494, 60 },
		{ 335, 495, 60 },
		{ 335, 496, 60 },
		{ 335, 497, 60 },
		{ 335, 498, 60 },
		{ 335, 499, 60 },
		{ 335, 500, 60 },
		{ 335, 501, 60 },
		{ 335, 502, 60 },
		{ 335, 503, 60 },
		{ 335, 504, 60 },
		{ 335, 505, 60 },
		{ 335, 506, 60 },
		{ 335, 507, 60 },
		{ 335, 508, 60 },
		{ 335, 509, 60 },
		{ 335, 510, 60 },
		{ 335, 511, 60 },
		{ 335, 512, 60 },
		{ 336, 468, 60 },
		{ 336, 469, 60 },
		{ 336, 470, 60 },
		{ 336, 471, 60 },
		{ 336, 472, 60 },
		{ 336, 473, 60 },
		{ 336, 474, 60 },
		{ 336, 475, 60 },
		{ 336, 476, 60 },
		{ 336, 477, 60 },
		{ 336, 478, 60 },
		{ 336, 479, 60 },
		{ 336, 480, 60 },
		{ 336, 481, 60 },
		{ 336, 482, 60 },
		{ 336, 483, 60 },
		{ 336, 484, 60 },
		{ 336, 485, 60 },
		{ 336, 486, 60 },
		{ 336, 487, 60 },
		{ 336, 488, 60 },
		{ 336, 489, 60 },
		{ 336, 490, 60 },
		{ 336, 491, 60 },
		{ 336, 492, 60 },
		{ 336, 493, 60 },
		{ 336, 494, 60 },
		{ 336, 495, 60 },
		{ 336, 496, 60 },
		{ 336, 497, 60 },
		{ 336, 498, 60 },
		{ 336, 499, 60 },
		{ 336, 500, 60 },
		{ 336, 501, 60 },
		{ 336, 502, 60 },
		{ 336, 503, 60 },
		{ 336, 504, 60 },
		{ 336, 505, 60 },
		{ 336, 506, 60 },
		{ 336, 507, 60 },
		{ 336, 508, 60 },
		{ 336, 509, 60 },
		{ 336, 510, 60 },
		{ 336, 511, 60 },
		{ 336, 512, 60 },
		{ 337, 471, 60 },
		{ 337, 472, 60 },
		{ 337, 473, 60 },
		{ 337, 474, 60 },
		{ 337, 475, 60 },
		{ 337, 476, 60 },
		{ 337, 477, 60 },
		{ 337, 478, 60 },
		{ 337, 479, 60 },
		{ 337, 480, 60 },
		{ 337, 481, 60 },
		{ 337, 482, 60 },
		{ 337, 483, 60 },
		{ 337, 484, 60 },
		{ 337, 485, 60 },
		{ 337, 486, 60 },
		{ 337, 487, 60 },
		{ 337, 488, 60 },
		{ 337, 489, 60 },
		{ 337, 490, 60 },
		{ 337, 491, 60 },
		{ 337, 492, 60 },
		{ 337, 493, 60 },
		{ 337, 494, 60 },
		{ 337, 495, 60 },
		{ 337, 496, 60 },
		{ 337, 497, 60 },
		{ 337, 498, 60 },
		{ 337, 499, 60 },
		{ 337, 500, 60 },
		{ 337, 501, 60 },
		{ 337, 502, 60 },
		{ 337, 503, 60 },
		{ 337, 504, 60 },
		{ 337, 505, 60 },
		{ 337, 506, 60 },
		{ 337, 507, 60 },
		{ 337, 508, 60 },
		{ 337, 509, 60 },
		{ 337, 510, 60 },
		{ 337, 511, 60 },
		{ 337, 512, 60 },
		{ 338, 474, 60 },
		{ 338, 475, 60 },
		{ 338, 476, 60 },
		{ 338, 477, 60 },
		{ 338, 478, 60 },
		{ 338, 479, 60 },
		{ 338, 480, 60 },
		{ 338, 481, 60 },
		{ 338, 482, 60 },
		{ 338, 483, 60 },
		{ 338, 484, 60 },
		{ 338, 485, 60 },
		{ 338, 486, 60 },
		{ 338, 487, 60 },
		{ 338, 488, 60 },
		{ 338, 489, 60 },
		{ 338, 490, 60 },
		{ 338, 491, 60 },
		{ 338, 492, 60 },
		{ 338, 493, 60 },
		{ 338, 494, 60 },
		{ 338, 495, 60 },
		{ 338, 496, 60 },
		{ 338, 497, 60 },
		{ 338, 498, 60 },
		{ 338, 499, 60 },
		{ 338, 500, 60 },
		{ 338, 501, 60 },
		{ 338, 502, 60 },
		{ 338, 503, 60 },
		{ 338, 504, 60 },
		{ 338, 505, 60 },
		{ 338, 506, 60 },
		{ 338, 507, 60 },
		{ 338, 508, 60 },
		{ 338, 509, 60 },
		{ 338, 510, 60 },
		{ 338, 511, 60 },
		{ 338, 512, 60 },
		{ 339, 476, 60 },
		{ 339, 477, 60 },
		{ 339, 478, 60 },
		{ 339, 479, 60 },
		{ 339, 480, 60 },
		{ 339, 481, 60 },
		{ 339, 482, 60 },
		{ 339, 483, 60 },
		{ 339, 484, 60 },
		{ 339, 485, 60 },
		{ 339, 486, 60 },
		{ 339, 487, 60 },
		{ 339, 488, 60 },
		{ 339, 489, 60 },
		{ 339, 490, 60 },
		{ 339, 491, 60 },
		{ 339, 492, 60 },
		{ 339, 493, 60 },
		{ 339, 494, 60 },
		{ 339, 495, 60 },
		{ 339, 496, 60 },
		{ 339, 497, 60 },
		{ 339, 498, 60 },
		{ 339, 499, 60 },
		{ 339, 500, 60 },
		{ 339, 501, 60 },
		{ 339, 502, 60 },
		{ 339, 503, 60 },
		{ 339, 504, 60 },
		{ 339, 505, 60 },
		{ 339, 506, 60 },
		{ 339, 507, 60 },
		{ 339, 508, 60 },
		{ 339, 509, 60 },
		{ 339, 510, 60 },
		{ 339, 511, 60 },
		{ 339, 512, 60 },
		{ 340, 479, 60 },
		{ 340, 480, 60 },
		{ 340, 481, 60 },
		{ 340, 482, 60 },
		{ 340, 483, 60 },
		{ 340, 484, 60 },
		{ 340, 485, 60 },
		{ 340, 486, 60 },
		{ 340, 487, 60 },
		{ 340, 488, 60 },
		{ 340, 489, 60 },
		{ 340, 490, 60 },
		{ 340, 491, 60 },
		{ 340, 492, 60 },
		{ 340, 493, 60 },
		{ 340, 494, 60 },
		{ 340, 495, 60 },
		{ 340, 496, 60 },
		{ 340, 497, 60 },
		{ 340, 498, 60 },
		{ 340, 499, 60 },
		{ 340, 500, 60 },
		{ 340, 501, 60 },
		{ 340, 502, 60 },
		{ 340, 503, 60 },
		{ 340, 504, 60 },
		{ 340, 505, 60 },
		{ 340, 506, 60 },
		{ 340, 507, 60 },
		{ 340, 508, 60 },
		{ 340, 509, 60 },
		{ 340, 510, 60 },
		{ 340, 511, 60 },
		{ 340, 512, 60 },
		{ 341, 482, 60 },
		{ 341, 483, 60 },
		{ 341, 484, 60 },
		{ 341, 485, 60 },
		{ 341, 486, 60 },
		{ 341, 487, 60 },
		{ 341, 488, 60 },
		{ 341, 489, 60 },
		{ 341, 490, 60 },
		{ 341, 491, 60 },
		{ 341, 492, 60 },
		{ 341, 493, 60 },
		{ 341, 494, 60 },
		{ 341, 495, 60 },
		{ 341, 496, 60 },
		{ 341, 497, 60 },
		{ 341, 498, 60 },
		{ 341, 499, 60 },
		{ 341, 500, 60 },
		{ 341, 501, 60 },
		{ 341, 502, 60 },
		{ 341, 503, 60 },
		{ 341, 504, 60 },
		{ 341, 505, 60 },
		{ 341, 506, 60 },
		{ 341, 507, 60 },
		{ 341, 508, 60 },
		{ 341, 509, 60 },
		{ 341, 510, 60 },
		{ 341, 511, 60 },
		{ 341, 512, 60 },
		{ 342, 484, 60 },
		{ 342, 485, 60 },
		{ 342, 486, 60 },
		{ 342, 487, 60 },
		{ 342, 488, 60 },
		{ 342, 489, 60 },
		{ 342, 490, 60 },
		{ 342, 491, 60 },
		{ 342, 492, 60 },
		{ 342, 493, 60 },
		{ 342, 494, 60 },
		{ 342, 495, 60 },
		{ 342, 496, 60 },
		{ 342, 497, 60 },
		{ 342, 498, 60 },
		{ 342, 499, 60 },
		{ 342, 500, 60 },
		{ 342, 501, 60 },
		{ 342, 502, 60 },
		{ 342, 503, 60 },
		{ 342, 504, 60 },
		{ 342, 505, 60 },
		{ 342, 506, 60 },
		{ 342, 507, 60 },
		{ 342, 508, 60 },
		{ 342, 509, 60 },
		{ 342, 510, 60 },
		{ 342, 511, 60 },
		{ 342, 512, 60 },
		{ 343, 487, 60 },
		{ 343, 488, 60 },
		{ 343, 489, 60 },
		{ 343, 490, 60 },
		{ 343, 491, 60 },
		{ 343, 492, 60 },
		{ 343, 493, 60 },
		{ 343, 494, 60 },
		{ 343, 495, 60 },
		{ 343, 496, 60 },
		{ 343, 497, 60 },
		{ 343, 498, 60 },
		{ 343, 499, 60 },
		{ 343, 500, 60 },
		{ 343, 501, 60 },
		{ 343, 502, 60 },
		{ 343, 503, 60 },
		{ 343, 504, 60 },
		{ 343, 505, 60 },
		{ 343, 506, 60 },
		{ 343, 507, 60 },
		{ 343, 508, 60 },
		{ 343, 509, 60 },
		{ 343, 510, 60 },
		{ 343, 511, 60 },
		{ 343, 512, 60 },
		{ 344, 490, 60 },
		{ 344, 491, 60 },
		{ 344, 492, 60 },
		{ 344, 493, 60 },
		{ 344, 494, 60 },
		{ 344, 495, 60 },
		{ 344, 496, 60 },
		{ 344, 497, 60 },
		{ 344, 498, 60 },
		{ 344, 499, 60 },
		{ 344, 500, 60 },
		{ 344, 501, 60 },
		{ 344, 502, 60 },
		{ 344, 503, 60 },
		{ 344, 504, 60 },
		{ 344, 505, 60 },
		{ 344, 506, 60 },
		{ 344, 507, 60 },
		{ 344, 508, 60 },
		{ 344, 509, 60 },
		{ 344, 510, 60 },
		{ 344, 511, 60 },
		{ 344, 512, 60 },
		{ 345, 492, 60 },
		{ 345, 493, 60 },
		{ 345, 494, 60 },
		{ 345, 495, 60 },
		{ 345, 496, 60 },
		{ 345, 497, 60 },
		{ 345, 498, 60 },
		{ 345, 499, 60 },
		{ 345, 500, 60 },
		{ 345, 501, 60 },
		{ 345, 502, 60 },
		{ 345, 503, 60 },
		{ 345, 504, 60 },
		{ 345, 505, 60 },
		{ 345, 506, 60 },
		{ 345, 507, 60 },
		{ 345, 508, 60 },
		{ 345, 509, 60 },
		{ 345, 510, 60 },
		{ 345, 511, 60 },
		{ 345, 512, 60 },
		{ 346, 495, 60 },
		{ 346, 496, 60 },
		{ 346, 497, 60 },
		{ 346, 498, 60 },
		{ 346, 499, 60 },
		{ 346, 500, 60 },
		{ 346, 501, 60 },
		{ 346, 502, 60 },
		{ 346, 503, 60 },
		{ 346, 504, 60 },
		{ 346, 505, 60 },
		{ 346, 506, 60 },
		{ 346, 507, 60 },
		{ 346, 508, 60 },
		{ 346, 509, 60 },
		{ 346, 510, 60 },
		{ 346, 511, 60 },
		{ 346, 512, 60 },
		{ 347, 498, 60 },
		{ 347, 499, 60 },
		{ 347, 500, 60 },
		{ 347, 501, 60 },
		{ 347, 502, 60 },
		{ 347, 503, 60 },
		{ 347, 504, 60 },
		{ 347, 505, 60 },
		{ 347, 506, 60 },
		{ 347, 507, 60 },
		{ 347, 508, 60 },
		{ 347, 509, 60 },
		{ 347, 510, 60 },
		{ 347, 511, 60 },
		{ 347, 512, 60 },
		{ 348, 500, 60 },
		{ 348, 501, 60 },
		{ 348, 502, 60 },
		{ 348, 503, 60 },
		{ 348, 504, 60 },
		{ 348, 505, 60 },
		{ 348, 506, 60 },
		{ 348, 507, 60 },
		{ 348, 508, 60 },
		{ 348, 509, 60 },
		{ 348, 510, 60 },
		{ 348, 511, 60 },
		{ 348, 512, 60 },
		{ 349, 503, 60 },
		{ 349, 504, 60 },
		{ 349, 505, 60 },
		{ 349, 506, 60 },
		{ 349, 507, 60 },
		{ 349, 508, 60 },
		{ 349, 509, 60 },
		{ 349, 510, 60 },
		{ 349, 511, 60 },
		{ 349, 512, 60 },
		{ 350, 506, 60 },
		{ 350, 507, 60 },
		{ 350, 508, 60 },
		{ 350, 509, 60 },
		{ 350, 510, 60 },
		{ 350, 511, 60 },
		{ 350, 512, 60 },
		{ 351, 508, 60 },
		{ 351, 509, 60 },
		{ 351, 510, 60 },
		{ 351, 511, 60 },
		{ 351, 512, 60 },
		{ 352, 511, 60 },
		{ 352, 512, 60 },
		{ 194, 1, 61 },
		{ 194, 2, 61 },
		{ 194, 3, 61 },
		{ 194, 4, 61 },
		{ 195, 1, 61 },
		{ 195, 2, 61 },
		{ 195, 3, 61 },
		{ 195, 4, 61 },
		{ 195, 5, 61 },
		{ 195, 6, 61 },
		{ 195, 7, 61 },
		{ 195, 8, 61 },
		{ 196, 1, 61 },
		{ 196, 2, 61 },
		{ 196, 3, 61 },
		{ 196, 4, 61 },
		{ 196, 5, 61 },
		{ 196, 6, 61 },
		{ 196, 7, 61 },
		{ 196, 8, 61 },
		{ 196, 9, 61 },
		{ 196, 10, 61 },
		{ 196, 11, 61 },
		{ 196, 12, 61 },
		{ 197, 1, 61 },
		{ 197, 2, 61 },
		{ 197, 3, 61 },
		{ 197, 4, 61 },
		{ 197, 5, 61 },
		{ 197, 6, 61 },
		{ 197, 7, 61 },
		{ 197, 8, 61 },
		{ 197, 9, 61 },
		{ 197, 10, 61 },
		{ 197, 11, 61 },
		{ 197, 12, 61 },
		{ 197, 13, 61 },
		{ 197, 14, 61 },
		{ 197, 15, 61 },
		{ 197, 16, 61 },
		{ 198, 1, 61 },
		{ 198, 2, 61 },
		{ 198, 3, 61 },
		{ 198, 4, 61 },
		{ 198, 5, 61 },
		{ 198, 6, 61 },
		{ 198, 7, 61 },
		{ 198, 8, 61 },
		{ 198, 9, 61 },
		{ 198, 10, 61 },
		{ 198, 11, 61 },
		{ 198, 12, 61 },
		{ 198, 13, 61 },
		{ 198, 14, 61 },
		{ 198, 15, 61 },
		{ 198, 16, 61 },
		{ 198, 17, 61 },
		{ 198, 18, 61 },
		{ 198, 19, 61 },
		{ 198, 20, 61 },
		{ 199, 1, 61 },
		{ 199, 2, 61 },
		{ 199, 3, 61 },
		{ 199, 4, 61 },
		{ 199, 5, 61 },
		{ 199, 6, 61 },
		{ 199, 7, 61 },
		{ 199, 8, 61 },
		{ 199, 9, 61 },
		{ 199, 10, 61 },
		{ 199, 11, 61 },
		{ 199, 12, 61 },
		{ 199, 13, 61 },
		{ 199, 14, 61 },
		{ 199, 15, 61 },
		{ 199, 16, 61 },
		{ 199, 17, 61 },
		{ 199, 18, 61 },
		{ 199, 19, 61 },
		{ 199, 20, 61 },
		{ 199, 21, 61 },
		{ 199, 22, 61 },
		{ 199, 23, 61 },
		{ 199, 24, 61 },
		{ 200, 1, 61 },
		{ 200, 2, 61 },
		{ 200, 3, 61 },
		{ 200, 4, 61 },
		{ 200, 5, 61 },
		{ 200, 6, 61 },
		{ 200, 7, 61 },
		{ 200, 8, 61 },
		{ 200, 9, 61 },
		{ 200, 10, 61 },
		{ 200, 11, 61 },
		{ 200, 12, 61 },
		{ 200, 13, 61 },
		{ 200, 14, 61 },
		{ 200, 15, 61 },
		{ 200, 16, 61 },
		{ 200, 17, 61 },
		{ 200, 18, 61 },
		{ 200, 19, 61 },
		{ 200, 20, 61 },
		{ 200, 21, 61 },
		{ 200, 22, 61 },
		{ 200, 23, 61 },
		{ 200, 24, 61 },
		{ 200, 25, 61 },
		{ 200, 26, 61 },
		{ 200, 27, 61 },
		{ 200, 28, 61 },
		{ 201, 1, 61 },
		{ 201, 2, 61 },
		{ 201, 3, 61 },
		{ 201, 4, 61 },
		{ 201, 5, 61 },
		{ 201, 6, 61 },
		{ 201, 7, 61 },
		{ 201, 8, 61 },
		{ 201, 9, 61 },
		{ 201, 10, 61 },
		{ 201, 11, 61 },
		{ 201, 12, 61 },
		{ 201, 13, 61 },
		{ 201, 14, 61 },
		{ 201, 15, 61 },
		{ 201, 16, 61 },
		{ 201, 17, 61 },
		{ 201, 18, 61 },
		{ 201, 19, 61 },
		{ 201, 20, 61 },
		{ 201, 21, 61 },
		{ 201, 22, 61 },
		{ 201, 23, 61 },
		{ 201, 24, 61 },
		{ 201, 25, 61 },
		{ 201, 26, 61 },
		{ 201, 27, 61 },
		{ 201, 28, 61 },
		{ 201, 29, 61 },
		{ 201, 30, 61 },
		{ 201, 31, 61 },
		{ 201, 32, 61 },
		{ 202, 1, 61 },
		{ 202, 2, 61 },
		{ 202, 3, 61 },
		{ 202, 4, 61 },
		{ 202, 5, 61 },
		{ 202, 6, 61 },
		{ 202, 7, 61 },
		{ 202, 8, 61 },
		{ 202, 9, 61 },
		{ 202, 10, 61 },
		{ 202, 11, 61 },
		{ 202, 12, 61 },
		{ 202, 13, 61 },
		{ 202, 14, 61 },
		{ 202, 15, 61 },
		{ 202, 16, 61 },
		{ 202, 17, 61 },
		{ 202, 18, 61 },
		{ 202, 19, 61 },
		{ 202, 20, 61 },
		{ 202, 21, 61 },
		{ 202, 22, 61 },
		{ 202, 23, 61 },
		{ 202, 24, 61 },
		{ 202, 25, 61 },
		{ 202, 26, 61 },
		{ 202, 27, 61 },
		{ 202, 28, 61 },
		{ 202, 29, 61 },
		{ 202, 30, 61 },
		{ 202, 31, 61 },
		{ 202, 32, 61 },
		{ 202, 33, 61 },
		{ 202, 34, 61 },
		{ 202, 35, 61 },
		{ 202, 36, 61 },
		{ 203, 1, 61 },
		{ 203, 2, 61 },
		{ 203, 3, 61 },
		{ 203, 4, 61 },
		{ 203, 5, 61 },
		{ 203, 6, 61 },
		{ 203, 7, 61 },
		{ 203, 8, 61 },
		{ 203, 9, 61 },
		{ 203, 10, 61 },
		{ 203, 11, 61 },
		{ 203, 12, 61 },
		{ 203, 13, 61 },
		{ 203, 14, 61 },
		{ 203, 15, 61 },
		{ 203, 16, 61 },
		{ 203, 17, 61 },
		{ 203, 18, 61 },
		{ 203, 19, 61 },
		{ 203, 20, 61 },
		{ 203, 21, 61 },
		{ 203, 22, 61 },
		{ 203, 23, 61 },
		{ 203, 24, 61 },
		{ 203, 25, 61 },
		{ 203, 26, 61 },
		{ 203, 27, 61 },
		{ 203, 28, 61 },
		{ 203, 29, 61 },
		{ 203, 30, 61 },
		{ 203, 31, 61 },
		{ 203, 32, 61 },
		{ 203, 33, 61 },
		{ 203, 34, 61 },
		{ 203, 35, 61 },
		{ 203, 36, 61 },
		{ 203, 37, 61 },
		{ 203, 38, 61 },
		{ 203, 39, 61 },
		{ 203, 40, 61 },
		{ 204, 1, 61 },
		{ 204, 2, 61 },
		{ 204, 3, 61 },
		{ 204, 4, 61 },
		{ 204, 5, 61 },
		{ 204, 6, 61 },
		{ 204, 7, 61 },
		{ 204, 8, 61 },
		{ 204, 9, 61 },
		{ 204, 10, 61 },
		{ 204, 11, 61 },
		{ 204, 12, 61 },
		{ 204, 13, 61 },
		{ 204, 14, 61 },
		{ 204, 15, 61 },
		{ 204, 16, 61 },
		{ 204, 17, 61 },
		{ 204, 18, 61 },
		{ 204, 19, 61 },
		{ 204, 20, 61 },
		{ 204, 21, 61 },
		{ 204, 22, 61 },
		{ 204, 23, 61 },
		{ 204, 24, 61 },
		{ 204, 25, 61 },
		{ 204, 26, 61 },
		{ 204, 27, 61 },
		{ 204, 28, 61 },
		{ 204, 29, 61 },
		{ 204, 30, 61 },
		{ 204, 31, 61 },
		{ 204, 32, 61 },
		{ 204, 33, 61 },
		{ 204, 34, 61 },
		{ 204, 35, 61 },
		{ 204, 36, 61 },
		{ 204, 37, 61 },
		{ 204, 38, 61 },
		{ 204, 39, 61 },
		{ 204, 40, 61 },
		{ 204, 41, 61 },
		{ 204, 42, 61 },
		{ 204, 43, 61 },
		{ 204, 44, 61 },
		{ 205, 1, 61 },
		{ 205, 2, 61 },
		{ 205, 3, 61 },
		{ 205, 4, 61 },
		{ 205, 5, 61 },
		{ 205, 6, 61 },
		{ 205, 7, 61 },
		{ 205, 8, 61 },
		{ 205, 9, 61 },
		{ 205, 10, 61 },
		{ 205, 11, 61 },
		{ 205, 12, 61 },
		{ 205, 13, 61 },
		{ 205, 14, 61 },
		{ 205, 15, 61 },
		{ 205, 16, 61 },
		{ 205, 17, 61 },
		{ 205, 18, 61 },
		{ 205, 19, 61 },
		{ 205, 20, 61 },
		{ 205, 21, 61 },
		{ 205, 22, 61 },
		{ 205, 23, 61 },
		{ 205, 24, 61 },
		{ 205, 25, 61 },
		{ 205, 26, 61 },
		{ 205, 27, 61 },
		{ 205, 28, 61 },
		{ 205, 29, 61 },
		{ 205, 30, 61 },
		{ 205, 31, 61 },
		{ 205, 32, 61 },
		{ 205, 33, 61 },
		{ 205, 34, 61 },
		{ 205, 35, 61 },
		{ 205, 36, 61 },
		{ 205, 37, 61 },
		{ 205, 38, 61 },
		{ 205, 39, 61 },
		{ 205, 40, 61 },
		{ 205, 41, 61 },
		{ 205, 42, 61 },
		{ 205, 43, 61 },
		{ 205, 44, 61 },
		{ 205, 45, 61 },
		{ 205, 46, 61 },
		{ 205, 47, 61 },
		{ 205, 48, 61 },
		{ 206, 1, 61 },
		{ 206, 2, 61 },
		{ 206, 3, 61 },
		{ 206, 4, 61 },
		{ 206, 5, 61 },
		{ 206, 6, 61 },
		{ 206, 7, 61 },
		{ 206, 8, 61 },
		{ 206, 9, 61 },
		{ 206, 10, 61 },
		{ 206, 11, 61 },
		{ 206, 12, 61 },
		{ 206, 13, 61 },
		{ 206, 14, 61 },
		{ 206, 15, 61 },
		{ 206, 16, 61 },
		{ 206, 17, 61 },
		{ 206, 18, 61 },
		{ 206, 19, 61 },
		{ 206, 20, 61 },
		{ 206, 21, 61 },
		{ 206, 22, 61 },
		{ 206, 23, 61 },
		{ 206, 24, 61 },
		{ 206, 25, 61 },
		{ 206, 26, 61 },
		{ 206, 27, 61 },
		{ 206, 28, 61 },
		{ 206, 29, 61 },
		{ 206, 30, 61 },
		{ 206, 31, 61 },
		{ 206, 32, 61 },
		{ 206, 33, 61 },
		{ 206, 34, 61 },
		{ 206, 35, 61 },
		{ 206, 36, 61 },
		{ 206, 37, 61 },
		{ 206, 38, 61 },
		{ 206, 39, 61 },
		{ 206, 40, 61 },
		{ 206, 41, 61 },
		{ 206, 42, 61 },
		{ 206, 43, 61 },
		{ 206, 44, 61 },
		{ 206, 45, 61 },
		{ 206, 46, 61 },
		{ 206, 47, 61 },
		{ 206, 48, 61 },
		{ 206, 49, 61 },
		{ 206, 50, 61 },
		{ 206, 51, 61 },
		{ 206, 52, 61 },
		{ 207, 1, 61 },
		{ 207, 2, 61 },
		{ 207, 3, 61 },
		{ 207, 4, 61 },
		{ 207, 5, 61 },
		{ 207, 6, 61 },
		{ 207, 7, 61 },
		{ 207, 8, 61 },
		{ 207, 9, 61 },
		{ 207, 10, 61 },
		{ 207, 11, 61 },
		{ 207, 12, 61 },
		{ 207, 13, 61 },
		{ 207, 14, 61 },
		{ 207, 15, 61 },
		{ 207, 16, 61 },
		{ 207, 17, 61 },
		{ 207, 18, 61 },
		{ 207, 19, 61 },
		{ 207, 20, 61 },
		{ 207, 21, 61 },
		{ 207, 22, 61 },
		{ 207, 23, 61 },
		{ 207, 24, 61 },
		{ 207, 25, 61 },
		{ 207, 26, 61 },
		{ 207, 27, 61 },
		{ 207, 28, 61 },
		{ 207, 29, 61 },
		{ 207, 30, 61 },
		{ 207, 31, 61 },
		{ 207, 32, 61 },
		{ 207, 33, 61 },
		{ 207, 34, 61 },
		{ 207, 35, 61 },
		{ 207, 36, 61 },
		{ 207, 37, 61 },
		{ 207, 38, 61 },
		{ 207, 39, 61 },
		{ 207, 40, 61 },
		{ 207, 41, 61 },
		{ 207, 42, 61 },
		{ 207, 43, 61 },
		{ 207, 44, 61 },
		{ 207, 45, 61 },
		{ 207, 46, 61 },
		{ 207, 47, 61 },
		{ 207, 48, 61 },
		{ 207, 49, 61 },
		{ 207, 50, 61 },
		{ 207, 51, 61 },
		{ 207, 52, 61 },
		{ 207, 53, 61 },
		{ 207, 54, 61 },
		{ 207, 55, 61 },
		{ 207, 56, 61 },
		{ 208, 1, 61 },
		{ 208, 2, 61 },
		{ 208, 3, 61 },
		{ 208, 4, 61 },
		{ 208, 5, 61 },
		{ 208, 6, 61 },
		{ 208, 7, 61 },
		{ 208, 8, 61 },
		{ 208, 9, 61 },
		{ 208, 10, 61 },
		{ 208, 11, 61 },
		{ 208, 12, 61 },
		{ 208, 13, 61 },
		{ 208, 14, 61 },
		{ 208, 15, 61 },
		{ 208, 16, 61 },
		{ 208, 17, 61 },
		{ 208, 18, 61 },
		{ 208, 19, 61 },
		{ 208, 20, 61 },
		{ 208, 21, 61 },
		{ 208, 22, 61 },
		{ 208, 23, 61 },
		{ 208, 24, 61 },
		{ 208, 25, 61 },
		{ 208, 26, 61 },
		{ 208, 27, 61 },
		{ 208, 28, 61 },
		{ 208, 29, 61 },
		{ 208, 30, 61 },
		{ 208, 31, 61 },
		{ 208, 32, 61 },
		{ 208, 33, 61 },
		{ 208, 34, 61 },
		{ 208, 35, 61 },
		{ 208, 36, 61 },
		{ 208, 37, 61 },
		{ 208, 38, 61 },
		{ 208, 39, 61 },
		{ 208, 40, 61 },
		{ 208, 41, 61 },
		{ 208, 42, 61 },
		{ 208, 43, 61 },
		{ 208, 44, 61 },
		{ 208, 45, 61 },
		{ 208, 46, 61 },
		{ 208, 47, 61 },
		{ 208, 48, 61 },
		{ 208, 49, 61 },
		{ 208, 50, 61 },
		{ 208, 51, 61 },
		{ 208, 52, 61 },
		{ 208, 53, 61 },
		{ 208, 54, 61 },
		{ 208, 55, 61 },
		{ 208, 56, 61 },
		{ 208, 57, 61 },
		{ 208, 58, 61 },
		{ 208, 59, 61 },
		{ 208, 60, 61 },
		{ 209, 1, 61 },
		{ 209, 2, 61 },
		{ 209, 3, 61 },
		{ 209, 4, 61 },
		{ 209, 5, 61 },
		{ 209, 6, 61 },
		{ 209, 7, 61 },
		{ 209, 8, 61 },
		{ 209, 9, 61 },
		{ 209, 10, 61 },
		{ 209, 11, 61 },
		{ 209, 12, 61 },
		{ 209, 13, 61 },
		{ 209, 14, 61 },
		{ 209, 15, 61 },
		{ 209, 16, 61 },
		{ 209, 17, 61 },
		{ 209, 18, 61 },
		{ 209, 19, 61 },
		{ 209, 20, 61 },
		{ 209, 21, 61 },
		{ 209, 22, 61 },
		{ 209, 23, 61 },
		{ 209, 24, 61 },
		{ 209, 25, 61 },
		{ 209, 26, 61 },
		{ 209, 27, 61 },
		{ 209, 28, 61 },
		{ 209, 29, 61 },
		{ 209, 30, 61 },
		{ 209, 31, 61 },
		{ 209, 32, 61 },
		{ 209, 33, 61 },
		{ 209, 34, 61 },
		{ 209, 35, 61 },
		{ 209, 36, 61 },
		{ 209, 37, 61 },
		{ 209, 38, 61 },
		{ 209, 39, 61 },
		{ 209, 40, 61 },
		{ 209, 41, 61 },
		{ 209, 42, 61 },
		{ 209, 43, 61 },
		{ 209, 44, 61 },
		{ 209, 45, 61 },
		{ 209, 46, 61 },
		{ 209, 47, 61 },
		{ 209, 48, 61 },
		{ 209, 49, 61 },
		{ 209, 50, 61 },
		{ 209, 51, 61 },
		{ 209, 52, 61 },
		{ 209, 53, 61 },
		{ 209, 54, 61 },
		{ 209, 55, 61 },
		{ 209, 56, 61 },
		{ 209, 57, 61 },
		{ 209, 58, 61 },
		{ 209, 59, 61 },
		{ 209, 60, 61 },
		{ 209, 61, 61 },
		{ 209, 62, 61 },
		{ 209, 63, 61 },
		{ 209, 64, 61 },
		{ 210, 1, 61 },
		{ 210, 2, 61 },
		{ 210, 3, 61 },
		{ 210, 4, 61 },
		{ 210, 5, 61 },
		{ 210, 6, 61 },
		{ 210, 7, 61 },
		{ 210, 8, 61 },
		{ 210, 9, 61 },
		{ 210, 10, 61 },
		{ 210, 11, 61 },
		{ 210, 12, 61 },
		{ 210, 13, 61 },
		{ 210, 14, 61 },
		{ 210, 15, 61 },
		{ 210, 16, 61 },
		{ 210, 17, 61 },
		{ 210, 18, 61 },
		{ 210, 19, 61 },
		{ 210, 20, 61 },
		{ 210, 21, 61 },
		{ 210, 22, 61 },
		{ 210, 23, 61 },
		{ 210, 24, 61 },
		{ 210, 25, 61 },
		{ 210, 26, 61 },
		{ 210, 27, 61 },
		{ 210, 28, 61 },
		{ 210, 29, 61 },
		{ 210, 30, 61 },
		{ 210, 31, 61 },
		{ 210, 32, 61 },
		{ 210, 33, 61 },
		{ 210, 34, 61 },
		{ 210, 35, 61 },
		{ 210, 36, 61 },
		{ 210, 37, 61 },
		{ 210, 38, 61 },
		{ 210, 39, 61 },
		{ 210, 40, 61 },
		{ 210, 41, 61 },
		{ 210, 42, 61 },
		{ 210, 43, 61 },
		{ 210, 44, 61 },
		{ 210, 45, 61 },
		{ 210, 46, 61 },
		{ 210, 47, 61 },
		{ 210, 48, 61 },
		{ 210, 49, 61 },
		{ 210, 50, 61 },
		{ 210, 51, 61 },
		{ 210, 52, 61 },
		{ 210, 53, 61 },
		{ 210, 54, 61 },
		{ 210, 55, 61 },
		{ 210, 56, 61 },
		{ 210, 57, 61 },
		{ 210, 58, 61 },
		{ 210, 59, 61 },
		{ 210, 60, 61 },
		{ 210, 61, 61 },
		{ 210, 62, 61 },
		{ 210, 63, 61 },
		{ 210, 64, 61 },
		{ 210, 65, 61 },
		{ 210, 66, 61 },
		{ 210, 67, 61 },
		{ 210, 68, 61 },
		{ 211, 1, 61 },
		{ 211, 2, 61 },
		{ 211, 3, 61 },
		{ 211, 4, 61 },
		{ 211, 5, 61 },
		{ 211, 6, 61 },
		{ 211, 7, 61 },
		{ 211, 8, 61 },
		{ 211, 9, 61 },
		{ 211, 10, 61 },
		{ 211, 11, 61 },
		{ 211, 12, 61 },
		{ 211, 13, 61 },
		{ 211, 14, 61 },
		{ 211, 15, 61 },
		{ 211, 16, 61 },
		{ 211, 17, 61 },
		{ 211, 18, 61 },
		{ 211, 19, 61 },
		{ 211, 20, 61 },
		{ 211, 21, 61 },
		{ 211, 22, 61 },
		{ 211, 23, 61 },
		{ 211, 24, 61 },
		{ 211, 25, 61 },
		{ 211, 26, 61 },
		{ 211, 27, 61 },
		{ 211, 28, 61 },
		{ 211, 29, 61 },
		{ 211, 30, 61 },
		{ 211, 31, 61 },
		{ 211, 32, 61 },
		{ 211, 33, 61 },
		{ 211, 34, 61 },
		{ 211, 35, 61 },
		{ 211, 36, 61 },
		{ 211, 37, 61 },
		{ 211, 38, 61 },
		{ 211, 39, 61 },
		{ 211, 40, 61 },
		{ 211, 41, 61 },
		{ 211, 42, 61 },
		{ 211, 43, 61 },
		{ 211, 44, 61 },
		{ 211, 45, 61 },
		{ 211, 46, 61 },
		{ 211, 47, 61 },
		{ 211, 48, 61 },
		{ 211, 49, 61 },
		{ 211, 50, 61 },
		{ 211, 51, 61 },
		{ 211, 52, 61 },
		{ 211, 53, 61 },
		{ 211, 54, 61 },
		{ 211, 55, 61 },
		{ 211, 56, 61 },
		{ 211, 57, 61 },
		{ 211, 58, 61 },
		{ 211, 59, 61 },
		{ 211, 60, 61 },
		{ 211, 61, 61 },
		{ 211, 62, 61 },
		{ 211, 63, 61 },
		{ 211, 64, 61 },
		{ 211, 65, 61 },
		{ 211, 66, 61 },
		{ 211, 67, 61 },
		{ 211, 68, 61 },
		{ 211, 69, 61 },
		{ 211, 70, 61 },
		{ 211, 71, 61 },
		{ 211, 72, 61 },
		{ 212, 1, 61 },
		{ 212, 2, 61 },
		{ 212, 3, 61 },
		{ 212, 4, 61 },
		{ 212, 5, 61 },
		{ 212, 6, 61 },
		{ 212, 7, 61 },
		{ 212, 8, 61 },
		{ 212, 9, 61 },
		{ 212, 10, 61 },
		{ 212, 11, 61 },
		{ 212, 12, 61 },
		{ 212, 13, 61 },
		{ 212, 14, 61 },
		{ 212, 15, 61 },
		{ 212, 16, 61 },
		{ 212, 17, 61 },
		{ 212, 18, 61 },
		{ 212, 19, 61 },
		{ 212, 20, 61 },
		{ 212, 21, 61 },
		{ 212, 22, 61 },
		{ 212, 23, 61 },
		{ 212, 24, 61 },
		{ 212, 25, 61 },
		{ 212, 26, 61 },
		{ 212, 27, 61 },
		{ 212, 28, 61 },
		{ 212, 29, 61 },
		{ 212, 30, 61 },
		{ 212, 31, 61 },
		{ 212, 32, 61 },
		{ 212, 33, 61 },
		{ 212, 34, 61 },
		{ 212, 35, 61 },
		{ 212, 36, 61 },
		{ 212, 37, 61 },
		{ 212, 38, 61 },
		{ 212, 39, 61 },
		{ 212, 40, 61 },
		{ 212, 41, 61 },
		{ 212, 42, 61 },
		{ 212, 43, 61 },
		{ 212, 44, 61 },
		{ 212, 45, 61 },
		{ 212, 46, 61 },
		{ 212, 47, 61 },
		{ 212, 48, 61 },
		{ 212, 49, 61 },
		{ 212, 50, 61 },
		{ 212, 51, 61 },
		{ 212, 52, 61 },
		{ 212, 53, 61 },
		{ 212, 54, 61 },
		{ 212, 55, 61 },
		{ 212, 56, 61 },
		{ 212, 57, 61 },
		{ 212, 58, 61 },
		{ 212, 59, 61 },
		{ 212, 60, 61 },
		{ 212, 61, 61 },
		{ 212, 62, 61 },
		{ 212, 63, 61 },
		{ 212, 64, 61 },
		{ 212, 65, 61 },
		{ 212, 66, 61 },
		{ 212, 67, 61 },
		{ 212, 68, 61 },
		{ 212, 69, 61 },
		{ 212, 70, 61 },
		{ 212, 71, 61 },
		{ 212, 72, 61 },
		{ 212, 73, 61 },
		{ 212, 74, 61 },
		{ 212, 75, 61 },
		{ 212, 76, 61 },
		{ 213, 1, 61 },
		{ 213, 2, 61 },
		{ 213, 3, 61 },
		{ 213, 4, 61 },
		{ 213, 5, 61 },
		{ 213, 6, 61 },
		{ 213, 7, 61 },
		{ 213, 8, 61 },
		{ 213, 9, 61 },
		{ 213, 10, 61 },
		{ 213, 11, 61 },
		{ 213, 12, 61 },
		{ 213, 13, 61 },
		{ 213, 14, 61 },
		{ 213, 15, 61 },
		{ 213, 16, 61 },
		{ 213, 17, 61 },
		{ 213, 18, 61 },
		{ 213, 19, 61 },
		{ 213, 20, 61 },
		{ 213, 21, 61 },
		{ 213, 22, 61 },
		{ 213, 23, 61 },
		{ 213, 24, 61 },
		{ 213, 25, 61 },
		{ 213, 26, 61 },
		{ 213, 27, 61 },
		{ 213, 28, 61 },
		{ 213, 29, 61 },
		{ 213, 30, 61 },
		{ 213, 31, 61 },
		{ 213, 32, 61 },
		{ 213, 33, 61 },
		{ 213, 34, 61 },
		{ 213, 35, 61 },
		{ 213, 36, 61 },
		{ 213, 37, 61 },
		{ 213, 38, 61 },
		{ 213, 39, 61 },
		{ 213, 40, 61 },
		{ 213, 41, 61 },
		{ 213, 42, 61 },
		{ 213, 43, 61 },
		{ 213, 44, 61 },
		{ 213, 45, 61 },
		{ 213, 46, 61 },
		{ 213, 47, 61 },
		{ 213, 48, 61 },
		{ 213, 49, 61 },
		{ 213, 50, 61 },
		{ 213, 51, 61 },
		{ 213, 52, 61 },
		{ 213, 53, 61 },
		{ 213, 54, 61 },
		{ 213, 55, 61 },
		{ 213, 56, 61 },
		{ 213, 57, 61 },
		{ 213, 58, 61 },
		{ 213, 59, 61 },
		{ 213, 60, 61 },
		{ 213, 61, 61 },
		{ 213, 62, 61 },
		{ 213, 63, 61 },
		{ 213, 64, 61 },
		{ 213, 65, 61 },
		{ 213, 66, 61 },
		{ 213, 67, 61 },
		{ 213, 68, 61 },
		{ 213, 69, 61 },
		{ 213, 70, 61 },
		{ 213, 71, 61 },
		{ 213, 72, 61 },
		{ 213, 73, 61 },
		{ 213, 74, 61 },
		{ 213, 75, 61 },
		{ 213, 76, 61 },
		{ 213, 77, 61 },
		{ 213, 78, 61 },
		{ 213, 79, 61 },
		{ 213, 80, 61 },
		{ 214, 1, 61 },
		{ 214, 2, 61 },
		{ 214, 3, 61 },
		{ 214, 4, 61 },
		{ 214, 5, 61 },
		{ 214, 6, 61 },
		{ 214, 7, 61 },
		{ 214, 8, 61 },
		{ 214, 9, 61 },
		{ 214, 10, 61 },
		{ 214, 11, 61 },
		{ 214, 12, 61 },
		{ 214, 13, 61 },
		{ 214, 14, 61 },
		{ 214, 15, 61 },
		{ 214, 16, 61 },
		{ 214, 17, 61 },
		{ 214, 18, 61 },
		{ 214, 19, 61 },
		{ 214, 20, 61 },
		{ 214, 21, 61 },
		{ 214, 22, 61 },
		{ 214, 23, 61 },
		{ 214, 24, 61 },
		{ 214, 25, 61 },
		{ 214, 26, 61 },
		{ 214, 27, 61 },
		{ 214, 28, 61 },
		{ 214, 29, 61 },
		{ 214, 30, 61 },
		{ 214, 31, 61 },
		{ 214, 32, 61 },
		{ 214, 33, 61 },
		{ 214, 34, 61 },
		{ 214, 35, 61 },
		{ 214, 36, 61 },
		{ 214, 37, 61 },
		{ 214, 38, 61 },
		{ 214, 39, 61 },
		{ 214, 40, 61 },
		{ 214, 41, 61 },
		{ 214, 42, 61 },
		{ 214, 43, 61 },
		{ 214, 44, 61 },
		{ 214, 45, 61 },
		{ 214, 46, 61 },
		{ 214, 47, 61 },
		{ 214, 48, 61 },
		{ 214, 49, 61 },
		{ 214, 50, 61 },
		{ 214, 51, 61 },
		{ 214, 52, 61 },
		{ 214, 53, 61 },
		{ 214, 54, 61 },
		{ 214, 55, 61 },
		{ 214, 56, 61 },
		{ 214, 57, 61 },
		{ 214, 58, 61 },
		{ 214, 59, 61 },
		{ 214, 60, 61 },
		{ 214, 61, 61 },
		{ 214, 62, 61 },
		{ 214, 63, 61 },
		{ 214, 64, 61 },
		{ 214, 65, 61 },
		{ 214, 66, 61 },
		{ 214, 67, 61 },
		{ 214, 68, 61 },
		{ 214, 69, 61 },
		{ 214, 70, 61 },
		{ 214, 71, 61 },
		{ 214, 72, 61 },
		{ 214, 73, 61 },
		{ 214, 74, 61 },
		{ 214, 75, 61 },
		{ 214, 76, 61 },
		{ 214, 77, 61 },
		{ 214, 78, 61 },
		{ 214, 79, 61 },
		{ 214, 80, 61 },
		{ 214, 81, 61 },
		{ 214, 82, 61 },
		{ 214, 83, 61 },
		{ 214, 84, 61 },
		{ 215, 1, 61 },
		{ 215, 2, 61 },
		{ 215, 3, 61 },
		{ 215, 4, 61 },
		{ 215, 5, 61 },
		{ 215, 6, 61 },
		{ 215, 7, 61 },
		{ 215, 8, 61 },
		{ 215, 9, 61 },
		{ 215, 10, 61 },
		{ 215, 11, 61 },
		{ 215, 12, 61 },
		{ 215, 13, 61 },
		{ 215, 14, 61 },
		{ 215, 15, 61 },
		{ 215, 16, 61 },
		{ 215, 17, 61 },
		{ 215, 18, 61 },
		{ 215, 19, 61 },
		{ 215, 20, 61 },
		{ 215, 21, 61 },
		{ 215, 22, 61 },
		{ 215, 23, 61 },
		{ 215, 24, 61 },
		{ 215, 25, 61 },
		{ 215, 26, 61 },
		{ 215, 27, 61 },
		{ 215, 28, 61 },
		{ 215, 29, 61 },
		{ 215, 30, 61 },
		{ 215, 31, 61 },
		{ 215, 32, 61 },
		{ 215, 33, 61 },
		{ 215, 34, 61 },
		{ 215, 35, 61 },
		{ 215, 36, 61 },
		{ 215, 37, 61 },
		{ 215, 38, 61 },
		{ 215, 39, 61 },
		{ 215, 40, 61 },
		{ 215, 41, 61 },
		{ 215, 42, 61 },
		{ 215, 43, 61 },
		{ 215, 44, 61 },
		{ 215, 45, 61 },
		{ 215, 46, 61 },
		{ 215, 47, 61 },
		{ 215, 48, 61 },
		{ 215, 49, 61 },
		{ 215, 50, 61 },
		{ 215, 51, 61 },
		{ 215, 52, 61 },
		{ 215, 53, 61 },
		{ 215, 54, 61 },
		{ 215, 55, 61 },
		{ 215, 56, 61 },
		{ 215, 57, 61 },
		{ 215, 58, 61 },
		{ 215, 59, 61 },
		{ 215, 60, 61 },
		{ 215, 61, 61 },
		{ 215, 62, 61 },
		{ 215, 63, 61 },
		{ 215, 64, 61 },
		{ 215, 65, 61 },
		{ 215, 66, 61 },
		{ 215, 67, 61 },
		{ 215, 68, 61 },
		{ 215, 69, 61 },
		{ 215, 70, 61 },
		{ 215, 71, 61 },
		{ 215, 72, 61 },
		{ 215, 73, 61 },
		{ 215, 74, 61 },
		{ 215, 75, 61 },
		{ 215, 76, 61 },
		{ 215, 77, 61 },
		{ 215, 78, 61 },
		{ 215, 79, 61 },
		{ 215, 80, 61 },
		{ 215, 81, 61 },
		{ 215, 82, 61 },
		{ 215, 83, 61 },
		{ 215, 84, 61 },
		{ 215, 85, 61 },
		{ 215, 86, 61 },
		{ 215, 87, 61 },
		{ 215, 88, 61 },
		{ 216, 1, 61 },
		{ 216, 2, 61 },
		{ 216, 3, 61 },
		{ 216, 4, 61 },
		{ 216, 5, 61 },
		{ 216, 6, 61 },
		{ 216, 7, 61 },
		{ 216, 8, 61 },
		{ 216, 9, 61 },
		{ 216, 10, 61 },
		{ 216, 11, 61 },
		{ 216, 12, 61 },
		{ 216, 13, 61 },
		{ 216, 14, 61 },
		{ 216, 15, 61 },
		{ 216, 16, 61 },
		{ 216, 17, 61 },
		{ 216, 18, 61 },
		{ 216, 19, 61 },
		{ 216, 20, 61 },
		{ 216, 21, 61 },
		{ 216, 22, 61 },
		{ 216, 23, 61 },
		{ 216, 24, 61 },
		{ 216, 25, 61 },
		{ 216, 26, 61 },
		{ 216, 27, 61 },
		{ 216, 28, 61 },
		{ 216, 29, 61 },
		{ 216, 30, 61 },
		{ 216, 31, 61 },
		{ 216, 32, 61 },
		{ 216, 33, 61 },
		{ 216, 34, 61 },
		{ 216, 35, 61 },
		{ 216, 36, 61 },
		{ 216, 37, 61 },
		{ 216, 38, 61 },
		{ 216, 39, 61 },
		{ 216, 40, 61 },
		{ 216, 41, 61 },
		{ 216, 42, 61 },
		{ 216, 43, 61 },
		{ 216, 44, 61 },
		{ 216, 45, 61 },
		{ 216, 46, 61 },
		{ 216, 47, 61 },
		{ 216, 48, 61 },
		{ 216, 49, 61 },
		{ 216, 50, 61 },
		{ 216, 51, 61 },
		{ 216, 52, 61 },
		{ 216, 53, 61 },
		{ 216, 54, 61 },
		{ 216, 55, 61 },
		{ 216, 56, 61 },
		{ 216, 57, 61 },
		{ 216, 58, 61 },
		{ 216, 59, 61 },
		{ 216, 60, 61 },
		{ 216, 61, 61 },
		{ 216, 62, 61 },
		{ 216, 63, 61 },
		{ 216, 64, 61 },
		{ 216, 65, 61 },
		{ 216, 66, 61 },
		{ 216, 67, 61 },
		{ 216, 68, 61 },
		{ 216, 69, 61 },
		{ 216, 70, 61 },
		{ 216, 71, 61 },
		{ 216, 72, 61 },
		{ 216, 73, 61 },
		{ 216, 74, 61 },
		{ 216, 75, 61 },
		{ 216, 76, 61 },
		{ 216, 77, 61 },
		{ 216, 78, 61 },
		{ 216, 79, 61 },
		{ 216, 80, 61 },
		{ 216, 81, 61 },
		{ 216, 82, 61 },
		{ 216, 83, 61 },
		{ 216, 84, 61 },
		{ 216, 85, 61 },
		{ 216, 86, 61 },
		{ 216, 87, 61 },
		{ 216, 88, 61 },
		{ 216, 89, 61 },
		{ 216, 90, 61 },
		{ 216, 91, 61 },
		{ 216, 92, 61 },
		{ 217, 1, 61 },
		{ 217, 2, 61 },
		{ 217, 3, 61 },
		{ 217, 4, 61 },
		{ 217, 5, 61 },
		{ 217, 6, 61 },
		{ 217, 7, 61 },
		{ 217, 8, 61 },
		{ 217, 9, 61 },
		{ 217, 10, 61 },
		{ 217, 11, 61 },
		{ 217, 12, 61 },
		{ 217, 13, 61 },
		{ 217, 14, 61 },
		{ 217, 15, 61 },
		{ 217, 16, 61 },
		{ 217, 17, 61 },
		{ 217, 18, 61 },
		{ 217, 19, 61 },
		{ 217, 20, 61 },
		{ 217, 21, 61 },
		{ 217, 22, 61 },
		{ 217, 23, 61 },
		{ 217, 24, 61 },
		{ 217, 25, 61 },
		{ 217, 26, 61 },
		{ 217, 27, 61 },
		{ 217, 28, 61 },
		{ 217, 29, 61 },
		{ 217, 30, 61 },
		{ 217, 31, 61 },
		{ 217, 32, 61 },
		{ 217, 33, 61 },
		{ 217, 34, 61 },
		{ 217, 35, 61 },
		{ 217, 36, 61 },
		{ 217, 37, 61 },
		{ 217, 38, 61 },
		{ 217, 39, 61 },
		{ 217, 40, 61 },
		{ 217, 41, 61 },
		{ 217, 42, 61 },
		{ 217, 43, 61 },
		{ 217, 44, 61 },
		{ 217, 45, 61 },
		{ 217, 46, 61 },
		{ 217, 47, 61 },
		{ 217, 48, 61 },
		{ 217, 49, 61 },
		{ 217, 50, 61 },
		{ 217, 51, 61 },
		{ 217, 52, 61 },
		{ 217, 53, 61 },
		{ 217, 54, 61 },
		{ 217, 55, 61 },
		{ 217, 56, 61 },
		{ 217, 57, 61 },
		{ 217, 58, 61 },
		{ 217, 59, 61 },
		{ 217, 60, 61 },
		{ 217, 61, 61 },
		{ 217, 62, 61 },
		{ 217, 63, 61 },
		{ 217, 64, 61 },
		{ 217, 65, 61 },
		{ 217, 66, 61 },
		{ 217, 67, 61 },
		{ 217, 68, 61 },
		{ 217, 69, 61 },
		{ 217, 70, 61 },
		{ 217, 71, 61 },
		{ 217, 72, 61 },
		{ 217, 73, 61 },
		{ 217, 74, 61 },
		{ 217, 75, 61 },
		{ 217, 76, 61 },
		{ 217, 77, 61 },
		{ 217, 78, 61 },
		{ 217, 79, 61 },
		{ 217, 80, 61 },
		{ 217, 81, 61 },
		{ 217, 82, 61 },
		{ 217, 83, 61 },
		{ 217, 84, 61 },
		{ 217, 85, 61 },
		{ 217, 86, 61 },
		{ 217, 87, 61 },
		{ 217, 88, 61 },
		{ 217, 89, 61 },
		{ 217, 90, 61 },
		{ 217, 91, 61 },
		{ 217, 92, 61 },
		{ 217, 93, 61 },
		{ 217, 94, 61 },
		{ 217, 95, 61 },
		{ 217, 96, 61 },
		{ 218, 1, 61 },
		{ 218, 2, 61 },
		{ 218, 3, 61 },
		{ 218, 4, 61 },
		{ 218, 5, 61 },
		{ 218, 6, 61 },
		{ 218, 7, 61 },
		{ 218, 8, 61 },
		{ 218, 9, 61 },
		{ 218, 10, 61 },
		{ 218, 11, 61 },
		{ 218, 12, 61 },
		{ 218, 13, 61 },
		{ 218, 14, 61 },
		{ 218, 15, 61 },
		{ 218, 16, 61 },
		{ 218, 17, 61 },
		{ 218, 18, 61 },
		{ 218, 19, 61 },
		{ 218, 20, 61 },
		{ 218, 21, 61 },
		{ 218, 22, 61 },
		{ 218, 23, 61 },
		{ 218, 24, 61 },
		{ 218, 25, 61 },
		{ 218, 26, 61 },
		{ 218, 27, 61 },
		{ 218, 28, 61 },
		{ 218, 29, 61 },
		{ 218, 30, 61 },
		{ 218, 31, 61 },
		{ 218, 32, 61 },
		{ 218, 33, 61 },
		{ 218, 34, 61 },
		{ 218, 35, 61 },
		{ 218, 36, 61 },
		{ 218, 37, 61 },
		{ 218, 38, 61 },
		{ 218, 39, 61 },
		{ 218, 40, 61 },
		{ 218, 41, 61 },
		{ 218, 42, 61 },
		{ 218, 43, 61 },
		{ 218, 44, 61 },
		{ 218, 45, 61 },
		{ 218, 46, 61 },
		{ 218, 47, 61 },
		{ 218, 48, 61 },
		{ 218, 49, 61 },
		{ 218, 50, 61 },
		{ 218, 51, 61 },
		{ 218, 52, 61 },
		{ 218, 53, 61 },
		{ 218, 54, 61 },
		{ 218, 55, 61 },
		{ 218, 56, 61 },
		{ 218, 57, 61 },
		{ 218, 58, 61 },
		{ 218, 59, 61 },
		{ 218, 60, 61 },
		{ 218, 61, 61 },
		{ 218, 62, 61 },
		{ 218, 63, 61 },
		{ 218, 64, 61 },
		{ 218, 65, 61 },
		{ 218, 66, 61 },
		{ 218, 67, 61 },
		{ 218, 68, 61 },
		{ 218, 69, 61 },
		{ 218, 70, 61 },
		{ 218, 71, 61 },
		{ 218, 72, 61 },
		{ 218, 73, 61 },
		{ 218, 74, 61 },
		{ 218, 75, 61 },
		{ 218, 76, 61 },
		{ 218, 77, 61 },
		{ 218, 78, 61 },
		{ 218, 79, 61 },
		{ 218, 80, 61 },
		{ 218, 81, 61 },
		{ 218, 82, 61 },
		{ 218, 83, 61 },
		{ 218, 84, 61 },
		{ 218, 85, 61 },
		{ 218, 86, 61 },
		{ 218, 87, 61 },
		{ 218, 88, 61 },
		{ 218, 89, 61 },
		{ 218, 90, 61 },
		{ 218, 91, 61 },
		{ 218, 92, 61 },
		{ 218, 93, 61 },
		{ 218, 94, 61 },
		{ 218, 95, 61 },
		{ 218, 96, 61 },
		{ 218, 97, 61 },
		{ 218, 98, 61 },
		{ 218, 99, 61 },
		{ 218, 100, 61 },
		{ 219, 1, 61 },
		{ 219, 2, 61 },
		{ 219, 3, 61 },
		{ 219, 4, 61 },
		{ 219, 5, 61 },
		{ 219, 6, 61 },
		{ 219, 7, 61 },
		{ 219, 8, 61 },
		{ 219, 9, 61 },
		{ 219, 10, 61 },
		{ 219, 11, 61 },
		{ 219, 12, 61 },
		{ 219, 13, 61 },
		{ 219, 14, 61 },
		{ 219, 15, 61 },
		{ 219, 16, 61 },
		{ 219, 17, 61 },
		{ 219, 18, 61 },
		{ 219, 19, 61 },
		{ 219, 20, 61 },
		{ 219, 21, 61 },
		{ 219, 22, 61 },
		{ 219, 23, 61 },
		{ 219, 24, 61 },
		{ 219, 25, 61 },
		{ 219, 26, 61 },
		{ 219, 27, 61 },
		{ 219, 28, 61 },
		{ 219, 29, 61 },
		{ 219, 30, 61 },
		{ 219, 31, 61 },
		{ 219, 32, 61 },
		{ 219, 33, 61 },
		{ 219, 34, 61 },
		{ 219, 35, 61 },
		{ 219, 36, 61 },
		{ 219, 37, 61 },
		{ 219, 38, 61 },
		{ 219, 39, 61 },
		{ 219, 40, 61 },
		{ 219, 41, 61 },
		{ 219, 42, 61 },
		{ 219, 43, 61 },
		{ 219, 44, 61 },
		{ 219, 45, 61 },
		{ 219, 46, 61 },
		{ 219, 47, 61 },
		{ 219, 48, 61 },
		{ 219, 49, 61 },
		{ 219, 50, 61 },
		{ 219, 51, 61 },
		{ 219, 52, 61 },
		{ 219, 53, 61 },
		{ 219, 54, 61 },
		{ 219, 55, 61 },
		{ 219, 56, 61 },
		{ 219, 57, 61 },
		{ 219, 58, 61 },
		{ 219, 59, 61 },
		{ 219, 60, 61 },
		{ 219, 61, 61 },
		{ 219, 62, 61 },
		{ 219, 63, 61 },
		{ 219, 64, 61 },
		{ 219, 65, 61 },
		{ 219, 66, 61 },
		{ 219, 67, 61 },
		{ 219, 68, 61 },
		{ 219, 69, 61 },
		{ 219, 70, 61 },
		{ 219, 71, 61 },
		{ 219, 72, 61 },
		{ 219, 73, 61 },
		{ 219, 74, 61 },
		{ 219, 75, 61 },
		{ 219, 76, 61 },
		{ 219, 77, 61 },
		{ 219, 78, 61 },
		{ 219, 79, 61 },
		{ 219, 80, 61 },
		{ 219, 81, 61 },
		{ 219, 82, 61 },
		{ 219, 83, 61 },
		{ 219, 84, 61 },
		{ 219, 85, 61 },
		{ 219, 86, 61 },
		{ 219, 87, 61 },
		{ 219, 88, 61 },
		{ 219, 89, 61 },
		{ 219, 90, 61 },
		{ 219, 91, 61 },
		{ 219, 92, 61 },
		{ 219, 93, 61 },
		{ 219, 94, 61 },
		{ 219, 95, 61 },
		{ 219, 96, 61 },
		{ 219, 97, 61 },
		{ 219, 98, 61 },
		{ 219, 99, 61 },
		{ 219, 100, 61 },
		{ 219, 101, 61 },
		{ 219, 102, 61 },
		{ 219, 103, 61 },
		{ 219, 104, 61 },
		{ 220, 1, 61 },
		{ 220, 2, 61 },
		{ 220, 3, 61 },
		{ 220, 4, 61 },
		{ 220, 5, 61 },
		{ 220, 6, 61 },
		{ 220, 7, 61 },
		{ 220, 8, 61 },
		{ 220, 9, 61 },
		{ 220, 10, 61 },
		{ 220, 11, 61 },
		{ 220, 12, 61 },
		{ 220, 13, 61 },
		{ 220, 14, 61 },
		{ 220, 15, 61 },
		{ 220, 16, 61 },
		{ 220, 17, 61 },
		{ 220, 18, 61 },
		{ 220, 19, 61 },
		{ 220, 20, 61 },
		{ 220, 21, 61 },
		{ 220, 22, 61 },
		{ 220, 23, 61 },
		{ 220, 24, 61 },
		{ 220, 25, 61 },
		{ 220, 26, 61 },
		{ 220, 27, 61 },
		{ 220, 28, 61 },
		{ 220, 29, 61 },
		{ 220, 30, 61 },
		{ 220, 31, 61 },
		{ 220, 32, 61 },
		{ 220, 33, 61 },
		{ 220, 34, 61 },
		{ 220, 35, 61 },
		{ 220, 36, 61 },
		{ 220, 37, 61 },
		{ 220, 38, 61 },
		{ 220, 39, 61 },
		{ 220, 40, 61 },
		{ 220, 41, 61 },
		{ 220, 42, 61 },
		{ 220, 43, 61 },
		{ 220, 44, 61 },
		{ 220, 45, 61 },
		{ 220, 46, 61 },
		{ 220, 47, 61 },
		{ 220, 48, 61 },
		{ 220, 49, 61 },
		{ 220, 50, 61 },
		{ 220, 51, 61 },
		{ 220, 52, 61 },
		{ 220, 53, 61 },
		{ 220, 54, 61 },
		{ 220, 55, 61 },
		{ 220, 56, 61 },
		{ 220, 57, 61 },
		{ 220, 58, 61 },
		{ 220, 59, 61 },
		{ 220, 60, 61 },
		{ 220, 61, 61 },
		{ 220, 62, 61 },
		{ 220, 63, 61 },
		{ 220, 64, 61 },
		{ 220, 65, 61 },
		{ 220, 66, 61 },
		{ 220, 67, 61 },
		{ 220, 68, 61 },
		{ 220, 69, 61 },
		{ 220, 70, 61 },
		{ 220, 71, 61 },
		{ 220, 72, 61 },
		{ 220, 73, 61 },
		{ 220, 74, 61 },
		{ 220, 75, 61 },
		{ 220, 76, 61 },
		{ 220, 77, 61 },
		{ 220, 78, 61 },
		{ 220, 79, 61 },
		{ 220, 80, 61 },
		{ 220, 81, 61 },
		{ 220, 82, 61 },
		{ 220, 83, 61 },
		{ 220, 84, 61 },
		{ 220, 85, 61 },
		{ 220, 86, 61 },
		{ 220, 87, 61 },
		{ 220, 88, 61 },
		{ 220, 89, 61 },
		{ 220, 90, 61 },
		{ 220, 91, 61 },
		{ 220, 92, 61 },
		{ 220, 93, 61 },
		{ 220, 94, 61 },
		{ 220, 95, 61 },
		{ 220, 96, 61 },
		{ 220, 97, 61 },
		{ 220, 98, 61 },
		{ 220, 99, 61 },
		{ 220, 100, 61 },
		{ 220, 101, 61 },
		{ 220, 102, 61 },
		{ 220, 103, 61 },
		{ 220, 104, 61 },
		{ 220, 105, 61 },
		{ 220, 106, 61 },
		{ 220, 107, 61 },
		{ 220, 108, 61 },
		{ 221, 1, 61 },
		{ 221, 2, 61 },
		{ 221, 3, 61 },
		{ 221, 4, 61 },
		{ 221, 5, 61 },
		{ 221, 6, 61 },
		{ 221, 7, 61 },
		{ 221, 8, 61 },
		{ 221, 9, 61 },
		{ 221, 10, 61 },
		{ 221, 11, 61 },
		{ 221, 12, 61 },
		{ 221, 13, 61 },
		{ 221, 14, 61 },
		{ 221, 15, 61 },
		{ 221, 16, 61 },
		{ 221, 17, 61 },
		{ 221, 18, 61 },
		{ 221, 19, 61 },
		{ 221, 20, 61 },
		{ 221, 21, 61 },
		{ 221, 22, 61 },
		{ 221, 23, 61 },
		{ 221, 24, 61 },
		{ 221, 25, 61 },
		{ 221, 26, 61 },
		{ 221, 27, 61 },
		{ 221, 28, 61 },
		{ 221, 29, 61 },
		{ 221, 30, 61 },
		{ 221, 31, 61 },
		{ 221, 32, 61 },
		{ 221, 33, 61 },
		{ 221, 34, 61 },
		{ 221, 35, 61 },
		{ 221, 36, 61 },
		{ 221, 37, 61 },
		{ 221, 38, 61 },
		{ 221, 39, 61 },
		{ 221, 40, 61 },
		{ 221, 41, 61 },
		{ 221, 42, 61 },
		{ 221, 43, 61 },
		{ 221, 44, 61 },
		{ 221, 45, 61 },
		{ 221, 46, 61 },
		{ 221, 47, 61 },
		{ 221, 48, 61 },
		{ 221, 49, 61 },
		{ 221, 50, 61 },
		{ 221, 51, 61 },
		{ 221, 52, 61 },
		{ 221, 53, 61 },
		{ 221, 54, 61 },
		{ 221, 55, 61 },
		{ 221, 56, 61 },
		{ 221, 57, 61 },
		{ 221, 58, 61 },
		{ 221, 59, 61 },
		{ 221, 60, 61 },
		{ 221, 61, 61 },
		{ 221, 62, 61 },
		{ 221, 63, 61 },
		{ 221, 64, 61 },
		{ 221, 65, 61 },
		{ 221, 66, 61 },
		{ 221, 67, 61 },
		{ 221, 68, 61 },
		{ 221, 69, 61 },
		{ 221, 70, 61 },
		{ 221, 71, 61 },
		{ 221, 72, 61 },
		{ 221, 73, 61 },
		{ 221, 74, 61 },
		{ 221, 75, 61 },
		{ 221, 76, 61 },
		{ 221, 77, 61 },
		{ 221, 78, 61 },
		{ 221, 79, 61 },
		{ 221, 80, 61 },
		{ 221, 81, 61 },
		{ 221, 82, 61 },
		{ 221, 83, 61 },
		{ 221, 84, 61 },
		{ 221, 85, 61 },
		{ 221, 86, 61 },
		{ 221, 87, 61 },
		{ 221, 88, 61 },
		{ 221, 89, 61 },
		{ 221, 90, 61 },
		{ 221, 91, 61 },
		{ 221, 92, 61 },
		{ 221, 93, 61 },
		{ 221, 94, 61 },
		{ 221, 95, 61 },
		{ 221, 96, 61 },
		{ 221, 97, 61 },
		{ 221, 98, 61 },
		{ 221, 99, 61 },
		{ 221, 100, 61 },
		{ 221, 101, 61 },
		{ 221, 102, 61 },
		{ 221, 103, 61 },
		{ 221, 104, 61 },
		{ 221, 105, 61 },
		{ 221, 106, 61 },
		{ 221, 107, 61 },
		{ 221, 108, 61 },
		{ 221, 109, 61 },
		{ 221, 110, 61 },
		{ 221, 111, 61 },
		{ 221, 112, 61 },
		{ 222, 1, 61 },
		{ 222, 2, 61 },
		{ 222, 3, 61 },
		{ 222, 4, 61 },
		{ 222, 5, 61 },
		{ 222, 6, 61 },
		{ 222, 7, 61 },
		{ 222, 8, 61 },
		{ 222, 9, 61 },
		{ 222, 10, 61 },
		{ 222, 11, 61 },
		{ 222, 12, 61 },
		{ 222, 13, 61 },
		{ 222, 14, 61 },
		{ 222, 15, 61 },
		{ 222, 16, 61 },
		{ 222, 17, 61 },
		{ 222, 18, 61 },
		{ 222, 19, 61 },
		{ 222, 20, 61 },
		{ 222, 21, 61 },
		{ 222, 22, 61 },
		{ 222, 23, 61 },
		{ 222, 24, 61 },
		{ 222, 25, 61 },
		{ 222, 26, 61 },
		{ 222, 27, 61 },
		{ 222, 28, 61 },
		{ 222, 29, 61 },
		{ 222, 30, 61 },
		{ 222, 31, 61 },
		{ 222, 32, 61 },
		{ 222, 33, 61 },
		{ 222, 34, 61 },
		{ 222, 35, 61 },
		{ 222, 36, 61 },
		{ 222, 37, 61 },
		{ 222, 38, 61 },
		{ 222, 39, 61 },
		{ 222, 40, 61 },
		{ 222, 41, 61 },
		{ 222, 42, 61 },
		{ 222, 43, 61 },
		{ 222, 44, 61 },
		{ 222, 45, 61 },
		{ 222, 46, 61 },
		{ 222, 47, 61 },
		{ 222, 48, 61 },
		{ 222, 49, 61 },
		{ 222, 50, 61 },
		{ 222, 51, 61 },
		{ 222, 52, 61 },
		{ 222, 53, 61 },
		{ 222, 54, 61 },
		{ 222, 55, 61 },
		{ 222, 56, 61 },
		{ 222, 57, 61 },
		{ 222, 58, 61 },
		{ 222, 59, 61 },
		{ 222, 60, 61 },
		{ 222, 61, 61 },
		{ 222, 62, 61 },
		{ 222, 63, 61 },
		{ 222, 64, 61 },
		{ 222, 65, 61 },
		{ 222, 66, 61 },
		{ 222, 67, 61 },
		{ 222, 68, 61 },
		{ 222, 69, 61 },
		{ 222, 70, 61 },
		{ 222, 71, 61 },
		{ 222, 72, 61 },
		{ 222, 73, 61 },
		{ 222, 74, 61 },
		{ 222, 75, 61 },
		{ 222, 76, 61 },
		{ 222, 77, 61 },
		{ 222, 78, 61 },
		{ 222, 79, 61 },
		{ 222, 80, 61 },
		{ 222, 81, 61 },
		{ 222, 82, 61 },
		{ 222, 83, 61 },
		{ 222, 84, 61 },
		{ 222, 85, 61 },
		{ 222, 86, 61 },
		{ 222, 87, 61 },
		{ 222, 88, 61 },
		{ 222, 89, 61 },
		{ 222, 90, 61 },
		{ 222, 91, 61 },
		{ 222, 92, 61 },
		{ 222, 93, 61 },
		{ 222, 94, 61 },
		{ 222, 95, 61 },
		{ 222, 96, 61 },
		{ 222, 97, 61 },
		{ 222, 98, 61 },
		{ 222, 99, 61 },
		{ 222, 100, 61 },
		{ 222, 101, 61 },
		{ 222, 102, 61 },
		{ 222, 103, 61 },
		{ 222, 104, 61 },
		{ 222, 105, 61 },
		{ 222, 106, 61 },
		{ 222, 107, 61 },
		{ 222, 108, 61 },
		{ 222, 109, 61 },
		{ 222, 110, 61 },
		{ 222, 111, 61 },
		{ 222, 112, 61 },
		{ 222, 113, 61 },
		{ 222, 114, 61 },
		{ 222, 115, 61 },
		{ 222, 116, 61 },
		{ 223, 1, 61 },
		{ 223, 2, 61 },
		{ 223, 3, 61 },
		{ 223, 4, 61 },
		{ 223, 5, 61 },
		{ 223, 6, 61 },
		{ 223, 7, 61 },
		{ 223, 8, 61 },
		{ 223, 9, 61 },
		{ 223, 10, 61 },
		{ 223, 11, 61 },
		{ 223, 12, 61 },
		{ 223, 13, 61 },
		{ 223, 14, 61 },
		{ 223, 15, 61 },
		{ 223, 16, 61 },
		{ 223, 17, 61 },
		{ 223, 18, 61 },
		{ 223, 19, 61 },
		{ 223, 20, 61 },
		{ 223, 21, 61 },
		{ 223, 22, 61 },
		{ 223, 23, 61 },
		{ 223, 24, 61 },
		{ 223, 25, 61 },
		{ 223, 26, 61 },
		{ 223, 27, 61 },
		{ 223, 28, 61 },
		{ 223, 29, 61 },
		{ 223, 30, 61 },
		{ 223, 31, 61 },
		{ 223, 32, 61 },
		{ 223, 33, 61 },
		{ 223, 34, 61 },
		{ 223, 35, 61 },
		{ 223, 36, 61 },
		{ 223, 37, 61 },
		{ 223, 38, 61 },
		{ 223, 39, 61 },
		{ 223, 40, 61 },
		{ 223, 41, 61 },
		{ 223, 42, 61 },
		{ 223, 43, 61 },
		{ 223, 44, 61 },
		{ 223, 45, 61 },
		{ 223, 46, 61 },
		{ 223, 47, 61 },
		{ 223, 48, 61 },
		{ 223, 49, 61 },
		{ 223, 50, 61 },
		{ 223, 51, 61 },
		{ 223, 52, 61 },
		{ 223, 53, 61 },
		{ 223, 54, 61 },
		{ 223, 55, 61 },
		{ 223, 56, 61 },
		{ 223, 57, 61 },
		{ 223, 58, 61 },
		{ 223, 59, 61 },
		{ 223, 60, 61 },
		{ 223, 61, 61 },
		{ 223, 62, 61 },
		{ 223, 63, 61 },
		{ 223, 64, 61 },
		{ 223, 65, 61 },
		{ 223, 66, 61 },
		{ 223, 67, 61 },
		{ 223, 68, 61 },
		{ 223, 69, 61 },
		{ 223, 70, 61 },
		{ 223, 71, 61 },
		{ 223, 72, 61 },
		{ 223, 73, 61 },
		{ 223, 74, 61 },
		{ 223, 75, 61 },
		{ 223, 76, 61 },
		{ 223, 77, 61 },
		{ 223, 78, 61 },
		{ 223, 79, 61 },
		{ 223, 80, 61 },
		{ 223, 81, 61 },
		{ 223, 82, 61 },
		{ 223, 83, 61 },
		{ 223, 84, 61 },
		{ 223, 85, 61 },
		{ 223, 86, 61 },
		{ 223, 87, 61 },
		{ 223, 88, 61 },
		{ 223, 89, 61 },
		{ 223, 90, 61 },
		{ 223, 91, 61 },
		{ 223, 92, 61 },
		{ 223, 93, 61 },
		{ 223, 94, 61 },
		{ 223, 95, 61 },
		{ 223, 96, 61 },
		{ 223, 97, 61 },
		{ 223, 98, 61 },
		{ 223, 99, 61 },
		{ 223, 100, 61 },
		{ 223, 101, 61 },
		{ 223, 102, 61 },
		{ 223, 103, 61 },
		{ 223, 104, 61 },
		{ 223, 105, 61 },
		{ 223, 106, 61 },
		{ 223, 107, 61 },
		{ 223, 108, 61 },
		{ 223, 109, 61 },
		{ 223, 110, 61 },
		{ 223, 111, 61 },
		{ 223, 112, 61 },
		{ 223, 113, 61 },
		{ 223, 114, 61 },
		{ 223, 115, 61 },
		{ 223, 116, 61 },
		{ 223, 117, 61 },
		{ 223, 118, 61 },
		{ 223, 119, 61 },
		{ 223, 120, 61 },
		{ 224, 1, 61 },
		{ 224, 2, 61 },
		{ 224, 3, 61 },
		{ 224, 4, 61 },
		{ 224, 5, 61 },
		{ 224, 6, 61 },
		{ 224, 7, 61 },
		{ 224, 8, 61 },
		{ 224, 9, 61 },
		{ 224, 10, 61 },
		{ 224, 11, 61 },
		{ 224, 12, 61 },
		{ 224, 13, 61 },
		{ 224, 14, 61 },
		{ 224, 15, 61 },
		{ 224, 16, 61 },
		{ 224, 17, 61 },
		{ 224, 18, 61 },
		{ 224, 19, 61 },
		{ 224, 20, 61 },
		{ 224, 21, 61 },
		{ 224, 22, 61 },
		{ 224, 23, 61 },
		{ 224, 24, 61 },
		{ 224, 25, 61 },
		{ 224, 26, 61 },
		{ 224, 27, 61 },
		{ 224, 28, 61 },
		{ 224, 29, 61 },
		{ 224, 30, 61 },
		{ 224, 31, 61 },
		{ 224, 32, 61 },
		{ 224, 33, 61 },
		{ 224, 34, 61 },
		{ 224, 35, 61 },
		{ 224, 36, 61 },
		{ 224, 37, 61 },
		{ 224, 38, 61 },
		{ 224, 39, 61 },
		{ 224, 40, 61 },
		{ 224, 41, 61 },
		{ 224, 42, 61 },
		{ 224, 43, 61 },
		{ 224, 44, 61 },
		{ 224, 45, 61 },
		{ 224, 46, 61 },
		{ 224, 47, 61 },
		{ 224, 48, 61 },
		{ 224, 49, 61 },
		{ 224, 50, 61 },
		{ 224, 51, 61 },
		{ 224, 52, 61 },
		{ 224, 53, 61 },
		{ 224, 54, 61 },
		{ 224, 55, 61 },
		{ 224, 56, 61 },
		{ 224, 57, 61 },
		{ 224, 58, 61 },
		{ 224, 59, 61 },
		{ 224, 60, 61 },
		{ 224, 61, 61 },
		{ 224, 62, 61 },
		{ 224, 63, 61 },
		{ 224, 64, 61 },
		{ 224, 65, 61 },
		{ 224, 66, 61 },
		{ 224, 67, 61 },
		{ 224, 68, 61 },
		{ 224, 69, 61 },
		{ 224, 70, 61 },
		{ 224, 71, 61 },
		{ 224, 72, 61 },
		{ 224, 73, 61 },
		{ 224, 74, 61 },
		{ 224, 75, 61 },
		{ 224, 76, 61 },
		{ 224, 77, 61 },
		{ 224, 78, 61 },
		{ 224, 79, 61 },
		{ 224, 80, 61 },
		{ 224, 81, 61 },
		{ 224, 82, 61 },
		{ 224, 83, 61 },
		{ 224, 84, 61 },
		{ 224, 85, 61 },
		{ 224, 86, 61 },
		{ 224, 87, 61 },
		{ 224, 88, 61 },
		{ 224, 89, 61 },
		{ 224, 90, 61 },
		{ 224, 91, 61 },
		{ 224, 92, 61 },
		{ 224, 93, 61 },
		{ 224, 94, 61 },
		{ 224, 95, 61 },
		{ 224, 96, 61 },
		{ 224, 97, 61 },
		{ 224, 98, 61 },
		{ 224, 99, 61 },
		{ 224, 100, 61 },
		{ 224, 101, 61 },
		{ 224, 102, 61 },
		{ 224, 103, 61 },
		{ 224, 104, 61 },
		{ 224, 105, 61 },
		{ 224, 106, 61 },
		{ 224, 107, 61 },
		{ 224, 108, 61 },
		{ 224, 109, 61 },
		{ 224, 110, 61 },
		{ 224, 111, 61 },
		{ 224, 112, 61 },
		{ 224, 113, 61 },
		{ 224, 114, 61 },
		{ 224, 115, 61 },
		{ 224, 116, 61 },
		{ 224, 117, 61 },
		{ 224, 118, 61 },
		{ 224, 119, 61 },
		{ 224, 120, 61 },
		{ 224, 121, 61 },
		{ 224, 122, 61 },
		{ 224, 123, 61 },
		{ 224, 124, 61 },
		{ 225, 1, 61 },
		{ 225, 2, 61 },
		{ 225, 3, 61 },
		{ 225, 4, 61 },
		{ 225, 5, 61 },
		{ 225, 6, 61 },
		{ 225, 7, 61 },
		{ 225, 8, 61 },
		{ 225, 9, 61 },
		{ 225, 10, 61 },
		{ 225, 11, 61 },
		{ 225, 12, 61 },
		{ 225, 13, 61 },
		{ 225, 14, 61 },
		{ 225, 15, 61 },
		{ 225, 16, 61 },
		{ 225, 17, 61 },
		{ 225, 18, 61 },
		{ 225, 19, 61 },
		{ 225, 20, 61 },
		{ 225, 21, 61 },
		{ 225, 22, 61 },
		{ 225, 23, 61 },
		{ 225, 24, 61 },
		{ 225, 25, 61 },
		{ 225, 26, 61 },
		{ 225, 27, 61 },
		{ 225, 28, 61 },
		{ 225, 29, 61 },
		{ 225, 30, 61 },
		{ 225, 31, 61 },
		{ 225, 32, 61 },
		{ 225, 33, 61 },
		{ 225, 34, 61 },
		{ 225, 35, 61 },
		{ 225, 36, 61 },
		{ 225, 37, 61 },
		{ 225, 38, 61 },
		{ 225, 39, 61 },
		{ 225, 40, 61 },
		{ 225, 41, 61 },
		{ 225, 42, 61 },
		{ 225, 43, 61 },
		{ 225, 44, 61 },
		{ 225, 45, 61 },
		{ 225, 46, 61 },
		{ 225, 47, 61 },
		{ 225, 48, 61 },
		{ 225, 49, 61 },
		{ 225, 50, 61 },
		{ 225, 51, 61 },
		{ 225, 52, 61 },
		{ 225, 53, 61 },
		{ 225, 54, 61 },
		{ 225, 55, 61 },
		{ 225, 56, 61 },
		{ 225, 57, 61 },
		{ 225, 58, 61 },
		{ 225, 59, 61 },
		{ 225, 60, 61 },
		{ 225, 61, 61 },
		{ 225, 62, 61 },
		{ 225, 63, 61 },
		{ 225, 64, 61 },
		{ 225, 65, 61 },
		{ 225, 66, 61 },
		{ 225, 67, 61 },
		{ 225, 68, 61 },
		{ 225, 69, 61 },
		{ 225, 70, 61 },
		{ 225, 71, 61 },
		{ 225, 72, 61 },
		{ 225, 73, 61 },
		{ 225, 74, 61 },
		{ 225, 75, 61 },
		{ 225, 76, 61 },
		{ 225, 77, 61 },
		{ 225, 78, 61 },
		{ 225, 79, 61 },
		{ 225, 80, 61 },
		{ 225, 81, 61 },
		{ 225, 82, 61 },
		{ 225, 83, 61 },
		{ 225, 84, 61 },
		{ 225, 85, 61 },
		{ 225, 86, 61 },
		{ 225, 87, 61 },
		{ 225, 88, 61 },
		{ 225, 89, 61 },
		{ 225, 90, 61 },
		{ 225, 91, 61 },
		{ 225, 92, 61 },
		{ 225, 93, 61 },
		{ 225, 94, 61 },
		{ 225, 95, 61 },
		{ 225, 96, 61 },
		{ 225, 97, 61 },
		{ 225, 98, 61 },
		{ 225, 99, 61 },
		{ 225, 100, 61 },
		{ 225, 101, 61 },
		{ 225, 102, 61 },
		{ 225, 103, 61 },
		{ 225, 104, 61 },
		{ 225, 105, 61 },
		{ 225, 106, 61 },
		{ 225, 107, 61 },
		{ 225, 108, 61 },
		{ 225, 109, 61 },
		{ 225, 110, 61 },
		{ 225, 111, 61 },
		{ 225, 112, 61 },
		{ 225, 113, 61 },
		{ 225, 114, 61 },
		{ 225, 115, 61 },
		{ 225, 116, 61 },
		{ 225, 117, 61 },
		{ 225, 118, 61 },
		{ 225, 119, 61 },
		{ 225, 120, 61 },
		{ 225, 121, 61 },
		{ 225, 122, 61 },
		{ 225, 123, 61 },
		{ 225, 124, 61 },
		{ 225, 125, 61 },
		{ 225, 126, 61 },
		{ 225, 127, 61 },
		{ 225, 128, 61 },
		{ 226, 1, 61 },
		{ 226, 2, 61 },
		{ 226, 3, 61 },
		{ 226, 4, 61 },
		{ 226, 5, 61 },
		{ 226, 6, 61 },
		{ 226, 7, 61 },
		{ 226, 8, 61 },
		{ 226, 9, 61 },
		{ 226, 10, 61 },
		{ 226, 11, 61 },
		{ 226, 12, 61 },
		{ 226, 13, 61 },
		{ 226, 14, 61 },
		{ 226, 15, 61 },
		{ 226, 16, 61 },
		{ 226, 17, 61 },
		{ 226, 18, 61 },
		{ 226, 19, 61 },
		{ 226, 20, 61 },
		{ 226, 21, 61 },
		{ 226, 22, 61 },
		{ 226, 23, 61 },
		{ 226, 24, 61 },
		{ 226, 25, 61 },
		{ 226, 26, 61 },
		{ 226, 27, 61 },
		{ 226, 28, 61 },
		{ 226, 29, 61 },
		{ 226, 30, 61 },
		{ 226, 31, 61 },
		{ 226, 32, 61 },
		{ 226, 33, 61 },
		{ 226, 34, 61 },
		{ 226, 35, 61 },
		{ 226, 36, 61 },
		{ 226, 37, 61 },
		{ 226, 38, 61 },
		{ 226, 39, 61 },
		{ 226, 40, 61 },
		{ 226, 41, 61 },
		{ 226, 42, 61 },
		{ 226, 43, 61 },
		{ 226, 44, 61 },
		{ 226, 45, 61 },
		{ 226, 46, 61 },
		{ 226, 47, 61 },
		{ 226, 48, 61 },
		{ 226, 49, 61 },
		{ 226, 50, 61 },
		{ 226, 51, 61 },
		{ 226, 52, 61 },
		{ 226, 53, 61 },
		{ 226, 54, 61 },
		{ 226, 55, 61 },
		{ 226, 56, 61 },
		{ 226, 57, 61 },
		{ 226, 58, 61 },
		{ 226, 59, 61 },
		{ 226, 60, 61 },
		{ 226, 61, 61 },
		{ 226, 62, 61 },
		{ 226, 63, 61 },
		{ 226, 64, 61 },
		{ 226, 65, 61 },
		{ 226, 66, 61 },
		{ 226, 67, 61 },
		{ 226, 68, 61 },
		{ 226, 69, 61 },
		{ 226, 70, 61 },
		{ 226, 71, 61 },
		{ 226, 72, 61 },
		{ 226, 73, 61 },
		{ 226, 74, 61 },
		{ 226, 75, 61 },
		{ 226, 76, 61 },
		{ 226, 77, 61 },
		{ 226, 78, 61 },
		{ 226, 79, 61 },
		{ 226, 80, 61 },
		{ 226, 81, 61 },
		{ 226, 82, 61 },
		{ 226, 83, 61 },
		{ 226, 84, 61 },
		{ 226, 85, 61 },
		{ 226, 86, 61 },
		{ 226, 87, 61 },
		{ 226, 88, 61 },
		{ 226, 89, 61 },
		{ 226, 90, 61 },
		{ 226, 91, 61 },
		{ 226, 92, 61 },
		{ 226, 93, 61 },
		{ 226, 94, 61 },
		{ 226, 95, 61 },
		{ 226, 96, 61 },
		{ 226, 97, 61 },
		{ 226, 98, 61 },
		{ 226, 99, 61 },
		{ 226, 100, 61 },
		{ 226, 101, 61 },
		{ 226, 102, 61 },
		{ 226, 103, 61 },
		{ 226, 104, 61 },
		{ 226, 105, 61 },
		{ 226, 106, 61 },
		{ 226, 107, 61 },
		{ 226, 108, 61 },
		{ 226, 109, 61 },
		{ 226, 110, 61 },
		{ 226, 111, 61 },
		{ 226, 112, 61 },
		{ 226, 113, 61 },
		{ 226, 114, 61 },
		{ 226, 115, 61 },
		{ 226, 116, 61 },
		{ 226, 117, 61 },
		{ 226, 118, 61 },
		{ 226, 119, 61 },
		{ 226, 120, 61 },
		{ 226, 121, 61 },
		{ 226, 122, 61 },
		{ 226, 123, 61 },
		{ 226, 124, 61 },
		{ 226, 125, 61 },
		{ 226, 126, 61 },
		{ 226, 127, 61 },
		{ 226, 128, 61 },
		{ 226, 129, 61 },
		{ 226, 130, 61 },
		{ 226, 131, 61 },
		{ 226, 132, 61 },
		{ 227, 1, 61 },
		{ 227, 2, 61 },
		{ 227, 3, 61 },
		{ 227, 4, 61 },
		{ 227, 5, 61 },
		{ 227, 6, 61 },
		{ 227, 7, 61 },
		{ 227, 8, 61 },
		{ 227, 9, 61 },
		{ 227, 10, 61 },
		{ 227, 11, 61 },
		{ 227, 12, 61 },
		{ 227, 13, 61 },
		{ 227, 14, 61 },
		{ 227, 15, 61 },
		{ 227, 16, 61 },
		{ 227, 17, 61 },
		{ 227, 18, 61 },
		{ 227, 19, 61 },
		{ 227, 20, 61 },
		{ 227, 21, 61 },
		{ 227, 22, 61 },
		{ 227, 23, 61 },
		{ 227, 24, 61 },
		{ 227, 25, 61 },
		{ 227, 26, 61 },
		{ 227, 27, 61 },
		{ 227, 28, 61 },
		{ 227, 29, 61 },
		{ 227, 30, 61 },
		{ 227, 31, 61 },
		{ 227, 32, 61 },
		{ 227, 33, 61 },
		{ 227, 34, 61 },
		{ 227, 35, 61 },
		{ 227, 36, 61 },
		{ 227, 37, 61 },
		{ 227, 38, 61 },
		{ 227, 39, 61 },
		{ 227, 40, 61 },
		{ 227, 41, 61 },
		{ 227, 42, 61 },
		{ 227, 43, 61 },
		{ 227, 44, 61 },
		{ 227, 45, 61 },
		{ 227, 46, 61 },
		{ 227, 47, 61 },
		{ 227, 48, 61 },
		{ 227, 49, 61 },
		{ 227, 50, 61 },
		{ 227, 51, 61 },
		{ 227, 52, 61 },
		{ 227, 53, 61 },
		{ 227, 54, 61 },
		{ 227, 55, 61 },
		{ 227, 56, 61 },
		{ 227, 57, 61 },
		{ 227, 58, 61 },
		{ 227, 59, 61 },
		{ 227, 60, 61 },
		{ 227, 61, 61 },
		{ 227, 62, 61 },
		{ 227, 63, 61 },
		{ 227, 64, 61 },
		{ 227, 65, 61 },
		{ 227, 66, 61 },
		{ 227, 67, 61 },
		{ 227, 68, 61 },
		{ 227, 69, 61 },
		{ 227, 70, 61 },
		{ 227, 71, 61 },
		{ 227, 72, 61 },
		{ 227, 73, 61 },
		{ 227, 74, 61 },
		{ 227, 75, 61 },
		{ 227, 76, 61 },
		{ 227, 77, 61 },
		{ 227, 78, 61 },
		{ 227, 79, 61 },
		{ 227, 80, 61 },
		{ 227, 81, 61 },
		{ 227, 82, 61 },
		{ 227, 83, 61 },
		{ 227, 84, 61 },
		{ 227, 85, 61 },
		{ 227, 86, 61 },
		{ 227, 87, 61 },
		{ 227, 88, 61 },
		{ 227, 89, 61 },
		{ 227, 90, 61 },
		{ 227, 91, 61 },
		{ 227, 92, 61 },
		{ 227, 93, 61 },
		{ 227, 94, 61 },
		{ 227, 95, 61 },
		{ 227, 96, 61 },
		{ 227, 97, 61 },
		{ 227, 98, 61 },
		{ 227, 99, 61 },
		{ 227, 100, 61 },
		{ 227, 101, 61 },
		{ 227, 102, 61 },
		{ 227, 103, 61 },
		{ 227, 104, 61 },
		{ 227, 105, 61 },
		{ 227, 106, 61 },
		{ 227, 107, 61 },
		{ 227, 108, 61 },
		{ 227, 109, 61 },
		{ 227, 110, 61 },
		{ 227, 111, 61 },
		{ 227, 112, 61 },
		{ 227, 113, 61 },
		{ 227, 114, 61 },
		{ 227, 115, 61 },
		{ 227, 116, 61 },
		{ 227, 117, 61 },
		{ 227, 118, 61 },
		{ 227, 119, 61 },
		{ 227, 120, 61 },
		{ 227, 121, 61 },
		{ 227, 122, 61 },
		{ 227, 123, 61 },
		{ 227, 124, 61 },
		{ 227, 125, 61 },
		{ 227, 126, 61 },
		{ 227, 127, 61 },
		{ 227, 128, 61 },
		{ 227, 129, 61 },
		{ 227, 130, 61 },
		{ 227, 131, 61 },
		{ 227, 132, 61 },
		{ 227, 133, 61 },
		{ 227, 134, 61 },
		{ 227, 135, 61 },
		{ 227, 136, 61 },
		{ 228, 1, 61 },
		{ 228, 2, 61 },
		{ 228, 3, 61 },
		{ 228, 4, 61 },
		{ 228, 5, 61 },
		{ 228, 6, 61 },
		{ 228, 7, 61 },
		{ 228, 8, 61 },
		{ 228, 9, 61 },
		{ 228, 10, 61 },
		{ 228, 11, 61 },
		{ 228, 12, 61 },
		{ 228, 13, 61 },
		{ 228, 14, 61 },
		{ 228, 15, 61 },
		{ 228, 16, 61 },
		{ 228, 17, 61 },
		{ 228, 18, 61 },
		{ 228, 19, 61 },
		{ 228, 20, 61 },
		{ 228, 21, 61 },
		{ 228, 22, 61 },
		{ 228, 23, 61 },
		{ 228, 24, 61 },
		{ 228, 25, 61 },
		{ 228, 26, 61 },
		{ 228, 27, 61 },
		{ 228, 28, 61 },
		{ 228, 29, 61 },
		{ 228, 30, 61 },
		{ 228, 31, 61 },
		{ 228, 32, 61 },
		{ 228, 33, 61 },
		{ 228, 34, 61 },
		{ 228, 35, 61 },
		{ 228, 36, 61 },
		{ 228, 37, 61 },
		{ 228, 38, 61 },
		{ 228, 39, 61 },
		{ 228, 40, 61 },
		{ 228, 41, 61 },
		{ 228, 42, 61 },
		{ 228, 43, 61 },
		{ 228, 44, 61 },
		{ 228, 45, 61 },
		{ 228, 46, 61 },
		{ 228, 47, 61 },
		{ 228, 48, 61 },
		{ 228, 49, 61 },
		{ 228, 50, 61 },
		{ 228, 51, 61 },
		{ 228, 52, 61 },
		{ 228, 53, 61 },
		{ 228, 54, 61 },
		{ 228, 55, 61 },
		{ 228, 56, 61 },
		{ 228, 57, 61 },
		{ 228, 58, 61 },
		{ 228, 59, 61 },
		{ 228, 60, 61 },
		{ 228, 61, 61 },
		{ 228, 62, 61 },
		{ 228, 63, 61 },
		{ 228, 64, 61 },
		{ 228, 65, 61 },
		{ 228, 66, 61 },
		{ 228, 67, 61 },
		{ 228, 68, 61 },
		{ 228, 69, 61 },
		{ 228, 70, 61 },
		{ 228, 71, 61 },
		{ 228, 72, 61 },
		{ 228, 73, 61 },
		{ 228, 74, 61 },
		{ 228, 75, 61 },
		{ 228, 76, 61 },
		{ 228, 77, 61 },
		{ 228, 78, 61 },
		{ 228, 79, 61 },
		{ 228, 80, 61 },
		{ 228, 81, 61 },
		{ 228, 82, 61 },
		{ 228, 83, 61 },
		{ 228, 84, 61 },
		{ 228, 85, 61 },
		{ 228, 86, 61 },
		{ 228, 87, 61 },
		{ 228, 88, 61 },
		{ 228, 89, 61 },
		{ 228, 90, 61 },
		{ 228, 91, 61 },
		{ 228, 92, 61 },
		{ 228, 93, 61 },
		{ 228, 94, 61 },
		{ 228, 95, 61 },
		{ 228, 96, 61 },
		{ 228, 97, 61 },
		{ 228, 98, 61 },
		{ 228, 99, 61 },
		{ 228, 100, 61 },
		{ 228, 101, 61 },
		{ 228, 102, 61 },
		{ 228, 103, 61 },
		{ 228, 104, 61 },
		{ 228, 105, 61 },
		{ 228, 106, 61 },
		{ 228, 107, 61 },
		{ 228, 108, 61 },
		{ 228, 109, 61 },
		{ 228, 110, 61 },
		{ 228, 111, 61 },
		{ 228, 112, 61 },
		{ 228, 113, 61 },
		{ 228, 114, 61 },
		{ 228, 115, 61 },
		{ 228, 116, 61 },
		{ 228, 117, 61 },
		{ 228, 118, 61 },
		{ 228, 119, 61 },
		{ 228, 120, 61 },
		{ 228, 121, 61 },
		{ 228, 122, 61 },
		{ 228, 123, 61 },
		{ 228, 124, 61 },
		{ 228, 125, 61 },
		{ 228, 126, 61 },
		{ 228, 127, 61 },
		{ 228, 128, 61 },
		{ 228, 129, 61 },
		{ 228, 130, 61 },
		{ 228, 131, 61 },
		{ 228, 132, 61 },
		{ 228, 133, 61 },
		{ 228, 134, 61 },
		{ 228, 135, 61 },
		{ 228, 136, 61 },
		{ 228, 137, 61 },
		{ 228, 138, 61 },
		{ 228, 139, 61 },
		{ 228, 140, 61 },
		{ 229, 1, 61 },
		{ 229, 2, 61 },
		{ 229, 3, 61 },
		{ 229, 4, 61 },
		{ 229, 5, 61 },
		{ 229, 6, 61 },
		{ 229, 7, 61 },
		{ 229, 8, 61 },
		{ 229, 9, 61 },
		{ 229, 10, 61 },
		{ 229, 11, 61 },
		{ 229, 12, 61 },
		{ 229, 13, 61 },
		{ 229, 14, 61 },
		{ 229, 15, 61 },
		{ 229, 16, 61 },
		{ 229, 17, 61 },
		{ 229, 18, 61 },
		{ 229, 19, 61 },
		{ 229, 20, 61 },
		{ 229, 21, 61 },
		{ 229, 22, 61 },
		{ 229, 23, 61 },
		{ 229, 24, 61 },
		{ 229, 25, 61 },
		{ 229, 26, 61 },
		{ 229, 27, 61 },
		{ 229, 28, 61 },
		{ 229, 29, 61 },
		{ 229, 30, 61 },
		{ 229, 31, 61 },
		{ 229, 32, 61 },
		{ 229, 33, 61 },
		{ 229, 34, 61 },
		{ 229, 35, 61 },
		{ 229, 36, 61 },
		{ 229, 37, 61 },
		{ 229, 38, 61 },
		{ 229, 39, 61 },
		{ 229, 40, 61 },
		{ 229, 41, 61 },
		{ 229, 42, 61 },
		{ 229, 43, 61 },
		{ 229, 44, 61 },
		{ 229, 45, 61 },
		{ 229, 46, 61 },
		{ 229, 47, 61 },
		{ 229, 48, 61 },
		{ 229, 49, 61 },
		{ 229, 50, 61 },
		{ 229, 51, 61 },
		{ 229, 52, 61 },
		{ 229, 53, 61 },
		{ 229, 54, 61 },
		{ 229, 55, 61 },
		{ 229, 56, 61 },
		{ 229, 57, 61 },
		{ 229, 58, 61 },
		{ 229, 59, 61 },
		{ 229, 60, 61 },
		{ 229, 61, 61 },
		{ 229, 62, 61 },
		{ 229, 63, 61 },
		{ 229, 64, 61 },
		{ 229, 65, 61 },
		{ 229, 66, 61 },
		{ 229, 67, 61 },
		{ 229, 68, 61 },
		{ 229, 69, 61 },
		{ 229, 70, 61 },
		{ 229, 71, 61 },
		{ 229, 72, 61 },
		{ 229, 73, 61 },
		{ 229, 74, 61 },
		{ 229, 75, 61 },
		{ 229, 76, 61 },
		{ 229, 77, 61 },
		{ 229, 78, 61 },
		{ 229, 79, 61 },
		{ 229, 80, 61 },
		{ 229, 81, 61 },
		{ 229, 82, 61 },
		{ 229, 83, 61 },
		{ 229, 84, 61 },
		{ 229, 85, 61 },
		{ 229, 86, 61 },
		{ 229, 87, 61 },
		{ 229, 88, 61 },
		{ 229, 89, 61 },
		{ 229, 90, 61 },
		{ 229, 91, 61 },
		{ 229, 92, 61 },
		{ 229, 93, 61 },
		{ 229, 94, 61 },
		{ 229, 95, 61 },
		{ 229, 96, 61 },
		{ 229, 97, 61 },
		{ 229, 98, 61 },
		{ 229, 99, 61 },
		{ 229, 100, 61 },
		{ 229, 101, 61 },
		{ 229, 102, 61 },
		{ 229, 103, 61 },
		{ 229, 104, 61 },
		{ 229, 105, 61 },
		{ 229, 106, 61 },
		{ 229, 107, 61 },
		{ 229, 108, 61 },
		{ 229, 109, 61 },
		{ 229, 110, 61 },
		{ 229, 111, 61 },
		{ 229, 112, 61 },
		{ 229, 113, 61 },
		{ 229, 114, 61 },
		{ 229, 115, 61 },
		{ 229, 116, 61 },
		{ 229, 117, 61 },
		{ 229, 118, 61 },
		{ 229, 119, 61 },
		{ 229, 120, 61 },
		{ 229, 121, 61 },
		{ 229, 122, 61 },
		{ 229, 123, 61 },
		{ 229, 124, 61 },
		{ 229, 125, 61 },
		{ 229, 126, 61 },
		{ 229, 127, 61 },
		{ 229, 128, 61 },
		{ 229, 129, 61 },
		{ 229, 130, 61 },
		{ 229, 131, 61 },
		{ 229, 132, 61 },
		{ 229, 133, 61 },
		{ 229, 134, 61 },
		{ 229, 135, 61 },
		{ 229, 136, 61 },
		{ 229, 137, 61 },
		{ 229, 138, 61 },
		{ 229, 139, 61 },
		{ 229, 140, 61 },
		{ 229, 141, 61 },
		{ 229, 142, 61 },
		{ 229, 143, 61 },
		{ 229, 144, 61 },
		{ 230, 1, 61 },
		{ 230, 2, 61 },
		{ 230, 3, 61 },
		{ 230, 4, 61 },
		{ 230, 5, 61 },
		{ 230, 6, 61 },
		{ 230, 7, 61 },
		{ 230, 8, 61 },
		{ 230, 9, 61 },
		{ 230, 10, 61 },
		{ 230, 11, 61 },
		{ 230, 12, 61 },
		{ 230, 13, 61 },
		{ 230, 14, 61 },
		{ 230, 15, 61 },
		{ 230, 16, 61 },
		{ 230, 17, 61 },
		{ 230, 18, 61 },
		{ 230, 19, 61 },
		{ 230, 20, 61 },
		{ 230, 21, 61 },
		{ 230, 22, 61 },
		{ 230, 23, 61 },
		{ 230, 24, 61 },
		{ 230, 25, 61 },
		{ 230, 26, 61 },
		{ 230, 27, 61 },
		{ 230, 28, 61 },
		{ 230, 29, 61 },
		{ 230, 30, 61 },
		{ 230, 31, 61 },
		{ 230, 32, 61 },
		{ 230, 33, 61 },
		{ 230, 34, 61 },
		{ 230, 35, 61 },
		{ 230, 36, 61 },
		{ 230, 37, 61 },
		{ 230, 38, 61 },
		{ 230, 39, 61 },
		{ 230, 40, 61 },
		{ 230, 41, 61 },
		{ 230, 42, 61 },
		{ 230, 43, 61 },
		{ 230, 44, 61 },
		{ 230, 45, 61 },
		{ 230, 46, 61 },
		{ 230, 47, 61 },
		{ 230, 48, 61 },
		{ 230, 49, 61 },
		{ 230, 50, 61 },
		{ 230, 51, 61 },
		{ 230, 52, 61 },
		{ 230, 53, 61 },
		{ 230, 54, 61 },
		{ 230, 55, 61 },
		{ 230, 56, 61 },
		{ 230, 57, 61 },
		{ 230, 58, 61 },
		{ 230, 59, 61 },
		{ 230, 60, 61 },
		{ 230, 61, 61 },
		{ 230, 62, 61 },
		{ 230, 63, 61 },
		{ 230, 64, 61 },
		{ 230, 65, 61 },
		{ 230, 66, 61 },
		{ 230, 67, 61 },
		{ 230, 68, 61 },
		{ 230, 69, 61 },
		{ 230, 70, 61 },
		{ 230, 71, 61 },
		{ 230, 72, 61 },
		{ 230, 73, 61 },
		{ 230, 74, 61 },
		{ 230, 75, 61 },
		{ 230, 76, 61 },
		{ 230, 77, 61 },
		{ 230, 78, 61 },
		{ 230, 79, 61 },
		{ 230, 80, 61 },
		{ 230, 81, 61 },
		{ 230, 82, 61 },
		{ 230, 83, 61 },
		{ 230, 84, 61 },
		{ 230, 85, 61 },
		{ 230, 86, 61 },
		{ 230, 87, 61 },
		{ 230, 88, 61 },
		{ 230, 89, 61 },
		{ 230, 90, 61 },
		{ 230, 91, 61 },
		{ 230, 92, 61 },
		{ 230, 93, 61 },
		{ 230, 94, 61 },
		{ 230, 95, 61 },
		{ 230, 96, 61 },
		{ 230, 97, 61 },
		{ 230, 98, 61 },
		{ 230, 99, 61 },
		{ 230, 100, 61 },
		{ 230, 101, 61 },
		{ 230, 102, 61 },
		{ 230, 103, 61 },
		{ 230, 104, 61 },
		{ 230, 105, 61 },
		{ 230, 106, 61 },
		{ 230, 107, 61 },
		{ 230, 108, 61 },
		{ 230, 109, 61 },
		{ 230, 110, 61 },
		{ 230, 111, 61 },
		{ 230, 112, 61 },
		{ 230, 113, 61 },
		{ 230, 114, 61 },
		{ 230, 115, 61 },
		{ 230, 116, 61 },
		{ 230, 117, 61 },
		{ 230, 118, 61 },
		{ 230, 119, 61 },
		{ 230, 120, 61 },
		{ 230, 121, 61 },
		{ 230, 122, 61 },
		{ 230, 123, 61 },
		{ 230, 124, 61 },
		{ 230, 125, 61 },
		{ 230, 126, 61 },
		{ 230, 127, 61 },
		{ 230, 128, 61 },
		{ 230, 129, 61 },
		{ 230, 130, 61 },
		{ 230, 131, 61 },
		{ 230, 132, 61 },
		{ 230, 133, 61 },
		{ 230, 134, 61 },
		{ 230, 135, 61 },
		{ 230, 136, 61 },
		{ 230, 137, 61 },
		{ 230, 138, 61 },
		{ 230, 139, 61 },
		{ 230, 140, 61 },
		{ 230, 141, 61 },
		{ 230, 142, 61 },
		{ 230, 143, 61 },
		{ 230, 144, 61 },
		{ 230, 145, 61 },
		{ 230, 146, 61 },
		{ 230, 147, 61 },
		{ 230, 148, 61 },
		{ 231, 1, 61 },
		{ 231, 2, 61 },
		{ 231, 3, 61 },
		{ 231, 4, 61 },
		{ 231, 5, 61 },
		{ 231, 6, 61 },
		{ 231, 7, 61 },
		{ 231, 8, 61 },
		{ 231, 9, 61 },
		{ 231, 10, 61 },
		{ 231, 11, 61 },
		{ 231, 12, 61 },
		{ 231, 13, 61 },
		{ 231, 14, 61 },
		{ 231, 15, 61 },
		{ 231, 16, 61 },
		{ 231, 17, 61 },
		{ 231, 18, 61 },
		{ 231, 19, 61 },
		{ 231, 20, 61 },
		{ 231, 21, 61 },
		{ 231, 22, 61 },
		{ 231, 23, 61 },
		{ 231, 24, 61 },
		{ 231, 25, 61 },
		{ 231, 26, 61 },
		{ 231, 27, 61 },
		{ 231, 28, 61 },
		{ 231, 29, 61 },
		{ 231, 30, 61 },
		{ 231, 31, 61 },
		{ 231, 32, 61 },
		{ 231, 33, 61 },
		{ 231, 34, 61 },
		{ 231, 35, 61 },
		{ 231, 36, 61 },
		{ 231, 37, 61 },
		{ 231, 38, 61 },
		{ 231, 39, 61 },
		{ 231, 40, 61 },
		{ 231, 41, 61 },
		{ 231, 42, 61 },
		{ 231, 43, 61 },
		{ 231, 44, 61 },
		{ 231, 45, 61 },
		{ 231, 46, 61 },
		{ 231, 47, 61 },
		{ 231, 48, 61 },
		{ 231, 49, 61 },
		{ 231, 50, 61 },
		{ 231, 51, 61 },
		{ 231, 52, 61 },
		{ 231, 53, 61 },
		{ 231, 54, 61 },
		{ 231, 55, 61 },
		{ 231, 56, 61 },
		{ 231, 57, 61 },
		{ 231, 58, 61 },
		{ 231, 59, 61 },
		{ 231, 60, 61 },
		{ 231, 61, 61 },
		{ 231, 62, 61 },
		{ 231, 63, 61 },
		{ 231, 64, 61 },
		{ 231, 65, 61 },
		{ 231, 66, 61 },
		{ 231, 67, 61 },
		{ 231, 68, 61 },
		{ 231, 69, 61 },
		{ 231, 70, 61 },
		{ 231, 71, 61 },
		{ 231, 72, 61 },
		{ 231, 73, 61 },
		{ 231, 74, 61 },
		{ 231, 75, 61 },
		{ 231, 76, 61 },
		{ 231, 77, 61 },
		{ 231, 78, 61 },
		{ 231, 79, 61 },
		{ 231, 80, 61 },
		{ 231, 81, 61 },
		{ 231, 82, 61 },
		{ 231, 83, 61 },
		{ 231, 84, 61 },
		{ 231, 85, 61 },
		{ 231, 86, 61 },
		{ 231, 87, 61 },
		{ 231, 88, 61 },
		{ 231, 89, 61 },
		{ 231, 90, 61 },
		{ 231, 91, 61 },
		{ 231, 92, 61 },
		{ 231, 93, 61 },
		{ 231, 94, 61 },
		{ 231, 95, 61 },
		{ 231, 96, 61 },
		{ 231, 97, 61 },
		{ 231, 98, 61 },
		{ 231, 99, 61 },
		{ 231, 100, 61 },
		{ 231, 101, 61 },
		{ 231, 102, 61 },
		{ 231, 103, 61 },
		{ 231, 104, 61 },
		{ 231, 105, 61 },
		{ 231, 106, 61 },
		{ 231, 107, 61 },
		{ 231, 108, 61 },
		{ 231, 109, 61 },
		{ 231, 110, 61 },
		{ 231, 111, 61 },
		{ 231, 112, 61 },
		{ 231, 113, 61 },
		{ 231, 114, 61 },
		{ 231, 115, 61 },
		{ 231, 116, 61 },
		{ 231, 117, 61 },
		{ 231, 118, 61 },
		{ 231, 119, 61 },
		{ 231, 120, 61 },
		{ 231, 121, 61 },
		{ 231, 122, 61 },
		{ 231, 123, 61 },
		{ 231, 124, 61 },
		{ 231, 125, 61 },
		{ 231, 126, 61 },
		{ 231, 127, 61 },
		{ 231, 128, 61 },
		{ 231, 129, 61 },
		{ 231, 130, 61 },
		{ 231, 131, 61 },
		{ 231, 132, 61 },
		{ 231, 133, 61 },
		{ 231, 134, 61 },
		{ 231, 135, 61 },
		{ 231, 136, 61 },
		{ 231, 137, 61 },
		{ 231, 138, 61 },
		{ 231, 139, 61 },
		{ 231, 140, 61 },
		{ 231, 141, 61 },
		{ 231, 142, 61 },
		{ 231, 143, 61 },
		{ 231, 144, 61 },
		{ 231, 145, 61 },
		{ 231, 146, 61 },
		{ 231, 147, 61 },
		{ 231, 148, 61 },
		{ 231, 149, 61 },
		{ 231, 150, 61 },
		{ 231, 151, 61 },
		{ 231, 152, 61 },
		{ 232, 1, 61 },
		{ 232, 2, 61 },
		{ 232, 3, 61 },
		{ 232, 4, 61 },
		{ 232, 5, 61 },
		{ 232, 6, 61 },
		{ 232, 7, 61 },
		{ 232, 8, 61 },
		{ 232, 9, 61 },
		{ 232, 10, 61 },
		{ 232, 11, 61 },
		{ 232, 12, 61 },
		{ 232, 13, 61 },
		{ 232, 14, 61 },
		{ 232, 15, 61 },
		{ 232, 16, 61 },
		{ 232, 17, 61 },
		{ 232, 18, 61 },
		{ 232, 19, 61 },
		{ 232, 20, 61 },
		{ 232, 21, 61 },
		{ 232, 22, 61 },
		{ 232, 23, 61 },
		{ 232, 24, 61 },
		{ 232, 25, 61 },
		{ 232, 26, 61 },
		{ 232, 27, 61 },
		{ 232, 28, 61 },
		{ 232, 29, 61 },
		{ 232, 30, 61 },
		{ 232, 31, 61 },
		{ 232, 32, 61 },
		{ 232, 33, 61 },
		{ 232, 34, 61 },
		{ 232, 35, 61 },
		{ 232, 36, 61 },
		{ 232, 37, 61 },
		{ 232, 38, 61 },
		{ 232, 39, 61 },
		{ 232, 40, 61 },
		{ 232, 41, 61 },
		{ 232, 42, 61 },
		{ 232, 43, 61 },
		{ 232, 44, 61 },
		{ 232, 45, 61 },
		{ 232, 46, 61 },
		{ 232, 47, 61 },
		{ 232, 48, 61 },
		{ 232, 49, 61 },
		{ 232, 50, 61 },
		{ 232, 51, 61 },
		{ 232, 52, 61 },
		{ 232, 53, 61 },
		{ 232, 54, 61 },
		{ 232, 55, 61 },
		{ 232, 56, 61 },
		{ 232, 57, 61 },
		{ 232, 58, 61 },
		{ 232, 59, 61 },
		{ 232, 60, 61 },
		{ 232, 61, 61 },
		{ 232, 62, 61 },
		{ 232, 63, 61 },
		{ 232, 64, 61 },
		{ 232, 65, 61 },
		{ 232, 66, 61 },
		{ 232, 67, 61 },
		{ 232, 68, 61 },
		{ 232, 69, 61 },
		{ 232, 70, 61 },
		{ 232, 71, 61 },
		{ 232, 72, 61 },
		{ 232, 73, 61 },
		{ 232, 74, 61 },
		{ 232, 75, 61 },
		{ 232, 76, 61 },
		{ 232, 77, 61 },
		{ 232, 78, 61 },
		{ 232, 79, 61 },
		{ 232, 80, 61 },
		{ 232, 81, 61 },
		{ 232, 82, 61 },
		{ 232, 83, 61 },
		{ 232, 84, 61 },
		{ 232, 85, 61 },
		{ 232, 86, 61 },
		{ 232, 87, 61 },
		{ 232, 88, 61 },
		{ 232, 89, 61 },
		{ 232, 90, 61 },
		{ 232, 91, 61 },
		{ 232, 92, 61 },
		{ 232, 93, 61 },
		{ 232, 94, 61 },
		{ 232, 95, 61 },
		{ 232, 96, 61 },
		{ 232, 97, 61 },
		{ 232, 98, 61 },
		{ 232, 99, 61 },
		{ 232, 100, 61 },
		{ 232, 101, 61 },
		{ 232, 102, 61 },
		{ 232, 103, 61 },
		{ 232, 104, 61 },
		{ 232, 105, 61 },
		{ 232, 106, 61 },
		{ 232, 107, 61 },
		{ 232, 108, 61 },
		{ 232, 109, 61 },
		{ 232, 110, 61 },
		{ 232, 111, 61 },
		{ 232, 112, 61 },
		{ 232, 113, 61 },
		{ 232, 114, 61 },
		{ 232, 115, 61 },
		{ 232, 116, 61 },
		{ 232, 117, 61 },
		{ 232, 118, 61 },
		{ 232, 119, 61 },
		{ 232, 120, 61 },
		{ 232, 121, 61 },
		{ 232, 122, 61 },
		{ 232, 123, 61 },
		{ 232, 124, 61 },
		{ 232, 125, 61 },
		{ 232, 126, 61 },
		{ 232, 127, 61 },
		{ 232, 128, 61 },
		{ 232, 129, 61 },
		{ 232, 130, 61 },
		{ 232, 131, 61 },
		{ 232, 132, 61 },
		{ 232, 133, 61 },
		{ 232, 134, 61 },
		{ 232, 135, 61 },
		{ 232, 136, 61 },
		{ 232, 137, 61 },
		{ 232, 138, 61 },
		{ 232, 139, 61 },
		{ 232, 140, 61 },
		{ 232, 141, 61 },
		{ 232, 142, 61 },
		{ 232, 143, 61 },
		{ 232, 144, 61 },
		{ 232, 145, 61 },
		{ 232, 146, 61 },
		{ 232, 147, 61 },
		{ 232, 148, 61 },
		{ 232, 149, 61 },
		{ 232, 150, 61 },
		{ 232, 151, 61 },
		{ 232, 152, 61 },
		{ 232, 153, 61 },
		{ 232, 154, 61 },
		{ 232, 155, 61 },
		{ 232, 156, 61 },
		{ 233, 1, 61 },
		{ 233, 2, 61 },
		{ 233, 3, 61 },
		{ 233, 4, 61 },
		{ 233, 5, 61 },
		{ 233, 6, 61 },
		{ 233, 7, 61 },
		{ 233, 8, 61 },
		{ 233, 9, 61 },
		{ 233, 10, 61 },
		{ 233, 11, 61 },
		{ 233, 12, 61 },
		{ 233, 13, 61 },
		{ 233, 14, 61 },
		{ 233, 15, 61 },
		{ 233, 16, 61 },
		{ 233, 17, 61 },
		{ 233, 18, 61 },
		{ 233, 19, 61 },
		{ 233, 20, 61 },
		{ 233, 21, 61 },
		{ 233, 22, 61 },
		{ 233, 23, 61 },
		{ 233, 24, 61 },
		{ 233, 25, 61 },
		{ 233, 26, 61 },
		{ 233, 27, 61 },
		{ 233, 28, 61 },
		{ 233, 29, 61 },
		{ 233, 30, 61 },
		{ 233, 31, 61 },
		{ 233, 32, 61 },
		{ 233, 33, 61 },
		{ 233, 34, 61 },
		{ 233, 35, 61 },
		{ 233, 36, 61 },
		{ 233, 37, 61 },
		{ 233, 38, 61 },
		{ 233, 39, 61 },
		{ 233, 40, 61 },
		{ 233, 41, 61 },
		{ 233, 42, 61 },
		{ 233, 43, 61 },
		{ 233, 44, 61 },
		{ 233, 45, 61 },
		{ 233, 46, 61 },
		{ 233, 47, 61 },
		{ 233, 48, 61 },
		{ 233, 49, 61 },
		{ 233, 50, 61 },
		{ 233, 51, 61 },
		{ 233, 52, 61 },
		{ 233, 53, 61 },
		{ 233, 54, 61 },
		{ 233, 55, 61 },
		{ 233, 56, 61 },
		{ 233, 57, 61 },
		{ 233, 58, 61 },
		{ 233, 59, 61 },
		{ 233, 60, 61 },
		{ 233, 61, 61 },
		{ 233, 62, 61 },
		{ 233, 63, 61 },
		{ 233, 64, 61 },
		{ 233, 65, 61 },
		{ 233, 66, 61 },
		{ 233, 67, 61 },
		{ 233, 68, 61 },
		{ 233, 69, 61 },
		{ 233, 70, 61 },
		{ 233, 71, 61 },
		{ 233, 72, 61 },
		{ 233, 73, 61 },
		{ 233, 74, 61 },
		{ 233, 75, 61 },
		{ 233, 76, 61 },
		{ 233, 77, 61 },
		{ 233, 78, 61 },
		{ 233, 79, 61 },
		{ 233, 80, 61 },
		{ 233, 81, 61 },
		{ 233, 82, 61 },
		{ 233, 83, 61 },
		{ 233, 84, 61 },
		{ 233, 85, 61 },
		{ 233, 86, 61 },
		{ 233, 87, 61 },
		{ 233, 88, 61 },
		{ 233, 89, 61 },
		{ 233, 90, 61 },
		{ 233, 91, 61 },
		{ 233, 92, 61 },
		{ 233, 93, 61 },
		{ 233, 94, 61 },
		{ 233, 95, 61 },
		{ 233, 96, 61 },
		{ 233, 97, 61 },
		{ 233, 98, 61 },
		{ 233, 99, 61 },
		{ 233, 100, 61 },
		{ 233, 101, 61 },
		{ 233, 102, 61 },
		{ 233, 103, 61 },
		{ 233, 104, 61 },
		{ 233, 105, 61 },
		{ 233, 106, 61 },
		{ 233, 107, 61 },
		{ 233, 108, 61 },
		{ 233, 109, 61 },
		{ 233, 110, 61 },
		{ 233, 111, 61 },
		{ 233, 112, 61 },
		{ 233, 113, 61 },
		{ 233, 114, 61 },
		{ 233, 115, 61 },
		{ 233, 116, 61 },
		{ 233, 117, 61 },
		{ 233, 118, 61 },
		{ 233, 119, 61 },
		{ 233, 120, 61 },
		{ 233, 121, 61 },
		{ 233, 122, 61 },
		{ 233, 123, 61 },
		{ 233, 124, 61 },
		{ 233, 125, 61 },
		{ 233, 126, 61 },
		{ 233, 127, 61 },
		{ 233, 128, 61 },
		{ 233, 129, 61 },
		{ 233, 130, 61 },
		{ 233, 131, 61 },
		{ 233, 132, 61 },
		{ 233, 133, 61 },
		{ 233, 134, 61 },
		{ 233, 135, 61 },
		{ 233, 136, 61 },
		{ 233, 137, 61 },
		{ 233, 138, 61 },
		{ 233, 139, 61 },
		{ 233, 140, 61 },
		{ 233, 141, 61 },
		{ 233, 142, 61 },
		{ 233, 143, 61 },
		{ 233, 144, 61 },
		{ 233, 145, 61 },
		{ 233, 146, 61 },
		{ 233, 147, 61 },
		{ 233, 148, 61 },
		{ 233, 149, 61 },
		{ 233, 150, 61 },
		{ 233, 151, 61 },
		{ 233, 152, 61 },
		{ 233, 153, 61 },
		{ 233, 154, 61 },
		{ 233, 155, 61 },
		{ 233, 156, 61 },
		{ 233, 157, 61 },
		{ 233, 158, 61 },
		{ 233, 159, 61 },
		{ 233, 160, 61 },
		{ 234, 1, 61 },
		{ 234, 2, 61 },
		{ 234, 3, 61 },
		{ 234, 4, 61 },
		{ 234, 5, 61 },
		{ 234, 6, 61 },
		{ 234, 7, 61 },
		{ 234, 8, 61 },
		{ 234, 9, 61 },
		{ 234, 10, 61 },
		{ 234, 11, 61 },
		{ 234, 12, 61 },
		{ 234, 13, 61 },
		{ 234, 14, 61 },
		{ 234, 15, 61 },
		{ 234, 16, 61 },
		{ 234, 17, 61 },
		{ 234, 18, 61 },
		{ 234, 19, 61 },
		{ 234, 20, 61 },
		{ 234, 21, 61 },
		{ 234, 22, 61 },
		{ 234, 23, 61 },
		{ 234, 24, 61 },
		{ 234, 25, 61 },
		{ 234, 26, 61 },
		{ 234, 27, 61 },
		{ 234, 28, 61 },
		{ 234, 29, 61 },
		{ 234, 30, 61 },
		{ 234, 31, 61 },
		{ 234, 32, 61 },
		{ 234, 33, 61 },
		{ 234, 34, 61 },
		{ 234, 35, 61 },
		{ 234, 36, 61 },
		{ 234, 37, 61 },
		{ 234, 38, 61 },
		{ 234, 39, 61 },
		{ 234, 40, 61 },
		{ 234, 41, 61 },
		{ 234, 42, 61 },
		{ 234, 43, 61 },
		{ 234, 44, 61 },
		{ 234, 45, 61 },
		{ 234, 46, 61 },
		{ 234, 47, 61 },
		{ 234, 48, 61 },
		{ 234, 49, 61 },
		{ 234, 50, 61 },
		{ 234, 51, 61 },
		{ 234, 52, 61 },
		{ 234, 53, 61 },
		{ 234, 54, 61 },
		{ 234, 55, 61 },
		{ 234, 56, 61 },
		{ 234, 57, 61 },
		{ 234, 58, 61 },
		{ 234, 59, 61 },
		{ 234, 60, 61 },
		{ 234, 61, 61 },
		{ 234, 62, 61 },
		{ 234, 63, 61 },
		{ 234, 64, 61 },
		{ 234, 65, 61 },
		{ 234, 66, 61 },
		{ 234, 67, 61 },
		{ 234, 68, 61 },
		{ 234, 69, 61 },
		{ 234, 70, 61 },
		{ 234, 71, 61 },
		{ 234, 72, 61 },
		{ 234, 73, 61 },
		{ 234, 74, 61 },
		{ 234, 75, 61 },
		{ 234, 76, 61 },
		{ 234, 77, 61 },
		{ 234, 78, 61 },
		{ 234, 79, 61 },
		{ 234, 80, 61 },
		{ 234, 81, 61 },
		{ 234, 82, 61 },
		{ 234, 83, 61 },
		{ 234, 84, 61 },
		{ 234, 85, 61 },
		{ 234, 86, 61 },
		{ 234, 87, 61 },
		{ 234, 88, 61 },
		{ 234, 89, 61 },
		{ 234, 90, 61 },
		{ 234, 91, 61 },
		{ 234, 92, 61 },
		{ 234, 93, 61 },
		{ 234, 94, 61 },
		{ 234, 95, 61 },
		{ 234, 96, 61 },
		{ 234, 97, 61 },
		{ 234, 98, 61 },
		{ 234, 99, 61 },
		{ 234, 100, 61 },
		{ 234, 101, 61 },
		{ 234, 102, 61 },
		{ 234, 103, 61 },
		{ 234, 104, 61 },
		{ 234, 105, 61 },
		{ 234, 106, 61 },
		{ 234, 107, 61 },
		{ 234, 108, 61 },
		{ 234, 109, 61 },
		{ 234, 110, 61 },
		{ 234, 111, 61 },
		{ 234, 112, 61 },
		{ 234, 113, 61 },
		{ 234, 114, 61 },
		{ 234, 115, 61 },
		{ 234, 116, 61 },
		{ 234, 117, 61 },
		{ 234, 118, 61 },
		{ 234, 119, 61 },
		{ 234, 120, 61 },
		{ 234, 121, 61 },
		{ 234, 122, 61 },
		{ 234, 123, 61 },
		{ 234, 124, 61 },
		{ 234, 125, 61 },
		{ 234, 126, 61 },
		{ 234, 127, 61 },
		{ 234, 128, 61 },
		{ 234, 129, 61 },
		{ 234, 130, 61 },
		{ 234, 131, 61 },
		{ 234, 132, 61 },
		{ 234, 133, 61 },
		{ 234, 134, 61 },
		{ 234, 135, 61 },
		{ 234, 136, 61 },
		{ 234, 137, 61 },
		{ 234, 138, 61 },
		{ 234, 139, 61 },
		{ 234, 140, 61 },
		{ 234, 141, 61 },
		{ 234, 142, 61 },
		{ 234, 143, 61 },
		{ 234, 144, 61 },
		{ 234, 145, 61 },
		{ 234, 146, 61 },
		{ 234, 147, 61 },
		{ 234, 148, 61 },
		{ 234, 149, 61 },
		{ 234, 150, 61 },
		{ 234, 151, 61 },
		{ 234, 152, 61 },
		{ 234, 153, 61 },
		{ 234, 154, 61 },
		{ 234, 155, 61 },
		{ 234, 156, 61 },
		{ 234, 157, 61 },
		{ 234, 158, 61 },
		{ 234, 159, 61 },
		{ 234, 160, 61 },
		{ 234, 161, 61 },
		{ 234, 162, 61 },
		{ 234, 163, 61 },
		{ 234, 164, 61 },
		{ 235, 1, 61 },
		{ 235, 2, 61 },
		{ 235, 3, 61 },
		{ 235, 4, 61 },
		{ 235, 5, 61 },
		{ 235, 6, 61 },
		{ 235, 7, 61 },
		{ 235, 8, 61 },
		{ 235, 9, 61 },
		{ 235, 10, 61 },
		{ 235, 11, 61 },
		{ 235, 12, 61 },
		{ 235, 13, 61 },
		{ 235, 14, 61 },
		{ 235, 15, 61 },
		{ 235, 16, 61 },
		{ 235, 17, 61 },
		{ 235, 18, 61 },
		{ 235, 19, 61 },
		{ 235, 20, 61 },
		{ 235, 21, 61 },
		{ 235, 22, 61 },
		{ 235, 23, 61 },
		{ 235, 24, 61 },
		{ 235, 25, 61 },
		{ 235, 26, 61 },
		{ 235, 27, 61 },
		{ 235, 28, 61 },
		{ 235, 29, 61 },
		{ 235, 30, 61 },
		{ 235, 31, 61 },
		{ 235, 32, 61 },
		{ 235, 33, 61 },
		{ 235, 34, 61 },
		{ 235, 35, 61 },
		{ 235, 36, 61 },
		{ 235, 37, 61 },
		{ 235, 38, 61 },
		{ 235, 39, 61 },
		{ 235, 40, 61 },
		{ 235, 41, 61 },
		{ 235, 42, 61 },
		{ 235, 43, 61 },
		{ 235, 44, 61 },
		{ 235, 45, 61 },
		{ 235, 46, 61 },
		{ 235, 47, 61 },
		{ 235, 48, 61 },
		{ 235, 49, 61 },
		{ 235, 50, 61 },
		{ 235, 51, 61 },
		{ 235, 52, 61 },
		{ 235, 53, 61 },
		{ 235, 54, 61 },
		{ 235, 55, 61 },
		{ 235, 56, 61 },
		{ 235, 57, 61 },
		{ 235, 58, 61 },
		{ 235, 59, 61 },
		{ 235, 60, 61 },
		{ 235, 61, 61 },
		{ 235, 62, 61 },
		{ 235, 63, 61 },
		{ 235, 64, 61 },
		{ 235, 65, 61 },
		{ 235, 66, 61 },
		{ 235, 67, 61 },
		{ 235, 68, 61 },
		{ 235, 69, 61 },
		{ 235, 70, 61 },
		{ 235, 71, 61 },
		{ 235, 72, 61 },
		{ 235, 73, 61 },
		{ 235, 74, 61 },
		{ 235, 75, 61 },
		{ 235, 76, 61 },
		{ 235, 77, 61 },
		{ 235, 78, 61 },
		{ 235, 79, 61 },
		{ 235, 80, 61 },
		{ 235, 81, 61 },
		{ 235, 82, 61 },
		{ 235, 83, 61 },
		{ 235, 84, 61 },
		{ 235, 85, 61 },
		{ 235, 86, 61 },
		{ 235, 87, 61 },
		{ 235, 88, 61 },
		{ 235, 89, 61 },
		{ 235, 90, 61 },
		{ 235, 91, 61 },
		{ 235, 92, 61 },
		{ 235, 93, 61 },
		{ 235, 94, 61 },
		{ 235, 95, 61 },
		{ 235, 96, 61 },
		{ 235, 97, 61 },
		{ 235, 98, 61 },
		{ 235, 99, 61 },
		{ 235, 100, 61 },
		{ 235, 101, 61 },
		{ 235, 102, 61 },
		{ 235, 103, 61 },
		{ 235, 104, 61 },
		{ 235, 105, 61 },
		{ 235, 106, 61 },
		{ 235, 107, 61 },
		{ 235, 108, 61 },
		{ 235, 109, 61 },
		{ 235, 110, 61 },
		{ 235, 111, 61 },
		{ 235, 112, 61 },
		{ 235, 113, 61 },
		{ 235, 114, 61 },
		{ 235, 115, 61 },
		{ 235, 116, 61 },
		{ 235, 117, 61 },
		{ 235, 118, 61 },
		{ 235, 119, 61 },
		{ 235, 120, 61 },
		{ 235, 121, 61 },
		{ 235, 122, 61 },
		{ 235, 123, 61 },
		{ 235, 124, 61 },
		{ 235, 125, 61 },
		{ 235, 126, 61 },
		{ 235, 127, 61 },
		{ 235, 128, 61 },
		{ 235, 129, 61 },
		{ 235, 130, 61 },
		{ 235, 131, 61 },
		{ 235, 132, 61 },
		{ 235, 133, 61 },
		{ 235, 134, 61 },
		{ 235, 135, 61 },
		{ 235, 136, 61 },
		{ 235, 137, 61 },
		{ 235, 138, 61 },
		{ 235, 139, 61 },
		{ 235, 140, 61 },
		{ 235, 141, 61 },
		{ 235, 142, 61 },
		{ 235, 143, 61 },
		{ 235, 144, 61 },
		{ 235, 145, 61 },
		{ 235, 146, 61 },
		{ 235, 147, 61 },
		{ 235, 148, 61 },
		{ 235, 149, 61 },
		{ 235, 150, 61 },
		{ 235, 151, 61 },
		{ 235, 152, 61 },
		{ 235, 153, 61 },
		{ 235, 154, 61 },
		{ 235, 155, 61 },
		{ 235, 156, 61 },
		{ 235, 157, 61 },
		{ 235, 158, 61 },
		{ 235, 159, 61 },
		{ 235, 160, 61 },
		{ 235, 161, 61 },
		{ 235, 162, 61 },
		{ 235, 163, 61 },
		{ 235, 164, 61 },
		{ 235, 165, 61 },
		{ 235, 166, 61 },
		{ 235, 167, 61 },
		{ 235, 168, 61 },
		{ 236, 1, 61 },
		{ 236, 2, 61 },
		{ 236, 3, 61 },
		{ 236, 4, 61 },
		{ 236, 5, 61 },
		{ 236, 6, 61 },
		{ 236, 7, 61 },
		{ 236, 8, 61 },
		{ 236, 9, 61 },
		{ 236, 10, 61 },
		{ 236, 11, 61 },
		{ 236, 12, 61 },
		{ 236, 13, 61 },
		{ 236, 14, 61 },
		{ 236, 15, 61 },
		{ 236, 16, 61 },
		{ 236, 17, 61 },
		{ 236, 18, 61 },
		{ 236, 19, 61 },
		{ 236, 20, 61 },
		{ 236, 21, 61 },
		{ 236, 22, 61 },
		{ 236, 23, 61 },
		{ 236, 24, 61 },
		{ 236, 25, 61 },
		{ 236, 26, 61 },
		{ 236, 27, 61 },
		{ 236, 28, 61 },
		{ 236, 29, 61 },
		{ 236, 30, 61 },
		{ 236, 31, 61 },
		{ 236, 32, 61 },
		{ 236, 33, 61 },
		{ 236, 34, 61 },
		{ 236, 35, 61 },
		{ 236, 36, 61 },
		{ 236, 37, 61 },
		{ 236, 38, 61 },
		{ 236, 39, 61 },
		{ 236, 40, 61 },
		{ 236, 41, 61 },
		{ 236, 42, 61 },
		{ 236, 43, 61 },
		{ 236, 44, 61 },
		{ 236, 45, 61 },
		{ 236, 46, 61 },
		{ 236, 47, 61 },
		{ 236, 48, 61 },
		{ 236, 49, 61 },
		{ 236, 50, 61 },
		{ 236, 51, 61 },
		{ 236, 52, 61 },
		{ 236, 53, 61 },
		{ 236, 54, 61 },
		{ 236, 55, 61 },
		{ 236, 56, 61 },
		{ 236, 57, 61 },
		{ 236, 58, 61 },
		{ 236, 59, 61 },
		{ 236, 60, 61 },
		{ 236, 61, 61 },
		{ 236, 62, 61 },
		{ 236, 63, 61 },
		{ 236, 64, 61 },
		{ 236, 65, 61 },
		{ 236, 66, 61 },
		{ 236, 67, 61 },
		{ 236, 68, 61 },
		{ 236, 69, 61 },
		{ 236, 70, 61 },
		{ 236, 71, 61 },
		{ 236, 72, 61 },
		{ 236, 73, 61 },
		{ 236, 74, 61 },
		{ 236, 75, 61 },
		{ 236, 76, 61 },
		{ 236, 77, 61 },
		{ 236, 78, 61 },
		{ 236, 79, 61 },
		{ 236, 80, 61 },
		{ 236, 81, 61 },
		{ 236, 82, 61 },
		{ 236, 83, 61 },
		{ 236, 84, 61 },
		{ 236, 85, 61 },
		{ 236, 86, 61 },
		{ 236, 87, 61 },
		{ 236, 88, 61 },
		{ 236, 89, 61 },
		{ 236, 90, 61 },
		{ 236, 91, 61 },
		{ 236, 92, 61 },
		{ 236, 93, 61 },
		{ 236, 94, 61 },
		{ 236, 95, 61 },
		{ 236, 96, 61 },
		{ 236, 97, 61 },
		{ 236, 98, 61 },
		{ 236, 99, 61 },
		{ 236, 100, 61 },
		{ 236, 101, 61 },
		{ 236, 102, 61 },
		{ 236, 103, 61 },
		{ 236, 104, 61 },
		{ 236, 105, 61 },
		{ 236, 106, 61 },
		{ 236, 107, 61 },
		{ 236, 108, 61 },
		{ 236, 109, 61 },
		{ 236, 110, 61 },
		{ 236, 111, 61 },
		{ 236, 112, 61 },
		{ 236, 113, 61 },
		{ 236, 114, 61 },
		{ 236, 115, 61 },
		{ 236, 116, 61 },
		{ 236, 117, 61 },
		{ 236, 118, 61 },
		{ 236, 119, 61 },
		{ 236, 120, 61 },
		{ 236, 121, 61 },
		{ 236, 122, 61 },
		{ 236, 123, 61 },
		{ 236, 124, 61 },
		{ 236, 125, 61 },
		{ 236, 126, 61 },
		{ 236, 127, 61 },
		{ 236, 128, 61 },
		{ 236, 129, 61 },
		{ 236, 130, 61 },
		{ 236, 131, 61 },
		{ 236, 132, 61 },
		{ 236, 133, 61 },
		{ 236, 134, 61 },
		{ 236, 135, 61 },
		{ 236, 136, 61 },
		{ 236, 137, 61 },
		{ 236, 138, 61 },
		{ 236, 139, 61 },
		{ 236, 140, 61 },
		{ 236, 141, 61 },
		{ 236, 142, 61 },
		{ 236, 143, 61 },
		{ 236, 144, 61 },
		{ 236, 145, 61 },
		{ 236, 146, 61 },
		{ 236, 147, 61 },
		{ 236, 148, 61 },
		{ 236, 149, 61 },
		{ 236, 150, 61 },
		{ 236, 151, 61 },
		{ 236, 152, 61 },
		{ 236, 153, 61 },
		{ 236, 154, 61 },
		{ 236, 155, 61 },
		{ 236, 156, 61 },
		{ 236, 157, 61 },
		{ 236, 158, 61 },
		{ 236, 159, 61 },
		{ 236, 160, 61 },
		{ 236, 161, 61 },
		{ 236, 162, 61 },
		{ 236, 163, 61 },
		{ 236, 164, 61 },
		{ 236, 165, 61 },
		{ 236, 166, 61 },
		{ 236, 167, 61 },
		{ 236, 168, 61 },
		{ 236, 169, 61 },
		{ 236, 170, 61 },
		{ 236, 171, 61 },
		{ 236, 172, 61 },
		{ 237, 1, 61 },
		{ 237, 2, 61 },
		{ 237, 3, 61 },
		{ 237, 4, 61 },
		{ 237, 5, 61 },
		{ 237, 6, 61 },
		{ 237, 7, 61 },
		{ 237, 8, 61 },
		{ 237, 9, 61 },
		{ 237, 10, 61 },
		{ 237, 11, 61 },
		{ 237, 12, 61 },
		{ 237, 13, 61 },
		{ 237, 14, 61 },
		{ 237, 15, 61 },
		{ 237, 16, 61 },
		{ 237, 17, 61 },
		{ 237, 18, 61 },
		{ 237, 19, 61 },
		{ 237, 20, 61 },
		{ 237, 21, 61 },
		{ 237, 22, 61 },
		{ 237, 23, 61 },
		{ 237, 24, 61 },
		{ 237, 25, 61 },
		{ 237, 26, 61 },
		{ 237, 27, 61 },
		{ 237, 28, 61 },
		{ 237, 29, 61 },
		{ 237, 30, 61 },
		{ 237, 31, 61 },
		{ 237, 32, 61 },
		{ 237, 33, 61 },
		{ 237, 34, 61 },
		{ 237, 35, 61 },
		{ 237, 36, 61 },
		{ 237, 37, 61 },
		{ 237, 38, 61 },
		{ 237, 39, 61 },
		{ 237, 40, 61 },
		{ 237, 41, 61 },
		{ 237, 42, 61 },
		{ 237, 43, 61 },
		{ 237, 44, 61 },
		{ 237, 45, 61 },
		{ 237, 46, 61 },
		{ 237, 47, 61 },
		{ 237, 48, 61 },
		{ 237, 49, 61 },
		{ 237, 50, 61 },
		{ 237, 51, 61 },
		{ 237, 52, 61 },
		{ 237, 53, 61 },
		{ 237, 54, 61 },
		{ 237, 55, 61 },
		{ 237, 56, 61 },
		{ 237, 57, 61 },
		{ 237, 58, 61 },
		{ 237, 59, 61 },
		{ 237, 60, 61 },
		{ 237, 61, 61 },
		{ 237, 62, 61 },
		{ 237, 63, 61 },
		{ 237, 64, 61 },
		{ 237, 65, 61 },
		{ 237, 66, 61 },
		{ 237, 67, 61 },
		{ 237, 68, 61 },
		{ 237, 69, 61 },
		{ 237, 70, 61 },
		{ 237, 71, 61 },
		{ 237, 72, 61 },
		{ 237, 73, 61 },
		{ 237, 74, 61 },
		{ 237, 75, 61 },
		{ 237, 76, 61 },
		{ 237, 77, 61 },
		{ 237, 78, 61 },
		{ 237, 79, 61 },
		{ 237, 80, 61 },
		{ 237, 81, 61 },
		{ 237, 82, 61 },
		{ 237, 83, 61 },
		{ 237, 84, 61 },
		{ 237, 85, 61 },
		{ 237, 86, 61 },
		{ 237, 87, 61 },
		{ 237, 88, 61 },
		{ 237, 89, 61 },
		{ 237, 90, 61 },
		{ 237, 91, 61 },
		{ 237, 92, 61 },
		{ 237, 93, 61 },
		{ 237, 94, 61 },
		{ 237, 95, 61 },
		{ 237, 96, 61 },
		{ 237, 97, 61 },
		{ 237, 98, 61 },
		{ 237, 99, 61 },
		{ 237, 100, 61 },
		{ 237, 101, 61 },
		{ 237, 102, 61 },
		{ 237, 103, 61 },
		{ 237, 104, 61 },
		{ 237, 105, 61 },
		{ 237, 106, 61 },
		{ 237, 107, 61 },
		{ 237, 108, 61 },
		{ 237, 109, 61 },
		{ 237, 110, 61 },
		{ 237, 111, 61 },
		{ 237, 112, 61 },
		{ 237, 113, 61 },
		{ 237, 114, 61 },
		{ 237, 115, 61 },
		{ 237, 116, 61 },
		{ 237, 117, 61 },
		{ 237, 118, 61 },
		{ 237, 119, 61 },
		{ 237, 120, 61 },
		{ 237, 121, 61 },
		{ 237, 122, 61 },
		{ 237, 123, 61 },
		{ 237, 124, 61 },
		{ 237, 125, 61 },
		{ 237, 126, 61 },
		{ 237, 127, 61 },
		{ 237, 128, 61 },
		{ 237, 129, 61 },
		{ 237, 130, 61 },
		{ 237, 131, 61 },
		{ 237, 132, 61 },
		{ 237, 133, 61 },
		{ 237, 134, 61 },
		{ 237, 135, 61 },
		{ 237, 136, 61 },
		{ 237, 137, 61 },
		{ 237, 138, 61 },
		{ 237, 139, 61 },
		{ 237, 140, 61 },
		{ 237, 141, 61 },
		{ 237, 142, 61 },
		{ 237, 143, 61 },
		{ 237, 144, 61 },
		{ 237, 145, 61 },
		{ 237, 146, 61 },
		{ 237, 147, 61 },
		{ 237, 148, 61 },
		{ 237, 149, 61 },
		{ 237, 150, 61 },
		{ 237, 151, 61 },
		{ 237, 152, 61 },
		{ 237, 153, 61 },
		{ 237, 154, 61 },
		{ 237, 155, 61 },
		{ 237, 156, 61 },
		{ 237, 157, 61 },
		{ 237, 158, 61 },
		{ 237, 159, 61 },
		{ 237, 160, 61 },
		{ 237, 161, 61 },
		{ 237, 162, 61 },
		{ 237, 163, 61 },
		{ 237, 164, 61 },
		{ 237, 165, 61 },
		{ 237, 166, 61 },
		{ 237, 167, 61 },
		{ 237, 168, 61 },
		{ 237, 169, 61 },
		{ 237, 170, 61 },
		{ 237, 171, 61 },
		{ 237, 172, 61 },
		{ 237, 173, 61 },
		{ 237, 174, 61 },
		{ 237, 175, 61 },
		{ 237, 176, 61 },
		{ 238, 1, 61 },
		{ 238, 2, 61 },
		{ 238, 3, 61 },
		{ 238, 4, 61 },
		{ 238, 5, 61 },
		{ 238, 6, 61 },
		{ 238, 7, 61 },
		{ 238, 8, 61 },
		{ 238, 9, 61 },
		{ 238, 10, 61 },
		{ 238, 11, 61 },
		{ 238, 12, 61 },
		{ 238, 13, 61 },
		{ 238, 14, 61 },
		{ 238, 15, 61 },
		{ 238, 16, 61 },
		{ 238, 17, 61 },
		{ 238, 18, 61 },
		{ 238, 19, 61 },
		{ 238, 20, 61 },
		{ 238, 21, 61 },
		{ 238, 22, 61 },
		{ 238, 23, 61 },
		{ 238, 24, 61 },
		{ 238, 25, 61 },
		{ 238, 26, 61 },
		{ 238, 27, 61 },
		{ 238, 28, 61 },
		{ 238, 29, 61 },
		{ 238, 30, 61 },
		{ 238, 31, 61 },
		{ 238, 32, 61 },
		{ 238, 33, 61 },
		{ 238, 34, 61 },
		{ 238, 35, 61 },
		{ 238, 36, 61 },
		{ 238, 37, 61 },
		{ 238, 38, 61 },
		{ 238, 39, 61 },
		{ 238, 40, 61 },
		{ 238, 41, 61 },
		{ 238, 42, 61 },
		{ 238, 43, 61 },
		{ 238, 44, 61 },
		{ 238, 45, 61 },
		{ 238, 46, 61 },
		{ 238, 47, 61 },
		{ 238, 48, 61 },
		{ 238, 49, 61 },
		{ 238, 50, 61 },
		{ 238, 51, 61 },
		{ 238, 52, 61 },
		{ 238, 53, 61 },
		{ 238, 54, 61 },
		{ 238, 55, 61 },
		{ 238, 56, 61 },
		{ 238, 57, 61 },
		{ 238, 58, 61 },
		{ 238, 59, 61 },
		{ 238, 60, 61 },
		{ 238, 61, 61 },
		{ 238, 62, 61 },
		{ 238, 63, 61 },
		{ 238, 64, 61 },
		{ 238, 65, 61 },
		{ 238, 66, 61 },
		{ 238, 67, 61 },
		{ 238, 68, 61 },
		{ 238, 69, 61 },
		{ 238, 70, 61 },
		{ 238, 71, 61 },
		{ 238, 72, 61 },
		{ 238, 73, 61 },
		{ 238, 74, 61 },
		{ 238, 75, 61 },
		{ 238, 76, 61 },
		{ 238, 77, 61 },
		{ 238, 78, 61 },
		{ 238, 79, 61 },
		{ 238, 80, 61 },
		{ 238, 81, 61 },
		{ 238, 82, 61 },
		{ 238, 83, 61 },
		{ 238, 84, 61 },
		{ 238, 85, 61 },
		{ 238, 86, 61 },
		{ 238, 87, 61 },
		{ 238, 88, 61 },
		{ 238, 89, 61 },
		{ 238, 90, 61 },
		{ 238, 91, 61 },
		{ 238, 92, 61 },
		{ 238, 93, 61 },
		{ 238, 94, 61 },
		{ 238, 95, 61 },
		{ 238, 96, 61 },
		{ 238, 97, 61 },
		{ 238, 98, 61 },
		{ 238, 99, 61 },
		{ 238, 100, 61 },
		{ 238, 101, 61 },
		{ 238, 102, 61 },
		{ 238, 103, 61 },
		{ 238, 104, 61 },
		{ 238, 105, 61 },
		{ 238, 106, 61 },
		{ 238, 107, 61 },
		{ 238, 108, 61 },
		{ 238, 109, 61 },
		{ 238, 110, 61 },
		{ 238, 111, 61 },
		{ 238, 112, 61 },
		{ 238, 113, 61 },
		{ 238, 114, 61 },
		{ 238, 115, 61 },
		{ 238, 116, 61 },
		{ 238, 117, 61 },
		{ 238, 118, 61 },
		{ 238, 119, 61 },
		{ 238, 120, 61 },
		{ 238, 121, 61 },
		{ 238, 122, 61 },
		{ 238, 123, 61 },
		{ 238, 124, 61 },
		{ 238, 125, 61 },
		{ 238, 126, 61 },
		{ 238, 127, 61 },
		{ 238, 128, 61 },
		{ 238, 129, 61 },
		{ 238, 130, 61 },
		{ 238, 131, 61 },
		{ 238, 132, 61 },
		{ 238, 133, 61 },
		{ 238, 134, 61 },
		{ 238, 135, 61 },
		{ 238, 136, 61 },
		{ 238, 137, 61 },
		{ 238, 138, 61 },
		{ 238, 139, 61 },
		{ 238, 140, 61 },
		{ 238, 141, 61 },
		{ 238, 142, 61 },
		{ 238, 143, 61 },
		{ 238, 144, 61 },
		{ 238, 145, 61 },
		{ 238, 146, 61 },
		{ 238, 147, 61 },
		{ 238, 148, 61 },
		{ 238, 149, 61 },
		{ 238, 150, 61 },
		{ 238, 151, 61 },
		{ 238, 152, 61 },
		{ 238, 153, 61 },
		{ 238, 154, 61 },
		{ 238, 155, 61 },
		{ 238, 156, 61 },
		{ 238, 157, 61 },
		{ 238, 158, 61 },
		{ 238, 159, 61 },
		{ 238, 160, 61 },
		{ 238, 161, 61 },
		{ 238, 162, 61 },
		{ 238, 163, 61 },
		{ 238, 164, 61 },
		{ 238, 165, 61 },
		{ 238, 166, 61 },
		{ 238, 167, 61 },
		{ 238, 168, 61 },
		{ 238, 169, 61 },
		{ 238, 170, 61 },
		{ 238, 171, 61 },
		{ 238, 172, 61 },
		{ 238, 173, 61 },
		{ 238, 174, 61 },
		{ 238, 175, 61 },
		{ 238, 176, 61 },
		{ 238, 177, 61 },
		{ 238, 178, 61 },
		{ 238, 179, 61 },
		{ 238, 180, 61 },
		{ 239, 1, 61 },
		{ 239, 2, 61 },
		{ 239, 3, 61 },
		{ 239, 4, 61 },
		{ 239, 5, 61 },
		{ 239, 6, 61 },
		{ 239, 7, 61 },
		{ 239, 8, 61 },
		{ 239, 9, 61 },
		{ 239, 10, 61 },
		{ 239, 11, 61 },
		{ 239, 12, 61 },
		{ 239, 13, 61 },
		{ 239, 14, 61 },
		{ 239, 15, 61 },
		{ 239, 16, 61 },
		{ 239, 17, 61 },
		{ 239, 18, 61 },
		{ 239, 19, 61 },
		{ 239, 20, 61 },
		{ 239, 21, 61 },
		{ 239, 22, 61 },
		{ 239, 23, 61 },
		{ 239, 24, 61 },
		{ 239, 25, 61 },
		{ 239, 26, 61 },
		{ 239, 27, 61 },
		{ 239, 28, 61 },
		{ 239, 29, 61 },
		{ 239, 30, 61 },
		{ 239, 31, 61 },
		{ 239, 32, 61 },
		{ 239, 33, 61 },
		{ 239, 34, 61 },
		{ 239, 35, 61 },
		{ 239, 36, 61 },
		{ 239, 37, 61 },
		{ 239, 38, 61 },
		{ 239, 39, 61 },
		{ 239, 40, 61 },
		{ 239, 41, 61 },
		{ 239, 42, 61 },
		{ 239, 43, 61 },
		{ 239, 44, 61 },
		{ 239, 45, 61 },
		{ 239, 46, 61 },
		{ 239, 47, 61 },
		{ 239, 48, 61 },
		{ 239, 49, 61 },
		{ 239, 50, 61 },
		{ 239, 51, 61 },
		{ 239, 52, 61 },
		{ 239, 53, 61 },
		{ 239, 54, 61 },
		{ 239, 55, 61 },
		{ 239, 56, 61 },
		{ 239, 57, 61 },
		{ 239, 58, 61 },
		{ 239, 59, 61 },
		{ 239, 60, 61 },
		{ 239, 61, 61 },
		{ 239, 62, 61 },
		{ 239, 63, 61 },
		{ 239, 64, 61 },
		{ 239, 65, 61 },
		{ 239, 66, 61 },
		{ 239, 67, 61 },
		{ 239, 68, 61 },
		{ 239, 69, 61 },
		{ 239, 70, 61 },
		{ 239, 71, 61 },
		{ 239, 72, 61 },
		{ 239, 73, 61 },
		{ 239, 74, 61 },
		{ 239, 75, 61 },
		{ 239, 76, 61 },
		{ 239, 77, 61 },
		{ 239, 78, 61 },
		{ 239, 79, 61 },
		{ 239, 80, 61 },
		{ 239, 81, 61 },
		{ 239, 82, 61 },
		{ 239, 83, 61 },
		{ 239, 84, 61 },
		{ 239, 85, 61 },
		{ 239, 86, 61 },
		{ 239, 87, 61 },
		{ 239, 88, 61 },
		{ 239, 89, 61 },
		{ 239, 90, 61 },
		{ 239, 91, 61 },
		{ 239, 92, 61 },
		{ 239, 93, 61 },
		{ 239, 94, 61 },
		{ 239, 95, 61 },
		{ 239, 96, 61 },
		{ 239, 97, 61 },
		{ 239, 98, 61 },
		{ 239, 99, 61 },
		{ 239, 100, 61 },
		{ 239, 101, 61 },
		{ 239, 102, 61 },
		{ 239, 103, 61 },
		{ 239, 104, 61 },
		{ 239, 105, 61 },
		{ 239, 106, 61 },
		{ 239, 107, 61 },
		{ 239, 108, 61 },
		{ 239, 109, 61 },
		{ 239, 110, 61 },
		{ 239, 111, 61 },
		{ 239, 112, 61 },
		{ 239, 113, 61 },
		{ 239, 114, 61 },
		{ 239, 115, 61 },
		{ 239, 116, 61 },
		{ 239, 117, 61 },
		{ 239, 118, 61 },
		{ 239, 119, 61 },
		{ 239, 120, 61 },
		{ 239, 121, 61 },
		{ 239, 122, 61 },
		{ 239, 123, 61 },
		{ 239, 124, 61 },
		{ 239, 125, 61 },
		{ 239, 126, 61 },
		{ 239, 127, 61 },
		{ 239, 128, 61 },
		{ 239, 129, 61 },
		{ 239, 130, 61 },
		{ 239, 131, 61 },
		{ 239, 132, 61 },
		{ 239, 133, 61 },
		{ 239, 134, 61 },
		{ 239, 135, 61 },
		{ 239, 136, 61 },
		{ 239, 137, 61 },
		{ 239, 138, 61 },
		{ 239, 139, 61 },
		{ 239, 140, 61 },
		{ 239, 141, 61 },
		{ 239, 142, 61 },
		{ 239, 143, 61 },
		{ 239, 144, 61 },
		{ 239, 145, 61 },
		{ 239, 146, 61 },
		{ 239, 147, 61 },
		{ 239, 148, 61 },
		{ 239, 149, 61 },
		{ 239, 150, 61 },
		{ 239, 151, 61 },
		{ 239, 152, 61 },
		{ 239, 153, 61 },
		{ 239, 154, 61 },
		{ 239, 155, 61 },
		{ 239, 156, 61 },
		{ 239, 157, 61 },
		{ 239, 158, 61 },
		{ 239, 159, 61 },
		{ 239, 160, 61 },
		{ 239, 161, 61 },
		{ 239, 162, 61 },
		{ 239, 163, 61 },
		{ 239, 164, 61 },
		{ 239, 165, 61 },
		{ 239, 166, 61 },
		{ 239, 167, 61 },
		{ 239, 168, 61 },
		{ 239, 169, 61 },
		{ 239, 170, 61 },
		{ 239, 171, 61 },
		{ 239, 172, 61 },
		{ 239, 173, 61 },
		{ 239, 174, 61 },
		{ 239, 175, 61 },
		{ 239, 176, 61 },
		{ 239, 177, 61 },
		{ 239, 178, 61 },
		{ 239, 179, 61 },
		{ 239, 180, 61 },
		{ 239, 181, 61 },
		{ 239, 182, 61 },
		{ 239, 183, 61 },
		{ 239, 184, 61 },
		{ 240, 1, 61 },
		{ 240, 2, 61 },
		{ 240, 3, 61 },
		{ 240, 4, 61 },
		{ 240, 5, 61 },
		{ 240, 6, 61 },
		{ 240, 7, 61 },
		{ 240, 8, 61 },
		{ 240, 9, 61 },
		{ 240, 10, 61 },
		{ 240, 11, 61 },
		{ 240, 12, 61 },
		{ 240, 13, 61 },
		{ 240, 14, 61 },
		{ 240, 15, 61 },
		{ 240, 16, 61 },
		{ 240, 17, 61 },
		{ 240, 18, 61 },
		{ 240, 19, 61 },
		{ 240, 20, 61 },
		{ 240, 21, 61 },
		{ 240, 22, 61 },
		{ 240, 23, 61 },
		{ 240, 24, 61 },
		{ 240, 25, 61 },
		{ 240, 26, 61 },
		{ 240, 27, 61 },
		{ 240, 28, 61 },
		{ 240, 29, 61 },
		{ 240, 30, 61 },
		{ 240, 31, 61 },
		{ 240, 32, 61 },
		{ 240, 33, 61 },
		{ 240, 34, 61 },
		{ 240, 35, 61 },
		{ 240, 36, 61 },
		{ 240, 37, 61 },
		{ 240, 38, 61 },
		{ 240, 39, 61 },
		{ 240, 40, 61 },
		{ 240, 41, 61 },
		{ 240, 42, 61 },
		{ 240, 43, 61 },
		{ 240, 44, 61 },
		{ 240, 45, 61 },
		{ 240, 46, 61 },
		{ 240, 47, 61 },
		{ 240, 48, 61 },
		{ 240, 49, 61 },
		{ 240, 50, 61 },
		{ 240, 51, 61 },
		{ 240, 52, 61 },
		{ 240, 53, 61 },
		{ 240, 54, 61 },
		{ 240, 55, 61 },
		{ 240, 56, 61 },
		{ 240, 57, 61 },
		{ 240, 58, 61 },
		{ 240, 59, 61 },
		{ 240, 60, 61 },
		{ 240, 61, 61 },
		{ 240, 62, 61 },
		{ 240, 63, 61 },
		{ 240, 64, 61 },
		{ 240, 65, 61 },
		{ 240, 66, 61 },
		{ 240, 67, 61 },
		{ 240, 68, 61 },
		{ 240, 69, 61 },
		{ 240, 70, 61 },
		{ 240, 71, 61 },
		{ 240, 72, 61 },
		{ 240, 73, 61 },
		{ 240, 74, 61 },
		{ 240, 75, 61 },
		{ 240, 76, 61 },
		{ 240, 77, 61 },
		{ 240, 78, 61 },
		{ 240, 79, 61 },
		{ 240, 80, 61 },
		{ 240, 81, 61 },
		{ 240, 82, 61 },
		{ 240, 83, 61 },
		{ 240, 84, 61 },
		{ 240, 85, 61 },
		{ 240, 86, 61 },
		{ 240, 87, 61 },
		{ 240, 88, 61 },
		{ 240, 89, 61 },
		{ 240, 90, 61 },
		{ 240, 91, 61 },
		{ 240, 92, 61 },
		{ 240, 93, 61 },
		{ 240, 94, 61 },
		{ 240, 95, 61 },
		{ 240, 96, 61 },
		{ 240, 97, 61 },
		{ 240, 98, 61 },
		{ 240, 99, 61 },
		{ 240, 100, 61 },
		{ 240, 101, 61 },
		{ 240, 102, 61 },
		{ 240, 103, 61 },
		{ 240, 104, 61 },
		{ 240, 105, 61 },
		{ 240, 106, 61 },
		{ 240, 107, 61 },
		{ 240, 108, 61 },
		{ 240, 109, 61 },
		{ 240, 110, 61 },
		{ 240, 111, 61 },
		{ 240, 112, 61 },
		{ 240, 113, 61 },
		{ 240, 114, 61 },
		{ 240, 115, 61 },
		{ 240, 116, 61 },
		{ 240, 117, 61 },
		{ 240, 118, 61 },
		{ 240, 119, 61 },
		{ 240, 120, 61 },
		{ 240, 121, 61 },
		{ 240, 122, 61 },
		{ 240, 123, 61 },
		{ 240, 124, 61 },
		{ 240, 125, 61 },
		{ 240, 126, 61 },
		{ 240, 127, 61 },
		{ 240, 128, 61 },
		{ 240, 129, 61 },
		{ 240, 130, 61 },
		{ 240, 131, 61 },
		{ 240, 132, 61 },
		{ 240, 133, 61 },
		{ 240, 134, 61 },
		{ 240, 135, 61 },
		{ 240, 136, 61 },
		{ 240, 137, 61 },
		{ 240, 138, 61 },
		{ 240, 139, 61 },
		{ 240, 140, 61 },
		{ 240, 141, 61 },
		{ 240, 142, 61 },
		{ 240, 143, 61 },
		{ 240, 144, 61 },
		{ 240, 145, 61 },
		{ 240, 146, 61 },
		{ 240, 147, 61 },
		{ 240, 148, 61 },
		{ 240, 149, 61 },
		{ 240, 150, 61 },
		{ 240, 151, 61 },
		{ 240, 152, 61 },
		{ 240, 153, 61 },
		{ 240, 154, 61 },
		{ 240, 155, 61 },
		{ 240, 156, 61 },
		{ 240, 157, 61 },
		{ 240, 158, 61 },
		{ 240, 159, 61 },
		{ 240, 160, 61 },
		{ 240, 161, 61 },
		{ 240, 162, 61 },
		{ 240, 163, 61 },
		{ 240, 164, 61 },
		{ 240, 165, 61 },
		{ 240, 166, 61 },
		{ 240, 167, 61 },
		{ 240, 168, 61 },
		{ 240, 169, 61 },
		{ 240, 170, 61 },
		{ 240, 171, 61 },
		{ 240, 172, 61 },
		{ 240, 173, 61 },
		{ 240, 174, 61 },
		{ 240, 175, 61 },
		{ 240, 176, 61 },
		{ 240, 177, 61 },
		{ 240, 178, 61 },
		{ 240, 179, 61 },
		{ 240, 180, 61 },
		{ 240, 181, 61 },
		{ 240, 182, 61 },
		{ 240, 183, 61 },
		{ 240, 184, 61 },
		{ 240, 185, 61 },
		{ 240, 186, 61 },
		{ 240, 187, 61 },
		{ 241, 1, 61 },
		{ 241, 2, 61 },
		{ 241, 3, 61 },
		{ 241, 4, 61 },
		{ 241, 5, 61 },
		{ 241, 6, 61 },
		{ 241, 7, 61 },
		{ 241, 8, 61 },
		{ 241, 9, 61 },
		{ 241, 10, 61 },
		{ 241, 11, 61 },
		{ 241, 12, 61 },
		{ 241, 13, 61 },
		{ 241, 14, 61 },
		{ 241, 15, 61 },
		{ 241, 16, 61 },
		{ 241, 17, 61 },
		{ 241, 18, 61 },
		{ 241, 19, 61 },
		{ 241, 20, 61 },
		{ 241, 21, 61 },
		{ 241, 22, 61 },
		{ 241, 23, 61 },
		{ 241, 24, 61 },
		{ 241, 25, 61 },
		{ 241, 26, 61 },
		{ 241, 27, 61 },
		{ 241, 28, 61 },
		{ 241, 29, 61 },
		{ 241, 30, 61 },
		{ 241, 31, 61 },
		{ 241, 32, 61 },
		{ 241, 33, 61 },
		{ 241, 34, 61 },
		{ 241, 35, 61 },
		{ 241, 36, 61 },
		{ 241, 37, 61 },
		{ 241, 38, 61 },
		{ 241, 39, 61 },
		{ 241, 40, 61 },
		{ 241, 41, 61 },
		{ 241, 42, 61 },
		{ 241, 43, 61 },
		{ 241, 44, 61 },
		{ 241, 45, 61 },
		{ 241, 46, 61 },
		{ 241, 47, 61 },
		{ 241, 48, 61 },
		{ 241, 49, 61 },
		{ 241, 50, 61 },
		{ 241, 51, 61 },
		{ 241, 52, 61 },
		{ 241, 53, 61 },
		{ 241, 54, 61 },
		{ 241, 55, 61 },
		{ 241, 56, 61 },
		{ 241, 57, 61 },
		{ 241, 58, 61 },
		{ 241, 59, 61 },
		{ 241, 60, 61 },
		{ 241, 61, 61 },
		{ 241, 62, 61 },
		{ 241, 63, 61 },
		{ 241, 64, 61 },
		{ 241, 65, 61 },
		{ 241, 66, 61 },
		{ 241, 67, 61 },
		{ 241, 68, 61 },
		{ 241, 69, 61 },
		{ 241, 70, 61 },
		{ 241, 71, 61 },
		{ 241, 72, 61 },
		{ 241, 73, 61 },
		{ 241, 74, 61 },
		{ 241, 75, 61 },
		{ 241, 76, 61 },
		{ 241, 77, 61 },
		{ 241, 78, 61 },
		{ 241, 79, 61 },
		{ 241, 80, 61 },
		{ 241, 81, 61 },
		{ 241, 82, 61 },
		{ 241, 83, 61 },
		{ 241, 84, 61 },
		{ 241, 85, 61 },
		{ 241, 86, 61 },
		{ 241, 87, 61 },
		{ 241, 88, 61 },
		{ 241, 89, 61 },
		{ 241, 90, 61 },
		{ 241, 91, 61 },
		{ 241, 92, 61 },
		{ 241, 93, 61 },
		{ 241, 94, 61 },
		{ 241, 95, 61 },
		{ 241, 96, 61 },
		{ 241, 97, 61 },
		{ 241, 98, 61 },
		{ 241, 99, 61 },
		{ 241, 100, 61 },
		{ 241, 101, 61 },
		{ 241, 102, 61 },
		{ 241, 103, 61 },
		{ 241, 104, 61 },
		{ 241, 105, 61 },
		{ 241, 106, 61 },
		{ 241, 107, 61 },
		{ 241, 108, 61 },
		{ 241, 109, 61 },
		{ 241, 110, 61 },
		{ 241, 111, 61 },
		{ 241, 112, 61 },
		{ 241, 113, 61 },
		{ 241, 114, 61 },
		{ 241, 115, 61 },
		{ 241, 116, 61 },
		{ 241, 117, 61 },
		{ 241, 118, 61 },
		{ 241, 119, 61 },
		{ 241, 120, 61 },
		{ 241, 121, 61 },
		{ 241, 122, 61 },
		{ 241, 123, 61 },
		{ 241, 124, 61 },
		{ 241, 125, 61 },
		{ 241, 126, 61 },
		{ 241, 127, 61 },
		{ 241, 128, 61 },
		{ 241, 129, 61 },
		{ 241, 130, 61 },
		{ 241, 131, 61 },
		{ 241, 132, 61 },
		{ 241, 133, 61 },
		{ 241, 134, 61 },
		{ 241, 135, 61 },
		{ 241, 136, 61 },
		{ 241, 137, 61 },
		{ 241, 138, 61 },
		{ 241, 139, 61 },
		{ 241, 140, 61 },
		{ 241, 141, 61 },
		{ 241, 142, 61 },
		{ 241, 143, 61 },
		{ 241, 144, 61 },
		{ 241, 145, 61 },
		{ 241, 146, 61 },
		{ 241, 147, 61 },
		{ 241, 148, 61 },
		{ 241, 149, 61 },
		{ 241, 150, 61 },
		{ 241, 151, 61 },
		{ 241, 152, 61 },
		{ 241, 153, 61 },
		{ 241, 154, 61 },
		{ 241, 155, 61 },
		{ 241, 156, 61 },
		{ 241, 157, 61 },
		{ 241, 158, 61 },
		{ 241, 159, 61 },
		{ 241, 160, 61 },
		{ 241, 161, 61 },
		{ 241, 162, 61 },
		{ 241, 163, 61 },
		{ 241, 164, 61 },
		{ 241, 165, 61 },
		{ 241, 166, 61 },
		{ 241, 167, 61 },
		{ 241, 168, 61 },
		{ 241, 169, 61 },
		{ 241, 170, 61 },
		{ 241, 171, 61 },
		{ 241, 172, 61 },
		{ 241, 173, 61 },
		{ 241, 174, 61 },
		{ 241, 175, 61 },
		{ 241, 176, 61 },
		{ 241, 177, 61 },
		{ 241, 178, 61 },
		{ 241, 179, 61 },
		{ 241, 180, 61 },
		{ 241, 181, 61 },
		{ 241, 182, 61 },
		{ 241, 183, 61 },
		{ 241, 184, 61 },
		{ 241, 185, 61 },
		{ 241, 186, 61 },
		{ 241, 187, 61 },
		{ 241, 188, 61 },
		{ 241, 189, 61 },
		{ 241, 190, 61 },
		{ 242, 1, 61 },
		{ 242, 2, 61 },
		{ 242, 3, 61 },
		{ 242, 4, 61 },
		{ 242, 5, 61 },
		{ 242, 6, 61 },
		{ 242, 7, 61 },
		{ 242, 8, 61 },
		{ 242, 9, 61 },
		{ 242, 10, 61 },
		{ 242, 11, 61 },
		{ 242, 12, 61 },
		{ 242, 13, 61 },
		{ 242, 14, 61 },
		{ 242, 15, 61 },
		{ 242, 16, 61 },
		{ 242, 17, 61 },
		{ 242, 18, 61 },
		{ 242, 19, 61 },
		{ 242, 20, 61 },
		{ 242, 21, 61 },
		{ 242, 22, 61 },
		{ 242, 23, 61 },
		{ 242, 24, 61 },
		{ 242, 25, 61 },
		{ 242, 26, 61 },
		{ 242, 27, 61 },
		{ 242, 28, 61 },
		{ 242, 29, 61 },
		{ 242, 30, 61 },
		{ 242, 31, 61 },
		{ 242, 32, 61 },
		{ 242, 33, 61 },
		{ 242, 34, 61 },
		{ 242, 35, 61 },
		{ 242, 36, 61 },
		{ 242, 37, 61 },
		{ 242, 38, 61 },
		{ 242, 39, 61 },
		{ 242, 40, 61 },
		{ 242, 41, 61 },
		{ 242, 42, 61 },
		{ 242, 43, 61 },
		{ 242, 44, 61 },
		{ 242, 45, 61 },
		{ 242, 46, 61 },
		{ 242, 47, 61 },
		{ 242, 48, 61 },
		{ 242, 49, 61 },
		{ 242, 50, 61 },
		{ 242, 51, 61 },
		{ 242, 52, 61 },
		{ 242, 53, 61 },
		{ 242, 54, 61 },
		{ 242, 55, 61 },
		{ 242, 56, 61 },
		{ 242, 57, 61 },
		{ 242, 58, 61 },
		{ 242, 59, 61 },
		{ 242, 60, 61 },
		{ 242, 61, 61 },
		{ 242, 62, 61 },
		{ 242, 63, 61 },
		{ 242, 64, 61 },
		{ 242, 65, 61 },
		{ 242, 66, 61 },
		{ 242, 67, 61 },
		{ 242, 68, 61 },
		{ 242, 69, 61 },
		{ 242, 70, 61 },
		{ 242, 71, 61 },
		{ 242, 72, 61 },
		{ 242, 73, 61 },
		{ 242, 74, 61 },
		{ 242, 75, 61 },
		{ 242, 76, 61 },
		{ 242, 77, 61 },
		{ 242, 78, 61 },
		{ 242, 79, 61 },
		{ 242, 80, 61 },
		{ 242, 81, 61 },
		{ 242, 82, 61 },
		{ 242, 83, 61 },
		{ 242, 84, 61 },
		{ 242, 85, 61 },
		{ 242, 86, 61 },
		{ 242, 87, 61 },
		{ 242, 88, 61 },
		{ 242, 89, 61 },
		{ 242, 90, 61 },
		{ 242, 91, 61 },
		{ 242, 92, 61 },
		{ 242, 93, 61 },
		{ 242, 94, 61 },
		{ 242, 95, 61 },
		{ 242, 96, 61 },
		{ 242, 97, 61 },
		{ 242, 98, 61 },
		{ 242, 99, 61 },
		{ 242, 100, 61 },
		{ 242, 101, 61 },
		{ 242, 102, 61 },
		{ 242, 103, 61 },
		{ 242, 104, 61 },
		{ 242, 105, 61 },
		{ 242, 106, 61 },
		{ 242, 107, 61 },
		{ 242, 108, 61 },
		{ 242, 109, 61 },
		{ 242, 110, 61 },
		{ 242, 111, 61 },
		{ 242, 112, 61 },
		{ 242, 113, 61 },
		{ 242, 114, 61 },
		{ 242, 115, 61 },
		{ 242, 116, 61 },
		{ 242, 117, 61 },
		{ 242, 118, 61 },
		{ 242, 119, 61 },
		{ 242, 120, 61 },
		{ 242, 121, 61 },
		{ 242, 122, 61 },
		{ 242, 123, 61 },
		{ 242, 124, 61 },
		{ 242, 125, 61 },
		{ 242, 126, 61 },
		{ 242, 127, 61 },
		{ 242, 128, 61 },
		{ 242, 129, 61 },
		{ 242, 130, 61 },
		{ 242, 131, 61 },
		{ 242, 132, 61 },
		{ 242, 133, 61 },
		{ 242, 134, 61 },
		{ 242, 135, 61 },
		{ 242, 136, 61 },
		{ 242, 137, 61 },
		{ 242, 138, 61 },
		{ 242, 139, 61 },
		{ 242, 140, 61 },
		{ 242, 141, 61 },
		{ 242, 142, 61 },
		{ 242, 143, 61 },
		{ 242, 144, 61 },
		{ 242, 145, 61 },
		{ 242, 146, 61 },
		{ 242, 147, 61 },
		{ 242, 148, 61 },
		{ 242, 149, 61 },
		{ 242, 150, 61 },
		{ 242, 151, 61 },
		{ 242, 152, 61 },
		{ 242, 153, 61 },
		{ 242, 154, 61 },
		{ 242, 155, 61 },
		{ 242, 156, 61 },
		{ 242, 157, 61 },
		{ 242, 158, 61 },
		{ 242, 159, 61 },
		{ 242, 160, 61 },
		{ 242, 161, 61 },
		{ 242, 162, 61 },
		{ 242, 163, 61 },
		{ 242, 164, 61 },
		{ 242, 165, 61 },
		{ 242, 166, 61 },
		{ 242, 167, 61 },
		{ 242, 168, 61 },
		{ 242, 169, 61 },
		{ 242, 170, 61 },
		{ 242, 171, 61 },
		{ 242, 172, 61 },
		{ 242, 173, 61 },
		{ 242, 174, 61 },
		{ 242, 175, 61 },
		{ 242, 176, 61 },
		{ 242, 177, 61 },
		{ 242, 178, 61 },
		{ 242, 179, 61 },
		{ 242, 180, 61 },
		{ 242, 181, 61 },
		{ 242, 182, 61 },
		{ 242, 183, 61 },
		{ 242, 184, 61 },
		{ 242, 185, 61 },
		{ 242, 186, 61 },
		{ 242, 187, 61 },
		{ 242, 188, 61 },
		{ 242, 189, 61 },
		{ 242, 190, 61 },
		{ 243, 1, 61 },
		{ 243, 2, 61 },
		{ 243, 3, 61 },
		{ 243, 4, 61 },
		{ 243, 5, 61 },
		{ 243, 6, 61 },
		{ 243, 7, 61 },
		{ 243, 8, 61 },
		{ 243, 9, 61 },
		{ 243, 10, 61 },
		{ 243, 11, 61 },
		{ 243, 12, 61 },
		{ 243, 13, 61 },
		{ 243, 14, 61 },
		{ 243, 15, 61 },
		{ 243, 16, 61 },
		{ 243, 17, 61 },
		{ 243, 18, 61 },
		{ 243, 19, 61 },
		{ 243, 20, 61 },
		{ 243, 21, 61 },
		{ 243, 22, 61 },
		{ 243, 23, 61 },
		{ 243, 24, 61 },
		{ 243, 25, 61 },
		{ 243, 26, 61 },
		{ 243, 27, 61 },
		{ 243, 28, 61 },
		{ 243, 29, 61 },
		{ 243, 30, 61 },
		{ 243, 31, 61 },
		{ 243, 32, 61 },
		{ 243, 33, 61 },
		{ 243, 34, 61 },
		{ 243, 35, 61 },
		{ 243, 36, 61 },
		{ 243, 37, 61 },
		{ 243, 38, 61 },
		{ 243, 39, 61 },
		{ 243, 40, 61 },
		{ 243, 41, 61 },
		{ 243, 42, 61 },
		{ 243, 43, 61 },
		{ 243, 44, 61 },
		{ 243, 45, 61 },
		{ 243, 46, 61 },
		{ 243, 47, 61 },
		{ 243, 48, 61 },
		{ 243, 49, 61 },
		{ 243, 50, 61 },
		{ 243, 51, 61 },
		{ 243, 52, 61 },
		{ 243, 53, 61 },
		{ 243, 54, 61 },
		{ 243, 55, 61 },
		{ 243, 56, 61 },
		{ 243, 57, 61 },
		{ 243, 58, 61 },
		{ 243, 59, 61 },
		{ 243, 60, 61 },
		{ 243, 61, 61 },
		{ 243, 62, 61 },
		{ 243, 63, 61 },
		{ 243, 64, 61 },
		{ 243, 65, 61 },
		{ 243, 66, 61 },
		{ 243, 67, 61 },
		{ 243, 68, 61 },
		{ 243, 69, 61 },
		{ 243, 70, 61 },
		{ 243, 71, 61 },
		{ 243, 72, 61 },
		{ 243, 73, 61 },
		{ 243, 74, 61 },
		{ 243, 75, 61 },
		{ 243, 76, 61 },
		{ 243, 77, 61 },
		{ 243, 78, 61 },
		{ 243, 79, 61 },
		{ 243, 80, 61 },
		{ 243, 81, 61 },
		{ 243, 82, 61 },
		{ 243, 83, 61 },
		{ 243, 84, 61 },
		{ 243, 85, 61 },
		{ 243, 86, 61 },
		{ 243, 87, 61 },
		{ 243, 88, 61 },
		{ 243, 89, 61 },
		{ 243, 90, 61 },
		{ 243, 91, 61 },
		{ 243, 92, 61 },
		{ 243, 93, 61 },
		{ 243, 94, 61 },
		{ 243, 95, 61 },
		{ 243, 96, 61 },
		{ 243, 97, 61 },
		{ 243, 98, 61 },
		{ 243, 99, 61 },
		{ 243, 100, 61 },
		{ 243, 101, 61 },
		{ 243, 102, 61 },
		{ 243, 103, 61 },
		{ 243, 104, 61 },
		{ 243, 105, 61 },
		{ 243, 106, 61 },
		{ 243, 107, 61 },
		{ 243, 108, 61 },
		{ 243, 109, 61 },
		{ 243, 110, 61 },
		{ 243, 111, 61 },
		{ 243, 112, 61 },
		{ 243, 113, 61 },
		{ 243, 114, 61 },
		{ 243, 115, 61 },
		{ 243, 116, 61 },
		{ 243, 117, 61 },
		{ 243, 118, 61 },
		{ 243, 119, 61 },
		{ 243, 120, 61 },
		{ 243, 121, 61 },
		{ 243, 122, 61 },
		{ 243, 123, 61 },
		{ 243, 124, 61 },
		{ 243, 125, 61 },
		{ 243, 126, 61 },
		{ 243, 127, 61 },
		{ 243, 128, 61 },
		{ 243, 129, 61 },
		{ 243, 130, 61 },
		{ 243, 131, 61 },
		{ 243, 132, 61 },
		{ 243, 133, 61 },
		{ 243, 134, 61 },
		{ 243, 135, 61 },
		{ 243, 136, 61 },
		{ 243, 137, 61 },
		{ 243, 138, 61 },
		{ 243, 139, 61 },
		{ 243, 140, 61 },
		{ 243, 141, 61 },
		{ 243, 142, 61 },
		{ 243, 143, 61 },
		{ 243, 144, 61 },
		{ 243, 145, 61 },
		{ 243, 146, 61 },
		{ 243, 147, 61 },
		{ 243, 148, 61 },
		{ 243, 149, 61 },
		{ 243, 150, 61 },
		{ 243, 151, 61 },
		{ 243, 152, 61 },
		{ 243, 153, 61 },
		{ 243, 154, 61 },
		{ 243, 155, 61 },
		{ 243, 156, 61 },
		{ 243, 157, 61 },
		{ 243, 158, 61 },
		{ 243, 159, 61 },
		{ 243, 160, 61 },
		{ 243, 161, 61 },
		{ 243, 162, 61 },
		{ 243, 163, 61 },
		{ 243, 164, 61 },
		{ 243, 165, 61 },
		{ 243, 166, 61 },
		{ 243, 167, 61 },
		{ 243, 168, 61 },
		{ 243, 169, 61 },
		{ 243, 170, 61 },
		{ 243, 171, 61 },
		{ 243, 172, 61 },
		{ 243, 173, 61 },
		{ 243, 174, 61 },
		{ 243, 175, 61 },
		{ 243, 176, 61 },
		{ 243, 177, 61 },
		{ 243, 178, 61 },
		{ 243, 179, 61 },
		{ 243, 180, 61 },
		{ 243, 181, 61 },
		{ 243, 182, 61 },
		{ 243, 183, 61 },
		{ 243, 184, 61 },
		{ 243, 185, 61 },
		{ 243, 186, 61 },
		{ 243, 187, 61 },
		{ 243, 188, 61 },
		{ 243, 189, 61 },
		{ 243, 190, 61 },
		{ 243, 191, 61 },
		{ 244, 1, 61 },
		{ 244, 2, 61 },
		{ 244, 3, 61 },
		{ 244, 4, 61 },
		{ 244, 5, 61 },
		{ 244, 6, 61 },
		{ 244, 7, 61 },
		{ 244, 8, 61 },
		{ 244, 9, 61 },
		{ 244, 10, 61 },
		{ 244, 11, 61 },
		{ 244, 12, 61 },
		{ 244, 13, 61 },
		{ 244, 14, 61 },
		{ 244, 15, 61 },
		{ 244, 16, 61 },
		{ 244, 17, 61 },
		{ 244, 18, 61 },
		{ 244, 19, 61 },
		{ 244, 20, 61 },
		{ 244, 21, 61 },
		{ 244, 22, 61 },
		{ 244, 23, 61 },
		{ 244, 24, 61 },
		{ 244, 25, 61 },
		{ 244, 26, 61 },
		{ 244, 27, 61 },
		{ 244, 28, 61 },
		{ 244, 29, 61 },
		{ 244, 30, 61 },
		{ 244, 31, 61 },
		{ 244, 32, 61 },
		{ 244, 33, 61 },
		{ 244, 34, 61 },
		{ 244, 35, 61 },
		{ 244, 36, 61 },
		{ 244, 37, 61 },
		{ 244, 38, 61 },
		{ 244, 39, 61 },
		{ 244, 40, 61 },
		{ 244, 41, 61 },
		{ 244, 42, 61 },
		{ 244, 43, 61 },
		{ 244, 44, 61 },
		{ 244, 45, 61 },
		{ 244, 46, 61 },
		{ 244, 47, 61 },
		{ 244, 48, 61 },
		{ 244, 49, 61 },
		{ 244, 50, 61 },
		{ 244, 51, 61 },
		{ 244, 52, 61 },
		{ 244, 53, 61 },
		{ 244, 54, 61 },
		{ 244, 55, 61 },
		{ 244, 56, 61 },
		{ 244, 57, 61 },
		{ 244, 58, 61 },
		{ 244, 59, 61 },
		{ 244, 60, 61 },
		{ 244, 61, 61 },
		{ 244, 62, 61 },
		{ 244, 63, 61 },
		{ 244, 64, 61 },
		{ 244, 65, 61 },
		{ 244, 66, 61 },
		{ 244, 67, 61 },
		{ 244, 68, 61 },
		{ 244, 69, 61 },
		{ 244, 70, 61 },
		{ 244, 71, 61 },
		{ 244, 72, 61 },
		{ 244, 73, 61 },
		{ 244, 74, 61 },
		{ 244, 75, 61 },
		{ 244, 76, 61 },
		{ 244, 77, 61 },
		{ 244, 78, 61 },
		{ 244, 79, 61 },
		{ 244, 80, 61 },
		{ 244, 81, 61 },
		{ 244, 82, 61 },
		{ 244, 83, 61 },
		{ 244, 84, 61 },
		{ 244, 85, 61 },
		{ 244, 86, 61 },
		{ 244, 87, 61 },
		{ 244, 88, 61 },
		{ 244, 89, 61 },
		{ 244, 90, 61 },
		{ 244, 91, 61 },
		{ 244, 92, 61 },
		{ 244, 93, 61 },
		{ 244, 94, 61 },
		{ 244, 95, 61 },
		{ 244, 96, 61 },
		{ 244, 97, 61 },
		{ 244, 98, 61 },
		{ 244, 99, 61 },
		{ 244, 100, 61 },
		{ 244, 101, 61 },
		{ 244, 102, 61 },
		{ 244, 103, 61 },
		{ 244, 104, 61 },
		{ 244, 105, 61 },
		{ 244, 106, 61 },
		{ 244, 107, 61 },
		{ 244, 108, 61 },
		{ 244, 109, 61 },
		{ 244, 110, 61 },
		{ 244, 111, 61 },
		{ 244, 112, 61 },
		{ 244, 113, 61 },
		{ 244, 114, 61 },
		{ 244, 115, 61 },
		{ 244, 116, 61 },
		{ 244, 117, 61 },
		{ 244, 118, 61 },
		{ 244, 119, 61 },
		{ 244, 120, 61 },
		{ 244, 121, 61 },
		{ 244, 122, 61 },
		{ 244, 123, 61 },
		{ 244, 124, 61 },
		{ 244, 125, 61 },
		{ 244, 126, 61 },
		{ 244, 127, 61 },
		{ 244, 128, 61 },
		{ 244, 129, 61 },
		{ 244, 130, 61 },
		{ 244, 131, 61 },
		{ 244, 132, 61 },
		{ 244, 133, 61 },
		{ 244, 134, 61 },
		{ 244, 135, 61 },
		{ 244, 136, 61 },
		{ 244, 137, 61 },
		{ 244, 138, 61 },
		{ 244, 139, 61 },
		{ 244, 140, 61 },
		{ 244, 141, 61 },
		{ 244, 142, 61 },
		{ 244, 143, 61 },
		{ 244, 144, 61 },
		{ 244, 145, 61 },
		{ 244, 146, 61 },
		{ 244, 147, 61 },
		{ 244, 148, 61 },
		{ 244, 149, 61 },
		{ 244, 150, 61 },
		{ 244, 151, 61 },
		{ 244, 152, 61 },
		{ 244, 153, 61 },
		{ 244, 154, 61 },
		{ 244, 155, 61 },
		{ 244, 156, 61 },
		{ 244, 157, 61 },
		{ 244, 158, 61 },
		{ 244, 159, 61 },
		{ 244, 160, 61 },
		{ 244, 161, 61 },
		{ 244, 162, 61 },
		{ 244, 163, 61 },
		{ 244, 164, 61 },
		{ 244, 165, 61 },
		{ 244, 166, 61 },
		{ 244, 167, 61 },
		{ 244, 168, 61 },
		{ 244, 169, 61 },
		{ 244, 170, 61 },
		{ 244, 171, 61 },
		{ 244, 172, 61 },
		{ 244, 173, 61 },
		{ 244, 174, 61 },
		{ 244, 175, 61 },
		{ 244, 176, 61 },
		{ 244, 177, 61 },
		{ 244, 178, 61 },
		{ 244, 179, 61 },
		{ 244, 180, 61 },
		{ 244, 181, 61 },
		{ 244, 182, 61 },
		{ 244, 183, 61 },
		{ 244, 184, 61 },
		{ 244, 185, 61 },
		{ 244, 186, 61 },
		{ 244, 187, 61 },
		{ 244, 188, 61 },
		{ 244, 189, 61 },
		{ 244, 190, 61 },
		{ 244, 191, 61 },
		{ 245, 1, 61 },
		{ 245, 2, 61 },
		{ 245, 3, 61 },
		{ 245, 4, 61 },
		{ 245, 5, 61 },
		{ 245, 6, 61 },
		{ 245, 7, 61 },
		{ 245, 8, 61 },
		{ 245, 9, 61 },
		{ 245, 10, 61 },
		{ 245, 11, 61 },
		{ 245, 12, 61 },
		{ 245, 13, 61 },
		{ 245, 14, 61 },
		{ 245, 15, 61 },
		{ 245, 16, 61 },
		{ 245, 17, 61 },
		{ 245, 18, 61 },
		{ 245, 19, 61 },
		{ 245, 20, 61 },
		{ 245, 21, 61 },
		{ 245, 22, 61 },
		{ 245, 23, 61 },
		{ 245, 24, 61 },
		{ 245, 25, 61 },
		{ 245, 26, 61 },
		{ 245, 27, 61 },
		{ 245, 28, 61 },
		{ 245, 29, 61 },
		{ 245, 30, 61 },
		{ 245, 31, 61 },
		{ 245, 32, 61 },
		{ 245, 33, 61 },
		{ 245, 34, 61 },
		{ 245, 35, 61 },
		{ 245, 36, 61 },
		{ 245, 37, 61 },
		{ 245, 38, 61 },
		{ 245, 39, 61 },
		{ 245, 40, 61 },
		{ 245, 41, 61 },
		{ 245, 42, 61 },
		{ 245, 43, 61 },
		{ 245, 44, 61 },
		{ 245, 45, 61 },
		{ 245, 46, 61 },
		{ 245, 47, 61 },
		{ 245, 48, 61 },
		{ 245, 49, 61 },
		{ 245, 50, 61 },
		{ 245, 51, 61 },
		{ 245, 52, 61 },
		{ 245, 53, 61 },
		{ 245, 54, 61 },
		{ 245, 55, 61 },
		{ 245, 56, 61 },
		{ 245, 57, 61 },
		{ 245, 58, 61 },
		{ 245, 59, 61 },
		{ 245, 60, 61 },
		{ 245, 61, 61 },
		{ 245, 62, 61 },
		{ 245, 63, 61 },
		{ 245, 64, 61 },
		{ 245, 65, 61 },
		{ 245, 66, 61 },
		{ 245, 67, 61 },
		{ 245, 68, 61 },
		{ 245, 69, 61 },
		{ 245, 70, 61 },
		{ 245, 71, 61 },
		{ 245, 72, 61 },
		{ 245, 73, 61 },
		{ 245, 74, 61 },
		{ 245, 75, 61 },
		{ 245, 76, 61 },
		{ 245, 77, 61 },
		{ 245, 78, 61 },
		{ 245, 79, 61 },
		{ 245, 80, 61 },
		{ 245, 81, 61 },
		{ 245, 82, 61 },
		{ 245, 83, 61 },
		{ 245, 84, 61 },
		{ 245, 85, 61 },
		{ 245, 86, 61 },
		{ 245, 87, 61 },
		{ 245, 88, 61 },
		{ 245, 89, 61 },
		{ 245, 90, 61 },
		{ 245, 91, 61 },
		{ 245, 92, 61 },
		{ 245, 93, 61 },
		{ 245, 94, 61 },
		{ 245, 95, 61 },
		{ 245, 96, 61 },
		{ 245, 97, 61 },
		{ 245, 98, 61 },
		{ 245, 99, 61 },
		{ 245, 100, 61 },
		{ 245, 101, 61 },
		{ 245, 102, 61 },
		{ 245, 103, 61 },
		{ 245, 104, 61 },
		{ 245, 105, 61 },
		{ 245, 106, 61 },
		{ 245, 107, 61 },
		{ 245, 108, 61 },
		{ 245, 109, 61 },
		{ 245, 110, 61 },
		{ 245, 111, 61 },
		{ 245, 112, 61 },
		{ 245, 113, 61 },
		{ 245, 114, 61 },
		{ 245, 115, 61 },
		{ 245, 116, 61 },
		{ 245, 117, 61 },
		{ 245, 118, 61 },
		{ 245, 119, 61 },
		{ 245, 120, 61 },
		{ 245, 121, 61 },
		{ 245, 122, 61 },
		{ 245, 123, 61 },
		{ 245, 124, 61 },
		{ 245, 125, 61 },
		{ 245, 126, 61 },
		{ 245, 127, 61 },
		{ 245, 128, 61 },
		{ 245, 129, 61 },
		{ 245, 130, 61 },
		{ 245, 131, 61 },
		{ 245, 132, 61 },
		{ 245, 133, 61 },
		{ 245, 134, 61 },
		{ 245, 135, 61 },
		{ 245, 136, 61 },
		{ 245, 137, 61 },
		{ 245, 138, 61 },
		{ 245, 139, 61 },
		{ 245, 140, 61 },
		{ 245, 141, 61 },
		{ 245, 142, 61 },
		{ 245, 143, 61 },
		{ 245, 144, 61 },
		{ 245, 145, 61 },
		{ 245, 146, 61 },
		{ 245, 147, 61 },
		{ 245, 148, 61 },
		{ 245, 149, 61 },
		{ 245, 150, 61 },
		{ 245, 151, 61 },
		{ 245, 152, 61 },
		{ 245, 153, 61 },
		{ 245, 154, 61 },
		{ 245, 155, 61 },
		{ 245, 156, 61 },
		{ 245, 157, 61 },
		{ 245, 158, 61 },
		{ 245, 159, 61 },
		{ 245, 160, 61 },
		{ 245, 161, 61 },
		{ 245, 162, 61 },
		{ 245, 163, 61 },
		{ 245, 164, 61 },
		{ 245, 165, 61 },
		{ 245, 166, 61 },
		{ 245, 167, 61 },
		{ 245, 168, 61 },
		{ 245, 169, 61 },
		{ 245, 170, 61 },
		{ 245, 171, 61 },
		{ 245, 172, 61 },
		{ 245, 173, 61 },
		{ 245, 174, 61 },
		{ 245, 175, 61 },
		{ 245, 176, 61 },
		{ 245, 177, 61 },
		{ 245, 178, 61 },
		{ 245, 179, 61 },
		{ 245, 180, 61 },
		{ 245, 181, 61 },
		{ 245, 182, 61 },
		{ 245, 183, 61 },
		{ 245, 184, 61 },
		{ 245, 185, 61 },
		{ 245, 186, 61 },
		{ 245, 187, 61 },
		{ 245, 188, 61 },
		{ 245, 189, 61 },
		{ 245, 190, 61 },
		{ 245, 191, 61 },
		{ 246, 1, 61 },
		{ 246, 2, 61 },
		{ 246, 3, 61 },
		{ 246, 4, 61 },
		{ 246, 5, 61 },
		{ 246, 6, 61 },
		{ 246, 7, 61 },
		{ 246, 8, 61 },
		{ 246, 9, 61 },
		{ 246, 10, 61 },
		{ 246, 11, 61 },
		{ 246, 12, 61 },
		{ 246, 13, 61 },
		{ 246, 14, 61 },
		{ 246, 15, 61 },
		{ 246, 16, 61 },
		{ 246, 17, 61 },
		{ 246, 18, 61 },
		{ 246, 19, 61 },
		{ 246, 20, 61 },
		{ 246, 21, 61 },
		{ 246, 22, 61 },
		{ 246, 23, 61 },
		{ 246, 24, 61 },
		{ 246, 25, 61 },
		{ 246, 26, 61 },
		{ 246, 27, 61 },
		{ 246, 28, 61 },
		{ 246, 29, 61 },
		{ 246, 30, 61 },
		{ 246, 31, 61 },
		{ 246, 32, 61 },
		{ 246, 33, 61 },
		{ 246, 34, 61 },
		{ 246, 35, 61 },
		{ 246, 36, 61 },
		{ 246, 37, 61 },
		{ 246, 38, 61 },
		{ 246, 39, 61 },
		{ 246, 40, 61 },
		{ 246, 41, 61 },
		{ 246, 42, 61 },
		{ 246, 43, 61 },
		{ 246, 44, 61 },
		{ 246, 45, 61 },
		{ 246, 46, 61 },
		{ 246, 47, 61 },
		{ 246, 48, 61 },
		{ 246, 49, 61 },
		{ 246, 50, 61 },
		{ 246, 51, 61 },
		{ 246, 52, 61 },
		{ 246, 53, 61 },
		{ 246, 54, 61 },
		{ 246, 55, 61 },
		{ 246, 56, 61 },
		{ 246, 57, 61 },
		{ 246, 58, 61 },
		{ 246, 59, 61 },
		{ 246, 60, 61 },
		{ 246, 61, 61 },
		{ 246, 62, 61 },
		{ 246, 63, 61 },
		{ 246, 64, 61 },
		{ 246, 65, 61 },
		{ 246, 66, 61 },
		{ 246, 67, 61 },
		{ 246, 68, 61 },
		{ 246, 69, 61 },
		{ 246, 70, 61 },
		{ 246, 71, 61 },
		{ 246, 72, 61 },
		{ 246, 73, 61 },
		{ 246, 74, 61 },
		{ 246, 75, 61 },
		{ 246, 76, 61 },
		{ 246, 77, 61 },
		{ 246, 78, 61 },
		{ 246, 79, 61 },
		{ 246, 80, 61 },
		{ 246, 81, 61 },
		{ 246, 82, 61 },
		{ 246, 83, 61 },
		{ 246, 84, 61 },
		{ 246, 85, 61 },
		{ 246, 86, 61 },
		{ 246, 87, 61 },
		{ 246, 88, 61 },
		{ 246, 89, 61 },
		{ 246, 90, 61 },
		{ 246, 91, 61 },
		{ 246, 92, 61 },
		{ 246, 93, 61 },
		{ 246, 94, 61 },
		{ 246, 95, 61 },
		{ 246, 96, 61 },
		{ 246, 97, 61 },
		{ 246, 98, 61 },
		{ 246, 99, 61 },
		{ 246, 100, 61 },
		{ 246, 101, 61 },
		{ 246, 102, 61 },
		{ 246, 103, 61 },
		{ 246, 104, 61 },
		{ 246, 105, 61 },
		{ 246, 106, 61 },
		{ 246, 107, 61 },
		{ 246, 108, 61 },
		{ 246, 109, 61 },
		{ 246, 110, 61 },
		{ 246, 111, 61 },
		{ 246, 112, 61 },
		{ 246, 113, 61 },
		{ 246, 114, 61 },
		{ 246, 115, 61 },
		{ 246, 116, 61 },
		{ 246, 117, 61 },
		{ 246, 118, 61 },
		{ 246, 119, 61 },
		{ 246, 120, 61 },
		{ 246, 121, 61 },
		{ 246, 122, 61 },
		{ 246, 123, 61 },
		{ 246, 124, 61 },
		{ 246, 125, 61 },
		{ 246, 126, 61 },
		{ 246, 127, 61 },
		{ 246, 128, 61 },
		{ 246, 129, 61 },
		{ 246, 130, 61 },
		{ 246, 131, 61 },
		{ 246, 132, 61 },
		{ 246, 133, 61 },
		{ 246, 134, 61 },
		{ 246, 135, 61 },
		{ 246, 136, 61 },
		{ 246, 137, 61 },
		{ 246, 138, 61 },
		{ 246, 139, 61 },
		{ 246, 140, 61 },
		{ 246, 141, 61 },
		{ 246, 142, 61 },
		{ 246, 143, 61 },
		{ 246, 144, 61 },
		{ 246, 145, 61 },
		{ 246, 146, 61 },
		{ 246, 147, 61 },
		{ 246, 148, 61 },
		{ 246, 149, 61 },
		{ 246, 150, 61 },
		{ 246, 151, 61 },
		{ 246, 152, 61 },
		{ 246, 153, 61 },
		{ 246, 154, 61 },
		{ 246, 155, 61 },
		{ 246, 156, 61 },
		{ 246, 157, 61 },
		{ 246, 158, 61 },
		{ 246, 159, 61 },
		{ 246, 160, 61 },
		{ 246, 161, 61 },
		{ 246, 162, 61 },
		{ 246, 163, 61 },
		{ 246, 164, 61 },
		{ 246, 165, 61 },
		{ 246, 166, 61 },
		{ 246, 167, 61 },
		{ 246, 168, 61 },
		{ 246, 169, 61 },
		{ 246, 170, 61 },
		{ 246, 171, 61 },
		{ 246, 172, 61 },
		{ 246, 173, 61 },
		{ 246, 174, 61 },
		{ 246, 175, 61 },
		{ 246, 176, 61 },
		{ 246, 177, 61 },
		{ 246, 178, 61 },
		{ 246, 179, 61 },
		{ 246, 180, 61 },
		{ 246, 181, 61 },
		{ 246, 182, 61 },
		{ 246, 183, 61 },
		{ 246, 184, 61 },
		{ 246, 185, 61 },
		{ 246, 186, 61 },
		{ 246, 187, 61 },
		{ 246, 188, 61 },
		{ 246, 189, 61 },
		{ 246, 190, 61 },
		{ 246, 191, 61 },
		{ 247, 1, 61 },
		{ 247, 2, 61 },
		{ 247, 3, 61 },
		{ 247, 4, 61 },
		{ 247, 5, 61 },
		{ 247, 6, 61 },
		{ 247, 7, 61 },
		{ 247, 8, 61 },
		{ 247, 9, 61 },
		{ 247, 10, 61 },
		{ 247, 11, 61 },
		{ 247, 12, 61 },
		{ 247, 13, 61 },
		{ 247, 14, 61 },
		{ 247, 15, 61 },
		{ 247, 16, 61 },
		{ 247, 17, 61 },
		{ 247, 18, 61 },
		{ 247, 19, 61 },
		{ 247, 20, 61 },
		{ 247, 21, 61 },
		{ 247, 22, 61 },
		{ 247, 23, 61 },
		{ 247, 24, 61 },
		{ 247, 25, 61 },
		{ 247, 26, 61 },
		{ 247, 27, 61 },
		{ 247, 28, 61 },
		{ 247, 29, 61 },
		{ 247, 30, 61 },
		{ 247, 31, 61 },
		{ 247, 32, 61 },
		{ 247, 33, 61 },
		{ 247, 34, 61 },
		{ 247, 35, 61 },
		{ 247, 36, 61 },
		{ 247, 37, 61 },
		{ 247, 38, 61 },
		{ 247, 39, 61 },
		{ 247, 40, 61 },
		{ 247, 41, 61 },
		{ 247, 42, 61 },
		{ 247, 43, 61 },
		{ 247, 44, 61 },
		{ 247, 45, 61 },
		{ 247, 46, 61 },
		{ 247, 47, 61 },
		{ 247, 48, 61 },
		{ 247, 49, 61 },
		{ 247, 50, 61 },
		{ 247, 51, 61 },
		{ 247, 52, 61 },
		{ 247, 53, 61 },
		{ 247, 54, 61 },
		{ 247, 55, 61 },
		{ 247, 56, 61 },
		{ 247, 57, 61 },
		{ 247, 58, 61 },
		{ 247, 59, 61 },
		{ 247, 60, 61 },
		{ 247, 61, 61 },
		{ 247, 62, 61 },
		{ 247, 63, 61 },
		{ 247, 64, 61 },
		{ 247, 65, 61 },
		{ 247, 66, 61 },
		{ 247, 67, 61 },
		{ 247, 68, 61 },
		{ 247, 69, 61 },
		{ 247, 70, 61 },
		{ 247, 71, 61 },
		{ 247, 72, 61 },
		{ 247, 73, 61 },
		{ 247, 74, 61 },
		{ 247, 75, 61 },
		{ 247, 76, 61 },
		{ 247, 77, 61 },
		{ 247, 78, 61 },
		{ 247, 79, 61 },
		{ 247, 80, 61 },
		{ 247, 81, 61 },
		{ 247, 82, 61 },
		{ 247, 83, 61 },
		{ 247, 84, 61 },
		{ 247, 85, 61 },
		{ 247, 86, 61 },
		{ 247, 87, 61 },
		{ 247, 88, 61 },
		{ 247, 89, 61 },
		{ 247, 90, 61 },
		{ 247, 91, 61 },
		{ 247, 92, 61 },
		{ 247, 93, 61 },
		{ 247, 94, 61 },
		{ 247, 95, 61 },
		{ 247, 96, 61 },
		{ 247, 97, 61 },
		{ 247, 98, 61 },
		{ 247, 99, 61 },
		{ 247, 100, 61 },
		{ 247, 101, 61 },
		{ 247, 102, 61 },
		{ 247, 103, 61 },
		{ 247, 104, 61 },
		{ 247, 105, 61 },
		{ 247, 106, 61 },
		{ 247, 107, 61 },
		{ 247, 108, 61 },
		{ 247, 109, 61 },
		{ 247, 110, 61 },
		{ 247, 111, 61 },
		{ 247, 112, 61 },
		{ 247, 113, 61 },
		{ 247, 114, 61 },
		{ 247, 115, 61 },
		{ 247, 116, 61 },
		{ 247, 117, 61 },
		{ 247, 118, 61 },
		{ 247, 119, 61 },
		{ 247, 120, 61 },
		{ 247, 121, 61 },
		{ 247, 122, 61 },
		{ 247, 123, 61 },
		{ 247, 124, 61 },
		{ 247, 125, 61 },
		{ 247, 126, 61 },
		{ 247, 127, 61 },
		{ 247, 128, 61 },
		{ 247, 129, 61 },
		{ 247, 130, 61 },
		{ 247, 131, 61 },
		{ 247, 132, 61 },
		{ 247, 133, 61 },
		{ 247, 134, 61 },
		{ 247, 135, 61 },
		{ 247, 136, 61 },
		{ 247, 137, 61 },
		{ 247, 138, 61 },
		{ 247, 139, 61 },
		{ 247, 140, 61 },
		{ 247, 141, 61 },
		{ 247, 142, 61 },
		{ 247, 143, 61 },
		{ 247, 144, 61 },
		{ 247, 145, 61 },
		{ 247, 146, 61 },
		{ 247, 147, 61 },
		{ 247, 148, 61 },
		{ 247, 149, 61 },
		{ 247, 150, 61 },
		{ 247, 151, 61 },
		{ 247, 152, 61 },
		{ 247, 153, 61 },
		{ 247, 154, 61 },
		{ 247, 155, 61 },
		{ 247, 156, 61 },
		{ 247, 157, 61 },
		{ 247, 158, 61 },
		{ 247, 159, 61 },
		{ 247, 160, 61 },
		{ 247, 161, 61 },
		{ 247, 162, 61 },
		{ 247, 163, 61 },
		{ 247, 164, 61 },
		{ 247, 165, 61 },
		{ 247, 166, 61 },
		{ 247, 167, 61 },
		{ 247, 168, 61 },
		{ 247, 169, 61 },
		{ 247, 170, 61 },
		{ 247, 171, 61 },
		{ 247, 172, 61 },
		{ 247, 173, 61 },
		{ 247, 174, 61 },
		{ 247, 175, 61 },
		{ 247, 176, 61 },
		{ 247, 177, 61 },
		{ 247, 178, 61 },
		{ 247, 179, 61 },
		{ 247, 180, 61 },
		{ 247, 181, 61 },
		{ 247, 182, 61 },
		{ 247, 183, 61 },
		{ 247, 184, 61 },
		{ 247, 185, 61 },
		{ 247, 186, 61 },
		{ 247, 187, 61 },
		{ 247, 188, 61 },
		{ 247, 189, 61 },
		{ 247, 190, 61 },
		{ 247, 191, 61 },
		{ 248, 1, 61 },
		{ 248, 2, 61 },
		{ 248, 3, 61 },
		{ 248, 4, 61 },
		{ 248, 5, 61 },
		{ 248, 6, 61 },
		{ 248, 7, 61 },
		{ 248, 8, 61 },
		{ 248, 9, 61 },
		{ 248, 10, 61 },
		{ 248, 11, 61 },
		{ 248, 12, 61 },
		{ 248, 13, 61 },
		{ 248, 14, 61 },
		{ 248, 15, 61 },
		{ 248, 16, 61 },
		{ 248, 17, 61 },
		{ 248, 18, 61 },
		{ 248, 19, 61 },
		{ 248, 20, 61 },
		{ 248, 21, 61 },
		{ 248, 22, 61 },
		{ 248, 23, 61 },
		{ 248, 24, 61 },
		{ 248, 25, 61 },
		{ 248, 26, 61 },
		{ 248, 27, 61 },
		{ 248, 28, 61 },
		{ 248, 29, 61 },
		{ 248, 30, 61 },
		{ 248, 31, 61 },
		{ 248, 32, 61 },
		{ 248, 33, 61 },
		{ 248, 34, 61 },
		{ 248, 35, 61 },
		{ 248, 36, 61 },
		{ 248, 37, 61 },
		{ 248, 38, 61 },
		{ 248, 39, 61 },
		{ 248, 40, 61 },
		{ 248, 41, 61 },
		{ 248, 42, 61 },
		{ 248, 43, 61 },
		{ 248, 44, 61 },
		{ 248, 45, 61 },
		{ 248, 46, 61 },
		{ 248, 47, 61 },
		{ 248, 48, 61 },
		{ 248, 49, 61 },
		{ 248, 50, 61 },
		{ 248, 51, 61 },
		{ 248, 52, 61 },
		{ 248, 53, 61 },
		{ 248, 54, 61 },
		{ 248, 55, 61 },
		{ 248, 56, 61 },
		{ 248, 57, 61 },
		{ 248, 58, 61 },
		{ 248, 59, 61 },
		{ 248, 60, 61 },
		{ 248, 61, 61 },
		{ 248, 62, 61 },
		{ 248, 63, 61 },
		{ 248, 64, 61 },
		{ 248, 65, 61 },
		{ 248, 66, 61 },
		{ 248, 67, 61 },
		{ 248, 68, 61 },
		{ 248, 69, 61 },
		{ 248, 70, 61 },
		{ 248, 71, 61 },
		{ 248, 72, 61 },
		{ 248, 73, 61 },
		{ 248, 74, 61 },
		{ 248, 75, 61 },
		{ 248, 76, 61 },
		{ 248, 77, 61 },
		{ 248, 78, 61 },
		{ 248, 79, 61 },
		{ 248, 80, 61 },
		{ 248, 81, 61 },
		{ 248, 82, 61 },
		{ 248, 83, 61 },
		{ 248, 84, 61 },
		{ 248, 85, 61 },
		{ 248, 86, 61 },
		{ 248, 87, 61 },
		{ 248, 88, 61 },
		{ 248, 89, 61 },
		{ 248, 90, 61 },
		{ 248, 91, 61 },
		{ 248, 92, 61 },
		{ 248, 93, 61 },
		{ 248, 94, 61 },
		{ 248, 95, 61 },
		{ 248, 96, 61 },
		{ 248, 97, 61 },
		{ 248, 98, 61 },
		{ 248, 99, 61 },
		{ 248, 100, 61 },
		{ 248, 101, 61 },
		{ 248, 102, 61 },
		{ 248, 103, 61 },
		{ 248, 104, 61 },
		{ 248, 105, 61 },
		{ 248, 106, 61 },
		{ 248, 107, 61 },
		{ 248, 108, 61 },
		{ 248, 109, 61 },
		{ 248, 110, 61 },
		{ 248, 111, 61 },
		{ 248, 112, 61 },
		{ 248, 113, 61 },
		{ 248, 114, 61 },
		{ 248, 115, 61 },
		{ 248, 116, 61 },
		{ 248, 117, 61 },
		{ 248, 118, 61 },
		{ 248, 119, 61 },
		{ 248, 120, 61 },
		{ 248, 121, 61 },
		{ 248, 122, 61 },
		{ 248, 123, 61 },
		{ 248, 124, 61 },
		{ 248, 125, 61 },
		{ 248, 126, 61 },
		{ 248, 127, 61 },
		{ 248, 128, 61 },
		{ 248, 129, 61 },
		{ 248, 130, 61 },
		{ 248, 131, 61 },
		{ 248, 132, 61 },
		{ 248, 133, 61 },
		{ 248, 134, 61 },
		{ 248, 135, 61 },
		{ 248, 136, 61 },
		{ 248, 137, 61 },
		{ 248, 138, 61 },
		{ 248, 139, 61 },
		{ 248, 140, 61 },
		{ 248, 141, 61 },
		{ 248, 142, 61 },
		{ 248, 143, 61 },
		{ 248, 144, 61 },
		{ 248, 145, 61 },
		{ 248, 146, 61 },
		{ 248, 147, 61 },
		{ 248, 148, 61 },
		{ 248, 149, 61 },
		{ 248, 150, 61 },
		{ 248, 151, 61 },
		{ 248, 152, 61 },
		{ 248, 153, 61 },
		{ 248, 154, 61 },
		{ 248, 155, 61 },
		{ 248, 156, 61 },
		{ 248, 157, 61 },
		{ 248, 158, 61 },
		{ 248, 159, 61 },
		{ 248, 160, 61 },
		{ 248, 161, 61 },
		{ 248, 162, 61 },
		{ 248, 163, 61 },
		{ 248, 164, 61 },
		{ 248, 165, 61 },
		{ 248, 166, 61 },
		{ 248, 167, 61 },
		{ 248, 168, 61 },
		{ 248, 169, 61 },
		{ 248, 170, 61 },
		{ 248, 171, 61 },
		{ 248, 172, 61 },
		{ 248, 173, 61 },
		{ 248, 174, 61 },
		{ 248, 175, 61 },
		{ 248, 176, 61 },
		{ 248, 177, 61 },
		{ 248, 178, 61 },
		{ 248, 179, 61 },
		{ 248, 180, 61 },
		{ 248, 181, 61 },
		{ 248, 182, 61 },
		{ 248, 183, 61 },
		{ 248, 184, 61 },
		{ 248, 185, 61 },
		{ 248, 186, 61 },
		{ 248, 187, 61 },
		{ 248, 188, 61 },
		{ 248, 189, 61 },
		{ 248, 190, 61 },
		{ 248, 191, 61 },
		{ 249, 1, 61 },
		{ 249, 2, 61 },
		{ 249, 3, 61 },
		{ 249, 4, 61 },
		{ 249, 5, 61 },
		{ 249, 6, 61 },
		{ 249, 7, 61 },
		{ 249, 8, 61 },
		{ 249, 9, 61 },
		{ 249, 10, 61 },
		{ 249, 11, 61 },
		{ 249, 12, 61 },
		{ 249, 13, 61 },
		{ 249, 14, 61 },
		{ 249, 15, 61 },
		{ 249, 16, 61 },
		{ 249, 17, 61 },
		{ 249, 18, 61 },
		{ 249, 19, 61 },
		{ 249, 20, 61 },
		{ 249, 21, 61 },
		{ 249, 22, 61 },
		{ 249, 23, 61 },
		{ 249, 24, 61 },
		{ 249, 25, 61 },
		{ 249, 26, 61 },
		{ 249, 27, 61 },
		{ 249, 28, 61 },
		{ 249, 29, 61 },
		{ 249, 30, 61 },
		{ 249, 31, 61 },
		{ 249, 32, 61 },
		{ 249, 33, 61 },
		{ 249, 34, 61 },
		{ 249, 35, 61 },
		{ 249, 36, 61 },
		{ 249, 37, 61 },
		{ 249, 38, 61 },
		{ 249, 39, 61 },
		{ 249, 40, 61 },
		{ 249, 41, 61 },
		{ 249, 42, 61 },
		{ 249, 43, 61 },
		{ 249, 44, 61 },
		{ 249, 45, 61 },
		{ 249, 46, 61 },
		{ 249, 47, 61 },
		{ 249, 48, 61 },
		{ 249, 49, 61 },
		{ 249, 50, 61 },
		{ 249, 51, 61 },
		{ 249, 52, 61 },
		{ 249, 53, 61 },
		{ 249, 54, 61 },
		{ 249, 55, 61 },
		{ 249, 56, 61 },
		{ 249, 57, 61 },
		{ 249, 58, 61 },
		{ 249, 59, 61 },
		{ 249, 60, 61 },
		{ 249, 61, 61 },
		{ 249, 62, 61 },
		{ 249, 63, 61 },
		{ 249, 64, 61 },
		{ 249, 65, 61 },
		{ 249, 66, 61 },
		{ 249, 67, 61 },
		{ 249, 68, 61 },
		{ 249, 69, 61 },
		{ 249, 70, 61 },
		{ 249, 71, 61 },
		{ 249, 72, 61 },
		{ 249, 73, 61 },
		{ 249, 74, 61 },
		{ 249, 75, 61 },
		{ 249, 76, 61 },
		{ 249, 77, 61 },
		{ 249, 78, 61 },
		{ 249, 79, 61 },
		{ 249, 80, 61 },
		{ 249, 81, 61 },
		{ 249, 82, 61 },
		{ 249, 83, 61 },
		{ 249, 84, 61 },
		{ 249, 85, 61 },
		{ 249, 86, 61 },
		{ 249, 87, 61 },
		{ 249, 88, 61 },
		{ 249, 89, 61 },
		{ 249, 90, 61 },
		{ 249, 91, 61 },
		{ 249, 92, 61 },
		{ 249, 93, 61 },
		{ 249, 94, 61 },
		{ 249, 95, 61 },
		{ 249, 96, 61 },
		{ 249, 97, 61 },
		{ 249, 98, 61 },
		{ 249, 99, 61 },
		{ 249, 100, 61 },
		{ 249, 101, 61 },
		{ 249, 102, 61 },
		{ 249, 103, 61 },
		{ 249, 104, 61 },
		{ 249, 105, 61 },
		{ 249, 106, 61 },
		{ 249, 107, 61 },
		{ 249, 108, 61 },
		{ 249, 109, 61 },
		{ 249, 110, 61 },
		{ 249, 111, 61 },
		{ 249, 112, 61 },
		{ 249, 113, 61 },
		{ 249, 114, 61 },
		{ 249, 115, 61 },
		{ 249, 116, 61 },
		{ 249, 117, 61 },
		{ 249, 118, 61 },
		{ 249, 119, 61 },
		{ 249, 120, 61 },
		{ 249, 121, 61 },
		{ 249, 122, 61 },
		{ 249, 123, 61 },
		{ 249, 124, 61 },
		{ 249, 125, 61 },
		{ 249, 126, 61 },
		{ 249, 127, 61 },
		{ 249, 128, 61 },
		{ 249, 129, 61 },
		{ 249, 130, 61 },
		{ 249, 131, 61 },
		{ 249, 132, 61 },
		{ 249, 133, 61 },
		{ 249, 134, 61 },
		{ 249, 135, 61 },
		{ 249, 136, 61 },
		{ 249, 137, 61 },
		{ 249, 138, 61 },
		{ 249, 139, 61 },
		{ 249, 140, 61 },
		{ 249, 141, 61 },
		{ 249, 142, 61 },
		{ 249, 143, 61 },
		{ 249, 144, 61 },
		{ 249, 145, 61 },
		{ 249, 146, 61 },
		{ 249, 147, 61 },
		{ 249, 148, 61 },
		{ 249, 149, 61 },
		{ 249, 150, 61 },
		{ 249, 151, 61 },
		{ 249, 152, 61 },
		{ 249, 153, 61 },
		{ 249, 154, 61 },
		{ 249, 155, 61 },
		{ 249, 156, 61 },
		{ 249, 157, 61 },
		{ 249, 158, 61 },
		{ 249, 159, 61 },
		{ 249, 160, 61 },
		{ 249, 161, 61 },
		{ 249, 162, 61 },
		{ 249, 163, 61 },
		{ 249, 164, 61 },
		{ 249, 165, 61 },
		{ 249, 166, 61 },
		{ 249, 167, 61 },
		{ 249, 168, 61 },
		{ 249, 169, 61 },
		{ 249, 170, 61 },
		{ 249, 171, 61 },
		{ 249, 172, 61 },
		{ 249, 173, 61 },
		{ 249, 174, 61 },
		{ 249, 175, 61 },
		{ 249, 176, 61 },
		{ 249, 177, 61 },
		{ 249, 178, 61 },
		{ 249, 179, 61 },
		{ 249, 180, 61 },
		{ 249, 181, 61 },
		{ 249, 182, 61 },
		{ 249, 183, 61 },
		{ 249, 184, 61 },
		{ 249, 185, 61 },
		{ 249, 186, 61 },
		{ 249, 187, 61 },
		{ 249, 188, 61 },
		{ 249, 189, 61 },
		{ 249, 190, 61 },
		{ 249, 191, 61 },
		{ 250, 1, 61 },
		{ 250, 2, 61 },
		{ 250, 3, 61 },
		{ 250, 4, 61 },
		{ 250, 5, 61 },
		{ 250, 6, 61 },
		{ 250, 7, 61 },
		{ 250, 8, 61 },
		{ 250, 9, 61 },
		{ 250, 10, 61 },
		{ 250, 11, 61 },
		{ 250, 12, 61 },
		{ 250, 13, 61 },
		{ 250, 14, 61 },
		{ 250, 15, 61 },
		{ 250, 16, 61 },
		{ 250, 17, 61 },
		{ 250, 18, 61 },
		{ 250, 19, 61 },
		{ 250, 20, 61 },
		{ 250, 21, 61 },
		{ 250, 22, 61 },
		{ 250, 23, 61 },
		{ 250, 24, 61 },
		{ 250, 25, 61 },
		{ 250, 26, 61 },
		{ 250, 27, 61 },
		{ 250, 28, 61 },
		{ 250, 29, 61 },
		{ 250, 30, 61 },
		{ 250, 31, 61 },
		{ 250, 32, 61 },
		{ 250, 33, 61 },
		{ 250, 34, 61 },
		{ 250, 35, 61 },
		{ 250, 36, 61 },
		{ 250, 37, 61 },
		{ 250, 38, 61 },
		{ 250, 39, 61 },
		{ 250, 40, 61 },
		{ 250, 41, 61 },
		{ 250, 42, 61 },
		{ 250, 43, 61 },
		{ 250, 44, 61 },
		{ 250, 45, 61 },
		{ 250, 46, 61 },
		{ 250, 47, 61 },
		{ 250, 48, 61 },
		{ 250, 49, 61 },
		{ 250, 50, 61 },
		{ 250, 51, 61 },
		{ 250, 52, 61 },
		{ 250, 53, 61 },
		{ 250, 54, 61 },
		{ 250, 55, 61 },
		{ 250, 56, 61 },
		{ 250, 57, 61 },
		{ 250, 58, 61 },
		{ 250, 59, 61 },
		{ 250, 60, 61 },
		{ 250, 61, 61 },
		{ 250, 62, 61 },
		{ 250, 63, 61 },
		{ 250, 64, 61 },
		{ 250, 65, 61 },
		{ 250, 66, 61 },
		{ 250, 67, 61 },
		{ 250, 68, 61 },
		{ 250, 69, 61 },
		{ 250, 70, 61 },
		{ 250, 71, 61 },
		{ 250, 72, 61 },
		{ 250, 73, 61 },
		{ 250, 74, 61 },
		{ 250, 75, 61 },
		{ 250, 76, 61 },
		{ 250, 77, 61 },
		{ 250, 78, 61 },
		{ 250, 79, 61 },
		{ 250, 80, 61 },
		{ 250, 81, 61 },
		{ 250, 82, 61 },
		{ 250, 83, 61 },
		{ 250, 84, 61 },
		{ 250, 85, 61 },
		{ 250, 86, 61 },
		{ 250, 87, 61 },
		{ 250, 88, 61 },
		{ 250, 89, 61 },
		{ 250, 90, 61 },
		{ 250, 91, 61 },
		{ 250, 92, 61 },
		{ 250, 93, 61 },
		{ 250, 94, 61 },
		{ 250, 95, 61 },
		{ 250, 96, 61 },
		{ 250, 97, 61 },
		{ 250, 98, 61 },
		{ 250, 99, 61 },
		{ 250, 100, 61 },
		{ 250, 101, 61 },
		{ 250, 102, 61 },
		{ 250, 103, 61 },
		{ 250, 104, 61 },
		{ 250, 105, 61 },
		{ 250, 106, 61 },
		{ 250, 107, 61 },
		{ 250, 108, 61 },
		{ 250, 109, 61 },
		{ 250, 110, 61 },
		{ 250, 111, 61 },
		{ 250, 112, 61 },
		{ 250, 113, 61 },
		{ 250, 114, 61 },
		{ 250, 115, 61 },
		{ 250, 116, 61 },
		{ 250, 117, 61 },
		{ 250, 118, 61 },
		{ 250, 119, 61 },
		{ 250, 120, 61 },
		{ 250, 121, 61 },
		{ 250, 122, 61 },
		{ 250, 123, 61 },
		{ 250, 124, 61 },
		{ 250, 125, 61 },
		{ 250, 126, 61 },
		{ 250, 127, 61 },
		{ 250, 128, 61 },
		{ 250, 129, 61 },
		{ 250, 130, 61 },
		{ 250, 131, 61 },
		{ 250, 132, 61 },
		{ 250, 133, 61 },
		{ 250, 134, 61 },
		{ 250, 135, 61 },
		{ 250, 136, 61 },
		{ 250, 137, 61 },
		{ 250, 138, 61 },
		{ 250, 139, 61 },
		{ 250, 140, 61 },
		{ 250, 141, 61 },
		{ 250, 142, 61 },
		{ 250, 143, 61 },
		{ 250, 144, 61 },
		{ 250, 145, 61 },
		{ 250, 146, 61 },
		{ 250, 147, 61 },
		{ 250, 148, 61 },
		{ 250, 149, 61 },
		{ 250, 150, 61 },
		{ 250, 151, 61 },
		{ 250, 152, 61 },
		{ 250, 153, 61 },
		{ 250, 154, 61 },
		{ 250, 155, 61 },
		{ 250, 156, 61 },
		{ 250, 157, 61 },
		{ 250, 158, 61 },
		{ 250, 159, 61 },
		{ 250, 160, 61 },
		{ 250, 161, 61 },
		{ 250, 162, 61 },
		{ 250, 163, 61 },
		{ 250, 164, 61 },
		{ 250, 165, 61 },
		{ 250, 166, 61 },
		{ 250, 167, 61 },
		{ 250, 168, 61 },
		{ 250, 169, 61 },
		{ 250, 170, 61 },
		{ 250, 171, 61 },
		{ 250, 172, 61 },
		{ 250, 173, 61 },
		{ 250, 174, 61 },
		{ 250, 175, 61 },
		{ 250, 176, 61 },
		{ 250, 177, 61 },
		{ 250, 178, 61 },
		{ 250, 179, 61 },
		{ 250, 180, 61 },
		{ 250, 181, 61 },
		{ 250, 182, 61 },
		{ 250, 183, 61 },
		{ 250, 184, 61 },
		{ 250, 185, 61 },
		{ 250, 186, 61 },
		{ 250, 187, 61 },
		{ 250, 188, 61 },
		{ 250, 189, 61 },
		{ 250, 190, 61 },
		{ 250, 191, 61 },
		{ 251, 1, 61 },
		{ 251, 2, 61 },
		{ 251, 3, 61 },
		{ 251, 4, 61 },
		{ 251, 5, 61 },
		{ 251, 6, 61 },
		{ 251, 7, 61 },
		{ 251, 8, 61 },
		{ 251, 9, 61 },
		{ 251, 10, 61 },
		{ 251, 11, 61 },
		{ 251, 12, 61 },
		{ 251, 13, 61 },
		{ 251, 14, 61 },
		{ 251, 15, 61 },
		{ 251, 16, 61 },
		{ 251, 17, 61 },
		{ 251, 18, 61 },
		{ 251, 19, 61 },
		{ 251, 20, 61 },
		{ 251, 21, 61 },
		{ 251, 22, 61 },
		{ 251, 23, 61 },
		{ 251, 24, 61 },
		{ 251, 25, 61 },
		{ 251, 26, 61 },
		{ 251, 27, 61 },
		{ 251, 28, 61 },
		{ 251, 29, 61 },
		{ 251, 30, 61 },
		{ 251, 31, 61 },
		{ 251, 32, 61 },
		{ 251, 33, 61 },
		{ 251, 34, 61 },
		{ 251, 35, 61 },
		{ 251, 36, 61 },
		{ 251, 37, 61 },
		{ 251, 38, 61 },
		{ 251, 39, 61 },
		{ 251, 40, 61 },
		{ 251, 41, 61 },
		{ 251, 42, 61 },
		{ 251, 43, 61 },
		{ 251, 44, 61 },
		{ 251, 45, 61 },
		{ 251, 46, 61 },
		{ 251, 47, 61 },
		{ 251, 48, 61 },
		{ 251, 49, 61 },
		{ 251, 50, 61 },
		{ 251, 51, 61 },
		{ 251, 52, 61 },
		{ 251, 53, 61 },
		{ 251, 54, 61 },
		{ 251, 55, 61 },
		{ 251, 56, 61 },
		{ 251, 57, 61 },
		{ 251, 58, 61 },
		{ 251, 59, 61 },
		{ 251, 60, 61 },
		{ 251, 61, 61 },
		{ 251, 62, 61 },
		{ 251, 63, 61 },
		{ 251, 64, 61 },
		{ 251, 65, 61 },
		{ 251, 66, 61 },
		{ 251, 67, 61 },
		{ 251, 68, 61 },
		{ 251, 69, 61 },
		{ 251, 70, 61 },
		{ 251, 71, 61 },
		{ 251, 72, 61 },
		{ 251, 73, 61 },
		{ 251, 74, 61 },
		{ 251, 75, 61 },
		{ 251, 76, 61 },
		{ 251, 77, 61 },
		{ 251, 78, 61 },
		{ 251, 79, 61 },
		{ 251, 80, 61 },
		{ 251, 81, 61 },
		{ 251, 82, 61 },
		{ 251, 83, 61 },
		{ 251, 84, 61 },
		{ 251, 85, 61 },
		{ 251, 86, 61 },
		{ 251, 87, 61 },
		{ 251, 88, 61 },
		{ 251, 89, 61 },
		{ 251, 90, 61 },
		{ 251, 91, 61 },
		{ 251, 92, 61 },
		{ 251, 93, 61 },
		{ 251, 94, 61 },
		{ 251, 95, 61 },
		{ 251, 96, 61 },
		{ 251, 97, 61 },
		{ 251, 98, 61 },
		{ 251, 99, 61 },
		{ 251, 100, 61 },
		{ 251, 101, 61 },
		{ 251, 102, 61 },
		{ 251, 103, 61 },
		{ 251, 104, 61 },
		{ 251, 105, 61 },
		{ 251, 106, 61 },
		{ 251, 107, 61 },
		{ 251, 108, 61 },
		{ 251, 109, 61 },
		{ 251, 110, 61 },
		{ 251, 111, 61 },
		{ 251, 112, 61 },
		{ 251, 113, 61 },
		{ 251, 114, 61 },
		{ 251, 115, 61 },
		{ 251, 116, 61 },
		{ 251, 117, 61 },
		{ 251, 118, 61 },
		{ 251, 119, 61 },
		{ 251, 120, 61 },
		{ 251, 121, 61 },
		{ 251, 122, 61 },
		{ 251, 123, 61 },
		{ 251, 124, 61 },
		{ 251, 125, 61 },
		{ 251, 126, 61 },
		{ 251, 127, 61 },
		{ 251, 128, 61 },
		{ 251, 129, 61 },
		{ 251, 130, 61 },
		{ 251, 131, 61 },
		{ 251, 132, 61 },
		{ 251, 133, 61 },
		{ 251, 134, 61 },
		{ 251, 135, 61 },
		{ 251, 136, 61 },
		{ 251, 137, 61 },
		{ 251, 138, 61 },
		{ 251, 139, 61 },
		{ 251, 140, 61 },
		{ 251, 141, 61 },
		{ 251, 142, 61 },
		{ 251, 143, 61 },
		{ 251, 144, 61 },
		{ 251, 145, 61 },
		{ 251, 146, 61 },
		{ 251, 147, 61 },
		{ 251, 148, 61 },
		{ 251, 149, 61 },
		{ 251, 150, 61 },
		{ 251, 151, 61 },
		{ 251, 152, 61 },
		{ 251, 153, 61 },
		{ 251, 154, 61 },
		{ 251, 155, 61 },
		{ 251, 156, 61 },
		{ 251, 157, 61 },
		{ 251, 158, 61 },
		{ 251, 159, 61 },
		{ 251, 160, 61 },
		{ 251, 161, 61 },
		{ 251, 162, 61 },
		{ 251, 163, 61 },
		{ 251, 164, 61 },
		{ 251, 165, 61 },
		{ 251, 166, 61 },
		{ 251, 167, 61 },
		{ 251, 168, 61 },
		{ 251, 169, 61 },
		{ 251, 170, 61 },
		{ 251, 171, 61 },
		{ 251, 172, 61 },
		{ 251, 173, 61 },
		{ 251, 174, 61 },
		{ 251, 175, 61 },
		{ 251, 176, 61 },
		{ 251, 177, 61 },
		{ 251, 178, 61 },
		{ 251, 179, 61 },
		{ 251, 180, 61 },
		{ 251, 181, 61 },
		{ 251, 182, 61 },
		{ 251, 183, 61 },
		{ 251, 184, 61 },
		{ 251, 185, 61 },
		{ 251, 186, 61 },
		{ 251, 187, 61 },
		{ 251, 188, 61 },
		{ 251, 189, 61 },
		{ 251, 190, 61 },
		{ 251, 191, 61 },
		{ 252, 1, 61 },
		{ 252, 2, 61 },
		{ 252, 3, 61 },
		{ 252, 4, 61 },
		{ 252, 5, 61 },
		{ 252, 6, 61 },
		{ 252, 7, 61 },
		{ 252, 8, 61 },
		{ 252, 9, 61 },
		{ 252, 10, 61 },
		{ 252, 11, 61 },
		{ 252, 12, 61 },
		{ 252, 13, 61 },
		{ 252, 14, 61 },
		{ 252, 15, 61 },
		{ 252, 16, 61 },
		{ 252, 17, 61 },
		{ 252, 18, 61 },
		{ 252, 19, 61 },
		{ 252, 20, 61 },
		{ 252, 21, 61 },
		{ 252, 22, 61 },
		{ 252, 23, 61 },
		{ 252, 24, 61 },
		{ 252, 25, 61 },
		{ 252, 26, 61 },
		{ 252, 27, 61 },
		{ 252, 28, 61 },
		{ 252, 29, 61 },
		{ 252, 30, 61 },
		{ 252, 31, 61 },
		{ 252, 32, 61 },
		{ 252, 33, 61 },
		{ 252, 34, 61 },
		{ 252, 35, 61 },
		{ 252, 36, 61 },
		{ 252, 37, 61 },
		{ 252, 38, 61 },
		{ 252, 39, 61 },
		{ 252, 40, 61 },
		{ 252, 41, 61 },
		{ 252, 42, 61 },
		{ 252, 43, 61 },
		{ 252, 44, 61 },
		{ 252, 45, 61 },
		{ 252, 46, 61 },
		{ 252, 47, 61 },
		{ 252, 48, 61 },
		{ 252, 49, 61 },
		{ 252, 50, 61 },
		{ 252, 51, 61 },
		{ 252, 52, 61 },
		{ 252, 53, 61 },
		{ 252, 54, 61 },
		{ 252, 55, 61 },
		{ 252, 56, 61 },
		{ 252, 57, 61 },
		{ 252, 58, 61 },
		{ 252, 59, 61 },
		{ 252, 60, 61 },
		{ 252, 61, 61 },
		{ 252, 62, 61 },
		{ 252, 63, 61 },
		{ 252, 64, 61 },
		{ 252, 65, 61 },
		{ 252, 66, 61 },
		{ 252, 67, 61 },
		{ 252, 68, 61 },
		{ 252, 69, 61 },
		{ 252, 70, 61 },
		{ 252, 71, 61 },
		{ 252, 72, 61 },
		{ 252, 73, 61 },
		{ 252, 74, 61 },
		{ 252, 75, 61 },
		{ 252, 76, 61 },
		{ 252, 77, 61 },
		{ 252, 78, 61 },
		{ 252, 79, 61 },
		{ 252, 80, 61 },
		{ 252, 81, 61 },
		{ 252, 82, 61 },
		{ 252, 83, 61 },
		{ 252, 84, 61 },
		{ 252, 85, 61 },
		{ 252, 86, 61 },
		{ 252, 87, 61 },
		{ 252, 88, 61 },
		{ 252, 89, 61 },
		{ 252, 90, 61 },
		{ 252, 91, 61 },
		{ 252, 92, 61 },
		{ 252, 93, 61 },
		{ 252, 94, 61 },
		{ 252, 95, 61 },
		{ 252, 96, 61 },
		{ 252, 97, 61 },
		{ 252, 98, 61 },
		{ 252, 99, 61 },
		{ 252, 100, 61 },
		{ 252, 101, 61 },
		{ 252, 102, 61 },
		{ 252, 103, 61 },
		{ 252, 104, 61 },
		{ 252, 105, 61 },
		{ 252, 106, 61 },
		{ 252, 107, 61 },
		{ 252, 108, 61 },
		{ 252, 109, 61 },
		{ 252, 110, 61 },
		{ 252, 111, 61 },
		{ 252, 112, 61 },
		{ 252, 113, 61 },
		{ 252, 114, 61 },
		{ 252, 115, 61 },
		{ 252, 116, 61 },
		{ 252, 117, 61 },
		{ 252, 118, 61 },
		{ 252, 119, 61 },
		{ 252, 120, 61 },
		{ 252, 121, 61 },
		{ 252, 122, 61 },
		{ 252, 123, 61 },
		{ 252, 124, 61 },
		{ 252, 125, 61 },
		{ 252, 126, 61 },
		{ 252, 127, 61 },
		{ 252, 128, 61 },
		{ 252, 129, 61 },
		{ 252, 130, 61 },
		{ 252, 131, 61 },
		{ 252, 132, 61 },
		{ 252, 133, 61 },
		{ 252, 134, 61 },
		{ 252, 135, 61 },
		{ 252, 136, 61 },
		{ 252, 137, 61 },
		{ 252, 138, 61 },
		{ 252, 139, 61 },
		{ 252, 140, 61 },
		{ 252, 141, 61 },
		{ 252, 142, 61 },
		{ 252, 143, 61 },
		{ 252, 144, 61 },
		{ 252, 145, 61 },
		{ 252, 146, 61 },
		{ 252, 147, 61 },
		{ 252, 148, 61 },
		{ 252, 149, 61 },
		{ 252, 150, 61 },
		{ 252, 151, 61 },
		{ 252, 152, 61 },
		{ 252, 153, 61 },
		{ 252, 154, 61 },
		{ 252, 155, 61 },
		{ 252, 156, 61 },
		{ 252, 157, 61 },
		{ 252, 158, 61 },
		{ 252, 159, 61 },
		{ 252, 160, 61 },
		{ 252, 161, 61 },
		{ 252, 162, 61 },
		{ 252, 163, 61 },
		{ 252, 164, 61 },
		{ 252, 165, 61 },
		{ 252, 166, 61 },
		{ 252, 167, 61 },
		{ 252, 168, 61 },
		{ 252, 169, 61 },
		{ 252, 170, 61 },
		{ 252, 171, 61 },
		{ 252, 172, 61 },
		{ 252, 173, 61 },
		{ 252, 174, 61 },
		{ 252, 175, 61 },
		{ 252, 176, 61 },
		{ 252, 177, 61 },
		{ 252, 178, 61 },
		{ 252, 179, 61 },
		{ 252, 180, 61 },
		{ 252, 181, 61 },
		{ 252, 182, 61 },
		{ 252, 183, 61 },
		{ 252, 184, 61 },
		{ 252, 185, 61 },
		{ 252, 186, 61 },
		{ 252, 187, 61 },
		{ 252, 188, 61 },
		{ 252, 189, 61 },
		{ 252, 190, 61 },
		{ 252, 191, 61 },
		{ 253, 1, 61 },
		{ 253, 2, 61 },
		{ 253, 3, 61 },
		{ 253, 4, 61 },
		{ 253, 5, 61 },
		{ 253, 6, 61 },
		{ 253, 7, 61 },
		{ 253, 8, 61 },
		{ 253, 9, 61 },
		{ 253, 10, 61 },
		{ 253, 11, 61 },
		{ 253, 12, 61 },
		{ 253, 13, 61 },
		{ 253, 14, 61 },
		{ 253, 15, 61 },
		{ 253, 16, 61 },
		{ 253, 17, 61 },
		{ 253, 18, 61 },
		{ 253, 19, 61 },
		{ 253, 20, 61 },
		{ 253, 21, 61 },
		{ 253, 22, 61 },
		{ 253, 23, 61 },
		{ 253, 24, 61 },
		{ 253, 25, 61 },
		{ 253, 26, 61 },
		{ 253, 27, 61 },
		{ 253, 28, 61 },
		{ 253, 29, 61 },
		{ 253, 30, 61 },
		{ 253, 31, 61 },
		{ 253, 32, 61 },
		{ 253, 33, 61 },
		{ 253, 34, 61 },
		{ 253, 35, 61 },
		{ 253, 36, 61 },
		{ 253, 37, 61 },
		{ 253, 38, 61 },
		{ 253, 39, 61 },
		{ 253, 40, 61 },
		{ 253, 41, 61 },
		{ 253, 42, 61 },
		{ 253, 43, 61 },
		{ 253, 44, 61 },
		{ 253, 45, 61 },
		{ 253, 46, 61 },
		{ 253, 47, 61 },
		{ 253, 48, 61 },
		{ 253, 49, 61 },
		{ 253, 50, 61 },
		{ 253, 51, 61 },
		{ 253, 52, 61 },
		{ 253, 53, 61 },
		{ 253, 54, 61 },
		{ 253, 55, 61 },
		{ 253, 56, 61 },
		{ 253, 57, 61 },
		{ 253, 58, 61 },
		{ 253, 59, 61 },
		{ 253, 60, 61 },
		{ 253, 61, 61 },
		{ 253, 62, 61 },
		{ 253, 63, 61 },
		{ 253, 64, 61 },
		{ 253, 65, 61 },
		{ 253, 66, 61 },
		{ 253, 67, 61 },
		{ 253, 68, 61 },
		{ 253, 69, 61 },
		{ 253, 70, 61 },
		{ 253, 71, 61 },
		{ 253, 72, 61 },
		{ 253, 73, 61 },
		{ 253, 74, 61 },
		{ 253, 75, 61 },
		{ 253, 76, 61 },
		{ 253, 77, 61 },
		{ 253, 78, 61 },
		{ 253, 79, 61 },
		{ 253, 80, 61 },
		{ 253, 81, 61 },
		{ 253, 82, 61 },
		{ 253, 83, 61 },
		{ 253, 84, 61 },
		{ 253, 85, 61 },
		{ 253, 86, 61 },
		{ 253, 87, 61 },
		{ 253, 88, 61 },
		{ 253, 89, 61 },
		{ 253, 90, 61 },
		{ 253, 91, 61 },
		{ 253, 92, 61 },
		{ 253, 93, 61 },
		{ 253, 94, 61 },
		{ 253, 95, 61 },
		{ 253, 96, 61 },
		{ 253, 97, 61 },
		{ 253, 98, 61 },
		{ 253, 99, 61 },
		{ 253, 100, 61 },
		{ 253, 101, 61 },
		{ 253, 102, 61 },
		{ 253, 103, 61 },
		{ 253, 104, 61 },
		{ 253, 105, 61 },
		{ 253, 106, 61 },
		{ 253, 107, 61 },
		{ 253, 108, 61 },
		{ 253, 109, 61 },
		{ 253, 110, 61 },
		{ 253, 111, 61 },
		{ 253, 112, 61 },
		{ 253, 113, 61 },
		{ 253, 114, 61 },
		{ 253, 115, 61 },
		{ 253, 116, 61 },
		{ 253, 117, 61 },
		{ 253, 118, 61 },
		{ 253, 119, 61 },
		{ 253, 120, 61 },
		{ 253, 121, 61 },
		{ 253, 122, 61 },
		{ 253, 123, 61 },
		{ 253, 124, 61 },
		{ 253, 125, 61 },
		{ 253, 126, 61 },
		{ 253, 127, 61 },
		{ 253, 128, 61 },
		{ 253, 129, 61 },
		{ 253, 130, 61 },
		{ 253, 131, 61 },
		{ 253, 132, 61 },
		{ 253, 133, 61 },
		{ 253, 134, 61 },
		{ 253, 135, 61 },
		{ 253, 136, 61 },
		{ 253, 137, 61 },
		{ 253, 138, 61 },
		{ 253, 139, 61 },
		{ 253, 140, 61 },
		{ 253, 141, 61 },
		{ 253, 142, 61 },
		{ 253, 143, 61 },
		{ 253, 144, 61 },
		{ 253, 145, 61 },
		{ 253, 146, 61 },
		{ 253, 147, 61 },
		{ 253, 148, 61 },
		{ 253, 149, 61 },
		{ 253, 150, 61 },
		{ 253, 151, 61 },
		{ 253, 152, 61 },
		{ 253, 153, 61 },
		{ 253, 154, 61 },
		{ 253, 155, 61 },
		{ 253, 156, 61 },
		{ 253, 157, 61 },
		{ 253, 158, 61 },
		{ 253, 159, 61 },
		{ 253, 160, 61 },
		{ 253, 161, 61 },
		{ 253, 162, 61 },
		{ 253, 163, 61 },
		{ 253, 164, 61 },
		{ 253, 165, 61 },
		{ 253, 166, 61 },
		{ 253, 167, 61 },
		{ 253, 168, 61 },
		{ 253, 169, 61 },
		{ 253, 170, 61 },
		{ 253, 171, 61 },
		{ 253, 172, 61 },
		{ 253, 173, 61 },
		{ 253, 174, 61 },
		{ 253, 175, 61 },
		{ 253, 176, 61 },
		{ 253, 177, 61 },
		{ 253, 178, 61 },
		{ 253, 179, 61 },
		{ 253, 180, 61 },
		{ 253, 181, 61 },
		{ 253, 182, 61 },
		{ 253, 183, 61 },
		{ 253, 184, 61 },
		{ 253, 185, 61 },
		{ 253, 186, 61 },
		{ 253, 187, 61 },
		{ 253, 188, 61 },
		{ 253, 189, 61 },
		{ 253, 190, 61 },
		{ 253, 191, 61 },
		{ 254, 1, 61 },
		{ 254, 2, 61 },
		{ 254, 3, 61 },
		{ 254, 4, 61 },
		{ 254, 5, 61 },
		{ 254, 6, 61 },
		{ 254, 7, 61 },
		{ 254, 8, 61 },
		{ 254, 9, 61 },
		{ 254, 10, 61 },
		{ 254, 11, 61 },
		{ 254, 12, 61 },
		{ 254, 13, 61 },
		{ 254, 14, 61 },
		{ 254, 15, 61 },
		{ 254, 16, 61 },
		{ 254, 17, 61 },
		{ 254, 18, 61 },
		{ 254, 19, 61 },
		{ 254, 20, 61 },
		{ 254, 21, 61 },
		{ 254, 22, 61 },
		{ 254, 23, 61 },
		{ 254, 24, 61 },
		{ 254, 25, 61 },
		{ 254, 26, 61 },
		{ 254, 27, 61 },
		{ 254, 28, 61 },
		{ 254, 29, 61 },
		{ 254, 30, 61 },
		{ 254, 31, 61 },
		{ 254, 32, 61 },
		{ 254, 33, 61 },
		{ 254, 34, 61 },
		{ 254, 35, 61 },
		{ 254, 36, 61 },
		{ 254, 37, 61 },
		{ 254, 38, 61 },
		{ 254, 39, 61 },
		{ 254, 40, 61 },
		{ 254, 41, 61 },
		{ 254, 42, 61 },
		{ 254, 43, 61 },
		{ 254, 44, 61 },
		{ 254, 45, 61 },
		{ 254, 46, 61 },
		{ 254, 47, 61 },
		{ 254, 48, 61 },
		{ 254, 49, 61 },
		{ 254, 50, 61 },
		{ 254, 51, 61 },
		{ 254, 52, 61 },
		{ 254, 53, 61 },
		{ 254, 54, 61 },
		{ 254, 55, 61 },
		{ 254, 56, 61 },
		{ 254, 57, 61 },
		{ 254, 58, 61 },
		{ 254, 59, 61 },
		{ 254, 60, 61 },
		{ 254, 61, 61 },
		{ 254, 62, 61 },
		{ 254, 63, 61 },
		{ 254, 64, 61 },
		{ 254, 65, 61 },
		{ 254, 66, 61 },
		{ 254, 67, 61 },
		{ 254, 68, 61 },
		{ 254, 69, 61 },
		{ 254, 70, 61 },
		{ 254, 71, 61 },
		{ 254, 72, 61 },
		{ 254, 73, 61 },
		{ 254, 74, 61 },
		{ 254, 75, 61 },
		{ 254, 76, 61 },
		{ 254, 77, 61 },
		{ 254, 78, 61 },
		{ 254, 79, 61 },
		{ 254, 80, 61 },
		{ 254, 81, 61 },
		{ 254, 82, 61 },
		{ 254, 83, 61 },
		{ 254, 84, 61 },
		{ 254, 85, 61 },
		{ 254, 86, 61 },
		{ 254, 87, 61 },
		{ 254, 88, 61 },
		{ 254, 89, 61 },
		{ 254, 90, 61 },
		{ 254, 91, 61 },
		{ 254, 92, 61 },
		{ 254, 93, 61 },
		{ 254, 94, 61 },
		{ 254, 95, 61 },
		{ 254, 96, 61 },
		{ 254, 97, 61 },
		{ 254, 98, 61 },
		{ 254, 99, 61 },
		{ 254, 100, 61 },
		{ 254, 101, 61 },
		{ 254, 102, 61 },
		{ 254, 103, 61 },
		{ 254, 104, 61 },
		{ 254, 105, 61 },
		{ 254, 106, 61 },
		{ 254, 107, 61 },
		{ 254, 108, 61 },
		{ 254, 109, 61 },
		{ 254, 110, 61 },
		{ 254, 111, 61 },
		{ 254, 112, 61 },
		{ 254, 113, 61 },
		{ 254, 114, 61 },
		{ 254, 115, 61 },
		{ 254, 116, 61 },
		{ 254, 117, 61 },
		{ 254, 118, 61 },
		{ 254, 119, 61 },
		{ 254, 120, 61 },
		{ 254, 121, 61 },
		{ 254, 122, 61 },
		{ 254, 123, 61 },
		{ 254, 124, 61 },
		{ 254, 125, 61 },
		{ 254, 126, 61 },
		{ 254, 127, 61 },
		{ 254, 128, 61 },
		{ 254, 129, 61 },
		{ 254, 130, 61 },
		{ 254, 131, 61 },
		{ 254, 132, 61 },
		{ 254, 133, 61 },
		{ 254, 134, 61 },
		{ 254, 135, 61 },
		{ 254, 136, 61 },
		{ 254, 137, 61 },
		{ 254, 138, 61 },
		{ 254, 139, 61 },
		{ 254, 140, 61 },
		{ 254, 141, 61 },
		{ 254, 142, 61 },
		{ 254, 143, 61 },
		{ 254, 144, 61 },
		{ 254, 145, 61 },
		{ 254, 146, 61 },
		{ 254, 147, 61 },
		{ 254, 148, 61 },
		{ 254, 149, 61 },
		{ 254, 150, 61 },
		{ 254, 151, 61 },
		{ 254, 152, 61 },
		{ 254, 153, 61 },
		{ 254, 154, 61 },
		{ 254, 155, 61 },
		{ 254, 156, 61 },
		{ 254, 157, 61 },
		{ 254, 158, 61 },
		{ 254, 159, 61 },
		{ 254, 160, 61 },
		{ 254, 161, 61 },
		{ 254, 162, 61 },
		{ 254, 163, 61 },
		{ 254, 164, 61 },
		{ 254, 165, 61 },
		{ 254, 166, 61 },
		{ 254, 167, 61 },
		{ 254, 168, 61 },
		{ 254, 169, 61 },
		{ 254, 170, 61 },
		{ 254, 171, 61 },
		{ 254, 172, 61 },
		{ 254, 173, 61 },
		{ 254, 174, 61 },
		{ 254, 175, 61 },
		{ 254, 176, 61 },
		{ 254, 177, 61 },
		{ 254, 178, 61 },
		{ 254, 179, 61 },
		{ 254, 180, 61 },
		{ 254, 181, 61 },
		{ 254, 182, 61 },
		{ 254, 183, 61 },
		{ 254, 184, 61 },
		{ 254, 185, 61 },
		{ 254, 186, 61 },
		{ 254, 187, 61 },
		{ 254, 188, 61 },
		{ 254, 189, 61 },
		{ 254, 190, 61 },
		{ 254, 191, 61 },
		{ 255, 1, 61 },
		{ 255, 2, 61 },
		{ 255, 3, 61 },
		{ 255, 4, 61 },
		{ 255, 5, 61 },
		{ 255, 6, 61 },
		{ 255, 7, 61 },
		{ 255, 8, 61 },
		{ 255, 9, 61 },
		{ 255, 10, 61 },
		{ 255, 11, 61 },
		{ 255, 12, 61 },
		{ 255, 13, 61 },
		{ 255, 14, 61 },
		{ 255, 15, 61 },
		{ 255, 16, 61 },
		{ 255, 17, 61 },
		{ 255, 18, 61 },
		{ 255, 19, 61 },
		{ 255, 20, 61 },
		{ 255, 21, 61 },
		{ 255, 22, 61 },
		{ 255, 23, 61 },
		{ 255, 24, 61 },
		{ 255, 25, 61 },
		{ 255, 26, 61 },
		{ 255, 27, 61 },
		{ 255, 28, 61 },
		{ 255, 29, 61 },
		{ 255, 30, 61 },
		{ 255, 31, 61 },
		{ 255, 32, 61 },
		{ 255, 33, 61 },
		{ 255, 34, 61 },
		{ 255, 35, 61 },
		{ 255, 36, 61 },
		{ 255, 37, 61 },
		{ 255, 38, 61 },
		{ 255, 39, 61 },
		{ 255, 40, 61 },
		{ 255, 41, 61 },
		{ 255, 42, 61 },
		{ 255, 43, 61 },
		{ 255, 44, 61 },
		{ 255, 45, 61 },
		{ 255, 46, 61 },
		{ 255, 47, 61 },
		{ 255, 48, 61 },
		{ 255, 49, 61 },
		{ 255, 50, 61 },
		{ 255, 51, 61 },
		{ 255, 52, 61 },
		{ 255, 53, 61 },
		{ 255, 54, 61 },
		{ 255, 55, 61 },
		{ 255, 56, 61 },
		{ 255, 57, 61 },
		{ 255, 58, 61 },
		{ 255, 59, 61 },
		{ 255, 60, 61 },
		{ 255, 61, 61 },
		{ 255, 62, 61 },
		{ 255, 63, 61 },
		{ 255, 64, 61 },
		{ 255, 65, 61 },
		{ 255, 66, 61 },
		{ 255, 67, 61 },
		{ 255, 68, 61 },
		{ 255, 69, 61 },
		{ 255, 70, 61 },
		{ 255, 71, 61 },
		{ 255, 72, 61 },
		{ 255, 73, 61 },
		{ 255, 74, 61 },
		{ 255, 75, 61 },
		{ 255, 76, 61 },
		{ 255, 77, 61 },
		{ 255, 78, 61 },
		{ 255, 79, 61 },
		{ 255, 80, 61 },
		{ 255, 81, 61 },
		{ 255, 82, 61 },
		{ 255, 83, 61 },
		{ 255, 84, 61 },
		{ 255, 85, 61 },
		{ 255, 86, 61 },
		{ 255, 87, 61 },
		{ 255, 88, 61 },
		{ 255, 89, 61 },
		{ 255, 90, 61 },
		{ 255, 91, 61 },
		{ 255, 92, 61 },
		{ 255, 93, 61 },
		{ 255, 94, 61 },
		{ 255, 95, 61 },
		{ 255, 96, 61 },
		{ 255, 97, 61 },
		{ 255, 98, 61 },
		{ 255, 99, 61 },
		{ 255, 100, 61 },
		{ 255, 101, 61 },
		{ 255, 102, 61 },
		{ 255, 103, 61 },
		{ 255, 104, 61 },
		{ 255, 105, 61 },
		{ 255, 106, 61 },
		{ 255, 107, 61 },
		{ 255, 108, 61 },
		{ 255, 109, 61 },
		{ 255, 110, 61 },
		{ 255, 111, 61 },
		{ 255, 112, 61 },
		{ 255, 113, 61 },
		{ 255, 114, 61 },
		{ 255, 115, 61 },
		{ 255, 116, 61 },
		{ 255, 117, 61 },
		{ 255, 118, 61 },
		{ 255, 119, 61 },
		{ 255, 120, 61 },
		{ 255, 121, 61 },
		{ 255, 122, 61 },
		{ 255, 123, 61 },
		{ 255, 124, 61 },
		{ 255, 125, 61 },
		{ 255, 126, 61 },
		{ 255, 127, 61 },
		{ 255, 128, 61 },
		{ 255, 129, 61 },
		{ 255, 130, 61 },
		{ 255, 131, 61 },
		{ 255, 132, 61 },
		{ 255, 133, 61 },
		{ 255, 134, 61 },
		{ 255, 135, 61 },
		{ 255, 136, 61 },
		{ 255, 137, 61 },
		{ 255, 138, 61 },
		{ 255, 139, 61 },
		{ 255, 140, 61 },
		{ 255, 141, 61 },
		{ 255, 142, 61 },
		{ 255, 143, 61 },
		{ 255, 144, 61 },
		{ 255, 145, 61 },
		{ 255, 146, 61 },
		{ 255, 147, 61 },
		{ 255, 148, 61 },
		{ 255, 149, 61 },
		{ 255, 150, 61 },
		{ 255, 151, 61 },
		{ 255, 152, 61 },
		{ 255, 153, 61 },
		{ 255, 154, 61 },
		{ 255, 155, 61 },
		{ 255, 156, 61 },
		{ 255, 157, 61 },
		{ 255, 158, 61 },
		{ 255, 159, 61 },
		{ 255, 160, 61 },
		{ 255, 161, 61 },
		{ 255, 162, 61 },
		{ 255, 163, 61 },
		{ 255, 164, 61 },
		{ 255, 165, 61 },
		{ 255, 166, 61 },
		{ 255, 167, 61 },
		{ 255, 168, 61 },
		{ 255, 169, 61 },
		{ 255, 170, 61 },
		{ 255, 171, 61 },
		{ 255, 172, 61 },
		{ 255, 173, 61 },
		{ 255, 174, 61 },
		{ 255, 175, 61 },
		{ 255, 176, 61 },
		{ 255, 177, 61 },
		{ 255, 178, 61 },
		{ 255, 179, 61 },
		{ 255, 180, 61 },
		{ 255, 181, 61 },
		{ 255, 182, 61 },
		{ 255, 183, 61 },
		{ 255, 184, 61 },
		{ 255, 185, 61 },
		{ 255, 186, 61 },
		{ 255, 187, 61 },
		{ 255, 188, 61 },
		{ 255, 189, 61 },
		{ 255, 190, 61 },
		{ 255, 191, 61 },
		{ 256, 1, 61 },
		{ 256, 2, 61 },
		{ 256, 3, 61 },
		{ 256, 4, 61 },
		{ 256, 5, 61 },
		{ 256, 6, 61 },
		{ 256, 7, 61 },
		{ 256, 8, 61 },
		{ 256, 9, 61 },
		{ 256, 10, 61 },
		{ 256, 11, 61 },
		{ 256, 12, 61 },
		{ 256, 13, 61 },
		{ 256, 14, 61 },
		{ 256, 15, 61 },
		{ 256, 16, 61 },
		{ 256, 17, 61 },
		{ 256, 18, 61 },
		{ 256, 19, 61 },
		{ 256, 20, 61 },
		{ 256, 21, 61 },
		{ 256, 22, 61 },
		{ 256, 23, 61 },
		{ 256, 24, 61 },
		{ 256, 25, 61 },
		{ 256, 26, 61 },
		{ 256, 27, 61 },
		{ 256, 28, 61 },
		{ 256, 29, 61 },
		{ 256, 30, 61 },
		{ 256, 31, 61 },
		{ 256, 32, 61 },
		{ 256, 33, 61 },
		{ 256, 34, 61 },
		{ 256, 35, 61 },
		{ 256, 36, 61 },
		{ 256, 37, 61 },
		{ 256, 38, 61 },
		{ 256, 39, 61 },
		{ 256, 40, 61 },
		{ 256, 41, 61 },
		{ 256, 42, 61 },
		{ 256, 43, 61 },
		{ 256, 44, 61 },
		{ 256, 45, 61 },
		{ 256, 46, 61 },
		{ 256, 47, 61 },
		{ 256, 48, 61 },
		{ 256, 49, 61 },
		{ 256, 50, 61 },
		{ 256, 51, 61 },
		{ 256, 52, 61 },
		{ 256, 53, 61 },
		{ 256, 54, 61 },
		{ 256, 55, 61 },
		{ 256, 56, 61 },
		{ 256, 57, 61 },
		{ 256, 58, 61 },
		{ 256, 59, 61 },
		{ 256, 60, 61 },
		{ 256, 61, 61 },
		{ 256, 62, 61 },
		{ 256, 63, 61 },
		{ 256, 64, 61 },
		{ 256, 65, 61 },
		{ 256, 66, 61 },
		{ 256, 67, 61 },
		{ 256, 68, 61 },
		{ 256, 69, 61 },
		{ 256, 70, 61 },
		{ 256, 71, 61 },
		{ 256, 72, 61 },
		{ 256, 73, 61 },
		{ 256, 74, 61 },
		{ 256, 75, 61 },
		{ 256, 76, 61 },
		{ 256, 77, 61 },
		{ 256, 78, 61 },
		{ 256, 79, 61 },
		{ 256, 80, 61 },
		{ 256, 81, 61 },
		{ 256, 82, 61 },
		{ 256, 83, 61 },
		{ 256, 84, 61 },
		{ 256, 85, 61 },
		{ 256, 86, 61 },
		{ 256, 87, 61 },
		{ 256, 88, 61 },
		{ 256, 89, 61 },
		{ 256, 90, 61 },
		{ 256, 91, 61 },
		{ 256, 92, 61 },
		{ 256, 93, 61 },
		{ 256, 94, 61 },
		{ 256, 95, 61 },
		{ 256, 96, 61 },
		{ 256, 97, 61 },
		{ 256, 98, 61 },
		{ 256, 99, 61 },
		{ 256, 100, 61 },
		{ 256, 101, 61 },
		{ 256, 102, 61 },
		{ 256, 103, 61 },
		{ 256, 104, 61 },
		{ 256, 105, 61 },
		{ 256, 106, 61 },
		{ 256, 107, 61 },
		{ 256, 108, 61 },
		{ 256, 109, 61 },
		{ 256, 110, 61 },
		{ 256, 111, 61 },
		{ 256, 112, 61 },
		{ 256, 113, 61 },
		{ 256, 114, 61 },
		{ 256, 115, 61 },
		{ 256, 116, 61 },
		{ 256, 117, 61 },
		{ 256, 118, 61 },
		{ 256, 119, 61 },
		{ 256, 120, 61 },
		{ 256, 121, 61 },
		{ 256, 122, 61 },
		{ 256, 123, 61 },
		{ 256, 124, 61 },
		{ 256, 125, 61 },
		{ 256, 126, 61 },
		{ 256, 127, 61 },
		{ 256, 128, 61 },
		{ 256, 129, 61 },
		{ 256, 130, 61 },
		{ 256, 131, 61 },
		{ 256, 132, 61 },
		{ 256, 133, 61 },
		{ 256, 134, 61 },
		{ 256, 135, 61 },
		{ 256, 136, 61 },
		{ 256, 137, 61 },
		{ 256, 138, 61 },
		{ 256, 139, 61 },
		{ 256, 140, 61 },
		{ 256, 141, 61 },
		{ 256, 142, 61 },
		{ 256, 143, 61 },
		{ 256, 144, 61 },
		{ 256, 145, 61 },
		{ 256, 146, 61 },
		{ 256, 147, 61 },
		{ 256, 148, 61 },
		{ 256, 149, 61 },
		{ 256, 150, 61 },
		{ 256, 151, 61 },
		{ 256, 152, 61 },
		{ 256, 153, 61 },
		{ 256, 154, 61 },
		{ 256, 155, 61 },
		{ 256, 156, 61 },
		{ 256, 157, 61 },
		{ 256, 158, 61 },
		{ 256, 159, 61 },
		{ 256, 160, 61 },
		{ 256, 161, 61 },
		{ 256, 162, 61 },
		{ 256, 163, 61 },
		{ 256, 164, 61 },
		{ 256, 165, 61 },
		{ 256, 166, 61 },
		{ 256, 167, 61 },
		{ 256, 168, 61 },
		{ 256, 169, 61 },
		{ 256, 170, 61 },
		{ 256, 171, 61 },
		{ 256, 172, 61 },
		{ 256, 173, 61 },
		{ 256, 174, 61 },
		{ 256, 175, 61 },
		{ 256, 176, 61 },
		{ 256, 177, 61 },
		{ 256, 178, 61 },
		{ 256, 179, 61 },
		{ 256, 180, 61 },
		{ 256, 181, 61 },
		{ 256, 182, 61 },
		{ 256, 183, 61 },
		{ 256, 184, 61 },
		{ 256, 185, 61 },
		{ 256, 186, 61 },
		{ 256, 187, 61 },
		{ 256, 188, 61 },
		{ 256, 189, 61 },
		{ 256, 190, 61 },
		{ 258, 324, 61 },
		{ 258, 325, 61 },
		{ 258, 326, 61 },
		{ 258, 327, 61 },
		{ 258, 328, 61 },
		{ 258, 329, 61 },
		{ 258, 330, 61 },
		{ 258, 331, 61 },
		{ 258, 332, 61 },
		{ 258, 333, 61 },
		{ 258, 334, 61 },
		{ 258, 335, 61 },
		{ 258, 336, 61 },
		{ 258, 337, 61 },
		{ 258, 338, 61 },
		{ 258, 339, 61 },
		{ 258, 340, 61 },
		{ 258, 341, 61 },
		{ 258, 342, 61 },
		{ 258, 343, 61 },
		{ 258, 344, 61 },
		{ 258, 345, 61 },
		{ 258, 346, 61 },
		{ 258, 347, 61 },
		{ 258, 348, 61 },
		{ 258, 349, 61 },
		{ 258, 350, 61 },
		{ 258, 351, 61 },
		{ 258, 352, 61 },
		{ 258, 353, 61 },
		{ 258, 354, 61 },
		{ 258, 355, 61 },
		{ 258, 356, 61 },
		{ 258, 357, 61 },
		{ 258, 358, 61 },
		{ 258, 359, 61 },
		{ 258, 360, 61 },
		{ 258, 361, 61 },
		{ 258, 362, 61 },
		{ 258, 363, 61 },
		{ 258, 364, 61 },
		{ 258, 365, 61 },
		{ 258, 366, 61 },
		{ 258, 367, 61 },
		{ 258, 368, 61 },
		{ 258, 369, 61 },
		{ 258, 370, 61 },
		{ 258, 371, 61 },
		{ 258, 372, 61 },
		{ 258, 373, 61 },
		{ 258, 374, 61 },
		{ 258, 375, 61 },
		{ 258, 376, 61 },
		{ 258, 377, 61 },
		{ 258, 378, 61 },
		{ 258, 379, 61 },
		{ 258, 380, 61 },
		{ 258, 381, 61 },
		{ 258, 382, 61 },
		{ 258, 383, 61 },
		{ 258, 384, 61 },
		{ 258, 385, 61 },
		{ 258, 386, 61 },
		{ 258, 387, 61 },
		{ 258, 388, 61 },
		{ 258, 389, 61 },
		{ 258, 390, 61 },
		{ 258, 391, 61 },
		{ 258, 392, 61 },
		{ 258, 393, 61 },
		{ 258, 394, 61 },
		{ 258, 395, 61 },
		{ 258, 396, 61 },
		{ 258, 397, 61 },
		{ 258, 398, 61 },
		{ 258, 399, 61 },
		{ 258, 400, 61 },
		{ 258, 401, 61 },
		{ 258, 402, 61 },
		{ 258, 403, 61 },
		{ 258, 404, 61 },
		{ 258, 405, 61 },
		{ 258, 406, 61 },
		{ 258, 407, 61 },
		{ 258, 408, 61 },
		{ 258, 409, 61 },
		{ 258, 410, 61 },
		{ 258, 411, 61 },
		{ 258, 412, 61 },
		{ 258, 413, 61 },
		{ 258, 414, 61 },
		{ 258, 415, 61 },
		{ 258, 416, 61 },
		{ 258, 417, 61 },
		{ 258, 418, 61 },
		{ 258, 419, 61 },
		{ 258, 420, 61 },
		{ 258, 421, 61 },
		{ 258, 422, 61 },
		{ 258, 423, 61 },
		{ 258, 424, 61 },
		{ 258, 425, 61 },
		{ 258, 426, 61 },
		{ 258, 427, 61 },
		{ 258, 428, 61 },
		{ 258, 429, 61 },
		{ 258, 430, 61 },
		{ 258, 431, 61 },
		{ 258, 432, 61 },
		{ 258, 433, 61 },
		{ 258, 434, 61 },
		{ 258, 435, 61 },
		{ 258, 436, 61 },
		{ 258, 437, 61 },
		{ 258, 438, 61 },
		{ 258, 439, 61 },
		{ 258, 440, 61 },
		{ 258, 441, 61 },
		{ 258, 442, 61 },
		{ 258, 443, 61 },
		{ 258, 444, 61 },
		{ 258, 445, 61 },
		{ 258, 446, 61 },
		{ 258, 447, 61 },
		{ 258, 448, 61 },
		{ 258, 449, 61 },
		{ 258, 450, 61 },
		{ 258, 451, 61 },
		{ 258, 452, 61 },
		{ 258, 453, 61 },
		{ 258, 454, 61 },
		{ 258, 455, 61 },
		{ 258, 456, 61 },
		{ 258, 457, 61 },
		{ 258, 458, 61 },
		{ 258, 459, 61 },
		{ 258, 460, 61 },
		{ 258, 461, 61 },
		{ 258, 462, 61 },
		{ 258, 463, 61 },
		{ 258, 464, 61 },
		{ 258, 465, 61 },
		{ 258, 466, 61 },
		{ 258, 467, 61 },
		{ 258, 468, 61 },
		{ 258, 469, 61 },
		{ 258, 470, 61 },
		{ 258, 471, 61 },
		{ 258, 472, 61 },
		{ 258, 473, 61 },
		{ 258, 474, 61 },
		{ 258, 475, 61 },
		{ 258, 476, 61 },
		{ 258, 477, 61 },
		{ 258, 478, 61 },
		{ 258, 479, 61 },
		{ 258, 480, 61 },
		{ 258, 481, 61 },
		{ 258, 482, 61 },
		{ 258, 483, 61 },
		{ 258, 484, 61 },
		{ 258, 485, 61 },
		{ 258, 486, 61 },
		{ 258, 487, 61 },
		{ 258, 488, 61 },
		{ 258, 489, 61 },
		{ 258, 490, 61 },
		{ 258, 491, 61 },
		{ 258, 492, 61 },
		{ 258, 493, 61 },
		{ 258, 494, 61 },
		{ 258, 495, 61 },
		{ 258, 496, 61 },
		{ 258, 497, 61 },
		{ 258, 498, 61 },
		{ 258, 499, 61 },
		{ 258, 500, 61 },
		{ 258, 501, 61 },
		{ 258, 502, 61 },
		{ 258, 503, 61 },
		{ 258, 504, 61 },
		{ 258, 505, 61 },
		{ 258, 506, 61 },
		{ 258, 507, 61 },
		{ 258, 508, 61 },
		{ 258, 509, 61 },
		{ 258, 510, 61 },
		{ 258, 511, 61 },
		{ 258, 512, 61 },
		{ 259, 323, 61 },
		{ 259, 324, 61 },
		{ 259, 325, 61 },
		{ 259, 326, 61 },
		{ 259, 327, 61 },
		{ 259, 328, 61 },
		{ 259, 329, 61 },
		{ 259, 330, 61 },
		{ 259, 331, 61 },
		{ 259, 332, 61 },
		{ 259, 333, 61 },
		{ 259, 334, 61 },
		{ 259, 335, 61 },
		{ 259, 336, 61 },
		{ 259, 337, 61 },
		{ 259, 338, 61 },
		{ 259, 339, 61 },
		{ 259, 340, 61 },
		{ 259, 341, 61 },
		{ 259, 342, 61 },
		{ 259, 343, 61 },
		{ 259, 344, 61 },
		{ 259, 345, 61 },
		{ 259, 346, 61 },
		{ 259, 347, 61 },
		{ 259, 348, 61 },
		{ 259, 349, 61 },
		{ 259, 350, 61 },
		{ 259, 351, 61 },
		{ 259, 352, 61 },
		{ 259, 353, 61 },
		{ 259, 354, 61 },
		{ 259, 355, 61 },
		{ 259, 356, 61 },
		{ 259, 357, 61 },
		{ 259, 358, 61 },
		{ 259, 359, 61 },
		{ 259, 360, 61 },
		{ 259, 361, 61 },
		{ 259, 362, 61 },
		{ 259, 363, 61 },
		{ 259, 364, 61 },
		{ 259, 365, 61 },
		{ 259, 366, 61 },
		{ 259, 367, 61 },
		{ 259, 368, 61 },
		{ 259, 369, 61 },
		{ 259, 370, 61 },
		{ 259, 371, 61 },
		{ 259, 372, 61 },
		{ 259, 373, 61 },
		{ 259, 374, 61 },
		{ 259, 375, 61 },
		{ 259, 376, 61 },
		{ 259, 377, 61 },
		{ 259, 378, 61 },
		{ 259, 379, 61 },
		{ 259, 380, 61 },
		{ 259, 381, 61 },
		{ 259, 382, 61 },
		{ 259, 383, 61 },
		{ 259, 384, 61 },
		{ 259, 385, 61 },
		{ 259, 386, 61 },
		{ 259, 387, 61 },
		{ 259, 388, 61 },
		{ 259, 389, 61 },
		{ 259, 390, 61 },
		{ 259, 391, 61 },
		{ 259, 392, 61 },
		{ 259, 393, 61 },
		{ 259, 394, 61 },
		{ 259, 395, 61 },
		{ 259, 396, 61 },
		{ 259, 397, 61 },
		{ 259, 398, 61 },
		{ 259, 399, 61 },
		{ 259, 400, 61 },
		{ 259, 401, 61 },
		{ 259, 402, 61 },
		{ 259, 403, 61 },
		{ 259, 404, 61 },
		{ 259, 405, 61 },
		{ 259, 406, 61 },
		{ 259, 407, 61 },
		{ 259, 408, 61 },
		{ 259, 409, 61 },
		{ 259, 410, 61 },
		{ 259, 411, 61 },
		{ 259, 412, 61 },
		{ 259, 413, 61 },
		{ 259, 414, 61 },
		{ 259, 415, 61 },
		{ 259, 416, 61 },
		{ 259, 417, 61 },
		{ 259, 418, 61 },
		{ 259, 419, 61 },
		{ 259, 420, 61 },
		{ 259, 421, 61 },
		{ 259, 422, 61 },
		{ 259, 423, 61 },
		{ 259, 424, 61 },
		{ 259, 425, 61 },
		{ 259, 426, 61 },
		{ 259, 427, 61 },
		{ 259, 428, 61 },
		{ 259, 429, 61 },
		{ 259, 430, 61 },
		{ 259, 431, 61 },
		{ 259, 432, 61 },
		{ 259, 433, 61 },
		{ 259, 434, 61 },
		{ 259, 435, 61 },
		{ 259, 436, 61 },
		{ 259, 437, 61 },
		{ 259, 438, 61 },
		{ 259, 439, 61 },
		{ 259, 440, 61 },
		{ 259, 441, 61 },
		{ 259, 442, 61 },
		{ 259, 443, 61 },
		{ 259, 444, 61 },
		{ 259, 445, 61 },
		{ 259, 446, 61 },
		{ 259, 447, 61 },
		{ 259, 448, 61 },
		{ 259, 449, 61 },
		{ 259, 450, 61 },
		{ 259, 451, 61 },
		{ 259, 452, 61 },
		{ 259, 453, 61 },
		{ 259, 454, 61 },
		{ 259, 455, 61 },
		{ 259, 456, 61 },
		{ 259, 457, 61 },
		{ 259, 458, 61 },
		{ 259, 459, 61 },
		{ 259, 460, 61 },
		{ 259, 461, 61 },
		{ 259, 462, 61 },
		{ 259, 463, 61 },
		{ 259, 464, 61 },
		{ 259, 465, 61 },
		{ 259, 466, 61 },
		{ 259, 467, 61 },
		{ 259, 468, 61 },
		{ 259, 469, 61 },
		{ 259, 470, 61 },
		{ 259, 471, 61 },
		{ 259, 472, 61 },
		{ 259, 473, 61 },
		{ 259, 474, 61 },
		{ 259, 475, 61 },
		{ 259, 476, 61 },
		{ 259, 477, 61 },
		{ 259, 478, 61 },
		{ 259, 479, 61 },
		{ 259, 480, 61 },
		{ 259, 481, 61 },
		{ 259, 482, 61 },
		{ 259, 483, 61 },
		{ 259, 484, 61 },
		{ 259, 485, 61 },
		{ 259, 486, 61 },
		{ 259, 487, 61 },
		{ 259, 488, 61 },
		{ 259, 489, 61 },
		{ 259, 490, 61 },
		{ 259, 491, 61 },
		{ 259, 492, 61 },
		{ 259, 493, 61 },
		{ 259, 494, 61 },
		{ 259, 495, 61 },
		{ 259, 496, 61 },
		{ 259, 497, 61 },
		{ 259, 498, 61 },
		{ 259, 499, 61 },
		{ 259, 500, 61 },
		{ 259, 501, 61 },
		{ 259, 502, 61 },
		{ 259, 503, 61 },
		{ 259, 504, 61 },
		{ 259, 505, 61 },
		{ 259, 506, 61 },
		{ 259, 507, 61 },
		{ 259, 508, 61 },
		{ 259, 509, 61 },
		{ 259, 510, 61 },
		{ 259, 511, 61 },
		{ 259, 512, 61 },
		{ 260, 323, 61 },
		{ 260, 324, 61 },
		{ 260, 325, 61 },
		{ 260, 326, 61 },
		{ 260, 327, 61 },
		{ 260, 328, 61 },
		{ 260, 329, 61 },
		{ 260, 330, 61 },
		{ 260, 331, 61 },
		{ 260, 332, 61 },
		{ 260, 333, 61 },
		{ 260, 334, 61 },
		{ 260, 335, 61 },
		{ 260, 336, 61 },
		{ 260, 337, 61 },
		{ 260, 338, 61 },
		{ 260, 339, 61 },
		{ 260, 340, 61 },
		{ 260, 341, 61 },
		{ 260, 342, 61 },
		{ 260, 343, 61 },
		{ 260, 344, 61 },
		{ 260, 345, 61 },
		{ 260, 346, 61 },
		{ 260, 347, 61 },
		{ 260, 348, 61 },
		{ 260, 349, 61 },
		{ 260, 350, 61 },
		{ 260, 351, 61 },
		{ 260, 352, 61 },
		{ 260, 353, 61 },
		{ 260, 354, 61 },
		{ 260, 355, 61 },
		{ 260, 356, 61 },
		{ 260, 357, 61 },
		{ 260, 358, 61 },
		{ 260, 359, 61 },
		{ 260, 360, 61 },
		{ 260, 361, 61 },
		{ 260, 362, 61 },
		{ 260, 363, 61 },
		{ 260, 364, 61 },
		{ 260, 365, 61 },
		{ 260, 366, 61 },
		{ 260, 367, 61 },
		{ 260, 368, 61 },
		{ 260, 369, 61 },
		{ 260, 370, 61 },
		{ 260, 371, 61 },
		{ 260, 372, 61 },
		{ 260, 373, 61 },
		{ 260, 374, 61 },
		{ 260, 375, 61 },
		{ 260, 376, 61 },
		{ 260, 377, 61 },
		{ 260, 378, 61 },
		{ 260, 379, 61 },
		{ 260, 380, 61 },
		{ 260, 381, 61 },
		{ 260, 382, 61 },
		{ 260, 383, 61 },
		{ 260, 384, 61 },
		{ 260, 385, 61 },
		{ 260, 386, 61 },
		{ 260, 387, 61 },
		{ 260, 388, 61 },
		{ 260, 389, 61 },
		{ 260, 390, 61 },
		{ 260, 391, 61 },
		{ 260, 392, 61 },
		{ 260, 393, 61 },
		{ 260, 394, 61 },
		{ 260, 395, 61 },
		{ 260, 396, 61 },
		{ 260, 397, 61 },
		{ 260, 398, 61 },
		{ 260, 399, 61 },
		{ 260, 400, 61 },
		{ 260, 401, 61 },
		{ 260, 402, 61 },
		{ 260, 403, 61 },
		{ 260, 404, 61 },
		{ 260, 405, 61 },
		{ 260, 406, 61 },
		{ 260, 407, 61 },
		{ 260, 408, 61 },
		{ 260, 409, 61 },
		{ 260, 410, 61 },
		{ 260, 411, 61 },
		{ 260, 412, 61 },
		{ 260, 413, 61 },
		{ 260, 414, 61 },
		{ 260, 415, 61 },
		{ 260, 416, 61 },
		{ 260, 417, 61 },
		{ 260, 418, 61 },
		{ 260, 419, 61 },
		{ 260, 420, 61 },
		{ 260, 421, 61 },
		{ 260, 422, 61 },
		{ 260, 423, 61 },
		{ 260, 424, 61 },
		{ 260, 425, 61 },
		{ 260, 426, 61 },
		{ 260, 427, 61 },
		{ 260, 428, 61 },
		{ 260, 429, 61 },
		{ 260, 430, 61 },
		{ 260, 431, 61 },
		{ 260, 432, 61 },
		{ 260, 433, 61 },
		{ 260, 434, 61 },
		{ 260, 435, 61 },
		{ 260, 436, 61 },
		{ 260, 437, 61 },
		{ 260, 438, 61 },
		{ 260, 439, 61 },
		{ 260, 440, 61 },
		{ 260, 441, 61 },
		{ 260, 442, 61 },
		{ 260, 443, 61 },
		{ 260, 444, 61 },
		{ 260, 445, 61 },
		{ 260, 446, 61 },
		{ 260, 447, 61 },
		{ 260, 448, 61 },
		{ 260, 449, 61 },
		{ 260, 450, 61 },
		{ 260, 451, 61 },
		{ 260, 452, 61 },
		{ 260, 453, 61 },
		{ 260, 454, 61 },
		{ 260, 455, 61 },
		{ 260, 456, 61 },
		{ 260, 457, 61 },
		{ 260, 458, 61 },
		{ 260, 459, 61 },
		{ 260, 460, 61 },
		{ 260, 461, 61 },
		{ 260, 462, 61 },
		{ 260, 463, 61 },
		{ 260, 464, 61 },
		{ 260, 465, 61 },
		{ 260, 466, 61 },
		{ 260, 467, 61 },
		{ 260, 468, 61 },
		{ 260, 469, 61 },
		{ 260, 470, 61 },
		{ 260, 471, 61 },
		{ 260, 472, 61 },
		{ 260, 473, 61 },
		{ 260, 474, 61 },
		{ 260, 475, 61 },
		{ 260, 476, 61 },
		{ 260, 477, 61 },
		{ 260, 478, 61 },
		{ 260, 479, 61 },
		{ 260, 480, 61 },
		{ 260, 481, 61 },
		{ 260, 482, 61 },
		{ 260, 483, 61 },
		{ 260, 484, 61 },
		{ 260, 485, 61 },
		{ 260, 486, 61 },
		{ 260, 487, 61 },
		{ 260, 488, 61 },
		{ 260, 489, 61 },
		{ 260, 490, 61 },
		{ 260, 491, 61 },
		{ 260, 492, 61 },
		{ 260, 493, 61 },
		{ 260, 494, 61 },
		{ 260, 495, 61 },
		{ 260, 496, 61 },
		{ 260, 497, 61 },
		{ 260, 498, 61 },
		{ 260, 499, 61 },
		{ 260, 500, 61 },
		{ 260, 501, 61 },
		{ 260, 502, 61 },
		{ 260, 503, 61 },
		{ 260, 504, 61 },
		{ 260, 505, 61 },
		{ 260, 506, 61 },
		{ 260, 507, 61 },
		{ 260, 508, 61 },
		{ 260, 509, 61 },
		{ 260, 510, 61 },
		{ 260, 511, 61 },
		{ 260, 512, 61 },
		{ 261, 323, 61 },
		{ 261, 324, 61 },
		{ 261, 325, 61 },
		{ 261, 326, 61 },
		{ 261, 327, 61 },
		{ 261, 328, 61 },
		{ 261, 329, 61 },
		{ 261, 330, 61 },
		{ 261, 331, 61 },
		{ 261, 332, 61 },
		{ 261, 333, 61 },
		{ 261, 334, 61 },
		{ 261, 335, 61 },
		{ 261, 336, 61 },
		{ 261, 337, 61 },
		{ 261, 338, 61 },
		{ 261, 339, 61 },
		{ 261, 340, 61 },
		{ 261, 341, 61 },
		{ 261, 342, 61 },
		{ 261, 343, 61 },
		{ 261, 344, 61 },
		{ 261, 345, 61 },
		{ 261, 346, 61 },
		{ 261, 347, 61 },
		{ 261, 348, 61 },
		{ 261, 349, 61 },
		{ 261, 350, 61 },
		{ 261, 351, 61 },
		{ 261, 352, 61 },
		{ 261, 353, 61 },
		{ 261, 354, 61 },
		{ 261, 355, 61 },
		{ 261, 356, 61 },
		{ 261, 357, 61 },
		{ 261, 358, 61 },
		{ 261, 359, 61 },
		{ 261, 360, 61 },
		{ 261, 361, 61 },
		{ 261, 362, 61 },
		{ 261, 363, 61 },
		{ 261, 364, 61 },
		{ 261, 365, 61 },
		{ 261, 366, 61 },
		{ 261, 367, 61 },
		{ 261, 368, 61 },
		{ 261, 369, 61 },
		{ 261, 370, 61 },
		{ 261, 371, 61 },
		{ 261, 372, 61 },
		{ 261, 373, 61 },
		{ 261, 374, 61 },
		{ 261, 375, 61 },
		{ 261, 376, 61 },
		{ 261, 377, 61 },
		{ 261, 378, 61 },
		{ 261, 379, 61 },
		{ 261, 380, 61 },
		{ 261, 381, 61 },
		{ 261, 382, 61 },
		{ 261, 383, 61 },
		{ 261, 384, 61 },
		{ 261, 385, 61 },
		{ 261, 386, 61 },
		{ 261, 387, 61 },
		{ 261, 388, 61 },
		{ 261, 389, 61 },
		{ 261, 390, 61 },
		{ 261, 391, 61 },
		{ 261, 392, 61 },
		{ 261, 393, 61 },
		{ 261, 394, 61 },
		{ 261, 395, 61 },
		{ 261, 396, 61 },
		{ 261, 397, 61 },
		{ 261, 398, 61 },
		{ 261, 399, 61 },
		{ 261, 400, 61 },
		{ 261, 401, 61 },
		{ 261, 402, 61 },
		{ 261, 403, 61 },
		{ 261, 404, 61 },
		{ 261, 405, 61 },
		{ 261, 406, 61 },
		{ 261, 407, 61 },
		{ 261, 408, 61 },
		{ 261, 409, 61 },
		{ 261, 410, 61 },
		{ 261, 411, 61 },
		{ 261, 412, 61 },
		{ 261, 413, 61 },
		{ 261, 414, 61 },
		{ 261, 415, 61 },
		{ 261, 416, 61 },
		{ 261, 417, 61 },
		{ 261, 418, 61 },
		{ 261, 419, 61 },
		{ 261, 420, 61 },
		{ 261, 421, 61 },
		{ 261, 422, 61 },
		{ 261, 423, 61 },
		{ 261, 424, 61 },
		{ 261, 425, 61 },
		{ 261, 426, 61 },
		{ 261, 427, 61 },
		{ 261, 428, 61 },
		{ 261, 429, 61 },
		{ 261, 430, 61 },
		{ 261, 431, 61 },
		{ 261, 432, 61 },
		{ 261, 433, 61 },
		{ 261, 434, 61 },
		{ 261, 435, 61 },
		{ 261, 436, 61 },
		{ 261, 437, 61 },
		{ 261, 438, 61 },
		{ 261, 439, 61 },
		{ 261, 440, 61 },
		{ 261, 441, 61 },
		{ 261, 442, 61 },
		{ 261, 443, 61 },
		{ 261, 444, 61 },
		{ 261, 445, 61 },
		{ 261, 446, 61 },
		{ 261, 447, 61 },
		{ 261, 448, 61 },
		{ 261, 449, 61 },
		{ 261, 450, 61 },
		{ 261, 451, 61 },
		{ 261, 452, 61 },
		{ 261, 453, 61 },
		{ 261, 454, 61 },
		{ 261, 455, 61 },
		{ 261, 456, 61 },
		{ 261, 457, 61 },
		{ 261, 458, 61 },
		{ 261, 459, 61 },
		{ 261, 460, 61 },
		{ 261, 461, 61 },
		{ 261, 462, 61 },
		{ 261, 463, 61 },
		{ 261, 464, 61 },
		{ 261, 465, 61 },
		{ 261, 466, 61 },
		{ 261, 467, 61 },
		{ 261, 468, 61 },
		{ 261, 469, 61 },
		{ 261, 470, 61 },
		{ 261, 471, 61 },
		{ 261, 472, 61 },
		{ 261, 473, 61 },
		{ 261, 474, 61 },
		{ 261, 475, 61 },
		{ 261, 476, 61 },
		{ 261, 477, 61 },
		{ 261, 478, 61 },
		{ 261, 479, 61 },
		{ 261, 480, 61 },
		{ 261, 481, 61 },
		{ 261, 482, 61 },
		{ 261, 483, 61 },
		{ 261, 484, 61 },
		{ 261, 485, 61 },
		{ 261, 486, 61 },
		{ 261, 487, 61 },
		{ 261, 488, 61 },
		{ 261, 489, 61 },
		{ 261, 490, 61 },
		{ 261, 491, 61 },
		{ 261, 492, 61 },
		{ 261, 493, 61 },
		{ 261, 494, 61 },
		{ 261, 495, 61 },
		{ 261, 496, 61 },
		{ 261, 497, 61 },
		{ 261, 498, 61 },
		{ 261, 499, 61 },
		{ 261, 500, 61 },
		{ 261, 501, 61 },
		{ 261, 502, 61 },
		{ 261, 503, 61 },
		{ 261, 504, 61 },
		{ 261, 505, 61 },
		{ 261, 506, 61 },
		{ 261, 507, 61 },
		{ 261, 508, 61 },
		{ 261, 509, 61 },
		{ 261, 510, 61 },
		{ 261, 511, 61 },
		{ 261, 512, 61 },
		{ 262, 323, 61 },
		{ 262, 324, 61 },
		{ 262, 325, 61 },
		{ 262, 326, 61 },
		{ 262, 327, 61 },
		{ 262, 328, 61 },
		{ 262, 329, 61 },
		{ 262, 330, 61 },
		{ 262, 331, 61 },
		{ 262, 332, 61 },
		{ 262, 333, 61 },
		{ 262, 334, 61 },
		{ 262, 335, 61 },
		{ 262, 336, 61 },
		{ 262, 337, 61 },
		{ 262, 338, 61 },
		{ 262, 339, 61 },
		{ 262, 340, 61 },
		{ 262, 341, 61 },
		{ 262, 342, 61 },
		{ 262, 343, 61 },
		{ 262, 344, 61 },
		{ 262, 345, 61 },
		{ 262, 346, 61 },
		{ 262, 347, 61 },
		{ 262, 348, 61 },
		{ 262, 349, 61 },
		{ 262, 350, 61 },
		{ 262, 351, 61 },
		{ 262, 352, 61 },
		{ 262, 353, 61 },
		{ 262, 354, 61 },
		{ 262, 355, 61 },
		{ 262, 356, 61 },
		{ 262, 357, 61 },
		{ 262, 358, 61 },
		{ 262, 359, 61 },
		{ 262, 360, 61 },
		{ 262, 361, 61 },
		{ 262, 362, 61 },
		{ 262, 363, 61 },
		{ 262, 364, 61 },
		{ 262, 365, 61 },
		{ 262, 366, 61 },
		{ 262, 367, 61 },
		{ 262, 368, 61 },
		{ 262, 369, 61 },
		{ 262, 370, 61 },
		{ 262, 371, 61 },
		{ 262, 372, 61 },
		{ 262, 373, 61 },
		{ 262, 374, 61 },
		{ 262, 375, 61 },
		{ 262, 376, 61 },
		{ 262, 377, 61 },
		{ 262, 378, 61 },
		{ 262, 379, 61 },
		{ 262, 380, 61 },
		{ 262, 381, 61 },
		{ 262, 382, 61 },
		{ 262, 383, 61 },
		{ 262, 384, 61 },
		{ 262, 385, 61 },
		{ 262, 386, 61 },
		{ 262, 387, 61 },
		{ 262, 388, 61 },
		{ 262, 389, 61 },
		{ 262, 390, 61 },
		{ 262, 391, 61 },
		{ 262, 392, 61 },
		{ 262, 393, 61 },
		{ 262, 394, 61 },
		{ 262, 395, 61 },
		{ 262, 396, 61 },
		{ 262, 397, 61 },
		{ 262, 398, 61 },
		{ 262, 399, 61 },
		{ 262, 400, 61 },
		{ 262, 401, 61 },
		{ 262, 402, 61 },
		{ 262, 403, 61 },
		{ 262, 404, 61 },
		{ 262, 405, 61 },
		{ 262, 406, 61 },
		{ 262, 407, 61 },
		{ 262, 408, 61 },
		{ 262, 409, 61 },
		{ 262, 410, 61 },
		{ 262, 411, 61 },
		{ 262, 412, 61 },
		{ 262, 413, 61 },
		{ 262, 414, 61 },
		{ 262, 415, 61 },
		{ 262, 416, 61 },
		{ 262, 417, 61 },
		{ 262, 418, 61 },
		{ 262, 419, 61 },
		{ 262, 420, 61 },
		{ 262, 421, 61 },
		{ 262, 422, 61 },
		{ 262, 423, 61 },
		{ 262, 424, 61 },
		{ 262, 425, 61 },
		{ 262, 426, 61 },
		{ 262, 427, 61 },
		{ 262, 428, 61 },
		{ 262, 429, 61 },
		{ 262, 430, 61 },
		{ 262, 431, 61 },
		{ 262, 432, 61 },
		{ 262, 433, 61 },
		{ 262, 434, 61 },
		{ 262, 435, 61 },
		{ 262, 436, 61 },
		{ 262, 437, 61 },
		{ 262, 438, 61 },
		{ 262, 439, 61 },
		{ 262, 440, 61 },
		{ 262, 441, 61 },
		{ 262, 442, 61 },
		{ 262, 443, 61 },
		{ 262, 444, 61 },
		{ 262, 445, 61 },
		{ 262, 446, 61 },
		{ 262, 447, 61 },
		{ 262, 448, 61 },
		{ 262, 449, 61 },
		{ 262, 450, 61 },
		{ 262, 451, 61 },
		{ 262, 452, 61 },
		{ 262, 453, 61 },
		{ 262, 454, 61 },
		{ 262, 455, 61 },
		{ 262, 456, 61 },
		{ 262, 457, 61 },
		{ 262, 458, 61 },
		{ 262, 459, 61 },
		{ 262, 460, 61 },
		{ 262, 461, 61 },
		{ 262, 462, 61 },
		{ 262, 463, 61 },
		{ 262, 464, 61 },
		{ 262, 465, 61 },
		{ 262, 466, 61 },
		{ 262, 467, 61 },
		{ 262, 468, 61 },
		{ 262, 469, 61 },
		{ 262, 470, 61 },
		{ 262, 471, 61 },
		{ 262, 472, 61 },
		{ 262, 473, 61 },
		{ 262, 474, 61 },
		{ 262, 475, 61 },
		{ 262, 476, 61 },
		{ 262, 477, 61 },
		{ 262, 478, 61 },
		{ 262, 479, 61 },
		{ 262, 480, 61 },
		{ 262, 481, 61 },
		{ 262, 482, 61 },
		{ 262, 483, 61 },
		{ 262, 484, 61 },
		{ 262, 485, 61 },
		{ 262, 486, 61 },
		{ 262, 487, 61 },
		{ 262, 488, 61 },
		{ 262, 489, 61 },
		{ 262, 490, 61 },
		{ 262, 491, 61 },
		{ 262, 492, 61 },
		{ 262, 493, 61 },
		{ 262, 494, 61 },
		{ 262, 495, 61 },
		{ 262, 496, 61 },
		{ 262, 497, 61 },
		{ 262, 498, 61 },
		{ 262, 499, 61 },
		{ 262, 500, 61 },
		{ 262, 501, 61 },
		{ 262, 502, 61 },
		{ 262, 503, 61 },
		{ 262, 504, 61 },
		{ 262, 505, 61 },
		{ 262, 506, 61 },
		{ 262, 507, 61 },
		{ 262, 508, 61 },
		{ 262, 509, 61 },
		{ 262, 510, 61 },
		{ 262, 511, 61 },
		{ 262, 512, 61 },
		{ 263, 323, 61 },
		{ 263, 324, 61 },
		{ 263, 325, 61 },
		{ 263, 326, 61 },
		{ 263, 327, 61 },
		{ 263, 328, 61 },
		{ 263, 329, 61 },
		{ 263, 330, 61 },
		{ 263, 331, 61 },
		{ 263, 332, 61 },
		{ 263, 333, 61 },
		{ 263, 334, 61 },
		{ 263, 335, 61 },
		{ 263, 336, 61 },
		{ 263, 337, 61 },
		{ 263, 338, 61 },
		{ 263, 339, 61 },
		{ 263, 340, 61 },
		{ 263, 341, 61 },
		{ 263, 342, 61 },
		{ 263, 343, 61 },
		{ 263, 344, 61 },
		{ 263, 345, 61 },
		{ 263, 346, 61 },
		{ 263, 347, 61 },
		{ 263, 348, 61 },
		{ 263, 349, 61 },
		{ 263, 350, 61 },
		{ 263, 351, 61 },
		{ 263, 352, 61 },
		{ 263, 353, 61 },
		{ 263, 354, 61 },
		{ 263, 355, 61 },
		{ 263, 356, 61 },
		{ 263, 357, 61 },
		{ 263, 358, 61 },
		{ 263, 359, 61 },
		{ 263, 360, 61 },
		{ 263, 361, 61 },
		{ 263, 362, 61 },
		{ 263, 363, 61 },
		{ 263, 364, 61 },
		{ 263, 365, 61 },
		{ 263, 366, 61 },
		{ 263, 367, 61 },
		{ 263, 368, 61 },
		{ 263, 369, 61 },
		{ 263, 370, 61 },
		{ 263, 371, 61 },
		{ 263, 372, 61 },
		{ 263, 373, 61 },
		{ 263, 374, 61 },
		{ 263, 375, 61 },
		{ 263, 376, 61 },
		{ 263, 377, 61 },
		{ 263, 378, 61 },
		{ 263, 379, 61 },
		{ 263, 380, 61 },
		{ 263, 381, 61 },
		{ 263, 382, 61 },
		{ 263, 383, 61 },
		{ 263, 384, 61 },
		{ 263, 385, 61 },
		{ 263, 386, 61 },
		{ 263, 387, 61 },
		{ 263, 388, 61 },
		{ 263, 389, 61 },
		{ 263, 390, 61 },
		{ 263, 391, 61 },
		{ 263, 392, 61 },
		{ 263, 393, 61 },
		{ 263, 394, 61 },
		{ 263, 395, 61 },
		{ 263, 396, 61 },
		{ 263, 397, 61 },
		{ 263, 398, 61 },
		{ 263, 399, 61 },
		{ 263, 400, 61 },
		{ 263, 401, 61 },
		{ 263, 402, 61 },
		{ 263, 403, 61 },
		{ 263, 404, 61 },
		{ 263, 405, 61 },
		{ 263, 406, 61 },
		{ 263, 407, 61 },
		{ 263, 408, 61 },
		{ 263, 409, 61 },
		{ 263, 410, 61 },
		{ 263, 411, 61 },
		{ 263, 412, 61 },
		{ 263, 413, 61 },
		{ 263, 414, 61 },
		{ 263, 415, 61 },
		{ 263, 416, 61 },
		{ 263, 417, 61 },
		{ 263, 418, 61 },
		{ 263, 419, 61 },
		{ 263, 420, 61 },
		{ 263, 421, 61 },
		{ 263, 422, 61 },
		{ 263, 423, 61 },
		{ 263, 424, 61 },
		{ 263, 425, 61 },
		{ 263, 426, 61 },
		{ 263, 427, 61 },
		{ 263, 428, 61 },
		{ 263, 429, 61 },
		{ 263, 430, 61 },
		{ 263, 431, 61 },
		{ 263, 432, 61 },
		{ 263, 433, 61 },
		{ 263, 434, 61 },
		{ 263, 435, 61 },
		{ 263, 436, 61 },
		{ 263, 437, 61 },
		{ 263, 438, 61 },
		{ 263, 439, 61 },
		{ 263, 440, 61 },
		{ 263, 441, 61 },
		{ 263, 442, 61 },
		{ 263, 443, 61 },
		{ 263, 444, 61 },
		{ 263, 445, 61 },
		{ 263, 446, 61 },
		{ 263, 447, 61 },
		{ 263, 448, 61 },
		{ 263, 449, 61 },
		{ 263, 450, 61 },
		{ 263, 451, 61 },
		{ 263, 452, 61 },
		{ 263, 453, 61 },
		{ 263, 454, 61 },
		{ 263, 455, 61 },
		{ 263, 456, 61 },
		{ 263, 457, 61 },
		{ 263, 458, 61 },
		{ 263, 459, 61 },
		{ 263, 460, 61 },
		{ 263, 461, 61 },
		{ 263, 462, 61 },
		{ 263, 463, 61 },
		{ 263, 464, 61 },
		{ 263, 465, 61 },
		{ 263, 466, 61 },
		{ 263, 467, 61 },
		{ 263, 468, 61 },
		{ 263, 469, 61 },
		{ 263, 470, 61 },
		{ 263, 471, 61 },
		{ 263, 472, 61 },
		{ 263, 473, 61 },
		{ 263, 474, 61 },
		{ 263, 475, 61 },
		{ 263, 476, 61 },
		{ 263, 477, 61 },
		{ 263, 478, 61 },
		{ 263, 479, 61 },
		{ 263, 480, 61 },
		{ 263, 481, 61 },
		{ 263, 482, 61 },
		{ 263, 483, 61 },
		{ 263, 484, 61 },
		{ 263, 485, 61 },
		{ 263, 486, 61 },
		{ 263, 487, 61 },
		{ 263, 488, 61 },
		{ 263, 489, 61 },
		{ 263, 490, 61 },
		{ 263, 491, 61 },
		{ 263, 492, 61 },
		{ 263, 493, 61 },
		{ 263, 494, 61 },
		{ 263, 495, 61 },
		{ 263, 496, 61 },
		{ 263, 497, 61 },
		{ 263, 498, 61 },
		{ 263, 499, 61 },
		{ 263, 500, 61 },
		{ 263, 501, 61 },
		{ 263, 502, 61 },
		{ 263, 503, 61 },
		{ 263, 504, 61 },
		{ 263, 505, 61 },
		{ 263, 506, 61 },
		{ 263, 507, 61 },
		{ 263, 508, 61 },
		{ 263, 509, 61 },
		{ 263, 510, 61 },
		{ 263, 511, 61 },
		{ 263, 512, 61 },
		{ 264, 323, 61 },
		{ 264, 324, 61 },
		{ 264, 325, 61 },
		{ 264, 326, 61 },
		{ 264, 327, 61 },
		{ 264, 328, 61 },
		{ 264, 329, 61 },
		{ 264, 330, 61 },
		{ 264, 331, 61 },
		{ 264, 332, 61 },
		{ 264, 333, 61 },
		{ 264, 334, 61 },
		{ 264, 335, 61 },
		{ 264, 336, 61 },
		{ 264, 337, 61 },
		{ 264, 338, 61 },
		{ 264, 339, 61 },
		{ 264, 340, 61 },
		{ 264, 341, 61 },
		{ 264, 342, 61 },
		{ 264, 343, 61 },
		{ 264, 344, 61 },
		{ 264, 345, 61 },
		{ 264, 346, 61 },
		{ 264, 347, 61 },
		{ 264, 348, 61 },
		{ 264, 349, 61 },
		{ 264, 350, 61 },
		{ 264, 351, 61 },
		{ 264, 352, 61 },
		{ 264, 353, 61 },
		{ 264, 354, 61 },
		{ 264, 355, 61 },
		{ 264, 356, 61 },
		{ 264, 357, 61 },
		{ 264, 358, 61 },
		{ 264, 359, 61 },
		{ 264, 360, 61 },
		{ 264, 361, 61 },
		{ 264, 362, 61 },
		{ 264, 363, 61 },
		{ 264, 364, 61 },
		{ 264, 365, 61 },
		{ 264, 366, 61 },
		{ 264, 367, 61 },
		{ 264, 368, 61 },
		{ 264, 369, 61 },
		{ 264, 370, 61 },
		{ 264, 371, 61 },
		{ 264, 372, 61 },
		{ 264, 373, 61 },
		{ 264, 374, 61 },
		{ 264, 375, 61 },
		{ 264, 376, 61 },
		{ 264, 377, 61 },
		{ 264, 378, 61 },
		{ 264, 379, 61 },
		{ 264, 380, 61 },
		{ 264, 381, 61 },
		{ 264, 382, 61 },
		{ 264, 383, 61 },
		{ 264, 384, 61 },
		{ 264, 385, 61 },
		{ 264, 386, 61 },
		{ 264, 387, 61 },
		{ 264, 388, 61 },
		{ 264, 389, 61 },
		{ 264, 390, 61 },
		{ 264, 391, 61 },
		{ 264, 392, 61 },
		{ 264, 393, 61 },
		{ 264, 394, 61 },
		{ 264, 395, 61 },
		{ 264, 396, 61 },
		{ 264, 397, 61 },
		{ 264, 398, 61 },
		{ 264, 399, 61 },
		{ 264, 400, 61 },
		{ 264, 401, 61 },
		{ 264, 402, 61 },
		{ 264, 403, 61 },
		{ 264, 404, 61 },
		{ 264, 405, 61 },
		{ 264, 406, 61 },
		{ 264, 407, 61 },
		{ 264, 408, 61 },
		{ 264, 409, 61 },
		{ 264, 410, 61 },
		{ 264, 411, 61 },
		{ 264, 412, 61 },
		{ 264, 413, 61 },
		{ 264, 414, 61 },
		{ 264, 415, 61 },
		{ 264, 416, 61 },
		{ 264, 417, 61 },
		{ 264, 418, 61 },
		{ 264, 419, 61 },
		{ 264, 420, 61 },
		{ 264, 421, 61 },
		{ 264, 422, 61 },
		{ 264, 423, 61 },
		{ 264, 424, 61 },
		{ 264, 425, 61 },
		{ 264, 426, 61 },
		{ 264, 427, 61 },
		{ 264, 428, 61 },
		{ 264, 429, 61 },
		{ 264, 430, 61 },
		{ 264, 431, 61 },
		{ 264, 432, 61 },
		{ 264, 433, 61 },
		{ 264, 434, 61 },
		{ 264, 435, 61 },
		{ 264, 436, 61 },
		{ 264, 437, 61 },
		{ 264, 438, 61 },
		{ 264, 439, 61 },
		{ 264, 440, 61 },
		{ 264, 441, 61 },
		{ 264, 442, 61 },
		{ 264, 443, 61 },
		{ 264, 444, 61 },
		{ 264, 445, 61 },
		{ 264, 446, 61 },
		{ 264, 447, 61 },
		{ 264, 448, 61 },
		{ 264, 449, 61 },
		{ 264, 450, 61 },
		{ 264, 451, 61 },
		{ 264, 452, 61 },
		{ 264, 453, 61 },
		{ 264, 454, 61 },
		{ 264, 455, 61 },
		{ 264, 456, 61 },
		{ 264, 457, 61 },
		{ 264, 458, 61 },
		{ 264, 459, 61 },
		{ 264, 460, 61 },
		{ 264, 461, 61 },
		{ 264, 462, 61 },
		{ 264, 463, 61 },
		{ 264, 464, 61 },
		{ 264, 465, 61 },
		{ 264, 466, 61 },
		{ 264, 467, 61 },
		{ 264, 468, 61 },
		{ 264, 469, 61 },
		{ 264, 470, 61 },
		{ 264, 471, 61 },
		{ 264, 472, 61 },
		{ 264, 473, 61 },
		{ 264, 474, 61 },
		{ 264, 475, 61 },
		{ 264, 476, 61 },
		{ 264, 477, 61 },
		{ 264, 478, 61 },
		{ 264, 479, 61 },
		{ 264, 480, 61 },
		{ 264, 481, 61 },
		{ 264, 482, 61 },
		{ 264, 483, 61 },
		{ 264, 484, 61 },
		{ 264, 485, 61 },
		{ 264, 486, 61 },
		{ 264, 487, 61 },
		{ 264, 488, 61 },
		{ 264, 489, 61 },
		{ 264, 490, 61 },
		{ 264, 491, 61 },
		{ 264, 492, 61 },
		{ 264, 493, 61 },
		{ 264, 494, 61 },
		{ 264, 495, 61 },
		{ 264, 496, 61 },
		{ 264, 497, 61 },
		{ 264, 498, 61 },
		{ 264, 499, 61 },
		{ 264, 500, 61 },
		{ 264, 501, 61 },
		{ 264, 502, 61 },
		{ 264, 503, 61 },
		{ 264, 504, 61 },
		{ 264, 505, 61 },
		{ 264, 506, 61 },
		{ 264, 507, 61 },
		{ 264, 508, 61 },
		{ 264, 509, 61 },
		{ 264, 510, 61 },
		{ 264, 511, 61 },
		{ 264, 512, 61 },
		{ 265, 323, 61 },
		{ 265, 324, 61 },
		{ 265, 325, 61 },
		{ 265, 326, 61 },
		{ 265, 327, 61 },
		{ 265, 328, 61 },
		{ 265, 329, 61 },
		{ 265, 330, 61 },
		{ 265, 331, 61 },
		{ 265, 332, 61 },
		{ 265, 333, 61 },
		{ 265, 334, 61 },
		{ 265, 335, 61 },
		{ 265, 336, 61 },
		{ 265, 337, 61 },
		{ 265, 338, 61 },
		{ 265, 339, 61 },
		{ 265, 340, 61 },
		{ 265, 341, 61 },
		{ 265, 342, 61 },
		{ 265, 343, 61 },
		{ 265, 344, 61 },
		{ 265, 345, 61 },
		{ 265, 346, 61 },
		{ 265, 347, 61 },
		{ 265, 348, 61 },
		{ 265, 349, 61 },
		{ 265, 350, 61 },
		{ 265, 351, 61 },
		{ 265, 352, 61 },
		{ 265, 353, 61 },
		{ 265, 354, 61 },
		{ 265, 355, 61 },
		{ 265, 356, 61 },
		{ 265, 357, 61 },
		{ 265, 358, 61 },
		{ 265, 359, 61 },
		{ 265, 360, 61 },
		{ 265, 361, 61 },
		{ 265, 362, 61 },
		{ 265, 363, 61 },
		{ 265, 364, 61 },
		{ 265, 365, 61 },
		{ 265, 366, 61 },
		{ 265, 367, 61 },
		{ 265, 368, 61 },
		{ 265, 369, 61 },
		{ 265, 370, 61 },
		{ 265, 371, 61 },
		{ 265, 372, 61 },
		{ 265, 373, 61 },
		{ 265, 374, 61 },
		{ 265, 375, 61 },
		{ 265, 376, 61 },
		{ 265, 377, 61 },
		{ 265, 378, 61 },
		{ 265, 379, 61 },
		{ 265, 380, 61 },
		{ 265, 381, 61 },
		{ 265, 382, 61 },
		{ 265, 383, 61 },
		{ 265, 384, 61 },
		{ 265, 385, 61 },
		{ 265, 386, 61 },
		{ 265, 387, 61 },
		{ 265, 388, 61 },
		{ 265, 389, 61 },
		{ 265, 390, 61 },
		{ 265, 391, 61 },
		{ 265, 392, 61 },
		{ 265, 393, 61 },
		{ 265, 394, 61 },
		{ 265, 395, 61 },
		{ 265, 396, 61 },
		{ 265, 397, 61 },
		{ 265, 398, 61 },
		{ 265, 399, 61 },
		{ 265, 400, 61 },
		{ 265, 401, 61 },
		{ 265, 402, 61 },
		{ 265, 403, 61 },
		{ 265, 404, 61 },
		{ 265, 405, 61 },
		{ 265, 406, 61 },
		{ 265, 407, 61 },
		{ 265, 408, 61 },
		{ 265, 409, 61 },
		{ 265, 410, 61 },
		{ 265, 411, 61 },
		{ 265, 412, 61 },
		{ 265, 413, 61 },
		{ 265, 414, 61 },
		{ 265, 415, 61 },
		{ 265, 416, 61 },
		{ 265, 417, 61 },
		{ 265, 418, 61 },
		{ 265, 419, 61 },
		{ 265, 420, 61 },
		{ 265, 421, 61 },
		{ 265, 422, 61 },
		{ 265, 423, 61 },
		{ 265, 424, 61 },
		{ 265, 425, 61 },
		{ 265, 426, 61 },
		{ 265, 427, 61 },
		{ 265, 428, 61 },
		{ 265, 429, 61 },
		{ 265, 430, 61 },
		{ 265, 431, 61 },
		{ 265, 432, 61 },
		{ 265, 433, 61 },
		{ 265, 434, 61 },
		{ 265, 435, 61 },
		{ 265, 436, 61 },
		{ 265, 437, 61 },
		{ 265, 438, 61 },
		{ 265, 439, 61 },
		{ 265, 440, 61 },
		{ 265, 441, 61 },
		{ 265, 442, 61 },
		{ 265, 443, 61 },
		{ 265, 444, 61 },
		{ 265, 445, 61 },
		{ 265, 446, 61 },
		{ 265, 447, 61 },
		{ 265, 448, 61 },
		{ 265, 449, 61 },
		{ 265, 450, 61 },
		{ 265, 451, 61 },
		{ 265, 452, 61 },
		{ 265, 453, 61 },
		{ 265, 454, 61 },
		{ 265, 455, 61 },
		{ 265, 456, 61 },
		{ 265, 457, 61 },
		{ 265, 458, 61 },
		{ 265, 459, 61 },
		{ 265, 460, 61 },
		{ 265, 461, 61 },
		{ 265, 462, 61 },
		{ 265, 463, 61 },
		{ 265, 464, 61 },
		{ 265, 465, 61 },
		{ 265, 466, 61 },
		{ 265, 467, 61 },
		{ 265, 468, 61 },
		{ 265, 469, 61 },
		{ 265, 470, 61 },
		{ 265, 471, 61 },
		{ 265, 472, 61 },
		{ 265, 473, 61 },
		{ 265, 474, 61 },
		{ 265, 475, 61 },
		{ 265, 476, 61 },
		{ 265, 477, 61 },
		{ 265, 478, 61 },
		{ 265, 479, 61 },
		{ 265, 480, 61 },
		{ 265, 481, 61 },
		{ 265, 482, 61 },
		{ 265, 483, 61 },
		{ 265, 484, 61 },
		{ 265, 485, 61 },
		{ 265, 486, 61 },
		{ 265, 487, 61 },
		{ 265, 488, 61 },
		{ 265, 489, 61 },
		{ 265, 490, 61 },
		{ 265, 491, 61 },
		{ 265, 492, 61 },
		{ 265, 493, 61 },
		{ 265, 494, 61 },
		{ 265, 495, 61 },
		{ 265, 496, 61 },
		{ 265, 497, 61 },
		{ 265, 498, 61 },
		{ 265, 499, 61 },
		{ 265, 500, 61 },
		{ 265, 501, 61 },
		{ 265, 502, 61 },
		{ 265, 503, 61 },
		{ 265, 504, 61 },
		{ 265, 505, 61 },
		{ 265, 506, 61 },
		{ 265, 507, 61 },
		{ 265, 508, 61 },
		{ 265, 509, 61 },
		{ 265, 510, 61 },
		{ 265, 511, 61 },
		{ 265, 512, 61 },
		{ 266, 323, 61 },
		{ 266, 324, 61 },
		{ 266, 325, 61 },
		{ 266, 326, 61 },
		{ 266, 327, 61 },
		{ 266, 328, 61 },
		{ 266, 329, 61 },
		{ 266, 330, 61 },
		{ 266, 331, 61 },
		{ 266, 332, 61 },
		{ 266, 333, 61 },
		{ 266, 334, 61 },
		{ 266, 335, 61 },
		{ 266, 336, 61 },
		{ 266, 337, 61 },
		{ 266, 338, 61 },
		{ 266, 339, 61 },
		{ 266, 340, 61 },
		{ 266, 341, 61 },
		{ 266, 342, 61 },
		{ 266, 343, 61 },
		{ 266, 344, 61 },
		{ 266, 345, 61 },
		{ 266, 346, 61 },
		{ 266, 347, 61 },
		{ 266, 348, 61 },
		{ 266, 349, 61 },
		{ 266, 350, 61 },
		{ 266, 351, 61 },
		{ 266, 352, 61 },
		{ 266, 353, 61 },
		{ 266, 354, 61 },
		{ 266, 355, 61 },
		{ 266, 356, 61 },
		{ 266, 357, 61 },
		{ 266, 358, 61 },
		{ 266, 359, 61 },
		{ 266, 360, 61 },
		{ 266, 361, 61 },
		{ 266, 362, 61 },
		{ 266, 363, 61 },
		{ 266, 364, 61 },
		{ 266, 365, 61 },
		{ 266, 366, 61 },
		{ 266, 367, 61 },
		{ 266, 368, 61 },
		{ 266, 369, 61 },
		{ 266, 370, 61 },
		{ 266, 371, 61 },
		{ 266, 372, 61 },
		{ 266, 373, 61 },
		{ 266, 374, 61 },
		{ 266, 375, 61 },
		{ 266, 376, 61 },
		{ 266, 377, 61 },
		{ 266, 378, 61 },
		{ 266, 379, 61 },
		{ 266, 380, 61 },
		{ 266, 381, 61 },
		{ 266, 382, 61 },
		{ 266, 383, 61 },
		{ 266, 384, 61 },
		{ 266, 385, 61 },
		{ 266, 386, 61 },
		{ 266, 387, 61 },
		{ 266, 388, 61 },
		{ 266, 389, 61 },
		{ 266, 390, 61 },
		{ 266, 391, 61 },
		{ 266, 392, 61 },
		{ 266, 393, 61 },
		{ 266, 394, 61 },
		{ 266, 395, 61 },
		{ 266, 396, 61 },
		{ 266, 397, 61 },
		{ 266, 398, 61 },
		{ 266, 399, 61 },
		{ 266, 400, 61 },
		{ 266, 401, 61 },
		{ 266, 402, 61 },
		{ 266, 403, 61 },
		{ 266, 404, 61 },
		{ 266, 405, 61 },
		{ 266, 406, 61 },
		{ 266, 407, 61 },
		{ 266, 408, 61 },
		{ 266, 409, 61 },
		{ 266, 410, 61 },
		{ 266, 411, 61 },
		{ 266, 412, 61 },
		{ 266, 413, 61 },
		{ 266, 414, 61 },
		{ 266, 415, 61 },
		{ 266, 416, 61 },
		{ 266, 417, 61 },
		{ 266, 418, 61 },
		{ 266, 419, 61 },
		{ 266, 420, 61 },
		{ 266, 421, 61 },
		{ 266, 422, 61 },
		{ 266, 423, 61 },
		{ 266, 424, 61 },
		{ 266, 425, 61 },
		{ 266, 426, 61 },
		{ 266, 427, 61 },
		{ 266, 428, 61 },
		{ 266, 429, 61 },
		{ 266, 430, 61 },
		{ 266, 431, 61 },
		{ 266, 432, 61 },
		{ 266, 433, 61 },
		{ 266, 434, 61 },
		{ 266, 435, 61 },
		{ 266, 436, 61 },
		{ 266, 437, 61 },
		{ 266, 438, 61 },
		{ 266, 439, 61 },
		{ 266, 440, 61 },
		{ 266, 441, 61 },
		{ 266, 442, 61 },
		{ 266, 443, 61 },
		{ 266, 444, 61 },
		{ 266, 445, 61 },
		{ 266, 446, 61 },
		{ 266, 447, 61 },
		{ 266, 448, 61 },
		{ 266, 449, 61 },
		{ 266, 450, 61 },
		{ 266, 451, 61 },
		{ 266, 452, 61 },
		{ 266, 453, 61 },
		{ 266, 454, 61 },
		{ 266, 455, 61 },
		{ 266, 456, 61 },
		{ 266, 457, 61 },
		{ 266, 458, 61 },
		{ 266, 459, 61 },
		{ 266, 460, 61 },
		{ 266, 461, 61 },
		{ 266, 462, 61 },
		{ 266, 463, 61 },
		{ 266, 464, 61 },
		{ 266, 465, 61 },
		{ 266, 466, 61 },
		{ 266, 467, 61 },
		{ 266, 468, 61 },
		{ 266, 469, 61 },
		{ 266, 470, 61 },
		{ 266, 471, 61 },
		{ 266, 472, 61 },
		{ 266, 473, 61 },
		{ 266, 474, 61 },
		{ 266, 475, 61 },
		{ 266, 476, 61 },
		{ 266, 477, 61 },
		{ 266, 478, 61 },
		{ 266, 479, 61 },
		{ 266, 480, 61 },
		{ 266, 481, 61 },
		{ 266, 482, 61 },
		{ 266, 483, 61 },
		{ 266, 484, 61 },
		{ 266, 485, 61 },
		{ 266, 486, 61 },
		{ 266, 487, 61 },
		{ 266, 488, 61 },
		{ 266, 489, 61 },
		{ 266, 490, 61 },
		{ 266, 491, 61 },
		{ 266, 492, 61 },
		{ 266, 493, 61 },
		{ 266, 494, 61 },
		{ 266, 495, 61 },
		{ 266, 496, 61 },
		{ 266, 497, 61 },
		{ 266, 498, 61 },
		{ 266, 499, 61 },
		{ 266, 500, 61 },
		{ 266, 501, 61 },
		{ 266, 502, 61 },
		{ 266, 503, 61 },
		{ 266, 504, 61 },
		{ 266, 505, 61 },
		{ 266, 506, 61 },
		{ 266, 507, 61 },
		{ 266, 508, 61 },
		{ 266, 509, 61 },
		{ 266, 510, 61 },
		{ 266, 511, 61 },
		{ 266, 512, 61 },
		{ 267, 323, 61 },
		{ 267, 324, 61 },
		{ 267, 325, 61 },
		{ 267, 326, 61 },
		{ 267, 327, 61 },
		{ 267, 328, 61 },
		{ 267, 329, 61 },
		{ 267, 330, 61 },
		{ 267, 331, 61 },
		{ 267, 332, 61 },
		{ 267, 333, 61 },
		{ 267, 334, 61 },
		{ 267, 335, 61 },
		{ 267, 336, 61 },
		{ 267, 337, 61 },
		{ 267, 338, 61 },
		{ 267, 339, 61 },
		{ 267, 340, 61 },
		{ 267, 341, 61 },
		{ 267, 342, 61 },
		{ 267, 343, 61 },
		{ 267, 344, 61 },
		{ 267, 345, 61 },
		{ 267, 346, 61 },
		{ 267, 347, 61 },
		{ 267, 348, 61 },
		{ 267, 349, 61 },
		{ 267, 350, 61 },
		{ 267, 351, 61 },
		{ 267, 352, 61 },
		{ 267, 353, 61 },
		{ 267, 354, 61 },
		{ 267, 355, 61 },
		{ 267, 356, 61 },
		{ 267, 357, 61 },
		{ 267, 358, 61 },
		{ 267, 359, 61 },
		{ 267, 360, 61 },
		{ 267, 361, 61 },
		{ 267, 362, 61 },
		{ 267, 363, 61 },
		{ 267, 364, 61 },
		{ 267, 365, 61 },
		{ 267, 366, 61 },
		{ 267, 367, 61 },
		{ 267, 368, 61 },
		{ 267, 369, 61 },
		{ 267, 370, 61 },
		{ 267, 371, 61 },
		{ 267, 372, 61 },
		{ 267, 373, 61 },
		{ 267, 374, 61 },
		{ 267, 375, 61 },
		{ 267, 376, 61 },
		{ 267, 377, 61 },
		{ 267, 378, 61 },
		{ 267, 379, 61 },
		{ 267, 380, 61 },
		{ 267, 381, 61 },
		{ 267, 382, 61 },
		{ 267, 383, 61 },
		{ 267, 384, 61 },
		{ 267, 385, 61 },
		{ 267, 386, 61 },
		{ 267, 387, 61 },
		{ 267, 388, 61 },
		{ 267, 389, 61 },
		{ 267, 390, 61 },
		{ 267, 391, 61 },
		{ 267, 392, 61 },
		{ 267, 393, 61 },
		{ 267, 394, 61 },
		{ 267, 395, 61 },
		{ 267, 396, 61 },
		{ 267, 397, 61 },
		{ 267, 398, 61 },
		{ 267, 399, 61 },
		{ 267, 400, 61 },
		{ 267, 401, 61 },
		{ 267, 402, 61 },
		{ 267, 403, 61 },
		{ 267, 404, 61 },
		{ 267, 405, 61 },
		{ 267, 406, 61 },
		{ 267, 407, 61 },
		{ 267, 408, 61 },
		{ 267, 409, 61 },
		{ 267, 410, 61 },
		{ 267, 411, 61 },
		{ 267, 412, 61 },
		{ 267, 413, 61 },
		{ 267, 414, 61 },
		{ 267, 415, 61 },
		{ 267, 416, 61 },
		{ 267, 417, 61 },
		{ 267, 418, 61 },
		{ 267, 419, 61 },
		{ 267, 420, 61 },
		{ 267, 421, 61 },
		{ 267, 422, 61 },
		{ 267, 423, 61 },
		{ 267, 424, 61 },
		{ 267, 425, 61 },
		{ 267, 426, 61 },
		{ 267, 427, 61 },
		{ 267, 428, 61 },
		{ 267, 429, 61 },
		{ 267, 430, 61 },
		{ 267, 431, 61 },
		{ 267, 432, 61 },
		{ 267, 433, 61 },
		{ 267, 434, 61 },
		{ 267, 435, 61 },
		{ 267, 436, 61 },
		{ 267, 437, 61 },
		{ 267, 438, 61 },
		{ 267, 439, 61 },
		{ 267, 440, 61 },
		{ 267, 441, 61 },
		{ 267, 442, 61 },
		{ 267, 443, 61 },
		{ 267, 444, 61 },
		{ 267, 445, 61 },
		{ 267, 446, 61 },
		{ 267, 447, 61 },
		{ 267, 448, 61 },
		{ 267, 449, 61 },
		{ 267, 450, 61 },
		{ 267, 451, 61 },
		{ 267, 452, 61 },
		{ 267, 453, 61 },
		{ 267, 454, 61 },
		{ 267, 455, 61 },
		{ 267, 456, 61 },
		{ 267, 457, 61 },
		{ 267, 458, 61 },
		{ 267, 459, 61 },
		{ 267, 460, 61 },
		{ 267, 461, 61 },
		{ 267, 462, 61 },
		{ 267, 463, 61 },
		{ 267, 464, 61 },
		{ 267, 465, 61 },
		{ 267, 466, 61 },
		{ 267, 467, 61 },
		{ 267, 468, 61 },
		{ 267, 469, 61 },
		{ 267, 470, 61 },
		{ 267, 471, 61 },
		{ 267, 472, 61 },
		{ 267, 473, 61 },
		{ 267, 474, 61 },
		{ 267, 475, 61 },
		{ 267, 476, 61 },
		{ 267, 477, 61 },
		{ 267, 478, 61 },
		{ 267, 479, 61 },
		{ 267, 480, 61 },
		{ 267, 481, 61 },
		{ 267, 482, 61 },
		{ 267, 483, 61 },
		{ 267, 484, 61 },
		{ 267, 485, 61 },
		{ 267, 486, 61 },
		{ 267, 487, 61 },
		{ 267, 488, 61 },
		{ 267, 489, 61 },
		{ 267, 490, 61 },
		{ 267, 491, 61 },
		{ 267, 492, 61 },
		{ 267, 493, 61 },
		{ 267, 494, 61 },
		{ 267, 495, 61 },
		{ 267, 496, 61 },
		{ 267, 497, 61 },
		{ 267, 498, 61 },
		{ 267, 499, 61 },
		{ 267, 500, 61 },
		{ 267, 501, 61 },
		{ 267, 502, 61 },
		{ 267, 503, 61 },
		{ 267, 504, 61 },
		{ 267, 505, 61 },
		{ 267, 506, 61 },
		{ 267, 507, 61 },
		{ 267, 508, 61 },
		{ 267, 509, 61 },
		{ 267, 510, 61 },
		{ 267, 511, 61 },
		{ 267, 512, 61 },
		{ 268, 323, 61 },
		{ 268, 324, 61 },
		{ 268, 325, 61 },
		{ 268, 326, 61 },
		{ 268, 327, 61 },
		{ 268, 328, 61 },
		{ 268, 329, 61 },
		{ 268, 330, 61 },
		{ 268, 331, 61 },
		{ 268, 332, 61 },
		{ 268, 333, 61 },
		{ 268, 334, 61 },
		{ 268, 335, 61 },
		{ 268, 336, 61 },
		{ 268, 337, 61 },
		{ 268, 338, 61 },
		{ 268, 339, 61 },
		{ 268, 340, 61 },
		{ 268, 341, 61 },
		{ 268, 342, 61 },
		{ 268, 343, 61 },
		{ 268, 344, 61 },
		{ 268, 345, 61 },
		{ 268, 346, 61 },
		{ 268, 347, 61 },
		{ 268, 348, 61 },
		{ 268, 349, 61 },
		{ 268, 350, 61 },
		{ 268, 351, 61 },
		{ 268, 352, 61 },
		{ 268, 353, 61 },
		{ 268, 354, 61 },
		{ 268, 355, 61 },
		{ 268, 356, 61 },
		{ 268, 357, 61 },
		{ 268, 358, 61 },
		{ 268, 359, 61 },
		{ 268, 360, 61 },
		{ 268, 361, 61 },
		{ 268, 362, 61 },
		{ 268, 363, 61 },
		{ 268, 364, 61 },
		{ 268, 365, 61 },
		{ 268, 366, 61 },
		{ 268, 367, 61 },
		{ 268, 368, 61 },
		{ 268, 369, 61 },
		{ 268, 370, 61 },
		{ 268, 371, 61 },
		{ 268, 372, 61 },
		{ 268, 373, 61 },
		{ 268, 374, 61 },
		{ 268, 375, 61 },
		{ 268, 376, 61 },
		{ 268, 377, 61 },
		{ 268, 378, 61 },
		{ 268, 379, 61 },
		{ 268, 380, 61 },
		{ 268, 381, 61 },
		{ 268, 382, 61 },
		{ 268, 383, 61 },
		{ 268, 384, 61 },
		{ 268, 385, 61 },
		{ 268, 386, 61 },
		{ 268, 387, 61 },
		{ 268, 388, 61 },
		{ 268, 389, 61 },
		{ 268, 390, 61 },
		{ 268, 391, 61 },
		{ 268, 392, 61 },
		{ 268, 393, 61 },
		{ 268, 394, 61 },
		{ 268, 395, 61 },
		{ 268, 396, 61 },
		{ 268, 397, 61 },
		{ 268, 398, 61 },
		{ 268, 399, 61 },
		{ 268, 400, 61 },
		{ 268, 401, 61 },
		{ 268, 402, 61 },
		{ 268, 403, 61 },
		{ 268, 404, 61 },
		{ 268, 405, 61 },
		{ 268, 406, 61 },
		{ 268, 407, 61 },
		{ 268, 408, 61 },
		{ 268, 409, 61 },
		{ 268, 410, 61 },
		{ 268, 411, 61 },
		{ 268, 412, 61 },
		{ 268, 413, 61 },
		{ 268, 414, 61 },
		{ 268, 415, 61 },
		{ 268, 416, 61 },
		{ 268, 417, 61 },
		{ 268, 418, 61 },
		{ 268, 419, 61 },
		{ 268, 420, 61 },
		{ 268, 421, 61 },
		{ 268, 422, 61 },
		{ 268, 423, 61 },
		{ 268, 424, 61 },
		{ 268, 425, 61 },
		{ 268, 426, 61 },
		{ 268, 427, 61 },
		{ 268, 428, 61 },
		{ 268, 429, 61 },
		{ 268, 430, 61 },
		{ 268, 431, 61 },
		{ 268, 432, 61 },
		{ 268, 433, 61 },
		{ 268, 434, 61 },
		{ 268, 435, 61 },
		{ 268, 436, 61 },
		{ 268, 437, 61 },
		{ 268, 438, 61 },
		{ 268, 439, 61 },
		{ 268, 440, 61 },
		{ 268, 441, 61 },
		{ 268, 442, 61 },
		{ 268, 443, 61 },
		{ 268, 444, 61 },
		{ 268, 445, 61 },
		{ 268, 446, 61 },
		{ 268, 447, 61 },
		{ 268, 448, 61 },
		{ 268, 449, 61 },
		{ 268, 450, 61 },
		{ 268, 451, 61 },
		{ 268, 452, 61 },
		{ 268, 453, 61 },
		{ 268, 454, 61 },
		{ 268, 455, 61 },
		{ 268, 456, 61 },
		{ 268, 457, 61 },
		{ 268, 458, 61 },
		{ 268, 459, 61 },
		{ 268, 460, 61 },
		{ 268, 461, 61 },
		{ 268, 462, 61 },
		{ 268, 463, 61 },
		{ 268, 464, 61 },
		{ 268, 465, 61 },
		{ 268, 466, 61 },
		{ 268, 467, 61 },
		{ 268, 468, 61 },
		{ 268, 469, 61 },
		{ 268, 470, 61 },
		{ 268, 471, 61 },
		{ 268, 472, 61 },
		{ 268, 473, 61 },
		{ 268, 474, 61 },
		{ 268, 475, 61 },
		{ 268, 476, 61 },
		{ 268, 477, 61 },
		{ 268, 478, 61 },
		{ 268, 479, 61 },
		{ 268, 480, 61 },
		{ 268, 481, 61 },
		{ 268, 482, 61 },
		{ 268, 483, 61 },
		{ 268, 484, 61 },
		{ 268, 485, 61 },
		{ 268, 486, 61 },
		{ 268, 487, 61 },
		{ 268, 488, 61 },
		{ 268, 489, 61 },
		{ 268, 490, 61 },
		{ 268, 491, 61 },
		{ 268, 492, 61 },
		{ 268, 493, 61 },
		{ 268, 494, 61 },
		{ 268, 495, 61 },
		{ 268, 496, 61 },
		{ 268, 497, 61 },
		{ 268, 498, 61 },
		{ 268, 499, 61 },
		{ 268, 500, 61 },
		{ 268, 501, 61 },
		{ 268, 502, 61 },
		{ 268, 503, 61 },
		{ 268, 504, 61 },
		{ 268, 505, 61 },
		{ 268, 506, 61 },
		{ 268, 507, 61 },
		{ 268, 508, 61 },
		{ 268, 509, 61 },
		{ 268, 510, 61 },
		{ 268, 511, 61 },
		{ 268, 512, 61 },
		{ 269, 323, 61 },
		{ 269, 324, 61 },
		{ 269, 325, 61 },
		{ 269, 326, 61 },
		{ 269, 327, 61 },
		{ 269, 328, 61 },
		{ 269, 329, 61 },
		{ 269, 330, 61 },
		{ 269, 331, 61 },
		{ 269, 332, 61 },
		{ 269, 333, 61 },
		{ 269, 334, 61 },
		{ 269, 335, 61 },
		{ 269, 336, 61 },
		{ 269, 337, 61 },
		{ 269, 338, 61 },
		{ 269, 339, 61 },
		{ 269, 340, 61 },
		{ 269, 341, 61 },
		{ 269, 342, 61 },
		{ 269, 343, 61 },
		{ 269, 344, 61 },
		{ 269, 345, 61 },
		{ 269, 346, 61 },
		{ 269, 347, 61 },
		{ 269, 348, 61 },
		{ 269, 349, 61 },
		{ 269, 350, 61 },
		{ 269, 351, 61 },
		{ 269, 352, 61 },
		{ 269, 353, 61 },
		{ 269, 354, 61 },
		{ 269, 355, 61 },
		{ 269, 356, 61 },
		{ 269, 357, 61 },
		{ 269, 358, 61 },
		{ 269, 359, 61 },
		{ 269, 360, 61 },
		{ 269, 361, 61 },
		{ 269, 362, 61 },
		{ 269, 363, 61 },
		{ 269, 364, 61 },
		{ 269, 365, 61 },
		{ 269, 366, 61 },
		{ 269, 367, 61 },
		{ 269, 368, 61 },
		{ 269, 369, 61 },
		{ 269, 370, 61 },
		{ 269, 371, 61 },
		{ 269, 372, 61 },
		{ 269, 373, 61 },
		{ 269, 374, 61 },
		{ 269, 375, 61 },
		{ 269, 376, 61 },
		{ 269, 377, 61 },
		{ 269, 378, 61 },
		{ 269, 379, 61 },
		{ 269, 380, 61 },
		{ 269, 381, 61 },
		{ 269, 382, 61 },
		{ 269, 383, 61 },
		{ 269, 384, 61 },
		{ 269, 385, 61 },
		{ 269, 386, 61 },
		{ 269, 387, 61 },
		{ 269, 388, 61 },
		{ 269, 389, 61 },
		{ 269, 390, 61 },
		{ 269, 391, 61 },
		{ 269, 392, 61 },
		{ 269, 393, 61 },
		{ 269, 394, 61 },
		{ 269, 395, 61 },
		{ 269, 396, 61 },
		{ 269, 397, 61 },
		{ 269, 398, 61 },
		{ 269, 399, 61 },
		{ 269, 400, 61 },
		{ 269, 401, 61 },
		{ 269, 402, 61 },
		{ 269, 403, 61 },
		{ 269, 404, 61 },
		{ 269, 405, 61 },
		{ 269, 406, 61 },
		{ 269, 407, 61 },
		{ 269, 408, 61 },
		{ 269, 409, 61 },
		{ 269, 410, 61 },
		{ 269, 411, 61 },
		{ 269, 412, 61 },
		{ 269, 413, 61 },
		{ 269, 414, 61 },
		{ 269, 415, 61 },
		{ 269, 416, 61 },
		{ 269, 417, 61 },
		{ 269, 418, 61 },
		{ 269, 419, 61 },
		{ 269, 420, 61 },
		{ 269, 421, 61 },
		{ 269, 422, 61 },
		{ 269, 423, 61 },
		{ 269, 424, 61 },
		{ 269, 425, 61 },
		{ 269, 426, 61 },
		{ 269, 427, 61 },
		{ 269, 428, 61 },
		{ 269, 429, 61 },
		{ 269, 430, 61 },
		{ 269, 431, 61 },
		{ 269, 432, 61 },
		{ 269, 433, 61 },
		{ 269, 434, 61 },
		{ 269, 435, 61 },
		{ 269, 436, 61 },
		{ 269, 437, 61 },
		{ 269, 438, 61 },
		{ 269, 439, 61 },
		{ 269, 440, 61 },
		{ 269, 441, 61 },
		{ 269, 442, 61 },
		{ 269, 443, 61 },
		{ 269, 444, 61 },
		{ 269, 445, 61 },
		{ 269, 446, 61 },
		{ 269, 447, 61 },
		{ 269, 448, 61 },
		{ 269, 449, 61 },
		{ 269, 450, 61 },
		{ 269, 451, 61 },
		{ 269, 452, 61 },
		{ 269, 453, 61 },
		{ 269, 454, 61 },
		{ 269, 455, 61 },
		{ 269, 456, 61 },
		{ 269, 457, 61 },
		{ 269, 458, 61 },
		{ 269, 459, 61 },
		{ 269, 460, 61 },
		{ 269, 461, 61 },
		{ 269, 462, 61 },
		{ 269, 463, 61 },
		{ 269, 464, 61 },
		{ 269, 465, 61 },
		{ 269, 466, 61 },
		{ 269, 467, 61 },
		{ 269, 468, 61 },
		{ 269, 469, 61 },
		{ 269, 470, 61 },
		{ 269, 471, 61 },
		{ 269, 472, 61 },
		{ 269, 473, 61 },
		{ 269, 474, 61 },
		{ 269, 475, 61 },
		{ 269, 476, 61 },
		{ 269, 477, 61 },
		{ 269, 478, 61 },
		{ 269, 479, 61 },
		{ 269, 480, 61 },
		{ 269, 481, 61 },
		{ 269, 482, 61 },
		{ 269, 483, 61 },
		{ 269, 484, 61 },
		{ 269, 485, 61 },
		{ 269, 486, 61 },
		{ 269, 487, 61 },
		{ 269, 488, 61 },
		{ 269, 489, 61 },
		{ 269, 490, 61 },
		{ 269, 491, 61 },
		{ 269, 492, 61 },
		{ 269, 493, 61 },
		{ 269, 494, 61 },
		{ 269, 495, 61 },
		{ 269, 496, 61 },
		{ 269, 497, 61 },
		{ 269, 498, 61 },
		{ 269, 499, 61 },
		{ 269, 500, 61 },
		{ 269, 501, 61 },
		{ 269, 502, 61 },
		{ 269, 503, 61 },
		{ 269, 504, 61 },
		{ 269, 505, 61 },
		{ 269, 506, 61 },
		{ 269, 507, 61 },
		{ 269, 508, 61 },
		{ 269, 509, 61 },
		{ 269, 510, 61 },
		{ 269, 511, 61 },
		{ 269, 512, 61 },
		{ 270, 323, 61 },
		{ 270, 324, 61 },
		{ 270, 325, 61 },
		{ 270, 326, 61 },
		{ 270, 327, 61 },
		{ 270, 328, 61 },
		{ 270, 329, 61 },
		{ 270, 330, 61 },
		{ 270, 331, 61 },
		{ 270, 332, 61 },
		{ 270, 333, 61 },
		{ 270, 334, 61 },
		{ 270, 335, 61 },
		{ 270, 336, 61 },
		{ 270, 337, 61 },
		{ 270, 338, 61 },
		{ 270, 339, 61 },
		{ 270, 340, 61 },
		{ 270, 341, 61 },
		{ 270, 342, 61 },
		{ 270, 343, 61 },
		{ 270, 344, 61 },
		{ 270, 345, 61 },
		{ 270, 346, 61 },
		{ 270, 347, 61 },
		{ 270, 348, 61 },
		{ 270, 349, 61 },
		{ 270, 350, 61 },
		{ 270, 351, 61 },
		{ 270, 352, 61 },
		{ 270, 353, 61 },
		{ 270, 354, 61 },
		{ 270, 355, 61 },
		{ 270, 356, 61 },
		{ 270, 357, 61 },
		{ 270, 358, 61 },
		{ 270, 359, 61 },
		{ 270, 360, 61 },
		{ 270, 361, 61 },
		{ 270, 362, 61 },
		{ 270, 363, 61 },
		{ 270, 364, 61 },
		{ 270, 365, 61 },
		{ 270, 366, 61 },
		{ 270, 367, 61 },
		{ 270, 368, 61 },
		{ 270, 369, 61 },
		{ 270, 370, 61 },
		{ 270, 371, 61 },
		{ 270, 372, 61 },
		{ 270, 373, 61 },
		{ 270, 374, 61 },
		{ 270, 375, 61 },
		{ 270, 376, 61 },
		{ 270, 377, 61 },
		{ 270, 378, 61 },
		{ 270, 379, 61 },
		{ 270, 380, 61 },
		{ 270, 381, 61 },
		{ 270, 382, 61 },
		{ 270, 383, 61 },
		{ 270, 384, 61 },
		{ 270, 385, 61 },
		{ 270, 386, 61 },
		{ 270, 387, 61 },
		{ 270, 388, 61 },
		{ 270, 389, 61 },
		{ 270, 390, 61 },
		{ 270, 391, 61 },
		{ 270, 392, 61 },
		{ 270, 393, 61 },
		{ 270, 394, 61 },
		{ 270, 395, 61 },
		{ 270, 396, 61 },
		{ 270, 397, 61 },
		{ 270, 398, 61 },
		{ 270, 399, 61 },
		{ 270, 400, 61 },
		{ 270, 401, 61 },
		{ 270, 402, 61 },
		{ 270, 403, 61 },
		{ 270, 404, 61 },
		{ 270, 405, 61 },
		{ 270, 406, 61 },
		{ 270, 407, 61 },
		{ 270, 408, 61 },
		{ 270, 409, 61 },
		{ 270, 410, 61 },
		{ 270, 411, 61 },
		{ 270, 412, 61 },
		{ 270, 413, 61 },
		{ 270, 414, 61 },
		{ 270, 415, 61 },
		{ 270, 416, 61 },
		{ 270, 417, 61 },
		{ 270, 418, 61 },
		{ 270, 419, 61 },
		{ 270, 420, 61 },
		{ 270, 421, 61 },
		{ 270, 422, 61 },
		{ 270, 423, 61 },
		{ 270, 424, 61 },
		{ 270, 425, 61 },
		{ 270, 426, 61 },
		{ 270, 427, 61 },
		{ 270, 428, 61 },
		{ 270, 429, 61 },
		{ 270, 430, 61 },
		{ 270, 431, 61 },
		{ 270, 432, 61 },
		{ 270, 433, 61 },
		{ 270, 434, 61 },
		{ 270, 435, 61 },
		{ 270, 436, 61 },
		{ 270, 437, 61 },
		{ 270, 438, 61 },
		{ 270, 439, 61 },
		{ 270, 440, 61 },
		{ 270, 441, 61 },
		{ 270, 442, 61 },
		{ 270, 443, 61 },
		{ 270, 444, 61 },
		{ 270, 445, 61 },
		{ 270, 446, 61 },
		{ 270, 447, 61 },
		{ 270, 448, 61 },
		{ 270, 449, 61 },
		{ 270, 450, 61 },
		{ 270, 451, 61 },
		{ 270, 452, 61 },
		{ 270, 453, 61 },
		{ 270, 454, 61 },
		{ 270, 455, 61 },
		{ 270, 456, 61 },
		{ 270, 457, 61 },
		{ 270, 458, 61 },
		{ 270, 459, 61 },
		{ 270, 460, 61 },
		{ 270, 461, 61 },
		{ 270, 462, 61 },
		{ 270, 463, 61 },
		{ 270, 464, 61 },
		{ 270, 465, 61 },
		{ 270, 466, 61 },
		{ 270, 467, 61 },
		{ 270, 468, 61 },
		{ 270, 469, 61 },
		{ 270, 470, 61 },
		{ 270, 471, 61 },
		{ 270, 472, 61 },
		{ 270, 473, 61 },
		{ 270, 474, 61 },
		{ 270, 475, 61 },
		{ 270, 476, 61 },
		{ 270, 477, 61 },
		{ 270, 478, 61 },
		{ 270, 479, 61 },
		{ 270, 480, 61 },
		{ 270, 481, 61 },
		{ 270, 482, 61 },
		{ 270, 483, 61 },
		{ 270, 484, 61 },
		{ 270, 485, 61 },
		{ 270, 486, 61 },
		{ 270, 487, 61 },
		{ 270, 488, 61 },
		{ 270, 489, 61 },
		{ 270, 490, 61 },
		{ 270, 491, 61 },
		{ 270, 492, 61 },
		{ 270, 493, 61 },
		{ 270, 494, 61 },
		{ 270, 495, 61 },
		{ 270, 496, 61 },
		{ 270, 497, 61 },
		{ 270, 498, 61 },
		{ 270, 499, 61 },
		{ 270, 500, 61 },
		{ 270, 501, 61 },
		{ 270, 502, 61 },
		{ 270, 503, 61 },
		{ 270, 504, 61 },
		{ 270, 505, 61 },
		{ 270, 506, 61 },
		{ 270, 507, 61 },
		{ 270, 508, 61 },
		{ 270, 509, 61 },
		{ 270, 510, 61 },
		{ 270, 511, 61 },
		{ 270, 512, 61 },
		{ 271, 323, 61 },
		{ 271, 324, 61 },
		{ 271, 325, 61 },
		{ 271, 326, 61 },
		{ 271, 327, 61 },
		{ 271, 328, 61 },
		{ 271, 329, 61 },
		{ 271, 330, 61 },
		{ 271, 331, 61 },
		{ 271, 332, 61 },
		{ 271, 333, 61 },
		{ 271, 334, 61 },
		{ 271, 335, 61 },
		{ 271, 336, 61 },
		{ 271, 337, 61 },
		{ 271, 338, 61 },
		{ 271, 339, 61 },
		{ 271, 340, 61 },
		{ 271, 341, 61 },
		{ 271, 342, 61 },
		{ 271, 343, 61 },
		{ 271, 344, 61 },
		{ 271, 345, 61 },
		{ 271, 346, 61 },
		{ 271, 347, 61 },
		{ 271, 348, 61 },
		{ 271, 349, 61 },
		{ 271, 350, 61 },
		{ 271, 351, 61 },
		{ 271, 352, 61 },
		{ 271, 353, 61 },
		{ 271, 354, 61 },
		{ 271, 355, 61 },
		{ 271, 356, 61 },
		{ 271, 357, 61 },
		{ 271, 358, 61 },
		{ 271, 359, 61 },
		{ 271, 360, 61 },
		{ 271, 361, 61 },
		{ 271, 362, 61 },
		{ 271, 363, 61 },
		{ 271, 364, 61 },
		{ 271, 365, 61 },
		{ 271, 366, 61 },
		{ 271, 367, 61 },
		{ 271, 368, 61 },
		{ 271, 369, 61 },
		{ 271, 370, 61 },
		{ 271, 371, 61 },
		{ 271, 372, 61 },
		{ 271, 373, 61 },
		{ 271, 374, 61 },
		{ 271, 375, 61 },
		{ 271, 376, 61 },
		{ 271, 377, 61 },
		{ 271, 378, 61 },
		{ 271, 379, 61 },
		{ 271, 380, 61 },
		{ 271, 381, 61 },
		{ 271, 382, 61 },
		{ 271, 383, 61 },
		{ 271, 384, 61 },
		{ 271, 385, 61 },
		{ 271, 386, 61 },
		{ 271, 387, 61 },
		{ 271, 388, 61 },
		{ 271, 389, 61 },
		{ 271, 390, 61 },
		{ 271, 391, 61 },
		{ 271, 392, 61 },
		{ 271, 393, 61 },
		{ 271, 394, 61 },
		{ 271, 395, 61 },
		{ 271, 396, 61 },
		{ 271, 397, 61 },
		{ 271, 398, 61 },
		{ 271, 399, 61 },
		{ 271, 400, 61 },
		{ 271, 401, 61 },
		{ 271, 402, 61 },
		{ 271, 403, 61 },
		{ 271, 404, 61 },
		{ 271, 405, 61 },
		{ 271, 406, 61 },
		{ 271, 407, 61 },
		{ 271, 408, 61 },
		{ 271, 409, 61 },
		{ 271, 410, 61 },
		{ 271, 411, 61 },
		{ 271, 412, 61 },
		{ 271, 413, 61 },
		{ 271, 414, 61 },
		{ 271, 415, 61 },
		{ 271, 416, 61 },
		{ 271, 417, 61 },
		{ 271, 418, 61 },
		{ 271, 419, 61 },
		{ 271, 420, 61 },
		{ 271, 421, 61 },
		{ 271, 422, 61 },
		{ 271, 423, 61 },
		{ 271, 424, 61 },
		{ 271, 425, 61 },
		{ 271, 426, 61 },
		{ 271, 427, 61 },
		{ 271, 428, 61 },
		{ 271, 429, 61 },
		{ 271, 430, 61 },
		{ 271, 431, 61 },
		{ 271, 432, 61 },
		{ 271, 433, 61 },
		{ 271, 434, 61 },
		{ 271, 435, 61 },
		{ 271, 436, 61 },
		{ 271, 437, 61 },
		{ 271, 438, 61 },
		{ 271, 439, 61 },
		{ 271, 440, 61 },
		{ 271, 441, 61 },
		{ 271, 442, 61 },
		{ 271, 443, 61 },
		{ 271, 444, 61 },
		{ 271, 445, 61 },
		{ 271, 446, 61 },
		{ 271, 447, 61 },
		{ 271, 448, 61 },
		{ 271, 449, 61 },
		{ 271, 450, 61 },
		{ 271, 451, 61 },
		{ 271, 452, 61 },
		{ 271, 453, 61 },
		{ 271, 454, 61 },
		{ 271, 455, 61 },
		{ 271, 456, 61 },
		{ 271, 457, 61 },
		{ 271, 458, 61 },
		{ 271, 459, 61 },
		{ 271, 460, 61 },
		{ 271, 461, 61 },
		{ 271, 462, 61 },
		{ 271, 463, 61 },
		{ 271, 464, 61 },
		{ 271, 465, 61 },
		{ 271, 466, 61 },
		{ 271, 467, 61 },
		{ 271, 468, 61 },
		{ 271, 469, 61 },
		{ 271, 470, 61 },
		{ 271, 471, 61 },
		{ 271, 472, 61 },
		{ 271, 473, 61 },
		{ 271, 474, 61 },
		{ 271, 475, 61 },
		{ 271, 476, 61 },
		{ 271, 477, 61 },
		{ 271, 478, 61 },
		{ 271, 479, 61 },
		{ 271, 480, 61 },
		{ 271, 481, 61 },
		{ 271, 482, 61 },
		{ 271, 483, 61 },
		{ 271, 484, 61 },
		{ 271, 485, 61 },
		{ 271, 486, 61 },
		{ 271, 487, 61 },
		{ 271, 488, 61 },
		{ 271, 489, 61 },
		{ 271, 490, 61 },
		{ 271, 491, 61 },
		{ 271, 492, 61 },
		{ 271, 493, 61 },
		{ 271, 494, 61 },
		{ 271, 495, 61 },
		{ 271, 496, 61 },
		{ 271, 497, 61 },
		{ 271, 498, 61 },
		{ 271, 499, 61 },
		{ 271, 500, 61 },
		{ 271, 501, 61 },
		{ 271, 502, 61 },
		{ 271, 503, 61 },
		{ 271, 504, 61 },
		{ 271, 505, 61 },
		{ 271, 506, 61 },
		{ 271, 507, 61 },
		{ 271, 508, 61 },
		{ 271, 509, 61 },
		{ 271, 510, 61 },
		{ 271, 511, 61 },
		{ 271, 512, 61 },
		{ 272, 324, 61 },
		{ 272, 325, 61 },
		{ 272, 326, 61 },
		{ 272, 327, 61 },
		{ 272, 328, 61 },
		{ 272, 329, 61 },
		{ 272, 330, 61 },
		{ 272, 331, 61 },
		{ 272, 332, 61 },
		{ 272, 333, 61 },
		{ 272, 334, 61 },
		{ 272, 335, 61 },
		{ 272, 336, 61 },
		{ 272, 337, 61 },
		{ 272, 338, 61 },
		{ 272, 339, 61 },
		{ 272, 340, 61 },
		{ 272, 341, 61 },
		{ 272, 342, 61 },
		{ 272, 343, 61 },
		{ 272, 344, 61 },
		{ 272, 345, 61 },
		{ 272, 346, 61 },
		{ 272, 347, 61 },
		{ 272, 348, 61 },
		{ 272, 349, 61 },
		{ 272, 350, 61 },
		{ 272, 351, 61 },
		{ 272, 352, 61 },
		{ 272, 353, 61 },
		{ 272, 354, 61 },
		{ 272, 355, 61 },
		{ 272, 356, 61 },
		{ 272, 357, 61 },
		{ 272, 358, 61 },
		{ 272, 359, 61 },
		{ 272, 360, 61 },
		{ 272, 361, 61 },
		{ 272, 362, 61 },
		{ 272, 363, 61 },
		{ 272, 364, 61 },
		{ 272, 365, 61 },
		{ 272, 366, 61 },
		{ 272, 367, 61 },
		{ 272, 368, 61 },
		{ 272, 369, 61 },
		{ 272, 370, 61 },
		{ 272, 371, 61 },
		{ 272, 372, 61 },
		{ 272, 373, 61 },
		{ 272, 374, 61 },
		{ 272, 375, 61 },
		{ 272, 376, 61 },
		{ 272, 377, 61 },
		{ 272, 378, 61 },
		{ 272, 379, 61 },
		{ 272, 380, 61 },
		{ 272, 381, 61 },
		{ 272, 382, 61 },
		{ 272, 383, 61 },
		{ 272, 384, 61 },
		{ 272, 385, 61 },
		{ 272, 386, 61 },
		{ 272, 387, 61 },
		{ 272, 388, 61 },
		{ 272, 389, 61 },
		{ 272, 390, 61 },
		{ 272, 391, 61 },
		{ 272, 392, 61 },
		{ 272, 393, 61 },
		{ 272, 394, 61 },
		{ 272, 395, 61 },
		{ 272, 396, 61 },
		{ 272, 397, 61 },
		{ 272, 398, 61 },
		{ 272, 399, 61 },
		{ 272, 400, 61 },
		{ 272, 401, 61 },
		{ 272, 402, 61 },
		{ 272, 403, 61 },
		{ 272, 404, 61 },
		{ 272, 405, 61 },
		{ 272, 406, 61 },
		{ 272, 407, 61 },
		{ 272, 408, 61 },
		{ 272, 409, 61 },
		{ 272, 410, 61 },
		{ 272, 411, 61 },
		{ 272, 412, 61 },
		{ 272, 413, 61 },
		{ 272, 414, 61 },
		{ 272, 415, 61 },
		{ 272, 416, 61 },
		{ 272, 417, 61 },
		{ 272, 418, 61 },
		{ 272, 419, 61 },
		{ 272, 420, 61 },
		{ 272, 421, 61 },
		{ 272, 422, 61 },
		{ 272, 423, 61 },
		{ 272, 424, 61 },
		{ 272, 425, 61 },
		{ 272, 426, 61 },
		{ 272, 427, 61 },
		{ 272, 428, 61 },
		{ 272, 429, 61 },
		{ 272, 430, 61 },
		{ 272, 431, 61 },
		{ 272, 432, 61 },
		{ 272, 433, 61 },
		{ 272, 434, 61 },
		{ 272, 435, 61 },
		{ 272, 436, 61 },
		{ 272, 437, 61 },
		{ 272, 438, 61 },
		{ 272, 439, 61 },
		{ 272, 440, 61 },
		{ 272, 441, 61 },
		{ 272, 442, 61 },
		{ 272, 443, 61 },
		{ 272, 444, 61 },
		{ 272, 445, 61 },
		{ 272, 446, 61 },
		{ 272, 447, 61 },
		{ 272, 448, 61 },
		{ 272, 449, 61 },
		{ 272, 450, 61 },
		{ 272, 451, 61 },
		{ 272, 452, 61 },
		{ 272, 453, 61 },
		{ 272, 454, 61 },
		{ 272, 455, 61 },
		{ 272, 456, 61 },
		{ 272, 457, 61 },
		{ 272, 458, 61 },
		{ 272, 459, 61 },
		{ 272, 460, 61 },
		{ 272, 461, 61 },
		{ 272, 462, 61 },
		{ 272, 463, 61 },
		{ 272, 464, 61 },
		{ 272, 465, 61 },
		{ 272, 466, 61 },
		{ 272, 467, 61 },
		{ 272, 468, 61 },
		{ 272, 469, 61 },
		{ 272, 470, 61 },
		{ 272, 471, 61 },
		{ 272, 472, 61 },
		{ 272, 473, 61 },
		{ 272, 474, 61 },
		{ 272, 475, 61 },
		{ 272, 476, 61 },
		{ 272, 477, 61 },
		{ 272, 478, 61 },
		{ 272, 479, 61 },
		{ 272, 480, 61 },
		{ 272, 481, 61 },
		{ 272, 482, 61 },
		{ 272, 483, 61 },
		{ 272, 484, 61 },
		{ 272, 485, 61 },
		{ 272, 486, 61 },
		{ 272, 487, 61 },
		{ 272, 488, 61 },
		{ 272, 489, 61 },
		{ 272, 490, 61 },
		{ 272, 491, 61 },
		{ 272, 492, 61 },
		{ 272, 493, 61 },
		{ 272, 494, 61 },
		{ 272, 495, 61 },
		{ 272, 496, 61 },
		{ 272, 497, 61 },
		{ 272, 498, 61 },
		{ 272, 499, 61 },
		{ 272, 500, 61 },
		{ 272, 501, 61 },
		{ 272, 502, 61 },
		{ 272, 503, 61 },
		{ 272, 504, 61 },
		{ 272, 505, 61 },
		{ 272, 506, 61 },
		{ 272, 507, 61 },
		{ 272, 508, 61 },
		{ 272, 509, 61 },
		{ 272, 510, 61 },
		{ 272, 511, 61 },
		{ 272, 512, 61 },
		{ 273, 324, 61 },
		{ 273, 325, 61 },
		{ 273, 326, 61 },
		{ 273, 327, 61 },
		{ 273, 328, 61 },
		{ 273, 329, 61 },
		{ 273, 330, 61 },
		{ 273, 331, 61 },
		{ 273, 332, 61 },
		{ 273, 333, 61 },
		{ 273, 334, 61 },
		{ 273, 335, 61 },
		{ 273, 336, 61 },
		{ 273, 337, 61 },
		{ 273, 338, 61 },
		{ 273, 339, 61 },
		{ 273, 340, 61 },
		{ 273, 341, 61 },
		{ 273, 342, 61 },
		{ 273, 343, 61 },
		{ 273, 344, 61 },
		{ 273, 345, 61 },
		{ 273, 346, 61 },
		{ 273, 347, 61 },
		{ 273, 348, 61 },
		{ 273, 349, 61 },
		{ 273, 350, 61 },
		{ 273, 351, 61 },
		{ 273, 352, 61 },
		{ 273, 353, 61 },
		{ 273, 354, 61 },
		{ 273, 355, 61 },
		{ 273, 356, 61 },
		{ 273, 357, 61 },
		{ 273, 358, 61 },
		{ 273, 359, 61 },
		{ 273, 360, 61 },
		{ 273, 361, 61 },
		{ 273, 362, 61 },
		{ 273, 363, 61 },
		{ 273, 364, 61 },
		{ 273, 365, 61 },
		{ 273, 366, 61 },
		{ 273, 367, 61 },
		{ 273, 368, 61 },
		{ 273, 369, 61 },
		{ 273, 370, 61 },
		{ 273, 371, 61 },
		{ 273, 372, 61 },
		{ 273, 373, 61 },
		{ 273, 374, 61 },
		{ 273, 375, 61 },
		{ 273, 376, 61 },
		{ 273, 377, 61 },
		{ 273, 378, 61 },
		{ 273, 379, 61 },
		{ 273, 380, 61 },
		{ 273, 381, 61 },
		{ 273, 382, 61 },
		{ 273, 383, 61 },
		{ 273, 384, 61 },
		{ 273, 385, 61 },
		{ 273, 386, 61 },
		{ 273, 387, 61 },
		{ 273, 388, 61 },
		{ 273, 389, 61 },
		{ 273, 390, 61 },
		{ 273, 391, 61 },
		{ 273, 392, 61 },
		{ 273, 393, 61 },
		{ 273, 394, 61 },
		{ 273, 395, 61 },
		{ 273, 396, 61 },
		{ 273, 397, 61 },
		{ 273, 398, 61 },
		{ 273, 399, 61 },
		{ 273, 400, 61 },
		{ 273, 401, 61 },
		{ 273, 402, 61 },
		{ 273, 403, 61 },
		{ 273, 404, 61 },
		{ 273, 405, 61 },
		{ 273, 406, 61 },
		{ 273, 407, 61 },
		{ 273, 408, 61 },
		{ 273, 409, 61 },
		{ 273, 410, 61 },
		{ 273, 411, 61 },
		{ 273, 412, 61 },
		{ 273, 413, 61 },
		{ 273, 414, 61 },
		{ 273, 415, 61 },
		{ 273, 416, 61 },
		{ 273, 417, 61 },
		{ 273, 418, 61 },
		{ 273, 419, 61 },
		{ 273, 420, 61 },
		{ 273, 421, 61 },
		{ 273, 422, 61 },
		{ 273, 423, 61 },
		{ 273, 424, 61 },
		{ 273, 425, 61 },
		{ 273, 426, 61 },
		{ 273, 427, 61 },
		{ 273, 428, 61 },
		{ 273, 429, 61 },
		{ 273, 430, 61 },
		{ 273, 431, 61 },
		{ 273, 432, 61 },
		{ 273, 433, 61 },
		{ 273, 434, 61 },
		{ 273, 435, 61 },
		{ 273, 436, 61 },
		{ 273, 437, 61 },
		{ 273, 438, 61 },
		{ 273, 439, 61 },
		{ 273, 440, 61 },
		{ 273, 441, 61 },
		{ 273, 442, 61 },
		{ 273, 443, 61 },
		{ 273, 444, 61 },
		{ 273, 445, 61 },
		{ 273, 446, 61 },
		{ 273, 447, 61 },
		{ 273, 448, 61 },
		{ 273, 449, 61 },
		{ 273, 450, 61 },
		{ 273, 451, 61 },
		{ 273, 452, 61 },
		{ 273, 453, 61 },
		{ 273, 454, 61 },
		{ 273, 455, 61 },
		{ 273, 456, 61 },
		{ 273, 457, 61 },
		{ 273, 458, 61 },
		{ 273, 459, 61 },
		{ 273, 460, 61 },
		{ 273, 461, 61 },
		{ 273, 462, 61 },
		{ 273, 463, 61 },
		{ 273, 464, 61 },
		{ 273, 465, 61 },
		{ 273, 466, 61 },
		{ 273, 467, 61 },
		{ 273, 468, 61 },
		{ 273, 469, 61 },
		{ 273, 470, 61 },
		{ 273, 471, 61 },
		{ 273, 472, 61 },
		{ 273, 473, 61 },
		{ 273, 474, 61 },
		{ 273, 475, 61 },
		{ 273, 476, 61 },
		{ 273, 477, 61 },
		{ 273, 478, 61 },
		{ 273, 479, 61 },
		{ 273, 480, 61 },
		{ 273, 481, 61 },
		{ 273, 482, 61 },
		{ 273, 483, 61 },
		{ 273, 484, 61 },
		{ 273, 485, 61 },
		{ 273, 486, 61 },
		{ 273, 487, 61 },
		{ 273, 488, 61 },
		{ 273, 489, 61 },
		{ 273, 490, 61 },
		{ 273, 491, 61 },
		{ 273, 492, 61 },
		{ 273, 493, 61 },
		{ 273, 494, 61 },
		{ 273, 495, 61 },
		{ 273, 496, 61 },
		{ 273, 497, 61 },
		{ 273, 498, 61 },
		{ 273, 499, 61 },
		{ 273, 500, 61 },
		{ 273, 501, 61 },
		{ 273, 502, 61 },
		{ 273, 503, 61 },
		{ 273, 504, 61 },
		{ 273, 505, 61 },
		{ 273, 506, 61 },
		{ 273, 507, 61 },
		{ 273, 508, 61 },
		{ 273, 509, 61 },
		{ 273, 510, 61 },
		{ 273, 511, 61 },
		{ 273, 512, 61 },
		{ 274, 327, 61 },
		{ 274, 328, 61 },
		{ 274, 329, 61 },
		{ 274, 330, 61 },
		{ 274, 331, 61 },
		{ 274, 332, 61 },
		{ 274, 333, 61 },
		{ 274, 334, 61 },
		{ 274, 335, 61 },
		{ 274, 336, 61 },
		{ 274, 337, 61 },
		{ 274, 338, 61 },
		{ 274, 339, 61 },
		{ 274, 340, 61 },
		{ 274, 341, 61 },
		{ 274, 342, 61 },
		{ 274, 343, 61 },
		{ 274, 344, 61 },
		{ 274, 345, 61 },
		{ 274, 346, 61 },
		{ 274, 347, 61 },
		{ 274, 348, 61 },
		{ 274, 349, 61 },
		{ 274, 350, 61 },
		{ 274, 351, 61 },
		{ 274, 352, 61 },
		{ 274, 353, 61 },
		{ 274, 354, 61 },
		{ 274, 355, 61 },
		{ 274, 356, 61 },
		{ 274, 357, 61 },
		{ 274, 358, 61 },
		{ 274, 359, 61 },
		{ 274, 360, 61 },
		{ 274, 361, 61 },
		{ 274, 362, 61 },
		{ 274, 363, 61 },
		{ 274, 364, 61 },
		{ 274, 365, 61 },
		{ 274, 366, 61 },
		{ 274, 367, 61 },
		{ 274, 368, 61 },
		{ 274, 369, 61 },
		{ 274, 370, 61 },
		{ 274, 371, 61 },
		{ 274, 372, 61 },
		{ 274, 373, 61 },
		{ 274, 374, 61 },
		{ 274, 375, 61 },
		{ 274, 376, 61 },
		{ 274, 377, 61 },
		{ 274, 378, 61 },
		{ 274, 379, 61 },
		{ 274, 380, 61 },
		{ 274, 381, 61 },
		{ 274, 382, 61 },
		{ 274, 383, 61 },
		{ 274, 384, 61 },
		{ 274, 385, 61 },
		{ 274, 386, 61 },
		{ 274, 387, 61 },
		{ 274, 388, 61 },
		{ 274, 389, 61 },
		{ 274, 390, 61 },
		{ 274, 391, 61 },
		{ 274, 392, 61 },
		{ 274, 393, 61 },
		{ 274, 394, 61 },
		{ 274, 395, 61 },
		{ 274, 396, 61 },
		{ 274, 397, 61 },
		{ 274, 398, 61 },
		{ 274, 399, 61 },
		{ 274, 400, 61 },
		{ 274, 401, 61 },
		{ 274, 402, 61 },
		{ 274, 403, 61 },
		{ 274, 404, 61 },
		{ 274, 405, 61 },
		{ 274, 406, 61 },
		{ 274, 407, 61 },
		{ 274, 408, 61 },
		{ 274, 409, 61 },
		{ 274, 410, 61 },
		{ 274, 411, 61 },
		{ 274, 412, 61 },
		{ 274, 413, 61 },
		{ 274, 414, 61 },
		{ 274, 415, 61 },
		{ 274, 416, 61 },
		{ 274, 417, 61 },
		{ 274, 418, 61 },
		{ 274, 419, 61 },
		{ 274, 420, 61 },
		{ 274, 421, 61 },
		{ 274, 422, 61 },
		{ 274, 423, 61 },
		{ 274, 424, 61 },
		{ 274, 425, 61 },
		{ 274, 426, 61 },
		{ 274, 427, 61 },
		{ 274, 428, 61 },
		{ 274, 429, 61 },
		{ 274, 430, 61 },
		{ 274, 431, 61 },
		{ 274, 432, 61 },
		{ 274, 433, 61 },
		{ 274, 434, 61 },
		{ 274, 435, 61 },
		{ 274, 436, 61 },
		{ 274, 437, 61 },
		{ 274, 438, 61 },
		{ 274, 439, 61 },
		{ 274, 440, 61 },
		{ 274, 441, 61 },
		{ 274, 442, 61 },
		{ 274, 443, 61 },
		{ 274, 444, 61 },
		{ 274, 445, 61 },
		{ 274, 446, 61 },
		{ 274, 447, 61 },
		{ 274, 448, 61 },
		{ 274, 449, 61 },
		{ 274, 450, 61 },
		{ 274, 451, 61 },
		{ 274, 452, 61 },
		{ 274, 453, 61 },
		{ 274, 454, 61 },
		{ 274, 455, 61 },
		{ 274, 456, 61 },
		{ 274, 457, 61 },
		{ 274, 458, 61 },
		{ 274, 459, 61 },
		{ 274, 460, 61 },
		{ 274, 461, 61 },
		{ 274, 462, 61 },
		{ 274, 463, 61 },
		{ 274, 464, 61 },
		{ 274, 465, 61 },
		{ 274, 466, 61 },
		{ 274, 467, 61 },
		{ 274, 468, 61 },
		{ 274, 469, 61 },
		{ 274, 470, 61 },
		{ 274, 471, 61 },
		{ 274, 472, 61 },
		{ 274, 473, 61 },
		{ 274, 474, 61 },
		{ 274, 475, 61 },
		{ 274, 476, 61 },
		{ 274, 477, 61 },
		{ 274, 478, 61 },
		{ 274, 479, 61 },
		{ 274, 480, 61 },
		{ 274, 481, 61 },
		{ 274, 482, 61 },
		{ 274, 483, 61 },
		{ 274, 484, 61 },
		{ 274, 485, 61 },
		{ 274, 486, 61 },
		{ 274, 487, 61 },
		{ 274, 488, 61 },
		{ 274, 489, 61 },
		{ 274, 490, 61 },
		{ 274, 491, 61 },
		{ 274, 492, 61 },
		{ 274, 493, 61 },
		{ 274, 494, 61 },
		{ 274, 495, 61 },
		{ 274, 496, 61 },
		{ 274, 497, 61 },
		{ 274, 498, 61 },
		{ 274, 499, 61 },
		{ 274, 500, 61 },
		{ 274, 501, 61 },
		{ 274, 502, 61 },
		{ 274, 503, 61 },
		{ 274, 504, 61 },
		{ 274, 505, 61 },
		{ 274, 506, 61 },
		{ 274, 507, 61 },
		{ 274, 508, 61 },
		{ 274, 509, 61 },
		{ 274, 510, 61 },
		{ 274, 511, 61 },
		{ 274, 512, 61 },
		{ 275, 330, 61 },
		{ 275, 331, 61 },
		{ 275, 332, 61 },
		{ 275, 333, 61 },
		{ 275, 334, 61 },
		{ 275, 335, 61 },
		{ 275, 336, 61 },
		{ 275, 337, 61 },
		{ 275, 338, 61 },
		{ 275, 339, 61 },
		{ 275, 340, 61 },
		{ 275, 341, 61 },
		{ 275, 342, 61 },
		{ 275, 343, 61 },
		{ 275, 344, 61 },
		{ 275, 345, 61 },
		{ 275, 346, 61 },
		{ 275, 347, 61 },
		{ 275, 348, 61 },
		{ 275, 349, 61 },
		{ 275, 350, 61 },
		{ 275, 351, 61 },
		{ 275, 352, 61 },
		{ 275, 353, 61 },
		{ 275, 354, 61 },
		{ 275, 355, 61 },
		{ 275, 356, 61 },
		{ 275, 357, 61 },
		{ 275, 358, 61 },
		{ 275, 359, 61 },
		{ 275, 360, 61 },
		{ 275, 361, 61 },
		{ 275, 362, 61 },
		{ 275, 363, 61 },
		{ 275, 364, 61 },
		{ 275, 365, 61 },
		{ 275, 366, 61 },
		{ 275, 367, 61 },
		{ 275, 368, 61 },
		{ 275, 369, 61 },
		{ 275, 370, 61 },
		{ 275, 371, 61 },
		{ 275, 372, 61 },
		{ 275, 373, 61 },
		{ 275, 374, 61 },
		{ 275, 375, 61 },
		{ 275, 376, 61 },
		{ 275, 377, 61 },
		{ 275, 378, 61 },
		{ 275, 379, 61 },
		{ 275, 380, 61 },
		{ 275, 381, 61 },
		{ 275, 382, 61 },
		{ 275, 383, 61 },
		{ 275, 384, 61 },
		{ 275, 385, 61 },
		{ 275, 386, 61 },
		{ 275, 387, 61 },
		{ 275, 388, 61 },
		{ 275, 389, 61 },
		{ 275, 390, 61 },
		{ 275, 391, 61 },
		{ 275, 392, 61 },
		{ 275, 393, 61 },
		{ 275, 394, 61 },
		{ 275, 395, 61 },
		{ 275, 396, 61 },
		{ 275, 397, 61 },
		{ 275, 398, 61 },
		{ 275, 399, 61 },
		{ 275, 400, 61 },
		{ 275, 401, 61 },
		{ 275, 402, 61 },
		{ 275, 403, 61 },
		{ 275, 404, 61 },
		{ 275, 405, 61 },
		{ 275, 406, 61 },
		{ 275, 407, 61 },
		{ 275, 408, 61 },
		{ 275, 409, 61 },
		{ 275, 410, 61 },
		{ 275, 411, 61 },
		{ 275, 412, 61 },
		{ 275, 413, 61 },
		{ 275, 414, 61 },
		{ 275, 415, 61 },
		{ 275, 416, 61 },
		{ 275, 417, 61 },
		{ 275, 418, 61 },
		{ 275, 419, 61 },
		{ 275, 420, 61 },
		{ 275, 421, 61 },
		{ 275, 422, 61 },
		{ 275, 423, 61 },
		{ 275, 424, 61 },
		{ 275, 425, 61 },
		{ 275, 426, 61 },
		{ 275, 427, 61 },
		{ 275, 428, 61 },
		{ 275, 429, 61 },
		{ 275, 430, 61 },
		{ 275, 431, 61 },
		{ 275, 432, 61 },
		{ 275, 433, 61 },
		{ 275, 434, 61 },
		{ 275, 435, 61 },
		{ 275, 436, 61 },
		{ 275, 437, 61 },
		{ 275, 438, 61 },
		{ 275, 439, 61 },
		{ 275, 440, 61 },
		{ 275, 441, 61 },
		{ 275, 442, 61 },
		{ 275, 443, 61 },
		{ 275, 444, 61 },
		{ 275, 445, 61 },
		{ 275, 446, 61 },
		{ 275, 447, 61 },
		{ 275, 448, 61 },
		{ 275, 449, 61 },
		{ 275, 450, 61 },
		{ 275, 451, 61 },
		{ 275, 452, 61 },
		{ 275, 453, 61 },
		{ 275, 454, 61 },
		{ 275, 455, 61 },
		{ 275, 456, 61 },
		{ 275, 457, 61 },
		{ 275, 458, 61 },
		{ 275, 459, 61 },
		{ 275, 460, 61 },
		{ 275, 461, 61 },
		{ 275, 462, 61 },
		{ 275, 463, 61 },
		{ 275, 464, 61 },
		{ 275, 465, 61 },
		{ 275, 466, 61 },
		{ 275, 467, 61 },
		{ 275, 468, 61 },
		{ 275, 469, 61 },
		{ 275, 470, 61 },
		{ 275, 471, 61 },
		{ 275, 472, 61 },
		{ 275, 473, 61 },
		{ 275, 474, 61 },
		{ 275, 475, 61 },
		{ 275, 476, 61 },
		{ 275, 477, 61 },
		{ 275, 478, 61 },
		{ 275, 479, 61 },
		{ 275, 480, 61 },
		{ 275, 481, 61 },
		{ 275, 482, 61 },
		{ 275, 483, 61 },
		{ 275, 484, 61 },
		{ 275, 485, 61 },
		{ 275, 486, 61 },
		{ 275, 487, 61 },
		{ 275, 488, 61 },
		{ 275, 489, 61 },
		{ 275, 490, 61 },
		{ 275, 491, 61 },
		{ 275, 492, 61 },
		{ 275, 493, 61 },
		{ 275, 494, 61 },
		{ 275, 495, 61 },
		{ 275, 496, 61 },
		{ 275, 497, 61 },
		{ 275, 498, 61 },
		{ 275, 499, 61 },
		{ 275, 500, 61 },
		{ 275, 501, 61 },
		{ 275, 502, 61 },
		{ 275, 503, 61 },
		{ 275, 504, 61 },
		{ 275, 505, 61 },
		{ 275, 506, 61 },
		{ 275, 507, 61 },
		{ 275, 508, 61 },
		{ 275, 509, 61 },
		{ 275, 510, 61 },
		{ 275, 511, 61 },
		{ 275, 512, 61 },
		{ 276, 334, 61 },
		{ 276, 335, 61 },
		{ 276, 336, 61 },
		{ 276, 337, 61 },
		{ 276, 338, 61 },
		{ 276, 339, 61 },
		{ 276, 340, 61 },
		{ 276, 341, 61 },
		{ 276, 342, 61 },
		{ 276, 343, 61 },
		{ 276, 344, 61 },
		{ 276, 345, 61 },
		{ 276, 346, 61 },
		{ 276, 347, 61 },
		{ 276, 348, 61 },
		{ 276, 349, 61 },
		{ 276, 350, 61 },
		{ 276, 351, 61 },
		{ 276, 352, 61 },
		{ 276, 353, 61 },
		{ 276, 354, 61 },
		{ 276, 355, 61 },
		{ 276, 356, 61 },
		{ 276, 357, 61 },
		{ 276, 358, 61 },
		{ 276, 359, 61 },
		{ 276, 360, 61 },
		{ 276, 361, 61 },
		{ 276, 362, 61 },
		{ 276, 363, 61 },
		{ 276, 364, 61 },
		{ 276, 365, 61 },
		{ 276, 366, 61 },
		{ 276, 367, 61 },
		{ 276, 368, 61 },
		{ 276, 369, 61 },
		{ 276, 370, 61 },
		{ 276, 371, 61 },
		{ 276, 372, 61 },
		{ 276, 373, 61 },
		{ 276, 374, 61 },
		{ 276, 375, 61 },
		{ 276, 376, 61 },
		{ 276, 377, 61 },
		{ 276, 378, 61 },
		{ 276, 379, 61 },
		{ 276, 380, 61 },
		{ 276, 381, 61 },
		{ 276, 382, 61 },
		{ 276, 383, 61 },
		{ 276, 384, 61 },
		{ 276, 385, 61 },
		{ 276, 386, 61 },
		{ 276, 387, 61 },
		{ 276, 388, 61 },
		{ 276, 389, 61 },
		{ 276, 390, 61 },
		{ 276, 391, 61 },
		{ 276, 392, 61 },
		{ 276, 393, 61 },
		{ 276, 394, 61 },
		{ 276, 395, 61 },
		{ 276, 396, 61 },
		{ 276, 397, 61 },
		{ 276, 398, 61 },
		{ 276, 399, 61 },
		{ 276, 400, 61 },
		{ 276, 401, 61 },
		{ 276, 402, 61 },
		{ 276, 403, 61 },
		{ 276, 404, 61 },
		{ 276, 405, 61 },
		{ 276, 406, 61 },
		{ 276, 407, 61 },
		{ 276, 408, 61 },
		{ 276, 409, 61 },
		{ 276, 410, 61 },
		{ 276, 411, 61 },
		{ 276, 412, 61 },
		{ 276, 413, 61 },
		{ 276, 414, 61 },
		{ 276, 415, 61 },
		{ 276, 416, 61 },
		{ 276, 417, 61 },
		{ 276, 418, 61 },
		{ 276, 419, 61 },
		{ 276, 420, 61 },
		{ 276, 421, 61 },
		{ 276, 422, 61 },
		{ 276, 423, 61 },
		{ 276, 424, 61 },
		{ 276, 425, 61 },
		{ 276, 426, 61 },
		{ 276, 427, 61 },
		{ 276, 428, 61 },
		{ 276, 429, 61 },
		{ 276, 430, 61 },
		{ 276, 431, 61 },
		{ 276, 432, 61 },
		{ 276, 433, 61 },
		{ 276, 434, 61 },
		{ 276, 435, 61 },
		{ 276, 436, 61 },
		{ 276, 437, 61 },
		{ 276, 438, 61 },
		{ 276, 439, 61 },
		{ 276, 440, 61 },
		{ 276, 441, 61 },
		{ 276, 442, 61 },
		{ 276, 443, 61 },
		{ 276, 444, 61 },
		{ 276, 445, 61 },
		{ 276, 446, 61 },
		{ 276, 447, 61 },
		{ 276, 448, 61 },
		{ 276, 449, 61 },
		{ 276, 450, 61 },
		{ 276, 451, 61 },
		{ 276, 452, 61 },
		{ 276, 453, 61 },
		{ 276, 454, 61 },
		{ 276, 455, 61 },
		{ 276, 456, 61 },
		{ 276, 457, 61 },
		{ 276, 458, 61 },
		{ 276, 459, 61 },
		{ 276, 460, 61 },
		{ 276, 461, 61 },
		{ 276, 462, 61 },
		{ 276, 463, 61 },
		{ 276, 464, 61 },
		{ 276, 465, 61 },
		{ 276, 466, 61 },
		{ 276, 467, 61 },
		{ 276, 468, 61 },
		{ 276, 469, 61 },
		{ 276, 470, 61 },
		{ 276, 471, 61 },
		{ 276, 472, 61 },
		{ 276, 473, 61 },
		{ 276, 474, 61 },
		{ 276, 475, 61 },
		{ 276, 476, 61 },
		{ 276, 477, 61 },
		{ 276, 478, 61 },
		{ 276, 479, 61 },
		{ 276, 480, 61 },
		{ 276, 481, 61 },
		{ 276, 482, 61 },
		{ 276, 483, 61 },
		{ 276, 484, 61 },
		{ 276, 485, 61 },
		{ 276, 486, 61 },
		{ 276, 487, 61 },
		{ 276, 488, 61 },
		{ 276, 489, 61 },
		{ 276, 490, 61 },
		{ 276, 491, 61 },
		{ 276, 492, 61 },
		{ 276, 493, 61 },
		{ 276, 494, 61 },
		{ 276, 495, 61 },
		{ 276, 496, 61 },
		{ 276, 497, 61 },
		{ 276, 498, 61 },
		{ 276, 499, 61 },
		{ 276, 500, 61 },
		{ 276, 501, 61 },
		{ 276, 502, 61 },
		{ 276, 503, 61 },
		{ 276, 504, 61 },
		{ 276, 505, 61 },
		{ 276, 506, 61 },
		{ 276, 507, 61 },
		{ 276, 508, 61 },
		{ 276, 509, 61 },
		{ 276, 510, 61 },
		{ 276, 511, 61 },
		{ 276, 512, 61 },
		{ 277, 338, 61 },
		{ 277, 339, 61 },
		{ 277, 340, 61 },
		{ 277, 341, 61 },
		{ 277, 342, 61 },
		{ 277, 343, 61 },
		{ 277, 344, 61 },
		{ 277, 345, 61 },
		{ 277, 346, 61 },
		{ 277, 347, 61 },
		{ 277, 348, 61 },
		{ 277, 349, 61 },
		{ 277, 350, 61 },
		{ 277, 351, 61 },
		{ 277, 352, 61 },
		{ 277, 353, 61 },
		{ 277, 354, 61 },
		{ 277, 355, 61 },
		{ 277, 356, 61 },
		{ 277, 357, 61 },
		{ 277, 358, 61 },
		{ 277, 359, 61 },
		{ 277, 360, 61 },
		{ 277, 361, 61 },
		{ 277, 362, 61 },
		{ 277, 363, 61 },
		{ 277, 364, 61 },
		{ 277, 365, 61 },
		{ 277, 366, 61 },
		{ 277, 367, 61 },
		{ 277, 368, 61 },
		{ 277, 369, 61 },
		{ 277, 370, 61 },
		{ 277, 371, 61 },
		{ 277, 372, 61 },
		{ 277, 373, 61 },
		{ 277, 374, 61 },
		{ 277, 375, 61 },
		{ 277, 376, 61 },
		{ 277, 377, 61 },
		{ 277, 378, 61 },
		{ 277, 379, 61 },
		{ 277, 380, 61 },
		{ 277, 381, 61 },
		{ 277, 382, 61 },
		{ 277, 383, 61 },
		{ 277, 384, 61 },
		{ 277, 385, 61 },
		{ 277, 386, 61 },
		{ 277, 387, 61 },
		{ 277, 388, 61 },
		{ 277, 389, 61 },
		{ 277, 390, 61 },
		{ 277, 391, 61 },
		{ 277, 392, 61 },
		{ 277, 393, 61 },
		{ 277, 394, 61 },
		{ 277, 395, 61 },
		{ 277, 396, 61 },
		{ 277, 397, 61 },
		{ 277, 398, 61 },
		{ 277, 399, 61 },
		{ 277, 400, 61 },
		{ 277, 401, 61 },
		{ 277, 402, 61 },
		{ 277, 403, 61 },
		{ 277, 404, 61 },
		{ 277, 405, 61 },
		{ 277, 406, 61 },
		{ 277, 407, 61 },
		{ 277, 408, 61 },
		{ 277, 409, 61 },
		{ 277, 410, 61 },
		{ 277, 411, 61 },
		{ 277, 412, 61 },
		{ 277, 413, 61 },
		{ 277, 414, 61 },
		{ 277, 415, 61 },
		{ 277, 416, 61 },
		{ 277, 417, 61 },
		{ 277, 418, 61 },
		{ 277, 419, 61 },
		{ 277, 420, 61 },
		{ 277, 421, 61 },
		{ 277, 422, 61 },
		{ 277, 423, 61 },
		{ 277, 424, 61 },
		{ 277, 425, 61 },
		{ 277, 426, 61 },
		{ 277, 427, 61 },
		{ 277, 428, 61 },
		{ 277, 429, 61 },
		{ 277, 430, 61 },
		{ 277, 431, 61 },
		{ 277, 432, 61 },
		{ 277, 433, 61 },
		{ 277, 434, 61 },
		{ 277, 435, 61 },
		{ 277, 436, 61 },
		{ 277, 437, 61 },
		{ 277, 438, 61 },
		{ 277, 439, 61 },
		{ 277, 440, 61 },
		{ 277, 441, 61 },
		{ 277, 442, 61 },
		{ 277, 443, 61 },
		{ 277, 444, 61 },
		{ 277, 445, 61 },
		{ 277, 446, 61 },
		{ 277, 447, 61 },
		{ 277, 448, 61 },
		{ 277, 449, 61 },
		{ 277, 450, 61 },
		{ 277, 451, 61 },
		{ 277, 452, 61 },
		{ 277, 453, 61 },
		{ 277, 454, 61 },
		{ 277, 455, 61 },
		{ 277, 456, 61 },
		{ 277, 457, 61 },
		{ 277, 458, 61 },
		{ 277, 459, 61 },
		{ 277, 460, 61 },
		{ 277, 461, 61 },
		{ 277, 462, 61 },
		{ 277, 463, 61 },
		{ 277, 464, 61 },
		{ 277, 465, 61 },
		{ 277, 466, 61 },
		{ 277, 467, 61 },
		{ 277, 468, 61 },
		{ 277, 469, 61 },
		{ 277, 470, 61 },
		{ 277, 471, 61 },
		{ 277, 472, 61 },
		{ 277, 473, 61 },
		{ 277, 474, 61 },
		{ 277, 475, 61 },
		{ 277, 476, 61 },
		{ 277, 477, 61 },
		{ 277, 478, 61 },
		{ 277, 479, 61 },
		{ 277, 480, 61 },
		{ 277, 481, 61 },
		{ 277, 482, 61 },
		{ 277, 483, 61 },
		{ 277, 484, 61 },
		{ 277, 485, 61 },
		{ 277, 486, 61 },
		{ 277, 487, 61 },
		{ 277, 488, 61 },
		{ 277, 489, 61 },
		{ 277, 490, 61 },
		{ 277, 491, 61 },
		{ 277, 492, 61 },
		{ 277, 493, 61 },
		{ 277, 494, 61 },
		{ 277, 495, 61 },
		{ 277, 496, 61 },
		{ 277, 497, 61 },
		{ 277, 498, 61 },
		{ 277, 499, 61 },
		{ 277, 500, 61 },
		{ 277, 501, 61 },
		{ 277, 502, 61 },
		{ 277, 503, 61 },
		{ 277, 504, 61 },
		{ 277, 505, 61 },
		{ 277, 506, 61 },
		{ 277, 507, 61 },
		{ 277, 508, 61 },
		{ 277, 509, 61 },
		{ 277, 510, 61 },
		{ 277, 511, 61 },
		{ 277, 512, 61 },
		{ 278, 342, 61 },
		{ 278, 343, 61 },
		{ 278, 344, 61 },
		{ 278, 345, 61 },
		{ 278, 346, 61 },
		{ 278, 347, 61 },
		{ 278, 348, 61 },
		{ 278, 349, 61 },
		{ 278, 350, 61 },
		{ 278, 351, 61 },
		{ 278, 352, 61 },
		{ 278, 353, 61 },
		{ 278, 354, 61 },
		{ 278, 355, 61 },
		{ 278, 356, 61 },
		{ 278, 357, 61 },
		{ 278, 358, 61 },
		{ 278, 359, 61 },
		{ 278, 360, 61 },
		{ 278, 361, 61 },
		{ 278, 362, 61 },
		{ 278, 363, 61 },
		{ 278, 364, 61 },
		{ 278, 365, 61 },
		{ 278, 366, 61 },
		{ 278, 367, 61 },
		{ 278, 368, 61 },
		{ 278, 369, 61 },
		{ 278, 370, 61 },
		{ 278, 371, 61 },
		{ 278, 372, 61 },
		{ 278, 373, 61 },
		{ 278, 374, 61 },
		{ 278, 375, 61 },
		{ 278, 376, 61 },
		{ 278, 377, 61 },
		{ 278, 378, 61 },
		{ 278, 379, 61 },
		{ 278, 380, 61 },
		{ 278, 381, 61 },
		{ 278, 382, 61 },
		{ 278, 383, 61 },
		{ 278, 384, 61 },
		{ 278, 385, 61 },
		{ 278, 386, 61 },
		{ 278, 387, 61 },
		{ 278, 388, 61 },
		{ 278, 389, 61 },
		{ 278, 390, 61 },
		{ 278, 391, 61 },
		{ 278, 392, 61 },
		{ 278, 393, 61 },
		{ 278, 394, 61 },
		{ 278, 395, 61 },
		{ 278, 396, 61 },
		{ 278, 397, 61 },
		{ 278, 398, 61 },
		{ 278, 399, 61 },
		{ 278, 400, 61 },
		{ 278, 401, 61 },
		{ 278, 402, 61 },
		{ 278, 403, 61 },
		{ 278, 404, 61 },
		{ 278, 405, 61 },
		{ 278, 406, 61 },
		{ 278, 407, 61 },
		{ 278, 408, 61 },
		{ 278, 409, 61 },
		{ 278, 410, 61 },
		{ 278, 411, 61 },
		{ 278, 412, 61 },
		{ 278, 413, 61 },
		{ 278, 414, 61 },
		{ 278, 415, 61 },
		{ 278, 416, 61 },
		{ 278, 417, 61 },
		{ 278, 418, 61 },
		{ 278, 419, 61 },
		{ 278, 420, 61 },
		{ 278, 421, 61 },
		{ 278, 422, 61 },
		{ 278, 423, 61 },
		{ 278, 424, 61 },
		{ 278, 425, 61 },
		{ 278, 426, 61 },
		{ 278, 427, 61 },
		{ 278, 428, 61 },
		{ 278, 429, 61 },
		{ 278, 430, 61 },
		{ 278, 431, 61 },
		{ 278, 432, 61 },
		{ 278, 433, 61 },
		{ 278, 434, 61 },
		{ 278, 435, 61 },
		{ 278, 436, 61 },
		{ 278, 437, 61 },
		{ 278, 438, 61 },
		{ 278, 439, 61 },
		{ 278, 440, 61 },
		{ 278, 441, 61 },
		{ 278, 442, 61 },
		{ 278, 443, 61 },
		{ 278, 444, 61 },
		{ 278, 445, 61 },
		{ 278, 446, 61 },
		{ 278, 447, 61 },
		{ 278, 448, 61 },
		{ 278, 449, 61 },
		{ 278, 450, 61 },
		{ 278, 451, 61 },
		{ 278, 452, 61 },
		{ 278, 453, 61 },
		{ 278, 454, 61 },
		{ 278, 455, 61 },
		{ 278, 456, 61 },
		{ 278, 457, 61 },
		{ 278, 458, 61 },
		{ 278, 459, 61 },
		{ 278, 460, 61 },
		{ 278, 461, 61 },
		{ 278, 462, 61 },
		{ 278, 463, 61 },
		{ 278, 464, 61 },
		{ 278, 465, 61 },
		{ 278, 466, 61 },
		{ 278, 467, 61 },
		{ 278, 468, 61 },
		{ 278, 469, 61 },
		{ 278, 470, 61 },
		{ 278, 471, 61 },
		{ 278, 472, 61 },
		{ 278, 473, 61 },
		{ 278, 474, 61 },
		{ 278, 475, 61 },
		{ 278, 476, 61 },
		{ 278, 477, 61 },
		{ 278, 478, 61 },
		{ 278, 479, 61 },
		{ 278, 480, 61 },
		{ 278, 481, 61 },
		{ 278, 482, 61 },
		{ 278, 483, 61 },
		{ 278, 484, 61 },
		{ 278, 485, 61 },
		{ 278, 486, 61 },
		{ 278, 487, 61 },
		{ 278, 488, 61 },
		{ 278, 489, 61 },
		{ 278, 490, 61 },
		{ 278, 491, 61 },
		{ 278, 492, 61 },
		{ 278, 493, 61 },
		{ 278, 494, 61 },
		{ 278, 495, 61 },
		{ 278, 496, 61 },
		{ 278, 497, 61 },
		{ 278, 498, 61 },
		{ 278, 499, 61 },
		{ 278, 500, 61 },
		{ 278, 501, 61 },
		{ 278, 502, 61 },
		{ 278, 503, 61 },
		{ 278, 504, 61 },
		{ 278, 505, 61 },
		{ 278, 506, 61 },
		{ 278, 507, 61 },
		{ 278, 508, 61 },
		{ 278, 509, 61 },
		{ 278, 510, 61 },
		{ 278, 511, 61 },
		{ 278, 512, 61 },
		{ 279, 346, 61 },
		{ 279, 347, 61 },
		{ 279, 348, 61 },
		{ 279, 349, 61 },
		{ 279, 350, 61 },
		{ 279, 351, 61 },
		{ 279, 352, 61 },
		{ 279, 353, 61 },
		{ 279, 354, 61 },
		{ 279, 355, 61 },
		{ 279, 356, 61 },
		{ 279, 357, 61 },
		{ 279, 358, 61 },
		{ 279, 359, 61 },
		{ 279, 360, 61 },
		{ 279, 361, 61 },
		{ 279, 362, 61 },
		{ 279, 363, 61 },
		{ 279, 364, 61 },
		{ 279, 365, 61 },
		{ 279, 366, 61 },
		{ 279, 367, 61 },
		{ 279, 368, 61 },
		{ 279, 369, 61 },
		{ 279, 370, 61 },
		{ 279, 371, 61 },
		{ 279, 372, 61 },
		{ 279, 373, 61 },
		{ 279, 374, 61 },
		{ 279, 375, 61 },
		{ 279, 376, 61 },
		{ 279, 377, 61 },
		{ 279, 378, 61 },
		{ 279, 379, 61 },
		{ 279, 380, 61 },
		{ 279, 381, 61 },
		{ 279, 382, 61 },
		{ 279, 383, 61 },
		{ 279, 384, 61 },
		{ 279, 385, 61 },
		{ 279, 386, 61 },
		{ 279, 387, 61 },
		{ 279, 388, 61 },
		{ 279, 389, 61 },
		{ 279, 390, 61 },
		{ 279, 391, 61 },
		{ 279, 392, 61 },
		{ 279, 393, 61 },
		{ 279, 394, 61 },
		{ 279, 395, 61 },
		{ 279, 396, 61 },
		{ 279, 397, 61 },
		{ 279, 398, 61 },
		{ 279, 399, 61 },
		{ 279, 400, 61 },
		{ 279, 401, 61 },
		{ 279, 402, 61 },
		{ 279, 403, 61 },
		{ 279, 404, 61 },
		{ 279, 405, 61 },
		{ 279, 406, 61 },
		{ 279, 407, 61 },
		{ 279, 408, 61 },
		{ 279, 409, 61 },
		{ 279, 410, 61 },
		{ 279, 411, 61 },
		{ 279, 412, 61 },
		{ 279, 413, 61 },
		{ 279, 414, 61 },
		{ 279, 415, 61 },
		{ 279, 416, 61 },
		{ 279, 417, 61 },
		{ 279, 418, 61 },
		{ 279, 419, 61 },
		{ 279, 420, 61 },
		{ 279, 421, 61 },
		{ 279, 422, 61 },
		{ 279, 423, 61 },
		{ 279, 424, 61 },
		{ 279, 425, 61 },
		{ 279, 426, 61 },
		{ 279, 427, 61 },
		{ 279, 428, 61 },
		{ 279, 429, 61 },
		{ 279, 430, 61 },
		{ 279, 431, 61 },
		{ 279, 432, 61 },
		{ 279, 433, 61 },
		{ 279, 434, 61 },
		{ 279, 435, 61 },
		{ 279, 436, 61 },
		{ 279, 437, 61 },
		{ 279, 438, 61 },
		{ 279, 439, 61 },
		{ 279, 440, 61 },
		{ 279, 441, 61 },
		{ 279, 442, 61 },
		{ 279, 443, 61 },
		{ 279, 444, 61 },
		{ 279, 445, 61 },
		{ 279, 446, 61 },
		{ 279, 447, 61 },
		{ 279, 448, 61 },
		{ 279, 449, 61 },
		{ 279, 450, 61 },
		{ 279, 451, 61 },
		{ 279, 452, 61 },
		{ 279, 453, 61 },
		{ 279, 454, 61 },
		{ 279, 455, 61 },
		{ 279, 456, 61 },
		{ 279, 457, 61 },
		{ 279, 458, 61 },
		{ 279, 459, 61 },
		{ 279, 460, 61 },
		{ 279, 461, 61 },
		{ 279, 462, 61 },
		{ 279, 463, 61 },
		{ 279, 464, 61 },
		{ 279, 465, 61 },
		{ 279, 466, 61 },
		{ 279, 467, 61 },
		{ 279, 468, 61 },
		{ 279, 469, 61 },
		{ 279, 470, 61 },
		{ 279, 471, 61 },
		{ 279, 472, 61 },
		{ 279, 473, 61 },
		{ 279, 474, 61 },
		{ 279, 475, 61 },
		{ 279, 476, 61 },
		{ 279, 477, 61 },
		{ 279, 478, 61 },
		{ 279, 479, 61 },
		{ 279, 480, 61 },
		{ 279, 481, 61 },
		{ 279, 482, 61 },
		{ 279, 483, 61 },
		{ 279, 484, 61 },
		{ 279, 485, 61 },
		{ 279, 486, 61 },
		{ 279, 487, 61 },
		{ 279, 488, 61 },
		{ 279, 489, 61 },
		{ 279, 490, 61 },
		{ 279, 491, 61 },
		{ 279, 492, 61 },
		{ 279, 493, 61 },
		{ 279, 494, 61 },
		{ 279, 495, 61 },
		{ 279, 496, 61 },
		{ 279, 497, 61 },
		{ 279, 498, 61 },
		{ 279, 499, 61 },
		{ 279, 500, 61 },
		{ 279, 501, 61 },
		{ 279, 502, 61 },
		{ 279, 503, 61 },
		{ 279, 504, 61 },
		{ 279, 505, 61 },
		{ 279, 506, 61 },
		{ 279, 507, 61 },
		{ 279, 508, 61 },
		{ 279, 509, 61 },
		{ 279, 510, 61 },
		{ 279, 511, 61 },
		{ 279, 512, 61 },
		{ 280, 350, 61 },
		{ 280, 351, 61 },
		{ 280, 352, 61 },
		{ 280, 353, 61 },
		{ 280, 354, 61 },
		{ 280, 355, 61 },
		{ 280, 356, 61 },
		{ 280, 357, 61 },
		{ 280, 358, 61 },
		{ 280, 359, 61 },
		{ 280, 360, 61 },
		{ 280, 361, 61 },
		{ 280, 362, 61 },
		{ 280, 363, 61 },
		{ 280, 364, 61 },
		{ 280, 365, 61 },
		{ 280, 366, 61 },
		{ 280, 367, 61 },
		{ 280, 368, 61 },
		{ 280, 369, 61 },
		{ 280, 370, 61 },
		{ 280, 371, 61 },
		{ 280, 372, 61 },
		{ 280, 373, 61 },
		{ 280, 374, 61 },
		{ 280, 375, 61 },
		{ 280, 376, 61 },
		{ 280, 377, 61 },
		{ 280, 378, 61 },
		{ 280, 379, 61 },
		{ 280, 380, 61 },
		{ 280, 381, 61 },
		{ 280, 382, 61 },
		{ 280, 383, 61 },
		{ 280, 384, 61 },
		{ 280, 385, 61 },
		{ 280, 386, 61 },
		{ 280, 387, 61 },
		{ 280, 388, 61 },
		{ 280, 389, 61 },
		{ 280, 390, 61 },
		{ 280, 391, 61 },
		{ 280, 392, 61 },
		{ 280, 393, 61 },
		{ 280, 394, 61 },
		{ 280, 395, 61 },
		{ 280, 396, 61 },
		{ 280, 397, 61 },
		{ 280, 398, 61 },
		{ 280, 399, 61 },
		{ 280, 400, 61 },
		{ 280, 401, 61 },
		{ 280, 402, 61 },
		{ 280, 403, 61 },
		{ 280, 404, 61 },
		{ 280, 405, 61 },
		{ 280, 406, 61 },
		{ 280, 407, 61 },
		{ 280, 408, 61 },
		{ 280, 409, 61 },
		{ 280, 410, 61 },
		{ 280, 411, 61 },
		{ 280, 412, 61 },
		{ 280, 413, 61 },
		{ 280, 414, 61 },
		{ 280, 415, 61 },
		{ 280, 416, 61 },
		{ 280, 417, 61 },
		{ 280, 418, 61 },
		{ 280, 419, 61 },
		{ 280, 420, 61 },
		{ 280, 421, 61 },
		{ 280, 422, 61 },
		{ 280, 423, 61 },
		{ 280, 424, 61 },
		{ 280, 425, 61 },
		{ 280, 426, 61 },
		{ 280, 427, 61 },
		{ 280, 428, 61 },
		{ 280, 429, 61 },
		{ 280, 430, 61 },
		{ 280, 431, 61 },
		{ 280, 432, 61 },
		{ 280, 433, 61 },
		{ 280, 434, 61 },
		{ 280, 435, 61 },
		{ 280, 436, 61 },
		{ 280, 437, 61 },
		{ 280, 438, 61 },
		{ 280, 439, 61 },
		{ 280, 440, 61 },
		{ 280, 441, 61 },
		{ 280, 442, 61 },
		{ 280, 443, 61 },
		{ 280, 444, 61 },
		{ 280, 445, 61 },
		{ 280, 446, 61 },
		{ 280, 447, 61 },
		{ 280, 448, 61 },
		{ 280, 449, 61 },
		{ 280, 450, 61 },
		{ 280, 451, 61 },
		{ 280, 452, 61 },
		{ 280, 453, 61 },
		{ 280, 454, 61 },
		{ 280, 455, 61 },
		{ 280, 456, 61 },
		{ 280, 457, 61 },
		{ 280, 458, 61 },
		{ 280, 459, 61 },
		{ 280, 460, 61 },
		{ 280, 461, 61 },
		{ 280, 462, 61 },
		{ 280, 463, 61 },
		{ 280, 464, 61 },
		{ 280, 465, 61 },
		{ 280, 466, 61 },
		{ 280, 467, 61 },
		{ 280, 468, 61 },
		{ 280, 469, 61 },
		{ 280, 470, 61 },
		{ 280, 471, 61 },
		{ 280, 472, 61 },
		{ 280, 473, 61 },
		{ 280, 474, 61 },
		{ 280, 475, 61 },
		{ 280, 476, 61 },
		{ 280, 477, 61 },
		{ 280, 478, 61 },
		{ 280, 479, 61 },
		{ 280, 480, 61 },
		{ 280, 481, 61 },
		{ 280, 482, 61 },
		{ 280, 483, 61 },
		{ 280, 484, 61 },
		{ 280, 485, 61 },
		{ 280, 486, 61 },
		{ 280, 487, 61 },
		{ 280, 488, 61 },
		{ 280, 489, 61 },
		{ 280, 490, 61 },
		{ 280, 491, 61 },
		{ 280, 492, 61 },
		{ 280, 493, 61 },
		{ 280, 494, 61 },
		{ 280, 495, 61 },
		{ 280, 496, 61 },
		{ 280, 497, 61 },
		{ 280, 498, 61 },
		{ 280, 499, 61 },
		{ 280, 500, 61 },
		{ 280, 501, 61 },
		{ 280, 502, 61 },
		{ 280, 503, 61 },
		{ 280, 504, 61 },
		{ 280, 505, 61 },
		{ 280, 506, 61 },
		{ 280, 507, 61 },
		{ 280, 508, 61 },
		{ 280, 509, 61 },
		{ 280, 510, 61 },
		{ 280, 511, 61 },
		{ 280, 512, 61 },
		{ 281, 354, 61 },
		{ 281, 355, 61 },
		{ 281, 356, 61 },
		{ 281, 357, 61 },
		{ 281, 358, 61 },
		{ 281, 359, 61 },
		{ 281, 360, 61 },
		{ 281, 361, 61 },
		{ 281, 362, 61 },
		{ 281, 363, 61 },
		{ 281, 364, 61 },
		{ 281, 365, 61 },
		{ 281, 366, 61 },
		{ 281, 367, 61 },
		{ 281, 368, 61 },
		{ 281, 369, 61 },
		{ 281, 370, 61 },
		{ 281, 371, 61 },
		{ 281, 372, 61 },
		{ 281, 373, 61 },
		{ 281, 374, 61 },
		{ 281, 375, 61 },
		{ 281, 376, 61 },
		{ 281, 377, 61 },
		{ 281, 378, 61 },
		{ 281, 379, 61 },
		{ 281, 380, 61 },
		{ 281, 381, 61 },
		{ 281, 382, 61 },
		{ 281, 383, 61 },
		{ 281, 384, 61 },
		{ 281, 385, 61 },
		{ 281, 386, 61 },
		{ 281, 387, 61 },
		{ 281, 388, 61 },
		{ 281, 389, 61 },
		{ 281, 390, 61 },
		{ 281, 391, 61 },
		{ 281, 392, 61 },
		{ 281, 393, 61 },
		{ 281, 394, 61 },
		{ 281, 395, 61 },
		{ 281, 396, 61 },
		{ 281, 397, 61 },
		{ 281, 398, 61 },
		{ 281, 399, 61 },
		{ 281, 400, 61 },
		{ 281, 401, 61 },
		{ 281, 402, 61 },
		{ 281, 403, 61 },
		{ 281, 404, 61 },
		{ 281, 405, 61 },
		{ 281, 406, 61 },
		{ 281, 407, 61 },
		{ 281, 408, 61 },
		{ 281, 409, 61 },
		{ 281, 410, 61 },
		{ 281, 411, 61 },
		{ 281, 412, 61 },
		{ 281, 413, 61 },
		{ 281, 414, 61 },
		{ 281, 415, 61 },
		{ 281, 416, 61 },
		{ 281, 417, 61 },
		{ 281, 418, 61 },
		{ 281, 419, 61 },
		{ 281, 420, 61 },
		{ 281, 421, 61 },
		{ 281, 422, 61 },
		{ 281, 423, 61 },
		{ 281, 424, 61 },
		{ 281, 425, 61 },
		{ 281, 426, 61 },
		{ 281, 427, 61 },
		{ 281, 428, 61 },
		{ 281, 429, 61 },
		{ 281, 430, 61 },
		{ 281, 431, 61 },
		{ 281, 432, 61 },
		{ 281, 433, 61 },
		{ 281, 434, 61 },
		{ 281, 435, 61 },
		{ 281, 436, 61 },
		{ 281, 437, 61 },
		{ 281, 438, 61 },
		{ 281, 439, 61 },
		{ 281, 440, 61 },
		{ 281, 441, 61 },
		{ 281, 442, 61 },
		{ 281, 443, 61 },
		{ 281, 444, 61 },
		{ 281, 445, 61 },
		{ 281, 446, 61 },
		{ 281, 447, 61 },
		{ 281, 448, 61 },
		{ 281, 449, 61 },
		{ 281, 450, 61 },
		{ 281, 451, 61 },
		{ 281, 452, 61 },
		{ 281, 453, 61 },
		{ 281, 454, 61 },
		{ 281, 455, 61 },
		{ 281, 456, 61 },
		{ 281, 457, 61 },
		{ 281, 458, 61 },
		{ 281, 459, 61 },
		{ 281, 460, 61 },
		{ 281, 461, 61 },
		{ 281, 462, 61 },
		{ 281, 463, 61 },
		{ 281, 464, 61 },
		{ 281, 465, 61 },
		{ 281, 466, 61 },
		{ 281, 467, 61 },
		{ 281, 468, 61 },
		{ 281, 469, 61 },
		{ 281, 470, 61 },
		{ 281, 471, 61 },
		{ 281, 472, 61 },
		{ 281, 473, 61 },
		{ 281, 474, 61 },
		{ 281, 475, 61 },
		{ 281, 476, 61 },
		{ 281, 477, 61 },
		{ 281, 478, 61 },
		{ 281, 479, 61 },
		{ 281, 480, 61 },
		{ 281, 481, 61 },
		{ 281, 482, 61 },
		{ 281, 483, 61 },
		{ 281, 484, 61 },
		{ 281, 485, 61 },
		{ 281, 486, 61 },
		{ 281, 487, 61 },
		{ 281, 488, 61 },
		{ 281, 489, 61 },
		{ 281, 490, 61 },
		{ 281, 491, 61 },
		{ 281, 492, 61 },
		{ 281, 493, 61 },
		{ 281, 494, 61 },
		{ 281, 495, 61 },
		{ 281, 496, 61 },
		{ 281, 497, 61 },
		{ 281, 498, 61 },
		{ 281, 499, 61 },
		{ 281, 500, 61 },
		{ 281, 501, 61 },
		{ 281, 502, 61 },
		{ 281, 503, 61 },
		{ 281, 504, 61 },
		{ 281, 505, 61 },
		{ 281, 506, 61 },
		{ 281, 507, 61 },
		{ 281, 508, 61 },
		{ 281, 509, 61 },
		{ 281, 510, 61 },
		{ 281, 511, 61 },
		{ 281, 512, 61 },
		{ 282, 358, 61 },
		{ 282, 359, 61 },
		{ 282, 360, 61 },
		{ 282, 361, 61 },
		{ 282, 362, 61 },
		{ 282, 363, 61 },
		{ 282, 364, 61 },
		{ 282, 365, 61 },
		{ 282, 366, 61 },
		{ 282, 367, 61 },
		{ 282, 368, 61 },
		{ 282, 369, 61 },
		{ 282, 370, 61 },
		{ 282, 371, 61 },
		{ 282, 372, 61 },
		{ 282, 373, 61 },
		{ 282, 374, 61 },
		{ 282, 375, 61 },
		{ 282, 376, 61 },
		{ 282, 377, 61 },
		{ 282, 378, 61 },
		{ 282, 379, 61 },
		{ 282, 380, 61 },
		{ 282, 381, 61 },
		{ 282, 382, 61 },
		{ 282, 383, 61 },
		{ 282, 384, 61 },
		{ 282, 385, 61 },
		{ 282, 386, 61 },
		{ 282, 387, 61 },
		{ 282, 388, 61 },
		{ 282, 389, 61 },
		{ 282, 390, 61 },
		{ 282, 391, 61 },
		{ 282, 392, 61 },
		{ 282, 393, 61 },
		{ 282, 394, 61 },
		{ 282, 395, 61 },
		{ 282, 396, 61 },
		{ 282, 397, 61 },
		{ 282, 398, 61 },
		{ 282, 399, 61 },
		{ 282, 400, 61 },
		{ 282, 401, 61 },
		{ 282, 402, 61 },
		{ 282, 403, 61 },
		{ 282, 404, 61 },
		{ 282, 405, 61 },
		{ 282, 406, 61 },
		{ 282, 407, 61 },
		{ 282, 408, 61 },
		{ 282, 409, 61 },
		{ 282, 410, 61 },
		{ 282, 411, 61 },
		{ 282, 412, 61 },
		{ 282, 413, 61 },
		{ 282, 414, 61 },
		{ 282, 415, 61 },
		{ 282, 416, 61 },
		{ 282, 417, 61 },
		{ 282, 418, 61 },
		{ 282, 419, 61 },
		{ 282, 420, 61 },
		{ 282, 421, 61 },
		{ 282, 422, 61 },
		{ 282, 423, 61 },
		{ 282, 424, 61 },
		{ 282, 425, 61 },
		{ 282, 426, 61 },
		{ 282, 427, 61 },
		{ 282, 428, 61 },
		{ 282, 429, 61 },
		{ 282, 430, 61 },
		{ 282, 431, 61 },
		{ 282, 432, 61 },
		{ 282, 433, 61 },
		{ 282, 434, 61 },
		{ 282, 435, 61 },
		{ 282, 436, 61 },
		{ 282, 437, 61 },
		{ 282, 438, 61 },
		{ 282, 439, 61 },
		{ 282, 440, 61 },
		{ 282, 441, 61 },
		{ 282, 442, 61 },
		{ 282, 443, 61 },
		{ 282, 444, 61 },
		{ 282, 445, 61 },
		{ 282, 446, 61 },
		{ 282, 447, 61 },
		{ 282, 448, 61 },
		{ 282, 449, 61 },
		{ 282, 450, 61 },
		{ 282, 451, 61 },
		{ 282, 452, 61 },
		{ 282, 453, 61 },
		{ 282, 454, 61 },
		{ 282, 455, 61 },
		{ 282, 456, 61 },
		{ 282, 457, 61 },
		{ 282, 458, 61 },
		{ 282, 459, 61 },
		{ 282, 460, 61 },
		{ 282, 461, 61 },
		{ 282, 462, 61 },
		{ 282, 463, 61 },
		{ 282, 464, 61 },
		{ 282, 465, 61 },
		{ 282, 466, 61 },
		{ 282, 467, 61 },
		{ 282, 468, 61 },
		{ 282, 469, 61 },
		{ 282, 470, 61 },
		{ 282, 471, 61 },
		{ 282, 472, 61 },
		{ 282, 473, 61 },
		{ 282, 474, 61 },
		{ 282, 475, 61 },
		{ 282, 476, 61 },
		{ 282, 477, 61 },
		{ 282, 478, 61 },
		{ 282, 479, 61 },
		{ 282, 480, 61 },
		{ 282, 481, 61 },
		{ 282, 482, 61 },
		{ 282, 483, 61 },
		{ 282, 484, 61 },
		{ 282, 485, 61 },
		{ 282, 486, 61 },
		{ 282, 487, 61 },
		{ 282, 488, 61 },
		{ 282, 489, 61 },
		{ 282, 490, 61 },
		{ 282, 491, 61 },
		{ 282, 492, 61 },
		{ 282, 493, 61 },
		{ 282, 494, 61 },
		{ 282, 495, 61 },
		{ 282, 496, 61 },
		{ 282, 497, 61 },
		{ 282, 498, 61 },
		{ 282, 499, 61 },
		{ 282, 500, 61 },
		{ 282, 501, 61 },
		{ 282, 502, 61 },
		{ 282, 503, 61 },
		{ 282, 504, 61 },
		{ 282, 505, 61 },
		{ 282, 506, 61 },
		{ 282, 507, 61 },
		{ 282, 508, 61 },
		{ 282, 509, 61 },
		{ 282, 510, 61 },
		{ 282, 511, 61 },
		{ 282, 512, 61 },
		{ 283, 362, 61 },
		{ 283, 363, 61 },
		{ 283, 364, 61 },
		{ 283, 365, 61 },
		{ 283, 366, 61 },
		{ 283, 367, 61 },
		{ 283, 368, 61 },
		{ 283, 369, 61 },
		{ 283, 370, 61 },
		{ 283, 371, 61 },
		{ 283, 372, 61 },
		{ 283, 373, 61 },
		{ 283, 374, 61 },
		{ 283, 375, 61 },
		{ 283, 376, 61 },
		{ 283, 377, 61 },
		{ 283, 378, 61 },
		{ 283, 379, 61 },
		{ 283, 380, 61 },
		{ 283, 381, 61 },
		{ 283, 382, 61 },
		{ 283, 383, 61 },
		{ 283, 384, 61 },
		{ 283, 385, 61 },
		{ 283, 386, 61 },
		{ 283, 387, 61 },
		{ 283, 388, 61 },
		{ 283, 389, 61 },
		{ 283, 390, 61 },
		{ 283, 391, 61 },
		{ 283, 392, 61 },
		{ 283, 393, 61 },
		{ 283, 394, 61 },
		{ 283, 395, 61 },
		{ 283, 396, 61 },
		{ 283, 397, 61 },
		{ 283, 398, 61 },
		{ 283, 399, 61 },
		{ 283, 400, 61 },
		{ 283, 401, 61 },
		{ 283, 402, 61 },
		{ 283, 403, 61 },
		{ 283, 404, 61 },
		{ 283, 405, 61 },
		{ 283, 406, 61 },
		{ 283, 407, 61 },
		{ 283, 408, 61 },
		{ 283, 409, 61 },
		{ 283, 410, 61 },
		{ 283, 411, 61 },
		{ 283, 412, 61 },
		{ 283, 413, 61 },
		{ 283, 414, 61 },
		{ 283, 415, 61 },
		{ 283, 416, 61 },
		{ 283, 417, 61 },
		{ 283, 418, 61 },
		{ 283, 419, 61 },
		{ 283, 420, 61 },
		{ 283, 421, 61 },
		{ 283, 422, 61 },
		{ 283, 423, 61 },
		{ 283, 424, 61 },
		{ 283, 425, 61 },
		{ 283, 426, 61 },
		{ 283, 427, 61 },
		{ 283, 428, 61 },
		{ 283, 429, 61 },
		{ 283, 430, 61 },
		{ 283, 431, 61 },
		{ 283, 432, 61 },
		{ 283, 433, 61 },
		{ 283, 434, 61 },
		{ 283, 435, 61 },
		{ 283, 436, 61 },
		{ 283, 437, 61 },
		{ 283, 438, 61 },
		{ 283, 439, 61 },
		{ 283, 440, 61 },
		{ 283, 441, 61 },
		{ 283, 442, 61 },
		{ 283, 443, 61 },
		{ 283, 444, 61 },
		{ 283, 445, 61 },
		{ 283, 446, 61 },
		{ 283, 447, 61 },
		{ 283, 448, 61 },
		{ 283, 449, 61 },
		{ 283, 450, 61 },
		{ 283, 451, 61 },
		{ 283, 452, 61 },
		{ 283, 453, 61 },
		{ 283, 454, 61 },
		{ 283, 455, 61 },
		{ 283, 456, 61 },
		{ 283, 457, 61 },
		{ 283, 458, 61 },
		{ 283, 459, 61 },
		{ 283, 460, 61 },
		{ 283, 461, 61 },
		{ 283, 462, 61 },
		{ 283, 463, 61 },
		{ 283, 464, 61 },
		{ 283, 465, 61 },
		{ 283, 466, 61 },
		{ 283, 467, 61 },
		{ 283, 468, 61 },
		{ 283, 469, 61 },
		{ 283, 470, 61 },
		{ 283, 471, 61 },
		{ 283, 472, 61 },
		{ 283, 473, 61 },
		{ 283, 474, 61 },
		{ 283, 475, 61 },
		{ 283, 476, 61 },
		{ 283, 477, 61 },
		{ 283, 478, 61 },
		{ 283, 479, 61 },
		{ 283, 480, 61 },
		{ 283, 481, 61 },
		{ 283, 482, 61 },
		{ 283, 483, 61 },
		{ 283, 484, 61 },
		{ 283, 485, 61 },
		{ 283, 486, 61 },
		{ 283, 487, 61 },
		{ 283, 488, 61 },
		{ 283, 489, 61 },
		{ 283, 490, 61 },
		{ 283, 491, 61 },
		{ 283, 492, 61 },
		{ 283, 493, 61 },
		{ 283, 494, 61 },
		{ 283, 495, 61 },
		{ 283, 496, 61 },
		{ 283, 497, 61 },
		{ 283, 498, 61 },
		{ 283, 499, 61 },
		{ 283, 500, 61 },
		{ 283, 501, 61 },
		{ 283, 502, 61 },
		{ 283, 503, 61 },
		{ 283, 504, 61 },
		{ 283, 505, 61 },
		{ 283, 506, 61 },
		{ 283, 507, 61 },
		{ 283, 508, 61 },
		{ 283, 509, 61 },
		{ 283, 510, 61 },
		{ 283, 511, 61 },
		{ 283, 512, 61 },
		{ 284, 366, 61 },
		{ 284, 367, 61 },
		{ 284, 368, 61 },
		{ 284, 369, 61 },
		{ 284, 370, 61 },
		{ 284, 371, 61 },
		{ 284, 372, 61 },
		{ 284, 373, 61 },
		{ 284, 374, 61 },
		{ 284, 375, 61 },
		{ 284, 376, 61 },
		{ 284, 377, 61 },
		{ 284, 378, 61 },
		{ 284, 379, 61 },
		{ 284, 380, 61 },
		{ 284, 381, 61 },
		{ 284, 382, 61 },
		{ 284, 383, 61 },
		{ 284, 384, 61 },
		{ 284, 385, 61 },
		{ 284, 386, 61 },
		{ 284, 387, 61 },
		{ 284, 388, 61 },
		{ 284, 389, 61 },
		{ 284, 390, 61 },
		{ 284, 391, 61 },
		{ 284, 392, 61 },
		{ 284, 393, 61 },
		{ 284, 394, 61 },
		{ 284, 395, 61 },
		{ 284, 396, 61 },
		{ 284, 397, 61 },
		{ 284, 398, 61 },
		{ 284, 399, 61 },
		{ 284, 400, 61 },
		{ 284, 401, 61 },
		{ 284, 402, 61 },
		{ 284, 403, 61 },
		{ 284, 404, 61 },
		{ 284, 405, 61 },
		{ 284, 406, 61 },
		{ 284, 407, 61 },
		{ 284, 408, 61 },
		{ 284, 409, 61 },
		{ 284, 410, 61 },
		{ 284, 411, 61 },
		{ 284, 412, 61 },
		{ 284, 413, 61 },
		{ 284, 414, 61 },
		{ 284, 415, 61 },
		{ 284, 416, 61 },
		{ 284, 417, 61 },
		{ 284, 418, 61 },
		{ 284, 419, 61 },
		{ 284, 420, 61 },
		{ 284, 421, 61 },
		{ 284, 422, 61 },
		{ 284, 423, 61 },
		{ 284, 424, 61 },
		{ 284, 425, 61 },
		{ 284, 426, 61 },
		{ 284, 427, 61 },
		{ 284, 428, 61 },
		{ 284, 429, 61 },
		{ 284, 430, 61 },
		{ 284, 431, 61 },
		{ 284, 432, 61 },
		{ 284, 433, 61 },
		{ 284, 434, 61 },
		{ 284, 435, 61 },
		{ 284, 436, 61 },
		{ 284, 437, 61 },
		{ 284, 438, 61 },
		{ 284, 439, 61 },
		{ 284, 440, 61 },
		{ 284, 441, 61 },
		{ 284, 442, 61 },
		{ 284, 443, 61 },
		{ 284, 444, 61 },
		{ 284, 445, 61 },
		{ 284, 446, 61 },
		{ 284, 447, 61 },
		{ 284, 448, 61 },
		{ 284, 449, 61 },
		{ 284, 450, 61 },
		{ 284, 451, 61 },
		{ 284, 452, 61 },
		{ 284, 453, 61 },
		{ 284, 454, 61 },
		{ 284, 455, 61 },
		{ 284, 456, 61 },
		{ 284, 457, 61 },
		{ 284, 458, 61 },
		{ 284, 459, 61 },
		{ 284, 460, 61 },
		{ 284, 461, 61 },
		{ 284, 462, 61 },
		{ 284, 463, 61 },
		{ 284, 464, 61 },
		{ 284, 465, 61 },
		{ 284, 466, 61 },
		{ 284, 467, 61 },
		{ 284, 468, 61 },
		{ 284, 469, 61 },
		{ 284, 470, 61 },
		{ 284, 471, 61 },
		{ 284, 472, 61 },
		{ 284, 473, 61 },
		{ 284, 474, 61 },
		{ 284, 475, 61 },
		{ 284, 476, 61 },
		{ 284, 477, 61 },
		{ 284, 478, 61 },
		{ 284, 479, 61 },
		{ 284, 480, 61 },
		{ 284, 481, 61 },
		{ 284, 482, 61 },
		{ 284, 483, 61 },
		{ 284, 484, 61 },
		{ 284, 485, 61 },
		{ 284, 486, 61 },
		{ 284, 487, 61 },
		{ 284, 488, 61 },
		{ 284, 489, 61 },
		{ 284, 490, 61 },
		{ 284, 491, 61 },
		{ 284, 492, 61 },
		{ 284, 493, 61 },
		{ 284, 494, 61 },
		{ 284, 495, 61 },
		{ 284, 496, 61 },
		{ 284, 497, 61 },
		{ 284, 498, 61 },
		{ 284, 499, 61 },
		{ 284, 500, 61 },
		{ 284, 501, 61 },
		{ 284, 502, 61 },
		{ 284, 503, 61 },
		{ 284, 504, 61 },
		{ 284, 505, 61 },
		{ 284, 506, 61 },
		{ 284, 507, 61 },
		{ 284, 508, 61 },
		{ 284, 509, 61 },
		{ 284, 510, 61 },
		{ 284, 511, 61 },
		{ 284, 512, 61 },
		{ 285, 370, 61 },
		{ 285, 371, 61 },
		{ 285, 372, 61 },
		{ 285, 373, 61 },
		{ 285, 374, 61 },
		{ 285, 375, 61 },
		{ 285, 376, 61 },
		{ 285, 377, 61 },
		{ 285, 378, 61 },
		{ 285, 379, 61 },
		{ 285, 380, 61 },
		{ 285, 381, 61 },
		{ 285, 382, 61 },
		{ 285, 383, 61 },
		{ 285, 384, 61 },
		{ 285, 385, 61 },
		{ 285, 386, 61 },
		{ 285, 387, 61 },
		{ 285, 388, 61 },
		{ 285, 389, 61 },
		{ 285, 390, 61 },
		{ 285, 391, 61 },
		{ 285, 392, 61 },
		{ 285, 393, 61 },
		{ 285, 394, 61 },
		{ 285, 395, 61 },
		{ 285, 396, 61 },
		{ 285, 397, 61 },
		{ 285, 398, 61 },
		{ 285, 399, 61 },
		{ 285, 400, 61 },
		{ 285, 401, 61 },
		{ 285, 402, 61 },
		{ 285, 403, 61 },
		{ 285, 404, 61 },
		{ 285, 405, 61 },
		{ 285, 406, 61 },
		{ 285, 407, 61 },
		{ 285, 408, 61 },
		{ 285, 409, 61 },
		{ 285, 410, 61 },
		{ 285, 411, 61 },
		{ 285, 412, 61 },
		{ 285, 413, 61 },
		{ 285, 414, 61 },
		{ 285, 415, 61 },
		{ 285, 416, 61 },
		{ 285, 417, 61 },
		{ 285, 418, 61 },
		{ 285, 419, 61 },
		{ 285, 420, 61 },
		{ 285, 421, 61 },
		{ 285, 422, 61 },
		{ 285, 423, 61 },
		{ 285, 424, 61 },
		{ 285, 425, 61 },
		{ 285, 426, 61 },
		{ 285, 427, 61 },
		{ 285, 428, 61 },
		{ 285, 429, 61 },
		{ 285, 430, 61 },
		{ 285, 431, 61 },
		{ 285, 432, 61 },
		{ 285, 433, 61 },
		{ 285, 434, 61 },
		{ 285, 435, 61 },
		{ 285, 436, 61 },
		{ 285, 437, 61 },
		{ 285, 438, 61 },
		{ 285, 439, 61 },
		{ 285, 440, 61 },
		{ 285, 441, 61 },
		{ 285, 442, 61 },
		{ 285, 443, 61 },
		{ 285, 444, 61 },
		{ 285, 445, 61 },
		{ 285, 446, 61 },
		{ 285, 447, 61 },
		{ 285, 448, 61 },
		{ 285, 449, 61 },
		{ 285, 450, 61 },
		{ 285, 451, 61 },
		{ 285, 452, 61 },
		{ 285, 453, 61 },
		{ 285, 454, 61 },
		{ 285, 455, 61 },
		{ 285, 456, 61 },
		{ 285, 457, 61 },
		{ 285, 458, 61 },
		{ 285, 459, 61 },
		{ 285, 460, 61 },
		{ 285, 461, 61 },
		{ 285, 462, 61 },
		{ 285, 463, 61 },
		{ 285, 464, 61 },
		{ 285, 465, 61 },
		{ 285, 466, 61 },
		{ 285, 467, 61 },
		{ 285, 468, 61 },
		{ 285, 469, 61 },
		{ 285, 470, 61 },
		{ 285, 471, 61 },
		{ 285, 472, 61 },
		{ 285, 473, 61 },
		{ 285, 474, 61 },
		{ 285, 475, 61 },
		{ 285, 476, 61 },
		{ 285, 477, 61 },
		{ 285, 478, 61 },
		{ 285, 479, 61 },
		{ 285, 480, 61 },
		{ 285, 481, 61 },
		{ 285, 482, 61 },
		{ 285, 483, 61 },
		{ 285, 484, 61 },
		{ 285, 485, 61 },
		{ 285, 486, 61 },
		{ 285, 487, 61 },
		{ 285, 488, 61 },
		{ 285, 489, 61 },
		{ 285, 490, 61 },
		{ 285, 491, 61 },
		{ 285, 492, 61 },
		{ 285, 493, 61 },
		{ 285, 494, 61 },
		{ 285, 495, 61 },
		{ 285, 496, 61 },
		{ 285, 497, 61 },
		{ 285, 498, 61 },
		{ 285, 499, 61 },
		{ 285, 500, 61 },
		{ 285, 501, 61 },
		{ 285, 502, 61 },
		{ 285, 503, 61 },
		{ 285, 504, 61 },
		{ 285, 505, 61 },
		{ 285, 506, 61 },
		{ 285, 507, 61 },
		{ 285, 508, 61 },
		{ 285, 509, 61 },
		{ 285, 510, 61 },
		{ 285, 511, 61 },
		{ 285, 512, 61 },
		{ 286, 374, 61 },
		{ 286, 375, 61 },
		{ 286, 376, 61 },
		{ 286, 377, 61 },
		{ 286, 378, 61 },
		{ 286, 379, 61 },
		{ 286, 380, 61 },
		{ 286, 381, 61 },
		{ 286, 382, 61 },
		{ 286, 383, 61 },
		{ 286, 384, 61 },
		{ 286, 385, 61 },
		{ 286, 386, 61 },
		{ 286, 387, 61 },
		{ 286, 388, 61 },
		{ 286, 389, 61 },
		{ 286, 390, 61 },
		{ 286, 391, 61 },
		{ 286, 392, 61 },
		{ 286, 393, 61 },
		{ 286, 394, 61 },
		{ 286, 395, 61 },
		{ 286, 396, 61 },
		{ 286, 397, 61 },
		{ 286, 398, 61 },
		{ 286, 399, 61 },
		{ 286, 400, 61 },
		{ 286, 401, 61 },
		{ 286, 402, 61 },
		{ 286, 403, 61 },
		{ 286, 404, 61 },
		{ 286, 405, 61 },
		{ 286, 406, 61 },
		{ 286, 407, 61 },
		{ 286, 408, 61 },
		{ 286, 409, 61 },
		{ 286, 410, 61 },
		{ 286, 411, 61 },
		{ 286, 412, 61 },
		{ 286, 413, 61 },
		{ 286, 414, 61 },
		{ 286, 415, 61 },
		{ 286, 416, 61 },
		{ 286, 417, 61 },
		{ 286, 418, 61 },
		{ 286, 419, 61 },
		{ 286, 420, 61 },
		{ 286, 421, 61 },
		{ 286, 422, 61 },
		{ 286, 423, 61 },
		{ 286, 424, 61 },
		{ 286, 425, 61 },
		{ 286, 426, 61 },
		{ 286, 427, 61 },
		{ 286, 428, 61 },
		{ 286, 429, 61 },
		{ 286, 430, 61 },
		{ 286, 431, 61 },
		{ 286, 432, 61 },
		{ 286, 433, 61 },
		{ 286, 434, 61 },
		{ 286, 435, 61 },
		{ 286, 436, 61 },
		{ 286, 437, 61 },
		{ 286, 438, 61 },
		{ 286, 439, 61 },
		{ 286, 440, 61 },
		{ 286, 441, 61 },
		{ 286, 442, 61 },
		{ 286, 443, 61 },
		{ 286, 444, 61 },
		{ 286, 445, 61 },
		{ 286, 446, 61 },
		{ 286, 447, 61 },
		{ 286, 448, 61 },
		{ 286, 449, 61 },
		{ 286, 450, 61 },
		{ 286, 451, 61 },
		{ 286, 452, 61 },
		{ 286, 453, 61 },
		{ 286, 454, 61 },
		{ 286, 455, 61 },
		{ 286, 456, 61 },
		{ 286, 457, 61 },
		{ 286, 458, 61 },
		{ 286, 459, 61 },
		{ 286, 460, 61 },
		{ 286, 461, 61 },
		{ 286, 462, 61 },
		{ 286, 463, 61 },
		{ 286, 464, 61 },
		{ 286, 465, 61 },
		{ 286, 466, 61 },
		{ 286, 467, 61 },
		{ 286, 468, 61 },
		{ 286, 469, 61 },
		{ 286, 470, 61 },
		{ 286, 471, 61 },
		{ 286, 472, 61 },
		{ 286, 473, 61 },
		{ 286, 474, 61 },
		{ 286, 475, 61 },
		{ 286, 476, 61 },
		{ 286, 477, 61 },
		{ 286, 478, 61 },
		{ 286, 479, 61 },
		{ 286, 480, 61 },
		{ 286, 481, 61 },
		{ 286, 482, 61 },
		{ 286, 483, 61 },
		{ 286, 484, 61 },
		{ 286, 485, 61 },
		{ 286, 486, 61 },
		{ 286, 487, 61 },
		{ 286, 488, 61 },
		{ 286, 489, 61 },
		{ 286, 490, 61 },
		{ 286, 491, 61 },
		{ 286, 492, 61 },
		{ 286, 493, 61 },
		{ 286, 494, 61 },
		{ 286, 495, 61 },
		{ 286, 496, 61 },
		{ 286, 497, 61 },
		{ 286, 498, 61 },
		{ 286, 499, 61 },
		{ 286, 500, 61 },
		{ 286, 501, 61 },
		{ 286, 502, 61 },
		{ 286, 503, 61 },
		{ 286, 504, 61 },
		{ 286, 505, 61 },
		{ 286, 506, 61 },
		{ 286, 507, 61 },
		{ 286, 508, 61 },
		{ 286, 509, 61 },
		{ 286, 510, 61 },
		{ 286, 511, 61 },
		{ 286, 512, 61 },
		{ 287, 378, 61 },
		{ 287, 379, 61 },
		{ 287, 380, 61 },
		{ 287, 381, 61 },
		{ 287, 382, 61 },
		{ 287, 383, 61 },
		{ 287, 384, 61 },
		{ 287, 385, 61 },
		{ 287, 386, 61 },
		{ 287, 387, 61 },
		{ 287, 388, 61 },
		{ 287, 389, 61 },
		{ 287, 390, 61 },
		{ 287, 391, 61 },
		{ 287, 392, 61 },
		{ 287, 393, 61 },
		{ 287, 394, 61 },
		{ 287, 395, 61 },
		{ 287, 396, 61 },
		{ 287, 397, 61 },
		{ 287, 398, 61 },
		{ 287, 399, 61 },
		{ 287, 400, 61 },
		{ 287, 401, 61 },
		{ 287, 402, 61 },
		{ 287, 403, 61 },
		{ 287, 404, 61 },
		{ 287, 405, 61 },
		{ 287, 406, 61 },
		{ 287, 407, 61 },
		{ 287, 408, 61 },
		{ 287, 409, 61 },
		{ 287, 410, 61 },
		{ 287, 411, 61 },
		{ 287, 412, 61 },
		{ 287, 413, 61 },
		{ 287, 414, 61 },
		{ 287, 415, 61 },
		{ 287, 416, 61 },
		{ 287, 417, 61 },
		{ 287, 418, 61 },
		{ 287, 419, 61 },
		{ 287, 420, 61 },
		{ 287, 421, 61 },
		{ 287, 422, 61 },
		{ 287, 423, 61 },
		{ 287, 424, 61 },
		{ 287, 425, 61 },
		{ 287, 426, 61 },
		{ 287, 427, 61 },
		{ 287, 428, 61 },
		{ 287, 429, 61 },
		{ 287, 430, 61 },
		{ 287, 431, 61 },
		{ 287, 432, 61 },
		{ 287, 433, 61 },
		{ 287, 434, 61 },
		{ 287, 435, 61 },
		{ 287, 436, 61 },
		{ 287, 437, 61 },
		{ 287, 438, 61 },
		{ 287, 439, 61 },
		{ 287, 440, 61 },
		{ 287, 441, 61 },
		{ 287, 442, 61 },
		{ 287, 443, 61 },
		{ 287, 444, 61 },
		{ 287, 445, 61 },
		{ 287, 446, 61 },
		{ 287, 447, 61 },
		{ 287, 448, 61 },
		{ 287, 449, 61 },
		{ 287, 450, 61 },
		{ 287, 451, 61 },
		{ 287, 452, 61 },
		{ 287, 453, 61 },
		{ 287, 454, 61 },
		{ 287, 455, 61 },
		{ 287, 456, 61 },
		{ 287, 457, 61 },
		{ 287, 458, 61 },
		{ 287, 459, 61 },
		{ 287, 460, 61 },
		{ 287, 461, 61 },
		{ 287, 462, 61 },
		{ 287, 463, 61 },
		{ 287, 464, 61 },
		{ 287, 465, 61 },
		{ 287, 466, 61 },
		{ 287, 467, 61 },
		{ 287, 468, 61 },
		{ 287, 469, 61 },
		{ 287, 470, 61 },
		{ 287, 471, 61 },
		{ 287, 472, 61 },
		{ 287, 473, 61 },
		{ 287, 474, 61 },
		{ 287, 475, 61 },
		{ 287, 476, 61 },
		{ 287, 477, 61 },
		{ 287, 478, 61 },
		{ 287, 479, 61 },
		{ 287, 480, 61 },
		{ 287, 481, 61 },
		{ 287, 482, 61 },
		{ 287, 483, 61 },
		{ 287, 484, 61 },
		{ 287, 485, 61 },
		{ 287, 486, 61 },
		{ 287, 487, 61 },
		{ 287, 488, 61 },
		{ 287, 489, 61 },
		{ 287, 490, 61 },
		{ 287, 491, 61 },
		{ 287, 492, 61 },
		{ 287, 493, 61 },
		{ 287, 494, 61 },
		{ 287, 495, 61 },
		{ 287, 496, 61 },
		{ 287, 497, 61 },
		{ 287, 498, 61 },
		{ 287, 499, 61 },
		{ 287, 500, 61 },
		{ 287, 501, 61 },
		{ 287, 502, 61 },
		{ 287, 503, 61 },
		{ 287, 504, 61 },
		{ 287, 505, 61 },
		{ 287, 506, 61 },
		{ 287, 507, 61 },
		{ 287, 508, 61 },
		{ 287, 509, 61 },
		{ 287, 510, 61 },
		{ 287, 511, 61 },
		{ 287, 512, 61 },
		{ 288, 382, 61 },
		{ 288, 383, 61 },
		{ 288, 384, 61 },
		{ 288, 385, 61 },
		{ 288, 386, 61 },
		{ 288, 387, 61 },
		{ 288, 388, 61 },
		{ 288, 389, 61 },
		{ 288, 390, 61 },
		{ 288, 391, 61 },
		{ 288, 392, 61 },
		{ 288, 393, 61 },
		{ 288, 394, 61 },
		{ 288, 395, 61 },
		{ 288, 396, 61 },
		{ 288, 397, 61 },
		{ 288, 398, 61 },
		{ 288, 399, 61 },
		{ 288, 400, 61 },
		{ 288, 401, 61 },
		{ 288, 402, 61 },
		{ 288, 403, 61 },
		{ 288, 404, 61 },
		{ 288, 405, 61 },
		{ 288, 406, 61 },
		{ 288, 407, 61 },
		{ 288, 408, 61 },
		{ 288, 409, 61 },
		{ 288, 410, 61 },
		{ 288, 411, 61 },
		{ 288, 412, 61 },
		{ 288, 413, 61 },
		{ 288, 414, 61 },
		{ 288, 415, 61 },
		{ 288, 416, 61 },
		{ 288, 417, 61 },
		{ 288, 418, 61 },
		{ 288, 419, 61 },
		{ 288, 420, 61 },
		{ 288, 421, 61 },
		{ 288, 422, 61 },
		{ 288, 423, 61 },
		{ 288, 424, 61 },
		{ 288, 425, 61 },
		{ 288, 426, 61 },
		{ 288, 427, 61 },
		{ 288, 428, 61 },
		{ 288, 429, 61 },
		{ 288, 430, 61 },
		{ 288, 431, 61 },
		{ 288, 432, 61 },
		{ 288, 433, 61 },
		{ 288, 434, 61 },
		{ 288, 435, 61 },
		{ 288, 436, 61 },
		{ 288, 437, 61 },
		{ 288, 438, 61 },
		{ 288, 439, 61 },
		{ 288, 440, 61 },
		{ 288, 441, 61 },
		{ 288, 442, 61 },
		{ 288, 443, 61 },
		{ 288, 444, 61 },
		{ 288, 445, 61 },
		{ 288, 446, 61 },
		{ 288, 447, 61 },
		{ 288, 448, 61 },
		{ 288, 449, 61 },
		{ 288, 450, 61 },
		{ 288, 451, 61 },
		{ 288, 452, 61 },
		{ 288, 453, 61 },
		{ 288, 454, 61 },
		{ 288, 455, 61 },
		{ 288, 456, 61 },
		{ 288, 457, 61 },
		{ 288, 458, 61 },
		{ 288, 459, 61 },
		{ 288, 460, 61 },
		{ 288, 461, 61 },
		{ 288, 462, 61 },
		{ 288, 463, 61 },
		{ 288, 464, 61 },
		{ 288, 465, 61 },
		{ 288, 466, 61 },
		{ 288, 467, 61 },
		{ 288, 468, 61 },
		{ 288, 469, 61 },
		{ 288, 470, 61 },
		{ 288, 471, 61 },
		{ 288, 472, 61 },
		{ 288, 473, 61 },
		{ 288, 474, 61 },
		{ 288, 475, 61 },
		{ 288, 476, 61 },
		{ 288, 477, 61 },
		{ 288, 478, 61 },
		{ 288, 479, 61 },
		{ 288, 480, 61 },
		{ 288, 481, 61 },
		{ 288, 482, 61 },
		{ 288, 483, 61 },
		{ 288, 484, 61 },
		{ 288, 485, 61 },
		{ 288, 486, 61 },
		{ 288, 487, 61 },
		{ 288, 488, 61 },
		{ 288, 489, 61 },
		{ 288, 490, 61 },
		{ 288, 491, 61 },
		{ 288, 492, 61 },
		{ 288, 493, 61 },
		{ 288, 494, 61 },
		{ 288, 495, 61 },
		{ 288, 496, 61 },
		{ 288, 497, 61 },
		{ 288, 498, 61 },
		{ 288, 499, 61 },
		{ 288, 500, 61 },
		{ 288, 501, 61 },
		{ 288, 502, 61 },
		{ 288, 503, 61 },
		{ 288, 504, 61 },
		{ 288, 505, 61 },
		{ 288, 506, 61 },
		{ 288, 507, 61 },
		{ 288, 508, 61 },
		{ 288, 509, 61 },
		{ 288, 510, 61 },
		{ 288, 511, 61 },
		{ 288, 512, 61 },
		{ 289, 386, 61 },
		{ 289, 387, 61 },
		{ 289, 388, 61 },
		{ 289, 389, 61 },
		{ 289, 390, 61 },
		{ 289, 391, 61 },
		{ 289, 392, 61 },
		{ 289, 393, 61 },
		{ 289, 394, 61 },
		{ 289, 395, 61 },
		{ 289, 396, 61 },
		{ 289, 397, 61 },
		{ 289, 398, 61 },
		{ 289, 399, 61 },
		{ 289, 400, 61 },
		{ 289, 401, 61 },
		{ 289, 402, 61 },
		{ 289, 403, 61 },
		{ 289, 404, 61 },
		{ 289, 405, 61 },
		{ 289, 406, 61 },
		{ 289, 407, 61 },
		{ 289, 408, 61 },
		{ 289, 409, 61 },
		{ 289, 410, 61 },
		{ 289, 411, 61 },
		{ 289, 412, 61 },
		{ 289, 413, 61 },
		{ 289, 414, 61 },
		{ 289, 415, 61 },
		{ 289, 416, 61 },
		{ 289, 417, 61 },
		{ 289, 418, 61 },
		{ 289, 419, 61 },
		{ 289, 420, 61 },
		{ 289, 421, 61 },
		{ 289, 422, 61 },
		{ 289, 423, 61 },
		{ 289, 424, 61 },
		{ 289, 425, 61 },
		{ 289, 426, 61 },
		{ 289, 427, 61 },
		{ 289, 428, 61 },
		{ 289, 429, 61 },
		{ 289, 430, 61 },
		{ 289, 431, 61 },
		{ 289, 432, 61 },
		{ 289, 433, 61 },
		{ 289, 434, 61 },
		{ 289, 435, 61 },
		{ 289, 436, 61 },
		{ 289, 437, 61 },
		{ 289, 438, 61 },
		{ 289, 439, 61 },
		{ 289, 440, 61 },
		{ 289, 441, 61 },
		{ 289, 442, 61 },
		{ 289, 443, 61 },
		{ 289, 444, 61 },
		{ 289, 445, 61 },
		{ 289, 446, 61 },
		{ 289, 447, 61 },
		{ 289, 448, 61 },
		{ 289, 449, 61 },
		{ 289, 450, 61 },
		{ 289, 451, 61 },
		{ 289, 452, 61 },
		{ 289, 453, 61 },
		{ 289, 454, 61 },
		{ 289, 455, 61 },
		{ 289, 456, 61 },
		{ 289, 457, 61 },
		{ 289, 458, 61 },
		{ 289, 459, 61 },
		{ 289, 460, 61 },
		{ 289, 461, 61 },
		{ 289, 462, 61 },
		{ 289, 463, 61 },
		{ 289, 464, 61 },
		{ 289, 465, 61 },
		{ 289, 466, 61 },
		{ 289, 467, 61 },
		{ 289, 468, 61 },
		{ 289, 469, 61 },
		{ 289, 470, 61 },
		{ 289, 471, 61 },
		{ 289, 472, 61 },
		{ 289, 473, 61 },
		{ 289, 474, 61 },
		{ 289, 475, 61 },
		{ 289, 476, 61 },
		{ 289, 477, 61 },
		{ 289, 478, 61 },
		{ 289, 479, 61 },
		{ 289, 480, 61 },
		{ 289, 481, 61 },
		{ 289, 482, 61 },
		{ 289, 483, 61 },
		{ 289, 484, 61 },
		{ 289, 485, 61 },
		{ 289, 486, 61 },
		{ 289, 487, 61 },
		{ 289, 488, 61 },
		{ 289, 489, 61 },
		{ 289, 490, 61 },
		{ 289, 491, 61 },
		{ 289, 492, 61 },
		{ 289, 493, 61 },
		{ 289, 494, 61 },
		{ 289, 495, 61 },
		{ 289, 496, 61 },
		{ 289, 497, 61 },
		{ 289, 498, 61 },
		{ 289, 499, 61 },
		{ 289, 500, 61 },
		{ 289, 501, 61 },
		{ 289, 502, 61 },
		{ 289, 503, 61 },
		{ 289, 504, 61 },
		{ 289, 505, 61 },
		{ 289, 506, 61 },
		{ 289, 507, 61 },
		{ 289, 508, 61 },
		{ 289, 509, 61 },
		{ 289, 510, 61 },
		{ 289, 511, 61 },
		{ 289, 512, 61 },
		{ 290, 390, 61 },
		{ 290, 391, 61 },
		{ 290, 392, 61 },
		{ 290, 393, 61 },
		{ 290, 394, 61 },
		{ 290, 395, 61 },
		{ 290, 396, 61 },
		{ 290, 397, 61 },
		{ 290, 398, 61 },
		{ 290, 399, 61 },
		{ 290, 400, 61 },
		{ 290, 401, 61 },
		{ 290, 402, 61 },
		{ 290, 403, 61 },
		{ 290, 404, 61 },
		{ 290, 405, 61 },
		{ 290, 406, 61 },
		{ 290, 407, 61 },
		{ 290, 408, 61 },
		{ 290, 409, 61 },
		{ 290, 410, 61 },
		{ 290, 411, 61 },
		{ 290, 412, 61 },
		{ 290, 413, 61 },
		{ 290, 414, 61 },
		{ 290, 415, 61 },
		{ 290, 416, 61 },
		{ 290, 417, 61 },
		{ 290, 418, 61 },
		{ 290, 419, 61 },
		{ 290, 420, 61 },
		{ 290, 421, 61 },
		{ 290, 422, 61 },
		{ 290, 423, 61 },
		{ 290, 424, 61 },
		{ 290, 425, 61 },
		{ 290, 426, 61 },
		{ 290, 427, 61 },
		{ 290, 428, 61 },
		{ 290, 429, 61 },
		{ 290, 430, 61 },
		{ 290, 431, 61 },
		{ 290, 432, 61 },
		{ 290, 433, 61 },
		{ 290, 434, 61 },
		{ 290, 435, 61 },
		{ 290, 436, 61 },
		{ 290, 437, 61 },
		{ 290, 438, 61 },
		{ 290, 439, 61 },
		{ 290, 440, 61 },
		{ 290, 441, 61 },
		{ 290, 442, 61 },
		{ 290, 443, 61 },
		{ 290, 444, 61 },
		{ 290, 445, 61 },
		{ 290, 446, 61 },
		{ 290, 447, 61 },
		{ 290, 448, 61 },
		{ 290, 449, 61 },
		{ 290, 450, 61 },
		{ 290, 451, 61 },
		{ 290, 452, 61 },
		{ 290, 453, 61 },
		{ 290, 454, 61 },
		{ 290, 455, 61 },
		{ 290, 456, 61 },
		{ 290, 457, 61 },
		{ 290, 458, 61 },
		{ 290, 459, 61 },
		{ 290, 460, 61 },
		{ 290, 461, 61 },
		{ 290, 462, 61 },
		{ 290, 463, 61 },
		{ 290, 464, 61 },
		{ 290, 465, 61 },
		{ 290, 466, 61 },
		{ 290, 467, 61 },
		{ 290, 468, 61 },
		{ 290, 469, 61 },
		{ 290, 470, 61 },
		{ 290, 471, 61 },
		{ 290, 472, 61 },
		{ 290, 473, 61 },
		{ 290, 474, 61 },
		{ 290, 475, 61 },
		{ 290, 476, 61 },
		{ 290, 477, 61 },
		{ 290, 478, 61 },
		{ 290, 479, 61 },
		{ 290, 480, 61 },
		{ 290, 481, 61 },
		{ 290, 482, 61 },
		{ 290, 483, 61 },
		{ 290, 484, 61 },
		{ 290, 485, 61 },
		{ 290, 486, 61 },
		{ 290, 487, 61 },
		{ 290, 488, 61 },
		{ 290, 489, 61 },
		{ 290, 490, 61 },
		{ 290, 491, 61 },
		{ 290, 492, 61 },
		{ 290, 493, 61 },
		{ 290, 494, 61 },
		{ 290, 495, 61 },
		{ 290, 496, 61 },
		{ 290, 497, 61 },
		{ 290, 498, 61 },
		{ 290, 499, 61 },
		{ 290, 500, 61 },
		{ 290, 501, 61 },
		{ 290, 502, 61 },
		{ 290, 503, 61 },
		{ 290, 504, 61 },
		{ 290, 505, 61 },
		{ 290, 506, 61 },
		{ 290, 507, 61 },
		{ 290, 508, 61 },
		{ 290, 509, 61 },
		{ 290, 510, 61 },
		{ 290, 511, 61 },
		{ 290, 512, 61 },
		{ 291, 394, 61 },
		{ 291, 395, 61 },
		{ 291, 396, 61 },
		{ 291, 397, 61 },
		{ 291, 398, 61 },
		{ 291, 399, 61 },
		{ 291, 400, 61 },
		{ 291, 401, 61 },
		{ 291, 402, 61 },
		{ 291, 403, 61 },
		{ 291, 404, 61 },
		{ 291, 405, 61 },
		{ 291, 406, 61 },
		{ 291, 407, 61 },
		{ 291, 408, 61 },
		{ 291, 409, 61 },
		{ 291, 410, 61 },
		{ 291, 411, 61 },
		{ 291, 412, 61 },
		{ 291, 413, 61 },
		{ 291, 414, 61 },
		{ 291, 415, 61 },
		{ 291, 416, 61 },
		{ 291, 417, 61 },
		{ 291, 418, 61 },
		{ 291, 419, 61 },
		{ 291, 420, 61 },
		{ 291, 421, 61 },
		{ 291, 422, 61 },
		{ 291, 423, 61 },
		{ 291, 424, 61 },
		{ 291, 425, 61 },
		{ 291, 426, 61 },
		{ 291, 427, 61 },
		{ 291, 428, 61 },
		{ 291, 429, 61 },
		{ 291, 430, 61 },
		{ 291, 431, 61 },
		{ 291, 432, 61 },
		{ 291, 433, 61 },
		{ 291, 434, 61 },
		{ 291, 435, 61 },
		{ 291, 436, 61 },
		{ 291, 437, 61 },
		{ 291, 438, 61 },
		{ 291, 439, 61 },
		{ 291, 440, 61 },
		{ 291, 441, 61 },
		{ 291, 442, 61 },
		{ 291, 443, 61 },
		{ 291, 444, 61 },
		{ 291, 445, 61 },
		{ 291, 446, 61 },
		{ 291, 447, 61 },
		{ 291, 448, 61 },
		{ 291, 449, 61 },
		{ 291, 450, 61 },
		{ 291, 451, 61 },
		{ 291, 452, 61 },
		{ 291, 453, 61 },
		{ 291, 454, 61 },
		{ 291, 455, 61 },
		{ 291, 456, 61 },
		{ 291, 457, 61 },
		{ 291, 458, 61 },
		{ 291, 459, 61 },
		{ 291, 460, 61 },
		{ 291, 461, 61 },
		{ 291, 462, 61 },
		{ 291, 463, 61 },
		{ 291, 464, 61 },
		{ 291, 465, 61 },
		{ 291, 466, 61 },
		{ 291, 467, 61 },
		{ 291, 468, 61 },
		{ 291, 469, 61 },
		{ 291, 470, 61 },
		{ 291, 471, 61 },
		{ 291, 472, 61 },
		{ 291, 473, 61 },
		{ 291, 474, 61 },
		{ 291, 475, 61 },
		{ 291, 476, 61 },
		{ 291, 477, 61 },
		{ 291, 478, 61 },
		{ 291, 479, 61 },
		{ 291, 480, 61 },
		{ 291, 481, 61 },
		{ 291, 482, 61 },
		{ 291, 483, 61 },
		{ 291, 484, 61 },
		{ 291, 485, 61 },
		{ 291, 486, 61 },
		{ 291, 487, 61 },
		{ 291, 488, 61 },
		{ 291, 489, 61 },
		{ 291, 490, 61 },
		{ 291, 491, 61 },
		{ 291, 492, 61 },
		{ 291, 493, 61 },
		{ 291, 494, 61 },
		{ 291, 495, 61 },
		{ 291, 496, 61 },
		{ 291, 497, 61 },
		{ 291, 498, 61 },
		{ 291, 499, 61 },
		{ 291, 500, 61 },
		{ 291, 501, 61 },
		{ 291, 502, 61 },
		{ 291, 503, 61 },
		{ 291, 504, 61 },
		{ 291, 505, 61 },
		{ 291, 506, 61 },
		{ 291, 507, 61 },
		{ 291, 508, 61 },
		{ 291, 509, 61 },
		{ 291, 510, 61 },
		{ 291, 511, 61 },
		{ 291, 512, 61 },
		{ 292, 398, 61 },
		{ 292, 399, 61 },
		{ 292, 400, 61 },
		{ 292, 401, 61 },
		{ 292, 402, 61 },
		{ 292, 403, 61 },
		{ 292, 404, 61 },
		{ 292, 405, 61 },
		{ 292, 406, 61 },
		{ 292, 407, 61 },
		{ 292, 408, 61 },
		{ 292, 409, 61 },
		{ 292, 410, 61 },
		{ 292, 411, 61 },
		{ 292, 412, 61 },
		{ 292, 413, 61 },
		{ 292, 414, 61 },
		{ 292, 415, 61 },
		{ 292, 416, 61 },
		{ 292, 417, 61 },
		{ 292, 418, 61 },
		{ 292, 419, 61 },
		{ 292, 420, 61 },
		{ 292, 421, 61 },
		{ 292, 422, 61 },
		{ 292, 423, 61 },
		{ 292, 424, 61 },
		{ 292, 425, 61 },
		{ 292, 426, 61 },
		{ 292, 427, 61 },
		{ 292, 428, 61 },
		{ 292, 429, 61 },
		{ 292, 430, 61 },
		{ 292, 431, 61 },
		{ 292, 432, 61 },
		{ 292, 433, 61 },
		{ 292, 434, 61 },
		{ 292, 435, 61 },
		{ 292, 436, 61 },
		{ 292, 437, 61 },
		{ 292, 438, 61 },
		{ 292, 439, 61 },
		{ 292, 440, 61 },
		{ 292, 441, 61 },
		{ 292, 442, 61 },
		{ 292, 443, 61 },
		{ 292, 444, 61 },
		{ 292, 445, 61 },
		{ 292, 446, 61 },
		{ 292, 447, 61 },
		{ 292, 448, 61 },
		{ 292, 449, 61 },
		{ 292, 450, 61 },
		{ 292, 451, 61 },
		{ 292, 452, 61 },
		{ 292, 453, 61 },
		{ 292, 454, 61 },
		{ 292, 455, 61 },
		{ 292, 456, 61 },
		{ 292, 457, 61 },
		{ 292, 458, 61 },
		{ 292, 459, 61 },
		{ 292, 460, 61 },
		{ 292, 461, 61 },
		{ 292, 462, 61 },
		{ 292, 463, 61 },
		{ 292, 464, 61 },
		{ 292, 465, 61 },
		{ 292, 466, 61 },
		{ 292, 467, 61 },
		{ 292, 468, 61 },
		{ 292, 469, 61 },
		{ 292, 470, 61 },
		{ 292, 471, 61 },
		{ 292, 472, 61 },
		{ 292, 473, 61 },
		{ 292, 474, 61 },
		{ 292, 475, 61 },
		{ 292, 476, 61 },
		{ 292, 477, 61 },
		{ 292, 478, 61 },
		{ 292, 479, 61 },
		{ 292, 480, 61 },
		{ 292, 481, 61 },
		{ 292, 482, 61 },
		{ 292, 483, 61 },
		{ 292, 484, 61 },
		{ 292, 485, 61 },
		{ 292, 486, 61 },
		{ 292, 487, 61 },
		{ 292, 488, 61 },
		{ 292, 489, 61 },
		{ 292, 490, 61 },
		{ 292, 491, 61 },
		{ 292, 492, 61 },
		{ 292, 493, 61 },
		{ 292, 494, 61 },
		{ 292, 495, 61 },
		{ 292, 496, 61 },
		{ 292, 497, 61 },
		{ 292, 498, 61 },
		{ 292, 499, 61 },
		{ 292, 500, 61 },
		{ 292, 501, 61 },
		{ 292, 502, 61 },
		{ 292, 503, 61 },
		{ 292, 504, 61 },
		{ 292, 505, 61 },
		{ 292, 506, 61 },
		{ 292, 507, 61 },
		{ 292, 508, 61 },
		{ 292, 509, 61 },
		{ 292, 510, 61 },
		{ 292, 511, 61 },
		{ 292, 512, 61 },
		{ 293, 402, 61 },
		{ 293, 403, 61 },
		{ 293, 404, 61 },
		{ 293, 405, 61 },
		{ 293, 406, 61 },
		{ 293, 407, 61 },
		{ 293, 408, 61 },
		{ 293, 409, 61 },
		{ 293, 410, 61 },
		{ 293, 411, 61 },
		{ 293, 412, 61 },
		{ 293, 413, 61 },
		{ 293, 414, 61 },
		{ 293, 415, 61 },
		{ 293, 416, 61 },
		{ 293, 417, 61 },
		{ 293, 418, 61 },
		{ 293, 419, 61 },
		{ 293, 420, 61 },
		{ 293, 421, 61 },
		{ 293, 422, 61 },
		{ 293, 423, 61 },
		{ 293, 424, 61 },
		{ 293, 425, 61 },
		{ 293, 426, 61 },
		{ 293, 427, 61 },
		{ 293, 428, 61 },
		{ 293, 429, 61 },
		{ 293, 430, 61 },
		{ 293, 431, 61 },
		{ 293, 432, 61 },
		{ 293, 433, 61 },
		{ 293, 434, 61 },
		{ 293, 435, 61 },
		{ 293, 436, 61 },
		{ 293, 437, 61 },
		{ 293, 438, 61 },
		{ 293, 439, 61 },
		{ 293, 440, 61 },
		{ 293, 441, 61 },
		{ 293, 442, 61 },
		{ 293, 443, 61 },
		{ 293, 444, 61 },
		{ 293, 445, 61 },
		{ 293, 446, 61 },
		{ 293, 447, 61 },
		{ 293, 448, 61 },
		{ 293, 449, 61 },
		{ 293, 450, 61 },
		{ 293, 451, 61 },
		{ 293, 452, 61 },
		{ 293, 453, 61 },
		{ 293, 454, 61 },
		{ 293, 455, 61 },
		{ 293, 456, 61 },
		{ 293, 457, 61 },
		{ 293, 458, 61 },
		{ 293, 459, 61 },
		{ 293, 460, 61 },
		{ 293, 461, 61 },
		{ 293, 462, 61 },
		{ 293, 463, 61 },
		{ 293, 464, 61 },
		{ 293, 465, 61 },
		{ 293, 466, 61 },
		{ 293, 467, 61 },
		{ 293, 468, 61 },
		{ 293, 469, 61 },
		{ 293, 470, 61 },
		{ 293, 471, 61 },
		{ 293, 472, 61 },
		{ 293, 473, 61 },
		{ 293, 474, 61 },
		{ 293, 475, 61 },
		{ 293, 476, 61 },
		{ 293, 477, 61 },
		{ 293, 478, 61 },
		{ 293, 479, 61 },
		{ 293, 480, 61 },
		{ 293, 481, 61 },
		{ 293, 482, 61 },
		{ 293, 483, 61 },
		{ 293, 484, 61 },
		{ 293, 485, 61 },
		{ 293, 486, 61 },
		{ 293, 487, 61 },
		{ 293, 488, 61 },
		{ 293, 489, 61 },
		{ 293, 490, 61 },
		{ 293, 491, 61 },
		{ 293, 492, 61 },
		{ 293, 493, 61 },
		{ 293, 494, 61 },
		{ 293, 495, 61 },
		{ 293, 496, 61 },
		{ 293, 497, 61 },
		{ 293, 498, 61 },
		{ 293, 499, 61 },
		{ 293, 500, 61 },
		{ 293, 501, 61 },
		{ 293, 502, 61 },
		{ 293, 503, 61 },
		{ 293, 504, 61 },
		{ 293, 505, 61 },
		{ 293, 506, 61 },
		{ 293, 507, 61 },
		{ 293, 508, 61 },
		{ 293, 509, 61 },
		{ 293, 510, 61 },
		{ 293, 511, 61 },
		{ 293, 512, 61 },
		{ 294, 406, 61 },
		{ 294, 407, 61 },
		{ 294, 408, 61 },
		{ 294, 409, 61 },
		{ 294, 410, 61 },
		{ 294, 411, 61 },
		{ 294, 412, 61 },
		{ 294, 413, 61 },
		{ 294, 414, 61 },
		{ 294, 415, 61 },
		{ 294, 416, 61 },
		{ 294, 417, 61 },
		{ 294, 418, 61 },
		{ 294, 419, 61 },
		{ 294, 420, 61 },
		{ 294, 421, 61 },
		{ 294, 422, 61 },
		{ 294, 423, 61 },
		{ 294, 424, 61 },
		{ 294, 425, 61 },
		{ 294, 426, 61 },
		{ 294, 427, 61 },
		{ 294, 428, 61 },
		{ 294, 429, 61 },
		{ 294, 430, 61 },
		{ 294, 431, 61 },
		{ 294, 432, 61 },
		{ 294, 433, 61 },
		{ 294, 434, 61 },
		{ 294, 435, 61 },
		{ 294, 436, 61 },
		{ 294, 437, 61 },
		{ 294, 438, 61 },
		{ 294, 439, 61 },
		{ 294, 440, 61 },
		{ 294, 441, 61 },
		{ 294, 442, 61 },
		{ 294, 443, 61 },
		{ 294, 444, 61 },
		{ 294, 445, 61 },
		{ 294, 446, 61 },
		{ 294, 447, 61 },
		{ 294, 448, 61 },
		{ 294, 449, 61 },
		{ 294, 450, 61 },
		{ 294, 451, 61 },
		{ 294, 452, 61 },
		{ 294, 453, 61 },
		{ 294, 454, 61 },
		{ 294, 455, 61 },
		{ 294, 456, 61 },
		{ 294, 457, 61 },
		{ 294, 458, 61 },
		{ 294, 459, 61 },
		{ 294, 460, 61 },
		{ 294, 461, 61 },
		{ 294, 462, 61 },
		{ 294, 463, 61 },
		{ 294, 464, 61 },
		{ 294, 465, 61 },
		{ 294, 466, 61 },
		{ 294, 467, 61 },
		{ 294, 468, 61 },
		{ 294, 469, 61 },
		{ 294, 470, 61 },
		{ 294, 471, 61 },
		{ 294, 472, 61 },
		{ 294, 473, 61 },
		{ 294, 474, 61 },
		{ 294, 475, 61 },
		{ 294, 476, 61 },
		{ 294, 477, 61 },
		{ 294, 478, 61 },
		{ 294, 479, 61 },
		{ 294, 480, 61 },
		{ 294, 481, 61 },
		{ 294, 482, 61 },
		{ 294, 483, 61 },
		{ 294, 484, 61 },
		{ 294, 485, 61 },
		{ 294, 486, 61 },
		{ 294, 487, 61 },
		{ 294, 488, 61 },
		{ 294, 489, 61 },
		{ 294, 490, 61 },
		{ 294, 491, 61 },
		{ 294, 492, 61 },
		{ 294, 493, 61 },
		{ 294, 494, 61 },
		{ 294, 495, 61 },
		{ 294, 496, 61 },
		{ 294, 497, 61 },
		{ 294, 498, 61 },
		{ 294, 499, 61 },
		{ 294, 500, 61 },
		{ 294, 501, 61 },
		{ 294, 502, 61 },
		{ 294, 503, 61 },
		{ 294, 504, 61 },
		{ 294, 505, 61 },
		{ 294, 506, 61 },
		{ 294, 507, 61 },
		{ 294, 508, 61 },
		{ 294, 509, 61 },
		{ 294, 510, 61 },
		{ 294, 511, 61 },
		{ 294, 512, 61 },
		{ 295, 410, 61 },
		{ 295, 411, 61 },
		{ 295, 412, 61 },
		{ 295, 413, 61 },
		{ 295, 414, 61 },
		{ 295, 415, 61 },
		{ 295, 416, 61 },
		{ 295, 417, 61 },
		{ 295, 418, 61 },
		{ 295, 419, 61 },
		{ 295, 420, 61 },
		{ 295, 421, 61 },
		{ 295, 422, 61 },
		{ 295, 423, 61 },
		{ 295, 424, 61 },
		{ 295, 425, 61 },
		{ 295, 426, 61 },
		{ 295, 427, 61 },
		{ 295, 428, 61 },
		{ 295, 429, 61 },
		{ 295, 430, 61 },
		{ 295, 431, 61 },
		{ 295, 432, 61 },
		{ 295, 433, 61 },
		{ 295, 434, 61 },
		{ 295, 435, 61 },
		{ 295, 436, 61 },
		{ 295, 437, 61 },
		{ 295, 438, 61 },
		{ 295, 439, 61 },
		{ 295, 440, 61 },
		{ 295, 441, 61 },
		{ 295, 442, 61 },
		{ 295, 443, 61 },
		{ 295, 444, 61 },
		{ 295, 445, 61 },
		{ 295, 446, 61 },
		{ 295, 447, 61 },
		{ 295, 448, 61 },
		{ 295, 449, 61 },
		{ 295, 450, 61 },
		{ 295, 451, 61 },
		{ 295, 452, 61 },
		{ 295, 453, 61 },
		{ 295, 454, 61 },
		{ 295, 455, 61 },
		{ 295, 456, 61 },
		{ 295, 457, 61 },
		{ 295, 458, 61 },
		{ 295, 459, 61 },
		{ 295, 460, 61 },
		{ 295, 461, 61 },
		{ 295, 462, 61 },
		{ 295, 463, 61 },
		{ 295, 464, 61 },
		{ 295, 465, 61 },
		{ 295, 466, 61 },
		{ 295, 467, 61 },
		{ 295, 468, 61 },
		{ 295, 469, 61 },
		{ 295, 470, 61 },
		{ 295, 471, 61 },
		{ 295, 472, 61 },
		{ 295, 473, 61 },
		{ 295, 474, 61 },
		{ 295, 475, 61 },
		{ 295, 476, 61 },
		{ 295, 477, 61 },
		{ 295, 478, 61 },
		{ 295, 479, 61 },
		{ 295, 480, 61 },
		{ 295, 481, 61 },
		{ 295, 482, 61 },
		{ 295, 483, 61 },
		{ 295, 484, 61 },
		{ 295, 485, 61 },
		{ 295, 486, 61 },
		{ 295, 487, 61 },
		{ 295, 488, 61 },
		{ 295, 489, 61 },
		{ 295, 490, 61 },
		{ 295, 491, 61 },
		{ 295, 492, 61 },
		{ 295, 493, 61 },
		{ 295, 494, 61 },
		{ 295, 495, 61 },
		{ 295, 496, 61 },
		{ 295, 497, 61 },
		{ 295, 498, 61 },
		{ 295, 499, 61 },
		{ 295, 500, 61 },
		{ 295, 501, 61 },
		{ 295, 502, 61 },
		{ 295, 503, 61 },
		{ 295, 504, 61 },
		{ 295, 505, 61 },
		{ 295, 506, 61 },
		{ 295, 507, 61 },
		{ 295, 508, 61 },
		{ 295, 509, 61 },
		{ 295, 510, 61 },
		{ 295, 511, 61 },
		{ 295, 512, 61 },
		{ 296, 414, 61 },
		{ 296, 415, 61 },
		{ 296, 416, 61 },
		{ 296, 417, 61 },
		{ 296, 418, 61 },
		{ 296, 419, 61 },
		{ 296, 420, 61 },
		{ 296, 421, 61 },
		{ 296, 422, 61 },
		{ 296, 423, 61 },
		{ 296, 424, 61 },
		{ 296, 425, 61 },
		{ 296, 426, 61 },
		{ 296, 427, 61 },
		{ 296, 428, 61 },
		{ 296, 429, 61 },
		{ 296, 430, 61 },
		{ 296, 431, 61 },
		{ 296, 432, 61 },
		{ 296, 433, 61 },
		{ 296, 434, 61 },
		{ 296, 435, 61 },
		{ 296, 436, 61 },
		{ 296, 437, 61 },
		{ 296, 438, 61 },
		{ 296, 439, 61 },
		{ 296, 440, 61 },
		{ 296, 441, 61 },
		{ 296, 442, 61 },
		{ 296, 443, 61 },
		{ 296, 444, 61 },
		{ 296, 445, 61 },
		{ 296, 446, 61 },
		{ 296, 447, 61 },
		{ 296, 448, 61 },
		{ 296, 449, 61 },
		{ 296, 450, 61 },
		{ 296, 451, 61 },
		{ 296, 452, 61 },
		{ 296, 453, 61 },
		{ 296, 454, 61 },
		{ 296, 455, 61 },
		{ 296, 456, 61 },
		{ 296, 457, 61 },
		{ 296, 458, 61 },
		{ 296, 459, 61 },
		{ 296, 460, 61 },
		{ 296, 461, 61 },
		{ 296, 462, 61 },
		{ 296, 463, 61 },
		{ 296, 464, 61 },
		{ 296, 465, 61 },
		{ 296, 466, 61 },
		{ 296, 467, 61 },
		{ 296, 468, 61 },
		{ 296, 469, 61 },
		{ 296, 470, 61 },
		{ 296, 471, 61 },
		{ 296, 472, 61 },
		{ 296, 473, 61 },
		{ 296, 474, 61 },
		{ 296, 475, 61 },
		{ 296, 476, 61 },
		{ 296, 477, 61 },
		{ 296, 478, 61 },
		{ 296, 479, 61 },
		{ 296, 480, 61 },
		{ 296, 481, 61 },
		{ 296, 482, 61 },
		{ 296, 483, 61 },
		{ 296, 484, 61 },
		{ 296, 485, 61 },
		{ 296, 486, 61 },
		{ 296, 487, 61 },
		{ 296, 488, 61 },
		{ 296, 489, 61 },
		{ 296, 490, 61 },
		{ 296, 491, 61 },
		{ 296, 492, 61 },
		{ 296, 493, 61 },
		{ 296, 494, 61 },
		{ 296, 495, 61 },
		{ 296, 496, 61 },
		{ 296, 497, 61 },
		{ 296, 498, 61 },
		{ 296, 499, 61 },
		{ 296, 500, 61 },
		{ 296, 501, 61 },
		{ 296, 502, 61 },
		{ 296, 503, 61 },
		{ 296, 504, 61 },
		{ 296, 505, 61 },
		{ 296, 506, 61 },
		{ 296, 507, 61 },
		{ 296, 508, 61 },
		{ 296, 509, 61 },
		{ 296, 510, 61 },
		{ 296, 511, 61 },
		{ 296, 512, 61 },
		{ 297, 418, 61 },
		{ 297, 419, 61 },
		{ 297, 420, 61 },
		{ 297, 421, 61 },
		{ 297, 422, 61 },
		{ 297, 423, 61 },
		{ 297, 424, 61 },
		{ 297, 425, 61 },
		{ 297, 426, 61 },
		{ 297, 427, 61 },
		{ 297, 428, 61 },
		{ 297, 429, 61 },
		{ 297, 430, 61 },
		{ 297, 431, 61 },
		{ 297, 432, 61 },
		{ 297, 433, 61 },
		{ 297, 434, 61 },
		{ 297, 435, 61 },
		{ 297, 436, 61 },
		{ 297, 437, 61 },
		{ 297, 438, 61 },
		{ 297, 439, 61 },
		{ 297, 440, 61 },
		{ 297, 441, 61 },
		{ 297, 442, 61 },
		{ 297, 443, 61 },
		{ 297, 444, 61 },
		{ 297, 445, 61 },
		{ 297, 446, 61 },
		{ 297, 447, 61 },
		{ 297, 448, 61 },
		{ 297, 449, 61 },
		{ 297, 450, 61 },
		{ 297, 451, 61 },
		{ 297, 452, 61 },
		{ 297, 453, 61 },
		{ 297, 454, 61 },
		{ 297, 455, 61 },
		{ 297, 456, 61 },
		{ 297, 457, 61 },
		{ 297, 458, 61 },
		{ 297, 459, 61 },
		{ 297, 460, 61 },
		{ 297, 461, 61 },
		{ 297, 462, 61 },
		{ 297, 463, 61 },
		{ 297, 464, 61 },
		{ 297, 465, 61 },
		{ 297, 466, 61 },
		{ 297, 467, 61 },
		{ 297, 468, 61 },
		{ 297, 469, 61 },
		{ 297, 470, 61 },
		{ 297, 471, 61 },
		{ 297, 472, 61 },
		{ 297, 473, 61 },
		{ 297, 474, 61 },
		{ 297, 475, 61 },
		{ 297, 476, 61 },
		{ 297, 477, 61 },
		{ 297, 478, 61 },
		{ 297, 479, 61 },
		{ 297, 480, 61 },
		{ 297, 481, 61 },
		{ 297, 482, 61 },
		{ 297, 483, 61 },
		{ 297, 484, 61 },
		{ 297, 485, 61 },
		{ 297, 486, 61 },
		{ 297, 487, 61 },
		{ 297, 488, 61 },
		{ 297, 489, 61 },
		{ 297, 490, 61 },
		{ 297, 491, 61 },
		{ 297, 492, 61 },
		{ 297, 493, 61 },
		{ 297, 494, 61 },
		{ 297, 495, 61 },
		{ 297, 496, 61 },
		{ 297, 497, 61 },
		{ 297, 498, 61 },
		{ 297, 499, 61 },
		{ 297, 500, 61 },
		{ 297, 501, 61 },
		{ 297, 502, 61 },
		{ 297, 503, 61 },
		{ 297, 504, 61 },
		{ 297, 505, 61 },
		{ 297, 506, 61 },
		{ 297, 507, 61 },
		{ 297, 508, 61 },
		{ 297, 509, 61 },
		{ 297, 510, 61 },
		{ 297, 511, 61 },
		{ 297, 512, 61 },
		{ 298, 422, 61 },
		{ 298, 423, 61 },
		{ 298, 424, 61 },
		{ 298, 425, 61 },
		{ 298, 426, 61 },
		{ 298, 427, 61 },
		{ 298, 428, 61 },
		{ 298, 429, 61 },
		{ 298, 430, 61 },
		{ 298, 431, 61 },
		{ 298, 432, 61 },
		{ 298, 433, 61 },
		{ 298, 434, 61 },
		{ 298, 435, 61 },
		{ 298, 436, 61 },
		{ 298, 437, 61 },
		{ 298, 438, 61 },
		{ 298, 439, 61 },
		{ 298, 440, 61 },
		{ 298, 441, 61 },
		{ 298, 442, 61 },
		{ 298, 443, 61 },
		{ 298, 444, 61 },
		{ 298, 445, 61 },
		{ 298, 446, 61 },
		{ 298, 447, 61 },
		{ 298, 448, 61 },
		{ 298, 449, 61 },
		{ 298, 450, 61 },
		{ 298, 451, 61 },
		{ 298, 452, 61 },
		{ 298, 453, 61 },
		{ 298, 454, 61 },
		{ 298, 455, 61 },
		{ 298, 456, 61 },
		{ 298, 457, 61 },
		{ 298, 458, 61 },
		{ 298, 459, 61 },
		{ 298, 460, 61 },
		{ 298, 461, 61 },
		{ 298, 462, 61 },
		{ 298, 463, 61 },
		{ 298, 464, 61 },
		{ 298, 465, 61 },
		{ 298, 466, 61 },
		{ 298, 467, 61 },
		{ 298, 468, 61 },
		{ 298, 469, 61 },
		{ 298, 470, 61 },
		{ 298, 471, 61 },
		{ 298, 472, 61 },
		{ 298, 473, 61 },
		{ 298, 474, 61 },
		{ 298, 475, 61 },
		{ 298, 476, 61 },
		{ 298, 477, 61 },
		{ 298, 478, 61 },
		{ 298, 479, 61 },
		{ 298, 480, 61 },
		{ 298, 481, 61 },
		{ 298, 482, 61 },
		{ 298, 483, 61 },
		{ 298, 484, 61 },
		{ 298, 485, 61 },
		{ 298, 486, 61 },
		{ 298, 487, 61 },
		{ 298, 488, 61 },
		{ 298, 489, 61 },
		{ 298, 490, 61 },
		{ 298, 491, 61 },
		{ 298, 492, 61 },
		{ 298, 493, 61 },
		{ 298, 494, 61 },
		{ 298, 495, 61 },
		{ 298, 496, 61 },
		{ 298, 497, 61 },
		{ 298, 498, 61 },
		{ 298, 499, 61 },
		{ 298, 500, 61 },
		{ 298, 501, 61 },
		{ 298, 502, 61 },
		{ 298, 503, 61 },
		{ 298, 504, 61 },
		{ 298, 505, 61 },
		{ 298, 506, 61 },
		{ 298, 507, 61 },
		{ 298, 508, 61 },
		{ 298, 509, 61 },
		{ 298, 510, 61 },
		{ 298, 511, 61 },
		{ 298, 512, 61 },
		{ 299, 426, 61 },
		{ 299, 427, 61 },
		{ 299, 428, 61 },
		{ 299, 429, 61 },
		{ 299, 430, 61 },
		{ 299, 431, 61 },
		{ 299, 432, 61 },
		{ 299, 433, 61 },
		{ 299, 434, 61 },
		{ 299, 435, 61 },
		{ 299, 436, 61 },
		{ 299, 437, 61 },
		{ 299, 438, 61 },
		{ 299, 439, 61 },
		{ 299, 440, 61 },
		{ 299, 441, 61 },
		{ 299, 442, 61 },
		{ 299, 443, 61 },
		{ 299, 444, 61 },
		{ 299, 445, 61 },
		{ 299, 446, 61 },
		{ 299, 447, 61 },
		{ 299, 448, 61 },
		{ 299, 449, 61 },
		{ 299, 450, 61 },
		{ 299, 451, 61 },
		{ 299, 452, 61 },
		{ 299, 453, 61 },
		{ 299, 454, 61 },
		{ 299, 455, 61 },
		{ 299, 456, 61 },
		{ 299, 457, 61 },
		{ 299, 458, 61 },
		{ 299, 459, 61 },
		{ 299, 460, 61 },
		{ 299, 461, 61 },
		{ 299, 462, 61 },
		{ 299, 463, 61 },
		{ 299, 464, 61 },
		{ 299, 465, 61 },
		{ 299, 466, 61 },
		{ 299, 467, 61 },
		{ 299, 468, 61 },
		{ 299, 469, 61 },
		{ 299, 470, 61 },
		{ 299, 471, 61 },
		{ 299, 472, 61 },
		{ 299, 473, 61 },
		{ 299, 474, 61 },
		{ 299, 475, 61 },
		{ 299, 476, 61 },
		{ 299, 477, 61 },
		{ 299, 478, 61 },
		{ 299, 479, 61 },
		{ 299, 480, 61 },
		{ 299, 481, 61 },
		{ 299, 482, 61 },
		{ 299, 483, 61 },
		{ 299, 484, 61 },
		{ 299, 485, 61 },
		{ 299, 486, 61 },
		{ 299, 487, 61 },
		{ 299, 488, 61 },
		{ 299, 489, 61 },
		{ 299, 490, 61 },
		{ 299, 491, 61 },
		{ 299, 492, 61 },
		{ 299, 493, 61 },
		{ 299, 494, 61 },
		{ 299, 495, 61 },
		{ 299, 496, 61 },
		{ 299, 497, 61 },
		{ 299, 498, 61 },
		{ 299, 499, 61 },
		{ 299, 500, 61 },
		{ 299, 501, 61 },
		{ 299, 502, 61 },
		{ 299, 503, 61 },
		{ 299, 504, 61 },
		{ 299, 505, 61 },
		{ 299, 506, 61 },
		{ 299, 507, 61 },
		{ 299, 508, 61 },
		{ 299, 509, 61 },
		{ 299, 510, 61 },
		{ 299, 511, 61 },
		{ 299, 512, 61 },
		{ 300, 430, 61 },
		{ 300, 431, 61 },
		{ 300, 432, 61 },
		{ 300, 433, 61 },
		{ 300, 434, 61 },
		{ 300, 435, 61 },
		{ 300, 436, 61 },
		{ 300, 437, 61 },
		{ 300, 438, 61 },
		{ 300, 439, 61 },
		{ 300, 440, 61 },
		{ 300, 441, 61 },
		{ 300, 442, 61 },
		{ 300, 443, 61 },
		{ 300, 444, 61 },
		{ 300, 445, 61 },
		{ 300, 446, 61 },
		{ 300, 447, 61 },
		{ 300, 448, 61 },
		{ 300, 449, 61 },
		{ 300, 450, 61 },
		{ 300, 451, 61 },
		{ 300, 452, 61 },
		{ 300, 453, 61 },
		{ 300, 454, 61 },
		{ 300, 455, 61 },
		{ 300, 456, 61 },
		{ 300, 457, 61 },
		{ 300, 458, 61 },
		{ 300, 459, 61 },
		{ 300, 460, 61 },
		{ 300, 461, 61 },
		{ 300, 462, 61 },
		{ 300, 463, 61 },
		{ 300, 464, 61 },
		{ 300, 465, 61 },
		{ 300, 466, 61 },
		{ 300, 467, 61 },
		{ 300, 468, 61 },
		{ 300, 469, 61 },
		{ 300, 470, 61 },
		{ 300, 471, 61 },
		{ 300, 472, 61 },
		{ 300, 473, 61 },
		{ 300, 474, 61 },
		{ 300, 475, 61 },
		{ 300, 476, 61 },
		{ 300, 477, 61 },
		{ 300, 478, 61 },
		{ 300, 479, 61 },
		{ 300, 480, 61 },
		{ 300, 481, 61 },
		{ 300, 482, 61 },
		{ 300, 483, 61 },
		{ 300, 484, 61 },
		{ 300, 485, 61 },
		{ 300, 486, 61 },
		{ 300, 487, 61 },
		{ 300, 488, 61 },
		{ 300, 489, 61 },
		{ 300, 490, 61 },
		{ 300, 491, 61 },
		{ 300, 492, 61 },
		{ 300, 493, 61 },
		{ 300, 494, 61 },
		{ 300, 495, 61 },
		{ 300, 496, 61 },
		{ 300, 497, 61 },
		{ 300, 498, 61 },
		{ 300, 499, 61 },
		{ 300, 500, 61 },
		{ 300, 501, 61 },
		{ 300, 502, 61 },
		{ 300, 503, 61 },
		{ 300, 504, 61 },
		{ 300, 505, 61 },
		{ 300, 506, 61 },
		{ 300, 507, 61 },
		{ 300, 508, 61 },
		{ 300, 509, 61 },
		{ 300, 510, 61 },
		{ 300, 511, 61 },
		{ 300, 512, 61 },
		{ 301, 434, 61 },
		{ 301, 435, 61 },
		{ 301, 436, 61 },
		{ 301, 437, 61 },
		{ 301, 438, 61 },
		{ 301, 439, 61 },
		{ 301, 440, 61 },
		{ 301, 441, 61 },
		{ 301, 442, 61 },
		{ 301, 443, 61 },
		{ 301, 444, 61 },
		{ 301, 445, 61 },
		{ 301, 446, 61 },
		{ 301, 447, 61 },
		{ 301, 448, 61 },
		{ 301, 449, 61 },
		{ 301, 450, 61 },
		{ 301, 451, 61 },
		{ 301, 452, 61 },
		{ 301, 453, 61 },
		{ 301, 454, 61 },
		{ 301, 455, 61 },
		{ 301, 456, 61 },
		{ 301, 457, 61 },
		{ 301, 458, 61 },
		{ 301, 459, 61 },
		{ 301, 460, 61 },
		{ 301, 461, 61 },
		{ 301, 462, 61 },
		{ 301, 463, 61 },
		{ 301, 464, 61 },
		{ 301, 465, 61 },
		{ 301, 466, 61 },
		{ 301, 467, 61 },
		{ 301, 468, 61 },
		{ 301, 469, 61 },
		{ 301, 470, 61 },
		{ 301, 471, 61 },
		{ 301, 472, 61 },
		{ 301, 473, 61 },
		{ 301, 474, 61 },
		{ 301, 475, 61 },
		{ 301, 476, 61 },
		{ 301, 477, 61 },
		{ 301, 478, 61 },
		{ 301, 479, 61 },
		{ 301, 480, 61 },
		{ 301, 481, 61 },
		{ 301, 482, 61 },
		{ 301, 483, 61 },
		{ 301, 484, 61 },
		{ 301, 485, 61 },
		{ 301, 486, 61 },
		{ 301, 487, 61 },
		{ 301, 488, 61 },
		{ 301, 489, 61 },
		{ 301, 490, 61 },
		{ 301, 491, 61 },
		{ 301, 492, 61 },
		{ 301, 493, 61 },
		{ 301, 494, 61 },
		{ 301, 495, 61 },
		{ 301, 496, 61 },
		{ 301, 497, 61 },
		{ 301, 498, 61 },
		{ 301, 499, 61 },
		{ 301, 500, 61 },
		{ 301, 501, 61 },
		{ 301, 502, 61 },
		{ 301, 503, 61 },
		{ 301, 504, 61 },
		{ 301, 505, 61 },
		{ 301, 506, 61 },
		{ 301, 507, 61 },
		{ 301, 508, 61 },
		{ 301, 509, 61 },
		{ 301, 510, 61 },
		{ 301, 511, 61 },
		{ 301, 512, 61 },
		{ 302, 438, 61 },
		{ 302, 439, 61 },
		{ 302, 440, 61 },
		{ 302, 441, 61 },
		{ 302, 442, 61 },
		{ 302, 443, 61 },
		{ 302, 444, 61 },
		{ 302, 445, 61 },
		{ 302, 446, 61 },
		{ 302, 447, 61 },
		{ 302, 448, 61 },
		{ 302, 449, 61 },
		{ 302, 450, 61 },
		{ 302, 451, 61 },
		{ 302, 452, 61 },
		{ 302, 453, 61 },
		{ 302, 454, 61 },
		{ 302, 455, 61 },
		{ 302, 456, 61 },
		{ 302, 457, 61 },
		{ 302, 458, 61 },
		{ 302, 459, 61 },
		{ 302, 460, 61 },
		{ 302, 461, 61 },
		{ 302, 462, 61 },
		{ 302, 463, 61 },
		{ 302, 464, 61 },
		{ 302, 465, 61 },
		{ 302, 466, 61 },
		{ 302, 467, 61 },
		{ 302, 468, 61 },
		{ 302, 469, 61 },
		{ 302, 470, 61 },
		{ 302, 471, 61 },
		{ 302, 472, 61 },
		{ 302, 473, 61 },
		{ 302, 474, 61 },
		{ 302, 475, 61 },
		{ 302, 476, 61 },
		{ 302, 477, 61 },
		{ 302, 478, 61 },
		{ 302, 479, 61 },
		{ 302, 480, 61 },
		{ 302, 481, 61 },
		{ 302, 482, 61 },
		{ 302, 483, 61 },
		{ 302, 484, 61 },
		{ 302, 485, 61 },
		{ 302, 486, 61 },
		{ 302, 487, 61 },
		{ 302, 488, 61 },
		{ 302, 489, 61 },
		{ 302, 490, 61 },
		{ 302, 491, 61 },
		{ 302, 492, 61 },
		{ 302, 493, 61 },
		{ 302, 494, 61 },
		{ 302, 495, 61 },
		{ 302, 496, 61 },
		{ 302, 497, 61 },
		{ 302, 498, 61 },
		{ 302, 499, 61 },
		{ 302, 500, 61 },
		{ 302, 501, 61 },
		{ 302, 502, 61 },
		{ 302, 503, 61 },
		{ 302, 504, 61 },
		{ 302, 505, 61 },
		{ 302, 506, 61 },
		{ 302, 507, 61 },
		{ 302, 508, 61 },
		{ 302, 509, 61 },
		{ 302, 510, 61 },
		{ 302, 511, 61 },
		{ 302, 512, 61 },
		{ 303, 442, 61 },
		{ 303, 443, 61 },
		{ 303, 444, 61 },
		{ 303, 445, 61 },
		{ 303, 446, 61 },
		{ 303, 447, 61 },
		{ 303, 448, 61 },
		{ 303, 449, 61 },
		{ 303, 450, 61 },
		{ 303, 451, 61 },
		{ 303, 452, 61 },
		{ 303, 453, 61 },
		{ 303, 454, 61 },
		{ 303, 455, 61 },
		{ 303, 456, 61 },
		{ 303, 457, 61 },
		{ 303, 458, 61 },
		{ 303, 459, 61 },
		{ 303, 460, 61 },
		{ 303, 461, 61 },
		{ 303, 462, 61 },
		{ 303, 463, 61 },
		{ 303, 464, 61 },
		{ 303, 465, 61 },
		{ 303, 466, 61 },
		{ 303, 467, 61 },
		{ 303, 468, 61 },
		{ 303, 469, 61 },
		{ 303, 470, 61 },
		{ 303, 471, 61 },
		{ 303, 472, 61 },
		{ 303, 473, 61 },
		{ 303, 474, 61 },
		{ 303, 475, 61 },
		{ 303, 476, 61 },
		{ 303, 477, 61 },
		{ 303, 478, 61 },
		{ 303, 479, 61 },
		{ 303, 480, 61 },
		{ 303, 481, 61 },
		{ 303, 482, 61 },
		{ 303, 483, 61 },
		{ 303, 484, 61 },
		{ 303, 485, 61 },
		{ 303, 486, 61 },
		{ 303, 487, 61 },
		{ 303, 488, 61 },
		{ 303, 489, 61 },
		{ 303, 490, 61 },
		{ 303, 491, 61 },
		{ 303, 492, 61 },
		{ 303, 493, 61 },
		{ 303, 494, 61 },
		{ 303, 495, 61 },
		{ 303, 496, 61 },
		{ 303, 497, 61 },
		{ 303, 498, 61 },
		{ 303, 499, 61 },
		{ 303, 500, 61 },
		{ 303, 501, 61 },
		{ 303, 502, 61 },
		{ 303, 503, 61 },
		{ 303, 504, 61 },
		{ 303, 505, 61 },
		{ 303, 506, 61 },
		{ 303, 507, 61 },
		{ 303, 508, 61 },
		{ 303, 509, 61 },
		{ 303, 510, 61 },
		{ 303, 511, 61 },
		{ 303, 512, 61 },
		{ 304, 446, 61 },
		{ 304, 447, 61 },
		{ 304, 448, 61 },
		{ 304, 449, 61 },
		{ 304, 450, 61 },
		{ 304, 451, 61 },
		{ 304, 452, 61 },
		{ 304, 453, 61 },
		{ 304, 454, 61 },
		{ 304, 455, 61 },
		{ 304, 456, 61 },
		{ 304, 457, 61 },
		{ 304, 458, 61 },
		{ 304, 459, 61 },
		{ 304, 460, 61 },
		{ 304, 461, 61 },
		{ 304, 462, 61 },
		{ 304, 463, 61 },
		{ 304, 464, 61 },
		{ 304, 465, 61 },
		{ 304, 466, 61 },
		{ 304, 467, 61 },
		{ 304, 468, 61 },
		{ 304, 469, 61 },
		{ 304, 470, 61 },
		{ 304, 471, 61 },
		{ 304, 472, 61 },
		{ 304, 473, 61 },
		{ 304, 474, 61 },
		{ 304, 475, 61 },
		{ 304, 476, 61 },
		{ 304, 477, 61 },
		{ 304, 478, 61 },
		{ 304, 479, 61 },
		{ 304, 480, 61 },
		{ 304, 481, 61 },
		{ 304, 482, 61 },
		{ 304, 483, 61 },
		{ 304, 484, 61 },
		{ 304, 485, 61 },
		{ 304, 486, 61 },
		{ 304, 487, 61 },
		{ 304, 488, 61 },
		{ 304, 489, 61 },
		{ 304, 490, 61 },
		{ 304, 491, 61 },
		{ 304, 492, 61 },
		{ 304, 493, 61 },
		{ 304, 494, 61 },
		{ 304, 495, 61 },
		{ 304, 496, 61 },
		{ 304, 497, 61 },
		{ 304, 498, 61 },
		{ 304, 499, 61 },
		{ 304, 500, 61 },
		{ 304, 501, 61 },
		{ 304, 502, 61 },
		{ 304, 503, 61 },
		{ 304, 504, 61 },
		{ 304, 505, 61 },
		{ 304, 506, 61 },
		{ 304, 507, 61 },
		{ 304, 508, 61 },
		{ 304, 509, 61 },
		{ 304, 510, 61 },
		{ 304, 511, 61 },
		{ 304, 512, 61 },
		{ 305, 450, 61 },
		{ 305, 451, 61 },
		{ 305, 452, 61 },
		{ 305, 453, 61 },
		{ 305, 454, 61 },
		{ 305, 455, 61 },
		{ 305, 456, 61 },
		{ 305, 457, 61 },
		{ 305, 458, 61 },
		{ 305, 459, 61 },
		{ 305, 460, 61 },
		{ 305, 461, 61 },
		{ 305, 462, 61 },
		{ 305, 463, 61 },
		{ 305, 464, 61 },
		{ 305, 465, 61 },
		{ 305, 466, 61 },
		{ 305, 467, 61 },
		{ 305, 468, 61 },
		{ 305, 469, 61 },
		{ 305, 470, 61 },
		{ 305, 471, 61 },
		{ 305, 472, 61 },
		{ 305, 473, 61 },
		{ 305, 474, 61 },
		{ 305, 475, 61 },
		{ 305, 476, 61 },
		{ 305, 477, 61 },
		{ 305, 478, 61 },
		{ 305, 479, 61 },
		{ 305, 480, 61 },
		{ 305, 481, 61 },
		{ 305, 482, 61 },
		{ 305, 483, 61 },
		{ 305, 484, 61 },
		{ 305, 485, 61 },
		{ 305, 486, 61 },
		{ 305, 487, 61 },
		{ 305, 488, 61 },
		{ 305, 489, 61 },
		{ 305, 490, 61 },
		{ 305, 491, 61 },
		{ 305, 492, 61 },
		{ 305, 493, 61 },
		{ 305, 494, 61 },
		{ 305, 495, 61 },
		{ 305, 496, 61 },
		{ 305, 497, 61 },
		{ 305, 498, 61 },
		{ 305, 499, 61 },
		{ 305, 500, 61 },
		{ 305, 501, 61 },
		{ 305, 502, 61 },
		{ 305, 503, 61 },
		{ 305, 504, 61 },
		{ 305, 505, 61 },
		{ 305, 506, 61 },
		{ 305, 507, 61 },
		{ 305, 508, 61 },
		{ 305, 509, 61 },
		{ 305, 510, 61 },
		{ 305, 511, 61 },
		{ 305, 512, 61 },
		{ 306, 454, 61 },
		{ 306, 455, 61 },
		{ 306, 456, 61 },
		{ 306, 457, 61 },
		{ 306, 458, 61 },
		{ 306, 459, 61 },
		{ 306, 460, 61 },
		{ 306, 461, 61 },
		{ 306, 462, 61 },
		{ 306, 463, 61 },
		{ 306, 464, 61 },
		{ 306, 465, 61 },
		{ 306, 466, 61 },
		{ 306, 467, 61 },
		{ 306, 468, 61 },
		{ 306, 469, 61 },
		{ 306, 470, 61 },
		{ 306, 471, 61 },
		{ 306, 472, 61 },
		{ 306, 473, 61 },
		{ 306, 474, 61 },
		{ 306, 475, 61 },
		{ 306, 476, 61 },
		{ 306, 477, 61 },
		{ 306, 478, 61 },
		{ 306, 479, 61 },
		{ 306, 480, 61 },
		{ 306, 481, 61 },
		{ 306, 482, 61 },
		{ 306, 483, 61 },
		{ 306, 484, 61 },
		{ 306, 485, 61 },
		{ 306, 486, 61 },
		{ 306, 487, 61 },
		{ 306, 488, 61 },
		{ 306, 489, 61 },
		{ 306, 490, 61 },
		{ 306, 491, 61 },
		{ 306, 492, 61 },
		{ 306, 493, 61 },
		{ 306, 494, 61 },
		{ 306, 495, 61 },
		{ 306, 496, 61 },
		{ 306, 497, 61 },
		{ 306, 498, 61 },
		{ 306, 499, 61 },
		{ 306, 500, 61 },
		{ 306, 501, 61 },
		{ 306, 502, 61 },
		{ 306, 503, 61 },
		{ 306, 504, 61 },
		{ 306, 505, 61 },
		{ 306, 506, 61 },
		{ 306, 507, 61 },
		{ 306, 508, 61 },
		{ 306, 509, 61 },
		{ 306, 510, 61 },
		{ 306, 511, 61 },
		{ 306, 512, 61 },
		{ 307, 458, 61 },
		{ 307, 459, 61 },
		{ 307, 460, 61 },
		{ 307, 461, 61 },
		{ 307, 462, 61 },
		{ 307, 463, 61 },
		{ 307, 464, 61 },
		{ 307, 465, 61 },
		{ 307, 466, 61 },
		{ 307, 467, 61 },
		{ 307, 468, 61 },
		{ 307, 469, 61 },
		{ 307, 470, 61 },
		{ 307, 471, 61 },
		{ 307, 472, 61 },
		{ 307, 473, 61 },
		{ 307, 474, 61 },
		{ 307, 475, 61 },
		{ 307, 476, 61 },
		{ 307, 477, 61 },
		{ 307, 478, 61 },
		{ 307, 479, 61 },
		{ 307, 480, 61 },
		{ 307, 481, 61 },
		{ 307, 482, 61 },
		{ 307, 483, 61 },
		{ 307, 484, 61 },
		{ 307, 485, 61 },
		{ 307, 486, 61 },
		{ 307, 487, 61 },
		{ 307, 488, 61 },
		{ 307, 489, 61 },
		{ 307, 490, 61 },
		{ 307, 491, 61 },
		{ 307, 492, 61 },
		{ 307, 493, 61 },
		{ 307, 494, 61 },
		{ 307, 495, 61 },
		{ 307, 496, 61 },
		{ 307, 497, 61 },
		{ 307, 498, 61 },
		{ 307, 499, 61 },
		{ 307, 500, 61 },
		{ 307, 501, 61 },
		{ 307, 502, 61 },
		{ 307, 503, 61 },
		{ 307, 504, 61 },
		{ 307, 505, 61 },
		{ 307, 506, 61 },
		{ 307, 507, 61 },
		{ 307, 508, 61 },
		{ 307, 509, 61 },
		{ 307, 510, 61 },
		{ 307, 511, 61 },
		{ 307, 512, 61 },
		{ 308, 462, 61 },
		{ 308, 463, 61 },
		{ 308, 464, 61 },
		{ 308, 465, 61 },
		{ 308, 466, 61 },
		{ 308, 467, 61 },
		{ 308, 468, 61 },
		{ 308, 469, 61 },
		{ 308, 470, 61 },
		{ 308, 471, 61 },
		{ 308, 472, 61 },
		{ 308, 473, 61 },
		{ 308, 474, 61 },
		{ 308, 475, 61 },
		{ 308, 476, 61 },
		{ 308, 477, 61 },
		{ 308, 478, 61 },
		{ 308, 479, 61 },
		{ 308, 480, 61 },
		{ 308, 481, 61 },
		{ 308, 482, 61 },
		{ 308, 483, 61 },
		{ 308, 484, 61 },
		{ 308, 485, 61 },
		{ 308, 486, 61 },
		{ 308, 487, 61 },
		{ 308, 488, 61 },
		{ 308, 489, 61 },
		{ 308, 490, 61 },
		{ 308, 491, 61 },
		{ 308, 492, 61 },
		{ 308, 493, 61 },
		{ 308, 494, 61 },
		{ 308, 495, 61 },
		{ 308, 496, 61 },
		{ 308, 497, 61 },
		{ 308, 498, 61 },
		{ 308, 499, 61 },
		{ 308, 500, 61 },
		{ 308, 501, 61 },
		{ 308, 502, 61 },
		{ 308, 503, 61 },
		{ 308, 504, 61 },
		{ 308, 505, 61 },
		{ 308, 506, 61 },
		{ 308, 507, 61 },
		{ 308, 508, 61 },
		{ 308, 509, 61 },
		{ 308, 510, 61 },
		{ 308, 511, 61 },
		{ 308, 512, 61 },
		{ 309, 466, 61 },
		{ 309, 467, 61 },
		{ 309, 468, 61 },
		{ 309, 469, 61 },
		{ 309, 470, 61 },
		{ 309, 471, 61 },
		{ 309, 472, 61 },
		{ 309, 473, 61 },
		{ 309, 474, 61 },
		{ 309, 475, 61 },
		{ 309, 476, 61 },
		{ 309, 477, 61 },
		{ 309, 478, 61 },
		{ 309, 479, 61 },
		{ 309, 480, 61 },
		{ 309, 481, 61 },
		{ 309, 482, 61 },
		{ 309, 483, 61 },
		{ 309, 484, 61 },
		{ 309, 485, 61 },
		{ 309, 486, 61 },
		{ 309, 487, 61 },
		{ 309, 488, 61 },
		{ 309, 489, 61 },
		{ 309, 490, 61 },
		{ 309, 491, 61 },
		{ 309, 492, 61 },
		{ 309, 493, 61 },
		{ 309, 494, 61 },
		{ 309, 495, 61 },
		{ 309, 496, 61 },
		{ 309, 497, 61 },
		{ 309, 498, 61 },
		{ 309, 499, 61 },
		{ 309, 500, 61 },
		{ 309, 501, 61 },
		{ 309, 502, 61 },
		{ 309, 503, 61 },
		{ 309, 504, 61 },
		{ 309, 505, 61 },
		{ 309, 506, 61 },
		{ 309, 507, 61 },
		{ 309, 508, 61 },
		{ 309, 509, 61 },
		{ 309, 510, 61 },
		{ 309, 511, 61 },
		{ 309, 512, 61 },
		{ 310, 470, 61 },
		{ 310, 471, 61 },
		{ 310, 472, 61 },
		{ 310, 473, 61 },
		{ 310, 474, 61 },
		{ 310, 475, 61 },
		{ 310, 476, 61 },
		{ 310, 477, 61 },
		{ 310, 478, 61 },
		{ 310, 479, 61 },
		{ 310, 480, 61 },
		{ 310, 481, 61 },
		{ 310, 482, 61 },
		{ 310, 483, 61 },
		{ 310, 484, 61 },
		{ 310, 485, 61 },
		{ 310, 486, 61 },
		{ 310, 487, 61 },
		{ 310, 488, 61 },
		{ 310, 489, 61 },
		{ 310, 490, 61 },
		{ 310, 491, 61 },
		{ 310, 492, 61 },
		{ 310, 493, 61 },
		{ 310, 494, 61 },
		{ 310, 495, 61 },
		{ 310, 496, 61 },
		{ 310, 497, 61 },
		{ 310, 498, 61 },
		{ 310, 499, 61 },
		{ 310, 500, 61 },
		{ 310, 501, 61 },
		{ 310, 502, 61 },
		{ 310, 503, 61 },
		{ 310, 504, 61 },
		{ 310, 505, 61 },
		{ 310, 506, 61 },
		{ 310, 507, 61 },
		{ 310, 508, 61 },
		{ 310, 509, 61 },
		{ 310, 510, 61 },
		{ 310, 511, 61 },
		{ 310, 512, 61 },
		{ 311, 474, 61 },
		{ 311, 475, 61 },
		{ 311, 476, 61 },
		{ 311, 477, 61 },
		{ 311, 478, 61 },
		{ 311, 479, 61 },
		{ 311, 480, 61 },
		{ 311, 481, 61 },
		{ 311, 482, 61 },
		{ 311, 483, 61 },
		{ 311, 484, 61 },
		{ 311, 485, 61 },
		{ 311, 486, 61 },
		{ 311, 487, 61 },
		{ 311, 488, 61 },
		{ 311, 489, 61 },
		{ 311, 490, 61 },
		{ 311, 491, 61 },
		{ 311, 492, 61 },
		{ 311, 493, 61 },
		{ 311, 494, 61 },
		{ 311, 495, 61 },
		{ 311, 496, 61 },
		{ 311, 497, 61 },
		{ 311, 498, 61 },
		{ 311, 499, 61 },
		{ 311, 500, 61 },
		{ 311, 501, 61 },
		{ 311, 502, 61 },
		{ 311, 503, 61 },
		{ 311, 504, 61 },
		{ 311, 505, 61 },
		{ 311, 506, 61 },
		{ 311, 507, 61 },
		{ 311, 508, 61 },
		{ 311, 509, 61 },
		{ 311, 510, 61 },
		{ 311, 511, 61 },
		{ 311, 512, 61 },
		{ 312, 478, 61 },
		{ 312, 479, 61 },
		{ 312, 480, 61 },
		{ 312, 481, 61 },
		{ 312, 482, 61 },
		{ 312, 483, 61 },
		{ 312, 484, 61 },
		{ 312, 485, 61 },
		{ 312, 486, 61 },
		{ 312, 487, 61 },
		{ 312, 488, 61 },
		{ 312, 489, 61 },
		{ 312, 490, 61 },
		{ 312, 491, 61 },
		{ 312, 492, 61 },
		{ 312, 493, 61 },
		{ 312, 494, 61 },
		{ 312, 495, 61 },
		{ 312, 496, 61 },
		{ 312, 497, 61 },
		{ 312, 498, 61 },
		{ 312, 499, 61 },
		{ 312, 500, 61 },
		{ 312, 501, 61 },
		{ 312, 502, 61 },
		{ 312, 503, 61 },
		{ 312, 504, 61 },
		{ 312, 505, 61 },
		{ 312, 506, 61 },
		{ 312, 507, 61 },
		{ 312, 508, 61 },
		{ 312, 509, 61 },
		{ 312, 510, 61 },
		{ 312, 511, 61 },
		{ 312, 512, 61 },
		{ 313, 482, 61 },
		{ 313, 483, 61 },
		{ 313, 484, 61 },
		{ 313, 485, 61 },
		{ 313, 486, 61 },
		{ 313, 487, 61 },
		{ 313, 488, 61 },
		{ 313, 489, 61 },
		{ 313, 490, 61 },
		{ 313, 491, 61 },
		{ 313, 492, 61 },
		{ 313, 493, 61 },
		{ 313, 494, 61 },
		{ 313, 495, 61 },
		{ 313, 496, 61 },
		{ 313, 497, 61 },
		{ 313, 498, 61 },
		{ 313, 499, 61 },
		{ 313, 500, 61 },
		{ 313, 501, 61 },
		{ 313, 502, 61 },
		{ 313, 503, 61 },
		{ 313, 504, 61 },
		{ 313, 505, 61 },
		{ 313, 506, 61 },
		{ 313, 507, 61 },
		{ 313, 508, 61 },
		{ 313, 509, 61 },
		{ 313, 510, 61 },
		{ 313, 511, 61 },
		{ 313, 512, 61 },
		{ 314, 486, 61 },
		{ 314, 487, 61 },
		{ 314, 488, 61 },
		{ 314, 489, 61 },
		{ 314, 490, 61 },
		{ 314, 491, 61 },
		{ 314, 492, 61 },
		{ 314, 493, 61 },
		{ 314, 494, 61 },
		{ 314, 495, 61 },
		{ 314, 496, 61 },
		{ 314, 497, 61 },
		{ 314, 498, 61 },
		{ 314, 499, 61 },
		{ 314, 500, 61 },
		{ 314, 501, 61 },
		{ 314, 502, 61 },
		{ 314, 503, 61 },
		{ 314, 504, 61 },
		{ 314, 505, 61 },
		{ 314, 506, 61 },
		{ 314, 507, 61 },
		{ 314, 508, 61 },
		{ 314, 509, 61 },
		{ 314, 510, 61 },
		{ 314, 511, 61 },
		{ 314, 512, 61 },
		{ 315, 490, 61 },
		{ 315, 491, 61 },
		{ 315, 492, 61 },
		{ 315, 493, 61 },
		{ 315, 494, 61 },
		{ 315, 495, 61 },
		{ 315, 496, 61 },
		{ 315, 497, 61 },
		{ 315, 498, 61 },
		{ 315, 499, 61 },
		{ 315, 500, 61 },
		{ 315, 501, 61 },
		{ 315, 502, 61 },
		{ 315, 503, 61 },
		{ 315, 504, 61 },
		{ 315, 505, 61 },
		{ 315, 506, 61 },
		{ 315, 507, 61 },
		{ 315, 508, 61 },
		{ 315, 509, 61 },
		{ 315, 510, 61 },
		{ 315, 511, 61 },
		{ 315, 512, 61 },
		{ 316, 494, 61 },
		{ 316, 495, 61 },
		{ 316, 496, 61 },
		{ 316, 497, 61 },
		{ 316, 498, 61 },
		{ 316, 499, 61 },
		{ 316, 500, 61 },
		{ 316, 501, 61 },
		{ 316, 502, 61 },
		{ 316, 503, 61 },
		{ 316, 504, 61 },
		{ 316, 505, 61 },
		{ 316, 506, 61 },
		{ 316, 507, 61 },
		{ 316, 508, 61 },
		{ 316, 509, 61 },
		{ 316, 510, 61 },
		{ 316, 511, 61 },
		{ 316, 512, 61 },
		{ 317, 498, 61 },
		{ 317, 499, 61 },
		{ 317, 500, 61 },
		{ 317, 501, 61 },
		{ 317, 502, 61 },
		{ 317, 503, 61 },
		{ 317, 504, 61 },
		{ 317, 505, 61 },
		{ 317, 506, 61 },
		{ 317, 507, 61 },
		{ 317, 508, 61 },
		{ 317, 509, 61 },
		{ 317, 510, 61 },
		{ 317, 511, 61 },
		{ 317, 512, 61 },
		{ 318, 502, 61 },
		{ 318, 503, 61 },
		{ 318, 504, 61 },
		{ 318, 505, 61 },
		{ 318, 506, 61 },
		{ 318, 507, 61 },
		{ 318, 508, 61 },
		{ 318, 509, 61 },
		{ 318, 510, 61 },
		{ 318, 511, 61 },
		{ 318, 512, 61 },
		{ 319, 506, 61 },
		{ 319, 507, 61 },
		{ 319, 508, 61 },
		{ 319, 509, 61 },
		{ 319, 510, 61 },
		{ 319, 511, 61 },
		{ 319, 512, 61 },
		{ 320, 510, 61 },
		{ 320, 511, 61 },
		{ 320, 512, 61 },
	};
}
